/*
 * LyMapBeahvior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package eu.w4.ae.app.behavior;

import java.io.*;
import java.net.*;

import leon.app.*;
import leon.app.behavior.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import eu.w4.ae.app.behaviorinterface.*;
import eu.w4.ae.control.*;


/**
 * Default application behavior for map controller interface.
 *
 * @author  Lyria S.A.
 * @version $Id: LyMapBehavior.java 41175 2012-10-15 08:02:27Z nhamon $
 * @see     eu.w4.ae.control.LyMapController
 * @see     eu.w4.ae.app.behaviorinterface.LyMapBehaviorInterface
 */
public class LyMapBehavior extends LyPanelBehavior
	implements LyMapBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyMapBehavior.java 41175 2012-10-15 08:02:27Z nhamon $";

	/**
	 * The default image height
	 */
    public static int       IMAGE_HEIGHT = 100;
    
	/**
	 * The default image node size.
	 */
    public static int       IMAGE_NODE_SIZE = 0;
	
	/** Indicates whether a Google map must be used. */
	protected boolean		_useGoogleMap;
    
    
	/**
	 * Constructor.
	 */
	public LyMapBehavior()
	{
	}
	
	/**
	 * Initializes information used within this behavior and is called immediately after its
	 * construction.<br />
	 * This implementation sets the <code>_useGoogleMap</code> attribute.
	 *
	 * @param controller The controller that requested the creation of this behavior.
	 */
	public void init(LyController controller)
	{
		super.init(controller);

		LyViewManager		viewManager = controller.getViewManager();
		LyAreaController	areaController = (LyAreaController)controller;
		String				style = (String)areaController.getParameter("_style", true);

		// Use Google map if view manager has the right property, if parameter is set and if there is
		// an available internet connection
		_useGoogleMap = (viewManager.hasProperty("DISPLAY_GOOGLE_MAP")) && (style != null);

		if (_useGoogleMap)
		{
			try
			{
				URL					url = new URL("http://maps.googleapis.com/maps/api/js?sensor=false&callback=importgoMapScript");
				HttpURLConnection	urlConnection = (HttpURLConnection)url.openConnection();
				
				urlConnection.setConnectTimeout(3000);
				urlConnection.connect();
			}
			catch (Exception e)
			{
				_useGoogleMap = false;
			}
		}
	}

	/**
	 * Gets the application objects list (which are {@link leon.data.LyObject LyObject} instances),
	 * that must be represented as <u>node</u> objects.<br>
	 * <b>IMPORTANT:</b> if the returned list is a dynamic workspace, the mapController will notify
	 * for node updates in {@link #updateNode(LyObject, LyMapNode, LyMapController)}. Association
	 * between application object and graphical object is made in the {@link #buildNode(LyObject,
	 * LyMapController)} method.<br />
	 * <i>The default implementation</i> returns all objects of the controller classInfo depending
	 * on the current context of the controller.
	 *
	 * @param  controller current LyMapController instance
	 * @return the list of node objects, or null if the map does not have nodes
	 * @see    #buildNode(LyObject, LyMapController)
	 * @see    #updateNode(LyObject, LyMapNode, LyMapController)
	 */
	public LyObjectList getNodeObjects(LyMapController controller)
	{
		LySession   session = controller.getSession();
		LyWorkSpace ws = null;
		LyWorkSpace context = controller.getContext();
		LyClassInfo classInfo = controller.getClassInfo();   
        LyFilter    filter = (controller.isFilterActive() ? controller.getFilter() : null);

		if (context != null)
		{
			// if classInfo is not null, building a mixed workspace
			if (classInfo != null)
				// Return a new dynamic workspace based on context
				ws = (LyWorkSpace)context.getList(session, classInfo, filter, true, false);
		}
		else
		{
			// Return all the objects of this class info
			if (classInfo != null)
				// TODO : optimize the loaded fieldinfos.
				ws = new LyClassWorkSpace(session, classInfo, filter, true);
		}

		if (ws != null)
			ws.setSortByName(true);
		return ws;
	}

	/**
	 * Gets the list of objects of the application (which are {@link leon.data.LyObject LyObject}
	 * instances), that must be represented as <u>link</u> objects.<br />
	 * <b>IMPORTANT:</b> if the returned list is a dynamic workspace, the mapController will notify
	 * for link updates in {@link #updateLink(LyObject, LyMapLink, LyMapController)}. Association
	 * between application object and graphical object is made in the {@link #buildLink(LyObject,
	 * LyMapController)} method.<br />
	 * <i>The default implementation</i> returns null. By default, no link object is created.
	 *
	 * @param  controller current LyMapController instance
	 * @return the list of links objects, or null if the map does not have links
	 * @see    #buildLink(LyObject, LyMapController)
	 * @see    #updateLink(LyObject, LyMapLink, LyMapController)
	 */
	public LyObjectList getLinkObjects(LyMapController controller)
	{
		return null;
	}

	/**
	 * Gets the list of objects of the application (which are {@link leon.data.LyObject LyObject}
	 * instances), that must be represented as <u>region</u> objects.<br />
	 * <b>IMPORTANT:</b> if the returned list is a dynamic workspace the mapController will notify
	 * for region updates in {@link #updateRegion(LyObject, LyMapRegion, LyMapController)}.
	 * Association between application object and graphical object is made in {@link
	 * #buildRegion(LyObject, LyMapController)} method.<br />
	 * <i>The default implementation</i> returns null. By Default, no region objects is created.
	 *
	 * @param  controller current LyMapController instance
	 * @return the list of region objects, or null if the map does not have regions
	 * @see    #buildRegion(LyObject, LyMapController)
	 * @see    #updateRegion(LyObject, LyMapRegion, LyMapController)
	 */
	public LyObjectList getRegionObjects(LyMapController controller)
	{
		return null;
	}

	/**
	 * Builds a node object instance from the given object.<br />
	 * <u>Note:</u> Node status (colors, borders, cross...) should be set in {@link
	 * #updateNode(LyObject, LyMapNode, LyMapController)}.<br />
	 * <i>The default implementation</i> creates a {@link LyMapNode map node} with the given node id
	 * and sets its style as an {@link LyMapNode.Style#IMAGE image}.
	 *
	 * @param  nodeObject application object, usually a LyObject instance but not mandatory
	 * @param  controller current LyMapController instance
	 * @return the built node view object.
	 */
	public LyMapNode buildNode(LyObject nodeObject, LyMapController controller)
	{
		if (nodeObject == null)
			return null;

		LyMapNode node = new LyMapNode();

		node.setId(nodeObject.getId());
		node.setStyle(LyMapNode.Style.IMAGE);

		return (node);
	}

	/**
	 * Builds a link object instance from the application object.<br />
	 * <u>Note:</u> Link status (colors, icons, ...) should be set in {@link #updateLink(LyObject,
	 * LyMapLink, LyMapController)} .<br />
	 * <i>The default implementation</i> does nothing and returns null.
	 *
	 * @param  linkObject application object, usually LyObject instance but not mandatory
	 * @param  controller current LyMapController instance
	 * @return the built link view object
	 */
	public LyMapLink buildLink(LyObject linkObject, LyMapController controller)
	{
		return null;
	}

	/**
	 * Builds a region object instance from the application object.<br />
	 * <u>Note:</u> Region status (colors, icons, ...) should be set in {@link
	 * #updateRegion(LyObject, LyMapRegion, LyMapController)} .<br />
	 * <i>The default implementation</i> does nothing and returns null.
	 *
	 * @param  regionObject application object, usually LyObject instance but not mandatory
	 * @param  controller   current LyMapController instance
	 * @return the built region view object
	 */
	public LyMapRegion buildRegion(LyObject regionObject, LyMapController controller)
	{
		return null;
	}

	/**
	 * Updates a the given mapNode object status (colors, label, border, style, ...) with the given
	 * nodeObject properties. The controller invokes this method when the given nodeObject is
	 * modified.<br />
	 * <i>The default implementation</i> updates the label, the image associated to the node and the
	 * hidden property.
	 *
	 * @param nodeObject application object (reference)
	 * @param mapNode    view objects (need update)
	 * @param controller current LyMapController instance
	 */
	public void updateNode(LyObject nodeObject, LyMapNode mapNode, LyMapController controller)
	{
		if ((nodeObject == null) || (mapNode == null))
			return;

		// Set the label with the object name and the image with
		// the class icon
		mapNode.setLabel(controller.getName(nodeObject));
		mapNode.setImage(getImage(controller, nodeObject));
		mapNode.setHidden(nodeObject.isHidden());
		
		// Geographical information
		LyFieldInfo		longitudeFieldInfo = (LyFieldInfo)controller.getParameter("_longitudeFieldInfo", true);
		LyFieldInfo		latitudeFieldInfo = (LyFieldInfo)controller.getParameter("_latitudeFieldInfo", true);
		LyFieldInfo		geolocationFieldInfo = (LyFieldInfo)controller.getParameter("_geolocationFieldInfo", true);
		
		if ((latitudeFieldInfo != null) && (longitudeFieldInfo != null))
		{
			Object	longitudeValue = nodeObject.getValue(longitudeFieldInfo);
			Object	latitudeValue = nodeObject.getValue(latitudeFieldInfo);
			float	longitude = (longitudeValue != null ? LyTools.floatFromString(longitudeValue.toString()) : 0);
			float	latitude = (latitudeValue != null ? LyTools.floatFromString(latitudeValue.toString()) : 0);
			
			mapNode.setX(longitude);
			mapNode.setY(latitude);
			mapNode.setMovable(longitudeFieldInfo.hasMark(LyFieldInfo.MARK_SET)
				&& latitudeFieldInfo.hasMark(LyFieldInfo.MARK_SET));
		}
		else
			mapNode.setMovable(false);
		
		if (geolocationFieldInfo != null)
		{
			Object	geolocationValue = nodeObject.getValue(geolocationFieldInfo);

			if (geolocationValue != null)
			{
				String	geolocation = geolocationFieldInfo.toName(geolocationValue);
	
				mapNode.addData("GEOLOCATION_VALUE", geolocation);
			}
		}
	}

    /**
     * Gets image corresponding to an object.
     * 
     * @param map the map controller
     * @param object the object we need to represent
     * @return Image that represents the given object.
     */
    public String getImage(LyMapController map, LyObject object)
    {
        // Try to find a file fieldInfo with image type
        LyClassInfo     classInfo = object.getClassInfo();
        LyFieldInfo     fieldInfo;
        int             i, n = classInfo.getFieldInfoCount();
        LyFileValue     value;
        String          image = null;
        LyEnvironment   env = map.getEnvironment();
        
        for (i = 0; i < n; i++)
        {
            fieldInfo = classInfo.getFieldInfo(i);
            if ((fieldInfo instanceof LyFileFieldInfo)
                && (((LyFileFieldInfo)fieldInfo).getType() == LyFileFieldInfo.Type.IMAGE))
            {
                value = object.getFileValue((LyFileFieldInfo)fieldInfo);
                if (!value.isNull())
                    image = ((LyFileFieldInfo)fieldInfo).getUrl(env, value.getString());
                break;
            }
        }
        
        if (image != null)
        {
            if (image.startsWith("file://"))
                image = image.substring(7);
            
            // adding the size information
            if (IMAGE_HEIGHT > 0)
            {
                LyString string = env.getString();

                image += string.VISIBLE_SEP3+ IMAGE_HEIGHT + string.VISIBLE_SEP3_BIS;
            }
        }
        else
        {
            // Retrieve the image corresponding to the object classInfo  
           image = map.getImageFromPrefix(classInfo.getFilePrefix(),
                    LyController.DEFAULT_IMAGE_FORMAT, IMAGE_NODE_SIZE);
        }
        
        return image;
    }
    
	/**
	 * Updates a the given mapLink object status (colors, label, border, style, ...) with the given
	 * linkObject properties. The controller invokes this method when the given linkObject is
	 * modified.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param linkObject application object (reference)
	 * @param mapLink    view objects (need update)
	 * @param controller current LyMapController instance
	 */
	public void updateLink(LyObject linkObject, LyMapLink mapLink, LyMapController controller)
	{
	}

	/**
	 * Updates a the given mapRegion object status (colors, label, border, style, ...) with the
	 * given regionObject properties. The controller invokes this method when the given regionObject
	 * is modified.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param regionObject application object (reference)
	 * @param mapRegion    view objects (need update)
	 * @param controller   current LyMapController instance
	 */
	public void updateRegion(LyObject regionObject, LyMapRegion mapRegion,
		LyMapController controller)
	{
	}

	/**
	 * This method is invoked by the controller when the given mapNode is selected on the map.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param nodeObject application object (reference)
	 * @param mapNode    view objects
	 * @param controller current LyMapController instance
	 */
	public void selectNode(LyObject nodeObject, LyMapNode mapNode, LyMapController controller)
	{
	}

	/**
	 * This method is invoked by the controller when the given mapNode is unselected on the map.
	 * <br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param nodeObject application object (reference)
	 * @param mapNode    view objects
	 * @param controller current LyMapController instance
	 */
	public void unselectNode(LyObject nodeObject, LyMapNode mapNode, LyMapController controller)
	{
	}

	/**
	 * This method is invoked by the controller when all nodes are unselected on the map.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param controller current LyMapController instance
	 */
	public void unselectNodes(LyMapController controller)
	{
	}

	/**
	 * This method is invoked by the controller when selected nodes on the map have been moved.
	 * <br />
	 * <i>The default implementation</i> does nothing, but children classes may set coordinates in
	 * the appObject for instance.
	 *
	 * @param controller current mapController instance
	 */
	public void selectionMoved(LyMapController controller)
	{
        LyMap       map = controller.getMapComponent();
        LyLayout    layout = map.getLayout();
        
        if ((layout != null)
            && (layout.getStyle() == LyLayout.Style.GEOGRAPHIC))
        {
            String  mapFile = (String)controller.getParameter("_mapFile", true);
            
            if (mapFile != null)
            {
	            try
	            {
	                PrintWriter pw = new PrintWriter(mapFile);
	                
	                pw.println("<?xml version=\"1.0\" encoding=\"" + LyString.DEFAULT_CHARSET + "\"?>");
	                pw.println("<!DOCTYPE MAP SYSTEM '/leon/conf/dtd/views.dtd'>");
	                map.writeXml(pw);
	                pw.close();
	            }
	            catch (Exception e)
	            {
	                LyEnvironment   environment = controller.getEnvironment();
	                
	                environment.writeException(e);
	            }
            }
        }

        // Geographical information
		LyFieldInfo		longitudeFieldInfo = (LyFieldInfo)controller.getParameter("_longitudeFieldInfo", true);
		LyFieldInfo		latitudeFieldInfo = (LyFieldInfo)controller.getParameter("_latitudeFieldInfo", true);
		
		if ((latitudeFieldInfo != null) && (longitudeFieldInfo != null))
		{
			LyObjectList  selection = controller.getSelection();
			LyObject	  nodeObject;
			LyMapNode     node;
			int			  i, n = (selection == null ? 0 : selection.getSize());

			controller.getMapComponent().setIdle(true);

			for (i = 0; i < n; i++)
			{
				nodeObject = selection.getObject(i);
				node = (LyMapNode)controller.getObjectAssoc(nodeObject);

				if (node != null)
				{
					nodeObject.setValue(longitudeFieldInfo, node.getX());
					nodeObject.setValue(latitudeFieldInfo, node.getY());
					nodeObject.set();
				}
			}
			
			controller.getMapComponent().setIdle(false);
		}
	}

	/**
	 * This method is invoked by the controller to invoke the default action when an object is
	 * activated on the map (e.g. by double click with graphical modality).<br />
	 * <i>The default implementation</i> does nothing and return false.
	 *
	 * @param      nodeObject application object (may be null)
	 * @param      mapObject  view objects (selected for action)
	 * @param      controller current LyMapController instance
	 * @return     true if the action was handled, false otherwise
	 * @deprecated replaced by {@link #invokeDefaultAction(LyObject, LyComponent, int, int,
	 *             LyMapController)}
	 */
	public boolean invokeDefaultAction(LyObject nodeObject, LyComponent mapObject,
		LyMapController controller)
	{
		return false;
	}

	/**
	 * This method is invoked by the controller to invoke the default action when an object is
	 * activated on the map (e.g. by double click with graphical modality).<br />
	 * <i>The default implementation</i> does nothing and return false.
	 *
	 * @param  nodeObject application object (may be null)
	 * @param  mapObject  view objects (selected for action)
	 * @param  x          the x position of the event on the map
	 * @param  y          the y position of the event on the map
	 * @param  controller current LyMapController instance
	 * @return true if the action was handled, false otherwise
	 */
	public boolean invokeDefaultAction(LyObject nodeObject, LyComponent mapObject, int x, int y,
		LyMapController controller)
	{
		return false;
	}

	/**
	 * This method is invoked when the view is ready to be shown. May be overriden to add extra
	 * components.<br />
	 * <i>The default implementation</i> sets the map layout to a matrix view and disables the
	 * viewer.
	 *
	 * @param controller current mapController instance
	 */
	public void completeMap(LyMapController controller)
	{
		LyMap map = controller.getMapComponent();

		// Set the layout
		if ((map != null) && (map.getLayout() == null))
		{
			// Create a default layout.
			LyLayout layout = new LyLayout();

			layout.setStyle(LyLayout.Style.MATRIX);
			layout.setDeltaNode(20);
			layout.setDeltaBorder(10);
			layout.setHAlign(LyLayout.HAlign.CENTER);
			layout.setVAlign(LyLayout.VAlign.TOP);
			layout.setFill(LyLayout.Fill.HORIZONTAL);

			map.setLayout(layout);
			map.setViewer(_useGoogleMap);
			map.setTranslation(_useGoogleMap);
			map.setZoom(_useGoogleMap);
			
			if (_useGoogleMap)
			{
				String		style = (String)controller.getParameter("_style", true);

				if ("HYBRID".equals(style))
					map.setStyle(LyMap.Style.HYBRID);
				else if ("ROADMAP".equals(style))
					map.setStyle(LyMap.Style.ROADMAP);
				else if ("SATELLITE".equals(style))
					map.setStyle(LyMap.Style.SATELLITE);
				else
					map.setStyle(LyMap.Style.TERRAIN);
			}
			
			map.setZoomFactor(_useGoogleMap ? 5 : 1);

			String		longitude = (String)controller.getParameter("_longitude", true);
			String		latitude = (String)controller.getParameter("_latitude", true);
			
			if ((longitude != null) && (latitude != null))
			{
				map.setLongitude(longitude);
				map.setLatitude(latitude);
			}
		}
	}

	/**
	 * This method is invoked when the content of the map has been modified.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param controller current mapController instance
	 * @param map        current map instance
	 * @param content    new map content instance
	 */
	public void setMapContent(LyMapController controller, LyMap map, LyMapContent content)
	{
	}

	/**
	 * Frees all resources.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param controller current mapController instance
	 */
	public void free(LyMapController controller)
	{
	}

	/**
	 * Method invoked at the beginning of the given object drag.<br />
	 * <i>The default implementation</i> keep in the controller session memory that the object is
	 * dragged and log a trace.
	 *
	 * @param controller    current LyMapController instance
	 * @param draggedObject dragged object
	 */
	public void dragStart(LyMapController controller, LyObject draggedObject)
	{
		LyEnvironment environment = controller.getEnvironment();
		LySession     session = controller.getSession();

		session.addValue(LySession.DRAGGED_OBJECT, draggedObject);
		environment.logTrace("Drag is starting with object: " + draggedObject);
	}

	/**
	 * Method called during a drag-and-drop, when the given object is dragged over the map
	 * component.<br />
	 * <i>The default implementation</i> only logs a trace.
	 *
	 * @param controller    current LyMapController instance
	 * @param draggedObject the dragged object
	 * @param x             The x position of the drag event
	 * @param y             The y position of the drag event
	 */
	public void dragOver(LyMapController controller, LyObject draggedObject, int x, int y)
	{
		LyEnvironment environment = controller.getEnvironment();

		environment.logTrace("Drag is over the map with object: " + draggedObject + " at position "
			+ x + ", " + y);
	}

	/**
	 * Method called when the given dragged object is dropped on the given dropped object.<br />
	 * <i>The default implementation</i> only logs a trace.
	 *
	 * @param controller    current LyMapController instance
	 * @param droppedObject object on where the dragged object is dropped
	 * @param draggedObject the dragged object
	 * @param x             The x position of drop event
	 * @param y             The y position of drop event
	 */
	public void drop(LyMapController controller, LyObject droppedObject, LyObject draggedObject,
		int x, int y)
	{
		LyEnvironment environment = controller.getEnvironment();

		environment.logTrace("Object: " + draggedObject + " is dropped on object " + droppedObject
			+ " at position " + x + ", " + y);
	}

	/**
	 * Method called at the end of a drag-and-drop.<br />
	 * <i>The default implementation</i> only logs a trace about the given drag-and-drop type.
	 *
	 * @param controller current LyMapController instance
	 * @param type       drag-and-drop type ({@link org.eclipse.swt.dnd.DND#DROP_COPY} or {@link
	 *                   org.eclipse.swt.dnd.DND#DROP_MOVE})
	 */
	public void dragFinished(LyMapController controller, int type)
	{
		// String dropType;
		// switch (type)
		// {
		// case DND.DROP_COPY:
		// dropType = "Copy (Ctrl + Click)";
		// break;
		// case DND.DROP_MOVE:
		// dropType = "Move (Shift + Click) or default";
		// break;
		// case DND.DROP_LINK:
		// dropType = "Link (Ctrl + Shift + Click)";
		// break;
		// default:
		// dropType = "None";
		// break;
		// }

		LyEnvironment environment = controller.getEnvironment();

		environment.logTrace("The drag-and-drop type is: " + type);
	}
	
	/**
	 * Method called when a widget action is performed (zoom, translation...).
	 * 
	 * @param controller current LyMapController instance
	 * @param source     identificator of the action
	 * @param value      value of the action
	 */
	public void widgetAction(LyMapController controller, String source, String value)
	{
		LyMap	map = controller.getMapComponent();
		
		if ("displayViewer".equals(source) && (value != null))
		{
			if (LyTools.booleanFromString(value))
				map.setViewerStatus(LyMap.ViewerStatus.SHOW);
			else
				map.setViewerStatus(LyMap.ViewerStatus.HIDE);
		}
		else if ("displayTranslation".equals(source) && (value != null))
		{
			if (LyTools.booleanFromString(value))
				map.setTranslationStatus(LyMap.TranslationStatus.SHOW);
			else
				map.setTranslationStatus(LyMap.TranslationStatus.HIDE);
		}
		else if ("displayZoom".equals(source) && (value != null))
		{
			if (LyTools.booleanFromString(value))
				map.setZoomStatus(LyMap.ZoomStatus.SHOW);
			else
				map.setZoomStatus(LyMap.ZoomStatus.HIDE);
		}
		else if ("moveViewer".equals(source) && (value != null))
		{
			// Remove 'px' strings
			value = LyString.replace(value, "px", "");
			
			String[]	values = LyString.stringToList(value, '_');
			
			if ((values != null) && (values.length == 2))
			{
				map.setViewerX(LyTools.intFromString(values[0]));
				map.setViewerY(LyTools.intFromString(values[1]));
			}
		}
	}
} // end class LyMapBehavior
