/*
 * LyTreemapBeahvior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package eu.w4.ae.app.behavior;

import java.util.*;

import leon.app.behavior.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import eu.w4.ae.app.behaviorinterface.*;
import eu.w4.ae.control.*;


/**
 * Default application behavior for treemap controller interface.
 *
 * @author  Lyria S.A.
 * @version $Id: LyTreemapBehavior.java 28997 2010-12-23 10:17:15Z bmichel $
 * @see     leon.control.LyTreemapController
 * @see     leon.app.behaviorinterface.LyTreemapBehaviorInterface
 */
public class LyTreemapBehavior extends LyActionBehavior
	implements LyTreemapBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyTreemapBehavior.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** Cache of colors used for threshold colors. */
	protected LyLastValues     _lastColors;

	/**
	 * Constructor. Sets the {@link #_lastColors cache} to null.
	 */
	public LyTreemapBehavior()
	{
		_lastColors = null;
	}

	/**
	 * Gets the list of fields that may be used by the treemap hierarchy.<br />
	 * <i>The default implementation</i> gets the controller class info fields with the hierarchy
	 * field marks from the metamodel. If this list is null or empty, gets fields of the class with
	 * marks hidden and secret or if there is not any hidden field, build a new {@link
	 * LyInfoList list}.
	 *
	 * @param  controller the treemap controller
	 * @return the list of fields that may be used by the treemap hierarchy
	 */
	public LyInfoList<LyFieldInfo> getHierarchyFieldList(LyTreemapController controller)
	{
		BitSet				    marks = controller.getParameterMarks("_hierarchyFieldMarks",
				LyBitSet.Type.F_MARKS);
		LyClassInfo			    classInfo = controller.getClassInfo();
		LyInfoList<LyFieldInfo> list = classInfo.getFieldInfos(marks);

		if ((list == null) || (list.getSize() < 1))
		{
			list = classInfo.getFieldInfos();
			list = list.getInfos(LyFieldInfo.MARK_HIDDEN, true);

			if (list != null)
				list = list.getInfos(LyFieldInfo.MARK_SECRET, true);
			if (list == null)
				list = new LyInfoList<LyFieldInfo>();
		}

		return list;
	}

	/**
	 * Gets the initial list of fields used to compute the treemap hierarchy, as defined in the
	 * metamodel.<br />
	 * <i>The default implementation</i> gets all fields "_hierarchyFieldInfo" defined in the
	 * metamodel.
	 *
	 * @param  controller the treemap controller
	 * @return the list of fields used to compute the treemap hierarchy.
	 */
	public LyInfoList<LyFieldInfo> getHierarchyFields(LyTreemapController controller)
	{
		LyInfoList<LyFieldInfo> hierarchyFields = new LyInfoList<LyFieldInfo>();
		int					    i = 1;
		LyFieldInfo			    fieldInfo = (LyFieldInfo)controller.getParameter(
				"_hierarchyFieldInfo" + i, true);

		while (fieldInfo != null)
		{
			fieldInfo = (LyFieldInfo)controller.getParameter("_hierarchyFieldInfo" + i, true);
			hierarchyFields.addInfo(fieldInfo);
			i++;
		}

		return hierarchyFields;
	}

	/**
	 * Gets the initial number of displayed levels of the treemap hierarchy, as defined in the
	 * metamodel.<br />
	 * <i>The default implementation</i> gets the "_hierarchyLevels" parameter from the metamodel.
	 * If the retrieve level is lower than 1 (means no level) then the {@link
	 * LyTreemapController#getHierarchyFields() number of fields used to compute the hierarchy} plus
	 * one is returned.
	 *
	 * @param  controller the treemap controller
	 * @return the number of displayed levels of the treemap hierarchy.
	 */
	public int getHierarchyLevels(LyTreemapController controller)
	{
		String levelParam = (String)controller.getParameter("_hierarchyLevels", true);
		int    level = LyTools.intFromString(levelParam);

		// If level is null, default level is the number of fields used to
		// compute the hierarchy plus one.
		if (level < 1)
		{
			LyInfoList<LyFieldInfo> list = controller.getHierarchyFields();

			level = list.getSize() + 1;
		}

		return level;
	}

	/**
	 * Gets the list of fields that may be used to compute the size of treemap nodes.<br />
	 * <i>The default implementation</i> only calls {@link #getNumberFieldInfos(LyTreemapController)
	 * getNumberFieldInfos(controller)} and reutrns its result.
	 *
	 * @param  controller the treemap controller
	 * @return the list of fields that may be used to compute the size of treemap nodes.
	 */
	public LyInfoList<LyFieldInfo> getSizeFieldList(LyTreemapController controller)
	{
		return getNumberFieldInfos(controller);
	}

	/**
	 * Gets the initial field used to compute the size of treemap nodes, as defined in the
	 * metamodel.<br />
	 * <i>The default implementation</i> returns the "_sizeFieldInfo" parameter from the metamodel.
	 *
	 * @param  controller the treemap controller
	 * @return the field used to compute the size of treemap nodes.
	 */
	public LyFieldInfo getSizeField(LyTreemapController controller)
	{
		LyFieldInfo fieldInfo = (LyFieldInfo)controller.getParameter("_sizeFieldInfo", true);

		return fieldInfo;
	}

	/**
	 * Gets the list of fields that may be used to compute the color of treemap nodes.<br />
	 * <i>The default implementation</i> only calls {@link #getNumberFieldInfos(LyTreemapController)
	 * getNumberFieldInfos(controller)} and reutrns its result.
	 *
	 * @param  controller the treemap controller
	 * @return the list of fields that may be used to compute the color of treemap nodes.
	 */
	public LyInfoList<LyFieldInfo> getColorFieldList(LyTreemapController controller)
	{
		return getNumberFieldInfos(controller);
	}

	/**
	 * Gets the initial field used to compute the color of treemap nodes, as defined in the
	 * metamodel.<br />
	 * <i>The default implementation</i> returns the "_colorFieldInfo" parameter from the metamodel.
	 *
	 * @param  controller the treemap controller
	 * @return the field used to compute the color of treemap nodes.
	 */
	public LyFieldInfo getColorField(LyTreemapController controller)
	{
		LyFieldInfo fieldInfo = (LyFieldInfo)controller.getParameter("_colorFieldInfo", true);

		return fieldInfo;
	}

	/**
	 * Gets the initial color used for the minimum value of treemap nodes.<br />
	 * <i>The default implementation</i> gets the LY_TREEMAP_MIN_COLOR_CODE resource value.
	 *
	 * @param  controller the treemap controller
	 * @return the color used for the minimum value of treemap nodes.
	 */
	public String getMinColor(LyTreemapController controller)
	{
		LyEnvironment env = controller.getEnvironment();
		String		  color = env.translate("LY_TREEMAP_MIN_COLOR_CODE");

		return color;
	}

	/**
	 * Gets the initial color used for the maximum value of treemap nodes.<br />
	 * <i>The default implementation</i> gets the LY_TREEMAP_MAX_COLOR_CODE resource value.
	 *
	 * @param  controller the treemap controller
	 * @return the color used for the maximum value of treemap nodes.
	 */
	public String getMaxColor(LyTreemapController controller)
	{
		LyEnvironment env = controller.getEnvironment();
		String		  color = env.translate("LY_TREEMAP_MAX_COLOR_CODE");

		return color;
	}

	/**
	 * Gets the list of supported aggregation algorithms.<br />
	 * <i>The default implementation</i> builds the following array of Strings:
	 *
	 * <table border='1'>
	 *   <tr>
	 *     <td> {@link LyTreemapController#MINIMUM_AGGREGATION}</td>
	 *     <td>LY_TREEMAP_MINIMUM_AGGREGATION key</td>
	 *   </tr>
	 *   <tr>
	 *     <td> {@link LyTreemapController#MAXIMUM_AGGREGATION}</td>
	 *     <td>LY_TREEMAP_MAXIMUM_AGGREGATION key</td>
	 *   </tr>
	 *   <tr>
	 *     <td> {@link LyTreemapController#AVERAGE_AGGREGATION}</td>
	 *     <td>LY_TREEMAP_AVERAGE_AGGREGATION key</td>
	 *   </tr>
	 *   <tr>
	 *     <td> {@link LyTreemapController#BALANCED_AGGREGATION}</td>
	 *     <td>LY_TREEMAP_BALANCED_AGGREGATION key</td>
	 *   </tr>
	 * </table>
	 *
	 * @param  controller the treemap controller
	 * @return Array of couple of strings that represents supported algorithms. Each couple of
	 *         strings is the algorithm code for the first string, and the displayable name of the
	 *         algorithm or a key to the dictionnary of this string for the second.
	 */
	public String[][] getAggregationTypes(LyTreemapController controller)
	{
		String[][] result = new String[4][2];

		result[0] = new String[]
			{
				LyTreemapController.MINIMUM_AGGREGATION,
				"LY_TREEMAP_MINIMUM_AGGREGATION"
			};
		result[1] = new String[]
			{
				LyTreemapController.MAXIMUM_AGGREGATION,
				"LY_TREEMAP_MAXIMUM_AGGREGATION"
			};
		result[2] = new String[]
			{
				LyTreemapController.AVERAGE_AGGREGATION,
				"LY_TREEMAP_AVERAGE_AGGREGATION"
			};
		result[3] = new String[]
			{
				LyTreemapController.BALANCED_AGGREGATION,
				"LY_TREEMAP_BALANCED_AGGREGATION"
			};

		return result;
	}

	/**
	 * Gets the code of the initial algorithm used for aggregation by the treemap.<br />
	 * <i>The default implementation</i> gets the "_aggregationType" parameter form the metamodel
	 * and depending on its value return the code:
	 *
	 * <table border='1'>
	 *   <tr>
	 *     <th>Parameter type</th>
	 *     <th>Returned algorithm code</th>
	 *   </tr>
	 *   <tr>
	 *     <td>MINIMUM</td>
	 *     <td> {@link LyTreemapController#MINIMUM_AGGREGATION}</td>
	 *   </tr>
	 *   <tr>
	 *     <td>MAXIMUM</td>
	 *     <td> {@link LyTreemapController#MAXIMUM_AGGREGATION}</td>
	 *   </tr>
	 *   <tr>
	 *     <td>AVERAGE</td>
	 *     <td> {@link LyTreemapController#AVERAGE_AGGREGATION}</td>
	 *   </tr>
	 *   <tr>
	 *     <td>Other</td>
	 *     <td> {@link LyTreemapController#BALANCED_AGGREGATION}</td>
	 *   </tr>
	 * </table>
	 *
	 * @param  controller the treemap controller
	 * @return the code of the algorithm used for aggregation by the treemap.
	 */
	public String getAggregationType(LyTreemapController controller)
	{
		String aggregation = (String)controller.getParameter("_aggregationType", true);

		if ("MINIMUM".equals(aggregation))
			aggregation = LyTreemapController.MINIMUM_AGGREGATION;
		else if ("MAXIMUM".equals(aggregation))
			aggregation = LyTreemapController.MAXIMUM_AGGREGATION;
		else if ("AVERAGE".equals(aggregation))
			aggregation = LyTreemapController.AVERAGE_AGGREGATION;
		else
			aggregation = LyTreemapController.BALANCED_AGGREGATION;

		return aggregation;
	}

	/**
	 * Gets the list of fields that may be used by the aggregation algorithms.<br />
	 * <i>The default implementation</i> only calls {@link #getNumberFieldInfos(LyTreemapController)
	 * getNumberFieldInfos(controller)} and reutrns its result.
	 *
	 * @param  controller the treemap controller
	 * @return the list of fields that may be used by the aggregation algorithms.
	 */
	public LyInfoList<LyFieldInfo> getAggregationFieldList(LyTreemapController controller)
	{
		return getNumberFieldInfos(controller);
	}

	/**
	 * Gets the initial field used by current aggregation algorithm, as defined in the metamodel.
	 * <br />
	 * <i>The default implementation</i> <i>The default implementation</i> gets the
	 * "_aggregationFieldInfo" parameter from the metamodel. If the retrieve field is null then gets
	 * the {@link #getSizeField(LyTreemapController) size field}.
	 *
	 * @param  controller the treemap controller
	 * @return the field used by current aggregation algorithm.
	 */
	public LyFieldInfo getAggregationField(LyTreemapController controller)
	{
		LyFieldInfo fieldInfo = (LyFieldInfo)controller.getParameter("_aggregationFieldInfo", true);

		// If aggregation field is not mentionned in the action description, use
		// size field
		if (fieldInfo == null)
			fieldInfo = getSizeField(controller);

		return fieldInfo;
	}

	/**
	 * Gets the displayable label of a map region from the given field info and the given object.
	 * <br />
	 * <i>The default implementation</i> returns the field info value as a {@link
	 * LyFieldInfo#toShortName(Object) short name} or the given object name if the field info is
	 * null.
	 *
	 * @param  controller the treemap controller
	 * @param  fieldInfo  the field info corresponding to the map region. If the region represents a
	 *                    final object of the hierarchy, this parameter is null. Otherwise it
	 *                    contains the field info used at the level of the map region (should be one
	 *                    of the field infos used in the hierarchy).
	 * @param  object     the object describing the map region. If map region does not represent a
	 *                    final object of the hierarchy, the object is the first one contained in
	 *                    the map region.
	 * @return the displayable name for the given map region
	 */
	public String getMapRegionLabel(LyTreemapController controller, LyFieldInfo fieldInfo,
		LyObject object)
	{
		if (fieldInfo == null)
		{
			if (object == null)
				return null;

			return object.getName();
		}
		else
		{
			Object value = object.getValue(fieldInfo);
			String name = fieldInfo.toShortName(value);

			return name;
		}
	}

	/**
	 * Gets the tooltip for a map region from the given parameters.
	 *
	 * @param  controller  the treemap controller
	 * @param  fieldInfo   the field info corresponding to the map region. If the region represents
	 *                     a final object of the hierarchy, this parameter is null. Otherwise it
	 *                     contains the field info used at the level of the map region (should be
	 *                     one of the field infos used in the hierarchy).
	 * @param  object      the object describing the map region. If map region does not represent a
	 *                     final object of the hierarchy, the object is the first one contained in
	 *                     the map region.
	 * @param  weightValue the value used to represent the weight of this region
	 * @param  colorValue  the value used to represent the color of this region
	 * @return the tooltip for the given map region
	 */
	public String getMapRegionTooltip(LyTreemapController controller, LyFieldInfo fieldInfo,
		LyObject object, float weightValue, float colorValue)
	{
		LyInfoList<LyFieldInfo> tooltipFields = controller.getTooltipFields();

		if (fieldInfo == null)
		{
			if (object == null)
				return null;

			// For a real object region, display its name and all the values
			// that must appear
			// in the tooltip
			StringBuffer statusTooltip = new StringBuffer();
			int			 i, n = ((tooltipFields == null) ? 0 : tooltipFields.getSize());
			Object		 value;
			String		 name;

			name = object.getClassInfo().getName();
			statusTooltip.append(name);
			statusTooltip.append(" : ");
			statusTooltip.append(object.getName());

			for (i = 0; i < n; i++)
			{
				fieldInfo = tooltipFields.getInfo(i);
				value = object.getValue(fieldInfo);

				if (value == null)
					continue;

				// Append field and value names
				statusTooltip.append('\n');

				name = fieldInfo.getName();
				statusTooltip.append(name);
				statusTooltip.append(" = ");

				name = fieldInfo.toName(value);
				statusTooltip.append(name);
			}

			return statusTooltip.toString();
		}
		else
		{
			// For a hierarchy region, display its name and the values of size
			// and
			// color fields
			LyFieldInfo   sizeField = controller.getSizeField();
			LyFieldInfo   colorField = controller.getColorField();
			StringBuffer  statusTooltip = new StringBuffer();
			Object		  value = object.getValue(fieldInfo);
			String		  name = fieldInfo.toShortName(value);
			LyEnvironment env = controller.getEnvironment();

			statusTooltip.append(fieldInfo.getName());
			statusTooltip.append(" = ");
			statusTooltip.append(name);

			// Append computed value for size field
			if ((tooltipFields != null)
					&& (tooltipFields.contains(sizeField)))
			{
				statusTooltip.append('\n');
				name = sizeField.getName();
				statusTooltip.append(name);
				statusTooltip.append(" (");
				statusTooltip.append(env.translate("LY_TREEMAP_COMPUTED_VALUE"));
				statusTooltip.append(") = ");
				statusTooltip.append(weightValue + "");

				if (sizeField instanceof LyNumberFieldInfo)
				{
					LyNumberFieldInfo numberField = (LyNumberFieldInfo)sizeField;

					if (numberField.getFieldUnit() != null)
					{
						LyUnit unit = numberField.getFieldUnit();

						statusTooltip.append(" " + env.translate(unit.getName()));
					}
				}
			}

			// Append computed value for color field
			if ((tooltipFields != null)
					&& (tooltipFields.contains(colorField)))
			{
				statusTooltip.append('\n');
				name = colorField.getName();
				statusTooltip.append(name);
				statusTooltip.append(" (");
				statusTooltip.append(env.translate("LY_TREEMAP_COMPUTED_VALUE"));
				statusTooltip.append(") = ");
				statusTooltip.append(colorValue + "");

				if (colorField instanceof LyNumberFieldInfo)
				{
					LyNumberFieldInfo numberField = (LyNumberFieldInfo)colorField;

					if (numberField.getFieldUnit() != null)
					{
						LyUnit unit = numberField.getFieldUnit();

						statusTooltip.append(" " + env.translate(unit.getName()));
					}
				}
			}

			return statusTooltip.toString();
		}
	} // end method getMapRegionTooltip

	/**
	 * Gets a small image for the map region associated to the given object.
	 *
	 * @param  controller the treemap controller
	 * @param  fieldInfo  the field info corresponding to the map region. If the region represents a
	 *                    final object of the hierarchy, this parameter is null. Otherwise it
	 *                    contains the field info used at the level of the map region (should be one
	 *                    of the field infos used in the hierarchy).
	 * @param  object     the object describing the map region. If map region does not represent a
	 *                    final object of the hierarchy, the object is the first one contained in
	 *                    the map region.
	 * @return the small image for the given map region.
	 */
	public String getMapRegionSmallImage(LyTreemapController controller, LyFieldInfo fieldInfo,
		LyObject object)
	{
		// Do not display any image for hierarchy regions
		if (fieldInfo != null)
			return null;

		LyInfoList<LyFieldInfo> pictoFields = controller.getPictoFields();

		if ((pictoFields == null) || (pictoFields.getSize() == 0))
			return null;

		if (object == null)
			return null;

		StringBuffer statusImage = null;
		int			 i, n = pictoFields.getSize();
		Object		 value;

		for (i = 0; i < n; i++)
		{
			fieldInfo = pictoFields.getInfo(i);
			value = object.getValue(fieldInfo);

			// Create new buffer or append if needed
			if (statusImage == null)
				statusImage = new StringBuffer();
			else
				statusImage.append('|');

			if (value != null)
				// Append image path
				statusImage.append(fieldInfo.toString(value));
		}

		String		  result = statusImage.toString();
		LyEnvironment env = controller.getEnvironment();
		String		  path = env.getImagePath(result);

		// If path ends with asked resource, may signify that the resource has
		// no associated
		// image : do not display anything.
		if ((path != null) && (path.endsWith(result)))
			result = null;

		return result;
	} // end method getMapRegionSmallImage

	/**
	 * Gets the tooltip for the small image of a map region.
	 *
	 * @param  controller the treemap controller
	 * @param  fieldInfo  the field info corresponding to the map region. If the region represents a
	 *                    final object of the hierarchy, this parameter is null. Otherwise it
	 *                    contains the field info used at the level of the map region (should be one
	 *                    of the field infos used in the hierarchy).
	 * @param  object     the object describing the map region. If map region does not represent a
	 *                    final object of the hierarchy, the object is the first one contained in
	 *                    the map region.
	 * @return the tooltip for the small image of the given map region
	 */
	public String getMapRegionSmallImageTooltip(LyTreemapController controller,
		LyFieldInfo fieldInfo, LyObject object)
	{
		// Do not display any image for hierarchy regions
		if (fieldInfo != null)
			return null;

		LyInfoList<LyFieldInfo> pictoFields = controller.getPictoFields();

		if ((pictoFields == null) || (pictoFields.getSize() == 0))
			return null;

		if (object == null)
			return null;

		// For a real object region, display its name and all the values that
		// must appear
		// in the small image tooltip
		StringBuffer statusTooltip = new StringBuffer();
		int			 i, n = pictoFields.getSize();
		Object		 value;
		String		 name;

		name = object.getClassInfo().getName();
		statusTooltip.append(name);
		statusTooltip.append(" : ");
		statusTooltip.append(object.getName());

		for (i = 0; i < n; i++)
		{
			fieldInfo = pictoFields.getInfo(i);
			value = object.getValue(fieldInfo);

			if (value == null)
				continue;

			// Append field and value names
			statusTooltip.append('\n');

			name = fieldInfo.getName();
			statusTooltip.append(name);
			statusTooltip.append(" = ");

			name = fieldInfo.toName(value);
			statusTooltip.append(name);
		}

		return statusTooltip.toString();
	} // end method getMapRegionSmallImageTooltip

	/**
	 * Computes the value used to represent the size of the given object.<br />
	 * <i>The default implementation</i> returns the value of the size field or null if the size
	 * field is null.
	 *
	 * @param  controller the treemap controller
	 * @param  sizeField  the field info used to compute sizes
	 * @param  object     the object used to compute size for a given map region
	 * @return the value used to represent the size for a given object.
	 */
	public Number computeSizeValue(LyTreemapController controller, LyFieldInfo sizeField,
		LyObject object)
	{
		Number number = null;

		if (sizeField != null)
			number = (Number)object.getValue(sizeField);

		return number;
	}

	/**
	 * Computes the value used to represent the color for a given list of objects.
	 *
	 * @param  controller  the treemap controller
	 * @param  colorField  the field info used to compute colors
	 * @param  objects     the list of objects used to compute color for a given map region
	 * @param  isHierarchy indicates if the color will be used for a hierarchy map region or a final
	 *                     map region
	 * @return the value used to represent the color for a given list of objects.
	 */
	public Number computeColorValue(LyTreemapController controller, LyFieldInfo colorField,
		LyObjectList objects, boolean isHierarchy)
	{
		LyObject object;
		int		 i, n = objects.getSize();

		if (n == 0)
			return new Float(0);
		else if (!isHierarchy)
		{
			object = objects.getObject(0);

			return (Number)object.getValue(colorField);
		}

		float[]     values = new float[n];
		float[]     balanceValues = new float[n];
		Number	    number;
		String	    aggregationType = controller.getAggregationType(false);
		LyFieldInfo aggregationField = controller.getAggregationField();
		int		    valuesNumber = 0;

		if (aggregationType == null)
			return null;

		for (i = 0; i < n; i++)
		{
			object = objects.getObject(i);
			number = (Number)object.getValue(colorField);

			if (number == null)
				continue;

			values[valuesNumber] = number.floatValue();

			if (LyTreemapController.BALANCED_AGGREGATION.equals(aggregationType)
					&& (aggregationField != null))
			{
				number = (Number)object.getValue(aggregationField);
				balanceValues[valuesNumber] = ((number == null) ? 0 : number.floatValue());
			}

			valuesNumber++;
		}

		return new Float(computeAggregationValue(controller, aggregationType, values, balanceValues,
				valuesNumber));
	} // end method computeColorValue

	/**
	 * Computes a numeric value for the given aggregation algorithm type and given numeric values.
	 *
	 * @param  controller      the treemap controller
	 * @param  aggregationType the code of the algorithm to be used.
	 * @param  values          the values that must be used
	 * @param  balanceValues   the values that must be used to balance the given values. Some
	 *                         algorithms do not use these values.
	 * @param  valuesNumber    indicate the number of values that must be taken into account
	 * @return a numeric value for a specific algorithm and given numeric values.
	 */
	public float computeAggregationValue(LyTreemapController controller, String aggregationType,
		float[] values, float[] balanceValues, int valuesNumber)
	{
		float minValue = Float.MAX_VALUE, maxValue = -Float.MAX_VALUE;
		int   i;

		if (LyTreemapController.MINIMUM_AGGREGATION.equals(aggregationType))
		{
			for (i = 0; i < valuesNumber; i++)
			{
				if (values[i] < minValue)
					minValue = values[i];
			}

			return minValue;
		}
		else if (LyTreemapController.MAXIMUM_AGGREGATION.equals(aggregationType))
		{
			for (i = 0; i < valuesNumber; i++)
			{
				if (values[i] > maxValue)
					maxValue = values[i];
			}

			return maxValue;
		}
		else if (LyTreemapController.AVERAGE_AGGREGATION.equals(aggregationType))
		{
			float total = 0;

			for (i = 0; i < valuesNumber; i++)
				total += values[i];

			return (total / valuesNumber);
		}
		else if (LyTreemapController.BALANCED_AGGREGATION.equals(aggregationType))
		{
			float   total = 0, totalBalanced = 0;
			boolean useBalanceValues = false;

			// Check if all balance values are the same (even 0)
			for (i = 0; i < (valuesNumber - 1); i++)
			{
				if (balanceValues[i] != balanceValues[i + 1])
				{
					useBalanceValues = true;
					break;
				}
			}

			for (i = 0; i < valuesNumber; i++)
			{
				if (useBalanceValues)
				{
					total += values[i] * balanceValues[i];
					totalBalanced += balanceValues[i];
				}
				else
				{
					total += values[i];
					totalBalanced++;
				}
			}

			if (total == 0)
				return 0;

			return (total / totalBalanced);
		}
		else
			return 0;
	} // end method computeAggregationValue

	/**
	 * Computes a color from the given values and the two given colors.
	 *
	 * @param  controller the treemap controller
	 * @param  value      the value used to compute the color
	 * @param  minValue   the value corresponding to the minimum color
	 * @param  maxValue   the value corresponding to the maximum color
	 * @param  minColor   the minimum color
	 * @param  maxColor   the maximum color
	 * @return the color (as a color key or an RGB code) computed from the given value and min and
	 *         max colors.
	 */
	public String computeColor(LyTreemapController controller, float value, float minValue,
		float maxValue, LyColor minColor, LyColor maxColor)
	{
		Hashtable<Float, String>   exceptionColors = controller.getExceptionColors();
		Hashtable<Float, String[]> thresholdColors = controller.getThresholdColors();
		float					   minValueTmp = minValue, maxValueTmp = maxValue;
		boolean					   isMinValueThreshold = false, isMaxValueThreshold = false;

		// Check if an exception color has been defined for this value
		if ((exceptionColors != null) && (exceptionColors.get(new Float(value)) != null))
			return exceptionColors.get(new Float(value));

		// Check if threshold colors have been defined
		if (thresholdColors != null)
		{
			Number	    threshold;
			float	    thresholdValue;
			Enumeration<Float> keys = thresholdColors.keys();
			String[]    colors;

			// Compute new min value, new max value and new colors to be used to
			// compute
			// the color
			while (keys.hasMoreElements())
			{
				threshold = (Number)keys.nextElement();
				thresholdValue = threshold.floatValue();

				if (((thresholdValue <= value) && (thresholdValue >= minValueTmp))
						|| ((thresholdValue <= minValue)
							&& ((thresholdValue >= minValueTmp) || !isMinValueThreshold)))
				{
					minValueTmp = thresholdValue;
					colors = thresholdColors.get(threshold);
					minColor = getLyColor(controller, colors[1]);
					isMinValueThreshold = true;
				}

				if (((thresholdValue >= value) && (thresholdValue <= maxValueTmp))
						|| ((thresholdValue >= maxValue)
							&& ((thresholdValue <= maxValueTmp) || !isMaxValueThreshold)))
				{
					maxValueTmp = thresholdValue;
					colors = thresholdColors.get(threshold);
					maxColor = getLyColor(controller, colors[0]);
					isMaxValueThreshold = true;
				}
			}
		}

		// Use a barycentric algorithm to compute the color
		float   ratio = (value - minValueTmp) / (maxValueTmp - minValueTmp);
		LyColor color = LyColors.getBarycentricColor(minColor, maxColor, ratio);

		return (color.getRed() + " " + color.getGreen() + " " + color.getBlue());
	} // end method computeColor

	/**
	 * Gets a color from the given string.
	 *
	 * @param  controller the treemap controller
	 * @param  color      the color string
	 * @return the LyColor instance for the given color string
	 */
	protected LyColor getLyColor(LyTreemapController controller, String color)
	{
		Object result;

		if (_lastColors == null)
			_lastColors = new LyLastValues();

		result = _lastColors.getValue(color);

		if (result == LyLastValues.NOT_FOUND)
		{
			LyEnvironment env = controller.getEnvironment();

			result = env.getLyColor(color);
			_lastColors.addValue(color, result);
		}

		return (LyColor)result;
	}

	/**
	 * Gets the treemap title from the digen array of labels. This title depends on the current
	 * position in the treemap hierarchy.
	 *
	 * @param  controller the treemap controller
	 * @param  labels     the list of the regions labels used to reach current position in the
	 *                    hierarchy
	 * @return the title of the treemap
	 */
	public String getTreemapTitle(LyTreemapController controller, String[][] labels)
	{
		// The map title represents the position in the hierarchy
		StringBuffer title = new StringBuffer();
		int			 i, n = labels.length;

		for (i = 0; i < n; i++)
		{
			if (i != 0)
				title.append(" / ");

			title.append(labels[i][0]).append(" : ").append(labels[i][1]);
		}

		return title.toString();
	}

	/**
	 * This method is invoked when the view is ready to be shown. May be overriden to add extra
	 * components.<br />
	 * <i>The default implementation</i> sets the map layout to a treemap viewand sets the
	 * selections mode depending on the one specified in the controller.
	 *
	 * @param controller the treemap controller
	 */
	public void completeTreemap(LyTreemapController controller)
	{
		LyMap  map = controller.getMapComponent();
		String borderSizes = (String)controller.getParameter("_borderSizes", true);
		String selectBorderSize = (String)controller.getParameter("_selectBorderSize", true);
		String labelStyles = (String)controller.getParameter("_labelStyles", true);

		// Set the layout
		if ((map != null) && (map.getLayout() == null))
		{
			// Create a default layout.
			LyLayout layout = new LyLayout();

			layout.setStyle(LyLayout.Style.TREEMAP);
			layout.setDeltaBorder(0);
			layout.setBorderSizes(borderSizes);
			layout.setSelectBorderSize(selectBorderSize);
			layout.setLabelStyles(labelStyles);

			map.setLayout(layout);
			map.setViewer(false);

			if (useMonoSelection(controller))
				map.setSelectionMode(LyMap.SelectionMode.SINGLE);
			else
				map.setSelectionMode(LyMap.SelectionMode.MULTIPLE);
		}
	} // end method completeTreemap

	/**
	 * Indicates if mono selection must be used in treemap.<br />
	 * <i>The default implementation</i> does nothing and returns true.
	 *
	 * @param  controller the treemap controller
	 * @return true if mono selection must be used, false if multi selection must be used
	 */
	public boolean useMonoSelection(LyTreemapController controller)
	{
		return true;
	}

	/**
	 * Indicates if treemap component must be refreshed when a selection occurs.<br />
	 * <i>The default implementation</i> returns true if the controller view manager has not the
	 * property "LOCAL_MAP_SELECTION", false otherwise.
	 *
	 * @param  controller the treemap controller
	 * @return true if treemap component must be refreshed when a selection occurs, false otherwise
	 */
	public boolean refreshViewOnSelection(LyTreemapController controller)
	{
		LyViewManager viewManager = controller.getViewManager();

		return (!viewManager.hasProperty("LOCAL_MAP_SELECTION"));
	}

	/**
	 * Indicates if the current user configuration must be saved and restored.<br />
	 * <i>The default implementation</i> returns true if the {@link LyTreemapController#getUser()
	 * controller user} is not null, false otherwise.
	 *
	 * @param  controller the treemap controller
	 * @return true if the configuration has to be saved, false otherwise.
	 */
	public boolean saveUserConfiguration(LyTreemapController controller)
	{
		// Save user configuration only if a user is logged
		LyObject user = controller.getUser();

		return (user != null);
	}

	/**
	 * Frees all resources to close the treemap controller.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param controller the treemap controller
	 */
	public void free(LyTreemapController controller)
	{
	}

	/**
	 * Gets the number field infos of current class.
	 *
	 * @param  controller the treemap controller
	 * @return the number field infos of current class.
	 */
	protected LyInfoList<LyFieldInfo> getNumberFieldInfos(LyTreemapController controller)
	{
		LyClassInfo			    classInfo = controller.getClassInfo();
		LyInfoList<LyFieldInfo> list = classInfo.getFieldInfos();

		list = list.getInfos(LyFieldInfo.MARK_HIDDEN, true);

		if (list != null)
			list = list.getInfos(LyFieldInfo.MARK_SECRET, true);
		if (list == null)
			list = new LyInfoList<LyFieldInfo>();

		LyFieldInfo field;
		int		    i, n = list.getSize();

		for (i = n - 1; i >= 0; i--)
		{
			field = list.getInfo(i);
			if (!(field instanceof LyNumberFieldInfo))
				list.removeInfo(field);
		}

		return list;
	}
} // end class LyTreemapBehavior
