/**
 * LyMapController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyMapController.java 36051 2012-06-27 13:58:05Z nhamon $
 */

package eu.w4.ae.control;

import java.util.*;

import eu.w4.ae.app.behaviorinterface.*;
import eu.w4.ae.control.builder.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;


/**
 * This class implements generic processing of the map. The default implementation handles
 * navigation between different related levels of class Info. This controller is associated with a
 * behavior class which implements a LyMapBehaviorInterface. Usually, the behavior is a
 * LyMapBehavior instance.
 *
 * @see leon.app.behaviorinterface.LyMapBehaviorInterface
 * @see leon.app.behavior.LyMapBehavior
 */
public class LyMapController extends LyPanelController
	implements LyAdvancedWorkSpaceListener,
		Runnable,
		LyFindInterface,
		LyPrintListenerInterface,
		LyFilterEndInterface
{
	/** File Version */
	public static final String		 __VERSION =
		"$Id: LyMapController.java 36051 2012-06-27 13:58:05Z nhamon $";

	/** Closing current area. Go back. */
	public static final String		 DEMAND_BACK = "_back".intern();

	/** Printing the map. */
	public static final String		 DEMAND_PRINT_MAP = "_printMap".intern();

	/**
	 * Association table mapObject, application object keys = map 'view' objects values =
	 * application objects : LyObject instances.
	 */
	protected LyDependencies		 _objectAssoc;

	/** Specific controller class to create and keep up to date the content of the view. */
	protected LyMapBehaviorInterface _mapBehavior;

	/** Caching : registered WorkSpace (to enable free impl). */
	protected LyWorkSpace			 _nodeList;

	/** Caching : registered WorkSpace (to enable free impl). */
	protected LyWorkSpace			 _linkList;

	/** Caching : registered WorkSpace (to enable free impl). */
	protected LyWorkSpace			 _regionList;

	/** Caching : map object. */
	protected LyMap					 _map;

	/**
	 * Initially present components may NON be removed with objects. Usefull when using Map template
	 * in an XML file.
	 */
	protected Vector<LyComponent>    _lockedMapObjects;

	/** Class info. */
	protected LyClassInfo			 _classInfo;

	/** List of classInfos to navigate between levels. */
	protected Vector<LyClassInfo>    _classInfos;

	/** List of contexts to navigate between levels. */
	protected Stack<LyWorkSpace>     _contexts;

	/** Stack of visited contents. */
	protected Stack<Object>			 _contents;

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo class info for this controller
	 * @param context   current context (may be null if the view is not contextual to a list of
	 *                  objects).
	 */
	public LyMapController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent, action, context);

		_classInfo = classInfo;

		init();
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyMapBehaviorInterface.class;
	}

	/**
	 * Initialize map controller.
	 */
	protected void init()
	{
		_objectAssoc = new LyDependencies();
		_nodeList = null;
		_linkList = null;
		_regionList = null;
		_lockedMapObjects = null;
		_mapBehavior = (LyMapBehaviorInterface)getBehavior();
		_contents = new Stack<Object>();
		_classInfo = (LyClassInfo)getParameter("_classInfo");
		_map = null;
		LyClassInfo classInfo;

		classInfo = (LyClassInfo)getParameter("_classInfo2", true);
		if ((_classInfo != null) && (classInfo != null))
		{
			_classInfos = new Vector<LyClassInfo>();
			classInfo = _classInfo;

			int i = 2;

			while (classInfo != null)
			{
				_classInfos.addElement(classInfo);
				classInfo = (LyClassInfo)getParameter("_classInfo" + i, true);
				i++;
			}

			_contexts = new Stack<LyWorkSpace>();
		}
		else
		{
			_classInfos = null;
			_contexts = null;
		}
        
        initPreference();
	} // end method init

	/**
	 * Gets current controller class to manage specific behavior of the map content.
	 *
	 * @return the specific controller of the map.
	 * @see    leon.app.behaviorinterface.LyMapBehaviorInterface
	 */
	public LyMapBehaviorInterface getMapBehaviorInterface()
	{
		return _mapBehavior;
	}

	/**
	 * Prepare component before showing. This method is called once, before first showing.
	 */
	protected void initArea()
	{
		if (_initialized)
			return;

		LyMap map = getMapComponent();

		if (map != null)
		{
			int i, n = map.getComponentCount();

			if (n > 0)
			{
				_lockedMapObjects = new Vector<LyComponent>(n);
				for (i = 0; i < n; i++)
					_lockedMapObjects.addElement(map.getComponent(i));
			}
		}

		super.initArea();
		if (isFreed()) // May be freed by super.initArea if startAction fails
		    return;
		
		completeMap();
	}

	/**
	 * Reset the entire view : everything is rebuilt.
	 *
	 * @param full Boolean indicating if a full reset (including reset of datas) is requiered or not
	 *             (only reset graphical area).
	 */
	public void reset(boolean full)
	{
		if (_component == null)
			return;

		synchronized (getMapComponent())
		{
			boolean idle = isIdle();

			_map.setIdle(true);

			updateContent();
			updateContextClient();
			updateTools();
			updateMessages();
			updateTitleBar();
            updateFilterMenu();
            updatePreferencesMenu();

			// Update the last modification time.
			setLastModified(new Date().getTime());

			_map.setIdle(idle);
		}

		// Call the behavior
		// Do not call super, because it will redo exactly the samle sequence without
		// completemap
		LyBehaviorInterface behavior = getBehavior();

		if (behavior instanceof LyActionBehaviorInterface)
			((LyActionBehaviorInterface)behavior).reset(this, full);

	} // end method reset

	/**
	 * Reset the content of the view (Nodes, links ...).
	 */
	public void updateContent()
	{
		if (_component == null)
			return;
		if (_mapBehavior == null)
			return;

		synchronized (getMapComponent())
		{
			LyMapNode   mapNode;
			LyMapLink   mapLink;
			LyMapRegion mapRegion;
			LyObject    appObject;
			int		    i, n;
            LyComponent node;

			// Retrieve internal map component
			LyMap	    map = getMapComponent();

			// Idle mode
			boolean     idle = isIdle();

			_map.setIdle(true);

			// Clear all
			if (_nodeList != null)
			{
				_nodeList.free(this);
				_nodeList = null;
			}
			if (_linkList != null)
			{
				_linkList.free(this);
				_linkList = null;
			}
			if (_regionList != null)
			{
				_regionList.free(this);
				_regionList = null;
			}

			clearMapObjects();

			// retrieve map content
			_nodeList = registerWorkSpace(_mapBehavior.getNodeObjects(this));
			_linkList = registerWorkSpace(_mapBehavior.getLinkObjects(this));
			_regionList = registerWorkSpace(_mapBehavior.getRegionObjects(this));

			// Get map objects : 1 Nodes.
			n = ((_nodeList == null) ? 0 : _nodeList.getSize());

			for (i = 0; i < n; i++)
			{
				appObject = _nodeList.getObject(i);

				// Build the node and sets its content
                node = _map.getComponent(appObject.getId());
                
                boolean	addNode = false;
                boolean	selected = appObject.isSelected();
                
                if (node instanceof LyMapNode)
                {
                    mapNode = (LyMapNode)node;
                    mapNode.setHidden(false);
                    selected = mapNode.isSelected();
                }
                else
                {
                    mapNode = _mapBehavior.buildNode(appObject, this);
                    
    				if (mapNode == null)
    					continue;
    				
    				addNode = true;
                }

				_mapBehavior.updateNode(appObject, mapNode, this);
                
				if (addNode)
					map.addMapNode(mapNode);

				// Add the node in the map
				_objectAssoc.addDepend(mapNode, appObject);

				// Restore selection status (if any)
				if ((_selections != null)
						&& (_selections.contains(appObject)))
					mapNode.setSelected(true);
				else if (selected)
				{
					if (_selections == null)
					{
						_selections = new LySimpleObjectList();
						_selections.setSorted(false);
					}
					_selections.addObject(appObject);
					mapNode.setSelected(true);
				}

				LyClassInfo classInfo = appObject.getClassInfo();
				LyInfoLabel label = getInfoLabel(classInfo, LyInfoLabel.Type.SEMANTIC);

				if (label != null)
					mapNode.setOverview(label.getString(appObject));
			}
            
            // Hide sensitive nodes that don't correspond to known objects
            

			// Get map objects : 2 Links.
			n = ((_linkList == null) ? 0 : _linkList.getSize());

			for (i = 0; i < n; i++)
			{
				appObject = _linkList.getObject(i);

				// Build the links and sets its content
				mapLink = _mapBehavior.buildLink(appObject, this);

				if (mapLink == null)
					continue;

				_mapBehavior.updateLink(appObject, mapLink, this);

				// Restore selection status (if any)
				if ((_selections != null)
						&& (_selections.contains(appObject)))
					mapLink.setSelected(true);
				else if (appObject.isSelected())
				{
					if (_selections == null)
					{
						_selections = new LySimpleObjectList();
						_selections.setSorted(false);
					}
					_selections.addObject(appObject);
					mapLink.setSelected(true);
				}

				LyClassInfo classInfo = appObject.getClassInfo();
				LyInfoLabel label = getInfoLabel(classInfo, LyInfoLabel.Type.SEMANTIC);

				if (label != null)
					mapLink.setOverview(label.getString(appObject));

				// Add the link in the map
				map.addMapLink(mapLink);
				_objectAssoc.addDepend(mapLink, appObject);
			}

			// Get map objects : 3 Regions.
			n = ((_regionList == null) ? 0 : _regionList.getSize());

			for (i = 0; i < n; i++)
			{
				appObject = _regionList.getObject(i);

				// Build the region and sets its content
				mapRegion = _mapBehavior.buildRegion(appObject, this);

				if (mapRegion == null)
					continue;

				_mapBehavior.updateRegion(appObject, mapRegion, this);

				// Restore selection status (if any)
				if ((_selections != null)
						&& (_selections.contains(appObject)))
					mapRegion.setSelected(true);
				else if (appObject.isSelected())
				{
					if (_selections == null)
					{
						_selections = new LySimpleObjectList();
						_selections.setSorted(false);
					}
					_selections.addObject(appObject);
					mapRegion.setSelected(true);
				}

				LyClassInfo classInfo = appObject.getClassInfo();
				LyInfoLabel label = getInfoLabel(classInfo, LyInfoLabel.Type.SEMANTIC);

				if (label != null)
					mapRegion.setOverview(label.getString(appObject));

				// Add the region in the map
				map.addMapRegion(mapRegion);
				_objectAssoc.addDepend(mapRegion, appObject);
			}

			// Idle mode
			_map.setIdle(idle);
		}
	} // end method updateContent
    
	/**
	 * @see leon.control.LyAreaController#completeMessages(leon.view.LyMessages, int)
	 */
	public void completeMessages(LyMessages messages, int level)
	{
		if (_action.hasMark(LyAction.MARK_NO_MESSAGES))
			return;
        
		LyMessage objectsMessage = new LyMessage(this.getId() + "_objectsMessage", null);

		objectsMessage.setSize(10);
		messages.addMessage(objectsMessage);
        
        LyMessage filterMessage = buildFilterMessage();

        if (filterMessage != null)
            messages.addMessage(filterMessage);
	}

	/**
	 * Reset the message bar status if any.
	 */
	public void updateMessages()
	{
		// XXX this part of code is a copy of LyTableController : a better way is
		// to factorize between area controllers.
		if (_component == null)
			return;
        
        super.updateMessages();
        
		LyView		  view = getView();
		
		if (view == null)
			return;
		
		LyMessage     objectsMessage = (LyMessage)view.getComponent(getId() + "_objectsMessage");
		int			  n = (_objectAssoc.getSize()) / 2;
		int			  nSelections = ((_selections == null) ? 0 : _selections.getSize());
		LyEnvironment environment = getEnvironment();
		StringBuffer  selectionsLabel;

		// Set message for objects
		if (objectsMessage != null)
		{
			selectionsLabel = new StringBuffer(environment.getMessage("LY_OBJECTS_COUNT", new Integer(n)));

			if (n > 0)
			{
				selectionsLabel.append(" (");
				selectionsLabel.append(environment.getMessage("LY_SELECTED_OBJECTS_COUNT", new Integer(nSelections)));
				selectionsLabel.append(")");
			}

			objectsMessage.setLabel(selectionsLabel.toString());
		}
        
		// Update the last modification time.
		setLastModified(new Date().getTime());
	} // end method updateMessages

	/**
	 * updates the status of Tool and menu items.
	 */
	protected void updateTools()
	{
		super.updateTools();

		if (_component == null)
			return;

		// Update 'Back' tool
		LyToolBar toolBar = (LyToolBar)_component.getComponent("_toolBar");

		if (toolBar == null)
			return;

		LyTool  tool = (LyTool)toolBar.getComponent(LyMapController.DEMAND_BACK);
		boolean enabled = (_contents != null) && (!_contents.empty());

		if (tool != null)
			tool.setEnabled(enabled);
	}

	/**
	 * To add extra processing before showing the map. invoke method on declared
	 * mapBehaviorInterface
	 */
	public void completeMap()
	{
		if (_mapBehavior != null)
			_mapBehavior.completeMap(this);

		// Reset the map
		LyMap map = getMapComponent();

		if (map != null)
			map.resetContent();
	}

	/**
	 * This method frees this controller.
	 *
	 * @param controller the parent controller
	 */
	public void free(LyController controller)
	{
		if (_parent != controller)
			return;

		if (_mapBehavior != null)
			_mapBehavior.free(this);

		// Free workspaces
		if (_nodeList != null)
			_nodeList.free(this);
		if (_linkList != null)
			_linkList.free(this);
		if (_regionList != null)
			_regionList.free(this);

		// Free memory (gc)
		_nodeList = null;
		_linkList = null;
		_regionList = null;
		_selections = null;

		// Retrieve internal map component
		if (_component != null)
		{
			LyMap map = getMapComponent();

			if (map != null)
				map.free();
		}

		_map = null;

		super.free(controller);
	} // end method free

	/**
	 * Method called when there is a selection in the map. It will update toolbar, messagesBar and
	 * menuBar and change the aspect of the selected node
	 *
	 * @param demand selection demand
	 * @see   LyController#selectionDemand
	 */
	public void selectionDemand(LySelectionDemand demand)
	{
		if (_component == null)
			return;

		boolean idle = isIdle();
		boolean extend = demand.isExtend();
		boolean select = false;
		LyMap   map = getMapComponent();
		int     i, n = map.getComponentCount();
		int     selectionType = demand.getType();

		_map.setIdle(true);

		if (!extend
				|| (selectionType == LySelectionDemand.TYPE_UNSELECT_ALL))
			selectAllObjects(false);

		if ((selectionType == LySelectionDemand.TYPE_SELECT)
				|| (selectionType == LySelectionDemand.TYPE_UNSELECT))
		{

			n = demand.getSelectionSize();

			if (selectionType == LySelectionDemand.TYPE_SELECT)
				select = true;
			else if (selectionType == LySelectionDemand.TYPE_UNSELECT)
				select = false;

			if (n != 0)
			{
				LyObject    object;
				LyComponent mapObject;
				String	    key;

				if (_selections == null)
				{
					_selections = new LySimpleObjectList();
					_selections.setSorted(false);
				}

				for (i = 0; i < n; i++)
				{
					key = demand.getSelectionId(i);
					mapObject = getMapObject(key);
					object = getObjectAssoc(mapObject);

					selectObject(mapObject, object, select, false, false);
				}
			}
		}

		updateContextClient();
		updateTools();
		updateMessages();
		_session.notifyOutline(this);

		setLastModified(System.currentTimeMillis());
		_map.setIdle(idle);
	} // end method selectionDemand

	/**
	 * Select all objects in panel.
	 *
	 * @param select new selection status
	 */
	public void selectAllObjects(boolean select)
	{
		if (!select)
		{
			LyMap		 map = getMapComponent();
			int			 i, m, n = map.getComponentCount();
			LyMapContent content = map.getMapContent();
			LyComponent  component;

			for (i = 0; i < n; i++)
			{
				component = map.getComponent(i);

				if (component instanceof LyMapNode)
				{
					if (((LyMapNode)component).isSelected())
						((LyMapNode)component).setSelected(false);
				}
				else if (component instanceof LyMapLink)
				{
					if (((LyMapLink)component).isSelected())
						((LyMapLink)component).setSelected(false);
				}
				else if (component instanceof LyMapRegion)
				{
					if (((LyMapRegion)component).isSelected())
						((LyMapRegion)component).setSelected(false);
				}
			}

			if (content != null)
			{
				m = content.getComponentCount();
				for (i = 0; i < m; i++)
				{
					component = content.getComponent(i);
					if (component instanceof LyMapNode)
					{
						if (((LyMapNode)component).isSelected())
							((LyMapNode)component).setSelected(false);
					}
					else if (component instanceof LyMapLink)
					{
						if (((LyMapLink)component).isSelected())
							((LyMapLink)component).setSelected(false);
					}
					else if (component instanceof LyMapRegion)
					{
						if (((LyMapRegion)component).isSelected())
							((LyMapRegion)component).setSelected(false);
					}
				}
			}

			if (_mapBehavior != null)
				_mapBehavior.unselectNodes(this);
		}

		super.selectAllObjects(select);
	} // end method selectAllObjects

	/**
	 * Select an object in the table.
	 *
	 * @param object selected or unselected object
	 * @param select new selection status
	 */
	public void selectObject(LyObject object, boolean select)
	{
		LyComponent mapObject = getObjectAssoc(object);

		if (mapObject != null)
			selectObject(mapObject, object, select, true, select);
	}

	/**
	 * Select a object in the map.
	 *
	 * @param mapObject
	 * @param object
	 * @param select          new selection status
	 * @param updateView      indicates if the view (tools, messages...) must be updated.
	 * @param resetMapContent indicates if the map must be reset.
	 */
	protected void selectObject(LyComponent mapObject, LyObject object, boolean select,
		boolean updateView, boolean resetMapContent)
	{
		if (mapObject instanceof LyMapNode)
		{
			if (((LyMapNode)mapObject).isSelected() != select)
				((LyMapNode)mapObject).setSelected(select);
		}
		else if (mapObject instanceof LyMapLink)
		{
			if (((LyMapLink)mapObject).isSelected() != select)
				((LyMapLink)mapObject).setSelected(select);
		}
		else if (mapObject instanceof LyMapRegion)
		{
			if (((LyMapRegion)mapObject).isSelected() != select)
				((LyMapRegion)mapObject).setSelected(select);
		}

		if (object == null)
			return;

		if (_selections == null)
		{
			_selections = new LySimpleObjectList();
			_selections.setSorted(false);
		}

		if (select)
		{
			// If object was not previously selected, update selection and notify behavior
			if (!_selections.contains(object))
			{
				_selections.addObject(object);
				if ((_mapBehavior != null) && (mapObject instanceof LyMapNode))
					_mapBehavior.selectNode(object, (LyMapNode)mapObject, this);
			}
		}
		else
		{
			// If object was previously selected, update selection and notify behavior
			if (_selections.contains(object))
			{
				_selections.removeObject(object);

				if ((_mapBehavior != null) && (mapObject instanceof LyMapNode))
					_mapBehavior.unselectNode(object, (LyMapNode)mapObject, this);
			}
		}
		object.setSelected(select);

		if (updateView)
		{
			updateContextClient();
			updateTools();
			updateMessages();
		}

		if (resetMapContent)
			_map.resetContent();

		setLastModified(System.currentTimeMillis());
	} // end method selectObject

	/**
	 * Hide an object in the map.
	 *
	 * @param mapObject
	 * @param object
	 * @param hide       new hidden status
	 * @param updateView indicates if the view (tools, messages...) must be updated.
	 */
	protected void hideObject(LyComponent mapObject, LyObject object, boolean hide,
		boolean updateView)
	{
		if (mapObject instanceof LyMapNode)
		{
			if (((LyMapNode)mapObject).isHidden() != hide)
				((LyMapNode)mapObject).setHidden(hide);
		}

		if (object == null)
			return;

		if (updateView)
		{
			updateContextClient();
			updateTools();
			updateMessages();
			_map.resetContent();
		}
		setLastModified(new Date().getTime());
	}

	/**
	 * accessor to the LyObject instance by its Id.
	 *
	 * @param  key the object id as a string
	 * @return the object id corresponding to the given id if any
	 */
	public LyObject findObject(String key)
	{
		if (key == null)
			return null;

		// Search for the object identified by key in the node list.
		LyObject object = null;

		if (_nodeList != null)
			// Search for the object identified by key in the nodes list.
			object = _nodeList.getObject(getObjectId(key));

		if ((object == null)
				&& (_linkList != null))
			// Search for the object identified by key in the links list.
			object = _linkList.getObject(getObjectId(key));

		if ((object == null)
				&& (_regionList != null))
			// Search for the object identified by key in the regions list.
			object = _regionList.getObject(getObjectId(key));

		return object;
	}

	/**
	 * Accessor to an object of the map (LyMapNode, LyMapLink, LyMapRegion) by its id.
	 *
	 * @param  id the map object id
	 * @return the object corresponding to the id.
	 */
	public LyComponent getMapObject(String id)
	{
		if (_component == null)
			return null;

		if (id == null)
			return null;

		LyMap			    map = getMapComponent();
		LyMapContent	    content = map.getMapContent();
		LyComponent		    component = null;
		String			    idComponent;
		Vector<LyComponent> components = new Vector<LyComponent>();
		int				    i, n;

		if (content != null)
		{
			n = content.getComponentCount();

			for (i = 0; i < n; i++)
				components.addElement(content.getComponent(i));
		}

		n = map.getComponentCount();

		for (i = 0; i < n; i++)
			components.addElement(map.getComponent(i));

		n = components.size();

		for (i = 0; i < n; i++)
		{
			component = components.elementAt(i);

			if (component instanceof LyGetIdInterface)
			{
				idComponent = ((LyGetIdInterface)component).getId();

				if (id.equals(idComponent))
					return component;
			}
		}

		return null;
	} // end method getMapObject

	/**
	 * Invoked when an action occurs. At this level, the implementation manages
	 *
	 * <ul>
	 *   <li>navigation between the different levels of a map (demand.getType with '_map' in it) or
	 *     (source == LyMapController.DEMAND_BACK)</li>
	 *   <li>translation of an object (demand.getType() == LyActionDemand.TYPE_MOVE)</li>
	 *   <li>deletion of an object (demand.getType() == LyActionDemand.TYPE_DELETE)</li>
	 * </ul>
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (_component == null)
			return;

		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			String   source = demand.getSource();
			String[] parameters = LyString.stringToList(source, ID_SEP);
			String   id = ((parameters == null) ? "" : parameters[0].intern());

			if (id.startsWith(getId() + "_map") || (id.startsWith("_map")))
			{
				// Default action on an item of the map
				String	    idObject = demand.getValue();
				LyComponent mapObject = getMapObject(idObject);
				LyObject    object;
				int		    x = demand.getIndex(0);
				int		    y = demand.getIndex(1);

				// Let the behavior handle this action ...
				if ((_mapBehavior != null)
						&& (_mapBehavior.invokeDefaultAction(getObjectAssoc(mapObject), mapObject,
								x, y, this)))
					return;

				// Check if action is on a region
				if (mapObject instanceof LyMapRegion)
				{
					openRegion((LyMapRegion)mapObject);
					return;
				}

				// If map object is not associated with an application object, try
				// to show the previous content
				if ((object = getObjectAssoc(mapObject)) == null)
				{
					closeRegion();
					return;
				}

				invokeDefaultAction(object);
				return;
			}
			else if (source.equals(LyMapController.DEMAND_BACK))
			{
				closeRegion();
				return;
			}
			else if (id == LyMapController.DEMAND_PRINT_MAP)
			{
				printMap();
				return;
			}
			else if ("displayViewer".equals(id)
					|| "displayTranslation".equals(id)
					|| "displayZoom".equals(id)
					|| "moveViewer".equals(id)
					|| "setMapZoomScale".equals(id)
					|| "translateMap".equals(id)
					|| "translateViewer".equals(id))
			{
				if (_mapBehavior != null)
					_mapBehavior.widgetAction(this, id, demand.getValue());
			}
			else
				super.actionDemand(demand);
		}
		else if (demand.getType() == LyActionDemand.TYPE_MOVE)
		{
			String source = demand.getSource();
			String value = demand.getValue();

			if ((source != null)
					&& (source.length() > 0)
					&& (value != null)
					&& (value.length() > 0))
			{

				String[] values = LyString.stringToList(value, '|');

				if (values.length == 2)
				{

					float	    x = Float.parseFloat(values[0]);
					float	    y = Float.parseFloat(values[1]);
					LyComponent mapObject = getMapObject(source);

					if (mapObject instanceof LyMapNode)
					{
						LyMapNode mapNode = (LyMapNode)mapObject;

						mapNode.setX(x);
						mapNode.setY(y);

						setLastModified(System.currentTimeMillis());
					}
				}
			}

            if (_selections != null)
            {
                int         i, n = _selections.getSize();
                LyObject    object;
                LyComponent mapObject;
                
                if (_lockedMapObjects == null)
                    _lockedMapObjects = new Vector<LyComponent>(n);
                
                for (i = 0; i < n; i++)
                {
                    object = _selections.getObject(i);
                    mapObject = getMapObject(object.getId());
                    if (mapObject instanceof LyMapNode)
                        _lockedMapObjects.addElement(mapObject);
                }
            }
            
			// Some nodes have been moved
			if (_mapBehavior != null)
				_mapBehavior.selectionMoved(this);
		}
		else if (demand.getType() == LyActionDemand.TYPE_DELETE)
		{
			if ((_selections == null) || (_selections.getSize() == 0))
				return;

			invokeDeleteAction(_selections, _classInfo);
		}
        else
            super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Invoked when a Drag & Drop occurs. (LyDemandListener interface) This method handles: opening
	 * and closing line, moving an interval, action
	 *
	 * @param demand drag-drop demand
	 */
	public void dragDropDemand(LyDragDropDemand demand)
	{
		if (demand.getType() == LyDragDropDemand.TYPE_DRAG_START)
		{
			LyObject    draggedObject = null;
			String	    elementId = demand.getSource();
			LyComponent mapObject = getMapObject(elementId);

			draggedObject = getObjectAssoc(mapObject);

			if (_mapBehavior != null)
				_mapBehavior.dragStart(this, draggedObject);
		}
		else if (demand.getType() == LyDragDropDemand.TYPE_DRAG_OVER)
		{
			LySession session = this.getSession();
			LyObject  draggedObject = (LyObject)session.getValue(LySession.DRAGGED_OBJECT);

			if (_mapBehavior != null)
			{
				int[] p = demand.getPosition();

				_mapBehavior.dragOver(this, draggedObject, p[0], p[1]);
			}
		}
		else if (demand.getType() == LyDragDropDemand.TYPE_DROP)
		{
			LyObject    droppedObject = null;
			String	    elementId = demand.getSource();
			LySession   session = this.getSession();
			LyObject    draggedObject = (LyObject)session.getValue(LySession.DRAGGED_OBJECT);
			LyComponent mapObject = getMapObject(elementId);

			droppedObject = getObjectAssoc(mapObject);

			if (_mapBehavior != null)
			{
				int[] p = demand.getPosition();

				_mapBehavior.drop(this, droppedObject, draggedObject, p[0], p[1]);
			}
		}
		else if (demand.getType() == LyDragDropDemand.TYPE_DRAG_FINISHED)
		{
			if (_mapBehavior != null)
				_mapBehavior.dragFinished(this, Integer.parseInt(demand.getValue()));
		}

		// Update the last modification time.
		setLastModified(System.currentTimeMillis());
	} // end method dragDropDemand

	/**
	 * Opens a region : check if this region has a content and if so set it as the map content.
	 *
	 * @param region the region to open
	 */
	protected void openRegion(LyMapRegion region)
	{
		if (_component == null)
			return;

		LyMap		 map = getMapComponent();
		LyMapContent mapContent = map.getMapContent();
		LyMapContent regionContent = region.getMapContent();

		// If no content under the region, do nothing
		if (regionContent == null)
			return;

		getViewManager().setWaiting(true);

		if (_contents == null)
			_contents = new Stack<Object>();

		_contents.push(mapContent);

		setMapContent(regionContent);
		_map.resetContent();

		getViewManager().setWaiting(false);

		setVisible(true);
	}

	/**
	 * Closes a region : set the previous content as the map content.
	 */
	protected void closeRegion()
	{
		Object parentContentObj = null;

		if ((_contents != null) && (!_contents.empty()))
			parentContentObj = _contents.pop();

		// If no content has been registered, return
		if (parentContentObj instanceof LyClassInfo)
		{
			// Maybe we have a hierarchy of classInfos ?
			if (_classInfos != null)
			{
				int index = _classInfos.indexOf(parentContentObj);

				if (index >= 0)
				{
					_classInfo = _classInfos.elementAt(index);

					LyWorkSpace workspace = _contexts.pop();

					if (workspace == null)
						setContext(null, false);
					else if (workspace.getSize() > 0)
					{
						setContext(workspace, false);
						workspace.unlock();
					}
					else
						closeRegion();
				}
			}

			setVisible(true);
			return;
		}

		LyMapContent parentContent = (LyMapContent)parentContentObj;

		if (parentContent == null)
			return;

		setMapContent(parentContent);
		_map.resetContent();

		setVisible(true);
	} // end method closeRegion

	/**
	 * Sets the content of the map.
	 *
	 * @param content the content to set
	 */
	protected void setMapContent(LyMapContent content)
	{
		if (_mapBehavior == null)
			return;

		synchronized (getMapComponent())
		{

			LyMap map = getMapComponent();

			_mapBehavior.setMapContent(this, map, content);

			map.setMapContent(content);

			reset();
		}
	}

	/**
	 * Empty the map from all map objects.
	 */
	public void clearMapObjects()
	{
		if (_component == null)
			return;

		synchronized (getMapComponent())
		{
			Enumeration<Object> e = _objectAssoc.keys();
			Object	    mapObject;

			// Retrieve internal map component
			LyMap	    map = getMapComponent();

			// Remove Everything
			while (e.hasMoreElements())
			{
				mapObject = e.nextElement();

				// Skip map objects
				if ((_lockedMapObjects != null)
                        && _lockedMapObjects.contains(mapObject))
					continue;

				// Removal depends on the nature of the map object
				if (mapObject instanceof LyMapNode)
					map.removeMapNode((LyMapNode)mapObject);
				else if (mapObject instanceof LyMapLink)
					map.removeMapLink((LyMapLink)mapObject);
				else if (mapObject instanceof LyMapRegion)
					map.removeMapRegion((LyMapRegion)mapObject);
			}
            
            // Hide sensitive map nodes and links
            if (_lockedMapObjects != null)
            {
                int     i, n = _lockedMapObjects.size();
                
                for (i = 0; i < n; i++)
                {
                    mapObject = _lockedMapObjects.get(i);
                    if ((mapObject instanceof LyMapNode)
                            && ((LyMapNode)mapObject).isSensitive())
                        ((LyMapNode)mapObject).setHidden(true);
                    else if ((mapObject instanceof LyMapLink)
                            && ((LyMapLink)mapObject).isSensitive())
                        ((LyMapLink)mapObject).setHidden(true);
                }
            }

			// Clear assoc table
			_objectAssoc = new LyDependencies();

			// Clear selections.
			_selections = null;
		}
	} // end method clearMapObjects

	/**
	 * Invoked when a menu action occurs.
	 *
	 * @param demand menu demand
	 */
	public void menuDemand(LyMenuDemand demand)
	{
		if (demand == null)
			return;

		LyMap map = getMapComponent();

		if ((map != null) && (demand.getSource().equals(map.getId())))
		{
			if (_component == null)
				return;

			// Find target object, or null if popup on 'background'
			String	    objId = demand.getItem();
			LyComponent mapObject = getMapObject(objId);
			LyObject    object;
			LyAction    action = getAction();

			if (action.hasMark(LyAction.MARK_NO_MENU))
				return;

			// If map object is not associated with an application object, return
			if ((object = getObjectAssoc(mapObject)) == null)
			{
				if (objId == null)
				{
					// Should enable map actions : zoom, rotate, create nodes if sel
					// empty, create lionk if two nodes selected
					getEnvironment().logWarning(
						"LyMapController : Popup on background not implemented yet.");
				}
				
				return;
			}

			LyClassInfo classInfo = object.getClassInfo();
			
			LyPopupMenu popupMenu = buildPopupMenu(demand, classInfo, true);
			
			// Coordinates
			int[] p = demand.getPosition();

			popupMenu.setX(p[0]);
			popupMenu.setY(p[1]);

			popupMenu.setController(getId());
			popupMenu.setEnvironment(map.getEnvironment());
			map.setPopupMenu(popupMenu);
		}
		else
			super.menuDemand(demand);
	} // end method menuDemand

	/**
	 * Accessor to objects contained in the map.
	 *
	 * @return the list of object of the map
	 */
	public LyObjectList getContent()
	{
		LyWorkSpace content = new LyMultiWorkSpace(getSession(), true);

		content.setSorted(false);

		if (_nodeList != null)
			content.addList(_nodeList);
		if (_linkList != null)
			content.addList(_linkList);
		if (_regionList != null)
			content.addList(_regionList);

		return content;
	}

	/* Implementation of LyWorkSpaceListener */

	/**
	 * Called when an object is added in the dynamic object list.
	 * 
	 * @param event the workspace event
	 * @return true if the demand was take into account
	 * @see    LyWorkSpaceListener#objectAdded
	 */
	public boolean objectAdded(LyWorkSpaceEvent event)
	{
		if ((_mapBehavior == null) || (isIdle()))
			return false;

		if (_component == null)
		{
			// Update the last modification time.
			setLastModified(new Date().getTime());
			return true;
		}

		boolean idle = isIdle();

		_map.setIdle(true);

		synchronized (getMapComponent())
		{

			LyWorkSpace source = event.getSource();
			LyObject    appObject = event.getObject();
			int		    index = event.getIndex();

			// Retrieve internal map component
			LyMap	    map = getMapComponent();

			if (source == _nodeList)
			{
				LyMapNode   mapNode;
                LyComponent node = _map.getComponent(appObject.getId());
                
                if (node instanceof LyMapNode)
                {
                    mapNode = (LyMapNode)node;
                    mapNode.setHidden(false);
                }
                else
                {                
    				// Add the node and keep the assoc table up to date
    				mapNode = _mapBehavior.buildNode(appObject, this);
    				if (mapNode == null)
    				{
    					_map.setIdle(idle);
    					return false;
    				}
                }

				LyClassInfo classInfo = appObject.getClassInfo();
				LyInfoLabel label = getInfoLabel(classInfo, LyInfoLabel.Type.SEMANTIC);

				if (label != null)
					mapNode.setOverview(label.getString(appObject));

				_mapBehavior.updateNode(appObject, mapNode, this);
				map.insertMapNode(mapNode, index);
				_objectAssoc.addDepend(mapNode, appObject);
			}
			else if (source == _linkList)
			{
				LyMapLink mapLink;

				// Add the node and keep the assoc table up to date
				mapLink = _mapBehavior.buildLink(appObject, this);

				if (mapLink == null)
				{
					_map.setIdle(idle);
					return false;
				}

				LyClassInfo classInfo = appObject.getClassInfo();
				LyInfoLabel label = getInfoLabel(classInfo, LyInfoLabel.Type.SEMANTIC);

				if (label != null)
					mapLink.setOverview(label.getString(appObject));

				_mapBehavior.updateLink(appObject, mapLink, this);
				map.insertMapLink(mapLink, index);
				_objectAssoc.addDepend(mapLink, appObject);
			}
			else if (source == _regionList)
			{
				LyMapRegion mapRegion;

				// Add the node and keep the assoc table up to date
				mapRegion = _mapBehavior.buildRegion(appObject, this);

				if (mapRegion == null)
				{
					_map.setIdle(idle);
					return false;
				}

				LyClassInfo classInfo = appObject.getClassInfo();
				LyInfoLabel label = getInfoLabel(classInfo, LyInfoLabel.Type.SEMANTIC);

				if (label != null)
					mapRegion.setOverview(label.getString(appObject));

				_mapBehavior.updateRegion(appObject, mapRegion, this);
				map.insertMapRegion(mapRegion, index);
				_objectAssoc.addDepend(mapRegion, appObject);
			}
			else
				// Should never happen
				getEnvironment().logWarning(
					"LyMapController does not know how to add (Notification) : " + appObject);

			updateContextClient();
			updateMessages();
		}

		// Update the last modification time.
		setLastModified(new Date().getTime());
		_map.setIdle(idle);

		return true;
	} // end method objectAdded

	/**
	 * Called when an object is removed from the dynamic object list.
	 *
	 * @param event the workspace event
	 * @return true if the demand was take into account
	 * @see    LyWorkSpaceListener#objectRemoved
	 */
	public boolean objectRemoved(LyWorkSpaceEvent event)
	{
		if ((_mapBehavior == null) || (isIdle()))
			return false;

		if (_component == null)
		{
			// Update the last modification time.
			setLastModified(new Date().getTime());
			return true;
		}

		boolean idle = isIdle();

		_map.setIdle(true);

		synchronized (getMapComponent())
		{
			LyWorkSpace source = event.getSource();
			LyObject    appObject = event.getObject();

			// Retrieve internal map component
			LyMap	    map = getMapComponent();

			if (source == _nodeList)
			{
				// Not as simple as stated below : may have impact on links !!
				_map.setIdle(idle);
				reset();
				return true;
			}
			else if (source == _linkList)
			{
				LyMapLink mapLink = (LyMapLink)getObjectAssoc(appObject);

				if (mapLink == null)
				{
					_map.setIdle(idle);
					return false;
				}

				map.removeMapLink(mapLink);

				if ((_selections != null) && (_selections.contains(appObject)))
					_selections.removeObject(appObject);

				_objectAssoc.removeObject(mapLink);
			}
			else if (source == _regionList)
			{
				LyMapRegion mapRegion = (LyMapRegion)getObjectAssoc(appObject);

				if (mapRegion == null)
				{
					_map.setIdle(idle);
					return false;
				}

				map.removeMapRegion(mapRegion);

				if ((_selections != null) && (_selections.contains(appObject)))
					_selections.removeObject(appObject);

				_objectAssoc.removeObject(mapRegion);
			}
			else
				// Should never happen
				getEnvironment().logWarning(
					"LyMapController does not know how to remove (Notification) : " + appObject);

			updateContextClient();
			updateTools();
			updateMessages();
		}

		// Update the last modification time.
		setLastModified(new Date().getTime());

		_map.setIdle(idle);
		return true;
	} // end method objectRemoved

	/**
	 * Called when an object has changed in the dynamic object list.
	 *
	 * @param event the workspace event
	 * @return true if the demand was take into account
	 * @see    LyWorkSpaceListener#objectChanged
	 */
	public boolean objectChanged(LyWorkSpaceEvent event)
	{
		if (isIdle())
			return false;

		if (_component == null)
		{
			// Update the last modification time.
			setLastModified(new Date().getTime());
			return true;
		}

		LyPanelBehaviorInterface behavior = (LyPanelBehaviorInterface)getBehavior();
		LyObject				 object = event.getObject();
		LyValueSet				 values = event.getValues();

		if (!behavior.needRefresh(this, object, values))
			return false;

		synchronized (getMapComponent())
		{
			LyWorkSpace source = event.getSource();
			LyObject    appObject = event.getObject();

			// Retrieve internal map component
			LyMap	    map = getMapComponent();
			boolean     idle = isIdle();

			_map.setIdle(true);

			if (source == _nodeList)
			{
				LyMapNode mapNode = (LyMapNode)getObjectAssoc(appObject);

				if (mapNode == null)
				{
					_map.setIdle(idle);
					return false;
				}
                
				if (values == null)
				{
					// Swap index
		            int         index = event.getIndex();
		            
					map.removeMapNode(mapNode);
					map.insertMapNode(mapNode, index);
				}
				else if (_mapBehavior != null)
					_mapBehavior.updateNode(appObject, mapNode, this);

				LyClassInfo classInfo = appObject.getClassInfo();
				LyInfoLabel label = getInfoLabel(classInfo, LyInfoLabel.Type.SEMANTIC);

				if (label != null)
					mapNode.setOverview(label.getString(appObject));
			}
			else if (source == _linkList)
			{
				LyMapLink mapLink = (LyMapLink)getObjectAssoc(appObject);

				if (mapLink == null)
				{
					_map.setIdle(idle);
					return false;
				}

				if (values == null)
				{
					// Swap index
		            int         index = event.getIndex();
		            
					map.removeMapLink(mapLink);
					map.insertMapLink(mapLink, index);
				}
				else if (_mapBehavior != null)
					_mapBehavior.updateLink(appObject, mapLink, this);

				LyClassInfo classInfo = appObject.getClassInfo();
				LyInfoLabel label = getInfoLabel(classInfo, LyInfoLabel.Type.SEMANTIC);

				if (label != null)
					mapLink.setOverview(label.getString(appObject));
			}
			else if (source == _regionList)
			{
				LyMapRegion mapRegion = (LyMapRegion)getObjectAssoc(appObject);

				if (mapRegion == null)
				{
					_map.setIdle(idle);
					return false;
				}

				if (values == null)
				{
					// Swap index
		            int         index = event.getIndex();
		            
					map.removeMapRegion(mapRegion);
					map.insertMapRegion(mapRegion, index);
				}
				else if (_mapBehavior != null)
					_mapBehavior.updateRegion(appObject, mapRegion, this);

				LyClassInfo classInfo = appObject.getClassInfo();
				LyInfoLabel label = getInfoLabel(classInfo, LyInfoLabel.Type.SEMANTIC);

				if (label != null)
					mapRegion.setOverview(label.getString(appObject));
			}

			_map.setIdle(idle);
		}

		// Update the last modification time.
		setLastModified(new Date().getTime());
		return true;
	} // end method objectChanged

	/**
	 * Called when an object state is changed.
	 *
	 * @param  event Event notifying object state change.
	 * @return True if the event is used by the listener, false otherwise.
	 */
	public boolean objectStateChanged(LyAdvancedWorkSpaceEvent event)
	{
		if (isIdle())
			return false;

		LyObject			  object = event.getObject();
		LyWorkSpaceEvent.Type type = event.getType();

		if (type == LyWorkSpaceEvent.Type.SELECTION_CHANGE)
		{
			LyComponent mapObject = getObjectAssoc(object);

			if (object.isSelected())
				selectObject(mapObject, object, true, true, false);
			else
				selectObject(mapObject, object, false, true, false);
			return true;
		}
		else if (type == LyWorkSpaceEvent.Type.VISIBILITY_CHANGE)
		{
			LyComponent mapObject = getObjectAssoc(object);

			if (object.isHidden())
				selectObject(mapObject, object, false, false, false);

			hideObject(mapObject, object, object.isHidden(), true);

			if (object.isSelected())
				selectObject(mapObject, object, true, true, false);
			else
				selectObject(mapObject, object, false, true, false);
			return true;
		}
		return false;
	} // end method objectStateChanged

	/**
	 * Accessor to the LyObject corresponding to the mapObject.
	 *
	 * @param  mapObject a map Object (view component)
	 * @return the application object or null if it was not found
	 */
	public LyObject getObjectAssoc(LyComponent mapObject)
	{
		Vector<Object> v = _objectAssoc.getDepends(mapObject);

		if ((v != null) && (v.size() > 0))
			return ((LyObject)v.firstElement());
		else
			return null;
	}

	/**
	 * Accessor to the map object for the given LyObject.
	 *
	 * @param  appObject an application object (LyObject instance)
	 * @return the map object or null if it was not found
	 */
	public LyComponent getObjectAssoc(LyObject appObject)
	{
		Vector<Object> v = _objectAssoc.getDepends(appObject);

		if ((v != null) && (v.size() > 0))
			return ((LyComponent)v.firstElement());
		else
			return null;
	}

	/**
	 * May add workspace to registration.
	 *
	 * @param objectList a list of object to register
	 * @return the registered workspace
	 */
	protected LyWorkSpace registerWorkSpace(LyObjectList objectList)
	{
		if (objectList == null)
			return null;

		LyWorkSpace workspace;

		if (objectList.isDynamic())
			workspace = (LyWorkSpace)objectList;
		else
			workspace = new LySimpleWorkSpace(getSession(), objectList, true);

		// Register
		workspace.addListener(this);
		return workspace;
	}

	/**
	 * Accessor to the node list.
	 *
	 * @return the list of nodes for the map
	 */
	public LyObjectList getNodeList()
	{
		return (_nodeList);
	}

	/**
	 * Accessor to the Links list.
	 *
	 * @return the list of links for the map
	 */
	public LyObjectList getLinkList()
	{
		return (_linkList);
	}

	/**
	 * Accessor to the Regions list.
	 *
	 * @return the list of regions for the map
	 */
	public LyObjectList getRegionList()
	{
		return (_regionList);
	}

	/**
	 * Accessor to the Map component.
	 *
	 * @return the map component
	 */
	public LyMap getMapComponent()
	{
		// Retrieve internal map component
		// Using XML template file : search without Id.
		if (_map == null)
		{
		    LyComponent    c = getComponent();
		    
    		if (c != null)
    		    _map = (LyMap)c.getComponent("_map" + getId());
		}
		
		return _map;
	}

	/**
	 * Accessor to panel component.
	 *
	 * @return the panel component
	 * @see LyMapController#getMapComponent
	 */
	public LyPanel getPanelComponent()
	{
		return getMapComponent();
	}

	/**
	 * Retrieves the default view Builder.
	 *
	 * @return a new instance of a LyMapBuilder
	 * @see    LyController#getDefaultBuilder
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return (new LyMapBuilder(this));
	}

	/**
	 * Sets visible status of map component.
	 *
	 * @param visible a boolean indicating the visible mode
	 */
	public void setVisible(boolean visible)
	{
		// Retrieve internal map component
		LyMap map = getMapComponent();

		if (map != null)
			map.setVisible(visible);
	}

	/**
	 * Class info : content description. (application class)
	 *
	 * @return the classInfo for this map
	 */
	public LyClassInfo getClassInfo()
	{
		return _classInfo;
	}

	/**
	 * Invoke default action on an object.
	 * @param object  the object on which the default action must apply
	 *
	 * @return the controller that take care of the default action
	 */
	public LyController invokeDefaultAction(LyObject object)
	{
		if (_classInfos != null)
		{
			int index = _classInfos.indexOf(_classInfo);

			index++;

			if (index < _classInfos.size())
			{
				_contents.push(_classInfo);
				setVisible(true);
				_classInfo = _classInfos.elementAt(index);

				LyWorkSpace newContext = new LySimpleWorkSpace(getSession(), object, true);

				boolean     needReset = false;

				if (_context != null)
				{
					_context.lock();

					/// FFT 810 : even if context unchanged, reset !
					if (newContext.equals(_context))
						needReset = true;
				}

				_contexts.push(_context);

				setContext(newContext, false);

				// FFT 810 Force reset if context content is unchanged, however we are one level deeped
				if (needReset)
					reset();

				return null;
			}
		}

		return super.invokeDefaultAction(object);
	} // end method invokeDefaultAction

	/**
	 * This method reproduced the behavior of the user when he clicks on the given object.
	 *
	 * @param object a LyObject: we show the object under it
	 */
	public void setLevel(LyObject object)
	{
		if (_classInfos != null)
		{
			int index = _classInfos.indexOf(object.getClassInfo());

			if (index >= 0)
			{
				index++;

				if (index < _classInfos.size())
				{
					_contents.push(_classInfo);
					setVisible(true);
					_classInfo = _classInfos.elementAt(index);
					if (_context != null)
						_context.lock();

					_contexts.push(_context);
					setContext(new LySimpleWorkSpace(getSession(), object, true), false);

				}
			}
		}
	}

	/**
	 * Set current controller context.
	 *
	 * @param context    Current context, or null if no context is set for the view.
	 * @param clearStack Boolean indicating if contexts stack must be cleared or not.
	 * @return true if successfull
	 */
	protected boolean setContext(LyWorkSpace context, boolean clearStack)
	{
		if (_action.hasMark(LyAction.MARK_NO_CONTEXT))
			return false;

		synchronized (getMapComponent())
		{
			if (clearStack && (_contexts != null))
			{
				_classInfo = _classInfos.elementAt(0);

				// Remove contexts from stack
				LyWorkSpace stackContext;

				while (!_contexts.isEmpty())
				{
					stackContext = _contexts.pop();
					stackContext.unlock();
				}
			}

			return super.setContext(context);
		}
	}

	/**
	 * Set current controller context.
	 *
	 * @param context Current context, or null if no context is set for the view
	 * @return true if successfull
	 * @see   LyContextInterface#setContext
	 */
	public boolean setContext(LyWorkSpace context)
	{
		if (_action.hasMark(LyAction.MARK_NO_CONTEXT))
			return false;

		return setContext(context, true);
	}

	/**
	 * Print the content of the map Run it asynchronous.
	 */
	public void printMap()
	{
		LyViewManager viewManager = getViewManager();

		viewManager.execGui(this, false);
	}

	/**
	 * Starting printMap request.
	 */
	public void run()
	{
		startRun();

		// Retrieve internal map component
		LyMap map = getMapComponent();

		map.print(this);

		endRun();
	}

	/**
	 * Implementation of LyPrintListenerInterface.
	 * @param page the page to print
	 */
	public void print(int page)
	{
	}

	/**
	 * Clears all cached data for this controller (especially the view and the component).
	 *
	 * <P>This method may be invoked for memory optimization, but the controller is still valid and
	 * may be used. (Unlike free) Further usage, needs to rebuild internal component & view which
	 * may be time consuming.</P>
	 */
	public void clearViewCache()
	{
		_objectAssoc = new LyDependencies();
		_lockedMapObjects = null;
		_selections = null;
		_contents = new Stack<Object>();
		_map = null;

		super.clearViewCache();
	}

	/**
	 * Method that indicates if controlled panel enables multi-selection or not.
	 *
	 * @return A boolean indicating if multi-selection is allowed or not.
	 */
	public boolean enablesMultiSelection()
	{
		LyMap map = getMapComponent();

		return ((map == null) ? true : (map.getSelectionMode() != LyMap.SelectionMode.SINGLE));
	}
} // end class LyMapController
