/*
 * LyPageSetupBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package eu.w4.ae.control.builder;


import java.text.*;
import java.util.*;

import eu.w4.ae.control.print.xslfo.*;


import leon.control.*;
import leon.control.builder.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;



/**
 * This is the builder for the page setup view.
 *
 * @version $Id: LyPageSetupBuilder.java 28702 2010-11-26 08:57:03Z bangiole $
 * @author  Lyria
 */
public final class LyPageSetupBuilder extends LyUserInputBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyPageSetupBuilder.java 28702 2010-11-26 08:57:03Z bangiole $";

	/** Top margin */
	public static double	   MARGIN_TOP_MAX = 10.0;
	/** Bottom margin */
	public static double	   MARGIN_BOTTOM_MAX = 10.0;
	/** Left margin */
	public static double	   MARGIN_LEFT_MAX = 10.0;
	/** Right margin */
	public static double	   MARGIN_RIGHT_MAX = 10.0;
	/** increment */
	public static double	   MARGIN_INCREMENT = 0.1;
	/** Nb of horizontal pages */
	public static double	   GANTT_PAGES_HORIZONTAL_MAX = 100.0;
	/** Nb of vertical pages */
	public static double	   GANTT_PAGES_VERTICAL_MAX = 100.0;

	/** A set of tabs when multiple view are printed in the same document (multiple properties). */
	private LyTabs			   _tabs;

	/**
	 * Contructor.
	 *
	 * @param parent The parent controller.
	 */
	public LyPageSetupBuilder(LyController parent)
	{
		super(parent);
		if (!(parent instanceof LyPageSetupController))
			throw new IllegalStateException(
				"BF AppEngine : invalid parent controller for LyPageSetupBuilder : "
				+ parent + "\n A leon.control.print.xslfo.LyPageSetupController is expected.");
		if (_form != null)
			_form.setBorder(false);
	}

	/**
	 * Return the form component.
	 *
	 * @return the form component
	 */
	public LyForm getFormComponent()
	{
		return _form;
	}

	/**
	 * Return the tabs component.
	 *
	 * @return the tabs component
	 */
	public LyTabs getTabsComponent()
	{
		return _tabs;
	}

	/**
	 * Add a form field that contains all the widgets for the selection of the properties of a page
	 * setup before printing.
	 *
	 * @param properties Set of initial properties.
	 * @param formId     Identifier of the form created (identifier of the related print
	 *                   controller).
	 * @param label      Label displayed for this form.
	 * @param isGantt    tells if the printed component is a gantt.
	 */
	public void addPropertiesForm(Properties properties, String formId, String label,
		boolean isGantt)
	{
		LyTab  tab = null;
		LyArea area = null;

		if (label != null)
		{
			if (_tabs == null)
			{
				area = _form.getParentArea();
				_tabs = new LyTabs();
				area.addPanel(_tabs);
			}

			tab = new LyTab();
			tab.setLabel(label);
			_tabs.addTab(tab);
		}
		else
			area = _form.getParentArea();

		if (tab != null)
		{
			area = new LyArea();
			area.setBorder(LyArea.Border.NONE);
			tab.setPanel(area);
		}

		area.setId(formId);

		// create form fields for page format
		createPageFormatFields(properties, area);

		// create the margin fields
		createPageMarginFields(properties, area);

		// create the font fields
		createPageFontFields(properties, area, (label == null), true,
			"LY_PAGE_SETUP_FONT_PROPERTIES");

		// add the specific gantt printing properties
		if (isGantt)
			createGanttFields(area);
		
		// Not in web mode
		if (getViewManager().hasProperty("LOCAL_FILE"))
			createPdfPreferencesFields(area);
		
	} // end method addPropertiesForm

	/**
	 * Create a form for the general properties of a view : the body font & the font size.
	 *
	 * @param properties
	 * @param formId
	 */
	public void addGeneralPropertiesForm(Properties properties, String formId)
	{
		LyViewManager viewManager = getViewManager();
		LyArea		  area;
		String		  title = "LY_PAGE_SETUP_FONT_PROPERTIES";

		if (!(viewManager.hasProperty("MULTI_FRAME_FORM")))
		{
			LyTab tab;

			if (_tabs == null)
			{
				area = _form.getParentArea();
				_tabs = new LyTabs();
				area.addPanel(_tabs);
			}

			tab = new LyTab();
			tab.setLabel("");
			_tabs.addTab(tab);
			area = new LyArea();
			area.setBorder(LyArea.Border.NONE);
			tab.setPanel(area);
			title = "LY_PAGE_SETUP_GLOBAL_FONT_PROPERTIES";
		}
		else
			area = _form.getParentArea();

		area.setId(formId);

		// create the font fields
		createPageFontFields(properties, area, true, false, title);
	} // end method addGeneralPropertiesForm

	/**
	 * Create the fields for the page format properties.
	 *
	 * @param properties Set of initial properties.
	 * @param parentArea The area in which fields are added.
	 */
	public void createPageFormatFields(Properties properties, LyArea parentArea)
	{
		// build a sub form
		LyArea area = new LyArea();

		area.setBorder(LyArea.Border.IN);
		area.setLabel("LY_PAGE_SETUP_PAGE_PROPERTIES");

		LyForm subForm = new LyForm("_pageFormatForm");

		subForm.setBorder(false);
		area.addPanel(subForm);

		// temporary store the main form
		LyForm mainForm = _form;

		_form = subForm;

		// build the field for the paper type
		createRelationField(parentArea.getId() + LyPageSetupController.PAPER_TYPE,
			"LY_PAGE_SETUP_PAPER_TYPE", LyPageSetupController.getAvailablePaperType(),
			properties.getProperty(LyPageSetupController.PAPER_TYPE));

		// build the field for the orientation
		createChoiceField(parentArea.getId() + LyPageSetupController.PAGE_ORIENTATION,
			"LY_PAGE_SETUP_PAGE_ORIENTATION", LyPageSetupController.getAvailablePageOrientation(),
			properties.getProperty(LyPageSetupController.PAGE_ORIENTATION));

		// set the main form
		_form = mainForm;
		parentArea.addPanel(area);
	} // end method createPageFormatFields

	/**
	 * Create the fields for the page margin properties.
	 *
	 * @param properties Set of initial properties.
	 * @param parentArea The area in which fields are added.
	 */
	public void createPageMarginFields(Properties properties, LyArea parentArea)
	{
		// build a sub form
		LyArea area = new LyArea();

		area.setBorder(LyArea.Border.IN);
		area.setLabel("LY_PAGE_SETUP_MARGIN_PROPERTIES");

		LyForm subForm = new LyForm("_pageMarginForm");

		subForm.setBorder(false);
		area.addPanel(subForm);

		// temporary store the main form
		LyForm mainForm = _form;

		_form = subForm;

		// build the margin top
		String		 value = properties.getProperty(LyPageSetupController.PAGE_MARGIN_TOP);
		LyNumber     number = createNumberField(parentArea.getId()
				+ LyPageSetupController.PAGE_MARGIN_TOP, "LY_PAGE_SETUP_PAGE_MARGIN_TOP",
				extractMarginValue(value), LyPageSetupController.PAGE_MARGIN_UNIT, 0.0,
				MARGIN_TOP_MAX, MARGIN_INCREMENT, 1);
		LyConstraint constraint = new LyConstraint();

		constraint.setRemainder(false);
		constraint.setColSpan(1);
		((LyFormField)number.getParent()).setConstraint(constraint);

		//      build the margin left
		value = properties.getProperty(LyPageSetupController.PAGE_MARGIN_LEFT);
		createNumberField(parentArea.getId() + LyPageSetupController.PAGE_MARGIN_LEFT,
			"LY_PAGE_SETUP_PAGE_MARGIN_LEFT", extractMarginValue(value),
			LyPageSetupController.PAGE_MARGIN_UNIT, 0.0, MARGIN_LEFT_MAX, MARGIN_INCREMENT, 1);

		// build the margin bottom
		value = properties.getProperty(LyPageSetupController.PAGE_MARGIN_BOTTOM);
		number = createNumberField(parentArea.getId() + LyPageSetupController.PAGE_MARGIN_BOTTOM,
				"LY_PAGE_SETUP_PAGE_MARGIN_BOTTOM", extractMarginValue(value),
				LyPageSetupController.PAGE_MARGIN_UNIT, 0.0, MARGIN_BOTTOM_MAX, MARGIN_INCREMENT,
				1);
		((LyFormField)number.getParent()).setConstraint(constraint);

		// build the margin right
		value = properties.getProperty(LyPageSetupController.PAGE_MARGIN_RIGHT);
		createNumberField(parentArea.getId() + LyPageSetupController.PAGE_MARGIN_RIGHT,
			"LY_PAGE_SETUP_PAGE_MARGIN_RIGHT", extractMarginValue(value),
			LyPageSetupController.PAGE_MARGIN_UNIT, 0.0, MARGIN_RIGHT_MAX, MARGIN_INCREMENT, 1);

		// set the main form
		_form = mainForm;
		parentArea.addPanel(area);
	} // end method createPageMarginFields

	/**
	 * Create the fields for the page font properties.
	 *
	 * @param properties Set of initial properties.
	 * @param parentArea The area in which fields are added.
	 * @param body       tells if the body and sizes properties is added.
	 * @param title      tells if the tile property is added.
	 * @param areaTitle  title of the area containing the fields.
	 */
	public void createPageFontFields(Properties properties, LyArea parentArea, boolean body,
		boolean title,
		String areaTitle)
	{
		// build a sub form
		LyArea area = new LyArea();

		area.setBorder(LyArea.Border.IN);
		area.setLabel(areaTitle);

		LyForm subForm = new LyForm("_pageFontForm");

		subForm.setBorder(false);
		area.addPanel(subForm);

		// temporary store the main form
		LyForm mainForm = _form;

		_form = subForm;

		if (body)
		{
			// build the body font family
			LyRelation   relation = createRelationField(parentArea.getId()
					+ LyPageSetupController.FONT_BODY_FAMILY, "LY_PAGE_SETUP_FONT_BODY_FAMILY",
					LyPageSetupController.getAvailableFonts(),
					properties.getProperty(LyPageSetupController.FONT_BODY_FAMILY));
			LyConstraint constraint = new LyConstraint();

			constraint.setRemainder(false);
			constraint.setColSpan(1);
			((LyFormField)relation.getParent()).setConstraint(constraint);

			// build the font size top
			String value = properties.getProperty(LyPageSetupController.FONT_SIZE);

			createNumberField(parentArea.getId() + LyPageSetupController.FONT_SIZE,
				"LY_PAGE_SETUP_FONT_SIZE", value, null, 6.0, 72.0, 1.0, 0);
		}

		if (title)
			// build the title font family
			createRelationField(parentArea.getId() + LyPageSetupController.FONT_TITLE_FAMILY,
				"LY_PAGE_SETUP_FONT_TITLE_FAMILY", LyPageSetupController.getAvailableFonts(),
				properties.getProperty(LyPageSetupController.FONT_TITLE_FAMILY));

		// set the main form
		_form = mainForm;
		parentArea.addPanel(area);
	} // end method createPageFontFields

	/**
	 * Create the fields for the gantt printing properties.
	 *
	 * @param parentArea The area in which fields are added.
	 */
	public void createGanttFields(LyArea parentArea)
	{
		// build a sub form
		LyArea area = new LyArea();

		area.setBorder(LyArea.Border.IN);
		area.setLabel("LY_PAGE_SETUP_GANTT_PROPERTIES");

		LyForm subForm = new LyForm("_ganttForm");

		subForm.setBorder(false);
		subForm.setController(_parent.getId());
		area.addPanel(subForm);

		// temporary store the main form
		LyForm mainForm = _form;

		_form = subForm;

		// build the field for the type of the printing (identical to screen view)
		LyBoolean    booleanC = createBooleanField(parentArea.getId()
				+ LyPageSetupController.GANTT_PRINTING_TYPE, "LY_PAGE_SETUP_GANTT_PRINTING_TYPE",
				LyBoolean.Status.UNCHECKED, "LY_PAGE_SETUP_GANTT_PRINTING_TYPE_USER",
				"LY_PAGE_SETUP_GANTT_PRINTING_TYPE_SCREEN", null, false);

		LyConstraint constraint = new LyConstraint();

		constraint.setRemainder(true);
		constraint.setColSpan(3);
		((LyFormField)booleanC.getParent()).setConstraint(constraint);
		((LyFormField)booleanC.getParent()).setNeedCheck(true);

		// build the fields for the end and startdate
		LyDate date = createDateField(parentArea.getId() + LyPageSetupController.GANTT_DATE_BEGIN,
				"LY_PAGE_SETUP_GANTT_BEGIN_DATE", null, LyEnvironment.TYPE_DATE_TIME);

		constraint = new LyConstraint();
		constraint.setRemainder(false);
		constraint.setColSpan(1);

		((LyFormField)date.getParent()).setEnabled(false);
		((LyFormField)date.getParent()).setConstraint(constraint);

		date = createDateField(parentArea.getId() + LyPageSetupController.GANTT_DATE_END,
				"LY_PAGE_SETUP_GANTT_END_DATE", null, LyEnvironment.TYPE_DATE_TIME);
		((LyFormField)date.getParent()).setEnabled(false);

		// build the fields for the number of pages
		LyNumber number = createNumberField(parentArea.getId()
				+ LyPageSetupController.GANTT_PAGES_HORIZONTAL,
				"LY_PAGE_SETUP_GANTT_PAGES_HORIZONTAL", "1", null, 1.0, GANTT_PAGES_HORIZONTAL_MAX,
				1.0, 0);

		constraint = new LyConstraint();
		constraint.setRemainder(false);
		constraint.setColSpan(1);

		((LyFormField)number.getParent()).setEnabled(false);
		((LyFormField)number.getParent()).setConstraint(constraint);

		number = createNumberField(parentArea.getId() + LyPageSetupController.GANTT_PAGES_VERTICAL,
				"LY_PAGE_SETUP_GANTT_PAGES_VERTICAL", "1", null, 1.0, GANTT_PAGES_VERTICAL_MAX, 1.0,
				0);
		((LyFormField)number.getParent()).setEnabled(false);

		// set the main form
		_form = mainForm;
		parentArea.addPanel(area);
	} // end method createGanttFields

	
	/**
	 * Create the fields for the output preferences.
	 * @param parentArea The area in which fields are added.
	 */
	public void createPdfPreferencesFields(LyArea parentArea)
	{
		// build a sub form
		LyArea area = new LyArea();

		area.setBorder(LyArea.Border.IN);
		area.setLabel("LY_PDF_PREFERENCES");

		LyForm subForm = new LyForm("_pdfPreferencesForm");

		subForm.setBorder(false);
		subForm.setController(_parent.getId());
		area.addPanel(subForm);

		// temporary store the main form
		LyForm mainForm = _form;

		_form = subForm;

		// build the boolean field for showing (or not) the PDF.
		LyBoolean    booleanC = createBooleanField("_pdfPreferencesFormShowPDF" , "LY_SHOW_PDF_CAPTION",
				LyBoolean.Status.CHECKED, "LY_SHOW_PDF_YES",
				"LY_SHOW_PDF_NO", null, false);
		((LyFormField)booleanC.getParent()).setEnabled(true);
		
		// Set the output file for pdf if you don't want to use the default output
		LyFile    fileC = createFileSelectorField("_pdfPreferencesFormOutputFile");
		((LyFormField)fileC.getParent()).setEnabled(true);
		((LyFormField)fileC.getParent()).setOptional(true);
		
		_form = mainForm;
		parentArea.addPanel(area);
	} // end method createGanttFields

	/**
	 * Create a file selector field
	 * 
	 * @param id the id of the field
	 * @return a file field
	 */
	protected LyFile createFileSelectorField(String id)
		{

			//	add the file text to the field
			LyFile fieldFile = new LyFile();

			LyFormField field = new LyFormField(id);
			
			LyCaption captionComponent = new LyCaption();
			captionComponent.setLabel("LY_FILE");
			captionComponent.setColon(true);

			field.setCaption(captionComponent);
			field.setFieldItem(fieldFile);
			field.setOptional(false);

			LyTip tip = new LyTip();
			tip.setLabel("LY_EXPORT_FILE_TIP");
			field.setTip(tip);


			//add the field in the form
			if (_form != null)
				_form.addFormField(field);

			return fieldFile;
		} // end method createBooleanField
	
	/**
	 * Change the enabled status of the printing properties for a gantt component.
	 *
	 * @param typeFieldId Identifier of the 'GANTT_PRINTING_TYPE' field.
	 * @param status      new enabled status of gantt properties
	 */
	public void setGanttPropertiesEnabled(String typeFieldId, boolean status)
	{
		// get the form field of the given type
		LyFormField typeField = (LyFormField)getComponent().getComponent(typeFieldId);

		// get the form of the given form field
		LyForm	    form = (LyForm)typeField.getParent();

		// change the enabled status of the fields
		LyFormField field;

		for (int i = 0; i < form.getFormFieldCount(); i++)
		{
			field = form.getFormField(i);
			if (field != typeField)
				field.setEnabled(status);
		}
	}

	/**
	 * Extract the value from a margin property (remove the unit part of the property).
	 *
	 * @param  value a margin property value.
	 * @return a number value.
	 */
	protected String extractMarginValue(String value)
	{
		if (value == null)
			return null;

		int idx = value.indexOf(LyPageSetupController.PAGE_MARGIN_UNIT);

		if (idx >= 0)
			return value.substring(0, idx);
		else
			return value;
	}

	/**
	 * Create a choice field that is added to the current form.
	 *
	 * @param  id           idenfier of the form field.
	 * @param  caption      label of the form field.
	 * @param  options      list of available options.
	 * @param  initialValue initial selected value.
	 * @return the created field item.
	 */
	protected LyChoice createChoiceField(String id, String caption, String[][] options,
		String initialValue)
	{
		LyChoice choice = addChoice(id, caption, false, true);
		LyOption option;

		for (int i = 0; i < options.length; i++)
		{
			option = new LyOption(options[i][0], options[i][1]);
			if (options[i][0].equals(initialValue))
				option.setChecked(true);
			choice.addOption(option);
		}
		return choice;
	}

	/**
	 * Create a relation field that is added to the current form.
	 *
	 * @param  id           idenfier of the form field.
	 * @param  caption      label of the form field.
	 * @param  options      list of available options.
	 * @param  initialValue initial selected value.
	 * @return the created field item.
	 */
	protected LyRelation createRelationField(String id, String caption, String[][] options,
		String initialValue)
	{
		LyRelation relation = addRelation(id, caption, false, true);
		LyLink     link;

		for (int i = 0; i < options.length; i++)
		{
			link = new LyLink(options[i][0], options[i][1]);
			if (options[i][0].equals(initialValue))
				link.setSelected(true);
			relation.addLink(link);
		}
		return relation;
	}

	/**
	 * Create a number field that is added to the current form.
	 *
	 * @param  id           idenfier of the form field.
	 * @param  caption      label of the form field.
	 * @param  initialValue initial selected value.
	 * @param  unit         unit of the field.
	 * @param  min          minimum value of the field.
	 * @param  max          maximum value of the field.
	 * @param  pitch        step of selection fo the component.
	 * @param  decimals     the number of decimals
	 * @return the created field item.
	 */
	protected LyNumber createNumberField(String id, String caption, String initialValue,
		String unit, double min, double max, double pitch, int decimals)
	{
		LyNumber number = addNumber(id, caption, initialValue,
				(decimals == 0) ? LyNumberFieldInfo.Type.INT : LyNumberFieldInfo.Type.DOUBLE, true);

		if (unit != null)
			number.setUnit(unit);
		number.setPitch(pitch);
		number.setVMin(min);
		number.setVMax(max);
		number.setDecimals(decimals);
		return number;
	}

	/**
	 * Create a boolean field that is added to the current form.
	 *
	 * @param  id             idenfier of the form field.
	 * @param  caption        label of the form field.
	 * @param  status         status of the boolean component : LyBoolean.Status.CHECKED,
	 *                        LyBoolean.Status.UNCHECKED or LyBoolean.Status.EMPTY
	 * @param  labelChecked   label used when component is in checked state.
	 * @param  labelUnchecked label used when component is in unchecked state.
	 * @param  labelEmpty     label used when component is in empty state.
	 * @param  not            indicates if the state is inverted
	 * @return the created field item.
	 */
	protected LyBoolean createBooleanField(String id, String caption,
		LyBoolean.Status status, String labelChecked, String labelUnchecked,
		String labelEmpty, boolean not)
	{
		// create the form field
		LyFormField field = new LyFormField(id);

		field.setOptional(false);

		// create the caption
		LyCaption captionComponent = new LyCaption();

		captionComponent.setLabel(caption);
		captionComponent.setColon(true);

		// create the boolean field
		LyBoolean booleanComponent = new LyBoolean();

		booleanComponent.setStatus(status);
		booleanComponent.setLabelChecked(labelChecked);
		booleanComponent.setLabelUnchecked(labelUnchecked);
		booleanComponent.setLabelEmpty(labelEmpty);
		booleanComponent.setNot(not);

		// complete the field
		field.setCaption(captionComponent);
		field.setFieldItem(booleanComponent);

		//add the field in the form
		if (_form != null)
			_form.addFormField(field);

		return booleanComponent;
	} // end method createBooleanField

	/**
	 * Create a date field
	 * @param id the id of the field
	 * @param caption the caption
	 * @param date the date value
	 * @param type the date type
	 * @return the date field item
	 */
	protected LyDate createDateField(String id, String caption, Date date, short type)
	{
		// get the date format
		LyEnvironment    env = _parent.getEnvironment();
		SimpleDateFormat format = (SimpleDateFormat)env.dateFormatFactory(type);

		String			 value = (date == null) ? null : format.format(date);

		LyDate			 dateComponent = addDate(id, caption, value, format.toPattern(), true);

		return dateComponent;
	}
} // end class LyPageSetupBuilder
