/**
 * LyApplication.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.app;

import java.io.*;
import java.util.*;

import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.data.*;
import leon.data.cache.*;
import leon.data.daemon.*;
import leon.info.*;
import leon.info.binding.*;
import leon.info.daemon.*;
import leon.info.infointerface.*;
import leon.info.location.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.extend.*;
import leon.peer.info.*;
import leon.peer.local.*;
import leon.peer.sql.*;
import leon.peer.union.*;


/**
 * This class contains all the data that are described in a meta-model. Several applications may be
 * launched in the same process using different meta-models (see leon.app.LyMain to see how to
 * create several applications). For each of them, several sessions may be created using different
 * environments. An application is described by :
 *
 * <UL>
 * <LI><B>an identifier</B><BR>
 * used by the environment of the application.</LI>
 * <LI><B>a list of sessions</B><BR>
 * the sessions registered in the application.</LI>
 * <LI><B>a root project</B><BR>
 * defined in the meta-model by the root tag (<CODE>&lt;PROJECT&gt;</CODE>).</LI>
 * <LI><B>a list of infos</B><BR>
 * these infos are objects such as actions, fields, filters, sorts...</LI>
 * <LI><B>a list of classes</B><BR>
 * the classes used in the application (these classes are defined in the meta-model).</LI>
 * <LI><B>a list of data sources</B><BR>
 * the data sources corresponding to the classes mentioned above.</LI>
 * <LI><B>a list of data providers</B><BR>
 * the data providers corresponding to the classes mentioned above.</LI>
 * <LI><B>a list of notifiers</B><BR>
 * the notifiers used by the application.</LI>
 * </UL>
 *
 * @author  Lyria
 * @see     leon.app.LyMain
 * @see     leon.app.LyDataSession
 */
public class LyApplication
    implements LyTimerClientInterface
{
    /** File version. */
    public static final String __VERSION = "$Id: LyApplication.java 41739 2012-12-07 10:41:26Z bangiole $";

    /** List of applications registered in the current process. */
    static Hashtable<String, Vector<LyApplication>> _applications;

    /** List of listeners */
    static List<LyApplicationListener> _listeners;

    /** Separator used by LyObjectId class. */
    public static char FIELD_SEP = ',';

    /** Separator used by LyObjectId class. */
    public static char VISIBLE_FIELD_SEP = ',';

    /** Separator used by LyObjectId class. */
    public static char CLASS_SEP = '\\';

    /** Separator used for composed title. */
    public static String LABEL_SEP = " : ";

    /** Separator used when multiple values are returned from the graphical input object. */
    public char VALUE_SEP = (char)4;

    /** Separator used by LyObjectId class. */
    public static char VISIBLE_VALUE_SEP = '|';

    /** Separator for objects with several names. */
    public char NAME_SEP = ' ';

    /** Internal separator used for array values. */
    public static char ARRAY_SEP = (char)5;

    /** Visible separator used for array values. */
    public static char VISIBLE_ARRAY_SEP = '\n';

    /** Visible separator used for vector values. */
    public static String VISIBLE_VECTOR_SEP = ";";

    /** Key for current data session. */
    static final String CURRENT_DATA_SESSION = "_currentDataSession";

    /** Key for current context. */
    static final String CURRENT_CONTEXT = "_currentContext";

    /** Key for current context. */
    static final String CURRENT_EVALUATOR = "_currentEvaluator";

    /**
     * Parent application : used when an application is a module. The primary module doesn't have
     * any parent application.
     */
    protected LyApplication _parent;

    /** Identifier of the application. */
    protected String _id;

    /** Identifier of the default location for this application. */
    protected String _defaultLocationId;

    /** List of sessions of the application. */
    protected Vector<LyDataSession> _sessions;

    /** Environment of the application. */
    protected LyEnvironment _environment;

    /** Project of the application. */
    protected LyProject _project;

    /** List of data providers used by the application. */
    protected LyDataProvider[] _providers;

    /**
     * Table of LyInfo used by the application. These instances may be instances of LyFieldInfo,
     * LyAction...
     */
    protected Hashtable<String, LyTag> _tagTable;

    /**
     * List of LyInfo used by the application. These instances may be instances of LyFieldInfo,
     * LyAction...
     */
    protected LyInfoList<LyInfo> _infoList;

    /** List of LyClassInfo used by the application. */
    protected LyInfoList<LyClassInfo> _classInfos;

    /**
     * Info provider of the application. This instance is used by data sources to transmit their
     * requests to an instance of RequestEncoder.
     */
    protected LyInfoProvider _infoProvider;

    /**
     * Union provider of the application. This instance is used by data sources to transmit their
     * requests to an instance of RequestEncoder for union mapping.
     */
    protected LyUnionDataProvider _unionProvider;

    /**
     * Extend provider of the application. This instance is used by data sources to transmit their
     * requests to an instance of RequestEncoder for extended classes.
     */
    protected LyExtendDataProvider _extendProvider;

    /**
     * Extend provider of the application. This instance is used by data sources to transmit their
     * requests to an instance of RequestEncoder for extended classes.
     */
    protected LyLocalDataProvider _localProvider;

    /**
     * Indicates if the application is asleep.
     *
     * @see  LyApplication#isSleeping()
     * @see  LyApplication#sleep()
     * @see  LyApplication#wakeUp()
     */
    protected boolean _isAsleep;

    /** Specific behavior of the application. */
    protected LyApplicationBehaviorInterface _behavior;

    /**
     * Indicates whether the application is valid or not. An application is valid if it has been
     * started.
     */
    protected boolean _valid;

    /** Daemons used by the application. */
    protected Hashtable<LyInfo, LyDaemon> _daemons;

    /**
     * Error message of the application. It is used when starting the application to get an
     * explanation of what happened in case of error.
     */
    protected String _error;

    /** Table of data sources used in the application corresponding to class id. */
    protected Hashtable<String, LyDataSource> _dataSources;

    /** Table of instances associated to Threads. */
    protected LyThreadData<Object> _threadInstances;

    /** Transaction manager of the application. */
    protected LyTransactionManager _transactionManager;

    /** Event manager of the application. */
    protected LyEventManager _eventManager;

    /** List of sub-applications of the application. */
    Vector<LyApplication> _subApplications;

    /** Timer thread for sleepingApp/closingApp. */
    protected LyTimer _closeTimer;

    /** Idle mode : no event while this mode is set. */
    protected boolean _idle;

    /** Closing state. */
    protected boolean _closing;

    /** Associations between keys and values used to valuate contextual values. */
    protected HashMap<Object, Object> _values;

    /** Edited state used when application is under edition in application composer. */
    protected boolean _edited;

    /** Session listeners. */
    protected Vector<LySessionListenerInterface> _sessionListeners;


    static
    {
        _applications = new Hashtable<String, Vector<LyApplication>>();
        _listeners = new ArrayList<LyApplicationListener>();
    }

    /**
     * Registers an application.
     *
     * @param  application  The application that must be registered.
     */
    public static void putApplication(LyApplication application)
    {
        synchronized (_applications)
        {
            String                id = application.getId();
            Vector<LyApplication> applications = _applications.get(id);

            if (applications == null)
                applications = new Vector<LyApplication>(1);

            applications.addElement(application);
            _applications.put(id, applications);

            Iterator<LyApplicationListener> it = _listeners.iterator();
            LyApplicationListener listener;
            while (it.hasNext())
            {
            	listener = it.next();
            	listener.applicationAdded(application);
            }
        }
    }

    /**
     * Retrieves an application.
     *
     * @param   id  The id of the application that must be retrieved.
     *
     * @return  The corresponding application or null if there is no application that matches the
     *          given id.
     */
    public static LyApplication getApplication(String id)
    {
        synchronized (_applications)
        {
            Vector<LyApplication> applications = _applications.get(id);

            if ((applications != null)
                    && (applications.size() > 0))
                return applications.elementAt(0);

            return null;
        }
    }

    /**
     * Retrieves an application from an info id.
     *
     * @param   id  The id of the info for which the application is searched.
     * @return  The corresponding application or null if there is no application that matches the
     *          given info id.
     */
    public static LyApplication getInfoApplication(String infoId)
    {
        // Try to find an info corresponding to given id
    	Vector<LyApplication> applications = getApplications();
        int				i, n = applications.size();
        LyApplication	application;
        LyInfo			info;

        for (i = 0; i < n; i++)
        {
        	application = applications.get(i);
        	info = application.getInfo(infoId);
        	if (info != null)
        		return application;
        }

        return null;
    }

    /**
     * Removes an application.
     *
     * @param  application  The application that must be removed.
     */
    public static void removeApplication(LyApplication application)
    {
    	if (_applications == null)
    		return;

        synchronized (_applications)
        {
            String id = application.getId();

            if (id != null)
            {
                Vector<LyApplication> applications = _applications.get(id);

                if ((applications != null)
                        && applications.contains(application))
                {
                    applications.remove(application);

                    if (applications.size() == 0)
                        _applications.remove(id);

                    Iterator<LyApplicationListener> it = _listeners.iterator();
                    LyApplicationListener listener;
                    while (it.hasNext())
                    {
                    	listener = it.next();
                    	listener.applicationRemoved(application);
                    }
                }
            }
        }
    }

    /**
     * Remove all applications.
     */
    public static void removeAll()
    {
        synchronized (_applications)
        {
            _applications = new Hashtable<String, Vector<LyApplication>>();
        }
    }

    /**
     * Gets list of all applications.
     *
     * @return  List of applications.
     */
    public static Vector<LyApplication> getApplications()
    {
        synchronized (_applications)
        {
            Enumeration<Vector<LyApplication>> elements = _applications.elements();
            Vector<LyApplication>              result = new Vector<LyApplication>();
            Vector<LyApplication>              applications;
            int                                i, n;

            while (elements.hasMoreElements())
            {
                applications = elements.nextElement();
                n = ((applications == null) ? 0 : applications.size());

                for (i = 0; i < n; i++)
                    result.add(applications.elementAt(i));
            }

            return result;
        }
    }

    /**
     * Registers a listener.
     *
     * @param  listener  The listener that must be registered.
     */
    public static void addListener(LyApplicationListener listener)
    {
        synchronized (_applications)
        {
            _listeners.add(listener);
        }
    }

    /**
     * Unregisters a listener.
     *
     * @param  listener  The listener that must be unregistered.
     */
    public static void removeListener(LyApplicationListener listener)
    {
        synchronized (_applications)
        {
        	if (_listeners.contains(listener))
        		_listeners.remove(listener);
        }
    }

    /**
     * Computes the default location id for the specified application id. This id is the
     * concatenation of the application id and LyLocation.DEFAULT_LOCATION prefixed by '_'.
     *
     * @param   applicationId  the application id
     *
     * @return  the default location id for this application.
     */
    public static String buildDefaultLocationId(String applicationId)
    {
        if (applicationId == null)
            return "_" + LyLocation.DEFAULT_LOCATION;

        String id = applicationId;
        int    dot = id.lastIndexOf(".");

        if (dot >= 0)
            id = id.substring(dot + 1);

        return ("_" + id + LyLocation.DEFAULT_LOCATION);
    }

    /**
     * Close all remaining applications.
     */
    public static void closeAll()
    {
        synchronized (_applications)
        {
            if (_applications.size() == 0)
                return;

            Enumeration<Vector<LyApplication>> elements = _applications.elements();
            Vector<LyApplication>              applications;
            LyApplication                      application;
            int                                i, n;

            while (elements.hasMoreElements())
            {
                applications = elements.nextElement();
                applications = new Vector<LyApplication>(applications); // clone the vector so it won't change
                n = ((applications == null) ? 0 : applications.size());

                for (i = 0; i < n; i++)
                {
                    try
                    {
                        application = applications.elementAt(i);
                        if (getApplication(application.getId()) == application)
                            application.close();
                    }
                    catch (Exception e)
                    {
                        // We are closing, no need to keep track of exception
                    }
                }
            }

            _applications = new Hashtable<String, Vector<LyApplication>>();
        }
    }

    /**
     * Constructor with a given id.
     *
     * @param  id  The identifier of the application.
     */
    public LyApplication(String id)
    {
        synchronized (_applications)
        {
            init();

            _id = id;

            LyApplication.putApplication(this);
            _environment = new LyEnvironment(_id, null, true);
        }
    }

    /**
     * Constructor with a given id.
     *
     * @param  id  The identifier of the application.
     */
    public LyApplication(String id, boolean edited)
    {
    	this(id);
    	_edited = edited;
    }

    /**
     * Constructor with a given id and configuration file.
     *
     * @param  id          The identifier of the application.
     * @param  configFile  The configuration file of the application.
     */
    public LyApplication(String id, String configFile)
    {
        this(id, configFile, null, null);
    }

    /**
     * Constructor with a given id and configuration file and a specific classloader.
     *
     * @param  id           The identifier of the application
     * @param  configFile   The configuration file for the application
     * @param  classLoader  A specific class loader used for loading specific classes and performing
     *                      introspection.
     */
    public LyApplication(String id, String configFile, ClassLoader classLoader)
    {
        this(id, configFile, null, classLoader);
    }

    /**
     * Constructor with a given id, configuration file and class loader.
     *
     * @param  id           The identifier of the application.
     * @param  configFile   The name of the config file. If null, the name of the file is built
     *                      using the identifier of the application followed by '.ini' (eg : for an
     *                      application whose identifier is 'sample' the default name of the
     *                      configuration file is 'sample.ini').
     * @param  path         The initial path for the application (LY_APP_DIR in LyEnvironment)
     * @param  classLoader  A specific class loader used for loading specific classes and performing
     *                      introspection.
     */
    public LyApplication(String id, String configFile, String path, ClassLoader classLoader)
    {
        this(null, id, configFile, path, classLoader);
    }

    /**
     * Constructor with a given id, configuration file and class loader.
     *
     * @param  parent       The parent application (null for the primary module).
     * @param  id           The identifier of the application.
     * @param  configFile   The name of the config file. If null, the name of the file is built
     *                      using the identifier of the application followed by '.ini' (eg : for an
     *                      application whose identifier is 'sample' the default name of the
     *                      configuration file is 'sample.ini').
     * @param  path         The initial path for the application (LY_APP_DIR in LyEnvironment)
     * @param  classLoader  A specific class loader used for loading specific classes and performing
     *                      introspection.
     */
    public LyApplication(LyApplication parent, String id, String configFile, String path, ClassLoader classLoader)
    {
        this(parent, null, id, configFile, path, classLoader);
    }

    /**
     * Constructor with a given id, configuration file and class loader.
     *
     * @param  parent       The parent application (null for the primary module).
     * @param  rootProject  The root project of this application.
     * @param  id           The identifier of the application.
     * @param  configFile   The name of the config file. If null, the name of the file is built
     *                      using the identifier of the application followed by '.ini' (eg : for an
     *                      application whose identifier is 'sample' the default name of the
     *                      configuration file is 'sample.ini').
     * @param  path         The initial path for the application (LY_APP_DIR in LyEnvironment)
     * @param  classLoader  A specific class loader used for loading specific classes and performing
     *                      introspection.
     */
    public LyApplication(LyApplication parent, LyProject rootProject, String id, String configFile, String path,
        ClassLoader classLoader)
    {
        this(parent, rootProject, id, configFile, path, classLoader, false);
    }

    /**
     * Constructor with a given id, configuration file and class loader.
     *
     * @param  parent       The parent application (null for the primary module).
     * @param  rootProject  The root project of this application.
     * @param  id           The identifier of the application.
     * @param  configFile   The name of the config file. If null, the name of the file is built
     *                      using the identifier of the application followed by '.ini' (eg : for an
     *                      application whose identifier is 'sample' the default name of the
     *                      configuration file is 'sample.ini').
     * @param  path         The initial path for the application (LY_APP_DIR in LyEnvironment)
     * @param  classLoader  A specific class loader used for loading specific classes and performing
     *                      introspection.
     * @param  isMobile		Indicates if the platform used is a mobile of not.
     */
    public LyApplication(LyApplication parent, LyProject rootProject, String id, String configFile, String path,
        ClassLoader classLoader, boolean isMobile)
    {
        Date startDate = new Date();

        // We can't do "this(id);" since initialization is done within a synchronized
        synchronized (_applications)
        {
            init();

            _parent = parent;
            _id = id;

            LyApplication.putApplication(this);

            // Add this application on its parent
            if (parent != null)
                parent.addSubApplication(this);

            LyEnvironment parentEnv = ((_parent == null) ? null : _parent.getEnvironment());
            //TODO : STAGE : Si jamais il y a un probleme, passer sur le constructeur de base si false
            _environment = new LyEnvironment(_id, parentEnv, true, isMobile);
            _environment.setClassLoader(classLoader);

            if (path != null)
                _environment.putEnv("LY_APP_DIR", path);

            if (parentEnv != null)
            {
                LyApplication appParent = parent;

                while (appParent != null)
                {
                    parentEnv = appParent.getEnvironment();
                    appParent = appParent.getParent();
                    if (parentEnv != null)
                        parentEnv.load(_id, getClassLoader());
                }

                // Init specific resources of this application.
                _environment.initResources();

                // Don't take following resources on parent environment
                _environment.putEnv("LY_PROJECT_FILE", "");
                _environment.putEnv("LY_PROJECT_CLASS", "");

                _environment.loadEnv(configFile, getClassLoader(), false);
            }
            else
                _environment.loadEnv(configFile, getClassLoader(), true);

            if (_environment.isDebugLevel(LyEnvironment.DEBUG_TRACE) && (path != null))
                _environment.logTrace("LY_APP_DIR=" + _environment.getEnv("LY_APP_DIR"));

            if (!isSplashScreenAllowed())
            	_environment.putEnv("LY_SPLASH_SCREEN", "_no");

            if (_parent == null)
            {
                LyViewManagerInterface viewManager = _environment.getViewManager(isMobile);

                if (viewManager != null)
                    viewManager.prepareGui();
            }

            // Check first if the associated project was already loaded
            if (rootProject == null)
            {
                LyInfo info = getInfo(_id);

                if (info instanceof LyProject)
                    rootProject = (LyProject)info;
            }

            start(rootProject);
        }

        if (isValid()
                && _environment.isDebugLevel(LyEnvironment.DEBUG_PERF))
        {
            Date endDate = new Date();

            System.gc();
            _environment.logOut("BF AppEngine : application " + this + " started in "
                + (endDate.getTime() - startDate.getTime()) + " ms."
                + "\n\tCurrent number of LyObject = " + LyObject._nbObjects);
        }
    }
    /**
     * Constructor with a given id.
     *
     * @param  parent  the parent application (can be null)
     * @param  id      The identifier of the application.
     */
    public LyApplication(LyApplication parent, String id)
    {
        synchronized (_applications)
        {
            init();

            _id = id;
            _parent = parent;

            // Add this application on its parent
            if (parent != null)
                parent.addSubApplication(this);

            LyApplication.putApplication(this);

            LyEnvironment parentEnv = ((_parent == null) ? null : _parent.getEnvironment());

            _environment = new LyEnvironment(_id, parentEnv, true);
        }
    }

    /**
     * Indicates whether the application is valid or not.
     *
     * @return  true if the application is valid, false otherwise.
     */
    public boolean isValid()
    {
        if (_applications == null)
            return false;

        // This synchronized is performed do avoid the use of this application
        // by another thread before it is completely started
        synchronized (_applications)
        {
            return _valid;
        }
    }

    /**
     * Accessor for the project of the application.
     *
     * @return  The project of the application.
     *
     * @see     leon.info.LyProject
     */
    public LyProject getProject()
    {
        return _project;
    }

    /**
     * Accessor for the identifier of the application.
     *
     * @return  The identifier of the application.
     */
    public String getId()
    {
        return _id;
    }

    /**
     * Indicates whether the application is asleep or not.
     *
     * @return  true if the application is asleep, false otherwise.
     *
     * @see     LyApplication#sleep()
     * @see     LyApplication#wakeUp()
     */
    public boolean isSleeping()
    {
        return (_isAsleep || isWaitingForClosing());
    }

    /**
     * Accessor for the error message of the application.
     *
     * @return  The error message of the application.
     */
    public String getError()
    {
        String error = _error;

        if ((error == null) && (_parent != null))
            error = _parent.getError();

        return error;
    }

    /**
     * Accessor for the specific class loader defined for this application.
     *
     * @return  The specific class loader defined for this application or null if not defined.
     *
     * @see     leon.misc.LyClassLoader
     */
    public ClassLoader getClassLoader()
    {
        return _environment.getClassLoader();
    }

    /**
     * Returns the behavior of this application.
     *
     * @return  The behavior of this application, may be null.
     */
    public LyApplicationBehaviorInterface getApplicationBehavior()
    {
    	if (_behavior != null)
    		return _behavior;
    	else if (isEdited())
    	{
    		_behavior = new LyDataApplicationBehavior();
    		return _behavior;
    	}
    	else
    		return null;
    }

    /**
     * Sets the error message of the application.
     *
     * @param  message  The new message that must be used.
     */
    public void setError(String message)
    {
        _error = message;
    }

    /**
     * Sets the specific class loader defined for this application with the given one. The specific
     * class loader could be, for example, a LyClassLoader.
     *
     * @param  loader  The new specific class loader defined for this application.
     *
     * @see    leon.misc.LyClassLoader
     */
    public void setClassLoader(ClassLoader loader)
    {
        _environment.setClassLoader(loader);
    }

    /**
     * Starts the application. There are four steps in this method : we parse the meta-model, create
     * the application behavior, create the providers and create the notifiers. If one of these
     * steps failed, we close the application. Otherwise, we create the daemons and the caches and
     * declare the application valid.
     *
     * @param  rootProject  The root project of this application. If this param is null the XML
     *                      model specified for this application is parsed.
     */
    public synchronized void start(LyProject rootProject)
    {
        try
        {
            if (!_environment.checkEnv())
                throw new IllegalStateException("environment is not valid.");

            if (rootProject == null)
            {
                if (!parse())
                    throw new IllegalStateException("parsing failed.");
            }
            else
            {
                if (!rootProject.isLoaded())
                    rootProject.load();

                _project = rootProject;
            }
            synchronized (_applications)
        	{
        		Iterator<LyApplicationListener> it = _listeners.iterator();
                LyApplicationListener listener;
                while (it.hasNext())
                {
                	listener = it.next();
                	listener.applicationStarted(this);
                }
        	}
            if (!createBehavior())
                throw new IllegalStateException("application behavior not ready.");

            // Clear resources of current view manager
            LyViewManagerInterface viewManager = _environment.getViewManager();

            if (viewManager != null)
                viewManager.clearTemporaryResources();

            // Load root project in the application
            if (rootProject == null)
                loadProject(_project);

            // Load extra infos
            loadExtraInfos();

            if ((_behavior != null)
            	&& !_behavior.started(this))
            {
                close();
                return; // Application start refused by the behavior.
            }

            LyInfoList<LyProject> projects = getInfoList(LyProject.class);
            LyProject project;
            for (int i = 0; i < projects.getSize(); i++)
            {
            	project = projects.getInfo(i);

            	if (project.getVersion() != null)
            		getEnvironment().logInfo("Version of " + project.getId() + ": " + project.getVersion());
            }
        }
        catch (Throwable e)
        {
            // Keep for showing error to the user when no console is available.
            StringWriter writer = new StringWriter();
            PrintWriter  pw = new PrintWriter(writer);

            if (e.getMessage() == null)
                e.printStackTrace(pw);
            else
            {
                pw.println("BF AppEngine : Application can't start, " + e.getMessage());

                if (!(e instanceof IllegalStateException))
                    e.printStackTrace(pw);
            }

            String error = getError();

            if (error != null)
                pw.println("Error Message : " + error);
            else
                _error = e.getMessage();

            pw.flush();

            _environment.logFatal(writer.toString());

            if (error == null)
                error = writer.toString();

            close();
            return;
        }

        // Sets application as valid
        _valid = true;
    } // end method start

    /** List of sub-projects discovered during parsing. */
    LyInfoList<LyProject> _subProjects = null;

    /** Accept loading of sub-projects. */
    boolean _acceptLoadingOfSubProjects = false;

    /**
     * Loads a project whithin an application. Create the providers and notifiers of the project.
     * Create the daemons and the caches.
     *
     * @param  project  the project to load
     */
    public void loadProject(LyProject project)
    {
    	if (project.getApplication() == this)
    		createProviders(project);

        if ((_parent != null)
            && !_parent._acceptLoadingOfSubProjects)
        {
            _parent.loadProject(project);
            return;
        }

        if (_acceptLoadingOfSubProjects)
            _acceptLoadingOfSubProjects = false;
        else if (project != _project)
        {
            // Root project is not yet loaded, load extern project afterward
            if (_subProjects == null)
                _subProjects = new LyInfoList<LyProject>();

            _subProjects.addInfo(project);
            return;
        }

        // Load classes of project
        LyInfoList<LyClassInfo> classes = project.getLoadedClasses();

        loadClasses(classes);
    }

    /**
     * Loads a list of classes whithin an application. Create the providers and notifiers of the
     * project. Create the daemons and the caches.
     *
     * @param  classes  a list of classes to load
     */
    public void loadClasses(LyInfoList<LyClassInfo> classes)
    {
        if ((_project == null) || (classes == null))
            return;

        // Validate classes
        validateClasses(classes);
        _acceptLoadingOfSubProjects = true;

        // Filter virtual classes
        LyInfoList<LyClassInfo> realClasses = new LyInfoList<LyClassInfo>();
        int                     i, nc = classes.getSize();
        LyClassInfo             classInfo;

        for (i = 0; i < nc; i++)
        {
            classInfo = classes.getInfo(i);
            if (!classInfo.isVirtual())
                realClasses.addInfo(classInfo);
        }

        // Create caches
        createCaches(realClasses);

        // Create daemons for all classes including virtual ones
        createDaemons(classes);

        // Create specific encoders.
        if (_behavior != null)
        {
            // Add specific encoders for new providers if needed
            int np = ((_providers == null) ? 0 : _providers.length);

            for (i = 0; i < np; i++)
                _behavior.addSpecificEncoders(this, _providers[i]);
        }

        // Now we can validate the notifiers
        validateNotifiers(realClasses);

        // Load data
        loadData(realClasses, null);

        // Find class infos that need to be notified
        LyClassInfo             classInfo2;
        LyInfoList<LyClassInfo> classInfos = getClassInfos();
        int                     j;

        for (j = classInfos.getSize() - 1; j >= 0; j--)
        {
            classInfo2 = classInfos.getInfo(j);
            if (classInfo2.isVirtual()
                    || !classInfo2.isLoaded()
                    || realClasses.contains(classInfo2))
                classInfos.removeInfo(classInfo2);
        }

        nc = classInfos.getSize();

        // Validate object sources
        LyObjectSource objectSource, objectSource2;
        int            n = realClasses.getSize();

        for (i = 0; i < n; i++)
        {
            classInfo = realClasses.getInfo(i);
            objectSource = getObjectSource(classInfo);

            // To have events on pointed objects.
            objectSource.validate();

            // Notify class loading to all existing object sources
            for (j = 0; j < nc; j++)
            {
                classInfo2 = classInfos.getInfo(j);
                objectSource2 = getObjectSource(classInfo2);
                if (objectSource2 != null)
                    objectSource2.notifyClassLoading(classInfo);
            }
        }
    } // end method loadClasses

    /**
     * Initialisation of the data of the instance. This method is called by the constructor.
     */
    protected void init()
    {
        // Initialize class attributes.
        _parent = null;
        _id = null;
        _defaultLocationId = null;
        _sessions = null;
        _environment = null;
        _project = null;
        _providers = null;
        _values = null;
        _tagTable = new Hashtable<String, LyTag>();
        // For optimization, we can let the infos list multiple to avoid checks
        // since the parser already did the control.
        _infoList = new LyInfoList<LyInfo>(true);
        _classInfos = new LyInfoList<LyClassInfo>(true);
        _dataSources = new Hashtable<String, LyDataSource>();
        _infoProvider = null;
        _unionProvider = null;
        _extendProvider = null;
        _isAsleep = false;
        _valid = false;
        _daemons = null;
        _error = null;
        _transactionManager = null;
        _eventManager = null;
        _threadInstances = null;
        _idle = false;
        _closing = false;
    }

    /**
     * Clears all data instance.
     */
    protected synchronized void free()
    {
        // Remove this application of its parent
        if (_parent != null)
            _parent.removeSubApplication(this);

        // Free sub-applications
        if (_subApplications != null)
        {
            int           i, n = _subApplications.size();
            LyApplication application;

            for (i = n - 1; i >= 0; i--)
            {
                application = _subApplications.elementAt(i);
                application.free();
            }

            _subApplications = null;
        }

        LyViewManagerInterface viewManager = _environment.getViewManager();

        // Reset class attributes.
        _parent = null;
        _id = null;
        _sessions = null;
        _environment.free();
        _environment = null;
        _project = null;
        _providers = null;
        _tagTable = null;
        _infoList = null;
        _dataSources = null;
        _infoProvider = null;
        _unionProvider = null;
        _extendProvider = null;
        _isAsleep = false;
        _valid = false;
        _daemons = null;
        _transactionManager = null;
        _eventManager = null;
        _threadInstances = null;
        _values = null;
        if (_closeTimer != null)
        {
            _closeTimer.stop();
            _closeTimer = null;
        }

        if (_applications.size() == 0)
        {
            LyEnvironment.freeRootEnvironment();

            if (viewManager != null)
                viewManager.exit(0);
        }
    } // end method free

    /**
     * Loads the precompiled meta-model if available, else parses the meta-model xml file. The root
     * tag of the meta-model must be a <CODE>&lt;PROJECT&gt;</CODE> tag.
     *
     * @return  false if no root project was found in the meta-model xml file, true otherwise.
     */
    public boolean parse()
    {
    	if (_project != null)
    		throw new IllegalStateException("Project is already loaded");

        LyInfoParser infoParser;
        String       className = _environment.getEnv("LY_PROJECT_CLASS");
        String       model = _environment.getEnv("LY_PROJECT_FILE");

        // Look for a compiled meta-model
        if ((className != null)
                && (className.length() > 0))
        {
            // We use the generated class to build the root project.
            _environment.logInfo("Using a compiled meta-model : " + className);

            try
            {
                // Create a new meta model.
            	@SuppressWarnings("unchecked")
                Class<LyMetaModel>  projectClass = (Class<LyMetaModel>)loadClass(className);
                LyMetaModel         metaModel = projectClass.newInstance();

                metaModel.load(this);

                // Get root project
                _project = metaModel.getRootProject();

            }
            catch (Exception e)
            {
                _environment.logFatal("Error loading class " + className);
                _environment.writeException(e);
            }
        }

        if (_project == null)
        {
            // We use the XML parser to build the root project.
            _environment.logInfo("Using the Leonardi parser...");
            _environment.logInfo("Begin of parsing " + new Date());

            infoParser = new LyInfoParser(this, model);

            _environment.logInfo("End of parsing   " + new Date());

            // Creating root project
            _project = infoParser.getRootProject();
        }

        if (_project == null)
            throw new IllegalStateException("no root project on application");

        // Lock project
        _project.lock();

        return true;
    } // end method parse

    /**
     * Method for parsing extra information such as charts or preferences.
     *
     * @param   prefix   File prefix of loaded infos.
     * @param   dirName  Name (or path) of the directory containing infos to be loaded.
     * @param   userId   User id (may be null when info is shared).
     *
     * @return  List of loaded infos.
     */
    public LyInfoList<LyInfo> parseExtraFiles(String prefix, String dirName, String userId)
    {
        if (dirName == null)
            return null;

        if (dirName != null)
        {
            dirName = _environment.getString().expand(dirName);
            if (dirName == null)
                return null;
        }

        File     directory = new File(_environment.getFile(dirName));
        String[] fileList = directory.list();

        if ((fileList == null) || (fileList.length == 0))
            return null;

        LyInfoList<LyInfo> infos = null;
        int                i, n = fileList.length, j, ni = _infoList.getSize(), p;
        LyInfoParser       infoParser;
        LyInfo             info;
        LyDataSession      session = getCurrentSession();
        LyObject           user = ((session != null) ? session.getUser() : null);
        String             sessionUserId = ((user == null) ? null : user.getObjectId().toString());
        String             infoUserId, filePath;
        boolean            loaded;

        for (i = 0; i < n; i++)
        {
            if (!fileList[i].startsWith(prefix))
                continue;

            if ((userId != null)
                    && !fileList[i].contains("_" + userId + "_"))
                continue;

            if ((userId == null)
                    && (sessionUserId != null)
                    && fileList[i].contains("_" + sessionUserId + "_"))
                continue;

            filePath = directory + "/" + fileList[i];

            // Check if this file was already loaded
            for (j = 0, loaded = false; (j < ni) && !loaded; j++)
            {
                info = _infoList.getInfo(j);
                if ((info.getFile() != null)
                        && info.getFile().equals(filePath))
                    loaded = true;
            }

            // Already loaded, skip.
            if (loaded)
                continue;

            try
            {
                infoParser = new LyInfoParser(this, filePath);
                info = (LyInfo)infoParser.getRootElement();

                if (info == null)
                    continue;

                if (info instanceof LyExtraInfoInterface)
                {
                    infoUserId = ((LyExtraInfoInterface)info).getUser();

                    if (infoUserId != null)
                    {
                        p = infoUserId.indexOf(LyObjectId.CLASS_SEP);
                        if (p >= 0)
                            infoUserId = infoUserId.substring(p + 1);
                    }

                    if (((infoUserId == null)
                                && (userId != null))
                            || ((infoUserId != null)
                                && (userId == null))
                            || ((infoUserId != null)
                                && (userId != null)
                                && !infoUserId.equals(userId)))
                    {
                        ((LyExtraInfoInterface)info).free();
                        continue;
                    }
                }

                if (infos == null)
                    infos = new LyInfoList<LyInfo>();

                infos.addInfo(info);

                // Set the file for this info since it has been loaded
                info.setFile(filePath);
            }
            catch (Exception e)
            {
                _environment.writeException(e);
            }
        }

        return infos;
    } // end method parseExtraFiles

    /**
     * Creates the application behavior. By default, this behavior is an instance of
     * LyApplicationBehaviorInterface. This but it may be customized by using the <CODE>
     * 'behavior'</CODE> attribute on the <CODE>&lt;PROJECT&gt;</CODE> tag.
     *
     * @return  false if an error occurred, true otherwise.
     */
    protected boolean createBehavior()
    {
        Class<?>    behaviorClass;

        try
        {
            behaviorClass = _project.getApplicationBehaviorClass();
            if (behaviorClass == null)
            {
                LyEnvironment environment = getEnvironment();

                if (environment.getViewManager() != null)
                    behaviorClass = loadClass("leon.app.behavior.LyApplicationBehavior", null);
                else
                    behaviorClass = LyDataApplicationBehavior.class;
            }
            else if (!LyApplicationBehaviorInterface.class.isAssignableFrom(behaviorClass))
                throw new IllegalStateException("BF AppEngine : invalid application behavior "
                    + behaviorClass + ", it must implement "
                    + LyApplicationBehaviorInterface.class);
        }
        catch (Exception e)
        {
            _environment.writeException(e);
            // Use default behavior
            behaviorClass = LyDataApplicationBehavior.class;
        }

        try
        {
            _behavior = (LyApplicationBehaviorInterface)behaviorClass.newInstance();
        }
        catch (Exception e)
        {
            _environment.writeException(e);
            return false;
        }

        addSessionListener(_behavior);
        return _behavior.start(this);
    } // end method createBehavior

    /**
     * Creates the data providers for the application.
     *
     * @param      project  a project
     *
     * @return     false if an error occured, true otherwise.
     *
     * @exception  java.lang.IllegalStateException  if an object source is null for a given class
     *                                              info.
     *
     * @see        leon.peer.LyDataProvider
     */
    protected boolean createProviders(LyProject project)
    {
        if (_parent == null)
        {
            if (_infoProvider == null)
                _infoProvider = new LyInfoProvider(this, new LyLocation(LyLocation.ID_L_INFO, null));

            if (_unionProvider == null)
                _unionProvider = new LyUnionDataProvider(this, new LyLocation(LyLocation.ID_L_UNION, null));

            if (_extendProvider == null)
                _extendProvider = new LyExtendDataProvider(this, new LyLocation(LyLocation.ID_L_EXTEND, null));

            if (_localProvider == null)
            {
                // Create local provider and validate it
                _localProvider = new LyLocalDataProvider(this, new LyLocation("_local", null));
                addDataProvider(_localProvider);
            }
        }

        String         defaultLocationId = getDefaultLocationId();
        LyLocation     location;
        LyDataProvider provider;
        LyApplication  rootApplication = getRootApplication();

        if (getInfo(defaultLocationId) == null)
        {
            // Create a default location for this project if needed.
            location = new LyFileLocation(defaultLocationId, "LY_DATA_DIR", LyFileLocation.Type.INFO, null, null);
            location.setApplication(rootApplication);
            provider = LyDataProviderFactory.getProvider(this, location);
            addDataProvider(provider);
        }

        // Create new providers
        int i, n = project.getLocationCount();

        for (i = 0; i < n; i++)
        {
            location = project.getLocation(i);
            provider = LyDataProviderFactory.getProvider(this, location);
            addDataProvider(provider);
        }

        // Create providers for sub-projects
        LyProject subProject;

        for (i = 0, n = project.getProjectCount(); i < n; i++)
        {
            subProject = project.getProject(i);
            createProviders(subProject);
        }

        return true;
    } // end method createProviders

    /**
     * Add a data provider to the current application.
     *
     * @param  provider  a data provider to add
     */
    public void addDataProvider(LyDataProvider provider)
    {
        if (provider == null)
            return;

        if (_parent != null)
        {
            _parent.addDataProvider(provider);
            return;
        }

        int        np = ((_providers == null) ? 0 : _providers.length);
        LyLocation oldLocation, location = provider.getLocation();
        String     providerId = location.getId();

        for (int i = 0; i < np; i++)
        {
            oldLocation = _providers[i].getLocation();
            if (providerId.equals(oldLocation.getId()))
                return;
        }

        // Resize providers table
        LyDataProvider[] providers = new LyDataProvider[np + 1];

        for (int i = 0; i < np; i++)
            providers[i] = _providers[i];

        _providers = providers;
        _providers[np] = provider;
    } // end method addDataProvider

    /**
     * Validate the given list of classes for existing providers.
     *
     * @param   classes  List of classes to validate.
     *
     * @return  false if validation failed on one of the clases, true otherwise.
     */
    public boolean validateClasses(LyInfoList<LyClassInfo> classes)
    {
        validateProviders(classes);

        // At least, all classInfos must have an object source.
        int            i, n;
        LyClassInfo    classInfo;
        LyObjectSource objectSource;

        for (i = 0, n = classes.getSize(); i < n; i++)
        {
            classInfo = classes.getInfo(i);
            if (classInfo.isVirtual())
                continue;

            objectSource = getObjectSource(classInfo);
            if (objectSource == null)
                throw new IllegalStateException("BF AppEngine : classInfo " + classInfo
                    + " has no data provider.");
        }

        // Validate classes of sub-projects discovered while validating current classes
        if (_subProjects != null)
        {
            LyInfoList<LyProject>   subProjects = new LyInfoList<LyProject>(_subProjects);
            LyProject               subProject;
            LyInfoList<LyClassInfo> subClasses;
            int                     j;

            _subProjects = null;
            for (i = 0, n = subProjects.getSize(); i < n; i++)
            {
                subProject = subProjects.getInfo(i);
                subClasses = subProject.getLoadedClasses();

                // Remove classes that have already been validated
                for (j = subClasses.getSize() - 1; j >= 0; j--)
                {
                    classInfo = subClasses.getInfo(j);
                    if (classes.contains(classInfo))
                        subClasses.removeInfo(classInfo);
                }

                if (subClasses.getSize() > 0)
                {
                    validateClasses(subClasses);
                    classes.addList(subClasses);
                }
            }
        }

        // Everything is ok
        return true;
    } // end method validateClasses

    /**
     * Validate the given list of classes for existing providers.
     *
     * @param  classes  List of classes to validate.
     */
    protected void validateProviders(LyInfoList<LyClassInfo> classes)
    {
        if (_parent != null)
        {
            _parent.validateProviders(classes);
            return;
        }

        int i, n = _providers.length;

        for (i = 0; i < n; i++)
        {
            if (!_providers[i].validate(classes))
            {
                String error = getError();

                throw new IllegalStateException("BF AppEngine : validation of data provider \""
                    + _providers[i].getLocation().getId() + "\" failed."
                    + ((error != null) ? ("\n" + error) : ""));
            }
        }

        // Validate union provider
        _unionProvider.validate(classes);

        // Validate info provider
        _infoProvider.validate(classes);

        // Validate extend provider
        _extendProvider.validate(classes);
    }

    /**
     * Creates the daemons for the application.
     *
     * @param  classInfos  a list of classes
     */
    protected void createDaemons(LyInfoList<LyClassInfo> classInfos)
    {
        // Create routed daemons.
        LyClassInfo classInfo;
        int         i, n = classInfos.getSize();

        for (i = 0; i < n; i++)
        {
            classInfo = classInfos.getInfo(i);
            createDaemon(classInfo);
        }

        if (_behavior != null)
            _behavior.createDaemons(this); // XXX Give list of classes ?
    }

    /**
     * Creates the daemons for the application of the given class info.
     *
     * @param  classInfo  a class
     *
     * @see    leon.data.daemon.LyRelationDaemon
     * @see    leon.data.daemon.LyJoinRelationDaemon
     * @see    leon.data.daemon.LyReverseRelationDaemon
     * @see    leon.data.daemon.LyRoutedRelationDaemon
     */
    protected void createDaemon(LyClassInfo classInfo)
    {
        // Create routed daemons.
        LyFieldInfo         fieldInfo;
        int                 i, n = classInfo.getFieldInfoCount();
        LyRelationFieldInfo relation;
        LyArrayFieldInfo    array;
        LyInfoDaemon        daemon;
        LyInfoLabel         label;

        for (i = 0; i < n; i++)
        {
            fieldInfo = classInfo.getFieldInfo(i);
            if (fieldInfo.isLocal()
                    && (fieldInfo instanceof LyRelationFieldInfo)
                    && !classInfo.isDerivated(fieldInfo))
            {
                relation = (LyRelationFieldInfo)fieldInfo;
                if (relation.getRoutes() != null)
                    new LyRoutedRelationDaemon(classInfo, relation, false);
                else if (relation.getDaemon() != null)
                {
                    daemon = relation.getDaemon();
                    if (daemon instanceof LyJoinDaemon)
                        new LyJoinRelationDaemon(classInfo, relation, ((LyJoinDaemon)daemon).getJoinClass(),
                            ((LyJoinDaemon)daemon).getSourceRelation(), ((LyJoinDaemon)daemon).getTargetRelation());
                    else if (daemon instanceof LyReverseDaemon)
                        new LyReverseRelationDaemon(classInfo, relation,
                            ((LyReverseDaemon)daemon).getReverseRelation());
                }
            }
            else if (fieldInfo.isLocal()
                    && (fieldInfo instanceof LyArrayFieldInfo)
                    && !classInfo.isDerivated(fieldInfo))
            {
                array = (LyArrayFieldInfo)fieldInfo;
                daemon = array.getDaemon();
                if (daemon instanceof LyArrayDaemon)
                    new LyArrayFieldInfoDaemon(classInfo, array, ((LyArrayDaemon)daemon).getArrayClassInfo());

                createLabelDaemons(classInfo, array.getFieldInfo());
            }
            else if (fieldInfo.getPath() != null)
                new LyPathDaemon(classInfo, fieldInfo);

            if (!classInfo.isDerivated(fieldInfo))
                createLabelDaemons(classInfo, fieldInfo);
        }

        label = classInfo.getLabel(LyInfoLabel.Type.NAME);

        if (label != null)
            new LyLabelDaemon(classInfo, label);
    } // end method createDaemon

    /**
     * Creates the label daemons for the application of the given info in the class info given as
     * parameter.
     *
     * @param  classInfo  a class
     * @param  fieldInfo  a field
     */
    protected void createLabelDaemons(LyClassInfo classInfo, LyFieldInfo fieldInfo)
    {
        // Create daemons for labels.
        LyInfoLabel label;

        label = fieldInfo.getLabel(LyInfoLabel.Type.VALUE);
        if (label != null)
            new LyLabelDaemon(classInfo, label);

        label = fieldInfo.getLabel(LyInfoLabel.Type.SHORT_VALUE);
        if (label != null)
            new LyLabelDaemon(classInfo, label);

        label = fieldInfo.getLabel(LyInfoLabel.Type.GRAMMATICAL);
        if (label != null)
            new LyLabelDaemon(classInfo, label);

        label = fieldInfo.getLabel(LyInfoLabel.Type.SEMANTIC);
        if (label != null)
            new LyLabelDaemon(classInfo, label);
    }

    /**
     * Validate the associated event notifier if exists.
     *
     * @param  classes  Classes to validate.
     */
    protected void validateNotifiers(LyInfoList<LyClassInfo> classes)
    {
        if (_parent != null)
        {
            _parent.validateNotifiers(classes);
            return;
        }

        int i, n = _providers.length;

        for (i = 0; i < n; i++)
        {
            if (!_providers[i].validateNotifier(classes))
                _environment.logError("BF AppEngine : validation of notifier for provider \""
                    + _providers[i].getLocation().getId() + "\" failed.");
        }
    }

    /**
     * Creates the caches for the application.
     *
     * @param  classes  a list of classes
     *
     * @see    leon.data.LyCache
     */
    protected void createCaches(LyInfoList<LyClassInfo> classes)
    {
        if (_parent != null)
        {
            _parent.createCaches(classes);
            return;
        }

        LyClassInfo classInfo;
        int         i, n = classes.getSize();

        for (i = 0; i < n; i++)
        {
            classInfo = classes.getInfo(i);
            createCache(classInfo);
        }

        _unionProvider.completeCaches(classes);
    }

    /**
     * Creates the cache for the application of the given class info.
     *
     * @param  classInfo  a class
     *
     * @see    leon.data.LyCache
     */
    public void createCache(LyClassInfo classInfo)
    {
        if (classInfo.isAbstract())
            // No cache on abstract classes.
            return;

        LyClassBinding classBinding = classInfo.getBinding();

        if (classBinding instanceof LyUnionBinding)
            // We let the union provider do the job.
            return;

        String      forcedCachePolicy = getEnvironment().getEnv("LY_CACHE_POLICY");

        LyInfoCache    cachePolicy = classInfo.getCachePolicy();
        if (forcedCachePolicy != null)
            cachePolicy = new LyInfoCache(LyInfoTokens.getClassInfoCachePolicy(forcedCachePolicy), null);

        LyLocation     location;
        String         locationId;
        LyObjectSource objectSource = getObjectSource(classInfo);

        if ((cachePolicy == null)
            || (cachePolicy.getCachePolicy() != LyFullCache.class))
        {
            locationId = (classBinding == null) ? null : classBinding.getLocation();
            if (locationId == null)
                locationId = getDefaultLocationId();

            location = (LyLocation)getInfo(locationId);

            if (((location != null) && !location.isPersistent())
                    || getEnvironment().testEnv("LY_USE_FILES"))
            {
                // If location can't save update on objects, we must use full caches.
                cachePolicy = new LyInfoCache(LyInfoCache.CachePolicy.FULL, null);
                if (!classInfo.isInternal())
	                _environment.logWarning("Forcing cache type to FULL for classInfo "
	                    + classInfo + " because the data provider is not persistent");
            }
        }

        objectSource.setCache(cachePolicy);
    } // end method createCache

    /**
     * Adds a session to the application.
     *
     * @param      session  The session that must be added to the application.
     *
     * @exception  java.lang.IllegalArgumentException  if the session is null.
     * @exception  java.lang.IllegalStateException     if the session has already been added to the
     *                                                 application.
     *
     * @see        LyApplication#removeSession(LyDataSession)
     * @see        leon.app.LyDataSession
     */
    protected void addSession(LyDataSession session)
    {
        if (session == null)
            throw new IllegalArgumentException("BF AppEngine : attemps to add a null session to an application");

        if (_sessions == null)
            _sessions = new Vector<LyDataSession>();

        // Check if sleeping app was in progress ?
        LyViewManagerInterface viewManager = _environment.getViewManager();

        if (viewManager != null)
        {
            viewManager.setPluginId(session.getPluginId());
            // Abort any application closing if in progress.
            applicationClosing(this, true /*abort*/);
        }

        if (_sessions.contains(session))
            throw new IllegalStateException("BF AppEngine : attemps to add an existent session to an application");

        synchronized (_sessions)
        {
            _sessions.addElement(session);
        }

        // Call listeners
        if (_sessionListeners != null)
        {
        	int		i, n = _sessionListeners.size();

        	for (i = 0; i < n; i++)
        		_sessionListeners.get(i).sessionAdded(this, session);
        }
    } // end method addSession

    /**
     * Removes a session from the application.
     *
     * @param      session  The session that must be removed.
     *
     * @exception  java.lang.IllegalArgumentException  if the session is null.
     * @exception  java.lang.IllegalArgumentException  if the session has not been added to the
     *                                                 application.
     *
     * @see        LyApplication#addSession(LyDataSession)
     * @see        leon.app.LyDataSession
     */
    protected void removeSession(LyDataSession session)
    {
        if (session == null)
            throw new IllegalArgumentException("BF AppEngine : attemps to remove "
                + "a null session to an application");

        if ((_sessions == null) || (!_sessions.contains(session)))
            throw new IllegalStateException("BF AppEngine : attemps to remove a "
                + "non existent session to an application");

        // Clean up threads/sessions table
        if (_threadInstances != null)
            _threadInstances.removeKeys(session);

        synchronized (_sessions)
        {
            // Remove session from sessions list
            _sessions.removeElement(session);
        }

        // Call listeners
        if (_sessionListeners != null)
        {
        	int		i, n = _sessionListeners.size();

        	for (i = 0; i < n; i++)
        		_sessionListeners.get(i).sessionRemoved(this, session);
        }

        // Close application if last session
        if ((_sessions.size() == 0)
                && !_environment.testResource("LY_APP_SERVER_MODE"))
        {
            LyViewManagerInterface viewManager = _environment.getViewManager();

            if (viewManager != null)
                viewManager.setPluginId(session.getPluginId());

            applicationClosing(this, false);
        }
    } // end method removeSession

    /**
     * Retreives current sessions count.
     *
     * @return  the number of session
     */
    public int getSessionCount()
    {
        return ((_sessions == null) ? 0 : _sessions.size());
    }

    /**
     * Get session at the given index.
     *
     * @param   index  an index
     *
     * @return  the corresponding session or no sessions are registered
     */
    public LyDataSession getSession(int index)
    {
        return ((_sessions == null) ? null : _sessions.elementAt(index));
    }

    /**
     * Get session at the given index.
     */
    public LyDataSession getSession(String sessionId)
    {
        if (_sessions == null)
            return null;

        synchronized (_sessions)
        {
            int           i, n = _sessions.size();
            LyDataSession session;

            for (i = 0; i < n; i++)
            {
                session = _sessions.elementAt(i);
                if (sessionId == session.getId())
                    return session;
            }
        }

        return null;
    }

    /**
     * Retreives current sessions list.
     *
     * @return  a list of sessions
     */
    public Vector<LyDataSession> getSessions()
    {
        if (_sessions == null)
            return new Vector<LyDataSession>();

        return new Vector<LyDataSession>(_sessions);
    }

    /**
     * Computes the default location id for this application. This id is the concatenation of the
     * application id and LyLocation.DEFAULT_LOCATION prefixed by '_'.
     *
     * @return  the default location id for this application.
     */
    public String getDefaultLocationId()
    {
        if (_defaultLocationId == null)
        {
            _defaultLocationId = _environment.getResource("LY_DEFAULT_LOCATION");
            if ((_defaultLocationId == null)
                    || (_defaultLocationId.length() == 0))
                _defaultLocationId = LyApplication.buildDefaultLocationId(getId());
        }

        return _defaultLocationId;
    }

    /**
     * To set the default lcation id for this application.
     *
     * @param  defaultLocationId  the new default location id for this application.
     */
    public void setDefaultLocationId(String defaultLocationId)
    {
        _defaultLocationId = defaultLocationId;
    }

    /**
     * To get all supported languages by the application.
     *
     * @return  Array of couple of strings that represents supported languages. Each couple of
     *          string is the code of the country for the language as defined in java.util.Locale
     *          for the first string, and the display name of the language or a key to the
     *          dictionnary of this string for the second.
     */
    public String[][] getSupportedLanguages()
    {
        String[][] 						result = null;
        LyApplicationBehaviorInterface	behavior = _behavior;

        if (behavior == null)
        	behavior = new LyDataApplicationBehavior();

        result = behavior.getSupportedLanguages(this);

        if (result == null)
        {
            String[]      language = new String[2];
            LyEnvironment environment = getEnvironment();
            Locale        locale = environment.getLocale();

            result = new String[1][2];

            language[0] = locale.getLanguage();
            language[1] = locale.getDisplayCountry();

            result[0] = language;
        }

        return result;
    }

    /**
     * To get all supported timeZone by the application.
     *
     * @return  Array of couple of strings that represents supported languages. Each couple of
     *          string is the code of the country for the language as defined in java.util.Locale
     *          for the first string, and the display name of the language or a key to the
     *          dictionnary of this string for the second.
     */
    public String[] getSupportedTimeZones()
    {
        String[] 						result = null;
        LyApplicationBehaviorInterface	behavior = _behavior;

        if (behavior == null)
        	behavior = new LyDataApplicationBehavior();

        result = behavior.getSupportedTimeZones(this);

        if (result == null)
        {
            result = new String[25];
            int i;
            int index = 0;
            for (i = 12; i >0 ; i--)
            {
                result[index] = "GMT-" + i + ":00";
                result[24 - index] = "GMT+" + i + ":00";
                index ++;
            }
            result[index] = "GMT";
        }

        return result;
    }

    /**
     * Load data and fill caches.
     *
     * @see  LyApplication#loadData(boolean,BitSet)
     * @see  LyApplication#unloadData()
     */
    public final void loadData()
    {
        loadData(null, null);
    }

    /**
     * Load data and fill caches.
     *
     * @param  classInfos           List of classes to load (null means all).
     * @param  marks                Indicates the marks that the classes to load must match. If
     *                              null, all the classes are loaded.
     *
     * @see    LyApplication#loadData()
     * @see    leon.info.LyClassInfo
     */
    public void loadData(LyInfoList<LyClassInfo> classInfos, BitSet marks)
    {
        if (classInfos == null)
            classInfos = _project.getLoadedClasses();

        if ((marks == null)
        	&& !_environment.testEnv("_LY_LOAD_DATA"))
        {
            marks = new BitSet();
            marks.set(LyClassInfo.MARK_LOAD);
        }

        LyInfoList<LyClassInfo> classes = (marks == null ? classInfos : classInfos.getInfos(marks));
        int                     n = (classes == null ? 0 : classes.getSize());

        if (n == 0)
            return;

        _idle = true;

        // Load content...
        LyObjectSource objectSource;
        LyClassInfo    classInfo;
        int            i;

        for (i = 0; i < n; i++)
        {
            classInfo = classes.getInfo(i);

            if ((_behavior != null) && !_behavior.enableLoadUnloadData(classInfo))
            	continue;

            _environment.logInfo("Loading " + classInfo);
            objectSource = getObjectSource(classInfo);
            objectSource.load();
            _environment.logInfo("Done");
        }

        _idle = false;

        if (_behavior != null)
            _behavior.loadData(this, classes);
    } // end method loadData

    /**
     * Clears all application data. WARNING : this method must never be invoked while views, or
     * workspaces are opened. Be sure to close all the windows and clear all the listeners before
     * cleaning data.
     *
     * @see  LyApplication#loadData()
     */
    public void unloadData()
    {
        if (_parent != null)
            return;

        if (_behavior != null)
            _behavior.unloadData(this);

        LyInfoList<LyClassInfo> classes = getClassInfos();
        LyClassInfo             classInfo;
        LyObjectSource          objectSource;
        int                     i, n = ((classes == null) ? 0 : classes.getSize());

        for (i = n - 1; i >= 0; i--)
        {
            classInfo = classes.getInfo(i);

            if ((_behavior != null) && !_behavior.enableLoadUnloadData(classInfo))
            	continue;

            objectSource = getObjectSource(classInfo);
            // FFT00949
            // A deletion event is 'loading' when it's related to an unload() request
            // This may occur when going into sleepingAPP mode or on
            // User request. Do not propagate on related objects to avoid loops.
            if (objectSource != null)
                objectSource.unload(false); // Do not notify deletions
        }

        if (_daemons != null)
        {
            Enumeration<LyDaemon> daemons = _daemons.elements();
            LyDaemon              daemon;

            while (daemons.hasMoreElements())
            {
                daemon = daemons.nextElement();
                daemon.unload();
            }
        }

        dataUnloaded();
    } // end method unloadData

    /**
     * Internal method called when data have been unloaded. This method calls the dataUnloaded
     * method for each sub application and calls dataUnloaded method on each session.
     */
    void dataUnloaded()
    {
        // Free dialogs for sub-applications
        int           i, n = ((_subApplications == null) ? 0 : _subApplications.size());
        LyApplication subApplication;

        for (i = 0; i < n; i++)
        {
            subApplication = _subApplications.elementAt(i);
            subApplication.dataUnloaded();
        }

        // freeDialogs for all sessions
        if (_sessions != null)
        {
            LyDataSession session;

            for (i = 0, n = _sessions.size(); i < n; i++)
            {
                session = _sessions.elementAt(i);
                session.dataUnloaded();
            }
        }
    }

    /**
     * To free all dialog controllers of all sessions.
     */
    public void freeDialogSessions()
    {
        if (_parent != null)
            _parent.freeDialogSessions();
        else
            dataUnloaded();
    }

    /**
     * Reload application data.
     *
     * @param  unload  Boolean indicating if datas have to be unlodaed first.
     *
     * @see    LyApplication#loadData()
     */
    public void reloadData(boolean unload)
    {
        if (unload)
            // Unload existing datas
            unloadData();

        // Reset sub-applications
        int           i, n = ((_subApplications == null) ? 0 : _subApplications.size());
        LyApplication subApplication;

        for (i = 0; i < n; i++)
        {
            subApplication = _subApplications.elementAt(i);
            subApplication.reloadData(false);
        }

        // Reset all sessions
        if (_sessions != null)
        {
            LyDataSession session;

            for (i = 0, n = _sessions.size(); i < n; i++)
            {
                session = _sessions.elementAt(i);
                session.reset();
            }
        }
    } // end method reloadData

    /**
     * Save all application data. Some of the providers do nothing when this method is called, but
     * some others do some stuff (like LyFileDataProvider that saves all the data on files).
     */
    public void saveData()
    {
        if (_parent != null)
            return;

        if (_behavior != null)
            _behavior.saveData(this);

        int i, n = ((_providers == null) ? 0 : _providers.length);

        for (i = 0; i < n; i++)
            _providers[i].save();
    }

    /**
     * Adds a new relation daemon to the application.
     *
     * @param  daemon  The relation daemon that must be added to the application.
     *
     * @see    LyApplication#getDaemon(LyFieldInfo)
     * @see    leon.data.daemon.LyRelationDaemon
     */
    public void addDaemon(LyDaemon daemon)
    {
        if (_parent != null)
        {
            _parent.addDaemon(daemon);
            return;
        }

        if (_daemons == null)
            _daemons = new Hashtable<LyInfo, LyDaemon>(1);

        _daemons.put(daemon.getInfo(), daemon);
    }

    /**
     * Retrieves the daemon in charge of the specified field info.
     *
     * @param   fieldInfo  The field info whose daemon we want to get.
     *
     * @return  the corresponding daemon
     */
    public LyFieldInfoDaemon getDaemon(LyFieldInfo fieldInfo)
    {
        if (fieldInfo == null)
            return null;

        if (_parent != null)
            return _parent.getDaemon(fieldInfo);

        if (_daemons == null)
            return null;

        return (LyFieldInfoDaemon)_daemons.get(fieldInfo);
    }

    /**
     * Retrieves the daemon that controls label of the given info label.
     *
     * @param   label  The infoLabel for which the daemon is searched.
     *
     * @return  the corresponding daemon.
     */
    public LyLabelDaemon getDaemon(LyInfoLabel label)
    {
        if (label == null)
            return null;

        if (_parent != null)
            return _parent.getDaemon(label);

        if (_daemons == null)
            return null;

        return (LyLabelDaemon)_daemons.get(label);
    }

    /**
     * Closing current application. This method closes current application behavior and data
     * providers Then frees all resources. This application should not be used anymore.
     *
     * @see  #sleep()
     */
    public void close()
    {
        if (_parent != null)
            return;

        close(null);
    }

    /**
     * Closing current application.
     *
     * @param  parent  the parent application
     */
    protected synchronized void close(LyApplication parent)
    {
        if (_parent != parent)
            return;

        _closing = true;
        if (_environment != null)
            _environment.logTrace("BF AppEngine : Closing application " + this);

        Date startDate = new Date();

        if (_behavior != null)
            _behavior.close(this);

        // Close all sessions
        if (_sessions != null)
        {
            int                    i, n = _sessions.size();
            LyDataSession          session;
            LyEnvironment          environment = getEnvironment();
            LyViewManagerInterface vm = environment.getViewManager();
            _LyCloseSession        runnable;

            for (i = n - 1; i >= 0; i--)
            {
                session = _sessions.elementAt(i);
                if (vm != null)
                {
                    runnable = new _LyCloseSession(session);
                    vm.execGui(runnable, true);
                }
                else
                    session.close();
            }
        }

        // Close sub-applications
        if (_subApplications != null)
        {
            int           i, n = _subApplications.size();
            LyApplication application;

            for (i = n - 1; i >= 0; i--)
            {
                application = _subApplications.elementAt(i);
                application.close(this);
            }
        }

        synchronized (this)
        {
            // Close providers
            if (_providers != null)
            {
                int i, n = _providers.length;

                for (i = 0; i < n; i++)
                {
                    if (_providers[i] != null)
                        _providers[i].close();
                }
            }

            if (_infoProvider != null)
                _infoProvider.close();

            if (_extendProvider != null)
                _extendProvider.close();

            if (_unionProvider != null)
                _unionProvider.close();

            if ((_environment != null)
                    && _environment.isDebugLevel(LyEnvironment.DEBUG_PERF))
            {
                Date endDate = new Date();

                System.gc();
                _environment.logOut("BF AppEngine : application " + this + " closed in "
                    + (endDate.getTime() - startDate.getTime()) + " ms."
                    + "\n\tCurrent number of LyDataSession = " + LyDataSession._nbSessions
                    + "\n\tCurrent number of LyObject = "
                    + LyObject._nbObjects
                /*+ "\n\tCurrent number of LyController = "      + LyController._nbControllers
                 * + "\n\tCurrent number of LyView = " + LyView._nbViews+ "\n\tCurrent number of LyComponent = " +
                 * LyComponent._nbComponents*/);
            }

            // Unload data
            unloadData();

            if (_environment != null)
            {
                // Unload temporary resources.
                LyViewManagerInterface viewManager = _environment.getViewManager();

                if (viewManager != null)
                    viewManager.clearTemporaryResources();
            }

            // If the id of the application is null, this application has been closed
            // when removing all the sessions.
            if (getId() != null)
            {
                LyApplication.removeApplication(this);
                if (_parent == null)
                    free();
            }
        }
    } // end method close

    /**
     * Sets the application in sleeping mode. This method clears all application data, clears all
     * the application views and sets all its providers and notifiers in sleeping mode.
     *
     * @see  LyApplication#wakeUp()
     * @see  #close()
     */
    public synchronized void sleep()
    {
        if (_isAsleep)
            return;

        _environment.logTrace("BF AppEngine : Trying to asleep application " + this);

        if (_behavior != null)
            _behavior.sleep(this);

        if (_parent == null)
        {
            int i;

            for (i = 0; i < _providers.length; i++)
                _providers[i].sleep();
        }

        // Unload all the datas.
        unloadData();

        // Unload all views.
        LyViewManagerInterface viewManager = _environment.getViewManager();

        if (viewManager != null)
        {
            viewManager.clearTemporaryResources();
            _environment.setViewManager(null);
        }

        _isAsleep = true;

        _environment.logTrace("BF AppEngine : Application " + this + " is now asleep.");

        if (_environment.isDebugLevel(LyEnvironment.DEBUG_PERF))
        {
            System.gc();
            _environment.logOut("BF AppEngine : application " + this + " is asleep"
                + "\n\tCurrent number of LyDataSession = " + LyDataSession._nbSessions
                + "\n\tCurrent number of LyObject = "
                + LyObject._nbObjects
            /*+ "\n\tCurrent number of LyController = "      + LyController._nbControllers
             * + "\n\tCurrent number of LyView = " + LyView._nbViews+ "\n\tCurrent number of LyComponent = " +
             * LyComponent._nbComponents */);
        }
    } // end method sleep

    /**
     * Wakes up an application which is in sleeping mode. This method loads data, fills cache and
     * wakes up all the application providers and notifiers.
     *
     * @return  true if wake up was successfull.
     *
     * @see     LyApplication#sleep()
     */
    public synchronized boolean wakeUp()
    {
        if (!_isAsleep)
        {
            abortClosing();
            return true;
        }

        _environment.logTrace("BF AppEngine : Starting to awake application " + this);

        if (_behavior != null)
        {
            try
            {
                _behavior.wakeUp(this);
            }
            catch (Exception e)
            {
                _environment.logError("BF AppEngine : Wake up of application " + this
                    + " failed when calling behavior.");
                return false;
            }
        }

        if (_parent == null)
        {
            int i;

            for (i = 0; i < _providers.length; i++)
            {
                if (!_providers[i].wakeUp())
                    return false;
            }
        }

        _isAsleep = false;

        // Load data
        loadData();

        _environment.logTrace("BF AppEngine : Application " + this + " is now awake.");

        if (_environment.isDebugLevel(LyEnvironment.DEBUG_PERF))
        {
            System.gc();
            _environment.logOut("BF AppEngine : application " + this + " is awake"
                + "\n\tCurrent number of LyDataSession = " + LyDataSession._nbSessions
                + "\n\tCurrent number of LyObject = "
                + LyObject._nbObjects
            /*+ "\n\tCurrent number of LyController = "      + LyController._nbControllers
             * + "\n\tCurrent number of LyView = " + LyView._nbViews+ "\n\tCurrent number of LyComponent = " +
             * LyComponent._nbComponents */);
        }

        return true;
    } // end method wakeUp

    /**
     * Retrieves a tag from its id.
     *
     * @param   id  The id of the searched tag.
     *
     * @return  The object corresponding to the id or null if no object is found.
     *
     * @see     leon.info.LyInfo
     * @see     LyApplication#putInfo(LyInfo)
     */
    public LyTag getTag(String id)
    {
        if (_parent != null)
            return _parent.getTag(id);

        if ((id == null) || (id.length() == 0))
            return null;

        return _tagTable.get(id);
    }

    /**
     * Retrieves an info from its id.
     *
     * @param   id  The id of the searched Info.
     *
     * @return  The object corresponding to the id or null if no object is found.
     *
     * @see     leon.info.LyInfo
     * @see     LyApplication#putInfo(LyInfo)
     */
    public LyInfo getInfo(String id)
    {
        LyTag tag = getTag(id);

        return ((tag instanceof LyInfo) ? (LyInfo)tag : null);
    }

    /**
     * Registers a tag in the application.
     *
     * @param  tag  a tag to register
     */
    public void putTag(LyTag tag)
    {
        if (_parent != null)
            _parent.putTag(tag);
        else
            _tagTable.put(tag.getId(), tag);
    }

    /**
     * Registers an info in the application. This may be an action, a class info, a field info...
     *
     * @param  info  an info to register
     *
     * @see    leon.info.LyInfo
     * @see    LyApplication#getInfo(String)
     */
    public void putInfo(LyInfo info)
    {
        if (_parent != null)
        {
            _parent.putInfo(info);
            return;
        }

        String id = info.getId();
        LyInfo oldInfo = getInfo(id);

        if (info == oldInfo)
            return;

        if (oldInfo != null)
            throw new IllegalStateException("BF AppEngine : an info with id " + id
                + " is already registered on application " + this);

        putTag(info);
        _infoList.addInfo(info);
        if (info instanceof LyClassInfo)
            _classInfos.addInfo((LyClassInfo)info);
    }

    /**
     * Unregisters a tag from this application.
     *
     * @param  tag  a tag to unregister
     */
    public void removeTag(LyTag tag)
    {
        if (_parent != null)
            _parent.removeTag(tag);
        else
            _tagTable.remove(tag.getId());
    }

    /**
     * Unregisters an info from this application.
     *
     * @param  info  an info to unregister
     *
     * @see    leon.info.LyInfo
     * @see    LyApplication#putInfo
     */
    public void removeInfo(LyInfo info)
    {
        if (_parent != null)
        {
            _parent.removeInfo(info);
            return;
        }

        removeTag(info);
        _infoList.removeInfo(info);
        if (info instanceof LyClassInfo)
            _classInfos.removeInfo((LyClassInfo)info);
    }

    /**
     * Shortcut to get a classInfo of the application.
     *
     * @param   id  The identifier of current class.
     *
     * @return  The requested classInfo, or null if it could not be found.
     */
    public LyClassInfo getClassInfo(String id)
    {
        if (_parent != null)
            return _parent.getClassInfo(id);

        LyInfo info = getInfo(id);

        if (info instanceof LyClassInfo)
            return (LyClassInfo)info;
        else
            return null;
    }

    /**
     * Find classinfos from the peer value (connector value defined in class binding). Since
     * multiple classes may be bound to the same data source binding, this method may return a list.
     *
     * @param   binding  the binding to search
     *
     * @return  the requested classinfos, an empty list if none is found or null if parameter is
     *          null
     */
    public LyInfoList<LyClassInfo> getClassInfosFromBinding(String binding)
    {
        if (binding == null)
            return null;

        LyClassInfo             classInfo;
        LyClassBinding          classBinding;
        LyInfoList<LyClassInfo> classInfoList = getClassInfos();
        LyInfoList<LyClassInfo> result = new LyInfoList<LyClassInfo>();

        for (int i = 0; (classInfoList != null) && (i < classInfoList.getSize()); i++)
        {
            classInfo = classInfoList.getInfo(i);
            classBinding = classInfo.getBinding();

            if (classBinding != null)
            {
                if (classBinding instanceof LyUnionBinding)
                {
                    List<String> bindings = ((LyUnionBinding)classBinding).getAllBindings();

                    if ((bindings != null) && (bindings.contains(binding)))
                        result.addInfo(classInfo);
                }
                else if (binding.equals(classBinding.getBinding()))
                    result.addInfo(classInfo);
            }
        }

        return result;
    }

    /**
     * Shortcut to get an action of the application.
     *
     * @param   id  The identifier of current action.
     *
     * @return  The requested action, or null if it could not be found.
     */
    public LyAction getAction(String id)
    {
        if (_parent != null)
            return _parent.getAction(id);

        LyInfo info = getInfo(id);

        if (info instanceof LyAction)
            return (LyAction)info;
        else
            return null;
    }

    /**
     * Get an action of the project with given template id.
     *
     * @param   id         The identifier of the template of searched action.
     * @param   classInfo  If not null, look first on class info before project.
     *
     * @return  The requested action, or null if it could not be found.
     */
    public LyAction getActionWithId(String id, LyClassInfo classInfo)
    {
        if (_parent != null)
            return _parent.getActionWithId(id, classInfo);

        if (classInfo != null)
        {
            LyAction action = classInfo.getActionWithId(id);

            if (action != null)
                return action;
        }

        LyAction template = getAction(id);

        if (template == null)
            return null;

        LyProject            project = getProject();
        LyInfoList<LyAction> actions = project.getActions();
        int                  i, n = ((actions == null) ? 0 : actions.getSize());
        LyAction             action;

        for (i = 0; i < n; i++)
        {
            action = actions.getInfo(i);
            if (action.hasTemplate(template))
                return action;
        }

        return template;
    }

    /**
     * Get an action of the application with given template id.
     *
     * @param   id         The identifier of the template of searched action.
     * @param   classInfo  If not null, look first on class info before project.
     *
     * @return  The requested action, or null if it could not be found.
     */
    public LyAction getActionWithTemplate(String id, LyClassInfo classInfo)
    {
        if (_parent != null)
            return _parent.getActionWithId(id, classInfo);

        if (classInfo != null)
        {
            LyAction action = classInfo.getActionWithId(id);

            if (action != null)
                return action;
        }

        LyAction template = getAction(id);

        if (template == null)
            return null;

        LyInfoList<LyAction> actions = getActions();
        int                  i, n = ((actions == null) ? 0 : actions.getSize());
        LyAction             action;

        for (i = 0; i < n; i++)
        {
            action = actions.getInfo(i);
            if (action.hasTemplate(template))
                return action;
        }

        return template;
    }

    /**
     * Returns an ordered list of all the tags of a given java class registered in the application.
     *
     * @param   <T>       the type of Tag to get
     * @param   tagClass  The java class.
     *
     * @return  List of all tags of the given java class registered on the application.
     */
	public <T extends LyTag> ArrayList<T> getTags(Class<T> tagClass)
    {
        if (_parent != null)
            return _parent.getTags(tagClass);

        if (_tagTable == null)
            return null;

        ArrayList<T>       result = new ArrayList<T>();
        Enumeration<LyTag> tags = _tagTable.elements();
        LyTag              tag;

        while (tags.hasMoreElements())
        {
            tag = tags.nextElement();
            if ((tagClass == null)
                    || tagClass.isAssignableFrom(tag.getClass()))
            {
            	@SuppressWarnings("unchecked")
            	T temp = (T)tag;
                result.add(temp);
            }
        }

        return result;
    }

    /**
     * Returns all the infos of a given java class registered in the application.
     *
     * @param       <T>        the type of LyInfo to return
     * @param       infoClass  a java class.
     *
     * @return      The requested info, or null if it could not be found.
     *
     * @see         leon.info.LyInfo
     * @deprecated  Use getInfoList
     */
    public <T extends LyInfo> LyInfoList<T> getInfos(Class<T> infoClass)
    {
        return getInfoList(infoClass);
    }

    /**
     * Returns an ordered list of all the infos of a given java class registered in the application.
     *
     * @param   <T>        the type of LyInfo to return
     * @param   infoClass  The java class.
     *
     * @return  List of all infos of the given java class registered on the application.
     *
     * @see     leon.info.LyInfo
     */
    public <T extends LyInfo> LyInfoList<T> getInfoList(Class<T> infoClass)
    {
        if (_parent != null)
            return _parent.getInfoList(infoClass);

        if (_infoList == null)
            return null;

        return _infoList.getInfos(infoClass);
    }

    /**
     * To get all the class infos registered in the application.
     *
     * @return  a list of classInfo
     *
     * @see     leon.info.LyClassInfo
     */
    public LyInfoList<LyClassInfo> getClassInfos()
    {
        if (_parent != null)
            return _parent.getClassInfos();

        if (_classInfos == null)
            return null;

        return new LyInfoList<LyClassInfo>(_classInfos);
    }

    /**
     * To get all the actions registered in the application.
     *
     * @return  a list of Actions
     *
     * @see     leon.info.LyAction
     */
    public LyInfoList<LyAction> getActions()
    {
        if (_parent != null)
            return _parent.getActions();

        return getInfos(LyAction.class);
    }

    /**
     * Returns the environment of the application.
     *
     * @return  The environment of the application.
     *
     * @see     leon.misc.LyEnvironment
     */
    public LyEnvironment getEnvironment()
    {
        return _environment;
    }

    /**
     * Shortcut to get a data source registered in the application.
     *
     * @param   classId  The identifier of the class whose data source we want to retrieve.
     *
     * @return  The data source corresponding to the given class identifier or null if it could not
     *          be found.
     */
    public LyDataSource getDataSource(String classId)
    {
        if (_parent != null)
            return _parent.getDataSource(classId);

        if (_dataSources.size() == 0)
            return null;

        LyDataSource dataSource = _dataSources.get(classId);

        if (dataSource != null)
            return dataSource;

        LyClassInfo classInfo = getClassInfo(classId);

        if (classInfo != null)
        {
            if (classInfo.isVirtual())
            {
                dataSource = getDataSource(classInfo.getRealClass());
                if (dataSource != null)
                {
                    _dataSources.put(classId, dataSource);
                    return dataSource;
                }
            }

            if (!classInfo.isLoaded())
            {
                classInfo.load();
                if (classInfo.isLoaded())
                    return classInfo.getObjectSource();
            }
        }

        // If classinfo is unknown (studio), don't throw an exception
        if (getClassInfo(classId) == null)
            return null;

        throw new IllegalStateException("BF AppEngine : no data source found for given id " + classId);
    } // end method getDataSource

    /**
     * Shortcut to get a data source registered in the application.
     *
     * @param   lyClass  The class whose data source we want to retrieve.
     *
     * @return  The data source corresponding to the given class or null if it could not be found.
     */
    public LyDataSource getDataSource(LyClass lyClass)
    {
        if (_parent != null)
            return _parent.getDataSource(lyClass);

        if (_dataSources == null)
            return null;

        if ((lyClass instanceof LyClassInfo)
                && ((LyClassInfo)lyClass).isVirtual())
            lyClass = ((LyClassInfo)lyClass).getRealClass();

        return _dataSources.get(lyClass.getId());
    }

    /**
     * Registers a data source in the application.
     *
     * @param  dataSource  The data source that must be registered.
     *
     * @see    LyApplication#getDataSource(String)
     * @see    LyApplication#getDataSource(LyClass)
     */
    public void putDataSource(LyDataSource dataSource)
    {
        if (_parent != null)
        {
            _parent.putDataSource(dataSource);
            return;
        }

        LyClass lyClass = dataSource.getDataClass();

        if (!(dataSource instanceof LyTemporaryObjectSource)
                && (getDataSource(lyClass) != null))
            throw new IllegalStateException("BF AppEngine : a data source for class " + lyClass
                + " is already registered on application " + getId());

        _dataSources.put(lyClass.getId(), dataSource);

        if (dataSource instanceof LyObjectSource)
        {
        	LyObjectSource objectSource = (LyObjectSource)dataSource;
        	synchronized (_applications)
        	{
        		Iterator<LyApplicationListener> it = _listeners.iterator();
                LyApplicationListener listener;
                while (it.hasNext())
                {
                	listener = it.next();
                	listener.objectSourceAdded(this, objectSource);
                }
        	}
        }
    }

    /**
     * Unregisters a data source from the application.
     *
     * @param  dataSource  The data source that must be unregistered.
     *
     * @see    LyApplication#getDataSource(String)
     * @see    LyApplication#getDataSource(LyClass)
     */
    public void removeDataSource(LyDataSource dataSource)
    {
        if (_parent != null)
        {
            _parent.removeDataSource(dataSource);
            return;
        }

        LyClass lyClass = dataSource.getDataClass();

        if (getDataSource(lyClass) == null)
            throw new IllegalStateException("BF AppEngine : a data source for class " + lyClass
                + " was not registered on application " + getId());

        _dataSources.remove(lyClass.getId());
    }

    /**
     * Shortcut to get an object source registered in the application.
     *
     * @param   classInfo  The class info whose object source we want to retrieve.
     *
     * @return  The object source corresponding to the given class or null if it could not be found.
     */
    public LyObjectSource getObjectSource(LyClassInfo classInfo)
    {
        if (_parent != null)
            return _parent.getObjectSource(classInfo);

        if (classInfo.isVirtual())
            classInfo = classInfo.getRealClass();

        return ((LyObjectSource)getDataSource(classInfo));
    }

    /**
     * Returns the info provider of the application.
     *
     * @return  The info provider of the application.
     */
    public LyInfoProvider getInfoProvider()
    {
        if (_parent != null)
            return _parent.getInfoProvider();

        return _infoProvider;
    }

    /**
     * Calls the loadClass(String className, String errorMessage) method with an error message not
     * defined. It implies that a generic message is displayed instead.
     *
     * @param   className  The fully qualified name of the desired class.
     *
     * @return  The loaded class or null if it could not be found.
     */
    public Class<?> loadClass(String className)
    {
        return loadClass(className, null);
    }

    /**
     * This method enables to find a class whose name matches the given name. If a specific class
     * loader is defined over this application then, its "loadClass" method is used, otherwise the
     * static "Class.forName" method is used.
     *
     * @param   className     The fully qualified name of the desired class.
     * @param   errorMessage  The error message displayed on the standard output stream when a
     *                        ClassNotFoundException is caught.
     *
     * @return  The loaded class or null if it could not be found.
     */
    public Class<?> loadClass(String className, String errorMessage)
    {
        // Substitute given class by the one found in environment
        // if a resource was set with the same name
        String appClassName = _environment.getResource("REPLACE-CLASS#" + className);

        if (appClassName != null)
        {
            Class<?> appClass = loadClass(appClassName, null);

            if (appClass != null)
                return appClass;
        }

        Class<?>    classe = null;
        ClassLoader classLoader = getClassLoader();

        try
        {
            if (classLoader == null)
                classe = Class.forName(className);
            else
                classe = classLoader.loadClass(className);
        }
        catch (ClassNotFoundException e)
        {
            if (errorMessage == null)
                errorMessage = "BF AppEngine : class not found " + className;

            if (_environment.testEnv("_LY_IGNORE_UNKNOWN_CLASSES"))
                _environment.logError(errorMessage);
            else
                throw new IllegalStateException(errorMessage);
        }

        return classe;
    } // end method loadClass

    /**
     * Gets the current session for the current thread.
     *
     * @return  Instance of the current session (may be null) if it was not set in the current
     *          thread.
     */
    public LyDataSession getCurrentSession()
    {
    	LyDataSession	currentSession = (LyDataSession)getCurrentInstance(CURRENT_DATA_SESSION);

    	if (currentSession == null)
    	{
			LyViewManagerInterface viewManager = _environment.getViewManager();

			if ((viewManager == null)
				|| !viewManager.isMultiSession())
			{
				if ((_sessions != null)
						&& (_sessions.size() == 1))
					// If only one session, consider it as the right one
					currentSession = _sessions.elementAt(0);
			}
    	}

        return currentSession;
    }

    /**
     * Sets the current session to the given one for the current thread.
     *
     * @param   session  Session to set as the current one for the current thread. If null, unset
     *                   the current session.
     *
     * @return  Previous registered session if it exists.
     */
    public LyDataSession setCurrentSession(LyDataSession session)
    {
        return (LyDataSession)setCurrentInstance(CURRENT_DATA_SESSION, session);
    }

    /**
     * Gets the transaction manager of this application.
     *
     * @return  The associated transaction manager.
     */
    public LyTransactionManager getTransactionManager()
    {
        if (_parent != null)
            return _parent.getTransactionManager();

        if (_transactionManager == null)
            _transactionManager = new LyTransactionManager(this);

        return _transactionManager;
    }

    /**
     * Tests if the speified session has somme pending transaction.
     *
     * @param   session  a session
     *
     * @return  true if the specified session has some pending transaction, else false.
     */
    public boolean hasTransaction(LyDataSession session)
    {
        LyTransactionManager tm = getTransactionManager();

        return (tm.getTransaction(session) != null);
    }

    /**
     * Gets the event manager of this application.
     *
     * @return  The associated event manager.
     */
    public LyEventManager getEventManager()
    {
        if (_parent != null)
            return _parent.getEventManager();

        if (_eventManager == null)
        {
            LyApplicationBehaviorInterface behavior = getApplicationBehavior();

            if (behavior != null)
            	_eventManager = behavior.getEventManager(this);
        }

        if (_eventManager == null)
            _eventManager = new LyEventManager(this);

        return _eventManager;
    }

    /**
     * To get the local data provider of this application.
     *
     * @return  the local data provider
     */
    public LyDataProvider getLocalDataProvider()
    {
        if (_parent != null)
            return _parent.getLocalDataProvider();

        return _localProvider;
    }

    /**
     * Retreives current providers count.
     *
     * @return  the count of data provider
     */
    public int getDataProviderCount()
    {
        return ((_providers == null) ? 0 : _providers.length);
    }

    /**
     * Get data provider at the given index.
     *
     * @param   index  an index
     *
     * @return  the corrresponding LyDataProvider or null if no data provider are registered
     */
    public LyDataProvider getDataProvider(int index)
    {
        return ((_providers == null) ? null : _providers[index]);
    }

    /**
     * Get the priority level of this application. This level is used by controls and rules to check
     * if a control or a rule is triggered. If the level of this application is lesser than 0 and
     * this application has a parent, the parent level is returned.
     *
     * @param   object  the object on which the control is applied (may be null).
     *
     * @return  the priority level of the application.
     */
    public int getPriorityLevel(LyObject object)
    {
        int level = ((_behavior == null) ? -1 : _behavior.getPriorityLevel(this, object));

        if ((_parent != null) && (level == -1))
            level = _parent.getPriorityLevel(object);
        else if (level == -1)
            level = Integer.MAX_VALUE;

        return level;
    }

    /**
     * Accessor to the the parent application.
     *
     * @return  null if the application is the primary module, the parent application otherwise.
     */
    public LyApplication getParent()
    {
        return _parent;
    }

    /**
     * Check if the given application is parent of this application.
     *
     * @param   application  the application
     *
     * @return  True if the given application is an ancestor of this application, false otherwise.
     */
    public boolean hasParent(LyApplication application)
    {
        if (_parent == null)
            return false;

        if (_parent == application)
            return true;

        return _parent.hasParent(application);
    }

    /**
     * Gets the context manager of the current session of this application.
     *
     * @return  The associated context manager of the current session of this application.
     */
    public LyContextManager getContextManager()
    {
        if (_parent != null)
            return _parent.getContextManager();

        LyDataSession session = getCurrentSession();

        if (session != null)
            return session.getContextManager();

        return null;
    }

    /**
     * Gets the current context for the current thread.
     *
     * @return  Instance of the current context (may be null) if it was not set in the current
     *          thread.
     */
    public LyContext getCurrentContext()
    {
        return (LyContext)getCurrentInstance(CURRENT_CONTEXT);
    }

    /**
     * Sets the current context to the given one for the current thread.
     *
     * @param   context  Context to set as the current one for the current thread. If null, unset
     *                   the current context.
     *
     * @return  Previous registered context if it exists.
     */
    public LyContext setCurrentContext(LyContext context)
    {
        return (LyContext)setCurrentInstance(CURRENT_CONTEXT, context);
    } // end method setCurrentContext

    /**
     * To get the root application of this application.
     *
     * @return  the root application of this application.
     */
    public LyApplication getRootApplication()
    {
        LyApplication application, parent;

        application = this;
        parent = application.getParent();

        while (parent != null)
        {
            application = parent;
            parent = application.getParent();
        }

        return application;
    }

    /**
     * Retreives current sub-applications count.
     *
     * @return  the count of sub application
     */
    public int getSubApplicationCount()
    {
        return ((_subApplications == null) ? 0 : _subApplications.size());
    }

    /**
     * Get sub-application at the given index.
     *
     * @param   index  an index
     *
     * @return  a corresponding sub application. Null if none
     */
    public LyApplication getSubApplication(int index)
    {
        return ((_subApplications == null) ? null : _subApplications.elementAt(index));
    }

    /**
     * Get sub-applications of this application.
     *
     * @return  A vector containing all sub-applications, null if no sub-applications found.
     */
    public Vector<LyApplication> getSubApplications()
    {
        return ((_subApplications == null) ? null : new Vector<LyApplication>(_subApplications));
    }

    /**
     * Add a sub-application to this application.
     *
     * @param  application  Sub-application added to this application.
     */
    protected void addSubApplication(LyApplication application)
    {
        if (_subApplications == null)
            _subApplications = new Vector<LyApplication>();

        _subApplications.add(application);
    }

    /**
     * Remove a sub-application to this application.
     *
     * @param  application  Sub-application removed to this application.
     */
    protected void removeSubApplication(LyApplication application)
    {
        _subApplications.remove(application);
    }

    /**
     * Change data provider location.
     *
     * @param      providerId   Id of the provider to be changed.
     * @param      newLocation  New location for the data provider corresponding to the given id.
     *
     * @return     Boolean indicating if the change was successfully completed or not.
     *
     * @exception  IllegalArgumentException  if the given id doesn't correspond to an existing
     *                                       provider. IllegalArgumentException if the given
     *                                       location hasn't the same type as the previous one
     *                                       associated to the provider.
     */
    public boolean setDataProviderLocation(String providerId, LyLocation newLocation)
    {
        if ((providerId == null)
                || (newLocation == null))
            return false;

        if (_parent != null)
            // Redirect demand on primary module
            return _parent.setDataProviderLocation(providerId, newLocation);

        // Find provider corresponding to given id
        LyDataProvider provider = null;
        int            i, np = ((_providers == null) ? 0 : _providers.length);

        for (i = 0; (i < np) && (provider == null); i++)
        {
            if (_providers[i].getId().equals(providerId))
                provider = _providers[i];
        }

        if (provider == null)
            throw new IllegalArgumentException("BF AppEngine : the given provider id "
                + providerId + " doesn't correspond to an existing provider");

        if (provider.getLocation().getClass() != newLocation.getClass())
            throw new IllegalArgumentException("BF AppEngine : the given location type "
                + newLocation.getClass() + " isn't not compatible with the existing one "
                + provider.getLocation().getClass());

        // Repost demand in GUI queue to avoid invalid thread accesses (on SWT)
        LyViewManagerInterface viewManager = _environment.getViewManager();
        LyLocation             location = provider.getLocation();

        if (viewManager != null)
            viewManager.execGui(new _LyCancelAbortUnload(provider, (location != newLocation)), true);
        else
            internalCancelAbortUnload(provider, (location != newLocation));

        try
        {
            // Set location on provider
            provider.setLocation(newLocation);
        }
        catch (Exception e)
        {
            _environment.logTrace("BF AppEngine : Failed to restart application " + this
                + " with location " + newLocation + " on provider " + providerId);
            _environment.writeException(e);
            if (location != newLocation)
            {
                // Reload data
                if (viewManager != null)
                    viewManager.execGui(new _LyReloadData(true), true);
                else
                    reloadData(true);
            }

            return false;
        }

        // Reload data and return once completed
        if (location != newLocation)
        {
            if (viewManager != null)
                viewManager.execGui(new _LyReloadData(true), true);
            else
                reloadData(true);
        }

        return true;
    } // end method setDataProviderLocation

    void internalCancelAbortUnload(LyDataProvider provider, boolean unload)
    {
        // Abort all context
        Vector<LyDataSession> sessions = getSessions();
        int                   n = sessions.size();
        LyDataSession         session;
        LyContextManager      cm;

        for (int i = 0; i < n; i++)
        {
            session = sessions.elementAt(i);
            cm = session.getContextManager();
            cm.cancelAll();
            cm.clearUndoRedo();
        }

        // Abort all remaining transactions on this provider.
        LyTransactionManager tm = getTransactionManager();

        tm.abort(provider);

        // Unload data.
        if (unload)
        {
            // Save data first for file providers
            saveData();

            // Unload data
            unloadData();
        }
    } // end method internalCancelAbortUnload

    /**
     * Indicates if a splash screen may be shown when starting the application.
     * @return true if splash screen must be shown, false otherwise
     */
    protected boolean isSplashScreenAllowed()
    {
    	return true;
    }

    /**
     * Method to test if all data providers of this application have some connection with the
     * physical layer.
     *
     * @return  true if all provider are OK, false if one of the providers has lost its connection
     *          with the physical layer.
     */
    public boolean checkProviders()
    {
        if (_parent != null)
            return _parent.checkProviders();

        // Check if each providers has already a connection.
        int i, np = ((_providers == null) ? 0 : _providers.length);

        for (i = 0; i < np; i++)
        {
            if (_providers[i].hasLostConnection())
                return false;
        }

        return true;
    }

    /**
     * Internal class used to close sessions.
     */
    class _LyCloseSession
        implements Runnable, LyGetSessionInterface
    {
        /** The session to close. */
        LyDataSession _session;

        _LyCloseSession(LyDataSession session)
        {
            _session = session;
        }

        public void run()
        {
            _session.close();
        }

		public final LyDataSession getSession()
		{
			return _session;
		}
    }

    /**
     * Internal class used to reload data.
     */
    class _LyReloadData
        implements Runnable
    {

        /** Flag indicated if the data may be unloaded before reloding. */
        boolean _unload;

        _LyReloadData(boolean unload)
        {
            _unload = unload;
        }

        public void run()
        {
            reloadData(_unload);
        }
    }

    /**
     * Internal class used to cancel all contexts, abort the specified data provider and to unload
     * data.
     */
    class _LyCancelAbortUnload
        implements Runnable
    {

        /** The provider to abort. */
        LyDataProvider _provider;

        /** Flag indicating if data may be unloaded or not. */
        boolean _unload;

        _LyCancelAbortUnload(LyDataProvider provider, boolean unload)
        {
            _provider = provider;
            _unload = unload;
        }

        public void run()
        {
            internalCancelAbortUnload(_provider, _unload);
        }
    }

    /**
     * When the application is closing. The application is not yet closed, but all sessions have
     * been closed by user or automatically. The implementation may go to sleeping app mode or to
     * close the application.
     *
     * @param  application  current application being closed.
     * @param  cancel       cancel indicates if the application is closing or if the closing request
     *                      is canceled. Some viewers (like web) may wait for a while before closing
     *                      the application. If parameter is false then the viewer may try to cancel
     *                      the closing procedure. Default implementation of cancel does <i>
     *                      Nothing</i>
     */
    protected void applicationClosing(LyApplication application, boolean cancel)
    {
        if (_closeTimer != null)
        {
            _closeTimer.stop();
            _closeTimer = null;
        }

        if (cancel)
            return;

        LyViewManagerInterface viewManager = _environment.getViewManager();

        if ((viewManager != null)
        	&& viewManager.hasProperty("DELAYED_CLOSING"))
        {
            // Eventually create a timer and/or go to Sleeping app mode
            String durationStr =  _environment.getResource("LY_SLEEPING_APP_DURATION");

            if (durationStr != null)
            {
                try
                {
                    long duration = Long.parseLong(durationStr);

                    if (duration > 0)
                    {
                        _environment.logTrace("BF AppEngine : waiting " + duration
                            + " sec. before closing application.");
                        _closeTimer = new LyTimer(this, duration * 1000L,
                                "BF AppEngine : application sleeping duration timer"); // Timer is in ms.
                        _closeTimer.setClientData(application);
                        _closeTimer.start();
                    }
                }
                catch (NumberFormatException e)
                {
                    _environment.writeException(e);
                }
            }
        }

        if (_closeTimer == null)
            // Call immediately, otherwise wait for notifications
            endTimer(null);
    } // end method applicationClosing

    /**
     * When a LyTimer ends this method is invoked.
     *
     * @param  timer  the timer which expired.
     */
    public void endTimer(LyTimer timer)
    {
        if (_environment.testEnv("LY_SLEEPING_APP"))
            sleep();
        else
            close();
    }

    /**
     * Indicates whether the application is idle or not.
     *
     * @return  true if the application is idle, false otherwise.
     */
    public boolean isIdle()
    {
        return !_valid || _idle;
    }

    /**
     * Sets idle mode on the application : no event on operation when set.
     *
     * @param  idle  true if the application is idle, false otherwise.
     */
    public void setIdle(boolean idle)
    {
        _idle = idle;
    }

    /**
     * Adds a pair key-value to the session. May be used for example to store the login used to
     * create the session.
     *
     * @param  key    The key that must be stored.
     * @param  value  The value that must be stored with the key.
     *
     * @see    #getValue
     */
    public void addValue(Object key, Object value)
    {
        if (_parent != null)
        {
        	_parent.addValue(key, value);
        	return;
        }
    	if (_values == null)
            _values = new HashMap<Object, Object>();

        _values.put(key, value);
    }

    /**
     * Retrieves a value from a registered key.
     *
     * @param   key  The key that must be used.
     *
     * @return  The value stored with the key or null if the key is not registered.
     *
     * @see     #addValue
     */
    public Object getValue(Object key)
    {
    	if (_parent != null)
    	{
    		return _parent.getValue(key);
    	}
        if ((_values == null) || (key == null))
            return null;

        return _values.get(key);
    }

    /**
     * Indicates whether the application is closing or not.
     *
     * @return  true if the application is closing, false otherwise.
     */
    public boolean isClosing()
    {
        return _closing;
    }

    /**
     * Indicates whether the application is waiting for closing or not. A timer was started to close
     * application.
     *
     * @return  true if the application is waiting for closing, false otherwise.
     */
    public boolean isWaitingForClosing()
    {
        return (_closeTimer != null);
    }

    /**
     * Abort the closing of an application.
     */
    public void abortClosing()
    {
        if (_closeTimer == null)
            return;

        _environment.logTrace("BF AppEngine : aborting diferred closing " + this);
        _closeTimer.stop();
        _closeTimer = null;
    }

    /**
     * Save given extra info like preferences, charts or treemap configurations.
     *
     * @param   info  Info to be saved.
     * @return  True if extra info was sucessfully saved.
     */
    public final boolean saveExtraInfo(LyExtraInfoInterface info)
    {
    	return saveExtraInfo(info, null);
    }

    /**
     * Save given extra info like preferences, charts or treemap configurations.
     *
     * @param   info  Info to be saved.
     * @return  True if extra info was sucessfully saved.
     */
    public boolean saveExtraInfo(LyExtraInfoInterface info, LyXmlWriterInterface writer)
    {
        if ((_behavior != null)
        	&& _behavior.saveExtraInfo(this, info, writer))
        {
            // Put info to the application
            putInfo((LyInfo)info);
            return true;
        }
        else
            return false;
    }

    /**
     * Load extra infos.
     * @param application Application on which extra infos (like charts or treemap configurations)
     * are loaded.
     * @return List of extra infos for the application.
     */
    public LyInfoList<LyInfo> loadExtraInfos()
    {
    	LyApplicationBehaviorInterface	behavior = getApplicationBehavior();

    	if (behavior != null)
    	{
    		// Load extra infos : charts, preferences and treemap configurations
    		LyInfoList<LyInfo> preferenceList = behavior.loadPreferences(this);
    		LyInfoList<LyInfo> extraInfoList = behavior.loadExtraInfos(this);
	        LyInfoList<LyInfo> result = null;
    		if (preferenceList != null)
    			result = preferenceList;
	        if (extraInfoList != null)
	        {
	        	if (result == null)
	        		result = extraInfoList;
	        	else
	        		result.addList(extraInfoList);
	        }

	        return result;
    	}
    	else
    		return null;
    }

    /**
     * Load extra info like preferences, charts, treemap configurations or form saves.
     *
     * @param   extraInfoClass  Class of the extra info.
     * @param   userId          Id of the user.
     * @param   extraInfoId     Id of extra info.
     *
     * @return  Info that was saved if found, null otherwise.
     */
    public Vector<LyExtraInfoInterface> loadExtraInfosFromDB(String extraInfoId, String userId)
    {
        if (_behavior == null)
            return null;

        return _behavior.loadExtraInfosFromDB(this, extraInfoId, userId);
    }

    /**
     * Delete given extra info like preferences, charts or treemap configurations.
     *
     * @param   info  Info to be deleted.
     *
     * @return  True if extra info was sucessfully deleted.
     */
    public boolean deleteExtraInfo(LyExtraInfoInterface info)
    {
        if (_behavior.deleteExtraInfo(this, info))
        {
            // Delete extra info
            info.free();
            return true;
        }
        else
            return false;
    }

    /**
     * Gets the current instance corresponding to specified name for the current thread.
     *
     * @param   name  Name of searched instance.
     *
     * @return  Instance associated to given in current thread (may be null).
     */
    public Object getCurrentInstance(String name)
    {
        if (_parent != null)
            return _parent.getCurrentInstance(name);

        if (_threadInstances == null)
            return null;
        else
            return _threadInstances.getData(name);
    }

    /**
     * Sets the current instance corresponding to specified name for the current thread.
     *
     * @param   name      Name of searched instance.
     * @param   instance  Instance associated to given in current thread (may be null).
     *
     * @return  Previous registered instance if it exists.
     */
    public Object setCurrentInstance(String name, Object instance)
    {
        if (_parent != null)
            return _parent.setCurrentInstance(name, instance);

        if (instance == null)
        {
            if (_threadInstances != null)
                return _threadInstances.removeData(name);
            else
                return null;
        }
        else
        {
            if (_threadInstances == null)
                _threadInstances = new LyThreadData<Object>();

            return _threadInstances.putData(name, instance);
        }
    } // end method setCurrentContext

    /**
     * Gets the current evaluator for the current thread.
     *
     * @return  Instance of the current evaluator. May be null if it was not set in the current
     *          thread.
     */
    public LyEvaluator getCurrentEvaluator()
    {
        LyEvaluator evaluator = (LyEvaluator)getCurrentInstance(CURRENT_EVALUATOR);

        return ((evaluator != null) ? evaluator : getCurrentSession());
    }

    /**
     * Sets the current evaluator to the given one for the current thread.
     *
     * @param   evaluator  Evaluator to set as the current one for the current thread. If null,
     *                     unset the current evaluator.
     *
     * @return  Previous registered evaluator if it exists.
     */
    public LyEvaluator setCurrentEvaluator(LyEvaluator evaluator)
    {
        return (LyEvaluator)setCurrentInstance(CURRENT_EVALUATOR, evaluator);
    } // end method setCurrentContext

    /**
     * Checks is application is being edited.
     * @return True if application is under edition, false otherwise.
     */
    public boolean isEdited()
    {
        return _edited;
    }

    /**
     * Adds a session listener to this application.
     * @param sessionListener
     */
    public void addSessionListener(LySessionListenerInterface sessionListener)
    {
    	if (_sessionListeners == null)
    		_sessionListeners = new Vector<LySessionListenerInterface>(1);
    	_sessionListeners.add(sessionListener);
    }

    /**
     * Removes a session listener from this application.
     * @param sessionListener
     */
    public void removeSessionListener(LySessionListenerInterface sessionListener)
    {
    	if (_sessionListeners != null)
    		_sessionListeners.remove(sessionListener);
    }

    /**
     * Method called on user login.
     */
	protected void userLoggedIn(LyDataSession session, LyObject user)
	{
        // Call listeners
        if (_sessionListeners != null)
        {
        	int		i, n = _sessionListeners.size();

        	for (i = 0; i < n; i++)
        		_sessionListeners.get(i).sessionLoggedIn(this, session, user);
        }
	}

	/**
	 * Find the SQL provider used to store the preferences into DB.
	 *
	 * @return  the SQL data provider to get the connection or null if it doesn't exist.
	 */
	public LySqlDataProvider getPrefDataProvider()
	{
    	String	providerId = getEnvironment().getResource("LY_PREFERENCES_LOCATION_ID");

    	return getSQLDataProvider(providerId);
	}

	/**
	 * find the SQL provider which could be used to store the preferences into DB.
	 *
	 * @param   application  Application on which extra info is saved.
	 * @param   providerId   the provider Id, by default stored in the LY_PREFERENCES_LOCATION_ID
	 *                       resource. Null or empty means : no provider and store info files. "any"
	 *                       is accepted to use the first provider
	 *
	 * @return  the SQL data provider to get the connection or
	 */
	public LySqlDataProvider getSQLDataProvider(String providerId)
	{
		if ((providerId == null)
				|| (providerId.trim().length() == 0))
			return null;

		int				i, n = getDataProviderCount();
		LyDataProvider	provider;

		for (i = 0; i < n; i++)
		{
			provider = getDataProvider(i);

			// return the first one here but could use an identifier
			if ((provider.getApplication() == this)
					&& (providerId.equals(provider.getId())
						|| providerId.equals("any"))
					&& (provider instanceof LySqlDataProvider))
				return ((LySqlDataProvider)provider);
		}

		return null;
	}
} // end class LyApplication
