/*
 * LyDataSession.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.app;

import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.data.*;
import leon.info.*;
import leon.info.misc.*;
import leon.misc.*;

import java.util.*;
import java.util.concurrent.locks.*;


/**
 * This class describes a user session. Each time a user connects itself on an application, a new
 * instance of this class is created. By default, in heavy client versions a session is created when
 * an application is started. In thin client versions (Servlet) a new session is created each time a
 * new user connects himself on the servlet. A session is described by :
 *
 * <UL>
 * <LI><B>an application</B><BR>
 * this application is the owner of this session.</LI>
 * <LI><B>an environment</B><BR>
 * the environment of this session. Having different environments for the different sessions
 * registered on an application allows for example multi-linguism in the same application.</LI>
 * <LI><B>a client interface (_dataBehavior)</B><BR>
 * defined in the meta-model by the <CODE>sessionBehavior</CODE> attribute on the root tag <CODE>
 * &lt;PROJECT&gt;</CODE>. This interface implements some specific behavior of the session.</LI>
 * <LI><B>a list of object sources</B><BR>
 * the object sources used to load the data of the class infos.</LI>
 * <LI><B>a list of controllers</B><BR>
 * this list contains all the controllers that are managed by this session.</LI>
 * <LI><B>a stack of controllers</B><BR>
 * this stack contains all the view controllers that were shown to the user corresponding to this
 * session (used in servlet mode).</LI>
 * </UL>
 *
 * @version  $Id: LyDataSession.java 41630 2012-11-28 13:04:13Z bangiole $
 * @author   Lyria
 * @see      leon.app.LyApplication
 */
public class LyDataSession
    implements LyEvaluator
{
	/**
	 * File version
	 */
	public static final String __VERSION = "$Id: LyDataSession.java 41630 2012-11-28 13:04:13Z bangiole $";

	/** States of the session. */
	public enum State
	{
		/**
		 *  init not yet started.
		 */
		INIT,
		/**
		 * starting : enter start method.
		 */
		STARTING,
		/**
		 *  started : start method exited.
		 */
		STARTED,
		/**
		 *  closing : entered close method.
		 */
		CLOSING,
		/**
		 *  closed : close method exited.
		 */
		CLOSED
	}

	/** Parent session : used for a module. Primary module session doesn't have any parent. */
	protected LyDataSession _dataParent;

	/** The application owner of the session. */
	protected LyApplication _application;

	/** Identifier of the session. */
	protected String _id;

	/** The environment of the session. */
	protected LyEnvironment _environment;

	/** The client interface of this session. */
	protected LyDataSessionBehaviorInterface _dataBehavior;

	/** Associations between keys and values used to evaluate contextual values. */
	protected HashMap<Object, Object> _values;

	/** Contextual object sources of this session. */
	protected Hashtable<String, LyObjectSource> _objectSources;

	/** State of the session. */
	protected State _state;

	/**
	 * Indicates if the session is valid or not. An application is valid if it has been started
	 * (after a login for example).
	 */
	protected boolean _valid;

	/** Indicates if the session is registered on the application. */
	protected boolean _registeredSession;

	/** Info listeners to notify when modifying info objects. */
	protected Vector<_LyInfoListener> _infoListeners;

	/** List of sub-sessions. */
	protected Vector<LyDataSession> _subSessions;

	/** Context manager of the session. */
	protected LyContextManager _contextManager;

	/** Number of sessions. */
	public static int _nbSessions;

	/** Sessions counter. */
    public static int _sessionsCounter;

    private synchronized static int getNextSessionId()
    {
        _sessionsCounter++;
        return _sessionsCounter;
    }

	/** Plugin id for this session. */
	protected String _pluginId;

	/** Time of creation. */
	protected long _creationTime;

	/**
	 * Reentrant lock.
	 * Use a fair lock: the longuest waiting thread will get the lock.
	 */
	ReentrantReadWriteLock _reentrantLock = new ReentrantReadWriteLock(true);


	/**
	 * Constructor with a given application. Registers the session on the application.
	 *
	 * @param  application  The application owner of the session.
	 */
	public LyDataSession(LyApplication application)
	{
		this(application, null, true);
	}

	/**
	 * Constructor with a given application.
	 *
	 * @param  application      The application owner of the session.
	 * @param  registerSession  Indicates if the session must be registered on the application
	 *                          (default is true), or if this is a temporary session which should be
	 *                          ignored.
	 */
	public LyDataSession(LyApplication application, boolean registerSession)
	{
		this(application, null, registerSession);
	}

	/**
	 * Constructor with a given application.
	 *
	 * @param  application      The application owner of the session.
	 * @param  parent           Parent session: used for a module.
	 * @param  registerSession  Indicates if the session must be registered on the application
	 *                          (default is true), or if this is a temporary session which should be
	 *                          ignored.
	 */
	public LyDataSession(LyApplication application, LyDataSession parent, boolean registerSession)
	{
		// Studio / FFT00870
		this(application, parent, registerSession, null);
	}

	// Studio / FFT00870
	/**
	 * Constructor with a given application.
	 *
	 * @param  application      The application owner of the session.
	 * @param  parent           Parent session: used for a module.
	 * @param  registerSession  Indicates if the session must be registered on the application
	 *                          (default is true), or if this is a temporary session which should be
	 *                          ignored.
	 * @param  perspectiveId    The perspective id for this session.
	 */
	public LyDataSession(LyApplication application, LyDataSession parent, boolean registerSession, String perspectiveId)
	{
		// XXX improve sessions and remove registerSession parameter.
		init();

		_application = application;
	    _id = (application.getId() + ".session" + LyDataSession.getNextSessionId()).intern();
		_dataParent = parent;
		_creationTime = System.currentTimeMillis();

		// Creating the session environment.
		// TODO : we must have an identifier for this session ?
		_environment = new LyEnvironment(null, _application.getEnvironment(), true);

		// the perspective for eclipse plugins must be set here !
		if (perspectiveId != null)
			_environment.putResource("LY_PERSPECTIVE", perspectiveId);

		// TODO : Loading specific environment for this session.
		createBehavior();

		if (registerSession)
			register(true);
		else
			_registeredSession = false;

		if (parent != null)
			_dataParent.addSubSession(this);
	}

	/**
	 * Finalize the session
	 */
	protected void finalize()
	{
		_nbSessions--;
	}

    /**
     * Accessor for the identifier of the session.
     *
     * @return  The identifier of the session.
     */
    public String getId()
    {
        return _id;
    }

	/**
	 * Gets state of this session. One of the static codes named STATE_XXX.
	 *
	 * @return  Current state of this session.
	 */
	public final State getState()
	{
		return _state;
	}

	/**
	 * Indicates whether the session is closed or not.
	 *
	 * @return  true if the session is closed, false otherwise.
	 *
	 * @see     LyDataSession#close()
	 */
	public final boolean isClosed()
	{
		return (_state == State.CLOSED);
	}

	/**
	 * Indicates whether the session is valid or not.
	 *
	 * @return  true if the session is valid, false otherwise.
	 */
	public boolean isValid()
	{
		return _valid;
	}

	/**
	 * Indicates whether the session is registered or not.
	 *
	 * @return  true if the session is registered on its application, false otherwise.
	 *
	 * @see     LyDataSession#register(boolean)
	 */
	public boolean isRegistered()
	{
		return _registeredSession;
	}

	/**
	 * Returns the environment of the session.
	 *
	 * @return  The environment of the session.
	 *
	 * @see     leon.misc.LyEnvironment
	 */
	public LyEnvironment getEnvironment()
	{
		if (_environment == null)
			return _application.getEnvironment();

		return _environment;
	}

	/**
	 * Returns the application owner of the session.
	 *
	 * @return  The application owner of the session.
	 *
	 * @see     leon.app.LyApplication
	 */
	public LyApplication getApplication()
	{
		return _application;
	}

	/**
	 * Returns the parent session.
	 *
	 * @return  The parent session.
	 */
	public LyDataSession getParent()
	{
		return _dataParent;
	}

	/**
	 * Returns the client interface of this session.
	 *
	 * @return  The client interface of this session.
	 *
	 * @see     leon.app.behaviorinterface.LyDataSessionBehaviorInterface
	 */
	public LyDataSessionBehaviorInterface getSessionBehavior()
	{
		return _dataBehavior;
	}

	/**
	 * Returns the time of creation of this session.
	 *
	 * @return  the creation time in milliseconds.
	 */
	public long getCreationTime()
	{
		return _creationTime;
	}

	/**
	 * Validates the session.
	 *
	 * @param valid a boolean indicating if the session is valid
	 */
	public void setValid(boolean valid)
	{
		_valid = valid;
	}

	/**
	 * Initialisation of the data of the instance. This method is called by the constructor.
	 */
	protected void init()
	{
		// Initialize class attributes.
		_dataParent = null;
		_application = null;
		_environment = null;
		_dataBehavior = null;
		_values = null;
		_objectSources = null;
		_state = State.INIT;
		_valid = false;
		_infoListeners = null;

		_nbSessions++;
	}

	/**
	 * Creates the client interface for this session. If no specific behavior is given for the
	 * session in the meta-model (using the <CODE>sessionBehavior</CODE> attribute on the root tag
	 * <CODE>&lt;PROJECT&gt;</CODE>), an instance of LyDataSessionBehavior is created. Otherwise an
	 * instance of the specified class is created.
	 *
	 * @see  leon.app.behaviorinterface.LyDataSessionBehaviorInterface
	 * @see  leon.app.behavior.LyDataSessionBehavior
	 */
	protected void createBehavior()
	{
		LyProject project = _application.getProject();
		Class<?>  behaviorClass = null;

		if (project != null)
			behaviorClass = project.getSessionBehaviorClass();

		if (behaviorClass == null)
			behaviorClass = LyDataSessionBehavior.class;
		else if (!LyDataSessionBehaviorInterface.class.isAssignableFrom(behaviorClass))
			throw new IllegalStateException("BF AppEngine : invalid session behavior "
				+ behaviorClass + ", it must implement " + LyDataSessionBehaviorInterface.class);

		try
		{
			_dataBehavior = (LyDataSessionBehaviorInterface)behaviorClass.newInstance();
		}
		catch (Exception e)
		{
			_environment.writeException(e);
		}
	}

	/**
	 * Registers the session on its application owner. This method starts the session if the given
	 * parameter is true by calling the start() method.
	 *
	 * @param      start  Indicates whether the session must be started or not.
	 *
	 * @exception  java.lang.IllegalStateException  if the session has already been registered.
	 */
	public void register(boolean start)
	{
		if (_registeredSession)
			throw new IllegalStateException("BF AppEngine : session is already registered");

		_registeredSession = true;
		_application.addSession(this);
	}

	/**
	 * Starts the session.
	 */
	protected void internalStart()
	{
		_state = State.STARTING;

		// Call behavior
		_dataBehavior.start(this);

		if (_state != State.CLOSED)
			_state = State.STARTED;
	}

	/**
	 * Test if the session has some pending transaction.
	 *
	 * @return  true if this session has somme pending transaction, else false.
	 */
	public boolean hasTransaction()
	{
		LyApplication application = getApplication();

		return application.hasTransaction(this);
	}

	/**
	 * Closes the session. Removes the session from its application, sets all the controllers, the
	 * object sources and the view manager free.
	 */
	public void close()
	{
		if (_state == State.CLOSED)
			return;

		LyDataSession oldCurrentSession = _application.getCurrentSession();

		try
		{
			_application.setCurrentSession(this);

			// Call behavior
			_dataBehavior.close(this);

			if (_state == State.CLOSING)
			{
				// Close sub-sessions
				if (_subSessions != null)
				{
					int			  i, n = _subSessions.size();
					LyDataSession session;

					for (i = n - 1; i >= 0; i--)
					{
						session = _subSessions.elementAt(i);
						session.close();
					}
				}

				// Remove all specific object sources.
				if (_objectSources != null)
				{
					LyObjectSource                 objectSource;
					Enumeration<LyObjectSource>    e = _objectSources.elements();

					while (e.hasMoreElements())
					{
						objectSource = e.nextElement();
						objectSource.free();
					}

					_objectSources = null;
				}

				// Free all resources.
				if (_environment != null)
				{
					// Free all specific environment resource bundles.
					LyResourceBundle.removeBundles(_environment);

					_environment.free();
				}
			}
		}
		catch (Exception e)
		{
			getEnvironment().writeException(e);
		}
		finally
		{
			if (_state == State.CLOSING)
			{
				// Remove this session on its parent
				if (_dataParent != null)
					_dataParent.removeSubSession(this);

				if ((oldCurrentSession != null)
						&& !oldCurrentSession.isClosed())
					// Unset current session
					_application.setCurrentSession(oldCurrentSession);

				// Unregister this session of application
				if (_registeredSession)
				{
					_application.removeSession(this);
					_registeredSession = false;
				}

				_state = State.CLOSED;
				_valid = false;
			}
		}
	} // end method close

	/**
	 * Adds a pair key-value to the session. May be used for example to store the login used to
	 * create the session.
	 *
	 * @param  key    The key that must be stored.
	 * @param  value  The value that must be stored with the key.
	 *
	 * @see    #getValue
	 */
	public void addValue(Object key, Object value)
	{
		if (_values == null)
			_values = new HashMap<Object, Object>();

		_values.put(key, value);
	}

	/**
	 * Evaluates the given expression.
	 *
	 * @param  evalExpression  the expression to evaluate.
	 */
	private Object evaluateExpression(final String expression)
	{
		if (expression == null)
            return null;

        int     p = expression.indexOf('.');

        if (p < 0)
            return null;

        String  key = expression.substring(0, p);
		Object  value = _values.get(key);
		String  fieldName = expression.substring(p + 1);

        if (value instanceof LyObjectId)
        {
            LyObject    object = ((LyObjectId)value).getLogicalObject();

    		return (object == null ? null : object.getValue(fieldName));
        }
        else
            return null;
	}

	/**
	 * Retrieves a value from a registered key.
	 *
	 * @param   key  The key that must be used.
	 *
	 * @return  The value stored with the key or null if the key is not registered.
	 *
	 * @see     #addValue
	 */
	public Object getValue(Object key)
	{
		if ((_values == null) || (key == null))
			return null;

		Object result = _values.get(key);

		if (result != null)
			return result;
        else if (key instanceof String)
			return evaluateExpression(key.toString());
		else
			return null;
	}

	/* Object Sources Methods */

	/**
	 * Registers an object source on this session.
	 *
	 * @param  objectSource  The object source that must be registered.
	 *
	 * @see    leon.data.LyObjectSource
	 * @see    #getObjectSource
	 */
	public void putObjectSource(LyObjectSource objectSource)
	{
		LyClassInfo classInfo = objectSource.getClassInfo();

		if (_objectSources == null)
			_objectSources = new Hashtable<String, LyObjectSource>();
		else if (_objectSources.get(classInfo.getId()) != null)
			throw new IllegalStateException("BF AppEngine : an object source for class " + classInfo
				+ " is already registered on session " + this);

		_objectSources.put(classInfo.getId(), objectSource);
	}

	/**
	 * Unregisters an object source on this session.
	 *
	 * @param  objectSource  The object source that must be unregistered.
	 *
	 * @see    leon.data.LyObjectSource
	 * @see    #getObjectSource
	 */
	public void removeObjectSource(LyObjectSource objectSource)
	{
		LyClassInfo classInfo = objectSource.getClassInfo();

		if (getObjectSource(classInfo) == null)
			throw new IllegalStateException("BF AppEngine : an object source for class " + classInfo
				+ " was not registered on session " + this);

		_objectSources.remove(classInfo.getId());
	}

	/**
	 * Retrieves an object source on this session.
	 *
	 * @param   classInfo  The class corresponding to the object source that must be returned.
	 *
	 * @return  The object source corresponding to the given class info.
	 *
	 * @see     leon.data.LyObjectSource
	 * @see     #putObjectSource
	 */
	public LyObjectSource getObjectSource(LyClassInfo classInfo)
	{
		return (LyObjectSource)getDataSource(classInfo);
	}

	/**
	 * Retrieves a list of objects by their names from a list of classes.
	 *
	 * @param   classInfos   the list of target classes used to retrieve the objects.
	 * @param   objectNames  the list of object names used to retrieve the objects.
	 *
	 * @return  the list of objects matching the given object names for given list of classes.
	 */
	public LyObjectList getObjectsByName(LyInfoList<LyClassInfo> classInfos, String[] objectNames)
	{
		int i, n = ((classInfos == null) ? 0 : classInfos.getSize());
		int j, m = ((objectNames == null) ? 0 : objectNames.length);

		// If no class info given or no name given, return null
		if ((n <= 0) || (m <= 0))
			return null;

		LySimpleObjectList	    result = null;
		LyClassInfo			    classInfo;
		LyInfoList<LyFieldInfo> nameFields;

		// For each class info, try to retrieve objects matching given names
		for (i = 0; i < n; i++)
		{
			classInfo = classInfos.getInfo(i);
			nameFields = classInfo.getNameFieldInfos();

			if ((nameFields == null) || (nameFields.getSize() == 0))
			{
				// Nothing to do : no name field specified
			}
			else if (nameFields.getSize() == 1)
			{
				// One field only : try to create a filter
				LyFilter		   filter = new LyFilter(LyFilter.Type.OR);
				LyFilterExpression exp = null;

				for (j = 0; j < m; j++)
				{
					exp = new LyFilterExpression(nameFields.getInfo(0), LyFilterExpression.Condition.EQUALS,
							objectNames[j]);

					filter.addElement(exp);
				}

				LyClassWorkSpace ws = new LyClassWorkSpace(this, classInfo, filter, false);

				if (ws.getSize() > 0)
				{
					if (result == null)
						result = new LySimpleObjectList();

					result.addList(ws);
				}
			}
			else
			{
				// Multiple fields : could we increase the efficiency ?
				LyClassWorkSpace ws = new LyClassWorkSpace(this, classInfo, null, false);

				if (ws.getSize() > 0)
				{
					int		 k, l = ws.getSize();
					LyObject obj;
					String   name;

					for (k = 0; k < l; k++)
					{
						obj = ws.getObject(k);
						name = obj.getName();

						for (j = 0; j < m; j++)
						{
							if (name.equals(objectNames[j]))
							{
								if (result == null)
									result = new LySimpleObjectList();

								result.addObject(obj);
							}
						}
					}
				}
			}
		}

		return result;
	} // end method getObjectsByName

	/**
	 * Retrieves a data source on this session.
	 *
	 * @param   classId  The identifier of the class corresponding to the object source that must be
	 *                   returned.
	 *
	 * @return  The object source corresponding to the given class identifier.
	 *
	 * @see     leon.data.LyDataSource
	 * @see     #getDataSource(LyClass)
	 */
	public LyDataSource getDataSource(String classId)
	{
		LyDataSource dataSource = null;

		if (_objectSources != null)
			dataSource = _objectSources.get(classId);

		if (dataSource == null)
			dataSource = _application.getDataSource(classId);

		return dataSource;
	}

	/**
	 * Retrieves a data source on this session.
	 *
	 * @param   lyClass  The class corresponding to the object source that must be returned.
	 *
	 * @return  The object source corresponding to the given class.
	 *
	 * @see     leon.data.LyDataSource
	 * @see     #getDataSource(String)
	 */
	public LyDataSource getDataSource(LyClass lyClass)
	{
		return getDataSource(lyClass.getId());
	}

	/**
	 * Called when an exception occured during proccessing.
	 *
	 * @param  t  The throwable instance that causes the stop of the processing.
	 */
	public void handleException(Throwable t)
	{
		LyEnvironment environment = getEnvironment();
		LyPlayer	  player = environment.getPlayer();

		if (player != null)
			player.handleException(t);

		LyDataSessionBehaviorInterface behavior = getSessionBehavior();

		if (behavior != null)
			behavior.handleException(this, t);
	}

	/**
	 * Adds an info listener to the application.
	 *
	 * @param  listener  Info listener that must be added.
	 * @param  prefix    Prefix of the files from which objects info were loaded.
	 */
	public void addInfoListener(LyInfoListener listener, String prefix)
	{
		if (_infoListeners == null)
			_infoListeners = new Vector<_LyInfoListener>();

		_LyInfoListener infoListener = new _LyInfoListener(prefix, listener);

		_infoListeners.addElement(infoListener);
	}

	/**
	 * Adds an info listener to the application.
	 *
	 * @param  listener  Info listener that must be removed.
	 */
	public void removeInfoListener(LyInfoListener listener)
	{
		if (_infoListeners == null)
			return;

		int			    i, n = _infoListeners.size();
		_LyInfoListener infoListener;

		for (i = n - 1; i >= 0; i--)
		{
			infoListener = _infoListeners.elementAt(i);

			// Remove all occurences of current listener
			if (infoListener._listener == listener)
				_infoListeners.removeElementAt(i);
		}
	}

	/**
	 * Notify info listeners on a class that an event has occured.
	 *
	 * @param  prefix  Prefix of the files from which objects info were loaded.
	 */
	public void notifyInfoListeners(String prefix)
	{
		if (_infoListeners == null)
			return;

		int			    i, n = _infoListeners.size();
		_LyInfoListener infoListener;

		for (i = 0; i < n; i++)
		{
			infoListener = _infoListeners.elementAt(i);

			if ((prefix == null)
					|| (prefix.equals(infoListener._prefix)))
				infoListener._listener.infoEvent(infoListener._prefix);
		}
	}

	/**
	 * Internal class to register a listener on info objects.
	 */
	class _LyInfoListener
	{

		/** Prefix of the info object. */
		protected String _prefix = null;

		/** Listener. */
		protected LyInfoListener _listener = null;

		/**
		 * Constructor.
		 */
		_LyInfoListener(String prefix, LyInfoListener listener)
		{
			_prefix = prefix;
			_listener = listener;
		}
	}

	/**
	 * Internal method called when data have been unloaded. This method frees dialog controllers if
	 * necessary.
	 */
	void dataUnloaded()
	{
	}

	/**
	 * Get sub-sessions of this session.
	 *
	 * @return  A vector containing all sub-sessions, null if no sub-sessions found.
	 */
	public Vector<LyDataSession> getSubSessions()
	{
		return ((_subSessions == null) ? null : new Vector<LyDataSession>(_subSessions));
	}

	/**
	 * Add a sub-session to this session.
	 *
	 * @param  session  Sub-session added to this session.
	 */
	protected void addSubSession(LyDataSession session)
	{
		if (_subSessions == null)
			_subSessions = new Vector<LyDataSession>();

		_subSessions.add(session);
	}

	/**
	 * Remove a sub-session to this session.
	 *
	 * @param  session  Sub-session removed to this session.
	 */
	protected void removeSubSession(LyDataSession session)
	{
		_subSessions.remove(session);
	}

	/**
	 * Gets the context manager of this application.
	 *
	 * @return  The associated context manager.
	 */
	public LyContextManager getContextManager()
	{
		if (_dataParent != null)
			return _dataParent.getContextManager();

		if (_contextManager == null)
			_contextManager = new LyContextManager(this);

		return _contextManager;
	}

	/**
	 * Reset the undo redo stack
	 */
	public void resetUndoRedo()
	{
	}

	/**
	 * Reset the session
	 */
	protected void reset()
	{
	}

	/**
	 * Set the plugin id for the session.
	 *
	 * @param  pluginId  The plugin id.
	 */
	public void setPluginId(String pluginId)
	{
		_pluginId = pluginId;
	}

	/**
	 * Get the plugin id for this session.
	 *
	 * @return  The plugin id.
	 */
	public String getPluginId()
	{
		return _pluginId;
	}

	/**
	 * Accessor to the user of current session.
	 *
	 * @return  The value of $USER value stored by the login controller if any, otherwise return
	 *          null
	 */
	public LyObject getUser()
	{
        LyObjectId userId = (LyObjectId)getValue("$USER");

        if (userId == null)
        {
            LyObject    user = _dataBehavior.getUser(this);

            if (user != null)
                setUser(user);
            return user;
        }
        else
            return userId.getLogicalObject();
	}

	/**
	 * Checks whether an object is visible in this session or not.
	 * Object may not be visible if a filtered object source has been created on its class info.
	 *
	 * @param object the object that must be checked
	 * @return true if object is visible, false otherwise
	 */
	public boolean isObjectVisible(LyObject object)
	{
		if (object == null)
			return true;

		LyClassInfo		classInfo = object.getClassInfo();
		LyDataSource	dataSource = getObjectSource(classInfo);

		if (dataSource instanceof LyFilteredObjectSource)
		{
			LyFilteredObjectSource	fos = (LyFilteredObjectSource)dataSource;
			LyFilter				filter = fos.getFilter();

			if ((filter != null) && !filter.match(object))
				return false;
		}

		return true;
	}

    /**
     * Retrieves a value from a registered key.
     *
     * @param   key  The key that must be used.
     * @return  The value stored with the key or null if the key is not registered.
     */
    public Object getKeyValue(String key)
    {
        if (_dataBehavior != null)
        {
            Object  value = _dataBehavior.getKeyValue(key);

            if (value != null)
                return value;
        }

        return getValue("$" + key);
    }

    public void dumpLocks(boolean before)
    {
    	if (!getEnvironment().isDebugLevel(LyEnvironment.DEBUG_LOCKS))
    		return;

    	if (before)
    		getEnvironment().logOut("Session = " + this.getId());
    	else
    		getEnvironment().logOut("-----");
    	getEnvironment().logOut("Read lock count " + _reentrantLock.getReadLockCount());
		getEnvironment().logOut("Write hold count " + _reentrantLock.getWriteHoldCount());

		if (!before && getEnvironment().isDebugLevel(LyEnvironment.DEBUG_ALL))
			getEnvironment().writeException(new Exception("Lock Stack"));
    }

	/**
	 * Get a soft lock on this session.
	 * Internal Use Only.
	 */
	public void getSoftLock()
	{
		if (!getEnvironment().testResource("LY_USE_LOCK_FOR_NOTIFIER_EVENTS"))
			return;
		dumpLocks(true);
		_reentrantLock.readLock().lock();
		dumpLocks(false);
	}

	/**
	 * Release a soft lock on this session.
	 * Internal Use Only.
	 */
	public void releaseSoftLock()
	{
		if (!getEnvironment().testResource("LY_USE_LOCK_FOR_NOTIFIER_EVENTS"))
			return;
		dumpLocks(true);
		_reentrantLock.readLock().unlock();
		dumpLocks(false);
	}

	/**
	 * Get a soft lock on this session.
	 * Internal Use Only.
	 */
	public void getHardLock()
	{
		if (!getEnvironment().testResource("LY_USE_LOCK_FOR_NOTIFIER_EVENTS"))
			return;
		dumpLocks(true);
		_reentrantLock.writeLock().lock();
		dumpLocks(false);
	}

	/**
	 * Release a soft lock on this session.
	 * Internal Use Only.
	 */
	public void releaseHardLock()
	{
		if (!getEnvironment().testResource("LY_USE_LOCK_FOR_NOTIFIER_EVENTS"))
			return;
		dumpLocks(true);
		_reentrantLock.writeLock().unlock();
		dumpLocks(false);
	}

	/**
	 * Sets the user of this session.
	 */
	public void setUser(LyObject user)
	{
		if (user == null)
			addValue("$USER", null);
		else
		{
			LyObjectId	currentUserId = (LyObjectId)getValue("$USER");

			if (currentUserId == null)
			{
				LyObjectId userId = user.getObjectId();

				addValue("$USER", userId);
			}
		}

		_application.userLoggedIn(this, user);
	}

	/**
	 * Set the current time zone for the user
	 * @param timeZoneId
	 */
	public void setTimeZone(String timeZone)
	{
		// register time zone in session environment
		getEnvironment().putEnv("LY_TIMEZONE", timeZone);
//		addValue("$TIMEZONE", timeZone);
	}

} // end class LyDataSession
