/*
 * LyMain.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app;

import java.io.*;
import java.lang.reflect.*;
import leon.misc.*;


/**
 * This class contains the main method used to create Leonardi applications. It must be used for fat
 * clients (win viewer, swing viewer...). The usage is:
 *
 * <UL>
 *   <li><CODE>java LyMain [appname &lt;project&gt;.ini]+</CODE></li>
 * </UL>
 *
 * <p>where <CODE>appname</CODE> is the name of the application and <CODE>&lt;project&gt;.ini</CODE>
 * the path of the ini file.</p>
 *
 * @author  Lyria S.A.
 * @version $Id: LyMain.java 34859 2012-03-19 14:32:04Z jlsaouli $
 * @see     leon.app.LyApplication
 * @see     leon.app.LySession
 */
public class LyMain
{
	/**
	 * File version
	 */
	public static final String __VERSION = "$Id: LyMain.java 34859 2012-03-19 14:32:04Z jlsaouli $";

	/**
	 * Main method. For each {@link LyApplication} started it creates a new {@link LySession} based
	 * on this application.
	 *
	 * @param argv Parameters that describe applications that must be created (see usage in the
	 *             class description).
	 * @see   leon.app.LyApplication
	 * @see   leon.app.LySession
	 */
	public static void main(String[] argv)
	{
		int nv = ((argv == null) ? 0 : argv.length);

		if ((nv == 0) || ((nv > 1) && ((nv % 2) == 1)) && (nv != 3))
		{
			System.err.println("usage: java LyMain [appname <project>.ini]+");
			System.exit(1);
		}

		if (argv.length == 1)
			startApplication(argv[0], null, null);
		else if( argv.length == 3 )
			startApplicationWithChosenLanguage(argv[0], argv[1], argv[2]);
		else
		{
			int i;

			for (i = 0; i < nv; i += 2)
				startApplication(argv[i], argv[i + 1], null);
		}
	}

	/**
	 * Method used to start an application.<br />
	 * Calls {@link #startApplication(String, String, String, ClassLoader) startApplication(appname,
	 * iniFile, null, null)}.
	 *
	 * @param appname Identifier of the application to start.
	 * @param iniFile Initialization file of the application (may be null).
	 * @see   leon.app.LyApplication
	 * @see   leon.app.LySession
	 * @see   #startApplication(String, String, String, ClassLoader)
	 */
	public static void startApplication(String appname, String iniFile)
	{
		startApplication(appname, iniFile, null, null, null);
	}
	
	/**
	 * Method used to start an application.<br />
	 * Calls {@link #startApplication(String, String, String, ClassLoader) startApplication(appname,
	 * iniFile, null, null)}.
	 *
	 * @param appname Identifier of the application to start.
	 * @param iniFile Initialization file of the application (may be null).
	 * @param chosenLanguage the desired language
	 * @see   leon.app.LyApplication
	 * @see   leon.app.LySession
	 * @see   #startApplication(String, String, String, ClassLoader)
	 */
	public static void startApplicationWithChosenLanguage(String appname, String iniFile, String chosenLanguage)
	{
		startApplication(appname, iniFile, null, null,  chosenLanguage);
	}
	

	/**
	 * Method used to start an application.<br />
	 * Calls {@link #startApplication(String, String, String, ClassLoader) startApplication(appname,
	 * iniFile, perspectiveId, null)}.
	 *
	 * @param appname       Identifier of the application to start.
	 * @param iniFile       Initialization file of the application (may be null).
	 * @param perspectiveId The persectiveId.
	 * @see   leon.app.LyApplication
	 * @see   leon.app.LySession
	 * @see   #startApplication(String, String, String, ClassLoader)
	 */
	public static void startApplication(String appname, String iniFile, String perspectiveId)
	{
		startApplication(appname, iniFile, perspectiveId, null, null);
	}

	/**
	 * Method used to start an application.
	 *
	 * @param appname       I dentifier of the application to start.
	 * @param iniFile        Initialization file of the application (may be null).
	 * @param perspectiveId  The persectiveId.
	 * @param classLoader    The applicationClassLoader.
	 * @see   leon.app.LyApplication
	 * @see   leon.app.LySession
	 */
	public static void startApplication(String appname, String iniFile, String perspectiveId,
		ClassLoader classLoader)
	{
		startApplication(appname, iniFile, perspectiveId, classLoader, null);
	}
	
	/**
	 * Method used to start an application.
	 *
	 * @param appname         Identifier of the application to start.
	 * @param iniFile         Initialization file of the application (may be null).
	 * @param perspectiveId   The persectiveId.
	 * @param classLoader     The applicationClassLoader.
	 * @param chosenLanguage  Language of the application to start (may be null).
	 * @see   leon.app.LyApplication
	 * @see   leon.app.LySession
	 */
	public static void startApplication(String appname, String iniFile, String perspectiveId,
		ClassLoader classLoader, String chosenLanguage)
	{
        if (iniFile == null)
        {
            String     name = appname.substring(appname.lastIndexOf('.') + 1);
            
            iniFile = appname.replace('.', File.separatorChar) + File.separatorChar + name + ".ini";
        }
	       
		LyApplication application = null;
		LySession     session = null;

		try
		{
			application = new LyApplication(appname, iniFile, classLoader);
			
			if (application.isValid())
			{
				LyEnvironment environment = application.getEnvironment();
				
				if (chosenLanguage != null)
					environment.setLanguage(chosenLanguage);
				
				if (environment.testEnv("LY_TEST_MONITOR"))
				{
					try
					{
						@SuppressWarnings("unchecked")
						Class<LySession> testSessionClass = (Class<LySession>)application.loadClass(
								"eu.w4.ae.test.monitor.LyTestSession");

						if (testSessionClass != null)
						{
							Constructor<LySession> constructor = testSessionClass.getConstructor(
									new Class[]
									{
										LyApplication.class, String.class
									});

							session = constructor.newInstance(
									new Object[]
									{
										application, perspectiveId
									});
						}
					}
					catch (Exception e)
					{
						environment.writeException(e);
					}
				}
				
				if (session == null)
					new LySession(application, null, perspectiveId, true);
			}
		}
		catch (RuntimeException e)
		{
			e.printStackTrace();

			// If anything occured, stop properly
			if (session != null)
				session.close();
			if (application != null)
				application.close();
		}
	} // end method startApplication
} // end class LyMain
