/*
 * LyGuiSession.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app;

import java.lang.reflect.*;
import java.util.*;

import leon.view.LyViewManager;

import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * This class describes a user session. Each time a user connects itself on an application, a new
 * instance of this class is created. By default, in heavy client versions a session is created when
 * an application is started. In thin client versions (Servlet) a new session is created each time a
 * new user connects himself on the servlet. A session is described by :
 *
 * <UL>
 * <LI><B>an application</B><BR>
 * this application is the owner of this session.</LI>
 * <LI><B>an environment</B><BR>
 * the environment of this session. Having different environments for the different sessions
 * registered on an application allows for example multi-linguism in the same application.</LI>
 * <LI><B>a client interface (_behavior)</B><BR>
 * defined in the meta-model by the <CODE>sessionBehavior</CODE> attribute on the root tag <CODE>
 * &lt;PROJECT&gt;</CODE>. This interface implements some specific behavior of the session.</LI>
 * <LI><B>a list of object sources</B><BR>
 * the object sources used to load the data of the class infos.</LI>
 * <LI><B>a list of controllers</B><BR>
 * this list contains all the controllers that are managed by this session.</LI>
 * <LI><B>a stack of controllers</B><BR>
 * this stack contains all the view controllers that were shown to the user corresponding to this
 * session (used in servlet mode).</LI>
 * </UL>
 *
 * @version  $Id: LySession.java 40832 2012-09-18 08:03:31Z smaupin $
 * @author   Lyria
 * @see      leon.app.LyApplication
 */
public class LySession extends LyDataSession
    implements LyTimerClientInterface,
        LyGuiClientInterface
{
	/** File version. */
    public static final String __VERSION = "$Id: LySession.java 40832 2012-09-18 08:03:31Z smaupin $";

    /** Default session time out (no time out). */
    public static final short SESSION_TIMEOUT = -1;

    /** The Exit demand. */
    public static final String DEMAND_EXIT = "_exit".intern();

    /** Default string. */
    public static final String DEFAULT = "_default".intern();

    /** DraggedObject string. */
    public static final String DRAGGED_OBJECT = "draggedObject".intern();

    /** Metric object id constant. */
    public static final String METRIC_OBJECT_ID_KEY = "$METRIC_OBJECT_ID_".intern();

    /** Popup view String. */
    public static final String POPUP_VIEW = "popup_view".intern();

    /** Parent session : used for a module. Primary module session doesn't have any parent. */
    protected LySession _parent;

    /** The client interface of this session. */
    protected LySessionBehaviorInterface _behavior;

    /** Table of controllers owned by this session (keys are controllers id). */
    protected Hashtable<String, LyController> _controllers;

    /** List of controllers owned by this session. */
    protected Vector<LyController> _controllersList;

    /**
     * Last controllers which have been removed from stack, but which are still in the
     * controllerList. (Used for clearViewCache)
     */
    protected Vector<LyViewController> _recentControllerStack;

    /**
     * Stacks of shown controllers. The stack is on per-view basis (default, target_popup_window
     * ...) This stack is used to navigate through the views shown to the user (used in servlet
     * mode).
     */
    protected Hashtable<String, Stack<LyViewController>> _controllersStackList;

    /**
     * Keep the view cache count for enhance further access. This value is LY_VIEW_CACHE_COUNT, read
     * from the ressource files and is for internal use only
     */
    protected int _viewCacheCount;

    /** Keep a reference of the timer for the session timeout. */
    protected LyTimer _timeoutTimer;

    /** The outline controllers for this session. */
    protected Hashtable<String, Vector<LyOutlineController>> _outlineControllers;

    /** Data listeners to notify on reset. */
    protected Vector<LyResetInterface> _resetListeners;

    /** Plugin id for this session. */
    // protected String                                              _pluginId;

    /** Indicators to detect when we must free dialog controllers when a session reset occured. */
    protected boolean _dialogsToFree;

    /** Current number of invoked actions. */
    protected int _invokedActions;

    /** Boolean set when the session is idle. */
    protected boolean _idle;

    /**
     * Master session. Used in web mode only to indicate that current session is a "slave" session
     * managed by another one.
     */
    protected LySession _masterSession;

    /**
     * List of slave sessions. Used in web mode only to indicate that current session is a "master"
     * session that manages other ones.
     */
    protected Vector<LySession> _slaveSessions;

    /**
     * A map of temporary file: store a key and an array of string containing: the complete path and
     * the name of the file.
     */
    protected Map<String, String[]> _tempFileMap;

    /**
     * Is session in chat.
     */
    protected boolean _chatting = false;

    /**
     * Counter for metrics idle mode.
     */
    protected int _metricsIdleMode;


    /**
     * Constructor with a given application. Registers the session on the application.
     *
     * @param  application  The application owner of the session.
     */
    public LySession(LyApplication application)
    {
        this(application, null, null, true);
    }

    /**
     * Constructor with a given application.
     *
     * @param  application      The application owner of the session.
     * @param  registerSession  Indicates if the session must be registered on the application
     *                          (default is true), or if this is a temporary session which should be
     *                          ignored.
     */
    public LySession(LyApplication application, boolean registerSession)
    {
        this(application, null, null, registerSession);
    }

    /**
     * Constructor with a given application.
     *
     * @param  application      The application owner of the session.
     * @param  parent           Parent session: used for a module.
     * @param  registerSession  Indicates if the session must be registered on the application
     *                          (default is true), or if this is a temporary session which should be
     *                          ignored.
     */
    public LySession(LyApplication application, LySession parent, boolean registerSession)
    {
        this(application, parent, null, registerSession);
    }

    /**
     * Constructor with a given application.
     *
     * @param  application      The application owner of the session.
     * @param  parent           Parent session: used for a module.
     * @param  perspectiveId    The perspective id for this session.
     * @param  registerSession  Indicates if the session must be registered on the application
     *                          (default is true), or if this is a temporary session which should be
     *                          ignored.
     */
    public LySession(LyApplication application, LySession parent, String perspectiveId, boolean registerSession)
    {
        // Studio / FFT00870
        super(application, parent, registerSession, perspectiveId);

        _parent = parent;
        _masterSession = null;
        _slaveSessions = null;
        _metricsIdleMode = 0;
    }

    /**
     * Returns the application owner of the session.
     *
     * @return  The application owner of the session.
     *
     * @see     leon.app.LyApplication
     */
    public LyApplication getApplication()
    {
        return _application;
    }

    /**
     * Returns the session timeout in seconds.
     *
     * @return  the session timeout
     */
    public int getSessionTimeout()
    {
        return _behavior.getSessionTimeout(this);
    }

    /**
     * Creates the client interface for this session. If no specific behavior is given for the
     * session in the meta-model (using the <CODE>sessionBehavior</CODE> attribute on the root tag
     * <CODE>&lt;PROJECT&gt;</CODE>), an instance of LySessionBehavior is created. Otherwise an
     * instance of the specified class is created.
     *
     * @see  leon.app.behaviorinterface.LySessionBehaviorInterface
     * @see  leon.app.behavior.LySessionBehavior
     */
    protected void createBehavior()
    {
        Class<?> behaviorClass = null;

        try
        {
            LyProject project = _application.getProject();

            if (project != null)
                behaviorClass = project.getSessionBehaviorClass();

            if (behaviorClass == null)
                behaviorClass = LySessionBehavior.class;
            else if (!LySessionBehaviorInterface.class.isAssignableFrom(behaviorClass))
                throw new IllegalStateException("BF AppEngine : invalid session behavior "
                    + behaviorClass + ", it must implement " + LySessionBehaviorInterface.class);
        }
        catch (Exception e)
        {
            _environment.writeException(e);
            behaviorClass = LySessionBehavior.class;
        }

        try
        {
            _behavior = (LySessionBehaviorInterface)behaviorClass.newInstance();
            _dataBehavior = _behavior;
        }
        catch (Exception e)
        {
            _environment.writeException(e);
        }
    } // end method createBehavior

    /**
     * Registers the session on its application owner. This method starts the session if the given
     * parameter is true by calling the start() method.
     *
     * @param      start  Indicates whether the session must be started or not.
     *
     * @exception  java.lang.IllegalStateException  if the session has already been registered.
     *
     * @see        #start
     */
    public void register(boolean start)
    {
        super.register(start);

        if (start)
        {
            // Asks GUI start to view manager
            LyViewManager viewManager = getViewManager();

            if (viewManager != null)
                viewManager.startGui(this);
            else
            {
                _state = State.STARTING;

                // Call behavior
                if (_behavior != null)
                    _behavior.start(this);

                _state = State.STARTED;
            }
        }
    }

    /**
     * Gets root action for this session.
     *
     * @return  the root action
     */
    protected LyAction getRootAction()
    {
        return _behavior.getRootAction(this);
    }

    /**
     * Starts the session by invoking the root action. If no root action is defined in the
     * meta-model then the '_login' action is started.
     *
     * @return  The controller of the root action or null if nothing was done.
     */
    public LyController start()
    {
        long		startTime = 0;

        if (LyMetricsTools.areMetricsActivated(this, LyMetricsTools.MetricType.SESSION))
        	startTime = System.currentTimeMillis();

        _state = State.STARTING;

        // Call behavior
        _behavior.start(this);

        if (_state == State.CLOSED)
            return null;

        LyAction     rootAction = getRootAction();
        LyController rootController = null;

        if (!tryAutoLogin(rootAction))
            // Invoke root action
            rootController = invokeAction(null, rootAction, null, null, true);

        _state = State.STARTED;

        // We must start the timer if needed
        resetTimeout();

        if (!_environment.testEnv("LY_TEST_MONITOR"))
        {
            try
            {
                Class<?> runnerClass = null;

                if (_environment.testEnv("LY_REPLAY"))
                    runnerClass = _application.loadClass("eu.w4.ae.test.LyTestPlayer");
                else if (_environment.testEnv("LY_RECORD"))
                    runnerClass = _application.loadClass("eu.w4.ae.test.LyTestRecorder");

                if (runnerClass != null)
                {
                    Constructor<?> constructor = runnerClass.getConstructor(new Class[] { LySession.class });

                    constructor.newInstance(new Object[] { this });
                }
            }
            catch (Exception e)
            {
                _environment.writeException(e);
            }
        }

		// Create session metric if needed
    	Long		startDuration = null;

    	if (startTime > 0)
    		startDuration = new Long(System.currentTimeMillis() - startTime);

		LyObject	metric = LyMetricsTools.createMetric(this, LyMetricsTools.MetricType.SESSION,
				new Date(), startDuration, null, null, null, null);

		if (metric != null)
			addValue(METRIC_OBJECT_ID_KEY, metric.getObjectId());

        return rootController;
    } // end method start

    /**
     * Indicates if an action is valid for a given controller. This method calls the same method on
     * the client interface.
     *
     * @param   action      The action that must be tested.
     * @param   controller  The controller used to control the action validity.
     *
     * @return  true if the action is valid, false otherwise.
     *
     * @see     leon.app.behavior.LySessionBehavior#validateAction
     */
    public boolean validateAction(LyAction action, LyController controller)
    {
        boolean result;

        result = _behavior.validateAction(this, action, controller);

        if ((result) && (_parent != null))
            result = _parent.validateAction(action, controller);

        return result;
    }

    /**
     * Indicates if an action is enabled for a given controller and a given objects list. This
     * method calls the same method on the client interface.
     *
     * @param   action      The action that must be tested.
     * @param   controller  The controller used to control the action validity.
     * @param   objects     The list of objects on which the action should be applied.
     * @param   showError   Indicates if an error message should be displayed in case of error.
     *
     * @return  true if the action is enabled, false otherwise.
     *
     * @see     leon.app.behavior.LySessionBehavior#enableAction
     */
    public boolean enableAction(LyAction action, LyController controller, LyObjectList objects, boolean showError)
    {
        boolean result;

        if (action.isAbstract())
        {
            LyInfoList<LyClassInfo> classInfos = ((objects == null) ? null : objects.getClasses());
            int                     nc = ((classInfos == null) ? 0 : classInfos.getSize());

            if (nc > 0)
            {
                LyClassInfo classInfo;
                int         i;

                for (i = 0; i < nc; i++)
                {
                    classInfo = classInfos.getInfo(i);
                    if (classInfo.getRealAction(action) != null)
                        break;
                }

                if (i == nc)
                    return false;
            }
            else
            {
                LyClassInfo classInfo = action.getClassInfo();

                if ((classInfo == null)
                        && (controller instanceof LyAreaController))
                    classInfo = ((LyAreaController)controller).getClassInfo();

                if ((classInfo == null)
                        || (classInfo.getRealAction(action) == null))
                    return false;
            }
        }

        result = _behavior.enableAction(this, action, controller, objects, showError);
        if (result && (_parent != null))
            result = _parent.enableAction(action, controller, objects, showError);

        return result;
    } // end method enableAction

    /**
     * Indicates if a transverse action is enabled for a given controller and a given objects list.
     * This method calls the same method on the client interface.
     *
     * @param   action      The action that must be tested.
     * @param   classInfo   The classInfo for which the action is checked.
     * @param   controller  The controller used to control the action validity.
     * @param   objects     The list of objects on which the action should be applied.
     * @param   showError   Indicates if an error message should be displayed in case of error.
     *
     * @return  true if the action is enabled, false otherwise.
     *
     * @see     leon.app.behavior.LySessionBehavior#enableTransverseAction
     */
    public boolean enableTransverseAction(LyAction action, LyClassInfo classInfo, LyController controller,
        LyObjectList objects, boolean showError)
    {
        if (classInfo == null)
            return enableAction(action, controller, objects, showError);

        if ((action.isAbstract())
                && (classInfo.getRealAction(action) == null))
            return false;

        boolean result = _behavior.enableTransverseAction(this, action, classInfo, controller, objects, showError);

        if (result && (_parent != null))
            result = _parent.enableTransverseAction(action, classInfo, controller, objects, showError);

        return result;
    }

    /**
     * Invokes an action that has no controller (an extra action).
     *
     * @param   parent     The source controller.
     * @param   action     The action to execute.
     * @param   classInfo  The class on which the action applies.
     * @param   objects    The list of objects on which the action applies.
     * @param   show       Indicates if the action must be shown or not.
     *
     * @return  true if the action was handled by the session, false otherwise.
     *
     * @see     leon.app.behavior.LySessionBehavior#invokeExtraAction
     */
    public boolean invokeExtraAction(LyController parent, LyAction action, LyClassInfo classInfo, LyObjectList objects,
        boolean show)
    {
        if ((parent != null)
                && parent.isFreed())
        {
            LyEnvironment environment = getEnvironment();

            environment.logWarning("BF AppEngine : call of invokeExtraAction for action "
                + action + " with a freed parent.");
        }

        // Try first to invoke behavior
        if (_behavior.invokeExtraAction(this, parent, action, classInfo, objects, show))
            return true;

        if (_parent != null)
        {
            LySessionBehaviorInterface parentBehavior = (LySessionBehaviorInterface)_parent.getSessionBehavior();

            if (parentBehavior.invokeExtraAction(_parent, parent, action, classInfo, objects, show))
                return true;
        }

        // Generic processing for standard actions
        if (processExtraAction(parent, action, classInfo, objects))
            return true;

        // Action not implemented
        if (show)
        {
            LyEnvironment environment = getEnvironment();
            String        message = environment.getMessage("LY_ER_ACTION", action.getName());

            environment.raiseError(message);
        }

        return false;
    } // end method invokeExtraAction

    /**
     * Processes an action that has no controller (an extra action) : processing of standard
     * actions.
     *
     * @param   parent     The source controller.
     * @param   action     The action to execute.
     * @param   classInfo  The class on which the action applies.
     * @param   objects    The list of objects on which the action applies.
     *
     * @return  true if the action was handled by the session, false otherwise.
     *
     * @see     leon.app.behavior.LySessionBehavior#invokeExtraAction
     */
    public boolean processExtraAction(LyController parent, LyAction action, LyClassInfo classInfo, LyObjectList objects)
    {
        String id = action.getRoot().getId();

        if ((id == LyAction.ID_EXIT)
                && (parent != null))
        {
            LyViewController parentViewController = parent.getViewController();

            parentViewController.exit();
            return true;
        }
        else if (id == LyAction.ID_EXEC)
        {
            String command = (String)action.getParameter("_command");

            _environment.executeCommand(command, false);
            return true;
        }
        else if ((id == LyAction.ID_ACTIVATE)
                || (id == LyAction.ID_MODIFY)
                || (id == LyAction.ID_SHOW)
                || (id == LyAction.ID_HIDE)
                || (id == LyAction.ID_COPY)
                || (id == LyAction.ID_PASTE))
        {
            Class<?> behaviorClass = action.getBehaviorClass();

            if (behaviorClass == null)
                throw new IllegalStateException("BF AppEngine : action "
                    + action + " requieres a behavior ");

            if (behaviorClass == LyClassBehavior.class)
            {
                // Try to find if we have a specific class behavior
                if ((classInfo != null)
                        && (classInfo.getBehaviorClass() != null))
                    behaviorClass = classInfo.getBehaviorClass();
            }

            LyClassBehavior behavior;

            try
            {
                behavior = (LyClassBehavior)behaviorClass.newInstance();
            }
            catch (Exception e)
            {
                _environment.writeException(e);
                return true;
            }

            if (id == LyAction.ID_ACTIVATE)
                behavior.changeState(parent, action, objects, null);
            else if (id == LyAction.ID_MODIFY)
                behavior.modify(parent, action, objects, null);
            else if (id == LyAction.ID_SHOW)
                behavior.show(parent, action, objects);
            else if (id == LyAction.ID_HIDE)
                behavior.hide(parent, action, objects);
            else if (id == LyAction.ID_COPY)
                behavior.copy(parent, action, objects);
            else if (id == LyAction.ID_PASTE)
            {
                String objectString = (String)getValue("$COPIED_OBJECT");

                behavior.paste(parent, action, objectString);
            }

            if (parent != null)
            	parent.showView();

            return true;
        }
        else if ((id == LyAction.ID_UNDO)
                || (id == LyAction.ID_REDO))
        {
            LyContextManager cm = getContextManager();

            if (id == LyAction.ID_UNDO)
            {
                if (!cm.undo(this))
                    _environment.raiseError("LY_ER_UNDO_FAILED");
            }
            else if (id == LyAction.ID_REDO)
            {
                if (!cm.redo(this))
                    _environment.raiseError("LY_ER_REDO_FAILED");
            }

            return true;
        }

        return false;
    } // end method processExtraAction

    /**
     * Invokes an action.
     *
     * @param   parent     The source controller.
     * @param   action     The action to execute.
     * @param   classInfo  The class on which the action applies.
     * @param   objects    The list of objects on which the action applies.
     * @param   show       Indicates if the action must be shown or not.
     *
     * @return  true if the action was handled by the session, false otherwise.
     *
     * @see     leon.app.behavior.LySessionBehavior#invokeAction
     */
    public LyController invokeAction(LyController parent, LyAction action, LyClassInfo classInfo, LyObjectList objects,
        boolean show)
    {
        if (action == null)
            return null;

        if ((parent != null)
                && parent.isFreed())
        {
            LyEnvironment environment = getEnvironment();

            environment.logWarning("BF AppEngine : call of invokeAction for action "
                + action + " with a freed parent.");
        }

        LyViewManager viewManager = getViewManager();

        if ((_invokedActions == 0) && (viewManager != null))
            viewManager.setWaiting(true);

        Date startDate = new Date();

        if (!action.isLoaded())
            action.load();

        if (classInfo == null)
            classInfo = action.getClassInfo();

        if ((classInfo != null) && !classInfo.isLoaded())
            classInfo.load();

        if (action.isAbstract())
        {
            LyAction realAction = null;

            if (classInfo != null)
                realAction = classInfo.getRealAction(action);

            if (realAction == null)
            {
                _environment.logWarning("BF AppEngine : can't find action corresponding to abstract action "
                    + action + " on class info" + classInfo);
                if ((_invokedActions == 0) && (viewManager != null))
                    viewManager.setWaiting(false);

                return null;
            }

            // Replace abstract action by real one
            action = realAction;
        }

        // Check first if action can be invoked
        if (!action.canBeInvoked((classInfo == null)))
            throw new IllegalStateException("BF AppEngine : action " + action
                + " can't be invoked directly because it has required parameters that are not valued");

        LyController  controller = null;
        LyDataSession oldCurrentSession = _application.getCurrentSession();
        LyEnvironment oldEnvironment = LyEnvironment.setCurrentEnvironment(getEnvironment());
        boolean		  actionInvoked = false;

        try
        {
            _metricsIdleMode++;

            // Set current session
            _application.setCurrentSession(this);

            // Increment counter of invoked actions.
            _invokedActions++;

            // Check if action is enabled
            if ((!(parent instanceof LyMultiAreaController) // XXX Find a better solution
                        || (parent instanceof LyDashboardController))
                    && !enableAction(action, parent, objects, true))
                return null;

            // Get controller class
            Class<?> controllerClass = action.getControllerClass();

            actionInvoked = true;

            if (controllerClass == null)
            {
				// Action is not handled by a controller
                invokeExtraAction(parent, action, classInfo, objects, show);
				return null;
            }

            // Try first to invoke behavior
            controller = _behavior.invokeAction(this, parent, action, classInfo, objects, show);
            if (controller != null)
                return controller;

            if (LyProcessActionInterface.class.isAssignableFrom(controllerClass))
            {
                LyProcessActionInterface processor = (LyProcessActionInterface)controllerClass.newInstance();

                controller = processor.processAction(this, parent, action, classInfo, objects);
                // If we process a webservice, we store the return value.
                if (processor instanceof LyProcessWSActionInterface)
                    _application.setCurrentInstance(action.getId(),
                        ((LyProcessWSActionInterface)processor).getWSReturnValue());
            }
            else
            {
                if (show)
                    controller = _behavior.findExistingController(this, parent, action, classInfo, objects, false);

                if (controller == null)
                    controller = processAction(parent, action, classInfo, objects);
            }

            if (show && (controller != null && !controller.isFreed()))
                showView(controller);
        }
        catch (Throwable t)
        {
            // Maybe we must do something.
            handleException(t);

            try
            {
                // If anything occurs, remove controller and free it if possible.
                if ((controller != null) && !controller.isFreed())
                    controller.free();
                else if (controller != null)
                    removeController(controller);

                // Unfreeze application
                if ((_invokedActions == 0) && (viewManager != null))
                    viewManager.setWaiting(false);
            }
            catch (Throwable t2)
            {
                handleException(t2);
            }

            return null;
        }
        finally
        {
        	_metricsIdleMode--;

            if ((oldCurrentSession != null)
                    && !oldCurrentSession.isClosed())
                // Unset current session
                _application.setCurrentSession(oldCurrentSession);

            LyEnvironment.setCurrentEnvironment(oldEnvironment);
            // Decrement counter of invoked actions.
            _invokedActions--;

            if ((_invokedActions == 0) && (_dialogsToFree))
            {
                freeDialogs();
                if (controller instanceof LyDialogController)
                {
                    if (!controller.isFreed())
                        controller.free();

                    controller = null;
                }
            }

            if (_environment.isDebugLevel(LyEnvironment.DEBUG_PERF))
            {
                Date endDate = new Date();

                _environment.logOut("BF AppEngine : action " + action + " handled in "
                    + (endDate.getTime() - startDate.getTime()) + " ms.");
            }

            Date		endDate = new Date();

            if ((_metricsIdleMode == 0) && actionInvoked && LyMetricsTools.areMetricsActivated(this, LyMetricsTools.MetricType.ACTION))
            {
				LyObject	metric = LyMetricsTools.createMetric(this, LyMetricsTools.MetricType.ACTION, new Date(),
						(endDate.getTime() - startDate.getTime()), (controller == null ? endDate : null),
						action.getId(), (classInfo == null ? null : classInfo.getId()), null);

				if ((metric != null) && (controller != null))
					addValue(METRIC_OBJECT_ID_KEY + controller.getId(), metric.getObjectId());
            }

            if ((_invokedActions == 0) && (viewManager != null))
                viewManager.setWaiting(false);
        }

        return controller;
    } // end method invokeAction

    /**
     * Processes the given action : creates the controller that handles the action.
     *
     * @param   parent     The source controller.
     * @param   action     The action to execute.
     * @param   classInfo  The class on which the action applies.
     * @param   objects    The list of objects on which the action applies.
     *
     * @return  the controller associated to this action
     *
     * @see     leon.app.behavior.LySessionBehavior#invokeAction
     */
    public LyController processAction(LyController parent, LyAction action, LyClassInfo classInfo, LyObjectList objects)
    {
        // Check if controller class is valid
        Class<?> controllerClass = action.getControllerClass();
        LyAction template = (LyAction)action.getTemplate();

        if (template != null)
        {
            Class<?> templateControllerClass = template.getControllerClass();

            if ((templateControllerClass != null)
                    && !templateControllerClass.isAssignableFrom(controllerClass))
                throw new IllegalStateException("BF AppEngine : invalid controller "
                    + controllerClass + " for action " + action
                    + ", controller class must extend " + templateControllerClass);
        }

        // Build parameters to find out the right constructor for this action call
        int np;

        if ((objects != null) && (classInfo != null))
            np = 5;
        else if ((objects != null) || (classInfo != null))
            np = 4;
        else
            np = 3;

        Class<?>[] types = new Class[np];
        Object[]   parameters = new Object[np];

        types[0] = LySession.class;
        parameters[0] = this;

        types[1] = LyController.class;
        parameters[1] = parent;

        types[2] = LyAction.class;
        parameters[2] = action;

        if (np == 5)
        {
            types[3] = LyClassInfo.class;
            parameters[3] = classInfo;

            // Context
            types[4] = LyObjectList.class;
            parameters[4] = objects;
        }
        else if (np == 4)
        {
            if (objects != null)
            {
                if ((action.getMax() == 1) && (objects.getSize() > 0))
                {
                    types[3] = LyObject.class;
                    parameters[3] = objects.getObject(0);
                }
                else
                {
                    types[3] = LyObjectList.class;
                    parameters[3] = objects;
                }
            }
            else if (classInfo != null)
            {
                types[3] = LyClassInfo.class;
                parameters[3] = classInfo;
            }
        }

        // Build controller
        try
        {
            Constructor<?> constructor = null;

            if (np == 5)
                constructor = controllerClass.getConstructor(types);
            else
            {
                try
                {
                    // Search first for a constructor with given types
                    constructor = controllerClass.getConstructor(types);
                }
                catch (NoSuchMethodException e)
                {
                    // Second chance, look for the constructor with all parameters
                    Class<?>[] types2 = new Class[5];

                    types2[0] = LySession.class;
                    types2[1] = LyController.class;
                    types2[2] = LyAction.class;
                    types2[3] = LyClassInfo.class;
                    types2[4] = LyObjectList.class;

                    constructor = controllerClass.getConstructor(types2);

                    // Constructor found, replace parameters
                    parameters = new Object[5];
                    parameters[0] = this;
                    parameters[1] = parent;
                    parameters[2] = action;
                    parameters[3] = classInfo;
                    parameters[4] = objects;
                }
            }

            LyController controller = (LyController)constructor.newInstance(parameters);

            if (controller.isFreed()) // The controller got suicide !
                return null;

            return controller;
        }
        catch (NoSuchMethodException e)
        {
            // Unknown controller
            String message = "BF AppEngine : constructor with parameters ("
                + types[0] + ", " + types[1] + ", " + types[2];

            if (np >= 4)
                message += ", " + types[3];

            if (np > 4)
                message += ", " + types[4];

            message += ") not found on class " + controllerClass;

            _environment.logError(message);
            _environment.writeException(e);
            return null;
        }
        catch (InvocationTargetException e)
        {
            _environment.writeException(e.getTargetException());
            return null;
        }
        catch (Exception e)
        {
            _environment.writeException(e);
            return null;
        }
    } // end method processAction

    /**
     * Show the result of the about action.
     *
     * @param  parent  the source controller
     */
    public void processAboutAction(LyController parent)
    {
        _behavior.showAbout(this, parent);
    }

    /**
     * Show the result of the log action.
     *
     * @param  parent  the source controller
     */
    public void processLogAction(LyController parent)
    {
        _behavior.showLog(this, parent);
    }

    /**
     * Show the result of the version action.
     *
     * @param  parent  the source controller
     */
    public void processVersionAction(LyController parent)
    {
        _behavior.showVersion(this, parent);
    }

    /**
     * Show the result of the help action.
     *
     * @param  parent  the source controller
     */
    public void processHelpAction(LyController parent)
    {
        _behavior.showHelp(this, parent);
    }

    /**
     * Show view corresponding to the given controller.
     *
     * @param  controller  the controller to show
     */
    public void showView(LyController controller)
    {
        if (controller == null)
            return;

        LyViewManager viewManager = getViewManager();

        if (viewManager == null)
            return;

        // Start waiting
        if (_invokedActions == 0)
            viewManager.setWaiting(true);

        // Show view
        controller.showView();

        if (viewManager instanceof LyAdvancedViewManagerInterface)
            ((LyAdvancedViewManagerInterface)viewManager).completeShowView(controller, _pluginId);

        // DEBUG_XML MODE : dump view
        LyEnvironment environment = getEnvironment();

        if (environment.isDebugLevel(LyEnvironment.DEBUG_XML)
                && !controller.isFreed())
        {
            LyView view = controller.getView();

            if (view != null)
            {
                if (environment.isDebugLevel(LyEnvironment.DEBUG_ALL))
                    view.writeXml(null, true, true);
                else if (controller instanceof LyAreaController)
                {
                    LyAreaController areaController = (LyAreaController)controller;

                    writeXml(areaController.getAction(), areaController.getClassInfo(), null, false);
                }
                else
                    view.writeXml(null, false, false);
            }
        }

        // End waiting
        if (_invokedActions == 0)
            viewManager.setWaiting(false);
    } // end method showView

    /**
     * Write XML view file corresponding to a given action.
     *
     * @param  action     The action to execute.
     * @param  classInfo  The class on which the action applies.
     * @param  file       Path of the output file.
     * @param  fullView   Dump of the full view or only of its content.
     *
     * @see    leon.app.behavior.LySessionBehavior#invokeAction
     */
    public void writeXml(LyAction action, LyClassInfo classInfo, String file, boolean fullView)
    {
        LyController controller = invokeAction(null, action, classInfo, null, false);

        if (controller == null)
            return;

        LyViewController viewController = controller.getViewController();

        if (viewController != null)
        {
            LyView           view = viewController.getView();

            view.writeXml(file, false, fullView);
            viewController.free();
        }
    }

    /**
     * Returns the number of printing threads registered on this session. Actually threads are
     * registered on controllers and this methods called all these controllers to retrieve printing
     * threads.
     *
     * @return  The number of printing threads registered on this session.
     */
    public int getPrintThreadCount()
    {
        int result = 0;

        if (_controllersList != null)
        {
            LyController controller;
            int          i, n = _controllersList.size();

            for (i = 0; i < n; i++)
            {
                controller = getController(i);

                if ((controller instanceof LyViewController)
                        && (((LyViewController)controller).hasPrintThread()))
                    result++;
            }
        }

        return result;
    }

    /**
     * Interrupts printing threads registered on this session. Actually threads are registered on
     * controllers and this methods called all these controllers to interrupt printing threads.
     */
    public void interruptPrintThreads()
    {
        if (_controllersList != null)
        {
            LyController controller;
            int          i, n = _controllersList.size();

            for (i = 0; i < n; i++)
            {
                controller = getController(i);

                if ((controller instanceof LyViewController)
                        && (((LyViewController)controller).hasPrintThread()))
                    ((LyViewController)controller).interruptPrintThread();
            }
        }
    }

    /**
     * Called by the timer when session is in time out.
     *
     * @param  timer  the time that call this method
     */
    public void endTimer(LyTimer timer)
    {
        sessionTimeout();
    }

    /**
     * Called to reset the timer (When a user action occured).
     */
    public synchronized void resetTimeout()
    {
        int timer = _behavior.getSessionTimeout(this);

        // Reset timer
        if (_timeoutTimer != null)
        {
            if (timer > 0)
            {
                timer = timer * 1000;
                _timeoutTimer.setDuration(timer);
                _timeoutTimer.restart();
            }
            else
            {
                _timeoutTimer.stop();
                _timeoutTimer = null;

            }
        }
        else if ((timer > 0)
                && (_state != State.CLOSING)
                && (_state != State.CLOSED))
        {
            // Do not start timeout timer if session is closed or being closed
            _timeoutTimer = new LyTimer(this, timer * 1000, "BF AppEngine : session timeout timer");
            _timeoutTimer.start();
        }
    } // end method resetTimeout

    /**
     * Called when the session is in time out.
     */
    public void sessionTimeout()
    {
        sessionTimeout(true);
    }

    /**
     * Called when the session is in time out.
     *
     * @param  asynchronous  indicates whether the exit is processed asynchronously
     */
    public void sessionTimeout(boolean asynchronous)
    {
        _behavior.sessionTimeout(this);
        confirmExit(asynchronous);
    }

    /**
     * Exit application.
     */
    public void exit()
    {
        LyViewManager viewManager = getViewManager();

        if (!(viewManager instanceof LyAdvancedViewManagerInterface)
                || ((LyAdvancedViewManagerInterface)viewManager).needConfirmOnClose(this))
            _behavior.exit(this);
        else
            confirmExit();
    }

    /**
     * Confirm the exit. Interrupt all print thread and close the session
     */
    public void confirmExit()
    {
        if (getViewManager().hasProperty("WEB_MANAGER"))
            confirmExit(false);
        else
            confirmExit(true);
    }

    /**
     * Confirm the exit. Interrupt all print thread and close the session
     *
     * @param  asynchronous  indicates whether the exit is processed asynchronously
     */
    public void confirmExit(boolean asynchronous)
    {
        _state = State.CLOSING;

        // Exit Asynchronously
        _LySessionExitThread t = new _LySessionExitThread(this);

        if (asynchronous)
            t.start();
        else
            t.run();
    }

    /**
     * Confirm the closing.
     */
    public void confirmClosing()
    {
        _state = State.CLOSING;
    }

    /**
     * Closes the session. Removes the session from its application, sets all the controllers, the
     * object sources and the view manager free.
     */
    public void close()
    {
        if (_state == State.CLOSED)
            return;

        LyDataSession oldCurrentSession = _application.getCurrentSession();

        try
        {
            _application.setCurrentSession(this);

            // Stop timeout timer
            if (_timeoutTimer != null)
            {
                _timeoutTimer.stop();
                _timeoutTimer = null;
            }

            // Call behavior
            _behavior.close(this);

            // Remove remaining controllers
            if (_controllersList != null)
            {
                LyController controller;

                while (_controllersList.size() > 0)
                {
                    controller = getController(0);
                    if (controller.isFreed())
                        getEnvironment().logWarning("BF AppEngine : controller " + controller
                            + " was already freed");
                    else
                        controller.free();

                    // If an error occurred and controllers list still contains controller
                    // force removal of controller to avoid loop
                    if (_controllersList.contains(controller))
                        _controllersList.removeElement(controller);
                }

                _recentControllerStack = null;
                _controllersList = null;
                _controllers = null;
            }

			if (_state == State.CLOSING)
			{
				// Close sub-sessions
				if (_subSessions != null)
				{
					Vector<LyDataSession> subSession = new Vector<LyDataSession>(_subSessions);
					int		  i, n = subSession.size();
					LySession session;

					for (i = n - 1; i >= 0; i--)
					{
						session = (LySession)subSession.elementAt(i);
						session.close();
					}
				}

				// Close slave-sessions
				if (_slaveSessions != null)
				{
					Vector<LySession> slaveSession = new Vector<LySession>( _slaveSessions);
					int		  i, n = slaveSession.size();
					LySession session;

					for (i = n - 1; i >= 0; i--)
					{
						session = slaveSession.elementAt(i);

						if (session != this)
							session.close();
					}
				}

                // Remove context manager
                if (_contextManager != null)
                {
                    _contextManager.free();
                    _contextManager = null;
                }

                // Remove all specific object sources.
                if (_objectSources != null)
                {
                    LyObjectSource              objectSource;
                    Enumeration<LyObjectSource> e = _objectSources.elements();

                    while (e.hasMoreElements())
                    {
                        objectSource = e.nextElement();
                        objectSource.free();
                    }

                    _objectSources = null;
                }

        		// Update session metric end date
        		LyObjectId	metricId = (LyObjectId)getValue(LySession.METRIC_OBJECT_ID_KEY);

        		if (metricId != null)
        		{
        			LyMetricsTools.setMetricEndDate(this, LyMetricsTools.MetricType.SESSION, metricId.getLogicalObject(), new Date());
        			LyMetricsTools.setMetricUserId(this, LyMetricsTools.MetricType.SESSION, metricId.getLogicalObject());
        		}

                // Free all resources.
                if (_environment != null)
                {
                    // Free all specific environment resource bundles.
                    LyResourceBundle.removeBundles(_environment);

                    _environment.free();
                }
            }
        }
        catch (Exception e)
        {
            getEnvironment().writeException(e);
        }
        finally
        {
            if (_state == State.CLOSING)
            {
                // Remove this session on its parent
                if (_parent != null)
                    _parent.removeSubSession(this);

                if ((oldCurrentSession != null)
                        && !oldCurrentSession.isClosed())
                    // Unset current session
                    _application.setCurrentSession(oldCurrentSession);

                // Unregister this session of application
                if (_registeredSession)
                {
                    _application.removeSession(this);
                    _registeredSession = false;
                }

                setMasterSession(null);

                _state = State.CLOSED;
                _values = null;
                _valid = false;
            }
        }
    } // end method close

    /**
     * Closes all views of the session except root.
     */
    public void closeViews()
    {
        closeViews(null);
    }

    /**
     * Closes all views of the session except root.
     *
     * @param  stackName  the name of the controller stack or null if default stack is considered
     *                    The name of a stack is the name of the window (Servlet mode)
     */
    public void closeViews(String stackName)
    {
        LyViewManager viewManager = getViewManager();

        if (viewManager == null)
            return;

        // Try to close every non root view controller
        if ((_controllersStackList != null)
                && !_controllersStackList.isEmpty())
        {
            Iterator<String>        keys = _controllersStackList.keySet().iterator();
            String                  key;
            Stack<LyViewController> controllersStack, stack;
            int                     i;
            LyViewController        controller;

            while (keys.hasNext())
            {
                key = keys.next();
                if ((stackName != null) && (!stackName.equals(key)))
                    continue;

                controllersStack = _controllersStackList.get(key);
                stack = new Stack<LyViewController>();
                stack.addAll(controllersStack);

                for (i = stack.size() - 1; i >= 0; i--)
                {
                    controller = stack.elementAt(i);
                    if (controller.getParent() != null)
                    {
                        if ((stackName != null)
                                && (!stackName.equals(POPUP_VIEW))
                                && (stackName.startsWith(POPUP_VIEW)))
                        {
                            if (!controller.hasSonShowing()
                                    && !controller.isInShowProcess())
                                controller.close();

                        }
                        else
                            controller.close();
                    }
                }
            }
        }

        if (POPUP_VIEW.equals(stackName))
            clearOldViews();

        updateViewManagerControllerStacks();
    } // end method closeViews

    /* Controllers Methods */

    /**
     * Method to check id of the given controller.
     *
     * @param   controller  The new controller.
     * @param   id          Desired id.
     *
     * @return  Id of the new controller once checked by this session.
     */
    public String checkControllerId(LyController controller, String id)
    {
        if (_behavior == null)
            return id;
        else
            return _behavior.checkControllerId(controller, id);
    }

    /**
     * Registers a controller on the session.
     *
     * @param      controller  The controller to register.
     *
     * @exception  java.lang.IllegalArgumentException  if the given controller is null.
     *
     * @see        #removeController
     */
    public void addController(LyController controller)
    {
        if (controller == null)
            throw new IllegalArgumentException("BF AppEngine : controller can't be null");

        if (_controllersList == null)
            _controllersList = new Vector<LyController>();
        else if (_controllersList.contains(controller))
        {
            // The controller is rebuilt
            if ((_viewCacheCount > 0) && (controller instanceof LyViewController))
                updateRecentControllerStack((LyViewController)controller);

            return;
        }

        if (_controllers == null)
            _controllers = new Hashtable<String, LyController>();

        _controllersList.addElement(controller);
        _controllers.put(controller.getId(), controller);

        // Notify behavior
        _behavior.controllerAdded(this, controller);
    }

    /**
     * Unregisters a controller from the session.
     *
     * @param      controller  The controller to unregister.
     *
     * @exception  java.lang.IllegalArgumentException  if the given controller is null.
     *
     * @see        #addController
     */
    public void removeController(LyController controller)
    {
        if (controller == null)
            throw new IllegalArgumentException("BF AppEngine : controller can't be null");

        if (_controllers != null)
        {
            LyController registeredController = _controllers.get(controller.getId());

            if (controller == registeredController)
                _controllers.remove(controller.getId());
        }

        if (_controllersList != null)
            _controllersList.removeElement(controller);

        if (_recentControllerStack != null)
            _recentControllerStack.removeElement(controller);

        // Remove controller from infoListeners list
        if (controller instanceof LyInfoListener)
            removeInfoListener((LyInfoListener)controller);

        // Notify behavior
        _behavior.controllerRemoved(this, controller);

		if (LyMetricsTools.areMetricsActivated(this, LyMetricsTools.MetricType.ACTION)
				&& (controller instanceof LyAreaController))
		{
			LyAreaController	areaController = (LyAreaController)controller;
			LyAction			action = areaController.getAction();

			if (action != null)
			{
				// Register action metric if needed
				LyObjectId	metricId = (LyObjectId)getValue(METRIC_OBJECT_ID_KEY + areaController.getId());

				if (metricId != null)
				{
					LyMetricsTools.setMetricEndDate(this, LyMetricsTools.MetricType.ACTION, metricId.getLogicalObject(), new Date());
					// Since user may not have been set when action was started (login action for example), set it again !
					LyMetricsTools.setMetricUserId(this, LyMetricsTools.MetricType.ACTION, metricId.getLogicalObject());
					_values.remove(METRIC_OBJECT_ID_KEY + areaController.getId());
				}
			}
		}
    }

    /**
     * Retrieves a controller by its identifier.
     *
     * @param      id  The id of the controller to get.
     *
     * @return     The controller which have the specified id.
     *
     * @exception  java.lang.IllegalArgumentException  if the given string is null.
     */
    public LyController getController(String id)
    {
        if (id == null)
            throw new IllegalArgumentException("BF AppEngine : id can't be null");

        if (_controllers == null)
            return null;

        return _controllers.get(id);
    }

    /**
     * Retrieves a controller by its index.
     *
     * @param   index  The index of the controller to get.
     *
     * @return  The controller which have the specified index in the list of this session.
     */
    public LyController getController(int index)
    {
        if (_controllersList == null)
            return null;

        if ((index < 0) || (index >= _controllersList.size()))
            return null;

        return _controllersList.elementAt(index);
    }

    /**
     * Gets number of controllers registered on this session.
     *
     * @return  Number of controllers of this session.
     */
    public int getControllerCount()
    {
        return ((_controllersList == null) ? 0 : _controllersList.size());
    }

    /**
     * Adds a view controller to the stack of shown controllers.
     *
     * @param  viewController  The view controller to add.
     */
    public void controllerShown(LyViewController viewController)
    {
        LyViewManager viewManager = getViewManager();

        if (viewManager == null)
            return;

        // Add or replace viewController on stack top
        Stack<LyViewController> stack = null;
        String                  stackName;
        LyView                  view = viewController.getView();

        if (view instanceof LyDialog)
            stackName = ((LyDialog)view).getTarget();
        else if (view instanceof LyFrame)
            stackName = ((LyFrame)view).getTarget();
        else
            stackName = viewManager.getCurrentWindowName();

        if (viewManager.hasProperty("ONLY_TARGET_POPUP"))
        {
            if ((stackName != null)
                    && !stackName.equals(POPUP_VIEW)
                    && !stackName.equals(DEFAULT)
                    && !stackName.endsWith("_specific_frame"))
                stackName = null;
        }

        if (stackName == null)
            stackName = DEFAULT;

        if (_controllersStackList == null)
            _controllersStackList = new Hashtable<String, Stack<LyViewController>>();
        else
        {
            stack = _controllersStackList.get(stackName);

            if ((!stackName.equals(POPUP_VIEW))
                    && (stack != null)
                    && (!stack.isEmpty())
                    && (stack.peek() != viewController))
            {
                // Another test (same as in LyServlet) : the pilot frame has a dashboard controller
                // which is not clearing the stack
                if (viewManager.hasProperty("MANAGE_DASHBOARD")
                        && ((viewController.getParent() != null)
                            || !(viewController.getAreaController() instanceof LyDashboardController)))
                {
                    // A new view is opened without popup : clear popup stack
                    // Do not close controller yet since they may have sons
                    // Move them to a dummy location (removed when next popup is opened in clearOldViews)
                    Stack<LyViewController> stackPopupView = _controllersStackList.get(POPUP_VIEW);

                    if ((stackPopupView != null) && (!stackPopupView.isEmpty()))
                    {
                        // Store old in its own controller
                        LyViewController formerViewController = stack.peek();

                        formerViewController.setControllerStack(stackPopupView);
                        // Get new from new controller
                        stackPopupView = viewController.getControllerStack();
                        if (stackPopupView == null)
                            _controllersStackList.remove(POPUP_VIEW);
                        else
                            _controllersStackList.put(POPUP_VIEW, stackPopupView);
                    }
                }
            }
            else if (stackName.equals(POPUP_VIEW))
                clearOldViews();
        }
        // a new controller stack
        if (stack == null)
        {
            stack = new Stack<LyViewController>();
            _controllersStackList.put(stackName, stack);
        }

        if (stack.contains(viewController))
            stack.removeElement(viewController);

        stack.push(viewController);

        // Clear the view cache in old controllers stack
        if (_viewCacheCount == 0)
        {
            String s = getEnvironment().getResource("LY_VIEW_CACHE_COUNT");

            if ((s == null) || (s.length() == 0))
                _viewCacheCount = -1; // Keep all
            else
            {
                try
                {
                    _viewCacheCount = Integer.parseInt(s);
                }
                catch (NumberFormatException e)
                {
                    _environment.logError(" Error in ressource LY_VIEW_CACHE_COUNT : " + s);
                    _environment.writeException(e);
                }
            }
        }
        // Clear cache if needed : FFT00170
        if (_viewCacheCount > 0)
        {
            updateRecentControllerStack(viewController);

            // Do not clear last viewcachecount controllers
            LyViewController oldViewController;
            int              i, n = getControllerCount();
            LyController     controller;
            int              j, ns;
            boolean          clear;

            for (i = 0; i < n; i++)
            {
                controller = _controllersList.elementAt(i);

                if (!(controller instanceof LyViewController))
                    continue;

                oldViewController = (LyViewController)controller;

                if (_recentControllerStack.contains(controller))
                    continue;

                // Do not clear cache for dashboard controllers
                if ((oldViewController.getParent() == null)
                        || (oldViewController.getAreaController() instanceof LyDashboardController))
                    continue;

                view = oldViewController.getView();

                // Do not clear cache for controllers displayed in the popup view
                if ((view instanceof LyDialog) && POPUP_VIEW.equals(((LyDialog)view).getTarget()))
                    continue;

                // Clear only viewControllers, and not current being shown
                // Keep also the last n=_viewCacheCount controllers.
                clear = true;
                ns = stack.size();

                for (j = Math.max(0, ns - _viewCacheCount); (j < ns) && (clear); j++)
                {
                    if (controller == stack.elementAt(j))
                        clear = false;
                }

                if (clear)
                    oldViewController.clearViewCache();
            }
        }

        // Notify controller
        _behavior.viewControllerShown(this, viewController);

        updateViewManagerControllerStacks();
    } // end method controllerShown

    /**
     * Utility method to clear old popup views from the stack (web mode).
     */
    protected void clearOldViews()
    {
        String           key, dateStr;
        long             date;
        Iterator<String> keys = _controllersStackList.keySet().iterator();

        while (keys.hasNext())
        {
            key = keys.next();
            if ((!key.equals(POPUP_VIEW)) && key.startsWith(POPUP_VIEW))
            {
                dateStr = key.substring(POPUP_VIEW.length());
                date = LyTools.longFromString(dateStr);
                if ((System.currentTimeMillis() - date) > 1000)
                    closeViews(key);
            }
        }
    }

    /**
     * Removes a view controller from the stack of shown controllers.
     *
     * @param  viewController  The view controller to remove.
     */
    public void controllerHidden(LyViewController viewController)
    {
        LyViewManager viewManager = getViewManager();

        if (viewManager == null)
            return;

        if (_controllersStackList != null)
        {
            Stack<LyViewController> stack;
            String                  stackName;
            LyView                  view = viewController.getView();

            if (view instanceof LyDialog)
                stackName = ((LyDialog)view).getTarget();
            else if (view instanceof LyFrame)
                stackName = ((LyFrame)view).getTarget();
            else
                stackName = viewManager.getCurrentWindowName();

            if (viewManager.hasProperty("ONLY_TARGET_POPUP"))
            {
                if ((stackName != null)
                        && !stackName.equals(POPUP_VIEW)
                        && !stackName.equals(DEFAULT)
                        && !stackName.endsWith("_specific_frame"))
                    stackName = null;
            }

            if (stackName == null)
                stackName = DEFAULT;

            stack = _controllersStackList.get(stackName);
            boolean done = false;

            if (stack != null)
                done = stack.remove(viewController);

            if (done)
            {
                if ((!stackName.equals(POPUP_VIEW))
                        && (stack != null)
                        && (!stack.isEmpty()))
                {
                    // A new view is closed outside the popup : restore the popup stack of the controller below
                    LyViewController        formerViewController = stack.peek();
                    Stack<LyViewController> stackPopupView = formerViewController.getControllerStack();

                    if ((stackPopupView != null) && (!stackPopupView.isEmpty()))
                    {
                        _controllersStackList.put(POPUP_VIEW, stackPopupView);

                        Iterator<LyViewController>  iter = stackPopupView.iterator();
                        ArrayList<LyViewController> toRemove = new ArrayList<LyViewController>();
                        LyViewController            controller;

                        while (iter.hasNext())
                        {
                            controller = iter.next();

                            if (controller.isFreed())
                                toRemove.add(controller);
                        }

                        // Remove freed controllers
                        iter = toRemove.iterator();

                        while (iter.hasNext())
                            stackPopupView.remove(iter.next());
                    }

                }
            }
            else
            {
                // Never leave the controller in the stack : in web mode, may be in a popup ?
                if (!_controllersStackList.isEmpty())
                {
                    Iterator<Stack<LyViewController>> keys = _controllersStackList.values().iterator();

                    while (keys.hasNext() && (!done))
                    {
                        stack = keys.next();
                        done = stack.remove(viewController);
                    }
                }
            }

            updateViewManagerControllerStacks();
        }

        LyAreaController controller = viewController.getAreaController();

        // Clear the outline if needed
        if (controller != null)
            notifyOutline(controller, true);
        // Notify controller
        _behavior.viewControllerHidden(this, viewController);
    } // end method controllerHidden

    /**
     * Returns the stack size (number of controller) for the given stack (or null for default).
     *
     * @param   stackName  the name of the controller stack or null if default stack is considered
     *
     * @return  the number of controllers or null if stack is empty.
     */
    public int getControllerStackSize(String stackName)
    {
        if (stackName == null)
            stackName = DEFAULT;

        Stack<LyViewController> stack = _controllersStackList.get(stackName);

        if (stack == null)
            return 0;
        else
            return stack.size();
    }

    /**
     * Returns the last controller for the given stack (or null for default).
     *
     * @param   stackName  the name of the controller stack or null if default stack is considered
     *
     * @return  the last controller of the stack or null if stack is empty.
     */
    public LyViewController getStackLastController(String stackName)
    {
        if (stackName == null)
            stackName = DEFAULT;

        Stack<LyViewController> stack = _controllersStackList.get(stackName);

        if ((stack == null) || stack.isEmpty())
            return null;
        else
            return stack.peek();
    }

    /**
     * Returns the view controller that was last shown.
     *
     * @return  The view controller that was last shown or null if no view controller had been shown
     *          yet.
     */
    public LyViewController getLastViewController()
    {
        return getLastViewController(0);
    }

    /**
     * Returns the view controller that was last shown.
     *
     * @param   steps  the number of steps used to retrieve desired view controller. For example,
     *                 getLastViewController(0) will return the view controller that was last shown,
     *                 getLastViewController(1) will return the view controller that was shown
     *                 before this one, etc.
     *
     * @return  The view controller that was last shown or null if no view controller had been shown
     *          yet.
     */
    public LyViewController getLastViewController(int steps)
    {
        LyViewManager viewManager = getViewManager();

        if (viewManager == null)
            return null;

        if (_controllersStackList == null)
            return null;

        Stack<LyViewController> stack;
        String                  stackName = viewManager.getCurrentWindowName();

        if (stackName == null)
            stackName = DEFAULT;

        stack = _controllersStackList.get(stackName);
        if ((stack == null) || stack.isEmpty())
            return null;

        int    n = stack.size();
        Object last = stack.get(n - 1);

        if ((stack == null) || stack.isEmpty())
        {
            updateViewManagerControllerStacks();
            return null;
        }

        LyViewController lastViewController = null;

        if (steps == 0)
            lastViewController = stack.peek();
        else if (steps < stack.size())
            lastViewController = stack.get(stack.size() - 1 - steps);

        if ((lastViewController != null) && lastViewController.isFreed())
        {
            _environment.logError("BF AppEngine : discarding last viewController found "
                + lastViewController + " because it was freed");

            stack.removeElement(last);
            updateViewManagerControllerStacks();

            return getLastViewController();
        }

        updateViewManagerControllerStacks();

        return lastViewController;
    } // end method getLastViewController

    /**
     * This method updates the view manager internal table of controller stacks. This method must be
     * called every time _controllersStackList is modified in order to let the view manager
     * consistent with current session.
     */
    private void updateViewManagerControllerStacks()
    {
        LyViewManager viewManager = getViewManager();

        if (viewManager == null)
            return;

        viewManager.updateControllersStackList(_controllersStackList);
    }

    /**
     * Add an outline controller for this session. This controller is usefull for selection
     * notification.
     *
     * @param  controller  An outline controller for this session.
     */
    public void addOutlineController(LyOutlineController controller)
    {
        if (_outlineControllers == null)
            _outlineControllers = new Hashtable<String, Vector<LyOutlineController>>();

        String perspectiveId = controller.getEnvironment().getResource("LY_PERSPECTIVE");

        if (perspectiveId == null)
            perspectiveId = "";

        Vector<LyOutlineController> outlines = _outlineControllers.get(perspectiveId);

        if (outlines == null)
        {
            outlines = new Vector<LyOutlineController>();
            _outlineControllers.put(perspectiveId, outlines);
        }

        outlines.add(controller);
    }

	/**
	 * Create a sticker window and add it to given parent window for the given sticker.
	 *
	 * @param parent             the parent controller whose view will get the sticker component.
	 * @param stickerController  the create sticker controller.
	 * @param sticker            the sticker definition.
	 * @deprecated
	 */
	public void createStickerWindow(LyController parent, LyCreateStickerController stickerController,
			LySticker sticker)
	{
	}

    /**
     * Remove an outline controller of this session for a given perspective.
     *
     * @param  controller  the outline controller to remove.
     */
    public void removeOutlineController(LyOutlineController controller)
    {
        String perspectiveId = controller.getEnvironment().getResource("LY_PERSPECTIVE");

        if (perspectiveId == null)
            perspectiveId = "";

        Vector<LyOutlineController> outlines = _outlineControllers.get(perspectiveId);

        outlines.remove(controller);
    }

    /**
     * Notifies the outline controller that the selection has changed in an other controller.
     *
     * @param  controller  The area controller where the selection has changed.
     */
    public void notifyOutline(LyAreaController controller)
    {
        notifyOutline(controller, false);
    }

    /**
     * Notifies the outline controller that the selection has changed in an other controller.
     *
     * @param  controller  The area controller where the selection has changed.
     */
    private void notifyOutline(LyAreaController controller, boolean clear)
    {
        String perspectiveId = controller.getEnvironment().getResource("LY_PERSPECTIVE");

        if (perspectiveId == null)
            perspectiveId = "";

        LySession session = this;

        while ((session != null)
                && ((session._outlineControllers == null)
                    || (session._outlineControllers.get(perspectiveId) == null)))
            session = (LySession)session.getParent();

        if (session != null)
        {
            Vector<LyOutlineController> outlines = session._outlineControllers.get(perspectiveId);
            LyOutlineController         outline;

            for (int i = 0; i < outlines.size(); i++)
            {
                outline = outlines.get(i);
                if ((outline.getComponent() == null) || outline.getComponent().isFreed())
                    return;

                if (!clear
                        || (outline.getLastOutlinedController() == controller))
                    outline.selectionChanged((clear ? null : controller));
            }
        }
    } // end method notifyOutline

    /**
     * Internal use : update the stack of recent controllers being used / rebuilt to avoid freeing
     * them all the time.
     *
     * @param  viewController  controller being shown or rebuilt
     */
    protected void updateRecentControllerStack(LyViewController viewController)
    {
        if (_recentControllerStack == null)
            _recentControllerStack = new Vector<LyViewController>(_viewCacheCount + 1);
        else
            _recentControllerStack.removeElement(viewController);

        _recentControllerStack.addElement(viewController);

        // Never keep more than _viewCacheCount controllers
        if (_recentControllerStack.size() > _viewCacheCount)
            _recentControllerStack.removeElementAt(0);
    }

    /**
     * internal class to close the session asynchronously.
     */
    class _LySessionExitThread extends Thread
    {
        protected LySession _session;

        _LySessionExitThread(LySession session)
        {
            _session = session;
            setName("BF AppEngine : exit session thread");
        }

        public void run()
        {
            if (_session != null)
            {
                // Asks GUI stop to view manager
                LyViewManager viewManager = getViewManager();

                if (viewManager != null)
                    viewManager.stopGui(_session);
            }
        }
    }

    /**
     * Starts GUI.
     */
    public void guiStart()
    {
        start();
    }

    /**
     * Stops GUI.
     */
    public void guiStop()
    {
        interruptPrintThreads();
        close();
    }

    /**
     * Check idle status of the table.
     *
     * @return  boolean indicating the idle mode
     */
    public final boolean isIdle()
    {
        return _idle;
    }

    /**
     * Freeze/unfreeze session.
     *
     * @param  idle  true to freeze the session, false to unfreeze
     */
    public void setIdle(boolean idle)
    {
        _idle = idle;

        if (!idle)
        {
            // Reset all controllers
            LyController controller;
            int          i, nc = _controllersList.size();

            for (i = 0; i < nc; i++)
            {
                controller = getController(i);
                if (controller instanceof LyAreaController)
                    ((LyAreaController)controller).reset(false);
            }
        }
    }

    /**
     * Check chat status.
     * @return boolean indicating the chat mode
     */
    public final boolean isChatting()
    {
    	return _chatting;
    }

    /**
     * Set chat status.
     * @param chat the new chat status.
     */
    public void setChatting(boolean chat)
    {
    	_chatting = chat;
    }

    /**
     * Reset session. Resets all view controllers of the session.
     */
    public void reset()
    {
        if (_controllersList == null)
            return;

        // Turn all controllers to idle mode
        setIdle(true);

        // Reset controllers
        LyController controller;
        int          i, nc = _controllersList.size();

        for (i = 0; i < nc; i++)
        {
            controller = getController(i);
            if (controller instanceof LyAreaController)
                ((LyAreaController)controller).reset();
        }

        // Reset listeners
        if (_resetListeners != null)
        {
            LyResetInterface listener;
            int              nr = _resetListeners.size();

            for (i = 0; i < nr; i++)
            {
                listener = _resetListeners.elementAt(i);
                listener.reset(this);
            }
        }

        // Awake controllers
        setIdle(false);
    } // end method reset

    /**
     * Internal method called when data have been unloaded. This method frees dialog controllers if
     * necessary.
     */
    void dataUnloaded()
    {
        _dialogsToFree = true;
        if (_invokedActions == 0)
            freeDialogs();

        super.dataUnloaded();
    }

    /**
     * Internal method used to free all dialog controllers of this session.
     */
    void freeDialogs()
    {
        if (!_dialogsToFree)
            return;

        if (_controllersList == null)
            return;

        // Free all dialog controllers.
        LyController               controller;
        int                        i, nc = _controllersList.size();
        Vector<LyDialogController> dialogs = new Vector<LyDialogController>();

        // Search for all dialog controllers
        for (i = 0; i < nc; i++)
        {
            controller = getController(i);
            if (controller instanceof LyDialogController)
            {
            	LyDialogController	dialogController = (LyDialogController)controller;
            	LyAreaController	areaController = dialogController.getAreaController();

            	if (!(areaController instanceof LyWaitController))
            		dialogs.addElement(dialogController);
            }
        }

        // Free controllers
        nc = dialogs.size();
        for (i = 0; i < nc; i++)
        {
            controller = dialogs.elementAt(i);
            if (!controller.isFreed())
            {
                try
                {
                    ((LyDialogController)controller).free();
                }
                catch (Exception e)
                {
                    LyEnvironment environment = getEnvironment();

                    environment.logWarning("Exception occured when freeing dialog");
                }
            }
        }

        _dialogsToFree = false;
    } // end method freeDialogs

    /**
     * Add reset listener.
     *
     * @param  listener  Listener to call on reset.
     */
    public void addResetListener(LyResetInterface listener)
    {
        if (_resetListeners == null)
            _resetListeners = new Vector<LyResetInterface>(1);

        _resetListeners.add(listener);
    }

    /**
     * Remove given reset listener.
     *
     * @param  listener  Listener removed from reset listeners list.
     */
    public void removeResetListener(LyResetInterface listener)
    {
        if (_resetListeners == null)
            return;

        _resetListeners.remove(listener);
    }

    /**
     * Set the plugin id for the session.
     *
     * @param  pluginId  The plugin id.
     */
    public void setPluginId(String pluginId)
    {
        _pluginId = pluginId;
    }

    /**
     * Get the plugin id for this session.
     *
     * @return  The plugin id.
     */
    public String getPluginId()
    {
        return _pluginId;
    }

    /**
     * Sets the language for this session. Default values are defined in LANGUAGE_xxx constants.
     *
     * @param  language  The new language.
     *
     * @see    leon.misc.LyEnvironment#setLanguage(String)
     */
    public void setLanguage(String language)
    {
		LyApplication	application = getApplication();

		application.setCurrentSession(this);

        // Change the language on the environment
        LyEnvironment environment = getEnvironment();

        if (environment != null)
        {
            environment.setLanguage(language);

// String skin = environment.getEnv("LY_SKIN");
// if (skin != null)
// environment.setSkin(skin, false);
        }

        LyInfoList<LyInclude>	includes = application.getInfoList(LyInclude.class);
        LyInclude				include;

        for (int i = 0; i < includes.getSize(); i++)
        {
        	include = includes.getInfo(i);
        	include.loadExtraLanguage(environment);
        }


        // Update all the controllers modification time in order to correctly update the
        // views in web viewer
        int          i, n = getControllerCount();
        LyController controller;

        for (i = 0; i < n; i++)
        {
            controller = getController(i);

            if (controller != null)
            {
                controller.setLastModified(System.currentTimeMillis());

				// Some views need to reset their components (for example message bar in tables)
				if (controller instanceof LySimpleTableController)
					((LySimpleTableController)controller).clearTranslatedStringMap();

				if (controller instanceof LyAreaController)
					((LyAreaController)controller).reset(false);

				LyEnvironment	subEnvironment = controller.getEnvironment();

				if ((subEnvironment != null) && (subEnvironment != environment))
					subEnvironment.setLanguage(language);
			}
		}
	} // end method setLanguage

    /**
     * Sets the skin for this session.
     *
     * @param  skin  The new skin.
     *
     * @see    leon.misc.LyEnvironment#setSkin(String)
     */
    public void setSkin(String skin)
    {
        LyEnvironment environment = getEnvironment();

        if (environment != null)
        {
            environment.clearResourceBundleCaches();
            environment.putEnv("LY_SKIN", skin);
            environment.setSkin(skin);
        }

        // Update all the controllers modification time in order to correctly update the
        // views in web viewer
        int          i, n = getControllerCount();
        LyController controller;

        for (i = 0; i < n; i++)
        {
            controller = getController(i);
            controller.setLastModified(System.currentTimeMillis());
            controller.getEnvironment().clearResourceBundleCaches();

            // Some views need to reset their components (for example message bar in tables)
            if (controller instanceof LyAreaController)
                ((LyAreaController)controller).reset(false);
        }
    }

    /**
     * Sets the master session for this session.<br />
     * Used in web mode only to indicate that current session is a "slave" session managed by
     * another one.
     *
     * @param  masterSession  the master session of this session
     */
    public void setMasterSession(LySession masterSession)
    {
        if (masterSession != null)
            masterSession.addSlaveSession(this);

        if (_masterSession != null)
            _masterSession.removeSlaveSession(this);

        _masterSession = masterSession;
    }

    /**
     * Return the master session for this session.<br />
     * Used in web mode only to indicate that current session is a "slave" session managed by
     * another one.
     *
     * @return  the master session for this session if set, null otherwise
     */
    public LySession getMasterSession()
    {
        return _masterSession;
    }

    /**
     * Add a slave session to current session.
     *
     * @param  session  the slave session that must be added
     */
    protected void addSlaveSession(LySession session)
    {
        if (_slaveSessions == null)
            _slaveSessions = new Vector<LySession>();

        if (!_slaveSessions.contains(session))
            _slaveSessions.add(session);
    }

    /**
     * Remove a slave session from current session.
     *
     * @param  session  the slave session that must be removed
     */
    protected void removeSlaveSession(LySession session)
    {
        if (_slaveSessions == null)
            return;

        if (_slaveSessions.contains(session))
            _slaveSessions.remove(session);

        if (_slaveSessions.isEmpty())
            _slaveSessions = null;
    }

    /**
     * Return the list of slave sessions. Used in web mode only to indicate the sessions managed by
     * current one.
     *
     * @return  a list of slave sessions
     */
    public Vector<LySession> getSlaveSessions()
    {
        return _slaveSessions;
    }

    /**
     * Get a root level context controller. This controller when created defines a top level session
     * context : GUI modification Like set controller should become subcontext of this context.
     * Usually this method returns null (no controller is a top level context controller). The
     * ImportTableController is defining a root level context to grab all user modifications in the
     * import context while the import table is active.
     *
     * @return  the root level context controller if any, or null if it could not be found.
     */
    public LyController getRootContextController()
    {
        if (_behavior != null)
            return _behavior.getRootContextController(this);

        return null;
    }

    /**
     * Dumps this session as a string.
     *
     * @return  the session dump as string
     */
    public String toString()
    {
        String str = super.toString();
        Object user = getValue("$USER");

        return ((user == null) ? str : (str + " (" + user + ") "));
    }

    /**
     * Reset undo redo stack.
     */
    public void resetUndoRedo()
    {
        if (_parent != null)
        {
            _parent.resetUndoRedo();
            return;
        }

        // Repost demand in GUI queue to avoid invalid thread accesses (on SWT)
        LyViewManager viewManager = getViewManager();

        if (viewManager != null)
            viewManager.execGui(new _LyResetUndoRedo(), true);
        else
            internalResetUndoRedo();
    }

    void internalResetUndoRedo()
    {
        // Call the session behavior
        getSessionBehavior().updateUndoRedo(this);

        if (_controllersList == null)
            return;

        // Reset undo/redo for all frame controllers.
        LyController controller;
        int          i, nc = _controllersList.size();

        // Search for all dialog controllers
        for (i = 0; i < nc; i++)
        {
            controller = getController(i);
            if (controller instanceof LyFrameController)
                ((LyFrameController)controller).updateUndoRedo();
        }

        Vector<LyDataSession> subSessions = getSubSessions();
        int                   n = ((subSessions == null) ? 0 : subSessions.size());
        LySession             subSession;

        for (i = 0; i < n; i++)
        {
            subSession = (LySession)subSessions.elementAt(i);
            subSession.internalResetUndoRedo();
        }
    } // end method internalResetUndoRedo

    class _LyResetUndoRedo
		implements Runnable, LyGetSessionInterface
    {
        _LyResetUndoRedo()
        {
        }

        public void run()
        {
            internalResetUndoRedo();
        }

		public final LyDataSession getSession()
		{
			return LySession.this;
		}
    }

    /**
     * Accessor to the view manager.
     *
     * @return  the view manager for this session
     */
    public LyViewManager getViewManager()
    {
        return (LyViewManager)_environment.getViewManager();
    }

    /**
     * Accessor to the view manager.
     * @param mobile : if the request comes from a mobile
     * @return  the view manager for this session
     */
    public LyViewManager getViewManager(boolean mobile)
    {
        return (LyViewManager)_environment.getViewManager(mobile);
    }

    /**
     * Method for loading user preferences.
     *
     * @return  a list of user preferences
     */
    public LyInfoList<?> loadUserPreferences()
    {
        if (_behavior == null)
            return null;

        LyObject user = getUser();
        String   userId = ((user == null) ? null : user.getObjectId().toString());

        return _behavior.loadUserPreferences(this, userId);
    }

	/**
	 * Method for loading user stickers.
	*/
	public LyInfoList<?> loadUserStickers()
	{
		if (_behavior == null)
			return null;

		if (getViewManager() == null)
		    return null;
		LyObject      user = getUser();
		String        userId = (user == null ? null : user.getObjectId().toString());
		LyAction	  stickerActionTemplate = getApplication().getAction("_createSticker");

        if (!validateAction(stickerActionTemplate, getController(0)))
        	return null;

        // Stickers available for web managers only
        if (!getViewManager().hasProperty("WEB_MANAGER"))
        	return null;

		LyInfoList<LyInfo>		result = _behavior.loadUserStickers(this, userId);
        LyInfoList<LySticker>	stickers = getApplication().getInfoList(LySticker.class);

		if ((stickers != null) && (stickers.getSize() > 0))
		{
			// If there are stickers, create controllers to manage them.
			Hashtable<String, Object>	parameters;
			LyAction					stickerAction;
			int							i, n = stickers.getSize();
			LySticker					sticker;
	        String                   	currentUserId = ((user == null) ? "" : user.getId());

			for (i = 0; i < n; i++)
			{
				sticker = stickers.getInfo(i);
	            userId = sticker.getUser();

	            // If current user does not own the preference do not show it
	            if ((userId != null)
	                    && !currentUserId.equals(userId))
	                continue;

	            parameters = new Hashtable<String, Object>();
				parameters.put("_stickerId", sticker.getId());
				stickerAction = new LyAction(stickerActionTemplate, parameters);
				invokeAction(getController(0), stickerAction, null, null, true);
			}
		}

		return result;
	}

    /**
     * Register a temporary file.
     *
     * @param   path  the path of the file
     * @param   name  the name of the file
     *
     * @return  a unique id for this file
     *
     * @see     #getTempFile(String)
     */
    public String storeTempFile(String path, String name)
    {
        if (_tempFileMap == null)
            _tempFileMap = new HashMap<String, String[]>();

        String   id = "" + LyTools.getNextIdentifier();
        String[] values = new String[2];

        values[0] = path;
        values[1] = name;
        _tempFileMap.put(id, values);

        getEnvironment().logTrace("Leonardi : LySession stores temp file "+name+" as id="+id+" in "+path);

        return id;
    }

    /**
     * Get a temporary file from its id.
     *
     * @param   id  the id of a file
     *
     * @return  an array of String: index 0 is the path, index 1 is the name. Can be null
     *
     * @see     #storeTempFile(String, String)
     */
    public String[] getTempFile(String id)
    {
        if (_tempFileMap == null)
            return null;

        return _tempFileMap.get(id);
    }

    /**
     * Gets login action.
     *
     * @return  The login action if found, null otherwise.
     */
    protected LyAction getLoginAction()
    {
        // Find login action
        LyAction rootAction = _application.getProject().getRootAction();
        LyAction root = ((rootAction == null) ? null : rootAction.getRoot());
        LyAction loginAction = _application.getAction("_login");

        if (root == loginAction)
            return rootAction;

        LyInfoList<LyAction> actions = _application.getActions();
        int                  i, n = actions.getSize();
        LyAction             action;

        for (i = 0; i < n; i++)
        {
            action = actions.getInfo(i);
            if ((action != loginAction)
                    && (action.getRoot() == loginAction))
                return action;
        }

        return null;
    }

    /**
     * Log user corresponding to given login and password.
     *
     * @param   login            User login.
     * @param   password         User password.
     * @param   loginParameters  Parameters string formated like param1=value1;param2=value2;....
     *
     * @return  True if log is successfull, false otherwise.
     */
    public boolean login(String login, String password, String loginParameters)
    {
        HashMap<String, String> values = null;

        if (loginParameters != null)
        {
            String[] parameters = LyString.stringToList(loginParameters, ';');
            int      i, p, n = parameters.length;

            values = new HashMap<String, String>(n);

            for (i = 0; i < n; i++)
            {
                p = parameters[i].indexOf('=');
                if (p > 0)
                    values.put(parameters[i].substring(0, p), parameters[i].substring(p + 1));
                else if (parameters[i].length() > 0)
                {
                    _environment.logError("BF AppEngine : invalid login parameter, " + parameters[i]
                        + ". Login parameters must be formated LY_LOGIN_PARAMETERS=param1=value1;param2=value2;...");
                    return false;
                }
            }
        }

        return login(login, password, values);
    }

    /**
     * Log user corresponding to given login and password.
     *
     * @param   login     User login.
     * @param   password  User password.
     * @param   values    Extra values if needed.
     *
     * @return  True if log is successfull, false otherwise.
     */
    public boolean login(String login, String password, HashMap<String, String> values)
    {
        LyEnvironment environment = getEnvironment();
        LyAction      loginAction = getLoginAction();

        if (loginAction == null)
        {
            // No login action
            environment.logError("BF AppEngine : can't check login, no login action found on project.");
            return false;
        }

        LyController controller = invokeAction(null, loginAction, null, null, false);

        if (!(controller instanceof LyLoginController))
        {
            environment.logError("BF AppEngine : can't check login, creation of login controller failed.");
            return false;
        }

        LySimpleParameterValues paramValues = new LySimpleParameterValues(null);

        if (values == null)
            values = new HashMap<String, String>();

        values.put(LyLoginController.PARAM_KEY_LOGIN, login);
        values.put(LyLoginController.PARAM_KEY_PASSWORD, password);
        paramValues.setValues(values);

        if (((LyLoginController)controller).checkLogin(paramValues))
            return true;
        else
            return false;
    }

    /**
     * Method that checks if login can be done with ini parameters.
     *
     * @param   rootAction  Root action of the project.
     *
     * @return  True if user can be logged successfully, false otherwise.
     */
    protected boolean tryAutoLogin(LyAction rootAction)
    {
        // Check if autologin can be performed
        String login = _environment.getEnv("LY_LOGIN");

        // make autologin in portlet mode.
        if (getViewManager().isPortletMode())
        {
            // force 'SSO' mode for portlet
            login = "user.name";
        }

        if (login == null)
            return false;

        LyAction loginAction = getLoginAction();

        if (loginAction != rootAction)
            return false;

        String password = _environment.getEnv("LY_PASSWORD");
        String loginParameters = _environment.getEnv("LY_LOGIN_PARAMETERS");

        return login(login, password, loginParameters);
    }
} // end class LySession
