/*
 * LyActionBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;


/**
 * This class implements the default behavior for actions that are controlled by an area controller.
 *
 * @author  Lyria S.A.
 * @version $Id: LyActionBehavior.java 28422 2010-11-03 12:03:38Z ehelmste $
 */
public abstract class LyActionBehavior extends LyBehavior
	implements LyActionBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyActionBehavior.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Construct a class behavior.
	 */
	public LyActionBehavior()
	{
	}

	/**
	 * Starts an action for the given controller. This method is the right place to allocate
	 * specific resources requiered for the given action. <i>The default implementation</i> does
	 * nothing and returns true.
	 *
	 * @param  areaController the current LyAreaController instance
	 * @param  action         the action that will be executed
	 * @return true
	 */
	public boolean startAction(LyAreaController areaController, LyAction action)
	{
		return true;
	}

	/**
	 * The controller resets its content. This method is the right place to add extra processing
	 * when the view is set.<br />
	 * <i>The default implementation</i> does nothing and returns true.
	 *
	 * @param  areaController the current LyAreaController instance
	 * @param  full           indicates if the controller performed a full reset of its content or a
	 *                        partial reset. The meaning of a full reset depends on controller
	 *                        implementation.
	 * @return true
	 */
	public boolean reset(LyAreaController areaController, boolean full)
	{
		return true;
	}

	/**
	 * Ends an action for the given controller. This method is the right place to free specific
	 * resources allocated for the given action.<br />
	 * <i>The default implementation</i> does nothing and returns true.
	 *
	 * @param  areaController the current LyAreaController instance
	 * @param  action         the action that was executed
	 * @return true
	 */
	public boolean endAction(LyAreaController areaController, LyAction action)
	{
		// Nothing to do to properly end a generic action
		return true;
	}

	/**
	 * Gets the image corresponding to the given object for the given controller. This image is
	 * mostly used in titlebar.<br />
	 * <i>The default implementation</i> does nothing and returns null.
	 *
	 * @param  areaController controller for which the image is required.
	 * @param  action         the action that was executed
	 * @return Logical name of the image for the given action.
     * null means action image, the string "_null" means no image.
	 */
	public String getImage(LyAreaController areaController, LyAction action)
	{
		return null;
	}

	/**
	 * Gets image tooltip corresponding to the given action.<br />
	 * Default implementation only calls {@link #getTitle(LyAreaController,  LyAction)
	 * getTitle(areaController, action)}.
	 *
	 * @param  areaController the current LyAreaController instance
	 * @param  action         the action that was executed
	 * @return image tooltip for the given action.
	 */
	public String getImageTooltip(LyAreaController areaController, LyAction action)
	{
		return getTitle(areaController, action);
	}

	/**
	 * Gets the menu image corresponding to the given action for the given controller.<br />
	 * <i>The default implementation</i> only calls {@link #getImage(LyAreaController,  LyAction)
	 * getImage(areaController, action)}.
	 *
	 * @param  areaController The controller for which the image is required.
	 * @param  action         The action for which the image is required
	 * @return Logical name of the image for the given action.
     * null means action image, the string "_null" means no image.
	 */
	public String getMenuItemImage(LyAreaController areaController, LyAction action)
	{
		return getImage(areaController, action);
	}

	/**
	 * Gets title corresponding to the given action.<br />
	 * <i>The default implementation</i> returns the action name.
	 *
	 * @param  areaController the current LyAreaController instance
	 * @param  action         the action that was executed
	 * @return the given action name
	 */
	public String getTitle(LyAreaController areaController, LyAction action)
	{
		return action.getName();
	}

	/**
	 * Gets title corresponding to the given action and context. Default implementation does nothing
	 * and return null.
	 *
	 * @param  areaController the current controller instance
	 * @param  action         the action that was executed
	 * @param  context        the action context
	 * @return null
	 */
	public String getContextTitle(LyAreaController areaController, LyAction action,
		LyObjectList context)
	{
		return null;
	}

	/**
	 * Gets help file for the given action. <i>The default implementation</i> does nothing and
	 * return null.
	 *
	 * @param  areaController the current controller instance
	 * @param  action         the action that was executed
	 * @return null
	 */
	public String getHelpFile(LyAreaController areaController, LyAction action)
	{
		return null;
	}

	/**
	 * Gets context to set on the controllers that have the given area controller as context.
	 *
	 * @param  areaController the current controller instance.
	 * @param  context  the context that is about to be set.
	 * @return The desired client context.
	 */
	public LyObjectList getClientContext(LyAreaController areaController, LyObjectList context)
	{
		return context;
	}
	
    /**
     * Retrieves a value from a registered key.
     *
     * @param   key  The key that must be used.
     * @return  The value stored with the key or null if the key is not registered.
     */
    public Object getKeyValue(String key)
    {
        return null;
    }
} // end class LyActionBehavior
