/*
 * LyApplicationBehavior.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.app.behavior;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.info.*;
import leon.peer.sql.LySqlDataProvider;


/**
 * This class implements the default behavior for applications.
 *
 * @author   Lyria S.A.
 * @version  $Id: LyApplicationBehavior.java 34913 2012-03-22 08:07:18Z jlsaouli $
 * @see      leon.app.behaviorinterface.LyApplicationBehaviorInterface
 * @see      leon.app.LyApplication
 */
public class LyApplicationBehavior extends LyDataApplicationBehavior
	implements LyApplicationBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION = "$Id: LyApplicationBehavior.java 34913 2012-03-22 08:07:18Z jlsaouli $";

	/**
	 * Constructor.
	 */
	public LyApplicationBehavior()
	{
	}

	/**
	 * Load application preferences.
	 *
	 * @param   application  Application on which preferences are loaded.
	 *
	 * @return  List of preferences of the application.
	 */
	public LyInfoList<LyInfo> loadPreferences(LyApplication application)
	{
		// Parse preferences files
	    LyInfoList<LyInfo> results;
    	LySqlDataProvider  prefDataProvider = application.getPrefDataProvider();
	    
		if (prefDataProvider != null)
			results = parseExtraFromDB(application, "preference", null);
		else
			results = application.parseExtraFiles("preference", "LY_PREFERENCE_DIR", null);
		
		return results;
	}

	/**
	 * Load extra infos.
	 *
	 * @param   application  Application on which extra infos (like charts or treemap
	 *                       configurations) are loaded.
	 *
	 * @return  List of extra infos for the application.
	 */
	public LyInfoList<LyInfo> loadExtraInfos(LyApplication application)
	{
		// Parse extra files : charts, preferences and treemap configurations
		LyInfoList<LyInfo> infos = new LyInfoList<LyInfo>();
    	LySqlDataProvider  prefDataProvider = application.getPrefDataProvider();

		if (prefDataProvider != null)
		{
			infos.addList(parseExtraFromDB(application, "chart", null));
			infos.addList(parseExtraFromDB(application, "treemap_configuration", null));
			infos.addList(parseExtraFromDB(application, "sticker", null));
		}
		else
		{
			infos.addList(application.parseExtraFiles("chart", "LY_CHART_DIR", null));
			infos.addList(application.parseExtraFiles("treemap_configuration", "LY_TREEMAP_DIR", null));
			infos.addList(application.parseExtraFiles("sticker", "LY_STICKER_DIR", null));
		}

		return infos;
	}

	/**
	 * Method called when a session is added to the application. If the data session is a LySession
	 * instance then calls
	 * {@link #sessionAdded(LyApplication, LySession) sessionAdded(application, (LySession) session)}
	 * else calls
	 * {@link LyDataApplicationBehavior#sessionAdded(LyApplication, LyDataSession) super.sessionAdded(application, session)}
	 *
	 * @param       application  Application in which a session is added.
	 * @param       session      Session that is added to the application.
	 *
	 * @deprecated  Use {@link #sessionAdded(LyApplication, LySession)}
	 */
	public void sessionAdded(LyApplication application, LyDataSession session)
	{
		if (session instanceof LySession)
			sessionAdded(application, (LySession)session);
		else
			super.sessionAdded(application, session);
	}

	/**
	 * Method called when a session is removed from the application. If the data session is a
	 * LySession instance then calls
	 * {@link #sessionRemoved(LyApplication, LySession) sessionRemoved(application, (LySession) session)}
	 * else calls
	 * {@link LyDataApplicationBehavior#sessionRemoved(LyApplication, LyDataSession) super.sessionRemoved(application, session)}
	 *
	 * @param       application  Application in which a session is removed.
	 * @param       session      Session that is removed from the application.
	 *
	 * @deprecated  Use sessionRemoved(LyApplication, LySession)
	 */
	public void sessionRemoved(LyApplication application, LyDataSession session)
	{
		if (session instanceof LySession)
			sessionRemoved(application, (LySession)session);
		else
			super.sessionRemoved(application, session);
	}

	/**
	 * Method called when a session is added to the application.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param  application  Application in which a session is added.
	 * @param  session      Session that is added to the application.
	 */
	public void sessionAdded(LyApplication application, LySession session)
	{
	}

	/**
	 * Method called when a session is removed from the application.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param  application  Application in which a session is removed.
	 * @param  session      Session that is removed from the application.
	 */
	public void sessionRemoved(LyApplication application, LySession session)
	{
	}

	/**
	 * Load data for the given classes list.
	 *
	 * @param  application  Application for which classes are loaded.
	 * @param  classes      List of classes to load
	 *
	 * @see    LyApplication#loadData()
	 * @see    leon.info.LyClassInfo
	 */
	public void loadData(LyApplication application, LyInfoList<LyClassInfo> classes)
	{
	} // end method loadData

} // end class LyApplicationBehavior
