/*
 * LyChartBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;


/**
 * This class implements the default behavior for actions that are controlled by a chart controller.
 *
 * @author  Lyria S.A.
 * @version $Id: LyChartBehavior.java 40675 2012-09-05 16:26:11Z jlsaouli $
 * @see     leon.control.LyChartController
 * @see     leon.control.LyGraphicalChartController
 * @see     leon.app.behaviorinterface.LyChartBehaviorInterface
 */
public class LyChartBehavior extends LyActionBehavior
	implements LyChartBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyChartBehavior.java 40675 2012-09-05 16:26:11Z jlsaouli $";

	/**
	 * Indicates whether a graphical chart view is resizable or not. Default implementation returns
	 * true if the controller is not a {@link leon.control.LyChartController  LyChartController},
	 * false otherwise.
	 *
	 * @param  controller the current chart controller or graphical chart controller
	 * @return true if the view is resizable, false otherwise
	 */
	public boolean isViewResizable(LyController controller)
	{
		if (controller instanceof LyChartController)
			return false;
		else
			return true;
	}

	/**
	 * Gets fields that may be used as X variables in the chart. Default implementation calls {@link
	 * #getFields(LyChartController) getFields(controller)}.
	 *
	 * @param  controller The current chart controller.
	 * @return fields that may be used as X variables.
	 * @see    #getFields(LyChartController)
	 */
	public LyInfoList<LyFieldInfo> getXFields(LyChartController controller)
	{
		return getFields(controller);
	}

	/**
	 * Gets fields that may be used as Y variables in the chart. Default implementation calls {@link
	 * #getFields(LyChartController) getFields(controller)}.
	 *
	 * @param  controller The current chart controller.
	 * @return fields that may be used as Y variables.
	 * @see    #getFields(LyChartController)
	 */
	public LyInfoList<LyFieldInfo> getYFields(LyChartController controller)
	{
		return getFields(controller);
	}

	/**
	 * Gets fields that may be used as group variables in the chart. Default implementation returns
	 * all fields that are not multiple choice or relation fields.
	 *
	 * @param  controller The current chart controller.
	 * @return the fields that may be used as group variables.
	 */
	public LyInfoList<LyFieldInfo> getGroupFields(LyChartController controller)
	{
		LyInfoList<LyFieldInfo>  list = controller.getFields();
		LyInfoList<LyFieldInfo>  result = new LyInfoList<LyFieldInfo>();
		LyFieldInfo	    fieldInfo;
		int			    i, n = list.getSize();

		// Get fieldinfo list
		for (i = 0; i < n; i++)
		{
			fieldInfo = list.getInfo(i);

			// Create relation to the fieldInfo if its type is choice or relation
			if (((fieldInfo instanceof LyChoiceFieldInfo)
						&& !((LyChoiceFieldInfo)fieldInfo).isMultiple())
					|| ((fieldInfo instanceof LyRelationFieldInfo)
						&& !((LyRelationFieldInfo)fieldInfo).isMultiple()))
				result.addInfo(fieldInfo);
		}

		return result;
	} // end method getGroupFields

	/**
	 * Gets fields that may be used as variables. Default implementation returns all fields that are
	 * not text fields.
	 *
	 * @param  controller The current chart controller.
	 * @return the fields that may be used as variables.
	 */
	protected LyInfoList<LyFieldInfo> getFields(LyChartController controller)
	{
		LyInfoList<LyFieldInfo>  list = controller.getFields();
		LyInfoList<LyFieldInfo>  result = new LyInfoList<LyFieldInfo>();
		LyFieldInfo	             fieldInfo;
		int			             i, n = list.getSize();

		// Get fieldinfo list
		for (i = 0; i < n; i++)
		{
			fieldInfo = list.getInfo(i);

			// Create relation to the fieldInfo if its type is not text
			// (because charts with text has no sense)
			if (!(fieldInfo instanceof LyTextFieldInfo))
				result.addInfo(fieldInfo);
		}

		return result;
	}

	/**
	 * Gets the image associated to the given object for the given form. Default implementation does
	 * nothing and return null.
	 *
	 * @param  formController the form controller
	 * @param  object         object whose image is looked for
	 * @return Logical name of the image.
	 */
	public String getImage(LyFormController formController, LyObject object)
	{
		return null;
	}

	/**
	 * Gets value of the given fieldInfo. <i>The default implementation</i> does nothing and return
	 * null.
	 *
	 * @param  form      the current LyFormController instance.
	 * @param  fieldInfo the fieldInfo that contains the value
	 * @return the value of the given fieldInfo as an Object
	 */
	public Object getValue(LyFormController form, LyFieldInfo fieldInfo)
	{
		return null;
	}

	/**
	 * Gets the chart context objects list.<br />
	 * <b>If</b> chart objects are selected (the given object list is not null) the return this list
	 * <br />
	 * <b>else if</b> the parent context is not null then return the parent context object list
	 * <br />
	 * <b>else if</b> the parent context is null then return a new worspace<br />
	 * <b>else</b> return null.
	 *
	 * @param  controller    the controller
	 * @param  session       the current session
	 * @param  classInfo     the classInfo that contains values
	 * @param  objects       the selected objects of the current class
	 * @param  parentContext the context of the parent class
	 * @return the context for the chart
	 */
	public LyObjectList getChartContext(LyController controller, LySession session,
		LyClassInfo classInfo, LyObjectList objects, LyWorkSpace parentContext)
	{
		LyAction	action = ((LyAreaController)controller).getAction();
		
		if (action.hasMark(LyAction.MARK_NO_CONTEXT))
			return null;
		
		LyObjectList result = null;

		// first case : the objects selection is not null
		if (null != objects)
			result = objects.getList(session, classInfo, false);

		// second case : the objects selection is null and there is no parent class
		else if (null == parentContext)
		{
			if (null == classInfo)
				controller.getEnvironment().logError("BF AppEngine : chart defined on action \"" +
				     controller.getId() + "\" has no classInfo");
			else
				// XXX TODO : optimize the fields loading.
				result = new LyClassWorkSpace(session, classInfo, false);
		}

		// third case : the objects selection is null but there is a parent
		// class with a selection
		else if (null != parentContext)
			result = parentContext.getList(session, classInfo, false);

		return result;
	}

	/**
	 * Indicates if a fieldInfo is displayed in this form or not. Default implementation does
	 * nothing and return true.
	 *
	 * @param  form      the current LyFormController instance.
	 * @param  fieldInfo the fieldInfo which is tested for addition in the view.
	 * @return true if the field should be added, false if it should be skipped.
	 */
	public boolean isDisplayed(LyFormController form, LyFieldInfo fieldInfo)
	{
		return true;
	}

	/**
	 * Computes a printable string (if the given typeVar is text, number or date) or a Vector of
	 * strings (if the given fieldInfo is a multiple choice or relation) from the given value.
	 *
	 * @param  controller the current controller instance
	 * @param  fieldInfo  the concerned field info
	 * @param  object     the concerned object (may be null)
	 * @param  value      the value to transform into a string. It comes from the given fieldInfo
	 * @param  typeVar    the field type defined in {@link leon.info.LyChartInfo LyChartInfo} such
	 *                    as {@link leon.info.LyChartInfo.Type#DATE}, {@link
	 *                    leon.info.LyChartInfo.Type#NUMBER} and {@link
	 *                    leon.info.LyChartInfo.Type#TEXT}
	 * @param  yAxis      indicates whether the value is computed for y or x axis
	 * @return a String corresponding to the value, or a Vector of Strings if it is a multiple
	 *         choice or relation.
	 */
	public Object getStringChartValue(LyGraphicalChartController controller, LyFieldInfo fieldInfo,
		LyObject object, Object value, LyChartInfo.Type typeVar, boolean yAxis)
	{
		if (value == null)
			return null;
		
		LyEnvironment environment = controller.getEnvironment();
		String		  stringValue = null;
			
		if ((typeVar == LyChartInfo.Type.TEXT)
				|| (typeVar == LyChartInfo.Type.NUMBER))
			stringValue = value.toString();
		else if ((typeVar == LyChartInfo.Type.DATE)
                && (value instanceof Date))
			stringValue = LyDateFormatter.format((Date)value,
					environment.getStringDateFormat(LyEnvironment.TYPE_DATE_TIME));

		// Special case : the type of variable is single choice or relation :
		// recover the name
		// there is no TYPECHOICE in LyChartInfo to permit mix between choice
		// and text or number or date
		if ((fieldInfo instanceof LyChoiceFieldInfo)
				&& !fieldInfo.isMultiple())
			stringValue = environment.translate(fieldInfo.toName(value));
		else if ((fieldInfo instanceof LyRelationFieldInfo)
                && !fieldInfo.isMultiple())
            stringValue = fieldInfo.toName(value);
		// Particular case for multiple relations :
		// return a vector that contains each name of related objects
		else if ((fieldInfo instanceof LyRelationFieldInfo)
             && fieldInfo.isMultiple())
		{
			// vect contains each name of related objects
			Vector<String>   vect = new Vector<String>();
			LyObjectId[]     objectIds = (LyObjectId[])value;
			LyObject         relatedObject;
			String		     objectName;
			int			     i, nObjectIds = objectIds.length;

			// Retrieve each name of related objects
			for (i = 0; i < nObjectIds; i++)
			{
				relatedObject = objectIds[i].getLogicalObject();
				objectName = ((relatedObject != null) ? controller.getName(relatedObject)
													  : objectIds[i].toName());
				vect.addElement(objectName);
			}

			return vect;
		}
		// Particular case for multiple choices :
		// return a vector that contains each name of related objects
		else if ((fieldInfo instanceof LyChoiceFieldInfo)
		        && fieldInfo.isMultiple())
		{
			// vect contains each name options
			Vector<String> vect = new Vector<String>();

			if (value instanceof BitSet)
			{
				BitSet		   objectIds = (BitSet)value;
				LyChoiceOption option;
				String		   optionName;
				short		   i;
				int			   nObjectIds = objectIds.size();

				// Retrieve each name of options
				for (i = 0; i < nObjectIds; i++)
				{
					if (objectIds.get(i))
					{
						option = ((LyChoiceFieldInfo)fieldInfo).getChoiceOption(i);
						optionName = ((option != null) ? (String)option.getName()
													   : new Short(i).toString());
						vect.addElement(optionName);
					}
				}
			}
			else
				vect.addElement(value.toString());

			return vect;
		}

		return stringValue;
	} // end method getStringChartValue

	/**
	 * Gets the label that must be shown in the legend of a bar chart. <i>The default
	 * implementation</i> does nothing and return null.
	 *
	 * @param  controller the current controller instance
	 * @param  fieldInfo  the concerned field info
	 * @param  object     the concerned object (may be null)
	 * @param  value      the value to transform into a string for the legend label. Comes from the
	 *                    given fieldInfo.
	 * @return null if the default label must be used, a string value otherwise
	 */
	public String getBarLegendLabel(LyGraphicalChartController controller, LyFieldInfo fieldInfo,
		LyObject object, Object value)
	{
		return null;
	}

	/**
	 * Gets the label that must be used for a point on a scatter plot chart.
	 *
	 * @param  controller the current controller instance
	 * @param  object     the object concerned
	 * @return the label that must be used for a given point
	 */
	public String getScatterPlotLabel(LyGraphicalChartController controller, LyObject object)
	{
		String result = controller.getName(object);

		return controller.getEnvironment().translate(result);
	}
	
	/**
	 * Return the label that must be used for a point on a scatter plot legend.
	 * @param controller the current controller instance
	 * @param object the object concerned
	 * @return null if label used on the chart must be used, a string value otherwise
	 */
	public String getScatterPlotLegendLabel(LyGraphicalChartController controller, LyObject object)
	{
		return null;
	}

	/**
	 * Indicates if the legend must be positionned on the right of the chart.
	 *
	 * @param  controller the current controller instance
	 * @return true if the legend must be positionned on the right of the chart, false otherwise
	 */
	public boolean isLegendRightPositionned(LyGraphicalChartController controller)
	{
		return false;
	}

	/**
	 * Indicates whether empty values must be displayed on the legend.
	 * 
	 * @param controller the current controller instance
	 * @return true if empty values must be displayed on the legend, false otherwise
	 */
	public boolean displayLegendEmptyValues(LyGraphicalChartController controller)
	{
		return true;
	}
	
	/**
	 * Indicates whether sensitive areas must be generated or not.
	 *
	 * @param  controller the current controller instance
	 * @return true if sensitive areas must be generated on the chart, false otherwise
	 */
	public boolean generateSensitiveAreas(LyGraphicalChartController controller)
	{
		return false;
	}

	/**
	 * Complete chart component before showing it.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param controller the current controller instance
	 * @param chart      the chart component
	 */
	public void completeChart(LyGraphicalChartController controller, LyChart chart)
	{
	}
} // end class LyChartBehavior
