package leon.app.behavior;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.setinterface.*;

import java.util.*;


/**
 * Behavior for chat action.
 */
public class LyChatBehavior extends LyUserInputBehavior implements LyQuestionAnswerInterface,
    LySessionListenerInterface
{

    /**
     * File version
     */
    public static final String __VERSION =
        "$Id: LyChatBehavior.java 31812 2011-07-28 12:45:13Z smaupin $";

    /** actors of conversation. */
    private LyLabel _chatActors;

    /** chat window. */
    private LyText _chatWindow;

    /** chat text input. */
    private LyText _chatInput;

    /** main controller. */
    private LyAreaController _controller;

    /** conversation with list of sessions. */
    private LyConversation _conversation = new LyConversation();

    /** LyTab of conversation. */
    private LyTab _conversationTab;

    /** color of sender msg */
    private String _senderColor;

    /** color of server msg */
    private String _serverColor;

    /**
     * Draw chat window.
     * @param areaController The area.
     */
    public void initChatWindow(LyAreaController areaController)
    {
        LyUserInputController chatController = (LyUserInputController) areaController;
        LyEnvironment         env = chatController.getEnvironment();
        _senderColor = LyColors.getHexString(env.getLyColor("LY_CHAT_SENDER_MSG"));
        _serverColor = LyColors.getHexString(env.getLyColor("LY_CHAT_SERVER_MSG"));


        _chatActors = chatController.addLabel("chat_label", null, "", 0);
        _chatActors.setNCols(30);

        _chatWindow = chatController.addText("chat_window", null, null, 26, 20, null, true, 1);
        _chatWindow.setType(LyText.Type.RICH);
        _chatWindow.setEditable(false);
        _chatWindow.setWrap(true);
        _chatWindow.setScroll(LyText.Scroll.BOTTOM);
        _chatInput = chatController.addText("chat_input", null, null, 24, 1, null, false, 2);
        _chatInput.setWrap(true);

        LyForm form = _chatInput.getParentForm();

        if (form != null)
        {

            for (int i = 0; i < form.getFormFieldCount(); i++)
            {
                LyFormField formField = form.getFormField(i);

                if ("chat_input".equals(formField.getId()))
                {
                    LyConstraint constraint = new LyConstraint();
                    constraint.setFill(LyConstraint.Fill.BOTH);
                    formField.setConstraint(constraint);

                    LyToolBar    toolBar = formField.getToolBar();
                    LyConstraint toolconstraint = new LyConstraint();
                    toolconstraint.setHAlign(LyConstraint.HAlign.LEFT);

                    LyTool envoyer = new LyTool(LyFormController.DEMAND_VALIDATE, null);
                    envoyer.setImage("LY_ACTION_CHAT_SEND");
                    envoyer.setConstraint(toolconstraint);

                    if (toolBar == null)
                    {
                        toolBar = new LyToolBar();
                        toolBar.setController(_controller.getId());
                    }

                    toolBar.addTool(envoyer);
                    formField.setToolBar(toolBar);
                }

                if ("chat_window".equals(formField.getId()))
                {
                    LyConstraint constraint = new LyConstraint();
                    constraint.setNCols(26 * 8);
                    constraint.setFill(LyConstraint.Fill.HORIZONTAL);
                    formField.setConstraint(constraint);
                }
            }
        }
    }


    /**
     * Write a message in chat window.
     * @param message The message
     * @param user The user
     * @param me true if i am the sender
     */
    public void writeMessage(String message, LyObject user, boolean serverMessage)
    {

        // if tab is inactive, add visual marker for new message
        LySession session = _controller.getSession();

        if ((session == null) || (session.getUser() == null))
            return;

        LyEnvironment env = _controller.getEnvironment();
        LyController  socialController = (LyController) session.getValue("social_controller");

        if (socialController != null)
        {
            LySocialBuilder socialBuilder = (LySocialBuilder) socialController.getBuilder();

            if ((socialBuilder != null) && (socialBuilder.getTabs() != null))
            {
                String selectedTab = socialBuilder.getTabs().getSelectedTab();

                if ((_conversationTab != null) && (selectedTab != null)
                        && !selectedTab.equals(_conversationTab.getId()))
                    _conversationTab.setColor("LY_CHAT_NEW_MSG");
            }
        }

        String   format = env.getStringDateFormat(LyDateFieldInfo.Type.HOUR_MIN.ordinal());
        String   date = LyDateFormatter.format(new Date(), format);
        String   oldLabel = _chatWindow.getLabel();
        LyObject currentUser = session.getUser();

        boolean me = false;

        if ((user != null) && (currentUser != null) && user.getId().equals(currentUser.getId()))
            me = true;

        StringBuffer newLabel = new StringBuffer();

        if (oldLabel != null)
            newLabel.append(oldLabel);

        if (me)
            newLabel.append("<strong>");

        if (serverMessage)
        {
            newLabel.append("<span style=\"color:#");
            newLabel.append(_serverColor);
            newLabel.append(";font-style:italic;\">");
        }
        else
        {
            newLabel.append("<span style=\"color:#");
            newLabel.append(_senderColor);
            newLabel.append("\">");
        }

        newLabel.append("[");
        newLabel.append(date);
        newLabel.append("] ");

        if (user != null)
        {

            if (!me)
                newLabel.append(user.getName());
            else
                newLabel.append(env.translate("LY_CHAT_USER"));
        }

        newLabel.append(" ");

        if (!serverMessage)
            newLabel.append(":</span><br />");

        newLabel.append(message);

        if (serverMessage)
            newLabel.append("</span>");

        if (me)
            newLabel.append("</strong>");

        newLabel.append("<br />");
        _chatWindow.setLabel(newLabel.toString());
    }

    @Override public boolean startAction(LyAreaController areaController, LyAction action)
    {
        LySession     currentSession = areaController.getSession();
        LyApplication application = areaController.getApplication();
        _controller = (LyAreaController) areaController;
        application.addSessionListener(this);
        _conversation.addSession(currentSession, _controller);
        initChatWindow(areaController);
        return super.startAction(areaController, action);
    }

    @Override public boolean validate(LyController controller, LyParameterValues values)
    {
        String inputValue = _chatInput.getValue();

        if ((inputValue == null) || "".equals(inputValue))
            return false;

        _conversation.sendMessage(inputValue, _controller.getSession(), false, false);
        _chatInput.setLabel("");
        return false;
    }

    @Override public boolean handleDemand(LyController controller, LyDemand demand)
    {

        // add validate on enter
        if (demand.getType() == LyActionDemand.TYPE_ACTION)
        {
            String source = demand.getSource();
            String value = ((LyActionDemand) demand).getValue();

            if ("chat_input".equals(source) && "_validateField".equals(value))
                validate(controller, null);
        }
        else if ((demand instanceof LyActionDemand)
                && (demand.getType() == LyActionDemand.TYPE_HREF))
        {
            String     value = ((LyActionDemand) demand).getValue();
            LyObjectId objectId = new LyObjectId(controller.getApplication(), value);
            LyObject   object = objectId.getLogicalObject();

            if (object != null)
            {
                LyClassInfo classInfo = object.getClassInfo();
                LyAction    consult = classInfo.getDefaultAction();

                if (consult == null)
                    consult = classInfo.getActionWithId(LyAction.ID_CONSULT);

                if (consult != null)
                {
                    LyAreaController areaController = (LyAreaController) controller;

                    controller.getSession().invokeAction(areaController.getParent(false), consult,
                        classInfo, new LySimpleObjectList(object), true);
                }
            }

            return true;
        }
        else if ((demand instanceof LyDragDropDemand)
                && (demand.getType() == LyDragDropDemand.TYPE_DROP))
        {
            String source = demand.getSource();

            if ((source != null) && ("chat_input".equals(source) || "chat_window".equals(source)))
            {
                // An object was dropped in the chat window
                LyDragDropDemand dropDemand = (LyDragDropDemand) demand;
                String           draggedObjectId = (String) dropDemand.getAttribute(
                        "_draggedObjectId");

                try
                {
                    LyObjectId       objectId = new LyObjectId(_controller.getApplication(),
                            draggedObjectId);
                    LyObject         object = objectId.getLogicalObject();
                    Vector<LyObject> users = _conversation.isObjectVisible(object,
                            _controller.getSession());

                    // Check if the users have an access to the sent object
                    if ((users != null) && !users.isEmpty())
                    {
                        String usersList = "";

                        for (LyObject user : users)
                        {

                            if (usersList.length() > 0)
                                usersList += ", ";

                            usersList += user.getName();
                        }

                        String message = _controller.getEnvironment().getMessage(
                                "LY_SEND_OBJECT_CONFIRM", object.getName(), usersList);

                        _controller.getViewManager().askQuestion(null, message, this, object);
                    }
                    else
                        _conversation.sendObject(object, _controller.getSession());
                }
                catch (Exception e)
                {
                    _controller.showError("LY_ER_SEND_OBJECT");
                }
            }

            return true;
        }

        return super.handleDemand(controller, demand);
    }

    /**
     * @see LyQuestionAnswerInterface#questionAnswer(boolean, Object)
     */
    public void questionAnswer(boolean result, Object clientData)
    {
        if (clientData instanceof LySession)
        {
            LySession     currentSession = _controller.getSession();
            LyEnvironment env = currentSession.getEnvironment();
            LySession     targetSession = (LySession)clientData;

            if (result)
            {
                if (targetSession != null)
                {
                    if (currentSession.getUser() == null)
                    {
                        LyViewManager viewManager = targetSession.getViewManager();

                        if (viewManager != null)
                            viewManager.showMessage(null, "", "LY_WARNING_ICON",
                                "LY_CHAT_DISCONNECTED", null, null, null);

                        return;
                    }

                    updateSessionStatus(targetSession, true);
                    showTargetChatWindow(targetSession);

                    // if something was typed before, send message
                    if ((_chatWindow.getLabel() != null) && !"".equals(_chatWindow.getLabel()))
                    {
                        LyAreaController targetChatController = _conversation.getController(
                                targetSession.getId());

                        if (targetChatController != null)
                        {
                            LyChatBehavior behavior = (LyChatBehavior) targetChatController
                                .getBehavior();

                            if ((behavior != null) && (behavior.getChatWindow() != null))
                                behavior.getChatWindow().setLabel(epureLabel(_chatWindow.getLabel(),
                                        env));
                        }
                    }

                    writeMessage(env.translate("LY_CHAT_USERINCOMM_MSG"), targetSession.getUser(),
                        true);
                }
            }
            else
                writeMessage(env.translate("LY_CHAT_REFUSED"), targetSession.getUser(), true);
        }
        else if (clientData instanceof LyObject)
        {

            if (result)
                _conversation.sendObject((LyObject) clientData, _controller.getSession());
        }
    }

    /**
     * Epure chat window.
     * @param label label of chat window to epure
     * @param env LyEnvironment
     * @return String Epured Label
     */
    private String epureLabel(String label, LyEnvironment env)
    {

        if ((label != null) && !"".equals(label))
        {
            LySession current = _controller.getSession();
            label = label.replace("<strong>", "");
            label = label.replace("</strong>", "");

            if ((current != null) && (current.getUser() != null))
                label = label.replace(env.translate("LY_CHAT_USER"), current.getUser().getName());
        }

        return label;
    }


    /**
     * init chat on target session.
     * @param targetSession target session
     */
    private void showTargetChatWindow(LySession targetSession)
    {
        // open chat window
        LyApplication         application = _controller.getApplication();
        LyDashboardController dashboardcontroller = null;

        // retrieve dashboard controller to open social panel
        for (int i = 0; i != targetSession.getControllerCount(); i++)
        {
            if (targetSession.getController(i) instanceof LyDashboardController)
            {
                dashboardcontroller = (LyDashboardController) targetSession.getController(i);
                break;
            }
        }

        if (dashboardcontroller == null)
            return;

        LyAction socialAction = application.getActionWithId("_social", null);
        LyAction chatAction = application.getActionWithId("_chat", null);

        // if chat is already launched
        LyAreaController targetSocialController = (LyAreaController)targetSession.getValue("social_controller");

        if ((targetSocialController == null)
        		&& (targetSession.getUser() != null))
            targetSocialController = (LyAreaController)targetSession.invokeAction(
                    dashboardcontroller, socialAction, targetSession.getUser().getClassInfo(), null, true);

        if (targetSocialController == null)
        	return;
        
        LySocialBuilder targetBuilder = (LySocialBuilder)targetSocialController.getBuilder();
        LySession       currentSession = _controller.getSession();

        if ((currentSession != null)
        	&& (currentSession.getUser() != null))
        {
            // sent the current user to open chat window. don't show before adding session to it.
            LyObjectList     userList = new LySimpleObjectList(currentSession.getUser());
            LyAreaController targetChatController = (LyAreaController) targetSession.invokeAction(
                    targetSocialController, chatAction, currentSession.getUser().getClassInfo(),
                    userList, false);
            targetChatController.initialize();

            LyTab targetTab = targetBuilder.addTab(targetChatController);
            _conversation.addSession(targetSession, targetChatController);

            LyChatBehavior targetBehavior = (LyChatBehavior) targetChatController.getBehavior();

            if (targetBehavior != null)
                targetBehavior.setConversationTab(targetTab);

            linkSessions();
        }
    }

    /**
     * Link conversation together
     */
    private void linkSessions()
    {
        LySession currentSession = _controller.getSession();

        for (LySession session : _conversation.getSessions().values())
        {

            if (session.getId().equals(currentSession.getId()))
            {
                updateActorsLabel();
                continue;
            }

            LyController controller = _conversation.getController(session.getId());

            if (controller != null)
            {
                LyChatBehavior behavior = (LyChatBehavior) controller.getBehavior();

                if (behavior != null)
                {
                    behavior.setConversation(_conversation);
                    behavior.updateActorsLabel();
                }
            }
        }
    }

    /**
     * update Actors Label.
     */
    public void updateActorsLabel()
    {

        if (_conversation != null)
        {
            LyEnvironment env = _controller.getEnvironment();
            LySession     currentSession = _controller.getSession();
            boolean       first = true;
            StringBuffer  actorLabel = new StringBuffer();
            actorLabel.append(env.translate("LY_CHAT_ACTORS"));
            actorLabel.append("\n");

            for (LySession session : _conversation.getSessions().values())
            {

                if (!currentSession.getUser().getId().equals(session.getUser().getId()))
                {

                    if (!first)
                        actorLabel.append(",");
                    else
                        first = false;

                    actorLabel.append(session.getUser().getName());
                }
            }

            _chatActors.setLabel(actorLabel.toString());
        }
    }


    /**
     * Update status of session.
     * @param session Session
     * @param chat new status.
     */
    public void updateSessionStatus(LySession session, boolean chat)
    {

        if (session == null)
            return;

        session.setChatting(chat);
    }

    /**
     * Remove session from conversation.
     * @param session session to remove
     */
    public void tabRemoveSession(LySession session)
    {

        if (session != null)
        {
            _conversation.sendMessage("LY_CHAT_QUIT", session, true, true);
            _conversation.removeSession(session);
            _conversation.updateActorLabels();
        }
    }

    /**
     * @see LySessionListenerInterface#sessionAdded(LyApplication, LyDataSession)
     */
    public void sessionAdded(LyApplication application, LyDataSession session)
    {
    }


    /**
     * @see LySessionListenerInterface#sessionLoggedIn(LyApplication, LyDataSession, LyObject)
     */
    public void sessionLoggedIn(LyApplication application, LyDataSession session, LyObject user)
    {
    }


    /**
     * @see LySessionListenerInterface#sessionRemoved(LyApplication, LyDataSession)
     */
    public void sessionRemoved(LyApplication application, LyDataSession session)
    {

        if (session != null)
        {
            LySession currentSession = _controller.getSession();

            // if user is in conversation
            if (!session.getId().equals(currentSession.getId())
                    && (_conversation.getTargetUserSession(session.getId()) != null))
            {
                LyEnvironment env = _controller.getEnvironment();
                writeMessage(env.translate("LY_CHAT_QUIT"), session.getUser(), true);
                updateActorsLabel();
                _conversation.removeSession((LySession) session);
            }
        }
    }

    /**
     * Conversation accessor.
     * @return conversation
     */
    public LyConversation getConversation()
    {
        return _conversation;
    }

    /**
     * Conversation setter.
     * @return conversation
     */
    public void setConversation(LyConversation conversation)
    {
        _conversation = conversation;
    }


    /**
     * Free this behavior. This method is provided to allow behavior to unregister itself from
     * object sources or workspaces if needed. It should be called only by the controller that built
     * this behavior.<br />
     * <i>The default implementation</i> does nothing.
     */
    public void free()
    {

        if (_controller != null)
        {
            LyApplication application = _controller.getApplication();

            if (application != null)
                application.removeSessionListener(this);
        }
    }

    /**
     * Set conversation tab.
     * @param tab Tab
     */
    public void setConversationTab(LyTab tab)
    {
        _conversationTab = tab;
    }

    public LyText getChatWindow()
    {
        return _chatWindow;
    }
}
