/*
 * LyClassBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.control.form.*;
import leon.data.*;
import leon.data.datainterface.*;
import leon.info.*;
import leon.info.save.*;
import leon.misc.*;


/**
 * This class implements the default behavior for consult, set, filter and delete actions. It can be
 * inherited by a specific user behavior.
 *
 * @author   Lyria S.A.
 * @version  $Id: LyClassBehavior.java 36510 2012-08-17 15:43:06Z jlsaouli $
 * @see      leon.app.behaviorinterface.LyFormBehaviorInterface
 * @see      leon.control.LyFormController
 * @see      leon.app.behaviorinterface.LyConsultBehaviorInterface
 * @see      leon.control.LyConsultController
 * @see      leon.app.behaviorinterface.LySetBehaviorInterface
 * @see      leon.control.LySetController
 * @see      leon.app.behaviorinterface.LyCreateBehaviorInterface
 * @see      leon.control.LyCreateController
 * @see      leon.app.behaviorinterface.LyFilterBehaviorInterface
 * @see      leon.control.LyFilterController
 * @see      leon.app.behaviorinterface.LyDeleteBehaviorInterface
 * @see      leon.control.LyDeleteController
 */
public class LyClassBehavior extends LyActionBehavior
    implements LyClassBehaviorInterface
{
    /** File version. */
    public static final String __VERSION = "$Id: LyClassBehavior.java 36510 2012-08-17 15:43:06Z jlsaouli $";

    /**
     * Code returned when there is no modification on the value for controlNewValue, checkNewValue
     * and controlNewValues.<br />
     * All codes are :
     *
     * <ul>
     * <li>{@link #STATUS_IGNORE}: the value is not modified</li>
     * <li>{@link #STATUS_KO}: the value is not correct</li>
     * <li>{@link #STATUS_OK}: the value is correct</li>
     * <li>{@link #STATUS_WARNING}: the value is on warning</li>
     * </ul>
     */
    public static final short STATUS_IGNORE = -1;

    /**
     * Code returned when the value is correct for controlNewValue, checkNewValue and
     * controlNewValues.
     *
     * @see  #STATUS_IGNORE
     */
    public static final short STATUS_OK = 0;

    /**
     * Code returned when the value is not correct for controlNewValue, checkNewValue and
     * controlNewValues.
     *
     * @see  #STATUS_IGNORE
     */
    public static final short STATUS_KO = 1;

    /**
     * Code returned when the value is on warning for controlNewValue, checkNewValue and
     * controlNewValues.
     *
     * @see  #STATUS_IGNORE
     */
    public static final short STATUS_WARNING = 2;

    /**
     * Construct a class behavior.
     */
    public LyClassBehavior()
    {
    }

    /**
     * Gets the image associated to the given object for the given form.
     *
     * @param   formController  the form controller
     * @param   object          object whose image is looked for
     *
     * @return  Logical name of the image.
     *
     * @see     leon.control.LyController#getImage(leon.info.LyObjectInfo)
     */
    public String getImage(LyFormController formController, LyObject object)
    {
        return ((object == null) ? null : formController.getImage(object.getClassInfo()));
    }

    /**
     * Gets title corresponding to the given action. Overriden to computes title with the action
     * name and the action classInfo name and the object name separated by " : ".
     *
     * @param   areaController  the current LyAreaController instance
     * @param   action          the action that was executed
     *
     * @return  Title for the given action.
     */
    public String getTitle(LyAreaController areaController, LyAction action)
    {
        if (!(areaController instanceof LyFormController))
            return super.getTitle(areaController, action);

        LyClassInfo classInfo = areaController.getClassInfo();
        LyObject    object = ((LyFormController)areaController).getObject();
        String      title = action.getName();

        if ((classInfo != null)
                && !title.endsWith(classInfo.getName()))
            title = action.getName() + " : " + classInfo.getName();

        LyAction rootAction = action.getRoot();
        String   actionId = rootAction.getId();

        if (actionId == LyAction.ID_CREATE)
            object = ((LySetController)areaController).getModel();

        if (object != null)
        {
            LyInfoLabel infoLabel = classInfo.getLabel(LyInfoLabel.Type.NAME);
            String      name = null;

            if (infoLabel != null)
                name = infoLabel.getString(object);

            if (name == null)
                name = areaController.getName(object);

            if (name != null)
                title += " : " + name;
        }

        return title;
    } // end method getTitle

    /**
     * Gets value of the given fieldInfo. This method must be efficient because it can be called
     * frequently.<br />
     * <i>The default implementation</i> gets the value of the object associated to the given form.
     *
     * @param   form       the current LyFormController instance.
     * @param   fieldInfo  the fieldInfo that contains the value.
     *
     * @return  the value of the given fieldInfo as an Object
     */
    public Object getValue(LyFormController form, LyFieldInfo fieldInfo)
    {
        LyObject object = form.getObject();

        if ((object == null) && (form instanceof LySetController)
                && !fieldInfo.hasMark(LyFieldInfo.MARK_PRIVATE)
                && !fieldInfo.hasMark(LyFieldInfo.MARK_SECRET))
            object = ((LySetController)form).getModel();

        if (object != null)
        {
            Object value = object.getValue(fieldInfo);

            if (value != null)
                return value;
        }

        if (((form instanceof LySetController)
                    && ((LySetController)form).isCreation())
                && (fieldInfo.getFormula() != null))
        {
            LyValue value = ((LySetController)form).computeFieldValue(fieldInfo);

            return value.getValue();
        }

        return null;
    }

    /**
     * Indicates if the given fieldInfo is displayed in the given form or not.<br />
     * <i>The default implementation</i> does nothing and return true.
     *
     * @param   form       the current LyFormController instance.
     * @param   fieldInfo  the fieldInfo which is tested for addition in the view.
     *
     * @return  true if the field should be added, false if it should be skipped.
     */
    public boolean isDisplayed(LyFormController form, LyFieldInfo fieldInfo)
    {
        return true;
    }

    /**
     * Gets a default value for the given fieldInfo. This method can be called:
     *
     * <ul>
     * <li>to initialize an input field in a creation form</li>
     * <li>to get a value for an empty input field that becomes mandatory</li>
     * <li>to get a value for notNull fields at the end of the creation process</li>
     * </ul>
     *
     * @param   set        the current LySetController instance.
     * @param   fieldInfo  the fieldInfo that contains the value.
     *
     * @return  the value of the given fieldInfo as an Object
     */
    public Object getDefaultValue(LySetController set, LyFieldInfo fieldInfo)
    {
        LySession     session = set.getSession();
        LyEnvironment environment = set.getEnvironment();
        Object        defaultValue = fieldInfo.getDefaultValue();
        LyValue       fieldValue;

        if ((fieldInfo instanceof LyChoiceFieldInfo)
                && (((LyChoiceFieldInfo)fieldInfo).getType() != LyChoiceFieldInfo.Type.BOOLEAN)
                && environment.testResource("LY_EMPTY_VALUE_FOR_CHOICES"))
            defaultValue = null;
        else
            defaultValue = fieldInfo.getDefaultValue();

        if ((defaultValue instanceof String) && (((String)defaultValue).charAt(0) == '$'))
        {
            defaultValue = session.getValue(defaultValue);

            if ((defaultValue != null) && fieldInfo.isMultiple())
            {
                Object[] tableValue = (Object[])java.lang.reflect.Array.newInstance(defaultValue.getClass(), 1);

                tableValue[0] = defaultValue;
                defaultValue = tableValue;
            }
        }

        // Build value from default value
        if (defaultValue != null)
        {
            // For text field, translate the default value to ensure it will be displayed correctly
            if ((defaultValue instanceof String) && (fieldInfo instanceof LyTextFieldInfo))
                defaultValue = environment.translate((String)defaultValue);

            fieldValue = fieldInfo.buildFieldValue(defaultValue);
            return fieldValue.getValue();
        }

        // Clone value of model object except for creation with private and
        // secret fields
        LyObject model = set.getModel();

        if ((model != null)
                && !(set.isCreation()
                    && (fieldInfo.hasMark(LyFieldInfo.MARK_PRIVATE)
                        || fieldInfo.hasMark(LyFieldInfo.MARK_SECRET))))
        {
            fieldValue = model.getFieldValue(fieldInfo);
            return fieldValue.getValue();
        }

        // Last attempt : try to get a valid value thanks to the context for
        // relations
        if (!(fieldInfo instanceof LyRelationFieldInfo))
            return null;

        if (fieldInfo instanceof LyStructFieldInfo)
            return null;

        LyRelationFieldInfo relation = (LyRelationFieldInfo)fieldInfo;

        if (((LyRelationFieldInfo)fieldInfo).ignoresContext())
            return null;

        LyObjectList context = getRelationContext(set, relation);
        LyClassInfo  relationClassInfo;
        int          i, n;

        if ((context == null) || (context.getSize() == 0))
        {
            context = set.getContext();
            if ((context == null) || (context.getSize() == 0))
            {
                if (set.isInitialized())
                {
                    LyController parent = set.getParent();
                    LyObject     object = null;

                    while ((parent != null)
                    		&& !(parent instanceof LyFormController))
                    	parent = parent.getParent();

                    if (parent != null)
                    {
                    	LyFormController	parentFormController = (LyFormController)parent;

                        object = parentFormController.getObject();

                        if (object != null)
                        	context = new LySimpleObjectList(object);
                        else
                        {
                        	LyClassInfo	parentClassInfo = parentFormController.getClassInfo();

                        	if (relation.relatedWith(parentClassInfo))
                        	{
                        		LyInfoList<LyFieldInfo>	parentIdFields = parentClassInfo.getIdFieldInfos();
                        		int						f, nf = parentIdFields.getSize();

                        		if (nf == 0)
                        			return null;

                        		LyFieldInfo				parentIdField;
                        		LyValue					parentIdValue;
                        		Object[]				parentIdValues = new Object[nf];

                        		for (f = 0; f < nf; f++)
                        		{
                        			parentIdField = parentIdFields.getInfo(f);
                        			parentIdValue = parentFormController.getFieldValue(parentIdField);
                        			parentIdValues[f] = (parentIdValue == null ? null : parentIdValue.getValue());
                        		}

                        		if (nf == 1)
                        			if (parentIdValues[0] == null)
                        				return null;
                        			else
                        				return new LyObjectId(parentClassInfo, parentIdValues[0]);
                        		else
                        			return new LyObjectId(parentClassInfo, parentIdValues);
                        	}
                        	else
                        			return null;
                        }
                    }
                    else
                    	return null;
                }
                else
                    return null;
            }
        }

        if (!fieldInfo.isMultiple())
        {
            LyObjectList list = null;

            for (i = 0, n = relation.getClassInfoCount(); i < n; i++)
            {
                relationClassInfo = relation.getClassInfo(i);
                list = context.getList(session, relationClassInfo);

                // Only one object could be accepted
                if ((list != null) && (list.getSize() == 1))
                {
                    fieldValue = fieldInfo.buildFieldValue(list);
                    return fieldValue.getValue();
                }
            }
        }

        return null;
    } // end method getDefaultValue

    /**
     * Checks if the given fieldInfo is optional or not for the given controller.<br />
     * <i>The default implementation</i> tests if the given fieldInfo has the optional mark.
     *
     * @param   set        the current LySetController instance (creation or set action)
     * @param   fieldInfo  the target fieldInfo to test
     *
     * @return  true if the given fieldInfo is optional, false otherwise
     */
    public boolean isOptional(LySetController set, LyFieldInfo fieldInfo)
    {
        return set.hasMark(fieldInfo, LyFieldInfo.MARK_OPTIONAL);
    }

    /**
     * Checks if the given fieldInfo is editable or not for the given controller.<br />
     * <i>The default implementation</i> return false if the given fieldInfo has the readOnly mark,
     * otherwise if the action is a creation and if the field info has a readOnly control then
     * return true if the read only control priority level is greater than the application priority
     * level on the set object.
     *
     * @param   set        the current LySetController instance (creation or set action)
     * @param   fieldInfo  the target fieldInfo to test
     *
     * @return  true if the given fieldInfo is editable, false otherwise
     */
    public boolean isEditable(LySetController set, LyFieldInfo fieldInfo)
    {
        if (fieldInfo.hasMark(LyFieldInfo.MARK_READ_ONLY))
            return false;

        return true;
    }

    /**
     * Checks if the given fieldInfo is enabled or not for the given controller.<br />
     * <i>The default implementation</i> checks if the given fieldInfo has not the disable mark.
     *
     * @param   set        the current LySetController instance
     * @param   fieldInfo  the target fieldInfo to test
     *
     * @return  true if the given fieldInfo is enabled, false otherwise
     */
    public boolean isEnabled(LySetController set, LyFieldInfo fieldInfo)
    {
        return !fieldInfo.hasMark(LyFieldInfo.MARK_DISABLE);
    }

    /**
     * Indicates the maximum number of values displayed in text components for auto-completion.
     * <br />
     * Default implementation returns the value of the LY_TEXT_AUTO_COMPLETION_MAXIMUM resource
     * except for fields that have the secret or private mark (returns 0).
     *
     * @param   set        the current LySetController instance
     * @param   fieldInfo  the target text fieldInfo to test
     *
     * @return  a positive number if auto-completion is activated, 0 if auto-completion is not
     *          activated, a negative value if there is no maximum.
     */
    public int getAutoCompletionMaximum(LySetController set, LyTextFieldInfo fieldInfo)
    {
        if (!fieldInfo.hasMark(LyFieldInfo.MARK_AUTO_COMPLETION))
            return 0;

        LyEnvironment environment = set.getEnvironment();
        boolean       secretField = fieldInfo.hasMark(LyFieldInfo.MARK_SECRET);
        boolean       privateField = fieldInfo.hasMark(LyFieldInfo.MARK_PRIVATE);

        // Do not authorize auto-completion for secret or private fields
        if (secretField || privateField)
            return 0;

        int autoCompletion = environment.getIntResource("LY_TEXT_AUTO_COMPLETION_MAXIMUM");

        return autoCompletion;
    }

    /**
     * Computes the list of values displayed in text components for auto-completion.<br />
     * Default implementation returns the list of values already loaded for given text field info.
     *
     * @param   set        the current LySetController instance.
     * @param   fieldInfo  the target text fieldInfo.
     *
     * @return  the list of values for auto-completion.
     */
    public String[] getAutoCompletionValues(LySetController set, LyTextFieldInfo fieldInfo)
    {
        // Retrieve all possible values for this text component.
        // Values are retrieved from objects already loaded (perform a local request).
        LyClassInfo classInfo = set.getClassInfo();

        // Check if class info is registered on current application before retrieving
        // its object source. A class info may not have been registered on the application
        // when it has been created on the fly for example.
        String        classId = ((classInfo == null) ? null : classInfo.getId());
        LyApplication application = set.getApplication();

        if (application.getClassInfo(classId) == null)
            return null;

        LySession           session = set.getSession();
        LyObjectSource      source = session.getObjectSource(classInfo);
        Vector<LyFieldInfo> fields = new Vector<LyFieldInfo>(1);

        if (source == null)
            return null;
        if (!classInfo.contains(fieldInfo))
            return null;

        fields.addElement(fieldInfo);

        LyResponse     response = source.getObjects(fields, null, null, -1, true, null);
        LyObjectList   objects = response.getObjects();
        LyObject       object;
        Object         textValue;
        Vector<String> values = new Vector<String>();
        int            i, n = ((objects == null) ? 0 : objects.getSize());

        for (i = 0; i < n; i++)
        {
            object = objects.getObject(i);
            textValue = object.getFromMemory(fieldInfo);

            if ((textValue instanceof String)
                    && (((String)textValue).length() > 0)
                    && !values.contains(textValue))
                values.addElement((String)textValue);
        }

        // Sort the values
        n = values.size();
        Collections.sort(values);

        String[] result = new String[n];

        for (i = 0; i < n; i++)
            result[i] = values.elementAt(i);

        return result;
    } // end method getAutoCompletionValues

    /**
     * Gets the list of all the currently selected options in the context field of the given choice
     * field.
     *
     * @param   set     the set controller
     * @param   choice  the choice fieldinfo
     *
     * @return  a LyInfoList containing the selected options.
     */
    protected LyInfoList<LyInfo> getChoiceContextOptions(LySetController set, LyChoiceFieldInfo choice)
    {
        LyChoiceFieldInfo             contextField = choice.getContextField();
        LyInfoList<LyChoiceFieldInfo> contextFields = new LyInfoList<LyChoiceFieldInfo>();

        contextFields.addInfo(contextField);

        // Build a list for all context fields;
        while (contextField.getContextField() != null)
        {
            contextField = contextField.getContextField();
            contextFields.addInfo(contextField);
        }

        LyChoiceValue      value;
        short              index;
        LyInfoList<LyInfo> contextOptions = new LyInfoList<LyInfo>();
        int                i;

        for (i = 0; i < contextFields.getSize(); i++)
        {
            contextField = contextFields.getInfo(i);
            value = (LyChoiceValue)set.getFieldValue(contextField);
            if ((value == null) || value.isNull())
                continue;

            if (contextField.isMultiple())
            {
                for (short j = 0; j < contextField.getChoiceItemCount(); j++)
                {
                    if (value.isSelected(j))
                        contextOptions.addInfo((LyInfo)contextField.getChoiceItem(j));
                }
            }
            else
            {
                index = value.getShortValue().shortValue();
                contextOptions.addInfo((LyInfo)contextField.getChoiceItem(index));
            }
        }

        return contextOptions;
    } // end method getChoiceContextOptions

    /**
     * Gets domain (list of allowed options) of the given choice fieldInfo.
     *
     * @param   set     the current LySetController instance
     * @param   choice  target choice fieldInfo
     *
     * @return  a list of {@link leon.info.LyChoiceOption} allowed for this choice field info
     */
    public Vector<LyChoiceOption> getChoiceDomain(LySetController set, LyChoiceFieldInfo choice)
    {
        short                      i, n = choice.getChoiceOptionCount();
        Vector<LyChoiceOption>     domain = new Vector<LyChoiceOption>(n);
        LyInfoList<LyChoiceOption> options = choice.getChoiceOptions(true);
        LyChoiceOption             option;
        LyChoiceFieldInfo          contextField = choice.getContextField();

        if (contextField == null)
        {
            for (i = 0; i < n; i++)
            {
                option = options.getInfo(i);
                if (enableChoiceOption(set, choice, option))
                    domain.addElement(option);
            }

            return domain;
        }

        LyInfoList<LyInfo>  contextOptions = getChoiceContextOptions(set, choice);
        LyInfoList<LyInfo>  neededOptions;
        LyChoiceOptionGroup group;
        LyChoiceOption      contextOption;

        // for each option in the choice, if they are valid (enableOption =
        // true)
        // compute the needed context options and check if they are present...
        LyInfoList<LyChoiceOptionGroup> groups;
        int                             j, ng;

        for (i = 0; i < n; i++)
        {
            option = options.getInfo(i);
            if (!enableChoiceOption(set, choice, option))
                continue;

            // build required values...
            groups = choice.getChoiceOptionGroups(option);
            for (j = 0, ng = groups.getSize(); j < ng; j++)
            {
                neededOptions = new LyInfoList<LyInfo>();
                group = groups.getInfo(j);
                do
                {
                    contextOption = group.getOption();
                    if (contextOption != null)
                        neededOptions.addInfo(contextOption);

                    if (group.getParent() instanceof LyChoiceOptionGroup)
                        group = (LyChoiceOptionGroup)group.getParent();
                    else
                        group = null;
                }
                while (group != null);

                // Now we have all needed options...
                if ((neededOptions.getSize() == 0)
                        || contextOptions.contains(neededOptions, true))
                    domain.addElement(option);
            }
        }

        return domain;
    } // end method getChoiceDomain

    /**
     * Checks if the given group is valid for the given option of the choice fieldinfo.
     *
     * @param   set     the current set controller
     * @param   choice  the choice fieldInfo
     * @param   option  the choiceOption
     * @param   group   the choiceOption group
     *
     * @return  true if the group is valid, false otherwise.
     */
    public boolean isValidGroup(LySetController set, LyChoiceFieldInfo choice, LyChoiceOption option,
        LyChoiceOptionGroup group)
    {
        // Build current values
        LyChoiceFieldInfo contextField = choice.getContextField();

        if (contextField == null)
            return true; // all groups are valid;

        LyInfoList<LyInfo> contextOptions = getChoiceContextOptions(set, choice);
        LyInfoList<LyInfo> neededOptions = new LyInfoList<LyInfo>();
        LyChoiceOption     contextOption;

        do
        {
            contextOption = group.getOption();
            if (contextOption != null)
                neededOptions.addInfo(contextOption);

            if (group.getParent() instanceof LyChoiceOptionGroup)
                group = (LyChoiceOptionGroup)group.getParent();
            else
                group = null;
        }
        while (group != null);

        // Now we have all needed options...
        if ((neededOptions.getSize() == 0) || contextOptions.contains(neededOptions, true))
            return true;

        return false;
    } // end method isValidGroup

    /**
     * Indicates if the given option is enabled on the given choice fieldInfo.<br />
     * <i>The default implementation</i> checks if the option is not disabled.
     *
     * @param   set     the current LySetController instance
     * @param   choice  choiceFieldInfo that contains this option
     * @param   option  the option to enabled
     *
     * @return  true if the option is enabled, false otherwise
     */
    public boolean enableChoiceOption(LySetController set, LyChoiceFieldInfo choice, LyChoiceOption option)
    {
        return (validateChoiceOption(set, choice, option)
                && !option.isDisable());
    }

    /**
     * Indicates if the given option is hidden on the given choice fieldInfo.<br />
     * <i>The default implementation</i> checks if the option is not disabled.
     *
     * @param   set     the current LySetController instance
     * @param   choice  choiceFieldInfo that contains this option
     * @param   option  the option to check
     *
     * @return  true if the option is hidden, false otherwise
     */
    public boolean validateChoiceOption(LySetController set, LyChoiceFieldInfo choice, LyChoiceOption option)
    {
        return true;
    }

    /**
     * Gets relation domain (list of available links) for the given relation fieldInfo.
     *
     * @param   set       the current LySetController instance
     * @param   relation  the target relation field
     *
     * @return  a list of LyObjects available for this relation
     */
    public LyObjectList getRelationDomain(LySetController set, LyRelationFieldInfo relation)
    {
        if (!set.isDisplayed(relation))
        {
            if (!(relation.getParent() instanceof LyArrayFieldInfo)
                    || !set.isDisplayed((LyFieldInfo)relation.getParent()))
                return null;
        }

        LyObjectList            links = null, classLinks;
        LyInfoList<LyClassInfo> linkClasses = getRelationClasses(set, relation);
        int                     i, n = ((linkClasses == null) ? 0 : linkClasses.getSize());
        LyClassInfo             linkClass;
        LyObjectList            context = getRelationContext(set, relation);
        LySession               session = set.getSession();
        LyFilter                filter = null;
        boolean                 filtered = false;
        LyEnvironment           environment = set.getEnvironment();
        String                  selectCountStr = environment.getResource("LY_SELECT_RELATION_COUNT");
        int                     selectCount = LyTools.intFromString(selectCountStr), count = 0;
        String                  nameSelectCountStr = environment.getResource("LY_SELECT_BY_NAME_RELATION_COUNT");
        int                     nameSelectCount = LyTools.intFromString(nameSelectCountStr);
        int                     globalSelectCount = -1;
        LyObjectList            currentSelection = null;

        // If auto-completion is used for given relation, do not use select action nor
        // select by name action.
        int completionMax = getAutoCompletionMaximum(set, relation);

        if (completionMax != 0)
        {
            selectCount = -1;
            nameSelectCount = -1;
        }

        // Retrieve the minimum positive of both variables
        globalSelectCount = Math.max(selectCount, nameSelectCount);

        if (globalSelectCount >= 0)
        {
            if (selectCount >= 0)
                globalSelectCount = Math.min(globalSelectCount, selectCount);

            if (nameSelectCount >= 0)
                globalSelectCount = Math.min(globalSelectCount, nameSelectCount);
        }
        // if we have a limit on the number of object to show, always allow objects value in domain
        if ((globalSelectCount >= 0)
                && set.contains(relation))
        {
            if ((set.getObject() != null) || (set.getModel() != null))
            {
                LyObject object = set.getObject();

                if (object == null)
                    object = set.getModel();

                if (!set.isChangingObject())
                {
                    // FFT Structis 146
                    // Keep the previously selected value if any.
                    LyRelationValue value = (LyRelationValue)set.getFieldValue(relation);
                    LyObjectList    list = ((value == null) ? null : value.getObjectList());

                    if (list != null)
                        currentSelection = list;
                    else
                        currentSelection = relation.getObjects(object);
                }
                else
                    currentSelection = relation.getObjects(object);
            }
            else
            {
                // If a default value has been set on the relation, return the object
                // associated to default value.
                Object defaultValue = relation.getDefaultValue();

                if (defaultValue instanceof LyObjectId)
                {
                    LyObjectId objectId = (LyObjectId)defaultValue;

                    LyObject object = objectId.getLogicalObject();

                    if (object != null)
                        currentSelection = new LySimpleObjectList(object);
                }
            }
        }

        if (set.contains(relation)
                && ((context == null) || relation.ignoresContext())
                && !relation.isToolbarDisabled())
        {
            if (globalSelectCount >= 0)
            {
                LyObjectSource source;

                for (i = 0; (i < n) && (count < globalSelectCount); i++)
                {
                    linkClass = linkClasses.getInfo(i);
                    // The source can be null when building a filter from studio.
                    source = session.getObjectSource(linkClass);
                    if (source != null)
                    {
                        filter = getRelationFilter(set, relation, linkClass);
                        if ((count + source.getSize(filter, true)) > globalSelectCount)
                            count = globalSelectCount;
                        else
                            count += source.getSize(filter);
                    }
                }

                if (count >= globalSelectCount)
                    return currentSelection;
            }
            else if (!useLocalCompletion(set, relation) && (completionMax != 0))
                return null;
        }

        LyObjectSource source;

        for (i = 0; i < n; i++)
        {
            linkClass = linkClasses.getInfo(i);
            source = session.getObjectSource(linkClass);
            if (source == null)
                continue;

            filter = getRelationFilter(set, relation, linkClass);
            if (filter != null)
                filtered = true;

            if (context != null)
            {
                // Get all links from the context
                try
                {
                    classLinks = context.getList(session, linkClass);

                    // FFT00497
                    if ((filter != null) && (classLinks != null))
                        classLinks = new LySubWorkSpace(session, classLinks, linkClass, filter, false);
                }
                catch (IllegalStateException e)
                {
                    // No route to context : show all
                    // Get all objects
                    classLinks = new LyClassWorkSpace(session, linkClass, new LyInfoList<LyFieldInfo>(), filter, null,
                            false, true);
                }

                if (classLinks == null)
                    // No route to context : show all
                    // Get all objects
                    classLinks = new LyClassWorkSpace(session, linkClass, new LyInfoList<LyFieldInfo>(), filter, null,
                            false, true);
            }
            else
                // Get all objects
                classLinks = new LyClassWorkSpace(session, linkClass, new LyInfoList<LyFieldInfo>(), filter, null,
                        false, true);

            if (links == null)
                links = classLinks;
            else
                links.addList(classLinks);
        }

        // use of enableRelationLink
        if ((links != null) && (links.getSize() > 0))
        {
            LySimpleObjectList enabledLinks = new LySimpleObjectList();
            LyObject           link;

            for (i = 0; i < links.getSize(); i++)
            {
                link = links.getObject(i);
                if (enableRelationLink(set, relation, link))
                    enabledLinks.addObject(link);
            }

            links = enabledLinks;
        }

        // Add extra objects
        if (set.contains(relation))
        {
            LyObjectList list = null;

            if (!set.isCreation() && filtered)
            {
                // On object set, always allow objects value in domain
                LyObject object = set.getObject();

                list = relation.getObjects(object);
            }
            else if (set.isCreation()
            	&& (set.getModel() == null)) // Not cloning
            {
                Object value = getValue(set, relation);

                if (value != null)
                    list = relation.getObjects(value);
            }

            if (list != null)
            {
                if (links == null)
                    links = list;
                else
                {
                    if (links.isDynamic())
                    {
                        LyObjectList oldLinks = links;

                        links = new LySimpleObjectList(links);
                        ((LyWorkSpace)oldLinks).free(null);
                    }

                    links.addList(list);
                }
            }
        }

        if (!relation.isToolbarDisabled())
        {
            if ((links != null)
                    && (links.getSize() >= globalSelectCount)
                    && (globalSelectCount >= 0))
                return currentSelection;
        }

        if (links != null)
            links.setSortByName(true);

        return links;
    } // end method getRelationDomain

    /**
     * Gets relation domain (list of available links in the filter context) for the given relation
     * fieldInfo for filter actions.
     *
     * @param   filter    the current LyFilterController instance
     * @param   relation  the target relation field
     *
     * @return  a list of LyObjects available for this relation in filtering context.
     */
    public LyObjectList getFilterRelationDomain(LyFilterController filter, LyRelationFieldInfo relation)
    {
        LyInfoList<LyClassInfo> linkClasses = getFilterRelationClasses(filter, relation);
        int                     i, n = ((linkClasses == null) ? 0 : linkClasses.getSize());
        LyClassInfo             linkClass;
        LySession               session = filter.getSession();
        LyObjectList            links = null, classLinks;

        if (!relation.isToolbarDisabled())
        {
            LyEnvironment environment = filter.getEnvironment();
            String        selectCountStr = environment.getResource("LY_SELECT_RELATION_COUNT");
            int           selectCount = LyTools.intFromString(selectCountStr), count = 0;
            String        nameSelectCountStr = environment.getResource("LY_SELECT_BY_NAME_RELATION_COUNT");
            int           nameSelectCount = LyTools.intFromString(nameSelectCountStr);
            int           globalSelectCount = -1;

            // Retrieve the minimum positive of both variables
            globalSelectCount = Math.max(selectCount, nameSelectCount);

            if (globalSelectCount >= 0)
            {
                if (selectCount >= 0)
                    globalSelectCount = Math.min(globalSelectCount, selectCount);

                if (nameSelectCount >= 0)
                    globalSelectCount = Math.min(globalSelectCount, nameSelectCount);
            }

            if (globalSelectCount >= 0)
            {
                LyObjectSource source;
                LyFilter       relationFilter;

                for (i = 0; (i < n) && (count < globalSelectCount); i++)
                {
                    linkClass = linkClasses.getInfo(i);
                    source = session.getObjectSource(linkClass);
                    if (source == null)
                        continue;

                    relationFilter = relation.getFilter(linkClass);
                    if ((count + source.getSize(relationFilter, true)) > globalSelectCount)
                        count = globalSelectCount;
                    else
                        count = source.getSize(relationFilter);

                    if (count < globalSelectCount)
                    {
                        classLinks = new LyClassWorkSpace(session, linkClass, new LyInfoList<LyFieldInfo>(), null, null,
                                false, true);
                        if (links == null)
                            links = classLinks;
                        else
                            links.addList(classLinks);
                    }
                }

                // ensure current values are added
                LyFilter currentFilter = filter.getDefaultFilter();

                if (currentFilter != null)
                {
                    // Always add current values
                    Vector<Object> values = new Vector<Object>();
                    Object         relationValue;
                    LyObjectList   tempList = null;

                    currentFilter.getValues(relation, values);
                    for (i = 0, n = values.size(); i < n; i++)
                    {
                        relationValue = values.elementAt(i);
                        // tempList = relation.getObjects(relationValue);
                        if (tempList == null)
                            tempList = relation.getObjects(relationValue);
                        else
                            tempList.addList(relation.getObjects(relationValue));
                    }
                    // add selected values
                    if (tempList != null)
                        if (links != null)
                            links.addList(tempList);
                        else
                            links = tempList;
                }

                return links;
            }
        }

        n = ((linkClasses == null) ? 0 : linkClasses.getSize());
        for (i = 0; i < n; i++)
        {
            linkClass = linkClasses.getInfo(i);

            // Get all objects
            classLinks = new LyClassWorkSpace(session, linkClass, new LyInfoList<LyFieldInfo>(), null, null, false,
                    true);
            if ((classLinks == null) || (classLinks.getSize() == 0))
                continue;

            if (links == null)
                links = classLinks;
            else
                links.addList(classLinks);
        }

        return links;
    } // end method getFilterRelationDomain

    /**
     * Gets the list of available class info for the given relation filtering.
     *
     * @param   filter    the current instance of "filter controller" This parameter could be both a
     *                    {@link leon.control.LyFilterController} or a
     *                    {@link leon.control.LyFilterRelationController}. That's why its type is
     *                    {@link leon.control.LyFormController} which is the common inheritance.
     * @param   relation  the target relation field
     *
     * @return  a list of classInfo available for filtering the relation
     */
    public LyInfoList<LyClassInfo> getFilterRelationClasses(LyFormController filter, LyRelationFieldInfo relation)
    {
        LyInfoList<LyClassInfo> linkClasses = new LyInfoList<LyClassInfo>();
        int                     i, n = relation.getClassInfoCount();

        for (i = 0; i < n; i++)
            linkClasses.addInfo(relation.getClassInfo(i));

        return linkClasses;
    }

    /**
     * Gets domain (list of allowed options) of the given choice fieldInfo.
     *
     * @param   filter  the current LyFilterController instance
     * @param   choice  target choice fieldInfo
     *
     * @return  a list of LyChoiceOption allowed for this choice field info
     */
    public Vector<LyChoiceOption> getFilterChoiceDomain(LyFilterController filter, LyChoiceFieldInfo choice)
    {
        short                      i, n = choice.getChoiceOptionCount();
        Vector<LyChoiceOption>     domain = new Vector<LyChoiceOption>(n);
        LyInfoList<LyChoiceOption> options = choice.getChoiceOptions(true);
        LyChoiceOption             option;

        for (i = 0; i < n; i++)
        {
            option = options.getInfo(i);
            domain.addElement(option);
        }

        return domain;
    }

    /**
     * Gets the filter of the given relation and the given class.<br />
     * This method is used by the {@link #getRelationDomain(LySetController, LyRelationFieldInfo)}
     * method when the target objets of the relation need to be filtered. The current implementation
     * return null if the relation is null, the relation filter (
     * {@link LyRelationFieldInfo#getFilter(LyClassInfo) relation.getFilter(classInfo)})
     *
     * @param   controller  the current LySetController instance
     * @param   relation    the relation field info we want to filter
     * @param   classInfo   ClassInfo for which we look for an associated filter.
     *
     * @return  The filter associated to the given classInfo, or null if no filter has been
     *          specified for this classInfo in this relation.
     */
    public LyFilter getRelationFilter(LySetController controller, LyRelationFieldInfo relation, LyClassInfo classInfo)
    {
        if (relation == null)
            return null;

        return relation.getFilter(classInfo);
    }

    /**
     * Allows the given link object as a possible value for the given relation field info.<br />
     * <i>The default implementation</i> only calls
     * {@link #enableRelationLink(LySetController, LyRelationFieldInfo, LyObject) enableRelationLink(set, relation, link)}
     * .
     *
     * @param   set        the current LySetController instance
     * @param   relation   the relation that contains the link
     * @param   link       the link to enabled
     * @param   showError  indicates if an error should be shown if the given link is rejected.
     *
     * @return  true if the link is enabled on this relation, false otherwise
     *
     * @see     #enableRelationLink(LySetController, LyRelationFieldInfo, LyObject)
     */
    public boolean enableRelationLink(LySetController set, LyRelationFieldInfo relation, LyObject link,
        boolean showError)
    {
        return enableRelationLink(set, relation, link);
    }

    /**
     * Allows the given link object as a possible value for the given relation field info.<br />
     * <i>The default implementation</i> always returns true.
     *
     * @param   set       the current LySetController instance
     * @param   relation  the relation that contains the link
     * @param   link      the link to enabled
     *
     * @return  true if the link is enabled on this relation, false otherwise
     */
    public boolean enableRelationLink(LySetController set, LyRelationFieldInfo relation, LyObject link)
    {
        return true;
    }

    /**
     * Enables the given href object on the given relation field info.<br />
     * <i>The default implementation</i> always returns true.
     *
     * @param   consult   the current LyConsultController instance
     * @param   relation  the relation that contains the href
     * @param   href      the href to enabled
     *
     * @return  true if the href is enabled on this relation, false otherwise
     */
    public boolean enableRelationHRef(LyConsultController consult, LyRelationFieldInfo relation, LyObject href)
    {
        return true;
    }

    /**
     * Gets the list of available classInfo of the given relation fieldInfo.
     *
     * @param   set       the current LySetController instance
     * @param   relation  the target relation fieldInfo
     *
     * @return  the list of classInfo available for this relation
     */
    public LyInfoList<LyClassInfo> getRelationClasses(LySetController set, LyRelationFieldInfo relation)
    {
        if (relation == null)
            return null;

        LyClassInfo   classInfo = set.getClassInfo();
        LyApplication application = classInfo.getApplication();

        // Check if the relation allows several classes and if it has a class
        // field
        // that specifies valid classes among all these classes.
        LyInfoList<LyClassInfo> classes = new LyInfoList<LyClassInfo>();
        LyClassInfo             linkClass;
        int                     i;

        if (relation.getClassInfoCount() > 1)
        {
            // Get class field
            LyFieldInfo classField = relation.getClassField();

            if (classField != null)
            {
                LyValue value = set.getFieldValue(classField);
                Object  fieldValue = null;

                if ((value == null) || value.isNull())
                {
                    LyObject object = set.getObject();

                    if (object != null)
                        fieldValue = object.getValue(classField);
                }
                else
                    fieldValue = value.getValue();

                if (fieldValue != null)
                {
                    LyInfoList<LyClassInfo> appClasses = application.getClassInfos();
                    String                  str = classField.toName(fieldValue);

                    if (!classField.isMultiple())
                    {
                        linkClass = appClasses.getInfo(str, true);
                        if ((linkClass != null) && relation.relatedWith(linkClass))
                            classes.addInfo(linkClass);
                    }
                    else
                    {
                        String[] strings = LyString.stringToList(str, application.VALUE_SEP);

                        for (i = 0; i < strings.length; i++)
                        {
                            // Retrieve chosen class
                            linkClass = appClasses.getInfo(strings[i]);
                            if ((linkClass == null) || !relation.relatedWith(linkClass))
                                continue;

                            classes.addInfo(linkClass);
                        }
                    }

                    if (classes.getSize() > 0)
                        return classes;
                }
            }
            else
            {
                LyFieldController fieldController = set.getFieldController(relation);

                if (fieldController instanceof LyCompositionFieldUnitSetController)
                {
                    LyClassInfo fieldType = ((LyCompositionFieldUnitSetController)fieldController).getFieldType();

                    if (fieldType != null)
                    {
                        classes.addInfo(fieldType);
                        return classes;
                    }
                }
            }
        }

        // Return the list of classes specified in the relation fieldInfo
        int n = relation.getClassInfoCount();

        for (i = 0; i < n; i++)
        {
            linkClass = relation.getClassInfo(i);
            classes.addInfo(linkClass);
        }

        return classes;
    } // end method getRelationClasses

    /**
     * Gets the list of objects of the given relation context.
     *
     * @param   set       the current LySetController instance
     * @param   relation  the target relation fieldInfo
     *
     * @return  a list of LyObject available in this relation field
     */
    public LyObjectList getRelationContext(LySetController set, LyRelationFieldInfo relation)
    {
        LyRelationFieldInfo contextField = relation.getContextField();

        // If the relation doesn't have a special context, return current
        // context
        if (contextField == null)
        {
            if ((set.isComposition(relation)) && (set.getObject() != null) && !set.isCreation())
                return new LySimpleObjectList(set.getObject());
            else if (set.getCompositionRelation() != null)
                return null;

            return relation.ignoresContext() ? null : set.getContext();
        }

        // Get context list
        LyValue value = null;

        if (set.contains(contextField))
        {
            value = set.getFieldValue(contextField);

            // If value is empty and our context field is optional
            // try to use the context of the context field
            if ((value != null) && value.isNull() && isOptional(set, contextField))
                return getRelationContext(set, contextField);
        }
        else
        {
            LyObject object = set.getObject();

            if (object != null)
                value = object.getFieldValue(contextField);

            // If value is empty, try to use the context of the context field
            if ((value != null) && value.isNull())
                return getRelationContext(set, contextField);
        }

        if ((value == null) || value.isNull())
            return new LySimpleObjectList();

        LyObjectList result = contextField.getObjects(value.getValue());

        if (result == null)
            result = new LySimpleObjectList();

        return result;
    } // end method getRelationContext

    /**
     * Indicates the maximum number of values displayed in relation components for auto-completion.
     * <br />
     * <i>The default implementation</i> returns the LY_MULTIPLE_RELATION_AUTO_COMPLETION_MAX
     * resource if relation is multiple or the LY_SIMPLE_RELATION_AUTO_COMPLETION_MAX resource if
     * relation is simple.
     *
     * @param   set       the current LySetController instance.
     * @param   relation  the target relation fieldInfo to test.
     *
     * @return  a positive number if auto-completion is activated, 0 if auto-completion is not
     *          activated, a negative value if there is no maximum.
     */
    public int getAutoCompletionMaximum(LySetController set, LyRelationFieldInfo relation)
    {
        if (!relation.hasMark(LyFieldInfo.MARK_AUTO_COMPLETION))
            return 0;

        LyEnvironment environment = set.getEnvironment();

        if (relation.isMultiple())
            return environment.getIntResource("LY_MULTIPLE_RELATION_AUTO_COMPLETION_MAX");
        else
            return environment.getIntResource("LY_SIMPLE_RELATION_AUTO_COMPLETION_MAX");
    }

    /**
     * Indicates if auto-completion on the given relation fieldInfo is managed by the graphical
     * component or by the field controller.
     *
     * @param   set       the current LySetController instance
     * @param   relation  the fieldInfo that may need to be checked
     *
     * @return  true if auto-completion is managed by the graphical component, false otherwise
     */
    public boolean useLocalCompletion(LySetController set, LyRelationFieldInfo relation)
    {
        // Multiple relations always perform auto completion on the graphical side
        if (relation.isMultiple())
            return true;

        int completionMax = getAutoCompletionMaximum(set, relation);

        if (completionMax < 0)
            return true;

        LyObjectList context = getRelationContext(set, relation);

        if (context != null)
            return true;

        if (completionMax != 0)
        {
            LyInfoList<LyFieldInfo> nameFields;
            LyInfoList<LyClassInfo> linkClasses = getRelationClasses(set, relation);
            int                     i, n = ((linkClasses == null) ? 0 : linkClasses.getSize());
            LyClassInfo             linkClass;
			boolean					allClassLoaded = true;
			LySession				session = set.getSession();
			LyObjectSource			objectSource;
			LyCache					cache;

            // Check if auto-completion may be performed on the controller side for this relation.
            // This is allowed only if :
            // - relation has target classes without dynamic labels
            // - relation has target classes with only one field info as the name field info
            // and this field info is a text field info or a relation field info with only
            // one target class info with one field info as the name field info and this
            // field info is a text field
            for (i = 0; i < n; i++)
            {
                linkClass = linkClasses.getInfo(i);
				objectSource = session.getObjectSource(linkClass);
				cache = (objectSource != null ? objectSource.getCache() : null);
				allClassLoaded = allClassLoaded && ((cache != null) && cache.isLoaded());
                nameFields = linkClass.getNameFieldInfos();

                if (linkClass.getLabel(LyInfoLabel.Type.NAME) != null)
                    return true;

                if (nameFields.getSize() != 1)
                    return true;

                LyFieldInfo nameField = nameFields.getInfo(0);

                if (nameField instanceof LyRelationFieldInfo)
                {
                    // If field info is a relation, check if this relation has only one target class
                    // and check that this class has only one text field info as the name field.
                    LyRelationFieldInfo relationField = (LyRelationFieldInfo)nameField;

                    if (relationField.getClassInfoCount() > 1)
                        return true;

                    linkClass = relationField.getClassInfo(0);
                    nameFields = linkClass.getNameFieldInfos();

                    if (linkClass.getLabel(LyInfoLabel.Type.NAME) != null)
                        return true;

                    if (nameFields.getSize() != 1)
                        return true;

                    nameField = nameFields.getInfo(0);

                    if (!(nameField instanceof LyTextFieldInfo))
                        return true;
                }
                else if (!(nameField instanceof LyTextFieldInfo))
                    return true;
            }

			// Optimization : if all classes are loaded, use local completion
			if (allClassLoaded)
				return true;
        }

        return false;
    } // end method useLocalCompletion

	/**
	 * Gets relation domain (list of available links) for the given relation fieldInfo with the
	 * given prefix for related objects.
	 *
	 * @param   set       the current LySetController instance
	 * @param   relation  the target relation field
	 * @param   prefix    the prefix for related objects
	 * @param   fill      indicates whether the domain must be filtered with given prefix or not
	 *
	 * @return  a list of LyObjects available for this relation and given prefix
	 */
	public LyObjectList getRelationDomain(LySetController set, LyRelationFieldInfo relation, String prefix,
			boolean fill)
	{
		LyObjectList domain = getRelationDomain(set, relation);

		// If local completion must be used, return the full domain.
		if (useLocalCompletion(set, relation)
				|| (getAutoCompletionMaximum(set, relation) == 0))
			return domain;

		// If completion may be performed by the controller side, retrieve all the objects from
		// related classes whose names start with given prefix.
		int					    limit = getAutoCompletionMaximum(set, relation);
		LyObjectList		    links = null, classLinksBefore, classLinksAfter;
		LyInfoList<LyClassInfo> linkClasses = getRelationClasses(set, relation);
		LyInfoList<LyFieldInfo> nameFields;
		LyFieldInfo			    nameField;
		int					    i, n = ((linkClasses == null) ? 0 : linkClasses.getSize());
		LyClassInfo			    linkClass;
		LyFilter			    filterBefore = null, filterAfter = null, relationFilter;
		LyFilterExpression	    expression;
		LySort				    sortBefore, sortAfter;
		int					    sizeBefore = 0;
		boolean                 emptyPrefix = ((prefix == null) || (prefix.length() == 0));

		for (i = 0; i < n; i++)
		{
			linkClass = linkClasses.getInfo(i);
			nameFields = linkClass.getNameFieldInfos();

			// We assume that related class only has one name field info.
			nameField = nameFields.getInfo(0);
			filterBefore = null;
			filterAfter = null;

			if (!emptyPrefix)
			{
				filterBefore = new LyFilter(LyFilter.Type.AND);
				filterAfter = new LyFilter(LyFilter.Type.AND);

				// Build filters
				if (nameField instanceof LyRelationFieldInfo)
				{
					// If field info is a relation, build a filter on the relation
					LyRelationFieldInfo     relationField = (LyRelationFieldInfo)nameField;
					LyClassInfo             targetClass = relationField.getClassInfo(0);
					LyInfoList<LyFieldInfo> targetFields = targetClass.getNameFieldInfos();
					LyFilterRelation        filterRelation;
					LyFilter                relatedFilter;
					LyFieldInfo             targetNameField = targetFields.getInfo(0);

					if (fill)
					{
						expression = new LyFilterExpression(targetNameField,
								LyFilterExpression.Condition.LOWER_OR_EQUALS, targetNameField.processValue(prefix));
						filterRelation = new LyFilterRelation(relationField);
						relatedFilter = new LyFilter();
						relatedFilter.addElement(expression);
						filterRelation.setFilter(relatedFilter);
						filterBefore.addElement(filterRelation);
					}

					expression = new LyFilterExpression(targetNameField, LyFilterExpression.Condition.GREATER_OR_EQUALS,
							targetNameField.processValue(prefix));
					filterRelation = new LyFilterRelation(relationField);
					relatedFilter = new LyFilter();
					relatedFilter.addElement(expression);
					filterRelation.setFilter(relatedFilter);
					filterAfter.addElement(filterRelation);
				}
				else
				{
					if (fill)
					{
						expression = new LyFilterExpression(nameField, LyFilterExpression.Condition.LOWER_OR_EQUALS,
								nameField.processValue(prefix));
						filterBefore.addElement(expression);
					}

					expression = new LyFilterExpression(nameField, LyFilterExpression.Condition.GREATER_OR_EQUALS,
							nameField.processValue(prefix));
					filterAfter.addElement(expression);
				}
			}

			if (domain == null)
			{
				// If domain is null (general case when auto-completion is not local), build a page workspace
				// with filter made of prefix filter and relation filter.
				relationFilter = getRelationFilter(set, relation, linkClass);
				if (!emptyPrefix)
				{
					if (fill)
					{
						sortBefore = new LySort();
						sortBefore.addCriterion(nameField, false);

						if (relationFilter != null)
							filterBefore.addElement(relationFilter);

						classLinksBefore = new LyPageWorkSpace(set.getSession(), linkClass, null, filterBefore,
								sortBefore, limit, false, false);
					}
					else
						classLinksBefore = null;

					if (relationFilter != null)
						filterAfter.addElement(relationFilter);
				}
				else
				{
					classLinksBefore = null;
					if (relationFilter != null)
					{
						if (filterAfter == null)
							filterAfter = new LyFilter(LyFilter.Type.AND);
						filterAfter.addElement(relationFilter);
					}
				}

				sortAfter = new LySort();
				sortAfter.addCriterion(nameField, true);

				classLinksAfter = new LyPageWorkSpace(set.getSession(), linkClass, null, filterAfter, sortAfter, limit,
						false, false);
			}
			else
			{
				// If a domain is given, filter this domain with the filter built from given prefix.
				if (fill && (prefix != null))
					classLinksBefore = new LySubWorkSpace(set.getSession(), domain, linkClass, filterBefore, false);
				else
					classLinksBefore = null;

				classLinksAfter = new LySubWorkSpace(set.getSession(), domain, linkClass, filterAfter, false);
			}

			if (links == null)
				links = new LyMultiWorkSpace(set.getSession(), false);

			if (classLinksBefore != null)
			{
				links.addList(classLinksBefore);
				sizeBefore += classLinksBefore.getSize();
			}

			links.addList(classLinksAfter);
		}

		LySimpleObjectList  result = null;

		if (links != null)
		{
			// Sort the list by name and only keep the first objects (using the limit)
			links.setSortByName(true);
			n = links.getSize();

			// If no more objects than given limit, return the links list.
			if ((limit >= 0) && (limit < n))
			{
				int		min = (sizeBefore / 2), max = Math.min(min + limit, n);

                if ((max - min) < limit)
					min = Math.max(0, max - limit);

				result = new LySimpleObjectList();
				result.setSortByName(true);

				for (i = min; i < max; i++)
					result.addObject(links.getObject(i));
			}
		}

		// Add extra objects if needed
		Object        value = getValue(set, relation);

		if (value != null)
		{
			LyObjectList  list = relation.getObjects(value);

			if (list != null)
			{
				if (result == null)
				{
					result = new LySimpleObjectList();
					result.setSortByName(true);
					if (links != null)
						result.addList(links);
				}

				result.addList(list);
			}
		}

		if (result == null)
			return links;

		return result;
	} // end method getRelationDomain

    /**
     * Gets the short label that must be used for the given relation fieldInfo. <i>The default
     * implementation</i> returns null.
     *
     * @param   set       the current LySetController instance
     * @param   relation  the target relation field
     * @param   link      the object represented by the link
     *
     * @return  the short name of the link (displayed when selected) or null if no short label must
     *          be used
     */
    public String getLinkShortLabel(LySetController set, LyRelationFieldInfo relation, LyObject link)
    {
        return null;
    }

    /**
     * Gets actions enabled by the given field. A list of pairs (action, classInfo) has to be
     * returned.<br />
     * <i>The default implementation</i> returns null if the given field is a not relation fied
     * info, the
     * {@link #getTransverseActions(LySetController, LyRelationFieldInfo) getTransverseActions(set, (LyRelationFieldInfo) fieldInfo)}
     * result otherwise
     *
     * @param   set        the current LySetController instance
     * @param   fieldInfo  the field that may enabled some actions
     *
     * @return  the list of actions enabled by the given field
     */
    public LyInfoDataList<LyAction, LyClassInfo> getTransverseActions(LySetController set, LyFieldInfo fieldInfo)
    {
        if (fieldInfo instanceof LyRelationFieldInfo)
            return getTransverseActions(set, (LyRelationFieldInfo)fieldInfo);
        else
            return null;
    }

    /**
     * Gets all actions enabled by the given relation. A list of pairs (action, classInfo) has to be
     * returned.
     *
     * @param   set       the current LySetController instance
     * @param   relation  the relation that may enabled some actions
     *
     * @return  a list of actions that are enabled by the given relation
     */
    public LyInfoDataList<LyAction, LyClassInfo> getTransverseActions(LySetController set, LyRelationFieldInfo relation)
    {
        LyApplication                         application = set.getApplication();
        LySession                             session = set.getSession();
        LyInfoDataList<LyAction, LyClassInfo> actions = null;
        LyClassInfo                           classInfo, classInfo2;
        LyObjectSource                        source;
        LyAction                              action;
        LyFilter                              filter;
        int                                   j, na;
        LyEnvironment                         environment = set.getEnvironment();
        String                                selectCountStr = environment.getResource("LY_SELECT_RELATION_COUNT");
        int                                   selectCount = LyTools.intFromString(selectCountStr);
        String                                nameSelectCountStr = environment.getResource(
                "LY_SELECT_BY_NAME_RELATION_COUNT");
        int                                   nameSelectCount = LyTools.intFromString(nameSelectCountStr);
        int                                   count = -1, nameCount = -1;
        LyInfoList<LyClassInfo>               classes = getRelationClasses(set, relation);
        int                                   i, nc = ((classes == null) ? 0 : classes.getSize());

        // See FFT02114 in LyFormfieldBuilder. Skipp adding bar for backward
        // compatibility
        LyFieldController fieldController = set.getFieldController(relation);

        if ((fieldController != null)
            && (fieldController.getAreaController() != null))
        {
            String needToolBar = (String)relation.getData("LY_NEED_TOOLBAR");

            if ((needToolBar == null)
                || ((!"true".equals(needToolBar))
                    && (!(("create".equals(needToolBar)
                        && (fieldController.getAreaController() instanceof LyCreateController))))
                    && (!(("set".equals(needToolBar)
                        && (fieldController.getAreaController() instanceof LySetController)
                        && !(fieldController.getAreaController() instanceof LyCreateController))))
                    && (!(("table".equals(needToolBar)
                        && (fieldController.getAreaController() instanceof LyTableController))))))
                return null;
        }

        // No select tool for a struct.
        if (relation instanceof LyStructFieldInfo)
        {
            selectCount = -1;
            nameSelectCount = -1;
        }

        // If auto-completion is used for given relation, do not use select by name action.
        int completionMax = getAutoCompletionMaximum(set, relation);

        if (completionMax != 0)
            nameSelectCount = -1;

        for (i = 0; i < nc; i++)
        {
            classInfo = classes.getInfo(i);

            // Add select action if the classInfo has many instances
            if ((selectCount >= 0) && (count < selectCount))
            {
                if (count < 0)
                    count = 0;

                if (selectCount > 0)
                {
                    filter = getRelationFilter(set, relation, classInfo);

                    // The source can be null for a preview action in studio.
                    source = session.getObjectSource(classInfo);
                    if (source != null)
                    {
                        if ((count + source.getSize(filter, true)) > selectCount)
                            count = selectCount;
                        else
                            count += source.getSize(filter);
                    }
                }

                if (count >= selectCount)
                {
                    for (j = 0; j < nc; j++)
                    {
                        classInfo2 = classes.getInfo(j);
                        action = application.getActionWithId(LyAction.ID_SELECT, classInfo2);
                        if (actions == null)
                            actions = new LyInfoDataList<LyAction, LyClassInfo>();

                        actions.addInfo(action, classInfo2);
                    }
                }
            }

            // Add select action if the classInfo has many instances
            if ((nameSelectCount >= 0) && (nameCount < nameSelectCount))
            {
                if (nameCount < 0)
                    nameCount = 0;

                if (nameSelectCount > 0)
                {
                    filter = getRelationFilter(set, relation, classInfo);

                    // The source can be null for a preview action in studio.
                    source = session.getObjectSource(classInfo);
                    if (source != null)
                    {
                        if ((nameCount + source.getSize(filter, true)) > nameSelectCount)
                            nameCount = nameSelectCount;
                        else
                            nameCount += source.getSize(filter);
                    }
                }

                if (nameCount >= nameSelectCount)
                {
                    for (j = 0; j < nc; j++)
                    {
                        classInfo2 = classes.getInfo(j);
                        action = application.getActionWithId(LyAction.ID_SELECT_BY_NAME, classInfo2);
                        if (actions == null)
                            actions = new LyInfoDataList<LyAction, LyClassInfo>();

                        actions.addInfo(action, classInfo2);
                    }
                }
            }

            // Add transverse actions of the classInfo
            if (classInfo.isAbstract())
            {
                LyInfoList<LyClassInfo> subClasses = classInfo.getSubClasses(true);

                if (subClasses == null)
                    continue;

                int                          n = subClasses.getSize();
                Vector<LyInfoList<LyAction>> actionsList = new Vector<LyInfoList<LyAction>>(n);
                LyInfoList<LyAction>         list;
                LyClassInfo                  subClass;
                int                          max = 0;

                // Get all actions for all subclasses.
                for (j = 0; j < n; j++)
                {
                    subClass = subClasses.getInfo(j);
                    list = subClass.getActions(LyAction.MARK_TRANSVERSE, false);
                    actionsList.add(list);
                    if (list == null)
                        continue;

                    if (list.getSize() > max)
                        max = list.getSize();
                }

                // Put all the actions in a LyInfoDataList.
                for (j = 0; j < max; j++)
                {
                    for (int k = 0; k < n; k++)
                    {
                        list = actionsList.elementAt(k);
                        if (list == null)
                            continue;

                        if (list.getSize() > j)
                        {
                            action = list.getInfo(j);

                            // At last, action must validated and enabled
                            if (session.validateAction(action, set))
                            {
                                classInfo2 = subClasses.getInfo(k);
                                if (actions == null)
                                    actions = new LyInfoDataList<LyAction, LyClassInfo>();

                                actions.addInfo(action, classInfo2);
                            }
                        }
                    }
                }
            }
            else
            {
                na = classInfo.getActionCount();

                for (j = 0; j < na; j++)
                {
                    action = classInfo.getAction(j);

                    // Action must have transverse mark
                    if (!action.hasMark(LyAction.MARK_TRANSVERSE))
                        continue;

                    // At last, action must validated and enabled
                    if (session.validateAction(action, set))
                    {
                        if (actions == null)
                            actions = new LyInfoDataList<LyAction, LyClassInfo>();

                        actions.addInfo(action, classInfo);
                    }
                }
            }
        }

        return actions;
    } // end method getTransverseActions

    /**
     * Gets all actions enabled by the given field for a filter action. A list of pairs (action,
     * classInfo) has to be returned.
     *
     * @param   filter     the current LyFilterController instance
     * @param   fieldInfo  the fieldInfo that may enabled some actions
     *
     * @return  a list of actions enabled by the given relation for a filter action
     */
    public LyInfoDataList<LyAction, LyClassInfo> getFilterTransverseActions(LyFilterController filter,
        LyFieldInfo fieldInfo)
    {
        if (fieldInfo instanceof LyRelationFieldInfo)
            return getFilterTransverseActions(filter, (LyRelationFieldInfo)fieldInfo);

        return null;
    }

    /**
     * Gets all actions enabled by the given relation for a filter action. A list of pairs (action,
     * classInfo) has to be returned.
     *
     * @param   filter    the current LyFilterController instance
     * @param   relation  the relation that may enabled some actions
     *
     * @return  a list of actions enabled by the given relation for a filter action
     */
    public LyInfoDataList<LyAction, LyClassInfo> getFilterTransverseActions(LyFilterController filter,
        LyRelationFieldInfo relation)
    {
        LyEnvironment environment = filter.getEnvironment();
        String        selectCountStr = environment.getResource("LY_SELECT_RELATION_COUNT");
        int           selectCount = LyTools.intFromString(selectCountStr);
        String        nameSelectCountStr = environment.getResource("LY_SELECT_BY_NAME_RELATION_COUNT");
        int           nameSelectCount = LyTools.intFromString(nameSelectCountStr);

        if ((selectCount < 0) && (nameSelectCount < 0))
            return null;

        LyInfoList<LyClassInfo> classes = getFilterRelationClasses(filter, relation);

        if (classes == null)
            return null;

        int                                   i, n = classes.getSize();
        LyApplication                         application = filter.getApplication();
        LySession                             session = filter.getSession();
        LyInfoDataList<LyAction, LyClassInfo> actions = null;
        LyClassInfo                           classInfo;
        LyObjectSource                        source;
        LyAction                              action;
        LyFilter                              relationFilter;

        for (i = 0; i < n; i++)
        {
            classInfo = classes.getInfo(i);

            // Add select action if the classInfo has many instances
            relationFilter = relation.getFilter(classInfo);

            // The source can be null when building a filter from studio.
            source = session.getObjectSource(classInfo);
            if ((source != null)
                    && (selectCount >= 0)
                    && ((source.getSize(relationFilter, true) >= selectCount)
                        || (source.getSize(relationFilter) >= selectCount)))
            {
                action = application.getActionWithId(LyAction.ID_SELECT, classInfo);
                if (actions == null)
                    actions = new LyInfoDataList<LyAction, LyClassInfo>();

                actions.addInfo(action, classInfo);
            }

            if ((source != null)
                    && (nameSelectCount >= 0)
                    && ((source.getSize(relationFilter, true) >= nameSelectCount)
                        || (source.getSize(relationFilter) >= nameSelectCount)))
            {
                action = application.getActionWithId(LyAction.ID_SELECT_BY_NAME, classInfo);
                if (actions == null)
                    actions = new LyInfoDataList<LyAction, LyClassInfo>();

                actions.addInfo(action, classInfo);
            }
        }

        return actions;
    } // end method getFilterTransverseActions

    /**
     * Indicates if the given fieldInfo needs to be checked when the value is modified (in creation
     * and set mode).<br />
     * <i>The default implementation</i> returns true if the given field has the mark NEED_POST or
     * has dependencies.
     *
     * @param   set        the current LySetController instance
     * @param   fieldInfo  the fieldInfo that may need to be checked
     *
     * @return  true if the field need to be checked, false otherwise
     */
    public boolean needCheck(LySetController set, LyFieldInfo fieldInfo)
    {
        // For an field contained in an array, post the data only if needed.
        if ((fieldInfo != null) && (fieldInfo.getParent() instanceof LyArrayFieldInfo))
        {
            // Return true if the field has the mark NEED_POST
            if (fieldInfo.hasMark(LyFieldInfo.MARK_NEED_POST))
                return true;

            // Or if the field has dependencies
            LyInfoList<LyFieldInfo> depends = fieldInfo.getDepends();

            return ((depends != null) && (depends.getSize() > 0));
        }

        return true;
    }

    /**
     * Progates the given field value to all fields.<br />
     * <i>The default implementation</i> gets the field associated to the value and propagates it on
     * all fields in the list of dependencies and that are managed by the given set controller.
     *
     * @param  set       the current LySetController instance
     * @param  newValue  new field info value to propagate.
     */
    public void propagate(LySetController set, LyFieldInfoValue newValue)
    {
        if (newValue == null)
            return;

        propagate(set, newValue.getFieldInfo(), newValue.getValue());
    }

    /**
     * Progates the given value on all fields in its the list of dependencies of the given one and
     * that are managed by the given set controller.
     *
     * @param       set        the current LySetController instance
     * @param       fieldInfo  field for which a change is propagated.
     * @param       newValue   value to propagate
     *
     * @deprecated  Use {@link #propagate(LySetController, LyFieldInfoValue)}
     */
    protected void propagate(LySetController set, LyFieldInfo fieldInfo, Object newValue)
    {
        if (set == null)
            return;

        LyInfoList<LyFieldInfo> depends = fieldInfo.getDepends();
        int                     i, n = ((depends == null) ? 0 : depends.getSize());
        LyFieldInfo             depend;

        for (i = 0; i < n; i++)
        {
            depend = depends.getInfo(i);
            if (set.contains(depend))
                set.update(depend, fieldInfo, true);
        }
    }

    /**
     * This method controls the given value before creating or modifying an object. It does not
     * checks its content but the coherence between it and its field marks.<br />
     * <i>The default implementation</i> checks that:
     *
     * <ul>
     * <li>if the field is not optional, its value is not null</li>
     * <li>the value has changed</li>
     * <li>if a field has the mark 'unique', it is really unique</li>
     * </ul>
     *
     * @param   set       the current LySetController instance
     * @param   newValue  field info value to check. This value cannot be null.
     *
     * @return  a code to indicates the control result. Possible values are:
     *
     *          <ul>
     *          <LI>{@link #STATUS_IGNORE} if the value has not change</LI>
     *          <LI>{@link #STATUS_OK} if the value is correct</LI>
     *          <LI>{@link #STATUS_KO} if the value is not correct</LI>
     *          <LI>{@link #STATUS_WARNING} if the value is on warning</LI>
     *          </ul>
     */
    public short controlNewValue(LySetController set, LyFieldInfoValue newValue)
    {
        if (newValue == null)
            return STATUS_KO;

        return controlNewValue(set, newValue.getFieldInfo(), newValue.getValue());
    }

    /**
     * This method controls the given value before creating or modifying an object. It does not
     * checks its content but the coherence between it and its field marks.<br />
     * <i>The default implementation</i> checks that:
     *
     * <ul>
     * <li>if the field is not optional, its value is not null</li>
     * <li>the value has changed</li>
     * <li>if a field has the mark 'unique', it is really is unique</li>
     * </ul>
     *
     * @param       set        the current LySetController instance
     * @param       fieldInfo  fieldInfo for which the value is checked.
     * @param       newValue   new value for the fieldInfo
     *
     * @return      a code to indicates the control result. Possible values are:
     *
     *              <ul>
     *              <LI>{@link #STATUS_IGNORE} if the value has not change</LI>
     *              <LI>{@link #STATUS_OK} if the value is correct</LI>
     *              <LI>{@link #STATUS_KO} if the value is not correct</LI>
     *              <LI>{@link #STATUS_WARNING} if the value is on warning</LI>
     *              </ul>
     *
     * @deprecated  Use {@link #controlNewValue(LySetController, LyFieldInfoValue)}
     */
    protected short controlNewValue(LySetController set, LyFieldInfo fieldInfo, Object newValue)
    {
        // Check if this fieldInfo is optional or not
        // This must be controlled first because it can change even when the
        // value is unchanged.
        boolean optional = set.isOptional(fieldInfo);

        if (!optional)
        {
            boolean emptyOrNull = false;

            if (newValue == null)
                emptyOrNull = true;
            else if ((fieldInfo instanceof LyTextFieldInfo)
                    && (((String)newValue).trim().length() == 0))
                emptyOrNull = true;

            if (emptyOrNull)
            {
                // Error : this fieldInfo doesn't allow a null value
                LyError  error;
                LyObject object = set.getObject();
                Object[] args = new Object[1];

                args[0] = fieldInfo.getName();

                if ("_null".equals(args[0]))
                    args[0] = fieldInfo.getAlias();

                if (!"_null".equals(args[0]))
                {
                    if ((fieldInfo instanceof LyRelationFieldInfo)
                            && fieldInfo.hasMark(LyFieldInfo.MARK_AUTO_COMPLETION))
                        error = new LyError(LyObject.LEVEL_ERROR, "LY_ER_FIELD_INVALID_RELATION_VALUE", args, fieldInfo,
                                object, fieldInfo);
                    else
                        error = new LyError(LyObject.LEVEL_ERROR, "LY_ER_FIELD_MANDATORY", args, fieldInfo, object, fieldInfo);
	                set.showError(object, error);
                }

                return STATUS_KO;
            }
        }

        // Compare with old value for object set
        LyObject object = set.getObject();
        boolean  creation = set.isCreation();

        if (!creation
                && (object != null))
        {
            Object value = object.getValue(fieldInfo);

            if (!object.hasNewValue(fieldInfo) && fieldInfo.equals(value, newValue))
            {
                // Control for type change
                LyAreaController areaController = set.getAreaController(fieldInfo);

                if ((areaController == null)
                        || (areaController.getAction().getRoot().getId() != LyAction.ID_CREATE))
                    // No attribute modification
                    return STATUS_IGNORE;
            }
            else if ((value instanceof String) && fieldInfo.hasMark(LyFieldInfo.MARK_SECRET))
            {
                // Specific case for encrypted password in Web mode
                int    c, nc = ((String)value).length();
                String stringValue = "";

                for (c = 0; c < nc; c++)
                    stringValue += '*';

                if (fieldInfo.equals(stringValue, newValue))
                    // No attribute modification
                    return STATUS_IGNORE;
            }
        }

        // Unicity control
        if (fieldInfo.hasMark(LyFieldInfo.MARK_UNIQUE) && (!optional || (newValue != null)))
        {
            // Retrieve class info in which the field info was defined
            LyClassInfo fieldClassInfo = set.getClassInfo();

            while (fieldClassInfo.isInherited(fieldInfo))
                fieldClassInfo = fieldClassInfo.getSuperClass();

            // Check unicity
            LySession          session = set.getSession();
            LyObjectSource     source = session.getObjectSource(fieldClassInfo);
            LyFilter           filter = new LyFilter(LyFilter.Type.AND);
            LyFilterExpression filterExp = new LyFilterExpression(fieldInfo, false, LyFilterExpression.Condition.EQUALS,
                    newValue);
            LyResponse         response;

            filter.addElement(filterExp);
            response = source.getObjects(filter);

            if ((response.getSize() > 1)
                    || ((response.getSize() == 1) && !response.contains(object)))
            {
                // Attribute authorize only one value
                // Add an error to the object or show the message.
                LyError  error;
                Object[] args = new Object[1];
                LyObject cause = null;
                int      i, n = response.getSize();

                for (i = 0; i < n; i++)
                {
                    cause = (LyObject)response.getResponse(i);
                    if (cause != object)
                        break;
                }

                args[0] = fieldInfo.getName();
                error = new LyError(LyObject.LEVEL_ERROR, "LY_ER_UNIQUE", args, fieldInfo, cause, fieldInfo);
                set.showError(object, error);
                return STATUS_KO;
            }
        }

        // Check value
        LyFieldInfoValue newFieldInfoValue = (LyFieldInfoValue)fieldInfo.buildFieldValue(newValue);
        short            result = checkNewValue(set, newFieldInfoValue);

        if (result == STATUS_KO)
            return result;

        // Control value
        if (newValue != null)
        {
            String errorMessage = fieldInfo.controlValue(set.getEnvironment(), newValue);

            if (errorMessage != null)
            {
                LyError error;

                error = new LyError(LyObject.LEVEL_ERROR, errorMessage, null, fieldInfo, object, fieldInfo);
                set.showError(object, error);
                return STATUS_KO;
            }
        }

        // Return final status
        if (newValue == null)
        {
            // in case of Struct, no created object, no control.
            LyAreaController areaController = set.getAreaController(fieldInfo);

            if (!creation && (areaController != null)
                    && (areaController.getAction().getRoot().getId().equals(LyAction.ID_CREATE)))
                return STATUS_IGNORE;

            return (creation ? STATUS_IGNORE : STATUS_OK);
        }

        return result;
    } // end method controlNewValue

    /**
     * Dynamicaly checks the given value for its associated field. This method is called when there
     * is a modification of the field in the user interface (Value changed, focus lost...).<br />
     * <i>The default implementation</i> call the
     * {@link leon.info.LyFieldInfo#checkValue(LyEnvironment, Object) checkValue} method of the
     * field associated to the new value.
     *
     * @param   set       the current LySetController instance
     * @param   newValue  field info value to check.
     *
     * @return  a code to indicates the check result. Possible values are:
     *
     *          <ul>
     *          <LI>{@link #STATUS_OK} if the value is correct</LI>
     *          <LI>{@link #STATUS_KO} if the value is not correct</LI>
     *          </ul>
     */
    public short checkNewValue(LySetController set, LyFieldInfoValue newValue)
    {
        if (newValue == null)
            return STATUS_KO;


        return checkNewValue(set, newValue.getFieldInfo(), newValue.getValue());
    }

    /**
     * Dynamic check the value for the given field. <i>The default implementation</i> call the
     * checkValue method of the given field. This method is called when there is a modification of
     * the field in the user interface (Value changed, focus lost...).
     *
     * @param       set        the current LySetController instance
     * @param       fieldInfo  fieldInfo for which the value is checked.
     * @param       newValue   value to check for the given fieldInfo. This value cannot be null.
     *
     * @return      a code to indicates the check result.Possible values are:
     *
     *              <ul>
     *              <LI>{@link #STATUS_OK} if the value is correct</LI>
     *              <LI>{@link #STATUS_KO} if the value is not correct</LI>
     *              </ul>
     *
     * @deprecated  Use {@link #checkNewValue(LySetController, LyFieldInfoValue)}
     */
    protected short checkNewValue(LySetController set, LyFieldInfo fieldInfo, Object newValue)
    {
        if (newValue == null)
            return STATUS_OK;


        String errorMessage = fieldInfo.checkValue(set.getEnvironment(), newValue);

        if ((errorMessage == null)
                && (fieldInfo instanceof LyRelationFieldInfo))
        {
            if (newValue instanceof LyObjectId)
            {
                // For a relation field info check if object id match a real object
                LyObjectId oid = (LyObjectId)newValue;

                if (oid.getLogicalObject() == null)
                {
                    LyEnvironment environment = set.getEnvironment();

                    errorMessage = environment.getMessage("LY_ER_FIELD_SET_VALUE",
                    		environment.translate(fieldInfo.getName()), oid.toName());
                }
            }
            else if (newValue instanceof LyObjectId[])
            {
                LyObjectId[] oids = (LyObjectId[])newValue;
                int          i, n = oids.length;

                for (i = 0; i < n; i++)
                {
                    // For a relation field info check if object id match a real object
                    if (oids[i].getLogicalObject() == null)
                    {
                        LyEnvironment environment = set.getEnvironment();

                        errorMessage = environment.getMessage("LY_ER_FIELD_SET_VALUE",
                        		environment.translate(fieldInfo.getName()), oids[i].toName());
                        break;
                    }
                }
            }
        }

        if (errorMessage != null)
        {
            LyError  error;
            LyObject object = set.getObject();

            error = new LyError(LyObject.LEVEL_ERROR, errorMessage, null, fieldInfo, object, fieldInfo);
            set.showError(object, error);
            return STATUS_KO;
        }

        return STATUS_OK;
    } // end method checkNewValue

    /**
     * Global control of values before creating or modifying an object. If needed, extra values may
     * be added during this control.<br />
     * <i>The default implementation</i> controls:
     *
     * <ul>
     * <LI>that the new values list is not null or empty,</LI>
     * <LI>that the object associated to the set controller is not null and that the control on its
     * class info ({@link leon.data.LyObject#callClassControls(LyValueSet)}) is ok</LI>
     * </ul>
     *
     * @param   set        the current LySetController instance
     * @param   newValues  Set of values to control
     *
     * @return
     *
     *          <ul>
     *          <LI>{@link #STATUS_IGNORE} if the value has not change</LI>
     *          <LI>{@link #STATUS_OK} if the value is correct</LI>
     *          <LI>{@link #STATUS_KO} if the value is not correct</LI>
     *          <LI>{@link #STATUS_WARNING} if the value is on warning</LI>
     *          </ul>
     */
    public short controlNewValues(LySetController set, LyValueSet newValues)
    {
        // If there is no new values (newValues.getSize == 0) we must already
        // control the object because it may have a dependency with other
        // objects for a class control. For example, when a value in a
        // struct has been changed, we must control the object containing
        // this struct.
        if (newValues == null)
            return STATUS_IGNORE;

        LyObject object = set.getObject();

        if (object != null)
        {
            short status = object.callClassControls(newValues);

            if (status != STATUS_OK)
            {
                LyError error = object.getLastError(status);

                set.showError(object, error);
            }

            return status;
        }
        else
            return STATUS_OK;
    } // end method controlNewValues

    /**
     * This method adds the given extra values before completing the current operation.<br />
     * <i>The default implementation</i> checks that every field with the NOT_NULL mark are not
     * null. If a field has a null value and is not optionnal, this method tries to fill it with a
     * default value.
     *
     * @param  set     the current LySetController instance
     * @param  values  Current set of values. Additionnal values must be add to this set.
     */
    public void addExtraValues(LySetController set, LyValueSet values)
    {
        if (set == null)
            return;

        if (values == null)
            return;

        // Add extra fields specified by parameters
        LyAction    action = set.getAction();
        LySession   session = set.getSession();
        LyFieldInfo fieldInfo;
        LyValue     fieldValue;
        Object      value;
        int         np;

        fieldInfo = (LyFieldInfo)action.getParameter("_setFieldInfo", true);
        value = action.getParameter("_value", true);
        np = ((fieldInfo != null) ? 2 : 1);

        do
        {
            if (fieldInfo != null)
            {
                if ((value instanceof String) && (((String)value).charAt(0) == '$'))
                    value = session.getValue(value);

                fieldValue = fieldInfo.buildFieldValue(value);
                values.addValue(fieldValue);
            }

            fieldInfo = (LyFieldInfo)action.getParameter("_setFieldInfo" + np, true);
            value = action.getParameter("_value" + np, true);
            np++;
        }
        while (fieldInfo != null);

        if (!set.isCreation())
            return;

        // Verify that all fieldInfos with the mark NOT_NULL have a value.
        LyClassInfo             classInfo = set.getClassInfo();
        Object                  defaultValue;
        LyInfoList<LyFieldInfo> fieldsNotNull;
        int                     i, n;

        fieldsNotNull = classInfo.getFieldInfos(LyFieldInfo.MARK_NOT_NULL);
        n = fieldsNotNull.getSize();

        for (i = 0; i < n; i++)
        {
            fieldInfo = fieldsNotNull.getInfo(i);
            fieldValue = values.getFieldValue(fieldInfo);
            if (fieldValue != null)
                continue;

            defaultValue = getDefaultValue(set, fieldInfo);
            if (defaultValue != null)
            {
                // Build value from default value
                fieldValue = fieldInfo.buildFieldValue(defaultValue);
                values.addValue(fieldValue);
            }
        }
    } // end method addExtraValues

    /**
     * Final step of the object creation process.<br />
     * <i>The default implementation</i> creates the object with the given values.
     *
     * @param   create  the current LyCreateController instance
     * @param   values  Set of values corresponding to the object to create
     *
     * @return  true if the create process is successfull, false otherwise
     */
    public boolean endCreation(LyCreateController create, LyValueSet values)
    {
        // Create object with the given list of values
        LyClassInfo classInfo = create.getClassInfo();
        LyContext   context = create.getControllerContext();
        LyObject    object = create.getObject();

        // Commit creation
        LyResponse        response;
        LyResponse.Status status = LyResponse.Status.COMPLETED;

        if (object == null)
        {
            object = classInfo.createInstance();

            context.setValues(object, values);

            // callClassControls was not called by controlNewValues because
            // object was null
            short objectStatus = object.callClassControls(values);

            if (objectStatus != LyControlInterface.STATUS_OK)
            {
                LyError error = object.getLastError(objectStatus);

                create.showError(object, error);
                create.setObjectInternal(object);
                return false;
            }
        }
        else
            context.setValues(object, values);

        // Maybe the object has already been created.
        // It occurs when the object is created in a sub-form and the validation of the principal form failed.
        if (context.isAlreadyCreated(object))
            response = context.setObject(object);
        else
            response = context.createObject(object);

        if (response != null)
            status = response.getStatus();

        create.setObjectInternal(object);

        if (status == LyResponse.Status.ERROR)
        {
            LyError  error;
            Object[] args = new Object[2];

            args[0] = classInfo.getName();
            args[1] = response.getError();
            error = new LyError(LyObject.LEVEL_ERROR, "LY_ER_CREATE", args, null, object, null);
            create.showError(null, error);
        }

        return (status == LyResponse.Status.COMPLETED);
    } // end method endCreation

    /**
     * Final step of the object set process.<br />
     * <i>The default implementation</i> commits modifications.
     *
     * @param   set     the current LySetController instance
     * @param   values  Set of values for the modified object
     *
     * @return  true if the modification process is done, false otherwise
     */
    public boolean endSet(LySetController set, LyValueSet values)
    {
        if ((values == null) || (values.getSize() == 0))
            return true; // Nothing to do

        // Set object with the given list of values
        LyObject  object = set.getObject();
        LyContext context = set.getControllerContext();

        // Commit set
        LyResponse        response;
        LyResponse.Status status = LyResponse.Status.COMPLETED;

        context.setValues(object, values);
        response = context.setObject(object);
        status = response.getStatus();

        if (status == LyResponse.Status.ERROR)
        {
            LyError  error;
            Object[] args = new Object[2];

            args[0] = set.getName(object);
            args[1] = response.getError();
            error = new LyError(LyObject.LEVEL_ERROR, "LY_ER_SET", args, null, object, null);
            set.showError(object, error);
        }

        return (status == LyResponse.Status.COMPLETED);
    } // end method endSet

    /**
     * Checks if the given object deletion is enabled or not (used before asking to the end-user if
     * he want to delete it).<br />
     * <i>The default implementation</i> does nothing and return null.
     *
     * @param   delete  current LyDeleteController instance
     * @param   object  Object checked before deletion.
     *
     * @return  Error message if deletion is not allowed, null otherwise.
     */
    public String checkDeletion(LyDeleteController delete, LyObject object)
    {
        return null;
    }

    /**
     * Checks if the given objects list deletion is enabled or not (used before asking to the
     * end-user if he want to delete them).<br />
     * <i>The default implementation</i> creates a new
     * {@link leon.data.LySimpleObjectList LySimpleObjectList(list)} and return it.
     *
     * @param   delete  current LyDeleteController instance
     * @param   list    list of objects to checked before deletion.
     *
     * @return  The list of allowed objects to delete.
     */
    public LyObjectList checkDeletions(LyDeleteController delete, LyObjectList list)
    {
        return new LySimpleObjectList(list);
    }

    /**
     * Asks for the deletion message for the given object.<br />
     * <i>The default implementation</i> does nothing and return null.
     *
     * @param   delete  current LyDeleteController instance
     * @param   object  Object checked before deletion.
     *
     * @return  Confirmation message to display to the user if necessary, null otherwise.
     */
    public String getConfirmDeletionMessage(LyDeleteController delete, LyObject object)
    {
        return null;
    }

    /**
     * Final step of object deletion process.<br />
     * <i>The default implementation</i> delete the object associated to the given controller.
     *
     * @param   delete  current LyDeleteController instance
     *
     * @return  true if the deletion is complete, false otherwise
     */
    public boolean endDeletion(LyDeleteController delete)
    {
        // Delete object
        LyObject object = delete.getObject();
        // XXX enhance performance with this info : idle ?

        // Get a context for the operation
        LyContext         context = delete.getControllerContext();
        LyResponse        response;
        LyResponse.Status status;

        response = context.deleteObject(object);
        status = response.getStatus();

        if (status == LyResponse.Status.ERROR)
        {
            LyError  error;
            Object[] args = new Object[2];

            args[0] = delete.getName(object);
            args[1] = response.getError();
            error = new LyError(LyObject.LEVEL_ERROR, "LY_ER_DELETE", args, null, object, null);
            delete.showError(object, error);
        }

        return (status == LyResponse.Status.COMPLETED);
    } // end method endDeletion

    /**
     * Method called when object shown in the form is changed.<br />
     * <i>The default implementation</i> does nothing.
     *
     * @param  controller  the current LyConsultController instance.
     * @param  object      the object shown in the form.
     */
    public void setObject(LyConsultController controller, LyObject object)
    {
        // Default implementation does nothing
    }

    /**
     * Method for changing state of an object.
     */
    /**
     * Gets an error message from the given list of LyError.
     *
     * @param   environment  the current environment
     * @param   errors       the list of errors (type LyError)
     *
     * @return  the message
     */
    protected String getErrorMessage(LyEnvironment environment, Vector<LyError> errors)
    {
        int          i, n = errors.size();
        LyError      error;
        StringBuffer buffer = new StringBuffer();
        String       message;
        LyObject     cause;

        for (i = 0; i < n; i++)
        {
            error = errors.elementAt(i);
            if (i != 0)
                buffer.append("\n");

            cause = error.getObjectCause();
            if (cause != null)
            {
                buffer.append(cause.getName());
                buffer.append(" : ");
            }

            buffer.append(error.getMessage(environment));
        }

        message = buffer.toString();
        return message;
    }

    /**
     * Generic method to handle modification of the state of an object (loop on values of a choice
     * field info).
     *
     * @param  parent     Parent controller from which the given action was invoked.
     * @param  action     Action of changing state.
     * @param  objects    Objects on which the state is changed.
     * @param  fieldInfo  Field info that symbolises state (Generic action only considers a
     *                    non-multiple choiceFieldInfo)
     */
    public void changeState(LyController parent, LyAction action, LyObjectList objects, LyFieldInfo fieldInfo)
    {
        if (objects == null)
            return;

        if (fieldInfo == null)
            fieldInfo = (LyFieldInfo)action.getParameter("_fieldInfo");

        if (!(fieldInfo instanceof LyChoiceFieldInfo) || fieldInfo.isMultiple())
            throw new IllegalArgumentException(
                "BF AppEngine : generic state modification is only supported for a non-multiple choice fieldInfo (field : "
                + fieldInfo + ")");

        // Change state : state = state + 1 % nbStates
        LyChoiceFieldInfo choice = (LyChoiceFieldInfo)fieldInfo;
        Short             value;
        Boolean           bvalue;
        short             state;
        boolean           bstate;
        int               i, n = objects.getSize();
        LyObject          object;
        LySession         session = parent.getSession();
        LyContextManager  cm = session.getContextManager();
        LyContext         context;
        LyContext.Status  status = LyContext.Status.COMPLETED;
        String            contextName;

        contextName = action.getName();
        if (n > 0)
        {
            object = objects.getObject(0);
            contextName += " ";
            contextName += object.getName();
        }

        if (n > 1)
            contextName += "...";

        context = cm.newContext(session, LyContext.Type.USER_NOT_CANCELLABLE);
        context.setName(contextName);

        if (parent instanceof LyControllerContextInterface)
        {
            LyContext parentContext = ((LyControllerContextInterface)parent).getControllerContext();

            parentContext.addSon(context);
        }

        for (i = 0; i < n; i++)
        {
            object = objects.getObject(i);
            if (choice.getType() == LyChoiceFieldInfo.Type.BOOLEAN)
            {
                bvalue = (Boolean)object.getValue(choice);
                bstate = ((bvalue == null) ? false : bvalue.booleanValue());
                context.setValue(object, choice, !bstate);
            }
            else
            {
                value = (Short)object.getValue(choice);
                state = ((value == null) ? 0 : value.shortValue());
                state = (short)((state + 1) % choice.getChoiceOptionCount());
                context.setValue(object, choice, new Short(state));
            }

            context.setObject(object);
        }

        status = context.commit();

        if (status == LyContext.Status.ROLLBACKED)
        {
            LyEnvironment environment = session.getEnvironment();
            String        cause = context.getErrorMessage();
            String        message;

            if (cause == null)
                cause = session.getApplication().getError();

            message = environment.getMessage("LY_ER_COMMIT_CONTEXT", cause);
            environment.raiseError(message);
        }
        else if (status == LyContext.Status.ERROR)
        {
            Vector<LyError> errors = context.getErrors(LyObject.LEVEL_ERROR);

            if ((errors != null) && (errors.size() != 0))
            {
                context.rollback();

                LyEnvironment environment = session.getEnvironment();
                String        message = getErrorMessage(environment, errors);

                message = environment.getMessage("LY_ER_COMMIT_CONTEXT", message);
                environment.raiseError(message);
            }
        }
    } // end method changeState

    /**
     * Method for changing state of an object.
     */

    /**
     * Generic method to handle modification of an objects list.
     *
     * @param  parent       Parent controller from which the given action was invoked.
     * @param  action       Action of modifying.
     * @param  objects      Objects to modify.
     * @param  fieldValues  Values to assign on given objects.
     */
    public void modify(LyController parent, LyAction action, LyObjectList objects, LyValueSet fieldValues)
    {
        if (objects == null)
            return;

        LySession   session = parent.getSession();
        LyFieldInfo fieldInfo;
        LyValue     fieldValue;

        if (fieldValues == null)
        {
            Object value;
            int    nf = 1;

            fieldValues = new LyValueSet();
            fieldInfo = (LyFieldInfo)action.getParameter("_fieldInfo", true);
            value = action.getParameter("_value", true);
            if (fieldInfo != null)
                nf = 2;

            do
            {
                if (fieldInfo != null)
                {
                    if ((value instanceof String) && (((String)value).charAt(0) == '$'))
                        value = session.getValue(value);

                    fieldValue = fieldInfo.buildFieldValue(value);
                    fieldValues.addValue(fieldValue);
                }

                fieldInfo = (LyFieldInfo)action.getParameter("_fieldInfo" + nf, true);
                value = action.getParameter("_value" + nf, true);
                nf++;
            }
            while (fieldInfo != null);
        }

        // Change state : state = state + 1 % nbStates
        int              i, no = objects.getSize();
        LyObject         object;
        int              j, nv = fieldValues.getSize();
        LyContextManager cm = session.getContextManager();
        LyContext        context;
        LyContext.Status status = LyContext.Status.COMPLETED;
        String           contextName;

        contextName = action.getName();
        if (no > 0)
        {
            object = objects.getObject(0);
            contextName += " ";
            contextName += object.getName();
        }

        if (no > 1)
            contextName += "...";

        context = cm.newContext(session, LyContext.Type.USER_NOT_CANCELLABLE);
        context.setName(contextName);

        if (parent instanceof LyControllerContextInterface)
        {
            LyContext parentContext = ((LyControllerContextInterface)parent).getControllerContext();

            parentContext.addSon(context);
        }

        LyValueSet values;

        for (i = 0; i < no; i++)
        {
            object = objects.getObject(i);
            values = new LyValueSet();
            for (j = 0; j < nv; j++)
            {
                fieldValue = fieldValues.getFieldValue(j);
                values.addValue(fieldValue);
            }

            context.setValues(object, values);
            context.setObject(object);
        }

        status = context.commit();

        if (status == LyContext.Status.ROLLBACKED)
        {
            LyEnvironment environment = session.getEnvironment();
            String        cause = context.getErrorMessage();
            String        message;

            if (cause == null)
                cause = session.getApplication().getError();

            message = environment.getMessage("LY_ER_COMMIT_CONTEXT", cause);
            environment.raiseError(message);
        }
        else if (status == LyContext.Status.ERROR)
        {
            Vector<LyError> errors = context.getErrors(LyObject.LEVEL_ERROR);

            if ((errors != null) && (errors.size() != 0))
            {
                context.rollback();

                LyEnvironment environment = session.getEnvironment();
                String        message = getErrorMessage(environment, errors);

                message = environment.getMessage("LY_ER_COMMIT_CONTEXT", message);
                environment.raiseError(message);
            }
        }
    } // end method modify

    /**
     * Method for showing/hidding an object.
     */

    /**
     * Generic method to show an objects list.
     *
     * @param  parent   Parent controller from which the given action was invoked.
     * @param  action   Action of showing.
     * @param  objects  Objects to show.
     */
    public void show(LyController parent, LyAction action, LyObjectList objects)
    {
        if (objects == null)
            return;

        int      i, no = objects.getSize();
        LyObject object;

        for (i = 0; i < no; i++)
        {
            object = objects.getObject(i);
            object.setHidden(false);
        }
    }

    /**
     * Generic method to hide an objects list.
     *
     * @param  parent   Parent controller from which the given action was invoked.
     * @param  action   Action of hidding.
     * @param  objects  Objects to hide.
     */
    public void hide(LyController parent, LyAction action, LyObjectList objects)
    {
        if (objects == null)
            return;

        int      i, no = objects.getSize();
        LyObject object;

        for (i = 0; i < no; i++)
        {
            object = objects.getObject(i);
            object.setHidden(true);
        }
    }

    /**
     * Generic method to copy an object.
     *
     * @param  parent   Parent controller from which the given action was invoked.
     * @param  action   Action of copying.
     * @param  objects  Objects to copy.
     */
    public void copy(LyController parent, LyAction action, LyObjectList objects)
    {
    	// To be implemented
    }

    /**
     * Generic method to paste an object.
     *
     * @param  parent        Parent controller from which the given action was invoked.
     * @param  action        pasting action
     * @param  objectString  description of objects to paste
     */
    public void paste(LyController parent, LyAction action, String objectString)
    {
    	// To be implemented
    }

    /**
     * Return the next available value for a given field info and class info. Only number and text
     * field infos are supported.
     *
     * @param   session    current session
     * @param   classInfo  class info of the object concerned
     * @param   fieldInfo  field info on which we want to get the next value
     * @param   oldValue   the last value for the given field info
     *
     * @return  the next available value for the given field info
     */
    protected Object getNextValue(LySession session, LyClassInfo classInfo, LyFieldInfo fieldInfo, Object oldValue)
    {
        if (fieldInfo == null)
            return null;

        // If number field info, retrieve the first available integer
        if (fieldInfo instanceof LyNumberFieldInfo)
        {
            LyNumberFieldInfo numberFieldInfo = (LyNumberFieldInfo)fieldInfo;
            int               count = 0;
            boolean           found = true;
            Integer           newValue = null;

            // Only manage integer values
            if ((numberFieldInfo.getType() != LyNumberFieldInfo.Type.INT)
                    && (numberFieldInfo.getType() != LyNumberFieldInfo.Type.SHORT)
                    && (numberFieldInfo.getType() != LyNumberFieldInfo.Type.LONG))
                return null;

            // Retrieve the first available integer
            LySort sort = new LySort();

            sort.addCriterion(fieldInfo, false);

            LyClassWorkSpace ws = new LyClassWorkSpace(session, classInfo, null, sort, false);

            // Retrieve the object with the greatest value for this field info
            if ((ws != null) && (ws.getSize() > 0))
            {
                LyObject object = ws.getObject(0);
                Integer  greatestValue = (Integer)object.getValue(fieldInfo);

                count = greatestValue.intValue();
            }

            while (found)
            {
                count++;
                newValue = new Integer(count);

                LyFilter           filter = new LyFilter();
                LyFilterExpression expression = new LyFilterExpression(fieldInfo, newValue);

                filter.addElement(expression);

                ws = new LyClassWorkSpace(session, classInfo, filter, false);

                found = (ws.getSize() > 0);
            }

            return newValue;
        }
        // If text field info, add "(n)" to the given value
        else if (fieldInfo instanceof LyTextFieldInfo)
        {
            String  stringValue = (String)oldValue;
            int     count = 0;
            boolean found = true;
            String  newValue = null;

            while (found)
            {
                count++;
                newValue = stringValue + " (" + count + ")";

                LyFilter           filter = new LyFilter();
                LyFilterExpression expression = new LyFilterExpression(fieldInfo, newValue);

                filter.addElement(expression);

                LyClassWorkSpace ws = new LyClassWorkSpace(session, classInfo, filter, false);

                found = (ws.getSize() > 0);
            }

            return newValue;
        }

        return null;
    } // end method getNextValue

    /**
     * Get action to use for consulting or editing the given relation with the specified class info
     * selected on it.
     *
     * @param   consult    the current LyConsultController instance.
     * @param   relation   Edited relation
     * @param   classInfo  Selected class info.
     *
     * @return  The action to use, if null let controller decides.
     */
    public LyAction getAction(LyConsultController consult, LyRelationFieldInfo relation, LyClassInfo classInfo)
    {
        return null;
    }

    /**
     * Checks if an edit table should be used for the given array field.
     *
     * @param   controller  The current LySetController instance.
     * @param   array       The array field info.
     *
     * @return  True if an edit table is requested, false otherwise.
     */
    public boolean useEditTable(LySetController controller, LyArrayFieldInfo array)
    {
        LyEnvironment environment = controller.getEnvironment();

        if (environment.testResource("LY_EDIT_TABLE_FOR_ARRAYS"))
            return true;

        LyFieldInfo fieldInfo = array.getFieldInfo();

        if (!(fieldInfo instanceof LyStructFieldInfo))
            return false;

        LyUnit      unit = fieldInfo.getFieldUnit();
        LyInfoLabel label = fieldInfo.getLabel(LyInfoLabel.Type.VALUE);

        if ((unit != null) || (label != null))
            return false;

        LyClassInfo             classInfo = ((LyStructFieldInfo)fieldInfo).getClassInfo(0);
        LyInfoList<LyFieldInfo> fields = classInfo.getFieldInfos(LyFieldInfo.MARK_TABLE);

        if (fields == null)
            fields = classInfo.getFieldInfos();

        if (fields.getSize() == 1)
        {
            LyFieldInfo field = fields.getInfo(0);

            if (!(field instanceof LyArrayFieldInfo) && !(field instanceof LyStructFieldInfo))
                return false;
        }

        return true;
    } // end method useEditTable

    /**
     * Checks if an edit table should be used for the given relation field.
     *
     * @param   controller  The current LySetController instance.
     * @param   relation    The relation field info.
     *
     * @return  True if an edit table is requested, false otherwise.
     */
    public boolean useEditTable(LySetController controller, LyRelationFieldInfo relation)
    {
        LyEnvironment environment = controller.getEnvironment();

        return ((environment.getResource("LY_EDIT_TABLE_FOR_RELATIONS") == null)
        	|| environment.testResource("LY_EDIT_TABLE_FOR_RELATIONS"));
    }

    /**
     * Builds a filter element with the given parameters.<br />
     * <i>The default implementation</i> does nothing.
     *
     * @param   filter     the current LyFilterController instance
     * @param   fieldInfo  FieldInfo on which this filter expression applies.
     * @param   not        Inverts filter condition or not.
     * @param   condition  Filter condition.
     * @param   value      Value of this expression (may be null). This value is an instance of the
     *                     class returned by the getValueClass method of LyFieldInfo.
     * @param   modifiers  List of modifiers (may be null).
     *
     * @return  The corresponding filter element.
     */
    public LyFilterElement buildFilterElement(LyFilterController filter, LyFieldInfo fieldInfo, boolean not,
        LyFilterExpression.Condition condition, Object value, BitSet modifiers)
    {
        // Accept default behavior
        return null;
    }

    /**
     * Get action used for fast edition of given field.
     *
     * @param   outline    the current LyOutlineController instance.
     * @param   fieldInfo  the fieldInfo to be edited.
     *
     * @return  The action to be used for fast editing or null if field id not editable.
     */
    public LyAction getSetAction(LyOutlineController outline, LyFieldInfo fieldInfo)
    {
        LyObject object = outline.getObject();

        return outline.getFastEditionAction(object, fieldInfo);
    }

    /**
     * Invoked when controller is starting to commit, default implementation does nothing.
     *
     * @param  controller  The current LySetController instance.
     */
    public void startCommit(LySetController controller)
    {
    }

    /**
     * Invoked when controller is ending commit, default implementation does nothing.
     *
     * @param  controller  The current LySetController instance.
     */
    public void endCommit(LySetController controller)
    {
    }

    /**
     * Indicates if the operations of the controller context must be inserted at the beginning of
     * the parent operations or at the end.
     *
     * @param   controller  the set controller
     * @param   relation    the relation field info
     *
     * @return  Position where operations must be inserted, may be null if left to controller
     *          decision.
     */
    public LyContext.Position getInsertPosition(LySetController controller, LyRelationFieldInfo relation)
    {
        return null;
    }

    /**
     * Method called on form save.
     *
     * @param  controller  The current LySetController instance.
     * @param  formSave    The formSave that was created.
     */
    public void formSaved(LySetController controller, LyFormSave formSave)
    {
    }

    /**
     * Method called on form restore.
     *
     * @param  controller  The current LySetController instance.
     * @param  formSave    The formSave that was restored.
     */
    public void formRestored(LySetController controller, LyFormSave formSave)
    {
    }

    /**
     * Indicates if the given controller must link its context to its parent context.
     *
     * @param  controller      The current LySetController instance.
     * @param  parentRelation  the parent relation for this controller (may be null)
     */
    public boolean useSingleContext(LySetController controller, LyRelationFieldInfo parentRelation)
    {
        return false;
    }

    /**
     * Checks if given object is locked for edition.
     * @param controller The current LySetController instance.
     * @param object Object to be checked.
     * @return Null if not locked, otherwise a string message giving information about the lock (user, date...).
     */
    public String isLocked(LySetController controller, LyObject object)
    {
    	LyAction		action = controller.getAction();
    	LyClassInfo		classInfo = controller.getClassInfo();
    	LyTextFieldInfo	lockFieldInfo = (LyTextFieldInfo)action.getParameter("_lockFieldInfo", true);
    	LyDateFieldInfo	lockDateFieldInfo = (LyDateFieldInfo)action.getParameter("_lockDateFieldInfo", true);

    	if ((lockFieldInfo == null)
    			|| !classInfo.contains(lockFieldInfo))
    		return null; // No lock management

    	LyInfoList<LyFieldInfo>		fieldInfos = new LyInfoList<LyFieldInfo>(lockFieldInfo);

    	if (lockDateFieldInfo != null)
    		fieldInfos.addInfo(lockDateFieldInfo);
	    object.reload(fieldInfos);

    	LyTextValue	lockValue = object.getTextValue(lockFieldInfo);

    	if (lockValue.isNull())
    		return null;

    	LyDateValue		lockDateValue = null;

    	if ((lockDateFieldInfo != null)
    			&& classInfo.contains(lockDateFieldInfo))
    		lockDateValue = object.getDateValue(lockDateFieldInfo);

    	LyEnvironment	environment = controller.getEnvironment();

    	return environment.getMessage("LY_OBJECT_LOCKED", object.getName(), lockValue.getValue(),
    		(lockDateValue == null ? null : lockDateValue.getValue()));
    }

    /**
     * Locks/Unlocks given object for edition.
     * @param controller The current LySetController instance.
     * @param object Object to be locked/unlocked.
     * @param lock True if lock is required, false for unlock.
     * @return True if successful, false otherwise.
     */
    public boolean lockObject(LySetController controller, LyObject object, boolean lock)
    {
    	if (!lock
    			&& !controller.isEditable())
    		return true;

    	LyAction		action = controller.getAction();
    	LyClassInfo		classInfo = controller.getClassInfo();
    	LyTextFieldInfo	lockFieldInfo = (LyTextFieldInfo)action.getParameter("_lockFieldInfo", true);

    	if ((lockFieldInfo == null)
    			|| !classInfo.contains(lockFieldInfo))
    		// No lock
    		return true;

    	// Lock given object
    	if (lock)
    	{
	    	LySession	session = controller.getSession();
	    	LyObject	user = session.getUser();

	    	object.setValue(lockFieldInfo, (user != null ? user.getName() : session.getId()));
    	}
    	else
    		object.setValue(lockFieldInfo, null);

    	LyDateFieldInfo	lockDateFieldInfo = (LyDateFieldInfo)action.getParameter("_lockDateFieldInfo", true);

    	if ((lockDateFieldInfo != null)
    			&& classInfo.contains(lockDateFieldInfo))
    		object.setValue(lockDateFieldInfo, (lock ? new Date() : null));

    	LyResponse	response = object.set();

    	return (response.getStatus() == LyResponse.Status.COMPLETED);
    }
} // end class LyClassBehavior
