/*
 * LyDashboardBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.info.misc.LyActionResource;
import leon.misc.*;
import leon.view.*;


/**
 * This class implements the default behavior for a LyDashboardController.
 *
 * @see LyDashboardController
 * @see LyDashboardBehaviorInterface
 */
public class LyDashboardBehavior extends LyActionBehavior
	implements LyDashboardBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyDashboardBehavior.java 35922 2012-06-15 16:34:36Z jlsaouli $";

	/**
	 * Default constructor.
	 */
	public LyDashboardBehavior()
	{
	}

	/**
	 * Indicates if the given sub view controller has to freed when closed.
	 *
	 * @param  dashboard      Dashboard controller.
	 * @param  viewController View controller.
	 * @return True if the given view controller is freed on closing, false otherwise.
	 * @deprecated Use {@link LyDashboardBehaviorInterface#isFreedOnClose(LyDashboardController, LyAction,
	 *             LyViewController)} instead.
	 * @see        LyDashboardBehaviorInterface#isFreedOnClose(LyDashboardController,LyAction,LyViewController)
	 */
	public boolean isFreedOnClose(LyDashboardController dashboard, LyViewController viewController)
	{
		return isFreedOnClose(dashboard, null, viewController);
	}

	/**
	 * Indicates if the given sub view controller has to be freed when closed.
	 *
	 * @param  dashboard      Dashboard controller.
	 * @param  action		  The action of the sub view controller.
	 * @param  viewController View controller.
	 * @return True if the given view controller is freed on closing, false otherwise.
	 */
	public boolean isFreedOnClose(LyDashboardController dashboard, LyAction action, LyViewController viewController)
	{
		// Check on environment if the LY_FREE_ON_CLOSE resource is set
		LyEnvironment	environment = dashboard.getEnvironment();
		LyViewManager	viewManager = dashboard.getViewManager();
		
		if (viewManager.hasProperty("FREE_ON_CLOSE")
				|| environment.testResource("LY_FREE_ON_CLOSE"))
			return true;
		
		Vector<LyActionResource>	resources = action.getResources();
		
		if (resources != null)
		{
			int		i, n = resources.size();
			
			for (i = 0; i < n; i++)
				if (resources.elementAt(i).getKey().equals("LY_FREE_ON_CLOSE"))
					return true;
		}
		
		return (viewController instanceof LyDialogController);
	}

	/**
	 * Indicates if the given action has to be opened in a new view.
	 * Default implementation uses the _useSingleView parameter of the dashboard action.
	 *
	 * @param  dashboard  Dashboard controller.
	 * @param  action	  The action to be opened.
	 * @return True if the given action has to be opened in a new view, false otherwise.
	 */
	public boolean openNewView(LyDashboardController dashboard, LyAction action)
	{
		if (!action.hasMark(LyAction.MARK_FRAME_VIEW))
			return true;
		
		return !dashboard.useSingleView();
	}
    
    /**
     * Method called when a new controller is (re)opened in the dasboard panel.
     * @param  dashboard      Dashboard controller.
     * @param  controller Area controller opened in the dashboard.
     */
    public void openController(LyDashboardController dashboard, LyAreaController controller)
    {
    }
    
    /**
     * Method called when a controller is closed in the dasboard panel.
     * @param  dashboard      Dashboard controller.
     * @param  controller Area controller closed in the dashboard.
     */
    public void closeController(LyDashboardController dashboard, LyAreaController controller)
    {
    }
	
	/**
	 * Builds a tab for given area controller when the useTabs mode is active.
	 * 
	 * @param dashboard  Dashboard controller.
	 * @param areaController Area controller opened in the dashboard.
	 * 
	 * @return the newly build tab
	 */
	public LyTab buildTab(LyDashboardController dashboard, LyAreaController areaController)
	{
		LyAction	action = areaController.getAction();
		LyTab		tab = new LyTab();
		
		tab.setId(LyDashboardController.TAB_ID_PREFIX + action.getId());
		tab.setLabel(action.getAlias());
		tab.setImage(areaController.getImage());
		
		return tab;
	}
	
	/**
	 * Get the list of actions displayed in the dashboard.
	 *
	 * @param dashboard  Dashboard controller.
	 * @param actions List of actions declared in the model. Can be null.
	 * @return
	 */
	public LyInfoList<LyAction> getActions(LyDashboardController dashboard, LyInfoList<LyAction> actions)
	{
		if ((actions != null) && (actions.getSize() != 0))
			return actions;
		
		// Build default actions list
		actions = new LyInfoList<LyAction>();

		LyApplication			  application = dashboard.getApplication();
		LyInfoList<LyClassInfo>   classes = application.getClassInfos();
		int						  i, n = classes.getSize();
		LyClassInfo				  appClassInfo;
		LyAction				  subAction;
		String					  subActionId;
		LyAction				  template;
		Hashtable<String, Object> parameters;
		String					  limit = "50";
		LyObjectSource			  source;

		for (i = 0; i < n; i++)
		{
			appClassInfo = classes.getInfo(i);
			if (appClassInfo.getApplication() != application)
				continue;
			if (appClassInfo.getId().startsWith("__"))
				continue;
			if (appClassInfo.isAbstract())
				continue;
			
			source = appClassInfo.getObjectSource();
			if (source == null || source.getSize() < 50)
			{
				template = application.getAction("_table");
				parameters = null;
			}
			else
			{
				template = application.getAction("_multiPageTable");
				parameters = new Hashtable<String, Object>();
				parameters.put("_limit", limit);
			}

			subActionId = appClassInfo.getId() + '.' + template.getId();
			subAction = application.getAction(subActionId);
			if (subAction == null)
				subAction = new LyAction(subActionId, template, appClassInfo, parameters);
			actions.addInfo(subAction);
		}
		return actions;
	}
	
	/**
	 * Get the classes list to be dispalyed in search tool.
	 * 	 
	 * @param dashboard  Dashboard controller.
	 * @return the classes list used for search, null to let the controller do the job.
	 */
	public LyInfoList<LyClassInfo>	getSearchClasses(LyDashboardController dashboard)
	{
		return null;
	}
} // end class LyDashboardBehavior
