/*
 * LyApplicationBehavior.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.app.behavior;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.data.*;
import leon.info.*;
import leon.info.infointerface.*;
import leon.info.location.*;
import leon.info.notifier.*;
import leon.info.save.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.sql.*;

import java.io.*;
import java.sql.*;
import java.util.*;


/**
 * Default implementation of the LyApplicationBehavior. This implementation does nothing.
 *
 * @author   Lyria S.A.
 * @version  $Id: LyDataApplicationBehavior.java 41407 2012-10-30 16:17:49Z bangiole $
 * @see      leon.app.behaviorinterface.LyApplicationBehaviorInterface
 * @see      leon.app.LyApplication
 */
public class LyDataApplicationBehavior
	implements LyApplicationBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyDataApplicationBehavior.java 41407 2012-10-30 16:17:49Z bangiole $";

	/** constant for table name for saving preferences in a blob. */
	public static String PREF_TABLE = "leon_preferences";
	/** constant for id column for saving preferences in a blob. */
	public static String PREF_ID_COL = "pref_id";
	/** constant for pref type column for saving preferences in a blob. */
	public static String PREF_TYPE_COL = "pref_type";
	/** constant for userId column for saving preferences in a blob. */
	public static String PREF_USER_COL = "pref_user";
	/** constant for date column for saving preferences in a blob. */
	public static String PREF_DATE_COL = "pref_date";
	/** constant for blob column for saving preferences in a blob. */
	public static String PREF_BLOB_COL = "pref_blob";

	/** Priority level. */
	protected Integer _priorityLevel;

	/**
	 * Constructor. Sets the priority level to null.
	 */
	public LyDataApplicationBehavior()
	{
		_priorityLevel = null;
	}

	/**
	 * Method called when an application is starting. This method is called <b>before</b> loading
	 * project and creating providers see {@link #started(LyApplication)} to be notified <b>
	 * after</b> generic staring process.<br />
	 * This implementation does nothing and return true.
	 *
	 * @param   application  Application that is starting.
	 *
	 * @see     #started(LyApplication)
	 *
	 * @return  true
	 */
	public boolean start(LyApplication application)
	{
		return true;
	}

	/**
	 * Method called when an application is started. This method is called <b>after</b> loading
	 * project and creating providers see {@link #start(LyApplication)} to be notified <b>before</b>
	 * generic staring process.<br />
	 * This implementation does nothing and return true.
	 *
	 * @param   application  Application that is started.
	 *
	 * @see     #start(LyApplication)
	 *
	 * @return  true
	 */
	public boolean started(LyApplication application)
	{
		return true;
	}

	/**
	 * Creates daemons to manage relations within an application.<br />
	 * This implementation does nothing.
	 *
	 * @param  application  Current application.
	 */
	public void createDaemons(LyApplication application)
	{
	}

	/**
	 * Load application preferences.
	 *
	 * @param   application  Application on which preferences are loaded.
	 *
	 * @return  List of preferences of the application.
	 */
	public LyInfoList<LyInfo> loadPreferences(LyApplication application)
	{
		return null;
	}

	/**
	 * Load extra infos.
	 *
	 * @param   application  Application on which extra infos (like charts or treemap
	 *                       configurations) are loaded.
	 *
	 * @return  List of extra infos for the application.
	 */
	public LyInfoList<LyInfo> loadExtraInfos(LyApplication application)
	{
		return null;
	}

	/**
	 * Method called when a session is added to the application.<br />
	 * This implementation does nothing.
	 *
	 * @param  application  Application in which a session is added.
	 * @param  session      Session added to the application.
	 */
	public void sessionAdded(LyApplication application, LyDataSession session)
	{
	}

	/**
	 * Method called when a user logged in a session.
	 *
	 * @param application Application in which a session is added.
	 * @param session     Session added to the application.
	 * @param user		  User that logged in.
	 */
	public void sessionLoggedIn(LyApplication application, LyDataSession session, LyObject user)
	{
	}

	/**
	 * Method called when a session is removed from the application.<br />
	 * This implementation does nothing.
	 *
	 * @param  application  Application in which a session is removed.
	 * @param  session      Session removed from the application.
	 */
	public void sessionRemoved(LyApplication application, LyDataSession session)
	{
	}

	/**
	 * Method called when datas are loaded for an application.<br />
	 * This implementation does nothing.
	 *
	 * @param  application  Application for which datas are loaded.
	 */
	public void loadData(LyApplication application)
	{
	}

	/**
	 * Method called when datas are unloaded for an application.<br />
	 * This implementation does nothing.
	 *
	 * @param  application  Application for which datas are unloaded.
	 */
	public void unloadData(LyApplication application)
	{
	}

	/**
	 * Method called when datas are saved for an application.<br />
	 * This implementation does nothing.
	 *
	 * @param  application  Application for which datas are saved.
	 */
	public void saveData(LyApplication application)
	{
	}

	/**
	 * Method called when an application is closed.<br />
	 * This implementation does nothing.
	 *
	 * @param  application  Application that is closed.
	 */
	public void close(LyApplication application)
	{
	}

	/**
	 * Method called when an application falls in sleeping mode.<br />
	 * This implementation does nothing.
	 *
	 * @param  application  Application for which datas are saved.
	 */
	public void sleep(LyApplication application)
	{
	}

	/**
	 * Method called when an application is awake from sleeping mode.<br />
	 * This implementation does nothing.
	 *
	 * @param  application  Application that is awake.
	 */
	public void wakeUp(LyApplication application)
	{
	}

	/**
	 * Gets all languages that the given application supports.<br />
	 * This implementation returns the if set, the values extracted from the LY_LANGUAGES ini variable
	 *
	 * @param   application  Application that is awake.
	 *
	 * @return  a list of supported languages
	 */
	public String[][] getSupportedLanguages(LyApplication application)
	{
		LyEnvironment environment = application.getEnvironment();
		String		  languages = environment.getEnv("LY_LANGUAGES");

		if (languages == null)
			return null;

		String[] list = LyString.stringToList(languages, '|');

		if ((list == null)
				|| (list.length == 0))
			return null;

		int		   i, n = list.length;
		String[][] result = new String[n][2];

		for (i = 0; i < n; i++)
		{
			result[i][0] = list[i];
			result[i][1] = "LY_LANGUAGE_" + list[i].toUpperCase();
		}

		return result;
	}

	/**
	 * Get all timezone supported by this application.
	 * If the LY_TIMEZONES env variable is set, this method returns the contents of this variable as a list.
	 * Otherwise, it returns null;
	 * Validity of a time zone depends on the java.util.TimeZone class.
	 * An unknow timezone for the TimeZone class is interpreted as GMT.
	 *
	 * @param application the current application
	 * @return a list of supported timezones.
	 */
	public String[] getSupportedTimeZones(LyApplication application)
	{
		LyEnvironment environment = application.getEnvironment();
		String		  timezones = environment.getEnv("LY_TIMEZONES");

		String[] list = (timezones == null ? null : LyString.stringToList(timezones, '|'));

		if ((list != null) && (list.length > 0))
			return list;

		return null;
	}

	/**
	 * Adds specific encoders for new providers if needed. An encoder is used to encode a request
	 * made on a data source and decode results to give a response. For instance, the
	 * LySqlRequestEncoder encodes and decodes SQL queries to a RDBMS.<br />
	 * This implementation does nothing.
	 *
	 * @param  application  Current application.
	 * @param  provider     provider which may need an encoder
	 */
	public void addSpecificEncoders(LyApplication application, LyDataProvider provider)
	{
	}

	/**
	 * Method called when a connection with a server
	 * ({@link leon.info.location.LyLocation LyLocation} or
	 * {@link leon.info.notifier.LyNotifier LyNotifier}) is lost.<br />
	 * This implementation shows a message indicating the connection is lost and closes all dialog
	 * windows.
	 *
	 * @param  application  Current application.
	 * @param  server       Logical server definition
	 */
	public void connectionLost(LyApplication application, LyInfo server)
	{
		LyDataSession session = application.getCurrentSession();
		LyEnvironment environment = ((session == null) ? application.getEnvironment() : session.getEnvironment());

		if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
		{
			environment.logOut("BF AppEngine : Connection lost with server " + server.getId());
			if (environment.isDebugLevel(LyEnvironment.DEBUG_ERROR))
				environment.writeException(new Exception("Stack trace"));
		}

		// Free dialogs
		application.freeDialogSessions();

		LyLocation location;

		if (server instanceof LyNotifier)
			location = ((LyNotifier)server).getLocation();
		else
			location = (LyLocation)server;

		// Try to reconnect
		boolean reconnected = application.setDataProviderLocation(location.getId(), location);

		if (reconnected)
			environment.logOut("BF AppEngine : Connection restored with server " + server.getId());
		else
			environment.logOut("BF AppEngine : Failed to restore with server " + server.getId());

		if (!reconnected || (location.getNotifier() != null))
		{
			// Display message
			String message;

			if (server instanceof LyLocation)
				message = environment.getMessage("LY_ER_CONNECTION_LOST", server.getId());
			else
				message = "LY_NOTIFIER_CONNECTION_LOST";

			environment.raiseError(message);
		}
	} // end method connectionLost

	/**
	 * Gets the control priority level of the application. This level is used to enable/disable
	 * controls.
	 *
	 * @param   application  Current application.
	 * @param   object       the object on which the control is applied.
	 *
	 * @return  the priority level of the application.
	 */
	public int getPriorityLevel(LyApplication application, LyObject object)
	{
		if (_priorityLevel != null)
			return _priorityLevel.intValue();

		int	priorityLevel = Integer.MAX_VALUE;

		_priorityLevel = new Integer(priorityLevel);
		return priorityLevel;
	}

	/**
	 * Method called when a transaction begins on a location.<br />
	 * This implementation print a trace about the begining of the transaction if the environment
	 * {@link LyEnvironment#isDebugLevel(int) debug level} is
	 * {@link LyEnvironment#DEBUG_TRANSACTION}.
	 *
	 * @param  application    Current application
	 * @param  idTransaction  Id of the begin transaction
	 * @param  locationId     Id of the location
	 */
	public void begin(LyApplication application, String idTransaction, String locationId)
	{
		LyEnvironment environment = application.getEnvironment();

		if (environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
			environment.logOut("[TRANSACTION] begin : transaction=" + idTransaction + " location="
				+ locationId);
	}

	/**
	 * Method called when a request is sent in a transaction for a location.<br />
	 * <i>The default implementation</i> print a trace about the transaction update if the
	 * environment {@link LyEnvironment#isDebugLevel(int) debug level} is
	 * {@link LyEnvironment#DEBUG_TRANSACTION}.
	 *
	 * @param  application    Current application
	 * @param  idTransaction  Id of the transaction
	 * @param  locationId     Id of the location
	 * @param  request        Sent request if the request has a string representation
	 */
	public void update(LyApplication application, String idTransaction, String locationId, String request)
	{
		LyEnvironment environment = application.getEnvironment();

		if (environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
			environment.logOut("[TRANSACTION] update : transaction=" + idTransaction + " location="
				+ locationId + " request=" + request);
	}

	/**
	 * Method called before a transaction should be committed on a location.<br />
	 * <i>The default implementation</i> print a trace about the transaction before commit if the
	 * environment {@link LyEnvironment#isDebugLevel(int) debug level} is
	 * {@link LyEnvironment#DEBUG_TRANSACTION}.
	 *
	 * @param  application    Current application
	 * @param  idTransaction  Id of the transaction
	 * @param  locationId     Id of the location
	 */
	public void beforeCommit(LyApplication application, String idTransaction, String locationId)
	{
		LyEnvironment environment = application.getEnvironment();

		if (environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
			environment.logOut("[TRANSACTION] beforeCommit : transaction=" + idTransaction
				+ " location=" + locationId);
	}

	/**
	 * Method called after a transaction has been comitted on a location.<br />
	 * <i>The default implementation</i> print a trace about the transaction after commit if the
	 * environment {@link LyEnvironment#isDebugLevel(int) debug level} is
	 * {@link LyEnvironment#DEBUG_TRANSACTION}.
	 *
	 * @param  application    Current application
	 * @param  idTransaction  Id of the transaction
	 * @param  locationId     Id of the location
	 */
	public void afterCommit(LyApplication application, String idTransaction, String locationId)
	{
		LyEnvironment environment = application.getEnvironment();

		if (environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
			environment.logOut("[TRANSACTION] afterCommit : transaction=" + idTransaction
				+ " location=" + locationId);
	}

	/**
	 * Method called before a transaction should be rollbacked on a location.<br />
	 * <i>The default implementation</i> print a trace about the transaction before roolback if the
	 * environment {@link LyEnvironment#isDebugLevel(int) debug level} is
	 * {@link LyEnvironment#DEBUG_TRANSACTION}.
	 *
	 * @param  application    Current application
	 * @param  idTransaction  Id of the transaction
	 * @param  locationId     Id of the location
	 */
	public void beforeRollback(LyApplication application, String idTransaction, String locationId)
	{
		LyEnvironment environment = application.getEnvironment();

		if (environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
			environment.logOut("[TRANSACTION] beforeRollback : transaction=" + idTransaction
				+ " location=" + locationId);
	}

	/**
	 * Method called after a transaction has been rollbacked on a location.<br />
	 * <i>The default implementation</i> print a trace about the transaction after rollback if the
	 * environment {@link LyEnvironment#isDebugLevel(int) debug level} is
	 * {@link LyEnvironment#DEBUG_TRANSACTION}.
	 *
	 * @param  application    Current application
	 * @param  idTransaction  Id of the transaction
	 * @param  locationId     Id of the location
	 */
	public void afterRollback(LyApplication application, String idTransaction, String locationId)
	{
		LyEnvironment environment = application.getEnvironment();

		if (environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
			environment.logOut("[TRANSACTION] aferRollback : transaction=" + idTransaction
				+ " location=" + locationId);
	}

	/**
	 * Method called after a transaction was successfully completed (commit and notification done).
	 * <br />
	 * <i>The default implementation</i> print a trace about transactions of the given list ending
	 * if the environment {@link LyEnvironment#isDebugLevel(int) debug level} is
	 * {@link LyEnvironment#DEBUG_TRANSACTION}.
	 *
	 * @param  application     the current application
	 * @param  idTransactions  List of involved transaction ids
	 * @param  locationIds     list of involved locations ids
	 */
	public void transactionCompleted(LyApplication application, Vector<String> idTransactions,
		Vector<String> locationIds)
	{
		LyEnvironment environment = application.getEnvironment();

		if (environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
		{
			int i, n = idTransactions.size();

			for (i = 0; i < n; i++)
				environment.logOut("[TRANSACTION] transactionCompleted : transaction="
					+ idTransactions.get(i) + " location=" + locationIds.get(i));
		}
	}

	/**
	 * Method used by abstract classes to define which subclass has to be loaded.<br />
	 * This implementation does nothing and return null.
	 *
	 * @param   application  current application.
	 * @param   objectId     the object identifier to be loaded
	 * @param   values       optional additional values we might know about the object : this
	 *                       parameter is usually null.
	 *
	 * @return  the list of classes if found or null to use generic processing. Note : an empty list
	 *          aborts generic processing (not found in any subclass)
	 */
	public LyInfoList<LyClassInfo> getSubClasses(LyApplication application, LyObjectId objectId, LyValueSet values)
	{
		return null;
	}

	/**
	 * Method used to find logical classes of a peer object identifier.<br />
	 * This implementation does nothing and return null.
	 *
	 * @param   application  current application.
	 * @param   objectId     an object identifier of a peer object (use
	 *                       {@link leon.info.LyClassInfo#getObjectClass()} to get the peer class :
	 *                       instance of {@link leon.data.LyClass})
	 * @param   values       optional additional values we might know about the object : this
	 *                       parameter is usually null.
	 *
	 * @return  the list of classes if found or null to use generic processing. Note : an empty
	 *          lists aborts generic processing (not found in any subclass)
	 */
	public LyInfoList<LyClassInfo> getLogicalClasses(LyApplication application, LyObjectId objectId, LyValueSet values)
	{
		return null;
	}

	/**
	 * Used to get an unique short value for the specified application.
	 *
	 * @param   application  current application
	 * @param   objectClass  the class for which a new object is created.
	 *
	 * @return  An unique value as short for the specified application.
	 */
	public short getNextShort(LyApplication application, LyClass objectClass)
	{
		return LyTools.nextShort();
	}

	/**
	 * Used to get an unique int value for the specified application.
	 *
	 * @param   application  current application
	 * @param   objectClass  the class for which a new object is created.
	 *
	 * @return  An unique value as int for the specified application.
	 */
	public int getNextInt(LyApplication application, LyClass objectClass)
	{
		return LyTools.nextInt();
	}

	/**
	 * Used to get an unique long value for the specified application.
	 *
	 * @param   application  current application
	 * @param   objectClass  the class for which a new object is created.
	 *
	 * @return  An unique value as long for the specified application.
	 */
	public long getNextLong(LyApplication application, LyClass objectClass)
	{
		return LyTools.nextLong();
	}

	/**
	 * Used to get an unique float value for the specified application.
	 *
	 * @param   application  current application
	 * @param   objectClass  the class for which a new object is created.
	 *
	 * @return  An unique value as float for the specified application.
	 */
	public float getNextFloat(LyApplication application, LyClass objectClass)
	{
		return LyTools.nextFloat();
	}

	/**
	 * Used to get an unique double value for the specified application.
	 *
	 * @param   application  current application
	 * @param   objectClass  the class for which a new object is created.
	 *
	 * @return  An unique value as double for the specified application.
	 */
	public double getNextDouble(LyApplication application, LyClass objectClass)
	{
		return LyTools.nextDouble();
	}

	/**
	 * Free this behavior.<br />
	 * This method is provided to allow behavior to unregister itself from object sources or
	 * workspaces if needed.<br />
	 * It should be called only by the controller that built this behavior.<br />
	 * This implementation does nothing.
	 */
	public void free()
	{
	}

	/**
	 * Load data for given classes.<br />
	 * This implementation does nothing.
	 *
	 * @param  application  Application for which classes are loaded.
	 * @param  classes      List of classes to load.
	 *
	 * @see    leon.app.LyApplication#loadData()
	 * @see    leon.info.LyClassInfo
	 */
	public void loadData(LyApplication application, LyInfoList<LyClassInfo> classes)
	{
	}

	/**
	 * Indicates whether data may be loaded / unloaded for given class info.
	 *
	 * @param classInfo class info for which data may be loaded / unloaded
	 * @return true if data may be loaded / unloaded
	 */
	public boolean enableLoadUnloadData(LyClassInfo classInfo)
	{
		return true;
	}

    /**
     * Method for parsing extra information such as charts or preferences.
     *
     * @param   application  Application for which classes are loaded.
     * @param   prefType     identifies the preference type : "preference", "chart", "sticker" or
     *                       "treemap_configuration"
     * @param   userId       User id (may be null when info is shared).
     *
     * @return  List of loaded infos.
     */
    public LyInfoList<LyInfo> parseExtraFromDB(LyApplication application, String prefType, String userId)
    {
        LyInfoList<LyInfo>            infos = new LyInfoList<LyInfo>();
        Vector<LyExtraInfoInterface>  extraInfos = loadExtraInfosFromDB(application, prefType, userId);

        if (extraInfos != null)
        {
            int   i, n = extraInfos.size();

            for (i = 0; i < n; i++)
                infos.addInfo((LyInfo)extraInfos.get(i));
        }

        return infos;
    }

	/**
	 * Save given extra info like preferences, charts or treemap configurations.
	 *
	 * @param   application  Application on which extra info is saved.
	 * @param   info         Info to be saved.
	 *
	 * @return  True if extra info was successfully saved.
	 */
	public boolean saveExtraInfo(LyApplication application, LyExtraInfoInterface info)
	{
		return saveExtraInfo(application, info, null);
	}

    /**
     * Save given extra info like preferences, charts or treemap configurations.
     * @param application Application on which extra info is saved.
     * @param info        Info to be saved.
     * @param writer	  Writer interface.
     * @return True if extra info was successfully saved.
     */
    public boolean saveExtraInfo(LyApplication application, LyExtraInfoInterface info,
    	LyXmlWriterInterface writer)
	{
    	LySqlDataProvider  prefDataProvider = application.getPrefDataProvider();

		if (prefDataProvider == null)
			return info.saveFile(writer);
		else
		{
			LySqlConnection sqlConn = (LySqlConnection)prefDataProvider.getConnection();

			try
			{
				// build and send QUERY
				String   request;
				Object[] params;
				String   prefType = "unknown";

				if (info instanceof LyPreference)
					prefType = "preference";
				else if (info instanceof LyChartInfo)
					prefType = "chart";
				else if (info instanceof LyTreemapConfiguration)
					prefType = "treemap_configuration";
				else if (info instanceof LyFormSave)
				    prefType = ((LyFormSaveFile)info).getId();
				else if (info instanceof LySticker)
					prefType = "sticker";

				String dir = application.getEnvironment().getFile("LY_PREFERENCE_DIR");

				if ((dir != null)
						&& ("LY_PREFERENCE_DIR".equals(dir) || (dir.length() == 0)))
					dir = null;

				File dirFile = null;

				if (dir != null)
				{
					dirFile = new File(dir);
					if (!dirFile.exists())
						dirFile.mkdirs();
				}

				File			   tmpFile = File.createTempFile("loadPref", ".xml", dirFile);
				FileOutputStream   output = new FileOutputStream(tmpFile);
				LyXmlPrintWriter   pw = new LyXmlPrintWriter(output, LyString.DEFAULT_CHARSET);

				if (writer == null)
					info.writeXml(pw, true, true);
				else
					writer.writeXml(pw, info);

				pw.close();

				InputStream iStream = new FileInputStream(tmpFile);

				byte[] dataFull = new byte[(int)tmpFile.length()];

				for (int i = 0, n = (int)tmpFile.length(); i < n; i++)
					dataFull[i] = (byte)iStream.read();

				// getUserId only (without class)
				String userId = info.getUser();

				if ((userId != null) && (userId.indexOf(leon.data.LyObjectId.CLASS_SEP) > 0))
					userId = userId.substring(userId.indexOf(leon.data.LyObjectId.CLASS_SEP) + 1);

				if (info.getFile() == null)
				{
					Integer id = new Integer(LyTools.nextInt());

					request = "insert into " + PREF_TABLE + " (" + PREF_ID_COL + ", " + PREF_TYPE_COL + ", "
						+ PREF_DATE_COL + ", " + PREF_USER_COL + ", " + PREF_BLOB_COL + ") values (?,?,?,?,?)";
					params = new Object[]
						{
							id, prefType, new java.sql.Date(System.currentTimeMillis()), userId, dataFull
						};
					info.setFile(id.toString());
				}
				else
				{
					request = "update " + PREF_TABLE + " set " + PREF_TYPE_COL + "=?, " + PREF_DATE_COL + "=?, "
						+ PREF_USER_COL + "=?, " + PREF_BLOB_COL + "=? where " + PREF_ID_COL + "=?";
					params = new Object[]
						{
							prefType, new java.sql.Date(System.currentTimeMillis()), userId, dataFull,
							LyTools.IntegerFromString(info.getFile())
						};
				}

				Connection		  cnx = sqlConn.getConnection();
				PreparedStatement prepStat = cnx.prepareStatement(request);

				// Set parameters
				for (int i = 0; (params != null) && (i < params.length); i++)
				{
					if (params[i] == null)
						prepStat.setNull(i + 1, java.sql.Types.VARCHAR);
					else if (params[i] instanceof InputStream)
						prepStat.setBinaryStream(i + 1, (InputStream)params[i], (int)tmpFile.length());
					else if (params[i] instanceof byte[])
						prepStat.setBytes(i + 1, (byte[])params[i]);
					else
						prepStat.setObject(i + 1, params[i]);
				}
				// execute SQL update
				int resu = prepStat.executeUpdate();

				// close all
				iStream.close();
				tmpFile.delete();
				return (resu == 1);
			}
			catch (Exception sqle)
			{
				application.getEnvironment().writeException(sqle);
			}
			finally
			{
				prefDataProvider.releaseConnection(sqlConn);
			}
		}

		return false;
	}

    /**
     * Load extra info like preferences, charts, treemap configurations or form saves.
     * @param   application  Application on which extra info is saved.
     * @param   type Extra info type.
     * @param   id Id of saved info.
     * @return  Info that was saved.
     */
    public Vector<LyExtraInfoInterface> loadExtraInfosFromDB(LyApplication application, String prefType, String userId)
    {
        // create table leon_preferences (pref_id int, pref_type varchar(254), pref_time timestamp, pref_user varchar(255),  pref_blob  BLOB, PRIMARY KEY(pref_id));
    	LySqlDataProvider  prefDataProvider = application.getPrefDataProvider();

        if (prefDataProvider == null)
            return null;

        LySqlConnection                 sqlConn = (LySqlConnection)prefDataProvider.getConnection();

        if (sqlConn == null)
            return null;

        ResultSet                       rs = null;
        Vector<LyExtraInfoInterface>    extraInfos = new Vector<LyExtraInfoInterface>();
        LyExtraInfoInterface            extraInfo;
        int                             i, n = 0;

        if ((userId != null)
                && (userId.indexOf(leon.data.LyObjectId.CLASS_SEP) > 0))
            userId = userId.substring(userId.indexOf(leon.data.LyObjectId.CLASS_SEP) + 1);

        try
        {
            // build and send QUERY
            String            request = "select " + PREF_BLOB_COL + ", " + PREF_ID_COL
                    + " from " + PREF_TABLE + " where "
                    + PREF_USER_COL + ((userId == null) ? " is null" : ("='" + userId + "'"))
                    + " and " + PREF_TYPE_COL  + "='" + prefType + "'";
            Connection        cnx = sqlConn.getConnection();
            PreparedStatement prepStat = cnx.prepareStatement(request);

            rs = prepStat.executeQuery();

            while (rs.next())
            {
                int         id = rs.getInt(PREF_ID_COL);
                boolean     loaded;

                // Check if this file was already loaded
                for (i = 0, loaded = false; (i < n) && !loaded; i++)
                {
                    extraInfo = extraInfos.get(i);
                    if ((extraInfo.getFile() != null)
                            && extraInfo.getFile().equals(Integer.toString(id)))
                        loaded = true;
                }

                // Already loaded, skip.
                if (loaded)
                    continue;

                // Extract BLOB to a temp file
                String dir = application.getEnvironment().getFile("LY_PREFERENCE_DIR");

                if ((dir != null) && ("LY_PREFERENCE_DIR".equals(dir) || (dir.length() == 0)))
                    dir = null;

                File dirFile = null;

                if (dir != null)
                {
                    dirFile = new File(dir);
                    if (!dirFile.exists())
                        dirFile.mkdirs();
                }

                File tmpFile = File.createTempFile("loadPref", ".xml", dirFile);

                // Parse XML Result
                try
                {
                    InputStream  	in = rs.getBinaryStream(PREF_BLOB_COL);

                    if (in == null)
                        continue;

                    OutputStream 	out = new BufferedOutputStream(new FileOutputStream(tmpFile));
                    int          	nr = in.read();

                    while (nr >= 0)
                    {
                        out.write(nr);
                        nr = in.read();
                    }

                    in.close();
                    out.close();

                    try
                    {
	                    LyInfoParser	infoParser = new LyInfoParser(application, tmpFile.getAbsolutePath());
	                    Object          root = infoParser.getRootElement();

	                    if (root instanceof LyExtraInfoInterface)
	                        extraInfo = (LyExtraInfoInterface)root;
	                    else if (root instanceof LyFormSave)
	                        extraInfo = new LyFormSaveFile(null, userId, (LyFormSave)root);
	                    else
	                        continue;

	                    extraInfo.setFile(Integer.toString(id));
	                    extraInfos.add(extraInfo);
                    }
                    catch (Exception e)
                    {
                        application.getEnvironment().writeException(e);
                    	handleExtraInfoException(application, Integer.toString(id));
                    }
                }
                catch (Exception e)
                {
                    application.getEnvironment().writeException(e);
                }
                finally
                {
                	if (tmpFile != null)
                	{
                		tmpFile.delete();
                		tmpFile = null;
                	}
                }
            }
        }
        catch (Exception sqle)
        {
            application.getEnvironment().writeException(sqle);
            return null;
        }
        finally
        {
            try
            {
            	if (rs != null)
            		rs.close();
            }
            catch (SQLException e)
            {
            }

            prefDataProvider.releaseConnection(sqlConn);
        }

        return extraInfos;
    }

    /**
     * Method that handles exception on extra info loading.
     */
    protected void handleExtraInfoException(LyApplication application, String extraInfoId)
    {
    	// Be careful when launching previous versions when preferences grammar has evolved
    	// deleteExtraInfoFromDB(application, extraInfoId);
    }

	/**
	 * Delete given extra info like preferences, charts or treemap configurations.
	 *
	 * @param   application  Application on which extra info is deleted.
	 * @param   info         Info to be deleted.
	 *
	 * @return  True if extra info was successfully deleted.
	 */
	public boolean deleteExtraInfo(LyApplication application, LyExtraInfoInterface info)
	{
    	LySqlDataProvider  prefDataProvider = application.getPrefDataProvider();

		if (prefDataProvider == null)
			return info.deleteFile();
		else
		{
			String	extraInfoId = info.getFile();

			if (extraInfoId != null)
				return deleteExtraInfoFromDB(application, extraInfoId);
		}

		return false;
	}

	/**
	 * Delete given extra info like preferences, charts or treemap configurations from database.
	 *
	 * @param   application  Application on which extra info is deleted.
	 * @param   extraInfoId  Id of the extra info to be deleted.
	 *
	 * @return  True if extra info was successfully deleted.
	 */
	public boolean deleteExtraInfoFromDB(LyApplication application, String extraInfoId)
	{
    	LySqlDataProvider	prefDataProvider = application.getPrefDataProvider();

    	if (prefDataProvider == null)
    		return false;

		LySqlConnection		sqlConn = (LySqlConnection)prefDataProvider.getConnection();

		try
		{
			// build and send QUERY
			String     request = "delete from " + PREF_TABLE + " where " + PREF_ID_COL + "=" + extraInfoId;
			Connection cnx = sqlConn.getConnection();
			Statement  stat = cnx.createStatement();
			int 	   count = stat.executeUpdate(request);

			return (count == 1);
		}
		catch (Exception sqle)
		{
			application.getEnvironment().writeException(sqle);
		}
		finally
		{
			prefDataProvider.releaseConnection(sqlConn);
		}

		return false;
	}

	/**
     * Get the event manager for the given application.
     * This implementation returns null.
     * @param application the current application
     * @return an event manager. If null is returned, default event manager will be instanciated.
     */
	public LyEventManager getEventManager(LyApplication application)
	{
		return null;
	}

} // end class LyDataApplicationBehavior
