/*
 * LyDataSessionBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.peer.*;


/**
 * This class implements the default behavior of a session.
 *
 * @author   Lyria S.A.
 * @version  $Id: LyDataSessionBehavior.java 33348 2011-11-30 16:18:01Z bangiole $
 * @see      leon.app.LyDataSession
 * @see      leon.app.behaviorinterface.LyDataSessionBehaviorInterface
 */
public class LyDataSessionBehavior
    implements LyDataSessionBehaviorInterface
{
    /** File version. */
    public static final String __VERSION = "$Id: LyDataSessionBehavior.java 33348 2011-11-30 16:18:01Z bangiole $";

    /**
     * Constructor.<br />
     * Does nothing.
     */
    public LyDataSessionBehavior()
    {
    }

    /**
     * Method called when a session is started.<br />
     * <i>The default implementation</i> checks that the session application has a root action and
     * {@link LyDataSession#setValid(boolean) validate} the session if the root action is not a
     * login action.
     *
     * @param  session  the started session.
     */
    public void start(LyDataSession session)
    {
        LyApplication application = session.getApplication();
        LyProject     project = application.getProject();
        LyAction      rootAction = project.getRootAction();

        if (rootAction == null)
            throw new IllegalStateException("BF AppEngine : application " + application.getId()
                + " has no root action");

        rootAction = rootAction.getRoot();

        // If root action is not a login action, validate the session.
        if (rootAction != application.getInfo("_login"))
            session.setValid(true);
    }

    /**
     * Method called when a session is closed.<br />
     * <i>The default implementation</i> does nothing.
     *
     * @param  session  the closed session.
     */
    public void close(LyDataSession session)
    {
    }

    /**
     * Gets user of this session.
     *
     * @param   session  Current session.
     *
     * @return  The object representing current user of the session.
     */
    public LyObject getUser(LyDataSession session)
    {
        return null;
    }

    /**
     * Method called when an exception occurred during the session processing.<br />
     * <i>The default implementation</i> writes the exception and
     * {@link leon.peer.LyTransactionManager#abort(LyDataSession) abort} transactions of the
     * session.
     *
     * @param  session  the current session
     * @param  t        The throwable instance that stopped the processing.
     */
    public void handleException(LyDataSession session, Throwable t)
    {
        // Dump exception to stderr
        LyEnvironment environment = session.getEnvironment();

        if (environment != null)
            environment.writeException(t);
        else
            t.printStackTrace();

        LyApplication        application = session.getApplication();
        LyTransactionManager tm = application.getTransactionManager();

        tm.abort(session);
    }

    /**
     * Free this behavior.<br />
     * This method is provided to allow behavior to unregister itself from object sources or
     * workspaces if needed.<br />
     * It should be called only by the controller that built this behavior.<br />
     * <i>The default implementation</i> does nothing.
     */
    public void free()
    {
    }

    /**
     * Method called when a context needs to control objects before commit.<br />
     * Result should be one of the following level:
     *
     * <ul>
     * <li>{@link leon.data.LyObject#LEVEL_ERROR}</li>
     * <li>{@link leon.data.LyObject#LEVEL_OK}</li>
     * <li>{@link leon.data.LyObject#LEVEL_WARNING}</li>
     * </ul>
     * <br />
     * <i>The default implementation</i> does nothing and return LyObject.LEVEL_OK.
     *
     * @param   session  the current session
     * @param   objects  the objects to control.
     *
     * @return  Result of calling controls.
     */
    public short controlBeforeCommit(LyDataSession session, Collection<LyObject> objects)
    {
        return LyObject.LEVEL_OK;
    }

    /**
     * /** Method called when the undo/redo status has changed.<br />
     * The status is available on the context manager of the application.<br />
     * <i>The default implementation</i> does nothing.
     *
     * @param  session  the current session.
     */
    public void updateUndoRedo(LyDataSession session)
    {
    }

    /**
     * Method called when a connection with a server
     * ({@link leon.info.location.LyLocation LyLocation} or
     * {@link leon.info.notifier.LyNotifier LyNotifier}) is lost.
     *
     * @param  session  Current session.
     * @param  server   Logical server definition
     */
    public void connectionLost(LyDataSession session, LyInfo server)
    {
        // nothing to do
    }

    /**
     * Retrieves a value from a registered key.
     *
     * @param   key  The key that must be used.
     *
     * @return  The value stored with the key or null if the key is not registered.
     */
    public Object getKeyValue(String key)
    {
        return null;
    }

    /**
     * Get the reference unit for the given field for the current session
     * 
     * @param session Current session
     * @param fieldInfo a field
     * @return a reference unit. Can be null.
     */
    public LyUnit getDefaultUnitForField(LyDataSession session, LyFieldInfo fieldInfo)
    {
    	return null;
    }
} // end class LyDataSessionBehavior
