/*
 * LyEditTableBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.view.*;


/**
 * This class implements the default behavior for an editable table controller.
 *
 * @author  Lyria S.A.
 * @version $Id: LyEditTableBehavior.java 29140 2011-01-12 10:23:03Z jlsaouli $
 * @see     leon.control.LyEditTableController
 * @see     leon.app.behavior.LyTableBehavior
 */
public class LyEditTableBehavior extends LyTableBehavior
	implements LyEditTableBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyEditTableBehavior.java 29140 2011-01-12 10:23:03Z jlsaouli $";

	/**
	 * Default constructor.
	 */
	public LyEditTableBehavior()
	{
	}

	/**
	 * Updates the label of the given cell for the given object and fieldinfo.<br />
	 * <i>The default implementation</i> updates the given cell label if needed, the cell object id
	 * if the given field is a relation field and if the view manager manages urls. If the given
	 * field is a file field, updates the cell label, Hlink and image.
	 *
	 * @param tableController the table controller that requires the cell label
	 * @param cell            the cell to update
	 * @param object          the object from which the field info value is represented in the table
	 *                        cell
	 * @param fieldInfo       the field info for which a cell label is required
	 */
	public void updateCell(LySimpleTableController tableController, LyCell cell, LyObject object,
		LyFieldInfo fieldInfo)
	{
		if (!(tableController instanceof LyEditTableController))
		{
			super.updateCell(tableController, cell, object, fieldInfo);
			return;
		}
		
		LyEditTableController		controller = (LyEditTableController)tableController;
		
		if (controller.isEditable(object, fieldInfo)
				&& controller.canBeEditable(fieldInfo)
				&& controller.isDynamicEditionMode())
			cell.setEditable(true);
		else
			cell.setEditable(false);
		
		LyEditTableSetController	setController = controller.getSetController(object);

		cell.setColor(null);
		
		if (controller.isDynamicEditionMode()
				&& (setController != null)
				&& setController.wasChanged()
				&& setController.getChangedFields().contains(fieldInfo)
				&& (!(fieldInfo instanceof LyChoiceFieldInfo)
						|| !fieldInfo.isMultiple()))
			cell.setColor(controller.getModificationsColor());

		super.updateCell(tableController, cell, object, fieldInfo);
	} // end method updateCell
	
	/**
     * Gets the cell label representing the value of the given field info for the given object.
     *
     * @param      controller Table controller that requires the cell label.
     * @param      object     Object from which the field info value is represented in the table
     *                        cell.
     * @param      fieldInfo  Field info for which a cell label is required.
     * @return     Label that will be put in the table cell for the given object and field info (may
     *             be null).
     */
    @SuppressWarnings("deprecation")
    public String getCellLabel(LySimpleTableController tableController, LyObject object,
        LyFieldInfo fieldInfo)
    {
        if (!(tableController instanceof LyEditTableController))
            return super.getCellLabel(tableController, object, fieldInfo);
        
        LyEditTableController       editTableController = (LyEditTableController)tableController;
        LyEditTableSetController    setController = editTableController.getSetController(object);
        
        if (setController == null)
            return super.getCellLabel(tableController, object, fieldInfo);
        
        LyValue                     formValue = setController.getFieldValue(fieldInfo), value = null;

        if (object != null)
        {
            // In dynamic edition mode, if cell is not being edited, retrieve
            // object value from setController then compute the label to be displayed.
            if (editTableController.isDynamicEditionMode()
                && setController.wasChanged()
                && setController.getChangedFields().contains(fieldInfo))
            {
                Object objectValue = ((object == null) ? null : object.getValue(fieldInfo));

                if (!(fieldInfo instanceof LyChoiceFieldInfo)
                    || !fieldInfo.isMultiple())
                {
                    if (((formValue == null) && (objectValue != null))
                            || ((formValue != null) && (objectValue == null))
                            || ((formValue != null) && !formValue.equals(objectValue)))
                        value = formValue;
                }
            }
            else if (!editTableController.isEditable(object, fieldInfo)
                || !editTableController.canBeEditable(fieldInfo))
            {
                // Case of multiple choice field info
                if (!(fieldInfo instanceof LyChoiceFieldInfo)
                        || !((LyChoiceFieldInfo)fieldInfo).isMultiple()
                        || !editTableController.canBeEditable(fieldInfo))
                    value = formValue;
            }
        }
        else
        {
            // Case of multiple choice field info
            if (!(fieldInfo instanceof LyChoiceFieldInfo)
                    || !((LyChoiceFieldInfo)fieldInfo).isMultiple()
                    || !editTableController.canBeEditable(fieldInfo))
                value = formValue;
        }
        
        if (value != null)
            return getCellLabel(tableController, object, fieldInfo, value.getValue());
        else
            return super.getCellLabel(tableController, object, fieldInfo);
    }
    
	/**
	 * Indicates if tools should be displayed at the end of an edit table row.<br />
	 * These tools allow the end-user to cancel or validate modifications made on a single line.
	 * <i>The default implementation</i> returns true.
	 * 
	 * @param tableController the table controller
	 * @return true is tools should be displayed, false otherwise.
	 */
	public boolean displayRowTools(LyEditTableController tableController)
	{
		return true;
	}
} // end class LyTableBehavior
