/*
 * LyExportBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import java.io.*;
import java.util.*;

import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;


/**
 * Default implementation of the export behavior.
 *
 * @author  Lyria S.A.
 * @version $Id: LyExportBehavior.java 40676 2012-09-05 16:27:18Z jlsaouli $
 * @see     leon.control.LyExportController
 */
public class LyExportBehavior extends LyActionBehavior
	implements LyExportBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyExportBehavior.java 40676 2012-09-05 16:27:18Z jlsaouli $";

    /**
     * Return the lines that will be written in the export file.
     *
     * @param  controller the current export controller
     * @param  objects    objects to export.
     * @param  fields     list of fieldInfos that should be exported to the file.
     * @param  fieldSep   separator between fields (columns)
     * @param  valueSep   separator between multiple values
     * @return the lines that will be written in the export file, null for standard export.
     * @param  exportType type of data to export : Possible values are SELECT_EXPORT_TYPE_TEXT
     *                    (.txt), SELECT_EXPORT_TYPE_RAW (.res) or SELECT_EXPORT_TYPE_SPREADSHEET
     *                    (.csv).
     */
    public Vector<String> getExportLines(LyExportController controller, Iterator<LyObject> objects,
        Vector<LyFieldInfo> fields, String fieldSep, String valueSep, String exportType)
    {
        return null;
    }
    
	/**
	 * Saves lines from the given list into the given file.
	 *
	 * @param  controller the current export controller
	 * @param  file       the file where data must be saved
	 * @param  lines      the list of lines to save
	 * @return true if file is save, false otherwise
	 */
	public boolean saveFile(LyExportController controller, File file, Vector<String> lines)
	{
		// Open the file
		Writer		  writer;
		LyEnvironment environment = controller.getEnvironment();
		OutputStream  out = null;

		try
		{
			if (file.exists())
				file.delete();

			out = new FileOutputStream(file);
			out = new BufferedOutputStream(out);
			String charset = controller.getExportCharset();
			
			if (charset == null)
			{
				if (controller.getExportType() == LyExportController.SELECT_EXPORT_TYPE_XML)
					charset = "UTF-8";
				else
					charset = "ISO-8859-1";
			}
			writer = new OutputStreamWriter(out, charset);
		}
		catch (IOException e)
		{
			environment.logError("Unable to open file :" + file + ". Cause : " + e.getMessage());
			if (!controller.isSilent())
				controller.showError(environment.getMessage("LY_ER_OPEN_FILE", file));
			return false;
		}

		environment.logInfo("Saving file :" + file);

		// Write the file
		PrintWriter ps = new PrintWriter(writer);
		int		    i, nl = lines.size();
		String	    line;

		for (i = 0; i < nl; i++)
		{
			line = lines.elementAt(i);
			ps.println(line);
		}

		ps.close();
		return true;
	} // end method saveFile
    
    /**
     * Returns the header mark, as a String, to be used at the beginning of header line<br />.
     * Default implementation returns the <code>LyString.EXPORT_HEADER_MARK</code> constant.
     * 
     * @param controller the export controller
     * @return the header mark
     */
    public String getHeaderMark(LyExportController controller)
    {
    	return controller.getEnvironment().getString().EXPORT_HEADER_MARK;
    }

    /**
     * Returns the default separator, as a String, of the separator to be used
     * between attributes. Used when the field is not available in the dialog.
     * 
     * @return the default separator for attributes
     */
    public String getDefaultAttributeSep()
    {
        return null;
    }

    /**
     * Returns the default separator, as a String, of the separator to be used
     * between attribute values. Used when the field is not available in the
     * dialog.
     * 
     * @return the default separator for attribute values
     */
    public String getDefaultAttributeValSep()
    {
        return null;
    }

    /**
     * Returns the default export type. Used when the field is not available in
     * the dialog.
     * 
     * @return the default separator
     */
    public String getDefaultExportType()
    {
        return LyExportController.SELECT_EXPORT_TYPE_SPREADSHEET;
    }

   /**
    * This method is called at the end of the export action.
    * In the web mode, it displays a new window from
    * which the resulting exported file can be downloaded.
    * 
    * @param controller The current export controller
    * @param fileName The neame of the exported file name
    */
   public void showExportedFileWindow(LyExportController controller, String fileName)
   {
        LyViewManager viewManager = controller.getViewManager();
        
        if ((viewManager != null)
            && !viewManager.hasProperty("LOCAL_FILE"))
        {
        	// Show a link to get the file
			String fileDir = viewManager.getDownloadAction(); 
			String filePath = viewManager.getTemporaryDir() + "/" + fileName;
			File   f = new File(filePath);
			String label = f.getName();
			String id = controller.getSession().storeTempFile(f.getAbsolutePath(), label);
			String url = fileDir + id;
           
			viewManager.showFile(null, url);
        }
   }

   /**
    * Indicates if the export form is displayed.<br />
    * Export form should not be displayed if all the fields values are given as
    * parameters of the export action.
    * 
    * @param controller the export controller
    * @return true if form is displayed, false otherwise.
    */
   public boolean isFormDisplayed(LyExportController controller) 
   {
	   return true;
   }
	
} // end class LyExportBehavior
