/*
 * LyImportBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import java.util.*;

import leon.app.LySession;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.control.content.LyImportContentController;
import leon.data.LyObject;
import leon.data.LyValue;
import leon.info.*;
import leon.misc.*;


/**
 * This class implements the default behavior for actions that are controlled by an import
 * controller.
 *
 * @author  Lyria S.A.
 * @version $Id: LyImportBehavior.java 34528 2012-03-13 12:42:36Z jlsaouli $
 * @see     leon.control.LyImportController
 */
public class LyImportBehavior extends LyActionBehavior
	implements LyImportBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyImportBehavior.java 34528 2012-03-13 12:42:36Z jlsaouli $";

	/**
	 * Checks if the import can be done with the given list of selected fields.<br />
	 * <i>The default implementation</i> checks that for the controller class info fields, all
	 * {@link leon.info.LyClassInfo#getIdFieldInfos() fields with the mark ID} are in the given list
	 * (if the {@link LyImportController#getImportType() controller import type} is not {@link
	 * LyImportController#SELECT_IMPORT_TYPE_SPREADSHEET}).
	 *
	 * @param  controller the current import controller
	 * @param  fields     the fields we want to check
	 * @return an error message if the import can not be done with the given selected files, null if
	 *         everything is fine
	 */
	public String checkFields(LyImportController controller, LyInfoList<LyFieldInfo> fields)
	{
		LyEnvironment environment = controller.getEnvironment();
		LyClassInfo   classInfo = controller.getClassInfo();
		String		  message = null;
		LyInfoList<LyFieldInfo>    idFields = classInfo.getIdFieldInfos();
		LyInfoList<LyFieldInfo>    nameFields = classInfo.getNameFieldInfos();
		int			  i, nFields = idFields.getSize();
		LyFieldInfo   fieldInfo;
		String		  importType = controller.getImportType();
		boolean		  allIdFields = true;

		for (i = 0; i < nFields; i++)
		{
			// Check if the id fields are selected
			fieldInfo = idFields.getInfo(i);
			if (!fields.contains(fieldInfo))
			{
				allIdFields = false;

				if (!LyImportController.SELECT_IMPORT_TYPE_SPREADSHEET.equals(importType))
					message = environment.getMessage("LY_ER_IMPORT_ID_FIELD", fieldInfo.getName(),
							classInfo.getName());
			}
		}

		// If the only selected operation is modification, check if at least all the id fields
		// or all the name fields are selected
		BitSet operations = controller.getImportOperations();

		if (!allIdFields
				&& LyImportController.SELECT_IMPORT_TYPE_SPREADSHEET.equals(importType)
				&& (operations != null)
				&& operations.get(LyImportController.SELECT_OPERATIONS_MODIFICATION)
				&& !operations.get(LyImportController.SELECT_OPERATIONS_CREATION)
				&& !operations.get(LyImportController.SELECT_OPERATIONS_DELETION))
		{
			nFields = nameFields.getSize();

			for (i = 0; i < nFields; i++)
			{
				// Check if the name fields are selected
				fieldInfo = nameFields.getInfo(i);

				if (!fields.contains(fieldInfo))
					message = environment.getMessage("LY_ER_IMPORT_NAME_FIELD",
							classInfo.getName());
			}
		}

		return message;
	} // end method checkFields


    /**
     * This is called by the LyImportTableController at import time.
     * Checks if the provided value for a fields is correct. Additional checks
     * can be performed here, provided an imported field value.
     * This generic behavior does nothing.
     * 
     * @param ctrl the initiating  LyImportTableController
     * @param importedObject the imported object
     * @param realObject the existing object if any, null otherwise
     * @param fieldInfo the LyFieldInfo
     * @param fieldValue the imported value to check
     * @return the error message, or null if everything is fine.
     */
    public String checkImportedObjectField(LyImportTableController ctrl, 
            LyObject importedObject, LyObject realObject, LyFieldInfo fieldInfo, 
            LyValue fieldValue)
    {
        return null;
    }
    
    /**
     * This is called by the LyImportTableController when import is finished
     * (and before commit is required). This is where additional messages can be
     * displayed to the user. This generic behavior does nothing.
     * 
     * @param ctrl
     *            the initiating LyImportTableController
     * @param nSuccess
     *            the number of successfully imported objects
     */
    public void importObjectsTerminated(LyImportTableController ctrl,
            int nSuccess)
    {
    }
    
    /**
     * This is called by the LyImportTableController when import is started.
     * This generic behavior does nothing.
     * 
     * @param ctrl
     *            the initiating LyImportTableController
     * @param nRequested
     *            the number of objects to be imported
     */
    public void importObjectsStarted(LyImportTableController ctrl,
            int nRequested)
    {
    }
    
    /**
     * This is called by the LyImportTableController when import is committed.
     * This is where additional messages can be displayed to the user. This
     * generic behavior does nothing.
     * 
     * @param ctrl
     *            the initiating LyImportTableController
     * @param nSuccess
     *            the number of successfully committed objects
     */
    public void importObjectsCommitted(LyImportTableController ctrl,
            int nSuccess)
    {
    }

    /**
     * Returns the default separator, as a String, of the separator to be used
     * between attributes. Used when the field is not available in the dialog.
     * 
     * @return the default separator for attributes
     */
    public String getDefaultAttributeSep()
    {
        return null;
    }

    /**
     * Returns the default separator, as a String, of the separator to be used
     * between attribute values. Used when the field is not available in the
     * dialog.
     * 
     * @return the default separator for attribute values
     */
    public String getDefaultAttributeValSep()
    {
        return null;
    }

    /**
     * Returns the default export type. Used when the field is not available in
     * the dialog.
     * 
     * @return the default separator
     */
    public String getDefaultImportType()
    {
        return LyImportController.SELECT_IMPORT_TYPE_XML;
    }

    /**
     * Returns the default operations. Used when the field is not available in
     * the dialog.
     * 
     * @return the default operations
     */
    public BitSet getDefaultOperations()
    {
        BitSet operations = new BitSet();

        operations.set(LyImportController.SELECT_OPERATIONS_CREATION);
        operations.set(LyImportController.SELECT_OPERATIONS_MODIFICATION);
        //operations.set(LyImportController.SELECT_OPERATIONS_DELETION);

        return operations;
    }

    /**
     * Returns the default start line number. Used when the field is not available in
     * the dialog.
     * 
     * @return the default start line number
     */
    public Integer getDefaultStartLineNumber()
    {
        return new Integer(0);
    }

    /**
     *  this method may be overridden to provide a custom class for encode / decode File content in import.
     *  Default implementation returns null if type is not found.
     * @param session user session
     * @param parent table to display results
     * @param action parent action
     * @param importType type of import requested by user
     * @return null if not supported type or the controller if available.
     */
	public LyImportContentController importContentFactory(LySession session, LyImportTableController parent, LyAction action, String importType)
	{
		LyImportContentController ctrl = null;
	
		if ((importType != null)
				&& importType.equals(LyImportController.SELECT_IMPORT_TYPE_XML))
			ctrl = parent.importContentXmlFactory(session, parent, action);
		
		return ctrl;
	}
} // end class LyImportBehavior
