/*
 * LyImportContentBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import leon.app.*;
import leon.control.*;
import leon.control.content.*;
import leon.data.*;
import leon.info.*;


/**
 * This class is used by the import action (parameter _importContentBehavior) to modify the content
 * which is imported during the import process.
 *
 * @author  Lyria S.A.
 * @version $Id: LyImportContentBehavior.java 29141 2011-01-12 11:22:10Z jlsaouli $
 */
public class LyImportContentBehavior extends LyBehavior
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyImportContentBehavior.java 29141 2011-01-12 11:22:10Z jlsaouli $";

	/**
	 * Construct a class behavior.
	 */
	public LyImportContentBehavior()
	{
	}

	/**
	 * Starts the import action for the given controller.<br />
	 * <i>The default implementation</i> does nothing and returns true.
	 *
	 * @param  controller the current Controller instance
	 * @return true if the action starts correctly, false otherwise.
	 */
	public boolean startImport(LyImportContentController controller)
	{
		return true;
	}

	/**
	 * Checks that the given imported value of the given field info is correct and modify it if
	 * necessary.<br />
	 * <i>The default implementation</i> does nothing and return the given value 'as is'.
	 *
	 * @param  controller the current Controller instance
	 * @param  classInfo  the field class info
	 * @param  fieldInfo  the imported field info
	 * @param  value      the imported value
	 * @return a correct field value
	 */
	public String getFieldValue(LyImportContentController controller, LyClassInfo classInfo,
		LyFieldInfo fieldInfo, String value)
	{
		return value;
	}

	/**
	 * Complete the import process : really create / set objects in current context.
	 *
	 * @param  controller      the current Controller instance
	 * @param  importedObjects object that where found in import, they are not yet created or setted
	 *                         for real
	 * @return the new list of imported objects which have been created/setted for real in the
	 *         application in current context
	 */
	public LyObjectList completeImport(LyImportContentController controller,
		LySimpleObjectList importedObjects)
	{
		LyObject		   importObject, realObject;
		LyObjectId		   realObjectId;
		LySimpleObjectList result = new LySimpleObjectList();
		LyContext		   context, subcontext = null;

		if (controller.getParent() instanceof LyControllerContextInterface)
			context = ((LyControllerContextInterface)controller.getParent())
					.getControllerContext();
		else
			context = null;

		int i, n1 = importedObjects.getSize();

		// **** Imported Objects **** Create them or set them for real
		for (i = 0; i < n1; i++)
		{
			importObject = importedObjects.getObject(i);

			importObject.resetObjectId();
			realObjectId = importObject.getObjectId();
			realObject = realObjectId.getLogicalObject();

			if (realObject == null)
			{
				// **** internalObjects Objects **** Create them or set them for
				// real.
				// A subcontext for each object in import
				subcontext = completeImportInternalObjects(controller, importObject, context, null);
				// *** internal objects are set now, lets go for the imported object

				// Actually create the object
				importObject.setTemporary(false);
				if (subcontext == null)
					subcontext = createSubcontext(controller, context);
				if (subcontext != null)
					subcontext.createObject(importObject);
				else
					importObject.create();

				importObject.control();

				result.addObject(importObject);
			}
			else
			{
				// **** internalObjects Objects **** Create them or set them for
				// real.
				// A subcontext for each object in import.
				subcontext = completeImportInternalObjects(controller, realObject, context, null);

				// *** internal objects are set now, lets go for the imported
				// object.
				// This object exists, sets it for real in current context.
				// We take all imported fields and do not import secret fields.
				LyValueSet values = getImportedValues(controller, importObject, false);

				if (!isEquals(values, realObject.getValues()))
				{
					if (subcontext == null)
						subcontext = createSubcontext(controller, context);
					realObject.skipEncoding(true);
					if (subcontext != null)
					{
						subcontext.setObject(realObject);
						subcontext.setValues(realObject, values);
					}
					else
					{
						realObject.set();
						realObject.setValues(values);
					}
					realObject.skipEncoding(false);
					realObject.control();
				}
				result.addObject(realObject);
			}
		}

		return result;
	} // end method completeImport

	/**
	 * Complete the import process : really create / set th given internal object in the given
	 * sub-context.
	 *
	 * @param  controller   the current Controller instance
	 * @param  importObject imported object from the import, not yet created or setted for real
	 * @param  context      current context for import
	 * @param  subcontext   subcontext (or null) for internal object
	 * @return the subcontext with the new object
	 */
	protected LyContext completeImportInternalObjects(LyImportContentController controller,
		LyObject importObject, LyContext context, LyContext subcontext)
	{
		LyObject     internalObject, realObject;
		LyObjectId   realObjectId;
		LyObjectList internalObjects;
		int			 j, n2;
		// **** internalObjects Objects **** Create them or set them for real
		internalObjects = controller.getInternalObjectList(importObject);

		n2 = ((internalObjects == null) ? 0 : internalObjects.getSize());

		for (j = 0; j < n2; j++)
		{
			internalObject = internalObjects.getObject(j);
			realObjectId = internalObject.getObjectId();
			realObject = realObjectId.getLogicalObject();

			// Recurse for strucs in structs ...
			subcontext = completeImportInternalObjects(controller, internalObject, context,
					subcontext);

			if (realObject == null)
			{
				internalObject.setTemporary(false);
				// This object does not exist, create it for real in current
				// context
				if (subcontext == null)
					subcontext = createSubcontext(controller, context);
				if (subcontext != null)
					subcontext.createObject(internalObject);
				else
					internalObject.create();
			}
			else
			{
				// This object exists, sets it for real in current context
				// We take all imported fields and do not import secret fields
				LyValueSet values = getImportedValues(controller, internalObject, true);

				if (!isEquals(values, realObject.getValues()))
				{
					if (subcontext == null)
						subcontext = createSubcontext(controller, context);
					realObject.skipEncoding(true);
					if (subcontext != null)
					{
						subcontext.setObject(realObject);
						subcontext.setValues(realObject, values);
					}
					else
					{
						realObject.set();
						realObject.setValues(values);
					}
					realObject.skipEncoding(false);
				}
			}
		}
		return subcontext;
	} // end method completeImportInternalObjects

	/**
	 * Creates a subcontext into the given context and return it.
	 *
	 * @param  controller the current import controller
	 * @param  context    the parent context
	 * @return the new subcontext
	 */
	protected LyContext createSubcontext(LyImportContentController controller, LyContext context)
	{
		LyContext subcontext;

		if (context != null)
		{
			// We create a SUBCONTEXT for each imported object (can cancel
			// object per object)
			LySession		 session = controller.getSession();
			LyContextManager manager = session.getContextManager();

			subcontext = manager.newContext(session, context.getType());
			subcontext.setInternal();
			// Will be commited or rollbacked with parent context
			context.addSon(subcontext);
		}
		else
			subcontext = null;

		return subcontext;
	}

	/**
	 * Compares the given source and target sets of values.
	 *
	 * @param  sourceValues the source values to compare in the set of targetValues
	 * @param  targetValues the targetValues as a reference, may contain more values than
	 *                      sourceValues
	 * @return true the source values are found into the target values, false otherwise
	 */
	protected boolean isEquals(LyValueSet sourceValues, LyValueSet targetValues)
	{
		if ((sourceValues == null) || (targetValues == null))
			return false;

		LyValue value1, value2;
		int     i, j;

		for (i = 0; i < sourceValues.getSize(); i++)
		{
			value1 = sourceValues.getFieldValue(i);
			for (j = 0; j < targetValues.getSize(); j++)
			{
				value2 = targetValues.getFieldValue(j);

				if (value1.getField() == value2.getField())
				{
					LyFieldInfo fieldInfo = (LyFieldInfo)value1.getField();

					// Ok found value to compare
					if (fieldInfo.equals(value1.getValue(), value2.getValue()))
						break; // Ok for this value, go on...
					else if ((fieldInfo instanceof LyStructFieldInfo)
							&& (((LyStructFieldInfo)fieldInfo).deepEquals(value1.getValue(),
									value2.getValue(), null)))
						break;
					else if ((fieldInfo instanceof LyArrayFieldInfo)
							&& (((LyArrayFieldInfo)fieldInfo).deepEquals(value1.getValue(),
									value2.getValue(), null)))
						break;
					else
						return false; // One value differ
				}
			}
			if (j >= targetValues.getSize())
				return false; // should never happen, not found in target values
		}
		return true;
	} // end method isEquals

	/**
	 * Fetchs the complete set of values of te given object.
	 *
	 * @param  controller     the current import controller
	 * @param  object         the object to complete
	 * @param  internalObject indicates if the given object is internal or not
	 * @return a set containing values for all fields info of the object classinfo
	 */
	public LyValueSet getImportedValues(LyImportContentController controller, LyObject object,
		boolean internalObject)
	{
		LyValueSet  valueSet = new LyValueSet();
		// For internal objects like struct objects, import ALL fields.
		LyInfoList<LyFieldInfo>  fields = (internalObject ? null : controller.getFields());
		LyClassInfo classInfo = object.getClassInfo();

		if (fields == null)
			fields = classInfo.getFieldInfos();

		LyFieldInfo fieldInfo;
		Object	    value;
		int		    i, n = fields.getSize();

		// Go through the fieldInfos and build the valueSet
		for (i = 0; i < n; i++)
		{
			fieldInfo = fields.getInfo(i);

			if (fieldInfo.hasMark(LyFieldInfo.MARK_SECRET))
			{
				if ((fieldInfo instanceof LyTextFieldInfo)
						&& (((LyTextFieldInfo)fieldInfo).getEncoding() == null))
				{
					controller.getEnvironment().logWarning("BF AppEngine : can't import text field "
						+ fieldInfo
						+ " which is secret and has no encoding");
					continue;
				}
			}

			if (fieldInfo.isLocal())
			{
				LyValueSet newValues = object.getNewValues();

				if ((newValues != null) && (newValues.contains(fieldInfo)))
					value = newValues.getFieldValue(fieldInfo).getValue();
				// FFT01596: take new value if available...
				else
					value = object.getLocalValue(fieldInfo, false);
					// FFT01596: override local computed fields in import
			}
			else
				value = object.getValue(fieldInfo);
			valueSet.addValue(fieldInfo, value);
		}

		return valueSet;
	} // end method getImportedValues

	/**
	 * Ends the import action of the given controller.<br />
	 * <i>The default implementation</i> does nothing and returns true.
	 *
	 * @param  controller the current LyAreaController instance
	 * @return true if the action performed correctly, false otherwise
	 */
	public boolean endImport(LyImportContentController controller)
	{
		return true;
	}
} // end class LyImportContentBehavior
