/*
 * LyImportTableBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.app.behavior;

import java.util.*;

import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;


/**
 * This class implements the default behavior for actions that are controlled by a import table
 * controller.
 *
 * @author   Lyria S.A.
 * @version  $Id: LyImportTableBehavior.java 28984 2010-12-22 15:45:10Z jlsaouli $
 * @see      LyImportTableController
 */
public class LyImportTableBehavior extends LyTableBehavior
    implements LyImportTableBehaviorInterface
{
    /** File version. */
    public static final String __VERSION = "$Id: LyImportTableBehavior.java 28984 2010-12-22 15:45:10Z jlsaouli $";

    /**
     * Checks if the import can be done with the given list of selected fields.<br />
     * <i>The default implementation</i> checks that for the controller class info fields selected
     * fields with the 'NOT NULL' mark have a value and that structure objects are unique.
     *
     * @param   controller  the current import table controller
     * @param   objects     the objects we want to check
     *
     * @return  the error message, or null if everything is fine
     */
    @SuppressWarnings("deprecation")
    public String checkFields(LyImportTableController controller, LyObjectList objects)
    {
        if (!controller.OLD_IMPORT)
            return null;

        LyEnvironment           environment = controller.getEnvironment();
        LyClassInfo             classInfo = controller.getClassInfo();
        LyClassInfo             realClassInfo = controller.getRealClassInfo();
        int                     i, j, nFields = classInfo.getFieldInfoCount();
        LyFieldInfo             fieldInfo, realFieldInfo;
        LyObject                object, realObject;
        LyValue                 fieldValue;
        LyInfoList<LyFieldInfo> fields = controller.getFields();

        for (i = 0; i < objects.getSize(); i++)
        {
            object = objects.getObject(i);

            for (j = 0; j < nFields; j++)
            {
                realFieldInfo = null;
                fieldInfo = classInfo.getFieldInfo(j);

                if (fieldInfo instanceof LyStructFieldInfo)
                    realFieldInfo = realClassInfo.getFieldInfo(fieldInfo.getId().substring(2));
                else
                    realFieldInfo = controller.checkRealFieldInfo(fieldInfo);

                fieldValue = object.getFieldValue(fieldInfo);
                realObject = controller.getRealObject(object);

                // Call import behavior for possible additional checks
                if ((realObject == null)
                        || ((realObject != null)
                            && controller.getFields().contains(fieldInfo)))
                {
                    LyImportBehavior behavior = controller.getImportBehavior();
                    String           err = behavior.checkImportedObjectField(controller, object, realObject, fieldInfo,
                            ((controller.getFields()).contains(fieldInfo)) ? fieldValue : null);

                    if (err != null)
                        return err;
                }

                // looking for id or not null fields
                if (fields.contains(fieldInfo))
                    fieldValue = object.getFieldValue(fieldInfo);
                else
                    fieldValue = null;

                if ((fieldValue == null) || (fieldValue.getValue() == null))
                {
                    if (realObject == null)
                    {
                        if ((realFieldInfo.hasMark(LyFieldInfo.MARK_NOT_NULL))
                                && (realFieldInfo.getDefaultValue() == null))
                            return environment.getMessage("LY_ER_IMPORT_NOTNULL_FIELD", realFieldInfo.getName(),
                                controller.getName(object), realClassInfo.getName());
                    }
                    else if (!realFieldInfo.hasMark(LyFieldInfo.MARK_OPTIONAL))
                    {

                        LyValue value = realObject.getFieldValue(realFieldInfo);
                        String  importStructClassInfoName;

                        if (fieldInfo instanceof LyStructFieldInfo)
                        {
                            importStructClassInfoName = ((LyStructFieldInfo)fieldInfo).getClassInfo(0).getId();

                            if (((LyRelationValue)value).getValue() != null)
                                value = new LyRelationValue((LyStructFieldInfo)fieldInfo,
                                        importStructClassInfoName + "\\"
                                        + ((LyRelationValue)value).getName());
                            else
                                value = new LyRelationValue((LyStructFieldInfo)fieldInfo, (LyObject)null);
                        }

                        object.setFieldValue(value);
                    }
                }
                // looking for unicity of the struct object, works also for
                // COMPOSITION relations
                else if ((fieldInfo instanceof LyRelationFieldInfo)
                        && (((LyRelationFieldInfo)fieldInfo).getType() == LyRelationFieldInfo.Type.COMPOSITION))
                {
                    LyRelationValue relationValue = (LyRelationValue)fieldValue;
                    LyObject        cachedObject;
                    LyObject        aNextImportObject;
                    LyRelationValue aNextImportRelationValue;
                    LyRelationValue cachedRelationValue;
                    Vector<Object>  cachedObjects = classInfo.getDataSource().getCache().getObjects();

                    for (int k = 0; k < cachedObjects.size(); k++)
                    {
                        cachedObject = (LyObject)cachedObjects.get(k);
                        if (object.equals(cachedObject))
                            continue;

                        cachedRelationValue = (LyRelationValue)cachedObject.getFieldValue(fieldInfo);
                        if (intersection(relationValue, cachedRelationValue) != null)
                            return environment.getMessage("LY_ER_IMPORT_STRUCT_DUPLICATE", controller.getName(object),
                                controller.getName(cachedObject), fieldInfo.getName());
                    }

                    for (int k = i + 1; k < objects.getSize(); k++)
                    {
                        aNextImportObject = objects.getObject(k);
                        aNextImportRelationValue = (LyRelationValue)aNextImportObject.getFieldValue(fieldInfo);

                        if (intersection(relationValue, aNextImportRelationValue) != null)
                            return environment.getMessage("LY_ER_IMPORT_STRUCT_DUPLICATE", controller.getName(object),
                                controller.getName(aNextImportObject), fieldInfo.getName());
                    }
                }
            }
        }

        return null;
    } // end method checkFields

    /**
     * Gets title corresponding to the given action.<br />
     * Default implementation returns the action name and the number of imported objects.
     *
     * @param   areaController  the current LyAreaController instance
     * @param   action          the action that was executed
     *
     * @return  the given action name and number of imported objects
     */
    public String getTitle(LyAreaController areaController, LyAction action)
    {
        LyImportTableController importTableController = (LyImportTableController)areaController;
        String                  result = super.getTitle(areaController, action);
        LyEnvironment           environment = importTableController.getEnvironment();
        LyObjectList            content = importTableController.getContent();
        int                     nbObjects = ((content == null) ? 0 : content.getSize());
        String                  message = environment.getMessage("LY_OBJECTS_COUNT", new Integer(nbObjects));

        result += " (" + message + ")";

        return result;
    }

    /**
     * Returns the message to be displayed when some lines of the import file contain errors.<br />
     * Default implementation returns a standard message with the indexes of lines in error.
     *
     * @param   controller  Import table controller that requires the error message.
     * @param   errorLines  a bitset with the indexes of lines containing errors.
     *
     * @return  the message to be displayed
     */
    public String getErrorLinesMessage(LyImportTableController controller, BitSet errorLines)
    {
        if ((errorLines != null) && !errorLines.isEmpty())
        {
            StringBuffer buffer = new StringBuffer();
            int          i, n = errorLines.length();

            for (i = 0; i < n; i++)
            {
                if (errorLines.get(i))
                {
                    if (buffer.length() > 0)
                        buffer.append(", ");

                    buffer.append(i);
                }
            }

            LyEnvironment env = controller.getEnvironment();
            String        message = env.getMessage("LY_ER_IMPORT_LINES_ERROR", new Integer(errorLines.cardinality()),
                    buffer);

            return message;
        }

        return null;
    }

    /**
     * Looks for intersection of ObjectId that are inside two relation values.
     *
     * @param   value1  1st relation value.
     * @param   value2  2nd relation value.
     *
     * @return  null if no intersection is detected, else a vector that contains instances of
     *          LyObjectId.
     */
    protected static Vector<LyObjectId> intersection(LyRelationValue value1, LyRelationValue value2)
    {
        if (value1.isNull() || value2.isNull())
            return null;

        Vector<LyObjectId> result = new Vector<LyObjectId>();

        for (int i = 0; i < value1.getValueCount(); i++)
            for (int j = 0; j < value2.getValueCount(); j++)
                if (value1.getObjectId(i).toName().equals(value2.getObjectId(j).toName()))
                    result.add(value1.getObjectId(i));

        if (!result.isEmpty())
            return result;
        else
            return null;
    }
} // end class LyImportTableBehavior
