/*
 * LyLoginBehavior.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.app.behavior;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.view.*;
import leon.view.web.*;
import leon.view.event.*;

import java.util.*;

/**
 * Default behavior for the login behavior interface. Applications may override this class or
 * implement their own instance of LyLoginBehaviorInterface
 *
 * @author   Lyria S.A.
 * @version  $Id: LyLoginBehavior.java 40832 2012-09-18 08:03:31Z smaupin $
 * @see      leon.control.LyLoginController
 * @see      leon.app.behaviorinterface.LyLoginBehaviorInterface
 */
public class LyLoginBehavior extends LyActionBehavior
	implements LyLoginBehaviorInterface,
		LyUserInputBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION = "$Id: LyLoginBehavior.java 40832 2012-09-18 08:03:31Z smaupin $";

	/**
	 * Default constructor : must have no parameter to enable automatic creation of instance from
	 * meta-model files.
	 */
	public LyLoginBehavior()
	{
	}

	/**
	 * Login initialization of manipulated classInfo, loginFieldInfo and passwordFieldInfo.<br />
	 * <i>The default implementation</i> does nothing, this initialization may be implemented in
	 * subclasses.
	 *
	 * @param  loginController  current controller
	 */
	public void initializeLogin(LyLoginController loginController)
	{
		// Nothing to do : may be implemented in subclasses
	}

	/**
	 * Complete login initialization with default values or extra fields adding for input.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param  loginController  current controller
	 */
	public void completeInitializeLogin(LyLoginController loginController)
	{
	}

	/**
	 * Validate the login.<br />
	 * This method is deprecated and only calls
	 * {@link #validate(LyLoginController, String, String, LyParameterValues, StringBuffer)}
	 *
	 * @param       loginController  current controller
	 * @param       login            value typed in by the user
	 * @param       password         value typed in by the user
	 * @param       values           extra values from the login controller validation.
	 *
	 * @return      true if login is ok, false for default login tries
	 *
	 * @deprecated  since V1.2. Replaced by
	 *              {@link #validate(LyLoginController, String, String, LyParameterValues, StringBuffer) validate(loginController, login, password, values, new StringBuffer())}
	 *              .
	 */
	public boolean validate(LyLoginController loginController, String login, String password, LyParameterValues values)
	{
		return validate(loginController, login, password, values, new StringBuffer());
	}

	/**
	 * Validate the login: checks that one unique user exists for the given values.<br />
	 *
	 * @param   loginController  current controller
	 * @param   login            value typed in by the user
	 * @param   password         value typed in by the user
	 * @param   values           extra values from the login controller validation.
	 * @param   error            error message to show in case of invalid login. It is a result
	 *                           parameter.
	 *
	 * @return  true if login is valid, false otherwise
	 */
	public boolean validate(LyLoginController loginController, String login, String password, LyParameterValues values,
		StringBuffer error)
	{
		LyObject user = getUser(loginController, values);

		if (user == null)
			return false;

		boolean		autoLogin = !loginController.isInitialized();

		return validateUser(loginController, user, values, error, autoLogin);
	}

	/**
	 * Validate the login: checks that user is valid.
	 *
	 * @param  loginController current controller
	 * @param  user			   user that logged in.
	 * @param  values          extra values from the login controller validation.
	 * @param  error           error message to show in case of invalid login. It is a result
	 *                         parameter.
	 * @param  autoLogin	   boolean indicating if the call results from an auto-login.
	 * @return true if login is valid, false otherwise
	 */
	public boolean validateUser(LyLoginController loginController, LyObject user, LyParameterValues values,
		StringBuffer error, boolean autoLogin)
	{
		LySession  session = loginController.getSession();

		session.setUser(user);
		return true;
	}

	/**
	 * Ends the login.<br />
	 * <i>The default implementation</i> closes user session if login is not valid
	 * ({@link LyLoginController.Status#LOGIN_FAILED}) or cancelled
	 * ({@link LyLoginController.Status#LOGIN_CANCELLED}) and if the login action is the root
	 * action, else if the login is ok or missing ({@link LyLoginController.Status#LOGIN_OK} or
	 * {@link LyLoginController.Status#LOGIN_MISSING}) then invoke the next action.
	 *
	 * @param  loginController  current controller
	 * @param  values           extra values from the login controller validation.
	 */
	public void endLogin(LyLoginController loginController, LyParameterValues values)
	{
		LySession session = loginController.getSession();

		if ((loginController.getStatus() == LyLoginController.Status.LOGIN_FAILED)
				|| (loginController.getStatus() == LyLoginController.Status.LOGIN_CANCELLED))
		{
			LyProject project = session.getApplication().getProject();
			LyAction  rootAction = project.getRootAction();

			if (loginController.getAction() == rootAction)
				// Close session
				session.close();
			else
				free(loginController);
		}
		else if ((loginController.getStatus() == LyLoginController.Status.LOGIN_OK)
				|| (loginController.getStatus() == LyLoginController.Status.LOGIN_MISSING))
		{
			// Validate session
			session.setValid(true);

			// Invoke actions which follow login
			loginController.invokeNextActions();
			free(loginController);
		}
	}

	/**
	 * The login controller is no longer useful. Invoke this method to free this controller from
	 * current session. This method may be overriden to avoid freeing login controller or to do
	 * final processing at this point.
	 *
	 * @param  controller  the login controller
	 */
	protected void free(LyController controller)
	{
		if (controller != null)
			controller.free();
	}

	/**
	 * Gets the user associated to the given login and password values.
	 *
	 * @param   loginController  current controller
	 * @param   values           values from the login controller.
	 *
	 * @return  The object that represents the user corresponding to given parameters if found, null
	 *          otherwise.
	 */
	public LyObject getUser(LyLoginController loginController, LyParameterValues values)
	{
		if (values == null)
			return null;

		LyViewManager	   viewManager = loginController.getViewManager();
		String			   login = values.getParameter(LyLoginController.PARAM_KEY_LOGIN);
		boolean			   systemLogin = false;
		LyFilter           filter;
        LyClassInfo        userClass;
        LyFilterExpression exp;


        if (login != null && "user.name".equals(login))
        {
            if (!viewManager.isMultiSession())
            {
                login = System.getProperty("user.name");
                systemLogin = true;
            }
            else
            {
                login = null;
                LySession session = loginController.getSession();
                LyWebSession webSession = (LyWebSession) session;
                if (webSession.getCurrentRequest() != null)
                {
                    if (viewManager.isPortletMode())
                    {
                        // get portal user information
                        login = webSession.getCurrentRequest().getParameter(LyLoginController.PARAM_KEY_PORTLET_LOGIN);
                    }
                    else
                    {
                        login =  webSession.getCurrentRequest().getRemoteUser();
                        if (login != null && login.indexOf("\\") > 0) // skip Microsoft domain
                        {
                            login = login.substring(login.lastIndexOf("\\") + 1, login.length());
                        }
                    }
                    if (login != null && !"".equals(login))
                    {
                        systemLogin = true;
                    }
                }
            }
        }

		// Validate user and password.
		// Create filter
		userClass = loginController.getUserClassInfo();
		filter = new LyFilter(LyFilter.Type.AND);

		// Build filter expression
		exp = new LyFilterExpression(loginController.getLoginFieldInfo(), LyFilterExpression.Condition.EQUALS, login);
		exp.setModifier(LyFilterExpression.MOD_CASE_SENSITIVE);
		filter.addElement(exp);

		if (!systemLogin)
		{
			String			   password = values.getParameter(LyLoginController.PARAM_KEY_PASSWORD);
			LyFieldInfo		   passwordFieldInfo = loginController.getPasswordFieldInfo();

			// Check if password fieldInfo is encrypted ?
			password = (String)passwordFieldInfo.encodeValue(password);

			exp = new LyFilterExpression(passwordFieldInfo, LyFilterExpression.Condition.EQUALS, password);
			exp.setModifier(LyFilterExpression.MOD_CASE_SENSITIVE);
			filter.addElement(exp);
		}

		// Fetch number of valid users
		LySession	   session = loginController.getSession();
		LyObjectSource source = session.getObjectSource(userClass);

		boolean    synchro = source.isSynchronous();
		LyResponse response;

		source.setSynchronous(true);
		response = source.getObjects(new Vector<LyFieldInfo>(), null, filter, -1, false, null);
		source.setSynchronous(synchro);
		if ((response.getSize() == 0)
				|| (response.getStatus() != LyResponse.Status.COMPLETED))
			return null;

		return (LyObject)response.getResponse(0);
	} // end method getUser

	/**
	 * Invoked when the user validates the input content.<br />
	 * <i>The default implementation</i>
	 * {@link LyLoginController#checkLogin(LyParameterValues) checks the login} if demand source is
	 * validate, {@link LyLoginController#close() closes} it if demand source is cancel.
	 *
	 * @param  controller  the current LyUserInputController instance
	 * @param  demand      demand parameters sent by the view component.
	 */
	public void userInputDemand(LyController controller, LyDemand demand)
	{
		if (demand != null)
		{
			if (LyFormController.DEMAND_VALIDATE.equals(demand.getSource()))
				((LyLoginController)controller).checkLogin(demand.getParameterValues());
			else if (LyFormController.DEMAND_CANCEL.equals(demand.getSource()))
				((LyLoginController)controller).close();
		}
	}

	/**
	 * This method adds a filtered object source for each class that has attributes having the given
	 * marks named markNames. The filter is created with the given value, the condition is equals
	 * for non multiple attribute and Cette mthode ajoute un filtre sur chaque source de donne de
	 * chaque classe pour tous les attributs qui portent la marque spcifique "ATT_CONTEXT". Pour
	 * chaque classe impacte, un Filtered Object Source est cr sur la session avec un filtre sur
	 * la valeur demande. Les valeurs nulles doivent tre acceptes (compatibilit).
	 *
	 * @param  session  instance de la session
	 * @param markNames an array of mark name
	 * @param  values   an array of values
	 */
	protected void setFilterContext(LySession session, String[] markNames, Object[] values)
	{
		if (markNames == null)
			return;

		LyApplication		    application = session.getApplication();
		LyInfoList<LyClassInfo> classes = application.getClassInfos();

		if (classes == null)
			return;

		// Find marks
		int    i, nm = markNames.length;
		BitSet marks = new BitSet();

		for (i = 0; i < nm; i++)
			marks.set(LyFieldInfo.getMark(markNames[i]));

		LyClassInfo					 classe;
		LyFieldInfo					 field;
		LyFilter					 filter, filter2;
		LyFilterExpression.Condition condition;
		LyInfoList<LyFieldInfo>		 attributes;
		int							 j, k, na;
		LyFilterExpression			 exp;
		Object						 value;
		int							 nc = classes.getSize();

		// Loop on all classes
		for (i = 0; i < nc; i++)
		{
			classe = classes.getInfo(i);

			// Find attributes having one of given marks
			attributes = classe.getFieldInfos(marks, false);
			na = ((attributes == null) ? 0 : attributes.getSize());
			if (na == 0)
				continue;

			// Create filter
			filter = new LyFilter();

			// Create a filter expression for each attribute
			for (j = 0; j < na; j++)
			{
				field = attributes.getInfo(j);
				value = null;

				for (k = 0; k < nm; k++)
					if (field.hasMark(LyFieldInfo.getMark(markNames[k])))
						value = values[k];

				// Decide filter condition if multiple or not
				if (field.isMultiple())
					condition = LyFilterExpression.Condition.CONTAINS;
				else
					condition = LyFilterExpression.Condition.EQUALS;

				exp = new LyFilterExpression(field, false, condition, value);

				// Add a condition for optional attributes
				if ((value != null)
						&& field.hasMark(LyFieldInfo.MARK_OPTIONAL))
				{
					filter2 = new LyFilter(LyFilter.Type.OR);
					filter2.addElement(exp);
					filter2.addElement(new LyFilterExpression(field, false, LyFilterExpression.Condition.EQUALS, null));
					filter.addElement(filter2);
				}
				else
					filter.addElement(exp);
			}

			// Finally, create filtered object source
			new LyFilteredObjectSource(session, classe, filter);
		}
	} // end method setFilterContext

	/**
	 * Gets any existing session for the given user.
	 *
	 * @param   loginController  current controller
	 * @param   currentUser      the current user
	 * @param   values           the values of the current view
	 * @return  a session with the current user as the logged-in user.
	 */
	public LySession getExistingSessionForUser(LyLoginController loginController, LyObject currentUser,
		LyParameterValues values)
	{
		if (currentUser == null)
			return null;

		LyApplication application = loginController.getApplication();
		Vector<LyDataSession>	sessions = application.getSessions();
		int			  size = ((sessions == null) ? 0 : sessions.size());

		if (size == 0)
			return null;

		LySession     session;
		LySession     currentSession = loginController.getSession();
		LyViewManager vm1 = currentSession.getViewManager();
		LyViewManager vm2;

		for (int i = 0; i < size; i++)
		{
            session = (LySession)sessions.get(i);
			if (session == currentSession)
				continue;

			if (session.isClosed())
				continue;

			if (!session.isValid())
				continue;

			// check view managers compatibility
			vm2 = session.getViewManager();

			if (vm2 == null)
				continue; // oups

			if (!vm1.areViewManagerCompatible(vm2))
				continue;

			// check the user
			if (currentUser.equals(session.getUser()))
				return session;
		}
		// no session found
		return null;
	} // end method getExistingSessionForUser

	/**
	 * Returns a boolean value that indicates if the reconnect must be confirmed or not.
	 *
	 * @param   loginController  the login controller
	 *
	 * @return  true if the reconnection has to be confirmed by the user.
	 */
	public boolean confirmReconnect(LyLoginController loginController)
	{
		return true;
	}
} // end class LyLoginBehavior
