/**
 * LyMailBehavior.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.app.behavior;

import com.sun.mail.smtp.*;

import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;

import java.io.*;
import java.util.*;

import javax.activation.*;
import javax.naming.*;
import javax.mail.*;
import javax.mail.internet.*;
import javax.mail.internet.MimeMessage.*;


/**
 * This class implements the default behavior for actions that sends mail with a mail controller.
 *
 * @author   Lyria S.A.
 * @version  $Id: LyMailBehavior.java 34729 2012-03-16 07:48:27Z jlsaouli $
 */
public class LyMailBehavior extends LyActionBehavior
	implements LyMailBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION = "$Id: LyMailBehavior.java 34729 2012-03-16 07:48:27Z jlsaouli $";

	/**
	 * Accessor to the String MAIL_JNDI
	 */
	public static final String MAIL_JNDI = "MAIL_JNDI";
	
	/** A mail message session. */
	protected Session _mailSession;

	/** The SMTP server to send the email message. */
	protected String _server;

	/** The port to connect to the SMTP server. */
	protected int _port;

	/** The current environment. */
	protected LyEnvironment _environment;

	/**
	 * Handle the given demand
	 * 
	 * @param controller the controller
	 * @param demand the demand
	 */
	public void userInputDemand(LyController controller, LyDemand demand)
	{
		String id = demand.getSource();

		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			if (id.equals(LyFormController.DEMAND_VALIDATE))
			{
				((LyMailController)controller).validateMail();
				return;
			}
			else if (id.equals(LyFormController.DEMAND_CANCEL))
			{
				controller.hideView();
				free();
				return;
			}
		}
		else if (demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
		{
			if (id.equals(LyMailController.FROM_USER_NAME) && ((LyMailController)controller).isSendSelected())
			{
				String	    newValue = ((LyActionDemand)demand).getValue();
				LyView	    view = controller.getView();
				LyFormField userField;
				LyText	    fieldText;

				userField = (LyFormField)view.getComponent(LyMailController.MAIL_PREVIEW_SENDER);
				fieldText = (LyText)userField.getFieldItem();

				String fromAddress = ((LyMailController)controller).getFromAddress();

				if (newValue.trim().equals(""))
				{
					if (fromAddress == null)
						fieldText.setLabel("");
					else
						fieldText.setLabel(fromAddress);
				}
				else
				{
					String label = newValue.trim();

					if (fromAddress != null)
						label += " <" + fromAddress + ">";

					fieldText.setLabel(label);
				}
			}
		}
	}

	/**
	 * This class stores a user login and a password for SMTP authentication. This class is a basic
	 * authenticator with the {@link #getPasswordAuthentication()} method added that permit to
	 * retrieve a new {@link PasswordAuthentication} object with the given user and password
	 * attributes.
	 *
	 * <p>Note that the CRAM-MD5 authentication on an SMTP server require a realm name that is not
	 * part of this authenticator and therefore must be provided from another method.</p>
	 */
	class _SmtpAuthenticator extends Authenticator
	{
		/** User login. */
		protected String _user;

		/** User password. */
		protected String _password;

		/**
		 * Default contructor.
		 *
		 * @param  user      the user login used afer an AUTH command
		 * @param  password  the password sent to the SMTP server
		 */
		public _SmtpAuthenticator(String user, String password)
		{
			_user = user;
			_password = password;
		}

		public PasswordAuthentication getPasswordAuthentication()
		{
			return new PasswordAuthentication(_user, _password);
		}
	}

	/**
	 * Sends the given email message through this behavior SMTP server.
	 *
	 * @param   email  the message to send through the SMTP protocol
	 *
	 * @return  true if the specified email message has been sent correctly, false otherwise. This
	 *          does not mean the message will be correctly forwarded to its recipients but only
	 *          accepted and correctly sent to the selected SMTP server.
	 */
	protected boolean sendMail(Message email)
	{
		boolean		  result = true;
		
		if (_environment.getEnv(LyMailBehavior.MAIL_JNDI) == null)
		{
			URLName		  url = new URLName("smtp", _server, _port, null, null, null);
			SMTPTransport smtp = new SMTPTransport(_mailSession, url);
	
			try
			{
				smtp.connect();
				smtp.sendMessage(email, email.getAllRecipients());
			}
			catch (SendFailedException e)
			{
				if (e.getValidUnsentAddresses() != null)
				{
					Address[] validUnsentAddresses = e.getValidUnsentAddresses();
	
					_environment.logError("The following valid email addresses were not sent");
					for (int i = 0; i < validUnsentAddresses.length; i++)
					{
						_environment.logError(validUnsentAddresses[i].toString());
					}
				}
	
				if (e.getInvalidAddresses() != null)
				{
					// The invalid email addresses has been discarded during the email message creation,
					// therefore, every address should be valid
					Address[] invalidAddresses = e.getInvalidAddresses();
	
					_environment.logWarning("The following invalid email addresses were not able to be sent");
					for (int i = 0; i < invalidAddresses.length; i++)
					{
						_environment.logWarning(invalidAddresses[i].toString());
					}
				}
	
				if (e.getValidUnsentAddresses() != null)
					result = false;
			}
			catch (MessagingException e)
			{
				_environment.logWarning(e.getMessage());
				result = false;
			}
			finally
			{
				try
				{
					smtp.close();
				}
				catch (MessagingException e)
				{
				}
			}
		}
		else
		{
			try
			{
				Transport.send(email);
			}
			catch (MessagingException e)
			{
				_environment.logError(e.getMessage());
				result = false;
			}
			catch (RuntimeException e)
			{
				_environment.logError(e.getMessage());
				result = false;
			}
		}
		if (result)
			_environment.logInfo("Mail sent");
		else
			_environment.logInfo("Error while sending email");

		return result;
	}

	/**
	 * Checks if the given string is a correct internet email address. The given address should be
	 * valid as the RFC822 with a "@domain" part.
	 *
	 * @param   address  the address to test
	 *
	 * @return  an InternetAddress corresponding to the given string address, or null if the string
	 *          address does not validate
	 */
	protected InternetAddress validateEmailAddress(String address)
	{
		try
		{
			if ((address != null) && (address.indexOf('@') > 0))
			{
				InternetAddress[] email = InternetAddress.parse(address, true);

				return email[0];
			}
		}
		catch (MessagingException e)
		{
		}

		return null;
	}

	/**
	 * Returns the index of the first occurence of a character in the <code>any</code> string that
	 * appears in the <code>s</code> string.
	 *
	 * @param   s    an arbitrary string that is tested
	 * @param   any  the list of characters to find in the first given string
	 *
	 * @return  the index of a character from the second string in the first, or -1 if no characters
	 *          were found.
	 */
	public static int indexOfAny(String s, String any)
	{
		try
		{
			int len = s.length();

			for (int i = 0; i < len; i++)
			{
				if (any.indexOf(s.charAt(i)) >= 0)
					return i;
			}

			return -1;
		}
		catch (StringIndexOutOfBoundsException e)
		{
			return -1;
		}
	}
	/**
	 * This method creates an initial email message for the specified server, with the given user
	 * name and email address. the name and address may be both null or empty, the email will be
	 * sent anyway if its other parts are correct. The shown user will not be set and no respond
	 * address will be available for the receiver of the email.<br>
	 * <b>Note</b>: the fact that an arbitrary user name and email address can be specified for the
	 * email to be sent does not signifies it will be anonymous and its origin will be masqued.
	 *
	 * @param   mailJndi        the resource name for the email JNDI ressource
	 * @param   fromName        the name of the sender. This attribute may be null or empty.
	 * @param   fromAddress     the email address of the sender. This attribute may be null or
	 *                          empty.
	 *
	 * @return  a new valid email message that will be sent through the given SMTP server, with the
	 *          given sender
	 */
	protected Message initMail(String mailJndi, String fromName, String fromAddress)
	{
        try
        {
            Context     context = new InitialContext();
            _mailSession = (Session)context.lookup(mailJndi);
            
            return initMail(null, -1, false, null, null, null, fromName, fromAddress);
        }
        catch (NamingException e)
        {
        	_environment.logError("Error during email ressource lookup : "+e.getMessage()+" (JNDI : "+mailJndi+")");
        }
        return null;
	}
    
	/**
	 * This method creates an initial email message for the specified server, with the given user
	 * name and email address. the name and address may be both null or empty, the email will be
	 * sent anyway if its other parts are correct. The shown user will not be set and no respond
	 * address will be available for the receiver of the email.<br>
	 * <b>Note</b>: the fact that an arbitrary user name and email address can be specified for the
	 * email to be sent does not signifies it will be anonymous and its origin will be masqued.
	 *
	 * @param   server          the SMTP server address, as an IP or with a DNS name
	 * @param   port            the port of the SMTP server
	 * @param   authentication  Indicates that the SMTP server requires authentication
	 * @param   login           the user login to use for authentication
	 * @param   password        the password needed on authentication
	 * @param   realm           the CRAM-MD5 realm for authentication
	 * @param   fromName        the name of the sender. This attribute may be null or empty.
	 * @param   fromAddress     the email address of the sender. This attribute may be null or
	 *                          empty.
	 *
	 * @return  a new valid email message that will be sent through the given SMTP server, with the
	 *          given sender
	 */
	protected Message initMail(String server, int port, boolean authentication, String login, String password,
		String realm, String fromName, String fromAddress)
	{
        _server = server;
        _port = port;
        
        if (_mailSession == null)
        {
    		Properties props = new Properties();
    
    		props.put("mail.transport.protocol", "smtp");
            if (_server != null)
                props.put("mail.smtp.host", _server);
    		props.put("mail.stmp.sendpartial", "true");
    
    		_SmtpAuthenticator auth = null;
    
    		if (authentication)
    		{
    			props.put("mail.smtp.auth", "true");
    			auth = new _SmtpAuthenticator(login, password);
    			if (realm != null)
    				props.put("mail.smtp.sasl.realm", realm);
    		}
    
    		if (_port != 25)
    			props.put("mail.smtp.port", Integer.toString(_port));
    
    		_mailSession = Session.getInstance(props, auth);
        }

		Message email = new MimeMessage(_mailSession);

		try
		{
			if ((fromName != null) || (fromAddress != null))
			{
				InternetAddress sender = new InternetAddress();
				String		    notEmailChars = " \t\n\r()<>,;:\\\"[]";
				String		    address = null;

				if (fromName != null)
					sender.setPersonal(MimeUtility.encodeText(fromName));

				if ((fromAddress != null) && (indexOfAny(fromAddress, notEmailChars) == -1)
						&& (fromAddress.indexOf('@') >= 0))
					address = fromAddress;

				if (address != null)
				{
					sender.setAddress(address);
					email.setFrom(sender);
				}
			}

			// The following header may cause false-positive problems through spam detection
			// softwares
			// XXX: email.setHeader("X-Mailer", "Leonardi");
			email.setSentDate(new Date());
		}
		catch (UnsupportedEncodingException e)
		{
			_environment.logError("Error during email initialization (encoding) : "+e.getMessage());
		}        	
		catch (MessagingException e)
		{
			_environment.logError("Error during email initialization (messaging) : "+e.getMessage());
		}

		return email;
	}

	/**
	 * Send a mail
	 * 
	 * @param controller the controller
	 * @param toClass the classInfo the classinfo corresponding to the recipient 
	 * @param toField the email field
	 * @param selectedObjects a list of selected objects
	 * 
	 * @return true if the mail was successfully send 
	 */
	public boolean sendMail(LyMailController controller, LyClassInfo toClass, LyFieldInfo toField,
		LyObjectList selectedObjects)
	{
		_environment = controller.getEnvironment();

		Message email;
		if (_environment.getEnv(LyMailBehavior.MAIL_JNDI) != null)
			email = initMail(_environment.getEnv(LyMailBehavior.MAIL_JNDI), controller.getFromUserName(), controller.getFromAddress());
		else
			email = initMail(controller.getSmtpServer(), controller.getSmtpPort(), controller.getAuthentication(),
				controller.getAuthLogin(), controller.getAuthPassword(), controller.getAuthRealm(),
				controller.getFromUserName(), controller.getFromAddress());

		// Retrieving the receipients list
		LyFormField field = (LyFormField)controller.getView().getComponent(LyMailController.SELECTED_RECIPIENTS);
		LyRelation  users = (LyRelation)field.getFieldItem();
		String	    id;

		try
		{
			for (int i = 0; i < users.getLinkCount(); i++)
			{
				LyLink user = users.getLink(i);

				if (user.isSelected())
				{
					id = user.getId();
					LyObjectId	    userId = controller.getObjectId(id);
					LyObject	    userObject = userId.getLogicalObject();
					LyValue		    mailField = userObject.getFieldValue(toField);
					InternetAddress recipient = validateEmailAddress(mailField.getString());

					if (recipient != null)
						email.addRecipient(RecipientType.TO, recipient);
				}
			}

			if ((email.getRecipients(RecipientType.TO) == null)
					|| (email.getRecipients(RecipientType.TO).length == 0))
				return false;

			email.setSubject(controller.getGeneratedSubject());

			String rawMessage = controller.getGeneratedRawContent();
			String htmlMessage = controller.getGeneratedHtmlContent();

			MimeBodyPart rawText = new MimeBodyPart();

			rawText.setText(rawMessage);

			Multipart    text = null;
			MimeBodyPart htmlText = null;

			text = new MimeMultipart("alternative");
			htmlText = new MimeBodyPart();
			htmlText.setContent(htmlMessage, "text/html");
			text.addBodyPart(rawText);
			text.addBodyPart(htmlText);

			email.setContent(text);
		}
		catch (MessagingException e)
		{
			_environment.logWarning("Error during email creation");
		}

		return sendMail(email);
	}

	/**
	 * Send the email
	 * 
	 * @param controller the mail controller
	 * @param selectedUsers the list of recipient
	 * @param toField the field corresponding to the email address
	 * @return true if the mail was succesfully send
	 */
	public boolean sendMail(LyMailController controller, LyObjectList selectedUsers, LyFieldInfo toField)
	{
		_environment = controller.getEnvironment();

		Message email = initMail(controller.getSmtpServer(), controller.getSmtpPort(), controller.getAuthentication(),
				controller.getAuthLogin(), controller.getAuthPassword(), controller.getAuthRealm(),
				controller.getFromUserName(), controller.getFromAddress());

		String subject = controller.getSubject();
		String rawMessage = controller.getRawMessage();
		File   htmlFile = controller.getHtmlMessage();
		File   attachedFile = controller.getAttachedFile();

		try
		{
			for (int i = 0; i < selectedUsers.getSize(); i++)
			{
				LyObject	    user = selectedUsers.getObject(i);
				LyValue		    mailField = user.getFieldValue(toField);
				InternetAddress recipient = validateEmailAddress(mailField.getString());

				if (recipient != null)
					email.addRecipient(RecipientType.TO, recipient);
			}

			if ((email.getRecipients(RecipientType.TO) == null)
					|| (email.getRecipients(RecipientType.TO).length == 0))
				return false;

			if (subject != null)
				email.setSubject(subject);

			// Getting html message
			String htmlMessage = null;

			try
			{
				if (htmlFile != null)
				{
					InputStream  iStream = new FileInputStream(htmlFile);
					StringBuffer sb = new StringBuffer();
					int			 b = iStream.read();

					while (b != -1) // EOF
					{
						sb.append((char)b);
						b = iStream.read();
					}

					htmlMessage = sb.toString();
					iStream.close();
				}
			}
			catch (IOException e)
			{
			}

			if ((htmlMessage == null)
                && ((attachedFile == null)
                    || !attachedFile.exists()))
				email.setText(rawMessage);
			else
			{
				MimeBodyPart rawText = new MimeBodyPart();
                Multipart    text = null;
                MimeBodyPart htmlText = null;
                
				rawText.setText(rawMessage);

				if (htmlMessage != null)
				{
					text = new MimeMultipart("alternative");
					htmlText = new MimeBodyPart();
					htmlText.setContent(htmlMessage, "text/html");
				}

				if (attachedFile != null)
				{
					Multipart mixed = new MimeMultipart("mixed");

					if (htmlMessage == null)
						mixed.addBodyPart(rawText);
					else
					{
						text.addBodyPart(rawText);
						text.addBodyPart(htmlText);

						MimeBodyPart mixedText = new MimeBodyPart();

						mixedText.setContent(text);

						mixed.addBodyPart(mixedText);
					}

					MimeBodyPart added_file = new MimeBodyPart();
					DataSource   ds = new FileDataSource(attachedFile);

					added_file.setDataHandler(new DataHandler(ds));
					added_file.setFileName(attachedFile.getName());
					added_file.setDisposition(Part.ATTACHMENT);
					mixed.addBodyPart(added_file);

					email.setContent(mixed);
				}
				else
				{
					text.addBodyPart(rawText);
					text.addBodyPart(htmlText);

					email.setContent(text);
				}
			}

		}
		catch (MessagingException e)
		{
			_environment.logWarning("Error during email creation");
		}

		return sendMail(email);
	}
}
