/*
 * LyMessageBehavior
 * Copyright (c) 2000-2011, W4.
 * All rights reserved.
 */
package leon.app.behavior;

import java.util.*;

import leon.app.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;

/**
 * Implements the message action.
 * This class handles the build of the userInput and show a message to the selected users.
 *
 */
public class LyMessageBehavior extends LyUserInputBehavior
{
	private static final String	MESSAGE_ACTION_ID	= "_message";
	private static final String	PARAM_SENDER	= "_sender";
	private static final String	PARAM_MESSAGE	= MESSAGE_ACTION_ID;
	private static final String	PARAM_DATE		= "_date";
	private static final String	MESSAGE_FIELD_ID	= PARAM_MESSAGE;
	private static final String	TO_FIELD_ID			= "_targetUsers";

	/** Store the relation field */
	private LyRelation		_targetUserField = null;

	/** Store the text field */
	private LyText			_messageField = null;

	/**
	 * Default constructor.
	 */
	public LyMessageBehavior()
	{
		super();
	}

	/**
	 * Handle form validation
	 */
	public boolean validate(LyController controller, LyParameterValues values)
	{
		if (!super.validate(controller, values))
			return false;

		String			message = _messageField.getValue();//values.getParameter(MESSAGE_FIELD_ID);
		LyObjectList	toList = null;
		LyApplication	application = controller.getApplication();
		boolean			sendToAll = false;

		if (_targetUserField ==null)
		{
			toList = ((LyAreaController)controller).getContext();
			sendToAll = false;
		}
		else
		{
			toList = new LySimpleObjectList();
			String dest = values.getParameter(TO_FIELD_ID);
			StringTokenizer st;
			String val;
			LyObjectId objId;
			LyObject user;

			if (dest != null)
			{
				st = new StringTokenizer(dest, "" + LyTools.OBJECT_SEP);

				while (st.hasMoreTokens())
				{
					val = st.nextToken();
					objId = new LyObjectId(controller.getApplication(), val);
					user = objId.getLogicalObject();
					if (user != null)
						((LySimpleObjectList)toList).addObject(user);
				}
			}

			if ((toList == null) || (toList.getSize() == 0))
				sendToAll = true;
		}

		int						  i, nbSessions = application.getSessionCount();
		LyDataSession			  session;
		boolean					  showMessage = true;
		LySession				  currentSession = controller.getSession();
		LyAction				  action = application.getAction("__showMessage");
		LyAction				  customAction = null;
		Hashtable<String, Object> map = new Hashtable<String, Object>();

		map.put(PARAM_MESSAGE, message);
		map.put(PARAM_SENDER, currentSession.getUser());
		map.put(PARAM_DATE, new Date());

		for (i = 0; i < nbSessions; i++)
		{
			session = application.getSession(i);

			if (!(session instanceof LySession))
				continue; // data only session

			if (session.getUser() == null)
				continue;

			// Don't send message to myself
			if (session == currentSession)
				continue;

			if (sendToAll || toList.contains(session.getUser()))
				showMessage = true;
			else
				showMessage = false;
			if (showMessage)
			{
				if (customAction == null)
				{
					customAction = new LyAction(action, map);
				}
				showMessage((LyAreaController)controller, (LySession)session, customAction);
			}
		}
		return true;
	}

	@Override
	public String getTitle(LyAreaController areaController, LyAction action)
	{
		LyAction messageAction = areaController.getApplication().getAction(MESSAGE_ACTION_ID);
		if (action.equals(messageAction) || action.hasTemplate(messageAction))
			return super.getTitle(areaController, action);

		LyObject		sender = (LyObject)action.getParameter(PARAM_SENDER);
		Date			sendDate = (Date)action.getParameter(PARAM_DATE);
		LyEnvironment	env = areaController.getEnvironment();
		String			format = env.getStringDateFormat(LyDateFieldInfo.Type.DATE_TIME.ordinal());
		String			date = LyDateFormatter.format(sendDate, format);
		String			title = date + "<br />" + env.getMessage("LY_MESSAGE_FROM", sender.getName());

		return title;
	}

	/**
	 * Format and show a message to a user
	 * @param controller the controller
	 * @param targetSession the session that will show the message
	 * @param message the message to show
	 */
	public void showMessage(LyAreaController controller, LySession targetSession, LyAction targetAction)
	{
		LyController targetRootController = targetSession.getController(0);
		targetSession.invokeAction(targetRootController, targetAction, null, null, true);
	}

	/**
	 * This method build dynamically the content of the message view.
	 * It always add a content text field
	 * If we cannot find users in the context of the controller, we add a relation field displaying currently
	 * connected users.
	 */
	public boolean startAction(LyAreaController areaController, LyAction action)
	{
		LyAction messageAction = areaController.getApplication().getAction(MESSAGE_ACTION_ID);
		if (action.equals(messageAction) || action.hasTemplate(messageAction))
		{
			LySession	session = areaController.getSession();
			LyObject	user = session.getUser();
			LyClassInfo userCls = (user == null ? null : user.getClassInfo());
			LyObjectList context = areaController.getContext();

			if (user == null)
			{
				areaController.getParentAreaController().showError("LY_NO_CONNECTED_USER_FOR_SOCIAL_ACTION");
				return false;
			}

			LyObjectList		  users = (context == null ? null : context.getList(session, userCls));
			LyUserInputController messageController = (LyUserInputController)areaController;

			if ((users == null) || (users.getSize() == 0))
			{
				// Add a relation field to show all connected users.
				LyApplication application = session.getApplication();
				int 		  i, nbSessions = application.getSessionCount();

				users = new LySimpleObjectList();
				for (i = 0; i < nbSessions; i++)
				{
					if (application.getSession(i) == session)
						continue;

					user = application.getSession(i).getUser();

					if (user == null)
						continue;

					((LySimpleObjectList)users).addObject(user);
				}
				if ( (users == null) || (users.getSize() == 0))
				{
					areaController.getParentAreaController().showError("LY_NOT_ENOUGH_CONNECTED_USER_FOR_SOCIAL_ACTION");
					return false;
				}

				_targetUserField = messageController.addRelation(TO_FIELD_ID, "LY_TARGET_USERS", true, true);
				_targetUserField.setNRows(4);

				LyLink	   link;

				for (i = 0; i < users.getSize(); i++)
				{
					user = users.getObject(i);

					link = new LyLink(user.getId(), user.getName());
					_targetUserField.addLink(link);
				}
			}

			// We always add a message field
			_messageField = messageController.addText(MESSAGE_FIELD_ID, "LY_MESSAGE_CONTENT", "", 50, 10, null, false, -1);
			_messageField.setType(LyTextBase.Type.RICH);
		}
		else
		{
			// we just have to display a message
			LyUserInputController messageController = (LyUserInputController)areaController;
			String message = (String)action.getParameter(PARAM_MESSAGE);
			_messageField = messageController.addText(MESSAGE_FIELD_ID, "", message, 50, 10, null, false, -1);
			_messageField.setType(LyTextBase.Type.RICH);
			_messageField.setEditable(false);
		}
		return super.startAction(areaController, action);
	}
}
