/*
 * LyMetricsBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import java.text.*;
import java.util.*;

import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.info.infointerface.*;
import leon.misc.*;
import leon.view.*;


/**
 * This class implements the default behavior for metrics chart actions.
 *
 * @author  Lyria S.A.
 * @version $Id: LyChartBehavior.java 30007 2011-03-10 17:58:16Z jlsaouli $
 */
public class LyMetricsChartBehavior extends LyChartBehavior
	implements LyComparatorInterface
{
	/** Identifier for sessions by date chart action. */
	public static final String	SESSIONS_BY_DATE_CHART_ACTION_ID = "__leonMetricSessionsByDate";

	/** Identifier for sessions by hour chart action. */
	public static final String	SESSIONS_BY_HOUR_CHART_ACTION_ID = "__leonMetricSessionsByHour";

	/** Identifier for sessions duration chart action. */
	public static final String	SESSIONS_DURATION_CHART_ACTION_ID = "__leonMetricSessionsDuration";

	/** Identifier for actions by date chart action. */
	public static final String	ACTIONS_BY_DATE_CHART_ACTION_ID = "__leonMetricActionsByDate";

	/** Identifier for most used actions chart action. */
	public static final String	MOST_USED_ACTIONS_CHART_ACTION_ID = "__leonMetricMostUsedActions";

	/** Identifier for max connected users chart action. */
	public static final String	MAX_CONNECTED_USERS_CHART_ACTION_ID = "__leonMetricMaxConnectedUsers";

	/** Max actions parameter name. */
	public static final String	MAX_ACTIONS_PARAMETER = "_max";

	/** Identifier for start date field. */
	public static final String	METRIC_START_DATE_FIELD_ID = "__leonMetricStartDate";

	/** Identifier for end date field. */
	public static final String	METRIC_END_DATE_FIELD_ID = "__leonMetricEndDate";

	/** Identifier for user field. */
	public static final String	METRIC_USER_FIELD_ID = "__leonMetricUserId";

	/** Identifier for type field. */
	public static final String	METRIC_TYPE_FIELD_ID = "__leonMetricType";

	/** Identifier for type session option. */
	public static final String	METRIC_TYPE_SESSION_OPTION_ID = "__leonMetricTypeSession";

	/** Identifier for action field. */
	public static final String	METRIC_ACTION_FIELD_ID = "__leonMetricActionId";

	/** Identifier for class field. */
	public static final String	METRIC_CLASS_FIELD_ID = "__leonMetricClassId";

	/** Identifier for value field. */
	public static final String	METRIC_VALUE_FIELD_ID = "__leonMetricValue";

	/** Identifier for sessions by hour sort criterion. */
	public static final String	SESSIONS_BY_HOUR_SORT_CRITERION_ID = "__leonMetricSessionsByHourSortCriterion";


	/**
	 * Computes a printable string (if the given typeVar is text, number or date) or a Vector of
	 * strings (if the given fieldInfo is a multiple choice or relation) from the given value.
	 */
	public Object getStringChartValue(LyGraphicalChartController controller, LyFieldInfo fieldInfo,
		LyObject object, Object value, LyChartInfo.Type typeVar, boolean yAxis)
	{
		if (value == null)
			return null;

		LyAction	action = controller.getAction();

		if ((fieldInfo != null)
				&& SESSIONS_BY_HOUR_CHART_ACTION_ID.equals(action.getId())
				&& METRIC_START_DATE_FIELD_ID.equals(fieldInfo.getId()))
		{
			// For sessions by hour chart action, transform start date into hour of day interval (eg 9H-10H)
			if (value instanceof Date)
			{
				Date		date = (Date)value;
				Calendar	calendar = GregorianCalendar.getInstance();

				calendar.setTime(date);
				value = calendar.get(Calendar.HOUR_OF_DAY) + "H-" + (calendar.get(Calendar.HOUR_OF_DAY) + 1) + "H";
			}
		}
		else if ((fieldInfo != null)
				&& SESSIONS_DURATION_CHART_ACTION_ID.equals(action.getId())
				&& METRIC_START_DATE_FIELD_ID.equals(fieldInfo.getId()))
		{
			// For sessions duration chart action, transform start date into first day of month
			if (value instanceof Date)
			{
				Date		date = (Date)value;
				Calendar	calendar = GregorianCalendar.getInstance();

				calendar.setTime(date);
				calendar.set(Calendar.SECOND, 0);
				calendar.set(Calendar.MINUTE, 0);
				calendar.set(Calendar.HOUR, 0);
				calendar.set(Calendar.HOUR_OF_DAY, 0);
				calendar.set(Calendar.DAY_OF_MONTH, 2);
				value = calendar.getTime();
			}
		}

		Object	result = super.getStringChartValue(controller, fieldInfo, object, value, typeVar, yAxis);

		if ((result != null)
				&& (fieldInfo != null))
		{
			if (METRIC_USER_FIELD_ID.equals(fieldInfo.getId()))
			{
				// Retrieve user name instead of user id
				LyObjectId	userId = new LyObjectId(controller.getApplication(), result.toString());
				LyObject	user = userId.getLogicalObject();

				if (user != null)
					result = user.getName();
			}
			else if (METRIC_ACTION_FIELD_ID.equals(fieldInfo.getId())
					|| METRIC_CLASS_FIELD_ID.equals(fieldInfo.getId()))
			{
				// Retrieve action name instead of user id
				LyObjectInfo	info = (LyObjectInfo)controller.getApplication().getInfo(result + "");

				if (info != null)
				{
					result = controller.getEnvironment().translate(info.getAlias());

					if (MOST_USED_ACTIONS_CHART_ACTION_ID.equals(action.getId())
							&& (info instanceof LyAction))
					{
						LyAction	actionInfo = (LyAction)info;
						Object		classInfoParam = actionInfo.getParameter("_classInfo", true);

						if (classInfoParam == null)
						{
							// Most used actions : add class name after action name if _classInfo parameter
							// is not defined on action
							LyTextValue	classValue = object.getTextValue(METRIC_CLASS_FIELD_ID);

							if ((classValue != null) && !classValue.isNull())
								result = (String)result + " : " + controller.getEnvironment().translate(classValue.getName());
						}
					}
				}
				else
					result = null;
			}
		}

		return result;
	} // end method getStringChartValue

	/**
	 * The controller resets its content. This method is the right place to add extra processing
	 * when the view is set.<br />
	 * <i>The default implementation</i> does nothing and returns true.
	 *
	 * @param  areaController the current LyAreaController instance
	 * @param  full           indicates if the controller performed a full reset of its content or a
	 *                        partial reset. The meaning of a full reset depends on controller
	 *                        implementation.
	 * @return true
	 */
	public boolean reset(LyAreaController areaController, boolean full)
	{
		LyAction					action = areaController.getAction();
		LyGraphicalChartController	gcController = (LyGraphicalChartController)areaController;
		LyChart						chart = gcController.getChartComponent();
		LyEnvironment				environment = areaController.getEnvironment();

		if (SESSIONS_BY_HOUR_CHART_ACTION_ID.equals(action.getId())
				|| SESSIONS_BY_DATE_CHART_ACTION_ID.equals(action.getId())
				|| ACTIONS_BY_DATE_CHART_ACTION_ID.equals(action.getId()))
		{
			// Invert data in order to display a column for each interval
			chart.setInvertData(true);
		}

		if (SESSIONS_DURATION_CHART_ACTION_ID.equals(action.getId()))
		{
			if (chart.getColumnCount() == 1)
			{
				LyColumn		column = chart.getColumn(0);
				int				i, nValue = column.getChartValueCount();
				LyChartValue	chartValue;

				// Remove data and create two columns with average and maximum durations
				LyColumn					averageColumn = new LyColumn();
				LyColumn					maxColumn = new LyColumn();
				LyChartValue				newChartValue;
				String						valX, nextValX = null;
				Integer						totalValY = 0, maxValY = 0;
				int							counter = 0;

				averageColumn.setLabel(environment.translate("LY_METRIC_AVERAGE_DURATION"));
				maxColumn.setLabel(environment.translate("LY_METRIC_MAXIMUM_DURATION"));

				for (i = 0; i < nValue; i++)
				{
					chartValue = column.getChartValue(i);
					valX = chartValue.getXValue();
					nextValX = (i == nValue - 1 ? null : column.getChartValue(i + 1).getXValue());
					totalValY += LyTools.IntegerFromString(chartValue.getYValue());
					maxValY = Math.max(maxValY, LyTools.IntegerFromString(chartValue.getYValue()));
					counter++;

					// Assuming that data are ordered by start date, if xVal has changed, we reached a new interval
					if ((nextValX == null) || !nextValX.equals(valX))
					{
						newChartValue = new LyChartValue((new Float(totalValY) / (counter * 3600)) + "");
						newChartValue.setXValue(valX);
						averageColumn.addChartValue(newChartValue);
						newChartValue = new LyChartValue((new Float(maxValY) / 3600) + "");
						newChartValue.setXValue(valX);
						maxColumn.addChartValue(newChartValue);
						maxValY = 0;
						totalValY = 0;
						counter = 0;
					}
				}

				chart.removeColumn(column);
				chart.addColumn(averageColumn);
				chart.addColumn(maxColumn);
			}
		}
		else if (MOST_USED_ACTIONS_CHART_ACTION_ID.equals(action.getId())
				&& !chart.isInvertData())
		{
			// Keep only the first n actions according to _max parameter
			String						maxValue = (String)action.getParameter(MAX_ACTIONS_PARAMETER);
			int							max = LyTools.intFromString(maxValue);

			if ((max > 0) && (chart.getColumnCount() == 1))
			{
				LyColumn					column = chart.getColumn(0);
				int							i, nValue = column.getChartValueCount();
				LyChartValue				chartValue;
				String						valY;
				Integer						counter;
				Hashtable<String, Integer>	valuesCount = new Hashtable<String, Integer>();
				Vector<String>				valueNames = new Vector<String>();

				// Retrieve number of occurences for each action
				for (i = 0; i < nValue; i++)
				{
					chartValue = column.getChartValue(i);
					valY = chartValue.getYValue();
					counter = valuesCount.get(valY);

					if (counter == null)
						counter = 1;
					else
						counter++;

					valuesCount.put(valY, counter);

					if (!valueNames.contains(valY))
						valueNames.add(valY);
				}

				// Order action names by their number of occurences
				_LyChartValueCountComparator	cvComparator = new _LyChartValueCountComparator(valuesCount);

				Collections.sort(valueNames, cvComparator);

				Hashtable<Integer, Vector<LyChartValue>>	topValues = new Hashtable<Integer, Vector<LyChartValue>>();
				Vector<LyChartValue>						chartValues;
				int											position;

				// Remove less used actions from the chart data and order most used actions
				for (i = nValue - 1; i >= 0; i--)
				{
					chartValue = column.getChartValue(i);
					valY = chartValue.getYValue();
					column.removeChartValue(chartValue);
					position = valueNames.indexOf(valY);
					chartValues = topValues.get(position);

					if (chartValues == null)
						chartValues = new Vector<LyChartValue>();

					chartValues.add(chartValue);

					if (position <= max)
						topValues.put(position, chartValues);
				}

				for (i = 0; i < max; i++)
				{
					chartValues = topValues.get(i);


					if (chartValues != null)
					{
						for (LyChartValue cv : chartValues)
						{
							column.addChartValue(cv);
						}
					}
				}

				chart.setInvertData(true);
			}
		}
		else if (MAX_CONNECTED_USERS_CHART_ACTION_ID.equals(action.getId()))
		{
			// Comute the maximum number of connected users with start and end dates
			LyChartValue	chartValue;
			int				i, nColumns = chart.getColumnCount();

			// Remove values computed by generic chart info
			for (i = nColumns - 1; i>= 0; i--)
				chart.removeColumn(chart.getColumn(i));

			LyObjectList	content = gcController.getContext();
			LyObject		metric;
			Date			startDate, endDate, date, nextDate;
			int				n = content.getSize(), usersConnected = 0, maxUsersConnected = 0;
			Vector<Long>	dates = new Vector<Long>();
			Long			time;
			LyChoiceValue	typeValue;

			// Store start and end dates (end times are stored as negative values)
			for (i = 0; i < n; i++)
			{
				metric = content.getObject(i);
				typeValue = metric.getChoiceValue(METRIC_TYPE_FIELD_ID);

				// Controller context may have been reinitialized : check if objects are real session metrics
				if ((typeValue == null)
						|| typeValue.isNull()
						|| !METRIC_TYPE_SESSION_OPTION_ID.equals(typeValue.getString()))
					continue;

				if ((metric.getDateValue(METRIC_END_DATE_FIELD_ID) == null) || metric.getDateValue(METRIC_END_DATE_FIELD_ID).isNull())
					continue;

				startDate = metric.getDateValue(METRIC_START_DATE_FIELD_ID).getDateValue();
				endDate = metric.getDateValue(METRIC_END_DATE_FIELD_ID).getDateValue();
				dates.add(startDate.getTime());
				dates.add(-endDate.getTime());
			}

			_LyDatesComparator	cvComparator = new _LyDatesComparator();

			Collections.sort(dates, cvComparator);

			n = dates.size();

			String				month, nextMonth;
			Calendar			calendar = GregorianCalendar.getInstance();
			DateFormatSymbols	dateFormatSymbols = new DateFormatSymbols(environment.getLocale());
			String[]			months = dateFormatSymbols.getMonths();
			LyColumn			column = new LyColumn();

			// Parse start and end times and compute number of connected users for each time interval.
			for (i = 0; i < n; i++)
			{
				time = dates.elementAt(i);
				date = new Date(Math.abs(time));
				calendar.setTime(date);
				month = months[calendar.get(Calendar.MONTH)] + " " + new Integer(calendar.get(Calendar.YEAR));
				nextDate = (i == n - 1 ? null : new Date(Math.abs(dates.elementAt(i + 1))));

				if (nextDate != null)
				{
					calendar.setTime(nextDate);
					nextMonth = months[calendar.get(Calendar.MONTH)] + " " + new Integer(calendar.get(Calendar.YEAR));
				}
				else
					nextMonth = null;

				if (time > 0)
					usersConnected++;
				else
					usersConnected--;

				maxUsersConnected = Math.max(maxUsersConnected, usersConnected);

				if ((nextMonth == null) || !nextMonth.equals(month))
				{
					chartValue = new LyChartValue(maxUsersConnected + "");
					chartValue.setXValue(month);
					column.addChartValue(chartValue);
					maxUsersConnected = 0;
					usersConnected = 0;
				}
			}

			column.setLabel("");
			chart.addColumn(column);
		}

		return true;
	}

	/**
	 * Implementation of the LyComparatorInterface interface.
	 */
	public int compare(LySortCriterion criterion, LyObject object1, LyObject object2)
	{
		if (SESSIONS_BY_HOUR_SORT_CRITERION_ID.equals(criterion.getId()))
		{
			// Implementation of the criterion for session by hour : compare hour of day only
			LyDateValue	dateValue1 = object1.getDateValue((LyDateFieldInfo)criterion.getFieldInfo());
			LyDateValue	dateValue2 = object2.getDateValue((LyDateFieldInfo)criterion.getFieldInfo());

			if ((dateValue1 != null) && !dateValue1.isNull()
					&& (dateValue2 != null) && !dateValue2.isNull())
			{
				Date		date1 = dateValue1.getDateValue();
				Date		date2 = dateValue2.getDateValue();
				Calendar	calendar = GregorianCalendar.getInstance();

				calendar.setTime(date1);

				int	hour1 = calendar.get(Calendar.HOUR_OF_DAY);

				calendar.setTime(date2);

				int	hour2 = calendar.get(Calendar.HOUR_OF_DAY);

				return (hour1 - hour2);
			}
		}

		return 0;
	}

	/**
	 * Comparator used to order chart values count.
	 */
	class _LyChartValueCountComparator implements Comparator<String>
	{
		Hashtable<String, Integer>	_valuesCount;

		_LyChartValueCountComparator(Hashtable<String, Integer> valuesCount)
		{
			_valuesCount = valuesCount;
		}

		public int compare(String name1, String name2)
		{
			Integer	val1 = _valuesCount.get(name1);
			Integer	val2 = _valuesCount.get(name2);

			return (val2 - val1);
		}
	} // end class _LyChartValueCountComparator

	/**
	 * Comparator used to order start and end times with end times as negative values.
	 */
	class _LyDatesComparator implements Comparator<Long>
	{
		_LyDatesComparator()
		{
		}

		public int compare(Long time1, Long time2)
		{
			if (Math.abs(time1) - Math.abs(time2) > 0)
				return 1;
			else if (Math.abs(time1) - Math.abs(time2) < 0)
				return -1;

			return 0;
		}
	} // end class _LyDatesComparator
}
