/*
 * LyPasswordBehaviorInterface.java Copyright (c) 2000-2009, Lyria-W4. All rights
 * reserved.
 */

package leon.app.behavior;

import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.event.*;


/**
 * Default behavior to handle changePassowrds request.<br />
 * <i>The default implementation</i> checks current Password and consistency between the new
 * password and the password confirmation field. This implementation does not make any checking on
 * password content (number of letters, number inside the paswword String): may be overriden to add
 * application rules.
 *
 * @author  Lyria S.A.
 * @version $Id: LyPasswordBehavior.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @see     leon.control.LyPasswordController
 * @see     leon.app.behaviorinterface.LyPasswordBehaviorInterface
 */
public class LyPasswordBehavior extends LyActionBehavior
	implements LyPasswordBehaviorInterface,
		LyUserInputBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyPasswordBehavior.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Password initialization of the manipulated classInfo and passwordFieldInfo.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param passwordController current controller
	 */
	public void initializePassword(LyPasswordController passwordController)
	{
		// Nothing to do
	}

	/**
	 * Complete password initialization with default values or extra fields adding for input.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param passwordController current controller
	 */
	public void completeInitializePassword(LyPasswordController passwordController)
	{
		// Nothing to do
	}

	/**
	 * Method invoked to change the given current password with the given new password value.
	 * Specific implementation may add extra rules here.<br />
	 * <i>The default implementation</i> gets the current user from the controller and the given
	 * values then checks that the user exists, that the given current password is the good one and
	 * that the given new password equals the given confirm password. If all those conditions are
	 * OK, the apply changes and returns true, else return false.
	 *
	 * @param  passwordController current controller
	 * @param  currentPassword    current password for validation.
	 * @param  newPassword        new password value typed in by the user
	 * @param  confirmPassword    confirm password request: should equals newPassword
	 * @param  values             user input values of the current, new and confirmed passwords
	 * @return true if action completed successfully, false if password is unchanged
	 */
	public boolean changePassword(LyPasswordController passwordController, String currentPassword,
		String newPassword, String confirmPassword, LyParameterValues values)
	{
		// ** Retrieve current user **
		LyObject user = getCurrentUser(passwordController, values);

		if (user == null)
		{
			// Nothing to do
			LyEnvironment environment = passwordController.getEnvironment();

			environment.logWarning(
				"LyPasswordBehavior is unable to get current user. Override getCurrentUser.");
			passwordController.showError("LY_ER_CURRENT_USER");

			return true;
		}

		LyFieldInfo passwordFieldInfo = passwordController.getPasswordFieldInfo();

		// Check current Password
		if ((passwordController.getUser() == null) || (currentPassword != null))
		{
			// We check on current user : check if current password is valid
			Object pass = user.getValue(passwordFieldInfo);

			if (passwordFieldInfo.compare(pass, currentPassword) != 0)
			{
				// Error : passwords do not match
				passwordController.showError("LY_ER_CURRENT_PASSWORD");
				return false;
			}
		}

		if (!passwordController.allowEmptyPassword() 
			&& (newPassword == null))
		{
			// Error : passwords is not optional
			passwordController.showError("LY_ER_CURRENT_PASSWORD_NOT_NULL");
			return false;
		}
		
		if (newPassword != null &&
			!newPassword.equals(confirmPassword))
		{
			// Error : passwords do not match
			passwordController.showError("LY_ER_CONFIRM_PASSWORD");
			return false;
		}

		if (checkPasswordValidity(newPassword))
		{
			// Check if anything has changed

			if (newPassword != null &&
				!newPassword.equals(currentPassword))
			{
				user.setValue(passwordFieldInfo, newPassword);
				user.set();

				passwordController.showWarning("LY_PASSWORD_CHANGED");
			}

			return true;
		}
        else
        {
            // Error : password invalid
            passwordController.showError("LY_ER_INVALID_NEW_PASSWORD");
            return false;
        }
	} // end method changePassword

	/**
	 * Checks the password content validity. May add application rules Here.<br />
	 * <i>The default implementation</i> does nothing and returns true.
	 *
	 * @param  newPassword passWord to check
	 * @return true if the new password pass the validity check, false otherwise.
	 */
	protected boolean checkPasswordValidity(String newPassword)
	{
		return true;
	}

	/**
	 * Gets the user corresponding to the password to change.<br />
	 * <i>The default implementation</i> try to get the user on the password controller. If it's
	 * unsuccessfull, try to get the current user from the session.
	 *
	 * @param  passwordController current controller
	 * @param  values             list of values given by the user for current, new and confirmed
	 *                            passwords
	 * @return the LyObject representation of the user if found, null otherwise
	 */
	protected LyObject getCurrentUser(LyPasswordController passwordController,
		LyParameterValues values)
	{
		// Attempt to get on controller
		LyObject user = passwordController.getUser();

		if (user != null)
			return user;

		// Attempt to fetch user value on the application context
		return passwordController.getSession().getUser();
	}

	/**
	 * Method invoked when the user validates the content of its inputs.<br />
	 * <i>The default implementation</i> {@link LyController#hideView() hides the controller view}
	 * and {@link LyController#free() frees} it, after successfully {@link
	 * LyPasswordController#checkPassword(LyParameterValues) checking the password} if he validated
	 * the input or directly if he canceled it.
	 *
	 * @param controller the current LyUserInputController instance
	 * @param demand     demand parameters sent by the view component.
	 */
	public void userInputDemand(LyController controller, LyDemand demand)
	{
		if (demand != null)
		{
			String id = demand.getSource();

			if (LyFormController.DEMAND_VALIDATE.equals(demand.getSource()))
			{
				if (((LyPasswordController)controller).checkPassword(demand.getParameterValues()))
				{
					controller.hideView();
					controller.free();
				}
			}
			else if (id.equals(LyFormController.DEMAND_CANCEL))
			{
				controller.hideView();
				controller.free();
			}
		}
	}

} // end class LyPasswordBehavior
