/*
 * LyPrintModelBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import leon.control.*;
import leon.control.print.*;
import leon.info.*;


/**
 * This class implements the default behavior for a LyPrintModelController.
 *
 * @author  Lyria S.A.
 * @version $Id: LyPrintModelBehavior.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @see     leon.control.LyPrintController
 */
public class LyPrintModelBehavior extends LyActionBehavior
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyPrintModelBehavior.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Default constructor.
	 */
	public LyPrintModelBehavior()
	{
	}

	/**
	 * Method invoked before printing the object or object list.<br />
	 * <i>The default implementation</i> does nothing and return false.
	 *
	 * @param  printer current LySimplePrintController instance
	 * @return true if the process is done (stop generic behavior) false otherwise.
	 */
	public boolean printBegin(LyPrintModelController printer)
	{
		return false;
	}

	/**
	 * Gets the model file name and path.<br />
	 * <i>The default implementation</i> gets the given controller action parameter defining the
	 * model, {@link leon.misc.LyString#expand(String) expands} it and returns it.
	 *
	 * @param  printer current LyPrintModelController instance
	 * @return the model file, by default the "_model" parameter of the controller action
	 */
	public String getModel(LyPrintModelController printer)
	{
		LyAction action = printer.getAction();

		String   modelParam = (String)action.getParameter("_model");

		modelParam = printer.getEnvironment().getString().expand(modelParam);

		return modelParam;
	}

	/**
	 * Gets the output file name and path.<br />
	 * <i>The default implementation</i> gets the given controller action parameter defining the
	 * output file, if it is not null {@link leon.misc.LyString#expand(String) expands} it, else
	 * creates a temporary one.
	 *
	 * @param  printer current LyPrintModelController instance
	 * @return the output file, by default the _output parameter of the action
	 */
	public String getOutput(LyPrintModelController printer)
	{
		LyAction action = printer.getAction();

		String   outputParam = (String)action.getParameter("_output");

		if (outputParam == null)
			outputParam = "tempPrintModel.tmp";
		else
			outputParam = printer.getEnvironment().getString().expand(outputParam);

		return outputParam;
	}

	/**
	 * Get the command string to execute at the end of the model printing.<br />
	 * <i>The default implementation</i> gets the given controller action parameter defining the
	 * command.
	 *
	 * @param  printer current LyPrintModelController instance
	 * @return the command string for exec(), by default the _command parameter of the action
	 */
	public String getCommand(LyPrintModelController printer)
	{
		LyAction action = printer.getAction();

		return (String)action.getParameter("_command");
	}

	/**
	 * Method invoked to configure the content replacer.<br />
	 * <i>The default implementation</i> does nothing and return the given replacer.
	 *
	 * @param  printer  current LyPrintModelController instance
	 * @param  replacer current replacer
	 * @return the configured content replacer
	 */
	public LyContentReplacer preparePrint(LyPrintModelController printer,
		LyContentReplacer replacer)
	{
		return replacer;
	}

	/**
	 * Method invoked to modify the result string.<br />
	 * <i>The default implementation</i> does nothing and return the given result.
	 *
	 * @param  printer current LyPrintModelController instance
	 * @param  result  string result
	 * @return the string result
	 */
	public String completePrint(LyPrintModelController printer, String result)
	{
		return result;
	}

	/**
	 * Method invoked after printing the object or object list.<br />
	 * <i>The default implementation</i> does nothing and return false.
	 *
	 * @param  printer current LySimplePrintController instance
	 * @return true if the process is done (stop generic behavior) false otherwise.
	 */
	public boolean printEnd(LyPrintModelController printer)
	{
		return false;
	}
} // end class LyPrintModelBehavior
