/*
 * LySendFileBehavior.java
 * Copyright (c) 2000-2011, W4.
 * All rights reserved.
 */
package leon.app.behavior;

import java.io.*;
import java.util.*;

import leon.app.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;

/**
 * Implements the send file action.
 * This class handles the build of the userInput and sends a file to the selected users.
 *
 */
public class LySendFileBehavior extends LyUserInputBehavior
{
	private static final String	SEND_FILE_ACTION_ID	= "_sendFile";
	private static final String	PARAM_SENDER		= "_sender";
	private static final String	PARAM_FILE			= SEND_FILE_ACTION_ID;
	private static final String	PARAM_DATE			= "_date";
	private static final String	FILE_FIELD_ID		= PARAM_FILE;

	/** Store the file field */
	private LyFile			_fileField = null;
	
	/** Path of the sent file. */
	private String			_filePath = null;
	
	/**
	 * Default constructor.
	 */
	public LySendFileBehavior()
	{
		super();
	}

	/**
	 * Handle form validation
	 */
	public boolean validate(LyController controller, LyParameterValues values)
	{
		boolean			result = super.validate(controller, values);
		
		if (result)
		{
			LyViewManager				viewManager = controller.getViewManager();
			String						filePath = viewManager.getTemporaryDir() + File.separator + _fileField.getValue();
			LyObjectList				context = ((LyAreaController)controller).getContext();
			LyObject					targetUser = context.getObject(0);
			LyApplication				application = controller.getApplication();
			int							i, nbSessions = application.getSessionCount();
			LyDataSession				session;
			LySession					currentSession = controller.getSession();
			LyAction					action = application.getAction("__sendFile");
			LyAction					customAction = null;
			Hashtable<String, Object>	map = new Hashtable<String, Object>();
			
			map.put(PARAM_FILE, filePath);
			map.put(PARAM_SENDER, currentSession.getUser());
			map.put(PARAM_DATE, new Date());
			
			for (i = 0; i < nbSessions; i++)
			{
				session = application.getSession(i);
	
				if (!(session instanceof LySession))
					continue; // data only session
	
				if (targetUser == session.getUser())
				{
					if (customAction == null)
						customAction = new LyAction(action, map);
	
					sendFile((LyAreaController)controller, (LySession)session, customAction);
				}
			}
		}
		
		return result;
	}

	@Override
	public void free()
	{
		if (_filePath != null)
		{
        	File	file = new File(_filePath);

        	file.delete();
		}
		
		super.free();
	}
	
	@Override
	public String getTitle(LyAreaController areaController, LyAction action)
	{
		LyAction messageAction = areaController.getApplication().getAction(SEND_FILE_ACTION_ID);

		if (action.equals(messageAction) || action.hasTemplate(messageAction))
			return super.getTitle(areaController, action);

		LyObject		sender = (LyObject)action.getParameter(PARAM_SENDER);
		Date			sendDate = (Date)action.getParameter(PARAM_DATE);
		LyEnvironment	env = areaController.getEnvironment();
		String			format = env.getStringDateFormat(LyDateFieldInfo.Type.DATE_TIME.ordinal());
		String			date = LyDateFormatter.format(sendDate, format);
		String			title = date + "<br />" + env.getMessage("LY_FILE_FROM", sender.getName());

		return title;
	}

	/**
	 * Sends a file to a user
	 * @param controller the controller
	 * @param targetSession the session that will show the message
	 * @param targetAction the action to invoke
	 */
	public void sendFile(LyAreaController controller, LySession targetSession, LyAction targetAction)
	{
		LyController targetRootController = targetSession.getController(0);
		targetSession.invokeAction(targetRootController, targetAction, null, null, true);
	}
	
	/**
	 * This method build dynamically the content of the message view.
	 * It always add a content text field
	 * If we cannot find users in the context of the controller, we add a relation field displaying currently
	 * connected users.
	 */
	public boolean startAction(LyAreaController areaController, LyAction action)
	{
		LyAction 		sendFileAction = areaController.getApplication().getAction(SEND_FILE_ACTION_ID);
		LySession		session = areaController.getSession();
		LyViewManager	viewManager = session.getViewManager();

		if (action.equals(sendFileAction) || action.hasTemplate(sendFileAction))
		{
			LyObject		user = session.getUser();
	
			if (user == null)
			{
				areaController.getParentAreaController().showError("LY_NO_CONNECTED_USER_FOR_SOCIAL_ACTION");
				return false;
			}
	
			LyUserInputController	messageController = (LyUserInputController)areaController;

			// We always add a message field
			_fileField = messageController.addFile(FILE_FIELD_ID, "LY_SEND_FILE_FILE", null, LyFile.Type.TEXT, false);
			_fileField.setAttachmentHandler("default");
			_fileField.setAttachmentDirectory(viewManager.getTemporaryDir());
		}
		else
		{
			LyUserInputController	fileController = (LyUserInputController)areaController;
			LyFormField				field = new LyFormField(FILE_FIELD_ID);
			LyCaption   			caption = new LyCaption();

			_filePath = (String)action.getParameter(PARAM_FILE);
			caption.setLabel("LY_SEND_FILE_SAVE_FILE");
			caption.setColon(true);
			
			File f = new File(_filePath);
			
			if (f.exists())
			{
				// have an URL to this temporary location
				String fileDir = viewManager.getDownloadAction(); 
				String label = f.getName();
				String id = session.storeTempFile(f.getAbsolutePath(), label);
				String url = fileDir + id;

				LyHRef	href = new LyHRef();

				href.setId(url);
				href.setLabel(label);
				field.setCaption(caption);
				field.setFieldItem(href);
				
				if (fileController.getFormComponent() != null)
					fileController.getFormComponent().addFormField(field);
			}
		}
		
		return super.startAction(areaController, action);
	}
}
