/*
 * LySessionBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.info.misc.LyFormula;
import leon.misc.*;
import leon.peer.sql.LySqlDataProvider;
import leon.view.*;
import leon.view.event.*;
import leon.view.setinterface.*;


/**
 * This class implements the default behavior of a session.
 *
 * @author  Lyria S.A.
 * @version $Id: LySessionBehavior.java 35897 2012-06-14 15:45:16Z jlsaouli $
 * @see     leon.app.LySession
 * @see     leon.app.behaviorinterface.LySessionBehaviorInterface
 */
public class LySessionBehavior extends LyDataSessionBehavior
	implements LySessionBehaviorInterface,
		LyQuestionAnswerInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LySessionBehavior.java 35897 2012-06-14 15:45:16Z jlsaouli $";

	public static final String	ENABLE_ACTION	= "_enableAction";
	public static final String	VALIDATE_ACTION	= "_validateAction";
	
	/**
	 * Constructor.
	 */
	public LySessionBehavior()
	{
	}

	/**
	 * Method called when a session is started.<br />
	 * <i>The default implementation</i> checks that the session application has a root action and
	 * {@link LyDataSession#setValid(boolean) validate} the session if the root action is not a
	 * login action.
	 *
	 * @param session the started session.
	 */
	public void start(LySession session)
	{
		LyApplication application = session.getApplication();
		LyProject     project = application.getProject();
		LyAction	  rootAction = project.getRootAction();

		if (rootAction == null)
		{
			rootAction = application.getAction("_dashboard");
			if (rootAction == null)
				throw new IllegalStateException("BF AppEngine : application " + application.getId()
					+ " has no root action");
		}

		rootAction = rootAction.getRoot();

		// If root action is not a login action, validate the session.
		if (rootAction != application.getInfo("_login"))
			session.setValid(true);
	}

	/**
	 * Gets root action for this session.
	 *
	 * @param session current session.
	 * @return the root action for the session
	 */
	public LyAction getRootAction(LySession session)
	{
		LyApplication application = session.getApplication();
		LyProject     project = application.getProject();
		LyAction	  rootAction = (LyAction)session.getValue("$ROOT_ACTION");

		if (rootAction == null)
		  rootAction = project.getRootAction();

		if (rootAction == null)
			rootAction = application.getAction("_dashboard");

		return rootAction;
	}

    /**
     * Load session preferences.
     * @param session Session on which preferences are loaded.
     * @param userId  Id of the session user (may be null if preference is shared).
     * @return List of preferences of the session.
     */
	public LyInfoList<LyInfo> loadUserPreferences(LySession session, String userId)
    {
        LyApplication      application = session.getApplication();
    	LySqlDataProvider  prefDataProvider = application.getPrefDataProvider();

		if (prefDataProvider != null)
		{
	        LyApplicationBehaviorInterface 	behavior = application.getApplicationBehavior();

        	return behavior.parseExtraFromDB(application, "preference", userId);
		}
        else
        	return application.parseExtraFiles("preference", "LY_USER_PREFERENCE_DIR", userId);
    }

    /**
     * Load session stickers.
     * @param session Session on which stickers are loaded.
     * @param userId  Id of the session user (may be null if preference is shared).
     * @return List of stickers of the session.
     */
    public LyInfoList<LyInfo> loadUserStickers(LySession session, String userId)
    {
        LyApplication      application = session.getApplication();
    	LySqlDataProvider  prefDataProvider = application.getPrefDataProvider();

        if (prefDataProvider != null)
        {
            LyApplicationBehaviorInterface 	behavior = application.getApplicationBehavior();

        	return behavior.parseExtraFromDB(application, "sticker", userId);
    	}
        else
        	return application.parseExtraFiles("sticker", "LY_USER_STICKER_DIR", userId);
    }

	/**
	 * Method called when a session is closed.<br />
	 * <i>The default implementation</i> only calls {@link LySession#confirmClosing()} method.
	 *
	 * @param session the closed session.
	 */
	public void close(LySession session)
	{
		session.confirmClosing();
	}

	/**
	 * Method called when the user ask to quit the given session.<br />
	 * <i>The default implementation</i> {@link LyViewManager#askQuestion(leon.view.LyComponent,
	 * String, leon.view.setinterface.LyQuestionAnswerInterface, Object) asks confirmation} to the
	 * user before exiting.
	 *
	 * @param session Session to exit from
	 */
	public void exit(LySession session)
	{
		LyViewManager viewManager = session.getViewManager();

		if (viewManager == null)
		{
			session.confirmExit();
			return;
		}

		LyEnvironment environment = session.getEnvironment();
		String		  message;
		int			  nThreads = session.getPrintThreadCount();

		if (nThreads <= 0)
			message = "LY_CONFIRM_EXIT";
		else
			message = environment.getMessage("LY_CONFIRM_EXIT_STOP_PRINT", new Integer(nThreads));

		// Ask confirmation
		viewManager.askQuestion(null, message, this, session);
		// Process goes on in questionAnswer
	}

	/**
	 * Answer to a askQuestion request. Here any answer terminates the confirmation.
	 *
	 * @param result     true if the user has clicked on Yes or Ok
	 * @param clientData data specified in the askQuestion method
	 * @see   leon.view.LyViewManager#askQuestion(leon.view.LyComponent, String,
	 *        LyQuestionAnswerInterface, Object)
	 */
	public void questionAnswer(boolean result, Object clientData)
	{
		if (clientData instanceof LySession)
		{
			LySession session = (LySession)clientData;

			if (result)
				session.confirmExit();
		}
	}

    /**
     * Method to compute id of the given controller.
     * @param controller The new controller.
     * @param id Desired id.
     * @return Id of the new controller.
     */
    public String checkControllerId(LyController controller, String id)
    {
        return id;
    }

	/**
	 * Method called when a controller is added to the session.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param session    Session in which a controller is added.
	 * @param controller Controller added to the session.
	 */
	public void controllerAdded(LySession session, LyController controller)
	{
	}

	/**
	 * Method called when a controller is removed from the session.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param session    Session in which a controller is removed.
	 * @param controller Controller removed from the session.
	 */
	public void controllerRemoved(LySession session, LyController controller)
	{
	}

	/**
	 * Method called when a view controller is shown within the given session.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param session        the session
	 * @param viewController The view controller that is shown.
	 */
	public void viewControllerShown(LySession session, LyViewController viewController)
	{
	}

	/**
	 * Method called when a view controller is hidden within the given session.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param session        the session
	 * @param viewController The view controller that is hidden.
	 */
	public void viewControllerHidden(LySession session, LyViewController viewController)
	{
	}

	/**
	 * Validates the given action for the given controller.<br />
	 * <i>The default implementation</i> does nothing and returns true because all actions are by
	 * default valid.
	 *
	 * @param  session    the current session
	 * @param  action     the action to validate
	 * @param  controller the controller for this action
	 * @return true if the action is valid, false otherwise.
	 */
	public boolean validateAction(LySession session, LyAction action, LyController controller)
	{
		if (action.hasMark(LyAction.MARK_MOBILEONLY)
		        && ( (session.getViewManager() == null)
		                || !session.getViewManager().hasProperty(LyAction.getMarkName(LyAction.MARK_MOBILEONLY).toUpperCase())))
		{
			return false;
		}

		if (action.hasMark(LyAction.MARK_NOTFORMOBILE)
		        && (session.getViewManager() != null)
		        && !session.getViewManager().hasProperty(LyAction.getMarkName(LyAction.MARK_NOTFORMOBILE).toUpperCase()))
		{
			return false;
		}

		if ((session.getViewManager() == null)
		        || !session.getViewManager().hasProperty("HANDLE_SOCIAL_BEHAVIOR"))
		{
			LyAction messageAction = session.getApplication().getAction("_message");
			LyAction chatAction = session.getApplication().getAction("_chat");
			LyAction shareAction = session.getApplication().getAction("_shareSession");
			LyAction stickerAction = session.getApplication().getAction("_createSticker");
			LyAction socialAction = session.getApplication().getAction("_social");

			if (action.equals(socialAction)
					|| action.hasTemplate(socialAction)
					|| action.hasTemplate(messageAction)
					|| action.hasTemplate(messageAction)
					|| action.hasTemplate(chatAction)
					|| action.hasTemplate(chatAction)
					|| action.hasTemplate(shareAction)
					|| action.hasTemplate(shareAction)
					|| action.hasTemplate(stickerAction)
					|| action.hasTemplate(stickerAction))
				return false;
		}
		
		// Try to validate action using a formula
		return checkFormula(session, action, VALIDATE_ACTION);
	}

	/**
	 * Internal method to check a formula on current user.
	 */
	protected boolean checkFormula(LySession session, LyAction action, String formulaName)
	{
		if (formulaName == null)
			return true;
		
		String		actionFormula = (String)action.getData(formulaName);
		
		if (actionFormula == null)
			return true;
		
		// Check for current user
		LyObject	user = session.getUser();
		
		if (user == null)
			return true;

		// Try first if formula is managed as a resource property
		LyEnvironment	environment = session.getEnvironment();
		String			formulaExp = environment.getResource(actionFormula);
		
		if (formulaExp == null)
			formulaExp = actionFormula;
		
		// Check on current user			
        LyFormula	formula = LyFormula.formulaFactory(action, formulaExp);
        Object		result = formula.getValue(user);
        
        return (result != null) && result.equals(true);
	}
	
	/**
	 * Enables the given action on the given list of objects for the given controller.<br />
	 * <i>The default implementation</i> checks if the number of selected objects match the min and
	 * max value of the action. If a filter is defined on the action, this behavior checks that the
	 * action is allowed on every object and that every object pass the filter.
	 *
	 * @param  session    the current session
	 * @param  action     the action to enable
	 * @param  controller the controller for this action
	 * @param  objects    list of objects on which the action has to be enabled
	 * @param  showError  report an error if the action is not enabled
	 * @return true if the action is enabled, false otherwise.
	 */
	public boolean enableAction(LySession session, LyAction action, LyController controller,
		LyObjectList objects, boolean showError)
	{
		if (controller == null)
			return true; // XXX : for XML writing
		if (session == null)
		    return false;
		if (session.getApplication().isEdited())
		    return true; // Studio

		// Check if number of selected objects is correct according to min and max
		int    min = action.getMin(), max = action.getMax();
		int    no = ((objects == null) ? 0 : objects.getSize());
		String message = null;

		if (max <= 0)
			max = Integer.MAX_VALUE;

		if ((no < min) || (no > max))
		{
			if (showError)
			{
				LyEnvironment environment = controller.getEnvironment();

				if ((min == max) && (no != min))
					message = environment.getMessage("LY_ER_SEL_OBJ", new Integer(min));
				else if (no < min)
					message = environment.getMessage("LY_ER_MIN_OBJ", new Integer(min));
				else if (no > max)
					message = environment.getMessage("LY_ER_MAX_OBJ", new Integer(max));

				controller.showError(message);
			}

			return false;
		}

        if (action.isInternal())
            return true;

		// check if the controller is associated to an action
		// and if that action has the 'noContext' mark.
		boolean   hasNoContextMark = action.hasMark(LyAction.MARK_NO_CONTEXT);

		if (!hasNoContextMark
			&& (controller != null)
            && (controller instanceof LyAreaController))
		{
			LyAreaController ctrl = (LyAreaController)controller;
			LyAction		 ctrlAction = ctrl.getAction();

			if ((ctrlAction != null)
                    && ctrlAction.hasMark(LyAction.MARK_NO_CONTEXT))
				hasNoContextMark = true;
		}

		// In a contextual view, when the context is empty, actions should be disabled
		// for area controller which is client context.
		if (!hasNoContextMark
                && (controller instanceof LyContextInterface)
				&& (((LyContextInterface)controller).getContext() != null)
				&& (((LyContextInterface)controller).getContext().getSize() == 0)
				&& (controller.getParent() instanceof LyCompoundAreaController)
				&& (((LyCompoundAreaController)controller.getParent()).getContextClient() == controller))
		{
			if (showError)
			{
				LyEnvironment environment = controller.getEnvironment();

				message = environment.getMessage("LY_ER_EMPTY_CONTEXT", action.getName());
				controller.showError(message);
			}

			return false;
		}

		// If a filter is associated to the action, check if selected objects
		// match this filter
		LyFilter filter = action.getFilter();

		if (filter != null)
		{
			LyObject object;

			for (int i = 0; i < no; i++)
			{
				object = objects.getObject(i);

				boolean resu;
				// FFT 1044 : the action may not be defined in the classinfo of
				// the object
				try
				{
					resu = filter.match(object);
				}
				catch (IllegalArgumentException e)
				{
					// The filter applies on another classInfo ;
					resu = false;
				}

				if (!resu)
				{
					if (showError)
					{
						LyEnvironment environment = controller.getEnvironment();

						message = environment.getMessage("LY_ER_ACTION_OBJ", action.getName(),
								object.getName());

						controller.showError(message);
					}

					return false;
				}
			}
		}

		// For an xref action, check first if it gives some results
		String actionId = action.getRoot().getId();

		if ((no > 0) && (actionId == LyAction.ID_XREF))
		{
			LyClassInfo  target = (LyClassInfo)action.getParameter("_classInfo");
			LyObjectList xrefs = objects.getList(session, target);

			if ((xrefs == null) || (xrefs.getSize() == 0))
				return false;
		}
		
		// Try to enable action using a formula
		return checkFormula(session, action, ENABLE_ACTION);
	} // end method enableAction

	/**
	 * Enables the given transverse action on the given list of objects for the given controller.
	 * <br />
	 * <i>The default implementation</i> only calls {@link #enableAction(LySession, LyAction,
	 * LyController, LyObjectList, boolean) enableAction(session, action, controller, objects,
	 * showError)} and returns its result.
	 *
	 * @param  session    the current session
	 * @param  action     the action to enable
	 * @param  classInfo  the classInfo in which the action is checked
	 * @param  controller the controller for this action
	 * @param  objects    list of objects on which the action has to be enabled
	 * @param  showError  report an error if the action is not enabled
	 * @return true if the action is enabled, false otherwise.
	 */
	public boolean enableTransverseAction(LySession session, LyAction action,
		LyClassInfo classInfo, LyController controller, LyObjectList objects, boolean showError)
	{
		return enableAction(session, action, controller, objects, showError);
	}

	/**
	 * Invoke an action that has no controller.<br />
	 * This function is the right place to implement an application specific action because it is
	 * not driven by a controller.<br />
	 * <i>The default implementation</i> does nothing and returns false.
	 *
	 * @param  session    the current session
	 * @param  controller the source controller
	 * @param  action     the action to execute
	 * @param  classInfo  class on which the action applies
	 * @param  objects    list of objects on which the action applies
	 * @param  show       indicates if the action must be shown or not
	 * @return true if the action was handled by this behavior, false otherwise.
	 */
	public boolean invokeExtraAction(LySession session, LyController controller, LyAction action,
		LyClassInfo classInfo, LyObjectList objects, boolean show)
	{
		return false;
	}

	/**
	 * Invoke an action.<br />
	 * This function is the right place to implement an application specific action.<br />
	 * <i>The default implementation</i> does nothing and returns null because it does not handle
	 * any Action.
	 *
	 * @param  session    the current session
	 * @param  controller the source controller
	 * @param  action     the action to execute
	 * @param  classInfo  class on which the action applies
	 * @param  objects    list of objects on which the action applies
	 * @param  show       indicates if the action must be shown or not
	 * @return the controller corresponding to the invoked action.
	 */
	public LyController invokeAction(LySession session, LyController controller, LyAction action,
		LyClassInfo classInfo, LyObjectList objects, boolean show)
	{
		return null;
	}

	/**
	 * Gets the session time out for the current session.
	 *
	 * @param  session the current session
	 * @return the session timeout in seconds
	 */
	public int getSessionTimeout(LySession session)
	{
		LyEnvironment environment = session.getEnvironment();
		String		  timeout = environment.getResource("LY_SESSION_TIMEOUT");
		int			  time = LySession.SESSION_TIMEOUT;

		try
		{
			time = Integer.parseInt(timeout);
		}
		catch (NumberFormatException e)
		{
			environment.logError(
				"BF AppEngine : Cannot parse session timeout value. Using default value");
		}

		return time;
	}

	/**
	 * Get a root level context controller. This controller when created defines a top level session
	 * context : GUI modification Like set controller should become subcontext of this context.
	 * Usually this method returns null (no controller is a top level context controller). The
	 * ImportTableController is defining a root level context to grab all user modifications in the
	 * import context while the import table is active.
	 *
	 * @param session the current session
	 * @return the root level context controller if any, or null if it could not be found.
	 */
	public LyController getRootContextController(LySession session)
	{
		// FFT01632
		LyController ctrl;

		for (int i = 0; i < session.getControllerCount(); i++)
		{
			ctrl = session.getController(i);

			if (ctrl instanceof LyControllerContextInterface2)
				return ctrl;
		}

		return null;
	}

	/**
	 * Method invoked when a session is in time out.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param session the current session
	 */
	public void sessionTimeout(LySession session)
	{
	}

	/**
	 * Method invoked when the user ask for the 'about' action.<br />
	 * <i>The default implementation</i> shows a popup to the user if the given session has a view
	 * manager, {@link LyEnvironment#logOut(String) logs out} the 'about' message otherwise.
	 *
	 * @param session the current session
	 * @param parent  the parent controller
	 */
	public void showAbout(LySession session, LyController parent)
	{
		LyViewManager viewManager = session.getViewManager();

		if (viewManager != null)
			viewManager.showMessage(((parent == null) ? null : parent.getView()), "LY_ITEM_ABOUT",
				"LY_WARNING_ICON", "LY_APPLICATION_ABOUT", null, null, null);
		else
		{
			LyEnvironment environment = session.getEnvironment();

			environment.logOut("LY_APPLICATION_ABOUT");
		}
	}

	/**
	 * Method invoked when the user ask for the 'help' action. This method must show the
	 * corresponding help to the user.
	 *
	 * @param session the current session
	 * @param parent  the parent controller
	 */
	public void showHelp(LySession session, LyController parent)
	{
		LyViewManager    viewManager = parent.getViewManager();
		LyAreaController areaController = null;

		if (parent instanceof LyAreaController)
			areaController = (LyAreaController)parent;
		else if (parent instanceof LyViewController)
			areaController = ((LyViewController)parent).getAreaController();

		if (areaController == null)
			return;

		String helpFile = areaController.getHelpFile();

		viewManager.showLocation(parent.getComponent(), helpFile, true);
		return;
	}

	/**
	 * Method invoked when the user ask for the 'version' action.<br />
	 * <i>The default implementation</i> shows a popup to the user if the given session has a view
	 * manager, {@link LyEnvironment#logOut(String) logs out} the 'version' message otherwise.
	 *
	 * @param session the current session
	 * @param parent  the parent controller
	 */
	public void showVersion(LySession session, LyController parent)
	{
		LyViewManager viewManager = session.getViewManager();

		if (viewManager != null)
			viewManager.showMessage(parent.getView(), "LY_ITEM_VERSION", "LY_WARNING_ICON",
				"LY_ROOT_VERSION", null, null, null);
		else
		{
			LyEnvironment environment = session.getEnvironment();

			environment.logOut("LY_ROOT_VERSION");
		}
	}

	/**
	 * Invoked when a demand is sent to this listener of this session.
	 *
	 * @param  session    the current session
	 * @param  controller controller that routed the demand.
	 * @param  demand     demand parameters sent by the view component.
	 * @return Return true if the demand was handled, false otherwise. If false, the demand is
	 *         transmitted to the controller to try to handle it.
	 */
	public boolean handleDemand(LySession session, LyController controller, LyDemand demand)
	{
		return false;
	}

	/**
	 * Searches for an existing controller being shown, and return it if found.
	 *
	 * @param  session            the current session
	 * @param  parent             The source controller.
	 * @param  action             The action to execute.
	 * @param  classInfo          The class on which the action applies.
	 * @param  objects            The list of objects on which the action applies.
	 * @param  recurseAllSessions indicates if we recurse into all sessions of the application.
	 * @return the controller if found, null otherwise
	 */
	public LyController findExistingController(LySession session, LyController parent,
		LyAction action, LyClassInfo classInfo, LyObjectList objects, boolean recurseAllSessions)
	{
		if (recurseAllSessions)
		{
			LyApplication			 application = session.getApplication();
			LyController             controller;
			Vector<LyDataSession>	 sessions = application.getSessions();
			int						 i, n = (sessions == null ? 0 : sessions.size());

			for (i = 0; i < n; i++)
			{
				session = (LySession)sessions.elementAt(i);
				controller = findExistingController(session, parent, action, classInfo, objects, false);
				if (controller != null)
					return controller; // FOUND !
			}
		}

		// Only take consult / create / set / clone into account by default
		// TODO : this could be a A_MARK on the action ?
		String rootActionId = action.getRoot().getId();

		if ((rootActionId != LyAction.ID_CONSULT)
				&& (rootActionId != LyAction.ID_SET)
				&& (rootActionId != LyAction.ID_CREATE)
				&& (rootActionId != LyAction.ID_CLONE))
			return null;

		int				 i, n = session.getControllerCount();
		LyFormController controller;
		LyObjectList     ctrlList;
		LyController     controllerParent;

		for (i = 0; i < n; i++)
		{
			if (session.getController(i) instanceof LyFormController)
			{
				controller = (LyFormController)session.getController(i);
				controllerParent = controller.getParent(false);
				if ((parent != null)
						&& (!(controllerParent instanceof LyViewController)
							|| (controllerParent.getParent() != parent)))
					continue;

				// Is it the same same action + same objects ?
				if ((controller.getAction() != null)
						&& controller.getAction().getId().equals(action.getId())
						&& (controller.getClassInfo() == classInfo))
				{
					// check objects
					ctrlList = controller.getObjectList();
					if (ctrlList == null)
						ctrlList = controller.getContent();
					if (((ctrlList == null) && (objects == null))
							|| ((objects != null) && objects.equals(ctrlList)))
						return controller;
				}
			}
		}

		return null; // Not found
	} // end method findExistingController

	/**
	 * Get an help file for the given session, the given controller and the given action. This
	 * default implementation returns null.
	 *
	 * @param  session    the current session.
	 * @param  controller the controller of the action.
	 * @param  action     the action for which the helpfile is required.
	 * @return the path to the help file, or null if we have to use the generic help file behavior.
	 */
	public String getHelpFile(LySession session, LyController controller, LyAction action)
	{
		return null;
	} // end method getHelpFile

	/**
	 * Show the log file. To override.
	 *
	 * @param session the current session
	 * @param controller  the parent controller
	 */
	public void showLog(LySession session, LyController controller)
	{
		return;
	}
} // end class LySessionBehavior
