/*
 * LyShareSessionBehavior.java
 * Copyright (c) 2000-2011, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import java.util.*;

import leon.app.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.setinterface.*;


/**
 * This class implements the default behavior for a share session controller.
 */
public class LyShareSessionBehavior extends LyUserInputBehavior
	implements LyQuestionAnswerInterface
{ 
	/** Parent controller */
	protected LyAreaController		_controller;
	
	/** Sessions sharing this session. */
	protected Vector<Object>		_sharingSessions;
	
	/** Number of positive answers from sharing sessions. */
	protected int					_positiveAnswers;
	
	/**
	 * Default constructor.
	 */
	public LyShareSessionBehavior()
	{
		_positiveAnswers = 0;
	}
	
	/**
	 * Initializes information used within this behavior and is called immediately after its
	 * construction.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param controller The controller that requested the creation of this behavior.
	 */
	public void init(LyController controller)
	{
		_controller = (LyAreaController)controller;
	}
	
	@Override
	public boolean startAction(LyAreaController areaController, LyAction action)
	{
    	LyViewManager		viewManager = _controller.getViewManager();
    	
    	// Check if current session is already shared. If so, do not authorize current action.
    	if (viewManager.isShared())
    	{
    		LyEnvironment	environment = areaController.getEnvironment();
    		String			message = environment.getMessage("LY_ER_SESSION_ALREADY_SHARED");
    		LyController	parentController = areaController.getParent();
    		
    		parentController.showError(message);
    		
    		return false;
    	}
    	
		return super.startAction(areaController, action);
	}
	
	@Override
	public void free()
	{
		super.free();
		
		if (_sharingSessions != null)
		{
			_sharingSessions.clear();
			_sharingSessions = null;
		}
	}

	/**
	 * Method invoked when the user validates the content of the input.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param controller the current LyUserInputController instance
	 * @param values     values set by the user.
     * @return True if validation is successful, false otherwise. 
	 */
	public boolean validate(LyController controller, LyParameterValues values)
	{
		boolean	result = super.validate(controller, values);
		
		if (result)
		{
        	LyViewManager		viewManager = _controller.getViewManager();
        	
        	if (viewManager != null)
        	{
        		LySession			currentSession = _controller.getSession();
        		LyApplication		application = _controller.getApplication();
        		LySession			session;
        		LyViewManager		sharingViewManager;
            	LyWorkSpace			context = _controller.getContext();
            	LyObject			currentUser = currentSession.getUser(), user, sessionUser;
            	int					i, nUsers = (context != null ? context.getSize() : 0);
        		int					j, nSessions = application.getSessionCount();

        		_sharingSessions = new Vector<Object>(nUsers);
            	
            	for (i = 0; i < nUsers; i++)
            	{
            		user = context.getObject(i);

            		for (j = 0; j < nSessions; j++)
            		{
            			session = (LySession)application.getSession(j);
            			
            			if (session != currentSession)
            			{
            				sessionUser = session.getUser();
            				
            				if (user == sessionUser)
            				{
            					sharingViewManager = session.getViewManager();

            					LyEnvironment	environment = sharingViewManager.getEnvironment();
            					String			message = environment.getMessage("LY_CONFIRM_ACCEPT_SHARED_SESSION", currentUser.getName());
            					
            					sharingViewManager.askQuestion(null, message, this, sharingViewManager);
            					_sharingSessions.add(session);
            					break;
            				}
            			}
            		}
            	}
        	}

        	// Just hide view, close process will be performed when answers are received from sharing sessions
        	controller.hideView();

        	return false;
		}
		
		return result;
	}

	public void questionAnswer(boolean result, Object clientData)
	{
		LyViewManager		viewManager = _controller.getViewManager();
		boolean				close = false;
		
		if (result)
		{
			_positiveAnswers++;

			if (_positiveAnswers == _sharingSessions.size())
			{
				close = true;
				viewManager.shareSession(_controller.getSession(), _sharingSessions, _controller.getApplication().getId());
			}
		}
		else
			close = true;
		
		if (close)
			close(_controller);
	}
} // end class LyShareSessionBehavior
