/*
 * LyShowUrlBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import java.io.*;
import java.net.*;

import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.info.*;
import leon.misc.*;


/**
 * This class implements the default behavior for action that is controlled by an show url
 * controller.
 *
 * @author  Lyria S.A.
 * @version $Id: LyShowUrlBehavior.java 31821 2011-07-28 16:19:04Z jlsaouli $
 * @see     leon.control.LyShowUrlController
 */
public class LyShowUrlBehavior extends LyActionBehavior
	implements LyShowUrlBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyShowUrlBehavior.java 31821 2011-07-28 16:19:04Z jlsaouli $";

	/**
	 * Construct a show url behavior.
	 */
	public LyShowUrlBehavior()
	{
	}
	
	/**
	 * Get the inputstream from the given parameters
	 * 
	 * @param controller the show url controller
	 * @param urlName the url name
	 * @param environment the current environment
	 * @param showError indicates if an error must be raise if needed
	 * 
	 * @return the input stream
	 */
	public InputStream getFileInputStream(LyShowUrlController controller, String urlName, 
			LyEnvironment environment, boolean showError)
	{
		URL url = null;
		
		try
		{
			url = new URL(urlName);
		}
		catch (MalformedURLException e)
		{
			String message = environment.getMessage("LY_ER_FILE", urlName);
			
			if (showError)
				controller.showError(message);
			
			if (e != null)
			{
				environment.logError(message);
				environment.writeException(e);
			}
			
		}
		
		if (url == null)
			return null;
	
		// create the connection
		URLConnection con;
		try
		{
			con = url.openConnection();
			con.setDoInput(true);
			con.setDoOutput(true);
			con.setUseCaches(false);
			
			// get response
			return con.getInputStream();
		}
		catch (IOException e)
		{
			String message = environment.getMessage("LY_ER_FILE", urlName);
			
			if (showError)
				controller.showError(message);
			
			if (e != null)
			{
				environment.logError(message);
				environment.writeException(e);
			}
		}
		
		return null;
	}


	/**
	 * Gets the URL from the file name. <br />
	 * Tries first to create the URL from the String.<br />
	 * If an exception occurs, tries to create the URL from a file. <br />
	 * 
	 * If a url is found, encode the URL name thaks to the method 
	 * {@link LyString#getXmlString(String, boolean, boolean)}.
	 * 
	 * @param controller	the controller
	 * @param fileName		the file whose URL is required
	 * @param environment	the current environment
	 * @param showError		indicates if an error message has to be shown to the user.
	 * @return the XML encoded file URL if found, null otherwise. 
	 */
	@SuppressWarnings("deprecation")
	public URL getFileUrl(LyShowUrlController controller, String fileName, LyEnvironment environment, boolean showError)
	{
		if (fileName == null)
			return null;
		
		URL url = null;
		Throwable exception = null;
		
		// Try to build the URL from the given String
		try
		{
			url = new URL(fileName);
		}
		catch (MalformedURLException e)
		{
			url = null;
			exception = e;
		}
		
		// This is not an URL, try to create it from a file
		if (url == null)
		{
			File file = new File(fileName);
			try
			{
				url = file.toURL();
			}
			catch (MalformedURLException e)
			{
				url = null;
				exception = e;
			}
		}
		
		if (url != null)
		{
			String urlStr = url.toString();
			
			urlStr = LyString.getXmlString(urlStr, true, true);
			
			try
			{
				url = new URL(urlStr);
			}
			catch (MalformedURLException e)
			{
				url = null;
				exception = e;
			}
		}
		
		// If no url is found, writes the last exception
		if (url == null)
		{
			String message = environment.getMessage("LY_ER_FILE", fileName);
			
			if (showError)
				controller.showError(message);
			
			if (exception != null)
			{
				environment.logError(message);
				environment.writeException(exception);
			}
			
			return null;
		}
		
		return url;
	}

	/**
	 * Get the url from the controller
	 * @param controller the controller
	 * @return an url as a string
	 */
	public String getUrl(LyShowUrlController controller)
	{
		LyAction action = controller.getAction();
		
		if (action == null)
			return null;
		
		return (String)action.getParameter("_url");
	}
	
	/**
	 * Indicates if the Url must be shown in a new window
	 * 
	 * @param controller the controller
	 * @return true if a new window must be opened
	 */
	public boolean openNewWindow(LyShowUrlController controller)
	{
		LyAction action = controller.getAction();
		
		if (action == null)
			return false;
		
		Object	newWindowParam = action.getParameter("_newWindow");
		
		return "true".equals(newWindowParam);
	}
}