/*
 * LySocialBehavior.java
 * Copyright (c) 2000-2011, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import java.io.*;
import java.sql.*;
import java.util.*;
import java.util.Date;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.peer.sql.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.setinterface.*;


/**
 * This class implements the default behavior for a social controller.
 */
public class LySocialBehavior extends LyTableBehavior
	implements LySessionListenerInterface, LyQuestionAnswerInterface
{
	public static final String	PHOTO = "photo_";
	
    // Parent controller
    protected LyController		_controller;

    // Connected users
    protected LySimpleWorkSpace	_users;
    
    // Associations between image keys and download identifiers
    protected Hashtable<String, String> _downloadIds;

    /**
     * Default constructor.
     */
    public LySocialBehavior()
    {
    }

    /**
     * Gets the content of the table corresponding of the given context.<br />
     * The workspace must be dynamic for dynamic table updates.<br />
     * If the returned workspace is null, the controller computes the content itself.<br />
     * <i>The default implementation</i> does nothing and returns null.
     *
     * @param  tableController Table controller that requires the content.
     * @param  context         Context for which the content is required.
     * @return the table content.
     */
    public LyWorkSpace getContent(LyTableController tableController, LyWorkSpace context)
    {
        // Get connected users
        LySession     currentSession = tableController.getSession();
        LyApplication application = tableController.getApplication();
        int           i, n = application.getSessionCount();
        LyDataSession session;
        LyObject      user;

        _users = new LySimpleWorkSpace(currentSession, true);
        _users.registerClassInfo(tableController.getClassInfo());

        for (i = 0; i < n; i++)
        {
            session = application.getSession(i);

            if (session != currentSession)
            {
                user = session.getUser();

                if ((user != null) && (user != currentSession.getUser()))
                    _users.addObject(user);
            }
        }

        return _users;
    }

    @Override public boolean startAction(LyAreaController areaController, LyAction action)
    {
        // add social controller to session, tchat will use it.
        LySession session = _controller.getSession();
        session.addValue("social_controller", _controller);
        return super.startAction(areaController, action);
    }

    @Override public boolean handleDemand(LyController controller, LyDemand demand)
    {
        LySocialBuilder socialBuilder = (LySocialBuilder) _controller.getBuilder();

        if (demand.getType() == LyActionDemand.TYPE_CLOSE_TAB)
        {

            if (socialBuilder != null)
            {
                int index = demand.getIndex(0);
                socialBuilder.removeTab(index);

                if (socialBuilder.getTabCount() == 0)
                {
                    _controller.getSession().setChatting(false);
                    socialBuilder.freeTabs();
                }
            }
        }
        else if (demand.getType() == LyActionDemand.TYPE_ACTION)
        {
            String source = demand.getSource();

            if ("_tabs".equals(source))
            {
                int    tabIndex = demand.getIndex(0);
                String tabId = ((LyActionDemand) demand).getValue();
                LyTabs tabs = socialBuilder.getTabs();
                LyTab  tab;

                if ((tabId != null) && !"".equals(tabId))
                    tab = (LyTab) tabs.getComponent(tabId);
                else
                    tab = tabs.getTab(tabIndex);

                if (tab != null)
                {
                    tabs.setSelectedTab(tab.getId());
                    tab.setColor(null);
                }
            }
        }
        else if ((demand instanceof LyDragDropDemand) && (demand.getType() == LyDragDropDemand.TYPE_DROP))
        {
            String	source = demand.getSource();
            LyTable	table = socialBuilder.getTable();
            
            if ((source != null) && source.equals(table.getId()))
            {
            	// An object was dropped in the chat window
            	LyDragDropDemand	dropDemand = (LyDragDropDemand)demand;
            	String				draggedObjectId = (String)dropDemand.getAttribute("_draggedObjectId");
            	String				value = ((LyDragDropDemand)demand).getValue();
            	
            	try
            	{
	            	LyObjectId			sentObjectId = new LyObjectId(_controller.getApplication(), draggedObjectId);
	            	LyObject			sentObject = sentObjectId.getLogicalObject();
	            	LyObjectId			userId = new LyObjectId(_controller.getApplication(), value);
	            	LyObject			user = userId.getLogicalObject();
	            	int					i, n = _controller.getApplication().getSessionCount();
	            	LyDataSession		targetSession = null;
	            	LyObject			targetSessionUser;
	            	boolean				objectVisible = true;
	            	
	            	for (i = 0; i < n; i++)
	            	{
	            		targetSession = _controller.getApplication().getSession(i);
						targetSessionUser = targetSession.getUser();
						
						if (targetSessionUser != null && user.getId().equals(targetSessionUser.getId()))
						{
							objectVisible = targetSession.isObjectVisible(sentObject);
							break;
						}
	            	}
	            	
	            	// Check if the user have an access to the sent object
	            	if (!objectVisible)
	            	{
	            		String	message = _controller.getEnvironment().getMessage("LY_SEND_OBJECT_CONFIRM", sentObject.getName(), user.getName());
	            		
	            		_controller.getViewManager().askQuestion(null, message, this, new Object[] {sentObject, targetSession});
	            	}
	            	else
	            		sendObject(sentObject, (LySession)targetSession);
            	}
            	catch (Exception e)
            	{
            		_controller.showError("LY_ER_SEND_OBJECT");
            	}
            }
        }

        return super.handleDemand(controller, demand);
    }
    
    /**
     * Sends an object to a given session.
     * 
     * @param object the object that must be sent.
     * @param targetSession the session that will receive the object
     */
    protected void sendObject(LyObject object, LySession targetSession)
    {
    	if (targetSession != null)
    	{
    		LyAction					action = targetSession.getApplication().getAction("__sendObject");
			LyAction					customAction = null;
			Hashtable<String, Object>	map = new Hashtable<String, Object>();
			
			map.put("_objectId", object.getObjectId().getFullString());
			map.put("_sender", _controller.getSession().getUser());
			map.put("_date", new Date());
			
			customAction = new LyAction(action, map);
			
			targetSession.invokeAction(targetSession.getController(0), customAction, null, null, true);
			
			String	message;
			String	objectId = object.getObjectId().getFullString();
			String	link = "<a class=hrefLabel style=text-decoration:none;color:blue" +
				" href=javascript:invokeHrefLink(\"" + _controller.getId() + "\",\"\",\"" +
				LyString.replace(objectId, "\\", "\\\\") + "\");>" + object.getName() + "</a>";

			message = _controller.getEnvironment().getMessage("LY_SENT_OBJECT_LINK_TO", link, targetSession.getUser().getName());
			
			_controller.showWarning(message);
    	}
    }

    /**
     * Initializes information used within this behavior and is called immediately after its
     * construction.<br />
     * <i>The default implementation</i> does nothing.
     *
     * @param controller The controller that requested the creation of this behavior.
     */
    public void init(LyController controller)
    {
        LyApplication application = controller.getApplication();

        application.addSessionListener(this);
        _controller = controller;
    }

    /**
     * Free this behavior. This method is provided to allow behavior to unregister itself from
     * object sources or workspaces if needed. It should be called only by the controller that built
     * this behavior.<br />
     * <i>The default implementation</i> does nothing.
     */
    public void free()
    {
    	// Unregister from application
        LyApplication application = _controller.getApplication();

        application.removeSessionListener(this);

        // Free chat controllers
        LySocialBuilder builder = (LySocialBuilder) _controller.getBuilder();

        if (builder != null)
            builder.freeChatControllers();
    }

    /**
     * Method called when a session is added to the application.<br />
     * This implementation does nothing.
     *
     * @param  application  Application in which a session is added.
     * @param  session      Session added to the application.
     */
    public void sessionAdded(LyApplication application, LyDataSession session)
    {
        // Wait for user login
    }

    /**
     * Method called when a user logged in a session.
     *
     * @param application Application in which a session is added.
     * @param session     Session added to the application.
     * @param user            User that logged in.
     */
    public void sessionLoggedIn(LyApplication application, LyDataSession session, LyObject user)
    {
		LySession	currentSession = _controller.getSession();

		if ((user != null)
				&& (user != currentSession.getUser()))
			_users.addObject(user);
    }

    /**
     * Method called when a session is removed from the application.<br />
     * This implementation does nothing.
     *
     * @param  application  Application in which a session is removed.
     * @param  session      Session removed from the application.
     */
    public void sessionRemoved(LyApplication application, LyDataSession session)
    {
        LyObject user = session.getUser();

        if (user != null)
            _users.removeObject(user);
    }

    /**
     * Gets the ordered list of pairs of (action, classInfo) to display in a popup menu.<br />
     * <i>The default implementation</i> does nothing and returns the given actions data list.
     *
     * @param  panelController Panel controller that displays the menu.
     * @param  actions         List of pairs (action, classInfo) that the controller is about to
     *                         present in popup menu on current selection.
     * @param  selection       Boolean indicating if popup menu is asked or not on selection.
     * @return List of pairs (action, classInfo) that the controller will display in menu.
     */
    public LyInfoDataList<LyAction, LyClassInfo> getPopupMenuActions(
        LyPanelController panelController, LyInfoDataList<LyAction, LyClassInfo> actions,
        boolean selection)
    {
        LyClassInfo                           userClassInfo = panelController.getClassInfo();
        LyAction                              tableAction = panelController.getAction();
        LyInfoList<LyAction>                  subActions = tableAction.getActions();
        LyAction                              subAction;
        LyInfoDataList<LyAction, LyClassInfo> result;
        int                                   i, n = subActions.getSize();

        result = new LyInfoDataList<LyAction, LyClassInfo>();

        for (i = 0; i < n; i++)
        {
            subAction = subActions.getInfo(i);

            if (subAction.getMin() != -1)
                result.addInfo(subAction, userClassInfo);
        }

        return result;
    }
    
	/**
	 * Gets the row image of the given table object used on display list mode.
	 *
	 * @param  tableController Table controller that requires the row image.
	 * @param  object          Object to which the row image is associated.
	 * @return Path to the image file or the name of an image resource if found, null otherwise.
	 */
	public String getRowImage(LySimpleTableController tableController, LyObject user)
	{
        String	imagePath = getRowImage(tableController, user, true);
        
        if (imagePath == null)
        	imagePath = "LY_DEFAULT_FILE_IMAGE";

        return imagePath;
	}
	
	/**
	 * Gets the row image of the given table object used on display list mode.
	 *
	 * @param  tableController Table controller that requires the row image.
	 * @param  object          Object to which the row image is associated.
	 * @return Path to the image file or the name of an image resource if found, null otherwise.
	 */
	public String getRowImage(LySimpleTableController tableController, LyObject user, boolean withDownload)
	{
        LyApplication   application = tableController.getApplication();
        LySession		session = tableController.getSession();
        LyEnvironment	environment = session.getEnvironment();
        String			userId = user.getObjectId().toString();
        String			imageKey = "__" + PHOTO + userId + "__";
        LyViewManager	vm = session.getViewManager();
        
        if (withDownload && (_downloadIds != null) && _downloadIds.containsKey(application.getId() + imageKey))
		{
        	String		downloadId = _downloadIds.get(application.getId() + imageKey);
			String[]	files = session.getTempFile(downloadId);
			
			if (files != null)
			{
				String	path = files[0];
				File	file = new File(path);
				
				if (file.exists())
				{
					String fileDir = vm.getDownloadAction();

					return fileDir + downloadId + "&time=" + System.currentTimeMillis();
				}
			}
			
			_downloadIds.remove(application.getId() + imageKey);
			environment.putImage(imageKey, null);
		}
        
        String			imagePath = environment.getImagePath(imageKey);
        
        if ((imagePath != null)
        		&& !imagePath.endsWith(imageKey))
        	return imagePath;
        		
        imagePath = loadImage(application, userId);
        
        if (imagePath != null)
        {
        	environment.putImage(imageKey, imagePath);
        	
        	LySqlDataProvider  prefDataProvider = application.getPrefDataProvider();

        	if (prefDataProvider != null)
        	{
	        	File file = new File(imagePath);
	        	
	            file.deleteOnExit();
        	}
        }
        
        if ((imagePath != null) && withDownload && vm.hasProperty("MANAGE_ATTACHMENT"))
		{
			String url = null;
			
			File f = new File(imagePath);
			
			if (f.exists())
			{
				// have an URL to this temporary location
				String fileDir = vm.getDownloadAction(); 
				
				if (fileDir  == null)
					return imagePath;
				else
				{
					String label = f.getName();
					String id = session.storeTempFile(f.getAbsolutePath(), label);
					url = fileDir + id;
					
					if (_downloadIds == null)
						_downloadIds = new Hashtable<String, String>();
					
					_downloadIds.put(application.getId() + imageKey, id);
				}
			}
			
			return url;
		}
        
		return imagePath;
	}
	
    /**
     * Save images from database.
     * @param   application  Application on which extra info is saved.
     * @param   type Extra info type.
     * @param   id Id of saved info.
     * @return  Info that was saved.
     */
    static public void saveImage(LyApplication application, String userId, String imageFilePath)
    {
    	if (imageFilePath == null)
    		return;
        
		String			   type = imageFilePath.substring(imageFilePath.lastIndexOf('.') + 1);
    	LySqlDataProvider  prefDataProvider = application.getPrefDataProvider();

		if (prefDataProvider == null)
		{
			// Copy images in data dir
			String	photosDir = application.getEnvironment().getEnv("LY_DATA_DIR");
			File	dirFile = new File(photosDir + File.separator + "photos");
			
			if (!dirFile.exists())
				dirFile.mkdirs();
			else
			{
				// Remove previous image if found
				String	imagePath = loadImage(application, userId);
				
				if (imagePath != null)
				{
					File file = new File(imagePath);
					
					file.delete();
				}
			}
			
			// Copy image
			File	file = new File(imageFilePath);
			File	newFile = new File(dirFile + File.separator + PHOTO + userId + '.' + type);
				
            try
            {
                InputStream  in = new BufferedInputStream(new FileInputStream(file));
                OutputStream out = new BufferedOutputStream(new FileOutputStream(newFile));
                int          nr;

                while ((nr = in.read()) >= 0)
                    out.write(nr);

                in.close();
                out.close();
            }
            catch (Exception e)
            {
                application.getEnvironment().writeException(e);
            }
            
			return;
		}
		
		// Save in database
		LySqlConnection		sqlConn = (LySqlConnection)prefDataProvider.getConnection();
		
    	try
    	{
            String          request = "delete from " + LyApplicationBehavior.PREF_TABLE
	            + " where " + LyApplicationBehavior.PREF_USER_COL + "='" + userId + "'"
            	+ " and " + LyApplicationBehavior.PREF_TYPE_COL  + " like '" + PHOTO + "%'";
            Connection        	cnx = sqlConn.getConnection();
            Statement		stat = cnx.createStatement();
            
            stat.executeUpdate(request);     
            
			Object[] 		params;
			File 			tmpFile = new File(imageFilePath);
			InputStream 	iStream = new FileInputStream(tmpFile);
			byte[] 			dataFull = new byte[(int)tmpFile.length()];

			for (int i = 0, n = (int)tmpFile.length(); i < n; i++)
				dataFull[i] = (byte)iStream.read();
			
			String			prefType = PHOTO + type;
			int				id = new Integer(LyTools.nextInt());
			
			request = "insert into " + LyApplicationBehavior.PREF_TABLE
				+ " (" + LyApplicationBehavior.PREF_ID_COL
				+ ", " + LyApplicationBehavior.PREF_TYPE_COL
				+ ", " + LyApplicationBehavior.PREF_DATE_COL
				+ ", " + LyApplicationBehavior.PREF_USER_COL
				+ ", " + LyApplicationBehavior.PREF_BLOB_COL + ") values (?,?,?,?,?)";
			params = new Object[]
				{
					id, prefType, new java.sql.Date(System.currentTimeMillis()), userId, dataFull
				};
	
			PreparedStatement	prepStat = cnx.prepareStatement(request);
	
			// Set parameters
			int		i, n = params.length;
			
			for (i = 0; (params != null) && (i < n); i++)
			{
				if (params[i] == null)
					prepStat.setNull(i + 1, java.sql.Types.VARCHAR);
				else if (params[i] instanceof InputStream)
					prepStat.setBinaryStream(i + 1, (InputStream)params[i], (int)tmpFile.length());
				else if (params[i] instanceof byte[])
					prepStat.setBytes(i + 1, (byte[])params[i]);
				else
					prepStat.setObject(i + 1, params[i]);
			}
			
			// execute SQL update
			prepStat.executeUpdate();
	
			// close all
			iStream.close();
		}
		catch (Exception sqle)
		{
			application.getEnvironment().writeException(sqle);
		}
		finally
		{
			prefDataProvider.releaseConnection(sqlConn);
		}
		
		// Remove previous image from cache
        LyEnvironment	environment = application.getEnvironment();
        String			imageKey = "__" + PHOTO + userId + "__";
        
        environment.putImage(imageKey, null);
    }
    
    /**
     * Load images from database.
     * @param   application  Application on which extra info is saved.
     * @param   type Extra info type.
     * @param   id Id of saved info.
     * @return  Info that was saved.
     */
    static public String loadImage(LyApplication application, String userId)
    {        
    	LySqlDataProvider  prefDataProvider = application.getPrefDataProvider();
    	
        if (prefDataProvider == null)
        {
        	// Load as file from preference dir
        	String	photosDir = application.getEnvironment().getEnv("LY_DATA_DIR") + File.separator + "photos";
        	File	dirFile = new File(photosDir);
        	
        	if (dirFile.exists())
        	{
            	String		photoName = PHOTO + userId + '.';
            	String[]	files = dirFile.list();
            	int			i, n = files.length;
            	
            	for (i = 0; i < n; i++)
            		if (files[i].startsWith(photoName))
            			return photosDir + File.separator + files[i];
        	}
			
			return null;
        }
        
        LySqlConnection   sqlConn = (LySqlConnection)prefDataProvider.getConnection();
       
        if (sqlConn == null)
            return null;
        
        ResultSet         rs = null;
        
        try
        {
            // build and send QUERY
            String            request = "select " + LyApplicationBehavior.PREF_TYPE_COL
            	+ "," + LyApplicationBehavior.PREF_BLOB_COL
                + " from " + LyApplicationBehavior.PREF_TABLE + " where "
                + LyApplicationBehavior.PREF_USER_COL + "='" + userId + "'"
                + " and " + LyApplicationBehavior.PREF_TYPE_COL  + " like '" + PHOTO + "%'";
            Connection        cnx = sqlConn.getConnection();
            PreparedStatement prepStat = cnx.prepareStatement(request);

            rs = prepStat.executeQuery();

            while (rs.next())
            {
            	String	type = rs.getString(LyApplicationBehavior.PREF_TYPE_COL);
            	
            	if ((type == null)
            			|| (type.length() != PHOTO.length() + 3))
            		continue;
            	
            	type = type.substring(PHOTO.length());
            	
                Blob 	blob = rs.getBlob(LyApplicationBehavior.PREF_BLOB_COL);
                
                if ((blob == null)
                		|| (blob.length() == 0))
                    continue;

                // Extract BLOB to a temp file
                LyViewManager	viewManager = (LyViewManager)application.getEnvironment().getViewManager();
                String			tempDir = viewManager.getTemporaryDir();
                File			dirFile = new File(tempDir);
                
                if (!dirFile.exists())
                    dirFile.mkdirs();

                String		photoPath = tempDir + File.separator + PHOTO + userId + "." + type;
                File 		photoFile = new File(photoPath);

                // Parse XML Result
                try
                {
                    InputStream  in = blob.getBinaryStream();
                    OutputStream out = new BufferedOutputStream(new FileOutputStream(photoFile));
                    int          nr;

                    while ((nr = in.read()) >= 0)
                        out.write(nr);

                    in.close();
                    out.close();
                    return photoPath;
                }
                catch (Exception e)
                {
                    application.getEnvironment().writeException(e);
                    return null;
                }
            }
        }
        catch (Exception sqle)
        {
            application.getEnvironment().writeException(sqle);
        }
        finally
        {
        	prefDataProvider.releaseConnection(sqlConn);
        }
        
        return null;
    }

	/**
	 * @see LyQuestionAnswerInterface#questionAnswer(boolean, Object)
	 */
	public void questionAnswer(boolean result, Object clientData)
	{
		if (result)
		{
			Object[]	array = (Object[])clientData;
			
			sendObject((LyObject)array[0], (LySession)array[1]);
		}
	}
} // end class LyTableBehavior
