/*
 * LyTabActionBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.info.*;


/**
 * This class implements the default behavior for tab actions.
 */
public class LyTabActionBehavior extends LyActionBehavior
	implements LyTabActionBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyTabActionBehavior.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Construct a tab action behavior.
	 */
	public LyTabActionBehavior()
	{
	}

	/**
	 * Get labels to display in left panel of the wizard.
	 * This method is only used for tab actions with wizard type.
	 * @param tabAreaController Controller of the tab action.
	 * @return A table of strings corresponding to labels, or null if no label is wanted.
	 */
	public String[] getStepLabels(LyTabAreaController tabAreaController)
	{
		LyTabAction		tabAction = (LyTabAction)tabAreaController.getAction();
		int				i, n = tabAction.getTabActionCount();
		String[]		labels = new String[n];
		LyAction		action;
		
		for (i = 0; i < n; i++)
		{
		    action = tabAction.getTabAction(i);
		    labels[i] = " " + (i + 1) + ". " + action.getName() + " ";
		}
		return labels;
	}
	
	/**
	 * Get image to display in left panel of the wizard.
	 * This method is only used for tab actions with wizard type.
	 * @param tabAreaController Controller of the tab action.
	 * @param index Index of the current step.
	 * @return A string that gives the name of the image to display.
	 */
	public String getImage(LyTabAreaController tabAreaController, int index)
	{
	    return "$LY_WIZARD_IMAGE";
	}
	
	/**
	 * Get the colors used for hyperlinks displayed in left panel of the wizard.
	 * This method is only used for tab actions with wizard type.
	 * @param tabAreaController Controller of the tab action.
	 * @return An array of strings that gives the colors for hyperlinks, highlighted hyperlinks,
	 * visited hyperlinks and disabled hyperlinks.
	 */
	public String[] getHrefColors(LyTabAreaController tabAreaController)
	{
	    return new String[] {"Red", "Red", "Blue", "Black"};
	}
	
	/**
	 * Get next action to be displayed in right panel of the wizard.
	 * This method is only used for tab actions with wizard type and is called when next button is selected
	 * the first time the step is reached.
	 * @param tabAreaController Controller of the tab action.
	 * @param index Index of the current step.
	 * @return The next action to be displayed.
	 */
	public LyAction getNextAction(LyTabAreaController tabAreaController, int index)
	{
		LyTabAction		action = (LyTabAction)tabAreaController.getAction();

		return action.getTabAction(index);
	}
	
	/**
	 * Builds area controller corresponding to the given action.
	 * This method is only used for tab actions with wizard type.
	 * @param tabAreaController Controller of the tab action.
	 * @param action Action for which a controller is requested. 
	 * @return The desired area controller, if null is returned the tabAreaController will construct it.
	 */
	public LyAreaController buildAreaController(LyTabAreaController tabAreaController, LyAction action)
	{
	    return null;
	}
	
	/**
	 * This method is only used for tab actions with wizard type.
	 * @param tabAreaController Controller of the tab action.
	 * @param index Index of the current step.
	 * @return A boolean indicating if the Next button is enabled or not.
	 */
	public boolean enableNext(LyTabAreaController tabAreaController, int index)
	{
		LyTabAction		action = (LyTabAction)tabAreaController.getAction();
		
	    return (index < action.getTabActionCount() - 1);
	}
	
	/**
	 * This method is only used for tab actions with wizard type.
	 * @param tabAreaController Controller of the tab action.
	 * @param index Index of the current step.
	 * @return A boolean indicating if the Terminate button is enabled or not.
	 */
	public boolean enableTerminate(LyTabAreaController tabAreaController, int index)
	{
		LyTabAction		action = (LyTabAction)tabAreaController.getAction();
		
	    return (index == action.getTabActionCount() - 1);
	}
	
	/**
	 * Method called when Terminate is selected before committing actions done in wizard.
	 * This method is only used for tab actions with wizard type.
	 * @param tabAreaController Controller of the tab action.
	 * @return A boolean indicating if the commit can be done.
	 */
	public boolean beforeCommit(LyTabAreaController tabAreaController)
	{
	    return true;
	}
	
	/**
	 * Method called when Terminate is selected after committing actions done in wizard.
	 * This method is only used for tab actions with wizard type.
	 * @param tabAreaController Controller of the tab action.
	 */
	public void afterCommit(LyTabAreaController tabAreaController)
	{
	}
	
	/**
	 * Method called when Cancel is selected before cancelling actions done in wizard.
	 * This method is only used for tab actions with wizard type.
	 * @param tabAreaController Controller of the tab action.
	 */
	public void cancel(LyTabAreaController tabAreaController)
	{
	}

	/**
	 * Method called when a tab has been selected.<br />
	 * The default implementation does nothing.
	 * 
	 * @param tabAreaController 	 the tab action controller
	 * @param selectedTab			 the index of the selected tab
	 * @param selectedAreaController the controller of the selected tab action
	 */
	public void selectedTabChanged(LyTabAreaController tabAreaController, int selectedTab, LyAreaController selectedAreaController)
	{
	}
	
} // end class LyTabActionBehavior
