/*
 * LyTableBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import java.io.*;
import java.text.DateFormat;
import java.util.Date;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;


/**
 * This class implements the default behavior for a table controller.
 *
 * @author  Lyria S.A.
 * @version $Id: LyTableBehavior.java 36619 2012-08-28 14:27:47Z jlsaouli $
 * @see     leon.control.LyTableController
 * @see     leon.app.behaviorinterface.LyTableBehaviorInterface
 */
public class LyTableBehavior extends LyPanelBehavior
	implements LyTableBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyTableBehavior.java 36619 2012-08-28 14:27:47Z jlsaouli $";

    /** Max row height. */
    public static short         MAX_ROW_HEIGHT = 10;

  
	/**
	 * Default constructor.
	 */
	public LyTableBehavior()
	{
	}

	/**
	 * Gets the list of fields to display in the table from the given list.<br />
	 * <i>The default implementation</i> does nothing and return the given list without any changes.
	 *
	 * @param  tableController the table controller that requires the list
	 * @param  fields          a list of fields from where fields to display are taken
	 * @return The list of fields to display in the table.
	 */
	public LyInfoList<LyFieldInfo> getFields(LySimpleTableController tableController,
		LyInfoList<LyFieldInfo> fields)
	{
		return fields;
	}

	/**
	 * Gets the status image of the given object.<br />
	 * <i>The default implementation</i> returns the image if it is found or null if
	 *
	 * <ul>
	 *   <li>the statusFields list is null or empty,</li>
	 *   <li>the list contain only one file field and the mime type of this file field is null or
	 *     starts withs "image",</li>
	 *   <li>the given object is null.</li>
	 * </ul>
	 *
	 * @param  controller   Table controller that requires the status image.
	 * @param  statusFields List of field infos that have the status mark.
	 * @param  object       Object to which the status image is associated.
	 * @return Path to the image file or the name of an image resource if found, null otherwise.
	 */
	public String getStatusImage(LySimpleTableController controller,
		LyInfoList<LyFieldInfo> statusFields,
		LyObject object)
	{
		if ((statusFields == null) || (statusFields.getSize() == 0))
		{
			if (controller.getEnvironment().testResource("LY_ERROR_STATE_ICON"))
				return "LY_EMPTY_IMAGES_STATE";
			else
				return null;
		}

		// Problems may occur if there is only one file field info as a status
		// (if file is not an image file). Ignore this case in generic behavior.
		if ((statusFields.getSize() == 1)
				&& (statusFields.getInfo(0) instanceof LyFileFieldInfo))
		{
			LyFileFieldInfo fileField = (LyFileFieldInfo)statusFields.getInfo(0);
			String		    mimeTypes = fileField.getMimeTypes();

			if ((mimeTypes == null) || !(mimeTypes.startsWith("image")))
				return null;
		}

		if (object == null)
			return null;

		StringBuffer statusImage = null;
		int			 i, n = statusFields.getSize();
		LyFieldInfo  fieldInfo;
		Object		 value;

		for (i = 0; i < n; i++)
		{
			fieldInfo = statusFields.getInfo(i);
			value = object.getValue(fieldInfo);

			if (value == null)
				continue;
			// Create new buffer or append if needed
			if (statusImage == null)
				statusImage = new StringBuffer();
			else
				statusImage.append('|'); // XXX

			// Append image path
			statusImage.append(fieldInfo.toString(value));
		}

		return (statusImage == null ? null : statusImage.toString());
	} // end method getStatusImage

	/**
	 * Gets status image tooltip for the given object.<br />
	 * This method is called only if the {@link #getStatusImage(LySimpleTableController, LyInfoList,
	 * LyObject)} method didn't return null.<br />
	 * <i>The default implementation</i> computes a string from the given object class info name,
	 * the object name, fields name and values of the given list found in the object.
	 *
	 * @param  tableController Table controller that requires the status image.
	 * @param  statusFields    List of field infos that have the status mark.
	 * @param  object          Object to which the status image is associated.
	 * @return String containing the tooltip associated to the status image.
	 */
	public String getStatusTooltip(LySimpleTableController tableController,
		LyInfoList<LyFieldInfo> statusFields, LyObject object)
	{
		if ((statusFields == null) || (statusFields.getSize() == 0))
			return null;
		if (object == null)
			return null;

		StringBuffer statusTooltip = new StringBuffer();
		int			 i, n = statusFields.getSize();
		LyFieldInfo  fieldInfo;
		Object		 value;
		String		 name;

		name = object.getClassInfo().getName();
		statusTooltip.append(name);
		statusTooltip.append(" : ");
		statusTooltip.append(tableController.getName(object));

		for (i = 0; i < n; i++)
		{
			fieldInfo = statusFields.getInfo(i);
			value = object.getValue(fieldInfo);
			if (value == null)
				continue;

			// Append field and value names
			statusTooltip.append('\n');

			name = fieldInfo.getName();
			statusTooltip.append(name);
			statusTooltip.append(" = ");

			name = fieldInfo.toName(value);
			statusTooltip.append(name);
		}

		return statusTooltip.toString();
	} // end method getStatusTooltip

	/**
	 * Gets row color associated to the given object.<br />
	 * <i>The default implementation</i> does nothing and returns null.
	 *
	 * @param  controller Table controller that requires the row color.
	 * @param  object     Object for which a row color is required.
	 * @return Color name or the name of a color resource if any, null otherwise.
	 */
	public String getRowColor(LySimpleTableController controller, LyObject object)
	{
		return null;
	}

	/**
	 * Gets the row image of the given table object used on display list mode.
	 *
	 * @param  tableController Table controller that requires the row image.
	 * @param  object          Object to which the row image is associated.
	 * @return Path to the image file or the name of an image resource if found, null otherwise.
	 */
	public String getRowImage(LySimpleTableController tableController,	LyObject object)
	{
		return null;
	}
	
	/**
	 * Gets the header label representing the given field info.<br />
	 * <i>The default implementation</i> returns the {@link
	 * leon.info.LyObjectInfo#getInternalAlias() field internal alias} is it is not null, else
	 * returns the field info label of type name if it is not null, else returns the {@link
	 * leon.info.LyObjectInfo#getAlias() field alias}.
	 *
	 * @param  controller Table controller that requires the header label.
	 * @param  fieldInfo  Field info for which a header label is required.
	 * @return Label that will be put in the table header for the given field info.
	 */
	public String getHeaderLabel(LySimpleTableController controller, LyFieldInfo fieldInfo)
	{
		String header = fieldInfo.getInternalAlias();

		if (header == null)
		{
			LyInfoLabel label = controller.getInfoLabel(fieldInfo, LyInfoLabel.Type.NAME);

			if (label != null)
				header = label.getString((LyObject)null);
		}

		if (header == null)
			header = fieldInfo.getAlias();
		return header;
	}

	/**
	 * Gets the header tooltip associated to the given field info.<br />
	 * <i>The default implementation</i> returns the name, the field unit if any and the field tip.
	 * If the tip is null, the name is used instead.
	 *
	 * @param  controller Table controller that requires the header tooltip.
	 * @param  fieldInfo  Field info for which a header tooltip is required.
	 * @return Label that will be put in the table header tooltip corresponding to the given field
	 *         info.
	 */
	public String getHeaderTooltip(LySimpleTableController controller, LyFieldInfo fieldInfo)
	{
		String		  name = fieldInfo.getName();
		LyInfoLabel   label = controller.getInfoLabel(fieldInfo, LyInfoLabel.Type.NAME);

		if (label != null)
			name = label.getString((LyObject)null);

		if ((name == null)
		        || name.equals("_null"))
		    return null;
		
        LyEnvironment environment = controller.getEnvironment();
	    String        tip = fieldInfo.getTip();
	    LyUnit        unit = fieldInfo.getFieldUnit();
	        
		if (unit != null)
		{
			String        str = unit.getName();

			name = environment.translate(name) + " (" + environment.translate(str) + ")";
		}

		if (tip == null)
			return name;

		return environment.translate(name) + " : " + environment.translate(tip);
	}

	/**
	 * Gets the cell label representing the value of the given field info for the given object.
	 * <br />
	 * <i>The default implementation</i> returns null if the given object is null, calls {@link
	 * #getCellLabel(LySimpleTableController, LyObject, LyFieldInfo, Object)
	 * getCellLabel(controller, object, fieldInfo, object.getValue(fieldInfo))} otherwise.
	 *
	 * @param      controller Table controller that requires the cell label.
	 * @param      object     Object from which the field info value is represented in the table
	 *                        cell.
	 * @param      fieldInfo  Field info for which a cell label is required.
	 * @return     Label that will be put in the table cell for the given object and field info (may
	 *             be null).
	 * @see        #getCellLabel(LySimpleTableController, LyObject, LyFieldInfo, Object)
	 * @deprecated Use {@link LyTableBehaviorInterface#updateCell(LySimpleTableController, LyCell,
	 *             LyObject, LyFieldInfo)} instead.
	 */
	public String getCellLabel(LySimpleTableController controller, LyObject object,
		LyFieldInfo fieldInfo)
	{
		if (object == null)
			return null;

		Object value = object.getValue(fieldInfo);

		return getCellLabel(controller, object, fieldInfo, value);
	}

	/**
	 * Gets the cell label representing the value of the given field info for the given object.
	 * <br />
	 * <i>The default implementation</i> returns null if the given value is null, else if the given
	 * field is a choice field, builds a label from the {@link
	 * leon.control.LyAreaController#getShortLabel(LyObject, LyFieldInfo, Object) short label)}
	 * list. Else if the field is an array, builds a label with sub fields label.
	 *
	 * @param  controller the table controller that requires the cell label
	 * @param  object     the object from which the field info value is represented in the table
	 *                    cell
	 * @param  fieldInfo  the field info for which a cell label is required
	 * @param  value      the given field value
	 * @return Label that will be put in the table cell for the given object and field info (may be
	 *         null).
	 * @see    LyTableBehaviorInterface#updateCell(LySimpleTableController,LyCell,LyObject,LyFieldInfo)
	 */
	public String getCellLabel(LySimpleTableController controller, LyObject object,
		LyFieldInfo fieldInfo, Object value)
	{
		if (value == null)
			return null;

		String label = controller.getShortLabel(object, fieldInfo, value);

		// Optimization : we can not translate all strings in LyCell
		if (fieldInfo.hasMark(LyFieldInfo.MARK_TRANSLATE))
		{
			if (fieldInfo.isMultiple())
			{
				LyViewManager viewManager = controller.getViewManager();
				LyEnvironment environment = controller.getEnvironment();
				String[] labels = LyString.stringToList(label, LyApplication.VISIBLE_FIELD_SEP);
				int		 i, n = ((labels == null) ? 0 : labels.length);
				String   sep = ((fieldInfo instanceof LyChoiceFieldInfo) ? ", " :
					(viewManager.hasProperty("MANAGE_URL") && environment.testResource("LY_WEB_TABLE_TRANSLATE_RELATION_TO_LINK") ?
							"\n" + LyTools.VALUE_SEP : "\n"));

				if ((fieldInfo instanceof LyRelationFieldInfo)
					&& (n > ((LyObjectId[])value).length))
				{
					labels = getShortLabel(controller, object, fieldInfo, value);
					n = labels.length;
				}

				for (i = 0, label = ""; i < n; i++)
				{
					if (i > 0)
						label += sep;
					label += controller.translate(labels[i]);
				}
			}
			else if (fieldInfo instanceof LyArrayFieldInfo)
            {
                String[] labels = LyString.stringToList(label, LyApplication.VISIBLE_FIELD_SEP);
                int      i, n = ((labels == null) ? 0 : labels.length);

                for (i = 0, label = ""; i < n; i++)
                {
                    if (i > 0)
                        label += ", ";
                    label += controller.translate(labels[i]);
                }
            }
			else
				label = controller.translate(label);
		}

		return label;
	} // end method getCellLabel

	/**
	 * Gets short label of an object, for specified field info and value.
	 *
	 * @param  controller the table controller
	 * @param  object     list of Object for which a short label is required.
	 * @param  fieldInfo  field info.
	 * @param  value      the given field value
	 * @return String[] list of label as a string.
	 */
	public String[] getShortLabel(LySimpleTableController controller, LyObject object,
		LyFieldInfo fieldInfo, Object value)
	{
		LyObjectId[] object_lst = (LyObjectId[])value;
		String[]     labels = new String[object_lst.length];
		String		 label;

		for (int i = 0; i < object_lst.length; i++)
		{
			LyObjectId   objectId = object_lst[i];
			LyObjectId[] obj = { objectId };

			label = controller.getShortLabel(object, fieldInfo, obj);
			labels[i] = label;
		}

		return labels;
	} // end method getShortLabel

	/**
	 * Gets the cell object id that displays the value of the given field info for the given object.
	 *
	 * @param  controller Table controller that requires the cell.
	 * @param  object     Object from which the field info value is represented in the table cell.
	 * @param  fieldInfo  Field info for which a cell is required.
	 * @return Object id that will be put in the table cell for the given object and field info (may
	 *         be null).
	 */
	public String getCellObjectId(LySimpleTableController controller, LyObject object,
		LyFieldInfo fieldInfo)
	{
		if (object == null)
			return null;

		Object value = object.getValue(fieldInfo);

		if (value == null)
			return null;

		LyViewManager viewManager = controller.getViewManager();

		if (!viewManager.hasProperty("MANAGE_URL"))
			return null;

		LyClassInfo relationClassInfo;
		LyAction    action;

		if (fieldInfo instanceof LyRelationFieldInfo)
		{
			if (((LyRelationFieldInfo)fieldInfo).isMultiple())
			{
				if (fieldInfo.hasMark(LyFieldInfo.MARK_TRANSLATE))
				{
					StringBuffer	result = null;
					LyObjectId[]	ids = (LyObjectId[])value;
					int				i, n = ids.length;
					
					ids = ((LyRelationFieldInfo)fieldInfo).getOrderedIds(ids);
					
					for (i = 0; i < n; i++)
					{
						relationClassInfo = ids[i].getClassInfo();
						action = relationClassInfo.getActionWithId(LyAction.ID_CONSULT);
						
						if (action == null)
							action = relationClassInfo.getDefaultAction();
	
						if (action != null)
						{
							if (result == null)
								result = new StringBuffer();
							else if (result.length() > 0)
								result.append(LyTools.VALUE_SEP);
							
							result.append(ids[i].getFullString());
						}
					}
					
					return (result == null ? null : result.toString());
				}
			}
			else
			{
				relationClassInfo = ((LyObjectId)value).getClassInfo();
	
				action = relationClassInfo.getActionWithId(LyAction.ID_CONSULT);
				
				if (action == null)
					action = relationClassInfo.getDefaultAction();
	
				if (action != null)
					return ((LyObjectId)value).getFullString();
			}
		}

		return null;
	} // end method getCellObjectId

	/**
	 * Indicates if escape characters must be replaced in labels of the given field info.<br />
	 * <i>The default implementation</i> returns <code>true</code> if:
	 *
	 * <ul>
	 *   <li>the given field is a multiple {@link LyTextFieldInfo text field} and the {@link
	 *     LyTextFieldInfo#getNRows() prefered number of rows} is greater than 0.</li>
	 *   <li>the given field is a {@link LyStructFieldInfo struct field} and if one of its sub
	 *     fields need to convert its escape characters.</li>
	 *   <li>the given field is a {@link LyRelationFieldInfo relation field} and if the given label
	 *     contains the new line character</li>
	 *   <li>the given field is an {@link LyArrayFieldInfo array field}</li>
	 * </ul>
	 *
	 * @param      fieldInfo the field info
	 * @param      label     the label displayed in the table
	 * @return     true if escape characters must be replaced, false otherwise
	 * @deprecated no longer used by generic code.
	 */
	protected boolean convertEscapeChars(LyFieldInfo fieldInfo, String label)
	{
		if (fieldInfo instanceof LyTextFieldInfo)
		{
			// Replace for multiple text only
			if (((LyTextFieldInfo)fieldInfo).isMultiple()
					|| (((LyTextFieldInfo)fieldInfo).getNRows() > 0))
				return true;
		}
		else if (fieldInfo instanceof LyStructFieldInfo)
		{
			// If the struct contains a field that needs the conversion of
			// escape characters, the struct needs conversion too.
			LyStructFieldInfo	    structFieldInfo = (LyStructFieldInfo)fieldInfo;
			LyInfoList<LyFieldInfo> fields = structFieldInfo.getFieldInfos();
			LyFieldInfo			    subFieldInfo;
			int					    i, n = ((fields == null) ? 0 : fields.getSize());

			for (i = 0; i < n; i++)
			{
				subFieldInfo = fields.getInfo(i);
				if (convertEscapeChars(subFieldInfo, label))
					return true;
			}
		}
		else if (fieldInfo instanceof LyRelationFieldInfo)
		{
			// Relations may contain escape characters due to another field of
			// the relation
			if (label.indexOf('\n') >= 0)
				return true;
		}
		else if (fieldInfo instanceof LyArrayFieldInfo)
			// Array needs conversion too
			return true;

		return false;
	} // end method convertEscapeChars

	/**
	 * Updates the label of the given cell for the given object and fieldinfo.<br />
	 * <i>The default implementation</i> updates the given cell label if needed, the cell object id
	 * if the given field is a relation field and if the view manager manages urls. If the given
	 * field is a file field, updates the cell label, Hlink and image.
	 *
	 * @param tableController the table controller that requires the cell label
	 * @param cell            the cell to update
	 * @param object          the object from which the field info value is represented in the table
	 *                        cell
	 * @param fieldInfo       the field info for which a cell label is required
	 */
	public void updateCell(LySimpleTableController tableController, LyCell cell, LyObject object,
		LyFieldInfo fieldInfo)
	{
		String newLabel = getCellLabel(tableController, object, fieldInfo);
		String image = getCellImage(tableController, object, fieldInfo, newLabel);

		LyViewManager viewManager = tableController.getViewManager();

		if ((fieldInfo instanceof LyRelationFieldInfo)
				&& viewManager.hasProperty("MANAGE_URL"))
		{
			String objectId = getCellObjectId(tableController, object, fieldInfo);

			cell.setObjectId(objectId);
		}

		if (image != null)
		{
			cell.setLabel(null);
			cell.setImage(image);
		}
		else
		{
			String label = cell.getLabel();

			if (((label == null) && (newLabel != null))
					|| ((label != null) && !label.equals(newLabel)))
			{
				// If field is multiple, avoid displaying values with more lines than nrows
				// of the field info (display entire value in the tooltip).
				if (fieldInfo.isMultiple()
					&& (!(fieldInfo instanceof LyTextFieldInfo) || (((LyTextFieldInfo)fieldInfo).getType() != LyTextFieldInfo.Type.RICH))
					&& (newLabel != null)
					&& (newLabel.indexOf('\n') >= 0))
				{
                    int          nRows = -1;
                    
                    if (fieldInfo instanceof LyTextFieldInfo)
                        nRows = ((LyTextFieldInfo)fieldInfo).getNRows();
                    else if (fieldInfo instanceof LyRelationFieldInfo)
                        nRows = ((LyRelationFieldInfo)fieldInfo).getNRows();
                    else if (fieldInfo instanceof LyArrayFieldInfo)
                        nRows = ((LyArrayFieldInfo)fieldInfo).getNRows();
                    
					String[]		labels = LyString.stringToList(newLabel, '\n');
					StringBuffer	buffer = new StringBuffer();
					
                    if ((nRows <= 0)
                            || (nRows > MAX_ROW_HEIGHT))
                        nRows = MAX_ROW_HEIGHT;
					
					if (labels.length > nRows)
					{
						for (int i = 0; i < nRows; i++)
						{			
							buffer.append(labels[i]);
                            buffer.append('\n');
						}

						buffer.append("...");
						cell.setLabel(buffer.toString());
						cell.setTooltip(newLabel);
					}
					else
					{
						cell.setLabel(newLabel);
						cell.setTooltip(null);
					}
				}
				else
					cell.setLabel(newLabel);
			}

			cell.setImage(null);
		}

		if ((fieldInfo instanceof LyTextFieldInfo)
				&& (((LyTextFieldInfo)fieldInfo).getType() == LyTextFieldInfo.Type.RICH))
			cell.setType(LyCell.Type.RICH);
		
		if (fieldInfo instanceof LyFileFieldInfo)
		{
			LyFileFieldInfo fileField = (LyFileFieldInfo)fieldInfo;
			LyEnvironment   env = tableController.getEnvironment();
			boolean		    computeURL = true; // Check for Hyperlink
			String		    mimeTypes = fileField.getMimeTypes();

			if (mimeTypes != null)
			{
				mimeTypes = mimeTypes.toLowerCase();

				if (!mimeTypes.startsWith("image"))
				{
					// Link to an attachment.
					if ((newLabel != null)
							&& (((LyFileFieldInfo)fieldInfo).getAttachmentHandler() != null))
					{
						if (mimeTypes.equals("text/x-url"))
							computeURL = false;
					}
					else
						computeURL = false; // Keep compatibility : XXX cleanup

					if (computeURL)
					{
						String docURL = env.getEnv("LY_APP_DOC");
						String fileDir = docURL + "/"
							+ ((LyFileFieldInfo)fieldInfo).getAttachmentDirectory();
						String url = env.getString().expand(fileDir + "/" + newLabel);

						url = url.replace('\\', '/');
						cell.setHLink(url);

						if (image == null)
						{
							// Clean up in display URL which may contain "__"+date+"__"
							int n1 = newLabel.indexOf("__");

							if (n1 >= 0)
							{
								int n2 = newLabel.indexOf("__", n1 + 1);

								if (n2 >= 0)
									cell.setLabel(newLabel.substring(0, n1)
										+ newLabel.substring(n2 + 2));
							}
						}
					}
					else if ((newLabel != null)	&& (newLabel.length() > 0)
							&& (((fieldInfo.getBinding() != null) && "BLOB".equals(fieldInfo.getBinding().getPeerType()) )
									|| "true".equals(fieldInfo.getData("BLOB"))))
					{
						LyViewManager vm = tableController.getViewManager();

						if (vm.hasProperty("MANAGE_ATTACHMENT"))
						{
							// file exists ?
							String path = vm.getTemporaryDir()+File.separator+newLabel;
							path = env.getString().expand(path);

							File f = new File(path);

							vm.getEnvironment().logTrace("Leonardi : LyTableBehavior link to the temporary file (exists "+f.exists()+") "+f.getAbsolutePath());

							if (f.exists())
							{
								String url;
								// We have a BLOB, this blob is saved to a temporary file and we shall
								// have an URL to this temporary location
								String fileDir = vm.getDownloadAction(); 

								if (fileDir  == null)
								{
									fileDir = vm.getTemporaryLocation();
									url = env.getString().expand(fileDir + "/" + newLabel);
									url = url.replace('\\', '/');
								}
								else
								{
									LySession session = tableController.getSession();

									String id = session.storeTempFile(f.getAbsolutePath(), newLabel);
									url = fileDir + id;
								}
								cell.setHLink(url);
							}

						}
					}	
					else if (fileField.getDirectory() != null)
					{
						String directory = fileField.getDirectory();
						String directoryLc = directory.toLowerCase();

						if (!directoryLc.startsWith("http://")
						        && !directoryLc.startsWith("https://")
								&& !directoryLc.startsWith("file://")
								&& !directoryLc.startsWith("ftp://"))
							newLabel = "file://" + directory + newLabel;
					}

					if (cell.getHLink() == null)
						cell.setHLink(newLabel);
				}
			}
		}
	} // end method updateCell

	/**
	 * Gets the cell image representing the value of the specified field info for the given object.
	 *
	 * @param  tableController Table controller that requires the cell image.
	 * @param  object          Object from which the field info value is represented in the table
	 *                         cell.
	 * @param  fieldInfo       Field info for which a cell image is required.
	 * @param  newLabel        Value of the field info given as a string.
	 * @return Image that will be put in the table cell for the given object and field info (may be
	 *         null).
	 */
	public String getCellImage(LySimpleTableController tableController, LyObject object,
		LyFieldInfo fieldInfo, String newLabel)
	{
		if ((newLabel == null)
				|| (newLabel.trim().length() == 0))
			return null;

		// If field is a file field with type image, replace label by an image
		if (!(fieldInfo instanceof LyFileFieldInfo))
			return null;

		LyFileFieldInfo fileField = (LyFileFieldInfo)fieldInfo;
		String		    mimeTypes = fileField.getMimeTypes();
		String		    image = null;

		if ((mimeTypes != null)
				&& mimeTypes.toLowerCase().startsWith("image"))
		{
			LyEnvironment env = tableController.getEnvironment();
			String		  heightValue = env.getResource("LY_CELL_IMAGE_HEIGHT");
			int			  height = LyTools.intFromString(heightValue);

			image = getImagePath(tableController, fileField, newLabel);
			if (height > 0)
				image += " (" + height + ")";
		}

		return image;
	} // end method getCellImage

	/**
	 * Utility method used to retrieve the path of the given file field from the file of the given
	 * label.
	 *
	 * @param  tableController Table controller that requires the cell image
	 * @param  fileField       the file fieldInfo
	 * @param  label           a file name
	 * @return the modified label for the field.
	 */
	protected String getImagePath(LySimpleTableController tableController,
		LyFileFieldInfo fileField, String label)
	{
		// Is it a local file or an external location
		if (label == null)
			return null;

		LyEnvironment environment = tableController.getEnvironment();
		String		  directory = fileField.getDirectory();
		String		  label2 = label;

		if ((directory != null) && !label.toUpperCase().startsWith(directory.toUpperCase()))
		{
			File file = new File(label);

			if (!file.exists())
			{
				if (!directory.endsWith("/") && !directory.endsWith("\\") 
						&& !label.startsWith("/") && !label.startsWith("\\"))
					directory = directory + File.separator;
				
				label2 = directory + label;
				label2 = environment.getString().expand(label2);
				file = new File(label2);
				
				if (!file.exists())
				{
					// TRY FROM CLASSPATH ?
					label2 = directory + label;
					if (label2.indexOf("$LY_APP_DOC_DIR$") >= 0)
					{
						String appId = tableController.getApplication().getId();

						appId = appId.replace('.', '/');
						label2 = LyString.replace(label2, "$LY_APP_DOC_DIR$", appId);
						label2 = label2.replace('\\', '/');

						InputStream iStream = environment.getResourceAsStream(label2, null);

						if (iStream == null)
							label2 = label; // not found
					}
					else
						label2 = label;

				}
			}
		}
		
		LyViewManager vm = tableController.getViewManager();
		if ((label != null)
				&& (fileField.getBinding() != null)
				&& ("BLOB".equals(fileField.getBinding().getPeerType())))
		{	
			if	(vm.hasProperty("MANAGE_ATTACHMENT"))
			{
				int			  n1;
				String url = null;
				label = label.replace('\\', '/');
				n1 = label.lastIndexOf('/');
				if (n1 > 0)
					label = label.substring(n1+1);
				
				String path = vm.getTemporaryDir()+File.separator+label;
				path = environment.getString().expand(path);
				
				File f = new File(path);
				
				if (f.exists() && !tableController.isFopMode())
				{
					// We have a BLOB, this blob is saved to a temporary file and we shall
					// have an URL to this temporary location
					String fileDir = vm.getDownloadAction(); 
					
					if (fileDir  == null)
					{
						fileDir = vm.getTemporaryLocation();
						url = environment.getString().expand(fileDir + "/" + label);
						url = url.replace('\\', '/');
					}
					else
					{
						LySession session = tableController.getSession();
						
						String id = session.storeTempFile(f.getAbsolutePath(), label);
						url = fileDir + id;
					}
				}
				else if (f.exists())
				{
					return f.getAbsolutePath();
				}
				
				
				return url;
			}
			else if (tableController.isFopMode())
			{
				// We have a blob in fop mode
				String fileDir = vm.getTemporaryLocation();
				fileDir = environment.getString().expand(fileDir + "/" + label);
				fileDir = fileDir.replace('\\', '/');
				File f = new File(fileDir);
				if (f.exists())
					return f.getAbsolutePath();
			}
		}
		label2 = label2.replace('\\', '/');

		String docDir = environment.getEnv("LY_APP_DOC_DIR");
		String appDoc = environment.getEnv("LY_APP_DOC");

		if ((docDir != null) && (appDoc != null))
		{
			docDir = docDir.replace('\\', '/');
			appDoc = appDoc.replace('\\', '/');

			if (label2.startsWith(docDir))
				return (appDoc + label2.substring(docDir.length()));
			else
			{
				if (vm.hasProperty("MANAGE_ATTACHMENT")
						&& (directory == null)
						&& (fileField.getAttachmentHandler() != null)
						&& ((fileField.getBinding() == null) || !"BLOB".equals(fileField.getBinding().getPeerType())))
				{
					// File of type image with no directory, no BLOB type and attachment managed by view manager :
					// add attachment directory before image name.
					String attachmentDir = fileField.getAttachmentDirectory();
					
					if (attachmentDir == null)
						attachmentDir = "files";
					
					String fileDir = appDoc + "/" + attachmentDir; // Should be : LyFilePartWriter.FILE_DIR;

					label2 = environment.getString().expand(fileDir + "/" + label2);
					label2 = label2.replace('\\', '/');
				}
				
				return label2;
			}
		}
		else
			return label2;
	} // end method getImagePath

	/**
	 * Gets the footer label associated to the given field info for the given list of objects
	 * presented in the table.
	 *
	 * @param  controller Table controller that requires the footer label.
	 * @param  fieldInfo  Field info for which a footer label is required.
	 * @param  objects    List of objects for which a footer label is computed.
	 * @return Label that will be put in the table footer that is corresponding to the given field
	 *         info for the given list of objects.
	 * @see    leon.info.LyFieldInfo#compute(LyObjectList)
	 */
	public String getFooterLabel(LySimpleTableController controller, LyFieldInfo fieldInfo,
		LyObjectList objects)
	{
		LyEnvironment environment = controller.getEnvironment();
		Object		  value = fieldInfo.compute(objects);

		return environment.translate(fieldInfo.toName(value));
	}

	/**
	 * Method called when the selection has changed in the view.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param controller the current table controller
	 */
	public void selectionChanged(LySimpleTableController controller)
	{
	}

	/**
	 * Gets the content of the table corresponding of the given context.<br />
	 * The workspace must be dynamic for dynamic table updates.<br />
	 * If the returned workspace is null, the controller computes the content itself.<br />
	 * <i>The default implementation</i> does nothing and returns null.
	 *
	 * @param  tableController Table controller that requires the content.
	 * @param  context         Context for which the content is required.
	 * @return the table content.
	 */
	public LyWorkSpace getContent(LyTableController tableController, LyWorkSpace context)
	{
		return null;
	}

	/**
	 * Gets the list of fieldinfos that must be loaded in the given table controller.<br />
	 * <i>The default implementation</i> returns all fieldinfos with the {@link
	 * LyFieldInfo#MARK_STATUS status mark} and all fieldinfos that are displayed in the table.
	 *
	 * @param  controller Table controller that requires the list.
	 * @return The list of fields info to load.
	 */
	public LyInfoList<LyFieldInfo> getLoadedFieldInfos(LySimpleTableController controller)
	{
		LyClassInfo classInfo = controller.getClassInfo();

		if (classInfo == null)
			return null;

		LyInfoList<LyFieldInfo> loadedFields = new LyInfoList<LyFieldInfo>();
		LyInfoList<LyFieldInfo> fields = controller.getFields();
		LyInfoList<LyFieldInfo> statusFields = classInfo.getFieldInfos(LyFieldInfo.MARK_STATUS);

		if (fields != null)
			loadedFields.addList(fields);
		if (statusFields != null)
			loadedFields.addList(statusFields);

		return loadedFields;
	}

	/**
	 * Gets the list of fieldinfos that will be used to compute/display values for the table.<br />
	 * <i>The default implementation</i> returns the list of {@link
	 * #getLoadedFieldInfos(LySimpleTableController) loaded} fields.
	 *
	 * @param  tableController the associated table controller
	 * @param  action          the table action
	 * @return a list of fieldinfos
	 * @see    #getLoadedFieldInfos(LySimpleTableController)
	 */
	public LyInfoList<LyFieldInfo> getDependsFieldInfos(LySimpleTableController tableController,
		LyAction action)
	{
		return getLoadedFieldInfos(tableController);
	}

	/**
	 * Gets the list of fieldinfos that may be displayed in the column chooser tool.<br />
	 * <i>The default implementation</i> does nothing and returns null.
	 *
	 * @param  controller Table controller that requires the list.
	 * @return the list of fields to display in the column chooser tool or null to let the
	 *         controller build this list.
	 */
	public LyInfoList<LyFieldInfo> getChooseColumnsFieldInfos(LySimpleTableController controller)
	{
		return null;
	}

	/**
	 * Gets the initial column width of a fieldInfo in a table.<br />
	 * <i>The default implementation</i> returns the "_colWidth" parameter from the table action.
	 *
	 * @param  controller Table controller that requires the list.
	 * @param  fieldInfo  fieldInfo associated to this column.
	 * @return the size (unit = number of characters) of the column or -1 (default) to compute
	 *         automatically
	 */
	public double getColumnWidth(LySimpleTableController controller, LyFieldInfo fieldInfo)
	{
		if ((controller != null) && (controller.getAction() != null))
		{
			LyAction action = controller.getAction();
			String   columnWithParameter = (String)action.getParameter("_columnWidth", true);

			if (columnWithParameter != null)
			{
				int     n = columnWithParameter.indexOf(fieldInfo.getId() + "=");

				if (n >= 0)
				{
					int        n1 = columnWithParameter.indexOf("##", n);

					if (n1 < 0)
						n1 = columnWithParameter.length(); // Not found, fetch
														   // till the end

					int        start = n + (fieldInfo.getId() + "=").length();
					String     valStr = columnWithParameter.substring(start, n1);
					int        l = n1 - start;
					boolean    percent = (valStr.charAt(l - 1) == '%');

					if (percent)
					    valStr = valStr.substring(0, l - 1);

					try
					{
						double    val = Double.parseDouble(valStr);

						return (percent ? val / 100 : val);
					}
					catch (NumberFormatException e)
					{
					}
				}
			}
		}
		
		return -1;
	} // end method getColumnWidth

	/**
	 * Method invoked at the beginning of the given object drag.<br />
	 * <i>The default implementation</i> keep in the controller session memory that the object is
	 * dragged and log a trace.
	 *
	 * @param tableController Table controller that displays the table
	 * @param draggedObject   dragged object
	 */
	public void dragStart(LySimpleTableController tableController, LyObject draggedObject)
	{
		LyEnvironment environment = tableController.getEnvironment();
		LySession     session = tableController.getSession();

		session.addValue(LySession.DRAGGED_OBJECT, draggedObject);
		environment.logTrace("Drag is starting with object: " + draggedObject);
	}

	/**
	 * Method called when the given dragged object is dropped on the given dropped object.<br />
	 * <i>The default implementation</i> only logs a trace.
	 *
	 * @param tableController table controller that displays the table
	 * @param droppedObject   object on where the dragged object is dropped
	 * @param draggedObject   the dragged object
	 */
	public void drop(LySimpleTableController tableController, LyObject droppedObject,
		LyObject draggedObject)
	{
		LyEnvironment environment = tableController.getEnvironment();

		environment.logTrace("Object: " + droppedObject + " is dropping.");
	}

	/**
	 * Method called at the end of a drag-and-drop.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param tableController table controller that displays the tree
	 * @param type            drag-and-drop type ({@link org.eclipse.swt.dnd.DND#DROP_COPY} or
	 *                        {@link org.eclipse.swt.dnd.DND#DROP_MOVE})
	 */
	public void dragFinished(LySimpleTableController tableController, int type)
	{
		/*
		 * String dropType;
		 *
		 * switch(type) { case DND.DROP_COPY: dropType = "Copy (Ctrl + Click)";
		 * break; case DND.DROP_MOVE: dropType = "Move (Shift + Click) or
		 * default"; break; //case DND.DROP_LINK: //dropType = "Link (Ctrl +
		 * Shift + Click)"; //break; default: dropType = "None"; break; }
		 *
		 * LyEnvironment environment = tableController.getEnvironment();
		 *
		 * environment.logTrace("The drag-and-drop type is: " + dropType);
		 */
	}
    
    /**
     * Method used to enable a sort.
     *
     * @param tableController Table controller that asks for the sort.
     * @param sort            The sort that is asked.
     * @return The enabled sort.
     */
    public LySort validateSort(LySimpleTableController tableController, LySort sort)
    {
        return sort;
    }
    
    /**
     * Enables groupBy feature on given field info.
     * 
     * @param tableController the table controller
     * @param fieldInfo Field info that is checked for groupBy feature.
     * @return True if enabled, false otherwise.
     */
    public boolean enableGroupBy(LySimpleTableController tableController, LyFieldInfo fieldInfo)
    {
        LyEnvironment   environment = tableController.getEnvironment();
        
        if (environment.testResource("LY_DISABLE_GROUP_BY_ON_PAGINATED_TABLES"))
        {
            String limitParameter = (String)tableController.getParameter("_limit");
    
            if (limitParameter != null) // Paginated table
                return false;
        }
        
        // Disable hidden and secret field
        if (fieldInfo.hasMark(LyFieldInfo.MARK_HIDDEN)
                && !fieldInfo.hasMark(LyFieldInfo.MARK_TABLE))
            return false;
        if (fieldInfo.hasMark(LyFieldInfo.MARK_SECRET))
            return false;
        
         // Enable in other cases
         return true;
    }
    
    /**
     * Gets group name for given value on specified object.
     * 
     * @param  tableController Table controller that requires the group value label.
     * @param  object          Object from which the group is requested.
     * @param  value           Field value for which the group is requested.
     * @return The group label for given value.
     */
    public String getGroupLabel(LySimpleTableController tableController, LyObject object, LyFieldInfoValue value)
    {
    	String result = null;
    	
    	if(value == null
    	   || value.getValue() == null)
    	{
    		return result;
    	}
    	else if(value instanceof LyDateValue)
    	{
    		Date date = (Date)value.getValue();
    		LyEnvironment env = tableController.getEnvironment();
    		
    		DateFormat dateFormat = env.dateFormatFactory(LyEnvironment.TYPE_DATE);
    		result = dateFormat.format(date);
    	}
    	else
    	{
   			result = value.getName();
    	}
    	return result;
    }
    
    /**
     * Gets row color associated to the given group.
     *
     * @param  tableController Table controller that requires the row color.
     * @param  object          Group for which a row color is asked.
     * @param  value           the value needed to compute the color
     * @return Color name or the name of a color resource if any, null otherwise.
     */
    public String getGroupColor(LySimpleTableController tableController, LyObject object, 
    		LyFieldInfoValue value)
    {
        return null; // Use default color
    }
    
    /**
     * Current builder of a new cell in a row.
     *
     * @param  tableController Table controller that requires the new row item.
     * @param  object object added in the contextual list, and which should set the content of row
     *                cells.
     * @param  fieldInfo  field value shown in the cell content
     * @return LyCell or LyFieldItem the new Cell description (XML description) or editable content
     *         to add in the LyRow component.
     */
    public LyComponent buildRowItem(LySimpleTableController tableController, LyObject object, 
    		LyFieldInfo fieldInfo)
    {
        return null;
    }
    
    /**
     * Updates the given component for the given object and fieldinfo.
     *
     * @param tableController the table controller that requires the update
     * @param component       the component to update.
     * @param object          the object from which the field info value is represented in the table.
     * @param fieldInfo       the field info for which a component is updated.
     */
    public void updateRowItem(LySimpleTableController tableController, LyComponent component, LyObject object,
        LyFieldInfo fieldInfo)
    {
    }
    
    /**
     * Completes the given row that corresponds to specified object.
     *
     * @param tableController the table controller that requires the completion.
     * @param row             the row to be completed.
     * @param object          the object that corresponds to the row.
     */
    public void completeRow(LySimpleTableController tableController, LyRow row, LyObject object)
    {
    }

    /**
     * Gives the image to display in first column in list display.
     * @param tableController the table controller.	  
     * @return a {@link LyFileFieldInfo} which contains the image else return null
     */
	public LyFileFieldInfo getDisplayListFirstImage(LySimpleTableController tableController)
	{
		return null;
	}
	
    /**
     * Gets label for object.
     * @param tableController	the table controller.
     * @param object          	Object for which the label is requested.
	 * @return The label for given object.
     */
    public String getLabel(LySimpleTableController tableController, LyObject object)
    {
    	return object.getName();
    }
   
    /**
     * Gets index for object.
     * @param tableController	the table controller.
     * @param object          	Object for which the index is requested.
	 * @return The index for given object.
     */
    public String getIndex(LySimpleTableController tableController, LyObject object)
    {
    	LySort	sort = tableController.getSort();
    	
        if ((sort != null)
        	&& (sort.getCriterionCount() > 0))
        {
        	// Set index if possible
        	LyFieldInfo	fieldInfo = sort.getFieldInfos().getInfo(0);
        	
            if (fieldInfo instanceof LyTextFieldInfo)
            {
            	String	indexValue = (String)object.getValue(fieldInfo);
            	
            	if ((indexValue != null)
            			&& (indexValue.length() > 0))
            		return indexValue.substring(0, 1);
            }
        }
        
        return null;
    }
    
    /**
     * Return default criterion for a given field info. If no default sort is defined on the class,
     * return null.
     * @param tableController	the table controller.
     * @param  fieldInfo the field info we want to get the criterion
     * @return the default sort criterion for the given field info
     */
    public LySortCriterion getDefaultCriterion(LySimpleTableController tableController, LyFieldInfo fieldInfo)
    {
    	return null;
    }
} // end class LyTableBehavior
