/*
 * LyTreeBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.LyActionDemand;
import leon.view.event.LyDemand;

import java.util.Vector;


/**
 * This class implements the default behavior for a LyTreeController.
 *
 * @author  Lyria S.A.
 * @version $Id: LyTreeBehavior.java 35914 2012-06-15 13:44:56Z jlsaouli $
 * @see     leon.control.LyTreeController
 * @see     leon.app.behaviorinterface.LyTreeBehaviorInterface
 */
public class LyTreeBehavior extends LyPanelBehavior implements LyTreeBehaviorInterface
{

    /**
     * File version
     */
    public static final String __VERSION =
        "$Id: LyTreeBehavior.java 35914 2012-06-15 13:44:56Z jlsaouli $";

    /**
     * Default constructor.<br />
     * <i>The default implementation</i> does nothing.
     */
    public LyTreeBehavior()
    {
    }

    /**
     * Indicates if data in the tree are sorted or not.<br />
     * <i>The default implementation</i> does nothing and return true.
     *
     * @param  controller current Tree Controller
     * @return true if the tree must be sorted false otherwise
     */
    public boolean isSorted(LyTreeController controller)
    {
        return true;
    }

    /**
     * Gets children of the given parent node.<br />
     * <i>The default implementation</i> does nothing and return null.
     *
     * @param      parentNodeObject parent of the requested objects.
     * @param      classInfos       target (only for data tree)
     * @param      controller       current LyTreeController instance
     * @return     the list of children objects, or null if the subtree is empty
     * @see        #getSons(LyObject, String, LyTreeController)
     * @deprecated use {@link #getSons(LyObject, String, LyTreeController)}
     */
    public LyObjectList getTreeSons(Object parentNodeObject, LyInfoList<LyClassInfo> classInfos,
        LyTreeController controller)
    {
        return null;
    }

    /**
     * Gets children objects of the given parent.<br />
     * A null parent object is considered as the root node.
     *
     * @param      parentNodeObject parent of the requested objects.
     * @param      level            the current level of the tree, as declared in the metamodel
     * @param      controller       the current tree controller
     * @return     the list of children objects, or null if the subtree is empty
     * @deprecated Since V3.3, use {@link #getSons(LyObject, LyTreeNode, String, LyTreeController)}.
     */
    public LyObjectList getSons(LyObject parentNodeObject, String level,
        LyTreeController controller)
    {
        return null;
    } // end method getSons

    /**
     * Gets all children objects of the given parent. A null parent object is considered as the root
     * node.
     *
     * @param  parentNodeObject parent of the requested objects.
     * @param  treeNode         the current tree node associated to given object
     * @param  level            the current level of the tree, as declared in the metamodel
     * @param  controller       the current tree controller
     * @return the list of children objects, or null if the subtree is empty
     */
    public LyObjectList getSons(LyObject parentNodeObject, LyTreeNode treeNode, String level,
        LyTreeController controller)
    {
        LySession               session = controller.getSession();
        LyObjectList            context = controller.getContext();
        LyEnvironment           environment = controller.getEnvironment();
        LyInfoList<LyClassInfo> classes = null;
        String                  currentLevel;
        int                     n, i = 1;
        LyClassInfo             classInfo;
        LyObjectList            result = null, list = null, parentList;
        Vector<LyFilter>        filters = null;
        LyFilter                filter;
        Vector<LySort>          sorts = null;
        LySort                  sort;
        boolean                 hasSort = false;

        if (controller.isDiscoverOnClick() && (treeNode != null))
            treeNode.setExpandable(true);

        if (level == null)
            level = "";

        // Gets Classes sorts and filters of the given tree level and sub levels.
        do
        {
            currentLevel = level + String.valueOf(i);
            classInfo = controller.getClassInfoFromLevel(currentLevel);

            if (classInfo == null)
            {
                currentLevel = level + "*";
                classInfo = controller.getClassInfoFromLevel(currentLevel);
            }

            filter = getFilter(controller, currentLevel);

            if (filter == null)
                filter = controller.getFilterFromLevel(currentLevel);

            sort = getSort(controller, currentLevel);

            if (sort == null)
                sort = controller.getSortFromLevel(currentLevel);

            if (classInfo != null)
            {

                if (classes == null)
                    classes = new LyInfoList<LyClassInfo>();
                else if (classes.contains(classInfo))
                    break;

                classes.addInfo(classInfo);

                if (filters == null)
                    filters = new Vector<LyFilter>();

                filters.add(filter);

                if (sorts == null)
                    sorts = new Vector<LySort>();

                sorts.add(sort);

                if (sort != null)
                    hasSort = true;
            }

            i++;
        }
        while (classInfo != null);

        // If no classes are defiend in the given level and sub levels, there can not be children.
        if (classes == null)
            return null;

        // Check if deprecated methods are used
        Object node;

        if (parentNodeObject != null)
            node = parentNodeObject;
        else
            node = getTreeRootObject(controller);

        // Deprecated method getTreeSons
        list = getTreeSons(node, classes, controller);

        if (list != null)
        {
            environment.logWarning("Using deprecated method getTreeSons in " + this.getClass());
            return list;
        }

        // Deprecated method getSons(LyObject, String, LyTreeController)
        list = getSons(parentNodeObject, level, controller);

        if (list != null)
        {
            environment.logWarning(
                "Using deprecated method getSons(LyObject, String, LyTreeController) in "
                + this.getClass());
            return list;
        }

        result = new LySimpleObjectList();
        result.setSorted(false);

        // Gets children of node
        n = classes.getSize();

        for (i = 0; i < n; i++)
        {
            classInfo = classes.getInfo(i);
            filter = filters.elementAt(i);
            sort = sorts.elementAt(i);

            if (filter != null)
            {

                if (parentNodeObject == null)
                {

                    if (context == null)
                        list = new LyClassWorkSpace(session, classInfo, filter, false);
                    else
                        list = context.getList(session, classInfo, filter, false);
                }
                else
                {
                    parentList = new LySimpleObjectList(parentNodeObject);
                    list = parentList.getList(session, classInfo, filter, false);
                }
            }
            else
            {

                if (parentNodeObject == null)
                {

                    if (context == null)
                        list = new LyClassWorkSpace(session, classInfo, false);
                    else
                        list = context.getList(session, classInfo, false);
                }
                else if (parentNodeObject.instanceOf(classInfo))
                {
                    // Look for sub-objects of the same class
                    LyInfoList<LyRoute> routes = classInfo.getRoutes(classInfo);
                    int                 r, nr = ((routes == null) ? 0 : routes.getSize());
                    LyRoute             route;

                    parentList = new LySimpleObjectList(parentNodeObject);

                    for (r = 0; r < nr; r++)
                    {
                        route = routes.getInfo(r);
                        list = parentList.getList(session, classInfo, route, false);
                    }
                }
                else
                    list = parentNodeObject.getList(session, classInfo, false,
                            isSorted(controller));
            }

            if (list != null)
            {

                if (isSorted(controller))
                {

                    if (sort != null)
                        list.setSort(sort);
                    else if (hasSort)
                        list.setSortByName(true);
                }

                if (result == null)
                    result = list;
                else
                    result.addList(list);
            }
        }

        // Only sort by name if the tree is sorted and if no other sort was defined.
        if ((result != null) && isSorted(controller))
        {

            if (this instanceof LyLabelBehaviorInterface)
            {
                LyInfoLabel label = ((LyLabelBehaviorInterface) this).getInfoLabel(controller,
                        classInfo, LyInfoLabel.Type.NAME);

                if (label != null)
                {
                    result = result.getSortedList(label);
                    hasSort = true;
                }
            }

            if (!hasSort)
            {
                result.setSorted(true);
                result.setSortByName(true);
            }
        }

        return result;
    } // end method getSons

    /**
     * Gets the given node parent objects list.
     *
     * @param  nodeObject the node object
     * @param  controller current LyTreeController instance
     * @return the list of possible given node parents
     */
    public LyObjectList getParents(Object nodeObject, LyTreeController controller)
    {
        Vector<LyFilter> filters = null;

        if (!(nodeObject instanceof LyObject))
            return null;

        LyObject                object = (LyObject) nodeObject;
        LyClassInfo             objectClassInfo = object.getClassInfo();
        LyInfoList<LyClassInfo> parentList = controller.getParentClassInfos(objectClassInfo);
        Vector<String>          levels = controller.getLevels(objectClassInfo);
        int                     i, n = ((levels == null) ? 0 : levels.size());
        String                  level, parentLevel;
        LyFilter                filter;
        LyClassInfo             parentClass;

        for (i = 0; i < n; i++)
        {

            if (filters == null)
                filters = new Vector<LyFilter>();

            level = levels.elementAt(i);

            filter = getFilter(controller, level);

            if (filter == null)
                filter = controller.getFilterFromLevel(level);

            parentLevel = level.substring(0, level.length() - 1);
            parentClass = controller.getClassInfoFromLevel(parentLevel);

            if ((filter != null) && !filter.match(nodeObject))
            {

                if (parentList.getSize() < 1)
                    parentList = null;
                else
                    parentList.removeInfo(parentClass);
            }
        }

        if (parentList == null)
            return new LySimpleObjectList();

        // retrieve parents
        LySession    session = controller.getSession();
        LyObjectList result = null;
        LyObjectList list = null;
        LyClassInfo  parentClassInfo;

        for (i = 0, n = parentList.getSize(); i < n; i++)
        {
            parentClassInfo = parentList.getInfo(i);
            list = object.getList(session, parentClassInfo, false, isSorted(controller));

            if (result == null)
                result = list;
            else
                result.addList(list);
        }

        return result;
    } // end method getParents

    /**
     * Gets the tree root node.<br />
     * <i>The default implementation</i> gets the <b>LY_TREE_ROOT</b> environment resource and
     * returns the String representation as show in the tree.
     *
     * @param  controller current LyTreeController instance
     * @return the object shown as the root of the tree, or null if the tree is empty
     */
    public Object getTreeRootObject(LyTreeBaseController controller)
    {
        LyEnvironment environment = controller.getEnvironment();
        String        root = environment.translate("LY_TREE_ROOT");

        return root.intern();
    }

    /**
     * Build a view object instance from the given application object. Node status (colors, borders,
     * ...) should be set in {@link #updateTreeNode(Object, LyTreeNode, LyTreeBaseController)}.
     * <br />
     * <i>The default implementation</i> just returns null if the given object is null or constructs
     * a {@link LyTreeNode#LyTreeNode() LyTreeNode} and returns it otherwise.
     *
     * @param  nodeObject the node to build, usually {@link LyObject} instance but not mandatory
     * @param  controller current LyTreeController instance
     * @return the built node view object.
     */
    public LyTreeNode buildTreeNode(Object nodeObject, LyTreeBaseController controller)
    {

        if (nodeObject == null)
            return null;

        // id is computed in controller
        LyTreeNode treeNode = new LyTreeNode();

        return treeNode;
    }

    /**
     * Updates the given tree node view object status (colors, label, border, style ...) from the
     * given node object properties. The controller invokes this method when the given nodeObject is
     * modified.<br />
     * <i>The default implementation</i>, if the given node object is a string, sets the label with
     * its value, sets image to null and open the node. If the given object is a {@link LyObject},
     * sets the label, the image, th open property and the hidden property with the given object
     * one.
     *
     * @param nodeObject application object (reference)
     * @param treeNode   view objects (need update)
     * @param controller current tree controller instance
     */
    public void updateTreeNode(Object nodeObject, LyTreeNode treeNode,
        LyTreeBaseController controller)
    {

        if ((nodeObject == null) || (treeNode == null))
            return;

        if (nodeObject instanceof String)
        {
            treeNode.setLabel((String) nodeObject);
            treeNode.setImage(null);
            treeNode.setOpen(true); // Default is opened.
        }
        else if (nodeObject instanceof LyObject)
        {
            LyObject    object = (LyObject) nodeObject;
            String      image = getImage(controller, object);
            Integer     status;

            treeNode.setLabel(controller.getName(object));

            // We may handle error icon in addition to the object icon
            if (object != null)
            {
                LyViewManager viewManager = controller.getViewManager();

                image = viewManager.buildErrorImage(image, object.getErrorLevel());
            }

            treeNode.setImage(image);

            status = ((LyTreeController) controller).getExpandedStatusFromNode(treeNode);

            if (status == LyTreeController.STATUS_OPENED)
                treeNode.setOpen(true);

            treeNode.setHidden(object.isHidden());
        }
    } // end method updateTreeNode

    /**
     * Gets image for given object.
     * @param controller Tree controller.
     * @param object	 Object for which an image is searched.
     * @return Image name.
     */
    protected String getImage(LyTreeBaseController controller, LyObject object)
    {
    	if (object == null)
    		return null;
    	
    	// Look for a specific image
    	LyClassInfo	 			classInfo = object.getClassInfo();
    	int					    imageMark = LyFieldInfo.getMark("_TREE_IMAGE");
    	LyInfoList<LyFieldInfo> fieldInfos = classInfo.getFieldInfos(imageMark);
    	
    	if (fieldInfos != null)
    	{
	    	int				i, n = fieldInfos.getSize();
			StringBuffer 	treeImage = null;
			LyFieldInfo  	fieldInfo;
			Object		 	value;
	
			for (i = 0; i < n; i++)
			{
				fieldInfo = fieldInfos.getInfo(i);
				value = object.getValue(fieldInfo);
				if (value == null)
					continue;
				
				// Create new buffer or append if needed
				if (treeImage == null)
					treeImage = new StringBuffer();
				else
					treeImage.append('|'); // XXX
	
				// Append image path
				treeImage.append(fieldInfo.toString(value));
			}
			
			if (treeImage != null)
				return controller.getImageFromPrefix(treeImage.toString());
    	}

		// Get generic image
	    return controller.getImage(classInfo);
    }
    
    /**
     * Gets the list of classInfo that represent the possibles children of the given classInfo at
     * the given level.<br />
     * <i>The default implementation</i> gets the {@link LyTreeController#getSonsClassInfos(String)
     * controller children class infos} and returns a list containing those classes and their {@link
     * LyClassInfo#getSubClasses() children} or null if the controller does not have children classe
     * info.
     *
     * @param  parentClassInfo the classInfo of the parent
     * @param  parentLevel     the current level of the parent
     * @param  controller      the current tree controller
     * @return a list of classInfo
     */
    public LyInfoList<LyClassInfo> getSonsClassInfos(LyClassInfo parentClassInfo,
        String parentLevel, LyTreeController controller)
    {
        LyInfoList<LyClassInfo> tempClassInfoList;
        LyInfoList<LyClassInfo> classInfoList = new LyInfoList<LyClassInfo>();

        if ((parentLevel != null) && (parentLevel.length() >= 1))
            tempClassInfoList = controller.getSonsClassInfos(parentLevel);
        else
            tempClassInfoList = controller.getRootClassInfos();

        if ((tempClassInfoList == null) || (tempClassInfoList.getSize() < 1))
            return tempClassInfoList;

        int         i, ni;
        LyClassInfo son;

        for (i = 0, ni = tempClassInfoList.getSize(); i < ni; i++)
        {
            son = tempClassInfoList.getInfo(i);
            classInfoList.addInfo(son);
            classInfoList.addList(son.getSubClasses(true));
        }

        return classInfoList;
    }

    /**
     * Gets the list of all classInfos that can be found in the tree.<br />
     * <i>The default implementation</i> retreives the {@link LyTreeController#getClassInfos()
     * controller class infos}.
     *
     * @param  controller the current tree controller
     * @return the list of classInfo
     */
    public LyInfoList<LyClassInfo> getClassInfos(LyTreeController controller)
    {
        return controller.getClassInfos();
    }

    /**
     * Method invoked when the selection has changed in the tree view.<br />
     * <i>The default implementation</i> does nothing.
     *
     * @param controller the current tree controller
     */
    public void selectionChanged(LyTreeController controller)
    {
    }

    /**
     * Gets the sort used in the tree for the given level.<br />
     * <i>The default implementation</i> does nothing and returns null.
     *
     * @param  controller the current tree controller
     * @param  level      the current level on tree
     * @return the found sort
     */
    public LySort getSort(LyTreeController controller, String level)
    {
        return null;
    }

    /**
     * Gets the filter used in the tree for the given level.<br />
     * <i>The default implementation</i> does nothing and returns null.
     *
     * @param  controller the current tree controller
     * @param  level      the current level on tree
     * @return the found filter
     */
    public LyFilter getFilter(LyTreeController controller, String level)
    {
        return null;
    }

    /**
     * Method invoked when a validity changed in a three state tree (valid, invalid, partially).
     * <br />
     * <u>Note:</u> a tree is activated as a three state validity tree by the resource :
     * LY_TREE_VALIDITY=true.<br />
     * <i>The default implementation</i> does nothing.
     *
     * @param treeNode   view objects (need update)
     * @param controller current LyTreeBaseController instance
     */
    public void validityModified(LyTreeNode treeNode, LyTreeBaseController controller)
    {
    }

    /**
     * Gets the ordered list of pairs of (action, classInfo) to display in popup menu.<br />
     * <i>The default implementation</i> does nothing and returns the given action.
     *
     * @param  treeController Tree controller that displays the menu.
     * @param  actions        List of pairs (action, classInfo) that the controller is about to
     *                        present in popup menu.
     * @param  classInfo      Class of the selected object.
     * @param  sons           List of classes declared as sons of the given class info.
     * @return List of pairs (action, classInfo) that the controller will display in menu.
     */
    public LyInfoDataList<LyAction, LyClassInfo> getExtraPopupMenuActions(
        LyTreeController treeController, LyInfoDataList<LyAction, LyClassInfo> actions,
        LyClassInfo classInfo, LyInfoList<LyClassInfo> sons)
    {
        return actions;
    }

    /**
     * Method invoked at the beginning of the given object drag.<br />
     * <i>The default implementation</i> keep in the controller session memory that the object is
     * dragged and log a trace.
     *
     * @param treeController current LyTreeController instance
     * @param draggedObject  dragged object
     */
    public void dragStart(LyTreeController treeController, LyObject draggedObject)
    {
        LyEnvironment environment = treeController.getEnvironment();
        boolean       debugTrace = environment.isDebugLevel(LyEnvironment.DEBUG_TRACE);
        LySession     session = treeController.getSession();

        session.addValue(LySession.DRAGGED_OBJECT, draggedObject);

        if (debugTrace)
            environment.logTrace("Drag is starting with object: " + draggedObject);
    }

    /**
     * Method called when the given dragged object is dropped on the given dropped object.<br />
     * <i>The default implementation</i> only logs a trace.
     *
     * @param treeController current LyTreeController instance
     * @param droppedObject  object on where the dragged object is dropped
     * @param draggedObject  the dragged object
     */
    public void drop(LyTreeController treeController, LyObject droppedObject,
        LyObject draggedObject)
    {
        LyEnvironment environment = treeController.getEnvironment();
        boolean       debugTrace = environment.isDebugLevel(LyEnvironment.DEBUG_TRACE);

        if (debugTrace)
            environment.logTrace("The object : " + droppedObject + " is dropping.");
    }

    /**
     * Handle a demand raised by the view
     *
     * @param controller the controller
     * @param demand the view demand
     *
     * @return true if the demand was handled
     */
    @Override public boolean handleDemand(LyController controller, LyDemand demand)
    {

        if (!(controller instanceof LyTreeController))
            return super.handleDemand(controller, demand);

        LyTreeController treeController = (LyTreeController) controller;

        if (treeController.isDiscoverOnClick())
        {

            if ("TYPE_EXPAND".equals(demand.getDemandType()))
            {
                LyActionDemand actionDemand = (LyActionDemand) demand;
                String         actionDemandValue = actionDemand.getValue();

                //In Swing mode, the action demand value ends with |true or |false to determinate if the node
                //is being opened or closed.
                //In Web2.0, there is not indication on the status of the node (opened or closed).

                String nodeId = null;

                if (actionDemandValue.endsWith("|true"))
                {
                    int pos = actionDemandValue.indexOf("|true");

                    nodeId = actionDemandValue.substring(0, pos);
                }
                else if (actionDemandValue.endsWith("|false"))
                {
                    int pos = actionDemandValue.indexOf("|false");

                    nodeId = actionDemandValue.substring(0, pos);
                }
                else
                    nodeId = actionDemandValue;

                LyObjectId parentNodeObjectId = treeController.getObjectIdFromTreeNodeId(nodeId);

                if (parentNodeObjectId == null)
                    return super.handleDemand(controller, demand);

                LyTreeNode treeNode = (LyTreeNode) treeController.getView().getComponent(nodeId);
                discoverTreeNode(treeController, treeNode);
            }
        }

        return super.handleDemand(controller, demand);
    }

    /**
     * Discover the content of a treenode
     *
     * @param treeController the tree controller
     * @param treeNode the tree node to discover
     */
    protected void discoverTreeNode(LyTreeController treeController, LyTreeNode treeNode)
    {
        String     currentLevel = treeController.getLevel(treeNode.getInternalId());
        LyObjectId parentNodeObjectId = treeController.getObjectIdFromTreeNodeId(treeNode.getId());

        if (currentLevel == null)
            currentLevel = "";

        treeController.setIgnoreIdle(true);

        LyObjectList sons = getSons(parentNodeObjectId.getLogicalObject(), treeNode, currentLevel,
                treeController);

        if ((sons == null) || (sons.getSize() == 0))
            treeNode.setExpandable(false);
        else
        {
            treeNode.setExpandable(true);

            LyTree tree = treeController.getTreeComponent();
            tree.setIdle(true);

            LyObject       son;
            Vector<String> levels;
            int            l, nl, i, nbSons = ((sons == null) ? 0 : sons.getSize());
            String         temp;
            int            size = currentLevel.length();
            String         nextLevel = null;

            for (i = 0; i < nbSons; i++)
            {
                son = sons.getObject(i);

                levels = treeController.getLevels(son.getClassInfo());
                nextLevel = null;

                if (levels != null)
                {

                    for (l = 0, nl = levels.size(); l < nl; l++)
                    {
                        temp = levels.elementAt(l);

                        if ((temp.length() == (size + 1))
                                && temp.substring(0, size).equals(currentLevel))
                        {
                            nextLevel = temp;
                            break;
                        }
                    }
                }
                else
                    //By default, if we haven't information about level to add the new node, we take the level of the parent
                    //node increasing by one.
                    nextLevel = currentLevel + "1";

                treeController.addTreeNode(treeNode, sons.getObject(i), nextLevel);
            }

            tree.setIdle(false);
        }

        treeController.setIgnoreIdle(false);
    }

    /**
     * Method called at the end of a drag-and-drop.<br />
     * <i>The default implementation</i> does nothing.
     *
     * @param treeController current LyMapController instance
     * @param type           drag-and-drop type ({@link org.eclipse.swt.dnd.DND#DROP_COPY} or {@link
     *                       org.eclipse.swt.dnd.DND#DROP_MOVE})
     */
    public void dragFinished(LyTreeController treeController, int type)
    {
        // LyEnvironment environment = treeController.getEnvironment();
        // boolean debugTrace =
        // environment.isDebugLevel(LyEnvironment.DEBUG_TRACE);
        // String dropType;
        //
        // switch (type)
        // {
        // case org.eclipse.swt.dnd.DND.DROP_COPY:
        // dropType = "Copy (Ctrl + Click)";
        // break;
        // case org.eclipse.swt.dnd.DND.DROP_MOVE:
        // dropType = "Move (Shift + Click or default)";
        // break;
        // //case DND.DROP_LINK:
        // //dropType = "Link (Ctrl + Shift + Click)";
        // //break;
        // default:
        // dropType = "None";
        // break;
        // }
        //
        // if (debugTrace)
        // environment.logTrace("The drag-and-drop type is: " + dropType);
    }
} // end class LyTreeBehavior
