/*
 * LyUserInputBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * This interface defines the default behavior when the data are submited.
 *
 * @author  Lyria S.A.
 * @version $Id: LyUserInputBehavior.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @see     leon.control.LyUserInputController
 */
public class LyUserInputBehavior extends LyActionBehavior
	implements LyUserInputBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyUserInputBehavior.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Method invoked when the user validates the content of the input.<br />
	 * <i>The default implementation</i>, depending on the user validation:
	 *
	 * <ul>
	 *   <li>{@link LyFormController#DEMAND_VALIDATE}: calls the {@link #validate(LyController,
	 *     LyParameterValues)} method then hides the controller view anfd frees it.</li>
	 *   <li>{@link leon.control.LyAreaController#DEMAND_CANCEL}: calls the {@link
	 *     #cancel(LyController)} method then hides the controller view anfd frees it.</li>
	 *   <li>{@link LyFormController#DEMAND_VALIDATE}: hides the controller view.</li>
	 * </ul>
	 *
	 * @param controller the current LyUserInputController instance
	 * @param demand     demand parameters sent by the view component.
	 */
	public void userInputDemand(LyController controller, LyDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			// Gets current action (source of the demand)
			String              id = demand.getSource();
            LyParameterValues   parameterValues = demand.getParameterValues();

			if (id.equals(LyFormController.DEMAND_VALIDATE))
			{
				if (validate(controller, parameterValues))
    				close(controller);
			}
            else if (id.equals(LyFormController.DEMAND_APPLY))
            {
                if (validate(controller, parameterValues))
                {
                    if (controller.hasViewController())
                        controller.hideView();
                }
            }
			else if (id.equals(LyFormController.DEMAND_CANCEL))
			{
				cancel(controller);
				close(controller);
			}
		}
	}

	/**
	 * Method invoked when the user validates the content of the input.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param controller the current LyUserInputController instance
	 * @param values     values set by the user.
     * @return True if validation is successfull, false otherwise. 
	 */
	public boolean validate(LyController controller, LyParameterValues values)
	{     
        if (!(controller instanceof LyUserInputController))
            return false;
        
        // Check that all mandatory fields have values
        return checkMandatoryFields((LyUserInputController)controller, values);
	}
    
    /**
     * Check mandatory fields.
     * @param controller the current LyUserInputController instance
     * @param values     values set by the user.
     * @return True if checking is successfull, false otherwise.
     */
    protected boolean checkMandatoryFields(LyUserInputController controller, LyParameterValues values)
    {
        LyForm          form = controller.getFormComponent();
        
        if (form == null)
            return false;
        
        LyFormField     field;
        LyCaption       caption;
        int             i, n = form.getFormFieldCount();
        LyEnvironment   environment = controller.getEnvironment();
        String          value, label, message;
        
        for (i = 0; i < n; i++)
        {
            field = form.getFormField(i);
            if (!field.isHidden()
                && !field.isOptional())
            {
                if (values == null)
                    value = null;
                else
                    value = values.getParameter(field.getId());
                
                if ((value == null)
                     || (value.length() == 0))
                {
                    field.setState(LyFormField.STATE_ERROR);
                    caption = field.getCaption();
                    label = (caption == null ? field.getId() : caption.getLabel());
                    
                    message = environment.getMessage("LY_ER_FIELD_MANDATORY", label);
                    environment.raiseError(message);
                    return false;
                }
            }

            field.setState(LyFormField.STATE_NORMAL);
        }
        
        return true;
    }

	/**
	 * Invoked when the user cancels the content of the input.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param controller the current LyUserInputController instance
	 */
	public void cancel(LyController controller)
	{
	}

	/**
	 * Invoked when the user closes the content of the input.<br />
	 * <i>The default implementation</i> free the controller.
	 *
	 * @param controller the current LyUserInputController instance
	 */
	public void close(LyController controller)
	{
		LyController	parent = controller.getParent();
		
		if ((controller instanceof LyUserInputController)
				&& (parent instanceof LyTabAreaController)
				&& !((LyTabAreaController)parent).isWizard())
		{
			LyUserInputController	userInputController = (LyUserInputController)controller;
			LyTabAreaController		parentTabAreaControntroller = (LyTabAreaController)parent;
			
			parentTabAreaControntroller.closeAction(userInputController.getAction());
			
			if (!controller.isFreed())
				controller.free();
		}
		else
			controller.free();
	}
} // end class LyUserInputBehavior
