/*
 * LyWaitBehavior.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behavior;

import java.util.*;
import leon.app.behaviorinterface.*;
import leon.control.*;


/**
 * Basic implementation of the wait behavior.
 *
 * @author  Lyria S.A.
 * @version $Id: LyWaitBehavior.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @see     leon.control.LyWaitController
 * @see     leon.app.behaviorinterface.LyWaitBehaviorInterface
 */
public class LyWaitBehavior extends LyActionBehavior
	implements LyWaitBehaviorInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyWaitBehavior.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Minimum duration for the waiting view before closing (in milliseconds). */
	public static long		   MIN_DURATION = 4000;

	/** Indicates if the program has to wait for screen update when a value is modified. */
	public static boolean	   WAIT_FOR_DISPLAY = true;

	/** Current displayed value. */
	protected int					  _value;

	/**
	 * Offset added to the current value: to simulate progressive advance we increase both current
	 * value and max value (percent apparently increase).
	 */
	protected float					  _offsetValue;

	/** The wait icon shown in the view. */
	protected String				  _icon;

	/** The wait label shown in the view. */
	protected String				  _label;

	/** The date and time of the wait begining. */
	protected Date					  _startTime;

	/** Indicates if the program has to wait until the view is displayed. */
	protected boolean				  _waitingForDisplay;

	/**
	 * Default construtor.<br />
	 * Calls the super implementation and sets the {@link #_value current value} to 0, the {@link
	 * #_offsetValue offset value} too, sets the {@link #_waitingForDisplay} attribute to false and
	 * store the current date into {@link #_startTime}.
	 */
	public LyWaitBehavior()
	{
		super();

		_value = 0;
		_offsetValue = 0;
		_waitingForDisplay = false;
		_startTime = new Date();
	}

	/**
	 * Gets the current progress value. This method is invoked when the view is updated.<br />
	 * <i>The default implementation</i> stops the wait (sets {@link #_waitingForDisplay} to false)
	 * and computes the progress value depending on the controller {@link LyWaitController#getNMax()
	 * maximum} and {@link LyWaitController#getNMin() minimum} values, the {@link #_value current
	 * progress value} and the {@link #_offsetValue offset value}.
	 *
	 * @param  controller the current LyWaitController instance
	 * @return the current progress value
	 */
	public int getWaitValue(LyWaitController controller)
	{
		int l = controller.getNMax() - controller.getNMin();

		// Stop waiting
		_waitingForDisplay = false;
		return (int)(((_value + _offsetValue) * l) / (l + _offsetValue));
	}

	/**
	 * Gets the current label for the waiting view. This method is invoked when the view is updated.
	 * May update the label of the waiting view.<br />
	 * <i>The default implementation</i> does nothing and return the {@link #_label wait label}.
	 *
	 * @param  controller the current LyWaitController instance
	 * @return the label or null if the label should not change
	 */
	public String getWaitLabel(LyWaitController controller)
	{
		return _label;
	}

	/**
	 * Gets the current icon for the waiting view. This method is invoked when the view is updated.
	 * May update the icon of the waiting view.<br />
	 * <i>The default implementation</i> does nothing and return the {@link #_icon wait icon}.
	 *
	 * @param  controller the current LyWaitController instance
	 * @return the current icon or null if the current icon should not change
	 */
	public String getWaitIcon(LyWaitController controller)
	{
		return _icon;
	}

	/**
	 * Increases the current offset. This method is used to simulate smoth progression of the bar.
	 * <br />
	 * <i>The default implementation</i> increments the {@link #_offsetValue}.
	 *
	 * @param controller the current LyWaitController instance
	 */
	public void increaseOffset(LyWaitController controller)
	{
		_offsetValue++;
	}

	/**
	 * Method called when the waiting starts. May be overriden to do extra work.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param controller current LyWaitController instance
	 */
	public void startWaiting(LyWaitController controller)
	{
	}

	/**
	 * Method invoked at the end of the wait, when it is over. Should complete progress bar because
	 * the controller has ended the progression. May do extra work here, or close the view.<br />
	 * <i>The default implementation</i> does nothing.
	 *
	 * @param controller the current LyWaitController instance
	 */
	public void endWaiting(LyWaitController controller)
	{
		// TODO ---- delete this comment if the following code is modified -----
		// the javadoc of the default implementation associated to the following
		// commented code is :
		// /** <i>The default implementation</i> waits until the minimum display
		// * time is over (defined by {@link #MIN_DURATION}).
		// */
		// TODO ----------------------------------------------------------------

		// waitForDisplay(-1);

		// Window should remain visible at least MIN_DURATION time.
		/*
		 * Date d = new Date(); long duration =
		 * d.getTime()-_startTime.getTime();
		 *
		 * if (duration < MIN_DURATION) { try {
		 * Thread.sleep(MIN_DURATION-duration); } catch(InterruptedException e) { } }
		 */
	}

	/**
	 * Lets current thread wait for next repaint request.
	 *
	 * @param value the new progress value
	 */
	protected void waitForDisplay(int value)
	{
		// If needed wait to display current value before update
		int step = 0;
		int max = 10;

		if (value > 0)
			max = Math.min(value - _value, 10);

		while ((_waitingForDisplay) && (step < max))
		{
			try
			{
				Thread.sleep(100);
			}
			catch (InterruptedException e)
			{
			}
			step++;
		}
		// Start waiting (if WAIT_FOR_DISPLAY is true)
		_waitingForDisplay = WAIT_FOR_DISPLAY;
	}

	/**
	 * Sets the value of the waiting view progress.<br />
	 * <i>The default implementation</i> calls {@link #waitForDisplay(int) waitForDisplay(value)}
	 * and set the {@link #_value current progress value} with the given value.
	 *
	 * @param controller the current LyWaitController instance
	 * @param value      new progress value
	 */
	public void setWaitValue(LyWaitController controller, int value)
	{
		waitForDisplay(value);
		_value = value;
	}

	/**
	 * Sets the label of the waiting view.
	 *
	 * @param controller the current LyWaitController instance
	 * @param label      new label to display
	 */
	public void setWaitLabel(LyWaitController controller, String label)
	{
		_label = label;
	}

	/**
	 * Sets the icon if the waiting view.
	 *
	 * @param controller the current LyWaitController instance
	 * @param icon       new icon to display
	 */
	public void setWaitIcon(LyWaitController controller, String icon)
	{
		_icon = icon;
	}
} // end class LyWaitBehavior
