/*
 * LyActionBehaviorInterface.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behaviorinterface;

import leon.control.*;
import leon.data.*;
import leon.info.*;


/**
 * Generic interface for all behaviors that applies on action.
 *
 * @author  Lyria S.A.
 * @version $Id: LyActionBehaviorInterface.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @see     LyAreaController
 */
public interface LyActionBehaviorInterface extends LyBehaviorInterface
{
	/**
	 * File version
	 */
	String __VERSION = "$Id: LyActionBehaviorInterface.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Starts an action for the given controller. This method is the right place to allocate
	 * specific resources requiered for the given action.
	 *
	 * @param  areaController the current LyAreaController instance
	 * @param  action         the action that will be executed
	 * @return true if the action starts correctly, false otherwise.
	 */
	boolean startAction(LyAreaController areaController, LyAction action);

	/**
	 * The controller resets its content. This method is the right place to add extra processing
	 * when the view is set.
	 *
	 * @param  areaController the current LyAreaController instance
	 * @param  full           indicates if the controller performed a full reset of its content or a
	 *                        partial reset. The meaning of a full reset depends on controller
	 *                        implementation.
	 * @return true if the action starts correctly, false otherwise.
	 */
	boolean reset(LyAreaController areaController, boolean full);

	/**
	 * Ends an action for the given controller. This method is the right place to free specific
	 * resources allocated for the given action.
	 *
	 * @param  areaController the current LyAreaController instance
	 * @param  action         the action that was executed
	 * @return true if the action performed correctly, false otherwise
	 */
	boolean endAction(LyAreaController areaController, LyAction action);

	/**
	 * Gets the image corresponding to the given object for the given controller. This image is
	 * mostly used in titlebar.
	 *
	 * @param  areaController controller for which the image is required.
	 * @param  action         the action that was executed
	 * @return Logical name of the image for the given action or null to let controller do the job.
	 *         The string "_null" means no image.
	 */
	String getImage(LyAreaController areaController, LyAction action);

	/**
	 * Gets image tooltip corresponding to the given action.
	 *
	 * @param  areaController the current LyAreaController instance
	 * @param  action         the action that was executed
	 * @return image tooltip for the given action.
	 */
	String getImageTooltip(LyAreaController areaController, LyAction action);

	/**
	 * Gets the menu image corresponding to the given action for the given controller.
	 *
	 * @param  areaController The controller for which the image is required.
	 * @param  action         The action for which the image is required
	 * @return Logical name of the image for the given action or null to let controller do the job.
	 *         The string "_null" means no image.
	 */
	String getMenuItemImage(LyAreaController areaController, LyAction action);

	/**
	 * Gets title corresponding to the given action.
	 *
	 * @param  areaController the current LyAreaController instance
	 * @param  action         the action that was executed
	 * @return Title for the given action or null to let controller do the job.
	 */
	String getTitle(LyAreaController areaController, LyAction action);

	/**
	 * Gets title corresponding to the given action and context (sub-title displayed on the right of
	 * the titlebar).
	 *
	 * @param  areaController the current controller instance
	 * @param  action         the action that was executed
	 * @param  context        the action context
	 * @return Context title for the given action or null to let controller do the job.
	 */
	String getContextTitle(LyAreaController areaController, LyAction action,
		LyObjectList context);

	/**
	 * Gets help file for the given action.
	 *
	 * @param  areaController the current controller instance
	 * @param  action         the action that was executed
	 * @return Help file for the given action or null to let controller do the job.
	 */
	String getHelpFile(LyAreaController areaController, LyAction action);

	/**
	 * Gets context to set on the controllers that have the given area controller as context.
	 *
	 * @param  areaController the current controller instance.
	 * @param  context  the context that is about to be set.
	 * @return The desired client context.
	 */
	LyObjectList getClientContext(LyAreaController areaController, LyObjectList context);
	
    /**
     * Retrieves a value from a registered key.
     *
     * @param   key  The key that must be used.
     * @return  The value stored with the key or null if the key is not registered.
     */
    public Object getKeyValue(String key);
}