/*
 * LyApplicationBehaviorInterface.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behaviorinterface;

import java.util.*;
import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.infointerface.*;
import leon.peer.*;


/**
 * This interface defines methods to implement in order to create a specific behavior for an
 * application.
 *
 * @author  Lyria S.A.
 * @version $Id: LyApplicationBehaviorInterface.java 41407 2012-10-30 16:17:49Z bangiole $
 * @see     leon.app.LyApplication
 */
public interface LyApplicationBehaviorInterface
	extends LySessionListenerInterface
{
	/**
	 * File version
	 */
	String __VERSION =
		"$Id: LyApplicationBehaviorInterface.java 41407 2012-10-30 16:17:49Z bangiole $";

	/**
	 * Method called when an application is starting. This method is called <b>before</b> loading
	 * project and creating providers see {@link #started(LyApplication)} to be notified <b>
	 * after</b> generic staring process.
	 *
	 * @param  application Application that is starting.
	 * @see    #started(LyApplication)
	 * @return True if the application is correctly started, false otherwise.
	 */
	boolean start(LyApplication application);

	/**
	 * Method called when an application is started. This method is called <b>after</b> loading
	 * project and creating providers see {@link #start(LyApplication)} to be notified <b>before</b>
	 * generic staring process.
	 *
	 * @param  application Application that is started.
	 * @see    #start(LyApplication)
	 * @return True if the application is correctly started, false otherwise to stop starting
	 *         process.
	 */
	boolean started(LyApplication application);

    /**
     * Load application preferences.
     * @param application Application on which preferences are loaded.
     * @return List of preferences of the application.
     */
    public LyInfoList<LyInfo> loadPreferences(LyApplication application);

    /**
     * Load extra infos.
     * @param application Application on which extra infos (like charts or treemap configurations)
     * are loaded.
     * @return List of extra infos for the application.
     */
    public LyInfoList<LyInfo> loadExtraInfos(LyApplication application);

    /**
     * Load extra info like preferences, charts, treemap configurations or form saves.
     * @param application Application on which extra info is saved.
     * @param type Type of the extra info.
     * @param id Id of extra info.
     * @return  Info that was saved if found, null otherwise.
     */
    public Vector<LyExtraInfoInterface> loadExtraInfosFromDB(LyApplication application, String type,
        String userId);

    /**
     * Save given extra info like preferences, charts or treemap configurations.
     * @param application Application on which extra info is saved.
     * @param info        Info to be saved.
     * @return True if extra info was sucessfully saved.
     */
    public boolean saveExtraInfo(LyApplication application, LyExtraInfoInterface info);

    /**
     * Save given extra info like preferences, charts or treemap configurations.
     * @param application Application on which extra info is saved.
     * @param info        Info to be saved.
     * @param writer	  Writer interface.
     * @return True if extra info was sucessfully saved.
     */
    public boolean saveExtraInfo(LyApplication application, LyExtraInfoInterface info,
    	LyXmlWriterInterface writer);

    /**
     * Delete given extra info like preferences, charts or treemap configurations.
     * @param application Application on which extra info is deleted.
     * @param info        Info to be deleted.
     * @return True if extra info was sucessfully deleted.
     */
    public boolean deleteExtraInfo(LyApplication application, LyExtraInfoInterface info);

	/**
	 * Creates daemons to manage relations within an application.
	 *
	 * @param application Current application.
	 */
	void createDaemons(LyApplication application);

	/**
	 * Method called when a session is added to the application.
	 *
	 * @param application Application in which a session is added.
	 * @param session     Session added to the application.
	 */
	void sessionAdded(LyApplication application, LyDataSession session);

	/**
	 * Method called when a session is removed from the application.
	 *
	 * @param application Application in which a session is removed.
	 * @param session     Session removed from the application.
	 */
	void sessionRemoved(LyApplication application, LyDataSession session);

	/**
	 * Method called when datas are loaded for an application.
	 *
	 * @param application Application for which datas are loaded.
	 */
	void loadData(LyApplication application);

	/**
	 * Method called when datas are unloaded for an application.
	 *
	 * @param application Application for which datas are unloaded.
	 */
	void unloadData(LyApplication application);

	/**
	 * Indicates whether data may be loaded / unloaded for given class info.
	 *
	 * @param classInfo class info for which data may be loaded / unloaded
	 * @return true if data may be loaded / unloaded
	 */
	boolean enableLoadUnloadData(LyClassInfo classInfo);

	/**
	 * Method called when datas are saved for an application.
	 *
	 * @param application Application for which datas are saved.
	 */
	void saveData(LyApplication application);

	/**
	 * Method called when an application is closed.
	 *
	 * @param application Application that is closed.
	 */
	void close(LyApplication application);

	/**
	 * Method called when an application falls in sleeping mode.
	 *
	 * @param application Application for which datas are saved.
	 */
	void sleep(LyApplication application);

	/**
	 * Method called when an application is awake from sleeping mode.
	 *
	 * @param application Application that is awake.
	 */
	void wakeUp(LyApplication application);

	/**
	 * Gets all languages that the given application supports.
	 *
	 * @param  application Application that is awake.
	 * @return Array of couple of strings that represents supported languages. Each couple of string
	 *         is the code of the country for the language as defined in {@link java.util.Locale}
	 *         for the first string, and the display name of the language or a key to the
	 *         dictionary of this string for the second.
	 */
	String[][] getSupportedLanguages(LyApplication application);

	/**
	 * Get all timezone supported by this application.
	 *
	 * @param application the current application
	 * @return a list of supported timezones.
	 */
	String[] getSupportedTimeZones(LyApplication application);

	/**
	 * Adds specific encoders for new providers if needed. An encoder is used to encode a request
	 * made on a data source and decode results to give a response. For instance, the
	 * LySqlRequestEncoder encodes and decodes SQL queries to a RDBMS.
	 *
	 * @param application Current application.
	 * @param provider    provider which may need an encoder
	 */
	void addSpecificEncoders(LyApplication application, LyDataProvider provider);

	/**
	 * Method called when a connection with a server ({@link leon.info.location.LyLocation
	 * LyLocation} or {@link leon.info.notifier.LyNotifier LyNotifier}) is lost.
	 *
	 * @param application Current application.
	 * @param server      Logical server definition
	 */
	void connectionLost(LyApplication application, LyInfo server);

	/**
	 * Gets the control priority level of the application. This level is used to enable/disable
	 * controls.
	 *
	 * @param  application Current application.
	 * @param  object      the object on which the control is applied.
	 * @return the priority level of the application.
	 */
	int getPriorityLevel(LyApplication application, LyObject object);

	/**
	 * Method called when a transaction begins on a location.
	 *
	 * @param application   Current application
	 * @param idTransaction Id of the begin transaction
	 * @param locationId    Id of the location
	 */
	void begin(LyApplication application, String idTransaction, String locationId);

	/**
	 * Method called when a request is sent in a transaction for a location.
	 *
	 * @param application   Current application
	 * @param idTransaction Id of the transaction
	 * @param locationId    Id of the location
	 * @param request       Sent request if the request has a string representation
	 */
	void update(LyApplication application, String idTransaction, String locationId,
		String request);

	/**
	 * Method called before a transaction should be committed on a location.
	 *
	 * @param application   Current application
	 * @param idTransaction Id of the transaction
	 * @param locationId    Id of the location
	 */
	void beforeCommit(LyApplication application, String idTransaction, String locationId);

	/**
	 * Method called after a transaction has been committed on a location.
	 *
	 * @param application   Current application
	 * @param idTransaction Id of the transaction
	 * @param locationId    Id of the location
	 */
	void afterCommit(LyApplication application, String idTransaction, String locationId);

	/**
	 * Method called before a transaction should be rollbacked on a location.
	 *
	 * @param application   Current application
	 * @param idTransaction Id of the transaction
	 * @param locationId    Id of the location
	 */
	void beforeRollback(LyApplication application, String idTransaction, String locationId);

	/**
	 * Method called after a transaction has been rollbacked on a location.
	 *
	 * @param application   Current application
	 * @param idTransaction Id of the transaction
	 * @param locationId    Id of the location
	 */
	void afterRollback(LyApplication application, String idTransaction, String locationId);

	/**
	 * Method used by abstract classes to define which subclass has to be loaded.
	 *
	 * @param  application current application.
	 * @param  objectId    the object identifier to be loaded
	 * @param  values      optional additional values we might know about the object : this
	 *                     parameter is usually null.
	 * @return the list of classes if found or null to use generic processing. Note : an empty list
	 *         aborts generic processing (not found in any subclass)
	 */
	LyInfoList<LyClassInfo> getSubClasses(LyApplication application, LyObjectId objectId,
		LyValueSet values);

	/**
	 * Method used to find logical classes of a peer object identifier.<br />
	 * This implementation does nothing and return null.
	 *
	 * @param  application current application.
	 * @param  objectId    an object identifier of a peer object (use {@link
	 *                     leon.info.LyClassInfo#getObjectClass()} to get the peer class : instance
	 *                     of {@link leon.data.LyClass})
	 * @param  values      optional additional values we might know about the object : this
	 *                     parameter is usually null.
	 * @return the list of classes if found or null to use generic processing. Note : an empty lists
	 *         aborts generic processing (not found in any subclass)
	 */
	LyInfoList<LyClassInfo> getLogicalClasses(LyApplication application, LyObjectId objectId,
		LyValueSet values);

	/**
	 * Method used to get an unique short value for the specified application.
	 *
	 * @param  application current application
	 * @param  objectClass the class for which a new object is created.
	 * @return An unique value as short for the specified application.
	 */
	short getNextShort(LyApplication application, LyClass objectClass);

	/**
	 * Method used to get an unique int value for the specified application.
	 *
	 * @param  application current application
	 * @param  objectClass the class for which a new object is created.
	 * @return An unique value as int for the specified application.
	 */
	int getNextInt(LyApplication application, LyClass objectClass);

	/**
	 * Method used to get an unique long value for the specified application.
	 *
	 * @param  application current application
	 * @param  objectClass the class for which a new object is created.
	 * @return An unique value as long for the specified application.
	 */
	long getNextLong(LyApplication application, LyClass objectClass);

	/**
	 * Method used to get an unique float value for the specified application.
	 *
	 * @param  application current application
	 * @param  objectClass the class for which a new object is created.
	 * @return An unique value as float for the specified application.
	 */
	float getNextFloat(LyApplication application, LyClass objectClass);

	/**
	 * Method used to get an unique double value for the specified application.
	 *
	 * @param  application current application
	 * @param  objectClass the class for which a new object is created.
	 * @return An unique value as double for the specified application.
	 */
	double getNextDouble(LyApplication application, LyClass objectClass);

	/**
	 * Method called after a transaction was successfully completed (commit and notification done).
	 *
	 * @param application    the current application
	 * @param idTransactions List of involved transaction ids
	 * @param locationIds    list of involved locations ids
	 */
	void transactionCompleted(LyApplication application, Vector<String> idTransactions,
		Vector<String> locationIds);

	/**
	 * Free this behavior.<br />
	 * This method is provided to allow behavior to unregister itself from object sources or
	 * workspaces if needed.<br />
	 * It should be called only by the controller that built this behavior.
	 */
	void free();

	/**
	 * Load data for given classes.
	 *
	 * @param application Application for which classes are loaded.
	 * @param classes     List of classes to load.
	 * @see   LyApplication#loadData()
	 * @see   leon.info.LyClassInfo
	 */
	void loadData(LyApplication application, LyInfoList<LyClassInfo> classes);

	/**
     * Method for parsing extra information such as charts or preferences.
  	 * @param application Application for which classes are loaded.
  	 * @param prefType identifies the preference type : "preference", "chart" or "treemap_configuration"
     * @param userId  User id (may be null when info is shared).
     * @return List of loaded infos.
     */
    LyInfoList<LyInfo> parseExtraFromDB(LyApplication application, String prefType,  String userId);

    /**
     * Get the event manager for the given application.
     * @param application the current application
     * @return an event manager. If null is returned, default event manager will be instanciated.
     */
    LyEventManager getEventManager(LyApplication application);
}