/*
 * LyChartBehaviorInterface.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behaviorinterface;

import leon.app.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.view.*;


/**
 * This interface defines the methods to implement in order to modify the behavior of a chart
 * action.
 *
 * @see leon.control.LyChartController
 * @see leon.control.LyGraphicalChartController
 */
public interface LyChartBehaviorInterface extends LyFormBehaviorInterface
{
	/**
	 * File version
	 */
	String __VERSION = "$Id: LyChartBehaviorInterface.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Indicates whether a graphical chart view is resizable or not.
	 *
	 * @param  controller the current chart controller or graphical chart controller
	 * @return true if the view is resizable, false otherwise
	 */
	boolean isViewResizable(LyController controller);

	/**
	 * Gets fields that may be used as X variables in the chart.
	 *
	 * @param  controller The current chart controller.
	 * @return the fields that may be used as X variables.
	 */
	LyInfoList<LyFieldInfo> getXFields(LyChartController controller);

	/**
	 * Gets fields that may be used as Y variables in the chart.
	 *
	 * @param  controller The current chart controller.
	 * @return the fields that may be used as Y variables.
	 */
	LyInfoList<LyFieldInfo> getYFields(LyChartController controller);

	/**
	 * Gets fields that may be used as group variables in the chart.
	 *
	 * @param  controller The current chart controller.
	 * @return the fields that may be used as group variables.
	 */
	LyInfoList<LyFieldInfo> getGroupFields(LyChartController controller);

	/**
	 * Gets the chart context objects list.
	 *
	 * @param  controller    the controller
	 * @param  session       the current session
	 * @param  classInfo     the classInfo that contains values
	 * @param  objects       the selected objects of the current class
	 * @param  parentContext the context of the parent class
	 * @return the context for the chart
	 */
	LyObjectList getChartContext(LyController controller, LySession session,
		LyClassInfo classInfo, LyObjectList objects, LyWorkSpace parentContext);

	/**
	 * Computes a printable string (if the given typeVar is text, number or date) or a Vector of
	 * strings (if the given fieldInfo is a multiple choice or relation) from the given value. This
	 * (these) string(s) will be inserted in LyChart.
	 *
	 * @param  controller the current controller instance
	 * @param  fieldInfo  the field info concerned
	 * @param  object     the object concerned (may be null)
	 * @param  value      the value to transform into string. It comes from the fieldInfo
	 * @param  typeVar    the type of the field, like date or numeric (defined in {@link
	 *                    leon.info.LyChartInfo})
	 * @param  yAxis      indicates whether the value is computed for y or x axis
	 * @return a String corresponding to the value, or a Vector of Strings if it is a multiple
	 *         choice or relation.
	 * @see    LyChartInfo
	 */
	Object getStringChartValue(LyGraphicalChartController controller, LyFieldInfo fieldInfo,
		LyObject object, Object value, LyChartInfo.Type typeVar, boolean yAxis);

	/**
	 * Gets the label that must be shown in the legend of a bar chart.
	 *
	 * @param  controller the current controller instance
	 * @param  fieldInfo  the field info concerned
	 * @param  object     the object concerned (may be null)
	 * @param  value      the value to transform into string. It comes from the fieldInfo
	 * @return null if default label must be used, a string value otherwise
	 */
	String getBarLegendLabel(LyGraphicalChartController controller, LyFieldInfo fieldInfo,
		LyObject object, Object value);

	/**
	 * Gets the label that must be used for a point on a scatter plot chart.
	 *
	 * @param  controller the current controller instance
	 * @param  object     the object concerned
	 * @return the label that must be used for a given point
	 */
	String getScatterPlotLabel(LyGraphicalChartController controller, LyObject object);

	/**
	 * Returns the label that must be used for a point on a scatter plot legend.
	 *
	 * @param  controller the current controller instance
	 * @param  object     the object concerned
	 * @return the label that must be used for a given point
	 */
	String getScatterPlotLegendLabel(LyGraphicalChartController controller, LyObject object);

	/**
	 * Indicates if the legend must be positionned on the right of the chart.
	 *
	 * @param  controller the current controller instance
	 * @return true if the legend must be positionned on the right of the chart, false otherwise
	 */
	boolean isLegendRightPositionned(LyGraphicalChartController controller);

	/**
	 * Indicates whether empty values must be displayed on the legend.
	 *
	 * @param  controller the current controller instance
	 * @return true if empty values must be displayed on the legend, false otherwise
	 */
	boolean displayLegendEmptyValues(LyGraphicalChartController controller);

	/**
	 * Indicates whether sensitive areas must be generated or not.
	 *
	 * @param  controller the current controller instance
	 * @return true if sensitive areas must be generated on the chart, false otherwise
	 */
	boolean generateSensitiveAreas(LyGraphicalChartController controller);

	/**
	 * Complete chart component before showing it.
	 *
	 * @param controller the current controller instance
	 * @param chart      the chart component
	 */
	void completeChart(LyGraphicalChartController controller, LyChart chart);
}