/*
 * LyDataSessionBehaviorInterface.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behaviorinterface;

import java.util.*;
import leon.app.*;
import leon.data.*;
import leon.info.*;


/**
 * This interface defines the methods to implement in order to change the default behavior of a
 * session.
 *
 * @author  Lyria S.A.
 * @version $Id: LyDataSessionBehaviorInterface.java 33348 2011-11-30 16:18:01Z bangiole $
 * @see     leon.app.LyDataSession
 */
public interface LyDataSessionBehaviorInterface
{
	/**
	 * File version
	 */
	String __VERSION =
		"$Id: LyDataSessionBehaviorInterface.java 33348 2011-11-30 16:18:01Z bangiole $";

	/**
	 * Method called when a session is started.
	 *
	 * @param session the started session.
	 */
	void start(LyDataSession session);

    /**
     * Gets user of this session.
     * @param session Current session.
     * @return The object representing current user of the session.
     */
    public LyObject getUser(LyDataSession session);
    
	/**
	 * Method called when a session is closed.
	 *
	 * @param session the closed session.
	 */
	void close(LyDataSession session);

	/**
	 * Method called when an exception occured during the session proccessing.
	 *
	 * @param session the current session
	 * @param t       The throwable instance that stopped the processing.
	 */
	void handleException(LyDataSession session, Throwable t);
	
	/**
	 * Method called when a context needs to control objects before commit.<br />
	 * Result should be one of the following level:
	 *
	 * <ul>
	 *   <li>{@link leon.data.LyObject#LEVEL_ERROR}</li>
	 *   <li>{@link leon.data.LyObject#LEVEL_OK}</li>
	 *   <li>{@link leon.data.LyObject#LEVEL_WARNING}</li>
	 * </ul>
	 *
	 * @param  session the current session
	 * @param  objects the objects to control.
	 * @return Result of calling controls.
	 */
	short controlBeforeCommit(LyDataSession session, Collection<LyObject> objects);

	/**
	 * Free this behavior.<br />
	 * This method is provided to allow behavior to unregister itself from object sources or
	 * workspaces if needed.<br />
	 * It should be called only by the controller that built this behavior.
	 */
	void free();

	/**
	 * Method called when the undo/redo status has changed.<br />
	 * The status is available on the context manager of the application.
	 *
	 * @param session the current session.
	 * @see   leon.data.LyContextManager#isUndoAvailable()
	 * @see   leon.data.LyContextManager#isRedoAvailable()
	 */
	void updateUndoRedo(LyDataSession session);
	
	/**
	 * Method called when a connection with a server ({@link leon.info.location.LyLocation
	 * LyLocation} or {@link leon.info.notifier.LyNotifier LyNotifier}) is lost.
	 *
	 * @param session Current session.
	 * @param server      Logical server definition
	 */
	void connectionLost(LyDataSession session, LyInfo server);
	
    /**
     * Retrieves a value from a registered key.
     *
     * @param   key  The key that must be used.
     * @return  The value stored with the key or null if the key is not registered.
     */
    public Object getKeyValue(String key);

    /**
     * Get the reference unit for the given field for the current session
     * 
     * @param session Current session
     * @param fieldInfo a field
     * @return a reference unit. Can be null.
     */
    public LyUnit getDefaultUnitForField(LyDataSession session, LyFieldInfo fieldInfo);
}