/*
 * LySessionBehaviorInterface.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behaviorinterface;

import leon.app.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.view.event.*;


/**
 * This interface defines the methods to implement in order to change the default behavior of a
 * session.
 *
 * @author  Lyria S.A.
 * @version $Id: LySessionBehaviorInterface.java 32981 2011-10-28 15:34:35Z jlsaouli $
 * @see     leon.app.LySession
 */
public interface LySessionBehaviorInterface extends LyDataSessionBehaviorInterface
{
	/**
	 * File version
	 */
	String __VERSION = "$Id: LySessionBehaviorInterface.java 32981 2011-10-28 15:34:35Z jlsaouli $";

	/**
	 * Method called when a session is started.
	 *
	 * @param session the started session.
	 */
	void start(LySession session);

    /**
     * Load session preferences.
     * @param session Session on which preferences are loaded.
     * @param userId  Id of the session user.
     * @return List of preferences of the session.
     */
    public LyInfoList<LyInfo> loadUserPreferences(LySession session, String userId);

    /**
     * Load session stickers.
     * @param session Session on which stickers are loaded.
     * @param userId  Id of the session user.
     * @return List of stickers of the session.
     */
    public LyInfoList<LyInfo> loadUserStickers(LySession session, String userId);
    
	/**
	 * Method called when a session is closed.
	 *
	 * @param session the closed session.
	 */
	void close(LySession session);

	/**
	 * Gets root action for this session.
	 *
	 * @param session current session.
	 * @return the root action for the session
	 */
	LyAction getRootAction(LySession session);

	/**
	 * Method called when the user ask to quit the given session.
	 *
	 * @param session Session to exit from
	 */
	void exit(LySession session);

    /**
     * Method to compute id of the given controller.
     * @param controller The new controller.
     * @param id Desired id.
     * @return Id of the new controller.
     */
    public String checkControllerId(LyController controller, String id);
    
	/**
	 * Method called when a controller is added to the session.
	 *
	 * @param session    Session in which a controller is added.
	 * @param controller Controller added to the session.
	 */
	void controllerAdded(LySession session, LyController controller);

	/**
	 * Method called when a controller is removed from the session.
	 *
	 * @param session    Session in which a controller is removed.
	 * @param controller Controller removed from the session.
	 */
	void controllerRemoved(LySession session, LyController controller);

	/**
	 * Method called when a view controller is shown within the given session.
	 *
	 * @param session        the session
	 * @param viewController The view controller that is shown.
	 */
	void viewControllerShown(LySession session, LyViewController viewController);

	/**
	 * Method called when a view controller is hidden within the given session.
	 *
	 * @param session        the session
	 * @param viewController The view controller that is hidden.
	 */
	void viewControllerHidden(LySession session, LyViewController viewController);

	/**
	 * Validates the given action for the given controller.
	 *
	 * @param  session    the current session
	 * @param  action     the action to validate
	 * @param  controller the controller for this action
	 * @return true if the action is valid, false otherwise.
	 */
	boolean validateAction(LySession session, LyAction action, LyController controller);

	/**
	 * Enables the given action on the given list of objects for the given controller.
	 *
	 * @param  session    the current session
	 * @param  action     the action to enable
	 * @param  controller the controller for this action
	 * @param  objects    list of objects on which the action has to be enabled
	 * @param  showError  report an error if the action is not enabled
	 * @return true if the action is enabled, false otherwise.
	 */
	boolean enableAction(LySession session, LyAction action, LyController controller,
		LyObjectList objects, boolean showError);

	/**
	 * Enables the given transverse action on the given list of objects for the given controller.
	 *
	 * @param  session    the current session
	 * @param  action     the action to enable
	 * @param  classInfo  the classInfo for which the action is checked
	 * @param  controller the controller for this action
	 * @param  objects    list of objects on which the action has to be enabled
	 * @param  showError  report an error if the action is not enabled
	 * @return true if the action is enabled, false otherwise.
	 */
	boolean enableTransverseAction(LySession session, LyAction action,
		LyClassInfo classInfo, LyController controller, LyObjectList objects, boolean showError);

	/**
	 * Invoke an action that has no controller.<br />
	 * This function is the right place to implement an application specific action because it is
	 * not driven by a controller.
	 *
	 * @param  session   the current session
	 * @param  order     the source controller
	 * @param  action    the action to execute
	 * @param  classInfo class on which the action applies
	 * @param  objects   list of objects on which the action applies
	 * @param  show      indicates if the action must be shown or not
	 * @return true if the action was handled by this behavior, false otherwise.
	 */
	boolean invokeExtraAction(LySession session, LyController order, LyAction action,
		LyClassInfo classInfo, LyObjectList objects, boolean show);

	/**
	 * Invoke an action.<br />
	 * This function is the right place to implement an application specific action.
	 *
	 * @param  session   the current session
	 * @param  order     the source controller
	 * @param  action    the action to execute
	 * @param  classInfo class on which the action applies
	 * @param  objects   list of objects on which the action applies
	 * @param  show      indicates if the action must be shown or not
	 * @return the controller corresponding to the invoked action.
	 */
	LyController invokeAction(LySession session, LyController order, LyAction action,
		LyClassInfo classInfo, LyObjectList objects, boolean show);

	/**
	 * Gets the session time out for the current session.
	 *
	 * @param  session the current session
	 * @return the session timeout in seconds
	 */
	int getSessionTimeout(LySession session);

	/**
	 * Method invoked when a session is in time out.
	 *
	 * @param session the current session
	 */
	void sessionTimeout(LySession session);

	/**
	 * Get a root level context controller. This controller when created defines a top level session
	 * context : GUI modification Like set controller should become subcontext of this context.
	 * Usually this method returns null (no controller is a top level context controller). The
	 * ImportTableController is defining a root level context to grab all user modifications in the
	 * import context while the import table is active.
	 *
	 * @param session the session
	 * @return the root level context controller if any, or null if it could not be found.
	 */
	LyController getRootContextController(LySession session);

	/**
	 * Method invoked when the user ask for the 'about' action. This method must give the 'about'
	 * message to the user.
	 *
	 * @param session the current session
	 * @param parent  the parent controller
	 */
	void showAbout(LySession session, LyController parent);

	/**
	 * Method invoked when the user ask for the 'log' action. This method must open the log
	 * file for the user.
	 *
	 * @param session the current session
	 * @param controler  the parent controller
	 */
	void showLog(LySession session, LyController controler);
	
	/**
	 * Method invoked when the user ask for the 'version' action. This method must give the
	 * 'version' message to the user.
	 *
	 * @param session the current session
	 * @param parent  the parent controller
	 */
	void showVersion(LySession session, LyController parent);

	/**
	 * Method invoked when the user ask for the 'help' action. This method must show the
	 * corresponding help to the user.
	 *
	 * @param session the current session
	 * @param parent  the parent controller
	 */
	void showHelp(LySession session, LyController parent);

	/**
	 * Invoked when a demand is sent to a listener of this session.
	 *
	 * @param  session    the current session
	 * @param  controller controller that routed the demand.
	 * @param  demand     demand parameters sent by the view component.
	 * @return Return true if the demand was handled, false otherwise. If false, the demand is
	 *         transmitted to the controller to try to handle it.
	 */
	boolean handleDemand(LySession session, LyController controller, LyDemand demand);

	/**
	 * Searches for an existing controller being shown, and return it if found.
	 *
	 * @param  session            the current session
	 * @param  parent             The source controller.
	 * @param  action             The action to execute.
	 * @param  classInfo          The class on which the action applies.
	 * @param  objects            The list of objects on which the action applies.
	 * @param  recurseAllSessions indicates if we recurse deeped into all sessions
	 * @return the controller if found, null otherwise
	 * @see    leon.app.behavior.LySessionBehavior#invokeAction
	 */
	LyController findExistingController(LySession session, LyController parent,
		LyAction action, LyClassInfo classInfo, LyObjectList objects, boolean recurseAllSessions);

	/**
	 * Get an help file for the given session, the given controller and the given action.
	 *
	 * @param  session    the current session.
	 * @param  controller the controller of the action.
	 * @param  action     the action for which the helpfile is required.
	 * @return the path to the help file, or null if we have to use the generic help file behavior.
	 */
	String getHelpFile(LySession session, LyController controller, LyAction action);
}