/*
 * LySetBehaviorInterface.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behaviorinterface;

import java.util.*;

import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.info.save.*;


/**
 * This interface defines the methods to implement in order to modify the behavior of a modification
 * action. A modification action includes create, set and clone.
 *
 * @author  Lyria S.A.
 * @version $Id: LySetBehaviorInterface.java 32982 2011-10-28 15:52:30Z jlsaouli $
 * @see     leon.control.LySetController
 */
public interface LySetBehaviorInterface extends LyConsultBehaviorInterface
{
	/**
	 * File version
	 */
	String __VERSION = "$Id: LySetBehaviorInterface.java 32982 2011-10-28 15:52:30Z jlsaouli $";

	/**
	 * Code returned when there is no modification on the value for controlNewValue, checkNewValue
	 * and controlNewValues.<br />
	 * All codes are :
	 *
	 * <ul>
	 *   <li>{@link #STATUS_CANCEL}: the operation is cancelled</li>
	 *   <li>{@link #STATUS_IGNORE}: the value is not modified</li>
	 *   <li>{@link #STATUS_OK}: the value is correct</li>
	 *   <li>{@link #STATUS_KO}: the value is not correct</li>
	 *   <li>{@link #STATUS_WARNING}: the value is on warning</li>
	 * </ul>
	 */
	short  STATUS_CANCEL = -2;

	/**
	 * Code returned when there is no modification on the value for controlNewValue, checkNewValue
	 * and controlNewValues.<br />
	 *
	 * @see #STATUS_CANCEL
	 */
	short  STATUS_IGNORE = -1;

	/**
	 * Code returned when the value is correct for controlNewValue, checkNewValue and
	 * controlNewValues.<br />
	 *
	 * @see #STATUS_CANCEL
	 */
	short  STATUS_OK = 0;

	/**
	 * Code returned when the value is not correct for controlNewValue, checkNewValue and
	 * controlNewValues.<br />
	 *
	 * @see #STATUS_CANCEL
	 */
	short  STATUS_KO = 1;

	/**
	 * Code returned when the value is on warning for controlNewValue, checkNewValue and
	 * controlNewValues.<br />
	 *
	 * @see #STATUS_CANCEL
	 */
	short  STATUS_WARNING = 2;

	/**
	 * Checks if the given fieldInfo is optional or not for the given controller.
	 *
	 * @param  set       the current LySetController instance
	 * @param  fieldInfo the target fieldInfo to test
	 * @return true if the given fieldInfo is optional, false otherwise
	 */
	boolean isOptional(LySetController set, LyFieldInfo fieldInfo);

	/**
	 * Checks if the given fieldInfo is editable or not for the given controller.
	 *
	 * @param  set       the current LySetController instance
	 * @param  fieldInfo the target fieldInfo to test
	 * @return true if the given fieldInfo is editable, false otherwise
	 */
	boolean isEditable(LySetController set, LyFieldInfo fieldInfo);

	/**
	 * Checks if the given fieldInfo is enabled or not for the given controller.
	 *
	 * @param  set       the current LySetController instance
	 * @param  fieldInfo the target fieldInfo to test
	 * @return true if the given fieldInfo is enabled, false otherwise
	 */
	boolean isEnabled(LySetController set, LyFieldInfo fieldInfo);

	/**
	 * Indicates the maximum number of values displayed in text components for auto-completion.
	 *
	 * @param  set       the current LySetController instance.
	 * @param  fieldInfo the target text fieldInfo to test.
	 * @return a positive number if auto-completion is activated, 0 if auto-completion is not
	 *         activated, a negative value if there is no maximum.
	 */
	int getAutoCompletionMaximum(LySetController set, LyTextFieldInfo fieldInfo);

	/**
	 * Computes the list of values displayed in text components for auto-completion.
	 *
	 * @param  set       the current LySetController instance.
	 * @param  fieldInfo the target text fieldInfo to test.
	 * @return the list of values for auto-completion.
	 */
	String[] getAutoCompletionValues(LySetController set, LyTextFieldInfo fieldInfo);

	/**
	 * Gets domain (list of allowed options) of the given choice fieldInfo.
	 *
	 * @param  set    the current LySetController instance
	 * @param  choice target choice fieldInfo
	 * @return the list of {@link leon.info.LyChoiceOption} allowed for this choice field info
	 */
	Vector<LyChoiceOption> getChoiceDomain(LySetController set, LyChoiceFieldInfo choice);

	/**
	 * Checks if the given options group is valid for the given option of the choice fieldinfo.
	 *
	 * @param  set    the current set controller
	 * @param  choice the choice fieldInfo
	 * @param  option the choiceOption
	 * @param  group  the choiceOption group
	 * @return true if the group is valid, false otherwise.
	 */
	boolean isValidGroup(LySetController set, LyChoiceFieldInfo choice,
		LyChoiceOption option, LyChoiceOptionGroup group);

	/**
	 * Indicates if the given option is enabled on the given choice fieldInfo.
	 *
	 * @param  set    the current LySetController instance
	 * @param  choice choiceFieldInfo that contains this option
	 * @param  option the option to enabled
	 * @return true if the option is enabled, false otherwise
	 */
	boolean enableChoiceOption(LySetController set, LyChoiceFieldInfo choice,
		LyChoiceOption option);

    /**
     * Indicates if the given option is hidden on the given choice fieldInfo.<br />
     * <i>The default implementation</i> checks if the option is not disabled.
     *
     * @param  set    the current LySetController instance
     * @param  choice choiceFieldInfo that contains this option
     * @param  option the option to check
     * @return true if the option is hidden, false otherwise
     */
    boolean validateChoiceOption(LySetController set, LyChoiceFieldInfo choice,
        LyChoiceOption option);
    
	/**
	 * Gets relation domain (list of available links) for the given relation fieldInfo.
	 *
	 * @param  set      the current LySetController instance
	 * @param  relation the target relation field
	 * @return a list of LyObjects available for this relation
	 */
	LyObjectList getRelationDomain(LySetController set, LyRelationFieldInfo relation);

	/**
	 * Allows the given link object as a possible value for the given relation field info.
	 *
	 * @param  set       the current LySetController instance
	 * @param  relation  the relation that contains the link
	 * @param  link      the link to enabled
	 * @param  showError indicates if an error should be shown if the given link is rejected.
	 * @return true if the link is enabled on this relation, false otherwise
	 */
	boolean enableRelationLink(LySetController set, LyRelationFieldInfo relation,
		LyObject link, boolean showError);

	/**
	 * Allows the given link object as a possible value for the given relation field info.
	 *
	 * @param  set      the current LySetController instance
	 * @param  relation the relation that contains the link
	 * @param  link     the link to enabled
	 * @return true if the link is enabled on this relation, false otherwise
	 */
	boolean enableRelationLink(LySetController set, LyRelationFieldInfo relation,
		LyObject link);

	/**
	 * Gets the list of available classInfo of the given relation fieldInfo.
	 *
	 * @param  set      the current LySetController instance
	 * @param  relation the target relation fieldInfo
	 * @return the list of classInfo available for this relation
	 */
	LyInfoList<LyClassInfo> getRelationClasses(LySetController set, LyRelationFieldInfo relation);

	/**
	 * Getsthe list of objects of the given relation context.
	 *
	 * @param  set      the current LySetController instance
	 * @param  relation the target relation fieldInfo
	 * @return a list of LyObject available in this relation field
	 */
	LyObjectList getRelationContext(LySetController set, LyRelationFieldInfo relation);

	/**
	 * Indicates the maximum number of values displayed in relation components for auto-completion.
	 * 
	 * @param  set       the current LySetController instance.
	 * @param  relation  the target relation fieldInfo to test.
	 * @return a positive number if auto-completion is activated, 0 if auto-completion is not
	 *         activated, a negative value if there is no maximum.
	 */
	int getAutoCompletionMaximum(LySetController set, LyRelationFieldInfo relation);
	
	/**
	 * Gets relation domain (list of available links) for the given relation fieldInfo with the
	 * given prefix for related objects.
	 *
	 * @param  set      the current LySetController instance
	 * @param  relation the target relation field
	 * @param  prefix   the prefix for related objects
	 * @param  fill     indicates whether the domain must be filtered with given prefix or not
	 * @return a list of LyObjects available for this relation and given prefix
	 */
	LyObjectList getRelationDomain(LySetController set, LyRelationFieldInfo relation, String prefix, boolean fill);
	
	/**
	 * Gets the short label that must be used for the given relation fieldInfo.
	 * 
	 * @param set      the current LySetController instance
	 * @param relation the target relation field
	 * @param link     the object represented by the link
	 * @return the short name of the link (displayed when selected) or null if no short label must be used
	 */
	String getLinkShortLabel(LySetController set, LyRelationFieldInfo relation, LyObject link);
	
	/**
	 * Indicates if auto-completion on the given relation fieldInfo is managed by the graphical
	 * component or by the field controller.
	 *
	 * @param  set       the current LySetController instance
	 * @param  relation  the fieldInfo that may need to be checked
	 * @return true if auto-completion is managed by the graphical component, false otherwise
	 */
	boolean useLocalCompletion(LySetController set, LyRelationFieldInfo relation);
	
	/**
	 * Gets actions enabled by the given relation. A list of pairs (action, classInfo) has to be
	 * returned.
	 *
	 * @param  set      the current LySetController instance
	 * @param  relation the relation that may enabled some actions
	 * @return the list of actions enabled by the given relation
	 */
	LyInfoDataList<LyAction, LyClassInfo> getTransverseActions(LySetController set, LyRelationFieldInfo relation);

	/**
	 * Gets actions enabled by the given field. A list of pairs (action, classInfo) has to be
	 * returned.
	 *
	 * @param  set       the current LySetController instance
	 * @param  fieldInfo the field that may enabled some actions
	 * @return a list of actions enabled by the given field
	 */
	LyInfoDataList<LyAction, LyClassInfo> getTransverseActions(LySetController set, LyFieldInfo fieldInfo);

	/**
	 * Gets the default value of the given fieldInfo. This method can be called:
	 *
	 * <ul>
	 *   <li>to initialize an input field in a creation form</li>
	 *   <li>to get a value for an empty input field that becomes mandatory</li>
	 *   <li>to get a value for notNull fields at the end of the creation process</li>
	 * </ul>
	 *
	 * @param  set       the current LySetController instance.
	 * @param  fieldInfo the fieldInfo that contains the value.
	 * @return the value of the given fieldInfo as an Object
	 */
	Object getDefaultValue(LySetController set, LyFieldInfo fieldInfo);

	/**
	 * Indicates if the given fieldInfo needs to be checked when the value is modified (in creation
	 * and set mode).
	 *
	 * @param  set       the current LySetController instance
	 * @param  fieldInfo the fieldInfo that may need to be checked
	 * @return true if the field need to be checked, false otherwise
	 */
	boolean needCheck(LySetController set, LyFieldInfo fieldInfo);

	/**
	 * Progates modifications of the given field on all other fields. Propagate modidication of a
	 * fieldInfo with the given value on other fields.
	 *
	 * @param set      the current LySetController instance
	 * @param newValue new field info value to propagate
	 */
	void propagate(LySetController set, LyFieldInfoValue newValue);

	/**
	 * This method controls the given value before creating or modifying an object. It does not
	 * checks its content but the coherence between it and its field marks.
	 *
	 * @param  set      the current LySetController instance
	 * @param  newValue field info value to check. This value cannot be null.
	 * @return a code to indicates the control result. Possible values are:
	 *
	 *         <ul>
	 *           <LI>{@link #STATUS_IGNORE} if the value has not change</LI>
	 *           <LI>{@link #STATUS_OK} if the value is correct</LI>
	 *           <LI>{@link #STATUS_KO} if the value is not correct</LI>
	 *           <LI>{@link #STATUS_WARNING} if the value is on warning</LI>
	 *         </ul>
	 */
	short controlNewValue(LySetController set, LyFieldInfoValue newValue);

	/**
	 * Dynamicaly checks the given value for its associated field. This method is called when there
	 * is a modification of the field in the user interface (Value changed, focus lost...).
	 *
	 * @param  set      the current LySetController instance
	 * @param  newValue field info value to check.
	 * @return a code to indicates the check result. Possible values are:
	 *
	 *         <ul>
	 *           <LI>STATUS_KO if the value is not correct</LI>
	 *           <LI>STATUS_OK if the value is correct</LI>
	 *         </ul>
	 */
	short checkNewValue(LySetController set, LyFieldInfoValue newValue);

	/**
	 * Global control of values before creating or modifying an object. If needed, extra values may
	 * be added during this control.
	 *
	 * @param  set       the current LySetController instance
	 * @param  newValues Set of values to control
	 * @return
	 *
	 *         <ul>
	 *           <LI>STATUS_IGNORE if the value has not change</LI>
	 *           <LI>STATUS_KO if the value is not correct</LI>
	 *           <LI>STATUS_OK if the value is correct</LI>
	 *         </ul>
	 */
	short controlNewValues(LySetController set, LyValueSet newValues);

	/**
	 * This method adds the given extra values before completing the current operation.
	 *
	 * @param set    the current LySetController instance
	 * @param values Current set of values. Additionnal values must be add to this set
	 */
	void addExtraValues(LySetController set, LyValueSet values);

	/**
	 * Final step of object set process.
	 *
	 * @param  set    the current LySetController instance
	 * @param  values Set of values for the modified object
	 * @return true if the modification process is done, false otherwise
	 */
	boolean endSet(LySetController set, LyValueSet values);

	/**
	 * Gets the filter of the given relation and the given class.<br />
	 * This method is used by the {@link #getRelationDomain(LySetController, LyRelationFieldInfo)}
	 * method when the target objets of the relation need to be filtered.
	 *
	 * @param  controller the current LyController instance
	 * @param  relation   the relation field info we want to filter
	 * @param  classInfo  ClassInfo for which we look for an associated filter.
	 * @return The filter associated to the given classInfo, or null if no filter has been specified
	 *         for this classInfo in this relation.
	 */
	LyFilter getRelationFilter(LySetController controller, LyRelationFieldInfo relation,
		LyClassInfo classInfo);

	/**
	 * Checks if an edit table should be used for the given array field.
	 *
	 * @param  controller The current LySetController instance.
	 * @param  array      The array field info.
	 * @return True if an edit table is requested, false otherwise.
	 */
	boolean useEditTable(LySetController controller, LyArrayFieldInfo array);

	/**
	 * Checks if an edit table should be used for the given relation field.
	 *
	 * @param  controller The current LySetController instance.
	 * @param  relation   The relation field info.
	 * @return True if an edit table is requested, false otherwise.
	 */
	boolean useEditTable(LySetController controller, LyRelationFieldInfo relation);
	
	/**
	 * Invoked when controller is starting to commit, default implementation does nothing
	 * @param  controller The current LySetController instance.
	 */
	void startCommit(LySetController controller);
	/**
	 * Invoked when controller is ending commit, default implementation does nothing
	 * @param  controller The current LySetController instance.
	 */
	void endCommit(LySetController controller);
	
	/**
     * Indicates if the operations of the controller context must be inserted at the beginning of the parent
     * operations or at the end.
     * 
     * @param controller the set controller
     * @param relation the relation field
     * @return Position where operations must be inserted, may be null if left to controller decision.
     */
    LyContext.Position getInsertPosition(LySetController controller, LyRelationFieldInfo relation);

    /**
     * Method called on form save.
     * @param  controller The current LySetController instance.
     * @param  formSave The formSave that was created.
     */
    void formSaved(LySetController controller, LyFormSave formSave);
    
    /**
     * Method called on form restore.
     * @param  controller The current LySetController instance.
     * @param  formSave The formSave that was restored.
     */
    void formRestored(LySetController controller, LyFormSave formSave);
    
    /**
     * Indicates if the given controller must link its context to its parent context
     * @param  controller The current LySetController instance.
     * @param parentRelation the parent relation for this controller (may be null)
     */
    boolean useSingleContext(LySetController controller, LyRelationFieldInfo parentRelation);

    /**
     * Checks if given object is locked for edition.
     * @param controller The current LySetController instance.
     * @param object Object to be checked.
     * @return Null if not locked, otherwise a string message giving information about the lock (user, date...).
     */
    String isLocked(LySetController controller, LyObject object);
    
    /**
     * Locks/Unlocks given object for edition.
     * @param controller The current LySetController instance.
     * @param object Object to be locked/unlocked.
     * @param lock True if lock is required, false for unlock.
     * @return True if successful, false otherwise.
     */
    boolean lockObject(LySetController controller, LyObject object, boolean lock);
}