/*
 * LyTabActionBehaviorInterface.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behaviorinterface;

import leon.control.*;
import leon.info.*;


/**
 * Interface for the controller of tab actions.
 *
 * @see LyTabAreaController
 */
public interface LyTabActionBehaviorInterface extends LyActionBehaviorInterface
{
	/**
	 * File version
	 */
	String __VERSION = "$Id: LyTabActionBehaviorInterface.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Get labels to display in left panel of the wizard. This method is only used for tab actions
	 * with wizard type.
	 *
	 * @param  tabAreaController Controller of the tab action.
	 * @return A table of strings corresponding to labels, or null if no label is wanted.
	 */
	String[] getStepLabels(LyTabAreaController tabAreaController);

	/**
	 * Get image to display in left panel of the wizard. This method is only used for tab actions
	 * with wizard type.
	 *
	 * @param  tabAreaController Controller of the tab action.
	 * @param  index             Index of the current step.
	 * @return A string that gives the name of the image to display.
	 */
	String getImage(LyTabAreaController tabAreaController, int index);

	/**
	 * Get the colors used for hyperlinks displayed in left panel of the wizard. This method is only
	 * used for tab actions with wizard type.
	 *
	 * @param  tabAreaController Controller of the tab action.
	 * @return An array of strings that gives the colors for hyperlinks, highlighted hyperlinks,
	 *         visited hyperlinks and disabled hyperlinks.
	 */
	String[] getHrefColors(LyTabAreaController tabAreaController);

	/**
	 * Get next action to be displayed in right panel of the wizard. This method is only used for
	 * tab actions with wizard type and is called when next button is selected the first time the
	 * step is reached.
	 *
	 * @param  tabAreaController Controller of the tab action.
	 * @param  index             Index of the current step.
	 * @return The next action to be displayed.
	 */
	LyAction getNextAction(LyTabAreaController tabAreaController, int index);

	/**
	 * Builds area controller corresponding to the given action. This method is only used for tab
	 * actions with wizard type.
	 *
	 * @param  tabAreaController Controller of the tab action.
	 * @param  action            Action for which a controller is requested.
	 * @return The desired area controller, if null is returned the tabAreaController will construct
	 *         it.
	 */
	LyAreaController buildAreaController(LyTabAreaController tabAreaController, LyAction action);

	/**
	 * This method is only used for tab actions with wizard type.
	 *
	 * @param  tabAreaController Controller of the tab action.
	 * @param  index             Index of the current step.
	 * @return A boolean indicating if the Next button is enabled or not.
	 */
	boolean enableNext(LyTabAreaController tabAreaController, int index);

	/**
	 * This method is only used for tab actions with wizard type.
	 *
	 * @param  tabAreaController Controller of the tab action.
	 * @param  index             Index of the current step.
	 * @return A boolean indicating if the Terminate button is enabled or not.
	 */
	boolean enableTerminate(LyTabAreaController tabAreaController, int index);

	/**
	 * Method called when Terminate is selected before committing actions done in wizard. This
	 * method is only used for tab actions with wizard type.
	 *
	 * @param  tabAreaController Controller of the tab action.
	 * @return A boolean indicating if the commit can be done.
	 */
	boolean beforeCommit(LyTabAreaController tabAreaController);

	/**
	 * Method called when Terminate is selected after committing actions done in wizard. This method
	 * is only used for tab actions with wizard type.
	 *
	 * @param tabAreaController Controller of the tab action.
	 */
	void afterCommit(LyTabAreaController tabAreaController);

	/**
	 * Method called when Cancel is selected before cancelling actions done in wizard. This method
	 * is only used for tab actions with wizard type.
	 *
	 * @param tabAreaController Controller of the tab action.
	 */
	void cancel(LyTabAreaController tabAreaController);

	/**
	 * Method called when a tab has been selected.
	 *
	 * @param tabAreaController      the tab action controller
	 * @param selectedTab            the index of the selected tab
	 * @param selectedAreaController the controller of the selected tab action
	 */
	void selectedTabChanged(LyTabAreaController tabAreaController, int selectedTab,
		LyAreaController selectedAreaController);
}