/*
 * LyTableBehaviorInterface.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behaviorinterface;

import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.view.*;


/**
 * This interface defines methods to implement in order to modify the default behavior of a table
 * controller.
 *
 * @author  Lyria S.A.
 * @version $Id: LyTableBehaviorInterface.java 36619 2012-08-28 14:27:47Z jlsaouli $
 * @see     leon.control.LyTableController
 */
public interface LyTableBehaviorInterface extends LyPanelBehaviorInterface
{
	/**
	 * File version
	 */
	String __VERSION = "$Id: LyTableBehaviorInterface.java 36619 2012-08-28 14:27:47Z jlsaouli $";

	/**
	 * Gets the list of fields to display in the table from the given list.
	 *
	 * @param  tableController the table controller that requires the list
	 * @param  fields          a list of fields from where fields to display are taken
	 * @return The list of fields to display in the table.
	 */
	LyInfoList<LyFieldInfo> getFields(LySimpleTableController tableController,
		LyInfoList<LyFieldInfo> fields);

	/**
	 * Gets the status image of the given table object.
	 *
	 * @param  tableController Table controller that requires the status image.
	 * @param  statusFields    List of field infos that have the status mark.
	 * @param  object          Object to which the status image is associated.
	 * @return Path to the image file or the name of an image resource if found, null otherwise.
	 */
	String getStatusImage(LySimpleTableController tableController,
		LyInfoList<LyFieldInfo> statusFields, LyObject object);

	/**
	 * Gets status image tooltip for the given object.<br />
	 * This method is called only if the {@link #getStatusImage(LySimpleTableController, LyInfoList,
	 * LyObject)} method didn't return null.
	 *
	 * @param  tableController Table controller that requires the status image.
	 * @param  statusFields    List of field infos that have the status mark.
	 * @param  object          Object to which the status image is associated.
	 * @return String containing the tooltip associated to the status image.
	 */
	String getStatusTooltip(LySimpleTableController tableController,
		LyInfoList<LyFieldInfo> statusFields, LyObject object);

	/**
	 * Gets row color associated to the given object.
	 *
	 * @param  tableController Table controller that requires the row color.
	 * @param  object          Object for which a row color is required.
	 * @return Color name or the name of a color resource if any, null otherwise.
	 */
	String getRowColor(LySimpleTableController tableController, LyObject object);

	/**
	 * Gets the row image of the given table object used on display list mode.
	 *
	 * @param  tableController Table controller that requires the row image.
	 * @param  object          Object to which the row image is associated.
	 * @return Path to the image file or the name of an image resource if found, null otherwise.
	 */
	String getRowImage(LySimpleTableController tableController,	LyObject object);
	
	/**
	 * Gets the header label representing the given field info.
	 *
	 * @param  controller Table controller that requires the header label.
	 * @param  fieldInfo  Field info for which a header label is required.
	 * @return Label that will be put in the table header for the given field info.
	 */
	String getHeaderLabel(LySimpleTableController controller, LyFieldInfo fieldInfo);

	/**
	 * Gets the header tooltip associated to the given field info.
	 *
	 * @param  controller Table controller that requires the header tooltip.
	 * @param  fieldInfo  Field info for which a header tooltip is required.
	 * @return Label that will be put in the table header tooltip corresponding to the given field
	 *         info.
	 */
	String getHeaderTooltip(LySimpleTableController controller, LyFieldInfo fieldInfo);

	/**
	 * Gets the cell label representing the value of the given field info for the given object.
	 *
	 * @param      tableController Table controller that requires the cell label.
	 * @param      object          Object from which the field info value is represented in the
	 *                             table cell.
	 * @param      fieldInfo       Field info for which a cell label is required.
	 * @return     Label that will be put in the table cell for the given object and field info (may
	 *             be null).
	 * @deprecated Use {@link LyTableBehaviorInterface#updateCell(LySimpleTableController, LyCell,
	 *             LyObject, LyFieldInfo)} instead.
	 * @see        LyTableBehaviorInterface#updateCell(LySimpleTableController,LyCell,LyObject,LyFieldInfo)
	 */
	String getCellLabel(LySimpleTableController tableController, LyObject object,
		LyFieldInfo fieldInfo);

	/**
	 * Gets the cell label representing the value of the given field info for the given object.
	 *
	 * @param  tableController Table controller that requires the cell label.
	 * @param  object          Object from which the field info value is represented in the table
	 *                         cell.
	 * @param  fieldInfo       Field info for which a cell label is required.
	 * @param  value           Value of the given field.
	 * @return Label that will be put in the table cell for the given object and field info (may be
	 *         null).
	 * @see    LyTableBehaviorInterface#updateCell(LySimpleTableController,LyCell,LyObject,LyFieldInfo)
	 */
	String getCellLabel(LySimpleTableController tableController, LyObject object,
		LyFieldInfo fieldInfo, Object value);

	/**
	 * Updates the label of the given cell for the given object and fieldinfo.
	 *
	 * @param tableController the table controller that requires the cell label
	 * @param cell            the cell to update
	 * @param object          the object from which the field info value is represented in the table
	 *                        cell
	 * @param fieldInfo       the field info for which a cell label is required
	 */
	void updateCell(LySimpleTableController tableController, LyCell cell, LyObject object,
		LyFieldInfo fieldInfo);

	/**
	 * Gets the footer label associated to the given field info for the given list of objects
	 * presented in the table.
	 *
	 * @param  tableController Table controller that requires the footer label.
	 * @param  fieldInfo       Field info for which a footer label is required.
	 * @param  objects         List of objects for which a footer label is computed.
	 * @return Label that will be put in the table footer that is corresponding to the given field
	 *         info for the given list of objects.
	 * @see    leon.info.LyFieldInfo#compute(LyObjectList)
	 */
	String getFooterLabel(LySimpleTableController tableController, LyFieldInfo fieldInfo,
		LyObjectList objects);

	/**
	 * Method called when the selection has changed in the view.
	 *
	 * @param controller the current table controller
	 */
	void selectionChanged(LySimpleTableController controller);

	/**
	 * Gets the content of the table corresponding of the given context.<br />
	 * The workspace must be dynamic for dynamic table updates.<br />
	 * If the returned workspace is null, the controller computes the content itself.
	 *
	 * @param  tableController Table controller that requires the content.
	 * @param  context         Context for which the content is required.
	 * @return the table content.
	 */
	LyWorkSpace getContent(LyTableController tableController, LyWorkSpace context);

	/**
	 * Gets the list of fieldinfos that must be loaded in the given table controller.<br />
	 * <i>The default implementation</i> returns all fieldinfos with the {@link
	 * LyFieldInfo#MARK_STATUS status mark} and all fieldinfos that are displayed in the table.
	 *
	 * @param  controller Table controller that requires the list.
	 * @return The list of fields info to load.
	 */
	LyInfoList<LyFieldInfo> getLoadedFieldInfos(LySimpleTableController controller);

	/**
	 * Gets the list of fieldinfos that will be used to compute/display values for the table.<br />
	 * <i>The default implementation</i> returns the list of {@link
	 * #getLoadedFieldInfos(LySimpleTableController) loaded} fields.
	 *
	 * @param  tableController the associated table controller
	 * @param  action          the table action
	 * @return a list of fieldinfos
	 */
	LyInfoList<LyFieldInfo> getDependsFieldInfos(LySimpleTableController tableController,
		LyAction action);

	/**
	 * Gets the list of fieldinfos that may be displayed in the column chooser tool.
	 *
	 * @param  controller Table controller that requires the list.
	 * @return the list of fields to display in the column chooser tool or null to let the
	 *         controller build this list.
	 */
	LyInfoList<LyFieldInfo> getChooseColumnsFieldInfos(LySimpleTableController controller);

	/**
     * Gets the initial column width of a fieldInfo in a table.<br />
     * <i>The default implementation</i> returns the "_colWidth" parameter from the table action.
     *
     * @param  controller Table controller that requires the list.
     * @param  fieldInfo  fieldInfo associated to this column.
     * @return the size (unit = number of characters or ratio of the full table width)
     *         of the column or -1 (default) to compute automatically.
     */
    double getColumnWidth(LySimpleTableController controller, LyFieldInfo fieldInfo);

	/**
	 * Method invoked at the beginning of the given object drag.
	 *
	 * @param tableController Table controller that displays the table
	 * @param sourceObject    dragged object
	 */
	void dragStart(LySimpleTableController tableController, LyObject sourceObject);

	/**
	 * Method called when the given dragged object is dropped on the given dropped object.
	 *
	 * @param tableController table controller that displays the table
	 * @param targetObject    object on where the dragged object is dropped
	 * @param sourceObject    the dragged object
	 */
	void drop(LySimpleTableController tableController, LyObject targetObject,
		LyObject sourceObject);

	/**
	 * Method called at the end of a drag-and-drop.
	 *
	 * @param tableController table controller that displays the tree
	 * @param type            drag-and-drop type ({@link org.eclipse.swt.dnd.DND#DROP_COPY} or
	 *                        {@link org.eclipse.swt.dnd.DND#DROP_MOVE})
	 */
	void dragFinished(LySimpleTableController tableController, int type);

    /**
     * Method used to enable a sort.
     *
     * @param tableController Table controller that asks for the sort.
     * @param sort            The sort that is asked.
     * @return The enabled sort.
     */
    LySort validateSort(LySimpleTableController tableController, LySort sort);
    
    /**
     * Enables groupBy feature on given field info.
     *
     * @param tableController Table controller that calls the checking.
     * @param fieldInfo Field info that is checked for groupBy feature.
     * @return True if enabled, false otherwise.
     */
    boolean enableGroupBy(LySimpleTableController tableController, LyFieldInfo fieldInfo);
    
    /**
     * Gets group name for given value on specified object.
     * 
     * @param  tableController Table controller that requires the group value label.
     * @param  object          Object for which the group is requested.
     * @param  value           Field value for which the group is requested.
     * @return The group label for given value.
     */
    String getGroupLabel(LySimpleTableController tableController, LyObject object, LyFieldInfoValue value);

    /**
     * Gets row color associated to the given group.
     *
     * @param  tableController Table controller that requires the row color.
     * @param  object an object displayed in a group
     * @param  value the LyFieldInfoValue used to compute the group color
     * @return Color name or the name of a color resource if any, null otherwise.
     */
    String getGroupColor(LySimpleTableController tableController, LyObject object, LyFieldInfoValue value);

    /**
     * Current builder of a new row item in a row.
     *
     * @param  tableController Table controller that requires the new row item.
     * @param  object object added in the contextual list, and which should set the content of row
     *                cells.
     * @param  fieldInfo  field value shown in the cell content
     * @return LyCell or LyFieldItem the new Cell description (XML description) or editable content
     *         to add in the LyRow component.
     */
    LyComponent buildRowItem(LySimpleTableController tableController, LyObject object, LyFieldInfo fieldInfo);

    /**
     * Updates the given component for the given object and fieldinfo.
     *
     * @param tableController the table controller that requires the update
     * @param component       the component to update.
     * @param object          the object from which the field info value is represented in the table.
     * @param fieldInfo       the field info for which a component is updated.
     */
    void updateRowItem(LySimpleTableController tableController, LyComponent component, LyObject object,
        LyFieldInfo fieldInfo);
    
    /**
     * Completes the given row that corresponds to spcified object.
     *
     * @param tableController the table controller that requires the completion.
     * @param row             the row to be completed.
     * @param object          the object that corresponds to the row.
     */
    void completeRow(LySimpleTableController tableController, LyRow row, LyObject object);
    
    /**
     * Gives the image to display in first column when display in list.
     * @param tableController the table controller.
     * @return a {@link LyFileFieldInfo} which contains the image else return null
     */
    LyFileFieldInfo getDisplayListFirstImage(LySimpleTableController tableController);
    
    /**
     * Gets label for object.
     * @param tableController	the table controller.
     * @param object          	Object for which the label is requested.
	 * @return The label for given object.
     */
    String getLabel(LySimpleTableController tableController, LyObject object);
   
    /**
     * Gets index for object.
     * @param tableController	the table controller.
     * @param object          	Object for which the index is requested.
	 * @return The index for given object.
     */
    String getIndex(LySimpleTableController tableController, LyObject object);
    
    /**
     * Return default criterion for a given field info. If no default sort is defined on the class,
     * return null.
     * @param tableController	the table controller.
     * @param  fieldInfo the field info we want to get the criterion
     * @return the default sort criterion for the given field info
     */
    public LySortCriterion getDefaultCriterion(LySimpleTableController tableController, LyFieldInfo fieldInfo);
}