/*
 * LyTreeBehaviorInterface.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.app.behaviorinterface;

import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.view.*;


/**
 * This interface defines the methods to implement in order to modify the default behavior of a tree
 * controller.
 *
 * @author  Lyria S.A.
 * @version $Id: LyTreeBehaviorInterface.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @see     leon.control.LyTreeController
 */
public interface LyTreeBehaviorInterface extends LyTreeBaseBehaviorInterface
{
	/**
	 * File version
	 */
	String __VERSION = "$Id: LyTreeBehaviorInterface.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Indicates if data in the tree are sorted or not.
	 *
	 * @param  controller current Tree Controller
	 * @return true if the tree must be sorted false otherwise
	 */
	boolean isSorted(LyTreeController controller);

	/**
	 * Gets children of the given parent node.
	 *
	 * @param      parentNodeObject parent of the requested objects.
	 * @param      classInfos       target (only for data tree)
	 * @param      controller       current LyTreeController instance
	 * @return     the list of children objects, or null if the subtree is empty
	 * @see        #getSons(LyObject, String, LyTreeController)
	 * @deprecated use {@link #getSons(LyObject, String, LyTreeController)}
	 */
	LyObjectList getTreeSons(Object parentNodeObject, LyInfoList<LyClassInfo> classInfos,
		LyTreeController controller);

	/**
	 * Gets all children objects of the given parent. A null parent object is considered as the root
	 * node.
	 *
	 * @param      parentNodeObject parent of the requested objects.
	 * @param      level            the current level of the tree, as declared in the metamodel
	 * @param      controller       the current tree controller
	 * @return     the list of children objects, or null if the subtree is empty
	 * @deprecated Since V3.3, use {@link #getSons(LyObject, String, LyTreeController)}.
	 */
	LyObjectList getSons(LyObject parentNodeObject, String level, LyTreeController controller);

	/**
	 * Gets all children objects of the given parent. A null parent object is considered as the root
	 * node.
	 *
	 * @param  parentNodeObject parent of the requested objects.
	 * @param  treeNode         the current tree node associated to given object
	 * @param  level            the current level of the tree, as declared in the metamodel
	 * @param  controller       the current tree controller
	 * @return the list of children objects, or null if the subtree is empty
	 */
	LyObjectList getSons(LyObject parentNodeObject, LyTreeNode treeNode, String level,
		LyTreeController controller);

	/**
	 * Gets the given node parent objects list.
	 *
	 * @param  nodeObject the node object
	 * @param  controller current LyTreeController instance
	 * @return the list of possible given node parents or null if the only parent is the root
	 */
	LyObjectList getParents(Object nodeObject, LyTreeController controller);

	/**
	 * Gets the list of classInfo that represent the possibles children of the given classInfo at
	 * the given level.
	 *
	 * @param  parentClassInfo the classInfo of the parent
	 * @param  parentLevel     the current level of the parent
	 * @param  controller      the current tree controller
	 * @return a list of classInfo
	 */
	LyInfoList<LyClassInfo> getSonsClassInfos(LyClassInfo parentClassInfo, String parentLevel,
		LyTreeController controller);

	/**
	 * Gets the list of all classInfos that can be found in the tree.
	 *
	 * @param  controller the current tree controller
	 * @return the list of classInfo
	 */
	LyInfoList<LyClassInfo> getClassInfos(LyTreeController controller);

	/**
	 * Method invoked when the selection has changed in the tree view.
	 *
	 * @param controller the current tree controller
	 */
	void selectionChanged(LyTreeController controller);

	/**
	 * Gets the sort used in the tree for the given level.
	 *
	 * @param  controller the current tree controller
	 * @param  level      the current level on tree
	 * @return the found sort
	 */
	LySort getSort(LyTreeController controller, String level);

	/**
	 * Gets the filter used in the tree for the given level.
	 *
	 * @param  controller the current tree controller
	 * @param  level      the current level on tree
	 * @return the found filter
	 */
	LyFilter getFilter(LyTreeController controller, String level);

	/**
	 * Gets the ordered list of pairs of (action, classInfo) to display in popup menu.
	 *
	 * @param  treeController Tree controller that displays the menu.
	 * @param  actions        List of pairs (action, classInfo) that the controller is about to
	 *                        present in popup menu.
	 * @param  classInfo      Class of the selected object.
	 * @param  sons           List of classes declared as sons of the given class info.
	 * @return List of pairs (action, classInfo) that the controller will display in menu.
	 */
	LyInfoDataList<LyAction, LyClassInfo> getExtraPopupMenuActions(LyTreeController treeController,
		LyInfoDataList<LyAction, LyClassInfo> actions, LyClassInfo classInfo,
        LyInfoList<LyClassInfo> sons);
}