/**
 * LyAreaController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyAreaController.java 40651 2012-09-04 10:38:24Z jlsaouli $
 */

package leon.control;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;

import leon.app.*;
import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.control.form.*;
import leon.data.*;
import leon.data.daemon.*;
import leon.info.*;
import leon.info.infointerface.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.mobile.*;


/**
 * An area controller is an abstract class used for factoring elementary view behavior. This class
 * is specialized into LyFormController, LyTableController, LyMapController for example. An
 * areaController is associated to an action and may have a context for the creation of the view.
 * (contextual views).
 *
 * <P>The area controller offer primitive to manage the content of of the view : menubar, titleBar,
 * messageBar.</P>
 *
 * <P>It provides description of current action, context, associated classes. The implementation of
 * all these primitive is usually overriden in subclasses.</P>
 */
public abstract class LyAreaController extends LyController
	implements LyGetActionInterface,
		LyContextInterface,
		LyInfoListener, LySelectFileListenerInterface
{
	/**
	 * File version
	 */
	public static final String	    __VERSION =
		"$Id: LyAreaController.java 40651 2012-09-04 10:38:24Z jlsaouli $";

	/** Closing area. */
	public static final String	    DEMAND_CLOSE = LyViewController.DEMAND_CLOSE;

	/** Interaction code on the command bar button : Apply inputs. Do not close the view. */
	public static final String	    DEMAND_APPLY = "_apply".intern();

	/** Cancel inputs. */
	public static final String	    DEMAND_CANCEL = DEMAND_CLOSE;

	/** Interaction code on the command bar button : Validate inputs. Apply and close. */
	public static final String	    DEMAND_VALIDATE = "_validate".intern();

	/** Menu id : delete preference. */
	public static final String	    DEMAND_DELETE_PREFERENCE = "_deletePreference".intern();

	/** Menu id : delete object info. */
	public static final String	    DEMAND_DELETE_OBJECTINFOS = "_deleteObjectInfos".intern();

	/** Creation of a new chart. */
	public static final String	    DEMAND_CREATE_CHART = "_chart".intern();

	/** Deletion of charts. */
	public static final String	    DEMAND_DELETE_CHART = "_chart_delete".intern();

	/** Update view demand. */
	public static final String	    DEMAND_UPDATE_VIEW = "_updateView".intern();

	/** Default size of image in menu items. */
	public static final int		    DEFAULT_MENU_ITEM_IMAGE_SIZE = 16;

	/** View controller that contains this area controller. */
	protected LyViewController	    _viewController;

	/** Action of this area controller. */
	protected LyAction			    _action;

	/** Target of the action. */
	protected String			    _target;

	/** Context of this area controller. */
	protected LyWorkSpace		    _context;

	/** Context client. */
	protected LyAreaController	    _contextClient;

	/** Multi-area controller managing this area controller (for dialog actions only). */
	protected LyMultiAreaController _areaStackController;

	/** Boolean indicating if this controller was initialized. */
	protected boolean			    _initialized;

	/** Boolean indicating if action was started or not. */
	protected boolean			    _started;

	/** Internal listener on context modifications. */
	protected LyWorkSpaceListener   _contextListener;

	/** Environment of this controller. */
	protected LyEnvironment		    _environment;

	/** Indicates if the environment of the controller was changed dynamically. */
	protected boolean			    _environmentChanged;

	/** The last data update time of this controller. */
	protected long				    _lastUpdated;

	/**
	 * Boolean indicating that the controller only builds sub-controllers (controllers that don't
	 * have their own view controller).
	 */
	protected boolean			    _buildInternalSubControllers;

	/**
	 * Boolean indicating that the controller has no view controller and uses the one of its parent.
	 */
	protected boolean			    _hasViewController;

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 * @param action  related action. If an action is specified, the controller uses action
	 *                parameters (action name, action specific XML view description ...)
	 * @param context current context (may be null if the view is not contextual to a list of
	 *                objects).
	 */

	public LyAreaController(LySession session, LyController parent, LyAction action,
		LyObjectList context)
	{
		super(session, parent,
			((action == null) ? null
				: ((action.getClassInfo() != null)
					? (action.getId() + CONTROLLER_ID_SEP + action.getClassInfo().getId())
					: action.getId())));

		if (action == null)
			throw new IllegalStateException("BF AppEngine : action parameter can't be null.");

		// Sets action
		_viewController = null;
		_action = action;
		_target = null;
		_contextClient = null;
		_areaStackController = null;
		_started = false;
		_lastUpdated = 0;
		_environmentChanged = false;
		_hasViewController = !(parent instanceof LyAreaController)
			 || !((LyAreaController)parent)._buildInternalSubControllers;

		// Init context
		initContext(context);

		setName(_action.getName());

		// Sets specific builder if one is specified for this action
		Class<?>	builderClass = action.getBuilderClass();

		setBuilderClass(builderClass);
	}

	/**
	 * Accessor to the internal builder. If no builder is set, the method returns the reference to a
	 * default builder (invokes getDefaultBuilder).
	 *
	 * @return the builder which creates the XML description of the view. Returned value is not null
	 *         and several invocation on getBuilder always return the same instance.
	 */
	public LyBuilder getBuilder()
	{
		if (_action == null)
			return null;

		LyEnvironment environment = getEnvironment();

		if (!environment.testEnv("_LY_IGNORE_APP_VIEWS"))
		{
			String view = _action.getView();

			if ((view != null) && (_builder == null))
				setBuilder(new LyFileBuilder(this, view, LyArea.class));
		}

		return super.getBuilder();
	}

	/**
	 * Prepare component before showing. This method is called once (unlike reset which may be
	 * called several times), before first showing. Default implementation invokes reset.
	 */
	protected void initArea()
	{
		if (_initialized)
			return;

        // Reset content view
        reset();

		// Start action on first access
		if (!startAction())
			free();
        else
            _lastUpdated = System.currentTimeMillis();

		_initialized = true;
	}

	/**
	 * Initializes this area controller.
	 */
	public void initialize()
	{
		initArea();
	}

	/**
	 * Checks if this controller was initialized or not.
	 *
	 * @return Boolean indicating intialization state or this controller.
	 */
	public boolean isInitialized()
	{
		return _initialized;
	}

	/**
	 * Clears all cached data for this controller (especially the view and the component).
	 *
	 * <P>This method may be invoked for memory optimization, but the controller is still valid and
	 * may be used. (Unlike free) Further usage, needs to rebuild internal component & view which
	 * may be time consuming.</P>
	 */
	public void clearViewCache()
	{
		super.clearViewCache();
		_initialized = false;
	}

	/**
	 * Perform a full reset (reset of datas and reset of graphical area). This method may be called
	 * several times.
	 *
	 * @see #reset(boolean)
	 */
	public void reset()
	{
        LyApplication   application = getApplication();
        LyEvaluator     evaluator = application.setCurrentEvaluator(this);

        reset(true);

        application.setCurrentEvaluator(evaluator);
	}

	/**
	 * Reset all the content of the view. This method may be called several times. This method
	 * should be overriden, default implementation makes nothing
	 *
	 * @param full Boolean indicating if a full reset (including reset of datas) is requiered or not
	 *             (only reset graphical area).
	 */
	public void reset(boolean full)
	{
		LyBehaviorInterface behavior = getBehavior();

		if (behavior instanceof LyActionBehaviorInterface)
			((LyActionBehaviorInterface)behavior).reset(this, full);
	}

	/**
	 * This method is called by the view controller when this area is shown to update its content.
	 *
	 * @return True if updated, false otherwise.
	 */
	public boolean update()
	{
		if (isIdle())
			return false;

		LyEnvironment environment = getEnvironment();
		String		  delayStr = environment.getResource("LY_DATA_UPDATE_DELAY");
		int			  delay = ((delayStr == null) ? -1 : LyTools.intFromString(delayStr));
	    long          time = System.currentTimeMillis();

		if ((delay >= 0) && (time > (_lastUpdated + (delay * 1000))))
		{
		    cleanObjectsForReload();
			reload();

			_lastUpdated = System.currentTimeMillis();
			return true;
		}
		else
			return false;
	}

	/**
	 * Clean objects before reloading.
	 */
	protected void cleanObjectsForReload()
	{
        LyObjectList   content = getContent();
        int            i, n = (content == null ? 0 : content.getSize());
        LyObject       object;

        for (i = 0; i < n; i++)
        {
            object = content.getObject(i);
            object.freeDaemonValues();
        }
	}

	/**
	 * Updates area controller content.
	 */
	protected void reload()
	{
		// Get content before reset
		LyObjectList oldContent = getContent();
		Vector<LyObject>		 objects = null;

		if (oldContent != null)
		{
			objects = oldContent.getObjects();
			// force unlock of objects to allow reloading
			oldContent = null;
			System.gc();
		}

		// Reset content
		reset();

		// Check objects that were in content and that are not found anymore
		LyObjectList   content = getContent();
		int			   i, n = ((objects == null) ? 0 : objects.size());
		LyObjectSource objectSource;
		LyObject	   object;
		LyResponse     response;

		for (i = 0; i < n; i++)
		{
			object = (LyObject)objects.elementAt(i);
			if ((content == null) || !content.contains(object))
			{
				// Check if object was deleted
				objectSource = object.getObjectSource();
				response = objectSource.reloadObject(object.getObjectId());
				if ((response.getStatus() == LyResponse.Status.COMPLETED)
					&& (response.getSize() == 0))
					object.delete(true); // delete must be local
			}
		}
	} // end method reload

	/**
	 * Prepare the component : default implementation invokes startAction on the behavior which may
	 * have extra processing.
	 *
	 * <p>This method is called at startup and each time a new object is shown in the form.</p>
	 *
	 * @see    leon.app.behavior.LyActionBehavior#startAction
	 * @return true if the action was successfully started, and false if the action failed / ended.
	 */
	protected boolean startAction()
	{
		if (_started)
			return true;

		LyBehaviorInterface behavior = getBehavior();

		if (behavior instanceof LyActionBehaviorInterface)
			_started = ((LyActionBehaviorInterface)behavior).startAction(this, _action);
		else
			_started = true;

		return _started;
	}

	/**
	 * Action ends. : default implementation invokes endAction on the behavior which may have extra
	 * processing.
	 *
	 * <p>This method is called at startup and each time the a new object is shown in the form.</p>
	 *
	 * @see    leon.app.behavior.LyActionBehavior#endAction
	 * @return true if the action was successfully stopped, and false if the action is still
	 *         'started'.
	 */
	protected boolean endAction()
	{
		if (!_started)
			return true;

		LyBehaviorInterface behavior = super.getBehavior(); // Don't construct it, if not

		if (behavior != null)
			_started = !((LyActionBehaviorInterface)behavior).endAction(this, _action);

		return !_started;
	}

	/**
	 * Update the titlebar content with default values Default implementation let the builder set
	 * the titlebar content if this service is supported by the builder.
	 *
	 * @see leon.control.builder.LyAreaBuilder#updateTitleBar
	 */
	protected void updateTitleBar()
	{
		if (_builder instanceof LyAreaBuilder)
			((LyAreaBuilder)_builder).updateTitleBar();

		updateTitleBar(null, null, null);
	}

	/**
	 * Update the title bar content (if any titlebar is associated with this area).
	 *
	 * @param  image        the image on the left side of the titlebar. If no image is given, try to
	 *                      get the image from the controller.
	 * @param  title        the string displayed on the left label of the titlebar. May be either a
	 *                      text message, or a String key found in the dictionary
	 *                      (StringXXX.properties) files. if to title is given, try to get a title
	 *                      from the behavior (@see LyActionBehavior#getTitle).
	 * @param  contextTitle the string displayed on the right side of the titlebar. May be either a
	 *                      text message, or a String key found in the dictionary
	 *                      (StringXXX.properties) files. If no string is given, try to construct
	 *                      one from the current context (if any).
	 * @return the titlebar view component
	 */
	public LyTitleBar updateTitleBar(String image, String title, String contextTitle)
	{
		if (_component == null)
			// Component not yet constructed or already freed
			return null;

		LyArea     area = (LyArea)_component;
		LyTitleBar titleBar = area.getTitleBar();

		if (titleBar == null)
		{
			if (_parent instanceof LyViewController)
			{
				LyToolBar     toolBar = area.getToolBar();
				LyEnvironment environment = getEnvironment();

				if (toolBar == null)
					toolBar = area.getLeftToolBar();

				if ((toolBar != null)
						&& !environment.testResource("LY_TOOL_LABEL_RIGHT_POSITION")
						&& environment.testResource("LY_TITLEBAR_LEFT_IMAGE"))
				{
					if (image == null)
						image = getImage();
					toolBar.setImage(image);
				}

				LyView view = getView();

				if (title == null)
					title = getTitle();
				view.setLabel(title);
			}
			return null;
		}

		if (_builder instanceof LyFileBuilder)
		{
			if (image == null)
				image = titleBar.getImage();
			if (title == null)
				title = titleBar.getLeftLabel();
//                      if (contextTitle == null)
//                              contextTitle = titleBar.getRightLabel();
		}

		if (image == null)
			image = getImage();

		if (title == null)
			title = getTitle();

		// check if the context title was not set previously
		if ((contextTitle == null)
				&& (_builder instanceof LyAreaBuilder))
			contextTitle = ((LyAreaBuilder)_builder).getRightLabel();

		LyEnvironment environment = getEnvironment();
		boolean		  rightLabelSensitive = true;

		if (contextTitle == null)
		{
			contextTitle = getContextTitle();

			if ((contextTitle == null)
					&& (_context != null))
			{
				int nc = _context.getSize();

				if (nc > 0)
				{
					LyObject object = _context.getObject(0);

					if (nc > 1)
						contextTitle = object.getName() + "...";
					else
					{
						contextTitle = object.getName();

						if (rightLabelSensitive)
						{
							LyAction defaultAction = _action.getDefaultAction();

							if (defaultAction == null)
							{
								LyClassInfo classInfo = object.getClassInfo();

								defaultAction = classInfo.getDefaultAction();
							}

							if (defaultAction == null)
								rightLabelSensitive = false;
						}
					}
				}
			}
		}

		if (environment.testResource("LY_TITLEBAR_LEFT_IMAGE"))
		{
			String imageTooltip = getImageTooltip();

			titleBar.setImage(image);
			titleBar.setImageToolTip(imageTooltip);
		}

		titleBar.setLeftLabel(title);
		titleBar.setRightImageToolTip(contextTitle);
		titleBar.setRightLabel(contextTitle);

		if ((contextTitle == null) || (contextTitle.length() == 0))
			rightLabelSensitive = false;

		titleBar.setRightLabelSensitive(rightLabelSensitive);

		return titleBar;
	} // end method updateTitleBar

	/**
	 * Update messages for the current controller.
	 */
	protected void updateMessages()
	{
	}

	/**
	 * Update the menu for the current controller.
	 */
	protected void updateMenuBar()
	{
	}

	/**
	 * This method is invoked to enable the the area to complete the menuBar of the view and add its
	 * own menu items. Default implementation invokes the method on the builder instance.
	 *
	 * @param menuBar the menuBar structure
	 * @param level   : level of current Controller. one of the LEVEL_XXX contants in
	 *                LyViewController
	 * @see   LyViewController
	 */
	public void completeMenuBar(LyMenuBar menuBar, int level)
	{
		LyBuilder builder = getBuilder();

		if (builder instanceof LyAreaBuilder)
			((LyAreaBuilder)builder).completeMenuBar(menuBar, level);
	}

	/**
	 * This method is invoked to enable the the area(s) to complete the message bar of the view and
	 * add its own message items. Default implementation invokes the method on the builder instance.
	 *
	 * @param messages the message bar structure
	 * @param level    : level of current Controller. one of the LEVEL_XXX contants. in
	 *                 LyViewController
	 * @see   LyViewController
	 */
	public void completeMessages(LyMessages messages, int level)
	{
		if (_action.hasMark(LyAction.MARK_NO_MESSAGES))
			return;

		if (_builder instanceof LyAreaBuilder)
			((LyAreaBuilder)_builder).completeMessages(messages, level);
	}

	/**
	 * Invoke the processing of an action on an object. This method determine the context of action
	 * execution, which depend on action MIN/MAX properties. The context may be (depending on
	 * selection and the action ) :
	 *
	 * <ul>
	 *   <LI>MAX < 0, MIN < 0 : Send selection or all content (min < 0). Unlimited selection.</LI>
	 *   <LI>MAX = 0, MIN < 0 : Independant from selection, send all content</LI>
	 *   <LI>MAX > 0, MIN < 0 : Send selection or all content. Limited selection (max).</LI>
	 *   <LI>MAX < 0, MIN = 0 : Send selection or empty. Unlimited selection (max).</LI>
	 *   <LI>MAX = 0, MIN = 0 : Independant from selection, not linked to any content. (class
	 *     action)</LI>
	 *   <LI>MAX > 0, MIN = 0 : Send selection or empty. Limited selection (max).</LI>
	 *   <LI>MAX < 0, MIN > 0 : Send selection or empty.</LI>
	 *   <LI>MAX = 0, MIN > 0 : Parse error : MIN < MAX</LI>
	 *   <LI>MAX > 0, MIN > 0 : Send selection or parse error : MIN < MAX.</LI>
	 * </ul>
	 *
	 * @param  action action being processed.
	 * @param  demand original demand (optional) for this action. May be null.
	 * @return the controller which was created by the action invocation or null if there is no
	 *         default action on this object.
	 * @see    leon.app.LySession#invokeAction
	 */
	protected LyController invokeAction(LyAction action, LyActionDemand demand)
	{
		LyObjectList context;

		if (action.hasMark(LyAction.MARK_NO_CONTEXT))
			context = null;
		else if (action.getMin() < 0)
		{
			// MIN < 0 : means any selection allowed, use content if nothing selected
			if (action.getMax() == 0)
				context = getContent();
			else
			{
				// take selection, even if any...
				context = getSelection();

				// ... otherwise take content
				if ((context == null) || (context.getSize() == 0))
					context = getContent();
			}
		}
		else if (action.getMin() == 0)
		{
			// MIN = 0 : means any selection allowed, use empty if nothing selected
			if (action.getMax() == 0)
			{
				// MIN = 0 and MAX = 0 action is independant from selection
				LyClassInfo classInfo = action.getClassInfo();

				if (classInfo != null)
					return invokeAction(action, classInfo);
				else
					return invokeAction(action, getClassInfo());
			}
			else
				// take selection, even if empty
				context = getSelection();
		}
		else
			// take selection, even if empty
			context = getSelection();

		// Do we have any content for action ?
		if ((context == null) || (context.getSize() == 0))
			return invokeAction(action, getClassInfo(), context, true);
		else
			return invokeAction(action, context);
	} // end method invokeAction

	/**
	 * Invoke the default action on an object. Default action is usually consultation, but the
	 * meta-model or a specific implementation can change the default action.
	 *
	 * <P>This method retrieve the default action then call invokeAction on parent class.</P>
	 *
	 * @param  object object of the action.
	 * @return the controller which was created by the action invocation or null if there is no
	 *         default action on this object.
	 * @see    leon.control.LyController#invokeAction
	 * @see    leon.app.LySession#invokeAction
	 */
	protected LyController invokeDefaultAction(LyObject object)
	{
		// Retrieve default action and invoke it on the given object
		LyClassInfo classInfo = object.getClassInfo();
		LyAction    defaultAction = getDefaultAction(object);
		LySession   session = getSession();

		if ((defaultAction != null) && session.validateAction(defaultAction, this))
			return invokeAction(defaultAction, classInfo, new LySimpleObjectList(object));
		else
			return null;
	}

	/**
	 * Invoke the delete action on an object (if there is one).
	 *
	 * @param  objects   objects of the action.
	 * @param  classInfo classInfo of the action.
	 * @return the controller which was created by the action invocation or null if there is no
	 *         delete action on these objects.
	 * @see    leon.control.LyController#invokeAction
	 * @see    leon.app.LySession#invokeAction
	 */
	protected LyController invokeDeleteAction(LyObjectList objects,
		LyClassInfo classInfo)
	{
		// Retrieve delete action and invoke it on the given objects
		LyAction deleteAction = classInfo.getActionWithId(LyAction.ID_DELETE);

		if (deleteAction != null)
			return invokeAction(deleteAction, classInfo, objects);

		if (!classInfo.isAbstract())
			return null;

		// Invoke delete action on each subList
		LyInfoList<LyClassInfo> classes = objects.getClasses();
		LyClassInfo			    subClassInfo;
		int					    i, n = classes.getSize();
		LyObjectList		    subList;
		LyController		    controller = null;
		LySession			    session = getSession();

		for (i = 0; i < n; i++)
		{
			subClassInfo = classes.getInfo(i);
			subList = objects.getList(session, subClassInfo);
			controller = invokeDeleteAction(subList, subClassInfo);
		}

		// Return last controller
		return controller;
	} // end method invokeDeleteAction

	/**
	 * Finds the best suited action for fast edition of given field for a given object.
	 *
	 * @param  object the object to edit
	 * @param  fieldInfo field info to be edited
	 * @return Action found for fast edition of specified field, could be null if not found.
	 */
	public LyAction getFastEditionAction(LyObject object, LyFieldInfo fieldInfo)
	{
		if (object == null)
			return null;

		if (fieldInfo.hasMark(LyFieldInfo.MARK_HIDDEN))
			return null;

		// Retrieve set action and invoke it on the given object
		LyClassInfo			    classInfo = object.getClassInfo();
		int					    i, n = classInfo.getActionCount();
		LyAction			    action, setAction = null, defaultSetAction = null;
		LyClassInfo			    actionClassInfo;
		LyInfoList<LyFieldInfo> fields;

		for (i = 0; (i < n) && (setAction == null); i++)
		{
			action = classInfo.getAction(i);
			actionClassInfo = action.getClassInfo();
			if ((actionClassInfo != null)
					&& (actionClassInfo != classInfo))
				continue;

			if (action.getRoot().getId().equals(LyAction.ID_SET))
			{
				fields = action.getFieldInfos(classInfo, true);
				if (fields == null)
				{
					if (defaultSetAction == null)
						defaultSetAction = action;
				}
				else if (fields.contains(fieldInfo))
					setAction = action;
			}
		}

		if (setAction == null)
		{
			if (defaultSetAction != null)
				setAction = defaultSetAction;
			else
				return null;
		}

		// Invoke the action but do not show the result before view is updated
		LySession session = getSession();

		if (!session.enableAction(setAction, this, new LySimpleObjectList(object), false))
			return null;

		return setAction;
	} // end method getFastEditionAction

	/**
	 * Starts the fast edition of an object.
	 *
	 * @param object    object to be edited
	 * @param fieldInfo field info to be edited
	 * @param x         x position where fast edition window must be displayed
	 * @param y         y position where fast edition window must be displayed
	 *
	 * @return the edition controller
	 */
	protected LyController invokeFastEdition(LyObject object, LyFieldInfo fieldInfo,
		int x, int y)
	{
		LyAction setAction = getFastEditionAction(object, fieldInfo);

		return invokeFastEdition(setAction, object, fieldInfo, x, y);
	}

	/**
	 * Starts the fast edition of an object.
	 *
	 * @param setAction action to use for fast edition.
	 * @param object    object to be edited
	 * @param fieldInfo field info to be edited
	 * @param x         x position where fast edition window must be displayed
	 * @param y         y position where fast edition window must be displayed
	 *
	 * @return the edition controller
	 */
	protected LyController invokeFastEdition(LyAction setAction, LyObject object,
		LyFieldInfo fieldInfo, int x, int y)
	{
		if ((setAction == null)
				|| (object == null)
				|| (fieldInfo == null))
			return null;

		LySession    session = getSession();
		LyClassInfo  classInfo = object.getClassInfo();
		LyController controller = session.invokeAction(this, setAction, classInfo,
				new LySimpleObjectList(object), false);
		boolean		 editable = (controller instanceof LySetController)
				&& ((LySetController)controller).setFastEditedField(fieldInfo);

		if (!editable)
		{
			if (controller instanceof LySetController)
			{
				// Force the construction of view controller, otherwise calling the free
				// method on the set controller will free current controller too !
				controller.getViewController();
				controller.free();
			}

			return null;
		}

		LySetController setController = (LySetController)controller;
		LyView		    view = setController.getView();

		if (view.getInternalView(false) != null)
		{
			setController.getViewController().clearViewCache();
			view = setController.getView();
		}

		LyArea area = view.getArea();

		// Update generated view:
		//  - remove decoration and set modal (if dialog)
		//  - remove toolbar
		//  - remove titlebar
		//  - remove any tab or border
		if (view instanceof LyDialog)
		{
			((LyDialog)view).setDecorated(false);
			((LyDialog)view).setModal(true);
		}

		LyForm     form = new LyForm("_fastEditionForm");
		LyCommands commands = area.getCommands();
		LyArea     centerArea = new LyArea();
		int		   i, n = area.getPanelCount();

		for (i = n - 1; i >= 0; i--)
			area.removePanel(area.getPanel(i));

		// No need to change component aspect in popup views
		LyViewManager viewManager = getViewManager();

		area.setToolBar(null);
		area.setTitleBar(null);
		form.setBorder(false);

		// Add the list of edited fields in the form
		LyFieldInfo			    field;
		LyFormField			    formField;
		LyInfoList<LyFieldInfo> fields = setController.getFields();

		for (i = 0, n = fields.getSize(); i < n; i++)
		{
			field = fields.getInfo(i);
			formField = setController.getFormField(field);

			if (formField != null)
				form.addFormField(formField);
		}

		if (viewManager.hasProperty("POPUP_VIEW"))
			area.addPanel(form);
		else
		{
			centerArea.addPanel(form);
			centerArea.setBorder(LyArea.Border.NONE);
			centerArea.setScrolling(LyArea.Scrolling.NO);
			area.setBorder(LyArea.Border.IN);
			area.setLabel(null);
			area.addPanel(centerArea);
			area.setScrolling(LyArea.Scrolling.NO);
			area.setCommands(null);

			if (commands == null)
			{
				commands = new LyCommands();

				LyCommand command;

				command = new LyCommand(LyFormController.DEMAND_VALIDATE, "LY_VALIDATE");
				command.setAction(LyCommand.Action.SUBMIT);
				commands.addCommand(command);

				command = new LyCommand(LyFormController.DEMAND_CANCEL, "LY_CANCEL");
				command.setAction(LyCommand.Action.CANCEL);
				commands.addCommand(command);
			}

			commands.setBorder(LyArea.Border.NONE);
			centerArea.setCommands(commands);
		}

		// Set the position of the filter window
		LyViewController viewController = setController.getViewController();

		viewController.initialize();

		// Set location before showing the view to avoid a clipping effect
		viewController.setLocation(x, y);

		setController.showView();

		int[] pos = view.getLocation();

		if ((x <= 0) && (y <= 0))
		{
			x = pos[0];
			y = pos[1];
		}

		// Ensure view is at the right position now that view is shown and has its size
		// computed (avoid view being out of screen)
		int[] screenSize = viewController.getViewManager().getScreenSize(view);
		int[] viewSize = view.getSize();
		int[] maxWindowSize = viewManager.getMaxWindowSize();
		int	  maxWidth = ((maxWindowSize != null) && (maxWindowSize[0] > 0) ? maxWindowSize[0] : screenSize[0]);
		int	  maxHeight = ((maxWindowSize != null) && (maxWindowSize[1] > 0) ? maxWindowSize[1] : screenSize[1]);

		if ((screenSize != null) && (viewSize != null))
		{
			x = Math.max(Math.min(x, maxWidth - viewSize[0]), 0);
			y = Math.max(Math.min(y, maxHeight - viewSize[1]), 0);
		}
		viewController.setLocation(x, y);

		return viewController;
	} // end method invokeFastEdition

	/**
	 * Invoke the create action for the given class info.
	 *
	 * @param  classInfo classInfo of the action.
	 * @return the controller which was created by the action invocation or null if there is no
	 *         create action on the given class.
	 * @see    leon.control.LyController#invokeAction
	 * @see    leon.app.LySession#invokeAction
	 */
	protected LyController invokeCreateAction(LyClassInfo classInfo)
	{
		// Retrieve create action and invoke it on the given objects
		LyAction createAction = classInfo.getActionWithId(LyAction.ID_CREATE);

		if (createAction != null)
			return invokeAction(createAction, classInfo);
		else
			return null;
	}

	/**
	 * Retrieve the default action for an object.
	 *
	 * @param  object object of the action.
	 * @return the default action for this object or null if there is no default action on this
	 *         object.
	 */
	public LyAction getDefaultAction(LyObject object)
	{
		if (object == null)
			return null;

		LyClassInfo classInfo = getClassInfo();
		LyAction    defaultAction = null;

		defaultAction = _action.getDefaultAction(); // search default action on current controller

		if ((defaultAction == null) || !object.instanceOf(classInfo))
		{
			if ((classInfo == null)
					|| !object.instanceOf(classInfo)
					|| classInfo.isAbstract())
			{
				classInfo = object.getClassInfo();
			}
			defaultAction = classInfo.getDefaultAction();
		}

		return defaultAction;
	}

	/**
	 * Set the area stack controller.<br />
	 * This method must be used when the LY_OPEN_DIALOGS_IN_SAME_VIEW resource is set.
	 *
	 * @param parent the multi-area controller that manages the stack of controllers.
	 */
	protected void setAreaStackController(LyMultiAreaController parent)
	{
		_areaStackController = parent;
	}

	/**
	 * Check if the area controller may be freed. This method may be overriden by subclasses. The
	 * current implementation always returns true.
	 * @param  showMessage indicates if the controller can show an warning message or not
	 * @return true if the free method can be called immediately, false otherwise.
	 */
	public boolean canFree(boolean showMessage)
	{
        if (_areaStackController != null)
            return _areaStackController.canFree(showMessage);
        else
            return canClose(showMessage);
    }

    /**
     * Check if the area controller may be closed. This method may be overriden by subclasses. The
     * current implementation always returns true.
     * @param showMessage indicates if the controller can show an warning message or not
     * @return true if the close method can be called immediately, false otherwise.
     */
    public boolean canClose(boolean showMessage)
    {
		return true;
	}

    /**
	 * Free this controller. Overrides method from LyController :<BR>
	 * If a parent is set, propagate the free request on parent<BR>
	 * If no parent is set, free current controller immediately.
	 *
	 * <P>The parent will then invoke free(LyController controller) Method is set final, since
	 * propagation in subclasses must use free(LyController controller)</P>
	 *
	 * @see #free(LyController)
	 */
	public void free()
	{
        if (isFreed())
            return;

		if (_areaStackController != null)
		{
			_areaStackController.openSubAreaController(this, false);
			_areaStackController = null;
			free(getParent(false));
		}
		else if (!_hasViewController)
		{
			if ((_action.getTarget() != null)
			        && (_session == _parent.getSession())) // AC
				hideView();
			free(_parent);
		}
		else if (_parent != null)
			_parent.free();
		else
			super.free();
	}

	/**
	 * Free this controller on request of the specified controller.
	 *
	 * @param controller controller instance which requested the freeing of this controller. If the
	 *                   controller is the parent, this controller is freed otherwise the method
	 *                   does nothing.
	 * @see   #free()
	 */
	public void free(LyController controller)
	{
		if (controller != _parent)
			return;

		// Free dialogs parented to this area controller
        LyController subController;
        int          i, nc = _session.getControllerCount();

        // Search for all dialog controllers
        for (i = nc - 1; i >= 0; i--)
        {
            subController = _session.getController(i);
            if ((subController instanceof LyDialogController)
                    && (subController.getParent() == this))
                ((LyDialogController)subController).free();
        }

        // Unregister controller on its parent
        if (_parent instanceof LyAreaController)
            ((LyAreaController)_parent).removeSon(this);

		// Stop properly current action
		endAction();

		// Generic part for free
		super.free();

		// Free context
		if (_context != null)
		{
			_context.free(_contextListener);
			_contextListener = null;
			_context = null;
		}

		// Free environment if specific
		if ((_environment != null)
				&& ((_action.getResources() != null)
					|| _environmentChanged))
			_environment.free();
		_environment = null;

		// Clear remaining attributes
		_contextClient = null;
		_viewController = null;
		_action = null;
		_target = null;
	} // end method free

	/**
	 * Invoked when an action occurs. At this level, the implementation manages
	 *
	 * <ul>
	 *   <li>HELP demands. (demand.getType = LyActionDemand.TYPE_ACTION AND demand.getSource() =
	 *     LyController.DEMAND_HELP)</li>
	 *   <li>user specific CHART demands. (demand.getType = LyActionDemand.TYPE_HREF AND
	 *     demand.getSource() = &quot;_selectFile&quot;</li>
	 * </ul>
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		String demandId = demand.getSource();

		if (demandId == null)
			return;

		boolean  selectFile = false;
		String   fileId = null;
		String[] params = LyString.stringToList(demandId, ID_SEP);

		if ((demand.getType() == LyActionDemand.TYPE_ACTION)
				&& (params != null)
				&& (params.length != 0)
				&& (params[0] != null))
		{
			String id = params[0].intern();

			if (id.equals("_tabs"))
				// Avoid warning on tabs selection
				return;

			if (id.equals(getId() + "_selectFile"))
			{
				selectFile = true;
				fileId = params[1];
			}
		}

		if ((demand.getType() == LyActionDemand.TYPE_HREF)
				&& demandId.equals("_selectFile"))
		{
			selectFile = true;
			fileId = demand.getValue();
		}

		if (selectFile)
		{
			LyApplication application = getApplication();
			LyInfo		  info = application.getInfo(fileId);

			// Loading a chart : create a _graphicalChart action.
			if (info instanceof LyChartInfo)
			{
				LyChartInfo  chartInfo = (LyChartInfo)info;
				LyClassInfo  chartClassInfo = application.getClassInfo(chartInfo.getClassInfoId());
				LyPreference preference = chartInfo.getPreference();

				if (preference != null)
					setPreference(preference);

				if (chartClassInfo == null)
				{
					LyEnvironment environment = getEnvironment();

					environment.logError("BF AppEngine : invalid LyClassInfo ID "
						+ chartInfo.getClassInfoId() + " on parsing ChartInfo");
				}
				else
				{
					String   chartActionId = info.getId() + '.' + "_graphicalChart";
					LyAction chartAction = application.getAction(chartActionId);

					if ((chartAction != null)
						&& (chartAction.getParameter("_chartInfo") != info))
					{
						chartAction.free();
						chartAction = null;
					}

					if (chartAction == null)
					{
						Hashtable<String, Object> parameters = new Hashtable<String, Object>(1);

						parameters.put("_chartInfo", info);
						chartAction = new LyAction(chartActionId,
							application.getAction("_graphicalChart"), chartClassInfo, parameters);
					}

					LyActionDemand newDemand = new LyActionDemand(LyActionDemand.TYPE_ACTION, chartAction.getId());

					actionDemand(newDemand);
					return;
				}
			}
			else if (fileId.equals(DEMAND_CREATE_CHART))
			{
                LyClassInfo    classInfo = getClassInfo();
                LyAction       chartAction = classInfo.getActionWithId(fileId);
				LyActionDemand newDemand = new LyActionDemand(LyActionDemand.TYPE_ACTION,
                        (chartAction == null ? fileId : chartAction.getId()));

				actionDemand(newDemand);
				return;
			}
			else if (fileId.equals(DEMAND_DELETE_CHART))
			{
				deleteCharts();
				return;
			}
		}

		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			if (demandId.startsWith("LY_DEBUG"))
			{
				handleDebugLevelChange(demandId);
				return;
			}
			else if (demandId.startsWith("LY_SHOW_JAVA_FILE_"))
            {
			    String   file = demandId.substring("LY_SHOW_JAVA_FILE_".length());

			    file = file.replace('.', File.separatorChar) + ".java";
			    showFile(file);
                return;
            }
	        else if (demandId.startsWith("LY_SHOW_JAVA_FILE_"))
            {
                String   file = demandId.substring("LY_SHOW_XML_FILE_".length());

                file = file.replace('.', File.separatorChar) + ".xml";
                showFile(file);
                return;
            }
			else if (demandId.equals("LY_DUMP_VIEW_WITH_CONTENT")
			         || demandId.equals("LY_DUMP_VIEW_WITHOUT_CONTENT")
			         || demandId.equals("LY_SNAPSHOT"))
			{
				LySelectFileController controller = new LySelectFileController(getSession(), this, getClassInfo());

				controller.addListener(this);
				controller.setUserData(demandId);
				if (demandId.equals("LY_SNAPSHOT"))
					controller.setMessage("LY_SELECT_JPEG_FILE");
				else
					controller.setMessage("LY_SELECT_XML_FILE");
				controller.showView();
				// --> Callback process in selectFileDemand (see : LySelectFileBehaviorInterface)

				return;
			}
			else if (demandId.equals(DEMAND_UPDATE_VIEW))
			{
				// When an update view demand occurs, nothing has to be done.
				// updateViewFromValues() has already been called in LyController.demandPerformed()
				return;
			}
			else if (demandId.equals(DEMAND_HELP))
			{
				// The session will handle this demand
				LySession session = getSession();

				session.processHelpAction(this);
				return;
			}
			else if (demandId.equals(DEMAND_CLOSE))
			{
				if (!_hasViewController)
				{
					if (canFree(true))
				    {
						hideView();
						free(_parent);
					}
				    return;
				}

				LyViewController viewController = getViewController();

				if ((viewController.getAreaController() instanceof LyDashboardController)
						|| (this instanceof LyDashboardController))
				{
					// Case of a dashboard controller with the single view mode :
					// close current view of the dashboard.
					LyDashboardController dashboardController;

					if (this instanceof LyDashboardController)
						dashboardController = (LyDashboardController)this;
					else
						dashboardController = (LyDashboardController)viewController
								.getAreaController();

					boolean singleView = dashboardController.useSingleView();

					if (singleView)
					{
						dashboardController.closeCurrentView(false);
						return;
					}
				}
				else if (_areaStackController != null)
				{
					if (canClose(true))
					{
						_areaStackController.openSubAreaController(this, false);
						_areaStackController = null;
						free(getParent(false));
					}
					return;
				}
			}
			else if (demandId.endsWith("_titleBar.leftImage")
					|| demandId.equals("leftImage")) // specific view
			{
				// CLOSING the view with the X to the right. The id is for former compatibility.
				LyViewController viewController = getViewController();
				LyController     parent = getParent();

				if ((viewController.getAreaController() instanceof LyDashboardController)
						|| (parent instanceof LyDashboardController))
				{
					// Case of a dashboard controller with the single view mode :
					// close current view of the dashboard.
					LyDashboardController dashboardController;

					if (parent instanceof LyDashboardController)
						dashboardController = (LyDashboardController)parent;
					else
						dashboardController = (LyDashboardController)viewController
								.getAreaController();

					if (dashboardController.getSocialAction() == getAction())
					{
						dashboardController.showSocialPanel();
						return;
					}

					boolean singleView = dashboardController.useSingleView();

					if (singleView)
					{
						dashboardController.closeCurrentView(true);
						return;
					}
				}

				if (viewController.getParent() == null)
				{
					LySession session = getSession();

					session.closeViews();
				}
				else
				{
					if (getViewManager().getCurrentWindowName() != null)
					{
						// We are in a popup views (can only occur in DHTML)
						// Close all the 'stack'
						LySession session = getSession();

						session.closeViews(getViewManager().getCurrentWindowName());
					}
					else
						viewController.close();
				}
				return;
			}
			else if (demandId.equals(getId() + "_titleBar.rightLabel")
					|| demandId.equals("rightLabel")) // specific view
			{
				// If action on the titlebar and context contains more than
				// one object, display context in a table
				LyObjectList context = getContext();
				int			 nc = ((context == null) ? 0 : context.getSize());

				if (nc < 1)
					return;
				else if (nc == 1)
					invokeDefaultAction(context.getObject(0));
				else
				{
					LySession			    session = getSession();
					LyApplication		    application = session.getApplication();
					LyContextListController contextController;
					LyAction			    action
					    = application.getActionWithId(LyAction.ID_CONTEXT, getClassInfo());

					contextController = new LyContextListController(session, this, action,
							getContext());
					contextController.showView();
				}
				return;
			}
			else if ("_validateField".equals(demand.getValue()))
			{
				// Test if the demand is a validation on an input field (may happen in web mode)
				LyComponent			component = getComponent();
				Vector<LyCommand>	commands = new Vector<LyCommand>();

				component.getComponents(LyCommand.class, commands);

				if ((commands != null) && (commands.size() > 0))
				{
					LyCommand		command = commands.elementAt(0);
					LyActionDemand	newDemand = new LyActionDemand(LyActionDemand.TYPE_ACTION, command.getId());

					newDemand.setParameterValues(demand.getParameterValues());
					actionDemand(newDemand);
					return;
				}
			}
			else if ("_closeAreaTip".equals(demand.getSource()))
			{
				LyComponent			component = getComponent();

				if (component instanceof LyArea)
					((LyArea)component).setTip(null);
			}
			else
			{
				// test if the demand identifer is an action and then execute it
				LyAction action = getApplication().getAction(demandId);

				LyComponent      comp = getComponent();
        LyComponent component = comp.getComponent(demandId);
				if (component instanceof LyArea)
				{
				  LyArea areacomponent = (LyArea)component;
				  if (areacomponent.isClosable())
				  {
				    areacomponent.setOpen(!areacomponent.isOpen());
				  }
				}

				if (action != null)
				{
					invokeAction(action);
					return;
				}
			}
		}

		// Tab demands are not managed (just avoid a warning message)
		if ((demand.getType() == LyActionDemand.TYPE_TAB_NEXT)
				|| (demand.getType() == LyActionDemand.TYPE_TAB_PREVIOUS))
			return;

		// Transmit demand to view controller that contains this area controller
		if (_viewController != null)
			_viewController.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Dump the view to an xml file.
	 *
	 * @param withContent true if the data are dumped with the view
	 * @param file the path of a file
	 */
	protected void dumpView(boolean withContent, String file)
	{
		LyView view = getView();

		if (view != null)
		{
			if (withContent)
				view.writeXml(file, true, true);
			else
				getSession().writeXml(getAction(), getClassInfo(), file, false);
		}
	}

	/**
	 * Take a snapshot of the view.
	 * @param file the path of the destination file
	 */
	protected void snapshot(String file)
	{
		if (file == null)
		{
			// If a file is not given, try to write to a file located in LY_APP_DIR
			String path = _environment.getEnv("LY_APP_DIR");

			if (_environment.testEnv("_LY_DATA_IN_TEMP_FILE"))
				path = _environment.getEnv("LY_TEMP_DIR") + "/data";

			if ((path != null) && (!path.equals("LY_APP_DIR")))
				file = path + "/" + "dump.jpg";
			else
				file = "dump.jpg";
		}
		LyView view = getView();

		getViewManager().takeSnapshot(view, file);
	}

	/**
	 * Handle the dynamic changes of the debug level.
	 *
	 * @param levelId the new debug level
	 */
	private void handleDebugLevelChange(String levelId)
	{
		if (levelId == null)
			return;

		if (_environment == null)
			getEnvironment();

		if ((_action.getResources() == null) && !_environmentChanged)
		{
			// build a new environment
			_environment = new LyEnvironment(null, getSession().getEnvironment(), false);
			_environmentChanged = true;
		}

		if (levelId.equals("LY_DEBUG_NONE"))
			_environment.setDebugLevel(0);
		else if (levelId.equals("LY_DEBUG_ALL"))
		{
			int debugLevel = LyEnvironment.DEBUG_ALL;

			_environment.setDebugLevel(debugLevel, !_environment.isDebugLevel(debugLevel));
		}
		else if (levelId.equals("LY_DEBUG_ERROR"))
		{
//                       remove debug all and upper debug level
			_environment.setDebugLevel(LyEnvironment.DEBUG_ALL, false);
			_environment.setDebugLevel(LyEnvironment.DEBUG_WARNING, false);
			_environment.setDebugLevel(LyEnvironment.DEBUG_INFO, false);
			_environment.setDebugLevel(LyEnvironment.DEBUG_TRACE, false);

			_environment.setDebugLevel(LyEnvironment.DEBUG_ERROR);

		}
		else if (levelId.equals("LY_DEBUG_WARNING"))
		{
//                       remove debug all and upper debug level
			_environment.setDebugLevel(LyEnvironment.DEBUG_ALL, false);
			_environment.setDebugLevel(LyEnvironment.DEBUG_INFO, false);
			_environment.setDebugLevel(LyEnvironment.DEBUG_TRACE, false);
			_environment.setDebugLevel(LyEnvironment.DEBUG_WARNING);
		}
		else if (levelId.equals("LY_DEBUG_INFO"))
		{
//                       remove debug all and upper debug level
			_environment.setDebugLevel(LyEnvironment.DEBUG_ALL, false);
			_environment.setDebugLevel(LyEnvironment.DEBUG_TRACE, false);
			_environment.setDebugLevel(LyEnvironment.DEBUG_INFO);
		}
		else if (levelId.equals("LY_DEBUG_TRACE"))
		{
//                       remove debug all and upper debug level
			_environment.setDebugLevel(LyEnvironment.DEBUG_ALL, false);
			_environment.setDebugLevel(LyEnvironment.DEBUG_TRACE);
		}
		else if (levelId.equals("LY_DEBUG_DBMS"))
		{
			int debugLevel = LyEnvironment.DEBUG_DBMS;

			_environment.setDebugLevel(LyEnvironment.DEBUG_ALL, false);
			_environment.setDebugLevel(debugLevel, !_environment.isDebugLevel(debugLevel));
		}
		else if (levelId.equals("LY_DEBUG_BUS"))
		{
			int debugLevel = LyEnvironment.DEBUG_BUS;

			_environment.setDebugLevel(LyEnvironment.DEBUG_ALL, false);
			_environment.setDebugLevel(debugLevel, !_environment.isDebugLevel(debugLevel));
		}
		else if (levelId.equals("LY_DEBUG_PERF"))
		{
			int debugLevel = LyEnvironment.DEBUG_PERF;

			_environment.setDebugLevel(debugLevel, !_environment.isDebugLevel(debugLevel));
		}
	} // end method handleDebugLevelChange

	/**
	 * Build the debug menu.
	 *
	 * @return the debug menu
	 */
	protected LyMenu getDebugMenu()
	{

		LyMenu menu = new LyMenu("LY_MENU_SWITCH_DEBUG_LEVEL");

		LyItem item = new LyItem("LY_DEBUG_ALL", "LY_ITEM_DEBUG_ALL");

		item.setController(getId());
		item.setEnabled(true);
		item.setCheckBox(true);
		item.setChecked(_environment.isDebugLevel(LyEnvironment.DEBUG_ALL));
		menu.addItem(item);

		item = new LyItem("LY_DEBUG_NONE", "LY_ITEM_DEBUG_NONE");
		item.setController(getId());
		item.setEnabled(true);
		item.setCheckBox(true);
		item.setChecked(_environment.isDebugLevel(0));
		menu.addItem(item);

		menu.addSep();
		boolean checked = _environment.isDebugLevel(LyEnvironment.DEBUG_ERROR)
				&& !_environment.isDebugLevel(LyEnvironment.DEBUG_WARNING);

		item = new LyItem("LY_DEBUG_ERROR", "LY_ITEM_DEBUG_ERROR");
		item.setController(getId());
		item.setEnabled(true);
		item.setCheckBox(true);
		item.setChecked(checked);
		menu.addItem(item);

		item = new LyItem("LY_DEBUG_WARNING", "LY_ITEM_DEBUG_WARNING");
		item.setController(getId());
		item.setEnabled(true);
		item.setCheckBox(true);
		checked = _environment.isDebugLevel(LyEnvironment.DEBUG_WARNING)
				&& !_environment.isDebugLevel(LyEnvironment.DEBUG_INFO);
		item.setChecked(checked);
		menu.addItem(item);

		item = new LyItem("LY_DEBUG_INFO", "LY_ITEM_DEBUG_INFO");
		item.setController(getId());
		item.setEnabled(true);
		item.setCheckBox(true);
		checked = _environment.isDebugLevel(LyEnvironment.DEBUG_INFO)
				&& !_environment.isDebugLevel(LyEnvironment.DEBUG_TRACE);
		item.setChecked(checked);
		menu.addItem(item);

		item = new LyItem("LY_DEBUG_TRACE", "LY_ITEM_DEBUG_TRACE");
		item.setController(getId());
		item.setEnabled(true);
		item.setCheckBox(true);
		item.setChecked(_environment.isDebugLevel(LyEnvironment.DEBUG_TRACE));
		menu.addItem(item);

		menu.addSep();

		item = new LyItem("LY_DEBUG_DBMS", "LY_ITEM_DEBUG_DBMS");
		item.setController(getId());
		item.setEnabled(true);
		item.setCheckBox(true);
		item.setChecked(_environment.isDebugLevel(LyEnvironment.DEBUG_DBMS));
		menu.addItem(item);

		item = new LyItem("LY_DEBUG_BUS", "LY_ITEM_DEBUG_BUS");
		item.setController(getId());
		item.setEnabled(true);
		item.setCheckBox(true);
		item.setChecked(_environment.isDebugLevel(LyEnvironment.DEBUG_BUS));
		menu.addItem(item);

		item = new LyItem("LY_DEBUG_PERF", "LY_ITEM_DEBUG_PERF");
		item.setController(getId());
		item.setEnabled(true);
		item.setCheckBox(true);
		item.setChecked(_environment.isDebugLevel(LyEnvironment.DEBUG_PERF));
		menu.addItem(item);

		return menu;
	} // end method getDebugMenu

	/**
	 * Invoked when an menu item notification is sent to this listener.
	 *
	 * <P>This method is called by demandPerformed and may be overriden in subclasses.</P>
	 *
	 * <P>Default implementation manages menud demands on titlebars (XML dumps, snapshots).</P>
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   #demandPerformed
	 */
	public void menuDemand(LyMenuDemand demand)
	{
		String demandId = demand.getSource();

		if (demandId == null)
			return;

		if (demand.getType() == LyMenuDemand.TYPE_POPUP)
		{
			// chek if it's a popup menu demand from titleBar
			if (demandId.equals(getId() + "_titleBar"))
			{
				LyPopupMenu popupMenu = new LyPopupMenu();
				LyMenu	    menu = new LyMenu("LY_MENU_DUMP_VIEW");
                LyItem      item;

				popupMenu.addMenu(menu);

                item = new LyItem("LY_DUMP_VIEW_WITH_CONTENT","LY_ITEM_DUMP_VIEW_WITH_CONTENT");
                item.setDialog(true);
				item.setController(getId());
				item.setEnabled(true);
				menu.addItem(item);

				item = new LyItem("LY_DUMP_VIEW_WITHOUT_CONTENT","LY_ITEM_DUMP_VIEW_WITHOUT_CONTENT");
                item.setDialog(true);
				item.setController(getId());
				item.setEnabled(true);
				menu.addItem(item);

				item = new LyItem("LY_SNAPSHOT", "LY_MENU_SNAPSHOT");
                item.setDialog(true);
				item.setController(getId());
				item.setEnabled(true);

				popupMenu.addItem(item);
				popupMenu.addSep();

				popupMenu.addMenu(getDebugMenu());
                popupMenu.addSep();

                item = new LyItem("LY_SHOW_JAVA_FILE_" + getClass().getName(),
                    "Controller : " + getClass().getName());
                popupMenu.addItem(item);

                if (getBehavior() != null)
                {
                    item = new LyItem("LY_SHOW_JAVA_FILE_" + getBehavior().getClass().getName(),
                        "Behavior : " + getBehavior().getClass().getName());
                    popupMenu.addItem(item);
                }

                if (_action.getView() != null)
                {
                    item = new LyItem("LY_SHOW_XML_FILE" + _action.getView(),
                            "View : " + _action.getView());
                    popupMenu.addItem(item);
                }
                else if (getBuilder() != null)
                {
                    item = new LyItem("LY_SHOW_JAVA_FILE_" + getBuilder().getClass().getName(),
                            "Builder : " + getBuilder().getClass().getName());
                    popupMenu.addItem(item);
                }

				int[] p = demand.getPosition();

				popupMenu.setX(p[0]);
				popupMenu.setY(p[1]);

				LyArea     area = (LyArea)getComponent();
				LyTitleBar titleBar = area.getTitleBar();

				if (titleBar == null)
					// An error occured, just return
					return;

				popupMenu.setController(getId());
				popupMenu.setEnvironment(titleBar.getEnvironment());
				titleBar.setPopupMenu(popupMenu);
			}
		}
	} // end method menuDemand

	/**
	 * Build a contextual menu with the given demand based on the given class info.
	 * @param demand the menu demand
	 * @param classInfo the classInfo
	 * @param selection true if this demand applies to a selection
	 * @return  the popup menu
	 */
	protected LyPopupMenu buildPopupMenu(LyMenuDemand demand,
		LyClassInfo classInfo, boolean selection)
	{
		if (classInfo == null)
			return null;

		// Get all the actions in a LyInfoDataList.
		LyInfoDataList<LyAction, LyClassInfo> actionsList = getPopupMenuActions(classInfo,
				selection);
		LyPopupMenu							  popupMenu = new LyPopupMenu();
		LyAction							  action;
		LyItem								  item;
		boolean								  enabled;
		LyAction							  action2;
		LyMenu								  subMenu;
		LyClassInfo							  classInfo2, classInfo3, selectedClass;
		LyInfoList<LyClassInfo>				  selectedClasses;
		boolean								  subMenuEnabled;
		int									  min;
		boolean								  multiple;
		int									  i, j, n = actionsList.getSize();
		String								  image;
		String								  group;
		Vector<String>						  groups = new Vector<String>();
		boolean								  nullGroup = false;
	    LyEnvironment                         environment = getEnvironment();
	    boolean                               groupByImage = environment.testResource("LY_GROUP_ACTIONS_BY_IMAGE");

		// Find all groups
		for (i = 0; i < n; i++)
		{
			action = actionsList.getInfo(i);
			group = action.getGroup();
			if ((group == null) && !nullGroup)
			{
				groups.add(null);
				nullGroup = true;
			}
			else if ((group != null) && !groups.contains(group))
				groups.add(group);
		}

		int			 g, ng = groups.size();
		LyObjectList selections = getSelection();

		for (g = 0; g < ng; g++)
		{
			group = groups.elementAt(g);
			if (g > 0)
				popupMenu.addSep();

			// For each action in the super class, search for equivalent action
			// in subclasses.
			for (i = 0; i < actionsList.getSize(); i++)
			{
				action = actionsList.getInfo(i);
				if (action.getGroup() != group)
					continue;

				classInfo3 = actionsList.getData(i);
				multiple = false;
				min = action.getMin();
				// Check if we have multiple occurrences of the same action image
				// if so, build a tool with a sub menu
				subMenu = null;
				for (j = i + 1; j < actionsList.getSize(); j++)
				{
					action2 = actionsList.getInfo(j);
					if (groupByImage
				        && (action.getRoot() == action2.getRoot())
						&& (action.getGroup() == action2.getGroup())
						&& (getImage(action).equals(getImage(action2))))
					{
						classInfo2 = actionsList.getData(j);
						if ((classInfo2 == classInfo)
								|| (min == 0))
						{
							if (subMenu == null)
							{
								subMenu = new LyMenu(action.getName());
								subMenu.setController(getId());
								subMenu.setId(action.getId());

								item = buildPopupMenuItem(action, classInfo3, action,
										((classInfo3 == classInfo) ? null : classInfo3));
								item.setController(getId());
								if (action.hasMark(LyAction.MARK_DIALOG))
									item.setDialog(true);
								// item.setShortcut(action.getShortcut());
								subMenu.addItem(item);

								// enable or disable ?
								enabled = enableAction(action);

								// for super class don't make this test.
								if (classInfo3 == classInfo)
									subMenuEnabled = enabled;
								else
								{
									selectedClasses = null;
									if (action.getMin() > 0)
										selectedClasses = ((selections != null)
												? selections.getClasses() : null);
									subMenuEnabled = enabled;
									if (selectedClasses != null)
									{
										if (classInfo3.isAbstract())
										{
											subMenuEnabled = enabled;
											for (int k = 0; k < selectedClasses.getSize(); k++)
											{
												selectedClass = selectedClasses.getInfo(k);
												subMenuEnabled = subMenuEnabled
														&& (selectedClass.inheritsFrom(classInfo3));
											}
										}
										else
										{
											subMenuEnabled = enabled
													&& (selectedClasses.getSize() == 1);
											selectedClass = selectedClasses.getInfo(0);

											subMenuEnabled = subMenuEnabled
													&& (selectedClass == classInfo3);
										}
									}
								}
								item.setEnabled(subMenuEnabled);
							}

							item = buildPopupMenuItem(action2, classInfo2, action2,
									((classInfo2 == classInfo) ? null : classInfo2));
							item.setController(getId());
							if (action.hasMark(LyAction.MARK_DIALOG))
								item.setDialog(true);
							// item.setShortcut(action2.getShortcut());
							subMenu.addItem(item);
							// enable or disable ?
							enabled = enableAction(action2);

							// for super class don't make this test.
							if (classInfo2 == classInfo)
								subMenuEnabled = enabled;
							else
							{
								selectedClasses = null;
								if (action2.getMin() > 0)
									selectedClasses = ((selections != null)
											? selections.getClasses() : null);
								subMenuEnabled = enabled;
								if (selectedClasses != null)
									subMenuEnabled = (enabled
												&& (selectedClasses.getSize() == 1)
												&& (selectedClasses.contains(classInfo2)));
							}

							item.setEnabled(subMenuEnabled);
						}
						multiple = true;
						actionsList.removeInfo(action2, classInfo2);
						j--;
					}
				}

				if (subMenu != null)
					popupMenu.addMenu(subMenu);
				else
				{
					if (multiple)
					{
						item = buildPopupMenuItem(action, null, action, null);
						item.setController(getId());
						if (action.hasMark(LyAction.MARK_DIALOG))
							item.setDialog(true);
						image = getMenuItemImage(action);
						item.setImage(image);
						// item.setShortcut(action.getShortcut());

						// enable or disable ?
						subMenuEnabled = enableAction(action);
						item.setEnabled(subMenuEnabled);
					}
					else
					{
						item = buildPopupMenuItem(action, classInfo3, action, null);
						item.setController(getId());
						if (action.hasMark(LyAction.MARK_DIALOG))
							item.setDialog(true);
						image = getMenuItemImage(action);
						item.setImage(image);
						// item.setShortcut(action.getShortcut());

						// enable or disable ?
						enabled = enableAction(action);

						// for super class don't make this test.
						if (classInfo3 == classInfo)
							subMenuEnabled = enabled;
						else
						{
							selectedClasses = null;
							if (action.getMin() > 0)
								selectedClasses = ((selections != null) ? selections.getClasses()
																		: null);
							subMenuEnabled = enabled;
							if (selectedClasses != null)
								subMenuEnabled = (enabled
											&& (selectedClasses.getSize() == 1)
											&& (selectedClasses.contains(classInfo3)));
						}
						item.setEnabled(subMenuEnabled);
					}
					popupMenu.addItem(item);
				}
			}
		}

		// Check cross references
		if (popupMenu != null)
		{
			String		  displayString = environment.translate("LY_ITEM_XREF");

			if (((displayString != null) && (displayString.length() > 0)))
			{
				// Add a submenu with all known classes
				LyApplication		    application = _session.getApplication();
				LyInfoList<LyClassInfo> classes = application.getClassInfos();
				String				    actionId;
				LyAction			    classAction;

				action = application.getAction("_xref");
				classes = classes.getInfosByName(environment);

				subMenu = new LyMenu("LY_ITEM_XREF");
				subMenu.setId(((_parent == null) ? "_menuXRef" : (_parent.getId() + "_menuXRef")));

				for (i = 0, n = classes.getSize(); i < n; i++)
				{
					classInfo2 = classes.getInfo(i);
					if (classInfo2 == classInfo)
						continue;
					if (classInfo2.getId().startsWith("__"))
					    continue;
					if (classInfo.getRoutes(classInfo2) == null)
						// No XREF between the two classes
						continue;

					// Check that xref is enabled
					actionId = "_xref" + LyAction.ACTION_ID_SEP + classInfo2.getId();
					classAction = application.getAction(actionId);

					if (classAction == null)
						classAction = new LyAction(actionId, action, classInfo2, null);
					if (!_session.validateAction(classAction, this))
						continue;

					item = buildPopupMenuItem(null, classInfo2, null, classInfo2);
					item.setImage(getImage(classInfo2));
					item.setEnabled(enableAction(classAction));
					subMenu.addItem(item);
				}

				if (subMenu.getComponentCount() > 0)
				{
					if (popupMenu.getComponentCount() > 0)
						popupMenu.addSep();

					popupMenu.addMenu(subMenu);
				}
			}
		}
		return popupMenu;
	} // end method buildPopupMenu

	/**
	 * Build a contextual menu item with given actions and class infos.
	 *
	 * @param  idAction       action used to generate item identifier.
	 * @param  idClassInfo    class info used to generate item identifier.
	 * @param  labelAction    action used to generate item label.
	 * @param  labelClassInfo class info used to generate item label.
	 * @return the menu item build from given actions and class infos.
	 */
	protected LyItem buildPopupMenuItem(LyAction idAction, LyClassInfo idClassInfo,
		LyAction labelAction, LyClassInfo labelClassInfo)
	{
		String id = ((idAction == null) ? "" : idAction.getId());
		String label = ((labelAction == null) ? "" : labelAction.getName());

		if ((idAction != null) && (idClassInfo != null))
			id += LyController.ID_SEP;

		if (idClassInfo != null)
			id += idClassInfo.getId();

		if ((labelAction != null) && (labelClassInfo != null))
			label += " : ";

		if (labelClassInfo != null)
			label += labelClassInfo.getName();

		return new LyItem(id, label);
	}

	/**
	 * Get a list of popup action for the given class
	 * @param classInfo a classInfo
	 * @param selection true if the demand applies to a selection
	 * @return a list of action
	 */
	protected LyInfoDataList<LyAction, LyClassInfo> getPopupMenuActions(LyClassInfo classInfo,
		boolean selection)
	{
		LySession			         session = getSession();
		LyInfoList<LyClassInfo>      subClasses = classInfo.getSubClasses(true);
        int                          i, j, ns, na;
        LyInfoList<LyAction>         list;
        LyAction                     action;
        LyClassInfo                  subClass;
        LyInfoDataList<LyAction, LyClassInfo> actionsList =
                new LyInfoDataList<LyAction, LyClassInfo>(true);

        if (subClasses == null)
            subClasses = new LyInfoList<LyClassInfo>();
        subClasses.addInfo(classInfo, 0);

        // Get all actions for all subclasses.
        for (i = 0, ns = subClasses.getSize(); i < ns; i++)
        {
            subClass = subClasses.getInfo(i);

            if (subClass.isAbstract())
                continue;

            list = subClass.getActions(LyAction.MARK_HIDDEN, true);
            na = (list == null ? 0 : list.getSize());

            for (j = 0; j < na; j++)
            {
                action = list.getInfo(j);

                // Skip actions which should not be here
                if (action.hasMark(LyAction.MARK_HIDDEN))
                    continue;
                // If clic on selection, the action must enable selected items
                if (selection && (action.getMax() == 0))
                    continue;
                // If clic on background, the action must enable 0 items
                if (!selection && (action.getMin() > 0))
                    continue;
                if (!session.validateAction(action, this))
                    continue;

                actionsList.addInfo(action, subClass);
            }
        }

        return actionsList;
    }

	/**
	 * Check if the given action is enabled in the context of this controller.
	 *
	 *@param action the action to check
	 * @return True if the action is enabled, false if disabled.
	 */
	protected boolean enableAction(LyAction action)
	{
		LySession session = getSession();

		return session.enableAction(action, this, getSelection(), false);
	}

	/**
	 * Accessor to the action associated to this controller. Area controllers are usually associated
	 * to an action, but this parameter may be null.
	 *
	 * @return The reference to associated action
	 */
	public LyAction getAction()
	{
		return _action;
	}

	/**
	 * Accessor for the action target specified when calling showView.
	 *
	 * @return Name of the action target.
	 */
	public String getTarget()
	{
		return _target;
	}

	/**
	 * Get current controller context.
	 *
	 * @return Current context, or null if no context is set for the view
	 * @see    LyContextInterface#getContext
	 */
	public LyWorkSpace getContext()
	{
		return _context;
	}

	/**
	 * Initialize the context for the controller.
	 *
	 * @param context Current context, or null if no context is set for the view
	 * @see   LyContextInterface#getContext
	 */
	protected void initContext(LyObjectList context)
	{
		if (_action.hasMark(LyAction.MARK_NO_CONTEXT))
			return;

		if (_context != null)
		{
			_context.free(_contextListener);
			_contextListener = null;
		}

		if ((context != null) && !context.isDynamic())
			_context = new LySimpleWorkSpace(getSession(), context, true);
		else
			_context = (LyWorkSpace)context;

		if (_context != null)
		{
			_contextListener = buildContextListener();
			_context.addListener(_contextListener);
		}
	}

	/**
	 * Set the context for the controller.
	 *
	 * @param context Current context, or null if no context is set for the view
	 * @return true if successfull
	 * @see   LyContextInterface#setContext
	 * @see   #setContext(LyWorkSpace)
	 */
	public final boolean setContext(LyObjectList context)
	{
		if (_action.hasMark(LyAction.MARK_NO_CONTEXT))
			return false;

		boolean result;

		if ((context != null)
				&& !context.isDynamic())
		{
			context = new LySimpleWorkSpace(getSession(), context, true);

			result = setContext((LyWorkSpace)context);

			LyWorkSpace newContext = getContext();

			if (newContext != context)
				// If the created context is not used we must free it !
				((LyWorkSpace)context).free(null);
		}
		else
			result = setContext((LyWorkSpace)context);
		return result;
	}

	/**
	 * Set the current context for the controller. If the context is unchanged, this method just
	 * call reset(). If there is a new context, this method free the current one and set the new
	 * one.
	 *
	 * @param context Current context, or null if no context is set for the view
	 * @return true if done
	 */
	public boolean setContext(LyWorkSpace context)
	{
		return changeContext(context);
	}

	/**
	 * Change the context
	 * @param context the new context
	 * @return true if done
	 */
	protected final boolean changeContext(LyWorkSpace context)
	{
		if (_action == null)
			return false;

		if (_action.hasMark(LyAction.MARK_NO_CONTEXT))
			return false;

		// Check if nothing has obviously changed
		boolean needUpdate = true;

		if (context != _context)
		{
			if (_context != null)
			{
				// Fast check if anything has changed : check changes if the list is 'small'
				// Not worth it for big lists
				if (_context.equals(context))
					needUpdate = false;

				_context.free(_contextListener);
				_contextListener = null;
			}
			if ((context != null) && (context.isDynamic()))
			{
				_contextListener = buildContextListener();
				context.addListener(_contextListener);
			}
			_context = context;
		}
		else
			needUpdate = false;

		if (!_initialized)
			return false;

		// FFT 726 : Reset content if anything may have changed
		if (needUpdate)
			reset();
		return needUpdate;
	} // end method changeContext

	/**
	 * Build a context listener
	 * @return the newly build context listener
	 */
	protected _LyContextListener buildContextListener()
	{
		return new _LyContextListener();
	}

	/**
	 * Update context client. This method update the context of the client area. The object list
	 * passed to the client depends of selection and action. If there is a non empty selection, this
	 * method send the selection. If there is no selection or an empty selection: If the action of
	 * the context client is null or has a MIN value equals to 0, this method sends an empty list,
	 * otherwise, it sends the content of the controller.
	 */
	protected void updateContextClient()
	{
		LyAreaController contextClient = getContextClient();

		if ((contextClient != null)
				&& !contextClient.isFreed())
		{
			LyObjectList clientContext = getClientContext();

			contextClient.setContext(clientContext);
		}
	}

	/**
	 * Gets context to set on client context.
	 *
	 * @return the client context
	 */
	protected LyObjectList getClientContext()
	{
		LyAreaController contextClient = getContextClient();

		if (contextClient == null)
			return null;

		LyObjectList selection = getSelection();
		LyObjectList clientContext;

		if ((selection == null) || (selection.getSize() == 0))
		{
			LyAction action = contextClient.getAction();

			if ((action != null) && (action.getMin() == 0))
				clientContext = new LySimpleObjectList();
            else if ((action != null) && (action.getMin() < 0))
                clientContext = null;
			else
				clientContext = new LySimpleObjectList(getContent());
		}
		else
			clientContext = selection;

		LyActionBehaviorInterface behavior = (LyActionBehaviorInterface)getBehavior();

		if (behavior != null)
			clientContext = behavior.getClientContext(this, clientContext);
		return clientContext;
	} // end method getClientContext

	/**
	 * Accessor to the context client area controller which will receive contextual notification
	 * when something occurs on this area controller.
	 *
	 * <P>Used for contextual views and compound actions.</P>
	 *
	 * @return the contextual controller, or null if no client is defined
	 */
	public LyAreaController getContextClient()
	{
		return _contextClient;
	}

	/**
	 * Set accessor to the context client area controller which will receive contextual notification
	 * when something occur on this area controller.
	 *
	 * <P>Used for contextual views and compound actions.</P>
	 *
	 * @param controller the contextual controller, or null if no client defined
	 */
	public void setContextClient(LyAreaController controller)
	{
		_contextClient = controller;
	}

	/**
	 * Gets value corresponding to the given parameter name. Shortcut for the getParameter(String,
	 * boolean) method with <i>allowUnknowValue</i> set to false.
	 *
	 * @param  name Parameter name.
	 * @return Value of the parameter if found, an exception if the parameter does not exists
	 * @see    leon.info.LyAction#getParameter
	 * @see    #getParameter(String, boolean)
	 */
	public Object getParameter(String name)
	{
		return getParameter(name, false);
	}

	/**
	 * Gets value corresponding to the given parameter.
	 *
	 * @param     name         Parameter name.
	 * @param     unknown_value Boolean indicating if an exception is raised if this parameter
	 *                         doesn't exist on the controller action.
	 * @return    untyped value of the parameter if found. Returns null if no parameter with the
	 *            given name is found on this action and <i>allowUnkownValue</i> is set to true
	 *            otherwise exceptions are thrown.
	 * @exception IllegalStateException    If no parameters are registered on this action (never
	 *                                     thrown if <i>allowUnkownValue</i> is on).
	 * @exception IllegalArgumentException If the action doesn't have a parameter with the given
	 *                                     name. (never thrown if <i>allowUnkownValue</i> is on).
	 */
	public Object getParameter(String name, boolean unknown_value)
	{
		Object parameter = _action.getParameter(name, unknown_value);

		if ((parameter instanceof String)
				&& (((String)parameter).charAt(0) == '$'))
		{
			// Retrieve value on session
			Object value = _session.getValue(parameter);

			if (value == null)
				// If not found, try with environment variables
				value = _session.getEnvironment().getEnv(((String)parameter).substring(1));

			return value;
		}
		else
			return parameter;
	}

	/**
	 * Tests if the given parameter has "true" as value.
	 *
	 * @param     name Parameter name.
	 *
	 * @return true if the parameter has the "true" value
	 * @exception IllegalArgumentException If the action doesn't have a parameter with the given
	 *                                     name. (never thrown if <i>allowUnkownValue</i> is on).
	 */
	public boolean testParameter(String name)
	{
		Object value = getParameter(name, false);

		if ((value instanceof String)
				&& ((String)value).equalsIgnoreCase("true"))
			return true;
		else
			return false;
	}

	/**
	 * Gets value corresponding to the given parameter mark. The result is a BitSet. A type is given
	 * to known if marks apply to action, classInfo, fieldInfo or filter.
	 *
	 * @param  name the name of the parameter
	 * @param  type the type of expected bitset (for validation) Possible values are :
	 *              LyBitSet.Type.A_MARKS, TYPE_F_MARKS, TYPE_C_MARKS...
	 * @return the marks
	 */
	public BitSet getParameterMarks(String name, LyBitSet.Type type)
	{
		return _action.getParameterMarks(name, type);
	}

	/**
	 * Get controller class info if any. This method is overriden if it makes sense to return a
	 * classInfo for this controller. Default returns null.
	 *
	 * @return a classinfo if any or null.
	 */
	public LyClassInfo getClassInfo()
	{
		return ((_action == null) ? null : _action.getClassInfo());
	}

	/**
	 * Gets title associated with this controller.
	 *
	 * @return the title corresponding to the action.
	 */
	public String getTitle()
	{
		LyActionBehaviorInterface behavior = (LyActionBehaviorInterface)getBehavior();
		String					  title = null;

		if (behavior != null)
			title = behavior.getTitle(this, _action);
		if (title == null)
			title = _action.getName();

		return title;
	}

	/**
	 * Gets context title associated with this controller.
	 *
	 * @return the title corresponding to the action.
	 */
	public String getContextTitle()
	{
		LyActionBehaviorInterface behavior = (LyActionBehaviorInterface)getBehavior();
		String					  contextTitle = null;

		if (behavior != null)
			contextTitle = behavior.getContextTitle(this, _action, _context);

		return contextTitle;
	}

	/**
	 * Gets image associated with this controller. The image is the one associated to the action.
	 *
	 * @return the image corresponding to the action.
	 */
	public String getImage()
	{
		LyActionBehaviorInterface behavior = (LyActionBehaviorInterface)getBehavior();
		String					  image = null;

		if (behavior != null)
			image = behavior.getImage(this, _action);
		if (image == null)
			image = getImage(_action);
		else if (image.equals("_null"))
			image = null;

		return image;
	}

	/**
	 * Gets image tooltip associated with this controller.
	 *
	 * @return the image tooltip corresponding to the action.
	 */
	public String getImageTooltip()
	{
		LyActionBehaviorInterface behavior = (LyActionBehaviorInterface)getBehavior();

		if (behavior != null)
			return behavior.getImageTooltip(this, _action);

		return getTitle();
	}

	/**
	 * Gets menu item image associated with this controller. The image is the one associated to the
	 * action.
	 *
	 * @param  action The action for which the image is needed.
	 * @return the image corresponding to the action.
	 */
	public String getMenuItemImage(LyAction action)
	{
		if (action == null)
			return null;

		LyActionBehaviorInterface behavior = (LyActionBehaviorInterface)getBehavior();
		String					  image = null;

		if (behavior != null)
			image = behavior.getMenuItemImage(this, action);
		if (image == null)
			image = getImage(action, DEFAULT_MENU_ITEM_IMAGE_SIZE, false);
		else if (image.equals("_null"))
			image = null;

		return image;
	}

	/**
	 * Gets help file associated with this controller.
	 *
	 * @return the help file corresponding to the action.
	 */
	public String getHelpFile()
	{
		LyActionBehaviorInterface behavior = (LyActionBehaviorInterface)getBehavior();
		String					  helpFile = null;

		if (behavior != null)
			helpFile = behavior.getHelpFile(this, _action);

		if (helpFile == null)
		{
			// Ask the session
			LySession				   session = getSession();
			LySessionBehaviorInterface sessionBehavior = (LySessionBehaviorInterface)session
					.getSessionBehavior();

			helpFile = sessionBehavior.getHelpFile(session, this, _action);
		}

		if (helpFile == null)
		{
			String		  actionId = _action.getId();
			LyAction	  rootAction = _action.getRoot();
			String		  rootActionId = ((rootAction == null) ? null : rootAction.getId());
			LyClassInfo   classInfo = getClassInfo();
			String		  classId = ((classInfo == null) ? null : classInfo.getId());
			LyEnvironment environment = getEnvironment();

			helpFile = environment.getHelpFile(actionId, rootActionId, classId);
		}

		return helpFile;
	} // end method getHelpFile

	/**
	 * Return list of objects presented in the area This method is overriden depending on the nature
	 * of the controller. Default returns null
	 *
	 * @return the list of objects or null if the area is empty.
	 */
	public LyObjectList getContent()
	{
		return null;
	}

	/**
	 * Return list of selected objects This method is overriden depending on the nature of the
	 * controller. Default returns null
	 *
	 * @return the list of selected objects or null if the selection is empty.
	 */
	public LyObjectList getSelection()
	{
		return null;
	}

    /**
     * Accessor to the reference to the parent controller. This method is added to allow direct
     * access to the _parent object without building the view.
     *
     * @param buildView true if the view must be build or not
     * @return the parent controller instance or null for root controllers.
     */
    public LyController getParent(boolean buildView)
    {
        if (buildView
            && _hasViewController
            && (_component != null)
            && (_viewController == null))
        {
            // To be sure that the view controller was built, because it may reparent
            // this area controller when built.
            getViewController();
        }

        return super.getParent();
    }

    /**
     * Accessor to the reference to the parent controller.
     * Warning ! This method may reparent controller on its first call by building the view
     * controller that contains this area controller.
     * To avoid side effects during construction or freeing caused by behaviors
     * that call this method at a wrong moment, use getParent(false) instead.
     *
     * @return the parent controller instance or null for root controllers.
     */
    public LyController getParent()
    {
        return getParent(true);
    }

	/**
	 * Accessor to the first parent area controller of this controller.
	 *
	 * @return the first parent area controller or null for root controllers.
	 */
	public LyAreaController getParentAreaController()
	{
		LyController parent = this;

		while (parent != null)
		{
			parent = parent.getParent();

			// If parent is an area controller, return it.
			if (parent instanceof LyAreaController)
				return (LyAreaController)parent;
		}

		// If no parent area controller was found, current controller may be a root
		// controller : return null.
		return null;
	}

	/**
	 * Accessor to the multi-area controller managing this area controller (for dialog actions only
	 * in LY_OPEN_DIALOGS_IN_SAME_VIEW mode).
	 *
	 * @return the multi-area controller managing this area controller or null if this area
	 *         controller is not managed by a multi-area controller.
	 */
	public LyMultiAreaController getAreaStackController()
	{
		return _areaStackController;
	}

	/**
	 * Accessor to the view controller of the area. The view controller is a Frame / dialog which
	 * contains this area.
	 *
	 * <P>If no container is set, a new Frame controller is created for this area. Method never
	 * return null, but several calls on this method will return the same reference.</P>
	 *
	 * <P>The default controller is a Frame controller, not a dialog : getViewController(false)</P>
	 *
	 * @return the controller which manages the parent frame / dialog
	 */
	public LyViewController getViewController()
	{
		if (_viewController != null)
			return _viewController;

		if (isFreed())
			return null;

		getBehavior();

		boolean dialog = !_action.hasMark(LyAction.MARK_FRAME_VIEW);

		return getViewController(dialog);
	}

	/**
	 * Accessor to the view controller of the area. The view controller is a Frame / dialog which
	 * contains this area.
	 *
	 * <P>If no container is set, a new Frame controller is created for this area. Method never
	 * return null, but several calls on this method will return the same reference.</P>
	 *
	 * @param  isDialog indicates wether the window is Dialog or Frame
	 * @return the controller which manages the parent frame / dialog
	 */
	public LyViewController getViewController(boolean isDialog)
	{
		if (_viewController != null)
			return _viewController;

		if (!_hasViewController)
			return ((_parent == null) ? null : _parent.getViewController());

		// If session is null, this controller was freed and view controller must not
		// be constructed again.
		LySession session = getSession();

		if (session == null)
			return null;

		LyBuilder builder = getBuilder();
		LyView    view = null;

		// If an XML view file is used, it may contains the view
		// in that case, determines whether a dialog or a frame is defined.
		if (builder instanceof LyFileBuilder)
		{
			LyArea	    area = (LyArea)getComponent();
			LyComponent parent = area.getParent();

			if (parent instanceof LyView)
			{
				view = (LyView)parent;
				isDialog = (view instanceof LyDialog);
			}
		}

		// Create view controller to contain this area controller
		if (isDialog)
			_viewController = new LyDialogController(session, _parent, this);
		else
			_viewController = new LyFrameController(session, _parent, this);

		if (view == null)
		{
			view = _viewController.getView();

			// Patch for Br
			LyEnvironment environment = getEnvironment();

			if (environment.testResource("LY_RESIZABLE_VIEWS"))
				view.setResizable(true);
			else
				view.setResizable(isViewResizable());

			if (view instanceof LyDialog)
			{
				boolean		 modal = _action.hasMark(LyAction.MARK_MODAL);

				((LyDialog)view).setModal(modal);
			}
		}
		else // FFT00139
			_viewController._component = view;

		return _viewController;
	} // end method getViewController

	/**
	 * Accessor to the view controller of the area. The view controller is a Frame / dialog which
	 * contains this area. This method is reserved for internal use and should not be used by the
	 * application.
	 *
	 * @param viewController the controller which manages the parent frame / dialog
	 */
	public void setViewController(LyViewController viewController)
	{
		if ((_viewController != null)
				&& (viewController != null) // XXX
				&& (_viewController != viewController))
			throw new IllegalStateException(
				"BF AppEngine : this area controller has already been added in an other view controller.");

		_viewController = viewController;
	}

	/**
	 * Accessor to the refresh time of this controller. Used by thin clients to know after a
	 * selection whether the controller view must be refreshed or not.
	 *
	 * @return A negative number if view must be refreshed or the refresh date.
	 */
	public long getRefreshTime()
	{
		return -1;
	}

	/**
	 * Indicates if the view for this area controller is resizable.
	 *
	 * @return true if the view is resizable, false otherwise.
	 */
	public boolean isViewResizable()
	{
		return true;
	}

	/**
	 * Indicates if the view for this area acts on the outline view. By default, always return true.
	 *
	 * @return true if this area acts on the outline view, false otherwise.
	 */
	public boolean isOutlined()
	{
		return true;
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */

	public Class<?> getBehaviorInterface()
	{
		return LyActionBehaviorInterface.class;
	}

	/**
	 * Accessor to controller behavior A behavior is associated to a controller to let the
	 * developper specify application logic in the behavior.
	 *
	 * @return associated behavior
	 */

	public LyBehaviorInterface getBehavior()
	{
		LyBehaviorInterface behavior = super.getBehavior();

		if (behavior != null)
			return behavior;

		if (_action == null)
			return null;

		// Sets specific behavior if one is specified for this action
		LyEnvironment environment = getEnvironment();
		Class<?>	  behaviorClass = _action.getBehaviorClass();
		Class<?>	  behaviorInterface = getBehaviorInterface();

		if (behaviorInterface == null)
		{
			// Controller doesn't accept any behavior
			if (behaviorClass != null)
				environment.logError("BF AppEngine : can't associate behavior "
					+ behaviorClass + " to controller " + getClass()
					+ ", because this controller class doesn't have a behavior.");
		}
		else if (behaviorClass == null)
		{
			if (!Modifier.isAbstract(behaviorInterface.getModifiers()))
				// No behavior but behavior is mandatory
				throw new IllegalStateException("BF AppEngine : controller " + getClass()
					+ " requires a behavior that implements " + behaviorInterface
					+ ((_action == null)
						? "" : (". Check the behavior attribute of your action " + _action.getId())));
		}
		else if (!behaviorInterface.isAssignableFrom(behaviorClass))
			// Controller has an invalid behavior
			throw new IllegalStateException("BF AppEngine : invalid behavior "
				+ behaviorClass + " for controller " + getClass()
				+ ", because it doesn't implement " + behaviorInterface
				+ ((_action == null)
					? "" : (". Check the behavior attribute of your action " + _action.getId())));
		else
		{
			if (isClassBehavior(behaviorClass))
			{
				// Try to find if we have a specific class behavior
				LyClassInfo classInfo = getClassInfo();

				if ((classInfo != null)
						&& (classInfo.getBehaviorClass() != null))
					behaviorClass = classInfo.getBehaviorClass();
			}

			try
			{
				behavior = (LyBehaviorInterface)behaviorClass.newInstance();
				super.setBehavior(behavior);
				behavior.init(this);
			}
			catch (Exception e)
			{
				environment.writeException(e);
				environment.logError("BF AppEngine : instanciation of behavior "
					+ behaviorClass + " failed on controller " + getClass()
					+ ((_action == null)
						? "" : (". Check the behavior attribute of your action " + _action.getId())));
			}
		}

		return behavior;
	} // end method getBehavior

	private boolean isClassBehavior(Class<?> behaviorClass)
	{
	    if (behaviorClass == LyClassBehavior.class)
	        return true;

	    String replaceClassName = _environment.getResource("REPLACE-CLASS#" + LyClassBehavior.class.getName());

	    if (behaviorClass.getName().equals(replaceClassName))
	        return true;

	    return false;
	}
	/**
	 * Get the behavior for an action
	 * @param action the action
	 * @return a behavior if available
	 */

	protected LyBehaviorInterface getBehavior(LyAction action)
	{
		if (action == null)
			return null;

		// Sets specific behavior if one is specified for this action
		LyEnvironment environment = getEnvironment();
		Class<?>	  behaviorClass = action.getBehaviorClass();

		if (behaviorClass == null)
			return null;

		LyBehaviorInterface behavior = null;

		if (isClassBehavior(behaviorClass))
		{
			// Try to find if we have a specific class behavior
			LyClassInfo classInfo = getClassInfo();

			if ((classInfo != null)
					&& (classInfo.getBehaviorClass() != null))
				behaviorClass = classInfo.getBehaviorClass();
		}

		try
		{
			behavior = (LyBehaviorInterface)behaviorClass.newInstance();
		}
		catch (Exception e)
		{
			environment.writeException(e);
			environment.logError("BF AppEngine : instanciation of behavior "
				+ behaviorClass + " failed on controller " + getClass());
		}

		return behavior;
	} // end method getBehavior

	/**
	 * Accessor to the current Environment instance This method returns the environment of the
	 * current Action or, if the previous one is null, the session environment.
	 *
	 * @return current environment instance.
	 */
	public LyEnvironment getEnvironment()
	{
		if (_environment != null)
			return _environment;

		LyEnvironment sessionEnvironment = super.getEnvironment();

		if (_action == null)
			_environment = sessionEnvironment;
		else
			_environment = _action.getEnvironment(sessionEnvironment);
		return _environment;
	}

	/**
	 * Accessor to the last modification date of this controller. The modification date is updated
	 * when anything in the view has changed. This parameter is used by view manager to decide
	 * whether the cache could be used or not.
	 *
	 * @return the last modification date (as described in java.util.Date.getTime())
	 */
	public long getLastModified()
	{
		long lastModified = super.getLastModified();

		return ((_lastUpdated > lastModified) ? _lastUpdated : lastModified);
	}

	/**
	 * Sets the preference for this controller.
	 *
	 * @param preference the preference to applu
	 */
	public void setPreference(LyPreference preference)
	{
	}

	/**
	 * Deleting preferences files.
	 *
	 * @param prefs the identifiers of the preferences
	 */
	protected void deletePreferencesFile(String prefs)
	{
	}

	/**
	 * Deleting charts.
	 */
	protected void deleteCharts()
	{
		// Create a dialog to type in the search parameters
		LySession			  session = getSession();
		LyUserInputController chartCtrl;
		LyApplication		  application = session.getApplication();
		LyClassInfo			  classInfo = null;

		if (this instanceof LyGetClassInfoInterface)
			classInfo = ((LyGetClassInfoInterface)this).getClassInfo();

	    LyAction   action = application.getActionWithId(LyAction.ID_DELETE_CHART, classInfo);

		chartCtrl = new LyUserInputController(session, this, action);

		_LyDeleteObjectInfosBehavior	behavior = new _LyDeleteObjectInfosBehavior();

		chartCtrl.setBehavior(behavior);
		behavior.init(chartCtrl);
		chartCtrl.setBuilder(new _LyObjectInfosInputBuilder(chartCtrl));
		chartCtrl.setTitleBar("LY_ACTION_DELETE");
		chartCtrl.setImage(getImage(_action));

		LyInfoList<LyChartInfo> charts = application.getInfoList(LyChartInfo.class);
		LyChartInfo			    chart;
		int					    i, n = ((charts == null) ? 0 : charts.getSize());
		LyOption			    option;
		LyChoice			    choice = chartCtrl.addChoice(DEMAND_DELETE_OBJECTINFOS, null, true,
				true);

		choice.setFill(LyChoice.Fill.VERTICAL);
		choice.setMaxSpan(6);

		if (n == 0)
			return;

		// Sort by name
		charts = charts.getInfosByName();

		for (i = 0; i < n; i++)
		{
			chart = charts.getInfo(i);

			// If chart has no file, it is not possible to delete this chart (chart may
			// be defined in the meta-model)
			if (chart.getFile() == null)
				continue;

			if (!chart.getId().startsWith("_chart_"))
				continue;

			if ((classInfo != null)
					&& (classInfo != chart.getClassInfo()))
				continue;

			option = new LyOption(chart.getId(), chart.getName());
			choice.addOption(option);
		}

		chartCtrl.showView();
		// Waiting for userInput event.
	} // end method deleteCharts

	/**
	 * Deleting object infos files.
	 *
	 * @param infos the identifiers of the object infos
	 */
	protected void deleteObjectInfosFile(String infos)
	{
		LySession     session = getSession();
		LyApplication application = session.getApplication();
		String[]	  ids = LyString.stringToList(infos, LyTools.OBJECT_SEP);
		String		  infoId;
		LyInfo		  info;
		int			  i, n = ids.length;

		for (i = 0; i < n; i++)
		{
			infoId = ids[i];
			info = application.getInfo(infoId);

			if (info instanceof LyExtraInfoInterface)
                application.deleteExtraInfo((LyExtraInfoInterface)info);
		}

		// Notify object infos listeners
		session.notifyInfoListeners(null);
	} // end method deleteObjectInfosFile

	/* Implementation of LyWorkSpaceListener */
	class _LyContextListener
		implements LyGuiListener,
			LyWorkSpaceListener
	{

		/**
		 * Called when an object is added in the dynamic object list.
		 *
		 * @param  event event of adding object
		 * @return the event is used by the listener
		 */
		public boolean objectAdded(LyWorkSpaceEvent event)
		{
			LyObject object = event.getObject();

			if (object == null)
			    return true;

			if (_context.getIndex(object) == 0)
				updateTitleBar();

			return true;
		}

		/**
		 * Called when an object is removed from the dynamic object list.
		 *
		 * @param  event event of removing object
		 * @return the event is used by the listener
		 */
		public boolean objectRemoved(LyWorkSpaceEvent event)
		{
			updateTitleBar();
			return true;
		}

		/**
		 * Called when an object has changed in the dynamic object list.
		 *
		 * @param  event event of changing object
		 * @return the event is used by the listener
		 */
		public boolean objectChanged(LyWorkSpaceEvent event)
		{
			updateTitleBar();
			return true;
		}

		/**
		 * Checks if this controller is idle.
		 *
		 * @return true if the controller is idle.
		 */
		public boolean isIdle()
		{
			return LyAreaController.this.isIdle();
		}
	} // end class _LyContextListener

	/* Implementation of LyInfoListener */

	/**
	 * Invoked when an action occurs on an info object (deletion or creation).
	 *
	 * @param prefix a file prefix
	 */
	public void infoEvent(String prefix)
	{
		if (_builder instanceof LyAreaBuilder)
			((LyAreaBuilder)_builder).updateSelectFile(prefix);
	}

	/**
	 * Get this controller as a string.
	 *
	 * @return this controller as a string
	 */
	public String toString()
	{
		LyClassInfo classInfo = getClassInfo();

		return getClass().toString() + ", action=" + getAction()
		+ ((classInfo != null) ? (", classInfo=" + getClassInfo()) : "");
	}

	/**
	 * Sets idle status of the table.
	 *
	 * @param idle a boolean indicating the idle mode
	 */
	public void setIdle(boolean idle)
	{
		super.setIdle(idle);

		if (!idle)
			// force refresh of view
			reset(false);
	}

	/**
	 * Gets name of an object in the context of the given area controller.
	 *
	 * @param  object Object for which its name is required.
	 * @return Name of the object.
	 */
	public String getName(LyObject object)
	{
		if (object == null)
			return null;

		LyBehaviorInterface behavior = getBehavior();
		String			    name = null;

		if (behavior instanceof LyLabelBehaviorInterface)
		{
			name = ((LyLabelBehaviorInterface)behavior).getName(this, object);
			if (name != null)
				return name;
		}

		LyClassInfo classInfo = object.getClassInfo();
		LyInfoLabel infoLabel = getInfoLabel(classInfo, LyInfoLabel.Type.NAME);

		if (infoLabel != null)
		{
			LyApplication application = getApplication();
			LyLabelDaemon daemon = application.getDaemon(infoLabel);

			if (daemon != null)
				name = daemon.getString(object);
			else
				name = infoLabel.getString(object);
		}

		if (name == null)
			name = object.getName();

		return name;
	} // end method getName

	/**
	 * Gets label of an object, for specified field info and value.
	 *
	 * @param  object     Object for which a label is required.
	 * @param  fieldInfo  field info.
	 * @param  fieldValue value of the field.
	 * @return Label as a string.
	 */
	public String getLabel(LyObject object, LyFieldInfo fieldInfo, Object fieldValue)
	{
		String	label = null;

		if(object != null)
		{
			LyClassInfo classInfo = object.getClassInfo();
			if (!classInfo.contains(fieldInfo))
				return fieldInfo.toName(fieldValue);
		}
		else
			label =  fieldInfo.toName(fieldValue);


		LyBehaviorInterface behavior = getBehavior();

		if (behavior instanceof LyLabelBehaviorInterface)
		{
			label = ((LyLabelBehaviorInterface)behavior).getLabel(this, object, fieldInfo,
					fieldValue);
			if (label != null)
				return label;
		}

		LyInfoLabel infoLabel = getInfoLabel(fieldInfo, LyInfoLabel.Type.VALUE);

		if (infoLabel != null)
		{
			LyApplication application = getApplication();
			LyLabelDaemon daemon = application.getDaemon(infoLabel);

			if (daemon != null)
				label = daemon.getString(object);
			else
				label = infoLabel.getString(object);
		}

		if (label == null)
			label = object.getLabel(fieldInfo, fieldValue);

		return label;
	} // end method getLabel

	/**
	 * Gets short label of an object, for specified field info and value.
	 *
	 * @param  object     Object for which a short label is required.
	 * @param  fieldInfo  field info.
	 * @param  fieldValue value of the field.
	 * @return Short label as a string.
	 */
	public String getShortLabel(LyObject object, LyFieldInfo fieldInfo, Object fieldValue)
	{
		LyBehaviorInterface behavior = getBehavior();
		String			    label = null;

		if (behavior instanceof LyLabelBehaviorInterface)
		{
			label = ((LyLabelBehaviorInterface)behavior).getShortLabel(this, object, fieldInfo,
					fieldValue);
			if (label != null)
				return label;
		}

		LyInfoLabel infoLabel = getInfoLabel(fieldInfo, LyInfoLabel.Type.SHORT_VALUE);

		if (infoLabel != null)
		{
			LyApplication application = getApplication();
			LyLabelDaemon daemon = application.getDaemon(infoLabel);

			if (daemon != null)
				label = daemon.getString(object);
			else
				label = infoLabel.getString(object);
		}

		if (label == null)
			if (object == null)
				return fieldInfo.toShortName(fieldValue);
			else
				label = object.getShortLabel(fieldInfo, fieldValue);

		return label;
	} // end method getShortLabel

	/**
	 * Gets info label of the given type for specified field info.
	 * @param fieldInfo the fieldinfo
	 * @param type  the type of label
	 *
	 * @return The info label if found, null otherwise.
	 */
	public LyInfoLabel getInfoLabel(LyFieldInfo fieldInfo, LyInfoLabel.Type type)
	{
		LyBehaviorInterface behavior = getBehavior();
		LyInfoLabel		    infoLabel = null;

		if (behavior instanceof LyLabelBehaviorInterface)
			infoLabel = ((LyLabelBehaviorInterface)behavior).getInfoLabel(this, fieldInfo, type);
		if (infoLabel == null)
			infoLabel = fieldInfo.getLabel(type);

		return infoLabel;
	}

	/**
	 * Gets info label of the given type for specified class info.
	 * @param classInfo a classinfo
	 * @param type the type of label
	 *
	 * @return The info label if found, null otherwise.
	 */
	public LyInfoLabel getInfoLabel(LyClassInfo classInfo, LyInfoLabel.Type type)
	{
		LyClassInfo actionClassInfo = getClassInfo();

		if ((actionClassInfo != null)
				&& actionClassInfo.isVirtual()
				&& (actionClassInfo.getRealClass() == classInfo))
			// Allow use of labels of virtual classes
			classInfo = actionClassInfo;

		LyBehaviorInterface behavior = getBehavior();
		LyInfoLabel		    infoLabel = null;

		if (behavior instanceof LyLabelBehaviorInterface)
			infoLabel = ((LyLabelBehaviorInterface)behavior).getInfoLabel(this, classInfo, type);
		if (infoLabel == null)
			infoLabel = classInfo.getLabel(type);

		return infoLabel;
	}

	/**
	 * Called when a son view controller has been added to the parent view controller.
	 *
	 * @param son The added son.
	 */
	protected void sonAdded(LyViewController son)
	{

	}

	/**
	 * Called when a son view controller has been removed from the parent view controller.
	 *
	 * @param son The removed son.
	 */
	protected void sonRemoved(LyViewController son)
	{

	}

	/**
	 * Internal class.
	 */
	class _LyObjectInfosInputBuilder extends LyUserInputBuilder
	{
		public _LyObjectInfosInputBuilder()
		{
			this(null);
		}

		public _LyObjectInfosInputBuilder(LyAreaController parent)
		{
			super(parent);
		}

		public LyComponent build()
		{
			LyComponent comp = super.build();

			if (comp != null)
				((LyArea)comp).setScrolling(LyArea.Scrolling.AUTO);

			_form.setBorder(false);
			_form.setAlign(LyForm.Align.LEFT);

			LyConstraint	constraint = new LyConstraint();

			constraint.setFill(LyConstraint.Fill.NONE);
			_form.setConstraint(constraint);

			return comp;
		}
	}


	/**
	 * Internal class.
	 */
	class _LyDeleteObjectInfosBehavior extends LyUserInputBehavior
	{
		/**
		 * Invoked when the user validates the content of the input.
		 *
		 * @param controller the current LyUserInputController instance
		 * @param demand     demand parameters sent by the view component.
		 */
		public void userInputDemand(LyController controller, LyDemand demand)
		{
			// Close find dialog
			String id = demand.getSource();

			if (id.equals(LyFormController.DEMAND_VALIDATE))
			{
				LyParameterValues paramValues = demand.getParameterValues();

				if (paramValues != null)
				{
					String pref = paramValues.getParameter(DEMAND_DELETE_PREFERENCE);
					String infos = paramValues.getParameter(DEMAND_DELETE_OBJECTINFOS);

					if ((pref != null) && (pref.length() != 0))
						deletePreferencesFile(pref);
					else if ((infos != null) && (infos.length() != 0))
						deleteObjectInfosFile(infos);

					if (controller.isFreed())
					    // Controller may be freed when deleting preference
					    return;
				}
			}

			super.userInputDemand(controller, demand);
		}
	} // end class _LyDeleteObjectInfosBehavior

    /**
     * Make the view visible to the user.
     */
    public void showView()
    {
        String      target = (_action == null ? null : _action.getTarget());
        if(getEnvironment().getViewManager().hasProperty(LyMobileViewManager.MOBILE_ONLY_PROPERTY))
        {
        	target = null;
        }
        showView(target);
    }

	/**
	 * Make the view visible to the user.
	 *
	 * @param target the name of the target where the view must be shown
	 */
	public void showView(String target)
	{
		// If controller is freed and is displayed in an area stack controller (may happen
		// when area controller is being initialized), there is no need to show the view.
		if ((_areaStackController != null) && isFreed())
			return;

		if (isFreed())
			throw new IllegalStateException("BF AppEngine : attempt to show controller "
				+ this + " that was freed");

		_target = target;

		if (target != null)
		{
			short   position = -1;
			boolean borderPanel = false;

			if (target.equals("_north"))
			{
				position = LyArea.NORTH;
				borderPanel = true;
			}
			else if (target.equals("_south"))
			{
				position = LyArea.SOUTH;
				borderPanel = true;
			}
			else if (target.equals("_east"))
			{
				position = LyArea.EAST;
				borderPanel = true;
			}
			else if (target.equals("_west"))
			{
				position = LyArea.WEST;
				borderPanel = true;
			}
			else
			{
				try
				{
					position = Short.parseShort(target);
				}
				catch (NumberFormatException e)
				{
					// Not a known target
				}
			}

			if (borderPanel || (position >= 0))
			{
				_hasViewController = false;

				LyArea area = (LyArea)getComponent();

				if (!_initialized)
					initArea();

				LyAreaController parent = getParentAreaController();

				if (borderPanel)
				{
					((LyArea)parent.getComponent()).setBorderPanel(area, position);
					handleFocus();
					parent.setLastModified(System.currentTimeMillis());
					return;
				}
				else
				{
					((LyArea)parent.getComponent()).insertPanel(area, position);
					parent.setLastModified(System.currentTimeMillis());
				}
			}
		}

		if (!_hasViewController)
			// Force hasViewController attribute to true since a showView is requested directly on this controller
			_hasViewController = true;

		super.showView();
	} // end method showView

	/**
	 * Hide the view. The effect of this command depends on the nature of the application. If the
	 * display is a fat client for example (windows (awt) / swing) this methods should close the
	 * window or hide the view.
	 *
	 * <P>This method should consider that the page is closed and should disappear.</P>
	 *
	 * @param free Indicates if only HIDDEN, not FREED (true).
	 */
	public void hideView(boolean free)
	{
		if (_target == null)
			super.hideView(free);
		else
		{
			LyArea area = (LyArea)getComponent();
			short  position;

			if (_target.equals("_north"))
				position = LyArea.NORTH;
			else if (_target.equals("_south"))
				position = LyArea.SOUTH;
			else if (_target.equals("_east"))
				position = LyArea.EAST;
			else if (_target.equals("_west"))
				position = LyArea.WEST;
			else
				position = -1;

			LyAreaController parent = getParentAreaController();

			if (parent.getComponent() != null)
			{
				if (position != -1)
					((LyArea)parent.getComponent()).setBorderPanel(null, position);
				else
					((LyArea)parent.getComponent()).removePanel(area);
			}

			parent.setLastModified(System.currentTimeMillis());

			// Clear view cache if needed
			if (free)
				clearViewCache();
		}
	} // end method hideView

	/**
	 * Check if the controller has a viewController.
	 *
	 * @return boolean if the controller has a viewController
	 */
	public boolean hasViewController()
	{
		return _hasViewController;
	}

	/**
	 * Invoked when the user validates the input content.
	 *
	 * @param controller the current LySelectFileController instance
	 * @param demand     demand parameters sent by the view component.
	 */
	public void selectFileDemand(LySelectFileController controller, LyDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			// Gets current action (source of the demand)
			String              id = demand.getSource();

			if (id.equals(LyFormController.DEMAND_VALIDATE) || (id.equals(LyFormController.DEMAND_APPLY)))
            {
				if ("LY_DUMP_VIEW_WITH_CONTENT".equals(controller.getUserData()))
					dumpView(true, controller.getSelectedPath());
				else if  ("LY_DUMP_VIEW_WITHOUT_CONTENT".equals(controller.getUserData()))
					dumpView(false, controller.getSelectedPath());
				else if  ("LY_SNAPSHOT".equals(controller.getUserData()))
					snapshot(controller.getSelectedPath());
            }
		}
	}

    /**
     * Gives the focus to the first editable field in the form.
     */
    protected void handleFocus()
    {
        Vector<LyFormField> result = new Vector<LyFormField>();
        LyArea              area = (LyArea)_component;

        if ((area != null) && (area.getPanelCount() > 0))
        {
            LyPanel panel = area.getPanel(0);

            panel.getComponents(LyFormField.class, result);
        }

        int         i, n = result.size();
        LyFormField field;

        for (i = 0; i < n; i++)
        {
            field = result.elementAt(i);
            if (field.isEditable() && field.isEnabled())
            {
                field.setFocus();
                break;
            }
        }
    }

    /**
     * Remove a son from this controller.
     *
     * @param controller son area controller to remove
     */
    protected void removeSon(LyAreaController controller)
    {
        // Nothing to do
    }

    /**
     * Checks if a mark is set for the given fieldInfo in the context
     * of the controlled action.
     * @param fieldInfo Field to be checked.
     * @param mark Mark to be checked.
     * @return True if the given field has the given mark, false otherwise.
     */
    public boolean hasMark(LyFieldInfo fieldInfo, int mark)
    {
        String          markName = LyFieldInfo.getMarkName(mark);
        String          actionMarkName;
        LyController    controller = this;
        LyAction        action;

        while (controller instanceof LyAreaController)
        {
            action = ((LyAreaController)controller).getAction();

            while (action != null)
            {
                actionMarkName = "__" + action.getId() + '.' + markName;

                if (fieldInfo.hasMark(actionMarkName))
                    return true;
                if (fieldInfo.hasMark(actionMarkName + "=false"))
                    return false;

                action = (LyAction)action.getTemplate();
            }

            controller = controller._parent;
            if (controller instanceof LyFieldController)
                controller = ((LyFieldController)controller).getFormController();
        }

        return fieldInfo.hasMark(mark);
    }

    /**
     * Checks if a mark is unset for the given fieldInfo in the context
     * of the controlled action.
     * @param fieldInfo Field to be checked.
     * @param mark Mark to be checked.
     * @return True if the given field has the given mark unset, false otherwise.
     */
    public boolean hasNotMark(LyFieldInfo fieldInfo, int mark)
    {
        String     markName = "__" + _action.getId()
            + "." + LyFieldInfo.getMarkName(mark) + "=false";

        if (fieldInfo.hasMark(markName))
            return true;

        return false;
    }

    /**
     * Retrieves a value from a registered key.
     *
     * @param   key  The key that must be used.
     * @return  The value stored with the key or null if the key is not registered.
     */
    public Object getKeyValue(String key)
    {
        Object              value = null;
        LyBehaviorInterface behavior = getBehavior();

        if (behavior instanceof LyActionBehaviorInterface)
        {
            value = ((LyActionBehaviorInterface)behavior).getKeyValue(key);
            if (value != null)
                return value;
        }

        if (_environment != null)
        {
            value = _environment.getResource(key);
            if (value != null)
                return value;
        }

        return super.getKeyValue(key);
    }

    /**
     * Show given file.
     * @param file File name.
     */
    protected void showFile(String file)
    {
        if (file == null)
            return;

        String  externalViewer = _environment.getFile("LY_EXTERNAL_VIEWER");

        if ((externalViewer != null)
            && !externalViewer.equals("LY_EXTERNAL_VIEWER"))
        {
            LyViewManager   viewManager = getViewManager();

            viewManager.showLocation(null, file, true);
        }
        else
        {
            LyApplication       application = getApplication();
            LyAction            action = application.getAction("_text");
            LyTextController    textController = new LyTextController(_session, this, action, file, 0);

            textController.showView();
        }

    }
} // end class LyAreaController
