/**
 * LyCompoundAreaController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyCompoundAreaController.java 34904 2012-03-21 15:55:41Z bangiole $
 */

package leon.control;

import java.util.*;

import leon.app.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.view.*;
import leon.view.LySplitter.ActivePart;
import leon.view.event.*;


/**
 * A compound area controller is associated to a compound action.
 *
 * <P>This controller is a view which gather two views controllers (subcontrollers) in the same
 * 'area'.</P>
 *
 * <P>The compoundAreaController provides the following services :</P>
 *
 * <UL>
 *   <LI>Management of the list of sub-controllers (initialize, free)</LI>
 *   <LI>context management between sub-controllers (views)</LI>
 * </UL>
 */
public class LyCompoundAreaController extends LyMultiAreaController
{
	/**
	 * File version
	 */
	public static final String					    __VERSION =
		"$Id: LyCompoundAreaController.java 34904 2012-03-21 15:55:41Z bangiole $";

	/** Hashtable of actions and controllers. */
	protected Hashtable<LyAction, LyAreaController> _controllersTable;

	/** Delete controller when action is set. */
	protected Boolean[]							    _deleteOnClose;

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. The action should be an instance of LyCompoundAction. If an
	 *                  action is specified, the controller uses action parameters (action name,
	 *                  action specific XML view description ...)
	 * @param classInfo classInfo : useless in this controller, should be null. Kept for
	 *                  compatibility.
	 * @param context   current context (may be null if the view is not contextual to a list of
	 *                  objects).
	 */
	public LyCompoundAreaController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent, action, null);

		_deleteOnClose = new Boolean[2];

		buildSubControllers(classInfo, context);
	}

	/**
	 * Free this controller. Current implementation overrides generic method to free table of
	 * controllers.
	 *
	 * @param controller controller instance which requested the freeing of this controller. If the
	 *                   controller is the parent, this controller is freed otherwise the method
	 *                   does nothing.
	 */
	public void free(LyController controller)
	{
		if (controller != _parent)
			return;

		if (_controllersTable != null)
		{
			LyAreaController areaController;
			Enumeration<LyAreaController>	controllers = _controllersTable.elements();

			// Free area controllers
			while (controllers.hasMoreElements())
			{
				areaController = controllers.nextElement();
				areaController.free(this);
			}
		}

		_controllersTable = null;

		super.free(controller);
	}

	/**
	 * Create the list of sub-controllers (content of the compound view). The methods retreives the
	 * information from the action and create sub-controllers from the list of sub-actions.
	 *
	 * <P>Should never be used directly, invoked by the constructor.</P>
	 * @param classInfo a classInfo
	 * @param context the context
	 */
	protected void buildSubControllers(LyClassInfo classInfo, LyObjectList context)
	{
		if (_action == null) // This occurs for a Xref action
			return;

		if (!(_action instanceof LyCompoundAction))
			throw new IllegalArgumentException("BF AppEngine : invalid action type "
				+ _action + " (expected action is leon.info.LyCompoundAction)");

		LyCompoundAction compoundAction = (LyCompoundAction)_action;
		LyAction		 subAction;
		int				 i, n = compoundAction.getSubActionCount();

		// Check if all actions are viewable.
		for (i = 0; i < n; i++)
		{
			subAction = compoundAction.getSubAction(i);
			if (!subAction.isViewable())
				throw new IllegalArgumentException("BF AppEngine : invalid action " + subAction
					+ " used in compound action " + _action
					+ ", only viewable actions can be used in a compound action");
		}

		LySession session = getSession();

		if (compoundAction.getStyle() == LyCompoundAction.Style.GRID)
		{
			LyAreaController areaController;

			_buildInternalSubControllers = true;

			for (i = 0; i < n; i++)
			{
				subAction = compoundAction.getSubAction(i);
				if (!session.validateAction(subAction, this))
					continue;

				areaController = (LyAreaController)session.invokeAction(this, subAction, classInfo,
						context, false);
				addAreaController(areaController);
			}

			_buildInternalSubControllers = false;
		}
		else
		{
			LyAction		 firstAction = compoundAction.getFirstAction();
			LyAction		 secondAction = compoundAction.getSecondAction();
			LyAreaController firstAreaController = null, secondAreaController = null;
			LyObjectList     selection = null;

			_buildInternalSubControllers = true;

			if (((LyCompoundAction)_action).getContextAction() == firstAction)
			{
				if (session.validateAction(firstAction, this))
					firstAreaController = (LyAreaController)session.invokeAction(this, firstAction,
							classInfo, context, false);

				// Get selection
				if (firstAreaController != null)
					selection = firstAreaController.getSelection();

				if ((selection == null) || (selection.getSize() == 0))
					if (secondAction.getMin() == 0)
						selection = new LySimpleObjectList();
					else if (firstAreaController != null)
						selection = firstAreaController.getContent();

				if (session.validateAction(secondAction, this))
					secondAreaController = (LyAreaController)session.invokeAction(this,
							secondAction, classInfo, selection, false);
			}
			else if (((LyCompoundAction)_action).getContextAction() == secondAction)
			{
				if (session.validateAction(secondAction, this))
					secondAreaController = (LyAreaController)session.invokeAction(this,
							secondAction, classInfo, context, false);

				// Get selection
				if (secondAreaController != null)
					selection = secondAreaController.getSelection();

				if ((selection == null) || (selection.getSize() == 0))
					if (firstAction.getMin() == 0)
						selection = new LySimpleObjectList();
					else if (secondAreaController != null)
						selection = secondAreaController.getContent();

				if (session.validateAction(firstAction, this))
					firstAreaController = (LyAreaController)session.invokeAction(this, firstAction,
							classInfo, selection, false);
			}
			else
			{
				if (session.validateAction(firstAction, this))
					firstAreaController = (LyAreaController)session.invokeAction(this, firstAction,
							classInfo, context, false);
				if (session.validateAction(secondAction, this))
					secondAreaController = (LyAreaController)session.invokeAction(this,
							secondAction, classInfo, context, false);
			}

			_buildInternalSubControllers = false;

			if (firstAreaController != null)
				addAreaController(firstAreaController);
			if (secondAreaController != null)
				addAreaController(secondAreaController);

			if ((firstAreaController != null) && (secondAreaController != null))
			{
				if (((LyCompoundAction)_action).getContextAction() == firstAction)
					firstAreaController.setContextClient(secondAreaController);
				else if (((LyCompoundAction)_action).getContextAction() == secondAction)
					secondAreaController.setContextClient(firstAreaController);
			}
		}
	} // end method buildSubControllers

	/**
	 * Prepare component before showing. This method is called once (unlike reset which may be
	 * called several times), before first showing. Overrides parent method to propagate the
	 * initialization on sub-controllers
	 */
	protected void initArea()
	{
		if (_initialized)
			return;

		LyAreaController areaController;
		int				 i, n = getAreaControllersCount();
		LyAction		 contextAction = null;

		if (getAction() instanceof LyCompoundAction)
		{
			LyCompoundAction action = (LyCompoundAction)getAction();

			contextAction = action.getContextAction();
		}

		for (i = 0; i < n; i++)
		{
			areaController = getAreaController(i);
			if ((contextAction == null)
					|| (areaController.getAction() == contextAction))
            {
				areaController.initArea();
                if (areaController.isFreed())
                {
                    removeAreaController(areaController);
                    i--;
                    n--;
                }
            }
		}

		if (contextAction != null)
		{
			for (i = 0; i < n; i++)
			{
				areaController = getAreaController(i);
				if (areaController.getAction() != contextAction)
				{
					areaController.initArea();
                    if (areaController.isFreed())
                    {
                        removeAreaController(areaController);
                        i--;
                        n--;
                    }
                }
			}
		}

		// Start action on first access
		if (!startAction())
			free();

		_initialized = true;
	} // end method initArea

	/**
	 * Create the default builder for this controller. Current implementation overrides generic
	 * method and returns an instance of leon.control.builder.LyCompoundAreaBuilder
	 *
	 * @return a new builder instance.
	 * @see    leon.control.LyController#getDefaultBuilder
	 * @see    leon.control.builder.LyCompoundAreaBuilder
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyCompoundAreaBuilder(this);
	}

	/**
	 * Reset the entire view : everything is rebuilt.
	 *
	 * @param full Boolean indicating if a full reset (including reset of datas) is requiered or not
	 *             (only reset graphical area).
	 */
	public void reset(boolean full)
	{
		updateTitleBar();
		super.reset(full);

		if (getAction() instanceof LyCompoundAction)
		{
			LyCompoundAction compoundAction = (LyCompoundAction)getAction();
			LyAction		 contextAction = compoundAction.getContextAction();

			if (contextAction != null)
			{
				short contextIndex = compoundAction.getContextIndex();

				getAreaController(contextIndex).reset(full);
			}
			else
			{
				int i, n = _areaControllers.size();

				for (i = 0; i < n; i++)
					getAreaController(i).reset(full);
			}
		}
	} // end method reset

	/**
	 * Set current controller context. This method set the context of the sub AreaController Two
	 * possibilities: One of the subareas is a context for the other one, so this method set its
	 * context. Both Areas are &quot;equal&quot; so the context is set for both.
	 *
	 * @param context context or null if none is specified
	 * @return true if successfull
	 * @see   LyContextInterface#setContext
	 */
	public boolean setContext(LyWorkSpace context)
	{
		if (getAction() instanceof LyCompoundAction)
		{
			LyCompoundAction compoundAction = (LyCompoundAction)getAction();

			if (compoundAction.getContextAction() != null)
			{
				short			 contextIndex = compoundAction.getContextIndex();
				LyAreaController areaController = getAreaController(contextIndex);

				if (areaController != null)
					return areaController.setContext(context);
			}
		}

		return super.setContext(context);
	}

	/**
	 * Accessor to the area controller which is the client context for the other area controller.
	 *
	 * @return the area controller whose context is defined by the other area controller.
	 */
	public LyAreaController getContextClient()
	{
		if (_action instanceof LyCompoundAction)
		{
	        LyCompoundAction compoundAction = (LyCompoundAction)_action;
	        LyAction         firstAction = compoundAction.getFirstAction();
	        LyAction         secondAction = compoundAction.getSecondAction();

	        // If context is given by first action, second action is the client context action
			if ((((LyCompoundAction)_action).getContextAction() == firstAction)
					&& (getAreaControllersCount() > 1))
				return getAreaController(1);
	        // If context is given by second action, first action is the client context action
			else if ((((LyCompoundAction)_action).getContextAction() == secondAction)
					&& (getAreaControllersCount() > 0))
				return getAreaController(0);
		}

		return null;
	}

	/**
	 * Accessor to set another area controller which will receive contextual notification when
	 * something occur on this area controller.
	 *
	 * <P>Method is overrident to take the internal sub-controller structure into account.</P>
	 *
	 * @param controller the contextual controller, or null if no client defined
	 */
	public void setContextClient(LyAreaController controller)
	{
		LyCompoundAction                  compoundAction = (LyCompoundAction)getAction();
        LyCompoundAction.ClientContext	  clientContext = compoundAction.getClientContext();

		if (clientContext == LyCompoundAction.ClientContext.NONE)
			return;

		LyAreaController areaController;

		if ((clientContext == LyCompoundAction.ClientContext.FIRST)
				|| (clientContext == LyCompoundAction.ClientContext.SECOND))
		{
            short   index = compoundAction.getClientContextIndex();

			areaController = getAreaController(index);
			areaController.setContextClient(controller);
		}
		else
		{
            LyCompoundAction.Context context = compoundAction.getContext();

			if (context == LyCompoundAction.Context.FIRST)
			{
				areaController = getAreaController(1);
				if (areaController == null)
					areaController = getAreaController(0);
				if (areaController != null)
					areaController.setContextClient(controller);
			}
			else if (context == LyCompoundAction.Context.SECOND)
			{
				areaController = getAreaController(0);
				if (areaController != null)
					areaController.setContextClient(controller);
			}
			else
				super.setContextClient(controller);
		}
	} // end method setContextClient

	/**
	 * Change client context action (or second action if no action is used for context).
	 *
	 * @param action        New action added in compound area.
	 * @param deleteOnClose Delete controller when action is changed.
	 */
	public void setAction(LyAction action, boolean deleteOnClose)
	{
		LyCompoundAction          compoundAction = (LyCompoundAction)getAction();
        LyCompoundAction.Context  context = compoundAction.getContext();
		short			          index;

		if (context == LyCompoundAction.Context.SECOND)
			index = 0;
		else
			index = 1;

		setAction(action, index, deleteOnClose);
	}

	/**
	 * Change client context action (or second action if no action is used for context).
	 *
	 * @param      action New action added in compound area.
	 * @deprecated Use setAction(LyAction, boolean)
	 */
	public final void setAction(LyAction action)
	{
		setAction(action, false);
	}

	/**
	 * Change action at the given position.
	 *
	 * @param action        New action added in compound area.
	 * @param context       The context of the new action.
	 * @param index         Index of the replaced action.
	 * @param deleteOnClose Delete controller when action is changed.
	 */
	public void setAction(LyAction action, LyObjectList context, short index, boolean deleteOnClose)
	{
		if ((index < 0) || (index >= getAreaControllersCount()))
			throw new IllegalStateException("BF AppEngine : invalid index parameter " + index);

		if ((action == null) || !action.isViewable())
			throw new IllegalStateException("BF AppEngine : invalid action parameter " + action
				+ ", given action is not viewable");

		LyCompoundAction compoundAction = (LyCompoundAction)getAction();

		if (_controllersTable == null)
		{
			int		 i, n = compoundAction.getSubActionCount();
			LyAction subAction;

			_controllersTable = new Hashtable<LyAction, LyAreaController>(n);
			for (i = 0; i < n; i++)
			{
				subAction = compoundAction.getSubAction(i);
				_controllersTable.put(subAction, getAreaController(i));
			}
		}

		// Replace area controller
		LySession		 session = getSession();
		LyAreaController contextAreaController = null;
		LyAreaController oldAreaController = getAreaController(index);
		if (context == null)
			context = getContext();
		short			 contextIndex = compoundAction.getContextIndex();

		if ((index != contextIndex)
		    && (compoundAction.getContext() != LyCompoundAction.Context.NONE))
		{
			contextAreaController = getAreaController(contextIndex);
			context = contextAreaController.getSelection();
		}

		LyAreaController newAreaController = _controllersTable.get(action);

		if ((oldAreaController == newAreaController)
				&& !newAreaController.isFreed())
			return;

		if ((newAreaController == null)
				|| newAreaController.isFreed())
		{
			_buildInternalSubControllers = true;
			newAreaController = (LyAreaController)session.invokeAction(this, action, null, context,
					false);
			_buildInternalSubControllers = false;

			newAreaController.setParent(this);
			newAreaController.setViewController(_viewController);
			_controllersTable.put(action, newAreaController);
		}
		else
			newAreaController.setIdle(false);

		_areaControllers.setElementAt(newAreaController, index);

		// Replace view
		LyArea newArea = (LyArea)newAreaController.getComponent();
		LyArea oldArea = (LyArea)oldAreaController.getComponent();
		LyArea parentArea = (LyArea)oldArea.getParent();

		parentArea.removePanel(oldArea);
		parentArea.addPanel(newArea);

		// Update context dependencies
		if (contextAreaController != null)
			contextAreaController.setContextClient(newAreaController);
		else if (index == contextIndex)
		{
			LyAreaController clientAreaController;

			if (compoundAction.getContext() == LyCompoundAction.Context.FIRST)
				clientAreaController = getAreaController(1);
			else
				clientAreaController = getAreaController(0);

			newAreaController.setContextClient(clientAreaController);
			newAreaController.updateContextClient();
		}

		// This method must be called after the complete initialization of
		// the newAreaController (include the context).
		newAreaController.initArea();

		// Clear old area controller
		boolean delete = ((_deleteOnClose[index] != null) ? _deleteOnClose[index].booleanValue()
														  : deleteOnClose);

		if (delete)
		{
			LyAction oldAction = oldAreaController.getAction();

			if ((oldAction != null)
					&& (_controllersTable.get(oldAction) == oldAreaController))
				_controllersTable.remove(oldAction);
			oldAreaController.free(this);
		}
		else
			oldAreaController.setIdle(true);

		_deleteOnClose[index] = Boolean.valueOf(deleteOnClose);

		// Update view
		updateMenus();
		updateMessageBar();

		// Update the last modification time.
		setLastModified(new Date().getTime());
	}
	/**
	 * Change action at the given position.
	 *
	 * @param action        New action added in compound area.
	 * @param index         Index of the replaced action.
	 * @param deleteOnClose Delete controller when action is changed.
	 */
	public void setAction(LyAction action, short index, boolean deleteOnClose)
	{
		setAction(action, null, index, deleteOnClose);
	} // end method setAction

	/**
	 * Change action at the given position.
	 *
	 * @param      action New action added in compound area.
	 * @param      index  Index of the replaced action.
	 * @deprecated Use setAction(LyAction, short, boolean)
	 */
	public final void setAction(LyAction action, short index)
	{
		setAction(action, index, false);
	}

	/**
	 * Invoked when an action occurs. At this level, the implementation manages
	 *
	 * <UL>
	 *   <LI>HREF demands. (demand.getType() = LyActionDemand.TYPE_HREF)</LI>
	 *   <LI>HREF demands. (demand.getType() = LyActionDemand.TYPE_ACTION AND demand.getSource() =
	 *     DEMAND_PREVIOUS_OBJECT or DEMAND_NEXT_OBJECT or DEMAND_SELECTION</LI>
	 *   <LI>User interaction on action bar.</LI>
	 * </UL>
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		String sourceId = demand.getSource().intern();
		LyComponent component = getComponent();
		Vector<LySplitter> vector = new Vector<LySplitter>();
		component.getComponents(LySplitter.class, vector);
		if (sourceId.equals("_splitter"))
		{
			LySplitter split = vector.firstElement();
			ActivePart select = split.getActivePart();
			if(select == ActivePart.FIRST)
			{
				split.setActivePart(ActivePart.SECOND);
			}
			else if(select == ActivePart.SECOND)
			{
				split.setActivePart(ActivePart.FIRST);
			}
		}

		super.actionDemand(demand);
	} // end method actionDemand
} // end class LyCompoundAreaController
