/**
 * LyCompoundFilterController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyCompoundFilterController.java 35456 2012-05-09 08:51:22Z nhamon $
 */
package leon.control;

import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;

import java.util.*;


/**
 * This controller implements a compound filter.
 * It is based on a coumpound area.
 *
 * @see LyCompoundAreaController
 */
public class LyCompoundFilterController extends LyCompoundAreaController
{
	/**
	 * File version
	 */
	public static final String __VERSION = "$Id: LyCompoundFilterController.java 35456 2012-05-09 08:51:22Z nhamon $";

	/** Demand id for adding a filter element. */
	public static final String DEMAND_ADD_FILTER = "_addFilter";

	/** Identifier of an input field : operation. */
	public static final String SELECT_TYPE = "_filterType";

	/** Id for operation raz. */
	public static final String ID_SIMPLE_FILTER = "_simpleFilter";

	/** Id for operation add. */
	public static final String		   ID_EXTENDED_FILTER = "_extendedFilter";
	
	/**
	 * The simple filter controller
	 */
	protected LySimpleFilterController _filterController;
	
	/**
	 * The filter editor controller
	 */
	protected LyFilterEditorController _filterEditorController;
	
	/**
	 * Extended filter state
	 */
	protected boolean				   _extendedFilter;

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo classInfo associated with this controller
	 * @param context   current context (may be null if the view is not contextual to a list of
	 *                  objects).
	 */
	public LyCompoundFilterController(LySession session, LyController parent, LyAction action, LyClassInfo classInfo,
		LyObjectList context)
	{
		super(session, parent, action, classInfo, context);
        
		_filterController = (LyFilterController)getAreaController(0);
		_filterEditorController = (LyFilterEditorController)getAreaController(1);
		_filterEditorController.updateFilterController();
        
		if (parent instanceof LyFilterEndInterface)
		{
			LyFilter filter = (action.getId().endsWith("New") ? null : ((LyFilterEndInterface)parent).getFilter());
			int		 nf = ((filter == null) ? 0 : filter.getSize());

			_filterEditorController.setFilterEndInterface((LyFilterEndInterface)parent);
			_extendedFilter = ((nf > 1)
			    || ((nf > 0) && !(filter.getElement(0) instanceof LyFilterExpression)));
		}
		else
			_extendedFilter = false;

		_filterController.setFilterEndInterface(_filterEditorController);
		if (_extendedFilter)
			openFilterEditor();
		else
			closeFilterEditor();
	}

	/**
	 * Checks whether extended filter is activated or not.
	 * @return True if extended filter is on, false if it's off.
	 */
	public boolean isExtendedFilter()
	{
		return _extendedFilter;
	}

	/**
	 * Change filter panel.
	 * @param filterController the filter controller to set 
	 * @return the final filter controller
	 */
	protected LySimpleFilterController setFilterController(LySimpleFilterController filterController)
	{
		// Replace area controller
		LyAreaController oldFilterController = getAreaController(0);

		_filterController = filterController;
		_filterController.setFilterEndInterface(_filterEditorController);
		_filterController.setParent(this);
		_filterController.setViewController(_viewController);
		_areaControllers.setElementAt(_filterController, 0);

		// Replace view
		LyArea newArea = (LyArea)_filterController.getComponent();
		LyArea oldArea = (LyArea)oldFilterController.getComponent();
		LyArea parentArea = (LyArea)oldArea.getParent();

		parentArea.removePanel(oldArea);
		parentArea.addPanel(newArea);

		// This method must be called after the complete initialization of
		// the newAreaController (include the context).
		_filterController.initArea();

		// Clear old area controller
		oldFilterController.free(this);

        // For Web mode, associate area to filter controller (not necessary otherwise)
        LyArea      area = (LyArea)getComponent();
        
        area.setController(_filterController.getId());
        
		// Update the last modification time.
		setLastModified(new Date().getTime());
		return _filterController;
	} // end method setFilterAction

	/**
	 * Gets controller of the filter element panel.
	 * @return First controller.
	 */
	public LySimpleFilterController getFilterController()
	{
		return _filterController;
	}

	/**
	 * Gets controller of the filter editor panel.
	 * @return Second controller.
	 */
	public LyFilterEditorController getFilterEditorController()
	{
		return _filterEditorController;
	}

	/**
	 * Invoked when an action notification is sent to this listener.
	 *
	 * <P>method implements VALIDATE, APPLY and CANCEL action events and subsequently invoke
	 * endFilter on the client.</P>
	 *
	 * <P>Additionally, the controller handles field changes and impact on form content. (The nature
	 * of components shown in the view is field dependant).</P>
	 *
	 * @param demand demand parameters sent by the view component.
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			// Menu or action bar
			String id = demand.getSource();

			if (id.equals(DEMAND_ADD_FILTER))
			{
				_filterController.validate(demand);
				openFilterEditor();
				return;
			}
			if (id.equals(LyFormController.DEMAND_CANCEL))
			{
				hideView();
				free(_parent);
				return;
			}

			boolean validate = id.equals(LyFormController.DEMAND_VALIDATE);

			if (!_extendedFilter || _filterController.checkIfModified())
				_filterController.validate(demand);

			if (validate && !_extendedFilter)
			{
				LyFilter	filter = ((LyFilterController)_filterController).buildFilter(demand);
				
				_filterEditorController.setFilter(filter);
			}
			
			_filterEditorController.actionDemand(demand);
			
			if (validate)
			{
				hideView();
				free(_parent);
			}

			return;
		}
		else if (demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
		{
			if (demand.getValue().equals(ID_SIMPLE_FILTER))
				closeFilterEditor();
			else
				openFilterEditor();

			return;
		}

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Open the filter editor
	 */
	protected void openFilterEditor()
	{
		LyEnvironment environment = getEnvironment();

		if (!environment.testResource("LY_EXTENDED_FILTER"))
			return;

		if (_extendedFilter && isInitialized())
			return; // Already open

		// Show second panel
		LyArea     area = (LyArea)getComponent();
		
		LyPanel	   firstPanel = area.getPanel(0);
		LySplitter splitter;
		LyArea     firstArea;
		
		if (firstPanel instanceof LySplitter)
		{
			splitter = (LySplitter)firstPanel;
			firstArea = (LyArea)splitter.getPanel(0);
		}
		else
			firstArea = (LyArea)firstPanel;
		
		LyArea     secondPanel = (LyArea)_filterEditorController.getComponent();
		LyConstraint  constraint = new LyConstraint();
		
		constraint.setFill(LyConstraint.Fill.BOTH);
		area.removePanel(firstPanel);
		splitter = new LySplitter();
		splitter.setConstraint(constraint);
		splitter.setResizable(true);
		splitter.setSizes("60 40");
		splitter.addPanel(firstArea);
		splitter.addPanel(secondPanel);
		area.addPanel(splitter);
		LyCommands commands = area.getCommands();
		LyCommand  command = commands.getCommand(0);

		command.setEnabled(true);
		_extendedFilter = true;
	}

	/**
	 * Close the filter editor
	 */
	protected void closeFilterEditor()
	{
		if (!_extendedFilter && isInitialized())
			return; // Already close

		// Hide second panel
		LyArea     area = (LyArea)getComponent();
		LySplitter splitter = (LySplitter)area.getPanel(0);
		LyArea     firstPanel = (LyArea)splitter.getPanel(0);
		LyConstraint  constraint = new LyConstraint();
		
		constraint.setFill(LyConstraint.Fill.BOTH);
		area.removePanel(splitter);
		area.addPanel(firstPanel);

		// Disable "Add" button
		LyCommands commands = area.getCommands();
		LyCommand  command = commands.getCommand(0);

		command.setEnabled(false);
		_extendedFilter = false;

		if ((_filterController instanceof LyFilterRelationController) && (_filterEditorController != null))
		{
			// If closing filter editor and relation filter was edited : reset filter controller
			LyAction	  action = (LyAction)getApplication().getInfo("__filter_item");

			_filterEditorController.createFilterController(action, null, _filterEditorController.getClassInfo());
		}
	}
}
