/**
 * LyConsultController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyConsultController.java 41807 2012-12-18 17:54:46Z smaupin $
 */

package leon.control;

import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.control.form.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * This class describes the controller associated to a consultation form. This class implements a
 * 'read-only' LyFormController and specializes method from its parent.
 *
 * <P>This form controller is always associated to an object or a list of objects.</P>
 *
 * <P>This class uses a LyFormBehavior (LyClassBehavior) instance which may implement specific
 * application logic, automatic field calculation... or dynamic impact on value changes...</P>
 *
 * @see LySetController
 * @see LyCreateController
 * @see leon.app.behaviorinterface.LyFormBehaviorInterface
 * @see leon.app.behavior.LyClassBehavior
 */
public class LyConsultController extends LyFormController
	implements LyContextInterface, LyControllerContextInterface
{
	/**
	 * File version
	 */
	public static final String     __VERSION =
		"$Id: LyConsultController.java 41807 2012-12-18 17:54:46Z smaupin $";

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo This parameter is not used. Kept from introspection compatibility. The
	 *                  classInfo is always the classInfo of the specified objects.
	 * @param objects   current list of objects being shown in the consultation form. One Object is
	 *                  shown, but the user may navigate between the object list through the GUI.
	 */
	public LyConsultController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		super(session, parent, action, classInfo, objects);

		// Try to retrieve context from the closest parent
		LyWorkSpace context;

		while (parent != null)
		{
			if (parent instanceof LyContextInterface)
			{
				context = ((LyContextInterface)parent).getContext();
				if ((context != null) && context.equals(_objects))
					context = null;
				initContext(context);
				break;
			}

			parent = parent.getParent();
		}

		if ((_objects == null)
				&& (parent != null)
				&& (objects == null)
				&& (action.getMin() == 0)
				&& (action.getMax() == -1)
				&& (_classInfo != null))
		{
			objects = new LyClassWorkSpace(session, _classInfo,
                new LyInfoList<LyFieldInfo>(), null, _classInfo.getSort(), true, true);
			setObjects(objects);
		}
	}

	/**
	 * Factory of the default builder for a controller. Overriden to create an instance of
	 * LyConsultBuilder
	 *
	 * @return the builder which creates the XML description of the view
	 * @see    leon.control.builder.LyConsultBuilder
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyConsultBuilder(this);
	}

	/**
	 * Factory for constructing the field controller of a field info.
	 *
	 * @param  fieldInfo Field info fo which a controller is built.
	 * @param forceConsult true to force consult
	 * @return The field controller of the given field.
	 */
	public LyFieldController fieldControllerFactory(LyFieldInfo fieldInfo, boolean forceConsult)
	{
		if (!isDisplayed(fieldInfo)
                && !forceConsult)
			return null;

		if (fieldInfo instanceof LyTextFieldInfo)
			return new LyTextFieldConsultController(this, (LyTextFieldInfo)fieldInfo);
		else if (fieldInfo instanceof LyFileFieldInfo)
			return new LyFileFieldConsultController(this, (LyFileFieldInfo)fieldInfo);
		else if (fieldInfo instanceof LyChoiceFieldInfo)
			return new LyChoiceFieldConsultController(this, (LyChoiceFieldInfo)fieldInfo);
		else if (fieldInfo instanceof LyStructFieldInfo)
		{
			// Test if we have a format or an value label.
			LyStructFieldInfo struct = (LyStructFieldInfo)fieldInfo;

			if ((struct.getClassInfoCount() > 1)
					&& (struct.fromFieldTag())
					&& !(struct.isMultiple()))
			{
				for (int i = 0; i < struct.getClassInfoCount(); i++)
				{
					LyClassInfo		  subStructClass = struct.getClassInfo(i);
					LyStructFieldInfo subStructFieldInfo = (LyStructFieldInfo)subStructClass
							.getParent();
					LyUnits			  units = subStructFieldInfo.getUnits();

					if (units != null)
						return new LyCompositionFieldUnitConsultController(this, struct);
				}
				return new LyCompositionFieldConsultController(this, struct);
			}
			else
			{
				// Test if we have a format or an value label.
				LyUnit	    unit = getDefaultFieldUnit(fieldInfo);
				LyInfoLabel label = fieldInfo.getLabel(LyInfoLabel.Type.VALUE);

				if ((struct.fromFieldTag())
						&& !(struct.isMultiple()))
				{
					LyClassInfo subStructClass = struct.getClassInfo();

					if ((_object != null)
							&& _object.getClassInfo().contains(struct))
					{
						LyRelationValue relationValue = _object.getRelationValue(struct);

						if (relationValue.getValueCount() > 0)
						{
							LyObject object = relationValue.getObject(0);

							if (object != null)
								subStructClass = object.getClassInfo();
						}
					}

					LyStructFieldInfo subStruct = (LyStructFieldInfo)subStructClass.getParent();

					unit = getDefaultFieldUnit(subStruct);
					label = subStruct.getLabel(LyInfoLabel.Type.VALUE);

					if ((unit != null) && (label != null))
						return new LyCompositionFieldUnitConsultController(this,
							(LyStructFieldInfo)fieldInfo);
				}

				if ((unit != null) || (label != null))
					return new LyFieldConsultController(this, fieldInfo);
				else
					return new LyCompositionFieldConsultController(this,
						(LyStructFieldInfo)fieldInfo);
			}
		}
		else if (fieldInfo instanceof LyRelationFieldInfo)
		{
			LyRelationFieldInfo         relation = (LyRelationFieldInfo)fieldInfo;
            LyRelationFieldInfo.Type	type = relation.getType();

			switch (type)
			{
			case COMPOSITION:
				if (fieldInfo.isMultiple())
					return new LyMultiCompositionFieldConsultController(this, relation);
				else
					return new LyCompositionFieldConsultController(this, relation);
			default:
				if (fieldInfo.isMultiple()
						&& (relation.getShowAs() == LyRelationFieldInfo.ShowAs.TABLE))
					return new LyMultiCompositionFieldConsultController(this, relation);
				else if (!fieldInfo.isMultiple()
                        && (relation.getShowAs() == LyRelationFieldInfo.ShowAs.FORM))
                    return new LyCompositionFieldConsultController(this, relation);
                else
					return new LyRelationFieldConsultController(this, relation);
			}
		}
		else if (fieldInfo instanceof LyArrayFieldInfo)
		{
			LyFieldInfo subFieldInfo = ((LyArrayFieldInfo)fieldInfo).getFieldInfo();

			if (subFieldInfo instanceof LyStructFieldInfo)
			{
				// Test if we have a format or an value label.
				LyUnit	    unit = getDefaultFieldUnit(subFieldInfo);
				LyInfoLabel label = fieldInfo.getLabel(LyInfoLabel.Type.VALUE);

				if ((unit != null) || (label != null))
					return new LyArrayFieldConsultController(this, (LyArrayFieldInfo)fieldInfo);
				else
					return new LyCompositionArrayFieldConsultController(this,
						(LyArrayFieldInfo)fieldInfo, (LyStructFieldInfo)subFieldInfo);
			}
			else if (isComposition(subFieldInfo))
				return new LyCompositionArrayFieldConsultController(this,
					(LyArrayFieldInfo)fieldInfo, (LyRelationFieldInfo)subFieldInfo);
			else
				return new LyArrayFieldConsultController(this, (LyArrayFieldInfo)fieldInfo);
		}
		else
			return new LyFieldConsultController(this, fieldInfo);
	} // end method fieldControllerFactory

	/**
	 * Called when current object has changed . Keep the overall content consistent. Overriden to
	 * call the behavior setObject() method.
	 *
	 * @param oldValues set of former values, or null if not specified.
	 */
	protected void updateObject(LyValueSet oldValues)
	{
		super.updateObject(oldValues);

		LyConsultBehaviorInterface behavior = getConsultBehavior();

		behavior.setObject(this, getObject());
	}

	/**
	 * Accessor to controler behavior : shortcut to (LyConsultBehaviorInterface)getBehavior().
	 *
	 * @return associated behavior
	 * @see    LyAreaController#getBehavior
	 */
	public LyConsultBehaviorInterface getConsultBehavior()
	{
		return (LyConsultBehaviorInterface)getBehavior();
	}

	/**
	 * Update object values. Demand by its composition relation.
	 *
	 * @param relationField The composition relation field.
	 */
	public void updateObjectFromCompositionRelation(LyRelationFieldInfo relationField)
	{
		if (relationField != _compositionRelation)
			return;
		updateObject(null);
	}

	/**
	 * Invoked when an menu item notification is sent to this listener. Manages menu demands on
	 * captions displayed in consult form.
	 *
	 * @param demand demand parameters sent by the view component.
	 */
	public void menuDemand(LyMenuDemand demand)
	{
		String demandCode = demand.getSource().intern();

		if (((demand.getType() == LyMenuDemand.TYPE_POPUP)
					|| (demand.getType() == LyMenuDemand.TYPE_EXTEND_POPUP))
				&& demandCode.startsWith("_caption_")
				&& allowFastEdition())
		{
			demandCode = demandCode.substring(demandCode.indexOf(FORM_FIELD_SEP)
					+ FORM_FIELD_SEP.length());

			LyClassInfo classInfo = getClassInfo();
			LyFieldInfo fieldInfo = classInfo.getFieldInfo(demandCode);
			int[]	    p = demand.getPosition();

			// An extended popup demand on a caption results in the fast edition of an object
			invokeFastEdition(getObject(), fieldInfo, p[0], p[1]);

			return;
		}

		super.menuDemand(demand);
	}

	/**
	 * Indicates if fast edition of a field info is allowed.
	 *
	 * @return true if fast edition of a field info is allowed, false otherwise.
	 */
	protected boolean allowFastEdition()
	{
		LyEnvironment environment = getEnvironment();

		return environment.testResource("LY_FAST_EDITION");
	}

	/**
	 * Invokes an action on the specified object. Closes the consult view if a set action is
	 * invoked.
	 * @param action the action to invoke
	 * @param object the object on which to invoke the action
	 * @return the resulting controller
	 */
	protected LyController invokeAction(LyAction action, LyObject object)
	{
		if ((object != _object)
				|| !(_parent instanceof LyViewController))
			return super.invokeAction(action, object);

		LyApplication application = getApplication();
		LyAction	  setAction = application.getAction("_set");

		if ((action != setAction)
				&& !action.hasTemplate(setAction))
			return super.invokeAction(action, object);

		LyClassInfo classInfo = object.getClassInfo();
		LySession   session = getSession();

		if ((action.getClassInfo() != null)
				&& action.getClassInfo().isVirtual())
			classInfo = action.getClassInfo();

		LyController parent = getViewController().getParent();
		LyController controller = session.invokeAction(parent, action, classInfo,
				new LySimpleObjectList(object), false);

		if ((controller instanceof LyAreaController)
				&& (_areaStackController != null))
			openSubAreaController((LyAreaController)controller, true);

	    controller.showView();

	    // Change tab if needed
		LyTabs tabs = getTabs();

		if ((tabs != null)
				&& (controller instanceof LyFormController))
		{
			LyComponent component = controller.getComponent();

			if (component != null)
			{
				LyTabs controllerTabs = ((LyFormController)controller).getTabs();

				if (controllerTabs != null)
				{
					String tab = tabs.getSelectedTab();

					if (tab != null)
						controllerTabs.setSelectedTab(tab);
				}
			}
		}

		free();
		return controller;
	} // end method invokeAction

	/**
	 * Gets the controller context.
	 * This implementation will return the parent controller context if any.
	 *
	 * @return The controller context.
	 */
	public LyContext getControllerContext()
	{
		// We don't have a context but our parent may have a context
		LyController parent = _parent;

		if (parent instanceof LyDialogController)
			parent = parent.getParent();
		else if (parent instanceof LyFieldSetController)
		{
			parent = parent.getParent();
		}

		// Last try to get a root level context controller (FFT01632)
		if (!(parent instanceof LyControllerContextInterface))
			parent = getSession().getRootContextController();

		LyContext context = null;
		if (parent instanceof LyControllerContextInterface)
		{
            // Be careful, the parent can return null even if it implements the interface
			if (parent instanceof LyControllerContextInterface2)
				context = ((LyControllerContextInterface2)parent).getControllerContext(this);
			else
				context = ((LyControllerContextInterface)parent).getControllerContext();
		}

		return context;
	} // end method getControllerContext

	/**
	 * Impact of a value change of a fieldInfo. The method invokes LyClassBehavior.propagate which
	 * may be used to implement specific impact on a value changed. This method is for internal use,
	 * application should use setFieldValue instead.
	 *
	 * @param fieldInfo current fieldInfo associated to the form field which has been modified.
	 * @param value     new Value of the field.
	 * @see   leon.app.behavior.LyClassBehavior#propagate
	 */
	protected void propagate(LyFieldInfo fieldInfo, Object value)
	{
		LyClassInfo   classInfo = getClassInfo();
		short		  nr = classInfo.getRuleCount();
		LyRule		  rule;
		LyFieldInfo   depend;
		boolean		  checked;
		LyApplication application = getApplication();
		int			  priority = application.getPriorityLevel(_object);
		LyObjectInfo  objectInfo;

		// Execute rules on class
		if (nr > 0)
		{
			// Look for rules to apply
			LyFilter			    filter;
			short				    j;
			LyValueSet			    values = null;
			int					    i, n;
			LyFieldInfo			    field;
			LyValue				    fieldValue;
			LyInfoList<LyFieldInfo> filterFields;

			for (j = 0; j < nr; j++)
			{
				rule = classInfo.getRule(j);
				if (rule.getPriority() > priority)
					continue;
				filter = rule.getFilter();
				if ((filter == null)
						|| !filter.contains(fieldInfo))
					continue;

				objectInfo = rule.getObjectInfo();
				if (!(objectInfo instanceof LyFieldInfo))
					continue;

                // Extract values from input fields and object for missing fields
				if (values == null)
					values = new LyValueSet(classInfo);

				filterFields = filter.getFieldInfos(false);

				for (i = 0, n = filterFields.getSize(); i < n; i++)
				{
					field = filterFields.getInfo(i);
					fieldValue = getFieldValue(field);
					if (fieldValue == null)
					{
						if (_object != null)
							values.addValue(field, _object.getFieldValue(field));
						else
							values.addValue(field, null);
					}
					else
						values.addValue(fieldValue);
				}

				checked = filter.match(values);
				execRule(rule, checked);

				// Replace value for depend if changed when applying rule
				depend = (LyFieldInfo)objectInfo;
				fieldValue = getFieldValue(depend);
				values.addValue(depend, fieldValue);
			}
		}

		LyFieldController fieldController = getFieldController(fieldInfo);

	    if (fieldController != null)
	        fieldController.propagate(value);
	} // end method propagate

	/**
	 * Exec given rule inside this controller.
	 *
	 * @param rule    Rule to apply.
	 * @param checked True if the rule is checked, false if unchecked.
	 */
	public void execRule(LyRule rule, boolean checked)
	{
		execRule(rule, checked, false);
	}

	/**
	 * Exec given rule inside this controller.
	 *
	 * @param rule    Rule to apply.
	 * @param checked True if the rule is checked, false if unchecked.
	 * @param valueChange indicates if this calls comes from a value changed event
	 */
	protected void execRule(LyRule rule, boolean checked, boolean valueChange)
	{
		LyFieldInfo depend = (LyFieldInfo)rule.getObjectInfo();

		depend = _classInfo.getRealFieldInfo(depend);

		if (contains(depend))
		{
			boolean			 onCheck = rule.isOnCheck();
			LyRule.Operation operation = rule.getOperation();

			if (operation == LyRule.Operation.SET_VISIBLE)
				setHidden(depend, (onCheck != checked));
			else if (operation == LyRule.Operation.SET_LABEL)
				setLabel(depend, rule.getLabel(), (onCheck == checked));
		}
		else if (_fieldControllers != null)
		{
			Enumeration<LyFieldController> elements = _fieldControllers.elements();
			LyFieldController              fieldController;
			LyAreaController               areaController;

			while (elements.hasMoreElements())
			{
				fieldController = elements.nextElement();
				areaController = fieldController.getAreaController();
				if (areaController instanceof LyConsultController)
				{
					LyClassInfo classInfo = ((LyConsultController)areaController).getClassInfo();
					if (classInfo.contains(depend))
						((LyConsultController)areaController).execRule(rule, checked, valueChange);
				}
			}
		}
	} // end method execRule

	/**
	 * Changes current field status to (non-)hidden. If the field is hidden, it is not submited with
	 * all other values. When the field is set hidden, it is also set enabled.
	 *
	 * @param fieldInfo the field which should be set hidden
	 * @param hidden    new fieldInfo visibility. (default is true)
	 */
	public void setHidden(LyFieldInfo fieldInfo, boolean hidden)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController instanceof LyFieldConsultController)
			((LyFieldConsultController)fieldController).setHidden(hidden);

	    String group = fieldInfo.getGroup();

        if ((group != null) && (group.length() > 0))
        {
            LyFormField formField = getFormField(fieldInfo);
            LyForm      form = (LyForm)formField.getParent();
            LyFormField field;
            int         i, n = (form == null ? 0 : form.getFormFieldCount());
            boolean     visible = false;

            for (i = 0; (i < n) && !visible; i++)
            {
                field = form.getFormField(i);
                visible = !field.isHidden();
            }

            if ((form != null)
                && (form.getParent() != null))
            {
                if (form.getParent() instanceof LyArea)
                    ((LyArea)form.getParent()).setHidden(!visible);

                form.getParent().setVisible(visible);
            }
        }

		String tabLabel = fieldInfo.getTab();

		if ((tabLabel == null) || (tabLabel.length() == 0))
			return;

		// If a least one field in the current tab is visible, the tab is enabled
		// otherwise, it is disabled...
		LyFieldInfo   field, firstVisibleField = null;
		boolean		  visible = false;
        int           i, n = _classInfo.getFieldInfoCount();

		for (i = 0; (i < n) && !visible; i++)
		{
            field = _classInfo.getFieldInfo(i);
            if (!isDisplayed(field))
                continue;

			if ((field.getTab() == null)
			    || !field.getTab().equals(tabLabel))
			{
				if ((firstVisibleField == null)
				       && (field.getTab() != null))
					firstVisibleField = field;
			}
			else
			    visible = !isHidden(field);
		}

		LyFormField formField = getFormField(fieldInfo);
		LyTabs	    tabs;
		LyTab	    tab;
		LyComponent component = getComponent();

		tabs = formField.getParentTabs(component);
		if (tabs != null)
		{
			for (i = 0, n = tabs.getTabCount(); i < n; i++)
			{
				tab = tabs.getTab(i);
				if ((tab.getLabel() != null)
						&& tab.getLabel().equals(tabLabel))
				{
					tab.setEnabled(visible);
					if (!visible
							&& (tabs.getSelectedTab() != null)
							&& tabs.getSelectedTab().equals(tabLabel)
							&& (firstVisibleField != null))
						tabs.setSelectedTab(firstVisibleField.getTab());
					break;
				}
			}
		}
	} // end method setHidden

	/**
	 * Accessor to the hidden status of a field in the form.
	 *
	 * @param  fieldInfo the specified field
	 * @return current field visibility
	 */
	public boolean isHidden(LyFieldInfo fieldInfo)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		return ((fieldController instanceof LyFieldConsultController)
				&& ((LyFieldConsultController)fieldController).isHidden());
	}

	@Override
	public void setIdle(boolean idle)
	{
	    if (!idle && !(this instanceof LySetController))
	        updateObject(null);
	    super.setIdle(idle);
	}

	/**
	 * Changes the label for the specified fieldinfo.
	 *
	 * @param fieldInfo the fieldInfo
	 * @param label the new label
	 * @param labelSet indicates if the label apply to set form
	 */
	public void setLabel(LyFieldInfo fieldInfo, LyInfoLabel label, boolean labelSet)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController instanceof LyFieldConsultController)
			((LyFieldConsultController)fieldController).setLabel(label, labelSet);
	}
} // end class LyConsultController
