/**
 * LyContextListController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyContextListController.java 28422 2010-11-03 12:03:38Z ehelmste $
 */

package leon.control;

import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.view.*;
import leon.view.event.*;


/**
 * The context list controller manages a list of objects which is the context of another view.
 *
 * <P>Optionally associated to a table controller (automatically created by a LyTableController when
 * the context is not empty), it creates a separated table view of objects belonging to the context.
 * </P>
 *
 * <P>Within a table controller, this view is located to the left of the table, and the user may
 * open it to see current context.</P>
 *
 * <P>This controller shows objects names from the context and enables usual table interactions
 * (popup menu, selection, double-clic ...)</P>
 *
 * @see leon.control.builder.LyContextListBuilder
 */
public class LyContextListController extends LySimpleTableController
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyContextListController.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Constructor.
	 *
	 * @param session    current session associated to the controller. The controller automatically
	 *                   registers on the specified session.
	 * @param parent     parent controller (or null for 'root' controller). Controller hierarchy
	 *                   indicates a child controller (for example a dialog created from a main
	 *                   view)
	 * @param action     related action. If an action is specified, the controller uses action
	 *                   parameters (action name, action specific XML view description ...)
	 * @param objectList list of objects shown in the contextual view.
	 */
	public LyContextListController(LySession session, LyController parent, LyAction action,
		LyObjectList objectList)
	{
		super(session, parent, action, null, objectList);

		if (_classInfo == null)
		{
			LyInfoList<LyClassInfo> classes = objectList.getClasses();

			if ((classes != null) && (classes.getSize() == 1))
				_classInfo = classes.getInfo(0);
		}

		_workSpace = new LySimpleWorkSpace(session, objectList, true);
		_workSpace.addListener(this);

		_initialWorkSpace = _workSpace;
	}

	/**
	 * Constructor.
	 *
	 * @param session    current session associated to the controller. The controller automatically
	 *                   registers on the specified session.
	 * @param parent     parent controller (or null for 'root' controller). Controller hierarchy
	 *                   indicates a child controller (for example a dialog created from a main
	 *                   view)
	 * @param action     related action. If an action is specified, the controller uses action
	 *                   parameters (action name, action specific XML view description ...)
	 * @param classInfo  class info for objects displayed in the list.
	 * @param objectList list of objects shown in the contextual view.
	 */
	public LyContextListController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objectList)
	{
		this(session, parent, action, objectList);

		if (_classInfo == classInfo)
			_classInfo = classInfo;
	}

	/**
	 * Add a new object in the table view.
	 *
	 * <P>Automatically invoked by the parent class when an object is added in in the context list.
	 * </P>
	 *
	 * <P>This method implements the impact on the XML description of the view.</P>
	 *
	 * @param object   object added in the contextual list, and which should be added in the view.
	 * @param index    position in the view (row number) where the object should be added. -1,
	 *                 appends the object at the end of the list.
	 * @param selected current state of the object : is is currently selected ?
	 */
	protected void insertRow(LyObject object, int index, boolean selected)
	{
		LyTable table = getTableComponent();
		LyRows  rows = table.getRows();

		if (rows == null)
		{
			rows = new LyRows();
			table.setRows(rows);
		}

		LyRow row = buildRow(object, null, null);

		if (selected)
			row.setSelected(true);
		rows.insertRow(row, index);
	}

	/**
	 * Current builder of a new row.
	 *
	 * @param  object       object added in the contextual list, and which should set the content of
	 *                      row cells.
	 * @param  fields       list of fields which should be seen.*Not* used, since the builder always
	 *                      shows the name of objects. Kept for compatibility and consistency with
	 *                      LyTableController.
	 * @param  statusFields list of fields which may set the status icon.*Not* used, since the
	 *                      contextual view is not showing status marks. Kept for compatibility and
	 *                      consistency with LyTableController.
	 * @return LyRow the new Row description (XML description) which is added in the LyTable
	 *         component.
	 * @see    #insertRow
	 */
	public LyRow buildRow(LyObject object, LyInfoList<LyFieldInfo> fields,
		LyInfoList<LyFieldInfo> statusFields)
	{
		if (object == null)
			return null;

		LyRow  row = new LyRow();
		LyCell cell = new LyCell();

		row.setId(object.getId());
		cell.setLabel(getName(object));
		row.addCell(cell);

		return row;
	}

	/**
	 * Invoked when an menu item notification is sent to this listener.
	 *
	 * <P>This method overriden to create a popup menu : buildPopupMenu. Method is overriden to
	 * fetch correct class info in the context.</P>
	 *
	 * @param demand demand parameters sent by the view component.
	 */
	public void menuDemand(LyMenuDemand demand)
	{
		if (demand == null)
			return;

		LyAction action = getAction();

		if (action.hasMark(LyAction.MARK_NO_MENU))
			return;
		if (demand.getSource().equals(getId() + "_table_headers"))
		    return;

		int		 index = demand.getIndex(0);
		LyObject object = ((index >= 0) ? _workSpace.getObject(index) : null);

		if (object == null)
			return;

		// Build popup menu
		LyPopupMenu popupMenu = buildPopupMenu(demand, object.getClassInfo(), (index >= 0));
		
		// Call the behavior to complete popup menu
		LyBehaviorInterface behavior = getBehavior();

		if (behavior instanceof LyPanelBehaviorInterface)
			((LyPanelBehaviorInterface)behavior).completePopupMenu(this, popupMenu);

		// Coordinates
		int[]	    p = demand.getPosition();

		popupMenu.setX(p[0]);
		popupMenu.setY(p[1]);

		LyTable table = getTableComponent();

		if (table == null)
			// An error occured, just return
			return;

		popupMenu.setController(getId());
		popupMenu.setEnvironment(table.getEnvironment());
		table.setPopupMenu(popupMenu);
	} // end method menuDemand

	/**
	 * Reset the content of the contextual view.
	 *
	 * <P>This method is called by the reset method of LySimpleTableController.</P>
	 *
	 * @see leon.control.LySimpleTableController#reset
	 */
	protected void updateContent()
	{
		if (_workSpace == null)
			return;

		LyTable table = getTableComponent();
		LyRows  rows = table.getRows();
		Object  lock = ((rows == null) ? new Object() : rows);

		// Remove existing rows
		int     i, n;

		synchronized (lock)
		{
			n = ((rows == null) ? 0 : rows.getRowCount());
			for (i = n - 1; i >= 0; i--)
				rows.removeRow(i);
			// Add new rows
			n = _workSpace.getSize();

			if (n > 0)
			{
				if (rows == null)
				{
					rows = new LyRows();
					table.setRows(rows);
				}

				LyObject object;
				LyRow    row;

				for (i = 0; i < n; i++)
				{
					object = _workSpace.getObject(i);
					row = buildRow(object, null, null);
					rows.addRow(row);
				}
			}
		}

		// Update the last modification time.
		setLastModified(new Date().getTime());
	} // end method updateContent

	/**
	 * Reset the headers of the contextual view.
	 *
	 * <P>This method is called by the reset method of LySimpleTableController, and is overriden to
	 * do 'nothing'. Headers are 'fixed' in a contextual view and created by the builder.</P>
	 *
	 * @see leon.control.LySimpleTableController#reset
	 */
	protected void updateHeaders()
	{
	}

	/**
	 * Factory of the default builder for a controller. Overriden to create an instance of
	 * LyConsultBuilder
	 *
	 * @return the builder which creates the XML description of the view
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyContextListBuilder(this);
	}
} // end class LyContextListController
