/**
 * LyController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyController.java 41509 2012-11-13 09:45:55Z bangiole $
 */

package leon.control;

import java.lang.reflect.*;
import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * The controller class is inherited by all view controllers. It provides de basic services and
 * accessors common to all implementations of view controllers.
 *
 * <P>A controller is usually the result of the invocation of an action. Generic actions (_create,
 * _consult, _table, _tree ...) create a new controller instance.</P>
 *
 * <P>A controller implements the generic processing associated to a view, the integration between
 * the view layer (graphical components) and the data Layer.</P>
 *
 * <P>A controllers also implements the generic response to user interactions and content (data)
 * modifications.</P>
 *
 * <P>Controllers enable the views to be 'dynamic', and they are usually associated to a LyBehavior
 * class which enable the user to override.</P>
 *
 * <P>A <b>LyBehavior</b> instance is usually associated to a controller : LyBehavior enables the
 * developper of a Leonardi aplication to change or extend the generic response of the controller to
 * create the logic of the application.</P>
 *
 * <P>Normally, simple developments tasks should be implemented as an inheritance of the LyBehavior.
 * Creating you own controllers or inheritance from a generic controller if of course possible, but
 * should be considered as a more advanced task not possible to achieve in the behavior.</P>
 *
 * <P>Main tasks of a controller :</P>
 *
 * <UL>
 *   <LI>Creation of the view structure (using a dedicated builder or from an XML File)</LI>
 *   <LI>Basic view operations : showView, hideView, free</LI>
 *   <LI>Several accessors on resources : session, application, environment ...
 *
 *     <P>Controller are registered on the user session.</P>
 *   </LI>
 * </UL>
 */
public abstract class LyController
	implements LyGuiListener,
		LyDemandListener,
		LyEvaluator,
		LyGetSessionInterface
{
	/**
	 * File version
	 */
	public static final String  __VERSION =
		"$Id: LyController.java 41509 2012-11-13 09:45:55Z bangiole $";

	/** Default image extension. The default is ".gif". Another possible value is '.jpg' */
	public static String	    DEFAULT_IMAGE_FORMAT = ".gif";

	/**
	 * Separator for view ids. The separator is used in the view ids to separate some information
	 * pieces.
	 */
	public static final String  ID_SEP = "::";

	/**
	 * Separator for controller ids. This separator is used to generate constant identifiers for
	 * controllers.
	 */
	public static final char    CONTROLLER_ID_SEP = '_';

	/* Demand codes */

	/** Number of controllers for debug purpose. */
	public static int		    _nbControllers;

	/** Used to compute id sequentially (in test mode). */
	public static long		    _nextId = 0;

	/**
	 * The key of the Help demand. When the help button is pressed in a view, the view event
	 * (demand) return this as the source object.
	 *
	 * @see leon.view.event.LyDemand#getSource
	 */
	public static final String  DEMAND_HELP = "_help".intern();

	/**
	 * Identifier of the controller. The identifier of the controller is unique within a user
	 * session The identifier is used to identify which controller is processing the demands (user
	 * events) of a view.
	 */
	protected String		    _id;

	/**
	 * Name of the controller. The name of the controller may be displayed to the user in a view.
	 */
	protected String		    _name;

	/** Reference to the parent controller. This reference is null for root controllers. */
	protected LyController	    _parent;

	/**
	 * Internal reference to the view component builder. All classes must use the getBuilder
	 * accessor and not this value directly. This value is only an internal cache for optimization.
	 */
	protected LyBuilder		    _builder;

	/** Internal reference to the specific builder class if exists. */
	protected Class<?>			    _builderClass;

	/** Behavior of this controller. */
	// This attribute must remain private because it's initialized by getBehavior method
	private LyBehaviorInterface _behavior;

	/**
	 * Component controlled by this controller. All classes must use the getComponent accessor and
	 * not this value directly. This value is only an internal cache for optimization.
	 */
	protected LyComponent	    _component;

	/** The current session of the controller. */
	protected LySession		    _session;

	/** The last modification time of this controller. */
	protected long			    _lastModified;

	/** Boolean set when the controller is freed. */
	protected boolean		    _freed;

	/** Boolean set when the controller is idle. */
	protected boolean		    _idle;

    /** User data. */
    protected Object            _userData;

	/**
	 * Constructor. The identifier of this controller is automatically generated (random id).
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 */
	public LyController(LySession session, LyController parent)
	{
		// Automatically create a random id
		this(session, parent, null);
	}

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 * @param id      identifier of the controller : can not be null. To generate the id
	 *                automatically, use LyString.getRandomId() or alternate constructor.
	 */
	public LyController(LySession session, LyController parent, String id)
	{
		_nbControllers++;

		if (session == null)
			throw new IllegalArgumentException("BF AppEngine : Invalid session (null)");
		if ((parent != null)
		        && (parent.getApplication() == session.getApplication()) // Added for studio preview
		        && (parent.getSession() != session))
			throw new IllegalArgumentException(
				"BF AppEngine : if a parent is given, it has to belong to the session given as parameter");

		// Set attributes
		_freed = false;
		_idle = false;
		_parent = parent;
		_id = computeId(session, id);

		// Set session
		setSession(session);

		// Update the last modified time.
		setLastModified(System.currentTimeMillis());
	}

	/**
	 * Internal method to compute id of this controller.
	 * @param session Session of the new controller.
	 * @param id Desired id (may be null).
	 * @return Id of the new controller.
	 */
	protected String computeId(LySession session, String id)
	{
	    String    newId = null;

        if (id == null)
        {
            // Old-fashion id
            if (LyEnvironment.TEST_MODE)
                id = Long.toString(_nextId++);
            else
                id = LyString.getRandomId();

            // Set id
            String name = getClass().getName();
            int    index = name.lastIndexOf('.');

            if (index >= 0)
                name = name.substring(index + 1);

            newId = name + '_' + id;
        }
        else
        {
            // Check if we don't have already this id
            int n = 1;

            newId = id;
            while (session.getController(newId) != null)
            {
                n++;
                newId = id + CONTROLLER_ID_SEP + n;
            }
        }

        newId = session.checkControllerId(this, newId);

        newId = newId.intern();
        return newId;
	}

	/**
	 * Finalize this controller
	 */
	protected void finalize()
	{
		_nbControllers--;
	}

	/**
	 * Make the view visible to the user. The effect of this command depends on the nature of the
	 * application. If the display is a fat client (windows (awt) / swing) this methods should open
	 * a window and show the view. In the case of thin client (html), this prepares the pages which
	 * will be returned and displayed in the client browser.
	 *
	 * <P>For any display, this method should consider that the page is ready to be shown and should
	 * be displayed.</P>
	 */
	public void showView()
	{
		if (isFreed())
			throw new IllegalStateException("BF AppEngine : attempt to show controller "
				+ this + " that was freed");

		LyViewController viewController = getViewController();

		if (viewController == null)
			throw new IllegalStateException("BF AppEngine : no view controller found for " + this);

		// Let view controller do the job...
		viewController.showView();
	}

	/**
	 * Hide the view. The effect of this command depends on the nature of the application. If the
	 * display is a fat client for example (windows (awt) / swing) this methods should close the
	 * window or hide the view.
	 *
	 * <P>This method should consider that the page is closed and should disappear.</P>
	 *
	 * <P>The view is only HIDDEN, not cleared (see free), so it should be possible to open it again
	 * with showView and resources allocated for the view are not freed.</P>
	 */
	public void hideView()
	{
		hideView(false);
	}

	/**
	 * Hide the view. The effect of this command depends on the nature of the application. If the
	 * display is a fat client for example (windows (awt) / swing) this methods should close the
	 * window or hide the view.
	 *
	 * <P>This method should consider that the page is closed and should disappear.</P>
	 *
	 * @param free Indicates if only HIDDEN, not FREED (true).
	 */
	public void hideView(boolean free)
	{
		LyViewController viewController = getViewController();

		if (viewController == null)
			throw new IllegalStateException("BF AppEngine : no view controller found for " + this);

		// Let view controller do the job...
		viewController.hideView(free);
	}

	/**
	 * Turns temporarily the view to visible/not visible.
	 *
	 * <P>The view is only HIDDEN, not cleared (see free), so it should be possible to open it again
	 * with setVisible and resources allocated for the view are not freed.</P>
	 * @param visible true to set the view visible
	 */
	protected void setVisible(boolean visible)
	{
		LyViewController viewController = getViewController();

		if (viewController == null)
			throw new IllegalStateException("BF AppEngine : no view controller found for " + this);

		// Let view controller do the job...
		viewController.setVisible(visible);
	}

	/**
	 * Notify a visibility change of the parent view.
	 * @param visible true when the view become visible
	 * @see #setVisible(boolean)
	 */
	protected void visibilityChanged(boolean visible)
	{
	}

	/**
	 * Free this controller. This method unregister the controller from the session and stop
	 * listening from events. Additionally, the method should free all internal resources allocated
	 * by the controller. The controller instance can not be used anymore after the invocation of
	 * the free method.
	 */
	public void free()
	{
		if (isFreed())
			throw new IllegalStateException("BF AppEngine : controller " + this + " was already freed");

		_idle = true;

		// Free behavior
		if (_behavior != null)
		{
			_behavior.free();
			_behavior = null;
		}

		// Unregister this controller from the demand transmitter.
		LyViewManager viewManager = getViewManager();

		if (viewManager != null)
			viewManager.removeDemandListener(this);

		// Unregister this controller fom the session.
		LySession session = getSession();

		if (session != null)
			session.removeController(this);

		// ** free internal references (cleanup)
		_builder = null;
		_builderClass = null;
		_behavior = null;
		_parent = null;
		if (_component != null)
			_component.free();
		_component = null;

		// Don't set session to null, it's still needed in endRun to unregister
		// current session on application for the current thread
		// _session = null;

		_freed = true;
	} // end method free

	/**
	 * Accessor to the identifier of the controller. The identifier of the controller is unique
	 * within a user session.
	 *
	 * <p>Additionally, this method is required for the LyDemandListener interface.</p>
	 *
	 * @return The identifier which is used to identify which controller is processing the demands
	 *         (user events) of a view.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Accessor to the name of the controller.
	 *
	 * @return The name of the controller may be displayed to the user in a view.
	 */
	public String getName()
	{
		return _name;
	}

	/**
	 * Accessor to the name of the controller.
	 *
	 * @param name The name of the controller may be displayed to the user in a view.
	 */
	public void setName(String name)
	{
		_name = name;
	}

	/**
	 * Accessor to the reference to the parent controller.
	 *
	 * @return the parent controller instance or null for root controllers.
	 */
	public LyController getParent()
	{
		return _parent;
	}

	/**
	 * Accessor to the reference to the parent controller.
	 *
	 * @param parent the parent controller instance or null for root controllers.
	 */
	protected void setParent(LyController parent)
	{
		_parent = parent;
	}

	/**
	 * Sets a builder for this controller. This method may be used to specify a specific builder for
	 * this controller.
	 *
	 * <P>If no builder was specified, the controller will create a default builder using the
	 * getDefaultBuilder() method.</P>
	 *
	 * @param builder the specified builder, or null to use default.
	 */
	public void setBuilder(LyBuilder builder)
	{
		_builder = builder;
	}

	/**
	 * Sets a builder class for this controller. This method may be used to specify a specific
	 * builder class for this controller.
	 *
	 * <P>If no builder class was specified, the controller will create a default builder using the
	 * getDefaultBuilder() method.</P>
	 *
	 * @param builderClass the specified builder class, or null to use default.
	 */
	public void setBuilderClass(Class<?> builderClass)
	{
		_builderClass = builderClass;
	}

	/**
	 * Accessor to the internal builder. If no builder is set, the method returns the reference to a
	 * default builder (invokes getDefaultBuilder).
	 *
	 * @return the builder which creates the XML description of the view. Returned value is not null
	 *         and several invocation on getBuilder always return the same instance.
	 */
	public LyBuilder getBuilder()
	{
		if (_builder != null)
			return _builder;

		if (_builderClass == null)
		{
			// No specific builder, use default one.
			_builder = getDefaultBuilder();
			if (_builder == null)
				return null;
		}
		else
		{
			// Build a new builder with the specific class.
			LyEnvironment environment = getEnvironment();
			Class<?>[]		  types = new Class[1];
			Object[]	  parameters = new Object[1];

			types[0] = LyController.class;
			parameters[0] = this;

			try
			{
				Constructor<?> constructor = _builderClass.getConstructor(types);

				setBuilder((LyBuilder)constructor.newInstance(parameters));
			}
			catch (InvocationTargetException e)
			{
				environment.writeException(e.getTargetException());

				_builderClass = null;
				getBuilder();
			}
			catch (NoSuchMethodException e)
			{
				// Unknown controller
				environment.logError("BF AppEngine : builder with parameter ("
					+ types[0] + ") not found on class " + _builderClass);

				_builderClass = null;
				getBuilder();
			}
			catch (Exception e)
			{
				environment.writeException(e);

				_builderClass = null;
				getBuilder();
			}
		}

		return _builder;
	} // end method getBuilder

	/**
	 * Factory of the default builder for a controller. The application should never invoke this
	 * method directly (internal use only : use getBuilder instead). This method is specialized in
	 * subclasses of LyController to return a valid builder instance.
	 *
	 * @return the builder which creates the XML description of the view
	 */
	protected abstract LyBuilder getDefaultBuilder();

	/**
	 * Accessor to controller behavior This method is accessor for internal use and subclasses only.
	 * A behavior is associated to a controller to let the developer specify application logic in
	 * the behavior.
	 * @param behavior the behavior to set
	 *
	 */
	protected void setBehavior(LyBehaviorInterface behavior)
	{
		_behavior = behavior;
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyBehaviorInterface.class;
	}

	/**
	 * Accessor to controller behavior A behavior is associated to a controller to let the
	 * Developer specify application logic in the behavior.
	 *
	 * @return associated behavior
	 */
	public LyBehaviorInterface getBehavior()
	{
		return _behavior;
	}

	/**
	 * Accessor to the view component instance. (XML description of the view) This instance is
	 * created by the builder, and multiple calls to this method always return the same instance.
	 *
	 * @return the LyComponent instance of the view package, memory representation (parsed) of the
	 *         XML description of the view.
	 */
	public LyComponent getComponent()
	{
		if (isFreed())
			return null;
		// If component is ready, return it
		if (_component != null)
			return _component;
		// Don't rebuild component while controller is idle
		if (_idle)
			return null;

		// Otherwise, build an empty component
		LyBuilder builder = getBuilder();

		if (builder == null)
		    return null;

		_component = builder.buildComponent();

		// Register this controller to receive user demands
		LyViewManager viewManager = getViewManager();

		// If the view was cleared, this method may be invoked several times
		if (viewManager != null)
		{
			viewManager.removeDemandListener(this);
			viewManager.addDemandListener(this);
		}

		// FFT 578 : To refresh recent controller stack -Usefull when cleared from memory cache)
		// To do : improve this memory management and to no 're-'add the controller
		_session.addController(this);

		// Return component once ready
		return _component;
	} // end method getComponent

	/**
	 * Accessor to the view controller (Frame or dialog controller) which is the container of this
	 * view. This method must be implemented in subclasses depending on the nature of the component.
	 *
	 * @return the reference to a view controller, never returns null if the controller is
	 *         associated to a view.
	 */
	public abstract LyViewController getViewController();

	/**
	 * Accessor view that contains the component (getComponent) of the controller. This is a
	 * shortcut to getViewController.getView(). Returned value depends on the nature of the
	 * controller.
	 *
	 * @return the LyView instance (memory representation of the XML view) which contains the
	 *         current LyComponent.
	 */
	public LyView getView()
	{
		LyViewController viewController = getViewController();

		return ((viewController == null) ? null : viewController.getView());
	}

	/**
	 * Accessor to the session instance associated to this controller.
	 *
	 * @return current session instance.
	 */
	public LySession getSession()
	{
		return _session;
	}

	/**
	 * Set current session and register as a controller. For internal use only.
	 *
	 * @param session current session instance.
	 */
	protected void setSession(LySession session)
	{
		if (_session == session)
			return;

		_session = session;
		_session.addController(this);
	}

	/**
	 * Accessor to the current application instance. (shortcut for getSession().getApplication()).
	 *
	 * @return current application instance.
	 */
	public LyApplication getApplication()
	{
		LySession session = getSession();

		return session.getApplication();
	}

	/**
	 * Accessor to the current Environment instance (shortcut for getSession().getEnvironment()).
	 *
	 * @return current environment instance.
	 */
	public LyEnvironment getEnvironment()
	{
		LySession session = getSession();

		return ((session == null) ? null : session.getEnvironment());
	}

	/**
	 * Accessor to the current view manager instance (shortcut for getViewManager()).
	 *
	 * @return current view manager instance.
	 */
	public LyViewManager getViewManager()
	{
		LyEnvironment environment = getEnvironment();

		return ((environment == null) ? null : (LyViewManager)environment.getViewManager());
	}

	/**
	 * Accessor to the last modification date of this controller. The modification date is updated
	 * when anything in the view has changed. This parameter is used by view manager to decide
	 * whether the cache could be used or not.
	 *
	 * @return the last modification date (as described in java.util.Date.getTime())
	 */
	public long getLastModified()
	{
		return _lastModified;
	}

	/**
	 * Accessor to set the last modification date of this controller. The modification date is
	 * updated when anything in the view has changed. The value should be (new Date()).getTime()
	 *
	 * @param time the new modification time
	 */
	public void setLastModified(long time)
	{
		_lastModified = time;
	}

	/**
	 * Override generic method to display the name of this controller. Default is identical to
	 * getName
	 * @return the name of this controller
	 *
	 * @see #getName
	 */
	public String toString()
	{
		return getName();
	}

	/**
	 * Internal method to check if this controller was freed.
	 *
	 * @return True if this controller was freed, false otherwise.
	 */
	public boolean isFreed()
	{
		return _freed;
	}

	/* LyDemandListener interface */

	/**
	 * Invoked when a notification is sent to this listener. Default implementation redirect the
	 * event depending on demand type :
	 *
	 * <UL>
	 *   <LI>actionDemand</LI>
	 *   <LI>selectionDemand</LI>
	 *   <LI>menuDemand</LI>
	 *   <LI>helpDemand</LI>
	 * </UL>
	 *
	 * <P>The listener should make the interpretation of the event, and update the view description
	 * if necessary.</P>
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   leon.view.event.LyDemandListener interface
	 */
	public synchronized void demandPerformed(LyDemand demand)
	{
		LySession     session = getSession();
		LyEnvironment environment = getEnvironment();

		// Do not handle demand if this controller is freed...
		if ((session == null) || (isFreed()))
		{
			if (environment != null)
				environment.logWarning(
					"A demand was received but the controller is already freed - Skipping it "
					+ demand + " session : " + session);
			else
				System.out.println(
					"A demand was received but the controller is already freed - Skipping it "
					+ demand + " session : " + session);
			return;
		}

		Date		  startDate = new Date();
		String		  toString = toString();
		LyApplication application = session.getApplication();
		LyDataSession oldCurrentSession = application.getCurrentSession();
		boolean		  hadTransaction = session.hasTransaction();
		LyContext     context = null, oldCurrentContext = null;
		LyViewManager viewManager = getViewManager();
		boolean		  multiSession = (viewManager != null) && viewManager.isMultiSession();
		LyEvaluator   oldCurrentEvaluator = application.getCurrentEvaluator();

		if (this instanceof LyControllerContextInterface)
		{
			oldCurrentContext = application.getCurrentContext();
			context = ((LyControllerContextInterface)this).getControllerContext();
			if (context != oldCurrentContext)
				application.setCurrentContext(context);
		}

		try
		{
			application.setCurrentSession(session);
			application.setCurrentEvaluator(this);

			// Reset session timeout
			session.resetTimeout();

			updateViewFromValues(demand);

			// First try to send the demand on the session behavior
			LySessionBehaviorInterface sessionBehavior = (LySessionBehaviorInterface)session
					.getSessionBehavior();

			if ((sessionBehavior != null)
					&& sessionBehavior.handleDemand(session, this, demand))
				return;

			// Try to call controller behavior to handle demand
			LyBehaviorInterface behavior = getBehavior();

			if ((behavior != null)
					&& behavior.handleDemand(this, demand))
				return;

			if (isFreed())
			    // Occur if a behavior frees the controller and return false
			    return;

			// Handle demand
			if (demand instanceof LyActionDemand)
				actionDemand((LyActionDemand)demand);
			else if (demand instanceof LySelectionDemand)
				selectionDemand((LySelectionDemand)demand);
			else if (demand instanceof LyMenuDemand)
				menuDemand((LyMenuDemand)demand);
			else if (demand instanceof LyHelpDemand)
				helpDemand((LyHelpDemand)demand);
			else if (demand instanceof LyActivateViewDemand)
				activateViewDemand((LyActivateViewDemand)demand);
			else if (demand instanceof LyDragDropDemand)
				dragDropDemand((LyDragDropDemand)demand);
			else
				throw new IllegalArgumentException("BF AppEngine : Unknown demand type "
					+ demand.getClass());

			// We mustn't have some pending transaction on the current session
			// if we hadn't one when receiving the demand.
			if (!hadTransaction
					&& session.hasTransaction())
			{
				String message = environment.getMessage("LY_ER_REMAINING_PENDING_TRANSACTION");

				if (isFreed())
					environment.raiseError(message);
				else
					showError(message);

				throw new IllegalStateException("BF AppEngine : Remaining pending transactions");
			}
		}
		catch (Throwable t)
		{
			// Trap exception to avoid application to be stopped
			if (multiSession
					&& isFreed())
			{
				// Demand performed aborted by freeing controller
				// produce non-significant exceptions
				environment.logTrace("BF AppEngine : demand aborted, controller freed during process.");
				if (environment.isDebugLevel(LyEnvironment.DEBUG_ALL))
					session.handleException(t);
			}
			else
				session.handleException(t);

			try
			{
				if (getViewManager() != null)
					getViewManager().setWaiting(false);
			}
			catch (Throwable t2)
			{
				session.handleException(t2);
			}
		}
		finally
		{
			if ((oldCurrentSession != null)
					&& (oldCurrentSession.getState() != LySession.State.CLOSED))
				application.setCurrentSession(oldCurrentSession);

			if (oldCurrentEvaluator != null)
			    application.setCurrentEvaluator(oldCurrentEvaluator);

			if (this instanceof LyControllerContextInterface)
			{
				if ((oldCurrentContext != null)
						&& !oldCurrentContext.isEnded())
				{
					if (oldCurrentContext != context)
						application.setCurrentContext(oldCurrentContext);
				}
				else
					application.setCurrentContext(null);
			}
		}

		if (environment.isDebugLevel(LyEnvironment.DEBUG_PERF))
		{
			Date endDate = new Date();

			System.gc();
			environment.logOut("BF AppEngine : [ " + demand.toShortString() + " ]"
				+ "\non controller : " + toString
				+ "\nhandled in " + (endDate.getTime() - startDate.getTime()) + " ms."
				+ "\n\tCurrent number of LySession = " + LySession._nbSessions
				+ "\n\tCurrent number of LyObject = " + LyObject._nbObjects
				+ "\n\tCurrent number of LyController = " + LyController._nbControllers
				+ "\n\tCurrent number of LyView = " + LyView._nbViews
				+ "\n\tCurrent number of LyComponent = " + LyComponent._nbComponents);
		}
	} // end method demandPerformed

	/**
	 * Update the view of the controller with the specified values from the demand (internal use) .
	 *
	 * @param demand demand parameters sent by the view component.
	 */
	public void updateViewFromValues(LyDemand demand)
	{
		// Update view when needed
		if (demand instanceof LyActionDemand)
		{
			LyParameterValues values = demand.getParameterValues();

			if (values != null)
			{
				values.updateViewFromValues(getComponent());
				setLastModified(System.currentTimeMillis());
			}
		}
	}

	/**
	 * Invoked when an action notification is sent to this listener.
	 *
	 * <P>This method is called by demandPerformed and may be overridden in subclasses.</P>
	 *
	 * <P>Default implementation makes nothing.</P>
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   #demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
			// Don't trace value changes because most of them are not handled
			// and to avoid displaying values of secret fields
			return;

		LyEnvironment environment = getEnvironment();

		environment.logWarning("BF AppEngine : demand not handled " + demand);
	}

	/**
	 * Invoked when an selection notification is sent to this listener.
	 *
	 * <P>This method is called by demandPerformed and may be overriden in subclasses.</P>
	 *
	 * <P>Default implementation makes nothing.</P>
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   #demandPerformed
	 */
	public void selectionDemand(LySelectionDemand demand)
	{
		LyEnvironment environment = getEnvironment();

		environment.logWarning("BF AppEngine : demand not handled " + demand);
	}

	/**
	 * Invoked when an menu item notification is sent to this listener.
	 *
	 * <P>This method is called by demandPerformed and may be overriden in subclasses.</P>
	 *
	 * <P>Default implementation makes nothing.</P>
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   #demandPerformed
	 */
	public void menuDemand(LyMenuDemand demand)
	{
		LyEnvironment environment = getEnvironment();

		environment.logWarning("BF AppEngine : demand not handled " + demand);
	}

	/**
	 * Invoked when an help notification is sent to this listener.
	 *
	 * <P>This method is called by demandPerformed and may be overridden in subclasses.</P>
	 *
	 * <P>Default implementation makes nothing.</P>
	 *
	 * @param demand demand parameters set by the view component.
	 * @see   #demandPerformed
	 */
	public void helpDemand(LyHelpDemand demand)
	{
		LyEnvironment environment = getEnvironment();

		environment.logWarning("BF AppEngine : demand not handled " + demand);
	}

	/**
	 * Invoked when an activate view notification is sent to this listener.
	 *
	 * <P>This method is called by demandPerformed and may be overridden in subclasses.</P>
	 *
	 * <P>Default implementation send a global event to the event manager.</P>
	 *
	 * @param demand demand parameters set by the view component.
	 * @see   #demandPerformed
	 */
	public void activateViewDemand(LyActivateViewDemand demand)
	{
		LyApplication  application = getApplication();
		LyEventManager em = application.getEventManager();
		LyGlobalEvent  event = new LyGlobalEvent(LyGlobalEvent.TYPE_ACTIVE_VIEW, this);

		em.postEvent(event);
	}

	/**
	 * Invoked when a Drag & Drop occurs. (LyDemandListener interface) This method handles: opening
	 * and closing line, moving an interval, action
	 *
	 * @param demand drag-drop demand
	 */
	public void dragDropDemand(LyDragDropDemand demand)
	{
		LyEnvironment environment = getEnvironment();

		environment.logWarning("BF AppEngine : demand not handled " + demand);
	}

	/* Invoke actions of the application */

	/**
	 * Invokes an action.
	 *
	 * <P>The action process is implemented on the session.</P>
	 *
	 * <P>This method is a shortcut (polymorphysm) to invokeAction(LyAction, LyClassInfo,
	 * LyObjectList, boolean)</P>
	 *
	 * @param action the action which is executed.
	 * @return the resulting controller
	 * @see   #invokeAction(LyAction, LyClassInfo, LyObjectList, boolean)
	 * @see   leon.app.LySession#invokeAction
	 */
	protected LyController invokeAction(LyAction action)
	{
		return invokeAction(action, null, null, true);
	}

	/**
	 * Invokes an action on the specified class.
	 *
	 * <P>The action process is implemented on the session.</P>
	 *
	 * <P>This method is a shortcut (polymorphysm) to invokeAction(LyAction, LyClassInfo,
	 * LyObjectList, boolean)</P>
	 *
	 * @param action    the action which is executed.
	 * @param classInfo the target class of the action.
	 * @return the resulting controller
	 * @see   #invokeAction(LyAction, LyClassInfo, LyObjectList, boolean)
	 * @see   leon.app.LySession#invokeAction
	 */
	protected LyController invokeAction(LyAction action, LyClassInfo classInfo)
	{
		return invokeAction(action, classInfo, null, true);
	}

	/**
	 * Invokes an action on the specified object.
	 *
	 * <P>The action process is implemented on the session.</P>
	 *
	 * <P>This method is a shortcut (polymorphysm) to invokeAction(LyAction, LyClassInfo,
	 * LyObjectList, boolean)</P>
	 *
	 * @param action the action which is executed.
	 * @param object specify on which object the action applies
	 * @return the resulting controller
	 * @see   #invokeAction(LyAction, LyClassInfo, LyObjectList, boolean)
	 * @see   leon.app.LySession#invokeAction
	 */
	protected LyController invokeAction(LyAction action, LyObject object)
	{
		return invokeAction(action, object.getClassInfo(), new LySimpleObjectList(object), true);
	}

	/**
	 * Invokes an action on the specified object list.
	 *
	 * <P>The action process is implemented on the session.</P>
	 *
	 * <P>This method is a shortcut (polymorphism) to invokeAction(LyAction, LyClassInfo,
	 * LyObjectList, boolean), but invokes the action on a per class basis. If the object list
	 * contains 5 objects, 2 from classInfo A and 3 from classInfo B, the method will call
	 * invokeAction(LyAction, LyClassInfo, LyObjectList, boolean) twice : once with the two items
	 * from class A, then once with the 3 items from class B.</P>
	 *
	 * @param action  the action which is executed.
	 * @param objects The list of objects on which the action applies. They may not belong to the
	 *                same application class (classInfo).
	 * @return the resulting controller
	 * @see   #invokeAction(LyAction, LyClassInfo, LyObjectList, boolean)
	 * @see   leon.app.LySession#invokeAction
	 */
	protected LyController invokeAction(LyAction action, LyObjectList objects)
	{
		// FFT 1251 : allow invoke the action on heterogeneous objects
		// for actions that have the _allowMulticlass parameter_decl
		if (action.getParameterDecl("_allowMulticlass") != null)
			return invokeAction(action, null, objects, true);

		// Split object list in sub-lists where all objects belong to the same class
		// and then invoke action on each sub-list
		LyInfoList<LyClassInfo> classes = objects.getClasses();
		int					    i, nc = classes.getSize();

		if (nc == 1)
			return invokeAction(action, classes.getInfo(0), objects, true);

		LyClassInfo  classInfo;
		LyObjectList subList;
		LyController controller = null;
		LySession    session = getSession();
		LyAction     action2 = action;

		for (i = 0; i < nc; i++)
		{
			classInfo = classes.getInfo(i);
			if (action.getClassInfo() == null)
				action2 = getActionOnClassInfo(action, classInfo);

			subList = new LySubWorkSpace(session, objects, classInfo, false);
			controller = invokeAction(action2, classInfo, subList, true);
		}

		// Return last controller
		return controller;
	} // end method invokeAction

	/**
	 * Get the good action, if action does not exist on classInfo.
	 *
	 * @param  action    the action which is checked.
	 * @param  classInfo the classInfo on which check is done.
	 * @return the good action on classInfo, same action other cases. see #invokeAction(LyAction,
	 *         LyObjectList)
	 */
	protected LyAction getActionOnClassInfo(LyAction action, LyClassInfo classInfo)
	{
		if (action == null)
			return null;

		LyAction action2 = classInfo.getAction(action.getId());

		//try to find better
		if (action2 == null)
		{
			LyInfoList<LyAction> actions = classInfo.getActions();

			for (int i = 0; ((actions != null) && (i < actions.getSize())); i++)
			{
				action2 = actions.getInfo(i);

				//find good action: same root and same image.
				if ((action.getRoot() == action2.getRoot())
						&& (getImage(action2) != null)
						&& (getImage(action) != null)
						&& (getImage(action2).equals(getImage(action))))
					return action2;
			}
		}

		return action;
	}

	/**
	 * Invokes an action on the specified class and object list.
	 *
	 * <P>The action process is implemented on the session.</P>
	 *
	 * <P>This method is a shortcut (polymorphysm) to invokeAction(LyAction, LyClassInfo,
	 * LyObjectList, boolean)</P>
	 *
	 * @param action    the action which is executed.
	 * @param classInfo the target class of the action.
	 * @param objects   The list of objects on which the action applies. They must belong to the
	 *                  specified classInfo. Otherwise, use invokeAction(LyAction,LyObjectList)
	 * @return the resulting controller
	 * @see   #invokeAction(LyAction, LyClassInfo, LyObjectList, boolean)
	 * @see   leon.app.LySession#invokeAction
	 */
	protected LyController invokeAction(LyAction action, LyClassInfo classInfo,
		LyObjectList objects)
	{
		return invokeAction(action, classInfo, objects, true);
	}

	/**
	 * Invokes an action on the specified class and object list.
	 *
	 * <P>The action process is implemented on the session. This is a shortcut to
	 * getSession().invokeAction All other invokeAction method invoke this method.</P>
	 *
	 * @param action    the action which is executed.
	 * @param classInfo the target class of the action.
	 * @param objects   The list of objects on which the action applies. They must belong to the
	 *                  specified classInfo. Otherwise, use invokeAction(LyAction,LyObjectList)
	 * @param show      indicates if the action must be showned
	 * @return the resulting controller
	 * @see   #invokeAction(LyAction, LyClassInfo, LyObjectList, boolean)
	 * @see   leon.app.LySession#invokeAction
	 */
	protected LyController invokeAction(LyAction action, LyClassInfo classInfo,
		LyObjectList objects, boolean show)
	{
		LySession session = getSession();

		if ((action.getClassInfo() != null)
				&& action.getClassInfo().isVirtual())
			classInfo = action.getClassInfo();

		return session.invokeAction(this, action, classInfo, objects, show);
	}

	/*Utility Methods */

	/**
	 * Shortcut to get a classInfo of the application this method is a utility shortcut to
	 * getApplication().getClassInfo(id);
	 *
	 * @param  id identifier of requested class.
	 * @return the requested classInfo, or null if it could not be found.
	 */
	public LyClassInfo getClassInfo(String id)
	{
		LyApplication application = getApplication();

		return application.getClassInfo(id);
	}

	/**
	 * Get the name of the image (key or path) representing an object instance. Default
	 * implementation returns the icon associated to the classInfo.
	 *
	 * @param  object source object for image association.
	 * @return the image path or key code.
	 */
	public String getImage(LyObject object)
	{
		return getImage(object.getClassInfo());
	}

	/**
	 * Get the name of the image (key or path) representing an objectinfo. Default implementation
	 * builds an image path from a prefix specified in the objectInfo.
	 *
	 * @param  objectInfo Info for which an image is searched.
	 * @return the image path or key code.
	 * @see    #getImageFromPrefix
	 */
	public String getImage(LyObjectInfo objectInfo)
	{
		return getImage(objectInfo, -1, true);
	}

	/**
	 * Get the name of the image (key or path) representing an objectinfo. Default implementation
	 * builds an image path from a prefix specified in the objectInfo.
	 *
	 * @param  objectInfo Info for which an image is searched.
	 * @param  size       Requested size.
	 * @return the image path or key code.
	 * @see    #getImageFromPrefix
	 */
	public String getImage(LyObjectInfo objectInfo, int size)
	{
		return getImage(objectInfo, size, true);
	}

	/**
	 * Get the name of the image (key or path) representing an objectinfo. Default implementation
	 * builds an image path from a prefix specified in the objectInfo.
	 *
	 * @param  objectInfo   Info for which an image is searched.
	 * @param  size         Requested size.
	 * @param  defaultImage Produce a default image if image not found.
	 * @return the image path or key code.
	 * @see    #getImageFromPrefix
	 */
	public String getImage(LyObjectInfo objectInfo, int size, boolean defaultImage)
	{
		LyEnvironment environment = getEnvironment();

		if ((objectInfo instanceof LyClassInfo)
				&& !environment.testResource("LY_SHOW_IMAGES_FOR_OBJECTS"))
			return null;

		String prefix = objectInfo.getFilePrefix();

		if ("_null".equals(prefix))
			return null;

		String image = getImageFromPrefix(prefix);

		// Check if image exists
		if (!environment.checkImage(image))
		{
			if (defaultImage)
			{
				if (objectInfo instanceof LyAction)
				{
					LyClassInfo classInfo = ((LyAction)objectInfo).getClassInfo();
					image = (classInfo == null ? null : getImage(classInfo, size, false));
					if (image != null)
						return image;
				}

				String name = objectInfo.getName();
				char   c = name.charAt(0);

				name = environment.translate(name);
				if (name.length() > 0)
                {
					name = LyTools.removeAccents(name);
                    c = name.charAt(0);
                }
                if(Character.isDigit(c))
                {
                	image = "LY_DIGIT_" + c;
                }
                else
	            {
					if (objectInfo instanceof LyAction)
						image = "LY_ACTION_" + Character.toUpperCase(c);
					else if (objectInfo instanceof LyClassInfo)
						image = "LY_CLASS_" + Character.toUpperCase(c);
					image = getImageFromPrefix(image);
	             }
			}
			else
				return null;
		}

		if (size > 0)
			image = getImageFromPrefix(prefix, DEFAULT_IMAGE_FORMAT, size);

		return image;
	} // end method getImage

	/**
	 * Build an image path from a prefix. (default extension, original size) (Prefix is usually
	 * specified in the data description : meta-model) It is possible to 'force' the default size of
	 * the image in the application resource LY_DEFAULT_IMAGE_HEIGHT (in resources.properties file).
	 *
	 * @param  prefix the prefix to identify the image resource
	 * @return the image path.
	 * @see    leon.misc.LyImages
	 */
	public String getImageFromPrefix(String prefix)
	{
		LyEnvironment environment = getEnvironment();
		String		  height = environment.getResource("LY_DEFAULT_IMAGE_HEIGHT");
		int			  defaultHeight = -1;

		if ((height != null) && (height.trim().length() > 0))
			defaultHeight = Integer.parseInt(height);

		return getImageFromPrefix(prefix, DEFAULT_IMAGE_FORMAT, defaultHeight, defaultHeight);
	}

	/**
	 * Build an image path from a prefix, an image extension and a specified height. (Prefix is
	 * usually specified in the data description) The image is resized, preserving the original
	 * aspect ratio. (Width is automatically calculated).
	 *
	 * @param  prefix the prefix to identify the image resource
	 * @param  format the extension associated to the image, ".gif", ".jpg" or
	 *                LyController.DEFAULT_IMAGE_FORMAT to use default value.
	 * @param  height requested height of the image.
	 * @return the image path with image size information to force the height and the width (width
	 *         is calculated with current aspect ratio from the height).
	 * @see    leon.misc.LyImages
	 */
	public final String getImageFromPrefix(String prefix, String format, int height)
	{
		return getImageFromPrefix(prefix, format, -1, height);
	}

	/**
	 * Build an image path from a prefix, extension, and size. (Prefix is usually specified in the
	 * data description)
	 *
	 * @param  prefix the prefix to identify the image resource
	 * @param  format the extension associated to the image, ".gif", ".jpg" or
	 *                LyController.DEFAULT_IMAGE_FORMAT to use default value.
	 * @param  width  requested height of the image.
	 * @param  height requested height of the image.
	 * @return the image path with image size information to force the size (width, height)
	 * @see    leon.misc.LyImages
	 */
	public String getImageFromPrefix(String prefix, String format, int width, int height)
	{
		StringBuffer  buffer = new StringBuffer();

		// looking for the filename of the image if it exists
		LyEnvironment environment = getEnvironment();
		String		  filename = environment.getImageFileName(prefix);

		// Adding the filename information
		if (filename != null)
			buffer.append(filename);
		else
		{
			// Don't add the default extension if there is already an extension
			buffer.append(prefix);
			if ((format != null) && (prefix.indexOf('.') < 0))
				buffer.append(format);
		}

		// adding the size information
		if (height > 0)
		{
			LyString string = environment.getString();

			buffer.append(string.VISIBLE_SEP3);
			if (width > 0)
			{
				buffer.append(width);
				buffer.append(string.VISIBLE_SEP2);
			}

			buffer.append(height);
			buffer.append(string.VISIBLE_SEP3_BIS);
		}

		return buffer.toString();
	} // end method getImageFromPrefix

	/**
	 * Factory for a LyObjectId instance for the specified String identifier This is a shortcut to :
	 * new LyObjectId(getApplication(), id).
	 *
	 * @param  id object identifier.
	 * @return the objectid instance. Never returns null.
	 */
	public LyObjectId getObjectId(String id)
	{
		LyApplication application = getApplication();

		return new LyObjectId(application, id);
	}

	/**
	 * Show a popup message warning box with the specified message. The message may be a text
	 * message, or the key to a String in the application disctionnary. (stringsXXX.properties
	 * files) This is a shortcut to getViewManager().showWarning()
	 * @param message the message to display
	 *
	 * @see leon.view.LyViewManager
	 * @see leon.misc.LyStrings
	 */
	public void showWarning(String message)
	{
		getViewManager().showWarning(getView(), message);
	}

	/**
	 * Show a popup message error box with the specified message. The message may be a text message,
	 * or the key to a String in the application disctionnary. (stringsXXX.properties files) This is
	 * a shortcut to getViewManager().showError()
	 * @param message the message to display
	 * @see leon.view.LyViewManager
	 * @see leon.misc.LyStrings
	 */
	public void showError(String message)
	{
		LyViewManager manager = getViewManager();
		if (manager != null)
			manager.showError(getView(), message);
	}

	/**
	 * Show an error for the specified object. If object is null or if the resource
	 * LY_SHOWN_ERRORS_ON_OBJECTS is set to true, the error message is shown in a popup window else
	 * the specified error is added to the object.
	 *
	 * @param object The object in error
	 * @param error  The error o show
	 * @see   #showError(String)
	 */
	public void showError(LyObject object, LyError error)
	{
		LyEnvironment environment = getEnvironment();

		if (error == null)
		{
			environment.logError("BF AppEngine : null error thrown on controller " + this);
			return;
		}

		String   message = error.getMessage();
		Object[] args = error.getArgs();

		message = environment.getMessage(message, args);
		showError(message);
	}

	/**
	 * Clears all cached data for this controller (especially the view and the component).
	 *
	 * <P>This method may be invoked for memory optimization, but the controller is still valid and
	 * may be used. (Unlike free) Further usage, needs to rebuild internal component & view which
	 * may be time consuming.</P>
	 */
	public void clearViewCache()
	{
		// Can not unregister from listening : thin client pages may be found in
		// cache and invoke actions even if they have been hidden.
		if (_component != null)
			_component.free();
		_component = null;
		_builder = null;
	}

	/**
	 * Method that should be called when starting the run method when the controller is Runnable.
	 * This method allows to associate the new thread with the current session. The endRun method
	 * must absolutely be called to allow thread to be collected by gc at the end of run method.
	 *
	 * @see #endRun
	 */
	protected void startRun()
	{
		// Associate Thread to session on application
		LyApplication application = getApplication();

		application.setCurrentSession(_session);
	}

	/**
	 * Method that should be called when ending the run method when the controller is Runnable. This
	 * method allows to remove association beetween the thread that was runned and the current
	 * session. This method must absolutely be called to allow thread to be collected by gc at the
	 * end of run method.
	 *
	 * @see #startRun
	 */
	protected void endRun()
	{
		// Associate Thread to session on application
		LyApplication application = getApplication();

		application.setCurrentSession(null);
	}

	/**
	 * Show errors view.
	 */
	protected void showErrors()
	{
		if (_parent != null)
			_parent.showErrors();
		else
		{
			// try with the parent application
			LyDataSession parentSession = getSession().getParent();

			if (parentSession instanceof LySession)
			{
				LyController controller = ((LySession)parentSession).getLastViewController();

				if (controller == null)
					return;

				while (controller.getParent() != null)
					controller = controller.getParent();

				controller.showErrors();
			}
		}
	}

	/**
	 * Sets idle status of the table.
	 *
	 * @param idle a boolean indicating the idle mode
	 */
	public void setIdle(boolean idle)
	{
		_idle = idle;
	}

	/**
	 * Check idle status of the table.
	 *
	 * @return boolean indicating the idle mode
	 */
	public boolean isIdle()
	{
		return _idle || _session.isIdle();
	}

	/**
	 * Check the controller have a viewController.
	 *
	 * @return boolean if the controller has a viewController
	 */
	public boolean hasViewController()
	{
		return true;
	}

    /**
     * Retrieves a value from a registered key.
     *
     * @param   key  The key that must be used.
     * @return  The value stored with the key or null if the key is not registered.
     */
    public Object getKeyValue(String key)
    {
        if (_parent != null)
            return _parent.getKeyValue(key);
        else
            return _session.getKeyValue(key);
    }

    /**
     * Gets user data  : data stored on controller by application.
     * @return Object stored as user data.
     */
    public final Object getUserData()
    {
        return _userData;
    }

    /**
     * Sets user data  : data stored on controller by application.
     */
    public final void setUserData(Object userData)
    {
        _userData = userData;
    }
} // end class LyController
