/**
 * LyCreateController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyCreateController.java 33237 2011-11-22 10:09:33Z nhamon $
 */

package leon.control;

import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;


/**
 * This class describes the controller associated to a creation form.
 *
 * <P>This class implements a 'read-write' LyFormController and specializes methods from its parent.
 * </P>
 *
 * <P>This form controller may have an object as a model (in that case the action is a creation
 * ('clone') keeping some current values of the model. (Creation by copy).</P>
 *
 * <P>Usually the creation form is not associated with any object but specified a LyClassInfo
 * parameter : the object will be created in this application class.</P>
 *
 * <P>Services supported by the LyCreateController :</P>
 *
 * <UL>
 *   <LI>Do not copy fields marked LyFieldInfo.MARK_PRIVATE or LyFieldInfo.MARK_SECRET.</LI>
 *   <LI>Uses a LyCreateBuilder (instead of LySetBuilder).</LI>
 *   <LI>Associated to a LyCreateBehaviorInterface, usually a LyCLassBehavior instance</LI>
 * </UL>
 *
 * @see leon.app.behavior.LyClassBehavior
 * @see LyConsultController
 * @see leon.control.builder.LySetBuilder
 * @see leon.control.builder.LyCreateBuilder
 * @see leon.app.behaviorinterface.LyCreateBehaviorInterface
 */
public class LyCreateController extends LySetController
{
	/**
	 * File version
	 */
	public static final String  __VERSION =
		"$Id: LyCreateController.java 33237 2011-11-22 10:09:33Z nhamon $";

	/** The model for a clone action. */
	protected LyObject		    _model;

	/**
	 * The creation is requested only if at least one field has been modified.
	 */
	protected boolean		    _createIfModified;

	/**  */
	protected LySimpleWorkSpace _workSpace;

	/**
	 * Flag indicating if this controller is used as an internal controller in a
	 * LyEditTableSetController.
	 */
	private boolean			    _internal;

	/**
	 * Last Object (if any) that was created by this controller. This is used for multiple creation
	 * to get last creation. unlike, _object and _newObject, the last created object is no longer
	 * used by this controller and is available until this controller is freed.
	 */
	protected LyObject		    _lastCreatedObject;

	/**
	 * Flag indicating if this controller is currently updating with the updateObject(LyValueSet)
	 * method to avoid stack overflow.
	 */
	boolean					    _updating = false;

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller).
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo this parameter is not used (introspection mechanisms).
	 * @param objects   model used as the base of the creation form. Only the first object in the
	 *                  list is considered. (creation by copy)
	 */
	public LyCreateController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		super(session, parent, action,
			((action.getClassInfo() != null) ? action.getClassInfo() : classInfo),
			((action.getClassInfo() != null) ? new LySimpleObjectList() : objects));

		if ((objects != null)
				&& (action.getClassInfo() != null))
		{
			LyInfoList<LyClassInfo> classes = objects.getClasses();

			if ((classes != null)
					&& (classes.getSize() == 1)
					&& (classes.getInfo(0) == action.getClassInfo()))
				setObjects(objects);
			else
				setContext(objects);
		}
		else if (objects == null)
			setObject(null);

		_createIfModified = false;
		_internal = false;
	}

	/**
	 * Internal constructor used by edit table controllers.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller).
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo this parameter is not used (introspection mechanisms).
	 * @param object    The object to create.
	 * @param context   The context to use.
	 */
	protected LyCreateController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObject object, LyContext context)
	{
		super(session, parent, action, classInfo, null);
		_controllerContext = context;
		_object = object;
		_internal = true;
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyCreateBehaviorInterface.class;
	}

	/**
	 * Free this controller on request of the specified controller.
	 *
	 * @param controller controller instance which requested the freeing of this controller. If the
	 *                   controller is the parent, this controller is freed with all its
	 *                   sub-controllers (relationControllers) otherwise the method does nothing.
	 */
	public void free(LyController controller)
	{
		if (_parent != controller)
			return;

		clearErrors();

		if (_workSpace != null)
		{
			try
			{
				_workSpace.free(this);
			}
			catch (Exception e)
			{
				getEnvironment().writeException(e);
			}
			_workSpace = null;
		}

		_model = null;
		_lastCreatedObject = null; // freed

		super.free(controller);
	}

	/**
	 * Free extra errors.
	 */
	protected void clearErrors()
	{
		if (_errors == null)
			return;
        
		int			   i, n = _errors.size();
		LyError		   error;
		LyClassInfo    classInfo = getClassInfo();
		LyObjectSource source = classInfo.getObjectSource();

		for (i = 0; i < n; i++)
		{
			error = _errors.elementAt(i);
			source.removeError(error, null);
		}

		_errors = null;
	}

	/**
	 * Accessor to the reference to the parent controller.
	 *
	 * @return the parent controller instance or null for root controllers.
	 */
	public LyController getParent()
	{
		if (_internal)
			return _parent;
		else
			return super.getParent();
	}

	/**
	 * Sets current model for the creation.
	 *
	 * <P>Reset default values of the form.</P>
	 *
	 * @param  object new object being modified.
	 * @return true if the object change was successful, false otherwise.
	 */
	public boolean setObject(LyObject object)
	{
		if ((object != null)
				&& (_object == object))
			return false;

		if (_ending
				&& !isApplied()
				&& !(_parent instanceof LyFormController))
		{
			// No need to update the view will be hidden
			_object = object;
			_model = object;
			return true;
		}

		//FFT00517
		if (_started
             && isApplied()
             && (!(_parent instanceof LySetController)
                 || ((LySetController)_parent).isCreation()))
		{
			// Previous object is now the model
			if (object == null)
			{
				object = _object;
				_object = null;
			}
			else
				_object = object;

			// End previous action
			if (!endAction())
				return false;

			// Start new action
			if (!startAction())
			{
				free();
				return false;
			}
		}

		_model = object;

		// Set the default value for all the fields that have the PRIVATE or SECRET mark
		LyClassInfo classInfo = getClassInfo();

		if (classInfo == null)
		{
			if (object == null)
			{
				_object = null;
				return false;
			}
			
			classInfo = object.getClassInfo();
		}

		LyValueSet  values = new LyValueSet();
		LyFieldInfo fieldInfo;
		int		    i, n = classInfo.getFieldInfoCount();
		Object	    value = null;

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			if (contains(fieldInfo)) // if the field is in the form, update the value.
			{
				if ((_model == null)
						|| fieldInfo.hasMark(LyFieldInfo.MARK_PRIVATE)
						|| fieldInfo.hasMark(LyFieldInfo.MARK_SECRET))
					value = getDefaultValue(fieldInfo);
				else
					// New value is model value
					value = _model.getValue(fieldInfo);

				if ((value != null) || isApplied())
					values.addValue(fieldInfo, value);
			}
		}

		_object = null;
		updateObject(values);
		return true;
	} // end method setObject

	/**
	 * Called when current object has changed . Keep the overall content consistent.
	 *
	 * @param oldValues set of former values, or null if not specified.
	 */
	public void updateObject(LyValueSet oldValues)
	{
		super.updateObject(oldValues);

		if (_object == null)
			return;
        if (!_initialized)
            return;

		// To avoid stack overflow.
		if (_updating)
			return;
		_updating = true;

		LyClassInfo classInfo = getClassInfo();
		LyValueSet  values = null;
		LyFieldInfo fieldInfo;
		LyValue     fieldValue;
		int		    i,
					n = ((oldValues != null) ? oldValues.getSize() : classInfo.getFieldInfoCount());
		Object	    newValue, oldValue;

		for (i = 0; i < n; i++)
		{
			if (oldValues != null)
			{
				fieldValue = oldValues.getFieldValue(i);
				fieldInfo = (LyFieldInfo)fieldValue.getField();
			}
			else
				fieldInfo = classInfo.getFieldInfo(i);

			if (contains(fieldInfo))
			{
				oldValue = _object.getValue(fieldInfo);
				newValue = getValue(fieldInfo);
				if (!fieldInfo.equals(newValue, oldValue))
				{
					if (values == null)
						values = new LyValueSet();
					values.addValue(fieldInfo, newValue);
				}
			}
		}

		_updating = false;
	} // end method updateObject

	/**
	 * Internal use : update the value of a fieldInfo in the form. Application should use
	 * setFieldValue instead. Overriden to set default values from the model, and to skip fieldInfos
	 * marked as LyFieldInfo.MARK_PRIVATE or LyFieldInfo.MARK_SECRET
	 *
	 * @param fieldInfo the corresponding field info.
	 * @param newValue  the new value of the field. This value may be null, a LyValue instance or
	 *                  the internal value associated to the fieldInfo. @See
	 *                  leon.data.LyValue#getValue()
	 * @param oldValue  the latest value of the field. This value may be null, a LyValue instance or
	 *                  the internal value associated to the fieldInfo. @See
	 *                  leon.data.LyValue#getValue()
	 * @param propagate Indicates whether the change is notified (propagated) to other fields or
	 *                  not. This is useless if you reset all fields for instance.
	 * @see   #setFieldValue
	 */
	protected void updateFieldItem(LyFieldInfo fieldInfo,
		Object newValue, Object oldValue, boolean propagate)
	{
		// This field is being updated from GUI in valueChanged()
		if (_valueChanging == fieldInfo)
			return;

		// In applyCreation mode only values that are different are updated.
		if (_ending && (fieldInfo.equals(newValue, oldValue)))
		{
			// FFT00531 : if the field is not null, a value may have been set automatically
			// ensure we report this in the form
			if (fieldInfo.hasMark(LyFieldInfo.MARK_NOT_NULL))
			{
				Object formValue = getFieldValue(fieldInfo).getValue();

				if (fieldInfo.equals(newValue, formValue))
					return; // Nothing changed
			}
			else
				return; // Nothing changed
		}

		if ((!_initialized || isApplied())
				&& (getModel() != null)
				&& (fieldInfo.hasMark(LyFieldInfo.MARK_PRIVATE)
					|| fieldInfo.hasMark(LyFieldInfo.MARK_SECRET))
				&& (oldValue == null))
			newValue = getDefaultValue(fieldInfo);

		super.updateFieldItem(fieldInfo, newValue, oldValue, propagate);
	} // end method updateFieldItem

	/**
	 * Start validation of the form. The form is closing and we check the values typed in by the
	 * user.
	 *
	 * <P>When closing the validate method is following these steps :</P>
	 *
	 * <UL>
	 *   <LI>Invoke the LyClassBehavior.addExtraValues (LySetBehaviorInterface) to let the
	 *     application fill in additional or calculated values to the result.</LI>
	 *   <LI>Validation of all the values using LyClassBehavior.controlNewValues</LI>
	 * </UL>
	 * @return the status of the controller
	 *
	 * @see leon.app.behavior.LyClassBehavior#addExtraValues
	 * @see leon.app.behavior.LyClassBehavior#controlNewValues
	 */
	public short startValidate()
	{
		// check if the validation makes sense (anything modified)
		if (_createIfModified)
		{
			if (!checkIfModified())
				return LyFormInterface.STATUS_IGNORE; // No change
		}

		return super.startValidate();
	}

	/**
	 * End validation of the form : commit edition.
	 * @return The status of the controller
	 */
	public short endValidate()
	{
		if (_createIfModified)
		{
			if (!checkIfModified())
				return LyFormInterface.STATUS_IGNORE; // No change
		}

		return super.endValidate();
	}

	/**
	 * Check if the form content has been modified and if the object need to be created.
	 * @return true if the form content was modified
	 */
	public boolean checkIfModified()
	{
		// Check if the validation makes sense (anything modified)
		return (_createIfModified && super.checkIfModified());
	}

	/**
	 * End of the current action. The creation completed, and the application may take control and
	 * continue or give up action. If false, the process of closing the window aborts.
	 *
	 * <P>The method returns LyClassBehavior.endCreation() (LyCreateBehaviorInterface)</P>
	 *
	 * @param  values Values returned by the controlled form.
	 * @return True if the action may end, false otherwise.
	 */
	protected boolean end(LyValueSet values)
	{
		LyCreateBehaviorInterface behavior = (LyCreateBehaviorInterface)getBehavior();
		boolean					  result;

		_ending = true;
		result = behavior.endCreation(this, values);
		updateStatus();
		_ending = false;
		return result;
	}

	/**
	 * Method called when all modifications are committed.
	 */
	public void commit()
	{
		if (_createIfModified)
		{
			if (!checkIfModified())
			{
				cancel(); // No change
				return;
			}
		}

		super.commit();

		if ((_controllerContext != null)
				&& (_controllerContext.getStatus() == LyContext.Status.ROLLBACKED))
			_object = null;
		
        // Register a metric for object creation if needed
		if (_object != null)
		{
			LyMetricsTools.createMetric(getSession(), LyMetricsTools.MetricType.DATA_CREATION, new Date(),
					null, null, _action.getId(), _object.getClassInfo().getId(), _object.getObjectId().getFullString());
		}
	}

	/**
	 * Commit sub-forms.
	 */
	public void subFormsCommit()
	{
		if (_createIfModified)
		{
			if (!checkIfModified())
				return;
		}
		super.subFormsCommit();
	}

	/**
	 * Factory of the default builder for a controller. Overriden to create an instance of
	 * LyCreateBuilder
	 *
	 * @return the builder which creates the XML description of the view
	 * @see    leon.control.builder.LyCreateBuilder
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return (new LyCreateBuilder(this));
	}

	/**
	 * Returns a boolean indicating if the action is or not a creation.
	 *
	 * @return true if the form is a creation form. Always return TRUE.
	 */
	public boolean isCreation()
	{
		return true;
	}

	/**
	 * Returns the objects of the form for a given context.
	 *
	 * @param  context The new context.
	 * @return The objects of the form for a given context.
	 */
	protected LyObjectList getObjectsFromContext(LyWorkSpace context)
	{
		// For a creation form, there are no objects for a given context
		// excepting if the form manages a composition relation.
		if (_compositionRelation != null)
			return super.getObjectsFromContext(context);
		else
			return null;
	}

	/**
	 * Accessor to the list of objects associated with this form. This list may be null.
	 *
	 * @return the list of objects for this controller
	 */
	public LyObjectList getObjectList()
	{
		return ((_object == null) ? null : new LySimpleObjectList(_object));
	}

	/**
	 * The creation is requested only if at least one field has been modified.
	 *
	 * @return current status
	 */
	public boolean isCreateIfModified()
	{
		return _createIfModified;
	}

	/**
	 * The creation is requested only if at least one field has been modified.
	 *
	 * @param b new status
	 */
	public void setCreateIfModified(boolean b)
	{
		_createIfModified = b;
	}

	/**
	 * Set current controller context.
	 *
	 * @param context context or null if none is specified
	 * @return true if successfull
	 * @see   LyContextInterface#getContext
	 */
	public boolean setContext(LyWorkSpace context)
	{
        if (isApplied())
            return super.setContext(context);
        else
            return changeContext(context);
	}

	/**
	 * Accessor to the model for a clone operation.
	 * @return the model object
	 */
	public LyObject getModel()
	{
		return _model;
	}

	/**
	 * The form is closing and we check the values typed in by the user.
	 *
	 * <P>This method is using the LyClassBehavior (LySetBehaviorInterface) to complete and validate
	 * the form.</P>
	 *
	 * <P>When closing the validate method is following these steps :</P>
	 *
	 * <UL>
	 *   <LI>Invoke the LyClassBehavior.addExtraValues (LySetBehaviorInterface) to let the
	 *     application fill in additional or calculated values to the result.</LI>
	 *   <LI>Validation of all the values using LyClassBehavior.controlNewValues</LI>
	 *   <LI>Invoke the end method (which will invoke LyClassBehavior.endSet</LI>
	 * </UL>
	 *
	 * @param free indicates if the form is closed and freed after a successful validation.
	 * @return the status of the controller
	 * @see   leon.app.behavior.LyClassBehavior#addExtraValues
	 * @see   leon.app.behavior.LyClassBehavior#controlNewValues
	 * @see   leon.app.behavior.LyClassBehavior#endSet
	 */
	public short validate(boolean free)
	{
		short status = super.validate(free);

		_lastCreatedObject = _object;

		if (!free
		    && isApplied()
			&& (status == LyFormInterface.STATUS_OK)
			&& !(_parent instanceof LyTabAreaController))
		{
			_changingObject = true;
			setObject(null);
			_changingObject = false;
		}

		return status;
	}

	/**
	 * Called when an object is removed from the dynamic object list. Keep content consistent : go
	 * to next available object if any or close the view.
	 *
	 * @param  event event of removing object
	 * @return the event is used by the listener
	 */
	public boolean objectRemoved(LyWorkSpaceEvent event)
	{
		if (_update // Don't free when cancelling
				&& (_object == event.getObject()))
			free();
		return true;
	}

	/**
	 * Show an error for the specified object.
	 *
	 * @param object The object in error
	 * @param error  The error to show
	 */
	public void showError(LyObject object, LyError error)
	{
		if ((_component == null)
				&& (_parent != null))
		{
			_parent.showError(object, error);
			return;
		}

		super.showError(object, error);

		if ((object == null)
				&& (_errors != null))
		{
			LyClassInfo    classInfo = getClassInfo();
			LyObjectSource source = classInfo.getObjectSource();

			error.setTransmitter(this);
			source.raiseError(error, null);
		}
	} // end method showError

	/**
	 * Reset fieldInfo errors on _object for fieldInfo.
	 *
	 * @param fieldInfo
	 */
	public void resetFieldInfoErrors(LyFieldInfo fieldInfo)
	{
        if ((_object != null)
            || (_errors == null))
		{
			super.resetFieldInfoErrors(fieldInfo);
			return;
		}
		
		int			   i, n = _errors.size();
		LyError		   error;
		LyClassInfo    classInfo = getClassInfo();
		LyObjectSource source = classInfo.getObjectSource();

		for (i = 0; i < n; i++)
		{
			error = _errors.elementAt(i);
			if ((error.getOrigin() == fieldInfo)
					|| ((error.getFieldInfosCause() != null)
						&& error.getFieldInfosCause().contains(fieldInfo)))
				source.removeError(error, null);
		}
	}

	/**
	 * Controls current edition.
	 *
	 * @return The error status.
	 */
	public short control()
	{
		if (!checkIfModified())
			return STATUS_OK;
		else
			return super.control();
	}

	/**
	 * Method called when a context is ended by a brother context.
	 *
	 * @param context Context ended.
	 * @param brother Brother context that ended the given context.
	 */
	public void contextEnded(LyContext context, LyContext brother)
	{
		if ((_controllerContext == context)
				&& (getParent() instanceof LyViewController))
			free();
	}

	/**
	 * Sets the form content from the specified object : all field controllers and form fields take
	 * the specified value. Any value modified by the user is the form is lost. Unlike setObject()
	 * and updateFielditem(), no control is done to prevent overriding existing values.
	 *
	 * @param object         new object used to reset the view content. If the param is null, a
	 *                       clone of the object is created.
	 * @param freeOldValue   indicates if the old value should be freed (deleted) when it is an
	 *                       object
	 * @param createNewValue indicates if a new empty value should created when object is null and
	 *                       when the value is an object
	 */
	public void resetFromObject(LyObject object, boolean freeOldValue, boolean createNewValue)
	{
		// recurse
		super.resetFromObject(object, freeOldValue, createNewValue);

		updateObject(null); // Check for consistency of struct objects
	}

	/**
	 * Last Object (if any) that was created by this controller. This is used for multiple creation
	 * to get last creation. unlike, _object and _newObject, the last created object is no longer
	 * used by this controller and is available until this controller is freed.
	 * @return the last created object (by this controller)
	 */
	public LyObject getLastCreatedObject()
	{
		return _lastCreatedObject;
	}

	/**
	 * Get default fields to display in form. This method all fields of the current class info
	 * expect hidden fields.
	 *
	 * @return List of fields to be displayed.
	 */
	protected LyInfoList<LyFieldInfo> getDefaultFields()
	{
        LyEnvironment           environment = getEnvironment();
        
        if (!environment.testResource("LY_DEFAULT_FIELDS_IN_FORMS_WHEN_EMPTY"))
            return null;
        
        LyInfoList<LyFieldInfo> fields = null;
		LyFieldInfo             fieldInfo;
	    int                     i, n = _classInfo.getFieldInfoCount();

		for (i = 0; i < n; i++)
		{
		    fieldInfo = _classInfo.getFieldInfo(i);
		    if (hasMark(fieldInfo, LyFieldInfo.MARK_HIDDEN))
		        // Remove hidden fields
		        continue;
		    if (hasMark(fieldInfo, LyFieldInfo.MARK_CREATE_CONSULT))
		        // Remove createConsult fields
		        continue;
            if (hasNotMark(fieldInfo, LyFieldInfo.MARK_CREATE))
                // Remove "notCreate" fields
                continue;
	        if (fieldInfo.isLocal()
	                && (fieldInfo.getImport() != null))
	            // Remove imported fields
	            continue;
	        
		    if (fields == null)
		        fields = new LyInfoList<LyFieldInfo>();
		    fields.addInfo(fieldInfo);
		}

		return fields;
	}
} // end class LyCreateController
