/**
 * LyStickerController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyStickerController.java 28245 2010-10-19 12:45:46Z bangiole $
 */

package leon.control;

import leon.app.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * This class implements a manager for stickers.
 */
public class LyCreateStickerController extends LyAreaController
{
	public static final String __VERSION =
		"$Id: LyStickerController.java 28245 2010-10-19 12:45:46Z bangiole $";

	/** Sticker instance managed by this controller. */
	protected LySticker		_sticker;

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo This parameter is not used. Kept from introspection compatibility.
	 * @param context   current context (may be null if the view is not contextual to a list of
	 *                  objects).
	 */
	public LyCreateStickerController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent, action, null);

		String	stickerId = (String)action.getParameter("_stickerId", true);

		// If a sticker id is given, retrieve sticker instance on application
		if ((stickerId != null) && (stickerId.length() > 0))
		{
			_sticker = (LySticker)getApplication().getInfo(stickerId);
		}
		else
		{
			// Create a new sticker and save its definition
			LyObject	user = session.getUser();
			String		userId = (user == null ? null : user.getObjectId().getFullString());

			_sticker = new LySticker("sticker_" + LyString.getRandomId(), userId, null, null, null, 300, 100, 200, 200);
			session.getApplication().saveExtraInfo(_sticker);
		}

		_sticker.setApplication(session.getApplication());

		// Register this controller to receive user demands
		LyViewManager viewManager = getViewManager();

		viewManager.addDemandListener(this);
	}

	/**
	 * Accessor to the sticker definition.
	 */
	public LySticker getSticker()
	{
		return _sticker;
	}

	/**
	 * Accessor to the view controller (Frame or dialog controller) which is the container of this
	 * view. This method must be implemented in subclasses depending on the nature of the component.
	 *
	 * @return the reference to a view controller, never returns null if the controller is
	 *         associated to a view.
	 */
	public LyViewController getViewController()
	{
		LyViewController	viewController = super.getViewController();
		LyDialog			dialog = (LyDialog)viewController.getView();

		viewController.setLocation(_sticker.getX(), _sticker.getY());
		dialog.setX(_sticker.getX());
		dialog.setY(_sticker.getY());
		dialog.setWidth(_sticker.getWidth());
		dialog.setHeight(_sticker.getHeight());
		dialog.setDecorated(false);
		dialog.setTarget("sticker_target");

		return viewController;
	}

	/**
	 * Factory of the default builder for a controller. The application should never invoke this
	 * method directly (internal use only : use getBuilder instead). This method is specialized in
	 * subclasses of LyController to return a valid builder instance.
	 *
	 * @return the builder which creates the XML description of the view
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return (new LyCreateStickerBuilder(this));
	}

	@Override
	public void actionDemand(LyActionDemand demand)
	{
		LyApplication	application = getApplication();
		LyComponent		component = getComponent();
		LyDialog		dialog = (LyDialog)component.getParentView();
		LyNotice		notice = (LyNotice)component.getComponent(_sticker.getId() + "_notice");

		if (LyViewController.DEMAND_CLOSE.equals(demand.getSource()))
		{
			application.deleteExtraInfo(_sticker);
			free();
			return;
		}
		else if ("move".equals(demand.getSource()))
		{
			String		value = demand.getValue();
			String[]	position = LyString.stringToList(value, '|');

			_sticker.setX(Integer.parseInt(position[0]));
			_sticker.setY(Integer.parseInt(position[1]));
			dialog.setX(Integer.parseInt(position[0]));
			dialog.setY(Integer.parseInt(position[1]));
			application.saveExtraInfo(_sticker);
			return;
		}
		else if ("resize".equals(demand.getSource()))
		{
			String		value = demand.getValue();
			String[]	size = LyString.stringToList(value, '|');

			_sticker.setWidth(Integer.parseInt(size[0]));
			_sticker.setHeight(Integer.parseInt(size[1]));
			dialog.setWidth(Integer.parseInt(size[0]));
			dialog.setHeight(Integer.parseInt(size[1]));
			application.saveExtraInfo(_sticker);
			return;
		}
		else if ((_sticker.getId() + "_notice").equals(demand.getSource()))
		{
			LyLabel		label = notice.getLabel();

			label.setLabel(demand.getValue());
			_sticker.setContent(demand.getValue());
			application.saveExtraInfo(_sticker);
			return;
		}

		super.actionDemand(demand);
	}
} // end class LyUserInputController
