/**
 * LyDashboardController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyDashboardController.java 41239 2012-10-16 12:46:40Z nhamon $
 */

package leon.control;

import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.graphic.*;
import leon.view.mobile.*;
import leon.view.setinterface.*;


/**
 * This class implements a dashboard view, usually associated to a '_dashboard' action in the
 * meta-model.
 *
 * <P>A Dashboard view is a 'toolbar' view which contains a set of other actions, or sub-dashboard
 * views (popup dashboards).</P>
 *
 * <P>The user may activate an action from the dashboard by a click on any button of this view.
 * Further click on the same action may either re-open existing window or create a new one if former
 * windows was freed :</P>
 *
 * @see LyViewController#setFreeOnClose for additional information.
 *
 *      <P>Additionally, the view supports image buttons, with or without labels and a small tooltip
 *      info on actions. Current implementation uses a LyDashboardBehaviorInterface (usually
 *      implemented by a LyDashboardBehavior) and uses a LyDashboardBuilder to create and set view
 *      content.</P>
 * @see leon.app.behavior.LyDashboardBehavior
 * @see leon.app.behaviorinterface.LyDashboardBehaviorInterface
 * @see leon.control.builder.LyDashboardBuilder
 */
public class LyDashboardController extends LyMultiAreaController
{
	/**
	 * File version
	 */
	public static final String					  __VERSION =
		"$Id: LyDashboardController.java 41239 2012-10-16 12:46:40Z nhamon $";

	/** Identifier for modifying filter. */
	public static final String					  DEMAND_CLOSE_ALL = "_closeAll".intern();

	/** Identifier for search tool. */
	public static final String					  DEMAND_SEARCH = "_search".intern();

	/** Internal prefix for tab identifiers. */
	public static final String					  TAB_ID_PREFIX = "_tab".intern();

	/** Identifier for search class. */
	public static final String					  SEARCH_CLASS = "_searchClass".intern();

	/** Identifier for search value. */
	public static final String					  SEARCH_VALUE = "_searchValue".intern();

	/** Search limit. */
	public static int					  	  	  SEARCH_LIMIT = 10;

	/** Content of the dashboard (list of actions). */
	protected LyInfoList<LyAction>				  _actions;

	/** List of view controllers opened from this controller. */
	protected Hashtable<String, LyViewController> _viewControllers;

	/** Currently displayed area controller (single view mode only). */
	protected LyAreaController					  _displayedAreaController;

	/** Indicates if a non-displayed view is being closed. */
	boolean										  _isClosing;

	/**
	 * Indicates if area controllers must be freed on close (single view mode only). The keys of
	 * this table are instances of LyAction. The values of this table boolean indicating if
	 * corresponding controllers must be freed when closing.
	 */
	protected Hashtable<LyAction, Boolean>		  _freeAreaControllers;

	/** Stack of actions that have been displayed in the dashboard view (single view mode only). */
	protected Stack<LyAction>					  _displayedActions;

	/** List of actions that have been opened in the dashboard view (single view mode only). */
	protected LyInfoList<LyAction>				  _openedActions;

	/** Boolean indicating if we have a singleview. */
	protected boolean							  _useSingleView;

	/** Instance of LyTabs used to display action in tabs (_useTabs mode only). */
	protected LyTabs							  _tabs;

	/** Indicates whether a tab area controller is used to display opened actions. */
	protected boolean							  _useTabAreaController;

	/** Tab area controller used to display opened actions. */
	protected LyTabAreaController				  _tabAreaController;

	protected String							  _searchClass;

	protected String							  _searchValue;

    /** Right border panel controller. */
    protected LyAreaController 					  _rightBorderPanelController;

    /** Social action. */
    protected LyAction							  _socialAction;


	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 * @param action  related action. If an action is specified, the controller uses action
	 *                parameters (action name, action specific XML view description ...)
	 * @param classInfo the associated classInfo
	 * @param objects This parameter is kept for compatibility and introspection (unused).
	 */
	public LyDashboardController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		super(session, parent, action, objects);

		_actions = action.getActions();
		_isClosing = false;

		if (_actions == null)
			setName("LY_APPLICATION");

		_actions = ((LyDashboardBehaviorInterface)getBehavior()).getActions(this, _actions);
		_viewControllers = null;
		_displayedAreaController = null;

        String	useTabs = (String)_action.getParameter("_useTabs", true);

        boolean manageDashboard = false;
        LyViewManager manager = getViewManager();
        if(getViewManager() != null)
        	manageDashboard = manager.hasProperty("MANAGE_DASHBOARD");

		_useTabAreaController = ("true".equalsIgnoreCase(useTabs) && manageDashboard);
	}

    /**
     * Free this controller.
     *
     * @param  controller  parent controller
     */
    public void free(LyController controller)
    {
        if (_parent != controller)
            return;

        if (_rightBorderPanelController != null)
        {
            if (!_rightBorderPanelController.isFreed())
                _rightBorderPanelController.free(this);

            _rightBorderPanelController = null;
        }

        super.free(controller);
    } // end method free

	/**
	 * Gets title associated with this controller.
	 *
	 * @return the title corresponding to the action.
	 */
	public String getTitle()
	{
		if (getName() != null)
			return getName();

		return super.getTitle();
	}

	/**
	 * Prepare component before showing. This method is called once (unlike reset which may be
	 * called several times), before first showing. Default implementation invokes reset.
	 */
	protected void initArea()
	{
		super.initArea();

		LyAction					 defaultAction = (LyAction)getParameter("_defaultAction", true);
		LyDashboardBehaviorInterface behavior = (LyDashboardBehaviorInterface)getBehavior();
		LyObjectList				 context = getContext();

		if (useSingleView())
		{
			// If a default action is set, display it when opening dashboard
			// (only works in single view mode).
			if (defaultAction != null)
			{
				// Call getViewController() to be sure that view controller is initialized
				// and that action will be correctly attached to this view controller.
				getViewController();
				setAction(defaultAction);

				// If use tabs, open all default actions
				if (useTabs())
				{
					int nb = 1;
					LyAction action = (LyAction) getParameter("_defaultAction" + nb, true);

					while (action != null)
					{
						// Call getViewController() to be sure that view controller is initialized
						// and that action will be correctly attached to this view controller.
						getViewController();
						setAction(action);

						nb++;
						action = (LyAction) getParameter("_defaultAction" + nb, true);
					}

					// Select the first default action tab
					setAction(defaultAction);
				}
			}
			
			if ((_action.getGeometry() == null) && !getEnvironment().testEnv("LY_SILENT_GUI"))
			{
				// If no geometry is given for dashboard action, occupy full screen
				LyController  parent = getParent();

				if (parent instanceof LyViewController)
				{
					LyView		  view = getView();
					LyViewManager viewManager = getViewManager();
					int[]		  screenSize = viewManager.getScreenSize(null);

					if ((view != null) && (screenSize != null))
					{
						view.setSize(screenSize[0], screenSize[1]);
						// FFT00435/Flandres
						view.setLocation(screenSize[2], screenSize[3]);
					}
				}
			}
		}
		else if (defaultAction != null)
		{
			String		 id = defaultAction.getId();
			LyController controller = invokeAction(defaultAction);

			if (controller != null)
			{
				LyViewController viewController = controller.getViewController();

				// View controller may be null for modal views
				if (viewController != null)
				{
					addViewController(id, null, viewController);

					if ((context != null)
							|| behavior.isFreedOnClose(this, defaultAction, viewController))
						viewController.setFreeOnClose(true);
					else
						viewController.setFreeOnClose(false);
				}
			}
		}
	} // end method initArea

	/**
	 * Reset the entire view : everything is rebuilt.
	 *
	 * @param full Boolean indicating if a full reset (including reset of datas) is requiered or not
	 *             (only reset graphical area).
	 */
	public void reset(boolean full)
	{
		updateTitleBar();

		super.reset(full);
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyDashboardBehaviorInterface.class;
	}

	/**
	 * Invoked when an action occurs. At this level, the implementation manages user interaction on
	 * tree nodes and actions buttons.
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
		{
			if (demand.getSource().equals(SEARCH_CLASS))
				_searchClass = demand.getValue();
			else if (demand.getSource().equals(SEARCH_VALUE))
				_searchValue = demand.getValue();
		}

		if (((demand.getType() == LyActionDemand.TYPE_ACTION)
				&& demand.getSource().equals(DEMAND_SEARCH))
				|| ((demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
						&& demand.getSource().equals(SEARCH_VALUE))
				|| ((demand.getType() == LyActionDemand.TYPE_ACTION)
						&& demand.getSource().equals(SEARCH_VALUE)))
		{
			if (((demand.getType() == LyActionDemand.TYPE_ACTION)
					&& demand.getSource().equals(DEMAND_SEARCH))
					|| (demand.getType() == LyActionDemand.TYPE_ACTION)
							&& demand.getSource().equals(SEARCH_VALUE))
			{
				LyParameterValues	values = demand.getParameterValues();

				if (values != null)
					_searchValue = values.getParameter(SEARCH_VALUE);
			}

			search(_searchClass, _searchValue);
			return;
		}

        // Social panel
		LyAction	socialAction = getSocialAction();

        if ((socialAction != null)
        	&& (demand.getType() == LyActionDemand.TYPE_ACTION)
        	&& (socialAction != null)
			&& demand.getSource().equals(socialAction.getId()))
        {
            showSocialPanel();
            return;
        }

		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			String	    id = demand.getSource();
			LyComponent comp = getComponent();
			LyComponent component = comp.getComponent(id);

			// If source is the id of a toolbar, the toolbar must be opened or closed.
			if (component instanceof LyToolBar)
			{
				String		  value = demand.getValue();
				Boolean		  openValue = Boolean.valueOf(value);
				boolean		  open = openValue.booleanValue();

				((LyToolBar)component).setOpen(open);

				setLastModified(new Date().getTime());

				if ((comp instanceof LyPanel) && _useTabAreaController)
					((LyPanel)comp).setVisible(true);

				return;
			}

			if (component instanceof LyTree)
				// Nothing to be done.
				return;

			if ("_tabs".equals(id))
			{
				if (_tabs != null)
				{
					int		index = demand.getIndex(0);
					LyTab	tab = _tabs.getTab(index);

					if (tab != null)
					{
						id = tab.getId();
						id = id.substring(TAB_ID_PREFIX.length());
					}
				}
			}

			// Implementation of the close all menu item
			if (DEMAND_CLOSE_ALL.equals(id))
			{
				closeAllViews();
				return;
			}

			LyApplication application = getApplication();
			boolean		  isMenuItem = false;

			if ((id != null)
				&& id.startsWith("_menuItem_"))
			{
				isMenuItem = true;
				id = id.substring("_menuItem_".length());
			}

			LyInfo info = application.getInfo(id);

			if (info instanceof LyAction)
			{
				LyViewController			 viewController = getViewController(id, null);
				LyDashboardBehaviorInterface behavior = (LyDashboardBehaviorInterface)getBehavior();
				LyAction					 rootAction = ((LyAction) info).getRoot();

				// Create a new controller for each sticker action
				if ((rootAction != null) && "_createSticker".equals(rootAction.getId()))
					viewController = null;

				if ((viewController != null)
						&& viewController.isFreed())
				{
					removeViewController(id, null);
					viewController = null;
				}

				getViewManager().setWaiting(true);

				if (!behavior.openNewView(this, (LyAction)info)
						&& ((demand.getAdditionalValue() == null)
							|| !"CTRL".equals(demand.getAdditionalValue())))
				{
					// If action is already displayed and demand comes from menubar, close its view and show
					// the previous action from the stack of displayed actions.
					if (isMenuItem
							&& (_displayedActions != null)
							&& (_displayedActions.size() > 0)
							&& (_displayedActions.peek() == info))
						closeCurrentView(false);
					else
						setAction((LyAction)info);
				}
				else
				{
					if (viewController != null)
						viewController.showView();
					else
					{
						LyAction     action = (LyAction)info;
						LyObjectList context = getContext();
						LyController controller;

						if (context == null)
							controller = invokeAction(action);
						else
							controller = invokeAction(action, context);

						if ((controller != null) && !controller.isFreed())
						{
							viewController = controller.getViewController();

							// View controller may be null for modal views or may be the same as dashboard
							// view controller for actions with a target such as _north, _south, _east, _west.
							if ((viewController != null) && (viewController != getViewController()))
							{
								addViewController(id, null, viewController);

								if ((context != null)
										|| behavior.isFreedOnClose(this, (LyAction)info, viewController))
									viewController.setFreeOnClose(true);
								else
									viewController.setFreeOnClose(false);
							}
						}
					}

                    if (viewController instanceof LyFrameController)
                        updateMenus();
				}

				// XXX
				// Correct a problem seen with the thin client on a solaris OS with
				// apache 1.3.12 and openssl:
				// When we ask for a new page using the dashboard, the base frameset is not
				// refreshed. To force the regeneration, we set the last modified date of the dashboard
				// controller equals to the last modified date of the generated viewController.
				if (viewController != null)
					setLastModified(viewController.getLastModified());

				// We must test the validity of the controller because the action may
				// have destroyed it.
				if (!isFreed())
				{
					LyViewManager vm = getViewManager();

					if (vm != null)
						vm.setWaiting(false);
				}

				return;
			}
		}
		else if (demand.getType() == LyActionDemand.TYPE_CLOSE_TAB)
		{
			// If close on a tab, close corresponding tab
			if (_tabs != null)
			{
				int				index = demand.getIndex(0);
				LyTab			tab = _tabs.getTab(index);
				String			actionId = tab.getId();

				actionId = actionId.substring(TAB_ID_PREFIX.length());

				LyApplication	application = getApplication();
				LyInfo			info = application.getInfo(actionId);

				if (info instanceof LyAction)
				{
					// The freeOnClose parameter has no effect since action is already displayed when closing its tab
					LyAction	displayedAction = _displayedActions.peek();

					_isClosing = (displayedAction != info);
					if (_isClosing)
						setAction((LyAction)info);
					closeCurrentView(false);
					_isClosing = false;
					return;
				}
			}
		}
		else if (demand.getType() == LyActionDemand.TYPE_EXPAND)
			// Tree expanding. Nothing to be done.
			return;

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Invoked when an selection notification is sent to this listener.
	 *
	 * <P>This method is called by demandPerformed and may be overriden in subclasses.</P>
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   leon.control.LyController#selectionDemand
	 */
	public void selectionDemand(LySelectionDemand demand)
	{
		LyViewManager viewManager = getViewManager();
		boolean		  perspectiveMode = viewManager.hasPerspective();
		String		  id = demand.getSelectionId(0);
		String		  perspectiveId = (perspectiveMode ? demand.getSelectionId(1) : null);

		if (id == null)
			return;

		LyTree     tree = (LyTree)getComponent().getComponent(getId() + "_tree");
		LyTreeNode node = (LyTreeNode)tree.getComponent(id);

		node.setSelected(true);

		LyApplication application = getApplication();
		LyInfo		  info = application.getInfo(id);

		if (info instanceof LyAction)
		{
			LyDashboardBehaviorInterface behavior = (LyDashboardBehaviorInterface)getBehavior();
			LyViewController			 viewController = getViewController(id, perspectiveId);
			boolean						 free = false;

			free = (viewController != null) && (viewController.isFreed());
			if ((!free) && (viewController != null))
			{
				LyComponent     comp = viewController.getComponent();
				LyViewComponent viewComp = ((comp == null) ? null : comp.getInternalView(false));

				free = (viewComp == null) || (((LyGraphicalObject)viewComp).isFreed());
			}

			if (free)
			{
				removeViewController(id, perspectiveId);
				if (!viewController.isFreed())
					viewController.free();
				viewController = null;
			}

			viewManager.setWaiting(true);

			if (!behavior.openNewView(this, (LyAction)info))
				setAction((LyAction)info);
			else
			{
				if (viewController != null)
					viewController.showView();
				else
				{
					LyAction     action = (LyAction)info;
					LyObjectList context = getContext();
					LyController controller;
					String		 sessionPerspective = getSession().getEnvironment().getResource(
							"LY_PERSPECTIVE");

					// FFT01207 Temporary change of the session perspectiveId
					if (perspectiveId != null)
						getSession().getEnvironment().putResource("LY_PERSPECTIVE", perspectiveId);

					if (context == null)
						controller = invokeAction(action);
					else
						controller = invokeAction(action, context);

					// FFT01207 Restore the perspective id of the session
					if ((perspectiveId != null) && (sessionPerspective != null))
						getSession().getEnvironment().putResource("LY_PERSPECTIVE",
							sessionPerspective);
					else if ((perspectiveId != null) && (sessionPerspective == null))
						getSession().getEnvironment().putResource("LY_PERSPECTIVE", "");

					if (controller != null)
					{
						viewController = controller.getViewController();

						// View controller may be null for modal views
						if (viewController != null)
						{
							addViewController(id, perspectiveId, viewController);

							if ((context != null)
									|| behavior.isFreedOnClose(this, action, viewController))
								viewController.setFreeOnClose(true);
							else
								viewController.setFreeOnClose(false);
						}
					}
				}
			}

			// XXX
			// Correct a problem seen with the thin client on a solaris OS with
			// apache 1.3.12 and openssl:
			// When we ask for a new page using the dashboard, the base frameset is not
			// refreshed. To force the regeneration, we set the last modified date of the dashboard
			// controller equals to the last modified date of the generated viewController.
			if (viewController != null)
				setLastModified(viewController.getLastModified());

			// We must test the validity of the controller because the action may
			// have destroyed it.
			if (!isFreed())
			{
				LyViewManager vm = getViewManager();

				if (vm != null)
					vm.setWaiting(false);
			}

			// Necessary for swing otherwise the node selection is not notified
			// if the node is selected a second time.
			node.setSelected(false);
			return;
		}

		super.selectionDemand(demand);
	} // end method selectionDemand

	/**
	 * Gets the target area : the panel where the action will be shown.
	 *
	 * @param  action Action that will be displayed in the target area.
	 * @return the target area for the given action.
	 */
	public LyArea getTargetArea(LyAction action)
	{
		String targetName = action.getTarget();
		LyArea topArea = (LyArea)getComponent();
		int    size = topArea.getPanelCount();

		if (size > 1)
		{
			// If one of the sub area id starts with the following prefix,
			// Use it as the target panel.
			LyPanel panel;
			int     i;

			for (i = 0; i < size; i++)
			{
				panel = topArea.getPanel(i);
				if ((panel instanceof LyArea)
						&& (((LyArea)panel).getId() != null)
						&& ((LyArea)panel).getId().equals(targetName))
					return (LyArea)panel;
			}
		}

		// If the panel has not been found use the first one.
		LyPanel panel = topArea.getPanel(0);
		LyArea  subArea;

		if (panel instanceof LySplitter) // showAsTree=true
			subArea = (LyArea)((LySplitter)panel).getPanel(1);
		else
			subArea = (LyArea)panel;

		return subArea;
	} // end method getTargetArea

	/**
	 * Change the main action for this controller.<br />
	 * Updates the action displayed in the dashboard view.
	 *
	 * @param action        New main action.
	 */
	public void setAction(LyAction action)
	{
		LyObjectList 				 context = getContext();
		LyDashboardBehaviorInterface behavior = (LyDashboardBehaviorInterface)getBehavior();
		boolean      				 freeOnClose = false;

		if ((context != null)
				|| behavior.isFreedOnClose(this, action, null))
			freeOnClose = true;

		setAction(action, freeOnClose);
	}

	/**
	 * Change the main action for this controller.<br />
	 * Updates the action displayed in the dashboard view.
	 *
	 * @param action        New main action.
	 * @param deleteOnClose Delete controller when action is changed.
	 */
	public void setAction(LyAction action, boolean deleteOnClose)
	{
        LySession        session = getSession();

        if (!session.validateAction(action, this))
            return;

		LyArea			 subArea = getTargetArea(action);
		LyAreaController newAreaController = getAreaController(action);

		if (_useTabAreaController && (_tabAreaController != null))
			newAreaController = _tabAreaController.getAreaController(action, true);

		if ((newAreaController == null) && (action != null))
		{
			_buildInternalSubControllers = true;
			newAreaController = (LyAreaController)invokeAction(action, null, null, false);
			_buildInternalSubControllers = false;

			// If no area controller was built, do not register it.
			if ((newAreaController == null)
					|| newAreaController.isFreed())
				return;
		}
		if (_useTabAreaController)
		{
				// If the tab area controller instance was not created or has been freed, build a new
				// one used to display opened actions.
				LyViewController	viewController;

				if (((_tabAreaController == null) || _tabAreaController.isFreed()))
				{
					LyDashboardBehaviorInterface behavior = (LyDashboardBehaviorInterface)getBehavior();
					boolean						 freeOnClose = behavior.isFreedOnClose(this, action, null);
					LyTabAreaController 		 old = _tabAreaController;
					LyTabAction					 tabAction = new LyTabAction(null, null, "leon.control.LyTabAreaController",
							"leon.app.behavior.LyTabActionBehavior", "leon.control.builder.LyTabAreaBuilder",
							(short)-1, null, LyTabAction.Type.TABS);

					tabAction.setApplication(getApplication());

					_tabAreaController = (LyTabAreaController)invokeAction(tabAction, null, null, false);
					_tabAreaController.reuseHiddenAreaControllers(old);
					viewController = _tabAreaController.getViewController();
					viewController.setName(getName());
					viewController.setFreeOnClose(freeOnClose);

					LyTabs	tabs = _tabAreaController.getTabs();

					tabs.setClosable(true);
				}

				viewController = _tabAreaController.getViewController();
				viewController.showView();
				_tabAreaController.addTabAction(newAreaController, deleteOnClose);
		}
		else if ((newAreaController == null)
					|| newAreaController.isFreed())
				return;
		else
			addAreaController(newAreaController);

		if (_freeAreaControllers == null)
			_freeAreaControllers = new Hashtable<LyAction, Boolean>();

		_freeAreaControllers.put(action, Boolean.valueOf(deleteOnClose));


		if (_useTabAreaController && (_tabAreaController != null))
		{
			_tabAreaController.selectAction(action);
			_tabAreaController.getViewController().showView();
			return;
		}

		LyViewManager viewManager = getViewManager();

		viewManager.setWaiting(true);

		if (newAreaController != _displayedAreaController)
		{
			LyArea		newArea = null;

			try
			{
				newArea = (LyArea)newAreaController.getComponent();
			}
			catch (Exception e)
			{
				getEnvironment().writeException(e);
			}
			if (newArea == null)
			{
				viewManager.setWaiting(false);
				return;
			}

			boolean		useTabs = useTabs();
			LyPanel		oldPanel = null;

			if ((subArea.getPanelCount() == 1) && !(subArea.getPanel(0) instanceof LyTabs))
			{
				oldPanel = subArea.getPanel(0);
				subArea.removePanel(oldPanel);

				if (useTabs)
				{
					LyConstraint  constraint = new LyConstraint();

					constraint.setFill(LyConstraint.Fill.BOTH);
					_tabs = new LyTabs();
					_tabs.setClosable(true);
					_tabs.setConstraint(constraint);
					subArea.addPanel(_tabs);
					subArea.setController(getId());
				}
			}

			boolean      update = false;

			if (!newAreaController.isInitialized())
			{
				if (useTabs)
				{
					LyDashboardBehaviorInterface	behavior = (LyDashboardBehaviorInterface)getBehavior();
					LyTab							tab = behavior.buildTab(this, newAreaController);

					newAreaController.initArea();
					if (newAreaController.isFreed())
					{
						viewManager.setWaiting(false);
						return;
					}
					tab.setPanel(newArea);
					_tabs.addTab(tab);
				}
				else
				{
					newAreaController.initArea();
					if (newAreaController.isFreed())
					{
						if (oldPanel != null)
							subArea.addPanel(oldPanel);
						viewManager.setWaiting(false);
						return;
					}
					subArea.addPanel(newArea);
				}
			}
			else
			{
				if (!useTabs)
				{
					subArea.addPanel(newArea);
				}
				else
				{
					LyTabs parentTabs = newAreaController.getComponent().getParentTabs();

					if (parentTabs == null)
					{
						LyDashboardBehaviorInterface	behavior = (LyDashboardBehaviorInterface)getBehavior();
						LyTab							tab = behavior.buildTab(this, newAreaController);

						tab.setPanel(newArea);
						_tabs.addTab(tab);
					}
				}
				if (!_isClosing)
					newAreaController.setIdle(false);
				update = true;
			}

			if (useTabs
			        && (oldPanel != null))
				oldPanel.free();

			if (!_isClosing
					&& (_displayedAreaController != null)
					&& (newAreaController != _displayedAreaController))
				_displayedAreaController.setIdle(true);

			_displayedAreaController = newAreaController;
			if (update)
			    _displayedAreaController.update();

			if (_displayedActions == null)
				_displayedActions = new Stack<LyAction>();

			if (_openedActions == null)
				_openedActions = new LyInfoList<LyAction>(false);

			if (_displayedActions.contains(action))
				_displayedActions.remove(action);

			_displayedActions.push(action);
			_openedActions.addInfo(action);

			if (useTabs && (_tabs != null))
				_tabs.setSelectedTab(TAB_ID_PREFIX + action.getId());

			// Update view
			updateMenus();
			updateMessageBar();

			// Notify opening to behavior
			LyDashboardBehaviorInterface behavior = (LyDashboardBehaviorInterface)getBehavior();

			behavior.openController(this, newAreaController);

			// DEBUG_XML MODE : dump view
			LyEnvironment environment = getEnvironment();

			if (environment.isDebugLevel(LyEnvironment.DEBUG_XML)
					&& !newAreaController.isFreed())
			{
				LyView view = getView();

				if (view != null)
					view.writeXml(null, true, true);
			}

			// Update the last modification time.
			setLastModified(new Date().getTime());
		}

		updateCloseTool();

		viewManager.setWaiting(false);
	} // end method setAction

	/**
	 * Close view currently being displayed in the dashboard view (single view mode only).
	 * @param closeTabs true to close all tabs
	 */
	public void closeCurrentView(boolean closeTabs)
	{
		if (_useTabAreaController && closeTabs && (_tabAreaController != null))
		{
			_tabAreaController.closeAll();
			return;
		}
		else if ((_displayedActions != null) && (_displayedActions.size() > 0))
		{
			// Remove currently displayed action
			LyAction		 oldAction = _displayedActions.pop();

			_openedActions.removeInfo(oldAction);
			closeController(_displayedAreaController);
		}
	} // end method closeCurrentView

	/**
	 * Close given controller view.
	 * @param areaController Controller to be closed.
	 */
	protected void closeController(final LyAreaController areaController)
	{
		if (_component == null)
			return;

		boolean          useTabs = useTabs();

		// Call behavior
		LyDashboardBehaviorInterface behavior = (LyDashboardBehaviorInterface)getBehavior();

		behavior.closeController(this, areaController);

		if (useTabs)
		{
			LyComponent component = areaController.getComponent();
			LyTab       currentTab = component.getParentTab();

			if ((currentTab != null) && (currentTab.getParent() == _tabs))
				_tabs.removeTab(currentTab);
		}

		if (_displayedActions.size() > 0)
		{
			// If there is at least another action that have been displayed, display
			// it again. Otherwise, display default area.
			LyAction     newAction = _displayedActions.pop();

			setAction(newAction);

			if (useTabs && !(_displayedActions.contains(newAction)))
				_displayedActions.push(newAction);
		}
		else
		{
			_displayedAreaController = null;

			LyDashboardBuilder builder = (LyDashboardBuilder)getBuilder();
			LyArea             defaultArea = builder.buildDefaultArea();
			LyArea             topArea = (LyArea)getComponent();
			LyPanel            panel = topArea.getPanel(0);
			LyArea             subArea;

			if (panel instanceof LySplitter) // showAsTree=true
				subArea = (LyArea)((LySplitter)panel).getPanel(1);
			else
				subArea = (LyArea)panel;

			if (subArea.getPanelCount() == 1)
			{
				subArea.removePanel(subArea.getPanel(0));
				subArea.addPanel(defaultArea);

				// Update view
				updateMenus();
				updateMessageBar();

				// Update the last modification time.
				setLastModified(new Date().getTime());
				updateCloseTool();
			}
		}

		// Free controller if needed
		LyAction	oldAction = areaController.getAction();
		Boolean		free = (_freeAreaControllers != null ?
				_freeAreaControllers.get(oldAction) : null);

		if ((free != null) && free.booleanValue())
			freeAreaController(areaController);
	}

	/**
	 * Close view currently being displayed in the dashboard view (single view mode only).
	 * @param areaController  the controller to be freed
	 */
	protected void freeAreaController(LyAreaController areaController)
	{
		LyAction oldAction = areaController.getAction();

		_displayedActions.remove(oldAction);
		_openedActions.removeInfo(oldAction);
		if (!areaController.isFreed())
			areaController.free(this);
		removeAreaController(areaController);
		_freeAreaControllers.remove(oldAction);
	}

	/**
	 * Closes all the views displayed in the dashboard.
	 */
	public void closeAllViews()
	{
		boolean		wasIdle = isIdle();

		setIdle(true);
		_isClosing = true;
		while ((_displayedActions != null) && (_displayedActions.size() > 0))
			closeCurrentView(false);

		getSession().closeViews();
		_isClosing = false;
		setIdle(wasIdle);
	}

	/**
	 * Update the state of the tool that closes the views.<br />
	 * This tool is enabled only if a view is displayed under the dashboard in single view mode.
	 */
	protected void updateCloseTool()
	{
		LyComponent component = getComponent();
		LyComponent closeComponent = component.getComponent(useTabs() ? DEMAND_CLOSE_ALL : DEMAND_CLOSE);

		if ((closeComponent != null) && (closeComponent instanceof LyTool))
		{
			LyTool tool = (LyTool)closeComponent;

			tool.setEnabled(_displayedAreaController != null);
		}
	}

	/**
	 * Updates the Window menu of the menu bar.
	 */
	public void updateWindowMenu()
	{
		// Recompute the content of the entire menu bar
		updateMenus();
	}

	/**
	 * Indicates whether the actions opened from this dashboard must be opened in a new view or not.
	 *
	 * @return true if the actions are opened in the same view as the dashboard, false otherwise.
	 */
	public boolean useSingleView()
	{
		LyViewManager viewManager = getViewManager();

		if(viewManager instanceof LyMobileViewManager)
		{
			return false;
		}

		if (_useTabAreaController)
			return true;

		// If dashboard is part of a compound action, do not use single view.
		//if (_parent instanceof LyCompoundAreaController)
		//    return false;

		String		  useSingleView = (String)_action.getParameter("_useSingleView", true);

		return ((useSingleView != null) && useSingleView.equalsIgnoreCase("true")
				&& !viewManager.hasProperty("MANAGE_DASHBOARD"));
	}

	/**
	 * Indicates whether the actions opened from this dashboard must be opened in tabs.<br />
	 * Used only if actions are opened in the same view as the dashboard.
	 *
	 * @return true if the actions are opened in tabs, false otherwise.
	 */
	public boolean useTabs()
	{
        String	useTabs = (String)_action.getParameter("_useTabs", true);
        boolean	useSingleView = useSingleView();

        return (useSingleView && useTabs.equalsIgnoreCase("true"));
	}

	/**
	 * Accessor to the area controller displayed in the dashboard.
	 *
	 * @return the area controller displayed in the dashboard, or null if no area controller is
	 *         displayed.
	 */
	public LyAreaController getDisplayedAreaController()
	{
		return (_useTabAreaController && (_tabAreaController != null) ?
			_tabAreaController.getSelectedAreaController() : _displayedAreaController);
	}

	/**
	 * Accessor to the list of actions opened in the dashboard.
	 *
	 * @return the list of actions opened in the dashboard, or null if no action has been opened.
	 */
	public LyInfoList<LyAction> getOpenedActions()
	{
		return _openedActions;
	}

	/**
	 * Accessor to the list of actions shown in the view.
	 *
	 * @return the list of actions
	 */
	public LyInfoList<LyAction> getActions()
	{
		return _actions;
	}

	/**
	 * Factory of the default builder for a controller. Overriden to create an instance of
	 * LyDashboardBuilder
	 *
	 * @return the builder which creates the XML description of the view
	 * @see    leon.control.builder.LyDashboardBuilder
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return (new LyDashboardBuilder(this));
	}

	/**
	 * Indicates if the view for this dashboard controller is resizable.
	 *
	 * @return false since dashboard views are not resizable.
	 */
	public boolean isViewResizable()
	{
		return useSingleView();
	}

	/**
	 * Gets image associated with this controller. The image is the one associated to the action.
	 * Overriden to keep the size of the titleBar icon
	 *
	 * @return the image corresponding to the action.
	 */
	public String getImage()
	{
		LyActionBehaviorInterface behavior = (LyActionBehaviorInterface)getBehavior();
		String					  image = null;

		if (behavior != null)
			image = behavior.getImage(this, _action);
		if (image == null)
			if (_action.getInternalFilePrefix() == null)
				image = getImageFromPrefix("LY_COMPANY_LOGO", DEFAULT_IMAGE_FORMAT, 0);
			else
				image = getImageFromPrefix(_action.getFilePrefix(), DEFAULT_IMAGE_FORMAT, 0);

		return image;
	}

	/**
	 * Retrieve a view controller kept.
	 *
	 * @param  id            The action id.
	 * @param  perspectiveId The perspective id.
	 * @return The kept view controller.
	 */
	LyViewController getViewController(String id, String perspectiveId)
	{
		if (_viewControllers == null)
			return null;
		else if (perspectiveId == null)
			return _viewControllers.get(id);
		else
			return _viewControllers.get(id + perspectiveId);
	}

	/**
	 * Remove a view controller kept.
	 *
	 * @param id            The action id.
	 * @param perspectiveId The perspective id.
	 */
	void removeViewController(String id, String perspectiveId)
	{
		if (perspectiveId == null)
			_viewControllers.remove(id);
		else
			_viewControllers.remove(id + perspectiveId);
	}

	/**
	 * Add a view controller to keep.
	 *
	 * @param id            The action id.
	 * @param perspectiveId The perspective id.
	 * @param controller    The view controller to keep.
	 */
	void addViewController(String id, String perspectiveId, LyViewController controller)
	{
		if (_viewControllers == null)
			_viewControllers = new Hashtable<String, LyViewController>(1);

		if (perspectiveId == null)
			_viewControllers.put(id, controller);
		else
			_viewControllers.put(id + perspectiveId, controller);
	}

	/**
	 * Add a view controller to keep.
	 *
	 * @param id         The action id.
	 * @param controller The view controller to keep.
	 */
	public void addViewController(String id, LyViewController controller)
	{
		addViewController(id, null, controller);
	}

	/**
	 * Return the list of view controllers opened from this controller.<br />
	 * Keys of the returned list are action identifiers and values are the view controllers
	 * themselves.
	 *
	 * @return the list of view controllers opened from this controller.
	 */
	public Hashtable<String, LyViewController> getViewControllers()
	{
		return _viewControllers;
	}

	/**
	 * Updates the action image representation in the controlled dashboard . This method may be used
	 * to change the icon of an action in the view.
	 *
	 * @param action action to be modified in the dashboard
	 * @param image  new image key or path.
	 */
	public void setImage(LyAction action, String image)
	{
		if (action == null)
			return;

		LyTool tool = (LyTool)getComponent().getComponent(action.getId());

		if (tool != null)
			tool.setImage(image);
	}

	/**
	 * Enables/Disables given objectInfo representation in the dashboard view.
	 *
	 * @param action action to be enabled/disabled in the dashboard
	 * @param enable new action status
	 */
	public void setEnabled(LyAction action, boolean enable)
	{
		if (action == null)
			return;

		LyComponent tool = getComponent().getComponent(action.getId());

		if ((tool != null) && (tool instanceof LySetEnabledInterface))
			((LySetEnabledInterface)tool).setEnabled(enable);
	}

	/**
	 * Called when a son view controller has been removed from the parent view controller.
	 *
	 * @param son The removed son.
	 */
	protected void sonRemoved(LyViewController son)
	{
		// Search in _viewController table to remove the son.
		if (_viewControllers != null)
		{
			Enumeration<String> keys = _viewControllers.keys();
			String			    key;
			Object			    controller;

			while (keys.hasMoreElements())
			{
				key = keys.nextElement();
				controller = _viewControllers.get(key);
				if (controller == son)
				{
					_viewControllers.remove(key);
					break;
				}
			}
		}
	}

	/**
	 * Gets help file associated with this controller. If we are in single view mode, return the
	 * dashboard help file is no child controller is started. Otherwise, returns the child
	 * controller help file.
	 *
	 * @return the help file corresponding to the action.
	 */
	public String getHelpFile()
	{
		if (_displayedAreaController == null)
			return super.getHelpFile();

		return _displayedAreaController.getHelpFile();
	}

    /**
     * Remove a son from this controller.
     *
     * @param areaController son area controller to remove
     */
    protected void removeSon(LyAreaController areaController)
    {
        // Unregister controller
        LyAction oldAction = areaController.getAction();

        if ((_openedActions == null) || (!_openedActions.contains(oldAction)))
            return;

        _displayedActions.remove(oldAction);
        _openedActions.removeInfo(oldAction);
        removeAreaController(areaController);
        _freeAreaControllers.remove(oldAction);

        // Close controller view
        closeController(areaController);
    }

    /**
     * Search and display objects corresponding to given parameters.
     * @param classInfoId	Class info id.
     * @param value			Part of id or name of searched objects.
     */
    protected void search(String classInfoId, String value)
    {
		LyApplication	application = getApplication();
		LyClassInfo		classInfo;

		if (classInfoId == null)
			classInfo = (LyClassInfo)getParameter("_searchClassInfo");
		else
			classInfo = application.getClassInfo(classInfoId);

		if (classInfo != null)
		{
			LyInfoList<LyFieldInfo>	fields = new LyInfoList<LyFieldInfo>();
			LyFilter				filter = buildSearchFilter(classInfo, value, fields);
			LyClassWorkSpace		objects = new LyPageWorkSpace(_session, classInfo, fields, filter, null,
				SEARCH_LIMIT, false, true);

			if (objects.getSize() == 0)
				showError("LY_ER_NO_SEARCH_RESULT");
			else
			{
				LyAction	action = classInfo.getActionWithId(LyAction.ID_CONSULT);

				if (action != null)
					_session.invokeAction(this, action, classInfo, objects, true);
			}
		}
    }

    /**
     * Builds the filter used to search and display objects corresponding to given parameters.
     * @param classInfo Class info used to build filter.
     * @param value     Part of id or name of searched objects.
     * @param fields    List of fields used to perform search.
     * @return          The filter to be used for given class info and given value.
     */
    protected LyFilter buildSearchFilter(LyClassInfo classInfo, String value, LyInfoList<LyFieldInfo> fields)
    {
		LyFieldInfo				fieldInfo;
		int						i, n = classInfo.getFieldInfoCount();
		LyFilter				filter = new LyFilter(LyFilter.Type.OR);
		LyFilterExpression		filterExp;

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			if ((fieldInfo.hasMark(LyFieldInfo.MARK_ID)
					&& !fieldInfo.hasMark(LyFieldInfo.MARK_HIDDEN))
				|| fieldInfo.hasMark(LyFieldInfo.MARK_NAME))
			{
				if (!fields.contains(fieldInfo) && (fieldInfo instanceof LyRelationFieldInfo))
				{
					LyRelationFieldInfo	relationFieldInfo = (LyRelationFieldInfo)fieldInfo;
					int					j, m = relationFieldInfo.getClassInfoCount();

					for (j = 0; j < m; j++)
					{
						LyClassInfo			relatedClassInfo = relationFieldInfo.getClassInfo(j);
						LyFilterRelation	relatedFilter = new LyFilterRelation(relationFieldInfo);

						relatedFilter.setFilter(buildSearchFilter(relatedClassInfo, value, fields));
						filter.addElement(relatedFilter);
					}
				}
				else
				{
					try
					{
						Object	val = fieldInfo.parse(value);

						filterExp = new LyFilterExpression(fieldInfo, LyFilterExpression.Condition.CONTAINS, val);
						filter.addElement(filterExp);
					}
					catch (Exception e)
					{
					}
				}

				fields.addInfo(fieldInfo);
			}
		}

		return filter;
    }

    /**
     * Fetch the social action. By default the action id is '_social' but this may be overridden with _findActionId
     * on the action.
     *
     * @return  the social action to open the widget
     */
    public LyAction getSocialAction()
    {
    	LyViewManager	viewManager = getViewManager();

    	if (!viewManager.isMultiSession())
    		return null;

    	if (!viewManager.hasProperty("HANDLE_SOCIAL_BEHAVIOR"))
    		return null;

        if (_session.getUser() == null)
        	return null;

    	if (_socialAction != null)
    		return _socialAction;

    	// Get standard social action
        LyApplication	application = getApplication();
        LyObject		user = _session.getUser();

        _socialAction = application.getActionWithId("_social", user.getClassInfo());
        return _socialAction;
    }

    /**
     * Implementation of find object menu action.
     */
    protected void showSocialPanel()
    {
        if ((_rightBorderPanelController instanceof LyAreaController)
            && !_rightBorderPanelController.isFreed())
        {
        	if (_rightBorderPanelController.isIdle())
        	{
        		_rightBorderPanelController.setIdle(false);
        		_rightBorderPanelController.showView();
        	}
        	else
        	{
        		_rightBorderPanelController.setIdle(true);
        		_rightBorderPanelController.hideView();
        	}
        }
        else
        {
	        LyAction 		action = getSocialAction();
	        LyObject		user = _session.getUser();
	        LyClassInfo		userClassInfo = user.getClassInfo();

	        _rightBorderPanelController = (LyAreaController)invokeAction(action, userClassInfo, null, false);
	        if (_rightBorderPanelController != null)
	        	_rightBorderPanelController.showView();
        }
    } // end method showSocialPanel

    @Override
    public void updateViewFromValues(LyDemand demand)
    {
		// Update view when needed
		if (demand instanceof LyActionDemand)
		{
			LyParameterValues values = demand.getParameterValues();

			if (values != null)
			{
				LyAreaController	areaController = getDisplayedAreaController();

				if (areaController != null)
				{
					LyComponent	component = areaController.getComponent();

					if (component != null)
					{
						values.updateViewFromValues(getComponent(), component, false);
						setLastModified(System.currentTimeMillis());
					}
				}
			}
		}
    }
} // end class LyDashboardController
