/**
 * LyDeleteController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyDeleteController.java 34887 2012-03-21 09:54:53Z jlsaouli $
 */

package leon.control;

import java.util.*;
import leon.app.*;
import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.setinterface.*;


/**
 * This class describes the controller associated to a delete action. This class asks the user for a
 * confirmation and accept single object or multiple objects deletion.
 *
 * <P>This class is associated to a LyDeleteBehaviorInterface, usually a LyClassBehavior instance.
 * </P>
 *
 * <P>The delete controller does not have any builder (no view is generated).</P>
 *
 * @see leon.app.behaviorinterface.LyDeleteBehaviorInterface
 * 
 * @see leon.app.behavior.LyClassBehavior
 */
public class LyDeleteController extends LyAreaController
	implements LyMessageAnswerInterface,
		LyQuestionAnswerInterface,
		LyControllerContextInterface
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyDeleteController.java 34887 2012-03-21 09:54:53Z jlsaouli $";

	/** Class info of deleted objects. */
	protected LyClassInfo      _classInfo;
	
	/** List of object to delete by this controller. */
	protected LyObjectList     _objects;

	/** Current object. */
	protected LyObject		   _currentObject;

	/** Is last object ? */
	protected boolean		   _last;

	/** Current object in list. */
	protected int			   _currentIndex;

	/** Context for this controller. */
	protected LyContext		   _controllerContext;

	/**
	 * Constructor.
	 *
	 * @param session    current session associated to the controller. The controller automatically
	 *                   registers on the specified session.
	 * @param parent     parent controller (or null for 'root' controller).
	 * @param action     related action. If an action is specified, the controller uses action
	 *                   parameters (action name, action specific XML view description ...) 
	 * @param classInfo  kept for compatibility (introspection). Not used.
	 * @param objects    current object list being deleted. Must be not null.
	 * @see   #getClassInfo
	 */
	public LyDeleteController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		super(session, parent, action, null);

		if (objects == null)
			throw new IllegalStateException(
				"Can't build a deleteController if deleted object list is null.");

		if ((classInfo == null)
		        && (action.getClassInfo() != null))
			classInfo = action.getClassInfo();
		
        _classInfo = classInfo;
        
        LyInfoList<LyClassInfo> classes = objects.getClasses();
        
		if ((classInfo != null)
				&& (classes != null)
				&& (classes.getSize() == 1)
				&& (classes.getInfo(0) != classInfo))
			_objects = objects.getList(getSession(), classInfo);
		else
			_objects = new LySimpleObjectList(objects);
	}

	String getErrorMessage(Vector<LyError> errors)
	{
		int			  i, n = errors.size();
		LyError		  error;
		StringBuffer  buffer = new StringBuffer();
		LyEnvironment environment = getEnvironment();
		String		  message;
		LyObject	  cause;

		for (i = 0; i < n; i++)
		{
			error = errors.elementAt(i);
			if (i != 0)
				buffer.append("\n");
			cause = error.getObjectCause();
			if (cause != null)
			{
				buffer.append(getName(cause));
				buffer.append(" : ");
			}
			buffer.append(error.getMessage(environment));
		}

		message = buffer.toString();
		return message;
	}

	/**
	 * Free this controller.
	 */
	public void free()
	{
		// Commit the context.
		if (_controllerContext != null)
		{
			Vector<LyError> errors = _controllerContext.getErrors(LyObject.LEVEL_ERROR);

			if ((errors != null) && (errors.size() != 0))
			{
				_controllerContext.rollback();

				String		  message = getErrorMessage(errors);
				LyEnvironment environment = getEnvironment();

				message = environment.getMessage("LY_ER_COMMIT_CONTEXT", message);
				getParent().showError(message);
			}
			else
			{
                LyContext.Status status = _controllerContext.commit();

				if (status == LyContext.Status.ROLLBACKED)
				{
					LyApplication application = getApplication();
					LyEnvironment environment = getEnvironment();
					String		  cause = _controllerContext.getErrorMessage();
					String		  message;

					if (cause == null)
						cause = application.getError();
					message = environment.getMessage("LY_ER_COMMIT_CONTEXT", cause);
					getParent().showError(message);
				}
				else if (status == LyContext.Status.ERROR)
				{
					errors = _controllerContext.getErrors(LyObject.LEVEL_ERROR);
					if ((errors != null) && (errors.size() != 0))
					{
						_controllerContext.rollback();

						String		  message = getErrorMessage(errors);
						LyEnvironment environment = getEnvironment();

						message = environment.getMessage("LY_ER_COMMIT_CONTEXT", message);
						getParent().showError(message);
					}
				}
			}
			_controllerContext = null;
		}

		// Free itself because no view controller
		free(_parent);
	} // end method free

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyDeleteBehaviorInterface.class;
	}
	
	/**
	 * Gets appropriate behavior for given object.
	 * @param object the object
	 * @return the associated behavior
	 */
	protected LyDeleteBehaviorInterface getBehavior(LyObject object)
	{
	    LyDeleteBehaviorInterface  behavior = (LyDeleteBehaviorInterface)super.getBehavior();
	    
        if ((object.getClassInfo() == _classInfo)
                || !(behavior instanceof LyClassBehavior))
            return behavior;

        LyClassInfo classInfo = object.getClassInfo();
        Class<?>    behaviorClass = classInfo.getBehaviorClass();
        
        if (behaviorClass != null)
        {
            try
            {
                // Try to find if we have a specific class behavior
                LyClassBehaviorInterface    classBehavior;
                
                classBehavior = (LyClassBehaviorInterface)behaviorClass.newInstance();
                classBehavior.init(this);
                return classBehavior;
            }
            catch (Exception e)
            {
                LyEnvironment  environment = getEnvironment();
                
                environment.writeException(e);
                environment.logError("BF AppEngine : instanciation of behavior " + behaviorClass
                    + " failed on controller " + getClass()
                    + ". Check the behavior attribute of your action " + _action.getId());
            }
        }
        
        return behavior;
	}
	
	/**
	 * Make the view visible to the user. This view raises a confirmation dialog.
	 */
	public void showView()
	{
		// confirm action
		// Build confirmation String
		if (_objects == null)
			return;

		// Start action on first access
		if (!startAction())
		{
			free();
			return;
		}

		// Check if deletion if allowed or not for the given objects
		LyDeleteBehaviorInterface behavior = (LyDeleteBehaviorInterface)getBehavior();

		if (behavior != null)
		    _objects = behavior.checkDeletions(this, _objects);
		
		if ((_objects == null)
			|| (_objects.getSize() == 0))
		{
			free();
			return;
		}

		_currentIndex = 0;

		confirmMessage();
	}

	/**
	 * Creation of a confirmation dialog view to confirm deletion.
	 */
	protected void confirmMessage()
	{
		int n = _objects.getSize();

		getControllerContext();

		if (_currentIndex >= n)
			return;

		LyViewManager			  viewManager = getViewManager();
		LyObject				  object = _objects.getObject(_currentIndex);

		// Check if deletion if allowed or not for the given object
		LyDeleteBehaviorInterface behavior = getBehavior(object);
        
		if (behavior != null)
		{   
			String message = behavior.checkDeletion(this, object);

			if (message != null)
			{
				if (_currentIndex == (n - 1))
				{
					getParent().showError(message);
					messageAnswer(LyViewManager.COMMAND_NO, null);
				}
				else
				{
					message += "\n\n" + "LY_CONFIRM_DELETION_PROCESS";
					viewManager.askQuestion(((_parent == null) ? null : _parent.getView()), message,
						this, null);
				}
				return;
			}
		}

		// Show Asynchronous popup message
		String  confirmMessage = null;
		boolean askMultipleDeletion = viewManager.hasProperty("ASK_MULTIPLE_OBJECTS_DELETION");

		if (behavior != null)
			confirmMessage = behavior.getConfirmDeletionMessage(this, object);

		if ((confirmMessage == null) || (confirmMessage.length() == 0))
		{
			StringBuffer question = new StringBuffer(100);

			if (!askMultipleDeletion)
			{
				question.append("LY_CONFIRM_DEL");
				question.append(" : ");
				question.append(object.getName());
			}
			else
				question.append("LY_CONFIRM_ALL_OBJECTS_DEL");

			confirmMessage = question.toString();
		}

		BitSet commands = new BitSet();

		if (!askMultipleDeletion)
		{
			commands.set(LyViewManager.COMMAND_YES);

			if (n > 1)
			{
				commands.set(LyViewManager.COMMAND_ALL);
				commands.set(LyViewManager.COMMAND_NONE);
			}

			commands.set(LyViewManager.COMMAND_NO);
		}
		else
		{
			commands.set(LyViewManager.COMMAND_ALL);
			commands.set(LyViewManager.COMMAND_NONE);
		}

		String messageTitle = null;

		if (behavior != null)
			messageTitle = behavior.getTitle(this, _action);

		if (messageTitle == null)
			messageTitle = _action.getName();

		String filePrefix = null;

		if (behavior != null)
			filePrefix = behavior.getImage(this, _action);

		if (filePrefix == null)
			filePrefix = _action.getFilePrefix();

		viewManager.showMessage(((_parent == null) ? null : _parent.getView()), messageTitle,
			filePrefix, confirmMessage, commands, this, null);

		// When window is closed, messageAnswer is called
	} // end method confirmMessage

	/**
	 * Answer to a askQuestion request.
	 *
	 * @param result one of the possible return code for LyViewManager : COMMAND_XXX
	 * @param clientData data specified in the askQuestion method
	 * @see   leon.view.setinterface.LyMessageAnswerInterface
	 */
	public void questionAnswer(boolean result, Object clientData)
	{
		if (result)
			messageAnswer(LyViewManager.COMMAND_NO, clientData);
		else
			messageAnswer(LyViewManager.COMMAND_NONE, clientData);
	}

	/**
	 * Answer to a askQuestion request. Internal use only.
	 *
	 * @param result one of the possible return code for LyViewManager : COMMAND_XXX
	 * @param clientData data specified in the askQuestion method
	 * @see   leon.view.setinterface.LyMessageAnswerInterface
	 */
	public void messageAnswer(int result, Object clientData)
	{
		int n = _objects.getSize();

		if (_currentIndex >= n)
			return;

		LySession     session = getSession();
		LyApplication application = session.getApplication();
		LyContext     context = null, oldCurrentContext = null;
		LyObject	  object;
		LyEnvironment environment = getEnvironment();
		LyViewManager viewManager = getViewManager();
		boolean		  askMultipleDeletion = viewManager.hasProperty(
				"ASK_MULTIPLE_OBJECTS_DELETION");

		if (askMultipleDeletion)
		{
			if (result == LyViewManager.COMMAND_NO)
				result = LyViewManager.COMMAND_NONE;
			else if (result == LyViewManager.COMMAND_YES)
				result = LyViewManager.COMMAND_ALL;
		}

		oldCurrentContext = application.getCurrentContext();
		context = getControllerContext();
		application.setCurrentContext(context);
		getViewManager().setWaiting(true);

		try
		{
			object = _objects.getObject(_currentIndex);
			close(object, result, (_currentIndex == (n - 1)));

			// Behavior may have freed the controller.
			// If yes we don't go any further
			if (!isFreed())
			{
				// What's next ?
				_currentIndex++;

				// Do not ask the question anymore : check if possible and delete all remaining objects
				if (result == LyViewManager.COMMAND_ALL)
				{
					LyDeleteBehaviorInterface behavior;
					String					  message;
					int						  i;

					for (i = _currentIndex; i < n; i++)
					{
						object = _objects.getObject(i);
						behavior = getBehavior(object);
						
						if ((behavior != null)
								&& ((message = behavior.checkDeletion(this, object)) != null))
							getParent().showError(message);
						else
						{
							close(object, result, (i == (n - 1)));
							if (isFreed())
								break;
						}
					}
					_currentIndex = n;
				}

				// Over
				if (result == LyViewManager.COMMAND_NONE)
					_currentIndex = n;

				if (_currentIndex < n)
					// Remaining objects : loop
					confirmMessage();
				else
                    
					free();
			}

			// We must'nt have some pending transaction on the current
			// session.
			if (session.hasTransaction())
			{
				String message = environment.getMessage("LY_ER_REMAINING_PENDING_TRANSACTION");

				if (isFreed())
					environment.raiseError(message);
				else
					getParent().showError(message);

				throw new IllegalStateException("BF AppEngine : Remaining pending transactions");
			}
		}
		catch (Throwable t)
		{
			session.handleException(t);
            
            try
            {
                if (_controllerContext != null)
                    _controllerContext.rollback();
            }
            catch (Throwable t2)
            {
                session.handleException(t2);
    		}
        }
		finally
		{
			if ((oldCurrentContext != null)
					&& !oldCurrentContext.isEnded())
			{
				if (oldCurrentContext != context)
					application.setCurrentContext(oldCurrentContext);
			}
			else
				application.setCurrentContext(null);

			viewManager.setWaiting(false);
		}
	} // end method messageAnswer

	/**
	 * Closing deletion controller.
	 *
	 * @param object current object being deleted
	 * @param result user response to the deletion confirmation (See LyViewManager)
	 * @param last   is it the last object which is being deleted.
	 * @see   leon.view.LyViewManager
	 */
	protected void close(LyObject object, int result, boolean last)
	{
		_last = last;

		switch (result)
		{
		case LyViewManager.COMMAND_YES:
		case LyViewManager.COMMAND_OK:
			end(object);
			break;

		case LyViewManager.COMMAND_NO:
		case LyViewManager.COMMAND_CANCEL:
		case LyViewManager.COMMAND_CLOSED:
			return;

		case LyViewManager.COMMAND_ALL:
			end(object);
			break;

		case LyViewManager.COMMAND_NONE:
			break;
		}
	}

	/**
	 * End of the current action. The deletion of specified object completed, and the application
	 * may take control.
	 *
	 * <P>The method calls LyClassBehavior.endDeletion() (LyDeleteBehaviorInterface)</P>
	 *
	 * @param object object which is being deleted.
	 */
	protected void end(LyObject object)
	{
		_currentObject = object;
		if (object.isDeleted())
			return;
		
	    LyDeleteBehaviorInterface behavior = getBehavior(object);
	      
		if (behavior != null)
		    behavior.endDeletion(this);
	}

	/**
	 * Accessor to the view controller (Frame or dialog controller) which is the container of this
	 * view. This method returns <b>null</b> since there is no view controller associated to the
	 * delete action.
	 *
	 * @return the reference to a view controller, returns null
	 */
	public LyViewController getViewController()
	{
		return null;
	}

	/**
	 * Accessor to the view that contains the component (getComponent) of the controller. Always
	 * return null (no view is associated)
	 *
	 * @return null
	 */
	public LyView getView()
	{
		return null;
	}

	/**
	 * Accessor to the builder of the view Always return null (no view is associated).
	 *
	 * @return null
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return (null);
	}

	/**
	 * Access to current object being deleted.
	 *
	 * @return current object reference. If multiple deletion, this value is updated before each
	 *         call to the behavior.
	 */
	public LyObject getObject()
	{
		return _currentObject;
	}

	/**
	 * Indicates wether this is last object in list.
	 *
	 * @return true if current object is the last one.
	 */
	public boolean isLast()
	{
		return _last;
	}

	/**
	 * Access to controlled object list. Never return null. If there is only one object being
	 * deleted, The list contains a single object.
	 *
	 * @return the list of object resquested for deletion.
	 */
	public LyObjectList getObjects()
	{
		return _objects;
	}

	/**
	 * Get class info of deleted object(s). The classInfo is obtained from the classInfo of the
	 * first object in the object list.
	 *
	 * @return the classInfo associated to this deletion request.
	 */
	public LyClassInfo getClassInfo()
	{
	    if (_classInfo != null)
	        return _classInfo;
	    
		if ((_objects == null) || (_objects.getSize() == 0))
			return null;

		LyObject object = _objects.getObject(0);

		_classInfo = object.getClassInfo();
		return _classInfo;
	}

	/**
	 * Gets the controller context.
	 *
	 * @return The controller context.
	 */
	public LyContext getControllerContext()
	{
		if ((_controllerContext == null)
				|| (_controllerContext.isEnded()))
		{
			LySession		 session = getSession();
			LyContextManager contextManager = session.getContextManager();
			LyController     parent = _parent;

			// Last try to get a root level context controller
			if (!(parent instanceof LyControllerContextInterface))
				parent = session.getRootContextController();

			_controllerContext = contextManager.newContext(session);

			if (parent instanceof LyControllerContextInterface)
			{
				// Be careful, the parent can return null even if it implements the interface
				LyContext context;

				if (parent instanceof LyControllerContextInterface2)
					context = ((LyControllerContextInterface2)parent).getControllerContext(this);
				else
					context = ((LyControllerContextInterface)parent).getControllerContext();

				if (context != null)
					context.addSon(_controllerContext);
			}

			_controllerContext.setName(getContextName());
		}
		return _controllerContext;
	} // end method getControllerContext

	String getContextName()
	{
		int    n = _objects.getSize();
		String name = _action.getName();

		if (n > 0)
		{
			LyObject object = _objects.getObject(0);

			name += " : ";
			name += getName(object);
		}
		if (n > 1)
			name += "...";

		return name;
	}
} // end class LyDeleteController
