/**
 * LyDialogController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyDialogController.java 41484 2012-11-09 17:48:14Z smaupin $
 */

package leon.control;

import java.util.*;
import leon.app.*;
import leon.control.builder.*;
import leon.info.*;
import leon.info.misc.*;
import leon.view.*;


/**
 * This controller manages a dialog window (or view). This dialog is used as the view controller of
 * other views (such as creation/modification/consultation forms).
 *
 * <P>Unlike a Frame, a dialog is a child window of a parent, not a main Window. A dialog does not
 * have a MenuBar and may be modal.</P>
 *
 * <P>The LyDialogController automatically calculates the dialog position and size. Size is
 * calculated from the content, and location (using VIEW_XOFFET, VIEW_YOFFSET) Position is
 * calculated from parent location.</P>
 *
 * <P>The dialog is always kept inside the screen, and a check is performed to be sure that de
 * dialog location fits in the screen.</P>
 *
 * <P>The other view controller is LyFrameController.</P>
 *
 * @see LyFrameController
 */
public final class LyDialogController extends LyViewController
{
	/** File Version */
	public static final String __VERSION =
		"$Id: LyDialogController.java 41484 2012-11-09 17:48:14Z smaupin $";

	/**
	 * Default X value for the offset (relative to parent) to determine view position when this
	 * dialog is opened.
	 */
	public static int		   VIEW_XOFFSET = 20;

	/**
	 * Default Y value for the offset (relative to parent) to determine view position when this
	 * dialog is opened.
	 */
	public static int		   VIEW_YOFFSET = 20;

	/** Maximum number of fields in a popup view, -1 means no limit. */
	public static int		   MAX_FIELDS_IN_POPUP = -1;

	/** Maximum number of tabs in a popup view, -1 means no limit. */
	public static int		   MAX_TABS_IN_POPUP = -1;

	/** Indicates that the bounds (size / location) are still unset for this view. */
	protected boolean		   _boundsSet = false;

	/**
	 * Construct a dialog controller with the specified parent.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller : this is never null for a dialog controller.
	 */
	public LyDialogController(LySession session, LyController parent)
	{
		this(session, parent, null);
	}

	/**
	 * Construct a dialog controller with the specified parent and a given area controller.
	 *
	 * @param session        current session associated to the controller. The controller
	 *                       automatically registers on the specified session.
	 * @param parent         parent controller : this is never null for a dialog controller.
	 * @param areaController the area which is inside the dialog : an area automatically builds a
	 *                       view controller (dialog or frame) when showView is called.
	 */
	public LyDialogController(LySession session, LyController parent,
		LyAreaController areaController)
	{
		super(session, parent, areaController);

		LyApplication application = session.getApplication();
		Class<?>	  builderClass = application.loadClass("leon.control.builder.LyDialogBuilder");

		setBuilderClass(builderClass);
	}

	/**
	 * Make the view visible to the user. Automatically calculate location on screen If the display
	 * is a fat client (windows (awt) / swing) this methods should open a window and show the view.
	 * In the case of thin client (html), this prepares the pages which will be returned and
	 * displayed in the client browser.
	 */
	public void showView()
	{
		LyView view = getView();

		if (!_viewInitialized)
			initView();

		// Check that this controller was not freed (initialisation failure)
		if (isFreed())
		{
			// If the view is not null, free it
			if (view != null)
			{
				view.free();
				view = null;
			}

			return;
		}

		boolean isPortlet = getViewManager().isPortletMode();
		// If view exists and is not visible, handle geometry
		if ((view != null) && !view.isVisible())
		{
			LyAreaController areaController = getAreaController();
			LyAction		 action = areaController.getAction();
			LyGeometry		 geometry = action.getGeometry();
			LyDialog         dialogView = (LyDialog)view;

			if (action.getTarget() != null && !isPortlet)
			{
			    if (!"_none".equals(action.getTarget()))
			        ((LyDialog)view).setTarget(action.getTarget());

			}
			else if (isPortlet
                    || (dialogView.getTarget() == null
                            && (getViewManager().hasProperty("FORCE_POPUP_VIEW")
                                  || (getViewManager().hasProperty("POPUP_VIEW")
                                        && action.hasMark(LyAction.MARK_POPUP)))))

			{
				// Check if view must be displayed in a popup
				boolean inPopUp = true;

				if (areaController instanceof LyConsultController)
				{
					LyConsultController controller = (LyConsultController)areaController;
					LyTabs			    tabs = controller.getTabs();
					int				    tabsCounter = (tabs == null ? 0 : tabs.getTabCount());

					// Check the number of tabs
					if ((MAX_TABS_IN_POPUP >= 0)
							&& (tabsCounter > MAX_TABS_IN_POPUP))
						inPopUp = false;
					else
					{
						// Check the number of fields
						LyComponent     component = controller.getComponent();

						if (component != null)
						{
							Vector<LyFormField> fields = new Vector<LyFormField>();

							component.getComponents(LyFormField.class, fields);
							if ((MAX_FIELDS_IN_POPUP >= 0)
									&& (fields.size() > MAX_FIELDS_IN_POPUP))
								inPopUp = false;
						}
					}
				}

				if (inPopUp
					|| getViewManager().hasProperty("FORCE_POPUP_VIEW"))
				{
					((LyDialog)view).setTarget(LySession.POPUP_VIEW);
					getViewManager().setCurrentWindowName(LySession.POPUP_VIEW);
				}
			}

			if ((geometry != null)
					&& (geometry.getX() >= 0)
					&& (geometry.getY() >= 0))
			{
				view.setLocation(geometry.getX(), geometry.getY());
				_boundsSet = true;
			}
			else if (!_boundsSet)
			{
				// set location according to parent location
				// XX MOVE Calculation in view/win package
				int[]  pt = null, dim, screen;
				LyView parentView = null;

				// FFT00445/Flandres
				if (_parent != null)
				{
					LyViewController parentViewController = _parent.getViewController();

					if (parentViewController != null)
						parentView = parentViewController.getView();
					else
						parentView = null;
				}

				dim = view.getSize();
				screen = getEnvironment().getScreenSize(parentView);

				// FFT00435/Flandres
				if (screen == null)
					screen = new int[] { 0, 0, 0, 0 };
				if (dim == null)
					dim = new int[] { 0, 0 };

				if ((parentView == null)
				        || ((VIEW_XOFFSET < 0) && (VIEW_YOFFSET < 0))
				        || ((LyDialog)view).isModal())
				{
					// Center the dialog
					pt = new int[]
						{
							Math.max(0, (screen[0] - dim[0]) / 2) + screen[2],
							Math.max(0, (screen[1] - dim[1]) / 2) + screen[3]
						};
				}
				else
				{
					pt = parentView.getLocation();
					if (pt == null)
						pt = new int[] { 0, 0 };
					pt[0] += VIEW_XOFFSET;
					pt[1] += VIEW_YOFFSET;
				}

				if ((pt[0] + dim[0]) > (screen[0] + screen[2]))
					pt[0] = (VIEW_XOFFSET < 0 ? screen[2] : VIEW_XOFFSET + screen[2]);
				if ((pt[1] + dim[1]) > screen[1] + screen[3])
					pt[1] = (VIEW_YOFFSET < 0 ? screen[3] : VIEW_YOFFSET + screen[3]);

				view.setLocation(pt[0], pt[1]);
				_boundsSet = true;
			}
		}

		super.showView();
	} // end method showView

	/**
	 * Set the location of the view.
	 *
	 * @param x the new x position of the view
	 * @param y the new y position of the view
	 */
	protected void setLocation(int x, int y)
	{
		super.setLocation(x, y);
		_boundsSet = true;
	}

	/**
	 * Factory of the default builder for a controller. Overriden to create an instance of
	 * LyDialogBuilder
	 *
	 * @return the builder which creates the XML description of the view
	 * @see    leon.control.builder.LyDialogBuilder
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return (new LyDialogBuilder(this));
	}
} // end class LyDialogController
