/**
 * LyEditTableController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyEditTableController.java 41317 2012-10-23 11:59:54Z bangiole $
 */

package leon.control;

import leon.app.*;
import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.control.form.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.mobile.*;
import leon.view.setinterface.*;

import java.util.*;

import leon.view.web.jquery.LyJQViewManager;


/**
 * This class implements an editable table controller. Fields displayed in the table may be modified
 * automatically. It is associated with a builder (LyTableBuilder) and a behavior that implements
 * LyTableBehaviorInterface. Usually, this behavior is implemented by an instance of LyTableBehavior
 *
 * @see LySimpleTableBuilder
 * @see LyTableController
 * @see leon.app.behaviorinterface.LyTableBehaviorInterface
 */
public class LyEditTableController extends LyMultiPageTableController
{
    private static final int ACTION_ICON_SIZE = 4;

    /** File Version */
    public static final String __VERSION =
        "$Id: LyEditTableController.java 41317 2012-10-23 11:59:54Z bangiole $";

    /** Indicates if a toolbar is visible for unchanged objects */
    public static boolean TOOLBAR_FOR_UNCHANGED_OBJECTS = false;

    /** Restore inputs. */
    public static final String DEMAND_RESTORE = "_restore".intern();

    /** Demand id : Validate row. */
    public static final String DEMAND_VALIDATE_ROW = "_validateRow".intern();

    /** Demand id : Cancel row. */
    public static final String DEMAND_CANCEL_ROW = "_cancelRow".intern();

    /** Demand id : Copy cells. */
    public static final String DEMAND_COPY_CELLS = "_copyCells".intern();

    /** Demand id : Paste cells. */
    public static final String DEMAND_PASTE_CELLS = "_pasteCells".intern();

    /** Demand id : ALT+UP key demand. */
    public static final String DEMAND_ALT_UP = "_altUp".intern();

    /** Demand id : ALT+DOWN key demand. */
    public static final String DEMAND_ALT_DOWN = "_altDown".intern();

    /** Demand id : ALT+LEFT key demand. */
    public static final String DEMAND_ALT_LEFT = "_altLeft".intern();

    /** Demand id : ALT+RIGHT key demand. */
    public static final String DEMAND_ALT_RIGHT = "_altRight".intern();

    /** Demand id : ALT+ADD key demand. */
    public static String DEMAND_ALT_ADD = "_altAdd".intern();

    /** Max column width */
    public static int MAX_COLUMN_WIDTH = 50;


    /** Enable showing messages. */
    protected boolean _showMessages;

    /** Current set controllers opened on row items. */
    protected Hashtable<LyObject, LyEditTableSetController> _setControllers;

    /** Current create controller. */
    protected LyEditTableSetController _createController;

    /** FieldInfo set marks (null means all fields). */
    protected BitSet _setFieldMarks;

    /** FieldInfo create marks (null means all fields). */
    protected BitSet _createFieldMarks;

    /** Editable status. */
    protected boolean _editable;

    /** Enabled status. */
    protected boolean _enabled;

    /** Create action (null if not enabled). */
    protected LyAction _createAction;

    /** Set action (null if not enabled). */
    protected LyAction _setAction;

    /** Indicates if dynamic edition mode is set. */
    protected boolean _dynamicEditionMode;

    /** Indicates if single clic is used when dynamic edition mode is set. */
    protected boolean _useSingleClickForEdition;

    /** Indicates if auto-commit mode is activated. */
    protected boolean _autoCommit;

    /** Object that is being edited (dynamic edition mode only). */
    protected LyObject _editedObject;

    /** Field that is being edited (dynamic edition mode only). */
    protected LyFieldInfo _editedFieldInfo;

    /** Boolean indicating if editable tables are use for relations. */
    protected boolean _editTableForRelations;

    /** Internal offset used when editing a multiple choice field info. */
    private int _multipleChoiceOffset;

    /** Last time when error was shown. */
    protected long _errorTime;

    /** List of fields whose columns are fixed in width. */
    protected LyInfoList<LyFieldInfo> _fixedColumnWidth;

    /**
     * Selected cells indexes.
     * First value is the row of the first selected cell.
     * Second value is the column of the first selected cell.
     * Third value is the row of the last selected cell.
     * Fourth value is the column of the last selected cell.
     */
    protected int[] _selectedCells;

    /** Array of values retrieved from cells copy action. */
    protected LyValue[][] _bufferedValues;

    /** Indicates if current row is valid */
    protected boolean _isCurrentRowValid;

    /** last edited index */
    private int _lastEditedIndex = -1;

    /**
     * Constructor.
     *
     * @param session   current session associated to the controller. The controller automatically
     *                  registers on the specified session.
     * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
     *                  indicates a child controller (for example a dialog created from a main view)
     * @param action    related action. If an action is specified, the controller uses action
     *                  parameters (action name, action specific XML view description ...)
     * @param classInfo class info for objects displayed in the list.
     * @param context   current context (may be null if the view is not contextual to a list of
     *                  objects).
     */
    public LyEditTableController(LySession session, LyController parent, LyAction action,
        LyClassInfo classInfo, LyObjectList context)
    {
        super(session, parent, action, classInfo, context);
    }

    /**
     * Init this instance.
     */
    protected void init()
    {
        super.init();

        _setFieldMarks = getParameterMarks("_setFieldMarks", LyBitSet.Type.F_MARKS);
        _createFieldMarks = getParameterMarks("_createFieldMarks", LyBitSet.Type.F_MARKS);
        _showMessages = false;
        _editable = true;
        _enabled = true;
        _dynamicEditionMode = true;
        _useSingleClickForEdition = true;
        _autoCommit = false;
        _editedObject = null;
        _editedFieldInfo = null;
        _multipleChoiceOffset = 0;
        _fixedColumnWidth = null;
        _isCurrentRowValid = true;
        _lastEditedIndex = -1;

        String dynamicEditionMode = (String) getParameter("_dynamicEditionMode", true);

        _selectedCells = null;
        _bufferedValues = null;

        if ((dynamicEditionMode != null) && dynamicEditionMode.equalsIgnoreCase("false"))
            _dynamicEditionMode = false;

        String useSingleClickForEdition = (String) getParameter("_useSingleClickForEdition", true);

        if ((useSingleClickForEdition != null)
                && useSingleClickForEdition.equalsIgnoreCase("false"))
            _useSingleClickForEdition = false;

        String autoCommit = (String) getParameter("_autoCommit", true);

        if ((autoCommit != null) && autoCommit.equalsIgnoreCase("true"))
            _autoCommit = true;

        if (_parent instanceof LyFieldSetController)
        {
            LyFieldSetController   fieldController = (LyFieldSetController) _parent;
            LySetController        setController = fieldController.getSetController();
            LySetBehaviorInterface setBehavior = setController.getSetBehavior();

            _editTableForRelations = setBehavior.useEditTable(setController,
                    (LyRelationFieldInfo) fieldController.getFieldInfo());
        }
        else
            _editTableForRelations = true;

        if(getEnvironment().getViewManager().hasProperty(LyMobileViewManager.MOBILE_ONLY_PROPERTY))
        {
        	_editTableForRelations = false;
        }

        checkSetAction(null);
        checkCreateAction(null);

    } // end method init

    /**
     * Overridden to clean set controllers table and to cancel edition state.
     *
     * @param full indicates if a full reset (view + data) must be performed
     */
    public void reset(boolean full)
    {
        super.reset(full);

        if (full)
        {

            if ((_setControllers != null) && !_setControllers.isEmpty())
            {
                Enumeration<LyObject> objects = _setControllers.keys();
                LyObject              object;

                while (objects.hasMoreElements())
                {
                    object = objects.nextElement();

                    if ((_workSpace == null) || !_workSpace.contains(object))
                    {
                        LySetController setController = _setControllers.get(object);

                        _setControllers.remove(object);

                        if ((setController != null) && !setController.isFreed())
                            setController.free(this);
                    }
                }
            }

            _editedObject = null;
            _editedFieldInfo = null;
            _selectedCells = null;
            _bufferedValues = null;
            _lastEditedIndex = -1;
        }
    }

    /**
     * Check if set action is enabled or not with the given list of fields.
     * @param fields a list of fields
     */
    protected void checkSetAction(LyInfoList<LyFieldInfo> fields)
    {
        // Check if a _setAction parameter was set
        _setAction = null;

        // Check if at least one set field is present in the table,
        // otherwise set is not allowed
        if ((fields == null) || (fields.getSize() == 0))
        {
            fields = _classInfo.getFieldInfos(_fieldMarks);

            if ((fields == null) || (fields.getSize() == 0))
                fields = _classInfo.getFieldInfos();

            LyEnvironment environment = getEnvironment();

            if ((_setFieldMarks != null)
                    && environment.testResource("LY_DEFAULT_FIELDS_IN_FORMS_WHEN_EMPTY"))
            {
                LyInfoList<LyFieldInfo> setFields = _classInfo.getFieldInfos(_setFieldMarks);

                if ((setFields == null) || (setFields.getSize() == 0))
                    _setFieldMarks = null;
            }
        }

        // Check if a _setAction parameter was set
        LyAction             setActionParam = (LyAction) getParameter("_setAction", true);
        LySession            session = getSession();
        LyAction             setActionTmp = null;
        LyInfoList<LyAction> setActions = _classInfo.getActionsWithId(LyAction.ID_SET);
        int                  i, n = ((setActions == null) ? 0 : setActions.getSize());

        for (i = 0; i < n; i++)
        {
            setActionTmp = setActions.getInfo(i);

            if (!session.validateAction(setActionTmp, this)
                    || ((setActionParam != setActionTmp)
                        && setActionTmp.hasMark(LyAction.MARK_HIDDEN)))
                continue;

            // Check if at least one editable field is present in the table, otherwise set is not allowed
            int         j, m = _classInfo.getFieldInfoCount();
            LyFieldInfo fieldInfo;

            for (j = 0; j < m; j++)
            {
                fieldInfo = _classInfo.getFieldInfo(j);

                if ((fields != null) && fields.contains(fieldInfo) && canBeEditable(fieldInfo)
                        && ((_setFieldMarks == null) || fieldInfo.hasMarks(_setFieldMarks, false)))
                {
                    _setAction = setActionTmp;
                    return;
                }
            }
        }
    } // end method checkSetAction

    /**
     * Check if creation action is enabled or not with the given list of fields.
     * @param fields a list of fields
     */
    protected void checkCreateAction(LyInfoList<LyFieldInfo> fields)
    {
        _createAction = null;

        // Check first that class info is not abstract
        if (_classInfo.isAbstract())
            return;

        // Check if all create and mandatory fields are present in the table,
        // otherwise creation is not allowed
        if ((fields == null) || (fields.getSize() == 0))
        {
            fields = _classInfo.getFieldInfos(_fieldMarks);

            if ((fields == null) || (fields.getSize() == 0))
            {
                fields = _classInfo.getFieldInfos();
                _fieldMarks = null;
            }

            LyEnvironment environment = getEnvironment();

            if ((_createFieldMarks != null)
                    && environment.testResource("LY_DEFAULT_FIELDS_IN_FORMS_WHEN_EMPTY"))
            {
                LyInfoList<LyFieldInfo> createFields = _classInfo.getFieldInfos(_createFieldMarks);

                if ((createFields == null) || (createFields.getSize() == 0))
                    _createFieldMarks = null;
            }
        }

        // Check if a _createAction parameter was set
        LyAction             createActionParam = (LyAction) getParameter("_createAction", true);
        LySession            session = getSession();
        LyAction             createActionTmp = null;
        LyInfoList<LyAction> createActions = _classInfo.getActionsWithId(LyAction.ID_CREATE);
        int                  i, n = ((createActions == null) ? 0 : createActions.getSize());
        boolean              match;

        for (i = 0; i < n; i++)
        {
            match = true;
            createActionTmp = createActions.getInfo(i);

            // Check if action is valid and if it is not hidden
            if (!session.validateAction(createActionTmp, this)
                    || ((createActionParam != createActionTmp)
                        && createActionTmp.hasMark(LyAction.MARK_HIDDEN)))
                continue;

            // Check that all mandatoty fields are present in edit table
            int         j, m = _classInfo.getFieldInfoCount();
            LyFieldInfo fieldInfo;

            for (j = 0; j < m; j++)
            {
                fieldInfo = _classInfo.getFieldInfo(j);

                if ((((fields != null) && !fields.contains(fieldInfo)) || !canBeEditable(fieldInfo))
                        && ((_createFieldMarks == null)
                            || fieldInfo.hasMarks(_createFieldMarks, false))
                        && !fieldInfo.hasMark(LyFieldInfo.MARK_OPTIONAL))
                {
                    match = false;
                    continue;
                }
            }

            if (match)
            {
                _createAction = createActionTmp;
                break;
            }
        }
    } // end method checkCreateAction

    /**
     * Gets the maximum number of objects allowed in this table.
     *
     * @return the maximum number of objects allowed in this table.
     */
    protected int getMaxObjects()
    {
        return -1;
    }

    /**
     * Go to next or previous page.
     *
     * @param next true if we go to next page.
     */
    public void loadPage(boolean next)
    {
        // If page is changed, free the old controllers (new ones will be rebuilt)
        freeControllers();

        super.loadPage(next);
    }

    /**
     * Indicates if the action has a view limit.
     *
     * @return true if the action has a view limit, false otherwise
     */
    public boolean hasViewLimit()
    {
        // MAZ SITA FFT00039
        String viewLimitParameter = (String) _action.getParameter("_viewLimit");

        if (viewLimitParameter != null)
            return true;
        else
            return false;
    }

    /**
     * Set current page being displayed If this page is greater than the lastPage value, the method
     * will take the page 'modulo' getLastPage() value. (Circular).
     *
     * @param page current page we want to show
     */
    public void setPage(int page)
    {
        boolean idle = isIdle();

        if (!idle)
            setIdle(true);

        // If page is changed, free the old controllers (new ones will be rebuilt)
        freeControllers();

        super.setPage(page);

        if (!idle)
            setIdle(false);
    }

    /**
     * Overridden to forbid fast edition from edit table.
     * @param object the object to edit
     * @param fieldInfo the fieldInfo to edit
     * @param x X position
     * @param y Y position
     * @return always null
     */
    protected LyController invokeFastEdition(LyObject object, LyFieldInfo fieldInfo, int x, int y)
    {
        return null;
    }

    /**
     * Set the current context for the controller.
     *
     * @param context Current context, or null if no context is set for the view
     * @return true if successfull
     */
    public boolean setContext(LyWorkSpace context)
    {

        // Rebuild the create controller when context is changed (may be needed
        // if controller contains contextual fields).
        if ((context != _context) && (_createController != null))
        {

            if (!_createController.isFreed())
                _createController.free(this);

            _createController = null;
        }

        return super.setContext(context);
    }

    /**
     * Reset the content of the table being shown. This method rebuild the XML structure of the
     * table.
     */
    protected void updateContent()
    {
        super.updateContent();

        // Add row for creation
        LySession session = getSession();

        if (_editTableForRelations && _editable && _enabled && (_createAction != null)
                && session.enableAction(_createAction, this, null, false))
        {
            // Add empty row
            LyInfoList<LyFieldInfo> fields = getFields();
            LyInfoList<LyFieldInfo> statusFields = getStatusFieldInfos();
            LyTable                 table = getTableComponent();
            LyRows                  rows = table.getRows();

            if (rows != null)
            {
                int nRows = rows.getRowCount();
                int maxRows = getMaxObjects();

                if ((maxRows > 0) && (nRows == maxRows))
                    return;
            }

            if (!(_parent instanceof LyMultiCompositionFieldConsultController))
            {

                if (_createController != null)
                {
                    LyObject object = _createController.getObject();

                    if ((object != null) && (_workSpace.contains(object)))
                        return;
                }

                LyRow row = buildRow(null, fields, statusFields);

                if (rows == null)
                {
                    rows = new LyRows();
                    table.setRows(rows);
                }

                if (row != null)
                    rows.addRow(row);
            }
        }
    } // end method updateContent

    /**
     * Current builder of a new row.
     *
     * @param  object       object added in the contextual list, and which should set the content of
     *                      row cells.
     * @param  fields       list of fields which should be seen.
     * @param  statusFields list of fields which may set the status icon.
     * @return LyRow the new Row description (XML description) which is added in the LyTable
     *         component.
     * @see    #insertRow
     */
    public LyRow buildRow(LyObject object, LyInfoList<LyFieldInfo> fields,
        LyInfoList<LyFieldInfo> statusFields)
    {

        if (!_editTableForRelations || !_editable || !_enabled)
        {
            // Always add a column status
            LyRow row = super.buildRow(object, fields, statusFields);

            return row;
        }

        LyEditTableSetController setController = getSetController(object);

        if (setController == null)
            setController = setControllerFactory(object);

        LyRow row = super.buildRow(object, fields, statusFields);

        if (setController == null)
            return row;

        // Now request service from setController : set content of field.
        StringBuffer label = new StringBuffer();
        StringBuffer tooltip = new StringBuffer();
        String       image = getStatusImage(setController, object, label, tooltip);

        setRowTool(row, object, image, label.toString(), tooltip.toString());

        return row;
    } // end method buildRow

    /**
     * Retrieve the status action for an object.
     *
     * @param  object object of the action.
     * @return the status action for this object or null if there is no status action on this
     *         object.
     */
    public LyAction getStatusAction(LyObject object)
    {
        LyClassInfo          classInfo = object.getClassInfo();
        LyInfoList<LyAction> actions = classInfo.getActions(LyAction.MARK_STATUS);

        if ((actions == null) || (actions.getSize() == 0))
            return null;

        return actions.getInfo(0);
    }

    /**
     * Updates the toolbar.
     */
    protected void updateTools()
    {

        if (_component == null)
            return;

        super.updateTools();

        // Update ToolBar
        LyToolBar toolBar = (LyToolBar) _component.getComponent("_toolBar");

        if (toolBar == null)
            return;

        // Update cancel row and validate row tools
        boolean enabled = _editable && _enabled;
        int     ns = ((_selections == null) ? 0 : _selections.getSize());
        boolean validateRowEnabled = enabled && (ns == 1);
        boolean cancelRowEnabled = validateRowEnabled;
        LyTool  validateTool = (LyTool) toolBar.getComponent(DEMAND_VALIDATE_ROW);
        LyTool  cancelTool = (LyTool) toolBar.getComponent(DEMAND_CANCEL_ROW);

        if (validateRowEnabled)
        {
            LyObject                 object = _selections.getObject(0);
            LyEditTableSetController setController = getSetController(object);

            if ((setController == null) || (setController.isFreed()))
            {
                validateRowEnabled = false;
                cancelRowEnabled = false;
            }
            else if (!setController.isEnabled() || !setController.isEditable())
            {
                validateRowEnabled = false;
                cancelRowEnabled = false;
            }
            else if (!setController.isValid() || !_isCurrentRowValid)
            {
                validateRowEnabled = false;
                cancelRowEnabled = true;
            }
            else if (setController.wasChanged())
            {
                validateRowEnabled = ((object != null) && (setController != _createController));
                cancelRowEnabled = true;
            }
            else
            {
                validateRowEnabled = false;
                cancelRowEnabled = false;
            }
        }

        if (validateTool != null)
            validateTool.setEnabled(validateRowEnabled);

        if (cancelTool != null)
            cancelTool.setEnabled(cancelRowEnabled);

        // Update "Copy cells" and "Paste cells" items
        boolean   hasMenu = hasMenu();
        LyMenuBar menuBar = null;

        if (hasMenu)
        {
            LyView view = getView();
            LyItem item = null;

            menuBar = ((view instanceof LyFrame) ? ((LyFrame) view).getMenuBar() : null);

            if (menuBar != null)
            {
                item = menuBar.findItem(DEMAND_COPY_CELLS, getId());

                if (item != null)
                    item.setEnabled(_selectedCells != null);

                item = menuBar.findItem(DEMAND_PASTE_CELLS, getId());

                if (item != null)
                    item.setEnabled((_selectedCells != null) && (_bufferedValues != null));
            }
        }
    } // end method updateTools

    /**
     * Update the given row with the new values.
     *
     * @see leon.control.LySimpleTableController#updateRow(leon.data.LyObject, int,
     *      leon.info.LyInfoList)
     */
    protected LyRow updateRow(LyObject object, int index, LyInfoList<LyFieldInfo> statusFields)
    {
        LyRow row = getRow(object, index);

        // If multi-page, object may not be found in the list : nothing to do
        if (row == null)
            return null;

        boolean status = isStatusColumn(statusFields);

        LyEditTableSetController setController = getSetController(object);

        if (!_editable || !_enabled || (setController == null) || _displayList)
            super.updateRow(object, index, statusFields);
        else
        {
            LyFieldInfo              fieldInfo;
            LyTable                  table = getTableComponent();
            LyHeaders                headers = table.getHeaders();
            int                      i, n = headers.getHeaderCount();
            LyComponent              comp;
            int                      pos;
            int                      offset = (status ? 1 : 0);
            LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();

            for (i = 0; i < n; i++)
            {
                pos = i + offset; // Skip status column and multiple choice fields

                fieldInfo = getFieldInfo(pos);
                comp = row.getComponent(pos);

                if ((fieldInfo instanceof LyChoiceFieldInfo)
                        && ((LyChoiceFieldInfo) fieldInfo).isMultiple() && canBeEditable(fieldInfo))
                {
                    LyChoiceFieldInfo          choiceInfo = (LyChoiceFieldInfo) fieldInfo;
                    LyInfoList<LyChoiceOption> tableOptions = choiceInfo.getTableOptions();

                    offset += tableOptions.getSize() - 1;
                }

                if (fieldInfo == null)
                    continue;

                if (!(comp instanceof LyCell))
                    continue;

                LyCell cell = (LyCell) comp;

                cell.setColor(null);

                behavior.updateCell(this, cell, object, fieldInfo);

                if (setController.getChangedFields().contains(fieldInfo))
                {

                    if ((object != null) && isEditable(object, fieldInfo)
                            && canBeEditable(fieldInfo))
                    {

                        if (_dynamicEditionMode && setController.wasChanged())
                            cell.setEditable(allowCellEdition(object, fieldInfo));
                    }
                    else
                        ((LyCell) comp).setColor(getModificationsColor());
                }
            }
        }

        updateRowStatusAndColor(row, object, statusFields);

        if (!_editTableForRelations || !_editable || !_enabled)
        {

            if (object != null)
                row.setHidden(object.isHidden());

            return row;
        }

        if (setController == null)
            setController = setControllerFactory(object);

        // Now request service from setController : set content of field.
        StringBuffer label = new StringBuffer();
        StringBuffer tooltip = new StringBuffer();
        String       image = getStatusImage(setController, object, label, tooltip);

        setRowTool(row, object, image, label.toString(), tooltip.toString());

        if (object != null)
            row.setHidden(object.isHidden());

        updateTools();

        return row;
    } // end method updateRow

    /**
     * Updates the cell for given object and field info with given value.
     *
     * @param object    the object used to retrieve cell to update.
     * @param fieldInfo the field used to retrieve cell to update.
     * @param value     the new value of the cell.
     */
    protected void updateCellValue(LyObject object, LyFieldInfo fieldInfo, Object value)
    {
        LyRow row = getRow(object, -1);

        if (row == null)
            return;

        int         colIndex = getColumnIndex(fieldInfo);
        LyComponent component = row.getComponent(colIndex);

        if (component instanceof LySetLabelInterface)
        {
            LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();
            String                   label = behavior.getCellLabel(this, object, fieldInfo, value);

            ((LySetLabelInterface) component).setLabel(label);
        }
    }

    /**
     * Updates the edition status of a given cell.
     *
     * @param object the object of the cell that must be updated.
     * @param fieldInfo the field info of the cell that must be updated.
     * @param editable true if cell must be editable, false otherwise.
     */
    protected void updateCellEditionStatus(LyObject object, LyFieldInfo fieldInfo, boolean editable)
    {
        LyRow row = getRow(object, -1);

        if (row == null)
            return;

        int         colIndex = getColumnIndex(fieldInfo);
        LyComponent component = row.getComponent(colIndex);

        if (component instanceof LyCell)
            ((LyCell) component).setEditable(editable);
    }

    /**
     * Changes the button at the end of the given row.<br />
     * This method call the super method only if table is editable and enabled.
     *
     * @see LySimpleTableController#setRowTool(LyRow, LyObject, String, String, String)
     */
    protected void setRowTool(LyRow row, LyObject object, String image, String label,
        String tooltip)
    {
        LyEditTableBehaviorInterface behavior = (LyEditTableBehaviorInterface) getBehavior();

        // Add a tool at the end of the row only if editable table
        if (!_editable || !_enabled || !behavior.displayRowTools(this))
            return;

        super.setRowTool(row, object, image, label, tooltip);
    }

    /**
     * Get status image, label and tooltip for the given object.
     *
     * @param setController the controller managing given object.
     * @param object        the object whose image must be retrieved.
     * @param label         the label for given object.
     * @param tooltip       the tooltip for given object.
     * @return the status image
     */
    protected String getStatusImage(LyEditTableSetController setController, LyObject object,
        StringBuffer label, StringBuffer tooltip)
    {
        String prefix = null;

        if ((setController == null) || !setController.isInitialized() || !setController.isEnabled()
                || !setController.isEditable() || setController.isFreed())
            prefix = null;
        else if (!setController.isValid() || !_isCurrentRowValid)
        {
            // Cancel a row with errors
            prefix = "LY_CANCEL_ROW";
            label.append("LY_EDIT_TABLE_CANCEL");
            tooltip.append("LY_CANCEL_CHANGES");
        }
        else if (setController.wasChanged())
        {

            // Validate a row (creation or modification)
            if ((object == null) || (setController == _createController))
            {
                prefix = "LY_VALIDATE_ROW_CREATION";
                label.append("LY_EDIT_TABLE_CREATE");
                tooltip.append("LY_CREATE_OBJECT");
            }
            else
            {
                prefix = "LY_VALIDATE_ROW";
                label.append("LY_EDIT_TABLE_VALIDATE");
                tooltip.append("LY_VALIDATE_CHANGES");
            }
        }
        else if ((object != null) && (setController != _createController))
        {

            if (TOOLBAR_FOR_UNCHANGED_OBJECTS)
            {
                // Delete a row
                LyClassInfo        classInfo = object.getClassInfo();
                LyAction           deleteAction = classInfo.getActionWithId(LyAction.ID_DELETE);
                LySimpleObjectList list = new LySimpleObjectList(object);
                LySession          session = getSession();

                if ((deleteAction != null) && (!deleteAction.hasMark(LyAction.MARK_HIDDEN))
                        && session.enableAction(deleteAction, this, list, false))
                {
                    prefix = "LY_DELETE_ROW";
                    label.append("LY_EDIT_TABLE_DELETE");
                    tooltip.append("LY_DELETE_OBJECT");
                }
            }
        }
        else if (_createAction != null)
        {
            // Create a new object
            prefix = "LY_VALIDATE_ROW_CREATION";
            label.append("LY_EDIT_TABLE_CREATE");
            tooltip.append("LY_CREATE_OBJECT");
        }

        if (prefix == null)
            return null;
        else
            return getImageFromPrefix(prefix, DEFAULT_IMAGE_FORMAT, 12);
    } // end method getStatusImage

    /**
     * Get the color resource for cells containing modifications.
     * @return the color for modified cells
     */
    public String getModificationsColor()
    {
    	if (_autoCommit)
    		return null;

        return "LY_TABLE_CHANGED_LINE_COLOR";
    }

    /**
     * Forces the width of a column.<br />
     * When a column is fixed, its width will not be computed with an approximation made with data.
     *
     * @param fieldInfo the field whose coulmn width must be fixed.
     */
    public void setFixedColumnWidth(LyFieldInfo fieldInfo)
    {

        if (_fixedColumnWidth == null)
            _fixedColumnWidth = new LyInfoList<LyFieldInfo>(false);

        _fixedColumnWidth.addInfo(fieldInfo);
    }

    /**
     * Current builder of a new cell in a row.
     *
     * @param  object object added in the contextual list, and which should set the content of row
     *                cells.
     * @param  fieldInfo  field value shown in the cell content
     * @return LyCell or LyFieldIntem the new Cell description (XML description) or editable content
     *         to add in the LyRow component.
     */
    protected LyComponent buildRowItem(LyObject object, LyFieldInfo fieldInfo)
    {

        if (!_editable || !_enabled)
            return super.buildRowItem(object, fieldInfo);

        LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();
        LyComponent              rowItem = behavior.buildRowItem(this, object, fieldInfo);

        if (rowItem != null)
            return rowItem;

        // Overriden to use editable field
        LyEditTableSetController setController = getSetController(object);

        if ((setController == null)
                || (!setController.isInitialized()
                    && (!(fieldInfo instanceof LyChoiceFieldInfo) || !fieldInfo.isMultiple())))
            // Build a simple label
            return super.buildRowItem(object, fieldInfo);

        if (isEditable(object, fieldInfo) && canBeEditable(fieldInfo))
        {
            LyFieldController fieldController = setController.getFieldController(fieldInfo);
            LyComponent       component = fieldController.getComponent();
            boolean           setColumnWidth = true;

            LyTable   table = getTableComponent();
            LyHeaders headers = table.getHeaders();
            int       index = getColumnIndex(fieldInfo);
            LyHeader  header = headers.getHeader(index);
            double         initialColumnWidth = -1;
            initialColumnWidth = header.getColumnWidth();

            // If column is fixed, do not compute its width from data
            if ((_fixedColumnWidth != null) && _fixedColumnWidth.contains(fieldInfo))
            {
                setColumnWidth = false;
            }

            if (initialColumnWidth  > 1)
            {
                setColumnWidth = false;
            }

            if (component == null)
                throw new IllegalStateException(
                    "BF AppEngine : can't retrieve formField for fieldInfo " + fieldInfo);

            LyFieldItem item = null;

            if (component instanceof LyFormField)
            {
                LyFormField field = (LyFormField) component;
                String      id = field.getId();
                int         columnWidth = -1;

                field.setCaption(null);
                field.setConstraint(null);
                item = field.getFieldItem();

                if (item instanceof LyText)
                {
                    // Reduce text width
                    LyText text = (LyText) item;
                    int    nRows = text.getNRows();
                    int    ncols = text.getNCols();
                    String label = text.getLabel();

                    if (ncols < 0)
                        ncols = LyText.DEFAULT_NCOLS;

                    if (ncols > 30)
                        text.setNCols(ncols / 2);

                    if (nRows > 3)
                        text.setNRows(3);

                    columnWidth = text.getNCols();

                    if( LyText.Type.RICH == text.getType())
                    {
                        columnWidth = Math.max(columnWidth, LyText.DEFAULT_RICH_NCOLS_EDITTABLE);
                    }

                    if (setColumnWidth)
                    {
                        columnWidth = Math.max(columnWidth,
                                ((label == null) ? 0
                                                 : Math.min(LyText.DEFAULT_NCOLS_EDITTABLE, label.length())));
                    }
                    else if (initialColumnWidth > 1)
                    {
                        text.setNCols((int)Math.floor(initialColumnWidth));
                    }

                    if (fieldInfo instanceof LyTextFieldInfo
                            && ((LyTextFieldInfo)fieldInfo).isMultilingual())
                    {
                        columnWidth = columnWidth + ACTION_ICON_SIZE;
                    }
                }
                else if ((fieldInfo instanceof LyFileFieldInfo) && (getViewManager() instanceof LyJQViewManager))
                {
                    columnWidth = 50;
                }
                else if (item instanceof LyNumber)
                {
                    // Remove unit
                    ((LyNumber) item).setUnit(null);
                    columnWidth = ((LyNumber) item).getMaxChars();
                }
                else if (item instanceof LyDate)
                {
                    columnWidth = LyDate.EDITTABLE_DATE_MIN_SIZE;
                }
                else if (item instanceof LyRelation)
                {
                    // Check that at least one link is selected
                    LyRelation relation = (LyRelation) item;
                    int        i, selectedIndex = -1, n = relation.getLinkCount();
                    String     label;
                    int maxLabelLength = -1;

                    for (i = 0; i < n; i++)
                    {
                        if (relation.getLink(i).isSelected())
                        {
                            selectedIndex = i;
                            if (_initialized)
                                break;
                        }

                    }

                    for (int j = 0; j < n; j++)
                    {
                        label = relation.getLink(j).getLabel();
                        // Only translate choice values
                        if ((fieldInfo instanceof LyChoiceFieldInfo))
                        {
                            label = translate(label);
                        }
                        if (label.length() > maxLabelLength)
                        {
                            maxLabelLength = label.length();
                        }
                    }

                    if (maxLabelLength > 0)
                    {
                        ((LyRelation) item).setNCols(maxLabelLength);

                        if (maxLabelLength > LyRelation.DEFAULT_WIDTH_EDITTABLE)
                            maxLabelLength = LyRelation.DEFAULT_WIDTH_EDITTABLE;

                        columnWidth = maxLabelLength;
                        if (fieldInfo instanceof LyRelationFieldInfo)
                            columnWidth = maxLabelLength + (2 * ACTION_ICON_SIZE);
                        if (fieldInfo instanceof LyChoiceFieldInfo)
                            columnWidth = maxLabelLength + ACTION_ICON_SIZE;
                    }

                    if (selectedIndex < 0)
                    {
                        int indexToSelect = -1;

                        // we have to set the correct value
                        if (object != null)
                        {
                            Object value = setController.getValue(fieldInfo);

                            if (value == null)
                                value = object.getValue(fieldInfo);

                            if (value != null)
                            {
                                if (value instanceof LyObjectId)
                                    value = ((LyObjectId) value).getFullString();
                                for (i = 0; i < n; i++)
                                {
                                    if (relation.getLink(i).getId().equals(value))
                                    {
                                        indexToSelect = i;
                                        break;
                                    }
                                }
                            }
                        }

                        if (indexToSelect < 0)
                            ((LyRelation) item).selectLink(0, true);
                        else
                            ((LyRelation) item).selectLink(indexToSelect, true);
                    }
                }

                // Set column width with the approximation made with data
                // doesn't work in list mode because there is no headers.
                if (_editable && _enabled && setColumnWidth && (columnWidth > 0) && !_displayList)
                {
                    int headerLabelLength = translate(header.getLabel()).length();

                    if (headerLabelLength < columnWidth)
                    {
                        header.setColumnWidth(columnWidth);
                    }
                }

                if (!id.endsWith(fieldController.getId()))
                    field.setId(id + fieldController.getId());
            }
            else if (component instanceof LyForm)
                item = (LyForm) component;

            boolean useComponent = (component instanceof LyForm);

            // In dynamic edition mode, if cell is not being edited, retrieve
            // object value from setController then compute the label to be displayed.
            if (_dynamicEditionMode && !useComponent
                    && ((_editedObject != object) || (_editedFieldInfo != fieldInfo)))
            {
                LyComponent result = super.buildRowItem(object, fieldInfo);

                if (result instanceof LyCell)
                {
                    LyCell cell = (LyCell) result;
                    Object formValue = setController.getValue(fieldInfo);
                    Object objectValue = ((object == null) ? null : object.getValue(fieldInfo));

                    if (!(fieldInfo instanceof LyChoiceFieldInfo) || !fieldInfo.isMultiple())
                    {
                        if (((formValue == null) && (objectValue != null))
                                || ((formValue != null) && (objectValue == null))
                                || ((formValue != null) && !formValue.equals(objectValue)))
                            cell.setLabel(behavior.getCellLabel(this, object, fieldInfo,
                                    formValue));

                        if (setController.getChangedFields().contains(fieldInfo)
                                && (cell.getColor() == null))
                            cell.setColor(getModificationsColor());
                    }
                    cell.setEditable(allowCellEdition(object, fieldInfo));
                }
                return result;
            }
            return component;
        }

        LyCell  cell = (LyCell) super.buildRowItem(object, fieldInfo);
        LyValue value = setController.getFieldValue(fieldInfo);

        // Do not override the displayed label unless the  value
        // is computed (it may be different from the object's value).
        if ((value != null)
                && (fieldInfo.isComputed() || fieldInfo.isImported()
                    || fieldInfo.hasMark(LyFieldInfo.MARK_READ_ONLY)))
            cell.setLabel(behavior.getCellLabel(this, object, fieldInfo, value.getValue()));

        return cell;
    } // end method buildRowItem


    @Override
    protected void setColumnSizes(String columnSizes)
    {
        super.setColumnSizes(columnSizes);
        reset(true);
    }
    /**
     * Adds a component at the end of a row.
     *
     * @param row       the row instance where the item must be added.
     * @param comp      the component to be added.
     * @param object    the object corresponding to the given row.
     * @param fieldInfo the fieldInfo corresponding to the added component.
     */
    protected void addRowItem(LyRow row, LyComponent comp, LyObject object, LyFieldInfo fieldInfo)
    {

        // Case of multiple choice field info
        if ((fieldInfo instanceof LyChoiceFieldInfo) && ((LyChoiceFieldInfo) fieldInfo)
                .isMultiple() && canBeEditable(fieldInfo))
        {
            LyChoiceFieldInfo          choiceInfo = (LyChoiceFieldInfo) fieldInfo;
            LyInfoList<LyChoiceOption> tableOptions = choiceInfo.getTableOptions();
            int                        i, n = tableOptions.getSize();

            for (i = 0; i < n; i++)
            {
                LyForm form = (LyForm) comp;

                super.addRowItem(row, form.getFormField(i), object, fieldInfo);
            }
        }
        else
            super.addRowItem(row, comp, object, fieldInfo);
    }

    /**
     * utility method (factory) to create setControllers related to objects being modified This
     * method may be overriden in subclasses.
     *
     * @param  object  current object being modified
     * @return current setController. If no set controller is currently related, a new one is
     *         created.
     */
    public LyEditTableSetController setControllerFactory(LyObject object)
    {
        // Keep caching on setControllers (do not create a new one each time)
        LyEditTableSetController setController = getSetController(object);

        if (setController != null)
            throw new IllegalStateException("BF AppEngine : a controller is already registered for "
                + object);

        LyAction           action;
        LySimpleObjectList list = null;
        LyClassInfo        classInfo;

        if (object != null)
        {
            classInfo = object.getClassInfo();
            action = classInfo.getActionWithId(LyAction.ID_SET);

            if (action == null)
                throw new IllegalStateException("BF AppEngine : can't edit objects for classInfo "
                    + classInfo + " because this classInfo doesn't have a set action");

            list = new LySimpleObjectList(object);
        }
        else
        {
            classInfo = getClassInfo();
            action = classInfo.getActionWithId(LyAction.ID_CREATE);

            if (action == null)
                throw new IllegalStateException("BF AppEngine : can't create objects for classInfo "
                    + classInfo + " because this classInfo doesn't have a create action");
        }

        // Create a new set Controller for the given object
        setController = new LyEditTableSetController(getSession(), this, action, classInfo, list);

        if (object == null)
            _createController = setController;
        else
        {

            if (_setControllers == null)
                _setControllers = new Hashtable<LyObject, LyEditTableSetController>();

            _setControllers.put(object, setController);
        }

        setController.setEnabled(_enabled);
        setController.setEditable(_editable);

        // We need to transfer the marks to the set/create controller
        if (object == null) // creation
        {
            setController.initArea();
            setController.setFieldMarks(_createFieldMarks);
        }
        else
            setController.setFieldMarks(_setFieldMarks);

        if (setController.isFreed())
        {

            if ((_setControllers != null) && _setControllers.contains(setController))
                _setControllers.remove(object);
            else if (setController == _createController)
                _createController = null;

            return null;
        }

        // Call the getFields() method on the controller to be sure that all the
        // fields will be loaded.
        LyInfoList<LyFieldInfo> fields = setController.getFields();
        LyFieldInfo             fieldInfo;
        boolean                 hasMultipleChoice = false;
        int                     i, n = ((fields == null) ? 0 : fields.getSize());

        for (i = 0; i < n; i++)
        {
            fieldInfo = fields.getInfo(i);

            if ((fieldInfo instanceof LyChoiceFieldInfo) && fieldInfo.isMultiple())
            {
                hasMultipleChoice = true;
                break;
            }
        }

        if (hasMultipleChoice)
            setController.initArea();

        return setController;
    } // end method setControllerFactory

    /**
     * Retrieves the view Builder.
     *
     * @return the builder instance;
     */
    protected LyBuilder getDefaultBuilder()
    {
        return new LyEditTableBuilder(this);
    }

    /**
     * Invoked when an action occurs.
     *
     * @param demand selection event created from the view. User interaction.
     * @see   LyController#selectionDemand
     */
    public void selectionDemand(LySelectionDemand demand)
    {
        // Reset any cell selection
        LyTable        table = getTableComponent();
        Vector<LyCell> selectedCells = new Vector<LyCell>();
        LyCell         cell;

        table.getComponents(LyCell.class, selectedCells);

        for (int i = 0; i < selectedCells.size(); i++)
        {
            cell = selectedCells.elementAt(i);

            if (cell.isSelected())
                cell.setSelected(false);
        }

        LyObjectList oldSelection = getSelection();
        LyObject     oldEditedObject = _editedObject;
        int[]        oldSelectedCells = ((_selectedCells == null) ? null : new int[]
                {
                    _selectedCells[0], _selectedCells[1], _selectedCells[2], _selectedCells[3]
                });

        _selectedCells = null;

        super.selectionDemand(demand);

        if (!_useSingleClickForEdition)
        {
            LyObjectList selection = getSelection();

            // If selection has changed, stop cell edition
            if ((oldSelection != null) && !(oldSelection.equals(selection)) && _dynamicEditionMode
                    && (oldEditedObject != null))
                startCellEdition(null, -1);
        }
        else
        {

            // If selection has changed, stop cell edition
            if (demand.getSubIndex() >= 0)
            {
                String      rowId = demand.getSelectionId(0);
                int         rowIndex = demand.getIndex(0);
                int         colIndex = demand.getSubIndex();
                LyFieldInfo oldEditedField = _editedFieldInfo;

                if ((oldSelection != null) && _dynamicEditionMode && (oldEditedObject != null))
                    startCellEdition(null, -1);

                LyRows rows = table.getRows();

                if (demand.isExtend() && ((oldEditedField != null) || (oldSelectedCells != null)))
                {
                    int oldRowIndex;
                    int oldColIndex;

                    if (oldEditedField != null)
                    {
                        LyRow oldRow = getRow(oldEditedObject, -1);

                        oldRowIndex = ((rows == null) ? 0 : rows.getRowIndex(oldRow));
                        oldColIndex = getColumnIndex(oldEditedField);

                        if (oldEditedObject == null)
                            oldRowIndex = rows.getRowCount() - 1;

                        if (rowIndex < 0)
                            rowIndex = oldRowIndex;
                    }
                    else
                    {
                        oldRowIndex = oldSelectedCells[0];
                        oldColIndex = oldSelectedCells[1];
                    }

                    // For an extended demand, select cells between old edited cell and selected cell
                    startCellEdition(null, -1);
                    selectObjects(null, false);

                    int   rowStart = Math.min(rowIndex, oldRowIndex);
                    int   rowEnd = Math.max(rowIndex, oldRowIndex);
                    int   colStart = Math.min(colIndex, oldColIndex);
                    int   colEnd = Math.max(colIndex, oldColIndex);
                    LyRow row;

                    for (int i = rowStart; i <= rowEnd; i++)
                    {

                        for (int j = colStart; j <= colEnd; j++)
                        {
                            row = rows.getRow(i);

                            if (row.getComponent(j) instanceof LyCell)
                            {
                                cell = (LyCell) row.getComponent(j);
                                cell.setSelected(true);
                            }
                        }
                    }

                    _selectedCells = new int[] { rowStart, colStart, rowEnd, colEnd };
                    updateTools();
                }
                else
                {
                    LyObjectId objectId = ((rowId == null) ? null : getObjectId(rowId));
                    LyObject   object = ((objectId == null) ? null
                                                            : _workSpace.getObject(objectId));
                    boolean    result = startCellEdition(object, colIndex);

                    if (!result)
                    {
                        // Cell is not edited : select the line
                        startCellEdition(null, -1);
                        super.selectionDemand(demand);
                    }
                }
            }
            else if (_autoCommit && (demand.getSelectionSize() <= 0) && !getViewManager().hasProperty("WEB_MANAGER"))
                startCellEdition(null, -1);
        }
    } // end method selectionDemand

    /**
     * Invoked when an action notification is sent to this listener.
     *
     * <P>This method is called by demandPerformed and may be overriden in subclasses.</P>
     *
     * <P>Default implementation makes nothing.</P>
     *
     * @param demand demand parameters sent by the view component.
     * @see   #demandPerformed
     */
    public void actionDemand(LyActionDemand demand)
    {

        if (!_editable || !_enabled)
        {
            super.actionDemand(demand);
            return;
        }

        // Also protect for empty or null pointer !
        if (demand.getType() == LyActionDemand.TYPE_ACTION)
        {
            String source = demand.getSource();

            if (source.equals(LyFormController.DEMAND_VALIDATE)
                    || source.equals(LyFormController.DEMAND_APPLY))
            {
                enableShowMessage(true); // May raise only one error
                setLastModified((new Date()).getTime());

                short result = validate();

                if ((result != LyFormInterface.STATUS_KO)
                        && source.equals(LyFormController.DEMAND_VALIDATE)
                        && !getAction().hasMark(LyAction.MARK_NO_COMMANDS)
                        && !getAction().hasMark(LyAction.MARK_FRAME_VIEW))
                    free();

                return;
            }
            else if (source.equals(LyFormController.DEMAND_CANCEL)
                    || source.equals(LyEditTableController.DEMAND_RESTORE))
            {
                setLastModified((new Date()).getTime());

                LyAction action = getAction();

                if ((action.hasMark(LyAction.MARK_FRAME_VIEW)
                            && source.equals(LyEditTableController.DEMAND_RESTORE))
                        || (!action.hasMark(LyAction.MARK_FRAME_VIEW)
                            && source.equals(LyEditTableController.DEMAND_CANCEL)))
                {
                    // If dialog, the controller will be freed
                    cancel(true);

                    if (_dynamicEditionMode)
                        startCellEdition(null, -1);

                    if (!action.hasMark(LyAction.MARK_FRAME_VIEW)
                            && source.equals(LyEditTableController.DEMAND_CANCEL))
                        free();

                    return;
                }
            }
            else if (source.equals(DEMAND_VALIDATE_ROW) || (source.equals(DEMAND_CANCEL_ROW)))
            {
                _showMessages = true;

                int ns = ((_selections == null) ? 0 : _selections.getSize());

                if (ns != 1)
                    return;

                LyObject                 object = _selections.getObject(0);
                LyEditTableSetController setController = getSetController(object);

                if (demand.getSource().equals(DEMAND_VALIDATE_ROW))
                    validateRow(setController);
                else if (demand.getSource().equals(DEMAND_CANCEL_ROW))
                    cancelRow(setController);

                _showMessages = false;
                return;
            }
            else if (source.startsWith(DEMAND_ROW_TOOL))
            {

                // Ignore event if an error resulting from a focusLost was shown some ms ago
                if (((new Date()).getTime() - _errorTime) < 500)
                    return;

                String     objectKey = source.substring(DEMAND_ROW_TOOL.length());
                LyObjectId objectId = ((objectKey.length() > 0) ? getObjectId(objectKey) : null);
                LyObject   object = ((objectId == null) ? null : objectId.getLogicalObject());

                _showMessages = true;

                if (!_editTableForRelations || !_editable || !_enabled)
                {
                    LyAction action = getStatusAction(object);

                    if (action == null)
                        action = getDefaultAction(object);

                    if (action != null)
                        invokeAction(action, object);
                }
                else
                {
                    LyEditTableSetController setController = getSetController(object);

                    if ((setController == null) || !setController.isEnabled()
                            || !setController.isEditable())
                        super.actionDemand(demand);
                    else if (!setController.isValid() || !_isCurrentRowValid)
                        // Cancel changes
                        cancelRow(setController);
                    else if (setController.wasChanged() || setController.isCreation())
                    {
                        // Call the valueChanged() method on the set controller to perform any needed propagation
                        // on other fields
                        if ((_editedFieldInfo != null)
                                && (_editedObject == setController.getObject()))
                        {
                            setController.valueChanged(_editedFieldInfo);
                            _showMessages = true;
                        }

                        // Validate changes
                        validateRow(setController);
                    }
                    else if ((object != null) && TOOLBAR_FOR_UNCHANGED_OBJECTS)
                    {

                        // Delete object
                        if (!deleteRow(setController))
                            super.actionDemand(demand);
                    }
                }

                _showMessages = false;
                return;
            }
            else if (source.equals(LyEditTableController.DEMAND_ALT_UP))
            {
                startCellEdition(false, true, demand.getValue());
                return;
            }
            else if (source.equals(LyEditTableController.DEMAND_ALT_DOWN))
            {
                startCellEdition(true, true, demand.getValue());
                return;
            }
            else if (source.equals(LyEditTableController.DEMAND_ALT_LEFT))
            {
                startCellEdition(false, false, demand.getValue());
                return;
            }
            else if (source.equals(LyEditTableController.DEMAND_ALT_RIGHT))
            {
                startCellEdition(true, false, demand.getValue());
                return;
            }
            else if (source.equals(LyEditTableController.DEMAND_ALT_ADD))
            {
                copyUpperValue();
                return;
            }
            else if (source.equals(LyEditTableController.DEMAND_COPY_CELLS))
            {

                if ((_selectedCells != null) && (_selectedCells.length == 4))
                {
                    LyObject                 object;
                    LyEditTableSetController setController;
                    LyFieldInfo              fieldInfo;

                    _bufferedValues =
                        new LyValue[_selectedCells[2] - _selectedCells[0] + 1][_selectedCells[3]
                            - _selectedCells[1] + 1];

                    for (int i = _selectedCells[0]; i <= _selectedCells[2]; i++)
                    {

                        for (int j = _selectedCells[1]; j <= _selectedCells[3]; j++)
                        {
                            object = getObject(i);
                            setController = getSetController(object);
                            fieldInfo = getFieldInfo(j);
                            _bufferedValues[i - _selectedCells[0]][j - _selectedCells[1]] =
                                setController.getFieldValue(fieldInfo);
                        }
                    }

                    updateTools();
                    return;
                }
            }
            else if (source.equals(LyEditTableController.DEMAND_PASTE_CELLS))
            {

                if ((_selectedCells != null) && (_selectedCells.length == 4)
                        && (_bufferedValues != null))
                {
                    LyTable                  table = getTableComponent();
                    LyRows                   rows = table.getRows();
                    LyRow                    row;
                    LyCell                   cell;
                    LyObject                 object;
                    LyEditTableSetController setController;
                    LyFieldInfo              fieldInfo;
                    int                      bufferRow, bufferColumn;
                    LyValue                  value;

                    for (int i = _selectedCells[0]; i <= _selectedCells[2]; i++)
                    {

                        for (int j = _selectedCells[1]; j <= _selectedCells[3]; j++)
                        {
                            row = rows.getRow(i);
                            cell = (LyCell) row.getComponent(j);

                            // Paste the buffered value only if cell is editable
                            if (cell.isEditable())
                            {
                                bufferRow = (i - _selectedCells[0]) % _bufferedValues.length;
                                bufferColumn = (j - _selectedCells[1]) % _bufferedValues[0].length;
                                object = getObject(i);
                                setController = getSetController(object);
                                fieldInfo = getFieldInfo(j);
                                value = _bufferedValues[bufferRow][bufferColumn];

                                // Pasted value may not be the right type for target field info. Catch the exception
                                // and send a warning.
                                try
                                {
                                    setController.setFieldValue(fieldInfo, value);
                                    setController.valueChanged(fieldInfo);
                                }
                                catch (IllegalArgumentException ex)
                                {
                                    getEnvironment().logWarning("BF AppEngine : fieldInfo "
                                        + fieldInfo + " does not accept value "
                                        + ((value == null) ? "null" : value.getString()));
                                }
                            }
                        }
                    }
                }

                return;
            }
            else if (_editable && _enabled)
            {
                String[] parameters = LyString.stringToList(source, ID_SEP);

                if ((parameters != null) && (parameters.length != 0) && (parameters[0] != null))
                {
                    String id = parameters[0].intern();
                    String key = demand.getValue();

                    if ((id.equals(getId() + "_table"))
                            || ((id.equals("_table")) && (key != null)
                                && (key.indexOf("_toolBar") < 0))) // "_table" action
                    {
                        // Action on a table cell : set the cell editable
                        String  rowId = demand.getValue();
                        int     rowIndex = demand.getIndex(0);
                        int     colIndex = demand.getSubIndex();
                        boolean forceReturn = false;

                        if ((key != null) && key.startsWith("_cellEdition"))
                        {
                            String[] values = LyString.stringToList(key, LyTools.PIPE_SEP);

                            rowIndex = LyTools.intFromString(values[1]);
                            colIndex = LyTools.intFromString(values[2]);

                            // If a cell edition has been asked, do not perform super.actionDemand()
                            // whatever happens.
                            forceReturn = true;

                            LyRow row = getRow(rowIndex);
                            rowId = row.getId();
                        }

                        if (_dynamicEditionMode && (colIndex >= 0) && (rowIndex >= 0))
                        {
                            LyObject object = null;

                            if (rowId != null)
                            {
                                LyObjectId objectId = getObjectId(rowId);
                                object = _workSpace.getObject(objectId);
                            }

                            boolean result = startCellEdition(object, colIndex, false);

                            if (result || forceReturn)
                                return;
                        }
                    }
                }
            }
        }
        else if ((demand.getType() == LyActionDemand.TYPE_SORT)
                || (demand.getType() == LyActionDemand.TYPE_EXTEND_SORT))
        {
            int       index = demand.getIndex(0);
            LyTable   table = getTableComponent();
            LyHeaders headers = table.getHeaders();
            LyHeader  header = headers.getHeader(index);
            String    headerId = header.getId();

            if (headerId == null)
                return;

            int         pos = headerId.indexOf(ID_SEP);
            LyFieldInfo fieldInfo;

            fieldInfo = _classInfo.getFieldInfo(headerId.substring(pos + ID_SEP.length()), false);

            if ((fieldInfo != null) || headerId.equals(DEMAND_STATUS_SORT))
                sort(fieldInfo, (demand.getType() == LyActionDemand.TYPE_EXTEND_SORT));

            return;
        }
        else if ((demand.getType() == LyActionDemand.TYPE_TAB_NEXT)
                || (demand.getType() == LyActionDemand.TYPE_TAB_PREVIOUS))
            startCellEdition(demand.getType() == LyActionDemand.TYPE_TAB_NEXT, demand.getValue());

        super.actionDemand(demand);
    } // end method actionDemand

    /**
     * Starts the edition of a cell (dynamic edition mode only).<br />
     * Depending on the <code>next <code/>parameter, the next or previous editable field of the
     * table will be edited. If the last (or first) editable field of an object is reached, the next
     * (or previous) object of the list will be edited.
     *
     * @param next  indicates if the next or previous editable field will be edited.
     * @param value value sent with the tab event.
     */
    protected void startCellEdition(boolean next, String value)
    {
        startCellEdition(next, false, value);
    }

    /**
     * Starts the edition of a cell (dynamic edition mode only).<br />
     * Depending on the <code>next<code/> and <code>changeRow<code/> parameters, the
     * next or previous editable field of the table will be edited or the same field
     * for next or previous row. If the last (or first) editable field of an object
     * is reached, the next (or previous) object of the list will be edited.
     *
     * @param next indicates if the next or previous editable field will be edited.
     * @param changeRow indicates if we must change edited row.
     * @param value value sent with the tab event.
     */
    protected void startCellEdition(boolean next, boolean changeRow, String value)
    {

        if (_editedFieldInfo != null)
        {
            LyFieldInfo field;
            LyTable     table = getTableComponent();
            LyHeaders   headers = table.getHeaders();
            int         i, n = headers.getHeaderCount();
            int         row = ((_editedObject == null) ? _workSpace.getSize()
                                                       : _workSpace.getIndex(_editedObject));
            int         fieldPos = getColumnIndex(_editedFieldInfo) + _multipleChoiceOffset;
            boolean     isEditable;

            LyEditTableSetController setController = getSetController(_editedObject);

            LyActionDemand demand = new LyActionDemand(LyActionDemand.TYPE_VALUE_CHANGED,
                    _editedFieldInfo.getId());

            if (value != null)
            {
                enableShowMessage(true); // May raise only one error
                demand.setValue(value);
                setController.getFieldController(_editedFieldInfo).actionDemand(demand);
                enableShowMessage(true); // May raise only one error
            }

            if (changeRow)
            {
                LyRows   rows = table.getRows();
                LyObject object;

                while ((row >= 0) && (row < rows.getRowCount()))
                {

                    if (next)
                        row++;
                    else
                        row--;

                    if (row >= 0)
                    {
                        object = ((row < _workSpace.getSize()) ? _workSpace.getObject(row) : null);
                        isEditable = isEditable(object, _editedFieldInfo)
                            && canBeEditable(_editedFieldInfo);

                        // For multiple choices, check that option is enabled
                        if ((_editedFieldInfo instanceof LyChoiceFieldInfo)
                                && _editedFieldInfo.isMultiple())
                        {
                            LyRow       rowInstance = rows.getRow(row);
                            LyFormField formField = (LyFormField) rowInstance.getComponent(
                                    fieldPos);
                            LyChoice    choice = (LyChoice) formField.getFieldItem();
                            LyOption    option = choice.getOption(0);

                            isEditable = option.isEnabled();
                        }

                        if (isEditable)
                        {
                            startCellEdition(object, fieldPos);
                            return;
                        }
                    }
                }

                return;
            }


            int start = (next ? (fieldPos + 1) : (fieldPos - 1));
            int offset = (next ? 1 : -1);

            // Try to find next editable field for currently edited object
            for (i = start; (next ? (i < n) : (i >= 0)); i = i + offset)
            {
                field = getFieldInfo(i);
                isEditable = (field != null) && isEditable(_editedObject, field)
                    && canBeEditable(field);

                // For multiple choices, check that option is enabled
                if ((field instanceof LyChoiceFieldInfo) && field.isMultiple())
                {
                    LyRows      rows = table.getRows();
                    LyRow       rowInstance = rows.getRow(row);
                    LyFormField formField = (LyFormField) rowInstance.getComponent(i);
                    LyChoice    choice = (LyChoice) formField.getFieldItem();
                    LyOption    option = choice.getOption(0);

                    isEditable = option.isEnabled();
                }

                if (isEditable)
                {
                    startCellEdition(_editedObject, i);
                    return;
                }
            }

            // Try next object if no other field was found
            if (next)
            {
                LyObject object = ((row < (_workSpace.getSize() - 1))
                        ? _workSpace.getObject(row + (next ? 1 : -1)) : null);
                short    status = LyFormInterface.STATUS_IGNORE;

                setController = getSetController(_editedObject);
                enableShowMessage(true); // May raise only one error

                if (setController.isValid(_editedFieldInfo))
                    status = validate(setController);

                for (i = 0; i < n; i++)
                {
                    field = getFieldInfo(i);
                    isEditable = (field != null) && canBeEditable(field)
                        && allowCellEdition(object, field);

                    if (isEditable)
                    {

                        if (status != LyFormInterface.STATUS_KO)
                            startCellEdition(object, i);

                        return;
                    }
                }
            }
            else if (!next && (row > 0))
            {
                LyObject object = _workSpace.getObject(row + (next ? 1 : -1));

                for (i = n - 1; i >= 0; i--)
                {
                    field = getFieldInfo(i);
                    isEditable = (field != null) && isEditable(object, field)
                        && canBeEditable(field);

                    if (isEditable)
                    {
                        startCellEdition(object, i);
                        return;
                    }
                }
            }
        }
    }

    /**
     * Copy value from upper cell in edited cell then start edition in the next row for
     * the same field info.
     */
    protected void copyUpperValue()
    {

        if (_editedFieldInfo == null)
            return;

        // Retrieve the value from upper cell
        int                      row = ((_editedObject == null)
                ? _workSpace.getSize() : _workSpace.getIndex(_editedObject));
        boolean                  isEditable;
        LyObject                 previousObject;
        LyEditTableSetController setController = getSetController(_editedObject);

        while (row > 0)
        {
            row--;
            previousObject = _workSpace.getObject(row);
            isEditable = isEditable(previousObject, _editedFieldInfo)
                && canBeEditable(_editedFieldInfo);

            if (isEditable)
            {
                LyEditTableSetController previousController = getSetController(previousObject);

                setController.setFieldValue(_editedFieldInfo,
                    previousController.getFieldValue(_editedFieldInfo));
                setController.valueChanged(_editedFieldInfo);
                break;
            }
        }

        if (_editedObject == null)
            // Start the edition of next cell for the same row if creation line
            startCellEdition(true, false, null);
        else
            // Start the edition of next row for the same field info
            startCellEdition(true, true, null);
    }

    /**
     * Starts the edition of a cell (dynamic edition mode only).
     * This method cancels the edition in the table if called with negative values
     * (for example a call to startCellEdition(null,-1) will cancel the edition).
     *
     * @param  object the object to edit.
     * @param  colIndex the index of the cell column to be edited.
     * @return true if the cell is in edition mode at the end of the treatment, false otherwise.
     */
    public boolean startCellEdition(LyObject object, int colIndex)
    {
    	return startCellEdition(object, colIndex, true);
    }

    /**
     * Starts the edition of a cell (dynamic edition mode only).
     * This method cancels the edition in the table if called with negative values
     * (for example a call to startCellEdition(null,-1) will cancel the edition).
     *
     * @param  object the object to edit.
     * @param  colIndex the index of the cell column to be edited.
     * @param  showRow indicates whether edited row must be shown on view manager.
     * @return true if the cell is in edition mode at the end of the treatment, false otherwise.
     */
    public boolean startCellEdition(LyObject object, int colIndex, boolean showRow)
    {

        // Only implemented in dynamic edition mode
        if (!_dynamicEditionMode)
            return false;


        LyFieldInfo fieldInfo;

        if (isDisplayList())
            fieldInfo = getFieldInfoDisplayList(colIndex);
        else
            fieldInfo = getFieldInfo(colIndex);

        if ((fieldInfo == null) && (colIndex >= 0))
            return false;

        if ((colIndex >= 0) && !allowCellEdition(object, fieldInfo))
            return false;

        boolean isMultipleChoice = (fieldInfo instanceof LyChoiceFieldInfo)
            && fieldInfo.isMultiple();

        // If field and object are the same, do nothing
        if ((object == _editedObject) && (fieldInfo == _editedFieldInfo) && !isMultipleChoice)
            return false;

        LyObject    lastEditedObject = _editedObject;
        LyFieldInfo lastEditedField = _editedFieldInfo;
        LyTable     table = getTableComponent();
        boolean     wasIdle = table.isIdle();

        table.setIdle(true);
        _editedObject = object;
        _editedFieldInfo = null;

        // Cancel the previous edited cell
        updateCellEditionStatus(lastEditedObject, lastEditedField);

        _editedFieldInfo = fieldInfo;
        _lastEditedIndex = colIndex;
        _multipleChoiceOffset = 0;

        // FIXME
        if (isMultipleChoice)
            _multipleChoiceOffset = colIndex - getColumnIndex(fieldInfo);

        // Update the edition status of the cell
        boolean         edited = false;
        LySetController setController = getSetController(_editedObject);

        if (setController != null)
            setController.initArea();

        edited = updateCellEditionStatus(_editedObject, _editedFieldInfo);

        table.setIdle(wasIdle);

        if (showRow && (object != null) && (table.getInternalView() instanceof LySetTableInterface))
        {
            int idx = getRowIndex(object);

            if (idx >= 0)
                ((LySetTableInterface) table.getInternalView()).showRow(idx);
        }

        return edited;
    } // end method startCellEdition

    /**
     * Updates the edition status of a cell corresponding to an object and a field.<br />
     * This method uses the <code>_editedObject</code> and <code>_editedFieldInfo</code> attributes
     * to check whether corresponding cell is editable or not.
     *
     * @param  object    the object of the cell that must be updated.
     * @param  fieldInfo the field info of the cell that must be updated.
     * @return true if the cell is in edition mode at the end of the treatment, false otherwise.
     */
    protected boolean updateCellEditionStatus(LyObject object, LyFieldInfo fieldInfo)
    {

        if (fieldInfo == null)
            return false;

        // Ignore status column
        int colIndex = getColumnIndex(fieldInfo);

        LyRow row = getRow(object, -1);


        if ((fieldInfo instanceof LyChoiceFieldInfo) && fieldInfo.isMultiple())
            colIndex += _multipleChoiceOffset;

        if ((object == null) && (_createController != null))
        {
            // If object is null, we may want to edit the creation line
            LyTable table = getTableComponent();
            LyRows  rows = table.getRows();

            if (rows != null)
                row = rows.getRow(rows.getRowCount() - 1);
        }

        if (row == null)
            return false;


        LyComponent newComponent = buildRowItem(object, fieldInfo);
        LyComponent oldComponent = row.getComponent(colIndex);

        if (isDisplayList())
        {

            if ((_lastEditedIndex == -1) || (_lastEditedIndex <= getDisplayListDataIndex()))
                return false;

            oldComponent = row.getComponent(_lastEditedIndex);
        }

        // Remove old component from the row
        if (oldComponent instanceof LyCell)
            row.removeCell((LyCell) oldComponent);
        else if (oldComponent != null)
        {
            row.removeFormField((LyFormField) oldComponent);
            oldComponent.reset();
        }

        // Insert new component
        if (newComponent instanceof LyCell)
        {
            row.insertCell((LyCell) newComponent, colIndex);

            // Need to update cell for multiple choice field info
            if ((fieldInfo instanceof LyChoiceFieldInfo)
                    && ((LyChoiceFieldInfo) fieldInfo).isMultiple() && canBeEditable(fieldInfo))
            {
                LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();

                behavior.updateCell(this, (LyCell) newComponent, object, fieldInfo);
            }
        }
        else if (newComponent instanceof LyFormField)
            row.insertFormField((LyFormField) newComponent, colIndex);
        else if (newComponent instanceof LyForm) // May occur if choice multiple
            // Insert old component (which should be an option) again.
            row.insertFormField((LyFormField) oldComponent, colIndex);

        return (newComponent instanceof LyFormField) || (newComponent instanceof LyForm);
    } // end method updateCellEditionStatus


    /**
     * Returns the column index of a given field info. If field info is a multiple choice field
     * info, this method returns the index of the first column displaying a value of this field
     * info.
     *
     * @param  fieldInfo the field info to be retrieved.
     * @return the column index of the given field info.
     */
    protected int getColumnIndex(LyFieldInfo fieldInfo)
    {
        LyTable   table = getTableComponent();
        LyHeaders headers = table.getHeaders();
        LyHeader  header;
        int       i, n = headers.getHeaderCount();

        // Try to retrieve field info with the headers id
        for (i = 0; i < n; i++)
        {
            header = headers.getHeader(i);

            if ((header.getId() != null)
                    && header.getId().equals("header" + LyController.ID_SEP + fieldInfo.getId()))
                return i;
        }

        // If field info was not found with the headers id, try to retrieve it from fields list.
        LyInfoList<LyFieldInfo> fields = getFields();
        int                     fieldPos = fields.getIndex(fieldInfo);
        boolean                 status = isStatusColumn(null);

        return fieldPos + (status ? 1 : 0);
    }

    /**
     * Returns the field info at given column index.
     *
     * @param  colIndex the column index.
     * @return the field info displayed at column with the given index.
     */
    protected LyFieldInfo getFieldInfo(int colIndex)
    {
        boolean status = isStatusColumn(null);

        if (colIndex < (status ? 1 : 0))
            return null;

        // Retrieve header at given column index
        LyTable                 table = getTableComponent();
        LyHeaders               headers = table.getHeaders();
        LyInfoList<LyFieldInfo> fields = getFields();

        if (headers == null)
            return ((colIndex <= fields.getSize()) ? fields.getInfo(colIndex - 1) : null);

        if (colIndex >= headers.getHeaderCount())
            return null;

        LyHeader header = headers.getHeader(colIndex);
        String   headerId = header.getId();

        headerId = ((headerId == null)
                ? "" : headerId.substring(("header" + LyController.ID_SEP).length()));

        LyFieldInfo fieldInfo = fields.getInfo(headerId);

        if (fieldInfo == null)
        {
            // If field info was not found, check if header id corresponds to an option
            // of a multiple choice field info
            LyApplication application = getApplication();
            LyInfo        info = application.getInfo(headerId);

            if ((info != null) && (info instanceof LyChoiceOption))
            {
                LyChoiceOption option = (LyChoiceOption) info;
                LyInfo         parent = option.getParent();

                ;

                while ((parent != null) && !(parent instanceof LyChoiceFieldInfo))
                    parent = parent.getParent();

                if (parent instanceof LyChoiceFieldInfo)
                    fieldInfo = (LyChoiceFieldInfo) parent;
            }
        }

        return fieldInfo;
    } // end method getFieldInfo

    /**
     * New sort or extended sort on given index.
     * @param index index of the column
     * @param extend indicates if we must extends the sort
     */
    protected void sort(int index, boolean extend)
    {

        if (index < 0)
            return;

        LyFieldInfo fieldInfo = getFieldInfo(index);

        sort(fieldInfo, extend);
    }

    /**
     * Public method to validate all editions.
     *
     * @return the validation status
     * @see LyFormInterface
     */
    public short validateChanges()
    {
        return validate();
    }

    /**
     * Validate all editions.
     * @return the validation status
     * @see LyFormInterface
     */
    protected short validate()
    {
        short result = startValidate();

        if (result != LyFormInterface.STATUS_OK)
            return result;
        else
            return endValidate();
    }

    /**
     * Start validating edition of given controller.
     * @param setController the controller to validate
     * @return the validation status
     * @see LyFormInterface
     */
    protected short startValidate(LyEditTableSetController setController)
    {
        return setController.startValidate();
    }

    /**
     * End Validating edition of given controller.
     * @param setController the controller to validate
     * @return the validation status
     * @see LyFormInterface
     */
    protected short endValidate(LyEditTableSetController setController)
    {
        short result = setController.endValidate();

        if (result != LyFormInterface.STATUS_IGNORE)
            valueChanged(setController, null);

        return result;
    }

    /**
     * Validate edition of given controller.
     * @param setController the controller to validate
     * @return the validation status
     * @see LyFormInterface
     */
    public short validate(LyEditTableSetController setController)
    {
        short result = startValidate(setController);

        if (result != LyFormInterface.STATUS_OK)
            return result;
        else
            result = endValidate(setController);

        if (result == LyFormInterface.STATUS_OK)
            setController.commit();

        if ((setController == _createController) && (result == LyFormInterface.STATUS_OK))
        {
            // If validating the creation line, free create controller and ensure that newly
            // created object is associated to a set controller
            LyObject object = setController.getObject();

            if (!_createController.isFreed())
                _createController.free(this);

            _createController = null;

            LyEditTableSetController newSetController = getSetController(object);

            if (newSetController == null)
                newSetController = setControllerFactory(object);

            reset(false);
        }

        return result;
    } // end method validate

    /**
     * Cancel all editions.
     */
    protected void cancel()
    {
        cancel(false);
    }

    /**
     * Cancel all editions.
     * @param update indicates if we need to update the view
     */
    protected void cancel(boolean update)
    {

        // Free any setController related to this controller
        if (_setControllers != null)
        {
            LyEditTableSetController              setController;
            Enumeration<LyEditTableSetController> e = _setControllers.elements();

            while (e.hasMoreElements())
            {
                setController = e.nextElement();
                cancel(setController, update);
            }
        }

        if (_createController != null)
            cancel(_createController, update);
    }

    /**
     * Internal use. Called when the parent form is starting validation.
     * @return the validation status
     * @see LyFormInterface
     */
    public short startValidate()
    {
        short result = LyFormInterface.STATUS_IGNORE, subResult;

        // Free any setController related to this controller
        if (_setControllers != null)
        {
            LyEditTableSetController              setController;
            Enumeration<LyEditTableSetController> e = _setControllers.elements();

            while (e.hasMoreElements())
            {
                setController = e.nextElement();

                if (!setController.wasChanged())
                    continue;

                subResult = startValidate(setController);

                if (subResult != LyFormInterface.STATUS_IGNORE)
                    valueChanged(setController, null);

                if (subResult == LyFormInterface.STATUS_OK)
                    result = subResult;
                else if (subResult == LyFormInterface.STATUS_KO)
                    return subResult;
            }
        }

        if ((_createController != null) && _createController.wasChanged())
        {
            subResult = startValidate(_createController);

            if (subResult != LyFormInterface.STATUS_IGNORE)
            {
                result = subResult;
                // Force Update of view
                // FFT01124
                valueChanged(_createController, null);
            }
        }

        return result;
    } // end method startValidate

    /**
     * Internal use. Called when the parent form is ending validation.
     * @return the validation status
     * @see LyFormInterface
     */
    public short endValidate()
    {
        short result = LyFormInterface.STATUS_OK;

        // Validate each setController
        if (_setControllers != null)
        {
            LyEditTableSetController              setController;
            Enumeration<LyEditTableSetController> e = _setControllers.elements();

            while (e.hasMoreElements())
            {
                setController = e.nextElement();

                if (endValidate(setController) == LyFormInterface.STATUS_KO)
                    result = LyFormInterface.STATUS_KO;
                else if (result == LyFormInterface.STATUS_OK)
                    setController.commit();
            }
        }

        if ((_createController != null) && _createController.wasChanged())
        {
            short status = endValidate(_createController);

            if (status == LyFormInterface.STATUS_KO)
                result = LyFormInterface.STATUS_KO;
            else
            {
                // Validate the creation line : free create controller and ensure that newly
                // created object is associated to a set controller
                LyObject object = _createController.getObject();

                if (!_createController.isFreed())
                    _createController.free(this);

                _createController = null;

                LyEditTableSetController newSetController = getSetController(object);

                if (newSetController == null)
                    newSetController = setControllerFactory(object);
            }
        }

        reset(false);

        if (_dynamicEditionMode)
            startCellEdition(null, -1);

        return result;
    } // end method endValidate

    /**
     * Cancel edition of given controller.
     * @param setController the set controller to cancel
     * @param update indicates if we must update the view
     */
    protected void cancel(LyEditTableSetController setController, boolean update)
    {
        setController.cancel(update);

        if (update)
            valueChanged(setController, null);
    }

    /**
     * Sets the displayed fields for this controller.
     *
     * @param fields List of fields to display in the table. If null, return to default list of
     *               fields.
     */
    public void setFields(LyInfoList<LyFieldInfo> fields)
    {
        Hashtable<Object, LyValueSet> oldValues = new Hashtable<Object, LyValueSet>();
        LyValueSet                    valueSet;

        if (_editable && _enabled)
        {

            // Store the list of modified values
            if (_setControllers != null)
            {
                LyEditTableSetController              setController;
                Enumeration<LyEditTableSetController> e = _setControllers.elements();

                while (e.hasMoreElements())
                {
                    setController = e.nextElement();

                    if (setController.wasChanged())
                    {
                        valueSet = setController.getValueSet(getFields());

                        LyValueSet hiddenValueSet = setController.getHiddenValues();

                        if (valueSet != null)
                        {

                            if (hiddenValueSet != null)
                                valueSet.addValues(hiddenValueSet);

                            oldValues.put(setController.getObject(), valueSet);
                        }
                        else if (hiddenValueSet != null)
                            oldValues.put(setController.getObject(), hiddenValueSet);
                    }
                }
            }

            // Store the list of modified values from the create controller
            if ((_createController != null) && _createController.wasChanged())
            {
                valueSet = _createController.getValueSet(getFields());

                if (valueSet != null)
                    oldValues.put("NULL_OBJECT", _createController.getValueSet(getFields()));
            }
        }

        // Free set controllers
        freeControllers();

        // Check set and create action
        checkSetAction(fields);
        checkCreateAction(fields);
        super.setFields(fields);

        // Retrieve modified values and repost them in the new set controllers
        Enumeration<Object> e = oldValues.keys();

        while (e.hasMoreElements())
        {
            Object                   key = e.nextElement();
            LyObject                 object = ((key instanceof LyObject) ? (LyObject) key : null);
            LyValueSet               values = oldValues.get(key);
            LyEditTableSetController setController = getSetController(object);
            int                      i, n = ((values == null) ? 0 : values.getSize());
            LyValue                  value;

            if (setController != null)
            {

                for (i = 0; i < n; i++)
                {
                    value = values.getFieldValue(i);
                    setController.setFieldValue((LyFieldInfo) value.getField(), value, false);
                    setController.valueChanged((LyFieldInfo) value.getField());
                    updateRowObject(object);
                }
            }
        }
    }

    /**
     * Free any setController related to this controller.
     */
    protected void freeControllers()
    {

        if (_setControllers != null)
        {
            LyEditTableSetController              setController;
            Enumeration<LyEditTableSetController> e = _setControllers.elements();

            while (e.hasMoreElements())
            {
                setController = e.nextElement();
                setController.free(this);
            }

            _setControllers = null;
        }

        if (_createController != null)
        {
            _createController.free(this);
            _createController = null;
        }
    }

    /**
     * Free this controller on request of the specified controller.
     *
     * @param controller controller instance which requested the freeing of this controller. If the
     *                   controller is the parent, this controller is freed otherwise the method
     *                   does nothing.
     * @see   #free()
     */
    public void free(LyController controller)
    {
        freeControllers();
        super.free(controller);
    }

    /**
     * Clears all cached data for this controller (especially the view and the component).
     *
     * <P>This method may be invoked for memory optimization, but the controller is still valid and
     * may be used. (Unlike free) Further usage, needs to rebuild internal component & view which
     * may be time consuming.</P>
     */
    public void clearViewCache()
    {
        super.clearViewCache();

        if (_setControllers != null)
        {
            LyEditTableSetController              setController;
            Enumeration<LyEditTableSetController> e = _setControllers.elements();

            while (e.hasMoreElements())
            {
                setController = e.nextElement();
                setController.clearViewCache();
            }
        }

        if (_createController != null)
            _createController.clearViewCache();
    }

    /**
     * Called by set controller when an object value is modified.
     *
     * @param set       set controller that notifies the change
     * @param fieldInfo field changed.
     */
    protected void valueChanged(LyEditTableSetController set, LyFieldInfo fieldInfo)
    {

        // Component may have been freed during cancel process : in this case, do nothing.
        if ((_component == null) || _component.isFreed())
            return;

        if (!_editable || !_enabled)
            return;

        if (!set.isInitialized())
            return;

        // Change status image to validate or cancel
        LyObject object = set.getModel();

        if (_autoCommit && (set != _createController)
                && set.isValid() && _isCurrentRowValid)
            validate(set);

        updateRowObject(object);
    } // end method valueChanged

    /**
     * Updates the row for a given object.
     *
     * @param object the object whose row must be updated.
     */
    public void updateRowObject(LyObject object)
    {
        int index;

        if (_workSpace == null)
            index = -1;
        else if (object == null)
            index = _workSpace.getSize();
        else
            index = _workSpace.getIndex(object);

        LyInfoList<LyFieldInfo> statusFields = getStatusFieldInfos();

        updateRow(object, index, statusFields);
    }

    /**
     * Called when an object is added in the dynamic object list.
     * @param event the workspace event
     * @return true if the event was handled
     *
     * @see LyWorkSpaceListener#objectAdded
     */
    public boolean objectAdded(LyWorkSpaceEvent event)
    {
        LyObject object = event.getObject();

        _selectedCells = null;
        _bufferedValues = null;

        if ((getClass() == LyEditTableController.class) && (_createController != null)
                && (_createController.getObject() == object))
            return false;
        else
            return super.objectAdded(event);
    }

    /**
     * Called when an object is removed from the dynamic object list.
     *
     * @param event the workspace event
     * @return true if the event was handled
     * @see LyWorkSpaceListener#objectRemoved
     */
    public boolean objectRemoved(LyWorkSpaceEvent event)
    {
        boolean removed = super.objectRemoved(event);

        if (removed)
        {
            LyObject        object = event.getObject();
            LySetController setController = getSetController(object);

            if (setController != null)
            {

                if (setController == _createController)
                {

                    if (!_createController.isFreed())
                        _createController.free(this);

                    _createController = null;
                }
                else
                {
                    _setControllers.remove(object);

                    if (!setController.isFreed())
                        setController.free(this);
                }
            }

            _selectedCells = null;
            _bufferedValues = null;
        }

        return removed;
    } // end method objectRemoved

    /**
     * For internal use, maintain a list of setControllers related to objects being modified.
     *
     * @param object  current object being modified
     * @return current setController. If no set controller is currently related, a new one is
     *         created.
     */
    public LyEditTableSetController getSetController(LyObject object)
    {

        if (object == null)
            return _createController;
        else if ((_createController != null) && (_createController.getObject() == object))
            return _createController;
        else if (_setControllers != null)
            return _setControllers.get(object);
        else
            return null;
    }

    /**
     * Indicates wether an error message may be shown or not.
     *
     * @return true if raising popup message is active, false otherwise
     */
    public boolean enableShowMessage()
    {
        return _showMessages;
    }

    /**
     * Indicates wether an error message may be shown or not.
     *
     * @param status true if raising popup message is active, false otherwise
     */
    public void enableShowMessage(boolean status)
    {
        _showMessages = status;
    }

    /**
     * Set edition field marks used to build the controlled view.
     *
     * @param fieldMarks marks to set
     */
    public void setSetFieldMarks(BitSet fieldMarks)
    {
        _setFieldMarks = fieldMarks;
    }

    /**
     * Access to edition field marks used to build the controlled view.
     *
     * @return the field marks
     */
    public BitSet getSetFieldMarks()
    {
        return _setFieldMarks;
    }

    /**
     * Checks if given field info is editable or not.
     * @param object the object to check
     * @param fieldInfo the field to check
     * @return true if the field for the given object is editable
     */
    public boolean isEditable(LyObject object, LyFieldInfo fieldInfo)
    {

        if (!_editable || !_enabled)
            return false;

        BitSet fieldMarks;

        if ((_createController != null) && (_createController.getObject() == object))
        {

            if (_createAction == null)
                return false;

            fieldMarks = _createFieldMarks;
        }
        else
        {

            if ((object == null) && (_createAction == null))
                return false;

            if ((object != null) && (_setAction == null))
                return false;

            fieldMarks = ((object == null) ? _createFieldMarks : _setFieldMarks);
        }

        // Check on the behavior if field is editable for given object
        LyEditTableSetController setController = getSetController(object);

        if (setController == null)
            return false;

        LyClassBehavior classBehavior = (LyClassBehavior) setController.getBehavior();
        boolean         editable = classBehavior.isEditable(setController, fieldInfo);

        return editable && ((fieldMarks == null) || fieldInfo.hasMarks(fieldMarks, false));
    }

    /**
     * Indicates in dynamic edition mode if a cell may be edited.
     *
     * @param  object    the object represented in the cell.
     * @param  fieldInfo the field info represented in the cell.
     * @return true if the cell may be edited, false otherwise.
     */
    protected boolean allowCellEdition(LyObject object, LyFieldInfo fieldInfo)
    {
        LySession session = getSession();
        boolean   editable = isEditable(object, fieldInfo);

        if ((object != null) && (_setAction != null)
                && !session.enableAction(_setAction, this, new LySimpleObjectList(object), false))
            editable = false;

        return editable;
    }

    /**
     * Checks if given field info can be editable or not.
     * @param fieldInfo the field to check
     * @return true if the field can be edited
     */
    public boolean canBeEditable(LyFieldInfo fieldInfo)
    {

        if (!_editable || !_enabled || !_editTableForRelations)
            return false;

        if (fieldInfo instanceof LyArrayFieldInfo)
            return false;

        if (fieldInfo instanceof LyStructFieldInfo)
            return false;

        if (fieldInfo instanceof LyRelationFieldInfo)
        {
            LyRelationFieldInfo.Type type = ((LyRelationFieldInfo) fieldInfo).getType();

            if (type == LyRelationFieldInfo.Type.COMPOSITION)
                return false;
        }

        // Edit multiple choice requires that the field can be set
        // otherwise do not create headers
        if ((fieldInfo instanceof LyChoiceFieldInfo) && (fieldInfo.isMultiple())
                && (_setFieldMarks != null) && !fieldInfo.hasMarks(_setFieldMarks, false))
            return false;

        if ((_setFieldMarks == null) || fieldInfo.hasMarks(_setFieldMarks, false))
            return true;

        if ((_createFieldMarks == null) || fieldInfo.hasMarks(_createFieldMarks, false))
            return true;

        return false;
    } // end method canBeEditable

    /**
     * Checks if the controlled form is editable or not.
     * @return the editable status of this controller
     */
    public boolean isEditable()
    {
        return _editable;
    }

    /**
     * Set boolean indicating if the controlled form is editable or not.
     * @param editable the editable status
     */
    public void setEditable(boolean editable)
    {

        if (editable == _editable)
            return;

        _editable = editable;

        if (_setControllers != null)
        {
            LyEditTableSetController              setController;
            Enumeration<LyEditTableSetController> e = _setControllers.elements();

            while (e.hasMoreElements())
            {
                setController = e.nextElement();
                setController.setEditable(editable);
            }
        }

        if (_createController != null)
            _createController.setEditable(editable);

        // Update commands state
        LyComponent root = getComponent();
        LyCommand   validate = ((root != null) ? (LyCommand) root.getComponent(DEMAND_VALIDATE)
                                               : null);
        LyCommand   apply = ((root != null) ? (LyCommand) root.getComponent(DEMAND_APPLY) : null);
        LyCommand   restore = ((root != null) ? (LyCommand) root.getComponent(DEMAND_RESTORE)
                                              : null);

        if (validate != null)
            validate.setEnabled(editable);

        if (apply != null)
            apply.setEnabled(editable);

        if (restore != null)
            restore.setEnabled(editable);

        if (_initialized)
            reset(false);
    }

    /**
     * Internal create controlle: for internal use only.
     * @return the internal create controller
     */
    public LyEditTableSetController getInternalCreateController()
    {
        return _createController;
    }

    /**
     * Checks if the controlled form is enabled or not.
     * @return the enabled status
     */
    public boolean isEnabled()
    {
        return _enabled;
    }

    /**
     * Checks if the dynamic edition mode is set or not.
     * @return the dynamic edition mode status
     */
    public boolean isDynamicEditionMode()
    {
        return _dynamicEditionMode;
    }

    /**
     * Indicates whether auto-commit is set or not.
     * @return true if auto-commit is set, false otherwise
     */
    public boolean isAutoCommit()
    {
        return _autoCommit;
    }

    /**
     * Set boolean indicating if the controlled form is enabled or not.
     * @param enabled the enabled status
     */
    public void setEnabled(boolean enabled)
    {

        if (enabled == _enabled)
            return;

        _enabled = enabled;

        if (_setControllers != null)
        {
            LyEditTableSetController              setController;
            Enumeration<LyEditTableSetController> e = _setControllers.elements();

            while (e.hasMoreElements())
            {
                setController = e.nextElement();
                setController.setEnabled(enabled);
            }
        }

        if (_createController != null)
            _createController.setEnabled(enabled);

        if (_initialized)
            reset(false);
    }

    /**
     * Get content of the table.
     *
     * @return the list of objects being displayed
     */
    public LyObjectList getContent()
    {

        if ((_createController != null) && _createController.wasChanged())
        {
            LyObject newObject = _createController.getObject();

            if ((newObject != null) && !_workSpace.contains(newObject))
            {
                LySimpleWorkSpace content = new LySimpleWorkSpace(getSession(), _workSpace, false);

                content.addObject(newObject);
                return content;
            }
        }

        return _workSpace;
    }

    /**
     * Checks if an edit table is used for editing composition relations.
     * @return true if we use edit table for relations
     */
    public boolean useEditTable()
    {
        return _editTableForRelations;
    }

    /**
     * Method for counting visible objects.
     * @return the number of visible objects
     */
    protected int getVisibleObjectsCount()
    {
        int n = super.getVisibleObjectsCount();

        if ((n > 0) && (_createController != null))
            return n - 1;
        else
            return n;
    }

    /**
     * Show an error for the specified object. If object is null or if the resource
     * LY_SHOWN_ERRORS_ON_OBJECTS is set to true, the error message is shown in a popup window else
     * the specified error is added to the object.
     *
     * @param object The object in error
     * @param error  The error o show
     * @see   #showError(String)
     */
    public void showError(LyObject object, LyError error)
    {

        if (!enableShowMessage())
            return;

        LyEnvironment environment = getEnvironment();

        _errorTime = new Date().getTime();

        if (error != null)
        {
            String   message = error.getMessage();
            Object[] args = error.getArgs();

            message = environment.getMessage(message, args);

            if (object != null)
            {
                args = new Object[] { getName(object) };
                message = environment.getMessage("LY_ER_ON_OBJECT", args) + '\n' + message;
            }
            else
                message = environment.getMessage("LY_ER_ON_NEW_OBJECT") + '\n' + message;

            showError(message);
        }
        else
            super.showError(object, error);
    } // end method showError

    /**
     * Validate the given controller
     * @param setController the controller to validate
     * @return the status of the validation
     * @see LyFormInterface
     */
    protected short validateRow(LyEditTableSetController setController)
    {
        // Validate changes
        short result = validate(setController);

        if (_dynamicEditionMode && (result == LyFormInterface.STATUS_OK))
            startCellEdition(null, -1);

        return result;
    }

    /**
     * Cancel the given controller
     * @param setController the controller to cancel
     */
    protected void cancelRow(LyEditTableSetController setController)
    {
        _isCurrentRowValid = true;
        // Validate changes
        cancel(setController, true);

        if (_dynamicEditionMode)
            startCellEdition(null, -1);
    }

    /**
     * Delete the given controller
     * @param setController the controller to delete.
     * @return True if successful.
     */
    protected boolean deleteRow(LyEditTableSetController setController)
    {
        // Delete object
        LyObject           object = setController.getObject();
        LyClassInfo        classInfo = object.getClassInfo();
        LyAction           deleteAction = classInfo.getActionWithId(LyAction.ID_DELETE);
        LySimpleObjectList list = new LySimpleObjectList(object);
        LySession          session = getSession();

        if ((deleteAction != null) && (!deleteAction.hasMark(LyAction.MARK_HIDDEN))
                && session.enableAction(deleteAction, this, list, false)
                && (getSession().validateAction(deleteAction, setController)))
        {
            invokeAction(deleteAction, object);
            return true;
        }
        else
            return false;
    }

    /**
     * Change the valid status of the current row
     * @param isValid the valid status
     */
    protected void setIsCurrentRowValid(boolean isValid)
    {
        _isCurrentRowValid = isValid;
    }

    /**
     * Get index of fieldInfo in fields when in list mode.
     * @param colIndex index of column.
     * @return fieldInfo in the column
     */
    public LyFieldInfo getFieldInfoDisplayList(int colIndex)
    {
        LyInfoList<LyFieldInfo> fields = getFields();
        LyClassInfo             classInfo = getClassInfo();

        if ((colIndex >= getColsNumberListDisplay()) || (colIndex <= getDisplayListDataIndex()))
            return null;

        if (isDisplayListShowFields())
            colIndex--;
        //for first data.
        colIndex--;

        LyFieldInfo fInfo;
        boolean     hasImage = isImageColumn(fields);

        for (int i = 0; i < fields.getSize(); i++)
        {
            fInfo = fields.getInfo(i);

            // if "normal field" , i.e, no complex, no footer, no first image
            if (classInfo.contains(fInfo))
            {

                // if field = first imageField, skip ( take count of other image fields )
                if ((fInfo instanceof LyFileFieldInfo)
                        && LyFileFieldInfo.Type.IMAGE.equals(((LyFileFieldInfo) fInfo).getType()))
                {

                    if (hasImage)
                        hasImage = false;
                    else
                        colIndex++;
                }

                if (!(fInfo instanceof LyNumberFieldInfo))
                    colIndex++;
                else if (((LyNumberFieldInfo) fInfo).getFunction()
                        == LyNumberFieldInfo.Function.NONE)
                    colIndex++;

            }
        }

        return fields.getInfo(colIndex);
    }

    /**
     * Get index of fields value column.
     * @return index of fields value column
     */
    public int getDisplayListDataIndex()
    {
        LyInfoList<LyFieldInfo> statusFields = getStatusFieldInfos();
        int 					pos = 1;

        if (isStatusColumn(statusFields))
            pos++;

        if (isDisplayListShowFields())
            pos++;

        return pos;
    }

	/**
	 * Sets display mode.
	 * @param displayList True if list mode, false if table mode.
	 */
	protected void setDisplayList(boolean displayList)
	{
		if (isDynamicEditionMode())
			startCellEdition(null, -1);

		super.setDisplayList(displayList);
	}
} // end class LyEditTableController
