/**
 * LyEditSetTableController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyEditTableSetController.java 34747 2012-03-16 12:26:54Z jlsaouli $
 */

package leon.control;

import leon.app.*;
import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.control.form.*;
import leon.data.*;
import leon.info.*;
import leon.view.*;
import leon.view.event.*;


/**
 * This class implements a set controller used by the LyEditTableController This class is for
 * internal use only.
 *
 * @see LyEditTableController
 */
public class LyEditTableSetController extends LySetController
{
	/** File Version */
	public static final String		  __VERSION =
		"$Id: LyEditTableSetController.java 34747 2012-03-16 12:26:54Z jlsaouli $";

	/** Related table controller. */
	protected LyEditTableController   _tableController;

	/** Boolean indicating if something was changed or not. */
	protected boolean				  _changed;

	/** The list of fields that have been changed in this controller. */
	protected LyInfoList<LyFieldInfo> _changedFields;

	/** Boolean indicating if update is being performed. */
	protected boolean				  _updating;

	/** Boolean indicating if the controlled form is enabled or not. */
	protected boolean				  _enabled;

	/** Boolean indicating if the controlled form is a creation form or a set form. */
	protected boolean				  _creation = false;

	/** Instance of LyCreateController used to create new objects. */
	protected LyCreateController	  _createController;
	
	/** List of values used for fields that are not displayed in the table. */
	protected LyValueSet			  _hiddenValues;


	/**
	 * Constructor. Modification form for a list of objects. ClassInfo parameter is ignored, kept
	 * for compatibility and introspection capabilities.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller).
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo this parameter is kept for introspection and compatibility. not used.
	 * @param objects   list of objects currently being modified.
	 */
	public LyEditTableSetController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		super(session, parent, action, classInfo, objects);

		_context = ((LyEditTableController)_parent).getContext();
		_tableController = (LyEditTableController)parent;
		_component = _tableController.getComponent();
		_changed = false;
		_changedFields = new LyInfoList<LyFieldInfo>();
		_updating = false;
		_enabled = true;
		_creation = (objects == null);
		_createController = null;

		// Register this controller to receive user demands
		LyViewManager viewManager = getViewManager();

		viewManager.addDemandListener(this);
	}

	/**
	 * Clears all cached data for this controller (especially the view and the component).
	 *
	 * <P>This method may be invoked for memory optimization, but the controller is still valid and
	 * may be used. (Unlike free) Further usage, needs to rebuild internal component & view which
	 * may be time consuming.</P>
	 */
	public void clearViewCache()
	{
		_component = null;
		super.clearViewCache();
	}

	public LyObject createObject()
	{
		if (_object != null)
			return _object;

		LyClassInfo classInfo = getClassInfo();

		_object = classInfo.createInstance();

		LyContext context = getControllerContext();

		context.setIgnoreUndo(true);
		context.createObject(_object);
		context.setIgnoreUndo(false);
		return _object;
	}
	
	/**
	 * Overriden to always return parent view controller.
	 * @return the view controller
	 */
	public LyViewController getViewController()
	{
		return ((_parent != null) ? _parent.getViewController() : null);
	}

	/**
	 * Factory for constructing the field controller of a field info.
	 *
	 * @param  fieldInfo Field info for which a controller is built.
	 * @param forceConsult indicates if the consult mode is forced for the field
	 * @return The field controller of the given field.
	 */
	public LyFieldController fieldControllerFactory(LyFieldInfo fieldInfo, boolean forceConsult)
	{
		if (!isDisplayed(fieldInfo))
			return null;

		if (fieldInfo instanceof LyChoiceFieldInfo)
			return new LyEditTableChoiceFieldSetController(this, (LyChoiceFieldInfo)fieldInfo);
		else
			return super.fieldControllerFactory(fieldInfo, forceConsult);
	}

	/**
	 * Invoked when an action occurs.
	 * @param demand the action demand
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if ((demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
				|| (demand.getType() == LyActionDemand.TYPE_UNIT_CHANGED))
		{
			String demandCode = demand.getSource().intern();
			int    pos = demandCode.indexOf(FORM_FIELD_SEP);

			if (pos >= 0)
				demandCode = demandCode.substring(pos + FORM_FIELD_SEP.length());
			pos = demandCode.indexOf(FORM_FIELD_SEP);
			if (pos > 0)
				demandCode = demandCode.substring(0, pos);

			LyClassInfo classInfo = getClassInfo();
			LyFieldInfo fieldInfo = classInfo.getFieldInfo(demandCode);

			if (demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
				valueChanged(fieldInfo);
			else
				unitChanged(fieldInfo, demand.getValue());

			//      Update parent values if needed (forms)
			// FFT01538
			LyParameterValues values = demand.getParameterValues();

			if (values != null)
			{
				LyController parent = getParent().getParent();

				if ((parent != null) && (parent instanceof LyAreaController))
					values.updateViewFromValues(((LyAreaController)parent).getComponent());
			}
			return;
		}

		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			String source = demand.getSource();

	        if (source.equals(LyAreaController.DEMAND_CLOSE))
            {
                cancel(true);
                _tableController.startCellEdition(null, -1);
            }
	        else if (source.equals(LyFormController.DEMAND_VALIDATE))
			    _tableController.validateRow(this);
			else if (source.equals(LyFormController.DEMAND_CANCEL))
			    _tableController.cancelRow(this);
			else if (source.equals(LyEditTableController.DEMAND_ALT_UP))
				_tableController.startCellEdition(false, true, demand.getValue());
			else if (source.equals(LyEditTableController.DEMAND_ALT_DOWN))
				_tableController.startCellEdition(true, true, demand.getValue());
			else if (source.equals(LyEditTableController.DEMAND_ALT_LEFT))
				_tableController.startCellEdition(false, false, demand.getValue());
			else if (source.equals(LyEditTableController.DEMAND_ALT_RIGHT))
				_tableController.startCellEdition(true, false, demand.getValue());
			else if (source.equals(LyEditTableController.DEMAND_ALT_ADD))
				_tableController.copyUpperValue();
		}
		else if (((demand.getType() == LyActionDemand.TYPE_TAB_NEXT)
				|| (demand.getType() == LyActionDemand.TYPE_TAB_PREVIOUS))
			&& (getParent() != null))
		{
			// Forward tab next in fields to parent
			((LyAreaController)getParent()).actionDemand(demand);
		}
		else
		    super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Accessor to the formField corresponding to the given fieldInfo.
	 *
	 * @param  fieldInfo the fieldInfo which is associated to the view component.
	 * @return the formField corresponding to the given fieldInfo
	 */
	public LyFormField getFormField(LyFieldInfo fieldInfo)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController == null)
			return null;

		LyFormField formField = fieldController.getFormField(fieldInfo);

		if (formField != null)
			formField.setNeedCheck(true);
		return formField;
	}

	/**
	 * Show a popup message warning box with the specified message. The message may be a text
	 * message, or the key to a String in the application disctionnary. (stringsXXX.properties
	 * files) This is a shortcut to getViewManager().showWarning()
	 *
	 * @param message the message to show
	 * @see leon.view.LyViewManager
	 * @see leon.misc.LyStrings
	 */
	public void showWarning(String message)
	{
		if (_tableController.enableShowMessage())
		{
			_tableController.showWarning(message);
			_tableController.enableShowMessage(false);
		}
	}

	/**
	 * Show a popup message error box with the specified message. The message may be a text message,
	 * or the key to a String in the application disctionnary. (stringsXXX.properties files) This is
	 * a shortcut to getViewManager().showError()
	 *
	 * @param message the message to show
	 * @see leon.view.LyViewManager
	 * @see leon.misc.LyStrings
	 */
	public void showError(String message)
	{
		if (_tableController.enableShowMessage())
		{
			_tableController.showError(message);
			_tableController.enableShowMessage(false);
		}
	}

	/**
	 * Free this controller.
	 */
	public void free()
	{
		// Don't free parent
		free(_parent);
	}

	/**
	 * Free this controller on request of the specified controller.
	 *
	 * @param controller controller instance which requested the freeing of this controller. If the
	 *                   controller is the parent, this controller is freed otherwise the method
	 *                   does nothing.
	 * @see   #free()
	 */
	public void free(LyController controller)
	{
		_component = null;

		super.free(controller);

		// Unregister this controller to receive user demands
		LyViewManager viewManager = getViewManager();

		viewManager.removeDemandListener(this);

		_tableController = null;
	}

	/**
	 * Start validating changes.
	 * @return the start validate status
	 */
	public short startValidate()
	{
		_tableController.setIsCurrentRowValid(true);
		
        if (_component == null)
            return LyFormInterface.STATUS_IGNORE;
       
		short result = super.startValidate();

		if (result == LyFormInterface.STATUS_IGNORE)
		{
			_changed = false;
			_changedFields.resetInfos();
			_tableController.valueChanged(this, null);
		}
		else if (result == LyFormInterface.STATUS_KO)
		{
			// Consider object as changed
			_changed = true;
			_tableController.setIsCurrentRowValid(false);
			_tableController.valueChanged(this, null);
			showErrors();
		}

		return result;
	}

	/**
	 * Overloaded to redirect the validation demand to the edit table controller.
	 * @param free indicates if this controller must be freed at the end
	 * @return the validation status
	 */
	public short validate(boolean free)
	{
		return _tableController.validate(this);
	}

	/**
	 * Validate changes.
	 * @return the end validate status
	 */
	public short endValidate()
	{	       
		if (!isCreation()
				&& ((_setValues == null) || (_setValues.getSize() == 0)))
		{
			_changed = false;
			_changedFields.resetInfos();
			return LyFormInterface.STATUS_IGNORE;
		}

		short result = super.endValidate();

		if (result != LyFormInterface.STATUS_KO)
		{
			updateObject(null);
			_changed = false;
			_changedFields.resetInfos();
		}

		return result;
	}

	/**
	 * Cancel changes.
	 * @param update indicates if an update must be performed
	 */
	public void cancel(boolean update)
	{
		_changed = false;
		_changedFields.resetInfos();
		cancel();
		if (isFreed())
			return;

		if (_creation)
			_object = null;

		if (update)
		{
			_update = true;
			if (_creation)
				// Force creation of a new object after deletion of the previous one
				getObject();
			updateObject(null);
		}
	}

	/**
	 * Internal method for control and propagate a value change.
	 * @param fieldInfo the fieldinfo
	 */
	public void valueChanged(LyFieldInfo fieldInfo)
	{
		if (_updating)
			// Ignore value changes during updates
			return;

		LyFieldInfoValue value = (LyFieldInfoValue)getFieldValue(fieldInfo);

		if ((_tableController instanceof LyRelationTableController)
				&& (!((LyRelationTableController)_tableController).canChange(fieldInfo)))
		{
			LyObject		 object = getObject();

			if (object == null)
				return;

			LyValue oldValue = object.getFieldValue(fieldInfo);

			updateFieldItem(fieldInfo, oldValue.getValue(), value.getValue(), false);
			return;
		}

		_tableController.enableShowMessage(true);

		super.valueChanged(fieldInfo);

		if (value == null)
			// An error occured when getting the value.
			return;

		short status = LyClassBehavior.STATUS_IGNORE;

		if (isDisplayed(fieldInfo))
		{
			if (isValid(fieldInfo))
			{
				// Control new value
				LySetBehaviorInterface behavior = getSetBehavior();
				LyFieldInfoValue	   newValue;

				newValue = (LyFieldInfoValue)getFieldValue(fieldInfo);
				status = behavior.controlNewValue(this, newValue);
				setState(fieldInfo, status);
			}
	        else
	            status = getState(fieldInfo);
		}

		if (status != LyClassBehavior.STATUS_IGNORE)
		{
			if (_initialized)
			{
				_changed = true;
				_changedFields.addInfo(fieldInfo);
			}

			_tableController.setIsCurrentRowValid(true);
			_tableController.valueChanged(this, fieldInfo);
		}
		else if (_changedFields.contains(fieldInfo))
		{
			_changedFields.removeInfo(fieldInfo);
			if (_changedFields.getSize() == 0)
				_changed = false;

			_tableController.valueChanged(this, fieldInfo);
		}
		else if (_changedFields.contains(fieldInfo))
		{
            _changedFields.removeInfo(fieldInfo);
            if (_changedFields.getSize() == 0)
                _changed = false;
    
            _tableController.valueChanged(this, fieldInfo);
		}

		_tableController.enableShowMessage(false);
	} // end method valueChanged

	/**
	 * Checks if something was changed.
	 * @return true if something was changed
	 */
	public boolean wasChanged()
	{
		return _changed;
	}

	/**
	 * Indicates which fields have been changed in this controller.
	 *
	 * @return the list of fields that have been changed or an empty list if no field has been
	 *         changed.
	 */
	public LyInfoList<LyFieldInfo> getChangedFields()
	{
		return _changedFields;
	}

	/**
	 * Get state of this controller.
	 * @return the state of this controller
	 * @see leon.app.behavior.LyClassBehavior
	 */
	public short getState()
	{
		LyClassInfo classInfo = getClassInfo();
		int		    i, n = classInfo.getFieldInfoCount();
		LyFieldInfo fieldInfo;
		short	    state = LyClassBehavior.STATUS_OK, tempState;

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			if (contains(fieldInfo))
			{
				tempState = getState(fieldInfo);
				if ((tempState != LyClassBehavior.STATUS_OK)
						&& ((state == LyClassBehavior.STATUS_OK)
							|| (tempState < state)))
					state = tempState;
			}
		}

		LyObject object = getObject();

		if (object != null)
		{
			tempState = object.getErrorLevel();
			if ((tempState != LyClassBehavior.STATUS_OK)
					&& ((state == LyClassBehavior.STATUS_OK)
						|| (tempState < state)))
				state = tempState;
		}

		return state;
	} // end method getState

	/**
	 * Checks if this controller is valid : all changes are correct.
	 * @return the valid status
	 */
	public boolean isValid()
	{
		return (getState() != LyClassBehavior.STATUS_KO);
	}

	/**
	 * Returns a boolean indicating if the action is or not a creation.
	 * @return the creation status
	 */
	public boolean isCreation()
	{
		return _creation;
	}

	/**
     * Is this table in a form consult or not.
	 * @return the consult state
     */
    public boolean isConsult()
    {
        LyController parent = _tableController.getParent();

        if (parent instanceof LyFieldController)
            return !(parent instanceof LyFieldSetController);
        else
            return !_tableController.isEditable();
    }
    
	/**
	 * End of the current action.
	 * @param values the values
	 * @return the status of the end operation
	 */
	protected boolean end(LyValueSet values)
	{
		if (!isCreation())
			return super.end(values);

		_tableController.setIdle(true);
		_changed = true;

		// Create dummy createController
		if (_createController == null)
			_createController = new LyCreateController(getSession(), this, getAction(),
					getClassInfo(), getObject(), getControllerContext());

		// End creation
		LyCreateBehaviorInterface behavior = (LyCreateBehaviorInterface)_createController
				.getBehavior();
		boolean					  created = behavior.endCreation(_createController, values);
		LyObject				  object = _object;

		if (object == null)
			object = _createController.getObject();

		if (created)
		{
			_createController.commit();

			// Check controller context before committing
			if ((_controllerContext != null)
					&& ((_controllerContext.getStatus() == LyContext.Status.ERROR)
						|| (_controllerContext.getStatus() == LyContext.Status.ROLLBACKED)
						|| (_controllerContext.getErrorLevel() == LyObject.LEVEL_ERROR)))
			{
				// Commit failed, there are problems
				showErrors();
				_object = object;
				_tableController.setIdle(false);
				return false;
			}

			_controllerContext = null;
			_creation = false;
			if (_formResultClient != null)
				_formResultClient.notifyFormResult(object, null);
		}
		else
		{
			_createController.cancel();
			_controllerContext = null;
			object = null;
			_tableController.setIsCurrentRowValid(false);
		}

		LyController parent = _createController.getParent();

		_createController.free(parent);
		_createController = null;
		_object = object;
		_tableController.setIdle(false);
		return created;
	} // end method end

	/**
	 * Set boolean indicating if the controlled form is enabled or not.
	 * @param enabled the enabled status
	 */
	public void setEnabled(boolean enabled)
	{
		_enabled = enabled;
		super.setEnabled(enabled);
	}

	/**
	 * Changes current field status to (non-)editable. If the field is not editable, the value is
	 * read-only but however submited with all other values. A disabled field is not submited with
	 * all other values and is just ignored.
	 *
	 * @param fieldInfo the field which should be marked editable
	 * @param editable new fieldInfo status. (default is true or obtained from the view
	 *                  description)
	 */
	public void setEditable(LyFieldInfo fieldInfo, boolean editable)
	{
		super.setEditable(fieldInfo, editable);
		
		_tableController.updateCellEditionStatus(_object, fieldInfo, editable);
	}

	/**
	 * Get boolean indicating if the controlled form is enabled or not.
	 * @return the enabled status
	 */
	public boolean isEnabled()
	{
		return _enabled;
	}
	   
	/**
	 * Checks if the specified fieldInfo is currently being shown in the form.
	 *
	 * @param  fieldInfo the fieldInfo we want to find
	 * @return true if the fieldInfo is present (modification/consultation), false otherwise.
	 */
	public boolean isDisplayed(LyFieldInfo fieldInfo)
	{
		LyInfoList<LyFieldInfo> fields = _tableController.getFields();

		if (fields != null)
			return fields.contains(fieldInfo);
		else
			return super.isDisplayed(fieldInfo);
	}
	
	/**
	 * Overridden to register values for fields that are not displayed in the edit table.
	 * @param fieldInfo the fieldinfo
	 * @param newValue the new value
	 * @param propagate indicates if we must propagate the changes
	 */
	public void setFieldValue(LyFieldInfo fieldInfo, LyValue newValue, boolean propagate)
	{
	    if (!_updating)
	        initArea();
		
		// If field is not displayed, register the new value in the list of hidden values.
		if (!isDisplayed(fieldInfo))
		{
			if (_hiddenValues == null)
				_hiddenValues = new LyValueSet();
			
			if (newValue != null)
				_hiddenValues.addValue(newValue);
			else
				_hiddenValues.addValue(fieldInfo.buildFieldValue(null));
			
			_changed = true;
			_changedFields.addInfo(fieldInfo);
		}
		else
			super.setFieldValue(fieldInfo, newValue, propagate);
	}
	
	/**
	 * Overridden to retrieve values for fields that are not displayed in the edit table.
	 * @param fieldInfo the fieldInfo
	 * @return the LyValue
	 */
	public LyValue getFieldValue(LyFieldInfo fieldInfo)
	{
		// If field value is registered in the list of hidden values, return it.
		if (!isDisplayed(fieldInfo)
				&& (_hiddenValues != null)
				&& _hiddenValues.contains(fieldInfo))
			return _hiddenValues.getFieldValue(fieldInfo);
		     
		if ((_object != null)
		        && (!isInitialized()
					|| !isDisplayed(fieldInfo)
					|| !super.isDisplayed(fieldInfo)))
			// If field is neither displayed in the controller nor in the table, return value from object.
			return _object.getFieldValue(fieldInfo);
				
		return super.getFieldValue(fieldInfo);
	}

	/**
	 * Overridden to add values from fields that are not displayed in the edit table.
	 * @param values the values set to complete
	 * @param fieldInfo the fieldInfo
	 * @return the status of the operation
	 */
	protected short addValue(LyValueSet values, LyFieldInfo fieldInfo)
	{
		if (!isDisplayed(fieldInfo)
			&& (_hiddenValues != null)
			&& _hiddenValues.contains(fieldInfo))
		{
			values.addValue(_hiddenValues.getFieldValue(fieldInfo));
			return LySetBehaviorInterface.STATUS_OK;
		}

		return super.addValue(values, fieldInfo);
	}
	
	/**
	 * Returns the list of values used for fields that are not displayed in the table.
	 * 
	 * @return the value of the _hiddenValues attribute.
	 */
	public LyValueSet getHiddenValues()
	{
		return _hiddenValues;
	}
	
	/**
	 * Set current controller context.
	 *
	 * @param context current context or null if none is specified
	 * @return true if successfull
	 * @see   LyContextInterface#getContext
	 */
	public boolean setContext(LyWorkSpace context)
	{
		if (isCreation())
			return changeContext(context);
		else
			return super.setContext(context);
	}

	/**
	 * Get the model object
	 * @return the model object
	 */
	public LyObject getModel()
	{
		return (_creation ? null : super.getModel());
	}

	/**
	 * Gets the controller context.
	 *
	 * @return The controller context.
	 */
	public LyContext getControllerContext()
	{
		if (isConsult())
			return null;
		else if ((_controllerContext == null)
				|| (_controllerContext.isEnded()))
		{
			LyContext context = super.getControllerContext();

			if ((_relationParent != null)
					&& (context != null))
				context.setInternal();
			return context;
		}
		else
			return _controllerContext;
	}

	/**
	 * Called when current object has changed . Keep the overall content consistent.
	 *
	 * @param oldValues set of former values, or null if not specified.
	 */
	public void updateObject(LyValueSet oldValues)
	{
	    if (!_update)
	        return;
	    
	    _updating = true;
	    
		super.updateObject(oldValues);

		if (!isCreation())
		{
		    _updating = false;
			return;
		}

		LyClassInfo classInfo = getClassInfo();
		LyValueSet  values = null;
		LyFieldInfo fieldInfo;
		LyValue     fieldValue;
		int		    i,
					n = ((oldValues != null) ? oldValues.getSize() : classInfo.getFieldInfoCount());
		Object	    newValue, oldValue;

		for (i = 0; i < n; i++)
		{
			if (oldValues != null)
			{
				fieldValue = oldValues.getFieldValue(i);
				fieldInfo = (LyFieldInfo)fieldValue.getField();
			}
			else
				fieldInfo = classInfo.getFieldInfo(i);

			if ((fieldInfo instanceof LyRelationFieldInfo)
					&& (((LyRelationFieldInfo)fieldInfo).getType() == LyRelationFieldInfo.Type.COMPOSITION))
				continue;

			if (contains(fieldInfo))
			{
				newValue = getValue(fieldInfo);
				if (_object == null)
					oldValue = getDefaultValue(fieldInfo);
				else
					oldValue = _object.getValue(fieldInfo);
				if (!fieldInfo.equals(newValue, oldValue))
				{
					if (values == null)
						values = new LyValueSet();
					values.addValue(fieldInfo, newValue);
				}
			}
		}

		if ((values != null) && (values.getSize() > 0))
		{
			if (_object != null)
			{
				LyContext context = getControllerContext();

				if (context != null)
				{
					context.setIgnoreUndo(true);
					context.setValues(_object, values);
					context.setIgnoreUndo(false);
				}
			}
		}
		
		_updating = false;
	} // end method updateObject

	/**
	 * Update the value of a fieldInfo shown in the view. The method sets the value of the
	 * associated fieldItem.
	 *
	 * @param fieldInfo the fieldInfo which should be updated in the form
	 * @param newValue  the new value of the field.
	 * @param oldValue  former value of the field,or null if unknown.
	 * @param propagate Indicates whether the change is notified (propagated) to other fields or
	 *                  not.
	 * @see   #getFieldItem
	 */
	protected void updateFieldItem(LyFieldInfo fieldInfo, Object newValue,
		Object oldValue, boolean propagate)
	{
		super.updateFieldItem(fieldInfo, newValue, oldValue, propagate);

		_tableController.updateCellValue(_object, fieldInfo, newValue);
	}

	/**
	 * Update the titlebar content with default values.<br />
	 * Overridden to avoid updating edit table titlebar.
	 *
	 * @see leon.control.LyAreaController#updateTitleBar
	 */
	protected void updateTitleBar()
	{
	} // end method updateTitleBar

	/**
     * Update the toolbar content with default values. <br />
     * Overridden to avoid updating edit table toolbar.
     *
     * @see leon.control.LyAreaController#updateTitleBar
     */
    protected void updateTools()
    {
    } // end method updateTools
    
    /**
     * Update context.
     * Overridden to avoid updating context menu.
     */
    protected void updateContext()
    { 
    } // end method updateContext

	/**
	 * Returns the objects of the form for a given context.
	 *
	 * @param  context The context.
	 * @return The objects of the form for the given context.
	 */
	protected LyObjectList getObjectsFromContext(LyWorkSpace context)
	{
		if ((context == null) || (context.getSize() == 0))
			return null;

		if (_compositionRelation != null)
		{
			// We are in case of a composition relation.
			// Only one object can be edited in a LyeditTableSetController.
			// So if _object is not null we just return it.
			if (_object != null)
				return new LySimpleObjectList(_object);
		}

		return super.getObjectsFromContext(context);
	}

	/**
	 * Creates the field controller for the given fieldInfo.
	 * @param fieldInfo the fieldInfo
	 * @return the created controller
	 */
	protected LyFieldController createFieldController(LyFieldInfo fieldInfo)
	{
		if ((_parent == null)
				|| (!((LyEditTableController)_parent).canBeEditable(fieldInfo)
					&& !isConsult(fieldInfo)))
			return null;

		return super.createFieldController(fieldInfo);
	}

	/**
	 * Show an error for the specified object. If object is null or if the resource
	 * LY_SHOWN_ERRORS_ON_OBJECTS is set to true, the error message is shown in a popup window else
	 * the specified error is added to the object.
	 *
	 * @param object The object in error
	 * @param error  The error o show
	 * @see  #showError(String)
	 */
	public void showError(LyObject object, LyError error)
	{
		if (_tableController.enableShowMessage())
		{
			_tableController.showError(object, error);
			_tableController.enableShowMessage(false);
		}
	}

	/**
	 * Check if the form content has been modified and if the object need to be created.
	 * @return the changed status
	 */
	public boolean checkIfModified()
	{
		return _changed;
	}

    /**
     * Returns a boolean indicating if update is being performed.
     * @return the updating status
     */
    public boolean isUpdating()
    {
    	return _updating;
    }
    
    /**
     * Sets the <code>_updating</code> attribute.
     * @param updating the updating value
     */
    public void setUpdating(boolean updating)
    {
    	_updating = updating;
    }
    
	/**
	 * @see leon.control.LyFormController#objectRemoved(leon.data.LyWorkSpaceEvent)
	 */
	public boolean objectRemoved(LyWorkSpaceEvent event)
	{
		if (event.getObject() != _object)
			return false;

		free();
		return true;
	}
} // end class LyEditTableSetController
