/**
 * LyExportController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyExportController.java 41546 2012-11-16 13:14:56Z bangiole $
 */

package leon.control;

import java.io.*;
import java.util.*;
import leon.app.*;
import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.data.cache.LyFullCache;
import leon.info.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.file.*;
import leon.peer.xml.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;


/**
 * This class implements the simple export edition area. The export area is used to create a export
 * of the data into a CSV like text file. This is an ASCII file with separators, one object per line
 * close to the save format of Microsoft Excel. (Excel uses commas or semicolons), as the controller
 * let the user choose these parameters.
 *
 * <P>Default separators are : FIELD_SEP (for columns) and VALUE_SEP (multiple values) Data may be
 * saved to an ascii file, then retreived with an import controller or an application such as MS
 * Excel</P>
 *
 * @see LyImportController
 */
public class LyExportController extends LyAreaController
	implements LyQuestionAnswerInterface
{
	/** File Version */
	public static final String __VERSION =
		"$Id: LyExportController.java 41546 2012-11-16 13:14:56Z bangiole $";

	/** Application end. */
	public static final String DEMAND_OVERWRITE_FILE = "_overwriteFile".intern();

	/** User demand identifier : show preview of saved data. */
	public static final String DEMAND_PREVIEW = "_preview".intern();

	/** Id of the preview area. */
	public static final String PREVIEW_AREA = "_exportPreview".intern();

	/** Preview area width. */
	public static final int    PREVIEW_WIDTH = 165;

	/** Preview area height. */
	public static final int    PREVIEW_HEIGHT = 30;

	/** Form field identifier : List of fields that should be saved. */
	public static final String SELECT_FIELDINFOS = "_fieldInfo".intern();

	/**
	 * Form field identifier : export type, either SELECT_EXPORT_TYPE_SPREADSHHET or
	 * SELECT_EXPORT_TYPE_TEXT or SELECT_EXPORT_TYPE_RAW or SELECT_EXPORT_TYPE_XML.
	 */
	public static final String SELECT_EXPORT_TYPE = "_exportType".intern();

	/** Form field value identifier : export text formated value. */
	public static final String SELECT_EXPORT_TYPE_TEXT = "_exportTypeText".intern();

	/** Form field value identifier : export text formated value. */
	public static final String SELECT_EXPORT_TYPE_SPREADSHEET = "_exportTypeSpreadSheet".intern();

	/** Form field value identifier : export text formated value. */
	public static final String SELECT_EXPORT_TYPE_RAW = "_exportTypeRaw".intern();

	/** Form field value identifier : export text formated value. */
	public static final String SELECT_EXPORT_TYPE_XML = "_exportTypeXml".intern();

	/** Form field identifier : select field separator (default is FIELD_SEP). */
	public static final String SELECT_FIELD_SEP = "_exportFieldSep".intern();

	/** Form field identifier : select value separator (default is VALUE_SEP). */
	public static final String SELECT_VALUE_SEP = "_exportValueSep".intern();

	/** Form field identifier */
    public static final String SELECT_FIELD_PREFIX = "_exportFieldPrefix".intern();

    /** Form field identifier */
    public static final String SELECT_FIELD_RENAME_EXTENSION = "_exportFieldRenameExtension".intern();

	/** Form field identifier : Checkbox only selected values (default is unchecked). */
	public static final String EXPORT_SELECTION = "_exportSelection".intern();

	/** Form field identifier : select output file path. */
	public static final String SELECT_FILE = "_exportFile".intern();

    /** Form field identifier : select output file path. */
    public static final String CHARSET = "_exportCharset".intern();

	/** Default field separator in saves. */
	public static String	   FIELD_SEP = "|";

	/** Default multiple value separator in saves. */
	public static String	   VALUE_SEP = ",";

	/** Default field separator in imports. */
	public static String	   SPREADSHEET_FIELD_SEP = ";";

	/** Default multiple value separator in imports. */
	public static String	   SPREADSHEET_VALUE_SEP = "~";

	/** Reference to the Class info. */
	protected LyClassInfo	   _classInfo;

	/** File where data must be written. */
	protected File			   _file;

	/** Data that must be written. */
	protected Vector<String>   _lines;

	/** Silent mode. */
	protected boolean		   _silent;

	/** Export type (for non GUI access). */
	protected String		   _exportType;

	/** Export charset. */
    protected String           _exportCharset;

	/** _fields (for non GUI access). */
	protected String[]		   _fields;

	/** Indicates if field separator has been changed by the user. */
	protected boolean		   _fieldSepChanged;

	/** Indicates if value separator has been changed by the user. */
	protected boolean		   _valueSepChanged;

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller).
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo This parameter is used when the objects list contains objects of different
	 *                  classes to known which class is exported.
	 * @param objects   list of objects that should be saved.
	 */
	public LyExportController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		super(session, parent, action, objects);

		_file = null;
		_lines = null;
		_exportType = null;
		_fields = null;
		_silent = (parent == null);

		if (classInfo == null)
		{
			LyInfoList<LyClassInfo> classes = ((objects == null) ? null : objects.getClasses());

			if ((classes != null) && (classes.getSize() == 1))
				_classInfo = classes.getInfo(0);
		}
		else
			_classInfo = classInfo;

		endInit();
	}

	/**
	 * Method to check if parameters are valid for export.
	 */
	protected void endInit()
	{
		if (_classInfo == null)
			throw new IllegalStateException(
				"BF AppEngine : class to export must be specified if objects list is null or contains more than one class");

		if (!isFormDisplayed())
			validateExport(false);
	}

	/**
	 * Overridden to ensure that export form is displayed.
	 */
	public void showView()
	{
		if (isFormDisplayed())
			super.showView();
		else
			return;
	}

	/**
	 * Indicates if the export form is displayed.<br />
	 * Export form should not be displayed if all the fields values are given as parameters of the
	 * export action.
	 *
	 * @return true if form is displayed, false otherwise.
	 */
	protected boolean isFormDisplayed()
	{
		if ((getParameterValue(SELECT_FIELDINFOS) != null)
				&& (getParameterValue(SELECT_FIELD_SEP) != null)
				&& (getParameterValue(SELECT_VALUE_SEP) != null)
				&& (getParameterValue(SELECT_EXPORT_TYPE) != null)
				&& (getParameterValue(SELECT_FILE) != null))
			return false;

        LyExportBehavior behavior = (LyExportBehavior)getBehavior();

        return behavior.isFormDisplayed(this);
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyExportBehaviorInterface.class;
	}

	/**
	 * Invoked when an action notification is sent to this listener.
	 *
	 * <P>This method is called by demandPerformed.</P>
	 *
	 * <P>At this level, the implementation handles demands of type LyActionDemand.TYPE_ACTION :
	 * LyFormController.DEMAND_VALIDATE, LyExportController.DEMAND_PREVIEW, and
	 * LyExportController.DEMAND_CANCEL</P>
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		String id = demand.getSource();

		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			if (getViewManager().hasProperty("RE_ORDER_EXPORT"))
				_fields = demand.getParameterValues().getParameterValues("_exportFieldInfo");

			if (id.equals(LyFormController.DEMAND_VALIDATE))
			{
				validateExport(false);
				_fields = null; //reinit value of _fields
				return;
			}
			else if (id.equals(LyExportController.DEMAND_PREVIEW))
			{
				validateExport(true);
				_fields = null; //reinit value of _fields
				return;
			}
			else if (id.equals(LyFormController.DEMAND_CANCEL))
			{
				hideView();
				free();
				return;
			}
		}
		else if (id.equals(LyExportController.SELECT_EXPORT_TYPE))
		{
			LyFormField fieldSep = getFormField(LyExportController.SELECT_FIELD_SEP);
			LyFormField valueSep = getFormField(LyExportController.SELECT_VALUE_SEP);

			if (demand.getValue().equals(LyExportController.SELECT_EXPORT_TYPE_XML))
			{
				if (fieldSep != null)
					fieldSep.setEnabled(false);

				if (valueSep != null)
					valueSep.setEnabled(false);
			}
			else
			{
				if (fieldSep != null)
					fieldSep.setEnabled(true);

				if (valueSep != null)
					valueSep.setEnabled(true);

				LyText fieldSepItem = ((fieldSep == null) ? null : (LyText)fieldSep.getFieldItem());
				LyText valueSepItem = ((valueSep == null) ? null : (LyText)valueSep.getFieldItem());

				if (demand.getValue().equals(LyExportController.SELECT_EXPORT_TYPE_SPREADSHEET)
						|| demand.getValue().equals(LyExportController.SELECT_EXPORT_TYPE_TEXT))
				{
					if (!_fieldSepChanged && (fieldSepItem != null))
						fieldSepItem.setLabel(SPREADSHEET_FIELD_SEP);

					if (!_valueSepChanged && (valueSepItem != null))
						valueSepItem.setLabel(SPREADSHEET_VALUE_SEP);
				}
				else
				{
					if (!_fieldSepChanged && (fieldSepItem != null))
						fieldSepItem.setLabel(FIELD_SEP);

					if (!_valueSepChanged && (valueSepItem != null))
						valueSepItem.setLabel(VALUE_SEP);
				}
			}
		}
		else if (id.equals(LyExportController.SELECT_FIELD_SEP))
			_fieldSepChanged = true;
		else if (id.equals(LyExportController.SELECT_VALUE_SEP))
			_valueSepChanged = true;
		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Validate the export. Internal use. This method creates the export content to a file, and
	 * eventually show the result in the preview.
	 *
	 * @param isPreview indicates if this is a final save to the file, or showing result in the
	 *                  preview.
	 */
	protected void validateExport(boolean isPreview)
	{
		// Reset export type to be able to change it
		_exportType = null;

		String		  fieldSep = (String)getParameterValue(SELECT_FIELD_SEP);
		String		  valueSep = (String)getParameterValue(SELECT_VALUE_SEP);
		String		  file = (String)getParameterValue(SELECT_FILE);
		String		  fileName = null;
		String		  error = null;
		String		  exportType = getExportType();
		LyFormField   field;
		LyFieldItem   fieldItem;
		LyViewManager viewManager = getViewManager();
		LyFormField   valuePrefix = getFormField(LyExportController.SELECT_FIELD_PREFIX);
        LyFormField   valueRenameExtension = getFormField(LyExportController.SELECT_FIELD_RENAME_EXTENSION);

		// Check if all the fields are fullfilled
		if (fieldSep == null)
		{
			field = getFormField(LyExportController.SELECT_FIELD_SEP);
			if (field != null)
			{
				fieldItem = field.getFieldItem();
				fieldSep = ((LyGetLabelInterface)fieldItem).getLabel();
			}
		}

		if (valueSep == null)
		{
			field = getFormField(LyExportController.SELECT_VALUE_SEP);
			if (field != null)
			{
				fieldItem = field.getFieldItem();
				valueSep = ((LyGetLabelInterface)fieldItem).getLabel();
			}
		}

		// Default values. Doesn't matter for the xml export.
		if (exportType == SELECT_EXPORT_TYPE_XML)
		{
			fieldSep = "|";
			valueSep = ",";
		}

		// If fields are not available in the dialog,
		// check if behavior is defining them
		LyExportBehavior behavior = (LyExportBehavior)getBehavior();

		if ((fieldSep == null) || (fieldSep.length() == 0))
			fieldSep = behavior.getDefaultAttributeSep();
		if ((valueSep == null) || (valueSep.length() == 0))
			valueSep = behavior.getDefaultAttributeValSep();

		if ((fieldSep == null) || (fieldSep.length() == 0))
			error = "LY_FIELD_SEP";

		if ((error == null) && ((valueSep == null) || (valueSep.length() == 0)))
			error = "LY_VALUE_SEP";

		if (!isPreview)
		{
			if ((viewManager == null)
					|| viewManager.hasProperty("LOCAL_FILE"))
			{
				if (file == null)
				{
					field = getFormField(LyExportController.SELECT_FILE);
                    if (field != null)
                    {
                        fieldItem = field.getFieldItem();
                        file = ((LyGetLabelInterface)fieldItem).getLabel();
                    }
				}

				if ((file != null) && (file.length() > 0) && (file.indexOf('.') < 0))
				{
					String fileType = "csv";

					if (exportType == SELECT_EXPORT_TYPE_TEXT)
						fileType = "txt";
					else if (exportType == SELECT_EXPORT_TYPE_RAW)
						fileType = "res";
					else if (exportType == SELECT_EXPORT_TYPE_XML)
						fileType = "xml";

					file += '.' + fileType;
				}
			}
			else
            {
            	String prefix = null;
        		String extension = null;

        		if ( valuePrefix != null)
        		{
        			fieldItem = valuePrefix.getFieldItem();
        			prefix = ((LyGetLabelInterface)fieldItem).getLabel();
        		}
        		else if (getParameterValue(SELECT_FIELD_PREFIX) != null)
        		{
        			prefix = (String)getParameterValue(SELECT_FIELD_PREFIX);
        		}

        		if ( valueRenameExtension != null)
        		{
        			fieldItem = valueRenameExtension.getFieldItem();
        			extension = ((LyGetLabelInterface)fieldItem).getLabel();
        		}
        		else if (getParameterValue(SELECT_FIELD_RENAME_EXTENSION) != null)
        		{
        			extension = (String)getParameterValue(SELECT_FIELD_RENAME_EXTENSION);
        		}

                // Get temporary file for web application
                fileName = getTemporaryFileName(exportType, prefix, extension);

                file = viewManager.getTemporaryDir() + "/" + fileName;
            }

			if ((error == null) && ((file == null) || (file.length() == 0)))
				error = "LY_FILE";
		}

		if (error != null)
		{
			LyEnvironment environment = getEnvironment();
			String		  message = environment.getMessage("LY_ER_NOTNULL", error);

			if (!_silent)
				showError(message);
			else
				environment.logError(message);
			return;
		}

		// FFT 00007 / SITA
		// Retrieve the selected fields
		@SuppressWarnings("unchecked")
		LyInfoList<LyFieldInfo> fieldsInfos = (LyInfoList<LyFieldInfo>)getParameterValue(SELECT_FIELDINFOS);

		Vector<LyFieldInfo>     fields;

		if (fieldsInfos == null)
		{
			if (_fields == null)
				fields = getSelectedFields();
			else
				fields = getFields();
		}
		else
		{
			int fieldsNb = fieldsInfos.getSize();

			fields = new Vector<LyFieldInfo>();

			for (int i = 0; i < fieldsNb; i++)
				fields.add(fieldsInfos.getInfo(i));
		}

		// If export action
		if (!isPreview)
		{
			export(file, fields, fieldSep, valueSep, exportType);
			behavior.showExportedFileWindow(this, fileName);
		}
		// Else if preview action
		else
			preview(fields, fieldSep, valueSep);
	} // end method validateExport

	/**
	 * Returns the value of a given parameter set on current export action.<br />
	 * The type of the returned value depends on the parameter.
	 *
	 * @param  parameterName the name of the parameter which value must be retrieved.
	 * @return the value of a given parameter set on current export action or null if the parameter
	 *         is not set on the export action.
	 */
	public Object getParameterValue(String parameterName)
	{
		Object parameterValue = getParameter(parameterName, true);

		if (parameterName == SELECT_FIELDINFOS)
		{
			LyFieldInfo			    field = (LyFieldInfo)parameterValue;
			LyInfoList<LyFieldInfo> result = null;
			int					    np = ((field == null) ? 1 : 2);

			do
			{
				if (field != null)
				{
					if (result == null)
						result = new LyInfoList<LyFieldInfo>();
					result.addInfo(field);
				}
				field = (LyFieldInfo)_action.getParameter(SELECT_FIELDINFOS + np, true);
				np++;
			}
			while (field != null);

			return result;
		}

        if (parameterValue == null)
            return null;
        else if (parameterName == SELECT_FIELD_SEP)
			return parameterValue;
		else if (parameterName == SELECT_VALUE_SEP)
			return parameterValue;
		else if (parameterName == SELECT_EXPORT_TYPE)
		{
			if ("SPREADSHEET".equals(parameterValue))
				return SELECT_EXPORT_TYPE_SPREADSHEET;
			else if ("TEXT".equals(parameterValue))
				return SELECT_EXPORT_TYPE_TEXT;
			else if ("RAW".equals(parameterValue))
				return SELECT_EXPORT_TYPE_RAW;
			else if ("XML".equals(parameterValue))
				return SELECT_EXPORT_TYPE_XML;
		}
		else if (parameterName == SELECT_FILE)
			return getEnvironment().getFile((String)parameterValue);
        else if(parameterName == SELECT_FIELD_PREFIX)
        	return parameterValue;
        else if(parameterName == SELECT_FIELD_RENAME_EXTENSION)
        	return parameterValue;
        else if ((parameterName == EXPORT_SELECTION)
                && (parameterValue instanceof String))
            return Boolean.valueOf(((String)parameterValue).equalsIgnoreCase("true"));

        return null;
	} // end method getParameterValue

	/**
     * Access to a temporary file path to store content.
	 * @param exportType an export type
	 * @return a temporary file name
     */
    public String getTemporaryFileName(String exportType)
    {
    	return getTemporaryFileName(exportType, null, null);
    }

    /**
     * Access to a temporary file path to store content.
     * @param exportType the export type
     * @param prefix a prefix
     * @param extension an extension
     * @return a temporary file name
     */
    public String getTemporaryFileName(String exportType, String prefix, String extension)
    {
        String fileType = "csv";

        if (exportType == SELECT_EXPORT_TYPE_TEXT)
            fileType = "txt";
        else if (exportType == SELECT_EXPORT_TYPE_RAW)
            fileType = "res";
        else if (exportType == SELECT_EXPORT_TYPE_XML)
            fileType = "xml";

        if (extension != null)
        	fileType = extension;

        // ask a temporary file to viewManager
        if(prefix == null)
        	return "_export" + System.currentTimeMillis() + "." + fileType;
        else
        	return prefix + "_export" + System.currentTimeMillis() + "." + fileType;
    }

	/**
	 * Extract the fieldInfo list selected by the user Internal use.
	 *
	 * @return the list of selected fields tha should be saved
	 */
	protected Vector<LyFieldInfo> getSelectedFields()
	{
		Vector<LyFieldInfo> fields = new Vector<LyFieldInfo>();
		LyFormField		    formField = getFormField(LyExportController.SELECT_FIELDINFOS);
        LyClassInfo         classInfo = getClassInfo();
        LyFieldInfo         fieldInfo;

		if (formField != null)
		{
    		LyRelation		    relation = (LyRelation)formField.getFieldItem();
    		int				    i, n = relation.getLinkCount();
    		LyLink			    link;
    		String			    value;

    		for (i = 0; i < n; i++)
    		{
    			link = relation.getLink(i);
    			if (link.isSelected())
    			{
    				value = link.getId();
    				fieldInfo = classInfo.getFieldInfo(value);
    				fields.addElement(fieldInfo);
    			}
    		}
		}

		if (fields.size() == 0)
		{
			// Return all valid fields
			BitSet		  fieldMarks = getParameterMarks("_fieldMarks", LyBitSet.Type.F_MARKS);
			LyEnvironment environment = getEnvironment();
			int           i, n = classInfo.getFieldInfoCount();

			for (i = 0; i < n; i++)
			{
				fieldInfo = classInfo.getFieldInfo(i);
				if ((fieldMarks != null) && !fieldInfo.hasMarks(fieldMarks, false))
					continue;
				if ((fieldMarks == null) && fieldInfo.hasMark(LyFieldInfo.MARK_SECRET))
				{
					if ((fieldInfo instanceof LyTextFieldInfo)
							&& (((LyTextFieldInfo)fieldInfo).getEncoding() == null))
					{
						environment.logWarning("BF AppEngine : can't export text field "
							+ fieldInfo + " which is secret and has no encoding");
						continue;
					}
				}

				fields.addElement(fieldInfo);
			}
		}

		return fields;
	} // end method getSelectedFields

	/**
	 * Extract the fieldInfo list of tabs _fields Internal use.
	 *
	 * @return the list of selected fields that should be saved
	 */
	protected Vector<LyFieldInfo> getFields()
	{
		Vector<LyFieldInfo> fields = new Vector<LyFieldInfo>();
		LyClassInfo		    classInfo = getClassInfo();

		if (_fields == null)
			return getSelectedFields();

		int		    i, n = _fields.length;
		LyFieldInfo fieldInfo;

		for (i = 0; i < n; i++)
		{
			String name = _fields[i];

			fieldInfo = classInfo.getFieldInfo(name);
			fields.addElement(fieldInfo);
		}

		return fields;
	}

	/**
	 * Export the objects as a single string.
	 *
	 * @param fields     list of fieldInfos that should be saveed to the file.
	 * @param fieldSep   separator between fields (columns)
	 * @param valueSep   separator between multiple values
	 * @param exportType type of data to export : Possible values are SELECT_EXPORT_TYPE_TEXT
	 *                   (.txt), SELECT_EXPORT_TYPE_RAW (.res), SELECT_EXPORT_TYPE_SPREADSHEET
	 *                   (.csv) or SELECT_EXPORT_TYPE_XML (.xml).
	 * @return a string corresponding to the export
	 */
	public String exportAsString(Vector<LyFieldInfo> fields, String fieldSep,
		String valueSep, String exportType)
	{
		_lines = getExportLines(fields, fieldSep, valueSep, exportType);

		StringBuffer sb = new StringBuffer();

		if (_lines != null)
		{
			for (int i = 0; i < _lines.size(); i++)
				sb.append(_lines.elementAt(i)).append('\n');
		}

		return sb.toString();
	}

	/**
	 * Save the objects to the file. If the file directory does not exist, this method attemps to
	 * create it.
	 *
	 * @param filePath   path of the save file
	 * @param fields     list of fieldInfos that should be saveed to the file.
	 * @param fieldSep   separator between fields (columns)
	 * @param valueSep   separator between multiple values
	 * @param exportType type of data to export : Possible values are SELECT_EXPORT_TYPE_TEXT
	 *                   (.txt), SELECT_EXPORT_TYPE_RAW (.res), SELECT_EXPORT_TYPE_SPREADSHEET
	 *                   (.csv) or SELECT_EXPORT_TYPE_XML (.xml).
	 */
	public void export(String filePath, Vector<LyFieldInfo> fields, String fieldSep,
		String valueSep, String exportType)
	{
		_exportType = exportType;
		_lines = getExportLines(fields, fieldSep, valueSep, exportType);

		// Get the path of file and directory
		_file = new File(filePath);

		String dirPath = _file.getParent();
		File   directory = null;

		if (dirPath != null)
			directory = new File(dirPath);

		LyEnvironment environment = getEnvironment();
		LyViewManager viewManager = getViewManager();

		// Try to create the directory if needed
		if ((directory != null) && (!directory.exists()))
		{
			try
			{
				environment.logInfo("Creating directory :" + dirPath);
				directory.mkdirs();
			}
			catch (SecurityException e)
			{
				environment.logError("Unable to create directory :" + dirPath);
				if (!_silent)
					showError(environment.getMessage("LY_ER_CREATE_DIR", dirPath));
				return;
			}
		}

		// If file already exists, ask if it must be overwritten
		if (_file.exists())
		{
			if (!_silent)
			{
				viewManager.askQuestion(getView(),
					environment.getMessage("LY_CONFIRM_OVERWRITE", _file), this,
					DEMAND_OVERWRITE_FILE);
				// Process goes on in questionAnswer
				return;
			}

			environment.logWarning("Overwriting existing file :" + _file);
		}

		saveFile();

		// If form is not displayed and file does not exist, notify file creation to the user
		if ((viewManager != null)
				&& viewManager.hasProperty("LOCAL_FILE")
				&& !isFormDisplayed())
			viewManager.showWarning(getView(),
				environment.getMessage("LY_EXPORT_FILE_CREATED", _file));

		free();
	} // end method export

	/**
	 * Save the export content to the file Internal use for the export method.
	 *
	 * @see #export
	 */
	protected void saveFile()
	{
		LyExportBehaviorInterface behavior = (LyExportBehaviorInterface)getBehavior();

		behavior.saveFile(this, _file, _lines);

		LyViewManager manager = getViewManager();

		if (manager != null)
			hideView();
	}

	/**
	 * Answer to a askQuestion request.
	 *
	 * @param result true if the user has clicked on Yes or Ok
	 * @param clientData data specified in the askQuestion method
	 * @see   LyQuestionAnswerInterface
	 */
	public void questionAnswer(boolean result, Object clientData)
	{
		if ((clientData == DEMAND_OVERWRITE_FILE) && result)
		{
			saveFile();
			free();
		}
	}

	/**
	 * Preview the exported data.
	 *
	 * @param fields   list of fieldInfos that should be exported to the file.
	 * @param fieldSep separator between fields (columns)
	 * @param valueSep separator between multiple values
	 */
	public void preview(Vector<LyFieldInfo> fields, String fieldSep, String valueSep)
	{
		Vector<String> lines = getExportLines(fields, fieldSep, valueSep, getExportType());
		LyViewManager  manager = getViewManager();

		if (!manager.hasProperty("MULTI_WINDOW"))
			hideView();

		LyApplication  application = getApplication();
		LyAction       action = application.getActionWithId(LyAction.ID_EXPORT_PREVIEW, _classInfo);

		_LyExportPreviewController preview = new _LyExportPreviewController(getSession(), this,
				action, _context, lines);

		preview.showView();
	}

	/**
	 * Shows the export form. Called by the preview controller when closing.
	 */
	public void showExportForm()
	{
		showView();
	}

    /**
     * Get content of the chart.
     *
     * @return the list of objects being displayed
     */
    public LyObjectList getContent()
    {
    	if (_context != null)
    		return _context;

    	LyAreaController	parentAreaController = getParentAreaController();

    	if (parentAreaController != null)
    	{
    		LyObjectList	content = parentAreaController.getContent();

    		if (content != null)
    			return content;
    	}

		LyInfoList<LyFieldInfo> fields = new LyInfoList<LyFieldInfo>(getFields());
		LyObjectSource			source = _classInfo.getObjectSource();
		LyCache					cache = source.getCache();
		LyObjectList			ws;

		if (cache.isLoaded()
				|| (cache instanceof LyFullCache))
	    	ws = new LyClassWorkSpace(_session, _classInfo, fields, null, null, false, true);
		else
			ws = new LyPageWorkSpace(_session, _classInfo, fields, null, null, 0, false, true);

    	return ws;
    }

	/**
	 * Return the lines that will be written in the export file.
	 *
	 * @param  fields     list of fieldInfos that should be exported to the file.
	 * @param  fieldSep   separator between fields (columns)
	 * @param  valueSep   separator between multiple values
	 * @return the lines that will be written in the export file
	 * @param  exportType type of data to export : Possible values are SELECT_EXPORT_TYPE_TEXT
	 *                    (.txt), SELECT_EXPORT_TYPE_RAW (.res) or SELECT_EXPORT_TYPE_SPREADSHEET
	 *                    (.csv).
	 */
	protected Vector<String> getExportLines(Vector<LyFieldInfo> fields, String fieldSep,
		String valueSep, String exportType)
	{
		boolean 			exportAll = true;
		LyComponent         cmpnt = getComponent().getComponent(EXPORT_SELECTION);

		if (cmpnt != null)
		{
			LyBoolean   bool = (LyBoolean)((LyFormField)cmpnt).getFieldItem();

			if ((bool  == null)
			        || (bool.getStatus() == LyBoolean.Status.UNCHECKED))
				exportAll = false;
			else
				exportAll = true;
		}
        else
        {
            Boolean       exportSelection = (Boolean)getParameterValue(LyExportController.EXPORT_SELECTION);

            exportAll = (exportSelection == null) || !exportSelection.booleanValue();
        }

		LyController 	parentController = getParent().getParent();
		LyObjectList    objects = null;

		if (!exportAll
				 && (parentController instanceof LyAreaController))
			objects =((LyAreaController)parentController).getSelection();

		if ((objects == null)
				|| (objects.getSize() == 0))
			objects = getContent();

		Iterator<LyObject>		  iterator = (objects == null ? null : objects.iterator());
        LyExportBehaviorInterface behavior = (LyExportBehaviorInterface)getBehavior();
        Vector<String>            lines = behavior.getExportLines(this, iterator,
            fields, fieldSep, valueSep, exportType);

        if (lines != null)
            return lines;

        LyDataSource         dataSource = _classInfo.getObjectSource();
        LyDataProvider       dataProvider = dataSource.getDataProvider();
        LyFileRequestEncoder requestEncoder = requestEncoderFactory(dataSource, dataProvider,
                fieldSep, valueSep, exportType);

		requestEncoder.setEnvironment(getEnvironment()); // Set current environment for export

        String		headerMark = behavior.getHeaderMark(this);

        requestEncoder.setHeaderMark(headerMark);

        int							limit = LyPageWorkSpace.ITERATOR_LIMIT;
        Vector<LyObject>			objectsVector = null;
        LyObject					object;
        String 						header;
        int							encodingType;
        LyXmlExportRequestEncoder	xmlEncoder = null;
        Vector<String>            	tmpLines;

		if (exportType == SELECT_EXPORT_TYPE_XML)
			xmlEncoder = (LyXmlExportRequestEncoder)requestEncoder;

        while (iterator.hasNext())
        {
        	object = iterator.next();

        	if (objectsVector == null)
        		objectsVector = new Vector<LyObject>(limit);
        	objectsVector.add(object);

        	if ((objectsVector.size() % limit == 0)
        		|| !iterator.hasNext())
        	{
		        // Read the file and create the objects
				if (xmlEncoder != null)
				{
					tmpLines = xmlEncoder.encodeValues(objectsVector, fields);

					if (lines == null)
					{
						header = xmlEncoder.writeHeader();
						tmpLines.add(0, header);
					}
					if (!iterator.hasNext())
						tmpLines.add(xmlEncoder.writeFooter());
				}
				else
		        {
		            if (exportType == SELECT_EXPORT_TYPE_TEXT)
		                encodingType = LyFileRequestEncoder.TEXT_ENCODING;
		            else if (exportType == SELECT_EXPORT_TYPE_RAW)
		                encodingType = LyFileRequestEncoder.EXPORT_ENCODING;
		            else if (exportType == SELECT_EXPORT_TYPE_SPREADSHEET)
		                encodingType = LyFileRequestEncoder.SPREADSHEET_ENCODING;
		            else
		                encodingType = LyFileRequestEncoder.XML_EXPORT_ENCODING;

		            tmpLines = requestEncoder.encodeValues(new Vector<Object>(objectsVector),
					    (fields == null ? null : new Vector<LyField>(fields)), encodingType);
		            if (lines != null)
		            	tmpLines.remove(0); // Remove header
		        }

	            if (lines == null)
	            	lines = tmpLines;
	            else
	            	lines.addAll(tmpLines);

				objectsVector = null;
        	}
        }

		int			   i, l = (lines == null ? 0 : lines.size());
		Vector<String> result = new Vector<String>(l);

		for (i = 0; i < l; i++)
			result.addElement(lines.elementAt(i));

		return result;
	} // end method getExportLines

	/**
	 * Factory for the export request encoder.
	 *
	 * @param  dataSource   collection of data from the classInfo to write
	 * @param  dataProvider connector
	 * @param  fieldSep     separator between fields
	 * @param  valueSep     separator between values (multiple)
	 * @param  exportType   one of the SELECT_xxx types
	 * @return the export request encoder
	 */
	protected LyFileRequestEncoder requestEncoderFactory(LyDataSource dataSource,
		LyDataProvider dataProvider, String fieldSep, String valueSep, String exportType)
	{
		if (exportType == SELECT_EXPORT_TYPE_XML)
			return new LyXmlExportRequestEncoder(dataProvider, dataSource, fieldSep, valueSep);
		else
			return new LyFileRequestEncoder(dataProvider, dataSource, fieldSep, valueSep);
	}

	/**
	 * Retrieve formField corresponding to id. Utility method : internal use.
	 *
	 * @param  id identifier of the fieldInfo
	 * @return the field corresponding to id
	 */
	protected LyFormField getFormField(String id)
	{
       if (_builder == null)
           return null;

		// Retrieve formField corresponding to id
		LyComponent root = _builder.getComponent();

		return (root != null ? (LyFormField)root.getComponent(id) : null);
	}

	/**
	 * Accessor to the classInfo associated to this form.
	 *
	 * @return the reference to the classInfo, never returns null. (mandatory in creation)
	 */
	public LyClassInfo getClassInfo()
	{
		return _classInfo;
	}

	/**
	 * Retrieves the view Builder.
	 *
	 * @return the builder instance;
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyExportBuilder(this);
	}

	/**
	 * Indicates if the view for this import controller is resizable.
	 *
	 * @return true if the view is resizable, false otherwise.
	 */
	public boolean isViewResizable()
	{
		return false;
	}

	/**
	 * Accessor to the export type selected in the GUI.
	 *
	 * @return the type of export from the GUI Possible values are SELECT_EXPORT_TYPE_TEXT (.txt),
	 *         SELECT_EXPORT_TYPE_RAW (.res) or SELECT_EXPORT_TYPE_SPREADSHEET (.csv).
	 */
	public String getExportType()
	{
		if (_exportType != null)
			return _exportType;

		String exportType = (String)getParameterValue(SELECT_EXPORT_TYPE);

		if (exportType != null)
			return exportType;

		LyExportBehavior behavior = (LyExportBehavior)getBehavior();

		exportType = behavior.getDefaultExportType();

		LyFormField field = getFormField(LyExportController.SELECT_EXPORT_TYPE);

		if (field != null)
		{
			LyChoice choice = (LyChoice)field.getFieldItem();
			LyOption option;
			int		 i, n = choice.getOptionCount();

			// Retrieve the id of the checked option
			for (i = 0; i < n; i++)
			{
				option = choice.getOption(i);

				if (option.isChecked())
					exportType = option.getId().intern();
			}
		}

		return exportType;
	} // end method getExportType

    /**
     * Accessor to the export charset.
     * @return the charset specified in the action. Default is null.
     */
    public String getExportCharset()
    {
    	if (_exportCharset != null)
    		return _exportCharset;

    	_exportCharset = (String)getParameter(CHARSET);
    	return _exportCharset;
    }
	/* Internal Classes */

	/**
	 * Internal class to show the preview.
	 */
	protected class _LyExportPreviewController extends LyAreaController
	{
		/** The lines to preview */
		protected Vector<String> _lines;

		/**
		 * Default constructor.
		 * @param session the current session
		 * @param parent the parent controller
		 * @param action the invoked action
		 * @param objects the objects to display
		 * @param lines the lines to display
		 */
		public _LyExportPreviewController(LySession session, LyController parent, LyAction action,
			LyObjectList objects, Vector<String> lines)
		{
			super(session, parent, action, objects);

			_lines = lines;
		}

		/**
		 * Method that gives the interface that must be implemented by the behavior associated to
		 * this controller.
		 *
		 * @return Class representing the behavior interface of this controller.
		 */
		public Class<?> getBehaviorInterface()
		{
			return LyExportBehaviorInterface.class;
		}

		/**
		 * Accessor to the list of lines (one line per object exported).
		 *
		 * @return the list of lines
		 */
		public Vector<String> getLines()
		{
			return _lines;
		}

		/**
		 * Retrieves the view Builder.
		 *
		 * @return the builder instance;
		 */
		protected LyBuilder getDefaultBuilder()
		{
			return new _LyExportPreviewBuilder(this);
		}

		/**
		 * Invoked when an action occurs. The implementation manages<BR>
		 * CLOSE demands. (demand.getType = LyActionDemand.TYPE_ACTION AND demand.getSource() =
		 * LyController.DEMAND_CLOSE)
		 *
		 * @param demand demand parameters sent by the view component.
		 * @see   LyController#demandPerformed
		 */
		public void actionDemand(LyActionDemand demand)
		{
			LyViewManager manager = this.getViewManager();

			if (!manager.hasProperty("MULTI_WINDOW"))
			{
				// Only close request for the moment...
				if (demand.getType() == LyActionDemand.TYPE_ACTION)
				{
					this.hideView();
					showExportForm();
				}
			}
			else
				super.actionDemand(demand);
		}

		/**
		 * Indicates if the view for this import controller is resizable.
		 *
		 * @return true if the view is resizable, false otherwise.
		 */
		public boolean isViewResizable()
		{
			return false;
		}
	} // end class _LyExportPreviewController


	/**
	 * Internal builder for the preview.
	 */
	protected class _LyExportPreviewBuilder extends LyAreaBuilder
	{

		/**
		 * Default constructor.
		 * @param parent the parent controller
		 */
		public _LyExportPreviewBuilder(_LyExportPreviewController parent)
		{
			super(parent);
		}

		/**
		 * Builds the view component for the specified parent controller. The component may create
		 * the XML Tree or the internal view.
		 *
		 * @return the view Component
		 */
		public LyComponent build()
		{
			/* The export preview is an area which contains :
			        - A titleBar (description)
			        - A form component
			*/

			LyArea area = new LyArea();

			area.setScrolling(LyArea.Scrolling.AUTO);
			area.setTitleBar(buildTitleBar());
			area.addPanel(buildPreviewForm());
			area.setCommands(buildCommands());
			area.setController(this._parent.getId());

			this._component = area;

			return (area);
		}

		/**
		 * Construct the title bar associated with the export form.
		 *
		 * @return the title bar component
		 * @see    leon.view.LyTitleBar
		 */
		protected LyTitleBar buildTitleBar()
		{
			LyTitleBar titleBar = new LyTitleBar();
			String     label = "LY_ACTION_EXPORT" + " : " + _classInfo.getName();

			if (_parent.getEnvironment().testResource("LY_TITLEBAR_LEFT_IMAGE"))
				titleBar.setImage(this._parent.getImage(_classInfo));
			titleBar.setLeftLabel(label);

			return titleBar;
		}

		/**
		 * Creates the preview form.
		 *
		 * @return the form component
		 * @see    leon.view.LyForm
		 */
		protected LyForm buildPreviewForm()
		{
			LyForm		 form = new LyForm(this._parent.getId());
			LyFormField  field;
			LyText		 fieldText = new LyText();
			StringBuffer buffer = new StringBuffer();
			Vector<String>	lines = ((_LyExportPreviewController)this._parent).getLines();
			int			 i, n = lines.size();

			// Write all the lines in the buffer
			for (i = 0; i < n; i++)
				buffer.append(lines.elementAt(i).toString() + '\n');

			fieldText.setNCols(PREVIEW_WIDTH);
			fieldText.setNRows(PREVIEW_HEIGHT);
			fieldText.setLabel(buffer.toString());
			fieldText.setWrap(false);

			field = new LyFormField(PREVIEW_AREA);
			field.setEditable(false);
			field.setFieldItem(fieldText);
			form.addFormField(field);

			LyConstraint constraint = new LyConstraint();

			constraint.setFill(LyConstraint.Fill.BOTH);

			field.setConstraint(constraint);

			return (form);
		} // end method buildPreviewForm

		/**
		 * Add a commands bar with a close button.
		 *
		 * @return the commands component
		 * @see    leon.view.LyCommands
		 */
		protected LyCommands buildCommands()
		{
			LyCommands commands = new LyCommands();
			LyCommand  command;

			command = new LyCommand(LyFormController.DEMAND_CLOSE, "LY_CLOSE");
			command.setAction(LyCommand.Action.CLOSE);
			commands.addCommand(command);

			return commands;
		}
	} // end class _LyExportPreviewBuilder

	/**
	 * Accessor to the file.
	 *
	 * @return current file to save
	 */
	public File getFile()
	{
		return _file;
	}

	/**
	 * Accessor to the file.
	 *
	 * @param f file to save
	 */
	public void setFile(File f)
	{
		_file = f;
	}

	/**
	 * Checks if this export is in silent mode.
	 * @return the silent mode status
	 */
	public boolean isSilent()
	{
		return _silent;
	}
} // end class LyExportController
