/*
 * LyFastFilterController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.control;

import java.util.*;
import leon.app.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.beaninterface.*;


/**
 * This is the controller for the fast filter action.
 *
 * @version $Id: LyFastFilterController.java 35839 2012-06-12 13:43:15Z nhamon $
 * @author  Lyria
 */
public class LyFastFilterController extends LyUserInputController
{
	/** File Version */
	public static final String		   __VERSION =
		"$Id: LyFastFilterController.java 35839 2012-06-12 13:43:15Z nhamon $";

	/** Search field identifier. */
    public static final String         DEMAND_CANCEL_FILTER = "_cancelFilter".intern();
    
	/** Search field identifier. */
	public static final String		   SEARCH_FIELD = "_searchField".intern();

	/** Current edited filter value. */
	protected String           		   _defaultFilterValue;

	/** Current field info being filtered. */
	protected LyFieldInfo			   _fieldInfo;

	/** Indicates if created filter is used to extend another filter. */
	protected boolean				   _extend;

	/** Current field item being displayed. */
	protected LyFieldItem			   _fieldItem;

	/**
	 * Client which may be notified when the filter forms is closed to update its content (usually a
	 * LyTableController).
	 */
	protected LyFastFilterEndInterface _filterEndInterface;

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo This parameter is not used. Kept from introspection compatibility.
	 * @param context   current context (may be null if the view is not contextual to a list of
	 *                  objects).
	 */
	public LyFastFilterController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent, action, classInfo, context);

		_defaultFilterValue = null;
		_fieldInfo = null;
		_fieldItem = null;
		_filterEndInterface = null;

		if (parent instanceof LyFastFilterEndInterface)
			setFilterEndInterface((LyFastFilterEndInterface)parent);

		// Check if field info may be filtered
		LyFieldInfo			    fieldInfo = _fieldInfo;
		BitSet				    fieldMarks = getParameterMarks("_fieldMarks",
				LyBitSet.Type.F_MARKS);
		LyInfoList<LyFieldInfo> fields = classInfo.getFieldInfos(fieldMarks);

		if ((fields == null)
				|| (fields.getSize() == 0)
				|| fields.contains(fieldInfo))
		{
			// Build a field item depending on the type of the field
			if (_fieldInfo instanceof LyArrayFieldInfo)
				fieldInfo = ((LyArrayFieldInfo)_fieldInfo).getFieldInfo();

			_fieldItem = buildFieldItem(fieldInfo);
		}

		// If no field item was built, action is not allowed on given field info :
		// Free this controller and cancel action.
		if (_fieldItem == null)
			free(_parent);
	}

	/**
	 * Complete content of the component of the controller.
	 */
	protected void completeComponent()
	{
        LyArea             area = (LyArea)getComponent();
        LyArea             subArea = (LyArea)area.getPanel(0);
        LyCommands         commands = area.getCommands();
        LyCommand          command;
        
        if ((_filterEndInterface != null)
        		&& _filterEndInterface.isFastFilterSet()
        		&& (_filterEndInterface.getFilter() != null))
        {
            command = new LyCommand(DEMAND_CANCEL_FILTER, "LY_CANCEL_FILTER");
            command.setAction(LyCommand.Action.CANCEL);
            commands.addCommand(command);
        }
        
		// No need to change component aspect in popup views
		LyViewManager viewManager = getViewManager();

		if (viewManager.hasProperty("POPUP_VIEW"))
			return;

		// Create a new top area with a nice border and remove scrolling
		LyArea			   topArea = new LyArea();
		LyUserInputBuilder builder = (LyUserInputBuilder)getBuilder();
        
        commands.setBorder(LyArea.Border.NONE);
        
		builder.setFormBorder(false);
		
		area.setBorder(LyArea.Border.NONE);
		area.setScrolling(LyArea.Scrolling.NO);

		topArea.setBorder(LyArea.Border.IN);
		topArea.addPanel(area);
		
		subArea.setBorder(LyArea.Border.NONE);
		subArea.setScrolling(LyArea.Scrolling.NO);
		
		_component = topArea;
	}

	/**
	 * Builds a field item according to the field info type.
	 *
	 * @param  fieldInfo the field info used to build the field item
	 * @return the field item built for given field info
	 */
	protected LyFieldItem buildFieldItem(LyFieldInfo fieldInfo)
	{
		// Build a field item according to given field info
		LyFieldItem fieldItem = null;

		if ((fieldInfo instanceof LyTextFieldInfo)
				|| (fieldInfo instanceof LyRelationFieldInfo)
				|| (fieldInfo instanceof LyFileFieldInfo))
		{
			if (fieldInfo instanceof LyRelationFieldInfo)
			{
				// For a relation field info, check if filtering is allowed (related classes must have at
				// least one text field with the NAME mark
				LyRelationFieldInfo     relation = (LyRelationFieldInfo)fieldInfo;
				LyInfoList<LyFieldInfo> fields = getFilteredFieldsForRelation(relation);

                if ((fields == null)
                        || (fields.getSize() != 1))
                    return null;
                else
                {
                    LyFieldInfo fieldInfoTmp = fields.getInfo(0);
                
                    if (fieldInfoTmp instanceof LyNumberFieldInfo)
                        return buildFieldItem(fieldInfoTmp);
                }
			}

			// Build a text field item
			fieldItem = addText(SEARCH_FIELD, "LY_FAST_FILTER", _defaultFilterValue, 20, null, false);
		}
		else if (fieldInfo instanceof LyNumberFieldInfo)
		{
			// Build a number field item
			LyNumberFieldInfo numberField = (LyNumberFieldInfo)fieldInfo;
			LyNumber		  number = addNumber(SEARCH_FIELD, "LY_FAST_FILTER",
			        _defaultFilterValue, LyNumberFieldInfo.Type.INT, false);
			String			  format = numberField.getFormat();
			Number			  vmin = numberField.getMin();
			Number			  vmax = numberField.getMax();
			Number			  pitch = numberField.getPitch();
			int				  decimals = numberField.getDecimals();

			if (format != null)
				number.setFormat(format);
			if (vmin != null)
				number.setVMin(vmin.doubleValue());
			if (vmax != null)
				number.setVMax(vmax.doubleValue());
			if (pitch != null)
				number.setPitch(pitch.doubleValue());

			number.setDecimals(decimals);
			fieldItem = number;
		}
		else if (fieldInfo instanceof LyDateFieldInfo)
		{
			// Build a date field item
			LyDate date = addDate(SEARCH_FIELD, "LY_FAST_FILTER", _defaultFilterValue, null, false);

			if (((LyDateFieldInfo)fieldInfo).getType() == LyDateFieldInfo.Type.TIME)
				date.setType(LyDate.Type.TIME);
			else if (((LyDateFieldInfo)fieldInfo).getType() == LyDateFieldInfo.Type.HOUR_MIN)
				date.setType(LyDate.Type.HOUR_MIN);
			else
				date.setType(LyDate.Type.DATE);

			fieldItem = date;
		}
		else if (fieldInfo instanceof LyChoiceFieldInfo)
		{
			// Build a relation field item with all the options of the choice
			LyRelation				   relation = addRelation(SEARCH_FIELD, "LY_FAST_FILTER", false,
					false);
			LyLink					   link;
			LyChoiceFieldInfo		   choiceFieldInfo = (LyChoiceFieldInfo)fieldInfo;
			LyInfoList<LyChoiceOption> options = choiceFieldInfo.getChoiceOptions(true);
			int						   i, n = ((options == null) ? 0 : options.getSize());
			LyChoiceOption			   option;
			String					   empty_value = "LY_CHOICE_EMPTY_VALUE";
			LyEnvironment			   environment = getEnvironment();
			String					   val = environment.translate(empty_value);

			// Add empty link
			if ((val != null) && val.equals("_null"))
				empty_value = null;

			link = new LyLink("LY_NO_CHOICE", empty_value);
			relation.addLink(link);

			// Add list of links
			for (i = 0; i < n; i++)
			{
				option = options.getInfo(i);
				link = new LyLink(option.getId(), option.getName());

				if ((_defaultFilterValue != null) && _defaultFilterValue.equals(option.getId()))
					link.setSelected(true);

				relation.addLink(link);
			}

			fieldItem = relation;
		}

	    if (fieldItem != null)
        {
            // Check if field info has a bean for filtering
            LyFieldBean bean = fieldInfo.getFieldBean(LyAction.ID_NEW_FILTER);

            if (bean != null)
            {
                fieldItem.setBean(bean.getJavaName());
                fieldItem.setDatas(bean.getDatas());
            }
        }
	      
		return fieldItem;
	} // end method buildFieldItem

	/**
	 * Build a filter for a given field info and a given string value. Depending on the type of the
	 * field info, filter may be built in different ways.
	 *
	 * @param  fieldInfo the field info used to build the filter
	 * @param  value     the value used to build the filter
	 * @return the filter built with both field info and value
	 */
	protected LyFilter buildFilter(LyFieldInfo fieldInfo, String value)
	{
        // if the field has a bean , send the value to the component and get it back
        if ((_fieldItem != null)
            && (value != null)
            && (fieldInfo.getFieldBean(LyAction.ID_NEW_FILTER) != null))
        {  
            LyFieldBeanInterface    component = _fieldItem.getBeanComponent();
            
            if (component instanceof LyGetLabelInterface)
            {
                component.setLabel(value);
                value = ((LyGetLabelInterface)component).getLabel();
            }
        }
        
		// Blank character is used as a separator for multiple filter values
		String[]		   values = tokenizeValue(value);
		LyFilter		   filter = new LyFilter(LyFilter.Type.OR);
		LyFilterExpression exp;

		if ((value == null) || (value.length() == 0) || (values == null))
		{
			// If no value was typed in, build a filter with null value
			exp = new LyFilterExpression(fieldInfo, LyFilterExpression.Condition.EQUALS, null);
			filter.addElement(exp);
		}
		else if ((fieldInfo instanceof LyTextFieldInfo)
				|| (fieldInfo instanceof LyFileFieldInfo)
				|| (fieldInfo instanceof LyArrayFieldInfo))
		{
			// For text and file field infos, check if value contains typed string
			int i, n = values.length;

			for (i = 0; i < n; i++)
			{
				exp = new LyFilterExpression(fieldInfo, LyFilterExpression.Condition.CONTAINS,
						values[i]);
				filter.addElement(exp);
			}
		}
		else if (fieldInfo instanceof LyNumberFieldInfo)
		{
			// For a number field info, check if value equals typed value
		    Number    numberValue = (Number)fieldInfo.parse(value);
		    
			exp = new LyFilterExpression(fieldInfo, LyFilterExpression.Condition.EQUALS, numberValue);
			filter.addElement(exp);
		}
		else if (fieldInfo instanceof LyChoiceFieldInfo)
		{
			// For a choice field info, check if value equals typed value for simple choice
			// and if value contains typed value for multiple choice
			if ("LY_NO_CHOICE".equals(value))
			{
				exp = new LyFilterExpression(fieldInfo, LyFilterExpression.Condition.EQUALS, null);
				filter.addElement(exp);
			}
			else
			{
				if (((LyChoiceFieldInfo)fieldInfo).isMultiple())
					exp = new LyFilterExpression(fieldInfo, LyFilterExpression.Condition.CONTAINS,
							value);
				else
					exp = new LyFilterExpression(fieldInfo, LyFilterExpression.Condition.EQUALS, value);
			}

			filter.addElement(exp);
		}
		else if (fieldInfo instanceof LyDateFieldInfo)
		{
			// For a date field, build max date as the day after and use a lower than condition with this date.
			LyDateFieldInfo dateFieldInfo = (LyDateFieldInfo)fieldInfo;
			LyFilter	    dateFilter = new LyFilter(LyFilter.Type.AND);
			Date		    minDate = (Date)dateFieldInfo.parse(value);
			int			    nbSeconds = (24 * 3600);

			// For a date field info, check the day only
			if ((dateFieldInfo.getType() == LyDateFieldInfo.Type.TIME)
					|| (dateFieldInfo.getType() == LyDateFieldInfo.Type.HOUR_MIN))
			{
				// For a time field, build max date with the following rules :
				//  - if seconds are set (e.g. 10:05:02), max date is one second after min date
				//  - if minutes are set but not seconds (e.g. 10:14:00), max date is one minute after min date
				//  - if only hours are set (e.g. 10:00:00), max date is one hour after min date
				Calendar calendar = Calendar.getInstance();

				calendar.setTime(minDate);

				if (calendar.get(Calendar.SECOND) != 0)
					nbSeconds = 1;
				else if (calendar.get(Calendar.MINUTE) != 0)
					nbSeconds = 60;
				else
					nbSeconds = 3600;
			}

			long maxTime = (minDate.getTime() + (nbSeconds * 1000));
			Date maxDate = new Date(maxTime);

			exp = new LyFilterExpression(fieldInfo, LyFilterExpression.Condition.GREATER_OR_EQUALS,
					minDate);
			dateFilter.addElement(exp);
			exp = new LyFilterExpression(fieldInfo, LyFilterExpression.Condition.LOWER_THAN, maxDate);
			dateFilter.addElement(exp);
			filter.addElement(dateFilter);
		}
		else if (fieldInfo instanceof LyRelationFieldInfo)
		{
			// For a relation, retrieve fields that must be filtered then build a relation filter
			// for each of them with given values.
			LyRelationFieldInfo     relation = (LyRelationFieldInfo)fieldInfo;
			LyInfoList<LyFieldInfo> fields = getFilteredFieldsForRelation(relation);
			LyFieldInfo			    fieldInfoTmp;
			int					    i, n = fields.getSize();
			int					    j, m = values.length;
			LyFilterRelation	    filterRelation;
			LyFilter			    relatedFilter;
			LyFilterExpression.Condition condition;
            Number                  numberValue;

			for (i = 0; i < n; i++)
			{
				filterRelation = new LyFilterRelation(relation);
				relatedFilter = new LyFilter(LyFilter.Type.OR);
				filterRelation.setFilter(relatedFilter);
				fieldInfoTmp = fields.getInfo(i);
				condition = LyFilterExpression.Condition.CONTAINS;

				for (j = 0; j < m; j++)
				{
					if (fieldInfoTmp instanceof LyNumberFieldInfo)
                    {
                        condition = LyFilterExpression.Condition.EQUALS;
                        numberValue = (Number)fieldInfoTmp.parse(values[j]);
                        exp = new LyFilterExpression(fieldInfoTmp, condition, numberValue);
                    }
                    else                    					
                        exp = new LyFilterExpression(fieldInfoTmp, condition, values[j]);
					
					relatedFilter.addElement(exp);
				}

				filter.addElement(filterRelation);
			}
		}

		return filter;
	} // end method buildFilter

	/**
	 * From a value, compute a list of sub values:
	 * String between double quotes are keeped as is.
	 * The rest is tokenized using space.
	 *
	 * @param value the value to tokenized
	 * @return the tokenized value
	 */
	protected String[] tokenizeValue(String value)
	{
		ArrayList<String> result = new ArrayList<String>();
		if ((value == null) || (value.trim().length() == 0))
			return null;
		
		StringTokenizer st = new StringTokenizer(value, "\"", true);
		boolean first = false;
		String val;
		StringTokenizer st2;
		while (st.hasMoreElements())
		{
			val = st.nextToken();
			val = val.trim();
			if (!first && val.equals("\""))
			{
				first = true;
				continue;
			}
			if (first && val.equals("\""))
			{
				first = false;
				continue;
			}
			if (first)
			{
				result.add(val);
			}
			else
			{
				st2 = new StringTokenizer(val, " ");
				while (st2.hasMoreElements())
				{
					result.add(st2.nextToken());
				}
			}
		}
		return result.toArray(new String[result.size()]);
	}
	
	/**
	 * Return the list of fields used to filter a relation field info. This method retrieves all the
	 * text or number fields that have the mark NAME in target classes.
	 *
	 * @param  relation the relation field info used to create the filter
	 * @return the list of field infos used to filter given relation field info
	 */
	protected LyInfoList<LyFieldInfo> getFilteredFieldsForRelation(LyRelationFieldInfo relation)
	{
		LyInfoList<LyFieldInfo> fields = new LyInfoList<LyFieldInfo>(), fieldsTmp;
		LyClassInfo			    classInfo;
		LyFieldInfo			    fieldInfo;
		int					    i, n = relation.getClassInfoCount(), j, m;

		// Parse the class list and retrieve fields that have the NAME mark
		for (i = 0; i < n; i++)
		{
			classInfo = relation.getClassInfo(i);
			fieldsTmp = classInfo.getNameFieldInfos();
			m = ((fieldsTmp == null) ? 0 : fieldsTmp.getSize());

			for (j = 0; j < m; j++)
			{
				// Add field to the list only if field is a text field info
				fieldInfo = fieldsTmp.getInfo(j);
				if (fieldInfo instanceof LyTextFieldInfo
				        || fieldInfo instanceof LyNumberFieldInfo)
					fields.addInfo(fieldInfo);
			}
		}

		return fields;
	}

	/**
	 * Accessor to the view controller of the area. The view controller is a Frame / dialog which
	 * contains this area.
	 *
	 * <P>If no container is set, a new Frame controller is created for this area. Method never
	 * return null, but several calls on this method will return the same reference.</P>
	 *
	 * @param  isDialog indicates whether the window is Dialog or Frame
	 * @return the controller which manages the parent frame / dialog
	 */
	public LyViewController getViewController(boolean isDialog)
	{
		LyViewController viewController = super.getViewController(isDialog);
		LyView			 view = viewController.getView();

		// Remove decoration from the filter window
		if (view instanceof LyDialog)
			((LyDialog)view).setDecorated(false);

		return viewController;
	}

	/**
	 * Invoked when an action occurs. (LyDemandListener interface) The current implementation calls
	 * the userInputDemand method of the behavior.
	 *
	 * @param demand the action demand.
	 */
	public void actionDemand(LyActionDemand demand)
	{
		String demandCode = demand.getSource().intern();

		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			// Validate the form : update parent filter
			if (demandCode == LyFormController.DEMAND_VALIDATE)
			{
				String			  id = demand.getSource();
				LyParameterValues paramValues = demand.getParameterValues();
				String			  value = paramValues.getParameter(SEARCH_FIELD);
				boolean			  checkValue = true;
				
				if (_fieldInfo instanceof LyRelationFieldInfo)
				{
					// For a relation field info, check if filtering is allowed (related classes must have at
					// least one text field with the NAME mark
					LyRelationFieldInfo     relation = (LyRelationFieldInfo)_fieldInfo;
					LyInfoList<LyFieldInfo> fields = getFilteredFieldsForRelation(relation);

					if ((fields != null) && (fields.getSize() > 0))
						checkValue = false;
				}

				if (checkValue)
				{
					try
					{
					    _fieldInfo.parse(value);
					}
					catch (Exception e)
					{ 
					    LyEnvironment   environment = getEnvironment();
					    String          message = environment.getMessage("LY_ER_MATCH_FORMAT",
					        new Object[] { value, _fieldInfo.getName() });
		                
					    showError(message);
		                return;
					}
				}
				
				LyFilter		  filter = buildFilter(_fieldInfo, value);

				if (_filterEndInterface != null)
					_filterEndInterface.endFastFilter(this, id, value, filter, _fieldInfo, _extend);

				free();
				return;
			}
			else if (demandCode == DEMAND_CANCEL_FILTER)
            {
                if (_filterEndInterface != null)
                    _filterEndInterface.endFastFilter(this, demand.getSource(), null, null, null, false);

                free();
                return;
            }
		}

		super.actionDemand(demand);
	}

	/**
	 * Sets client (listener) for closing notification.
	 *
	 * @param endInterface client notified when the dialog closes.
	 */
	public void setFilterEndInterface(LyFastFilterEndInterface endInterface)
	{
		_filterEndInterface = endInterface;
		_fieldInfo = (_filterEndInterface != null ?
		    _filterEndInterface.getFastFilterFieldInfo() : null);
		_defaultFilterValue = (_filterEndInterface != null ?
		    _filterEndInterface.getFastFilterValue() : null);
		_extend = ((_filterEndInterface != null) ?
		    _filterEndInterface.isFastFilterExtended() : false);
	}

	/**
	 * Gets client (listener) for closing notification.
	 *
	 * @return client notified when the dialog closes. May be null.
	 */
	public LyFastFilterEndInterface getFastFilterEndInterface()
	{
		return _filterEndInterface;
	}
} // end class LyFastFilterController
