/**
 * LyFilterController.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved. Current version: $Id:
 * LyFilterController.java,v 6.21.4.2 2009/02/09 13:04:28 jls Exp $
 */

package leon.control;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.view.*;
import leon.view.beaninterface.*;
import leon.view.event.*;
import leon.view.getinterface.*;

import java.util.*;


/**
 * This class implements the simple filter edition area. This form enables the user to type in
 * elementary filtering criteria (fieldInfo / condition / value).
 *
 * <P>This controller is associated to a behavior which implements LyFilterBehaviorInterface.
 * Usually the behavior is an instance of LyClassBehavior.</P>
 *
 * <P>A simple filter can only create a simple filter criteria (1 field info / 1 condition (lower
 * than, greater than, equals ...) and 1 value). To build advanced filters, use
 * LyFilterEditorController.</P>
 *
 * @see  LyFilterEditorController
 * @see  LyFilterEndInterface
 */
public class LyFilterController extends LySimpleFilterController
	implements LyFormResultInterface
{
	/** File Version */
	public static final String __VERSION = "$Id: LyFilterController.java 41407 2012-10-30 16:17:49Z bangiole $";

	/** Identifier of an input field : selected fieldInfo. */
	public static final String SELECT_FIELDINFO = "_filterfieldinfo";

	/** Identifier of an input field : selected condition. */
	public static final String SELECT_CONDITION = "_filtercondition";

	/** Identifier of an input field : user specified value. */
	public static final String SELECT_VALUE = "_filtervalue";

	/** Identifier of an input field : selected day. */
	public static final String SELECT_DATE_DAY_VALUE = "_filterdatedayvalue";

	/** Identifier of an input field : selected modifiers. */
	public static final String SELECT_MODIFIERS = "_filtermodifiers";

	/** Identifier of an input field : selected today. */
	public static final String SELECT_DATE_SPECIAL_VALUE = "_filterdatespecialvalue";

	/** Id for modifier NOT. */
	public static final String ID_MODIFIER_NOT = "_notModifierId";

	/** Id for modifier DATE_NORMAL. */
	public static final String ID_DATE_NORMAL = "_dateNormal";

	/** Id for modifier TODAY. */
	public static final String ID_TODAY = "_today";

	/** Id for modifier NOW. */
	public static final String ID_NOW = "_now";

	/** Id for modifier DAY. */
	public static final String ID_DAY = "_day";

	/** Id for operation remove. */
	public static final String ID_OPERATION_REMOVE = "_filterOperationRemove";

	/** The last selected field info */
	protected LyFieldInfo _lastFieldInfo;

	/** The last date special value */
	protected String _lastDateSpecialValue;

	/**
	 * Constructor.
	 *
	 * @param  session    current session associated to the controller. The controller automatically
	 *                    registers on the specified session.
	 * @param  parent     parent controller. If the parent implements LyFilterEndInterface, it will
	 *                    be notified when the filter form is closed.
	 * @param  action     related action. (action based on '_filter' action)
	 * @param  classInfo  the application class associated to the filter
	 * @param  objects    the current context for this action
	 */
	public LyFilterController(LySession session, LyController parent, LyAction action, LyClassInfo classInfo,
		LyObjectList objects)
	{
		super(session, parent, action, classInfo, objects);

		// If new null initial filter
		String id_action = action.getId();

		if ((id_action.indexOf("new") != -1)
				|| (id_action.indexOf("New") != -1))
			_defaultFilter = null;
	}

	/**
	 * Internal method to compute id of this controller. This method will always return an id based
	 * on this parent
	 *
	 * @param   session  Session of the new controller.
	 * @param   id       Desired id (may be null).
	 *
	 * @return  Id of the new controller.
	 */
	protected String computeId(LySession session, String id)
	{
		if ((_parent instanceof LyCompoundFilterController)
				|| (_parent instanceof LyFilterEditorController))
		{
			// Redirect demand on compound filter
			// Get the parent id
			LyCompoundFilterController compound = null;

			if (_parent instanceof LyCompoundFilterController)
				compound = (LyCompoundFilterController)_parent;
			else if ((_parent instanceof LyFilterEditorController)
					&& (_parent.getParent() instanceof LyCompoundFilterController))
				compound = (LyCompoundFilterController)_parent.getParent();

			if (compound != null)
				return "FilterControllerId" + compound.getId();
		}

		return super.computeId(session, id);
	}

	/**
	 * Accessor to the view component instance. (XML description of the view) This instance is
	 * created by the builder, and multiple calls to this method always return the same instance.
	 *
	 * @return  the LyComponent instance of the view package, memory representation (parsed) of the
	 *          XML description of the view.
	 */
	public LyComponent getComponent()
	{
		if (_component == null)
		{
			super.getComponent();

			LyFormField formField = (LyFormField)_component.getComponent(LyFilterController.SELECT_FIELDINFO);

			if (formField == null) // Multi-criteria filter
				return _component;

			LyRelation  relation = (LyRelation)formField.getFieldItem();
			int		    i, n = relation.getLinkCount();
			LyLink	    link;
			LyClassInfo classInfo = getClassInfo();
			String	    fieldId;

			for (i = 0; i < n; i++)
			{
				link = relation.getLink(i);
				if (link.isSelected())
				{
					fieldId = link.getId();
					_lastFieldInfo = classInfo.getFieldInfo(fieldId);
					break;
				}
			}
		}

		return super.getComponent();
	} // end method getComponent

	/**
	 * Invoked when an action notification is sent to this listener.
	 *
	 * <P>method implements VALIDATE, APPLY and CANCEL action events and subsequently invoke
	 * endFilter on the client.</P>
	 *
	 * <P>Additionally, the controller handles field changes and impact on form content. (The nature
	 * of components shown in the view is field dependant).</P>
	 *
	 * @param  demand  demand parameters sent by the view component.
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			String id = demand.getSource();

			if (_parent instanceof LyCompoundFilterController)
			{
				if (id.equals(LyFormController.DEMAND_APPLY)
						|| id.equals(LyFormController.DEMAND_VALIDATE)
						|| id.equals(LyFormController.DEMAND_CANCEL)
						|| id.equals(LyCompoundFilterController.DEMAND_ADD_FILTER))
				{
					// Redirect demand on compound filter
					((LyCompoundFilterController)_parent).actionDemand(demand);
					return;
				}
			}

			// Menu or action bar
			if (id.equals(LyFormController.DEMAND_CANCEL))
			{
				if (_hasViewController)
				{
					if (_filterEndInterface != null)
						_filterEndInterface.endFilter(this, id, null);

					free();
				}
				else
				{
					if (_filterEndInterface != null)
						_filterEndInterface.endFilter(this, DEMAND_VALIDATE, null);

					hideView();
					free(_parent);
				}

				return;
			}

			if (!id.equals(LyFormController.DEMAND_VALIDATE)
					&& !id.equals(LyFormController.DEMAND_APPLY))
			{
				super.actionDemand(demand);
				return;
			}

			validate(demand);
			return;
		}
		else if (demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
		{
			_modified = true;

			if (LyFilterController.SELECT_FIELDINFO.equals(demand.getSource()))
			{
				// Refresh view
				LyFilterBuilder builder = (LyFilterBuilder)getBuilder();
				String		    fieldId = demand.getValue();
				LyClassInfo     classInfo = getClassInfo();
				LyFieldInfo     fieldInfo = classInfo.getFieldInfo(fieldId);

				if (!areComparable(fieldInfo, _lastFieldInfo))
				{
					builder.resetConditionField(fieldId);
					builder.resetValueField(fieldId);
					builder.resetModifiersField(fieldId);
					builder.resetDateSpecialValues(fieldId);
					builder.resetForm();
				}

				_lastFieldInfo = fieldInfo;
			}
			else if (LyFilterController.SELECT_CONDITION.equals(demand.getSource()))
			{
				// Refresh only if text and text has a given format
				LyParameterValues values = demand.getParameterValues();
				String			  fieldId = values.getParameter(LyFilterController.SELECT_FIELDINFO);

				if (fieldId != null)
				{
					LyClassInfo classInfo = getClassInfo();
					LyFieldInfo fieldInfo = classInfo.getFieldInfo(fieldId);

					if ((fieldInfo instanceof LyTextFieldInfo)
							&& (((LyTextFieldInfo)fieldInfo).getFormat() != null))
					{
						// Refresh view
						LyFilterBuilder builder = (LyFilterBuilder)getBuilder();

						builder.resetValueField(fieldId);
						builder.resetForm();
					}
				}
			}
			else if(LyFilterController.SELECT_DATE_SPECIAL_VALUE.equals(demand.getSource()))
			{
				LyFilterBuilder builder = (LyFilterBuilder)getBuilder();
				String		    fieldId = demand.getValue();

				if (!fieldId.equals(_lastDateSpecialValue)) {

					if (fieldId.equals(ID_TODAY) || fieldId.equals(ID_NOW))
					{
						builder.enableValueField(false);
						builder.showDayValueField(false);
					}
					else if(fieldId.startsWith(ID_DAY))
					{
						builder.enableValueField(true);
						builder.showDayValueField(true);
					}
					else
					{
						builder.enableValueField(true);
						builder.showDayValueField(false);
					}

					builder.resetForm();
				}

				_lastDateSpecialValue = fieldId;
			}

			return;
		}

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Internal method to validate filter creation.
	 * @param demand the action demand
	 */
	protected void validate(LyActionDemand demand)
	{
		LyFilter filter = null;

		try
		{
			filter = buildFilter(demand);
		}
		catch (IllegalStateException e)
		{
			if ((e.getMessage() != null)
					&& (e.getMessage().equals("LyValidateInterface.STATUS_KO")))
				return;

			throw e;
		}

		String id = demand.getSource();

		if (filter == null)
		{
			if (id.equals(LyFormController.DEMAND_VALIDATE))
				// Nothing done : action is considered as cancelled
				free();

			if (_filterEndInterface != null)
				_filterEndInterface.endFilter(this, id, null);

			return;
		}

		if (filter.getSize() == 0) // REMOVE dummy filter
		{
			filter = null;
			if (_filterEndInterface instanceof LySimpleTableController)
				((LySimpleTableController)_filterEndInterface).setFilter(null);
		}

		LyFilter newFilter = filter;

		if (_filterEndInterface != null)
		{
			_modified = false;
			_filterEndInterface.endFilter(this, id, filter);
			newFilter = _filterEndInterface.getFilter();
		}
		else
		{
			LySession		 session = getSession();
			LyAreaController areaController;
			LyAction		 action;
			int				 i, n = _action.getActionCount();

            if (n == 0)
            {
                action = getApplication().getAction("_multiPageTable");
                areaController = (LyAreaController)session.invokeAction(_parent.getParent(), action, getClassInfo(), null, false);

                if (areaController instanceof LyFilterEndInterface)
                    ((LyFilterEndInterface)areaController).endFilter(this, id, filter);

                areaController.showView();
            }
            else
            {
                for (i = 0; i < n; i++)
                {
                    action = _action.getAction(i);
                    areaController = (LyAreaController)session.invokeAction(_parent.getParent(), action, getClassInfo(),
                            null, false);

                    if (areaController instanceof LyFilterEndInterface)
                        ((LyFilterEndInterface)areaController).endFilter(this, id, filter);

                    areaController.showView();
                }
            }
		}

		if (!(_parent instanceof LyCompoundAreaController))
			if (id.equals(LyFormController.DEMAND_VALIDATE)
					|| (_hasViewController && (newFilter != filter)))
				free();
	}

	/**
	 * Internal method used to compare 2 fieldInfos when a fieldInfo is selected in the view. This
	 * method returns false if one of the fieldInfo is null or is a relation or if the 2 fieldInfos
	 * are not of the same type.
	 *
	 * @param   fieldInfo1  First fieldInfo
	 * @param   fieldInfo2  Second fieldInfo
	 *
	 * @return  true if the 2 fieldInfos are comaparable.
	 */
	protected boolean areComparable(LyFieldInfo fieldInfo1, LyFieldInfo fieldInfo2)
	{
		if ((fieldInfo1 instanceof LyRelationFieldInfo)
				|| (fieldInfo2 instanceof LyRelationFieldInfo)
				|| (fieldInfo1 == null)
				|| (fieldInfo2 == null)
				|| (fieldInfo1.getClass() != (fieldInfo2.getClass())))
			return false;

		// Check beans used for filter actions
		LyFieldBean fieldBean1 = fieldInfo1.getFieldBean(LyAction.ID_NEW_FILTER);
		LyFieldBean fieldBean2 = fieldInfo2.getFieldBean(LyAction.ID_NEW_FILTER);

		if (fieldBean1 != fieldBean2)
			return false;

		if (fieldInfo1 instanceof LyChoiceFieldInfo)
			return false;

		// Check types and multiple
		if (fieldInfo1.isMultiple() != fieldInfo2.isMultiple())
			return false;
			// if (fieldInfo1.getType() != fieldInfo2.getType())
			// return false;

		// For number fields, check min and max
		if ((fieldInfo1 instanceof LyNumberFieldInfo)
				&& (fieldInfo2 instanceof LyNumberFieldInfo))
		{
			Number min1 = ((LyNumberFieldInfo)fieldInfo1).getMin();
			Number min2 = ((LyNumberFieldInfo)fieldInfo2).getMin();

			if (((min1 == null) && (min2 != null))
					|| ((min1 != null) && (min2 == null))
					|| ((min1 != null) && (min2 != null)
						&& (min1.doubleValue() != min2.doubleValue())))
				return false;

			Number max1 = ((LyNumberFieldInfo)fieldInfo1).getMax();
			Number max2 = ((LyNumberFieldInfo)fieldInfo2).getMax();

			if (((max1 == null) && (max2 != null))
					|| ((max1 != null) && (max2 == null))
					|| ((max1 != null) && (max2 != null)
						&& (max1.doubleValue() != max2.doubleValue())))
				return false;
		}

		// For text fields, check the formats
		if ((fieldInfo1 instanceof LyTextFieldInfo)
				&& (fieldInfo2 instanceof LyTextFieldInfo))
		{
			String format1 = ((LyTextFieldInfo)fieldInfo1).getFormat();
			String format2 = ((LyTextFieldInfo)fieldInfo2).getFormat();

			if (((format1 == null) && (format2 != null))
					|| ((format1 != null) && (format2 == null))
					|| ((format1 != null) && (format2 != null) && !format1.equals(format2)))
				return false;

			int ncols1 = ((LyTextFieldInfo)fieldInfo1).getNCols();
			int ncols2 = ((LyTextFieldInfo)fieldInfo2).getNCols();

			if (ncols1 != ncols2)
				return false;
		}

		// For date fields, check the type
		if ((fieldInfo1 instanceof LyDateFieldInfo)
				&& (fieldInfo2 instanceof LyDateFieldInfo))
		{
			if (((LyDateFieldInfo)fieldInfo1).getType()
					!= ((LyDateFieldInfo)fieldInfo2).getType())
				return false;
			if (((LyDateFieldInfo)fieldInfo1).isWithTimeZone()
					!= ((LyDateFieldInfo)fieldInfo2).isWithTimeZone())
				return false;
		}

		if ((fieldInfo1 instanceof LyArrayFieldInfo) && (fieldInfo2 instanceof LyArrayFieldInfo))
		{
			LyFieldInfo internalField1 = ((LyArrayFieldInfo)fieldInfo1).getFieldInfo();
			LyFieldInfo internalField2 = ((LyArrayFieldInfo)fieldInfo2).getFieldInfo();

			if (!areComparable(internalField1, internalField2))
				return false;
		}

		return true;
	} // end method areComparable

	/**
	 * Get current unit for the specified fieldInfo in the form.
	 *
	 * @param fieldInfo the fieldInfo
	 * @param formField the associated formfield
	 *
	 * @return  current unit, or null if not found or field is disabled.
	 */
	public leon.info.LyUnit getUnit(LyFieldInfo fieldInfo, LyFormField formField)
	{
		LyUnits units = fieldInfo.getUnits();

		if (units == null)
			return null;

		if ((formField == null) || !formField.isEnabled())
			return getDefaultFieldUnit(fieldInfo);

		LyFieldUnit formFieldUnit = formField.getFieldUnit();

		if (formFieldUnit == null)
			return null;

		LyFieldItem unitItem = formFieldUnit.getFieldItem();
		String	    unitId = unitItem.getValue();

		if ((unitId == null) && (unitItem instanceof LyRelation))
		{
			LyRelation relation = (LyRelation)unitItem;
			int		   nLink = relation.getLinkCount();
			LyLink     link;

			for (int j = 0; (j < nLink) && (unitId == null); j++)
			{
				link = relation.getLink(j);
				if (link.isSelected())
					unitId = link.getId();
			}
		}

		LyUnit unit = units.getUnit(unitId);

		if (unit == null)
			unit = getDefaultFieldUnit(fieldInfo);

		return unit;
	} // end method getUnit

	/**
	 * Builds the filter from the data Structure This methods extracts values typed in by the user
	 * and create the LyFilter instance.
	 *
	 * @param   demand  the demand which contains all values submited by the form.
	 *
	 * @return  the LyFilter instance or null if an error occured.
	 */
	protected LyFilter buildFilter(LyDemand demand)
	{
		if (demand == null)
			return null;

		LyParameterValues paramValues = demand.getParameterValues();

		if (paramValues == null)
		{
			getEnvironment().writeException(new Exception("Stack trace"));
			return null;
		}

		// Retrieve field
		String fieldId = paramValues.getParameter(LyFilterController.SELECT_FIELDINFO);

		if (fieldId == null)
			return null;

		// Build condition
		String						 condId = paramValues.getParameter(LyFilterController.SELECT_CONDITION);
		LyFilterExpression.Condition condition;
        boolean                      not = condId.startsWith("_not.");

        if (not)
            condId = condId.substring(5);

		condition = LyFilterExpression.getCondition(condId);

		// Build modifiers BitSet
		BitSet	    modifiersBS = new BitSet();
		String[]    modifiers = paramValues.getParameterValues(LyFilterController.SELECT_MODIFIERS);
		LyClassInfo classInfo = getClassInfo();
		LyFieldInfo fieldInfo = classInfo.getFieldInfo(fieldId);

		if ((modifiers != null)
			&& (fieldInfo.getFilterModifiers() != null))
		{
			int   m, nm = modifiers.length;
			short modifIndex;

			for (m = 0; m < nm; m++)
			{
				modifIndex = LyFilterExpression.getModifier(modifiers[m]);
				if (modifIndex >= 0)
					modifiersBS.set(modifIndex);
			}
		}

        not ^= ((modifiers != null)
                && (modifiers.length > 0)
                && modifiers[0].equals(ID_MODIFIER_NOT));

		// For a relation or a choice, values may be empty but not null.
		// Built expression must contain the null value.

		// Create filter
		LyFilter filter = new LyFilter(not ? LyFilter.Type.AND : LyFilter.Type.OR);

		// Build filter expression
		String[]	    values;
		String dateSpecialValue = paramValues.getParameter(LyFilterController.SELECT_DATE_SPECIAL_VALUE);

		if(!(fieldInfo instanceof LyDateFieldInfo))
		{
			values = paramValues.getParameterValues(LyFilterController.SELECT_VALUE);
		}
		else if((dateSpecialValue == null) || (dateSpecialValue.equals(LyFilterController.ID_DATE_NORMAL)))
		{
			values = paramValues.getParameterValues(LyFilterController.SELECT_VALUE);
		}
		else
		{
			values = paramValues.getParameterValues(LyFilterController.SELECT_DATE_SPECIAL_VALUE);
		}

		LyFilterElement filterElem = null;

		// Validate input if a bean is set
		LyFieldItem fieldItem = getFieldItem(fieldInfo);
		short	    resu = fieldItem.validateInput();

		if (resu == LyValidateInterface.STATUS_KO)
			// STOP : an error in user input occurred.
			throw new IllegalStateException("LyValidateInterface.STATUS_KO");

		if (resu == LyValidateInterface.STATUS_CANCEL)
			return null; // CANCEL Filter (and close)

		if (values == null)
		{
			filterElem = buildFilterElement(fieldInfo, not, condition, null, modifiersBS);
			filter.addElement(filterElem);
		}
		// If choice field and multiple, use only one filter expression
		// with all selected values.
		else if (((fieldInfo instanceof LyRelationFieldInfo)
					|| (fieldInfo instanceof LyChoiceFieldInfo))
				&& fieldInfo.isMultiple())
		{
			LyApplication app = getSession().getApplication();
			boolean       nullValue = false;

	        if (values.length == 1)
	        {
                if ((fieldInfo instanceof LyChoiceFieldInfo)
                        && "LY_NO_CHOICE".equals(values[0]))
                    nullValue = true;
                else if ((fieldInfo instanceof LyRelationFieldInfo)
                        && ("LY_NO_CHOICE".equals(values[0])
                            || "LY_FILTER_COMPOSITION_EMPTY_VALUE".equals(values[0])))
                    nullValue = true;
	        }

			Object   parsedValue = (nullValue ? null : fieldInfo.parse(LyString.listToString(values, app.VALUE_SEP)));

			filterElem = buildFilterElement(fieldInfo, not, condition, parsedValue, modifiersBS);
			filter.addElement(filterElem);
		}
		else
		{
			Object				 parsedValue;
			int                  j, n = values.length;
			LyFilterBuilder		 builder = (LyFilterBuilder)getBuilder();
			LyFieldItem			 item;
			LyFieldBeanInterface component;
			LyFormField			 formField = (LyFormField)fieldItem.getParent();
			LyUnit				 unit = getUnit(fieldInfo, formField);

			for (j = 0; j < n; j++)
			{
				// if the field has a bean , send the value to the component and get it back
				if ((fieldInfo.getFieldBean(LyAction.ID_NEW_FILTER) != null)
					&& (values[j] != null))
				{
					item = getFieldItem(fieldInfo);
					component = item.getBeanComponent();
					if (component instanceof LyGetLabelInterface)
					{
						component.setLabel(values[j]);
						values[j] = ((LyGetLabelInterface)component).getLabel();
					}
				}

				try
				{
					// If we have a textFieldInfo, and the filter condition is contains, we
					// must take directly the given value. If there is a format (or a regex)
					// the parse will failed
					if ((fieldInfo instanceof LyTextFieldInfo)
							&& (condition == LyFilterExpression.Condition.CONTAINS))
						parsedValue = values[j];
					else if (fieldInfo instanceof LyDateFieldInfo)
					{
						LyDateFieldInfo dateField = (LyDateFieldInfo)fieldInfo;
						if ("now".equals(values[j]))
							values[j] = ID_NOW; // Be sure to be consistent
						else if ("today".equals(values[j]))
							values[j] = ID_TODAY; // Be sure to be consistent)
						else if(values[j].startsWith(ID_DAY))
						{
							String[] tVals = values = paramValues.getParameterValues(LyFilterController.SELECT_DATE_DAY_VALUE);
							if((tVals != null) && (tVals.length ==1))
								values[j] = ID_DAY+"_"+tVals[0].replace('/', '_').split(" ")[0];
						}

						String tz = null;
						if (dateField.isWithTimeZone())
						{
							LyFieldUnit formFieldUnit = formField.getFieldUnit();

							if (formFieldUnit == null)
								return null;

							LyFieldItem unitItem = formFieldUnit.getFieldItem();
							tz = unitItem.getValue();
						}
						if (values == null)
							parsedValue = null;
						else if (ID_NOW.equals(values[j])
								|| ID_TODAY.equals(values[j])
								|| values[j].startsWith(ID_DAY))
							parsedValue = values[j];
						else
							parsedValue = dateField.parse(values[j], tz);
					}
					else if ((fieldInfo instanceof LyChoiceFieldInfo)
							&& "LY_NO_CHOICE".equals(values[j]))
						parsedValue = null;
	                else if ((fieldInfo instanceof LyRelationFieldInfo)
	                        && ("LY_NO_CHOICE".equals(values[j])
	                            || "LY_EMPTY_RELATION_VALUE".equals(values[j])))
                        parsedValue = null;
					else
						parsedValue = fieldInfo.parse(values[j]);
				}
				catch (LyFieldInfoFormatException e)
				{
					builder.setValueFieldStatus(false);
					showError(e.getMessage());
					return null;
				}

				// FFT00058 : allow filter on multiple text field
				if ((fieldInfo instanceof LyTextFieldInfo)
						&& fieldInfo.isMultiple())
					parsedValue = LyString.replace((String)parsedValue, "\r", "");

				if (unit != null)
					parsedValue = fieldInfo.convertFromFieldUnit(parsedValue, unit);

				filterElem = buildFilterElement(fieldInfo, not, condition, parsedValue, modifiersBS);
				filter.addElement(filterElem);
			}
		}

		_defaultFilter = filter;
		return filter;
	} // end method buildFilter

	/**
	 * Builds filter element for given parameters.
	 *
	 * @param   fieldInfo  FieldInfo on which this filter expression applies.
	 * @param   not        Inverts filter condition or not.
	 * @param   condition  Filter condition.
	 * @param   value      Value of this expression (may be null). This value is an instance of the
	 *                     class returned by the getValueClass method of LyFieldInfo.
	 * @param   modifiers  List of modifiers (may be null).
	 *
	 * @return  The corresponding filter element.
	 */
	public LyFilterElement buildFilterElement(LyFieldInfo fieldInfo, boolean not,
		LyFilterExpression.Condition condition, Object value, BitSet modifiers)
	{
		LyFilterBehaviorInterface behavior = (LyFilterBehaviorInterface)getBehavior();
		LyFilterElement			  filterElement;

		if (behavior != null)
		{
			filterElement = behavior.buildFilterElement(this, fieldInfo, not, condition, value, modifiers);
			if (filterElement != null)
				return filterElement;
		}

		filterElement = new LyFilterExpression(fieldInfo, not, condition, value, modifiers);
		return filterElement;
	}

	/**
	 * Indicates if the view for this import controller is resizable.
	 *
	 * @return  true if the view is resizable, false otherwise.
	 */
	public boolean isViewResizable()
	{
		return false;
	}

	/**
	 * Factory of the default builder for a controller. Current implementation returns a
	 * LyFilterBuilder instance.
	 *
	 * @return  the builder which creates the XML description of the view
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyFilterBuilder(this);
	}

	/**
	 * Called when current object has changed . Keep the overall content consistent.
	 *
	 * @param  oldValues  set of former values, or null if not specified.
	 */
	protected void updateObject(LyValueSet oldValues)
	{
		super.updateObject(oldValues);
	}

	/**
	 * Accessor to the list of objects associated with this form. The user may navigate through this
	 * list of objects.
	 *
	 * @param   objects  list of objects which could be shown in the form.
	 *
	 * @return  true if successfull
	 */
	public boolean setObjects(LyObjectList objects)
	{
		// Not used in current version
		// Could be used to have field domains that depend on context
		return false;
	}

	/**
	 * Accessor to the formField (View component) associated to the fieldInfo.
	 *
	 * @param   fieldInfo  the fieldInfo which is associated to the view component.
	 *
	 * @return  the formField corresponding to the given fieldInfo
	 */
	public LyFormField getFormField(LyFieldInfo fieldInfo)
	{
		LyFormField formField = (LyFormField)_component.getComponent(LyFilterController.SELECT_FIELDINFO);

		if (formField == null) // Multi-criteria filter
			return super.getFormField(fieldInfo);

		LyRelation relation = (LyRelation)formField.getFieldItem();
		int		   i, n = relation.getLinkCount();
		LyLink     link;

		for (i = 0; i < n; i++)
		{
			link = relation.getLink(i);
			if (link.isSelected()
					&& link.getId().equals(fieldInfo.getId()))
				return (LyFormField)_component.getComponent(LyFilterController.SELECT_VALUE);
		}

		return null;
	}

	/**
	 * Internal use. When an object is created in a popup creation form for a relation, the child
	 * controller invokes this method to set the newly created object as the current value of the
	 * relation (if a new object is created we automatically select this value). This method is
	 * never used by the application : use setFieldValue instead
	 *
	 * @param  object         newly created object. The new value of the relation fieldInfo.
	 * @param  relationField  the relation
	 */
	public void notifyFormResult(LyObject object, LyRelationFieldInfo relationField)
	{
		if (relationField == null)
			return;

		LyFieldItem fieldItem = getFieldItem(relationField);

		if (!(fieldItem instanceof LyRelation))
			return;

		LyRelation relation = (LyRelation)fieldItem;
		LyLink     link;
		int		   i, n = relation.getLinkCount();
		String     id;
		boolean    multiple = relation.isMultiple();
		boolean    found = false;

		for (i = 0; i < n; i++)
		{
			link = relation.getLink(i);
			id = link.getId();

			if ((object != null)
					&& (id != null)
					&& id.equals(object.getId()))
			{
				link.setSelected(true);
				found = true;
				if (multiple)
					break;
			}
			else if ((!multiple) && (link.isSelected()))
				link.setSelected(false);
		}

		if (!found)
		{
		    if (relationField.hasMark(LyFieldInfo.MARK_TRANSLATE))
		        link = new LyLink(object.getId(), getName(object));
		    else
		        // $$ is added to avoid translation of the name
		        link = new LyLink(object.getId(), "$$" + getName(object));

			// Add images only if multiple classes
			int     nClasses = relationField.getClassInfoCount();
			boolean addImage = (nClasses
						>= getEnvironment().getIntResource("LY_SHOW_RELATION_IMAGES"));

			if (addImage)
				link.setImage(getImage(object));

			relation.addLink(link);
			link.setSelected(true);
		}
	} // end method notifyFormResult
} // end class LyFilterController
