/**
 * LyFilterCriterionController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyFilterCriterionController.java 34469 2012-03-12 10:06:54Z bangiole $
 */

package leon.control;

import leon.app.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.event.*;


/**
 * This class implements the criterion filter edition area. This form enables the user to choose a
 * filter criterion.
 *
 * <P>This controller is associated to a behavior which implements LyFilterBehaviorInterface.
 * Usually the behavior is an instance of LyClassBehavior.</P>
 *
 * @see LyFilterEditorController
 * @see LyFilterEndInterface
 */
public class LyFilterCriterionController extends LySimpleFilterController
{
	/** File Version */
	public static final String     __VERSION =
		"$Id: LyFilterCriterionController.java 34469 2012-03-12 10:06:54Z bangiole $";

	/** Identifier of an input field : user specified value. */
	public static final String     SELECT_CRITERION = "_criterionvalue";


	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller. If the parent implements LyFilterEndInterface, it will be
	 *                  notified when the filter form is closed.
	 * @param action    related action. (action based on '_filter' action)
	 * @param classInfo the application class associated to the filter
	 * @param objects   current context for this action
	 */
	public LyFilterCriterionController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		super(session, parent, action, classInfo, objects);
	}

	/**
	 * Invoked when an action notification is sent to this listener.
	 *
	 * <P>method implements VALIDATE, APPLY and CANCEL action events and subsequently invoke
	 * endFilter on the client.</P>
	 *
	 * <P>Additionally, the controller handles field changes and impact on form content.</P>
	 *
	 * @param demand demand parameters sent by the view component.
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			// Menu or action bar
			String id = demand.getSource();
			if (_parent instanceof LyCompoundFilterController)
			{
				if (id.equals(LyFormController.DEMAND_APPLY)
						|| id.equals(LyFormController.DEMAND_VALIDATE)
						|| id.equals(LyFormController.DEMAND_CANCEL)
						|| id.equals(LyCompoundFilterController.DEMAND_ADD_FILTER))
				{
					// Redirect demand on compound filter
					((LyCompoundFilterController)_parent).actionDemand(demand);
					return;
				}
			}
			if (id.equals(LyFormController.DEMAND_CANCEL))
			{
				if (_filterEndInterface != null)
					_filterEndInterface.endFilter(this, id, null);

				free();
				return;
			}

			if (!id.equals(LyFormController.DEMAND_VALIDATE)
					&& !id.equals(LyFormController.DEMAND_APPLY))
				return;

			validate(demand);

			if (id.equals(LyFormController.DEMAND_VALIDATE))
				free();
			return;
		}

		super.actionDemand(demand);
	} // end method actionDemand

    /**
     * Internal method to validate filter creation.
     * @param demand the action demand
     */
    protected void validate(LyActionDemand demand)
    {
        LyFilterCriterion filterCriterion = getFilterCriterion();

        // If no filter has been chosen, show an error
        if (filterCriterion == null)
        {
            LyEnvironment environment = getEnvironment();

            showError(environment.getMessage("LY_ER_NOTNULL", "LY_FILTER_CRITERION"));
            return;
        }

        // Notify filter interface of the creation
        if (_filterEndInterface != null)
        {
            String      id = demand.getSource();
            LyFilter    filter = new LyFilter();

            filter.addElement(filterCriterion.cloneFilterCriterion());
            _filterEndInterface.endFilter(this, id, filter);
        }
    }
    
	/**
	 * Returns the created criterion filter.
	 *
	 * @return the created criterion filter
	 */
	protected LyFilterCriterion getFilterCriterion()
	{
		LyFilterCriterionBuilder builder = (LyFilterCriterionBuilder)getBuilder();
		LyFilterCriterion		 criterion = builder.getSelectedCriterion();

		return criterion;
	}

	/**
	 * Indicates if the view for this import controller is resizable.
	 *
	 * @return true if the view is resizable, false otherwise.
	 */
	public boolean isViewResizable()
	{
		return false;
	}

	/**
	 * Factory of the default builder for a controller. Current implementation returns a
	 * LyFilterCriterionBuilder instance.
	 *
	 * @return the builder which creates the XML description of the view
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyFilterCriterionBuilder(this);
	}

	/**
	 * Called when current object has changed . Keep the overall content consistent.
	 *
	 * @param oldValues set of former values, or null if not specified.
	 */
	protected void updateObject(LyValueSet oldValues)
	{
	}

	/**
	 * Accessor to the list of objects associated with this form. The user may navigate through this
	 * list of objects.
	 *
	 * @param  objects list of objects which could be shown in the form.
	 * @return true if successfull
	 */
	public boolean setObjects(LyObjectList objects)
	{
		// Not used in current version
		// Could be used to have field domains that depend on context
		return false;
	}
} // end class LyFilterCriterionController
