/**
 * LyFilterEditorController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyFilterEditorController.java 34904 2012-03-21 15:55:41Z bangiole $
 */
package leon.control;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.setinterface.*;

import java.util.*;


/**
 * This class implements an extended filter edition form. The extended filter is a 'tree'-like view
 * which is used as a GUI to edit extended filters. An extended filter is a set of elementary
 * conditions associated with operators AND / OR.
 *
 * <P>The extended filter view is based on a LyFilterController (elementary edition area) to create
 * conditions and add them into the tree.</P>
 *
 * @see LyFilterController
 * @see LyFilterEndInterface
 */
public class LyFilterEditorController extends LyMultiAreaController implements LyFilterEndInterface,
	LyQuestionAnswerInterface
{
	/** File Version */
	public static final String __VERSION = "$Id: LyFilterEditorController.java 34904 2012-03-21 15:55:41Z bangiole $";

	/** String used when modifying a filter expression. */
	public static final String CURRENT_FILTER_EXPRESSION = "_currentFilterExpression";

	/** Associated Filter. */
	protected LyFilter _filter;

	/** Selected objects. */
	protected Vector<LyFilterElement> _selections;

	/** Controller table used to build LyFilterExpression. */
	protected LyClassInfo _classInfo;

	/** Filter parent used to build LyFilterExpression. */
	protected LyFilter _filterParent;

	/**
	 * Client which may be notified when the filter forms is closed to update its content (usually a LyTableController).
	 */
	protected LyFilterEndInterface _filterEndInterface;

	/** Filter in buffer (cut or copy action). */
	protected LyFilterElement _bufferFilter;

	/** Association table treeObject, application object. */
	protected LyDependencies _objectAssoc;

	/** Current filter controller. */
	protected LySimpleFilterController _filterController;

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller. If the parent implements LyFilterEndInterface, it will be
	 *                  notified when the filter form is closed.
	 * @param action    related action. (action based on '_filter' action)
	 * @param classInfo the application class associated to the filter
	 * @param objects   the current context
	 */
	public LyFilterEditorController(LySession session, LyController parent, LyAction action, LyClassInfo classInfo,
		LyObjectList objects)
	{
		super(session, parent, action, null);
		init();
		if (parent instanceof LyFilterEndInterface)
			setFilterEndInterface((LyFilterEndInterface)parent);

		if (action.getClassInfo() != null)
			_classInfo = action.getClassInfo();
		else
			_classInfo = classInfo;
	}

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller. If the parent implements LyFilterEndInterface, it will be
	 *                notified when the filter form is closed.
	 * @param action  related action. (action based on '_filter' action)
	 */
	public LyFilterEditorController(LySession session, LyController parent, LyAction action)
	{
		this(session, parent, action, null, null);
	}

	/**
	 * Free this controller.
	 *
	 * @param controller parent controller
	 */
	public void free(LyController controller)
	{
		if (_parent != controller)
			return;

		freeFilterController();
		super.free(controller);
	}

	/**
	 * Initialize controller.
	 */
	protected void init()
	{
		_filter = new LyFilter(LyFilter.Type.AND);
		_selections = new Vector<LyFilterElement>(1);
		_objectAssoc = new LyDependencies();
	}

	/**
	 * Reset the entire view : everything is rebuilt.
	 *
	 * @param full Boolean indicating if a full reset (including reset of datas) is requiered or not
	 *             (only reset graphical area).
	 */
	public void reset(boolean full)
	{
		updateContent();
		updateTools();

		// Update the last modification time.
		setLastModified(new Date().getTime());
		super.reset(full);
	}

	/**
	 * Update the content of the view (Nodes).
	 */
	public void updateContent()
	{
		LyTreeNode treeNode;
		LyTree     tree = getTreeComponent();

		// Clear all
		tree.setIdle(true);
		clearTreeContent();

		// build tree content (null, null : root level, then recurse)
		treeNode = addTreeNode(null, _filter, -1);
		if (treeNode != null)
		{
			treeNode.setOpen(true);
			treeNode.setSelected(true);
			tree.setTreeNode(treeNode);
		}

		// Select root node
		Vector<LyFilterElement> v = new Vector<LyFilterElement>(1);

		v.addElement(_filter);
		setSelections(v);
		tree.setIdle(false);
		updateTools();
	} // end method updateContent

	/**
	 * Empty the tree (remove all nodes).
	 */
	public void clearTreeContent()
	{

		// Retrieve internal tree component
		LyTree tree = getTreeComponent();

		tree.setTreeNode(null);

		// Clear assoc table
		_objectAssoc = new LyDependencies();
	}

	/**
	 * Create a node subtree for the parent node.
	 *
	 * @param  parent    of the subtree, or null if create subtree a root level
	 * @param  appObject the application object which is the node of the subtree, or null to fetch
	 *                   root component
	 * @param  position  Positon of the new tree node.
	 * @return subtreenode associated with Application object and its own subtree nodes
	 */
	public LyTreeNode addTreeNode(LyTreeNode parent, Object appObject, int position)
	{
		LyTreeNode treeNode = null;

		// Attempt to build the node from the tree controller
		treeNode = buildTreeNode(appObject);
		// Found a subtree ?
		if (treeNode == null)
			return null;

		treeNode.setOpen(true);

		// Keep association in dependencies table
		_objectAssoc.addDepend(treeNode, appObject);

		// Fill content and status of the node
		updateTreeNode(appObject, treeNode);

		// Retrieve subtree for appObject
		Vector<LyFilterElement> treeSons;
		int					    i, n;

		treeSons = getTreeSons(appObject);
		n = ((treeSons == null) ? 0 : treeSons.size());

		// Recurse on subtree sons
		for (i = 0; i < n; i++)
			addTreeNode(treeNode, treeSons.elementAt(i), -1);

		if ((parent != null) && (treeNode != null))
			parent.insertTreeNode(treeNode, position);

		return treeNode;
	} // end method addTreeNode

	/**
	 * Fetch the subtree objects for the specified parent node.
	 *
	 * @param  parentNodeObject parentNodeObject parent of the requested objects.
	 * @return the son list, or null if the subtree is empty
	 * @see    LyTreeBehaviorInterface#getTreeSons
	 */
	public Vector<LyFilterElement> getTreeSons(Object parentNodeObject)
	{
		if (parentNodeObject instanceof LyFilter)
		{
			LyFilter			    parentFilter = (LyFilter)parentNodeObject;
			int					    i, n = parentFilter.getSize();
			Vector<LyFilterElement> resu = new Vector<LyFilterElement>(n);

			// Fill in the list of sub filter items
			for (i = 0; i < n; i++)
				resu.addElement(parentFilter.getElement(i));

			return resu;
		}
		else if (parentNodeObject instanceof LyFilterRelation)
		{
			Vector<LyFilterElement> resu = new Vector<LyFilterElement>(1);

			resu.addElement(((LyFilterRelation)parentNodeObject).getFilter());
			return resu;
		}

		// No child
		return null;
	}

	/**
	 * Invoked when a demand is sent to this listener. Default implementation does nothing.
	 *
	 * @param  controller controller that routed the demand.
	 * @param  demand     demand parameters sent by the view component.
	 * @return Return true if the demand was handled, false otherwise. If false, the demand is
	 *         transmitted to the controller to try to handle it.
	 */
	public boolean handleDemand(LyController controller, LyDemand demand)
	{
		return false;
	}

	/**
	 * Method called to allows to execute specific action in toolbar and tree. Following actions are
	 * coded: Add filter item on filter Add AND or OR filter delete subItem expand node
	 *
	 * @param demand action demand
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		String		source = (demand == null ? "__filter_item" : demand.getSource());
		
		if (source == null)
			return;
		
		Vector<LyFilterElement>	selections = getSelections();
		LyEnvironment 			environment = getEnvironment();
		LySession     			session = getSession();
		LyApplication 			application = session.getApplication();
        LyClassInfo   			classInfo = _classInfo;

		// Add filter item
		if (source.startsWith("__filter_item"))
		{
			if ((_filterController instanceof LyFilterController) && !_filterController.isFreed())
			{
				_filterController.showView();
				return;
			}

			freeFilterController();

			LyAction action = application.getActionWithId(source, classInfo);

			if ((selections.firstElement() instanceof LyFilterExpression)
				|| (selections.firstElement() instanceof LyFilterRelation)
				|| (selections.firstElement() instanceof LyFilterCriterion))
			{
				String message = environment.getMessage("LY_ER_FILTER_ACTION_EXP", action.getName());

				showError(message);
				return;
			}

			// Keep parent filter for endFilter method
			_filterParent = (LyFilter)selections.firstElement();

			// Create a new filter controller to create Expression
			LyFilterRelation filterRelation = getParentFilterRelation(_filterParent);

			if (filterRelation != null)
				classInfo = filterRelation.getClassInfo();

			createFilterController(action, null, classInfo);
		}

		// Add relation filter item
		else if (source.startsWith("__filter_relation_item"))
		{
			if ((_filterController instanceof LyFilterRelationController) && !_filterController.isFreed())
			{
				_filterController.showView();
				return;
			}

			freeFilterController();

			LyAction action = application.getActionWithId(source, classInfo);

			// Check if class info contains at least one relation
			if (!enableRelationCriterion(action))
			{
				String message = environment.getMessage("LY_ER_FILTER_ACTION_RELATION", _classInfo.getName());

				showError(message);
				return;
			}

			if ((selections.firstElement() instanceof LyFilterExpression)
				|| (selections.firstElement() instanceof LyFilterRelation)
				|| (selections.firstElement() instanceof LyFilterCriterion))
			{
				String message = environment.getMessage("LY_ER_FILTER_ACTION_EXP", action.getName());

				showError(message);
				return;
			}

			// Keep parent filter for endFilter method
			_filterParent = (LyFilter)selections.firstElement();

			// Create a new filter controller to create Expression
			LyFilterRelation filterRelation = getParentFilterRelation(_filterParent);

			if (filterRelation != null)
				classInfo = filterRelation.getClassInfo();

			createFilterRelationController(action, null, classInfo);
		}

		// Add criterion filter item
		else if (source.startsWith("__filter_criterion_item"))
		{
			LyAction action = application.getActionWithId(source, classInfo);

			if ((selections.firstElement() instanceof LyFilterExpression)
				|| (selections.firstElement() instanceof LyFilterRelation)
				|| (selections.firstElement() instanceof LyFilterCriterion))
			{
				String message = environment.getMessage("LY_ER_FILTER_ACTION_EXP", action.getName());

				showError(message);
				return;
			}

			// Check if there is some criterion to add
			LyInfoList<LyFilterCriterion> criterions = getCriterionList();

			if ((criterions == null) || (criterions.getSize() == 0))
			{
				String message = environment.getMessage("LY_ER_FILTER_ACTION_CRITERION", _classInfo.getName());

				showError(message);
				return;
			}

			// Keep parent filter for endFilter method
			_filterParent = (LyFilter)selections.firstElement();

			// Create a new filter controller to create Expression
			createFilterCriterionController(action, null, _classInfo);
		}

		// Add filter AND or OR
		else if (source.startsWith("__filter_and")
			|| source.startsWith("__filter_or"))
		{
			if ((selections.firstElement() instanceof LyFilterExpression)
				|| (selections.firstElement() instanceof LyFilterRelation)
				|| (selections.firstElement() instanceof LyFilterCriterion))
			{
				LyAction action = application.getActionWithId(source, classInfo);
				String   message = environment.getMessage("LY_ER_FILTER_ACTION_EXP", action.getName());

				showError(message);
				return;
			}

			LyFilter	  currentFilter = (LyFilter)selections.firstElement();
			LyFilter.Type type = (source.startsWith("__filter_and") ? LyFilter.Type.AND : LyFilter.Type.OR);

			if (currentFilter != null)
			{

				//add the new filter
				LyFilter	   newFilter = new LyFilter(type);
				LyTree		   tree = getTreeComponent();
				LyDependencies old_depends = getDependencies();

				currentFilter.addElement(newFilter);

				//update tree
				tree.setIdle(true);

				// build the new node
				LyTreeNode treeNode = buildTreeNode(newFilter);

				//addDependencies
				old_depends.addDepend(treeNode, newFilter);

				//Retrieve treeNodeParent and add son
				LyTreeNode treeNodeParent = (LyTreeNode)old_depends.getDepends(currentFilter).firstElement();

				treeNodeParent.setOpen(true);
				updateTreeNode(currentFilter, treeNodeParent);

				// update labels and expands
				treeNode.setOpen(true);
				updateTreeNode(newFilter, treeNode);
				treeNodeParent.addTreeNode(treeNode);

				// Update the last modification time.
				setLastModified(new Date().getTime());

				//idle mode
				tree.setIdle(false);
				tree.setVisible(true);
			}
		}

		// Change filter type
		else if (source.startsWith("__filter_change_type"))
		{
			if ((selections.firstElement() instanceof LyFilterExpression)
				|| (selections.firstElement() instanceof LyFilterRelation)
				|| (selections.firstElement() instanceof LyFilterCriterion))
			{
				LyAction action = application.getActionWithId(source, classInfo);
				String   message = environment.getMessage("LY_ER_FILTER_ACTION_EXP", action.getName());

				showError(message);
				return;
			}

			LyFilter currentFilter = (LyFilter)selections.firstElement();

			if (currentFilter != null)
			{

				//Retrieve treeNode
				LyDependencies depends = getDependencies();
				LyTreeNode     treeNode = (LyTreeNode)depends.getDepends(currentFilter).firstElement();
				LyTree		   tree = getTreeComponent();
				LyFilter.Type  type = currentFilter.getType();

				tree.setIdle(true);
				if (type == LyFilter.Type.AND)
					type = LyFilter.Type.OR;
				else
					type = LyFilter.Type.AND;

				currentFilter.setType(type);
				updateTreeNode(currentFilter, treeNode);

				//idle mode
				tree.setIdle(false);
				tree.setVisible(true);
			}
		}

		// Deletion of an item
		else if (source.startsWith("__filter_delete"))
		{
			if (selections.firstElement() == _filter)
			{
				LyAction action = application.getActionWithId(source, classInfo);
				String   message = environment.getMessage("LY_ER_FILTER_ACTION_ROOT", action.getName());

				showError(message);
			}

			deleteSelectedItem();
		}

		// Set an item
		else if (source.startsWith("__filter_set_expression"))
		{
			LyAction action = application.getActionWithId(source, classInfo);

			editFilterElement(selections.firstElement(), action);
		}

		// Cut selection
		else if (source.startsWith("__filter_cut"))
		{
			if (selections.firstElement() == _filter)
			{
				LyAction action = application.getActionWithId(source, classInfo);
				String   message = environment.getMessage("LY_ER_FILTER_ACTION_ROOT", action.getName());

				showError(message);
				return;
			}

			copyFilter(true);
		}

		// Copy selection
		else if (source.startsWith("__filter_copy"))
		{
			if (selections.firstElement() == _filter)
			{
				LyAction action = application.getActionWithId(source, classInfo);
				String   message = environment.getMessage("LY_ER_FILTER_ACTION_ROOT", action.getName());

				showError(message);
				return;
			}

			copyFilter(false);
		}

		// Paste selection
		else if (source.startsWith("__filter_paste"))
		{
			if (_bufferFilter == null)
			{
				showError("LY_ER_FILTER_NO_BUFFER");
				return;
			}

			if ((selections.firstElement() instanceof LyFilterExpression)
				|| (selections.firstElement() instanceof LyFilterRelation)
				|| (selections.firstElement() instanceof LyFilterCriterion))
			{
				LyAction action = application.getActionWithId(source, classInfo);
				String   message = environment.getMessage("LY_ER_FILTER_ACTION_EXP", action.getName());

				showError(message);
				return;
			}

			LyFilter currentFilter = (LyFilter)selections.firstElement();

			if (currentFilter != null)
			{

				// Check if paste is allowed
				if (checkPaste(currentFilter, _bufferFilter))
					paste(currentFilter);
			}
		}
		else if (source.equals(LyFormController.DEMAND_VALIDATE)
				|| source.equals(LyFormController.DEMAND_APPLY))
		{
			LyFilter errorFilter = checkFilter(_filter, true);

			if (errorFilter != null)
			{
				showError("LY_ER_FILTER_CHECK");
				return;
			}

			boolean  validate = source.equals(LyFormController.DEMAND_VALIDATE);
			LyFilter filter;

			if (validate)
			{
				filter = _filter;

				// Before setting the filter, simplify it by removing filters that only contains one filter
				simplifyFilter(filter);
			}
			else

				// Filter on client must be different
				filter = _filter.cloneFilter();

			if (_filterEndInterface != null)
			{
				_filterEndInterface.endFilter(this, source, filter);

				boolean	updateFilter = true;

				if (_filterEndInterface instanceof LyPanelController)
					updateFilter = !((LyPanelController)_filterEndInterface).confirmNotPhysicalFilter(filter);

				if (updateFilter)
					filter = _filterEndInterface.getFilter();
			}
			else
			{
				LyAreaController areaController;
				LyAction		 action;
				int				 i, n = _action.getActionCount();

				for (i = 0; i < n; i++)
				{
					action = _action.getAction(i);
					areaController = (LyAreaController)session.invokeAction(_parent.getParent(), action, getClassInfo(),
							null, false);
					if (areaController instanceof LyFilterEndInterface)
					{
						((LyFilterEndInterface)areaController).endFilter(this, source, filter);
					    areaController.showView();
					}
					else
					    areaController.free(_parent.getParent());
				}
			}

			if (validate)
			{
				if (!(_parent instanceof LyCompoundFilterController))
					free();
			}
			else if (_filter != filter)
			{
				_filter = ((filter == null) ? new LyFilter() : filter.cloneFilter(session));
				reset(true);
			}
		}
		else if (source.equals(LyFormController.DEMAND_CANCEL))
		{
			if (_hasViewController)
			{
				if (_filterEndInterface != null)
					_filterEndInterface.endFilter(this, source, null);

				free();
			}
			else
			{
				if (_filterEndInterface != null)
					_filterEndInterface.endFilter(this, DEMAND_VALIDATE, null);

				hideView();
				free(_parent);
			}
		}
		else if (demand.getType() == LyActionDemand.TYPE_EXPAND)
		{
			String value = demand.getValue();

			if (value != null)
			{
				String[] array = LyString.stringToList(value, LyActionDemand.SEPARATOR);

				if ((array != null) && (array.length <= 2))
				{
					String  key = array[0];
					Boolean bool = null;

					if (array.length == 2)
						bool = Boolean.valueOf(array[1]);

					LyTree		   tree = getTreeComponent();
					LyDependencies old_depends = getDependencies();

					//we refresh all (open status is changed by Canvas in WIN)
					tree.setIdle(true);

					// We refresh impacted node
					// search it
					LyTreeNode changeNode = null;

					for (Enumeration<Object> e = old_depends.keys(); e.hasMoreElements();)
					{
						Object obj = e.nextElement();

						if ((obj instanceof LyTreeNode) && ((LyTreeNode)obj).getId().equals(key))
						{
							changeNode = (LyTreeNode)obj;
							break;
						}
					}

					if (changeNode == null)
					{
						tree.setIdle(false);
						return;
					}

					// change expand value
					changeNode.setOpen(((bool == null) ? (!changeNode.isOpen()) : bool.booleanValue()));

					Vector<Object>	depends;

					LyFilterElement element;

					depends = old_depends.getDepends(changeNode);
					if (depends != null)
					{
						element = (LyFilterElement)depends.firstElement();
						updateTreeNode(element, changeNode);
						LyFilter   parentFilter = getFilterParent(element, null);
						LyTreeNode parentNode;

						depends = old_depends.getDepends(parentFilter);
						if (depends != null)
						{
							parentNode = (LyTreeNode)depends.firstElement();
							updateTreeNode(parentFilter, parentNode);
						}
					}

					tree.setIdle(false);
				}
			}

			setLastModified(new Date().getTime());
		}
		else
		{
			super.actionDemand(demand);
			return;
		}

		updateTools();
		setLastModified(new Date().getTime());
	} // end method actionDemand

	/**
	 * Invoked when a Drag & Drop occurs. (LyDemandListener interface) This method handles: opening
	 * and closing line, moving an interval, action
	 *
	 * @param demand drag-drop demand
	 */
	public void dragDropDemand(LyDragDropDemand demand)
	{
		String		    demandId = demand.getSource();
		LyFilterElement element = findObject(_filter, demandId);

		if (element == null)
			return;

		int type = demand.getType();

		switch (type)
		{

		case LyDragDropDemand.TYPE_DRAG_START:

			// Root filter may not be dragged
			if (element == _filter)
				element = null;
			else
			{
				LyFilterRelation filterRelation = getParentFilterRelation(element);

				// Internal filter of relation filter may not be dragged
				if ((element instanceof LyFilter) && (filterRelation != null)
					&& (filterRelation.getFilter() == element))
					element = null;
			}

			_bufferFilter = element;
			break;

		case LyDragDropDemand.TYPE_DROP:
			if (_bufferFilter == null)
				return;

			// When a drop occurs on a node, first check if dragged element is not dropped on
			// on of its child. If not, paste dragged element.
			LyDependencies dependencies = getDependencies();
			LyTreeNode     dropTreeNode = (LyTreeNode)dependencies.getDepends(element).firstElement();
			LyTreeNode     dragTreeNode = (LyTreeNode)dependencies.getDepends(_bufferFilter).firstElement();
			LyComponent    parent = dropTreeNode.getParent();
			boolean		   isAncestor = (dragTreeNode == dropTreeNode);

			while (!isAncestor && !(parent instanceof LyTree))
			{
				if (parent == dragTreeNode)
					isAncestor = true;

				parent = parent.getParent();
			}

			// Check if paste is allowed
			boolean allowPaste = !isAncestor && checkPaste(element, _bufferFilter);

			if (allowPaste)
			{
				copyFilter(true);
				paste(element);
			}

			break;

		default:
			super.dragDropDemand(demand);
		}
	} // end method dragDropDemand

	/**
	 * Checks if an element may be pasted on a parent element.
	 *
	 * @param  parent  the parent element.
	 * @param  element the element that will be pasted if allowed.
	 * @return true if element may be pasted on parent element, false otherwise.
	 */
	protected boolean checkPaste(LyFilterElement parent, LyFilterElement element)
	{

		// Paste is not allowed on a relation filter
		if (parent instanceof LyFilterRelation)
			return false;

		LyFilterRelation filterRelation = getParentFilterRelation(element);
		LyFilterRelation parentFilterRelation = getParentFilterRelation(parent);
		LyClassInfo		 classInfo = ((filterRelation == null) ? _classInfo : filterRelation.getClassInfo());
		LyClassInfo		 parentClassInfo = ((parentFilterRelation == null) ? _classInfo
				: parentFilterRelation.getClassInfo());
		boolean			 result = (classInfo == parentClassInfo);

		return result;
	}

	/**
	 * Invoked when an menu item notification is sent to this listener.
	 * @param demand the menu demand
	 */
	public void menuDemand(LyMenuDemand demand)
	{
		LyTree tree = getTreeComponent();

		if (tree.getId().equals(demand.getSource()))
		{
			LyPopupMenu			 menu = new LyPopupMenu();
			LyItem				 item;
			LyInfoList<LyAction> actions = _action.getActions();
			int					 i, n = actions.getSize();
			LyAction			 action;

			for (i = 0; i < n; i++)
			{
				action = actions.getInfo(i);
				if (!enableAction(action))
					continue;

				item = new LyItem(action.getId(), action.getName());
				item.setController(getId());
				item.setDialog(true);
				item.setImage(getImageFromPrefix(action.getFilePrefix(), null, 12));
				menu.addItem(item);
			}

			// Coordinates
			int[] p = demand.getPosition();

			menu.setX(p[0]);
			menu.setY(p[1]);
			menu.setController(getId());
			menu.setEnvironment(getEnvironment());
			tree.setPopupMenu(menu);
		}
		else
			super.menuDemand(demand);
	} // end method menuDemand

	/**
	 * Method called when there is a selection in tree. It will update toolbar and change the aspect
	 * of the selected node
	 *
	 * @param demand LySelectionDemand selection demand
	 * @see   LyController#selectionDemand
	 */
	public void selectionDemand(LySelectionDemand demand)
	{

		// Check if modifications have been performed on form. If so, ask the user
		// whether modifications must be cancelled or not.
		boolean modifiedData = false;

		if (_filterController != null)
			modifiedData = _filterController.checkIfModified();

		if (modifiedData)
		{
			String message = getEnvironment().getMessage("LY_CONFIRM_FILTER_ABORT");

			getViewManager().askQuestion(getComponent(), message, this, demand);
		}
		else
			updateSelection(demand, false);
	}

	/**
	 * Update the tree selection with the given selection demand.
	 *
	 * @param demand            The selection demand.
	 * @param resetOldSelection Indicates if old selection must be set after selection demand.
	 *                          Needed when user cancels a new selection demand.
	 */
	protected void updateSelection(LySelectionDemand demand, boolean resetOldSelection)
	{
		Vector<LyFilterElement> v = new Vector<LyFilterElement>();
		LyTree				    tree = getTreeComponent();
		LyDependencies		    depends = getDependencies();
		int					    i, n = demand.getSelectionSize();
		String				    id;
		LyFilter			    filt;
		LyFilterElement		    elt;
		LyTreeNode			    changeNode;
		Object				    obj;

		tree.setIdle(true);
		if (n == 0)
		{

			// If no selection, remove any selection from the tree
			Vector<LyFilterElement>	selection = getSelections();
			Object		   			object;
			LyTreeNode     			treeNode;
			LyDependencies 			dependencies = getDependencies();
			int			   			nSelection = ((selection == null) ? 0 : selection.size());

			for (i = 0; i < nSelection; i++)
			{
				object = selection.elementAt(i);
				treeNode = (LyTreeNode)dependencies.getDepends(object).firstElement();
				treeNode.setSelected(false);
			}
		}

		for (i = 0; i < n; i++)
		{
			id = demand.getSelectionId(i);
			filt = _filter;
			elt = findObject(filt, id);
			v.addElement(elt);
			for (Enumeration<Object> e = depends.keys(); e.hasMoreElements();)
			{
				obj = e.nextElement();
				if (obj instanceof LyTreeNode)
				{
					changeNode = (LyTreeNode)obj;
					changeNode.setSelected(resetOldSelection ? false : changeNode.getId().equals(id));
				}
			}
		}

		if (resetOldSelection)
		{
			Vector<LyFilterElement>	selection = getSelections();
			Object		   			object;
			LyTreeNode     			treeNode;
			LyDependencies 			dependencies = getDependencies();

			n = ((selection == null) ? 0 : selection.size());
			for (i = 0; i < n; i++)
			{
				object = selection.elementAt(i);
				treeNode = (LyTreeNode)dependencies.getDepends(object).firstElement();
				treeNode.setSelected(true);
			}

			tree.setIdle(false);
			return;
		}

		setSelections(v);
		updateTools();

		// Update the last modification time.
		setLastModified(new Date().getTime());
		tree.setIdle(false);
	} // end method updateSelection

	/**
	 * Answer to a askQuestion request.
	 *
	 * @param result true if the user has clicked on Yes or Ok
	 * @param clientData data specified in the askQuestion method
	 * @see   LyQuestionAnswerInterface
	 */
	public void questionAnswer(boolean result, Object clientData)
	{
		updateSelection((LySelectionDemand)clientData, !result);
	}

	/**
	 * Edit given filter element.
	 * @param element the element to edit
	 * @param action the action to use
	 */
	protected void editFilterElement(LyFilterElement element, LyAction action)
	{
		if (element == null)
			return;

		LyEnvironment environment = getEnvironment();

		if (element instanceof LyFilter)
		{
			if (element == _filter)
			{
				String message = environment.getMessage("LY_ER_FILTER_ACTION_ROOT", action.getName());

				showError(message);
			}
			else
			{
				String message = environment.getMessage("LY_ER_FILTER_ACTION_FILTER", action.getName());

				showError(message);
			}

			return;
		}
		else if (element instanceof LyFilterCriterion)
		{
			String message = environment.getMessage("LY_ER_FILTER_ACTION_FILTER", action.getName());

			showError(message);
			return;
		}

		freeFilterController();

		// Create a new filter controller to set Expression
		Vector<LyFilterElement>	selections = getSelections();
		LyFilterElement 		currentFilter = selections.firstElement();

		// Create a temporary filter used by the filter controller
		// (a call to getFilter() is done by the new controller)
		_filterParent = new LyFilter(CURRENT_FILTER_EXPRESSION, null, LyFilter.Type.AND);
		_filterParent.addElement(currentFilter);
		LyClassInfo		 classInfo = _classInfo;
		LyFilterRelation filterRelation = getParentFilterRelation(currentFilter);

		if (filterRelation != null)
			classInfo = filterRelation.getClassInfo();

		if (currentFilter instanceof LyFilterExpression)
			_filterController = createFilterController(action, (LyFilterExpression)currentFilter, classInfo);
		else if (currentFilter instanceof LyFilterRelation)
			_filterController = createFilterRelationController(action, (LyFilterRelation)currentFilter, classInfo);
	} // end method editFilterElement

	/**
	 * Create a filter controller based on the arguments
	 * @param action an action
	 * @param filterExpression a filter expression
	 * @param classInfo a classInfo
	 * @return a LyFilterController
	 */
	protected LyFilterController createFilterController(LyAction action, LyFilterExpression filterExpression,
		LyClassInfo classInfo)
	{
		_filterController = new LyFilterController(getSession(), this, action, classInfo, null);
		if (_parent instanceof LyCompoundFilterController)
			((LyCompoundFilterController)_parent).setFilterController(_filterController);
		else
		{
			_filterController.setFilterEndInterface(this);
			if (_filterController.hasViewController())
				openSubAreaController(_filterController, true);

			_filterController.showView();
		}

		return (LyFilterController)_filterController;
	}

	/**
	 * Create a filter relation controller based on the given arguments
	 * @param action an action
	 * @param filterRelation the filter relation to edit
	 * @param classInfo a classInfo
	 * @return a filter relation controller
	 */
	protected LyFilterRelationController createFilterRelationController(LyAction action,
		LyFilterRelation filterRelation, LyClassInfo classInfo)
	{
		_filterController = new LyFilterRelationController(getSession(), this, action, classInfo, null);
		((LyFilterRelationController)_filterController).setDisplayFilterRelation(false);
		if (_parent instanceof LyCompoundFilterController)
			((LyCompoundFilterController)_parent).setFilterController(_filterController);
		else
		{
			_filterController.setFilterEndInterface(this);
			if (_filterController.hasViewController())
				openSubAreaController(_filterController, true);

			_filterController.showView();
		}

		if (filterRelation == null)
			return (LyFilterRelationController)_filterController;

		LyClassInfo selectedClassInfo = filterRelation.getClassInfo();

		if (selectedClassInfo != null)
			((LyFilterRelationController)_filterController).setClassInfo(selectedClassInfo);

		// Do not allow the modification of a relation filter containing filter expressions.
		LyFilter subFilter = filterRelation.getFilter();

		if ((subFilter != null) && (subFilter.getElements() != null))
			((LyFilterRelationController)_filterController).setFormEditable(false);

		return (LyFilterRelationController)_filterController;
	}

	/**
	 * Create a filter criterion controller based on the given arguments
	 * @param action an actoion
	 * @param filterCriterion the filter criterion to edit
	 * @param classInfo a classInfo
	 * @return a filter criterion controller
	 */
	protected LyFilterCriterionController createFilterCriterionController(LyAction action,
		LyFilterCriterion filterCriterion, LyClassInfo classInfo)
	{
		_filterController = new LyFilterCriterionController(getSession(), this, action, classInfo, null);
		if (_parent instanceof LyCompoundFilterController)
			((LyCompoundFilterController)_parent).setFilterController(_filterController);
		else
		{
			_filterController.setFilterEndInterface(this);
			openSubAreaController(_filterController, true);
			_filterController.showView();
		}

		return (LyFilterCriterionController)_filterController;
	}

	/**
	 * Free the filter controller
	 */
	protected void freeFilterController()
	{
		if (_parent instanceof LyCompoundFilterController)
			return;

		if (_filterController != null)
		{
			_filterController.free();
			_filterController = null;
		}
	}

	/**
	 * Returns the first parent filter relation of given filter element.
	 *
	 * @param  element the filter element whose parent must be retrieved.
	 * @return the first parent filter relation of given filter element, or null if not found.
	 */
	protected LyFilterRelation getParentFilterRelation(Object element)
	{
		if (element == null)
			return null;

		LyDependencies depends = getDependencies();
		Vector<Object> dependency = depends.getDepends(element);

		if (dependency == null)
			return null;

		LyTreeNode  treeNode = (LyTreeNode)dependency.firstElement();
		LyComponent treeNodeParent = treeNode.getParent();

		if (!(treeNodeParent instanceof LyTreeNode))
			return null;

		Object parentObject = ((treeNodeParent instanceof LyTreeNode)
				? depends.getDepends(treeNodeParent).firstElement() : null);

		if (parentObject instanceof LyFilterRelation)
			return (LyFilterRelation)parentObject;

		return getParentFilterRelation(parentObject);
	}

	/**
	 * Reset the tool bar status (depends on current selection).
	 */
	public void updateTools()
	{
		if (isIdle())
			return;

		LyAction			 action;
		LyComponent			 comp;
		LyTool				 tool = null;
		LyToolBar			 toolBar = (LyToolBar)getComponent().getComponent("_toolBar");
		boolean				 enable;
		LyInfoList<LyAction> actions = _action.getActions();
		int					 i, n = actions.getSize();

		if (toolBar == null)
			return;

		for (i = 0; i < n; i++)
		{
			action = actions.getInfo(i);
			enable = enableAction(action);
			comp = toolBar.getComponent(action.getId());
			if (comp instanceof LyTool)
			{
				tool = (LyTool)comp;
				if (enable != tool.isEnabled())
					tool.setEnabled(enable);
			}
		}
	} // end method updateTools

	/**
	 * Method that checks if the given action is present or not.
	 * @param action the action to validate
	 * @return the status
	 */
	public boolean validateAction(LyAction action)
	{
		LySession session = getSession();

		if (!session.validateAction(action, this))
			return false;

		if (action.getId().equals("__filter_relation_item") && !enableRelationCriterion(action))
			return false;

		if (action.getId().equals("__filter_criterion_item") && !enableCriterion())
			return false;

		return true;
	}

	/**
	 * Checks if an action is allowed for current selection.
	 *
	 * @param  action the action that must be checked.
	 * @return true if the action is allowed, false otherwise.
	 */
	protected boolean enableAction(LyAction action)
	{
		if (!validateAction(action))
			return false;

		LySession session = getSession();

		if (!session.enableAction(action, this, null, false))
			return false;

		if (action.getId().equals("__filter_relation_item")
			&& (_filterController instanceof LyFilterRelationController))
			return false;

		if (action.getId().equals("__filter_item") && (_filterController instanceof LyFilterController))
			return false;

		Vector<LyFilterElement> selections = getSelections();
		int					    size = ((selections == null) ? 0 : selections.size());
		boolean				    enable = false;

		// Check with min max value
		if ((size > 0) && (size >= action.getMin()) && (size <= action.getMax()))
			enable = true;

		// Check specific action
		if (enable && (size > 0))
		{
			Object firstSelection = selections.firstElement();

			if (firstSelection instanceof LyFilterCriterion)
			{

				// Only cut, copy and delete for criterion
				if (!action.getId().startsWith("__filter_delete") && !action.getId().startsWith("__filter_cut")
					&& !action.getId().startsWith("__filter_copy"))
					enable = false;
			}

			// By default a leaf has no action except deletion.
			if ((firstSelection instanceof LyFilterExpression) || (firstSelection instanceof LyFilterRelation))
			{
				if (!action.getId().startsWith("__filter_delete")
					&& !action.getId().startsWith("__filter_set_expression")
					&& !action.getId().startsWith("__filter_cut") && !action.getId().startsWith("__filter_copy"))
					enable = false;
			}

			// We can not delete, copy, cut root object.
			if ((firstSelection == _filter)
				&& (action.getId().startsWith("__filter_delete") || action.getId().startsWith("__filter_copy")
					|| action.getId().startsWith("__filter_cut")))
				enable = false;

			// We can not delete, copy, cut internal filter for relation filter.
			LyFilterRelation filterRelation = getParentFilterRelation(firstSelection);

			if ((firstSelection instanceof LyFilter) && (filterRelation != null)
				&& (filterRelation.getFilter() == firstSelection)
				&& (action.getId().startsWith("__filter_delete") || action.getId().startsWith("__filter_copy")
					|| action.getId().startsWith("__filter_cut")))
				enable = false;

			// Only LyFilterExpression and LyFilterRelation can be set
			if (!(firstSelection instanceof LyFilterExpression) && !(firstSelection instanceof LyFilterRelation)
				&& action.getId().startsWith("__filter_set_expression"))
				enable = false;

			// If buffer is null, paste is not allowed
			if ((_bufferFilter == null) && (action.getId().equals("__filter_paste")))
				enable = false;
		}

		return enable;
	} // end method enableAction

	/**
	 * Utility method that check if there are some criterion available.
	 *
	 * @return true if there are predifined criterion, false otherwise.
	 */
	protected boolean enableCriterion()
	{
		if ((getCriterionList() == null) || (getCriterionList().getSize() == 0))
			return false;

		return true;
	}

	/**
	 * Utility method that check if we can have a relation filter for the current class.
	 *
	 * @param  action the relation criterion filter action
	 * @return true if we can have a relation filter
	 */
	protected boolean enableRelationCriterion(LyAction action)
	{
		if (!_parent.getEnvironment().testResource("LY_RELATION_FILTER"))
			return false;

		// Retrieve class info concerned by selection
		Vector<LyFilterElement>	selections = getSelections();
		LyClassInfo		 		classInfo = _classInfo;
		LyFilterElement  		currentFilter = (((selections != null) && (selections.size() > 0))
				? (LyFilterElement)selections.firstElement() : null);
		LyFilterRelation 		filterRelation = getParentFilterRelation(currentFilter);

		if (filterRelation != null)
			classInfo = filterRelation.getClassInfo();

		// Check if class info contains at least one relation
		LyBitSet			    fieldMarks = (LyBitSet)action.getParameter("_fieldMarks");
		LyInfoList<LyFieldInfo> fields = classInfo.getFieldInfos(fieldMarks.getBitSet());

		if ((fields == null) || (fields.getSize() == 0))
			fields = classInfo.getFieldInfos();

		LyFieldInfo field;
		int		    i, n = fields.getSize();

		for (i = 0; i < n; i++)
		{
			field = fields.getInfo(i);
			if (field instanceof LyRelationFieldInfo)
				return true;
		}

		return false;
	} // end method enableRelationCriterion

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyActionBehaviorInterface.class;
	}

	/**
	 * Finds an object by identifier Returned item may be null, a LyFilter or a LyFilterElement.
	 *
	 * @param  filter root element for the search
	 * @param  id     identifier of the search element.
	 * @return the filter element if found, null otherwise
	 */
	protected LyFilterElement findObject(LyFilter filter, String id)
	{
		if ((filter.getId() != null) && (filter.getId().equals(id)))
			return filter;
		else
		{
			String id2 = filter.getClass() + "@" + filter.hashCode();

			if (id2.equals(id))
				return filter;
		}

		Vector<LyFilterElement> sons = getTreeSons(filter);
		int					    i, n = sons.size();
		Object				    obj;
		String				    id2;

		for (i = 0; i < n; i++)
		{
			obj = sons.elementAt(i);
			id2 = obj.getClass() + "@" + obj.hashCode();
			if (id2.equals(id))
				return ((LyFilterElement)obj);
			else if (obj instanceof LyFilter)
			{
				LyFilterElement res = findObject(((LyFilter)obj), id);

				if (res != null)
					return res;
			}
			else if (obj instanceof LyFilterRelation)
			{
				LyFilterRelation relation = (LyFilterRelation)obj;
				LyFilter		 subFilter = relation.getFilter();
				LyFilterElement  res = findObject(subFilter, id);

				if (res != null)
					return res;
			}
		}

		return null;
	} // end method findObject

	/**
	 * Allows to get the parent filter of an expression or a filter.
	 *
	 * @param  current The filter element son.
	 * @param  begin   the start branch (may be null, starts at root)
	 * @return the Filter parent or null if current is root filter.
	 */
	protected LyFilter getFilterParent(Object current, LyFilter begin)
	{
		if (current == null)
			return null;

		// case from root
		if (begin == null)
		{
			LyFilter root = _filter;

			if (root == current)
				return null;

			begin = root;
		}

		// case from other filter
		Object element;

		for (int jj = 0; jj < begin.getSize(); jj++)
		{
			element = begin.getElement(jj);
			if (element == current)
				return begin;

			// if son is not a leaf, we see on its sons.
			if (element instanceof LyFilter)
			{
				LyFilter result = getFilterParent(current, (LyFilter)element);

				if (result != null)
					return result;
			}
			else if (element instanceof LyFilterRelation)
			{
				LyFilter result = getFilterParent(current, ((LyFilterRelation)element).getFilter());

				if (result != null)
					return result;
			}
		}

		return null;
	} // end method getFilterParent

	/**
	 * Builds the displayed label of a LyFilterExpression.
	 *
	 * @param  exp         the LyFilterExpression.
	 * @param  environment the environment used to translate strings.
	 * @return Label to display.
	 */
	public static String getNameExpression(LyFilterExpression exp, LyEnvironment environment)
	{
		String	result = exp.toString(true);

		if (environment != null)
			return environment.translate(result);

		return result;
	} // end method getNameExpression

	/**
	 * Builds the displayed label of a LyFilter. The label may be different depending on the status
	 * (expanded / collapsed) of the filter in the tree. (Collapse may show the content for
	 * instance).
	 *
	 * @param  filter the filter
	 * @param environment current environment
	 * @param  expand indicates if tree node associated to the filter is expanded or not
	 * @return Label which is displayed
	 */
	public static String getNameFilter(LyFilter filter, LyEnvironment environment, boolean expand)
	{
		StringBuffer    resu = new StringBuffer();
		LyFilterElement element;
		int			    i, n = filter.getSize();
		boolean		    type_and = (filter.getType() == LyFilter.Type.AND);

		if ((expand) || (n == 0))
		{

			//search name of field info
			String label = (type_and ? "LY_FILTER_AND" : "LY_FILTER_OR");

			if (environment != null)
				label = environment.translate(label);

			return label;
		}

		String filterAndSep = "LY_AND_FILTER_SEP";
		String filterOrSep = "LY_OR_FILTER_SEP";

		if (environment != null)
		{
			filterAndSep = environment.translate(filterAndSep);
			filterOrSep = environment.translate(filterOrSep);
		}

		resu.append('(');

		// Go through the element list
		for (i = 0; i < n; i++)
		{
			element = filter.getElement(i);
			int nb_sub_filter = -1;

			if (element instanceof LyFilter)
				nb_sub_filter = ((LyFilter)element).getSize();

			// Add operator
			if ((i != 0) && (nb_sub_filter != 0))
			{
				if (type_and)
				{
					resu.append(" ");
					resu.append(filterAndSep);
					resu.append(" ");
				}
				else
				{
					resu.append(" ");
					resu.append(filterOrSep);
					resu.append(" ");
				}
			}

			if ((n > 1) && !(element instanceof LyFilter))
				resu.append('(');

			if (element instanceof LyFilter)
			{
				if (nb_sub_filter != 0)
					resu.append(getNameFilter((LyFilter)element, environment, false));
			}
			else if (element instanceof LyFilterExpression)
				resu.append(environment.translate(getNameExpression((LyFilterExpression)element, environment)));
			else if (element instanceof LyFilterRelation)
				resu.append(environment.translate(
						getNameFilterRelation((LyFilterRelation)element, environment, false)));
			else if (element instanceof LyFilterCriterion)
				resu.append(environment.translate(getNameFilterCriterion((LyFilterCriterion)element, environment)));

			if ((n > 1) && !(element instanceof LyFilter))
				resu.append(')');
		}

		resu.append(')');
		return resu.toString();
	} // end method getNameFilter

	/**
	 * Builds the displayed label of a LyFilterRelation.
	 *
	 * @param  filter      the filter.
	 * @param  environment the environment used to translate strings.
	 * @return Label which is displayed.
	 */
	public static String getNameFilterRelation(LyFilterRelation filter, LyEnvironment environment)
	{
		return LyFilterEditorController.getNameFilterRelation(filter, environment, false);
	}

	/**
	 * Builds the displayed label of a LyFilterRelation.
	 *
	 * @param  filter      the filter.
	 * @param  environment the environment used to translate strings.
	 * @param  expand      indicates if tree node associated to the filter is expanded or not.
	 * @return Label which is displayed
	 */
	public static String getNameFilterRelation(LyFilterRelation filter, LyEnvironment environment, boolean expand)
	{
		StringBuffer result = new StringBuffer();
		LyFieldInfo  fieldInfo = filter.getRelationFieldInfo();

		// Search name of field info
		String id = null;

		if (environment != null)
			id = environment.translate(fieldInfo.getAlias());

		if (id == null)
			return null;

		if (environment != null)
		{
			result.append(environment.translate("LY_FILTER_ON_FIELD"));
			result.append(" ");
		}

		// add Attribute info
		result.append(id);
		LyClassInfo relatedClass = filter.getClassInfo();
		int		    n = filter.getRelationFieldInfo().getClassInfoCount();

		if ((n > 1) && (environment != null) && (relatedClass != null))
		{
			result.append(" ");
			result.append("LY_FILTER_ON_CLASS");
			result.append(" ");
			result.append(environment.translate(relatedClass.getName()));
		}

		if (!expand)
		{

			// Add filter
			result.append(" ");
			LyFilter subFilter = filter.getFilter();

			if ((subFilter.getName() == null) || (subFilter.getName().startsWith("_")))
				result.append(getNameFilter(subFilter, environment, false));
			else
				result.append("\"" + environment.translate(subFilter.getName()) + "\"");
		}

		return result.toString();
	} // end method getNameFilterRelation

	/**
	 * Builds the displayed label of a LyFilterCriterion.
	 *
	 * @param  criterion   the criterion.
	 * @param  environment the environment used to translate strings.
	 * @return Label which is displayed.
	 */
	public static String getNameFilterCriterion(LyFilterCriterion criterion, LyEnvironment environment)
	{
		return environment.translate(criterion.getName());
	}

	/**
	 * Sets the list of selected items. Internal use only.
	 *
	 * @param a list of selected nodes (tree nodes) or new Vector() if selection is empty.
	 */
	void setSelections(Vector<LyFilterElement> selections)
	{
		_selections = selections;

		updateTools();

		LyAction	setExpAction = getApplication().getActionWithId(LyAction.ID_SET_FILTER, _classInfo);

		if ((setExpAction.getTarget() != null) && (_selections.size() == 1))
		{
			LyFilterElement element = _selections.firstElement();

			if ((element instanceof LyFilterExpression) || (element instanceof LyFilterRelation))
				editFilterElement(element, setExpAction);
			else
				resetFilterController(element);
		}
	}

	/**
	 * Copy current selection in buffer filter.
	 *
	 * @param cut Indicates if selected node should be cut after copy.
	 */
	protected void copyFilter(boolean cut)
	{
		LyTree				    tree = getTreeComponent();
		Vector<LyFilterElement> selections = getSelections();
		Object				    template = selections.firstElement();

		if (template instanceof LyFilter)
			_bufferFilter = ((LyFilter)template);
		else if (template instanceof LyFilterExpression)
			_bufferFilter = ((LyFilterExpression)template);
		else if (template instanceof LyFilterRelation)
			_bufferFilter = ((LyFilterRelation)template);
		else if (template instanceof LyFilterCriterion)
			_bufferFilter = ((LyFilterCriterion)template);

		if (cut)
			deleteSelectedItem();

		updateTools();
		tree.setVisible(true);
	}

	/**
	 * Paste to selection.
	 * @param filterElement the target element
	 */
	protected void paste(LyFilterElement filterElement)
	{
		LyFilter filter = null;
		int		 index = -1;

		if (filterElement instanceof LyFilter)
			filter = (LyFilter)filterElement;
		else
		{
			filter = getFilterParent(filterElement, null);

			if (filter == null)
				return;

			index = filter.getIndex(filterElement);
		}

		// Add the new filter
		LyFilterElement newFilter = null;

		if (_bufferFilter instanceof LyFilter)
			newFilter = ((LyFilter)_bufferFilter).cloneFilter();
		else if (_bufferFilter instanceof LyFilterExpression)
			newFilter = ((LyFilterExpression)_bufferFilter).cloneFilterExpression();
		else if (_bufferFilter instanceof LyFilterRelation)
			newFilter = ((LyFilterRelation)_bufferFilter).cloneFilterRelation();
		else if (_bufferFilter instanceof LyFilterCriterion)
			newFilter = ((LyFilterCriterion)_bufferFilter).cloneFilterCriterion();

		if (index >= 0)
			filter.addElement(newFilter, index);
		else
			filter.addElement(newFilter);

		//update tree
		LyTree tree = getTreeComponent();

		tree.setIdle(true);

		//Retrieve treeNodeParent and add son
		LyDependencies old_depends = getDependencies();
		LyTreeNode     treeNodeParent = (LyTreeNode)old_depends.getDepends(filter).firstElement();
		LyTreeNode     treeNode = addTreeNode(treeNodeParent, newFilter, index);

		// update labels and expands
		treeNodeParent.setOpen(true);
		updateTreeNode(filter, treeNodeParent);
		treeNode.setOpen(true);
		updateTreeNode(newFilter, treeNode);

		// Update the last modification time.
		setLastModified(new Date().getTime());

		//idle mode
		tree.setIdle(false);
		tree.setVisible(true);
	} // end method paste

	/**
	 * Delete the selected item in the tree.
	 */
	protected void deleteSelectedItem()
	{
		Object   currentFilter = getSelections().firstElement();
		LyFilter parent = getFilterParent(currentFilter, null);

		if (parent != null)
		{

			// Retrieve internal tree component
			LyTree tree = getTreeComponent();

			tree.setIdle(true);

			// Remove filter
			parent.removeElement((LyFilterElement)currentFilter);

			// Remove tree node
			LyDependencies depends = getDependencies();
			LyTreeNode     treeNode = (LyTreeNode)depends.getDepends(currentFilter).firstElement();
			LyTreeNode     treeNodeParent = (LyTreeNode)treeNode.getParent();

			depends.removeObject(currentFilter);

			// TreeNode is destroyed -> not selected
			treeNode.setSelected(false);
			treeNodeParent.removeTreeNode(treeNode);
			depends.removeObject(treeNode);

			// Remove all son nodes...
			LyTreeNode	    son;
			LyFilterElement subFilter;

			for (int i = 0; i < treeNode.getTreeNodeCount(); i++)
			{
				son = treeNode.getTreeNode(i);
				subFilter = (LyFilterElement)depends.getDepends(son).firstElement();
				depends.removeObject(subFilter);
				depends.removeObject(son);
			}

			treeNode = null;

			//idle mode
			tree.setIdle(false);

			//tree.setVisible(true);
			setSelections(new Vector<LyFilterElement>());
			resetFilterController(null);
		}
	} // end method deleteSelectedItem

	/**
	 * Checks a filter. Checks recursively that every filter has at least one element.
	 *
	 * @param  filter    The filter that must be checked
	 * @param  showError Indicates whether the first tree node in error should be changed.
	 * @return the first filter that has no element.
	 */
	protected LyFilter checkFilter(LyFilter filter, boolean showError)
	{
		if (filter == null)
			return null;

		LyDependencies	depends = getDependencies();
		Vector<Object>	dependency = depends.getDepends(filter);

		if ((dependency == null) || (dependency.size() == 0))
			return null;

		LyTreeNode	    treeNode = (LyTreeNode)dependency.firstElement();
		int			    i, n = filter.getSize();
		LyFilterElement element;
		LyFilter	    result = ((n > 0) ? null : filter);

		for (i = 0; (i < n) && (result == null); i++)
		{
			element = filter.getElement(i);

			// If one subFilter is not checked return subFilter
			if ((element instanceof LyFilter) && (checkFilter((LyFilter)element, showError) != null))
				result = (LyFilter)element;

			// If one subFilter is a relation filter, check its internal filter
			else if ((element instanceof LyFilterRelation)
				&& (checkFilter(((LyFilterRelation)element).getFilter(), showError) != null))
				result = ((LyFilterRelation)element).getFilter();
		}

		if (result == filter)
		{
			if (showError)
				treeNode.setFontColor("LY_LABEL_ERROR_STATE");
		}
		else
			treeNode.setFontColor(null);

		return result;
	} // end method checkFilter

	/**
	 * Recursively simplify a filter by removing filters that only contains one filter.
	 *
	 * @param filter The filter to be simplified.
	 */
	protected void simplifyFilter(LyFilter filter)
	{
		if (filter == null)
			return;

		LyFilterElement subElement;
		int			    i, n = filter.getSize();

		for (i = 0; i < n; i++)
		{
			subElement = filter.getElement(i);
			if (subElement instanceof LyFilter)
				simplifyFilter((LyFilter)subElement);
		}

		LyFilterElement element = filter.getElement(0);

		if ((n == 1) && (element instanceof LyFilter))
		{
			filter.removeElement(element);
			filter.setType(((LyFilter)element).getType());
			n = ((LyFilter)element).getSize();
			for (i = 0; i < n; i++)
			{
				subElement = ((LyFilter)element).getElement(i);
				filter.addElement(subElement);
			}
		}
	} // end method simplifyFilter

    /**
     * Set accessor to the context client area controller which will receive contextual notification
     * when something occur on this area controller. This method is overriden to set the given
     * controller as filter client when the filter controller is used in a compound action.
     *
     * @param controller the contextual controller.
     */
    public void setContextClient(LyAreaController controller)
    {
        super.setContextClient(controller);

        if ((_filterEndInterface == null)
                && (controller instanceof LyFilterEndInterface))
            setFilterEndInterface((LyFilterEndInterface)controller);
    }

	/**
	 * Gets client (listener) for closing notification.
	 *
	 * @return client notified when the dialog closes. May be null.
	 */
	public LyFilterEndInterface getFilterEndInterface()
	{
		return _filterEndInterface;
	}

	/**
	 * Sets client (listener) for closing notification.
	 *
	 * @param endInterface client notified when the dialog closes.
	 */
	public void setFilterEndInterface(LyFilterEndInterface endInterface)
	{
		_filterEndInterface = endInterface;

		// If new clear initial filter
		String id_action = _action.getId();

		if ((id_action.indexOf("new") != -1) || (id_action.indexOf("New") != -1))
			_filter = null;
		else if (_filterEndInterface != null)
		{
			LyFilter filter = _filterEndInterface.getFilter();

			//we clone the filter if it exists.
			if (filter != null)
			{
				if (!filter.isExtended() && (filter.getSize() > 1))
				{

					// Add a level to really extend the filter.
					_filter = new LyFilter(LyFilter.Type.AND);
					_filter.addElement(filter);
				}
				else
					_filter = filter;

				_filter = _filter.cloneFilter();
			}
		}
		else
			_filter = null;

		// Check if root filter is ready
		if (_filter == null)
			_filter = new LyFilter(LyFilter.Type.AND);
	}

	/**
	 * Accessor to the list of selected tree nodes.
	 *
	 * @return a list of selected nodes (tree nodes) or new Vector() in selection is empty.
	 */
	Vector<LyFilterElement> getSelections()
	{
		return _selections;
	}

	/**
	 * Accessor to the class info for which this filter is created.
	 *
	 * @return the requested classInfo
	 */
	public LyClassInfo getClassInfo()
	{
		return _classInfo;
	}

	/**
	 * Gets the list of valid filter criterion.
	 * @return a list of filter criterion
	 */
	protected LyInfoList<LyFilterCriterion> getCriterionList()
	{
		LyFilterCriterion			  criterion;
		LyApplication				  application = getApplication();
		LyInfoList<LyFilterCriterion> list = application.getInfoList(LyFilterCriterion.class);
		int							  i, n;

		n = list.getSize();
		for (i = n - 1; i >= 0; i--)
		{
			criterion = list.getInfo(i);
			if (criterion.isInternal())
				list.removeInfo(criterion);
		}

		return list;
	}
	/* LyFilterEndInterface */

	/**
	 * Accessor to initial filter for the creation of an elementary filter.
	 *
	 * @return the current filter
	 */
	public LyFilter getFilter()
	{
		if ((_filterParent != null) && (CURRENT_FILTER_EXPRESSION.equals(_filterParent.getId())))
			return _filterParent;
		else
			return null;
	}

	/**
	 * Modifier for current editor filter.
	 */
	protected void setFilter(LyFilter filter)
	{
		if (filter == null)
			_filter = null;
		else
			_filter = filter.cloneFilter();
	}

	/**
	 * The filter edition form is closed : notify registered listener.
	 *
	 * @param controller   reference to the filter controller (filter form which is closing)
	 * @param actionDemand actionDemand on filter controller : id of the command such as
	 *                     LyFormController.DEMAND_VALIDATE, LyFormController.DEMAND_APPLY,
	 *                     LyFormController.DEMAND_CANCEL
	 * @param filter       the filter that was created in the form
	 */
	public void endFilter(LyController controller, String actionDemand, LyFilter filter)
	{
		if (_filterParent == null)
		{
			Vector<LyFilterElement> selections = getSelections();

			if (selections.size() > 0)
				_filterParent = (LyFilter)selections.firstElement();
			else
				_filterParent = _filter;
		}

		if ((filter != null) && (_filterParent != null))
		{
			int     pos = -1;
			boolean updateSelection = false;

			// If an expression is being set, remove old expression first
			if (CURRENT_FILTER_EXPRESSION.equals(_filterParent.getId())
                && (_filterParent.getSize() > 0))
			{
				LyDependencies  old_depends = getDependencies();
				LyFilterElement element = _filterParent.getElement(0);
				LyTreeNode	    treeNode = (LyTreeNode)old_depends.getDepends(element).firstElement();
				LyTreeNode	    treeNodeParent = (LyTreeNode)treeNode.getParent();

				_filterParent = (LyFilter)old_depends.getDepends(treeNodeParent).firstElement();

				// Retrieve old position
				int i, n = treeNodeParent.getTreeNodeCount();

				for (i = 0; i < n; i++)
				{
					if (treeNodeParent.getTreeNode(i) == treeNode)
					{
						pos = i;
						break;
					}
				}

				_filterParent.removeElement(element);
				if (_selections.contains(element))
				{
					setSelections(new Vector<LyFilterElement>());
					updateSelection = true;
				}

				treeNodeParent.removeTreeNode(treeNode);
				old_depends.removeObject(element);
			}

			// Retrieve internal tree component
			LyTree		    tree = getTreeComponent();
			LyDependencies  old_depends = getDependencies();
			LyTreeNode	    treeNode = null, treeNode2 = null;
			LyFilterElement element;
			LyTreeNode	    treeNodeParent;

			tree.setIdle(true);

			//If parent filter is AND type and the expression is multiple, we add filter OR
			if ((_filterParent.getType() == LyFilter.Type.AND) && (filter.getSize() > 1))
			{
				int i, n = filter.getSize();

				_filterParent.addElement(filter);
				treeNode = buildTreeNode(filter);
				for (i = 0; i < n; i++)
				{
					element = filter.getElement(i);
					treeNode2 = buildTreeNode(element);

					//addDependencies
					old_depends.addDepend(treeNode2, element);

					//Retrieve treeNodeParent and add son
					treeNode.addTreeNode(treeNode2);
					updateTreeNode(element, treeNode2);
				}

				//addDependencies
				old_depends.addDepend(treeNode, filter);

				//Retrieve treeNodeParent and add son
				treeNodeParent = (LyTreeNode)old_depends.getDepends(_filterParent).firstElement();
				treeNodeParent.setOpen(true);
				treeNodeParent.insertTreeNode(treeNode, pos);
				updateTreeNode(_filterParent, treeNodeParent);

				// update labels and expands
				treeNode.setOpen(true);
				updateTreeNode(filter, treeNode);
				if (updateSelection)
				{
					treeNode.setSelected(true);
					Vector<LyFilterElement> sel = new Vector<LyFilterElement>();

					sel.addElement(filter);
					setSelections(sel);
				}
			}
			else
			{
				int i, n = filter.getSize();

				//Retrieve treeNodeParent and add son
				if ((old_depends != null) && (old_depends.getDepends(_filterParent) != null))
				{
					treeNodeParent = (LyTreeNode)old_depends.getDepends(_filterParent).firstElement();

					for (i = 0; i < n; i++)
					{
						element = filter.getElement(i);
						if ((element instanceof LyFilterExpression) || (element instanceof LyFilterRelation)
							|| (element instanceof LyFilterCriterion))
						{
							_filterParent.addElement(element);
							treeNode = buildTreeNode(element);

							//addDependencies
							old_depends.addDepend(treeNode, element);

							//Retrieve treeNodeParent and add son
							treeNodeParent.setOpen(true);
							updateTreeNode(_filterParent, treeNodeParent);
							treeNode.setOpen(true);
							updateTreeNode(element, treeNode);
							treeNodeParent.insertTreeNode(treeNode, ((pos > 0) ? (pos + i) : pos));
							if (element instanceof LyFilterRelation)
							{
								LyFilter   subFilter = ((LyFilterRelation)element).getFilter();
								LyTreeNode subFilterTreeNode = buildTreeNode(subFilter);

								//addDependencies
								old_depends.addDepend(subFilterTreeNode, subFilter);

								//Retrieve treeNodeParent and add son
								updateTreeNode(subFilter, subFilterTreeNode);
								treeNode.addTreeNode(subFilterTreeNode);
							}

							if (updateSelection && (n == 1)) // force only selection if 1 element
							{
								treeNode.setSelected(true);
								Vector<LyFilterElement> sel = new Vector<LyFilterElement>();

								sel.addElement(element);
								setSelections(sel);
							}
						}
						else if (element instanceof LyFilter)
						{
							_filterParent.addElement(element);

							// Build the treeNode and the subTrees for this filter
							treeNode = addTreeNode(null, element, -1);

							//addDependencies
							old_depends.addDepend(treeNode, element);

							//Retrieve treeNodeParent and add son
							treeNodeParent.setOpen(true);
							updateTreeNode(_filterParent, treeNodeParent);
							treeNode.setOpen(true);
							updateTreeNode(element, treeNode);
							treeNodeParent.insertTreeNode(treeNode, ((pos > 0) ? (pos + i) : pos));
							if (updateSelection && (n == 1)) // force only selection if 1 element
							{
								treeNode.setSelected(true);
								Vector<LyFilterElement> sel = new Vector<LyFilterElement>();

								sel.addElement(element);
								setSelections(sel);
							}
						}

						if (updateSelection && (n > 1)) // select parent
						{
							treeNodeParent.setSelected(true);
							Vector<LyFilterElement> sel = new Vector<LyFilterElement>();

							sel.addElement(_filterParent);
							setSelections(sel);
						}
					}
				}
			}

			// Check the filter in order to reset tree nodes colors
			checkFilter(_filterParent, false);
			tree.setIdle(false);

			// Update the last modification time.
			setLastModified(new Date().getTime());
		}
		else

			// reinitialize buffer if cancel
			_filterParent = null;
	} // end method endFilter

	/* LyTreeBehaviorInterface */
	/**
	 * Starts an action for the given controller.
	 *
	 * @param  form           the current LyAreaController instance
	 * @param  action         the action that will be executed
	 * @return true if the action starts correctly, false otherwise.
	 */
	public boolean startAction(LyAreaController form, LyAction action)
	{
		return true;
	}

	/**
	 * Ends an action for the given controller.
	 *
	 * @param  form     the current LyAreaController instance
	 * @param  action   the action that was executed
	 * @return true if the action performed correctly, false otherwise
	 */
	public boolean endAction(LyAreaController form, LyAction action)
	{
		return true;
	}

	/**
	 * Gets the image corresponding to the given object for the given controller. This image is
	 * mostly used in titlebar.
	 *
	 * @param  areaController controller for which the image is required.
	 * @param  action         the action that was executed
	 * @return Logical name of the image for the given action.
	 */
	public String getImage(LyAreaController areaController, LyAction action)
	{
		return areaController.getImage(action);
	}

	/**
	 * Gets title corresponding to the given action.
	 *
	 * @param  areaController the current LyAreaController instance
	 * @param  action         the action that was executed
	 * @return Title for the given action.
	 */
	public String getTitle(LyAreaController areaController, LyAction action)
	{
		return action.getName();
	}

	/**
	 * Build a view object instance from the application object Node status (colors, borders, ...)
	 * should be set in updateTreeNode.
	 *
	 * @param  nodeObject     usually LyObject instance but not mandatory
	 * @return node view object.
	 */
	public LyTreeNode buildTreeNode(Object nodeObject)
	{
		LyTreeNode treeNode = null;

		if ((nodeObject instanceof LyFilter) || (nodeObject instanceof LyFilterElement))
		{
			treeNode = new LyTreeNode();
			treeNode.setId(nodeObject.getClass() + "@" + nodeObject.hashCode());
		}

		return (treeNode);
	}

	/**
	 * Update view object status (colors, label, border, style ...) the tree controller invokes this
	 * method when nodeObject is modified.
	 *
	 * @param nodeObject application object (reference)
	 * @param treeNode   view objects (need update)
	 */
	public void updateTreeNode(Object nodeObject, LyTreeNode treeNode)
	{
		if ((nodeObject == null) || (treeNode == null))
			return;

		if (nodeObject instanceof LyFilter)
		{
			LyFilter filter = (LyFilter)nodeObject;

			//if root object: specific label and expanded node
			if (filter == _filter)
			{
				if (treeNode.getLabel() == null)
				{
					String		  id = null;
					LyEnvironment env = getEnvironment();

					if (env != null)
						id = env.translate("LY_FILTER");

					treeNode.setLabel(id);
				}
			}
			else

				// Add AND or OR subtree filter
				treeNode.setLabel(getNameFilter((LyFilter)nodeObject, getFilterEnvironment(), treeNode.isOpen()));

			treeNode.setImage((filter.getType() == LyFilter.Type.AND) ? "LY_FILTER_AND" : "LY_FILTER_OR");
		}
		else if (nodeObject instanceof LyFilterExpression)
		{

			// Add filter expression (tree leaf)
			treeNode.setLabel(getNameExpression((LyFilterExpression)nodeObject, getFilterEnvironment()));
			treeNode.setImage("LY_FILTER_EXPRESSION");
		}
		else if (nodeObject instanceof LyFilterRelation)
		{

			// Add filter expression (tree leaf)
			treeNode.setLabel(getNameFilterRelation((LyFilterRelation)nodeObject, getFilterEnvironment(), treeNode.isOpen()));
			treeNode.setImage("LY_FILTER_RELATION");
		}
		else if (nodeObject instanceof LyFilterCriterion)
		{
			treeNode.setLabel(getNameFilterCriterion((LyFilterCriterion)nodeObject, getFilterEnvironment()));
			treeNode.setImage("LY_FILTER_CRITERION");
		}
	} // end method updateTreeNode

	/**
	 * Returns the environement that must be used to display filter data.
	 */
	protected LyEnvironment getFilterEnvironment()
	{
		return getEnvironment();
	}

	/**
	 * Accessor to the Tree component (view).
	 *
	 * @return The tree view component
	 */
	public LyTree getTreeComponent()
	{

		// Retrieve internal tree component
		LyTree tree = (LyTree)getComponent().getComponent(getId() + "_tree");

		return tree;
	}

	/**
	 * Retrieves the default view Builder.
	 *
	 * @return the default builder instance
	 * @see    LyController#getDefaultBuilder
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyFilterEditorBuilder(this);
	}

	/**
	 * Gets dependencies of the tree controller.
	 *
	 * @return Dependencies of the tree controller.
	 * @see    LyDependencies
	 */
	public LyDependencies getDependencies()
	{
		return _objectAssoc;
	}

	/**
	 * Updates filter controller.
	 * Method called by LyCompoundFilterController on initialization.
	 */
	protected void updateFilterController()
	{
		if (_parent instanceof LyCompoundFilterController)
			_filterController = ((LyCompoundFilterController)_parent).getFilterController();
	}

	/**
	 * Clear edited filter and reset filter controller panel.
	 */
	protected void clearFilter()
	{
		_filter = new LyFilter(LyFilter.Type.AND);
		reset(true);
		resetFilterController(null);
	}

	/**
	 * Reset filter controller panel with given selected element.
	 * @param element Filter element that is selected on editor panel.
	 */
	protected void resetFilterController(LyFilterElement element)
	{
		if (!(_parent instanceof LyCompoundFilterController))
		{
			freeFilterController();
			return;
		}

		LyApplication application = getApplication();
	    LyClassInfo   classInfo = _classInfo;
		LyAction	  action = application.getActionWithId("__filter_item", classInfo);

		if (element != null)
		{
			LyFilterRelation filterRelation = getParentFilterRelation(element);

			if (filterRelation != null)
				classInfo = filterRelation.getClassInfo();

			if (element instanceof LyFilter)
			{
				if (((LyCompoundFilterController)_parent).isExtendedFilter())
					_filterParent = (LyFilter)element;
				else if (((LyFilter)element).getSize() > 0)
				{
					LyFilterElement filterElement = ((LyFilter)element).getElement(0);

					// Create a temporary filter used by the filter controller
					// (a call to getFilter() is done by the new controller)
					_filterParent = new LyFilter(CURRENT_FILTER_EXPRESSION, null, LyFilter.Type.AND);
					_filterParent.addElement(filterElement);
				}
			}
		}
		else
			_filterParent = null;

		// Create a new filter controller to create Expression
		createFilterController(action, null, classInfo);
	}

	/**
	 * Translate the values displayed in filter form.
	 *
	 * @param value the value that must be translated
	 * @return the value that must be displayed
	 */
	public String translate(String value)
	{
		return value;
	}
} // end class LyFilterEditorController
