/**
 * LyFilterRelationController.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved. Current version: $Id:
 * LyFilterRelationController.java,v 6.10 2009/01/06 11:03:37 jls Exp $
 */

package leon.control;

import leon.app.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * This class implements the relation filter edition area. This form enables the user to type in
 * relation filter (fieldInfo / filter).
 *
 * <P>This controller is associated to a behavior which implements LyFilterBehaviorInterface.
 * Usually the behavior is an instance of LyClassBehavior.</P>
 *
 * @see  LyFilterEditorController
 * @see  LyFilterEndInterface
 */
public class LyFilterRelationController extends LySimpleFilterController
	implements LyFilterEndInterface
{
	/** File Version */
	public static final String __VERSION = "$Id: LyFilterRelationController.java 34618 2012-03-14 10:10:21Z jlsaouli $";

	/** Identifier of an input field : selected fieldInfo. */
	public static final String SELECT_FIELDINFO = "_filterfieldinfo";

	/** Identifier of an input field : user specified value. */
	public static final String SELECT_CLASS = "_filterclass";

	/** Identifier of an input field : user specified value. */
	public static final String SELECT_FILTER = "_filtervalue";

	/** Identifier of create simple filter button. */
	public static final String CREATE_SIMPLE_FILTER = "_createSimpleFilter";

	/** Identifier of create extended filter button. */
	public static final String CREATE_EXTENDED_FILTER = "_createExtendedFilter";

	/** Identifier of set filter button. */
	public static final String SET_FILTER = "_setFilter";

	/** Identifier of extend simple filter button. */
	public static final String EXTEND_SIMPLE_FILTER = "_extendSimpleFilter";

	/** List of created filters. */
	protected LyInfoList<LyFilter> _createdFilters;

	/** Indicates if a filter is being set. */
	protected boolean _setFilter;

	/** Indicates if relation containing the list of possible filters must be displayed. */
	protected boolean _displayFilterRelation;

	/**
	 * Constructor.
	 *
	 * @param  session    current session associated to the controller. The controller automatically
	 *                    registers on the specified session.
	 * @param  parent     parent controller. If the parent implements LyFilterEndInterface, it will
	 *                    be notified when the filter form is closed.
	 * @param  action     related action. (action based on '_filter' action)
	 * @param  classInfo  the application class associated to the filter
	 * @param  objects    current context
	 */
	public LyFilterRelationController(LySession session, LyController parent, LyAction action, LyClassInfo classInfo,
		LyObjectList objects)
	{
		super(session, parent, action, classInfo, objects);

		_setFilter = false;
		_displayFilterRelation = true;
	}

	/**
	 * Internal method to compute id of this controller. This method will always return an id based
	 * on this parent
	 *
	 * @param   session  Session of the new controller.
	 * @param   id       Desired id (may be null).
	 *
	 * @return  Id of the new controller.
	 */
	protected String computeId(LySession session, String id)
	{
		if ((_parent instanceof LyCompoundFilterController)
				|| (_parent instanceof LyFilterEditorController))
		{
			// Redirect demand on compound filter
			// Get the parent id
			LyCompoundFilterController compound = null;

			if (_parent instanceof LyCompoundFilterController)
				compound = (LyCompoundFilterController)_parent;
			else if ((_parent instanceof LyFilterEditorController)
					&& (_parent.getParent() instanceof LyCompoundFilterController))
				compound = (LyCompoundFilterController)_parent.getParent();

			if (compound != null)
				return "FilterControllerId" + compound.getId();
		}

		return super.computeId(session, id);
	}

	/**
	 * Invoked when an action notification is sent to this listener.
	 *
	 * <P>method implements VALIDATE, APPLY and CANCEL action events and subsequently invoke
	 * endFilter on the client.</P>
	 *
	 * <P>Additionally, the controller handles field changes and impact on form content.</P>
	 *
	 * @param  demand  demand parameters sent by the view component.
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			if (_parent instanceof LyCompoundFilterController)
			{
				// Redirect demand on compound filter
				((LyCompoundFilterController)_parent).actionDemand(demand);
				return;
			}

			// Menu or action bar
			String id = demand.getSource();

			if (id.equals(CREATE_SIMPLE_FILTER)
				|| id.equals(CREATE_EXTENDED_FILTER)
				|| id.equals(SET_FILTER)
				|| id.equals(EXTEND_SIMPLE_FILTER))
			{
				// Create a new filter controller
				String				    actionId = null;
				LyFilterRelationBuilder builder = (LyFilterRelationBuilder)getBuilder();
				LyFilter			    filter = builder.getSelectedFilter();

				if (id.equals(CREATE_SIMPLE_FILTER))
					actionId = "__filter_new";
				else if (id.equals(CREATE_EXTENDED_FILTER))
					actionId = "__filter_extended_new";
				else if (id.equals(SET_FILTER))
				{
					if (filter.isExtended())
						actionId = "__filter_extended_set";
					else
						actionId = "__filter_set";

					_setFilter = true;
				}
				else if (id.equals(EXTEND_SIMPLE_FILTER))
				{
					actionId = "__filter_extended_set";
					_setFilter = true;
				}

				LySession		 session = getSession();
				LyApplication    application = session.getApplication();
                LyClassInfo      classInfo = ((LyFilterRelationBuilder)getBuilder()).getSelectedClassInfo();
				LyAction		 action = application.getActionWithId(actionId, classInfo);
				LyAreaController controller = (LyAreaController)invokeAction(action, classInfo, null, false);

				controller.showView(null);
				return;
			}

			if (id.equals(LyFormController.DEMAND_CANCEL))
			{
				if (_hasViewController)
				{
					if (_filterEndInterface != null)
						_filterEndInterface.endFilter(this, id, null);

					free();
				}
				else
				{
					if (_filterEndInterface != null)
						_filterEndInterface.endFilter(this, DEMAND_VALIDATE, null);

					hideView();
					free(_parent);
				}

				return;
			}

			if ((id.equals(LyFormController.DEMAND_VALIDATE)
						|| id.equals(LyFormController.DEMAND_APPLY))
					&& ((_defaultFilter == null) || checkIfModified()))
			{
				validate(demand);
				if (id.equals(LyFormController.DEMAND_VALIDATE))
					free();

				return;
			}
		}
		else if (demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
		{
			_modified = true;

			// Refresh view
			LyFilterRelationBuilder builder = (LyFilterRelationBuilder)getBuilder();

			if (SELECT_FIELDINFO.equals(demand.getSource()))
			{
				builder.resetClassField();

				if (_displayFilterRelation)
					builder.resetFilterField();

				builder.resetForm();
			}
			else if (SELECT_CLASS.equals(demand.getSource()))
			{
				if (_displayFilterRelation)
				{
					builder.resetFilterField();
					builder.resetForm();
				}
			}
			else if (SELECT_FILTER.equals(demand.getSource()))
				builder.updateTools();

			return;
		}

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Internal method to validate filter creation.
	 * @param demand the action demand
	 */
	protected void validate(LyActionDemand demand)
	{
		LyFilterRelation filterRelation = getFilterRelation();

		// If no filter has been chosen, show an error
		if (_displayFilterRelation
				&& ((filterRelation == null) || (filterRelation.getFilter() == null)))
		{
			LyEnvironment environment = getEnvironment();

			showError(environment.getMessage("LY_ER_NOTNULL", "LY_FILTER"));
			return;
		}

		// Notify filter interface of the creation
		if (_filterEndInterface != null)
		{
			String   id = demand.getSource();
			LyFilter filter = new LyFilter();

			if (filterRelation.getFilter() == null)
			{
				LyFilter internalFilter = new LyFilter();

				filterRelation.setFilter(internalFilter);
			}

			filter.addElement(filterRelation);
			_modified = false;
			_filterEndInterface.endFilter(this, id, filter);
		}
	}

	/**
	 * Forces the field that contains the possible classes to the given class.
	 *
	 * @param  classInfo  the class info that must be set on the field.
	 */
	public void setClassInfo(LyClassInfo classInfo)
	{
		boolean				    oldState = _modified;
		LyFilterRelationBuilder builder = (LyFilterRelationBuilder)getBuilder();

		builder.setClassInfo(classInfo);
		_modified = oldState;
	}

	/**
	 * Adds a filter in the list of created filters.
	 *
	 * @param  filter  new filter.
	 */
	public void addFilter(LyFilter filter)
	{
		if (filter == null)
			return;

		if (_createdFilters == null)
			_createdFilters = new LyInfoList<LyFilter>();

		_createdFilters.addInfo(filter);
	}

	/**
	 * Returns the created relation filter.
	 *
	 * @return  the created relation filter
	 */
	protected LyFilterRelation getFilterRelation()
	{
		LyFilterRelationBuilder builder = (LyFilterRelationBuilder)getBuilder();
		LyRelationFieldInfo     fieldInfo = builder.getSelectedFieldInfo();
		LyFilter			    filter = builder.getSelectedFilter();
		LyFilterRelation	    filterRelation = new LyFilterRelation(fieldInfo, filter);
		LyClassInfo			    classInfo = builder.getSelectedClassInfo();

		filterRelation.setClassInfo(classInfo);

		return filterRelation;
	}

	/**
	 * Indicates if relation containing the list of possible filters must be displayed.
	 *
	 * @param  displayFilterRelation  true if relation must be displayed, false otherwise.
	 */
	public void setDisplayFilterRelation(boolean displayFilterRelation)
	{
		_displayFilterRelation = displayFilterRelation;
	}

	/**
	 * Indicates if relation containing the list of possible filters is displayed.
	 *
	 * @return true if relation is displayed, false otherwise.
	 */
	public boolean isDisplayFilterRelation()
	{
		return _displayFilterRelation;
	}

	/**
	 * Sets the editable state of the form component.
	 *
	 * @param  editable  true if form must be editable, false otherwise.
	 */
	public void setFormEditable(boolean editable)
	{
		// Retrieve the form component
		LyForm form = (LyForm)_component.getComponent(getId());

		form.setEditable(editable);
	}

	/**
	 * Indicates if the view for this import controller is resizable.
	 *
	 * @return  true if the view is resizable, false otherwise.
	 */
	public boolean isViewResizable()
	{
		return false;
	}

	/**
	 * Factory of the default builder for a controller. Current implementation returns a
	 * LyFilterBuilder instance.
	 *
	 * @return  the builder which creates the XML description of the view
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyFilterRelationBuilder(this);
	}

	/**
	 * Called when current object has changed . Keep the overall content consistent.
	 *
	 * @param  oldValues  set of former values, or null if not specified.
	 */
	protected void updateObject(LyValueSet oldValues)
	{
	}

	/**
	 * Accessor to the list of objects associated with this form. The user may navigate through this
	 * list of objects.
	 *
	 * @param   objects  list of objects which could be shown in the form.
	 *
	 * @return  true if successfull
	 */
	public boolean setObjects(LyObjectList objects)
	{
		// Not used in current version
		// Could be used to have field domains that depend on context
		return false;
	}

	/**
	 * Gets the created filters (may be simple or extended filters).
	 *
	 * @return  the list of created filters.
	 */
	public LyInfoList<LyFilter> getCreatedFilters()
	{
		return _createdFilters;
	}

	/* Implementation of LyFilterEndInterface */

	/**
	 * Gets initial : indicates to the filter controller the default filter parameters.
	 *
	 * @return  current filter or null if no filter is selected
	 */
	public LyFilter getFilter()
	{
		LyFilterRelationBuilder builder = (LyFilterRelationBuilder)getBuilder();
		LyFilter			    filter = builder.getSelectedFilter();

		return filter;
	}

	/**
	 * The filter edition form is closed : notify registred listener.
	 *
	 * @param  parent        reference to the filter controller (filter form which is closing)
	 * @param  actionDemand  actionDemand on filter controller : id of the command such as
	 *                       LyFormController.DEMAND_VALIDATE, LyFormController.DEMAND_APPLY,
	 *                       LyFormController.DEMAND_CANCEL
	 * @param  filter        the filter that was created in the form
	 */
	public void endFilter(LyController parent, String actionDemand, LyFilter filter)
	{
		if (filter == null)
		{
			_setFilter = false;
			return;
		}

		LyFilterRelationBuilder builder = (LyFilterRelationBuilder)getBuilder();
		LyFilter			    oldFilter = builder.getSelectedFilter();

		if (_setFilter && (_createdFilters != null) && (_createdFilters.contains(oldFilter)))
		{
			_createdFilters.removeInfo(oldFilter);

			if (!DEMAND_APPLY.equals(actionDemand))
				_setFilter = false;
		}

		addFilter(filter);

		builder.resetFilterField();
		builder.resetForm();
	}
} // end class LyFilterRelationController
