/*
 * LyFindController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.control;

import java.util.*;

import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.view.*;


/**
 * This is the controller for the find action.
 *
 * @version $Id: LyFindController.java 28997 2010-12-23 10:17:15Z bmichel $
 * @author  Lyria
 */
public class LyFindController extends LyUserInputController
{
	/** File Version */
	public static final String		  __VERSION =
		"$Id: LyFindController.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** Menu id : Find. */
	public static final String		  DEMAND_FIND = "_find".intern();

	/** Menu id : Find. */
	public static final String		  DEMAND_CLOSE_FIND = "_closeFind".intern();

	/** Menu id : Find next. */
	public static final String		  DEMAND_NEXT = "_next".intern();

	/** Menu id : Find previous. */
	public static final String		  DEMAND_PREVIOUS = "_previous".intern();

	/** Find popup : Find pattern. */
	public static final String		  DEMAND_FIND_PATTERN = "_findPattern".intern();

	/** Find popup : Find field. */
	public static final String		  DEMAND_FIND_FIELD = "_findField".intern();

	/** Find popup : Whole word. */
	public static final String		  DEMAND_FIND_WHOLE_WORD = "_findWholeWord".intern();

	/** Find popup : Case sensitive. */
	public static final String		  DEMAND_FIND_CASE_SENSITIVE = "_findCaseSensitive".intern();

	/** Search pattern. */
	protected String				  _pattern;

	/** Field info used for search. */
	protected LyFieldInfo			  _fieldInfo;

	/** Whole word. */
	protected boolean				  _wholeWord;

	/** Case-sensitive. */
	protected boolean				  _caseSensitive;

	/** Visible state. */
	protected boolean				  _visible;

	/** List of fields used for search. */
	protected LyInfoList<LyFieldInfo> _fields;

	/**
	 * Constructor for the find action
	 * @param session the current session
	 * @param parent the parent action
	 * @param action the action to invoke
	 * @param classInfo a classinfo
	 * @param context the current context
	 */
	public LyFindController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent, action, classInfo, context);

		_pattern = null;
		_fieldInfo = null;
		_wholeWord = false;
		_caseSensitive = false;
		_visible = false ;
		_hasViewController = false;
		_fields = null;

		//setBuilder(new LyFindBuilder(this));
		//setBehavior(new LyFindBehavior());
	}

	/**
	 * Search pattern.
	 * @return the search pattern
	 */
	public String getPattern()
	{
		return _pattern;
	}

	/**
	 * Search pattern.
	 * @param pattern the pattern to set
	 */
	public void setPattern(String pattern)
	{
		 _pattern=pattern;
	}
	
	/**
	 * Field info used for search.
	 * @return the current field
	 */
	public LyFieldInfo getFieldInfo()
	{
		return _fieldInfo;
	}
	
	/**
	 * Field info used for search.
	 * @param fieldInfo the fieldInfo to use
	 */
	public void setFieldInfo(LyFieldInfo fieldInfo)
	{
		 _fieldInfo=fieldInfo;
	}

	/**
	 * Whole word.
	 * @return true if the search is on whole word
	 */
	public boolean isWholeWord()
	{
		return _wholeWord;
	}

	/**
	 * Set the whole word search mode
	 * @param word the whole word search mode
	 */
	public void setWholeWord(boolean word) 
	{
		_wholeWord = word;
	}

	/**
	 * Case-sensitive.
	 * @return the case sensitive status
	 */
	public boolean isCaseSensitive()
	{
		return _caseSensitive;
	}

	/**
	 * Set the case sensitive status
	 * @param sensitive the status
	 */
	public void setCaseSensitive(boolean sensitive) 
	{
		_caseSensitive = sensitive;
	}
	
	/**
	 * Gets fields available for find action.
	 * @return a list of fields
	 */
	public LyInfoList<LyFieldInfo> getFields()
	{
		if (_fields != null)
			return _fields;

		LyInfoList<LyFieldInfo> fields = null;
		LyFieldInfo			    fieldInfo = (LyFieldInfo)_action.getParameter("_fieldInfo", true);
		int					    np = ((fieldInfo != null) ? 2 : 1);

		do
		{
			if (fieldInfo != null)
			{
				if (fields == null)
					fields = new LyInfoList<LyFieldInfo>();
				fields.addInfo(fieldInfo);
			}
			fieldInfo = (LyFieldInfo)_action.getParameter("_fieldInfo" + np, true);
			np++;
		}
		while (fieldInfo != null);

		if (fields == null)
		{
			BitSet fieldMarks = getParameterMarks("_fieldMarks", LyBitSet.Type.F_MARKS);

			if (fieldMarks != null)
				fields = _classInfo.getFieldInfos(fieldMarks);
		}

		if ((fields == null) || (fields.getSize() < 1))
		{
			fields = _classInfo.getFieldInfos();
			fields = fields.getInfos(LyFieldInfo.MARK_HIDDEN, true);
			fields = fields.getInfos(LyFieldInfo.MARK_SECRET, true);
		}

		return fields;
	} // end method getFields

	/**
	 * Make the view visible to the user.
	 */
	public void showView()
	{
		if (_visible)
			return;

		_visible = true;
		super.showView();

		LyAreaController parent = getParentAreaController();

		updateFindMenu(parent);
	}

	/**
	 * Hide the view.
	 * @param free indicates if this controller must be freed
	 */
	public void hideView(boolean free)
	{
		if (!_visible)
			return;

		_visible = false;
		super.hideView(free);

		LyAreaController parent = getParentAreaController();

		updateFindMenu(parent);
	}

	/**
	 * Returns visible state.
	 * @return the visible status
	 */
	public boolean isVisible()
	{
		return _visible;
	}

	/**
	 * Update the find menuitem status.
	 * @param parent parent controller
	 */
	public void updateFindMenu(LyAreaController parent)
	{
		// If panel is Idle, no need to refresh
		if (parent.isIdle())
			return;

		LyViewManager viewManager = getViewManager();
		boolean		  hasMenu = viewManager.hasProperty("MENU");
		// ** Find menu status **
		LyView		  view = parent.getView();
		LyItem		  item = null;
		LyComponent   comp = null;
        LyMenuBar     menuBar = (view instanceof LyFrame ? ((LyFrame)view).getMenuBar() : null);

		if (hasMenu && (menuBar != null))
		{
			item = menuBar.findItem(DEMAND_FIND, parent.getId());
			if (item != null)
				item.setChecked(_visible);

			item = menuBar.findItem(DEMAND_NEXT, parent.getId());
			if (item != null)
				item.setEnabled(_pattern != null);

			item = menuBar.findItem(DEMAND_PREVIOUS, parent.getId());
			if (item != null)
				item.setEnabled(_pattern != null);
		}
		else if (view != null)
		{
			comp = view.getComponent(DEMAND_NEXT + LyController.ID_SEP + parent.getId());
			if ((comp != null) && (comp instanceof LyTool))
				((LyTool)comp).setEnabled(_pattern != null);

			comp = view.getComponent(DEMAND_PREVIOUS + LyController.ID_SEP + parent.getId());
			if ((comp != null) && (comp instanceof LyTool))
				((LyTool)comp).setEnabled(_pattern != null);
		}
	} // end method updateFindMenu

} // end class LyFindController