/**
 * LyFormController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyFormController.java 40599 2012-08-30 08:51:17Z bangiole $
 */

package leon.control;

import java.util.*;

import leon.app.*;
import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.control.form.*;
import leon.data.*;
import leon.info.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * The form controller is an area which describes a dialog view with a set of fields and values.
 * User may just consult or modify data. Usually (not necessary) the form is associated to one or
 * several Leonardi objects (LyObjects).
 *
 * <P>This class is specialized in consult/create/set and other controllers.</P>
 *
 * <P>Services offered at this level of implementation :</P>
 *
 * <UL>
 *   <LI>Content management as a single object or a list of objects.</LI>
 *   <LI>Navigation between objects (previous/next)</LI>
 *   <LI>Context management</LI>
 *   <LI>Basic user interaction on command bar buttons and on toolbar icons</LI>
 * </UL>
 *
 * <P>The form controller is also receiving updates notification on object changes and may refresh
 * the information shown to the user : titlebar, toobar, field values ...</P>
 *
 * <P>The behavior associated to a LyFormController is shared by several implementation of this
 * controller : the behavior must implement LyFormBehaviorInterface and is usually an instance of
 * <b>LyClassBehavior</b>.</P>
 *
 * @see leon.app.behaviorinterface.LyFormBehaviorInterface
 * @see leon.app.behavior.LyClassBehavior
 */
public abstract class LyFormController extends LyMultiAreaController
	implements LyAdvancedWorkSpaceListener,
		LyGetClassInfoInterface
{
	/** File Version */
	public static final String						    __VERSION =
		"$Id: LyFormController.java 40599 2012-08-30 08:51:17Z bangiole $";

	/** Separator used to build form field identifiers. */
	public static final String						    FORM_FIELD_SEP = "..";

	/* Demand codes */

	/** Interaction code on the command bar button : Validate inputs. Apply and close. */
	public static final String						    DEMAND_VALIDATE = "_validate".intern();

	/** Interaction code on the command bar button : Apply inputs. Do not close the view. */
	public static final String						    DEMAND_APPLY = "_apply".intern();

    /** Interaction code on the command bar button : Save inputs. */
    public static final String                          DEMAND_SAVE = "_save".intern();

	/**
	 * User action on toolbar : Show previous object. (Available only if multiple objects are shown
	 * in the form)
	 */
	public static final String						    DEMAND_PREVIOUS_OBJECT = "_previousObject";

	/**
	 * User action on toolbar : Show next object. (Available only if multiple objects are shown in
	 * the form)
	 */
	public static final String						    DEMAND_NEXT_OBJECT = "_nextObject";

	/**
	 * User action on toolbar : select an object in the list. (Available only if multiple objects
	 * are shown in the form)
	 */
	public static final String						    DEMAND_SELECTION = "_selection".intern();

	/** Show previous window. */
	public static final String						    DEMAND_PREVIOUS_STEP = "_previousStep";

	/** Show next window. */
	public static final String						    DEMAND_NEXT_STEP = "_nextStep";

	/** Class info to which the form corresponds to. */
	protected LyClassInfo							    _classInfo;

	/**
	 * List of objects presented in this form. Usually one object is shown, and the user may
	 * navigate between objects.
	 */
	protected LyWorkSpace							    _objects;

	/** Current object presented in this form. */
	protected LyObject								    _object;

	/** List of fields displayed in form. */
	protected LyInfoList<LyFieldInfo>				    _fields;

	/** List of field marks used to configure this controller. */
	private BitSet									    _fieldMarks;

	/** Hashtable of fieldInfo controllers. */
	protected Hashtable<LyFieldInfo, LyFieldController> _fieldControllers;

	/**
	 * Composition relation : this field is set when this form is included in another form
	 * controller to manage a composition relation.
	 */
	protected LyRelationFieldInfo					    _compositionRelation;

	/** Indicates if the controller is changing its object. */
	protected boolean								    _changingObject;

	/** String used as a suffix for popup menu items (internal use only). */
	private String									    _contextualMenuSuffix;

	/** Store the last active field **/
	protected LyFieldInfo			                    _lastActiveField;

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 * @param action  related action. If an action is specified, the controller uses action
	 *                parameters (action name, action specific XML view description ...)
	 * @param classInfo the edited classInfo
	 * @param objects Object list currently being shown in the form (showing multiple objects).
	 */
	protected LyFormController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		super(session, parent, action, objects);

		if (action.getClassInfo() != null)
			_classInfo = action.getClassInfo();
		else
			_classInfo = classInfo;

		_objects = null;
		_object = null;
		try
		{
			_fieldMarks = getParameterMarks("_fieldMarks", LyBitSet.Type.F_MARKS);
		}
		catch (IllegalArgumentException e)
		{
			getEnvironment().logWarning(e.getMessage());
		}

		_changingObject = false;
		_contextualMenuSuffix = null;
		_hasViewController = _hasViewController && (action.getTarget() == null);

		Object id = getParameter("_object", true);

		if (id != null)
		{
			LyObjectId objectId = null;

			if (id instanceof LyObjectId)
				objectId = (LyObjectId)id;
			else if ((_classInfo == null)
					&& (id instanceof String))
				objectId = new LyObjectId(getApplication(), (String)id);
			else
				objectId = new LyObjectId(_classInfo, id);

			LyObject object = objectId.getLogicalObject();

			if (object != null)
			{
				if (object.instanceOf(_classInfo))
					setObject(object);
				else
					setObjects(new LySimpleObjectList(object));
			}
		}
		else if (objects != null)
			setObjects(objects);
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */

	public Class<?> getBehaviorInterface()
	{
		return LyFormBehaviorInterface.class;
	}

	/**
	 * Get the field controller of the given fieldInfo.
	 * @param fieldInfo a field
	 * @return the associated field controller
	 *
	 */
	public LyFieldController getFieldController(LyFieldInfo fieldInfo)
	{
		if (fieldInfo == null)
			return null;

		if (_fieldControllers == null)
			_fieldControllers = new Hashtable<LyFieldInfo, LyFieldController>();
		else
		{
			LyFieldController fieldController = _fieldControllers.get(fieldInfo);

			if (fieldController != null)
				return fieldController;
		}

		return createFieldController(fieldInfo);
	}

	/**
	 * Creates the field controller for the given fieldInfo.
	 * @param fieldInfo a field
	 * @return the field controller
	 */
	protected LyFieldController createFieldController(LyFieldInfo fieldInfo)
	{
		LyFieldController fieldController;

		if (isDisplayed(fieldInfo))
			fieldController = fieldControllerFactory(fieldInfo, false);
		else
			fieldController = null;

		if (fieldController != null)
			_fieldControllers.put(fieldInfo, fieldController);
		return fieldController;
	}

	/**
	 * Factory for constructing the field controller of a field info.
	 *
	 * @param  fieldInfo Field info for which a controller is built.
	 * @param forceConsult true to force consult mode
	 * @return The field controller of the given field.
	 */
	public LyFieldController fieldControllerFactory(LyFieldInfo fieldInfo, boolean forceConsult)
	{
		if (!isDisplayed(fieldInfo))
			return null;

		if (fieldInfo instanceof LyRelationFieldInfo)
			return new LyRelationFieldController(this, (LyRelationFieldInfo)fieldInfo);
		return new LyFieldController(this, fieldInfo);
	}

	/**
	 * Prepare component before showing. This method is called once (unlike reset which may be
	 * called several times), before first showing. Default implementation invokes updateTitleBar,
	 * updateObject then startAction.
	 */
	protected void initArea()
	{
	   if (_initialized)
	       return;

		super.initArea();

		if (isFreed())
			return;

		_initialized = false;

		// Fill component with current object values
		updateObject(null);

		// Update context
		updateContext();

		_initialized = true;

        if (_parent instanceof LyViewController)
            // Restore content when possible
            restore();
	}

	/**
	 * Restore the last active field set by the method setActiveField.
	 * @see setActiveField
	 */
	private void restoreLastActiveField()
	{
		if (_lastActiveField != null)
			setActiveField(_lastActiveField);

		_lastActiveField = null;
	}
	/**
	 * Activates a field. Retrieve its tab and selects it. If there is no tab in the view this
	 * method does nothing.
	 *
	 * @param fieldInfo The field info that must be selected.
	 */
	public void setActiveField(LyFieldInfo fieldInfo)
	{
		if ((_areaControllerStack != null ) && (_areaControllerStack.size() > 0))
		{
			// We are not on top, store the future active field for a later use...
			_lastActiveField = fieldInfo;
			return;
		}
		_lastActiveField = null;
		LyFormField formField = null;
		LyTabs	    tabs = null;

		// No field is active. Retrieve the tabs component (if existing)
		// and set selection to null.
		if (fieldInfo == null)
		{
			tabs = getTabs();
			if (tabs != null)
				tabs.setSelectedTab(null);
		}
		// Else, select the tab containing the field item
		else
		{
			formField = getFormField(fieldInfo);
			if (formField == null)
				return;

			LyComponent parent = formField.getParent();
			LyTab	    tab;

			while (parent != null)
			{
				if (parent instanceof LyTab)
				{
					tab = (LyTab)parent;
					tabs = (LyTabs)tab.getParent();
					tabs.setSelectedTab(tab.getLabel());
					parent = tabs;
				}

				parent = parent.getParent();
			}
			formField.setFocus();
		}
	} // end method setActiveField

	/**
	 * Retrieves the tabs component of this controller view.
	 *
	 * @return The tabs component of the view or null if there is no tabs component in the
	 *         controller view.
	 */
	public LyTabs getTabs()
	{
		if (_component == null)
			return null;

		// Try with global component
		int		    i, n = _component.getChildCount();
		LyComponent child;

		for (i = 0; i < n; i++)
		{
			child = _component.getChild(i);
			if (child instanceof LyTabs)
				return (LyTabs)child;
		}

		// Try with fields of the class
		Vector<LyTabs> tabs = new Vector<LyTabs>();

		_component.getComponents(LyTabs.class, tabs);

		if (tabs.size() > 0)
			return tabs.elementAt(0);

		return null;
	}

	/**
	 * Navigate to the next object in the list of objects associated to this action. The target
	 * object is usually shown in the view. If the object is the last in the list, this action
	 * navigates to the first object in the list.
	 * @param next true to edit next object, false for previous
	 *
	 * @see #setObject(LyObject)
	 */
	protected void nextObject(boolean next)
	{
		int no = ((_objects == null) ? 0 : _objects.getSize());

		if (no == 0)
			return;

		LyObjectList list = null;
		LyController grandParent = null;

		if (no == 1)
		{
			grandParent = getViewController().getParent();
			if (grandParent instanceof LySimpleTableController)
				list = ((LyAreaController)grandParent).getContent();
			else
				list = _objects;
		}
		else
			list = _objects;

		if (list == null)
			return;

		int index = list.getIndex(_object);

		LyObject  object = null;
		int		  n = list.getSize();
		LySession session = getSession();

		if ((index < 0) && (n > 0))
		{
			object = list.getObject(0);
			if (!session.enableAction(_action, this, new LySimpleObjectList(object), false))
				object = null;
		}
		else
		{
			while (object != _object)
			{
				if (next)
				{
					if (index < (n - 1))
						index++;
					else
						index = 0;
				}
				else
				{
					if (index > 0)
						index--;
					else
						index = n - 1;
				}

				object = list.getObject(index);
				if (session.enableAction(_action, this, new LySimpleObjectList(object), false))
					break;
			}
		}

		if ((object != null)
		    && (object != _object))
		{
            if ((object.getClassInfo() == getClassInfo()))
            {
                _changingObject = true;

    			if (no > 1)
    				setObject(object);
    			else
    			{
    				if (grandParent instanceof LySimpleTableController)
    					((LySimpleTableController)grandParent).selectObject(_object, false);
    				setObjects(new LySimpleObjectList(object));
    				if (grandParent instanceof LySimpleTableController)
    					((LySimpleTableController)grandParent).selectObject(object, true);
    			}

                _changingObject = false;
            }
            else if (_parent instanceof LyViewController)
            {
                if ((no == 1) && (grandParent instanceof LySimpleTableController))
                    ((LySimpleTableController)grandParent).selectObject(_object, false);

                LyFormController formController = (LyFormController)session.invokeAction(getParentAreaController(),
                     _action, object.getClassInfo(), _objects, false);

                formController.setObject(object);
                free();
                formController.showView();

                if ((no == 1) && (grandParent instanceof LySimpleTableController))
                    ((LySimpleTableController)grandParent).selectObject(object, true);
            }
		}
		else
		    updateTools();
	} // end method nextObject

	/**
	 * Clears all cached data for this controller (especially the view and the component).
	 *
	 * <P>This method may be invoked for memory optimization, but the controller is still valid and
	 * may be used. (Unlike free) Further usage, needs to rebuild internal component & view which
	 * may be time consuming.</P>
	 */
	public void clearViewCache()
	{
		// Clear field controllers
		if (_fieldControllers != null)
		{
			Enumeration<LyFieldController>   elements = _fieldControllers.elements();
			LyFieldController                fieldController;

			while (elements.hasMoreElements())
			{
				fieldController = elements.nextElement();
				fieldController.clearViewCache();
			}
		}

		super.clearViewCache();
	}

	/**
	 * Free this controller on request of the specified controller. This method invokes endAction.
	 *
	 * @param controller controller instance which requested the freeing of this controller. If the
	 *                   controller is the parent, this controller is freed otherwise the method
	 *                   does nothing.
	 */
	public void free(LyController controller)
	{
		if (_parent != controller)
			return;

		// Stop properly current edition
		endAction();

		// Free field controllers
		if (_fieldControllers != null)
		{
			Enumeration<LyFieldController>   elements = _fieldControllers.elements();
			LyFieldController                fieldController;

			while (elements.hasMoreElements())
			{
				fieldController = elements.nextElement();
				fieldController.free();
			}

			_fieldControllers = null;
		}

		// Unregister as client of object list
		if (_objects != null)
		{
			try
			{
				_objects.free(this);
			}
			catch (Exception e)
			{
				getEnvironment().writeException(e);
			}
			_objects = null;
		}

		_object = null;
		_fieldMarks = null;
		_fieldControllers = null;
		_fields = null;

		super.free(controller);
	} // end method free

	/**
	 * Invoked when an action occurs. At this level, the implementation manages
	 *
	 * <UL>
	 *   <LI>HREF demands. (demand.getType() = LyActionDemand.TYPE_HREF)</LI>
	 *   <LI>HREF demands. (demand.getType() = LyActionDemand.TYPE_ACTION AND demand.getSource() =
	 *     DEMAND_PREVIOUS_OBJECT or DEMAND_NEXT_OBJECT or DEMAND_SELECTION</LI>
	 *   <LI>User interaction on action bar.</LI>
	 * </UL>
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		String sourceId = demand.getSource().intern();

		if ((demand.getType() == LyActionDemand.TYPE_HREF)
				&& sourceId.startsWith("_formObjects"))
		{
			String     	stringId = demand.getValue();
			LyObjectId 	oid = getObjectId(stringId);
			LyFormField formField = (LyFormField)_component.getComponent(sourceId);
			LyHRef		hRef = (LyHRef)formField.getComponent(stringId);

			hRef.setVisited(true);

			// check if object is in the current object list
			LyObject   object = ((_objects == null) ? null : _objects.getObject(oid));

			if (object == null)
				object = (LyObject)oid.getObject();
			if (object != null)
				setObject(object);
			return;
		}
		else if (sourceId.equals("_tabs"))
		{
			LyParameterValues values = demand.getParameterValues();
			int				  tabIndex = demand.getIndex(0);
			LyComponent		  component = _component.getComponent(demand.getValue());
			LyTab			  tab = ((component instanceof LyTab) ? (LyTab)component : null);
			LyTabs			  tabs = ((tab == null) ? getTabs() : (LyTabs)tab.getParent());

			if (values != null)
			{
				setLastModified(new Date().getTime());

				// Check if there are tabs within tabs (composition RELATION fields) and we
				// may need to forward the event
				String targetControllerId = values.getParameter("_target_controller");

				if ((targetControllerId != null)
						&& (targetControllerId.length() > 0)
						&& (!targetControllerId.equals(getId())))
				{
					// Forward event !
					LyFormController formCtrl = (LyFormController)getSession().getController(
							targetControllerId);
					LyActionDemand   demandFwd = new LyActionDemand(demand.getType(), "_tabs");

					demandFwd.addIndex(tabIndex);

					formCtrl.actionDemand(demandFwd);
					return;
				}
			}

			if (tabs != null)
			{
				tab = tabs.getTab(tabIndex);

				// In replay mode, retrieve tab by its id and not by its index.
				// In web mode, there may be an offset in tabs indexes due to a first virtual tab
				// built for fields that are not in a specific tab.
				LyEnvironment environment = getEnvironment();
				LyPlayer	  player = environment.getPlayer();

				if ((player != null)
						&& player.isPlaying()
						&& (component instanceof LyTab))
					tab = (LyTab)component;

				if (tab != null)
					tabs.setSelectedTab(tab.getLabel());
			}
		}
		else if (demand.getType() == LyActionDemand.TYPE_CLOSE_TAB)
		{
			// Received a submit from a tab actions : force callback on the right controller
			// When a tab is closed on web mode on a tabareacontroller, we need to submit the form
			// however the right target of the demand is the sourceId controller itself !
			if ((getSession() != null)
					&& (getSession().getController(sourceId) instanceof LyTabAreaController))
			{
				LyTabAreaController tabAreaCtrl = (LyTabAreaController)getSession().getController(
						sourceId);
				int				    tabIndex = demand.getIndex(0);
				LyActionDemand	    demandFwd = new LyActionDemand(demand.getType(), "_tabs");

				demandFwd.addIndex(tabIndex);

				tabAreaCtrl.actionDemand(demandFwd);
				return;
			}
		}
		else if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{

			if (sourceId == DEMAND_PREVIOUS_OBJECT)
			{
				nextObject(false);
				return;
			}
			else if (sourceId == DEMAND_NEXT_OBJECT)
			{
				nextObject(true);
				return;
			}
			else if (sourceId == DEMAND_SELECTION)
			{
				selectObject();
				return;
			}
			else if (sourceId == DEMAND_HELP)
			{
				LyClassInfo   classInfo = getClassInfo();
				LyViewManager viewManager = getViewManager();
				String		  fichierAide = null;

				// if we know the classInfo and if there is an associated help file, we show it
				// else we let the area controller handle this case
				if (classInfo != null)
				{
					fichierAide = classInfo.getTip();

					if (fichierAide != null)
					{
						viewManager.showLocation(_parent.getComponent(), fichierAide, true);
						return;
					}
				}
			}
			// Received a submit from a tab actions : force callback on the right controller
			// When a tab is selected on web mode on a tabareacontroller, we need to submit the form
			// however the right target of the demand is the sourceId controller itself !
			else if ((getSession() != null)
					&& (getSession().getController(sourceId) instanceof LyTabAreaController))
			{
				LyTabAreaController tabAreaCtrl = (LyTabAreaController)getSession().getController(
						sourceId);
				int				    tabIndex = demand.getIndex(0);
				LyActionDemand	    demandFwd = new LyActionDemand(demand.getType(), "_tabs");

				demandFwd.addIndex(tabIndex);

				tabAreaCtrl.actionDemand(demandFwd);
				return;
			}
			else if (getSession() != null)
			{
				// Menu or action bar
				String[]	  parameters = LyString.stringToList(sourceId, ID_SEP);
				LyApplication application = getApplication();
				LyInfo		  info = application.getInfo(parameters[0]);

				if (info instanceof LyAction)
				{
					LyAction action = (LyAction)info;

					if (action != null)
					{
						String classInfoId = null;
						String fieldInfoId = null;

						if (parameters.length > 2)
						{
							classInfoId = parameters[1];
							fieldInfoId = parameters[2];
						}

						LyObject contextualObject = null;

						if (parameters.length > 3)
						{
							LyObjectId objectId = new LyObjectId(getApplication(), parameters[3]);

							contextualObject = objectId.getLogicalObject();
						}

						if (classInfoId != null)
						{
							LyClassInfo		    classInfo = getClassInfo(classInfoId, fieldInfoId);
							LyFieldInfo fieldInfo = ((fieldInfoId != null)
									? (LyFieldInfo)application.getInfo(fieldInfoId) : null);

							if (contextualObject != null)
								invokeAction(action, classInfo,
									new LySimpleObjectList(contextualObject));
							else
								invokeTransverseAction(fieldInfo, action, classInfo);
						}
						else if (_object == null)
							// Invoke the action on classInfo
							invokeAction(action, getClassInfo());
						else
							// Invoke the action on selections
							invokeAction(action, _object);
						return;
					}
				}

				if (info instanceof LyClassInfo)
				{
					String		 actionId = "_xref" + LyAction.ACTION_ID_SEP + parameters[0];
					LyAction     action = application.getAction("_xref");
					LyAction     classAction = application.getAction(actionId);
					LyController controller;
					LyObject     contextualObject;

					if (parameters.length > 2)
					{
						LyObjectId objectId = new LyObjectId(getApplication(), parameters[2]);

						contextualObject = objectId.getLogicalObject();

						if (classAction == null)
							classAction = new LyAction(actionId, action, (LyClassInfo)info, null);

						controller = new LyXRefAreaController(getSession(), this, classAction, null,
								new LySimpleObjectList(contextualObject));

						getSession().showView(controller);
						return;
					}
				}
			}
		}

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Invoked when a menu demand occurs. At this level, the implementation manages demands on
	 * objects represented by an hyperlink.
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   LyController#demandPerformed
	 */
	public void menuDemand(LyMenuDemand demand)
	{
		String	    item = demand.getItem();
		LyComponent component = getComponent();
		LyComponent sourceComponent = component.getComponent(item);

		if (sourceComponent instanceof LyHRef && (item != null) && (item.indexOf(LyObjectId.CLASS_SEP) >= 0))
		{
            LyObjectId objectId = new LyObjectId(getApplication(), item);
            LyObject   contextualObject = objectId.getLogicalObject();

            if (contextualObject == null)
                return;

			// Build popup menu
			LyHRef     href = (LyHRef)sourceComponent;
			String     fieldId = href.getParentFormFieldId();
			String     field = fieldId.substring(fieldId.indexOf(FORM_FIELD_SEP)
					+ FORM_FIELD_SEP.length());
			LyForm     parentForm = href.getParentForm();

			_contextualMenuSuffix = LyController.ID_SEP + field + LyController.ID_SEP + item;

			LyPopupMenu popupMenu = buildPopupMenu(demand, contextualObject.getClassInfo(), true);

			_contextualMenuSuffix = null;

			// Coordinates
			int[] p = demand.getPosition();

			popupMenu.setX(p[0]);
			popupMenu.setY(p[1]);

			if (parentForm == null)
				// An error occured, just return
				return;

			popupMenu.setController(getId());
			popupMenu.setEnvironment(parentForm.getEnvironment());
			parentForm.setPopupMenu(popupMenu);

			return;
		}

		super.menuDemand(demand);
	} // end method menuDemand

	/**
	 * Check if the given action is enabled in the context of this controller.
	 * @param action the action to check
	 *
	 * @return True if the action is enabled, false if disabled.
	 */
	protected boolean enableAction(LyAction action)
	{
		if (_contextualMenuSuffix != null)
		{
			String[]   parameters = LyString.stringToList(_contextualMenuSuffix, ID_SEP);
			LyObjectId objectId = new LyObjectId(getApplication(), parameters[2]);
			LyObject   contextualObject = objectId.getLogicalObject();
			LySession  session = getSession();

			return session.enableAction(action, this, new LySimpleObjectList(contextualObject),
				false);
		}
		else
			return super.enableAction(action);
	}

	/**
	 * Build a contextual menu item with given actions and class infos.
	 *
	 * @param  idAction       action used to generate item identifier.
	 * @param  idClassInfo    class info used to generate item identifier.
	 * @param  labelAction    action used to generate item label.
	 * @param  labelClassInfo class info used to generate item label.
	 * @return the menu item build from given actions and class infos.
	 */
	protected LyItem buildPopupMenuItem(LyAction idAction, LyClassInfo idClassInfo,
		LyAction labelAction, LyClassInfo labelClassInfo)
	{
		String id = ((idAction == null) ? "" : idAction.getId());
		String label = ((labelAction == null) ? "" : labelAction.getName());

		if ((idAction != null) && (idClassInfo != null))
			id += LyController.ID_SEP;

		if (idClassInfo != null)
			id += idClassInfo.getId();

		id += _contextualMenuSuffix;

		if ((labelAction != null) && (labelClassInfo != null))
			label += " : ";

		if (labelClassInfo != null)
			label += labelClassInfo.getName();

		return new LyItem(id, label);
	}

	/**
	 * Internal method to invoke on a relation a transverse action for a given classInfo.
	 *
	 * @param fieldInfo  field on which the transverse action is invoked.
	 * @param action    Transverse action.
	 * @param classInfo Class info on which the action is applied.
	 */
	protected void invokeTransverseAction(LyFieldInfo fieldInfo, LyAction action,
		LyClassInfo classInfo)
	{
		LyFieldController controller = getFieldController(fieldInfo);

		if (controller == null)
			return;

		controller.invokeTransverseAction(action, classInfo);
	}

	/**
	 * Sets composition relation : this method is invoked when this form manages a relation in
	 * another form.
	 *
	 * @param relation the relation fieldInfo
	 */
	public void setCompositionRelation(LyRelationFieldInfo relation)
	{
		_compositionRelation = relation;
	}

	/**
	 * Gets the composition relation managed by this form.
	 *
	 * @return Composition relation managed by this form (may be null).
	 */
	public LyRelationFieldInfo getCompositionRelation()
	{
		return _compositionRelation;
	}

	/**
	 * Called when current object has changed . Keep the overall content consistent.
	 *
	 * @param oldValues set of former values, or null if not specified.
	 */
	protected void updateObject(LyValueSet oldValues)
	{
		if (_component == null)
			return;

		if (_component instanceof LyPanel)
			// Freeze the panel
			((LyPanel)_component).setIdle(true);

		LyClassInfo classInfo = getClassInfo();

		if (classInfo == null)
			return;

		// Update view content
		LyValue     fieldValue;
		Object	    newValue, oldValue;
		LyFieldInfo fieldInfo;
		int		    i, n;
		LyValueSet  newValues = new LyValueSet();

		if ((oldValues == null) || (oldValues.getSize() == 0))
		{
			oldValues = null;
			n = classInfo.getFieldInfoCount();
		}
		else
			n = oldValues.getSize();

		// Update titleBar
		updateTitleBar();

		// Update fields
		for (i = 0; i < n; i++)
		{
			if (oldValues == null)
			{
				oldValue = null;
				fieldInfo = classInfo.getFieldInfo(i);
			}
			else
			{
				fieldValue = oldValues.getFieldValue(i);
				oldValue = fieldValue.getValue();
				fieldInfo = (LyFieldInfo)fieldValue.getField();
			}

			if (!classInfo.contains(fieldInfo)
					&& classInfo.isVirtual())
			{
				fieldInfo = classInfo.getVirtualFieldInfo(fieldInfo);
				if (fieldInfo == null)
					continue;
			}

			if (!contains(fieldInfo))
				// This field is not present in the form
				continue;

			// Update field item
			newValue = getFormValue(fieldInfo);
			updateFieldItem(fieldInfo, newValue, oldValue, false);

			fieldValue = getFieldValue(fieldInfo); // new value may have change on update
			if (fieldValue != null)
				newValues.addValue(fieldValue);
		}

		// Propagate modifications at the end of update
		for (i = 0, n = classInfo.getFieldInfoCount(); i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			if (newValues.contains(fieldInfo))
				fieldValue = newValues.getFieldValue(fieldInfo);
			else
			{
				fieldValue = getFieldValue(fieldInfo);
				if ((fieldValue == null)
						&& (_object != null))
					fieldValue = _object.getFieldValue(fieldInfo);
			}

			newValue = (fieldValue == null ? null : fieldValue.getValue());
			propagate(fieldInfo, newValue);
		}

		// Update status
		updateStatus();

		// Update toolBar
		updateTools();

		if (_component instanceof LyPanel)
			// Unfreeze the panel
			((LyPanel)_component).setIdle(false);
	} // end method updateObject

	/**
	 * Update the status.
	 */
	protected void updateStatus()
	{
		if (_object == null)
			return;

		LyClassInfo classInfo = getClassInfo();
		int		    n = classInfo.getFieldInfoCount();
		LyFieldInfo fieldInfo;
		short	    errorLevel, state;

		for (int i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			if (!contains(fieldInfo))
				// This field is not present in the form
				continue;
			errorLevel = _object.getErrorLevel(fieldInfo);
			switch (errorLevel)
			{
			case LyObject.LEVEL_ERROR:
				state = LyFormField.STATE_ERROR;
				break;
			case LyObject.LEVEL_WARNING:
				state = LyFormField.STATE_WARNING;
				break;
			case LyObject.LEVEL_OK:
				state = LyFormField.STATE_NORMAL;
				break;
			default:
				state = errorLevel;
				break;
			}
			setState(fieldInfo, state);
		}
	} // end method updateStatus

	/**
	 * Get object value to put it in the form.
	 *
	 * @param  fieldInfo FieldInfo associated to the requested value
	 * @return Value of the field info.
	 */
	protected Object getFormValue(LyFieldInfo fieldInfo)
	{
        LyApplication             application = _session.getApplication();
        LyEvaluator               evaluator = application.setCurrentEvaluator(this);
        LyFormBehaviorInterface   behavior = (LyFormBehaviorInterface)getBehavior();
        Object                    value;

        try
        {
    		value = behavior.getValue(this, fieldInfo);
        }
        finally
        {
            application.setCurrentEvaluator(evaluator);
        }

		return value;
	}

	/**
	 * Impact of a value change of a fieldInfo.
	 *
	 * @param fieldInfo current fieldInfo associated to the form field which has been modified.
	 * @param value     new Value of the field.
	 */
	protected void propagate(LyFieldInfo fieldInfo, Object value)
	{
	}

	/**
	 * Extract current value for the specified fieldInfo in the form.
	 *
	 * @param  id fieldInfo id associated to the requested value
	 * @return current value, or null if not found or field is disabled.
	 */
	public LyValue getFieldValue(String id)
	{
		if ((getClassInfo() == null) || (id == null))
			return null;

		LyFieldInfo fieldInfo = getClassInfo().getFieldInfo(id);

		return getFieldValue(fieldInfo);
	}

	/**
	 * Extract current value for the specified fieldInfo in the form.
	 *
	 * @param  fieldInfo fieldInfo associated to the requested value
	 * @return current value, or null if not found or field is disabled.
	 */
	public LyValue getFieldValue(LyFieldInfo fieldInfo)
	{
		return ((_object == null) ? null : _object.getFieldValue(fieldInfo));
	}

	/**
	 * Gets content of this controller.
	 *
	 * @return the current object in a list
	 */
	public LyObjectList getContent()
	{
		if (_object == null)
			return null;

		return new LySimpleObjectList(_object);
	}

	/**
	 * Gets selection of this controller. The current implementation return getContent
	 *
	 * @return return the current selection as a list
	 * @see    #getContent
	 */
	public LyObjectList getSelection()
	{
		return getContent();
	}

	/**
	 * Access to object currently beeing shown in the view. Depending on nature of the controller,
	 * this object MAY be null.
	 *
	 * @return the reference to current object being shown or null if no objects are associated to
	 *         this form.
	 */
	public LyObject getObject()
	{
		return _object;
	}

	/**
	 * Set current object currently being 'shown' in the view. The method calls compute the
	 * oldValues that were displayed with the values of the current object, then call
	 * setObject(LyObject object, LyValueSet oldValues) to update the view.
	 *
	 * @param  object the reference to current object being shown or null if no objects are
	 *                associated to this form.
	 * @return true if the operation was sucessful.
	 */
	protected boolean setObject(LyObject object)
	{
		// get Old values from the current object
		LyValueSet oldValues = ((_object == null) ? null : _object.getValues());

		// FFT00808
		if (oldValues != null)
		{
			for (int i = oldValues.getSize() - 1; i >= 0; i--)
			{
				LyValue     fieldValue = oldValues.getFieldValue(i);
				LyFieldInfo fieldInfo = (LyFieldInfo)fieldValue.getField();

				if (!_classInfo.contains(fieldInfo))
					oldValues.removeValue(fieldInfo);
			}
		}

		return setObject(object, oldValues);
	}

	/**
	 * Set current object currently being 'shown' in the view. The method calls endAction,
	 * startAction then updateObject. Depending on nature of the controller, this object MAY be
	 * null.
	 *
	 * <P>This method calls endAction, then startAction and finally updateObject</P>
	 *
	 * @param  object    the reference to current object being shown or null if no objects are
	 *                   associated to this form.
	 * @param  oldValues the list of old values that were displayed inside the form.
	 * @return true if the operation was sucessful.
	 */
	protected final boolean setObject(LyObject object, LyValueSet oldValues)
	{
		// if the new object to display is equals to the old one, do nothing
		if (_object == object)
			return true;

		if (_started)
		{
			// End previous action
			if (!endAction())
				return false;

			_object = object;

			// Start new action
			if (!startAction())
			{
				free();
				return false;
			}
		}
		else
			_object = object;

        if ((_object != null)
            && _object.isFullyLoaded())
        {
            _lastUpdated = 0; // Force update
             update();
        }

		// Update form with new object values
		updateObject(oldValues);
		updateContextClient();

		// Check if object is in the objects list
		if ((object != null) && ((_objects == null) || (!_objects.contains(object))))
			setObjects(new LySimpleObjectList(object));

        if (_initialized)
            restore();

		return true;
	} // end method setObject

	/**
	 * Accessor to the list of objects associated with this form. This list may be null.
	 *
	 * @return the list of objects for this controller
	 */
	public LyObjectList getObjectList()
	{
		return _objects;
	}

	/**
	 * Accessor to the list of objects associated with this form. The user may navigate through this
	 * list of objects.
	 *
	 * @param  objects list of objects which could be shown in the form.
	 * @return true if successfull
	 */
	public boolean setObjects(LyObjectList objects)
	{
		if (objects != null)
		{
			LyInfoList<LyClassInfo> classes = objects.getClasses();
			LyClassInfo			    classInfo = getClassInfo();

			if ((classInfo != null) && classInfo.isVirtual())
				classInfo = classInfo.getRealClass();

			if ((classes != null)
					&& (classes.getSize() == 1)
					&& (classes.getInfo(0) != classInfo))
				objects = objects.getList(getSession(), classInfo);
		}

		LyObject object = (((objects == null) || (objects.getSize() == 0)) ? null
																		   : objects.getObject(0));

		if ((object != _object) && !setObject(object))
			if (object != _object)
				return false;

		// Unregister as client of workspace
		if (_objects != null)
			_objects.free(this);

		if ((objects != null) && (objects.getSize() > 0))
		{
			if (!objects.isDynamic())
			{
				_objects = new LySimpleWorkSpace(getSession(), true);
				_objects.setSorted(false);
				_objects.addList(objects);
			}
			else
				_objects = (LyWorkSpace)objects;

			// Register as client of workspace
			_objects.addListener(this);
		}
		else
			_objects = null;

		resetObjects();

		setLastModified(new Date().getTime());

		return true;
	} // end method setObjects

	/**
	 * Gets associated classInfo. The classInfo is the class of the object if any, or null if no
	 * object is being shown.
	 *
	 * @return the classInfo of current object or null if no object is specified.
	 */
	public LyClassInfo getClassInfo()
	{
		if (_classInfo != null)
			return _classInfo;

		if (_action.getClassInfo() != null)
		{
			_classInfo = _action.getClassInfo();
			return _classInfo;
		}

		if (_object != null)
		{
			_classInfo = _object.getClassInfo();
			return _classInfo;
		}

		_classInfo = super.getClassInfo();
		return _classInfo;
	}

	/**
	 * Gets image associated with this form. The image is the image associated with the classInfo of
	 * current object.
	 *
	 * @return the image if an object or a classInfo is found, null otherwise.
	 */
	public String getImage()
	{
		String image = getImage(_object);

		if (image == null)
		{
			LyClassInfo classInfo = getClassInfo();

			if (classInfo != null)
				image = getImage(classInfo);
			else
				image = super.getImage();
		}

		return image;
	}

	/**
	 * Get the name of the image (key or path) representing an object instance. Default
	 * implementation returns the icon associated to the classInfo.
	 *
	 * @param  object source object for image association.
	 * @return the image path or key code.
	 */
	public String getImage(LyObject object)
	{
		LyFormBehaviorInterface behavior = (LyFormBehaviorInterface)getBehavior();

		String				    imageKey = behavior.getImage(this, object);

		// We may handle error icon in addition to the object icon
		if (object != null)
			imageKey = getViewManager().buildErrorImage(imageKey, object.getErrorLevel());

		return imageKey;
	}

	/**
	 * Retrieves the classInfo for the specified classInfoId and the specfified RelationId.
	 *
	 * <P>Default return the classInfo specified in parameter classInfoId, but this method may be
	 * overriden when the name of the classInfo is given by another field.</P>
	 *
	 * @param  classInfoId current classInfo Identifier.
	 * @param  relationId  current field (relation field) associated to the request. *Not used* at
	 *                     this level.
	 * @return the classInfo of objects currenlty being shown in the relation field info.
	 * @see    LySetController#getClassInfo(String, String)
	 */
	protected LyClassInfo getClassInfo(String classInfoId, String relationId)
	{
		return getApplication().getClassInfo(classInfoId);
	}

	/**
	 * Set current controller context.
	 *
	 * @param context current context or null if none is specified
	 * @return true if successfull
	 * @see   LyContextInterface#getContext
	 */
	public boolean setContext(LyWorkSpace context)
	{
		LyWorkSpace previousContext = getContext();

		if ((context == previousContext)
				|| ((context != null)
					&& (previousContext != null)
					&& context.equals(previousContext)))
			return false;

		// Get all objects of _classInfo related with context.
		LyObjectList objects;

		objects = getObjectsFromContext(context);

		// Keep context listener
		LyWorkSpaceListener contextListener = _contextListener;

		// Trying to set the new objects list
		_context = context;
		_contextListener = null;
		_changingObject=true;

		if (setObjects(objects))
		{
			_changingObject = false;
			_context = previousContext;
			_contextListener = contextListener;
            return changeContext(context);
		}

		_changingObject = false;
		_context = previousContext;
        _contextListener = contextListener;
		return false;
	}

	/**
	 * Returns the objects of the form for a given context.
	 *
	 * @param  context The context.
	 * @return The objects of the form for the given context.
	 */
	protected LyObjectList getObjectsFromContext(LyWorkSpace context)
	{
		if ((context == null) || (context.getSize() == 0))
			return null;
		else if (_compositionRelation != null)
		{
			// We are in a case of a composition relation
			// If the context has one classInfo and if this classInfo contains
			// the relation we just take the content of the relation.
			LyInfoList<LyClassInfo> classes = context.getClasses();

			if ((classes != null) && (classes.getSize() == 1))
			{
				LyClassInfo classInfo = classes.getInfo(0);

				if (classInfo.contains(_compositionRelation))
				{
					int			 i, n = context.getSize();
					LyObject     object;
					LyObjectList result = null, list;

					for (i = 0; i < n; i++)
					{
						object = context.getObject(i);
						list = _compositionRelation.getObjects(object);
						if (result == null)
							result = list;
						else if (list != null)
							result.addList(list);
					}

					return result;
				}
			}
		}

		LyClassInfo classInfo = getClassInfo();

		if (classInfo == null)
			return null;

		return context.getList(getSession(), classInfo, true);
	} // end method getObjectsFromContext

	/**
	 * Set field marks required on fields added in the view of this controller.
	 *
	 * @param fieldMarks the fieldMark list
	 */
	public void setFieldMarks(BitSet fieldMarks)
	{
		_fields = null;
		_fieldMarks = fieldMarks;
	}

	/**
	 * Gets the list of fields being displayed in the form. The default implementation uses the
	 * specified fieldsMarks or if specified the lists of fields given by the setFields method.
	 *
	 * @return the list of fields being displayed
	 */
	public LyInfoList<LyFieldInfo> getFields()
	{
	    if (_action == null)
	        return null;
		if (_fields != null)
			return _fields;

		LyInfoList<LyFieldInfo> fields = _action.getFieldInfos(_classInfo, false);

		if ((fields == null) || (fields.getSize() == 0))
		{
            BitSet    fieldMarks = getFieldMarks();

            if (fieldMarks == null)
                fieldMarks = getParameterMarks("_fieldMarks", LyBitSet.Type.F_MARKS);

            if ((_classInfo != null)
                 && (fieldMarks != null))
            {
                int         m, nm = fieldMarks.size();
                int         i, n = _classInfo.getFieldInfoCount();
                LyFieldInfo fieldInfo;

                for (m = 0; m < nm; m++)
                {
                    if (!fieldMarks.get(m))
                        continue;

                    for (i = 0; i < n; i++)
                    {
                        fieldInfo = _classInfo.getFieldInfo(i);
                        if (hasMark(fieldInfo, m))
                        {
                            if (fields == null)
                                fields = new LyInfoList<LyFieldInfo>();
                            fields.addInfo(fieldInfo);
                        }
                    }
                }
            }

		    if ((fields == null) || (fields.getSize() == 0))
    		{
    			fields = getDefaultFields();
    			if (fields == null)
    				fields = new LyInfoList<LyFieldInfo>();
    		}
		}

		// Filter fields with action behavior
		LyFormBehaviorInterface behavior = (LyFormBehaviorInterface)getBehavior();

		if (behavior != null)
		{
			LyFieldInfo fieldInfo;
			int		    i, n = fields.getSize();

			for (i = n - 1; i >= 0; i--)
			{
				fieldInfo = fields.getInfo(i);
				if (!behavior.isDisplayed(this, fieldInfo))
					fields.removeInfo(fieldInfo);
			}
		}

		_fields = fields;
		return _fields;
	} // end method getFields

	/**
	 * Get default fields to display in form. This method all fields of the current class info
	 * expect hidden and secret fields.
	 *
	 * @return List of fields to be displayed.
	 */
	protected LyInfoList<LyFieldInfo> getDefaultFields()
	{
        int                     i, n = _classInfo.getFieldInfoCount();
        LyInfoList<LyFieldInfo> fields = null;
        LyFieldInfo             fieldInfo;

        for (i = 0; i < n; i++)
        {
            fieldInfo = _classInfo.getFieldInfo(i);
            if (hasMark(fieldInfo, LyFieldInfo.MARK_HIDDEN))
                // Remove hidden fields
                continue;
            if (fieldInfo.hasMark(LyFieldInfo.MARK_SECRET))
                // Remove secret fields
                continue;

            if (fields == null)
                fields = new LyInfoList<LyFieldInfo>();
            fields.addInfo(fieldInfo);
        }

        return fields;
	}

	/**
	 * Accessor to the field marks required on fields added in the view of this controller.
	 *
	 * @return fieldMarks the fieldMark list, or null if none specified
	 */
	public BitSet getFieldMarks()
	{
		return _fieldMarks;
	}

	/**
	 * Utility method to know if a fieldInfo is displayed in this form or not. Default
	 * implementation check if the specified fieldInfo has the marks requested in getFieldMark. If
	 * no marksList is specified (null), the method returns true.
	 *
	 * @param  fieldInfo the fieldInfo which is tested for addition in the view.
	 * @return true if the field should be added, false if it should be skipped.
	 */
	public boolean isDisplayed(LyFieldInfo fieldInfo)
	{
		LyInfoList<LyFieldInfo> fields = getFields();

		if ((fields == null)
		        || fields.contains(fieldInfo))
		    return true;

		LyFieldInfo			    field;
		int					    i, n = ((fields == null) ? 0 : fields.getSize());

		// Retrieve the array fields and check if their field info is the given field info.
		for (i = 0; i < n; i++)
		{
			field = fields.getInfo(i);
			if ((field instanceof LyArrayFieldInfo)
					&& (((LyArrayFieldInfo)field).getFieldInfo() == fieldInfo))
				return true;
		}

		return false;
	}

	/**
	 * Indicates if the view for this form controller is resizable.
	 *
	 * @return false since form views are not resizable.
	 */
	public boolean isViewResizable()
	{
		LyViewManager viewManager = getViewManager();

		return ((viewManager != null) && viewManager.hasProperty("ALLOW_FORM_RESIZE"));
	}

	/**
	 * Checks if the given fieldInfo is present in the controlled form. The field is present if a
	 * formField was created
	 * @param fieldInfo the field to check
	 * @return true if the field is added (shown) in the form, false otherwise.
	 */
	public boolean contains(LyFieldInfo fieldInfo)
	{
		if (fieldInfo == null)
			return false;

		return ((isDisplayed(fieldInfo)
					&& ((getFormField(fieldInfo) != null)
						|| (getAreaController(fieldInfo) != null)))
				|| (getFieldController(fieldInfo) != null));
	}

	/**
	 * Result of a selection demand : set current selection (in the tool bar) as the current object.
	 */
	protected void selectObject()
	{
		LyObject object = _objects.getObject(0);

		setObject(object);
	}

	/**
	 * Update the value of a fieldInfo shown in the view. The method sets the value of the
	 * associated fieldItem.
	 *
	 * @param fieldInfo the fieldInfo which should be updated in the form
	 * @param newValue  the new value of the field.
	 * @param oldValue  former value of the field,or null if unknown.
	 * @param propagate Indicates whether the change is notified (propagated) to other fields or
	 *                  not.
	 * @see   #getFieldItem
	 */
	protected void updateFieldItem(LyFieldInfo fieldInfo, Object newValue,
		Object oldValue, boolean propagate)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController != null)
		{
			fieldController.updateFieldItem(newValue, oldValue, propagate);
			setLastModified(new Date().getTime());
		}
	}

	/**
	 * Accessor to the formField (View component) associated to the fieldInfo. The formField
	 * contains a caption (label of the fieldInfo), a fieldItem (value of the fieldInfo) and the
	 * constraints definition (how the component are shown in the view).
	 *
	 * @param  fieldInfo the fieldInfo which is associated to the view component.
	 * @return the formField corresponding to the given fieldInfo
	 */
	public LyFormField getFormField(LyFieldInfo fieldInfo)
	{
		if (_builder == null)
			return null; // Not yet built

		LyFieldController fieldController = getFieldController(fieldInfo);

		return ((fieldController == null) ? null : fieldController.getFormField());
	}

	/**
	 * Accessor to the fieldItem (View component) associated to the fieldInfo. This is a shortcut
	 * for getFormField().getFieldItem()
	 *
	 * @param  fieldInfo the fieldInfo which is associated to the view component.
	 * @return the fieldItem corresponding to the given fieldInfo
	 * @see    #getFormField to access directly to the parent view structure.
	 */
	protected LyFieldItem getFieldItem(LyFieldInfo fieldInfo)
	{
		return getFieldItem(fieldInfo, null);
	}

	/**
	 * Accessor to the fieldItem (View component) associated to the fieldInfo.
	 *
	 * @param  fieldInfo      FieldInfo which is associated to the view component.
	 * @param  fieldItemClass Expected class for the returned field item.
	 * @return the fieldItem corresponding to the given fieldInfo and to the expected class, null
	 *         otherwise.
	 */

	protected LyFieldItem getFieldItem(LyFieldInfo fieldInfo, Class<?> fieldItemClass)
	{
		// Retrieve fieldItem corresponding to fieldInfo
		LyFormField formField = getFormField(fieldInfo);

		if (formField == null)
			return null;

		LyFieldItem fieldItem = formField.getFieldItem();

		if ((fieldItemClass != null)
				&& (fieldItem != null)
				&& (fieldItem.getClass() != fieldItemClass))
		{
			// Warns user and return null
			Exception e = new IllegalStateException(
					"BF AppEngine : component corresponding to field info "
					+ fieldInfo + " (class " + fieldItem.getClass()
					+ ") doesn't match the expected class " + fieldItemClass
					+ ", returning null");

			getEnvironment().writeException(e);
			return null;
		}

		return fieldItem;
	}

	/**
	 * Update the toolbar content.
	 */
	protected void updateTools()
	{
		if (_component == null)
			return;

		LyArea area = (LyArea)_component;
		int    nbToolbar = area.getToolBarCount();

		if (nbToolbar == 0)
			return;

		// get all the actions of the toolbar
		LyAction			 rootAction = getAction();
		LyInfoList<LyAction> actions = new LyInfoList<LyAction>();

		if (rootAction != null)
			actions.addList(rootAction.getActions()); // add sub-actions of the root action

		LyClassInfo classInfo = getClassInfo();

		if (classInfo.getActionCount() > 0)
			actions.addList(classInfo.getActions()); // add actions of the class info

		LySession session = getSession();
		LyAction  action;
		LyTool    tool;
		LyToolBar toolBar;
		int		  i, k, n = actions.getSize();
		boolean   enabled;

		// we update all the toolbars of the area
		for (k = 0; k < nbToolbar; k++)
		{
			toolBar = area.getToolBar(k);

			for (i = 0; i < n; i++)
			{
				action = actions.getInfo(i);
				if ((action.getMin() > 1) // Needs more than one object
						|| (action.getMax() == 0)) // Asks for no selection
					continue;

				tool = (LyTool)toolBar.getComponent(action.getId());
				if (tool != null)
				{
					enabled = session.enableAction(action, this,
							((_object == null) ? null : new LySimpleObjectList(_object)), false);
					tool.setEnabled(enabled);
				}
			}
		}

        LyEnvironment environment = _parent.getEnvironment();

        if (!environment.testResource("LY_NAVIGATION_IN_FORMS"))
            return;

		toolBar = area.getToolBar(0);
		if (toolBar == null)
		    return;

        LyTool      previousTool = (LyTool)toolBar.getComponent(DEMAND_PREVIOUS_OBJECT);
        LyTool      nextTool = (LyTool)toolBar.getComponent(DEMAND_NEXT_OBJECT);
        LyController grandParent = getViewController().getParent();

        if (!(grandParent instanceof LySimpleTableController))
            return;
        LyObjectList content = ((LySimpleTableController)grandParent).getContent();
        int     no = (content == null ? 0 :content.getSize());

        if (previousTool != null)
            previousTool.setEnabled(no > 1);
        if (nextTool != null)
            nextTool.setEnabled(no > 1);
	} // end method updateTools

	/**
	 * Update the context component (showing the objectList) when an object is changed/removed.
	 *
	 * @param object current object from the current object list.
	 * @see   #getObjectList()
	 */
	protected void updateObjects(LyObject object)
	{
		LyComponent component = getComponent();

		if (!LyFormBuilder.SHOW_CONTEXT_AREA)
		{
			LyFormField formField = (LyFormField)component.getComponent("_formObjects");

			if (formField == null)
				return;

			LyHRefs hRefs = (LyHRefs)formField.getFieldItem();
			LyHRef  hRef;
			int     i, n = hRefs.getHRefCount();
			String  id = object.getId();

			for (i = 0; i < n; i++)
			{
				hRef = hRefs.getHRef(i);
				if (hRef.getId().equals(id))
				{
					if (!_objects.contains(object))
						hRefs.removeHRef(hRef);
					else
						hRef.setLabel(getName(object));
					return;
				}
			}
		}
		else
		{
			LyForm form = (LyForm)component.getComponent("_formContext");

			if (form == null)
				return;

			int			 i, n = form.getFormFieldCount();
			LyFormField	 formField;
			LyHRef		 hRef;
			String		 id = object.getId();

			// Remove all current fields.
			for (i = 0; i < n; i++)
			{
				formField = form.getFormField(i);
				hRef = (LyHRef)formField.getFieldItem();
				if (hRef.getId().equals(id))
				{
					if (!_objects.contains(object))
					{
						form.removeFormField(i);
						form.resetContent();
					}
					else
						hRef.setLabel(getName(object));
					return;
				}
			}
		}
	}

	/**
	 * Update view when resetting the objects list.
	 */
	protected void resetObjects()
	{
		if (!_initialized)
			return;

		updateContext();
		updateTitleBar();
	}

	/**
	 * Update context.
	 */
	protected void updateContext()
	{
		if (_component == null)
			return;

		// Getting hrefs
		LyArea		 area = (LyArea)_component;
		LyObjectList list = getObjectList();
		int			 no = ((list == null) ? 0 : list.getSize());

		if (!LyFormBuilder.SHOW_CONTEXT_AREA)
		{
			LyFormField formField = (LyFormField)area.getComponent("_formObjects");

			if (formField == null)
				return;

			LyHRefs		 hRefs = (LyHRefs)formField.getFieldItem();
			LyHRef		 hRef;
			int			 i, nh = hRefs.getHRefCount();

			// Remove all current hrefs.
			for (i = nh - 1; i >= 0; i--)
			{
				hRef = hRefs.getHRef(i);
				hRefs.removeHRef(hRef);
			}

			// Add new hrefs from the list.
			LyObject object;

			for (i = 0; i < no; i++)
			{
				object = list.getObject(i);
				hRef = new LyHRef();

				hRef.setId(object.getId());
				hRef.setLabel(getName(object));
				hRef.setImage(getImage(object));

				hRefs.addHRef(hRef);
			}
		}
		else
		{
			LyForm form = (LyForm)area.getComponent("_formContext");

			if (form == null)
				return;

			int			 i, nff = form.getFormFieldCount();
			LyFormField	 formField;
			LyHRef		 href;

			// Remove all current fields.
			for (i = nff - 1; i >= 0; i--)
			{
				form.removeFormField(i);
			}

			// Add new hrefs from the list.
			LyObject object;

			for (i = 0; i < no; i++)
			{
				object = list.getObject(i);
				href = new LyHRef();

				href.setId(object.getId());
				href.setLabel(getName(object));
				href.setImage(_parent.getImage(object));

				formField = new LyFormField("_formObjects" + i);
				formField.setFieldItem(href);
				formField.setEditable(false);
				form.addFormField(formField);
			}

			if ((nff > 0) || (no > 0))
				form.resetContent();
		}

		LyToolBar toolBar = area.getToolBar();

		if (toolBar != null)
		{
			LyTool previous = (LyTool)toolBar.getComponent(DEMAND_PREVIOUS_OBJECT);
			LyTool next, selection;

			if ((previous == null) && (no > 1))
			{
				LyBuilder builder = getBuilder();

				if (builder instanceof LyFormBuilder)
					((LyFormBuilder)builder).completeToolBarForContext(toolBar);
				else
					((LyFormBuilder)getDefaultBuilder()).completeToolBarForContext(toolBar);

				previous = (LyTool)toolBar.getComponent(DEMAND_PREVIOUS_OBJECT);
			}

			next = (LyTool)toolBar.getComponent(DEMAND_NEXT_OBJECT);
			selection = (LyTool)toolBar.getComponent(DEMAND_SELECTION);

			if (previous != null)
				previous.setEnabled(no > 1);
			if (next != null)
				next.setEnabled(no > 1);
			if (selection != null)
				selection.setEnabled(no > 1);
		}
	} // end method updateContext

	/* Implementation of LyWorkSpaceListener */

	/**
	 * Called when an object is added in the dynamic object list. Default implementation does
	 * nothing, and returns false
	 *
	 * @param  event event of adding object
	 * @return the event is used by the listener
	 */
	public boolean objectAdded(LyWorkSpaceEvent event)
	{
		// Nothing to do
		return false;
	}

	/**
	 * Called when an object is removed from the dynamic object list. Keep content consistent : go
	 * to next available object if any or close the view.
	 *
	 * @param  event event of removing object
	 * @return the event is used by the listener
	 */
	public boolean objectRemoved(LyWorkSpaceEvent event)
	{
		if ((_objects == null) || (_objects.getSize() == 0))
		{
			if (_parent instanceof LyViewController)
			{
				showWarning("LY_FORM_CLOSED_OBJECT_DELETED");
				free();
			}
			else
				setObjects(null);
			return true;
		}

		LyObject object = event.getObject();

		if (_object == object)
			nextObject(true);

		updateObjects(object);
		return true;
	}

	/**
	 * Called when an object has changed in the dynamic object list. Keep the content consistent and
	 * invoke when needed updateObjects and updateObject.
	 *
	 * @param  event event of changing object
	 * @return the event is used by the listener
	 */
	public boolean objectChanged(LyWorkSpaceEvent event)
	{
		LyObject object = event.getObject();

		if (object != _object)
		{
			// Update context if object name changes
			updateObjects(object);
			return false;
		}

		updateObject(event.getValues());
		return true;
	}

	/**
	 * Called when an object state is changed.
	 *
	 * @param  event Event notifying object state change.
	 * @return True if the event is used by the listener, false otherwise.
	 */
	public boolean objectStateChanged(LyAdvancedWorkSpaceEvent event)
	{
		// We must refresh the toolbar icon (an error occured)
		updateTitleBar();

		return true;
	}

	/**
	 * Called when an object user event occured.
	 *
	 * @param  event Event notifying object user event.
	 * @return True if the event is used by the listener, false otherwise.
	 */
	public boolean objectUserEvent(LyAdvancedWorkSpaceEvent event)
	{
		return false;
	}

	/**
	 * Called when grouped events are sent.
	 *
	 * @param  events Sorted list of events.
	 * @return True if the events are used by the listener, false otherwise.
	 */

	public boolean workSpaceEventsGrouped(ArrayList<LyWorkSpaceEvent> events)
	{
		if ((events == null) || (events.isEmpty()))
			return false;

		int					  i, n = events.size();
		LyWorkSpaceEvent	  event;
		LyWorkSpaceEvent.Type type;
		boolean				  used = false;

		for (i = 0; i < n; i++)
		{
			event =events.get(i);
			type = event.getType();
			switch (type)
			{
			case ADD:
				used |= objectAdded(event);
				break;
			case CHANGE:
				used |= objectChanged(event);
				break;
			case REMOVE:
				used |= objectRemoved(event);
				break;
			default:
				break;
			}
		}

		return used;
	} // end method workSpaceEventsGrouped

	/**
	 * Gets the specific areaController for the given field info.
	 *
	 * @param  fieldInfo Field info for which an area controller is requested.
	 * @return Area controller corresponding to the given field info, null in most cases.
	 */
	public LyAreaController getAreaController(LyFieldInfo fieldInfo)
	{
		if (_builder == null)
			return null; // Not yet built

		LyFieldController fieldController = getFieldController(fieldInfo);

		return ((fieldController == null) ? null : fieldController.getAreaController());
	}

	/**
	 * Reloads area controller content.
	 */
	protected void reload()
	{
		if (_object == null)
			return;

		LyObjectSource objectSource = _session.getObjectSource(_classInfo);

		objectSource.reloadObject(_object.getObjectId());
	}

	/**
	 * Tests if the specified fieldInfo must be represented as a composition relation in forms.
	 *
	 * @param  fieldInfo
	 * @return true if the fieldinfo is a composition field
	 */
	public boolean isComposition(LyFieldInfo fieldInfo)
	{
		if ((fieldInfo instanceof LyRelationFieldInfo)
		    && (((LyRelationFieldInfo)fieldInfo).getType() == LyRelationFieldInfo.Type.COMPOSITION))
			return true;
		else if (fieldInfo instanceof LyArrayFieldInfo)
		{
			fieldInfo = ((LyArrayFieldInfo)fieldInfo).getFieldInfo();
			return isComposition(fieldInfo);
		}
		return false;
	}

	/**
	 * Reset form content.
	 */
	public void resetFormContent()
	{
		// Reset the form
		LyComponent root = getComponent();
		LyComponent form = ((root == null) ? null : root.getComponent(getId(), true));

		if (form instanceof LyForm)
			((LyForm)form).resetContent();
	}

	/**
	 * Sets the valid status of a field in the form.
	 *
	 * @param fieldInfo the specified field
	 * @param status    current validity status
	 */
	public void setState(LyFieldInfo fieldInfo, short status)
	{
		if (status < 0)
			return;

		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController == null)
			return;

		if (getState(fieldInfo) != status)
		{
			getEnvironment().logTrace("\tsetState " + fieldInfo + " : " + status);
			fieldController.setState(status);

			if ((status != LyClassBehavior.STATUS_OK)
					&& (status <= LyClassBehavior.STATUS_WARNING))
				setActiveField(fieldInfo);
		}
	}

	/**
	 * Accessor to the state for a field in the form.
	 *
	 * @param  fieldInfo the specified field
	 * @return current field state
	 */
	public short getState(LyFieldInfo fieldInfo)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

        return (fieldController != null ? fieldController.getState() : LySetBehaviorInterface.STATUS_OK);
	}

	/**
	 * Indicates if the controller is changing its object.
	 * @return true if the object is changing
	 */
	public boolean isChangingObject()
	{
		return _changingObject;
	}

	/**
	 * Internal use. Sets the form content from the specified object : all field controllers and
	 * form fields take the specified value. Any value modified by the user is the form is lost.
	 * Unlike setObject() and updateFielditem(), no control is done to prevent overriding existing
	 * values.
	 *
	 * @param object         new object used to reset the view content.
	 * @param freeOldValue   indicates if the old value should be freed (deleted) when it is an
	 *                       object
	 * @param createNewValue indicates if a new empty value should created when object is null and
	 *                       when the value is an object
	 */
	public void resetFromObject(LyObject object, boolean freeOldValue, boolean createNewValue)
	{
		// The goal of this method is to be able to change recursively the values in the form by code
		// Do not add complex logic in this method
		_object = object;

		// Now we update the fieldControllers values
		LyClassInfo		  classInfo = getClassInfo();
		int				  i, n = classInfo.getFieldInfoCount();
		LyFieldInfo		  fieldInfo;
		LyFieldController fieldController;
		Object			  value;

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			fieldController = getFieldController(fieldInfo);

			if (fieldController == null)
				continue;

			value = ((object == null) ? null : object.getValue(fieldInfo));
			fieldController.resetFromValue(value, freeOldValue, createNewValue);
		}
	} // end method resetFromObject

	/**
	 * Find recursively the formField corresponding to the given fieldInfo.
	 *
	 * @param  fieldInfo the fieldInfo which is associated to the view component.
	 * @return the formField corresponding to the given fieldInfo
	 */
	public LyFormField findFormField(LyFieldInfo fieldInfo)
	{
		if (_builder == null)
			return null; // Not yet built

		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController != null)
			return fieldController.getFormField();

		Enumeration<LyFieldController>    elements = _fieldControllers.elements();
		LyAreaController                  areaController;
		LyFormField		                  formField;

		while (elements.hasMoreElements())
		{
			fieldController = elements.nextElement();
			areaController = fieldController.getAreaController();
			if (areaController instanceof LyFormController)
			{
				formField = ((LyFormController)areaController).findFormField(fieldInfo);
				if (formField != null)
					return formField;
			}
		}

		return null;
	} // end method findFormField

	/**
	 * Accessor to the current Environment instance (shortcut for getSession().getEnvironment()).
	 *
	 * @return current environment instance.
	 */
	public LyEnvironment getEnvironment()
	{
		if (_environment != null)
			return _environment;

		if ((_parent != null) && !(_parent instanceof LyViewController))
		{
			LyEnvironment parentEnvironment = _parent.getEnvironment();

			_environment = _action.getEnvironment(parentEnvironment);
		}
		else
			_environment = super.getEnvironment();
		return _environment;
	}

	/**
	 * Reset all the content of the view.
	 *
	 * @param full Boolean indicating if a full reset (including reset of datas) is requiered or not
	 *             (only reset graphical area).
	 */
	public void reset(boolean full)
	{
		super.reset(full);

		if (_initialized)
		{
			updateTitleBar();
			updateContext();
		}
	}

	/**
	 * Updates the view according to the area controllers stack and restore the last active field.
	 * @see LyMultiAreaController#updateAreaStack(LyAreaController)
	 */
	protected void updateAreaStack(LyAreaController displayedAreaController)
	{
		super.updateAreaStack(displayedAreaController);
		getViewManager().execGui(new _LyFormControllerUpdater());
	}

	class _LyFormControllerUpdater
		implements Runnable, LyGetSessionInterface
	{
		public void run()
		{
			restoreLastActiveField();
		}

		public LyDataSession getSession()
		{
			return LyFormController.this.getSession();
		}
	}

    /**
     * Retrieves a value from a registered key.
     *
     * @param   key  The key that must be used.
     * @return  The value stored with the key or null if the key is not registered.
     */
    public Object getKeyValue(String key)
    {
        LyClassInfo     	classInfo = getClassInfo();
        LyObject			object = getObject();
        LyFieldInfo			fieldInfo;
        LyFieldController	fieldController;

        while (classInfo != null)
        {
            if (key.equalsIgnoreCase(classInfo.getId()))
                return this;

    		fieldInfo = getClassInfo().getFieldInfo(key, false);
    		if (fieldInfo != null)
    		{
    			fieldController = getFieldController(fieldInfo);
    			if (fieldController instanceof LyCompositionFieldSetController)
    				return fieldController.getAreaController();

    			if (fieldController != null)
    				return fieldController.getFieldValue();

    			if (object != null)
    				return object.getValue(fieldInfo);
    		}

            classInfo = classInfo.getSuperClass();
        }

        return super.getKeyValue(key);
    }

    /**
     * Show an error for the specified object.
     *
     * @param object The object in error
     * @param error  The error to show
     */
    public void showError(LyObject object, LyError error)
    {
        LyEnvironment environment = getEnvironment();
        String		  resource = environment.getResource("LY_FORM_SHOW_ERRORS_IN_FORM");
        boolean		  showErrorsOnField = "field".equals(resource);
        LyFieldInfo	  fieldInfo = error.getOrigin();

        if (showErrorsOnField && (fieldInfo != null))
        {
            LyFormField formField = getFormField(fieldInfo);

            if (formField != null)
            {
                String        message = error.getMessage(environment);

                formField.setMessage(message);
            }
        }

        if (!showErrorsOnField)
        	super.showError(object, error);
    } // end method showError

    /**
     * Restore form inputs.
     */
    protected boolean restore()
    {
        return false;
    }

	/**
	 * Get the default unit for the given fieldInfo.
	 * This method ask the session behavior to compute this unit.
	 * If the session behavior returns null, it will return the reference unit.
	 * @param fieldInfo a fieldInfo
	 * @return a unit (can be null)
	 */
	public LyUnit getDefaultFieldUnit(LyFieldInfo fieldInfo)
	{
		LySession session = getSession();
		LyDataSessionBehaviorInterface sessionBehaviorInterface = session.getSessionBehavior();

		LyUnit unit = sessionBehaviorInterface.getDefaultUnitForField(session, fieldInfo);

		if (unit == null)
			unit = fieldInfo.getFieldUnit();
		return unit;
	}
} // end class LyFormController
