/**
 * LyFrameController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyFrameController.java 41094 2012-10-09 16:21:57Z smaupin $
 */

package leon.control;

import leon.app.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.info.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * This controller manages a frame window (or view). This Frame is used as the view controller of
 * other views (such as creation/modification/consultation forms). A Frame usually contains a
 * 'major' view such as tables, maps, trees, not an edition form. However, the content is defined in
 * the internal area.
 *
 * <P>Unlike a Dialog, a frame is a main window which may or may not have a parent window. If the
 * frame has no parent, this is a 'root' level Frame. A Frame has a MenuBar, and usually contains a
 * titlebar and toolbar.</P>
 *
 * <P>The LyFrameController automatically calculates the window position and size.</P>
 *
 * <P>The other view controller is LyDialogController.</P>
 *
 * @see LyViewController
 * @see LyDialogController
 */
public class LyFrameController extends LyViewController
{
	/** File Version */
	public static final String __VERSION =
		"$Id: LyFrameController.java 41094 2012-10-09 16:21:57Z smaupin $";

	/**
	 * Default X value for the offset (relative to parent) to determine view position when this
	 * frame is opened.
	 */
	public static int		   VIEW_XOFFSET = 20;

	/**
	 * Default Y value for the offset (relative to parent) to determine view position when this
	 * frame is opened.
	 */
	public static int		   VIEW_YOFFSET = 20;

	/** Identifier of a action in the view : Saving data. */
	public static final String DEMAND_SAVE_DATA = "_saveData".intern();

	/** Identifier of a action in the view : Loading data. */
	public static final String DEMAND_LOAD_DATA = "_loadData".intern();

	/** Identifier of a action in the view : Run garbage collector. */
	public static final String DEMAND_RUN_GC = "_runGC".intern();

	/** Indicates that the bounds (size / location) are still unset for this view. */
	protected boolean		   _boundsSet = false;

	/** The last known location when the view is cleared. */
	protected int[]			   _lastLocation = null;

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller : this parameter may be null for 'root' controller
	 */
	public LyFrameController(LySession session, LyController parent)
	{
		this(session, parent, null);
	}

	/**
	 * Constructor.
	 *
	 * @param session        current session associated to the controller. The controller
	 *                       automatically registers on the specified session.
	 * @param parent         parent controller : this parameter may be null for 'root' controller
	 * @param areaController the area which is inside the dialog : an area automatically builds a
	 *                       view controller (dialog or frame) when showView is called.
	 */
	public LyFrameController(LySession session, LyController parent,
		LyAreaController areaController)
	{
		super(session, parent, areaController);

		if (areaController != null)
			setName(areaController.getName());

		LyApplication application = session.getApplication();
		Class<?>	  builderClass = application.loadClass("leon.control.builder.LyFrameBuilder");

		setBuilderClass(builderClass);
	}

	/**
	 * Invoked when an action occurs. At this level, the implementation manages
	 *
	 * <UL>
	 *   <LI>HREF demands. demand.getSource() = DEMAND_SAVE_DATA</LI>
	 *   <LI>HREF demands. demand.getSource() = DEMAND_RUN_GC</LI>
	 *   <LI>HREF demands. demand.getSource() = DEMAND_HELP</LI>
	 *   <LI>HREF demands. demand.getSource() = LyViewController.DEMAND_ABOUT</LI>
	 *   <LI>HREF demands. demand.getSource() = LyViewController.DEMAND__version</LI>
	 * </UL>
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			String id = demand.getSource().intern();

			if (id != null)
				id = id.intern();

			if (id == DEMAND_SAVE_DATA)
			{
				LyApplication application = getApplication();

				application.saveData();
				return;
			}
			else if (id == DEMAND_LOAD_DATA)
			{
				LyApplication application = getApplication();

				application.loadData();
				return;
			}
			else if (id == DEMAND_RUN_GC)
			{
				System.runFinalization();
				System.gc();
				return;
			}
			else if (id == LyViewController.DEMAND_ABOUT)
			{
				// The session will handle this
				LySession session = getSession();

				session.processAboutAction(this);
				return;
			}
			else if (id == LyViewController.DEMAND_LOG)
			{
				// The session will handle this
				LySession session = getSession();

				session.processLogAction(this);
				return;
			}
			else if (id == LyViewController.DEMAND_VERSION)
			{
				// The session will handle this
				LySession session = getSession();

				session.processVersionAction(this);
				return;
			}
			else if (id.startsWith(DEMAND_HELP))
			{
//                               The session will handle this
				LySession session = getSession();

				session.processHelpAction(this);
				return;

			}
			else if ((id == LyAction.ID_UNDO)
					|| (id == LyAction.ID_REDO))
			{
				LySession     session = getSession();
				LyApplication application = session.getApplication();
				LyAction	  action = application.getAction(id);

				session.invokeAction(this, action, null, null, false);
				return;
			}
		}
		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Clears all cached data for this controller (especially the view and the component).
	 *
	 * <P>This method may be invoked for memory optimization, but the controller is still valid and
	 * may be used. (Unlike free) Further usage, needs to rebuild internal component & view which
	 * may be time consuming.</P>
	 */
	public void clearViewCache()
	{
		if (_component != null)
		{
			LyView view = (LyView)_component;
			int[]  p = view.getLocation();

			super.clearViewCache();

			if (!_viewInitialized)
				_lastLocation = p;
			else
				_lastLocation = null;
		}
		else
			super.clearViewCache();
	}

	/**
	 * Make the view visible to the user. Automatically calculate location on screen If the display
	 * is a fat client (windows (awt) / swing) this methods should open a window and show the view.
	 * In the case of thin client (html), this prepares the pages which will be returned and
	 * displayed in the client browser.
	 */
	public void showView()
	{
		LyView		  view = getView();
		LyViewManager viewManager = getViewManager();

		if (viewManager instanceof LyAdvancedViewManagerInterface)
			((LyAdvancedViewManagerInterface)viewManager).handleShowView(this);

		if (!_viewInitialized)
		{
			initView();

			updateUndoRedo();
		}

		// Check that this controller was not freed (initialisation failure)
		if (isFreed())
		{
			// If the view is not null, free it
			if (view != null)
			{
				view.free();
				view = null;
			}

			return;
		}

		// If view exists and is not visible, handle geometry
		if ((view != null) && !view.isVisible())
		{
			LyAreaController areaController = getAreaController();
			LyAction		 action = areaController.getAction();
			LyGeometry		 geometry = action.getGeometry();

			boolean isPortlet = getViewManager().isPortletMode();

			if (action.getTarget() != null && !isPortlet)
				((LyFrame)view).setTarget(action.getTarget());
			else if (((((LyFrame)view).getTarget() == null)
			    && getViewManager().hasProperty("POPUP_VIEW")
			    && action.hasMark(LyAction.MARK_POPUP)) || isPortlet)
			{
				((LyFrame)view).setTarget(LySession.POPUP_VIEW);
				getViewManager().setCurrentWindowName(LySession.POPUP_VIEW);
			}

			// set location according to parent location
			if (_lastLocation != null)
			{
				view.setLocation(_lastLocation[0], _lastLocation[1]);
				_boundsSet = true;
				_lastLocation = null;
			}
			else if ((geometry != null)
					&& (geometry.getX() >= 0)
					&& (geometry.getY() >= 0))
			{
				view.setLocation(geometry.getX(), geometry.getY());
				_boundsSet = true;
			}
			else if (!_boundsSet && (_parent != null))
			{
				// Attemp to find parent location on screen
				LyViewController parentViewController = _parent.getViewController();
				LyView			 parentView = ((parentViewController == null)
						? null : parentViewController.getView());

				// FFT00435/Flandres
				if (parentView != null)
				{
					int[] pt = null;
					int[] dim, screen;

					if ((VIEW_XOFFSET > 0) && (VIEW_YOFFSET > 0))
					{
		                int   nsons = parentViewController.getSonsCount();
	                    int[] pdim = parentView.getSize();
	                    int[] ppos = parentView.getLocation();

    					// Show the view under the parent window1
    					if ((ppos != null) && (pdim != null))
    						pt = new int[]
    							{
    								ppos[0] + (VIEW_XOFFSET * (nsons - 1)),
    								ppos[1] + (VIEW_YOFFSET * (nsons - 1))
    							};
					}

					dim = view.getSize();
					screen = getEnvironment().getScreenSize(parentView);

					if (screen == null)
						screen = new int[] { 0, 0, 0, 0 };

					if (dim == null)
						dim = new int[] { 0, 0 };

					if (pt == null)
						// Center the Frame
						pt = new int[]
							{
								Math.max(0, (screen[0] - dim[0]) / 2) + screen[2],
								Math.max(0, (screen[1] - dim[1]) / 2) + screen[3]
							};

					if ((pt[0] + dim[0]) > (screen[0] + screen[2]))
						pt[0] = (VIEW_XOFFSET < 0 ? screen[2] : VIEW_XOFFSET + screen[2]);
					if ((pt[1] + dim[1]) > screen[1] + screen[3])
						pt[1] = (VIEW_XOFFSET < 0 ? screen[3] : VIEW_YOFFSET + screen[3]);

					view.setLocation(pt[0], pt[1]);
					_boundsSet = true;
				}

				// FFT 00089 IPBOSS only for one screen
				int screen[] = getEnvironment().getScreenSize(parentView);

				if ((screen != null) &&
					((screen[2] == 0)
					&& (screen[3] == 0)))
				{
					int location[] = view.getLocation();
					int viewSize[] = view.getSize();

					if ( (viewSize != null)
							&& (location != null) )
					{
						int w = viewSize[0];
						int h = viewSize[1];
						boolean modifySize = false;

						if ( (location[0] + viewSize[0]) > screen[0])
						{
							w = screen[0] - location[0];
							modifySize = true;
						}
						if ( (location[1] + viewSize[1]) > screen[1])
						{
							h = screen[1] - location[1];
							modifySize = true;
						}

						if (modifySize)
							view.setSize(w, h);

					}
				}
			}
		}

		super.showView();

		_updated = false;
	} // end method showView

	/**
	 * @see leon.control.LyViewController#demandPerformed(leon.view.event.LyDemand)
	 */
	public void demandPerformed(LyDemand demand)
	{
		LySession session = getSession();
		// outline
		if (demand instanceof LyActivateViewDemand)
		{
			LyAreaController areaController = getAreaController();

			if (areaController.isOutlined())
				session.notifyOutline(areaController);
		}
		super.demandPerformed(demand);
	}

	/**
	 * @see leon.control.LyViewController#exit()
	 */
	public void exit()
	{
		getSession().exit();
	}

	/**
	 * Update the undo redo stack
	 */
	public void updateUndoRedo()
	{
		if (_component == null)
			return;

		LyViewManager viewManager = getViewManager();
		boolean		  hasMenu = viewManager.hasProperty("MENU");

		if (!hasMenu)
			return;

		LyView    view = getView();
        LyMenuBar menuBar = (view == null ? null : ((LyFrame)view).getMenuBar());

		if (menuBar == null)
			return;

		// Update items
		LyItem			 item;

		// update the menu item for the  status sort
		LyApplication    application = getApplication();
		LySession		 session = getSession();
		LyContextManager cm = session.getContextManager();
		LyAction		 action;
		boolean			 enable;
		String			 actionName, name;

		action = application.getAction(LyAction.ID_UNDO);
		item = menuBar.findItem(action.getId(), getId());
		if (item != null)
		{
			enable = cm.isUndoAvailable();
			actionName = action.getName();
			name = cm.getUndoName();
			if (name != null)
			{
				actionName += " : ";
				actionName += name;
			}
			item.setLabel(actionName);
			item.setEnabled(enable);
		}

		action = application.getAction(LyAction.ID_REDO);
		item = menuBar.findItem(action.getId(), getId());
		if (item != null)
		{
			enable = cm.isRedoAvailable();
			actionName = action.getName();
			name = cm.getRedoName();
			if (name != null)
			{
				actionName += " : ";
				actionName += name;
			}
			item.setLabel(actionName);
			item.setEnabled(enable);
		}
	} // end method updateUndoRedo

	/* ------------------ Accessors ----------------- */

	/**
	 * Factory of the default builder for a controller. Overriden to create an instance of
	 * LyFrameBuilder
	 *
	 * @return the builder which creates the XML description of the view
	 * @see    leon.control.builder.LyDialogBuilder
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return (new LyFrameBuilder(this));
	}
} // end class LyFrameController
