/**
 * LyGraphicalChartController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyGraphicalChartController.java 41109 2012-10-10 12:20:36Z jlsaouli $
 */

package leon.control;

import java.io.*;
import java.text.*;
import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.data.cache.*;
import leon.info.*;
import leon.info.infointerface.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;


/**
 * This class implements the graphical chart area. This controller is associated with a builder
 * LyGraphicalChartBuilder). It is also related to a behavior that implements
 * LyChartBehaviorInterface. Usually, this behavior is an instance of LyChartBehavior.
 *
 * @see LyGraphicalChartBuilder
 * @see leon.app.behaviorinterface.LyChartBehaviorInterface
 * @see leon.app.behavior.LyChartBehavior
 */
public class LyGraphicalChartController extends LyAreaController
	implements Runnable,
		LyPrintListenerInterface
{
	/** File Version */
	public static final String  __VERSION =
		"$Id: LyGraphicalChartController.java 41109 2012-10-10 12:20:36Z jlsaouli $";

	/** Saving the chart image. */
	public static final String  DEMAND_PRINT_CHART = "_printChart".intern();

	/** Saving the chart image. */
	public static final String  DEMAND_SAVE_IMG_CHART = "_saveImgChart".intern();

	/** Saving the chart definition. */
	public static final String  DEMAND_SAVE_DEF_CHART = "_saveDefChart".intern();

	/** View chart data. */
	public static final String  DEMAND_CHART_DATA = "_chartData".intern();

	/** Application end. */
	public static final String  DEMAND_OVERWRITE_FILE = "_overwriteFile".intern();

	/** Class info. */
	protected LyClassInfo	    _classInfo;

	/** Current chart info. */
	protected LyChartInfo	    _chartInfo;

	/** A handle on the parameter windows Controller. */
	protected LyChartController _chartController;

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo always the classInfo of the specified objects.
	 * @param objects   current context (not null because the view is contextual to a list of
	 *                  objects).
	 * @param chartController the parent chart controller
	 */
	public LyGraphicalChartController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects, LyChartController chartController)
	{
		super(session, parent, action, objects);

		// If a classInfo is given in parameter, take it for the chart
		_classInfo = (LyClassInfo)action.getParameter("_classInfo", true);
		if (_classInfo == null)
			_classInfo = classInfo;

		// Set the context if needed
		if (!action.hasMark(LyAction.MARK_NO_CONTEXT))
		{
			LyChartBehaviorInterface behavior = (LyChartBehaviorInterface)getBehavior();
			LyWorkSpace 			 parentContext = null;

			if (parent instanceof LyContextInterface)
				parentContext = ((LyContextInterface)parent).getContext();

			setContext(behavior.getChartContext(this, session, _classInfo, objects, parentContext));
		}

		if (chartController != null)
			_chartController = chartController;
		else if (parent instanceof LyChartController)
			_chartController = (LyChartController)parent;

		if (chartController == null)
		{
			LyChartInfo chartInfo = (LyChartInfo)action.getParameter("_chartInfo", true);
			String	    selectFile = (String)action.getParameter("_selectFile", true);

			// Be sure to have the right chart

			if (chartInfo != null)
				_chartInfo = chartInfo.cloneChartInfo(chartInfo.getId());
			else if (selectFile != null)
			{
				LyApplication		    application = getApplication();
				LyEnvironment		    environment = getEnvironment();
				LyInfoList<LyChartInfo> charts = application.getInfoList(LyChartInfo.class);
				int					    i, n = ((charts == null) ? 0 : charts.getSize());
				String				    file;
				File				    selectFileAsFile = null;

				selectFile = environment.getString().expand(selectFile);
				selectFileAsFile = new File(selectFile);

				for (i = 0; (i < n) && (_chartInfo == null); i++)
				{
					chartInfo = charts.getInfo(i);
					file = chartInfo.getFile();

					if (file != null)
					{
						file = environment.getString().expand(file);

						File f1 = new File(file);

						if (f1.equals(selectFileAsFile))
							_chartInfo = chartInfo.cloneChartInfo(chartInfo.getId());
					}
				}

				if (_chartInfo == null)
				{
					try
					{
						LyInfoParser infoParser = new LyInfoParser(session.getApplication(), selectFile);

						_chartInfo = (LyChartInfo)infoParser.getRootElement();
						_chartInfo.setFile(selectFile);
					}
					catch (Exception e)
					{
						_environment.writeException(e);
					}
				}
			}
			else
			{
			    _chartInfo = new LyChartInfo();
			    _chartInfo.setApplication(getApplication());
			}
		}
		else if (_chartController != null)
			_chartInfo = chartController.getChartInfo();

		if ((_chartInfo != null)
				&& (_chartInfo.getName() != null))
			setName(_chartInfo.getName());
	}

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo always the classInfo of the specified objects.
	 * @param objects   current context (not null because the view is contextual to a list of
	 *                  objects).
	 */
	public LyGraphicalChartController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		this(session, parent, action, classInfo, objects, null);
	}

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo always the classInfo of the specified objects.
	 */
	public LyGraphicalChartController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo)
	{
		this(session, parent, action, classInfo, null);
	}

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 * @param action  related action. If an action is specified, the controller uses action
	 *                parameters (action name, action specific XML view description ...)
	 */
	public LyGraphicalChartController(LySession session, LyController parent, LyAction action)
	{
		this(session, parent, action, action.getClassInfo());
	}

	/**
	 * Retrieves the view Builder : The builder of the chart window.
	 *
	 * @return the builder instance;
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyGraphicalChartBuilder(this);
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyChartBehaviorInterface.class;
	}

    /**
     * Get content of the chart.
     *
     * @return the list of objects being displayed
     */
    public LyObjectList getContent()
    {
    	if (_context != null)
    		return _context;

    	LyAreaController	parentAreaController = getParentAreaController();

    	if (parentAreaController != null)
    	{
    		LyObjectList	content = parentAreaController.getContent();

    		if (content != null)
    			return content;
    	}

		LyPreference		    pref = _chartInfo.getPreference();
		LyFilter			    filter = (pref == null ? null : pref.getFilter());
		LySort				    sort = (pref == null ? null : pref.getSort());
		LyInfoList<LyFieldInfo> fields = _chartInfo.getFields();
		LyObjectSource			source = _classInfo.getObjectSource();
		
		if (source == null)
			return null;
		
		LyCache					cache = source.getCache();
		LyObjectList			ws;

		if ((cache != null) && cache.isLoaded()
				|| (cache instanceof LyFullCache))
	    	ws = new LyClassWorkSpace(_session, _classInfo, fields, filter, sort, false, true);
		else
			ws = new LyPageWorkSpace(_session, _classInfo, fields, filter, sort, 0, false, true);

    	return ws;
    }

	/**
	 * Prepare component before showing. This method is called once (unlike reset which may be
	 * called several times), before first showing. Default implementation invokes reset.
	 */
	protected void initArea()
	{
		super.initArea();

		if (isFreed())
			return;

		_initialized = false;

		updateTitleBar();

		_initialized = true;
	}

	/**
	 * @see leon.control.LyAreaController#reset(boolean)
	 */
	public void reset(boolean full)
	{
		LyGraphicalChartBuilder builder = (LyGraphicalChartBuilder)getBuilder();

		super.reset(full);

		LyChartBehaviorInterface behavior = (LyChartBehaviorInterface)getBehavior();

		builder.updateSpecificBehaviors();
		behavior.completeChart(this, getChartComponent());
	}

	/**
	 * Reloads the chart component in the parent view.
	 */
	public void reloadChartComponent()
	{
		LyGraphicalChartBuilder		builder = (LyGraphicalChartBuilder)getBuilder();
		LyChartBehaviorInterface	behavior = (LyChartBehaviorInterface)getBehavior();
	    LyArea    area = (LyArea)getComponent();
	    LyPanel   panel = area.getPanel(0);

	    area.removePanel(panel);

	    LyChart	newChart = builder.buildChart();

	    area.addPanel(newChart);
	    panel.free();
		reset(true);
		builder.updateSpecificBehaviors();
		behavior.completeChart(this, getChartComponent());
	}

	/**
	 * @see leon.control.LyAreaController#setContext(leon.data.LyWorkSpace)
	 */
	public boolean setContext(LyWorkSpace context)
	{
		if (_action.hasMark(LyAction.MARK_NO_CONTEXT))
			return false;

		if (getBehavior() instanceof LyChartBehaviorInterface)
		{
			LyObjectList	newContext = ((LyChartBehaviorInterface)getBehavior()).getChartContext(this,
					getSession(), _classInfo, null, context);

			if (newContext instanceof LyWorkSpace)
				return super.setContext((LyWorkSpace)newContext);
		}

		return super.setContext(context);
	}

	/**
	 * Indicates if the view for this import controller is resizable.
	 *
	 * @return true if the view is resizable, false otherwise.
	 */
	public boolean isViewResizable()
	{
		LyChartBehaviorInterface behavior = (LyChartBehaviorInterface)getBehavior();

		return behavior.isViewResizable(this);
	}

	/**
	 * Retrieves the LyChartController parent if there is one (ie the chart is defined by
	 * parametring windows).
	 *
	 * @return the chartController instance;
	 */
	public LyChartController getChartController()
	{
		return _chartController;
	}

	/**
	 * Retrieves the ClassInfo.
	 *
	 * @return the ClassInfo instance;
	 */
	public LyClassInfo getClassInfo()
	{
		return _classInfo;
	}

	/**
	 * Retrieves the ChartInfo.
	 *
	 * @return the ChartInfo instance;
	 */
	public LyChartInfo getChartInfo()
	{
		return _chartInfo;
	}

	/**
	 * Accessor to the chart component (view).
	 *
	 * @return The chart view component
	 */
	public LyChart getChartComponent()
	{
		// Retrieve internal chart component
		LyGraphicalChartBuilder chartBuilder = (LyGraphicalChartBuilder)getBuilder();

		return chartBuilder.getChart();
	}

	/**
	 * Sets the ChartInfo.
	 *
	 * @param chartInfo the ChartInfo instance;
	 */
	public void setChartInfo(LyChartInfo chartInfo)
	{
		_chartInfo = chartInfo;

		if ((_chartInfo != null) && (_chartInfo.getName() != null))
			setName(_chartInfo.getName());
	}

	/**
	 * Invoked when an action occurs.
	 *
	 * @param demand the demand caught by the listener
	 */
	public void actionDemand(LyActionDemand demand)
	{
		// Update the last modification time of the controller.
		setLastModified(new Date().getTime());

		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			String id = demand.getSource();

			if (id.equals(LyChartController.DEMAND_PREVIOUS_STEP))
			{
				if (null != _chartController)
					_chartController.demandPrevious4();

				// Update the last modification time of the controller.
				setLastModified(new Date().getTime());
			}
			else if (id.equals(LyChartController.DEMAND_MODIFY_STEP))
			{
				// Hide the current view.
				hideView();

				LyApplication application = getApplication();
				LyAction	  action = application.getAction("_chart");

				_chartController = new LyChartController(_session, this, action, _classInfo,
						getContext(), _chartInfo, this);

				_chartController.setGraphicalChartAction(getAction());
				_chartController.showView();

				// disable the modify button
				LyCommands commands = ((LyArea)this.getBuilder().getComponent()).getCommands();
				LyCommand  command = commands.getCommand(0);

				if (command.getId() == LyChartController.DEMAND_MODIFY_STEP)
					command.setEnabled(false);

				// Update the last modification time of the controller.
				setLastModified(new Date().getTime());
			}
			// Printing.
			else if (id == DEMAND_PRINT_CHART)
				printChart();
			// Saving image.
			else if (id == DEMAND_SAVE_IMG_CHART)
				selectFile(id, null);
			// Saving image.
			else if (id == DEMAND_SAVE_DEF_CHART)
				selectFile(id, null);
			// View data.
			else if (id == DEMAND_CHART_DATA)
				viewChartData();
			else
				super.actionDemand(demand);
		}
		else
			super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Free this controller on request of the specified controller.
	 *
	 * @param controller controller instance which requested the freeing of this controller. If the
	 *                   controller is the parent, this controller is freed otherwise the method
	 *                   does nothing.
	 * @see   #free()
	 */
	public void free(LyController controller)
	{
		super.free(controller);

		_classInfo = null;
		_chartInfo = null;

		if (_chartController != null)
		{
			_chartController._graphicalChartController = null;
			_chartController = null;
		}
	}

	/**
	 * Save the image of the chart.
	 *
	 * @param demandId the ID of the demand
	 * @param parent the parent controller
	 */
	protected void selectFile(String demandId, LyController parent)
	{
		// Retrieve internal chart component
		LyChart		  chart = ((LyGraphicalChartBuilder)this.getBuilder()).getChart();
		LyController  newParent = ((parent == null) ? this : parent);
		LyApplication application = getApplication();
		LyAction      action = application.getActionWithId(LyAction.ID_CHART_FILE, _classInfo);

		_LyChartFileController fileController = new _LyChartFileController(getSession(), newParent,
				action, _classInfo, getContext(), chart, demandId);

		fileController.showView();
	}

	/**
	 * Print the content of the chart Run it asynchronous.
	 */
	protected void printChart()
	{
		LyViewManager viewManager = getViewManager();

		viewManager.execGui(this, false);
	}

	/**
	 * Show the chart data view
	 */
	protected void viewChartData()
	{
		LyChart				   chart = ((LyGraphicalChartBuilder)this.getBuilder()).getChart();
		LyAction			   action = (LyAction)getApplication().getInfo("__chartDataTable");

		_LyChartDataController dataController = new _LyChartDataController(getSession(), this,
				action, chart);

		dataController.showView();
	}

	/**
	 * Method that should be called when starting the run method when the controller is Runnable.
	 * Modified to register the printing thread on the controller.
	 *
	 * @see #endRun
	 */
	protected void startRun()
	{
		super.startRun();

		getViewController().setPrintThread(Thread.currentThread());
	}

	/**
	 * Method that should be called when ending the run method when the controller is Runnable.
	 * Modified to unregister the printing thread on the controller.
	 *
	 * @see #startRun
	 */
	protected void endRun()
	{
		getViewController().setPrintThread(null);

		super.endRun();
	}

	/**
	 * Starting printChart request.
	 */
	public void run()
	{
		startRun();

		// Retrieve internal chart component
		LyChart chart = ((LyGraphicalChartBuilder)getBuilder()).getChart();

		chart.print(this);

		endRun();
	}

	/**
	 * Implementation of LyPrintListenerInterface.
	 *
	 * @param page the page to print
	 */
	public void print(int page)
	{
	}

	/* Internal Classes */

	/** Preview inputs. */
	public static final String SELECT_FILE = "_chartImgFile".intern();

	/** Preview inputs. */
	public static final String OVERWRITE_DEF = "_overwriteDef".intern();

	/** Preview inputs. */
	public static final String OVERWRITE_DEF_YES = "_overwriteDefYes".intern();

	/** Preview inputs. */
	public static final String OVERWRITE_DEF_NO = "_overwriteDefNo".intern();

	/** Preview inputs. */
	public static final String SAVE_PREFERENCE = "_savePref".intern();

	/** Preview inputs. */
	public static final String SAVE_PREFERENCE_YES = "_savePrefYes".intern();

	/** Preview inputs. */
	public static final String SAVE_PREFERENCE_NO = "_savePrefNo".intern();

	/** Length of the field to enter the name of the save file (for chart definition). */
	public static final int    FILE_NAME_WIDTH = 40;

	/**
	 * Internal class to show a File Selector.
	 */
	protected class _LyChartFileController extends LyUserInputController
		implements LyQuestionAnswerInterface
	{

		/** Chart. */
		protected LyChart _chart;

		/** Type of demand : Save chart image or Save chart definition. */
		protected String  _demandType;

		/* Constructor */

		/**
		 * Default constructor.
		 *
		 * @param session    current session associated to the controller. The controller
		 *                   automatically registers on the specified session.
		 * @param parent     parent controller Controller hierarchy indicates a child controller
		 *                   (for example a dialog created from a main view)
		 * @param action     related action. If an action is specified, the controller uses action
		 *                   parameters (action name, action specific XML view description ...)
		 * @param classInfo  the associated classinfo
		 * @param objects    current context (not null because the view is contextual to a list of
		 *                   objects).
		 * @param chart      the chart component
		 * @param demandType the type of demand : DEMAND_SAVE_IMG_CHART or DEMAND_SAVE_DEF_CHART.
		 */
		public _LyChartFileController(LySession session, LyController parent,
			LyAction action, LyClassInfo classInfo, LyObjectList objects,
			LyChart chart, String demandType)
		{
			super(session, parent, action, classInfo, objects);
			_chart = chart;
			_demandType = demandType;
		}

		/**
		 * Invoked when an action occurs.
		 *
		 * @param demand demand parameters sent by the view component.
		 * @see   LyController#demandPerformed
		 */
		public void actionDemand(LyActionDemand demand)
		{
			if (demand.getType() == LyActionDemand.TYPE_ACTION)
			{
				String id = demand.getSource();

				if (id.equals(LyFormController.DEMAND_VALIDATE))
				{
					LyFormField field = getFormField(SELECT_FILE);
					LyFieldItem fieldItem = field.getFieldItem();
					String	    filePath = ((LyGetLabelInterface)fieldItem).getLabel();

					if (filePath == null)
					{
						String message;

						if (_demandType == DEMAND_SAVE_DEF_CHART)
							message = this.getEnvironment().getMessage("LY_ER_NOTNULL", "LY_CHART_NAME");
						else
							message = this.getEnvironment().getMessage("LY_ER_NOTNULL", "LY_FILE");

						LyViewManager viewManager = this.getViewManager();
						LyComponent   root = this.getBuilder().getComponent();

						viewManager.showError(root, message);
						return;
					}

					if (_demandType == DEMAND_SAVE_DATA)
						saveData(filePath);
					else if (_demandType == DEMAND_SAVE_IMG_CHART)
						saveImgChart(filePath);
					else if (_demandType == DEMAND_SAVE_DEF_CHART)
						saveDefChart(filePath);

					return;
				}
				else if (id.equals(LyFormController.DEMAND_CANCEL))
				{
					this.hideView();
					this.free();

					return;
				}
			}
		} // end method actionDemand

		/**
		 * Method used to save the image to the file specified in the form.
		 *
		 * @param filePath the path to the save file.
		 */
		protected void saveImgChart(String filePath)
		{
			if (filePath == null)
		        return;
		    if (!filePath.endsWith(".png"))
		        filePath += ".png";

			// If file already exists, ask if it must be overwritten
			File dataFile = new File(filePath);

			if (dataFile.exists())
				this.getViewManager().askQuestion(this.getView(),
					this.getEnvironment().getMessage("LY_CONFIRM_OVERWRITE", filePath), this,
					DEMAND_OVERWRITE_FILE);
			// Process goes on in questionAnswer
			else
				questionAnswer(true, DEMAND_OVERWRITE_FILE);
		}

		/**
		 * Answer to a askQuestion request.
		 *
		 * @param result true if the user has clicked on Yes or Ok
		 * @param clientData data specified in the askQuestion method
		 * @see   LyQuestionAnswerInterface
		 */
		public void questionAnswer(boolean result, Object clientData)
		{
			String filePath = null;

			if (result)
			{
				LyFormField field = getFormField(SELECT_FILE);
				LyFieldItem fieldItem = field.getFieldItem();

				filePath = ((LyGetLabelInterface)fieldItem).getLabel();
			}

			if ((clientData == DEMAND_OVERWRITE_FILE) && result)
			{
			    if ((filePath != null)
		            	&& !filePath.endsWith(".png"))
		        	filePath += ".png";

				// Retrieve internal view
				_chart.saveImg("", filePath);
				this.free();
			}
			else if ((clientData == DEMAND_SAVE_DATA) && result)
			{
            	if ((filePath != null)
                    	&& !filePath.endsWith(".csv"))
                	filePath += ".csv";

				((_LyChartDataController)this._parent).writeDataToFile(filePath);
				this.free();
			}
		}

		/**
		 * Method used to save the chart definition to an xml file.
		 *
		 * @param name the name of the statistic.
		 */
		protected void saveDefChart(String name)
		{
			LyEnvironment environment = this.getEnvironment();
			boolean		  overwrite = false;
			LyFormField   field = getFormField(OVERWRITE_DEF);
			LyPreference  preference = null;

			// Check if file must be overwritten
			if (field != null)
			{
				LyChoice choice = (LyChoice)field.getFieldItem();
				LyOption option = null;
				int		 j, n = choice.getOptionCount();

				for (j = 0; j < n; j++)
				{
					option = choice.getOption(j);
					if (option.isChecked())
						break;
				}

				if ((option != null)
                        && (option.getId().equals(OVERWRITE_DEF_YES)))
					overwrite = true;
			}

			// Check if preferences must be saved
			field = getFormField(SAVE_PREFERENCE);

			if (field != null)
			{
				LyChoice choice = (LyChoice)field.getFieldItem();
				LyOption option = null;
				int		 j, n = choice.getOptionCount();

				for (j = 0; j < n; j++)
				{
					option = choice.getOption(j);

					if (option.isChecked())
						break;
				}

				// If preference must be saved, retrieve table sort and filter
				if ((option != null) && (option.getId().equals(SAVE_PREFERENCE_YES)))
				{
					LySimpleTableController tc = getParentTableController();

					if (tc != null)
					{
						LySort		 sort = tc.getContent().getSort();
						LyFilter     filter = tc.getFilter();
						boolean		 isFilterActive = tc.isFilterActive();
						LyAction     action = tc.getAction();
						LyObjectInfo objectInfo;

						if (action.isActionRef())
							objectInfo = tc.getClassInfo();
						else
							objectInfo = action;

						preference = new LyPreference("_preference" + LyString.getRandomId(),
								objectInfo, "Chart preference", null, 0, null, false);

						if ((isFilterActive) && (filter != null))
							preference.setFilter(filter.cloneFilter());

						if (sort != null)
							preference.setSort(sort.cloneSort());
					}
				}
			}

            if (_chartInfo != null)
            {
    			// Create xml file for the given classInfo
    			try
    			{
					String     id;

					// If we have save the same chart for a second time with another id, we must duplicate it.
					if (overwrite)
						id = _chartInfo.getId();
					else
                    {
                        id = "_chart_" + LyString.getRandomId();
						_chartInfo = _chartInfo.cloneChartInfo(id);
                        _chartInfo.setFile(null);
                    }

					// Affect the name to the LyChartInfo
					_chartInfo.setName(name);
					this.setName(name);

					// Affect the current classInfo to the LyChartInfo
					_chartInfo.setClassInfoId(this._classInfo.getId());

					if (preference != null)
						_chartInfo.setPreference(preference);
					else
						_chartInfo.setPreference(null);

					// Register this chart on the application
					LyApplication application = this.getApplication();
					LySession     session = this.getSession();

					// If it is a duplication we must register the new chart info.
					LyInfo info = application.getInfo(id);

					if ((info != _chartInfo)
						&& (info instanceof LyExtraInfoInterface))
					{
						application.deleteExtraInfo((LyExtraInfoInterface)info);
						_chartInfo.setFile(null);
					}

					application.saveExtraInfo(_chartInfo);

					if ((info == null)
							|| (_chartInfo.getApplication() == null))
						_chartInfo.setApplication(application);

					// Notify controllers that a new chartInfo has been created
					session.notifyInfoListeners("_chart");

					this.free();
				}
    			catch (Exception e)
    			{
                    String  fileName = _chartInfo.getFile();

    				environment.writeException(e);

    				environment.logError("Unable to open file :" + fileName);
    				this.showError(environment.getMessage("LY_ER_OPEN_FILE", fileName));
    				return;
    			}
            }
		} // end method saveDefChart

		/**
		 * Save the data to the given file path
		 * @param filePath a file path
		 */
		protected void saveData(String filePath)
		{
			if (filePath == null)
		        return;
		    if (!filePath.endsWith(".csv"))
		        filePath += ".csv";

			// If file already exists, ask if it must be overwritten
			File dataFile = new File(filePath);

			if (dataFile.exists())
				this.getViewManager().askQuestion(this.getView(),
					this.getEnvironment().getMessage("LY_CONFIRM_OVERWRITE", filePath), this,
					DEMAND_SAVE_DATA);
			// Process goes on in questionAnswer
			else
				questionAnswer(true, DEMAND_SAVE_DATA);
		}

		/**
		 * Retrieve formField corresponding to id. Utility method : internal use.
		 *
		 * @param  id identifier of the fieldInfo
		 * @return the formField corresponding to id
		 */
		protected LyFormField getFormField(String id)
		{
			// Retrieve formField corresponding to id
			LyComponent root = this._builder.getComponent();

			return (LyFormField)root.getComponent(id);
		}

		/**
		 * Prepare component before showing. This method is called once, before first showing.
		 */
		protected void initArea()
		{
			LyEnvironment environment = this.getEnvironment();

			if (this._action != null)
				setImage(this.getImage(this._action));

			if (_demandType == DEMAND_SAVE_IMG_CHART)
			{
				addFile(SELECT_FILE, "LY_FILE", null, LyFile.Type.TEXT, false);
			}
			else if (_demandType == DEMAND_SAVE_DEF_CHART)
			{
				addText(SELECT_FILE, "LY_CHART_NAME", environment.translate(_chartInfo.getTitle()),
					FILE_NAME_WIDTH, null, false);

				if (_chartInfo.getFile() != null)
				{
					LyChoice    choice = addChoice(OVERWRITE_DEF, "LY_OVERWRITE_CHART", false, false);
					LyOption    option;
					LyFormField formField = (LyFormField)choice.getParent();
					LyTip	    tip = new LyTip();

					tip.setLabel("LY_OVERWRITE_CHART_TIP");
					formField.setTip(tip);

					option = new LyOption(OVERWRITE_DEF_YES, "LY_YES");
					option.setChecked(true);
					choice.addOption(option);

					option = new LyOption(OVERWRITE_DEF_NO, "LY_NO");
					choice.addOption(option);
				}

				// Retrieve the area controller from which this controller was invoked.
				LySimpleTableController tc = getParentTableController();

				if (tc != null)
				{
					LyChoice    choice = addChoice(SAVE_PREFERENCE, "LY_SAVE_CHART_PREFERENCE", false, false);
					LyOption    option;
					LyFormField formField = (LyFormField)choice.getParent();
					LyTip	    tip = new LyTip();

					tip.setLabel("LY_SAVE_CHART_PREFERENCE_TIP");
					formField.setTip(tip);

					option = new LyOption(SAVE_PREFERENCE_YES, "LY_YES");
					choice.addOption(option);

					option = new LyOption(SAVE_PREFERENCE_NO, "LY_NO");
					option.setChecked(true);
					choice.addOption(option);
				}
			}
			else if (_demandType == DEMAND_SAVE_DATA)
			{
				addFile(SELECT_FILE, "LY_FILE", null, LyFile.Type.TEXT, false);
			}

			LyCommands commands = new LyCommands();
			LyCommand  command = new LyCommand(LyFormController.DEMAND_VALIDATE, "LY_VALIDATE");

			command.setAction(LyCommand.Action.SUBMIT);
			commands.addCommand(command);
			((LyArea)_component).setCommands(commands);

			super.initArea(); // Let parent do its job
		} // end method initArea
	} // end class _LyChartFileController

	/**
	 * Return the closest simple table controller of this controller. Return null if there is no
	 * parent of this type.
	 * @return the associated table controller
	 */
	protected LySimpleTableController getParentTableController()
	{
		LyController parent = this.getParent();

		while ((parent != null) && !(parent instanceof LyTableController))
			parent = parent.getParent();

		return (LySimpleTableController)parent;
	}

	/** Save chart data. */
	public static final String DEMAND_SAVE_DATA = "_saveData".intern();

	/**
	 * Internal class to show chart data.
	 */
	protected class _LyChartDataController extends LyAreaController
	{

		/** Chart. */
		protected LyChart _chart;

		/* Constructor */

		/**
		 * Default constructor.
		 *
		 * @param session   current session associated to the controller. The controller
		 *                  automatically registers on the specified session.
		 * @param parent    parent controller Controller hierarchy indicates a child controller (for
		 *                  example a dialog created from a main view)
		 * @param action    related action. If an action is specified, the controller uses action
		 *                  parameters (action name, action specific XML view description ...)
		 * @param chart     the chart component
		 */
		public _LyChartDataController(LySession session, LyGraphicalChartController parent,
			LyAction action, LyChart chart)
		{
			super(session, parent, action, null);

			_chart = chart;
		}

		/**
		 * Invoked when an action occurs.
		 *
		 * @param demand demand parameters sent by the view component.
		 * @see   LyController#demandPerformed
		 */
		public void actionDemand(LyActionDemand demand)
		{
			if (demand.getType() == LyActionDemand.TYPE_ACTION)
			{
				String		  id = demand.getSource();
				LyViewManager viewManager = getViewManager();

				if (id.equals(DEMAND_SAVE_DATA))
				{
					if (viewManager.hasProperty("LOCAL_FILE"))
						selectFile(id, this);
					else
					{
						// ask a temporary file to viewManager
						String fileName = "_chartData" + System.currentTimeMillis() + ".csv";
						String file = viewManager.getTemporaryDir() + "/" + fileName;
						String link = viewManager.getTemporaryLocation() + "/" + fileName;

						writeDataToFile(file);
						viewManager.showFile(null, link);
					}

					return;
				}
				else
					super.actionDemand(demand);
			}
		} // end method actionDemand

		/**
		 * Generates a file containing the chart data.
		 * @param filePath the file path
		 */
		protected void writeDataToFile(String filePath)
		{
			_LyChartDataBuilder builder = (_LyChartDataBuilder)this._builder;
			String[][]		    dataValues = builder.getDataValues();

			// Open the file
			FileOutputStream    output;
			LyEnvironment	    environment = this.getEnvironment();

			try
			{
				output = new FileOutputStream(filePath);
			}
			catch (Exception e)
			{
				environment.logError("Unable to open file :" + filePath);
				this.showError(environment.getMessage("LY_ER_OPEN_FILE", filePath));
				return;
			}

			environment.logInfo("Saving file :" + filePath);

			// Write the file
			PrintWriter ps = new PrintWriter(output);
			String	    line;
			int		    i, nl = dataValues.length;

			for (i = 0; i < nl; i++)
			{
				line = LyString.listToString(dataValues[i], ';');

				ps.println(line);
			}

			ps.close();
		} // end method writeDataToFile

		/**
		 * Retrieves the view Builder.
		 *
		 * @return the builder instance;
		 */
		protected LyBuilder getDefaultBuilder()
		{
			return new _LyChartDataBuilder(this);
		}

		/**
		 * Accessor to the refresh time of this controller. Used by thin clients to know after a
		 * selection whether the controller view must be refreshed or not.
		 *
		 * @return A negative number if view must be refreshed or the refresh date.
		 */
		public long getRefreshTime()
		{
			return 0;
		}

		/**
		 * Accessor to the chart component
		 * @return the chart component
		 */
		public LyChart getChart()
		{
			return _chart;
		}

		/**
		 * Method that gives the interface that must be implemented by the behavior associated to
		 * this controller.
		 *
		 * @return Class representing the behavior interface of this controller.
		 */
		public Class<?> getBehaviorInterface()
		{
			return LyChartBehaviorInterface.class;
		}

		/**
		 * Gets title associated with this controller. The method is overriden to check the name of
		 * the chartinfo if available
		 *
		 * @return the title corresponding to the action.
		 */
		public String getTitle()
		{
			if ((_chart != null)
					&& (_chart.getTitle() != null)
					&& (_chart.getTitle().length() > 0))
				return _chart.getTitle();
			else
				return super.getTitle();
		}
	} // end class _LyChartDataController


	/**
	 * Internal class to build the view to show chart data.
	 */
	protected class _LyChartDataBuilder extends LyAreaBuilder
	{

		/** Data values. */
		protected String[][] _dataValues;

		/**
		 * Default constructor.
		 *
		 * @param parent parent controller
		 */
		public _LyChartDataBuilder(_LyChartDataController parent)
		{
			super(parent);

			_dataValues = null;
		}

		/**
		 * Builds the view component for the specified parent controller. The component may create
		 * the XML Tree or the internal view.
		 *
		 * @return the view Component
		 */
		@SuppressWarnings("deprecation")
		public LyComponent build()
		{
			LyArea area = new LyArea(); //(LyArea)super.build();

			// set content
			area.setController(this._parent.getId());
			area.setScrolling(LyArea.Scrolling.AUTO);
			area.setId(this._parent.getId());
			setTitleBarParameters(((LyAreaController)this._parent).getTitle(), null, null, null,
				null);
			area.setTitleBar(buildTitleBar());
			area.addPanel(buildTable());
			area.setCommands(buildCommands());

			this._component = area;

			// Update default content
			this.updateTitleBar();

			return area;
		}

		/**
		 * Builds the table component.
		 *
		 * @return the table component
		 */
		protected LyTable buildTable()
		{
			LyTable		 table = new LyTable(this._parent.getId() + "_table");
			LyChart		 chart = ((_LyChartDataController)this._parent).getChart();
			LyColumn     column;
			LyChartValue chartValue;
			LyRows		 rows = new LyRows();
			LyRow		 row;
			LyCell		 cell;
			String[][]   dataValues = null;
			int			 i, j, nValues = 0, nCols = chart.getColumnCount();

			//no selection are authorized see FFT0963
			table.setSelectionMode(LyTable.SelectionMode.NONE);

			// Retrieve the values from the columns of the chart
			for (i = 0; i < nCols; i++)
			{
				column = chart.getColumn(i);
				nValues = column.getChartValueCount();

				if (dataValues == null)
					dataValues = new String[nCols + 1][nValues + 1];

				dataValues[i + 1][0] = column.getLabel();

				for (j = 0; j < nValues; j++)
				{
					chartValue = column.getChartValue(j);

					dataValues[i + 1][j + 1] = format(chartValue.getYValue());

					if (i == 0)
						dataValues[0][j + 1] = chartValue.getXValue();
				}
			}

			int ni = nCols + 1, nj = nValues + 1;

			// Invert data if needed
			if (nCols <= nValues)
			{
				ni = nValues + 1;
				nj = nCols + 1;

				_dataValues = new String[nValues + 1][nCols + 1];

				if (dataValues != null)
					for (i = 0; i < (nCols + 1); i++)
					{
						for (j = 0; j < (nValues + 1); j++)
							_dataValues[j][i] = dataValues[i][j];
					}
			}
			else
				_dataValues = dataValues;

			// Use a chart converter to transform the labels.
			// Dates are converted
			LyViewManager    vm = getViewManager();
			LyChartConverter conv = vm.buildChartConverter(chart);
			Vector<String>   xSerie;

			conv.setIntervals();

			if (nCols <= nValues)
			{
				xSerie = new Vector<String>(ni - 1);

				for (i = 1; i < ni; i++)
					xSerie.addElement(_dataValues[i][0]);

				try
				{
					String[] labels = conv.buildXLabels(xSerie, chart.getXVal(), chart.getXType(),
							chart.getXDateType(), false);

					for (i = 0; i < labels.length; i++)
						_dataValues[i + 1][0] = LyString.replace(labels[i], " \n ", " ");
				}
				catch (NumberFormatException nfe)
				{
				}
			}

			if (nCols > nValues)
			{
				xSerie = new Vector<String>(nj - 1);

				for (j = 1; j < nj; j++)
					xSerie.addElement(_dataValues[0][j]);

				try
				{
					String[] labels = conv.buildXLabels(xSerie, chart.getXVal(), chart.getXType(),
							chart.getXDateType(), false);

					for (j = 0; j < labels.length; j++)
						_dataValues[0][j + 1] = LyString.replace(labels[j], " \n ", " ");
				}
				catch (NumberFormatException nfe)
				{
				}
			}

			LyHeaders headers = new LyHeaders();
			LyHeader  header = new LyHeader();

			// Build headers
			for (j = 0; j < nj; j++)
			{
				header = new LyHeader();
				header.setSortable(false);

				if (_dataValues[0][j] != null)
					header.setLabel(_dataValues[0][j]);
				else
					header.setLabel("");

				headers.addHeader(header);
			}

			// Fill table with data
			for (i = 1; i < ni; i++)
			{
				row = new LyRow();

				for (j = 0; j < nj; j++)
				{
					cell = new LyCell();
					cell.setLabel(_dataValues[i][j]);
					row.addCell(cell);
				}

				rows.addRow(row);
			}

			table.setHeaders(headers);
			table.setRows(rows);

			return table;
		} // end method buildTable

		/**
		 * Formats a string for a data value.
		 *
		 * @param  str String that must be formatted.
		 * @return the formatted string.
		 */
		protected String format(String str)
		{
			String		  result;
			double		  doubleValue = LyTools.doubleFromString(str);
			DecimalFormat format = new DecimalFormat();

			// Avoid blanks in string
			format.setGroupingSize(1000);

			result = format.format(doubleValue);

			return result;
		}

		/**
		 * Return data values.
		 *
		 * @return the data values
		 */
		protected String[][] getDataValues()
		{
			return _dataValues;
		}

		/**
		 * Builds the commands bar.
		 *
		 * @return the commands component
		 */
		protected LyCommands buildCommands()
		{
			LyCommands commands = new LyCommands();
			LyCommand  command;

			commands.setId("_commands");

			command = new LyCommand(DEMAND_SAVE_DATA, "LY_SAVE_CHART_DATA");
			command.setAction(LyCommand.Action.SUBMIT);
			commands.addCommand(command);

			command = new LyCommand(LyFormController.DEMAND_CANCEL, "LY_CLOSE");
			command.setAction(LyCommand.Action.CLOSE);
			commands.addCommand(command);

			command = buildHelpCommand();

			if (command != null)
				commands.addCommand(command);

			return commands;
		}
	} // end class _LyChartDataBuilder

	/**
	 * Gets title associated with this controller. The method is overriden to check the name of the
	 * chartinfo if available
	 *
	 * @return the title corresponding to the action.
	 */
	public String getTitle()
	{
		if ((_chartInfo != null)
				&& (_chartInfo.getName() != null)
				&& (_chartInfo.getName().length() > 0))
			return _chartInfo.getName();
		else
			return super.getTitle();
	}

} // end class LyGraphicalChartController
