/**
 * LyImportController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyImportController.java 34887 2012-03-21 09:54:53Z jlsaouli $
 */

package leon.control;

import java.io.*;
import java.util.*;
import leon.app.*;
import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;


/**
 * This class implements the import edition form. The user may specify import parameters :
 *
 * <UL>
 *   <LI>The path to the file which contains data</LI>
 *   <LI>The list of fieldInfos to import</LI>
 *   <LI>Separators expected in the imported file between fields, and between values if
 *     multiple</LI>
 * </UL>
 *
 * <p>The data file is usually created from an export controller : LyExportController or an
 * application such as MS Excel. This class only implements the import form : the import process is
 * done in a LyImportTableController instance. This controller is associated with a
 * behaviorInterface (LyImportBehaviorInterface). Usually, this behavior is an instance of
 * LyImportBehavior</p>
 *
 * @see LyExportController
 * @see LyImportTableController
 * @see leon.app.behaviorinterface.LyImportBehaviorInterface
 * @see leon.app.behavior.LyImportBehavior
 */
public class LyImportController extends LyAreaController
{
	/** File Version */
	public static final String __VERSION =
		"$Id: LyImportController.java 34887 2012-03-21 09:54:53Z jlsaouli $";

	/** Form field identifier : List of fields for import. */
	public static final String SELECT_FIELDINFOS = "_fieldInfo";

	/** Form field identifier : select field separator (default is FIELD_SEP). */
	public static final String SELECT_FIELD_SEP = "_importFieldSep";

	/** Form field identifier : select value separator (default is VALUE_SEP). */
	public static final String SELECT_VALUE_SEP = "_importValueSep";

	/** Form field identifier : select input file path. */
	public static final String SELECT_FILE = "_importFile";

	/** Form field identifier : select input start line number. */
	public static final String SELECT_START_LINE_NUMBER = "_importStartLineNumber";

	/** Default field separator in imports. */
	public static String	   FIELD_SEP = "|";

	/** Default multiple value separator in imports. */
	public static String	   VALUE_SEP = ",";

	/** Default field separator in imports. */
	public static String	   SPREADSHEET_FIELD_SEP = ";";

	/** Default multiple value separator in imports. */
	public static String	   SPREADSHEET_VALUE_SEP = "~";

	/**
	 * Form field identifier : import type, either SELECT_IMPORT_TYPE_RAW or SELECT_IMPORT_TYPE_XML
	 * or SELECT_IMPORT_TYPE_SPREADSHEET.
	 */
	public static final String SELECT_IMPORT_TYPE = "_importType".intern();

	/** Form field value identifier : import raw data. */
	public static final String SELECT_IMPORT_TYPE_RAW = "_importTypeRaw".intern();

	/** Form field value identifier : import XML data. */
	public static final String SELECT_IMPORT_TYPE_XML = "_importTypeXml".intern();

	/** Form field value identifier : import spreadsheet data. */
	public static final String SELECT_IMPORT_TYPE_SPREADSHEET = "_importTypeSpreadsheet".intern();

	/** Form field identifier : operations to be executed. */
	public static final String SELECT_OPERATIONS = "_importOperations";

	/** Form field value identifier : creation operation. */
	public static final int    SELECT_OPERATIONS_CREATION = 0;

	/** Form field value identifier : creation operation. */
	public static final int    SELECT_OPERATIONS_MODIFICATION = 1;

	/** Form field value identifier : deletion operation. */
	public static final int    SELECT_OPERATIONS_DELETION = 2;

	/** Form field value identifier : consultation operation. */
	public static final int    SELECT_OPERATIONS_CONSULTATION = 3;

	/** Prefix used for the name of the preferences saved for import actions. */
	public static final String PREFERENCE_PREFIX = "__exportPreference:".intern();

	/** Reference to the Class info. */
	protected LyClassInfo	   _classInfo;

	/** Silent mode. */
	protected boolean		   _silent;

	/** Indicates if field separator has been changed by the user. */
	protected boolean		   _fieldSepChanged;

	/** Indicates if value separator has been changed by the user. */
	protected boolean		   _valueSepChanged;
	
	
	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller).
	 * @param action    related action.
	 * @param classInfo classInfo which indicates which kind of objects we expect in import
	 * @param objects   current context
	 */
	public LyImportController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		super(session, parent, action, null);

		_classInfo = classInfo;
		_silent = (parent == null);
		_fieldSepChanged = false;
		_valueSepChanged = false;
		
		if (!isFormDisplayed())
			validateImport();
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyImportBehaviorInterface.class;
	}

	/**
	 * Overriden to update the list of field infos with the export preference.
	 * @return true is action can be started
	 */
	protected boolean startAction()
	{
		boolean		 result = super.startAction();

		// Try to retrieve import preference for this class info
		LyPreference preference = getImportPreference();
		
		if (preference != null)
		{
			LyInfoList<LyFieldInfo> fields = preference.getFieldInfos();
			LyFormField			    formField = getFormField(LyImportController.SELECT_FIELDINFOS);

			if (formField != null)
			{
				LyRelation  relation = (LyRelation)formField.getFieldItem();
				LyLink	    link;
				LyFieldInfo fieldInfo;
				int		    i, n = ((fields == null) ? 0 : fields.getSize());
				int		    j, m = relation.getLinkCount();

				for (i = 0; i < n; i++)
				{
					fieldInfo = fields.getInfo(i);

					for (j = 0; j < m; j++)
					{
						link = relation.getLink(j);

						if (fieldInfo.getId().equals(link.getId()))
						{
							link.setSelected(true);
							relation.removeLink(link);
							relation.addLink(link);
							break;
						}
					}
				}
			}
			
		    LyViewManager viewManager = getViewManager();

		    if (!viewManager.hasProperty("MANAGE_ATTACHMENT"))
			{
    			// We set the last open file name present in preference file.
		        // Not in web mode, the file was changed by attachment management.
    			LyFormField			    selectFormField = getFormField(LyImportController.SELECT_FILE);
    			
    			if ((preference.getLastFileName() != null) 
    				&& (selectFormField != null))
    			{
    				LyFile file = (LyFile)selectFormField.getFieldItem();
    				
    				file.setLabel(preference.getLastFileName());
    			}
			}
		}

		return result;
	} // end method startAction

	/**
	 * Overridden to ensure that export form is displayed.
	 */
	public void showView()
	{
		if (isFormDisplayed())
			super.showView();
		else
			return;
	}

	/**
	 * Indicates if the export form is displayed.<br />
	 * Export form should not be displayed if all the fields values are given as parameters of the
	 * export action.
	 *
	 * @return true if form is displayed, false otherwise.
	 */
	protected boolean isFormDisplayed()
	{
		if ((getParameterValue(SELECT_FIELDINFOS) != null)
				&& (getParameterValue(SELECT_FIELD_SEP) != null)
				&& (getParameterValue(SELECT_VALUE_SEP) != null)
				&& (getParameterValue(SELECT_IMPORT_TYPE) != null)
				&& (getParameterValue(SELECT_FILE) != null)
				&& (getParameterValue(SELECT_START_LINE_NUMBER) != null))
			return false;

		return true;
	}

	/**
	 * Invoked when an action notification is sent to this listener.
	 *
	 * <P>This method is called by demandPerformed.</P>
	 *
	 * <P>At this level, the implementation handles demands of type LyActionDemand.TYPE_ACTION :
	 * LyFormController.DEMAND_VALIDATE, and LyExportController.DEMAND_CANCEL</P>
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		String id = demand.getSource();

		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			if (id.equals(LyFormController.DEMAND_VALIDATE))
			{
				validateImport();
				return;
			}
			else if (id.equals(LyFormController.DEMAND_CANCEL))
			{
				hideView();
				free();
				return;
			}
		}
		else if (id.equals(LyImportController.SELECT_IMPORT_TYPE))
		{
			LyFormField fieldSep = getFormField(LyImportController.SELECT_FIELD_SEP);
			LyFormField valueSep = getFormField(LyImportController.SELECT_VALUE_SEP);
			LyFormField startLine = getFormField(LyImportController.SELECT_START_LINE_NUMBER);

			// Field and value separators make no sense for XML import
			if (demand.getValue().equals(LyImportController.SELECT_IMPORT_TYPE_XML))
			{
				if (fieldSep != null)
					fieldSep.setEnabled(false);

				if (valueSep != null)
					valueSep.setEnabled(false);
			}
			else
			{
				if (fieldSep != null)
					fieldSep.setEnabled(true);

				if (valueSep != null)
					valueSep.setEnabled(true);

				LyText fieldSepItem = ((fieldSep == null) ? null : (LyText)fieldSep.getFieldItem());
				LyText valueSepItem = ((valueSep == null) ? null : (LyText)valueSep.getFieldItem());

				if (demand.getValue().equals(LyImportController.SELECT_IMPORT_TYPE_SPREADSHEET))
				{
					if (!_fieldSepChanged && (fieldSepItem != null))
						fieldSepItem.setLabel(SPREADSHEET_FIELD_SEP);

					if (!_valueSepChanged && (valueSepItem != null))
						valueSepItem.setLabel(SPREADSHEET_VALUE_SEP);
				}
				else
				{
					if (!_fieldSepChanged && (fieldSepItem != null))
						fieldSepItem.setLabel(FIELD_SEP);

					if (!_valueSepChanged && (valueSepItem != null))
						valueSepItem.setLabel(VALUE_SEP);
				}
			}

			// Start line is authorized for spreadsheet import only
			if (demand.getValue().equals(LyImportController.SELECT_IMPORT_TYPE_SPREADSHEET))
				startLine.setEnabled(true);
			else
				startLine.setEnabled(false);
		}
		else if (id.equals(LyImportController.SELECT_FIELD_SEP))
			_fieldSepChanged = true;
		else if (id.equals(LyImportController.SELECT_VALUE_SEP))
			_valueSepChanged = true;
		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Validate the first step of the import. This method called first checkFields on the behavior.
	 * If no error occured, we check that all required fields are correctly filled. Then, we check
	 * that the given file really exists. At end, we import the file
	 *
	 * @see #buildImportTable
	 */
	protected void validateImport()
	{
		LyEnvironment			  environment = getEnvironment();
		String					  message;
		String					  type;
		String					  fieldSep = (String)getParameterValue(SELECT_FIELD_SEP);
		String					  valueSep = (String)getParameterValue(SELECT_VALUE_SEP);
		String					  file = (String)getParameterValue(SELECT_FILE);
		boolean					  fileParameterSet = (file != null);
		String					  error = null;
		LyFormField				  field;
		LyFieldItem				  fieldItem;
		boolean					  isXmlImport = false;

		// Check if the chosen fields are corrects
		LyInfoList<LyFieldInfo>   fields = getSelectedFields();
		LyImportBehaviorInterface behavior = (LyImportBehaviorInterface)getBehavior();

		if ((message = behavior.checkFields(this, fields)) != null)
		{
			if (!_silent)
				showError(message);
			else
				environment.logError(message);
			return;
		}

		// Check if type is xml.
		type = getImportType();

		if (type == LyImportController.SELECT_IMPORT_TYPE_XML)
			isXmlImport = true;

		// Check if all the fields are fulfilled
		if (file == null)
		{
			field = getFormField(LyImportController.SELECT_FILE);
			fieldItem = field.getFieldItem();
			file = ((LyGetLabelInterface)fieldItem).getLabel();
		}

		if ((file == null) || (file.length() == 0))
			error = "LY_FILE";

		LyViewManager viewManager = getViewManager();
		String        filePath;

		if (!fileParameterSet
		    && viewManager.hasProperty("MANAGE_ATTACHMENT"))
		{
			// XXX Should use the file attachment handler instance here,
			// however, can access view/web packages from here.
			// XX Create interfaces.
			LyEnvironment env = getEnvironment();
			String		  docURL = env.getEnv("LY_APP_DOC_DIR");
			String		  fileDir = docURL + "/files"; // Should be : LyFilePartWriter.FILE_DIR;

			filePath = env.getString().expand(fileDir + "/" + file);
			filePath = filePath.replace('\\', '/');
		}
		else
		    filePath = file;

		if (fieldSep == null)
		{
			field = getFormField(LyImportController.SELECT_FIELD_SEP);
			if (field == null)
			{
				LyImportBehavior beh = (LyImportBehavior)getBehavior();

				fieldSep = beh.getDefaultAttributeSep();
			}
			else
			{
				fieldItem = field.getFieldItem();
				fieldSep = ((LyGetLabelInterface)fieldItem).getLabel();
			}
		}

		// Default value to pass the tests.
		if (isXmlImport)
			fieldSep = "|";

		// If fields are not available in the dialog,
		// check if behavior is defining them
		LyImportBehavior behav = (LyImportBehavior)getBehavior();

		if ((fieldSep == null) || (fieldSep.length() == 0))
			fieldSep = behav.getDefaultAttributeSep();
		if ((valueSep == null) || (valueSep.length() == 0))
			valueSep = behav.getDefaultAttributeValSep();

		if ((error == null) && ((fieldSep == null) || (fieldSep.length() == 0)))
			error = "LY_FIELD_SEP";

		if (valueSep == null)
		{
			field = getFormField(LyImportController.SELECT_VALUE_SEP);
			fieldItem = field.getFieldItem();
			valueSep = ((LyGetLabelInterface)fieldItem).getLabel();
		}

		// Default value to pass the tests.
		if (isXmlImport)
			valueSep = ",";

		if ((error == null) && ((valueSep == null) || (valueSep.length() == 0)))
			error = "LY_VALUE_SEP";

		if (error != null)
		{
			message = environment.getMessage("LY_ER_NOTNULL", error);
			if (!_silent)
				showError(message);
			else
				environment.logError(message);
			return;
		}

		// Check if at least one operation is selected
		BitSet operations = getImportOperations();

		// Check if the file really exists
		File   f = new File(filePath);

		if (!f.exists())
		{
			InputStream stream = environment.getResourceAsStream(filePath, null);

			if (stream == null)
			{
				message = environment.getMessage("LY_ER_FILE", filePath);
				if (!_silent)
					showError(message);
				else
					environment.logError(message);
				return;
			}
		}

		// Build the list of import objects
		fields = sortFields(fields);

		int     startLine = getStartLineNumber();
		BitSet  operationsParameter = (BitSet)getParameterValue(LyImportController.SELECT_OPERATIONS);

		boolean result = buildImportTable(filePath, fieldSep, valueSep, type, fields, operations,
				startLine, (operationsParameter == null));
		
		if (result)
		{
			savePreference(file);
			hideView();
			free();
		}
	} // end method validateImport

	/**
	 * Save the preference for the import action. This method generates a file in the data directory
	 * containing the preference associated with this import action.
	 */
	protected void savePreference()
	{
		savePreference(null);
	}
	
	/**
	 * Save the preference for the import action. This method generates a file in the data directory
	 * containing the preference associated with this import action.
	 * @param lastFileName file name
	 */
	protected void savePreference(String lastFileName)
	{
		LySession               session = getSession();
		LyApplication           application = getApplication();
		LyObject	            user = session.getUser();
		String		            userFullId = ((user == null) ? null : user.getId().toString());
		LyInfoList<LyFieldInfo> fields = getGraphicalSelectedFields();
		LyPreference		    preference = getImportPreference();

		if (preference == null)
		{
			String preferenceName = PREFERENCE_PREFIX + _classInfo.getId();

			preference = new LyPreference("_preference" + LyString.getRandomId(), _classInfo,
					preferenceName, null, 0, userFullId, false);
			preference.setApplication(application);
		}

		preference.setFieldInfos(fields);
		
		// We save the path and the name of the last open file.
		if (lastFileName != null)
			preference.setLastFileName(lastFileName);
        
		 getApplication().saveExtraInfo(preference);
	} // end method savePreference

	/**
	 * Retrieve the import preference for current class info.
	 *
	 * @return the import preference saved for current class info if it exists, otherwise return null.
	 */
	@SuppressWarnings("deprecation")
	protected LyPreference getImportPreference()
	{
		// Try to retrieve import preference for this class info
		LySession				 session = getSession();
		LyApplication			 application = getApplication();
		LyObject				 user = session.getUser();
		String					 userFullId = ((user == null) ? null : user.getId().toString());
		LyPreference			 preference = null, preferenceTmp;
		String					 preferenceName = PREFERENCE_PREFIX + _classInfo.getId();
		LyInfoList<LyPreference> prefs = application.getInfos(LyPreference.class);
		int						 j, n = ((prefs == null) ? 0 : prefs.getSize());
		
		for (j = 0; j < n; j++)
		{
			preferenceTmp = prefs.getInfo(j);
			if (preferenceTmp.getName().equals(preferenceName)
				&& ((userFullId == preferenceTmp.getUser())
					|| userFullId.equals(preferenceTmp.getUser())))
			{
				preference = preferenceTmp;
				break;
			}
		}

		return preference;
	}

	/**
	 * Utility method to create the LyImportController instance and load data Parameters are
	 * submitted by the user in the import form.
	 *
	 * @param  file       path to the imported file.
	 * @param  fieldSep   separator between fields.
	 * @param  valueSep   separator between multiple values.
	 * @param  importType the import type.
	 * @param  fields     list of fields imported from the file.
	 * @param  operations list of operations to be executed.
	 * @param  startLine  the first line of the imported file containing real data.
	 * @param allowOperations indicates if operation are valid in the table
	 * @return true if everything is correct
	 */
	protected boolean buildImportTable(String file, String fieldSep, String valueSep,
		String importType, LyInfoList<LyFieldInfo> fields, BitSet operations, int startLine,
		boolean allowOperations)
	{
		LyAction action = (LyAction)_action.getParameter("_importTableAction");

		getViewManager().setWaiting(true);

		// Create an import table controller with the parent of the current
		// view controller
		LyImportTableController importTableCtrl = new LyImportTableController(getSession(),
				getViewController().getParent(), action, _classInfo);

		importTableCtrl.setImportBehavior((LyImportBehavior)getBehavior());

		// Set the parameters
		importTableCtrl.setImportControllerParent(this);
		importTableCtrl.setFile(file);
		importTableCtrl.setFields(fields);
		importTableCtrl.setFieldSep(fieldSep);
		importTableCtrl.setValueSep(valueSep);
		importTableCtrl.setImportType(importType);
		importTableCtrl.setOperations(operations);
		importTableCtrl.setAllowOperations(allowOperations);
		importTableCtrl.setStartLine(startLine);

		// Load the file containing the data and show the view
		boolean result = importTableCtrl.loadData();

		if (result)
			importTableCtrl.showView();
		else
			importTableCtrl.free();

		getViewManager().setWaiting(false);

		return result;
	} // end method buildImportTable

	/**
	 * Accessor to the view component associated to the id.
	 *
	 * @param  id on of the field identifiers defined as statics : SELECT_FIELDINFO, SELECT_FILE,
	 *            SELECT_VALUE_SEP, SELECT_FIELD_SEP
	 * @return the view component if any, or null if it could not be found
	 */
	protected LyFormField getFormField(String id)
	{
		// Retrieve formField corresponding to id
		LyComponent root = _builder.getComponent();

		return (LyFormField)root.getComponent(id);
	}

	/**
	 * Returns the value of a given parameter set on current import action.<br />
	 * The type of the returned value depends on the parameter.
	 *
	 * @param  parameterName the name of the parameter which value must be retrieved.
	 * @return the value of a given parameter set on current import action or null if the parameter
	 *         is not set on the import action.
	 */
	public Object getParameterValue(String parameterName)
	{
		Object parameterValue = getParameter(parameterName, true);

		if ((parameterValue == null) && (parameterName != SELECT_FIELDINFOS))
			return null;

		if (parameterName == SELECT_FIELDINFOS)
		{
			LyFieldInfo			    field = (LyFieldInfo)parameterValue;
			LyInfoList<LyFieldInfo> result = null;
			int					    np = ((field == null) ? 1 : 2);

			do
			{
				if (field != null)
				{
					if (result == null)
						result = new LyInfoList<LyFieldInfo>();
					result.addInfo(field);
				}
				field = (LyFieldInfo)_action.getParameter(SELECT_FIELDINFOS + np, true);
				np++;
			}
			while (field != null);

			return result;
		}
		else if (parameterName == SELECT_FIELD_SEP)
			return parameterValue;
		else if (parameterName == SELECT_VALUE_SEP)
			return parameterValue;
		else if (parameterName == SELECT_IMPORT_TYPE)
		{
			if ("SPREADSHEET".equals(parameterValue))
				return SELECT_IMPORT_TYPE_SPREADSHEET;
			else if ("RAW".equals(parameterValue))
				return SELECT_IMPORT_TYPE_RAW;
			else if ("XML".equals(parameterValue))
				return SELECT_IMPORT_TYPE_XML;
		}
		else if (parameterName == SELECT_FILE)
			return getEnvironment().getFile((String)parameterValue);
		else if (parameterName == SELECT_START_LINE_NUMBER)
			return new Integer(LyTools.intFromString((String)parameterValue) - 1);
		else if (parameterName == SELECT_OPERATIONS)
		{
			String   operationsValue = (String)parameterValue;
			String[] operations = LyString.stringToList(operationsValue, " ");

			if ((operations == null) || (operations.length == 0))
				return null;

			BitSet importType = new BitSet();
			int    i, n = operations.length;

			for (i = 0; i < n; i++)
			{
				if ("CREATION".equals(operations[i]))
					importType.set(SELECT_OPERATIONS_CREATION);
				else if ("MODIFICATION".equals(operations[i]))
					importType.set(SELECT_OPERATIONS_MODIFICATION);
				else if ("DELETION".equals(operations[i]))
					importType.set(SELECT_OPERATIONS_DELETION);
				else if ("CONSULTATION".equals(operations[i]))
					importType.set(SELECT_OPERATIONS_CONSULTATION);
			}

			return importType;
		}

		return null;
	} // end method getParameterValue

	/**
	 * Accessor to the list of fields graphically selected in the form. This methods converts the
	 * values selected in the view into a list of fields which should be imported.
	 *
	 * @return the list of fieldInfos for import.
	 */
	protected LyInfoList<LyFieldInfo> getGraphicalSelectedFields()
	{
		LyInfoList<LyFieldInfo> fields = new LyInfoList<LyFieldInfo>();
		LyFormField			    formField = getFormField(LyImportController.SELECT_FIELDINFOS);

		if (formField == null)
			return null;

		LyRelation relation = (LyRelation)formField.getFieldItem();
		int		   i, n = relation.getLinkCount();
		LyLink     link;
		String     value;
		LyField    field;

		for (i = 0; i < n; i++)
		{
			link = relation.getLink(i);

			if (link.isSelected())
			{
				value = link.getId();
				field = _classInfo.getField(value);
				fields.addInfo((LyFieldInfo)field);
			}
		}

		return fields;
	} // end method getGraphicalSelectedFields

	/**
	 * Accessor to the list of selected fields in the form. This methods converts the values
	 * selected in the view into a list of fields which should be imported. If no field is selected
	 * in the view, return all the fields.
	 *
	 * @return the list of fieldInfos for import.
	 */
	protected LyInfoList<LyFieldInfo> getSelectedFields()
	{
		@SuppressWarnings("unchecked")
		LyInfoList<LyFieldInfo> fields = (LyInfoList<LyFieldInfo>)getParameterValue(
				SELECT_FIELDINFOS);

		if (fields != null)
			return fields;

		fields = getGraphicalSelectedFields();

		LyFormField formField = getFormField(LyImportController.SELECT_FIELDINFOS);
		LyRelation  relation = (LyRelation)formField.getFieldItem();
		int		    i, n = relation.getLinkCount();
		LyLink	    link;
		String	    value;
		LyField     field;

		// If no field selected, import all the fields
		if (fields.getSize() == 0)
		{
			for (i = 0; i < n; i++)
			{
				link = relation.getLink(i);
				value = link.getId();
				field = _classInfo.getField(value);
				fields.addInfo((LyFieldInfo)field);
			}
		}

		return fields;
	} // end method getSelectedFields

	/**
	 * Accessor to the classinfo Application class associated to the import action.
	 *
	 * @return the classInfo associated to the import form
	 */
	public LyClassInfo getClassInfo()
	{
		return _classInfo;
	}

	/**
	 * Retrieves the view Builder.
	 *
	 * @return the builder instance;
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyImportBuilder(this);
	}

	/**
	 * Indicates if the view for this import controller is resizable.
	 *
	 * @return true if the view is resizable, false otherwise.
	 */
	public boolean isViewResizable()
	{
		return false;
	}

	/**
	 * Accessor to the import type.
	 *
	 * @return the import type selected in the form. The result is one of the following:
	 *
	 *         <ul>
	 *           <li>{@link #SELECT_IMPORT_TYPE_RAW}</li>
	 *           <li>{@link #SELECT_IMPORT_TYPE_XML}</li>
	 *           <li>{@link #SELECT_IMPORT_TYPE_SPREADSHEET}</li>
	 *         </ul>
	 */
	public String getImportType()
	{
		String importType = (String)getParameterValue(SELECT_IMPORT_TYPE);

		if (importType != null)
			return importType;

		LyImportBehavior behavior = (LyImportBehavior)getBehavior();

		importType = behavior.getDefaultImportType();
		// importType = SELECT_IMPORT_TYPE_XML;

		LyFormField field = getFormField(LyImportController.SELECT_IMPORT_TYPE);

		if (field == null)
		{
			LyImportBehavior beh = (LyImportBehavior)getBehavior();

			importType = beh.getDefaultImportType();
		}
		else
		{
			LyChoice choice = (LyChoice)field.getFieldItem();
			LyOption option;
			int		 i, n = choice.getOptionCount();

			// Retrieve the id of the checked option
			for (i = 0; i < n; i++)
			{
				option = choice.getOption(i);

				if (option.isChecked())
					importType = option.getId().intern();
			}
		}

		return importType;
	} // end method getImportType

	/**
	 * Accessor to the import operations.
	 *
	 * @return a bitset containing selected operations. Operations may be one of the following:
	 *
	 *         <ul>
	 *           <li>{@link #SELECT_OPERATIONS_CREATION}</li>
	 *           <li>{@link #SELECT_OPERATIONS_DELETION}</li>
	 *           <li>{@link #SELECT_OPERATIONS_MODIFICATION}</li>
	 *         </ul>
	 */
	public BitSet getImportOperations()
	{
		BitSet operations = (BitSet)getParameterValue(LyImportController.SELECT_OPERATIONS);

		if (operations != null)
			return operations;

		LyFormField field = getFormField(LyImportController.SELECT_OPERATIONS);

		if (field == null)
		{
			LyImportBehavior behavior = (LyImportBehavior)getBehavior();

			operations = behavior.getDefaultOperations();
			return operations;
		}

		return null;
	}

	/**
	 * Accessor to the start line number.
	 *
	 * @return the index of the first line of the imported file containing real data.
	 */
	public int getStartLineNumber()
	{
		Integer startLineNumber = (Integer)getParameterValue(SELECT_START_LINE_NUMBER);

		if (startLineNumber != null)
			return startLineNumber.intValue();

		LyFormField field = getFormField(LyImportController.SELECT_START_LINE_NUMBER);

		if (field == null)
		{
			LyImportBehavior behavior = (LyImportBehavior)getBehavior();

			startLineNumber = behavior.getDefaultStartLineNumber();
			return startLineNumber.intValue();
		}
		LyNumber number = (LyNumber)field.getFieldItem();
		String   value = number.getValue();

		// Internal index is one less than the value typed in by the end-user
		return LyTools.intFromString(value) - 1;
	}

	/**
	 * Method used to sort the fields to import.
	 *
	 * @param  fields The fields to import.
	 * @return A sorted infos list which contains the fields to import.
	 */
	protected LyInfoList<LyFieldInfo> sortFields(LyInfoList<LyFieldInfo> fields)
	{
		// In spreadsheet mode fields may be in a different order
		if (getImportType() == SELECT_IMPORT_TYPE_SPREADSHEET)
			return new LyInfoList<LyFieldInfo>(fields);

		LyInfoList<LyFieldInfo> sortedFields = new LyInfoList<LyFieldInfo>();
		LyInfoList<LyFieldInfo> classFields = _classInfo.getFieldInfos();
		LyFieldInfo			    fieldInfo;

		for (int i = 0; i < classFields.getSize(); i++)
		{
			fieldInfo = classFields.getInfo(i);
			if (fields.contains(fieldInfo))
				sortedFields.addInfo(fieldInfo);
		}

		return sortedFields;
	}

	/**
	 * Checks if this import is in silent mode.
	 * @return true if in silent mode
	 */
	public boolean isSilent()
	{
		return _silent;
	}

} // end class LyImportController
