/**
 * LyImportTableController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyImportTableController.java 40621 2012-08-31 12:55:37Z nhamon $
 */

package leon.control;

import leon.app.*;
import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.control.content.*;
import leon.data.*;
import leon.info.*;
import leon.info.daemon.*;
import leon.info.location.*;
import leon.misc.*;
import leon.peer.file.*;
import leon.peer.xml.*;
import leon.view.*;
import leon.view.event.*;

import java.lang.reflect.*;

import java.util.*;


/**
 * This class implements a table view to import data. Data obtained from the import file are shown
 * as objects in the table : this is only an 'import preview' since data have not yet been added
 * into the view.
 *
 * <P>The user may check if the import procedure was successful and finally decide to import all
 * data or only selected items.</P>
 *
 * <P>It is possible to apply filters and sort as in any table then select the requested items.
 * Usually this view is the result of the _import action associated to the LyImportController.</P>
 *
 * @see LyImportController
 * @see LyExportController
 * @see LySimpleTableController
 */
public class LyImportTableController extends LyMultiPageTableController implements LyControllerContextInterface2
{
	/** File Version */
	public static final String __VERSION = "$Id: LyImportTableController.java 40621 2012-08-31 12:55:37Z nhamon $";

	/** Default image size for icons. */
	public static int DEFAULT_ROW_IMAGE_SIZE = 15;

	/** Internal action : select all. */
	public static final String DEMAND_IMPORT_SELECT_ALL = "__importSelectAll".intern();

	/** Internal action : import selected. */
	public static final String DEMAND_IMPORT_SELECTION = "__importImportSelection".intern();

	/** Internal action : import all items. */
	public static final String DEMAND_IMPORT_ALL = "__importImportAll".intern();

	/** Internal action : view imported objects. */
	public static final String DEMAND_VIEW_IMPORTED = "__importViewImported".intern();

	/** Class info for imported objects. */
	protected LyClassInfo _importClassInfo;

	/** OLD_IMPORT : Object source for imported objects. */
	protected LyTemporaryObjectSource _importObjectSource;

	/** Import fields to use for import. */
	protected LyInfoList<LyFieldInfo> _importFields;

	/** List of operations to be executed. */
	protected BitSet _operations;

	/** File path to use for import. */
	protected String _file;

	/** Buffer to use for import. */
	protected String _buffer;

	/** Separator used between fields. */
	protected String _fieldSep;

	/** Separator used between values. */
	protected String _valueSep;

	/** Returns a boolean that indicates if operations field must be displayed. */
	protected boolean _allowOperations;

	/**
	 * The import type. Either LyImportController.SELECT_IMPORT_TYPE_RAW or
	 * LyImportController.SELECT_IMPORT_TYPE_XML
	 */
	protected String _importType;

	/** The index of the first line of the imported file containing real data. */
	protected int _startLine;

	/** OLD_IMPORT List of the struct classInfos of _importClassInfo. */
	protected LyInfoList<LyClassInfo> _structClassInfos;

	/** OLD_IMPORT List of the struct classInfos of _importClassInfo. */
	protected LyInfoList<LyClassInfo> _importStructClassInfos;

	/** OLD_IMPORT The objectSource of the structClassInfos. */
	protected Vector<LyObjectSource> _structObjectSource;

	/** OLD_IMPORT List of associations between imported struct objects and real struct objects. */
	protected Hashtable<LyObject, LyObject> _realStructObjects;

	/** List of associations between objects and import errors. */
	protected Hashtable<LyObject, String> _importErrors;

	/** OLD_IMPORT List of temporary objects imported during the process. */
	protected LyObjectList _temporaryImportedObjects;

	/** OLD_IMPORT List of real objects imported during the process. */
	protected LySimpleObjectList _realImportedObjects;

	/** List of objects that may be deleted during the import process. */
	protected LySimpleObjectList _objectsToDelete;

	/** OLD_IMPORT List of associations between imported (temporary) objects and real objects. */
	protected Hashtable<LyObject, LyObject> _realObjects;

	/** Context for this controller. */
	protected LyContext _controllerContext;

	/** OLD_IMPORT Association between local imported fields and temporary fields. */
	protected Hashtable<LyFieldInfo, LyFieldInfo> _localImportedFields;

	/** Indicates if something was imported in new_IMPORT to enable the validate button. */
	protected boolean _somethingWasImported;

	/**
	 * TEMPORARY parameter due to import rewriting. Indicates if the import is an old import or a
	 * new import This attribute and all code sequence between if (OLD_IMPORT) { } will be
	 * deprecated and removed
	 */
	public boolean OLD_IMPORT = true;

	/**
	 * Last error that was shown during import This may be used to check by code if import was Ok.
	 */
	protected String _lastErrorMessage;

	/**
	 * Indicates Objects modified outside the table (for example commited / modified by code or by
	 * GUI) : won't be rollbacked.
	 */
	protected LySimpleObjectList _modifiedObjects;

	/** Indicates if there is a Gui */
	protected boolean _withoutGui;

	/**
	 * The import behavior associated with the import action
	 */
	protected LyImportBehavior _importBehavior;

	/**
	 * Import controller.
	 */
	protected LyImportController _importControllerParent;

	/**
	 * Constructor. Must be used for import action with GUI.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller).
	 * @param action    related action.
	 * @param classInfo classInfo which indicates which kind of objects we expect in import
	 */
	public LyImportTableController(LySession session, LyController parent, LyAction action, LyClassInfo classInfo)
	{
		this(session, parent, action, classInfo, false);
	}

	/**
	 * Constructor. Must be used for import action without GUI.
	 *
	 * @param session    current session associated to the controller. The controller automatically
	 *                   registers on the specified session.
	 * @param parent     parent controller (or null for 'root' controller).
	 * @param action     related action.
	 * @param classInfo  classInfo which indicates which kind of objects we expect.
	 * @param withoutGui Indicates if the import has some GUI or not. This parameter must be set to
	 *                   true to execute an import action by code. in import
	 */
	public LyImportTableController(LySession session, LyController parent, LyAction action, LyClassInfo classInfo,
		boolean withoutGui)
	{
		super(session, parent, action, classInfo, null);

		_structObjectSource = new Vector<LyObjectSource>();

		_withoutGui = withoutGui;
		_classInfo = classInfo;
		_file = null;
		_buffer = null;
		_fieldSep = null;
		_valueSep = null;
		_workSpace = null;
		_selections = null;
		_realObjects = null;
		_objectsToDelete = null;
		_importType = null;
		_startLine = 0;
		_operations = new BitSet();
		_operations.set(LyImportController.SELECT_OPERATIONS_CREATION);
		_operations.set(LyImportController.SELECT_OPERATIONS_MODIFICATION);
		_allowOperations = true;
		_currentStatusSort = 0;
		_temporaryImportedObjects = null;
		_somethingWasImported = false;
		_modifiedObjects = null;

		if (_action != null)
		{
			// The behavior MUST be an instance of LyImportTableBehavioir and not only an ImportBehavior
			Class<?>    behaviorClass = _action.getBehaviorClass();
			Class<?>    behaviorInterface = getBehaviorInterface();

			if ((behaviorClass == null) || (!behaviorInterface.isAssignableFrom(behaviorClass)))
			{

				// The import table controller is expecting an ImportTableBehavior (not an ImportBehavior).
				// This may be confusing when importing without a GUI
				if (session != null)
					session.getEnvironment().logTrace("BF AppEngine : invalid behavior " + behaviorClass
						+ " for controller " + getClass() + ", because it doesn't implement " + behaviorInterface
						+ ((_action == null) ? ""
							: (". Check the behavior attribute of your action " + _action.getId())));

				_action = getApplication().getActionWithId(LyAction.ID_IMPORT_TABLE, _classInfo); // Use internal importTable action
			}
		}
	}

	/**
	 * Free this controller on request of the specified controller.
	 *
	 * @param controller controller instance which requested the freeing of this controller. If the
	 *                   controller is the parent, this controller is freed otherwise the method
	 *                   does nothing.
	 */
	public void free(LyController controller)
	{
		if (controller != _parent)
			return;

		// If not committed yet, just forget modifications
		rollback();

		if (OLD_IMPORT)
		{
			LyInfoList<LyFieldInfo> importFields = _importClassInfo.getFieldInfos();
			LyFieldInfo			    fieldInfo;
			LyClassInfo			    classInfo;

			for (int i = 0; i < importFields.getSize(); i++)
			{
				fieldInfo = importFields.getInfo(i);
				if (fieldInfo instanceof LyStructFieldInfo)
					getApplication().removeInfo(fieldInfo);
			}

			_importClassInfo.free();

			for (int i = 0; i < _importStructClassInfos.getSize(); i++)
			{
				classInfo = _importStructClassInfos.getInfo(i);
				importFields = classInfo.getFieldInfos();

				for (int j = 0; j < importFields.getSize(); j++)
				{
					fieldInfo = importFields.getInfo(j);
					if (fieldInfo instanceof LyStructFieldInfo)
						getApplication().removeInfo(fieldInfo);
				}

				classInfo.free();
			}

			_temporaryImportedObjects = null;
			_realImportedObjects = null;
			_realObjects = null;
		}

		super.free(controller);

		if (_importObjectSource != null)
			_importObjectSource.free();

		_localImportedFields = null;
		_modifiedObjects = null;
	} // end method free

	/**
	 * Accessor to the internal builder. If no builder is set, the method returns the reference to a
	 * default builder (invokes getDefaultBuilder).
	 *
	 * @return the builder which creates the XML description of the view. Returned value is not null
	 *         and several invocation on getBuilder always return the same instance.
	 */
	public LyBuilder getBuilder()
	{
		if (_withoutGui)
			return null;

		return super.getBuilder();
	}

	/**
	 * Accessor to the view component instance. (XML description of the view) This instance is
	 * created by the builder, and multiple calls to this method always return the same instance.
	 *
	 * @return the LyComponent instance of the view package, memory representation (parsed) of the
	 *         XML description of the view.
	 */
	public LyComponent getComponent()
	{
		if (_withoutGui)
			return null;

		return super.getComponent();
	}

	/**
	 * Accessor view that contains the component (getComponent) of the controller. This is a
	 * shortcut to getViewController.getView(). Returned value depends on the nature of the
	 * controller.
	 *
	 * @return the LyView instance (memory representation of the XML view) which contains the
	 *         current LyComponent.
	 */
	public LyView getView()
	{
		if (_withoutGui)
			return null;

		return super.getView();
	}

	/**
	 * Reset component.
	 *
	 * @param full Boolean indicating if a full reset (including reset of datas) is requiered or not
	 *             (only reset graphical area).
	 */
	public void reset(boolean full)
	{
		if (_withoutGui)
			return;

		// Reset must not reload data (performed in loadData() method)
		super.reset(false);
	}

	/**
	 * Accessor to the view controller (Frame or dialog controller) which is the container of this
	 * view. This method must be implemented in subclasses depending on the nature of the component.
	 *
	 * @return the reference to a view controller, never returns null if the controller is
	 *         associated to a view.
	 */
	public LyViewController getViewController()
	{
		if (_withoutGui)
			return null;

		return super.getViewController();
	}

	/**
	 * Hide the view. The effect of this command depends on the nature of the application. If the
	 * display is a fat client for example (windows (awt) / swing) this methods should close the
	 * window or hide the view.
	 *
	 * <P>This method should consider that the page is closed and should disappear.</P>
	 *
	 * @param free Indicates if only HIDDEN, not FREED (true).
	 */
	public void hideView(boolean free)
	{
		if (_withoutGui)
			return;

		super.hideView(free);
	}

	/**
	 * Initialize this instance.
	 */
	protected void init()
	{
		init(false);
	}

	/**
	 * Initialize this instance.
	 * @param withoutGui indicates if this controller has a gui
	 */
	protected void init(boolean withoutGui)
	{
		if (!withoutGui)
			super.init();

		_filter = (LyFilter)getParameter("_filter", true);
		_filterActive = false;
		_currentStatusSort = 0;
		_localImportedFields = new Hashtable<LyFieldInfo, LyFieldInfo>();
	}

	/**
	 * Invokes an action on the specified class and object list.
	 *
	 * @param action    the action which is executed.
	 * @param classInfo the target class of the action.
	 * @param objects   The list of objects on which the action applies.
	 * @see   LyController#invokeAction(LyAction, LyClassInfo, LyObjectList, boolean)
	 */
	protected LyController invokeAction(LyAction action, LyClassInfo classInfo, LyObjectList objects, boolean show)
	{
		if (OLD_IMPORT)
		{
			if (classInfo == _classInfo)

				// Redirect on import class info
				classInfo = _importClassInfo;
		}

		return super.invokeAction(action, classInfo, objects, show);
	}

	/**
	 * Initialize the import fields
	 */
	protected void initImportFields()
	{
		if (OLD_IMPORT)
		{
			LyFieldInfo fieldInfo;

			_importFields = new LyInfoList<LyFieldInfo>();
			_importErrors = new Hashtable<LyObject, String>();

			if (_fields != null)
			{
				for (int i = 0; i < _fields.getSize(); i++)
				{
					fieldInfo = _fields.getInfo(i);
					if (fieldInfo instanceof LyStructFieldInfo)
						fieldInfo = _importClassInfo.getFieldInfo("__" + fieldInfo.getId());

					_importFields.addInfo(fieldInfo);
				}
			}
		}
	}

	/**
	 * Factory to build the import content controller for XML import.
	 *
	 * @param  session The current session.
	 * @param  parent  The parent of the build controller.
	 * @param  action  The action definition for import.
	 * @return The inmportContentXmlController.
	 */
	public LyImportContentXmlController importContentXmlFactory(LySession session, LyController parent, LyAction action)
	{
		return new LyImportContentXmlController(session, parent, action);
	}

	/**
	 * Sets import controller to be able to get parameters set on it.
	 * @param importController the import controller
	 */
	protected void setImportControllerParent(LyImportController importController)
	{
		_importControllerParent = importController;
	}

	/**
	 * Gets import controller parent.
	 * @return the import controller
	 */
	public LyImportController getImportControllerParent()
	{
		return _importControllerParent;
	}

	/**
	 * Gets import action.
	 * @return the import action
	 */
	public LyAction getImportAction()
	{
		return ((_importControllerParent == null) ? null : _importControllerParent.getAction());
	}

	/**
	 * Internal method to get file data provider.
	 * @return the import file data provider
	 */

	protected LyImportFileDataProvider getImportFileDataProvider()
	{
		if (_importType == LyImportController.SELECT_IMPORT_TYPE_XML)
			return null;

		LyFileLocation			 location = new LyFileLocation(_importClassInfo.getId(), null, LyFileLocation.Type.INFO,
				_fieldSep, _valueSep);
		LyImportFileDataProvider ifdp = null;
		LyApplication			 application = getApplication();
		LyAction				 action = getImportAction();
		String					 providerClassName = ((action == null) ? null
				: (String)action.getParameter("_importFileProvider"));

		if (providerClassName != null)
		{
			try
			{
				Class<?>   c = application.loadClass(providerClassName);
				Class<?>[] types = new Class[2];

				types[0] = LyApplication.class;
				types[1] = LyLocation.class;

				Constructor<?> constructor = c.getConstructor(types);
				Object[]    parameters = new Object[2];

				parameters[0] = application;
				parameters[1] = location;
				ifdp = (LyImportFileDataProvider)constructor.newInstance(parameters);
			}
			catch (Exception e)
			{
				getEnvironment().writeException(e);
			}
		}

		if (ifdp == null)
			ifdp = new LyImportFileDataProvider(application, location);

		ifdp.setFilePath(_file);
		ifdp.validate();

		if (_importType == LyImportController.SELECT_IMPORT_TYPE_SPREADSHEET)
		{
			ifdp.setDecodingType(LyFileRequestEncoder.SPREADSHEET_ENCODING);

			int			   i, n = ((_importFields == null) ? 0 : _importFields.getSize());
			Vector<String> fieldIds = new Vector<String>(n);
			LyInfo		   info;

			for (i = 0; i < n; i++)
			{
				info = _importFields.getInfo(i);
				fieldIds.addElement(info.getId());
			}

			ifdp.setFileFieldIds(fieldIds);
			ifdp.setLogicalClass(_classInfo);
			ifdp.setStartLine(_startLine);
		}

		return ifdp;
	}

	/**
	 * Load the data contained in the file.
	 * @return true if successfull
	 */
	public boolean loadData()
	{
		return loadData(true);
	}

	/**
	 * Load the data contained in the file.
	 *
	 * @param full indicates if a full reload must be performed.
	 * @return true if successfull
	 */
	public boolean loadData(boolean full)
	{
		LyObjectList  importedObjects;
		LyApplication application = getApplication();
		String		  errorMessage = null;

		_importErrors = new Hashtable<LyObject, String>();

		LyImportContentController ctrl = null;

		if (_importBehavior != null)
			ctrl = _importBehavior.importContentFactory(getSession(), this, _action, _importType);

		if (ctrl != null)
		{
			if (_environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
			{
				_environment.logOut(">>> Xml Import start : ");
				if (_environment.isDebugLevel(LyEnvironment.DEBUG_ALL))
					_environment.writeException(new Throwable());
			}
			java.util.Date date = new java.util.Date();

			// NEW IMPORT
			OLD_IMPORT = false;
			_importFields = _fields;
			if (_importFields == null)
				_importFields = super.getFields();

			ctrl.setClassInfo(_classInfo);
			ctrl.setFields(_fields);

			if (_buffer != null)
				importedObjects = ctrl.loadData(_buffer);
			else
				importedObjects = ctrl.loadDataFile(_file);

			setLastErrorMessage(ctrl.getLastErrorMessage());
			updateErrors(importedObjects);
			ctrl.free();

			if (_environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
				_environment.logOut("<<< End of Xml Import ("
						+ ((new java.util.Date()).getTime() - date.getTime()) + " ms)");
		}
		else
		{
			if (_environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
			{
				_environment.logOut(">>> Import start : ");
				if (_environment.isDebugLevel(LyEnvironment.DEBUG_ALL))
					_environment.writeException(new Throwable());
			}
			java.util.Date date = new java.util.Date();

			_realObjects = new Hashtable<LyObject, LyObject>();

			// OLD Import indicates code which is obsolete when using a logical level Import
			OLD_IMPORT = true; // Should be removed later

			// Create a new class info for imported objects from the given one
			if (_importClassInfo == null)
				_importClassInfo = createClassInfo(_classInfo);

			initImportFields();

			LyImportFileDataProvider ifdp = getImportFileDataProvider();

			// We load and start with empty imported lists
			_temporaryImportedObjects = new LySimpleObjectList();
			_realImportedObjects = new LySimpleObjectList();

			_importObjectSource = (LyTemporaryObjectSource)application.getObjectSource(_importClassInfo);
			_importObjectSource.setHandleErrorEvents(false); // Do not propagate errors on temp. objects
			_importObjectSource.setCache(new LyInfoCache(LyInfoCache.CachePolicy.FULL, null));

			LyClassInfo			    importStructClassInfo;
			LyTemporaryObjectSource objectSource;

			for (int i = 0; i < _importStructClassInfos.getSize(); i++)
			{
				importStructClassInfo = _importStructClassInfos.getInfo(i);
				objectSource = (LyTemporaryObjectSource)getApplication().getObjectSource(importStructClassInfo);
				_structObjectSource.add(objectSource);
				objectSource.setCache(new LyInfoCache(LyInfoCache.CachePolicy.FULL, null));
			}

			// Read the file and create the objects
			try
			{
				importedObjects = new LyClassWorkSpace(getSession(), _importClassInfo, _importFields, null, null, false,
						true);
			}
			catch (Exception e)
			{
				getEnvironment().writeException(e);

				String		 message = getEnvironment().getMessage("LY_ER_IMPORT_FILE", _file);
				LyController messageController = this;

				if ((getParent() != null) && (getParent().getParent() != null))
					messageController = getParent().getParent();

				messageController.showError(message);

				_importObjectSource.free();
				_importObjectSource = null;
				return false;
			}
			finally
			{
				if (_environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
					_environment.logOut("<<< End of Import ("
							+ ((new java.util.Date()).getTime() - date.getTime()) + " ms)");
			}
			//  swap our empty list with the new imported list
			_temporaryImportedObjects = importedObjects;

			LyImportTableBehaviorInterface behavior = (LyImportTableBehaviorInterface)getBehavior();

			errorMessage = behavior.getErrorLinesMessage(this, ifdp.getErrorLines());


		}

		// check if there is some results
		if (errorMessage != null)
			showError(errorMessage);

		if ((importedObjects == null) || (importedObjects.getSize() == 0))
		{
			if (errorMessage == null)
			{
				String message;

				if (getLastErrorMessage() == null)
					message = getEnvironment().getMessage("LY_ER_IMPORT_CLASSINFO", _file, _classInfo.getName());
				else
					message = getEnvironment().getMessage("LY_ER_IMPORT_MESSAGE", _file, getLastErrorMessage());

				LyController messageController = this;

				if ((getParent() != null) && (getParent().getParent() != null))
					messageController = getParent().getParent();

				messageController.showError(message);
			}

			if (OLD_IMPORT)
			{
				_importObjectSource.free();
				_importObjectSource = null;
			}

			return false;
		}

		LyObject		   importObject;
		LyObjectId		   realObjectId;
		LyObject		   realObject;
		LyObject		   realSessionObject;
		int				   nHiddenObjects = 0;
		boolean			   isHidden;
		LyDataSource	   realDataSource = getSession().getDataSource(_classInfo);
		LySimpleObjectList context = new LySimpleObjectList();

		if (OLD_IMPORT)
		{
			_realStructObjects = new Hashtable<LyObject, LyObject>();
			LyClassInfo importStructClass, structClass;
			LyObject    importStructObject;
			Vector<Object>	    importStructObjects;

			// Imported Struct Objects
			for (int i = 0; i < _importStructClassInfos.getSize(); i++)
			{
				importStructClass = _importStructClassInfos.getInfo(i);
				structClass = _structClassInfos.getInfo(i);
				importStructObjects = importStructClass.getDataSource().getCache().getObjects();

				for (int j = 0; j < importStructObjects.size(); j++)
				{
					importStructObject = (LyObject)importStructObjects.get(j);
					realObjectId = new LyObjectId(structClass, importStructObject.getObjectId());
					realObject = realObjectId.getLogicalObject();

					// Check values for new object
					checkImportObject(importStructObject, importStructClass);

					if (realObject != null)
						_realStructObjects.put(importStructObject, realObject);
				}
			}

			// Now update local fields from "imported local fields"
			/*n = _importClassInfo.getFieldInfoCount();
			                for (i = 0; i < n; i++)
			                {
			                        fieldInfo = _importClassInfo.getFieldInfo(i);

			                        if (fieldInfo.getId().endsWith(LyXmlExportRequestEncoder.LOCAL_FIELD_IMPORT_EXTENSION))
			                        {
			                                // This is a "dummy" fieldInfo : get the related local field
			                                int p = fieldInfo.getId().length() - LyXmlExportRequestEncoder.LOCAL_FIELD_IMPORT_EXTENSION.length();
			                                LyFieldInfo fieldInfo2 = _importClassInfo.getFieldInfo(fieldInfo.getId().substring(0,p));

			                                int j, q = importedObjects.getSize();
			                                for (j = 0; j< q; j++)
			                                {
			                                        importObject = importedObjects.getObject(j);

			                                        importObject.setValue(fieldInfo2, importObject.getValue(fieldInfo));
			                                }
			                        }
			                }
			*/
		}

		String message;
		int    i, n = importedObjects.getSize();

		// Build the list of real objects that are not in the import file
		if (full)
		{
			_objectsToDelete = new LySimpleObjectList();

			if (_operations.get(LyImportController.SELECT_OPERATIONS_DELETION))
			{
				LyClassWorkSpace realObjectsWs = new LyClassWorkSpace(getSession(), _classInfo, _fields, null, null, false,
						false);
				int				 j, m = realObjectsWs.getSize();

				for (j = 0; j < m; j++)
					_objectsToDelete.addObject(realObjectsWs.getObject(j));
			}
		}

		// Imported Objects
		for (i = 0; i < n; i++)
		{
			importObject = importedObjects.getObject(i);
			realObjectId = new LyObjectId(_classInfo, importObject.getObjectId());

			realObject = realObjectId.getLogicalObject();
			getEnvironment().logTrace("## Try to find realObject : classInfo " + _classInfo.getId() + " import : "
				+ importObject.getObjectId() + " in " + realObjectId.getFullString() + " --> " + realObject);
			realSessionObject = realObject;
			isHidden = false;

			if (realDataSource instanceof LyFilteredObjectSource)
			{
				realSessionObject = (LyObject)realDataSource.getObject(realObjectId);

				// If an object exists on the application but not in the
				// session (for instance a FilteredObjectSource may have
				// been created on the session), show a message to the user
				// and do not display this object.
				if ((realObject != null) && (realSessionObject == null))
				{
					nHiddenObjects++;
					realObject = null;
					isHidden = true;
				}
			}

			if ((_objectsToDelete != null) && (realObject != null) && _objectsToDelete.contains(realObject))
				_objectsToDelete.removeObject(realObject);

			// Check values for new object
			if (OLD_IMPORT)
			{
				message = checkImportObject(importObject);

				if (message != null)
				{
					LyController messageController = this;

					if ((getParent() != null) && (getParent().getParent() != null))
						messageController = getParent().getParent();

					messageController.showError(message);

					_importObjectSource.free();
					_importObjectSource = null;

					for (int j = 0; j < _structObjectSource.size(); j++)
						((LyTemporaryObjectSource)_structObjectSource.get(j)).free();

					_structObjectSource = null;
					_realStructObjects = null;
					_realObjects = null;

					return false;
				}
			}
			if (OLD_IMPORT)
			{
				// Check if creation operation is requested
				if (!_operations.get(LyImportController.SELECT_OPERATIONS_CREATION) && (realObject == null))
					isHidden = true;

				// Check if creation modification is requested
				if (!_operations.get(LyImportController.SELECT_OPERATIONS_MODIFICATION) && (realObject != null))
				{
					LyInfoList<LyFieldInfo> differentValues = getDifferences(importObject, realObject);

					if ((differentValues != null) && (differentValues.getSize() > 0))
						isHidden = true;
				}

				// Check if consultation operation is requested
				if (!_operations.get(LyImportController.SELECT_OPERATIONS_CONSULTATION) && (realObject != null))
				{
					LyInfoList<LyFieldInfo> differentValues = getDifferences(importObject, realObject);

					if ((differentValues == null) || (differentValues.getSize() == 0))
						isHidden = true;
				}
			}
			else
			{
				isHidden = true;
				if (isNewObject(importObject))
				{
					if (_operations.get(LyImportController.SELECT_OPERATIONS_CREATION))
						isHidden = false;
				}
				// Check if creation modification is requested
				else
				{
					LyInfoList<LyFieldInfo> differentValues = isSameObject(realObject);
					if ((differentValues != null) && _operations.get(LyImportController.SELECT_OPERATIONS_MODIFICATION))
						isHidden = false;

					if ((differentValues == null) && _operations.get(LyImportController.SELECT_OPERATIONS_CONSULTATION))
						isHidden = false;
				}
			}
			if (!isHidden)
				context.addObject(importObject);

			if (OLD_IMPORT)
			{
				if (realObject != null)
					_realObjects.put(importObject, realObject);
			}
		}

		if (full
			&& (_objectsToDelete != null)
			&& (_objectsToDelete.getSize() > 0))
		{
			if (_importClassInfo == null)
				_importClassInfo = createClassInfo(_classInfo);

			try
			{
				LyObject   objectToDelete, deletedImportObject;
				LyValueSet values;
				LyValue    value;

				for (i = _objectsToDelete.getSize() - 1; i >= 0; i--)
				{
					deletedImportObject = _importClassInfo.createInstance();
					objectToDelete = _objectsToDelete.getObject(i);
					values = objectToDelete.getValues();

					for (int j = 0; j < values.getSize(); j++)
					{
						value = values.getFieldValue(j);
						deletedImportObject.setValue(value.getField().getId(), value.getValue());
					}

					deletedImportObject.create();
					context.addObject(deletedImportObject);
					_objectsToDelete.removeObject(objectToDelete);
					_objectsToDelete.addObject(deletedImportObject, i);

					if (OLD_IMPORT)
						_realObjects.put(deletedImportObject, objectToDelete);
				}
			}
			catch (Exception e)
			{
				LyController messageController = this;

				if ((getParent() != null) && (getParent().getParent() != null))
					messageController = getParent().getParent();

				getEnvironment().writeException(e);

				message = getEnvironment().getMessage("LY_ER_IMPORT_FILE", _file);

				messageController.showError(message);

				_importObjectSource.free();
				_importObjectSource = null;
				return false;
			}
		}

		//Check the visibility of objects to delete
		for (i = _objectsToDelete.getSize() - 1; i >= 0; i--)
		{
			LyObject objectToDelete = _objectsToDelete.getObject(i);

			if (!_operations.get(LyImportController.SELECT_OPERATIONS_DELETION) && context.contains(objectToDelete))
				context.removeObject(objectToDelete);
			else if (_operations.get(LyImportController.SELECT_OPERATIONS_DELETION)
				&& !context.contains(objectToDelete))
				context.addObject(objectToDelete);
		}

		if (!OLD_IMPORT)
			_realImportedObjects = context;

		_workSpace = new LySimpleWorkSpace(getSession(), context, true);
		_workSpace.addListener(this);

		if (full || (getSort() == null))
			_workSpace.setSort(getDefaultSort());
		else
			_workSpace.setSort(getSort());

		if (nHiddenObjects > 0)
		{
			LyEnvironment environment = getEnvironment();

			message = environment.getMessage("LY_ER_IMPORT_HIDDEN_OBJECTS",
					new Object[] { _file, new Integer(nHiddenObjects), _classInfo.getName() });

			showWarning(message);
		}

		return true;
	} // end method loadData

	/**
	 * Updates the errors on imported objects.
	 * @param list a list of objects
	 */
	protected void updateErrors(LyObjectList list)
	{
		if (list == null)
			return;

		for (int i = 0; i < list.getSize(); i++)
		{
			LyObject object = list.getObject(i);

			if (object.getErrorLevel() == LyObject.LEVEL_ERROR)
			{
				StringBuffer sb = new StringBuffer();
				LyError		 error;
				int			 n1 = object.getErrorsCount();

				for (int i1 = n1 - 1; i1 >= 0; i1--)
				{
					error = object.getError(i1);
					if (error.getLevel() == LyObject.LEVEL_ERROR)
					{
						if (sb.length() > 0)
							sb.append('\n');

						sb.append(error.getMessage(getEnvironment()));
					}
				}

				if (sb.length() > 0)
					_importErrors.put(object, sb.toString());
				else
					_importErrors.remove(object);
			}
		}
	} // end method updateErrors

	/**
	 * Checks the values for the newly created object. Parse all the relations of the object and
	 * check if related objects are "visible" in this session.
	 *
	 * @param      importObject object that must be checked.
	 * @return     An error message or null
	 * @deprecated Will not be used in newImport
	 */
	protected String checkImportObject(LyObject importObject)
	{
		if (OLD_IMPORT)
		{
			LyDataSource	    realDataSource;
			LyFieldInfo		    fieldInfo;
			LyRelationFieldInfo relation;
			LySimpleObjectList  relatedObjects;
			LyObject		    relatedObject, realSessionObject;
			LyObjectId		    realObjectId;
			Object			    defaultValue;
			boolean			    deleteFieldValue;
			int				    i, j, nObjects, nFields = _importClassInfo.getFieldInfoCount();
			Object			    lyValue;

			for (i = 0; i < nFields; i++)
			{
				deleteFieldValue = false;
				fieldInfo = _importClassInfo.getFieldInfo(i);

				if (!_importFields.contains(fieldInfo))
				{
					defaultValue = getDefaultValue(fieldInfo);
					if (defaultValue != null)
					{
						importObject.setValue(fieldInfo, defaultValue);
						importObject.set();
					}

					getControllerContext();
					continue;
				}

				lyValue = importObject.getValue(fieldInfo);

				if (lyValue != null)
					importObject.setValue(fieldInfo, lyValue);

				if (fieldInfo instanceof LyRelationFieldInfo)
				{

					// If relation, retrieve the related objects
					relation = (LyRelationFieldInfo)fieldInfo;
					relatedObjects = relation.getObjects(importObject);
					defaultValue = getDefaultValue(relation);

					if ((relatedObjects == null) || (relatedObjects.getSize() == 0))
						continue;

					nObjects = relatedObjects.getSize();

					for (j = 0; j < nObjects; j++)
					{

						// For each object, check if existing in the current session
						relatedObject = relatedObjects.getObject(j);
						realDataSource = getSession().getDataSource(relatedObject.getClassInfo());
						realObjectId = relatedObject.getObjectId();
						realSessionObject = (LyObject)realDataSource.getObject(realObjectId);

						if (realSessionObject == null)
						{
							deleteFieldValue = true;
							break;
						}
					}

					if (deleteFieldValue)
					{
						// If related object is not visible in the current session
						// set value for import object with its default value.

						importObject.setValue(relation, defaultValue);
						importObject.set();
					}
				}
			}
		}

		return null;
	} // end method checkImportObject

	/**
	 * Checks the values for the newly created object. Parse all the relations of the object and
	 * check if related objects are "visible" in this session.
	 *
	 * @param      importObject object that must be checked.
	 * @param      classInfo    classInfo of the object.
	 * @deprecated Will not be used in newImport
	 */
	protected void checkImportObject(LyObject importObject, LyClassInfo classInfo)
	{
		if (OLD_IMPORT)
		{
			LyDataSource	    realDataSource;
			LyFieldInfo		    fieldInfo;
			LyRelationFieldInfo relation;
			LySimpleObjectList  relatedObjects;
			LyObject		    relatedObject, realSessionObject;
			LyObjectId		    realObjectId;
			Object			    defaultValue;
			boolean			    deleteFieldValue;
			int				    i, j, nObjects, nFields = classInfo.getFieldInfoCount();

			for (i = 0; i < nFields; i++)
			{
				deleteFieldValue = false;
				fieldInfo = classInfo.getFieldInfo(i);
				String value;
				Date   date;

				if (fieldInfo instanceof LyChoiceFieldInfo)
				{
					LyChoiceValue v = importObject.getChoiceValue((LyChoiceFieldInfo)fieldInfo);

					value = v.getString();
					importObject.setValue(fieldInfo, fieldInfo.parse(value));
				}
				else if (fieldInfo instanceof LyDateFieldInfo)
				{
					date = (Date)importObject.getValue(fieldInfo);
					importObject.setValue(fieldInfo, date);
				}
				else
				{
					value = importObject.getValue(fieldInfo).toString();
					importObject.setValue(fieldInfo, fieldInfo.parse(value));
				}

				if (fieldInfo instanceof LyRelationFieldInfo)
				{

					// If relation, retrieve the related objects
					relation = (LyRelationFieldInfo)fieldInfo;
					relatedObjects = relation.getObjects(importObject);
					defaultValue = getDefaultValue(relation);

					if ((relatedObjects == null) || (relatedObjects.getSize() == 0))
						continue;

					nObjects = relatedObjects.getSize();

					for (j = 0; j < nObjects; j++)
					{

						// For each object, check if existing in the current session
						relatedObject = relatedObjects.getObject(j);
						realDataSource = getSession().getDataSource(relatedObject.getClassInfo());
						realObjectId = relatedObject.getObjectId();
						realSessionObject = (LyObject)realDataSource.getObject(realObjectId);

						if (realSessionObject == null)
						{
							deleteFieldValue = true;
							break;
						}
					}

					if (deleteFieldValue)
					{
						// If related object is not visible in the current session
						// set value for import object with its default value.

						importObject.setValue(relation, defaultValue);
						importObject.set();
					}
				}
			}
		}
	} // end method checkImportObject

	/**
	 * Invoked when an action notification is sent to this listener.
	 *
	 * <P>This method is called by demandPerformed.</P>
	 *
	 * <P>At this level, the implementation handles demands of type LyActionDemand.TYPE_ACTION :
	 * DEMAND_IMPORT_SELECT_ALL, DEMAND_IMPORT_SELECTION, DEMAND_IMPORT_ALL,
	 * LyFormController.DEMAND_CANCEL</P>
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		String message = null;

		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			String   source = demand.getSource();
			String[] parameters = LyString.stringToList(source, ID_SEP);
			String   id = parameters[0].intern();

			if (id == DEMAND_IMPORT_SELECT_ALL)
			{
				if ((_selections != null) && (_selections.getSize() == _workSpace.getSize()))
					selectObjects(null, false);
				else
					selectObjects(null, true);

				return;
			}
			else if (id == DEMAND_IMPORT_SELECTION)
			{
				message = checkFields(_selections);
				if (message != null)
				{
					showError(message);
					return;
				}

				// Check controller context before committing
				if ((!_withoutGui) && (_controllerContext != null)
					&& ((_controllerContext.getStatus() == LyContext.Status.ERROR)
						|| (_controllerContext.getErrorLevel() == LyObject.LEVEL_ERROR)))
					showErrors();

				importObjects(_selections, false);
				return;
			}
			else if (id == DEMAND_IMPORT_ALL)
			{
				message = checkFields(_workSpace);
				if (message != null)
				{
					showError(message);
					return;
				}

				// Check controller context before committing
				if ((!_withoutGui) && (_controllerContext != null)
					&& ((_controllerContext.getStatus() == LyContext.Status.ERROR)
						|| (_controllerContext.getErrorLevel() == LyObject.LEVEL_ERROR)))
					showErrors();

				importObjects(_workSpace, false);
				return;
			}
			else if (id == DEMAND_VIEW_IMPORTED)
			{
				if ((getRealImportedObjects() == null) || (getRealImportedObjects().getSize() == 0))
				{
					message = "LY_NO_IMPORTED_OBJECT";
					showError(message);
					return;
				}

				LyAction			    tableAction = getApplication().getAction("_table");
				LyObjectList		    list = getRealImportedObjects();
				LyInfoList<LyClassInfo> classInfos = list.getClasses();

				for (int i = 0; i < classInfos.getSize(); i++)
				{
					LyClassInfo		   classInfo = classInfos.getInfo(i);
					LySimpleObjectList subList = new LySimpleObjectList();

					for (int j = 0; j < list.getSize(); j++)
					{
						LyObject object = list.getObject(j);

						if (object.instanceOf(classInfo))
							subList.addObject(object);
					}

					getSession().invokeAction(this, tableAction, classInfo, subList, true);
				}

				return;
			}
			else if (id == LyFormController.DEMAND_CANCEL)
			{
				cancel();
				return;
			}
			else if (id == LyFormController.DEMAND_VALIDATE)
			{
				rollbackNonImportedObjects();

				// Check controller context before committing
				if ((!_withoutGui) && (_controllerContext != null)
					&& ((_controllerContext.getStatus() == LyContext.Status.ERROR)
						|| (_controllerContext.getErrorLevel() == LyObject.LEVEL_ERROR)))
				{
					showErrors();
					return;
				}

				// FFT00081/ipboss: Check the number of objects successfully committed
				int				 nop = _controllerContext.getOperations().size();
				Vector<LyObject> updatedObjects = new Vector<LyObject>();
				LyObject		 updatedObject;

				for (int k = 0; k < nop; k++)
				{
					updatedObject = _controllerContext.getOperations().get(k).getObject();
					if (!updatedObjects.contains(updatedObject))
						updatedObjects.addElement(updatedObject);
				}

				commit();
				hideView();

				// FFT00081/ipboss: Check controller context before committing.
				// Else, when successful, display a confirmation
				if ((_controllerContext != null)
					&& ((_controllerContext.getStatus() == LyContext.Status.ERROR)
						|| (_controllerContext.getStatus() == LyContext.Status.ROLLBACKED)
						|| (_controllerContext.getErrorLevel() == LyObject.LEVEL_ERROR)))
				{
					showErrors();
					return;
				}

				_importBehavior.importObjectsCommitted(this, updatedObjects.size());

				free();
				return;
			}
			else
				super.actionDemand(demand);
		}
		else if (demand.getType() == LyActionDemand.TYPE_STATUS)
		{
			String     demandId = demand.getValue();
			LyObjectId oid = getObjectId(demandId);
			LyObject   object = (LyObject)oid.getObject();

			if (object != null)
			{
				LySimpleObjectList objectList = new LySimpleObjectList(object);

				message = checkFields(objectList);
				if (message != null)
				{
					showError(message);
					return;
				}

				importObjects(objectList, false);
				return;
			}
		}
		else if ((demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
			&& (demand.getSource().intern() == LyImportController.SELECT_OPERATIONS))
		{
			LyFormField field = (LyFormField)getComponent().getComponent(LyImportController.SELECT_OPERATIONS);

			if (field == null)
			{
				LyImportBehavior behavior = (LyImportBehavior)getBehavior();

				_operations = behavior.getDefaultOperations();
				return;
			}

			LyChoice choice = (LyChoice)field.getFieldItem();
			LyOption option;
			int		 i, n = choice.getOptionCount();

			_operations.clear();

			// Retrieve the checked options
			for (i = 0; i < n; i++)
			{
				option = choice.getOption(i);

				if (option.isChecked())
					_operations.set(Integer.parseInt(option.getId()));
			}

			setIdle(true);
			loadData(false);
			reset(true);
			setIdle(false);
		}
		else
			super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Do the final import operation : this will commit the import context for the specified
	 * objects. The result is STILL in the cancellable context of this controller. Only subcontext
	 * are commited. This method is a shortcut for importObjects(objects, true) for backward
	 * compatibility
	 *
	 * @param objects the list of objects to import
	 * @see   #commit
	 * @see   #rollbackNonImportedObjects
	 * @see   #rollback
	 */
	public void importObjects(LyObjectList objects)
	{
		importObjects(objects, true);
	}

	/**
	 * Do the final import operation : this will commit the import context for the specified
	 * objects. The result is STILL in the cancellable context of this controller. Only subcontext
	 * are commited.
	 *
	 * @param objects      the list of objects to import
	 * @param importErrors indicates if we accept to import objects in error.
	 * @see   #commit
	 * @see   #rollbackNonImportedObjects
	 * @see   #rollback
	 */
	public void importObjects(LyObjectList objects, boolean importErrors)
	{
		boolean					   wasIdle = isIdle();
		LyController			   parent = getParent();
		Map<LyController, Boolean> wasParentIdle = new HashMap<LyController, Boolean>();

		while (parent != null)
		{
			if (parent instanceof LyPanelController)
			{
				if (!parent.isIdle())
				{
					wasParentIdle.put(parent, Boolean.valueOf(parent.isIdle()));
					parent.setIdle(true);
				}
			}

			parent = parent.getParent();
		}

		setIdle(true);
		getViewManager().setWaiting(true);

		if (OLD_IMPORT)
		{
			int		 i, n = objects.getSize();
			LyObject importObject, realObject;
			boolean  deleteObject;

			// FFT00116/ipboss
			_importBehavior.importObjectsStarted(this, n);

			if (_realImportedObjects == null)
				_realImportedObjects = new LySimpleObjectList();

			for (i = n - 1; i >= 0; i--)
			{
				importObject = objects.getObject(i);

				if (importObject.getClassInfo().getTemplate() != _classInfo)
					continue; // Skip

				realObject = getRealObject(importObject);

				// realObject may have been deleted AFTER opening import view
				if ((realObject != null) && (realObject.isDeleted()))
				{
					_realObjects.remove(importObject);
					realObject = null;
				}

				deleteObject = true;

				// If real object doesn't exist, create it
				if (isDeletedObject(importObject) && (realObject != null))
					deleteObject = deleteRealObject(importObject);
				else if (realObject == null)
					deleteObject = createRealObject(importObject);

				// If real object exists and needs to be set, set it
				else if (isSameObject(importObject) != null)
					deleteObject = setRealObject(realObject, importObject);

				// If real object exists and is the same as the imported object
				// just add it to the list of imported objects
				if (realObject != null)
					_realImportedObjects.addObject(realObject);

				if ((deleteObject) && (realObject != null))
				{
					if (realObject.getErrorLevel() == LyObject.LEVEL_ERROR)
					{
						deleteObject = false;

						StringBuffer sb = new StringBuffer();
						LyError		 error;
						int			 n1 = realObject.getErrorsCount();

						for (int i1 = n1 - 1; i1 >= 0; i1--)
						{
							error = realObject.getError(i1);
							if (error.getLevel() == LyObject.LEVEL_ERROR)
							{
								if (sb.length() > 0)
									sb.append('\n');

								sb.append(error.getMessage(getEnvironment()));
							}
						}

						if (sb.length() > 0)
							_importErrors.put(importObject, sb.toString());
						else
							_importErrors.remove(importObject);

						updateRow(importObject, -1, null);
					}
				}

				if (deleteObject)
					deleteImportObject(importObject);
				else if (realObject != null)

					// Restore uncommited values when failed
					realObject.removeNewValues(null);
			}

			// FFT00081/ipboss
			_importBehavior.importObjectsTerminated(this, n);
		}
		else
		{
			updateErrors(objects);

			// Actually importing an object means : going through the internal subcontext and validate them
			// Compare object before and after the import here
			// As we work with real objects, look in the context
			LyContext context = getControllerContext();

			if ((context != null) && (context.getInternalSons() != null))
			{

				// Do we have a son context for this import ?
				Iterator<LyContext>    iter = context.getInternalSons().iterator();
				LyContext   subcontext;
				LyOperation operation;
				ArrayList<LyOperation>   operations;

				while (iter.hasNext())
				{
					subcontext = iter.next();
					operations = subcontext.getOperations();

					for (int i = 0;
						((i < operations.size())
							&& ((subcontext.getStatus() == LyContext.Status.STARTED)
								|| (subcontext.getStatus() == LyContext.Status.RUNNING))); i++)
					{
						operation = operations.get(i);
						if (objects.contains(operation.getObject()))
						{

							// Be sure there is no pending errors
							if ((!importErrors) && (_importErrors.get(operation.getObject()) != null))
								continue; // Can not import this one

							subcontext.commit(); // It is still  undoable in parent context

							if (!_somethingWasImported)
							{
								_somethingWasImported = true;
								updateTools();
							}

							updateRow(operation.getObject(), -1, null);
						}
					}
				}
			}
		}

		setIdle(wasIdle);
		getViewManager().setWaiting(false);
		Iterator<LyController> parentIter = wasParentIdle.keySet().iterator();

		while (parentIter.hasNext())
		{
			parent = parentIter.next();
			parent.setIdle(wasParentIdle.get(parent));
		}
	} // end method importObjects

	/**
	 * This methods cancel the import for objects which where not validated by importObjects in the
	 * graphical view or by code.
	 *
	 * @see #importObjects
	 * @see #commit
	 * @see #rollback
	 */
	protected void rollbackNonImportedObjects()
	{
		if (OLD_IMPORT)
			return;

		// Actually importing an object means : going through the internal subcontext and validate them
		// Compare object before and after the import here
		// As we work with real objects, look in the context
		LyContext context = getControllerContext();

		if ((context != null) && (context.getInternalSons() != null))
		{

			// Do we have a son context for this import ?
			Iterator<LyContext>  iter = context.getInternalSons().iterator();
			LyContext subcontext;

			while (iter.hasNext())
			{
				subcontext = iter.next();

				boolean rollback = true;

				// Check if object was not changed outside :
				if (_modifiedObjects != null)
				{
					LyOperation operation;
					ArrayList<LyOperation>   operations = subcontext.getOperations();

					for (int i = 0; (i < operations.size()) && rollback; i++)
					{
						operation = operations.get(i);
						if (_modifiedObjects.contains(operation.getObject()))
							rollback = false;
					}
				}

				if (rollback)
					subcontext.rollback();
			}
		}
	} // end method rollbackNonImportedObjects

	/**
	 * Delete an imported object.
	 *
	 * @param object the temporary imported object to delete
	 */
	protected void deleteImportObject(LyObject object)
	{
		if (OLD_IMPORT)
		{

			// Passing the composition relation values to null
			removeCompositionRelationValues(object);

			// Remove the import object
			object.delete();
		}
	}

	/**
	 * Delete a real object from the specified temporary object.
	 *
	 * @param  importObject a temporary object
	 * @return true if the operation was completed succesfully
	 */
	protected boolean deleteRealObject(LyObject importObject)
	{
		if (OLD_IMPORT)
		{

			// Retrieve real object
			LyObjectId realObjectId = new LyObjectId(_classInfo, importObject.getObjectId());
			LyObject   realObject = realObjectId.getLogicalObject();
			LyContext  context = getControllerContext();

			getEnvironment().logTrace("## Import object is DELETED : retrieve real object : " + realObject);

			// Commit deletion
			LyResponse		  response = context.deleteObject(realObject);
			LyResponse.Status status = response.getStatus();

			if (status == LyResponse.Status.ERROR)
			{
				LyClassInfo   classInfo = getRealClassInfo();
				LyEnvironment environment = getEnvironment();
				String		  message = environment.getMessage("LY_ER_DELETE", classInfo.getName(),
						response.getError());

				_importErrors.put(importObject, message);

				updateRow(importObject, -1, null);

				showError(message);
			}

			return (status == LyResponse.Status.COMPLETED);
		}

		return false;
	} // end method deleteRealObject

	/**
	 * Create a new real object from the specified temporary object.
	 *
	 * @param  object a temporary object
	 * @return true if the operation was completed succesfully
	 */
	protected boolean createRealObject(LyObject object)
	{
		if (OLD_IMPORT)
		{

			// Create object with the given list of values
			LyObject realObject = createObjectFromImportedObject(object);

			getEnvironment().logTrace("## Import object is NEW : create real object : " + realObject);
			LyContext context = getControllerContext();

			// Commit creation
			LyResponse		  response = context.createObject(realObject);
			LyResponse.Status status = response.getStatus();

			if (status == LyResponse.Status.ERROR)
			{
				LyClassInfo   classInfo = getRealClassInfo();
				LyEnvironment environment = getEnvironment();
				String		  message = environment.getMessage("LY_ER_CREATE", classInfo.getName(),
						response.getError());

				_importErrors.put(object, message);

				updateRow(object, -1, null);

				showError(message);
			}

			if (_realImportedObjects == null)
				_realImportedObjects = new LySimpleObjectList();

			_realImportedObjects.addObject(realObject);

			if (_realObjects == null)
				_realObjects = new Hashtable<LyObject, LyObject>();

			_realObjects.put(object, realObject);

			return (status == LyResponse.Status.COMPLETED);
		}

		return false;
	} // end method createRealObject

	/**
	 * A getter to retrieve the LyImportBehavior Be careful: this is not the bahavior of the tablle
	 * (import table) but the behavior of the import action itself.
	 *
	 * @return the LyImportBehavior
	 */
	public LyImportBehavior getImportBehavior()
	{
		return _importBehavior;
	}

	/**
	 * A setter to retrieve the LyImportBehavior Be careful: this is not the bahavior of the tablle
	 * (import table) but the behavior of the import action itself.
	 *
	 * @param importBehavior the LyImportBehavior
	 */
	public void setImportBehavior(LyImportBehavior importBehavior)
	{
		_importBehavior = importBehavior;
	}

	/**
	 * Create a real object from an imported object
	 * @param object an imported object
	 * @return the real object
	 */

	public LyObject createObjectFromImportedObject(LyObject object)
	{
		if (OLD_IMPORT)
		{
			// Create object with the given list of values
			LyClassInfo			    classInfo = getRealClassInfo();
			LyObject			    realObject;
			LyValue				    value;
			LyValueSet			    values, realValues;
			LyInfoList<LyFieldInfo> fields = new LyInfoList<LyFieldInfo>(_importFields);
			LyInfoList<LyFieldInfo> idFields = _importClassInfo.getIdFieldInfos();

			fields.addList(idFields);

			realObject = classInfo.createInstance();
			realValues = new LyValueSet();
			values = object.getValues(fields);
			for (int i = 0; i < values.getSize(); i++)
			{
				value = values.getFieldValue(i);
				if (value.getField() instanceof LyStructFieldInfo)
					value = createRealStructValue(value, _fields, fields);
				else if (value.getField() instanceof LyArrayFieldInfo)
					value = createRealArrayValue(value, _fields, fields);

				// If the field is associated to a local imported field, get rid of its value
				if (_localImportedFields.containsValue(value.getField()))
				{
					LyFieldInfo realFieldInfo = checkRealFieldInfo((LyFieldInfo)value.getField());

					value = realFieldInfo.buildFieldValue(value.getValue());
				}

				realValues.addValue(value);

				// For spreadsheet mode, add extra values from the class behavior
				if (_importType == LyImportController.SELECT_IMPORT_TYPE_SPREADSHEET)
				{
					Class<?>				 classBehavior = classInfo.getBehaviorClass();
					LyClassBehaviorInterface behavior = null;

					if (classBehavior != null)
					{
						try
						{
							behavior = (LyClassBehaviorInterface)classBehavior.newInstance();
						}
						catch (Exception e)
						{
							_environment.writeException(e);
							return null;
						}

						behavior.init(this);
						behavior.addExtraValues(null, realValues);
					}
				}
			}

			realObject.skipEncoding(true);
			realObject.setValues(realValues);
			realObject.skipEncoding(false);

			//FFT00360
			// Use to complete values (for example not null field)
			completeValues(realObject);
			return realObject;
		}
		else
			return object;
	} // end method createObjectFromImportedObject

	/**
	 * Create a new real struct object from the specified temporary struct object.
	 *
	 * @param  importStructObject a temporary object
	 * @return true if the operation was completed succesfully
	 */
	protected boolean createRealStructObject(LyObject importStructObject)
	{
		if (OLD_IMPORT)
		{

			// Create object with the given list of values
			LyClassInfo importStructClassInfo = importStructObject.getClassInfo();

			//int                                   classIndex = _importStructClassInfos.getIndex(importStructClassInfo);
			//LyClassInfo                   structClassInfo = (LyClassInfo)_structClassInfos.getInfo(classIndex);
			LyClassInfo			    structClassInfo = (LyClassInfo)importStructClassInfo.getParent();
			LyInfoList<LyFieldInfo> fields = structClassInfo.getFieldInfos();
			LyInfoList<LyFieldInfo> importFields = importStructClassInfo.getFieldInfos();
			LyValue				    value;
			LyObject			    realStructObject;
			LyValueSet			    values, realValues;

			realValues = new LyValueSet();
			realStructObject = structClassInfo.createInstance();
			values = importStructObject.getValues(importFields);

			for (int i = 0; i < values.getSize(); i++)
			{
				value = values.getFieldValue(i);
				if (value.getField() instanceof LyStructFieldInfo)
					value = createRealStructValue(value, fields, importFields);

				realValues.addValue(value);
			}

			realStructObject.skipEncoding(true);
			realStructObject.setValues(realValues);
			realStructObject.skipEncoding(false);

			//FFT00360
			// Use to complete values (for example not null field)
			completeStructValues(realStructObject);

			// Commit creation
			LyResponse		  response = realStructObject.create();
			LyResponse.Status status = response.getStatus();

			if (status == LyResponse.Status.ERROR)
			{
				LyEnvironment environment = getEnvironment();
				String		  message = environment.getMessage("LY_ER_CREATE", structClassInfo.getName(),
						response.getError());

				showError(message);
			}

			return (status == LyResponse.Status.COMPLETED);
		}

		return false;
	} // end method createRealStructObject

	/**
	 * Sets a real object from the specified temporary object.
	 *
	 * @param  realObject   the real object which should be modified
	 * @param  importObject a temporary object
	 * @return true if the operation was completed succesfully
	 */
	protected boolean setRealObject(LyObject realObject, LyObject importObject)
	{
		if (OLD_IMPORT)
		{

			// Set object with the given list of values (keep only values
			// for selected fields).
			LyInfoList<LyFieldInfo> importFields = _importClassInfo.getFieldInfos();
			int					    i, n = ((importFields == null) ? 0 : importFields.getSize());
			boolean				    resetClassIndexes = false;

			// Remove all local fields. It is not necessary to load their values and for
			// some of them, it may result in reloading import file (for field info with
			// a daemon for example).
			for (i = n - 1; i >= 0; i--)
			{
				if (importFields.getInfo(i).hasMark(LyFieldInfo.MARK_LOCAL))
				{
					importFields.removeInfo(importFields.getInfo(i));
					resetClassIndexes = true;
				}
			}

			if (resetClassIndexes)
				_importClassInfo.resetFieldIndexes();

			LyValueSet  values = importObject.getValues(importFields);
			LyValueSet  realValues = new LyValueSet();
			LyValue     value;
			LyFieldInfo field;

			n = values.getSize();

			for (i = n - 1; i >= 0; i--)
			{

				// Retrieve the field in the list of values
				field = (LyFieldInfo)values.getFieldValue(i).getField();

				// If the field is associated to a local imported field, get rid of its value
				if (_localImportedFields.containsValue(field))
				{
					LyFieldInfo realFieldInfo = checkRealFieldInfo(field);

					value = realFieldInfo.buildFieldValue(values.getFieldValue(i).getValue());

					values.removeValue(field);
					values.removeValue(realFieldInfo);
					values.addValue(value);
				}
			}

			n = values.getSize();

			for (i = n - 1; i >= 0; i--)
			{

				// Retrieve the field in the list of values
				field = (LyFieldInfo)values.getFieldValue(i).getField();

				// Check if this field is selected or is associated to a selected field
				if ((_importFields.contains(field))
					|| ((_localImportedFields.get(field) != null)
						&& (_importFields.contains(_localImportedFields.get(field)))))
					continue;
				else
					values.removeValue(field);
			}

			for (i = 0; i < values.getSize(); i++)
			{
				value = values.getFieldValue(i);
				if (value.getField() instanceof LyStructFieldInfo)
					value = createRealStructValue(value, _fields, _importFields);
				else if (value.getField() instanceof LyArrayFieldInfo)
					value = createRealArrayValue(value, _fields, _importFields);

				realValues.addValue(value);
			}

			LyContext context = getControllerContext();

			realObject.skipEncoding(true);
			context.setValues(realObject, realValues);
			realObject.skipEncoding(false);

			// Use to complete values (for example not null field)
			completeValues(realObject);

			// Commit modification
			LyResponse		  response = context.setObject(realObject);
			LyResponse.Status status = response.getStatus();

			if (status == LyResponse.Status.ERROR)
			{
				LyClassInfo   classInfo = getRealClassInfo();
				LyEnvironment environment = getEnvironment();
				String		  message = environment.getMessage("LY_ER_SET", classInfo.getName(), response.getError());

				_importErrors.put(importObject, message);

				updateRow(importObject, -1, null);
				showError(message);
			}

			return (status == LyResponse.Status.COMPLETED);
		}

		return false;
	} // end method setRealObject

	/**
	 * Sets a real struct object from the specified temporary struct object.
	 *
	 * @param  realStructObject   the real struct object which should be modified
	 * @param  importStructObject a temporary struct object
	 * @return true if the operation was completed succesfully
	 */
	protected boolean setRealStructObject(LyObject realStructObject, LyObject importStructObject)
	{
		if (OLD_IMPORT)
		{

			// Set object with the given list of values (keep only values
			// for selected fields).
			LyValueSet			    values = importStructObject.getValues();
			LyValueSet			    realValues = new LyValueSet();
			int					    i, n = values.getSize();
			LyInfoList<LyFieldInfo> importFields = importStructObject.getClassInfo().getFieldInfos();
			LyInfoList<LyFieldInfo> fields = realStructObject.getClassInfo().getFieldInfos();
			LyValue				    value;

			for (i = 0; i < n; i++)
			{
				value = values.getFieldValue(i);
				if (value.getField() instanceof LyStructFieldInfo)
					value = createRealStructValue(value, fields, importFields);

				realValues.addValue(value);
			}

			realStructObject.skipEncoding(true);
			realStructObject.setValues(realValues);
			realStructObject.skipEncoding(false);

			// Use to complete values (for example not null field)
			completeValues(realStructObject);

			// Commit modification
			LyContext		  context = getControllerContext();
			LyResponse		  response = context.setObject(realStructObject);
			LyResponse.Status status = response.getStatus();

			if (status == LyResponse.Status.ERROR)
			{
				LyClassInfo   classInfo = getRealClassInfo();
				LyEnvironment environment = getEnvironment();
				String		  message = environment.getMessage("LY_ER_SET", classInfo.getName(), response.getError());

				showError(message);
			}

			return (status == LyResponse.Status.COMPLETED);
		}

		return false;
	} // end method setRealStructObject

	/**
	 * Return a real struct value of the temp struct value give in parameter. It also creates real
	 * the sub struct objects.
	 *
	 * @param  value        The temp struct value.
	 * @param  realFields   The real object fields.
	 * @param  importFields The temp object fields.
	 * @return The real struct value.
	 */
	protected LyValue createRealStructValue(LyValue value, LyInfoList<LyFieldInfo> realFields,
		LyInfoList<LyFieldInfo> importFields)
	{
		if (OLD_IMPORT)
		{
			String			  fieldId;
			LyStructFieldInfo structFieldInfo;
			LyClassInfo		  realStructClassInfo;
			LyObjectId		  realStructObjectId;
			LyDataSource	  dataSource;
			Vector<Object>	  importStructObjects;
			LyObject		  importStructObject, realStructObject;
			LyObjectId		  objectId = null;

			fieldId = value.getField().getId().substring(2);
			structFieldInfo = (LyStructFieldInfo)realFields.getInfo(fieldId);

			if (structFieldInfo == null) // a struct in array may not clone fieldInfo
				structFieldInfo = (LyStructFieldInfo)value.getField();

			if (value.getValue() == null)
				return new LyRelationValue(structFieldInfo, objectId);

			objectId = ((LyRelationValue)value).getObjectId(0);
			LyClassInfo clonedClassInfo = objectId.getClassInfo();

			realStructClassInfo = (LyClassInfo)clonedClassInfo.getParent(); //_structClassInfos.getInfo(_importStructClassInfos.getIndex(clonedClassInfo));

			realStructObjectId = new LyObjectId(realStructClassInfo, objectId.toStrings());

			value = new LyRelationValue(structFieldInfo, realStructObjectId);

			// getting the import struct objects of this import struct class created
			dataSource = clonedClassInfo.getDataSource();
			importStructObjects = dataSource.getCache().getObjects();

			for (int j = 0; j < importStructObjects.size(); j++)
			{
				importStructObject = (LyObject)importStructObjects.get(j);

				// retrieving the good import struct object
				if (((LyRelationValue)value).getObjectId(0).toString().equals(
						importStructObject.getObjectId().toString()))
				{
					realStructObject = _realStructObjects.get(importStructObject);
					if ((realStructObject == null) || (realStructObject.isDeleted()))
						createRealStructObject(importStructObject);
					else if (isSameObject(importStructObject) != null)
						setRealStructObject(realStructObject, importStructObject);
					else
						importStructObject.delete();

					break;
				}
			}

			return value;
		}

		return null;
	} // end method createRealStructValue

	/**
	 * Return a real array value of the temp struct value give in parameter.
	 *
	 * @param  value        The temp struct value.
	 * @param  realFields   The real object fields.
	 * @param  importFields The temp object fields.
	 * @return The real struct value.
	 */
	protected LyValue createRealArrayValue(LyValue value, LyInfoList<LyFieldInfo> realFields,
		LyInfoList<LyFieldInfo> importFields)
	{
		if (OLD_IMPORT)
		{
			LyArrayFieldInfo arrayFieldInfo = (LyArrayFieldInfo)value.getField();
			LyFieldInfo		 fieldInfo = arrayFieldInfo.getFieldInfo();

			if ((fieldInfo instanceof LyStructFieldInfo) || (fieldInfo instanceof LyArrayFieldInfo))
			{
				Object[] importValues = (Object[])value.getValue();

				if (importValues == null)
					importValues = new Object[0];

				Object[] realValues = new Object[importValues.length];
				LyValue  importValue, realValue;

				// for each sublevel : recurse
				for (int i = 0; i < importValues.length; i++)
				{
					importValue = fieldInfo.buildFieldValue(importValues[i]);
					if (importValue.getField() instanceof LyStructFieldInfo)
						realValue = createRealStructValue(importValue, _fields, _importFields);
					else if (importValue.getField() instanceof LyArrayFieldInfo)
						realValue = createRealArrayValue(importValue, _fields, _importFields);
					else

						// Should NEVER happen
						realValue = importValue;

					realValues[i] = realValue.getValue();
				}

				value = new LyArrayValue(arrayFieldInfo, realValues);
			}

			return value;
		}

		return null;
	} // end method createRealArrayValue

	/**
	 * Complete the object with values that are not in the import file Actually, this method looks
	 * for each field with the not_null flag and fill in the value with the default value.
	 *
	 * @param realObject the object to be set or created
	 */
	protected void completeValues(LyObject realObject)
	{
		if (OLD_IMPORT)
		{
			LyClassInfo classInfo = getRealClassInfo();

			// For each field with a not_null flags, check if it has got a value
			// If not use default value
			LyInfoList<LyFieldInfo> fieldsList = classInfo.getFieldInfos(LyFieldInfo.MARK_NOT_NULL);

			// for each field, get the current value, if null or start with '$', change value with
			// default value
			for (int i = 0; i < fieldsList.getSize(); i++)
			{
				LyFieldInfo currentfield = fieldsList.getInfo(i);
				Object	    currentValue = realObject.getValue(currentfield);

				if (currentValue == null)
				{
					Object defaultValue = getDefaultValue(currentfield);

					realObject.setValue(currentfield, defaultValue);
				}
			}
		}
	}

	/**
	 * Complete the object with values that are not in the import file Actually, this method looks
	 * for each field with the not_null flag and fill in the value with the default value.
	 *
	 * @param realObject the object to be set or created
	 */
	protected void completeStructValues(LyObject realObject)
	{
		if (OLD_IMPORT)
		{
			LyClassInfo classInfo = realObject.getClassInfo();

			// For each field with a not_null flags, check if it has got a value
			// If not use default value
			LyInfoList<LyFieldInfo> fieldsList = classInfo.getFieldInfos(LyFieldInfo.MARK_NOT_NULL);

			// for each field, get the current value, if null or start with '$', change value with
			// default value
			for (int i = 0; i < fieldsList.getSize(); i++)
			{
				LyFieldInfo currentfield = fieldsList.getInfo(i);
				Object	    currentValue = realObject.getValue(currentfield);

				if (currentValue == null)
				{
					Object defaultValue = getDefaultValue(currentfield);

					realObject.setValue(currentfield, defaultValue);
				}
			}
		}
	}

	/**
	 * Returns the default value for a given fieldInfo.
	 *
	 * @param  fieldInfo the fieldInfo.
	 * @return the default value for the given fieldInfo.
	 */
	protected Object getDefaultValue(LyFieldInfo fieldInfo)
	{
		Object defaultValue = fieldInfo.getDefaultValue();

		if ((defaultValue instanceof String) && (((String)defaultValue).charAt(0) == '$'))
		{
			LySession session = getSession();

			defaultValue = session.getValue(defaultValue);
			if ((defaultValue != null) && fieldInfo.isMultiple())
			{
				try
				{
					Object valTmp;

					valTmp = java.lang.reflect.Array.newInstance(defaultValue.getClass(), 1);
					((Object[])valTmp)[0] = defaultValue;
					defaultValue = valTmp;
				}
				catch (NegativeArraySizeException e)
				{
				}
			}
		}

		return defaultValue;
	} // end method getDefaultValue

	/**
	 * Cancel action This method frees the controller.
	 */
	protected void cancel()
	{
		setIdle(true);

		if (_workSpace != null)
		{
			synchronized (_workSpace)
			{
				Vector<LyObject> list = _workSpace.getObjects();
				int				 i, n = ((list == null) ? 0 : list.size());
				LyObject		 object;

				for (i = 0; i < n; i++)
				{
					object = list.elementAt(i);
					deleteImportObject(object);
				}
			}
		}

		// Rollback modifications
		rollback();
		setIdle(false);
		free();
	}

	/**
	 * Commit data modifications on imported objects This method closes current context and
	 * validates all objects in import It is possible to commit() partially the import with
	 * importObjects() method or to rollback() part of the import with rollbackNonImportedObjects.
	 * After this operation the context is not accessible.
	 *
	 * @return The status of the committed context.
	 * @see    #importObjects
	 * @see    #rollbackNonImportedObjects
	 * @see    #rollback
	 */
	public LyContext.Status commit()
	{
		boolean wasIdle = isIdle();

		setIdle(true);
		getViewManager().setWaiting(true);

		LyContext.Status result = LyContext.Status.COMPLETED;

		if (_controllerContext != null)
		{
			LyApplication    application = getApplication();
			LyEnvironment    environment = getEnvironment();
			LyController     grandParent = getParent().getParent();
			LyContext.Status status = _controllerContext.commit();

			if (status == LyContext.Status.ROLLBACKED)
			{
				String cause = _controllerContext.getErrorMessage();
				String message;

				if (cause == null)
					cause = application.getError();

				message = environment.getMessage("LY_ER_COMMIT_CONTEXT", cause);
				grandParent.showError(message);
			}
			else if (status != LyContext.Status.ERROR)
				_controllerContext = null;
		}

		getViewManager().setWaiting(false);
		setIdle(wasIdle);

		return result;
	} // end method commit

	/**
	 * Rollback all data modifications on imported objects, even those imported with
	 * importObjects().
	 *
	 * @see #importObjects
	 * @see #rollbackNonImportedObjects
	 * @see #commit
	 */
	public void rollback()
	{
		if (_controllerContext != null)
		{
			Collection<LyContext> sons1 = _controllerContext.getSons();
			Collection<LyContext> sonsInternal = _controllerContext.getInternalSons();
			Collection<LyContext> sons = new ArrayList<LyContext>();

			if (sons1 != null)
				sons.addAll(sons1);

			if (sonsInternal != null)
				sons.addAll(sonsInternal);

			if ((sons != null) && (!sons.isEmpty()))
			{
				LyContext subContext;
				Iterator<LyContext>  iter = sons.iterator();

				while (iter.hasNext())
				{
					subContext = iter.next();
					subContext.commit(); // Flaten the operation list (to sort them properly) then ROLLBACK ALL
				}
			}

			_controllerContext.rollback();
			_controllerContext = null;
		}
	} // end method rollback

	/**
	 * Return the real object corresponding to the imported object.
	 *
	 * @param      importObject the temporary object read from import file
	 * @return     the 'real' object instance if any, or null if it could not be found
	 * @deprecated getRealObject is not used any more with new import. Method returns importObject.
	 */
	public LyObject getRealObject(LyObject importObject)
	{
		if (OLD_IMPORT)
		{
			if (_realObjects == null)
			{
				getEnvironment().logTrace(" NULL car _realObjects est null");
				return null;
			}

			return _realObjects.get(importObject);
		}
		else
			return importObject;
	}

	/**
	 * Indicates if the real object corresponding to the imported object has exactly the same values
	 * or not.
	 *
	 * @param importObject the 'real' object instance for content comparizon
	 * @return null if all values are identical, the list of fields with different values or an
	 *         empty list.
	 */
	public LyInfoList<LyFieldInfo> isSameObject(LyObject importObject)
	{
		LyInfoList<LyFieldInfo> result = new LyInfoList<LyFieldInfo>();

		if (OLD_IMPORT)
		{
			if (_realObjects == null)
				return result;

			LyObject realObject = getRealObject(importObject);

			if (realObject == null)
				return result;

			return getDifferences(importObject, realObject);
		}
		else
		{

			// Compare object before and after the import here
			// As we work with real objects, look in the context
			if (((importObject.getNewValues() != null) && (importObject.getNewValues().getSize() > 0))
				|| ((importObject.getPendingValues() != null) && (importObject.getPendingValues().getSize() > 0)))
			{

				// Do we have a son context for this import ? Otherwise the subcontext is actually commited
				LyContext context = getControllerContext();

				if ((context != null) && (context.getInternalSons() != null))
				{
					Iterator<LyContext>    iter = context.getInternalSons().iterator();
					LyContext   subcontext;
					LyOperation operation;
					ArrayList<LyOperation>   operations;

					while (iter.hasNext())
					{
						subcontext = iter.next();
						operations = subcontext.getOperations();

						for (int i = 0;
							((i < operations.size())
								&& ((subcontext.getStatus() == LyContext.Status.STARTED)
									|| (subcontext.getStatus() == LyContext.Status.RUNNING))); i++)
						{
							operation = operations.get(i);
							if (importObject.equals(operation.getObject()))
							{
								LyValueSet values = importObject.getNewValues();

								// We can add new and pending values...
								if (values != null)
								{
									for (int j = 0; j < values.getSize(); j++)
									{
										result.addInfo((LyFieldInfo)values.get(j).getField());
									}
								}
								values = importObject.getPendingValues();

								// We can add new and pending values...
								if (values != null)
								{
									for (int j = 0; j < values.getSize(); j++)
									{
										result.addInfo((LyFieldInfo)values.get(j).getField());
									}
								}

								return result;
							}
						}
					}
				}

				return null;
			}
			else
				return null;
		}
	} // end method isSameObject

	/**
	 * Returns the differences between an imported object and a real object.
	 *
	 * @param  importObject the imported object.
	 * @param  realObject   the real object.
	 * @return null if all values are identical, the list of fields with different values or an
	 *         empty list.
	 */
	protected LyInfoList<LyFieldInfo> getDifferences(LyObject importObject, LyObject realObject)
	{
		LyInfoList<LyFieldInfo> result = new LyInfoList<LyFieldInfo>();
		int					    i, n = _fields.getSize();
		LyFieldInfo			    fieldInfo, importFieldInfo;
		Object				    importValue, realValue;
		boolean				    hasDifference = false;

		for (i = 0; i < n; i++)
		{
			fieldInfo = _fields.getInfo(i);

			// If we have a temporary fieldinfo, use it
			if (getLocalFieldInfo(fieldInfo) != null)
				fieldInfo = getLocalFieldInfo(fieldInfo);

			if (fieldInfo instanceof LyStructFieldInfo)
			{
				importFieldInfo = _importFields.getInfo("__" + fieldInfo.getId());
				importValue = importObject.getValue(importFieldInfo);
			}
			else
				importValue = importObject.getValue(fieldInfo);

			realValue = realObject.getValue(checkRealFieldInfo(fieldInfo)); // Go back if necessary to the real fieldinfo

			if (fieldInfo instanceof LyStructFieldInfo)
			{
				if (!((LyStructFieldInfo)fieldInfo).deepEquals(importValue, realValue, null))
				{
					hasDifference = true;
					result.addInfo(fieldInfo);
				}
			}
			else if (fieldInfo instanceof LyArrayFieldInfo)
			{
				if (!((LyArrayFieldInfo)fieldInfo).deepEquals(importValue, realValue, null))
				{
					hasDifference = true;
					result.addInfo(fieldInfo);
				}
			}
			else if (!fieldInfo.equals(importValue, realValue))
			{
				hasDifference = true;
				result.addInfo(fieldInfo);
			}
		}

		return (hasDifference ? result : null);
	} // end method getDifferences

	/**
	 * Indicates if the imported object may be deleted or not.
	 *
	 * @param importObject the imported object instance for the check
	 * @return true if the object may be deleted, false otherwise
	 */
	public boolean isDeletedObject(LyObject importObject)
	{
		if (_objectsToDelete == null)
			return false;

		return _objectsToDelete.contains(importObject);
	}

	/**
	 * Indicates if the imported object has been created or not.
	 *
	 * @param importObject the imported object instance for the check
	 * @return true if the object was created, false otherwise
	 */
	public boolean isNewObject(LyObject importObject)
	{
		if (OLD_IMPORT)
			return (getRealObject(importObject) == null);
		else
		{

			// Compare object before and after the import here
			// As we work with real objects, look in the context
			LyContext context = getControllerContext();

			if ((context != null) && (context.getInternalSons() != null))
			{

				// Do we have a son context for this import ?
				Iterator<LyContext>    iter = context.getInternalSons().iterator();
				LyContext   subcontext;
				LyOperation operation;
				ArrayList<LyOperation>   operations;

				while (iter.hasNext())
				{
					subcontext = iter.next();
					operations = subcontext.getOperations();

					for (int i = 0; i < operations.size(); i++)
					{
						operation = operations.get(i);
						if (operation.getObject() == importObject)
						{

							// Something happened !
							if (operation.getType() == LyOperation.Type.CREATE)
								return true;
						}
					}
				}
			}

			return false;
		}
	} // end method isNewObject

	/**
	 * @see leon.control.LySimpleTableController#sort(leon.info.LyFieldInfo, boolean)
	 */
	protected void sort(LyFieldInfo fieldInfo, boolean extend)
	{
		if (fieldInfo != null)
			super.sort(checkRealFieldInfo(fieldInfo), extend);
	}

	/**
	 * Converts a temporary locally imported fieldInfo to the real fieldInfo Should be used
	 * internally by this class (only).
	 *
	 * @param  fieldInfo the fieldInfo to convert
	 * @return the original fieldInfo or the parameter fieldInfo if not found
	 */
	public LyFieldInfo checkRealFieldInfo(LyFieldInfo fieldInfo)
	{
		if (OLD_IMPORT)
		{
			if (!_localImportedFields.containsValue(fieldInfo))
				return fieldInfo;

			// Maybe we have a local field imported (like reverse relation) in a dummy fieldInfo
			Enumeration<LyFieldInfo> keys = _localImportedFields.keys();

			LyFieldInfo realFieldInfo = null;
			Object	    key;

			while ((realFieldInfo == null) && keys.hasMoreElements())
			{
				key = keys.nextElement();

				if (_localImportedFields.get(key) == fieldInfo)
					realFieldInfo = (LyFieldInfo)key;
			}

			if (realFieldInfo == null)
				realFieldInfo = fieldInfo; // should never happen

			return realFieldInfo;
		}
		else
			return fieldInfo;
	} // end method checkRealFieldInfo

	/**
	 * Converts a real fieldInfo to the a temporary imported fieldInfo if it exists Should be used
	 * internally by this class (only).
	 *
	 * @param  fieldInfo the fieldInfo to convert
	 * @return the temporary fieldInfo or null if it could not be found
	 */
	public LyFieldInfo getLocalFieldInfo(LyFieldInfo fieldInfo)
	{
		if (OLD_IMPORT)
			return _localImportedFields.get(fieldInfo);
		else
			return fieldInfo;
	}

	/**
	 * Updates the status of tools and menu items.
	 */
	protected void updateTools()
	{
		if ((_withoutGui) || (_component == null))
			return;

		// Update ToolBar
		LyToolBar toolBar = (LyToolBar)_component.getComponent("_toolBar");

		if (toolBar == null)
			return;

		LySession			 session = getSession();
		LyInfoList<LyAction> actions = _action.getActions();
		LyAction			 action;
		LyTool				 tool = null;
		boolean				 enabled;
		int					 i, n = actions.getSize();
		LyComponent			 comp;

		// Update toolbar
		for (i = 0; i < n; i++)
		{
			action = actions.getInfo(i);
			enabled = session.enableAction(action, this, _selections, false);

			// action tool ?
			comp = toolBar.getComponent(action.getId());
			if (comp instanceof LyTool)
			{
				tool = (LyTool)comp;
				tool.setEnabled(enabled);
			}
		}

		// Update Commands
		LyCommands commands = (LyCommands)_component.getComponent("_commands");

		if (commands == null)
			return;

		comp = commands.getComponent(DEMAND_IMPORT_SELECTION);

		if (comp instanceof LyCommand)
		{
			LyCommand command = (LyCommand)comp;

			if ((_selections != null) && (_selections.getSize() != 0))
				command.setEnabled(true);
			else
				command.setEnabled(false);
		}

		// The validate buttons has no meaning when nothing was imported yet
		if (OLD_IMPORT)
		{
			// Nothing to DO
		}
		else
		{
			comp = commands.getComponent(DEMAND_VALIDATE);

			if (comp instanceof LyCommand)
			{
				LyCommand command = (LyCommand)comp;

				if (_somethingWasImported)
					command.setEnabled(true);
				else
					command.setEnabled(false);
			}
		}

		super.updateTools();
	} // end method updateTools

	/**
	 * Used to check the fields of the objects.
	 *
	 * @param  objects the ojbects to check.
	 * @return the error message or null if there's not error.
	 */
	public String checkFields(LyObjectList objects)
	{
		LyImportTableBehaviorInterface behavior = (LyImportTableBehaviorInterface)getBehavior();

		return behavior.checkFields(this, objects);
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyImportTableBehaviorInterface.class;
	}

	/**
	 * Retrieves the view Builder.
	 *
	 * @return the builder instance;
	 */
	protected LyBuilder getDefaultBuilder()
	{
		if (_withoutGui)
			return null;

		return new LyImportTableBuilder(this);
	}

	/**
	 * Return the classInfo used for the import.
	 *
	 * @return The temp classInfo.
	 */
	public LyClassInfo getRealClassInfo()
	{
		return _classInfo;
	}

	/**
	 * Sets the file to be used.
	 *
	 * @param file File path
	 */
	public void setFile(String file)
	{
		_file = file;
	}

	/**
	 * Set the buffer to be used to load objects.
	 *
	 * @param buffer the buffer containing the objects description
	 */
	public void setBuffer(String buffer)
	{
		_buffer = buffer;
	}

	/**
	 * Accessor to the list of fields for import.
	 *
	 * @return the list of fields to be displayed in the import view
	 */
	public LyInfoList<LyFieldInfo> getFields()
	{
		if (_importFields == null)
			return _classInfo.getFieldInfos();

		if (OLD_IMPORT)
			return _importFields;
		else
		{

			// This is the intersect oif classical table fields and imported fields
			LyInfoList<LyFieldInfo> result = new LyInfoList<LyFieldInfo>();
			LyInfoList<LyFieldInfo> tableFields = super.getFields();
			LyFieldInfo			    field;

			for (int i = 0; i < _importFields.getSize(); i++)
			{
				field = _importFields.getInfo(i);
				if (tableFields.contains(field))
					result.addInfo(field);
			}

			return result;
		}
	}

	/**
	 * Class info : content description. (application class)
	 *
	 * @return the classInfo being shown in this table
	 */
	public LyClassInfo getClassInfo()
	{
		if (OLD_IMPORT)
			return _importClassInfo;
		else
			return super.getClassInfo();
	}

	/**
	 * Accessor to the list of fields for import.
	 *
	 * @param fields the list of fields to be read from the import file
	 */
	public void setFields(LyInfoList<LyFieldInfo> fields)
	{
		if (_localImportedFields.size() > 0)
		{
			LyInfoList<LyFieldInfo> newFields = new LyInfoList<LyFieldInfo>();
			LyFieldInfo			    info;

			for (int i = 0; i < fields.getSize(); i++)
			{
				info = fields.getInfo(i);
				if (_localImportedFields.get(info) == null)
					newFields.addInfo(info);
				else
					newFields.addInfo(_localImportedFields.get(info));
			}

			fields = newFields;
		}

		_fields = fields;
	}

	/**
	 * Sets the field separator to be used.
	 *
	 * @param fieldSep the field separator
	 */
	public void setFieldSep(String fieldSep)
	{
		_fieldSep = fieldSep;
	}

	/**
	 * Sets the value separator to be used.
	 *
	 * @param valueSep the multiple value separator
	 */
	public void setValueSep(String valueSep)
	{
		_valueSep = valueSep;
	}

	/**
	 * Sets the import type.
	 *
	 * @param importType The import type.
	 */
	public void setImportType(String importType)
	{
		_importType = importType;
	}

	/**
	 * Sets the operations to be executed.
	 *
	 * @param operations a bitset containing operations to be executed. Operations may be one of the
	 *                   following:
	 *
	 *                   <ul>
	 *                     <li>{@link LyImportController#SELECT_OPERATIONS_CREATION}</li>
	 *                     <li>{@link LyImportController#SELECT_OPERATIONS_DELETION}</li>
	 *                     <li>{@link LyImportController#SELECT_OPERATIONS_MODIFICATION}</li>
	 *                   </ul>
	 */
	public void setOperations(BitSet operations)
	{
		_operations = operations;
	}

	/**
	 * Sets the parameter that indicates if operations field must be displayed.
	 *
	 * @param allowOperations true if operations field must be displayed, false otherwise.
	 */
	public void setAllowOperations(boolean allowOperations)
	{
		_allowOperations = allowOperations;
	}

	/**
	 * Returns a boolean that indicates if operations field must be displayed.
	 *
	 * @return true if operations field must be displayed, false otherwise.
	 */
	public boolean allowOperations()
	{
		return _allowOperations;
	}

	/**
	 * Sets the start line.
	 *
	 * @param startLine the index of the first line of the imported file containing real data.
	 */
	public void setStartLine(int startLine)
	{
		_startLine = startLine;
	}

	/**
	 * The list of temporary imported objects (not the real objects) To get the list of realObjects,
	 * importObjects(xxx) should be invoked with this list (or part of this list) of temporary
	 * objects.
	 *
	 * @deprecated replaced by getTemporaryImportedObjects()
	 * @return     the list of temporary objects
	 */
	public LyObjectList getImportedObjects()
	{
		if (OLD_IMPORT)
			return getTemporaryImportedObjects();
		else
			return getRealImportedObjects();
	}

	/**
	 * The list imported objects.
	 *
	 * @return     the list of imported objects
	 * @deprecated Used by OLD_IMPORT only,
	 */
	public LyObjectList getTemporaryImportedObjects()
	{
		if (OLD_IMPORT)
			return _temporaryImportedObjects;
		else
			return getRealImportedObjects();
	}

	/**
	 * The list of real imported objects since last loadData.
	 *
	 * @return the list of real objects
	 */
	public LyObjectList getRealImportedObjects()
	{
		if (_realImportedObjects == null)
			_realImportedObjects = new LySimpleObjectList();

		return _realImportedObjects;
	}

	/**
	 * Return the list of objects that may be deleted during the import process.
	 *
	 * @return the list of objects that may be deleted during the import process.
	 */
	public LyObjectList getObjectsToDelete()
	{
		return _objectsToDelete;
	}

	/**
	 * Current builder of a new row.
	 *
	 * @param  object       object added in the contextual list, and which should set the content of
	 *                      row cells.
	 * @param  fields       list of fields which should be seen.
	 * @param  statusFields list of fields which may set the status icon.
	 * @return LyRow the new Row description (XML description) which is added in the LyTable
	 *         component.
	 */
	public LyRow buildRow(LyObject object, LyInfoList<LyFieldInfo> fields, LyInfoList<LyFieldInfo> statusFields)
	{
		if ((_withoutGui) || (fields == null))
			return null;

		LyRow				    row = new LyRow();
		int					    i, n = fields.getSize();
		LyFieldInfo			    fieldInfo;
		LyCell				    cell;
		LyInfoList<LyFieldInfo> differentValues = isSameObject(object);
		LyObject			    realObject = getRealObject(object);

		row.setId(object.getId());

		// Adding an empty cell for status column
		cell = new LyCell();
		cell.setLabel("");
		row.addCell(cell);

		if (_importErrors.get(object) != null)
		{
			row.setTooltip(_importErrors.get(object));
			row.setImage("LY_IMPORT_ERROR");
		}
		else if (isDeletedObject(object))
		{
			row.setImage("LY_IMPORT_DELETE");
			row.setTooltip("LY_IMPORT_DELETE_OBJECT");
		}
		else if (isNewObject(object))
		{
			row.setImage("LY_IMPORT_NEW");
			row.setTooltip("LY_IMPORT_NEW_OBJECT");
		}
		else if (differentValues == null)
		{
			row.setImage("LY_IMPORT_SAME");
			row.setTooltip("LY_IMPORT_SAME_OBJECT");
		}
		else
		{
			row.setImage("LY_IMPORT_UPDATE");
			row.setTooltip("LY_IMPORT_UPDATE_OBJECT");
		}

		LyTableBehaviorInterface behavior = (LyTableBehaviorInterface)getBehavior();

		for (i = 0; i < n; i++)
		{
			fieldInfo = fields.getInfo(i);

			cell = new LyCell();
			behavior.updateCell(this, cell, object, fieldInfo);

			if ((differentValues != null) && differentValues.contains(fieldInfo))
			{
				cell.setColor("LY_IMPORT_TABLE_DIFFERENCE_COLOR");

				if (realObject != null)
					cell.setTooltip(getEnvironment().getMessage("LY_IMPORT_TABLE_ACTUAL_VALUE",
							fieldInfo.toName(realObject.getValue(fieldInfo))));
			}

			row.addCell(cell);
		}

		return row;
	} // end method buildRow

	/**
	 * Current builder of a new cell in a row. At this level, the implementation creates a LyCell
	 * instance, but generic behavior may be overriden.
	 *
	 * @param  object object added in the contextual list, and which should set the content of row
	 *                cells.
	 * @param  fieldInfo  field value shown in the cell content
	 * @return LyCell or LyFieldIntem the new Cell description (XML description) or editable content
	 *         to add in the LyRow component.
	 */
	protected LyComponent buildRowItem(LyObject object, LyFieldInfo fieldInfo)
	{
		if (_withoutGui)
			return null;

		return super.buildRowItem(object, fieldInfo);
	}

	/**
	 * Check if the status column exists.
	 *
	 * @param  statusFields (optional) list of status fields
	 * @return true
	 */
	public boolean isStatusColumn(LyInfoList<LyFieldInfo> statusFields)
	{
		return true;
	}

	/**
	 * @see leon.control.LyTableController#completeMessages(leon.view.LyMessages, int)
	 */
	public void completeMessages(LyMessages messages, int level)
	{
		if (_withoutGui)
			return;

		super.completeMessages(messages, level);
	}

	/**
	 * This method is invoked to enable the the area to complete the menuBar of the view and add its
	 * own menu items. Default implementation invokes the method on the builder instance.
	 *
	 * @param menuBar the menuBar structure
	 * @param level   : level of current Controller. one of the LEVEL_XXX contants in
	 *                LyViewController
	 * @see   LyViewController
	 */
	public void completeMenuBar(LyMenuBar menuBar, int level)
	{
		if (_withoutGui)
			return;

		super.completeMenuBar(menuBar, level);
	}

	/**
	 * Prepare component before showing. This method is called once (unlike reset which may be
	 * called several times), before first showing. Default implementation invokes reset.
	 */
	protected void initArea()
	{
		if (_withoutGui)
			return;

		super.initArea();
	}

	/**
	 * Accessor to the view controller of the area. The view controller is a Frame / dialog which
	 * contains this area. This method is reserved for internal use and should not be used by the
	 * application.
	 *
	 * @param viewController the controller which manages the parent frame / dialog
	 */
	public void setViewController(LyViewController viewController)
	{
		if (_withoutGui)
			return;

		super.setViewController(viewController);
	}

	/**
	 * Update the titlebar content with default values Default implementation let the builder set
	 * the titlebar content if this service is supported by the builder.
	 *
	 * @see leon.control.builder.LyAreaBuilder#updateTitleBar
	 */
	protected void updateTitleBar()
	{
		if (_withoutGui)
			return;

		super.updateTitleBar();
	}

	/**
	 * Update the title bar content (if any titlebar is associated with this area).
	 *
	 * @param  image        the image on the left side of the titlebar. If no image is given, try to
	 *                      get the image from the controller.
	 * @param  title        the string displayed on the left label of the titlebar. May be either a
	 *                      text message, or a String key found in the dictionnary
	 *                      (StringXXX.properties) files. if to title is given, try to get a title
	 *                      from the behavior (@see LyActionBehavior#getTitle).
	 * @param  contextTitle the string displayed on the right side of the titlebar. May be either a
	 *                      text message, or a String key found in the dictionnary
	 *                      (StringXXX.properties) files. If no string is given, try to construct
	 *                      one from the current context (if any).
	 * @return the titlebar view component
	 */
	public LyTitleBar updateTitleBar(String image, String title, String contextTitle)
	{
		if (_withoutGui)
			return null;

		return super.updateTitleBar(image, title, contextTitle);
	}

	/**
	 * Freeze or unFreeze a column.
	 *
	 * @param column column index to fix
	 */
	public void fixColumn(int column)
	{
		if (_withoutGui)
			return;

		super.fixColumn(column);
	}

	/**
	 * Accessor to panel component.
	 * @return the panel component
	 * @see LySimpleTableController#getTableComponent
	 */
	public LyPanel getPanelComponent()
	{
		if (_withoutGui)
			return null;

		return getTableComponent();
	}

	/**
	 * @see leon.control.LySimpleTableController#getRow(leon.data.LyObject, int)
	 */
	protected LyRow getRow(LyObject object, int index)
	{
		if (_withoutGui)
			return null;

		return super.getRow(object, index);
	}

	/**
	 * Accessor to the Table component.
	 * @return the table component
	 */
	public LyTable getTableComponent()
	{
		if (_withoutGui)
			return null;

		return super.getTableComponent();
	}

	/**
	 * Hide a row in a table (graphical component). Internal use of the Controller. This method only
	 * switch the hidden status in the view (graphical user interface impact).
	 *
	 * @param object hidden or shown object. May not be null
	 * @param hide   new hidden status
	 */
	protected void hideTableRow(LyObject object, boolean hide)
	{
		if (_withoutGui)
			return;

		super.hideTableRow(object, hide);
	}

	/**
	 * Internal method to add a new row for the specified object, in the specified line.
	 *
	 * @param object   current object being inserted in the table
	 * @param index    current index it the view, or -1 to add at the end
	 * @param selected indicates wether this line is selected or not
	 */
	protected void insertRow(LyObject object, int index, boolean selected)
	{
		if (_withoutGui)
			return;

		super.insertRow(object, index, selected);
	}

	/**
	 * Invoked when an menu item notification is sent to this listener.
	 *
	 * <P>This method overriden to create a popup menu : buildPopupMenu.</P>
	 *
	 * @param demand demand parameters sent by the view component.
	 */
	public void menuDemand(LyMenuDemand demand)
	{
		if (_withoutGui)
			return;

		super.menuDemand(demand);
	}

	/**
	 * Internal method to remove a row at the specified line.
	 *
	 * @param object the object to remove
	 * @param  index current index of the line to remove
	 * @return the row which was removed, or null if it could nott be found
	 */
	protected LyRow removeRow(LyObject object, int index)
	{
		if (_withoutGui)
			return null;

		return super.removeRow(object, index);
	}

	/**
	 * Invoked when a selection occurs. This method handles: select all, unselect all and objects by
	 * objects selection
	 *
	 * @param demand selection event created from the view. User interaction.
	 */
	public void selectionDemand(LySelectionDemand demand)
	{
		if (_withoutGui)
			return;

		super.selectionDemand(demand);
	}

	/**
	 * Select an object in the table.
	 *
	 * @param object       selected or unselected object
	 * @param select       new selection status
	 * @param updateView   indicates if the view (tools, messages...) must be updated.
	 * @param showSelected indicates wether the view should scroll to the new selection status
	 */
	public void selectObject(LyObject object, boolean select, boolean updateView, boolean showSelected)
	{
		if (_withoutGui)
			return;

		super.selectObject(object, select, updateView, showSelected);
	}

	/**
	 * @see leon.control.LyMultiPageTableController#selectRow(leon.info.LyFieldInfo, java.lang.String, int, boolean, boolean, boolean)
	 */
	public int selectRow(LyFieldInfo fieldInfo, String pattern, int from, boolean next, boolean substring,
		boolean caseSensitive)
	{
		if (_withoutGui)
			return -1;

		return super.selectRow(fieldInfo, pattern, from, next, substring, caseSensitive);
	}

	/**
	 * Selection of a row in a table (graphical component). Internal use of the Controller. This
	 * method only switch the selection status in the view (graphical user interface impact). The
	 * internal selection list (getSelection()) is not set by this method. Use selectObject to
	 * change the selection status of an object:
	 *
	 * @see   #selectObject
	 * @param object       selected or unselected object. May not be null
	 * @param select       new selection status
	 * @param showSelected indicates wether the view should scroll to the new selection status
	 */
	protected void selectTableRow(LyObject object, boolean select, boolean showSelected)
	{
		if (_withoutGui)
			return;

		super.selectTableRow(object, select, showSelected);
	}

	/**
	 * @see leon.control.LySimpleTableController#showObjectNotInViewError(java.lang.String)
	 */
	protected void showObjectNotInViewError(String id)
	{
		if (_withoutGui)
			return;

		super.showObjectNotInViewError(id);
	}

	/**
	 * Internal method to swap two rows at the specified lines.
	 *
	 * @param index    new index of the first line to move
	 * @param oldIndex current index of the line to move
	 */
	protected void swapRows(int index, int oldIndex)
	{
		if (_withoutGui)
			return;

		super.swapRows(index, oldIndex);
	}

	/**
	 * Reset the content of the table being shown. This method rebuild the XML structure of the
	 * table.
	 */
	protected void updateContent()
	{
		if (_withoutGui)
			return;

		super.updateContent();
	}

	/**
	 * Update the filter menuitem status.
	 */
	protected void updateFilterMenu()
	{
		if (_withoutGui)
			return;

		super.updateFilterMenu();
	}

	/**
	 * Update the footers of the table (content at the bottom of the columns).
	 */
	protected void updateFooters()
	{
		if (_withoutGui)
			return;

		super.updateFooters();
	}

	/**
	 * Update the headers of the table (content of the column titles).
	 */
	protected void updateHeaders()
	{
		if (_withoutGui)
			return;

		super.updateHeaders();
	}

	/**
	 * Update the menu for the current controller.
	 */
	protected void updateMenuBar()
	{
		if (_withoutGui)
			return;

		super.updateMenuBar();
	}

	/**
	 * This method is invoked to refresh the message bar of the view.
	 *
	 * @see LyAreaBuilder#completeMessages
	 */
	protected void updateMessages()
	{
		if (_withoutGui)
			return;

		super.updateMessages();
	}

	/**
	 * Update the preferences menuitem.
	 */
	public void updatePreferencesMenu()
	{
		if (_withoutGui)
			return;

		super.updatePreferencesMenu();
	}

	/**
	 * Update the sort menuitem status.
	 */
	protected void updateSortMenu()
	{
		if (_withoutGui)
			return;

		super.updateSortMenu();
	}

	/**
	 * Build a contextual menu with the given demand based on the given class info. Creates the
	 * actionList then invokes buildPopupMenu(LyDemand, Vector)
	 *
	 * @param demand the notification from the view component : request of a popup menu
	 * @param classInfo the classInfo associated to the popup menu
	 * @return a popup menu
	 * @see   leon.control.LySimpleTableController#buildPopupMenu(LyMenuDemand, LyClassInfo, boolean)
	 */
	protected LyPopupMenu buildPopupMenu(LyMenuDemand demand, LyClassInfo classInfo)
	{
		if (_withoutGui)
			return null;

		LySession			 session = getSession();
		LyInfoList<LyAction> actionsList = _action.getActions();
		LyAction			 action;
		int					 index = demand.getIndex(0);
		int					 i, n = actionsList.getSize();
		Vector<LyAction>     actions = new Vector<LyAction>(n);

		for (i = 0; i < n; i++)
		{
			action = actionsList.getInfo(i);

			// If clic on selection, the action must enable selected items
			if ((index >= 0) && (action.getMax() == 0))
				continue;

			// If clic on background, the action must enable 0 items
			if ((index < 0) && (action.getMin() > 0))
				continue;

			if (!session.validateAction(action, this))
				continue;

			actions.addElement(action);
		}

		LyPopupMenu popupMenu = new LyPopupMenu();

		completePopupMenu(popupMenu, actions);
		return popupMenu;
	} // end method buildPopupMenu

	/**
	 * Build a contextual menu with the given demand based on the given class info.
	 * @param demand a menu demand
	 * @param classInfo  a classInfp
	 */
	protected void buildPopupMenuItem(LyMenuDemand demand, LyClassInfo classInfo)
	{
		if (_withoutGui)
			return;

		super.buildPopupMenuItem(demand, classInfo);
	}

	/**
	 * Complete contextual menu with the given actions list.
	 * @param popupMenu the popup menu demand
	 * @param actions a list of actions
	 */
	protected void completePopupMenu(LyPopupMenu popupMenu, Vector<LyAction> actions)
	{
		if (_withoutGui)
			return;

		super.completePopupMenu(popupMenu, actions);
	}

	/**
	 * Called when an object has changed in the dynamic object list.
	 * @param event the workspace event
	 * @return true if event was handled
	 *
	 * @see LyWorkSpaceListener#objectChanged
	 */
	public boolean objectChanged(LyWorkSpaceEvent event)
	{
		if (!OLD_IMPORT)
		{
			LyObject object = event.getObject();

			if ((event.getType() == LyWorkSpaceEvent.Type.CHANGE) && (object != null) && (event.getValues() != null))
			{

				// The import context was modified outside (in the GUI) : this becomes a candidate for import
				if (_modifiedObjects == null)
					_modifiedObjects = new LySimpleObjectList(object);

				if (!_modifiedObjects.contains(object))
					_modifiedObjects.addObject(object);

				if (!_somethingWasImported)
				{
					_somethingWasImported = true;
					updateTools();
				}

			}
		}

		return super.objectChanged(event);
	}

	/**
	 * Update the row corresponding to the object.
	 *
	 * @param object       the object that was modified
	 * @param index        the position of the object
	 * @param statusFields the list of status fields
	 * @return the updated row
	 */
	protected LyRow updateRow(LyObject object, int index, LyInfoList<LyFieldInfo> statusFields)
	{
		if (_withoutGui)
			return null;

		LyRow row = getRow(object, index);

		if ((row == null) || (!(getBehavior() instanceof LyTableBehaviorInterface)))
			return null; // May occur when using MultiImportController and import without display.

		LyInfoList<LyFieldInfo>  fields = getFields();
		LyFieldInfo				 fieldInfo;
		LyTableBehaviorInterface behavior = (LyTableBehaviorInterface)getBehavior();
		LyComponent				 comp;
		int						 i, pos, n = fields.getSize();
		LyInfoList<LyFieldInfo>  differentValues = isSameObject(object);
		LyObject				 realObject = getRealObject(object);

		for (i = 0; i < n; i++)
		{
			pos = i + 1;

			fieldInfo = fields.getInfo(i);
			comp = row.getChild(pos);

			if (comp instanceof LyCell)
			{
				behavior.updateCell(this, (LyCell)comp, object, fieldInfo);

				if ((differentValues != null) && differentValues.contains(fieldInfo))
				{
					((LyCell)comp).setColor("LY_IMPORT_TABLE_DIFFERENCE_COLOR");

					if (realObject != null)
						((LyCell)comp).setTooltip(getEnvironment().getMessage("LY_IMPORT_TABLE_ACTUAL_VALUE",
								fieldInfo.toName(realObject.getValue(fieldInfo))));
				}
			}
		}

		if (_importErrors.get(object) != null)
		{
			row.setTooltip(_importErrors.get(object));
			row.setImage("LY_IMPORT_ERROR");
		}
		else if (isDeletedObject(object))
		{
			row.setImage("LY_IMPORT_DELETE");
			row.setTooltip("LY_IMPORT_DELETE_OBJECT");
		}
		else if (isNewObject(object))
		{
			row.setImage("LY_IMPORT_NEW");
			row.setTooltip("LY_IMPORT_NEW_OBJECT");
		}
		else if (differentValues == null)
		{
			row.setImage("LY_IMPORT_SAME");
			row.setTooltip("LY_IMPORT_SAME_OBJECT");
		}
		else
		{
			row.setImage("LY_IMPORT_UPDATE");
			row.setTooltip("LY_IMPORT_UPDATE_OBJECT");
		}

		String rowColor = behavior.getRowColor(this, object);
		String oldColor = row.getColor();

		if (((rowColor != null) && (oldColor == null)) || ((oldColor != null) && !oldColor.equals(rowColor)))
			row.setColor(rowColor);

		return row;
	} // end method updateRow

	/**
	 * Create a clone classInfo and update the relation fields.
	 *
	 * @param      classInfo The classInfo to clone.
	 * @return     The new classInfo.
	 * @deprecated is used by OLD_IMPORT only
	 */
	protected LyClassInfo createClassInfo(LyClassInfo classInfo)
	{
		if (OLD_IMPORT)
		{
			if (classInfo == null)
				return null;

			// Do not use javaName and behavior for the clonedClassInfo
			LyClassInfo			    clonedClassInfo = new LyClassInfo(classInfo, null, null, false, true, false, false,
					LyObjectInfo.PART_DATA | LyObjectInfo.PART_GUI);
			LyInfoList<LyFieldInfo> fieldInfos = classInfo.getFieldInfos();
			LyInfoList<LyFieldInfo> importFieldInfos = clonedClassInfo.getFieldInfos();
			LyFieldInfo			    fieldInfo;
			LyStructFieldInfo	    structFieldInfo;
			LyClassInfo			    structClassInfo, clonedStructClassInfo;

			LyImportMetaModel metaModel = new LyImportMetaModel(getApplication());

			// we need a specific behavior
			metaModel.setBehavior(clonedClassInfo, LyImportClassBehavior.class.getName());

			if (_structClassInfos == null)
			{
				_structClassInfos = new LyInfoList<LyClassInfo>();
				_importStructClassInfos = new LyInfoList<LyClassInfo>();
			}

			for (int i = 0; i < fieldInfos.getSize(); i++)
			{
				fieldInfo = fieldInfos.getInfo(i);
				if (fieldInfo instanceof LyStructFieldInfo)
				{
					structFieldInfo = (LyStructFieldInfo)fieldInfo;

					for (int j = 0; j < structFieldInfo.getClassInfoCount(); j++)
					{
						structClassInfo = structFieldInfo.getClassInfo(j);
						_structClassInfos.addInfo(structClassInfo);

						clonedStructClassInfo = ((LyStructFieldInfo)importFieldInfos.getInfo(i)).getClassInfo(j);
						_importStructClassInfos.addInfo(clonedStructClassInfo);

						completeCreation(clonedStructClassInfo, structClassInfo);
					}
				}

				// We also import (when available) fields defined by a reverse or a join relation even if they are local
				if (fieldInfo.isLocal() && (fieldInfo instanceof LyRelationFieldInfo)
					&& ((((LyRelationFieldInfo)fieldInfo).getDaemon() instanceof LyReverseDaemon)
						|| (((LyRelationFieldInfo)fieldInfo).getDaemon() instanceof LyJoinDaemon)))
				{
					LyRelationFieldInfo computedRelation = (LyRelationFieldInfo)fieldInfo;

					String key = fieldInfo.getId() + LyXmlExportRequestEncoder.LOCAL_FIELD_IMPORT_EXTENSION;

					LyRelationFieldInfo newRelation = (LyRelationFieldInfo)getApplication().getInfo(key);

					if (newRelation == null)
					{
						newRelation = metaModel.newRelationFieldInfo(key, null, null, computedRelation, true, false,
								computedRelation.getDirection(), computedRelation.getType(), null, true, true,
								LyRelationFieldInfo.ShowAs.DEFAULT);

						metaModel.updateClassInfosFromParent(newRelation, computedRelation);

						// clears local mark
						BitSet bs = new BitSet();

						bs.set(LyFieldInfo.MARK_TABLE);
						bs.set(LyFieldInfo.MARK_CONSULT);
						bs.set(LyFieldInfo.MARK_LOCAL);
						metaModel.setMarks(newRelation, bs);
					}

					_localImportedFields.put(computedRelation, newRelation);

					// add this "dummy" fieldinfo to import.
					metaModel.addFieldInfo(clonedClassInfo, newRelation);
				}
			}

			return clonedClassInfo;
		}
		else
			return classInfo;
	} // end method createClassInfo

	/**
	 * Create the temp struct classInfos and the corresponding struct classInfo.
	 *
	 * @param      classInfo       The temp struct classInfo.
	 * @param      importClassInfo The struct classInfo.
	 * @deprecated is used by OLD_IMPORT only
	 */
	protected void completeCreation(LyClassInfo classInfo, LyClassInfo importClassInfo)
	{
		LyInfoList<LyFieldInfo> fieldInfos = importClassInfo.getFieldInfos();
		LyFieldInfo			    fieldInfo;
		LyStructFieldInfo	    structFieldInfo;
		LyClassInfo			    structClassInfo, clonedStructClassInfo;

		for (int i = 0; i < fieldInfos.getSize(); i++)
		{
			fieldInfo = fieldInfos.getInfo(i);

			if (fieldInfo instanceof LyStructFieldInfo)
			{
				structFieldInfo = (LyStructFieldInfo)fieldInfo;
				for (int j = 0; j < structFieldInfo.getClassInfoCount(); j++)
				{
					structClassInfo = structFieldInfo.getClassInfo(j);
					_structClassInfos.addInfo(structClassInfo);
					clonedStructClassInfo = ((LyStructFieldInfo)classInfo.getFieldInfo(i)).getClassInfo(j);
					_importStructClassInfos.addInfo(clonedStructClassInfo);
					completeCreation(clonedStructClassInfo, structClassInfo);
				}
			}
		}
	}

	/**
	 * Removing the relation composition values before removing the import object.
	 *
	 * @param      object The import object.
	 * @deprecated is used by OLD_IMPORT only
	 */
	protected void removeCompositionRelationValues(LyObject object)
	{
		if (object == null)
			return;

		LyValueSet		    values = object.getValues(_importFields);
		LyValue			    value;
		LyRelationValue     relationValue;
		LyRelationFieldInfo relationField;

		for (int j = values.getSize() - 1; j >= 0; j--)
		{
			value = values.getFieldValue(j);
			if (value instanceof LyRelationValue)
			{
				relationValue = (LyRelationValue)value;
				relationField = (LyRelationFieldInfo)relationValue.getFieldInfo();

				if ((relationField instanceof LyStructFieldInfo) && (relationValue.getValue() != null))
					removeCompositionRelationValues(relationValue.getObject(0));
				else if (relationField.getType() == LyRelationFieldInfo.Type.COMPOSITION)
				{
					values.removeValue(relationField);
					object.setValue(relationField, null);
				}
			}
		}
	} // end method removeCompositionRelationValues

	/**
	 * Gets the controller context.
	 *
	 * @return The controller context.
	 */
	public LyContext getControllerContext()
	{
		if ((_controllerContext == null) || (_controllerContext.isEnded()))
		{
			LySession		 session = getSession();
			LyContextManager contextManager = session.getContextManager();

			_controllerContext = contextManager.newContext(session, LyContext.Type.USER_CANCELLABLE);
			if (_parent instanceof LyControllerContextInterface)
			{
				LyContext context = ((LyControllerContextInterface)_parent).getControllerContext();

				context.addSon(_controllerContext);
			}
		}

		return _controllerContext;
	} // end method getControllerContext

	/**
	 * Get current controller context.
	 *
	 * @param  controller Controller wich is requesting the context
	 * @return current context or null if none is specified
	 * @see    LyControllerContextInterface#getControllerContext
	 */
	public LyContext getControllerContext(LyController controller)
	{

		// If the context is modifying an imported object,
		// The parent context is the subcontext related to the object, not current context
		// FFT 1696
		LyContext context = getControllerContext();

		if ((context != null) && (context.getInternalSons() != null) && (controller instanceof LySetController))
		{
			LyObject currentObject = ((LySetController)controller).getObject();

			if (currentObject != null)
			{

				// Do we have a son context for this import ?
				Iterator<LyContext>  iter = context.getInternalSons().iterator();
				LyContext subcontext;

				while (iter.hasNext())
				{
					subcontext = iter.next();

					LyOperation operation;
					ArrayList<LyOperation>   operations = subcontext.getOperations();

					for (int i = 0; i < operations.size(); i++)
					{
						operation = operations.get(i);
						if (currentObject.equals(operation.getObject()))
							return subcontext;
					}
				}
			}
		}

		return context;
	} // end method getControllerContext

	/**
	 * Show a popup message error box with the specified message. The message may be a text message,
	 * or the key to a String in the application disctionnary. (stringsXXX.properties files) This is
	 * a shortcut to getViewManager().showError()
	 * @param message the message to show
	 * @see leon.view.LyViewManager
	 * @see leon.misc.LyStrings
	 */
	public void showError(String message)
	{
		if (_withoutGui)
			return;

		super.showError(message);
	}

	/**
	 * Show errors view.
	 */
	protected void showErrors()
	{
		if (_withoutGui)
			return;

		super.showErrors();
	}

	/**
	 * Make the view visible to the user. The effect of this command depends on the nature of the
	 * application. If the display is a fat client (windows (awt) / swing) this methods should open
	 * a window and show the view. In the case of thin client (html), this prepares the pages which
	 * will be returned and displayed in the client browser.
	 *
	 * <P>For any display, this method should consider that the page is ready to be shown and should
	 * be displayed.</P>
	 */
	public void showView()
	{
		if (_withoutGui)
			return;

		super.showView();
	}

	/**
	 * Show a popup message warning box with the specified message. The message may be a text
	 * message, or the key to a String in the application disctionnary. (stringsXXX.properties
	 * files) This is a shortcut to getViewManager().showWarning()
	 * @param message the message to show
	 * @see leon.view.LyViewManager
	 * @see leon.misc.LyStrings
	 */
	public void showWarning(String message)
	{
		if (_withoutGui)
			return;

		super.showWarning(message);
	}

	/**
	 * Update the view of the controller with the specified values from the demand (internal use) .
	 *
	 * @param demand demand parameters sent by the view component.
	 */
	public void updateViewFromValues(LyDemand demand)
	{
		if (_withoutGui)
			return;

		super.updateViewFromValues(demand);
	}

	/**
	 * Accessor to the Last error that was shown during import This may be used to check by code if
	 * import was Ok.
	 *
	 * @return the last message or null if no error occured
	 */
	public String getLastErrorMessage()
	{
		return _lastErrorMessage;
	}

	/**
	 * Last error that was shown during import This may be used to check by code if import was Ok.
	 * @param msg the message to store
	 */
	protected void setLastErrorMessage(String msg)
	{
		_lastErrorMessage = msg;
	}

	/**
	 * Called when grouped events are sent.
	 *
	 * @param  events Sorted list of events.
	 * @return True if the events are used by the listener, false otherwise.
	 */
	public boolean workSpaceEventsGrouped(ArrayList<LyWorkSpaceEvent> events)
	{
		int					  i, n = events.size();
		LyWorkSpaceEvent	  event;
		LyWorkSpaceEvent.Type type;
		boolean				  used = false;

		for (i = 0; i < n; i++)
		{
			event = events.get(i);
			type = event.getType();
			switch (type)
			{

			case ADD:
				used |= objectAdded(event);
				break;

			case REMOVE:
				used |= objectRemoved(event);
				break;

			case CHANGE:
				used |= objectChanged(event);
				break;

			case SELECTION_CHANGE:
			case VISIBILITY_CHANGE:
				used |= objectStateChanged((LyAdvancedWorkSpaceEvent)event);
				break;

			case USER_EVENT:
				used |= objectUserEvent((LyAdvancedWorkSpaceEvent)event);
				break;

			default:
				break;
			}
		}

		return used;
	} // end method workSpaceEventsGrouped

	/**
	 * THIS CLASS IS FOR OLD_IMPORT only (deprecated).
	 *
	 * @deprecated is used by OLD_IMPORT only
	 */
	protected class LyImportMetaModel extends LyMetaModel
	{
		/**
		 * Constructor
		 * @param application the application
		 */
		public LyImportMetaModel(LyApplication application)
		{
			super();
			_application = application;
		}

		/**
		 * @see leon.info.LyMetaModel#completeInfos()
		 */
		protected void completeInfos()
		{
		}

		/**
		 * @see leon.info.LyMetaModel#createInfos()
		 */
		protected void createInfos()
		{
		}

		/**
		 * @see leon.info.LyMetaModel#newRelationFieldInfo(java.lang.String, java.lang.String, leon.info.LyRelationFieldInfo, leon.info.LyRelationFieldInfo, boolean, boolean, leon.info.LyRelationFieldInfo.Direction, leon.info.LyRelationFieldInfo.Type, leon.info.LyRelationFieldInfo, java.lang.Object, boolean, boolean, leon.info.LyRelationFieldInfo.ShowAs)
		 */
		public LyRelationFieldInfo newRelationFieldInfo(String id, String javaName,
			LyRelationFieldInfo extendedRelation, LyRelationFieldInfo model, boolean multiple, boolean ordered,
			LyRelationFieldInfo.Direction direction, LyRelationFieldInfo.Type type,
			Object userData, boolean noContext, boolean noToolbar, LyRelationFieldInfo.ShowAs showAs)
		{
			return super.newRelationFieldInfo(id, javaName, extendedRelation, model, multiple, ordered, direction, type,
				userData, noContext, noToolbar, showAs);
		}

		/**
		 * @see leon.info.LyMetaModel#setDaemon(leon.info.LyRelationFieldInfo, leon.info.daemon.LyInfoDaemon)
		 */
		public void setDaemon(LyRelationFieldInfo relation, LyInfoDaemon daemon)
		{
			super.setDaemon(relation, daemon);
		}

		/**
		 * @see leon.info.LyMetaModel#setMarks(leon.info.LyObjectInfo, java.util.BitSet)
		 */
		public void setMarks(LyObjectInfo objectInfo, BitSet marks)
		{
			super.setMarks(objectInfo, marks);
		}

		/**
		 * @see leon.info.LyMetaModel#addFieldInfo(leon.info.LyClassInfo, leon.info.LyFieldInfo)
		 */
		protected void addFieldInfo(LyClassInfo classInfo, LyFieldInfo fieldInfo)
		{
			super.addFieldInfo(classInfo, fieldInfo);
		}

		/**
		 * Update a classInfo
		 * @param relation a relation
		 * @param extendedRelation the extended relation
		 */
		public void updateClassInfosFromParent(LyRelationFieldInfo relation, LyRelationFieldInfo extendedRelation)
		{
			int		    i, n = extendedRelation.getClassInfoCount();
			LyClassInfo classInfo;

			for (i = 0; i < n; i++)
			{
				classInfo = extendedRelation.getClassInfo(i);
				addClassInfo(relation, classInfo, extendedRelation.getFilter(classInfo));
			}

			setName(relation, extendedRelation.getName());
			setAlias(relation, extendedRelation.getAlias());
		}

		/**
		 * @see leon.info.LyMetaModel#setBehavior(leon.info.LyClassInfo, java.lang.String)
		 */
		public void setBehavior(LyClassInfo classInfo, String behavior)
		{
			super.setBehavior(classInfo, behavior);
		}
	} // end class LyImportMetaModel

} // end class LyImportTableController

/**
 * THIS CLASS IS FOR OLD_IMPORT only (deprecated).
 *
 * @deprecated is used by OLD_IMPORT only
 */
class LyImportClassBehavior extends LyClassBehavior
{
	public static final String __VERSION = "$Id: LyImportTableController.java 40621 2012-08-31 12:55:37Z nhamon $";

	public boolean isEditable(LySetController set, LyFieldInfo fieldInfo)
	{
		if (fieldInfo.hasMark(LyFieldInfo.MARK_READ_ONLY))
			return true; // This is allowed in temporary objects

		// TODO Auto-generated method stub
		return super.isEditable(set, fieldInfo);
	}

	protected LyImportTableController getImportTableController(LyController controller)
	{
		while (controller != null)
		{
			if (controller instanceof LyImportTableController)
				return ((LyImportTableController)controller);

			controller = controller.getParent();
		}

		return null;
	}

	/**
	 * Utility method to know if a fieldInfo is displayed in this form or not.
	 *
	 * @param  form       the current LyFormController instance.
	 * @param  fieldInfo  the fieldInfo which is tested for addition in the view.
	 * @return true if the field should be added, false if it should be skipped.
	 */
	public boolean isDisplayed(LyFormController form, LyFieldInfo fieldInfo)
	{
		LyImportTableController parent = getImportTableController(form);

		// do not display the "real" fieldInfo
		if (parent.getLocalFieldInfo(fieldInfo) != null)
			return false;

		return super.isDisplayed(form, fieldInfo);
	}
} // end class LyImportClassBehavior
