/**
 * LyLoginController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyLoginController.java 36545 2012-08-22 15:24:34Z smaupin $
 */

package leon.control;

import java.lang.reflect.*;
import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.setinterface.*;


/**
 * The login controller is an user input controller which describes a login view This controller
 * handles:
 *
 * <UL>
 *   <LI>a basic login view with a user login, a user password field and language selection.</LI>
 *   <LI>a maximum number of retries</LI>
 *   <LI>If login was successfull, it calls the next action</LI>
 * </UL>
 *
 * <p>The login controller is associated to a behavior which implements LyLoginBehaviorInterface.
 * Usually, this interface is implemented by a LyLoginBehavior class.</p>
 *
 * @see LyUserInputController
 * @see leon.app.behaviorinterface.LyLoginBehaviorInterface
 * @see leon.app.behavior.LyLoginBehavior
 */
public class LyLoginController extends LyUserInputController implements LyQuestionAnswerInterface
{
	/** File Version */
	public static final String __VERSION =
		"$Id: LyLoginController.java 36545 2012-08-22 15:24:34Z smaupin $";

	/** Define the status of the login */
	public enum Status
	{
		/**
		 *  The login is successful.
		 */
		LOGIN_OK,
		/**
		 *  The login invalid, but further attempts exists.
		 */
		LOGIN_ERROR,
		/**
		 *  The login has failed (too many unsuccessful attempt).
		 */
		LOGIN_FAILED,
		/**
		 *  The login was cancelled.
		 */
		LOGIN_CANCELLED,
		/**
		 *  The login could not execute : no login classInfo or fieldInfos.
		 */
		LOGIN_MISSING,
		/**
		 *  The login view has not been shown yet.
		 */
		LOGIN_INIT,
		/**
		 *  The login is still in progress. View should be visible.
		 */
		LOGIN_IN_PROGRESS
	}

	/** Default id for user class. */
	public static String	   USER_CLASS_ID = "User";

	/** Default width of string input fields. */
	public static int		   DEFAULT_COLS = 30;

	/** parameter key for login value (received in parameter values). */
	public static String	   PARAM_KEY_LOGIN = "_login";

	/** parameter key for password value (received in parameter values). */
	public static String	   PARAM_KEY_PASSWORD = "_password";

	/** parameter key for language value (received in parameter values). */
	public static String	   PARAM_KEY_LANGUAGE = "_language";

	/** parameter key for target value (must match LyWebConstants.TARGET_PARAM). */
	public static String	   PARAM_KEY_TARGET = "target_session";

	/** parameter key for portlet login value */
	public static String      PARAM_KEY_PORTLET_LOGIN = "_portlet_user_login";


	/** User classInfo (default is USER_CLASS_ID). */
	protected LyClassInfo	   _userClassInfo;

	/** Login fieldInfo (attribute) (default is first ID fieldInfo). */
	protected LyFieldInfo	   _loginFieldInfo;

	/** Password fieldInfo (attribute) (default is first SECRET fieldInfo). */
	protected LyFieldInfo	   _passwordFieldInfo;

	/** Number of max retry for this login. */
	protected int			   _maxLoginRetry;

	/** Login status. */
	protected Status		   _loginStatus;

	/** Retry counts. */
	protected int			   _userRetries;

	/** Allow reconnection **/
	protected boolean		   _allowReconnect;

	/** Display **/
	protected boolean          _alwaysDisplay;

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller).
	 * @param action  related action (_login).
	 * @param classInfo related classInfo
	 * @param context current action context. This context is used for the next actions.
	 */
	public LyLoginController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent, action, classInfo, context);

		_userClassInfo = (LyClassInfo)getParameter("_userClassInfo");
		_loginFieldInfo = (LyFieldInfo)getParameter("_loginFieldInfo");
		_passwordFieldInfo = (LyFieldInfo)getParameter("_passwordFieldInfo");

		String maxRetryParameter = (String)getParameter("_maxLoginRetry");

		_maxLoginRetry = 3;
		if (maxRetryParameter != null)
		{
			_maxLoginRetry = LyTools.intFromString(maxRetryParameter);
			if ((_maxLoginRetry != -1) &&(_maxLoginRetry <= 0))
				_maxLoginRetry = 3;
		}

		String reconnect = (String)getParameter("_allowReconnect");

		_allowReconnect = (reconnect == null ? false : Boolean.parseBoolean(reconnect));
		_loginStatus = Status.LOGIN_INIT;
		_userRetries = 0;

		String display = (String)getParameter("_display");

		_alwaysDisplay = ((display != null) && display.equalsIgnoreCase("ALWAYS"));
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyLoginBehaviorInterface.class;
	}

	/**
	 * Shows the Window which contains the view.
	 */
	public void showView()
	{
		// Initialize the controller
		if (getStatus() == Status.LOGIN_INIT)
			init();

		// Initialization may fail or could be already completed
		if (getStatus() == Status.LOGIN_INIT)
			super.showView();
	}

	/**
	 * Indicates login status : to know if login was successful.
	 *
	 * @return one of the Status.xxx constants
	 */
	public Status getStatus()
	{
		return _loginStatus;
	}

	/**
	 * Indicates how many times the user has tried to log in.
	 *
	 * @return a number between 0 and _maxLoginRetry
	 */
	public int getUserRetries()
	{
		return _userRetries;
	}

	/**
	 * Check the the login is available (number of user ...) This method is called once, before
	 * first showing.
	 */
	protected void init()
	{
		LyLoginBehaviorInterface behavior = (LyLoginBehaviorInterface)getBehavior();
		LySession				 session = getSession();
		LyObject                 user = getUser();

		if ((user != null)
			&& !_alwaysDisplay)
		{
			loginWithUser(user);
			return;
		}

		LyApplication			 application = session.getApplication();
		LyEnvironment			 environment = getEnvironment();

		// Use default
		if (_action == null)
			_action = application.getAction(LyAction.ID_LOGIN);

		if (behavior != null)
			behavior.initializeLogin(this);

		// Ensure we have the classInfo and fieldInfos
		if (_userClassInfo == null)
			// Default is USER_CLASS_ID
			_userClassInfo = application.getClassInfo(LyLoginController.USER_CLASS_ID);

		if (_userClassInfo != null)
		{
			LyInfoList<LyFieldInfo> fields;

			if (_loginFieldInfo == null)
			{
				// Default is firdst ID field
				fields = _userClassInfo.getIdFieldInfos();
				_loginFieldInfo = fields.getInfo(0);
			}

			if (_passwordFieldInfo == null)
			{
				// Default is first SECRET field
				fields = _userClassInfo.getFieldInfos(LyFieldInfo.MARK_SECRET);

				if ((fields != null) && (fields.getSize() > 0))
					_passwordFieldInfo = fields.getInfo(0);
				else
					environment.logWarning(
						"LoginController : fieldInfo password not found in class "
						+ _userClassInfo);
			}
		}
		else
			environment.logWarning("LoginController : classInfo not found.");

		// Check the number of users in the user class
		int nUsers = 0;

		if (_userClassInfo != null)
		{
			LyObjectSource objectSource = getSession().getObjectSource(_userClassInfo);

			nUsers = objectSource.getSize();
		}

		// Check that everything is correct
        if (((nUsers == 0) && (getCreateAction() == null))
				|| (_userClassInfo == null)
				|| (_loginFieldInfo == null)
				|| (_passwordFieldInfo == null))
		{
			_loginStatus = Status.LOGIN_MISSING;

			// Notify
			if (behavior != null)
				behavior.endLogin(this, null);
		}
	} // end method init

	/**
	 * Prepare component before showing. This method is called once, before first showing.
	 */
	protected void completeComponent()
	{
		if ((getStatus() == Status.LOGIN_FAILED)
				|| (getStatus() == Status.LOGIN_MISSING))
			return;

		_loginStatus = Status.LOGIN_IN_PROGRESS;

		// Client may do extra work here ...
		LyLoginBehaviorInterface behavior = (LyLoginBehaviorInterface)getBehavior();

		if (behavior != null)
			behavior.completeInitializeLogin(this);
	} // end method completeComponent

	/**
	 * Accessor to the view component instance. (XML description of the view) This instance is
	 * created by the builder, and multiple calls to this method always return the same instance.
	 *
	 * @return the LyComponent instance of the view package, memory representation (parsed) of the
	 *         XML description of the view.
	 */
	public LyComponent getComponent()
	{
		// If view is built by a file builder, fields have their controllers set
		// (see method build() in LyFileBuilder). It may result in strange behaviors
		// with current controller (login controller).
		LyBuilder builder = getBuilder();

		if ((_component == null) && (builder instanceof LyFileBuilder))
		{
			// If login and password may have their controllers set, set them to null
			LyComponent comp = super.getComponent();
			LyComponent login = comp.getComponent(PARAM_KEY_LOGIN);
			LyComponent password = comp.getComponent(PARAM_KEY_PASSWORD);

			if (login instanceof LySetControllerInterface)
				((LySetControllerInterface)login).setController(null);

			if (password instanceof LySetControllerInterface)
				((LySetControllerInterface)password).setController(null);
		}

		return super.getComponent();
	}

	/**
	 * This method checks if the login entered by the user is valid. It increments the retries
	 * counter and use the login method to check the value
	 *
	 * @param  values The values entered by the user
	 * @return true if login is successfull
	 */
	public boolean checkLogin(LyParameterValues values)
	{
		// Increment tries count
		_userRetries++;

		StringBuffer error = new StringBuffer();
		LyObject	 user = getUser();
		boolean		 login = false;

		if (user == null)
			login = login(values, error);
		else
		{
			// Validate auto-login
			LyLoginBehaviorInterface behavior = (LyLoginBehaviorInterface)getBehavior();
			boolean					 autoLogin = !isInitialized();

			login = ((behavior != null)
				&& behavior.validateUser(this, user, values, error, autoLogin));
		}

		boolean		 genericResult = true; // Close window ?

		if (login)
		{
			_loginStatus = Status.LOGIN_OK;

			// Load user preferences
			_session.loadUserPreferences();
		}
		else // No result or more than one result
		{
			if (_component != null)
			{
				LyFormField passwordFormField = (LyFormField)(_component.getComponent(PARAM_KEY_PASSWORD));

				if (passwordFormField != null)
				{
					LyFieldItem passwordField = passwordFormField.getFieldItem();
					passwordField.setLabel(null); // reset password field
				}

				if (_maxLoginRetry == -1)
				{
					// User can always reconnect
					if (error.length() > 0)
						showError(error.toString());
					else
						showError("LY_ER_LOGIN");
					_loginStatus = Status.LOGIN_ERROR;
					genericResult = false;
				}
				else if (_userRetries >= _maxLoginRetry)
				{
					if (error.length() > 0)
						showError(error.toString());
					else
						showError("LY_ER_MAX_LOGIN");
					_loginStatus = Status.LOGIN_FAILED;
				}
				else
				{
					if (_userRetries == (_maxLoginRetry - 1))
					{
						if (error.length() > 0)
							showError(error.toString());
						else
							showError("LY_ER_LAST_LOGIN");
					}
					else
					{
						if (error.length() > 0)
							showError(error.toString());
						else
							showError("LY_ER_LOGIN");
					}

					_loginStatus = Status.LOGIN_ERROR;
					genericResult = false;
				}
			}
			else
			{
			    _loginStatus = Status.LOGIN_ERROR;
				return false;
			}
		}

		LyLoginBehaviorInterface behavior = (LyLoginBehaviorInterface)getBehavior();

		// if needed try to reconnect
		if (_allowReconnect)
		{
			// Before notifying behavior check if we can attach to an existing session
			LySession existingSession = getSessionWithCurrentUser(values);
			boolean askQuestion = behavior.confirmReconnect(this);

			if (existingSession != null)
			{
				if (askQuestion)
				{
					Object[] questionValues = new Object[2];
					questionValues[0] = existingSession;
					// In web mode, the values will be reseted by the web server.
					if (values instanceof LyAdvancedParameterValues)
						((LyAdvancedParameterValues)values).keepValues();
					questionValues[1] = values;
					getViewManager().askQuestion(_component, "LY_CONFIRM_RECONNECT", this, questionValues);
				}
				else
					swapSession(existingSession);
				return true;
			}
		}

		// Notify behavior.
		if (behavior != null)
			behavior.endLogin(this, values);

		if (login)
		{
			// Load user stickers
			_session.loadUserStickers();
		}

		return genericResult;
	} // end method checkLogin


	/**
	 * Get an existing session with the user defined in the given view values
	 * @param values the view values
	 * @return a session if found
	 */
	protected LySession getSessionWithCurrentUser(LyParameterValues values)
	{
		LyViewManager viewManager = (LyViewManager)getEnvironment().getViewManager();
		if (viewManager != null && !viewManager.isMultiSession())
			return null;
		// Notify behavior.
		LyLoginBehaviorInterface behavior = (LyLoginBehaviorInterface)getBehavior();
		LyObject user = getSession().getUser();

		if (behavior != null)
			return behavior.getExistingSessionForUser(this, user, values);
		return null;
	}
	/**
	 * Check if the login is correct. This method retrieves the login and password value and calls
	 * the login behavior with the values.
	 *
	 * @param  values the value entered by the user
	 * @param  error  the error message to show in case of invalid login.
	 * @return true if the login is successfull
	 */
	protected boolean login(LyParameterValues values, StringBuffer error)
	{
		String login = values.getParameter(PARAM_KEY_LOGIN);
		String password = values.getParameter(PARAM_KEY_PASSWORD);
		String language = values.getParameter(PARAM_KEY_LANGUAGE);
		Object parsedLogin = login;
		Object parsedPassword = password;

		if (language != null)
		{
			LySession     session = getSession();
			session.setLanguage(language);
		}

		try
		{
			parsedLogin = _loginFieldInfo.parse(login);
			parsedPassword = _passwordFieldInfo.parse(password);

			if ((!(parsedLogin instanceof String))
					&& (parsedLogin != null))
				parsedLogin = parsedLogin.toString();

			if ((!(parsedPassword instanceof String))
					&& (parsedPassword != null))
				parsedPassword = parsedPassword.toString();
		}
		catch (Exception e)
		{
			getEnvironment().writeException(e);
		}

		LyLoginBehaviorInterface behavior = (LyLoginBehaviorInterface)getBehavior();

		if ((behavior != null)
				&& behavior.validate(this, (String)parsedLogin, (String)parsedPassword, values, error))
			return true;
		else
			return false;

	} // end method login

	/**
	 * Proceed to next actions after a login (login should be valid at that point).
	 */
	public void invokeNextActions()
	{
		LySession session = getSession(), newSession;
		LyAction  action;
		int		  i, n = _action.getActionCount();

		// Check if multiple actions must be invoked in web mode
		LyViewManager	viewManager = getViewManager();
		boolean			multiWindowMode = false;
		boolean			mainWindowOpened = false;
		LyController	controller;

		if (viewManager == null)
		{
			return;
		}

		multiWindowMode = viewManager.hasProperty("MULTI_WINDOW");

		for (i = 0; i < n; i++)
		{
			action = _action.getAction(i);
			controller = null;

			if (mainWindowOpened && !multiWindowMode)
			{
				// In web mode, open new windows if there are more than one sub-action
				if (session.enableAction(action, getViewController().getParent(), getContext(), false)
						&& session.validateAction(action, getViewController().getParent()))
				{
					String	slaveSessionClassName = viewManager.getSlaveSessionClassName();

					if (slaveSessionClassName != null)
					{
						Class<?>[]  types = new Class[2];
						Object[]    parameters = new Object[2];

						types[0] = LyApplication.class;
						parameters[0] = session.getApplication();

						types[1] = boolean.class;
						parameters[1] = Boolean.FALSE;

						try
						{
							Class<?>	   viewClass = Class.forName(slaveSessionClassName);
							Constructor<?> constructor = viewClass.getConstructor(types);

							newSession = (LySession)constructor.newInstance(parameters);
						}
						catch (InvocationTargetException e1)
						{
							_environment.writeException(e1.getTargetException());
							_environment.writeException(e1);
							throw new IllegalArgumentException("BF AppEngine : can't build slave session class : "
								+ slaveSessionClassName);
						}
						catch (Exception e)
						{
							_environment.writeException(e);
							throw new IllegalArgumentException("BF AppEngine : can't build slave session class : "
								+ slaveSessionClassName);
						}
					}
					else
						newSession = new LySession(session.getApplication(), false);

					newSession.setValid(session.isValid());
					newSession.addValue("$USER", session.getValue("$USER"));
					newSession.addValue("$ROOT_ACTION", action);
					newSession.register(session.isRegistered());
					newSession.setMasterSession(session);

					LyGeometry					geometry = action.getGeometry();
					Hashtable<Short, String>	params = new Hashtable<Short, String>();

					if (geometry != null)
					{
						params.put(new Short(LyViewManager.SHOW_LOCATION_PARAMETER_X), geometry.getX() + "");
						params.put(new Short(LyViewManager.SHOW_LOCATION_PARAMETER_Y), geometry.getY() + "");
						params.put(new Short(LyViewManager.SHOW_LOCATION_PARAMETER_WIDTH), geometry.getWidth() + "");
						params.put(new Short(LyViewManager.SHOW_LOCATION_PARAMETER_HEIGHT), geometry.getHeight() + "");
					}
					else
						params.put(new Short(LyViewManager.SHOW_LOCATION_PARAMETER_FULL_SCREEN), Boolean.TRUE.toString());

					if (viewManager != null)
						viewManager.showLocation(null, PARAM_KEY_TARGET + '=' + newSession.getId(), true, params);
				}
			}
			else
			{
				LyController    parent = (_viewController == null ? getParent() : _viewController.getParent());

				controller = session.invokeAction(parent, action, null, getContext(), true);
			}

			if (!multiWindowMode && (controller != null))
				mainWindowOpened = true;
		}
	}

	/**
	 * This method handles closing of the login controller. It is called when the user presses the
	 * cancel button.
	 */
	public void close()
	{
		LyLoginBehaviorInterface behavior = (LyLoginBehaviorInterface)getBehavior();

		// Cancel button was pressed
		_loginStatus = Status.LOGIN_CANCELLED;

		if (behavior != null)
			behavior.endLogin(this, null);
	}

	/**
	 * Accessor to User classInfo.
	 *
	 * @return User classInfo
	 */
	public LyClassInfo getUserClassInfo()
	{
		return (_userClassInfo);
	}

	/**
	 * Accessor to the login field in userClassInfo.
	 *
	 * @return login fieldInfo
	 */
	public LyFieldInfo getLoginFieldInfo()
	{
		return (_loginFieldInfo);
	}

	/**
	 * Accessor to password fieldInfo.
	 *
	 * @return password fieldInfo
	 */
	public LyFieldInfo getPasswordFieldInfo()
	{
		return (_passwordFieldInfo);
	}

	/**
	 * Accessor to User classInfo.
	 *
	 * @param userClassInfo User classInfo
	 */
	public void setUserClassInfo(LyClassInfo userClassInfo)
	{
		_userClassInfo = userClassInfo;
	}

	/**
	 * Accessor to Login fieldInfo (attribute).
	 *
	 * @param loginFieldInfo Login fieldInfo
	 */
	public void setLoginFieldInfo(LyFieldInfo loginFieldInfo)
	{
		_loginFieldInfo = loginFieldInfo;
	}

	/**
	 * Accessor to password fieldInfo.
	 *
	 * @param passwordFieldInfo password fieldInfo
	 */
	public void setPasswordFieldInfo(LyFieldInfo passwordFieldInfo)
	{
		_passwordFieldInfo = passwordFieldInfo;
	}

	/**
	 * Invoked when an action occurs.
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			String   sourceId = demand.getSource();
			LyAction action = _userClassInfo.getActionWithId(sourceId);

			if (action != null)
			{
				// Invoke the action on classInfo
				invokeAction(action, _userClassInfo);
				return;
			}
		}

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Swap the current session with a currently existing one
	 * @param existingSession the existing session
	 */
	protected void swapSession(LySession existingSession)
	{

		LyViewManager vm = getViewManager();
		vm.swapSession(getSession(), existingSession, getApplication().getId());
		LyController topController = existingSession.getController(0);
		if (topController != null)
			topController.setLastModified(System.currentTimeMillis());
		//topController.showView();
	}

	/**
	 * Called when the user answer about reconnecting to an existing session
	 *
	 * @param result the result (true or false)
	 * @param clientData an array of object: 0 an existing session, 1 view values
	 */
	public void questionAnswer(boolean result, Object clientData)
	{
		if (!(clientData instanceof Object[])
				&& ( ((Object[])clientData).length != 2 ))
			return;

		Object[] values = (Object[])clientData;

		if (result)
		{
			// reconnect to session
			LySession existingSession = (LySession)values[0];

			swapSession(existingSession);
		}
		else
		{
			LyParameterValues params = (LyParameterValues)values[1];
			// Notify behavior.
			LyLoginBehaviorInterface behavior = (LyLoginBehaviorInterface)getBehavior();

			if (behavior != null)
				behavior.endLogin(this, params);
		}
	}

	/**
	 * Gets user connected on current session (through login or with SSO).
	 *
	 * @return  The user connected by the login controller, null if not connected or invalid login.
	 */
	public LyObject getUser()
	{
		LySession   session = getSession();

		return session.getUser();
	}

	/**
     * Complete login with given user.
     * @param user User to log in.
     */
    public void loginWithUser(LyObject user)
    {
        if (user == null)
            return;

        LyLoginBehaviorInterface behavior = (LyLoginBehaviorInterface)getBehavior();

        _loginStatus = Status.LOGIN_OK;

        if (behavior != null)
            behavior.initializeLogin(this);

        // Notify
        if (behavior != null)
            behavior.endLogin(this, null);
    }

    /**
     * Get creation action if it is enabled in login form.
     * @return The action to be used for create a new user, null if it doesn't exist or if not enabled.
     */
    public LyAction getCreateAction()
    {
        if (!testParameter("_allowCreation"))
            return null;

        LyClassInfo userClassInfo = getUserClassInfo();

        if (userClassInfo == null)
            return null;

        LyAction    create = userClassInfo.getActionWithId("_create");

        if ((create != null)
                && _session.validateAction(create, this))
            return create;
        else
            return null;
    }
} // end class LyLoginController
