/**
  * LyMailController.java
  * Copyright (c) 2000-2009, Lyria-W4.
  * All rights reserved.
  */
  
package leon.control;

import java.io.*;
import java.util.*;
import leon.app.*;
import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;


/**
 * This class implements a sending email input dialog/window to ask the requested values from the
 * user. This controller is associated with a builder ({@link LyMailBuilder}) and a behavior ({@link
 * LyMailBehavior}) that implements {@link LyMailBehaviorInterface}.
 *
 * @author  Lyria S.A.
 * @version $Id: LyMailController.java 30940 2011-05-11 15:09:35Z nhamon $
 */
public class LyMailController extends LyUserInputController
{
	/** File Version */
	public static final String __VERSION =
		"$Id: LyMailController.java 30940 2011-05-11 15:09:35Z nhamon $";

	/** Form field identifier: smtp server to send the email message. */
	public static final String SMTP_SERVER = "_smtpServer".intern();

	/** Form field identifier: server port to use for the connection. */
	public static final String SMTP_PORT = "_smtpPort".intern();

	/** Form field identifier: smtp server requires authentication. */
	public static final String AUTHENTICATION = "_authentication".intern();

	/** Form field identifier: smtp user name for authentication. */
	public static final String AUTH_LOGIN = "_authLogin".intern();

	/** Form field identifier: smtp password for authentication. */
	public static final String AUTH_PASSWORD = "_authPassword".intern();

	/** Form field identifier: smtp realm for CRAM-MD5 authentication. */
	public static final String AUTH_REALM = "_authRealm".intern();

	/**
	 * Form field identifier: sets the controller to send the selected objects or use the selected
	 * objects as the receivers.
	 */
	public static final String SEND_SELECTED = "_sendSelected".intern();

	/** Form field identifier: email subject field. */
	public static final String MAIL_SUBJECT = "_mailSubject".intern();

	/** Form field identifier: user name of the sender. */
	public static final String FROM_USER_NAME = "_fromUserName".intern();

	/** Form field identifier: email address of the sender. */
	public static final String FROM_MAIL_ADDRESS = "_fromMailAddress".intern();

	/** Form field identifier: the class that holds the receivers' users as objects. */
	public static final String RECIPIENT_USER_CLASS = "_recipientClassInfo".intern();

	/** Form field identifier: the field of the class that contains the email address. */
	public static final String RECIPIENT_USER_FIELD = "_recipientFieldInfo".intern();

	/** Form field identifier: the relation field that holds the recipient users. */
	public static final String SELECTED_RECIPIENTS = "_selectedRecipients".intern();

	/** Form field identifier: a file that can be sent with the email message. */
	public static final String ATTACHED_FILE = "_attachedFile".intern();

	/** Form field identifier: a file that contains the HTML content of the email message. */
	public static final String HTML_MESSAGE = "_htmlMessage".intern();

	/** Form field identifier: the unformatted email message. */
	public static final String RAW_MESSAGE = "_rawMessage".intern();

	/** Form field identifier: the generated name of the sender, with its email address. */
	public static final String MAIL_PREVIEW_SENDER = "_mailPreviewSender".intern();

	/** Form field identifier: the generated subject of a consultation email message. */
	public static final String MAIL_PREVIEW_SUBJECT = "_mailPreviewSubject".intern();

	/** Form field identifier: the generated content of a consultation email message. */
	public static final String MAIL_PREVIEW_CONTENT = "_mailPreviewContent".intern();

	/** Silent mode. */
	protected boolean		   _silent;

	/** The SMTP server to use for sending the email message. */
	protected String		   _smtpServer;

	/** The port used to connect the SMTP server. */
	protected int			   _smtpPort;

	/** Uses authentication for the SMTP server. */
	protected boolean		   _authentication;

	/** User name for the SMTP server authentication. */
	protected String		   _authLogin;

	/** Password for the SMTP server authentication. */
	protected String		   _authPassword;

	/** Realm for CRAM-MD5 authentication on the SMTP server. */
	protected String		   _authRealm;

	/**
	 * Sends the selected objects by email.<br>
	 * If this attribute is true, the selected objects will be used for the email message content:
	 * the email subject is the name of the selected object if there is only one, or the number of
	 * objects and their classes if there's multiple.<br>
	 * If this attribute is false, then the selected objects will be used to obtain the email
	 * addresses of the receivers, and a custom message will be send to them.
	 */
	protected boolean		   _sendSelected;

	/** Subject of the custom message. */
	protected String		   _subject;

	/** Sender user name. */
	protected String		   _fromName;

	/** Sender email address. */
	protected String		   _fromAddress;

	/** Unformatted message of the email. */
	protected String		   _rawMessage;

	/** Reiceivers objects class. */
	protected String		   _toUserClass;

	/** Reiceivers field that contains the email addresses. */
	protected String		   _toUserField;

	/** File that contains the formatted content of the email message. */
	protected File			   _htmlMessage;

	/** Attached file. */
	protected File			   _attachedFile;

	/** List of selected objects at action start. */
	protected LyWorkSpace	   _selectedObjects = getContext();

	/**
	 * This constructor creates a new LyMailController object.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 * @param action  related action. If an action is specified, the controller uses action
	 *                parameters (action name, action specific XML view description ...)
	 * @param context current context (may be null if the view is not contextual to a list of
	 *                objects).
	 */
	public LyMailController(LySession session, LyController parent, LyAction action,
		LyObjectList context)
	{
		this(session, parent, action, null, context);
	}

	/**
	 * This constructor creates a new LyMailController object.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo This parameter is not used. Kept from introspection compatibility.
	 * @param context   current context (may be null if the view is not contextual to a list of
	 *                  objects).
	 */
	public LyMailController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent, action, classInfo, context);
		_silent = (parent == null);

		if ((_selectedObjects == null) && (classInfo != null))
			_selectedObjects = new LyClassWorkSpace(session, classInfo, false);
	}

	/**
	 * @see leon.control.LyUserInputController#getBehaviorInterface()
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyMailBehaviorInterface.class;
	}

	/**
	 * 	 * @see leon.control.LyUserInputController#getDefaultBuilder()
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return (new LyMailBuilder(this));
	}

	/**
	 * Sets the visibility of the specified field. The visibility concept is based on using the
	 * "hidden" mark on the given field.
	 *
	 * @param view    the common view description of this controller
	 * @param id      the identifier of a field that needs its "hidden" mark changed to the given
	 *                value
	 * @param visible the state of the visibility for the given field
	 */
	protected void setVisibility(LyView view, String id, boolean visible)
	{
		LyFormField field = (LyFormField)view.getComponent(id);

		if (field != null)
			field.setHidden(!visible);
	}

	/**
	 * Gets the value of the specified field for the given view. This method starts by asking this
	 * controller the original value of the parameter in the meta-model. Then the form field
	 * associated to the given identifier is requested through the view. If the field is null, or if
	 * its value doesn't correspond to a type available in the _mail definition, then the original
	 * value is returned, or null if nothing can be casted to the String object.
	 *
	 * @param  id   the identifier of the field to retreive
	 * @param  view the current view that holds the specified field
	 * @return a String describing the value of the specified id in the given view, or null if it is
	 *         not available
	 */
	protected String getValue(String id, LyView view)
	{
		Object	    originalValue = getParameter(id);

		LyFormField field = (LyFormField)view.getComponent(id);

		if (field == null)
			if (originalValue instanceof String)
				return (String)originalValue;
			else if (originalValue instanceof LyClassInfo)
				return ((LyClassInfo)originalValue).getId();
			else if (originalValue instanceof LyFieldInfo)
				return ((LyFieldInfo)originalValue).getId();
			else
				return null;
		return field.getFieldItem().getValue();
	}

	/**
	 * Retrieves all the values described for this controller.
	 */
	protected void retrieveValues()
	{
		LyView view = getView();

		_smtpServer = getValue(SMTP_SERVER, view);

		String port = getValue(SMTP_PORT, view);

		if (port == null)
			_smtpPort = 25;
		else
			_smtpPort = Integer.valueOf(getValue(SMTP_PORT, view)).intValue();

		String sendSelectedValue = getValue(SEND_SELECTED, view);

		_sendSelected = "true".equals(sendSelectedValue);

		String authValue = getValue(AUTHENTICATION, view);

		_authentication = "true".equals(authValue);
		_authLogin = getValue(AUTH_LOGIN, view);
		_authPassword = getValue(AUTH_PASSWORD, view);
		_authRealm = getValue(AUTH_REALM, view);

		_fromAddress = getValue(FROM_MAIL_ADDRESS, view);
		_fromName = getValue(FROM_USER_NAME, view);
		_subject = getValue(MAIL_SUBJECT, view);
		_rawMessage = getValue(RAW_MESSAGE, view);
		String htmlMessageLocation = getValue(HTML_MESSAGE, view);

		if (htmlMessageLocation != null)
			_htmlMessage = new File(htmlMessageLocation);
		if (_sendSelected && (_rawMessage == null) && (_htmlMessage == null))
		{
			LyFormField field = (LyFormField)view.getComponent(MAIL_PREVIEW_CONTENT);

			_rawMessage = field.getFieldItem().getValue();
		}

		String attachedFileLocation = getValue(ATTACHED_FILE, view);

		if (attachedFileLocation != null)
			_attachedFile = new File(attachedFileLocation);
		_toUserClass = getValue(RECIPIENT_USER_CLASS, view);
		_toUserField = getValue(RECIPIENT_USER_FIELD, view);
	} // end method retrieveValues

	/**
	 * Checks every aspect of all the values already retrieved with {@link #retrieveValues()} and
	 * generates the appropriate error if some values are missing.
	 */
	public void validateMail()
	{
		retrieveValues();

		String error = null;

		if ((_rawMessage == null) && (_htmlMessage == null))
			if (_rawMessage == null)
				error = "LY_MAIL_RAW_MESSAGE";
			else
				error = "LY_MAIL_HTML_MESSAGE";

		if (_toUserField == null)
			error = "LY_MAIL_USER_FIELD";

		if (error != null)
		{
			LyEnvironment environment = getEnvironment();
			String		  message = environment.getMessage("LY_ER_NOTNULL", error);

			if (!_silent)
				showError(message);
			else
				environment.logError(message);
			return;
		}
		sendMail();
	} // end method validateMail

	/**
	 * Sends a message associated to the given objects. If the {@link #_sendSelected} attribute is
	 * true, then the given objects will be sent through the {@link
	 * LyMailBehavior#sendMail(LyMailController, LyClassInfo, LyFieldInfo, LyObjectList)} method.
	 * Else the objects will specify email users to send a custom email message with the {@link
	 * LyMailBehavior#sendMail(LyMailController, LyObjectList, LyFieldInfo)} method.
	 */
	public void sendMail()
	{
		LyMailBehaviorInterface behavior = (LyMailBehaviorInterface)getBehavior();

		if ((getSmtpServer() == null) && (getEnvironment().getResource(LyMailBehavior.MAIL_JNDI) == null))
			_environment.logError("The SMTP server or a MAIL_JNDI ressource mail must be defined. Mail not sent.");
		else
		{
			if (_sendSelected)
				behavior.sendMail(this, getClassInfo(_toUserClass),
					getClassInfo(_toUserClass).getFieldInfo(_toUserField), _selectedObjects);
			else
				behavior.sendMail(this, _selectedObjects,
					getClassInfo().getFieldInfo(_toUserField));
		}
		LyViewManager manager = getViewManager();

		if (manager != null)
			hideView();
	}

	/**
	 * Generates the subject of an email message when the selected objects are sent to multiple
	 * recipients. In this case, the subject will be based on the number of selected items as well
	 * as the associated classes, or simply its name if the object is unique.
	 *
	 * @return a String resuming the selected objects content
	 */
	public String getGeneratedSubject()
	{
		String				    subject = "";
		LyObject			    object;
		LyClassInfo			    objectClassInfo;
		
		if (_selectedObjects != null)
		{
			LyInfoList<LyClassInfo> classes = _selectedObjects.getClasses();
	
			object = _selectedObjects.getObject(0);
			objectClassInfo = object.getClassInfo();
			Object[] args =
			{
				new Integer(_selectedObjects.getSize()),
				object.getName(),
				new Integer(classes.getSize()),
				_environment.translate(objectClassInfo.getName())
			};
	
			subject = _environment.getString("LY_MAIL_SUBJECT_MULTIPLE_OBJECTS", args);
		}
		
		return subject;
	}

	/**
	 * Generates the textual content of a generated email message. This message will be a list of
	 * all the object's fields that appear in consultation.
	 *
	 * @return a String containing the consultation text of the objects.
	 */
	public String getGeneratedRawContent()
	{
		StringBuffer		    content = new StringBuffer();
		Object				    valueObject;
		LyObject			    object;
		LyValue				    value;
		LyValueSet			    values;
		LyClassInfo			    objectClassInfo;
		LyFieldInfo			    objectFieldInfo;
		
		if (_selectedObjects != null)
		{
			LyInfoList<LyClassInfo> classes = _selectedObjects.getClasses();
	
			for (int i = 0; i < _selectedObjects.getSize(); i++)
			{
				object = _selectedObjects.getObject(i);
				objectClassInfo = object.getClassInfo();
				values = object.getValues();
	
				if (i > 0)
					content.append("\n----------\n");
	
				if (classes.getSize() == 1)
					content.append(_environment.getString("LY_MAIL_RAW_OBJECT", object.getName()));
				else
				{
					String[] args =
					{
						object.getName(),
						_environment.translate(object.getClassInfo().getName())
					};
	
					content.append(_environment.getString("LY_MAIL_RAW_OBJECT_CLASS", args));
				}
				content.append(":\n");
	
				for (int j = 0; j < values.getSize(); j++)
				{
					value = values.getFieldValue(j);
					objectFieldInfo = objectClassInfo.getFieldInfo(value.getField().getId());
					valueObject = value.getValue();
					if ((valueObject != null) && isConsultable(objectFieldInfo))
					{
						content.append(_environment.translate(objectFieldInfo.getName()));
						content.append(": ");
						content.append(_environment.translate(objectFieldInfo.toName(valueObject)));
						content.append("\n");
					}
				}
			}
		}
		
		return content.toString();
	} // end method getGeneratedRawContent

	/**
	 * Generates a formatted message content that represents the selected objects.
	 *
	 * @return a String that has HTML content for the representation of the objects to be sent
	 */
	public String getGeneratedHtmlContent()
	{
		StringBuffer		    content = new StringBuffer();
		Object				    valueObject;
		LyObject			    object;
		LyClassInfo			    objectClassInfo;
		LyInfoList<LyClassInfo> classes = _selectedObjects.getClasses();

		content.append("<TABLE border=\"1\" style=\"border-collapse:collapse\">\n<THEAD><TR><TH>");
		content.append(_environment.translate("LY_HEADER_HIDDEN_OBJECT_LABEL"));
		content.append("</TH>");
		if (classes.getSize() > 1)
			content.append("<TH>").append(_environment.translate("LY_HEADER_CLASSINFO_LABEL"))
				.append("</TH>");
		content.append("<TH>");
		content.append(_environment.translate("LY_MAIL_FIELD_NAME"));
		content.append("</TH><TH>");
		content.append(_environment.translate("LY_MAIL_FIELD_VALUE"));
		content.append("</TH></TR></THEAD>\n<TBODY>\n");

		for (int i = 0; i < _selectedObjects.getSize(); i++)
		{
			object = _selectedObjects.getObject(i);
			objectClassInfo = object.getClassInfo();
			LyValue     value;
			LyFieldInfo objectFieldInfo;
			LyValueSet  values = object.getValues();
			int		    span = 0;

			for (int j = 0; j < values.getSize(); j++)
			{
				value = values.getFieldValue(j);
				objectFieldInfo = objectClassInfo.getFieldInfo(value.getField().getId());
				valueObject = value.getValue();
				if ((valueObject != null) && isConsultable(objectFieldInfo))
					span++;
			}

			content.append("<TR BGCOLOR=\"#");
			if ((i % 2) == 0)
				content.append("F9F9F9");
			else
				content.append("F3F3F3");
			content.append("\"><TD");
			if (span > 1)
				content.append(" ROWSPAN=\"").append(span).append("\"");
			content.append(">");
			content.append(object.getName());
			content.append("</TD>");

			if (classes.getSize() > 1)
			{
				content.append("<TD");
				if (span > 1)
					content.append(" ROWSPAN=\"").append(span).append("\"");
				content.append(">");
				content.append(_environment.translate(object.getClassInfo().getName()));
				content.append("</TD>");
			}

			for (int j = 0, k = 0; j < values.getSize(); j++)
			{
				value = values.getFieldValue(j);
				objectFieldInfo = objectClassInfo.getFieldInfo(value.getField().getId());
				valueObject = value.getValue();
				if ((valueObject != null) && isConsultable(objectFieldInfo))
				{
					if (k > 0)
					{
						content.append("<TR BGCOLOR=\"#");
						if ((i % 2) == 0)
							content.append("F9F9F9");
						else
							content.append("F3F3F3");
						content.append("\">");
					}
					content.append("<TD>");
					content.append(_environment.translate(objectFieldInfo.getName()));
					content.append("</TD><TD>");
					content.append(_environment.translate(objectFieldInfo.toName(valueObject)));
					content.append("</TD></TR>");
					k++;
				}
			}

			content.append("</TR>\n");

		}
		content.append("</TBODY>\n</TABLE>\n");
		return content.toString();
	} // end method getGeneratedHtmlContent

	/**
	 * Generates the sender name of an email message using the given user name and email address in
	 * the meta-model.
	 *
	 * @return a String representing an email address with its associated name
	 */
	public String getGeneratedSender()
	{
		String  originalName = (String)getParameter(FROM_USER_NAME);
		String  originalAddress = (String)getParameter(FROM_MAIL_ADDRESS);
		boolean name = false;

		String  sender = "";

		if (_fromName != null)
		{
			sender = _fromName.trim();
			name = true;
		}
		else if (_session.getUser() != null)
		{
			sender = _session.getUser().getName();
			name = true;
		}
		else if (originalName != null)
		{
			sender = originalName;
			name = true;
		}

		if (_fromAddress != null)
			sender += (name ? " <" : "") + _fromAddress + (name ? ">" : "");
		else if (originalAddress != null)
			sender += (name ? " <" : "") + originalAddress + (name ? ">" : "");
		return sender.trim();
	} // end method getGeneratedSender

	/**
	 * Checks that the given field description has the right marks to be displayed in a consultation
	 * form. In this case, the consultation form is the content of an email that shows the content
	 * of the selected objects. The field is consultable only if the {@link LyFieldInfo#MARK_HIDDEN}
	 * and {@link LyFieldInfo#MARK_SECRET} marks are not set.
	 *
	 * @param  objectFieldInfo the field information of a specific object that is tested for its
	 *                         consultation marks
	 * @return true if the given field can be displayed on a consultation form, false otherwise
	 */
	protected boolean isConsultable(LyFieldInfo objectFieldInfo)
	{
		BitSet properties = new BitSet();

		properties.set(LyFieldInfo.MARK_HIDDEN);
		properties.set(LyFieldInfo.MARK_SECRET);

		return !objectFieldInfo.hasMarks(properties, false);
	}

	/**
	 * @see leon.control.LyAreaController#showView()
	 */
	public void showView()
	{
		retrieveValues();

		LyView  view = getView();

		boolean showBuilder = ((_sendSelected && (_toUserClass == null))
					|| (!_sendSelected && (_rawMessage == null) && (_htmlMessage == null))
					|| _sendSelected
					|| (_toUserField == null)
					|| (_authentication && ((_authLogin == null) || (_authPassword == null)))
					|| (getValue(AUTHENTICATION, view) == null));

		if (showBuilder)
			super.showView();
		else
			validateMail();
	}

	/**
	 * Accessor to the authentication parameter.
	 *
	 * @return true if the SMTP server needs authentication, false otherwise.
	 */
	public boolean getAuthentication()
	{
		return _authentication;
	}

	/**
	 * Accessor to the login used on SMTP authentication.
	 *
	 * @return the SMTP authentication login
	 */
	public String getAuthLogin()
	{
		return _authLogin;
	}

	/**
	 * Accessor to the password used on SMTP authentication.
	 *
	 * @return the SMTP authentication password
	 */
	public String getAuthPassword()
	{
		return _authPassword;
	}

	/**
	 * Accessor to the realm used on CRAM-MD5 authentication.
	 *
	 * @return the authentication realm
	 */
	public String getAuthRealm()
	{
		return _authRealm;
	}

	/**
	 * Accessor to the SMTP server address.
	 *
	 * @return the SMTP server address, either at a full name or as an IP
	 */
	public String getSmtpServer()
	{
		return _smtpServer;
	}

	/**
	 * Accessor to the SMTP port number.
	 *
	 * @return the port number used for SMTP connection. By default, it is 25.
	 */
	public int getSmtpPort()
	{
		return _smtpPort;
	}

	/**
	 * Accessor to the user name of the sender.
	 *
	 * @return the name of the sending user. This value may be null or empty.
	 */
	public String getFromUserName()
	{
		return _fromName;
	}

	/**
	 * Accessor to the email address of the sender.
	 *
	 * @return the email address of the sending user. This value may be null or empty.
	 */
	public String getFromAddress()
	{
		return _fromAddress;
	}

	/**
	 * Accessor to the identifier of the class that contains the recipients for the email message.
	 *
	 * @return a class identifier
	 */
	public String getToUserClass()
	{
		return _toUserClass;
	}

	/**
	 * Accessor to the identifier of the field that contains the email addresses for the email
	 * message's recipients.
	 *
	 * @return a field identifier
	 */
	public String getToUserField()
	{
		return _toUserField;
	}

	/**
	 * Accessor to the subject of a custom email message.
	 *
	 * @return the subject field of the custom message to send. This value may be null or empty.
	 */
	public String getSubject()
	{
		return _subject;
	}

	/**
	 * Accessor to the raw message of the custom email message.
	 *
	 * @return the text for the unformatted part of the custom email message.
	 */
	public String getRawMessage()
	{
		return _rawMessage;
	}

	/**
	 * Accessor to the file that contains the formatted part of the custom email message.
	 *
	 * @return a file with HTML content.
	 */
	public File getHtmlMessage()
	{
		return _htmlMessage;
	}

	/**
	 * Accessor to the attached file of the custom email message.
	 *
	 * @return a file that is associated to the email message and that is sent as part of the
	 *         message.
	 */
	public File getAttachedFile()
	{
		return _attachedFile;
	}

	/**
	 * Indicates that the selected objects have to be sent.
	 *
	 * @return true if the selected objects are sent, false if the selected objects represents users
	 *         to send a message
	 */
	public boolean isSendSelected()
	{
		return _sendSelected;
	}
} // end class LyMailController
