/**
 * LyMultiAreaController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
* Current version: $Id: LyMultiAreaController.java 35030 2012-03-23 17:12:33Z jlsaouli $
 */

package leon.control;

import java.util.*;
import leon.app.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.setinterface.*;


/**
 * A compound area controller is associated to a compound action.
 *
 * <P>This controller is a view which gather two views controllers (subcontrollers) in the same
 * 'area'.</P>
 *
 * <P>The compoundAreaController provides the following services :</P>
 *
 * <UL>
 *   <LI>Management of the list of sub-controllers (initialize, free)</LI>
 *   <LI>context management between sub-controllers (views)</LI>
 * </UL>
 */
public abstract class LyMultiAreaController extends LyAreaController
{
	/** File Version */
	public static final String		   __VERSION =
		"$Id: LyMultiAreaController.java 35030 2012-03-23 17:12:33Z jlsaouli $";

	/** List of sub-area controllers contained in this view. */
	protected Vector<LyAreaController> _areaControllers;

	/**
	 * List of sub-area controllers contained in the stack. This list contains all the controllers
	 * opened from current controller in the LY_OPEN_DIALOGS_IN_SAME_VIEW mode.
	 */
	protected Vector<LyAreaController> _areaControllerStack;

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 * @param action  related action. The action should be an instance of LyMultiAction. If an
	 *                action is specified, the controller uses action parameters (action name,
	 *                action specific XML view description ...)
	 * @param context current context (may be null if the view is not contextual to a list of
	 *                objects).
	 */
	public LyMultiAreaController(LySession session, LyController parent, LyAction action,
		LyObjectList context)
	{
		super(session, parent, action, context);

		_areaControllers = new Vector<LyAreaController>(5);
		_areaControllerStack = new Vector<LyAreaController>(5);
	}

	/**
	 * Prepare component before showing. This method is called once (unlike reset which may be
	 * called several times), before first showing. Overrides parent method to propagate the
	 * initialization on sub-controllers
	 */
	protected void initArea()
	{
		if (_initialized)
			return;

		super.initArea();

		if (isFreed())
			return;

		_initialized = false;

		LyAreaController areaController;
		int				 i, n = getAreaControllersCount();

		for (i = 0; i < n; i++)
		{
			areaController = getAreaController(i);
			areaController.initArea();
            if (areaController.isFreed())
            {
                removeAreaController(areaController);
                i--;
                n--;
            }
		}

		_initialized = true;
	}

	/**
	 * Set current controller context. This method set the context of the sub AreaController Two
	 * possibilities: One of the subareas is a context for the other one, so this method set its
	 * context. Both Areas are &quot;equal&quot; so the context is set for both.
	 *
	 * @param context current context or null if none is specified
	 * @return true if successfull
	 * @see   LyContextInterface#setContext
	 */
	public boolean setContext(LyWorkSpace context)
	{
		if ((_action != null) && _action.hasMark(LyAction.MARK_NO_CONTEXT))
			return false;

		boolean			 resetted = super.setContext(context);
		LyAreaController areaController = null;
		int				 i, n = getAreaControllersCount();

		for (i = 0; i < n; i++)
		{
			areaController = getAreaController(i);
			if (areaController != null)
				resetted |= areaController.setContext(context);
		}

		return resetted;
	}

	/**
	 * Accessor to set another area controller which will receive contextual notification when
	 * something occur on this area controller.
	 *
	 * <P>Method is overrident to take the internal sub-controller structure into account.</P>
	 *
	 * @param controller the contextual controller, or null if no client defined
	 */
	public void setContextClient(LyAreaController controller)
	{
		super.setContextClient(controller);

		LyAreaController areaController = null;
		int				 i, n = getAreaControllersCount();

		for (i = 0; i < n; i++)
		{
			areaController = getAreaController(i);
			areaController.setContextClient(controller);
		}
	}

	/**
	 * Invoked when an action occurs. At this level, the implementation manages
	 *
	 * <UL>
	 *   <LI>HREF demands. (demand.getType() = LyActionDemand.TYPE_ACTION AND demand.getSource()
	 *     starts with "_formStack"</LI>
	 * </UL>
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		String sourceId = demand.getSource().intern();

		if ((demand.getType() == LyActionDemand.TYPE_HREF)
				&& sourceId.startsWith("_formStack"))
		{
			// If demand comes from the stack form, display controller corresponding
			// to this demand.
			String			 controllerId = sourceId.substring("_formStack".length());
			int				 i, n = getAreaControllersCount();
			LyAreaController areaController = null;

			for (i = 0; i < n; i++)
			{
				areaController = getAreaController(i);

				if (controllerId.equals(getAreaController(i).getId()))
					break;
			}

			// If no controller was found, try with current one.
			if ((i == n) && controllerId.equals(getId()))
				areaController = this;

			if (areaController != null)
				updateAreaStack(areaController);

			return;
		}

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Accessor to the number of sub-controllers.
	 *
	 * @return the number of sub-controllers, which should be 2 or 0 if still empty.
	 */
	public int getAreaControllersCount()
	{
		return ((_areaControllers == null) ? 0 : _areaControllers.size());
	}

	/**
	 * Accessor to the sub-controllers at the specified index.
	 *
	 * @param  index specified index, from >=0 and < getAreaControllersCount()
	 * @return the reference to the specified controller, or throws an exception if the index is
	 *         invalid.
	 */
	public LyAreaController getAreaController(int index)
	{
		if (_areaControllers == null)
			return null;
		if ((index < 0) || (index >= _areaControllers.size()))
			return null;
		
		return _areaControllers.elementAt(index);
	}

	/**
	 * Accessor to the number of sub-controllers in the stack.
	 *
	 * @return the number of sub-controllers in the stack.
	 */
	public int getAreaControllerStackCount()
	{
		return ((_areaControllerStack == null) ? 0 : _areaControllerStack.size());
	}

	/**
	 * Accessor to the sub-controllers in the stack at the specified index.
	 *
	 * @param  index specified index, from >=0 and < getAreaControllerStackCount()
	 * @return the reference to the specified controller, or throws an exception if the index is
	 *         invalid.
	 */
	public LyAreaController getAreaControllerStack(int index)
	{
		return _areaControllerStack.elementAt(index);
	}

	/**
	 * Accessor to the sub-controllers with the specified action.
	 *
	 * @param  action specified action whose controller must be returned.
	 * @return the reference to the specified controller, or null if no controller was found for
	 *         specified action.
	 */
	public LyAreaController getAreaController(LyAction action)
	{
		int				 i, n = getAreaControllersCount();
		LyAreaController areaController;

		for (i = 0; i < n; i++)
		{
			areaController = _areaControllers.elementAt(i);

			if (areaController.getAction() == action)
				return areaController;
		}

		return null;
	}

	/**
	 * Append a new sub-controllers at the children list This method should not be used directly,
	 * use action description instead (LyMultiAction).
	 *
	 * @param areaController the new controller we want to add.
	 */
	public void addAreaController(LyAreaController areaController)
	{
		if (!_areaControllers.contains(areaController))
			_areaControllers.addElement(areaController);
		areaController.setParent(this);
		areaController.setViewController(_viewController);
	}

	/**
	 * Remove a sub-controllers at the children list This method should not be used directly, use
	 * action description instead (LyMultiAction).
	 *
	 * @param areaController the new controller we want to remove.
	 */
	public void removeAreaController(LyAreaController areaController)
	{
		_areaControllers.removeElement(areaController);
		areaController.setParent(null);
		areaController.setViewController(null);
	}

	/**
	 * Append a new sub-controllers at the stack list.
	 *
	 * @param areaController the new controller we want to add in the stack.
	 */
	public void addAreaControllerStack(LyAreaController areaController)
	{
		_areaControllerStack.addElement(areaController);

	}

	/**
	 * Remove a sub-controllers at the stack list.
	 *
	 * @param areaController the new controller we want to remove from the stack.
	 */
	public void removeAreaControllerStack(LyAreaController areaController)
	{
		_areaControllerStack.removeElement(areaController);
	}

	/**
	 * Accessor to the view controller of the area. The view controller is a Frame / dialog which
	 * contains this area. This method is reserved for internal use and should not be used by the
	 * application. Current implementation overrides generic method to set the parent value on
	 * sub-controllers.
	 *
	 * @param viewController the controller which manages the parent frame / dialog
	 */
	public void setViewController(LyViewController viewController)
	{
		super.setViewController(viewController);

		LyAreaController areaController;
		int				 i, n = getAreaControllersCount();

		for (i = 0; i < n; i++)
		{
			areaController = getAreaController(i);
			areaController.setViewController(viewController);
		}
	}

	/**
	 * Change the main action for this controller.<br />
	 * Default implementation does nothing.
	 *
	 * @param action        New main action.
	 * @param deleteOnClose Delete controller when action is changed.
	 */
	public void setAction(LyAction action, boolean deleteOnClose)
	{
	}

	/**
	 * Invokes an action on the specified class and object list.
	 *
	 * <P>The action process is implemented on the session. This is a shortcut to
	 * getSession().invokeAction All other invokeAction method invoke this method.</P>
	 *
	 * @param action    the action which is executed.
	 * @param classInfo the target class of the action.
	 * @param objects   The list of objects on which the action applies. They must belong to the
	 *                  specified classInfo. Otherwise, use invokeAction(LyAction,LyObjectList)
	 * @param show      indicates if the invoked action must be shown
	 * @return the resulting controller
	 * @see   #invokeAction(LyAction, LyClassInfo, LyObjectList, boolean)
	 * @see   leon.app.LySession#invokeAction
	 */
	protected LyController invokeAction(LyAction action, LyClassInfo classInfo,
		LyObjectList objects, boolean show)
	{
		LySession session = getSession();

		if ((action.getClassInfo() != null)
				&& action.getClassInfo().isVirtual())
			classInfo = action.getClassInfo();

		LyController controller = session.invokeAction(this, action, classInfo, objects, false);

		if (controller instanceof LyAreaController)
		{
		    LyAreaController  subAreaController = (LyAreaController)controller;
		    
		    if ((subAreaController.getContext() == null)
		            && (_context != null))
		        subAreaController.setContext(_context);
		      
			openSubAreaController(subAreaController, true);
		}

		if (show)
			session.showView(controller);

		return controller;
	}

	/**
	 * Indicates if this controller allows sub-areas management (LY_OPEN_DIALOGS_IN_SAME_VIEW mode).
	 * <br />
	 * Default implementation return true but some controllers may not support sub-areas management
	 * (such as tab area controllers that implement their own sub areas policy).
	 *
	 * @return true if this controller allows sub-areas management, false otherwise.
	 */
	protected boolean allowSubAreas()
	{
		return true;
	}

	/**
	 * Opens a sub-area controller and updates the state of the view.<br />
	 * Depending on the LY_OPEN_DIALOGS_IN_SAME_VIEW, this method :
	 *
	 * <UL>
	 *   <LI>does nothing</LI>
	 *   <LI>adds an hyperlink area to navigate through areas</LI>
	 * </UL>
	 *
	 * @param areaController areaController opened from current form controller.
	 * @param open           indicates whether areaController opened from current form controller is
	 *                       opening or closing.
	 */
	public void openSubAreaController(LyAreaController areaController, boolean open)
	{
		// Retrieve the resource that indicates if transverse actions must be displayed in the same window
		String		  dialogsMode = getEnvironment().getResource("LY_OPEN_DIALOGS_IN_SAME_VIEW");
		boolean		  horizontal = "HORIZONTAL".equals(dialogsMode);
		boolean		  vertical = "VERTICAL".equals(dialogsMode);
		boolean		  useSameWindow = horizontal || vertical;
		LyAction	  action = getAction();

		if (action != null)
			useSameWindow &= !action.hasMark(LyAction.MARK_FRAME_VIEW);

		useSameWindow &= allowSubAreas();

		if (useSameWindow)
		{
			LyMultiAreaController formStackController = ((_areaStackController == null)
					? this : _areaStackController);
			LyViewController	  viewController = getViewController();

			// Check if parent is a dialog controller or if an area stack controller is already
			// set on current controller
			if ((_areaStackController == null) && !(getParent(false) instanceof LyDialogController))
				return;

			if (open)
			{
				// Register new controller as the new area controller for the view and add it to the
				// formStackController list of managed controllers.
				LyViewController tmpViewController = areaController.getViewController();

				// If the new controller has no view, there is no need to register it.
				if (tmpViewController == null)
					return;

				// Set current controller as the new parent for opened controller
				areaController.setParent(this);
				areaController.setAreaStackController(formStackController);
				areaController.initArea();

				// Controller may have been freed during init process
				if (areaController.isFreed())
				{
					areaController.setParent(null);
					areaController.setViewController(null);
					areaController.setAreaStackController(formStackController);

					// Free temporary view controller
					tmpViewController.setAreaController(null);
					tmpViewController.free();

					return;
				}

				areaController.setViewController(null);

				// Update stack controller with the newly opened controller
				formStackController.addAreaController(areaController);
				formStackController.addAreaControllerStack(areaController);

				// Update both view controller and view component
				viewController.setAreaController(areaController);
				areaController.setViewController(viewController);
				areaController.setParent(this);
				formStackController.updateAreaStack(null);

				// Free temporary view controller
				tmpViewController.setAreaController(null);
				tmpViewController.free();
			}
			else
			{
				LyController oldAreaControllerParent = areaController.getParent();

				// Check if removed area controller was displayed on the form stack controller.
				// If not, there is no need to update the area.
				Vector<LyAreaController>	stack = formStackController._areaControllerStack;
				boolean		 wasDisplayed = ((stack == null) ? false
															 : stack.contains(areaController));

				// Remove controller from the list of managed controllers on the
				// formStackController then set the last area controller as the new area
				// controller for the view.
				formStackController.removeAreaController(areaController);
				formStackController.removeAreaControllerStack(areaController);

				LyAreaController lastAreaController = this;

				if (getAreaControllerStackCount() > 0)
					lastAreaController = getAreaControllerStack(getAreaControllerStackCount() - 1);

				LyController oldParent = lastAreaController.getParent();

				// Update the view controller with the last area controller without
				// changing the last area controller's parent
				areaController.setParent(oldAreaControllerParent);
				viewController.setAreaController(lastAreaController);
				lastAreaController.setParent(oldParent);

				if (wasDisplayed)
					formStackController.updateAreaStack(null);
			}
		}
	} // end method openSubAreaController

	/**
	 * Updates the view according to the area controllers stack :
	 *
	 * <UL>
	 *   <LI>if area controllers stack is empty, just display current view</LI>
	 *   <LI>if area controllers stack contains at least one element, adds a form at the top of the
	 *     view with one href for each area controller</LI>
	 * </UL>
	 *
	 * @param displayedAreaController area controller being displayed.
	 */
	protected void updateAreaStack(LyAreaController displayedAreaController)
	{
		// First check if the stack form must be displayed
		int     i, n = getAreaControllerStackCount();
		boolean displayStack = (n > 0);

		if (displayedAreaController == null)
			displayedAreaController = (displayStack ? getAreaControllerStack(n - 1) : this);

		LyArea subAreaRoot = (LyArea)displayedAreaController.getComponent();
		LyView view = getView();
		LyArea topArea;
		LyArea formStackArea = (LyArea)view.getComponent("_formStackArea");
		LyArea formArea = (LyArea)view.getComponent("_formArea");

		if (displayStack)
		{
			// If stack must be displayed, build a form with one href for each area controller.
			String			 dialogsMode = getEnvironment().getResource(
					"LY_OPEN_DIALOGS_IN_SAME_VIEW");
			boolean			 vertical = "VERTICAL".equals(dialogsMode);
			LyAreaController areaController;
			LyArea			 controllerArea;
			LyForm			 form = new LyForm("_hrefsForm");
			LyFormField		 formField;
			LyHRef			 href;
			LyConstraint     constraint;

			constraint = new LyConstraint();
			constraint.setFill(LyConstraint.Fill.NONE);
			form.setConstraint(constraint);
			form.setBorder(false);
			form.setController(getId());
			form.setEditable(false);
			constraint = new LyConstraint();
			constraint.setRemainder(false);
			constraint.setColSpan(1);

			for (i = -1; i < n; i++)
			{
				if (i == -1)
					areaController = this;
				else
					areaController = getAreaControllerStack(i);

				// Build href for current controller
				href = new LyHRef();
				href.setId("_formStackHRef" + areaController.getId());
				href.setLabel(areaController.getTitle());
				href.setImage("LY_DIALOG_STACK_SEPARATOR");

				// Href is sensitive only if area controller is not being displayed.
				href.setSensitive(displayedAreaController != areaController);
				formField = new LyFormField("_formStack" + areaController.getId());
				formField.setEditable(false);
				formField.setFieldItem(href);

				if (!vertical)
					formField.setConstraint(constraint);

				form.addFormField(formField);
				controllerArea = (LyArea)areaController.getComponent();
				controllerArea.freeze(i != (n - 1));
			}

			if ((formArea != null) && (formStackArea != null)
					&& !formArea.isFreed() && !formStackArea.isFreed())
			{
				// If form and bottom area already exists, just update their content.
				formStackArea.removePanel(formStackArea.getPanel(0));
				formStackArea.addPanel(form);
				formArea.removePanel(formArea.getPanel(0));
				formArea.addPanel(subAreaRoot);

				return;
			}

			// Build the view structure, a splitter component with the form on the top and
			// the area containing the view of the controller view on the bottom.
			formStackArea = new LyArea();
			formStackArea.setId("_formStackArea");
			formStackArea.addPanel(form);
			formStackArea.setScrolling(LyArea.Scrolling.NO);

			formArea = new LyArea();
			formArea.setId("_formArea");
			formArea.setScrolling(LyArea.Scrolling.NO);
			formArea.addPanel(subAreaRoot);
			
			constraint = new LyConstraint();
			constraint.setColSpan(1);
			constraint.setRemainder(true);
			constraint.setFill(LyConstraint.Fill.NONE);

			if (vertical)
				constraint.setRemainder(false);

			formStackArea.setConstraint(constraint);

			topArea = new LyArea();
			topArea.setScrolling(LyArea.Scrolling.NO);
			topArea.setBorder(LyArea.Border.NONE);
			topArea.addPanel(formStackArea);
			topArea.addPanel(formArea);

			displayedAreaController.updateTitleBar();
		}
		else
		{
			// If stack is not displayed, just display the view of displayed area controller
			topArea = subAreaRoot;
			subAreaRoot.freeze(false);

			if (formArea != null)
				formArea.free();

			if (formStackArea != null)
				formStackArea.free();

			view.resetChildrenCache();
		}

		view.setArea(topArea);
	} // end method updateAreaStack

	/**
	 * Free the area controllers contained in the stack (LY_OPEN_DIALOGS_IN_SAME_VIEW mode only).
	 */
	protected void freeAreaStack()
	{
		int				 i, n = getAreaControllerStackCount();
		LyAreaController areaController;

		// Free area controllers
		for (i = n - 1; i >= -1; i--)
		{
			if (i == -1)
				areaController = this;
			else
				areaController = getAreaControllerStack(i);

			areaController.clearViewCache();
			areaController.free(areaController.getParent(false));
		}
	}

	/**
	 * Free this controller. Current implementation overrides generic method to free sub-controllers
	 * recursively
	 *
	 * @param controller controller instance which requested the freeing of this controller. If the
	 *                   controller is the parent, this controller is freed otherwise the method
	 *                   does nothing.
	 */
	public void free(LyController controller)
	{
		if (controller != _parent)
			return;

		LyAreaController areaController;
		int				 i, n = getAreaControllersCount();

		// Free area controllers
		for (i = n - 1; i >= 0; i--)
		{
			areaController = getAreaController(i);
			areaController.free(this);
		}

		_areaControllers = null;

		super.free(controller);
	}

	/**
	 * Check if the area controller may be closed. This method recursively check canClose on
	 * sub-controllers.
	 *
	 * @param  showMessage indicates if the controller can show an warning message or not
	 * @return return the status of sub-controllers
	 */
	public boolean canClose(boolean showMessage)
	{
		LyAreaController controller;
		int				 i, n = getAreaControllersCount();
		boolean			 canClose = true;

		for (i = 0; (i < n) && canClose; i++)
		{
			controller = getAreaController(i);
			canClose &= controller.canClose(showMessage);
		}

		return canClose;
	}

	/**
	 * Accessor to the last modification date of this controller. The modification date is the
	 * latest modification date on sub-controllers
	 *
	 * @return the last modification date of sub-controllers (as described in
	 *         java.util.Date.getTime())
	 */
	public long getLastModified()
	{
		// The last modification date of a view controller is the most recent
		// last modification date of the area controllers.
		int				 i, n = getAreaControllersCount();
		LyAreaController areaController;
		long			 lastModified;

		for (i = 0; i < n; i++)
		{
			areaController = getAreaController(i);
			lastModified = areaController.getLastModified();
			if (_lastModified < lastModified)
				_lastModified = lastModified;
		}

		return _lastModified;
	}

	/**
	 * Clears all cached data for this controller (especially the view and the component).
	 *
	 * <P>This method may be invoked for memory optimization, but the controller is still valid and
	 * may be used. (Unlike free) Further usage, needs to rebuild internal component & view which
	 * may be time consuming.</P>
	 */
	public void clearViewCache()
	{
		super.clearViewCache();

		LyAreaController areaController;
		int				 i, n = getAreaControllersCount();

		for (i = 0; i < n; i++)
		{
			areaController = getAreaController(i);
			areaController.clearViewCache();
		}
	}

	/**
	 * Indicates if the view for this compound controller is resizable. View is resizable if at
	 * least one sub-view is resizable.
	 *
	 * @return true if the view is resizable, false otherwise.
	 */
	public boolean isViewResizable()
	{
		LyAreaController areaController;
		int				 i, n = getAreaControllersCount();

		for (i = 0; i < n; i++)
		{
			areaController = getAreaController(i);

			// If at least one view is resizable, compound view is too.
			if (areaController.isViewResizable())
				return true;
		}

		return false;
	}

	/**
	 * Accessor to all the sub-controllers that are not instances of LyMultiAreaController.
	 *
	 * @return A vector that contains all the sub-controllers.
	 */
	public Vector<LyController> getFinalAreaControllers()
	{
		Vector<LyController> result = new Vector<LyController>();
		int					 i, j, n = getAreaControllersCount();
		Vector<LyController> subResult = new Vector<LyController>();
		LyAreaController     subController;

		for (i = 0; i < n; i++)
		{
			subController = getAreaController(i);

			if (!(subController instanceof LyMultiAreaController))
				result.addElement(subController);
			else
			{
				subResult = ((LyMultiAreaController)subController).getFinalAreaControllers();

				for (j = 0; j < subResult.size(); j++)
					result.addElement(subResult.elementAt(j));
			}
		}

		return result;
	}

	/**
	 * Update the menu for the current controller.
	 */
	protected void updateMenuBar()
	{
		int i, n = getAreaControllersCount();

		for (i = 0; i < n; i++)
			getAreaController(i).updateMenuBar();

		super.updateMenuBar();
	}

	/**
	 * Update menus of menu bar.
	 */
	protected void updateMenus()
	{
		if (_component == null)
			return;

		LyViewManager viewManager = getViewManager();
		boolean		  hasMenu = viewManager.hasProperty("MENU");

		if (!hasMenu)
			return;

		LyView    view = getView();
        LyMenuBar menuBar = (view instanceof LyFrame ? ((LyFrame)view).getMenuBar() : null);

		if (menuBar == null)
			return;

		// Remove all menus from menubar
		boolean wasIdle = menuBar.isIdle();

		if (!wasIdle)
			menuBar.setIdle(true);

		int    i, n = menuBar.getMenuCount();
		LyMenu menu;

		for (i = n - 1; i >= 0; i--)
		{
			menu = menuBar.getMenu(i);
			menuBar.removeMenu(menu);
		}

		LyFrameController frameController = (LyFrameController)getViewController();
		LyFrameBuilder    frameBuilder = (LyFrameBuilder)frameController.getBuilder();
		LyMenuBar		  menuBar1 = new LyMenuBar();

		menuBar1.setSorted(true);
		menuBar1.setId("_menuBar");
		frameBuilder.completeMenuBar(menuBar1, LyViewController.LEVEL_ROOT);

		// Don't know why, but if we use the current menubar, the order of the menu / item is
		// wrong and there is no separator
		while (menuBar1.getMenuCount() > 0)
		{
			menu = menuBar1.getMenu(0);
			// changing menu parent removes menu from previous menubar
			menuBar1.removeMenu(menu);
			menuBar.addMenu(menu);
		}

		menuBar1 = null;

		// Update the menubar for the top most area controller
		// (because the tabAreaController can be in a multiAreaController
		LyAreaController areaController = frameController.getAreaController();

		areaController.updateMenuBar();

		if (!wasIdle)
			menuBar.setIdle(false);
	} // end method updateMenus

	/**
	 * Reset the message bar status if any The current implementation does nothing.
	 */
	protected void updateMessages()
	{
		int i, n = getAreaControllersCount();

		for (i = 0; i < n; i++)
			getAreaController(i).updateMessages();

		super.updateMessages();
	}

	/**
	 * Update messages of message bar.
	 */
	protected void updateMessageBar()
	{
		if (_component == null)
			return;

		LyView view = getView();

		if (!(view instanceof LyFrame))
			return;

		// Get message bar
		LyMessages messages = ((LyFrame)view).getMessages();

		if (messages == null)
		{
			messages = new LyMessages();
			((LyFrame)view).setMessages(messages);
		}

		// If internal view manages idle mode, set this mode to avoid graphical
		// refresh during message bar update.
		Object internalMessages = messages.getInternalView(false);

		if (internalMessages instanceof LySetIdleInterface)
			((LySetIdleInterface)internalMessages).setIdle(true);

		// Remove all messages from message bar
		int		  i, n = messages.getMessageCount();
		LyMessage message;

		for (i = n - 1; i >= 0; i--)
		{
			message = messages.getMessage(i);
			messages.removeMessage(message);
			message.free();
		}

		// Rebuild message bar
		LyFrameController frameController = (LyFrameController)getViewController();
		LyFrameBuilder    frameBuilder = (LyFrameBuilder)frameController.getBuilder();

		frameBuilder.completeMessages(messages, LyViewController.LEVEL_ROOT);

		// Update the messagebar : we must use the high level area controller
		// because this multi area may be in another multi area.
		LyAreaController areaController = frameController.getAreaController();

		areaController.updateMessages();

		if (internalMessages instanceof LySetIdleInterface)
			((LySetIdleInterface)internalMessages).setIdle(false);
	} // end method updateMessageBar

	/**
	 * This method is called by the view controller when this area is shown to update its content.
	 * Default implementation calls update on each sub-controller.
	 * 
	 * @return true if successfull
	 */
	public boolean update()
	{
		LyAreaController areaController;
		int				 i, n = getAreaControllersCount();
		boolean			 updated = false;

		for (i = 0; i < n; i++)
		{
			areaController = getAreaController(i);
			updated = areaController.update() & updated;
		}

		return super.update() & updated;
	}

	/**
	 * Sets idle mode on this controller. Default implementation calls setIdle on each
	 * sub-controller.
	 * @param idle the idle status
	 */
	public void setIdle(boolean idle)
	{
		LyAreaController areaController;
		int				 i, n = getAreaControllersCount();

		for (i = 0; i < n; i++)
		{
			areaController = getAreaController(i);
			areaController.setIdle(idle);
		}

		super.setIdle(idle);
	}
	
	/**
	 * Notify a visibility change of the parent view.
	 * @param visible the visible status
	 * @see #setVisible(boolean)
	 */
	protected void visibilityChanged(boolean visible)
	{
		LyAreaController areaController;
		int				 i, n = getAreaControllersCount();

		for (i = 0; i < n; i++)
		{
			areaController = getAreaController(i);

			// If at least one view is resizable, compound view is too.
			if (areaController != null)
				areaController.visibilityChanged(visible);
		}

	}
} // end class LyMultiAreaController
