/**
 * LyMultiClassExportController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyMultiClassExportController.java 28422 2010-11-03 12:03:38Z ehelmste $
 */

package leon.control;

import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.data.*;
import leon.info.*;
import leon.peer.*;
import leon.peer.file.*;
import leon.peer.xml.*;


/**
 * This class implements an export controller (no GUI) to xporta set of data from multiple
 * classInfos into XML (only XML is available).
 *
 * @see LyExportController
 */
public class LyMultiClassExportController extends LyExportController
{
	/** File Version */
	public static final String __VERSION =
		"$Id: LyMultiClassExportController.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The first class in the list. */
	protected boolean		   _first;

	/** The last class in the list. */
	protected boolean		   _last;

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param action  related action. If an action is specified, the controller uses action
	 *                parameters (action name, action specific XML view description ...)
	 * @param objects list of objects that should be saved.
	 */
	public LyMultiClassExportController(LySession session, LyAction action, LyObjectList objects)
	{
		this(session, null, action, null, objects);
	}

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller).
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo kept for compatibility. This parameter is *NOT* used.
	 * @param objects   list of objects that should be saved.
	 */
	public LyMultiClassExportController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		super(session, parent, action, classInfo, objects);

		_classInfo = null;
	}

	/**
	 * Method to check if parameters are valid for export.
	 */
	protected void endInit()
	{
		// Nothing to do : do not check _classInfoValidity
	}

	/**
	 * Return the lines that will be written in the export file.
	 *
	 * @param  fields     list of fieldInfos that should be exported to the file.
	 * @param  fieldSep   separator between fields (columns)
	 * @param  valueSep   separator between multiple values
	 * @return the lines that will be written in the export file
	 * @param  exportType type of data to export : Possible values are SELECT_EXPORT_TYPE_TEXT
	 *                    (.txt), SELECT_EXPORT_TYPE_RAW (.res) or SELECT_EXPORT_TYPE_SPREADSHEET
	 *                    (.csv).
	 */
	protected Vector<String> getExportLines(Vector<LyFieldInfo> fields, String fieldSep, String valueSep,
		String exportType)
	{
		if (exportType != SELECT_EXPORT_TYPE_XML)
			throw new IllegalArgumentException(
				"BF AppEngine : The exportType parameter must be SELECT_EXPORT_TYPE_XML");

		// Default values. Doesn't matter for the xml export.
		if (exportType == SELECT_EXPORT_TYPE_XML)
		{
			fieldSep = "|";
			valueSep = ",";
		}

		LyInfoList<LyClassInfo> classes = _context.getClasses();
		LyObjectList		    exportContext = _context;
		Vector<String>		    result = new Vector<String>();
		Vector<LyObject>	    initContext = _context.getObjects();
		LySimpleObjectList	    contextTmp = new LySimpleObjectList();
		LyObject			    object;

		for (int i = 0; i < classes.getSize(); i++)
		{
			// Init general parameters for parent
			_classInfo = classes.getInfo(i);
			_first = (i == 0);
			_last = (i == (classes.getSize() - 1));

			// If fields specified keep only fields of this classInfo
			Vector<LyFieldInfo> currentFields = null;

			if (fields != null)
			{
				currentFields = new Vector<LyFieldInfo>(fields);
				for (int j = currentFields.size() - 1; j >= 0; j--)
				{
					LyFieldInfo fieldInfo = currentFields.elementAt(j);

					if (!_classInfo.contains(fieldInfo))
						currentFields.removeElement(fieldInfo);
				}
			}

			contextTmp = new LySimpleObjectList();

			for (int j = 0; j < initContext.size(); j++)
			{
				object = initContext.elementAt(j);

				if (object.getClassInfo() == _classInfo)
					contextTmp.addObject(object);
			}

			setContext(contextTmp);

			Vector<String> v = super.getExportLines(currentFields, fieldSep, valueSep, exportType);

			for (int j = 0; j < v.size(); j++)
				result.addElement(v.elementAt(j));
		}

		setContext(exportContext);
		return result;
	} // end method getExportLines

	/**
	 * Factory for the export request encoder.
	 *
	 * @param  dataSource   collection of data from the classInfo to write
	 * @param  dataProvider connector
	 * @param  fieldSep     separator betwen fields
	 * @param  valueSep     separator between values (multiple)
	 * @param  exportType   one of the SELECT_xxx types
	 * @return the export request encoder
	 */
	protected LyFileRequestEncoder requestEncoderFactory(LyDataSource dataSource,
		LyDataProvider dataProvider, String fieldSep, String valueSep, String exportType)
	{
		if (exportType == SELECT_EXPORT_TYPE_XML)
		{
			LyXmlExportRequestEncoder encoder = new LyXmlExportRequestEncoder(dataProvider,
					dataSource, fieldSep, valueSep);

			if (!_first)
				encoder.skipHeader(true);
			if (!_last)
				encoder.skipFooter(true);
			return encoder;
		}
		else
			return super.requestEncoderFactory(dataSource, dataProvider, fieldSep, valueSep,
				exportType);
	}

	/**
	 * Hide the view. The effect of this command depends on the nature of the application. If the
	 * display is a fat client for example (windows (awt) / swing) this methods should close the
	 * window or hide the view.
	 *
	 * <P>This method should consider that the page is closed and should disappear.</P>
	 *
	 * @param free Indicates if only HIDDEN, not FREED (true).
	 */
	public void hideView(boolean free)
	{
		// NO GUI YET
	}

	/**
	 * Accessor to the view controller (Frame or dialog controller) which is the container of this
	 * view. This method must be implemented in subclasses depending on the nature of the component.
	 *
	 * @return the reference to a view controller, never returns null if the controller is
	 *         associated to a view.
	 */
	public LyViewController getViewController()
	{
		// No Gui yet
		return null;
	}

	/**
	 * Free this controller. Overrides method from LyController :<BR>
	 * If a parent is set, propagate the free request on parent<BR>
	 * If no parent is set, free current controller immediately.
	 *
	 * <P>The parent will then invoke free(LyController controller) Method is set final, since
	 * propagation in subclasses must use free(LyController controller)</P>
	 *
	 * @see #free(LyController)
	 */
	public void free()
	{
		free(_parent);
	}

	/**
	 * Make the view visible to the user. The effect of this command depends on the nature of the
	 * application. If the display is a fat client (windows (awt) / swing) this methods should open
	 * a window and show the view. In the case of thin client (html), this prepares the pages which
	 * will be returned and displayed in the client browser.
	 *
	 * <P>For any display, this method should consider that the page is ready to be shown and should
	 * be displayed.</P>
	 */
	public void showView()
	{
		LyBehaviorInterface behavior = getBehavior();

		if (behavior instanceof LyActionBehaviorInterface)
			((LyActionBehaviorInterface)behavior).startAction(this, getAction());
	}
} // end class LyMultiClassExportController
