/**
 * LyMultiClassExportController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyMultiClassImportController.java 28997 2010-12-23 10:17:15Z bmichel $
 */

package leon.control;

import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.peer.xml.*;


/**
 * This class implements an export controller (no GUI) to xporta set of data from multiple
 * classInfos into XML (only XML is available).
 *
 * @see LyExportController
 */
public class LyMultiClassImportController extends LyImportTableController
{
	/** File Version */
	public static final String		  __VERSION =
		"$Id: LyMultiClassImportController.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** Indicates if import is in progress */
	protected boolean				  _running;

	/** List of the (real) classInfos to be imported. */
	protected LyInfoList<LyClassInfo> _classInfos;

	/** List of the (temporary) classInfos to be imported. */
	protected Hashtable<LyClassInfo, LyClassInfo>	_importClassInfos;

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param action  related action. If an action is specified, the controller uses action
	 *                parameters (action name, action specific XML view description ...)
	 */
	public LyMultiClassImportController(LySession session, LyAction action)
	{
		this(session, null, action, null, null);
	}

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller).
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo kept for compatibility. This parameter is *NOT* used.
	 * @param objects   list of objects that should be saved.
	 */
	public LyMultiClassImportController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		// No GUI yet
		super(session, parent, action, classInfo, true);

		_running = false;
		_classInfos = null;
		_importClassInfos = null;
	}

	/**
	 * Init this instance.
	 */
	protected void init()
	{
		init(true);
	}

	/**
	 * Method to check if parameters are valid for export.
	 */
	protected void endInit()
	{
		// Nothing to do : do not check _classInfoValidity
	}

	/**
	 * Import objects from an XML file for given class infos.
	 *
	 *@return true if successfull
	 */
	public boolean loadClassInfos()
	{
		_classInfos = new LyInfoList<LyClassInfo>();

		Vector<String> classInfosNames;

		if (_buffer != null)
			classInfosNames = LyXmlExportRequestEncoder.getClassInfosNamesFromBuffer(_buffer);
		else
			classInfosNames = LyXmlExportRequestEncoder.getClassInfosNames(_file);

		if ((classInfosNames != null) && (classInfosNames.size() > 0))
		{
			String	    classInfoName;
			LyClassInfo classInfo;
			int		    i, n = classInfosNames.size();

			for (i = 0; i < n; i++)
			{
				classInfoName = classInfosNames.elementAt(i);
				classInfo = getClassInfo(classInfoName);
				_classInfos.addInfo(classInfo);
			}
		}

		return true;
	}

	/**
	 * Import objects from an XML file for given class infos.
	 *
	 * @return true if sucessfull
	 */
	public boolean loadData()
	{
		// If no class info is given, retrieve the list of class infos from the
		// XML request encoder
		if ((_classInfos == null) || (_classInfos.getSize() == 0))
		{
			boolean resu = loadClassInfos();

			if (!resu)
				return false;
		}
		_importClassInfos = new Hashtable<LyClassInfo,LyClassInfo>();

		int i, n = _classInfos.getSize();

		_importType = LyImportController.SELECT_IMPORT_TYPE_XML;
		_running = true;
		OLD_IMPORT = false;

		LySimpleObjectList importedObjectsFull = new LySimpleObjectList();
		LyEnvironment	   environment = getEnvironment();

		environment.logTrace("## Loading " + n + " classinfos in multiclass import.");
		String				    errorMsg = null;

		LyInfoList<LyFieldInfo> globalFieldList = _fields;
		LyInfoList<LyFieldInfo> fieldList;

		for (i = 0; i < n; i++)
		{
			_classInfo = _classInfos.getInfo(i);
			fieldList = _classInfo.getFieldInfos();

			// If we have a list of fields to import, use it to restreint the final field list for a given class.
			if (globalFieldList != null)
			{
				_fields = null;
				for (int j = 0; j < globalFieldList.getSize(); j++)
				{
					if (fieldList.contains(globalFieldList.getInfo(j)))
					{
						if (_fields == null)
							_fields = new LyInfoList<LyFieldInfo>();

						_fields.addInfo(globalFieldList.getInfo(j));
					}
				}
				if (_fields == null)
					_fields = fieldList;
			}
			else
				_fields = fieldList;

			environment.logTrace("##   -Loading classInfo in IMPORT (multiClass) "
				+ _classInfo.getId());

			super.loadData();

			if (errorMsg == null)
				errorMsg = getLastErrorMessage(); // To keep an error message if it occurs
			if (_realImportedObjects != null)
			{
				importedObjectsFull.addList(_realImportedObjects);
				environment.logTrace("##   - Import found " + _realImportedObjects.getSize()
					+ " . Cumulative found so far " + importedObjectsFull.getSize());
			}
		}
		_fields = globalFieldList;
		if (errorMsg != null)
			setLastErrorMessage(errorMsg);

		_realImportedObjects = importedObjectsFull;
		_running = false;
		environment.logTrace(" ############### End of LOAD ##############");
		environment.logTrace("##   - Imported objects " + importedObjectsFull.getSize());
		return true;
	} // end method loadData

	/**
	 * Do the final import operation : this will commit the import context for the specified
	 * objects. The result is STILL in the cancellable context of this controller. Only subcontext
	 * are commited.
	 *
	 * @param objects       the temporary object list
	 * @param importErrors  indicates if we accept to import objects in error.
	 * @param allClassInfos indicates if we import all classInfos
	 * @see   #createRealObject
	 * @see   #setRealObject
	 */
	public void importObjects(LyObjectList objects, boolean importErrors, boolean allClassInfos)
	{
		if (allClassInfos)
			importObjects(objects, importErrors);
		else
			super.importObjects(objects, importErrors);
	}

	/**
	 * Do the final import operation : this will commit the import context for the specified
	 * objects. The result is STILL in the cancellable context of this controller. Only subcontext
	 * are commited.
	 *
	 * @param objects      the temporary object list
	 * @param importErrors indicates if we accept to import objects in error.
	 * @see   #createRealObject
	 * @see   #setRealObject
	 */
	public void importObjects(LyObjectList objects, boolean importErrors)
	{
		int i, n = _classInfos.getSize();

		for (i = 0; i < n; i++)
		{
			_classInfo = _classInfos.getInfo(i);
			_fields = _classInfo.getFieldInfos();
			if (OLD_IMPORT)
			{
				_importClassInfo = _importClassInfos.get(_classInfo);

				if (_importClassInfo == null)
				{
					getEnvironment().logWarning(
						"[LyMultiClassInportController]ClassInfo ignored in import : no virtual class.");
					continue;
				}
			}
			initImportFields();

			super.importObjects(objects, importErrors);
		}
	}

	/**
	 * Indicates if the real object corresponding to the imported object has exactly the same values
	 * or not.
	 *
	 * @param  importObject the 'real' object instance for content comparizon
	 * @return null if all values are identical, the list of fields with different values or an
	 *         empty list.
	 */
	@SuppressWarnings("deprecation")
	public LyInfoList<LyFieldInfo> isSameObject(LyObject importObject)
	{
		if (importObject == null)
			return new LyInfoList<LyFieldInfo>();

		LyObject realObject = getRealObject(importObject);

		if (realObject == null)
			return new LyInfoList<LyFieldInfo>();

		_classInfo = realObject.getClassInfo();
		_fields = _classInfo.getFieldInfos();

		return super.isSameObject(importObject);

	}

	/**
	 * Set the list of class infos to be imported (null if all the classes found in file should be
	 * imported).
	 *
	 * @param classInfos list of class infos to be imported
	 */
	public void setClassInfos(LyInfoList<LyClassInfo> classInfos)
	{
		_classInfos = classInfos;
	}

	/**
	 * Get the list of class infos to be imported (null if all the classes found in file should be
	 * imported).
	 *
	 * @return classInfos list of class infos to be imported
	 */
	public LyInfoList<LyClassInfo> getClassInfos()
	{
		return _classInfos;
	}

	/**
	 * Reset component.
	 *
	 * @param full Boolean indicating if a full reset (including reset of datas) is requiered or not
	 *             (only reset graphical area).
	 */
	public void reset(boolean full)
	{
		if (_withoutGui)
			return;
			// Nothing to do (no view)
	}

	/**
	 * Hide the view. The effect of this command depends on the nature of the application. If the
	 * display is a fat client for example (windows (awt) / swing) this methods should close the
	 * window or hide the view.
	 *
	 * <P>This method should consider that the page is closed and should disappear.</P>
	 *
	 * @param free Indicates if only HIDDEN, not FREED (true).
	 */
	public void hideView(boolean free)
	{
		if (_withoutGui)
			return;
			// NO GUI YET
	}

	/**
	 * Accessor to the view controller (Frame or dialog controller) which is the container of this
	 * view. This method must be implemented in subclasses depending on the nature of the component.
	 *
	 * @return the reference to a view controller, never returns null if the controller is
	 *         associated to a view.
	 */
	public LyViewController getViewController()
	{
		if (_withoutGui)
			return null;
		// No Gui yet
		return null;
	}

	/**
	 * Cancel action This method frees the controller.
	 */
	protected void cancel()
	{
		if (_running)
			return;
		else
			super.cancel();
	}

	/**
	 * Free this controller.
	 */
	public void free()
	{
		if (_running)
			return;

		free(_parent);
	}

	/**
	 * Make the view visible to the user. The effect of this command depends on the nature of the
	 * application. If the display is a fat client (windows (awt) / swing) this methods should open
	 * a window and show the view. In the case of thin client (html), this prepares the pages which
	 * will be returned and displayed in the client browser.
	 *
	 * <P>For any display, this method should consider that the page is ready to be shown and should
	 * be displayed.</P>
	 */
	public void showView()
	{
		LyBehaviorInterface behavior = getBehavior();

		if (behavior instanceof LyActionBehaviorInterface)
			((LyActionBehaviorInterface)behavior).startAction(this, getAction());
	}
} // end class LyMultiClassImportController
