/**
 * LyMultiFilterController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyMultiFilterController.java 36407 2012-07-31 13:07:37Z nhamon $
 */

package leon.control;

import leon.app.*;
import leon.control.builder.*;
import leon.control.form.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * This class implements the controller associated for the multi-filter action.
 *
 * @see  LyFilterController
 * @see  LyMultiFilterBuilder
 * @see  LyFilterEndInterface
 */
public class LyMultiFilterController extends LyFilterController
{
	/** File Version. */
    public static final String __VERSION = "$Id: LyMultiFilterController.java 36407 2012-07-31 13:07:37Z nhamon $";

    /**
     * Constructor.
     *
     * @param  session    current session associated to the controller. The controller automatically
     *                    registers on the specified session.
     * @param  parent     parent controller. If the parent implements LyFilterEndInterface, it will
     *                    be notified when the filter form is closed.
     * @param  action     related action. (action based on '_filter' action)
     * @param  classInfo  the application class associated to the filter
     * @param  objects    current context
     */
    public LyMultiFilterController(LySession session, LyController parent, LyAction action, LyClassInfo classInfo,
        LyObjectList objects)
    {
        super(session, parent, action, classInfo, objects);
    }

    /**
     * Invoked when an action occurs. At this level, the implementation manages<BR>
     * DEMAND_VALIDATE demands. (demand.getType = LyActionDemand.TYPE_ACTION AND demand.getSource()
     * = LyController.DEMAND_VALIDATE)<BR>
     * DEMAND_APPLY demands. (demand.getType = LyActionDemand.TYPE_ACTION AND demand.getSource() =
     * LyController.DEMAND_APPLY)<BR>
     * Default action (double click)<BR>
     * Default action (double click)<BR>
     * VALUE_CHANGED demands. (demand.getType = LyActionDemand.TYPE_VALUE_CHANGED
     *
     * @param  demand  the action demand
     *
     * @see    LyController#demandPerformed
     */
    public void actionDemand(LyActionDemand demand)
    {
        String demandCode = demand.getSource().intern();

        if (demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
        {
            boolean checkbox = demandCode.startsWith("_checkbox_");

            demandCode = demandCode.substring(demandCode.indexOf(FORM_FIELD_SEP) + FORM_FIELD_SEP.length());

            LyClassInfo classInfo = getClassInfo();
            LyFieldInfo fieldInfo = classInfo.getFieldInfo(demandCode);

            if (fieldInfo == null)
                throw new IllegalStateException("Leonardi : invalid demand received " + demand);

            LyFormField formField = getFormField(fieldInfo);

            if (checkbox)
            {
                boolean enabled = formField.isChecked();

                setChecked(fieldInfo, enabled);
            }
            else
                formField.setChecked(true);

            setLastModified(System.currentTimeMillis());
            return;
        }
        else if ((demand.getType() == LyActionDemand.TYPE_ACTION)
                && (demandCode == DEMAND_VALIDATE)
                && !(getParent() instanceof LyViewController))
        {

            LyActionDemand de = new LyActionDemand(LyActionDemand.TYPE_ACTION, DEMAND_APPLY);

            de.setParameterValues(demand.getParameterValues());
            de.setValue(demand.getValue());
            super.actionDemand(de);
        }
        else
            super.actionDemand(demand);
    } // end method actionDemand

    /**
     * Indicates whether the form field corresponding to the given field info is checked or not.
     *
     * @param   fieldInfo  the field info
     *
     * @return  true if the form field corresponding to the given field info is checked, false
     *          otherwise
     */
    public boolean isChecked(LyFieldInfo fieldInfo)
    {
        LyFormField formField = getFormField(fieldInfo);

        return formField.isChecked();
    }

    /**
     * Checks or unchecks the form field corresponding to the given field info.
     *
     * @param  fieldInfo  the field info
     * @param  checked    true if corresponding form field must be checked, false otherwise.
     */
    public void setChecked(LyFieldInfo fieldInfo, boolean checked)
    {
        LyFormField formField = getFormField(fieldInfo);

        LyFieldItem item = formField.getFieldItem();

        if (item instanceof LyForm)
        {
            LyForm form = (LyForm)item;

            form.setEditable(checked);
        }

        formField.setEditable(true);
        formField.setEnabled(checked);
        formField.setChecked(checked);
    }

    /**
     * Builds the filter from the data structure This methods extracts values typed in by the user
     * and create the LyFilter instance.
     *
     * @param   demand  the demand which contains all values submited by the form.
     *
     * @return  the LyFilter instance or null if an error occured.
     */
    public LyFilter buildFilter(LyDemand demand)
    {
        // Create field Values corresponding to field items
        LyParameterValues paramValues = demand.getParameterValues();
        LyFilter          filter = new LyFilter();
        LyClassInfo       classInfo = getClassInfo();
        int               i, n = classInfo.getFieldInfoCount();
        LyFieldInfo       fieldInfo;
        LyFilterElement   filterElement;
        LyFormField       formField;

        for (i = 0; i < n; i++)
        {
            fieldInfo = classInfo.getFieldInfo(i);
            if (!isDisplayed(fieldInfo))
                continue; // Skipped

            // We must set the form field to valid to know if getFieldValue sets it to invalid.
            formField = getFormField(fieldInfo);

            if ((formField == null)
                    || !formField.isEnabled())
                continue;

            if ((formField.isCheckBox())
                    && !formField.isChecked())
                continue;

            formField.setState(LyFormField.STATE_NORMAL);

            // Get new value
            filterElement = getFilterElement(fieldInfo, paramValues);

            // Check validity
            if (formField.getState() == LyFormField.STATE_ERROR)
                return null;

            if (filterElement == null)
                continue;

            filter.addElement(filterElement);
        }

        return filter;
    } // end method buildFilter

    /**
     * Build a filter element for the given fieldinfo and the given values.
     *
     * @param   fieldInfo    a fieldInfo
     * @param   paramValues  the current view value
     *
     * @return  a filter element
     */
    protected LyFilterElement getFilterElement(LyFieldInfo fieldInfo, LyParameterValues paramValues)
    {
        boolean checkBox = hasCheckBoxes();
        String  fieldId;

        if ((fieldInfo instanceof LyNumberFieldInfo)
                || (fieldInfo instanceof LyDateFieldInfo))
        {
            String value;
            Object parsedValue;

            fieldId = fieldInfo.getId() + "_1";
            value = paramValues.getParameter(fieldId);

            if (!checkBox
                    && (value == null))
                return null;

            LyFilterExpression exp1 = null, exp2 = null;

            if ((value != null) && (value.trim().length() != 0))
            {
                parsedValue = fieldInfo.parse(value);
                exp1 = new LyFilterExpression(fieldInfo, LyFilterExpression.Condition.GREATER_OR_EQUALS, parsedValue);
            }

            fieldId = fieldInfo.getId() + "_2";
            value = paramValues.getParameter(fieldId);
            if ((value != null) && (value.trim().length() != 0))
            {
                parsedValue = fieldInfo.parse(value);
                exp2 = new LyFilterExpression(fieldInfo, LyFilterExpression.Condition.LOWER_OR_EQUALS, parsedValue);
            }

            if ((exp1 != null) && (exp2 != null))
            {
                LyFilter filter = new LyFilter();

                filter.addElement(exp1);
                filter.addElement(exp2);
                return filter;
            }
            else if (exp1 != null)
                return exp1;
            else if (exp2 != null)
                return exp2;
            else
                return null;
        }

        // Build value
        LyClassInfo classInfo = getClassInfo();
        String[]    values;

        fieldId = classInfo.getId() + LyFormController.FORM_FIELD_SEP + fieldInfo.getId();
        values = paramValues.getParameterValues(fieldId);

        if (!checkBox
                && (values == null))
            return null;

        // Build filter expression
        LyFilterExpression exp = null;

        if (values == null)
        {
            exp = new LyFilterExpression(fieldInfo, null);
            return exp;
        }
        else if (((fieldInfo instanceof LyRelationFieldInfo)
                    || (fieldInfo instanceof LyChoiceFieldInfo))
                && fieldInfo.isMultiple())
        {
            LyApplication app = getSession().getApplication();
            boolean       nullValue = false;

            if (values.length == 1)
            {
                if ((fieldInfo instanceof LyChoiceFieldInfo)
                        && "LY_NO_CHOICE".equals(values[0]))
                    nullValue = true;
                else if ((fieldInfo instanceof LyRelationFieldInfo)
                        && ("LY_NO_CHOICE".equals(values[0])
                            || "LY_FILTER_COMPOSITION_EMPTY_VALUE".equals(values[0])))
                    nullValue = true;
            }

            if (nullValue)
                exp = new LyFilterExpression(fieldInfo, LyFilterExpression.Condition.EQUALS, null);
            else
            {
                Object parsedValue = fieldInfo.parse(LyString.listToString(values, app.VALUE_SEP));

                exp = new LyFilterExpression(fieldInfo, LyFilterExpression.Condition.CONTAINS, parsedValue);
            }

            return exp;
        }
        else
        {
        	if ((fieldInfo instanceof LyRelationFieldInfo)
        			&& !fieldInfo.isMultiple()
        			&& (values.length == 1)
        			&& "LY_NO_CHOICE".equals(values[0]))
        		return new LyFilterExpression(fieldInfo, LyFilterExpression.Condition.EQUALS, null);

        	Object                       parsedValue;
            int                          i, n = values.length;
            LyFilter                     subFilter = null;
            LyFilterExpression.Condition condition;

            for (i = 0; i < n; i++)
            {
                try
                {
                    // If we have a textFieldInfo, and the filter condition is contains, we
                    // must take directly the given value. If there is a format (or a regex)
                    // the parse will failed
                    if (fieldInfo instanceof LyTextFieldInfo)
                    {
                        parsedValue = LyString.replace(values[i], "\r", "");
                        condition = LyFilterExpression.Condition.CONTAINS;
                    }
                    else
                    {
                        parsedValue = fieldInfo.parse(values[i]);
                        condition = LyFilterExpression.Condition.EQUALS;
                    }
                }
                catch (LyFieldInfoFormatException e)
                {
                    LyFormField formField = getFormField(fieldInfo);

                    formField.setState(LyFormField.STATE_ERROR);
                    showError(e.getMessage());
                    return null;
                }

                exp = new LyFilterExpression(fieldInfo, condition, parsedValue);

                if (n == 1)
                    return exp;

                if (subFilter == null)
                    subFilter = new LyFilter(LyFilter.Type.OR);

                subFilter.addElement(exp);
            }

            return subFilter;
        }
    } // end method getFilterElement

    /**
     * Factory of the default builder for a controller. Current implementation returns a
     * LyMultiFilterBuilder instance.
     *
     * @return  the builder which creates the XML description of the view
     */
    protected LyBuilder getDefaultBuilder()
    {
        return new LyMultiFilterBuilder(this);
    }

    /**
     * Utility method to know if a fieldInfo is displayed in this form or not.
     */
    public boolean isDisplayed(LyFieldInfo fieldInfo)
    {
        if ((fieldInfo instanceof LyRelationFieldInfo)
                && (((LyRelationFieldInfo)fieldInfo).getType() == LyRelationFieldInfo.Type.COMPOSITION)
                && !hasCheckBoxes())
            return false;
        else
            return super.isDisplayed(fieldInfo);
    }

    /**
     * Utility method to know if a fieldInfo is displayed in this form or not.
     */
    public final boolean hasCheckBoxes()
    {
        return testParameter("_checkBoxes");
    }

	@Override
	public LyFieldController fieldControllerFactory(LyFieldInfo fieldInfo, boolean forceConsult)
	{
		if (fieldInfo instanceof LyRelationFieldInfo)
			return new LyRelationFieldFilterController(this, (LyRelationFieldInfo)fieldInfo);
		return super.fieldControllerFactory(fieldInfo, forceConsult);
	}

	@Override
	public void reset(boolean full)
	{
		super.reset(full);
		updateObject(null);
	}
} // end class LyMultiFilterController
