/**
 * LyMultiPageTableController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyMultiPageTableController.java 34121 2012-02-10 16:56:23Z nhamon $
 */

package leon.control;

import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.setinterface.*;


/**
 * This table controller has the capability of showing data partially, using a page mechanism. This
 * is only a rendering effect when creating the views, since the controller is 'aware' of the entire
 * content of the table.
 *
 * <P>Actions (select, filter, sort ...) apply on the whole content of the view.</P>
 *
 * @see LyTableController
 */
public class LyMultiPageTableController extends LyTableController
{
	/** File Version */
	public static final String __VERSION =
		"$Id: LyMultiPageTableController.java 34121 2012-02-10 16:56:23Z nhamon $";

	/** AUTO SCROLL if create or remove outside of view.
	 Warning : many events may cause scrollings */
	public static boolean	   IMPACT_OUTSIDE_OF_CURRENT_VIEW = true;

	/** Select page string */
	public static final String SELECT_PAGE = "_select_page_".intern(); // must end with '_'

	/** Indicates if pages number must be displayed in select bar. */
	protected boolean		   _displayPagesNumber;

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo class info for this controller
	 * @param context   current context (may be null if the view is not contextual to a list of
	 *                  objects).
	 */
	public LyMultiPageTableController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent, action, classInfo, context);
	}

	/**
	 * Retrieves the view Builder.
	 *
	 * @return the builder instance;
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyMultiPageTableBuilder(this);
	}

	/**
	 * Initialize this instance.
	 */
	protected void init()
	{
	    _currentPage = 0;
        _maxRowCount = -1;
        _displayPagesNumber = true;
	        
		super.init();

		int	   limit = super.getLimit();
		String viewLimitParameter = (String)_action.getParameter("_viewLimit");
		String displayPagesNumberParameter = (String)_action.getParameter("_displayPagesNumber");

		if ((_maxRowCount < 0)
		        && (viewLimitParameter != null)
		        && (limit <= 0))
			_maxRowCount = LyTools.intFromString(viewLimitParameter);

		if ((displayPagesNumberParameter != null)
				&& displayPagesNumberParameter.equalsIgnoreCase("false"))
			_displayPagesNumber = false;
		
		if (!_displayPagesNumber && (limit > 0))
		{
			getEnvironment().logWarning("BF AppEngine : _displayPagesNumber=false and _limit parameters can not be set at the same time. Setting "
					+ "_displayPagesNumber=true on action " + _action.getId());
			_displayPagesNumber = true;
		}
	}

	/**
	 * Indicates if the action has a view limit.
	 *
	 * @return true if the action has a view limit, false otherwise
	 */
	public boolean hasViewLimit()
	{
        return ((_maxRowCount > 0)
            && !(_workSpace instanceof LyPageWorkSpace));
	}
	
	/**
	 * Indicates if there is a pagination on this table controller.
	 * 
	 * @return true if a pagination is used, false otherwise
	 */
	public boolean isPaginated()
	{
		String limitParameter = (String)_action.getParameter("_limit");
		String viewLimitParameter = (String)_action.getParameter("_viewLimit");

		return ((viewLimitParameter != null) || (limitParameter != null));
	}

	/**
	 * Indicates if pages number must be displayed in select bar.
	 *
	 * @return true if pages number must be displayed, false otherwise
	 */
	public boolean isDisplayPagesNumber()
	{
		return _displayPagesNumber;
	}
	
	/**
	 * Indicates whether previous / next tools must be added to the top toolbar.
	 * 
	 * @return true if the _limit parameter is set on this table, false otherwise
	 */
	public boolean isDisplayNextPreviousTools()
	{
		String	displayNextPreviousTools = (String)_action.getParameter("_displayNextPreviousTools", true);
		
		return ("true".equalsIgnoreCase(displayNextPreviousTools));
	}

	/**
	 * Invoked when an action occurs. (LyDemandListener interface) This method implements:
	 *
	 * <UL>
	 *   <LI>default action on object</LI>
	 *   <LI>filters</LI>
	 *   <LI>Find</LI>
	 *   <LI>XRefs</LI>
	 *   <LI>...</LI>
	 * </UL>
	 *
	 * @param demand action event created from the view. User interaction.
	 */
	public void actionDemand(LyActionDemand demand)
	{
		String source = demand.getSource();

		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			if ((source != null) && (source.startsWith(SELECT_PAGE)))
			{
				int		    n = source.lastIndexOf("_");
				LyTable		table = getTableComponent();
				LyComponent component = getComponent();
				LySelectBar selectBar = (LySelectBar)component.getComponent(getId() + "_selectBar");
				String	    parameter = source.substring(n + 1);

				selectBar.setIdle(true);
				//table.setIdle(true);
				// Update the last modification time.
				setLastModified(new Date().getTime());

				if (parameter.equals("first"))
				{
					selectBar.setStartItem(0);
					setPage(0);
					reset(false);
				}
				else if (parameter.equals("previous"))
				{
					int	page = Math.max(0,
							selectBar.getStartItem() - selectBar.getItemCount());
					
					selectBar.setStartItem(page);
					setPage(page);
					reset(false);
				}
				else if (parameter.equals("showprevious"))
				{
					loadPage(false);
					table.setVisible(true);
				}
				else if (parameter.equals("shownext"))
				{
					loadPage(true);
					table.setVisible(true);
				}
				else if (parameter.equals("next"))
				{
					int pos = Math.min(selectBar.getComponentCount() - 1,
							selectBar.getStartItem() + selectBar.getItemCount());
					int lastPos = (selectBar.getComponentCount() / selectBar.getItemCount())
							* selectBar.getItemCount();

					selectBar.setStartItem(Math.min(pos, lastPos));
					setPage(Math.min(pos, lastPos));
					reset(false);
				}
				else if (parameter.equals("last"))
				{
					int pos = (selectBar.getComponentCount() / selectBar.getItemCount())
							* selectBar.getItemCount();

					// If component count is a multiple of item count, do not show nothing : show last pages !
					if (pos == selectBar.getComponentCount())
						pos -= selectBar.getItemCount();

					selectBar.setStartItem(pos);
					setPage(pos);
					reset(false);
				}
				else
				{
					int page = LyTools.intFromString(source.substring(n + 1));

					setPage(page);

					reset(false);
				}

				selectBar.setIdle(false);
				//table.setIdle(false);

				return;
			}
		}

		// Implements the modification of the view limit
		if (((demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
					&& (source != null) && (source.equals(getId() + "_viewLimit")))
				|| ((demand.getType() == LyActionDemand.TYPE_ACTION)
					&& (source != null) && (source.equals(getId() + "_refreshViewLimit")
							|| source.equals(getId() + "_viewLimit"))))
		{
			LyComponent component = getComponent();
			LyFormField formField = (LyFormField)component.getComponent(getId() + "_viewLimit");
			LyNumber    viewLimit = (LyNumber)formField.getFieldItem();
			int		    viewLimitValue = LyTools.intFromString(viewLimit.getLabel());
			int			maxPerView = getEnvironment().getIntResource("LY_MAX_OBJECTS_PER_PAGE");
			
			if (viewLimitValue < 1)
				viewLimitValue = 1;

			if (viewLimitValue > maxPerView)
				viewLimitValue = maxPerView;

			if (_maxRowCount != viewLimitValue)
			{
				_maxRowCount = viewLimitValue;
				setPage(0);
				reset(true);
			}
			else if (getEnvironment().testResource("LY_FORCE_REFRESH_ON_MULTIPAGE_TABLE"))
				update();

			return;
		}

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Indicates whether previous / next page tools may be disabled.
	 * 
	 * @return Default implementation returns false.
	 */
	protected boolean allowDisablePreviousNextTools()
	{
		return false;
	}

	/**
	 * Reset the content of the table being shown. This method rebuilds the XML structure of the
	 * tableLySelectI and manages the page structure of the list.
	 */
	protected void updateContent()
	{
		if (_component == null)
			return;

		if (!hasViewLimit())
		{
			super.updateContent();
			return;
		}

		if (_workSpace == null)
			return;

		synchronized (_workSpace)
		{
			// Add new rows
			int     n = _workSpace.getSize();

			if (n > 0)
			{
				int	    startIdx = _currentPage * _maxRowCount;
				int		endIdx;

				// Check if page is still in boundaries
				if (startIdx >= n)
				{
					// Take last page
					_currentPage = n / _maxRowCount;
					startIdx = _currentPage * _maxRowCount;
				}

				endIdx = Math.min(n, (_currentPage + 1) * _maxRowCount);

				if (_printMode)
				{
					startIdx = 0;
					endIdx = n;
				}

                buildContent(startIdx, endIdx);
			}
            else
                buildContent(0, 0);

			// Update the last modification time.
			if (!_printMode)
				setLastModified(new Date().getTime());
			
			updateTableNRows();
		}
	} // end method updateContent

	/**
	 * This method updates the content of the message bar. Current implementation indicates
	 * selection, filters ...
	 */
	protected void updateMessages()
	{
		super.updateMessages();

		if (_component == null)
			return;

		LyComponent component = getComponent();
		LySelectBar selectBar = (LySelectBar)component.getComponent(getId() + "_selectBar");

		// Set message for objects
		if (selectBar != null)
		{
			LySelectItem item;
			int			 lastPage = getLastPage();

			if (selectBar.getComponentCount() > lastPage)
			{
				selectBar.setIdle(true);

				// Clear additional pages
				for (int i = selectBar.getComponentCount() - 1; i >= lastPage; i--)
				{
					item = (LySelectItem)selectBar.getComponent(i);

					selectBar.removeSelectItem(item);
				}

				// Keep sure we are still in range
				int lastPos = (selectBar.getComponentCount() / selectBar.getItemCount())
						* selectBar.getItemCount();

				if (lastPos < selectBar.getStartItem())
					selectBar.setStartItem(lastPos);

			}
			else if (selectBar.getComponentCount() < lastPage)
			{
				selectBar.setIdle(true);
				// Add additional pages
				for (int i = selectBar.getComponentCount(); i < lastPage; i++)
				{
					item = new LySelectItem(SELECT_PAGE + i, Integer.toString(i + 1));

					selectBar.addSelectItem(item);
				}
			}

			// If pages are not displayed by name, update items labels
			if (!isDisplayPagesNumber())
			{
				String[] labels = getItemLabels();

				if (labels != null)
				{
					selectBar.setIdle(true);

					int i, n = selectBar.getComponentCount();

					for (i = 0; (i < n) && (i < labels.length); i++)
					{
						item = (LySelectItem)selectBar.getComponent(i);
						item.setLabel(labels[i]);
					}
				}
			}

			// Check if current page is still valid
			if (_currentPage >= getLastPage())
			{
				loadPage(false);
			}

			// Ensure current page is selected
			item = (LySelectItem)selectBar.getComponent(_currentPage);
			if ((item != null) && !item.isSelected())
			{
				selectBar.setIdle(true);
				item.setSelected(true);
			}

			// Update the number of objects per page
			LyFormField viewLimit = selectBar.getFormField();

			if (viewLimit != null)
			{
				LyNumber number = (LyNumber)viewLimit.getFieldItem();

				number.setLabel(_maxRowCount + "");
			}

			// End of idle if set above
			selectBar.setIdle(false);
		}
	} // end method updateMessages

	/**
	 * Return the list of labels to be used for select items.
	 *
	 * @return the list of labels to be used for select items
	 */
	protected String[] getItemLabels()
	{
		if (isDisplayPagesNumber())
			return null;
		else
		{
			// Retrieve the first field used to sort the objects
			LySort	    sort = getSort();
			LyFieldInfo fieldInfo = null;
			boolean     usePageNumbers = false;

			if ((sort != null) && (sort.getCriterionCount() > 0))
			{
				LySortCriterion criterion = sort.getCriterion(0);

				fieldInfo = criterion.getFieldInfo();

				if (!(fieldInfo instanceof LyRelationFieldInfo)
						&& !(fieldInfo instanceof LyTextFieldInfo))
					usePageNumbers = true;
			}

			int				   i, n = ((_workSpace == null) ? 0 : _workSpace.getSize());
			LySimpleObjectList list = new LySimpleObjectList();
			LyObject		   object = null;

			list.setSorted(false);

			// Retrieve the list of objects displayed at first positions
			for (i = 0; i < n; i += _maxRowCount)
			{
				object = _workSpace.getObject(i);
				list.addObject(object);
			}

			n = list.getSize();

			String[] labels = new String[n];

			// Build the list of full labels
			for (i = 0; i < n; i++)
			{
				object = list.getObject(i);

				if (usePageNumbers)
					labels[i] = Integer.toString(i + 1);
				else if (fieldInfo == null)
					labels[i] = object.getName();
				else
				{
					Object value = object.getValue(fieldInfo);

					labels[i] = getShortLabel(object, fieldInfo, value);
				}

				if ((labels[i] == null) || (labels[i].length() == 0))
					labels[i] = Integer.toString(i + 1);
			}

			// Shorten the labels with only needed characters
			if (!usePageNumbers)
				computeItemLabels(labels);

			return labels;
		}
	} // end method getItemLabels

	/**
	 * Shorten the given labels to the minimum number of characters sufficient to differenciate
	 * them. For example for labels 'abcdef' and 'abrty' return 'abc' and 'abr'.
	 *
	 * @param labels the list of labels that will be modified
	 */
	protected void computeItemLabels(String[] labels)
	{
		int    i, j, n = labels.length;
		String label1, label2;
		int    length, length1, length2;
		int    pos = 0;

		for (i = 0; i < (n - 1); i++)
		{
			// Retrieve and compare two consecutive labels
			label1 = labels[i];
			label2 = labels[i + 1];
			length1 = label1.length();
			length2 = label2.length();
			length = Math.min(length1, length2);

			for (j = 0; j < length; j++)
			{
				if (label1.charAt(j) != label2.charAt(j))
				{
					pos = Math.max(pos, j);
					break;
				}
			}
		}

		// Now that we have the minimum number of characters, shorten the labels
		for (i = 0; i < n; i++)
		{
			if (labels[i].length() > pos)
				labels[i] = labels[i].substring(0, pos + 1);
		}
	} // end method computeItemLabels

	/**
	 * Gets the selection text for the specified number of items in the view May be overriden to add
	 * additional info (Multipage for instance).
	 *
	 * @param  n number of rows currently being shown
	 * @return the label shown in message bar.
	 */
	protected String getMessageSelectionLabel(int n)
	{
		if (!hasViewLimit())
			return super.getMessageSelectionLabel(n);

		int s = ((_workSpace == null) ? 0 : _workSpace.getSize());

		return getEnvironment().getMessage("LY_OBJECTS_MULTIPAGE_COUNT", new Integer(n),
			new Integer(s));
	}

	/**
	 * Utility method to check if the index is inside the current viewing area. Return the index if
	 * inside the area (returns the index within the area), otherwise returned value is null.<BR>
	 * Return -2 if outside viewing area, and BEFORE current view.<BR>
	 * Return -3 if outside viewing area, and AFTER current view.<BR>
	 *
	 * @param  index the index to check
	 * @return < 0 if out of the viewing area, or a positive index if inside the viewing area : the
	 *         value is the viewing area index.
	 */
	protected int getViewIndex(int index)
	{
		if (index < 0)
			return index;

		int startIndex = _currentPage * _maxRowCount;

		if (index < startIndex)
			return -2;
		else if (index >= (startIndex + _maxRowCount))
			return -3;
		else
			return (index - startIndex);
	}

	/**
	 * Utility method to get the page number from the index of an object in a workspace. Return the
	 * page index,
	 *
	 * @param  index the index to check
	 * @return the page index
	 */
	protected int getPageIndex(int index)
	{
		int value = index / _maxRowCount;

		return value;

	}

	/**
	 * Invoked when an action occurs.
	 *
	 * @param demand selection event created from the view. User interaction.
	 * @see   LyController#selectionDemand
	 */
	public void selectionDemand(LySelectionDemand demand)
	{
		if (!hasViewLimit())
		{
			super.selectionDemand(demand);
			return;
		}

		int selectionType = demand.getType();
		// We only consider extend demands (demand using control + clic)
		// or next page
		if ((!demand.isExtend())
				|| (selectionType == LySelectionDemand.TYPE_SELECT_ALL)
				|| (selectionType == LySelectionDemand.TYPE_UNSELECT_ALL))
		{
			int		 n = demand.getSelectionSize();
			LyObject extendSelectionStart = null;

			if ((n > 1) && (_selections != null) && (_selections.getSize() == 1))
				// This case is a shift + click : maybee we have chaned the page between shift and clic
				extendSelectionStart = _selections.getObject(0);
			super.selectionDemand(demand);
			if ((extendSelectionStart != null)
					&& (_selections.getSize() > 1)
					&& (!_selections.contains(extendSelectionStart)))
			{
				int		 idxOld = _workSpace.getIndex(extendSelectionStart);
				int		 selBegin = _workSpace.getIndex(_selections.getObject(0));
				int		 selLast = _workSpace.getIndex(_selections.getObject(
							_selections.getSize() - 1));
				LyObject object;
				boolean  refresh = false;

				if (idxOld < selBegin)
				{
					// Select all until selbegin
					for (int i = idxOld; i < selBegin; i++)
					{
						object = _workSpace.getObject(i);

						selectObject(object, true, false, false);
					}
					refresh = true;
				}
				else if (idxOld > selLast)
				{
					// Select all after selLast
					for (int i = selLast; i < idxOld; i++)
					{
						object = _workSpace.getObject(i);

						selectObject(object, true, false, false);
					}
					refresh = true;
				}
				if (refresh)
				{
					LyTableBehaviorInterface behavior = (LyTableBehaviorInterface)getBehavior();

					behavior.selectionChanged(this);

					updateContextClient();
					updateTools();
					updateMessages();
				}
			}
		}
		else
		{
			LySimpleObjectList oldSelections = _selections;

			super.selectionDemand(demand);

			// Keep objects out of the view selected, even if they are not
			// in the demand. The graphical component is not aware of these
			// selected objects out of its current scope.

			if (oldSelections != null)
			{
				int		 n = oldSelections.getSize();

				LyObject object;
				int		 i, idx, viewIndex;
				boolean  foundExternal = false;

				for (i = 0; i < n; i++)
				{
					object = oldSelections.getObject(i);

					idx = _workSpace.getIndex(object);
					viewIndex = getViewIndex(idx);

					// We are only concerned about former selected objects
					// OPut of the current scope. That is IN the workspace and OUT of
					// the view
					// objects IN the view use parent method in super.selectionDemand()
					if ((idx >= 0) && (viewIndex < 0))
					{
						if (_selections == null)
							_selections = new LySimpleObjectList();
						_selections.addObject(object);
						foundExternal = true;
					}
				}

				if (foundExternal)
				{
					updateTools();
					updateMessages();
				}
			}
		}
	} // end method selectionDemand

	/**
	 * Select a row in the table for given parameters.
	 * @param fieldInfo the fieldInfo to select
	 * @param pattern the search pattern
	 * @param from the start index for the search
	 * @param next true to find next row, false for previous
	 * @param substring true to find in substring
	 * @param caseSensitive true for case sensitive search 
	 *
	 * @return the index of the selected row, -1 if none has been found
	 */
	public int selectRow(LyFieldInfo fieldInfo, String pattern,
		int from, boolean next, boolean substring, boolean caseSensitive)
	{
		if (!hasViewLimit())
			return super.selectRow(fieldInfo, pattern, from, next, substring, caseSensitive);

		if ((fieldInfo == null) || (pattern == null))
			return -1;

		//FFT00574
		if (from >= 0)
			from = from + (_maxRowCount * _currentPage);

		int idx = super.selectRow(fieldInfo, pattern, from, next, substring, caseSensitive);

		return getViewIndex(idx);
	}

	/**
	 * Selection of a row in a table (graphical component) Internal use of the Controller. This
	 * method only switch the selection status in the view (graphical user interface impact). The
	 * internal selection list (getSelection()) is not set by this method. Use selectObject to
	 * change the selection status of an object:
	 *
	 * @param object       selected or unselected object. May not be null
	 * @param select       new selection status
	 * @param showSelected indicates wether the view should scroll to the new selection status
	 * @see   #selectObject
	 */
	protected void selectTableRow(LyObject object, boolean select, boolean showSelected)
	{
		if (object == null)
			return;

		if (!hasViewLimit())
		{
			super.selectTableRow(object, select, showSelected);
			return;
		}

		LyTable   table = getTableComponent();
		LyRows    rows = table.getRows();

		LySetTableInterface tableInterface = null;

		if (table.getInternalView() instanceof LySetTableInterface)
			tableInterface = (LySetTableInterface)table.getInternalView();

		synchronized (_workSpace)
		{
			int idx = _workSpace.getIndex(object);
			int pageIndex = getPageIndex(idx);

			if (showSelected && (pageIndex != _currentPage))
			{
				setPage(pageIndex);
				reset(false);
			}

            LyRow    row = findRow(rows, object.getId());

			if (row != null)
			{
                int      rowIndex = getRowIndex(object);
                
				row.setSelected(select);
				if (tableInterface != null)
					tableInterface.selectRow(rowIndex, select);

				if (showSelected && select && (tableInterface != null))
					tableInterface.showRow(rowIndex);
			}
		}
	} // end method selectTableRow

	/**
	 * Insert a row representing an object at a given position. It is possible to preselect this row
	 *
	 * @param object   the object to display in the row
	 * @param index    position of the row
	 * @param selected true if the row is selected
	 */
	protected void insertRow(LyObject object, int index, boolean selected)
	{
		if (_component == null)
			return;

		if (!hasViewLimit())
		{
			super.insertRow(object, index, selected);
			return;
		}

		LyTable table = getTableComponent();
		LyRows  rows = table.getRows();
		int     n, viewIndex = getViewIndex(index);

		if (rows == null)
		{
			rows = new LyRows();
			table.setRows(rows);
		}

		synchronized (_workSpace)
		{
			n = _workSpace.getSize();

			if (viewIndex < 0)
			{
				// We only insert the row if within current view OR
				// if append and we are in last page
				if ((index >= 0) || (((_currentPage + 1) * _maxRowCount) < n))
				{
					if (IMPACT_OUTSIDE_OF_CURRENT_VIEW)
						reset(false);
					else
					{
						updateMessages();
						updateFooters();
					}

					return;
				}
				if (IMPACT_OUTSIDE_OF_CURRENT_VIEW)
					reset(false);
			}

			super.insertRow(object, viewIndex, selected);

			synchronized (rows)
			{
				// Remove last if last page is full
				n = rows.getRowCount();

				if (n > _maxRowCount)
				{
					rows.removeRow(n - 1);
					// FFT00333
					reset(false);
				}
			}
		}
	} // end method insertRow

	/**
	 * Update the row corresponding to the object.
	 *
	 * @param object       the object that was modified
	 * @param index        the position of the object
	 * @param statusFields the list of status fields
	 * @return the updated row
	 */
	protected LyRow updateRow(LyObject object, int index, LyInfoList<LyFieldInfo> statusFields)
	{
		if (_component == null)
			return null;

		if (!hasViewLimit())
			return super.updateRow(object, index, statusFields);

		int viewIndex = getViewIndex(index);

		if (viewIndex < 0)
			return null; // Outside viewing area
		else
			return super.updateRow(object, index, statusFields);
	}

	/**
	 * Remove the row that is at the specified position.
	 *
	 * @param object object to remove
	 * @param  index position of the row to remove
	 * @return the removed row
	 */
	protected LyRow removeRow(LyObject object, int index)
	{
		if (_component == null)
			return null;

		if (!hasViewLimit())
			return super.removeRow(object, index);

		int viewIndex = getViewIndex(index);

		if (viewIndex < 0)
		{
			// Check if current page is still valid
			if (_currentPage >= getLastPage())
				loadPage(false);
			else
			{
				if (IMPACT_OUTSIDE_OF_CURRENT_VIEW)
					reset(false);
				else
				{
					updateMessages();
					updateFooters();
				}
			}

			return null; // Outside viewing area
		}
		else
		{
			// Append NEXT line instead (if further lines available)
			LyRow row = super.removeRow(object, viewIndex);
			int   n = _workSpace.getSize();

			if (n >= ((_currentPage + 1) * _maxRowCount))
			{
				int		 idx = ((_currentPage + 1) * _maxRowCount) - 1;
				LyObject nextObject = _workSpace.getObject(idx);
				LyTable  table = getTableComponent();
				LyRows   rows = table.getRows();

				if ((nextObject == object) || (rows.getRow(nextObject.getId()) != null))
					// nextObject may already be in list because objectChanged
					// on LySimpleTableController is implemented by removeRow
					// followed by insertRow, this causes a desynchronization
					// between rows and workspace contains
					return row;

				insertRow(nextObject, idx, false);
			}

			// Check if current page is still valid
			if (_currentPage >= getLastPage())
				loadPage(false);

			return row;
		}
	} // end method removeRow

	/**
	 * Swap two rows.
	 *
	 * @param index    new position
	 * @param oldIndex old position
	 */
	protected void swapRows(int index, int oldIndex)
	{
		if (_component == null)
			return;

		if (!hasViewLimit())
		{
			super.swapRows(index, oldIndex);
			return;
		}

		LyTable table = getTableComponent();
		LyRows  rows = table.getRows();

		synchronized (_workSpace)
		{
			int viewIndex = getViewIndex(index);
			int oldViewIndex = getViewIndex(oldIndex);

			if ((viewIndex < 0) && (oldViewIndex < 0))
			{
				// Nothing to do
			}
			else if ((viewIndex >= 0) && (oldViewIndex >= 0))
				// both are in viewing area
				super.swapRows(viewIndex, oldViewIndex);
			else
			{
				if ((viewIndex < 0) && (oldViewIndex >= 0))
				{
					// One is thrown out of viewing area
					super.removeRow(null, oldViewIndex);

					// Other is inserted from outside (create and insert)
					LyObject			    object = _workSpace.getObject(oldIndex);
					LyInfoList<LyFieldInfo> fields = getFields();
					LyInfoList<LyFieldInfo> statusFields = getStatusFieldInfos();
					LyRow				    row = buildRow(object, fields, statusFields);

					if ((_selections != null) && (_selections.contains(object)))
						row.setSelected(true);

					rows.insertRow(row, oldViewIndex);
				}
				else if ((viewIndex >= 0) && (oldViewIndex < 0))
				{
					// One is thrown out of viewing area
					super.removeRow(null, viewIndex);

					// Other is inserted from outside (create and insert)
					LyObject			    object = _workSpace.getObject(index);

					LyInfoList<LyFieldInfo> fields = getFields();
					LyInfoList<LyFieldInfo> statusFields = getStatusFieldInfos();
					LyRow				    row = buildRow(object, fields, statusFields);

					if ((_selections != null) && (_selections.contains(object)))
						row.setSelected(true);

					rows.insertRow(row, viewIndex);
				}
			}
		}
	} // end method swapRows

	/**
	 * Go to next or previous page.
	 *
	 * @param next true if we go to next page.
	 */
	public void loadPage(boolean next)
	{
//		if (!hasViewLimit())
//		{
//			super.loadPage(next);
//			return;
//		}

		synchronized (_workSpace)
		{
			if (getLastPage() <= 1)
			{
				// FFT00327 : if there is only one page, check if we are at page 0 !
				if (_currentPage != 0)
				{
					_currentPage = 0;
					setPage(_currentPage);
					reset(false);
				}
				return; // Only one single page
			}
			
			if ((_currentPage == 0) && !next)
			{
				LyComponent component = getComponent();
				LySelectBar selectBar = (LySelectBar)component.getComponent(getId() + "_selectBar");
				int			pos = (selectBar.getComponentCount() / selectBar.getItemCount()) * selectBar.getItemCount();

				// If component count is a multiple of item count, do not show nothing : show last pages !
				if (pos == selectBar.getComponentCount())
					pos -= selectBar.getItemCount();

				selectBar.setStartItem(pos);
				setPage(selectBar.getComponentCount() - 1);
			}
			else
			{
				if (next)
					setPage(_currentPage + 1);
				else
					setPage(_currentPage - 1);
			}
			
			if (!update())
				reset(false);
		}
	} // end method loadPage

	/**
	 * Accessor the number (index) of the page being shown.
	 *
	 * @return the current page index, the first page is assumed at index 0.
	 */
	public int getPage()
	{
		return (_currentPage);
	}

	/**
	 * Set current page being displayed If this page is greater than the lastPage value, the method
	 * will take the page 'modulo' getLastPage() value. (Circular).
	 *
	 * @param page current page we want to show
	 */
	public void setPage(int page)
	{
		if (!isInitialized())
			return;

		getViewManager().setWaiting(true);

		int lastPage = getLastPage(); // be careful getLastPage starts at index 1

		// Ensure the page is >= 0
		while (page < 0)
			page += Math.max(1, lastPage);

		// Unselect current
		LyComponent  component = getComponent();
		LySelectBar  selectBar = (LySelectBar)component.getComponent(getId() + "_selectBar");
		LySelectItem item = null;

		if (selectBar != null)
		{
			item = (LySelectItem)selectBar.getComponent(_currentPage);

			if ((item != null) && item.isSelected())
			{
				selectBar.setIdle(true);
				item.setSelected(false);
			}
		}

		// Change page
		_currentPage = page % lastPage;

		if (selectBar != null)
		{
			if (_currentPage == 0)
				selectBar.setStartItem(0);
			else if (_currentPage < selectBar.getStartItem())
				selectBar.setStartItem(Math.max(0,
						selectBar.getStartItem() - selectBar.getItemCount()));
			else if (_currentPage >= (selectBar.getStartItem() + selectBar.getItemCount()))
			{
				int pos = Math.min(selectBar.getComponentCount() - 1,
						selectBar.getStartItem() + selectBar.getItemCount());
				int lastPos = (selectBar.getComponentCount() / selectBar.getItemCount())
						* selectBar.getItemCount();

				selectBar.setStartItem(Math.min(pos, lastPos));
			}

			// Select current
			item = (LySelectItem)selectBar.getComponent(_currentPage);

			if ((item != null) && !item.isSelected())
			{
				selectBar.setIdle(true);
				item.setSelected(true);
			}

			selectBar.setIdle(false);
		}

		if (_workSpace instanceof LyPageWorkSpace)
		{
			LyPageWorkSpace	pageWs = (LyPageWorkSpace)_workSpace;
			boolean			wasIdle = isIdle();
			
			setIdle(true);			
			pageWs.loadPage((short)_currentPage);
			setIdle(wasIdle);
		}

		getViewManager().setWaiting(false);
	} // end method setPage

	/**
	 * Accessor the number of the last available page.
	 *
	 * @return page count of pages, the first page is assumed at index 1.
	 */
	public int getLastPage()
	{
		if (_workSpace instanceof LyPageWorkSpace)
		{
			int	total = getTotalCount();
			
			if (total == 0)
				return 1;
			
			int pageCount = ((total + _maxRowCount - 1) / _maxRowCount);
			
			return pageCount;
		}
		
		int n = ((_workSpace == null) ? 0 : _workSpace.getSize());
		int lastPage;

		if (n == 0)
			return 1;
		if (((n / _maxRowCount) * _maxRowCount) == n)
			lastPage = n / _maxRowCount;
		else
			lastPage = (n / _maxRowCount) + 1; // Partialy fill last page

		return lastPage;
	}

	/**
	 * Accessor to set the last modification date of this controller. The modification date is
	 * updated when anything in the view has changed. The value should be (new Date()).getTime()
	 *
	 * @param time the new modification time
	 */
	public void setLastModified(long time)
	{
		super.setLastModified(time);

		if (_component != null)
		{
			LyTable table = getTableComponent();

			table.setLastModified(time);
		}
	}

	/**
	 * Set the current context for the controller. If the context is unchanged, this method just
	 * call the parent method. If there is a new context, this method defines the new page selected
	 * in this table : the 1st page.
	 *
	 * @param context Current context, or null if no context is set for the view
	 * @return true if successfull
	 */
	public boolean setContext(LyWorkSpace context)
	{
		if ((_action != null)
				&& _action.hasMark(LyAction.MARK_NO_CONTEXT))
			return false;

		if (!hasViewLimit())
			return super.setContext(context);

		if (context != _context)
		{
			// unselect current page : the view will be updated in the reset method
			LyComponent component = getComponent();

			if (component != null)
			{
    			LySelectBar selectBar = (LySelectBar)component.getComponent(getId() + "_selectBar");
    
    			if (selectBar != null)
    			{
    				LySelectItem item = (LySelectItem)selectBar.getComponent(_currentPage);
    
    				if ((item != null) && (item.isSelected()))
    					item.setSelected(false);
    			}
            }
            
			// define the new page
			_currentPage = 0;
		}

		return super.setContext(context);
	}

	/**
	 * Activate or deactivate the filter This method is overriden to reset the page number to 0.
	 *
	 * @param status true if the current filter is active (apply), or is inactive (no filter, but
	 *               filter is kept for possible further use).
	 */
	protected void activateFilter(boolean status)
	{
		setPage(0); // Do not need to call reset here

		super.activateFilter(status);
	}

	/**
	 * Set current sort criteria on the view.
	 *
	 * @param sort new sort criteria
	 */
	public void setSort(LySort sort)
	{
		setPage(0); // Do not need to call reset here
		if (hasViewLimit())
		{
			reset(false);
		}

		super.setSort(sort);
	}

	/**
	 * Accessor on internal classFilter : filter always applied on data shown in view. The class
	 * filter applies wether a user filter is set or not. This is a low level filter always set on
	 * data. This value is usually set in the table action by the '_classFilter' parameter.
	 *
	 * @param filter  current classFilter value, or null if no filter is set (default).
	 * @param reset   indicates wheter the content is immediately calculated and view is refreshed.
	 *                This parameter should be true to recalculate content. This is useless only if
	 *                the content is explicitely reset by the application, using setContext for
	 *                instance.
	 */
	public void setClassFilter(LyFilter filter, boolean reset)
	{
		setPage(0);

		super.setClassFilter(filter, reset);
	}

	/**
	 * (De)Activate print mode. This method is invoked with true before printing the to false at the
	 * end of the print task. The view/controller may need to change the content at that time.
	 *
	 * @param status true if printing is active, false otherwise.
	 */
	public void setPrintMode(boolean status)
	{
        if (status != _printMode)
        {
            super.setPrintMode(status);
            
            if (!hasViewLimit())
                updateContent();
        }
	}

	/**
	 * Updates area controller content.
	 */
	protected void reload()
	{
		if (!hasViewLimit())
			super.reload();

		int currentPage = _currentPage;

		reset();
		setPage(currentPage);
	}

	/**
	 * Reset component.
	 *
	 * @param full Boolean indicating if a full reset (including reset of datas) is requiered or not
	 *             (only reset graphical area).
	 */
	public void reset(boolean full)
	{
		if (full)
		{
			if (_maxRowCount <= 0)
			{
				int	   limit = super.getLimit();
				String viewLimitParameter = (String)_action.getParameter("_viewLimit");

				if ((viewLimitParameter != null) && (limit <= 0))
					_maxRowCount = LyTools.intFromString(viewLimitParameter);
			}

			if (!hasViewLimit())
				setPage(0);
		}

		super.reset(full);
	}
} // end class LyMultiPageTableController
