/*
 * LyOutlineController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyOutlineController.java 34619 2012-03-14 10:10:55Z jlsaouli $
 */
package leon.control;

import java.util.*;

import leon.app.*;
import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.control.form.*;
import leon.data.*;
import leon.info.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.setinterface.*;


/**
 * Controller for the ouline view.
 *
 * @author nla
 */
public class LyOutlineController extends LyFormController
{
	/** File Version */
	public static final String			    __VERSION =
		"$Id: LyOutlineController.java 34619 2012-03-14 10:10:55Z jlsaouli $";

	/** Demand id : Row tool. */
	public static final String			    DEMAND_ROW_TOOL = "rowTool".intern();

	/** List of objects that are outlined. This is usefull for prevent infinite loop. */
	protected LySimpleWorkSpace			    _outlinedObjects;

	/** Association of row-fieldInfo. */
	protected Hashtable<LyRow, LyFieldInfo> _rowsFieldInfoAssociation;

	/** Temporary Vector of selected row. */
	protected Vector<LyRow>				    _tempSelectedRowVector;

	/** Listener for the outline workspace. */
	protected _LyAdvancedWorkSpaceListener  _listener;

	/** Hashtable which contains row - table index association. */
	protected Hashtable<LyRow, Integer>     _rowTableIndexAssoc;

	/** Store the last outlined controller. */
	protected LyAreaController			    _lastOutlinedController;

	/** Class info that can be shown in the outline. */
	protected LyInfoList<LyClassInfo> 		_displayedClasses;
	
	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 * @param action  related action. If an action is specified, the controller uses action
	 *                parameters (action name, action specific XML view description ...)
	 * @param classInfo the associated classInfo
	 * @param context current context (may be null if the view is not contextual to a list of
	 *                objects).
	 */
	public LyOutlineController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent, action, classInfo, context);

		session.addOutlineController(this);

		_tempSelectedRowVector = new Vector<LyRow>();
		_rowTableIndexAssoc = new Hashtable<LyRow, Integer>();
		_listener = new _LyAdvancedWorkSpaceListener();
		_lastOutlinedController = null;
		
		_displayedClasses = getDisplayedClassInfos();
	}


	/**
	 * Reset the entire view : everything is rebuilt.
	 *
	 * @param full Boolean indicating if a full reset (including reset of datas) is requiered or not
	 *             (only reset graphical area).
	 */
	public void reset(boolean full)
	{
		if (_component == null)
			return;
		if (full)
			_fields = null;

		super.reset(full);
		updateContent();
		updateTitleBar();
	}

	/**
	 * @see leon.control.LyAreaController#getContextTitle()
	 */
	public String getContextTitle()
	{
		if (_object == null || !showObject(_object))
			return "";
		
		return super.getContextTitle();
	}
	
	/**
	 * @see leon.control.LyController#getDefaultBuilder()
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return (new LyOutlineBuilder(this));
	}

	/**
	 * Gets list of classes concerned by the action if specified.
	 *
	 * @return Info list containing classes concerned by the action, null if not specified.
	 */
	public LyInfoList<LyClassInfo> getDisplayedClassInfos()
	{
		LyInfoList<LyClassInfo> classes = null;
		LyClassInfo			    classInfo = (LyClassInfo)getParameter("_displayedClass", true);
		int					    np = 1;

		while (classInfo != null)
		{
			if (classes == null)
				classes = new LyInfoList<LyClassInfo>();

			classes.addInfo(classInfo);
			classInfo = (LyClassInfo)getParameter("_displayedClass" + np, true);
			np++;
		}

		return classes;
	} // end method getClassInfos
	
	/**
	 * Method used to update the outline view when the selection changed.
	 *
	 * @param controller The controller where the selection has changed.
	 */
	public void selectionChanged(LyAreaController controller)
	{
		String	notifiedByOtherViewsValue = (String)_action.getParameter("_notifiedByOtherViews", true);
		boolean	notifiedByOtherViews = !"false".equals(notifiedByOtherViewsValue);
		
		if (!notifiedByOtherViews)
		{
			LyViewController	outlineViewController = getViewController();
			LyViewController	controllerViewController = (controller != null ? controller.getViewController() : null);
			
			if ((outlineViewController != null)
					&& (controllerViewController != null)
					&& (controllerViewController != outlineViewController))
				return;
		}
		
		LyObjectList objects = ((controller != null) ? controller.getSelection() : null);

		if (objects != null)
		{
			_lastOutlinedController = controller;
			setContext(objects);
		}
	}

	/**
	 * Return the last outlined controller.
	 *
	 * @return the last outlined controller.
	 */
	public LyAreaController getLastOutlinedController()
	{
		return _lastOutlinedController;
	}

	/**
	 * Set current controller context.
	 *
	 * @param context current context or null if none is specified
	 * @return true if successfull
	 * @see   LyContextInterface#getContext
	 */
	public boolean setContext(LyWorkSpace context)
	{
		if ((context == null)
				|| (context.getSize() == 0))
			return false;

		LyObject object = context.getObject(0);

		if (object == _object)
			return false;

		if (!showObject(object))
			object = null;
		
		_object = object;
		_classInfo = (object == null ? null : object.getClassInfo());

		// As classInfo changes, behavior has to be reconstructed
		setBehavior(null);
		
		return changeContext(context);
	}

	/**
	 * Update the table content. Called when the selection change, or when the object selected is
	 * modified.
	 */
	public void updateContent()
	{
		if (_outlinedObjects != null)
		{
			_outlinedObjects.free(_listener);
			_outlinedObjects = null;
		}

		if (_component == null)
			return;

		LyTable			   table = getTableComponent();
		LyRows			   rows = table.getRows();
		LySetIdleInterface tableInterface = null;

		if (table.getInternalView(false) instanceof LySetIdleInterface)
		{
			tableInterface = (LySetIdleInterface)table.getInternalView();
			if (tableInterface != null)
				tableInterface.setIdle(true);
		}

		// Remove existing rows
		removeRows(rows);
		_rowsFieldInfoAssociation = new Hashtable<LyRow, LyFieldInfo>(1);

		if (_object == null || !showObject(_object))
		{ 
			checkTableRows();

			if (tableInterface != null)
				tableInterface.setIdle(false);
			return;
		}

		_outlinedObjects = new LySimpleWorkSpace(getSession(), true);

		if (rows == null)
		{
			rows = new LyRows();
			table.setRows(rows);
		}

		LyRow			 row;
		LyFieldInfo		 fieldInfo;
		LyClassInfo		 classInfo = getClassInfo();
		int				 i, n = classInfo.getFieldInfoCount();
		LyFieldInfoValue fieldValue;
		Object			 value;

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			if (!isDisplayed(fieldInfo))
				continue;

			value = getFormValue(fieldInfo);
			fieldValue = (LyFieldInfoValue)fieldInfo.buildFieldValue(value);

			row = buildRow(fieldInfo, fieldValue, _object, -1);
			rows.addRow(row);
			_rowsFieldInfoAssociation.put(row, fieldInfo);
		}

		_outlinedObjects.addListener(_listener);
		updateTitleBar();
		checkTableRows();

		if (tableInterface != null)
			tableInterface.setIdle(false);

		// Update the last modification time.
		setLastModified(new Date().getTime());
	} // end method updateContent

	/**
	 * Indicates if the class of the object is currently shown
	 * @param object the object
	 * @return true if currently shown
	 */
	protected boolean showObject(LyObject object)
	{
		if (_displayedClasses == null || _displayedClasses.getSize() == 0)
			return true;
		
		LyClassInfo objectClass = object.getClassInfo();
		
		// If the class is in the list of authorized classes
		if (_displayedClasses.contains(objectClass))
			return true;
		
		return false;
	}


	/**
	 * Checks if at least one row is displayed in the table.<br />
	 * Since there is no header, this has to be done because some viewers are not able to determine
	 * the number of columns when there is no header and no row.
	 */
	protected void checkTableRows()
	{
		if (_component == null)
			return;

		LyTable table = getTableComponent();
		LyRows  rows = table.getRows();

		if ((rows != null) && (rows.getChildCount() > 0))
			return;

		if (rows == null)
		{
			rows = new LyRows();
			table.setRows(rows);
		}

		LyRow  row = new LyRow();
		LyCell cell1 = new LyCell();
		LyCell cell2 = new LyCell();
		LyCell cell3 = new LyCell();

		row.addCell(cell1);
		row.addCell(cell2);
		row.addCell(cell3);
		rows.addRow(row);
	}

	/**
	 * Remove rows of the LyRows component.
	 *
	 * @param rows The LyRow component from which the rows are removed.
	 */
	protected void removeRows(LyRows rows)
	{
		LyRow  row;
		LyRows subRows;
		int    i, n = ((rows == null) ? 0 : rows.getRowCount());

		for (i = n - 1; i >= 0; i--)
		{
			row = rows.getRow(i);
			if ((subRows = row.getRows()) != null)
            {
				removeRows(subRows);
                subRows.free();
            }

			rows.removeRow(i);
            row.free();
		}
	}

	/**
	 * Build a group of row for a given object.
	 *
	 * @param  object The object to outline.
	 * @return The group of row corresponding to the given object.
	 */
	protected LyRows buildRows(LyObject object)
	{
		LyRows				    rows = new LyRows();
		LyInfoList<LyFieldInfo> fields = getFields();
		LyFieldInfo			    fieldInfo;
		LyFieldInfoValue	    value;
		LyRow				    row;
		int					    i, n = fields.getSize();

		_outlinedObjects.addObject(object);

		if (_rowsFieldInfoAssociation == null)
			_rowsFieldInfoAssociation = new Hashtable<LyRow, LyFieldInfo>(1);

		for (i = 0; i < n; i++)
		{
			fieldInfo = fields.getInfo(i);
			value = (LyFieldInfoValue)object.getFieldValue(fieldInfo);

			row = buildRow(fieldInfo, value, object, -1);
			rows.addRow(row);
			_rowsFieldInfoAssociation.put(row, fieldInfo);
		}

		return rows;
	} // end method buildRows

	/**
	 * Build a row of the outline table.
	 *
	 * @param  fieldInfo The fieldInfo which is represented by the row.
	 * @param  fieldValue     The fieldInfo value which is represented by the row.
	 * @param  object    The object outlined.
	 * @param  idx       The index of the value (used for arrays only).
	 * @return The row corresponding to the value of the object for the given fieldInfo.
	 */
	protected LyRow buildRow(LyFieldInfo fieldInfo, LyFieldInfoValue fieldValue,
		LyObject object, int idx)
	{
		LyRow		  row = new LyRow();
		LyCell		  cell;
		LyInfoLabel   infoLabel;
		String		  label;
		String		  valueLabel;
		Object		  value = ((fieldValue == null) ? null : fieldValue.getValue());
		LyEnvironment environment = getEnvironment();
		boolean		  subRow = false;

		row.setId(fieldInfo.getId());

		// Field name
		infoLabel = getInfoLabel(fieldInfo, LyInfoLabel.Type.NAME);
		if (infoLabel != null)
			label = infoLabel.getString(object);
		else
		{
			label = fieldInfo.getAlias();
			label = environment.translate(label);
		}

		if (idx > 0)
			label = label + " " + idx;
		cell = new LyCell();
		cell.setLabel(label);
		row.addCell(cell);

		infoLabel = getInfoLabel(fieldInfo, LyInfoLabel.Type.VALUE);

		// Field value
		if (fieldInfo instanceof LyStructFieldInfo)
		{
			valueLabel = getStructLabel(object, (LyStructFieldInfo)fieldInfo, (LyObjectId)value);
			if ((valueLabel == null) || (valueLabel.equals("")))
				subRow = true;
		}
		else if ((infoLabel == null)
				&& (fieldInfo instanceof LyArrayFieldInfo))
		{
			valueLabel = "";
			subRow = true;
		}
		else
		{
			valueLabel = getLabel(object, fieldInfo, value);
			if (valueLabel != null)
			{
				if (fieldInfo instanceof LyChoiceFieldInfo)
				{
					if (fieldInfo.isMultiple())
					{
						String[] labels = LyString.stringToList(valueLabel,
								LyApplication.VISIBLE_FIELD_SEP);
						int		 i, n = labels.length;

						for (i = 0, valueLabel = ""; i < n; i++)
						{
							if (i > 0)
								valueLabel += ", ";
							valueLabel += environment.translate(labels[i]);
						}
					}
					else
						valueLabel = environment.translate(valueLabel);
				}
				else
				{
					// FFT00158 : pbs when displaying '\n' '\t' in cells (win, swing).
					// We could use the replace method in the viewer, but we have no
					// information about field type and it would be less efficient
					// to parse all the strings in the cells.
					// The problem is that some viewer are able to display multi-line
					// text in tables (web). To be improved...
					if (fieldInfo instanceof LyTextFieldInfo)
					{
						if (((LyTextFieldInfo)fieldInfo).isMultiple()
								|| (((LyTextFieldInfo)fieldInfo).getNRows() > 0))
						{
							valueLabel = LyString.replace(valueLabel, "\r\n", " ");
							valueLabel = valueLabel.replace('\n', ' ');
						}

						valueLabel = valueLabel.replace('\t', ' ');
					}

					valueLabel = environment.translate(valueLabel);
				}
			}
		}

		cell = new LyCell();
		cell.setLabel(valueLabel);
		row.addCell(cell);

		//  Struct
		if (subRow)
		{
			if ((fieldInfo instanceof LyStructFieldInfo)
				&& (((LyRelationFieldInfo)fieldInfo).getType() == LyRelationFieldInfo.Type.COMPOSITION)
				&& (fieldValue != null))
			{
				LyRelationValue relationValue = (LyRelationValue)fieldValue;
				LyObject	    relation = (relationValue.getValueCount() == 0)
					? null : relationValue.getObject(0);

				if ((relation != null)
						&& !(_outlinedObjects.contains(relation)))
				{
					row.setRows(buildRows(relation));
					row.setId(fieldInfo.getId() + "_row");
				}
			}
			else if ((fieldInfo instanceof LyArrayFieldInfo)
					&& (fieldValue != null))
			{
				LyArrayValue	   arrayValue = (LyArrayValue)fieldValue;
				LyFieldInfoValue[] values = arrayValue.getArrayFieldValue();
				LyRows			   rows = new LyRows();
				LyRow			   lineRow;
				LyFieldInfo		   subFieldInfo = ((LyArrayFieldInfo)fieldInfo).getFieldInfo();

				for (int i = 0; (values != null) && (i < values.length); i++)
				{
					lineRow = buildRow(subFieldInfo, values[i], object, i + 1);
					lineRow.setId(fieldInfo.getId() + "_row" + i);
					rows.addRow(lineRow);
					_rowsFieldInfoAssociation.put(lineRow, subFieldInfo);
				}
				row.setRows(rows);
				row.setId(fieldInfo.getId() + "_row");
			}
		}

		LyAction setAction = getSetAction(object, fieldInfo);

		if (setAction != null)
		{
			LyToolBar toolbar = new LyToolBar();
			LyTool    tool;

			toolbar.setController(getId());
			tool = new LyTool(DEMAND_ROW_TOOL + setAction.getId() + ID_SEP + fieldInfo.getId(),
					"...");
			tool.setTooltip("LY_ACTION_SET");
			toolbar.addTool(tool);
			row.addToolBar(toolbar);
		}

		return row;
	} // end method buildRow

	/**
	 * Internal method to compute struct label.
	 *
	 * @param  object    The object owner of the struct.
	 * @param  fieldInfo The struct fieldInfo
	 * @param  value     The value of the fieldInfo
	 * @return The label of the struct.
	 */
	protected String getStructLabel(LyObject object, LyStructFieldInfo fieldInfo, LyObjectId value)
	{
		if (value == null)
			return null;

		LyObject struct = value.getLogicalObject();

		if (struct == null)
			return null;

		LyClassInfo		  classInfo = struct.getClassInfo();
		LyStructFieldInfo subStruct = (LyStructFieldInfo)classInfo.getParent();
		LyUnit			  unit = subStruct.getFieldUnit();
		LyFormat		  format = ((unit != null) ? unit.getFormat() : null);
		String			  label;

		if (format == null)
		{
			// Search for a value label
			LyInfoLabel valueLabel = getInfoLabel(fieldInfo, LyInfoLabel.Type.VALUE);

			if ((valueLabel != null) && (object != null))
				label = getLabel(object, fieldInfo, value);
			else
				label = "";
		}
		else
			label = format.toString(fieldInfo, value, unit);
		return label;
	} // end method getStructLabel

	/**
	 * Accessor to the Table component.
	 * @return the table component
	 */
	public LyTable getTableComponent()
	{
		// Retrieve internal table component
		LyTable table = (LyTable)getComponent().getComponent(getId() + "_table");

		// For file builder
		if (table == null)
			table = (LyTable)getComponent().getComponent("_table");

		if (table == null)
			throw new IllegalStateException("BF AppEngine : can't retrieve table for controller "
				+ this);

		return table;
	}

	/**
	 * Get the row vector for a given row index. The vector contains the tree of row to reach the
	 * needed row.
	 *
	 * @param  index The row index to reach.
	 * @return A vector which contains the tree of row to reach the needed row.
	 */
	protected Vector<LyRow> getRowVector(int index)
	{
		_rowTableIndexAssoc = new Hashtable<LyRow, Integer>();

		LyTable table = getTableComponent();
		LyRows  rows = table.getRows();

		if (rows != null)
			findRow(rows, index);

		Vector<LyRow> rowVector = _tempSelectedRowVector;

		_tempSelectedRowVector = new Vector<LyRow>();
		return rowVector;
	}

	/**
	 * Find a row at the given index
	 * @param rows the table rows
	 * @param index an index
	 * @return the resulting index
	 */
	protected int findRow(LyRows rows, int index)
	{
		for (int i = 0; i < rows.getRowCount(); i++)
		{
			LyRow row = rows.getRow(i);

			if (index == 0)
			{
				_tempSelectedRowVector.add(row);
				_rowTableIndexAssoc.put(row, new Integer(i));
				return -1;
			}

			index--;

			if ((row.getRows() != null) && (row.isOpen()))
			{
				_tempSelectedRowVector.add(row);
				index = findRow(row.getRows(), index);
				if (index == -1)
				{
					_rowTableIndexAssoc.put(row, new Integer(i));
					return -1;
				}

				_tempSelectedRowVector.remove(row);
			}
		}

		return index;
	} // end method findRow

	/**
	 * Invoked when an action occurs. At this level, the implementation manages user interaction on
	 * tree nodes.
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		String source = demand.getSource();

		if ((demand.getType() == LyActionDemand.TYPE_ACTION)
			&& source.equals(_id + "_table"))
		{
			int              index = demand.getIndex(0);
			Vector<LyRow>    rowVector = getRowVector(index);

			invokeOutlineAction(((_context == null) ? null : _object), rowVector);
			return;
		}
		else if (source.startsWith(DEMAND_ROW_TOOL))
		{
			source = source.substring(DEMAND_ROW_TOOL.length());

			String[] parameters = LyString.stringToList(source, ID_SEP);

			if ((parameters != null)
					&& (parameters.length == 2))
			{
				if (_object != null)
				{
					LyClassInfo classInfo = _object.getClassInfo();
					LyAction    action = classInfo.getAction(parameters[0]);
					LyFieldInfo fieldInfo = classInfo.getFieldInfo(parameters[1]);

					if (fieldInfo != null)
						invokeFastEdition(action, _object, fieldInfo, -1, -1);
				}
			}
			return;
		}

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Invoke the outline action for given object on the given rows
	 * @param object an object
	 * @param rowVector a vector of LyRow
	 */
	protected void invokeOutlineAction(LyObject object, Vector<LyRow> rowVector)
	{
		if (object == null)
			return;

		LyClassInfo classInfo = object.getClassInfo();
		LyAction    action = getOutlineAction(classInfo, object);

		if (action == null)
			invokeDefaultAction(object);
		else
		{
			LyController controller = invokeAction(action, object);

			if (controller instanceof LyFormController)
			{
				LyFormController  formController;
				LyRow			  row;
				LyFieldInfo		  fieldInfo;
				LyFieldController fieldController;
				LyAreaController  fieldAreaController;
				LyFormField		  formField;
				LyTabs			  tabs;
				String			  tab;
				LyComponent		  comp;
				LyViewComponent   viewComp;

				while ((rowVector != null)
						&& (!rowVector.isEmpty())
						&& (controller instanceof LyFormController))
				{
					formController = (LyFormController)controller;
					row = rowVector.remove(0);
					fieldInfo = _rowsFieldInfoAssociation.get(row);
					fieldController = formController.getFieldController(fieldInfo);
					fieldAreaController = ((fieldController == null)
							? null : fieldController.getAreaController());
					formField = formController.getFormField(fieldInfo);

					// Select the right tab
					if (formField != null)
					{
						comp = formField.getParent();

						while ((comp != null) && !(comp instanceof LyTab))
							comp = comp.getParent();

						if (comp != null)
						{
							tab = ((LyTab)comp).getLabel();
							tabs = formController.getTabs();
							if (tabs != null)
								tabs.setSelectedTab(tab);
						}
					}

					// Simple field
					if (fieldAreaController == null)
					{
						if ((fieldController instanceof LyArrayFieldSetController)
								&& !rowVector.isEmpty())
						{
							// select the right row
							LyRow			  tableRow = rowVector.remove(0);
							int				  index = _rowTableIndexAssoc.get(tableRow).intValue();
							LyFieldController subFieldController;

							subFieldController = ((LyArrayFieldSetController)fieldController)
									.getFieldController(index);
							formField = subFieldController.getFormField();
							if ((formField != null)
									&& (formField.getInternalView(false) == null))
							{
								// Get controller corresponding to creation line
								subFieldController = ((LyArrayFieldSetController)fieldController)
										.getFieldController(-1);
								if (subFieldController != null)
									formField = subFieldController.getFormField();
							}
						}

						if ((formField != null)
								&& (formField.getInternalView(false) != null))
						{
							// For simple fields, set the focus
							formField.setFocus();
							break;
						}
					}

					// Complex field
					if (fieldAreaController instanceof LyFormController)
					{
						controller = fieldAreaController;
						comp = controller.getComponent();
						if (comp != null)
						{
							viewComp = comp.getInternalView(false);
							if ((viewComp != null)
									&& (viewComp instanceof LySetFocusInterface))
								((LySetFocusInterface)viewComp).setFocus();
						}
						continue;
					}
					else if (fieldAreaController instanceof LyRelationTableController)
					{
						formField.setFocus();

						if (rowVector.isEmpty())
							return;

						LyRelationTableController tableController = (LyRelationTableController)
							fieldAreaController;
						LyRow					  tableRow = rowVector.remove(0);
						int						  index = _rowTableIndexAssoc.get(tableRow)
								.intValue();

						if ((tableController._workSpace == null)
								|| (tableController._workSpace.getSize() <= index))
							return;

						LyObject tableObject = tableController.getObject(index);

						// unselect all
						tableController.selectObject(null, false);

						// select the good row
						tableController.selectObject(tableObject, true);
						controller = tableController.getSetController(tableObject);

						// if row vector is empty, set the caret to the first field.
						if ((rowVector.isEmpty()) && (controller != null))
						{
							LyComponent child;
							int		    i, n;

							comp = controller.getComponent();
							while ((comp != null) && !(comp instanceof LyTable))
							{
								n = comp.getChildCount();
								for (i = 0; i < n; i++)
								{
									child = comp.getChild(i);
									if ((child instanceof LyArea)
											|| (child instanceof LyTable))
									{
										comp = child;
										continue;
									}
								}
								continue;
							}
							if (comp == null)
								continue;

							LyTable tableComp = (LyTable)comp;
							LyRows  rows = tableComp.getRows();
							LyRow   fieldRow = rows.getRow(index);

							n = fieldRow.getChildCount();
							for (i = 0; i < n; i++)
							{
								LyComponent     rowChildComp = fieldRow.getChild(i);
								LyViewComponent rowChildViewComp = rowChildComp.getInternalView(
										false);

								if ((rowChildViewComp != null)
										&& (rowChildViewComp instanceof LySetFocusInterface))
								{
									((LySetFocusInterface)rowChildViewComp).setFocus();
									return;
								}
							}
						}
					}
				}
			}
		}
	} // end method invokeOutlineAction

	/**
	 * Get the outline action for a classInfo and an object
	 * @param classInfo a classInfo
	 * @param object an object
	 * @return the outline action
	 */
	protected LyAction getOutlineAction(LyClassInfo classInfo, LyObject object)
	{
		LyAction		  actionSet = classInfo.getActionWithId(LyAction.ID_SET);
		LyAction		  actionCreate = classInfo.getActionWithId(LyAction.ID_CREATE);
		LyAction		  action = null;
		LySession		  session = getSession();
		LySessionBehavior behavior = (LySessionBehavior)session.getSessionBehavior();
		LyController	  ctrl;
		LyObjectList	  list = new LySimpleObjectList(object);

		if ((object.getPeer() == null)
				&& (session.enableAction(actionCreate, this, list, false)))
		{
			action = actionCreate;
			ctrl = behavior.findExistingController(session, _parent, actionCreate, classInfo, list,
					true);
			if (ctrl == null)
				action = actionSet;
		}
		else
			action = actionSet;

		if ((action == null)
				|| !session.enableAction(action, this, list, false))
		{
			action = classInfo.getActionWithId(LyAction.ID_CONSULT);
			if (!session.enableAction(action, this, list, false))
				action = null;
		}
		return action;
	} // end method getOutlineAction

	/**
	 * Finds the best suited action for fast edition of given field.
	 *
	 * @param object an object
	 * @param  fieldInfo field info to be edited
	 * @return Action found for fast edition of specified field, could be null if not found.
	 */
	public LyAction getSetAction(LyObject object, LyFieldInfo fieldInfo)
	{
		LyOutlineBehaviorInterface behavior = (LyOutlineBehaviorInterface)getBehavior();
		LyAction				   action = behavior.getSetAction(this, fieldInfo);

		return action;
	}

	/**
	 * Invoked when an menu item notification is sent to this listener.
	 *
	 * <P>This method overriden to create a popup menu : buildPopupMenu.</P>
	 *
	 * @param demand demand parameters sent by the view component.
	 */
	public void menuDemand(LyMenuDemand demand)
	{
		if ((demand.getType() == LyMenuDemand.TYPE_EXTEND_POPUP)
				&& getEnvironment().testResource("LY_FAST_EDITION"))
		{
			if ((_context == null)
					|| (_context.getSize() == 0))
				return;

			// An extended popup demand on a table results in the fast edition of an object
			String item = demand.getItem();

			if (_object != null)
			{
				LyClassInfo classInfo = _object.getClassInfo();
				LyFieldInfo fieldInfo = classInfo.getFieldInfo(item);

				if (fieldInfo != null)
				{
					LyAction setAction = getSetAction(_object, fieldInfo);
					int[]    p = demand.getPosition();

					invokeFastEdition(setAction, _object, fieldInfo, p[0], p[1]);
				}
			}

			return;
		}
	} // end method menuDemand

	/**
	 * Invoked when a selection occurs. This method handles: select all, unselect all and objects by
	 * objects selection
	 *
	 * @param demand selection event created from the view. User interaction.
	 */
	public void selectionDemand(LySelectionDemand demand)
	{
		return;
	}

	/**
	 * Indicates if the view for this area acts on the outline view. By default, always return
	 * false.
	 *
	 * @return true if this area acts on the outline view, false otherwise.
	 */
	public boolean isOutlined()
	{
		return false;
	}

	/**
	 * Free this controller.
	 */
	public void free()
	{
		_session.removeOutlineController(this);
		super.free();
	}

	/**
	 * @see leon.control.LyAreaController#buildContextListener()
	 */
	protected _LyContextListener buildContextListener()
	{
		return new _LyOutlineContextListener();
	}
	
	/**
	 * Clear the content of this controller
	 */
	public void empty()
	{
		_object = null;
		reset(true);
	}

	/**
	 * implementation of LyContextListener 
	 *
	 */
	class _LyOutlineContextListener extends _LyContextListener
	{

		/**
		 * Called when an object has changed in the dynamic object list.
		 *
		 * @param  event event of changing object
		 * @return the event is used by the listener
		 */
		public boolean objectChanged(LyWorkSpaceEvent event)
		{
		    LyObject  object = event.getObject();
		    
		    if (object.isDeleted()) 
		        return false; // Ignore if deleted
		    
			updateContent();
			return true;
		}
	}


	class _LyAdvancedWorkSpaceListener
		implements LyGuiListener,
			LyAdvancedWorkSpaceListener
	{

		/**
		 * Called when an object state is changed.
		 *
		 * @param  event Event notifying object state change.
		 * @return True if the event is used by the listener, false otherwise.
		 */
		public boolean objectStateChanged(LyAdvancedWorkSpaceEvent event)
		{
			return false;
		}

		/**
		 * Called when an object user event occured.
		 *
		 * @param  event Event notifying object user event.
		 * @return True if the event is used by the listener, false otherwise.
		 */
		public boolean objectUserEvent(LyAdvancedWorkSpaceEvent event)
		{
			return false;
		}

		/**
		 * Called when grouped events are sent.
		 *
		 * @param  events Sorted list of events.
		 * @return True if the events are used by the listener, false otherwise.
		 */
		public boolean workSpaceEventsGrouped(ArrayList<LyWorkSpaceEvent> events)
		{
			updateContent();
			return true;
		}

		/**
		 * Implementation of LyWorkSpaceListener
		 * 
		 * @param  event the workspace event
		 * @return true if the event was handled.
		 */
		public boolean objectAdded(LyWorkSpaceEvent event)
		{
			updateContent();
			return true;
		}

		/**
		 * Implementation of LyWorkSpaceListener
		 * 
		 * @param  event the workspace event
		 * @return true if the event was handled.
		 */
		public boolean objectChanged(LyWorkSpaceEvent event)
		{
			updateContent();
			return true;
		}

		/**
		 * Implementation of LyWorkSpaceListener
		 * 
		 * @param  event the workspace event
		 * @return true if the event was handled.
		 */
		public boolean objectRemoved(LyWorkSpaceEvent event)
		{
			updateContent();
			return true;
		}

		/**
		 * Checks if this controller is idle.
		 * 
		 * @return the idle status
		 */
		public boolean isIdle()
		{
			return LyOutlineController.this.isIdle();
		}
	} // end class _LyAdvancedWorkSpaceListener
} // end class LyOutlineController
