/**
 * LyPanelController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyPanelController.java 41013 2012-10-03 10:39:30Z jlsaouli $
 */

package leon.control;

import java.util.*;

import leon.app.*;
import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.data.*;
import leon.info.*;
import leon.info.daemon.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.mobile.*;
import leon.view.setinterface.*;


/**
 * This class implements the basics of a table controller. It is used to factorize code between all
 * table controllers. It is associated with a builder (LySimpleTableBuilder) and a behavior that
 * implements LySimpleTableBehaviorInterface. There is no default implementation of this behavior.
 *
 * @see  leon.control.builder.LySimpleTableBuilder
 * @see  leon.app.behaviorinterface.LyTableBehaviorInterface
 * @see  LyTableController
 * @see  LyMultiPageTableController
 */
public abstract class LyPanelController extends LyAreaController
    implements LyQuestionAnswerInterface
{
    /** File Version. */
    public static final String __VERSION = "$Id: LyPanelController.java 41013 2012-10-03 10:39:30Z jlsaouli $";

    /** Menu id : Select all. */
    public static final String DEMAND_SELECT_ALL = "_selectAll".intern();

    /** Menu id : Deselect all. */
    public static final String DEMAND_UNSELECT_ALL = "_unselectAll".intern();

    /** Menu id : save preference. */
    public static final String DEMAND_SAVE_PREFERENCE = "_savePreference".intern();

    /** Menu id : default preference. */
    public static final String DEMAND_DEFAULT_PREFERENCE = "_defaultPreference".intern();
    
    /** Menu id : default preference for mobile view. */
    public static final String DEMAND_DEFAULT_FOR_MOBILE_PREFERENCE = "_defaultForMobilePreference".intern();

    /** Menu id : default preference. */
    public static final String DEMAND_SHARED_PREFERENCE = "_sharedPreference".intern();

    /** Menu id : preference target. */
    public static final String DEMAND_PREFERENCE_TARGET = "_preferenceTarget".intern();

    /** Preference target. */
    public static final String PREFERENCE_TARGET_ACTION = "_preferenceTargetAction".intern();

    /** Preference target. */
    public static final String PREFERENCE_TARGET_CLASS = "_preferenceTargetClass".intern();

    /** Menu id : Fast filter. */
    public static final String FAST_FILTER = "_fastFilter".intern();

    /** Menu id : Filter new. */
    public static final String SIMPLE_FILTER_NEW = "_simpleFilterNew".intern();

    /** Menu id : Extended filter new. */
    public static final String EXTENDED_FILTER_NEW = "_extendedFilterNew".intern();

    /** Menu id : Filter set. */
    public static final String FILTER_SET = "_filterSet".intern();

    /** Menu id : Extend simple filter. */
    public static final String EXTEND_SIMPLE_FILTER = "_extendSimpleFilter".intern();

    /** Menu id : Filter active. */
    public static final String FILTER_ACTIVE = "_filterActive".intern();

    /** Selection list. */
    protected LySimpleObjectList _selections = null;

    /** Last refresh time. */
    protected long _refreshTime;

    /** Intern status to avoid update content twice. */
    private boolean _isFullResetWorking;

    /** Current filter. */
    protected LyFilter _filter;

    /** Indicates whether current filter is active. */
    protected boolean _filterActive;

    /** Border panel controller. */
    protected LyAreaController _borderPanelController;

    /** Filter editor controller. */
    protected LyFilterEditorController _filterEditorController;

    /** Current preference. */
    protected LyPreference _preference;

    /**
     * Constructor.
     *
     * @param  session  current session associated to the controller. The controller automatically
     *                  registers on the specified session.
     * @param  parent   parent controller (or null for 'root' controller). Controller hierarchy
     *                  indicates a child controller (for example a dialog created from a main view)
     * @param  action   related action. If an action is specified, the controller uses action
     *                  parameters (action name, action specific XML view description ...)
     * @param  context  current context (may be null if the view is not contextual to a list of
     *                  objects).
     */
    public LyPanelController(LySession session, LyController parent, LyAction action, LyObjectList context)
    {
        super(session, parent, action, context);

        _refreshTime = -1;
        _isFullResetWorking = false;
        _borderPanelController = null;
        _filterEditorController = null;
        _preference = null;
    }

    /**
     * Init preference on this panel.
     */
    protected void initPreference()
    {
        LyInfoList<LyPreference> prefs = getPreferences();
        int                      i, j, n = ((prefs == null) ? 0 : prefs.getSize());
        LyPreference             preference, userPreference;
        LySession                session = getSession();
    	boolean					 mobile = getViewManager().hasProperty(LyMobileViewManager.MOBILE_ONLY_PROPERTY);

    	if (mobile)
    	{
            for (i = 0; i < n; i++)
            {
                preference = prefs.getInfo(i);
                if (preference.isDefaultForMobile())
                {
                    // Select user preference if it exists
                    for (j = i; j < n; j++)
                    {
                        userPreference = prefs.getInfo(j);
                        if (userPreference.isDefaultForMobile()
                                && (userPreference.getUser() != null)
                                && (session.getUser() != null)
                                && userPreference.getUser().equals(session.getUser().getId()))
                        {
                            preference = userPreference;
                            break;
                        }
                    }

                    _idle = true;
                    setPreference(preference);
                    _idle = false; // Avoid reset if calling setIdle method
                    return;
                }
            }
    	}
    	
        for (i = 0; i < n; i++)
        {
            preference = prefs.getInfo(i);
            if (preference.isDefault())
            {
                // Select user preference if it exists
                for (j = i; j < n; j++)
                {
                    userPreference = prefs.getInfo(j);
                    if (userPreference.isDefault()
                            && (userPreference.getUser() != null)
                            && (session.getUser() != null)
                            && userPreference.getUser().equals(session.getUser().getId()))
                    {
                        preference = userPreference;
                        break;
                    }
                }

                _idle = true;
                setPreference(preference);
                _idle = false; // Avoid reset if calling setIdle method
                return;
            }
        }
    }

    /**
     * Free this controller.
     *
     * @param  controller  parent controller
     */
    public void free(LyController controller)
    {
        if (_parent != controller)
            return;

        if (_borderPanelController != null)
        {
            if (!_borderPanelController.isFreed())
                _borderPanelController.free(this);

            _borderPanelController = null;
        }

        if (_filterEditorController != null)
        {
            if (!_filterEditorController.isFreed())
                _filterEditorController.free();

            _filterEditorController = null;
        }

        super.free(controller);
    } // end method free

    /**
     * Internal method to open filter panel.
     */
    protected void openFilter()
    {
        if (_borderPanelController instanceof LyCompoundFilterController)
        {
            if (_borderPanelController.isFreed())
                closeBorderPanel();
            else
                return; // Already open
        }

        if (!(_parent instanceof LyCompoundAreaController)
                || !(((LyCompoundAreaController)_parent).getAreaController(0) instanceof LyFilterController))
        {
            String openFilter = (String)getParameter("_openFilter", true);

            if (((_filter == null) || !_filterActive)
                    && (openFilter != null)
                    && (openFilter.equalsIgnoreCase("TRUE")
                        || openFilter.equalsIgnoreCase("IF_EMPTY")))
                buildFilter(_filter);
        }
    }

    /**
     * Reset component.
     *
     * @param  full  Boolean indicating if a full reset (including reset of datas) is requiered or
     *               not (only reset graphical area).
     */
    public void reset(boolean full)
    {
        if (_component == null)
            return;

        if (isIdle())
            return;

        LyPanel panel = getPanelComponent();

        synchronized (panel)
        {
            // Do the test, in synchronized code!!!
            if (_isFullResetWorking)
                return;

            // Freeze the panel
            boolean wasIdle = isIdle();
            boolean wasPanelIdle = panel.isIdle();

            // In this case, method reset() is called twice.
            // FFT00051 (Flandres)
            _isFullResetWorking = full && !wasIdle;

            if (full)
                setIdle(true);

            panel.setIdle(true);

            updateContent();
            updateContextClient();
            updateTools();
            updateMessages();
            updateTitleBar();
            updatePreferencesMenu();

            // Update the last modification time.
            setLastModified(new Date().getTime());

            panel.setIdle(wasPanelIdle);
            if (full)
                setIdle(wasIdle);

            _isFullResetWorking = false;
        }

        super.reset(full);
    } // end method reset

    /**
     * Invoked when an action occurs. (LyDemandListener interface) This method implements:
     *
     * <UL>
     * <LI>default action on object</LI>
     * <LI>filters</LI>
     * <LI>Find</LI>
     * <LI>XRefs</LI>
     * <LI>...</LI>
     * </UL>
     *
     * @param  demand  action event created from the view. User interaction.
     */
    public void actionDemand(LyActionDemand demand)
    {
        if (demand.getType() == LyActionDemand.TYPE_ACTION)
        {
            String   source = demand.getSource();
            String[] parameters = LyString.stringToList(source, ID_SEP);

            if ((parameters == null) || (parameters.length == 0) || (parameters[0] == null))
                return;

            String id = parameters[0].intern();

            if (id.equals(getId() + "_filterMessage"))
            {
                if (_filter == null)
                    buildFilter(null);
                else if ("true".equals(demand.getAttribute("LY_PORTLET_DEMAND")))
                    buildFilter(_filter);
                else
                    activateFilter(!(_filterActive));

                return;
            }
            // Filtering menu items
            else if (id == SIMPLE_FILTER_NEW)
            {
                buildFilter(null);
                return;
            }
            else if (id == EXTENDED_FILTER_NEW)
            {
                buildExtendedFilter(null);
                return;
            }
            else if (id == FILTER_SET)
            {
                buildFilter(_filter);
                return;
            }
            else if (id == EXTEND_SIMPLE_FILTER)
            {
                buildExtendedFilter(_filter);
                return;
            }
            else if (id == FILTER_ACTIVE)
            {
                activateFilter((!_filterActive) && (_filter != null));
                return;
            }
            // Find menu items
            else if ((id == LyFindController.DEMAND_FIND)
                    || (id == LyFindController.DEMAND_NEXT)
                    || (id == LyFindController.DEMAND_PREVIOUS))
            {
                findDemand(id);
                return;
            }
            // Selection items
            else if (id == DEMAND_SELECT_ALL)
            {
                selectAllObjects(true);
                return;
            }
            else if (id == DEMAND_UNSELECT_ALL)
            {
                selectAllObjects(false);
                return;
            }
            // save preferences demand
            else if (id == DEMAND_SAVE_PREFERENCE)
            {
                savePreference();
                return;
            }
            // delete preferences demand
            else if (id == DEMAND_DELETE_PREFERENCE)
            {
                deletePreferences();
                return;
            }
            // Actions
            else
            {
                // Menu or action bar
                LyApplication application = getApplication();
                LyInfo        info = application.getInfo(id);

                if (info instanceof LyAction)
                {
                    LyAction action = (LyAction)info;

                    invokeAction(action, demand);
                    return;
                }

                if (info instanceof LyClassInfo)
                {
                    String       actionId = "_xref" + LyAction.ACTION_ID_SEP + id;
                    LyAction     action = application.getAction("_xref");
                    LyAction     classAction = application.getAction(actionId);
                    LyController controller;

                    if (classAction == null)
                        classAction = new LyAction(actionId, action, (LyClassInfo)info, null);

                    if ((_selections == null) || (_selections.getSize() == 0))
                    {
                        LyObjectList content = getContent();

                        controller = new LyXRefAreaController(getSession(), this, classAction, null, content);
                    }
                    else
                        controller = new LyXRefAreaController(getSession(), this, classAction, null, _selections);

                    getSession().showView(controller);
                    return;
                }
            }
        }
        else if (demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
        {
            String source = demand.getSource();

            if ((source != null)
                    && source.equals(getId() + "_preferencesField"))
            {
                String value = demand.getValue();

                restorePreference(value);
            }
        }

        super.actionDemand(demand);
    } // end method actionDemand

    /**
     * Updates the status of tools contained in the toolbar.<br />
     * The state of each tool is computed by calling the enableAction() method on current session
     * with the action corresponding to the tool as a parameter.
     */
    protected void updateTools()
    {
        if (_component == null)
            return;

        // Update ToolBar
        LyToolBar toolBar = (LyToolBar)_component.getComponent("_toolBar");

        if (toolBar == null)
            return;

        LySession            session = getSession();
        LyAction             action = getAction();
        LyInfoList<LyAction> actions = action.getActions();
        LyTool               tool = null;
        boolean              enabled;
        int                  i, n = ((actions == null) ? 0 : actions.getSize());
        LyComponent          comp;

        // Update toolbar and corresponding menu items
        for (i = 0; i < n; i++)
        {
            action = actions.getInfo(i);
            enabled = session.enableAction(action, this, _selections, false);

            // action tool ?
            comp = toolBar.getComponent(action.getId());
            if (comp instanceof LyTool)
            {
                tool = (LyTool)comp;
                tool.setEnabled(enabled);
            }
        }

    } // end method updateTools

    /**
     * Reset the content of the view (Nodes) The current implementation does nothing.
     */
    protected void updateContent()
    {
    }

    /**
     * Get content of the table.
     *
     * @return  the content of the table
     */
    public abstract LyObjectList getContent();

    /**
     * Accessor to the Table component.
     *
     * @return  the panel component
     */
    public abstract LyPanel getPanelComponent();

    /**
     * Accessor to the refresh time of this controller. Used by thin clients to know after a
     * selection whether the controller view must be refreshed or not.
     *
     * @return  A negative number if view must be refreshed or the refresh date.
     */
    public long getRefreshTime()
    {
        if (_refreshTime > 0)
        {
            long oldValue = _refreshTime;

            setRefreshTime(0);
            return oldValue;
        }
        else
            return 0;
    }

    /**
     * Sets the refresh time of this controller.
     *
     * @param  time  the refres time in ms
     */
    public void setRefreshTime(long time)
    {
        _refreshTime = time;
    }

    /**
     * @see  leon.control.LyAreaController#invokeAction(leon.info.LyAction, leon.view.event.LyActionDemand)
     */
    protected LyController invokeAction(LyAction action, LyActionDemand demand)
    {
        String       sourceId = demand.getSource().intern();
        String[]     parameters = LyString.stringToList(sourceId, ID_SEP);
        int          np = parameters.length;
        LyObjectList context = null;
        LyClassInfo  classInfo = getClassInfo();

        if (np > 1)
        {
            LyApplication application = getApplication();
            int           i = 1;

            // Check if first parameter is a class info id
            classInfo = application.getClassInfo(parameters[1]);
            if (classInfo != null)
                i++;

            // Do we have object ids ?
            LyObjectId objectId;
            LyObject   object;

            for (; i < np; i++)
            {
                if (classInfo != null)
                    objectId = new LyObjectId(classInfo, parameters[i]);
                else
                    objectId = new LyObjectId(application, parameters[i]);

                object = objectId.getLogicalObject();
                if (object != null)
                    context = new LySimpleObjectList(object);
                else
                    ((LySimpleObjectList)context).addObject(object);
            }
        }

        if (action.hasMark(LyAction.MARK_NO_CONTEXT))
        	context = null;
        else if (action.getMin() < 0)
        {
            // MIN < 0 : means any selection allowed, use content if nothing selected
            if (action.getMax() == 0)
                context = getContent();
            else
            {
                // take selection, even if any...
                if (context == null)
                    context = getSelection();

                // ... otherwise take content
                if ((context == null) || (context.getSize() == 0))
                    context = getContent();
            }
        }
        // MIN = 0 : means any selection allowed, use empty if nothing selected
        else if ((action.getMin() == 0) && (action.getMax() == 0))
        {
            // MIN = 0 and MAX = 0 action is independant from selection
            LyClassInfo classInfo2 = getClassInfoForAction(action);

            if (classInfo2 == null)
                classInfo2 = classInfo;

            if (classInfo2 != null)
                return invokeAction(action, classInfo2);
            else
                return invokeAction(action, getClassInfo());
        }
        else
        {
            // take selection, even if empty
            if (context == null)
                context = getSelection();

            // Additional test to check if some objects might have disappeared meanwhile
            // Warning : demand.getViewSelection() is ONLY indicative !
            if ((demand.getViewSelection() != null) && (context != null))
            {
                if (context.getSize() < demand.getViewSelection().size())
                {
                    showError(getEnvironment().getMessage("LY_ER_LOST_SELECTION"));
                    return null;
                }
            }
        }

        // Do we have any content for action ?
        if ((context == null)
        		|| (context.getSize() == 0))
            return invokeAction(action, getClassInfo(), context, true);
        else
            return invokeAction(action, context);
    } // end method invokeAction

    /**
     * Internal method to get the classinfo related to the action.
     *
     * @param   action  current action
     *
     * @return  action.getClassInfo but may be overridden (to take into account a class field for
     *          instance)
     */
    protected LyClassInfo getClassInfoForAction(LyAction action)
    {
        return action.getClassInfo();
    }

    /**
     * Sets idle status of the table.
     *
     * @param  idle  a boolean indicating the idle mode
     */
    public void setIdle(boolean idle)
    {
        LyPanel panel = getPanelComponent();

        if (idle && (panel != null))
            panel.setIdle(true);

        super.setIdle(idle);

        if (!idle && (panel != null))
            panel.setIdle(false);
    }

    /**
     * @see  leon.control.LyAreaController#getPopupMenuActions(leon.info.LyClassInfo, boolean)
     */
    protected LyInfoDataList<LyAction, LyClassInfo> getPopupMenuActions(LyClassInfo classInfo, boolean selection)
    {
        LyInfoDataList<LyAction, LyClassInfo> actionsList = super.getPopupMenuActions(classInfo, selection);
        LyBehaviorInterface                   behavior = getBehavior();

        if (behavior instanceof LyPanelBehaviorInterface)
            actionsList = ((LyPanelBehaviorInterface)behavior).getPopupMenuActions(this, actionsList, selection);

        return actionsList;
    } // end method getPopupMenuActions

    /**
     * Build a popup menu item.
     *
     * @param  demand     the menu demand
     * @param  classInfo  a classInfo
     */
    protected void buildPopupMenuItem(LyMenuDemand demand, LyClassInfo classInfo)
    {
    }

    /**
     * Complete contextual menu with the given actions list.
     *
     * @param  popupMenu  the popup menu to complete
     * @param  actions    a list of actions
     */
    protected void completePopupMenu(LyPopupMenu popupMenu, Vector<LyAction> actions)
    {
        // Find all groups
        int            i, n = actions.size();
        LyAction       action;
        String         group;
        Vector<String> groups = new Vector<String>();
        boolean        nullGroup = false;

        for (i = 0; i < n; i++)
        {
            action = actions.elementAt(i);
            group = action.getGroup();
            if ((group == null) && !nullGroup)
            {
                groups.add(null);
                nullGroup = true;
            }
            else if ((group != null) && !groups.contains(group))
                groups.add(group);
        }

        // Find target object, or idx = -1 if popup on 'background'
        LyItem  item;
        boolean enabled;
        int     g, ng = groups.size();
        String  image;

        // Update toolbar and corresponding menu items
        for (g = 0; g < ng; g++)
        {
            group = groups.elementAt(g);
            if (g > 0)
                popupMenu.addSep();

            for (i = 0; i < n; i++)
            {
                action = actions.elementAt(i);
                if (action.getGroup() != group)
                    continue;

                enabled = enableAction(action);

                item = new LyItem(action.getId(), action.getName());
                item.setEnabled(enabled);
                item.setController(getId());
                item.setDialog(action.hasMark(LyAction.MARK_DIALOG));
                image = getMenuItemImage(action);
                item.setImage(image);

                // Add in Popup menu
                popupMenu.addItem(item);
            }
        }
    } // end method completePopupMenu

    /* Implementation of LyWorkSpaceListener */

    /**
     * Called when an object is added in the dynamic object list.
     *
     * @param   event  the workspace event
     *
     * @return  true if the demand was take into account
     *
     * @see     LyWorkSpaceListener#objectAdded
     */
    public boolean objectAdded(LyWorkSpaceEvent event)
    {
        return false;
    }

    /**
     * Called when an object is removed from the dynamic object list.
     *
     * @param   event  the workspace event
     *
     * @return  true if the demand was take into account
     *
     * @see     LyWorkSpaceListener#objectRemoved
     */
    public boolean objectRemoved(LyWorkSpaceEvent event)
    {
        return false;
    }

    /**
     * Called when an object has changed in the dynamic object list.
     *
     * @param   event  the workspace event
     *
     * @return  true if the demand was take into account
     *
     * @see     LyWorkSpaceListener#objectChanged
     */
    public boolean objectChanged(LyWorkSpaceEvent event)
    {
        return false;
    }

    /**
     * Called when an object state is changed.
     *
     * @param   event  Event notifying object state change.
     *
     * @return  True if the event is used by the listener, false otherwise.
     *
     * @see     LyAdvancedWorkSpaceListener#objectStateChanged
     */
    public boolean objectStateChanged(LyAdvancedWorkSpaceEvent event)
    {
        return false;
    }

    /**
     * Called when an object user event occured.
     *
     * @param   event  Event notifying object user event.
     *
     * @return  True if the event is used by the listener, false otherwise.
     *
     * @see     LyAdvancedWorkSpaceListener#objectUserEvent
     */
    public boolean objectUserEvent(LyAdvancedWorkSpaceEvent event)
    {
        return false;
    }

    /**
     * Called when grouped events are sent.
     *
     * @param   events  Sorted list of events.
     *
     * @return  True if the events are used by the listener, false otherwise.
     *
     * @see     LyAdvancedWorkSpaceListener#workSpaceEventsGrouped
     */

    public boolean workSpaceEventsGrouped(ArrayList<LyWorkSpaceEvent> events)
    {
        int                i, n = events.size();
        LyObject           object;
        LySimpleObjectList objects = new LySimpleObjectList();
        LyWorkSpaceEvent   event;

        for (i = 0; i < n; i++)
        {
            event = events.get(i);
            object = event.getObject();
            if (object == null)
                continue;
            if (!objects.contains(object))
                objects.addObject(object);
        }

        // If less than 5 objects are impacted, handle events one by one
        if (objects.getSize() < 5)
        {
            LyWorkSpaceEvent.Type type;

            for (i = 0; i < n; i++)
            {
                event = (LyWorkSpaceEvent)events.get(i);
                type = event.getType();
                switch (type)
                {
                case ADD:
                    objectAdded(event);
                    break;

                case CHANGE:
                    objectChanged(event);
                    break;

                case REMOVE:
                    objectRemoved(event);
                    break;

                case SELECTION_CHANGE:
                case VISIBILITY_CHANGE:
                case ERROR_CHANGE:
                    objectStateChanged((LyAdvancedWorkSpaceEvent)event);
                    break;

                case USER_EVENT:
                    objectUserEvent((LyAdvancedWorkSpaceEvent)event);
                    break;
                    
                default:
                	break;
                }
            }

            return true;
        }

        reset(false);
        return true;
    } // end method workSpaceEventsGrouped

    /**
     * Implementation of find object menu action.
     *
     * @param  demandId  the id of the demand
     */
    protected void findDemand(String demandId)
    {
        if (!(this instanceof LyFindInterface))
            return;

        if (demandId == null)
            return;

        demandId = demandId.intern();

        if (demandId == LyFindController.DEMAND_FIND)
        {
            boolean close = (_borderPanelController instanceof LyFindController)
                    && !_borderPanelController.isFreed();

            closeBorderPanel();
            if (close)
                return;

            LyAction action = getFindAction();

            _borderPanelController = (LyFindController)invokeAction(action, getClassInfo(), null, false);
            _borderPanelController.showView();
        }
        else if (demandId == LyFindController.DEMAND_NEXT)
            ((LyFindInterface)this).findObject((LyFindController)_borderPanelController, -1, true);
        else if (demandId == LyFindController.DEMAND_PREVIOUS)
            ((LyFindInterface)this).findObject((LyFindController)_borderPanelController, -1, false);
    } // end method findDemand

    /**
     * Fetch the find action to open a search item (for example in tables). By default the action id
     * is '__find' but this may be overriden with _findActionId on the action which is being shown
     * (getAction).
     *
     * @return  the find action to open the widget
     */
    protected LyAction getFindAction()
    {
        String   findActionId = null;
        LyAction currentControllerAction = getAction();

        if (currentControllerAction != null)
            findActionId = (String)currentControllerAction.getParameter("_findActionId", true);

        if (findActionId == null)
            findActionId = "__find";

        LyApplication application = getApplication();
        LyAction      action = application.getActionWithId(findActionId, getClassInfo());

        return action;
    }

    /**
     * Update the filter menuitem status.
     */
    protected void updateFilterMenu()
    {
        if (_component == null)
            return;

        // If table is Idle, no need to refresh
        if (isIdle())
            return;

        boolean hasMenu = hasMenu();

        // ** Filtering menu status **
        LyView      view = getView();
        LyComponent comp = null;
        boolean     extendedFilter = ((_filter != null) && (_filter.isExtended()));

        // update filter items
        if (hasMenu && !_action.hasMark(LyAction.MARK_NO_MENUBAR))
        {
            LyMenuBar menuBar = ((view instanceof LyFrame) ? ((LyFrame)view).getMenuBar() : null);

            if (menuBar != null)
            {
                LyItem item = null;

                item = menuBar.findItem(FILTER_SET, getId());
                if (item != null)
                    item.setEnabled(_filter != null);

                item = menuBar.findItem(EXTEND_SIMPLE_FILTER, getId());
                if (item != null)
                    item.setEnabled((_filter != null) && (!extendedFilter));

                item = menuBar.findItem(FILTER_ACTIVE, getId());
                if (item != null)
                {
                    item.setEnabled(_filter != null);
                    item.setChecked(_filterActive);
                }
            }
        }

        if (view != null)
        {
            comp = view.getComponent(FILTER_SET + LyController.ID_SEP + getId());
            if (comp instanceof LyTool)
                ((LyTool)comp).setEnabled(_filter != null);

            comp = view.getComponent(EXTEND_SIMPLE_FILTER + LyController.ID_SEP + getId());
            if (comp instanceof LyTool)
                ((LyTool)comp).setEnabled((_filter != null) && (!extendedFilter));
            
            comp = view.getComponent(FILTER_ACTIVE);
            if (comp instanceof LyTool)
            {
                ((LyTool)comp).setEnabled((_filter != null) && (!extendedFilter));
                if(_filterActive)
                {
                	((LyTool)comp).setLabel("LY_DEACTIVATE_FILTER");
                	((LyTool)comp).setImage("LY_UNACTIVE_FILTER");
                	((LyTool)comp).setTooltip("LY_DEACTIVATE_FILTER");
                }
                else
                {
                	((LyTool)comp).setLabel("LY_ACTIVATE_FILTER");
                	((LyTool)comp).setImage("LY_ACTIVE_FILTER");
                	((LyTool)comp).setTooltip("LY_ACTIVATE_FILTER");
                }
                
            	
            }
        }
    } // end method updateFilterMenu

    /**
     * Update the find menuitem status.
     */
    protected void updateFindMenu()
    {
        if ((_borderPanelController instanceof LyFindController)
                && !_borderPanelController.isFreed())
            ((LyFindController)_borderPanelController).updateFindMenu(this);
    }

    /**
     * Method that indicates if controlled panel enables multi-selection or not.
     *
     * @return  A boolean indicating if multi-selection is allowed or not.
     */
    public boolean enablesMultiSelection()
    {
        return false;
    }

    /**
     * Method that indicates if controlled panel is horizontal.
     *
     * @return  A boolean indicating if the panel component is horizontal.
     */
    public boolean isHorizontal()
    {
        return true;
    }

    /**
     * Select all objects in panel.
     *
     * @param  select  new selection status
     */
    public void selectAllObjects(boolean select)
    {
        if (!select)
        {
            if (_selections == null)
                return;

            LyPanel panel = getPanelComponent();

            panel.setIdle(true);

            LyObject     object;
            LyObjectList selections = new LySimpleObjectList(_selections);
            int          i, n = selections.getSize();

            for (i = 0; i < n; i++)
            {
                object = selections.getObject(i);
                selectObject(object, false);
            }

            panel.setIdle(false);
        }
        else
        {
            LyObjectList content = getContent();

            if (content == null)
                return;

            LyPanel panel = getPanelComponent();

            panel.setIdle(true);

            LyObject object;
            int      i, n = content.getSize();

            for (i = 0; i < n; i++)
            {
                object = content.getObject(i);
                if (!object.isSelected())
                    selectObject(object, true);
            }

            panel.setIdle(false);
        }
    } // end method selectAllObjects

    /**
     * Selected objects. Highlighted in the table.
     *
     * @return  the list of selected items
     */
    public LyObjectList getSelection()
    {
        return ((_selections == null) ? null : new LySimpleObjectList(_selections, true));
    }

    /**
     * Find object in the table and select it.
     *
     * @param   find  the find controller
     * @param   from  index to start from (-1 : begin)
     * @param   next  find commant is searching forward (next selection) or previous.
     *
     * @return  the index of the found object
     */
    public int findObject(LyFindController find, int from, boolean next)
    {
        LyFieldInfo fieldInfo = find.getFieldInfo();
        String      pattern = find.getPattern();
        boolean     caseSensitive = find.isCaseSensitive();
        boolean     wholeWord = find.isWholeWord();

        if ((fieldInfo == null) || (pattern == null))
            return -1;

        LyObjectList content = getContent();
        int          n = content.getSize();

        if (n == 0)
            return -1;

        // Search item
        LyObject object;
        Object   value;
        int      idx;
        String   str;

        if (!caseSensitive)
            pattern = pattern.toLowerCase();

        if ((from < 0) || (from >= n))
        {
            // Retrieve first selection
            if ((_selections != null)
                    && (_selections.getSize() > 0))
            {
                LyObject first = _selections.getObject(0);

                from = content.getIndex(first);
            }
        }

        boolean wrap = (from >= 0) && (from < n);

        if (next)
            idx = ((!wrap || (from == (n - 1))) ? 0 : (from + 1));
        else
            idx = ((!wrap || (from == 0)) ? (n - 1) : (from - 1));

        LyClassInfo classInfo = find.getClassInfo();

        while ((idx >= 0)
                && (idx != from)
                && (idx < n))
        {
            object = content.getObject(idx);
            if (object.instanceOf(classInfo))
            {
                value = object.getValue(fieldInfo);
                str = fieldInfo.toName(value);

                if (str == null)
                    continue;

                if (!caseSensitive)
                    str = str.toLowerCase();

                if ((!wholeWord && (str.indexOf(pattern) >= 0))
                        || (wholeWord && str.equals(pattern)))
                    break;
            }

            if (next)
                idx = ((wrap && (idx == (n - 1))) ? 0 : (idx + 1));
            else
                idx = ((wrap && (idx == 0)) ? (n - 1) : (idx - 1));
        }

        selectAllObjects(false);

        if ((idx >= 0) && (idx < n))
        {
            LyObject selectedObject = content.getObject(idx);

            selectObject(selectedObject, true);
            return idx;
        }
        else
            // Not found
            return -1;
    } // end method findObject

    /**
     * Select an object in the table.
     *
     * @param  object  selected or unselected object
     * @param  select  new selection status
     */
    public void selectObject(LyObject object, boolean select)
    {
        object.setSelected(select);
    }

    /**
     * Set or create a new Filter.
     *
     * @param  currentFilter  current filter or null if a new Filter
     */
    public void buildFilter(LyFilter currentFilter)
    {
        // Create a new filter controller
        boolean close = (_borderPanelController instanceof LyCompoundFilterController)
                && !_borderPanelController.isFreed();

        closeBorderPanel();
        if (close)
            return;

        getViewManager().setWaiting(true);

        String        actionId;
        LyEnvironment environment = _parent.getEnvironment();

        if (!environment.testResource("LY_EXTENDED_FILTER")
        	|| environment.getViewManager().hasProperty(LyMobileViewManager.MOBILE_ONLY_PROPERTY))
        {
            if (currentFilter != null)
            {
                if (currentFilter.isExtended())
                {
                    buildExtendedFilter(currentFilter);
                    getViewManager().setWaiting(false);
                    return;
                }

                actionId = "_filter_set";
            }
            else
            {
                if (_filterActive)
                    activateFilter(false);

                actionId = "_filter_new";
            }
        }
        else
        {
            if (currentFilter != null)
                actionId = "__compoundFilterSet";
            else
            {
                if (_filterActive)
                    activateFilter(false);

                actionId = "__compoundFilterNew";
            }
        }

        LySession     session = getSession();
        LyApplication application = session.getApplication();
        LyAction      action = application.getActionWithId(actionId, getClassInfo());

        _borderPanelController = (LyAreaController)invokeAction(action, getClassInfo());

        getViewManager().setWaiting(false);
    } // end method buildFilter

    /**
     * Set or create a new Filter.
     *
     * @param  currentFilter  current filter or null if a new Filter
     */
    protected void buildExtendedFilter(LyFilter currentFilter)
    {
        String    actionId = "__filter_extended";
        LySession session = getSession();

        getViewManager().setWaiting(true);

        if (currentFilter != null)
            actionId = actionId + "_set";
        else
        {
            activateFilter(false);
            actionId = actionId + "_new";
        }

        if ((_filterEditorController != null)
                && !_filterEditorController.isFreed())
        {
            _filterEditorController.showView();
            getViewManager().setWaiting(false);
            return;
        }

        LyApplication application = session.getApplication();
        LyAction      action = application.getActionWithId(actionId, getClassInfo());

        _filterEditorController = (LyFilterEditorController)invokeAction(action, getClassInfo());

        getViewManager().setWaiting(false);
    } // end method buildExtendedFilter

    /**
     * The filter edition form is closed : notify registered listener.
     *
     * @param  controller    reference to the filter controller (filter form which is closing)
     * @param  actionDemand  actionDemand on filter controller : id of the command such as
     *                       LyFormController.DEMAND_VALIDATE, LyFormController.DEMAND_APPLY,
     *                       LyFormController.DEMAND_CANCEL
     * @param  filter        the filter that was created in the form
     */
    public void endFilter(LyController controller, String actionDemand, LyFilter filter)
    {
        if ((filter != null)
                || !actionDemand.equals(DEMAND_CANCEL))
        {
            if ((filter == null)
                    || (filter.getSize() != 0))
            {
                if (confirmNotPhysicalFilter(filter))
                {
                    LyViewManager vm = getViewManager();

                    vm.askQuestion(_component, "LY_CONFIRM_NOT_PHYSICAL_FILTER", this, filter);
                }
                else
                    setFilter(filter);
            }
            else
            {
                // An empty filter is considered as a removal of the current filter
                if (_filterActive)
                    activateFilter(false);
            }
        }
        else
            // Update the last modification time : we must change this time
            // for the next call.
            setLastModified(new Date().getTime());
    }
    
    /**
     * Indicates whether user must confirm the usage of given filter.
     * 
     * @param filter filter to be checked
     * @return true if user must confirm filter, false otherwise
     */
    protected boolean confirmNotPhysicalFilter(LyFilter filter)
    {
    	return (getContent() instanceof LyPageWorkSpace) && !checkFilterForPhysicalLayer(filter);
    }

    /**
     * Apply the filter to the view This method sets the specified filter as the default filter then
     * invokes activateFilter.
     *
     * @param  newFilter  current filter
     *
     * @see    #activateFilter
     */
    public void setFilter(LyFilter newFilter)
    {
        LyPanelBehaviorInterface behavior = (LyPanelBehaviorInterface)getBehavior();

        newFilter = behavior.validateFilter(this, newFilter);

        _filter = newFilter;
        activateFilter(_filter != null);
    }

    /**
     * Activate or deactivate the filter.
     *
     * @param  status  true if the current filter is active (apply), or is inactive (no filter, but
     *                 filter is kept for possible further use).
     */
    protected void activateFilter(boolean status)
    {
        _filterActive = status;

        if (isIdle())
            return;

        getViewManager().setWaiting(true);
        reset(true);
        getViewManager().setWaiting(false);
    }

    /**
     * Gets initial : indicates to the filter controller the default filter parameters.
     *
     * @return  the initial filter or null if no filter is set
     */
    public LyFilter getFilter()
    {
        return _filter;
    }

    /**
     * Indicates whether the filter is active.
     *
     * @return  true if a filter is active
     */
    public boolean isFilterActive()
    {
        return _filterActive;
    }

    /**
     * Builds the filter message description.
     *
     * @return  the component for the objects message description
     */
    protected LyMessage buildFilterMessage()
    {
        LyEnvironment environment = getEnvironment();
        String        displayString = environment.translate("LY_MENU_FILTER");

        if ((displayString == null)
                || (displayString.trim().length() == 0))
            return null;

        LyMessage filterMessage = new LyMessage(this.getId() + "_filterMessage", " ");

        displayString = environment.translate("LY_ITEM_FILTER_ACTIVE");
        if ((displayString != null) && (displayString.trim().length() > 0))
            filterMessage.setController(this.getId());

        filterMessage.setSize(10);
        return filterMessage;
    }

    /**
     * This method updates the content of the message bar. Current implementation indicates
     * selection, filters ...
     */
    protected void updateMessages()
    {
        if (_component == null)
            return;

        // If table is Idle, no need to refresh
        if (isIdle())
            return;

        LyView view = getView();

        if (!(view instanceof LyFrame))
            return;

        LyMessages messages = ((LyFrame)view).getMessages();

        if (messages == null)
            return;

        LyMessage     filterMessage = ((view == null) ? null
                : (LyMessage)messages.getComponent(getId() + "_filterMessage"));
        LyEnvironment environment = getEnvironment();

        if (filterMessage != null)
        {
            StringBuffer filterLabel;

            if (_filter == null)
            {
                filterMessage.setImage("LY_UNACTIVE_FILTER");
                filterLabel = new StringBuffer(environment.getMessage("LY_NO_FILTER"));
            }
            else if (isFilterActive())
            {
                filterMessage.setImage("LY_ACTIVE_FILTER");
                filterLabel = new StringBuffer(environment.getMessage("LY_ACTIVE_FILTER"));
            }
            else
            {
                filterMessage.setImage("LY_UNACTIVE_FILTER");
                filterLabel = new StringBuffer(environment.getMessage("LY_UNACTIVE_FILTER"));
            }

            filterMessage.setLabel(filterLabel.toString());
        }
    } // end method updateMessages

    /**
     * Check if the table has a menubar to update.
     *
     * @return  check status on the view manager, but may be overriden
     */
    protected boolean hasMenu()
    {
        LyViewManager viewManager = getViewManager();
        boolean       hasMenu = viewManager.hasProperty("MENU");

        return hasMenu;
    }

    /**
     * Saving a preference.
     */
    protected void savePreference()
    {
        boolean close = ((_borderPanelController instanceof LyUserInputController)
                    && _borderPanelController.getId().startsWith("__savePreference")
                    && !_borderPanelController.isFreed());

        closeBorderPanel();
        if (close)
            return;

        // Create a dialog to type in the search parameters
        LySession             session = getSession();
        LyApplication         application = session.getApplication();
        LyAction              action = application.getActionWithId("__savePreference", getClassInfo());
        String                title = null;
        LyEnvironment         environment = getEnvironment();
        LyUserInputController savePrefCtrl;

        if (_preference != null)
            title = environment.translate(_preference.getName());

        savePrefCtrl = new LyUserInputController(session, this, action);
		
		_LySavePreferenceBehavior	behavior = new _LySavePreferenceBehavior(savePrefCtrl);
		
		savePrefCtrl.setBehavior(behavior);
		behavior.init(savePrefCtrl);

        // FFT 000947
        savePrefCtrl.setBuilder(new _LyObjectInfosInputBuilder(savePrefCtrl));

        savePrefCtrl.setTitleBar("LY_ACTION_CREATE" + " : " + "LY_ACTION_PREFERENCE");
        savePrefCtrl.setImage(getImage(_action));

        savePrefCtrl.addText(savePrefCtrl.getId() + DEMAND_SAVE_PREFERENCE, "LY_PREFERENCE_NAME", title, 40, null,
            false);

        String str = environment.getString("LY_DEFAULT_PREFERENCE");

        if ((str != null) && (str.length() > 0))
            savePrefCtrl.addBoolean(savePrefCtrl.getId() + DEMAND_DEFAULT_PREFERENCE, "LY_DEFAULT_PREFERENCE",
                (((_preference != null) && _preference.isDefault()) ? LyBoolean.Status.CHECKED
                    : LyBoolean.Status.UNCHECKED), false);

        str = environment.getString("LY_DEFAULT_FOR_MOBILE_PREFERENCE");
        if ((str != null) && (str.length() > 0))
            savePrefCtrl.addBoolean(savePrefCtrl.getId() + DEMAND_DEFAULT_FOR_MOBILE_PREFERENCE,
            	"LY_DEFAULT_FOR_MOBILE_PREFERENCE",
                (((_preference != null) && _preference.isDefaultForMobile()) ?
                	LyBoolean.Status.CHECKED : LyBoolean.Status.UNCHECKED), false);
        
        if (session.getUser() != null)
        {
            str = environment.getString("LY_SHARED_PREFERENCE");
            if ((str != null) && (str.length() > 0))
                savePrefCtrl.addBoolean(savePrefCtrl.getId() + DEMAND_SHARED_PREFERENCE, "LY_SHARED_PREFERENCE",
                    (((_preference != null) && (_preference.getUser() == null)) ? LyBoolean.Status.CHECKED
                        : LyBoolean.Status.UNCHECKED), false);
        }

        if (_action.getClassInfo() != null)
        {
            str = environment.getString("LY_PREFERENCE_TARGET");
            if ((str != null) && (str.length() > 0))
            {
                LyChoice choice = savePrefCtrl.addChoice(savePrefCtrl.getId() + DEMAND_PREFERENCE_TARGET,
                        "LY_PREFERENCE_TARGET", false, false);
                LyOption option;
                boolean  targetAction = (_preference == null)
                        || !(_preference.getObjectInfo() instanceof LyClassInfo);

                option = new LyOption(PREFERENCE_TARGET_ACTION, "LY_PREFERENCE_TARGET_ACTION");
                option.setChecked(targetAction);
                choice.addOption(option);

                option = new LyOption(PREFERENCE_TARGET_CLASS, "LY_PREFERENCE_TARGET_CLASS");
                option.setChecked(!targetAction);
                choice.addOption(option);
            }
        }

        savePrefCtrl.showView();
        _borderPanelController = savePrefCtrl;

        // Waiting for userInput event.
    } // end method savePreference

    /**
     * Deleting preferences.
     */
    protected void deletePreferences()
    {
        boolean close = ((_borderPanelController instanceof LyUserInputController)
                    && _borderPanelController.getId().startsWith("__deletePreference")
                    && !_borderPanelController.isFreed());

        closeBorderPanel();
        if (close)
            return;

        // Create a dialog to type in the search parameters
        LySession             session = getSession();
        LyApplication         application = session.getApplication();
        LyAction              action = application.getActionWithId("__deletePreference", getClassInfo());
        LyUserInputController deletePrefCtrl;

        deletePrefCtrl = new LyUserInputController(session, this, action);
        
        _LyDeleteObjectInfosBehavior	behavior = new _LyDeleteObjectInfosBehavior();
        
        deletePrefCtrl.setBehavior(behavior);
        behavior.init(deletePrefCtrl);
        deletePrefCtrl.setBuilder(new _LyObjectInfosInputBuilder(deletePrefCtrl));
        deletePrefCtrl.setTitleBar("LY_ACTION_DELETE" + " : " + "LY_ACTION_PREFERENCE");
        deletePrefCtrl.setImage(getImage(_action));

        LyInfoList<LyPreference> prefs = getPreferences();
        LyPreference             preference;
        int                      i, n = ((prefs == null) ? 0 : prefs.getSize());
        LyOption                 option;
        LyChoice                 choice = deletePrefCtrl.addChoice(DEMAND_DELETE_PREFERENCE, "LY_PREFERENCE_DELETE",
                true, false);

        choice.setFill(LyChoice.Fill.VERTICAL);
        choice.setMaxSpan(6);

        if (n == 0)
            return;

        for (i = 0; i < n; i++)
        {
            preference = prefs.getInfo(i);
            if (preference.isReadOnly())
                continue;

            option = new LyOption(preference.getId(), preference.getName());
            choice.addOption(option);
        }

        deletePrefCtrl.showView();
        _borderPanelController = deletePrefCtrl;

        // Waiting for userInput event.
    } // end method deletePreferences

    /**
     * Deleting preferences files.
     *
     * @param  prefs  the identifiers of the preferences
     */
    protected void deletePreferencesFile(String prefs)
    {
        LySession     session = getSession();
        LyApplication application = session.getApplication();
        String[]      ids = LyString.stringToList(prefs, LyTools.OBJECT_SEP);
        String        prefId;
        LyPreference  preference;
        int           i, n = ids.length;
        LyArea        area = (LyArea)getComponent();
        LyTitleBar    titleBar = area.getTitleBar();
        LyToolBar     rightToolBar;
        LyFormField   prefsField = null;

        if (titleBar != null)
        {
            rightToolBar = titleBar.getToolBar();
            if (rightToolBar != null)
                prefsField = (LyFormField)rightToolBar.getComponent(getId() + "_preferencesField");
        }

        if (prefsField == null)
        {
            rightToolBar = area.getTopRightToolBar();
            if (rightToolBar != null)
                prefsField = (LyFormField)rightToolBar.getComponent(getId() + "_preferencesField");
        }

        boolean clearPref = false;

        if (prefsField != null)
        {
            LyRelation relation = (LyRelation)prefsField.getFieldItem();
            LyLink     link;

            for (i = 0; i < n; i++)
            {
                prefId = ids[i];
                preference = (LyPreference)application.getInfo(prefId);

                if (preference != null)
                {
                    // Reset current preference if deleted
                    if (preference == _preference)
                        clearPref = true;

                    if (application.deleteExtraInfo(preference))
                    {
                        link = (LyLink)relation.getComponent(prefId);
                        relation.removeLink(link);
                    }
                }
            }
        }

        // Update preferences menu
        updatePreferencesMenu();

        if (clearPref)
            setPreference(null);
    } // end method deletePreferencesFile

    /**
     * Restoring preference.
     *
     * @param  id  identifier of the preference.
     */
    protected void restorePreference(String id)
    {
        LySession     session = getSession();
        LyApplication application = session.getApplication();
        LyPreference  preference = (LyPreference)application.getInfo(id);

        setPreference(preference);
    }

    /**
     * Sets the preference for this controller.
     *
     * @param  preference  the preference to apply
     */
    public void setPreference(LyPreference preference)
    {
        if (preference == _preference)
            return;

        // Update filter
        LyFilter filter = ((preference == null) ? null : preference.getFilter());

        if (filter != null)
            setFilter(filter.cloneFilter());
        else
        {
            filter = (LyFilter)getParameter("_filter", true);
            if (filter != null)
                filter = filter.cloneFilter();

            setFilter(filter);
        }

        LyInfoList<LyPreference>	preferences = getPreferences();
        boolean						visible = ((preferences != null) && preferences.contains(preference));
        		
        if (visible)
        	_preference = preference;
        else
        	_preference = null;

        // Reset
        if (!isIdle())
        {
            reset(false);
            
            if (visible)
            	updatePreferencesMenu();
        }
    } // end method setPreference

    /**
     * Gets current preference.
     *
     * @return  the current preference
     */
    public LyPreference getPreference()
    {
        return _preference;
    }

    class _LySavePreferenceBehavior extends LyUserInputBehavior
        implements LyQuestionAnswerInterface
    {
        LyUserInputController _controller;
        String                _name;
        boolean               _default;
        boolean               _defaultForMobile;
        boolean               _shared;
        boolean               _targetAction;

        public _LySavePreferenceBehavior(LyUserInputController controller)
        {
            super();

            _controller = controller;
            _name = null;
            _default = false;
            _defaultForMobile = false;
            _shared = false;
            _targetAction = true;
        }

        /**
         * Invoked when the user validates the content of the input.
         *
         * @param  controller  the current LyUserInputController instance
         * @param  demand      demand parameters sent by the view component.
         */
        public void userInputDemand(LyController controller, LyDemand demand)
        {
            // Close find dialog
            String id = demand.getSource();

            if (id.equals(LyFormController.DEMAND_VALIDATE))
            {
                LyParameterValues paramValues = demand.getParameterValues();

                if (paramValues != null)
                {
                    String defaultPref = paramValues.getParameter(controller.getId() + DEMAND_DEFAULT_PREFERENCE);
                    String defaultForMobilePref = paramValues.getParameter(controller.getId() + DEMAND_DEFAULT_FOR_MOBILE_PREFERENCE);
                    String sharedPref = paramValues.getParameter(controller.getId() + DEMAND_SHARED_PREFERENCE);
                    String targetAction = paramValues.getParameter(controller.getId() + DEMAND_PREFERENCE_TARGET);

                    _name = paramValues.getParameter(controller.getId() + DEMAND_SAVE_PREFERENCE);
                    _default = (defaultPref != null)
                            && defaultPref.equals(LyBoolean.Status.CHECKED.toString());
                    _defaultForMobile = (defaultForMobilePref != null)
                    	&& defaultForMobilePref.equals(LyBoolean.Status.CHECKED.toString());
                    _shared = (sharedPref != null)
                            && sharedPref.equals(LyBoolean.Status.CHECKED.toString());
                    _targetAction = (targetAction == null)
                            || targetAction.equals(PREFERENCE_TARGET_ACTION);

                    if ((_name == null) || (_name.length() == 0))
                    {
                        String        message = getEnvironment().getMessage("LY_ER_NOTNULL", "LY_PREFERENCE_NAME");
                        LyViewManager viewManager = getViewManager();
                        LyComponent   root = getBuilder().getComponent();

                        viewManager.showError(root, message);
                        return;
                    }

                    // Find start
                    writePreference();
                }
            }
            else
                super.userInputDemand(controller, demand);
        } // end method userInputDemand

        /**
         * Writing a preference.
         */
        protected void writePreference()
        {
            // Search if a preference with the given name already exists
            LyPreference             preference = null;
            LyInfoList<LyPreference> allPrefs = getPreferences(false, false);
            LyInfoList<LyPreference> currentUserPrefs = getPreferences(false, true);
            LyInfoList<LyPreference> currentUserAndActionPrefs = getPreferences(true, true);
            LyInfoList<LyPreference> currentActionPrefs = getPreferences(true, false);
            LyPreference             oldPreference = null;
            int                      j, n = ((allPrefs == null) ? 0 : allPrefs.getSize());
            String                   name;
            LyEnvironment            environment = getEnvironment();
            boolean					 otherUserExistingPref = false, otherViewExistingPref = false;
            LyObject                 user = getSession().getUser();
            String                   currentUserId = ((user == null) ? "" : user.getId());
            String					 allPreference = environment.translate("LY_ALL");

            if ((allPreference != null)
            	&& allPreference.equals(_name))
            {
                String  message = environment.getMessage("LY_ER_ALL_PREFERENCE_RESERVED", _name);

                getViewManager().showError(getView(), message);
                return;
            }
            
            for (j = 0; j < n; j++)
            {
                preference = allPrefs.getInfo(j);
                name = environment.translate(preference.getName());
                
                if (name.equals(_name))
                {
                	// If preference is shared, check if another preference already exists for another user
                	if (_shared
            			&& (preference.getUser() != null)
            			&& !currentUserId.equals(preference.getUser())
            			&& (currentActionPrefs != null)
            			&& currentActionPrefs.contains(preference))
                	{
                		otherUserExistingPref = true;
	                    oldPreference = preference;
	                    break;
	                }
	            	// If target of preference is not current action, check if another preference already exists
	            	// for another view with the same class info
	            	else if (!_targetAction
	        			&& (preference.getObjectInfo() instanceof LyAction)
	        			&& (((LyAction)preference.getObjectInfo()).getClassInfo() == _action.getClassInfo())
	        			&& (currentUserPrefs != null)
	        			&& currentUserPrefs.contains(preference))
	            	{
	            		otherViewExistingPref = true;
	                    oldPreference = preference;
	            		break;
	            	}
	            	else if ((currentUserAndActionPrefs != null)
	            			&& currentUserAndActionPrefs.contains(preference))
	            	{
	                    oldPreference = preference;
	            		break;
	            	}
	            }
	        }

            if (oldPreference != null)
            {
                if (oldPreference.isReadOnly())
                {
                    String        message = environment.getMessage("LY_ER_PREFERENCE_READONLY", _name);
                    LyViewManager viewManager = getViewManager();
                    LyComponent   root = getBuilder().getComponent();

                    viewManager.showError(root, message);
                    return;
                }

                if (otherUserExistingPref)
                	getViewManager().showWarning(getView(),
                		environment.getMessage("LY_ER_PREFERENCE_ALREADY_EXISTS_FOR_USER", oldPreference.getName()));
                else if (otherViewExistingPref)
                	getViewManager().showWarning(getView(),
                		environment.getMessage("LY_ER_PREFERENCE_ALREADY_EXISTS_FOR_VIEW", oldPreference.getName()));
                else
                getViewManager().askQuestion(getView(),
                    environment.getMessage("LY_CONFIRM_PREFERENCE_OVERWRITE", oldPreference.getName()), this,
                    oldPreference);
            }
            // Process goes on in questionAnswer
            else
                questionAnswer(true, null);
        } // end method writePreference

        /**
         * Answer to a askQuestion request.
         *
         * @param  result      true if the user has clicked on Yes or Ok
         * @param  clientData  data specified in the askQuestion method
         *
         * @see    LyQuestionAnswerInterface
         */
        public void questionAnswer(boolean result, Object clientData)
        {
            if (!result)
                return;

            LySession     session = getSession();
        	LyViewManager viewManager = session.getViewManager();
        	boolean		  portletMode = viewManager.isPortletMode();
            LyApplication application = session.getApplication();
            LyObject      user = (_shared ? null : session.getUser());
            String        userFullId = ((user == null) ? null : user.getId().toString());
            LyPreference  preference = null;
            String        prefId = "_preference" + LyString.getRandomId();

            if (clientData instanceof LyPreference)
            {
                preference = (LyPreference)clientData;

                if ((_shared && (preference.getUser() != null))
                        || (!_shared && (preference.getUser() == null)))
                {
                    prefId = preference.getId();
                    application.deleteExtraInfo(preference);
                    preference = null;
                }
            }

            // retrieve the name of the data directory
            LyObjectInfo objectInfo;

            if (!_targetAction
                    && (_action.getClassInfo() != null))
                objectInfo = _action.getClassInfo();
            else
                objectInfo = _action;

            if (preference != null)
            {
                preference.setObjectInfo(objectInfo);
                preference.setDefault(_default);
            }
            else
            {
                preference = new LyPreference(prefId, objectInfo, _name, null, -1, userFullId, _default);
                preference.setApplication(application);
            }

            preference.setDefaultForMobile(_defaultForMobile);
            preference.setForPortlet(portletMode);
            completePreference(preference);

            // Save preference
            getApplication().saveExtraInfo(preference);

            setPreference(preference);

            _controller.free();
        } // end method questionAnswer
    } // end class _LySavePreferenceBehavior

    /**
     * Update the find menuitem status.
     */
    public void updatePreferencesMenu()
    {
        // If table is Idle, no need to refresh
        if (isIdle())
            return;

        LyArea      area = (LyArea)getComponent();
        LyTitleBar  titleBar = area.getTitleBar();
        LyToolBar   rightToolBar;
        LyFormField prefsField = null;

        if (titleBar != null)
        {
            rightToolBar = titleBar.getToolBar();
            if (rightToolBar != null)
                prefsField = (LyFormField)rightToolBar.getComponent(getId() + "_preferencesField");
        }

        if (prefsField == null)
        {
            rightToolBar = area.getTopRightToolBar();
            if (rightToolBar != null)
                prefsField = (LyFormField)rightToolBar.getComponent(getId() + "_preferencesField");
        }

        if (prefsField == null)
            return;

        LyInfoList<LyPreference> prefs = getPreferences();
        int                      i, n = ((prefs == null) ? 0 : prefs.getSize()), nd = 0;
        String                   id, label;
        LyPreference             preference;
        LyRelation               relation = (LyRelation)prefsField.getFieldItem();
        LyLink                   link;

        prefsField.setHidden(n == 0);

        for (i = 0; i < n; i++)
        {
            preference = prefs.getInfo(i);
            id = preference.getId();
            label = preference.getName();
            if (!preference.isReadOnly())
                nd++;

            link = (LyLink)relation.getComponent(id);
            if (link == null)
            {
                link = new LyLink(id, label);
                relation.addLink(link);
            }

            link.setSelected(preference == _preference);
        }

        LyView    view = getView();
        LyMenuBar menuBar = ((view instanceof LyFrame) ? ((LyFrame)view).getMenuBar() : null);

        LyMenu menu = null;

        if (menuBar != null)
            menu = menuBar.findMenu("_preferencesMenu", getId());

        if (menu != null)
        {
            LyItem item = menuBar.findItem(DEMAND_DELETE_PREFERENCE, getId());

            if (item != null)
                item.setEnabled(nd > 0);
        }

        // Maybe we have a toolbar with the correct tool
        rightToolBar = area.getTopRightToolBar();
        if (rightToolBar != null)
        {
            LyTool deleteTool = (LyTool)rightToolBar.getComponent(LyPanelController.DEMAND_DELETE_PREFERENCE
                    + LyController.ID_SEP + getId());

            if (deleteTool != null)
                deleteTool.setEnabled(nd > 0);
        }

        LyLink emptyLink = relation.getLink(0);

        if ((emptyLink != null)
                && emptyLink.getId().equals("LY_NO_CHOICE"))
            emptyLink.setSelected(_preference == null);
    } // end method updatePreferencesMenu

    /**
     * Complete preference with panel parameters.
     *
     * @param  preference  the preference to complete
     */
    protected void completePreference(LyPreference preference)
    {
        boolean  isFilterActive = isFilterActive();
        LyFilter filter = getFilter();

        if (isFilterActive && (filter != null))
            preference.setFilter(filter);
    }

    /**
     * Get a list of preferences for this action.
     *
     * @return  a list of preferences
     */
    protected LyInfoList<LyPreference> getPreferences()
    {
    	return getPreferences(true, true);
    }
    
    /**
     * Return a list of existing preferences.
     * 
     * @param currentAction indicates whether preferences of current action only must be returned
     * @param currentUser   indicates whether preferences of current user only must be returned
     */
    @SuppressWarnings("deprecation")
	protected LyInfoList<LyPreference> getPreferences(boolean currentAction, boolean currentUser)
    {
        LyApplication            application = getApplication();
        LyInfoList<LyPreference> prefs = application.getInfos(LyPreference.class);
        int                      n = ((prefs == null) ? 0 : prefs.getSize());

        if (n == 0)
            return null;

        LySession                session = getSession();
    	LyViewManager			 viewManager = session.getViewManager();
    	boolean					 portletMode = viewManager.isPortletMode();
        LyObject                 user = session.getUser();
        String                   userId, currentUserId = ((user == null) ? "" : user.getId());
        LyClassInfo              classInfo = _action.getClassInfo();
        LyPreference             preference;
        int                      i;
        LyInfoList<LyPreference> panelPrefs = null;
        LyObjectInfo             objectInfo;

        for (i = 0; i < n; i++)
        {
            preference = prefs.getInfo(i);
            userId = preference.getUser();

            // If preference is not applied on current action or class info, do not add it.
            objectInfo = preference.getObjectInfo();

            // If preference doesn't apply on current action, do not add it.
            if (currentAction)
            {
            	if (objectInfo == null)
                	continue;

	            if ((_action != objectInfo)
	            		&& !_action.hasTemplate(objectInfo)
	                    && ((classInfo == null)
	                        || ((classInfo != objectInfo)
	                        	&& !classInfo.hasTemplate(objectInfo))))
	                continue;
            }

            // Special case for preferences saved for import actions
            if ((preference.getName() != null)
                    && preference.getName().startsWith(LyImportController.PREFERENCE_PREFIX))
                continue;

            // If no file for this preference (happens when preference of
            // a chart info)
            if ((preference.getFile() == null)
            		&& ((preference.getId() == null) || !preference.getId().startsWith("__leonMetrics")))
                continue;
            
            // If preference is for portlet, do not display it in other modes
            if (!portletMode && preference.isForPortlet())
            	continue;

            // If current user does not own the preference do not show it
            if (currentUser
            		&& (userId != null)
                    && !currentUserId.equals(userId))
                continue;

            if (panelPrefs == null)
                panelPrefs = new LyInfoList<LyPreference>();

            panelPrefs.addInfo(preference);
        }

        if (panelPrefs != null)
            panelPrefs = panelPrefs.getInfosByName();

        return panelPrefs;
    } // end method getPreferences

    /**
     * Accessor (internal use) to the internal filter controller displaying (if exists) the filter
     * view of the panel.
     *
     * @return  the filter controller
     */
    public LyAreaController getFilterController()
    {
        if (_borderPanelController instanceof LyFilterController)
            return _borderPanelController;
        else if (_borderPanelController instanceof LyCompoundFilterController)
            return ((LyCompoundFilterController)_borderPanelController).getFilterController();
        else
            return null;
    }

    /**
     * Accessor (internal use) to the internal filter editor controller displaying (if exists) the
     * filter view of the panel.
     *
     * @return  the filter editor controller
     */
    public LyFilterEditorController getFilterEditorController()
    {
        return _filterEditorController;
    }

    /**
     * Internal method for closing border panel.
     */
    public void closeBorderPanel()
    {
        if ((_borderPanelController != null)
                && !_borderPanelController.isFreed())
        {
            _borderPanelController.hideView();
            _borderPanelController.free(this);
            _borderPanelController = null;
        }
    }

    /**
     * Answer to a askQuestion request.
     *
     * @param  result  true if the user has clicked on Yes or Ok
     * @param  client  data specified in the askQuestion method
     *
     * @see    LyQuestionAnswerInterface
     */
    public void questionAnswer(boolean result, Object clientData)
    {
        if (result)
        {
            if (clientData instanceof LyFilter)
            {
                setFilter((LyFilter)clientData);
                setLastModified(new Date().getTime());
            }
        }
    }

    /**
     * Check if the given sort is valid for physical layer. All its fields must be valid for the
     * physical layer.
     *
     * @param   sort  the sort to check
     *
     * @return  true if the sort is valid.
     */
    protected boolean checkFilterForPhysicalLayer(LyFilter filter)
    {
        // check if a criterion
        if (filter == null)
            return true;

        LyInfoList<LyFieldInfo> fields = filter.getFieldInfos(true);

        if (fields == null)
            return true;

        int         i, nbField = fields.getSize();
        LyFieldInfo field;

        for (i = 0; i < nbField; i++)
        {
            field = fields.getInfo(i);
            if (isFieldValidForSortOrFilter(field))
                continue;

            // if a single field is not valid, stop checkin and return false;
            return false;
        }

        return true;
    }

    /**
     * Check if a field is valid for a filter or a sort used on the physical layer. A field is valid
     * if it is not local or, if local, if it has a join or reverse daemon.
     *
     * @return  true if the field can be used in a physical sort or filter
     */
    protected boolean isFieldValidForSortOrFilter(LyFieldInfo fieldInfo)
    {
        if (!fieldInfo.hasMark(LyFieldInfo.MARK_LOCAL))
            return true; // non local field

        if (!(fieldInfo instanceof LyRelationFieldInfo)
                || (fieldInfo instanceof LyStructFieldInfo))
            return false; // only relation can be valid

        LyRelationFieldInfo relation = (LyRelationFieldInfo)fieldInfo;
        LyInfoDaemon        daemon = relation.getDaemon();

        if (daemon == null)
            return false;

        if (daemon instanceof LyJoinDaemon)
            return true;

        if (daemon instanceof LyReverseDaemon)
            return true;

        return false;
    }
} // end class LyPanelController
