/**
 * LyPasswordController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version:$Id: LyPasswordController.java 28997 2010-12-23 10:17:15Z bmichel $
 */

package leon.control;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * This class implements a user input dialog/window to change the user password. This controller is
 * associated with a behavior that implements LyPasswordBehaviorInterface. Usually, the behavior is
 * an instance of LyPasswordBehavior class.
 *
 * @see leon.app.behavior.LyPasswordBehavior
 * @see leon.app.behaviorinterface.LyPasswordBehaviorInterface
 */
public class LyPasswordController extends LyUserInputController
{
	/** File Version */
	public static final String __VERSION =
		"$Id: LyPasswordController.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** Current Password (for validation). */
	public static String	   PARAM_CURRENT_PASSWORD = "_currentPassword";

	/** New Password. */
	public static String	   PARAM_NEW_PASSWORD = "_newPassword";

	/** Confirm Password. */
	public static String	   PARAM_CONFIRM_PASSWORD = "_confirmPassword";

	/** Current user. */
	protected LyObject		   _user;

	/** User classInfo. default is LyLoginController.USER_CLASS_ID */
	protected LyClassInfo	   _userClassInfo;

	/** Password FieldInfo. */
	protected LyFieldInfo	   _passwordFieldInfo;

	/** Indicates if a password can be null **/
	protected boolean		   _allowEmptyPassword;
	
	/**
	 * Default constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller).
	 * @param action  action associated to the password. (parameters of view)
	 * @param classInfo related classInfo
	 * @param objects current context
	 */
	public LyPasswordController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		super(session, parent, action, classInfo, objects);

		_userClassInfo = ((classInfo == null) ? (LyClassInfo)getParameter("_userClassInfo")
											  : classInfo);
		_passwordFieldInfo = (LyFieldInfo)getParameter("_passwordFieldInfo");

		if (objects != null)
		{
			LyObject object = objects.getObject(0);

			if ((object != null)
					&& (object.getClassInfo() == _userClassInfo))
				_user = object;
		}
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyPasswordBehaviorInterface.class;
	}

	/**
	 * Prepare component before showing. This method is called once, before first showing.
	 */
	protected void completeComponent()
	{
		LySession     session = getSession();
		LyApplication application = session.getApplication();
		LyEnvironment environment = getEnvironment();

		// Use default
		if (_action == null)
			_action = application.getAction(LyAction.ID_PASSWORD);

		LyPasswordBehaviorInterface behavior = (LyPasswordBehaviorInterface)getBehavior();

		if (behavior != null)
			behavior.initializePassword(this);

		// Ensure we have the classInfo and fieldInfos
		if (_userClassInfo == null)
			_userClassInfo = application.getClassInfo(LyLoginController.USER_CLASS_ID);

		if (_userClassInfo == null)
		{
			environment.logWarning("PasswordController : classInfo not found.");
			return;
		}

		if (_passwordFieldInfo == null)
		{
			// Default is first SECRET field
			LyInfoList<LyFieldInfo> fields = _userClassInfo.getFieldInfos(LyFieldInfo.MARK_SECRET);

			if ((fields != null) && (fields.getSize() > 0))
				_passwordFieldInfo = fields.getInfo(0);
			else
				environment.logWarning("PasswordController : fieldInfo password not found in class "
					+ _userClassInfo);
		}

		// Update titlebar and label
		if (_action != null)
		{
			setImage(getImage(_action));

			if (_user == null)
				setTitleBar(_action.getName());
			else
				setTitleBar(_action.getName() + " : " + getName(_user));
		}

		_allowEmptyPassword = false;
		if (_passwordFieldInfo.hasMark(LyFieldInfo.MARK_OPTIONAL))
			_allowEmptyPassword = true;
		
		LyText currentPText, newPText, confirmPText;
		int    nCols = 30;

		// Current Password : ask if the user is not specified -use current user-
		if (_user == null)
		{
			currentPText = addText(PARAM_CURRENT_PASSWORD, "LY_CURRENT_PASSWORD", null, nCols, null,
				_allowEmptyPassword);
			currentPText.setSecret(true);
		}

		// New Password
		newPText = addText(PARAM_NEW_PASSWORD, "LY_NEW_PASSWORD", null, nCols, null, _allowEmptyPassword);
		newPText.setSecret(true);

		// Confirm Password
		confirmPText = addText(PARAM_CONFIRM_PASSWORD, "LY_CONFIRM_PASSWORD", null, nCols, null,
			_allowEmptyPassword);
		confirmPText.setSecret(true);

		// Client may do extra work here ...
		if (behavior != null)
			behavior.completeInitializePassword(this);

		super.completeComponent(); // Let parent do its job
	} // end method completeComponent

	/**
	 * Checks the values submitted in the password and return true if everything is valid.
	 *
	 * @param  values submitted values.
	 * @return true if validation is fine
	 */
	public boolean checkPassword(LyParameterValues values)
	{
		Object currentPassword = values.getParameter(PARAM_CURRENT_PASSWORD);
		Object newPassword = values.getParameter(PARAM_NEW_PASSWORD);
		Object confirmPassword = values.getParameter(PARAM_CONFIRM_PASSWORD);

		try
		{
			currentPassword = _passwordFieldInfo.parse((String)currentPassword);
			newPassword = _passwordFieldInfo.parse((String)newPassword);
			confirmPassword = _passwordFieldInfo.parse((String)confirmPassword);
		}
		catch (Exception e)
		{
			getEnvironment().writeException(e);
		}

		LyPasswordBehaviorInterface behavior = (LyPasswordBehaviorInterface)getBehavior();

		if ((behavior != null)
				&& behavior.changePassword(this, (String)currentPassword, (String)newPassword,
					(String)confirmPassword, values))
			return true;

		LyFieldItem currentPText = null;
		if (_user == null)
		{
			currentPText = ((LyFormField)(_component.getComponent(PARAM_CURRENT_PASSWORD)))
				.getFieldItem();
		}
		LyFieldItem newPText = ((LyFormField)(_component.getComponent(PARAM_NEW_PASSWORD)))
				.getFieldItem();
		LyFieldItem confirmPText = ((LyFormField)(_component.getComponent(PARAM_CONFIRM_PASSWORD)))
				.getFieldItem();

		if (currentPText != null)
			currentPText.setLabel(null);
		newPText.setLabel(null);
		confirmPText.setLabel(null);

		return false;
	} // end method checkPassword

	/**
	 * Proceed to next actions after a password (password should be valid at that point).
	 */
	public void invokeNextActions()
	{
		LySession session = getSession();
		LyAction  action;
		int		  i, n = _action.getActionCount();

		for (i = 0; i < n; i++)
		{
			action = _action.getAction(i);
			session.invokeAction(null, action, null, null, true);
		}
	}

	/**
	 * Accessor to the user to modify password.
	 *
	 * @return the user or null if not defined.
	 */
	public LyObject getUser()
	{
		return (_user);
	}

	/**
	 * Accessor to User classInfo.
	 *
	 * @return User classInfo
	 */
	public LyClassInfo getUserClassInfo()
	{
		return (_userClassInfo);
	}

	/**
	 * Accessor to password fieldInfo.
	 *
	 * @return password fieldInfo
	 */
	public LyFieldInfo getPasswordFieldInfo()
	{
		return (_passwordFieldInfo);
	}

	/**
	 * Accessor to User classInfo.
	 *
	 * @param userClassInfo User classInfo
	 */
	public void setUserClassInfo(LyClassInfo userClassInfo)
	{
		_userClassInfo = userClassInfo;
	}

	/**
	 * Accessor to password fieldInfo.
	 *
	 * @param passwordFieldInfo password fieldInfo
	 */
	public void setPasswordFieldInfo(LyFieldInfo passwordFieldInfo)
	{
		_passwordFieldInfo = passwordFieldInfo;
	}
	
	/**
	 * Getter for the allow empty password value.
	 * @return true if the password can be null, fall otherwise
	 */
	public boolean allowEmptyPassword()
	{
		return _allowEmptyPassword;
	}
} // end class LyPasswordController
