/**
 * LyPrintController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyPrintController.java 35128 2012-04-05 12:13:01Z nhamon $
 */

package leon.control;

import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.view.*;


/**
 * Print a list of objects.
 */
public class LyPrintController extends LySimplePrintController
{
	/** File Version */
	public static final String			  __VERSION =
		"$Id: LyPrintController.java 35128 2012-04-05 12:13:01Z nhamon $";

	/** Consult controllers used to create the print view (if needed by the view manager). */
	protected Vector<LyConsultController> _consultControllers;

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller).
	 * @param action    related action (_login).
	 * @param classInfo related class info
	 * @param objects   current action context
	 */
	public LyPrintController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		this(session, parent, action, classInfo, objects, false);
	}
	
	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller).
	 * @param action    related action (_login).
	 * @param classInfo related class info
	 * @param objects   current action context
	 * @param byList    print objects as a list
	 */
	public LyPrintController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects, boolean byList)
	{
		super(session, parent, action, objects);

        if  (action.getMin() == 0)
        {
            setByList(true);
            if  ((_objects == null)
                    && (parent instanceof LyAreaController))
                _objects = ((LyAreaController)parent).getContent();
            if (parent instanceof LyTableController)
                setFields(((LyTableController)parent).getFields()); 
        }
        else
            setByList(byList);
        
		LyViewManager viewManager = getViewManager();

		_consultControllers = null;

		if ((viewManager != null)
			&& (viewManager.hasProperty("MANAGE_PRINT")
			&& !action.getId().equals(LySimpleTableController.DEMAND_PRINT_TABLE))
			&& !_byList)
		{
			if (objects == null)
			{
				free(getParent());
				return;
			}
			
			_consultControllers = new Vector<LyConsultController>();

			LyClassInfo			 objectClassInfo;
			LyAction			 consultAction;
			LyController		 controller;
			LyInfoList<LyAction> actions;
			int					 i, n = objects.getSize();
			int					 j, m;
			LyObject			 object;

			for (i = 0; i < n; i++)
			{
				object = objects.getObject(i);
				objectClassInfo = object.getClassInfo();
				actions = objectClassInfo.getActionsWithId(LyAction.ID_CONSULT);
				m = ((actions == null) ? 0 : actions.getSize());

				for (j = 0; j < m; j++)
				{
					consultAction = actions.getInfo(j);
					// don't make a showView for printing !
					controller = invokeAction(consultAction, objectClassInfo,
							new LySimpleObjectList(object), false);

					if (controller instanceof LyConsultController)
					{
						_consultControllers.addElement((LyConsultController)controller);
						break;
					}
				}
			}
		}
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<LyPrintBehaviorInterface> getBehaviorInterface()
	{
		return LyPrintBehaviorInterface.class;
	}

	/**
	 * Returns default builder.
	 * @return the default print builder
	 */
	public LyBuilder getDefaultBuilder()
	{
		if (_consultControllers != null)
			return (new LyPrintBuilder(this));
		else
			return null;
	}

	/**
	 * Reset all the content of the view.
	 *
	 * @param full Boolean indicating if a full reset (including reset of datas) is requiered or not
	 *             (only reset graphical area).
	 */
	public void reset(boolean full)
	{
		LyConsultController consultCont;
		int				    i, n = _consultControllers.size();

		for (i = 0; i < n; i++)
		{
			consultCont = _consultControllers.elementAt(i);
			resetFormController(consultCont);
		}

		super.reset(full);
	} // end method reset
	
	/**
	 * Reset a given form controller recursively.
	 */
	protected void resetFormController(LyFormController formController)
	{
		LyFieldInfo		    fieldInfo;
		LyClassInfo		    classInfo;
		int				    j, nFields;
		LyAreaController    fieldController;

		formController.updateObject(null);
		classInfo = formController.getClassInfo();
		nFields = classInfo.getFieldInfoCount();

		for (j = 0; j < nFields; j++)
		{
			fieldInfo = classInfo.getFieldInfo(j);
			fieldController = formController.getAreaController(fieldInfo);

			if (fieldController instanceof LyFormController)
			{
				((LyFormController)fieldController).updateObject(null);
				resetFormController((LyFormController)fieldController);
			}
			else if (fieldController != null)
				fieldController.reset(true);
		}
	}

	/**
	 * Returns view controller.
	 * @return the view controller
	 */
	public LyViewController getViewController()
	{
		if (_consultControllers != null)
			return (getViewController(true));
		else
			return (super.getViewController());
	}

	/**
	 * Shows the Window which contains the view.
	 */
	public void showView()
	{
		if (_consultControllers != null)
		{
			LyViewController    viewController = getViewController();

			LyConsultController consultCont;
			LyViewController    consultViewController;
			LyView			    view;
			LyFieldInfo		    fieldInfo;
			LyClassInfo		    classInfo;
			LyAreaController    fieldController;
			int				    i, j, nFields, n = _consultControllers.size();

			for (i = 0; i < n; i++)
			{
				// For each consult controller, build its view (dialog) even
				// if not shown. This is mandatory in order to free the
				// consult controllers.
				consultCont = _consultControllers.elementAt(i);
				consultViewController = consultCont.getViewController(true);

				// Set the same controller id for these controllers. If not,
				// web viewer may throw an exception (forms with different
				// controllers in the same view).
				view = consultViewController.getView();
				if (view instanceof LyDialog)
					((LyDialog)view).setController(viewController.getId());
				else
					((LyFrame)view).setController(viewController.getId());
				
				// Retrieve the sub area controllers (composite relations)
				// and build their views too.
				classInfo = consultCont.getClassInfo();
				nFields = classInfo.getFieldInfoCount();

				for (j = 0; j < nFields; j++)
				{
					fieldInfo = classInfo.getFieldInfo(j);
					fieldController = consultCont.getAreaController(fieldInfo);

					if (fieldController != null)
					{
						// Structis / FFT00048
						LyController parent = fieldController.getParent();

						consultViewController = fieldController.getViewController();
						fieldController.setParent(parent);
						view = consultViewController.getView();
					}
				}
			}

			viewController.showView();
		}
		else
			super.showView();
	} // end method showView

	/**
	 * Return the consult controllers built for printing.
	 * @return a vector of consult controller
	 */
	public Vector<LyConsultController> getSonControllers()
	{
		return _consultControllers;
	}

	/**
	 * Printing objects starts.
	 */
	public void run()
	{
		// Create a new printjob
		Properties properties = new Properties();

		properties.put("orientation", (_byList ? "landscape" : "portrait"));

		LyViewManager viewManager = getViewManager();

		_printManager = viewManager.getPrintManager(this, properties);
		if (_printManager == null) // Cancelled or refused
		{
			freePrinting();
			return;
		}

		// Check printing parameters
		_inFile = _printManager.isInFile();
		_supportName = _printManager.getSupportName();
		_format = _printManager.getFormat();

		// Process printing action
		boolean over = false;

		startRun();

		// Check if an objectPrinterInterface is registered
		if (_behavior != null)
			over = _behavior.printBegin(this);

		if (!over)
			printEnd();

		// Call behavior to complete printing
		if (!over && (_behavior != null))
			over = _behavior.printEnd(this);

		endRun();

		freePrinting();
	} // end method run
} // end class LyPrintController
