/**
 * LyPrintModelController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyPrintModelController.java 34716 2012-03-15 14:52:58Z jlsaouli $
 */

package leon.control;

import java.io.*;
import leon.app.*;
import leon.app.behavior.*;
import leon.control.builder.*;
import leon.control.print.*;
import leon.data.*;
import leon.info.*;


/**
 * This class describes the controller associated to the print action of a list of objects from a
 * model file. This controller doesn't have a builder.
 *
 * @see LyAreaController
 * @see leon.app.behaviorinterface.LyPrintBehaviorInterface
 * @see leon.app.behavior.LyPrintBehavior
 */
public class LyPrintModelController extends LyAreaController
	implements Runnable
{
	/** File Version */
	public static final String __VERSION =
		"$Id: LyPrintModelController.java 34716 2012-03-15 14:52:58Z jlsaouli $";

	/** Object list to print. */
	protected LyObjectList     _objects;

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param classInfo related class info
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param objects   current context (may be null if the view is not contextual to a list of
	 *                  objects).
	 */
	public LyPrintModelController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		super(session, parent, action, objects);

		_action = action;
		_objects = objects;

		init();
	}

	/**
	 * Init this instance.
	 */
	protected void init()
	{
	}

	/**
	 * Returns default builder. This method overrides the standard method to return. There is no
	 * builder associated with this controller
	 *
	 * @return the default builder
	 */
	public LyBuilder getDefaultBuilder()
	{
		return null;
	}

	/**
	 * Returns view controller. This controller doesn't have a view controller
	 *
	 * @return the associated view Controller
	 */
	public LyViewController getViewController()
	{
		return null;
	}

	/**
	 * Shows the Window which contains the view.
	 */
	public void showView()
	{
		if ((_action != null)
				&& (_action.getParameter("_synchronous", true) != null)
				&& (_action.getParameter("_synchronous", true).equals("true")))
			run(); // Run Synchronous : mandatory for web mode when a result is shown.
		else
		{
			Thread t = new Thread(this, "BF AppEngine : printing from model action thread");

			t.start(); // Run asynchronous
		}
	}

	/**
	 * Printing objects starts.
	 */
	public void run()
	{
		LyPrintModelBehavior behavior = (LyPrintModelBehavior)getBehavior();

		if (behavior.printBegin(this))
			return;

		String			  modelParam = behavior.getModel(this);
		String			  outputParam = behavior.getOutput(this);
		String			  commandParam = behavior.getCommand(this);

		LyContentReplacer replacer = new LyContentReplacer(getEnvironment());

		replacer.setModelFile(modelParam);

		boolean modeXml = (replacer.getMode() == LyContentReplacer.MODE_XML);

		replacer = behavior.preparePrint(this, replacer);

		String result = replacer.getContentFromModel(_objects);

		result = behavior.completePrint(this, result);

		FileOutputStream os = null;
		Writer			 wr = null;
		PrintWriter		p = null;
		
		try
		{
			getEnvironment().logTrace("[LyPrintModelControler] writing print file to : "
				+ outputParam);
			File outputFile = new File(outputParam);
			// force creation of directory
			if (outputFile.getParentFile() != null)
				outputFile.getParentFile().mkdirs(); // force creation of parents directory
			

			if (modeXml)
			{
				os = new FileOutputStream(outputParam);
				wr = new OutputStreamWriter(os, "UTF8");
			}
			else
				wr = new FileWriter(outputParam);

			p = new PrintWriter(wr);
			p.print(result);

			if (commandParam != null)
			{
				String cmd = getEnvironment().getString().compose(commandParam,
						"\"" + outputParam + "\"", modelParam);

				getEnvironment().logTrace("[LyPrintModelControler] sending command : " + cmd);

				_environment.executeCommand(cmd, false);
			}
		}
		catch (IOException e)
		{
			getEnvironment().writeException(e);
		}
		finally
		{
			if (p != null)
				p.close();

			try
			{
				if (wr != null)
					wr.close();
			}
			catch (IOException e)
			{
				getEnvironment().writeException(e);
			}
			
			try
			{
				if (os != null)
					os.close();
			}
			catch (IOException e)
			{
				getEnvironment().writeException(e);
			}
		}

		if (behavior.printEnd(this))
			return;

		free();
	} // end method run

	/**
	 * Free this controller.
	 */
	public void free()
	{
		free(_parent);
	}
} // end class LyPrintModelController
