/*
 * LySelectByNameController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.control;

import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.setinterface.*;


/**
 * This is the controller for the select by name action.
 *
 * @version $Id: LySelectByNameController.java 31257 2011-05-31 14:07:18Z jlsaouli $
 * @author  Lyria
 */
public class LySelectByNameController extends LyUserInputController
	implements LyTransverseInterface
{
	/** File Version */
	public static final String	    __VERSION =
		"$Id: LySelectByNameController.java 31257 2011-05-31 14:07:18Z jlsaouli $";

	/** Search text identifier. */
	public static final String	    SEARCH_TEXT = "_searchText".intern();

	/** Refresh search text identifier. */
	public static final String	    REFRESH_SEARCH_TEXT = "_refreshSearchText".intern();

	/** Search results identifier. */
	public static final String	    SEARCH_RESULTS = "_searchResults".intern();

	/** Ignored characters in regular expressions. */
	public static final char[]	    IGNORED_CHARACTERS =
	{ '.', '[', ']', '{', '}', '(', ')', '+', '?', '^', '\\' };

	/** Text component containing the search text. */
	protected LyText			    _text;

	/** List component displaying the list of available objects. */
	protected LyComponent		    _list;

	/** Interface used to notify results. */
	protected LyFormResultInterface _formResultClient;

	/** Parent relation of the select controller. */
	protected LyRelationFieldInfo   _relationParent;

	/** Initial filter set on objects displayed in the list. */
	protected LyFilter			    _filter;

	/** Boolean indicating if context may be set. */
	protected boolean			    _allowSetContext;

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 * @param action  related action. If an action is specified, the controller uses action
	 *                parameters (action name, action specific XML view description ...)
	 */
	public LySelectByNameController(LySession session, LyController parent, LyAction action)
	{
		// This constructor is kept because it's called directly by other controllers
		this(session, parent, action, null, null);
	}

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo This parameter is not used. Kept from introspection compatibility.
	 */
	public LySelectByNameController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo)
	{
		// This constructor is kept because it's called directly by other controllers
		this(session, parent, action, classInfo, null);
	}

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 * @param action  related action. If an action is specified, the controller uses action
	 *                parameters (action name, action specific XML view description ...)
	 * @param context current context (may be null if the view is not contextual to a list of
	 *                objects).
	 */
	public LySelectByNameController(LySession session, LyController parent, LyAction action,
		LyObjectList context)
	{
		// This constructor is kept because it's called directly by other controllers
		this(session, parent, action, null, context);
	}

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo This parameter is not used. Kept from introspection compatibility.
	 * @param context   current context (may be null if the view is not contextual to a list of
	 *                  objects).
	 */
	public LySelectByNameController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent, action, classInfo, context);

		_allowSetContext = false;
	}

	/**
	 * Complete content of the component of the controller.
	 */
	protected void completeComponent()
	{
		_text = addText(SEARCH_TEXT, "LY_SELECT_BY_NAME_SEARCH", null, 30, null, false);

		LyFormField searchField = (LyFormField)_text.getParent();
		LyTip	    tip = new LyTip();

		tip.setLabel("LY_SELECT_BY_NAME_TIP");
		searchField.setTip(tip);
		_text.setNotifyValueChanged(true);

		// If relation is multiple or client is a filter controller, use a multiple relation
		if ((_relationParent != null) && _relationParent.isMultiple())
			_list = addRelation(SEARCH_RESULTS, "LY_SELECT_BY_NAME_RESULTS", true, false);
		else
			_list = addList(SEARCH_RESULTS, "LY_SELECT_BY_NAME_RESULTS", true, 10, false);

		updateResultList(false);
	} // end method completeComponent

	/**
	 * Invoked when an action occurs. (LyDemandListener interface) The current implementation calls
	 * the userInputDemand method of the behavior.
	 *
	 * @param demand the action demand.
	 */
	public void actionDemand(LyActionDemand demand)
	{
		String demandCode = demand.getSource().intern();

		if (demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
		{
			// Update result list when search text is changed
			if (demandCode == SEARCH_TEXT)
			{
				updateResultList(true);
				return;
			}
		}
		else if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			// Validate the form : update parent relation
			if (demandCode == REFRESH_SEARCH_TEXT)
			{
				updateResultList(true);
				return;
			}
			else if (demandCode == LyFormController.DEMAND_VALIDATE)
			{
				boolean result = notifyClient();

				if (result)
					free();

				return;
			}
			else if (demandCode == SEARCH_RESULTS)
			{
				// If relation is multiple, select the link first then notify
				if (_list instanceof LyRelation)
				{
					String value = demand.getValue();
					int    i, n = ((LyRelation)_list).getLinkCount();
					LyLink link;

					for (i = 0; i < n; i++)
					{
						link = ((LyRelation)_list).getLink(i);

						if ((value != null) && (value.equals(link.getId())))
							link.setSelected(true);
					}
				}

				boolean result = notifyClient();

				if (result)
					free();

				return;
			}
		}
		else if (demand.getType() == LyActionDemand.TYPE_MOUSE_OVER)
			return;

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Notify the client interface by calling the.{@link
	 * LyFormResultInterface#notifyFormResult(LyObject, LyRelationFieldInfo)}
	 * @return true if successfull
	 */
	protected boolean notifyClient()
	{
		if (_formResultClient != null)
		{
			LyListOption	   option;
			LyLink			   link;
			LyObjectId		   objectId;
			LyObject		   object;
			LySimpleObjectList selection = new LySimpleObjectList();
			LyApplication	   application = getApplication();
			String			   id;
			int				   i, n = 0;

			if (_list instanceof LyList)
				n = ((LyList)_list).getListOptionCount();
			else if (_list instanceof LyRelation)
				n = ((LyRelation)_list).getLinkCount();

			// Retrieve selected objects by their ids
			for (i = 0; i < n; i++)
			{
				id = null;

				if (_list instanceof LyList)
				{
					option = ((LyList)_list).getListOption(i);

					if (option.isSelected())
						id = option.getId();
				}
				else if (_list instanceof LyRelation)
				{
					link = ((LyRelation)_list).getLink(i);

					if (link.isSelected())
						id = link.getId();
				}

				if (id != null)
				{
					objectId = new LyObjectId(application, id);
					object = objectId.getLogicalObject();
					selection.addObject(object);
				}
			}

			if ((selection == null) || (selection.getSize() == 0))
			{
				// No result found
				LyEnvironment environment = getEnvironment();
				String		  message = environment.getMessage("LY_ER_NO_RESULT");

				showError(message);
				return false;
			}

			LySetController		   setController = null;
			LySetBehaviorInterface setBehavior = null;

			// Notify parent relation
			if (_formResultClient instanceof LySetController)
			{
				setController = (LySetController)_formResultClient;
				setBehavior = (LySetBehaviorInterface)setController.getBehavior();
			}

			n = selection.getSize();

			for (i = 0; i < n; i++)
			{
				object = selection.getObject(i);

				if ((setBehavior == null)
						|| setBehavior.enableRelationLink(setController, _relationParent, object,
							true))
					_formResultClient.notifyFormResult(object, _relationParent);
			}
		}

		return true;
	} // end method notifyClient

	/**
	 * Update the result list according to the search text.
	 *
	 * @param setIdle indicates whether component must be idled or not
	 */
	protected void updateResultList(boolean setIdle)
	{
		String search = _text.getLabel();

		if (search == null)
			search = "";

		search = search.toLowerCase();

		// First, empty list component
		LyListOption   option;
		LyLink		   link;
		int			   i, n = 0;
		Vector<String> selectedIds = new Vector<String>();

		if (_list instanceof LyList)
			n = ((LyList)_list).getListOptionCount();
		else if (_list instanceof LyRelation)
			n = ((LyRelation)_list).getLinkCount();

		if (setIdle)
		{
			LyViewComponent component = _list.getInternalView();

			if (component instanceof LySetIdleInterface)
				((LySetIdleInterface)component).setIdle(true);
		}

		for (i = n - 1; i >= 0; i--)
		{
			if (_list instanceof LyList)
			{
				option = ((LyList)_list).getListOption(i);
				((LyList)_list).removeListOption(option);
			}
			else if (_list instanceof LyRelation)
			{
				link = ((LyRelation)_list).getLink(i);

				// Do not remove link if selected for multiple case
				if (!link.isSelected())
					((LyRelation)_list).removeLink(link);
				else
					selectedIds.addElement(link.getId());
			}
		}

		// Get all the objects matching initial filter
		LyClassInfo   classInfo = getClassInfo();
		LySession     session = getSession();
		LyEnvironment env = getEnvironment();
		LyObject	  object;
		String		  objectName, objectNameLower;
		boolean		  fixSearch = true, addObject;
		LyWorkSpace   context = getContext();
        LyObjectList  ws;
		boolean		  useSimpleString = env.testResource("LY_IGNORE_TEXT_ACCENT");

		// If a context has been set, use it as a restriction of initial list
		if ((context != null))
			ws = context.getList(session, classInfo, _filter, false);
        else
            ws = new LyClassWorkSpace(session, classInfo, _filter, false);

		if ((search != null) && (search.indexOf('*') >= 0))
		{
			fixSearch = false;

			for (i = 0; i < IGNORED_CHARACTERS.length; i++)
				search = LyString.replace(search, String.valueOf(IGNORED_CHARACTERS[i]),
						"\\" + IGNORED_CHARACTERS[i]);

			search = LyString.replace(search, "*", ".*") + ".*";
		}

		// Manage mode with no accent
		if (useSimpleString && (search != null))
			search = LyString.getSimpleString(search);

		if (ws != null)
		{
			ws.setSortByName(true);
			ws.setSorted(true);
			n = ws.getSize();
		}
		else
			n = 0;

		// Retrieve LY_SELECT_BY_NAME_RELATION_COUNT resource to check the maximum
		// number of objects to be displayed in the list.
		LyEnvironment environment = getEnvironment();
		int			  max = environment.getIntResource("LY_SELECT_BY_NAME_RELATION_COUNT");
		int			  count = 0;

		// Keep only objects matching search text
		for (i = 0; i < n; i++)
		{
			addObject = false;
			object = ws.getObject(i);
			objectName = getName(object);

			if ("".equals(search) || "*".equals(search))
				addObject = true;
			else
			{
				objectNameLower = objectName.toLowerCase();

				if (useSimpleString)
					objectNameLower = LyString.getSimpleString(objectNameLower);

				if (fixSearch && objectNameLower.startsWith(search))
					addObject = true;
				else if (!fixSearch)
				{
					// Use a regular expression
					if (LyRegExpUtils.matchFormat(env, search, objectNameLower))
						addObject = true;
				}
			}

			// Do not add object if already selected
			if (selectedIds.contains(object.getId()))
				addObject = false;

			if (addObject)
			{
				if (_list instanceof LyList)
				{
					option = new LyListOption(object.getId(), objectName);
					((LyList)_list).addListOption(option);
				}
				else if (_list instanceof LyRelation)
				{
					link = new LyLink(object.getId(), objectName);
					((LyRelation)_list).addLink(link);
				}

				// Check if the maximum number of objects has been reached
				count++;

				if ((max > 0) && (count >= max))
					break;
			}
		}

		if (setIdle)
		{
			LyViewComponent component = _list.getInternalView();

			if (component instanceof LySetIdleInterface)
				((LySetIdleInterface)component).setIdle(false);
		}
	} // end method updateResultList

	/**
	 * Sets relation parent : this method is invoked when this controller is a child of another form
	 * (dynamic popup creation of an object in a relation field).
	 * 
	 * @param client the client of this action
	 * @param relation the relation fieldInfo which is updated as the result of the current action
	 *                 (usually dynamic popup creation)
	 */
	public void setRelationParent(LyFormResultInterface client, LyRelationFieldInfo relation)
	{
		_formResultClient = client;
		_relationParent = relation;

		if (relation == null)
			return;

		// Sets class filter if a filter is specified on relation for classInfo
		LyClassInfo classInfo = getClassInfo();

		if (client instanceof LySetController)
		{
			LySetController		   setController = (LySetController)client;
			LySetBehaviorInterface setBehavior = (LySetBehaviorInterface)setController
					.getBehavior();

			_filter = setBehavior.getRelationFilter(setController, relation, classInfo);
		}
		else
			_filter = relation.getFilter(classInfo);

		// Sets context from relation context
		if (client instanceof LySetController)
		{
			LySetController		   setController = (LySetController)client;
			LySetBehaviorInterface setBehavior = (LySetBehaviorInterface)setController
					.getBehavior();
			LyObjectList		   context = setBehavior.getRelationContext(setController,
					relation);

			if ((context != null))
			{
				_allowSetContext = true;
				setContext(context);
				_allowSetContext = false;
			}
		}

		// Sets single selection mode if relation is not multiple
		if (!relation.isMultiple()
				&& !(client instanceof LyFilterController))
		{
			// Call getComponent() method to ensure component (and also _list) is built
			getComponent();
			if (_list != null)
			    ((LyList)_list).setSelectionMode(LyList.SelectionMode.SINGLE);
		}
	} // end method setRelationParent

	/**
	 * Overridden to use the _allowSetContext parameter to check if context may be set.
	 *
	 * @see LyAreaController#setContext(LyWorkSpace)
	 */
	public boolean setContext(LyWorkSpace context)
	{
		boolean result = false;

		if (_allowSetContext)
			result = changeContext(context);

		return result;
	}
} // end class LySelectByNameController
