/**
 * LySelectController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LySelectController.java 34746 2012-03-16 12:26:11Z jlsaouli $
 */

package leon.control;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 */
public class LySelectController extends LyCompoundAreaController
	implements LyTransverseInterface
{
	/** File Version */
	public static final String	    __VERSION =
		"$Id: LySelectController.java 34746 2012-03-16 12:26:11Z jlsaouli $";

	/** Instance of LyFormResultInterface to be notified at the end of the process. */
	protected LyFormResultInterface _formResultClient;

	/** Relation field info using the select action. */
	protected LyRelationFieldInfo   _relationParent;

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. The action should be an instance of LyCompoundAction. If an
	 *                  action is specified, the controller uses action parameters (action name,
	 *                  action specific XML view description ...)
	 * @param classInfo classInfo : useless in this controller, should be null. Kept for
	 *                  compatibility.
	 * @param context   current context (may be null if the view is not contextual to a list of
	 *                  objects).
	 */
	public LySelectController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent, action, classInfo, context);
	}

	/**
	 * Sets relation parent : this method is invoked when this controller is a child of another form
	 * (dynamic popup creation of an object in a relation field).
	 *
	 * @param client the client of the current action
	 * @param relation the relation fieldInfo which is updated as the result of the current action
	 *                 (usually dynamic popup creation)
	 */
	public void setRelationParent(LyFormResultInterface client, LyRelationFieldInfo relation)
	{
		_formResultClient = client;
		_relationParent = relation;

		if (relation == null)
			return;

		// Sets class filter if a filter is specified on relation for classInfo
		LyTableController tableController = (LyTableController)getAreaController(1);
		LyClassInfo		  classInfo = tableController.getClassInfo();
		LyFilter		  filter;

		if (client instanceof LySetController)
		{
			LySetController		   setController = (LySetController)client;
			LySetBehaviorInterface setBehavior = (LySetBehaviorInterface)setController.getBehavior();

			filter = setBehavior.getRelationFilter(setController, relation, classInfo);
		}
		else
			filter = relation.getFilter(classInfo);

		tableController.setClassFilter(filter, false);

		// Sets context from relation context
		if (client instanceof LySetController)
		{
			LySetController		   setController = (LySetController)client;
			LySetBehaviorInterface setBehavior = (LySetBehaviorInterface)setController
					.getBehavior();
			LyObjectList		   context = setBehavior.getRelationContext(setController,
					relation);

			if ((context != null) && (context.getSize() > 0))
				tableController.setContext(context);
		}

		// Sets single selection mode if relation is not multiple
		if (!relation.isMultiple()
				&& !(client instanceof LyFilterController))
		{
			LyTable table = tableController.getTableComponent();

			table.setSelectionMode(LyTable.SelectionMode.SINGLE);
		}
	} // end method setRelationParent

	/**
	 * Invoked when an action occurs.
	 *
	 * @param demand the action demand
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			String demandCode = demand.getSource().intern();

			if (demandCode == LyFormController.DEMAND_VALIDATE)
			{
				if (_formResultClient != null)
				{
					LyAreaController areaController = getAreaController(1);
					LyObjectList     list = areaController.getSelection();

					if ((list == null) || (list.getSize() == 0))
					{
						// No result found
						LyEnvironment environment = getEnvironment();
						String		  message = environment.getMessage("LY_ER_NO_RESULT");

						showError(message);
						return;
					}

					LySetController		   setController = null;
					LySetBehaviorInterface setBehavior = null;

					if (_formResultClient instanceof LySetController)
					{
						setController = (LySetController)_formResultClient;
						setBehavior = (LySetBehaviorInterface)setController.getBehavior();
					}

					int		 i, n = list.getSize();
					LyObject object;

					for (i = 0; i < n; i++)
					{
						object = list.getObject(i);
						if ((setBehavior == null)
								|| setBehavior.enableRelationLink(setController, _relationParent,
									object, true))
							_formResultClient.notifyFormResult(object, _relationParent);
					}
				}
				else
				{
					// Invoke default action
					LyAreaController areaController = getAreaController(1);
					LyClassInfo		 classInfo = areaController.getClassInfo();
					LyObjectList     list = areaController.getSelection();
					LyAction		 action = classInfo.getDefaultAction();

					if (action != null)
					{
						LySession    session = getSession();
						LyController client = getParent().getParent();

						session.invokeAction(client, action, classInfo, list, true);
					}
				}

				free();
				return;
			}
		}

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Accessor to the LyFormResultInterface instance that must be notified at the end of the
	 * process.
	 * 
	 * @return the form result client
	 */
	public LyFormResultInterface getFormResultClient()
	{
		return _formResultClient;
	}

	/**
	 * Accessor to the relation field info using the select action.
	 * @return the parent relation
	 */
	public LyRelationFieldInfo getRelationParent()
	{
		return _relationParent;
	}
} // end class LySelectController
