/**
  * LySelectFileController.java
  * Copyright (c) 2000-2009, Lyria-W4.
  * All rights reserved.
  */
  
package leon.control;

import java.util.*;
import leon.app.*;
import leon.data.*;
import leon.info.*;

import leon.view.*;
import leon.view.event.*;


/**
 * This class implements a file/directory selection
 *
 * @author  Lyria S.A.
 * @version $Id: LySelectFileController.java 30098 2011-03-16 12:35:25Z jlsaouli $
 */
public class LySelectFileController extends LyUserInputController
{
	/** File Version */
	public static final String __VERSION =
		"$Id: LySelectFileController.java 30098 2011-03-16 12:35:25Z jlsaouli $";

	/** Default Path key */
	public static final String LY_SELECT_FILE_DEFAULT_PATH_KEY = "LY_SELECT_FILE_DEFAULT_PATH";
    
	/** Default file width */
    public static int          FILE_WIDTH  = 80;
	
	/** default opening directory or file. */
	protected String		   _defaultPath;

	/** Selected directory or path. */
	protected String		   _selectedPath;

	/** Custom message displayed. */
	protected String		   _message;

	/** Accessor file type in selector */
	protected LyFile.Type		_type;
    
    /** Show file state. */
    protected boolean          _showFile;
    

	/**
	 * Listeners for callback
	 */
	protected List<LySelectFileListenerInterface> _listeners;
	
	/**
	 * This constructor creates a new controller object.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 * @param classInfo  related classInfo.
	 */
	public LySelectFileController(LySession session, LyController parent, LyClassInfo classInfo)
	{
		this(session, parent,
		    session.getApplication().getActionWithId(LyAction.ID_SELECT_FILE, classInfo),
		    classInfo, null);
	}

	/**
	 * This constructor creates a new controller object.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 * @param action  related action. If an action is specified, the controller uses action
	 *                parameters (action name, action specific XML view description ...)
	 * @param context current context (may be null if the view is not contextual to a list of
	 *                objects).
	 */
	public LySelectFileController(LySession session, LyController parent, LyAction action,
		LyObjectList context)
	{
		this(session, parent, action, null, context);
	}

	/**
	 * This constructor creates a new controller object.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo This parameter is not used. Kept from introspection compatibility.
	 * @param context   current context (may be null if the view is not contextual to a list of
	 *                  objects).
	 */
	public LySelectFileController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent, action, classInfo, context);
	}

	/**
	 * Reset all the content of the view. This method may be called several times. This method
	 * should be overridden, default implementation makes nothing
	 *
	 * @param full Boolean indicating if a full reset (including reset of datas) is required or not
	 *             (only reset graphical area).
	 */
	public void reset(boolean full)
	{
		if (full)
		{
			if (getMessage() != null)
				addLabel("__message", "", getEnvironment().getString(getMessage()));
			
			LyFile file = addFile("__selectedPath", "LY_SELECT_PATH", null, getType(), false);
            
            file.setNCols(FILE_WIDTH);
            
            LyChoice    choice = addChoice("__showFile", "LY_SHOW_FILE", true, true);
            LyOption    option;

            option = new LyOption("__showFile.yes", null);
            option.setChecked(false);
            choice.addOption(option);
		}
        
		super.reset(full);
	}

	/**
	 * @see leon.control.LyAreaController#free()
	 */
	@Override
	public void free() 
	{
		super.free();
		_listeners = null;
	}

	/**
	 * Invoked when an action occurs. (LyDemandListener interface) The current implementation calls
	 * the userInputDemand method of the behavior.
	 *
	 * @param demand the action demand.
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (!isFreed())
		{
			LyParameterValues paramValues = demand.getParameterValues();
			
			_selectedPath = paramValues.getParameter("__selectedPath");
			if (_selectedPath != null)
				getSession().addValue(LY_SELECT_FILE_DEFAULT_PATH_KEY, _selectedPath);
            
            String showFile = paramValues.getParameter("__showFile");
            
            _showFile = (showFile != null);
		}
        
		List<LySelectFileListenerInterface> listeners = _listeners;
        LyAreaController                    parent = getParentAreaController();
        
		super.actionDemand(demand);
		
		if (listeners != null)
		{
			Iterator<LySelectFileListenerInterface> iter = listeners.iterator();
			
			while(iter.hasNext())
				iter.next().selectFileDemand(this, demand);
		}
		
        if (_showFile
            && (demand.getType() == LyActionDemand.TYPE_ACTION)
            && demand.getSource().equals(LyFormController.DEMAND_VALIDATE)
            && (_selectedPath != null))
        {
            LyApplication       application = getApplication();
            LyAction            action = application.getAction("_text");
            LyTextController    textController = new LyTextController(_session, parent, action, _selectedPath, 0);
            
            textController.showView();
        }
	}

	/** 
	 * default opening directory or file. 
	 * @return the path or null
	 */
	public String getDefaultPath() 
	{
		if (_defaultPath != null)
			return _defaultPath;
		else if (getSession().getValue(LY_SELECT_FILE_DEFAULT_PATH_KEY) != null)
			return (String)getSession().getValue(LY_SELECT_FILE_DEFAULT_PATH_KEY);
		else
			return null;
	}

	/** 
	 * Set the default path
	 * @param path the default path
	 */
	public void setDefaultPath(String path) 
	{
		_defaultPath = path;
	}

	/**
	 * Return user path selection in the input dialog
	 * @return user path
	 */
	public String getSelectedPath() 
	{
		return _selectedPath;
	}

	/**
	 * Accessor user path selection in the input dialog
	 * @param path user path
	 */
	public void setSelectedPath(String path) 
	{
		_selectedPath = path;
	}

	/**
	 * Accessor detailed message in the view
	 * @return the requested message
	 */
	public String getMessage() 
	{
		return _message;
	}

	/**
	 * Accessor detailed message in the view
	 * @param message the requested message
	 */
	public void setMessage(String message) 
	{
		this._message = message;
	}

	/**
	 * Accessor file type in selector
	 * @return one of the File.Type values
	 */
	public LyFile.Type getType() 
	{
		if (_type == null)
			return LyFile.Type.TEXT;
		return _type;
	}

	/**
	 * Accessor file type in selector
	 * @param type one of the File.Type values
	 */
	public void setType(LyFile.Type type) 
	{
		this._type = type;
	}
	
	/**
	 * Add a listener for callback
	 * @param listener
	 */
	public void addListener(LySelectFileListenerInterface listener)
	{
		if (_listeners == null)
			_listeners = new ArrayList<LySelectFileListenerInterface>();
		_listeners.add(listener);
	}
	/**
	 * remove a listener for callback
	 * @param listener
	 */
	public void removeListener(LySelectFileListenerInterface listener)
	{
		if (_listeners != null && listener != null)
			_listeners.remove(listener);
	}
} // end class LySelectFieController
