/**
 * LySetController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LySetController.java 41448 2012-11-07 10:37:47Z bangiole $
 */

package leon.control;

import java.io.*;
import java.util.*;

import leon.app.*;
import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.control.form.*;
import leon.data.*;
import leon.data.datainterface.*;
import leon.info.*;
import leon.info.misc.*;
import leon.info.save.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;


/**
 * This class describes the controller associated to a modification form.
 *
 * <P>This class implements a 'read-write' LyFormController and specializes methods from its parent.
 * </P>
 *
 * <P>This form controller is associated to an object or a list of objects, however this is not
 * mandatory. The LySetController is also inherited by LyCreateController which may not have any
 * object.</P>
 *
 * <P>The LySetController is associated to a LySetBuilder class which sets the content and the
 * overall structure of the view.</P>
 *
 * <P>This class uses a LyClassBehavior instance which may implement specific application logic,
 * additional value validation, automatic field calculation, value filtering or dynamic impact on
 * value changes...</P>
 *
 * <P>Services supported by the LySetController :</P>
 *
 * <UL>
 *   <LI>Toggle mode to editable/non-editable : can have a read-only modification form (a
 *     consultation form with the same look-and-feel).</LI>
 *   <LI>User interaction on command bar : validate, cancel...</LI>
 *   <LI>Dynamically keep the content consistent if object is modified</LI>
 *   <LI>Extract (retrieve) values set by the user from the form</LI>
 *   <LI>Set a value into the form</LI>
 *   <LI>Manage impact fields changes. CONTEXT_FIELD : the modification of the target relation
 *     fieldInfo forces the context of the relation field.</LI>
 *   <LI>Manage impact fields changes. CLASS_FIELD : the modification of the target field forces the
 *     classInfo of the relation field.</LI>
 *   <LI>Manage impact fields changes. RULES : the modification of the field changes the status of
 *     the target field.</LI>
 *   <LI>Validation and error messages when the form is closed with 'validate'.</LI>
 * </UL>
 *
 * @see leon.app.behavior.LyClassBehavior
 * @see leon.app.behaviorinterface.LySetBehaviorInterface
 * @see leon.control.builder.LySetBuilder
 * @see LyConsultController
 * @see LyCreateController
 */
public class LySetController extends LyConsultController
	implements LyQuestionAnswerInterface,
		LyContextInterface,
		LyFormInterface,
		LyFormResultInterface,
		LySubmitTab,
		LyContextListener,
		LyTransverseInterface
{
	/** File Version */
	public static final String		  __VERSION =
		"$Id: LySetController.java 41448 2012-11-07 10:37:47Z bangiole $";

	/** Indicate whether this form is read-only or read-write. Default value is true = read-write */
	protected boolean				  _editable;

	/** Indicate if the current action is ending or not. */
	protected boolean				  _ending;

	/** Indicate if the current action is an apply or a validate. */
	protected boolean				  _apply;

	/** Value set of set values Filled in after startValidate and cleared in endValidate. */
	protected LyValueSet			  _setValues;

	/** Indicate if propagation is caused by a value change or not. */
	public boolean					  _valueChange;

	/**
	 * Keep last modification date for internal use in tab-actions Handle conflict when user has
	 * changed data in forms and object is modified. Changing tab must not ignore objectChanged
	 */
	protected long					  _last_object_changed;

	/** Boolean indicating that update is no longer needed. */
	protected boolean				  _update;

	/** Context for this controller. */
	protected LyContext				  _controllerContext;

	/** Indicate if a updateViewFromValues is in process. */
	protected boolean				  _inProcess;

	/**
	 * Internal use : avoid overriding from content in creation form when actually creating the
	 * temporary LyObject.
	 */
	protected LyFieldInfo			  _valueChanging;

	/** Indicate when an error last occurred in value changed. */
	protected long					  _errorInValueChanged;

	/** Field infos being edited (fast edition mode). */
	protected LyInfoList<LyFieldInfo> _fastEditedFields;

	/**
	 * The client to call back when the action is completed.
	 *
	 * @see #_relationParent
	 */
	protected LyFormResultInterface   _formResultClient;

	/**
	 * Relation parent : this field is set when this form is the son of another form controller.
	 * Usually, when a popup creation form is opened to add a new object in a relation field. (See
	 * _relationFieldInfo to see which field is the target of the creation.
	 */
	protected LyRelationFieldInfo     _relationParent;

	/** Errors list. */
	protected Vector<LyError>         _errors;

	/**
	 * List of temp files created by file controller in this controller : delete them on exit
	 */
	protected List<String>            _tempFilesPath;

    /**
     * Form save to be used for restoring form.
     */
    protected LyFormSave              _formSave;


	/**
	 * Constructor. Modification form for a list of objects. ClassInfo parameter is ignored, kept
	 * for compatibility and introspection capabilities.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller).
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo this parameter is kept for introspection and compatibility. not used.
	 * @param objects   list of objects currently being modified.
	 */
	public LySetController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		super(session, parent, action, classInfo, objects);

		if (isCreation())
			_editable = true;
		_ending = false;
		_apply = false;
		_valueChange = false;
		_update = true;
		_fastEditedFields = null;
		_errorInValueChanged = -1;
		_relationParent = null;
		_compositionRelation = null;
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LySetBehaviorInterface.class;
	}

	/**
	 * Factory for constructing the field controller of a field info.
	 *
	 * @param  fieldInfo Field info for which a controller is built.
	 * @param forceConsult indicates if the given field is forced in consult mode
	 * @return The field controller of the given field.
	 */
	public LyFieldController fieldControllerFactory(LyFieldInfo fieldInfo, boolean forceConsult)
	{
		if (forceConsult
				|| (isConsult(fieldInfo)
						&& ((fieldInfo instanceof LyRelationFieldInfo)
								|| (fieldInfo.getFormula() == null))))
			return super.fieldControllerFactory(fieldInfo, forceConsult);
		else if (fieldInfo instanceof LyFileFieldInfo)
			return new LyFileFieldSetController(this, (LyFileFieldInfo)fieldInfo);
		else if (fieldInfo instanceof LyChoiceFieldInfo)
			return new LyChoiceFieldSetController(this, (LyChoiceFieldInfo)fieldInfo);
		else if (fieldInfo instanceof LyStructFieldInfo)
		{
			// Test if we have a format or an value label.
			LyStructFieldInfo	   struct = (LyStructFieldInfo)fieldInfo;
			LyInfo				   parent = fieldInfo.getParent();
			LySetBehaviorInterface behavior = getSetBehavior();

			if ((struct.getClassInfoCount() > 1)
					&& (struct.fromFieldTag())
					&& !(struct.isMultiple()))
			{
				LyClassInfo		  subStructClass;
				LyStructFieldInfo subStructFieldInfo;
				LyUnits			  units;
				int				  i, n = struct.getClassInfoCount();

				for (i = 0; i < n; i++)
				{
					subStructClass = struct.getClassInfo(i);
					subStructFieldInfo = (LyStructFieldInfo)subStructClass.getParent();
					units = subStructFieldInfo.getUnits();
					if (units != null)
						return new LyCompositionFieldUnitSetController(this, struct);
				}

				if ((parent instanceof LyArrayFieldInfo)
						&& !behavior.useEditTable(this, (LyArrayFieldInfo)parent))
					return new LyStructInArraySetController(this, (LyArrayFieldInfo)parent, struct);
				else
					return new LyCompositionFieldSetController(this, struct);
			}
			else
			{
				// Test if we have a format or an value label.
				LyUnit	    unit = getDefaultFieldUnit(fieldInfo);
				LyInfoLabel label = fieldInfo.getLabel(LyInfoLabel.Type.VALUE);

				if ((struct.fromFieldTag())
						&& !(struct.isMultiple())
						&& (!(parent instanceof LyArrayFieldInfo)))
				{
					LyClassInfo subStructClass = struct.getClassInfo();

					if ((_object != null)
							&& _object.getClassInfo().contains(struct))
					{
						LyRelationValue relationValue = _object.getRelationValue(struct);

						if (relationValue.getValueCount() > 0)
						{
							LyObject object = relationValue.getObject(0);

							if (object != null)
								subStructClass = object.getClassInfo();
						}
					}

					LyStructFieldInfo subStruct = (LyStructFieldInfo)subStructClass.getParent();

					unit = getDefaultFieldUnit(subStruct);
					label = subStruct.getLabel(LyInfoLabel.Type.VALUE);
					if (unit != null)
						return new LyCompositionFieldUnitSetController(this,
							(LyStructFieldInfo)fieldInfo);
				}

				if ((unit != null) || (label != null))
					return new LyStructFieldSetController(this, (LyStructFieldInfo)fieldInfo);
				else if ((parent instanceof LyArrayFieldInfo)
						&& !behavior.useEditTable(this, (LyArrayFieldInfo)parent))
					return new LyStructInArraySetController(this, (LyArrayFieldInfo)parent,
						(LyStructFieldInfo)fieldInfo);
				else
					return new LyCompositionFieldSetController(this, (LyStructFieldInfo)fieldInfo);
			}
		}
		else if (fieldInfo instanceof LyRelationFieldInfo)
		{
			LyRelationFieldInfo		 relation = (LyRelationFieldInfo)fieldInfo;
			LyRelationFieldInfo.Type type = relation.getType();

			switch (type)
			{
			case COMPOSITION:
				if (fieldInfo.isMultiple())
					return new LyMultiCompositionFieldSetController(this, relation);
				else if (relation.getShowAs() == LyRelationFieldInfo.ShowAs.LINK)
					return new LyRelationAsLinkFieldSetController(this, relation);
				else
					return new LyCompositionFieldSetController(this, relation);
			default:
				if (fieldInfo.isMultiple()
						&& (relation.getShowAs() == LyRelationFieldInfo.ShowAs.TABLE))
					return new LyMultiCompositionFieldSetController(this, relation);
				else if (!fieldInfo.isMultiple()
						&& (relation.getShowAs() == LyRelationFieldInfo.ShowAs.FORM))
					return new LyCompositionFieldSetController(this, relation);
				else if (relation.getShowAs() == LyRelationFieldInfo.ShowAs.LINK)
					return new LyRelationAsLinkFieldSetController(this, relation);
				else
					return new LyRelationFieldSetController(this, relation);
			}
		}
		else if (fieldInfo instanceof LyArrayFieldInfo)
		{
			LyArrayFieldInfo	   array = (LyArrayFieldInfo)fieldInfo;
			LySetBehaviorInterface behavior = getSetBehavior();

			if (behavior.useEditTable(this, array))
			{
				LyFieldInfo subFieldInfo = array.getFieldInfo();

				if (subFieldInfo instanceof LyStructFieldInfo)
				{
					LyUnit	    unit = getDefaultFieldUnit(subFieldInfo);
					LyInfoLabel label = subFieldInfo.getLabel(LyInfoLabel.Type.VALUE);

					if ((unit != null) || (label != null))
						return new LyArrayTableStructFieldSetController(this, array);
					else
						return new LyMultiCompositionArrayFieldSetController(this, array,
							(LyRelationFieldInfo)subFieldInfo);
				}
				else if (isComposition(subFieldInfo))
					return new LyMultiCompositionArrayFieldSetController(this, array,
						(LyRelationFieldInfo)subFieldInfo);
				else
					return new LyArrayTableFieldSetController(this, array);
			}
			else
				return new LyArrayFieldSetController(this, array);
		}
		else if (fieldInfo instanceof LyDateFieldInfo)
			return new LyDateFieldSetController(this, fieldInfo);
		else if (fieldInfo instanceof LyTextFieldInfo)
			return new LyTextFieldSetController(this, fieldInfo);
	    else if (fieldInfo instanceof LyNumberFieldInfo)
	        return new LyNumberFieldSetController(this, (LyNumberFieldInfo)fieldInfo);
		else
			return new LyFieldSetController(this, fieldInfo);
	} // end method fieldControllerFactory

	/**
	 * Update the view of the controller with the specified values from the demand (internal use) .
	 *
	 * @param demand demand parameters sent by the view component.
	 */
    public void updateViewFromValues(LyDemand demand)
	{
		if (!_inProcess && (!isFreed()))
		{
			boolean currentInProcess = _inProcess;

			_inProcess = true;
			super.updateViewFromValues(demand);
			_inProcess = currentInProcess;
		}
	}

	/**
	 * Invoked when an action occurs. At this level, the implementation manages<BR>
	 * DEMAND_VALIDATE demands. (demand.getType = LyActionDemand.TYPE_ACTION AND demand.getSource()
	 * = LyController.DEMAND_VALIDATE)<BR>
	 * DEMAND_APPLY demands. (demand.getType = LyActionDemand.TYPE_ACTION AND demand.getSource() =
	 * LyController.DEMAND_APPLY)<BR>
	 * Default action (double click)<BR>
	 * Default action (double click)<BR>
	 * VALUE_CHANGED demands. (demand.getType = LyActionDemand.TYPE_VALUE_CHANGED
	 *
	 * @param demand the action demand
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		LyParameterValues values = demand.getParameterValues();
		String			  demandCode = demand.getSource().intern();

		// FFT01069
		if (demandCode.equals("_tabs"))
			super.actionDemand(demand);

		if (values != null)
		{
			// Warning : in this is a tabaction change and a modification has just been received
			// Ignore updates and keep modifications
			if ((_last_object_changed == getLastModified())
					&& (getSession().getController(demandCode) instanceof LyTabAreaController))
				_last_object_changed = 0; // Restore to 0 :take this into account first time.
			else
			{
				setLastModified(System.currentTimeMillis());

				if (demand.getType() != LyActionDemand.TYPE_VALUE_CHANGED)
				{
					boolean currentInProcess = _inProcess;

					_inProcess = true;
                    values.updateViewFromValues(getView(), !currentInProcess);
					_inProcess = currentInProcess;
				}
			}
		}

		if (demandCode.equals("_tabs"))
			return;

		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			if (demandCode == DEMAND_VALIDATE)
			{
				setIdle(true);
				validate(true);
				if (!isFreed())
					setIdle(false);
				return;
			}
			else if (demandCode == DEMAND_APPLY)
			{
			    apply();
				return;
			}
	        else if (demandCode == DEMAND_SAVE)
	        {
                saveForm(true);
                return;
	        }
		}

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Sets the value of a fieldInfo.
	 *
	 * @param fieldInfo the field which should be modified
	 * @param newValue     the new value of the field. This value may be null to set the value to blank
	 *                  (empty).
	 * @param propagate Indicates whether the change is notified (propagated) to other fields or
	 *                  not. This is useless if you reset all fields for instance.
	 * @see   #getFieldValue
	 */
	public void setFieldValue(LyFieldInfo fieldInfo, LyValue newValue, boolean propagate)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController != null)
			fieldController.setFieldValue(newValue, propagate);
	}

    /**
     * Method to recompute the field value of a field.
     * @param  fieldInfo Field to compute.
     * @return Computed field value.
     */
    public LyValue computeFieldValue(LyFieldInfo fieldInfo)
    {
        return computeFieldValue(fieldInfo, null);
    } // end method computeValue

    /**
     * Internal method to compute the field value of a field.
     * @param  fieldInfo Field to compute.
     * @param  expression the expression to evaluate
     * @return Computed field value.
     */
	protected LyValue computeFieldValue(LyFieldInfo fieldInfo, String expression)
	{
        if (!contains(fieldInfo)
            || ((expression == null)
                && (!isEnabled(fieldInfo) || (fieldInfo.getFormula() == null))))
            return getFieldValue(fieldInfo);

        // First of all, we have to fill a LyValueSet which contains all the values
        // needed to calculate this field.
	    LyFormula               formula = null;
        LyInfoList<LyFieldInfo> formulaFields;

	    if (expression != null)
	    {
	        formula = LyFormula.formulaFactory(fieldInfo, expression);
	        formulaFields = formula.getFormulaFields();

	        if (formulaFields.getSize() == 0)
	            return fieldInfo.buildFieldValue(expression);
	    }
	    else
	        formulaFields = fieldInfo.getFormulaFields();

        int                     i, n = (formulaFields != null ? formulaFields.getSize() : 0);
        LyFieldInfo             formulaField;
        LyClassInfo             classInfo = getClassInfo();
        LyValueSet              values = new LyValueSet(classInfo);
        LyValue                 dependValue;

        for (i = 0; i < n; i++)
        {
            formulaField = formulaFields.getInfo(i);
            if (isConsult(formulaField)
                    && formulaField.isComputed())
                dependValue = computeFieldValue(formulaField, null);
            else
                dependValue = getFieldValue(formulaField);

            if (dependValue != null)
                values.addValue(dependValue);
        }

        Object  fieldValue;

        if (expression == null)
            fieldValue = fieldInfo.compute(values);
        else
            fieldValue = formula.getValue(values);

        if (isOptional(fieldInfo)
                && (fieldValue instanceof Number)
                && (((Number)fieldValue).doubleValue() == 0))
            fieldValue = null;

        return fieldInfo.buildFieldValue(fieldValue);
	}

	/**
	 * Sets the value of a fieldInfo Shortcut for setFieldValue(fieldInfo, value, true).
	 *
	 * @param fieldInfo the field which should be modified
	 * @param value     the new value of the field. This value may be <b>null</b> (to set the value
	 *                  to blank (empty)), a <b>primitive object type</b> (String, Integer...) or a
	 *                  <b>LyValue</b>
	 * @see   #setFieldValue(LyFieldInfo, Object, boolean)
	 */
	public void setFieldValue(LyFieldInfo fieldInfo, Object value)
	{
		setFieldValue(fieldInfo, value, true);
	}

	/**
	 * Sets the value of a fieldInfo.
	 *
	 * @param fieldInfo the field which should be modified
	 * @param value     the new value of the field. This value may be null, a LyValue instance or
	 *                  the internal value associated to the fieldInfo. See LyValue.getValue()
	 * @param propagate Indicates whether the change is notified (propagated) to other fields or
	 *                  not. This is useless if you reset all fields for instance.
	 * @see   #getFieldValue
	 * @see   leon.data.LyValue#getValue
	 */
	public void setFieldValue(LyFieldInfo fieldInfo, Object value, boolean propagate)
	{
		LyValue fieldValue = null;

		if (value instanceof LyValue)
			fieldValue = (LyValue)value;
		else if (value != null)
			fieldValue = fieldInfo.buildFieldValue(value);

		setFieldValue(fieldInfo, fieldValue, propagate);
	}

	/**
	 * Extract current value for the specified fieldInfo in the form.
	 *
	 * @param  fieldInfo fieldInfo associated to the requested value
	 * @return current value, or null if not found or field is disabled.
	 */
	public LyValue getFieldValue(LyFieldInfo fieldInfo)
	{
		if (_fields == null)
			return null; // Not added yet, do not call getFields and isDisplayed while form is being built

		LyFieldController fieldController = getFieldController(fieldInfo);

		return ((fieldController == null) ? null : fieldController.getFieldValue());
	}

	/**
	 * Extract current value for the specified fieldInfo in the form.
	 *
	 * @param  formField form field from which the value is expected.
	 * @param  fieldInfo fieldInfo associated to the requested value
	 * @param  unit      Unit of the value.
	 * @return current value, or null if not found or field is disabled.
	 */
	protected final LyValue getFieldValue(LyFormField formField, LyFieldInfo fieldInfo,
		LyUnit unit)
	{
		if (_fields == null)
			return
			null; // Not added yet, do not call getFields and isDisplayed while form is beeing built

		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController instanceof LyFieldSetController)
			return ((LyFieldSetController)fieldController).getFieldValue(unit);
		else
			return null;
	}

	/**
	 * Return the list of all values set in the current form, return null, if an error occurs. This
	 * method performs a validity check before returning the values.
	 *
	 *@param fieldInfos a list of fieldInfos
	 * @return the list of values, as getFieldValue for all fields.
	 * @see    #getFieldValue
	 */
	protected LyValueSet getValueSet(LyInfoList<LyFieldInfo> fieldInfos)
	{
		LyEnvironment environment = getEnvironment();
		boolean		  stopOnError = !environment.testResource("LY_FORM_SHOW_ALL_ERRORS");
		Vector<LyError>	subErrors = null;

        if (!stopOnError)
        {
        	if (_errors != null)
        	{
        		// Keep errors from simple sub-relations and put them at the end of the errors list
    			LyError	error;
    			Object	transmitter;
        		int		i, n  = _errors.size();

        		for (i = 0; i < n; i++)
        		{
        			error = _errors.elementAt(i);
        			transmitter = error.getTransmitter();

        			if (transmitter instanceof LySetController)
        			{
	        			LyFieldInfo	errorField = ((LySetController)transmitter).getCompositionRelation();

	        			if ((errorField instanceof LyRelationFieldInfo)
	        					&& !errorField.isMultiple())
	        			{
	        				if (subErrors == null)
	        					subErrors = new Vector<LyError>(1);

	        				subErrors.add(_errors.get(i));
	        			}
        			}
        		}
        	}

        	_errors = new Vector<LyError>(1);
        }

		if (_component instanceof LyPanel)
			// Freeze the panel
			((LyPanel)_component).setIdle(true);

		// Create field Values corresponding to field items
		LyClassInfo classInfo = getClassInfo();
		int		    i, n;
		LyFieldInfo fieldInfo;
		LyValueSet  values = new LyValueSet(classInfo);
		short	    status;
		boolean     errorFound = false;

		if (fieldInfos == null)
			fieldInfos = classInfo.getFieldInfos();
		n = fieldInfos.getSize();

		for (i = 0; i < n; i++)
		{
			fieldInfo = fieldInfos.getInfo(i);
			status = addValue(values, fieldInfo);
			if (status == LySetBehaviorInterface.STATUS_IGNORE)
				continue;
			else if (status == LySetBehaviorInterface.STATUS_KO)
			{
			    errorFound = true;
				if (stopOnError)
					break;
			}
			else if (status == LySetBehaviorInterface.STATUS_CANCEL)
			{
				// Cancel this edition, free this controller and return null
				free();
				return null;
			}
		}

		if (_component instanceof LyPanel)
			// Refresh the panel
			((LyPanel)_component).setIdle(false);

		if (subErrors != null)
		{
			if (_errors == null)
				_errors = new Vector<LyError>(1);

			_errors.addAll(subErrors);
		}

		if (_errors != null)
            showSetErrors();

        if (errorFound)
            // There was an error, return null
            return null;

        return values;
    } // end method getValueSet

    /**
     * Reset errors.
     */
    protected void resetErrors()
    {
        LyEnvironment environment = getEnvironment();
        String		  resource = environment.getResource("LY_FORM_SHOW_ERRORS_IN_FORM");
        boolean		  showErrorsOnField = "field".equals(resource);

        if (_component != null)
        {
            LyArea		  area = (LyArea)_component;
            LyPanel		  borderPanel = area.getBorderPanel(LyArea.NORTH);

            if (borderPanel != null)
            {
                borderPanel.setVisible(false);
                area.setBorderPanel(null, LyArea.NORTH);
                borderPanel.free();
            }
        }

        if (showErrorsOnField)
        {
    		LyClassInfo		classInfo = getClassInfo();
    		int				i, n = classInfo.getFieldInfoCount();
    		LyFieldInfo		fieldInfo;
    		LyFormField		formField;

    		for (i = 0; i < n; i++)
    		{
    			fieldInfo = classInfo.getFieldInfo(i);
    			formField = getFormField(fieldInfo);

    			if (formField != null)
    				formField.setMessage(null);
    		}
        }

        _errors = null;
    }

    /**
     * Show errors view.
     */
    protected void showSetErrors()
    {
        if (_errors == null)
            return;

        if (_errors.size() == 0)
        {
            _errors = null;
            return;
        }

        if ((_compositionRelation != null)
            && (_parent instanceof LySetController))
        {
            LySetController   setControllerParent = (LySetController)_parent;

            if (setControllerParent._errors == null)
                setControllerParent._errors = new Vector<LyError>(_errors);
            else
                setControllerParent._errors.addAll(_errors);
        }
        else
        {
            // Show errors and reset errors
            LyEnvironment environment = getEnvironment();
            LyError       error;
            String        message;
            StringBuffer  buffer = new StringBuffer();
            Object[]      args;
            int           i, n = _errors.size();
            String		  resource = environment.getResource("LY_FORM_SHOW_ERRORS_IN_FORM");
            boolean		  showErrorsOnField = "field".equals(resource);

            for (i = 0; i < n; i++)
            {
                error = _errors.elementAt(i);
                message = error.getMessage();
                args = error.getArgs();

                if (showErrorsOnField)
                {
                	LyFieldInfo	fieldInfo = error.getOrigin();
                	LyFormField	formField = getFormField(fieldInfo);

                	if (formField != null)
                		formField.setMessage(environment.getMessage(message, args));
                }
                else
                {
	                if (i > 0)
	                    buffer.append('\n');
	                buffer.append(environment.getMessage(message, args));
                }
            }

            if (!showErrorsOnField)
            	showError(buffer.toString());
        }

        _errors = null;
    }

	/**
     * Show an error for the specified object.
     *
     * @param object The object in error
     * @param error  The error to show
     */
    public void showError(LyObject object, LyError error)
    {
        if (_errors == null)
            super.showError(object, error);
        else
            _errors.add(error);
    } // end method showError

	/**
	 * Show a popup message error box with the specified message. The message may be a text message,
	 * or the key to a String in the application dictionnary. (stringsXXX.properties files) This is
	 * a shortcut to getViewManager().showError()
	 *
	 * @param message the message to show
	 * @see leon.view.LyViewManager
	 * @see leon.misc.LyStrings
	 */
	public void showError(String message)
	{
		LyEnvironment environment = getEnvironment();

		if (!environment.testResource("LY_FORM_SHOW_ERRORS_IN_FORM"))
		{
		    super.showError(message);
		    return;
		}

		LyController	parent = getParent();
		LyController	lastSetController = parent;

		while (parent instanceof LySetController)
		{
			lastSetController = parent;
			parent = parent.getParent();
		}

		if (lastSetController instanceof LySetController)
		{
			lastSetController.showError(message);
			return;
		}

		LyArea			area = (LyArea)_component;
		LyArea			messageArea = new LyArea();
		LyForm			messageForm = new LyForm("_errorForm");
		LyFormField		messageField = new LyFormField("_errorField");
		LyCaption		messageCaption = new LyCaption();
		LyConstraint	constraint = new LyConstraint();
		LyLabel			label = new LyLabel();

		constraint.setHAlign(LyConstraint.HAlign.LEFT);
		constraint.setRemainder(false);
		constraint.setColSpan(4);
		messageCaption.setLabel(message);
		messageCaption.setColon(false);
		messageCaption.setConstraint(constraint);
		messageField.setState(LyFormField.STATE_ERROR);
		messageField.setCaption(messageCaption);
		messageField.setFieldItem(label);
		messageForm.addFormField(messageField);
		messageArea.setBorder(LyArea.Border.IN);
		messageArea.addPanel(messageForm);

		constraint = new LyConstraint();
		constraint.setHAlign(LyConstraint.HAlign.LEFT);
		constraint.setRemainder(true);
		constraint.setFill(LyConstraint.Fill.HORIZONTAL);
		messageArea.setConstraint(constraint);
		messageForm.setConstraint(constraint);

		area.setBorderPanel(messageArea, LyArea.NORTH);
		messageForm.resetContent();
	}

	/**
	 * Complete a list of values with the value of the given field. This
	 * method performs a validity check before returning the values.
	 * @param values a list of values to complete
	 * @param fieldInfo the field
	 *
	 * @return the status of the add operation.
	 * @see    #getFieldValue
	 */
	protected short addValue(LyValueSet values, LyFieldInfo fieldInfo)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		if (!(fieldController instanceof LyFieldSetController)
		        || isConsult(fieldInfo))
			return LySetBehaviorInterface.STATUS_IGNORE;

		// Create field Values corresponding to field items
		LyFieldSetController   fieldSetController = (LyFieldSetController)fieldController;
		LySetBehaviorInterface behavior = getSetBehavior();
		LyFieldInfoValue	   newValue;
		short				   status;
		LyFormField			   formField;
		LyFieldItem			   fieldItem;

		// Reset errors on fieldInfo for the object.
		resetFieldInfoErrors(fieldInfo);

		// We must set the form field to valid to know if getFieldValue sets it to invalid.
		fieldSetController.setState(LyFormField.STATE_NORMAL);

		// Get new value
		newValue = (LyFieldInfoValue)fieldSetController.getFinalValue();

		// Check validity
		if (fieldSetController.getState() == LyFormField.STATE_ERROR)
			return LySetBehaviorInterface.STATUS_KO;
		if (newValue == null)
			return LySetBehaviorInterface.STATUS_IGNORE;

		formField = fieldController.getFormField();
		if (formField != null)
		{
			fieldItem = formField.getFieldItem();
			status = fieldItem.validateInput();

			// If status is KO, try to get error message from field item
			if (status == LySetBehaviorInterface.STATUS_KO)
			{
				String message = fieldItem.getErrorMessage();

				if ((message != null) && (message.length() > 0))
					showError(message);
			}
		}
		else
			status = LySetBehaviorInterface.STATUS_OK;

		if (status == LySetBehaviorInterface.STATUS_OK)
			// Control this new value
			status = behavior.controlNewValue(this, newValue);

		if (status == LySetBehaviorInterface.STATUS_IGNORE)
			return status;
		else if (status == LySetBehaviorInterface.STATUS_CANCEL)
		{
            if (_setValues != null)
                _setValues.removeValue(fieldInfo);
			return status;
		}
		else if (status == LySetBehaviorInterface.STATUS_OK)
		{
			// Call controls on fieldInfo : this call must'nt be removed because
			// some viewer may not have called valueChanged (HTML viewer by example).
			if (_object != null)
				status = _object.callFieldControls(fieldInfo, newValue.getValue());
		}

		fieldSetController.setState(status);

		if ((status == LySetBehaviorInterface.STATUS_OK)
				|| (status == LySetBehaviorInterface.STATUS_WARNING))
			values.addValue(newValue);
		else if (status == LySetBehaviorInterface.STATUS_KO)
			checkFieldController(fieldController);

		return status;
	} // end method addValue

	/**
	 * Called when current object has changed . Overriden to avoid refresh of the view when the form
	 * is validated.
	 *
	 * @param oldValues set of former values, or null if not specified.
	 */
	public void updateObject(LyValueSet oldValues)
	{
		if (!_update)
			return;
		if (_ending && !isApplied())
			// No need to update the view will be discarded
			return;

		super.updateObject(oldValues);
	}

	/**
	 * Internal use : update the value of a fieldInfo in the form. <B>Application should use
	 * setFieldValue instead.</B> For relation fieldInfo, this method invokes updateRelation
	 * (specialization). For choice fieldInfo, this method invokes updateChoice (specialization).
	 *
	 * @param fieldInfo the corresponding field info.
	 * @param newValue  the new value of the field. This value may be null, a LyValue instance or
	 *                  the internal value associated to thge fieldInfo. See LyValue.getValue()
	 * @param oldValue  the latest value of the field. This value may be null, a LyValue instance or
	 *                  the internal value associated to thge fieldInfo. See LyValue.getValue()
	 * @param propagate Indicates whether the change is notified (propagated) to other fields or
	 *                  not. This is useless if you reset all fields for instance.
	 * @see   #setFieldValue
	 */
	protected void updateFieldItem(LyFieldInfo fieldInfo,
		Object newValue, Object oldValue, boolean propagate)
	{
		super.updateFieldItem(fieldInfo, newValue, oldValue, propagate);

		// Update the last modification time.
		setLastModified(System.currentTimeMillis());
	}

	/**
	 * Exec given rule inside this controller.
	 *
	 * @param rule    Rule to apply.
	 * @param checked True if the rule is checked, false if unchecked.
	 */
	public void execRule(LyRule rule, boolean checked)
	{
		execRule(rule, checked, _valueChange);
	}

	/**
	 * Exec given rule inside this controller.
	 *
	 * @param rule    Rule to apply.
	 * @param checked True if the rule is checked, false if unchecked.
	 * @param valueChange indicates if this calls comes from a value changed event
	 */
	protected void execRule(LyRule rule, boolean checked, boolean valueChange)
	{
		LyFieldInfo depend = (LyFieldInfo)rule.getObjectInfo();

		depend = _classInfo.getRealFieldInfo(depend);

		if (contains(depend))
		{
			boolean			 onCheck = rule.isOnCheck();
			LyRule.Operation operation = rule.getOperation();
			Object			 oldValue = null;
			boolean			 checkDepend = false;
			String           formula = rule.getFormula();

			if (_object != null)
				oldValue = _object.getValue(depend);

			if (operation == LyRule.Operation.SET_ENABLED)
				setEnabled(depend, (onCheck == checked));
			else if (operation == LyRule.Operation.SET_EDITABLE)
				setEditable(depend, (onCheck == checked));
			else if (operation == LyRule.Operation.SET_OPTIONAL)
			{
				setOptional(depend, (onCheck == checked));
				checkDepend = true;
			}
			else if (operation == LyRule.Operation.SET_VISIBLE)
				setHidden(depend, (onCheck != checked));
			else if ((operation == LyRule.Operation.SHOW)
					&& (onCheck == checked))
				setActiveField(depend);
			else if (operation == LyRule.Operation.SET_LABEL)
				setLabel(depend, rule.getLabel(), (onCheck == checked));
			else if (onCheck == checked)
			{
				if (operation == LyRule.Operation.CLEAR)
					clear(depend);
				else if (operation == LyRule.Operation.RESET)
					reset(depend);
				else if (operation == LyRule.Operation.SET_VALUE)
				{
				    if (formula == null)
				        setFieldValue(depend, rule.getValue());
				    else
				        setFieldValue(depend, computeFieldValue(depend, rule.getFormula()));
				}

				valueChanged(depend);
			}

			if ((_object != null)
					&& isEnabled(depend))
			{
				LyValue fieldValue = getFieldValue(depend);
				Object  value = ((fieldValue == null) ? null : fieldValue.getValue());

				// If value was changed by the rule, control field and update state
				if (checkDepend || (!depend.equals(value, oldValue)))
				{
					_object.callFieldControls(depend, value);
					setState(depend, _object.getErrorLevel(depend));
				}
			}
		}
		else if (_fieldControllers != null)
		{
			Enumeration<LyFieldController> elements = _fieldControllers.elements();
			LyFieldController              fieldController;
			LyAreaController               areaController;

			while (elements.hasMoreElements())
			{
				fieldController = elements.nextElement();
				areaController = fieldController.getAreaController();
				if (areaController instanceof LySetController)
					((LySetController)areaController).execRule(rule, checked, valueChange);
			}
		}
	} // end method execRule

	/**
	 * Internal method for control and propagate a value change. This methods is called when the
	 * field has been changed by the user and checks current value validity.
	 *
	 * <P>This validity is called for possible specific application implementation on the
	 * LyClassBehavior instance.</P>
	 *
	 * <P>The LySetController additionaly impact the view using the setValid method.</P>
	 *
	 * @param fieldInfo fieldInfo which was modified by the user.
	 * @see   leon.app.behavior.LyClassBehavior#controlNewValue
	 * @see   #setValid
	 */
	public void valueChanged(LyFieldInfo fieldInfo)
	{
		LySetBehaviorInterface behavior = getSetBehavior();
		LyFieldInfoValue	   value = (LyFieldInfoValue)getFieldValue(fieldInfo);

		if (value == null)
			// An error occured when getting the value.
			return;

		// Keep value changing since when the object has not yet been created, getObject(true) will
		// create a new one and send an event to the GUI to update the form with default value.
		_valueChanging = fieldInfo;
		// Reset errors on fieldInfo for the object
		LyObject object = getObject();

		_valueChanging = null;

		resetFieldInfoErrors(fieldInfo);

		short status = behavior.checkNewValue(this, value);

		if (status == LySetBehaviorInterface.STATUS_KO)
			_errorInValueChanged = System.currentTimeMillis();

		if ((object == null) || object.isDeleted())
		{
			if ((status != LySetBehaviorInterface.STATUS_KO)
					&& (_inProcess == false))
			{
				boolean	valueChange = _valueChange;

				_valueChange = true;
				propagate(fieldInfo, ((value == null) ? null : value.getValue()));
				_valueChange = valueChange;
			}

			setState(fieldInfo, status);
			return;
		}

		LyValue oldValue = object.getFieldValue(fieldInfo);

		if (!fieldInfo.equals(oldValue.getValue(), value.getValue()))
		{
		    object.callFieldControls(fieldInfo, value.getValue());
			if (isFreed())
				return;
		}

		if (status != LySetBehaviorInterface.STATUS_KO)
		{
	        if (object.getErrorLevel(fieldInfo) != 0)
	        {
	            // Error occurred during object modification (and not during behavior.checkValue()
	            int        i, n = object.getErrorsCount();
	            LyError    error;

	            for (i = 0; i < n; i++)
	            {
	                error = object.getError(i);
	                if (error.getOrigin() == fieldInfo)
	                {
	                    showError(null, error);
	                    break;
	                }
	            }
	        }
		    else if (!_inProcess)
    		{
				boolean	valueChange = _valueChange;

    			_valueChange = true;
    			propagate(fieldInfo, ((value == null) ? null : value.getValue()));
    			_valueChange = valueChange;
    		}
		}

		updateStatus();

		if (status != LySetBehaviorInterface.STATUS_OK)
		{
			short currentState = getState(fieldInfo);

			if ((currentState == LySetBehaviorInterface.STATUS_OK)
					|| (status < currentState))
				setState(fieldInfo, status);
		}

		if ((_compositionRelation != null)
				&& (_compositionRelation instanceof LyStructFieldInfo)
				&& (_parent instanceof LySetController))
		{
			LyContext parentContext = ((LySetController)_parent).getControllerContext();

			parentContext.setLinkUndo(true);
			((LySetController)_parent).valueChanged(_compositionRelation);
			parentContext.setLinkUndo(false);
		}
	} // end method valueChanged

	/**
	 * Internal method to handle a unit change. This methods is called when the field unit was
	 * changed by the user.
	 *
	 * @param fieldInfo fieldInfo which was modified by the user.
	 * @param unit      id of the new unit.
	 */
	protected void unitChanged(LyFieldInfo fieldInfo, String unit)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController instanceof LyFieldSetController)
			((LyFieldSetController)fieldController).unitChanged(unit, null);
	}

	/**
     * Apply form.
     * @return Result status.
     */
    public short apply()
    {
        short  result;

        _apply = true;
        result = validate(false);
        _apply = false;
        return result;
    }

	/**
	 * The form is closing and we check the values typed in by the user.
	 *
	 * <P>This method is using the LyClassBehavior (LySetBehaviorInterface) to complete and validate
	 * the form.</P>
	 *
	 * <P>When closing the validate method is following these steps :</P>
	 *
	 * <UL>
	 *   <LI>Invoke the LyClassBehavior.addExtraValues (LySetBehaviorInterface) to let the
	 *     application fill in additional or calculated values to the result.</LI>
	 *   <LI>Validation of all the values using LyClassBehavior.controlNewValues</LI>
	 *   <LI>Invoke the end method (which will invoke LyClassBehavior.endSet</LI>
	 * </UL>
	 *
	 * @param free indicates if the form is closed and freed after a successful validation.
	 * @return the status of the validation
	 * @see   leon.app.behavior.LyClassBehavior#addExtraValues
	 * @see   leon.app.behavior.LyClassBehavior#controlNewValues
	 * @see   leon.app.behavior.LyClassBehavior#endSet
	 * @see   LyFormInterface
	 */
	public short validate(boolean free)
	{
		if ((System.currentTimeMillis() - _errorInValueChanged) < 500)
		{
			_errorInValueChanged = -1;
			return LyFormInterface.STATUS_KO;
		}

		// Start validation of the form
		short status = startValidate();

		if (status == LySetBehaviorInterface.STATUS_KO)
		{
			// There was a problem, we keep the view opened
			showErrors();
			return LyFormInterface.STATUS_KO;
		}

		if (status == LySetBehaviorInterface.STATUS_CANCEL)
			// Edition was cancelled
			return LyFormInterface.STATUS_CANCEL;

		// Nothing done ?
		if (status == LySetBehaviorInterface.STATUS_IGNORE)
		{
			if (free)
				// The action is considered as cancelled
				free();
			else if (!isCreation())
				nextObject(true);

			return LyFormInterface.STATUS_IGNORE;
		}

		// End validation
		status = endValidate();

		if (status == LySetBehaviorInterface.STATUS_KO)
		{
			// There was a problem, we keep the view opened
			showErrors();
			return LyFormInterface.STATUS_KO;
		}

		if (status == LySetBehaviorInterface.STATUS_CANCEL)
			// Edition was cancelled
			return LyFormInterface.STATUS_CANCEL;

		if (_action == null)
			// This controller was freed during validation, just return
			return LyFormInterface.STATUS_OK;

		// Get the object.
		LyObject   object = getObject();
        LyValueSet newValues = (object == null ? null : object.getNewValues());

		// Commit modifications
    	LyApplication	application	= getApplication();
    	LyContext		currentContext = application.getCurrentContext();

        try
        {
        	if (currentContext == null)
        		// Be sure to have a context otherwise it could cause problems to daemons
        		application.setCurrentContext(_controllerContext);

        	commit();
        }
        finally
        {
        	if (currentContext == null)
        		// Restore context if it was set
        		application.setCurrentContext(null);
        }

		// Check controller context before committing
		if ((_controllerContext != null)
			&& ((_controllerContext.getStatus() == LyContext.Status.ERROR)
				|| (_controllerContext.getStatus() == LyContext.Status.ROLLBACKED)
				|| (_controllerContext.getErrorLevel() == LyObject.LEVEL_ERROR)))
		{
			if ((newValues != null)
			    && (_controllerContext.getStatus() != LyContext.Status.ROLLBACKED))
			{
				// Create a new object when needed
				object = getObject();
				if (object != null)
				{
					// And restore values
					_controllerContext.setValues(object, newValues);
					updateObject(newValues);
				}
			}

			if (_relationParent == null)
				// Commit failed, there are problems
				showErrors();
			return LyFormInterface.STATUS_KO;
		}

		// Do we have a result client for creation ?
		if ((_formResultClient != null)
				&& (object != null))
		{
			if (free)
				_update = false;
			if (_action.hasMark(LyAction.MARK_TRANSVERSE))
				_formResultClient.notifyFormResult(object, _relationParent);
			else if ((_formResultClient instanceof LyStructFieldSetController)
					|| (_formResultClient instanceof LyCompositionFieldUnitSetController)
					|| (_formResultClient instanceof LyRelationTableController))
				// For a change of the struct object
				_formResultClient.notifyFormResult(object, _relationParent);
		}

		if (free)
		{
			// Free this controller
			if (!isFreed())
				free();
		}
		else if (!isCreation())
			// Go to next object
			nextObject(true);

		return LyFormInterface.STATUS_OK;
	} // end method validate

	/**
	 * Commit sub-forms.
	 */
	public void subFormsCommit()
	{
		if (_fieldControllers != null)
		{
			Enumeration<LyFieldController> elements = _fieldControllers.elements();
			LyFieldController              fieldController;

			while (elements.hasMoreElements())
			{
				fieldController = elements.nextElement();
				if (fieldController instanceof LyFieldSetController)
					((LyFieldSetController)fieldController).commit();
			}
		}
	}

	/**
	 * Cancel sub-forms.
	 */
	public void subFormsCancel()
	{
		_update = false;
		if (_fieldControllers != null)
		{
		    Enumeration<LyFieldController> elements = _fieldControllers.elements();
			LyFieldController              fieldController;

			while (elements.hasMoreElements())
			{
				fieldController = elements.nextElement();
				if (fieldController instanceof LyFieldSetController)
					((LyFieldSetController)fieldController).cancel();
			}
		}
	}

	/**
	 * Update all values...
	 */
	public void updateValues()
	{
		getValueSet(null);
	}

	/**
	 * Start validation of the form. The form is closing and we check the values typed in by the
	 * user.
	 *
	 * <P>When closing the validate method is following these steps :</P>
	 *
	 * <UL>
	 *   <LI>Invoke the LyClassBehavior.addExtraValues (LySetBehaviorInterface) to let the
	 *     application fill in additional or calculated values to the result.</LI>
	 *   <LI>Validation of all the values using LyClassBehavior.controlNewValues</LI>
	 * </UL>
	 * @return the status of this operation
	 * @see leon.app.behavior.LyClassBehavior#addExtraValues
	 * @see leon.app.behavior.LyClassBehavior#controlNewValues
	 * @see LyFormInterface
	 */
	public short startValidate()
	{
		LyObject object = getObject();

		if ((object == null)
		        && !isCreation())
			return LyFormInterface.STATUS_IGNORE;

		// Start validation of sub-forms
		LyClassInfo		  classInfo = getClassInfo();
		int				  i, n = classInfo.getFieldInfoCount();
		LyFieldInfo		  fieldInfo;
		LyFieldController fieldController;
		short			  tempStatus, subFormsStatus, status;
	    LyEnvironment     environment = getEnvironment();
	    boolean           stopOnError = !environment.testResource("LY_FORM_SHOW_ALL_ERRORS");

	    // Reset errors
        resetErrors();

		subFormsStatus = LyFormInterface.STATUS_IGNORE;
		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			fieldController = getFieldController(fieldInfo);
			if (!(fieldController instanceof LyFieldSetController))
				continue;

			tempStatus = ((LyFieldSetController)fieldController).startValidate();
			if (tempStatus == LyFormInterface.STATUS_KO)
			{
				subFormsStatus = LyFormInterface.STATUS_KO;
				if (stopOnError)
					break;
			}
			else if (tempStatus == LyFormInterface.STATUS_OK)
			{
				if (!isOptional(fieldInfo))
					tempStatus = ((LyFieldSetController)fieldController).endValidate();
				if (tempStatus == LyFormInterface.STATUS_KO)
					subFormsStatus = LyFormInterface.STATUS_KO;
			}
		}

		short formStatus = LyFormInterface.STATUS_IGNORE;

		if (i < n)
			// Loop was stop on error
			_setValues = null;
		else
			// Retrieve values
			_setValues = getValueSet(null);

		if (isFreed())
			return LyFormInterface.STATUS_CANCEL;

		if (_setValues == null)
			// At least an error occurred, we can't go any further
			formStatus = LyFormInterface.STATUS_KO;
		else
		{
			// Add some extra values of the behavior.
			LySetBehaviorInterface behavior = getSetBehavior();

			behavior.addExtraValues(this, _setValues);

			// No error occured, control values all together
			tempStatus = controlNewValues(_setValues);

			if (tempStatus != LyFormInterface.STATUS_IGNORE)
				formStatus = tempStatus;
		}

		if ((subFormsStatus == LyFormInterface.STATUS_KO)
				|| (formStatus == LyFormInterface.STATUS_KO))
			status = LyFormInterface.STATUS_KO;
		else if ((subFormsStatus == LyFormInterface.STATUS_IGNORE)
				&& (formStatus == LyFormInterface.STATUS_IGNORE))
			status = LyFormInterface.STATUS_IGNORE;
		else
			status = LyFormInterface.STATUS_OK;
		return status;
	} // end method startValidate

	/**
	 * Internal method to control a set of values for this controller.
	 *
	 * @param values Controlled values.
	 * @return the status of the control
	 * @see LyFormInterface
	 */
	protected short controlNewValues(LyValueSet values)
	{
		// Reset class errors on the object
		resetClassErrors();

		LySetBehaviorInterface behavior = getSetBehavior();
		short				   result;

		result = behavior.controlNewValues(this, values);
		updateStatus();

		if (result == LyControlInterface.STATUS_ERROR)
			return LyFormInterface.STATUS_KO;
		else if (result == LyControlInterface.STATUS_WARNING)
			return LyFormInterface.STATUS_OK; // XXX
		else
			return LyFormInterface.STATUS_OK;
	}

	/**
	 * End validation of the form : commit edition.
	 * @return the status of the validation
	 * @see LyFormInterface
	 */
	public short endValidate()
	{
		// <BEGIN OF TRANSACTION>

		// Maybe this controller has been canceled.
		if ((_object == null) && !isCreation())
			return LyFormInterface.STATUS_IGNORE;

		// End validation of sub-forms
		LyClassInfo			    classInfo = getClassInfo();
		int					    i, n = classInfo.getFieldInfoCount();
		LyFieldInfo			    fieldInfo;
		LyFieldController	    fieldController;
		short				    result;
		LyInfoList<LyFieldInfo> fieldInfos = new LyInfoList<LyFieldInfo>();

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			if (!isOptional(fieldInfo))
				continue;

			fieldController = getFieldController(fieldInfo);
			if (!(fieldController instanceof LyFieldSetController))
				continue;

			result = ((LyFieldSetController)fieldController).endValidate();
			if (result == LyFormInterface.STATUS_KO)
			{
				setActiveField(fieldInfo);
				return LyFormInterface.STATUS_KO;
			}
			else if (result == LyFormInterface.STATUS_OK)
			{
				if ((fieldInfo instanceof LyRelationFieldInfo)
						&& ((LyRelationFieldInfo)fieldInfo).getType().equals(LyRelationFieldInfo.Type.COMPOSITION))
					// Value for the fieldInfo may have changed
					fieldInfos.addInfo(fieldInfo);
			}
		}

		// Maybe this controller has been canceled.
		if ((_object == null) && (!isCreation()))
			return LyFormInterface.STATUS_CANCEL;

		if (fieldInfos.getSize() > 0)
		{
			LyValueSet newValues = getValueSet(fieldInfos);

			if (isFreed())
				return LyFormInterface.STATUS_CANCEL;
			if (newValues == null)
				return LyFormInterface.STATUS_KO;

			// Merge old and new value set
			for (i = 0; i < newValues.getSize(); i++)
			{
				if (_setValues == null)
					_setValues = new LyValueSet();
				_setValues.addValue(newValues.getFieldValue(i));
			}


			result = controlNewValues(_setValues);
			if (result == LyFormInterface.STATUS_KO)
				return LyFormInterface.STATUS_KO;
		}

		// Commit edition
		if (!end(_setValues))
			return LyFormInterface.STATUS_KO;

		// FFT01173
		if (!isCreation()
				&& ((_setValues == null) || (_setValues.getSize() == 0)))
			return LyFormInterface.STATUS_IGNORE;

		// <END OF TRANSACTION>

		// Clear values
		_setValues = null;

		return LyFormInterface.STATUS_OK;
	} // end method endValidate

	/**
	 * Method called when all modifications are committed.
	 */
	public void commit()
	{
	    if (isFreed())
	        return;

		_update = false;
		LySetBehaviorInterface behavior = getSetBehavior();

		behavior.startCommit(this);
		// Commit sub-forms
		subFormsCommit();

		// Commit modifications
		if (_controllerContext != null)
		{
			LyApplication application = getApplication();
			LyEnvironment environment = getEnvironment();

			if ((_relationParent == null)
			    && (_controllerContext.getErrorLevel() == LyObject.LEVEL_ERROR))
			{
				String cause = _controllerContext.getErrorMessage();
				String message = null;

				if (cause == null)
				{
					Vector<LyError> errors = _controllerContext.getErrors(LyObject.LEVEL_ERROR);

					if ((errors != null)
						&& (errors.size() > 0))
					{
						LyError  error = errors.elementAt(0);
						Object[] args = error.getArgs();
						LyObject object = error.getObjectCause();

						cause = environment.getMessage(error.getMessage(), args);

						if ((_relationParent != null)
						        && (object != null)
								&& (object != _object))
						    showError(object, error);
					}
					else
					{
				        LyController  grandParent = getParent().getParent();

						cause = application.getError();
			              if (message == null)
			                    message = environment.getMessage("LY_ER_COMMIT_CONTEXT", cause);
			            grandParent.showError(message);
					}
				}
			}
			else
			{
                LyContext.Status status = _controllerContext.commit();

				if (status == LyContext.Status.ROLLBACKED)
				{
					String cause = _controllerContext.getErrorMessage();
					String message;

					if (cause == null)
						cause = application.getError();
					message = environment.getMessage("LY_ER_COMMIT_CONTEXT", cause);
					showError(message);
				}
				else if ((_relationParent != null)
				        || ((status != LyContext.Status.ERROR)
				                && (_controllerContext.getErrorLevel() != LyObject.LEVEL_ERROR)))
				{
					_controllerContext.setListener(null);
					_controllerContext = null;
				}
			}
		}
		behavior.endCommit(this);

		_update = true;
	} // end method commit

	/**
	 * Method called when the form is cancelled.
	 */
	public void cancel()
	{
		subFormsCancel();

		// Rollback modifications
		if (_controllerContext != null)
		{
			_controllerContext.rollback();
			_controllerContext = null;
		}
	}

	/**
	 * End of the current action. The modifications are completed, and the application may take
	 * control and continue or give up action. If false, the process od closing the window aborts.
	 *
	 * <P>The method returns LyClassBehavior.endSet() (LySetBehaviorInterface)</P>
	 *
	 * @param  values Values returned by the controlled form.
	 * @return True if the action may end, false otherwise.
	 */
	protected boolean end(LyValueSet values)
	{
		LySetBehaviorInterface behavior = getSetBehavior();
		boolean				   result;

		_ending = true;
		result = behavior.endSet(this, values);
		updateStatus();
		_ending = false;
		return result;
	}

	/**
	 * Internal use. When an object is created in a popup creation form for a relation, the child
	 * controller invokes this method to set the newly created object as the current value of the
	 * relation (if a new object is created we automatically select this value). This method is
	 * never used by the application : use setFieldValue instead
	 *
	 * @param object        newly created object. The new value of the relation fieldInfo.
	 * @param relationField the relation
	 */
	public void notifyFormResult(LyObject object, LyRelationFieldInfo relationField)
	{
		LyFieldController fieldController = getFieldController(relationField);

		if (fieldController instanceof LyFormResultInterface)
			((LyFormResultInterface)fieldController).notifyFormResult(object, relationField);

		// Update the state of the field when update from a popup creation.
		valueChanged(relationField);
	}

	/**
	 * Factory of the default builder for a controller. Overriden to create an instance of
	 * LySetBuilder
	 *
	 * @return the builder which creates the XML description of the view
	 * @see    leon.control.builder.LySetBuilder
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return (new LySetBuilder(this));
	}

	/**
	 * Accessor to controler behavior : shortcut to (LySetBehaviorInterface)getBehavior().
	 *
	 * @return associated behavior
	 * @see    LyAreaController#getBehavior
	 */
	public LySetBehaviorInterface getSetBehavior()
	{
		return (LySetBehaviorInterface)getBehavior();
	}

	/**
	 * Set boolean indicating if the controlled form is editable or not. Current implementation is *not
	 * * dynamic. This method must be called before showing the view.
	 *
	 * @param editable new editable status.
	 */
	public void setEditable(boolean editable)
	{
		if (_initialized
				&& (editable == _editable))
			return;

		_editable = editable;

		if (_component == null)
			return;

		boolean			  enabled = editable && ((_object != null) || (_action.getMin() <= 0));

		// Change field controllers
		LyClassInfo		  classInfo = getClassInfo();
		int				  i, n = classInfo.getFieldInfoCount();
		LyFieldInfo		  fieldInfo;
		LyFieldController fieldController;
		LySetBehaviorInterface behavior = getSetBehavior();

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			fieldController = getFieldController(fieldInfo);
			if (fieldController instanceof LyFieldSetController)
			{
			    if (editable && behavior.isEditable(this, fieldInfo))
			        ((LyFieldSetController)fieldController).setEditable(editable);
			    else
			        ((LyFieldSetController)fieldController).setEditable(false);
			}
		}

		// Enable/disable apply and validate buttons
		LyComponent root = getComponent();
		LyCommand   validate = (LyCommand)root.getComponent(DEMAND_VALIDATE);
		LyCommand   apply = (LyCommand)root.getComponent(DEMAND_APPLY);

		if (validate != null)
			validate.setEnabled(enabled);
		if (apply != null)
			apply.setEnabled(enabled);

		reset(false);
	} // end method setEditable

	/**
	 * Set boolean indicating if the controlled form is enabled or not.
	 *
	 * @param enabled new enabled status.
	 */
	public void setEnabled(boolean enabled)
	{
		setEditable(enabled); // XXX
	}

	/**
	 * Boolean indicating if the controlled form is editable or not.
	 *
	 * @return true if the form is editable, false otherwise
	 */
	public boolean isEditable()
	{
		return _editable;
	}

	/**
	 * Set the fast edited field info.<br />
	 * If a field info is set for fast edition, it will be displayed in the form with fields
	 * depending on this field. This method must be called before the view is built.
	 *
	 * @param  fieldInfo the field info that must be edited in fast edition mode.
	 * @return true if the field may be edited in fast edition mode, false otherwise.
	 */
	public boolean setFastEditedField(LyFieldInfo fieldInfo)
	{
		LyClassInfo classInfo = getClassInfo();

		fieldInfo = classInfo.getRealFieldInfo(fieldInfo, false);

		if (!isDisplayed(fieldInfo))
			return false;

		// Force the build process by calling initView on view controller
		getViewController().initView();

		if (!isEditable(fieldInfo)
				|| !isEnabled(fieldInfo))
		{
			_fastEditedFields = null;
			return false;
		}

		LyInfoList<LyFieldInfo> fields = new LyInfoList<LyFieldInfo>();

		fields.addInfo(fieldInfo);

		_fastEditedFields = fields;
		return true;
	}

	/**
	 * Indicates if fast edition of a field info is allowed.
	 *
	 * @return true if fast edition of a field info is allowed, false otherwise.
	 */
	protected boolean allowFastEdition()
	{
		return false;
	}

	/**
	 * Gets the list of fields being displayed in the form.
	 *
	 * @return Fast edited field infos if set, otherwise return the result of the super method.
	 * @see    leon.control.LyFormController#getFields()
	 */
	public LyInfoList<LyFieldInfo> getFields()
	{
		if (_fastEditedFields != null)
			return _fastEditedFields;

		return super.getFields();
	}

	/**
	 * Accessor to the internal value associated to a filedInfo. This is a shortcut to
	 * getFieldValue(fieldInfo).getValue()
	 *
	 * @param fieldInfo the fieldInfo
	 * @return the value if found and set, null otherwise.
	 */
	public Object getValue(LyFieldInfo fieldInfo)
	{
		LyValue value = getFieldValue(fieldInfo);

		return ((value == null) ? null : value.getValue());
	}

	/**
	 * Sets current object being modified.
	 *
	 * <P>If changes have been made, this method asks the user for confirmation that changes on
	 * former object will be lost.</P>
	 *
	 * <P>The the method updates all fields for the new object.</P>
	 *
	 * @param  object new object being modified.
	 * @return true if the object change was successful, false otherwise.
	 * @see    #questionAnswer
	 */
	public boolean setObject(LyObject object)
	{
        if (object != null)
        {
            if (object == _object)
                return false;

            if (_initialized
                    && !isApplied()
                    && !allowChange(object, true))
                return false;
        }

    	if (isCreation())
    		// No lock on creation
    		return super.setObject(object);

    	// First check if object is not already locked
		LyObject			   previousObject = _object;
		LySetBehaviorInterface behavior = getSetBehavior();
		String				   lockMessage = behavior.isLocked(this, object);

		if (lockMessage == null)
		{
			if (!super.setObject(object))
				return false;

			// Lock new object
			if (!lockObject(object, true))
			{
				setEditable(false);
				return false;
			}

			if (previousObject != null)
				// Unlock previous object
				lockObject(previousObject, false);

			// Enable edition
			setEditable(true);
			return true;
		}

		// Set object
    	if (!super.setObject(object))
    		return false;

    	// Inhibits edition
    	setEditable(false);

    	// Show messages
		boolean		unlockByUser = testParameter("_unlockByUser");

		if (!unlockByUser)
		{
			// Just notify lock and give up
			showError(lockMessage);
		}
		else
		{
    		// Ask to end-user if he wants to go on
            LyViewManager	viewManager = getViewManager();
    		LyEnvironment	environment	= getEnvironment();

    		lockMessage += "\n" + environment.getMessage("LY_CONFIRM_UNLOCK");
			viewManager.askQuestion(getView(), lockMessage, this,
				(previousObject != null ? previousObject : object));
		}

    	return true;
	} // end method setObject

	/**
	 * Answer to a askQuestion request. Current implementation proceed the setObject method after
	 * user confirmation that changes may be lost.
	 *
	 * @param result true if the user has clicked on Yes or Ok
	 * @param clientData data specified in the askQuestion method
	 * @see   leon.view.setinterface.LyQuestionAnswerInterface#questionAnswer
	 * @see   #setObject
	 */

	public void questionAnswer(boolean result, Object clientData)
	{
        if (clientData instanceof LyObject)
        {
        	if (!result)
				// Disable edition
				setEditable(false);
        	else
        	{
				// Lock new object
				if (!lockObject(_object, true))
					setEditable(false);
				else
				{
		        	LyObject	previousObject = (LyObject)clientData;

					if (previousObject != _object)
						// Unlock previous object
						lockObject(previousObject, false);

					// Enable edition
					setEditable(true);
				}
        	}

			return;
        }

		if (result)
		{
			if (clientData == null)
			{
				free();
				return;
			}
            else if (clientData instanceof LyController)
            {
                ((LyController)clientData).free();
                return;
            }

			// Cancel current edition
			cancel();
			_update = true;

			// Update object
			Vector<?> v = (Vector<?>)clientData;
			LyObject    newObject = (LyObject)v.get(0);
			LyWorkSpace context = (LyWorkSpace)v.get(1);

			super.setObject(newObject);

			if (context != getContext())
				setContext(context);
		}
	}

	/**
	 * Checks if the specified fieldInfo is currently being shown in the form.
	 *
	 * @param  fieldInfo the fieldInfo we want to find
	 * @return true if the fieldInfo is present (modification/consultation), false otherwise.
	 */
	public boolean isDisplayed(LyFieldInfo fieldInfo)
	{
		LyClassInfo classInfo = getClassInfo();

		if (!classInfo.contains(fieldInfo))
			return false;

		boolean				    isConsult = isConsult(fieldInfo);
		LyInfoList<LyFieldInfo> fields = getFields();
		boolean				    result = ((fields == null)
					|| fields.contains(fieldInfo)
					|| isConsult);

		if (result)
		{
			LyFormBehaviorInterface behavior = (LyFormBehaviorInterface)getBehavior();

			if (behavior != null)
				result = result & behavior.isDisplayed(this, fieldInfo);
		}

		return result;
	} // end method isDisplayed

	/**
	 * Checks if the specified fieldInfo is currently being shown as a 'read-only' field in the
	 * form.
	 *
	 * @param  fieldInfo the fieldInfo we want to find
	 * @return true if the fieldInfo is present (consultation), false otherwise.
	 */
	public boolean isConsult(LyFieldInfo fieldInfo)
	{
		boolean isSet = super.isDisplayed(fieldInfo);

		if (!isSet)
		{
		    BitSet consultMarks = getParameterMarks("_consultMarks", LyBitSet.Type.F_MARKS);
            int    consultMark = (isCreation() ? LyFieldInfo.MARK_CREATE_CONSULT : LyFieldInfo.MARK_SET_CONSULT);

		    if (consultMarks != null)
		    {
                if (consultMarks.get(LyFieldInfo.MARK_CONSULT))
                    consultMarks.set(consultMark);

    		    if (consultMarks.get(consultMark))
    		        if (hasMark(fieldInfo, consultMark))
    		            return true;
    		        else
    		            consultMarks.clear(consultMark);

    			if ((consultMarks.size() > 0)
    			        && fieldInfo.hasMarks(consultMarks, false))
    			    return true;
		    }

			return false;
		}
		else
		    return false;
	} // end method isConsult

	/**
	 * Returns a boolean indicating if the action is or not a creation.
	 *
	 * @return true if the form is a creation form. Always return FALSE.
	 */
	public boolean isCreation()
	{
		return false;
	}

	/**
	 * This method may be used at the end of the action to know if the user pushed Apply or Validate
	 * button.
	 *
	 * @return Returns true if Apply was requested, false otherwise.
	 */
	public boolean isApplied()
	{
        if (_apply)
            return true;
        if (_parent instanceof LySetController)
            return ((LySetController)_parent).isApplied();
        else
            return false;
	}

	/* Utility methods */

	/**
	 * Clears specified fieldInfo value : setting value to null.
	 *
	 * @param fieldInfo fieldInfo which is being cleared
	 */
	public void clear(LyFieldInfo fieldInfo)
	{
		getEnvironment().logTrace("\tclear " + fieldInfo);

		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController instanceof LyFieldSetController)
			((LyFieldSetController)fieldController).clear();
	}

	/**
	 * Resets specified fieldInfo value : the default value is fetched from the associated behavior.
	 * LyClassBehavior.getValue (LySetBehaviorInterface) The field is restored to its original value
	 * when the view was opened.
	 *
	 * @param fieldInfo fieldInfo which is being restored
	 */
	public void reset(LyFieldInfo fieldInfo)
	{
		getEnvironment().logTrace("\treset " + fieldInfo);

		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController instanceof LyFieldSetController)
			((LyFieldSetController)fieldController).reset();
	}

	/**
	 * Utility method (internal use for behavior) to calculate formulas of fieldInfos.
	 * Application should use setFieldValue instead.
	 *
	 * @param fieldInfo the field which was modified
	 * @param origin    the field that propagates the modification
	 * @param propagate Indicates whether the change is notified (propagated) to other fields or
	 *                  not.
	 * @see   #setFieldValue
	 */
	public void update(LyFieldInfo fieldInfo, LyFieldInfo origin, boolean propagate)
	{
		getEnvironment().logTrace("\tupdate " + fieldInfo);

		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController instanceof LyFieldConsultController)
			((LyFieldConsultController)fieldController).update(origin, propagate, _valueChange);
	}

	/**
	 * Changes current field status to (non-)optional. An optional field may accept an empty (null)
	 * value and its label is not shown as bold.
	 *
	 * @param fieldInfo the field which should be marked optional
	 * @param optional new fieldInfo status. (default is from the view description)
	 */
	public void setOptional(LyFieldInfo fieldInfo, boolean optional)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController instanceof LyFieldSetController)
		{
			if (!optional
					&& (getValue(fieldInfo) == null))
				checkFieldController(fieldController);

			((LyFieldSetController)fieldController).setOptional(optional);
		}
	}

	/**
	 * Accessor to the optional status for a field in the form.
	 *
	 * @param  fieldInfo the specified field
	 * @return current field status
	 */
	public boolean isOptional(LyFieldInfo fieldInfo)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController instanceof LyFieldSetController)
			return ((LyFieldSetController)fieldController).isOptional();
		else
		{
			// XXX What is it for ?
			LySetBehaviorInterface behavior = getSetBehavior();

			return behavior.isOptional(this, fieldInfo);
		}
	}

	/**
	 * Changes current field status to (non-)editable. If the field is not editable, the value is
	 * read-only but however submited with all other values. A disabled field is not submited with
	 * all other values and is just ignored.
	 *
	 * @param fieldInfo the field which should be marked editable
	 * @param editable new fieldInfo status. (default is true or obtained from the view
	 *                  description)
	 */
	public void setEditable(LyFieldInfo fieldInfo, boolean editable)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController instanceof LyFieldSetController)
			((LyFieldSetController)fieldController).setEditable(editable);
	}

	/**
	 * Accessor to the editable/ non editable status for a field in the form.
	 *
	 * @param  fieldInfo the specified field
	 * @return current field status
	 */
	public boolean isEditable(LyFieldInfo fieldInfo)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		return ((fieldController instanceof LyFieldSetController)
				&& ((LyFieldSetController)fieldController).isEditable());
	}

	/**
	 * Changes current field status to (non-)enabled.
	 * A non-enabled field is not submitted with all other values and it's value is ignored even if it was modified
	 * during the edition.
	 *
	 * @param fieldInfo the field which should be marked enabled
	 * @param enabled  new fieldInfo status. (default is true or obtained from the view
	 *                  description)
	 */
	public void setEnabled(LyFieldInfo fieldInfo, boolean enabled)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController instanceof LyFieldSetController)
			((LyFieldSetController)fieldController).setEnabled(enabled);
	}

	/**
	 * Accessor to the enabled/disabled status for a field in the form.
	 *
	 * @param  fieldInfo the specified field
	 * @return current field status
	 */
	public boolean isEnabled(LyFieldInfo fieldInfo)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController instanceof LyFieldSetController)
			return ((LyFieldSetController)fieldController).isEnabled();
		else
		{
			LySetBehaviorInterface behavior = getSetBehavior();

			return behavior.isEnabled(this, fieldInfo);
		}
	}

	/**
	 * Changes current field status to (non-)hidden. If the field is hidden, it is not submited with
	 * all other values. When the field is set hidden, it is also set enabled.
	 *
	 * @param fieldInfo the field which should be set hidden
	 * @param hidden    new fieldInfo visibility. (default is true)
	 */
	public void setHidden(LyFieldInfo fieldInfo, boolean hidden)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController instanceof LyFieldConsultController)
			((LyFieldConsultController)fieldController).setHidden(hidden);

	    String group = fieldInfo.getGroup();

        if ((group != null) && (group.length() > 0))
        {
            LyFormField formField = getFormField(fieldInfo);
            LyForm      form = (LyForm)formField.getParent();
            LyFormField field;
            int         i, n = (form == null ? 0 : form.getFormFieldCount());
            boolean     visible = false;

            for (i = 0; (i < n) && !visible; i++)
            {
                field = form.getFormField(i);
                visible = !field.isHidden();
            }

            if ((form != null)
                && (form.getParent() != null))
            {
                if (form.getParent() instanceof LyArea)
                    ((LyArea)form.getParent()).setHidden(!visible);

                form.getParent().setVisible(visible);
            }
        }

		String tabLabel = fieldInfo.getTab();

		if ((tabLabel == null) || (tabLabel.length() == 0))
			return;

		// If a least one field in the current tab is visible, the tab is enabled
		// otherwise, it is disabled...
		LyFieldInfo   field, firstVisibleField = null;
		boolean		  visible = false;
        int           i, n = _classInfo.getFieldInfoCount();

		for (i = 0; (i < n) && !visible; i++)
		{
            field = _classInfo.getFieldInfo(i);
            if (!isDisplayed(field))
                continue;

			if ((field.getTab() == null)
			    || !field.getTab().equals(tabLabel))
			{
				if ((firstVisibleField == null)
				       && (field.getTab() != null))
					firstVisibleField = field;
			}
			else
			    visible = !isHidden(field);
		}

		LyFormField formField = getFormField(fieldInfo);
		LyTabs	    tabs;
		LyTab	    tab;
		LyComponent component = getComponent();

		tabs = formField.getParentTabs(component);
		if (tabs != null)
		{
			for (i = 0, n = tabs.getTabCount(); i < n; i++)
			{
				tab = tabs.getTab(i);
				if ((tab.getLabel() != null)
						&& tab.getLabel().equals(tabLabel))
				{
					tab.setEnabled(visible);
					if (!visible
							&& (tabs.getSelectedTab() != null)
							&& tabs.getSelectedTab().equals(tabLabel)
							&& (firstVisibleField != null))
						tabs.setSelectedTab(firstVisibleField.getTab());
					break;
				}
			}
		}
	} // end method setHidden

	/**
	 * Accessor to the hidden status of a field in the form.
	 *
	 * @param  fieldInfo the specified field
	 * @return current field visibility
	 */
	public boolean isHidden(LyFieldInfo fieldInfo)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		return ((fieldController instanceof LyFieldSetController)
				&& ((LyFieldSetController)fieldController).isHidden());
	}

	/**
	 * Changes the label for the specified fieldinfo.
	 *
	 * @param fieldInfo the fieldInfo
	 * @param label the new label
	 * @param labelSet indicates if the label apply to set form
	 */
	public void setLabel(LyFieldInfo fieldInfo, LyInfoLabel label, boolean labelSet)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController instanceof LyFieldSetController)
			((LyFieldSetController)fieldController).setLabel(label, labelSet);
	}

	/**
	 * Sets the valid status of a field in the form.
	 *
	 * @param fieldInfo the specified field
	 * @param valid     current validity status
	 */
	public void setValid(LyFieldInfo fieldInfo, boolean valid)
	{
		setState(fieldInfo, (valid ? LyFormField.STATE_NORMAL : LyFormField.STATE_ERROR));
	}

	/**
	 * Accessor to the valid status for a field in the form.
	 *
	 * @param  fieldInfo the specified field
	 * @return current field status
	 */
	public boolean isValid(LyFieldInfo fieldInfo)
	{
		return (getState(fieldInfo) != LyFormField.STATE_ERROR);
	}

	/**
	 * Get object value to put it in the form.
	 *
	 * @param  fieldInfo FieldInfo associated to the requested value
	 * @return Value of the reqested field info.
	 */
	protected Object getFormValue(LyFieldInfo fieldInfo)
	{
		Object    value = super.getFormValue(fieldInfo);

		if ((value == null)
				&& !_valueChange)
			value = getDefaultValue(fieldInfo);

		// FFT00785
		LyFieldController fieldController = getFieldController(fieldInfo);
		LyUnit			  unit = null;

		if ((fieldController != null) && (fieldController instanceof LyFieldSetController))
		{
			LyFormField formField = fieldController.getFormField();

			if (formField != null)
				unit = ((LyFieldSetController)fieldController).getUnit();

			if (unit != null)
			{
				if ((fieldInfo instanceof LyStructFieldInfo)
						&& (((LyStructFieldInfo)fieldInfo).fromFieldTag())
						&& (((LyStructFieldInfo)fieldInfo).getClassInfoCount() > 1)
						&& !(((LyStructFieldInfo)fieldInfo).isMultiple()))
				{
					LyClassInfo		  subStructClass =
						((LyCompositionFieldUnitSetController)fieldController).getFieldType();
					LyStructFieldInfo subStruct = (LyStructFieldInfo)subStructClass.getParent();

					value = subStruct.convertToFieldUnit(value, unit);
				}
				else
					value = fieldInfo.convertToFieldUnit(value, unit);
			}
		}

		return value;
	} // end method getFormValue

	/**
	 * Internal method to get default value for a field info.
	 *
	 * @param  fieldInfo Field info for which default value is requested.
	 * @return The default value for the given field.
	 */
	protected Object getDefaultValue(LyFieldInfo fieldInfo)
	{
		if (!isEnabled(fieldInfo) || isOptional(fieldInfo))
			return null;

		LySetBehaviorInterface behavior = (LySetBehaviorInterface)getBehavior();

		return behavior.getDefaultValue(this, fieldInfo);
	}

	/**
	 * Clears all cached data for this controller (especially the view and the component).
	 *
	 * <P>This method may be invoked for memory optimization, but the controller is still valid and
	 * may be used. (Unlike free) Further usage, needs to rebuild internal component & view which
	 * may be time consuming.</P>
	 */
	public void clearViewCache()
	{
		// Clear field controllers
		if (_fieldControllers != null)
		{
		    Enumeration<LyFieldController> elements = _fieldControllers.elements();
		    LyFieldController              fieldController;

		    while (elements.hasMoreElements())
		    {
		        fieldController = elements.nextElement();
				fieldController.clearViewCache();
			}
		}

		super.clearViewCache();
	}

	/**
	 * Free this controller on request of the specified controller.
	 *
	 * @param controller controller instance which requested the freeing of this controller. If the
	 *                   controller is the parent, this controller is freed with all its
	 *                   sub-controllers (relationControllers) otherwise the method does nothing.
	 */
	public void free(LyController controller)
	{
		if (_parent != controller)
			return;

		// Don't update anymore
		_valueChanging = null;
		setIdle(true);
		_update = false;

		// Cancel edition
		cancel();

		// Unlock object
		lockObject(_object, false);

		_relationParent = null;
		_compositionRelation = null;
		_errors = null;
		_formSave = null;

		// Delete Temporary file path if needed
		deleteTempFiles();

		super.free(controller);
	}

	/**
	 * Reset the context component (showing the objectList).
	 *
	 * @see #getObjectList()
	 */
	protected void resetObjects()
	{
		if (!_initialized)
			return;

		setEditable(_editable);

		super.resetObjects();
	}

	/**
	 * Called when an object has changed in the dynamic object list. This method is overriden for
	 * tabactions
	 *
	 * @param  event event of changing object
	 * @return the event is used by the listener
	 */
	public boolean objectChanged(LyWorkSpaceEvent event)
	{
		if (!_update)
			return true;

		boolean status;
		boolean	valueChange = _valueChange;

		_valueChange = true;
		status = super.objectChanged(event);
		_valueChange = valueChange;

		// Handle conflict when user has changed data in forms and object is modified.
		// Changing tab must not ignore objectChanged
		_last_object_changed = getLastModified();

		return status;
	}

	/**
	 * Get the state of the set controller.
	 *
	 * @return current state
	 */
	public short getState()
	{
		// Get max state
		Enumeration<LyFieldController> elements = _fieldControllers.elements();
		LyFieldController              fieldController;
		short			               state = LyClassBehavior.STATUS_OK, fieldState;

		while (elements.hasMoreElements())
		{
			fieldController = elements.nextElement();
			if (fieldController instanceof LyFieldSetController)
			{
				fieldState = ((LyFieldSetController)fieldController).getState();
				if ((fieldState != LyClassBehavior.STATUS_OK)
						&& ((state == LyClassBehavior.STATUS_OK)
							|| (fieldState < state)))
					state = fieldState;
			}
		}

		return state;
	}

	/**
	 * Internal use. Gets the classInfo for the specified relation or classInfoId. This method is
	 * used to managed the possible use cases of access to the classInfo of a relationFieldInfo.
	 *
	 * @param  classInfoId current classInfo Identifier. May be null, and the value is fetched from
	 *                     the relationId (eventually using a possible CLASS_FIELD info in the
	 *                     meta-model)
	 * @param  relationId  current field (relation field) associated to the request. This value may
	 *                     be null if classInfoId is *not* null.
	 * @return the classInfo of objects currenlty being shown in the relation field info.
	 */
	protected LyClassInfo getClassInfo(String classInfoId, String relationId)
	{
		if ((relationId == null)
				|| ((classInfoId != null)
					&& !classInfoId.equals(LyFieldSetController.STANDARD_CLASSINFO)))
			return super.getClassInfo(classInfoId, relationId);

		LyApplication	    application = getApplication();
		LyRelationFieldInfo relation = (LyRelationFieldInfo)application.getInfo(relationId);

		if (relation == null)
			return super.getClassInfo(classInfoId, null);

		LySetBehaviorInterface  behavior = getSetBehavior();
		LyInfoList<LyClassInfo> classes = behavior.getRelationClasses(this, relation);

		if ((classes == null) || (classes.getSize() == 0))
			return null;

		return classes.getInfo(0);
	}

	/**
	 * Get current unit for the specified fieldInfo in the form.
	 *
	 * @param  fieldInfo fieldInfo associated to the requested unit relation
	 * @return current unit, or null if not found or field is disabled.
	 */
	public LyUnit getUnit(LyFieldInfo fieldInfo)
	{
		LyFieldController fieldController = getFieldController(fieldInfo);

		if (fieldController instanceof LyFieldSetController)
			return ((LyFieldSetController)fieldController).getUnit();
		else
			return null;
	}

	/**
	 * Reset fieldInfo errors on _object for fieldInfo.
	 *
	 * @param fieldInfo
	 */
	public void resetFieldInfoErrors(LyFieldInfo fieldInfo)
	{
		// Reset valid state
		setState(fieldInfo, LyFormField.STATE_NORMAL);

		if (_object != null)
			_object.resetFieldInfoErrors(fieldInfo);
	}

	/**
	 * Reset class errors on _object.
	 */
	protected void resetClassErrors()
	{
        if (_controllerContext != null)
        {
            Vector<LyObject>  objects = _controllerContext.getObjectsInError();

            if (objects != null)
            {
                int         i, n = objects.size();
                LyObject    object;
                int         j, nr;

                for (i = 0; i < n; i++)
                {
                    object = objects.elementAt(i);
                    for (j = 0, nr = object.getErrorsCount(); j < nr; j++)
                        object.removeError(j);
                }
            }
        }

		if (_object != null)
			_object.resetClassErrors();
	}

	/**
	 * Gets the controller context.
	 *
	 * @return The controller context.
	 */
	public LyContext getControllerContext()
	{
		if ((_controllerContext == null)
			|| (_controllerContext.isEnded()))
		{
			LySession		 session = getSession();
			LyContextManager contextManager = session.getContextManager();

			_controllerContext = contextManager.newContext(session, LyContext.Type.USER_CANCELLABLE);
			_controllerContext.setListener(this);

			LyController parent = _parent;

			if (parent instanceof LyDialogController)
				parent = parent.getParent();
			else if (parent instanceof LyFieldSetController)
			{
				_controllerContext.setInternal();
				parent = parent.getParent();
			}

			// Last try to get a root level context controller (FFT01632)
			if (!(parent instanceof LyControllerContextInterface))
				parent = getSession().getRootContextController();

			LySetBehaviorInterface behavior = getSetBehavior();
			LyRelationFieldInfo parentRelation = (_compositionRelation != null ? _compositionRelation : _relationParent);

			if ((parent instanceof LyControllerContextInterface)
				&& !behavior.useSingleContext(this, parentRelation))
			{
				boolean  singleContext = (_compositionRelation == null)
    				&& ((_relationParent == null)
    					|| (_relationParent.getType() != LyRelationFieldInfo.Type.COMPOSITION))
    				&& !(_parent instanceof LyRelationTableController)
    				&& !(_parent instanceof LyCompositionFieldSetController)
    				&& !(_parent instanceof LyTabAreaController);

				if (!singleContext)
				{
    				// Be careful, the parent can return null even if it implements the interface
    				LyContext context;

                    if (parent instanceof LyControllerContextInterface2)
                        context = ((LyControllerContextInterface2)parent).getControllerContext(this);
                    else
                        context = ((LyControllerContextInterface)parent).getControllerContext();

                    if (context != null)
                        context.addSon(_controllerContext);
                }
            }

			_controllerContext.setName(getTitle());
		}

		return _controllerContext;
	} // end method getControllerContext

	/**
	 * Get the model of the current form
	 *
	 * @return the model object
	 */
	public LyObject getModel()
	{
		return _object;
	}

	/**
	 * Sets composition relation : this method is invoked when this form manages a relation in
	 * another form.
	 *
	 */
	public void setContextInternal()
	{
		LyContext context = getControllerContext();

		context.setInternal();
	}

	String getContextName()
	{
		return getTitle();
	}

	/**
	 * Controls current edition.
	 *
	 * @return The error status.
	 */
	public short control()
	{
		// Retrieve values
		LyValueSet setValues = getValueSet(null);

		if (isFreed())
			return LyFormInterface.STATUS_CANCEL;

		short status = STATUS_OK;

		if (setValues == null)
			// At least an error occurred
			status = LyFormInterface.STATUS_KO;
		else
		{
			// Add some extra values of the behavior.
			LySetBehaviorInterface behavior = getSetBehavior();

			behavior.addExtraValues(this, setValues);

			// No error occured, control values all together
			status = controlNewValues(setValues);
		}

		LyClassInfo		  classInfo = getClassInfo();
		int				  i, n = classInfo.getFieldInfoCount();
		LyFieldInfo		  fieldInfo;
		LyFieldController fieldController;
		short			  fieldStatus;

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			fieldController = getFieldController(fieldInfo);
			if ((fieldController instanceof LyFieldSetController)
					&& ((LyFieldSetController)fieldController).checkIfModified())
			{
				fieldStatus = ((LyFieldSetController)fieldController).control();
				if ((status == STATUS_OK)
						|| ((fieldStatus != STATUS_OK) && (fieldStatus < status)))
					status = fieldStatus;
			}
		}

		if (status != STATUS_OK)
			showErrors();
		return status;
	} // end method control

	/**
	 * Method called when a context is ended by a brother context.
	 *
	 * @param context Context ended.
	 * @param brother Brother context that ended the given context.
	 */
	public void contextEnded(LyContext context, LyContext brother)
	{
	}

	/**
	 * Check (when it makes sense) the graphical components to see if they have a pending value
	 * typed by the user, and not yet refreshed in the LyView components. Like a textfield, which
	 * value may have changed in the graphical component and this modification has not been
	 * notified. This method may do nothing, depending on the nature of the graphical component.
	 */
	public void updateViewContent()
	{
		// Start validation of sub-forms
		LyClassInfo		  classInfo = getClassInfo();
		int				  i, n = classInfo.getFieldInfoCount();
		LyFieldInfo		  fieldInfo;
		LyFieldController fieldController;
		LyValue			  valueBefore, valueAfter, valueInObject;

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			fieldController = getFieldController(fieldInfo);
			if (!(fieldController instanceof LyFieldSetController))
				continue;

			// Be sure There is no pending values due to a possible focus lost not yet notified
			valueBefore = getFieldValue(fieldInfo);
			((LyFieldSetController)fieldController).updateViewContent();
			valueAfter = getFieldValue(fieldInfo);

			// Keep internal object consistent
			if ((valueAfter != null)
					&& (valueBefore != null)
					&& !valueAfter.equals(valueBefore))
				valueChanged(fieldInfo); // Something may have changed, but usually does nothing

			if ((_object != null)
					&& _object.getClassInfo().contains(fieldInfo))
			{
				// Be sure our internal value is consistent
				valueInObject = _object.getFieldValue(fieldInfo);

				if ((valueAfter != null)
						&& (valueInObject != null)
						&& !valueAfter.equals(valueInObject))
					valueChanged(fieldInfo);
			}
		}
	} // end method updateViewContent

	/**
	 * Get default fields to display in form. This method all fields of the current class info
	 * expect hidden and id fields.
	 *
	 * @return List of fields to be displayed.
	 */
	protected LyInfoList<LyFieldInfo> getDefaultFields()
	{
        LyEnvironment           environment = getEnvironment();

        if (!environment.testResource("LY_DEFAULT_FIELDS_IN_FORMS_WHEN_EMPTY"))
            return null;

        LyInfoList<LyFieldInfo> fields = null;
        int                     i, n = _classInfo.getFieldInfoCount();
        LyFieldInfo             fieldInfo;

        for (i = 0; i < n; i++)
        {
            fieldInfo = _classInfo.getFieldInfo(i);
            if (hasMark(fieldInfo, LyFieldInfo.MARK_HIDDEN))
                // Remove hidden fields
                continue;
            if (fieldInfo.hasMark(LyFieldInfo.MARK_ID))
                // Remove id fields
                continue;
            if (hasMark(fieldInfo, LyFieldInfo.MARK_SET_CONSULT))
                // Remove setConsult fields
                continue;
            if (hasNotMark(fieldInfo, LyFieldInfo.MARK_SET))
                // Remove "notSet" fields
                continue;
            if (fieldInfo.isLocal()
                    && (fieldInfo.getImport() != null))
                // Remove imported fields
                continue;

            if (fields == null)
                fields = new LyInfoList<LyFieldInfo>();
            fields.addInfo(fieldInfo);
        }

        return fields;
	}

	/**
	 * Invokes an action on the specified object. Closes the consult view if a set action is
	 * invoked.
	 * @param action the action to invoke
	 * @param object the selected object
	 * @return the resulting controller
	 */
	protected LyController invokeAction(LyAction action, LyObject object)
	{
		return invokeAction(action, object.getClassInfo(), new LySimpleObjectList(object), true);
	}

	/**
	 * Check if input field corresponding to given controller is present in fast edition mode.
	 *
	 * @param fieldController The field controller.
	 */
	protected void checkFieldController(LyFieldController fieldController)
	{
		if (_fastEditedFields == null)
			return;

		LyFormField formField = fieldController.getFormField();

		if ((formField != null)
				&& (formField.getInternalView(false) == null))
		{
			// In fast edition mode, add formfield in form if not present to allow end-user to fill in mandatory fields
			LyComponent root = getComponent();
			LyComponent form = ((root == null) ? null
											   : root.getComponent("_fastEditionForm", true));

			if (form instanceof LyForm)
			{
				((LyForm)form).addFormField(formField);
				((LyForm)form).resetContent();
			}
		}
	}

	/**
	 * Check if the area controller may be freed. This method may be overriden by subclasses. The
	 * current implementation always returns true
	 *
	 * @param  showMessage indicates if the controller can show an warning message or not
	 * @return true if the free method can be called immediately, false otherwise.
	 */
	public boolean canFree(boolean showMessage)
	{
		if (!isEditable())
			// Edition was locked
    		return true;

        LyEnvironment environment = getEnvironment();

        if (!environment.testResource("LY_CONFIRM_FORM_FREE"))
            return true;

        boolean     multiple = (_areaStackController != null)
            || ((_areaControllers != null) && (_areaControllers.size() > 0));

        if (super.canFree(showMessage && !multiple))
            return true;

        if (showMessage && multiple)
        {
            // Modifications were made (even erroneous), before changing to an other object
            // ask first to the user if he wants to abort his edition.
            String        message;
            LyViewManager viewManager = getViewManager();

            message = environment.getMessage("LY_CONFIRM_COMPLETE_ABORT");

            viewManager.askQuestion(getView(), message, this, getViewController());
        }

        return false;
    }

    /**
     * Check if the area controller may be freed. This method may be overriden by subclasses. The
     * current implementation always returns true
     * @param showMessage indicates if the controller can show an warning message or not
     * @return true if the free method can be called immediately, false otherwise.
     */
    public boolean canClose(boolean showMessage)
    {
        LyEnvironment environment = getEnvironment();

        if (!environment.testResource("LY_CONFIRM_FORM_FREE"))
            return true;

        return super.canClose(showMessage)
            && (_ending || allowChange(null, showMessage));
    }

    /**
     * Check if the form content has been modified and if the object need to be created.
     * @return true if the form was modified
     */
    public boolean checkIfModified()
    {
        return !allowChange(null, false);
    }

    /**
	 * Internal method to check if there was some changes in form before cancelling or changing
	 * edited object.
	 *
	 * @param  newObject new object about to be set in form (null on cancel).
	 * @param highlightChanges indicates if we change the look of modified fields
	 * @return Boolean indicating if change is allowed or not.
	 */
	protected boolean allowChange(LyObject newObject, boolean highlightChanges)
	{
		// Start validation of sub-forms
		LyClassInfo		  classInfo = getClassInfo();
		int				  i, n = classInfo.getFieldInfoCount();
		LyFieldInfo		  fieldInfo;
		LyFieldController fieldController;
		boolean			  modified = false;

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			fieldController = getFieldController(fieldInfo);
			if ((fieldController instanceof LyFieldSetController)
					&& ((LyFieldSetController)fieldController).checkIfModified())
			{
				modified = true;
				if (highlightChanges
						&& isValid(fieldInfo))
					((LyFieldSetController)fieldController).setState(LyFormField.STATE_WARNING);
			}
		}

		if (modified
				&& highlightChanges)
		{
			// Modifications were made (even erroneous), before changing to an other object
			// ask first to the user if he wants to abort his edition.
			LyEnvironment  environment = getEnvironment();
			String		   message;
			LyViewManager  viewManager = getViewManager();
			Vector<Object> clientData = null;

			if (isCreation())
				message = environment.getMessage("LY_CONFIRM_CREATE_ABORT", _classInfo.getName());
			else
				message = environment.getMessage("LY_CONFIRM_SET_ABORT", getName(_object));

			if (newObject != null)
			{
				clientData = new Vector<Object>();
				clientData.add(newObject);
				clientData.add(getContext());
			}

			viewManager.askQuestion(getView(), message, this, clientData);
			return false;
		}

		return !modified;
	} // end method allowChange

	/**
	 * Sets relation parent : this method is invoked when this form is a child of another form
	 * (dynamic popup creation of an object in a relation field).
	 * @param client the client for the resulting operation
	 * @param relation the relation fieldInfo which is updated as the result of the current action
	 *                 (usually dynamic popup creation)
	 */
	public void setRelationParent(LyFormResultInterface client, LyRelationFieldInfo relation)
	{
		_formResultClient = client;
		_relationParent = relation;
	}

    /**
     * Checks if this controller needs update.
     * @return true if the controller needs to be updated
     */
    public boolean needUpdate()
    {
        return _update;
    }

	/**
	 * This method is called by the view controller when this area is shown to update its content.
	 *
	 * @return True if updated, false otherwise.
	 */
	public boolean update()
	{
	 return false;// do not update forms
	}

	/**
	 * Accessor to the _object parameter (with no control) Applications should use setObject(object)
	 * and not invoke this method directly.
	 *
	 * @param object new object being created.
	 */
	public void setObjectInternal(LyObject object)
	{
		_object = object;
	}

	/**
	 * Add a file to the delete on free list
	 * @param path the file path to add to the delete list
	 */
	public void addTempFilePath(String path)
	{
		if (_parent instanceof LySetController)
		{
			((LySetController) _parent).addTempFilePath(path);
			return;
		}

		if (_formResultClient != null)
		{
			if (_formResultClient instanceof LyFieldSetController)
			{
				((LyFieldSetController)_formResultClient).addTempFilePath(path);
				return;
			}
			else if (_formResultClient instanceof LyController)
			{
				LyController parent = ((LyController)_formResultClient).getParent();

				if (parent instanceof LyFieldSetController)
				{
					((LyFieldSetController)parent).addTempFilePath(path);
					return;
				}
				else if (parent instanceof LySetController)
				{
					((LySetController)parent).addTempFilePath(path);
					return;

				}
			}
		}

		// FFT01210
		if (path != null)
		{
			if (_tempFilesPath == null)
				_tempFilesPath = new ArrayList<String>();

			_tempFilesPath.add(path);
		}
	}

	/**
	 * Delete all temporary files
	 */
	protected void deleteTempFiles()
	{
		int size = (_tempFilesPath == null ? 0 : _tempFilesPath.size());

		for (int i=0; i< size; i++)
		{
			String path = _tempFilesPath.get(i);

			try
			{
				File f = new File (path);
				if (f.exists())
					f.delete();
			}
			catch (RuntimeException e)
			{
				if (getEnvironment() != null)
					getEnvironment().writeException(e);
			}
		}
		_tempFilesPath = null;
	}

    /**
     * Save form inputs.
     */
    protected LyFormSaveFile saveForm(boolean free)
    {
        LyFormSave      innerFormSave = save(false);
        LyFormSaveFile  formSave = null;

        if (innerFormSave != null)
        {
            LySession       session = getSession();
            LyObject        user = session.getUser();
            String          userId = (user == null ? null : user.getId());

            if ((userId != null)
                    && (userId.indexOf(leon.data.LyObjectId.CLASS_SEP) > 0))
                userId = userId.substring(userId.indexOf(leon.data.LyObjectId.CLASS_SEP) + 1);

            formSave = new LyFormSaveFile(_action, userId, innerFormSave);

            if (formSave != null)
                getApplication().saveExtraInfo(formSave);
        }

        if (free)
            free();

        return formSave;
    }

    /**
     * Save form inputs.
     */
    public LyFormSave save(boolean full)
    {
        LyClassInfo         classInfo = getClassInfo();
        int                 i, n = classInfo.getFieldInfoCount();
        LyFieldInfo         fieldInfo;
        LyFieldController   fieldController;
        LyAreaController	areaController;
        LyFieldSave         fieldSave;
        Vector<LyFieldSave> fieldSaves = null;
        LyObject            object = getObject();
        LyValue             newValue;

        for (i = 0; i < n; i++)
        {
            fieldInfo = classInfo.getFieldInfo(i);
            fieldController = getFieldController(fieldInfo);

            if (((fieldController == null)
                    || isConsult(fieldInfo))
                && ((object != null)
                    && (object.hasNewValue(fieldInfo)
                        || object.hasPendingValue(fieldInfo))))
            {
                if ((fieldInfo instanceof LyRelationFieldInfo)
                    && (((LyRelationFieldInfo)fieldInfo).getType() == LyRelationFieldInfo.Type.COMPOSITION))
                {
                    // We have to create a field controller to ensure save of the relation
                    if (fieldInfo.isMultiple())
                        fieldController = new LyMultiCompositionFieldSetController(this, (LyRelationFieldInfo)fieldInfo);
                    else
                        fieldController = new LyCompositionFieldSetController(this, (LyRelationFieldInfo)fieldInfo);

                    areaController = fieldController.getAreaController();
                    areaController.getComponent(); // Force construction of component
                    areaController.initialize();

                    fieldSave = fieldController.save(true);
                }
                else
                {
                    newValue = object.getFieldValue(fieldInfo);
                    fieldSave = new LyFieldSave(fieldInfo, newValue.getValue());
                }
            }
            else if (fieldController != null)
                fieldSave = fieldController.save(full);
            else
                fieldSave = null;

            if (fieldSave != null)
            {
                if (fieldSaves == null)
                    fieldSaves = new Vector<LyFieldSave>(1);
                fieldSaves.add(fieldSave);
            }
        }

        LyFormSave  formSave;

        if (_object == null)
            formSave = new LyFormSave(_classInfo, fieldSaves);
        else
            formSave = new LyFormSave(_object, fieldSaves);

        LySetBehaviorInterface behavior = getSetBehavior();

        if (behavior != null)
            behavior.formSaved(this, formSave);
        return formSave;
    }

    /**
     * Restore form inputs from saved file.
     */
    protected boolean restore()
    {
        LyFormSave  formSave = getFormSave();

        if (formSave == null)
            return false;

        restore(formSave);
        showSetErrors();

        if (formSave instanceof LyFormSaveFile)
        {
            LyApplication   application = getApplication();

            application.deleteExtraInfo((LyFormSaveFile)formSave);
        }

        _formSave = null;
        return true;
    }

    /**
     * Restore form inputs.
     */
    public boolean restore(LyFormSave formSave)
    {
        if (formSave == null)
            return false;
        if (formSave._classInfo != _classInfo)
            return false;

        if (!_initialized)
        {
            // Keep form save to restore on initialization
            _formSave = formSave;
            return true;
        }

        Vector<LyFieldSave>     fieldSaves = formSave._fieldSaves;
        int                     n = (fieldSaves == null ? 0 : fieldSaves.size());

        if (n == 0)
            return true;

        if (_component instanceof LyPanel)
            // Freeze the panel
            ((LyPanel)_component).setIdle(true);

        // Update fields
        LyFieldInfo         fieldInfo;
        LyFieldController   fieldController;
        LyFieldSave         fieldSave;
        int                 i;
        LyObject            object = getObject();
        LyContext           context = getControllerContext();
        boolean             set = false;

        for (i = 0; i < n; i++)
        {
            fieldSave = fieldSaves.get(i);
            fieldInfo = fieldSave._fieldInfo;
            fieldController = getFieldController(fieldInfo);

            if ((fieldController == null)
                || isConsult(fieldInfo))
            {
                if ((fieldInfo instanceof LyRelationFieldInfo)
                    && (((LyRelationFieldInfo)fieldInfo).getType() == LyRelationFieldInfo.Type.COMPOSITION))
                {
                    // We have to create a field controller to ensure restore of the relation
                    if (fieldInfo.isMultiple())
                        fieldController = new LyMultiCompositionFieldSetController(this, (LyRelationFieldInfo)fieldInfo);
                    else
                        fieldController = new LyCompositionFieldSetController(this, (LyRelationFieldInfo)fieldInfo);
                    fieldController.restore(fieldSave);
                }

                if ((object != null)
                    && !object.hasNewValue(fieldInfo)
                    && !object.hasPendingValue(fieldInfo))
                {
                    // Restore pending value
                    context.setValue(object, fieldInfo, fieldSave._newValue);
                    set = true;
                }
            }
            else
                fieldController.restore(fieldSave);
        }

        if (set)
            context.setObject(object);

        LySetBehaviorInterface behavior = getSetBehavior();

        if (behavior != null)
            behavior.formRestored(this, formSave);

        if (_component instanceof LyPanel)
            // Unfreeze the panel
            ((LyPanel)_component).setIdle(false);
        return true;
    }

    /**
     * Gets form save associated to this set controller.
     * @return The current form save, can be null.
     */
    public LyFormSave getFormSave()
    {
        if (_formSave != null)
            return _formSave;

        LySession   session = getSession();
        LyObject    user = session.getUser();
        String      userId = (user == null ? null : user.getId());

        if ((userId != null)
                && (userId.indexOf(leon.data.LyObjectId.CLASS_SEP) > 0))
            userId = userId.substring(userId.indexOf(leon.data.LyObjectId.CLASS_SEP) + 1);

        _formSave = LyFormSaveFile.loadSave(_action, _classInfo, _object, userId);
        return _formSave;
    }

    /**
     * Locks/Unlocks given object for edition.
     * @param object Object to be locked/unlocked.
     * @param lock True if lock is required, false for unlock.
     * @return True if successful, false otherwise.
     */
    protected boolean lockObject(LyObject object, boolean lock)
    {
		LySetBehaviorInterface	behavior = (LySetBehaviorInterface)getBehavior();

		// Lock new object
		if (behavior.lockObject(this, object, lock))
			return true;
		else
		{
			LyEnvironment environment = getEnvironment();
			String		  message = environment.getMessage((lock ? "LY_ER_LOCK_FAILED" : "LY_ER_UNLOCK_FAILED"),
				object.getName());

			showError(message);
			return false;
		}
    }

    /**
     * Cancel all changes done in the form.
     */
    public void cancelChanges()
    {
		LyClassInfo		  classInfo = getClassInfo();
		int				  i, n = classInfo.getFieldInfoCount();
		LyFieldInfo		  fieldInfo;
		LyFieldController fieldController;

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			fieldController = getFieldController(fieldInfo);

			if (fieldController instanceof LyFieldSetController)
				((LyFieldSetController)fieldController).reset();
		}
    }
} // end class LySetController
