/*
 * LyShowUrlController.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.control;

import leon.app.*;
import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;

import java.io.*;
import java.net.*;
import java.util.*;


/**
 * This class implements a dialog window that displays a text chosen by the user or the content of a
 * text file. The generated window can have a titleBar and an image. If a file is used, the user can
 * specify a refresh time ( in seconds ). If the time interval is set to zero, there is no refresh.
 */
public class LyShowUrlController extends LyPanelController
{
	/** File Version */
	public static final String __VERSION = "$Id: LyShowUrlController.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** File path, or null if this controller doesn't display a file. */
	protected String _url;

	/** Refreshing interval (in milliseconds). */
	protected int _interval;

	/** Length of file. */
	protected long _fileLength;

	/** Temporary file for console mode. */
	protected File _temporaryFile;

	/**
	 * Specific constructor.
	 *
	 * @param  session   Session for this controller
	 * @param  parent    Parent controller
	 * @param  action    Action
	 * @param classInfo  corresponding classinfo
	 * @param context    current context
	 */
	public LyShowUrlController(LySession session, LyController parent, LyAction action, LyClassInfo classInfo,
		LyObjectList context)
	{
		super(session, parent, action, context);
		initialize(action);
	}

	/**
	 * Initialize the controller for the given action
	 * @param action an action
	 */
	public void initialize(LyAction action)
	{
		String url = getUrl();

		setUrl(url);
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return  Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyShowUrlBehaviorInterface.class;
	}

	/**
	 * Invoked when the controller is destroyed.
	 *
	 * @param  controller  parent controller
	 *
	 * @see    LyAreaController
	 */
	public void free(LyController controller)
	{
		super.free(controller);
	}

	/**
	 * Sets the url to display in the text window.
	 *
	 * @param   request     Url to display
	 * @param   fieldLabel  Label of the text field
	 * @param   nCols       Number of columns of the text field
	 * @param   nRows       Number of rows of the text field
	 *
	 * @return  the text view component
	 */
	public LyText setUrl(String request, String fieldLabel, int nCols, int nRows)
	{
		LyText text = null;

		_url = request;

		LyEnvironment environment = getEnvironment();
		LyViewManager viewManager = getViewManager();

		if (viewManager.hasProperty("MANAGE_URL"))
			return null;

		String externalViewer = _environment.getFile("LY_EXTERNAL_VIEWER");

		if ((externalViewer != null)
				&& !externalViewer.equals("LY_EXTERNAL_VIEWER"))
			return null;

		if (request == null)
		{
			text = setText(fieldLabel, null, nCols, nRows, null);
			return null;
		}

		String content = getFileContent(request, environment);

		text = setText(fieldLabel, content, nCols, nRows, null);

		return text;
	} // end method setUrl

	/**
	 * Gets the URL from the file name.<br />
	 * Tries first to create the URL from the String.<br />
	 * If an exception occurs, tries to create the URL from a file.<br />
	 *
	 * <p>If a url is found, encode the URL name thaks to the method
	 * {@link LyString#getXmlString(String, boolean, boolean)}.</p>
	 *
	 * @param   fileName     the file whose URL is required
	 * @param   environment  the current environment
	 * @param   showError    indicates if an error message has to be shown to the user.
	 *
	 * @return  the XML encoded file URL if found, null otherwise.
	 */
	public URL getFileUrl(String fileName, LyEnvironment environment, boolean showError)
	{
		LyShowUrlBehavior behavior = (LyShowUrlBehavior)getBehavior();
		URL				  url = behavior.getFileUrl(this, fileName, environment, showError);

		return url;
	}

	/**
	 * Gets the Input Stream associated to the given file name.
	 *
	 * @param   urlName     the file to read given as an url string
	 * @param   environment  the current environment
	 * @param   showError    if true, show errors thrown to the user
	 *
	 * @return  the input stream of the file if found, null otherwise
	 *
	 * @see     LyShowUrlBehavior#getFileUrl(LyShowUrlController, String, LyEnvironment, boolean)
	 */
	public InputStream getFileInputStream(String urlName, LyEnvironment environment, boolean showError)
	{
		LyShowUrlBehavior behavior = (LyShowUrlBehavior)getBehavior();
		InputStream		  in = behavior.getFileInputStream(this, urlName, environment, showError);

		return in;
	}

	/**
	 * Gets the given file content.
	 *
	 * @param   fileName     the file to read
	 * @param   environment  the current environment
	 * @param   showError    indicates if an error has to be shown to the user. If false, only print
	 *                       error in log files
	 *
	 * @return  the file content if the file is found and could be read, null otherwise
	 */
	public String getFileContent(String fileName, LyEnvironment environment, boolean showError)
	{
		URL url = getFileUrl(fileName, environment, showError);

		if (url == null)
			return null;

		InputStream in = getFileInputStream(url.toString(), environment, showError);

		if (in == null)
			return null;

		return getFileContent(in, environment, showError);
	}

	/**
	 * Get the file content
	 * @param in the source input stream
	 * @param environment the current environment
	 * @param showError indicates if an error must be shown to the user if needed
	 * @return the file content
	 */
	public String getFileContent(InputStream in, LyEnvironment environment, boolean showError)
	{
		try
		{
			if (in == null)
				return null;

			StringBuffer res = new StringBuffer();
			int			 nbchars = in.available();
			char[]		 content = new char[nbchars];

			// creating the reader
			BufferedReader reader = new BufferedReader(new InputStreamReader(in));

			// start reading
			reader.read(content, 0, nbchars);
			res.append(content);
			reader.close();
			in.close();

			return res.toString();
		}
		catch (IOException e)
		{
			String message = environment.getMessage("LY_ER_FILE", _url);

			if (showError)
				showError(message);

			environment.logError(message);
			environment.writeException(e);

			return null;
		}
	} // end method getFileContent

	/**
	 * Gets the given file content.
	 *
	 * @param   fileName     the file to read
	 * @param   environment  the current environment
	 *
	 * @return  the file content if the file is found and could be read, null otherwise
	 *
	 * @see     #getFileContent(String, LyEnvironment, boolean)
	 */
	public String getFileContent(String fileName, LyEnvironment environment)
	{
		return getFileContent(fileName, environment, true);
	}

	/**
	 * Sets the file to display in the text window. In this case there is no label, and the text
	 * field has 100 columns and 20 rows.
	 *
	 * @param   url  Url to display
	 *
	 * @return  the text view component
	 */
	public LyText setUrl(String url)
	{
		return setUrl(url, null, 100, 20);
	}

	/**
	 * Sets the text to display in the text window.
	 *
	 * @param   fieldLabel  Label of the text field
	 * @param   itemLabel   Text to display
	 * @param   nCols       Number of columns of the text field
	 * @param   nRows       Number of rows of the text field
	 * @param   format      Format of the text (See also tag <TEXT> of leon.dtd)
	 *
	 * @return  the text view component
	 */
	public LyText setText(String fieldLabel, String itemLabel, int nCols, int nRows, String format)
	{
		LyShowUrlBuilder builder = (LyShowUrlBuilder)getBuilder();

		if (builder == null)
			return null;

		String fieldId = getId() + "_UrlId";

		return builder.setText(fieldId, fieldLabel, itemLabel, nCols, nRows, format, true);
	}

	/**
	 * Updates the text in the field with the given id.
	 *
	 * @param  itemLabel  New text for the field
	 */
	public void updateText(String itemLabel)
	{
		String	    fieldId = getId() + "_UrlId";
		LyFormField formField = (LyFormField)getComponent().getComponent(fieldId);
		LyText	    text = (LyText)formField.getFieldItem();

		text.setLabel(itemLabel);

		// Update the last modified time.
		Date date = new Date();

		setLastModified(date.getTime());
	}

	/**
	 * @see leon.control.LyController#getDefaultBuilder()
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyShowUrlBuilder(this);
	}

	/**
	 * This method is overriden to return null because there is not any selection.
	 *
	 * @return  always null
	 */
	public LyObjectList getSelection()
	{
		return null;
	}

	/**
	 * This method is overriden to return null because there is not any content.
	 *
	 * @return  always null
	 */
	public LyObjectList getContent()
	{
		return null;
	}

	/**
	 * @see leon.control.LyPanelController#getPanelComponent()
	 */
	public LyPanel getPanelComponent()
	{
		return (LyPanel)getComponent();
	}

	/**
	 * Make the view visible to the user. The effect of this command depends on the nature of the
	 * application. If the display is a fat client (windows (awt) / swing) this methods should open
	 * a window and show the view. In the case of thin client (html), this prepares the pages which
	 * will be returned and displayed in the client browser.
	 *
	 * <P>For any display, this method should consider that the page is ready to be shown and should
	 * be displayed.</P>
	 */
	public void showView()
	{
		LyViewManager viewManager = getViewManager();

		if (viewManager.hasProperty("MANAGE_URL"))
		{
			LyShowUrlBehavior behavior = (LyShowUrlBehavior)getBehavior();
			boolean			  newWindow = false;

			if (behavior != null)
				newWindow = behavior.openNewWindow(this);

			// web mode + target specified
			if (((getAction().getTarget() != null) && (_url != null))
					|| newWindow)
			{
				LyEnvironment env = getEnvironment();
				String		  path = env.getFile(_url);

				path = env.getString().expand(path);
				path = path.replaceAll("\\\\\\\\", "\\\\");

				viewManager.showLocation(null, path, true);
				// We don't need to keep the view controller
				LyViewController vc = getViewController();
				if (vc != null)
					vc.free();
				return;
			}
		}
		else
		{
			String externalViewer = _environment.getFile("LY_EXTERNAL_VIEWER");

			if ((externalViewer != null)
					&& !externalViewer.equals("LY_EXTERNAL_VIEWER"))
			{
				viewManager.showLocation(null, _url, true);
				return;
			}
		}

		super.showView();
	}

	/**
	 * Sets the current url and update the view.
	 *
	 * @param  url  the new url
	 */
	public void updateUrl(String url)
	{
		if (url == null)
			return;

		setUrl(url);

		LyShowUrlBuilder builder = (LyShowUrlBuilder)getBuilder();

		builder.setUrl(url);
	}

	/**
	 * Gets the current url. Gets the behavior url if not null, the first url otherwise.
	 *
	 * @return  the current url
	 *
	 * @see     LyShowUrlBehavior#getUrl(LyShowUrlController)
	 */
	public String getUrl()
	{
		if (_url != null)
			return _url;

		LyAction action = getAction();

		if (action == null)
			return null;

		LyShowUrlBehavior behavior = (LyShowUrlBehavior)getBehavior();

		if (behavior != null)
			return behavior.getUrl(this);

		return null;
	}

} // end class LyShowUrlController
