/**
 * LySimplePrintController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LySimplePrintController.java 34624 2012-03-14 10:14:54Z jlsaouli $
 */

package leon.control;

import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.print.*;


/**
 * This class describes the controller associated to the print action of a list of objects. This
 * controller doesn't have a builder. The controller is associted with a behavior that mus
 * implementes LyPrintBehaviorInterface. Usually, this behavior is an instance of the class
 * LyPrintBehavior
 *
 * @see LyAreaController
 * @see leon.app.behaviorinterface.LyPrintBehaviorInterface
 * @see leon.app.behavior.LyPrintBehavior
 */
public abstract class LySimplePrintController extends LyAreaController
	implements Runnable,
		LyPrintDataInterface
{
	/** File Version */
	public static final String		   __VERSION =
		"$Id: LySimplePrintController.java 34624 2012-03-14 10:14:54Z jlsaouli $";

	/** Format when printing to File. */
	protected static String			   FORMAT_ASCII = "LY_ASCII".intern();

	/** Format when printing to File. */
	protected static String			   FORMAT_POSTSCRIPT = "LY_POSTSCRIPT".intern();

	/** Format when printing to File. */
	protected static String			   FORMAT_HTML = "LY_HTML".intern();

	/** Object list to print. */
	protected LyObjectList			   _objects;

	/** Current print manager. */
	protected LyPrintManager		   _printManager;

	/** Print handler support (file or printer name). */
	public String					   _supportName;

	/** Specific printing behavior. */
	protected LyPrintBehaviorInterface _behavior;

	/** Print objects as a list. */
	protected boolean				   _byList;

	/** Print objects in a file. */
	protected boolean				   _inFile;

	/** Print format (ASCII / POSTSCRIPT / HTML). */
	protected String				   _format;

	/** Supported print formats ASCII. */
	public String					   PRINT_FORMATS;

	/** FieldInfo marks (null means all fields). */
	protected BitSet				   _fieldMarks;

	/** List of fields to print. */
	protected LyInfoList<LyFieldInfo>  _fields;

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 * @param action  related action. If an action is specified, the controller uses action
	 *                parameters (action name, action specific XML view description ...)
	 * @param objects current context (may be null if the view is not contextual to a list of
	 *                objects).
	 */
	public LySimplePrintController(LySession session, LyController parent,
		LyAction action, LyObjectList objects)
	{
		super(session, parent, action, objects);

		_action = action;
		_objects = objects;

		init();
	}

	/**
	 * Init this instance.
	 */
	protected void init()
	{
		_byList = false;
		_inFile = false;
		_fieldMarks = getParameterMarks("_fieldMarks", LyBitSet.Type.F_MARKS);
		_fields = null;

		LyString string = getEnvironment().getString();

		PRINT_FORMATS = FORMAT_ASCII;

		if (LyEnvironment.isOperatingSystem(LyEnvironment.OS_UNIX))
			PRINT_FORMATS += string.SEP + FORMAT_POSTSCRIPT;
		PRINT_FORMATS += string.SEP + FORMAT_HTML;
	}
	
	/**
     * Free the current print job.
     */
    public void free()
    {
        freePrinting();
    }

	/**
	 * this method free the current print job.
	 */
	protected void freePrinting()
	{
		if (_printManager != null)
			_printManager.end();

		// Do as if this controller was freed by its parent
		super.free(_parent);
	}

	/**
	 * Returns default builder. This method overrides the standard method to return. There is no
	 * builder associated with this controller
	 *
	 * @return the default builder
	 */
	public LyBuilder getDefaultBuilder()
	{
		return null;
	}

	/**
	 * Returns view controller. This controller doesn't have a view controller
	 *
	 * @return the associated view Controller
	 */
	public LyViewController getViewController()
	{
		return null;
	}

	/**
	 * Shows the Window which contains the view.
	 */
	public void showView()
	{
		LyViewManager viewManager = getViewManager();

		viewManager.execGui(this, false);
	}

	/**
	 * Printing objects starts.
	 */
	public abstract void run();

	/**
	 * Callback for end of print action.
	 * 
	 * @return true if successfull
	 */
	public boolean printEnd()
	{
		if (_objects == null)
			return false;

		if (_byList)
			print(_objects);
		else
		{
			LyObject object;
			int		 i, n = _objects.getSize(), startPage = 1, np;

			for (i = 0; i < n; i++)
			{
				object = _objects.getObject(i);
				np = print(object, _fieldMarks, LyPrintManager.PRINT_TABLE_FONT_REDUCTION, startPage);
				if (!resetPageNumberForObject())
					startPage += np;
			}
		}

		return true;
	}

	/**
	 * Print a LyObject instance.
	 *
	 * @param  object        object to print
	 * @param  marks         marks required for fieldInfo added to the print output
	 * @param  fontReduction font reduction factor
	 * @param startPage      the index of the first page
	 * @return Number of pages drawn for the given object.
	 */
	protected int print(LyObject object, BitSet marks, int fontReduction, int startPage)
	{
		LyClassInfo			    classInfo = object.getClassInfo();
		LyInfoList<LyFieldInfo> fields;

		if (marks == null)
			fields = classInfo.getFieldInfos(marks);
		else
			fields = getFields(classInfo);

		return _printManager.print(object, getEnabledFields(fields), fontReduction, startPage);
	}

	/**
	 * Print a LyObject.
	 *
	 * @param  object object to print
	 * @param  startPage the start page
	 * @return Number of pages drawn for the given object.
	 */
	protected int print(LyObject object, int startPage)
	{
		LyClassInfo			    classInfo = object.getClassInfo();
		LyInfoList<LyFieldInfo> fields = getFields(classInfo);

		return _printManager.print(object, getEnabledFields(fields),
			LyPrintManager.PRINT_TABLE_FONT_REDUCTION, startPage);
	}

	/**
	 * Print a list of LyObjects.
	 *
	 * @param objects       list of objects to print
	 * @param marks         mark required for fieldInfo added to the print output
	 * @param fontReduction Font reduction factor
	 */
	protected void print(LyObjectList objects, BitSet marks, int fontReduction)
	{
		LyClassInfo			    classInfo = objects.getObject(0).getClassInfo();
		Vector<LyObject>	    list = objects.getObjects();
		LyInfoList<LyFieldInfo> fields;

		if (marks != null)
			fields = classInfo.getFieldInfos(marks);
		else
			fields = getFields(classInfo);

		_printManager.print(list, getEnabledFields(fields, true), fontReduction);
	}

	/**
	 * Print a list of LyObjects.
	 *
	 * @param objects list of objects to print
	 */
	protected void print(LyObjectList objects)
	{
		LyClassInfo			    classInfo = objects.getObject(0).getClassInfo();
		LyInfoList<LyFieldInfo> fields = getFields(classInfo);
		Vector<LyObject>	    list = objects.getObjects();

		_printManager.print(list, getEnabledFields(fields, true),
			LyPrintManager.PRINT_TABLE_FONT_REDUCTION);
	}

	/**
	 * Utility methods to filter secret or hidden fields from given list of fields.
	 * @param fieldInfos a list of fields
	 * @return a vector of enabled fields
	 */
	protected Vector<Object> getEnabledFields(LyInfoList<LyFieldInfo> fieldInfos)
	{
		return getEnabledFields(fieldInfos, false);
	}

	/**
	 * Utility methods to filter secret or hidden fields from given list of fields.
	 * @param fieldInfos a list of fields
	 * @param table indicates if we print a table
	 * @return a vector of enabled fields
	 */
	protected Vector<Object> getEnabledFields(LyInfoList<LyFieldInfo> fieldInfos, boolean table)
	{
		int			   i, n = fieldInfos.getSize();
		Vector<Object> fields = new Vector<Object>(n);
		LyFieldInfo    fieldInfo;

		for (i = 0; i < n; i++)
		{
			fieldInfo = fieldInfos.getInfo(i);
			if (fieldInfo.hasMark(LyFieldInfo.MARK_SECRET)
					|| fieldInfo.hasMark(LyFieldInfo.MARK_HIDDEN))
				continue;

			if (!(fieldInfo instanceof LyRelationFieldInfo) || table)
			{
				fields.add(fieldInfo);
				continue;
			}

			LyRelationFieldInfo		 relation = (LyRelationFieldInfo)fieldInfo;
			LyRelationFieldInfo.Type type = relation.getType();

			if (!(relation.isMultiple())
				&& (type == LyRelationFieldInfo.Type.COMPOSITION))
			{
				Vector<Object> subFields = getEnabledFields(relation.getClassInfo(0)
							.getFieldInfos());

				for (int j = 1; j < relation.getClassInfoCount(); j++)
					subFields = intersect(subFields,
							getEnabledFields(relation.getClassInfo(j).getFieldInfos()));

				for (int j = 0; j < subFields.size(); j++)
					fields.add(new LyPrintRelationDesc(relation, subFields.get(j)));
			}
			else
				fields.add(fieldInfo);
		}

		return fields;
	} // end method getEnabledFields

	private static Vector<Object> intersect(Vector<Object> v1, Vector<Object> v2)
	{
		if ((v1 == null) || (v2 == null))
			return null;

		Vector<Object> toReturn = new Vector<Object>();

		for (int i = 0; i < v1.size(); i++)
			if (v2.contains(v1.get(i)))
				toReturn.add(v1.get(i));

		return toReturn;
	}

	/**
	 * Set the specified behavior for this controller.
	 *
	 * @param behavior behavior
	 */
	public void setPrintBehaviorInterface(LyPrintBehaviorInterface behavior)
	{
		_behavior = behavior;
	}

	/**
	 * Accessor to controller behavior A behavior is associated to a controller to let the
	 * developper specify application logic in the behavior.
	 *
	 * @return associated behavior
	 */
	public LyBehaviorInterface getBehavior()
	{
		return _behavior;
	}

	/**
	 * Sets printing by list mode.
	 * @param byList the byList status
	 */
	public void setByList(boolean byList)
	{
		_byList = byList;
	}

	/**
	 * Sets the lits of fields to print.
	 * @param fields a list of fields
	 */
	public void setFields(LyInfoList<LyFieldInfo> fields)
	{
		_fields = fields;
	}

	/**
	 * Gets the lits of fields to print.
	 * @param classInfo a classInfo
	 * @return a list of fields of the given classInfo
	 */
	public LyInfoList<LyFieldInfo> getFields(LyClassInfo classInfo)
	{
		if (_fields != null)
			return _fields;

		LyInfoList<LyFieldInfo> fields = null;
		LyFieldInfo			    fieldInfo = (LyFieldInfo)_action.getParameter("_fieldInfo", true);
		int					    np = ((fieldInfo != null) ? 2 : 1);

		do
		{
			if (fieldInfo != null)
			{
				if (fields == null)
					fields = new LyInfoList<LyFieldInfo>();
				fields.addInfo(fieldInfo);
			}
			fieldInfo = (LyFieldInfo)_action.getParameter("_fieldInfo" + np, true);
			np++;
		}
		while (fieldInfo != null);

		if ((fields == null) && (_fieldMarks != null))
			fields = classInfo.getFieldInfos(_fieldMarks);

		if ((fields == null) || (fields.getSize() < 1))
		{
			fields = classInfo.getFieldInfos();
			// Remove hidden fields
			fields = fields.getInfos(LyFieldInfo.MARK_HIDDEN, true);
			// Remove secret fields
			fields = fields.getInfos(LyFieldInfo.MARK_SECRET, true);
		}

		_fields = fields;
		return _fields;
	} // end method getFields

	/**
	 * Access to field marks used to build the controlled view.
	 *
	 * @return the field marks
	 */
	public BitSet getFieldMarks()
	{
		return _fieldMarks;
	}

	/**
	 * Method that should be called when starting the run method when the controller is Runnable.
	 * Modified to register the printing thread on the controller.
	 */
	protected void startRun()
	{
		super.startRun();

		if ((_parent != null) && (_parent.getViewController() != null))
			_parent.getViewController().setPrintThread(Thread.currentThread());
	}

	/**
	 * Method that should be called when ending the run method when the controller is Runnable.
	 * Modified to unregister the printing thread on the controller.
	 *
	 * @see #startRun
	 */
	protected void endRun()
	{
		if ((_parent != null) && (_parent.getViewController() != null))
			_parent.getViewController().setPrintThread(null);

		super.endRun();
	}

	/**
	 * Indicates if print job is a pdf job.
	 * @return true if in pdf mode
	 */
	protected boolean isPdfMode()
	{
		return false;
	}

	/**
	 * Checks if printing is done in a file.
	 *
	 * @return True if printing is done in a file, false otherwise.
	 */
	public boolean isInFile()
	{
		return _inFile;
	}

	/**
	 * Gets support name (either printer name or file name if printing is done in a file).
	 *
	 * @return Support name.
	 */
	public String getSupportName()
	{
		return _supportName;
	}

	/**
	 * Gets printing format.
	 *
	 * @return Print format.
	 */
	public String getFormat()
	{
		return _format;
	}

	/**
	 * Indicates if page number must be reset when printing a new object.
	 * @return true if page number must be reset when printing a new object.
	 */
	protected boolean resetPageNumberForObject()
	{
		return true;
	}

	/**
	 * @see leon.view.print.LyPrintDataInterface#getTitle(java.lang.Object)
	 */
	public String getTitle(Object object)
	{
		if (object == null)
			if (_parent instanceof LyAreaController)
				return ((LyAreaController)_parent).getTitle();
			else
				return null; // XXX

		LyObject    logicalObject = (LyObject)object;
		LyClassInfo classInfo = logicalObject.getClassInfo();

		return classInfo.getName() + " : " + getName(logicalObject);
	}

	/**
	 * @see leon.view.print.LyPrintDataInterface#getImage(java.lang.Object)
	 */
	public String getImage(Object object)
	{
		if (object == null)
			if (_parent instanceof LyAreaController)
				return ((LyAreaController)_parent).getImage();
			else
				return null; // XXX

		LyObject    logicalObject = (LyObject)object;
		LyClassInfo classInfo = logicalObject.getClassInfo();

		return super.getImage(classInfo);
	}

	/**
	 * @see leon.view.print.LyPrintDataInterface#getTab(java.lang.Object)
	 */
	public String getTab(Object field)
	{
        Object fieldTemp = field;

        while (fieldTemp instanceof LyPrintRelationDesc)
            fieldTemp = ((LyPrintRelationDesc)fieldTemp).getField();

        LyFieldInfo fieldInfo = (LyFieldInfo)fieldTemp;
	        
		return fieldInfo.getTab();
	}

	/**
	 * @see leon.view.print.LyPrintDataInterface#getLabel(java.lang.Object)
	 */
	public String getLabel(Object field)
	{
		Object fieldTemp = field;

		while (fieldTemp instanceof LyPrintRelationDesc)
			fieldTemp = ((LyPrintRelationDesc)fieldTemp).getField();

		LyFieldInfo fieldInfo = (LyFieldInfo)fieldTemp;

		return (_byList ? fieldInfo.getAlias() : fieldInfo.getName());
	}

	/**
	 * @see leon.view.print.LyPrintDataInterface#getLabel(java.lang.Object, java.lang.Object)
	 */
	public String getLabel(Object object, Object field)
	{
		LyObject		    logicalObject = (LyObject)object;
		LyRelationValue     relationValue;
		LyRelationFieldInfo relation;

		while (field instanceof LyPrintRelationDesc)
		{
			relation = (LyRelationFieldInfo)((LyPrintRelationDesc)field).getRelation();
			field = ((LyPrintRelationDesc)field).getField();
			relationValue = logicalObject.getRelationValue(relation);
			logicalObject = ((relationValue.getValueCount() == 0) ? null
																  : relationValue.getObject(0));
		}

		LyEnvironment environment = getEnvironment();
		LyFieldInfo   fieldInfo = (LyFieldInfo)field;
		Object		  value = ((logicalObject == null) ? null : logicalObject.getValue(fieldInfo));
		String		  str = null;

		if (logicalObject != null)
			str = getLabel(logicalObject, fieldInfo, value);

		if (str == null)
			str = "";

		if ((fieldInfo instanceof LyChoiceFieldInfo)
				|| (fieldInfo instanceof LyNumberFieldInfo))
			str = environment.translate(str);

		return str;
	} // end method getLabel

	/**
	 * @see leon.view.print.LyPrintDataInterface#getShortLabel(java.lang.Object, java.lang.Object)
	 */
	public String getShortLabel(Object object, Object field)
	{
		LyObject		    logicalObject = (LyObject)object;
		LyRelationValue     relationValue;
		LyRelationFieldInfo relation;

		while (field instanceof LyPrintRelationDesc)
		{
			relation = (LyRelationFieldInfo)((LyPrintRelationDesc)field).getRelation();
			field = ((LyPrintRelationDesc)field).getField();
			relationValue = logicalObject.getRelationValue(relation);
			logicalObject = ((relationValue.getValueCount() == 0) ? null
																  : relationValue.getObject(0));
		}

		LyEnvironment environment = getEnvironment();
		LyFieldInfo   fieldInfo = (LyFieldInfo)field;
		Object		  value = ((logicalObject == null) ? null : logicalObject.getValue(fieldInfo));
		String		  str = ((logicalObject == null)
				? "" : getShortLabel(logicalObject, fieldInfo, value));

		if (fieldInfo instanceof LyChoiceFieldInfo)
			str = environment.translate(str);

		return str;
	}

	/**
	 * @see leon.view.print.LyPrintDataInterface#hasFooters()
	 */
	public boolean hasFooters()
	{
		if (_objects == null)
			return false;

		LyTableBehaviorInterface behavior = null;

		if (_parent instanceof LyAreaController)
		{
			LyBehaviorInterface parentBehavior = ((LyAreaController)_parent).getBehavior();

			if (parentBehavior instanceof LyTableBehaviorInterface)
				behavior = (LyTableBehaviorInterface)parentBehavior;
		}

		LyClassInfo			    classInfo = _objects.getObject(0).getClassInfo();
		LyInfoList<LyFieldInfo> fields = getFields(classInfo);
		int					    c, nc = fields.getSize();
		LyFieldInfo			    fieldInfo;

		if (nc == 0)
			return false;

		for (c = 0; c < nc; c++)
		{
			fieldInfo = fields.getInfo(c);
			if (fieldInfo.hasMark(LyFieldInfo.MARK_SECRET)
					|| fieldInfo.hasMark(LyFieldInfo.MARK_HIDDEN))
				continue;

			if ((behavior != null)
					&& (behavior.getFooterLabel((LyTableController)_parent, fieldInfo, _objects)
						!= null))
				return true;
		}

		return false;
	} // end method hasFooters

	/**
	 * @see leon.view.print.LyPrintDataInterface#getSortIndex(java.lang.Object)
	 */
	public int getSortIndex(Object field)
	{
		LySort sort = _objects.getSort();

		if (field instanceof LyPrintRelationDesc)
			field = ((LyPrintRelationDesc)field).getRelation();

		return ((sort == null) ? -1 : sort.getIndex((LyFieldInfo)field));
	}

	/**
	 * @see leon.view.print.LyPrintDataInterface#isIncreasing(java.lang.Object)
	 */
	public boolean isIncreasing(Object field)
	{
		LySort sort = _objects.getSort();

		if (sort == null)
			return false;

		if (field instanceof LyPrintRelationDesc)
			field = ((LyPrintRelationDesc)field).getRelation();

		LySortCriterion criterion = sort.getCriterion((LyFieldInfo)field);

		return criterion.isIncreasing();
	}

	/**
	 * @see leon.view.print.LyPrintDataInterface#getFooterLabel(java.lang.Object)
	 */
	public String getFooterLabel(Object field)
	{
		if (field instanceof LyPrintRelationDesc)
			field = ((LyPrintRelationDesc)field).getRelation();

		LyTableBehaviorInterface behavior = null;
		LyFieldInfo				 fieldInfo = (LyFieldInfo)field;

		if (_parent instanceof LyAreaController)
		{
			LyBehaviorInterface parentBehavior = ((LyAreaController)_parent).getBehavior();

			if (parentBehavior instanceof LyTableBehaviorInterface)
				behavior = (LyTableBehaviorInterface)parentBehavior;
		}

		return ((behavior == null)
			? null : behavior.getFooterLabel((LyTableController)_parent, fieldInfo, _objects));
	}

	/**
	 * @see leon.view.print.LyPrintDataInterface#isLeftAligned(java.lang.Object)
	 */
	public boolean isLeftAligned(Object field)
	{
		if (field instanceof LyPrintRelationDesc)
			field = ((LyPrintRelationDesc)field).getRelation();

		LyFieldInfo fieldInfo = (LyFieldInfo)field;

		return (!(fieldInfo instanceof LyNumberFieldInfo));
	}

	/**
	 * @see leon.view.print.LyPrintDataInterface#getParentView()
	 */
	public LyView getParentView()
	{
		return _parent.getView();
	}
} // end class LySimplePrintController
