/**
 * LySimpleTableController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LySimpleTableController.java 41549 2012-11-16 13:32:03Z bangiole $
 */

package leon.control;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.info.daemon.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.mobile.*;
import leon.view.setinterface.*;

import java.util.*;



/**
 * This class implements the basics of a table controller. It is used to factorize code between all
 * table controllers. It is associated with a builder (LySimpleTableBuilder) and a behavior that
 * implements LySimpleTableBehaviorInterface. There is no default implementation of this behavior.
 *
 * @see LySimpleTableBuilder
 * @see leon.app.behaviorinterface.LyTableBehaviorInterface
 * @see LyTableController
 * @see LyMultiPageTableController
 */
public abstract class LySimpleTableController extends LyPanelController implements Runnable,
    LyAdvancedWorkSpaceListener, LyPrintListenerInterface, LyFindInterface, LyFilterEndInterface,
    LyGetClassInfoInterface, LySortEndInterface, LyFastFilterEndInterface, LySetPrintModeInterface,
    LyQuestionAnswerInterface
{

    /** File Version */
    public static final String __VERSION =
        "$Id: LySimpleTableController.java 41549 2012-11-16 13:32:03Z bangiole $";

    /** Menu id : Previous page. */
    public static final String DEMAND_PREVIOUS_PAGE = "_previousPage".intern();

    /** Menu id : Next page. */
    public static final String DEMAND_NEXT_PAGE = "_nextPage".intern();

    /** Menu id : Fix column. */
    public static final String DEMAND_FIX_COLUMN = "_fixColumn_".intern();

    /** Menu id : Unfix column. */
    public static final String DEMAND_UNFIX_COLUMN = "_unfixColumn".intern();

    /** Menu id : Reverse list. */
    public static final String DEMAND_REVERSE = "_reverseList".intern();

    /** Menu id : Simple Sort. */
    public static final String DEMAND_SORT_BY = "_sortBy_".intern();

    /** Menu id : Extended sort. */
    public static final String DEMAND_SORT_EXTEND = "_sortExtend_".intern();

    /** Menu id : Reverse sort. */
    public static final String DEMAND_SORT_REVERSE = "_sortReverse_".intern();

    /** Menu id : Default sort. */
    public static final String DEMAND_SORT_DEFAULT = "_sortDefault_".intern();

    /** Menu id : Status sort. */
    public static final String DEMAND_STATUS_SORT = "_statusSort_".intern();

    /** Menu id : Invert status sort. */
    public static final String DEMAND_INVERT_STATUS_SORT = "_invertStatusSort_".intern();

    /** Menu id : Group. */
    public static final String DEMAND_GROUP_BY = "_groupBy_".intern();

    /** Message id : Objects count. */
    public static final String MESSAGE_OBJECTS_COUNT = "_objectsCount".intern();

    /** Menu id : Print list. */
    public static final String DEMAND_PRINT_TABLE = "_printTable".intern();

    /** Tool id : subtoolbar action. */
    public static final String DEMAND_SUB_TOOL_ACTION = "_subToolBarAction".intern();

    /** Menu id : Choose columns. */
    public static final String DEMAND_CHOOSE_COLUMNS = "_chooseColumns".intern();

    /** Menu id : Sort Wizard. */
    public static final String DEMAND_SORT_WIZARD = "_sortWizard".intern();

    /** Sort Wizard id : Sort Wizard direction. */
    protected static final String DEMAND_SORT_WIZARD_DIRECTION = "_sortWizardDirection".intern();

    /** Sort wizard id : direction decreasing */
    protected static final String DEMAND_SORT_WIZARD_DIR_DECREASING = "sortWizardDirDecreasing"
        .intern();

    /** Sort wizard id : direction increasing */
    protected static final String DEMAND_SORT_WIZARD_DIR_INCREASING = "sortWizardDirIncreasing"
        .intern();

    /** Demand id : Row tool. */
    public static final String DEMAND_ROW_TOOL = "rowTool".intern();

    /** Demand id : Copy cell content. */
    public static final String DEMAND_COPY_CELL_CONTENT = "_copyCellContent".intern();

    /** Prefix used for section row identifiers. */
    public static final String SECTION_ROW_PREFIX = "SECTION_ROW_PREFIX".intern();

    /** Prefix used for footer row identifiers. */
    public static final String FOOTER_ROW_PREFIX = "FOOTER_ROW_PREFIX".intern();

    /** Empty cell id. */
    public static final String EMPTY_CELL = "_emptyCell".intern();

    /** Display List */
    public static final String DEMAND_LIST_DISPLAY = "_demandListDisplay".intern();

    /** prefix for displaylist fields ID's. */
    protected static final String DISPLAY_LIST_PREFIX = "_displayList_".intern();

    /** Class info. */
    protected LyClassInfo _classInfo;

    /** Object workSpace. */
    protected LyWorkSpace _initialWorkSpace;

    /** Object workSpace. */
    protected LyWorkSpace _workSpace;

    /** FieldInfo marks (null means all fields). */
    protected BitSet _fieldMarks;

    /** FieldInfo status marks (null means no fields). */
    protected BitSet _statusFieldMarks;

    /** Class filter. */
    protected LyFilter _classFilter;

    /** Current fast filter value. */
    protected String _fastFilterValue;

    /** Current fast filter field info. */
    protected LyFieldInfo _fastFilterFieldInfo;

    /** Indicates if extension is used for fast filter creation. */
    protected boolean _fastFilterExtend;

    /** Last fast filter field info. */
    protected LyFieldInfo _lastFastFilterFieldInfo;

    /** Current sort. */
    protected LySort _sort;

    /** Indication on the sort of the status. */
    protected int _currentStatusSort;

    /** Index of fixed column. */
    protected int _fixedColumn;

    /** List of fields displayed in table. */
    protected LyInfoList<LyFieldInfo> _fields;

    /**
     * List of fields used for computing values in the table. When one of this fields is modified in
     * an object, the table will be updated.
     */
    protected LyInfoList<LyFieldInfo> _dependsFields;

    /** Internal String table used for optimization. */
    protected Hashtable<String, String> _strings;

    /** Specific controller class to create and keep up to date the content of the view. */
    protected LyTableBehaviorInterface _tableBehaviorInterface;

    /** Current page being shown. */
    protected int _currentPage;

    /** Maximal number of rows in a page. */
    protected int _maxRowCount;

    /** Row tool environment. */
    protected LyEnvironment _rowToolEnvironment;

    /** Boolean indictaing if filter panel has to be opened. */
    protected boolean _openFilterIfEmpty;

    /** Boolean indicating if workspace is or not dynamic. */
    protected boolean _dynamic;

    /** Indicates whether the print mode is activated or not. */
    protected boolean _printMode;

    /** List of closed rows. */
    protected Vector<String> _closedRows;

    /** Indicates whether the fop mode is activated or not. */
    protected boolean                  _fopMode;

    /** State for keeping selection during specific events. */
    private boolean _keepSelection = false;

    /** List Display */
    protected boolean _displayList = false;

    /** has List Display toggle button ? */
    protected boolean _displayListToggleButton = false;

    /** show fields in display List ? */
    protected boolean _displayListShowFieldLabels = true;

    /** String representing the sizes of the columns. */
    protected String _columnSizes;

    /** Index of the field used by the groupBy action. */
    protected int	_groupByFieldOldIndex;


    /**
     * Constructor.
     *
     * @param session   current session associated to the controller. The controller automatically
     *                  registers on the specified session.
     * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
     *                  indicates a child controller (for example a dialog created from a main view)
     * @param action    related action. If an action is specified, the controller uses action
     *                  parameters (action name, action specific XML view description ...)
     * @param classInfo class info for objects displayed in the list.
     * @param context   current context (may be null if the view is not contextual to a list of
     *                  objects).
     */
    public LySimpleTableController(LySession session, LyController parent, LyAction action,
        LyClassInfo classInfo, LyObjectList context)
    {
        super(session, parent, action, context);

        _classInfo = classInfo;
        init();
    }

    /**
     * Method that gives the interface that must be implemented by the behavior associated to this
     * controller.
     *
     * @return Class representing the behavior interface of this controller.
     */
    public Class<?> getBehaviorInterface()
    {
        return LyTableBehaviorInterface.class;
    }

    /**
     * Set field marks used to build the controlled view.
     *
     * @param fieldMarks marks to set
     */
    public void setFieldMarks(BitSet fieldMarks)
    {
        _fieldMarks = fieldMarks;
    }

    /**
     * Access to field marks used to build the controlled view.
     *
     * @return the field marks
     */
    public BitSet getFieldMarks()
    {
        return _fieldMarks;
    }

    /**
     * Init this instance.
     */
    protected void init()
    {

        if (_classInfo == null)
            _classInfo = (LyClassInfo) getParameter("_classInfo");
        else
        {
            LyClassInfo classInfo = (LyClassInfo) getParameter("_classInfo", true);

            if (classInfo != null)
                _classInfo = classInfo;
        }

        _classFilter = (LyFilter) getParameter("_classFilter");
        _filter = (LyFilter) getParameter("_filter");
        _fastFilterValue = null;
        _fastFilterFieldInfo = null;
        _fastFilterExtend = false;
        _lastFastFilterFieldInfo = null;
        _filterActive = (_filter != null);
        _fieldMarks = getParameterMarks("_fieldMarks", LyBitSet.Type.F_MARKS);
        _statusFieldMarks = getParameterMarks("_statusFieldMarks", LyBitSet.Type.F_MARKS);

        // If no status mark is given, do not consider any field as status
        if ((_statusFieldMarks != null) && (_statusFieldMarks.isEmpty()))
            _statusFieldMarks = null;

        _fields = null;
        _currentStatusSort = 0;
        _refreshTime = -1;
        _fixedColumn = -1;
        _strings = new Hashtable<String, String>();
        _currentPage = 0;
        _maxRowCount = -1;

        _tableBehaviorInterface = (LyTableBehaviorInterface) getBehavior();
        _dependsFields = _tableBehaviorInterface.getDependsFieldInfos(this, _action);
        _openFilterIfEmpty = false;
        _dynamic = true;
        _printMode = false;
        _closedRows = null;
        _columnSizes = null;
        _groupByFieldOldIndex = -1;

        String str = (String) getParameter("_openFilter", true);

        if ((str != null) && str.equalsIgnoreCase("IF_EMPTY"))
            _openFilterIfEmpty = true;

        str = (String) getParameter("_dynamic", true);

        if ((str != null) && str.equalsIgnoreCase("false"))
            _dynamic = false;

        // SMA : init list mode

        String display = (String) getParameter("_displayList", true);

        if(getViewManager().hasProperty(LyMobileViewManager.MOBILE_ONLY_PROPERTY))
        {
        	display = "LIST";
        }

        if (display != null)
        {
            String showFields = (String) getParameter("_displayListShowFieldLabels", true);

            if ("false".equalsIgnoreCase(showFields))
                _displayListShowFieldLabels = false;

            if ("LIST".equalsIgnoreCase(display))
            {
                _displayList = true;
                _displayListToggleButton = false;
            }
            else if ("CHOOSE".equalsIgnoreCase(display))
            {
                _displayList = false;
                _displayListToggleButton = true;
            }
            else
            {
                _displayList = false;
                _displayListToggleButton = false;
            }
        }

        initPreference();
    } // end method init

    /**
             * Reset component.
             *
             * @param full Boolean indicating if a full reset (including reset of datas) is requiered or not
             *             (only reset graphical area).
             */
    public void reset(boolean full)
    {

        if (isIdle())
            return;

        // Freeze the table
        LyTable table = getTableComponent();
        boolean wasIdle = table.isIdle();

        table.setIdle(true);

        if (full)
            updatePreferencesMenu();

        updateContent();
        updateSelections();
        updateContextClient();
        updateTools();
        updateHeaders();
        updateFooters();
        updateMessages();
        updateSortMenu();
        updateGroupByMenu();
        updateFilterMenu();
        updateTitleBar();


        _keepSelection = false;
        // Update the last modification time.
        setLastModified(new Date().getTime());

        LyBehaviorInterface behavior = getBehavior();

        if (behavior instanceof LyActionBehaviorInterface)
            ((LyActionBehaviorInterface) behavior).reset(this, full);

        table.setIdle(wasIdle);
    } // end method reset

    /**
     * Update the menu for the current controller.
     */
    protected void updateMenuBar()
    {

        if (_workSpace != null)
        {
            updateSortMenu();
            updateGroupByMenu();
            updateFilterMenu();
            updatePreferencesMenu();
        }

        super.updateMenuBar();
    }

    /**
     * Internal method to free workspace.
     */
    protected void freeData()
    {

        if (_workSpace != null)
        {
            try
            {
                _workSpace.free(this);
            }
            catch (Exception e)
            {
                getEnvironment().writeException(e);
            }

            _workSpace = null;
        }

        // We don't need to free _initialWorkSpace because if it is different
        // from _workSpace, _workSpace is a subWorkSpace and subWorkSpaces free their parent.
        _initialWorkSpace = null;

        _selections = null;
    }

    /**
     * Free this controller.
     *
     * @param controller parent controller
     */
    public void free(LyController controller)
    {

        if (_parent != controller)
            return;

        if (_rowToolEnvironment != null)
        {
            _rowToolEnvironment.free();
            _rowToolEnvironment = null;
        }

        freeData();

        _strings = null;
        _fields = null;
        super.free(controller);
    } // end method free

    /**
     * Clears all cached data for this controller (especially the view and the component). Updated
     * to clear workspace and avoid events treatments.
     */
    public void clearViewCache()
    {
        super.clearViewCache();
        freeData();
    }

    /**
     * Invoked when a selection occurs. This method handles: select all, unselect all and objects by
     * objects selection
     *
     * @param demand selection event created from the view. User interaction.
     */
    public void selectionDemand(LySelectionDemand demand)
    {
        int selectionType = demand.getType();

        if (_workSpace == null)
            return;

        LyObjectList	oldSelection = getSelection();

        if (selectionType == LySelectionDemand.TYPE_SELECT_ALL)
            selectObjects(null, true);
        else if (selectionType == LySelectionDemand.TYPE_UNSELECT_ALL)
            selectObjects(null, false);
        else
        {
            // Get new selection
            int n = demand.getSelectionSize();

            if (n == 0)
                selectObjects(null, false);
            else
            {
                LyObject           object;
                String             key;
                int                i;
                LyObjectId         oid;
                LySimpleObjectList selections = null;

                for (i = 0; i < n; i++)
                {
                    key = demand.getSelectionId(i);

                    if (key == null)
                        continue;

                    if (key.startsWith(SECTION_ROW_PREFIX) || key.startsWith(FOOTER_ROW_PREFIX))
                    {
                        selectObjects(null, false);
                        return;
                    }

                    oid = getObjectId(key);
                    object = _workSpace.getObject(oid);

                    if (object != null)
                    {

                        if (selections == null)
                        {
                            selections = new LySimpleObjectList();
                            selections.setSorted(false);
                        }

                        selections.addObject(object);
                    }
                    else
                        showObjectNotInViewError(oid.getFullString());
                }

                if (selections != null)
                    selectObjects(selections, true);
            }

            LyEnvironment	environment = getEnvironment();

            // Display a popup menu when LY_DISPLAY_LIGHT_TOOLBAR_ON_SELECTION resource is set
            if (environment.testResource("LY_DISPLAY_LIGHT_TOOLBAR_ON_SELECTION")
            		&& (getSelection() != null)
            		&& (!getSelection().equals(oldSelection) || !getViewManager().hasProperty("WEB_MANAGER"))
            		&& (demand.getAttribute("x") != null)
            		&& (demand.getAttribute("y") != null))
            {
                LyClassInfo classInfo = getClassInfo();

                if (classInfo.isAbstract())
                {
                    // The menu applies on selection if there is only one object selected, transmit its classInfo
                    LyObjectList selection = getSelection();

                    if ((selection != null) && (selection.getSize() == 1))
                    {
                        LyObject selectedObject = selection.getObject(0);

                        classInfo = selectedObject.getClassInfo();
                    }
                }

                // Remove XRef items and do not group actions by their image
                String	xref = environment.getString("LY_ITEM_XREF");
                String	groupActions = environment.getResource("LY_GROUP_ACTIONS_BY_IMAGE");

                environment.putString("LY_ITEM_XREF", "");
                environment.putResource("LY_GROUP_ACTIONS_BY_IMAGE", "false");

                LyMenuDemand	menuDemand = new LyMenuDemand(LyMenuDemand.TYPE_POPUP, demand.getSource());
                LyPopupMenu		popupMenu = super.buildPopupMenu(menuDemand, classInfo, true);

                environment.putString("LY_ITEM_XREF", xref);
                environment.putResource("LY_GROUP_ACTIONS_BY_IMAGE", groupActions);

                // Call the behavior to complete popup menu
                LyBehaviorInterface behavior = getBehavior();

                if (behavior instanceof LyPanelBehaviorInterface)
                    ((LyPanelBehaviorInterface) behavior).completePopupMenu(this, popupMenu);

                // Coordinates
                String	posX = demand.getAttribute("x").toString();
                String	posY = demand.getAttribute("y").toString();

                popupMenu.setStyle(LyPopupMenu.Style.LIGHT);
                popupMenu.setX(LyTools.intFromString(posX) + 5);
                popupMenu.setY(LyTools.intFromString(posY) + 3);

                LyTable table = getTableComponent();

                if (table != null)
                {
                	popupMenu.setController(getId());
                    popupMenu.setEnvironment(table.getEnvironment());
                    table.setPopupMenu(popupMenu);
                }
            }
        }
    } // end method selectionDemand

    /**
     * Invoked when a selection on an object that no longer exists is called.
     * @param id  id of the object
     */
    protected void showObjectNotInViewError(String id)
    {
        LyEnvironment env = getEnvironment();
        String        message = env.getMessage("LY_ER_OBJECT_NOT_IN_VIEW", id);

        showError(message);

        // Refresh table
        reset(false);
    }

    /**
     * Invoked when an action occurs. (LyDemandListener interface) This method implements:
     *
     * <UL>
     *   <LI>default action on object</LI>
     *   <LI>filters</LI>
     *   <LI>Find</LI>
     *   <LI>XRefs</LI>
     *   <LI>...</LI>
     * </UL>
     *
     * @param demand action event created from the view. User interaction.
     */
    public void actionDemand(LyActionDemand demand)
    {
        String source = demand.getSource();

        if (demand.getType() == LyActionDemand.TYPE_ACTION)
        {
            String[] parameters = LyString.stringToList(source, ID_SEP);

            if ((parameters == null) || (parameters.length == 0) || (parameters[0] == null))
                return;

            String id = parameters[0].intern();
            String key = demand.getValue();

            if ((id.equals(getId() + "_table"))
                    || ((id.equals("_table")) && (key != null) && (key.indexOf("_toolBar") < 0))) // "_table" action
            {

                if ((key != null) && key.endsWith(EMPTY_CELL))
                    key = key.substring(0, key.indexOf(EMPTY_CELL));

                // Default action on an item of the table
                LyObjectId oid = ((key == null) || (key.trim().length() == 0) ? null : getObjectId(key));
                LyObject   object = ((oid == null) ? null : _workSpace.getObject(oid));

                if (object != null)
                {
                    invokeDefaultAction(object);
                    return;
                }
                else if (oid != null)
                    showObjectNotInViewError(oid.getFullString());
            }
            else if (id.equals(getId() + "_selectFile" + DEMAND_DELETE_OBJECTINFOS))
                deleteObjectInfos(parameters[1]);
            // *** Fast filter **
            else if (id == FAST_FILTER)
            {

                if (enableFastFilter())
                {
                    String      fieldId = parameters[1].intern();
                    LyClassInfo classInfo = getClassInfo();
                    LyFieldInfo fieldInfo = classInfo.getFieldInfo(fieldId);

                    buildFastFilter(fieldInfo, -1, -1, false);
                }
            }
            // *** Sort menu items **
            else if (id == DEMAND_REVERSE)
                reverseSort();
            else if (id == DEMAND_CHOOSE_COLUMNS)
                chooseColumns();
            else if (id == DEMAND_UNFIX_COLUMN)
                fixColumn(-1);
            else if (id == DEMAND_FIX_COLUMN)
            {
                int column = LyTools.intFromString(parameters[1]);

                fixColumn(column);
            }
            else if (id == DEMAND_SORT_WIZARD)
                sortWizard();
            else if (id == DEMAND_SORT_BY)
            {
                LyFieldInfo     fieldInfo = _classInfo.getFieldInfo(parameters[1]);
                LySortCriterion defaultCriterion = getDefaultCriterion(fieldInfo);
                LySort          sort = new LySort();

                if (defaultCriterion == null)
                    sort.addCriterion(fieldInfo, true);
                else
                    sort.addCriterion(defaultCriterion);

                setSort(sort);
            }
            else if (id == DEMAND_SORT_EXTEND)
            {
                LyFieldInfo fieldInfo = _classInfo.getFieldInfo(parameters[1]);
                LySort      sort = getSort();

                if (sort != null)
                {
                    LySortCriterion criterion = sort.getCriterion(fieldInfo);

                    // Add sort criterion on fieldInfo in current sort
                    if (criterion == null)
                    {
                        LySortCriterion defaultCriterion = getDefaultCriterion(fieldInfo);

                        if (defaultCriterion == null)
                            sort.addCriterion(fieldInfo, true);
                        else
                            sort.addCriterion(defaultCriterion);

                        setSort(sort);
                    }
                }
            }
            else if (id == DEMAND_SORT_REVERSE)
            {
                LyFieldInfo fieldInfo = _classInfo.getFieldInfo(parameters[1]);
                LySort      sort = getSort();

                if (sort != null)
                {
                    LySortCriterion criterion = sort.getCriterion(fieldInfo);

                    if (criterion != null)
                    {
                        // Invert sort criterion on fieldInfo
                        criterion.setIncreasing(!criterion.isIncreasing());
                        setSort(sort);
                    }
                }
            }
            else if (id == DEMAND_SORT_DEFAULT)
                setSort(getDefaultSort());
            else if (id == DEMAND_STATUS_SORT)
                setStatusSort(false);
            else if (id == DEMAND_INVERT_STATUS_SORT)
                setStatusSort(true);
            // Group by
            else if (id == DEMAND_GROUP_BY)
            {
                LyFieldInfo fieldInfo = _classInfo.getFieldInfo(parameters[1]);

                groupBy(fieldInfo);
            }
            // Page loading.
            else if (id == LyTableController.DEMAND_NEXT_PAGE)
            {
                loadPage(true);

                LyTable table = getTableComponent();

                table.setVisible(true);
            }
            else if (id == LyTableController.DEMAND_PREVIOUS_PAGE)
            {
                loadPage(false);

                LyTable table = getTableComponent();

                table.setVisible(true);
            }
            // Printing.
            else if (id == LyTableController.DEMAND_PRINT_TABLE)
                printList();
            // subTool demand
            else if (id == LyTableController.DEMAND_SUB_TOOL_ACTION)
            {
                LyToolBar   toolBar = (LyToolBar) _component.getComponent("_toolBar");
                LyTool      tool = (LyTool) toolBar.getComponent(source);
                LyComponent component = tool.getComponent();

                if (component instanceof LyArea)
                {
                    int            i;
                    LyActionDemand myDemand;

                    toolBar = ((LyArea) component).getToolBar();

                    for (i = 0; i < toolBar.getChildCount(); i++)
                    {
                        tool = (LyTool) toolBar.getChild(i);

                        if (tool.isEnabled())
                        {
                            myDemand = new LyActionDemand(demand.getType(), tool.getId());
                            actionDemand(myDemand);
                            break;
                        }
                    }
                }
            }
            // copy cell content to clipboard
            else if ((id != null) && (id.startsWith(DEMAND_COPY_CELL_CONTENT)))
            {
                String        text = id.substring(DEMAND_COPY_CELL_CONTENT.length());
                LyViewManager viewManager = getViewManager();

                text = getEnvironment().translate(text);
                viewManager.copyToClipboard(text);
            }
            // Implements the modification of the view limit
            else if ((source != null)
                    && (source.equals(getId() + "_refreshViewLimit")
                        || source.equals(getId() + "_viewLimit")))
            {
                LyComponent component = getComponent().getParent(); // Parent is for __select action

                if (demand.getParameterValues() != null)
                    demand.getParameterValues().updateViewFromValues(component, false);

                LyFormField formField = (LyFormField) component.getComponent(getId()
                        + "_viewLimit");
                LyNumber    viewLimit = (LyNumber) formField.getFieldItem();
                int         vMax = (int) viewLimit.getVMax();
                int         viewLimitValue = LyTools.intFromString(viewLimit.getLabel());

                if (viewLimitValue < 1)
                    viewLimitValue = 1;

                if (viewLimitValue > vMax)
                    viewLimitValue = vMax;

                if (_maxRowCount != viewLimitValue)
                {
                    _maxRowCount = viewLimitValue;
                    reset(true);
                }
                else if (getEnvironment().testResource("LY_FORCE_REFRESH_ON_MULTIPAGE_TABLE"))
                    update();

                return;
            }
            else if (LyTableController.DEMAND_LIST_DISPLAY.equals(id))
                // Toggle display
                setDisplayList(!_displayList);
            else if ("resizeColumn".equals(id))
            {
                // Parse columnSizes parameter and add field identifiers
                String   columnSizes = demand.getValue();
                String[] sizes = LyString.stringToList(columnSizes, "##");

                if ((sizes != null) && (sizes.length > 0))
                {
                    String                  fieldSizes = "";
                    LyInfoList<LyFieldInfo> fields = getFields();
                    LyInfoList<LyFieldInfo> tableFields = getTableComponentFields(false);

                    if (tableFields != null)
                        fields = tableFields;

                    for (int i = 0; (i < sizes.length) && (i < fields.getSize()); i++)
                    {

                        if (i > 0)
                            fieldSizes += "##";

                        fieldSizes += fields.getInfo(i).getId() + "=" + sizes[i];
                    }

                    setColumnSizes(fieldSizes);
                }
            }
            else if ("dragColumn".equals(id))
            {
                String   columnIds = demand.getValue();
                String[] headerIds = LyString.stringToList(columnIds, "##");

                if ((headerIds != null) && (headerIds.length > 0))
                {
                    LyInfoList<LyFieldInfo> fields = getFields(headerIds, false);
                    setFields(fields);
                }
            }
            else
            {
                LyApplication application = getApplication();
                LyAction      action = application.getAction(id);

                if (action != null)
                {
                    LyObjectInfo template = action.getTemplate();

                    if ((template != null)
                            && template.getId().equals(LyTableController.DEMAND_PRINT_TABLE))
                        printList();
                    else
                        super.actionDemand(demand);
                }
                else
                    super.actionDemand(demand);
            }
        }
        else if (demand.getType() == LyActionDemand.TYPE_SORT)
            sort(demand.getIndex(0), false);
        else if (demand.getType() == LyActionDemand.TYPE_EXTEND_SORT)
            sort(demand.getIndex(0), true);
        else if (demand.getType() == LyActionDemand.TYPE_HREF)
        {

            // for exemple : the loading of the charts
            if (source.equals("_selectFile"))
                super.actionDemand(demand);
            else
            {
                // *** Manage Hyper text Selection : usually Consult ***
                // Invoke default action when href demanded
                String     demandId = demand.getValue();
                LyObjectId oid = getObjectId(demandId);
                LyObject   object = (LyObject) oid.getObject();

                if (object != null)
                    invokeDefaultAction(object);
            }
        }
        else if (demand.getType() == LyActionDemand.TYPE_DELETE)
        {

            if ((_selections == null) || (_selections.getSize() == 0))
                return;

            invokeDeleteAction(_selections, _classInfo);
        }
        else if (demand.getType() == LyActionDemand.TYPE_INSERT)
            invokeCreateAction(_classInfo);
        else if (demand.getType() == LyActionDemand.TYPE_STATUS)
        {
            String demandId = demand.getValue();

            if (demandId == null)
                return;

            LyObjectId oid = getObjectId(demandId);
            LyObject   object = (LyObject) oid.getObject();

            if (object != null)
            {
                LyClassInfo          classInfo = object.getClassInfo();
                LyInfoList<LyAction> actions = classInfo.getActions(LyAction.MARK_STATUS);

                if ((actions == null) || (actions.getSize() == 0))
                    invokeDefaultAction(object);
                else
                    invokeAction(actions.getInfo(0), object);
            }
            else
                showObjectNotInViewError(oid.getFullString());
        }
        // Implements the modification of the view limit
        else if ((demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED) && (source != null)
                && (source.equals(getId() + "_viewLimit")))
        {
            LyComponent component = getComponent().getParent(); // Parent is for __select action
            LyFormField formField = (LyFormField) component.getComponent(getId() + "_viewLimit");
            LyNumber    viewLimit = (LyNumber) formField.getFieldItem();
            int         viewLimitValue = LyTools.intFromString(viewLimit.getLabel());

            if (viewLimitValue < 1)
                viewLimitValue = 1;

            if (_maxRowCount != viewLimitValue)
            {
                _maxRowCount = viewLimitValue;
                reset(true);
            }

            return;
        }
        else if (demand.getType() == LyActionDemand.TYPE_OPEN)
        {
            String rowId = demand.getSource();
            String open = demand.getValue();
            LyRow  row = getRow(rowId, -1);

            if (row != null)
            {
                boolean state = "true".equals(open);

                if ((open == null) || (open.length() == 0))
                    state = !row.isOpen();

                row.setOpen(state);

                if (_closedRows == null)
                    _closedRows = new Vector<String>();

                if (state && _closedRows.contains(rowId))
                    _closedRows.remove(rowId);
                else if (!state && !_closedRows.contains(rowId))
                    _closedRows.add(rowId);
            }
        }
        else
            super.actionDemand(demand);
    } // end method actionDemand

    /**
     * Invoked when a Drag & Drop occurs. (LyDemandListener interface) This method handles: opening
     * and closing line, moving an interval, action
     *
     * @param demand drag-drop demand
     */
    public void dragDropDemand(LyDragDropDemand demand)
    {
        String demandId = demand.getValue();

        if (demandId == null)
            return;

        if (demand.getType() == LyDragDropDemand.TYPE_DRAG_START)
        {
            LyObjectId oid = ((demandId != null) ? getObjectId(demandId) : null);
            LyObject   draggedObject = ((oid != null) ? (LyObject) oid.getObject() : null);

            if ((_tableBehaviorInterface != null) && (draggedObject != null))
                _tableBehaviorInterface.dragStart(this, draggedObject);
        }
        else if (demand.getType() == LyDragDropDemand.TYPE_DROP)
        {
            LySession  session = getSession();
            LyObject   draggedObject = (LyObject) session.getValue(LySession.DRAGGED_OBJECT);
            LyObjectId oid = getObjectId(demandId);
            LyObject   droppedObject = (LyObject) oid.getObject();

            if (_tableBehaviorInterface != null)
                _tableBehaviorInterface.drop(this, droppedObject, draggedObject);
        }
        else if (demand.getType() == LyDragDropDemand.TYPE_DRAG_FINISHED)
        {

            if (_tableBehaviorInterface != null)
                _tableBehaviorInterface.dragFinished(this, Integer.parseInt(demandId));
        }
    } // end method dragDropDemand

    /**
     * Invoked when an menu item notification is sent to this listener.
     *
     * <P>This method overriden to create a popup menu : buildPopupMenu.</P>
     *
     * @param demand demand parameters sent by the view component.
     */
    public void menuDemand(LyMenuDemand demand)
    {

        if (demand == null)
            return;

        if (demand.getSource().equals(getId() + "_table_headers"))
        {

            if (!enableFastFilter())
                return;

            // Popup menu on headers of the table : display the fast filter action
            // on corresponding field info
            LyAction action = getAction();

            if (action.hasMark(LyAction.MARK_NO_MENU))
                return;

            // Retrieve field info used to build fast filter
            int                     index = demand.getIndex(0);
            LyInfoList<LyFieldInfo> fields = getFields();

            if (isStatusColumn(null))
                index--;

            if ((index >= 0) && (index < fields.getSize()))
            {
                LyFieldInfo fieldInfo = fields.getInfo(index);
                int[]       p = demand.getPosition();
                int         x = -1, y = -1;
                boolean     extend = (demand.getType() == LyMenuDemand.TYPE_EXTEND_POPUP);

                if (p != null)
                {
                    x = p[0];
                    y = p[1];
                }

                buildFastFilter(fieldInfo, x, y, extend);
            }
        }
        else if (demand.getSource().equals(getId() + "_table"))
        {

            if ((demand.getType() == LyMenuDemand.TYPE_EXTEND_POPUP) && enableFastEdition())
            {
                // An extended popup demand on a table results in the fast edition of an object
                String     item = demand.getItem();
                LyObjectId oid = ((item == null) ? null : getObjectId(item));
                LyObject   object = ((oid == null) ? null : _workSpace.getObject(oid));

                if (object != null)
                {
                    int                     index = demand.getSubIndex();
                    LyInfoList<LyFieldInfo> fields = getFields();

                    if (isStatusColumn(null))
                        index--;

                    if ((index >= 0) && (index < fields.getSize()))
                    {
                        LyFieldInfo fieldInfo = fields.getInfo(index);
                        int[]       p = demand.getPosition();

                        invokeFastEdition(object, fieldInfo, p[0], p[1]);
                    }
                }

                return;
            }

            LyAction action = getAction();

            if (action.hasMark(LyAction.MARK_NO_MENU))
                return;

            // If the table classInfo is not abstract, send it
            LyClassInfo classInfo = getClassInfo();

            if (classInfo.isAbstract())
            {
                // The menu applies on selection
                // if there is only one object selected, transmit its classInfo
                LyObjectList selection = getSelection();

                if ((selection != null) && (selection.getSize() == 1))
                {
                    LyObject selectedObject = selection.getObject(0);

                    classInfo = selectedObject.getClassInfo();
                }
            }

            // Build popup menu
            LyPopupMenu popupMenu = buildPopupMenu(demand, classInfo, (demand.getIndex(0) >= 0));

            // Call the behavior to complete popup menu
            LyBehaviorInterface behavior = getBehavior();

            if (behavior instanceof LyPanelBehaviorInterface)
                ((LyPanelBehaviorInterface) behavior).completePopupMenu(this, popupMenu);

            // Coordinates
            int[] p = demand.getPosition();

            popupMenu.setX(p[0]);
            popupMenu.setY(p[1]);

            LyTable table = getTableComponent();

            if (table == null)
                // An error occured, just return
                return;

            popupMenu.setController(getId());
            popupMenu.setEnvironment(table.getEnvironment());
            table.setPopupMenu(popupMenu);
        }
        else
            super.menuDemand(demand);
    } // end method menuDemand

    /**
     * Build a contextual menu with the given demand based on the given class info. Overridden to
     * add the 'Copy cell content' item.
     * @param demand the menu demand
     * @param classInfo a classInfo
     * @param selection indicates if the demand applies to a selection
     * @return a popup menu
     */
    protected LyPopupMenu buildPopupMenu(LyMenuDemand demand, LyClassInfo classInfo,
        boolean selection)
    {
        LyPopupMenu   popupMenu = super.buildPopupMenu(demand, classInfo, selection);
        LyViewManager viewManager = getViewManager();

        // Add 'Copy cell content' item only if view manager is not a web one
        if ((popupMenu != null) && !viewManager.hasProperty("WEB_MANAGER"))
        {
            // Retrieve cell content
            int rowIndex = demand.getIndex(0);
            int colIndex = demand.getSubIndex();

            if ((rowIndex >= 0) && (colIndex >= 0))
            {
                LyTable table = getTableComponent();
                LyRows  rows = table.getRows();
                LyRow   row = rows.getRow(rowIndex);

                if (row != null)
                {
                    LyComponent component = row.getChild(colIndex);
                    String      text = null;

                    if (component instanceof LyCell)
                        text = ((LyCell) component).getLabel();

                    if ((text != null) && (text.length() > 0))
                    {
                        String label = translate("LY_ITEM_COPY_CELL_CONTENT");

                        if ((label != null) && (label.length() > 0) && !"_null".equals(label))
                        {
                            LyItem item = new LyItem(DEMAND_COPY_CELL_CONTENT + text,
                                    "LY_ITEM_COPY_CELL_CONTENT");

                            item.setEnabled(true);
                            item.setImage(getImageFromPrefix("LY_ACTION_COPY_CELL_CONTENT",
                                    DEFAULT_IMAGE_FORMAT, DEFAULT_MENU_ITEM_IMAGE_SIZE));

                            if (popupMenu.getComponentCount() > 0)
                                popupMenu.addSep();

                            popupMenu.addItem(item);
                        }
                    }
                }
            }
        }

        return popupMenu;
    } // end method buildPopupMenu

    /**
     * New sort or extended sort on given index.
     * @param index index of the column
     * @param extend indicates if we extend the current sort
     */
    protected void sort(int index, boolean extend)
    {
        LyInfoList<LyFieldInfo> fields = getFields();
        LyFieldInfo             fieldInfo;

        if (isStatusColumn(null))
            index--;

        if (index >= 0)
            fieldInfo = fields.getInfo(index);
        else
            fieldInfo = null;

        sort(fieldInfo, extend);
    }

    /**
     * check if the status column exists.
     *
     * @param  statusFields (optional) list of status fields
     * @return true if there is a status column
     */
    public boolean isStatusColumn(LyInfoList<LyFieldInfo> statusFields)
    {

        if (getEnvironment().testResource("LY_ERROR_STATE_ICON"))
            return true;

        if (statusFields == null)
            statusFields = getStatusFieldInfos();

        return ((statusFields != null) && (statusFields.getSize() > 0));
    }

    /**
     * Check if there is an image to display first in list display.
     * @param fields Fields
     * @return True if there is an image field
     */
    public boolean isImageColumn(LyInfoList<LyFieldInfo> fields)
    {

        for (int i = 0; i < fields.getSize(); i++)
        {

            if ((fields.getInfo(i) instanceof LyFileFieldInfo)
                    && LyFileFieldInfo.Type.IMAGE.equals(
                        ((LyFileFieldInfo) fields.getInfo(i)).getType()))
                return true;
        }

        return false;
    }

    /**
     * Returns the list of field infos that represent the object status for this table.
     *
     * @return a list of field infos or null if no field is used for object status.
     */
    public LyInfoList<LyFieldInfo> getStatusFieldInfos()
    {

        if (_statusFieldMarks == null)
            return null;

        return _classInfo.getFieldInfos(_statusFieldMarks, true);
    }

    /**
     * New sort or extended sort on given field.
     * @param fieldInfo the new fieldinfo
     * @param extend indicates if we extend the current sort
     */
    protected void sort(LyFieldInfo fieldInfo, boolean extend)
    {
        if ((fieldInfo != null) && fieldInfo.hasMark(LyFieldInfo.MARK_SECRET))
        	return;

		if (!allowSortOnField(fieldInfo))
			return;

        LySort sort = getSort();

        // If fieldInfo is null, we must sort by status...
        if (fieldInfo != null)
        {
            LySortCriterion criterion = ((sort == null) ? null : sort.getCriterion(fieldInfo));

            if ((_statusFieldMarks != null) && fieldInfo.hasMarks(_statusFieldMarks, true))
                _currentStatusSort = 0;

            if ((sort != null) && ((criterion != null) && (criterion.isLocked())))
            {
                // Duplicate criteria FFT003648
                LySort          newSort = new LySort();
                int             i, n = sort.getCriterionCount();
                LySortCriterion oldCriterion, newCriterion;

                for (i = 0; i < n; i++)
                {
                    oldCriterion = sort.getCriterion(i);
                    newCriterion = new LySortCriterion(oldCriterion.getFieldInfo(),
                            oldCriterion.isIncreasing());
                    newCriterion.setGroupBy(oldCriterion.isGroupBy());
                    newSort.addCriterion(newCriterion);

                    if (criterion == oldCriterion)
                        criterion = newCriterion;
                }

                sort = newSort;
            }

            if ((sort == null)
                    || (!extend && ((sort.getCriterionCount() > 1) || (criterion == null))))
            {

                if (sort == null)
                    sort = new LySort();
                else
                {
                    // Keep groupBy criteria
                    LySort          newSort = new LySort();
                    int             i, n = sort.getCriterionCount();
                    LySortCriterion newCriterion;

                    for (i = 0; i < n; i++)
                    {
                        newCriterion = sort.getCriterion(i);

                        if (newCriterion.isGroupBy())
                            newSort.addCriterion(newCriterion);
                    }

                    sort = newSort;
                }

                if (criterion == null)
                {
                    LySortCriterion defaultCriterion = getDefaultCriterion(fieldInfo);

                    if (defaultCriterion == null)
                        sort.addCriterion(fieldInfo, true);
                    else
                        sort.addCriterion(defaultCriterion);
                }
                else
                {
                    criterion.setIncreasing(!criterion.isIncreasing());
                    sort.addCriterion(criterion);
                }

                _currentStatusSort = 0;
            }
            else if (criterion != null)
                // Invert sort criterion on fieldInfo
                criterion.setIncreasing(!criterion.isIncreasing());
            else
            {
                // Add sort criterion on fieldInfo in current sort
                LySortCriterion defaultCriterion = getDefaultCriterion(fieldInfo);

                if (defaultCriterion == null)
                    sort.addCriterion(fieldInfo, true);
                else
                    sort.addCriterion(defaultCriterion);
            }
        }
        else
        {
            int oldStatus = _currentStatusSort;

            if ((sort == null) || !extend)
            {
                sort = new LySort();
                oldStatus = 0;
            }

            if (_currentStatusSort == 0)
                _currentStatusSort = 1;
            else
                _currentStatusSort = -_currentStatusSort;

            LyInfoList<LyFieldInfo> fieldInfos = getStatusFieldInfos();
            int                     i, size = ((fieldInfos != null) ? fieldInfos.getSize() : 0);
            LySortCriterion         criterion;

            for (i = 0; i < size; i++)
            {
                fieldInfo = fieldInfos.getInfo(i);
                criterion = ((sort == null) ? null : sort.getCriterion(fieldInfo));

                if (oldStatus == 0)
                    sort.addCriterion(fieldInfo, (_currentStatusSort > 0));
                else
                    criterion.setIncreasing((_currentStatusSort > 0));
            }
        }

        if ((_workSpace instanceof LyPageWorkSpace) && (!checkSortForPhysicalLayer(sort)))
        {
            // ask the user if he really want to apply this filter
            LyViewManager vm = getViewManager();
            vm.askQuestion(_component, "LY_CONFIRM_NOT_PHYSICAL_SORT", this, sort);
        }
        else
            setSort(sort);
    } // end method sort

	/**
	 * Indicates whether sort is allowed on multiple relations.
	 */
	public boolean allowSortOnField(LyFieldInfo fieldInfo)
	{
		if ((_workSpace instanceof LyPageWorkSpace)
    			&& (fieldInfo instanceof LyRelationFieldInfo)
    			&& fieldInfo.isMultiple())
			return false;

		return true;
	}

    /**
     * Set current sort criteria on the view.
     *
     * @param sort new sort criteria
     */
    public void setSort(LySort sort)
    {
        LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();

        sort = behavior.validateSort(this, sort);

        if ((sort == null) && (_sort == null))
            return; // nothing to do

        if ((_sort != null) && (_sort.equals(sort)))
            return; // nothing to do

        LySort oldSort = _sort;

        _sort = sort;

        if (isIdle())
            return;

        getViewManager().setWaiting(true);

        // Freeze the table
        setIdle(true);

        // Sort data : list will be sorted when receiving object list events
        _workSpace.setSort(_sort);

        updateFields(oldSort);

        if ((_borderPanelController instanceof LySortWizardController)
                && !_borderPanelController.isFreed())
            ((LySortWizardController) _borderPanelController).setSort(_sort);

        // Update the last modification time.
        setLastModified(new Date().getTime());

        setIdle(false);

        getViewManager().setWaiting(false);
    } // end method setSort

    /**
     * Get the colum size for a fieldInfo.
     * @param fieldInfo a fieldInfo
     * @return a positive double for a fixed sized a negative one otherwise
     */
    public double getColumnSize(LyFieldInfo fieldInfo)
    {
        LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();
        double                   colWidth = behavior.getColumnWidth(this, fieldInfo);

        if (colWidth >= 0)
            return colWidth;

        colWidth = -1.0;

        if (_columnSizes != null)
        {
            String[] sizes = LyString.stringToList(_columnSizes, "##");

            if (sizes != null)
            {
                int      i, n = sizes.length;
                String   size;
                String[] fieldIdSize;

                for (i = 0; i < n; i++)
                {
                    size = sizes[i];

                    if (!size.startsWith(fieldInfo.getId() + "="))
                        continue;

                    fieldIdSize = LyString.stringToList(sizes[i], '=');

                    if ((fieldIdSize == null) || (fieldIdSize.length != 2))
                        continue;

                    colWidth = LyTools.doubleFromString(fieldIdSize[1]);
                    break;
                }
            }
        }

        return colWidth;
    }

    /**
     * Sets the size of the headers.
     */
    protected void setColumnSizes(String columnSizes)
    {
        _columnSizes = columnSizes;

        // Parse columnSizes parameter
        String[] sizes = LyString.stringToList(columnSizes, "##");

        if ((sizes != null) && (sizes.length > 0))
        {
            LyTable table = getTableComponent();

            if (table == null)
                return;

            LyHeaders headers = table.getHeaders();

            if (headers != null)
            {
                LyHeader header;

                for (int i = 0; i < sizes.length; i++)
                {
                    String[] fieldIdSize = LyString.stringToList(sizes[i], '=');

                    if ((fieldIdSize == null) || (fieldIdSize.length != 2))
                        continue;

                    header = (LyHeader) headers.getComponent("header" + ID_SEP + fieldIdSize[0]);

                    if (header == null)
                    {
                        LyInfo fieldInfo = getApplication().getInfo(fieldIdSize[0]);

                        if (fieldInfo != null)
                            header = (LyHeader) headers.getComponent("header" + ID_SEP
                                    + fieldIdSize[0] + "_" + fieldInfo.hashCode());
                    }

                    if (header == null)
                        continue;

                    header.setColumnWidth(LyTools.doubleFromString(fieldIdSize[1]));
                }
            }
        }
    }

    /**
     * Update fields according to sort.
     */
    protected void updateFields(LySort oldSort)
    {
        LyInfoList<LyFieldInfo> fields = null;
        int                     i, n = ((_fields == null) ? 0 : _fields.getSize());
        LyFieldInfo             fieldInfo;
        int                     index;

        if (_sort != null)
        {

            // Remove columns for grouped fields
            for (i = n - 1; i >= 0; i--)
            {
                fieldInfo = _fields.getInfo(i);
                index = _sort.getIndex(fieldInfo);

                if ((index >= 0) && _sort.getCriterion(index).isGroupBy())
                {

                    if (fields == null)
                        fields = new LyInfoList<LyFieldInfo>(_fields);

                    fields.removeInfo(fieldInfo);
                }
            }
        }

        if (oldSort != null)
        {
            // Add columns for ungrouped fields
            LyInfoList<LyFieldInfo> groupByFields = oldSort.getGroupByFieldInfos();
            int                     pos = 0;

            n = ((groupByFields == null) ? 0 : groupByFields.getSize());

            if ((n == 1) && (_groupByFieldOldIndex >= 0))
            	pos = _groupByFieldOldIndex;

            for (i = 0; i < n; i++)
            {
                fieldInfo = groupByFields.getInfo(i);
                index = ((_sort == null) ? -1 : _sort.getIndex(fieldInfo));

                if ((index < 0)
                        || (oldSort.getCriterion(fieldInfo).isGroupBy()
                            && !_sort.getCriterion(index).isGroupBy()))
                {

                    if (fields == null)
                        fields = new LyInfoList<LyFieldInfo>(_fields);

                    if (!fields.contains(fieldInfo))
                    {
                        fields.addInfo(fieldInfo, pos);
                        pos++;
                    }
                }
            }
        }

        if (fields != null)
            setFields(fields);
    }

    /**
     * Sort on the status icon (represents on or several fields as an incon to the left of the
     * view). This might be a multiple sort criteria on all status fields.
     *
     * @param invert true indicates that we want to have a reverse sort on the status icons
     */
    protected void setStatusSort(boolean invert)
    {
        LyInfoList<LyFieldInfo> fieldInfos = getStatusFieldInfos();
        LySort                  sort = getSort();
        int                     oldStatus = _currentStatusSort;

        if (!invert)
        {
            sort = new LySort();
            oldStatus = 0;
            _currentStatusSort = 1;
        }
        else
            _currentStatusSort = -_currentStatusSort;

        int size = ((fieldInfos == null) ? 0 : fieldInfos.getSize());

        for (int i = 0; i < size; i++)
        {
            LyFieldInfo     fieldInfo = fieldInfos.getInfo(i);
            LySortCriterion criterion = ((sort == null) ? null : sort.getCriterion(fieldInfo));

            if (oldStatus == 0)
                sort.addCriterion(fieldInfo, (_currentStatusSort > 0));
            else
                criterion.setIncreasing((_currentStatusSort > 0));
        }

        setSort(sort);
    } // end method setStatusSort

    /**
     * Group rows by given field info values.
     * @param fieldInfo Field info used for grouping table rows.
     */
    protected void groupBy(LyFieldInfo fieldInfo)
    {
        LySort	sort = null;
        int		groupByFieldOldIndex = -1;

        if ((_sort != null) && (_sort.getCriterionCount() > 0)
                && (_sort.getCriterion(0).getFieldInfo() == fieldInfo)
                && (_sort.getCriterion(0).isGroupBy()))
            fieldInfo = null;
        else if (allowSortOnField(fieldInfo))
        {
            // Add field info on sort
            LySortCriterion sortCriterion;

            sort = new LySort();
            sortCriterion = sort.addCriterion(fieldInfo, true);
            sortCriterion.setGroupBy(true);

            if ((_fields != null) && _fields.contains(fieldInfo))
            	groupByFieldOldIndex = _fields.getIndex(fieldInfo);
        }

        if (_closedRows != null)
            _closedRows.clear();

        _closedRows = null;

        // Store sort and groupBy field
        setSort(sort);

        // Update menu
        updateGroupByMenu();

        // Update content
        reset(false);

        _groupByFieldOldIndex = groupByFieldOldIndex;
    }

    /**
     * Select an object in the table.
     *
     * @param object selected or unselected object
     * @param select new selection status
     */
    public void selectObject(LyObject object, boolean select)
    {
        selectObject(object, select, true, true);
    }

    /**
     * Select an object in the table.
     *
     * @param object       selected or unselected object
     * @param select       new selection status
     * @param updateView   indicates if we must update the views
     * @param showSelected indicates wether the view should scroll to the new selection
     */
    public void selectObject(LyObject object, boolean select, boolean updateView,
        boolean showSelected)
    {

        if (object == null)
        {
            selectObjects(null, select);
            return;
        }

        if ((select && (_selections != null) && _selections.contains(object))
                || (!select && ((_selections == null) || !_selections.contains(object))))
            return;

        selectTableRow(object, select, showSelected);

        if (select)
        {

            if (_selections == null)
            {
                _selections = new LySimpleObjectList(object);
                _selections.setSorted(false);
            }
            else
                _selections.addObject(object);
        }
        else if ((_selections != null) && (_selections.contains(object)))
            _selections.removeObject(object);

        object.setSelected(select);

        if (updateView)
        {
            updateContextClient();
            updateTools();
            updateMessages();
        }

        // Update the last modification time.
        setLastModified(new Date().getTime());
    } // end method selectObject

    /**
     * Select all objects in panel.
     *
     * @param select new selection status
     */
    public void selectAllObjects(boolean select)
    {
        selectObjects(null, select);
    }

    /**
     * Select all given objects in the table.
     *
     * @param objects Objects to select/deselect
     * @param select  new selection status
     */
    public void selectObjects(LyObjectList objects, boolean select)
    {
        int          no = ((objects == null) ? 0 : objects.getSize());
        LyObjectList selections = ((_selections == null) ? null
                                                         : new LySimpleObjectList(_selections));

        if (no == 0)
        {
            LyTable table = getTableComponent();
            LyRows  rows = table.getRows();

            synchronized (_workSpace)
            {
                // If there is a limit but no page workspace (may happen when a context
                // is set on the table with a limit)
                int limit = getLimit();

                if (select)
                {

                    if ((limit > 0) && !(_initialWorkSpace instanceof LyPageWorkSpace))
                    {
                        int i, startIndex = _currentPage * _maxRowCount;
                        int endIndex = Math.min(startIndex + _maxRowCount, _workSpace.getSize());

                        _selections = new LySimpleObjectList();

                        for (i = startIndex; i < endIndex; i++)
                            _selections.addObject(_workSpace.getObject(i));
                    }
                    else
                        _selections = new LySimpleObjectList(_workSpace);

                    _selections.setSorted(false);
                }
                else
                    _selections = null;

                if (rows != null)
                {
                    Vector<LyRow> subrows = new Vector<LyRow>();

                    rows.getDisplayedSubRows(subrows);

                    if (subrows != null)
                    {

                        for (LyRow subrow : subrows)
                        {

                            if ((subrow.getId() != null)
                                    && !subrow.getId().startsWith(SECTION_ROW_PREFIX)
                                    && !subrow.getId().startsWith(FOOTER_ROW_PREFIX))
                                subrow.setSelected(select);
                        }
                    }
                }

                LyObject object;
                int      i, n = _workSpace.getSize();

                for (i = 0; i < n; i++)
                {
                    object = _workSpace.getObject(i);
                    object.setSelected((_selections != null) && (_selections.contains(object)));
                }
            }

            if (table.getInternalView() instanceof LySetTableInterface)
            {
                LySetTableInterface tableInterface = (LySetTableInterface) table.getInternalView();

                if (tableInterface != null)
                    tableInterface.selectAll(select, true);
            }
        }
        else
        {
            int      i;
            LyObject object;

            // First unselect all the objects from the workspace that are not in the selection
            for (i = 0; i < _workSpace.getSize(); i++)
            {
                object = _workSpace.getObject(i);

                if (!objects.contains(object))
                    selectObject(object, false, false, false);
            }

            if (select)
            {

                // Now select all the objects from the selection list
                for (i = 0; i < no; i++)
                {
                    object = objects.getObject(i);
                    selectObject(object, true, false, false); //(i == 0));
                }
            }
        }

        // Notify selection if selections list changed
        if ((_selections != selections)
                && ((selections == null) || (_selections == null)
                    || !_selections.equals(selections)))
        {
            updateContextClient();
            updateTools();
            updateMessages();

            // Update the last modification time.
            setLastModified(new Date().getTime());

            LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();

            behavior.selectionChanged(this);

            if (notifyOutline())
                _session.notifyOutline(this);
        }
    } // end method selectObjects

    /**
     * Selection of a row in a table (graphical component). Internal use of the Controller. This
     * method only switch the selection status in the view (graphical user interface impact). The
     * internal selection list (getSelection()) is not set by this method. Use selectObject to
     * change the selection status of an object:
     *
     * @see   #selectObject
     * @param object       selected or unselected object. May not be null
     * @param select       new selection status
     * @param showSelected indicates wether the view should scroll to the new selection status
     */
    protected void selectTableRow(LyObject object, boolean select, boolean showSelected)
    {

        if (object == null)
            return;

        LyTable             table = getTableComponent();
        LySetTableInterface tableInterface = null;

        if (table.getInternalView() instanceof LySetTableInterface)
            tableInterface = (LySetTableInterface) table.getInternalView();

        synchronized (_workSpace)
        {
            LyRows rows = table.getRows();
            LyRow  row = findRow(rows, object.getId());
            int    idx = getRowIndex(object);

            if (idx < 0)
                return;

            if ((row == null) || (row.isSelected() == select))
            {

                if (showSelected && select && (tableInterface != null))
                    tableInterface.showRow(idx);

                return;
            }

            row.setSelected(select);

            if (tableInterface != null)
                tableInterface.selectRow(idx, select);

            if (showSelected && select && (tableInterface != null))
                tableInterface.showRow(idx);
        }
    } // end method selectTableRow

    /**
     * Scrolls list to make the given object visible.
     *
     * @param object Object to be shown. If object is null, the first row is shown.
     */
    public void showObject(LyObject object)
    {

        if (object == null)
            return;

        LyTable             table = getTableComponent();
        LySetTableInterface tableInterface = null;

        if (table.getInternalView() instanceof LySetTableInterface)
            tableInterface = (LySetTableInterface) table.getInternalView();

        if (tableInterface == null)
            return;

        synchronized (_workSpace)
        {
            int idx = _workSpace.getIndex(object);

            if (idx < 0)
                idx = 0;

            tableInterface.showRow(idx);
        }
    }

    /**
     * Hide a row in a table (graphical component). Internal use of the Controller. This method only
     * switch the hidden status in the view (graphical user interface impact).
     *
     * @param object hidden or shown object. May not be null
     * @param hide   new hidden status
     */
    protected void hideTableRow(LyObject object, boolean hide)
    {

        if (object == null)
            return;

        LyTable             table = getTableComponent();
        LySetTableInterface tableInterface = null;

        if (table.getInternalView() instanceof LySetTableInterface)
            tableInterface = (LySetTableInterface) table.getInternalView();

        synchronized (_workSpace)
        {
            int idx = _workSpace.getIndex(object);

            if (idx < 0)
                return;

            LyRow row = getRow(object, idx);

            row.setHidden(hide);

            if (tableInterface != null)
                tableInterface.hideRow(idx, hide);
        }

        // Update the last modification time.
        setLastModified(new Date().getTime());
    } // end method hideTableRow

    /**
     * Reverse Sort.
     */
    protected void reverseSort()
    {
        LySort sort = getSort();

        if (sort != null)
        {
            int i, n = sort.getCriterionCount();

            for (i = 0; i < n; i++)
            {
                LySortCriterion criterion = sort.getCriterion(i);

                if (criterion != null)
                // Invert sort criterion on fieldInfo
                    criterion.setIncreasing(!criterion.isIncreasing());
            }

            setSort(sort);
        }
    }

    /**
     * Set or create a new fast Filter.
     *
     * @param fieldInfo     field info used to build the new fast filter
     * @param x             x position where fast filter window must be displayed
     * @param y             y position where fast filter window must be displayed
     * @param extend        indicates if current filter must be extended with new one
     */
    protected void buildFastFilter(LyFieldInfo fieldInfo, int x, int y, boolean extend)
    {
        // Create a new fast filter controller
        String                  actionId = LyAction.ID_FAST_FILTER;
        LySession               session = getSession();
        LyApplication           application = session.getApplication();
        LyAction                action = application.getActionWithId(actionId, _classInfo);
        LyInfoList<LyFieldInfo> fields = _classInfo.getFieldInfos(LyFieldInfo.MARK_FILTER);

        if ((fields == null) || (fields.getSize() == 0))
        {
            fields = _classInfo.getFieldInfos();
            fields = fields.getInfos(LyFieldInfo.MARK_HIDDEN, true);
            fields = fields.getInfos(LyFieldInfo.MARK_SECRET, true);
        }

        // Check if field really belongs to class info (may happen in complex table actions)
        if ((fields != null) && (fields.getSize() > 0) && !fields.contains(fieldInfo))
            return;

        getViewManager().setWaiting(true);

        _fastFilterFieldInfo = fieldInfo;
        _fastFilterExtend = extend;

        // Create the fast filter controller and display it
        LyController filterController = invokeAction(action, _classInfo, null, false);

        if ((filterController != null) && !filterController.isFreed())
        {
            // Set the position of the filter window
            LyViewController viewController = filterController.getViewController();

            viewController.initialize();

            // Set location before showing the view to avoid a clipping effect
            if ((x >= 0) && (y >= 0))
                viewController.setLocation(x, y);

            filterController.showView();

            LyView view = viewController.getView();
            int[]  pos = view.getLocation();

            if ((x <= 0) && (y <= 0))
            {
                x = pos[0];
                y = pos[1];
            }

            // Ensure view is at the right position now that view is shown and has its size
            // computed (avoid view being out of screen)
            int[] screenSize = viewController.getViewManager().getScreenSize(view);
            int[] viewSize = view.getSize();

            if ((viewSize != null) && (screenSize != null))
            {
                x = Math.max(Math.min(x, screenSize[0] + screenSize[2] - viewSize[0]), 0);
                y = Math.max(Math.min(y, screenSize[1] + screenSize[3] - viewSize[1]), 0);

                viewController.setLocation(x, y);
            }
        }

        getViewManager().setWaiting(false);
    } // end method buildFastFilter

    /**
     * Gets initial fast filter value : indicates to the fast filter controller the default filter
     * value.
     *
     * @return the initial fast filter value
     */
    public String getFastFilterValue()
    {
        if (_fastFilterFieldInfo == _lastFastFilterFieldInfo)
            return _fastFilterValue;

        return null;
    }

    /**
     * Gets initial fast filter field info : indicates to the fast filter controller the default
     * filter field info.
     *
     * @return the initial fast filter field info or null if no filter is set
     */
    public LyFieldInfo getFastFilterFieldInfo()
    {
        return _fastFilterFieldInfo;
    }

    /**
     * Indicates if fast filter creation is an extension of current filter.
     *
     * @return true if fast filter creation is an extension of current filter, false otherwise
     */
    public boolean isFastFilterExtended()
    {
        return _fastFilterExtend;
    }

	/**
	 * Indicates if a fast filter is set.
	 *
	 * @return true if a fast filter is set, false otherwise
	 */
	public boolean isFastFilterSet()
	{
		return (_lastFastFilterFieldInfo != null);
	}

    /**
     * Activate or deactivate the filter.
     *
     * @param status true if the current filter is active (apply), or is inactive (no filter, but
     *               filter is kept for possible further use).
     */
    protected void activateFilter(boolean status)
    {
        _filterActive = status;

        if (isIdle())
            return;

        if (_workSpace == null)
            return;

        if ((_initialWorkSpace == null) || _openFilterIfEmpty) // not initialized or openFilterIfEmpty
        {
            getViewManager().setWaiting(true);
            reset(true);
            getViewManager().setWaiting(false);
            return;
        }

        // Freeze the table
        setIdle(true);

        getViewManager().setWaiting(true);

        if (!(_initialWorkSpace instanceof LyPageWorkSpace))
        {
            if (status)
            {
                _workSpace.removeListener(this);

                if (_workSpace == _initialWorkSpace)
                {
                    _workSpace = new LySubWorkSpace(getSession(), _initialWorkSpace, _filter,
                            true);

                    if (getSort() == null)
                        _workSpace.setSort(_initialWorkSpace.getSort());
                    else
                    {
                        LySort sort = getSort(), wsSort = _workSpace.getSort();

                        if ((sort != wsSort)
                                && ((sort == null)
                                    || ((sort != null) && (!sort.equals(wsSort)))))
                            _workSpace.setSort(sort); // sort has changed
                    }
                }
                else
                {
                    ((LySubWorkSpace) _workSpace).setFilter(_filter);

                    LySort sort = getSort(), wsSort = _workSpace.getSort();

                    if ((sort != wsSort)
                            && ((sort == null)
                                    || !sort.equals(wsSort)))
                        _workSpace.setSort(sort); // sort has changed
                }

                _workSpace.addListener(this);
            }
            else if (_workSpace != _initialWorkSpace)
            {
                // Keep the current workspace
                //FFT00540
                LyWorkSpace temp = _workSpace;

                LySort sort = getSort(), wsSort = _initialWorkSpace.getSort();

                if ((sort != wsSort)
                        && ((sort == null)
                                || !sort.equals(wsSort)))
                    _initialWorkSpace.setSort(getSort()); // sort has changed

                _workSpace = _initialWorkSpace;

                _workSpace.addListener(this);
                //FFT00540
                temp.free(this);
                temp = null;
            }
        }
        else // case of a page workspace
        {

            if (status)
            {

                // set the filter of the workspace
                if (_classFilter != null)
                { // create a new filter that has the conditions of the class filter and the current filter

                    LyFilter newFilter = new LyFilter(LyFilter.Type.AND);

                    newFilter.addElement(_classFilter);
                    newFilter.addElement(_filter);
                    // set this new filter to the workspace
                    ((LyPageWorkSpace) _workSpace).setFilterAndSort(newFilter, getSort(), true);
                }
                else
                    ((LyPageWorkSpace) _workSpace).setFilterAndSort(_filter, getSort(), true);
            }
            else // When the filtering  is removed by the user, the class filter is put on the current workspace
                ((LyPageWorkSpace) _workSpace).setFilterAndSort(_classFilter, getSort(), true);
        }

        //updateFilterMenu();
        // Update the last modification time.
        setLastModified(new Date().getTime());

        // Restore the previous idle state
        setIdle(false);

        getViewManager().setWaiting(false);
    } // end method activateFilter

    /**
     * Choose columns.
     */
    protected void chooseColumns()
    {
        // Build view to choose columns
        boolean close = ((_borderPanelController instanceof LyChooseFieldsController)
                && !_borderPanelController.isFreed());

        closeBorderPanel();

        if (close)
            return;

        getViewManager().setWaiting(true);

        LyApplication application = getApplication();
        LyClassInfo   classInfo = getClassInfo();
        LyAction      action = application.getActionWithId(LyAction.ID_CHOOSE, classInfo);

        _borderPanelController = (LyChooseFieldsController) invokeAction(action, classInfo, null,
                false);

        if (_borderPanelController != null)
            _borderPanelController.showView();

        getViewManager().setWaiting(false);
    } // end method chooseColumns

    /**
     * Sort Wizard.
     */
    protected void sortWizard()
    {
        boolean close = ((_borderPanelController instanceof LySortWizardController)
                && !_borderPanelController.isFreed());

        closeBorderPanel();

        if (close)
            return;

        // Create a new sort controller
        getViewManager().setWaiting(true);

        LyApplication application = getApplication();
        LyAction      action = application.getActionWithId(LyAction.ID_SORT_WIZARD, _classInfo);

        _borderPanelController = (LySortWizardController) invokeAction(action, _classInfo);

        if (_borderPanelController != null)
            _borderPanelController.showView();

        getViewManager().setWaiting(false);
    } // end method sortWizard

    /**
     * Freeze or unFreeze a column.
     *
     * @param column column index to fix
     */
    public void fixColumn(int column)
    {
        _fixedColumn = column;

        if (isIdle())
            return;

        LyTable             table = getTableComponent();
        LySetTableInterface tableInterface = (LySetTableInterface) table.getInternalView();
        // FFT00132 : wrong column fixed if status column
        int status = 0;

        if (isStatusColumn(null))
            status = 1;

        tableInterface.freezeColumn(column + status);

        boolean hasMenu = hasMenu();

        if (hasMenu)
        {
            // ** Filtering menu status **
            LyView    view = getView();
            LyMenuBar menuBar = ((view instanceof LyFrame) ? ((LyFrame) view).getMenuBar() : null);

            if (menuBar != null)
            {
                LyItem item;
                int    i;

                // Unfix column item
                for (i = 0;; i++)
                {
                    item = menuBar.findItem(DEMAND_FIX_COLUMN + ID_SEP + i, this.getId());

                    if (item == null)
                        break;
                    else
                        item.setEnabled(true);
                }

                item = menuBar.findItem(DEMAND_UNFIX_COLUMN, this.getId());

                if (item != null)
                    item.setEnabled((column >= 0));

                item = menuBar.findItem(DEMAND_FIX_COLUMN + ID_SEP + column, this.getId());

                if (item != null)
                    item.setEnabled(false);

                // FFT00113 : do not display the item for last column
                // Last item always inactive : can not fix last column
                /*item = menuBar.findItem(DEMAND_FIX_COLUMN + ID_SEP + (i-1), this.getId());
                if (item != null)
                        item.setEnabled(false);*/
            }
        }
        // Update the last modification time.
        setLastModified(new Date().getTime());
    } // end method fixColumn

    /**
     * Find object in the table and select it.
     *
     * @param find      the find controller
     * @param from      index to start from (-1 : begin)
     * @param next      find command is searching forward (next selection) or previous.
     * @return index of the found object
     */
    public int findObject(LyFindController find, int from, boolean next)
    {
        LyTable            table = getTableComponent();
        LySetIdleInterface tableInterface = null;

        if (table.getInternalView() instanceof LySetIdleInterface) // FORCE REFRESH
            tableInterface = (LySetIdleInterface) table.getInternalView();

        if (tableInterface != null)
            tableInterface.setIdle(true);

        int idx = selectRow(find.getFieldInfo(), find.getPattern(), from, next, !find.isWholeWord(),
                find.isCaseSensitive());

        LyObject object = (((_selections != null) && (_selections.getSize() > 0))
                ? _selections.getObject(0) : null);

        if (tableInterface != null)
            tableInterface.setIdle(false); // FORCE REFRESH

        // Force selection of first selected object (needed for example in multi-page
        // tables in web mode to change currently selected page).
        if (object != null)
            selectTableRow(object, true, true);

        return idx;
    } // end method findObject

    /**
     * Select object in the panel for given parameters.
     * @param fieldInfo the fieldinfo
     * @param pattern the pattern to select
     * @param from the start index
     * @param next indicates the direction of the search
     * @param substring indicates if we look at substring
     * @param caseSensitive indicates if the search is case sensitive
     *
     * @return the index of the selected row, -1 if none has been found
     */
    public int selectRow(LyFieldInfo fieldInfo, String pattern, int from, boolean next,
        boolean substring, boolean caseSensitive)
    {

        if ((fieldInfo == null) || (pattern == null))
            return -1;

        int n = _workSpace.getSize();

        if (n == 0)
            return -1;

        getViewManager().setWaiting(true);

        // Search item
        if (!caseSensitive)
            pattern = pattern.toLowerCase();

        if ((from < 0) || (from >= n))
        {

            // Retrieve first selection
            if ((_selections != null) && (_selections.getSize() > 0))
            {
                LyObject first = _selections.getObject(0);

                from = _workSpace.getIndex(first);
            }
        }

        LyObject                 object = null;
        Object                   value;
        String                   str1, str2;
        boolean                  paginated = (_workSpace instanceof LyPageWorkSpace);
        int                      idx = (((from < 0) || (from > n)) ? (next ? -1 : n) : from);
        int                      count;
        LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();
        boolean                  wasIdle = isIdle();
        int                      nt = (paginated ? ((LyPageWorkSpace) _workSpace).getTotalCount()
                                                 : n);
        boolean                  found = false;
        boolean                  ignoreAccent = getEnvironment().testResource("LY_IGNORE_TEXT_ACCENT");

        for (count = 0; (count < nt) && !found; count++)
        {

            if (next)
            {

                if (idx == (n - 1))
                {

                    if (paginated)
                    {
                        setIdle(true);
                        loadPage(next);
                        n = _workSpace.getSize();
                    }

                    idx = 0;
                }
                else
                    idx++;
            }
            else
            {

                if (idx == 0)
                {

                    if (paginated)
                    {
                        setIdle(true);
                        loadPage(next);
                        n = _workSpace.getSize();
                    }

                    idx = n - 1;
                }
                else
                    idx--;
            }

            object = _workSpace.getObject(idx);
            value = object.getValue(fieldInfo);
            str1 = behavior.getCellLabel(this, object, fieldInfo, value);
            str2 = fieldInfo.toName(value);

            // If field is a rich text, remove tags before doing comparison
            if ((fieldInfo instanceof LyTextFieldInfo)
                    && (((LyTextFieldInfo) fieldInfo).getType() == LyTextFieldInfo.Type.RICH))
            {
                str1 = LyTools.removeHtmlTags(str1);
                str2 = LyTools.removeHtmlTags(str2);
            }

            if (ignoreAccent)
            {
            	str1 = LyString.getSimpleString(str1);
            	str2 = LyString.getSimpleString(str2);
            	pattern = LyString.getSimpleString(pattern);
            }

            if (str1 != null)
            {

                if (!caseSensitive)
                    str1 = str1.toLowerCase();

                found = ((substring && (str1.indexOf(pattern) >= 0)) || str1.equals(pattern));
            }
            else if (str2 != null)
            {

                if (!caseSensitive)
                    str2 = str2.toLowerCase();

                found = ((substring && (str2.indexOf(pattern) >= 0)) || str2.equals(pattern));
            }
        }

        if (found)
            selectObjects(new LySimpleObjectList(object), true);
        else
        {
	        if (paginated && (from < 0))
                loadPage(next);

            selectObjects(null, false);
        }

        if (!wasIdle && isIdle())
            setIdle(false);

        getViewManager().setWaiting(false);
        return (found ? idx : -1);
    } // end method selectRow

    /**
     * Update the sort menuitem status.
     */
    protected void updateSortMenu()
    {
    }

    /**
     * Check if the given action is enabled in the context of this controller.
     * @param action the action to check
     *
     * @return True if the action is enabled, false if disabled.
     */
    protected boolean enableAction(LyAction action)
    {
        LySession session = getSession();

        return session.enableAction(action, this, _selections, false);
    }

    /**
     * Class info : content description. (application class)
     *
     * @return the classInfo being shown in this table
     */
    public LyClassInfo getClassInfo()
    {
        return _classInfo;
    }

    /**
     * Get content of the table.
     *
     * @return the list of objects being displayed
     */
    public LyObjectList getContent()
    {
        return _workSpace;
    }

    /**
     * Gets list of selected objects.
     *
     * @return the list of objects currently selected or null if no object is selected
     */
    public LyObjectList getSelection()
    {
        return getSelection(true);
    }

    /**
     * Gets list of selected objects.
     *
     * @param sorted True if selection list is given with current table sort, false if selection order is desired.
     * @return the list of objects currently selected or null if no object is selected
     */
    public LyObjectList getSelection(boolean sorted)
    {

        if (_selections == null)
            return null;

        LyObjectList selection = new LySimpleObjectList(_selections);

        if (sorted)
            selection.setSort(getSort());

        return selection;
    }

    /**
     * Gets current sort.
     *
     * @return the initial sort or null if no sort is set
     */
    public LySort getSort()
    {

        if (_sort == null)
            return _sort;
        else
            return _sort.cloneSort();
    }

    /**
     * Accessor to the Table component.
     * @return the table component
     */
    public LyTable getTableComponent()
    {
        // Retrieve internal table component
        LyComponent component = getComponent();

        if (component == null)
            return null;

        LyTable table = (LyTable) component.getComponent(getId() + "_table");

        // For file builder
        if (table == null)
            table = (LyTable) getComponent().getComponent("_table");

        if (table == null)
        {
        	LySession	session = getSession();

        	if ((session != null)
        			&& (session.getState() != LySession.State.CLOSING)
        			&& (session.getState() != LySession.State.CLOSED))
        		throw new IllegalStateException("BF AppEngine : can't retrieve table for controller "
        				+ this);
        }

        return table;
    }

    /**
     * Accessor to panel component.
     * @return the panel component
     * @see LySimpleTableController#getTableComponent
     */
    public LyPanel getPanelComponent()
    {
        return getTableComponent();
    }

    /**
     * Return the list of fields displayed in the table component. This list may be different from
     * the real field info list since graphical component may allow the end-user to change the order
     * of the headers (in Swing viewer for example).
     *
     * @return the list of fields displayed in the table component or null if no specific order has
     *         been returned by the table component.
     */
    public LyInfoList<LyFieldInfo> getTableComponentFields()
    {
        return getTableComponentFields(true);
    }

    /**
     * Return the list of fields displayed in the table component. This list may be different from
     * the real field info list since graphical component may allow the end-user to change the order
     * of the headers (in Swing viewer for example).
     *
     * @param classFieldsOnly indicates whether only fields of current class info must be returned.
     * @return the list of fields displayed in the table component or null if no specific order has
     *         been returned by the table component.
     */
    public LyInfoList<LyFieldInfo> getTableComponentFields(boolean classFieldsOnly)
    {
        LyTable         table = getTableComponent();
        LyViewComponent viewComponent = table.getInternalView();

        // Table component may return a value to indicate the new positions of headers.
        if (viewComponent instanceof LyGetValueInterface)
        {
            LyGetValueInterface getValue = (LyGetValueInterface) viewComponent;
            String              value = getValue.getValue();
            String[]            headerIds = LyString.stringToList(value, LyTools.OBJECT_SEP);

            return getFields(headerIds, classFieldsOnly);
        }

        return null;
    } // end method getTableComponentFields

    /**
     * Return the list of fields from an array of header identifiers.
     *
     * @param classFieldsOnly indicates whether only fields of current class info must be returned.
     * @return the list of fields computed from the array of header identifiers.
     */
    protected LyInfoList<LyFieldInfo> getFields(String[] headerIds, boolean classFieldsOnly)
    {
        LyTable                 table = getTableComponent();
        LyFieldInfo             fieldInfo;
        LyInfoList<LyFieldInfo> fields = new LyInfoList<LyFieldInfo>(false);
        int                     j, n = headerIds.length;
        String                  headerId;

        for (j = 0; j < n; j++)
        {
            headerId = headerIds[j];

            if (headerId.startsWith("header" + ID_SEP))
                headerIds[j] = headerId.substring(("header" + ID_SEP).length());

            fieldInfo = _classInfo.getFieldInfo(headerIds[j], false);

            if (fieldInfo == null)
            {

                if (classFieldsOnly)
                {
                    // Try to retrieve real field from parent top headers
                    LyComponent component = table.getComponent(headerId);
                    LyHeaders   headers = null;

                    while ((component != null) && !(component instanceof LyTable))
                    {
                        component = component.getParent();

                        if ((component instanceof LyHeaders) && (component.getId() != null))
                            headers = (LyHeaders) component;
                    }

                    if (headers != null)
                    {
                        String headersId = headers.getId();

                        if ((headersId != null) && headersId.startsWith("headers" + ID_SEP))
                            headersId = headersId.substring(("headers" + ID_SEP).length());

                        fieldInfo = _classInfo.getFieldInfo(headersId, false);
                    }
                }
                else
                {
                    int pos = headerIds[j].lastIndexOf('_');

                    if (pos > 0)
                    {
                        headerIds[j] = headerIds[j].substring(0, pos);
                        fieldInfo = (LyFieldInfo) getApplication().getInfo(headerIds[j]);
                    }
                }
            }

            if (fieldInfo != null)
                fields.addInfo(fieldInfo);
        }

        return fields;
    }

    /**
     * Invokes an action on the specified object list.
     *
     * <P>The action process is implemented on the session.</P>
     *
     * @param action  the action which is executed.
     * @param objects The list of objects on which the action applies. They may not belong to the
     *                same application class (classInfo).
     * @return the resulting controller
     * @see   #invokeAction(LyAction, LyClassInfo, LyObjectList, boolean)
     * @see   leon.app.LySession#invokeAction
     */
    protected LyController invokeAction(LyAction action, LyObjectList objects)
    {

        if ((objects == null) || (objects.getSize() == 0))
            return invokeAction(action, _classInfo, objects, true);
        else
        {

            // Check if the action is on the super class
            if ((objects.getSize() != 1) && !_classInfo.contains(action))
            {
                LyInfoList<LyAction> actions = _classInfo.getActions(LyAction.MARK_HIDDEN, true);
                LyAction             action2;
                int                  i, n = ((actions == null) ? 0 : actions.getSize());

                for (i = 0; i < n; i++)
                {
                    action2 = actions.getInfo(i);

                    if ((action.getRoot() == action2.getRoot())
                            && getImage(action).equals(getImage(action2)))
                        return invokeAction(action, _classInfo, objects, true);
                }
            }

            return super.invokeAction(action, objects);
        }
    }

    /* Implementation of LyWorkSpaceListener */

    /**
     * Called when an object is added in the dynamic object list.
     * @param event the workspace event
     * @return true if the event was handled
     *
     * @see LyWorkSpaceListener#objectAdded
     */
    public boolean objectAdded(LyWorkSpaceEvent event)
    {
        if (isIdle())
            return false;

        if (_component != null)
        {
        	LyObject object = event.getObject();
            int      index = event.getIndex();
            LyWorkSpace ws = _workSpace;

            if (ws == null)
                return true;

            if (index >= 0)
            {
	            synchronized (ws)
	            {
                    // insert new row
                    insertRow(object, index, false);

                    // show it
                    LyTable table = getTableComponent();

                    if (table.getInternalView() instanceof LySetTableInterface)
                        ((LySetTableInterface) table.getInternalView()).showRow(index);
                }
            }
            else if (index == -10)
			{
				// We should store the selection list
				if ((_selections != null)
						&& (_selections.getSize() != 0))
					_keepSelection = true;
			}

            if ((_selections == null)
            	|| (_selections.getSize() == 0))
			{
				updateContextClient();
			}
			else
			{
				if (_keepSelection
					&& (object != null)
					&& _selections.contains(object))
				{
					_selections.removeObject(object);
					selectObject(object, true);
				}
			}

            updateTools();
            updateMessages();
            updateFooters();
        }

        // Update the last modification time.
        setLastModified(new Date().getTime());
        setRefreshTime(new Date().getTime());

        return true;
    } // end method objectAdded

    /**
     * Called when an object is removed from the dynamic object list.
     * @param event the workspace event
     * @return true if the event was handled
     * @see LyWorkSpaceListener#objectRemoved
     */
    public boolean objectRemoved(LyWorkSpaceEvent event)
    {
        if (isIdle())
            return false;

        if (_component != null)
        {
            int      index = event.getIndex();
            LyWorkSpace ws = _workSpace;
            if (ws == null)
            	return true;

            if (index >= 0)
            {
	            synchronized (ws)
	            {
	            	LyObject object = event.getObject();

                    removeRow(object, index);

                    if (!_keepSelection
                        && (_selections != null)
                        && _selections.contains(object))
                    {
                        _selections.removeObject(object);
                        if (_selections.getSize() == 0)
                            _selections = null;

                        updateContextClient();
                    }
                }
            }
            else if (index == -10)
            {
				// Remove unknown objects from selections
                if (_selections != null)
                {
                    int     	i, ns = _selections.getSize();
                    LyObject	object;

                    for (i = ns - 1; i >= 0; i--)
                    {
                        object = _selections.getObject(i);
                        if (!ws.contains(object))
                            _selections.removeObject(object);
                    }
                }

                _keepSelection = false;
            }

            updateTools();
            updateMessages();
            updateFooters();
        }

        // Update the last modification time.
        setLastModified(new Date().getTime());
        setRefreshTime(new Date().getTime());

        return true;
    } // end method objectRemoved

    /**
     * Called when an object has changed in the dynamic object list.
     *
     * @param event the workspace event
     * @return true if the event was handled
     * @see LyWorkSpaceListener#objectChanged
     */
    public boolean objectChanged(LyWorkSpaceEvent event)
    {

        if (isIdle())
            return false;

        if (_component == null)
        {
            // Update the last modification time.
            setLastModified(new Date().getTime());
            return true;
        }

        LyPanelBehaviorInterface behavior = (LyPanelBehaviorInterface) getBehavior();
        LyObject                 object = event.getObject();
        LyValueSet               values = event.getValues();

        if (!behavior.needRefresh(this, object, values))
            return false;

        int index = event.getIndex();
        int oldIndex = event.getOldIndex();

        if (index < 0)
        {
            // Just update messages
            updateMessages();
            setLastModified(new Date().getTime());
            return true;
        }

        if ((_sort != null) && _sort.hasGroups())
        {

            if (index != oldIndex)
            {
                reset(false);
                return true;
            }

            LyInfoList<LyFieldInfo> groupFieldInfos = _sort.getGroupByFieldInfos();
            int                     i, n = groupFieldInfos.getSize();
            LyFieldInfo             fieldInfo;

            for (i = 0; i < n; i++)
            {
                fieldInfo = groupFieldInfos.getInfo(i);

                if (values.contains(fieldInfo))
                {
                    reset(false);
                    return true;
                }
            }
        }

        if (values == null)
        {

            synchronized (_workSpace)
            {
                // Table is resorted
                swapRows(index, oldIndex);
            }
        }
        else
        {
            // If a fieldInfo with mark ID has changed, we must reset the table.

            // Check if an update is necessary or not
            int         i, n = values.getSize();
            LyValue     value;
            LyFieldInfo fieldInfo;
            LyRow       row = null;

            if (values.hasMark(LyFieldInfo.MARK_ID) && ((row = getRow(object, index)) == null))
            {
                LyObjectId oldObjectId = object.getObjectId(values);

                if (oldObjectId != null)
                    row = getRow(oldObjectId.getFullString(), index);

                if (row != null)
                {
                    row.setId(object.getId());
                    oldIndex = -1;
                }
                else
                {
                    reset(false);
                    return true;
                }
            }

            if ((index == oldIndex) && (values.getSize() != 0))
            {

                for (i = 0; i < n; i++)
                {
                    value = values.getFieldValue(i);
                    fieldInfo = (LyFieldInfo) value.getField();

                    if (!_classInfo.contains(fieldInfo) && _classInfo.isVirtual())
                    {
                        fieldInfo = _classInfo.getVirtualFieldInfo(fieldInfo);

                        if (fieldInfo == null)
                            continue;
                    }

                    if (needsUpdateForField(fieldInfo))
                        break;
                }

                if (i == n)
                {
                    // Table update is useless but may be required for toolbar
                    updateTools();
                    return false;
                }
            }

            boolean selected = false;

            if (_workSpace != null)
            {
                synchronized (_workSpace)
                {

                    // Update concerned object
                    if ((_selections != null) && _selections.contains(object))
                        selected = true;

                    if ((index != oldIndex)
                    	|| _displayList)
                    {
                        row = removeRow(object, oldIndex);

                        if (row != null)
                            insertRow(object, index, selected);
                    }
                    else
                    {
                        LyInfoList<LyFieldInfo> statusFields = getStatusFieldInfos();

                        row = updateRow(object, oldIndex, statusFields);
                    }
                }
            }
            if (row != null)
            {

                if (selected)
                    updateTools();

                updateMessages();
                updateFooters();

                setRefreshTime(new Date().getTime());
            }
            else
            {
                reset(false);
                return true;
            }
        }

        // Update the last modification time.
        setLastModified(new Date().getTime());
        return true;
    } // end method objectChanged

    /**
     * Called when an object state is changed.
     *
     * @param  event Event notifying object state change.
     * @return True if the event is used by the listener, false otherwise.
     */
    public boolean objectStateChanged(LyAdvancedWorkSpaceEvent event)
    {

        if (isIdle())
            return false;

        LyObject              object = event.getObject();
        LyWorkSpaceEvent.Type type = event.getType();

        if (type == LyWorkSpaceEvent.Type.SELECTION_CHANGE)
        {

            if (object.isSelected())
                selectObject(object, true, true, false);
            else
                selectObject(object, false, true, false);

            return true;
        }
        else if (type == LyWorkSpaceEvent.Type.VISIBILITY_CHANGE)
        {

            if (object.isHidden())
                selectObject(object, false, true, false);

            hideTableRow(object, object.isHidden());
            return true;
        }
        else if (type == LyWorkSpaceEvent.Type.ERROR_CHANGE)
        {

            if (getEnvironment().testResource("LY_ERROR_STATE_ICON"))
            {
                LyInfoList<LyFieldInfo> statusFields = getStatusFieldInfos();

                updateRow(object, event.getIndex(), statusFields);
            }
        }

        return false;
    } // end method objectStateChanged

    /**
     * Indicates whether a value change on the given field info should perform
     * an update on the table.
     *
     * @param fieldInfo the field info concerned by the value change
     * @return true if an update must be performed, false otherwise
     */
    protected boolean needsUpdateForField(LyFieldInfo fieldInfo)
    {
        LySort sort = getSort();

        if (((_fields != null) && _fields.contains(fieldInfo))
                || ((_dependsFields != null) && _dependsFields.contains(fieldInfo))
                || ((_fieldMarks != null) && fieldInfo.hasMarks(_fieldMarks, false))
                || ((_statusFieldMarks != null) && fieldInfo.hasMarks(_statusFieldMarks, true))
                || ((sort != null) && sort.contains(fieldInfo)))
            return true;

        return false;
    }

    /**
     * Update the headers of the table (content of the column titles).
     */
    protected void updateHeaders()
    {

        if (_component == null)
            return;

        // If table is Idle, no need to refresh
        if (isIdle())
            return;

        // SMA : no real headers to update in display list
        if (_displayList)
            return;

        LyInfoList<LyFieldInfo> fields = getRealFields();
        int                     i, n = ((fields == null) ? 0 : fields.getSize());
        LyFieldInfo             fieldInfo;
        LySort                  sort = getSort();
        LyInfoList<LyFieldInfo> groupByFields = ((sort == null) ? null
                                                                : sort.getGroupByFieldInfos());
        int                     index, visibleIndex;
        LySortCriterion         criterion;
        LyTable                 table = getTableComponent();
        LyHeaders               headers = table.getHeaders();
        LyHeader                header;

        if (headers == null)
            return;

        if (isStatusColumn(null))
        {
            header = headers.getHeader(0);

            if ((_currentStatusSort != 0) && (header.isSortable()))
            {
                // we change sort by status
                header.setSort(0);
                header.setIncreasing((_currentStatusSort > 0));
            }
            else
                header.setSort(-1);
        }

        for (i = 0; i < n; i++)
        {
            fieldInfo = fields.getInfo(i);
            index = ((sort == null) ? -1 : sort.getIndex(fieldInfo));
            visibleIndex = ((index < 0)
                    ? index : (index - ((groupByFields == null) ? 0 : groupByFields.getSize())));

            header = (LyHeader) headers.getComponent("header" + ID_SEP + fieldInfo.getId());

            if (header == null)
                continue;

            if (header.getSort() != visibleIndex)
                header.setSort(visibleIndex);

            if (index >= 0)
            {
                criterion = sort.getCriterion(index);

                if (header.isIncreasing() != criterion.isIncreasing())
                    header.setIncreasing(criterion.isIncreasing());
            }
        }

        // Update the last modification time.
        setLastModified(new Date().getTime());
    } // end method updateHeaders

    /**
     * Update the footers of the table (content at the bottom of the columns).
     */
    protected void updateFooters()
    {
    }

    /**
     * Internal method to add a new row for the specified object, in the specified line.
     *
     * @param object   current object being inserted in the table
     * @param index    current index it the view, or -1 to add at the end
     * @param selected indicates wether this line is selected or not
     */
    protected void insertRow(LyObject object, int index, boolean selected)
    {

        if (_component == null)
            return;

        if ((_sort != null) && _sort.hasGroups())
        {
            reset(false);
            return;
        }

        LyTable table = getTableComponent();
        LyRows  rows = table.getRows();

        if (rows == null)
        {
            rows = new LyRows();
            table.setRows(rows);
        }

        synchronized (_workSpace)
        {
            LyInfoList<LyFieldInfo> fields = getFields();
            LyInfoList<LyFieldInfo> statusFields = getStatusFieldInfos();
            LyRow                   row = buildRow(object, fields, statusFields);

            if (selected)
                row.setSelected(true);

            if (index > rows.getRowCount())
                index = rows.getRowCount();

            rows.insertRow(row, index);
        }
    } // end method insertRow

    /**
     * Get the row corresponding to the object or the index if object is null
     * @param object an object
     * @param index an index
     * @return a row
     */
    protected LyRow getRow(LyObject object, int index)
    {
        return getRow(((object == null) ? null : object.getId()), index);
    }

    /**
     * Get the row corresponding to the objectId and the index
     * @param objectId an object
     * @param index an index
     * @return a row
     */
    protected LyRow getRow(String objectId, int index)
    {

        if (_component == null)
            return null;

        LyTable table = getTableComponent();
        LyRows  rows = table.getRows();

        if (rows == null)
            return null;

        LyRow  row = rows.getRow(index);
        String rowId = ((row == null) ? null : row.getId());

        if ((row != null) && ((objectId == null) || ((rowId != null) && rowId.equals(objectId)))) // Be sure to remove the right row
            return row;
        else if (objectId != null)
            return findRow(rows, objectId);
        else
            return null;
    } // end method getRow

    /**
     * Internal method to remove a row at the specified line or for the specified object.
     *
     * @param object the object to remove
     * @param  index current index of the line to remove
     * @return the row which was removed, or null if it could nott be found
     */
    protected LyRow removeRow(LyObject object, int index)
    {
        LyRow row = getRow(object, index);

        if (row != null)
            removeRow(row);
        else
        {
            reset(false);
            return null;
        }

        LyRows rows = (LyRows) row.getParent();

        if ((rows != null) && (rows.getParent() instanceof LyRow))
        {
            // Update footer row if needed
            int    nr = rows.getRowCount();
            LyRow  lastRow = ((nr > 1) ? rows.getRow(nr - 1) : null);
            String lastRowId = ((lastRow == null) ? null : lastRow.getId());

            if ((lastRowId != null)
                    && (lastRowId.startsWith(SECTION_ROW_PREFIX)
                        || lastRowId.startsWith(FOOTER_ROW_PREFIX)))
            {
                String                  firstId = rows.getRow(0).getId();
                String                  lastId = rows.getRow(nr - 2).getId(); // last minus 1 for footer row
                LyClassInfo             classInfo = getClassInfo();
                LyObjectId              firstObjectId = new LyObjectId(classInfo, firstId);
                LyObjectId              lastObjectId = new LyObjectId(classInfo, lastId);
                int                     firstIdx = _workSpace.getIndex(firstObjectId
                        .getLogicalObject());
                int                     lastIdx = _workSpace.getIndex(lastObjectId
                        .getLogicalObject());
                LyInfoList<LyFieldInfo> statusFields = getStatusFieldInfos();
                boolean                 status = ((statusFields != null)
                        && (statusFields.getSize() > 0));

                rows.removeRow(nr - 1);
                row = buildFooterRow(firstIdx, lastIdx + 1, status);

                if (row != null)
                    rows.addRow(row);
            }
        }

        return row;
    }

    /**
     * Update the row corresponding to the object.
     *
     * @param object       the object that was modified
     * @param index        the position of the object
     * @param statusFields the list of status fields
     * @return the updated row
     */
    protected LyRow updateRow(LyObject object, int index, LyInfoList<LyFieldInfo> statusFields)
    {
        LyRow row = getRow(object, index);

        if (row == null) // the row has disappeared
        {
            LyEnvironment environment = getEnvironment();

            environment.logWarning("UpdateRow, no row was found for " + object + " at position "
                + index);
            return row;
        }

        LyInfoList<LyFieldInfo>  fields = getFields();
        LyFieldInfo              fieldInfo;
        LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();
        LyComponent              comp;
        int                      i, pos, n = fields.getSize();
        boolean                  status = isStatusColumn(statusFields);

        for (i = 0; i < n; i++)
        {
            pos = i;

            if (status)
                pos++;

            fieldInfo = fields.getInfo(i);
            comp = row.getChild(pos);

            // in display list mode, get comp by id.
            if (isDisplayList())
                comp = row.getComponent(object.getId() + DISPLAY_LIST_PREFIX + fieldInfo.getId());

            if (comp instanceof LyCell)
                behavior.updateCell(this, (LyCell) comp, object, fieldInfo);
            else
                behavior.updateRowItem(this, comp, object, fieldInfo);
        }

        updateRowStatusAndColor(row, object, statusFields);

        row.setLabel(behavior.getLabel(this, object));
        row.setIndex(behavior.getIndex(this, object));
        return row;
    } // end method updateRow

    /**
     * Updates the status image and tooltip and the color for a given object.
     *
     * @param row          the row of the object.
     * @param object       the object that was modified
     * @param statusFields the list of status fields
     */
    protected void updateRowStatusAndColor(LyRow row, LyObject object,
        LyInfoList<LyFieldInfo> statusFields)
    {
        LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();
        LyFieldInfo              fieldInfo;

        // Set the row image if needed
        String statusImage = behavior.getStatusImage(this, statusFields, object);
        String oldToolTip = row.getTooltip();
        String statusTooltip = null;
        String image = row.getImage();

        if (statusImage != null)
        {
            LyEnvironment environment = getEnvironment();
            boolean       found = environment.checkImage(statusImage);

            if (!found
            	&& environment.checkImage(statusImage.toUpperCase()))
            {
                found = true;
                statusImage = statusImage.toUpperCase();
            }

            // Check if status image exists
            if (!found
            	&& (statusFields.getSize() == 1))
            {
                fieldInfo = statusFields.getInfo(0);

                if (!fieldInfo.isMultiple()
                	&& (fieldInfo instanceof LyChoiceFieldInfo))
                {
                    Object value = ((object == null) ? null : object.getValue(fieldInfo));

                    if (value == null)
                        statusImage = "LY_STATUS_10";
                    else if (value instanceof Short)
                        statusImage = "LY_STATUS_" + ((Short) value).shortValue();
                    else if (value instanceof Boolean)
                        statusImage = (((Boolean) value).booleanValue() ? "LY_STATUS_0"
                                                                        : "LY_STATUS_3");
                }
            }
        }

        // We may handle error icon in addition to the object icon
        if ((object != null) && (statusImage != null))
            statusImage = getViewManager().buildErrorImage(statusImage, object.getErrorLevel());

        if (((image == null) && (statusImage != null))
                || ((image != null) && !image.equals(statusImage)))
            row.setImage(statusImage);

        image = ((statusImage == null) ? null : statusImage);

        if (image != null)
            statusTooltip = behavior.getStatusTooltip(this, statusFields, object);
        else if (object != null)
            statusTooltip = "$$" + getName(object);

        // If the tooltip has been modified
        if (((oldToolTip == null) && (statusTooltip != null))
                || ((oldToolTip != null) && !oldToolTip.equals(statusTooltip)))
            row.setTooltip(statusTooltip);

        // Set the row image if needed
        String rowColor = behavior.getRowColor(this, object);
        String oldColor = row.getColor();

        if (((rowColor != null) && (oldColor == null))
                || ((oldColor != null) && !oldColor.equals(rowColor)))
            row.setColor(rowColor);

        if (object != null)
        {
            row.setHidden(object.isHidden());

            LyClassInfo classInfo = object.getClassInfo();
            LyInfoLabel label = getInfoLabel(classInfo, LyInfoLabel.Type.SEMANTIC);

            if (label != null)
                row.setOverview(label.getString(object));
        }
    } // end method updateRowStatusAndColor

    /**
     * Changes the button at the end of the given row.
     *
     * @param row     the row to be updated.
     * @param object  the object represented by given row.
     * @param image   the new image to be used for given row.
     * @param label   the new label to be used for given row.
     * @param tooltip the new tooltip to be used for given row.
     */
    protected void setRowTool(LyRow row, LyObject object, String image, String label,
        String tooltip)
    {

        if (row == null)
            return;

        // Add a tool at the end of the row
        LyTable     table = getTableComponent();
        LyHeaders   headers = table.getHeaders();
        int         n = headers.getHeaderCount();
        LyToolBar   toolbar = null;
        LyTool      tool = null;
        LyComponent lastComponent = row.getComponent(n - 1);
        boolean     addEmptyCell = false;
        boolean     isLastComponentToolbar = (lastComponent instanceof LyToolBar);
        boolean     isLastComponentEmptyCell = ((lastComponent instanceof LyCell)
                && (row.getId() + EMPTY_CELL).equals(((LyCell) lastComponent).getObjectId()));

        // In web mode, force the construction of toolbar to avoid refresh issues.
        LyViewManager viewManager = getViewManager();
        boolean       forceToolbar = viewManager.hasProperty("WEB_MANAGER");

        if (isLastComponentToolbar)
        {
            toolbar = (LyToolBar) lastComponent;
            tool = (LyTool) toolbar.getComponent(0);
        }
        else if ((image != null) || forceToolbar)
        {

            if (isLastComponentEmptyCell)
            {
                LyCell emptyCell = (LyCell) lastComponent;

                row.removeCell(emptyCell);
                emptyCell.free();
            }

            // Change the toolbar environment to display force tool label to be displayed
            // on the right.
            if (_rowToolEnvironment == null)
            {
                _rowToolEnvironment = new LyEnvironment(getEnvironment());
                _rowToolEnvironment.putResource("LY_WRITE_TOOL_LABEL", "true");
                _rowToolEnvironment.putResource("LY_TOOL_LABEL_RIGHT_POSITION", "true");
            }

            toolbar = new LyToolBar();
            toolbar.setController(getId());
            tool = new LyTool(DEMAND_ROW_TOOL
                    + ((object == null) ? "" : object.getObjectId().getFullString()), "");
            toolbar.setEnvironment(_rowToolEnvironment);
            toolbar.addTool(tool);

            int componentCount = row.getComponentCount();

            if (_displayList)
                componentCount = getColsNumberListDisplay();

            row.insertToolBar(toolbar, componentCount);
        }
        else if (!isLastComponentEmptyCell)
            addEmptyCell = true;

        if (image != null)
        {
            tool.setImage(image);
            tool.setLabel(label);
            tool.setTooltip(tooltip);
            tool.setEnabled(true);
        }
        else if (forceToolbar)
        {
            tool.setImage(null);
            tool.setLabel(label);
            tool.setTooltip(tooltip);
            tool.setEnabled(false);
        }
        else if (isLastComponentToolbar)
        {
            row.removeToolBar(toolbar);
            toolbar.free();
            addEmptyCell = true;
        }

        if (addEmptyCell)
        {
            LyCell emptyCell = new LyCell();

            emptyCell.setObjectId(row.getId() + EMPTY_CELL);
            row.insertCell(emptyCell, row.getComponentCount());
        }
    } // end method setRowTool

    /**
     * Internal method to swap two rows at the specified lines.
     *
     * @param index    new index of the first line to move
     * @param oldIndex current index of the line to move
     */
    protected void swapRows(int index, int oldIndex)
    {

        if (_component == null)
            return;

        LyTable table = getTableComponent();
        LyRows  rows = table.getRows();

        if (rows == null)
            return;

        synchronized (_workSpace)
        {
            rows.swapRows(oldIndex, index);
        }
    }

    /**
     * Accessor to the object from it's line number in the view.
     *
     * @param  viewLineNumber current index in the view. (line number)
     * @return current object to retrieve.
     */
    protected LyObject getObject(int viewLineNumber)
    {

        if ((_workSpace == null) || (viewLineNumber < 0)
                || (viewLineNumber > (_workSpace.getSize() - 1)))
            return null;

        if ((_maxRowCount > 0) && (viewLineNumber >= _maxRowCount))
            return null;

        int workSpaceIndex = ((_workSpace instanceof LyPageWorkSpace)
                ? viewLineNumber : ((_currentPage * _maxRowCount) + viewLineNumber));

        if (workSpaceIndex > (_workSpace.getSize() - 1))
            return null;

        // Get the right index
        if ((_sort != null) && _sort.hasGroups())
        {
            LyTable table = getTableComponent();
            LyRows  rows = table.getRows();

            if (rows == null)
                return null;

            LyRow      row = getRow(viewLineNumber);
            String     rowId = ((row == null) ? null : row.getId());
            LyObjectId objectId = getObjectId(rowId);

            return _workSpace.getObject(objectId);
        }
        else
            return _workSpace.getObject(workSpaceIndex);
    }

    /**
     * Current builder of a new row.
     *
     * @param  object       object added in the contextual list, and which should set the content of
     *                      row cells.
     * @param  fields       list of fields which should be seen.
     * @param  statusFields list of fields which may set the status icon.
     * @return LyRow the new Row description (XML description) which is added in the LyTable
     *         component.
     * @see    #insertRow
     */
    public LyRow buildRow(LyObject object, LyInfoList<LyFieldInfo> fields,
        LyInfoList<LyFieldInfo> statusFields)
    {
        if (fields == null)
            return null;

        LyRow row = new LyRow();

        // SMA : if display list, build row in list
        if (_displayList)
            return buildRowDisplayList(object, fields, statusFields);

        LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();

        if (object != null)
        {
            row.setId(object.getId());
            row.setLabel(behavior.getLabel(this, object));
            row.setIndex(behavior.getIndex(this, object));
        }

        if (isStatusColumn(statusFields))
        {
            LyCell cell = new LyCell();

            cell.setLabel("");
            row.addCell(cell);
        }

        int         i, n = fields.getSize();
        LyFieldInfo fieldInfo;
        LyComponent comp;

        for (i = 0; i < n; i++)
        {
            fieldInfo = fields.getInfo(i);
            comp = buildRowItem(object, fieldInfo);
            addRowItem(row, comp, object, fieldInfo);
        }

        updateRowStatusAndColor(row, object, statusFields);

        behavior.completeRow(this, row, object);
        return row;
    } // end method buildRow

    /**
     * Says if table is in display list or not.
     * @return true if display list
     */
    public boolean isDisplayList()
    {
        return _displayList;
    }

    /**
     * Says if we have to show fields in display list.
     * @return true if fields must be show
     */
    public boolean isDisplayListShowFields()
    {
        return _displayListShowFieldLabels;
    }


    /**
     * List row representation.[status if exist | image if exist | data in one cell | footers]
     * @param object object added in the contextual list, and which should set the content of row cells.
     * @param fields list of fields which should be seen.
     * @param statusFields list of fields which may set the status icon.
     * @return LyRow the new Row description (XML description) which is added in the LyTable component.
     * @see #insertRow
     */
    protected LyRow buildRowDisplayList(LyObject object, LyInfoList<LyFieldInfo> fields,
        LyInfoList<LyFieldInfo> statusFields)
    {

        // if no fields or no object do nothing
        if ((fields == null) || (object == null))
            return null;

        LyRow       row = new LyRow();
        LyFieldInfo imageField = null;
        // cell for field name
        LyCell fieldNamesCell = new LyCell();
        // cell for datas
        LyCell      dataCell = new LyCell();
        int         i, n = fields.getSize();
        LyFieldInfo fieldInfo;
        LyClassInfo classInfo = getClassInfo();
        LyComponent comp;
        // list for complex fields
        LyInfoList<LyFieldInfo> complexFields = getComplexFields();
        // list for footers fields
        LyInfoList<LyFieldInfo>  footerFields = new LyInfoList<LyFieldInfo>();
        LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();

        if (object != null)
        {
            row.setId(object.getId());
            row.setLabel(behavior.getLabel(this, object));
            row.setIndex(behavior.getIndex(this, object));
        }

        // check if we have an image parameter, if so, put image in first column
        // if not, display first image found.
        fieldInfo = (LyFieldInfo) this.getParameter("_displayListImgField", true);

        // be sure to have an image
        if ((fieldInfo instanceof LyFileFieldInfo)
                && LyFileFieldInfo.Type.IMAGE.equals(((LyFileFieldInfo) fieldInfo).getType()))
            imageField = fieldInfo;
        // if we have a behavior which gives the image field
        fieldInfo = behavior.getDisplayListFirstImage(this);

        if (fieldInfo != null)
            imageField = fieldInfo;

        // add object name in first data cell
        LyCell nameCell = new LyCell();
        nameCell.setLabel(object.getName());
        nameCell.setFont("LY_DEFAULT_BOLD_FONT");

        if (_displayListShowFieldLabels)
        {
        	fieldNamesCell.addCell(nameCell);
        	dataCell.addCell(new LyCell());
        }
        else
        {
        	dataCell.addCell(nameCell);
        }

        for (i = 0; i < n; i++)
        {
            fieldInfo = fields.getInfo(i);

            // skip non class members fields
            if (!classInfo.contains(fieldInfo))
                continue;

            // skip name marked fields
    		if (fieldInfo.hasMark(LyFieldInfo.MARK_NAME))
    			continue;

            // look for image
            if ((fieldInfo instanceof LyFileFieldInfo)
                    && LyFileFieldInfo.Type.IMAGE.equals(((LyFileFieldInfo) fieldInfo).getType())
                    && (imageField == null))
                imageField = fieldInfo;
            else if ((fieldInfo instanceof LyNumberFieldInfo)
                    && (((LyNumberFieldInfo) fieldInfo).getFunction()
                        != LyNumberFieldInfo.Function.NONE))
                // look for footer's fields
                footerFields.addInfo(fieldInfo);
            else
            {

                // skip image field displayed first
                if ((imageField != null) && fieldInfo.getId().equals(imageField.getId()))
                    continue;


                // get field's value
                comp = buildRowItem(object, fieldInfo);

                if (comp instanceof LyCell)
                {
                    LyCell c = (LyCell) comp;
                    c.setFont("LY_DEFAULT_FONT");
                    if (object != null)
                        c.setId(object.getId() + DISPLAY_LIST_PREFIX + fieldInfo.getId());
                    if ((c.getLabel() == null)
                            && (c.getImage() == null))
                        continue;

                    // look for other fields
                    // get field's name
                    if (_displayListShowFieldLabels)
                    {
                        LyCell hcell = new LyCell();
                        hcell.setLabel(translate(fieldInfo.getName()));
                        hcell.setFont("LY_DEFAULT_FONT");
                        fieldNamesCell.addCell(hcell);
                    }

                    dataCell.addCell(c);
                }
            }
        }

        if (isStatusColumn(statusFields))
        {
            LyCell cell = new LyCell();
            cell.setLabel("");
            row.addCell(cell);
        }

        // cell 2 => add first image field found
        if ((imageField != null) && (object.getValue(imageField) != null))
        {

            comp = buildRowItem(object, imageField);

            if (comp instanceof LyCell)
                ((LyCell) comp).setId(object.getId() + DISPLAY_LIST_PREFIX + imageField.getId());

            addRowItem(row, comp, object, imageField);
        }
        else
            addRowItem(row, getClassInfoImgField(classInfo, object), object, imageField);

        // cell 3 => add headers label
        if (_displayListShowFieldLabels)
            addRowItem(row, fieldNamesCell, object, null);

        // cell 4 => add datas
        addRowItem(row, dataCell, object, null);

        // footers fields
        for (int j = 0; j < footerFields.getSize(); j++)
        {
            fieldInfo = footerFields.getInfo(j);
            comp = buildRowItem(object, fieldInfo);

            if (comp instanceof LyCell)
                ((LyCell) comp).setId(object.getId() + DISPLAY_LIST_PREFIX + fieldInfo.getId());

            if (this instanceof LyEditTableController)
            {
                boolean editable = ((LyEditTableController) this).isEditable(object, fieldInfo);

                if (comp instanceof LyCell)
                    ((LyCell) comp).setEditable(editable);
            }

            addRowItem(row, comp, object, footerFields.getInfo(j));
        }

        // complex fields
        for (int k = 0; k < complexFields.getSize(); k++)
        {
            fieldInfo = complexFields.getInfo(k);
            comp = buildRowItem(object, fieldInfo);

            if (comp instanceof LyCell)
                ((LyCell) comp).setId(object.getId() + DISPLAY_LIST_PREFIX + fieldInfo.getId());

            addRowItem(row, comp, object, fieldInfo);
        }

        // status
        updateRowStatusAndColor(row, object, statusFields);

        behavior.completeRow(this, row, object);

        return row;
    }

    /**
     * Get class image or default if not exist.
     * @param classInfo classInfo
     * @param object object
     * @return cell with image
     */
    protected LyComponent getClassInfoImgField(LyClassInfo classInfo, LyObject object)
    {
        LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();
    	String					 rowImage = behavior.getRowImage(this, object);
        String      			 heightValue = getEnvironment().getResource("LY_CELL_IMAGE_HEIGHT");
        int         			 height = LyTools.intFromString(heightValue);

    	if (rowImage == null)
    	{
	        LyFieldInfo imgField = null, fInfo;
	        int			i, n = classInfo.getFieldInfoCount();

	        for (i = 0; (i < n) && (imgField == null); i++)
	        {
	        	fInfo = classInfo.getFieldInfo(i);
	            if ((fInfo instanceof LyFileFieldInfo)
	            		&& (((LyFileFieldInfo)fInfo).getType() == LyFileFieldInfo.Type.IMAGE))
	                imgField = fInfo;
	        }

	        if ((imgField == null)
	        		|| (object.getValue(imgField) == null))
	            rowImage = this.getImage(classInfo, height, true);
	        else
	            return buildRowItem(object, imgField);
    	}
    	else
    		rowImage = getImageFromPrefix(rowImage, null, height);

        LyCell      cell = new LyCell();

        cell.setImage(rowImage);
        return cell;
    }

    /**
     * Get complex fields.
     * @return list of complex fields.
     */
    public LyInfoList<LyFieldInfo> getComplexFields()
    {
        LyInfoList<LyFieldInfo> fields = getFields();
        LyClassInfo             classInfo = getClassInfo();
        LyInfoList<LyFieldInfo> complexFields = new LyInfoList<LyFieldInfo>();

        for (int i = 0; i < fields.getSize(); i++)
        {
            LyFieldInfo fieldInfo = fields.getInfo(i);

            if (!classInfo.contains(fieldInfo))
                complexFields.addInfo(fieldInfo);
        }

        return complexFields;
    }

    /**
     * Current builder of a new cell in a row. At this level, the implementation creates a LyCell
     * instance, but generic behavior may be overriden.
     *
     * @param  object object added in the contextual list, and which should set the content of row
     *                cells.
     * @param  fieldInfo  field value shown in the cell content
     * @return LyCell or LyFieldIntem the new Cell description (XML description) or editable content
     *         to add in the LyRow component.
     */
    protected LyComponent buildRowItem(LyObject object, LyFieldInfo fieldInfo)
    {
        if (fieldInfo == null)
            return null;

        LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();
        LyComponent              component = behavior.buildRowItem(this, object, fieldInfo);

        if (component != null)
            return component;

        Object      value;
        LyObjectId  objectId;
        LyClassInfo relationClassInfo;
        LyCell      cell = new LyCell();

        behavior.updateCell(this, cell, object, fieldInfo);

        String label = cell.getLabel();

        if (label != null)
        {
            cell.setLabel(label);

            // If relation, set the id of the cell with the object id
            if ((fieldInfo instanceof LyRelationFieldInfo) && getClassInfo().contains(fieldInfo)) // Complex table case
            {
                value = ((object == null) ? null : object.getValue(fieldInfo));

                if (value instanceof LyObjectId)
                {
                    objectId = (LyObjectId) value;

                    // Check if consultation is authorized on the relation class info
                    relationClassInfo = objectId.getClassInfo();

                    if (relationClassInfo.getActionWithId(LyAction.ID_CONSULT) != null)
                    {
                        LyViewManager viewManager = getViewManager();

                        if (viewManager.hasProperty("MANAGE_URL"))
                            cell.setObjectId(objectId.getFullString());
                    }
                }
                //else if (value instanceof Array)
                // XXX TO DO: multiple relation
            }
            // If FileFieldInfo :  Moved into LyTableBehavior for attachment handling (modified)
        }

        return cell;
    } // end method buildRowItem

    /**
     * Adds a component at the end of a row.
     *
     * @param row       the row instance where the item must be added.
     * @param comp      the component to be added.
     * @param object    the object corresponding to the given row.
     * @param fieldInfo the fieldInfo corresponding to the added component.
     */
    protected void addRowItem(LyRow row, LyComponent comp, LyObject object, LyFieldInfo fieldInfo)
    {

        if (comp instanceof LyCell)
            row.addCell((LyCell) comp);
        else if (comp instanceof LyFormField)
            row.addFormField((LyFormField) comp);
        else if (comp instanceof LyToolBar)
            row.addToolBar((LyToolBar) comp);
        else
            throw (new IllegalArgumentException(
                    "BF AppEngine : Component Class not supported in LyRow (" + comp
                    + "). Expected : LyCell | LyFormField."));
    }

    /**
     * Builds a new row corresponding to a grouping value.
     *
     * @param  group    Group label.
     * @param  rowColor Background color of the row.
     * @param  level    Group level.
     * @return LyRow the new row description which is added in the table component.
     */
    protected LyRow buildGroupRow(String group, String rowColor, int level)
    {
        LyRow   row = new LyRow();
        String  rowId = SECTION_ROW_PREFIX + '_' + group;
        boolean isOpen = ((_closedRows == null) || !_closedRows.contains(rowId));

        row.setId(rowId);
        row.setLabel(group);
        row.setOpen(isOpen);

        if (rowColor == null)
            rowColor = "LY_GROUP_BY_COLOR" + level;

        row.setColor(rowColor);
        row.setFontColor("LY_GROUP_BY_FONT_COLOR" + level);
        row.setFont("LY_GROUP_BY_FONT" + level);
        row.setSensitive(false);

        LyRows rows = new LyRows();

        row.setRows(rows);

        return row;
    }

    /**
     * Builds a row corresponding to footers when objects are grouped.
     * @param from start index in the object list
     * @param to end index in the object list
     * @param status indicates if there is a status column
     * @return the footer row
     */
    protected LyRow buildFooterRow(int from, int to, boolean status)
    {

        if ((to - from) <= 1)
            return null;

        if (isDisplayList())
        {
            LySimpleObjectList objectList = new LySimpleObjectList();
            int                i;

            objectList.setSorted(false);

            for (i = from; i < to; i++)
                objectList.addObject(_workSpace.getObject(i));

            LyRow                    row = new LyRow();
            LyInfoList<LyFieldInfo>  fields = getFields();
            LyCell                   cell;
            String                   label = null;
            LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();
            LyFieldInfo              fieldInfo;

            int footerStart = 2;

            if (status)
                footerStart++;

            if (isDisplayListShowFields())
                footerStart++;

            // add Empty footers
            for (i = 0; i < footerStart; i++)
            {
                cell = new LyCell();
                cell.setLabel("");
                row.addCell(cell);
            }

            // add real footers field
            for (int j = 0; j < fields.getSize(); j++)
            {

                if ((fields.getInfo(j) instanceof LyNumberFieldInfo)
                        && (((LyNumberFieldInfo) fields.getInfo(j)).getFunction()
                            != LyNumberFieldInfo.Function.NONE))
                {
                    fieldInfo = fields.getInfo(j);
                    label = behavior.getFooterLabel(this, fieldInfo, objectList);
                    cell = new LyCell();

                    if (label != null)
                        cell.setLabel(label);

                    row.addCell(cell);
                }
            }

            row.setId(FOOTER_ROW_PREFIX + from + "_" + to);
            row.setSensitive(false);
            row.setFont("LY_GROUP_FOOTER_FONT");
            row.setColor("LY_HEADER_COLOR");
            row.setFontColor("LY_TABLE_HEADER_FOREGROUND");

            return row;
        }

        return null;
    }

    /**
     * This method is invoked to enable the the area(s) to complete the message bar of the view and
     * add its own message items.
     * @param messages the messages to update
     * @param level the level
     *
     * @see LyAreaBuilder#completeMessages
     */
    public void completeMessages(LyMessages messages, int level)
    {
        super.completeMessages(messages, level);
    }

    /**
     * This method is invoked to refresh the message bar of the view.
     *
     * @see LyAreaBuilder#completeMessages
     */
    protected void updateMessages()
    {
    }

    /**
     * Reset the content of the table being shown. This method rebuild the XML structure of the
     * table.
     */
    protected void updateContent()
    {

        if (_component == null)
            return;

        if (_workSpace == null)
            return;

        synchronized (_workSpace)
        {
            // Add new rows
            int n = _workSpace.getSize();

            if (n > 0)
            {
                LyInfoList<LyFieldInfo> fields = getFields();
                LyObject                object;

                //GTA
                // Do not declare this vector out of bounds of the next code block because
                // lists must'nt be garbaged during the table filling in.
                Vector<LyObjectList> lists = null;

                if (!(_initialWorkSpace instanceof LyPageWorkSpace))
                {
                    int                     j, m = ((fields == null) ? 0 : fields.getSize());
                    LyFieldInfo             fi, idf;
                    LyInfoList<LyFieldInfo> idFields;
                    LyArrayFieldInfo        array;
                    LyObjectList            list;
                    LyClassInfo             ci, ai;
                    LySession               session = getSession();
                    LyJoinFilter            join;
                    LyFilterRelation        r;
                    LyFilter                filter, arrayFilter;
                    LyFilter                sourceFilter = null;
                    LyArrayDaemon           daemon;

                    if (_classFilter != null)
                    {
                        sourceFilter = new LyFilter();
                        sourceFilter.addElement(_classFilter);
                    }

                    if ((_filter != null) && (_filterActive))
                    {

                        if (sourceFilter == null)
                            sourceFilter = new LyFilter();

                        sourceFilter.addElement(_filter);
                    }

                    for (j = 0; j < m; j++)
                    {
                        fi = fields.getInfo(j);

                        if (fi instanceof LyArrayFieldInfo)
                        {
                            array = (LyArrayFieldInfo) fi;
                            daemon = array.getDaemon();
                            fi = array.getFieldInfo();

                            if (daemon != null)
                            {
                                // Load arrays
                                ai = daemon.getArrayClassInfo();
                                idf = ai.getFieldInfo(0);

                                if (sourceFilter != null)
                                {
                                    filter = new LyFilter();
                                    r = new LyFilterRelation((LyRelationFieldInfo) idf,
                                            sourceFilter);
                                    filter.addElement(r);
                                }
                                else
                                    filter = null;

                                list = new LyClassWorkSpace(session, ai, null, filter, null, false,
                                        false, false);

                                if (lists == null)
                                    lists = new Vector<LyObjectList>();

                                lists.addElement(list);

                                arrayFilter = filter;

                                if (fi instanceof LyStructFieldInfo)
                                {
                                    // Load structs
                                    ci = ((LyStructFieldInfo) fi).getStructClassInfo();
                                    idFields = ci.getIdFieldInfos();
                                    idf = idFields.getInfo(0);
                                    fi = ai.getFieldInfo(1);
                                    filter = new LyFilter();
                                    join = new LyJoinFilter(ai, arrayFilter, fi, idf);
                                    filter.addElement(join);
                                    list = new LyClassWorkSpace(session, ci, null, filter, null,
                                            false, false, false);
                                    lists.addElement(list);
                                }
                            }
                        }
                        else if (fi instanceof LyStructFieldInfo)
                        {
                            ci = ((LyStructFieldInfo) fi).getStructClassInfo();
                            idFields = ci.getIdFieldInfos();

                            if ((idFields != null) && (idFields.getSize() > 0))
                            {
                                idf = idFields.getInfo(0);
                                filter = new LyFilter();
                                join = new LyJoinFilter(getClassInfo(), sourceFilter, fi, idf);
                                filter.addElement(join);
                                list = new LyClassWorkSpace(session, ci, null, filter, null, false,
                                        false, false);

                                if (lists == null)
                                    lists = new Vector<LyObjectList>();

                                lists.addElement(list);
                            }
                        }
                    }
                }
                // END GTA

                // If there is a limit but no page workspace (may happen when a context
                // is set on the table with a limit)
                int startIdx = 0;
                int endIdx = n;
                int limit = getLimit();

                if ((limit > 0) && !(_initialWorkSpace instanceof LyPageWorkSpace))
                {
                    startIdx = _currentPage * _maxRowCount;

                    // Check if page is still in boundaries
                    if (startIdx >= n)
                    {
                        // Take last page
                        _currentPage = n / _maxRowCount;
                        startIdx = _currentPage * _maxRowCount;
                    }

                    endIdx = Math.min(n, (_currentPage + 1) * _maxRowCount);
                }

                // Keep selections synchronized with workSpace
                if (_selections != null)
                {
                    int i, ns = _selections.getSize();

                    for (i = ns - 1; i >= 0; i--)
                    {
                        object = _selections.getObject(i);

                        if (!_workSpace.contains(object))
                            _selections.removeObject(object);
                    }
                }

                buildContent(startIdx, endIdx);
            }
            else
                buildContent(0, 0);

            updateTableNRows();
        }

        // Update the last modification time.
        setLastModified(new Date().getTime());
    } // end method updateContent

    /**
     * Updates the nRows attribute of the table component.
     * This update will be performed only if the
     */
    protected void updateTableNRows()
    {

        if (_maxRowCount > 0)
        {
            String displayEmptyRows = (String) _action.getParameter("_displayEmptyRows", true);

            if ("true".equalsIgnoreCase(displayEmptyRows))
            {
                LyTable table = getTableComponent();

                table.setNRows(_maxRowCount);
            }
        }
    }

    /**
      * Builds content for objects beetween given indexes.
      * @param startIdx Start index.
      * @param endIdx End index.
      */
    protected void buildContent(int startIdx, int endIdx)
    {
        // Remove existing rows
        LyTable table = getTableComponent();
        LyRows  rows = table.getRows();
        int     i, n = ((rows == null) ? 0 : rows.getRowCount());

        for (i = n - 1; i >= 0; i--)
            rows.removeRow(i);

        if ((startIdx < 0) || (endIdx < 0) || (startIdx >= endIdx))
            return;

        LyInfoList<LyFieldInfo>  fields = getFields();
        LyInfoList<LyFieldInfo>  statusFields = getStatusFieldInfos();
        LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();
        LyRow                    row;
        LyObject                 object;
        LyFieldInfoValue         value;
        LyInfoList<LyFieldInfo>  groupFieldInfos = ((_sort != null) ? _sort.getGroupByFieldInfos()
                                                                    : null);
        int                      j, k,
                                 ng = ((groupFieldInfos == null) ? 0 : groupFieldInfos.getSize());
        String[]                 previousGroups = ((ng > 0) ? new String[ng] : null);
        LyRows[]                 groupRows = ((ng > 0) ? new LyRows[ng] : null);
        LyFieldInfo              groupFieldInfo;
        String                   group, color;
        int                      groupStartIdx = -1;
        boolean                  hasFooters = (table.getFooters() != null);
        boolean                  status = isStatusColumn(statusFields);

        if (rows == null)
        {
            rows = new LyRows();
            table.setRows(rows);
        }

        for (i = startIdx; i < endIdx; i++)
        {
            object = _workSpace.getObject(i);

            for (j = 0; j < ng; j++)
            {
                groupFieldInfo = groupFieldInfos.getInfo(j);
                value = (LyFieldInfoValue) object.getFieldValue(groupFieldInfo);
                group = behavior.getGroupLabel(this, object, value);

                if (group == null)
                {
                    group = value.getName();

                    if (group == null)
                        group = "LY_NONE";
                }

                if ((previousGroups[j] == null) || !previousGroups[j].equals(group))
                {

                    if ((groupStartIdx >= 0) && (groupRows[j] != null))
                    {
                        row = buildFooterRow(groupStartIdx, i, status);

                        if (row != null)
                            groupRows[ng - 1].addRow(row);
                    }

                    color = behavior.getGroupColor(this, object, value);
                    row = buildGroupRow(group, color, j);

                    if (j == 0)
                        rows.addRow(row);
                    else
                    {
                        LyRows      parentRows = groupRows[j - 1];
                        LyComponent parentRow = parentRows.getParent();

                        if (parentRow instanceof LyRow)
                            row.setId(((LyRow) parentRow).getId() + row.getId());

                        parentRows.addRow(row);
                    }

                    groupRows[j] = row.getRows();
                    previousGroups[j] = group;

                    if (hasFooters)
                        groupStartIdx = i;

                    for (k = j + 1; k < ng; k++)
                    {
                        previousGroups[k] = null;
                        groupRows[k] = null;
                    }
                }
            }

            row = buildRow(object, fields, statusFields);

            if ((_selections != null) && (_selections.contains(object)))
                row.setSelected(true);
            else if (object.isSelected())
            {
                if (_selections == null)
                {
                    _selections = new LySimpleObjectList();
                    _selections.setSorted(false);
                }

                _selections.addObject(object);
                row.setSelected(true);
            }

            if (groupRows != null)
                groupRows[ng - 1].addRow(row);
            else
                rows.addRow(row);
        }

        if ((groupStartIdx >= 0) && (groupRows != null))
        {
            row = buildFooterRow(groupStartIdx, i, status);

            if (row != null)
                groupRows[ng - 1].addRow(row);
        }
    }

    /**
     * Keep selections synchronized with workSpace.
     */
    protected void updateSelections()
    {

        if ((_selections == null) || (_selections.getSize() == 0))
            return;

        if (_workSpace == null)
            _selections = null;
        else
        {
            int      i, n = _selections.getSize();
            LyObject object;

            for (i = n - 1; i >= 0; i--)
            {
                object = _selections.getObject(i);

                if (!_workSpace.contains(object))
                    _selections.removeObject(object);
            }
        }
    }

    /**
     * Get the default sort for the current action.
     * First try to get it from action parameter _sort
     * then try to get it from the classInfo
     * finally build a sort based on the first column with the sort mark
     * @return the default sort for this table.
     */
    protected LySort getDefaultSort()
    {
        LySort sort = (LySort) getParameter("_sort");

        if (sort != null)
            return sort.cloneSort();

        sort = _classInfo.getSort();

        if (sort != null)
            return sort.cloneSort();

        // Use default sort : first fieldInfo with mark SORT
        LyInfoList<LyFieldInfo> fields = getFields();
        int                     i, n = ((fields == null) ? 0 : fields.getSize());
        LyFieldInfo             fieldInfo;

        for (i = 0; i < n; i++)
        {
            fieldInfo = fields.getInfo(i);

            if (fieldInfo.hasMark(LyFieldInfo.MARK_SORT))
            {
                sort = new LySort();
                sort.addCriterion(fieldInfo, true);
                return sort;
            }
        }

        return null;
    }

    /**
     * Return default criterion for a given field info. If no default sort is defined on the class,
     * return null.
     *
     * @param  fieldInfo the field info we want to get the criterion
     * @return the default sort criterion for the given field info
     */
    protected LySortCriterion getDefaultCriterion(LyFieldInfo fieldInfo)
    {
    	LySort	sort = _classInfo.getSort();

        if (sort != null)
        {
	        int             i, n = sort.getCriterionCount();
	        LySortCriterion criterion;

	        // Retrieve the criterion for the given field info
	        for (i = 0; i < n; i++)
	        {
	            criterion = sort.getCriterion(i);
	            if (criterion.getFieldInfo() == fieldInfo)
	                return new LySortCriterion(fieldInfo, criterion.isIncreasing(), criterion.getJavaName());
	        }
        }

		LyTableBehaviorInterface behavior = (LyTableBehaviorInterface)getBehavior();

    	return behavior.getDefaultCriterion(this, fieldInfo);
    }

    /**
     * Gets the list of fields being displayed in the columns of the table The default
     * implementation uses the specified fieldsMarks or if specified the lists of fields given by
     * the setFields method.
     *
     * @return the list of fields being displayed
     * @see    #setFields(LyInfoList)
     */
    public LyInfoList<LyFieldInfo> getFields()
    {
        if (_fields != null)
            return _fields;

        LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();
        LyFieldInfo              field = (LyFieldInfo) _action.getParameter("_fieldInfo", true);
        int                      np = ((field != null) ? 2 : 1);

        do
        {
            if (field != null)
            {
                if (_fields == null)
                    _fields = new LyInfoList<LyFieldInfo>();

                _fields.addInfo(field);
            }

            field = (LyFieldInfo) _action.getParameter("_fieldInfo" + np, true);
            np++;
        }
        while (field != null);

        if (_fields == null)
        {
            if (_fieldMarks != null)
            {
                int         m, nm = _fieldMarks.size();
                int         i, n = _classInfo.getFieldInfoCount();
                LyFieldInfo fieldInfo;

                for (m = 0; m < nm; m++)
                {

                    if (!_fieldMarks.get(m))
                        continue;

                    for (i = 0; i < n; i++)
                    {
                        fieldInfo = _classInfo.getFieldInfo(i);

                        if (hasMark(fieldInfo, m)
                                || ((m == LyFieldInfo.MARK_TABLE)
                                    && hasMark(fieldInfo, LyFieldInfo.MARK_MAIN)))
                        {

                            if (_fields == null)
                                _fields = new LyInfoList<LyFieldInfo>();

                            _fields.addInfo(fieldInfo);
                        }
                    }
                }
            }

            if (_fields == null)
                _fields = getDefaultFields();
        }

        if (behavior != null)
        {
            _fields = behavior.getFields(this, _fields);
            _fields.lock();
        }

        if (_fields.getSize() == 0)
            _fields = null;

        return _fields;
    } // end method getFields

    /**
     * Gets default fields.
     * @return List of default fields to display in table.
     */
    protected LyInfoList<LyFieldInfo> getDefaultFields()
    {
        LyInfoList<LyFieldInfo> fields = _classInfo.getFieldInfos().getInfos(
                LyFieldInfo.MARK_HIDDEN, true);

        // Remove multiple fields, arrays and structs
        LyFieldInfo field;
        int         i, n = fields.getSize();

        for (i = n - 1; i >= 0; i--)
        {
            field = fields.getInfo(i);

            if (field.hasMark(LyFieldInfo.MARK_SECRET) || field.isMultiple()
                    || (field instanceof LyArrayFieldInfo) || (field instanceof LyStructFieldInfo)
                    || hasNotMark(field, LyFieldInfo.MARK_TABLE))
                fields.removeInfo(field);
        }

        if (fields.getSize() == 0)
        {
            // Take all fields except secret ones
            fields = _classInfo.getFieldInfos().getInfos(LyFieldInfo.MARK_HIDDEN, true);
            n = fields.getSize();

            for (i = n - 1; i >= 0; i--)
            {
                field = fields.getInfo(i);

                if (field.hasMark(LyFieldInfo.MARK_SECRET))
                    fields.removeInfo(field);
            }

        }

        return fields;
    }

    /**
     * Gets the list of real fields being displayed in the columns of the table The default
     * implementation returns the list of fields given by the getFields method.
     *
     * @return the list of real fields being displayed
     * @see    #getFields()
     */
    public LyInfoList<LyFieldInfo> getRealFields()
    {
        return getFields();
    }

    /**
     * Load next/previous page when a page loading is active. This method does noting if the limit
     * parameter is not set in the meta model. If the parameter is set, the the next (preceding)
     * block of data (maximum count of item specified in limit parameter) are shown
     *
     * @param next indicate we want the next pages (items after the last item of the list) If this
     *             parameter is false, previous page is loaded
     */
    protected void loadPage(boolean next)
    {
        LyTable table = getTableComponent();

        if (table == null)
            // An error occurred, just return
            return;

        synchronized (_workSpace)
        {
            getViewManager().setWaiting(true);

            // Store the current idle state
            boolean wasIdle = table.isIdle();

            // Freeze the table
            setIdle(true);

            // Loads datas.
            if (!(_initialWorkSpace instanceof LyPageWorkSpace))
            {

                if (next)
                    _currentPage++;
                else if (_currentPage > 0)
                    _currentPage--;
            }
            else
                ((LyPageWorkSpace) _initialWorkSpace).loadPage(next);

            // Update the last modification time.
            setLastModified(new Date().getTime());

            // Restore the idle state
            setIdle(wasIdle);
            getViewManager().setWaiting(false);
        }
    } // end method loadPage

    /**
     * Translate the given key to a displayable string. This method is provided for optimization.
     *
     * @param  key the string to translate
     * @return the translated string if it could be found, the key parameter otherwise
     */
    public final String translate(String key)
    {

        if (key == null)
            return null;

        if (_strings == null)
            _strings = new Hashtable<String, String>();

        String value = _strings.get(key);

        if (value != null)
            return value;

        LyEnvironment environment = getEnvironment();

        value = environment.translate(key);
        _strings.put(key, value);
        return value;
    }

    /**
     * Clear the translated string map.
     */
    public final void clearTranslatedStringMap()
    {
        _strings = null;
    }

    /**
     * The fast filter edition form is closed : notify registered listener.
     *
     * @param controller   reference to the fast filter controller (filter form which is closing)
     * @param actionDemand actionDemand on filter controller : id of the command such as
     *                     LyFormController.DEMAND_VALIDATE, LyFormController.DEMAND_CANCEL
     * @param filterValue  the filter value that was typed in the form
     * @param filter       the filter that was created in the form
     * @param fieldInfo    the fast filter field info
     * @param extend       indicates if fast filter creation is an extension of current filter.
     */
    public void endFastFilter(LyController controller, String actionDemand, String filterValue,
        LyFilter filter, LyFieldInfo fieldInfo, boolean extend)
    {
        _fastFilterFieldInfo = null;

        if (filter != null)
        {
            _fastFilterValue = filterValue;
            _lastFastFilterFieldInfo = fieldInfo;

            // If extension on current filter, create a new filter of type FILTER_AND
            if (extend)
            {
                LyFilter oldFilter = getFilter();

                if (oldFilter != null)
                {
                    LyFilter newFilter = oldFilter;

                    if (newFilter.getType() != LyFilter.Type.AND)
                    {
                        newFilter = new LyFilter(LyFilter.Type.AND);
                        newFilter.addElement(oldFilter);
                    }

                    newFilter.addElement(filter);
                    filter = newFilter;
                }
            }
        }
        else
        {
            _fastFilterValue = null;
            _lastFastFilterFieldInfo = null;

            // If a preference is set on the table, set the preference filter
            LyPreference	preference = getPreference();

            filter = ((preference == null) ? null : preference.getFilter());
        }

        if ((_workSpace instanceof LyPageWorkSpace) && !checkFilterForPhysicalLayer(filter))
        {
            // ask the user if he really want to apply this filter
            LyViewManager vm = getViewManager();
            vm.askQuestion(_component, "LY_CONFIRM_NOT_PHYSICAL_FILTER", this, filter);
        }
        else
            setFilter(filter);
    } // end method endFastFilter

    /**
     * The sort edition form is closed : notify registered listener.
     *
     * @param parent       reference to the sort controller (sort form which is closing)
     * @param actionDemand actionDemand on sort controller : id of the command such as
     *                     LyFormController.DEMAND_VALIDATE, LyFormController.DEMAND_APPLY,
     *                     LyFormController.DEMAND_CANCEL
     * @param sort         the sort that was created in the form
     */
    public void endSort(LyController parent, String actionDemand, LySort sort)
    {

        if (!actionDemand.equals(LyFormController.DEMAND_CANCEL))
        {

            // if table has _limit param, we should check if the sort is valid for the physical layer
            if ((_workSpace instanceof LyPageWorkSpace) && !checkSortForPhysicalLayer(sort))
            {
                // ask the user if he really want to apply this filter
                LyViewManager vm = getViewManager();
                vm.askQuestion(_component, "LY_CONFIRM_NOT_PHYSICAL_SORT", this, sort);
            }
            else
            {
                setSort(sort);
                setLastModified(new Date().getTime());
            }
        }
    }


    /**
     * Check if the given sort is valid for physical layer.
     * All its fields must be valid for the physical layer.
     * @param sort the sort to check
     * @return true if the sort is valid.
     */
    protected boolean checkSortForPhysicalLayer(LySort sort)
    {

        // check if a criterion
        if (sort == null)
            return true;

        int             i, nbCriterion = sort.getCriterionCount();
        LySortCriterion criterion;

        for (i = 0; i < nbCriterion; i++)
        {
            criterion = sort.getCriterion(i);

            if (isFieldValidForSortOrFilter(criterion.getFieldInfo()))
                continue;

            // if a single field is not valid, stop checkin and return false;
            return false;
        }

        return true;
    }


    /**
     * Print the content of the list Run it asynchronous.
     */
    public void printList()
    {
        LyApplication application = getApplication();
        LyAction      action = application.getAction(DEMAND_PRINT_TABLE);

        if (action == null)
            action = new LyAction(DEMAND_PRINT_TABLE, application.getAction(LyAction.ID_PRINT_TABLE),
                    _classInfo, null);

        LyController controller = _session.invokeAction(this, action, _classInfo, _workSpace,
                false);

        if (controller instanceof LyPrintController)
        {
            LyPrintController printController = (LyPrintController) controller;

            _printMode = true;

            printController.setFields(getFields());
            printController.setByList(true);

            _printMode = false;

        }

        controller.showView();
    }

    /**
     * Method that should be called when starting the run method when the controller is Runnable.
     * Modified to register the printing thread on the controller.
     *
     * @see #endRun
     */
    protected void startRun()
    {
        super.startRun();

        if (getViewController() != null)
            getViewController().setPrintThread(Thread.currentThread());
    }

    /**
     * Method that should be called when ending the run method when the controller is Runnable.
     * Modified to unregister the printing thread on the controller.
     *
     * @see #startRun
     */
    protected void endRun()
    {

        if (getViewController() != null)
            getViewController().setPrintThread(null);

        super.endRun();
    }

    /**
     * Starting printList request.
     */
    public void run()
    {
        startRun();

        // Print table
        LyTable table = getTableComponent();

        table.print(this);

        endRun();
    }

    /**
     * Delete object infos associated to a prefix.
     *
     * @param prefix Prefix associated to object infos that must be deleted.
     */
    protected void deleteObjectInfos(String prefix)
    {
        // Create a dialog to type in the search parameters
        LySession             session = getSession();
        LyUserInputController infoCtrl;
        LyApplication         application = session.getApplication();
        LyAction              action = application.getAction("_delete");

        infoCtrl = new LyUserInputController(session, this, action);

        _LyDeleteObjectInfosBehavior behavior = new _LyDeleteObjectInfosBehavior();

        infoCtrl.setBehavior(behavior);
        behavior.init(infoCtrl);

        // FFT 000947
        infoCtrl.setBuilder(new _LyObjectInfosInputBuilder(infoCtrl));

        infoCtrl.setTitleBar("LY_ACTION_DELETE");
        infoCtrl.setImage(getImage(_action));

        LyInfoList<LyObjectInfo> infos = application.getInfoList(LyObjectInfo.class);
        LyObjectInfo             info;
        int                      i, n = ((infos == null) ? 0 : infos.getSize());
        LyOption                 option;
        LyChoice                 choice = infoCtrl.addChoice(DEMAND_DELETE_OBJECTINFOS, null, true,
                true);

        choice.setFill(LyChoice.Fill.VERTICAL);
        choice.setMaxSpan(6);

        if (n == 0)
            return;

        // Sort by name
        infos = infos.getInfosByName();

        for (i = 0; i < n; i++)
        {
            info = infos.getInfo(i);

            if (!info.getId().startsWith(prefix + '_'))
                continue;

            if ((info instanceof LyGetClassInfoInterface)
                    && (getClassInfo() != ((LyGetClassInfoInterface) info).getClassInfo()))
                continue;

            option = new LyOption(info.getId(), info.getName());

            choice.addOption(option);
        }

        infoCtrl.showView();

        // Waiting for userInput event.
    } // end method deleteObjectInfos

    /**
     * Implementation of LyPrintListenerInterface.
     * @param page the page to print
     */
    public void print(int page)
    {
    }

    /**
     * Accessor on internal classFilter : filter always applied on data shown in view. The class
     * filter applies wether a user filter is set or not. This is a low level filter always set on
     * data. This value is usually set in the table action by the '_classFilter' parameter. Default
     * is null.
     *
     * @return current classFilter value, or null if no filter is set (default).
     */
    public LyFilter getClassFilter()
    {
        return _classFilter;
    }

    /**
     * Accessor on internal classFilter : filter always applied on data shown in view. The class
     * filter applies wether a user filter is set or not. This is a low level filter always set on
     * data. This value is usually set in the table action by the '_classFilter' parameter.
     *
     * @param filter  classFilter value, or null if no filter is set (default).
     * @param reset   indicates wheter the content is immediately calculated and view is refreshed.
     *                This parameter should be true to recalculate content. This is useless only if
     *                the content is explicitely reset by the application, using setContext for
     *                instance.
     */
    public void setClassFilter(LyFilter filter, boolean reset)
    {
        _classFilter = filter;

        if (reset)
            // This forces to reset the internal workspace.
            reset(true);
    }

    /**
     * Sets the displayed fields for this controller.
     *
     * @param fields List of fields to display in the table. If null, return to default list of
     *               fields.
     */
    public void setFields(LyInfoList<LyFieldInfo> fields)
    {

        if (fields != null)
            fields.lock();

        LyFieldInfo fixedField = null;

        if (_fixedColumn >= 0)
        {
            LyInfoList<LyFieldInfo> tableFields = getFields();

            fixedField = tableFields.getInfo(_fixedColumn);
        }

        _fields = fields;

        if (_component == null)
            return;

        // Rebuild table
        LyBuilder builder = getBuilder();

        // the builder may be a fileBuilder, then create
        // a table builder to create a new table component
        if (!(builder instanceof LySimpleTableBuilder))
            builder = new LyTableBuilder(this);

        // If view manager allows table headers to be updated, just change the table headers.
        // Otherwise, we need to rebuild the entire table.
        if (getViewManager().hasProperty("ALLOW_TABLE_HEADERS_UPDATE"))
        {
            ((LySimpleTableBuilder) builder).rebuildHeaders();
            ((LySimpleTableBuilder) builder).rebuildFooters();
        }
        else
            ((LySimpleTableBuilder) builder).rebuildTable(getTableComponent());

        // Update "Fix column" sub-menu
        if (builder instanceof LyTableBuilder)
        {
            LyMenuBar menuBar = null;

            if (hasMenu())
            {
                LyView view = getView();

                menuBar = ((view instanceof LyFrame) ? ((LyFrame) view).getMenuBar() : null);
            }

            if (menuBar != null)
                ((LyTableBuilder) builder).updateFixColumnSubMenu(menuBar);

            if (fixedField != null)
            {
                LyInfoList<LyFieldInfo> tableFields = getFields();
                int                     column = tableFields.getIndex(fixedField);

                if (column >= 0)
                    fixColumn(column);
                else
                    fixColumn(-1);
            }

            setColumnSizes(_columnSizes);
        }

        // Reset content
        if (_initialized && !isIdle())
        {
            LyObjectSource source = _classInfo.getObjectSource();
            boolean        full = (_fields == null) ? false : (!source.areFieldsLoaded(_fields));

            reset(full);
        }
    } // end method setFields

    /**
     * Indicates if this table notifies the outline.
     *
     * @return A boolean which indicates if this table notifies the outline. Always true here.
     */
    protected boolean notifyOutline()
    {
        return true;
    }

    /**
     * Method that indicates if controlled panel enables multi-selection or not.
     *
     * @return A boolean indicating if multi-selection is allowed or not.
     */
    public boolean enablesMultiSelection()
    {
        LyTable table = getTableComponent();

        return ((table == null) ? true
                                : (table.getSelectionMode() != LyTable.SelectionMode.SINGLE));
    }

    /**
     * Complete given preference.
     * @param preference the preference to complete
     */
    protected void completePreference(LyPreference preference)
    {
        super.completePreference(preference);

        LySort      sort = getSort();
        LyFieldInfo fixedColumn = null;

        if (_fixedColumn >= 0)
        {
            LyInfoList<LyFieldInfo> fields = getFields();

            fixedColumn = fields.getInfo(_fixedColumn);
            preference.setFixedColumn(fixedColumn);
        }

        if (sort != null)
            preference.setSort(sort);

        if (_fields != null)
        {
            LyInfoList<LyFieldInfo> tableFields = getTableComponentFields(true);

            if ((tableFields != null) && (tableFields.getSize() > 0))
                setFields(tableFields);

            preference.setFieldInfos(getRealFields());
        }

        if (_maxRowCount > 0)
            preference.setNbObjects(_maxRowCount);

        preference.setColumnSizes(_columnSizes);

        if (_displayList)
            preference.setShowAs(LyPreference.ShowAs.LIST);
    } // end method completePreference

    /**
     * Sets the preference for this controller.
     * @param preference the preference to set
     */
    public void setPreference(LyPreference preference)
    {
        boolean idle = isIdle();

        if (!idle)
            setIdle(true);

        closeBorderPanel();

        boolean hasField = (_preference == null ? false : _preference.getFieldInfos() != null);
        super.setPreference(preference);

        LyInfoList<LyFieldInfo> oldFields = _fields;

        // Update fields only if needed
        if (preference != null)
        {
            LyInfoList<LyFieldInfo> fields = preference.getFieldInfos();

            if ((fields != null) && !fields.equals(oldFields))
                setFields(new LyInfoList<LyFieldInfo>(fields));
            else if (hasField && (fields == null))
            {
                _fields = null; // to force recomputing

                fields = getFields();

                if ((fields != null) && !fields.equals(oldFields))
                    setFields(fields);
            }
        }
        else
        {
            _fields = null; // to force recomputing

            LyInfoList<LyFieldInfo> fields = getFields();

            if ((fields != null) && !fields.equals(oldFields))
                setFields(fields);
        }

        // Update sort
        LySort sort = ((preference == null) ? getDefaultSort() : preference.getSort());

        if (sort != null)
            setSort(sort.cloneSort());

        // Unfix column
        LyFieldInfo fieldInfo = ((preference == null) ? null : preference.getFixedColumn());

        if ((preference == null) || (fieldInfo != null))
            fixColumn(-1);

        // Set number of objects
        int nbObjects = ((preference == null) ? -1 : preference.getNbObjects());

        _maxRowCount = nbObjects;

        // Fix column
        if (fieldInfo != null)
        {
            LyInfoList<LyFieldInfo> fields = getFields();
            int                     pos = fields.getIndex(fieldInfo);

            if (pos >= 0)
                fixColumn(pos);
        }

        // Set column sizes
        String columnSizes = ((preference != null) ? preference.getColumnSizes() : null);

        setColumnSizes(columnSizes);

        // Restore display
        if ((preference == null) && !_displayListToggleButton)
        {
        	// Set initial display list value
            String display = (String) getParameter("_displayList", true);

            if (getViewManager().hasProperty(LyMobileViewManager.MOBILE_ONLY_PROPERTY))
            	display = "LIST";

            if (display != null)
            {
                if ("LIST".equalsIgnoreCase(display))
                	setDisplayList(true);
                else if ("TABLE".equalsIgnoreCase(display))
                	setDisplayList(false);
            }
        }
        else
        {
            boolean displayList = (preference != null)
            	&& (preference.getShowAs() == LyPreference.ShowAs.LIST);

        	setDisplayList(displayList);
        }

        if (!idle)
        {
            reset(true);
            setIdle(false);

            updatePreferencesMenu();
        }

        // Cancel any fast filter
        _fastFilterFieldInfo = null;
        _fastFilterValue = null;
        _lastFastFilterFieldInfo = null;
    } // end method setPreference

    /**
     * Method for allowing fast filter.
     *
     * @return True if fast filter is enabled, false otherwise.
     */
    protected boolean enableFastFilter()
    {
        return getEnvironment().testResource("LY_FAST_FILTER");
    }

    /**
     * Method for allowing fast edition.
     *
     * @return True if fast edition is enabled, false otherwise.
     */
    protected boolean enableFastEdition()
    {
        return getEnvironment().testResource("LY_FAST_EDITION");
    }

    /**
     * Updates groupBy menu according to current sort
     */
    protected void updateGroupByMenu()
    {
    }

    /**
     * Get the row index of an object
    * @param object an object
    * @return the row index
    */
    protected final int getRowIndex(LyObject object)
    {
        LyTable table = getTableComponent();

        return getRowIndex(table.getRows(), object, new Integer[] { 0 });
    }

    /**
     * Method for getting real row from rowIndex or object.
     * @param rows List of rows where row is searched.
     * @param object the object to find
     * @param rowIndex Index of searched row.
     *                 Note : an array with 1 entry is used to allow rowIndex modification during recursive search.
     * @return Found row or null if not found.
     */
    protected final int getRowIndex(LyRows rows, LyObject object, Integer[] rowIndex)
    {

        if (rows == null)
            return -1;

        int    i, n = rows.getRowCount();
        LyRow  row = null;
        LyRows subRows;
        int    result;
        String objectId = object.getId();
        String rowId;

        for (i = 0; i < n; i++)
        {
            row = rows.getRow(i);
            rowId = row.getId();

            if ((rowId != null) && rowId.equals(objectId))
                return rowIndex[0];

            rowIndex[0]++;

            subRows = row.getRows();

            if ((subRows != null) && row.isOpen())
            {
                result = getRowIndex(subRows, object, rowIndex);

                if (result >= 0)
                    return result;
            }
        }

        return -1;
    }

    /**
     * Method for getting real row from rowIndex.
     * @param rowIndex Index of searched row.
     *                 Note : an array with 1 entry is used to allow rowIndex modification during recursive search.
     * @return Found row or null if not found.
     */
    protected final LyRow getRow(int rowIndex)
    {
        LyTable table = getTableComponent();

        return getRow(table.getRows(), new Integer[] { rowIndex });
    }

    /**
     * Method for getting real row from rowIndex.
     * @param rows List of rows where row is searched.
     * @param rowIndex Index of searched row.
     *                 Note : an array with 1 entry is used to allow rowIndex modification during recursive search.
     * @return Found row or null if not found.
     */
    protected final LyRow getRow(LyRows rows, Integer[] rowIndex)
    {

        if (rows == null)
            return null;

        int    i, n = rows.getRowCount();
        LyRow  row = null;
        LyRows subRows;
        LyRow  result;

        for (i = 0; i < n; i++)
        {
            row = rows.getRow(i);

            if (rowIndex[0] == 0)
                return row;

            rowIndex[0]--;

            subRows = row.getRows();

            if ((subRows != null) && row.isOpen())
            {
                result = getRow(subRows, rowIndex);

                if (result != null)
                    return result;
            }
        }

        return null;
    }

    /**
     * Method for counting real rows.
    * @param rows the LyRows
    * @return a number of row for the given LyRows
     */
    protected final int getRowCount(LyRows rows)
    {

        if (rows == null)
            return 0;

        int    i, n = rows.getRowCount();
        LyRow  row;
        int    count = 0;
        String rowId;

        for (i = 0; i < n; i++)
        {
            row = rows.getRow(i);
            if (row == null)
            	continue;

            rowId = row.getId();
            if ((rowId != null) && !rowId.startsWith(SECTION_ROW_PREFIX)
                    && !rowId.startsWith(FOOTER_ROW_PREFIX))
                count++;

            count += getRowCount(row.getRows());
        }

        return count;
    }

    /**
     * Find a row from an object id.
    * @param rows a LyRows
    * @param objectId an objectId
    * @return the corresponding LyRow
     */
    protected LyRow findRow(LyRows rows, String objectId)
    {

        if (rows == null)
            return null;

        // Get the right row
        int    i, n = rows.getRowCount();
        LyRow  row;
        String rowId;

        for (i = 0; i < n; i++)
        {
            row = rows.getRow(i);
            rowId = row.getId();

            if ((rowId != null) && rowId.equals(objectId))
                return row;
            else
            {
                row = findRow(row.getRows(), objectId);
                if (row != null)
                    return row;
            }
        }

        return null;
    }

    /**
     * Removes a row from its parent.
     * @param row the row to remove
     */
    protected void removeRow(LyRow row)
    {
        LyRows rows = (LyRows) row.getParent();

        rows.removeRow(row);

        int nr = rows.getRowCount();

        if ((nr == 0) && (rows.getParent() instanceof LyRow))
            removeRow((LyRow) rows.getParent());
    }

    /**
     * Gets limit parameter for pagination.
     * @return -1 means no limit, a positive value indicates the number of objects per page.
     */
    public int getLimit()
    {
        String limitParameter = (String) _action.getParameter("_limit", true);
        int    limit = ((limitParameter == null) ? -1 : LyTools.intFromString(limitParameter)); // -1 means no limit

        if (limit == 0)
            limit = 100;
        else if (limit < 0)
            limit = -1;

        return limit;
    }

    /**
     * (De)Activate print mode. This method is invoked with true before printing the to false at the
     * end of the print task. The view/controller may need to change the content at that time.
     *
     * @param status true if printing is active, false otherwise.
     */
    public void setPrintMode(boolean status)
    {

        if (status != _printMode)
        {
            _printMode = status;
        }
    }

    /**
     * Checks if print mode is activated.
     * @return True if print mode is on, false otherwise.
     */
    public final boolean hasPrintMode()
    {
        return _printMode;
    }

    /**
     * Set the fop mode to the given status
     * @param status the new fop status
     */
    public final void setFopMode(boolean status)
    {
        _fopMode = status;
    }

    /**
     * Checks if fop mode is activated.
     * @return True if fop mode is on, false otherwise.
     */
    public final boolean isFopMode()
    {
        return _fopMode;
    }
    /**
     * Checks if we have to add toggle button.
     * @return True if toggle button, false otherwise.
     */
    public final boolean hasDisplayListToggleButton()
    {
        return _displayListToggleButton;
    }

    /**
     * Answer to a askQuestion request.
     *
     * @param result true if the user has clicked on Yes or Ok
     * @param client data specified in the askQuestion method
     * @see   LyQuestionAnswerInterface
     */
    public void questionAnswer(boolean result, Object clientData)
    {

        if (result)
        {

            if (clientData instanceof LySort)
            {
                setSort((LySort) clientData);
                setLastModified(new Date().getTime());
            }
            else if (clientData instanceof LyFilter)
                setFilter((LyFilter) clientData);
        }
    }

    /**
     * Return the number of columns in display list.
     * 1 column for status, 1 for first image field, 1 for field name, 1 for field value, 1 for each footer field.
     * @return Int number of columns
     */
    public int getColsNumberListDisplay()
    {
        // SMA
        LyInfoList<LyFieldInfo> fields = getFields();
        LyInfoList<LyFieldInfo> statusFields = getStatusFieldInfos();
        LyClassInfo             classInfo = getClassInfo();
        LyFieldInfo             fieldInfo;
        int                     colsNumber = 0;

        if (fields != null)
        {
            if (_displayListShowFieldLabels)
                // 3 columns : Image fields ( always present ) , one for fields name, 1 for fields values
                colsNumber = 3;
            else
                // 1 for image fields, 1 for fields values
                colsNumber = 2;

            for (int i = 0; i < fields.getSize(); i++)
            {
                fieldInfo = fields.getInfo(i);

                if (!classInfo.contains(fieldInfo))
                    // complex table
                    colsNumber++;

                // how many footer fields ?
                if ((fieldInfo instanceof LyNumberFieldInfo)
                        && (((LyNumberFieldInfo) fieldInfo).getFunction()
                            != LyNumberFieldInfo.Function.NONE) && classInfo.contains(fieldInfo))
                    colsNumber++;
            }

            // status column ?
            if (isStatusColumn(statusFields))
                colsNumber++;
        }

        return colsNumber;
    }

    /**
     * Sets display mode.
     * @param displayList True if list mode, false if table mode.
     */
    protected void setDisplayList(boolean displayList)
    {
        // Set display
        _displayList = displayList;

        LySimpleTableBuilder builder = (LySimpleTableBuilder) getBuilder();

        if (getViewManager().hasProperty("ALLOW_TABLE_HEADERS_UPDATE"))
        {
            builder.rebuildHeaders();
            builder.rebuildFooters();
        }
        else
            builder.rebuildTable(getTableComponent());

        // Update display list tool state
        if (hasDisplayListToggleButton())
        {
        	LyComponent	component = getComponent();
        	if (component != null)
        	{
        	    LyTool		tool = (LyTool)component.getComponent(DEMAND_LIST_DISPLAY + LyController.ID_SEP + getId());

        	    if (tool != null)
        	        tool.setPressed(displayList);
        	}
        }

        reset(false);
    }

} // end class LySimpleTableController
