/**
 * LySortWizardController.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved. Current version: $Id:
 * LySortWizardController.java,v 6.11 2009/01/06 11:03:39 jls Exp $
 */

package leon.control;

import leon.app.*;
import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;

import java.util.*;


/**
 * This class implements the sort wizard edition area. This form enables the user to type in
 * elementary sort criterion (fieldInfo / direction).
 */
public class LySortWizardController extends LyAreaController
{
	/** File Version */
	public static final String __VERSION = "$Id: LySortWizardController.java 36510 2012-08-17 15:43:06Z jlsaouli $";

	/** Sort Wizard id : Sort Wizard field. */
	public static final String FIELD = "_sortWizardField".intern();
	/** Sort Wizard id : Sort Wizard direction. */
	public static final String DIRECTION = "_sortWizardDirection".intern();
	/** Sort Wizard id : Sort Wizard direction decreasing. */
	public static final String DIR_DECREASING = "_sortWizardDirDecreasing".intern();
	/** Sort Wizard id : Sort Wizard direction increasing. */
	public static final String DIR_INCREASING = "_sortWizardDirIncreasing".intern();
	/** Sort Wizard id : Sort Wizard group by. */
	public static final String GROUP_BY = "_sortWizardGroupBy".intern();
	/** Sort Wizard id : Sort Wizard grouped. */
	public static final String GROUPED = "_sortWizardGrouped".intern();

	/** Default number of fields */
	public static int  DEFAULT_FIELD_NUMBER = 3;

	/** Default number of Group by */
	public static int  MAX_GROUP_BY = 3;

	/** The classInfo */
	protected LyClassInfo _classInfo;

	/** The current sort */
	protected LySort _sort;

	/** The target interface */
	protected LySortEndInterface _sortEndInterface;

	/** Store the maximum number of sort field for the dialog. */
	protected int _maxItem;

	/** List of fields used for sort. */
	protected LyInfoList<LyFieldInfo> _fields;

	/**
	 * Default constructor
	 * @param session the current session
	 * @param parent parent controller
	 * @param action the corresponding action
	 * @param classInfo the associated classInfo
	 * @param objects the context
	 */
	public LySortWizardController(LySession session, LyController parent, LyAction action, LyClassInfo classInfo,
		LyObjectList objects)
	{
		super(session, parent, action, objects);

		_maxItem = -1;
		_fields = null;

		if (parent instanceof LySortEndInterface)
			setSortEndInterface((LySortEndInterface)parent);

		if (action.getClassInfo() != null)
			_classInfo = action.getClassInfo();
		else
			_classInfo = classInfo;
	}

	/**
	 * Gets fields available for sort action.
	 * @return a list of fields
	 */
	protected LyInfoList<LyFieldInfo> getFields()
	{
		if (_fields != null)
			return _fields;

		LyInfoList<LyFieldInfo> fields = null;
		LyFieldInfo			    fieldInfo = (LyFieldInfo)_action.getParameter("_fieldInfo", true);
		int					    np = ((fieldInfo != null) ? 2 : 1);

		do
		{
			if (fieldInfo != null)
			{
				if (fields == null)
					fields = new LyInfoList<LyFieldInfo>();

				fields.addInfo(fieldInfo);
			}

			fieldInfo = (LyFieldInfo)_action.getParameter("_fieldInfo" + np, true);
			np++;
		}
		while (fieldInfo != null);

		if (fields == null)
		{
			BitSet fieldMarks = getParameterMarks("_fieldMarks", LyBitSet.Type.F_MARKS);

			if (fieldMarks != null)
				fields = _classInfo.getFieldInfos(fieldMarks);
		}

		if ((fields == null) || (fields.getSize() < 1))
		{
			fields = _classInfo.getFieldInfos();
			fields = fields.getInfos(LyFieldInfo.MARK_HIDDEN, true);
			fields = fields.getInfos(LyFieldInfo.MARK_SECRET, true);
		}
	    
	    if ((fields != null)
	    		&&(_sortEndInterface instanceof LyTableController))
	    {
	    	int		i, n = fields.getSize();
	    	
	    	for (i = n - 1; i >= 0; i--)
	    	{
	    		fieldInfo = fields.getInfo(i);
	    		
	    		if (!((LyTableController)_sortEndInterface).allowSortOnField(fieldInfo))
	    			fields.removeInfo(fieldInfo);
	    	}
	    }

		return fields;
	} // end method getFields

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return  Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyActionBehaviorInterface.class;
	}

	/**
	 * Indicates if the view for this form controller is resizable.
	 *
	 * @return  false since form views are not resizable.
	 */
	public boolean isViewResizable()
	{
		return false;
	}

	/**
	 * Setter for the interface that will receive the new sort.
	 *
	 * @param  sortInterface  the interface
	 */
	public void setSortEndInterface(LySortEndInterface sortInterface)
	{
		_sortEndInterface = sortInterface;
		_sort = _sortEndInterface.getSort();
		if (_sort != null)
			if (_initialized)
				updateFieldsFromSort();
	}

	/**
	 * Accessor to the interface that will receive the new sort notification.
	 *
	 * @return  the interface
	 */
	public LySortEndInterface getSortEndInterface()
	{
		return _sortEndInterface;
	}

	/**
	 * Initialize the area.
	 */
	protected void initArea()
	{
		super.initArea();

		if (isFreed())
			return;

		_initialized = false;
		updateFieldsFromSort();
		_initialized = true;
	}

	/**
	 * Get the max number of item for the sort wizard.
	 *
	 * @return  the max number of item
	 */
	public int getMaxItem()
	{
		if (_maxItem < 0)
		{
			LyEnvironment		    environment = getEnvironment();
			LyInfoList<LyFieldInfo> sortFields = getFields();
			String				    maxItemAsString = environment.getResource("LY_MAX_ITEM");

			try
			{
				_maxItem = Integer.parseInt(maxItemAsString);
			}
			catch (NumberFormatException e)
			{
				_maxItem = DEFAULT_FIELD_NUMBER;
			}

			if ((_sort != null) && (_sort.getCriterionCount() > _maxItem))
				_maxItem = _sort.getCriterionCount();

			if (sortFields.getSize() < _maxItem)
				_maxItem = sortFields.getSize();
		}

		return _maxItem;
	}

	/**
	 * Factory of the default builder for a controller. Current implementation returns a
	 * LyFilterBuilder instance.
	 *
	 * @return  the builder which creates the XML description of the view
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LySortWizardBuilder(this);
	}

	/**
	 * Update the fields of the form from the current sort.
	 */
	protected void updateFieldsFromSort()
	{
		if (_maxItem < 0)
			getMaxItem();

		Vector<String>  fields = new Vector<String>(_maxItem);
		Vector<String>  directions = new Vector<String>(_maxItem);
		Vector<Boolean> groups = new Vector<Boolean>(_maxItem);

		if (_sort != null)
		{
			int             n = _sort.getCriterionCount();
			LySortCriterion criterion;
			boolean		    increasing, groupBy;
			LyFieldInfo     fieldInfo;

			for (int i = 0; i < n; i++)
			{
				criterion = _sort.getCriterion(i);
				fieldInfo = criterion.getFieldInfo();
				increasing = criterion.isIncreasing();
				groupBy = criterion.isGroupBy();
				fields.addElement(fieldInfo.getId());

				if (increasing)
					directions.addElement(DIR_INCREASING);
				else
					directions.addElement(DIR_DECREASING);

				groups.addElement(groupBy);
			}
		}

		updateFields(fields, directions, groups);
	} // end method updateFieldsFromSort

	/**
	 * Update the fields from the current sort.
	 *
	 * @param  fields      a vector containing all the name of fields already selected
	 * @param  directions  the sort direction for the fields
	 * @param  groups      the groupBy feature for the fields
	 */
	protected void updateFields(Vector<String> fields, Vector<String> directions, Vector<Boolean> groups)
	{
        if (isIdle())
            return;

        setIdle(true);

		String	    field, dir;
		Boolean     grouped;
		LyFormField relationFormField, choiceFormField, groupFormField;
		LyFieldInfo fieldInfo;
		LyRelation  relation;
		LyChoice    choice, group;
		LyLink	    link;
		LyComponent root = getComponent();
		String	    empty_value = "LY_RELATION_EMPTY_VALUE";
		String	    val = getEnvironment().translate(empty_value);
		int		    i, n = _maxItem;
        boolean     enableGroupBy = true;

		if ((val != null) && val.equals("_null"))
			empty_value = null;

		for (i = 0; i < n; i++)
		{
			relationFormField = (LyFormField)root.getComponent(FIELD + i);
			relation = (LyRelation)relationFormField.getFieldItem();

			choiceFormField = (LyFormField)root.getComponent(DIRECTION + i);
			choice = (LyChoice)choiceFormField.getFieldItem();

			groupFormField = (LyFormField)root.getComponent(GROUP_BY + i);
            group = (groupFormField != null ? (LyChoice)groupFormField.getFieldItem() : null);

			// Clear current relation
			relation.setIdle(true);
			for (int j = relation.getLinkCount() - 1; j >= 0; j--)
			{
				link = relation.getLink(j);
				relation.removeLink(link);
			}

			relation.setIdle(false);
			// Add empty field
			relation.addLink(new LyLink("LY_NO_CHOICE", empty_value));

			// Add current selected field
			if (i < fields.size())
			{
				field = fields.elementAt(i);
				fieldInfo = _classInfo.getFieldInfo(field, false);

				if (fieldInfo != null)
				{
					link = new LyLink(fieldInfo.getId(), fieldInfo.getAlias());
					relation.addLink(link);
					link.setSelected(true);
				}
			}
			else
				fieldInfo = null;

			LyInfoList<LyFieldInfo> list = getFields();

			// Sort by name depending on a resource
			list = list.getInfosByName(getEnvironment());

			int j, nl = list.getSize();

            enableGroupBy = enableGroupBy
	        	&& (list.getSize() - i > 1)
	        	&& enableGroupBy(fieldInfo);

			relationFormField.setEnabled(i <= fields.size());
			relationFormField.setEditable(i <= fields.size());
			choiceFormField.setEnabled(i < fields.size());
			choiceFormField.setEditable(i < fields.size());

			if (groupFormField != null)
            {
                groupFormField.setEnabled(enableGroupBy);
                groupFormField.setEditable(enableGroupBy);
            }

			if ((i < directions.size())
                && (fieldInfo != null))
			{
				dir = directions.elementAt(i);

				if ((dir == null) || (dir.equals(DIR_INCREASING)))
					choice.getOption(0).setChecked(true);
				else
					choice.getOption(1).setChecked(true);
			}
			else
				choice.getOption(0).setChecked(true);

            if (group != null)
            {
                grouped = (enableGroupBy && (i < groups.size()) && groups.get(i));
                group.getOption(0).setChecked(grouped);
                enableGroupBy = grouped;
            }

			for (j = 0; j < nl; j++)
			{
				fieldInfo = list.getInfo(j);
				if (fields.contains(fieldInfo.getId()))
					continue;

				relation.addLink(new LyLink(fieldInfo.getId(), fieldInfo.getAlias()));
			}
		}

        setIdle(false);
	} // end method updateFields

	/**
	 * Invoked when the user validates the content of the input.
	 *
	 * @param  demand      demand parameters sent by the view component.
	 */
	public void actionDemand(LyActionDemand demand)
	{
		String id = demand.getSource();

		if (id.equals(LyFormController.DEMAND_VALIDATE)
				|| id.equals(LyFormController.DEMAND_APPLY))
		{
			LyParameterValues paramValues = demand.getParameterValues();

			if (paramValues != null)
			{
				LySort result = buildSortFromValues(paramValues);

				if (_sortEndInterface != null)
					_sortEndInterface.endSort(this, id, result);

				if (id.equals(LyFormController.DEMAND_VALIDATE))
					free();
				else
					updateFieldsFromSort();
			}

			return;
		}

		if (id.equals(LyFormController.DEMAND_CANCEL))
		{
			if (_sortEndInterface != null)
				_sortEndInterface.endSort(this, id, null);

			free();
			return;
		}
		else if (demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
		{
			LyParameterValues paramValues = demand.getParameterValues();

			if (paramValues != null)
			{
				Vector<String>  fields = new Vector<String>(_maxItem);
				Vector<String>  directions = new Vector<String>(_maxItem);
                Vector<Boolean> groups = new Vector<Boolean>(_maxItem);
				String		    field, choice, group;
				int			    i;

				for (i = 0; i < _maxItem; i++)
				{
					// Update fields list
					field = paramValues.getParameter(FIELD + i);
                    if ((field != null)
                        && !field.equals("LY_NO_CHOICE"))
                    {
                        fields.add(field);

    					choice = paramValues.getParameter(DIRECTION + i);
    					if (choice != null)
    						directions.add(choice);
    					else
    						directions.add("");

                        group = paramValues.getParameter(GROUP_BY + i);
                        groups.add(Boolean.valueOf(group != null && group.length() > 0));
                    }
				}

				updateFields(fields, directions, groups);

				if (!_hasViewController)
				{
					LySort result = buildSortFromValues(paramValues);

					if (_sortEndInterface != null)
						_sortEndInterface.endSort(this, id, result);
				}
			}

			return;
		}

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Construct a new sort with the given values.
	 *
	 * @param paramValues the parameter values
	 * @return  the new sort
	 */
	protected LySort buildSortFromValues(LyParameterValues paramValues)
	{
		LySort		    sort = null;
        LySortCriterion criterion;
		Vector<String>  fields = new Vector<String>(_maxItem);
		Vector<String>  directions = new Vector<String>(_maxItem);
		Vector<Boolean> groups = new Vector<Boolean>(_maxItem);
		String		    field, direction, choice, group;
		boolean		    increasing;
        Boolean         grouped;
		LyFieldInfo     fieldInfo;
		int			    i;

		for (i = 0; i < _maxItem; i++)
		{
			// Update fields list
			field = paramValues.getParameter(FIELD + i);
			if (field != null)
				fields.add(field);

			choice = paramValues.getParameter(DIRECTION + i);
			if (choice != null)
				directions.add(choice);
			else
				directions.add("");

            group = paramValues.getParameter(GROUP_BY + i);
            groups.add(Boolean.valueOf(group != null && group.length() > 0));
		}

		for (i = 0; i < fields.size(); i++)
		{
			field = fields.elementAt(i);
			fieldInfo = _classInfo.getFieldInfo(field, false);
			if (fieldInfo != null)
			{
				direction = directions.elementAt(i);
				increasing = direction.equals(DIR_INCREASING);

                grouped = groups.elementAt(i);

				if (sort == null)
					sort = new LySort();
				criterion = new LySortCriterion(fieldInfo, increasing);
				criterion.setGroupBy((grouped != null) && grouped.booleanValue());
                sort.addCriterion(criterion);
			}
		}

		return sort;
	} // end method buildSortFromValues

	/**
	 * Check if the group by is enabled for the given field
	 * @param fieldInfo the fieldInfo
	 * @return true if enabled
	 */
    protected boolean enableGroupBy(LyFieldInfo fieldInfo)
    {
        if (fieldInfo == null)
            return false;
        if (!(_sortEndInterface instanceof LyTableController))
            return false;

        LyTableController   tableController = (LyTableController)_sortEndInterface;
        LyTableBehavior     behavior = (LyTableBehavior)tableController.getBehavior();

        return behavior.enableGroupBy(tableController, fieldInfo);
    }

    /**
     * Sets sort of this wizard.
     * @param sort New sort.
     */
    public void setSort(LySort sort)
    {
    	if (_sort == sort)
    		return;

    	_sort = sort;
    	updateFieldsFromSort();
    }
} // end class LySortWizardController
