/**
 * LyTabAreaController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyTabAreaController.java 34904 2012-03-21 15:55:41Z bangiole $
 */

package leon.control;

import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;


/**
 * A tab area controller is associated to a tab action.
 *
 * <P>This controller is a view which gather two views controllers (subcontrollers) in the same
 * 'area'.</P>
 *
 * <P>The tabAreaController provides the following services :</P>
 *
 * <UL>
 *   <LI>Management of the list of sub-controllers (initialize, free)</LI>
 *   <LI>context management between sub-controllers (views)</LI>
 * </UL>
 */
public class LyTabAreaController extends LyMultiAreaController
	implements LyControllerContextInterface,
		LyQuestionAnswerInterface
{
	/** File Version */
	public static final String __VERSION =
		"$Id: LyTabAreaController.java 34904 2012-03-21 15:55:41Z bangiole $";

	/** Demand code : Previous step. */
	public static final String DEMAND_PREVIOUS = "_previous".intern();

	/** Demand code : Next step. */
	public static final String DEMAND_NEXT = "_next".intern();

	/** Demand code : Terminate wizard process. */
	public static final String DEMAND_TERMINATE = "_terminate".intern();

	/** Class info to which the wizard corresponds to. */
	protected LyClassInfo	   _classInfo;

	/** Index of the current controller (tab or step). */
	protected int			   _index;

	/** Context for this controller. */
	protected LyContext		   _controllerContext;

	/** Specific environment used for wizard form. */
	protected LyEnvironment    _wizardFormEnvironment;

	/**
	 * Indicates if area controllers must be freed on close (not used in wizard mode). The keys of
	 * this table are instances of LyAction. The values of this table boolean indicating if
	 * corresponding controllers must be freed when closing.
	 */
	protected Hashtable<LyAction, Boolean>		  _freeAreaControllers;

	/**
	 * tabs closed but not freed (see LY_FREE_ON_CLOSE)
	 */
	protected List<LyAreaController> 		  	  _hiddenAreaControllers;

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. The action should be an instance of LyTabAction. If an
	 *                  action is specified, the controller uses action parameters (action name,
	 *                  action specific XML view description ...)
	 * @param classInfo classInfo : useless in this controller, should be null. Kept for
	 *                  compatibility.
	 * @param context   current context (may be null if the view is not contextual to a list of
	 *                  objects).
	 */
	public LyTabAreaController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent, action, context);

		if (action.getClassInfo() != null)
			_classInfo = action.getClassInfo();
		else
			_classInfo = classInfo;

		buildSubControllers(classInfo);
		_index = 0;
	}

	/**
	 * Get controller class info if any. This method is overriden if it makes sense to return a
	 * classInfo for this controller. Default returns null.
	 *
	 * @return a classinfo if any or null.
	 */
	public LyClassInfo getClassInfo()
	{
		return _classInfo;
	}

	/**
	 * Create the list of sub-controllers (content of the compound view). The methods retreives the
	 * information from the action and create sub-controllers from the list of sub-actions.
	 *
	 * <P>Should never be used directly, invoked by the constructor.</P>
	 * @param classInfo a classInfo
	 */
	protected void buildSubControllers(LyClassInfo classInfo)
	{
		if (!(getAction() instanceof LyTabAction))
			throw new IllegalArgumentException("BF AppEngine : invalid action type "
				+ _action + " (expected action is leon.info.LyTabAction)");

        LyTabAction      tabAction = (LyTabAction)getAction();
        int              i, n = tabAction.getTabActionCount();
        LyAction         action;
        LySession        session = getSession();
        LyAreaController areaController;

        if (isWizard()
                && (n > 1))
            n = 1;

		// Check if all actions are viewable.
		for (i = 0; i < n; i++)
		{
			action = tabAction.getTabAction(i);
			if (!action.isViewable())
				throw new IllegalArgumentException("BF AppEngine : invalid action " + action
					+ " used in tab action " + _action
					+ ", only viewable actions can be used in a tab action");
		}

		for (i = 0; i < n; i++)
		{
			action = tabAction.getTabAction(i);

			if (action.hasMark(LyAction.MARK_HIDDEN))
				continue;

			if (!session.validateAction(action, this))
				continue;

			areaController = buildAreaController(action);
			if (areaController != null)
				addAreaController(areaController);
		}
	} // end method buildSubControllers

	/**
	 * Prepare component before showing. Overrides parent method to freeze hidden controllers.
	 */
	protected void initArea()
	{
		super.initArea();

		int              i, n = getAreaControllersCount();
		LyAreaController areaController;

		for (i = 1; i < n; i++)
		{
	        areaController = getAreaController(i);
	        areaController.setIdle(true);
		}

		updateCommands();
		updateSteps();
	}

	/**
	 * Create the default builder for this controller. Current implementation overrides generic
	 * method and returns an instance of leon.control.builder.LyTabAreaBuilder
	 *
	 * @return a new builder instance.
	 * @see    leon.control.LyController#getDefaultBuilder
	 * @see    leon.control.builder.LyTabAreaBuilder
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyTabAreaBuilder(this);
	}

	/**
	 * Invoked when an action occurs. At this level, the implementation manages
	 *
	 * <UL>
	 *   <LI>HREF demands. (demand.getType() = LyActionDemand.TYPE_HREF)</LI>
	 *   <LI>HREF demands. (demand.getType() = LyActionDemand.TYPE_ACTION AND demand.getSource() =
	 *     DEMAND_PREVIOUS_OBJECT or DEMAND_NEXT_OBJECT or DEMAND_SELECTION</LI>
	 *   <LI>User interaction on action bar.</LI>
	 * </UL>
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		String sourceId = demand.getSource().intern();

		if (!isWizard())
		{
			getViewManager().setWaiting(true);

			if (demand.getType() == LyActionDemand.TYPE_CLOSE_TAB)
			{
				int					index = demand.getIndex(0);
				LyAreaController	areaController = getAreaController(index);
				LyAction			oldAction = areaController.getAction();

				setLastModified(System.currentTimeMillis());
				closeAction(oldAction);

				getViewManager().setWaiting(false);
				return;
			}
			else if (sourceId.equals("_tabs"))
			{
				String tabId = demand.getValue();
				int    tabIndex = demand.getIndex(0);
				LyTabs tabs = getTabs();

				if (tabs != null)
				{
					LyTab tab;

					if ((tabId != null) && (tabId.length() > 0))
						tab = (LyTab)tabs.getComponent(tabId);
					else
						tab = tabs.getTab(tabIndex);

					if (tab != null)
					{
						setLastModified(new Date().getTime());
						tabs.setSelectedTab(tab.getId());

						// Check if a form may be submitted in one of the areas
						LyController toSubmit = getAreaController(tabIndex);

						// XXX Should go THROUGH sub controllers here and search for interface
						// XXX At this step just keep first
						if (toSubmit instanceof LyCompoundAreaController)
							toSubmit = ((LyCompoundAreaController)toSubmit).getAreaController(0);

						if (toSubmit instanceof LySubmitTab)
						{
							for (int j = 0; j < tabs.getTabCount(); j++)
							{
								tab = tabs.getTab(j);
								tab.setTargetFormId(toSubmit.getId());
							}
						}
					}

					if (_component instanceof LyPanel)
						((LyPanel)_component).setVisible(true);
				}

				// Update current controller, menu bar and message bar
				LyAreaController oldController, currentController;

				oldController = getSelectedAreaController();
				_index = tabIndex;
				currentController = getSelectedAreaController();

				currentController.setContext(_context);
				currentController.setIdle(false);

				LyTabActionBehaviorInterface behavior = (LyTabActionBehaviorInterface)getBehavior();

				if (behavior != null)
					behavior.selectedTabChanged(this, getSelectedTab(), getSelectedAreaController());

				updateMenus();
				updateMessageBar();

				if ((oldController != null) && (oldController != currentController))
					oldController.setIdle(true);

				getViewManager().setWaiting(false);
			}
			else
			{
				LyApplication application = getApplication();
				LyInfo		  info = application.getInfo(sourceId);

				if (info instanceof LyAction)
				{
					invokeAction((LyAction)info, demand);
					return;
				}
			}
		}
		else
		{
			if (sourceId.startsWith("_step"))
			{
				int index = Integer.parseInt(sourceId.substring(5));

				if ((index != _index)
						&& ((index > 0) || (index < getAreaControllersCount())))
					setAction(index);
				return;
			}
			else if (sourceId == DEMAND_PREVIOUS)
			{
				setAction(_index - 1);
				return;
			}
			else if (sourceId == DEMAND_NEXT)
			{
				setAction(_index + 1);
				return;
			}
			else if (sourceId == DEMAND_TERMINATE)
			{
				commit();
				return;
			}
			else if (sourceId == DEMAND_CANCEL)
			{
				if (!canClose(true))
					return;
				cancel();
			}
		}

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Get the LyTabs component
	 * @return the LyTabs component
	 */
	protected LyTabs getTabs()
	{
		LyArea area = (LyArea)_component;
		LyTabs tabs = (LyTabs)area.getPanel(0);

		return tabs;
	}

	/**
	 * Adds a new action to this tab action.
	 *
	 * @param areaController the area controller managing the action that must be added.
	 */
	public void addTabAction(LyAreaController areaController)
	{
		addTabAction(areaController, false);
	}

	/**
	 * Adds a new action to this tab action.
	 *
	 * @param areaController the area controller managing the action that must be added.
	 * @param deleteOnClose Delete controller when action is changed.
	 */
	protected void addTabAction(LyAreaController areaController, boolean deleteOnClose)
	{
		if (isWizard())
			return;

		// Check if action is already registered on this tab controller.
		LyAction	action = areaController.getAction();

		if (getAreaController(action) != null)
			return;

		addAreaController(areaController);

		LyArea	subArea = (LyArea)areaController.getComponent();

		if (subArea == null)
			// There was a problem when building the component
			return;

		subArea.setScrolling(LyArea.Scrolling.NO);

		LyArea area = (LyArea)_component;

		if (area != null)
		{
			LyTabs tabs = (LyTabs)area.getPanel(0);
			LyTab	tab = new LyTab();

			tab.setId("_tab" + areaController.getId());
			//tab.setId(areaController.getId());
			tab.setLabel(areaController.getName());
			tab.setImage(areaController.getImage());
			tab.setPanel(subArea);
			tabs.addTab(tab);

			areaController.initArea();

			if (_freeAreaControllers == null)
				_freeAreaControllers = new Hashtable<LyAction, Boolean>();

			_freeAreaControllers.put(action, Boolean.valueOf(deleteOnClose));
		}
	}

	/**
	 * Selects an action registered on the tab controller.<br />
	 * Selects the tab containing the view of the action.
	 *
	 * @param action the action that must be selected.
	 */
	public void selectAction(LyAction action)
	{
		if (isWizard())
			return;

		// Retrieve which tab must be selected.
		int					i, n = getAreaControllersCount(), selectedIndex = -1;
		LyAreaController	areaController;

		for (i = 0; i < n; i++)
		{
			areaController = getAreaController(i);

			if (areaController.getAction() == action)
			{
				selectedIndex = i;
				break;
			}
		}

		// Send a new action demand to simulate a tab selection.
		LyActionDemand	actionDemand = new LyActionDemand(LyActionDemand.TYPE_ACTION, "_tabs");

		actionDemand.addIndex(selectedIndex);
		actionDemand(actionDemand);
	}

	/**
     * Closes all tabs.
     */
    public void closeAll()
    {
        if (isWizard())
            return;

        // Retrieve which tab must be closed.
        int                 i, n = getAreaControllersCount();
        LyAreaController    areaController;

        for (i = n - 1; i >= 0; i--)
        {
            areaController = getAreaController(i);
            closeAction(areaController.getAction());
        }
    }

	/**
	 * Closes an action registered on the tab controller.<br />
	 * Removes the tab containing the view of the action.
	 *
	 * @param action the action that must be closed.
	 */
	public void closeAction(LyAction action)
	{
		if (isWizard())
			return;

		// Retrieve which tab must be closed.
		int					i, n = getAreaControllersCount(), selectedIndex = -1;
		LyAreaController	areaController = null;

		for (i = 0; i < n; i++)
		{
			areaController = getAreaController(i);

			if (areaController.getAction() == action)
			{
				selectedIndex = i;
				break;
			}
		}

		if (areaController == null)
			return;

		Boolean	free = ((_freeAreaControllers != null)
				? (Boolean)_freeAreaControllers.get(action) : null);

		if (selectedIndex >= 0)
		{
			LyTabs	tabs = getTabs();
			LyTab	tab = tabs.getTab(selectedIndex);

			tabs.removeTab(tab);
		}

		if ((free == null) || free.booleanValue())
			areaController.free(this);

		removeAreaController(areaController);

		if (getAreaControllersCount() == 0)
			getViewController().free();
		else if (selectedIndex == _index)
		{
			// If closed tab was selected, select next action if possible.
			int					count = getAreaControllersCount();
			LyAreaController	firstController = getAreaController(selectedIndex >= count ? count - 1 : selectedIndex);

			selectAction(firstController.getAction());
		}
	}

	/**
	 * Get the index of the selected tab
	 * @return index of the selected tab
	 */
	public int getSelectedTab()
	{
		return _index;
	}

	/**
	 * Get the selected LyAreaController
	 * @return the selected LyAreaController
	 */
	public LyAreaController getSelectedAreaController()
	{
		if (_index < getAreaControllersCount())
			return getAreaController(_index);

		return null;
	}

	/**
	 * Gets help file associated with this controller.
	 *
	 * @return the help file corresponding to the action.
	 * If no help file is associated to the tab action, retrieve the help file associated to
	 * selected action in this tab controller.
	 */
	public String getHelpFile()
	{
		String	helpFile = super.getHelpFile();

		if (helpFile == null)
		{
			LyAreaController	areaController = getSelectedAreaController();

			if (areaController != null)
				return areaController.getHelpFile();
		}

		return helpFile;
	}

	/**
	 * Returns the specific environment used for wizard form. This method builds this specific
	 * environment if it was not already created.
	 *
	 * @return the <code>_wizardFormEnvironment</code> attribute.
	 */
	public LyEnvironment getWizardFormEnvironment()
	{
		if (_wizardFormEnvironment == null)
		{
            LyTabActionBehaviorInterface tabBehavior = (LyTabActionBehaviorInterface)getBehavior();

            if (tabBehavior == null)
                return null;

             // Put specific resources on this environment
			String[] colors = tabBehavior.getHrefColors(this);
			String   image = tabBehavior.getImage(this, 0);

			_wizardFormEnvironment = new LyEnvironment(getEnvironment());
			_wizardFormEnvironment.putImage("LY_AREA_BACKGROUND_IMAGE", image);
			_wizardFormEnvironment.putFont("LY_HREF_FONT", "Arial-bold-16");
			_wizardFormEnvironment.putColor("LY_HREF_COLOR", colors[0]);
			_wizardFormEnvironment.putColor("LY_HREF_HIGHLIGHT", colors[1]);
			_wizardFormEnvironment.putColor("LY_HREF_VISITED", colors[2]);
			_wizardFormEnvironment.putColor("LY_HREF_DISABLE", colors[3]);
		}

		return _wizardFormEnvironment;
	}

	/**
	 * Accessor to set another area controller which will receive contextual notification when
	 * something occur on this area controller.
	 *
	 * <P>Method is overriden to take the internal sub-controller structure into account.</P>
	 *
	 * @param controller the contextual controller, or null if no client defined
	 */
	public void setContextClient(LyAreaController controller)
	{
		LyTabAction tabAction = (LyTabAction)getAction();
		short	    clientContextIndex = tabAction.getClientContextIndex();

		if (clientContextIndex == LyTabAction.CONTEXT_NONE)
			return;

		if (clientContextIndex == LyTabAction.CONTEXT_ALL)
			super.setContextClient(controller);
		else
		{
			LyAreaController areaController = getAreaController(clientContextIndex);

			areaController.setContextClient(controller);
		}
	}

	/**
	 * Reset the entire view : everything is rebuilt.
	 *
	 * @param full Boolean indicating if a full reset (including reset of datas) is requiered or not
	 *             (only reset graphical area).
	 */
	public void reset(boolean full)
	{
		updateTitleBar();
		updateMessages();
		updateCommands();
		updateSteps();

		super.reset(full);
	}

	/**
	 * This method is called by the view controller when this area is shown to update its content.
	 * Default implementation updates selected tab.
	 * @return true if update was successfull
	 */
	public boolean update()
	{
		LyAreaController current = null;

		if ((_index >= 0) && (_index < getAreaControllersCount()))
			current = getAreaController(_index);

		return ((current == null) ? false : current.update());
	}

	/**
	 * Change action at given index.
	 * @param index the index
	 */
	protected void setAction(int index)
	{
		if ((index < 0)
				|| (index > getAreaControllersCount())
				|| (index == _index))
			return;

		// Replace area controller
		LyTabActionBehaviorInterface tabBehavior = (LyTabActionBehaviorInterface)getBehavior();
		LyAreaController			 oldAreaController = getAreaController(_index);
		boolean						 next = (index > _index);

		if (next
				&& (validate(oldAreaController) == LyFormInterface.STATUS_KO))
			// Can't go forward if we have errors
			return;

		_index = index;

		LyAreaController newAreaController = null;
		LyAction		 action;

		if ((_index >= 0) && (_index < getAreaControllersCount()))
			newAreaController = getAreaController(_index);

		if (next)
		{
			action = tabBehavior.getNextAction(this, _index);
			if ((action == null) || !action.isViewable())
				throw new IllegalStateException("BF AppEngine : invalid action parameter " + action
					+ ", given action is not viewable");

			if ((newAreaController != null)
					&& (newAreaController.getAction() != action))
			{
				LyAreaController areaController;
				int				 i, n = getAreaControllersCount();

				for (i = n - 1; i >= _index; i--)
				{
					areaController = getAreaController(i);
					areaController.free();
					removeAreaController(areaController);
				}

				newAreaController = null;
			}
		}
		else
			action = newAreaController.getAction();

		if ((newAreaController == null)
				|| newAreaController.isFreed())
		{
			newAreaController = buildAreaController(action);
			newAreaController.setParent(this);
			newAreaController.setViewController(_viewController);
			addAreaController(newAreaController);
		}
		else
			newAreaController.setIdle(false);

		// Replace view
		LyArea newArea = (LyArea)newAreaController.getComponent();
		LyArea oldArea = (LyArea)oldAreaController.getComponent();
		LyArea parentArea = (LyArea)oldArea.getParent();
		LyArea grandParentArea = (LyArea)parentArea.getParent();
		LyArea newParentArea = new LyArea();

		grandParentArea.removePanel(parentArea);
		newParentArea.addPanel(newArea);
		newArea.setCommands(null);
		grandParentArea.addPanel(newParentArea);

		// This method must be called after the complete initialization of
		// the newAreaController (include the context).
		newAreaController.initArea();

		// Clear old area controller
		//oldAreaController.clearViewCache();
		oldAreaController.setIdle(true);

		// Update view
		updateMenus();
		updateMessageBar();
		updateCommands();
		updateSteps();

		// Update the last modification time.
		setLastModified(new Date().getTime());
	} // end method setAction

	/**
	 * Builds area controller corresponding to the given action.
	 * @param action an action
	 * @return the resulting area controller
	 */
	protected LyAreaController buildAreaController(LyAction action)
	{
		_buildInternalSubControllers = true;

        LyAreaController              areaController = null;
        LyTabActionBehaviorInterface  tabBehavior = (LyTabActionBehaviorInterface)getBehavior();

        if (tabBehavior != null)
        {
            areaController = tabBehavior.buildAreaController(this, action);
            if (areaController != null)
            {
                _buildInternalSubControllers = false;
                return areaController;
            }
        }

		if (!isWizard())
			areaController = (LyAreaController)invokeAction(action, getClassInfo(), getContext(), false);
		else
			areaController = (LyAreaController)invokeAction(action, (LyActionDemand)null);

		_buildInternalSubControllers = false;
		return areaController;
	}

	/**
	 * Invoke given actions with specified classInfo and context. This method is overridden to set
	 * show parameter to false.
	 * @param action the action to invoke
	 * @param classInfo the related classinfo
	 * @param objects the context
	 * @param show true to show the controller
	 * @return the resulting controller
	 */
	protected LyController invokeAction(LyAction action, LyClassInfo classInfo,
		LyObjectList objects, boolean show)
	{
		// Don't show, the resulting controller will be integrated in a tab or in wizard panel
		return super.invokeAction(action, classInfo, objects, false);
	}

	/**
	 * @see leon.control.LyMultiAreaController#allowSubAreas()
	 */
	protected boolean allowSubAreas()
	{
		return !isWizard();
	}

	/**
	 * Updates command bar for wizard mode.
	 */
	public void updateCommands()
	{
		if (_component == null)
			return;

		LyCommands commands = (LyCommands)_component.getComponent("_commands");

		if (commands == null)
			return;

		LyCommand command = null;

		command = (LyCommand)commands.getComponent(DEMAND_PREVIOUS);
		command.setEnabled(_index > 0);

		LyTabActionBehaviorInterface tabBehavior = (LyTabActionBehaviorInterface)getBehavior();

		command = (LyCommand)commands.getComponent(DEMAND_NEXT);
		command.setEnabled(tabBehavior.enableNext(this, _index));

		command = (LyCommand)commands.getComponent(DEMAND_TERMINATE);
		command.setEnabled(tabBehavior.enableTerminate(this, _index));
	}

	/**
	 * Updates steps for wizard mode.
	 */
	protected void updateSteps()
	{
		if (_component == null)
			return;

		LyForm labelsForm = (LyForm)_component.getComponent("_labels");

		if (labelsForm == null)
			return;

		LyTabActionBehaviorInterface tabBehavior = (LyTabActionBehaviorInterface)getBehavior();
		String[]					 labels = tabBehavior.getStepLabels(this);
		int							 i, n = ((labels == null) ? 0 : labels.length);
		int							 nl = labelsForm.getFormFieldCount();
		LyFormField					 field;
		LyHRef						 href;

		for (i = 0; i < n; i++)
		{
			if (i >= nl)
			{
				field = new LyFormField("_step" + i);
				href = new LyHRef();
				field.setFieldItem(href);
				labelsForm.addFormField(field);
			}
			else
			{
				field = labelsForm.getFormField(i);
				href = (LyHRef)field.getFieldItem();
				field.setVisible(true);
			}

			href.setLabel(labels[i]);
			href.setSensitive(i <= _index);
			href.setVisited(i < _index);
		}

		for (; i < nl; i++)
		{
			field = labelsForm.getFormField(i);
			field.setVisible(false);
		}
	} // end method updateSteps

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyTabActionBehaviorInterface.class;
	}

	/**
	 * Gets the controller context.
	 *
	 * @return The controller context.
	 */
	public LyContext getControllerContext()
	{
		if (!isWizard())
			return null;

		if ((_controllerContext != null)
				&& !_controllerContext.isEnded())
			return _controllerContext;

		LySession		 session = getSession();
		LyContextManager contextManager = session.getContextManager();

		_controllerContext = contextManager.newContext(session, LyContext.Type.USER_CANCELLABLE);
		_controllerContext.setName(getTitle());
		return _controllerContext;
	} // end method getControllerContext

	/**
	 * Validate given area controller.
	 * @param areaController the areaController to validate
	 * @return the result of validation
	 * @see LyFormInterface
	 */
	protected short validate(LyAreaController areaController)
	{
		if (!(areaController instanceof LyFormInterface))
			return LyFormInterface.STATUS_IGNORE;

	    // Apply form
        if (areaController instanceof LySetController)
            ((LySetController)areaController).apply();

		return ((LyFormInterface)areaController).control();
	}

	/**
	 * Method called when all modifications are committed.
	 */
	public void commit()
	{
		LyTabActionBehaviorInterface tabBehavior = (LyTabActionBehaviorInterface)getBehavior();

		if (!tabBehavior.beforeCommit(this))
			return;

		if (validate(getAreaController(_index)) == LyFormInterface.STATUS_KO)
				return;

		// Commit modifications
		if ((_controllerContext != null)
				&& (_controllerContext.getErrorLevel() != LyObject.LEVEL_ERROR))
		{
			LyApplication    application = getApplication();
			LyEnvironment    environment = getEnvironment();
            LyContext.Status status = _controllerContext.commit();

			if (status == LyContext.Status.ROLLBACKED)
			{
				String cause = _controllerContext.getErrorMessage();
				String message;

				if (cause == null)
					cause = application.getError();
				message = environment.getMessage("LY_ER_COMMIT_CONTEXT", cause);
				showError(message);
			}
			else if (status != LyContext.Status.ERROR)
				_controllerContext = null;
		}

		tabBehavior.afterCommit(this);

		// Check controller context after committing
		if ((_controllerContext != null)
				&& ((_controllerContext.getStatus() == LyContext.Status.ERROR)
					|| (_controllerContext.getErrorLevel() == LyObject.LEVEL_ERROR)))
			// Commit failed, there are problems
			showErrors();
		else
			free();
	} // end method commit


	/**
	 * Method called when the form is cancelled.
	 */
	public void cancel()
	{
		LyTabActionBehaviorInterface tabBehavior = (LyTabActionBehaviorInterface)getBehavior();

		tabBehavior.cancel(this);

		LyAreaController areaController;
		int				 i, n = getAreaControllersCount();

		for (i = 0; i < n; i++)
		{
			areaController = getAreaController(i);
			if (areaController instanceof LyFormInterface)
				((LyFormInterface)areaController).cancel();
		}

		// Rollback modifications
		if (_controllerContext != null)
		{
			_controllerContext.rollback();
			_controllerContext = null;
		}
	}

	/**
	 * Free this controller. Overrides method from LyMultiAreaController :<BR>
	 * If a specific wizard environment has been created, it is freed.
	 *
	 * @see LyMultiAreaController#free(LyController)
	 */
	public void free(LyController controller)
	{
		super.free(controller);

		if (_wizardFormEnvironment != null)
		{
			_wizardFormEnvironment.free();
			_wizardFormEnvironment = null;
		}
	}

	/**
	 * Return list of selected objects This method is overriden depending on the nature of the
	 * controller. Default returns null
	 *
	 * @return the list of selected objects or null if the selection is empty.
	 */
	public LyObjectList getSelection()
	{
		if ((_areaControllers.size() == 0) && isWizard())
		{
			// If the action is a wizard action and first area controller is
			// not built yet, try to retrieve selection from parent controller.
			LyController parent = _parent;

			while ((parent != null) && !(parent instanceof LyAreaController))
				parent = parent.getParent();

			if (parent instanceof LyAreaController)
			{
				LyAreaController parentController = (LyAreaController)parent;

				return parentController.getSelection();
			}
		}

		return getAreaController(0).getSelection();
	}

	/**
	 * Checks if the managed action is a wizard action.
	 *
	 * @return True if the controlled action is a wizard action, false otherwise.
	 */
	public boolean isWizard()
	{
		return (_action instanceof LyTabAction)
			&& (((LyTabAction)_action).getType() == LyTabAction.Type.WIZARD);
	}

	/**
	 * Check if the area controller may be closed. This method recursively check canClose on
	 * sub-controllers
	 *
	 * @param  showMessage indicates if the controller can show an warning message or not
	 * @return return the status of sub-controllers
	 */
	public boolean canClose(boolean showMessage)
	{
		boolean canClose = super.canClose(false);

		if (!canClose && showMessage)
		{
			// Modifications were made (even erroneous), before changing to an other object
			// ask first to the user if he wants to abort his edition.
			LyEnvironment environment = getEnvironment();
			String		  message;
			LyViewManager viewManager = getViewManager();

			message = environment.getMessage("LY_CONFIRM_TAB_ABORT");

			viewManager.askQuestion(getView(), message, this, null);
		}
		return canClose;
	}

	/**
	 * Answer to a askQuestion request. Current implementation proceed the setObject method after
	 * user confirmation that changes may be lost.
	 *
	 * @param result true if the user has clicked on Yes or Ok
	 * @param clientData data specified in the askQuestion method
	 * @see   leon.view.setinterface.LyQuestionAnswerInterface#questionAnswer
	 */
	public void questionAnswer(boolean result, Object clientData)
	{
		if (result)
		{
			// Cancel current edition
			cancel();
			free();
		}
	}

	/**
	 * Notify a visibility change of the parent view.
	 * @param visible the visible status
	 * @see #setVisible(boolean)
	 */
	protected void visibilityChanged(boolean visible)
	{
	    if (!isWizard())
	    {
    		// close tabs : FFT IPBOSS 00253
    		if (!visible)
    		{
    			while (getAreaControllersCount() > 0)
    			{
    				LyAreaController	areaController = getAreaController(0);
    				LyAction			oldAction = areaController.getAction();

    				closeAction(oldAction);
    			}
    		}

    		setLastModified(System.currentTimeMillis());
	    }

		super.visibilityChanged(visible);
	}

	/**
	 * Append a new sub-controllers at the children list This method should not be used directly,
	 * use action description instead (LyMultiAction).
	 *
	 * @param areaController the new controller we want to add.
	 */
	public void addAreaController(LyAreaController areaController)
	{
		super.addAreaController(areaController);

		if ((_hiddenAreaControllers != null) && (_hiddenAreaControllers.contains(areaController)))
		{
			_hiddenAreaControllers.remove(areaController);
		}
	}

	/**
	 * Remove a sub-controllers at the children list This method should not be used directly, use
	 * action description instead (LyMultiAction).
	 *
	 * @param areaController the new controller we want to remove.
	 */
	public void removeAreaController(LyAreaController areaController)
	{
		super.removeAreaController(areaController);

		if (!areaController.isFreed())
		{
			if (_hiddenAreaControllers == null)
				_hiddenAreaControllers = new ArrayList<LyAreaController>();
			_hiddenAreaControllers.add(areaController);
		}
	}

	/**
	 * Accessor to the sub-controllers with the specified action.
	 *
	 * @param  action specified action whose controller must be returned.
	 * @param searchHidden search also in hidden tabs (kept for LY_FREE_ON_CLOSE)
	 * @return the reference to the specified controller, or null if no controller was found for
	 *         specified action.
	 */
	public LyAreaController getAreaController(LyAction action, boolean searchHidden)
	{
		LyAreaController areaController = super.getAreaController(action);

		// If the controller was found, not need to search in hidden controllers.
		if (areaController != null)
			return areaController;
		// may be it is in hidden tab ?
		if ((searchHidden) && (areaController == null) && (_hiddenAreaControllers != null))
		{
			Iterator<LyAreaController> iter = _hiddenAreaControllers.iterator();

			while(iter.hasNext())
			{
				areaController = iter.next();

				if (areaController.getAction() == action)
					return areaController;
			}
		}
		// The controller was not found
		return null;
	}

	/**
	 * Internal utility method to reuse existing hidden tabs from another controller (beeing closed)
	 * Used for LY_FREE_ON_CLOSE mode
	 * @param old old tab area being freed.
	 */
	protected void reuseHiddenAreaControllers(LyTabAreaController old)
	{
		if ((old != null) && (old._hiddenAreaControllers != null))
		{
			if (_hiddenAreaControllers == null)
				_hiddenAreaControllers = old._hiddenAreaControllers;
			else
				_hiddenAreaControllers.addAll(old._hiddenAreaControllers);
			old._hiddenAreaControllers = null;
		}

	}

	   /**
     * Set current controller context. This method set the context of the sub AreaController Two
     * possibilities: One of the subareas is a context for the other one, so this method set its
     * context. Both Areas are &quot;equal&quot; so the context is set for both.
     *
     * @param context current context or null if none is specified
     * @return true if successfull
     * @see   LyContextInterface#setContext
     */
    public boolean setContext(LyWorkSpace context)
    {
        if (!isInitialized())
            return false;
        if ((_action != null)
                && _action.hasMark(LyAction.MARK_NO_CONTEXT))
            return false;

        boolean          resetted = changeContext(context);
        LyAreaController areaController = null;
        int              i, n = getAreaControllersCount();

        for (i = 0; i < n; i++)
        {
            areaController = getAreaController(i);
            if ((areaController != null)
                    && !areaController.isIdle())
                resetted |= areaController.setContext(context);
        }

        return resetted;
    }
} // end class LyTabAreaController
