/**
 * LyTableController.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved. Current version: $Id:
 * LyTableController.java,v 6.55 2009/07/20 12:25:58 jls Exp $
 */

package leon.control;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;

import java.util.*;


/**
 * This class implements a table controller. It's the standard implementation for a table It is
 * associated with a builder (LyTableBuilder) and a behavior that implements
 * LyTableBehaviorInterface. Usually, this behavior is implemented by an instance of LyTableBehavior
 *
 * @see  LySimpleTableBuilder
 * @see  leon.app.behaviorinterface.LyTableBehaviorInterface
 */
public class LyTableController extends LySimpleTableController
	implements LyContextInterface
{
	/** File Version */
	public static final String __VERSION = "$Id: LyTableController.java 41549 2012-11-16 13:32:03Z bangiole $";

	/** Define the max objects in the context to build a context filter */
	public static int MAX_OBJECTS_COUNT_FOR_CONTEXT_FILTER = 10;

	/** Boolean used to avoid calls to source.getSize on selections */
	protected boolean _selection;

	/** Store the old filter from context */
	protected LyFilter _oldFilterFromContext;

	/**
	 * Constructor.
	 *
	 * @param  session    current session associated to the controller. The controller automatically
	 *                    registers on the specified session.
	 * @param  parent     parent controller (or null for 'root' controller). Controller hierarchy
	 *                    indicates a child controller (for example a dialog created from a main
	 *                    view)
	 * @param  action     related action. If an action is specified, the controller uses action
	 *                    parameters (action name, action specific XML view description ...)
	 * @param  classInfo  class info for objects displayed in the list.
	 * @param  context    current context (may be null if the view is not contextual to a list of
	 *                    objects).
	 */
	public LyTableController(LySession session, LyController parent, LyAction action, LyClassInfo classInfo,
		LyObjectList context)
	{
		super(session, parent, action, classInfo, context);

		_selection = false;

		if (_classInfo == null)
			_classInfo = classInfo;

		// classInfo could be null when using LyMultiClassImportController
		if (_classInfo != null)
		{
			// Check if we have columns (fields) in the table
			LyInfoList<LyFieldInfo> fields = getFields();

			if ((fields == null) || (fields.getSize() <= 0))
			{
				getEnvironment().logWarning("attempt to show table for class "
					+ _classInfo.getId()
					+ " which has no fields with the 'table=true' mark. Using all fields");
				_fieldMarks = null; // SHOW ALL !

				fields = getFields();
				if ((fields == null) || (fields.getSize() <= 0))
					getEnvironment().logWarning("BF AppEngine : can't find any field for table "
						+ _classInfo.getId());
			}

			setName(getAction().getName());
		}
	}

	/**
	 * Reset component.
	 *
	 * @param  full  Boolean indicating if a full reset (including reset of datas) is requiered or
	 *               not (only reset graphical area).
	 */
	public void reset(boolean full)
	{
		if (full)
		{
			// Calculate workspaces containing datas
			LySession session = getSession();

			if (_workSpace != null)
			{
				try
				{
					_workSpace.free(this);
				}
				catch (Exception e)
				{
					getEnvironment().writeException(e);
				}

				_workSpace = null;
			}

			if (_sort == null)
			{
				_sort = getDefaultSort();
				updateFields(null);
			}

			LyWorkSpace				 context = getContext();
			LyTableBehaviorInterface behavior = (LyTableBehaviorInterface)getBehavior();
			LyWorkSpace				 content = behavior.getContent(this, context);
			LyInfoList<LyFieldInfo>  fieldsToLoad = behavior.getLoadedFieldInfos(this);

			if (content != null)
				context = content;

			int    limit = getLimit(); // No limit

			if (limit > 0)
			{
				if (context != null)
				{
					LyFilter filter = getFilterFromContext(context);

					if (_oldFilterFromContext != null)
					{
						if (_classFilter == _oldFilterFromContext)
							_classFilter = null;
						else if (_classFilter != null)
							_classFilter.removeElement(_oldFilterFromContext);
					}

					if (filter != null)
					{
						if (_classFilter == null)
							_classFilter = filter;
						else
						{
						    _classFilter = _classFilter.cloneFilter();
							_classFilter.addElement(filter);
						}

						context = null;
					}

					_oldFilterFromContext = filter;
				}
			}

			if (((_filter == null) || !_filterActive)
					&& _openFilterIfEmpty)
			{
				_workSpace = new LySimpleWorkSpace(session, true);
				_workSpace.addListener(this);
				_maxRowCount = limit;
			}

			openFilter();

			if (_workSpace == null)
			{
				if ((limit > 0) && (_maxRowCount <= 0))
					_maxRowCount = limit;

				if (context == null)
				{
					LyFilter filter = _classFilter;

					if (_openFilterIfEmpty
							&& _filterActive
							&& (_filter != null))
					{
						if (_classFilter == null)
							filter = _filter;
						else
						{
							filter = new LyFilter();
							filter.addElement(_classFilter);
							filter.addElement(_filter);
						}
					}

					if (limit > 0)
						_workSpace = new LyPageWorkSpace(session, _classInfo, fieldsToLoad,
							filter, _sort, _maxRowCount, _dynamic, true);
					else
						_workSpace = new LyClassWorkSpace(session, _classInfo, fieldsToLoad,
							filter, _sort, _dynamic, true);


					_initialWorkSpace = _workSpace;

					if (_filterActive && !_openFilterIfEmpty)
					{
						if (limit > 0)
						{
							if (_classFilter != null)
							{
								// create a new filter that has the conditions of
								// the class filter and the current filter
								LyFilter newFilter = new LyFilter(LyFilter.Type.AND);

								newFilter.addElement(_classFilter);
								newFilter.addElement(_filter);
								// set this new filter to the workspace
								((LyPageWorkSpace)_workSpace).setFilter(newFilter);
							}
							else
								((LyPageWorkSpace)_workSpace).setFilter(_filter);
						}
						else
						{
							_workSpace = new LySubWorkSpace(session, _initialWorkSpace, _filter, _dynamic);
							_workSpace.setSort(_sort);
						}
					}

					if (!_dynamic)
					{
						_workSpace = new LySimpleWorkSpace(session, _workSpace, true);
						_workSpace.setSort(_sort);
						_initialWorkSpace = _workSpace;
					}
				}
				else
				{
				    LyObjectList contextList = context.getList(session, _classInfo, _dynamic);

				    if ((contextList != null) && !(contextList instanceof LyWorkSpace))
				        contextList = new LySimpleWorkSpace(session, contextList, _dynamic);

					_workSpace = (LyWorkSpace)contextList;

					if (_workSpace == null)
					{
						LyInfoList<LyClassInfo> classes = context.getClasses();

						if ((classes != null)
								&& (classes.getSize() > 0)
								&& (_action.getMin() < 0))
						{
							// No route : take all
							if (limit > 0)
								_workSpace = new LyPageWorkSpace(session, _classInfo, fieldsToLoad,
									_classFilter, _sort, _maxRowCount, _dynamic, true);
							else
								_workSpace = new LyClassWorkSpace(session, _classInfo, fieldsToLoad,
									_classFilter, _sort, _dynamic, true);
						}
						else
							_workSpace = new LySimpleWorkSpace(session, _dynamic); // XXX
					}
					else
					{
						if (_workSpace.equals(context))
						{
							// Look for sub-objects of the same class
							LyInfoList<LyRoute>  routes = _classInfo.getRoutes(_classInfo);
							int                  r, nr = (routes == null ? 0 : routes.getSize());
							LyRoute              route;
							LyMultiWorkSpace     mw = null;
							LyWorkSpace          temp = null;

							for (r = 0; r < nr; r++)
							{
								route = routes.getInfo(r);
								if (route.isGenerated())
									continue;

								temp = (LyWorkSpace)context.getList(session, _classInfo, route, _dynamic, false, true);
								if (r == 0)
									_workSpace = temp;
								else
								{
									if (mw == null)
									{
										mw = new LyMultiWorkSpace(session, _dynamic);
										mw.addList(_workSpace);
									}

									mw.addList(temp);
								}
							}
							if (mw != null)
								_workSpace = mw;
						}
						if (_classFilter != null)
							_workSpace = new LySubWorkSpace(session, _workSpace, _classFilter, _dynamic);
					}

					if (!_dynamic)
					{
						_workSpace = new LySimpleWorkSpace(session, _workSpace, true);
						_workSpace.setSort(_sort);
					}

					_initialWorkSpace = _workSpace;

					if (_filterActive)
						_workSpace = new LySubWorkSpace(session, _initialWorkSpace, _filter, true);

					// sort after applying filter : perf. improved.
					_workSpace.setSort(_sort);
				}

				_workSpace.addListener(this);
			}
		}

		super.reset(full);
	} // end method reset

	/**
	 * Invoked when a selection occurs. This method handles: select all, unselect all and objects by
	 * objects selection
	 *
	 * @param  demand  selection event created from the view. User interaction.
	 */
	public void selectionDemand(LySelectionDemand demand)
	{
		_selection = true;
		super.selectionDemand(demand);
		_selection = false;
	}

	/**
	 * Select an object in the table.
	 *
	 * @param  object        selected or unselected object
	 * @param  select        new selection status
	 * @param  updateView    indicates if the view (tools, messages...) must be updated.
	 * @param  showSelected  indicates wether the view should scroll to the new selection status
	 */
	public void selectObject(LyObject object, boolean select, boolean updateView, boolean showSelected)
	{
		boolean selection = _selection;

		_selection = true;
		super.selectObject(object, select, updateView, showSelected);
		_selection = selection;
	}

	/**
	 * Select all given objects in the table.
	 *
	 * @param  objects  Objects to select/deselect
	 * @param  select   new selection status
	 */
	public void selectObjects(LyObjectList objects, boolean select)
	{
		boolean selection = _selection;

		_selection = true;
		super.selectObjects(objects, select);
		_selection = selection;
	}

	/**
	 * Indicates whether previous / next tools must be added to the top toolbar.
	 *
	 * @return true if the _limit parameter is set on this table, false otherwise
	 */
	public boolean isDisplayNextPreviousTools()
	{
		return (_action.getParameter("_limit") != null);
	}

	/**
	 * updates the status of Tool and menu items.
	 */
	protected void updateTools()
	{
		if (_component == null)
			return;

		// If table is Idle, no need to refresh
		if (isIdle())
			return;

		boolean   hasMenu = hasMenu();
		LyMenuBar menuBar = null;

		if (hasMenu)
		{
			// Update "Select all" and "Unselect all" items
			LyView view = getView();
			LyItem item = null;
			int    ns = ((_selections == null) ? 0 : _selections.getSize());
			int    wssize = ((_workSpace == null) ? 0 : _workSpace.getSize());

            menuBar = (view instanceof LyFrame ? ((LyFrame)view).getMenuBar() : null);

			if (menuBar != null)
			{
				item = menuBar.findItem(DEMAND_SELECT_ALL, getId());
				if (item != null)
					item.setEnabled(ns != wssize);

				item = menuBar.findItem(DEMAND_UNSELECT_ALL, getId());
				if (item != null)
					item.setEnabled(ns != 0);
			}

			// Update "Print list" and "Search" items
			boolean empty = ((_workSpace == null) || (_workSpace.getSize() == 0));

			if (hasMenu && (menuBar != null))
			{
				item = menuBar.findItem(DEMAND_PRINT_TABLE, getId());
				if (item != null)
					item.setEnabled(!empty);

				item = menuBar.findItem(LyFindController.DEMAND_FIND, getId());
				if (item != null)
					item.setEnabled(!empty);
			}
		}
		// Update ToolBar
		LyArea	  area = (LyArea)_component;
		LyToolBar toolBar = (LyToolBar)area.getToolBar();

		if (toolBar == null)
			return;

		// get all the actions of the toolbar
		LyAction			 rootAction = getAction();
		LyInfoList<LyAction> actions = new LyInfoList<LyAction>();

		if (rootAction != null)
			actions.addList(rootAction.getActions()); // add sub-actions of the root action

		LyClassInfo			    classInfo = getClassInfo();
		LyInfoList<LyClassInfo> subClasses = classInfo.getSubClasses(true);

		if (subClasses == null)
			subClasses = new LyInfoList<LyClassInfo>();

		subClasses.addInfo(classInfo, 0);

		LyClassInfo subClass;

		// Add actions of all subclasses.
		for (int i = 0; i < subClasses.getSize(); i++)
		{
			subClass = subClasses.getInfo(i);
			actions.addList(subClass.getActions());
		}

		// Update toolbar and corresponding menu items
		LyAction action;
		boolean  enabled;
		int		 i, n = actions.getSize();

		for (i = 0; i < n; i++)
		{
			action = actions.getInfo(i);
			if (action.hasMark(LyAction.MARK_HIDDEN))
				continue;

			enabled = enableAction(action);

			if (menuBar != null)
				updateMenu(menuBar, classInfo, subClasses, action, enabled);

			updateTool(toolBar, classInfo, subClasses, action, enabled);
		}

		String limitParameter = (String)_action.getParameter("_limit");

		if (limitParameter != null)
		{
			LyTool previousTool = (LyTool)toolBar.getComponent(DEMAND_PREVIOUS_PAGE);
			LyTool nextTool = (LyTool)toolBar.getComponent(DEMAND_NEXT_PAGE);

			// Update the number of objects per page
			LyFormField viewLimit = (LyFormField)toolBar.getComponent(getId() + "_viewLimit");

			if (viewLimit != null)
			{
				LyNumber number = (LyNumber)viewLimit.getFieldItem();

				number.setLabel(_maxRowCount + "");
			}

			if ((_workSpace instanceof LyPageWorkSpace)
					&& !_selection
					&& allowDisablePreviousNextTools())
			{
				int page = ((LyPageWorkSpace)_workSpace).getPage();

				if (previousTool != null)
					previousTool.setEnabled(page > 0);

				if (nextTool != null)
				{
					if (!getAction().hasMark(LyAction.MARK_NO_MESSAGES))
					{
						int   limit = LyTools.intFromString(limitParameter);

						// We change dynamically the limit from the field item
						if (viewLimit != null)
						{
							LyNumber number = (LyNumber)viewLimit.getFieldItem();

							limit = Integer.parseInt(number.getValue());
						}

						int count = getTotalCount();
						int pageCount = ((count + limit - 1) / limit);

						nextTool.setEnabled(page < (pageCount - 1));
					}
					else
					    nextTool.setEnabled(true);
				}

				return;
			}

			// Structis / FFT00047
			if (!(_workSpace instanceof LyPageWorkSpace))
			{
				if (previousTool != null)
					previousTool.setEnabled(_currentPage > 0);

				int count = getTotalCount();
				int pageCount = ((count + _maxRowCount - 1) / _maxRowCount);

				if (nextTool != null)
					nextTool.setEnabled(_currentPage < (pageCount - 1));
			}
		}
	} // end method updateTools

	/**
	 * Indicates whether previous / next page tools may be disabled.
	 *
	 * @return Default implementation returns true.
	 */
	protected boolean allowDisablePreviousNextTools()
	{
		return true;
	}

	/**
	 * Updates status of menus when the class is a super class.
	 *
	 * @param  menuBar     the menubar that contains the menus to update.
	 * @param  classInfo   the super class
	 * @param  subClasses  the subclasses of the super class
	 * @param  action      the action for which the menus must be updated.
	 * @param  enabled     the status of the menu.
	 */
	protected void updateMenu(LyMenuBar menuBar, LyClassInfo classInfo, LyInfoList<LyClassInfo> subClasses,
		LyAction action, boolean enabled)
	{
		// If table is Idle, no need to refresh
		if (isIdle())
			return;

		LyItem	    item;
		LyClassInfo subClass;

		// Item for superclass
		item = menuBar.findItem(action.getId(), getId());
		if (item != null)
			item.setEnabled(enabled);

		// Items for subclasses
		LyInfoList<LyClassInfo> selectedClasses = null;

		if (action.getMin() > 0)
			selectedClasses = ((_selections != null) ? _selections.getClasses() : null);

		boolean     subMenuEnabled = enabled;
		LyClassInfo selectedClass;
		int		    i, n = subClasses.getSize();
		int		    j, nc = ((selectedClasses == null) ? 0 : selectedClasses.getSize());

		for (i = 0; i < n; i++)
		{
			subClass = subClasses.getInfo(i);
			if (!subClass.contains(action))
				continue;

			item = menuBar.findItem(action.getId() + LyController.ID_SEP
					+ subClass.getId(), getId());

			if (item != null)
			{
				if (nc > 0)
				{
					if (subClass.isVirtual())
						subClass = subClass.getRealClass();

					if (subClass.isAbstract())
					{
						subMenuEnabled = enabled;
						for (j = 0; j < nc; j++)
						{
							selectedClass = selectedClasses.getInfo(j);
							subMenuEnabled = subMenuEnabled
									&& (selectedClass.inheritsFrom(subClass));
						}
					}
					else
					{
						subMenuEnabled = enabled && (nc == 1);
						selectedClass = selectedClasses.getInfo(0);
						subMenuEnabled = subMenuEnabled && (selectedClass == subClass);
					}
				}
				else
				{
					if (action.getMin() > 0)
						subMenuEnabled = false;
				}

				item.setEnabled(subMenuEnabled);
			}
		}
	} // end method updateMenu

	/**
	 * Updates status of tools when the class is a super class.
	 *
	 * @param  toolBar  the toolbar that contains the tools to update.
	 * @param  classInfo   the super class
	 * @param  subClasses  the subclasses of the super class
	 * @param  action      the action for which the tools must be updated.
	 * @param  enabled     the status of the tool.
	 */
	protected void updateTool(LyToolBar toolBar, LyClassInfo classInfo, LyInfoList<LyClassInfo> subClasses,
		LyAction action, boolean enabled)
	{
		if (!action.hasMark(LyAction.MARK_TOOL))
			return;

		// If table is Idle, no need to refresh
		if (isIdle())
			return;

		LyComponent comp;
		LyTool	    tool;

		// Tool for superclass
		comp = toolBar.getComponent(action.getId());

		if ((comp != null) && (comp instanceof LyTool))
		{
			tool = (LyTool)comp;
			tool.setEnabled(enabled);
		}

		// Tools for subclasses
		LyInfoList<LyClassInfo> selectedClasses = null;
		LyClassInfo			    subClass;

		if (action.getMin() > 0)
			selectedClasses = ((_selections != null) ? _selections.getClasses() : null);

		boolean     subToolEnabled = enabled;
		LyClassInfo selectedClass;
		int		    j, n;
		int		    i, ns;

		for (j = 0, n = subClasses.getSize(); j < n; j++)
		{
			subClass = subClasses.getInfo(j);
			if (!subClass.contains(action))
				continue;

			comp = toolBar.getComponent(action.getId() + LyController.ID_SEP + subClass.getId());
			if (comp == null)
				continue;

			if (comp instanceof LyTool)
			{
				if ((selectedClasses != null) && (selectedClasses.getSize() > 0))
				{
					if (subClass.isVirtual())
						subClass = subClass.getRealClass();

					if (subClass.isAbstract())
					{
						subToolEnabled = enabled;
						for (i = 0, ns = selectedClasses.getSize(); i < ns; i++)
						{
							selectedClass = selectedClasses.getInfo(i);
							subToolEnabled = subToolEnabled
									&& (selectedClass.inheritsFrom(subClass));
						}
					}
					else
					{
						subToolEnabled = enabled && (selectedClasses.getSize() == 1);
						selectedClass = selectedClasses.getInfo(0);

						subToolEnabled = subToolEnabled && (selectedClass == subClass);
					}
				}

				tool = (LyTool)comp;
				tool.setEnabled(subToolEnabled);
			}
		}

		StringBuffer toolId = new StringBuffer();

		toolId.append(LySimpleTableController.DEMAND_SUB_TOOL_ACTION);
		toolId.append(LyController.ID_SEP);
		toolId.append(action.getRoot().getId());
		toolId.append(getImage(action));

		comp = toolBar.getComponent(toolId.toString());
		if ((comp != null) && (comp instanceof LyTool))
		{
			tool = (LyTool)comp;

			boolean   toolEnabled = false;
			LyArea    area = (LyArea)tool.getComponent();
			LyToolBar subToolBar = area.getToolBar();
			LyTool    subTool = null;

			for (j = 0, n = subToolBar.getChildCount(); (j < n) && !toolEnabled; j++)
			{
				subTool = (LyTool)subToolBar.getChild(j);
				toolEnabled = toolEnabled || subTool.isEnabled();
			}

			tool.setEnabled(toolEnabled);
		}
	} // end method updateTool

	/**
	 * Retrieves the view Builder.
	 *
	 * @return  the builder instance;
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyTableBuilder(this);
	}

	/**
	 * This method updates the footers of the table.
	 */
	protected void updateFooters()
	{
		if (_component == null)
			return;

		if (isDisplayList()) {
			updateDisplayListFooters();
			return;
		}

		// If table is Idle, no need to refresh
		if (isIdle() || (_workSpace == null))
			return;

		LyTable   table = getTableComponent();
		LyFooters footers = table.getFooters();

		if (footers == null)
			return;

		LyInfoList<LyFieldInfo> fields = getFields();
		int					    i, n = ((fields == null) ? 0 : fields.getSize());
		LyFieldInfo			    fieldInfo;
		LyFooter			    footer;
		String				    label = null;
		LySimpleObjectList	    objectList = null;

		if (n > 0)
		{
			objectList = new LySimpleObjectList(); // FFT1524
			objectList.setSorted(false);
			objectList.addList(_workSpace);
		}

		LyTableBehaviorInterface behavior = (LyTableBehaviorInterface)getBehavior();

		for (i = 0; i < n; i++)
		{
			fieldInfo = fields.getInfo(i);
			label = behavior.getFooterLabel(this, fieldInfo, objectList);
			footer = (LyFooter)footers.getComponent("footer" + ID_SEP + fieldInfo.getId());
			if (footer == null)
				continue;

			if (label == null)
				label = "";

			footer.setLabel(label);
		}

		// Update the last modification time.
		setLastModified(new Date().getTime());
	} // end method updateFooters


	/**
	 * This method updates the footers of the table displayed in list.
	 */
	protected void updateDisplayListFooters()
	{
		// If table is Idle, no need to refresh
		if (isIdle() || (_workSpace == null))
			return;

		LyTable table = getTableComponent();
		LyFooters footers = table.getFooters();

		if (footers == null)
			return;

		LyInfoList<LyFieldInfo> fields = getFields();
		LyInfoList<LyFieldInfo> footerFields = new LyInfoList<LyFieldInfo>();
		LyClassInfo				classInfo = getClassInfo();
		LyFieldInfo fieldInfo;

		for (int i = 0; i < fields.getSize(); i++)
		{
			fieldInfo = fields.getInfo(i);
			if ((fieldInfo instanceof LyNumberFieldInfo)
					&& (((LyNumberFieldInfo) fieldInfo).getFunction() != LyNumberFieldInfo.Function.NONE)
						&& classInfo.contains(fieldInfo))
			{
				footerFields.addInfo(fieldInfo);
			}
		}

		int n = footerFields.getSize();

		LyFooter footer;
		String label = null;
		LySimpleObjectList objectList = null;

		if (n > 0) {
			objectList = new LySimpleObjectList(); // FFT1524
			objectList.setSorted(false);
			objectList.addList(_workSpace);
		}

		LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) getBehavior();

		for (int j = 0; j < n; j++) {
			fieldInfo = footerFields.getInfo(j);
			label = behavior.getFooterLabel(this, fieldInfo, objectList);
			footer = (LyFooter) footers.getComponent("footer" + ID_SEP + fieldInfo.getId());
			if (footer == null)
				continue;

			if (label == null)
				label = "";

			footer.setLabel(label);
		}

		// Update the last modification time.
		setLastModified(new Date().getTime());
	}

	/**
	 * This method is invoked to enable the the area(s) to complete the message bar of the view and
	 * add its own message items.
	 *
	 * @param  messages  messages bar
	 * @param  level     index of message
	 *
	 * @see    LyAreaBuilder#completeMessages
	 */
	public void completeMessages(LyMessages messages, int level)
	{
		if (_action.hasMark(LyAction.MARK_NO_MESSAGES))
			return;

		LyMessage objectsMessage = buildObjectsMessage();
		LyMessage filterMessage = buildFilterMessage();
		LyMessage printMessage = buildPrintMessage();

		if (objectsMessage != null)
			messages.addMessage(objectsMessage);

		if (filterMessage != null)
			messages.addMessage(filterMessage);

		if (printMessage != null)
			messages.addMessage(printMessage);

		super.completeMessages(messages, level);
	}

	/**
	 * Builds the objects message description.
	 *
	 * @return  the component for the objects message description
	 */
	protected LyMessage buildObjectsMessage()
	{
		LyMessage objectsMessage = new LyMessage(this.getId() + "_objectsMessage", " ");

		objectsMessage.setSize(10);
		return objectsMessage;
	}

	/**
	 * Builds the print message description.
	 *
	 * @return  the component for the objects message description
	 */
	protected LyMessage buildPrintMessage()
	{
		LyMessage printMessage = new LyMessage(this.getId() + "_printMessage", " ");

		printMessage.setSize(2);
		return printMessage;
	}

	/**
	 * This method updates the content of the message bar. Current implementation indicates
	 * selection, filters ...
	 */
	protected void updateMessages()
	{
		if (_component == null)
			return;

		// If table is Idle, no need to refresh
		if (isIdle())
			return;

		LyView view = getView();

		if (!(view instanceof LyFrame))
			return;

		LyMessages messages = ((LyFrame)view).getMessages();

		if (messages == null)
			return;

		LyMessage     objectsMessage = ((view == null) ? null
				: (LyMessage)messages.getComponent(getId() + "_objectsMessage"));
		LyMessage     filterMessage = ((view == null) ? null
				: (LyMessage)messages.getComponent(getId() + "_filterMessage"));
		int			  n = getVisibleObjectsCount();
		int			  nSelections = ((_selections == null) ? 0 : _selections.getSize());
		LyEnvironment environment = getEnvironment();

		if (!_selection && (filterMessage != null))
		{
			StringBuffer filterLabel;

			if (_filter == null)
			{
				filterMessage.setImage("LY_UNACTIVE_FILTER");
				filterLabel = new StringBuffer(environment.getMessage("LY_NO_FILTER"));
			}
			else if (isFilterActive())
			{
				filterMessage.setImage("LY_ACTIVE_FILTER");
				filterLabel = new StringBuffer(environment.getMessage("LY_ACTIVE_FILTER"));
				filterLabel.append(" ");
				if (_workSpace != null)
					filterLabel.append(getMessageFilterLabel(_workSpace.getSize()));
			}
			else
			{
				filterMessage.setImage("LY_UNACTIVE_FILTER");
				filterLabel = new StringBuffer(environment.getMessage("LY_UNACTIVE_FILTER"));
			}

			filterMessage.setLabel(filterLabel.toString());
		}

		// Set message for objects
		if (objectsMessage != null)
		{
			StringBuffer selectionsLabel = null;
			String		 message = objectsMessage.getLabel();

			if (_selection && (message != null))
			{
				int p = message.lastIndexOf(" (");

				if (p >= 0)
					message = message.substring(0, p);

				selectionsLabel = new StringBuffer(message);
			}
			else
				selectionsLabel = new StringBuffer(getMessageSelectionLabel(n));

			if (n > 0)
			{
				selectionsLabel.append(" (");
				selectionsLabel.append(environment.getMessage("LY_SELECTED_OBJECTS_COUNT", new Integer(nSelections)));
				selectionsLabel.append(")");
			}

			objectsMessage.setLabel(selectionsLabel.toString());
		}

		// Update the last modification time.
		// setLastModified(new Date().getTime());
	} // end method updateMessages

	/**
	 * Gets the selection text for the specified number of items in the view May be overriden to add
	 * additional info (Multipage for instance).
	 *
	 * @param   n  number of rows currently being shown
	 *
	 * @return  the label shown in message bar.
	 */
	protected String getMessageSelectionLabel(int n)
	{
		String limitParameter = (String)getAction().getParameter("_limit");

		if (limitParameter != null)
		{
			int count = getTotalCount();

			return getEnvironment().getMessage("LY_OBJECTS_MULTIPAGE_COUNT", new Integer(n), new Integer(count));
		}
		else
			return getEnvironment().getMessage("LY_OBJECTS_COUNT", new Integer(n));
	}

	/**
	 * Method for counting visible objects.
	 * @return the number of visible objects
	 */
	protected int getVisibleObjectsCount()
	{
		LyTable table = getTableComponent();
		LyRows  rows = table.getRows();

		return getRowCount(rows);
	}

	/**
	 * Method for counting viewable objects.
	 * @return the total number of objects
	 */
	protected int getTotalCount()
	{
		if (_workSpace instanceof LyPageWorkSpace)
			return ((LyPageWorkSpace)_workSpace).getTotalCount();

		if (_context != null)
			return ((_workSpace == null) ? 0 : _workSpace.getSize());

		LySession	   session = getSession();
		LyObjectSource source = session.getObjectSource(_classInfo);
		LyFilter	   filter;

		if ((isFilterActive()) && (_classFilter != null))
		{
			filter = new LyFilter(LyFilter.Type.AND);
			filter.addElement(_classFilter);
			filter.addElement(_filter);
		}
		else if (isFilterActive())
			filter = _filter;
		else
			filter = _classFilter;

		// If the filter applies on local fields, contacting the object source
		// is useless, use the workspace size.
		if ((filter != null) && filter.hasLocalFields())
			return ((_workSpace == null) ? 0 : _workSpace.getSize());

		return source.getSize(filter);
	}

	/**
	 * Gets the selection text for the specified number of items in the view May be overriden to add
	 * additional info (Multipage for instance).
	 *
	 * @param   n  number of objects matching filter
	 *
	 * @return  the label shown in message bar.
	 */
	protected String getMessageFilterLabel(int n)
	{
		if (_initialWorkSpace == null)
			return null;

		String limitParameter = (String)getAction().getParameter("_limit");
		int    total;

		if (limitParameter != null)
		{
			LySession	   session = getSession();
			LyObjectSource source = session.getObjectSource(_classInfo);

			n = getTotalCount();
			total = source.getSize(_classFilter);
		}
		else
			total = _initialWorkSpace.getSize();

		return getEnvironment().getMessage("LY_FILTER_OBJECTS_COUNT", new Integer(n), new Integer(total));
	}

	/**
	 * Implementation of LyPrintListenerInterface.
	 *
	 * @param  page  page to print
	 */
	public void print(int page)
	{
		super.print(page);

		// Set message for printjob
		LyView    view = getView();
		LyMessage printMessage = (LyMessage)view.getComponent(getId() + "_printMessage");

		if (page <= 0)
		{
			printMessage.setImage(null);
			printMessage.setLabel("");
		}
		else
		{
			printMessage.setImage("LY_ACTION_PRINT");
			printMessage.setLabel(" " + page);
		}
	}

	/**
	 * Update the menu for the current controller.
	 */
	protected void updateMenuBar()
	{
		if (_workSpace != null)
		{
			updateTools();
			updateFindMenu();
		}

		super.updateMenuBar();
	}

	/**
	 * Update the sort menu item.
	 */
	protected void updateSortMenu()
	{
		if (_component == null)
			return;

		// If table is Idle, no need to refresh
		if (isIdle())
			return;

		boolean hasMenu = hasMenu();

		if (!hasMenu)
			return;

	    // Get menu bar
		LyView    view = getView();
        LyMenuBar menuBar = (view instanceof LyFrame ? ((LyFrame)view).getMenuBar() : null);

		if (menuBar == null)
			return;

		// Get all the columns in the table
		LyInfoList<LyFieldInfo> fields = getFields();
		LySort				    sort = getSort();

		if ((fields == null) || (sort == null))
			return;

		// Update items
		LyItem item;

		// update the menu item for the  status sort
		item = menuBar.findItem(DEMAND_STATUS_SORT, getId());
		if (item != null)
		{
			if (_currentStatusSort == 1)
				item.setEnabled(false);
			else
				item.setEnabled(true);
		}

		// update the menu item for the invert status sort
		item = menuBar.findItem(DEMAND_INVERT_STATUS_SORT, getId());
		if (item != null)
		{
			if (_currentStatusSort != 1)
				item.setEnabled(false);
			else
				item.setEnabled(true);
		}

		int		    extendSortItem = 0;
		int		    i, criterionIndex;
		LyFieldInfo fieldInfo;

		for (i = 0; i < fields.getSize(); i++)
		{
			fieldInfo = fields.getInfo(i);
			criterionIndex = sort.getIndex(fieldInfo);
			item = menuBar.findItem(DEMAND_SORT_BY + LyController.ID_SEP
					+ fieldInfo.getId(), getId());
			if (item != null)
			{
				if (criterionIndex >= 0)
					item.setEnabled(false);
				else
					item.setEnabled(true);
			}

			item = menuBar.findItem(DEMAND_SORT_REVERSE + LyController.ID_SEP
					+ fieldInfo.getId(), getId());
			if (item != null)
			{
				if (criterionIndex >= 0)
					item.setEnabled(true);
				else
					item.setEnabled(false);
			}

			item = menuBar.findItem(DEMAND_SORT_EXTEND + LyController.ID_SEP
					+ fieldInfo.getId(), getId());
			if (item != null)
			{
				if (criterionIndex >= 0)
				{
					item.setEnabled(false);
					extendSortItem++;
				}
				else
					item.setEnabled(true);
			}
		}

		boolean useExtendSort = (extendSortItem > 1);

		for (i = 0; i < fields.getSize(); i++)
		{
			fieldInfo = fields.getInfo(i);
			criterionIndex = sort.getIndex(fieldInfo);

			item = menuBar.findItem(DEMAND_SORT_BY + LyController.ID_SEP + fieldInfo.getId(), getId());
			if (item != null)
			{
				if ((criterionIndex >= 0) && !useExtendSort)
					item.setEnabled(false);
				else
					item.setEnabled(true);
			}
		}
	} // end method updateSortMenu

	/**
	 * Updates area controller content.
	 */
	protected void reload()
	{
		if ((_initialWorkSpace instanceof LyPageWorkSpace) && _dynamic)
		{
			boolean wasIdle = isIdle();
			setIdle(true);
			((LyPageWorkSpace)_initialWorkSpace).reload();
			setIdle(wasIdle);
		}
		else
			super.reload();
	}

	/**
	 * Try to transform given context to a filter to optimize requests when paging is on.
	 * If the amount of objects in the context is greater than MAX_OBJECTS_COUNT_FOR_CONTEXT_FILTER,
	 * null is returned.
	 * @param context the context list
	 * @return a filter corresponding to the context (may be null)
	 */
	protected LyFilter getFilterFromContext(LyObjectList context)
	{
		int no = ((context == null) ? 0 : context.getSize());

		if ((no == 0) || (no > MAX_OBJECTS_COUNT_FOR_CONTEXT_FILTER))
			return null;

		LyInfoList<LyClassInfo> classes = context.getClasses();
		int					    nc = ((classes == null) ? 0 : classes.getSize());

		if (nc == 0)
			return null;

		// For each classes of this list
		int					    c;
		LyClassInfo			    classInfo;
		LyInfoList<LyRoute>     routes;
		LyRoute				    route;
		int					    r, nr;
		Vector<LyStep>		    steps;
		LyStep				    step;
		int					    s, ns;
		LyInfoList<LyFieldInfo> relations = null, stepRelations;
		LyFieldInfo			    relation;

		for (c = 0; c < nc; c++)
		{
			classInfo = classes.getInfo(c);
			routes = classInfo.getRoutes(_classInfo);
			if (routes == null)
				return null;

			for (r = 0, nr = routes.getSize(); r < nr; r++)
			{
				route = routes.getInfo(r);
				if (route.isReverse())
					return null;

				steps = route.getSteps();
				if (steps == null)
					return null;

				for (s = 0, ns = steps.size(); s < ns; s++)
				{
					step = steps.elementAt(s);
					if (!step.isReverse())
						return null;

					stepRelations = step.getRelations();
					if (stepRelations == null)
						return null;

					if (relations == null)
						relations = new LyInfoList<LyFieldInfo>();

					relations.addList(stepRelations);
				}
			}
		}

		// Build filter on each object
		LyFilter					 filter = new LyFilter(LyFilter.Type.OR);
		LyFilterExpression			 filterExp;
		int							 o;
		LyObject					 object;
		LyFilterExpression.Condition condition;

		for (r = 0, nr = relations.getSize(); r < nr; r++)
		{
			relation = relations.getInfo(r);
			if (!_classInfo.contains(relation))
				return null;

			if (!relation.isMultiple())
				condition = LyFilterExpression.Condition.EQUALS;
			else
				condition = LyFilterExpression.Condition.CONTAINS;

			for (o = 0; o < no; o++)
			{
				object = context.getObject(o);
				filterExp = new LyFilterExpression(relation, false, condition, object.getObjectId());
				filterExp.setModifier(LyFilterExpression.MOD_CASE_SENSITIVE);
				filter.addElement(filterExp);
			}
		}

		return filter;
	} // end method getFilterFromContext

	/**
	 * Updates groupBy menu according to current sort.
	 */
	protected void updateGroupByMenu()
	{
		if (_component == null)
			return;
		// If table is Idle, no need to refresh
		if (isIdle())
			return;

		if (!hasMenu())
			return;

		LyView view = getView();

		if (view == null)
			return;

		// Get menu bar
		LyMenuBar menuBar = (LyMenuBar)view.getComponent("_menuBar");

		if (menuBar == null)
			return;

		LyClassInfo			    classInfo = getClassInfo();
		LyInfoList<LyFieldInfo> groupByFieldInfos = ((_sort == null) ? null : _sort.getGroupByFieldInfos());
		int					    i, n = classInfo.getFieldInfoCount();
		LyFieldInfo			    fieldInfo;
		LyItem				    item;

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			item = menuBar.findItem(DEMAND_GROUP_BY + LyController.ID_SEP + fieldInfo.getId(), getId());
			if (item != null)
			{
				if ((groupByFieldInfos != null)
						&& groupByFieldInfos.contains(fieldInfo))
					item.setChecked(true);
				else
					item.setChecked(false);
			}
		}
	}

	/**
	 * Builds a row corresponding to footers when objects are grouped.
	 * @param from start index for the object list
	 * @param to end index for the object list
	 * @param status indicate if there is a status column
	 * @return the footer row
	 */
	protected LyRow buildFooterRow(int from, int to, boolean status)
	{
		/*if (isDisplayList())
		{
			return super.buildFooterRow(from, to, status);
		}*/
		if (to - from < 1)
			return null;

		LySimpleObjectList      objectList = new LySimpleObjectList();
		int                     i;

		objectList.setSorted(false);
		for (i = from; i < to; i++)
			objectList.addObject(_workSpace.getObject(i));

		LyRow                    row = new LyRow();
		LyInfoList<LyFieldInfo>  fields = getFields();
		int                      n = ((fields == null) ? 0 : fields.getSize());
		LyFieldInfo              fieldInfo;
		LyCell                   cell;
		String                   label = null;
		LyTableBehaviorInterface behavior = (LyTableBehaviorInterface)getBehavior();

		if (status)
		{
			cell = new LyCell();
			cell.setLabel("");
			row.addCell(cell);
		}

		for (i = 0; i < n; i++)
		{
			fieldInfo = fields.getInfo(i);
			label = behavior.getFooterLabel(this, fieldInfo, objectList);
			cell = new LyCell();
			if (label != null)
				cell.setLabel(label);
			row.addCell(cell);
		}

		row.setId(FOOTER_ROW_PREFIX + from + "_" + to);
		row.setSensitive(false);
		row.setFont("LY_GROUP_FOOTER_FONT");
		row.setColor("LY_HEADER_COLOR");
		row.setFontColor("LY_TABLE_HEADER_FOREGROUND");

		return row;
	}
} // end class LyTableController
