/**
 * LyTextController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyTextController.java 31261 2011-05-31 14:45:11Z jlsaouli $
 */

package leon.control;

import java.io.*;
import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * This class implements a dialog window that displays a text chosen by the user or the content of a
 * text file. The generated window can have a titleBar and an image. If a file is used, the user can
 * specify a refresh time ( in seconds ). If the time interval is set to zero, there is no refresh.
 */
public class LyTextController extends LyAreaController
{
	/** File Version */
	public static final String __VERSION =
		"$Id: LyTextController.java 31261 2011-05-31 14:45:11Z jlsaouli $";

	/** "Save as" demand. */
    public static final String  DEMAND_SAVE_AS = "_saveAs";
  
    /** "Clear" demand. */
    public static final String  DEMAND_CLEAR = "_clear";
    
	/** File path, or null if this controller doesn't display a file. */
	protected String		   _file;

	/** Refreshing interval (in milliseconds). */
	protected int			   _interval;

	/** Thread used to refresh the content of the text window. */
	protected _LyRefreshThread _refreshingThread;

	/** Length of file. */
	protected long			   _fileLength;

	/** Process viewer for console mode. */
	protected LyProcessViewer  _processViewer;

	/** Temporary file for console mode. */
	protected File			   _temporaryFile;
    
    /** Start of displayed text. */
	protected int              _start;
    

	/**
	 * Specific constructor.
	 *
	 * @param session  Session for this controller
	 * @param parent   Parent controller
	 * @param action   Action
	 * @param file     File to display in the text window
	 * @param interval refresh interval
	 */
	public LyTextController(LySession session, LyController parent, LyAction action,
		String file, int interval)
	{
		this(session, parent, action, file, interval, true);
	}

	/**
	 * Specific constructor.
	 *
	 * @param session  Session for this controller
	 * @param parent   Parent controller
	 * @param action   Action
	 * @param file     File to display in the text window
	 * @param interval refresh interval
	 * @param seconds  true to set the unit of the refresh interval in second, false for millisecond
	 */
	public LyTextController(LySession session, LyController parent, LyAction action,
		String file, int interval, boolean seconds)
	{
		super(session, parent, action, null);

		_refreshingThread = null;
        _start = 0;

		setFile(file);
		if (seconds)
			setInterval(interval);
		else
			setIntervalInMillis(interval);
	}

	/**
	 * Specific constructor for console mode.
	 *
	 * @param session       Session for this controller
	 * @param parent        Parent controller
	 * @param action        Action
	 * @param command       Command to execute
	 * @param commandString Command string to display to the end-user
	 */
	public LyTextController(LySession session, LyController parent, LyAction action,
		String command, String commandString)
	{
		this(session, parent, action, command, commandString, null);
	}

	/**
	 * Specific constructor for console mode.
	 *
	 * @param session       Session for this controller
	 * @param parent        Parent controller
	 * @param action        Action
	 * @param commands      Command list to execute
	 * @param commandString Command string to display to the end-user
	 */
	public LyTextController(LySession session, LyController parent, LyAction action,
		String[] commands, String commandString)
	{
		this(session, parent, action, commands, commandString, null);
	}

	/**
	 * Specific constructor for console mode.
	 *
	 * @param session       Session for this controller
	 * @param parent        Parent controller
	 * @param action        Action
	 * @param commands      Command list to execute
	 * @param commandString Command string to display to the end-user
	 * @param errorMessage  Error message to show to the end-user if needed
	 */
	public LyTextController(LySession session, LyController parent, LyAction action,
		String[] commands, String commandString, String errorMessage)
	{
		super(session, parent, action, null);

		// Create a new environment and set its standard and error output
		LyEnvironment environment = new LyEnvironment(parent.getEnvironment());
		String		  dir = environment.getEnv("LY_DATA_DIR");
		String		  fileName = dir + "/" + LyString.getRandomId();

		try
		{
			_temporaryFile = new File(fileName);

			FileOutputStream output = new FileOutputStream(_temporaryFile);
			PrintStream		 printStream = new PrintStream(output);

			environment.setOutputStream(printStream);
			environment.setErrorStream(printStream);
		}
		catch (IOException e)
		{
			environment.writeException(e);
			free();
			return;
		}

		// Execute command and show results
		if (commandString != null)
			environment.logOut(">>> " + commandString);
		else
		{
			StringBuffer messageLog = new StringBuffer(">>>");

			for (int i = 0; i < commands.length; i++)
				messageLog.append(" ").append(commands[i]);
			environment.logOut(messageLog.toString());
		}

		_processViewer = new LyProcessViewer(environment, commands, false, 10, errorMessage);

		setFile(fileName);
		setIntervalInMillis(50);
	}

	/**
	 * Specific constructor for console mode.
	 *
	 * @param session       Session for this controller
	 * @param parent        Parent controller
	 * @param action        Action
	 * @param command       Command to execute
	 * @param commandString Command string to display to the end-user
	 * @param errorMessage  error message to show to the end-user if needed
	 */
	public LyTextController(LySession session, LyController parent, LyAction action,
		String command, String commandString, String errorMessage)
	{
		super(session, parent, action, null);

		// Create a new environment and set its standard and error outputs
		LyEnvironment environment = new LyEnvironment(parent.getEnvironment());
		String		  dir = environment.getEnv("LY_DATA_DIR");
		String		  fileName = dir + "/" + LyString.getRandomId();

		try
		{
			_temporaryFile = new File(fileName);

			FileOutputStream output = new FileOutputStream(_temporaryFile);
			PrintStream		 printStream = new PrintStream(output);

			environment.setOutputStream(printStream);
			environment.setErrorStream(printStream);
		}
		catch (IOException e)
		{
			environment.writeException(e);
			free();
			return;
		}

		// Execute command and show results
		if (commandString != null)
			environment.logOut(">>> " + commandString);
		else
			environment.logOut(">>> " + command);

		_processViewer = new LyProcessViewer(environment, command, false, 10, errorMessage);

		setFile(fileName);
		setIntervalInMillis(50);
	}

	/**
	 * Default constructor.
	 *
	 * @param session Session for this controller
	 * @param parent  Parent controller
	 * @param action  Action
	 * @param classInfo related classinfo
	 * @param objects context of the action
	 */
	public LyTextController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		super(session, parent, action, null);

		String file = (String)action.getParameter("_file");
		String parameter = (String)action.getParameter("_refresh");
		int    interval = 0;

		if (parameter != null)
			interval = LyTools.intFromString(parameter);

		setFile(file);
		setInterval(interval);
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyTextBehaviorInterface.class;
	}

	/**
	 * Invoked when an action occurs.
	 *
	 * @param demand Paramters of the action to perform
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			// Menu or action bar
			String id = demand.getSource();

			if (id.equals(LyFormController.DEMAND_VALIDATE))
			{
				hideView();
				if (_refreshingThread != null)
				{
					_refreshingThread.askStop();
					_refreshingThread = null;
				}
				free();
			}
            else if (id.equals(DEMAND_SAVE_AS))
            {
                save(null);
                return;
            }
            else if (id.equals(DEMAND_CLEAR))
            {
                clear();
                return;
            }
		}

		super.actionDemand(demand);
	}

	/**
	 * Invoked when the controller is destroyed.
	 *
	 * @param controller parent controller
	 * @see   LyAreaController
	 */
	public void free(LyController controller)
	{
		if (_refreshingThread != null)
		{
			_refreshingThread.askStop();
			_refreshingThread = null;
		}

		if (_processViewer != null)
		{
			LyEnvironment env = _processViewer.getEnvironment();

			env.free();

			_processViewer.abort();
			_processViewer = null;
		}

		if (_temporaryFile != null)
		{
			_temporaryFile.delete();
			_temporaryFile = null;
		}

		super.free(controller);
	}

	/**
	 * Retrieve the refresh time.
	 *
	 * @return Refresh time interval in seconds
	 */
	public int getInterval()
	{
		return getIntervalInMillis() / 1000;
	}

	/**
	 * Retrieve the refresh time.
	 *
	 * @return Refresh time interval in milliseconds
	 */
	public int getIntervalInMillis()
	{
		return _interval;
	}

	/**
	 * Set the refresh time.
	 *
	 * @param interval Refresh time interval (in seconds). the 0 value disables the refresh
	 */
	public void setInterval(int interval)
	{
		setIntervalInMillis(interval * 1000);
	}

	/**
	 * Set the refresh time.
	 *
	 * @param interval Refresh time interval (in seconds). the 0 value disables the refresh
	 */
	public void setIntervalInMillis(int interval)
	{
		_interval = interval;

		if ((_interval != 0) && (_file != null) && (_file.length() > 0))
		{
			// if thread already exist, just change time interval
			if (_refreshingThread != null)
				_refreshingThread.setInterval(_interval);
			else
			{
				_refreshingThread = new _LyRefreshThread(this, _interval);
				_refreshingThread.start();
			}
			LyArea area = (LyArea)getComponent();

			area.setRefresh(interval);
		}
		else if (_refreshingThread != null)
		{
			// stop the current threa
			_refreshingThread.askStop();
			_refreshingThread = null;

			LyArea area = (LyArea)getComponent();

			area.setRefresh(0);
		}
	} // end method setIntervalInMillis

	/**
	 * Set the file to display in the text window.
	 *
	 * @param  file       Path of the file to display
	 * @param  fieldLabel Label of the text field
	 * @param  nCols      Number of columns of the text field
	 * @param  nRows      Number of rows of the text field
	 * @return the text view component
	 */
	public LyText setFile(String file, String fieldLabel, int nCols, int nRows)
	{
		LyText text = null;

		_file = file;

		LyEnvironment environment = getEnvironment();

		if (file == null)
		{
			text = setText(fieldLabel, null, nCols, nRows, null);
			return null;
		}

		_file = environment.getFile(file);
		_file = environment.getString().expand(_file);

		// First read the file
		try
		{
			BufferedReader in = new BufferedReader(new FileReader(_file));
			StringBuffer   sb = new StringBuffer(1000);
			String		   line;

			while ((line = in.readLine()) != null)
				sb.append(line + "\n");

			in.close();

			text = setText(fieldLabel, sb.toString(), nCols, nRows, null);
		}
		catch (FileNotFoundException e)
		{
			try
			{
				InputStream  stream = environment.getResourceAsStream(_file, null);
				StringBuffer sb = new StringBuffer();
				int			 b = stream.read();

				while (b != -1) // EOF
				{
					sb.append((char)b);
					b = stream.read();
				}

				stream.close();
				text = setText(fieldLabel, sb.toString(), nCols, nRows, null);
			}
			catch (Exception e2)
			{
				String message = environment.getMessage("LY_ER_FILE", _file);

                text = setText(fieldLabel, message, nCols, nRows, null);
			}
		}
		catch (IOException e)
		{
			environment.writeException(e);

			text = setText(fieldLabel, null, nCols, nRows, null);
			String message = environment.getMessage("LY_ER_FILE", _file);

			showError(message);
		}

		return text;
	} // end method setFile

	/**
	 * Set the file to display in the text window. In this case there is no label, and the text
	 * field has 100 columns and 20 rows
	 *
	 * @param  file Path of the file to display
	 * @return the text view component
	 */
	public LyText setFile(String file)
	{
		return setFile(file, null, -1, -1);
	}

	/**
	 * Refresh the content of the current file.
	 */
	protected void refreshFile()
	{
		if ((_file == null) || (_file.length() == 0))
			return;

		try
		{
			try
			{
				File file = new File(_file);

				if (_fileLength == file.length())
					return;

				_fileLength = file.length();

				BufferedReader in = new BufferedReader(new FileReader(file));
				StringBuffer   sb = new StringBuffer(1000);
				String		   line;

				while ((line = in.readLine()) != null)
					sb.append(line + "\n");

				in.close();

				updateText(sb.toString());
			}
			catch (FileNotFoundException e)
			{
				try
				{
					LyEnvironment environment = getEnvironment();
					InputStream   stream = environment.getResourceAsStream(_file, null);
					StringBuffer  sb = new StringBuffer();
					int			  b = stream.read();

					while (b != -1) // EOF
					{
						sb.append((char)b);
						b = stream.read();
					}

					stream.close();
					updateText(sb.toString());
				}
				catch (Exception e2)
				{
					LyEnvironment environment = getEnvironment();

					environment.writeException(e);
				}
			}
		}
		catch (Exception e)
		{
			// An exception can occur is controller is freed during refresh
		}
	} // end method refreshFile

	/**
	 * Set the text to display in the text window.
	 *
	 * @param  fieldLabel Label of the text field
	 * @param  itemLabel  Text to display
	 * @param  nCols      Number of columns of the text field
	 * @param  nRows      Number of rows of the text field
	 * @param  format     Format of the text (See also tag <TEXT> of leon.dtd)
	 * @return the text view component
	 */
	public LyText setText(String fieldLabel, String itemLabel,
		int nCols, int nRows, String format)
	{
		LyTextBuilder builder = (LyTextBuilder)getBuilder();

		if (builder == null)
			return null;

		String fieldId = getId() + "_FileId";

		return builder.setText(fieldId, fieldLabel, itemLabel, nCols, nRows, format, true);
	}

	/**
	 * Update the text in the field with the given id.
	 *
	 * @param content New text for the field
	 */
	public void updateText(String content)
	{        
        if ((_start > 0)
                && (content != null)
                && (_start < content.length()))
            content = content.substring(_start);
        
        updateTextField(content);
        
		// Update the last modified time.
		Date date = new Date();

		setLastModified(date.getTime());
	}

    /**
     * Update text field label.
     * @param label New field label.
     */
    protected void updateTextField(final String label)
    {
        LyViewManager viewManager = super.getViewManager();
        
        viewManager.execGui(new _LyTextControllerUpdater(label));
    }
    
	class _LyTextControllerUpdater
		implements Runnable, LyGetSessionInterface
	{
		String		_label;
		
		_LyTextControllerUpdater(String label)
		{
			_label = label;
		}
		
		public void run()
		{
	        LyText  text = getTextField();
            
	        if (text != null)	        
	        	text.setLabel(_label);
		}
		
		public LyDataSession getSession()
		{
			return LyTextController.this.getSession();
		}
	}
    
    /**
     * Clear text content.
     */
    public void clear()
    {
        LyText      text = getTextField();
        String      content = (text == null ? null : text.getLabel());
        
        if (content != null)
            _start += content.length();
        updateText(null);
    }
    
    /**
     * Open file chooser and save current text in chosen file.
     * @param file path of the file used to save the content
     */
    public void save(String file)
    {
        if (file == null)
        {
            LySelectFileController controller = new LySelectFileController(getSession(), this, getClassInfo());
            
            controller.addListener(this);
            controller.setUserData(DEMAND_SAVE_AS);
            controller.showView();
            return;
        }
        
        LyText      text = getTextField();
        String      content = text.getLabel();
        
        try
        {
            FileOutputStream output = new FileOutputStream(file);
            PrintWriter      pw = new PrintWriter(output);

            pw.write(content);
            pw.close();
        }
        catch (Exception e)
        {
            _environment.writeException(e);
        }
    }
    
    /**
     * Invoked when the user validates the input content.
     *
     * @param controller the current LySelectFileController instance
     * @param demand     demand parameters sent by the view component.
     */
    public void selectFileDemand(LySelectFileController controller, LyDemand demand)
    {
        String  demandId = demand.getSource();
        
        if ((demand.getType() == LyActionDemand.TYPE_ACTION)
            && (demandId != null)
            && demandId.equals(LyFormController.DEMAND_VALIDATE))
        {
            String  filePath = controller.getSelectedPath();
            
            if (filePath != null)
                save(filePath);
        }
    }
    
	/**
	 * Retrieves the view Builder.
	 *
	 * @return the builder instance;
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyTextBuilder(this);
	}

    /**
     * Gets text field.
     * @return the text field
     */
    protected LyText getTextField()
    {
        if (_component == null)
            return null;
        
        String      fieldId = getId() + "_FileId";
        LyFormField formField = (LyFormField)getComponent().getComponent(fieldId);
        
        if (formField == null)
            return null;
        
        return (LyText)formField.getFieldItem();
    }
    
	/**
	 * Gets process corresponding to command execution when used as console.
	 * @return the process
	 */
	public Process getProcess()
	{
		if (_processViewer == null)
			return null;

		return _processViewer.getProcess();
	}

	/**
	 * This class implements a thread that refresh the text of the controller at a given time
	 * interval.
	 */
	class _LyRefreshThread extends Thread
	{
		LyTextController _textController;
		int				 _interval;
		boolean			 _stopRequest;

		/**
		 * Contructor for the class.
		 *
		 * @param textController Parent controller for this thread
		 * @param interval       Refresh time interval
		 */
		public _LyRefreshThread(LyTextController textController, int interval)
		{
			_textController = textController;
			_interval = interval;
			_stopRequest = false;

			this.setName("BF AppEngine : text update thread");
		}

		/**
		 * Change the refresh time interval.
		 *
		 * @param interval Refresh time interval
		 */
		public void setInterval(int interval)
		{
			_interval = interval;
		}

		/**
		 * Implementation of the run method of the java.lang.Thread class This function refresh the
		 * content of the text window every interval seconds.
		 */
		public void run()
		{
			startRun();

			while (!_stopRequest)
			{
				if ((_processViewer != null)
						&& !_processViewer.isRunning())
				{
					LyEnvironment env = _processViewer.getEnvironment();

					_processViewer.getEnvironment().logOut("\n<<<");
					env.err.close();
					env.out.close();

					_processViewer = null;
					askStop();
				}

				_textController.refreshFile();

				try
				{
					sleep(_interval);
				}
				catch (InterruptedException e)
				{
					return;
				}
			}

			endRun();
		} // end method run

		/**
		 * This method is used to stop the thread.
		 */
		public void askStop()
		{
			_stopRequest = true;
		}
	} // end class _LyRefreshThread
} // end class LyTextController
