/**
 * LyTreeController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyTreeController.java 41017 2012-10-03 13:01:02Z gtachon $
 */

package leon.control;

import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;


/**
 * The LyTreeController implements the generic behavior for a tree. This controller is associated
 * with a builder (LyTreeBuilder) and a behavior. The behavior must implements
 * LyTreeBehaviorInterfave. Usually, this behavior is an instance of the LyTreeBehavior class
 *
 * @see LyTreeBaseController
 * @see leon.app.behaviorinterface.LyTreeBehaviorInterface
 * @see leon.app.behavior.LyTreeBehavior
 */
public class LyTreeController extends LyTreeBaseController
	implements LyAdvancedWorkSpaceListener //, LyFindInterface
{
	/** File Version */
	public static final String							 __VERSION =
		"$Id: LyTreeController.java 41017 2012-10-03 13:01:02Z gtachon $";

	/** Workspace that store all the LyObject of the tree. */
	protected LySimpleWorkSpace							 _workSpace;

	/**
	 * used to store the objectId of open node Cannot use treeNode or TreeNodeId because they change
	 * when an objet is added, removed or modified.
	 */
	protected Vector<LyObjectId>						 _openNode = null;

	/** Used to store the selected treeNode. */
	protected Vector<LyComponent>						 _selectedNodes = null;

	/** Boolean set when the idle must be ignored. */
	protected boolean			_ignoreIdle = false;

	/** Form field identifier: flag to indicate if the tree must be fully discovered or
	 * discovered on node open action. */
	public static final String DISCOVER_ON_CLICK = "_discoverOnclick".intern();

	/** Allows fast access to a list of treenode from a LyObjectId */
	protected Hashtable<LyObjectId, Vector<LyComponent>> _objectIdToTreeNode = null;

	/** Allows fast access from a treeNode Id to a LyTreeNode */
	protected Hashtable<String, LyComponent>			 _treeNodeIdToTreeNode = null;

	/** Allows fast access to the depth level from a treeNode id */
	protected Hashtable<String, String>					 _treeNodeIdToLevel = null;

	/** Allows fast access to a list of tree element from a level*/
	protected Hashtable<String, _LyTreeElement>			 _level;
	/** Allows fast access to a list of level from a classInfo */
	protected Hashtable<LyClassInfo, Vector<String>>	 _classes;
	/** Allows fast access to the sort associated to a level */
	protected Hashtable<String, LySort>					 _sorts;
	/** Allows fast access to the filter associated to a level */
	protected Hashtable<String, LyFilter>				 _filters;
	/** Indicate if the tree is build on the fly */
	protected boolean			_discoverOnClick;
	/** separator id for treeNode. */
	public final static String							 SEP_ID = "".intern();

	/** Close state for a node. */
	public static final Integer							 STATUS_NONE = new Integer(0);

	/** Open state for a node. */
	public static final Integer							 STATUS_OPENED = new Integer(1);

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo classInfo for this controller (unused)
	 * @param context   current context (may be null if the view is not contextual to a list of
	 *                  objects).
	 */
	public LyTreeController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent, action, context);
		_discoverOnClick = Boolean.valueOf((String)getParameter(DISCOVER_ON_CLICK, true));
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyTreeBehaviorInterface.class;
	}

	/**
	 * Initialize Tree controller We aspect the following parameters: _classInfoA _classInfoAA
	 * _classInfoAB ... _classInfoB ... _classInfoN If a parameter is _classInfo, we consider it as
	 * _classInfo1 and all are: _classInfo1A
	 */
	protected void initTreeController()
	{
		_treeBehaviorInterface = (LyTreeBehaviorInterface)getBehavior();

		LyClassInfo rootClass = (LyClassInfo)getParameter(LyAction.PARAM_CLASS, true);
		LyFilter    filterRootClass = (LyFilter)getParameter(LyAction.PARAM_FILTER, true);
		LySort	    sortRootClass = (LySort)getParameter(LyAction.PARAM_SORT + "0", true);

		if (rootClass != null)
		{
			_LyTreeElement root = new _LyTreeElement(rootClass, null, "1");

			addLevel(root);
			addClasses(rootClass, "1");

			if (sortRootClass != null)
				addSort(sortRootClass, "1");

			if (filterRootClass != null)
				addFilter(filterRootClass, "1");

			initTreeClassInfo("1", "", root);
		}
		else
		{
			initTreeClassInfo("", "", null);

			// MAY Get the classinfos from the behavior
			if ((_classes == null) && (_treeBehaviorInterface != null))
			{
				LyInfoList<LyClassInfo> list = ((LyTreeBehaviorInterface)_treeBehaviorInterface)
						.getClassInfos(this);

				if (list != null)
				{
					for (int i = 0; i < list.getSize(); i++)
						addClasses(list.getInfo(i), "1");
				}
			}
		}

		_refreshTime = -1;
	} // end method initTreeController

	/**
	 * Initialize Tree class Info recurse method.
	 *
	 * @param level  tree level
	 * @param depth  Current Deep in parameter
	 * @param parent parent
	 */
	protected void initTreeClassInfo(String level, String depth, _LyTreeElement parent)
	{
		String   current_level, end;
		boolean  next = true;
		int		 i = 1; //, depth;
		Object   param;
		LyFilter filter;
		LySort   sort;

		while (next)
		{
			// current level = _classInfo<i>
			current_level = level + String.valueOf(i);
			end = depth + String.valueOf(i);
			param = getParameter(LyAction.PARAM_CLASS + end, true);
			sort = (LySort)getParameter(LyAction.PARAM_SORT + end, true);
			filter = (LyFilter)getParameter(LyAction.PARAM_FILTER + end, true);

			if (param != null)
			{
				LyClassInfo    classInfo = (LyClassInfo)param;

				_LyTreeElement treeElement = new _LyTreeElement(classInfo, parent, current_level);

				addLevel(treeElement);
				addClasses(classInfo, current_level);

				if (sort != null)
					addSort(sort, current_level);

				if (filter != null)
					addFilter(filter, current_level);

				initTreeClassInfo(current_level, end, treeElement);
			}
			else
				next = false;

			i++;
		}
	} // end method initTreeClassInfo

	/**
	 * add a level
	 * @param element the element to add
	 */
	protected void addLevel(_LyTreeElement element)
	{
		if (element == null)
			return;

		if (_level == null)
			_level = new Hashtable<String, _LyTreeElement>();

		_level.put(element.getLevel(), element);
	}

	/**
	 * Internal method to add a classInfo at the specified level.
	 *
	 * @param classInfo reference to the classInfo
	 * @param level     level as a string ("1" , ...) as in meta-model
	 */
	protected void addClasses(LyClassInfo classInfo, String level)
	{
		if (_classes == null)
			_classes = new Hashtable<LyClassInfo, Vector<String>>();

		if (classInfo.isVirtual())
			classInfo = classInfo.getRealClass();

		Vector<String> levels;

		if (_classes.containsKey(classInfo))
			levels = _classes.get(classInfo);
		else
			levels = new Vector<String>();

		if (!levels.contains(level))
			levels.addElement(level);

		_classes.put(classInfo, levels);
	}

	/**
	 * Internal method to associate a sort at the specified level.
	 *
	 * @param sort  reference to the sort
	 * @param level level as a string ("1" , ...) as in meta-model
	 */
	protected void addSort(LySort sort, String level)
	{
		if (_sorts == null)
			_sorts = new Hashtable<String, LySort>();

		_sorts.put(level, sort);
	}

	/**
	 * Internal method to associate a filter at the specified level.
	 *
	 * @param filter  reference to the filter
	 * @param level level as a string ("1" , ...) as in model
	 */
	protected void addFilter(LyFilter filter, String level)
	{
		if (_filters == null)
			_filters = new Hashtable<String, LyFilter>();

		_filters.put(level, filter);
	}

	/**
	 * Get all levels that have the given classinfo.
	 * @param classInfo a classInfo
	 * @return  a list of levels
	 */
	public Vector<String> getLevels(LyClassInfo classInfo)
	{
		if (_classes == null)
			return null;

		Vector<String> result = _classes.get(classInfo);

		if ((result == null) && (classInfo.getSuperClass() != null))
			result = getLevels(classInfo.getSuperClass());

		return result;
	}

	/**
	 * Get a list of available classInfo for the sons at the given level.
	 *
	 * @param  level Tree level
	 * @return a list of classInfo
	 */
	public LyInfoList<LyClassInfo> getSonsClassInfos(String level)
	{
		if (_level == null)
			return null;

		_LyTreeElement treeElement = _level.get(level);

		return ((treeElement == null) ? null : treeElement.getSons());
	}

	/**
	 * Get the classInfo associated with a given level.
	 *
	 * @param  level the level
	 * @return the classInfo
	 */
	public LyClassInfo getClassInfoFromLevel(String level)
	{
		_LyTreeElement treeElement = (_level == null ? null : _level.get(level));

		if (treeElement != null)
			return treeElement.getClassInfo();

		if (level.endsWith("*"))
			return (LyClassInfo)getParameter(LyAction.PARAM_CLASS + level, true);
		else
			return null;
	}

	/**
	 * Get the sort associated with a given level.
	 *
	 * @param  level     the level
	 * @return the sort
	 */
	public LySort getSortFromLevel(String level)
	{
		if ((_sorts != null) && _sorts.containsKey(level))
			return _sorts.get(level);
		else
			return (LySort)getParameter(LyAction.PARAM_SORT + level, true);
	}

	/**
	 * Get the filter associated with a given level.
	 *
	 * @param  level     the level
	 * @return the filter
	 */
	public LyFilter getFilterFromLevel(String level)
	{
		if ((_filters != null) && _filters.containsKey(level))
			return _filters.get(level);
		else
			return (LyFilter)getParameter(LyAction.PARAM_FILTER + level, true);
	}

	/**
	 * Retrieves a level from a component id.
	 *
	 * @param  compId the id of the component
	 * @return the level of the given component
	 */
	public String getLevel(String compId)
	{
		if ((compId == null) || (_treeNodeIdToLevel == null))
			return null;

		return _treeNodeIdToLevel.get(compId);
	}

	/**
	 * Return a list of all classes in the tree.
	 *
	 * @return the list of classInfo
	 */
	public LyInfoList<LyClassInfo> getClassInfos()
	{
		LyInfoList<LyClassInfo> list = new LyInfoList<LyClassInfo>();

		if (_classes != null)
		{
			for (Enumeration<LyClassInfo> e = _classes.keys(); e.hasMoreElements();)
			{
				LyClassInfo classInfo = e.nextElement();

				list.addInfo(classInfo);
			}
		}
		return list;
	}

	/**
	 * Accessor to the list of root classInfos of the tree.
	 *
	 * @return the list of all root classes (may be null)
	 */
	public LyInfoList<LyClassInfo> getRootClassInfos()
	{
		LyInfoList<LyClassInfo> list = null;
		LyClassInfo			    classInfo;
		String				    level;
		int					    i = 1;

		do
		{
			level = String.valueOf(i);
			classInfo = getClassInfoFromLevel(level);

			if (classInfo != null)
			{
				if (list == null)
					list = new LyInfoList<LyClassInfo>();
				else if (list.contains(classInfo))
					break;

				list.addInfo(classInfo);
			}
			i++;
		}
		while (classInfo != null);

		return list;
	}

	/**
	 * Empty the tree (remove all nodes).
	 */
	public void clearTreeContent()
	{
		if (_component == null)
			return;

		boolean isSorted = true;

		// Retrieve internal tree component
		LyTree  tree = getTreeComponent();

		tree.setTreeNode(null);

		if (_workSpace != null)
			_workSpace.free(this);

		LyInfoList<LyClassInfo> classInfos = null;

		if (_treeBehaviorInterface != null)
		{
			classInfos = ((LyTreeBehaviorInterface)_treeBehaviorInterface).getClassInfos(this);
			isSorted = ((LyTreeBehaviorInterface)_treeBehaviorInterface).isSorted(this);
		}

		if (classInfos == null)
			classInfos = getClassInfos();

		_workSpace = new LySimpleWorkSpace(_session, true);
		_workSpace.setNotifyCreation(true);
		_workSpace.setNotifyChange(true);
		_workSpace.setSorted(isSorted);


		int		    i, n = classInfos.getSize();
		LyClassInfo classInfo;

		for (i = 0; i < n; i++)
		{
			classInfo = classInfos.getInfo(i);
			_workSpace.registerClassInfo(classInfo);
		}

		_workSpace.addListener(this);

		_objectIdToTreeNode = new Hashtable<LyObjectId, Vector<LyComponent>>();
		_treeNodeIdToTreeNode = new Hashtable<String, LyComponent>();
		_treeNodeIdToLevel = new Hashtable<String, String>();

		_selections = null;
	} // end method clearTreeContent

	/**
	 * Reset the content of the view (Nodes).
	 */
	public void updateContent()
	{
		if (_component == null)
			return;

		if (_treeBehaviorInterface != null)
		{
			// store current selections
			LySimpleObjectList currentSelections = _selections;

			// Clear current list of selected tree node
			_selectedNodes = null;

			clearTreeContent();
			// Retrieve internal tree component
			LyTree  tree = getTreeComponent();
			boolean idle = tree.isIdle();

			// Idle mode
			tree.setIdle(true);

			// Get root node
			Object     appObject = _treeBehaviorInterface.getTreeRootObject(this);

			// build tree content
			LyTreeNode treeNode = addTreeNode(null, appObject, "");

			if (treeNode != null)
				tree.setTreeNode(treeNode);

			// Update the open node list, removing nodes that are not yet
			// in the tree.
			if (_openNode != null)
			{
				LyObjectId openNodeId;
				Vector<LyComponent>     openNodes;

				for (int i = _openNode.size() - 1; i >= 0; i--)
				{
					openNodeId = _openNode.elementAt(i);
					openNodes = getTreeNodesFromObjectId(openNodeId);
					if (openNodes == null)
						_openNode.removeElementAt(i);
				}

				if (_openNode.size() == 0)
					_openNode = null;
			}

			// If nothing opened, open at least the root node
			if ((_openNode == null) && (appObject instanceof LyObject))
				expandNode(treeNode, true);

			//FFT00544
			// Restore selection
			if ((currentSelections != null) && (currentSelections.getSize() > 0))
			{
				if (_selections != null)
					currentSelections.addList(_selections);
				_selections = null;
				_selections = new LySimpleObjectList();

				for (int i = 0; i < currentSelections.getSize(); i++)
				{
					LyObject selectedObject = currentSelections.getObject(i);

					if (selectedObject != null)
					{
						Vector<LyComponent> treeNodes = getTreeNodesFromObjectId(selectedObject.getObjectId());

						if (treeNodes != null)
						{
							for (int j = 0; j < treeNodes.size(); j++)
							{
								treeNode = (LyTreeNode)treeNodes.elementAt(j);
								if (treeNode != null)
									selectObject(selectedObject, treeNode.getId(), false);
							}
						}
					}
				}
			}

			// Idle mode
			tree.setIdle(idle);
		}
	} // end method updateContent

	/**
	 * Free this controller.
	 *
	 * @param controller parent controller
	 */
	public void free(LyController controller)
	{
		if (_parent != controller)
			return;

		_treeBehaviorInterface = null;

		if (_workSpace != null)
			_workSpace.free(this);

		// Free memory (gc)
		_selections = null;

		if (_objectIdToTreeNode != null)
		{
			_objectIdToTreeNode.clear();
			_objectIdToTreeNode = null;
		}

		if (_treeNodeIdToTreeNode != null)
		{
			_treeNodeIdToTreeNode.clear();
			_treeNodeIdToTreeNode = null;
		}

		if (_treeNodeIdToLevel != null)
		{
			_treeNodeIdToLevel.clear();
			_treeNodeIdToLevel = null;
		}

		// Retrieve internal tree component
		if (_component != null)
		{
			LyTree tree = getTreeComponent();

			if (tree != null)
				tree.free();
		}

		super.free(controller);
	} // end method free

	/**
	 * Update context client. This method update the context of the client area. The object list
	 * passed to the client depends of selection and action. If there is a non empty selection, this
	 * method send the selection. If there is no selection or an empty selection: If the action of
	 * the context client is null or has a MIN value equals to 0, this method sends an empty list,
	 * otherwise, it sends the content of the controller.
	 * @return the client context
	 */
	protected LyObjectList getClientContext()
	{
		LyAreaController contextClient = getContextClient();

		if (contextClient == null)
			return null;

		LyObjectList selection = getSelection();
		LyObjectList clientContext;

		if ((selection == null) || (selection.getSize() == 0))
		{
			LyAction action = contextClient.getAction();

			if ((action != null) && (action.getMin() == 0))
				clientContext = new LySimpleObjectList();
			else if ((action != null) && (action.getMin() < 0))
				clientContext = null;
			else
			{
				// We don't need to send all datas, just the first level and
				// we ask the behavior to compute the list...
				clientContext = ((LyTreeBehaviorInterface)_treeBehaviorInterface).getSons(null,
						null, null, this);
				if (clientContext != null)
					clientContext.setSorted(true);
			}
		}
		else
		{
			clientContext = selection;
			clientContext.setSorted(true);
		}

		LyActionBehaviorInterface behavior = (LyActionBehaviorInterface)getBehavior();

		clientContext = behavior.getClientContext(this, clientContext);
		return clientContext;
	} // end method getClientContext

	/**
	 * Create a node subtree for the parent node.
	 *
	 * @param  parentNode    of the subtree, or null if create subtree a root level
	 * @param  appObject the application object which is the node of the subtree, or null to fetch
	 *                   root component
	 * @param level deepth level
	 * @return subtreenode associated with Application object and its own subtree nodes
	 */
	public LyTreeNode addTreeNode(LyTreeNode parentNode, Object appObject, String level)
	{
		if (_treeBehaviorInterface == null)
			return null;

		// compute node Id
		String     nodeId = computeId(appObject, parentNode);

		if (nodeId == null)
		    return null;

		LyTreeNode treeNode = (LyTreeNode)_treeNodeIdToTreeNode.get(nodeId);

		if (treeNode != null)
			return treeNode;

		// Attempt to build the node from the tree controller
		treeNode = _treeBehaviorInterface.buildTreeNode(appObject, this);

		if (treeNode == null)
			return null;

		// set TreeNode Id
		treeNode.setId(nodeId);

		// add new node to his parent.
		if (parentNode != null)
			parentNode.addTreeNode(treeNode);

		//Register treeNodeId to TreeNode
		_treeNodeIdToTreeNode.put(nodeId, treeNode);
		// Set all others parameters
		_treeBehaviorInterface.updateTreeNode(appObject, treeNode, this);

		if (level != null)
			_treeNodeIdToLevel.put(nodeId, level);

		if (appObject instanceof LyObject)
		{
			LyObject object = (LyObject)appObject;

			addTreeNodeFromObjectId(object.getObjectId(), treeNode);
			if (object.isSelected())
				selectObject(object, nodeId, false);

			_workSpace.addObject(object);
		}

		if (isDiscoverOnClick() && (treeNode != null))
			treeNode.setExpandable(true);

	//We have just added a new node. In some cases, we have to insert the hierarchy of the new node :
	//  - if we don't use the discoveryOnclick : we want to load the whole tree in memory (so the
	//        _discoverOnclick parameter is not set) we call recursively the addTreeNode
	//  - if the node we have added is expanded (in order to see its children) : we discover the nodes
	//         already open (for example, when the user switch from a tab with a tree to another tab, and then returns to
	//         the tab containing the three
	//  - if we add the root node, it may not be expanded because not yet filled with hierrarchy, so we have to fill with
	//         children (first level of the three)

	Object     rootId = _treeBehaviorInterface.getTreeRootObject(this);
	LyTreeNode rootTreeNode = (LyTreeNode)_treeNodeIdToTreeNode.get(rootId);

	if(!isDiscoverOnClick()
	   || ((isDiscoverOnClick()
		&& (getExpandedStatusFromNode(treeNode) == LyTreeController.STATUS_OPENED
		    || (treeNode.equals(rootTreeNode))))))
	{
		// get workspace for current list of classInfo
		LyObjectList sons = null;

  		if (appObject instanceof LyObject)
  			sons = ((LyTreeBehaviorInterface)_treeBehaviorInterface).getSons((LyObject)appObject,
  					treeNode, level, this);
  		else
  			sons = ((LyTreeBehaviorInterface)_treeBehaviorInterface).getSons(null, treeNode, level,
  					this);

  		// We look at all the node Childs
  		if ((sons != null) && (sons.getSize() > 0))
  		{
  			//add Tree Node on sons
  			int		 i, n = sons.getSize();
  			LyObject object;

  			if (level == null)
  				level = "";

  			String           currentLevel;
  			int              size = level.length();
  			Vector<String>   levels;
  			int              l, nl;
  			String           temp;

  			for (i = 0; i < n; i++)
  			{
  				// Add the child as a subtree node at current level
  				object = sons.getObject(i);
  				currentLevel = level;

  				levels = getLevels(object.getClassInfo());
  				if (levels != null)
  				{
  					for (l = 0, nl = levels.size(); l < nl; l++)
  					{
  						temp = levels.elementAt(l);
  						if ((temp.length() == (size + 1))
  							&& temp.substring(0, size).equals(level))
  						{
  							currentLevel = temp;
  							break;
  						}
  					}
  				}

  				addTreeNode(treeNode, object, currentLevel);
  			}
		  }
		}

		return treeNode;
	} // end method addTreeNode

	/**
	 * Compute Id of the given object.
	 *
	 * @param  appObject  object
	 * @param  parentNode parent node for this object
	 * @return the id of the object
	 */
	protected String computeId(Object appObject, LyTreeNode parentNode)
	{
		String id;

		if (appObject instanceof LyObject)
			id = ((LyObject)appObject).getId();
		else if (appObject instanceof LyObjectId)
			id = ((LyObjectId)appObject).getFullString();
		else
			id = null;

		if (parentNode != null)
			id = parentNode.getId() + SEP_ID + id;
		else if (_treeBehaviorInterface != null)
		{
		    Object    rootObject = _treeBehaviorInterface.getTreeRootObject(this);

			id = (rootObject == null ? null : rootObject.toString());
		}
		else if (appObject == null)
			id = "_tree".intern();

		return id;
	}

	/**
	 * Invoked when an action occurs.
	 *
	 * @param demand menu demand
	 */
	public void menuDemand(LyMenuDemand demand)
	{
		if (demand == null)
			return;

		if (demand.getSource().equals(getId() + "_tree"))
		{
			String	    key = demand.getItem();
			LyObject    object = getObjectFromTreeNodeId(key);
			LyPopupMenu popupMenu = buildPopupMenu(demand, object);

			// Call the behavior to complete popup menu
			LyBehaviorInterface behavior = getBehavior();

			if (behavior instanceof LyPanelBehaviorInterface)
				((LyPanelBehaviorInterface)behavior).completePopupMenu(this, popupMenu);

			if (popupMenu == null)
				return;

			LyTree tree = getTreeComponent();

			if (tree == null)
				// An error occured, just return
				return;

			popupMenu.setController(getId());
			popupMenu.setEnvironment(tree.getEnvironment());
			tree.setPopupMenu(popupMenu);
		}
		else
			super.menuDemand(demand);
	}

	/**
	 * Build a contextual menu with the given demand based on the given object.
	 * @param demand the menu demand
	 * @param object the associated object
	 * @return the generated popup menu
	 */
	protected LyPopupMenu buildPopupMenu(LyMenuDemand demand, LyObject object)
	{
		if (demand == null)
			return null;

		String   key = demand.getItem();
		LyAction action = getAction();

		if (action.hasMark(LyAction.MARK_NO_MENU))
			return null;

		if (key == null)
			key = "";

		LyPopupMenu popupMenu = null;
		String	    nodeLevel = _treeNodeIdToLevel.get(key);
		LyClassInfo classInfo = null;
		LySession   session = getSession();

		if (object != null)
			classInfo = object.getClassInfo();
		else if (nodeLevel != null)
			classInfo = getClassInfoFromLevel(nodeLevel);

		if (classInfo == null)
			popupMenu = new LyPopupMenu();
		else
			popupMenu = buildPopupMenu(demand, classInfo, true);

		LyInfoList<LyClassInfo>				  sons =
			((LyTreeBehaviorInterface)_treeBehaviorInterface).getSonsClassInfos(classInfo, nodeLevel, this);
		LyInfoDataList<LyAction, LyClassInfo> actions = getExtraPopupMenuActions(classInfo, sons);

		if (actions != null)
		{
			LyItem	    item;
			boolean     enabled;
			LyClassInfo son;
			int		    i, n = actions.getSize();
			String	    id, image, name;

			for (i = 0; i < n; i++)
			{
				action = actions.getInfo(i);
				son = actions.getData(i);
				id = action.getId() + LyController.ID_SEP + son.getId();

				// Add in Popup menu
				if (popupMenu == null)
					popupMenu = new LyPopupMenu();
				else
				{
					item = (LyItem)popupMenu.getComponent(id);
					if (item != null)
						popupMenu.removeItem(item);
				}

				enabled = session.enableAction(action, this, _selections, false);

				name = (action.getClassInfo() == son ? action.getName()
						: action.getName() + " : " + son.getName());
				item = new LyItem(id, name);
				item.setEnabled(enabled);
				item.setController(getId());
				item.setDialog(action.hasMark(LyAction.MARK_DIALOG));
				image = getMenuItemImage(action);
				item.setImage(image);

				popupMenu.insertItem(item, i);
			}

			if ((popupMenu.getComponentCount() > n)
					&& (popupMenu.getComponent(n) != null)
					&& (popupMenu.getComponent(n).getId() != LyComponent.SEP_ID))
				popupMenu.insertSep(n);
		}

		if (popupMenu == null)
			return null;

		// Coordinates
		int[] p = demand.getPosition();

		popupMenu.setX(p[0]);
		popupMenu.setY(p[1]);

		return popupMenu;
	} // end method buildPopupMenu

	/**
	 * Get the actions to build a popup menu
	 * @param classInfo the classinfo of the selected node
	 * @param sons a list of classInfo (sons)
	 * @return a list of actions
	 */
	protected LyInfoDataList<LyAction, LyClassInfo> getExtraPopupMenuActions(LyClassInfo classInfo,
		LyInfoList<LyClassInfo> sons)
	{
		LyInfoDataList<LyAction, LyClassInfo> actions = null;
		LySession							  session = getSession();
		LyAction							  action;

		if (sons != null)
		{
			LyClassInfo son;
			int		    i, ni, j, nj;

			for (i = 0, ni = sons.getSize(); i < ni; i++)
			{
				son = sons.getInfo(i);

				for (j = 0, nj = son.getActionCount(); j < nj; j++)
				{
					action = son.getAction(j);
					if (action.getRoot().getId().equals(LyAction.ID_CREATE)
							&& (action.getMin() == 0)
							&& !son.isAbstract())
					{
						// Skip actions which should not be here
						if (action.hasMark(LyAction.MARK_HIDDEN))
							continue;
						if (!session.validateAction(action, this))
							continue;
						if (action.getMax() != 0)
							continue;

						if (actions == null)
							actions = new LyInfoDataList<LyAction, LyClassInfo>(true);
						actions.addInfo(action, son);
					}
				}
			}
		}

		LyBehaviorInterface behavior = getBehavior();

		if (behavior instanceof LyPanelBehaviorInterface)
			actions = ((LyTreeBehaviorInterface)behavior).getExtraPopupMenuActions(this, actions,
					classInfo, sons);

		return actions;
	} // end method getExtraPopupMenuActions

	/**
	 * Invoked when an action occurs. (LyDemandListener interface) This method handles: opening and
	 * closing node, action
	 *
	 * @param demand action demand
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (_component == null)
			return;

		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			String   source = demand.getSource();
			String[] parameters = LyString.stringToList(source, ID_SEP);
			String   id = parameters[0].intern();

			if (id.equals(getId() + "_tree"))
			{
				// Default action on an item of the tree
				String   key = demand.getValue();
				LyObject object = getObjectFromTreeNodeId(key);

				if (object != null)
				{
					invokeDefaultAction(object);
					return;
				}
			}
			else
			{
				// Menu or action bar
				LyApplication application = getApplication();
				LyInfo		  info = application.getInfo(id);

				if (info instanceof LyAction)
				{
					LyAction    action = (LyAction)info;
					boolean     hasSelection = (_selections != null)
							&& (_selections.getSize() != 0);
					LyClassInfo classInfo = null;

					if (parameters.length > 1)
						classInfo = application.getClassInfo(parameters[1]);

					if ((action.getMin() > 0) || (hasSelection && (action.getMax() > 0)))
						// Call action on selection
						invokeAction(action, demand);
					else
					{
						LyController controller = invokeAction(action, classInfo, null, false);

						if (controller == null)
							return;

						// Force construction of the viewController.
						controller.getViewController();

						if (hasSelection
								&& (controller instanceof LyContextInterface))
						{
							LySimpleWorkSpace context = new LySimpleWorkSpace(getSession(),
									_selections, true);

							((LyContextInterface)controller).setContext(context);
							if (context.getListenerCount() == 0)
								context.free(null);
						}

						if (controller != null)
							controller.showView();
					}
				}
				else
					super.actionDemand(demand);
			}

			// Update tools and messages bars.
			updateContextClient();
			updateTools();
			updateMessages();
		}
		else if (demand.getType() == LyActionDemand.TYPE_EXPAND)
		{
			boolean    isOpen;
			String     key = demand.getValue();
			int		   i = key.indexOf('|');
			LyTree     tree = getTreeComponent();
			String     open;
			LyTreeNode treeNode;

			if (i >= 0)
			{
				open = key.substring(i + 1);
				key = key.substring(0, i);
				treeNode = (LyTreeNode)_treeNodeIdToTreeNode.get(key);
			}
			else
			{
				treeNode = (LyTreeNode)_treeNodeIdToTreeNode.get(key);
				open = !treeNode.isOpen() + "";
			}

			if (treeNode == null)
				treeNode = getTreeComponent().getTreeNode();

			isOpen = open.equals("true");

			expandNode(treeNode, isOpen);

			// Activate the node (scrolling in web)
			tree.setActiveNode(treeNode.getId());
		}
		else if (demand.getType() == LyActionDemand.TYPE_DELETE)
		{
			String   key = demand.getValue();
			LyObject object = getObjectFromTreeNodeId(key);

			if (object != null)
			{
				LySimpleObjectList list = new LySimpleObjectList(object);

				invokeDeleteAction(list, object.getClassInfo());

				// Update tools and messages bars.
				updateContextClient();
				updateTools();
				updateMessages();
			}
		}
		else if (demand.getType() == LyActionDemand.TYPE_STATUS)
		{
			String     key = demand.getValue();
			LyTreeNode treeNode = (LyTreeNode)_treeNodeIdToTreeNode.get(key);

			if (_treeBehaviorInterface != null)
				_treeBehaviorInterface.validityModified(treeNode, this);
		}
		else if (demand.getType() == LyActionDemand.TYPE_INSERT)
		{
			LyClassInfo classInfo = (LyClassInfo)getParameter("_classInfo", true);

			invokeCreateAction(classInfo);
		}

		// Update the last modification time.
		setLastModified(System.currentTimeMillis());
	} // end method actionDemand

	/**
	 * Invoked when a Drag & Drop occurs. (LyDemandListener interface) This method handles: opening
	 * and closing line, moving an interval, action
	 *
	 * @param demand drag-drop demand
	 */
	public void dragDropDemand(LyDragDropDemand demand)
	{
		if (demand.getType() == LyDragDropDemand.TYPE_DRAG_START)
		{
			LyObject draggedObject = null;
			String   nodeId = demand.getSource();

			draggedObject = getObjectFromTreeNodeId(nodeId);

			if (_treeBehaviorInterface != null)
				_treeBehaviorInterface.dragStart(this, draggedObject);
		}
		else if (demand.getType() == LyDragDropDemand.TYPE_DROP)
		{
			LyObject  droppedObject = null;
			String    nodeId = demand.getSource();
			LySession session = this.getSession();
			LyObject  draggedObject = (LyObject)session.getValue(LySession.DRAGGED_OBJECT);

			droppedObject = getObjectFromTreeNodeId(nodeId);

			if (_treeBehaviorInterface != null)
				_treeBehaviorInterface.drop(this, droppedObject, draggedObject);
		}
		else if (demand.getType() == LyDragDropDemand.TYPE_DRAG_FINISHED)
		{
			if (_treeBehaviorInterface != null)
				_treeBehaviorInterface.dragFinished(this, Integer.parseInt(demand.getValue()));
		}
		else if (demand.getType() == LyActionDemand.TYPE_INSERT)
		{
			LyClassInfo classInfo = (LyClassInfo)getParameter("_classInfo", true);

			invokeCreateAction(classInfo);
		}

		// Update the last modification time.
		setLastModified(System.currentTimeMillis());
	} // end method dragDropDemand

	/**
	 * This method expand (or close) the given node.
	 *
	 * @param treeNode the tree node to expand or close
	 * @param expand   boolean: true if node must be expanded, false if it must be closed
	 */
	public void expandNode(LyTreeNode treeNode, boolean expand)
	{
		if (_openNode == null)
			_openNode = new Vector<LyObjectId>();

		LyObjectId objectId = getObjectIdFromTreeNodeId(treeNode.getId());

		if (objectId == null)
			return;

		if (expand)
		{
			// the node ask to be opened: open it and had it to the open node list
			treeNode.setOpen(true);
			_openNode.addElement(objectId);
		}
		else
		{
			treeNode.setOpen(false);
			if (_openNode.contains(objectId))
				_openNode.removeElement(objectId);
		}
	}

	/**
	 * Add a reference to a treeNode for the objectId.
	 *
	 * @param objectId LyObjectId of the object associated with this treeNode
	 * @param treeNode LyTreeNode associated to the object Id,
	 */
	public void addTreeNodeFromObjectId(LyObjectId objectId, LyTreeNode treeNode)
	{
		// check if there are already treeNodes attached to this Id
		Vector<LyComponent> treeNodes = _objectIdToTreeNode.get(objectId);

		if (treeNodes == null)
			treeNodes = new Vector<LyComponent>(1);

		// add object to current vector
		if (!treeNodes.contains(treeNode))
			treeNodes.addElement(treeNode);

		_objectIdToTreeNode.put(objectId, treeNodes);
	}

	/**
	 * Remove a reference.
	 *
	 * @param object object previously associated with this treeNode
	 * @param treeNode LyTreeNode associated to the object Id, or null to remove all association for this
	 *                   object Id
	 */
	public void removeTreeNodeFromObjectId(LyObject object, LyTreeNode treeNode)
	{
		// check if there are already treeNodes attached to this Id
		LyObjectId		    objectId = object.getObjectId();
		Vector<LyComponent> treeNodes = _objectIdToTreeNode.get(objectId);

		if (treeNodes == null)
			return;

		if (treeNode == null)
			treeNodes.removeAllElements();
		else
			// remove treenode from current vector
			treeNodes.removeElement(treeNode);

		if (treeNodes.size() == 0)
			_objectIdToTreeNode.remove(objectId);
		else
			_objectIdToTreeNode.put(objectId, treeNodes);


		_workSpace.removeObject(object);
	}

	/**
	 * get all references associated to this objectId.
	 *
	 * @param  objectId LyObjectId of the object
	 * @return a vector of treeNodes associated with this objectId or null if none
	 */
	public Vector<LyComponent> getTreeNodesFromObjectId(LyObjectId objectId)
	{
		return _objectIdToTreeNode.get(objectId);
	}

	/**
	 * Method used to get the object associated with a TreeNode, using the node Id.
	 *
	 * @param  treeNodeId the node id
	 * @return the object corresponding to the key
	 */
	public LyObjectId getObjectIdFromTreeNodeId(String treeNodeId)
	{
		if (treeNodeId.indexOf(SEP_ID) < 0)
		{
			// Check if root node is an object id
			if (_treeBehaviorInterface != null)
			{
				// Get root node
				Object appObject = _treeBehaviorInterface.getTreeRootObject(this);

				if (appObject instanceof LyObject)
				{
					String nodeId = computeId(appObject, null);

					if (nodeId.equals(treeNodeId))
						return ((LyObject)appObject).getObjectId();
				}
			}

			return null;
		}

		String[]   params = LyString.stringToList(treeNodeId, SEP_ID);
		LyObjectId objectId = new LyObjectId(getApplication(), params[params.length - 1]);

		return objectId;
	}

	/**
	 * Method used to get the object associated with a TreeNode, using the node Id.
	 *
	 * @param  treeNodeId the node id
	 * @return the object corresponding to the key
	 */
	public LyObject getObjectFromTreeNodeId(String treeNodeId)
	{
		if (treeNodeId == null)
			return null;

		if (treeNodeId.indexOf(SEP_ID) < 0)
		{
			// Check if root node is an object id
			if (_treeBehaviorInterface != null)
			{
				// Get root node
				Object appObject = _treeBehaviorInterface.getTreeRootObject(this);

				if (appObject instanceof LyObject)
				{
					String nodeId = computeId(appObject, null);

					if (nodeId.equals(treeNodeId))
						return ((LyObject)appObject);
				}
			}

			return null;
		}

		String[]   params = LyString.stringToList(treeNodeId, SEP_ID);

		LyObjectId objectId = new LyObjectId(getApplication(), params[params.length - 1]);

		return _workSpace.getObject(objectId);
	} // end method getObjectFromTreeNodeId

	/**
	 * This method returns a vector of Strings Each string represent an object The first element is
	 * the tree root. The last element is the LyObjectId (as a string) of the given treenode All
	 * element between are grandparents and parent of the treenode
	 *
	 * @param  treeNode the given treeNode
	 * @return the vector of Id
	 */
	public Vector<String> getBranchFromTreeNode(LyTreeNode treeNode)
	{
		String key = treeNode.getId();

		if (key.indexOf(SEP_ID) < 0)
			return null;

		String[]	   params = LyString.stringToList(key, SEP_ID);

		Vector<String> result = new Vector<String>();

		for (int i = 0; i < params.length; i++)
			result.addElement(params[i]);

		return result;
	}

	/**
	 * Invoked when an selection occurs.
	 *
	 * @param demand selection demand
	 */
	public void selectionDemand(LySelectionDemand demand)
	{
		if (_component == null)
			return;

		// Unselect all the objects
		String   key = null;
		LyObject object;
		LyTree   tree = getTreeComponent();
		boolean  idle = tree.isIdle();

		// Idle mode
		tree.setIdle(true);

		// Find selected objects
		LySimpleObjectList selection = new LySimpleObjectList();
		int				   i, n = demand.getSelectionSize();

		// Get key from selected treeNode
		Vector<String>     keyList = new Vector<String>();

		for (i = 0; i < n; i++)
		{
			key = demand.getSelectionId(i);
			keyList.add(key);
			object = getObjectFromTreeNodeId(key);
			if (object != null)
				selection.addObject(object);
			else
			{
				// Display a message (the object was deleted) except if
				// selection on the root node (SEP_ID not in the key).
				if ((n == 1) && (key != null) && (key.indexOf(SEP_ID) > 0))
				{
					LyEnvironment env = getEnvironment();
					String		  message = env.getMessage("LY_ER_OBJECT_NOT_IN_VIEW");

					getViewManager().showError(tree, message);

					// Refresh tree
					updateContent();
					tree.setVisible(true);
					break;
				}
			}
		}

		// first, we have to deselect all selected items
		if (_selectedNodes != null)
			selectAllObjects(selection, false);

		_selectedNodes = new Vector<LyComponent>();
		_selections = new LySimpleObjectList();

		// Then we handle the new selection
		for (i = 0, n = demand.getSelectionSize(); i < n; i++)
		{
			key = demand.getSelectionId(i);
			object = getObjectFromTreeNodeId(key);
			if ((object == null) || !selection.contains(object))
				// do not select an unselected object
				continue;
			selectObject(object, key, false);
		}

		// Activate the node (scrolling in web)
		tree.setActiveNode(key);

		// Update context, tools and messages bars.
		updateContextClient();
		updateTools();
		updateMessages();
		_session.notifyOutline(this);

		if (_treeBehaviorInterface != null)
			((LyTreeBehaviorInterface)_treeBehaviorInterface).selectionChanged(this);

		tree.setIdle(idle);
		// Update the last modification time.
		setLastModified(System.currentTimeMillis());
	} // end method selectionDemand

	/**
	 * Select an object in the table.
	 *
	 * @param object selected or unselected object
	 * @param select new selection status
	 */
	public void selectObject(LyObject object, boolean select)
	{
		Vector<LyComponent> v = getTreeNodesFromObjectId(object.getObjectId());

		if (v != null)
		{
			int				 i, n = v.size();
			LyGetIdInterface getIdObject;

			for (i = 0; i < n; i++)
			{
				getIdObject = (LyGetIdInterface)v.elementAt(i);
				if (select)
					selectObject(object, getIdObject.getId(), (i == (n - 1)));
				else
					unselectObject(object, getIdObject.getId(), (i == (n - 1)));
			}
		}

		updateContextClient();
		updateTools();
		updateMessages();
	}

	/**
	 * Select the given node.
	 * @param object the object to select
	 * @param nodeId the Id of the node to select
	 * @param updateView true to update the view
	 */
	protected void selectObject(LyObject object, String nodeId, boolean updateView)
	{
		LyComponent component = _treeNodeIdToTreeNode.get(nodeId);

		if (component instanceof LySetSelectedInterface)
			((LySetSelectedInterface)component).setSelected(true);

		if (component != null)
		{
			if (_selectedNodes == null)
				_selectedNodes = new Vector<LyComponent>();

			_selectedNodes.addElement(component);
		}

		if (_selections == null)
			_selections = new LySimpleObjectList();
		_selections.addObject(object);
		object.setSelected(true);
	}

	/**
	 * Unselect the given node.
	 * @param object object to unselect
	 * @param nodeId the Id of the node to unselect
	 * @param updateView true to update the view
	 */
	protected void unselectObject(LyObject object, String nodeId, boolean updateView)
	{
		LyComponent component = _treeNodeIdToTreeNode.get(nodeId);

		if (component instanceof LySetSelectedInterface)
			((LySetSelectedInterface)component).setSelected(false);

		if ((_selectedNodes != null) && (component != null))
			_selectedNodes.removeElement(component);

		if (_selections != null)
			_selections.removeObject(object);
		object.setSelected(false);
	}

	/**
	 * Hide or show the given node.

	 * @param object impacted object
	 * @param nodeId the Id of the node to hide or show
	 * @param hide true to hide the object, false to show
	 * @param updateView true to update the view
	 */
	protected void hideObject(LyObject object, String nodeId, boolean hide, boolean updateView)
	{
		LySetHiddenInterface setHiddenObject = null;
		LyComponent			 component = _treeNodeIdToTreeNode.get(nodeId);

		if (component instanceof LySetHiddenInterface)
			setHiddenObject = (LySetHiddenInterface)component;

		if (setHiddenObject != null)
			setHiddenObject.setHidden(hide);

		// Update the last modification time.
		setLastModified(System.currentTimeMillis());
		setRefreshTime(System.currentTimeMillis());
	}

	/**
	 * Select/unselect all nodes.
	 * @param select true to select all objects, false to unselect
	 */
	public void selectAllObjects(boolean select)
	{
		selectAllObjects(null, select);
	}

	/**
	 * Select/unselect all nodes except given ones.
	 * @param selection a list of object
	 * @param select true to select
	 */
	protected void selectAllObjects(LySimpleObjectList selection, boolean select)
	{
		if (!select)
		{
			Vector<LyComponent> 	    selectedNodes = _selectedNodes;  // copy to avoid reset of _selectedNodes in the loop
			int		    i, n = ((selectedNodes == null) ? 0 : selectedNodes.size());
			LyComponent component;

			for (i = 0; i < n; i++)
			{
				component = selectedNodes.elementAt(i);
				if (component instanceof LySetSelectedInterface)
					((LySetSelectedInterface)component).setSelected(false);
			}
		}

		LySimpleObjectList objects = null;
		LyObject		   object;
		int				   i, no;

		if (_selections == null)
			no = 0;
		else
		{
			objects = new LySimpleObjectList(_selections);
			no = objects.getSize();
		}

		for (i = 0; i < no; i++)
		{
			object = objects.getObject(i);
			if (select)
			{
				if ((selection == null) || selection.contains(object))
					object.setSelected(true);
			}
			else
			{
				if ((selection == null) || !selection.contains(object))
					object.setSelected(false);
			}
		}

		_selectedNodes = null;
		_selections = null;
	} // end method unselectAllObjects

	/**
	 * Get the selected treeNodes.
	 *
	 * @return A vector of selected treeNodes
	 */
	public Vector<LyComponent> getSelectedTreeNodes()
	{
		return _selectedNodes;
	}

	/**
	 * Gets current expand status of a node.
	 *
	 * @param  getIdObject view component
	 * @return status of the view component (Opened -> STATUS_OPENED or closed -> STATUS_NONE)
	 */
	public Integer getExpandedStatusFromNode(LyGetIdInterface getIdObject)
	{
		LyObjectId object = getObjectIdFromTreeNodeId(getIdObject.getId());

		if ((_openNode != null) && (_openNode.contains(object)))
			return STATUS_OPENED;
		else
			return STATUS_NONE;
	}

	/**
	 * Get content of the tree.
	 *
	 * @return the list of objects in the node
	 */
	public LyObjectList getContent()
	{
		return ((_workSpace == null) ? null : new LySimpleObjectList(_workSpace, true));
	}

	/**
	 * Add a node (to a parent node).
	 *
	 * @param parentNode parent Node
	 * @param newObject  object to add in the tree
	 */
	protected void addNode(LyTreeNode parentNode, LyObject newObject)
	{
		if (parentNode == null)
			return;
		if (newObject == null)
			return;

		String           parentLevel = _treeNodeIdToLevel.get(parentNode.getId());
		LyFilter         filter;
		LySort           sort;
		String           level = null;
		Vector<String>   levels = getLevels(newObject.getClassInfo());

		if (levels == null)
			level = "";
		else
		{
			int   i, n = ((levels == null) ? 0 : levels.size());
			int   pll, ll;

			for (i = 0; i < n; i++)
			{
				level = levels.elementAt(i);

				// Add just after the root
				if ((parentLevel == null) && (level != null) && (level.length() == 0))
					break;
				else if ((parentLevel != null) && (level != null))
				{
					ll = level.length();
					pll = parentLevel.length();
					if ((pll == 0) && (ll == 1))
						break;
					if ((ll == (pll + 1))
							&& (level.substring(pll).equals(parentLevel)))
						break;
				}
			}

			if  (level == null)
				return;
		}

		filter = ((LyTreeBehaviorInterface)_treeBehaviorInterface).getFilter(this, level);
		if (filter == null)
			filter = getFilterFromLevel(level);

		sort = ((LyTreeBehaviorInterface)_treeBehaviorInterface).getSort(this, level);
		if (sort == null)
			sort = getSortFromLevel(level);

		if ((filter != null)
				&& !filter.match(newObject))
			return;

		LyTreeNode treeNode = addTreeNode(parentNode, newObject, level);

		if (treeNode == null)
			return;

		boolean isSorted = ((LyTreeBehaviorInterface)_treeBehaviorInterface).isSorted(this);

		if ((sort != null) && isSorted)
			reset();
		else if (isSorted)
		{
			parentNode.removeTreeNode(treeNode);

			String label = treeNode.getLabel();

			if (_treeBehaviorInterface instanceof LyLabelBehaviorInterface)
			{
				LyInfoLabel infoLabel = ((LyLabelBehaviorInterface)_treeBehaviorInterface)
						.getInfoLabel(this, newObject.getClassInfo(), LyInfoLabel.Type.NAME);

				if (infoLabel != null)
					label = infoLabel.getString(newObject);
			}

			if (label == null)
				label = "";

			boolean    inserted = false;
			LyTreeNode currentNode;
			String     currentLabel;
			int		   i, n = parentNode.getTreeNodeCount();

			for (i = 0; i < n; i++)
			{
				currentNode = parentNode.getTreeNode(i);
				if (currentNode == null)
					continue;
				currentLabel = currentNode.getLabel();
				if (currentLabel == null)
					continue;

				if (label.compareTo(currentLabel) < 0)
				{
					// insert treeNode before this one
					parentNode.insertTreeNode(treeNode, i);
					inserted = true;
					break;
				}
			}
			if (!inserted)
				// insert treeNode at the end
				parentNode.addTreeNode(treeNode);
		}
	} // end method addNode

	/* Implementation of LyWorkSpaceListener */

	/**
	 * Called when an object is added in the dynamic object list.
	 *
	 * @param  event event
	 * @return true is demand was successfull.
	 * @see    LyWorkSpaceListener#objectAdded
	 */
	public boolean objectAdded(LyWorkSpaceEvent event)
	{
		if(isIgnoreIdle())
			return false;

		if (isIdle())
			return false;
		if (_treeBehaviorInterface == null)
			return false;

		if (_component == null)
		{
			// Update the last modification time.
			setLastModified(System.currentTimeMillis());
			setRefreshTime(System.currentTimeMillis());
			return true;
		}

		LyObject addedObject = event.getObject();

		if (_objectIdToTreeNode.containsKey(addedObject.getObjectId()))
			return true;

		LyObjectList parents = ((LyTreeBehaviorInterface)_treeBehaviorInterface).getParents(
				addedObject, this);

		if ((parents != null)
				&& (parents.getSize() == 0))
			// No parents available : object is not in tree
			return true;

		LyTree     tree = getTreeComponent();
		boolean    wasIdle = tree.isIdle();
		LyTreeNode parentNode = null;

		if ((parents == null)
				|| isTopLevelClassInfo(addedObject.getClassInfo()))
		{
			// May be added at top level
			tree.setIdle(true);
			parentNode = tree.getTreeNode();
			addNode(parentNode, addedObject);
			tree.setIdle(wasIdle);
			return true;
		}

		// May be added somewhere in the tree
		if ((parents != null) && (parents.getSize() > 0))
		{
			int		 i, n = parents.getSize();
			int		 j, m;
			LyObject currentParent;
			Vector<LyComponent>   parentNodes;

			for (i = 0; i < n; i++)
			{
				currentParent = parents.getObject(i);
				parentNodes = getTreeNodesFromObjectId(currentParent.getObjectId());
				if (parentNodes == null)
					continue;

				if (!tree.isIdle())
					tree.setIdle(true);

				// We must work on a cloned list because the list could
				// be modified when adding the node.
				parentNodes = new Vector<LyComponent>(parentNodes);

				for (j = 0, m = parentNodes.size(); j < m; j++)
				{
					parentNode = (LyTreeNode)parentNodes.elementAt(j);
					addNode(parentNode, addedObject);
				}
			}
		}

		tree.setIdle(wasIdle);

		// Update the last modification time.
		setLastModified(System.currentTimeMillis());
		setRefreshTime(System.currentTimeMillis());

		return true;
	} // end method objectAdded

	/**
	 * Called when an object is removed from the dynamic object list.
	 *
	 * @param  event event
	 * @return true is demand was successfull.
	 * @see    LyWorkSpaceListener#objectRemoved
	 */
	public boolean objectRemoved(LyWorkSpaceEvent event)
	{
		if(isIgnoreIdle())
			return false;

		if (isIdle())
			return false;

		if (_treeBehaviorInterface == null)
			return false;

		if (_component == null)
		{
			// Update the last modification time.
			setLastModified(System.currentTimeMillis());
			setRefreshTime(System.currentTimeMillis());
			return true;
		}

		LyTree  tree = getTreeComponent();
		boolean idle = tree.isIdle();

		// Idle mode
		// If we use setIdle, the window gain focus in web mode -> not good (very very bad in fact)
		tree.setIdle(true);

		LyObject removedObject = event.getObject();

		// FFT00462
		// Remove object from selection
		if (_selections != null)
			_selections.removeObject(removedObject);

		Vector<LyComponent> treeNodes = getTreeNodesFromObjectId(removedObject.getObjectId());

		if (treeNodes == null)
		{
			tree.setIdle(idle);
			return false;
		}

		int		   i, n = treeNodes.size();
		LyTreeNode treeNode;

		for (i = 0; i < n; i++)
		{
			treeNode = (LyTreeNode)treeNodes.elementAt(i);
			if (treeNode != null)
				removeTreeNode(treeNode);
		}

		removeTreeNodeFromObjectId(removedObject, null);

		tree.setIdle(idle);
		// Update the last modification time.
		setLastModified(System.currentTimeMillis());
		setRefreshTime(System.currentTimeMillis());

		// Update controller
		updateContextClient();
		updateTools();
		updateMessages();

		return true;
	} // end method objectRemoved

	/**
	 * Remove given tree node.
	 * @param treeNode treenode to remove
	 */
	protected void removeTreeNode(LyTreeNode treeNode)
	{
		if (treeNode.isSelected())
		{
			treeNode.setSelected(false);
			if (_selectedNodes != null)
				_selectedNodes.removeElement(treeNode);
		}

		LyComponent parent = treeNode.getParent();

		if (parent instanceof LyTreeNode)
			((LyTreeNode)parent).removeTreeNode(treeNode);
		else if (parent instanceof LyTree)
			((LyTree)parent).setTreeNode(null);
		else
			reset();

		_treeNodeIdToTreeNode.remove(treeNode.getId());
		treeNode.free();
		reset();
	}

	/**
	 * Called when an object has changed in the dynamic object list.
	 *
	 * @param  event event
	 * @return true is demand was successfull.
	 * @see    LyWorkSpaceListener#objectChanged
	 */
	public boolean objectChanged(LyWorkSpaceEvent event)
	{
		if(isIgnoreIdle())
			return false;

		if (isIdle())
			return false;
		if (_treeBehaviorInterface == null)
			return false;

		if (_component == null)
		{
			// Update the last modification time.
			setLastModified(System.currentTimeMillis());
			setRefreshTime(System.currentTimeMillis());
			return true;
		}

		LyPanelBehaviorInterface behavior = (LyPanelBehaviorInterface)getBehavior();
		LyObject				 object = event.getObject();
		LyValueSet				 values = event.getValues();

		if (!behavior.needRefresh(this, object, values))
			return false;

		boolean useUpdateContent = false;
		LyTree  tree = getTreeComponent();
		boolean idle = tree.isIdle();

		// for simple modification (label...) we do it efficiently
		// for more complicated modification (parent modification / child evolution)
		//  -> updateContent
		Vector<LyComponent>  treeNodes = getTreeNodesFromObjectId(object.getObjectId());

		if (treeNodes == null)
		{
			if ((values != null)
					&& values.hasMark(LyFieldInfo.MARK_ID))
			{
				LyObjectId oldObjectId = object.getObjectId(values);

				if (oldObjectId != null)
					treeNodes = getTreeNodesFromObjectId(oldObjectId);
			}

			if (treeNodes == null)
			{
				if (!_workSpace.contains(object))
				{
					LyObjectList parents = ((LyTreeBehaviorInterface)_treeBehaviorInterface)
							.getParents(object, this);

					if ((parents != null)
							&& (parents.getSize() > 0))
						return objectAdded(event);

					Vector<String>   levels = getLevels(object.getClassInfo());
					int		 i, n = ((levels == null) ? 0 : levels.size());
					String   level;
					LyFilter filter;

					for (i = 0; i < n; i++)
					{
						level = levels.elementAt(i);
						filter = getFilterFromLevel(level);
						if ((filter != null)
								&& filter.match(object))
							return objectAdded(event);
					}

					return false;
				}

				if (_initialized)
				{
					// Idle mode
					tree.setIdle(true);

					// XXX Find when a best way when object id changes
					reset();

					tree.setIdle(idle);
				}

				// Update the last modification time.
				setLastModified(System.currentTimeMillis());
				setRefreshTime(System.currentTimeMillis());
				return true;
			}
		}

		// Idle mode
		tree.setIdle(true);

		LyTreeNode		 treeNode;
		LyObjectList     parents = ((LyTreeBehaviorInterface)_treeBehaviorInterface).getParents(
				object, this);
		Vector<LyObject> currentParents = new Vector<LyObject>();

		// We should check:
		// - if the name has changed
		// - if level = "" et filter not match
		// - if one of the sons has changed
		// - if parents are different

		int				 i, ni, j, nj;
		String			 level;
		LyComponent		 parent;
		LyFilter		 filter;
		LyTreeNode		 parentNode;

		for (i = 0, ni = treeNodes.size(); i < ni; i++)
		{
			treeNode = (LyTreeNode)treeNodes.elementAt(i);
			parent = treeNode.getParent();
			level = _treeNodeIdToLevel.get(treeNode.getId());
			parentNode = null;
			filter = ((LyTreeBehaviorInterface)_treeBehaviorInterface).getFilter(this, level);

			if (filter == null)
				filter = getFilterFromLevel(level);

			if ((filter != null)
					&& !filter.match(object))
			{
				removeTreeNode(treeNode);
				removeTreeNodeFromObjectId(object, treeNode);
				continue;
			}

			String previousLabel = treeNode.getLabel();

			if (previousLabel == null)
				previousLabel = "";
			// first we update the node
			_treeBehaviorInterface.updateTreeNode(object, treeNode, this);

			String label = treeNode.getLabel();

			if (label == null)
				label = "";

			if (parent instanceof LyTreeNode)
			{
				parentNode = (LyTreeNode)parent;

				LyObject parentObject = getObjectFromTreeNodeId(parentNode.getId());

				currentParents.addElement(parentObject);

				if ((parentObject != null)
						&& ((parents == null)
							|| (parents.getSize() == 0)
							|| !parents.contains(parentObject)))
				{
					useUpdateContent = true;
					break;
				}

				//FFT0941
				if ((parents == null) && (parentObject != null))
				{
					useUpdateContent = true;
					break;
				}
			}

			if ((values != null)
					&& values.hasMark(LyFieldInfo.MARK_ID))
			{
				LyObjectId oldObjectId = object.getObjectId(values);
				String     oldNodeId = computeId(oldObjectId, parentNode);
				String     nodeId = computeId(object, parentNode);

				// set TreeNode Id
				treeNode.setId(nodeId);

				//Register treeNodeId to TreeNode
				if (oldNodeId != null)
					_treeNodeIdToTreeNode.remove(oldNodeId);
				_treeNodeIdToTreeNode.put(nodeId, treeNode);

				if (level != null)
				{
					if (oldNodeId != null)
						_treeNodeIdToLevel.remove(oldNodeId);
					_treeNodeIdToLevel.put(nodeId, level);
				}

				if (oldObjectId != null)
					_objectIdToTreeNode.remove(oldObjectId);
				addTreeNodeFromObjectId(object.getObjectId(), treeNode);
			}

			LySort sort = ((LyTreeBehaviorInterface)_treeBehaviorInterface).getSort(this, level);

			if (sort == null)
				sort = getSortFromLevel(level);

			boolean isSorted = ((LyTreeBehaviorInterface)_treeBehaviorInterface).isSorted(this);

			if ((sort != null) && isSorted)
			{
				LyInfoList<LyFieldInfo> fieldsInfosSort = sort.getFieldInfos();
				int					    k, nk;

				for (k = 0, nk = values.getSize(); k < nk; k++)
				{
					LyFieldInfo currentFieldInfo = (LyFieldInfo)values.getFieldValue(k).getField();

					if (fieldsInfosSort.contains(currentFieldInfo))
					{
						useUpdateContent = true;
						break;
					}
				}
			}
			else
			{
				if (!label.equals(previousLabel) && isSorted)
				{
					// label has changed, we may have to move the node
					// -> we keep the list sorted alphabetically
					if ((parent != null) && (parent instanceof LyTreeNode))
					{
						label = treeNode.getLabel(); //.toLowerCase();

						if (label == null)
							label = "";

						parentNode.removeTreeNode(treeNode);

						boolean    inserted = false;
						LyTreeNode currentNode;
						String     currentLabel;

						for (j = 0, nj = parentNode.getTreeNodeCount(); j < nj; j++)
						{
							currentNode = parentNode.getTreeNode(j);
							currentLabel = currentNode.getLabel();
							if (label.compareTo(currentLabel) < 0)
							{
								// insert treeNode before this one
								parentNode.insertTreeNode(treeNode, j);
								inserted = true;
								break;
							}
						}

						if (!inserted)
							parentNode.addTreeNode(treeNode);
					}
				}
			}

			if ((level != null)
				&& (level.length() == 1))
			{
				LyFilter filterRootClass = (LyFilter)getParameter(LyAction.PARAM_FILTER, true);

				if (filterRootClass != null)
				{
					boolean match = filterRootClass.match(object);

					if (!match)
						useUpdateContent = true;
				}
			}

			LyObjectList sons = ((LyTreeBehaviorInterface)_treeBehaviorInterface).getSons(object,
					treeNode, level, this);

			if (sons != null)
			{
				LyTreeNode sonNode;
				LyObject   sonObject;

				// we look into each sons to see if the associated object belong to the list
				for (j = 0, nj = treeNode.getTreeNodeCount(); j < nj; j++)
				{
					sonNode = treeNode.getTreeNode(j);
					sonObject = getObjectFromTreeNodeId(sonNode.getId());

					if ((sonObject == null) || !sons.contains(sonObject))
						useUpdateContent = true;
				}

				// now, look into each object of the workspace to see if they all have the current
				// node as parent
				Vector<LyComponent> sonTreeNodes;

				if (sons != null)
				{
					boolean newOne;
					int     k, nk;

					for (j = 0, nj = sons.getSize(); j < nj; j++)
					{
						newOne = true;
						sonObject = sons.getObject(j);
						sonTreeNodes = getTreeNodesFromObjectId(sonObject.getObjectId());

						if ((sonTreeNodes == null) || (sonTreeNodes.size() == 0))
						{
							useUpdateContent = true;
							continue;
						}

						for (k = 0, nk = sonTreeNodes.size(); k < nk; k++)
						{
							sonNode = (LyTreeNode)sonTreeNodes.elementAt(k);
							if (sonNode.getParent() == treeNode)
							{
								newOne = false;
								break;
							}
						}

						if (newOne)
							useUpdateContent = true;
					}
				}
			}
		}

		// for each element in the parent list, check that there is the equivalent in
		// the vector list
		if (parents != null)
		{
			LyObject parentObject;

			for (i = 0, ni = parents.getSize(); i < ni; i++)
			{
				parentObject = parents.getObject(i);
				if (!currentParents.contains(parentObject))
				{
					useUpdateContent = true;
					break;
				}
			}
		}

		if (useUpdateContent)
			reset();

		tree.setIdle(idle);
		// Update the last modification time.
		setLastModified(System.currentTimeMillis());
		setRefreshTime(System.currentTimeMillis());
		return true;
	} // end method objectChanged

	/**
	 * Called when an object state is changed.
	 *
	 * @param  event Event notifying object state change.
	 * @return True if the event is used by the listener, false otherwise.
	 */
	public boolean objectStateChanged(LyAdvancedWorkSpaceEvent event)
	{
		if (isIdle())
			return false;

		LyObject                 object = event.getObject();
		LyWorkSpaceEvent.Type    type = event.getType();

		if (type == LyWorkSpaceEvent.Type.SELECTION_CHANGE)
		{
			selectObject(object, object.isSelected());
			return true;
		}
		else if (type == LyWorkSpaceEvent.Type.VISIBILITY_CHANGE)
		{
			Vector<LyComponent> v = getTreeNodesFromObjectId(object.getObjectId());

			if (v != null)
			{
				int				 i, n = v.size();
				LyGetIdInterface getIdObject;

				for (i = 0; i < n; i++)
				{
					getIdObject = (LyGetIdInterface)v.elementAt(i);

					if (object.isHidden())
						unselectObject(object, getIdObject.getId(), (i == (n - 1)));

					hideObject(object, getIdObject.getId(), object.isHidden(), (i == (n - 1)));
				}
			}

			updateContextClient();
			updateTools();
			updateMessages();
			return (v != null);
		}
		else if (type == LyWorkSpaceEvent.Type.ERROR_CHANGE)
		{
			Vector<LyComponent> v = getTreeNodesFromObjectId(object.getObjectId());

			if ((v != null) && (_treeBehaviorInterface != null))
			{
				int		   i, n = v.size();
				LyTreeNode treeNode;

				for (i = 0; i < n; i++)
				{
					treeNode = (LyTreeNode)v.elementAt(i);
					_treeBehaviorInterface.updateTreeNode(object, treeNode, this);
				}
			}

			return (v != null);
		}
		return false;
	} // end method objectStateChanged

	/**
	 * Return the possible sons of a classInfo in the tree.
	 *
	 * @param  classInfo class Info
	 * @return a list of valid classInfo (as sons) for a classInfo
	 */
	protected LyInfoList<LyClassInfo> getSonClassInfos(LyClassInfo classInfo)
	{
		LyInfoList<LyClassInfo> result = new LyInfoList<LyClassInfo>();

		if (_level != null)
		{
			Vector<String>   levels = getLevels(classInfo);
			int              i, n = ((levels == null) ? 0 : levels.size());

			for (i = 0; i < n; i++)
			{
				String		   level = levels.elementAt(i);
				_LyTreeElement treeElement = _level.get(level);

				result.addList(treeElement.getSons());
			}
		}

		return result;
	}

	/**
	 * Search the list of possible parent classInfos in the tree.
	 *
	 * @param  classInfo current classInfo
	 * @return the list of parent classInfo (above in the tree structure)
	 * @see    #isTopLevelClassInfo(LyClassInfo)
	 */
	public LyInfoList<LyClassInfo> getParentClassInfos(LyClassInfo classInfo)
	{
		LyInfoList<LyClassInfo> result = new LyInfoList<LyClassInfo>();

		if (_level != null)
		{
			Vector<String> levels = getLevels(classInfo);
			int            i, n = ((levels == null) ? 0 : levels.size());

			for (i = 0; i < n; i++)
			{
				String		   level = levels.elementAt(i);
				_LyTreeElement treeElement = _level.get(level);

				if (treeElement.getParent() != null)
					result.addInfo(treeElement.getParent().getClassInfo());
			}
		}

		return result;
	}

	/**
	 * Search if the object is suitable to be added witout a parent classInfo.
	 *
	 * @param  classInfo current classInfo
	 * @return true if the object may be added at top level
	 * @see    #getParentClassInfos(LyClassInfo)
	 */
	public boolean isTopLevelClassInfo(LyClassInfo classInfo)
	{
		if (_level != null)
		{
			Vector<String> levels = getLevels(classInfo);

			if ((levels != null)
					&& (levels.size() == 1)
					&& levels.contains("1"))
			{
				_LyTreeElement treeElement = _level.get("1");

				if (treeElement.getParent() == null)
					return true;
			}
		}

		return false;
	}

	/**
	 * Clears all cached data for this controller (especially the view and the component).
	 *
	 * <P>This method may be invoked for memory optimization, but the controller is still valid and
	 * may be used. (Unlike free) Further usage, needs to rebuild internal component & view which
	 * may be time consuming.</P>
	 */
	public void clearViewCache()
	{
		_objectIdToTreeNode = new Hashtable<LyObjectId, Vector<LyComponent>>();
		_treeNodeIdToTreeNode = new Hashtable<String, LyComponent>();

		// Clear assoc table
		_selections = null;

		super.clearViewCache();
	}

	/**
	 * Internal class to build the tree
	 *
	 */
	protected class _LyTreeElement
	{
		/** Classinfo of the element */
		protected LyClassInfo			  _classInfo;

		/** Parent element */
		protected _LyTreeElement		  _elementParent;

		/** Level of element */
		protected String				  _elementLevel;

		/** list of sons */
		protected LyInfoList<LyClassInfo> _sons = null;

		/**
		 * Constructor
		 * @param classInfo classinfo of element
		 * @param parent parent element
		 * @param level associated level
		 */
		public _LyTreeElement(LyClassInfo classInfo, _LyTreeElement parent, String level)
		{
			_classInfo = classInfo;
			_elementParent = parent;
			_elementLevel = level;

			if (parent != null)
				parent.addSon(classInfo);
		}

		/**
		 * Get the elmement classInfo
		 * @return the associated classInfo
		 */
		protected LyClassInfo getClassInfo()
		{
			return _classInfo;
		}

		/**
		 * Get the parent element
		 * @return the parent element (null for root node)
		 */
		protected _LyTreeElement getParent()
		{
			return _elementParent;
		}

		/**
		 * Get level of element
		 * @return the level
		 */
		protected String getLevel()
		{
			return _elementLevel;
		}

		/**
		 * Add a son classInfoto the element
		 * @param son the son to add
		 */
		protected void addSon(LyClassInfo son)
		{
			if (_sons == null)
				_sons = new LyInfoList<LyClassInfo>();

			_sons.addInfo(son);
		}

		/**
		 * Accessor to the sons classinfo
		 * @return a list of LyClassInfo
		 */
		protected LyInfoList<LyClassInfo> getSons()
		{
			return _sons;
		}

		/**
		 * @see java.lang.Object#toString()
		 */
		public String toString()
		{
			String result = "Level : " + _elementLevel + " classInfo : " + _classInfo;

			return result;
		}
	} // end class _LyTreeElement

	/**
	 * Method that indicates if controlled panel enables multi-selection or not.
	 *
	 * @return A boolean indicating if multi-selection is allowed or not.
	 */
	public boolean enablesMultiSelection()
	{
		LyTree tree = getTreeComponent();

		return ((tree == null) ? false
							   : (tree.getSelectionMode() == LyTree.SelectionMode.MULTIPLE));
	}

	/**
	 * Method that indicates if controlled panel is horizontal.
	 *
	 * @return A boolean indicating if the panel component is horizontal.
	 */
	public boolean isHorizontal()
	{
		return false;
	}


	/**
	 * Check ignore idle status of the tree controller.
	 *
	 * @return boolean indicating the ignore idle mode
	 */
	public boolean isIgnoreIdle()
	{
		return _ignoreIdle;
	}

	/**
	 * Set the ignore idle status of tree controller.
	 * @param mode true to ignore idle, false else
	 */
	public void setIgnoreIdle(boolean mode)
	{
		_ignoreIdle = mode;
	}

	/**
	 * Check discovering mode of the tree (fully or on node click).
	 *
	 * @return boolean indicating if the discovering is made on click
	 */
	public boolean isDiscoverOnClick()
	{
		return _discoverOnClick;
	}

} // end class LyTreeController
