/**
 * LyUserInputController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyUserInputController.java 28997 2010-12-23 10:17:15Z bmichel $
 */

package leon.control;

import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * This class implements a user input dialog/window to ask some values from the user. Values may be
 * read when window/dialog is closed This controller is associated with a builder
 * (LyUserInputBuilder) and a behavior that implements LyUserInputBehaviorInterface. Usually, this
 * behavior is an instance of the LyUserInputBehavior class.
 *
 * @see LyUserInputBuilder
 * @see leon.app.behaviorinterface.LyUserInputBehaviorInterface
 * @see leon.app.behavior.LyUserInputBehavior
 */
public class LyUserInputController extends LyAreaController
{
	/** File Version */
	public static final String __VERSION =
		"$Id: LyUserInputController.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** Indicates whether the window is a frame (window) or a dialog. */
	protected boolean		   _isDialog;

	/** Input id set by the listener to identify this dialog. */
	protected String		   _inputId;

	/** Window titlebar label. */
	protected String		   _titleBar;

	/** Window titlebar image. */
	protected String		   _image;

	/** Class info given on construction (may be null). */
	protected LyClassInfo	   _classInfo;

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 * @param action  related action. If an action is specified, the controller uses action
	 *                parameters (action name, action specific XML view description ...)
	 */
	public LyUserInputController(LySession session, LyController parent, LyAction action)
	{
		// This constructor is kept because it's called directly by other controllers
		this(session, parent, action, null, null);
	}

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo This parameter is not used. Kept from introspection compatibility.
	 */
	public LyUserInputController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo)
	{
		// This constructor is kept because it's called directly by other controllers
		this(session, parent, action, classInfo, null);
	}

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 * @param action  related action. If an action is specified, the controller uses action
	 *                parameters (action name, action specific XML view description ...)
	 * @param context current context (may be null if the view is not contextual to a list of
	 *                objects).
	 */
	public LyUserInputController(LySession session, LyController parent, LyAction action,
		LyObjectList context)
	{
		// This constructor is kept because it's called directly by other controllers
		this(session, parent, action, null, context);
	}

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo This parameter is not used. Kept from introspection compatibility.
	 * @param context   current context (may be null if the view is not contextual to a list of
	 *                  objects).
	 */
	public LyUserInputController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent,
			((action == null) ? session.getApplication().getActionWithId(LyAction.ID_USER_INPUT, classInfo) : action),
			context);

		_isDialog = !_action.hasMark(LyAction.MARK_FRAME_VIEW);
		_inputId = _action.getId();
		_titleBar = _action.getName();
		_image = getImage(_action);
		_classInfo = classInfo;
	}

	/**
	 * Get controller class info if any. This method is overriden if it makes sense to return a
	 * classInfo for this controller. Default returns null.
	 *
	 * @return a classinfo if any or null.
	 */
	public LyClassInfo getClassInfo()
	{
		return ((_classInfo != null) ? _classInfo : super.getClassInfo());
	}

	/**
	 * Accessor to the view component instance. (XML description of the view) This instance is
	 * created by the builder, and multiple calls to this method always return the same instance.
	 *
	 * @return the LyComponent instance of the view package, memory representation (parsed) of the
	 *         XML description of the view.
	 */
	public LyComponent getComponent()
	{
		// If component is ready, return it
		if (_component != null)
			return _component;

		// Build component and complete it
		super.getComponent();

		LyBuilder builder = getBuilder();

		if (!(builder instanceof LyFileBuilder))
			completeComponent();

		return _component;
	}

	/**
	 * Complete content of the component of the controller.
	 */
	protected void completeComponent()
	{
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyUserInputBehaviorInterface.class;
	}

	/**
	 * Invoked when an action occurs. (LyDemandListener interface) The current implementation calls
	 * the userInputDemand method of the behavior.
	 *
	 * @param demand the action demand.
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (getBehavior() != null)
			((LyUserInputBehaviorInterface)getBehavior()).userInputDemand(this, demand);

		if (!isFreed())
		{
			String sourceId = demand.getSource().intern();

			if (sourceId.equals("_tabs"))
			{
				LyParameterValues values = demand.getParameterValues();
				int				  tabIndex = demand.getIndex(0);
				LyComponent		  component = _component.getComponent(demand.getValue());
				LyTab			  tab = ((component instanceof LyTab) ? (LyTab)component : null);
				LyTabs			  tabs = ((tab == null) ? null : (LyTabs)tab.getParent());

				if (tabs == null)
				{
					Vector<LyTabs> componentTabs = new Vector<LyTabs>(1);

					_component.getComponents(LyTabs.class, componentTabs);

					if (componentTabs.size() > 0)
						tabs = componentTabs.elementAt(0);
				}

				if (values != null)
				{
					setLastModified(new Date().getTime());

					// Check if there are tabs within tabs (composition RELATION fields) and we
					// may need to forward the event
					String targetControllerId = values.getParameter("_target_controller");

					if ((targetControllerId != null)
							&& (targetControllerId.length() > 0)
							&& (!targetControllerId.equals(getId())))
					{
						// Forward event !
						LyFormController formCtrl = (LyFormController)getSession().getController(
								targetControllerId);
						LyActionDemand   demandFwd = new LyActionDemand(demand.getType(), "_tabs");

						demandFwd.addIndex(tabIndex);

						formCtrl.actionDemand(demandFwd);
						return;
					}
				}

				if (tabs != null)
				{
					tab = tabs.getTab(tabIndex);

					// In replay mode, retrieve tab by its id and not by its index.
					// In web mode, there may be an offset in tabs indexes due to a first virtual tab
					// built for fields that are not in a specific tab.
					LyEnvironment environment = getEnvironment();
					LyPlayer	  player = environment.getPlayer();

					if ((player != null)
							&& player.isPlaying()
							&& (component instanceof LyTab))
						tab = (LyTab)component;

					if (tab != null)
						tabs.setSelectedTab(tab.getLabel());
				}
			}
			super.actionDemand(demand);
		}
	} // end method actionDemand

	/**
	 * Retrieve the identification freely set by the listener.
	 *
	 * @return the identification
	 */
	public String getInputId()
	{
		return _inputId;
	}

	/**
	 * Sets the identification freely set by the listener.
	 *
	 * @param inputId the identification
	 */
	public void setInputId(String inputId)
	{
		_inputId = inputId;
	}

	/**
	 * Retrieve the label of the titleBar.
	 *
	 * @return the label of the titlebar
	 */
	public String getTitleBar()
	{
		if (_titleBar != null)
			return _titleBar;

		if (getBehavior() != null)
			return ((LyActionBehaviorInterface)getBehavior()).getTitle(this, _action);

		return null;
	}

	/**
	 * Sets the label of the titleBar.
	 *
	 * @param titleBar the label of the titleBar
	 */
	public void setTitleBar(String titleBar)
	{
		_titleBar = titleBar;

		// Force update if needed
		if (_builder instanceof LyAreaBuilder)
			((LyAreaBuilder)_builder).updateTitleBar();
	}

	/**
	 * Retrieve the image shown in the window titleBar.
	 *
	 * @return the associated image
	 */
	public String getImage()
	{
		return _image;
	}

	/**
	 * Sets dialog status. default is true
	 *
	 * @param isDialog dialog status (view is a dialog)
	 */
	public void setDialog(boolean isDialog)
	{
		_isDialog = isDialog;
	}

	/**
	 * Gets dialog status. default is true
	 *
	 * @return dialog status (view is a dialog)
	 */
	public boolean isDialog()
	{
		return _isDialog;
	}

	/**
	 * Sets the image shown in the window titleBar.
	 *
	 * @param image the image to set
	 */
	public void setImage(String image)
	{
		_image = image;
	}

	/**
	 * Accessor to the associated view controller.
	 *
	 * @return the view controller
	 */
	public LyViewController getViewController()
	{
		return getViewController(_isDialog);
	}

	/**
	 * Add a label in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  itemLabel  content of the field
	 * @return the label component
	 */
	public LyLabel addLabel(String fieldId, String fieldLabel, String itemLabel)
	{
		return addLabel(fieldId, fieldLabel, itemLabel, -1);
	}

	/**
	 * Add a label in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  itemLabel  content of the field
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the label component
	 */
	public LyLabel addLabel(String fieldId, String fieldLabel, String itemLabel, int pos)
	{
		LyUserInputBuilder builder = (LyUserInputBuilder)getBuilder();

		return builder.addLabel(fieldId, fieldLabel, itemLabel, pos);
	}

	/**
	 * Add a text field in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  itemLabel  content of the field
	 * @param  ncols      number of columns for the field
	 * @param  format     format for the field (f.e. IP Address format...)
	 * @param  optional   indicates if the field is optional or not
	 * @return the text component
	 */
	public LyText addText(String fieldId, String fieldLabel, String itemLabel,
		int ncols, String format, boolean optional)
	{
		return addText(fieldId, fieldLabel, itemLabel, ncols, format, optional, -1);
	}

	/**
	 * Add a text field in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  itemLabel  content of the field
	 * @param  ncols      number of columns for the field
	 * @param  format     format for the field (f.e. IP Address format...)
	 * @param  optional   indicates if the field is optional or not
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the text component
	 */
	public LyText addText(String fieldId, String fieldLabel, String itemLabel,
		int ncols, String format, boolean optional, int pos)
	{
		LyUserInputBuilder builder = (LyUserInputBuilder)getBuilder();

		return builder.addText(fieldId, fieldLabel, itemLabel, ncols, format, optional, pos);
	}

	/**
	 * Add a text field in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  itemLabel  content of the field
	 * @param  ncols      number of columns for the field
	 * @param  nrows      number of rows for the field
	 * @param  format     format for the field (f.e. IP Address format...)
	 * @param  optional   indicates if the field is optional or not
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the text component
	 */
	public LyText addText(String fieldId, String fieldLabel, String itemLabel,
		int ncols, int nrows, String format, boolean optional, int pos)
	{
		LyUserInputBuilder builder = (LyUserInputBuilder)getBuilder();

		return builder.addText(fieldId, fieldLabel, itemLabel, ncols, nrows, format, optional, pos);
	}

	/**
	 * Add a relation field (empty) in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  multiple   indicates if this is a multiple relation
	 * @param  optional   indicates if the field is optional or not
	 * @return the relation component
	 */
	public LyRelation addRelation(String fieldId, String fieldLabel, boolean multiple,
		boolean optional)
	{
		return addRelation(fieldId, fieldLabel, multiple, optional, -1);
	}

	/**
	 * Add a relation field (empty) in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  multiple   indicates if this is a multiple relation
	 * @param  optional   indicates if the field is optional or not
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the relation component
	 */
	public LyRelation addRelation(String fieldId, String fieldLabel, boolean multiple,
		boolean optional, int pos)
	{
		LyUserInputBuilder builder = (LyUserInputBuilder)getBuilder();

		return builder.addRelation(fieldId, fieldLabel, multiple, optional, pos);
	}

	/**
	 * Add a choice field (empty) in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  multiple   indicates if this is a multiple relation
	 * @param  optional   indicates if the field is optional or not
	 * @return the choice component
	 */
	public LyChoice addChoice(String fieldId, String fieldLabel, boolean multiple,
		boolean optional)
	{
		return addChoice(fieldId, fieldLabel, multiple, optional, -1);
	}

	/**
	 * Add a choice field (empty) in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  multiple   indicates if this is a multiple relation
	 * @param  optional   indicates if the field is optional or not
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the choice component
	 */
	public LyChoice addChoice(String fieldId, String fieldLabel, boolean multiple,
		boolean optional, int pos)
	{
		LyUserInputBuilder builder = (LyUserInputBuilder)getBuilder();

		return builder.addChoice(fieldId, fieldLabel, multiple, optional, pos);
	}

	/**
	 * Add a number field in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  itemLabel  the original value of the field
	 * @param  type       the type of the number field
	 * @param  optional   indicates if the field is optional or not
	 * @return the number component
	 */
	public LyNumber addNumber(String fieldId, String fieldLabel, String itemLabel,
		LyNumberFieldInfo.Type type, boolean optional)
	{
		return addNumber(fieldId, fieldLabel, itemLabel, type, optional, -1);
	}

	/**
	 * Add a number field in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  itemLabel  the original value of the field
	 * @param  type       the type of the number field
	 * @param  optional   indicates if the field is optional or not
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the number component
	 */
	public LyNumber addNumber(String fieldId, String fieldLabel, String itemLabel,
        LyNumberFieldInfo.Type type, boolean optional, int pos)
	{
		LyUserInputBuilder builder = (LyUserInputBuilder)getBuilder();

		return builder.addNumber(fieldId, fieldLabel, itemLabel, type, optional, pos);
	}

	/**
	 * Add a date field in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  itemLabel  content of the field
	 * @param  format     date format for the field
	 * @param  optional   indicates if the field is optional or not
	 * @return the date component
	 */
	public LyDate addDate(String fieldId, String fieldLabel, String itemLabel,
		String format, boolean optional)
	{
		return addDate(fieldId, fieldLabel, itemLabel, format, optional, -1);
	}

	/**
	 * Add a date field in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  itemLabel  content of the field
	 * @param  format     date format for the field
	 * @param  optional   indicates if the field is optional or not
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the date component
	 */
	public LyDate addDate(String fieldId, String fieldLabel, String itemLabel,
		String format, boolean optional, int pos)
	{
		LyUserInputBuilder builder = (LyUserInputBuilder)getBuilder();

		return builder.addDate(fieldId, fieldLabel, itemLabel, format, optional, pos);
	}

	/**
	 * Add a text field in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  itemLabel  original content of the field
	 * @param  type       type of the file
	 * @param  optional   indicates if the field is optional or not
	 * @return the date component
	 */
	public LyFile addFile(String fieldId, String fieldLabel, String itemLabel,
		LyFile.Type type, boolean optional)
	{
		return addFile(fieldId, fieldLabel, itemLabel, type, optional, -1);
	}

	/**
	 * Add a text field in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  itemLabel  original content of the field
	 * @param  type       type of the file
	 * @param  optional   indicates if the field is optional or not
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the date component
	 */
	public LyFile addFile(String fieldId, String fieldLabel, String itemLabel,
		LyFile.Type type, boolean optional, int pos)
	{
		LyUserInputBuilder builder = (LyUserInputBuilder)getBuilder();

		return builder.addFile(fieldId, fieldLabel, itemLabel, type, optional, pos);
	}

	/**
	 * Add a list field (empty) in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  multiple   indicates if selection mode is multiple in the list
	 * @param  height     the height of the list
	 * @param  optional   indicates if the field is optional or not
	 * @return the list component
	 */
	public LyList addList(String fieldId, String fieldLabel, boolean multiple,
		int height, boolean optional)
	{
		return addList(fieldId, fieldLabel, multiple, height, optional, -1);
	}

	/**
	 * Add a list field (empty) in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel field label
	 * @param  multiple   indicates if selection mode is multiple in the list
	 * @param  height     the height of the list
	 * @param  optional   indicates if the field is optional or not
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the list component
	 */
	public LyList addList(String fieldId, String fieldLabel, boolean multiple,
		int height, boolean optional, int pos)
	{
		LyUserInputBuilder builder = (LyUserInputBuilder)getBuilder();

		return builder.addList(fieldId, fieldLabel, multiple, height, optional, pos);
	}

	/**
	 * Add a boolean field in the user input window.
	 *
	 * @param  fieldId    the id of the field
	 * @param  fieldLabel the label of the field
	 * @param  status     the boolean status
	 * @param  optional   indicates if the choice is optional
	 * @return the boolean component
	 */
	public LyBoolean addBoolean(String fieldId, String fieldLabel, LyBoolean.Status status,
		boolean optional)
	{
		return addBoolean(fieldId, fieldLabel, status, optional, -1);
	}

	/**
	 * Add a boolean field in the user input window.
	 *
	 * @param  fieldId    the id of the field
	 * @param  fieldLabel the label of the field
	 * @param  status     the boolean status
	 * @param  optional   indicates if the choice is optional
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the boolean component
	 */
	public LyBoolean addBoolean(String fieldId, String fieldLabel,
		LyBoolean.Status status, boolean optional, int pos)
	{
		LyUserInputBuilder builder = (LyUserInputBuilder)getBuilder();

		return builder.addBoolean(fieldId, fieldLabel, status, optional, pos);
	}

	/**
	 * Retrieves the view Builder.
	 *
	 * @return the builder instance;
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return (new LyUserInputBuilder(this));
	}

	/**
	 * Indicates if the view for this input controller is resizable.
	 *
	 * @return true if the view is resizable, false otherwise.
	 */
	public boolean isViewResizable()
	{
		return false;
	}

	/**
	 * Setter to controller builder A builder is associated to a controller to let the developper
	 * construct the desired view.
	 *
	 * @param builder associated builder
	 */
	public void setBuilder(LyBuilder builder)
	{
		super.setBuilder(builder);
	}

	/**
	 * Setter to controller behavior A behavior is associated to a controller to let the
	 * developper specify application logic in the behavior.
	 *
	 * @param behavior associated behavior
	 */
	public void setBehavior(LyBehaviorInterface behavior)
	{
		super.setBehavior(behavior);
	}

	/**
	 * Retrieves form that contains input fields.
	 * @return the form component
	 */
	public LyForm getFormComponent()
	{
		// Retrieve internal form component
		LyForm form = (LyForm)getComponent().getComponent(getId() + "_form");

		// For file builder
		if (form == null)
        {
			form = (LyForm)getComponent().getComponent("_form");
            
            if (form == null)
            {
                Vector<LyForm>  forms = new Vector<LyForm>(1);
                
                getComponent().getComponents(LyForm.class, forms);
                if (forms.size() > 0)
                    form = forms.elementAt(0);
            }
        }
        
		return form;
	}

	/**
	 * Reset all the content of the view. This method may be called several times. This method
	 * should be overriden, default implementation makes nothing
	 *
	 * @param full Boolean indicating if a full reset (including reset of datas) is requiered or not
	 *             (only reset graphical area).
	 */
	public void reset(boolean full)
	{
		super.reset(full);
		updateTitleBar();
	}
	
	
} // end class LyUserInputController
