/**
 * LyViewController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyViewController.java 41460 2012-11-08 16:34:27Z nhamon $
 */

package leon.control;

import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.data.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.setinterface.*;


/**
 * This class implements the basis of all view controller.
 *
 * @see LyController
 * @see LyFrameController
 * @see LyDialogController
 */
public abstract class LyViewController extends LyController
	implements LyContextInterface,
		LyQuestionAnswerInterface
{
	/** File Version */
	public static final String		  __VERSION =
		"$Id: LyViewController.java 41460 2012-11-08 16:34:27Z nhamon $";

	/** Indicates the root level controller, usually a "Frame" (generic). */
	public static final int			  LEVEL_ROOT = 0;

	/**
	 * Indicates the primary level controller, the view controller if the view is simple (Table,
	 * Tree, Map) or the main controller in the case of multi view.
	 */
	public static final int			  LEVEL_PRIMARY = 1;

	/**
	 * Indicates the secondary level controller. Content depends on the primary controller, and this
	 * controller may not add some menus to the menubar.
	 */
	public static final int			  LEVEL_SECONDARY = 2;

	/** Third level controller. May be used within specific implementations */
	public static final int			  LEVEL_THIRD = 3;

	/* Demand codes */

	/** Identifier of a action in the view : Application end. */
	public static final String		  DEMAND_EXIT = "_exit".intern();

	/** Closing view (deletes window and associated controller). */
	public static final String		  DEMAND_CLOSE = "_close".intern();

	/** Hiding view. */
	public static final String		  DEMAND_HIDE = "_hide".intern();

	/** Show version. */
	public static final String		  DEMAND_VERSION = "_version".intern();

	/** Show about. */
	public static final String		  DEMAND_ABOUT = "_about".intern();

	/** Show log. */
	public static final String		  DEMAND_LOG = "_log".intern();

	/** Update. */
	public static final String		  DEMAND_UPDATE = "_update".intern();

	/** Area controller contained in this window. */
	LyAreaController				  _areaController;

	/** Controllers opened from this window. */
	Vector<LyViewController>		  _sons;

	/** Boolean indicating if this controller must be freed on closing. */
	boolean							  _freeOnClose;

	/** Initialized. */
	boolean							  _viewInitialized;

	/** Boolean indicating if the controlled view is shown or not. */
	boolean							  _shown;

	/** Printing thread. */
	protected Thread				  _printThread;

	/** In shown process. */
	protected boolean				  _inShowProcess;

	/** Store the popup view stack. */
	protected Stack<LyViewController> _subControllerStack;

	/** Update state. */
	protected boolean                 _updated;

	/**
	 * Default constructor.
	 *
	 * @param session   session for the controller
	 * @param parent the parent controller
	 */
	public LyViewController(LySession session, LyController parent)
	{
		this(session, parent, null);
	}

	/**
	 * Default constructor.
	 *
	 * @param session        session for the controller
	 * @param parent         the parent controller
	 * @param areaController the associated area controller
	 */
	public LyViewController(LySession session, LyController parent, LyAreaController areaController)
	{
		super(session, parent,
			((areaController == null) ? null : ("view_" + areaController.getId())));

		_sons = null;
		_areaController = null;
		_freeOnClose = true;
		_viewInitialized = false;
		_shown = false;
		_printThread = null;
		_inShowProcess = true;

		// If not root controller, add this controller on parent
		if (parent != null)
		{
			LyViewController parentViewController = parent.getViewController();

			if (parentViewController != null)
				parentViewController.addSon(this);
		}

		if (areaController != null)
			setAreaController(areaController);
	}

	/**
	 * Get Accessor to the associated area controller.
	 *
	 * @return the associated area controller
	 */
	public LyAreaController getAreaController()
	{
		return _areaController;
	}

	/**
	 * Set Accessor to the associated area controller.
	 *
	 * @param areaController the area controller
	 */
	public void setAreaController(LyAreaController areaController)
	{
		_areaController = areaController;

		if (_areaController != null)
		{
			_areaController.setViewController(this);
			_areaController.setParent(this);

			setName(_areaController.getName());
		}
	}

	/**
	 * Initializes this area controller.
	 */
	public void initialize()
	{
		initView();
	}

	/**
	 * Method used to prepare components before showing. It's the right time to fill up the
	 * component with real datas.
	 */
	protected void initView()
	{
		if (_viewInitialized)
			return;

		// Init area controller
		if (_areaController != null)
			_areaController.initArea();

		_viewInitialized = true;
		_updated = true;
	}

	/**
	 * Free this controller.
	 *
	 * @see LyController#free
	 */
	public void free()
	{
		if (isFreed())
			throw new IllegalStateException("BF AppEngine : controller " + this
				+ " was already freed");

		// Freeze controllers
		setIdle(true);

		// Hide and free of the associated view
		hideView(true);

		// Free recursively son controllers
		LyViewController controller;
		int				 i, n = getSonsCount();

		for (i = n - 1; i >= 0; i--)
		{
			controller = getSon(i);

			if (!controller.isFreed())
				controller.free();
		}
		_subControllerStack = null;
		_sons = null;

		// Remove this controller from its parent (except for root controller)
		if (_parent != null)
		{
			LyViewController parentViewController = _parent.getViewController();

			if (parentViewController != null)
				parentViewController.removeSon(this);
		}

		// Free area controller
		if (_areaController != null)
		{
			// If area controller has a stack area controller, free this stack area controller
			// which will result in a recursive freeing of area controllers contained in this
			// view controller.
			if (_areaController.getAreaStackController() != null)
				_areaController.getAreaStackController().freeAreaStack();
			else
				_areaController.free(this);

			_areaController = null;
		}

		// Free on this controller may have been called when hiding view
		if (!isFreed())
			super.free();
	} // end method free

	/**
	 * Add a son to this controller.
	 *
	 * @param controller son view controller
	 */
	protected void addSon(LyViewController controller)
	{
		if (controller == null)
			throw new IllegalArgumentException("BF AppEngine : controller can't be null");
		if (_sons == null)
			_sons = new Vector<LyViewController>(1);

		// Add son
		_sons.addElement(controller);

		// Call the area controller.
		if (_areaController != null)
			_areaController.sonAdded(controller);
	}

	/**
	 * Remove a son from this controller.
	 *
	 * @param controller son view controller to remove
	 */
	protected void removeSon(LyViewController controller)
	{
		if (controller == null)
			throw new IllegalArgumentException("BF AppEngine : controller can't be null");
		if ((_sons == null) || !_sons.contains(controller))
		{
			getEnvironment().logError("BF AppEngine : controller " + controller + " unknown");
			return;
			//throw new IllegalArgumentException("BF AppEngine : controller " + controller + " unknown");
		}
		// Remove son
		_sons.removeElement(controller);

		if (_sons.size() == 0)
			_sons = null;

		// Call the area controller.
		if (_areaController != null)
			_areaController.sonRemoved(controller);
	}

	/**
	 * Number of son controller.
	 *
	 * @return the number of sons controllers
	 */
	public int getSonsCount()
	{
		return ((_sons == null) ? 0 : _sons.size());
	}

	/**
	 * Get the son controller at the given position.
	 *
	 * @param  index the position of the controller to get
	 * @return the son view controller
	 */
	public LyViewController getSon(int index)
	{
		return _sons.elementAt(index);
	}

	/**
	 * Asks if printing jobs shall be stopped.
	 *
	 * @param exit true if exit is in progress
	 */
	protected void askPrintStop(boolean exit)
	{
		LyEnvironment environment = getEnvironment();
		LyViewManager viewManager = getViewManager();
		String		  message = environment.getMessage("LY_CONFIRM_PRINT_STOP");

		viewManager.askQuestion(getView(), message, this, Boolean.valueOf(exit));
	}

	/**
	 * Answer to a askQuestion request.
	 *
	 * @param result one of the possible return code for LyViewManager : COMMAND_XXX
	 * @param clientData data specified in the askQuestion method
	 * @see   leon.view.setinterface.LyMessageAnswerInterface
	 */
	public void questionAnswer(boolean result, Object clientData)
	{
		Boolean data = (Boolean)clientData;
		boolean exit = data.booleanValue();

		if (result)
		{
			if (!exit)
			{
				// If not an exit, interrupt thread for this view and close.
				interruptPrintThread();
				close();
			}
			else
			{
				// Stop all the threads registered on current session.
				LySession session = getSession();

				session.interruptPrintThreads();
				exit();
			}
		}
	}

	/**
	 * Returns view controller that contains this controller. This controller IS a view Controller.
	 * Return this.
	 *
	 * @return this
	 */
	public LyViewController getViewController()
	{
		return this;
	}

	/**
	 * Creates a root view which contains the view.
	 *
	 * @return the root view
	 * @see    LyController#getView
	 */
	public LyView getView()
	{
		return (LyView)getComponent();
	}

	/**
	 * Shows the Window which contains the view.
	 */
	public void showView()
	{
		// Check that this controller was not freed (initialisation failure)
		if (isFreed())
			return;

		if (!_viewInitialized)
			initView();
		else if (!_updated)
    		// Update area controller content if needed
    		_areaController.update();

		// Notify session that controller is shown
		LySession session = getSession();

		setInShowProcess(true);
		session.controllerShown(this);
		setInShowProcess(false);

		// Show view
		setVisible(true);
	}

	/**
	 * Hides the Window which contains the view.
	 *
	 * @param free true if this controller must be freed
	 * @see   LyController#hideView
	 */
	public void hideView(boolean free)
	{
		if (_shown)
		{
			// Notify session that controller is hidden
			LySession session = getSession();

			session.controllerHidden(this);

			// Close son views
			if (_sons != null)
			{
				Vector<LyViewController>    sons = new Vector<LyViewController>(_sons);
				LyViewController            son;
				int				            i, n = sons.size();

				for (i = n - 1; i >= 0; i--)
				{
					son = sons.elementAt(i);
					if (_sons.contains(son))
						// If son is still contained in sons list, close it
						son.close();
				}
			}

			// Hide view
			setVisible(false);
		}

		// Clear view cache if needed
		if (free)
			clearViewCache();
	} // end method hideView

	/**
	 * Tests if controlled view is shown or not.
	 *
	 * @return True if shown, false otherwise.
	 */
	public boolean isShown()
	{
		return _shown;
	}

	/**
	 * Close this controller If a print thread is in progress, ask the user if the print must be
	 * cancelled.
	 */
	public void close()
	{
		LyViewManager viewManager = getViewManager();

		// Closing root view controller means exit
		if ((_parent == null) //if the mode of viewManager is not mod server
				&& !viewManager.hasProperty("MOD_APP_SERVER")
				&& ((getEnvironment() == null)
						|| !getEnvironment().testEnv("LY_SILENT_GUI")))
			getSession().exit();
		// Delete this controller on closing of its view
		else
		{
			if (hasPrintThread())
			{
				askPrintStop(false);
				return;
			}

			LyController parent = getParent();

			if (getFreeOnClose())
			{
				// Check if area controller can be freed
				if ((_areaController == null)
						|| _areaController.canFree(true))
					free();
			}
			else
				// Hide view
				hideView();

			// Specific case for a dashboard parent : update window menu in the menu bar.
			if (parent instanceof LyDashboardController)
				((LyDashboardController)parent).updateWindowMenu();
		}
	} // end method close

	/**
	 * Exit.
	 */
	public void exit()
	{
		LySession session = getSession();
		int		  nThreads = session.getPrintThreadCount();

		if (nThreads <= 0)
			// Get the current session and close it.
			session.close();
		else
			askPrintStop(true);
	}

	/**
	 * Invoked when a notification is sent to this listener. Overriden to call behvior of the area
	 * controller.
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   LyController#demandPerformed
	 */
	public synchronized void demandPerformed(LyDemand demand)
	{
		if (isFreed())
			return;

		LyBehaviorInterface behavior = getBehavior();
		LySession		    session = getSession();
		LyApplication	    application = session.getApplication();
		LyDataSession	    oldCurrentSession = application.getCurrentSession();

		try
		{
			application.setCurrentSession(session);
			if (behavior == null)
				behavior = _areaController.getBehavior();

			if ((behavior != null)
					&& behavior.handleDemand(this, demand))
				return;

			super.demandPerformed(demand);

			// We must'nt have some pending transaction on the current
			// session.
			if (!(demand instanceof LyActivateViewDemand)
					&& session.hasTransaction())
			{
				LyEnvironment environment = getEnvironment();
				String		  message = environment.getMessage(
						"LY_ER_REMAINING_PENDING_TRANSACTION");

				if (isFreed())
					getViewManager().showError(null, message);
				else
					showError(message);

				throw new IllegalStateException("BF AppEngine : Remaining pending transaction(s)");
			}
		}
		catch (Throwable t)
		{
			session.handleException(t);
		}
		finally
		{
			if ((oldCurrentSession != null)
					&& !oldCurrentSession.isClosed())
				application.setCurrentSession(oldCurrentSession);
		}
	} // end method demandPerformed

	/**
	 * Invoked when an action occurs. This method handles close and help demand
	 *
	 * @param demand action demand
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (isFreed())
			return;

		if (demand.getType() == LyActionDemand.TYPE_WINDOW_CLOSED)
		{
			close();
			return;
		}
		else if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			String demandId = demand.getSource().intern();

			if (demandId == DEMAND_CLOSE)
			{
				if ((_parent == null)
						&& !getViewManager().hasProperty("MOD_APP_SERVER"))
					exit();
				else
					close();
				return;
			}
			else if (demandId == DEMAND_HELP)
			{
				_areaController.actionDemand(demand);
				return;
			}
			else if (demandId == DEMAND_EXIT)
			{
				getSession().exit();
				return;
			}
			else if (demandId == DEMAND_UPDATE)
			{
				_areaController.update();
				return;
			}
		}

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Returns boolean indicating if this controller is freed on close.
	 *
	 * @return true if the controller is freed on close
	 */
	public boolean getFreeOnClose()
	{
		return _freeOnClose;
	}

	/**
	 * Sets boolean indicating if this controller is freed on close.
	 *
	 * @param freeOnClose true if this controller must be freed on close
	 */
	public void setFreeOnClose(boolean freeOnClose)
	{
		_freeOnClose = freeOnClose;
	}

	/**
	 * Get view area.
	 *
	 * @return the view area
	 */
	public LyArea getViewArea()
	{
		LyView view = getView();

		return view.getArea();
	}

	// XXX
	/**
	 * Set view area.
	 *
	 * @param area the view area
	 */
	public void setViewArea(LyArea area)
	{
		LyView view = getView();

		view.setArea(area);
	}

	/**
	 * To get the last modification date of this controller.
	 *
	 * @return the last modification date
	 */
	public long getLastModified()
	{
		// The last modification date of a view controller is the
		// last modification date of the area controller.
		if (_areaController != null)
			_lastModified = _areaController.getLastModified();

		return _lastModified;
	}

	/**
	 * Set current controller context.
	 *
	 * @param context current context or null if none is specified
	 * @return true if the context was correctly set
	 * @see   LyContextInterface#getContext
	 */
	public boolean setContext(LyWorkSpace context)
	{
		if (_areaController != null)
			return _areaController.setContext(context);
		else
			return false;
	}

	/**
	 * Get current controller context.
	 *
	 * @return current context or null if none is specified
	 * @see    LyContextInterface#getContext
	 */
	public LyWorkSpace getContext()
	{
		if (_areaController != null)
			return _areaController.getContext();
		else
			return null;
	}

	/**
	 * Clears all cached data for this controller (especially the view and the component).
	 *
	 * <P>This method may be invoked for memory optimization, but the controller is still valid and
	 * may be used. (Unlike free) Further usage, needs to rebuild internal component & view which
	 * may be time consuming.</P>
	 */
	public void clearViewCache()
	{
		if ((_areaController != null)
				&& (_component != null))
		{
			if (!((LyView)_component).isVisible())
			{
				// Free all view components
				LyView view = (LyView)_component;

				view.free();
				_viewInitialized = false;

				super.clearViewCache();

				if (_areaController != null)
					_areaController.clearViewCache();
			}
		}
	}

	/**
	 * Interrupts printing thread registered on this controller.
	 */
	@SuppressWarnings("deprecation")
	public void interruptPrintThread()
	{
		if (_printThread != null)
		{
			_printThread.stop();
			_printThread = null;
		}
	}

	/**
	 * Sets the printing thread for this controller.
	 *
	 * @param thread printing thread
	 */
	public void setPrintThread(Thread thread)
	{
		_printThread = thread;
	}

	/**
	 * Indicates if there is a printing thread registered on this controller.
	 *
	 * @return true if there is a printing thread registered on this controller, false otherwise.
	 */
	public boolean hasPrintThread()
	{
		return (_printThread != null);
	}

	/**
	 * Accessor to the current Environment instance This method returns the environment of the area
	 * controller displayed by this view controller.
	 *
	 * @return current environment instance.
	 */
	public LyEnvironment getEnvironment()
	{
		if ((_areaController != null) && (!_areaController.isFreed()))
			return _areaController.getEnvironment();
		else
			return super.getEnvironment();
	}

	/**
	 * Sets idle mode on this controller. Calls setIdle on the area controller.
	 * @param idle the desired idle mode
	 */
	public void setIdle(boolean idle)
	{
		if (_areaController != null)
			_areaController.setIdle(idle);

		super.setIdle(idle);
	}

	/**
	 * Turn temporarely the view to visible/not visible.
	 *
	 * <P>The view is only HIDDEN, not cleared (see free), so it should be possible to open it again
	 * with setVisible and ressources allocated for the view are not freed.</P>
	 * @param visible indicates if the view must be shown
	 */
	protected void setVisible(boolean visible)
	{
		if (!_viewInitialized)
			return;

		LyView view = (LyView)_component;

		if ((view == null)
				|| view.isFreeing()
				|| view.isFreed())
			return;

		view.setVisible(visible);
		_shown = visible;

		visibilityChanged(visible);

		if ((_areaController != null) && (!_areaController.isFreed()))
			_areaController.visibilityChanged(visible);
	}

	/**
	 * Set the location of the view.
	 *
	 * @param x the new x position of the view
	 * @param y the new y position of the view
	 */
	protected void setLocation(int x, int y)
	{
		if (!_viewInitialized)
			return;

		LyView view = (LyView)_component;

		if ((view == null)
				|| view.isFreeing()
				|| view.isFreed())
			return;

		view.setLocation(x, y);
	}

	/**
	 * Check if this controller as son window showing (may decide if we free it or not).
	 * @return true if a son is visible
	 */
	public boolean hasSonShowing()
	{
		// Check recursively son controllers
		LyViewController controller;
		int				 i, n = getSonsCount();

		for (i = n - 1; i >= 0; i--)
		{
			controller = getSon(i);
			if ((controller != null) && (!controller.isFreed()) && (controller.isShown()))
				return true;
		}
		return false;
	}

	/**
	 * @return Returns the _inShowProcess.
	 */
	public boolean isInShowProcess()
	{
		return _inShowProcess;
	}

	/**
	 * @param showProcess The _inShowProcess to set.
	 */
	public void setInShowProcess(boolean showProcess)
	{
		_inShowProcess = showProcess;
	}

	/**
	 * Accessor to popup view controller stack (internal use).
	 *
	 * @return the sub controller stack
	 */
	public Stack<LyViewController> getControllerStack()
	{
		return _subControllerStack;
	}

	/**
	 * Accessor to popup view controller stack (internal use).
	 *
	 * @param controllerStack the sub controller stack
	 */
	public void setControllerStack(Stack<LyViewController> controllerStack)
	{
		_subControllerStack = controllerStack;
	}

    /**
     * Sets view name.
     */
    public void setName(String name)
    {
       LyView view = getView();

       super.setName(name);
       view.setLabel(name);
    }
} // end class LyViewController
