/**
 * LyWaitController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyWaitController.java 31261 2011-05-31 14:45:11Z jlsaouli $
 */

package leon.control;

import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.view.*;


/**
 * This class implements a wait controller. This controller waits for a given amount of time, before
 * executing a new action. It is associated with a builder (LyWaitBuilder) and a behavior that
 * implements LyWaitBehaviorInterface. Usually, this behavior is an instance of LyWaitBehavior
 *
 * @see LyWaitBuilder
 * @see leon.app.behaviorinterface.LyWaitBehaviorInterface
 * @see leon.app.behavior.LyWaitBehavior
 */
public class LyWaitController extends LyAreaController
	implements Runnable
{
	/** File Version */
	public static final String __VERSION =
		"$Id: LyWaitController.java 31261 2011-05-31 14:45:11Z jlsaouli $";

	/** Periodic check of client for value (if client specified). */
	protected long			   _updateDelay;

	/** Minimum value for the progress bar. */
	protected int			   _min;

	/** Maximum value for the progress. */
	protected int			   _max;

	/** Number of columns for the progress bar. */
	protected int			   _nCols;

	/** Number of rows for the progress bar. */
	protected int			   _nRows;

	/** The thread to manage the progress bar. */
	protected Thread		   _thread;

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller). Controller hierarchy
	 *                indicates a child controller (for example a dialog created from a main view)
	 * @param action  related action. If an action is specified, the controller uses action
	 *                parameters (action name, action specific XML view description ...)
	 * @param classInfo related classInfo.
	 * @param context current context (may be null if the view is not contextual to a list of
	 *                objects).
	 */
	public LyWaitController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent, action, context);

		initWaitController();

		// If action is null, use default
		if (_action != null)
		{
			_min = _action.getMin();
			_max = _action.getMax();
		}
	}

	/**
	 * Default constructor.
	 * 
	 * @param session current session 
	 * @param parent window controller
	 * @param action associated to the wait window.
	 * @param min    minimum value.
	 * @param max    maximum value.
	 * @param nCols  preferred number of columns -chars- in notice. -1 (default) is auto-size.
	 * @param nRows  preferred number of rows -lines- in notice. -1 (default) is auto-size.
	 */
	public LyWaitController(LySession session, LyController parent, LyAction action,
		int min, int max, int nCols, int nRows)
	{
		super(session, parent, action, null);

		initWaitController();

		_min = min;
		_max = max;
		_nRows = nRows;
		_nCols = nCols;
	}

	/**
	 * Initialise the default values of this controller.
	 *
	 * <UL>
	 *   <LI>update delay: 150 ms</LI>
	 *   <LI>minimum value: 0%</LI>
	 *   <LI>maximum value: 100%</LI>
	 *   <LI>number of rows: auto (-1)</LI>
	 *   <LI>number of columns: auto (-1)</LI>
	 * </UL>
	 */
	protected void initWaitController()
	{
		_updateDelay = 150; // (ms)
		_min = 0;
		_max = 100;
		_nRows = -1; // Autosize
		_nCols = -1; // Autosize

		setName(_action.getName());
	}

	/**
	 * Method that gives the interface that must be implemented by the behavior associated to this
	 * controller.
	 *
	 * @return Class representing the behavior interface of this controller.
	 */
	public Class<?> getBehaviorInterface()
	{
		return LyWaitBehaviorInterface.class;
	}

	/**
	 * Prepare component before showing. This method is called once, before first showing.
	 */
	protected void initArea()
	{
		super.initArea();

		if (isFreed())
			return;

		_initialized = false;

		// if client is set, start Thread
		if (getBehavior() != null)
		{
			_thread = new Thread(this, "BF AppEngine : waiting action thread");
			_thread.start();
		}

		_initialized = true;
	}

	/**
	 * Reset all. This method may be called several times.
	 */
	public synchronized void reset()
	{
		if (isFreed())
			return;

		super.reset();

		LyWaitBehaviorInterface behavior = (LyWaitBehaviorInterface)getBehavior();

		if (behavior != null)
		{
			LyGauge gauge = getGaugeComponent();
			LyLabel label = getLabelComponent();
			LyIcon  icon = getIconComponent();
			String  str, image;
			int     value;

			value = behavior.getWaitValue(this);
			str = behavior.getWaitLabel(this);
			image = behavior.getWaitIcon(this);

			if (gauge != null)
				gauge.setNValue(value);
			if (label != null)
				label.setLabel(str);
			if ((icon != null) && (image != null))
				icon.setImage(image);
		}
	} // end method reset

	/**
	 * Show the view and start the waiting thread.
	 */
	public void showView()
	{
		LyWaitBehaviorInterface behavior = (LyWaitBehaviorInterface)getBehavior();

		if (behavior != null)
			behavior.startWaiting(this);

		super.showView();
	}

	/**
	 * This method is called when the wait is over If the ok parameter is 'true', this method invoke
	 * the next action.
	 *
	 * @param ok true if the wait was successull
	 */
	public void endWait(boolean ok)
	{
		hideView();

		if (ok)
			invokeNextActions();

		free();
	}

	/**
	 * Proceed to next actions after a login (login should be valid at that point).
	 * @return a list of invoked controllers
	 */
	public LyController[] invokeNextActions()
	{
		LySession		 session = getSession();
		LyAction		 action;
		int				 i, n = _action.getActionCount();
		LyViewController viewController = getViewController();
		LyController[]   controllers = new LyController[n];

		for (i = 0; i < n; i++)
		{
			action = _action.getAction(i);
			controllers[i] = session.invokeAction(viewController.getParent(), action, null, null,
					true);
		}

		return controllers;
	}

	/**
	 * Free this controller.
	 *
	 * @param controller parent controller
	 */
	public synchronized void free(LyController controller)
	{
		if (_parent != controller)
			return;

		_updateDelay = -1; // End Thread whatever

		LyWaitBehaviorInterface behavior = (LyWaitBehaviorInterface)getBehavior();

		if (behavior != null)
			behavior.endWaiting(this);

		super.free(controller);
	}

	/**
	 * Implementation of the runnable interface Periodic check of the client : invoke free to stop.
	 */
	public void run()
	{
		if (isFreed())
			return;

		startRun();

		_LyWaitControllerUpdater updater = new _LyWaitControllerUpdater();
		LyViewManager			 viewManager = getViewManager();

		while (_updateDelay > 0)
		{
			viewManager.execGui(updater, true);

			try
			{
				// updateDelay can change if controller is freed
				if (_updateDelay > 0)
					Thread.sleep(_updateDelay);
			}
			catch (Exception e)
			{
				getEnvironment().writeException(e);
			}

			setLastModified((new Date()).getTime());
		}

		endRun();
	} // end method run

	class _LyWaitControllerUpdater
		implements Runnable, LyGetSessionInterface
	{
		public void run()
		{
			LyWaitBehaviorInterface behavior = (LyWaitBehaviorInterface)getBehavior();

			// Refresh update
			reset();

			if (behavior != null)
				behavior.increaseOffset(LyWaitController.this);
		}
		
		public LyDataSession getSession()
		{
			return LyWaitController.this.getSession();
		}
	}

	/**
	 * Indicates if the view for this wait controller is resizable.
	 *
	 * @return true if the view is resizable, false otherwise.
	 */
	public boolean isViewResizable()
	{
		return false;
	}

	/**
	 * Retrieves the view Builder.
	 *
	 * @return the builder instance;
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyWaitBuilder(this);
	}

	/**
	 * Accessor to the Gauge component.
	 *
	 * @return the gauge view component
	 */
	public LyGauge getGaugeComponent()
	{
		// Retrieve internal map component
		LyComponent component = getComponent();
		String	    id = getId() + "_notice";

		LyNotice    notice = (LyNotice)component.getComponent(id);

		return ((notice == null) ? null : notice.getGauge());
	}

	/**
	 * Accessor to the Icon component.
	 *
	 * @return the icon view component
	 */
	public LyIcon getIconComponent()
	{
		LyComponent component = getComponent();
		String	    id = getId() + "_notice";

		LyNotice    notice = (LyNotice)component.getComponent(id);

		return ((notice == null) ? null : notice.getIcon());
	}

	/**
	 * Accessor to the Label component.
	 *
	 * @return the label view component
	 */
	public LyLabel getLabelComponent()
	{
		LyComponent component = getComponent();
		String	    id = getId() + "_notice";

		LyNotice    notice = (LyNotice)component.getComponent(id);

		return ((notice == null) ? null : notice.getLabel());
	}

	/**
	 * Set periodic update delay on client.
	 *
	 * @param delay refresh interval (in ms)
	 */
	public void setUpdateDelay(long delay)
	{
		_updateDelay = delay;
	}

	/**
	 * Accessor to current minimum value.
	 *
	 * @return the minimum value
	 */
	public int getNMin()
	{
		return _min;
	}

	/**
	 * Accessor to current maximum value.
	 *
	 * @return the maximum value
	 */
	public int getNMax()
	{
		return _max;
	}

	/**
	 * Accessor to current number of rows.
	 *
	 * @return the number of rows
	 */
	public int getNRows()
	{
		return _nRows;
	}

	/**
	 * Accessor to current number of columns.
	 *
	 * @return the number of columns
	 */
	public int getNCols()
	{
		return _nCols;
	}
    
    /**
     * Wait panel is closable only when action has a commands bar.
     * @param showMessage indicates if a message must be shown
     * @return true if the controller can be closes
     */
    public boolean canClose(boolean showMessage)
    {
        return !_action.hasMark(LyAction.MARK_NO_COMMANDS);
    }
} // end class LyWaitController
