/**
 * LyAreaBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyAreaBuilder.java 40687 2012-09-06 15:07:47Z jlsaouli $
 */

package leon.control.builder;

import java.util.*;

import leon.app.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.mobile.*;
import leon.view.web.mobile.*;


/**
 * This class implements a basic builder for all controller based on a LyAreaController. It handles
 * all common components (toolbar, titlebar messages bar, menubar...)
 *
 * @see LyAreaController
 */
public abstract class LyAreaBuilder extends LyBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyAreaBuilder.java 40687 2012-09-06 15:07:47Z jlsaouli $";

	/** The left label of the titleBar. */
	String					   _leftLabel;

	/** The center label of the titleBar. */
	String					   _centerLabel;

	/** The right label of the titleBar. */
	String					   _rightLabel;

	/** The left image of the titleBar. */
	protected String			_leftImage;

	/** The right image of the titleBar. */
	protected String			_rightImage;

	/** Associations between actions prefix and instances of LyArea. */
	Hashtable<String, LyArea>  _actionAreas;

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyAreaBuilder(LyController parent)
	{
		super(parent);

		// By default, all labels and image are computed
		_leftLabel = null;
		_centerLabel = null;
		_rightLabel = null;
		_leftImage = null;
		_rightImage = null;
		_actionAreas = null;
	}

	/**
	 * Builds the view component for the specified parent controller.
	 *
	 * @return the view Component
	 */
	public LyComponent buildComponent()
	{
		if (_component != null)
			return _component;

		LyArea		  area = (LyArea)super.buildComponent();
		LyEnvironment environment = ((LyAreaController)_parent).getEnvironment();
		LyAction	  action = ((LyAreaController)_parent).getAction();

		if ((action != null)
				&& (action.getTip() != null))
			area.setTip(action.getTip());
		
				
		if(action != null && action.hasMark(LyAction.MARK_MOBILEWEBVIEW))
		{
			area.addData(LyMobileConstants.WEBVIEW_MOBILE_KEY, LyMobileConstants.TRUE);
		}
		else
		{
			LyAreaController parentController = ((LyAreaController)_parent).getParentAreaController();
			boolean mobileAction = false;
			
			while(parentController != null && ! mobileAction)
			{
				LyAction parentAction = parentController.getAction();
				if(parentAction != null && parentAction.hasMark(LyAction.MARK_MOBILEWEBVIEW))
				{
					area.addData(LyMobileConstants.WEBVIEW_MOBILE_KEY, LyMobileConstants.TRUE);
					mobileAction = true;
				}
				parentController = parentController.getParentAreaController();
			}
		}

		area.setEnvironment(environment);
		return area;
	}

	/**
	 * This method allows to set title labels and icons.
	 *
	 * @param      leftLabel   the left label
	 * @param      centerLabel the center label
	 * @param      rightLabel  the right label
	 * @param      leftImage   the left image
	 * @param      rightImage  the right image
	 * @deprecated Use behavior
	 */
	public void setTitleBarParameters(String leftLabel, String centerLabel, String rightLabel,
		String leftImage, String rightImage)
	{
		_leftLabel = leftLabel;
		_centerLabel = centerLabel;
		_rightLabel = rightLabel;
		_leftImage = leftImage;
		_rightImage = rightImage;

		// Update the titlebar
		updateTitleBar();
	}

	/**
	 * Returns the main titlebar of the area component.
	 */
	protected LyTitleBar getTitleBar()
	{
        LyArea     area = (LyArea)_component;
        
        return area.getTitleBar();
	}
	
	/**
	 * Update titlebar and set default content.
	 */
	public void updateTitleBar()
	{
		if (_component == null)
			return;

        LyArea     area = (LyArea)_component;
        LyTitleBar titleBar = getTitleBar();
        
		if (titleBar != null)
		{
			LyAreaController parent = (LyAreaController)_parent;
			LyAction		 action = parent.getAction();
			LyEnvironment    environment = parent.getEnvironment();

			if (_rightImage != null)
				titleBar.setRightImage(_rightImage);
			else
			{
				String key = "LY_TITLE_BAR_RIGHT_IMAGE";
				String rightImage = environment.getImagePath(key);

				if ((rightImage != null)
						&& !rightImage.endsWith(key)) // XXX SEE LyImage. Remove This
					titleBar.setRightImage(key);
			}

			if (_leftLabel != null)
				titleBar.setLeftLabel(_leftLabel);
			else if (action != null)
				titleBar.setLeftLabel(parent.getTitle());

			if (_centerLabel != null)
				titleBar.setCenterLabel(_centerLabel);

			if (_rightLabel != null)
				titleBar.setRightLabel(_rightLabel);

			titleBar.setRightLabelSensitive(true);

			if (environment.testResource("LY_TITLEBAR_LEFT_IMAGE"))
			{
				if (_leftImage != null)
				{
					titleBar.setImage(_leftImage);
					titleBar.setImageSensitive(true);
				}
				else if (action != null)
				{
					titleBar.setImage(parent.getImage());
					titleBar.setImageSensitive(true);
				}
			}

			// Merge toolbar and titlebar
			if (allowMergeTitlebarToolbar()
					&& environment.testResource("LY_MERGE_TOOLBAR_TITLEBAR"))
			{
				LyToolBar toolbar = area.getToolBar();

				if (toolbar != null)
				{
					area.setToolBar(null);
					titleBar.setToolBar(toolbar);
				}
			}
		}
	} // end method updateTitleBar

	/**
	 * Indicates if merging of toolbar and titlebar is allowed.
	 *
	 * @return true if merging of toolbar and titlebar is allowed, false otherwise.
	 */
	protected boolean allowMergeTitlebarToolbar()
	{
		return true;
	}

	/**
	 * Create a titlebar for this Area.
	 *
	 * @return the titlebar Component
	 */
	protected LyTitleBar buildTitleBar()
	{
		LyAction action = ((LyAreaController)_parent).getAction();

		// FFT00109.
		if ((action == null) || (action.hasMark(LyAction.MARK_NO_TITLEBAR)))
			return null;

		LyTitleBar titleBar = new LyTitleBar();

		titleBar.setId(_parent.getId() + "_titleBar");

		return titleBar;
	}

	/**
	 * Builds the toolbar for this area.
	 *
	 * @return the toolbar component
	 */
	protected LyToolBar buildToolBar()
	{
		LyAction action = ((LyAreaController)_parent).getAction();

		if (action.hasMark(LyAction.MARK_NO_TOOLBAR))
			return null;

		return completeToolBar(null);
	}

	/**
     * Builds the right toolbar for this area.
     *
     * @return the right toolbar component.
     */
    protected LyToolBar buildRightToolBar()
    {
        LyEnvironment environment = _parent.getEnvironment();
        
        if (!environment.testResource("LY_RIGHT_TOOLBAR"))
            return null;
        
        LyAction action = ((LyAreaController)_parent).getAction();

        if (action.hasMark(LyAction.MARK_NO_TOOLBAR))
            return null;

        return completeRightToolBar(null);
    }
    
	/**
	 * Builds the toolbar for this area.
	 *
	 * @param  toolBar the tolbar to complete
	 * @return the completed toolbar component
	 */
	protected LyToolBar completeToolBar(LyToolBar toolBar)
	{
		LyInfoList<LyAction> actions = getToolbarActions((LyAreaController)_parent, null);
		toolBar = completeToolBar(toolBar, actions);
		return toolBar;
	}

	/**
	 * Get a list of action to build a toolbar for a given info
	 * @param controller the controller
	 * @param info the info
	 * @return a list of action
	 */
	protected LyInfoDataList<LyAction, LyClassInfo> getToolbarActions(LyAreaController controller, LyInfo info)
	{
		LyInfoList<LyAction> actions = null;
		
		if (info == null)
			actions = controller.getAction().getActions();
		else if (info instanceof LyAction)
			actions = ((LyAction) info).getActions();
		else if (info instanceof LyClassInfo)
			actions = ((LyClassInfo)info).getActions();
		
		if (actions == null)
			return null;
		
		LyInfoDataList<LyAction, LyClassInfo> actionsList = new LyInfoDataList<LyAction, LyClassInfo>(true);
		actionsList.addList(actions);
		
		return actionsList;
	}
	
	
	/**
	 * Builds the toolbar for this area.
	 *
	 * @param  toolBar the tolbar to complete
	 * @param  actions List of actions to add in the toolbar.
	 * @return the completed toolbar component
	 */
	protected LyToolBar completeToolBar(LyToolBar toolBar, LyInfoList<LyAction> actions)
	{
		LyAction  action = ((LyAreaController)_parent).getAction();
		LyToolBar result;

		if (actions == null)
			result = toolBar;
		else if (toolBar == null)
			result = buildToolBar(actions);
		else
		{
			// Check if a separator must be added
			int		 i, j, m = toolBar.getComponentCount(), n = actions.getSize();
			LyAction subAction;
			LyTool   tool;
			String   group;
			String   iconPath;
			boolean  addSep = false;

			for (i = 0; (i < n) && !addSep && (m > 0); i++)
			{
				subAction = actions.getInfo(i);
				if (!validateAction(subAction))
					continue;

				group = subAction.getGroup();

				if (group != null)
					addSep = true;

				iconPath = _parent.getImage(subAction);

				// Check if we already have a tool with the same image
				for (j = 0; j < m; j++)
				{
					if (toolBar.getComponent(j) instanceof LyTool)
					{
						tool = (LyTool)toolBar.getComponent(j);
						if ((tool != null)
								&& (tool.getImage() != null)
								&& tool.getImage().equals(iconPath))
							break;
					}
				}

				LyEnvironment env = ((LyAreaController)_parent).getEnvironment();

				if ((j == m) || !env.testResource("LY_GROUP_ACTIONS_BY_IMAGE"))
					addSep = true;
			}

			if (addSep)
				toolBar.addSep();

			addTools(toolBar, actions);

			result = toolBar;
		}

		// Check if help button must be added. Add it if action has no menu bar and
		// if action has a help file.
		if ((result != null) && action.hasMark(LyAction.MARK_NO_MENUBAR))
		{
			String helpFile = ((LyAreaController)_parent).getHelpFile();

			if ((helpFile != null) && (helpFile.trim().length() > 0))
			{
				LyTool helpTool = new LyTool(LyController.DEMAND_HELP, "LY_HELP");

				helpTool.setImage("LY_HELP");
				helpTool.setDialog(true);
				helpTool.setTooltip("LY_HELP");
				result.addSep();
				result.addTool(helpTool);
			}
		}

		return result;
	} // end method completeToolBar

	/**
	 * Builds the toolbar for this area.
	 *
	 * @param  toolbar the toolbar to complete
	 * @return the completed toolbar component
	 */
	protected LyToolBar completeRightToolBar(LyToolBar toolbar)
	{
		LyAction	  action = ((LyAreaController)_parent).getAction();
		LyEnvironment environment = _parent.getEnvironment();
		boolean		  isDashboard = (action != null)
				&& (action.getRoot().getId() == LyAction.ID_DASHBOARD);
		LyTool		  tool;

		if (!isDashboard)
		{
			String delayStr = environment.getResource("LY_DATA_UPDATE_DELAY");
			int    delay = ((delayStr == null) ? -1 : LyTools.intFromString(delayStr));

			if (delay >= 0)
			{
				// We add them in the right toolBar
				if (toolbar == null)
				{
					toolbar = new LyToolBar();
					toolbar.setFill(LyToolBar.Fill.HORIZONTAL);
					toolbar.setStyle(LyToolBar.Style.SEPARATOR);
					toolbar.setHAlign(LyToolBar.HAlign.RIGHT);
					toolbar.setVAlign(LyToolBar.VAlign.TOP);
				}
				else if (toolbar.getComponentCount() > 0)
					toolbar.addSep();

				String refreshLabel = environment.translate(environment.translate("LY_TOOL_UPDATE"));
				if(refreshLabel != null && refreshLabel.trim().length() > 0)
				{
					tool = new LyTool(LyViewController.DEMAND_UPDATE, "LY_TOOL_UPDATE");
					tool.setImage("LY_TOOL_UPDATE");
					tool.setDialog(false);
					tool.setEnabled(true);
					toolbar.addTool(tool);
				}
			}
		}
		return toolbar;
	} // end method completeRightToolBar

	/**
	 * Builds the toolbar from a list of object infos. Methods recurses into inner toolbars.
	 *
	 * @param  actions description of the actions and action groups for the current application
	 *                 class.
	 * @return the toolBar
	 */
	protected LyToolBar buildToolBar(LyInfoList<LyAction> actions)
	{
		if ((actions == null) || (actions.getSize() == 0))
			return null;

		LyToolBar toolBar = new LyToolBar();

		toolBar.setId("_toolBar");
		toolBar.setFill(LyToolBar.Fill.HORIZONTAL);
		toolBar.setHAlign(LyToolBar.HAlign.LEFT);
		toolBar.setVAlign(LyToolBar.VAlign.TOP);
		addTools(toolBar, actions);

		return ((toolBar.getComponentCount() == 0) ? null : toolBar);

	}

	/**
	 * Add tools to the given toolBar.
	 *
	 * @param toolBar the toolBar to complete
	 * @param actions List of the actions for which corresponding tools are added.
	 */
	protected void addTools(LyToolBar toolBar, LyInfoList<LyAction> actions)
	{
		LyAction	   action;
		String		   group;
		Vector<String> groups = new Vector<String>();
		int			   i, n = actions.getSize();
		LyToolBar	   subToolBar;

		for (i = 0; i < n; i++)
		{
			action = actions.getInfo(i);
			if (!validateAction(action))
				continue;

			group = action.getGroup();
			
			if ((action.getRoot().getId() == LyAction.ID_DASHBOARD)
				&& expandDashboards())
			{
				toolBar.addSep();
				addTools(toolBar, getToolbarActions((LyAreaController)_parent, action));
				continue;
			}

			if (group == null)
				buildTool(toolBar, action, true, null, null);
			else if (!groups.contains(group))
			{
				subToolBar = buildToolBar(group, actions);
				toolBar.addToolBar(subToolBar);
				groups.addElement(group);
			}
		}

		if (groups.size() == 0)
			toolBar.setStyle(LyToolBar.Style.SEPARATOR);

		toolBar.setFill(LyToolBar.Fill.HORIZONTAL);
	} // end method addTools

	/**
	 * Indicates if dashboards are expanded.
	 * 
	 * @return True if expanded, false otherwise
	 */
	protected boolean expandDashboards()
	{
		return false;
	}
	
	/**
	 * Add items to the given menuBar.
	 *
	 * @param menuBar the menubar to complete
	 * @param actions List of the actions for which corresponding items are added in menu bar.
	 */
	protected void addMenuItems(LyMenuBar menuBar, LyInfoList<LyAction> actions)
	{
		LyAction		 action;
		String			 group, menuName, name, translatedName;
		int				 i, j, n = actions.getSize();
		LyMenu			 menu;
		LyItem			 item;
		LyComponent      lastComp;
		Vector<LyAction> actionItems = new Vector<LyAction>();
		Vector<String>   newGroups;
		LyEnvironment    environment = _parent.getEnvironment();

		for (i = 0; i < n; i++)
		{
			action = actions.getInfo(i);
			if (actionItems.contains(action))
				continue;

			menuName = action.getMenu();
			group = action.getGroup();
			newGroups = new Vector<String>();

			for (j = i; j < n; j++)
			{
				action = actions.getInfo(j);
				if ((action.getMenu() != menuName)
						|| (action.getGroup() != group))
					continue;

				if (!validateAction(action, false))
				{
					actionItems.add(action);
					continue;
				}

				if (menuName == null)
				{
					if (action.getMax() == 0)
						// Default menu is File menu
						name = "LY_MENU_FILE".intern();
					else
						// Default menu is Actions menu
						name = "LY_MENU_ACTIONS".intern();
				}
				else
					name = menuName;

				translatedName = environment.translate(name);
				if ((translatedName == null) || (translatedName.length() == 0))
					continue;

				menu = (LyMenu)menuBar.getComponentByLabel(name);
				
				if (menu == null)
				{
					menu = new LyMenu(name);
					menuBar.addMenu(menu);
					newGroups.add(name);
				}

				if ((name == "LY_MENU_ACTIONS".intern()) && (group != null))
				{
					LyMenu	groupMenu = (LyMenu)menu.getComponent(group + "_menuGroup");
					
					if (groupMenu == null)
					{
						groupMenu = new LyMenu(group);
						groupMenu.setId(group + "_menuGroup");
						menu.addMenu(groupMenu);
					}

					menu = groupMenu;
				}
				else if (!newGroups.contains(name))
				{
				    lastComp = menu.getComponent(menu.getComponentCount() - 1);
                    if ((lastComp != null)
                            && (lastComp.getId() != LyComponent.SEP_ID))
						menu.addSep();
					
					newGroups.add(name);
				}

				item = buildItem(action);
				menu.addItem(item);

				actionItems.add(action);
			}
		}
	} // end method addMenuItems

	/**
	 * Check if the action should be added in the toolbar or the menubar Return true if not hidden,
	 * has the tool mark (or not notMenu mark) set and if session validates the action
	 * (validateAction).
	 *
	 * @param  action the action we want to test
	 * @param  tool   true if action is used as tool
	 * @return true if the action should be added, false otherwise
	 */
	protected boolean validateAction(LyAction action, boolean tool)
	{
		LySession session = _parent.getSession();

		if (action.hasMark(LyAction.MARK_HIDDEN))
			return false;
		if (tool && !action.hasMark(LyAction.MARK_TOOL))
			return false;
		if (!tool && action.hasMark(LyAction.MARK_NO_MENU))
			return false;
		if (!session.validateAction(action, _parent))
			return false;

		return true;
	}

	/**
	 * Check if the action should be added in the toolbar Return true if not hidden, has the tool
	 * mark set and if session validates the action (validateAction).
	 *
	 * @param  action the action we want to test
	 * @return true if the action should be added, false otherwise
	 */
	protected boolean validateAction(LyAction action)
	{
		return validateAction(action, true);
	}

	/**
	 * Builds the toolbar from a list of object infos. Methods recurses into inner toolbars.
	 *
	 * @param  group   name of the group of action
	 * @param  actions list of actions
	 * @return the toolBar
	 */
	protected LyToolBar buildToolBar(String group, LyInfoList<LyAction> actions)
	{
		LyToolBar toolBar;
		LyAction  action;
		int		  i, n = actions.getSize();

		String grouptrimed = null;
		if(group != null)
			grouptrimed = group.replace(' ', '_').replace('\\', '_').replace("/", "_");
		
		toolBar = new LyToolBar();
		toolBar.setId("_toolBar" + grouptrimed + "_" + _parent.getId());
		toolBar.setLabel(group);
		toolBar.setStyle(LyToolBar.Style.BORDER);
		toolBar.setFill(LyToolBar.Fill.HORIZONTAL);

		for (i = 0; i < n; i++)
		{
			action = actions.getInfo(i);
			if (action.getGroup() != group)
				continue;
			if (!validateAction(action))
				continue;

			buildTool(toolBar, action, true, null, null);
		}

		return toolBar;
	} // end method buildToolBar

	/**
	 * Builds a tool description for a specified action.
	 *
	 * @param  toolBar      toolbar in which the new tool is added.
	 * @param  action       the action for the tool
	 * @param  enable       enabled state for the tool
	 * @param  parameter    parameters for the associated action
	 * @param  iconSizeInfo size of the associated icon
	 * @return the tool
	 */
	protected LyTool buildTool(LyToolBar toolBar, LyAction action, boolean enable,
		String parameter, String iconSizeInfo)
	{
		if (toolBar == null)
			return buildTool(action, enable, parameter, iconSizeInfo);

		String iconPath = _parent.getImage(action);
		int    i, n = toolBar.getComponentCount();
		LyTool tool = null;

		// Check if we already have a tool with the same image
		for (i = 0; i < n; i++)
		{
			if (toolBar.getComponent(i) instanceof LyTool)
			{
				tool = (LyTool)toolBar.getComponent(i);
				if ((tool != null)
						&& (tool.getImage() != null)
						&& tool.getImage().equals(iconPath))
					break;
			}
		}

		LyEnvironment env = ((LyAreaController)_parent).getEnvironment();

		if ((i == n) || !env.testResource("LY_GROUP_ACTIONS_BY_IMAGE") || getViewManager().hasProperty(LyMobileViewManager.MOBILE_ONLY_PROPERTY))
		{
			tool = buildTool(action, true);
			toolBar.addTool(tool);
			return tool;
		}

		LyToolBar subToolBar = ((tool.getComponent() instanceof LyArea)
				? ((LyArea)tool.getComponent()).getLeftToolBar() : null);
		LyTool    newTool;

		if (subToolBar == null)
		{
			LyArea area = new LyArea();

			subToolBar = new LyToolBar();
			subToolBar.setId("_toolBar" + LyTools.getNextIdentifier()
				+ "_" + _parent.getId());
			subToolBar.setFill(LyToolBar.Fill.VERTICAL);
			subToolBar.setHAlign(LyToolBar.HAlign.LEFT);
			subToolBar.setVAlign(LyToolBar.VAlign.MIDDLE);
			subToolBar.setController(_parent.getId());

			newTool = new LyTool(tool.getId(), tool.getLabel());
			newTool.setTooltip(tool.getTooltip());
			newTool.setImage(tool.getImage());
			newTool.setEnabled(tool.isEnabled());
			newTool.setDialog(tool.isDialog());
			subToolBar.addTool(newTool);

			env = new LyEnvironment(env);
			env.putResource("LY_WRITE_TOOL_LABEL", "true");
			env.putResource("LY_TOOL_LABEL_RIGHT_POSITION", "true");
			area.setEnvironment(env);

			LyConstraint	constraint = new LyConstraint();
			
			constraint.setFill(LyConstraint.Fill.NONE);
			area.addToolBar(subToolBar);
			area.setConstraint(constraint);
			tool.setDialog(false);
			tool.setComponent(area);
		}

		newTool = buildTool(action, true);
		subToolBar.addTool(newTool);
		tool.setTooltip(tool.getTooltip() + '\n' + newTool.getTooltip());

		return newTool;
	} // end method buildTool
	
	/**
	 * Builds a tool description for a specified action.
	 *
	 * @param  action       the action for the tool
	 * @param  enable       enabled state for the tool
	 * @param  parameter    parameters for the associated action
	 * @param  iconSizeInfo size of the associated icon
	 * @return the tool
	 */
	protected LyTool buildTool(LyAction action, boolean enable, String parameter,
		String iconSizeInfo)
	{
		String toolId = action.getId();

		if (parameter != null)
			toolId += LyController.ID_SEP + parameter;

		LyTool tool = new LyTool(toolId, action.getAlias());
		String iconPath = _parent.getImage(action);

		if (iconSizeInfo != null)
			iconPath = iconPath + iconSizeInfo;

		tool.setTooltip(action.getName());
		tool.setImage(iconPath);
		tool.setEnabled(enable);
		tool.setDialog(action.hasMark(LyAction.MARK_DIALOG));

		if (action.getRoot().getId() == LyAction.ID_DASHBOARD)
		{
			LyArea    area = new LyArea();
			LyToolBar toolBar = buildToolBar(getToolbarActions((LyAreaController)_parent, action));

			toolBar.setFill(LyToolBar.Fill.HORIZONTAL);
			toolBar.setController(_parent.getId());
			toolBar.setId("_toolBar" + toolId + "_" + _parent.getId());
			toolBar.setHAlign(LyToolBar.HAlign.LEFT);
			toolBar.setVAlign(LyToolBar.VAlign.TOP);
			LyDashboardBuilder.setSubToolBarFill(toolBar, LyToolBar.Fill.DEFAULT);

			area.setToolBar(toolBar);
			tool.setComponent(area);
		}

		// Make a list with all the files that begin by the name of the parameter _selectFile
		// (if there is one)
		String prefix = (String)action.getParameter("_selectFile", true);

		if (prefix != null)
				tool.setComponent(buildSelectFile(prefix));

		return tool;
	} // end method buildTool

	/**
	 * Builds a tool description for a specified action.
	 *
	 * @param  action the action for the tool
	 * @param  enable enabled state for the tool
	 * @return the tool
	 */
	protected LyTool buildTool(LyAction action, boolean enable)
	{
		return buildTool(action, enable, null, null);
	}

	/**
	 * Builds a menu item for a specified action.
	 *
	 * @param  action    the action for the item
	 * @param  classInfo the classInfo on which action applied to.
	 * @return the menu item
	 */
	protected LyItem buildItem(LyAction action, LyClassInfo classInfo)
	{
		String id = ((classInfo == null)
				? action.getId() : (action.getId() + LyController.ID_SEP + classInfo.getId()));
		String name = action.getName();
		LyItem item = new LyItem(id, name);

		item.setController(_parent.getId());
		if (action.hasMark(LyAction.MARK_DIALOG))
			item.setDialog(true);
		item.setShortcut(action.getShortcut());

		item.setImage(((LyAreaController)_parent).getMenuItemImage(action));

		return item;
	}

	/**
	 * Builds a menu item for a specified action.
	 *
	 * @param  action the action for the item
	 * @return the menu item
	 */
	protected LyItem buildItem(LyAction action)
	{
		return buildItem(action, null);
	}

	/**
	 * Build a list of all data files.
	 *
	 * @param  prefix The prefix of the files.
	 * @return an area containing the list of files.
	 */
	protected LyArea buildSelectFile(String prefix)
	{
		// Create hrefs
		LyArea    area = updateSelectFile(prefix);

		// Register controller on session to listen to info objects events
		LySession session = _parent.getSession();

		session.addInfoListener((LyAreaController)_parent, prefix);

		return area;
	}

	/**
	 * Updates the list of all data files.
	 *
	 * @param  prefix The prefix of the files.
	 * @return an area containing the list of files.
	 */
	
	public LyArea updateSelectFile(String prefix)
	{
		LyArea area = null;

		if ((_actionAreas != null)
				&& (_actionAreas.get(prefix) != null))
			area = _actionAreas.get(prefix);

		// Create hrefs
		LyApplication		    application = _parent.getApplication();
		LyInfoList<LyChartInfo> infos = application.getInfoList(LyChartInfo.class);
		int					    i, n = infos.getSize();
		LyChartInfo			    info;
		Class<?>			    c = null;
		LyHRefs				    hrefs = new LyHRefs();
		LyHRef				    href;
		LyClassInfo			    classInfo = null;

		if (_parent instanceof LyGetClassInfoInterface)
			classInfo = ((LyGetClassInfoInterface)_parent).getClassInfo();

		for (i = n - 1; i >= 0; i--)
		{
			info = infos.getInfo(i);
			if (!info.getId().startsWith(prefix + '_')
					|| ((classInfo != null)
						&& (classInfo != ((LyGetClassInfoInterface)info).getClassInfo())))
				infos.removeInfo(info);
		}

		n = infos.getSize();

		// Add create menu
		if (prefix.equals("_chart"))
		{
			boolean addDeleteHRef = false;

			href = new LyHRef();
			href.setId(LyAreaController.DEMAND_CREATE_CHART);
			href.setLabel("LY_NEW_CHART");
			hrefs.addHRef(href);

			for (i = 0; i < n; i++)
			{
				info = infos.getInfo(i);

				// If chart has no file, it is not possible to delete this chart (chart may
				// be defined in the meta-model)
				if (info.getFile() != null)
				{
					addDeleteHRef = true;
					break;
				}
			}

			// Add delete menu if at least one chart that may be deleted
			if (addDeleteHRef)
			{
				href = new LyHRef();
				href.setId(LyAreaController.DEMAND_DELETE_CHART);
				href.setLabel("LY_DELETE_CHART");
				hrefs.addHRef(href);
			}
		}

		// Sort by name
		infos = infos.getInfosByName();

		for (i = 0; i < n; i++)
		{
			info = infos.getInfo(i);

			// Register current object info class to register
			if (c == null)
				c = info.getClass();

			href = new LyHRef();
			href.setId(info.getId());
			href.setLabel(((LyObjectInfo)info).getName());
			href.setImage(info.getImage());

			hrefs.addHRef(href);
		}

		if (hrefs.getHRefCount() == 0)
			return null;

		// Adjust hrefs height to its content with a limit of 15
		hrefs.setNRows(Math.min(15, hrefs.getHRefCount()));
		hrefs.setBorder(false);

		if (area == null)
			area = new LyArea();
		else
			area.removePanel(area.getPanel(0));

		LyForm		 form = new LyForm(prefix);
		LyFormField  formField = new LyFormField("_selectFile");
		LyConstraint constraint = new LyConstraint();

		formField.setFieldItem(hrefs);
		formField.setConstraint(constraint);
		form.addFormField(formField);
		form.setController(_parent.getId());
		form.setBorder(false);
		area.addPanel(form);

		constraint = new LyConstraint();
		constraint.setFill(LyConstraint.Fill.NONE);
		area.setConstraint(constraint);

		if (_actionAreas == null)
			_actionAreas = new Hashtable<String, LyArea>(1);

		_actionAreas.put(prefix, area);

		return area;
	} // end method updateSelectFile

	/**
	 * This method is invoked to enable the the area to complete the menuBar of the view and add its
	 * own menu items. Default implementation invokes the method on the builder instance.
	 *
	 * @param menuBar the menuBar structure
	 * @param level   : level of current Controller. one of the LEVEL_XXX contants in
	 *                LyViewController
	 * @see   LyViewController
	 */
	public void completeMenuBar(LyMenuBar menuBar, int level)
	{
		if (menuBar == null)
			return;

		LyAction action = ((LyAreaController)_parent).getAction();

		if (action.hasMark(LyAction.MARK_NO_MENUBAR))
			return;

		LyInfoList<LyAction> actions = action.getActions();

		if (actions == null)
			return;

		addMenuItems(menuBar, actions);
	}

	/**
	 * This method is invoked to enable the the area(s) to complete the message bar of the view and
	 * add its own message items. Default implementation invokes the method on the builder instance.
	 *
	 * @param messages the message bar structure
	 * @param level    : level of current Controller. one of the LEVEL_XXX contants. in
	 *                 LyViewController
	 * @see   LyViewController
	 */
	public void completeMessages(LyMessages messages, int level)
	{
	}

	/**
	 * Create help button if necessary. Method is shared by many implementation of LyAreaBuilder
	 *
	 * @return the Help command button
	 */
	protected LyCommand buildHelpCommand()
	{
		String helpFile = ((LyAreaController)_parent).getHelpFile();

		if ((helpFile == null) || (helpFile.length() == 0))
			return null;

		LyEnvironment environment = _parent.getEnvironment();
		LyCommand	  command;

		command = new LyCommand(LyController.DEMAND_HELP, "LY_HELP");
		command.setAction(LyCommand.Action.HELP);
		
		if (environment.checkImage("LY_COMMAND_HELP"))
			command.setImage("LY_COMMAND_HELP");

		return command;
	}

	/**
	 * Create help tool if necessary. Method is shared by many implementation of LyAreaBuilder
	 *
	 * @return the help tool
	 */
	protected LyTool buildHelpTool()
	{
		String helpFile = ((LyAreaController)_parent).getHelpFile();

		if ((helpFile == null) || (helpFile.length() == 0))
			return null;

		LyTool tool;

		tool = new LyTool(LyController.DEMAND_HELP, "LY_HELP");
		tool.setImage("LY_HELP");
		return tool;
	}

	/**
	 * Gets right label set by setTitleBarParameters.
	 *
	 * @return the right label
	 */
	public final String getRightLabel()
	{
		return _rightLabel;
	}

	/**
	 * Gets name of an object in the context of the given area controller.
	 *
	 * @param  object Object for which its name is required.
	 * @return Name of the object.
	 */
	protected String getName(LyObject object)
	{
		return ((LyAreaController)_parent).getName(object);
	}
} // end class LyAreaBuilder
