/**
 * LyCompoundAreaBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyCompoundAreaBuilder.java 35768 2012-06-07 08:56:24Z kchenal $
 */

package leon.control.builder;

import leon.control.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.mobile.*;
import leon.view.web.mobile.*;


/**
 * This builder construct the view for a coumpound Area Controller.
 *
 * @see LyCompoundAreaController
 */
public class LyCompoundAreaBuilder extends LyMultiAreaBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyCompoundAreaBuilder.java 35768 2012-06-07 08:56:24Z kchenal $";

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyCompoundAreaBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * This method is invoked to enable the the area(s) to complete the message bar of the view and
	 * add its own message items. Default implementation invokes the method on the builder instance.
	 *
	 * @param messages the message bar structure
	 * @param level    : level of current Controller. one of the LEVEL_XXX contants. in
	 *                 LyViewController
	 * @see   LyViewController
	 */
	public void completeMessages(LyMessages messages, int level)
	{
		LyCompoundAreaController compoundController = (LyCompoundAreaController)_parent;
		LyAction				 action = compoundController.getAction();

		if (action.hasMark(LyAction.MARK_NO_MESSAGES))
			return;

		// Add messages for the areas
		LyAreaController areaController;
		int				 i, n = compoundController.getAreaControllersCount();

		for (i = 0; i < n; i++)
		{
			areaController = compoundController.getAreaController(i);
			areaController.completeMessages(messages, level + 1);
		}
	}

	/**
	 * Builds the toolbar for this area.
	 *
	 * @return the toolbar component
	 */
	public LyToolBar buildToolBar()
	{
		LyCompoundAreaController compoundController = (LyCompoundAreaController)_parent;
		LyAction				 action = compoundController.getAction();

		if (action.hasMark(LyAction.MARK_NO_TOOLBAR))
			return null;

		int n = compoundController.getAreaControllersCount();

		if (n == 0)
			return null;

		// Add sub-toolbars for the areas
		LyAreaController areaController;
		LyAreaBuilder    areaBuilder;
		LyToolBar		 toolBar = null, subToolBar;
		int				 i;

		for (i = 0; i < n; i++)
		{
			areaController = compoundController.getAreaController(i);
			areaBuilder = (LyAreaBuilder)areaController.getBuilder();
			subToolBar = areaBuilder.buildToolBar();

			if (n == 1)
				return subToolBar;

			if (subToolBar == null)
				continue;

			if (toolBar == null)
				toolBar = new LyToolBar();

			toolBar.addToolBar(subToolBar);
		}

		if (toolBar != null)
		{
			toolBar.setStyle(LyToolBar.Style.SEPARATOR);
			toolBar.setColor("LY_DASHBOARD_HORIZONTAL_BACKGROUND");
		}

		return toolBar;
	} // end method buildToolBar

	/**
	 * Builds the view component for the specified parent controller. The component may create the
	 * XML Tree or the internal view.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{
		LyCompoundAreaController compoundController = (LyCompoundAreaController)_parent;
		int						 i, n = compoundController.getAreaControllersCount();

		// Add areas for the sub-areas
		LyCompoundAction		 compoundAction = (LyCompoundAction)compoundController.getAction();

		LyAreaController areaController;
		LyArea			 area = null, subArea, firstArea = null, panel;
		LySplitter		 splitter = null;
		boolean			 initialized = false;
		int[]			 sizes = compoundAction.getIntSizes();
		LyConstraint     constraint;

		area = new LyArea();
		_component = area;

		// FFT00468.
	    LyEnvironment environment = _parent.getEnvironment();
	    LyViewManager viewManager = (LyViewManager) environment.getViewManager();

		for (i = 0; i < n; i++)
		{
			areaController = compoundController.getAreaController(i);
			subArea = (LyArea)areaController.getComponent();
			if (subArea == null)
				// There was a problem when building the component
				continue;

			constraint = new LyConstraint();
			constraint.setFill(LyConstraint.Fill.BOTH);

			if (compoundAction.getStyle() == LyCompoundAction.Style.GRID)
			{
				panel = new LyArea();
				panel.setBorder(LyArea.Border.IN);
				panel.addPanel(subArea);

				constraint.setColSpan(sizes[3 * i]);
				constraint.setRowSpan(sizes[(3 * i) + 1]);
				constraint.setRemainder(((i == (n - 1)) || (sizes[(3 * i) + 2] == 1)) ? true
																					  : false);
				panel.setConstraint(constraint);
				area.addPanel(panel);

				LyConstraint	areaConstraint = new LyConstraint();

				areaConstraint.setFill(LyConstraint.Fill.BOTH);
				area.setConstraint(areaConstraint);
			}
			else
			{
				if (!initialized)
				{
					initialized = true;
					firstArea = subArea;
					if (i == (n - 1))
					{
						area.addPanel(subArea);
						break;
					}

					splitter = new LySplitter();

					if (compoundAction.getStyle() == LyCompoundAction.Style.HORIZONTAL)
						splitter.setStyle(LySplitter.Style.HORIZONTAL);
					else
						splitter.setStyle(LySplitter.Style.VERTICAL);

					splitter.setConstraint(constraint);
					splitter.setSizes(compoundAction.getSizes());
					// splitter may not be resizable
					splitter.setResizable(true);

					if(viewManager.hasProperty(LyMobileViewManager.MOBILE_ONLY_PROPERTY))
					{
						splitter.setActivePart(LySplitter.ActivePart.FIRST);
					}
					else
					{
						splitter.setActivePart(LySplitter.ActivePart.AUTO);
					}
					area.addPanel(splitter);
				}

				// Always add an intermediary level for web viewer to avoid graphical issues
				panel = new LyArea();
				panel.addPanel(subArea);
				panel.setConstraint(constraint);
				splitter.addPanel(panel);

				if (i == 0)
					splitter.setFirstTooltip("LY_ACTION_HIDE " + areaController.getName());
				else if (i == 1)
					splitter.setSecondTooltip("LY_ACTION_HIDE " + areaController.getName());
			}

			panel.setConstraint(constraint);
		}

		constraint = new LyConstraint();
		constraint.setFill(LyConstraint.Fill.BOTH);
		area.setConstraint(constraint);

		LyAction action = ((LyAreaController)_parent).getAction();
		
		if(action != null && action.hasMark(LyAction.MARK_MOBILEWEBVIEW))
		{
			area.addData(LyMobileConstants.WEBVIEW_MOBILE_KEY, LyMobileConstants.TRUE);
		}
		else
		{
			LyAreaController parentController = ((LyAreaController)_parent).getParentAreaController();
			boolean mobileAction = false;
			
			while(parentController != null && ! mobileAction)
			{
				LyAction parentAction = parentController.getAction();
				if(parentAction != null && parentAction.hasMark(LyAction.MARK_MOBILEWEBVIEW))
				{
					area.addData(LyMobileConstants.WEBVIEW_MOBILE_KEY, LyMobileConstants.TRUE);
					mobileAction = true;
				}
				parentController = parentController.getParentAreaController();
			}
		}

	    if ((action == null)
	        || !action.hasMark(LyAction.MARK_NO_TOOLBAR))
		{
			LyToolBar toolbar = ((firstArea == null) ? null : firstArea.getToolBar());

			completeToolBar(toolbar);
		}

		LyTitleBar titleBar = buildTitleBar();

		area.setTitleBar(titleBar);
		updateTitleBar();

		area.setController(_parent.getId());
		area.setId(_parent.getId());
		
		return area;
	} // end method build
} // end class LyCompoundAreaBuilder
