/**
 * LyDashboardBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyDashboardBuilder.java 36510 2012-08-17 15:43:06Z jlsaouli $
 */

package leon.control.builder;

import java.util.*;

import leon.app.*;
import leon.app.behavior.LyDashboardBehavior;
import leon.control.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.mobile.*;
import leon.view.web.*;


/**
 * This builder is used to construct the view for the dashboard controller.
 *
 * @see LyDashboardController
 */
public class LyDashboardBuilder extends LyMultiAreaBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyDashboardBuilder.java 36510 2012-08-17 15:43:06Z jlsaouli $";

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyDashboardBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Builds the view component for the specified parent controller. The component may create the
	 * XML Tree or the internal view.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{
		/* The dashboard view is an area which contains :
		        - A titleBar (description)
		        - A toolbar (actions)
		*/
		LyArea  		area = new LyArea();		
		boolean  		useExpandBar = ((LyDashboardController)_parent).testParameter("_useExpandBar");
		boolean  		showAsTree = ((LyDashboardController)_parent).testParameter("_showAsTree");
		LyViewManager	viewManager = getViewManager();
		LyConstraint	constraint = new LyConstraint();
		
		constraint.setFill(LyConstraint.Fill.BOTH);
		area.setConstraint(constraint);
		area.setId(_parent.getId());
		area.setController(_parent.getId());

		// set content
		area.setTitleBar(buildTitleBar());

		if (showAsTree)
			area.addPanel(buildTree());
		else
		{
			LyToolBar     toolBar = buildToolBar();

			toolBar.setHAlign(LyToolBar.HAlign.LEFT);

			if (useExpandBar
					&& (!viewManager.hasProperty("WEB_MANAGER")
							|| !getViewManager().hasProperty("MANAGE_DASHBOARD")))
				toolBar.setVAlign(LyToolBar.VAlign.MIDDLE);
			else
				toolBar.setVAlign(LyToolBar.VAlign.TOP);

			area.setToolBar(toolBar);
		}

		String redirect = (String)((LyDashboardController)_parent).getParameter("_redirect", true);

		area.setRedirect(redirect);
		_component = area;

		// Update content
		updateTitleBar();

		// If dashboard uses a single window, create a splitter to place the dashboard toolbar
		// and the actions views.
		if (((LyDashboardController)_parent).useSingleView()
				 && !viewManager.hasProperty("MANAGE_DASHBOARD"))
		{
			LyArea secondArea = buildDefaultArea();

            if (showAsTree)
            {
                LyArea          topArea = new LyArea();
                LySplitter      splitter = new LySplitter();
                
                splitter.setStyle(LySplitter.Style.HORIZONTAL);
                splitter.setSizes("-1 70");
                splitter.setResizable(true);
                splitter.addPanel(area);
                splitter.addPanel(secondArea);
                topArea.addPanel(splitter);
                
                _component = topArea;
            }
            else
            {
        		// Add a tool to close the views
            	if (addCloseViewsTool())
            	{
        			LyToolBar    rightToolbar = buildCloseToolBar();
        			
        			// If expand bar mode, add the close tool at the top of the left toolbar
                	if (!useExpandBar)
            			area.addToolBar(rightToolbar);
                	else if (useExpandBar && (!viewManager.hasProperty("WEB_MANAGER")
    						|| !getViewManager().hasProperty("MANAGE_DASHBOARD")))
                		area.getLeftToolBar().insertToolBar(rightToolbar, 0);
                	else
                		area.getTopLeftToolBar().insertToolBar(rightToolbar, 0);
            	}
            	
                area.addPanel(secondArea);
                _component = area;
            }

            if (secondArea != null)
            	secondArea.setConstraint(constraint);

			// Remove titlebar when using single window
			area.setTitleBar(null);
		}

		return _component;
	} // end method build

	protected LyToolBar buildCloseToolBar()
	{
		LyToolBar    rightToolbar = new LyToolBar();
		
		rightToolbar.setFill(LyToolBar.Fill.HORIZONTAL);
		rightToolbar.setStyle(LyToolBar.Style.SEPARATOR);
		rightToolbar.setHAlign(LyToolBar.HAlign.RIGHT);
		rightToolbar.setVAlign(LyToolBar.VAlign.TOP);
		
		boolean	  useTabs = ((LyDashboardController)_parent).useTabs();
		LyTool	  tool = new LyTool(useTabs ? LyDashboardController.DEMAND_CLOSE_ALL : LyDashboardController.DEMAND_CLOSE, "");
		
		tool.setImage("LY_TOOL_CLOSE_ALL_VIEWS");
		tool.setTooltip(useTabs ? "LY_CLOSE_ALL_VIEWS" : "LY_CLOSE_VIEW");
		tool.setEnabled(false);
		rightToolbar.addTool(tool);
		return rightToolbar;
	}
	
	/**
	 * Builds the area displayed when opening the dashboard (single window mode only).<br />
	 * Default implementation returns a map containing the application image and a welcome message.
	 *
	 * @return the default area or null if not single window mode.
	 */
	public LyArea buildDefaultArea()
	{
		LyArea		  area = new LyArea();
		LyViewManager viewManager = getViewManager();
		LyConstraint  constraint = new LyConstraint();
		
		constraint.setFill(LyConstraint.Fill.BOTH);
		area.setConstraint(constraint);

		if (viewManager.hasProperty("WEB_MANAGER"))
		{
			LyWebViewManager	webViewManager = (LyWebViewManager)viewManager;
			LyArea				subArea = new LyArea();
			
			subArea.setRedirect(webViewManager.getDefaultPageResource());
			area.addPanel(subArea);
			subArea.setConstraint(constraint);
		}
		else
		{
			LyMap	map = buildDefaultMap();

			map.setConstraint(constraint);
			area.addPanel(map);
		}

		return area;
	}
	
	protected LyMap buildDefaultMap()
	{
		LyMap		  map = new LyMap();
		LyMapNode     imageNode = buildDefaultMapNode();
		LyLayout	  layout = new LyLayout();

		map.setId("_dashboardMap");
		map.addMapNode(imageNode);
		map.setViewer(false);
		map.setLayout(layout);
		return map;
	}
	
	protected LyMapNode buildDefaultMapNode()
	{
		LyMapNode     imageNode = new LyMapNode();

		imageNode.setImage("LY_WELCOME_IMAGE");
		imageNode.setStyle(LyMapNode.Style.IMAGE);
		imageNode.setId("_dashboardMapImageNode");
		imageNode.setSensitive(false);
		imageNode.setMovable(false);
		return imageNode;
	}

	/**
	 * Builds the toolbar of the table view.
	 *
	 * @return the view Component
	 */
	protected LyToolBar buildToolBar()
	{
		LyInfoList<LyAction> actions = ((LyDashboardController)_parent).getActions(); 
		LyToolBar			 toolBar = completeToolBar(null, actions);

		if (toolBar == null)
		{
	        LyEnvironment environment = _parent.getEnvironment();
	        
			environment.logError("BF AppEngine : Cannot construct a DashBoard without Actions");
			toolBar = new LyToolBar();
		}

		String	useExpandBarParam = (String)((LyDashboardController)_parent).getParameter( "_useExpandBar", true);
		boolean	useExpandBar = "true".equals(useExpandBarParam);

		// Set default fill
		LyToolBar.Fill fill = (useExpandBar ? LyToolBar.Fill.VERTICAL : LyToolBar.Fill.DEFAULT);

		toolBar.setFill(fill);
		toolBar.setColor(null);
		setSubToolBarFill(toolBar, fill);

		// Toolbars are closed except with expand bar
		toolBar.setOpen(useExpandBar);

		// Set toolbars style
		if (useExpandBar)
		{
			toolBar.setStyle(LyToolBar.Style.EXPAND);
			setSubToolBarOpen(toolBar, false);

			int		    i, n = toolBar.getComponentCount();
			LyComponent component;

			for (i = 0; i < n; i++)
			{
				component = toolBar.getComponent(i);

				if (component instanceof LyToolBar)
				{
					((LyToolBar)component).setStyle(LyToolBar.Style.EXPAND);
					((LyToolBar)component).setOpen(getSubToolbarsOpenState());
					
					// Close toolbars at level 2 or greater
					setSubToolBarOpen(((LyToolBar)component), false);
				}
			}

			if (writeToolLabel())
			{
				LyEnvironment environment = _parent.getEnvironment();
				LyEnvironment subEnvironment = new LyEnvironment(environment);
	
				subEnvironment.putResource("LY_WRITE_TOOL_LABEL", "true");
				subEnvironment.putResource("LY_TOOL_LABEL_RIGHT_POSITION", "true");
	
				toolBar.setEnvironment(subEnvironment);
			}
		}
		else
			toolBar.setStyle(LyToolBar.Style.DEFAULT);
		
		LyToolBar	searchToolBar = buildSearchToolBar();
		
		if (searchToolBar != null)
			toolBar.addToolBar(searchToolBar);
		
		LyTool	socialTool = buildSocialTool(toolBar);
		
		if (socialTool != null)
			toolBar.addTool(socialTool);
		
		return toolBar;
	} // end method buildToolBar

	/**
	 * Builds the sub-toolbar for search tool.
	 */
	protected LyToolBar buildSearchToolBar()
	{
		LyDashboardController	dashboardController = (LyDashboardController)_parent;
		
		if (!dashboardController.testParameter("_searchTool") || 
				getViewManager().hasProperty(LyMobileViewManager.MOBILE_ONLY_PROPERTY))
			return null;

		LyToolBar	subToolBar = new LyToolBar();
		
		subToolBar.setId("_searchTool");
		subToolBar.setLabel("LY_TOOL_SEARCH");
		subToolBar.setStyle(LyToolBar.Style.EXPAND);
		
		LyFormField	formField;
		LyRelation	relation;
		
		formField = new LyFormField(LyDashboardController.SEARCH_CLASS);
		formField.setNeedCheck(true);
		relation = new LyRelation();
		
		LyDashboardBehavior		behavior = (LyDashboardBehavior)dashboardController.getBehavior();
		LyInfoList<LyClassInfo>	classes = behavior.getSearchClasses(dashboardController);
		
		if (classes == null)
		{
			LyApplication	application = dashboardController.getApplication();
			
			classes = application.getClassInfos();
			classes = classes.getInfosByName();
		}
		
		if (classes == null)
			return null;
		
		int				i, n = classes.getSize();
		LyClassInfo		classInfo;
		LySession		session = _parent.getSession();
		LyLink			link;
		LyAction		action;
        LyClassInfo 	searchClassInfo = (LyClassInfo)dashboardController.getParameter("_searchClassInfo");
		
		link = new LyLink("LY_NO_CHOICE", "LY_TOOL_SEARCH_EMPTY_LINK");
		relation.addLink(link);
		
		for (i = 0; i < n; i++)
		{
			classInfo = classes.getInfo(i);
			if (classInfo.isInternal())
				continue;
			action = classInfo.getActionWithId(LyAction.ID_CONSULT);
			if (action == null)
				continue;
			if (!session.validateAction(action, dashboardController))
				continue;
			
			link = new LyLink(classInfo.getId(), classInfo.getName());
			relation.addLink(link);
			if (searchClassInfo == classInfo)
				link.setSelected(true);
		}
		
		//relation.setNCols(25);
		formField.setFieldItem(relation);
		subToolBar.addFormField(formField);
		
		LyText		text;
		LyToolBar	textToolBar;
		LyTool		tool;
		
		formField = new LyFormField(LyDashboardController.SEARCH_VALUE);
		formField.setNeedCheck(false);
		text = new LyText();
		text.setNCols(20);
		formField.setFieldItem(text);
		
		textToolBar = new LyToolBar();
		tool = new LyTool(LyDashboardController.DEMAND_SEARCH, null);
		tool.setImage("LY_TOOL_SEARCH");
		tool.setTooltip("LY_TOOL_SEARCH_TIP");
		textToolBar.setController(_parent.getId());
		textToolBar.addTool(tool);
		
		formField.setToolBar(textToolBar);
		
		LyEnvironment	parentEnvironment = _parent.getEnvironment();
		LyEnvironment	environment = new LyEnvironment(parentEnvironment);
		
		environment.putResource("LY_COMBO_MIN_WIDTH", "-1");
		subToolBar.setEnvironment(environment);

		subToolBar.addFormField(formField);
		
		return subToolBar;
	} // end method buildToolBar

	/**
	 * Add social tool to given toolbar.
	 */
	protected LyTool buildSocialTool(LyToolBar toolBar)
	{
		if (!((LyDashboardController)_parent).testParameter("_socialTool"))
			return null;
		
		LyAction	action = ((LyDashboardController)_parent).getSocialAction();
		
		return (action != null ? buildTool(action, true) : null);
	}
	
	/**
	 * Indicates whether sub-toolbars are opened (expand mode only).
	 *
	 * @return true if sub-toolbars must be opened, false otherwise.
	 */
	protected boolean getSubToolbarsOpenState()
	{
		String  subToolbarsOpenState = (String)((LyDashboardController)_parent).getParameter(
				"_openExpandBars", false);
		
		if(subToolbarsOpenState == null 
				|| subToolbarsOpenState.equals("true"))
			return true;
		
		return false;
	}

	/**
	 * Indicates if a close tool must be added in expand bar to close all the views.
	 * 
	 * @return true if such a tool must be added, false otherwise
	 */
	protected boolean addCloseViewsTool()
	{
		return true;
	}
	
	/**
	 * Indicates if labels are written in expand mode.
	 * 
	 * @return true if labels are written, false otherwise
	 */
	protected boolean writeToolLabel()
	{
		return true;
	}

	/**
	 * Indicates if merging of toolbar and titlebar is allowed.
	 *
	 * @return true if merging of toolbar and titlebar is allowed, false otherwise.
	 */
	protected boolean allowMergeTitlebarToolbar()
	{
		LyViewManager viewManager = getViewManager();

		return (!viewManager.hasProperty("WEB_MANAGER"));
	}

	/**
	 * Builds the tree of the dashboard.
	 *
	 * @return the view Component
	 */
	protected LyTree buildTree()
	{
		LyTree				 tree = new LyTree(_parent.getId() + "_tree");
		LyInfoList<LyAction> actions = ((LyDashboardController)_parent).getActions();
		LyTreeNode			 subTreeNode1, subTreeNode2, treeNode = new LyTreeNode();
		LyAction			 action, groupAction;
		String				 group, group2;

		tree.setController(_parent.getId());
		tree.setSelectionType(LyTree.SelectionType.LEAVES_ONLY);
		tree.setSelectionMode(LyTree.SelectionMode.SINGLE);
		treeNode.setLabel(_parent.getApplication().getId());
		treeNode.setImage(_parent.getImage(((LyDashboardController)_parent).getAction()));

		if (actions == null)
			throw new IllegalArgumentException(
				"BF AppEngine : Cannot construct a DashBoard without Actions");

		if (!actions.isWritable())
			actions = new LyInfoList<LyAction>(actions);

		for (int i = 0; i < actions.getSize(); i++)
		{
			subTreeNode1 = new LyTreeNode();
			action = actions.getInfo(i);

			if (action.getGroup() != null)
			{
				group = action.getGroup();
				subTreeNode1.setLabel(group);
				subTreeNode1.setOpen(true);

				subTreeNode2 = new LyTreeNode();
				subTreeNode2.setId(action.getId());
				subTreeNode2.setLabel(action.getAlias());
				subTreeNode2.setImage(_parent.getImage(action));
				subTreeNode1.addTreeNode(subTreeNode2);

				for (int j = i + 1; j < actions.getSize(); j++)
				{
					subTreeNode2 = new LyTreeNode();
					groupAction = actions.getInfo(j);
					group2 = groupAction.getGroup();
					if ((group2 != null) && (group2.equals(group)))
					{
						subTreeNode2.setId(groupAction.getId());
						subTreeNode2.setLabel(groupAction.getAlias());
						subTreeNode2.setImage(_parent.getImage(groupAction));
						subTreeNode1.addTreeNode(subTreeNode2);
						actions.removeInfo(groupAction);
						j--;
					}
				}
			}
			else
			{
				subTreeNode1.setId(action.getId());
				subTreeNode1.setLabel(action.getAlias());
				subTreeNode1.setImage(_parent.getImage(action));
			}
			treeNode.addTreeNode(subTreeNode1);
			treeNode.setOpen(true);
		}

		tree.setTreeNode(treeNode);

		return tree;
	} // end method buildTree

	/**
	 * This method is invoked to enable the the area(s) to complete the message bar of the view and
	 * add its own message items. Default implementation invokes the method on the builder instance.
	 *
	 * @param messages the message bar structure
	 * @param level    : level of current Controller. one of the LEVEL_XXX contants. in
	 *                 LyViewController
	 * @see   LyViewController
	 */
	public void completeMessages(LyMessages messages, int level)
	{
		LyDashboardController dashboardController = (LyDashboardController)_parent;
		LyAction			  action = dashboardController.getAction();

		if (action.hasMark(LyAction.MARK_NO_MESSAGES))
			return;

		// Add messages for the areas
		LyAreaController areaController = dashboardController.getDisplayedAreaController();

		if (areaController != null)
			areaController.completeMessages(messages, level + 1);
	}
	
	/**
	 * Builds the window menu. <br />
	 * This menu is used to allow users seeing which dashboard actions are opened.
	 * 
	 * @return the window menu
	 */
	public LyMenu buildWindowMenu()
	{
		LyEnvironment environment = _parent.getEnvironment();
		String		  displayString = environment.translate("LY_MENU_WINDOW");

		if ((displayString == null) || (displayString.length() == 0))
			return null;

		// Add the window menu with all the opened windows
		LyMenu windowMenu = new LyMenu("LY_MENU_WINDOW");
		LyItem item;

		windowMenu.setId("_windowMenu");
		windowMenu.setController(_parent.getId());

		item = new LyItem(LyDashboardController.DEMAND_CLOSE_ALL, "LY_ITEM_CLOSE_ALL");

		item.setController(_parent.getId());
		item.setEnabled(false);
		windowMenu.addItem(item);

		LyDashboardController			    dashboardController = (LyDashboardController)_parent;
		Vector<LyController>			    subControllers = new Vector<LyController>();
		LyInfoList<LyAction>			    actions = dashboardController.getOpenedActions();
		Hashtable<String, LyViewController> viewControllers = dashboardController
				.getViewControllers();
		LyAction						    subAction;
		LyAreaController				    controller;

		if ((actions != null) && (actions.getSize() > 0))
		{
			int i, n = actions.getSize();

			for (i = 0; i < n; i++)
			{
				subAction = actions.getInfo(i);
				controller = dashboardController.getAreaController(subAction);
				if (controller != null)
					subControllers.addElement(controller);
			}
		}

		if ((viewControllers != null) && (viewControllers.size() > 0))
		{
			Enumeration<LyViewController> enumeration = viewControllers.elements();
			LyViewController			  viewController;

			while (enumeration.hasMoreElements())
			{
				viewController = enumeration.nextElement();
				subControllers.addElement(viewController);
			}
		}

		int i, n = subControllers.size();

		if (n > 0)
		{
			LyController     subController;
			LyAreaController subAreaController;
			LyAreaController displayedController = dashboardController.getDisplayedAreaController();
			boolean			 checked;

			item.setEnabled(true);
			windowMenu.addSep();

			for (i = 0; i < n; i++)
			{
				subController = subControllers.elementAt(i);

				if (subController instanceof LyViewController)
				{
					subAreaController = ((LyViewController)subController).getAreaController();
					if (subAreaController == null)
						continue;
					checked = ((LyViewController)subController).isShown();
				}
				else
				{
					subAreaController = (LyAreaController)subController;
					checked = (subAreaController == displayedController);
				}

				subAction = subAreaController.getAction();
				item = new LyItem("_menuItem_" + subAction.getId(), subAreaController.getTitle());
				item.setImage(dashboardController.getMenuItemImage(subAction));
				item.setDialog(false);
				item.setCheckBox(true);
				item.setChecked(checked);
				item.setController(_parent.getId());
				item.setEnabled(true);
				windowMenu.addItem(item);
			}
		}
		
		return windowMenu;
	}

	/**
	 * Complete the given menu bar.
	 *
	 * @param menuBar the menubar to complete
	 * @param level   the current level
	 */
	public void completeMenuBar(LyMenuBar menuBar, int level)
	{
		if (menuBar == null)
			return;

		LyAction action = ((LyAreaController)_parent).getAction();

		if (action.hasMark(LyAction.MARK_NO_MENUBAR))
			return;

		LyMenu windowMenu = buildWindowMenu();

        if (windowMenu != null)
            menuBar.addMenu(windowMenu);
        
		// Recurse in the client area
		super.completeMenuBar(menuBar, level);
	} // end method completeMenuBar

	/**
	 * Check if the action should be added in the toolbar Return true if not hidden, and if session
	 * validates the action (validateAction). DO NOT check tool mark in a dashboard.
	 *
	 * @param  action the action we want to test
	 * @return true if the action should be added, false otherwise
	 */
	protected boolean validateAction(LyAction action)
	{
		LySession session = _parent.getSession();

		if (action.hasMark(LyAction.MARK_HIDDEN))
			return false;

		// We Do *NOT* check the tool mark in a dashboard
		// Within a dashboard, action are explicitely added.

		if (!session.validateAction(action, _parent))
			return false;

		return true;
	}

	/**
	 * Change a toolbar fill mode.
	 *
	 * @param toolBar target toolbar
	 * @param fill    fill mode
	 */
	protected static void setSubToolBarFill(LyToolBar toolBar, LyToolBar.Fill fill)
	{
		Vector<LyToolBar> subToolBars = new Vector<LyToolBar>();
		LyToolBar		  subToolBar;
		int				  i, n;

		toolBar.getComponents(LyToolBar.class, subToolBars);
		n = subToolBars.size();

		for (i = 0; i < n; i++)
		{
			subToolBar = subToolBars.elementAt(i);
			if (subToolBar.getParent() instanceof LyToolBar)
				subToolBar.setFill(fill);
		}
	}

	/**
	 * Change a toolbar open mode.
	 *
	 * @param toolBar target toolbar
	 * @param open    open mode
	 */
	protected static void setSubToolBarOpen(LyToolBar toolBar, boolean open)
	{
		Vector<LyToolBar> subToolBars = new Vector<LyToolBar>();
		LyToolBar		  subToolBar;
		int				  i, n;

		toolBar.getComponents(LyToolBar.class, subToolBars);
		n = subToolBars.size();

		for (i = 0; i < n; i++)
		{
			subToolBar = subToolBars.elementAt(i);
			subToolBar.setOpen(open);
		}
	}

} // end class LyDashboardBuilder
