/**
 * LyEditTableBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyEditTableBuilder.java 35042 2012-03-26 07:25:39Z jlsaouli $
 */

package leon.control.builder;

import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;


/**
 * This builder creates an editable table view for the EditTable controller It is based on the table
 * builder.
 *
 * @see LyTableBuilder
 * @see LyEditTableController
 */
public class LyEditTableBuilder extends LyMultiPageTableBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyEditTableBuilder.java 35042 2012-03-26 07:25:39Z jlsaouli $";

	/** Default width for columns of editable tables. */
	public static int		DEFAULT_COLUMN_WIDTH = 17;
	
	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyEditTableBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Builds the view component for the specified parent controller. The component may create the
	 * XML Tree or the internal view.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{
		LyArea area = (LyArea)super.build();

		// set content
		area.setCommands(buildCommands());
		return area;
	}

	/**
	 * Complete the toolbar.
	 *
	 * @param  toolBar the toolbar to complete
	 * @return the completed toolbar
	 */
	protected LyToolBar completeToolBar(LyToolBar toolBar)
	{
		// Add tools for validating and canceling rows
		LyEditTableController tableController = (LyEditTableController)_parent;
		LyToolBar			  tools = super.completeToolBar(toolBar);

		if (!tableController.useEditTable())
			return tools;

		tools = completeEditionTools(tools);
		
		return tools;
	} // end method completeToolBar
	
	/**
	 * Complete the toolbar with edition tools.<br />
	 * Default implementation adds two tools to validate and cancel the rows.
	 * 
	 * @param  toolBar the toolbar to complete
	 * @return the completed toolbar
	 */
	protected LyToolBar completeEditionTools(LyToolBar toolBar)
	{
		if (toolBar == null)
		{
			toolBar = new LyToolBar();
			toolBar.setId("_toolBar");
			toolBar.setFill(LyToolBar.Fill.HORIZONTAL);
			toolBar.setColor("LY_DASHBOARD_HORIZONTAL_BACKGROUND");
		}
		else if (toolBar.getComponentCount() > 0)
			toolBar.addSep();

		String image;
		LyTool tool;

		tool = new LyTool(LyEditTableController.DEMAND_VALIDATE_ROW, "LY_VALIDATE");
		image = _parent.getImageFromPrefix("LY_VALIDATE_ROW");
		tool.setImage(image);
        tool.setTooltip("LY_VALIDATE_CHANGES");
		toolBar.addTool(tool);

		tool = new LyTool(LyEditTableController.DEMAND_CANCEL_ROW, "LY_CANCEL");
		image = _parent.getImageFromPrefix("LY_CANCEL_ROW");
		tool.setImage(image);
        tool.setTooltip("LY_CANCEL_CHANGES");
		toolBar.addTool(tool);

		return toolBar;
	}

	/**
	 * Complete the menubar with edition items.<br />
	 * Default implementation adds two items to copy and paste cells.
	 * 
	 * @param  menuBar the menuBar to complete
	 */
    protected void completeEditionMenu(LyMenuBar menuBar)
    {
    	super.completeEditionMenu(menuBar);
    	
		LyViewManager	  viewManager = getViewManager();

		if (!viewManager.hasProperty("WEB_MANAGER"))
		{
	        LyEnvironment environment = _parent.getEnvironment();
	        String        displayString = environment.translate("LY_MENU_EDITION");
	        
	        if ((displayString != null) && (displayString.trim().length() > 0))
	        {
	            // Search for Edition Menu
	            LyMenu editionMenu = (LyMenu)menuBar.getComponentByLabel("LY_MENU_EDITION");

	            // If it doesn't exist, create it
	            if (editionMenu == null)
	            {
	                editionMenu = new LyMenu("LY_MENU_EDITION");
	                menuBar.addMenu(editionMenu);
	            }
	            else
	                editionMenu.addSep();
	            
	            // Copy cells
	            String	id;
	            LyItem	item;
	            
	            displayString = environment.translate("LY_ITEM_COPY_CELLS");
	            
	            if ((displayString != null) && (displayString.trim().length() > 0))
	            {
	                id = LyEditTableController.DEMAND_COPY_CELLS;
	                item = new LyItem(id, "LY_ITEM_COPY_CELLS");
	                item.setController(_parent.getId());
	                item.setShortcut("LY_SHORTCUT_COPY_CELLS");
	                editionMenu.addItem(item);
	            }
	            
	            displayString = environment.translate("LY_ITEM_PASTE_CELLS");
	            
	            if ((displayString != null) && (displayString.trim().length() > 0))
	            {
	                id = LyEditTableController.DEMAND_PASTE_CELLS;
	                item = new LyItem(id, "LY_ITEM_PASTE_CELLS");
	                item.setController(_parent.getId());
	                item.setShortcut("LY_SHORTCUT_PASTE_CELLS");
	                editionMenu.addItem(item);
	            }
	        }
		}
    }

    /**
     * Builds the table header line description.
     *
     * @return the headers line
     */
    protected LyHeaders buildHeaders()
    {
        LyEditTableController tableController = (LyEditTableController)_parent;
        LyInfoList<LyFieldInfo>           fields = tableController.getFields();
        int                   i, n = ((fields == null) ? 0 : fields.getSize());
        LyFieldInfo           fieldInfo;
        LyHeaders             headers = new LyHeaders();
        LyHeader              header;
        LyChoiceFieldInfo     choiceInfo;
        LyInfoList<LyChoiceOption>            tableOptions;
        LyChoiceOption        option;
        int                   j, no;

        // SMA : empty headers in list display
        if (tableController.isDisplayList())
        {
            return super.buildHeaders();
        }

        // if there are fields with status in this classinfo, we add a header.
        if (tableController.isStatusColumn(null))
        {
            header = buildStatusHeader();

            if (header == null)
            {
                // Always add a status column
                header = new LyHeader();
                header.setLabel("LY_STATUS_SORT");
            }

            headers.addHeader(header);
        }

        LyEditTableBehaviorInterface behavior = (LyEditTableBehaviorInterface)_parent.getBehavior();
        LyHeaders lastSubHeaderForGroups = null;

        for (i = 0; i < n; i++)
        {
            fieldInfo = fields.getInfo(i);

            /** ONLY THE MULTIPLE CHOICE HAS BEEN MODIFIED **/
            if ((fieldInfo instanceof LyChoiceFieldInfo)
                    && ((LyChoiceFieldInfo)fieldInfo).isMultiple()
                    && tableController.canBeEditable(fieldInfo))
            {
                // Build a headers for given choice with a header for each option
                LyHeaders subHeaders = new LyHeaders();
                String    label = behavior.getHeaderLabel(tableController, fieldInfo);
                String    tooltip = behavior.getHeaderTooltip(tableController, fieldInfo);

                subHeaders.setId("headers" + LyController.ID_SEP + fieldInfo.getId());
                subHeaders.setLabel(label);
                subHeaders.setTooltip(tooltip);
                subHeaders.setAlign(LyHeader.Align.CENTER);
                headers.addHeaders(subHeaders);

                choiceInfo = (LyChoiceFieldInfo)fieldInfo;
                tableOptions = choiceInfo.getTableOptions();

                for (j = 0, no = tableOptions.getSize(); j < no; j++)
                {
                    option = tableOptions.getInfo(j);
                    header = buildHeader(option);
                    subHeaders.addHeader(header);
                }
            }
            else
            {
                if ((fieldInfo.getGroup() != null)
                        && ("true".equalsIgnoreCase((String) tableController
                                .getAction().getParameter(
                                        "_showGroupsInHeaders", true))))
                {
                    if ((lastSubHeaderForGroups != null)
                            && (!fieldInfo.getGroup().equals(
                                    lastSubHeaderForGroups.getLabel())))
                    {
                        lastSubHeaderForGroups = null;
                    }
                    if (lastSubHeaderForGroups == null)
                    {
                        lastSubHeaderForGroups = new LyHeaders();
                        String label = fieldInfo.getGroup();

                        lastSubHeaderForGroups.setId("headers"
                                + LyController.ID_SEP + fieldInfo.getId());
                        lastSubHeaderForGroups.setLabel(label);
                        lastSubHeaderForGroups.setAlign(LyHeader.Align.CENTER);
                        headers.addHeaders(lastSubHeaderForGroups);
                    }
                    header = buildHeader(fieldInfo);
                    lastSubHeaderForGroups.addHeader(header);
                }
                else
                {
                    lastSubHeaderForGroups = null;
                    header = buildHeader(fieldInfo);
                    headers.addHeader(header);
                }
            }
        }

        // Add a header for row images
        if (tableController.isEditable()
                && tableController.isEnabled()
                && tableController.useEditTable()
                && behavior.displayRowTools(tableController))
        {
            header = new LyHeader();
            header.setId("_imagesHeader");
            header.setLabel("");
            header.setSortable(false);
            header.setResizable(true);
            header.setAlign(LyHeader.Align.LEFT);

            // Set the column width     
            LyAction action = tableController.getAction();
            String   columnWithParameter = (String)action.getParameter("_columnWidth", true);

            if (columnWithParameter != null)
            {
                int pos = columnWithParameter.indexOf("_imagesHeader=");

                if (pos >= 0)
                {
                    int n1 = columnWithParameter.indexOf("##", pos);

                    if (n1 < 0)
                        n1 = columnWithParameter.length();

                    int    start = pos + "_imagesHeader=".length();
                    String valStr = columnWithParameter.substring(start, n1);

                    try
                    {
                        int val = Integer.parseInt(valStr);

                        header.setColumnWidth(val);
                    }
                    catch (NumberFormatException e)
                    {
                    }
                }
            }

            headers.addHeader(header);
        }

        headers.setReorderingAllowed(false);

        return headers;
    } // end method buildHeaders

	/**
	 * Build a header for a fieldInfo.
	 *
	 * @param fieldInfo Field info for which a header is built.
	 * @return the built header
	 */
	public LyHeader buildHeader(LyFieldInfo fieldInfo)
	{
		LyHeader			  header = super.buildHeader(fieldInfo);
		double 			      colWidth = ((header != null) ? header.getColumnWidth() : -1);
		LyEditTableController tableController = (LyEditTableController)_parent;
		LyViewManager		  viewManager = tableController.getViewManager();

		if (tableController.isEditable()
				&& tableController.isEnabled()
				&& (colWidth < 0)
				&& !viewManager.hasProperty("WEB_MANAGER")
				&& ((fieldInfo instanceof LyTextFieldInfo)
					|| ((fieldInfo instanceof LyChoiceFieldInfo)
						&& (((LyChoiceFieldInfo)fieldInfo).getChoiceOptionCount() > 2)
						&& !fieldInfo.isMultiple())
					|| (fieldInfo instanceof LyRelationFieldInfo)))
			header.setColumnWidth(DEFAULT_COLUMN_WIDTH);
		else if (colWidth > 0)
			tableController.setFixedColumnWidth(fieldInfo);
		
		return header;
	} // end method buildHeader

	/**
	 * Build header for a choice option.
	 * @param option the choice option
	 * @return the built header
	 */
	protected LyHeader buildHeader(LyChoiceOption option)
	{
		LyHeader header = new LyHeader();
		String   label;

		header.setId("header" + LyController.ID_SEP + option.getId());
		label = option.getAlias();
		if (label != null)
		{
			label = _parent.getEnvironment().translate(label);
			label = ((label.length() > 2) ? (label.substring(0, 2) + '.') : label);
		}
		header.setLabel(label);
		header.setTooltip(option.getName());
		header.setSortable(false);
		header.setAlign(LyHeader.Align.LEFT);

		return header;
	}

	/**
	 * Builds the table footer line description.
	 *
	 * @return the footer line
	 */
	protected LyFooters buildFooters()
	{
		LyEditTableController tableController = (LyEditTableController)_parent;
		if (tableController.isDisplayList())
		{
			return super.buildFooters();
		}
		LyInfoList<LyFieldInfo>			  fields = tableController.getFields();
		int					  i, n = ((fields == null) ? 0 : fields.getSize());
		LyFieldInfo			  fieldInfo;

		for (i = 0; i < n; i++)
		{
			fieldInfo = fields.getInfo(i);
			if ((fieldInfo instanceof LyNumberFieldInfo)
					&& (((LyNumberFieldInfo)fieldInfo).getFunction()
						!= LyNumberFieldInfo.Function.NONE))
				break;
		}

		if (i == n)
			return null;

		// There is almost one footer. We must add one for each column.
		LyFooters footers = new LyFooters();
		LyFooter  footer;

		// Add a footer for status column if needed
		if (tableController.isStatusColumn(null))
		{
			footer = new LyFooter();
			footer.setLabel("");
			footers.addFooter(footer);
		}

		LyChoiceFieldInfo choiceInfo;
		LyInfoList<LyChoiceOption>	tableOptions;
		int				  j, no;

		for (i = 0; i < n; i++)
		{
			fieldInfo = fields.getInfo(i);
			if ((fieldInfo instanceof LyChoiceFieldInfo)
					&& ((LyChoiceFieldInfo)fieldInfo).isMultiple()
					&& tableController.canBeEditable(fieldInfo))
			{
				choiceInfo = (LyChoiceFieldInfo)fieldInfo;
				tableOptions = choiceInfo.getTableOptions();

				for (j = 0, no = tableOptions.getSize(); j < no; j++)
				{
					footer = new LyFooter();
					footer.setLabel("");
					footers.addFooter(footer);
				}
			}
			else
			{
				footer = new LyFooter();
				footer.setId("footer" + LyController.ID_SEP + fieldInfo.getId());
				footer.setLabel("");
				footers.addFooter(footer);
			}
		}

		// Add a footer for row images
		if (tableController.isEditable()
				&& tableController.isEnabled()
				&& tableController.useEditTable())
		{
			footer = new LyFooter();
			footer.setId("_imagesFooter");
			footer.setLabel("");
			footers.addFooter(footer);
		}

		return footers;
	} // end method buildFooters

	/**
	 * Builds commands bar.
	 *
	 * @return the commands component
	 */
	protected LyCommands buildCommands()
	{
		LyAction action = ((LyEditTableController)_parent).getAction();
		boolean  autoCommit = ((LyEditTableController)_parent).isAutoCommit();

		if (action.hasMark(LyAction.MARK_NO_COMMANDS))
			return null;

		LyCommands commands = new LyCommands();
		LyCommand  command;

		if (!action.hasMark(LyAction.MARK_FRAME_VIEW))
		{
			command = new LyCommand(LyFormController.DEMAND_VALIDATE, "LY_VALIDATE");
			command.setAction(LyCommand.Action.SUBMIT);
			commands.addCommand(command);
		}

		LyEnvironment environment = _parent.getEnvironment();

		if (!autoCommit
				&& (environment.testResource("LY_APPLY_BUTTON")
					|| action.hasMark(LyAction.MARK_FRAME_VIEW)))
		{
			command = new LyCommand(LyFormController.DEMAND_APPLY, "LY_APPLY");
			command.setAction(LyCommand.Action.SUBMIT);
			commands.addCommand(command);
		}

		if (!autoCommit)
		{
			if (action.hasMark(LyAction.MARK_FRAME_VIEW))
				command = new LyCommand(LyEditTableController.DEMAND_RESTORE, "LY_RESTORE");
			else
				command = new LyCommand(LyFormController.DEMAND_CANCEL, "LY_CANCEL");
			
			command.setAction(LyCommand.Action.CANCEL);
			commands.addCommand(command);
		}

		if (commands.getCommandCount() == 0)
			return null;
		
		return commands;
	} // end method buildCommands
} // end class LyEditTableBuilder
