/**
 * LyExportBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyExportBuilder.java 40687 2012-09-06 15:07:47Z jlsaouli $
 */

package leon.control.builder;

import java.util.*;
import leon.control.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;


/**
 * This builder create the view for the export controller.
 *
 * @see LyExportController
 */
public class LyExportBuilder extends LyAreaBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyExportBuilder.java 40687 2012-09-06 15:07:47Z jlsaouli $";

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyExportBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Builds the view component for the specified parent controller. The component may create the
	 * XML Tree or the internal view.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{

		/* The export view is an area which contains :
		        - A toolbar (actions)
		        - A titleBar (description)
		        - A form component
		*/
		LyArea area = new LyArea();

		area.setScrolling(LyArea.Scrolling.AUTO);
		area.setTitleBar(buildTitleBar());
		area.addPanel(buildExportForm());
		area.setCommands(buildCommands());
		area.setController(_parent.getId());

		_component = area;

		return (area);
	}

	/**
	 * Build the titlebar.
	 *
	 * @return the titlebar component
	 */
	protected LyTitleBar buildTitleBar()
	{
		LyAction action = ((LyExportController)_parent).getAction();

		if (action.hasMark(LyAction.MARK_NO_TITLEBAR))
			return null;

		LyTitleBar  titleBar = new LyTitleBar();
		LyClassInfo classInfo = ((LyExportController)_parent).getClassInfo();
		String	    label = "LY_ACTION_EXPORT" + " : " + classInfo.getName();

		if (_parent.getEnvironment().testResource("LY_TITLEBAR_LEFT_IMAGE"))
			titleBar.setImage(_parent.getImage(classInfo));
		titleBar.setLeftLabel(label);

		return titleBar;
	}

	/**
	 * Build the attributes field
	 * @return the form field
	 */
	protected LyFormField buildFieldAttributes()
	{
		LyFormField   field;
		LyCaption     caption;
		LyTip		  tip;
		LyEnvironment environment = _parent.getEnvironment();
		LyClassInfo   classInfo = ((LyExportController)_parent).getClassInfo();
		LyConstraint  constraint;
		BitSet		  fieldMarks = ((LyExportController)_parent).getParameterMarks("_fieldMarks",
				LyBitSet.Type.F_MARKS);

		// **** Add attribute List
		LyRelation    fieldRelation = new LyRelation();
		int			  i, n = classInfo.getFieldInfoCount();
		LyFieldInfo   fieldInfo;
		LyLink		  link;

		// Get fieldinfo list
		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			if ((fieldMarks != null)
					&& !fieldInfo.hasMarks(fieldMarks, false))
				continue;

			if ((fieldMarks == null)
				&& fieldInfo.hasMark(LyFieldInfo.MARK_SECRET))
			{
				if ((fieldInfo instanceof LyTextFieldInfo)
					&& (((LyTextFieldInfo)fieldInfo).getEncoding() == null))
				{
					environment.logWarning("BF AppEngine : can't export text field "
						+ fieldInfo + " which is secret and has no encoding");
					continue;
				}
			}

			// Create relation to the fieldInfo
			link = new LyLink(fieldInfo.getId(), fieldInfo.getName());
			fieldRelation.addLink(link);
		}

		fieldRelation.setMultiple(true);

		field = new LyFormField(LyExportController.SELECT_FIELDINFOS);

		caption = new LyCaption();
		caption.setLabel("LY_EXPORT_FIELD");
		caption.setColon(true);

		constraint = new LyConstraint();
		constraint.setColSpan(LyConstraint.REMAINDER);
		constraint.setRemainder(true);
		constraint.setHAlign(LyConstraint.HAlign.LEFT);
		caption.setConstraint(constraint);

		field.setCaption(caption);
		field.setFieldItem(fieldRelation);
		field.setOptional(true);

		tip = new LyTip();
		tip.setLabel("LY_EXPORT_ATT_TIP");
		field.setTip(tip);

		constraint = new LyConstraint();
		constraint.setHAlign(LyConstraint.HAlign.CENTER);
		field.setConstraint(constraint);

		return field;
	} // end method buildFieldAttributes

    /**
     * Build the attribute prefix field
     * @return the form field
     */
    protected LyFormField buildFieldAttributePrefix()
    {
        LyFormField field;
        LyCaption caption;
        LyTip tip;
        LyText fieldText;

        fieldText = new LyText();
        field = new LyFormField(LyExportController.SELECT_FIELD_PREFIX);

        caption = new LyCaption();
        caption.setLabel("LY_SELECT_FIELD_PREFIX_TEXT");
        caption.setColon(true);

        fieldText.setNCols(30);
        field.setCaption(caption);
        field.setFieldItem(fieldText);
        field.setOptional(true);
        field.setNeedCheck(true);

        tip = new LyTip();
        tip.setLabel("LY_SELECT_FIELD_PREFIX_TIP");
        field.setTip(tip);

        LyConstraint constraint = new LyConstraint();
        constraint.setRemainder(false);
        constraint.setColSpan(1);
        constraint.setVAlign(LyConstraint.VAlign.TOP);
        
        field.setConstraint(constraint);
        return field;
    }
    
    /**
     * Build the attribute rename extension field
     * @return the form field
     */
    protected LyFormField buildFieldAttributeRenameExtension()
    {
        LyFormField field;
        LyCaption caption;
        LyTip tip;
        LyText fieldText;

        fieldText = new LyText();
        field = new LyFormField(LyExportController.SELECT_FIELD_RENAME_EXTENSION);

        caption = new LyCaption();
        caption.setLabel("LY_SELECT_FIELD_RENAME_EXTENSION_TEXT");
        caption.setColon(true);
        fieldText.setNCols(3);
        field.setCaption(caption);
        field.setFieldItem(fieldText);
        field.setOptional(true);
        field.setNeedCheck(true);

        tip = new LyTip();
        tip.setLabel("LY_SELECT_FIELD_RENAME_EXTENSION_TIP");
        field.setTip(tip);

        LyConstraint constraint = new LyConstraint();

        constraint.setVAlign(LyConstraint.VAlign.TOP);
		constraint.setColSpan(1);
        constraint.setRemainder(true);
        field.setConstraint(constraint);

        return field;
    }
	
	/**
	 * build the attribute separator field
	 * @param fieldSep the field separator value
	 * @param valueSep the value separator value
	 * @return the form field
	 */
	protected LyFormField buildFieldAttributeSep(String fieldSep,
		String valueSep)
	{
		LyFormField  field;
		LyCaption    caption;
		LyTip		 tip;
		LyConstraint constraint;
		LyText		 fieldText;

		fieldText = new LyText();
		fieldText.setLabel(LyExportController.SPREADSHEET_FIELD_SEP);
		fieldText.setNCols(1);
		fieldText.setNChars(1);
		field = new LyFormField(LyExportController.SELECT_FIELD_SEP);

		if (valueSep == null)
		{
			constraint = new LyConstraint();

			constraint.setColSpan(1);
			constraint.setRemainder(false);
			constraint.setFill(LyConstraint.Fill.NONE);

			field.setConstraint(constraint);
		}

		caption = new LyCaption();
		caption.setLabel("LY_FIELD_SEP");
		caption.setColon(true);

		field.setCaption(caption);
		field.setFieldItem(fieldText);
		field.setOptional(false);
		field.setNeedCheck(true);

		tip = new LyTip();
		tip.setLabel("LY_EXPORT_FIELD_SEP_TIP");
		field.setTip(tip);

		return field;
	} // end method buildFieldAttributeSep

	/**
	 * Build the attribute value separator field
	 * @param fieldSep the field separator value
	 * @param valueSep the value separator value
	 * @return the form field
	 */
	protected LyFormField buildFieldAttributeValSep(String fieldSep,
		String valueSep)
	{
		LyFormField field;
		LyCaption   caption;
		LyTip	    tip;
		LyText	    fieldText;

		fieldText = new LyText();
		field = new LyFormField(LyExportController.SELECT_VALUE_SEP);

		fieldText.setLabel(LyExportController.SPREADSHEET_VALUE_SEP);
		fieldText.setNCols(1);
		fieldText.setNChars(1);
		caption = new LyCaption();
		caption.setLabel("LY_VALUE_SEP");
		caption.setColon(true);

		field.setCaption(caption);
		field.setFieldItem(fieldText);
		field.setOptional(false);
		field.setNeedCheck(true);

		tip = new LyTip();
		tip.setLabel("LY_EXPORT_VALUE_SEP_TIP");
		field.setTip(tip);

		return field;
	} // end method buildFieldAttributeValSep

	/**
	 * Build the export type field
	 * @param exportType the export type
	 * @return the form field
	 */
	protected LyFormField buildFieldExportType(String exportType)
	{
		LyFormField  field = new LyFormField(LyExportController.SELECT_EXPORT_TYPE);
		LyCaption    caption;
		LyTip		 tip;
		LyConstraint constraint;

		LyChoice     fieldChoice = new LyChoice();
		LyOption     fieldOption;

		caption = new LyCaption();
		caption.setLabel("LY_EXPORT_TYPE");
		caption.setColon(true);

		fieldChoice.setMultiple(false);
		fieldChoice.setFill(LyChoice.Fill.VERTICAL);

		// FFT00020 : adding spreadsheet export mode
		fieldOption = new LyOption(LyExportController.SELECT_EXPORT_TYPE_SPREADSHEET,
				"LY_SELECT_EXPORT_TYPE_SPREADSHEET");
		fieldOption.setChecked(true);
		fieldChoice.addOption(fieldOption);

		fieldOption = new LyOption(LyExportController.SELECT_EXPORT_TYPE_TEXT,
				"LY_SELECT_EXPORT_TYPE_TEXT");
		fieldChoice.addOption(fieldOption);

		fieldOption = new LyOption(LyExportController.SELECT_EXPORT_TYPE_RAW,
				"LY_SELECT_EXPORT_TYPE_RAW");
		fieldChoice.addOption(fieldOption);

		fieldOption = new LyOption(LyExportController.SELECT_EXPORT_TYPE_XML,
				"LY_SELECT_EXPORT_TYPE_XML");
		fieldChoice.addOption(fieldOption);

		field.setCaption(caption);
		field.setFieldItem(fieldChoice);
		field.setOptional(false);
		field.setNeedCheck(true);

		tip = new LyTip();
		tip.setLabel("LY_EXPORT_TYPE_TIP");
		field.setTip(tip);

		constraint = new LyConstraint();

		//constraint.setRowSpan(2);
		field.setConstraint(constraint);
		return field;
	} // end method buildFieldExportType

	/**
	 * get the attribute separator value
	 * @param controller the controller
	 * @return the attribute separator value
	 */
	protected String getAttributeSep(LyExportController controller)
	{
		return (String)controller.getParameterValue(LyExportController.SELECT_FIELD_SEP);
	}

	/**
	 * Get the attribute value separator
	 * @param controller the controller
	 * @return the attribute value separator value
	 */
	protected String getAttributeValSep(LyExportController controller)
	{
		return (String)controller.getParameterValue(LyExportController.SELECT_VALUE_SEP);
	}

	/**
	 * get the export type
	 * @param controller the parent controller
	 * @return the export type
	 */
	protected String getExportType(LyExportController controller)
	{
		return (String)controller.getParameterValue(LyExportController.SELECT_EXPORT_TYPE);
	}

	/**
	 * get the export file
	 * @param controller the parent controller
	 * @return the export file
	 */
	protected String getFile(LyExportController controller)
	{
		return (String)controller.getParameterValue(LyExportController.SELECT_FILE);
	}
	
	/**
	 * get the attribute field prefix
	 * @param controller the parent controller
	 * @return the prefix
	 */
	protected String getAttributeFieldPrefix(LyExportController  controller)
	{
		return (String)controller.getParameterValue(LyExportController.SELECT_FIELD_PREFIX);
	}

	/**
	 * Get the rename extension
	 * @param controller the parent controller
	 * @return the rename extension
	 */
	protected String getAttributeFieldRenameExtension(LyExportController  controller)
	{
		return (String)controller.getParameterValue(LyExportController.SELECT_FIELD_RENAME_EXTENSION);
	}
	
	/**
	 * get the attributes
	 * @param controller the parent controller
	 * @return a list of attributes
	 */
	protected LyInfoList<LyFieldInfo> getAttributes(LyExportController controller)
	{
		@SuppressWarnings("unchecked")
		LyInfoList<LyFieldInfo> temp = (LyInfoList<LyFieldInfo>)controller.getParameterValue(
				LyExportController.SELECT_FIELDINFOS);
		return temp;
	}
	
	/**
	 * get the export selection value
	 * @param controller the parent controller
	 * @return the export selection value
	 */
	protected Boolean getExportSelection(LyExportController controller)
	{
		return (Boolean)controller.getParameterValue(LyExportController.EXPORT_SELECTION);
	}

	/**
	 * Build the bottom form
	 * @return the bottom form
	 */
	protected LyForm getBottomForm()
	{
		LyForm			   bottomForm = new LyForm(_parent.getId());
		LyFormField		   field;
		LyViewManager	   viewManager = getViewManager();

		// Add attribute List
		LyExportController controller = (LyExportController)_parent;

		// add the field separator text to the field
		String			   fieldSep = getAttributeSep(controller);
		String			   valueSep = getAttributeValSep(controller);

		if (fieldSep == null)
		{
			field = buildFieldAttributeSep(fieldSep, valueSep);
			if (field != null)
				bottomForm.addFormField(field);
		}

		bottomForm.setBorder(false);

		// add the value separator text to the field
		if (valueSep == null)
		{
			field = buildFieldAttributeValSep(fieldSep, valueSep);
			if (field != null)
				bottomForm.addFormField(field);
		}

		// add the export type choice to the field
		String exportType = getExportType(controller);

		if (exportType == null)
		{
			field = buildFieldExportType(exportType);
			if (field != null)
				bottomForm.addFormField(field);
		}

		if (viewManager.hasProperty("LOCAL_FILE"))
		{
			// add the file text to the field
			String file = getFile(controller);

			if (file == null)
			{
				field = buildFieldFile(file);
				if (field != null)
					bottomForm.addFormField(field);
			}
		}
		else
        {
        	String prefix = getAttributeFieldPrefix(controller);
        	
        	if (prefix == null)
        	{
	        	field = buildFieldAttributePrefix();
	            if (field != null)
	                bottomForm.addFormField(field);
        	}
            
        	String renameExtension = getAttributeFieldRenameExtension(controller);
        	
        	if (renameExtension == null)
        	{
	            field = buildFieldAttributeRenameExtension();
	            if (field != null)
	                bottomForm.addFormField(field);
        	}
        }
		
        Boolean     exportSelection = getExportSelection(controller);
        
        if (exportSelection == null)
        {
    		// We have to get the LyAreaController
    		LyController parentController = controller.getParentAreaController();
    		
    		// To export only selection list
    		if ((parentController instanceof LyAreaController)
    			&& (((LyAreaController)parentController).getContent() != null))
    		{
                LyFormField    booleanC = buildFieldSelectedValues();
                
                if (booleanC != null)
                {
        			LyTip    tip = new LyTip();
                    
        			tip.setLabel("LY_TIP_EXPORT_ONLY_SELECTED_VALUES");
        			booleanC.setTip(tip);
        			
        			// This checkbox is enabled only if the areaController have a selection 
        			if(((LyAreaController)parentController).getSelection() == null 
        					|| ((LyAreaController)parentController).getSelection().getSize() == 0)
        				booleanC.setEditable(false);
        			
        			bottomForm.addFormField(booleanC);
                }
    		}
        }
		
		return bottomForm;
	} // end method getBottomForm

	/**
	 * Build the bottom area
	 * @param bottomForm the bottom form
	 * @return the bottom area
	 */
	protected LyArea getBottomArea(LyForm bottomForm)
	{
		LyArea bottomArea = new LyArea();

		bottomArea.setBorder(LyArea.Border.IN);
		bottomArea.setLabel("LY_IMPORT_GROUP_PARAMETERS");
		bottomArea.addPanel(bottomForm);
		return bottomArea;
	}

	/**
	 * build the top form
	 * @return the top form
	 */
	protected LyForm getTopForm()
	{
		LyForm	    topForm = new LyForm(_parent.getId());
		LyFormField field = buildFieldAttributes();

		if (field != null)
			topForm.addFormField(field);
		topForm.setBorder(false);
		return topForm;
	}

	/**
	 * build the top area
	 * @return the top area
	 */
	protected LyArea getTopArea()
	{
		LyArea			   topArea = null;

		// Add attribute List
		LyExportController        controller = (LyExportController)_parent;
		LyInfoList<LyFieldInfo>	  fields = getAttributes(controller);

		if (fields == null)
		{
			LyForm topForm = getTopForm();

			topArea = new LyArea();
			topArea.setBorder(LyArea.Border.IN);
			topArea.setLabel("LY_IMPORT_GROUP_FIELDS");
			topArea.addPanel(topForm);
		}
		
		return topArea;
	}

	/**
	 * Build the file field
	 * @param file the file value
	 * @return the form field
	 */
	protected LyFormField buildFieldFile(String file)
	{
		LyFormField field = new LyFormField(LyExportController.SELECT_FILE);
		LyCaption   caption;
		LyTip	    tip;
		LyFile	    fieldFile = new LyFile();

		caption = new LyCaption();
		caption.setLabel("LY_FILE");
		caption.setColon(true);

		field.setCaption(caption);
		field.setFieldItem(fieldFile);
		field.setOptional(false);

		tip = new LyTip();
		tip.setLabel("LY_EXPORT_FILE_TIP");
		field.setTip(tip);

		return field;
	}

	/**
	 * Creates the export form.
	 *
	 * @return the area component
	 */
	protected LyArea buildExportForm()
	{
		LyForm bottomForm = getBottomForm();
		LyArea area = new LyArea();
		LyArea topArea = getTopArea();

		if (topArea != null)
			area.addPanel(topArea);

		if (bottomForm.getFormFieldCount() > 0)
		{
			LyArea bottomArea = getBottomArea(bottomForm);

			if (bottomArea != null)
				area.addPanel(bottomArea);
		}
		
		
		//** Export area is ready
		return (area);
	} // end method buildExportForm

	/**
	 * Creates the command bar and sets its content.
	 *
	 * @return the command bar component
	 */
	protected LyCommands buildCommands()
	{
		LyEnvironment environment = _parent.getEnvironment();
		LyCommands commands = new LyCommands();
		LyCommand  command;

		command = new LyCommand(LyFormController.DEMAND_VALIDATE, "LY_VALIDATE");
		command.setAction(LyCommand.Action.SUBMIT);
		
		if (environment.checkImage("LY_COMMAND_VALIDATE"))
			command.setImage("LY_COMMAND_VALIDATE");
		
		commands.addCommand(command);

		command = new LyCommand(LyExportController.DEMAND_PREVIEW, "LY_PREVIEW");
		command.setAction(LyCommand.Action.OTHER);
		
		if (environment.checkImage("LY_COMMAND_PREVIEW"))
			command.setImage("LY_COMMAND_PREVIEW");
		
		commands.addCommand(command);

		command = new LyCommand(LyFormController.DEMAND_CANCEL, "LY_CANCEL");
		command.setAction(LyCommand.Action.CANCEL);
		
		if (environment.checkImage("LY_COMMAND_CANCEL"))
			command.setImage("LY_COMMAND_CANCEL");
		
		commands.addCommand(command);

		command = buildHelpCommand();

		if (command != null)
			commands.addCommand(command);

		return commands;
	}
	
	/**
	 * Create a boolean field that is added to the current form to indicate if selection must
	 * be exported from the parent list.
	 *
	 * @return the created field item.
	 */
	protected LyFormField buildFieldSelectedValues()
	{
		// create the form field
		LyFormField field = new LyFormField(LyExportController.EXPORT_SELECTION);

		field.setOptional(false);

		// create the caption
		LyCaption captionComponent = new LyCaption();

		captionComponent.setLabel("LY_EXPORT_ONLY_SELECTED_VALUES");
		captionComponent.setColon(true);

		// create the boolean field
		LyBoolean booleanComponent = new LyBoolean();

		booleanComponent.setStatus(LyBoolean.Status.CHECKED);
		booleanComponent.setLabelChecked("");
		booleanComponent.setLabelUnchecked("");
		booleanComponent.setLabelEmpty(null);
		booleanComponent.setNot(true);

		// complete the field
		field.setCaption(captionComponent);
		field.setFieldItem(booleanComponent);
		field.setEnabled(true);

		return field;
	} // end method buildFieldSelectedValues
} // end class LyExportBuilder
