/**
 * LyFilterBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyFilterBuilder.java 41498 2012-11-12 13:20:36Z bangiole $
 */

package leon.control.builder;

import leon.control.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;


/**
 * Creates a view to type in a new simple filter expression.
 *
 * @see  LyFilterController
 */
public class LyFilterBuilder extends LySimpleFilterBuilder
{
    /** File version. */
    public static final String __VERSION = "$Id: LyFilterBuilder.java 41498 2012-11-12 13:20:36Z bangiole $";

    /**
     * Default constructor.
     *
     * @param  parent  parent controller
     */
    public LyFilterBuilder(LyController parent)
    {
        super(parent);
    }

    /**
     * Builds the view component for the specified parent controller. The component may create the
     * XML Tree or the internal view.
     *
     * @return  the view Component
     */
    public LyComponent build()
    {
        LyArea       area = new LyArea();
        LyForm       form = buildPanel();
        LyConstraint constraint = new LyConstraint();

        if (!_parent.getEnvironment().testResource("LY_EXTENDED_FILTER"))
        	constraint.setFill(LyConstraint.Fill.HORIZONTAL);
        else
        	constraint.setFill(LyConstraint.Fill.NONE);

        area.setConstraint(constraint);
        area.setScrolling(LyArea.Scrolling.AUTO);
        area.setController(_parent.getId());
        area.setTitleBar(buildTitleBar());
        area.addPanel(form);
        area.setCommands(buildCommands());
        area.setController(_parent.getId());
        _component = area;

        return area;
    }

    /**
     * Creates the form.
     *
     * @return  the form component
     */
    protected LyForm buildPanel()
    {
        // Get first expression
        LyFilter           filter = ((LyFilterController)_parent).getDefaultFilter();
        LyFilterExpression filterExp = getLastFilterExpression(filter);
        LyForm             form = new LyForm(_parent.getId());
        boolean            extended = false;
        LyConstraint       constraint = new LyConstraint();

        if (_parent.getParent() instanceof LyCompoundFilterController)
            extended = ((LyCompoundFilterController)_parent.getParent()).isExtendedFilter();
        else
            extended = ((filter != null)
                        && ((filter.getSize() > 1)
                            || !(filter.getElement(0) instanceof LyFilterExpression)));

        addFilterTypeField(form, extended);

        // Create a relation with fieldInfo list
        LyInfoList<LyFieldInfo> list = ((LyFilterController)_parent).getFields();

        // Sort by name depending on a resource
        list = list.getInfosByName(_parent.getEnvironment());

        int         i, n = list.getSize();
        LyFieldInfo fieldInfo;
        LyFieldInfo selectedField = null;
        LyLink      link;
        boolean     selected = false;
        LyRelation  fieldRelation = new LyRelation();

        for (i = 0; i < n; i++)
        {
            fieldInfo = list.getInfo(i);

            if ((selectedField == null)
                    && ((filterExp == null) || (filterExp.getFieldInfo() == fieldInfo)))
            {
                selectedField = fieldInfo;
                selected = true;
            }
            else
                selected = false;

            // Create relation to the fieldInfo
            link = new LyLink(fieldInfo.getId(), translate(fieldInfo.getAlias()));
            link.setSelected(selected);

            fieldRelation.addLink(link);
        }

        // Add the relation to the field
        LyFormField field;
        LyCaption   caption;

        field = new LyFormField(LyFilterController.SELECT_FIELDINFO);

        caption = new LyCaption();
        caption.setLabel("LY_FIELDINFO");
        caption.setColon(true);

        field.setCaption(caption);
        field.setFieldItem(fieldRelation);
        field.setOptional(false);
        field.setNeedCheck(true);

        form.addFormField(field);

        // Create list of all possible conditions
        LyRelation condRelation = buildConditionItem(selectedField, filterExp);

        field = new LyFormField(LyFilterController.SELECT_CONDITION);

        caption = new LyCaption();
        caption.setLabel("LY_CONDITION");
        caption.setColon(true);

        field.setCaption(caption);
        field.setFieldItem(condRelation);
        field.setOptional(false);
        field.setNeedCheck(true);

        form.addFormField(field);

        // Build Value field
        LyFieldItem item = buildFieldItem(selectedField);
        LyFieldUnit unit = buildUnitsList(selectedField);

        // add the relation
        LyToolBar toolBar = buildToolBar(selectedField);

        LyFormField valueField = new LyFormField(LyFilterController.SELECT_VALUE);

        caption = new LyCaption();
        caption.setLabel("LY_FILTER_VALUE");
        caption.setColon(true);

        valueField.setCaption(caption);
        valueField.setFieldItem(item);
        valueField.setOptional(true);
        valueField.setToolBar(toolBar);
        valueField.setFieldUnit(unit);

        form.addFormField(valueField);

        // special day value
        LyFormField dayField = new LyFormField(LyFilterController.SELECT_DATE_DAY_VALUE);

        caption = new LyCaption();
        caption.setLabel("LY_FILTER_VALUE");
        caption.setColon(true);

        LyDate dateItem = new LyDate();

        if (!LyDate.SHOW_ARROWS)
            dateItem.showArrows(false);

        dateItem.setType(LyDate.getType(LyDateFieldInfo.Type.DATE.ordinal()));
        dateItem.setLabel(null);
        dateItem.setFormat("dd/MM/yyyy");

        dayField.setCaption(caption);
        dayField.setFieldItem(dateItem);
        dayField.setOptional(true);
        dayField.setToolBar(toolBar);
        dayField.setFieldUnit(null);
        dayField.setHidden(true);

        form.addFormField(dayField);

        // add the modifiers choice
		LyEnvironment environment = _parent.getEnvironment();
        String		  displayString = environment.translate("LY_MODIFIERS");

		if ((displayString != null)
			&& (displayString.trim().length() > 0))
		{
	        // ****** Modifiers selection
	        LyChoice modifChoice = buildModifiersItem(selectedField, filterExp);

	        field = new LyFormField(LyFilterController.SELECT_MODIFIERS);

	        caption = new LyCaption();
	        caption.setLabel("LY_MODIFIERS");
	        caption.setColon(true);

	        field.setCaption(caption);
	        field.setFieldItem(modifChoice);
	        field.setOptional(false);

	        form.addFormField(field);
		}

		displayString = environment.translate("LY_DATE_SPECIAL_VALUE");
		if ((displayString != null)
				&& (displayString.trim().length() > 0))
		{
	        LyChoice choice = buildDateSpecialValuesItem(selectedField, filterExp, valueField, dayField);

	        field = new LyFormField(LyFilterController.SELECT_DATE_SPECIAL_VALUE);
	        caption = new LyCaption();
	        caption.setLabel("LY_DATE_SPECIAL_VALUE");
	        caption.setColon(true);

	        field.setCaption(caption);
	        field.setFieldItem(choice);
	        field.setOptional(true);
	        field.setNeedCheck(true);

	        if (!(selectedField instanceof LyDateFieldInfo))
	            field.setHidden(true);

	        form.addFormField(field);
		}
        // ** Filter form is ready
        form.setController(_parent.getId());

        constraint = new LyConstraint();
        constraint.setFill(LyConstraint.Fill.NONE);
        form.setConstraint(constraint);
        form.setAlign(LyForm.Align.LEFT);
        form.setBorder(false);

        return form;
    } // end method buildPanel

    /**
     * Build a field item for a text fieldInfo.
     *
     * @param   textField  the text fieldInfo corresponding to the built field item
     * @param   value      the value of the field item
     *
     * @return  the field item corresponding to the fieldInfo
     */
    public LyFieldItem buildText(LyTextFieldInfo textField, Object value)
    {
        LyText text = (LyText)super.buildText(textField, value);
        String format = textField.getFormat();

        // translate format
        LyEnvironment environment = _parent.getEnvironment();
        String        translatedFormat = environment.getString(format);

        if (translatedFormat != null)
            format = translatedFormat;

        if (format == null)
            return text;

        // Set format if the filtering condition is "equals"
        // We initialize the var with the good condition, because it is
        // the first selected when we open a new form.
        LyFilterExpression.Condition filterCondition = LyFilterExpression.Condition.EQUALS;

        if (_component != null)
        {
            LyFormField condField = null;
            String      condValue = null;

            condField = (LyFormField)_component.getComponent(LyFilterController.SELECT_CONDITION);
            condValue = condField.getFieldItem().getValue();

            if ((condValue != null) && (condValue.length() > 0))
                filterCondition = LyFilterExpression.Condition.valueOf(condValue);
        }

        if (filterCondition == LyFilterExpression.Condition.EQUALS)
            text.setFormat(format);

        return text;
    } // end method buildText

    /**
     * Build all conditions available for a given field.
     *
     * @param   fieldInfo  the field
     * @param   filterExp  filter expression
     *
     * @return  a relation component
     */
    protected LyRelation buildConditionItem(LyFieldInfo fieldInfo, LyFilterExpression filterExp)
    {
        LyRelation condRelation = new LyRelation();

        if (fieldInfo == null)
            return condRelation;

        LyLink                         link;
        LyFilterExpression.Condition[] conditions = fieldInfo.getFilterConditions();
        LyFilterExpression.Condition   s;
        int                            i, n = conditions.length;
        boolean                        selected = false;
        String                         name;
        boolean                        not = (filterExp != null) && filterExp.isInverted();

        // For all conditions enabled by the fieldInfo
        for (i = 0; i < n; i++)
        {
            s = conditions[i];

            selected = false; // TO DO : default

            if ((filterExp == null) && (i == 0))
                selected = true;

            if ((filterExp != null) && (filterExp.getCondition() == s))
                selected = true;

            // Create a link corresponding to condition
            name = LyFilterExpression.getCondition(s);
            link = new LyLink(s.toString(), name);
            link.setSelected(selected);
            condRelation.addLink(link);

            // Add not condition if not already present
            name = LyFilterExpression.getNotCondition(s);
            if ((name != null)
                    && name.contains("_NOT_"))
            {
                if (selected && not)
                    link.setSelected(false);

                link = new LyLink("_not." + s.toString(), name);
                if (selected && not)
                    link.setSelected(true);

                condRelation.addLink(link);
            }
        }

        return condRelation;
    } // end method buildConditionItem

    /**
     * Build the special values options for date filter.
     *
     * @param   fieldInfo   the field
     * @param   filterExp   filter expression
     * @param   valueField  value field
     * @param   dayField    day field
     *
     * @return  a choice component
     */
    protected LyChoice buildDateSpecialValuesItem(LyFieldInfo fieldInfo, LyFilterExpression filterExp,
        LyFormField valueField, LyFormField dayField)
    {
        LyChoice choice = new LyChoice();

        choice.setMultiple(false);

        LyOption optionNormal = new LyOption(LyFilterController.ID_DATE_NORMAL, "LY_FILTER_DATE_NORMAL");

        choice.addOption(optionNormal);

        LyOption optionToday = new LyOption(LyFilterController.ID_TODAY, "LY_FILTER_TODAY");

        choice.addOption(optionToday);

        LyOption optionNow = new LyOption(LyFilterController.ID_NOW, "LY_FILTER_NOW");

        choice.addOption(optionNow);

        LyOption optionOtherDay = new LyOption(LyFilterController.ID_DAY, "LY_FILTER_DAY");

        choice.addOption(optionOtherDay);

        if (filterExp == null)
            optionNormal.setChecked(true);
        else if (LyDateFieldInfo.VALUE_TODAY.equals(filterExp.getInitValue()))
        {
            optionToday.setChecked(true);
            valueField.setEnabled(false);
        }
        else if (LyDateFieldInfo.VALUE_NOW.equals(filterExp.getInitValue()))
        {
            optionNow.setChecked(true);
            valueField.setEnabled(false);
        }
        else if ((filterExp.getInitValue() != null)
                && filterExp.getInitValue().toString().startsWith(LyDateFieldInfo.VALUE_DAY))
        {
            optionOtherDay.setChecked(true);
            valueField.setHidden(true);
            dayField.setHidden(false);

            String stringDate = filterExp.getInitValue().toString().replace("_day_", "");

            stringDate = stringDate.replace('_', '/');

            dayField.getFieldItem().setLabel(stringDate + " 00:00:00");
        }
        else
        	optionNormal.setChecked(true);
        return choice;
    }

    /**
     * Build all modifiers available for a given field.
     *
     * @param   fieldInfo  the field
     * @param   filterExp  filter expression
     *
     * @return  a choice component
     */
    protected LyChoice buildModifiersItem(LyFieldInfo fieldInfo, LyFilterExpression filterExp)
    {
        LyChoice modifChoice = new LyChoice();

        if (fieldInfo == null)
            return modifChoice;

        LyOption option;
        short[]  modifiers = fieldInfo.getFilterModifiers();

        modifChoice.setMultiple(true);

        option = new LyOption(LyFilterController.ID_MODIFIER_NOT, "LY_FILTER_MODIF_NOT");

        if ((filterExp != null) && filterExp.isInverted())
        {
            String name = LyFilterExpression.getNotCondition(filterExp.getCondition());

            if ((name == null)
                    || !name.contains("_NOT_"))
                option.setChecked(true);
        }

        modifChoice.addOption(option);

        if (modifiers != null)
        {
            int    m, nm = modifiers.length;
            String modifierName;

            for (m = 0; m < nm; m++)
            {
                modifierName = LyFilterExpression.getModifier(modifiers[m]);
                option = new LyOption(modifierName, modifierName);

                if ((filterExp != null) && filterExp.hasModifier(modifiers[m]))
                    option.setChecked(true);

                modifChoice.addOption(option);
            }
        }

        return modifChoice;
    } // end method buildModifiersItem

    /**
     * Gets the last filter expression from filter.
     *
     * @param   filter  A filter.
     *
     * @return  The last found filter expression.
     */
    protected LyFilterExpression getLastFilterExpression(LyFilter filter)
    {
        if (filter == null)
            return null;

        LyFilterElement    element;
        LyFilterExpression exp;
        int                i, n = filter.getSize();

        for (i = n - 1; i >= 0; i--)
        {
            element = filter.getElement(i);

            if (element instanceof LyFilterExpression)
                return (LyFilterExpression)element;
            else if (element instanceof LyFilter)
            {
                exp = getLastFilterExpression((LyFilter)element);
                if (exp != null)
                    return exp;
            }
        }

        return null;
    }

    /**
     * Update the field that contains the possible filter values.
     *
     * @param  fieldId  The id of the new selected field.
     */
    public void resetValueField(String fieldId)
    {
        LyClassInfo classInfo = ((LyFilterController)_parent).getClassInfo();
        LyFieldInfo fieldInfo = classInfo.getFieldInfo(fieldId);
        LyFieldItem newItem = buildFieldItem(fieldInfo, null);
        LyFieldUnit unit = buildUnitsList(fieldInfo);
        LyToolBar   toolBar = buildToolBar(fieldInfo);

        // Retrieve currentField item container (formfield)
        LyFormField formField = (LyFormField)getComponent().getComponent(LyFilterController.SELECT_VALUE);

        // Replace the item in the XML Structure
        formField.setState(LyFormField.STATE_NORMAL);
        formField.setFieldItem(newItem);
        formField.setFieldUnit(unit);
        formField.setToolBar(toolBar);
        formField.resetContent();
    }

    /**
     * Enable or disable value field.
     *
     * @param  enable  if true, enable the value field, disable otherwise
     */
    public void enableValueField(boolean enable)
    {
        // Retrieve currentField item container (formfield)
        LyFormField formField = (LyFormField)getComponent().getComponent(LyFilterController.SELECT_VALUE);

        formField.setEnabled(enable);
    }

    /**
     * Update the field status : field is valid or not.
     *
     * @param  valid  the new valid status
     */
    public void setValueFieldStatus(boolean valid)
    {
        LyFormField formField = (LyFormField)getComponent().getComponent(LyFilterController.SELECT_VALUE);

        if (valid)
            formField.setState(LyFormField.STATE_NORMAL);
        else
            formField.setState(LyFormField.STATE_ERROR);
    }

    /**
     * Update the field that contains the filter conditions.
     *
     * @param  fieldId  The id of the new selected field.
     */
    public void resetConditionField(String fieldId)
    {
        LyClassInfo classInfo = ((LyFilterController)_parent).getClassInfo();
        LyFieldInfo fieldInfo = classInfo.getFieldInfo(fieldId);
        LyRelation  newItem = buildConditionItem(fieldInfo, null);

        // Retrieve currentField item container (formfield)
        LyFormField formField = (LyFormField)getComponent().getComponent(LyFilterController.SELECT_CONDITION);

        // Replace the item in the XML Structure
        formField.setFieldItem(newItem);
        formField.resetContent();
    }

    /**
     * Update the field that contains the filter modifiers.
     *
     * @param  fieldId  The id of the new selected field.
     */
    public void resetModifiersField(String fieldId)
    {
        // Retrieve currentField item container (formfield)
        LyFormField formField = (LyFormField)getComponent().getComponent(LyFilterController.SELECT_MODIFIERS);

        if (formField == null)
        	return;

        LyClassInfo classInfo = ((LyFilterController)_parent).getClassInfo();
        LyFieldInfo fieldInfo = classInfo.getFieldInfo(fieldId);
        LyChoice    newItem = buildModifiersItem(fieldInfo, null);

        // Replace the item in the XML Structure
        formField.setFieldItem(newItem);
        formField.resetContent();
    }

    /**
     * Update the field that contains the filter date special values.
     *
     * @param  fieldId  The id of the new selected field.
     */
    public void resetDateSpecialValues(String fieldId)
    {
        LyClassInfo classInfo = ((LyFilterController)_parent).getClassInfo();
        LyFieldInfo fieldInfo = classInfo.getFieldInfo(fieldId);

        // Retrieve currentField item container (formfield)
        LyFormField formField = (LyFormField)getComponent().getComponent(LyFilterController.SELECT_DATE_SPECIAL_VALUE);

        if (formField == null)
        	return;
        if (!(fieldInfo instanceof LyDateFieldInfo))
        {
            formField.setHidden(true);
            formField.setOptional(true);
            enableValueField(true);
            showDayValueField(false);
        }
        else
        {
            LyChoice choices = (LyChoice)formField.getFieldItem();
            int      i = 0;
            boolean  checked = false;
            LyOption option = null;
            boolean  date_time = ((((LyDateFieldInfo)fieldInfo).getType() == LyDateFieldInfo.Type.DATE_HOUR_MIN)
                        || (((LyDateFieldInfo)fieldInfo).getType() == LyDateFieldInfo.Type.DATE_TIME));

            option = choices.getOption(LyFilterController.ID_DAY);
            option.setEnabled(date_time);

            formField.setHidden(false);
            formField.setOptional(false);

            while (!checked && (i < choices.getOptionCount()))
            {
                option = choices.getOption(i);
                checked = option.isChecked();
                i++;
            }

            if (checked
                    && (LyFilterController.ID_TODAY.equals(option.getId())
                        || LyFilterController.ID_NOW.equals(option.getId())))
                enableValueField(false);
        }

        formField.resetContent();
    }

    /**
     * Reset form once all modifications were done.
     */
    public void resetForm()
    {

        // Force a refresh of the 'internal' structure
        LyForm form = (LyForm)_component.getComponent(_parent.getId());

        form.resetContent();
    }

    /**
     * Show the Day value field.
     *
     * @param  show  true to show
     */
    public void showDayValueField(boolean show)
    {
        // Retrieve currentField item container (formfield)
        LyFormField formField = (LyFormField)getComponent().getComponent(LyFilterController.SELECT_DATE_DAY_VALUE);

        formField.setHidden(!show);

        // Retrieve currentField item container (formfield)
        formField = (LyFormField)getComponent().getComponent(LyFilterController.SELECT_VALUE);
        formField.setHidden(show);
    }
} // end class LyFilterBuilder
