/**
 * LyFilterCriterionBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyFilterCriterionBuilder.java 28422 2010-11-03 12:03:38Z ehelmste $
 */

package leon.control.builder;

import leon.app.*;
import leon.control.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;


/**
 * Creates a view to choose a criterion filter.
 *
 * @see LyFilterCriterionController
 */
public class LyFilterCriterionBuilder extends LySimpleFilterBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyFilterCriterionBuilder.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Maximum length of displayed name for filters criterions. */
	public static final int    FILTER_NAME_MAX_LENGTH = 80;

	/** Relation containing the list of possible criterions. */
	protected LyRelation	   _criterionRelation = null;

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyFilterCriterionBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Builds the view component for the specified parent controller. The component may create the
	 * XML Tree or the internal view.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{
		LyArea area = new LyArea();

		area.setScrolling(LyArea.Scrolling.AUTO);
		area.setController(_parent.getId());
		area.setTitleBar(buildTitleBar());
		area.addPanel(buildPanel());
		area.setCommands(buildCommands());
		area.setController(_parent.getId());
		_component = area;

		return area;
	}

	/**
	 * Creates the form.
	 *
	 * @return the form component
	 */
	protected LyForm buildPanel()
	{        
		LyForm	    form = new LyForm(_parent.getId());
		LyFormField field;
		LyCaption   caption;

        addFilterTypeField(form, true);
        
		// Create a relation with criteria
		_criterionRelation = buildCriterionRelation();

		// Add the relation to the field
		field = new LyFormField(LyFilterCriterionController.SELECT_CRITERION);

		caption = new LyCaption();
		caption.setLabel("LY_FILTER_CRITERION");
		caption.setColon(true);

		field.setCaption(caption);
		field.setFieldItem(_criterionRelation);
		field.setOptional(false);
		field.setNeedCheck(true);

		form.addFormField(field);

		// Filter form is ready
		form.setController(_parent.getId());

		return form;
	} // end method buildPanel

	/**
	 * Builds the relation containing the list of filters.
	 *
	 * @return the relation containing the list of filters
	 */
	protected LyRelation buildCriterionRelation()
	{
		// Retrieve currentField item container (formfield)
		LyEnvironment     env = ((LyFilterCriterionController)_parent).getEnvironment();
		String			  name;
		LyRelation		  relation = new LyRelation();
		LyFilterCriterion criterion;
		LyInfoList<LyFilterCriterion>		  list = getCriterionList();
		int				  i, n;
		LyLink			  link;
		boolean			  selected = false, noSelection = true;

		n = list.getSize();

		for (i = n - 1; i >= 0; i--)
		{
			criterion = list.getInfo(i);

			selected = false;

			if (noSelection)
			{
				selected = true;
				noSelection = false;
			}

			name = LyFilterEditorController.getNameFilterCriterion(criterion, env);

			if (name.length() > FILTER_NAME_MAX_LENGTH)
				name = name.substring(0, FILTER_NAME_MAX_LENGTH) + "...";

			link = new LyLink(criterion.getId(), name);
			link.setSelected(selected);

			relation.addLink(link);
		}

		return relation;
	} // end method buildCriterionRelation

	/**
	 * Gets list of valid filter criterions to put in the relation.
	 * @return A list of filter criterion
	 */
	protected LyInfoList<LyFilterCriterion> getCriterionList()
	{
		LyFilterCriterion			  criterion;
		LyApplication				  application = _parent.getApplication();
		LyInfoList<LyFilterCriterion> list = application.getInfoList(LyFilterCriterion.class);
		int							  i, n;

		n = list.getSize();

		for (i = n - 1; i >= 0; i--)
		{
			criterion = list.getInfo(i);
			if (criterion.isInternal())
				list.removeInfo(criterion);
		}

		return list;
	}

	/**
	 * Build and fill the title bar.
	 *
	 * @return the titlebar component
	 */
	protected LyTitleBar buildTitleBar()
	{
		LyAction action = ((LyGetActionInterface)_parent).getAction();

		if (action.hasMark(LyAction.MARK_NO_TITLEBAR))
			return null;
        
        if ((((LyAreaController)_parent).getTarget() != null)
                || (((LyAreaController)_parent).getParent() instanceof LyCompoundFilterController))
            return null;

		LyTitleBar  titleBar = new LyTitleBar();
		LyClassInfo classInfo = ((LyGetClassInfoInterface)_parent).getClassInfo();
		String	    label = "LY_FILTER_CRITERION" + " : " + classInfo.getName();

		if (_parent.getEnvironment().testResource("LY_TITLEBAR_LEFT_IMAGE"))
			titleBar.setImage(_parent.getImage(classInfo));
		titleBar.setLeftLabel(label);

		return titleBar;
	}

	/**
	 * Creates the command bar and sets its content.
	 *
	 * @return the commands bar component
	 */
	protected LyCommands buildCommands()
	{
        LyAction action = ((LyAreaController)_parent).getAction();

        if (action.hasMark(LyAction.MARK_NO_COMMANDS))
            return null;
        
        if (((LyAreaController)_parent).getParent() instanceof LyCompoundFilterController)
            return null;
        
		LyCommands commands = new LyCommands();
		LyCommand  command;

		if (_parent.getParent() instanceof LyViewController)
		{
			command = new LyCommand(LyFormController.DEMAND_VALIDATE, "LY_VALIDATE");
			command.setAction(LyCommand.Action.SUBMIT);
			commands.addCommand(command);
		}

		LyEnvironment environment = _parent.getEnvironment();

		if (environment.testResource("LY_APPLY_BUTTON")
				|| !(_parent.getParent() instanceof LyViewController))
		{
			command = new LyCommand(LyFormController.DEMAND_APPLY, "LY_APPLY");
			command.setAction(LyCommand.Action.SUBMIT);
			commands.addCommand(command);
		}

		if (_parent.getParent() instanceof LyViewController)
		{
			command = new LyCommand(LyFormController.DEMAND_CANCEL, "LY_CANCEL");
			command.setAction(LyCommand.Action.CANCEL);
			commands.addCommand(command);
		}

		command = buildHelpCommand();

		if (command != null)
			commands.addCommand(command);

		return commands;
	} // end method buildCommands

	/**
	 * Resets form once all modifications were done.
	 */
	public void resetForm()
	{
		// Force a refresh of the 'internal' structure
		LyForm form = (LyForm)_component.getComponent(_parent.getId());

		form.resetContent();
	}

	/**
	 * Returns selected filter in the filter list.
	 *
	 * @return selected filter in the filter list
	 */
	public LyFilterCriterion getSelectedCriterion()
	{
		LyLink link;
		int    i, n = _criterionRelation.getLinkCount();

		for (i = 0; i < n; i++)
		{
			link = _criterionRelation.getLink(i);

			if (link.isSelected())
			{
				String			  criterionId = link.getId();
				LyFilterCriterion criterion = null;

				criterion = (LyFilterCriterion)_parent.getApplication().getInfo(criterionId);

				return criterion;
			}
		}

		return null;
	}
} // end class LyFilterCriterionBuilder
