/**
 * LyFilterEditorBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyFilterEditorBuilder.java 31005 2011-05-13 14:54:03Z nhamon $
 */

package leon.control.builder;

import leon.control.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;


/**
 * This builder builds a view for the extended filter controller. This builder is based on a tree
 * builder
 *
 * @see LyFilterEditorController
 */
public class LyFilterEditorBuilder extends LyAreaBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyFilterEditorBuilder.java 31005 2011-05-13 14:54:03Z nhamon $";

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyFilterEditorBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Builds the view component for the specified parent controller. The component may create the
	 * XML Tree or the internal view.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{

		/* The filter editor view is an area which contains :
		        - A titleBar (description)
		        - A toolbar (actions)
		        - A tree component
		*/
		LyArea			area = new LyArea();
		LyTree			tree = buildTree();
		LyConstraint 	constraint = new LyConstraint();
		
		constraint.setFill(LyConstraint.Fill.BOTH);
		area.setConstraint(constraint);
		tree.setConstraint(constraint);

		area.setId(_parent.getId());
		area.setController(_parent.getId());
		area.setScrolling(LyArea.Scrolling.AUTO);

		// set content
		area.setToolBar(buildToolBar());
		area.setTitleBar(buildTitleBar());
		area.addPanel(tree);
		area.setCommands(buildCommands());

		_component = area;

		// Update default content
		updateTitleBar();

		return area;
	} // end method build

	/**
	 * Builds the tree description.
	 *
	 * @return the tree component
	 */
	protected LyTree buildTree()
	{
		LyTree tree = new LyTree(_parent.getId() + "_tree");

		tree.setController(_parent.getId());
		tree.setIsContext(((LyAreaController)_parent).getContextClient() != null);
		return tree;
	}

	/**
	 * Build and fill the title bar.
	 *
	 * @return the titlebar component
	 */
	protected LyTitleBar buildTitleBar()
	{
		if (((LyAreaController)_parent).getTarget() != null)
			return null;

		return super.buildTitleBar();
	}

	/**
	 * Update the titlebar.
	 */
	public void updateTitleBar()
	{
		LyTitleBar titleBar = (LyTitleBar)_component.getComponent(_parent.getId() + "_titleBar"); // XXX

		if (titleBar != null)
		{
			LyClassInfo classInfo = ((LyFilterEditorController)_parent).getClassInfo();
			String	    label = "LY_EXTENDED_FILTER" + " : " + ((LyFilterEditorController)_parent).translate(classInfo.getName());

			if (_parent.getEnvironment().testResource("LY_TITLEBAR_LEFT_IMAGE"))
				titleBar.setImage(_parent.getImage(classInfo));
			titleBar.setLeftLabel(label);
		}
	}

	/**
	 * Creates the command bar and sets its content.
	 *
	 * @return the commands bar component
	 */
	protected LyCommands buildCommands()
	{
        LyAction action = ((LyAreaController)_parent).getAction();

        if (action.hasMark(LyAction.MARK_NO_COMMANDS))
            return null;
        
		LyCommands    commands = new LyCommands();
		LyCommand     command;
		LyViewManager viewManager = getViewManager();
		boolean		  haveValidateButton = false;
		boolean		  hasView = (_parent.getParent() instanceof LyViewController);
		boolean		  hasTarget = (((LyAreaController)_parent).getTarget() != null);

		if (hasView)
		{
			command = new LyCommand(LyFormController.DEMAND_VALIDATE, "LY_VALIDATE");
			command.setAction(LyCommand.Action.SUBMIT);
			commands.addCommand(command);
			haveValidateButton = true;
		}

		LyEnvironment environment = _parent.getEnvironment();

		if (((environment.testResource("LY_APPLY_BUTTON")
						&& (((LyFilterEditorController)_parent).getFilterEndInterface() != null))
					|| !hasView))
		{
			if ((haveValidateButton == false) || !viewManager.hasProperty("NOT_APPLY"))
			{
				command = new LyCommand(LyFormController.DEMAND_APPLY, "LY_APPLY");
				command.setAction(LyCommand.Action.SUBMIT);
				commands.addCommand(command);
			}
		}

		if (hasView || hasTarget)
		{
			command = new LyCommand(LyFormController.DEMAND_CANCEL, "LY_CANCEL");
			command.setAction(LyCommand.Action.CANCEL);
			commands.addCommand(command);
		}

		command = buildHelpCommand();

		if (command != null)
			commands.addCommand(command);

		return commands;
	} // end method buildCommands

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.control.builder.LyAreaBuilder#validateAction(leon.info.LyAction, boolean)
	 */
	protected boolean validateAction(LyAction action, boolean tool)
	{
		if (!super.validateAction(action, tool))
			return false;

		LyFilterEditorController editor = (LyFilterEditorController)getParent();

		return editor.validateAction(action);
	}

} // end class LyFilterEditorBuilder
