/**
 * LyFilterRelationBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyFilterRelationBuilder.java 36510 2012-08-17 15:43:06Z jlsaouli $
 */

package leon.control.builder;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;


/**
 * Creates a view to type in a new relation filter.
 *
 * @see LyFilterRelationController
 */
public class LyFilterRelationBuilder extends LySimpleFilterBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyFilterRelationBuilder.java 36510 2012-08-17 15:43:06Z jlsaouli $";

	/** Maximum length of displayed name for filters. */
	public static final int    FILTER_NAME_MAX_LENGTH = 80;

	/** Relation containing the list of possible field infos. */
	protected LyRelation	   _fieldRelation;

	/** Item containing the list of possible class infos. */
	protected LyFieldItem	   _classItem;

	/** Relation containing the list of possible filters. */
	protected LyRelation	   _filterRelation;

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyFilterRelationBuilder(LyController parent)
	{
		super(parent);
		_fieldRelation = null;
		_classItem = null;
		_filterRelation = null;
	}

	/**
	 * Builds the view component for the specified parent controller. The component may create the
	 * XML Tree or the internal view.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{
		LyArea area = new LyArea();

		area.setScrolling(LyArea.Scrolling.AUTO);
		area.setController(_parent.getId());
		area.setTitleBar(buildTitleBar());
		area.addPanel(buildPanel());
		area.setCommands(buildCommands());
		area.setController(_parent.getId());
		_component = area;

		updateTools();

		return area;
	}

	/**
	 * Creates the form.
	 *
	 * @return the form component
	 */
	protected LyForm buildPanel()
	{
		LyFilter			filter = ((LyFilterRelationController)_parent).getDefaultFilter();
		LyFilterRelation	initialRelation = ((filter == null) ? null : (LyFilterRelation)filter.getElement(0));
		LyForm				form = new LyForm(_parent.getId());
		LyFormField			field;
		LyCaption			caption;
        
        addFilterTypeField(form, true);
        
		// Create a relation with relation fieldInfo list
		LyInfoList<LyFieldInfo> list = ((LyFilterRelationController)_parent).getFields();

		// Sort by name depending on a resource
		if (list != null)
			list = list.getInfosByName(_parent.getEnvironment());

		int		    i, n = ((list == null) ? 0 : list.getSize());
		LyFieldInfo fieldInfo;
		LyLink	    link;
		boolean     selected = false, noSelection = true;

		_fieldRelation = new LyRelation();

		for (i = 0; i < n; i++)
		{
			fieldInfo = list.getInfo(i);

			// We keep only relation fieldInfos
			if (!(fieldInfo instanceof LyRelationFieldInfo)
					|| !enableRelation((LyRelationFieldInfo)fieldInfo))
				continue;

			selected = false;

			if (noSelection)
			{
				if ((initialRelation == null)
						|| (initialRelation.getRelationFieldInfo() == fieldInfo))
				{
					noSelection = false;
					selected = true;
				}
			}

			// Create relation to the fieldInfo
			link = new LyLink(fieldInfo.getId(), translate(fieldInfo.getName()));
			link.setSelected(selected);

			_fieldRelation.addLink(link);
		}

		// Add the relation to the field
		field = new LyFormField(LyFilterRelationController.SELECT_FIELDINFO);

		caption = new LyCaption();
		caption.setLabel("LY_FIELDINFO");
		caption.setColon(true);

		field.setCaption(caption);
		field.setFieldItem(_fieldRelation);
		field.setOptional(false);
		field.setNeedCheck(true);

		form.addFormField(field);

		// Create a choice with classes corresponding to selected fieldInfo
		buildClassItem(initialRelation);

		field = new LyFormField(LyFilterRelationController.SELECT_CLASS);
		caption = new LyCaption();
		caption.setLabel("LY_CLASS");
		caption.setColon(true);

		field.setCaption(caption);
		field.setFieldItem(_classItem);
		field.setOptional(false);
		field.setNeedCheck(true);

		form.addFormField(field);

		// Create a relation with filters corresponding to selected fieldInfo
		if (((LyFilterRelationController)_parent).isDisplayFilterRelation())
		{
			buildFilterRelation(initialRelation);

			// Add the relation to the field
			field = new LyFormField(LyFilterRelationController.SELECT_FILTER);
			field.setToolBar(buildToolBar());

			caption = new LyCaption();
			caption.setLabel("LY_FILTER");
			caption.setColon(true);

			field.setCaption(caption);
			field.setFieldItem(_filterRelation);
			field.setOptional(false);
			field.setNeedCheck(true);

			form.addFormField(field);
		}

		// Filter form is ready
		form.setController(_parent.getId());

		return form;
	} // end method buildPanel

	/**
	 * @see leon.control.builder.LySimpleFilterBuilder#buildToolBar()
	 */
	protected LyToolBar buildToolBar()
	{
		// Add a toolbar to the field to create filters (simple or extended)
		LyEnvironment environment = ((LyFilterRelationController)_parent).getEnvironment();
		boolean		  showLabel = environment.testResource("LY_WRITE_FIELD_TOOL_LABEL");
		LyToolBar     toolbar = new LyToolBar();
		LyTool		  tool;

		// Add a tool for the creation of a simple filter
		tool = new LyTool(LyFilterRelationController.CREATE_SIMPLE_FILTER,
				(showLabel ? "LY_ITEM_FILTER_NEW" : null));
		tool.setImage("LY_TOOL_FILTER_NEW");
		tool.setTooltip("LY_ITEM_FILTER_NEW");
		tool.setDialog(true);
		toolbar.setFill(LyToolBar.Fill.HORIZONTAL);
		toolbar.addTool(tool);
		toolbar.setController(_parent.getId());

		if (environment.testResource("LY_EXTENDED_FILTER"))
		{
			// Add a tool for the creation of an extended filter
			tool = new LyTool(LyFilterRelationController.CREATE_EXTENDED_FILTER,
					(showLabel ? "LY_ITEM_EXTENDED_FILTER_NEW" : null));
			tool.setTooltip("LY_ITEM_EXTENDED_FILTER_NEW");
			tool.setImage("LY_TOOL_EXTENDED_FILTER_NEW");
			tool.setDialog(true);
			toolbar.addTool(tool);
		}

		// Add a tool for the modification of a simple filter
		tool = new LyTool(LyFilterRelationController.SET_FILTER,
				(showLabel ? "LY_TOOL_FILTER_SET" : null));
		tool.setTooltip("LY_TOOL_FILTER_SET");
		tool.setImage("LY_TOOL_FILTER_SET");
		tool.setDialog(true);
		toolbar.setFill(LyToolBar.Fill.HORIZONTAL);
		toolbar.addTool(tool);
		toolbar.setController(_parent.getId());

		if (environment.testResource("LY_EXTENDED_FILTER"))
		{
			// Add a tool for the extension of a simple filter
			tool = new LyTool(LyFilterRelationController.EXTEND_SIMPLE_FILTER,
					(showLabel ? "LY_TOOL_EXTEND_SIMPLE_FILTER" : null));
			tool.setTooltip("LY_TOOL_EXTEND_SIMPLE_FILTER");
			tool.setImage("LY_TOOL_EXTEND_SIMPLE_FILTER");
			tool.setDialog(true);
			toolbar.addTool(tool);
		}

		return toolbar;
	} // end method buildToolBar

	/**
	 * Check if the given relation is enabled
	 * @param relation the relation
	 * @return true if the relation is enabled
	 */
	protected boolean enableRelation(LyRelationFieldInfo relation)
	{
		LyClassInfo classInfo;
		int		    i, n = relation.getClassInfoCount();

		// Check that we can filter on related classes
		for (i = 0; i < n; i++)
		{
			classInfo = relation.getClassInfo(i);
			if (enableClassInfo(classInfo))
				return true;
		}

		return false;
	}

	/**
	 * Check if the given classinfo is enabled
	 * @param classInfo a classInfo
	 * @return always true
	 */
	protected boolean enableClassInfo(LyClassInfo classInfo)
	{
		return true;
	}

	/**
	 * Builds the relation containing the list of class info.
	 *
	 * @param  initialFilter Filter used to set the view data.
	 * @return the relation containing the list of class info
	 */
	protected LyFieldItem buildClassItem(LyFilterRelation initialFilter)
	{
		LyRelationFieldInfo		   field = getSelectedFieldInfo();
		LyFilterRelationController parent = (LyFilterRelationController)_parent;
		LyFilterBehaviorInterface  behavior = (LyFilterBehaviorInterface)parent.getBehavior();
		LyInfoList<LyClassInfo>    classes = behavior.getFilterRelationClasses(parent, field);
		int						   i, n = ((classes == null) ? 0 : classes.getSize());
		LyClassInfo				   classInfo;
		boolean					   checked = false;
		LyChoice 				   choice = new LyChoice();
		LyOption				   option;

		for (i = 0; i < n; i++)
		{
			classInfo = classes.getInfo(i);
			if (!enableClassInfo(classInfo))
				continue;

			option = new LyOption(classInfo.getId(), translate(classInfo.getName()));
			checked = false;

			// If a filter is being set, retrieve a class that can be used
			// for current filter
			if (initialFilter != null)
			{
				if (initialFilter.getFilter().matchClass(classInfo))
				{
					checked = true;
					initialFilter = null;
				}
			}
			else
				checked = (i == 0);

			option.setChecked(checked);
			choice.addOption(option);
		}

		choice.setFill(LyChoice.Fill.VERTICAL);
		_classItem = choice;

		return _classItem;
	} // end method buildClassItem

	/**
	 * Builds the relation containing the list of filters.
	 *
	 * @param  initialFilter Filter used to set the view data.
	 * @return the relation containing the list of filters
	 */
	protected LyRelation buildFilterRelation(LyFilterRelation initialFilter)
	{
		// Retrieve currentField item container (formfield)
		LyApplication		 application = ((LyFilterRelationController)_parent).getApplication();
		LyInfoList<LyFilter> createdFilters = ((LyFilterRelationController)_parent)
				.getCreatedFilters();
		LyEnvironment		 env = ((LyFilterRelationController)_parent).getEnvironment();
		LyFilter			 appliFilter;
		String				 name;
		LyRelation			 relation = new LyRelation();
		LyInfoList<LyFilter> list = application.getInfoList(LyFilter.class);
		int					 i, n;
		LyLink				 link;
		boolean				 selected = false, noSelection = true;
		LyClassInfo			 classInfo = getSelectedClassInfo();

		list.addList(createdFilters);

		n = list.getSize();

		for (i = n - 1; i >= 0; i--)
		{
			appliFilter = list.getInfo(i);

			if (!appliFilter.matchClass(classInfo))
				continue;

			// If this filter is not a preference and was not created, ignore it
			if (!(appliFilter.getParent() instanceof LyPreference)
					&& ((createdFilters == null)
						|| (!createdFilters.contains(appliFilter))))
				continue;

			selected = false;

			if (initialFilter != null)
			{
				LyFilter filter = initialFilter.getFilter();

				if ((filter.getName() != null) && (filter.getName().equals(appliFilter.getName())))
				{
					selected = true;
					initialFilter = null;
					noSelection = false;
				}
			}
			else if (noSelection)
			{
				selected = true;
				noSelection = false;
			}

			if ((appliFilter.getName() == null)
					|| (appliFilter.getName().startsWith("_")))
			{
				name = LyFilterEditorController.getNameFilter(appliFilter, env, false);

				if (name.length() > FILTER_NAME_MAX_LENGTH)
					name = name.substring(0, FILTER_NAME_MAX_LENGTH) + "...";
			}
			else
				name = appliFilter.getName();

			link = new LyLink(appliFilter.getId(), name);
			link.setSelected(selected);

			relation.addLink(link);
		}

		// If nothing selected and an initial filter is set, it means that
		// filter was not found in the filter list. We must create it and add
		// it to the list.
		if (noSelection && (initialFilter != null))
		{
			((LyFilterRelationController)_parent).addFilter(initialFilter.getFilter());
			buildFilterRelation(null);
		}
		else
			_filterRelation = relation;

		return relation;
	} // end method buildFilterRelation

	/**
	 * Build and fill the title bar.
	 *
	 * @return the titlebar component
	 */
	protected LyTitleBar buildTitleBar()
	{
		LyAction action = ((LyGetActionInterface)_parent).getAction();

		if (action.hasMark(LyAction.MARK_NO_TITLEBAR))
			return null;
        
		if ((((LyAreaController)_parent).getTarget() != null)
                || (((LyAreaController)_parent).getParent() instanceof LyCompoundFilterController))
			return null;

		LyTitleBar  titleBar = new LyTitleBar();
		LyClassInfo classInfo = ((LyGetClassInfoInterface)_parent).getClassInfo();
		String	    label = "LY_FILTER_RELATION" + " : " + classInfo.getName();

		if (_parent.getEnvironment().testResource("LY_TITLEBAR_LEFT_IMAGE"))
			titleBar.setImage(_parent.getImage(classInfo));
		titleBar.setLeftLabel(label);

		return titleBar;
	}

	/**
	 * Creates the command bar and sets its content.
	 *
	 * @return the commands bar component
	 */
	protected LyCommands buildCommands()
	{
        LyAction action = ((LyAreaController)_parent).getAction();

        if (action.hasMark(LyAction.MARK_NO_COMMANDS))
            return null;
        
        if (((LyAreaController)_parent).getParent() instanceof LyCompoundFilterController)
            return null;
        
		LyCommands    commands = new LyCommands();
		LyCommand     command;
		LyViewManager viewManager = getViewManager();
		boolean		  haveValidateButton = false;

		if (_parent.getParent() instanceof LyViewController)
		{
			command = new LyCommand(LyFormController.DEMAND_VALIDATE, "LY_VALIDATE");
			command.setAction(LyCommand.Action.SUBMIT);
			commands.addCommand(command);
			haveValidateButton = true;
		}

		LyEnvironment environment = _parent.getEnvironment();

		if ((environment.testResource("LY_APPLY_BUTTON")
					|| !(_parent.getParent() instanceof LyViewController)))
		{
			if ((haveValidateButton == false) || !viewManager.hasProperty("NOT_APPLY"))
			{
				command = new LyCommand(LyFormController.DEMAND_APPLY, "LY_APPLY");
				command.setAction(LyCommand.Action.SUBMIT);
				commands.addCommand(command);
			}
		}

		if (_parent.getParent() instanceof LyViewController)
		{
			command = new LyCommand(LyFormController.DEMAND_CANCEL, "LY_CANCEL");
			command.setAction(LyCommand.Action.CANCEL);
			commands.addCommand(command);
		}

		command = buildHelpCommand();

		if (command != null)
			commands.addCommand(command);

		return commands;
	} // end method buildCommands

	/**
	 * Updates the field that contains the possible classes.
	 */
	public void resetClassField()
	{
		LyFormField formField = (LyFormField)_classItem.getParent();

		buildClassItem(null);
		formField.setFieldItem(_classItem);
		formField.resetContent();
	}

	/**
	 * Forces the field that contains the possible classes to the given class.
	 *
	 * @param classInfo the class info that must be set on the field.
	 */
	public void setClassInfo(LyClassInfo classInfo)
	{
		if (classInfo == null)
			return;

		if (_classItem instanceof LyChoice)
		{
			LyChoice classChoice = (LyChoice)_classItem;
			LyOption option;
			int		 i, n = classChoice.getOptionCount();
			String   classId = classInfo.getId();

			for (i = 0; i < n; i++)
			{
				option = classChoice.getOption(i);
				option.setChecked(classId.equals(option.getId()));
			}
		}
		else if (_classItem instanceof LyRelation)
		{
			LyRelation classRelation = (LyRelation)_classItem;
			LyLink     link;
			int		   i, n = classRelation.getLinkCount();
			String     classId = classInfo.getId();

			for (i = 0; i < n; i++)
			{
				link = classRelation.getLink(i);
				link.setSelected(classId.equals(link.getId()));
			}
		}
	} // end method setClassInfo

	/**
	 * Updates the field that contains the possible filters.
	 */
	public void resetFilterField()
	{
		if ((_filterRelation == null)
				|| !((LyFilterRelationController)_parent).isDisplayFilterRelation())
			return;

		// Retrieve currentField item container (formfield)
		LyFormField formField = (LyFormField)_filterRelation.getParent();

		buildFilterRelation(null);
		formField.setFieldItem(_filterRelation);
		updateTools();
		formField.resetContent();
	}

	/**
	 * Resets form once all modifications were done.
	 */
	public void resetForm()
	{
		// Force a refresh of the 'internal' structure
		LyForm form = (LyForm)_component.getComponent(_parent.getId());

		form.resetContent();
	}

	/**
	 * Updates state of filtering tools.
	 */
	public void updateTools()
	{
		if (_filterRelation == null)
			return;

		LyFilter			 filter = getSelectedFilter();
		LyInfoList<LyFilter> createdFilters = ((LyFilterRelationController)_parent)
				.getCreatedFilters();
		LyTool				 setFilterTool = (LyTool)_component.getComponent(
				LyFilterRelationController.SET_FILTER);
		LyTool				 extendFilterTool = (LyTool)_component.getComponent(
				LyFilterRelationController.EXTEND_SIMPLE_FILTER);
		boolean				 extendedFilter = ((filter != null) && (filter.isExtended()));
		boolean				 createdFilter = ((createdFilters != null)
					&& (createdFilters.contains(filter)));

		setFilterTool.setEnabled(createdFilter && (filter != null));
		extendFilterTool.setEnabled(createdFilter && (filter != null) && !extendedFilter);
	}

	/**
	 * Returns selected field info in the class info list.
	 *
	 * @return selected field info in the class info list
	 */
	public LyRelationFieldInfo getSelectedFieldInfo()
	{
		LyLink	    link;
		LyClassInfo classInfo = ((LyFilterRelationController)_parent).getClassInfo();
		int		    i, n = _fieldRelation.getLinkCount();

		for (i = 0; i < n; i++)
		{
			link = _fieldRelation.getLink(i);

			if (link.isSelected())
			{
				String fieldId = link.getId();

				return (LyRelationFieldInfo)classInfo.getFieldInfo(fieldId);
			}
		}

		return null;
	}

	/**
	 * Returns selected class info in the class info list.
	 *
	 * @return selected class info in the class info list
	 */
	public LyClassInfo getSelectedClassInfo()
	{
		LyApplication application = _parent.getApplication();
        LyClassInfo   classInfo = null;

		if (_classItem instanceof LyChoice)
		{
			LyChoice		    classChoice = (LyChoice)_classItem;
			LyOption		    option;
			int				    i, n = classChoice.getOptionCount();
			String			    classId;
			LyRelationFieldInfo field = getSelectedFieldInfo();

			for (i = 0; i < n; i++)
			{
				option = classChoice.getOption(i);
				if (option.isChecked())
				{
					classId = option.getId();
                    classInfo = application.getClassInfo(classId);

                    // If not found, try to get the classInfo on the relation (Studio)
					if ((classInfo == null)
                            && (field != null)
							&& (i < field.getClassInfoCount())
							&& (field.getClassInfo(i).getId().equals(classId)))
						return field.getClassInfo(i);
					else
						return classInfo;
				}
			}
		}
		else if (_classItem instanceof LyRelation)
		{
			LyRelation		    classRelation = (LyRelation)_classItem;
			LyLink			    link;
			int				    i, n = classRelation.getLinkCount();
			String			    classId;
			LyRelationFieldInfo field = getSelectedFieldInfo();

			for (i = 0; i < n; i++)
			{
				link = classRelation.getLink(i);
				if (link.isSelected())
				{
					classId = link.getId();
                    classInfo = application.getClassInfo(classId);

					// If not found, try to get the classInfo on the relation (Studio)
					if ((classInfo == null)
                            && (field != null)
							&& (i < field.getClassInfoCount())
							&& (field.getClassInfo(i).getId().equals(classId)))
						return field.getClassInfo(i);
					else
						return classInfo;
				}
			}
		}

		return null;
	} // end method getSelectedClassInfo

	/**
	 * Returns selected filter in the filter list.
	 *
	 * @return selected filter in the filter list
	 */
	public LyFilter getSelectedFilter()
	{
		if (_filterRelation == null)
			return null;

		LyLink link;
		int    i, n = _filterRelation.getLinkCount();

		for (i = 0; i < n; i++)
		{
			link = _filterRelation.getLink(i);

			if (link.isSelected())
			{
				String   filterId = link.getId();
				LyFilter filter = null;

				filter = (LyFilter)_parent.getApplication().getInfo(filterId);

				if (filter == null)
				{
					// If filter was not found on the application, search in the list of
					// created filters.
					LyInfoList<LyFilter> createdFilters = ((LyFilterRelationController)_parent)
							.getCreatedFilters();

					if (createdFilters != null)
						filter = createdFilters.getInfo(filterId);
				}

				return filter;
			}
		}

		return null;
	} // end method getSelectedFilter
} // end class LyFilterRelationBuilder
