/**
 * LyFormBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyFormBuilder.java 36257 2012-07-11 09:54:33Z bangiole $
 */

package leon.control.builder;

import java.util.*;

import leon.app.*;
import leon.control.*;
import leon.control.form.*;
import leon.control.form.builder.*;
import leon.data.*;
import leon.info.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.view.*;


/**
 * This builder creates a form view.
 *
 * @see LyFormController
 */
public abstract class LyFormBuilder extends LyAreaBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyFormBuilder.java 36257 2012-07-11 09:54:33Z bangiole $";

	static final String		   NO_TAB_MODE = "NO_TAB";

	/** Default size for Transverse icon (actions) */
	public static int		   DEFAULT_TRANSVERSE_ICON_WIDTH = 15;

	/** Position of context area */
	public static boolean	   SHOW_CONTEXT_AREA = true;

	/** Sub area of main area (used only when SHOW_CONTEXT_AREA is true and multiple context is set). */
	protected LyArea		   _subArea;

	/**
	 * Convert a LyFieldConstraint to a LyConstraint.Fill
	 * @param fill the LyFieldConstraint
	 * @return the corresponding LyConstraint.Fill
	 */
	public static LyConstraint.Fill getFill(LyFieldConstraint.Fill fill)
	{
		if (fill == LyFieldConstraint.Fill.HORIZONTAL)
			return LyConstraint.Fill.HORIZONTAL;
		else if (fill == LyFieldConstraint.Fill.VERTICAL)
			return LyConstraint.Fill.VERTICAL;
		else if (fill == LyFieldConstraint.Fill.BOTH)
			return LyConstraint.Fill.BOTH;
		else if (fill == LyFieldConstraint.Fill.NONE)
			return LyConstraint.Fill.NONE;
		else if (fill == LyFieldConstraint.Fill.DEFAULT)
			return LyConstraint.Fill.DEFAULT;
		else
			return null;
	}

	/**
	 * Convert a LyFieldConstraint.HAlign to a LyConstraint.HAlign
	 * @param hAlign source HAlign
	 * @return converted HAlign
	 */
	public static LyConstraint.HAlign getHAlign(LyFieldConstraint.HAlign hAlign)
	{
		if (hAlign == LyFieldConstraint.HAlign.LEFT)
			return LyConstraint.HAlign.LEFT;
		else if (hAlign == LyFieldConstraint.HAlign.CENTER)
			return LyConstraint.HAlign.CENTER;
		else if (hAlign == LyFieldConstraint.HAlign.RIGHT)
			return LyConstraint.HAlign.RIGHT;
		else if (hAlign == LyFieldConstraint.HAlign.DEFAULT)
			return LyConstraint.HAlign.DEFAULT;
		else
			return null;
	}

	/**
	 * Convert a LyFieldConstraint.VAlign to a  LyConstraint.VAlign
	 * @param vAlign source VAlign
	 * @return converted VAlign
	 */
	public static LyConstraint.VAlign getVAlign(LyFieldConstraint.VAlign vAlign)
	{
		if (vAlign == LyFieldConstraint.VAlign.TOP)
			return LyConstraint.VAlign.TOP;
		else if (vAlign == LyFieldConstraint.VAlign.MIDDLE)
			return LyConstraint.VAlign.MIDDLE;
		else if (vAlign == LyFieldConstraint.VAlign.BOTTOM)
			return LyConstraint.VAlign.BOTTOM;
		else if (vAlign == LyFieldConstraint.VAlign.DEFAULT)
			return LyConstraint.VAlign.DEFAULT;
		else
			return null;
	}

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyFormBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Builds the view component.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{
		LyAction action = ((LyFormController)_parent).getAction();
		LyArea   area = new LyArea();
		LyPanel  noTabPanel;
		LyArea	 contextArea = (SHOW_CONTEXT_AREA ? buildContext() : null);

		LyConstraint	c = new LyConstraint();

		// Add a fill constraint to ensure form will not be displayed with a full height
		// in web viewer.
		if (getViewManager().hasProperty("WEB_MANAGER"))
			c.setFill(LyConstraint.Fill.HORIZONTAL);
		else
			c.setFill(LyConstraint.Fill.BOTH);

		area.setConstraint(c);

		if (contextArea != null)
		{
			LyArea			topArea = new LyArea();
			LyConstraint	topConstraint = new LyConstraint();

			topConstraint.setFill(LyConstraint.Fill.BOTH);
			contextArea.setConstraint(topConstraint);
			topArea.setConstraint(topConstraint);
			topArea.addPanel(area);
			topArea.setBorderPanel(contextArea, LyArea.WEST);
			_component = topArea;
			_subArea = area;
		}
		else
			_component = area;

		if (!(_parent.getParent() instanceof LyFormController))
			setFormComponents(area);

		if (action.hasMark(LyAction.MARK_NO_TABS))
		{
			noTabPanel = buildArea(NO_TAB_MODE);
			if (noTabPanel != null)
				area.addPanel(noTabPanel);
		}
		else
			setTabs(area);

		area.setController(_parent.getId());
		area.setId(_parent.getId());
		area.setScrolling(LyArea.Scrolling.AUTO);
		// Update default content
		updateTitleBar();

		return _component;
	} // end method build

	/**
	 * Returns the main titlebar of the area component.
	 */
	protected LyTitleBar getTitleBar()
	{
		if (_subArea != null)
			return _subArea.getTitleBar();

		return super.getTitleBar();
	}

	/**
	 * Method used to set the tabs of the form if needed.
	 *
	 * @param area The area on which the tabs will be added.
	 */
	protected void setTabs(LyArea area)
	{
		LyViewManager viewManager = getViewManager();
		LyPanel		  noTabPanel, tabs;

		noTabPanel = buildArea(null);
		tabs = buildTabs();

		if (noTabPanel != null)
		{
			if ((tabs != null) && !(viewManager.hasProperty("MULTI_FRAME_FORM")))
			{
				LyTab tab = new LyTab();

				tab.setId(_parent.getId() + "_tab0");
				tab.setLabel("");
				tab.setPanel(noTabPanel);
				((LyTabs)tabs).insertTab(tab, 0);
			}
			else
				area.addPanel(noTabPanel);
		}

		if (tabs != null)
			area.addPanel(tabs);
	}

	/**
	 * Method used to set the toolbar, the titlebar and the commands of the form.
	 *
	 * @param area The area on which the tabs will be added.
	 */
	protected void setFormComponents(LyArea area)
	{
		area.setToolBar(buildToolBar());
        area.setTopRightToolBar(buildRightToolBar());
		area.setTitleBar(buildTitleBar());

        LyCommands commands = buildCommands();

		area.setCommands(commands);
	}

	/**
	 * Complete the toolbar with the form context. If there is a context on the form, three buttons
	 * are added to the toolbar : previous, next and selection.
	 *
	 * @param toolBar the toolbar to complete
	 */
	public void completeToolBarForContext(LyToolBar toolBar)
	{
		if (toolBar == null)
			return;

		LyArea	 contextArea = (SHOW_CONTEXT_AREA ? null : buildContext());

		if (contextArea == null)
		{
			LyEnvironment environment = _parent.getEnvironment();

			if (!environment.testResource("LY_NAVIGATION_IN_FORMS"))
				return;

			LyClassInfo  classInfo = ((LyFormController)_parent).getClassInfo();
			LyController grandParent = ((LyFormController)_parent).getViewController().getParent();

			if (!(grandParent instanceof LySimpleTableController)
					|| (classInfo != ((LySimpleTableController)grandParent).getClassInfo())
                        && !classInfo.inheritsFrom(((LySimpleTableController)grandParent).getClassInfo()))
				return;

			LyObjectList list = ((LyAreaController)grandParent).getContent();

			if ((list == null)
			        || (list.getSize() < 2))
			    return;
		}

		LyTool tool;

		if (toolBar.getComponentCount() > 0)
			toolBar.addSep();

		LyViewManager	viewManager = _parent.getViewManager();
		boolean			isRtl = viewManager.isRtlMode();

		tool = new LyTool(LyFormController.DEMAND_PREVIOUS_OBJECT, "LY_PREVIOUS");
		tool.setImage(isRtl ? "LY_RIGHT" : "LY_LEFT");
		tool.setEnabled(true);
		tool.setDialog(false);
		toolBar.addTool(tool);

		tool = new LyTool(LyFormController.DEMAND_NEXT_OBJECT, "LY_NEXT");
		tool.setImage(isRtl ? "LY_LEFT" : "LY_RIGHT");
		tool.setEnabled(true);
		tool.setDialog(false);
		toolBar.addTool(tool);

		if (contextArea != null)
		{
			tool = new LyTool(LyFormController.DEMAND_SELECTION, "LY_SELECTION");
			tool.setImage("LY_ACTION_MENU");
			tool.setEnabled(true);
			tool.setDialog(false);
			tool.setComponent(contextArea);
			toolBar.addTool(tool);
		}
	} // end method completeToolBarForContext

	/**
	 * Build the toolbar of the form.
	 *
	 * @return the built toolbar
	 */
	protected LyToolBar buildToolBar()
	{
		LyFormController formController = (LyFormController)_parent;
		LyAction		 formAction = formController.getAction();

		if (formAction.hasMark(LyAction.MARK_NO_TOOLBAR))
			return null;

		LyClassInfo classInfo = formController.getClassInfo();

		if (classInfo == null)
			return null;

		// Build toolBar
		LyInfoList<LyAction> actions = getToolbarActions(formController, classInfo);

		LyToolBar			 toolBar = buildToolBar(actions);

		// Complete toolBar
		toolBar = completeToolBar(toolBar);

		// Add extra tools to in this toolBar if necessary
		toolBar = addExtraTools(toolBar);

		// If the toolbar is null and if we have more than 1 object, we must create a toolbar...
		LyObjectList list = formController.getObjectList();

		if ((toolBar == null)
				&& ((formAction.getMax() < 0) || (formAction.getMax() > 1))
				&& ((list == null) || (list.getSize() > 1)))
		{
			toolBar = new LyToolBar();
			toolBar.setId("_toolBar");
			toolBar.setFill(LyToolBar.Fill.HORIZONTAL);
			toolBar.setColor("LY_DASHBOARD_HORIZONTAL_BACKGROUND");
		}

		// Add context part
		completeToolBarForContext(toolBar);

        LyEnvironment    environment = _parent.getEnvironment();

		if ((toolBar != null)
			&& (formAction.hasMark(LyAction.MARK_NO_TITLEBAR))
				&& environment.testResource("LY_TITLEBAR_LEFT_IMAGE"))
		{
			String logo = formController.getImage();

			toolBar.setImage(logo);
		}

		return toolBar;
	} // end method buildToolBar

	/**
	 * Check if the action should be added in the form toolbar Return true if not hidden, has the
	 * form mark set and if sessions validates the action (validateAction).
	 *
	 * @param  action the action we want to test
	 * @return true if the action should be added, false otherwise
	 */
	protected boolean validateAction(LyAction action)
	{
		if (action.hasMark(LyAction.MARK_HIDDEN))
			return false;
		if (!action.hasMark(LyAction.MARK_FORM))
			return false;
		if ((action.getMin() > 1) // Needs more than one object
				|| (action.getMax() == 0)) // Asks for no selection
			return false;
		if (action == ((LyFormController)_parent).getAction())
			return false;

		LySession session = _parent.getSession();

		if (!session.validateAction(action, _parent))
			return false;

		return true;
	}

	/**
	 * Add extra tools to the toolbar of the form.
	 *
	 * @param toolBar the toolbar that must be updated
	 * @return the completed toolbar
	 */
	protected LyToolBar addExtraTools(LyToolBar toolBar)
	{
		return toolBar;
	}

	/**
	 * Build a tool of the toolbar.
	 *
	 * @param  action the action which tool must be built
	 * @param  enable indicates if tool must be enabled
	 * @return a tool corresponding to the given action
	 */
	protected LyTool buildTool(LyAction action, boolean enable)
	{
		LyObject object = ((LyFormController)_parent).getObject();
		LyFilter filter = action.getFilter();

		enable = enable && ((filter == null) || !filter.match(object));

		return super.buildTool(action, enable);
	}

	/**
	 * Build area for the form context.
	 *
	 * @return the area for the form context
	 */
	protected LyArea buildContext()
	{
		LyFormController formController = (LyFormController)_parent;
		LyAction		 formAction = formController.getAction();
		LyObjectList     list = formController.getObjectList();

		//FFT01109 formAction with max=0 may have a context
		if ((formAction.getMax() == 1)
					|| (list == null)
					|| ((list != null) && (list.getSize() < 2)))
			return null;

		// Create hrefs
		LyArea	 area = new LyArea();
		LyForm	 form = new LyForm("_formContext");
		LyHRefs  hrefs = new LyHRefs();
		LyHRef   href;
		LyObject object;
		int		 i, n = ((list == null) ? 0 : list.getSize());

		for (i = 0; i < n; i++)
		{
			object = list.getObject(i);
			href = new LyHRef();

			href.setId(object.getId());
			href.setLabel(getName(object));
			href.setImage(_parent.getImage(object));

			if (SHOW_CONTEXT_AREA)
			{
				LyFormField formField = new LyFormField("_formObjects" + i);

				formField.setFieldItem(href);
				formField.setEditable(false);
				form.addFormField(formField);
			}
			else
				hrefs.addHRef(href);
		}

		// Adjust hrefs height to its content with a limit of 15
		hrefs.setNRows(Math.min(15, hrefs.getHRefCount()));
		hrefs.setBorder(false);

		if (!SHOW_CONTEXT_AREA)
		{
			LyFormField formField = new LyFormField("_formObjects");

			formField.setFieldItem(hrefs);
			formField.setEditable(false);
			form.addFormField(formField);
		}
		else
		{
			area.setBorder(LyArea.Border.IN);
			area.setScrolling(LyArea.Scrolling.AUTO);
		}

		form.setController(_parent.getId());
		form.setBorder(false);
		form.setEditable(false);
		area.addPanel(form);

		return area;
	} // end method buildContext

	/**
	 * Build the tabs panel of the form.
	 *
	 * @return the tabs panel of the form
	 */
	protected LyPanel buildTabs()
	{
		LyClassInfo					  classInfo = ((LyFormController)_parent).getClassInfo();
		int							  i, n = ((classInfo == null) ? 0 : classInfo.getFieldInfoCount());
		LyFieldInfo					  fieldInfo;
		String						  tabName;
		Vector<String>				  tabList = new Vector<String>();
		Hashtable<String, LyFieldTab> fieldtabList = new Hashtable<String, LyFieldTab>();

		// Get tabs list
		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			if (!((LyFormController)_parent).isDisplayed(fieldInfo))
				continue;
			tabName = fieldInfo.getTab();

			if ((tabName != null) && !tabList.contains(tabName))
			{
				tabList.addElement(tabName);

				if (fieldInfo.getFieldTab() != null)
					fieldtabList.put(tabName, fieldInfo.getFieldTab());
			}
			else if ((tabName == null) && (needTab(fieldInfo)))
				tabList.addElement(fieldInfo.getName());
		}

		if (tabList.size() == 0)
			return null;

		// Build tabs
		LyTabs		tabs = null;
		int			t, nt = tabList.size();
		LyTab		tab;
		LyFieldTab	fieldTab;
		String		tabId;

		for (t = 0; t < nt; t++)
		{
			tabName = tabList.elementAt(t);
			tabId = _parent.getId() + "_tab_" + t;
			fieldTab = fieldtabList.get(tabName);
			tab = buildTab(tabId, tabName, (fieldTab == null ? null : fieldTab.getPrefix()));

			if (tab != null)
			{
				if (tabs == null)
					tabs = new LyTabs();
				tabs.addTab(tab);
			}
		}

		LyConstraint	constraint = new LyConstraint();

		constraint.setFill(LyConstraint.Fill.BOTH);
		constraint.setRemainder(true);
		tabs.setConstraint(constraint);

		return tabs;
	} // end method buildTabs

	/**
	 * Method used to know if a field need a tab.
	 *
	 * @param  fieldInfo The fieldInfo concerned.
	 * @return true if the fieldInfo needs a tab, false otherwise.
	 */
	protected boolean needTab(LyFieldInfo fieldInfo)
	{
		return false;
	}

	/**
	 * Build a tab of the tabs panel.
	 *
	 * @param  tabId   the id of the tab to build
	 * @param  tabName the label of the tab to build
	 * @return the tab with the given label for the tabs panel
	 */
	protected LyTab buildTab(String tabId, String tabName)
	{
		return buildTab(tabId, tabName, null);
	}

	/**
	 * Build a tab of the tabs panel.
	 *
	 * @param  tabId    the id of the tab to build
	 * @param  tabName  the label of the tab to build
	 * @param  tabImage the image of the tab to build
	 * @return the tab with the given label for the tabs panel
	 */
	protected LyTab buildTab(String tabId, String tabName, String tabImage)
	{
		LyTab  tab = new LyTab();
		LyArea area = buildArea(tabName);

		if ((tab != null) && (area != null))
		{
			// Add an intermediate area with an horizontal fill only. With this area, groups will be
			// closed properly (they will be displayed at the top of the tab when closed).
			LyArea			topTabArea = new LyArea();
			LyConstraint	constraint = new LyConstraint();

			constraint.setFill(LyConstraint.Fill.HORIZONTAL);
			constraint.setVAlign(LyConstraint.VAlign.TOP);
			area.setConstraint(constraint);
			topTabArea.addPanel(area);
			tab.setId(tabId);
			tab.setLabel(tabName);
			tab.setImage(tabImage);
			tab.setPanel(topTabArea);
		}
		else
			return null;

		return tab;
	}

	/**
	 * Build the area contained in a tab of the tabs panel.
	 *
	 * @param  tabName the label of the tab containing the area
	 * @return the area contained in the tab with the given label
	 */
	protected LyArea buildArea(String tabName)
	{
		LyFormController formController = (LyFormController)_parent;
		LyClassInfo		 classInfo = formController.getClassInfo();
		LyEnvironment    environment = _parent.getEnvironment();

		if (classInfo == null)
			environment.logWarning(
				"LyFormBuilder : no classInfo was specified to build the form. Please check action parameters : "
				+ formController.getAction());

		int			   i, n = ((classInfo == null) ? 0 : classInfo.getFieldInfoCount());
		LyFieldInfo    fieldInfo;
		Vector<String> groupList = new Vector<String>();
		String		   group;
		String		   NULL_GROUP = new String();
		LyArea		   area = null, subArea;
		LyForm		   form;
		boolean		   needTab;
		boolean		   assemble = environment.testResource("LY_GROUP_FIELDS_WITH_NO_GROUP");

		// Build fields off the given tab
		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			needTab = needTab(fieldInfo);

			if ((tabName == null) && needTab)
				continue;

			if ((tabName != NO_TAB_MODE)
					&& (fieldInfo.getTab() != tabName))
			{
				if ((!needTab)
						|| (needTab && !(fieldInfo.getName().equalsIgnoreCase(tabName))))
					continue;
			}

			if (!formController.isDisplayed(fieldInfo))
				continue;

			group = fieldInfo.getGroup();
			if (group == null)
				group = NULL_GROUP;

			if (!groupList.contains(group))
			{
				if (group == NULL_GROUP)
				{
					if (!assemble)
						form = buildForm(tabName, null, i, false);
					else
						form = buildForm(tabName, null);
				}
				else
					form = buildForm(tabName, group);

				if (form == null)
					continue;

                // Check visibility
                int            f, nf = form.getFormFieldCount();
                boolean        visible = false;

                for (f = 0; (f < nf) && !visible; f++)
                    visible = !form.getFormField(f).isHidden();

                if (!visible)
                   continue;

				if (area == null)
				{
					LyConstraint  constraint = new LyConstraint();

					constraint.setFill(LyConstraint.Fill.BOTH);
					constraint.setRemainder(true);
					area = new LyArea();
					area.setScrolling(LyArea.Scrolling.AUTO);
					area.setConstraint(constraint);
				}

				if (group == NULL_GROUP)
				{
					area.addPanel(form);

					LyConstraint		constraint = new LyConstraint();

					constraint.setFill(LyConstraint.Fill.HORIZONTAL);
					constraint.setRemainder(true);
					form.setConstraint(constraint);
				}
				else
				{
					LyFieldGroup		fieldGroup = fieldInfo.getFieldGroup();
					LyFieldGroup.State	state = fieldGroup.getState();
					boolean				closable = (state != LyFieldGroup.State.FIXED);

					subArea = new LyArea();
					subArea.setId("group_" + fieldGroup.hashCode());
					subArea.setBorder(LyArea.Border.IN);
					subArea.setLabel(group);
					subArea.addPanel(form);
					subArea.setClosable(closable);
					subArea.setOpen((state == LyFieldGroup.State.OPENED) || (!closable && (state == LyFieldGroup.State.FIXED)));

					if (fieldGroup.getConstraint() != null)
					{
						LyFieldConstraint	groupConstraint = fieldGroup.getConstraint();
						LyConstraint		constraint = new LyConstraint();

						constraint.setRowSpan(groupConstraint.getRowSpan());
						constraint.setColSpan(groupConstraint.getColSpan());
						constraint.setRemainder(groupConstraint.isRemainder());
						constraint.setFill(LyFormBuilder.getFill(groupConstraint.getFill()));
						constraint.setHAlign(LyFormBuilder.getHAlign(groupConstraint.getHAlign()));
						constraint.setVAlign(LyFormBuilder.getVAlign(groupConstraint.getVAlign()));
						subArea.setConstraint(constraint);
					}
					else
					{
						LyConstraint		constraint = new LyConstraint();

						constraint.setRemainder(true);
						constraint.setFill(LyConstraint.Fill.BOTH);
						subArea.setConstraint(constraint);
					}

					area.addPanel(subArea);
				}

				groupList.addElement(group);
			}

			if (!assemble
					&& (group != NULL_GROUP)
					&& groupList.contains(NULL_GROUP))
				groupList.removeElement(NULL_GROUP);
		}

		if (area != null)
		{
			LyConstraint	constraint = new LyConstraint();

			constraint.setFill(LyConstraint.Fill.BOTH);
			constraint.setRemainder(true);
			area.setConstraint(constraint);
		}

		return area;
	} // end method buildArea

	/**
	 * Build the form contained in a tab for a given group.
	 *
	 * @param  tabName the label of the tab containing the form
	 * @param  group   the label of the group containing the form
	 * @return the form contained in the tab with the given label and in the group with the given
	 *         label
	 */
	protected LyForm buildForm(String tabName, String group)
	{
		return buildForm(tabName, group, 0, true);
	}

	/**
	 * Build a form
	 * @param tabName tab name
	 * @param group group name
	 * @param startIndex start index in the field list
	 * @param assemble true to assemble field
	 * @return a form
	 */
	protected LyForm buildForm(String tabName, String group, int startIndex, boolean assemble)
	{
		LyClassInfo		  classInfo = ((LyFormController)_parent).getClassInfo();
		LyForm			  form = null;
		LyFormField		  formField;
		int				  i, n = ((classInfo == null) ? 0 : classInfo.getFieldInfoCount());
		LyFieldInfo		  fieldInfo;
		LyFormController  formController = (LyFormController)_parent;
		boolean			  editable = !(formController instanceof LyConsultController) || (formController instanceof LySetController);
		String			  formId;
		boolean			  needTab;
		LyFieldController fieldController;

		for (i = startIndex; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);

			needTab = needTab(fieldInfo);

			if ((tabName == null) && needTab)
				continue;

			if ((tabName != NO_TAB_MODE) && (fieldInfo.getTab() != tabName))
			{
				if ((!needTab)
						|| (needTab && !(fieldInfo.getName().equalsIgnoreCase(tabName))))
					continue;
			}

			if (!formController.isDisplayed(fieldInfo))
				continue; // Skipped
			if (fieldInfo.getGroup() != group)
			{
				if ((form != null) && !assemble)
					return form;
				else
					continue;
			}

			fieldController = formController.getFieldController(fieldInfo);
			if (fieldController == null)
				continue;

			formField = fieldController.getFormField();

            if (form == null)
            {
                formId = _parent.getId();
                if (tabName != null)
                    formId += LyController.ID_SEP + tabName;
                if (group != null)
                    formId += LyController.ID_SEP + group;

                // Replace space by underscore
                formId = LyString.replace(formId, " ", "_");
                form = new LyForm(formId);
                form.setController(_parent.getId());
                form.setBorder(false);
                form.setEditable(editable);

                LyConstraint constraint = new LyConstraint();

                constraint.setFill(LyConstraint.Fill.HORIZONTAL);
        		constraint.setRemainder(true);
                form.setConstraint(constraint);
            }

			if (formField != null)
				form.addFormField(formField);
		}

		return form;
	} // end method buildForm

	/**
	 * Build a form field for a given fieldInfo.
	 *
	 * @param  fieldInfo the fieldInfo corresponding to the built form field
	 * @param  editable  indicates whether the form field must be editable or not
	 * @return the form field corresponding to the fieldInfo
	 */
	public LyFormField buildFormField(LyFieldInfo fieldInfo, boolean editable)
	{
		LyFormFieldBuilder builder = getFormFieldBuilder(fieldInfo);

		return builder.buildFormField(editable);
	}

	/**
	 * Get caption constraint.
	 * @param fieldInfo the fieldInfo
	 * @param formField the formField
	 * @return the computed constraint
	 */
	public LyConstraint getCaptionConstraint(LyFieldInfo fieldInfo,
		LyFormField formField)
	{
		LyFormFieldBuilder builder = getFormFieldBuilder(fieldInfo);

		return builder.getCaptionConstraint(formField);
	}

	/**
	 * Get field constraint.
	 * @param fieldInfo the fieldInfo
	 * @param formField the formField
	 * @return the computed constraint
	 */
	public LyConstraint getFieldConstraint(LyFieldInfo fieldInfo,
		LyFormField formField)
	{
		LyFormFieldBuilder builder = getFormFieldBuilder(fieldInfo);

		return builder.getFieldConstraint(formField);
	}

	/**
	 * Build a field item for a given fieldInfo.
	 *
	 * @param  fieldInfo the fieldInfo corresponding to the built field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildFieldItem(LyFieldInfo fieldInfo)
	{
		return buildFieldItem(fieldInfo, null);
	}

	/**
	 * Build a field item for a given fieldInfo.
	 *
	 * @param  fieldInfo the fieldInfo corresponding to the built field item
	 * @param  value     the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildFieldItem(LyFieldInfo fieldInfo, Object value)
	{
		LyFormFieldBuilder builder = getFormFieldBuilder(fieldInfo);

		return builder.buildFieldItem();
	}

	/**
	 * Build a field item for a text fieldInfo.
	 *
	 * @param  textField the text fieldInfo corresponding to the built field item
	 * @param  value     the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildText(LyTextFieldInfo textField, Object value)
	{
		LyFormController   form = (LyFormController)_parent;
		LyFormFieldBuilder builder = (LyFormFieldBuilder)form.getFieldController(textField)
				.getBuilder();

		return builder.buildText();
	}

	/**
	 * Build a field item for a date fieldInfo.
	 *
	 * @param  dateField the date fieldInfo corresponding to the built field item
	 * @param  value     the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildDate(LyDateFieldInfo dateField, Object value)
	{
		LyFormController   form = (LyFormController)_parent;
		LyFormFieldBuilder builder = (LyFormFieldBuilder)form.getFieldController(dateField)
				.getBuilder();

		return builder.buildDate();
	}

	/**
	 * Build a field item for a number fieldInfo.
	 *
	 * @param  numberField the number fieldInfo corresponding to the built field item
	 * @param  value       the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildNumber(LyNumberFieldInfo numberField, Object value)
	{
		LyFormController   form = (LyFormController)_parent;
		LyFormFieldBuilder builder = (LyFormFieldBuilder)form.getFieldController(numberField)
				.getBuilder();

		return builder.buildNumber();
	}

	/**
	 * Build a field item for a relation fieldInfo.
	 *
	 * @param  relationField the relation fieldInfo corresponding to the built field item
	 * @param  value         the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildRelation(LyRelationFieldInfo relationField, Object value)
	{
		LyFormController   form = (LyFormController)_parent;
		LyFormFieldBuilder builder = (LyFormFieldBuilder)form.getFieldController(relationField)
				.getBuilder();

		return builder.buildRelation();
	}

	/**
	 * Build a field item for a choice fieldInfo.
	 *
	 * @param  choiceField the choice fieldInfo corresponding to the built field item
	 * @param  value       the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildChoice(LyChoiceFieldInfo choiceField, Object value)
	{
		LyFormController   form = (LyFormController)_parent;
		LyFormFieldBuilder builder = (LyFormFieldBuilder)form.getFieldController(choiceField)
				.getBuilder();

		return builder.buildChoice();
	}

	/**
	 * Build a field item for a file fieldInfo.
	 *
	 * @param  fileField the file fieldInfo corresponding to the built field item
	 * @param  value     the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildFile(LyFileFieldInfo fileField, Object value)
	{
		LyFormController   form = (LyFormController)_parent;
		LyFormFieldBuilder builder = (LyFormFieldBuilder)form.getFieldController(fileField)
				.getBuilder();

		return builder.buildFile();
	}

	/**
	 * Build a field item for an array fieldInfo.
	 *
	 * @param  arrayField the array fieldInfo corresponding to the built field item
	 * @param  value      the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildArray(LyArrayFieldInfo arrayField, Object value)
	{
		LyFormController   form = (LyFormController)_parent;
		LyFormFieldBuilder builder = (LyFormFieldBuilder)form.getFieldController(arrayField)
				.getBuilder();

		return builder.buildArray();
	}

	/**
	 * Build a field item for an struct fieldInfo.
	 *
	 * @param  structField the struct fieldInfo corresponding to the built field item
	 * @param  value       the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildStruct(LyStructFieldInfo structField, Object value)
	{
		LyFormController   form = (LyFormController)_parent;
		LyFormFieldBuilder builder = (LyFormFieldBuilder)form.getFieldController(structField)
				.getBuilder();

		return builder.buildStruct();
	}

	/**
	 * Build a fieldItem for a fieldinfo to present its value.
	 *
	 * @param  fieldInfo the fieldInfo corresponding to the built field item
	 * @param  value     the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildValueLabel(LyFieldInfo fieldInfo, Object value)
	{
		LyFormFieldBuilder builder = getFormFieldBuilder(fieldInfo);

		return builder.buildValueLabel();
	}

	/**
	 * Build a toolbar for a given fieldInfo. The built toolbar is added to the form field
	 * corresponding to the given fieldInfo
	 *
	 * @param  fieldInfo the fieldInfo corresponding to the built toolbar
	 * @return the toolbar for the given fieldInfo
	 */
	public LyToolBar buildToolBar(LyFieldInfo fieldInfo)
	{
		LyFormFieldBuilder builder = getFormFieldBuilder(fieldInfo);

		return builder.buildToolBar();
	}

	/**
	 * Builds the toolbar for the given field and list of actions.
	 *
	 * @param  fieldInfo the field
	 * @param  actions   List of pairs (action, classInfo)
	 * @return the toolbar for the given field
	 */
	public LyToolBar buildToolBar(LyFieldInfo fieldInfo,
		LyInfoDataList<LyAction, LyClassInfo> actions)
	{
		LyFormFieldBuilder builder = getFormFieldBuilder(fieldInfo);

		return builder.buildToolBar(actions);
	}

	/**
	 * Tells if we must ignore the class fields of relation while building toolbars.<br>
	 * This is used only for filter builders.
	 *
	 * @return true;
	 */
	boolean ignoreRelationClassField()
	{
		return false;
	}

	/**
	 * Build a command bar for the form.
	 *
	 * @return a command bar for the form
	 */
	protected abstract LyCommands buildCommands();

	/**
	 * Accessor to the builder associated with the current fieldinfo.
	 *
	 * @param  fieldInfo the fieldinfo
	 * @return the LyFormFieldBuilder associated with the given fieldInfo.
	 */
	protected LyFormFieldBuilder getFormFieldBuilder(LyFieldInfo fieldInfo)
	{
		LyFormController   form = (LyFormController)_parent;

		LyFormFieldBuilder builder = (LyFormFieldBuilder)form.getFieldController(fieldInfo)
				.getBuilder();

		return builder;
	}
} // end class LyFormBuilder
