/**
 * LyFrameBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyFrameBuilder.java 35018 2012-03-23 15:17:40Z jlsaouli $
 */

package leon.control.builder;

import java.io.*;
import leon.app.*;
import leon.control.*;
import leon.info.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.view.*;


/**
 * This builder creates a view for a frame controller.
 *
 * @see LyFrameController
 * @see LyViewBuilder
 */
public class LyFrameBuilder extends LyViewBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyFrameBuilder.java 35018 2012-03-23 15:17:40Z jlsaouli $";

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyFrameBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Builds the view component for the specified parent controller. The component may create the
	 * XML Tree or the internal view.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{
		LyArea area = buildArea();

		if (_component != null)
			// The building of the view was done by a call to getView while constructing area
			return _component;

		LyFrame frame = (LyFrame)area.getParent();

		if (frame == null)
		{
			LyEnvironment env = _parent.getEnvironment();

			frame = new LyFrame(_parent.getId(), env);
		}

		if (frame.getLabel() == null)
			frame.setLabel(_parent.getName());
		frame.setController(_parent.getId());

		frame.setIcon(buildIcon());
		frame.setMenuBar(buildMenuBar());
		frame.setArea(area);
		frame.setMessages(buildMessages());

		LyViewController viewController = (LyViewController)_parent;
		LyAreaController areaController = viewController.getAreaController();
		LyAction		 action = areaController.getAction();

		frame.setDecorated(!action.hasMark(LyAction.MARK_NO_DECORATION));

		// Sets geometry
		LyGeometry geometry = action.getGeometry();

		if (geometry != null)
		{
			frame.setX(geometry.getX());
			frame.setY(geometry.getY());
			frame.setWidth(geometry.getWidth());
			frame.setHeight(geometry.getHeight());
			frame.setMinWidth(geometry.getMinWidth());
			frame.setMinHeight(geometry.getMinHeight());
		}

		_component = frame;
		frame.setRootFrame(viewController.getParent() == null);

		return frame;
	} // end method build

	/**
	 * Builds the icon of the frame.
	 *
	 * @return the icon component description
	 */
	protected LyIcon buildIcon()
	{
		LyViewController viewController = (LyViewController)_parent;
		LyAreaController areaController = viewController.getAreaController();
		LyArea			 area = (LyArea)areaController.getComponent();
		LyTitleBar		 titleBar = area.getTitleBar();
		String			 image = null;
		String			 label = null;

		if (titleBar != null)
		{
			image = titleBar.getImage();
			label = titleBar.getLeftLabel();
		}
		if (image == null)
		{
			LyAction action = areaController.getAction();

			image = _parent.getImage(action);
		}

		if (image == null)
			return null;
		else
			return new LyIcon(label, image);
	}

	/**
	 * Builds the menuBar of the frame Method recurses into sub-Areas (invokes completeMenuBar()).
	 *
	 * @return the icon component description
	 */
	protected LyMenuBar buildMenuBar()
	{
		LyViewController viewController = (LyViewController)_parent;
		LyAreaController areaController = viewController.getAreaController();
		LyAction		 action = areaController.getAction();

		if (action.hasMark(LyAction.MARK_NO_MENUBAR))
			return null;

		LyViewManager viewManager = getViewManager();
		LyMenuBar     menuBar = null;

		if ((viewManager != null) && (viewManager.hasProperty("MENU")))
		{
			menuBar = new LyMenuBar();
			menuBar.setSorted(true);
			menuBar.setId("_menuBar");
			completeMenuBar(menuBar, LyViewController.LEVEL_ROOT);
		}

		return menuBar;
	}

	/**
	 * This method is invoked to enable the area to complete the menuBar of the view and add its own
	 * menu items. Default implementation invokes the method on the builder instance.
	 *
	 * @param menuBar the menuBar structure
	 * @param level   : level of current Controller. one of the LEVEL_XXX contants.
	 */
	public void completeMenuBar(LyMenuBar menuBar, int level)
	{
		LyFrameController frameController = (LyFrameController)_parent;
		LyAreaController  areaController = frameController.getAreaController();
		LyAction		  action = areaController.getAction();

		if (action.hasMark(LyAction.MARK_NO_MENUBAR))
			return;

		LyMenu   menu;
		LyMenu   fileMenu = null;
		LyItem   item;
		String   id = null;
		LyAction actionItem = null;
		String   image = null;

		// Add menu for the frame
		if (level == LyViewController.LEVEL_ROOT)
		{
			// FILE
			fileMenu = new LyMenu("LY_MENU_FILE");
			fileMenu.setId("_menu_" + _parent.getId());

			menuBar.addMenu(fileMenu);

			LyEnvironment environment = _parent.getEnvironment();
			boolean		  undoRedoMenu = (environment.testResource("LY_ENABLE_UNDO_REDO")
						|| (frameController.getParent() == null));

			if (undoRedoMenu)
			{
		        String	undoRedoStr = environment.getResource("LY_NB_UNDO_REDO");
		        
		        if ((undoRedoStr != null)
		        	&& (undoRedoStr.length() > 0))
		        {
					int nbUndo = environment.getIntResource("LY_NB_UNDO_REDO");
	
					if (nbUndo != 0)
					{
						LyMenu editionMenu = new LyMenu("LY_MENU_EDITION");
	
						menuBar.addMenu(editionMenu);
	
						LyApplication		 application = _parent.getApplication();
						LyProject			 project = (LyProject)application.getInfo(
								"__leonUndoRedoActions");
						LyInfoList<LyAction> actions = null;
	
						if (project != null)
							actions = project.getActions();
	
						if (actions != null)
						{
							int    i, n = actions.getSize();
							String shortcut;
	
							for (i = 0; i < n; i++)
							{
								actionItem = actions.getInfo(i);
								id = actionItem.getId();
								item = new LyItem(id, actionItem.getName());
								image = areaController.getMenuItemImage(actionItem);
								item.setImage(image);
								item.setEnabled(false);
								item.setController(_parent.getId());
								shortcut = actionItem.getShortcut();
								if (shortcut != null)
									item.setShortcut(shortcut);
								editionMenu.addItem(item);
							}
	
							editionMenu.addSep();
						}
					}
				}
			}
		}
	      
		// Recurse in the client area
		super.completeMenuBar(menuBar, level);

		if (level == LyViewController.LEVEL_ROOT)
		{
			if (frameController.getParent() == null)
			{
				LyEnvironment	environment = _parent.getEnvironment();

				if (environment.testEnv("LY_USE_FILES")
					|| environment.testEnv("LY_SAVE_FILES"))
				{
					id = LyFrameController.DEMAND_LOAD_DATA;
					item = new LyItem(id, "LY_ITEM_LOAD_DATA");
					actionItem = (LyAction)_parent.getApplication().getInfo(id);
					image = areaController.getMenuItemImage(actionItem);
					item.setImage(image);
					fileMenu.addItem(item);
					id = LyFrameController.DEMAND_SAVE_DATA;
					item = new LyItem(id, "LY_ITEM_SAVE_DATA");
					actionItem = (LyAction)_parent.getApplication().getInfo(id);
					image = areaController.getMenuItemImage(actionItem);
					item.setImage(image);
					fileMenu.addItem(item);
					fileMenu.addSep();
				}

				if (environment.isDebugLevel(LyEnvironment.DEBUG_PERF))
				{
					id = LyFrameController.DEMAND_RUN_GC;
					item = new LyItem(id, "LY_ITEM_RUN_GC");
					actionItem = (LyAction)_parent.getApplication().getInfo(id);
					image = areaController.getMenuItemImage(actionItem);
					item.setImage(image);
					fileMenu.addItem(item);
					fileMenu.addSep();
				}

				LyComponent     lastComp = fileMenu.getComponent(fileMenu.getComponentCount() - 1);
				
                if ((lastComp != null)
                        && (lastComp.getId() != LyComponent.SEP_ID))
                    fileMenu.addSep();

				id = LyFrameController.DEMAND_EXIT;
				item = new LyItem(id, "LY_ITEM_QUIT");
				actionItem = (LyAction)_parent.getApplication().getInfo(id);
				image = areaController.getMenuItemImage(actionItem);
				item.setImage(image);
				item.setDialog(true);
				item.setShortcut("LY_SHORTCUT_QUIT");
				fileMenu.addItem(item);
			}
			else
			{
				id = LyFrameController.DEMAND_CLOSE;
				item = new LyItem(id, "LY_ITEM_CLOSE");
				actionItem = (LyAction)_parent.getApplication().getInfo(id);
				image = areaController.getMenuItemImage(actionItem);
				item.setImage(image);
				item.setShortcut("LY_SHORTCUT_CLOSE");
				fileMenu.addItem(item);
			}

			// HELP
			LyEnvironment environment = _parent.getEnvironment();
			String		  helpStr = environment.getString("LY_MENU_HELP");

			if ((helpStr == null)
					|| (helpStr.length() == 0))
				return;

			boolean     showHelpMenu = false;
			boolean     helpMenuEverExists = false;
			
			menu = (LyMenu)menuBar.getComponentByLabel("LY_MENU_HELP");
			if (menu == null)
				menu = new LyMenu("LY_MENU_HELP");
			else
			{
			    menu.addSep();
				helpMenuEverExists = true;
			}
			
			LyAction console = _parent.getApplication().getAction("_console");
			
			if(console != null)
			{
				String 	filePath = environment.getEnv("LY_OUTPUT_FILE");
				
				id = LyViewController.DEMAND_LOG;
				item = new LyItem(id, "LY_ITEM_LOG");
				actionItem = (LyAction)_parent.getApplication().getInfo(id);
				image = areaController.getMenuItemImage(actionItem);
				item.setImage(image);
				item.setDialog(true);
				menu.addItem(item);
				showHelpMenu = true;
					
				if(filePath != null)
				{
					filePath = _parent.getEnvironment().getString().expand(filePath);
				
					File	logfile = new File(filePath);
					if(!logfile.exists())
						item.setEnabled(false);
				}
				else
					item.setEnabled(false);
			}
			
	        String version = environment.translate("LY_ROOT_VERSION");
	            
			if ((version != null) && (version.trim().length() > 0))
			{
				item = new LyItem(LyViewController.DEMAND_VERSION, "LY_ITEM_VERSION");
				item.setDialog(true);
				menu.addItem(item);
				showHelpMenu = true;
			}

	        String about = environment.translate("LY_APPLICATION_ABOUT");
	         
			if ((about != null) && (about.trim().length() > 0))
			{

				id = LyViewController.DEMAND_ABOUT;
				item = new LyItem(id, "LY_ITEM_ABOUT");
				actionItem = (LyAction)_parent.getApplication().getInfo(id);
				image = areaController.getMenuItemImage(actionItem);
				item.setImage(image);
				item.setDialog(true);
				menu.addItem(item);
				showHelpMenu = true;
			}
			
			if (showHelpMenu && !helpMenuEverExists)
				menuBar.addMenu(menu);

			menu.setId("_menu_" + _parent.getId());
		}
	} // end method completeMenuBar

	/**
	 * builds the message bar of the frame.
	 *
	 * @return the messageBar if any
	 */
	public LyMessages buildMessages()
	{
		LyMessages messages = new LyMessages();

		// Nothing to be done at LEVEL_ROOT

		// recurse
		completeMessages(messages, LyViewController.LEVEL_PRIMARY);

		int messageCount = messages.getMessageCount();

		if (messageCount > 0)
			return messages;

		messages.free();
		return null;
	}
} // end class LyFrameBuilder
