/**
 * LyGraphicalChartBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyGraphicalChartBuilder.java 41023 2012-10-04 09:19:37Z bangiole $
 */

package leon.control.builder;

import java.util.*;

import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.getinterface.*;
import leon.view.mobile.*;


/**
 * This class builds the graphical chart area.
 */
public class LyGraphicalChartBuilder extends LyAreaBuilder
{
    /** File version. */
    public static final String __VERSION = "$Id: LyGraphicalChartBuilder.java 41023 2012-10-04 09:19:37Z bangiole $";

    /** Separator used in legends between group field and y fields. */
    public static String GROUP_FIELD_SEP = " / ";

    /** Current chart. */
    protected LyChart _chart;

    /** Specific legend labels. */
    protected Hashtable<String, String> _specificLegendLabels;

    /**
     * Set a Date Format for Y variables. For example : if there is one Y var DATE_TIME and one Y
     * var DATE, the result is DATE_TIME. if there is one Y var TIME and one Y var HOUR, the result
     * is TIME.
     *
     * @param   yFields    the Y variables selected (type date)
     * @param   classInfo  corresponds to the class of yFields
     * @param   chartInfo  corresponds to the current chart
     *
     * @return  The Date Type of the whole Y variables.
     */
    public static String makeYDateType(Vector<String> yFields, LyClassInfo classInfo, LyChartInfo chartInfo)
    {
        LyDateFieldInfo.Type dateType = null, yFieldType;
        LyFieldInfo          yField = null;
        int                  i, nFields = yFields.size();

        // Initialisation of dateType : the  date type of the first Y variable
        yField = classInfo.getFieldInfo(yFields.elementAt(0));

        if ((chartInfo.getTypeVarY() == LyChartInfo.Type.DATE)
                && (yField instanceof LyDateFieldInfo))
            dateType = ((LyDateFieldInfo)yField).getType();

        for (i = 1; i < nFields; i++)
        {
            yField = classInfo.getFieldInfo(yFields.elementAt(i));

            // fill the Y Date Type parameter (step by step)
            if ((chartInfo.getTypeVarY() == LyChartInfo.Type.DATE)
                    && (yField instanceof LyDateFieldInfo))
            {
                yFieldType = ((LyDateFieldInfo)yField).getType();
                if (dateType == LyDateFieldInfo.Type.DATE_HOUR_MIN)
                {
                    if ((yFieldType == LyDateFieldInfo.Type.DATE_TIME)
                            || (yFieldType == LyDateFieldInfo.Type.TIME))
                        dateType = LyDateFieldInfo.Type.DATE_TIME;
                }

                if (dateType == LyDateFieldInfo.Type.DATE)
                {
                    if ((yFieldType == LyDateFieldInfo.Type.DATE_TIME)
                            || (yFieldType == LyDateFieldInfo.Type.TIME))
                        dateType = LyDateFieldInfo.Type.DATE_TIME;
                    else if ((yFieldType == LyDateFieldInfo.Type.DATE_HOUR_MIN)
                            || (yFieldType == LyDateFieldInfo.Type.HOUR_MIN))
                        dateType = LyDateFieldInfo.Type.DATE_HOUR_MIN;
                }
                else if (dateType == LyDateFieldInfo.Type.TIME)
                {
                    if ((yFieldType == LyDateFieldInfo.Type.DATE_TIME)
                            || (yFieldType == LyDateFieldInfo.Type.DATE_HOUR_MIN)
                            || (yFieldType == LyDateFieldInfo.Type.DATE))
                        dateType = LyDateFieldInfo.Type.DATE_TIME;
                }
                else if (dateType == LyDateFieldInfo.Type.HOUR_MIN)
                {
                    if (yFieldType == LyDateFieldInfo.Type.TIME)
                        dateType = LyDateFieldInfo.Type.TIME;
                    else if (yFieldType == LyDateFieldInfo.Type.DATE_TIME)
                        dateType = LyDateFieldInfo.Type.DATE_TIME;
                    else if ((yFieldType == LyDateFieldInfo.Type.DATE_HOUR_MIN)
                            || (yFieldType == LyDateFieldInfo.Type.DATE))
                        dateType = LyDateFieldInfo.Type.DATE_HOUR_MIN;
                }
                else
                    dateType = yFieldType;
            }
        }

        return dateType.toString();
    } // end method makeYDateType

    /**
     * Default constructor.
     *
     * @param  parent  parent controller
     */
    public LyGraphicalChartBuilder(LyController parent)
    {
        super(parent);
    }

    /**
     * Builds the view component for the specified parent controller.
     *
     * @return  the view Component
     */
    public LyComponent build()
    {
        // The chart 4th window is an area which contains the chart
        LyArea area = new LyArea();

        _component = area;

        area.setScrolling(LyArea.Scrolling.AUTO);
        area.setTitleBar(buildTitleBar());
        area.setCommands(buildCommands());
        area.addPanel(buildChart());
        area.setController(_parent.getId());
        area.setToolBar(buildToolBar());
        area.setTopRightToolBar(buildRightToolBar());

        LyConstraint	constraint = new LyConstraint();

        constraint.setFill(LyConstraint.Fill.BOTH);

        area.setConstraint(constraint);

        return (area);
    }

    /**
     * Builds the chart graphic.
     *
     * @return  the component for the chart graphic
     */
    public LyChart buildChart()
    {
        LyGraphicalChartController chartController = (LyGraphicalChartController)_parent;
        String                     chartFile = (String)chartController.getParameter("_chartFile", true);

        // If a file is given for the chart use xml description
        if (chartFile != null)
        {
            LyFileBuilder fileBuilder = new LyFileBuilder(_parent, chartFile, LyChart.class);

            _chart = (LyChart)fileBuilder.build();
        }
        else
            _chart = new LyChart();

        _chart = fillChart();
        _chart.setController(_parent.getId());

        return _chart;
    }

    /**
     * Shortcut to the chart description.
     *
     * @return  the component for the chart description
     */
    public LyChart getChart()
    {
        return _chart;
    }

    /**
     * Fill the LyChart object with LyChartInfo.
     *
     * @return  the chart filled all required parameters
     */
    public LyChart fillChart()
    {
        LyChart chart = _chart;

        chart.setId(_parent.getId() + "_chart");

        chart = fillChartOptions(chart);
        chart = fillChartColumns(chart);
        chart = fillChartIntervals(chart);
        chart = fillChartAxis(chart);

		if (!chart.isNoXField()
				&& (chart.isPieBy() && (chart.getGraphType() == LyChart.GraphType.PIE)))
			chart.setInvertData(true);

        return chart;
    }

    /**
     * Builds the toolbar of the chart view.
     *
     * @return  the toolbar Component
     */
    protected LyToolBar buildToolBar()
    {
        LyAction action = ((LyAreaController)_parent).getAction();

        if (action.hasMark(LyAction.MARK_NO_TOOLBAR))
            return null;

        LyTool        tool = null;
        LyToolBar     toolBar = new LyToolBar();
        LyViewManager viewManager = _parent.getViewManager();

        if(viewManager.hasProperty(LyMobileViewManager.MOBILE_ONLY_PROPERTY))
        	return null;

        toolBar.setId("_toolBar");
        toolBar.setFill(LyToolBar.Fill.HORIZONTAL);
        toolBar.setStyle(LyToolBar.Style.SEPARATOR);
        toolBar.setColor("LY_DASHBOARD_HORIZONTAL_BACKGROUND");

        // Add printing action
        tool = new LyTool(LyGraphicalChartController.DEMAND_PRINT_CHART, "LY_ITEM_PRINT_CHART");
        tool.setImage("LY_ACTION_PRINT");
        tool.setDialog(true);
        toolBar.addTool(tool);

        // Add separator
        toolBar.addSep();

        // Add chart definition save
        tool = new LyTool(LyGraphicalChartController.DEMAND_SAVE_DEF_CHART, "LY_ITEM_SAVE_DEF_CHART");
        tool.setImage("LY_TOOL_SAVE");
        tool.setDialog(true);
        toolBar.addTool(tool);

        if (viewManager.hasProperty("LOCAL_FILE"))
        {
            // Add image save
            tool = new LyTool(LyGraphicalChartController.DEMAND_SAVE_IMG_CHART, "LY_ITEM_SAVE_IMG_CHART");
            tool.setImage("LY_TOOL_SAVE_CHART");
            tool.setDialog(true);
            toolBar.addTool(tool);
        }

        // Add view data action
        toolBar.addSep();

        tool = new LyTool(LyGraphicalChartController.DEMAND_CHART_DATA, "LY_CHART_DATA");
        tool.setImage("LY_ACTION_CHART_DATA");
        tool.setDialog(true);
        toolBar.addTool(tool);

        return completeToolBar(toolBar);
    } // end method buildToolBar

    /**
     * Add information bar into the chart form.
     *
     * @return  the title bar
     */
    protected LyTitleBar buildTitleBar()
    {
        LyAction action = ((LyAreaController)_parent).getAction();

        // FFT00109.
        if ((action == null) || (action.hasMark(LyAction.MARK_NO_TITLEBAR)))
            return null;

        LyTitleBar    titleBar = new LyTitleBar();
        LyEnvironment environment = _parent.getEnvironment();

        if (_parent instanceof LyGraphicalChartController)
        {
            // Default title is the name of the LyChartInfo if there is one,
            LyGraphicalChartController chartController = (LyGraphicalChartController)_parent;
            LyChartInfo                chartInfo = chartController.getChartInfo();
            LyClassInfo                classInfo = chartController.getClassInfo();
            String                     label = chartInfo.getName();

            // The title is translated if it is a property
            if (label != null)
            {
                LyViewManager viewManager = getViewManager();

                if ((viewManager != null) && !viewManager.hasProperty(label))
                    label = environment.translate(label);
            }
            // if null, the title is the name of the action
            else
                label = environment.translate("LY_ACTION_CHART") + " : "
                    + environment.translate(classInfo.getName());

            if (environment.testResource("LY_TITLEBAR_LEFT_IMAGE"))
            {
                titleBar.setImage(_parent.getImage(classInfo));
                titleBar.setImageSensitive(true);
            }

            titleBar.setLeftLabel(label);
                titleBar.setImageSensitive(true);
        }

        return titleBar;
    } // end method buildTitleBar

    /**
     * Creates the command bar and sets its content.
     *
     * @return  a command bar with one button : previous
     */
    protected LyCommands buildCommands()
    {
        LyAction action = ((LyAreaController)_parent).getAction();

        // FFT00109.
        if ((action == null) || (action.hasMark(LyAction.MARK_NO_COMMANDS)) || (getViewManager() != null && getViewManager().hasProperty(LyMobileViewManager.MOBILE_ONLY_PROPERTY)))
            return null;

        LyViewManager manager = _parent.getViewManager();

        // no commands for charts in portlet mode
        if (manager.isPortletMode())
          return null;

        LyCommands    commands = new LyCommands();
        LyCommand     command;

        LyGraphicalChartController chartController = (LyGraphicalChartController)_parent;

        // if there is only one window allowed (like in a web browser) and if there where parameter windows, make a "previous" button.
        if (chartController.getChartController() != null)
        {
            command = new LyCommand(LyChartController.DEMAND_PREVIOUS_STEP, "LY_PREVIOUS");
            command.setAction(LyCommand.Action.SUBMIT);
            commands.addCommand(command);
        }

        // if it is a predefined chart, make a "modify" button.
        if (chartController.getChartController() == null)
        {
            command = new LyCommand(LyChartController.DEMAND_MODIFY_STEP, "LY_MODIFY");
            command.setAction(LyCommand.Action.OTHER);
            commands.addCommand(command);
        }

        if (_parent.hasViewController())
        {
            command = new LyCommand(LyFormController.DEMAND_CANCEL, "LY_CLOSE");
            command.setAction(LyCommand.Action.CLOSE);
            commands.addCommand(command);
        }

        command = buildHelpCommand();

        if (command != null)
            commands.addCommand(command);

        return commands;
    } // end method buildCommands

    /**
     * Update specific legend labels on the chart converter.
     */
    public void updateSpecificBehaviors()
    {
        LyViewComponent internalChart = ((_chart == null) ? null : _chart.getInternalView());

        if (!(internalChart instanceof LyGetChartConverterInterface))
            return;

        LyGraphicalChartController   chartController = (LyGraphicalChartController)_parent;
        LyGetChartConverterInterface getConverter = (LyGetChartConverterInterface)internalChart;
        LyChartConverter             converter = getConverter.getChartConverter();
        LyChartBehaviorInterface     behavior = (LyChartBehaviorInterface)_parent.getBehavior();

        // Set specific legend labels
        converter.setSpecificLegendValue(_specificLegendLabels);

        // Set legend position
        converter.setLegendRightPosition(behavior.isLegendRightPositionned(chartController));

        // Generate sensitive areas if needed
        converter.generateSensitiveAreas(behavior.generateSensitiveAreas(chartController));
    }

    /**
     * Fill the LyChart Columns with LyChartInfo, and defines the type of date variables.
     *
     * @param   chart  the chart to fill
     *
     * @return  the chart filled with columns values
     */
    protected LyChart fillChartColumns(LyChart chart)
    {
        LyEnvironment              environment = _parent.getEnvironment();
        LyGraphicalChartController chartController = (LyGraphicalChartController)_parent;
        LyChartBehaviorInterface   behavior = (LyChartBehaviorInterface)chartController.getBehavior();

        if (_specificLegendLabels != null)
            _specificLegendLabels.clear();

        _specificLegendLabels = new Hashtable<String, String>();

        LyChartInfo chartInfo = chartController.getChartInfo();
        LyClassInfo classInfo = chartController.getClassInfo();
        String      selectedValue = chartInfo.getSelectedXFieldInfo();
        LyFieldInfo selectedX = null;

        if (selectedValue != null)
            selectedX = classInfo.getFieldInfo(selectedValue, false);

        selectedValue = chartInfo.getSelectedGroupFieldInfo();

        LyFieldInfo groupFieldInfo = null;

        if (selectedValue != null)
            groupFieldInfo = classInfo.getFieldInfo(selectedValue, false);

        LyFieldInfo    	currentYFieldInfo;
        Vector<String>	selectedY = chartInfo.getSelectedYFieldsInfo();
        LyObject       	currentObject;
        Object         	xValue;
        Object         	yValue;
        Object         	objectValue = null;
        String         	xString;
        String         	yString;
        Object         	xObject;
        Object        	yObject;
        LyColumn      	col = null;
        short        	n, k;
        int           	p, r;
        boolean        	foundChoice;
        int            	i, nSelectedY = (selectedY == null ? 0 : selectedY.size());
        int            	t, nGroup = 1;
        Vector<String>	groupIds = null, groupLabels = null;
        boolean        	useEmptyGroup = false;

        // If a group field has been selected, retrieve group field values.
        // These values are choice options if choice field or objects pointed
        // by the relation if relation field.
        if (groupFieldInfo instanceof LyChoiceFieldInfo)
        {
            LyChoiceFieldInfo          choice = (LyChoiceFieldInfo)groupFieldInfo;
            LyInfoList<LyChoiceOption> options = choice.getChoiceOptions(true);
            LyChoiceOption             option;
            String                     label;

            nGroup = choice.getChoiceOptionCount();
            groupIds = new Vector<String>(nGroup);
            groupLabels = new Vector<String>(nGroup);

            for (i = 0; i < nGroup; i++)
            {
                option = options.getInfo(i);
                groupIds.addElement(new Integer(i).toString());
                label = option.getName();
                label = environment.translate(label);
                groupLabels.addElement(label);
            }
        }
        else if (groupFieldInfo instanceof LyRelationFieldInfo)
        {
            LyRelationFieldInfo relation = (LyRelationFieldInfo)groupFieldInfo;
            LyObjectList        content = chartController.getContent();
            Iterator<LyObject>	iterator = content.iterator();
            LyObjectList        relationsTmp, relations = new LySimpleObjectList();
            LyObject            object;

            while (iterator.hasNext())
            {
            	object = iterator.next();
                relationsTmp = relation.getObjects(object);
                if (relationsTmp != null)
                    relations.addList(relationsTmp);
            }

            nGroup = relations.getSize();
            groupIds = new Vector<String>(nGroup);
            groupLabels = new Vector<String>(nGroup);

            for (i = 0; i < nGroup; i++)
            {
                object = relations.getObject(i);
                groupIds.addElement(object.getObjectId().getFullString());
                groupLabels.addElement(getName(object));
            }
        }
        else if (groupFieldInfo != null)
        {
            // Retrieve all the group field values
            LyObjectList		content = chartController.getContent();
            Iterator<LyObject>	iterator = content.iterator();
            LyObject    		object;
            Object      		fieldValue;
            String      		strValue;

            groupIds = new Vector<String>();
            groupLabels = new Vector<String>();

            while (iterator.hasNext())
            {
                object = iterator.next();
                fieldValue = classInfo.getValue(object, groupFieldInfo);

                if (fieldValue == null)
                {
                    fieldValue = "";
                    useEmptyGroup = true;
                }

                strValue = fieldValue.toString();

                if (!groupIds.contains(strValue))
                {
                    groupIds.addElement(strValue);
                    groupLabels.addElement(strValue);
                }
            }

            nGroup = groupIds.size();
        }

        LyInfoList<LyChoiceOption> options = null;
        short                      index = -1;
        LyChartInfo.GraphType      chartGraphType = chartInfo.getSelectedGraphType();
        boolean                    multiple = ((chartGraphType != LyChartInfo.GraphType.BAR)
            && (chartGraphType != LyChartInfo.GraphType.PIE));
        LyObjectList			   content = chartController.getContent();
        Iterator<LyObject>  	   iterator;

        // For each group value and each y field create a column.
        // If no group field has been selected there is one column for one y field.
        for (t = 0; t < nGroup; t++)
        {
            for (i = 0; i < nSelectedY; i++)
            {
                currentYFieldInfo = classInfo.getFieldInfo(selectedY.elementAt(i), false);
                if ((i == 0) || multiple)
                {
                    col = new LyColumn();

                    if (groupIds == null)
                    {
                        if ((nSelectedY > 1) || multiple)
                            col.setLabel(environment.translate(currentYFieldInfo.getName()));
                        else
                            col.setLabel("");
                    }
                    else
                    {
                        if (nSelectedY > 1)
                            col.setLabel(environment.translate(currentYFieldInfo.getName())
                                + GROUP_FIELD_SEP + groupLabels.elementAt(t).toString());
                        else
                            col.setLabel(groupLabels.elementAt(t).toString());
                    }
                }

                // If the type of variable is choice : keep the order of the options :
                // for each option search all Y values of the context which equals to this option
                if (currentYFieldInfo instanceof LyChoiceFieldInfo)
                {
                    n = ((LyChoiceFieldInfo)currentYFieldInfo).getChoiceOptionCount();
                    options = ((LyChoiceFieldInfo)currentYFieldInfo).getChoiceOptions(true);
                }
                else
                    n = 1;

                for (k = 0; k < n; k++)
                {
                    // reset foundChoice : if no object correspond to this choice, create a line with x=null and y=this choice.
                    // useful when COUNT is selected, to make a zero appear on the graphic.
                    foundChoice = false;

                    iterator = (content == null ? null : content.iterator());
                    if (iterator != null)
                    {
                        while (iterator.hasNext())
                        {
                            currentObject = iterator.next();

                            if (groupIds != null)
                            {
                                if (groupFieldInfo instanceof LyRelationFieldInfo)
                                {
                                    LyRelationValue relationValue = currentObject.getRelationValue((LyRelationFieldInfo)
                                            groupFieldInfo);

                                    // As group field info are not multiple, retrieve first object id
                                    if ((relationValue != null)
                                            && (relationValue.getValueCount() > 0))
                                        objectValue = relationValue.getObjectId(0).getFullString();
                                    else
                                        objectValue = null;
                                }
                                else
                                    objectValue = classInfo.getValue(currentObject, groupFieldInfo);

                                if ((objectValue == null) && !useEmptyGroup)
                                    continue;

                                if (objectValue == null)
                                    objectValue = "";

                                if (!objectValue.toString().equals(groupIds.elementAt(t).toString()))
                                    continue;
                            }

                            yValue = classInfo.getValue(currentObject, currentYFieldInfo);
                            yString = null;

                            if (currentYFieldInfo instanceof LyChoiceFieldInfo)
                                index = ((LyChoiceFieldInfo)currentYFieldInfo).getIndex(options.getInfo(k));
                            else
                                index = -1;

                            // To keep the right order :
                            // For Choices, search all Y values of the context which equals to the current option
                            // For others, ok (no particular order).
                            if (!(currentYFieldInfo instanceof LyChoiceFieldInfo)
                                    || ((yValue instanceof Short)
                                        && (((Short)yValue).intValue() == index))
                                    || ((yValue instanceof BitSet)
                                        && ((BitSet)yValue).get(index))
                                    || (yValue instanceof Boolean))
                            {
                                foundChoice = true;
                                yString = null;

                                yObject = getStringChartValue(currentYFieldInfo, currentObject, yValue,
                                        chartInfo.getTypeVarY(), true);

                                if (yObject instanceof String)
                                    yString = (String)yObject;
                                else if (index >= 0)
                                    yString = options.getInfo(k).getName();

                                if (selectedX != null)
                                {
                                    xValue = classInfo.getValue(currentObject, selectedX);
                                    xString = null;
                                    xObject = getStringChartValue(selectedX, currentObject, xValue,
                                            chartInfo.getTypeVarX(), false);

                                    if (xObject instanceof String)
                                    {
                                        xString = (String)xObject;
                                        String legendLabel = behavior.getBarLegendLabel(chartController, selectedX,
                                                currentObject, xValue);

                                        if ((xValue != null) && (legendLabel != null))
                                            _specificLegendLabels.put(xString, legendLabel);
                                    }
                                }
                                else
                                {
                                    // if no X field is selected :
                                    // if type of chart is pie, X field is equal to blank
                                    // if type of chart is not pie, X field is equal to Y variable title
                                    if (chartInfo.getSelectedGraphType() == LyChartInfo.GraphType.PIE)
                                        xString = "";
                                    else
                                        xString = col.getLabel();

                                    xObject = xString;
                                }

                                // These loops permit to decompose multiple choices or relations (else the cardinal of the loop is 1).
                                if ((yObject instanceof Vector<?>)
                                        && (yString == null))
                                    p = ((Vector<?>)yObject).size();
                                else
                                    p = 1;

                                // y loop
                                for (int q = 0; q < p; q++)
                                {
                                    if ((yObject instanceof Vector<?>)
                                            && (yString == null))
                                        yString = (String)((Vector<?>)yObject).elementAt(q);

                                    if (xObject instanceof Vector<?>)
                                        r = ((Vector<?>)xObject).size();
                                    else
                                        r = 1;

                                    // x loop
                                    for (int s = 0; s < r; s++)
                                    {
                                        if (xObject instanceof Vector<?>)
                                            xString = (String)((Vector<?>)xObject).elementAt(s);

                                        if ((xString != null) & (yString != null))
                                        {
                                            LyChartValue chartValue = new LyChartValue(yString);

                                            chartValue.setXValue(xString);
                                            col.addChartValue(chartValue);
                                        }
                                    }
                                }
                            }
                        }
                    }

                    // if no object correspond to this choice, create a line with x=null and y=this choice.
                    // usefull when COUNT is selected, to make a zero appear on the graphic.
                    if ((foundChoice == false) && (currentYFieldInfo instanceof LyChoiceFieldInfo))
                    {
                        LyChoiceFieldInfo fieldInfo = (LyChoiceFieldInfo)currentYFieldInfo;

                        if (!fieldInfo.isMultiple())
                        {
                            yValue = (fieldInfo.getChoiceOption(k).getValue());

                            if (fieldInfo.getType().equals(LyChoiceFieldInfo.Type.BOOLEAN))
                            {
                                if (k == 0)
                                    yValue = Boolean.FALSE;
                                else
                                    yValue = Boolean.TRUE;
                            }
                            else
                                yValue = new Short(k);
                        }
                        else
                            yValue = null;

                        Object option = getStringChartValue(fieldInfo, null, yValue, chartInfo.getTypeVarY(), true);

                        if (null != option)
                        {
                            LyChartValue chartValue = new LyChartValue(option.toString());

                            chartValue.setXValue(null);
                            col.addChartValue(chartValue);
                        }
                    }
                }

                chart.insertColumn(col, (nSelectedY * t) + i);
            }
        }

        /* For a scatter plot chart with only one y field, used a point for each object of the chart.
        // Labels may be changed by the behavior.
        if (chartInfo.getSelectedGraphType() == LyChartInfo.GraphType.SCATTER_PLOT)
        {
            int colCount = chart.getColumnCount();

            if (colCount == 1)
            {
                LyColumn column = chart.getColumn(0), newColumn;
                int      valueIndex, valueCount = column.getChartValueCount();

                chart.removeColumn(column);

                for (valueIndex = 0; valueIndex < valueCount; valueIndex++)
                {
                    currentObject = objects.getObject(valueIndex);
                    newColumn = new LyColumn();

                    String label = behavior.getScatterPlotLabel(chartController, currentObject);
                    String legendLabel = behavior.getScatterPlotLegendLabel(chartController, currentObject);

                    if ((label != null) && (legendLabel != null))
                        _specificLegendLabels.put(label, legendLabel);

                    newColumn.setLabel(label);
                    // when adding chartValue, it is removed from source column.
                    // So, the 0 index value is always the interesting one.
                    newColumn.addChartValue(column.getChartValue(0));

                    chart.addColumn(newColumn);
                }
            }
        } */

        // fill the final Y Date Type parameter
        if (chartInfo.getTypeVarY() == LyChartInfo.Type.DATE)
            chart.setYDateType(makeYDateType(selectedY, classInfo, chartInfo));

        // fill the X Date Type parameter
        if ((chartInfo.getTypeVarX() == LyChartInfo.Type.DATE)
                && (selectedX instanceof LyDateFieldInfo))
            chart.setXDateType(((LyDateFieldInfo)selectedX).getType().toString());

        if (selectedY != null)
        {
            // fill the Y Format
            LyFieldInfo yField = classInfo.getFieldInfo(selectedY.elementAt(0), false);

            chart.setYFormat(yField.getFormat());
        }

        // fill the X Format
        if (selectedX != null)
            chart.setXFormat(selectedX.getFormat());

        return chart;
    } // end method fillChartColumns

    /**
     * Make a String or a Vector or Strings (for multiple choices or relations) from a value of an
     * Object. This (these) string(s) will be inserted in LyChart. The format of the dates is
     * mm/dd/yyyy hh:mm:ss
     *
     * @param   fieldInfo  the field info concerned
     * @param   object     the object concerned (may be null)
     * @param   value      the value to transform into String. It comes from the fieldInfo
     * @param   typeVar    the type of the field, like date or numeric (defined in LyChartInfo)
     * @param   yAxis      indicates whether the value is computed for y or x axis
     *
     * @return  a String corresponding to the value, or a Vector of Strings if it is a multiple
     *          choice or relation.
     *
     * @see     LyChartInfo
     */
    protected Object getStringChartValue(LyFieldInfo fieldInfo, LyObject object, Object value, LyChartInfo.Type typeVar,
        boolean yAxis)
    {
        LyChartBehaviorInterface behavior = (LyChartBehaviorInterface)_parent.getBehavior();

        return behavior.getStringChartValue((LyGraphicalChartController)_parent, fieldInfo, object, value, typeVar,
            yAxis);
    }

    /**
     * Make a String from a value of an Object. This string will be inserted in LyChart. The dates
     * are transformated in milliseconds
     *
     * @param   value    the value to transform into String.
     * @param   typeVar  the type of the field, like date or numeric (defined in LyChartInfo)
     *
     * @see     LyChartInfo
     *
     * @return  a String corresponding to the value.
     */
    protected String getStringValue(Object value, LyChartInfo.Type typeVar)
    {
        if (value == null)
            return null;

        String stringValue = null;

        if (typeVar == LyChartInfo.Type.TEXT)
            stringValue = value.toString();
        else if (typeVar == LyChartInfo.Type.NUMBER)
            stringValue = value.toString();
        else if ((typeVar == LyChartInfo.Type.DATE)
                && (value instanceof Date))
        {
            long time = ((Date)value).getTime();

            stringValue = new Long(time).toString();
        }
        else
            stringValue = value.toString();

        return stringValue;
    }

    /**
     * Fill the LyChart Serie of Options with LyChartInfo.
     *
     * @param   chart  the chart to fill
     *
     * @return  the chart filled with option values
     */
    protected LyChart fillChartOptions(LyChart chart)
    {
        // Recover the graph type
        LyGraphicalChartController chartController = (LyGraphicalChartController)_parent;
        LyChartInfo                chartInfo = chartController.getChartInfo();
        LyChartInfo.GraphType      chartGraphType = chartInfo.getSelectedGraphType();
        LyClassInfo                classInfo = chartController.getClassInfo();

        switch (chartGraphType)
        {
        case STACK_BAR:
            chart.setGraphType(LyChart.GraphType.STACK_BAR);
            break;

        case BAR:
            chart.setGraphType(LyChart.GraphType.BAR);
            break;

        case PIE:
            chart.setGraphType(LyChart.GraphType.PIE);
            break;

        case SCATTER_PLOT:
            chart.setGraphType(LyChart.GraphType.SCATTER_PLOT);
            break;

        case PLOT:
            chart.setGraphType(LyChart.GraphType.PLOT);
            break;

        case STACK_AREA:
            chart.setGraphType(LyChart.GraphType.STACK_AREA);
            break;

        default:
            chart.setGraphType(LyChart.GraphType.PLOT);
        }

        // Recover the type of values for X (real, intervals).
        if (chartInfo.getXVal() == LyChartInfo.ValType.REAL)
            chart.setXVal(LyChart.XVal.X_REAL);
        else if (chartInfo.getXVal() == LyChartInfo.ValType.INTERVAL)
            chart.setXVal(LyChart.XVal.X_INTERVAL);

        // Recover the type of values for Y (real, sum, count, count intervals).
        if (chartInfo.getYVal() == LyChartInfo.ValType.REAL)
            chart.setYVal(LyChart.YVal.Y_REAL);
        else if (chartInfo.getYVal() == LyChartInfo.ValType.SUM_REAL)
            chart.setYVal(LyChart.YVal.Y_SUM_REAL);
        else if (chartInfo.getYVal() == LyChartInfo.ValType.COUNT_REAL)
            chart.setYVal(LyChart.YVal.Y_COUNT_REAL);
        else if (chartInfo.getYVal() == LyChartInfo.ValType.COUNT_INTERVAL)
            chart.setYVal(LyChart.YVal.Y_COUNT_INTERVAL);
        else if (chartInfo.getYVal() != null)
            throw new IllegalArgumentException("BF AppEngine : Illegal Y chart intervals type : " + chartInfo.getYVal());

        // Recover the type pies : one pie by X value or one pie by Y variable.
        chart.setPieBy(chartInfo.getPieBy());

        // Recover the titles.
        LyEnvironment environment = _parent.getEnvironment();

        chart.setTitle(environment.translate(chartInfo.getTitle()));
        chart.setXTitle(environment.translate(chartInfo.getXTitle()));
        chart.setYTitle(environment.translate(chartInfo.getYTitle()));

        // Recover the type of values for X and Y : number, text, date.
        if (chartInfo.getTypeVarX() == LyChartInfo.Type.NUMBER)
            chart.setXType(LyChart.Type.CHARTNUMBER);
        else if (chartInfo.getTypeVarX() == LyChartInfo.Type.DATE)
            chart.setXType(LyChart.Type.CHARTDATE);
        else
            chart.setXType(LyChart.Type.CHARTTEXT);

        if (chartInfo.getTypeVarY() == LyChartInfo.Type.NUMBER)
            chart.setYType(LyChart.Type.CHARTNUMBER);
        else if (chartInfo.getTypeVarY() == LyChartInfo.Type.DATE)
            chart.setYType(LyChart.Type.CHARTDATE);
        else
            chart.setYType(LyChart.Type.CHARTTEXT);

        // Recover the type of X labels : horizontal or vertical.
        chart.setXVerticalLabels(chartInfo.getXVerticalLabels());

        // Show/hide X Grid.
        chart.setXGrid(chartInfo.getXGrid());

        // Show/hide Y Grid.
        chart.setYGrid(chartInfo.getYGrid());

        // Retrieve the type of X labels : horizontal or vertical.
        chart.setD3Effect(chartInfo.get3dEffect());

        // Retrieve if the legend has to be showed
        chart.setShowLegend(chartInfo.getShowLegend());

        // Retrieve if the tip labels have all to be showed (and not only when the mouse is over the point)
        chart.setShowTips(chartInfo.getShowTips());

        // Retrieve if the datas have to be transform to percentiles
        chart.setRatios(chartInfo.getRatios());

        // Retrieve if there is no selected "function of" variable
        String xField = chartInfo.getSelectedXFieldInfo();

        if ((xField == null)
                || (classInfo.getFieldInfo(xField, false) == null))
            chart.setNoXField(true);
        else
            chart.setNoXField(false);

        // Check on the behavior if empty values must be displayed on the legend.
        LyChartBehaviorInterface behavior = (LyChartBehaviorInterface)_parent.getBehavior();

        chart.setDisplayLegendEmptyValues(behavior.displayLegendEmptyValues(chartController));

        return chart;
    } // end method fillChartOptions

    /**
     * Fill the LyChart Intervals with LyChartInfo.
     *
     * @param   chart  the chart to fill
     *
     * @return  the chart filled with intervals values
     */
    protected LyChart fillChartIntervals(LyChart chart)
    {
        LyGraphicalChartController chartController = (LyGraphicalChartController)_parent;
        LyChartInfo                chartInfo = chartController.getChartInfo();
        LyClassInfo                classInfo = chartController.getClassInfo();
        String[]                   xIntervals = chartInfo.getXIntervals();
        String[]                   yIntervals = chartInfo.getYIntervals();

        if ((xIntervals != null) || (yIntervals != null))
            chart.setIntervals(new LyIntervals());

        // X intervals
        if ((xIntervals != null)
                && (chartInfo.getXVal() == LyChartInfo.ValType.INTERVAL))
        {
            String xBegin = "0";

            if (xIntervals[0] != null)
            {
                if (chartInfo.getTypeVarX() == LyChartInfo.Type.DATE)
                {
                    // Retrieve the format of the date
                    LyDateFieldInfo dateField = (LyDateFieldInfo)classInfo.getFieldInfo(chartInfo
                                .getSelectedXFieldInfo(), false);

                    if (dateField == null)
                        dateField = (LyDateFieldInfo)classInfo.getFieldInfo(chartInfo.getSelectedYFieldsInfo()
                                    .elementAt(0), false);

                    try
                    {
                        Date date = LyDateFormatter.parse((xIntervals[0]).toString(), dateField.getStringDateFormat());

                        xBegin = getStringValue(date, chartInfo.getTypeVarX());
                    }
                    catch (Exception e) // when parse error, set 0 as the beginning
                    {
                        xBegin = "0";
                    }
                }
                else
                    xBegin = getStringValue(xIntervals[0], chartInfo.getTypeVarX());
            }

            // update x begin of intervals in LyChart
            chart.getIntervals().setXBegin(LyTools.doubleFromString(xBegin));

            String xWidth = "1";

            if ((xIntervals[1] != null)
                    && (LyTools.floatFromString(xIntervals[1]) != 0))
                xWidth = xIntervals[1];

            chart.getIntervals().setXWidth(LyTools.floatFromString(xWidth));

            // if type of values is date, stock the time unit (year, day, hour ...) in LyChart
            if (chartInfo.getTypeVarX() == LyChartInfo.Type.DATE)
            {
                LyIntervals.TimeUnit xTimeUnit = null;

                if (xIntervals[2] != null)
                    xTimeUnit = LyIntervals.getTimeUnit(xIntervals[2]);

                chart.getIntervals().setXTimeUnit(xTimeUnit);
            }
        }

        // Y intervals
        if ((yIntervals != null)
                && (chartInfo.getYVal() == LyChartInfo.ValType.COUNT_INTERVAL))
        {
            String yBegin = "0";

            if (yIntervals[0] != null)
            {
                if (chartInfo.getTypeVarY() == LyChartInfo.Type.DATE)
                {
                    // Retrieve the format of the date
                    LyDateFieldInfo dateField = (LyDateFieldInfo)classInfo.getFieldInfo(chartInfo
                                .getSelectedYFieldsInfo().elementAt(0), false);
                    Date            date = LyDateFormatter.parse(yIntervals[0], dateField.getStringDateFormat());

                    yBegin = getStringValue(date, chartInfo.getTypeVarY());
                }
                else
                    yBegin = getStringValue(yIntervals[0], chartInfo.getTypeVarY());
            }

            // update y begin of intervals in LyChart
            chart.getIntervals().setYBegin(LyTools.doubleFromString(yBegin));

            String yWidth = "1";

            if ((yIntervals[1] != null)
                    && (LyTools.floatFromString(yIntervals[1]) != 0))
                yWidth = yIntervals[1];

            chart.getIntervals().setYWidth(LyTools.floatFromString(yWidth));

            // if type of values is date, stock the time unit (year, day, hour ...) in LyChart
            if (chartInfo.getTypeVarY() == LyChartInfo.Type.DATE)
            {
                LyIntervals.TimeUnit yTimeUnit = null;

                if (yIntervals[2] != null)
                    yTimeUnit = LyIntervals.getTimeUnit(yIntervals[2]);

                chart.getIntervals().setYTimeUnit(yTimeUnit);
            }
        }

        return chart;
    } // end method fillChartIntervals

    /**
     * Fill the LyChart Axis Bounds and intervals with LyChartInfo.
     *
     * @param   chart  the chart to fill
     *
     * @return  the chart filled with axis parameters
     */
    protected LyChart fillChartAxis(LyChart chart)
    {
        LyGraphicalChartController chartController = (LyGraphicalChartController)_parent;
        LyChartInfo                chartInfo = chartController.getChartInfo();
        LyClassInfo                classInfo = chartController.getClassInfo();
        String[]                   xAxis = chartInfo.getXAxis();
        String[]                   yAxis = chartInfo.getYAxis();

        if ((xAxis != null) || (yAxis != null))
            chart.setAxis(new LyAxis());

        // X axis
        if (xAxis != null)
        {
            if (chartInfo.getTypeVarX() == LyChartInfo.Type.DATE)
            {
                // if type of values is date, no width for the axis
                // Retrieve the format of the date
                LyDateFieldInfo dateField = (LyDateFieldInfo)classInfo.getFieldInfo(chartInfo.getSelectedXFieldInfo(),
                        false);

                if (dateField == null)
                    dateField = (LyDateFieldInfo)(classInfo.getFieldInfo(
                                chartInfo.getSelectedYFieldsInfo().elementAt(0)));

                if (xAxis[0] != null)
                {
                    Date   date = LyDateFormatter.parse(xAxis[0], dateField.getStringDateFormat());
                    String xMin = getStringValue(date, chartInfo.getTypeVarX());

                    chart.getAxis().setXBegin(xMin);
                }

                if (xAxis[1] != null)
                {
                    Date   date = LyDateFormatter.parse(xAxis[1], dateField.getStringDateFormat());
                    String xMax = getStringValue(date, chartInfo.getTypeVarX());

                    chart.getAxis().setXEnd(xMax);
                }

                if (xAxis[2] != null)
                {
                    String xWidth = getStringValue(xAxis[2], chartInfo.getTypeVarX());

                    chart.getAxis().setXWidth(xWidth);
                }
            }
            else // numeric values
            {
                if (xAxis[0] != null)
                {
                    String xMin = getStringValue(xAxis[0], chartInfo.getTypeVarX());

                    chart.getAxis().setXBegin(xMin);
                }

                if (xAxis[1] != null)
                {
                    String xMax = getStringValue(xAxis[1], chartInfo.getTypeVarX());

                    chart.getAxis().setXEnd(xMax);
                }

                if ((chartInfo.getTypeVarX() == LyChartInfo.Type.NUMBER)
                        && (xAxis[2] != null))
                {
                    String xWidth = getStringValue(xAxis[2], chartInfo.getTypeVarX());

                    chart.getAxis().setXWidth(xWidth);
                }
            }

        }

        // Y axis
        if (yAxis != null)
        {
            // real date variables
            if ((chartInfo.getTypeVarY() == LyChartInfo.Type.DATE)
                    && (chartInfo.getYVal() != LyChartInfo.ValType.COUNT_REAL)
                    && (chartInfo.getYVal() != LyChartInfo.ValType.COUNT_INTERVAL))
            {
                // if type of values is date, no width for the axis
                // Retrieve the format of the date (based on the first Y variable)
                LyDateFieldInfo dateField = (LyDateFieldInfo)classInfo.getFieldInfo(chartInfo.getSelectedYFieldsInfo()
                            .elementAt(0));

                try
                {
                    if (yAxis[0] != null)
                    {
                        Date   date = LyDateFormatter.parse(yAxis[0], dateField.getStringDateFormat());
                        String yMin = getStringValue(date, chartInfo.getTypeVarY());

                        chart.getAxis().setYBegin(yMin);
                    }

                    if (yAxis[1] != null)
                    {
                        Date   date = LyDateFormatter.parse(yAxis[1], dateField.getStringDateFormat());
                        String yMax = getStringValue(date, chartInfo.getTypeVarY());

                        chart.getAxis().setYEnd(yMax);
                    }
                }
                catch (Exception e) // when Y types of dates are not the same
                {
                    chart.getAxis().setYBegin(null);
                    chart.getAxis().setYEnd(null);
                }
            }
            else // numeric values (numeric variables or variables count)
            {
                if (yAxis[0] != null)
                {
                    String yMin = getStringValue(yAxis[0], chartInfo.getTypeVarY());

                    chart.getAxis().setYBegin(yMin);
                }

                if (yAxis[1] != null)
                {
                    String yMax = getStringValue(yAxis[1], chartInfo.getTypeVarY());

                    chart.getAxis().setYEnd(yMax);
                }

                if ((chartInfo.getTypeVarY() == LyChartInfo.Type.NUMBER)
                        && (yAxis[2] != null))
                {
                    String yWidth = getStringValue(yAxis[2], chartInfo.getTypeVarY());

                    chart.getAxis().setYWidth(yWidth);
                }
            }
        }

        return chart;
    } // end method fillChartAxis
} // end class LyGraphicalChartBuilder
