/**
 * LyImportBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyImportBuilder.java 40687 2012-09-06 15:07:47Z jlsaouli $
 */

package leon.control.builder;

import java.util.*;
import leon.control.*;
import leon.info.*;
import leon.info.daemon.*;
import leon.misc.*;
import leon.view.*;


/**
 * This builder creates a view for the import controller.
 *
 * @see LyImportController
 */
public class LyImportBuilder extends LyAreaBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyImportBuilder.java 40687 2012-09-06 15:07:47Z jlsaouli $";

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyImportBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Builds the view component for the specified parent controller. The component may create the
	 * XML Tree or the internal view.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{

		/* The import view is an area which contains :
		        - A toolbar (actions)
		        - A titleBar (description)
		        - A form component
		*/
		LyArea area = new LyArea();

		area.setScrolling(LyArea.Scrolling.AUTO);
		area.setTitleBar(buildTitleBar());
		area.addPanel(buildImportForm());
		area.setCommands(buildCommands());
		area.setController(_parent.getId());

		_component = area;

		// Update default content
		updateTitleBar();

		return (area);
	}

	/**
	 * Update the titlebar of the form.
	 */
	public void updateTitleBar()
	{
		LyTitleBar titleBar = (LyTitleBar)_component.getComponent(_parent.getId() + "_titleBar"); // XXX

		if (titleBar != null)
		{
			LyClassInfo classInfo = ((LyImportController)_parent).getClassInfo();
			String	    label = "LY_ACTION_IMPORT" + " : " + classInfo.getName();

			if (_parent.getEnvironment().testResource("LY_TITLEBAR_LEFT_IMAGE"))
				titleBar.setImage(_parent.getImage(classInfo));
			titleBar.setLeftLabel(label);
		}
	}

	/**
	 * Build the attributes field
	 * @return the attributes field
	 */
	protected LyFormField buildFieldAttributes()
	{
		LyFormField  field;
		LyCaption    caption;
		LyTip		 tip;
		LyClassInfo  classInfo = ((LyImportController)_parent).getClassInfo();
		LyConstraint constraint;
		BitSet		 fieldMarks = ((LyImportController)_parent).getParameterMarks("_fieldMarks",
				LyBitSet.Type.F_MARKS);

		// **** Add attribute List
		LyRelation   fieldRelation = new LyRelation();
		int			 i, n = classInfo.getFieldInfoCount();
		LyFieldInfo  fieldInfo;
		LyLink		 link;

		// Get fieldinfo list
		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			if ((fieldMarks != null) && !fieldInfo.hasMarks(fieldMarks, false))
				continue;

			// skip import on local fields (except for reverse or join relation fieldInfos)
			if (fieldInfo.isLocal()
					&& (!((fieldInfo instanceof LyRelationFieldInfo)
							&& ((((LyRelationFieldInfo)fieldInfo).getDaemon()
									instanceof LyReverseDaemon)
								|| (((LyRelationFieldInfo)fieldInfo).getDaemon()
									instanceof LyJoinDaemon)))))
				continue;
			// Create relation to the fieldInfo
			link = new LyLink(fieldInfo.getId(), fieldInfo.getName());

			fieldRelation.addLink(link);
		}

		fieldRelation.setMultiple(true);

		field = new LyFormField(LyImportController.SELECT_FIELDINFOS);

		caption = new LyCaption();
		caption.setLabel("LY_IMPORT_FIELD");
		caption.setColon(true);

		constraint = new LyConstraint();
		constraint.setColSpan(LyConstraint.REMAINDER);
		constraint.setRemainder(true);
		constraint.setHAlign(LyConstraint.HAlign.LEFT);
		caption.setConstraint(constraint);

		field.setCaption(caption);
		field.setFieldItem(fieldRelation);
		field.setOptional(true);
		field.setNeedCheck(false);

		tip = new LyTip();
		tip.setLabel("LY_IMPORT_ATT_TIP");
		field.setTip(tip);

		constraint = new LyConstraint();
		constraint.setHAlign(LyConstraint.HAlign.CENTER);
		field.setConstraint(constraint);

		return field;
	} // end method buildFieldAttributes

	/**
	 * Build the attribute sep field 
	 * @param fieldSep the field separator value
	 * @param valueSep the value separator value
	 * @return the field item
	 */
	protected LyFormField buildFieldAttributeSep(String fieldSep,
		String valueSep)
	{
		LyFormField  field;
		LyCaption    caption;
		LyTip		 tip;
		LyConstraint constraint;
		LyText		 fieldText;

		fieldText = new LyText();
		fieldText.setLabel(LyImportController.SPREADSHEET_FIELD_SEP);
		fieldText.setNCols(1);
		fieldText.setNChars(1);
		field = new LyFormField(LyImportController.SELECT_FIELD_SEP);

		if (valueSep == null)
		{
			constraint = new LyConstraint();

			constraint.setColSpan(1);
			constraint.setRemainder(false);
			constraint.setFill(LyConstraint.Fill.NONE);

			field.setConstraint(constraint);
		}

		caption = new LyCaption();
		caption.setLabel("LY_FIELD_SEP");
		caption.setColon(true);

		field.setCaption(caption);
		field.setFieldItem(fieldText);
		field.setOptional(false);
		field.setNeedCheck(true);

		tip = new LyTip();
		tip.setLabel("LY_IMPORT_FIELD_SEP_TIP");
		field.setTip(tip);

		return field;
	} // end method buildFieldAttributeSep

	/**
	 * Build the attribute value separator field
	 * @param fieldSep the field sep value
	 * @param valueSep the value sep value
	 * @return the field item
	 */
	protected LyFormField buildFieldAttributeValSep(String fieldSep,
		String valueSep)
	{
		LyFormField field;
		LyCaption   caption;
		LyTip	    tip;
		LyText	    fieldText;

		fieldText = new LyText();
		field = new LyFormField(LyImportController.SELECT_VALUE_SEP);

		fieldText.setLabel(LyImportController.SPREADSHEET_VALUE_SEP);
		fieldText.setNCols(1);
		fieldText.setNChars(1);
		caption = new LyCaption();
		caption.setLabel("LY_VALUE_SEP");
		caption.setColon(true);

		field.setCaption(caption);
		field.setFieldItem(fieldText);
		field.setOptional(false);
		field.setNeedCheck(true);

		tip = new LyTip();
		tip.setLabel("LY_IMPORT_VALUE_SEP_TIP");
		field.setTip(tip);

		return field;
	} // end method buildFieldAttributeValSep

	/**
	 * Build the start line number field
	 * @return the field item
	 */
	protected LyFormField buildFieldStartLineNumber()
	{
		LyNumber fieldNumber = new LyNumber();

		fieldNumber.setVMin(1);
		fieldNumber.setPitch(1);
		fieldNumber.setLabel("1");
		fieldNumber.setDecimals(0);
		LyFormField field = new LyFormField(LyImportController.SELECT_START_LINE_NUMBER);

		LyCaption   caption = new LyCaption();

		caption.setLabel("LY_IMPORT_START_LINE");
		caption.setColon(true);

		field.setCaption(caption);
		field.setFieldItem(fieldNumber);
		field.setOptional(false);
		field.setNeedCheck(false);

		LyTip tip = new LyTip();

		tip.setLabel("LY_IMPORT_START_LINE_TIP");
		field.setTip(tip);

		return field;
	}

	/**
	 * Build the import type field
	 * @param importType the import type
	 * @return the field item
	 */
	protected LyFormField buildFieldImportType(String importType)
	{
		LyFormField  field = new LyFormField(LyImportController.SELECT_IMPORT_TYPE);
		LyCaption    caption;
		LyTip		 tip;

		LyConstraint constraint;

		LyChoice     fieldChoice = new LyChoice();
		LyOption     fieldOption;

		caption = new LyCaption();
		caption.setLabel("LY_IMPORT_TYPE");
		caption.setColon(true);

		fieldChoice.setMultiple(false);
		fieldChoice.setFill(LyChoice.Fill.VERTICAL);

		// FFT00020 : adding spreadsheet import mode
		fieldOption = new LyOption(LyImportController.SELECT_IMPORT_TYPE_SPREADSHEET,
				"LY_SELECT_IMPORT_TYPE_SPREADSHEET");
		fieldOption.setChecked(true);
		fieldChoice.addOption(fieldOption);

//        fieldOption = new LyOption(LyImportController.SELECT_IMPORT_TYPE_TEXT,
//                "LY_SELECT_IMPORT_TYPE_TEXT");
//        fieldChoice.addOption(fieldOption);

		fieldOption = new LyOption(LyImportController.SELECT_IMPORT_TYPE_RAW,
				"LY_SELECT_IMPORT_TYPE_RAW");
		fieldChoice.addOption(fieldOption);

		fieldOption = new LyOption(LyImportController.SELECT_IMPORT_TYPE_XML,
				"LY_SELECT_IMPORT_TYPE_XML");
		fieldChoice.addOption(fieldOption);

		field.setCaption(caption);
		field.setFieldItem(fieldChoice);
		field.setOptional(false);
		field.setNeedCheck(true);

		tip = new LyTip();
		tip.setLabel("LY_IMPORT_TYPE_TIP");
		field.setTip(tip);

		constraint = new LyConstraint();

		//constraint.setRowSpan(2);
		field.setConstraint(constraint);
		return field;
	} // end method buildFieldImportType

	/**
	 * Builds the form field used to display operations to be performed.
	 * @return a form field
	 * @deprecated No longer used
	 */
	protected LyFormField buildFieldOperations()
	{
		return null;
	}

	/**
	 * Get the attribute separator
	 * @param controller the parent controller
	 * @return the attribute separator
	 */
	protected String getAttributeSep(LyImportController controller)
	{
		return (String)controller.getParameterValue(LyImportController.SELECT_FIELD_SEP);
	}

	/**
	 * Get the attribute value separator
	 * @param controller the parent controller
	 * @return the attribute value separator
	 */
	protected String getAttributeValSep(LyImportController controller)
	{
		return (String)controller.getParameterValue(LyImportController.SELECT_VALUE_SEP);
	}

	/**
	 * Get the import type
	 * @param controller the parent controller
	 * @return the import type
	 */
	protected String getImportType(LyImportController controller)
	{
		return (String)controller.getParameterValue(LyImportController.SELECT_IMPORT_TYPE);
	}

	/**
	 * Get the available operation
	 * @param controller the parent controller
	 * @return the available operations
	 */
	protected BitSet getOperations(LyImportController controller)
	{
		return (BitSet)controller.getParameterValue(LyImportController.SELECT_OPERATIONS);
	}

	/**
	 * Get the import file
	 * @param controller parent controller
	 * @return the import file
	 */
	protected String getFile(LyImportController controller)
	{
		return (String)controller.getParameterValue(LyImportController.SELECT_FILE);
	}

	/**
	 * get the start line number
	 * @param controller the parent controller
	 * @return the start line number
	 */
	protected Integer getStartLineNumber(LyImportController controller)
	{
		return (Integer)controller.getParameterValue(LyImportController.SELECT_START_LINE_NUMBER);
	}

	/**
	 * get all attributes
	 * @param controller parent controller
	 * @return a list of attributes
	 */
	protected LyInfoList<LyFieldInfo> getAttributes(LyImportController controller)
	{
		@SuppressWarnings("unchecked")
		LyInfoList<LyFieldInfo> temp = (LyInfoList<LyFieldInfo>)controller.getParameterValue(
				LyImportController.SELECT_FIELDINFOS);
		return temp;
	}

	/**
	 * Get the bottom form
	 * @return the bottom form
	 */
	protected LyForm getBottomForm()
	{
		LyForm			   bottomForm = new LyForm(_parent.getId());
		LyFormField		   field;

		// Add attribute List
		LyImportController controller = (LyImportController)_parent;

		// add the field separator text to the field
		String			   fieldSep = getAttributeSep(controller);
		String			   valueSep = getAttributeValSep(controller);

		if (fieldSep == null)
		{
			field = buildFieldAttributeSep(fieldSep, valueSep);
			if (field != null)
				bottomForm.addFormField(field);
		}

		bottomForm.setBorder(false);

		// add the value separator text to the field
		if (valueSep == null)
		{
			field = buildFieldAttributeValSep(fieldSep, valueSep);
			if (field != null)
				bottomForm.addFormField(field);
		}

		// add the import type choice
		String importType = getImportType(controller);

		if (importType == null)
		{
			field = buildFieldImportType(importType);
			if (field != null)
				bottomForm.addFormField(field);
		}

		// add the file text
		String file = getFile(controller);

		if (file == null)
		{
			field = buildFieldFile(file);
			if (field != null)
				bottomForm.addFormField(field);
		}

		// add the start line number
		Integer line = getStartLineNumber(controller);

		if (line == null)
		{
			field = buildFieldStartLineNumber();
			if (field != null)
				bottomForm.addFormField(field);
		}

		return bottomForm;
	} // end method getBottomForm

	/**
	 * Build the bottom area
	 * @param bottomForm the bottom form
	 * @return the bottom area
	 */
	protected LyArea getBottomArea(LyForm bottomForm)
	{
		LyArea bottomArea = new LyArea();

		bottomArea.setBorder(LyArea.Border.IN);
		bottomArea.setLabel("LY_IMPORT_GROUP_PARAMETERS");
		bottomArea.addPanel(bottomForm);
		return bottomArea;
	}

	/**
	 * Build the top form
	 * @return the top form
	 */
	protected LyForm getTopForm()
	{
		LyForm	    topForm = new LyForm(_parent.getId());
		LyFormField field = buildFieldAttributes();

		if (field != null)
			topForm.addFormField(field);
		topForm.setBorder(false);
		return topForm;
	}

	/**
	 * Build the top area
	 * @return the top area
	 */
	protected LyArea getTopArea()
	{
		LyArea			   topArea = null;

		// Add attribute List
		LyImportController controller = (LyImportController)_parent;
		LyInfoList<LyFieldInfo>		   fields = getAttributes(controller);

		if (fields == null)
		{
			LyForm topForm = getTopForm();

			topArea = new LyArea();
			topArea.setBorder(LyArea.Border.IN);
			topArea.setLabel("LY_IMPORT_GROUP_FIELDS");
			topArea.addPanel(topForm);
		}
		return topArea;
	}

	/**
	 * Build the file field
	 * @param file the file
	 * @return the file field
	 */
	protected LyFormField buildFieldFile(String file)
	{
		LyFormField   field = new LyFormField(LyImportController.SELECT_FILE);
		LyCaption     caption;
		LyTip		  tip;
		LyFile		  fieldFile = new LyFile();

		LyViewManager viewManager = getViewManager();

		if (viewManager.hasProperty("MANAGE_ATTACHMENT"))
			fieldFile.setAttachmentHandler("default");

		caption = new LyCaption();
		caption.setLabel("LY_FILE");
		caption.setColon(true);

		field.setCaption(caption);
		field.setFieldItem(fieldFile);
		field.setOptional(false);
		field.setNeedCheck(false);

		tip = new LyTip();
		tip.setLabel("LY_IMPORT_FILE_TIP");
		field.setTip(tip);

		return field;
	}

	/**
	 * Creates the import form and sets its content.
	 *
	 * @return the area component
	 */
	protected LyArea buildImportForm()
	{
		LyForm bottomForm = getBottomForm();
		LyArea area = new LyArea();
		LyArea topArea = getTopArea();

		if (topArea != null)
			area.addPanel(topArea);

		if (bottomForm.getFormFieldCount() > 0)
		{
			LyArea bottomArea = getBottomArea(bottomForm);

			if (bottomArea != null)
				area.addPanel(bottomArea);
		}

		//** Import area is ready
		return (area);
	} // end method buildImportForm

	/**
	 * Creates the command bar and sets its content.
	 *
	 * @return the commands component
	 */
	protected LyCommands buildCommands()
	{
		LyEnvironment environment = _parent.getEnvironment();
		LyCommands commands = new LyCommands();
		LyCommand  command;

		command = new LyCommand(LyFormController.DEMAND_VALIDATE, "LY_VALIDATE");
		command.setAction(LyCommand.Action.SUBMIT);
		
		if (environment.checkImage("LY_COMMAND_VALIDATE"))
			command.setImage("LY_COMMAND_VALIDATE");
		
		commands.addCommand(command);

		command = new LyCommand(LyFormController.DEMAND_CANCEL, "LY_CANCEL");
		command.setAction(LyCommand.Action.CANCEL);
		
		if (environment.checkImage("LY_COMMAND_CANCEL"))
			command.setImage("LY_COMMAND_CANCEL");
		
		commands.addCommand(command);

		command = buildHelpCommand();

		if (command != null)
			commands.addCommand(command);

		return commands;
	}
} // end class LyImportBuilder
