/**
  * LyMailBuilder.java
  * Copyright (c) 2000-2009, Lyria-W4.
  * All rights reserved.
  */
  
package leon.control.builder;

import leon.app.behavior.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.view.*;


/**
 * This class implements a builder for the mail controller. This builder creates the needed form
 * fields if the controller does not have their values. If the values are already given in the
 * meta-model, the fields will not be created, so a meta-model with all the values already completed
 * will be created as an empty frame with this builder. It is for the controller to test this
 * possibility and don't show the input frame in this case in the {@link LyController#showView()}
 * method.
 *
 * @author  Lyria S.A.
 * @version $Id: LyMailBuilder.java 30940 2011-05-11 15:09:35Z nhamon $
 */
public class LyMailBuilder extends LyAreaBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyMailBuilder.java 30940 2011-05-11 15:09:35Z nhamon $";

	/** Indicates that a preview area must be added to the frame. */
	protected boolean		   _showPreview;

	/**
	 * Default controller.
	 *
	 * @param parent parent controller
	 */
	public LyMailBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * @see leon.control.builder.LyBuilder#build()
	 */
	public LyComponent build()
	{
		LyArea area = new LyArea();

		area.setTitleBar(buildTitleBar());
		
		LyArea	subArea = new LyArea();
		
		subArea.setScrolling(LyArea.Scrolling.AUTO);
		subArea.addPanel(buildMailForm());

		if (_showPreview)
			subArea.addPanel(buildPreviewForm());
		
		area.addPanel(subArea);
		area.setCommands(buildCommands());
		area.setController(_parent.getId());
		area.setScrolling(LyArea.Scrolling.AUTO);

		_component = area;

		return (area);
	}

	/**
	 * @see leon.control.builder.LyAreaBuilder#buildTitleBar()
	 */
	protected LyTitleBar buildTitleBar()
	{
		LyAction action = ((LyMailController)_parent).getAction();

		if (action.hasMark(LyAction.MARK_NO_TITLEBAR))
			return null;

		LyTitleBar  titleBar = new LyTitleBar();
		LyClassInfo classInfo = ((LyMailController)_parent).getClassInfo();
		String	    label = "LY_ACTION_MAIL";
		
		if  (classInfo != null)
		{
		    label += " : " + classInfo.getName();
		    if (_parent.getEnvironment().testResource("LY_TITLEBAR_LEFT_IMAGE"))
		        titleBar.setImage(_parent.getImage(classInfo));
		}

		titleBar.setLeftLabel(label);
		return titleBar;
	}

	/**
	 * Creates an area that shows the email message that will be sent to all the users.
	 *
	 * @return an area with non-editable fields filled with the generated content
	 */
	protected LyArea buildPreviewForm()
	{
		String			 label;
		LyArea			 area;
		LyArea			 bottomArea;
		LyForm			 bottomForm;
		LyFormField		 field;
		LyText			 fieldText;
		LyCaption		 caption;
		LyMailController controller = (LyMailController)_parent;

		bottomForm = new LyForm(controller.getId());
		bottomForm.setBorder(false);

		// Creates a consultation fields for the email preview
		field = new LyFormField(LyMailController.MAIL_PREVIEW_SENDER);
		fieldText = new LyText();
		label = controller.getGeneratedSender();
		fieldText.setLabel(label);
		fieldText.setFieldValue(label);

		caption = new LyCaption();
		caption.setLabel("LY_MAIL_SENDER");
		caption.setColon(true);

		field.setCaption(caption);
		field.setFieldItem(fieldText);
		field.setEditable(false);

		bottomForm.addFormField(field);

		field = new LyFormField(LyMailController.MAIL_PREVIEW_SUBJECT);
		fieldText = new LyText();
		label = controller.getGeneratedSubject();
		fieldText.setLabel(label);
		fieldText.setFieldValue(label);

		caption = new LyCaption();
		caption.setLabel("LY_MAIL_SUBJECT");
		caption.setColon(true);

		field.setCaption(caption);
		field.setFieldItem(fieldText);
		field.setEditable(false);

		bottomForm.addFormField(field);

		field = new LyFormField(LyMailController.MAIL_PREVIEW_CONTENT);
		fieldText = new LyText();
		fieldText.setWrap(true);
		fieldText.setNCols(60);
		fieldText.setNRows(10);
		label = controller.getGeneratedRawContent();
		fieldText.setLabel(label);
		fieldText.setFieldValue(label);

		caption = new LyCaption();
		caption.setLabel("LY_MAIL_RAW_MESSAGE");
		caption.setColon(true);

		field.setCaption(caption);
		field.setFieldItem(fieldText);
		field.setEditable(false);

		bottomForm.addFormField(field);

		area = new LyArea();

		if (bottomForm.getFormFieldCount() > 0)
		{
			bottomArea = new LyArea();

			bottomArea.setBorder(LyArea.Border.IN);
			bottomArea.setLabel("LY_MAIL_PREVIEW");
			bottomArea.addPanel(bottomForm);

			area.addPanel(bottomArea);
		}

		return area;
	} // end method buildPreviewForm

	/**
	 * Creates the area for all the form fields. If the values already exists on the meta-model, the
	 * associated form field is not created.
	 *
	 * @return the area with the needed form fields
	 */
	protected LyArea buildMailForm()
	{
		LyArea			 area;
		LyArea			 centerArea;
		LyForm			 centerForm;
		LyText			 fieldText;
		LyFile			 fieldFile;
		LyRelation		 fieldRelation;
		LyCaption		 caption;
		LyFormField		 field;
		LyLink			 link;
		LyValue			 mailValue;
		LyMailController controller = (LyMailController)_parent;
		LyObject		 currentUser = controller.getSession().getUser();

		String			 fromUserName = (String)controller.getParameter(
				LyMailController.FROM_USER_NAME);
		String			 fromMailAddress = (String)controller.getParameter(
				LyMailController.FROM_MAIL_ADDRESS);
		String			 sendSelected = (String)controller.getParameter(
				LyMailController.SEND_SELECTED);

		String			 mailSubject = (String)controller.getParameter(
				LyMailController.MAIL_SUBJECT);
		String			 attachedFile = (String)controller.getParameter(
				LyMailController.ATTACHED_FILE);
		String			 rawMessage = (String)controller.getParameter(LyMailController.RAW_MESSAGE);
		String			 htmlContent = (String)controller.getParameter(
				LyMailController.HTML_MESSAGE);

		LyClassInfo		 recipientUserClassInfo = (LyClassInfo)controller.getParameter(
				LyMailController.RECIPIENT_USER_CLASS);

		_showPreview = "true".equals(sendSelected);

		centerForm = new LyForm(controller.getId());
		centerForm.setBorder(false);

		// Sender user name
		if (fromUserName == null)
		{
			field = new LyFormField(LyMailController.FROM_USER_NAME);
			fieldText = new LyText();
			fieldText.setNCols(30);

			if (currentUser != null)
			{
				fieldText.setLabel(currentUser.getName());
				fieldText.setFieldValue(currentUser.getName());
			}

			caption = new LyCaption();
			caption.setLabel("LY_MAIL_FROM_USER_NAME");
			caption.setColon(true);

			field.setCaption(caption);
			field.setFieldItem(fieldText);

			centerForm.addFormField(field);
		}

		// Sender mail address
		if (fromMailAddress == null)
		{
			field = new LyFormField(LyMailController.FROM_MAIL_ADDRESS);
			fieldText = new LyText();

			/*
			 * Following SMTP IETF draft, the local part of a mail address can have a maximum of 64
			 * characters and the domain is up to 255, plus the @ separator.
			 */
			fieldText.setNChars(320);
			fieldText.setNCols(40);

			if (currentUser != null)
			{
				// Finding, through the associated values of the current user if any of these can
				// be used as an email address, and use it if available.
				for (int i = 0; i < currentUser.getValues().getSize(); i++)
				{
					mailValue = currentUser.getValues().getFieldValue(i);
					if (mailValue.getValue() != null)
					{
						String mail = mailValue.getString();

						if ((mail.indexOf('@') >= 0)
								&& (LyMailBehavior.indexOfAny(mail, " \t\n\r()<>,;:\\\"[]") == -1))
						{
							fieldText.setLabel(mail);
							fieldText.setFieldValue(mail);
							break;
						}
					}
				}
			}

			caption = new LyCaption();
			caption.setLabel("LY_MAIL_FROM_MAIL_ADDRESS");
			caption.setColon(true);

			field.setCaption(caption);
			field.setFieldItem(fieldText);

			centerForm.addFormField(field);
		}

		if (_showPreview && (recipientUserClassInfo != null))
		{
			field = new LyFormField(LyMailController.SELECTED_RECIPIENTS);
			fieldRelation = new LyRelation();

			fieldRelation.setMultiple(true);

			LyObjectList userList = new LyClassWorkSpace(controller.getSession(),
					recipientUserClassInfo, false);

			for (int i = 0; i < userList.getSize(); i++)
			{
				LyObject user = userList.getObject(i);

				link = new LyLink(user.getId(), user.getName());

				fieldRelation.addLink(link);
			}

			caption = new LyCaption();
			caption.setLabel("LY_MAIL_SELECTED_RECIPIENTS");
			caption.setColon(true);

			field.setCaption(caption);
			field.setFieldItem(fieldRelation);
			field.setOptional(false);

			centerForm.addFormField(field);
		}

		if (mailSubject == null)
		{
			field = new LyFormField(LyMailController.MAIL_SUBJECT);
			fieldText = new LyText();
			fieldText.setNChars(2048);
			fieldText.setNCols(40);

			caption = new LyCaption();
			caption.setLabel("LY_MAIL_SUBJECT");
			caption.setColon(true);

			field.setCaption(caption);
			field.setFieldItem(fieldText);

			field.setHidden(_showPreview);

			centerForm.addFormField(field);
		}

		if (attachedFile == null)
		{
			field = new LyFormField(LyMailController.ATTACHED_FILE);
			fieldFile = new LyFile();
			fieldFile.setFilter("*.*");

			caption = new LyCaption();
			caption.setLabel("LY_MAIL_ATTACHED_FILE");
			caption.setColon(true);

			field.setCaption(caption);
			field.setFieldItem(fieldFile);

			field.setHidden(_showPreview);

			centerForm.addFormField(field);
		}

		if (rawMessage == null)
		{
			field = new LyFormField(LyMailController.RAW_MESSAGE);
			fieldText = new LyText();
			fieldText.setWrap(true);
			fieldText.setNCols(60);
			fieldText.setNRows(5);

			caption = new LyCaption();
			caption.setLabel("LY_MAIL_RAW_MESSAGE");
			caption.setColon(true);

			field.setCaption(caption);
			field.setFieldItem(fieldText);

			field.setHidden(_showPreview);

			centerForm.addFormField(field);
		}

		if (htmlContent == null)
		{
			field = new LyFormField(LyMailController.HTML_MESSAGE);
			fieldFile = new LyFile();
			fieldFile.setFilter("*.html *.htm");

			caption = new LyCaption();
			caption.setLabel("LY_MAIL_HTML_MESSAGE");
			caption.setColon(true);

			field.setCaption(caption);
			field.setFieldItem(fieldFile);

			field.setHidden(_showPreview);

			centerForm.addFormField(field);
		}

		area = new LyArea();

		if (centerForm.getFormFieldCount() > 0)
		{
			centerArea = new LyArea();

			centerArea.setBorder(LyArea.Border.IN);
			centerArea.setLabel("LY_MAIL_PARAMETERS");
			centerArea.addPanel(centerForm);

			area.addPanel(centerArea);
		}

		return area;
	} // end method buildMailForm

	/**
	 * Creates the command bar and sets its content.
	 *
	 * @return the command bar component
	 */
	protected LyCommands buildCommands()
	{
		LyCommands commands = new LyCommands();
		LyCommand  command;

		command = new LyCommand(LyFormController.DEMAND_VALIDATE, "LY_VALIDATE");
		command.setAction(LyCommand.Action.SUBMIT);
		commands.addCommand(command);

		command = new LyCommand(LyFormController.DEMAND_CANCEL, "LY_CANCEL");
		command.setAction(LyCommand.Action.CANCEL);
		commands.addCommand(command);

		command = buildHelpCommand();

		if (command != null)
			commands.addCommand(command);

		return commands;
	}
} // end class LyMailBuilder
