/**
 * LyMultiAreaBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyMultiAreaBuilder.java 35018 2012-03-23 15:17:40Z jlsaouli $
 */

package leon.control.builder;

import java.util.*;
import leon.control.*;
import leon.info.*;
import leon.view.*;


/**
 * This builder construct the view for a multi Area Controller.
 *
 * @see LyMultiAreaController
 */
public abstract class LyMultiAreaBuilder extends LyAreaBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyMultiAreaBuilder.java 35018 2012-03-23 15:17:40Z jlsaouli $";

	/** Displayable delimiter. */
	public static char		   VISIBLE_SEP = '|';

	/** Group string */
	public static String	   GROUP = "GROUP_";

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyMultiAreaBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * For a given menu, fill the tree with the correct information for this menu all its items and
	 * call recursively on each submenu.
	 *
	 * @param menu     the current menu
	 * @param tree     the storage tree for the menus
	 * @param parentId the identification string of the parent menu
	 * @param group    the current group for this menu
	 * @param name     the alternative name for the menu label
	 */
	protected void computeMenu(LyMenu menu, _Tree tree, String parentId, int group, String name)
	{
		// First add the menu to the tree;

		// If parent is root, ignore parent id
		if (!parentId.equals(""))
			parentId += VISIBLE_SEP;

		// Compute the menu identification string
		String menuPath = parentId + GROUP + group + VISIBLE_SEP + menu.getLabel();

		// Generic case
		if (menu.getParent() instanceof LyMenu)
			tree.addElement(menuPath, menu, name);
		// first menu in the menubar case
		else
		{
			// if the node exist, continue
			_Node node = tree.findElement(menuPath);

			if (node == null)
				tree.addElement(menuPath, menu, name);
		}

		_Node  node = tree.findElement(menuPath);

		// Get each sons and compute Group count
		int    count;

		// The group in the current menu
		int    currentGroup = 0;
		// Get all component
		String label;

		for (int i = 0; i < menu.getComponentCount(); i++)
		{
			LyComponent comp = menu.getComponent(i);

			// Separator, increments group count
			if (comp == null) // Sep
			{
				currentGroup++;
				node.setGroupsCount(currentGroup);
			}
			// item, compute item id and add it to the tree
			else if (comp instanceof LyItem)
			{
				LyItem  item = (LyItem)comp;
                boolean found = false;
                _Node   tempNode;
                
				// Before adding the element try to find it with a different group number
				count = node.getGroupsCount();
				label = item.getLabel();
				if (label == null) // Sep
	            {
	                currentGroup++;
	                node.setGroupsCount(currentGroup);
	                continue;
	            }

				label = menuPath + VISIBLE_SEP + GROUP + currentGroup + VISIBLE_SEP + label;

				for (int j = 0; j <= count; j++)
				{
					label = menuPath + VISIBLE_SEP + GROUP + j + VISIBLE_SEP + item.getLabel();
					tempNode = tree.findElement(label);

					if (tempNode != null)
					{
						found = true;
						break;
					}
				}
				if (!found)
					label = menuPath + VISIBLE_SEP + GROUP + currentGroup + VISIBLE_SEP
						+ item.getLabel();

				tree.addElement(label, item, name);
			}
			// menu, call recursively computeMenu
			else if (comp instanceof LyMenu)
				computeMenu((LyMenu)comp, tree, menuPath, currentGroup, name);
		}
	} // end method computeMenu

	/**
	 * Fill a menu with the elements in the current node.
	 *
	 * @param menu        the menu to fill
	 * @param currentNode the currentNode
	 */
	protected void fillMenu(LyMenu menu, _Node currentNode)
	{
		// A sub menu consists of one or more group and sons of the group
		_Node	    group;
		_Node	    element;
		LyMenu	    subMenu;
		Object[]    datas;
		String	    label;
		LyComponent comp;

		for (int i = 0; i < currentNode.getSonsCount(); i++)
		{
			// If it's not the first group, add separator
			if (i > 0)
				menu.addSep();

			group = currentNode.getSonAt(i);

			// for each element in the group
			for (int j = 0; j < group.getSonsCount(); j++)
			{
				element = group.getSonAt(j);

				// If the current element is multiple, we must insert a new menu
				// and rename the current one
				if (element.isMultiple())
				{
					// Create a fake subMenu with the element label
					subMenu = new LyMenu(element.getLabel());
					subMenu.setController(_parent.getId());

					for (int k = 0; k < element.getDataSize(); k++)
					{
						datas = element.getDataAt(k);
						label = (String)datas[1];
						comp = (LyComponent)datas[0];

						if (comp instanceof LyItem)
						{
							((LyItem)comp).setLabel(label);
							//((LyItem)comp).setShortcut(null);
							subMenu.addItem((LyItem)comp);
						}
						else if (comp instanceof LyMenu)
						{
							((LyMenu)comp).setLabel(label);
							subMenu.addMenu((LyMenu)comp);
						}
					}
					menu.addMenu(subMenu);
				}
				else
				{
					datas = element.getDataAt(0);
					label = (String)datas[1];
					comp = (LyComponent)datas[0];
					if (comp instanceof LyItem)
						menu.addItem((LyItem)comp);
					else if (comp instanceof LyMenu)
					{
						//menu.addMenu((LyMenu)comp);
						LyMenu newMenu = new LyMenu(((LyMenu)comp).getLabel());

						newMenu.setId(((LyMenu)comp).getId());
						newMenu.setController(((LyMenu)comp).getController());
						menu.addMenu(newMenu);
						fillMenu(newMenu, element);
					}
				}
			}
		}
	} // end method fillMenu

	/**
	 * This method is invoked to enable the area to complete the menuBar of the view and add its own
	 * menu items. Default implementation invokes the method on the builder instance.
	 *
	 * @param menuBar the menuBar structure
	 * @param level   : level of current Controller. one of the LEVEL_XXX contants. in
	 *                LyViewController
	 * @see   LyViewController
	 */
	public void completeMenuBar(LyMenuBar menuBar, int level)
	{
		LyMultiAreaController multiController = (LyMultiAreaController)_parent;
		LyAction			  action = multiController.getAction();

		if (action.hasMark(LyAction.MARK_NO_MENUBAR))
			return;

		// Add menus for the areas and store the list of existing menus
		LyAreaController		   areaController;
		int						   i, j, n;
		LyMenuBar				   newMenuBar;
		LyMenu					   menu;
		String					   label;
		HashMap<String, LyMenu>    menus = new HashMap<String, LyMenu>();
		HashMap<String, ArrayList<Object>> subMenus = new HashMap<String, ArrayList<Object>>();

		n = menuBar.getMenuCount();

		for (i = 0; i < n; i++)
		{
			menu = menuBar.getMenu(i);
			menus.put(menu.getLabel(), menu);
			subMenus.put(menu.getLabel(), new ArrayList<Object>());
		}

		// Search each sub-controllers that are not compound controllers.
		// in this way, the menu will offer a "flatten" selection of views
		// appearing in this compound view.
		LyMultiAreaController			  startingController;
		LyCompoundAreaController		  compoundAreaController;
		LyTabAreaController				  tabAreaController;
		LyDashboardController			  dashboardController;
		Vector<LyMenuBar>				  menuBarList;
		LinkedList<LyMultiAreaController> compoundControllers =
			new LinkedList<LyMultiAreaController>();
		ArrayList<LyAreaController>		  subControllers = new ArrayList<LyAreaController>();

		compoundControllers.addLast(multiController);

		while (!compoundControllers.isEmpty())
		{
			startingController = compoundControllers.removeFirst();

			if (startingController instanceof LyCompoundAreaController)
			{
				compoundAreaController = (LyCompoundAreaController)startingController;
				n = compoundAreaController.getAreaControllersCount();

				for (i = 0; i < n; i++)
				{
					areaController = compoundAreaController.getAreaController(i);

					if (areaController instanceof LyMultiAreaController)
					{
						compoundControllers.addLast((LyMultiAreaController)areaController);

						// Add this controller to the list of sub-controllers as it may itself
						// update the menu bar (for example dashboard controllers are multi area
						// controllers that need to update the menu bar).
						if (areaController instanceof LyDashboardController)
							subControllers.add(areaController);
					}
					else
						subControllers.add(areaController);
				}
			}
			else if (startingController instanceof LyTabAreaController)
			{
				// Assuming that this is a tabAreaController
				tabAreaController = (LyTabAreaController)startingController;

				areaController = tabAreaController.getSelectedAreaController();
				if (areaController instanceof LyMultiAreaController)
					compoundControllers.addLast((LyMultiAreaController)areaController);
				else
					subControllers.add(areaController);
			}
			else if (startingController instanceof LyDashboardController)
			{
				// Assuming that this is a dashboardController
				dashboardController = (LyDashboardController)startingController;

				areaController = dashboardController.getDisplayedAreaController();

				if (areaController != null)
				{
					if (areaController instanceof LyMultiAreaController)
						compoundControllers.addLast((LyMultiAreaController)areaController);
					else
						subControllers.add(areaController);
				}
			}
		}

		// look for each menu of sub-controllers
		n = subControllers.size();
		LyMenu rootMenu;

		menuBarList = new Vector<LyMenuBar>(n);

		// First generate all menuBars
		for (i = 0; i < n; i++)
		{
			areaController = subControllers.get(i);
			newMenuBar = new LyMenuBar();
			newMenuBar.setSorted(true);
			newMenuBar.addMenu(new LyMenu("LY_MENU_FILE"));
			areaController.completeMenuBar(newMenuBar, level + 1);

			menuBarList.addElement(newMenuBar);
		}

		_Tree menuTree = new _Tree();

		for (i = 0; i < n; i++)
		{
			areaController = subControllers.get(i);
			newMenuBar = menuBarList.elementAt(i);

			for (j = 0; j < newMenuBar.getMenuCount(); j++)
			{
				rootMenu = newMenuBar.getMenu(j);
				computeMenu(rootMenu, menuTree, "", 0, areaController.getName());
			}
		}

		// Browse the tree and create the associated menu (if they don't exist)
		_Node root = menuTree.getRootNode();

		// there is a group (only one because its the menuBar level)
		if (root.getSonsCount() > 0)
		{
			root = root.getSonAt(0);

			// First create all root menu (if they don't exists)
			_Node currentNode;

			for (i = 0; i < root.getSonsCount(); i++)
			{
				currentNode = root.getSonAt(i);
				label = currentNode.getLabel();

				menu = (LyMenu)menuBar.getComponentByLabel(label);

				if (menu == null)
				{
					menu = new LyMenu(label);
					menu.setId("_menu_" + _parent.getId());
					menu.setController(_parent.getId());
					menuBar.addMenu(menu);
				}
				fillMenu(menu, currentNode);
			}
		}

		// Add menus for the actions of the compound action
		super.completeMenuBar(menuBar, level);
	} // end method completeMenuBar

	/**
	 * Internal class to store the menu tree It will be of the following form: ROOT GROUP_0
	 * ROOT_MENU_1 GROUP_0 SUB_MENU_1 GROUP_0 ITEM_1 ITEM_2 GROUP_1 ITEM_3 ITEM_4 SUB_MENU_2 GROUP_0
	 * ITEM_5 ITEM_6 GROUP_1 ITEM_7 ITEM_8 GROUP_1 SUB_MENU_3 GROUP_0 ITEM_9 ITEM_10 ROOT_MENU_2
	 * GROUP_0 ITEM_11 ITEM_12.
	 *
	 * <p>Each group change will result in a separator in the final menu An item/menu can be
	 * multiple, in this case, in the final menu construction, a menu will be inserted with the name
	 * of the menu and the menu will be renamed with the controller name (done in the main class)
	 * </p>
	 *
	 * <p>This tree class allows to add an element, find an element, get the root element and it can
	 * dump itself A node is identify by it's identification string This id looks like:
	 * PARENT_1|PARENT_2|LABEL</p>
	 */
	protected class _Tree
	{
		/** the root node */
		protected _Node _rootNode;

		/**
		 * Constructor
		 */
		public _Tree()
		{
			_rootNode = new _Node("", null, null, null);
		}

		/**
		 * Add a new element to the tree Use the label to know where to add it.
		 *
		 * @param  multiLabel the identification of the node
		 * @param  data1      first data
		 * @param  data2      second data
		 * @return the new node
		 */
		public _Node addElement(String multiLabel, Object data1, Object data2)
		{
			// If the node already exist set to multiple
			_Node newNode = findElement(multiLabel);

			if (newNode != null)
			{
				newNode.setMultiple(true);

				newNode.addData(data1, data2);

				return newNode;
			}

			// get the parent node
			int    lastSep = multiLabel.lastIndexOf(VISIBLE_SEP);
			String parentId;
            _Node  parentNode;

			if (lastSep <= 0)
            {
				parentId = "";
                parentNode = _rootNode;
            }
			else
            {
				parentId = multiLabel.substring(0, lastSep);
				parentNode = findElement(parentId);
            }

			if (parentNode == null)
				// The parent element doesnot exists
				// must build everything
				parentNode = addElement(parentId, null, null);

            String label = multiLabel.substring(lastSep + 1);
            
			return parentNode.addSon(label, data1, data2);
		} // end method addElement

		/**
		 * Return the root node for the tree.
		 *
		 * @return the root node
		 */
		public _Node getRootNode()
		{
			return _rootNode;
		}

		/**
		 * Find an element from its id.
		 *
		 * @param  multiLabel id string
		 * @return the node or null if not found
		 */
		public _Node findElement(String multiLabel)
		{
			return _rootNode.findElement(multiLabel);
		}

		/**
		 * Dump the tree.
		 *
		 * @return the dump string
		 */
		public String toString()
		{
			return _rootNode.toString(0);
		}
	} // end class _Tree


	/**
	 * The node is an element of the tree, it has a label, a parent, multiple sons, can be multiple,
	 * and can store datas (up to two Object).
	 */
	protected class _Node
	{
		/** Id of the node */
		protected String		   _id;

		/** Label of the node */
		protected String		   _label;

		/** The parent node */
		protected _Node			   _parent;

		/** The sons (null if terminal element) */
		protected _Node[]          _sons;

		/** s the node multiple */
		protected boolean		   _multiple;

		/** Storage array */
		protected Vector<Object[]> _datas;

		/** The number of groups */
		protected int			   _groupsCount;

		/**
		 * Constructor.
		 *
		 * @param label  the label of the node
		 * @param parent the parent node (null if root element)
		 * @param data1  first data
		 * @param data2  second data
		 */
		public _Node(String label, _Node parent, Object data1, Object data2)
		{
			_label = label;
			_parent = parent;
			_multiple = false;
			_groupsCount = 0;

			if ((data1 != null) || (data2 != null))
			{
				_datas = new Vector<Object[]>();
				Object[] datas = new Object[2];

				datas[0] = data1;
				datas[1] = data2;

				_datas.addElement(datas);
			}
		}

		/**
		 * Set the number of groups for this node.
		 *
		 * @param count the new group count. If this value is less than the previous one, don't
		 *                   change it
		 */
		public void setGroupsCount(int count)
		{
			if (count > _groupsCount)
				_groupsCount = count;
		}

		/**
		 * Return the number of groups for this node.
		 *
		 * @return the number of groups
		 */
		public int getGroupsCount()
		{
			return _groupsCount;
		}

		/**
		 * Add a son to the current node.
		 *
		 * @param label label of the son
		 * @param data1 user data
		 * @param data2 user data
		 * @return a node
		 */
		public _Node addSon(String label, Object data1, Object data2)
		{
			_Node son = new _Node(label, this, data1, data2);

            if (_sons == null)
            {
                _sons = new _Node[1];
                _sons[0] = son;
                return son;
            }

            Object[] oldSons = _sons;
            int      n = _sons.length;

            _sons = new _Node[n + 1];
            System.arraycopy(oldSons, 0, _sons, 0, n);

            _sons[n] = son;
			return son;
		}

		/**
		 * Return the number of sons for the node.
		 *
		 * @return the number of sons
		 */
		public int getSonsCount()
		{
			return ((_sons == null) ? 0 : _sons.length);
		}

		/**
		 * Accessor to one son.
		 *
		 * @param  index index of the son
		 * @return the son at the given index
		 */
		public _Node getSonAt(int index)
		{
			return _sons[index];
		}

		/**
		 * Add data to the current node.
		 *
		 * @param data1 user data
		 * @param data2 user data
		 */
		public void addData(Object data1, Object data2)
		{
			if ((data1 != null) || (data2 != null))
			{
				if (_datas == null)
					_datas = new Vector<Object[]>();
				Object[] datas = new Object[2];

				datas[0] = data1;
				datas[1] = data2;

				_datas.addElement(datas);
			}
		}

		/**
		 * Find a node using its id.
		 *
		 * @param  multiLabel id of the node to find
		 * @return the node if found, null otherwise
		 */
		public _Node findElement(String multiLabel)
		{
			// if we are here, we don't match
			if (multiLabel == null)
				return null;

			if (_sons == null)
				return null;

			_Node   son;
            int     i, n = _sons.length;

            multiLabel = multiLabel.intern();
			for (i = 0; i < n; i++)
			{
				// check each son
				son = _sons[i];
				if (son.getIdString() == multiLabel)
					return son;
			}

			_Node find = null;
			// No son match, ask each son if one of its son matches
			for (i = 0; i < n; i++)
			{
				// check each son
				son = _sons[i];
				find = son.findElement(multiLabel);
				if (find != null)
					break;
			}

			return find;
		} // end method findElement

		/**
		 * Change multiplicity for the node.
		 *
		 * @param multiple multiplicity
		 */
		public void setMultiple(boolean multiple)
		{
			_multiple = multiple;
		}

		/**
		 * Is the node multiple.
		 *
		 * @return true if the node is multiple, false otherwise
		 */
		public boolean isMultiple()
		{
			return _multiple;
		}

		/**
		 * Build the id string for the current node.
		 *
		 * @return the id string of this node
		 */
		protected String getIdString()
		{
			if (_id != null)
				return _id;

			_id = _label;

			// Root node -> no need for an id
			if (_id.length() == 0)
				return _id;

			String parentId = null;

			if (_parent != null)
				parentId = _parent.getIdString();

			if ((parentId != null) && (parentId.length() > 0))
				_id = parentId + VISIBLE_SEP + _id;

            _id = _id.intern();
			return _id;
		}

		/**
		 * Return the number of data for this node.
		 *
		 * @return the size of the data
		 */
		public int getDataSize()
		{
			return ((_datas == null) ? 0 : _datas.size());
		}

		/**
		 * Return the data at the given index.
		 *
		 * @param  index the index of the data to fetch
		 * @return the data at the given index (if they exists)
		 */
		public Object[] getDataAt(int index)
		{
			Object[] result = null;

			if ((_datas != null) && (index < _datas.size()))
				result = _datas.elementAt(index);

			return result;
		}

		/**
		 * Return the label of the node.
		 *
		 * @return the label of the node
		 */
		public String getLabel()
		{
			return _label;
		}

		/**
		 * Dump the node.
		 *
		 * @param  pos the depth of this node
		 * @return the dump string
		 */
		public String toString(int pos)
		{
			StringBuffer buffer = new StringBuffer();

			for (int i = 0; i < pos; i++)
				buffer.append("\t");

			String size = ((_sons == null) ? "0" : ("" + _sons.length));
			String dataSize = ((_datas == null) ? "0" : ("" + _datas.size()));

			buffer.append(_label + "( multiple=" + _multiple + " ) ( " + size + " , " + dataSize
				+ " )\n");

			if (_sons != null)
			{
                int     i, n = _sons.length;

				for (i = 0; i < n; i++)
					buffer.append(_sons[i].toString(pos + 1));
			}

			return buffer.toString();
		}
	} // end class _Node
} // end class LyMultiAreaBuilder
