/**
 * LyMultiFilterBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyMultiFilterBuilder.java 33722 2012-01-06 08:25:08Z bangiole $
 */

package leon.control.builder;

import java.util.*;
import leon.control.*;
import leon.info.*;
import leon.view.*;


/**
 * Creates a view to type in a filter with multiple criteria.
 *
 * @see LyFilterController
 */
public class LyMultiFilterBuilder extends LySimpleFilterBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyMultiFilterBuilder.java 33722 2012-01-06 08:25:08Z bangiole $";

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyMultiFilterBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Builds the view component.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{
		LyComponent	result = super.build();
		
		if (result instanceof LyArea)
		{
			LyArea				area = (LyArea)result;
			Vector<LyFormField>	fields = new Vector<LyFormField>();
			
			area.getComponents(LyFormField.class, fields);
			
			int		nFields = fields.size();
			
			// If there are only a few components, remove scrolling from main area
			if (nFields < 4)
				area.setScrolling(LyArea.Scrolling.NO);
		}
		
		return result;
	}
	
	/**
	 * Build a form field for a given fieldInfo.
	 *
	 * @param  fieldInfo the fieldInfo corresponding to the built form field
	 * @param  editable  indicates whether the form field must be editable or not
	 * @return the form field corresponding to the fieldInfo
	 */
	public LyFormField buildFormField(LyFieldInfo fieldInfo, boolean editable)
	{
        LyFormField      formField = super.buildFormField(fieldInfo, editable);
        LyFieldItem      item = formField.getFieldItem();
        boolean          checkBox = ((LyMultiFilterController)_parent).hasCheckBoxes();
        
        if (!checkBox)
        {
            formField.setCheckBox(false);
            formField.setEnabled(true);
            formField.setEditable(true);
        }
        else
        {
        	formField.setEnabled(false);
            formField.setEditable(false);
        }

        if (item instanceof LyForm)
        {
            LyForm form = (LyForm)item;

            form.setEditable(formField.isEnabled());
        }

        return formField;
	}

	/**
	 * Build an intervalfor a date fieldInfo.
	 *
	 * @param  fieldInfo the date fieldInfo corresponding to the built field item
	 * @param  fieldItem1     the first value field item
	 * @param  fieldItem2     the second value field item
	 * @return the field item corresponding to the fieldInfo
	 */
	protected LyFieldItem buildInterval(LyFieldInfo fieldInfo, LyFieldItem fieldItem1,
		LyFieldItem fieldItem2)
	{
		LyForm	    form = new LyForm(fieldInfo.getId());
		LyFormField formField1 = new LyFormField(fieldInfo.getId() + "_1");
		LyFormField formField2 = new LyFormField(fieldInfo.getId() + "_2");
		LyCaption   caption;

		caption = new LyCaption();
		caption.setLabel("LY_STARTS_FROM");
		formField1.setCaption(caption);
		formField1.setFieldItem(fieldItem1);
		form.addFormField(formField1);

		LyConstraint constraint = new LyConstraint();

		constraint.setColSpan(1);
		constraint.setRemainder(false);
		formField1.setConstraint(constraint);

		caption = new LyCaption();
		caption.setLabel("LY_ENDS_TO");
		formField2.setCaption(caption);
		formField2.setFieldItem(fieldItem2);
		form.addFormField(formField2);

		form.setBorder(false);

		return form;
	} // end method buildInterval

	/**
	 * Gets all filter values corresponding to the given field info.
	 *
	 * @param fieldInfo A field info.
	 * @param filter    A filter.
	 * @param result    result list of filter values.
	 */
	protected void getFilterValues(LyFieldInfo fieldInfo, LyFilter filter, Vector<Object> result)
	{
		if (filter == null)
			return;

		if ((fieldInfo instanceof LyNumberFieldInfo)
				|| (fieldInfo instanceof LyDateFieldInfo))
		{
			LyFilterElement element;
			int			    i, n = filter.getSize();
			Object		    value1 = null, value2 = null;
            LyFilterExpression.Condition		    condition;

			for (i = 0; i < n; i++)
			{
				element = filter.getElement(i);
				if ((element instanceof LyFilterExpression)
						&& (((LyFilterExpression)element).getFieldInfo() == fieldInfo))
				{
					condition = ((LyFilterExpression)element).getCondition();
					if (condition == LyFilterExpression.Condition.GREATER_OR_EQUALS)
						value1 = ((LyFilterExpression)element).getValue();
					else if (condition == LyFilterExpression.Condition.LOWER_OR_EQUALS)
						value2 = ((LyFilterExpression)element).getValue();
				}
				else if (element instanceof LyFilter)
					getFilterValues(fieldInfo, (LyFilter)element, result);
			}

			if ((value1 != null) || (value2 != null))
			{
				result.addElement(value1);
				result.addElement(value2);
			}
		}
		else
			super.getFilterValues(fieldInfo, filter, result);
	} // end method getFilterValues

	/**
	 * Build a field item for a date fieldInfo.
	 *
	 * @param  dateField the date fieldInfo corresponding to the built field item
	 * @param  value     the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildDate(LyDateFieldInfo dateField, Object value)
	{
		Object value1 = null, value2 = null;

		if ((value instanceof Vector<?>)
				&& (((Vector<?>)value).size() == 2))
		{
			value1 = ((Vector<?>)value).elementAt(0);
			value2 = ((Vector<?>)value).elementAt(1);
		}

		LyFieldItem date1 = super.buildDate(dateField, value1);
		LyFieldItem date2 = super.buildDate(dateField, value2);

		return buildInterval(dateField, date1, date2);
	}

	/**
	 * Build a field item for a number fieldInfo.
	 *
	 * @param  numberField the number fieldInfo corresponding to the built field item
	 * @param  value       the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildNumber(LyNumberFieldInfo numberField, Object value)
	{
		Object value1 = null, value2 = null;

		if ((value instanceof Vector<?>)
				&& (((Vector<?>)value).size() == 2))
		{
			value1 = ((Vector<?>)value).elementAt(0);
			value2 = ((Vector<?>)value).elementAt(1);
		}

		LyFieldItem number1 = super.buildNumber(numberField, value1);
		LyFieldItem number2 = super.buildNumber(numberField, value2);

		return buildInterval(numberField, number1, number2);
	}
} // end class LyMultiFilterBuilder
