/**
 * LySimpleTableBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyPanelBuilder.java 40953 2012-09-27 14:05:40Z bangiole $
 */

package leon.control.builder;

import leon.control.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.mobile.*;


/**
 * This builder builds the basic view for a simple table controller.
 *
 * @see LySimpleTableController
 */
public abstract class LyPanelBuilder extends LyAreaBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyPanelBuilder.java 40953 2012-09-27 14:05:40Z bangiole $";

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyPanelBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Complete the given menu bar.
	 *
	 * @param menuBar the menubar to complete
	 * @param level   the current level
	 */
	public void completeMenuBar(LyMenuBar menuBar, int level)
	{
		if (menuBar == null)
			return;

		LyAction action = ((LyPanelController)_parent).getAction();

		if (action.hasMark(LyAction.MARK_NO_MENUBAR))
			return;

		int nFile = completeActionsMenu(menuBar);

		completeFileMenu(menuBar, nFile);

		completeEditionMenu(menuBar);

		completeDisplayMenu(menuBar);

		// Recurse in the client area
		super.completeMenuBar(menuBar, level);

		completeHelpMenu(menuBar);
	}

	/**
	 * Complete the action menu
	 * @param menuBar the menu bar
	 * @return the count of added action
	 */
	protected int completeActionsMenu(LyMenuBar menuBar)
	{
		return 0;
	}

	/**
	 * Complete the file menu
	 * @param menuBar the menu bar
	 * @param nFile
	 */
	protected void completeFileMenu(LyMenuBar menuBar, int nFile)
	{
	}

	/**
	 * Complete the help menu
	 * @param menuBar the menu bar
	 */
	protected void completeHelpMenu(LyMenuBar menuBar)
	{
		LyPanelController	panelController = (LyPanelController)_parent;
		LyAction			action = panelController.getAction();

		if (action != null && action.hasMark(LyAction.MARK_NO_MENUBAR))
			return;

		String helpFile = panelController.getHelpFile();

		// Shall we add the help content
		//if ((action == null) && (classInfo == null) && (helpFile == null))
		if (helpFile == null)
			return;

		// If we have an action component, add its contextual help
		LyAction    rootAction;
		LyClassInfo classInfo = null;

		// Check Help Button if help file exists
		rootAction = ((action == null) ? null : action.getRoot());

		if (panelController instanceof LyGetClassInfoInterface)
			classInfo = ((LyGetClassInfoInterface)panelController).getClassInfo();

		String classId = ((classInfo == null) ? null : classInfo.getId());
		String actionId = ((action == null) ? null : action.getId());
		String rootActionId = ((rootAction == null) ? null : rootAction.getId());

		// Build help menu info (to be able to show this help)
		String info = LyViewController.DEMAND_HELP + ":";

		if (classId != null)
			info += classId;

		info += ":";

		if (actionId != null)
			info += actionId;

		info += ":";

		if (rootActionId != null)
			info += rootActionId;

		LyEnvironment	environment = _parent.getEnvironment();
		LyItem			item;

		if (action != null)
			item = new LyItem(info, action.getName());
		else if (classInfo != null)
			item = new LyItem(info, classInfo.getName());
		else
			item = new LyItem(info, "LY_HELP");

		String	helpStr = environment.getString("LY_MENU_HELP");

		if ((helpStr == null)
				|| (helpStr.length() == 0))
			return;

		LyMenu	helpMenu = (LyMenu)menuBar.getComponentByLabel("LY_MENU_HELP");

		if (helpMenu == null)
		{
			helpMenu = new LyMenu("LY_MENU_HELP");
			menuBar.addMenu(helpMenu);
		}
		else
			helpMenu.addSep();

		item.setDialog(true);
		item.setShortcut("LY_SHORTCUT_HELP");
		helpMenu.addItem(item);
	}

	/**
	 * Complete the display menu
	 * @param menuBar the menu bar
	 */
	protected void completeDisplayMenu(LyMenuBar menuBar)
	{
		LyEnvironment environment = _parent.getEnvironment();
		String		  displayString = environment.translate("LY_MENU_DISPLAY");

		if ((displayString != null) && (displayString.length() != 0))
		{
			// Search for display Menu
			LyMenu displayMenu = (LyMenu)menuBar.getComponentByLabel("LY_MENU_DISPLAY");

			// if menu display does not exist : create it
			if (displayMenu == null)
			{
				displayMenu = new LyMenu("LY_MENU_DISPLAY");
				displayMenu.setId("_displayMenu");
				menuBar.addMenu(displayMenu);
			}
			else
				displayMenu.addSep();

			addFilterItems(displayMenu);
		}
	}

	/**
	 * Complete the edition menu
	 * @param menuBar the menu bar
	 */
	protected void completeEditionMenu(LyMenuBar menuBar)
	{
		LyEnvironment environment = _parent.getEnvironment();
		String		  displayString;

		displayString = environment.translate("LY_MENU_EDITION");
		if ((displayString != null) && (displayString.trim().length() > 0))
		{
			// Search for Edition Menu
			LyMenu editionMenu = (LyMenu)menuBar.getComponentByLabel("LY_MENU_EDITION");

			// If it doesn't exist, create it
			if (editionMenu == null)
			{
				editionMenu = new LyMenu("LY_MENU_EDITION");
				menuBar.addMenu(editionMenu);
			}
			else
				editionMenu.addSep();

			addFindItems(editionMenu);

			addSelectionItems(editionMenu);
		}
	}

	/**
	 * Add the find item to the edition menu
	 * @param editionMenu the edition menu
	 */
	protected void addFindItems(LyMenu editionMenu)
	{
		LyPanelController panel = (LyPanelController)_parent;

		if (!(panel instanceof LyFindInterface))
			return;

		// Add class actions in Actions menu
		LyClassInfo			    classInfo = panel.getClassInfo();
		LyItem				    item;
		LyEnvironment		    environment = _parent.getEnvironment();
		LyInfoList<LyFieldInfo> fields;
		String				    displayString;
		String				    id = null, image = null;
		LyAction			    actionItem = null;

		// Add find menu if at least one fieldinfos has the find Mark
		fields = classInfo.getFieldInfos(LyFieldInfo.MARK_FIND);
		if ((fields == null) || (fields.getSize() == 0))
		{
			fields = classInfo.getFieldInfos();
			fields = fields.getInfos(LyFieldInfo.MARK_HIDDEN, true);
			fields = fields.getInfos(LyFieldInfo.MARK_SECRET, true);
		}

		if ((fields != null) && (fields.getSize() > 0))
		{
			// Search
			displayString = environment.translate("LY_ITEM_FIND");
			if ((displayString != null) && (displayString.trim().length() > 0))
			{
				id = LyFindController.DEMAND_FIND;
				actionItem = (LyAction)_parent.getApplication().getInfo(id);
				image = ((LyPanelController)_parent).getMenuItemImage(actionItem);
				item = new LyItem(id, "LY_ITEM_FIND");
				item.setCheckBox(true);
				item.setChecked(false);
				item.setImage(image);
				item.setDialog(true);
				item.setController(_parent.getId());
				item.setShortcut("LY_SHORTCUT_FIND");
				editionMenu.addItem(item);
			}
			displayString = environment.translate("LY_ITEM_NEXT");
			if ((displayString != null) && (displayString.trim().length() > 0))
			{
				id = LyFindController.DEMAND_NEXT;
				actionItem = (LyAction)_parent.getApplication().getInfo(id);
				image = ((LyPanelController)_parent).getMenuItemImage(actionItem);
				item = new LyItem(id, "LY_ITEM_NEXT");
				item.setImage(image);
				item.setEnabled(false);
				item.setController(_parent.getId());
				item.setShortcut("LY_SHORTCUT_NEXT");
				editionMenu.addItem(item);
			}

			displayString = environment.translate("LY_ITEM_PREVIOUS");
			if ((displayString != null) && (displayString.trim().length() > 0))
			{
				id = LyFindController.DEMAND_PREVIOUS;
				actionItem = (LyAction)_parent.getApplication().getInfo(id);
				image = ((LyPanelController)_parent).getMenuItemImage(actionItem);
				item = new LyItem(id, "LY_ITEM_PREVIOUS");
				item.setImage(image);
				item.setEnabled(false);
				item.setController(_parent.getId());
				editionMenu.addItem(item);
			}
		}
	} // end method addFindItems

    /**
     * Add the find tools to a toolbar
     * @param toolbar the toolbar
     */
    protected void addFindTools(LyToolBar toolbar)
    {
        LyPanelController panel = (LyPanelController)_parent;

        if (!(panel instanceof LyFilterEndInterface))
            return;

        LyClassInfo             classInfo = ((LyPanelController)_parent).getClassInfo();
        LyInfoList<LyFieldInfo> fields;

        fields = classInfo.getFieldInfos(LyFieldInfo.MARK_FIND);
        if ((fields == null) || (fields.getSize() == 0))
        {
            fields = classInfo.getFieldInfos();
            fields = fields.getInfos(LyFieldInfo.MARK_HIDDEN, true);
            fields = fields.getInfos(LyFieldInfo.MARK_SECRET, true);
        }

        if ((fields == null) || (fields.getSize() == 0))
            return;

        LyEnvironment           environment = _parent.getEnvironment();
        String                  displayString;

        displayString = environment.translate("LY_TOOL_FIND");
        if ((displayString != null)
            && (displayString.trim().length() > 0))
        {
            LyTool  tool = new LyTool(LyFindController.DEMAND_FIND, "");

            tool.setImage("LY_TOOL_FIND");
            tool.setTooltip("LY_ITEM_FIND");
            tool.setDialog(true);
            toolbar.addTool(tool);
        }
    }

	/**
	 * Add selection item to the edition menu
	 * @param editionMenu the edition menu
	 */
	protected void addSelectionItems(LyMenu editionMenu)
	{
		LyPanelController panel = (LyPanelController)_parent;

		if (!(panel instanceof LyFindInterface)) // XXX
			return;

		// Add class actions in Actions menu
		LyItem		  item;
		LyEnvironment environment = _parent.getEnvironment();
		String		  displayString;
		String		  id = null, image = null;
		LyAction	  actionItem = null;

		// Selection
		boolean		  addSep = true;

		if (panel.enablesMultiSelection())
		{
			displayString = environment.translate("LY_ITEM_SELECT_ALL");
			if ((displayString != null) && (displayString.trim().length() > 0))
			{
				editionMenu.addSep();
				addSep = false;
				id = LyTableController.DEMAND_SELECT_ALL;
				actionItem = (LyAction)_parent.getApplication().getInfo(id);
				image = ((LyPanelController)_parent).getMenuItemImage(actionItem);
				item = new LyItem(id, "LY_ITEM_SELECT_ALL");
				item.setImage(image);
				item.setController(_parent.getId());
				item.setShortcut("LY_SHORTCUT_SELECT_ALL");
				editionMenu.addItem(item);
			}

			displayString = environment.translate("LY_ITEM_UNSELECT_ALL");
			if ((displayString != null) && (displayString.trim().length() > 0))
			{
				id = LyTableController.DEMAND_UNSELECT_ALL;
				actionItem = (LyAction)_parent.getApplication().getInfo(id);
				image = ((LyPanelController)_parent).getMenuItemImage(actionItem);
				item = new LyItem(id, "LY_ITEM_UNSELECT_ALL");
				item.setImage(image);
				item.setController(_parent.getId());
				if (addSep)
					editionMenu.addSep();
				editionMenu.addItem(item);
			}
		}
	} // end method addSelectionItems


	/**
	 * Add the filter items to the display menu
	 * @param displayMenu the display menu
	 */
	protected void addFilterItems(LyMenu displayMenu)
	{
		LyPanelController panel = (LyPanelController)_parent;

		if (!(panel instanceof LyFilterEndInterface))
			return;

		LyClassInfo			    classInfo = ((LyPanelController)_parent).getClassInfo();
		LyItem				    item;
		LyEnvironment		    environment = _parent.getEnvironment();
		LyInfoList<LyFieldInfo> fields;
		String				    displayString;
		String				    id = null, image = null;
		LyAction			    actionItem = null;

		displayString = environment.translate("LY_MENU_FILTER");
		if ((displayString != null) && (displayString.trim().length() > 0))
		{
			// Create Filtering menu
			fields = classInfo.getFieldInfos(LyFieldInfo.MARK_FILTER);
			if ((fields == null) || (fields.getSize() == 0))
			{
				fields = classInfo.getFieldInfos();
				fields = fields.getInfos(LyFieldInfo.MARK_HIDDEN, true);
				fields = fields.getInfos(LyFieldInfo.MARK_SECRET, true);
			}

			if ((fields != null) && (fields.getSize() > 0))
			{
				// Within display, add submenu LY_MENU_FILTER
				LyMenu menuFilter;

				menuFilter = new LyMenu("LY_MENU_FILTER");
				displayMenu.addMenu(menuFilter);

				addFastFilterItems(menuFilter);

				displayString = environment.translate("LY_ITEM_FILTER_NEW");
				if ((displayString != null) && (displayString.trim().length() > 0))
				{
					id = LyTableController.SIMPLE_FILTER_NEW;
					actionItem = (LyAction)_parent.getApplication().getInfo(id);
					image = ((LyPanelController)_parent).getMenuItemImage(actionItem);
					item = new LyItem(id, "LY_ITEM_FILTER_NEW");
					item.setImage(image);
					item.setDialog(true);
					item.setController(_parent.getId());
                    item.setShortcut("LY_SHORTCUT_NEW_FILTER");
					menuFilter.addItem(item);
				}

				displayString = environment.translate("LY_ITEM_FILTER_SET");
				if ((displayString != null) && (displayString.trim().length() > 0))
				{
					id = LyTableController.FILTER_SET;
					actionItem = (LyAction)_parent.getApplication().getInfo(id);
					image = ((LyPanelController)_parent).getMenuItemImage(actionItem);
					item = new LyItem(id, "LY_ITEM_FILTER_SET");
					item.setImage(image);
					item.setEnabled(false);
					item.setDialog(true);
					item.setController(_parent.getId());
                    item.setShortcut("LY_SHORTCUT_SET_FILTER");
					menuFilter.addItem(item);
				}

				displayString = environment.translate("LY_ITEM_FILTER_ACTIVE");
				if ((displayString != null) && (displayString.trim().length() > 0))
				{
					id = LyTableController.FILTER_ACTIVE;
					actionItem = (LyAction)_parent.getApplication().getInfo(id);
					image = ((LyPanelController)_parent).getMenuItemImage(actionItem);
					item = new LyItem(id, "LY_ITEM_FILTER_ACTIVE");
					item.setImage(image);
					item.setEnabled(false);
					item.setCheckBox(true);
					item.setController(_parent.getId());
					menuFilter.addItem(item);
				}
			}
		}
	} // end method addFilterItems

	/**
	 * Add the fast filter items to the filter menu
	 * @param menuFilter the filter menu
	 */
	protected void addFastFilterItems(LyMenu menuFilter)
	{
	}

	/**
	 * Add the preferences tools
	 * @param toolbar a toolbar
	 */
	protected void addPreferencesTools(LyToolBar toolbar)
	{
		LyEnvironment environment = _parent.getEnvironment();

		String		  displayString = environment.translate("LY_MENU_PREFERENCES");

		if ((displayString == null)
		        || (displayString.length() == 0)
		        || "_null".equals(displayString))
			return;

		// Don't add save/delete tools if there's nothing in the toolbar
		if (toolbar.getChildCount() == 0)
		    return;

		// Create menu items
		displayString = environment.translate("LY_TOOL_SAVE_PREFERENCE");

		if ((displayString == null) || (displayString.length() == 0) || "_null".equals(displayString))
			return;

		// add items SAVE_PREFERENCES
		LyTool tool = new LyTool(LyPanelController.DEMAND_SAVE_PREFERENCE + LyController.ID_SEP + _parent.getId(), "");

		tool.setTooltip("LY_TOOL_SAVE_PREFERENCE");
		tool.setImage("LY_TOOL_SAVE_PREFERENCE");
		tool.setDialog(true);
		toolbar.addTool(tool);

		// add items DELETE_PREFERENCES
		tool = new LyTool(LyPanelController.DEMAND_DELETE_PREFERENCE + LyController.ID_SEP + _parent.getId(), "");
		tool.setTooltip("LY_TOOL_DELETE_PREFERENCE");
		tool.setImage("LY_TOOL_DELETE_PREFERENCE");
		tool.setDialog(true);

		tool.setEnabled(false);
		toolbar.addTool(tool);
	} // end method addPreferencesTool


	/**
	 * Add the filter tools
	 * @param toolbar a toolbar
	 */
	protected void addFilterTools(LyToolBar toolbar)
	{
		LyPanelController panel = (LyPanelController)_parent;

		if (!(panel instanceof LyFilterEndInterface))
			return;

		LyClassInfo				classInfo = ((LyPanelController)_parent).getClassInfo();
		LyEnvironment			environment = _parent.getEnvironment();
		LyInfoList<LyFieldInfo>	fields;
		String					displayString;
		String					id = null;
		LyTool					tool;

		fields = classInfo.getFieldInfos(LyFieldInfo.MARK_FILTER);
		if ((fields == null) || (fields.getSize() == 0))
		{
			fields = classInfo.getFieldInfos();
			fields = fields.getInfos(LyFieldInfo.MARK_HIDDEN, true);
			fields = fields.getInfos(LyFieldInfo.MARK_SECRET, true);
		}

		if ((fields == null) || (fields.getSize() == 0))
			return;

		displayString = environment.translate("LY_TOOL_FILTER_NEW");
		if ((displayString != null)
				&& (displayString.trim().length() > 0)
				&& !"_null".equals(displayString))
		{
			id = LyTableController.SIMPLE_FILTER_NEW + LyController.ID_SEP + _parent.getId();
			tool = new LyTool(id, "");
			tool.setImage("LY_TOOL_FILTER_NEW");
			tool.setTooltip("LY_TOOL_FILTER_NEW");
			tool.setDialog(true);
			toolbar.addTool(tool);
		}

		displayString = environment.translate("LY_TOOL_FILTER_SET");
		if ((displayString != null)
				&& (displayString.trim().length() > 0)
				&& !"_null".equals(displayString))
		{
			id = LyTableController.FILTER_SET + LyController.ID_SEP + _parent.getId();
			tool = new LyTool(id, "");
			tool.setImage("LY_TOOL_FILTER_SET");
			tool.setTooltip("LY_TOOL_FILTER_SET");
			tool.setEnabled(false);

			tool.setDialog(true);

			toolbar.addTool(tool);
		}

		if(getViewManager().hasProperty(LyMobileViewManager.MOBILE_ONLY_PROPERTY))
		{
			displayString = environment.translate("LY_ITEM_FILTER_ACTIVE");
			if ((displayString != null) && (displayString.trim().length() > 0))
			{
				id = LyTableController.FILTER_ACTIVE;
				tool = new LyTool(id, "");
				tool.setImage("LY_UNACTIVE_FILTER");
				tool.setTooltip("LY_ACTIVATE_FILTER");
				tool.setLabel("LY_ACTIVATE_FILTER");
				tool.setEnabled(false);
				toolbar.addTool(tool);
			}
		}
	}

	/**
	 * Complete the panel
	 * @param area the area to complte
	 */
	protected void completePanel(LyArea area)
	{
		updateTitleBar();

		LyEnvironment environment = _parent.getEnvironment();

		if (!environment.testResource("LY_RIGHT_TOOLBAR"))
			return;

		LyFormField preferenceField = buildPreferenceField();
		if (preferenceField == null)
			return;

		LyTitleBar  titleBar = area.getTitleBar();
		LyAction    action = ((LyAreaController)_parent).getAction();

		// If there is no titlebar on the area, use the right toolbar to display preferences field.
		// Except if action has the noToolbar mark.
		if ((titleBar == null)
			&& (action.hasMark(LyAction.MARK_NO_TOOLBAR)
			    || !environment.testResource("LY_RIGHT_TOOLBAR")))
			return;

		LyToolBar   toolBar = null;

		if (titleBar != null)
        {
			toolBar = titleBar.getToolBar();
            if (toolBar == null)
                toolBar = area.getRightToolBar();
        }
        else
            toolBar = area.getTopRightToolBar();

		if (toolBar == null)
		{
			toolBar = new LyToolBar();
			toolBar.setFill(LyToolBar.Fill.HORIZONTAL);
			toolBar.setStyle(LyToolBar.Style.SEPARATOR);
			toolBar.setHAlign(LyToolBar.HAlign.RIGHT);
			toolBar.setVAlign(LyToolBar.VAlign.TOP);
			toolBar.setController(_parent.getId());
		}
		else
			toolBar.addSep();

		toolBar.addFormField(preferenceField);
		if (action.hasMark(LyAction.MARK_NO_TOOLBAR))
			completeRightToolBar(toolBar);

		if (titleBar != null)
		{
			if (titleBar.getToolBar() == null)
				titleBar.setToolBar(toolBar);
		}
		else
		{
			if (area.getTopRightToolBar() == null)
				area.setTopRightToolBar(toolBar);
		}
	} // end method completePanel


	/**
	 * Builds field to hold preferences.
	 * @return the preference field
	 */
	protected LyFormField buildPreferenceField()
	{
		LyFormField   formField = new LyFormField(_parent.getId() + "_preferencesField");
		LyRelation    relation = new LyRelation();
		LyEnvironment environment = _parent.getEnvironment();
		String        displayString = environment.translate("LY_ITEM_FILTER_ACTIVE");

		if ((displayString != null) && (displayString.trim().length() > 0))
		{
			LyLink link = new LyLink("LY_NO_CHOICE", "LY_ALL");

			relation.addLink(link);
		}

		formField.setFieldItem(relation);
		formField.setHidden(true);
		formField.setNeedCheck(true);
		formField.setController(_parent.getId());

		return formField;
	}

	/**
	 * Add the preference items to the display menu
	 * @param displayMenu the display menu
	 */
	protected void addPreferencesMenu(LyMenu displayMenu)
	{
		LyEnvironment environment = _parent.getEnvironment();

		String		  displayString = environment.translate("LY_MENU_PREFERENCES");

		if ((displayString == null) || (displayString.length() == 0))
			return;

		// Create menu items
		LyMenu menu = new LyMenu("LY_MENU_PREFERENCES");

		menu.setId("_preferencesMenu");
		menu.setController(_parent.getId());

		displayString = environment.translate("LY_ITEM_SAVE_PREFERENCE");

		if ((displayString == null) || (displayString.length() == 0))
			return;

		// add items SAVE_PREFERENCES
		LyItem item = new LyItem(LyPanelController.DEMAND_SAVE_PREFERENCE,
				"LY_ITEM_SAVE_PREFERENCE");

		item.setDialog(true);
		item.setController(_parent.getId());
		menu.addItem(item);

		// add items DELETE_PREFERENCES
		item = new LyItem(LyPanelController.DEMAND_DELETE_PREFERENCE, "LY_ITEM_DELETE");
		item.setDialog(true);
		item.setController(_parent.getId());
		item.setEnabled(false);
		menu.addItem(item);

		// Add menu
		displayMenu.addSep();
		displayMenu.addMenu(menu);
	} // end method addPreferencesMenu

	/**
	 * Complete the right toolbar
	 * @param toolbar the right toolbar
	 * @return a toolbar
	 */
	protected LyToolBar completeRightToolBar(LyToolBar toolbar)
	{
		LyAction action = ((LyAreaController)_parent).getAction();

		if (action.hasMark(LyAction.MARK_NO_TOOLBAR))
			return toolbar;

		boolean isDashboard = (action.getRoot().getId() == LyAction.ID_DASHBOARD);

		if (isDashboard)
			return toolbar;

		LyEnvironment environment = _parent.getEnvironment();

		if (environment.testResource("LY_RIGHT_TOOLBAR"))
		{
			if (toolbar == null)
			{
				toolbar = new LyToolBar();
				toolbar.setFill(LyToolBar.Fill.HORIZONTAL);
				toolbar.setStyle(LyToolBar.Style.SEPARATOR);
				toolbar.setHAlign(LyToolBar.HAlign.RIGHT);
				toolbar.setVAlign(LyToolBar.VAlign.TOP);
			}

			LyToolBar	groupToolbar = null;
			LyTool		groupTool = null;

			if (environment.testResource("LY_RIGHT_TOOLBAR_GROUP"))
			{
				LyArea	area = new LyArea();

				area.setController(_parent.getId());
				groupTool = new LyTool(_parent.getId() + "_rightToolbarGroup", null);
				toolbar.addTool(groupTool);
				groupToolbar = new LyToolBar();
				groupToolbar.setFill(LyToolBar.Fill.HORIZONTAL);
				groupToolbar.setStyle(LyToolBar.Style.SEPARATOR);
				area.setToolBar(groupToolbar);
				groupTool.setComponent(area);
			}
			else
				groupToolbar = toolbar;

			addRightToolbarTools(groupToolbar);

			if ((groupTool != null) && (groupToolbar.getComponent(0) instanceof LyTool))
			{
				LyTool	firstTool = (LyTool)groupToolbar.getComponent(0);

				groupTool.setImage(firstTool.getImage());
				groupTool.setLabel(firstTool.getLabel());
				groupTool.setTooltip(firstTool.getTooltip());
				groupTool.setDialog(true);
			}
			else if (groupTool != null)
				toolbar.removeTool(groupTool);
		}

		return super.completeRightToolBar(toolbar);
	} // end method completeRightToolBar

	/**
	 * Add tools in the right toolbar.
	 * @param toolbar a toolbar
	 */
	protected void addRightToolbarTools(LyToolBar toolbar)
	{
		addFindTools(toolbar);
		addFilterTools(toolbar);
		addPreferencesTools(toolbar);
	}
} // end class LyPanelBuilder
