/**
 * LySetBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LySetBuilder.java 40687 2012-09-06 15:07:47Z jlsaouli $
 */

package leon.control.builder;

import leon.control.*;
import leon.control.form.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;


/**
 * This builder creates the modification form for the set controller It is based on the consult
 * builder.
 *
 * @see LyConsultBuilder
 * @see LySetController
 */
public class LySetBuilder extends LyConsultBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LySetBuilder.java 40687 2012-09-06 15:07:47Z jlsaouli $";

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LySetBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Build a field item for a given fieldInfo and unit.
	 *
	 * @param  fieldInfo the fieldInfo corresponding to the built field item
	 * @param value      the current value
	 * @param  unit      the unit corresponding to the built field item
	 * @return the field item corresponding to the fieldInfo and unit.
	 */
	public LyFieldItem buildFieldItem(LyFieldInfo fieldInfo, Object value, LyUnit unit)
	{
		LyFieldSetBuilder set = (LyFieldSetBuilder)getFormFieldBuilder(fieldInfo);

		return set.buildFieldItem(unit);
	}

	/**
	 * Build a relation that contains all units of the given field.
	 *
	 * @param  fieldInfo FieldInfo for which a units list is required.
	 * @return The relation containing all possible units for the given field info.
	 */
	public LyFieldUnit buildUnitsList(LyFieldInfo fieldInfo)
	{
		LyFieldSetBuilder set = (LyFieldSetBuilder)getFormFieldBuilder(fieldInfo);

		return set.buildUnitsList();
	}

	/**
	 * Builds a relation to change the relation type.
	 * @param fieldInfo a fieldInfo
	 * @return the built relation item
	 */
	public LyRelation buildTypesRelation(LyFieldInfo fieldInfo)
	{
		LyFieldSetBuilder set = (LyFieldSetBuilder)getFormFieldBuilder(fieldInfo);

		return set.buildTypesRelation();
	}

	/**
	 * Indicates if the current field is in consult mode.
	 *
	 * @param  fieldInfo the field
	 * @return true if the field is in consult mode
	 */
	public boolean isConsult(LyFieldInfo fieldInfo)
	{
		LyFieldSetBuilder set = (LyFieldSetBuilder)getFormFieldBuilder(fieldInfo);

		return set.isConsult();
	}

	/**
	 * Indicates if the current field is enabled.
	 *
	 * @param  fieldInfo the field
	 * @return true if the field is enabled
	 */
	public boolean isEnabled(LyFieldInfo fieldInfo)
	{
		LyFieldSetBuilder set = (LyFieldSetBuilder)getFormFieldBuilder(fieldInfo);

		return set.isEnabled();
	}

	/**
	 * Indicates if the current field is editable.
	 *
	 * @param  fieldInfo the field
	 * @return true if the field is editable
	 */
	public boolean isEditable(LyFieldInfo fieldInfo)
	{
		LyFieldSetBuilder set = (LyFieldSetBuilder)getFormFieldBuilder(fieldInfo);

		return set.isEditable();
	}

	/**
	 * Update the toolbar for the given field.
	 *
	 * @param  fieldInfo the field
	 * @return the toolbar for the given field
	 */
	public LyToolBar updateToolBar(LyFieldInfo fieldInfo)
	{
		LyFieldSetBuilder set = (LyFieldSetBuilder)getFormFieldBuilder(fieldInfo);

		return set.updateToolBar();
	}

	/**
	 * Builds commands bar.
	 *
	 * @return the commands component
	 */
	protected LyCommands buildCommands()
	{
		LyAction action = ((LyAreaController)_parent).getAction();

		if (action.hasMark(LyAction.MARK_NO_COMMANDS))
			return null;

		LyCommands	  commands = new LyCommands();
		LyCommand	  command;
		boolean		  hasView = (_parent.getParent() instanceof LyViewController);
		boolean		  hasTarget = (((LyAreaController)_parent).getTarget() != null);
		LyEnvironment environment = _parent.getEnvironment();

		if (hasView)
		{
			command = new LyCommand(LyFormController.DEMAND_VALIDATE, "LY_VALIDATE");
			command.setAction(LyCommand.Action.SUBMIT);
			
			if (environment.checkImage("LY_COMMAND_VALIDATE"))
				command.setImage("LY_COMMAND_VALIDATE");
			
			commands.addCommand(command);
		}

		LySetController setController = (LySetController)_parent;
		LyObjectList    objects = setController.getObjectList();

		if (!hasView
				|| (objects == null)
				|| (objects.getSize() > 1)
				|| (environment.testResource("LY_NAVIGATION_IN_FORMS")
				&& environment.testResource("LY_APPLY_BUTTON")))
		{
			command = new LyCommand(LyFormController.DEMAND_APPLY, "LY_APPLY");
			command.setAction(LyCommand.Action.SUBMIT);
			
			if (environment.checkImage("LY_COMMAND_APPLY"))
				command.setImage("LY_COMMAND_APPLY");
			
			commands.addCommand(command);
		}

        if (hasView
            && environment.testResource("LY_SAVE_BUTTON"))
        {
            command = new LyCommand(LyFormController.DEMAND_SAVE, "LY_SAVE");
            command.setAction(LyCommand.Action.SUBMIT);
            
            if (environment.checkImage("LY_COMMAND_SAVE"))
                command.setImage("LY_COMMAND_SAVE");
            
            commands.addCommand(command);
        }
        
		if (hasTarget)
		{
			command = new LyCommand(LyFilterController.DEMAND_CLOSE, "LY_CLOSE");
			command.setAction(LyCommand.Action.CLOSE);
			
			if (environment.checkImage("LY_COMMAND_CLOSE"))
				command.setImage("LY_COMMAND_CLOSE");
			
			commands.addCommand(command);
		}
		else if (hasView)
		{
			command = new LyCommand(LyFormController.DEMAND_CANCEL, "LY_CANCEL");
			command.setAction(LyCommand.Action.CANCEL);
			
			if (environment.checkImage("LY_COMMAND_CANCEL"))
				command.setImage("LY_COMMAND_CANCEL");
			
			commands.addCommand(command);
		}

		command = buildHelpCommand();
		if (command != null)
			commands.addCommand(command);

		return commands;
	} // end method buildCommands
} // end class LySetBuilder
