/**
 * LyShowUrlBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyShowUrlBuilder.java 34963 2012-03-22 14:55:31Z jlsaouli $
 */

package leon.control.builder;

import leon.control.*;
import leon.view.*;


/**
 * Creates a view to show Text.
 *
 * @see LyTextController
 */
public class LyShowUrlBuilder extends LyAreaBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyShowUrlBuilder.java 34963 2012-03-22 14:55:31Z jlsaouli $";

	/** The associated form */
	protected LyForm		   _form;

	/**
	 * Default constructor.
	 *
	 * @param parent Parent controller of this builder
	 */
	public LyShowUrlBuilder(LyController parent)
	{
		super(parent);

		// Get ready to recieve content input fields

		_form = new LyForm(_parent.getId());
	}

	/**
	 * Builds the view component for the specified parent controller. The component may create the
	 * XML Tree or the internal view.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{
		/* The text view is an area which contains :
		        - A titleBar (description)
		        - A text component
		*/

		LyArea area = new LyArea();

		area.setId(_parent.getId());
		area.setController(_parent.getId());
		area.setScrolling(LyArea.Scrolling.AUTO);
		area.setTitleBar(buildTitleBar());

		LyShowUrlController parent = (LyShowUrlController)_parent;
		LyViewManager    	viewManager = parent.getViewManager();
		String			 	url = parent.getUrl();
		
		setUrl(url, area);
				
		_component = area;

		if (!viewManager.hasProperty("MANAGE_URL"))
			area.addPanel(_form);
		
		updateTitleBar();

		LyConstraint  constraint = new LyConstraint();
		
		constraint.setFill(LyConstraint.Fill.BOTH);
		area.setConstraint(constraint);

		return area;
	} // end method build

	/**
	 * Sets a new url on the main component.
	 * 
	 * @param url the new url to set
	 */
	public void setUrl(String url)
	{
		if (_component != null && _component instanceof LyArea)
			setUrl(url, (LyArea)_component);
	}
	
	/**
	 * Sets the given url (if not null) on the given area.
	 * 
	 * @param url  the new url 
	 * @param area the area
	 */
	public void setUrl(String url, LyArea area)
	{
		if (url != null && area != null)
			area.setRedirect(url);
	}
	
	/**
	 * Add a text field in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel Label of the field
	 * @param  itemLabel  Text to display
	 * @param  nCols      Number of columns of the text field
	 * @param  nRows      Number of rows of the text field
	 * @param  format     Format of the text Field (See also tag <TEXT> of leon.dtd)
	 * @param  optional   In user input mode, tells if the field is optional
	 * @return The text component created.
	 */
	public LyText setText(String fieldId,
		String fieldLabel, String itemLabel,
		int nCols, int nRows, String format, boolean optional)
	{
		LyFormField  field = new LyFormField(fieldId);

		LyConstraint constraint = new LyConstraint();

		constraint.setFill(LyConstraint.Fill.BOTH);
		constraint.setRemainder(true);

		field.setConstraint(constraint);

		if (fieldLabel != null)
		{
			LyCaption caption = new LyCaption();

			caption.setLabel(fieldLabel);
			caption.setColon(true);
			field.setCaption(caption);
		}

		// The command depends on the type of the fieldInfo
		LyText text = new LyText();

		text.setNCols((nCols <= 0) ? 20 : nCols);
		text.setNRows((nRows <= 0) ? 1 : nRows);
		text.setFormat(format);

		if (!LyText.SHOW_ARROWS)
			text.showArrows(false);

		if (itemLabel != null)
			text.setLabel(itemLabel);

		field.setFieldItem(text);
		field.setOptional(optional);

		// XXX To Remove when setEditable will be implemented
		field.setEditable(false);

		_form.addFormField(field);
		_form.setConstraint(constraint);
		
		return (text);
	} // end method setText
} // end class LyShowUrlBuilder
