/**
 * LySimpleFilterBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LySimpleFilterBuilder.java 41446 2012-11-07 10:16:19Z jlsaouli $
 */

package leon.control.builder;

import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.control.form.*;
import leon.data.*;
import leon.data.daemon.*;
import leon.info.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.view.*;


/**
 * Creates a view to type in a filter with multiple criteria.
 *
 * @see LyFilterController
 */
public abstract class LySimpleFilterBuilder extends LyFormBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LySimpleFilterBuilder.java 41446 2012-11-07 10:16:19Z jlsaouli $";

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LySimpleFilterBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Build and fill the title bar.
	 *
	 * @return the titlebar component
	 */
	protected LyTitleBar buildTitleBar()
	{
		LyAction action = ((LyFilterController)_parent).getAction();

		if (action.hasMark(LyAction.MARK_NO_TITLEBAR))
			return null;

		if ((((LyAreaController)_parent).getTarget() != null)
                || (((LyAreaController)_parent).getParent() instanceof LyCompoundFilterController))
			return null;

		LyTitleBar  titleBar = new LyTitleBar();
		LyClassInfo classInfo = ((LyFilterController)_parent).getClassInfo();
		String	    label = "LY_FILTER" + " : " + translate(classInfo.getName());

		if (_parent.getEnvironment().testResource("LY_TITLEBAR_LEFT_IMAGE"))
			titleBar.setImage(_parent.getImage(classInfo));
		titleBar.setLeftLabel(label);

		return titleBar;
	}

	/**
	 * Build the toolbar of the form.
	 *
	 * @return the built toolbar
	 */
	protected LyToolBar buildToolBar()
	{
		return null;
	}

	/**
	 * Build a form field for a given fieldInfo.
	 *
	 * @param  fieldInfo the fieldInfo corresponding to the built form field
	 * @param  editable  indicates whether the form field must be editable or not
	 * @return the form field corresponding to the fieldInfo
	 */
	public LyFormField buildFormField(LyFieldInfo fieldInfo, boolean editable)
	{
		LyFormField		 formField;
		LyCaption		 caption;
		LyFieldItem		 fieldItem;
		LyFormController formController = (LyFormController)_parent;
		LyClassInfo		 classInfo = formController.getClassInfo();
		LyObject		 object = formController.getObject();

		formField = new LyFormField(classInfo.getId()
				+ LyFormController.FORM_FIELD_SEP + fieldInfo.getId());
		formField.setEditable(editable);
		formField.setOptional(false);

		caption = new LyCaption();
		LyInfoLabel infoLabel = fieldInfo.getLabel(LyInfoLabel.Type.NAME);
		String	    label = fieldInfo.getName();

		if (infoLabel != null)
			label = infoLabel.getString(object);

		caption.setLabel(label);
		formField.setCaption(caption);

		fieldItem = buildFieldItem(fieldInfo);
		formField.setFieldItem(fieldItem);

		LyFieldUnit unit = buildUnitsList(fieldInfo);

		formField.setFieldUnit(unit);

		// Set constraints
		LyConstraint captionConstraint = getCaptionConstraint(fieldInfo, formField);

		if (captionConstraint != null)
			caption.setConstraint(captionConstraint);

		LyConstraint fieldConstraint = getFieldConstraint(fieldInfo, formField);

		if (fieldConstraint != null)
			formField.setConstraint(fieldConstraint);

		// Add toolbar when needed
		LyAreaController areaController = formController.getAreaController(fieldInfo);
		LyToolBar		 toolBar = ((areaController != null) ? null : buildToolBar(fieldInfo));

		if (toolBar != null)
			formField.setToolBar(toolBar);
		else
		{
			if ((fieldInfo instanceof LyRelationFieldInfo)
					&& !fieldInfo.isMultiple()
					&& (fieldItem instanceof LyArea))
				((LyArea)fieldItem).setScrolling(LyArea.Scrolling.NO);
		}

		// Set tip
		infoLabel = fieldInfo.getLabel(LyInfoLabel.Type.GRAMMATICAL);
		LyLabelDaemon daemon = ((infoLabel == null)
				? null : formController.getApplication().getDaemon(infoLabel));
		String		  tip, overview;

		if (infoLabel != null)
			tip = ((daemon == null) ? infoLabel.getString(object) : daemon.getString(object));
		else
			tip = fieldInfo.getTip();

		infoLabel = fieldInfo.getLabel(LyInfoLabel.Type.SEMANTIC);
		daemon = ((infoLabel == null) ? null
									  : formController.getApplication().getDaemon(infoLabel));
		if (infoLabel != null)
			overview = ((daemon == null) ? infoLabel.getString(object) : daemon.getString(object));
		else
			overview = null;

		if ((tip != null) || (overview != null))
		{
			LyTip viewTip = new LyTip();

			viewTip.setLabel(tip);
			viewTip.setOverview(overview);
			formField.setTip(viewTip);
		}

		LyFilter filter = ((LyFilterController)_parent).getDefaultFilter();
		boolean  enabled = ((filter != null) && filter.contains(fieldInfo));

		toolBar = buildToolBar(fieldInfo);
		formField.setToolBar(toolBar);
		formField.setEnabled(enabled);
		formField.setCheckBox(true);
		return formField;
	} // end method buildFormField

	/**
	 * Get caption constraint.
	 * @param fieldInfo the fieldinfo
	 * @param formField the formfield
	 * @return the constraint
	 */
	public LyConstraint getCaptionConstraint(LyFieldInfo fieldInfo,
		LyFormField formField)
	{
		LyFormController  formController = (LyFormController)_parent;
		LyAreaController  areaController = formController.getAreaController(fieldInfo);
		LyConstraint	  constraint = new LyConstraint();
		LyFieldConstraint captionConstraint = fieldInfo.getCaptionConstraint();

		if (captionConstraint != null)
		{
			constraint.setRowSpan(captionConstraint.getRowSpan());
			constraint.setColSpan(captionConstraint.getColSpan());
			constraint.setRemainder(captionConstraint.isRemainder());
			constraint.setFill(LyFormBuilder.getFill(captionConstraint.getFill()));
			constraint.setHAlign(LyFormBuilder.getHAlign(captionConstraint.getHAlign()));
			constraint.setVAlign(LyFormBuilder.getVAlign(captionConstraint.getVAlign()));
			// XX to do colWeight rowWeight
		}
		else if (areaController != null)
		{
			constraint.setColSpan(-1);
			constraint.setRemainder(true);
			constraint.setHAlign(LyConstraint.HAlign.LEFT);
		}
		else
		{
			LyFieldItem fieldItem = formField.getFieldItem();

			constraint.setHAlign(LyConstraint.HAlign.RIGHT);
			if ((!(fieldInfo instanceof LyChoiceFieldInfo)
						&& fieldInfo.isMultiple())
					|| ((fieldInfo instanceof LyTextFieldInfo)
						&& (((LyTextFieldInfo)fieldInfo).getNRows() > 1))
					|| ((fieldInfo instanceof LyChoiceFieldInfo)
						&& (((LyChoiceFieldInfo)fieldInfo).getNCols() > 0))
					|| ((fieldInfo instanceof LyFileFieldInfo)
						&& (fieldItem instanceof LyIcon))
					|| (fieldInfo instanceof LyArrayFieldInfo))
				constraint.setVAlign(LyConstraint.VAlign.TOP);
			else
				constraint.setVAlign(LyConstraint.VAlign.MIDDLE);
		}

		return constraint;
	} // end method getCaptionConstraint

	/**
	 * Get field constraint.
	 * @param fieldInfo the fieldinfo
	 * @param formField the form field
	 * @return the built constraint
	 */
	public LyConstraint getFieldConstraint(LyFieldInfo fieldInfo,
		LyFormField formField)
	{
		LyFormController  formController = (LyFormController)_parent;
		LyAreaController  areaController = formController.getAreaController(fieldInfo);
		LyFieldConstraint fieldConstraint = fieldInfo.getConstraint();
		LyConstraint	  constraint = new LyConstraint();

		if (fieldConstraint != null)
		{
			constraint.setRowSpan(fieldConstraint.getRowSpan());
			constraint.setColSpan(fieldConstraint.getColSpan());

			if (fieldConstraint.isRemainder())
				constraint.setRemainder(true);
			else if (fieldConstraint.getColSpan() == LyFieldConstraint.REMAINDER)
			{
				// Below if kept for former version compatibility.
				_parent.getEnvironment().logWarning(
					"LyFormBuilder : LyFieldConstraint.REMAINDER value for colspan is deprecated. \nUse remainder attribute on CONSTRAINT instead.");

				constraint.setRemainder(true);
			}
			else
				constraint.setRemainder(false);

			constraint.setFill(LyFormBuilder.getFill(fieldConstraint.getFill()));
			constraint.setHAlign(LyFormBuilder.getHAlign(fieldConstraint.getHAlign()));
			constraint.setVAlign(LyFormBuilder.getVAlign(fieldConstraint.getVAlign()));
		}
		else if (areaController != null)
		{
			constraint.setColWeight(100); // 100% width
			constraint.setRemainder(true);
		}
		else
		{
			LyCaption    caption = formField.getCaption();
			LyConstraint captionConstraint = caption.getConstraint();

			constraint.setHAlign(LyConstraint.HAlign.LEFT);
			if (captionConstraint != null)
				constraint.setVAlign(captionConstraint.getVAlign());

			constraint.setRemainder(true);

			if (fieldInfo.isMultiple())
			{
				if (fieldInfo instanceof LyTextFieldInfo)
					constraint.setFill(LyConstraint.Fill.BOTH);
				else if (fieldInfo instanceof LyRelationFieldInfo)
					constraint.setFill(LyConstraint.Fill.HORIZONTAL);
			}
		}

		return constraint;
	} // end method getFieldConstraint

	/**
	 * Build a field item for a given fieldInfo.
	 *
	 * @param  fieldInfo the fieldInfo corresponding to the built field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildFieldItem(LyFieldInfo fieldInfo)
	{
		LyFilter filter = ((LyFilterController)_parent).getDefaultFilter();

		if(filter != null && filter.getElements().size() == 1 &&
		    filter.getElement(0) instanceof LyFilterExpression &&
				(LyDateFieldInfo.VALUE_TODAY.equals(((LyFilterExpression)filter.getElement(0)).getInitValue())
					|| LyDateFieldInfo.VALUE_NOW.equals(((LyFilterExpression)filter.getElement(0)).getInitValue())
					|| (((LyFilterExpression)filter.getElement(0)).getInitValue() != null && ((LyFilterExpression)filter.getElement(0)).getInitValue().toString().startsWith(LyDateFieldInfo.VALUE_DAY))))
				filter = null;

		Object   value = null;

		if (filter != null)
		{
			Vector<Object> values = new Vector<Object>();

			getFilterValues(fieldInfo, filter, values);

			if (values.size() == 1)
				value = values.elementAt(0);
			else if (values.size() > 1)
				value = values;
		}

		LyFieldItem fieldItem = buildFieldItem(fieldInfo, value);

		if (fieldItem != null)
		{
			// Check if field info has a bean for filtering
			LyFieldBean bean = fieldInfo.getFieldBean(LyAction.ID_NEW_FILTER);

			if (bean != null)
			{
				fieldItem.setBean(bean.getJavaName());
				fieldItem.setDatas(bean.getDatas());
			}
		}

		return fieldItem;
	} // end method buildFieldItem

	/**
	 * Build a field item for a given fieldInfo.
	 *
	 * @param  fieldInfo the fieldInfo corresponding to the built field item
	 * @param  value     the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildFieldItem(LyFieldInfo fieldInfo, Object value)
	{
		LyFormController  formController = (LyFormController)_parent;
		LyFieldController fieldController = formController.getFieldController(fieldInfo);
		LyAreaController  areaController = ((fieldController == null)
				? null : fieldController.getAreaController());

		if (areaController != null)
		{
			LyArea area = (LyArea)areaController.getComponent();

			// From field tag --> adding a parent area to contain the area fieldItem
			if ((fieldInfo.fromFieldTag())
					&& !(fieldInfo.isMultiple()))
			{
				LyArea parentArea = new LyArea();

				parentArea.setEnvironment(area.getEnvironment());

				if (area.getBorder() == LyArea.Border.NONE)
				{
					parentArea.setLabel(null);
					parentArea.setBorder(LyArea.Border.IN);
					parentArea.setScrolling(LyArea.Scrolling.NO);

					area.setLabel(null);
					area.setScrolling(LyArea.Scrolling.NO);
					LyComponent areaChild = area.getChild(0);

					if (areaController instanceof LySetController)
						fieldController.setAreaNoScroll(areaChild);
				}
				parentArea.addPanel(area);
				return parentArea;
			}
			else
			{
				if (area.getBorder() == LyArea.Border.NONE)
				{
					area.setLabel(null);
					area.setBorder(LyArea.Border.IN);
				}
				return area;
			}
		}

		LyFieldItem fieldItem = null;

		if (fieldInfo instanceof LyTextFieldInfo)
			fieldItem = buildText((LyTextFieldInfo)fieldInfo, value);
		else if (fieldInfo instanceof LyNumberFieldInfo)
			fieldItem = buildNumber((LyNumberFieldInfo)fieldInfo, value);
		else if (fieldInfo instanceof LyDateFieldInfo)
			fieldItem = buildDate((LyDateFieldInfo)fieldInfo, value);
		else if (fieldInfo instanceof LyRelationFieldInfo)
			fieldItem = buildRelation((LyRelationFieldInfo)fieldInfo, value);
		else if (fieldInfo instanceof LyChoiceFieldInfo)
			fieldItem = buildChoice((LyChoiceFieldInfo)fieldInfo, value);
		else if (fieldInfo instanceof LyFileFieldInfo)
			fieldItem = buildFile((LyFileFieldInfo)fieldInfo, value);
		else if (fieldInfo instanceof LyArrayFieldInfo)
		{
			// If field is an array, retrieve array field info and build a value
			// with the first given value
			LyArrayFieldInfo array = (LyArrayFieldInfo)fieldInfo;

			if ((value != null) && value.getClass().isArray())
				value = ((Object[])value)[0];

			fieldItem = buildFieldItem(array.getFieldInfo(), value);
		}
		else
		{
			_parent.getEnvironment().logWarning("FormBuilder : Unable to build fieldInfo : "
				+ fieldInfo + " - value : " + value);
			return null;
		}

		// Check if field info has a bean
		LyAction action = ((LyAreaController)_parent).getAction();
		LyAction rootAction = action.getRoot();
		String   rootActionType = rootAction.getId();
		boolean  editable = (((formController instanceof LySetController)
						&& !((LySetController)formController).isConsult(fieldInfo))
					|| (formController instanceof LyFilterController));

		// FFT01204 : if field is not editable, use consult bean
		if (!editable)
			rootActionType = LyAction.ID_CONSULT;

		LyFieldBean bean = fieldInfo.getFieldBean(rootActionType);

		if (bean != null)
		{
			fieldItem.setBean(bean.getJavaName());
			fieldItem.setDatas(bean.getDatas());
		}

		fieldItem.setFieldValue(value);
		return fieldItem;
	} // end method buildFieldItem

	/**
	 * Gets all filter values corresponding to the given field info.
	 *
	 * @param fieldInfo A field info.
	 * @param filter    A filter.
	 * @param result    result list.
	 */
	protected void getFilterValues(LyFieldInfo fieldInfo, LyFilter filter, Vector<Object> result)
	{
		if (filter == null)
			return;

		filter.getValues(fieldInfo, result);
	}

	/**
	 * Build a relation that contains all units of the given field.
	 * @param  fieldInfo FieldInfo for which a units list is required.
	 * @return The relation containing all possible units for the given field info.
	 */
	public LyFieldUnit buildUnitsList(LyFieldInfo fieldInfo)
	{
		if (fieldInfo == null)
			return null;
		
		if ((fieldInfo instanceof LyDateFieldInfo)
				&& ((LyDateFieldInfo)fieldInfo).isWithTimeZone())
		{
			LyFieldUnit unit = new LyFieldUnit();

			LyApplication application = fieldInfo.getApplication();
			String timeZones[] = application.getSupportedTimeZones();
			String userTz = getParent().getEnvironment().getEnv("LY_TIMEZONE");
			LyRelation relation = new LyRelation();
			LyLink link;
			int i, n = (timeZones == null ? 0 : timeZones.length);
			for (i = 0; i < n; i++)
			{
				link = new LyLink(timeZones[i], timeZones[i]);
				if (timeZones[i].equals(userTz))
					link.setSelected(true);
				else
				link.setSelected(false);
				relation.addLink(link);
			}
			unit.setFieldItem(relation);
			return unit;
		}

		LyUnits			  units = fieldInfo.getUnits();
		LyUnit			  fieldInfoUnit = ((LySimpleFilterController)_parent).getDefaultFieldUnit(fieldInfo);

		if (fieldInfo instanceof LyStructFieldInfo)
			return null;
		if (units == null)
			return null;

		int		    nbUnits = units.getUnitCount();
		LyFieldUnit fieldUnit = new LyFieldUnit();
		String	    name, unitName, unitAlias;

		if (nbUnits == 1)
		{
			LyLabel label = new LyLabel();

			unitAlias = fieldInfoUnit.getAlias();
			unitName = fieldInfoUnit.getName();
            if ((unitAlias != null) && (unitAlias.length() > 0))
                name = unitAlias;
            else if ((unitName != null) && (unitName.length() > 0))
                name = unitName;
            else
                name = null;

			if (name != null)
				label.setLabel(" " + name);
			fieldUnit.setFieldItem(label);
			return fieldUnit;
		}

		LyRelation relation = new LyRelation();
		int		   i;
		LyUnit     unit;
		LyLink     link;

		for (i = 0; i < nbUnits; i++)
		{
			unit = units.getUnit(i);

			unitAlias = unit.getAlias();
			unitName = unit.getName();
			if (((unitAlias != null) && (unitAlias.length() > 0))
					&& ((unitName != null) && (unitName.length() > 0)))
				name = unitAlias + " : " + unitName;
			else if ((unitName != null) && (unitName.length() > 0))
				name = unitName;
			else if ((unitAlias != null) && (unitAlias.length() > 0))
				name = unitAlias;
			else
				name = null;

			link = new LyLink(unit.getId(), name);
			relation.addLink(link);
			if (unit == fieldInfoUnit)
				relation.selectLink(i, true);
		}

		fieldUnit.setFieldItem(relation);
		return fieldUnit;
	} // end method buildUnitsList


	/**
	 * Build a field item for a text fieldInfo.
	 *
	 * @param  textField the text fieldInfo corresponding to the built field item
	 * @param  value     the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildText(LyTextFieldInfo textField, Object value)
	{
		LyText  text = new LyText();
		boolean multiple = textField.isMultiple();
		int     n = textField.getNCols();
		boolean editor = (_parent.getParent() instanceof LyFilterEditorController);

		if (multiple) // FFT00058 : allow filter on multiple text field
			text.setNRows(3);

		// If filter is displayed in a filter editor, do not authorize text greater
		// than 60 to avoid the tree part to be hidden
		if (editor)
			n = Math.min(n, 60);

		text.setNCols((n <= 0) ? 20 : n);

		if (value != null)
			text.setLabel(value.toString());

		return text;
	}

	/**
	 * Build a field item for a date fieldInfo.
	 *
	 * @param  dateField the date fieldInfo corresponding to the built field item
	 * @param  value     the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildDate(LyDateFieldInfo dateField, Object value)
	{
		String				 label = ((value == null) ? null : dateField.toString(value));
		LyDateFieldInfo.Type type = dateField.getType();
		String				 format = dateField.getFormat();
		LyDate				 date = new LyDate();

		if (!LyDate.SHOW_ARROWS)
			date.showArrows(false);

		date.setType(LyDate.getType(type.ordinal()));
		date.setLabel(label);
		date.setFormat(format);

		return date;
	}

	/**
	 * Build a field item for a number fieldInfo.
	 *
	 * @param  numberField the number fieldInfo corresponding to the built field item
	 * @param  value       the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	@SuppressWarnings("deprecation")
	public LyFieldItem buildNumber(LyNumberFieldInfo numberField, Object value)
	{
		String   format = numberField.getFormat();
		Number   vmin = numberField.getMin();
		Number   vmax = numberField.getMax();
		Number   pitch = numberField.getPitch();
		int		 decimals = numberField.getDecimals();
		LyUnit   fieldUnit = ((LySimpleFilterController)_parent).getDefaultFieldUnit(numberField);
		String   unit = ((fieldUnit == null) ? numberField.getUnit() : fieldUnit.getName());
		LyNumber number = new LyNumber();

		if (!LyNumber.SHOW_ARROWS)
			number.showArrows(false);

		if (format != null)
			number.setFormat(format);
		if (vmin != null)
			number.setVMin(vmin.doubleValue());
		if (vmax != null)
			number.setVMax(vmax.doubleValue());
		if (pitch != null)
			number.setPitch(pitch.doubleValue());
		number.setDecimals(decimals);
		if ((unit != null) && (fieldUnit == null))
			number.setUnit(unit);

		if (value != null)
			number.setLabel(value.toString());

		return number;
	} // end method buildNumber

	/**
	 * Build a field item for a relation fieldInfo.
	 *
	 * @param  relationField the relation fieldInfo corresponding to the built field item
	 * @param  value         the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildRelation(LyRelationFieldInfo relationField, Object value)
	{
        if (((LyFilterController) _parent).isComposition(relationField))
        {
            LyLabel   label = new LyLabel();

            label.setLabel("LY_FILTER_COMPOSITION_EMPTY_VALUE");
            return label;
        }

		LyRelation relation = new LyRelation();

        if (_parent instanceof LyMultiFilterController)
            relation.setMultiple(true);
        else
            relation.setMultiple(relationField.isMultiple());

		// Get relation domain and selected objects
		LyObjectList list = null, tempo_list;
		Object		 relationValue;

		if ((value != null) && (value instanceof Vector<?>))
		{
			Vector<?> vect = (Vector<?>)value;

			for (int vv = 0; vv < vect.size(); vv++)
			{
				relationValue = vect.elementAt(vv);
				tempo_list = relationField.getObjects(relationValue);
				if ((tempo_list == null)
						|| (tempo_list.getSize() == 0))
					continue;
				if (list == null)
					list = tempo_list;
				else
					list.addList(tempo_list);
			}
		}
		else
			list = relationField.getObjects(value);

		LyFilterBehaviorInterface behavior = (LyFilterBehaviorInterface)((LyFilterController)_parent).getBehavior();
		LyObjectList			  domain = behavior.getFilterRelationDomain(((LyFilterController)_parent), relationField);

		// FFT00251 : sort links as in the set and creation forms
		if (domain != null)
			domain.setSortByName(true);

		int			  i, n = ((domain == null) ? 0 : domain.getSize());
		LyLink		  link;
		boolean		  selected;

		// Add images only if multiple classes
		int			  nClasses = relationField.getClassInfoCount();
		LyEnvironment environment = _parent.getEnvironment();
		boolean		  addImage = (nClasses >= environment.getIntResource("LY_SHOW_RELATION_IMAGES"));
		int			  nList = ((list == null) ? 0 : list.getSize());
		LyLink[]	  links = new LyLink[nList];
		LyObject	  object;
		int			  pos;

		if  (!relation.isMultiple())
		{
	        String empty_value = "LY_RELATION_EMPTY_VALUE";
            String val = environment.translate(empty_value);

            if ((val != null) && !val.equals("_null"))
            {
                if ((val != null)
                        && val.equals("_null"))
                    empty_value = null;

                link = new LyLink("LY_NO_CHOICE", empty_value);
                relation.addLink(link);
            }
		}

		for (i = 0; i < n; i++)
		{
			object = domain.getObject(i);
			selected = ((list != null) && list.contains(object));

			if (relationField.hasMark(LyFieldInfo.MARK_TRANSLATE))
			    link = new LyLink(object.getId(), getName(object));
			else
			    link = new LyLink(object.getId(), "$$" + getName(object));

			if (selected)
			{
				pos = list.getIndex(object);
				links[pos] = link;
			}

			if (addImage)
				link.setImage(_parent.getImage(object));

			// FFT00252 : keep order in relation
			if (!selected)
				relation.addLink(link);
		}

		// FFT00252 : keep order in relation
		for (i = 0; i < nList; i++)
		{
			if (links[i] != null)
			{
				relation.addLink(links[i]);
				links[i].setSelected(true);
			}
		}

		return relation;
	} // end method buildRelation

	/**
	 * Builds the toolbar for the given field.
	 *
	 * @param  fieldInfo the field
	 * @return the toolbar for the given field
	 */
	public LyToolBar buildToolBar(LyFieldInfo fieldInfo)
	{
		if ((fieldInfo instanceof LyRelationFieldInfo)
				&& ((LyRelationFieldInfo)fieldInfo).isToolbarDisabled())
			return null;

		// Add toolbar for transverse actions if needed
		LyFilterBehaviorInterface			  behavior = (LyFilterBehaviorInterface)_parent
				.getBehavior();
		LyInfoDataList<LyAction, LyClassInfo> actions = behavior.getFilterTransverseActions(
				(LyFilterController)_parent, fieldInfo);

		return buildToolBar(fieldInfo, actions);
	}

	/**
	 * Builds the toolbar for the given relation field and list of actions.
	 *
	 * @param  fieldInfo the field
	 * @param  actions       List of pairs (action, classInfo)
	 * @return the toolbar for the given relation field
	 */
	public LyToolBar buildToolBar(LyFieldInfo fieldInfo,
		LyInfoDataList<LyAction, LyClassInfo> actions)
	{
		if ((fieldInfo instanceof LyRelationFieldInfo)
				&& ((LyRelationFieldInfo)fieldInfo).isToolbarDisabled())
			return null;

		int   n = ((actions == null) ? 0 : actions.getSize());

		if (n == 0)
			return null;

		int		  i;
		LyAction  action;
		LySession session = _parent.getSession();

		// for each action, check if it is valid
		for (i = n - 1; i >= 0; i--)
		{
			action = actions.getInfo(i);
			if (!session.validateAction(action, _parent))
				// the current action is not valid, remove it from the list
				actions.removeInfo(action);
		}

		n = ((actions == null) ? 0 : actions.getSize());
		if (n == 0)
			return null;

		int			  j;
		LyAction	  action2;
		LyToolBar     toolBar = new LyToolBar();
		LyTool		  tool = null;
		// Keep subtoolbars table while building transverse actions
		LyToolBar     subToolBar;
		LyArea		  area;
		LyClassInfo   classInfo, classInfo2;
		LyEnvironment environment = _parent.getEnvironment();

		toolBar.setController(_parent.getId());
		if (fieldInfo.isMultiple())
			toolBar.setFill(LyToolBar.Fill.VERTICAL);
		else
			toolBar.setFill(LyToolBar.Fill.HORIZONTAL);

		for (i = 0; i < n; i++)
		{
			action = actions.getInfo(i);
			classInfo = actions.getData(i);

			// Check if we have multiple occurences of the same action image
			// if so, build a tool with a sub menu
			subToolBar = null;
			if (!(fieldInfo instanceof LyRelationFieldInfo)
					|| (((LyRelationFieldInfo)fieldInfo).getClassField() == null)
					|| ignoreRelationClassField())
			{
				for (j = i + 1; j < n; j++)
				{
					action2 = actions.getInfo(j);

					if (_parent.getImage(action).equals(_parent.getImage(action2))
							&& environment.testResource("LY_GROUP_ACTIONS_BY_IMAGE"))
					{
						if (subToolBar == null)
						{
							subToolBar = new LyToolBar();
							subToolBar.setController(_parent.getId());
							subToolBar.setFill(LyToolBar.Fill.VERTICAL);
							tool = buildTool(action, true,
									classInfo.getId() + LyController.ID_SEP + fieldInfo.getId(), null);
							tool.setLabel(action.getAlias() + " : " + classInfo.getAlias());
							tool.setImage(_parent.getImage(classInfo));
							subToolBar.addTool(tool);
						}

						classInfo2 = actions.getData(j);
						if (classInfo2 != null)
						{
							tool = buildTool(action2, true,
									classInfo2.getId() + LyController.ID_SEP + fieldInfo.getId(), null);
							tool.setLabel(action2.getAlias() + " : " + classInfo2.getAlias());
							tool.setImage(_parent.getImage(classInfo2));
						}
						else
							tool = buildTool(action2, true,
									LyFieldSetController.STANDARD_CLASSINFO + LyController.ID_SEP + fieldInfo.getId(), null);

						subToolBar.addTool(tool);
						actions.removeInfo(action2, classInfo2);
						j--;
						n--;
					}
				}

				if (classInfo != null)
					tool = buildTool(action, true,
							classInfo.getId() + LyController.ID_SEP + fieldInfo.getId(),
							environment.getImageSizeInfo(DEFAULT_TRANSVERSE_ICON_WIDTH, DEFAULT_TRANSVERSE_ICON_WIDTH));
				else
					tool = buildTool(action, true,
							LyFieldSetController.STANDARD_CLASSINFO + LyController.ID_SEP + fieldInfo.getId(),
							environment.getImageSizeInfo(DEFAULT_TRANSVERSE_ICON_WIDTH, DEFAULT_TRANSVERSE_ICON_WIDTH));
			}
			else
			{
				for (j = i + 1; j < n; j++)
				{
					action2 = actions.getInfo(j);
					if (_parent.getImage(action).equals(_parent.getImage(action2)))
					{
						actions.removeInfo(action2, actions.getData(j));
						j--;
						n--;
					}
				}

				tool = buildTool(action, true,
						LyFieldSetController.STANDARD_CLASSINFO + LyController.ID_SEP + fieldInfo.getId(),
						environment.getImageSizeInfo(DEFAULT_TRANSVERSE_ICON_WIDTH, DEFAULT_TRANSVERSE_ICON_WIDTH));
			}

			toolBar.addTool(tool);
			if (subToolBar != null)
			{
				area = new LyArea();
				area.setToolBar(subToolBar);
				tool.setComponent(area);
			}
			else if (environment.testResource("LY_WRITE_FIELD_TOOL_LABEL"))
			{
				if (classInfo != null)
					tool.setLabel(action.getAlias() + LyFieldSetController.TRANSVERSE_ACTION_SEP + classInfo.getAlias());
				else
					tool.setLabel(action.getAlias());
			}
		}

		return toolBar;
	} // end method buildToolBar

    /**
     * Builds a tool description for a specified action.
     *
     * @param  action       the action for the tool
     * @param  enable       enabled state for the tool
     * @param  parameter    parameters for the associated action
     * @param  iconSizeInfo size of the associated icon
     * @return the tool
     */
    protected LyTool buildTool(LyAction action, boolean enable, String parameter, String iconSizeInfo)
    {
        String toolId = action.getId();

        if (parameter != null)
            toolId += LyController.ID_SEP + parameter;

        LyEnvironment environment = _parent.getEnvironment();
        boolean       showLabel = environment.testResource("LY_WRITE_FIELD_TOOL_LABEL");
        LyTool        tool = new LyTool(toolId, (showLabel ? action.getAlias() : null));
        String        iconPath = _parent.getImage(action);

        if (iconSizeInfo != null)
            iconPath = iconPath + iconSizeInfo;

        tool.setTooltip(action.getName());
        tool.setImage(iconPath);
        tool.setEnabled(enable);
        tool.setDialog(action.hasMark(LyAction.MARK_DIALOG));

        return tool;
    } // end method buildTool

	/**
	 * Build a field item for an array fieldInfo.
	 *
	 * @param  arrayField the array fieldInfo corresponding to the built field item
	 * @param  value      the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildArray(LyArrayFieldInfo arrayField, Object value)
	{
		LyTable table = new LyTable(arrayField.getId());
		LyRows  rows = new LyRows();
		int     nrows = arrayField.getNRows();

		if (nrows <= 0)
			nrows = 3;

		table.setSelectionMode(LyTable.SelectionMode.NONE);
		table.setRows(rows);
		table.setNRows(nrows);

		return table;
	} // end method buildArray

	/**
	 * /** Build a field item for a choice fieldInfo.
	 *
	 * @param  choiceField the choice fieldInfo corresponding to the built field item
	 * @param  value       the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildChoice(LyChoiceFieldInfo choiceField, Object value)
	{
		LyChoiceOption		   choiceOption;
		boolean				   checked = false;
		boolean				   enable = true;
		Vector<LyChoiceOption> selectedChoiceOptions = new Vector<LyChoiceOption>();

		//For update checked option
		// get selected options
		if (value != null)
		{
			if (value instanceof Vector<?>)
			{
				Vector<?> v = (Vector<?>)value;
				Short  s;
				int    z, n = v.size();

				for (z = 0; z < n; z++)
				{
					s = (Short)v.elementAt(z);
					if (s != null)
					{
						choiceOption = choiceField.getChoiceOption(s.shortValue());
						if (choiceOption != null)
							selectedChoiceOptions.addElement(choiceOption);
					}
				}
			}
			else if (value instanceof BitSet)
			{
				BitSet bs = (BitSet)value;
				int    kk, bsSize = bs.size();

				for (kk = 0; kk < bsSize; kk++)
				{
					if (bs.get(kk))
					{
						choiceOption = choiceField.getChoiceOption((short)kk);
						if (choiceOption != null)
							selectedChoiceOptions.addElement(choiceOption);
					}
				}
			}
			else if (value instanceof Boolean)
			{
				short index = 0;
				if (((Boolean)value).booleanValue())
					index = 1;

				choiceOption = choiceField.getChoiceOption(index);

				if (choiceOption != null)
					selectedChoiceOptions.addElement(choiceOption);
			}
			else
			{
				choiceOption = choiceField.getChoiceOption(((Short)value).shortValue());

				if (choiceOption != null)
					selectedChoiceOptions.addElement(choiceOption);
			}
		}

		int		nCols = choiceField.getNCols();

		LyFilterBehaviorInterface behavior = (LyFilterBehaviorInterface)
			((LyFilterController)_parent).getBehavior();
		Vector<LyChoiceOption>	  domain = behavior.getFilterChoiceDomain(((LyFilterController)
					_parent), choiceField);
		int						  i, n = ((domain == null) ? 0 : domain.size());

		if (nCols != 0)
		{
			LyChoice choice = new LyChoice();
			LyOption option;

			choice.setMultiple(choiceField.getType() != LyChoiceFieldInfo.Type.BOOLEAN);

			if (nCols < 0)
				nCols = 5;

			choice.setMaxSpan(nCols);

			for (i = 0; i < n; i++)
			{
				choiceOption = (LyChoiceOption)domain.elementAt(i);
				checked = (selectedChoiceOptions.contains(choiceOption));

				option = new LyOption(choiceOption.getId(), translate(choiceOption.getAlias()));
				option.setChecked(checked);
				option.setEnabled(enable);

				choice.addOption(option);
			}

			if (choiceField.hasMark(LyFieldInfo.MARK_OPTIONAL) && (choiceField.getType() == LyChoiceFieldInfo.Type.BOOLEAN))
			{
				option = new LyOption("LY_NO_CHOICE", "LY_NO_CHOICE");
				option.setChecked(value == null);
				choice.addOption(option);
			}

			return choice;
		}
		else
		{
			LyRelation relation = new LyRelation();
			LyLink     link;

			relation.setMultiple(true);

			for (i = 0; i < n; i++)
			{
				choiceOption = (LyChoiceOption)domain.elementAt(i);
				checked = (selectedChoiceOptions.contains(choiceOption));

				link = new LyLink(choiceOption.getId(), translate(choiceOption.getName()));
				link.setSelected(checked);

				relation.addLink(link);
			}

			return relation;
		}
	} // end method buildChoice

	/**
	 * Translate the values displayed in filter form.
	 *
	 * @param value the value that must be translated
	 * @return the value that must be displayed
	 */
	public String translate(String value)
	{
		boolean editor = (_parent.getParent() instanceof LyFilterEditorController);

		if (editor)
			return ((LyFilterEditorController)_parent.getParent()).translate(value);

		return value;
	}

	/**
	 * Build a field item for a file fieldInfo.
	 *
	 * @param  fileField the file fieldInfo corresponding to the built field item
	 * @param  value     the value of the field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildFile(LyFileFieldInfo fileField, Object value)
	{
		String mimeTypes = fileField.getMimeTypes();
		if((mimeTypes != null) && (mimeTypes.equals("text/x-url")))
		{
			LyText text = new LyText();
			text.setWrap(false);

			if (value != null)
				text.setLabel(value.toString());

			return text;
		}

		LyFile file = new LyFile();

		if (value != null)
			file.setLabel(value.toString());

		return file;
	}

	/**
	 * Creates the command bar and sets its content.
	 *
	 * @return the commands bar component
	 */
	protected LyCommands buildCommands()
	{
        LyAction action = ((LyAreaController)_parent).getAction();

        if (action.hasMark(LyAction.MARK_NO_COMMANDS))
            return null;

        if (((LyAreaController)_parent).getParent() instanceof LyCompoundFilterController)
            return null;

		LyCommands    commands = new LyCommands();
		LyCommand     command;
		LyEnvironment environment = _parent.getEnvironment();
		LyViewManager viewManager = getViewManager();
		boolean		  haveValidateButton = false;
		boolean		  hasView = (_parent.getParent() instanceof LyViewController);
		boolean		  hasTarget = (((LyAreaController)_parent).getTarget() != null);
		boolean		  editor = (_parent.getParent() instanceof LyFilterEditorController);

		if (hasView || !_parent.getEnvironment().testResource("LY_EXTENDED_FILTER"))
		{
			command = new LyCommand(LyFormController.DEMAND_VALIDATE, "LY_VALIDATE");
			command.setAction(LyCommand.Action.SUBMIT);

			if (environment.checkImage("LY_COMMAND_VALIDATE"))
				command.setImage("LY_COMMAND_VALIDATE");

			commands.addCommand(command);
			haveValidateButton = true;
		}

		if (((environment.testResource("LY_APPLY_BUTTON")
						&& (((LyFilterController)_parent).getFilterEndInterface() != null))
					|| !hasView))
		{
			if ((haveValidateButton == false) || !viewManager.hasProperty("NOT_APPLY"))
			{
				command = new LyCommand(LyFormController.DEMAND_APPLY, "LY_APPLY");
				command.setAction(LyCommand.Action.SUBMIT);

				if (environment.checkImage("LY_COMMAND_APPLY"))
					command.setImage("LY_COMMAND_APPLY");

				commands.addCommand(command);
			}
		}

		if (!editor
            && (hasView || hasTarget))
		{
			command = new LyCommand(LyFormController.DEMAND_CANCEL, "LY_CANCEL");
			command.setAction(LyCommand.Action.CANCEL);

			if (environment.checkImage("LY_COMMAND_CANCEL"))
				command.setImage("LY_COMMAND_CANCEL");

			commands.addCommand(command);
		}

		command = buildHelpCommand();

		if (command != null)
			commands.addCommand(command);

		return commands;
	} // end method buildCommands

	/**
	 * Tells if we must ignore the class fields of relation while building toolbars.
	 *
	 * @return true;
	 */
	boolean ignoreRelationClassField()
	{
		return true;
	}
	/**
	 * Add a filter type field to a form
	 * @param form the form
	 * @param extended true if we are in an extended filter
	 * @return the field
	 */
    protected LyFormField addFilterTypeField(LyForm form, boolean extended)
    {
        if (!(_parent.getParent() instanceof LyCompoundFilterController))
            return null;

        LyEnvironment   environment = _parent.getEnvironment();

        if (!environment.testResource("LY_EXTENDED_FILTER"))
            return null;

        // add the operation choice
        LyFormField     field;
        LyCaption       caption;
        LyChoice        choice = new LyChoice();
        LyOption        option;

        field = new LyFormField(LyCompoundFilterController.SELECT_TYPE);

        caption = new LyCaption();
        caption.setLabel("LY_FILTER_TYPE");
        caption.setColon(true);

        option = new LyOption(LyCompoundFilterController.ID_SIMPLE_FILTER, "LY_SIMPLE_FILTER");
        option.setChecked(!extended);
        choice.addOption(option);
        option = new LyOption(LyCompoundFilterController.ID_EXTENDED_FILTER, "LY_EXTENDED_FILTER");
        option.setChecked(extended);
        choice.addOption(option);

        field.setCaption(caption);
        field.setFieldItem(choice);
        field.setOptional(false);
        field.setNeedCheck(true);
        field.setController(_parent.getParent().getId());

        form.addFormField(field);
        return field;
    }
} // end class LySimpleFilterBuilder
