/**
 * LySimpleTableBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LySimpleTableBuilder.java 36542 2012-08-21 15:24:36Z jlsaouli $
 */

package leon.control.builder;

import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.control.form.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;

import java.util.*;


/**
 * This builder builds the basic view for a simple table controller.
 *
 * @see LySimpleTableController
 */
public abstract class LySimpleTableBuilder extends LyPanelBuilder
{

    /**
     * File version
     */
    public static final String __VERSION =
        "$Id: LySimpleTableBuilder.java 36542 2012-08-21 15:24:36Z jlsaouli $";

    /** The table component. */
    protected LyTable _table;

    /**
     * Default constructor.
     *
     * @param parent parent controller
     */
    public LySimpleTableBuilder(LyController parent)
    {
        super(parent);
    }

    /**
     * Builds the view component for the specified parent controller. The component may create the
     * XML Tree or the internal view.
     *
     * @return the view Component
     */
    public LyComponent build()
    {

        /* The table view is an area which contains :
                - A toolbar (actions)
                - A titleBar (description)
                - A table component
        */
        LyArea area = new LyArea();

        area.setId(_parent.getId());
        area.setController(_parent.getId());
        area.setScrolling(LyArea.Scrolling.AUTO);

        // Should be parametrized in a property file
        if ("w4".equals(_parent.getEnvironment().getSkin())
                && getViewManager().hasProperty("WEB_MANAGER"))
        {
            LyConstraint constraint = new LyConstraint();

            constraint.setFill(LyConstraint.Fill.HORIZONTAL);
            area.setConstraint(constraint);
            area.setBorder(LyArea.Border.IN);
        }

        // set content
        _table = buildTable();
        area.addPanel(_table);

        LyConstraint constraint = new LyConstraint();

        constraint.setFill(LyConstraint.Fill.BOTH);
        area.setConstraint(constraint);
        _table.setConstraint(constraint);
        _component = area;
        return area;
    }

    /**
     * Builds the toolbar of the table view.
     *
     * @return the toolbar Component
     */
    protected LyToolBar buildToolBar()
    {
        LyAction action = ((LyGetActionInterface) _parent).getAction();

        if (action.hasMark(LyAction.MARK_NO_TOOLBAR))
            return null;


        LyClassInfo classInfo = ((LyGetClassInfoInterface) _parent).getClassInfo();
        LyToolBar   toolBar = buildToolBar(classInfo);

        return completeToolBar(toolBar);
    }

    /**
     * @see leon.control.builder.LyAreaBuilder#getToolbarActions(leon.control.LyAreaController, leon.info.LyInfo)
     */
    protected LyInfoDataList<LyAction, LyClassInfo> getToolbarActions(LyAreaController controller,
        LyInfo info)
    {
        return getMenuActions(controller, info, true);
    }

    /**
     * Get the action
     * @param controller the parent controller
     * @param info an info
     * @param checkTool true to check tool
     * @return a list of action
     */
    protected LyInfoDataList<LyAction, LyClassInfo> getMenuActions(LyAreaController controller,
        LyInfo info, boolean checkTool)
    {

        if ((info != null) && (info instanceof LyClassInfo))
        {
            LyClassInfo             classInfo = (LyClassInfo) info;
            LyInfoList<LyClassInfo> subClasses = null;
            LyEnvironment           environment = _parent.getEnvironment();
            LySimpleTableController tableController = (LySimpleTableController) _parent;
            LyController            parent = tableController.getParent();
            LyFieldInfo             fieldInfo = null;

            if ((parent != null) && (parent instanceof LyCompositionArrayFieldSetController))
                fieldInfo = ((LyCompositionArrayFieldSetController) parent).getFieldInfo();
            else if ((parent != null) && (parent instanceof LyArrayFieldSetController))
                fieldInfo = ((LyArrayFieldSetController) parent).getFieldInfo();

            if ((fieldInfo != null) && (fieldInfo instanceof LyStructFieldInfo)
                    && (((LyStructFieldInfo) fieldInfo).fromFieldTag()))
            {
                LyStructFieldInfo       struct = (LyStructFieldInfo) fieldInfo;
                LyInfoList<LyClassInfo> classes = struct.getClassInfos();

                subClasses = new LyInfoList<LyClassInfo>(classes);
            }
            else
            {
                subClasses = classInfo.getSubClasses(true);

                // Sort subClasses by name
                if (subClasses == null)
                    subClasses = new LyInfoList<LyClassInfo>();
                else
                    subClasses = subClasses.getInfosByName(environment);

                // Put the classInfo in the first position
                subClasses.addInfo(classInfo, 0);
            }

            int                                   i, j, ns = subClasses.getSize(), na;
            LyInfoList<LyAction>                  list;
            LyAction                              action;
            LyClassInfo                           subClass;
            LyInfoDataList<LyAction, LyClassInfo> actionsList =
                new LyInfoDataList<LyAction, LyClassInfo>(true);
            boolean                               onlyAbstract = false;

            // Get all actions for all subclasses.
            for (i = 0; i < ns; i++)
            {
                subClass = subClasses.getInfo(i);

                if (subClass.isAbstract())
                    onlyAbstract = true;
                else
                    onlyAbstract = false;

                list = subClass.getActions();

                for (j = 0, na = list.getSize(); j < na; j++)
                {
                    action = list.getInfo(j);

                    // For abstract classes, only keep abstract actions
                    if (onlyAbstract && !action.isAbstract())
                        continue;

                    if (!validateAction(action, checkTool))
                        continue;

                    actionsList.addInfo(action, subClass);
                }
            }

            return actionsList;
        }

        return super.getToolbarActions(controller, info);
    } // end method getToolbarActions

    /**
     * Build a toolbar from a classInfo
     * @param classInfo the classInfo
     * @return a toolbar
     */
    protected LyToolBar buildToolBar(LyClassInfo classInfo)
    {
        LyInfoDataList<LyAction, LyClassInfo> actionsList = getToolbarActions((LyAreaController) _parent, classInfo);
        LyToolBar                             toolBar = new LyToolBar();
        
        toolBar.setId("_toolBar");
        
        if (actionsList == null)
        	return toolBar;

        LyEnvironment                         environment = _parent.getEnvironment();
        LyAction                              action, action2;
        LyToolBar                             subToolBar;
        LyTool                                tool;
        LyArea                                area;
        LyClassInfo                           classInfo2, classInfo3;
        String                                group;
        Hashtable<String, LyToolBar>          groupToolsBars = new Hashtable<String, LyToolBar>();
        int                                   min;
        LyInfoList<LyClassInfo>				  subClasses = classInfo.getSubClasses();

        toolBar.setFill(LyToolBar.Fill.HORIZONTAL);
        toolBar.setColor("LY_DASHBOARD_HORIZONTAL_BACKGROUND");

        boolean multiple;
        String  image;
        int     i, j;
        boolean testGroups;

        // Now, we can add the rest of the actions from the subclasses
        // If we find the same action more than once, build a toolbar if min == 0, else
        // just add tool
        for (i = 0; i < actionsList.getSize(); i++)
        {
            action = actionsList.getInfo(i);
            classInfo3 = actionsList.getData(i);
            min = action.getMin();
            subToolBar = null;
            multiple = false;

            for (j = i + 1; j < actionsList.getSize(); j++)
            {
                action2 = actionsList.getInfo(j);
                testGroups = false;

                // FFT 00520 FLANDRES
                if ((action.getGroup() != null) && (action2.getGroup() != null))
                {

                    if (action.getGroup().equals(action2.getGroup()))
                        testGroups = true;
                }
                else if ((action.getGroup() != null) || (action2.getGroup() != null))
                    testGroups = false;
                else
                    testGroups = true;

                if ((action == action2)
                        || ((action.getRoot() == action2.getRoot())
                            && (min == action2.getMin()) // FFT01229 : some actions may have the same root but different MIN
                            && (_parent.getImage(action).equals(_parent.getImage(action2)))
                            && environment.testResource("LY_GROUP_ACTIONS_BY_IMAGE")
                            && testGroups))
                {
                    classInfo2 = actionsList.getData(j);

                    if (((classInfo2 == classInfo) || (min == 0))
                            && !action.hasMark(LyAction.MARK_NO_CONTEXT)
                            && acceptSubAction(action, classInfo3, action2, classInfo2))
                    {

                        if (subToolBar == null)
                        {
                            subToolBar = new LyToolBar();
                            subToolBar.setController(_parent.getId());

                            if (environment.testResource("LY_WRITE_TOOL_LABEL"))
                                subToolBar.setFill(LyToolBar.Fill.VERTICAL);
                            else
                                subToolBar.setFill(LyToolBar.Fill.HORIZONTAL);

                            toolBar.setColor("LY_DASHBOARD_HORIZONTAL_BACKGROUND");
                            tool = buildTool(action, true, classInfo3.getId(), null);

                            if (classInfo3 != classInfo)
                            {

                                if (action.getAlias().equals(action2.getAlias()))
                                    tool.setLabel(action.getAlias() + " : "
                                        + classInfo3.getAlias());

                                if (action.getName().equals(action2.getName()))
                                    tool.setTooltip(action.getName() + " : "
                                        + classInfo3.getName());

                                image = _parent.getImage(classInfo3);

                                if (image != null)
                                    tool.setImage(image);
                            }

                            subToolBar.addTool(tool);
                        }

                        tool = buildTool(action2, true, classInfo2.getId(), null);

                        if (classInfo2 != classInfo)
                        {

                            if (action.getAlias().equals(action2.getAlias()))
                                tool.setLabel(action2.getAlias() + " : " + classInfo2.getAlias());

                            if (action.getName().equals(action2.getName()))
                                tool.setTooltip(action2.getName() + " : " + classInfo2.getName());

                            image = _parent.getImage(classInfo2);

                            if (image != null)
                                tool.setImage(image);
                        }

                        subToolBar.addTool(tool);
                    }

                    actionsList.removeInfo(action2, classInfo2);
                    multiple = true;
                    j--;
                }
            }

            if (subToolBar != null)
            {
                StringBuffer toolId = new StringBuffer();
                LyAction     template = (LyAction) action.getTemplate();

                if (template == null)
                    template = action;

                toolId.append(LySimpleTableController.DEMAND_SUB_TOOL_ACTION);
                toolId.append(LyController.ID_SEP);
                toolId.append(template.getId());
                toolId.append(_parent.getImage(template));

                tool = new LyTool(toolId.toString(), template.getAlias());
                tool.setTooltip(template.getName());
                tool.setImage(_parent.getImage(template));
                tool.setEnabled(true);
                tool.setDialog(template.hasMark(LyAction.MARK_DIALOG));

                area = new LyArea();
                area.setToolBar(subToolBar);
                tool.setComponent(area);
            }
            else
            {

                if (multiple)
                    tool = buildTool(action, true, null, null);
                else
                {
                    tool = buildTool(action, true, classInfo3.getId(), null);

                    if ((classInfo3 != classInfo)
                    		&& (subClasses.getSize() > 1))
                        tool.setLabel(action.getAlias() + " : " + classInfo3.getAlias());
                }
            }
            // check if the action should be added to a groupe
            group = action.getGroup();

            if (group == null)
                toolBar.addTool(tool);
            else
            {
                subToolBar = groupToolsBars.get(group);

                if (subToolBar == null)
                {
                    subToolBar = buildToolBar(group);
                    groupToolsBars.put(group, subToolBar);
                    toolBar.addToolBar(subToolBar);
                }

                subToolBar.addTool(tool);
            }
        }

        return toolBar;
    } // end method buildToolBar

    /**
     * Internal use only (toolbar build).
     * @param action1
     * @param classInfo1
     * @param action2
     * @param classInfo2
     *
     * @return true (always if not overriden for internal use only)
     */
    protected boolean acceptSubAction(LyAction action1, LyClassInfo classInfo1, LyAction action2,
        LyClassInfo classInfo2)
    {
        return true;
    }

    /**
     * Build a toolbar from a group name.
     * @param group the group name
     * @return a toolbar
     */
    protected LyToolBar buildToolBar(String group)
    {
        LyToolBar toolBar = new LyToolBar();

        toolBar.setId("_toolBar" + group + "_" + _parent.getId());
        toolBar.setLabel(group);
        toolBar.setStyle(LyToolBar.Style.BORDER);
        toolBar.setFill(LyToolBar.Fill.HORIZONTAL);

        return toolBar;
    }

    /**
     * Builds the table description.
     *
     * @return the table component
     */
    protected LyTable buildTable()
    {
        _table = new LyTable(_parent.getId() + "_table");

        LyEnvironment environment = _parent.getEnvironment();

        String selectionMode = environment.getResource("LY_SELECTION_MODE");

        if (selectionMode != null)
        {

            if (selectionMode.equals("MULTIPLE"))
                _table.setSelectionMode(LyTable.SelectionMode.MULTIPLE);
            else if (selectionMode.equals("SINGLE"))
                _table.setSelectionMode(LyTable.SelectionMode.SINGLE);
            else if (selectionMode.equals("NONE"))
                _table.setSelectionMode(LyTable.SelectionMode.NONE);
            else
                environment.logWarning("Unexpected value for property LY_SELECTION_MODE : "
                    + selectionMode + " expected MULTIPLE, SINGLE or NONE.");
        }

        _table.setIsContext(((LyAreaController) _parent).getContextClient() != null);

        _table.setController(_parent.getId());
        _table.setHeaders(buildHeaders());
        _table.setHLines(true);
        _table.setVLines(true);

        // Set the constraint on the table (buildTable() method may be called separately).
        LyConstraint constraint = new LyConstraint();

        constraint.setFill(LyConstraint.Fill.BOTH);
        _table.setConstraint(constraint);

        return _table;
    } // end method buildTable

    /**
     * Builds the table description.
     * @param oldTable  the old table
     *
     * @return the table component
     */
    public LyTable rebuildTable(LyTable oldTable)
    {

        if (oldTable == null)
            return null;

        LyArea area = (LyArea) oldTable.getParent();

        area.removePanel(oldTable);
        oldTable.free();

        if ((_table != null) && (_table != oldTable))
            _table.free();

        _table = buildTable();
        area.addPanel(_table);
        return _table;
    }

    /**
     * Rebuilds the headers of the table.
     *
     * @return the headers component
     */
    public LyHeaders rebuildHeaders()
    {
        LyHeaders headers = null;

        if (_table != null)
        {
            LyHeaders oldHeaders = _table.getHeaders();

            if (oldHeaders != null)
                oldHeaders.free();

            headers = buildHeaders();
            _table.setHeaders(headers);
        }

        return headers;
    }

    /**
     * Rebuilds the footers of the table.
     *
     * @return the footers component
     */
    public LyFooters rebuildFooters()
    {
        LyFooters footers = null;

        if (_table != null)
        {
            LyFooters oldFooters = _table.getFooters();

            if (oldFooters != null)
                oldFooters.free();

            footers = buildFooters();
            _table.setFooters(footers);
        }

        return footers;
    }

    /**
     * Shortcut to the table description.
     *
     * @return the table component
     */
    public LyTable getTable()
    {
        return _table;
    }

    /**
     * Builds the table header line description.
     *
     * @return the headers line
     */
    protected LyHeaders buildHeaders()
    {

        LySimpleTableController tableController = (LySimpleTableController) _parent;
        LyHeader                header;
        LyHeaders               headers = new LyHeaders();

        // SMA : No headers in list display, but LySwingTable needs empty headers
        if (tableController.isDisplayList())
        {
            LyInfoList<LyFieldInfo> complexFields = tableController.getComplexFields();
            int                     nbCols = tableController.getColsNumberListDisplay();
            int                     i = 0;

            // header for status column
            if (tableController.isStatusColumn(tableController.getStatusFieldInfos()))
            {
                header = new LyHeader();
                header.setLabel("LY_STATUS_SORT");
                header.setHidden(true);
                headers.addHeader(header);
                i = 1;
            }

            boolean hidden = true;
            int     complexSize = complexFields.getSize();

            if (complexSize > 0)
            {
                hidden = false;
                nbCols = nbCols - complexSize;
            }

            // add non complex headers, complex headers will be added by complexbuilder
            for (; i < nbCols; i++)
            {
                header = new LyHeader();
                header.setLabel("");
                header.setHidden(hidden);
                header.setSortable(false);
                headers.addHeader(header);
            }
        }
        else
        {
            LyInfoList<LyFieldInfo> fields = tableController.getFields();
            int                     i, n = ((fields == null) ? 0 : fields.getSize());
            LyFieldInfo             fieldInfo;

            // if there are field with status in this classinfo, we add a
            // header...
            header = buildStatusHeader();

            if (header != null)
                headers.addHeader(header);

            LyHeaders lastSubHeaderForGroups = null;

            for (i = 0; i < n; i++)
            {
                fieldInfo = fields.getInfo(i);
                if ((fieldInfo.getGroup() != null)
                        && ("true".equalsIgnoreCase((String) tableController
                                .getAction().getParameter(
                                        "_showGroupsInHeaders", true))))
                {
                    if ((lastSubHeaderForGroups != null)
                            && (!fieldInfo.getGroup().equals(
                                    lastSubHeaderForGroups.getLabel())))
                    {
                        lastSubHeaderForGroups = null;
                    }
                    if (lastSubHeaderForGroups == null)
                    {
                        lastSubHeaderForGroups = new LyHeaders();
                        String label = fieldInfo.getGroup();

                        lastSubHeaderForGroups.setId("headers"
                                + LyController.ID_SEP + fieldInfo.getId());
                        lastSubHeaderForGroups.setLabel(label);
                        lastSubHeaderForGroups.setAlign(LyHeader.Align.CENTER);
                        headers.addHeaders(lastSubHeaderForGroups);
                    }
                    header = buildHeader(fieldInfo);
                    lastSubHeaderForGroups.addHeader(header);
                } else
                {
                    lastSubHeaderForGroups = null;
                    header = buildHeader(fieldInfo);
                    headers.addHeader(header);
                }
            }
        }
        return headers;
    }

    /**
     * Build a header for a fieldInfo.
     *
     * @param fieldInfo Field info for which a header is built.
     * @return the header
     */
    public LyHeader buildHeader(LyFieldInfo fieldInfo)
    {
        LySimpleTableController  tableController = (LySimpleTableController) _parent;
        LyTableBehaviorInterface behavior = (LyTableBehaviorInterface) _parent.getBehavior();
        LyHeader                 header = new LyHeader();
        String                   label, tooltip;

        header.setId("header" + LyController.ID_SEP + fieldInfo.getId());
        label = behavior.getHeaderLabel(tableController, fieldInfo);
        header.setLabel(label);

        LyClassInfo             classInfo = tableController.getClassInfo();
        LyInfoList<LyFieldInfo> sortFields = classInfo.getFieldInfos(LyFieldInfo.MARK_SORT);

        if ((sortFields == null) || (sortFields.getSize() == 0) || sortFields.contains(fieldInfo))
            header.setSortable(true);
        else
            header.setSortable(false);

        tooltip = behavior.getHeaderTooltip(tableController, fieldInfo);
        header.setTooltip(tooltip);

        if (fieldInfo instanceof LyNumberFieldInfo)
            header.setAlign(LyHeader.Align.RIGHT);
        else
            header.setAlign(LyHeader.Align.LEFT);

        double colWidth = tableController.getColumnSize(fieldInfo);

        if (colWidth >= 0)
            header.setColumnWidth(colWidth);

        return header;
    } // end method buildHeader

    /**
     * Build a status header.
     *
     * @return the headers
     */
    protected LyHeader buildStatusHeader()
    {
        LyInfoList<LyFieldInfo> statusFields = ((LySimpleTableController) _parent)
            .getStatusFieldInfos();
        int                     n = ((statusFields == null) ? 0 : statusFields.getSize());

        if ((n == 0) && (!_parent.getEnvironment().testResource("LY_ERROR_STATE_ICON")))
            return null;

        // the header is sortable if at least one status field
        // is sortable.
        boolean  sortable = false;
        int      i;
        LyHeader header;

        for (i = 0; i < n; i++)
            sortable = sortable || statusFields.getInfo(i).hasMark(LyFieldInfo.MARK_SORT);

        header = new LyHeader();
        header.setId(LySimpleTableController.DEMAND_STATUS_SORT);
        header.setLabel("LY_STATUS_SORT");
        header.setSortable(sortable);
        header.setResizable(false);
        header.setFixed(true);
        header.setAlign(LyHeader.Align.CENTER);
        return header;
    } // end method buildStatusHeader

    /**
     * Builds the table footer line description.
     *
     * @return the footer line
     */
    protected LyFooters buildFooters()
    {
        LySimpleTableController tableController = (LySimpleTableController) _parent;
        LyInfoList<LyFieldInfo> fields = tableController.getFields();
        int                     i, n = ((fields == null) ? 0 : fields.getSize());
        LyFieldInfo             fieldInfo;
        LyFooter                footer;
        LyFooters               footers = null;

        // SMA
        if (tableController.isDisplayList())
            return buildDisplayListFooters();

        for (i = 0; i < n; i++)
        {
            fieldInfo = fields.getInfo(i);

            if ((fieldInfo instanceof LyNumberFieldInfo)
                    && (((LyNumberFieldInfo) fieldInfo).getFunction()
                        != LyNumberFieldInfo.Function.NONE))
                break;
        }

        if (i < n)
        {
            // There is almost one footer. We must add one for each column.
            footers = new LyFooters();

            // if there are field with status in this classinfo, we add a footer...
            LyInfoList<LyFieldInfo> statusFields = tableController.getStatusFieldInfos();

            if (((statusFields != null) && (statusFields.getSize() > 0))
                    || (_parent.getEnvironment().testResource("LY_ERROR_STATE_ICON")))
            {
                footer = new LyFooter();
                footers.addFooter(footer);
            }

            for (i = 0; i < n; i++)
            {
                fieldInfo = fields.getInfo(i);
                footer = new LyFooter();
                footer.setId("footer" + LyController.ID_SEP + fieldInfo.getId());
                footer.setLabel("");
                footers.addFooter(footer);
            }
        }

        return footers;
    } // end method buildFooters

    /**
     * Builds the table footer line description.
     * In list display, we have to know how many columns we are going to show.
     * @return the footer line
     */
    protected LyFooters buildDisplayListFooters()
    {
        LySimpleTableController tableController = (LySimpleTableController) _parent;
        LyInfoList<LyFieldInfo> fields = tableController.getFields();
        LyClassInfo             classInfo = tableController.getClassInfo();
        LyInfoList<LyFieldInfo> statusFields = tableController.getStatusFieldInfos();
        LyFooter                footer;
        LyFooters               footers = new LyFooters();
        LyFieldInfo             fieldInfo;

        // default : there is 2 columns, image and datas.
        int footerStart = 2;

        if (tableController.isDisplayListShowFields())
            footerStart++;

        // if there is a status column, empty footer for it
        if (tableController.isStatusColumn(statusFields))
            footerStart++;

        // add empty footers
        for (int i = 0; i < footerStart; i++)
        {
            footer = new LyFooter();
            footer.setLabel("");
            footers.addFooter(footer);
        }

        boolean footerAdded = false;

        // add real footers
        for (int j = 0; j < fields.getSize(); j++)
        {
            fieldInfo = fields.getInfo(j);

            if ((fieldInfo instanceof LyNumberFieldInfo)
                    && (((LyNumberFieldInfo) fieldInfo).getFunction()
                        != LyNumberFieldInfo.Function.NONE) && classInfo.contains(fieldInfo))
            {
                footer = new LyFooter();
                footer.setId("footer" + LyController.ID_SEP + fieldInfo.getId());
                footer.setLabel("");
                footers.addFooter(footer);
                footerAdded = true;
            }
        }

        // if no footer fields, don't show footers.
        if (!footerAdded)
            return null;

        return footers;
    } // end method buildDisplayListFooters

} // end class LySimpleTableBuilder
