/**
 * LySocialBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyTableBuilder.java 31812 2011-07-28 12:45:13Z smaupin $
 */

package leon.control.builder;

import java.util.HashMap;

import leon.app.behavior.LyChatBehavior;
import leon.control.*;
import leon.info.LyAction;
import leon.info.LyClassInfo;
import leon.info.LyInfo;
import leon.info.LyInfoDataList;
import leon.misc.LyEnvironment;
import leon.misc.LyTools;
import leon.view.*;


/**
 * This builder creates a table view for the social table controller It is based on the table builder.
 *
 * @see LyTableBuilder
 */
public class LySocialBuilder extends LyTableBuilder
{
	private static final String CHAT_TAB_ID = "chat_tab_";

	/**
	 * File version
	 */
	public static final String		 __VERSION =
		"$Id: LyTableBuilder.java 31812 2011-07-28 12:45:13Z smaupin $";

	/**
	 * Tabs for chat.
	 */
	private LyTabs	_chatTabs;
	
	/**
	 * Main area.
	 */
	private LyArea _mainArea;
	
	/**
	 * Storage for controller associated with tabs.
	 */
	private HashMap<String, LyController> _controllerList;
	
	
	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LySocialBuilder(LyController parent)
	{
		super(parent);
		_controllerList = new HashMap<String, LyController>();
		
	}

	/**
	 * Builds the table description.
	 *
	 * @return the table component
	 */
	protected LyTable buildTable()
	{
		LyTable table = super.buildTable();
		String	nRowsStr = (String)((LyAreaController)_parent).getParameter("_nRows");
		int		nRows = LyTools.intFromString(nRowsStr);

		table.setHeaders(null);
		LyConstraint constraint = new LyConstraint();
		constraint.setFill(LyConstraint.Fill.BOTH);
		table.setConstraint(constraint);
		table.setNRows(nRows);
		
		return table;
	}
	
	/**
	 * Get a list of action to build a toolbar for a given info
	 * @param controller the controller
	 * @param info the info
	 * @return a list of action
	 */
	protected LyInfoDataList<LyAction, LyClassInfo> getToolbarActions(LyAreaController controller, LyInfo info)
	{
		if (info != null)
			// First call with class info is to get class info actions
			return null;
		
		// Second call for extra actions declared directly on social action
		return super.getToolbarActions(controller, info);
	}
	
	/**
	 * Complete the right toolbar
	 * @param toolbar the right toolbar
	 * @return a toolbar
	 */
	protected LyToolBar completeRightToolBar(LyToolBar toolbar)
	{
		// No right toolbar
		return toolbar;
	}
	
	/**
	 * Add chat tab to chat tabs.
	 * @param controller the chat controller to add
	 * @return The Tab
	 */
	public LyTab addTab(LyController controller)
	{
		_mainArea = (LyArea) super.getComponent();
		if (_chatTabs == null && _mainArea != null)
		{
			_chatTabs = new LyTabs();
			_chatTabs.setClosable(true);
			LyConstraint constraint = new LyConstraint();
			constraint.setFill(LyConstraint.Fill.HORIZONTAL);
			_mainArea.setConstraint(constraint);
			_chatTabs.setConstraint(constraint);
			_mainArea.setBorderPanel(_chatTabs, LyArea.SOUTH);
			
		}
		int index = getTabIndex();
		LyTab tab = new LyTab();
		tab.setLabel(String.valueOf(index + 1));
		
		String tabId = CHAT_TAB_ID + String.valueOf(index);
		tab.setId(tabId);
		tab.setPanel((LyArea) controller.getComponent());
		_controllerList.put(tabId, controller);
		_chatTabs.insertTab(tab, index);
		_chatTabs.setSelectedTab(tab.getId());
		return tab;
	}
	
	/**
	 * Get tab index.
	 * @return tab index
	 */
	private int getTabIndex()
	{
		LyEnvironment env = _parent.getEnvironment();
		String maxChat = env.getResource("LY_CHAT_MAX_CONVERSATIONS");
		int maxChatTab = LyTools.intFromString(maxChat);
		for (int i = 0; i < maxChatTab; i++)
		{
			if (_controllerList.get(CHAT_TAB_ID + String.valueOf(i)) == null)
			{
				return i;
			}
		}
		return 0;
	}
	
	/**
	 * Get tabs.
	 * @return LyTabs tabs
	 */
	public LyTabs getTabs()
	{
		return _chatTabs;
	}
	
	/**
	 * Get tab count.
	 * @return int Tab Count
	 */
	public int getTabCount()
	{
		if (_chatTabs != null)
		{
			return _chatTabs.getTabCount();
		}
		return -1;
	}
	
	/**
	 * Remove tab.
	 * @param index index of tab.
	 */
	public void removeTab(int index)
	{
		if (_chatTabs != null)
		{
			LyTab tab = _chatTabs.getTab(index);
			_chatTabs.removeTab(tab);
			tab.free();
			LyAreaController controller = (LyAreaController)_controllerList.get(tab.getId());
			if (controller != null)
			{
				LyChatBehavior behavior = (LyChatBehavior)controller.getBehavior();
				if (behavior != null)
				{
					behavior.tabRemoveSession(controller.getSession());
					behavior.free();
				}
				controller.free(controller.getParent(false));
				_controllerList.remove(tab.getId());
			}
		}
	}
	
	/**
	 * Free tab component.
	 */
	public void freeTabs()
	{
		if (_chatTabs != null)
		{
			_mainArea.setBorderPanel(null, LyArea.SOUTH);
			_chatTabs.free();
			_chatTabs = null;
		}
	}
	
	/**
	 * Free controllers.
	 */
	public void freeChatControllers()
	{
		for (LyController c : _controllerList.values())
		{
			LyAreaController controller = (LyAreaController)c;
			controller.free(controller.getParent(false));
		}
	}
	
	
} // end class LySocialBuilder
