/**
 * LySortWizardBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LySortWizardBuilder.java 40687 2012-09-06 15:07:47Z jlsaouli $
 */

package leon.control.builder;

import leon.control.*;
import leon.misc.*;
import leon.view.*;


/**
 * Creates a view to type in a new simple filter expression.
 *
 * @see LyFilterController
 */
public class LySortWizardBuilder extends LyAreaBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LySortWizardBuilder.java 40687 2012-09-06 15:07:47Z jlsaouli $";

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LySortWizardBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Builds the view component for the specified parent controller. The component may create the
	 * XML Tree or the internal view.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{
		LyArea			area = new LyArea();
		LyForm			form = buildPanel();
		LyConstraint	constraint = new LyConstraint();
		
		constraint.setFill(LyConstraint.Fill.BOTH);
		area.setScrolling(LyArea.Scrolling.AUTO);
		area.setController(_parent.getId());
		area.setTitleBar(buildTitleBar());
		area.setConstraint(constraint);
		constraint = new LyConstraint();
		constraint.setFill(LyConstraint.Fill.NONE);
		form.setConstraint(constraint);
		area.addPanel(form);
		area.setCommands(buildCommands());
		area.setController(_parent.getId());
        area.setBorder(LyArea.Border.IN);
		_component = area;
		updateTitleBar();
		return area;
	}

	/**
	 * Creates the form.
	 *
	 * @return the form component
	 */
	protected LyForm buildPanel()
	{
        LySortWizardController parent = (LySortWizardController)_parent;
        LyForm                 form = new LyForm(_parent.getId());
        LyFormField            field;
        LyRelation             relation;
        LyChoice               choice;
        LyCaption              caption;
        LyConstraint           constraint;
        int                    maxGroupBy = LySortWizardController.MAX_GROUP_BY;
        
        if (maxGroupBy > 0)
        {
            LySortEndInterface    controller = parent.getSortEndInterface();
            
            if ((controller instanceof LyTableController)
                && (((LyTableController)controller).getParameter("_limit") != null)) // Paginated table
            {
                LyEnvironment   environment = ((LyController)controller).getEnvironment();
                
                if (environment.testResource("LY_DISABLE_GROUP_BY_ON_PAGINATED_TABLES"))
                    maxGroupBy = 0;
            }
        }
	        
		field = new LyFormField(LyFilterController.SELECT_FIELDINFO);

		// Get the max number of item from parent
		int maxItem = ((LySortWizardController)_parent).getMaxItem();

		for (int i = 0; i < maxItem; i++)
		{
			field = new LyFormField(LySortWizardController.FIELD + i);
			field.setNeedCheck(true);
			constraint = new LyConstraint();
			constraint.setRemainder(false);
			constraint.setColSpan(1);
			field.setConstraint(constraint);

			caption = new LyCaption();
			if (i == 0)
				caption.setLabel("LY_SORT_BY");
			else
				caption.setLabel("LY_SORT_NEXT");

			relation = new LyRelation();

			field.setCaption(caption);
			field.setFieldItem(relation);
			form.addFormField(field);

			field = new LyFormField(LySortWizardController.DIRECTION + i);
			field.setNeedCheck(true);
            if (i < maxGroupBy)
                field.setConstraint(constraint);

			choice = new LyChoice();
			choice.addOption(new LyOption(LySortWizardController.DIR_INCREASING, "LY_SORT_INCREASING"));
			choice.addOption(new LyOption(LySortWizardController.DIR_DECREASING, "LY_SORT_DECREASING"));

			field.setFieldItem(choice); // Add Choice

			form.addFormField(field);
            
            if (i < maxGroupBy)
            {
                field = new LyFormField(LySortWizardController.GROUP_BY + i);
                field.setNeedCheck(true);
    
                choice = new LyChoice();
                choice.setMultiple(true);
                choice.addOption(new LyOption(LySortWizardController.GROUPED + i, "LY_SORT_GROUP_BY"));
    
                field.setFieldItem(choice); // Add Choice
                field.setEnabled(false);
                form.addFormField(field);
            }
		}
		
		constraint = new LyConstraint();
		constraint.setFill(LyConstraint.Fill.NONE);
		form.setConstraint(constraint);
		form.setAlign(LyForm.Align.LEFT);
		form.setBorder(false);
		
		return form;
	} // end method buildPanel

	/**
	 * Creates the command bar and sets its content.
	 *
	 * @return the commands bar component
	 */
	protected LyCommands buildCommands()
	{
		LyCommands    commands = new LyCommands();
		LyCommand     command;
		LyViewManager viewManager = getViewManager();
		boolean		  hasView = (_parent.getParent() instanceof LyViewController);
		LyEnvironment environment = _parent.getEnvironment();

		if (hasView)
		{
			command = new LyCommand(LyFormController.DEMAND_VALIDATE, "LY_VALIDATE");
			command.setAction(LyCommand.Action.SUBMIT);
			
			if (environment.checkImage("LY_COMMAND_VALIDATE"))
				command.setImage("LY_COMMAND_VALIDATE");
			
			commands.addCommand(command);
		}

		if (hasView
			&& environment.testResource("LY_APPLY_BUTTON")
			&& (((LySortWizardController)_parent).getSortEndInterface() != null))
		{
			if (!hasView || !viewManager.hasProperty("NOT_APPLY"))
			{
				command = new LyCommand(LyFormController.DEMAND_APPLY, "LY_APPLY");
				command.setAction(LyCommand.Action.SUBMIT);
				
				if (environment.checkImage("LY_COMMAND_APPLY"))
					command.setImage("LY_COMMAND_APPLY");
				
				commands.addCommand(command);
			}
		}

		if (hasView)
		{
			command = new LyCommand(LyFormController.DEMAND_CANCEL, "LY_CANCEL");
			command.setAction(LyCommand.Action.CANCEL);
			
			if (environment.checkImage("LY_COMMAND_CANCEL"))
				command.setImage("LY_COMMAND_CANCEL");
			
			commands.addCommand(command);
		}
		else
		{
			command = new LyCommand(LyFormController.DEMAND_CLOSE, "LY_CLOSE");
			command.setAction(LyCommand.Action.CLOSE);
			
			if (environment.checkImage("LY_COMMAND_CLOSE"))
				command.setImage("LY_COMMAND_CLOSE");
			
			commands.addCommand(command);
		}

		command = buildHelpCommand();

		if (command != null)
			commands.addCommand(command);

		return commands;
	} // end method buildCommands
} // end class LySortWizardBuilder
