/**
 * LyTabAreaBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyTabAreaBuilder.java 31040 2011-05-17 13:28:17Z nhamon $
 */

package leon.control.builder;

import leon.control.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.getinterface.*;


/**
 * This builder construct the view for a tab Area Controller.
 *
 * @see LyTabAreaController
 */
public class LyTabAreaBuilder extends LyAreaBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyTabAreaBuilder.java 31040 2011-05-17 13:28:17Z nhamon $";

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyTabAreaBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Builds the view component for the specified parent controller. The component may create the
	 * XML Tree or the internal view.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{
		LyTabAreaController tabController = (LyTabAreaController)_parent;
		LyArea			    area;

		if (tabController.isWizard())
			area = buildWizardArea();
		else
			area = buildTabsArea();

		LyTitleBar titleBar = buildTitleBar();

		area.setTitleBar(titleBar);
		area.setToolBar(buildToolBar());

		// Update content
		updateTitleBar();

		area.setController(_parent.getId());

		LyConstraint  constraint = new LyConstraint();
		
		constraint.setFill(LyConstraint.Fill.BOTH);
		area.setConstraint(constraint);
		
		_component = area;
		return area;
	} // end method build

	/**
	 * Builds tabs area.
	 * @return an area
	 */
	protected LyArea buildTabsArea()
	{
		LyTabAreaController tabController = (LyTabAreaController)_parent;
		int				    n = tabController.getAreaControllersCount();

		// Add areas for the sub-areas
		LyArea			    area = (LyArea)_component, subArea;
		LyTabs			    tabs = null;

		if (area == null)
		{
			area = new LyArea();
			_component = area;
			tabs = new LyTabs();
			area.addPanel(tabs);
		}
		else
			tabs = (LyTabs)area.getPanel(0);

		String			 submitTabId = null;
		int				 i;
		LyTab			 tab = null;
		LyAreaController areaController;

		for (i = 0; i < n; i++)
		{
			areaController = tabController.getAreaController(i);
			subArea = (LyArea)areaController.getComponent();
			if (subArea == null)
				// There was a problem when building the component
				continue;

			subArea.setScrolling(LyArea.Scrolling.NO);

			tab = new LyTab();

			tab.setId("_tab" + areaController.getId());
			//tab.setId(areaController.getId());

			tab.setLabel(areaController.getName());
			tab.setImage(areaController.getImage());
			tab.setPanel(subArea);

			if (i == 0)
			{
				// Check if a form may be submitted in one of the areas
				LyController toSubmit = areaController;

				// XXX Should go THROUGH sub controllers here and search for interface
				// XXX At this step just keep first
				if (toSubmit instanceof LyCompoundAreaController)
					toSubmit = ((LyCompoundAreaController)toSubmit).getAreaController(0);

				if (toSubmit instanceof LySubmitTab)
					submitTabId = toSubmit.getId();
			}
			if (submitTabId != null)
				tab.setTargetFormId(submitTabId);

			tabs.addTab(tab);
		}
		
		LyConstraint  constraint = new LyConstraint();
		
		constraint.setFill(LyConstraint.Fill.BOTH);
		tabs.setConstraint(constraint);

		return area;
	} // end method buildTabsArea

	/**
	 * Builds wizard area.
	 * @return an area
	 */
	protected LyArea buildWizardArea()
	{
		LyTabAreaController tabController = (LyTabAreaController)_parent;
		LyArea			    area = new LyArea(), subArea, parentArea;
		LyForm			    form;

		subArea = new LyArea();
		form = new LyForm("_labels");

		form.setBorder(true);
		form.setController(_parent.getId());
		subArea.addPanel(form);

		LyConstraint  constraint = new LyConstraint();
		LyEnvironment environment = tabController.getWizardFormEnvironment();

		subArea.setEnvironment(environment);
		constraint.setRemainder(false);
		constraint.setColSpan(1);
		constraint.setColWeight(0);
		constraint.setFill(LyConstraint.Fill.VERTICAL);
		subArea.setConstraint(constraint);

		area.addPanel(subArea);
		area.setController(_parent.getId());

		if (tabController.getAreaControllersCount() >= 1)
		{
    		LyAreaController areaController = tabController.getAreaController(0);
    
            parentArea = new LyArea();
            subArea = (LyArea)areaController.getComponent();
            subArea.setScrolling(LyArea.Scrolling.NO);
            subArea.setCommands(null);
            parentArea.addPanel(subArea);
            area.addPanel(parentArea);
        }

		LyArea topArea = new LyArea();

		topArea.addPanel(area);
		topArea.setScrolling(LyArea.Scrolling.NO);
		topArea.setCommands(buildCommands());

		return topArea;
	} // end method buildWizardArea

	/**
	 * Clear the tabs content
	 *
	 */
	public void clearTabs()
	{
		if (_component != null)
		{
			LyArea area = ((LyArea)_component);
			LyTabs tabs = (LyTabs)area.getPanel(0);
			LyTab  tab;

			while (tabs.getTabCount() > 0)
			{
				tab = tabs.getTab(0);
				tabs.removeTab(tab);
			}
		}
	}

	/**
	 * This method is invoked to enable the the area to complete the menuBar of the view and add its
	 * own menu items. Default implementation invokes the method on the builder instance.
	 *
	 * @param menuBar the menuBar structure
	 * @param level   : level of current Controller. one of the LEVEL_XXX constants
	 * @see   LyViewController
	 */
	public void completeMenuBar(LyMenuBar menuBar, int level)
	{
		LyTabAreaController tabController = (LyTabAreaController)_parent;
		LyAction		    action = tabController.getAction();

		if (action.hasMark(LyAction.MARK_NO_MENUBAR))
			return;

		int currentTab = tabController.getSelectedTab();

		if (tabController.getAreaControllersCount() > 0)
		{
			LyAreaController areaController = tabController.getAreaController(currentTab);
			LyBuilder		 builder = areaController.getBuilder();

			if (builder instanceof LyAreaBuilder)
				((LyAreaBuilder)builder).completeMenuBar(menuBar, level);
		}
		// Add menus for the actions of the tab action
		super.completeMenuBar(menuBar, level);
	}

	/**
	 * This method is invoked to enable the the area to complete the messageBar of the view and add
	 * its own messages. Default implementation invokes the method on the builder instance.
	 *
	 * @param messages the message bar structure
	 * @param level    : level of current Controller. one of the LEVEL_XXX constants.
	 * @see   LyViewController
	 */
	public void completeMessages(LyMessages messages, int level)
	{
		LyTabAreaController tabController = (LyTabAreaController)_parent;
		LyAction		    action = tabController.getAction();

		if (action.hasMark(LyAction.MARK_NO_MESSAGES))
			return;

		if (tabController.getAreaControllersCount() > 0)
		{
			int				 currentTab = tabController.getSelectedTab();
			LyAreaController areaController = tabController.getAreaController(currentTab);

			areaController.completeMessages(messages, level);
		}
	}

	/**
	 * Build the command bar.
	 *
	 * @return the command bar component
	 */
	protected LyCommands buildCommands()
	{
		LyCommands commands = new LyCommands();
		LyCommand  command;

		command = new LyCommand(LyTabAreaController.DEMAND_PREVIOUS, "LY_PREVIOUS");
		command.setEnabled(false);
		commands.addCommand(command);

		command = new LyCommand(LyTabAreaController.DEMAND_NEXT, "LY_NEXT");
		command.setEnabled(true);
		commands.addCommand(command);

		command = new LyCommand(LyTabAreaController.DEMAND_TERMINATE, "LY_TERMINATE");
		command.setEnabled(false);
		commands.addCommand(command);

		command = new LyCommand(LyFormController.DEMAND_CANCEL, "LY_CANCEL");
		command.setAction(LyCommand.Action.CANCEL);
		command.setEnabled(true);
		commands.addCommand(command);

		commands.setAlign(LyCommands.Align.RIGHT);
		commands.setId("_commands");
		return commands;
	}
} // end class LyTabAreaBuilder
