 /**
 * LyTextBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyTextBuilder.java 34963 2012-03-22 14:55:31Z jlsaouli $
 */

package leon.control.builder;

import leon.control.*;
import leon.info.*;
import leon.view.*;


/**
 * Creates a view to show Text.
 *
 * @see LyTextController
 */
public class LyTextBuilder extends LyAreaBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyTextBuilder.java 34963 2012-03-22 14:55:31Z jlsaouli $";
    
	/** Default text width */
    public static int       DEFAULT_TEXT_WIDTH = 100;
    
    /** Default text height */
    public static int       DEFAULT_TEXT_HEIGHT = 30;
    
    /** The associated form */
	protected LyForm		_form;

	/**
	 * Default constructor.
	 *
	 * @param parent Parent controller of this builder
	 */
	public LyTextBuilder(LyController parent)
	{
		super(parent);

		// Get ready to recieve content input fields

		_form = new LyForm(_parent.getId() + "_form");
	}

	/**
	 * Builds the view component for the specified parent controller. The component may create the
	 * XML Tree or the internal view.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{
		/* The text view is an area which contains :
        - A titleBar (description)
        - A text component
		*/
		
		LyArea area = new LyArea();
		LyArea subArea = new LyArea();
		
		area.setId(_parent.getId());
		area.setController(_parent.getId());
		area.setTitleBar(buildTitleBar());
		subArea.addPanel(_form);
		area.addPanel(subArea);
		
		if (getViewManager().hasProperty("WEB_MANAGER"))
		{
			// In DHTML mode, use a fake command bar to avoid scroll issues.
			LyCommands	commands = new LyCommands();

			subArea.setCommands(commands);
			subArea.setScrolling(LyArea.Scrolling.AUTO);
		}
		
		_component = area;
		
		updateTitleBar();
		
		return area;
	}

	/**
	 * Add a text field in the user input window.
	 *
	 * @param  fieldId    Id of the field
	 * @param  fieldLabel Label of the field
	 * @param  itemLabel  Text to display
	 * @param  nCols      Number of columns of the text field
	 * @param  nRows      Number of rows of the text field
	 * @param  format     Format of the text Field (See also tag <TEXT> of leon.dtd)
	 * @param  optional   In user input mode, tells if the field is optional
	 * @return The text component created.
	 */
	public LyText setText(String fieldId,
		String fieldLabel, String itemLabel,
		int nCols, int nRows, String format, boolean optional)
	{
		LyFormField  field = new LyFormField(fieldId);

		LyConstraint constraint = new LyConstraint();

		constraint.setFill(LyConstraint.Fill.BOTH);
		constraint.setRowWeight(1);
		constraint.setColWeight(1);
		constraint.setRemainder(true);

		field.setConstraint(constraint);

		if (fieldLabel != null)
		{
			LyCaption caption = new LyCaption();

			caption.setLabel(fieldLabel);
			caption.setColon(true);
			field.setCaption(caption);
		}

		// The command depends on the type of the fieldInfo
		LyText text = new LyText();

		text.setNCols(nCols <= 0 ? DEFAULT_TEXT_WIDTH : nCols);
		text.setNRows(nRows <= 0 ? DEFAULT_TEXT_HEIGHT : nRows);
		text.setFormat(format);

		if (!LyText.SHOW_ARROWS)
			text.showArrows(false);
			//text.setWrap(false);

		if (itemLabel != null)
			text.setLabel(itemLabel);

		field.setFieldItem(text);
		field.setOptional(optional);

		// XXX To Remove when setEditable will be implemented
		field.setEditable(false);

		_form.addFormField(field);
		return (text);
	} // end method setText
    
    /**
     * This method is invoked to enable the the area to complete the menuBar of the view and add its
     * own menu items. Default implementation invokes the method on the builder instance.
     *
     * @param menuBar the menuBar structure
     * @param level   : level of current Controller. one of the LEVEL_XXX contants in
     *                LyViewController
     * @see   LyViewController
     */
    public void completeMenuBar(LyMenuBar menuBar, int level)
    {
        if (menuBar == null)
            return;

        LyAction action = ((LyAreaController)_parent).getAction();

        if (action.hasMark(LyAction.MARK_NO_MENUBAR))
            return;
        
        super.completeMenuBar(menuBar, level);
        
        LyMenu  menu = (LyMenu)menuBar.getComponentByLabel("LY_MENU_FILE");
        LyItem  item;
        
        
        item = new LyItem(LyTextController.DEMAND_CLEAR, "LY_CLEAR");
        item.setController(_parent.getId());
        menu.addItem(item);
        
        menu.addSep();
        
        item = new LyItem(LyTextController.DEMAND_SAVE_AS, "LY_SAVE_AS");
        item.setDialog(true);
        item.setController(_parent.getId());
        menu.addItem(item);
    }
} // end class LyTextBuilder
