/**
 * LyUserInputBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyUserInputBuilder.java 40687 2012-09-06 15:07:47Z jlsaouli $
 */

package leon.control.builder;

import leon.control.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;


/**
 * Creates a view to type in a new simple find selection view.
 *
 * @see LyUserInputController
 */
public class LyUserInputBuilder extends LyAreaBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION =
		"$Id: LyUserInputBuilder.java 40687 2012-09-06 15:07:47Z jlsaouli $";

	/** Internal form with usr input. */
	protected LyForm		   _form;

	/**
	 * Default constructor.
	 *
	 */
	public LyUserInputBuilder()
	{
		this(null);
	}

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyUserInputBuilder(LyController parent)
	{
		super(parent);

		// Get ready to receive content input fields
		_form = ((parent == null) ? null : new LyForm(_parent.getId() + "_form"));

		if ((_form != null) && (_parent != null))
			_form.setController(_parent.getId());
	}

	/**
	 * Builds the view component for the specified parent controller. The component may create the
	 * XML Tree or the internal view.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{
		if (_parent != null)
		{
			LyArea			area = new LyArea();
			LyArea			subArea = new LyArea();
			LyConstraint	constraint = new LyConstraint();
			
			constraint.setFill(LyConstraint.Fill.BOTH);
			area.setController(_parent.getId());
			if(!_parent.hasViewController())
			    area.setBorder(LyArea.Border.OUT);
			_form.setConstraint(constraint);
			area.setConstraint(constraint);
			subArea.setConstraint(constraint);
			area.setScrolling(LyArea.Scrolling.NO);
			area.setTitleBar(buildTitleBar());
			area.setToolBar(buildToolBar());
			subArea.setScrolling(LyArea.Scrolling.AUTO);
			subArea.setController(_parent.getId());
			subArea.addPanel(_form);
			area.addPanel(subArea);
			area.setCommands(buildCommands());

			_component = area;
			buildFormContent();

			// Update default content
			updateTitleBar();
			return (area);
		}
		return null;
	}

	/**
	 * Creates the form content.
	 */
	protected void buildFormContent()
	{
	}

	/**
	 * Creates the command bar and sets its content.
	 *
	 * @return the commands component
	 */
	protected LyCommands buildCommands()
	{
		LyAction action = ((LyAreaController)_parent).getAction();

		if (action.hasMark(LyAction.MARK_NO_COMMANDS))
			return null;

		LyEnvironment environment = _parent.getEnvironment();
		LyCommands commands = new LyCommands();
		LyCommand  command;

		if (_parent.hasViewController()
		        || (action.getTarget() != null))
		{
			command = new LyCommand(LyFormController.DEMAND_VALIDATE, "LY_VALIDATE");
			if (environment.checkImage("LY_COMMAND_VALIDATE"))
				command.setImage("LY_COMMAND_VALIDATE");
		}
		else
		{
			command = new LyCommand(LyFormController.DEMAND_APPLY, "LY_APPLY");
			if (environment.checkImage("LY_COMMAND_APPLY"))
				command.setImage("LY_COMMAND_APPLY");
		}
		
		command.setAction(LyCommand.Action.SUBMIT);
		commands.addCommand(command);

		if (_parent.hasViewController()
	        || (action.getTarget() != null))
		{
			command = new LyCommand(LyFormController.DEMAND_CANCEL, "LY_CANCEL");
			command.setAction(LyCommand.Action.CANCEL);
			if (environment.checkImage("LY_COMMAND_CANCEL"))
				command.setImage("LY_COMMAND_CANCEL");
			commands.addCommand(command);
		}

		command = buildHelpCommand();
		if (command != null)
			commands.addCommand(command);

		return commands;
	} // end method buildCommands

	/**
	 * Add a label in the user input window.
	 *
	 * @param  fieldId    the field id
	 * @param  fieldLabel the field label
	 * @param  itemLabel  the initial value
	 * @return the label component
	 */
	public LyLabel addLabel(String fieldId, String fieldLabel,
		String itemLabel)
	{
		return addLabel(fieldId, fieldLabel, itemLabel, -1);
	}

	/**
	 * Add a label in the user input window.
	 *
	 * @param  fieldId    the field id
	 * @param  fieldLabel the field label
	 * @param  itemLabel  the initial value
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the label component
	 */
	public LyLabel addLabel(String fieldId, String fieldLabel,
		String itemLabel, int pos)
	{
		LyFormField field = new LyFormField(fieldId);
		LyCaption   caption = new LyCaption();

		caption.setLabel(fieldLabel);
		caption.setColon(true);

		// The command depends on the type of the fieldInfo
		LyLabel label = new LyLabel();

		if (itemLabel != null)
			label.setLabel(itemLabel);

		field.setCaption(caption);
		field.setFieldItem(label);

		if (_form != null)
			_form.insertFormField(field, pos);

		return label;
	}

	/**
	 * Add a text field in the user input window.
	 *
	 * @param  fieldId    the field id
	 * @param  fieldLabel the field label
	 * @param  itemLabel  the initial value
	 * @param  ncols      the number of columns
	 * @param  format     the format of the field
	 * @param  optional   indicates if the field is optional
	 * @return the text component
	 */
	public LyText addText(String fieldId, String fieldLabel,
		String itemLabel, int ncols, String format, boolean optional)
	{
		return addText(fieldId, fieldLabel, itemLabel, ncols, 1, format, optional, -1);
	}

	/**
	 * Add a text field in the user input window.
	 *
	 * @param  fieldId    the field id
	 * @param  fieldLabel the field label
	 * @param  itemLabel  the initial value
	 * @param  ncols      the number of columns
	 * @param  format     the format of the field
	 * @param  optional   indicates if the field is optional
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the text component
	 */
	public LyText addText(String fieldId, String fieldLabel,
		String itemLabel, int ncols,
		String format, boolean optional, int pos)
	{
		return addText(fieldId, fieldLabel, itemLabel, ncols, 1, format, optional, pos);
	}

	/**
	 * Add a text field in the user input window.
	 *
	 * @param  fieldId    the field id
	 * @param  fieldLabel the field label
	 * @param  itemLabel  the initial value
	 * @param  ncols      the number of columns
	 * @param  nrows      the number of rows
	 * @param  format     the format of the field
	 * @param  optional   indicates if the field is optional
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the text component
	 */
	public LyText addText(String fieldId, String fieldLabel,
		String itemLabel, int ncols, int nrows,
		String format, boolean optional, int pos)
	{
		LyFormField field = new LyFormField(fieldId);
		LyCaption   caption = new LyCaption();

		caption.setLabel(fieldLabel);
		caption.setColon(true);

		// The command depends on the type of the fieldInfo
		LyText text = new LyText();

		text.setNCols((ncols <= 0) ? 20 : ncols);
		text.setNRows((nrows <= 0) ? 1 : nrows);
		text.setFormat(format);

		if (!LyText.SHOW_ARROWS)
			text.showArrows(false);

		if (itemLabel != null)
			text.setLabel(itemLabel);

		field.setCaption(caption);
		field.setFieldItem(text);
		field.setOptional(optional);

		if (_form != null)
			_form.insertFormField(field, pos);

		return (text);
	} // end method addText

	/**
	 * Add a relation field (empty) in the user input window.
	 *
	 * @param  fieldId    the id of the field
	 * @param  fieldLabel the label of the field
	 * @param  multiple   indicates if the relation is multiple
	 * @param  optional   indicates if the relation is optional
	 * @return the relation component
	 */
	public LyRelation addRelation(String fieldId, String fieldLabel,
		boolean multiple, boolean optional)
	{
		return addRelation(fieldId, fieldLabel, multiple, optional, -1);
	}

	/**
	 * Add a relation field (empty) in the user input window.
	 *
	 * @param  fieldId    the id of the field
	 * @param  fieldLabel the label of the field
	 * @param  multiple   indicates if the relation is multiple
	 * @param  optional   indicates if the relation is optional
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the relation component
	 */
	public LyRelation addRelation(String fieldId, String fieldLabel,
		boolean multiple, boolean optional, int pos)
	{
		LyFormField field = new LyFormField(fieldId);
		LyCaption   caption = new LyCaption();

		caption.setLabel(fieldLabel);
		caption.setColon(true);

		LyRelation relation = new LyRelation();

		relation.setMultiple(multiple);

		field.setCaption(caption);
		field.setFieldItem(relation);
		field.setOptional(optional);

		if (_form != null)
			_form.insertFormField(field, pos);

		return (relation);
	}

	/**
	 * Add a choice field (empty) in the user input window.
	 *
	 * @param  fieldId    the id of the field
	 * @param  fieldLabel the label of the field
	 * @param  multiple   indicates if the choice is multiple
	 * @param  optional   indicates if the choice is optional
	 * @return the choice component
	 */
	public LyChoice addChoice(String fieldId, String fieldLabel,
		boolean multiple, boolean optional)
	{
		return addChoice(fieldId, fieldLabel, multiple, optional, -1);
	}

	/**
	 * Add a choice field (empty) in the user input window.
	 *
	 * @param  fieldId    the id of the field
	 * @param  fieldLabel the label of the field
	 * @param  multiple   indicates if the choice is multiple
	 * @param  optional   indicates if the choice is optional
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the choice component
	 */
	public LyChoice addChoice(String fieldId, String fieldLabel,
		boolean multiple, boolean optional, int pos)
	{
		LyFormField field = new LyFormField(fieldId);
		LyCaption   caption = new LyCaption();

		caption.setLabel(fieldLabel);
		caption.setColon(true);

		LyChoice choice = new LyChoice();

		choice.setMultiple(multiple);

		field.setCaption(caption);
		field.setFieldItem(choice);
		field.setOptional(optional);

		if (_form != null)
			_form.insertFormField(field, pos);

		return (choice);
	}

	/**
	 * Add a number field in the user input window.
	 *
	 * @param  fieldId    the field id
	 * @param  fieldLabel the field label
	 * @param  itemLabel  the initial value
	 * @param  type       the number type
	 * @param  optional   indicates if the field is optional
	 * @return the number component
	 */
	public LyNumber addNumber(String fieldId, String fieldLabel,
		String itemLabel, LyNumberFieldInfo.Type type, boolean optional)
	{
		return addNumber(fieldId, fieldLabel, itemLabel, type, optional, -1);
	}

	/**
	 * Add a number field in the user input window.
	 *
	 * @param  fieldId    the field id
	 * @param  fieldLabel the field label
	 * @param  itemLabel  the initial value
	 * @param  type       the number type
	 * @param  optional   indicates if the field is optional
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the number component
	 */
	public LyNumber addNumber(String fieldId, String fieldLabel,
		String itemLabel, LyNumberFieldInfo.Type type, boolean optional, int pos)
	{
		LyFormField field = new LyFormField(fieldId);
		LyCaption   caption = new LyCaption();

		caption.setLabel(fieldLabel);
		caption.setColon(true);

		LyNumber number = new LyNumber();

		if (!LyNumber.SHOW_ARROWS)
			number.showArrows(false);

		//number.setType(type);
		number.setLabel(itemLabel);

		field.setCaption(caption);
		field.setFieldItem(number);
		field.setOptional(optional);

		if (_form != null)
			_form.insertFormField(field, pos);

		return (number);
	}

	/**
	 * Add a date field in the user input window.
	 *
	 * @param  fieldId    the field id
	 * @param  fieldLabel the field label
	 * @param  itemLabel  the initial value
	 * @param  format     the date format
	 * @param  optional   indicates if the field is optional
	 * @return the date component
	 */
	public LyDate addDate(String fieldId, String fieldLabel,
		String itemLabel, String format,
		boolean optional)
	{
		return addDate(fieldId, fieldLabel, itemLabel, format, optional, -1);
	}

	/**
	 * Add a date field in the user input window.
	 *
	 * @param  fieldId    the field id
	 * @param  fieldLabel the field label
	 * @param  itemLabel  the initial value
	 * @param  format     the date format
	 * @param  optional   indicates if the field is optional
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the date component
	 */
	public LyDate addDate(String fieldId, String fieldLabel,
		String itemLabel, String format,
		boolean optional, int pos)
	{
		LyFormField field = new LyFormField(fieldId);
		LyCaption   caption = new LyCaption();

		caption.setLabel(fieldLabel);
		caption.setColon(true);

		LyDate date = new LyDate();

		if (!LyDate.SHOW_ARROWS)
			date.showArrows(false);

		date.setFormat(format);
		date.setLabel(itemLabel);

		field.setCaption(caption);
		field.setFieldItem(date);
		field.setOptional(optional);

		if (_form != null)
			_form.insertFormField(field, pos);

		return (date);
	}

	/**
	 * Add a file field in the user input window.
	 *
	 * @param  fieldId    the field id
	 * @param  fieldLabel the field label
	 * @param  itemLabel  the initial value
	 * @param  type       the file type
	 * @param  optional   indicates if the field is optional
	 * @return the file component
	 */
	public LyFile addFile(String fieldId, String fieldLabel,
		String itemLabel, LyFile.Type type, boolean optional)
	{
		return addFile(fieldId, fieldLabel, itemLabel, type, optional, -1);
	}

	/**
	 * Add a file field in the user input window.
	 *
	 * @param  fieldId    the field id
	 * @param  fieldLabel the field label
	 * @param  itemLabel  the initial value
	 * @param  type       the file type
	 * @param  optional   indicates if the field is optional
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the file component
	 */
	public LyFile addFile(String fieldId, String fieldLabel,
		String itemLabel, LyFile.Type type, boolean optional, int pos)
	{
		LyFormField field = new LyFormField(fieldId);
		LyCaption   caption = new LyCaption();

		caption.setLabel(fieldLabel);
		caption.setColon(true);

		// The command depends on the type of the fieldInfo
		LyFile file = new LyFile();

		file.setType(type);

		if (itemLabel != null)
			file.setLabel(itemLabel);

		field.setCaption(caption);
		field.setFieldItem(file);
		field.setOptional(optional);

		if (_form != null)
			_form.insertFormField(field, pos);

		return (file);
	}

	/**
	 * Add a list field in the user input window.
	 *
	 * @param  fieldId    the field id
	 * @param  fieldLabel the field label
	 * @param  multiple   indicates if selection mode is multiple in the list
	 * @param  height     the height of the list
	 * @param  optional   indicates if the field is optional
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the list component
	 */
	public LyList addList(String fieldId, String fieldLabel,
		boolean multiple, int height,
		boolean optional, int pos)
	{
		LyFormField field = new LyFormField(fieldId);
		LyCaption   caption = new LyCaption();

		caption.setLabel(fieldLabel);
		caption.setColon(true);

		LyList list = new LyList();

		list.setSelectionMode(multiple ? LyList.SelectionMode.MULTIPLE
									   : LyList.SelectionMode.SINGLE);
		list.setNRows(height);

		field.setCaption(caption);
		field.setFieldItem(list);
		field.setOptional(optional);

		if (_form != null)
			_form.insertFormField(field, pos);

		return (list);
	}

	/**
	 * Add a boolean field in the user input window.
	 *
	 * @param  fieldId    the id of the field
	 * @param  fieldLabel the label of the field
	 * @param  status     the boolean status
	 * @param  optional   indicates if the choice is optional
	 * @return the boolean component
	 */
	public LyBoolean addBoolean(String fieldId, String fieldLabel,
		LyBoolean.Status status, boolean optional)
	{
		return addBoolean(fieldId, fieldLabel, status, optional, -1);
	}

	/**
	 * Add a boolean field in the user input window.
	 *
	 * @param  fieldId    the id of the field
	 * @param  fieldLabel the label of the field
	 * @param  status     the boolean status
	 * @param  optional   indicates if the choice is optional
	 * @param  pos        indicates the position of the field, -1 for the last position
	 * @return the boolean component
	 */
	public LyBoolean addBoolean(String fieldId, String fieldLabel,
		LyBoolean.Status status, boolean optional, int pos)
	{
		LyFormField field = new LyFormField(fieldId);
		LyCaption   caption = new LyCaption();

		caption.setLabel(fieldLabel);
		caption.setColon(true);

		LyBoolean booleanComp = new LyBoolean();

		booleanComp.setStatus(status);

		field.setCaption(caption);
		field.setFieldItem(booleanComp);
		field.setOptional(optional);

		if (_form != null)
			_form.insertFormField(field, pos);

		return booleanComp;
	}

    /**
     * Build and fill the title bar.
     *
     * @return the titlebar component
     */
    protected LyTitleBar buildTitleBar()
    {
        if (((LyAreaController)_parent).getTarget() != null)
            return null;

        return super.buildTitleBar();
    }
    
	/**
	 * Update titlebar ans set default content.
	 */
	public void updateTitleBar()
	{
		if (!(_parent instanceof LyUserInputController))
		{
			super.updateTitleBar();
			return;
		}

		LyUserInputController parent = (LyUserInputController)_parent;

		if (_component == null)
			return;

		LyTitleBar titleBar = (LyTitleBar)_component.getComponent(_parent.getId() + "_titleBar"); // XXX

		if (titleBar != null)
		{
			titleBar.setLeftLabel(parent.getTitleBar());
			if (parent.getEnvironment().testResource("LY_TITLEBAR_LEFT_IMAGE"))
				titleBar.setImage(parent.getImage());
		}
	}

	/**
	 * Sets the alignment of the form.
	 *
	 * @param align The new alignment based on the LyForm ALIGN constants.
	 */
	public void setFormAlignment(LyForm.Align align)
	{
		_form.setAlign(align);
	}

	/**
	 * Sets the border of the form.
	 *
	 * @param border Indicates if border must be set on the form.
	 */
	public void setFormBorder(boolean border)
	{
		_form.setBorder(border);
	}
} // end class LyUserInputBuilder
