/**
 * LyViewBuilder.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved. Current version: $Id: LyViewBuilder.java,v
 * 6.6.4.2 2009/07/22 10:00:11 jls Exp $
 */

package leon.control.builder;

import leon.app.*;
import leon.control.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.mobile.*;

import java.util.*;


/**
 * This builder is a basic class.
 *
 * @see  LyViewController
 */
public abstract class LyViewBuilder extends LyBuilder
{
	/**
	 * File version
	 */
	public static final String __VERSION = "$Id: LyViewBuilder.java 41145 2012-10-11 12:26:13Z smaupin $";

	// Flandres / FFT00048
	// Update once toolbar in construction.
	private boolean _underConstruction;

	/**
	 * Default constructor.
	 *
	 * @param  parent  parent controller
	 */
	public LyViewBuilder(LyController parent)
	{
		super(parent);
		_underConstruction = false;
	}

	/**
	 * This method is invoked to enable the the area to complete the menuBar of the view and add its
	 * own menu items. Default implementation invokes the method on the builder instance.
	 *
	 * @param  menuBar  the menuBar structure
	 * @param  level    : level of current Controller. one of the LEVEL_XXX contants. in
	 *                  LyViewController
	 *
	 * @see    LyViewController
	 */
	public void completeMenuBar(LyMenuBar menuBar, int level)
	{
		LyViewController viewController = (LyViewController)_parent;
		LyAreaController areaController = viewController.getAreaController();

		areaController.completeMenuBar(menuBar, level);
	}

	/**
	 * This method is invoked to enable the the area(s) to complete the message bar of the view and
	 * add its own message items. Default implementation invokes the method on the builder instance.
	 *
	 * @param  messages  the message bar structure
	 * @param  level     : level of current Controller. one of the LEVEL_XXX contants. in
	 *                   LyViewController
	 *
	 * @see    LyViewController
	 */
	public void completeMessages(LyMessages messages, int level)
	{
		LyViewController viewController = (LyViewController)_parent;
		LyAreaController areaController = viewController.getAreaController();

		areaController.completeMessages(messages, level);
	}

	/**
	 * Builds the toolbar for the view.
	 *
	 * @return  the toolbar component
	 */
	public LyToolBar buildToolBar()
	{
		LyViewController viewController = (LyViewController)_parent;
		LyAreaBuilder    areaBuilder = (LyAreaBuilder)viewController.getBuilder();

		return areaBuilder.buildToolBar();
	}

	/**
	 * Builds the area for this view.
	 *
	 * @return  the area component
	 */
	public LyArea buildArea()
	{
		// Update once toolbar in construction.
		_underConstruction = true;

		LyViewController viewController = (LyViewController)_parent;
		LyAreaController areaController = viewController.getAreaController();
		LyArea			 area = (LyArea)areaController.getComponent();

		if (_underConstruction)
		{
			LyToolBar leftToolbar = area.getTopLeftToolBar();
			LyToolBar rightToolbar = area.getTopRightToolBar();

			// Workaround for Defining which is the toolbar of the compound action
			// search through all sub-controllers of the view
			if (areaController instanceof LyCompoundAreaController)
			{
				leftToolbar = getCompoundToolBar((LyCompoundAreaController)areaController);
				rightToolbar = getCompoundRightToolBar((LyCompoundAreaController)areaController);
			}

			completeToolBar(leftToolbar);
			completeRightToolBar(rightToolbar);

			_underConstruction = false;
		}

		return area;
	} // end method buildArea

	/**
	 * Recursive method to find the left toolbar of a compound controller.
	 *
	 * @param   compoundController  compound controller
	 *
	 * @return  the toolbar for the given controller
	 */
	protected LyToolBar getCompoundToolBar(LyCompoundAreaController compoundController)
	{
		return getCompoundToolBar(compoundController, true);
	}

	/**
	 * Recursive method to find the left toolbar of a compound controller.
	 *
	 * @param   compoundController  compound controller
	 *
	 * @return  the toolbar for the given controller
	 */
	protected LyToolBar getCompoundRightToolBar(LyCompoundAreaController compoundController)
	{
		return getCompoundToolBar(compoundController, false);
	}

	/**
	 * Recursive method to find the toolbar of a compound controller.
	 *
	 * @param   compoundController  compound controller
	 * @param left true to find the left toolbar, false to find the right toolbar
	 *
	 * @return  the toolbar for the given controller
	 */
	protected LyToolBar getCompoundToolBar(LyCompoundAreaController compoundController, boolean left)
	{
		LyToolBar		 toolbar = null;
		int				 n = compoundController.getAreaControllersCount();
		LyAreaController areaController;

		// search the toolbar through sub-controllers
		for (int i = 0; i < n; i++)
		{
			areaController = compoundController.getAreaController(i);
			if (areaController instanceof LyCompoundAreaController)
				toolbar = getCompoundToolBar((LyCompoundAreaController)areaController, left);
			else
			{
				LyArea area = (LyArea)areaController.getComponent();
				// Return the first topLeft (if left == true) or
				// topRight if left == false
				LyToolBar tmp;

				if (left)
					tmp = area.getTopLeftToolBar();
				else
					tmp = area.getTopRightToolBar();

				if (tmp != null)
					toolbar = tmp;
			}

			if (toolbar != null)
				break;
		}

		return toolbar;
	} // end method getCompoundToolBar

	/**
	 * Complete the given toolbar.
	 *
	 * @param toolbar  the toolbar to complete
	 *
	 * @return  the completed toolbar
	 */
	public LyToolBar completeToolBar(LyToolBar toolbar)
	{
		return toolbar;
	}

	/**
	 * Complete the given toolbar.
	 *
	 * @param toolbar  the toolbar to complete
	 *
	 * @return  the completed toolbar
	 */
	public LyToolBar completeRightToolBar(LyToolBar toolbar)
	{
		LyViewController viewController = (LyViewController)_parent;
		LyAreaController areaController = viewController.getAreaController();
		LyArea			 area = (LyArea)areaController.getComponent();

		if (area.getCommands() != null)
			return toolbar;

		LyAction action = areaController.getAction();

		if ((toolbar == null) && (areaController instanceof LyCompoundAreaController))
		{
			LyToolBar leftToolBar = getCompoundToolBar((LyCompoundAreaController)areaController);

			if (leftToolBar != null)
				area = leftToolBar.getParentArea();
			else
			{
				LyArea temp = getAreaFromCompoundController((LyCompoundAreaController)areaController);

				if (temp != null)
					area = temp;
			}
		}

		LyEnvironment environment = _parent.getEnvironment();
		boolean		  isDashboard = ((action != null) ? (action.getRoot().getId() == LyAction.ID_DASHBOARD) : false);
		LyViewManager viewManager = getViewManager();

		if ((viewManager != null) && !viewManager.hasProperty(LyMobileViewManager.MOBILE_ONLY_PROPERTY)
				&& (!viewManager.hasProperty("MENU")
					|| (viewManager.hasProperty("PARTIAL_MENU")
						&& isDashboard
						&& viewManager.hasProperty("MANAGE_DASHBOARD"))
					|| ((action.hasMark(LyAction.MARK_NO_MENUBAR)
							|| !action.hasMark(LyAction.MARK_FRAME_VIEW))
						&& !isDashboard
						&& (!LySession.POPUP_VIEW.equals(viewManager.getCurrentWindowName())))))
		{
			// Add tools for help and closing the frame.
			boolean needToolbar = false;

			if ((_parent.getParent() == null)
					&& action.hasMark(LyAction.MARK_FRAME_VIEW))
				needToolbar = true;
			else if (viewManager.hasProperty("WEB_MANAGER"))
			{
				// Check if there is at least one titlebar in the area. If there is one, viewer
				// will manage itself the addition of close button. Otherwise we need to add the button here.
				Vector<LyTitleBar> titlebars = new Vector<LyTitleBar>();

				area.getComponents(LyTitleBar.class, titlebars);
				if (titlebars.size() == 0)
					needToolbar = true;
			}

			if (needToolbar)
			{
				LyTool aboutTool = null;
				LyTool exitTool = null;

				if (_parent.getParent() == null)
				{
					String about = environment.getString("LY_TOOL_ABOUT");

					if ((about != null)
						&& (about.length() > 0))
					{
						aboutTool = new LyTool(LyFrameController.DEMAND_ABOUT, "LY_TOOL_ABOUT");
						aboutTool.setImage("LY_TOOL_ABOUT");
						aboutTool.setDialog(true);
						aboutTool.setEnabled(true);
					}

					String quit = environment.getString("LY_TOOL_QUIT");

					if ((quit != null)
							&& (quit.trim().length() > 0))
					{
						exitTool = new LyTool(LyFrameController.DEMAND_EXIT, "LY_TOOL_QUIT");
						exitTool.setImage("LY_TOOL_QUIT");
						exitTool.setDialog(true);
						exitTool.setEnabled(true);
					}
					else
						exitTool = null;
				}
				else
				{
					String close = environment.getString("LY_TOOL_CLOSE");

					if ((close != null)
							&& (close.length() > 0))
					{
						exitTool = new LyTool(LyViewController.DEMAND_CLOSE, "LY_TOOL_CLOSE");
						exitTool.setImage("LY_TOOL_CLOSE");
						exitTool.setEnabled(true);
					}
				}

				// We add them in the right toolBar
				if (((exitTool != null)
						|| (aboutTool != null)) && !viewManager.isPortletMode())
				{
					if (toolbar == null)
					{
				        if (!environment.testResource("LY_RIGHT_TOOLBAR"))
				            return null;
						if (action.hasMark(LyAction.MARK_NO_TOOLBAR))
							return null;

						toolbar = new LyToolBar();
						toolbar.setFill(LyToolBar.Fill.HORIZONTAL);
						toolbar.setStyle(LyToolBar.Style.SEPARATOR);
						// toolbar.setColor("LY_DASHBOARD_HORIZONTAL_BACKGROUND");
						toolbar.setHAlign(LyToolBar.HAlign.RIGHT);
						toolbar.setVAlign(LyToolBar.VAlign.TOP);

						area.addToolBar(toolbar);
					}
					else if (toolbar.getComponentCount() > 0)
						toolbar.addSep();

					if (aboutTool != null)
						toolbar.addTool(aboutTool);
					if (exitTool != null)
						toolbar.addTool(exitTool);
				}
			}
		}

		return toolbar;
	}

	/**
	 * This method returns the area of the first areaController that can have a toolbar.
	 *
	 * @param   compound  the compound controller
	 *
	 * @return  an area or null if nothing found
	 */
	private LyArea getAreaFromCompoundController(LyCompoundAreaController compound)
	{
		LyArea			 area;
		LyAreaController controller;
		LyAction		 action;
		int				 size = compound.getAreaControllersCount();

		for (int i = 0; i < size; i++)
		{
			controller = compound.getAreaController(i);
			action = controller.getAction();

			if (action.hasMark(LyAction.MARK_NO_TOOLBAR))
				continue;

			if (controller instanceof LyCompoundAreaController)
				area = getAreaFromCompoundController((LyCompoundAreaController)controller);
			else
				area = (LyArea)controller.getComponent();

			if (area != null)
				return area;
		}

		return null;
	}

} // end class LyViewBuilder
