/**
 * LyImportController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyImportContentController.java 28422 2010-11-03 12:03:38Z ehelmste $
 */

package leon.control.content;

import java.io.*;
import leon.app.*;
import leon.app.behavior.*;
import leon.control.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;


/**
 * This class is the shared code between import on logical level.
 * @see LyImportController
 * @see LyImportTableController
 */
public abstract class LyImportContentController extends LyController
{
	/**
	 * File version
	 */
	public static final String		  __VERSION =
		"$Id: LyImportContentController.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Action of this area controller. */
	protected LyAction				  _action;

	/** Behavior to override import behavior. */
	protected LyImportContentBehavior _behavior;

	/** Internal objects (Struct objects) generated during the import. */
	protected LySimpleObjectList	  _internalObjectList;

	/** Class info. */
	protected LyClassInfo			  _classInfo;

	/** Fields to use for import (may be null for all fields). */
	protected LyInfoList<LyFieldInfo>	 _fields;

	/**
	 * Last error that was shown during import This may be used to check by code if import was Ok.
	 */
	protected String				  _lastErrorMessage;

	/**
	 * Constructor.
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller).
	 * @param action  related action.
	 */
	public LyImportContentController(LySession session, LyController parent, LyAction action)
	{
		super(session, parent, ((action == null) ? null : action.getId()));

		_action = action;
	}

	/**
	 * Accessor to the action associated to this controller. Area controllers are usually associated
	 * to an action, but this parameter may be null.
	 * @return The reference to associated action
	 */
	public LyAction getAction()
	{
		return _action;
	}

	/**
	 * Load the XML data from the specified buffer as an XML String.
	 * @param  buffer current buffer to read
	 * @return the list of imported objects or null if an error occured
	 */
	public LyObjectList loadData(String buffer)
	{
		if (buffer != null)
			return loadData(new StringReader(buffer));
		return null;
	}

	/**
	 * Load the XML data from the specified file.
	 * @param  filePath filename to read.
	 * @return the list of imported objects or null if an error occured
	 */
	public LyObjectList loadDataFile(String filePath)
	{
		LyEnvironment environment = getApplication().getEnvironment();
		//Reader                        reader = null;
		InputStream   istream;
		File		  file = new File(filePath);

		try
		{
			FileInputStream in = new FileInputStream(file);
//                       buffering for efficiency

			istream = new BufferedInputStream(in);
			//reader = new LyUnicodeDetectReader(in, null);
		}
		catch (FileNotFoundException notFound)
		{
			getParent().showError(environment.getMessage("LY_ER_GET_INPUT_STREAM", filePath));
			return null;
		}

		return loadData(istream);
	}

	/**
	 * Load the XML data from the specified stream.
	 * @param  istream current XML Stream to read
	 * @return the list of imported objects or null if an error occured
	 */
	public abstract LyObjectList loadData(InputStream istream);

	/**
	 * Load the XML data from the specified stream.
	 * @param  reader current XML Stream to read
	 * @return the list of imported objects or null if an error occured
	 */
	public abstract LyObjectList loadData(Reader reader);

	/**
	 * Accessor to the Internal objects (Struct objects) generated during the import This objects
	 * are used by imported objects in relations.
	 * @param  importedObject real imported object which caused the creation of these internal
	 *                        objects
	 * @return internal objects
	 */
	public abstract LyObjectList getInternalObjectList(LyObject importedObject);

	/**
	 * Class info : content description. (application class)
	 * @param classInfo the classInfo being imported
	 */
	public void setClassInfo(LyClassInfo classInfo)
	{
		_classInfo = classInfo;
	}

	/**
	 * Class info : content description. (application class)
	 * @return the classInfo being shown in this table
	 */
	public LyClassInfo getClassInfo()
	{
		return _classInfo;
	}

	/**
	 * Accessor to the list of fields for import.
	 * @param fields the list of fields to be read from the import file
	 */
	public void setFields(LyInfoList<LyFieldInfo> fields)
	{
		_fields = fields;
	}

	/**
	 * Accessor to the list of fields for import.
	 * @return the list of fields to be read from the import file
	 */
	public LyInfoList<LyFieldInfo> getFields()
	{
		return _fields;
	}

	/**
	 * Accessor to the view controller (Frame or dialog controller) which is the container of this
	 * view.
	 * @return the reference to a view controller, default returns null since no view is related to
	 *         this controller.
	 */
	public LyViewController getViewController()
	{
		return null;
	}

	/**
	 * Factory of the default builder for a controller.
	 * @return null since no view is related to this controller
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return null;
	}

	/**
	 * Accessor to the Last error that was shown during import This may be used to check by code if
	 * import was Ok.
	 * @return the last message or null if no error occured
	 */
	public String getLastErrorMessage()
	{
		return _lastErrorMessage;
	}

	/**
	 * Last error that was shown during import This may be used to check by code if import was Ok.
	 * @param msg the error message to set
	 */
	protected void setLastErrorMessage(String msg)
	{
		_lastErrorMessage = msg;
	}

} // end class LyImportContentController
