/**
 * LyImportXmlController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyImportContentXmlController.java 35179 2012-04-10 13:07:22Z jlsaouli $
 */

package leon.control.content;

import java.io.*;
import java.text.*;
import java.util.*;

import javax.xml.parsers.*;

import leon.app.*;
import leon.app.behavior.*;
import leon.control.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;

import org.w3c.dom.*;
import org.xml.sax.*;


/**
 * This class reads an XML File from export and create LyObjects from the import. The import is done
 * within a context (may cancel) all the import. Imported objects may change the identifier of the
 * imported objects (generate new identifiers) if overriden by the application Imported objetcs
 * returned by the loadData or loadDataFile are juste created / set in memory : it is possible to
 * commit/cancel import for objects following an object per object basis
 *
 * @see LyImportController
 * @see LyImportTableController
 */
public class LyImportContentXmlController extends LyImportContentController
{
	/**
	 * File version
	 */
	public static final String						  __VERSION =
		"$Id: LyImportContentXmlController.java 35179 2012-04-10 13:07:22Z jlsaouli $";

	/**
	 * Internal objects (Struct objects) generated during the import : key is imported object, value
	 * is an objectlist.
	 */
	protected Hashtable<LyObject, LySimpleObjectList> _internalObjectTable;

	/** Boolean indicating if the parser is validating or not. */
	protected boolean								  _validating;

	/**
	 * Constructor.
	 *
	 * @param session current session associated to the controller. The controller automatically
	 *                registers on the specified session.
	 * @param parent  parent controller (or null for 'root' controller).
	 * @param action  related action.
	 */
	public LyImportContentXmlController(LySession session, LyController parent,
		LyAction action)
	{
		super(session, parent, action);
	}

	/**
	 * Load the XML data from the specified stream.
	 *
	 * @param  istream current XML Stream to read
	 * @return the list of imported objects or null if an error occured
	 */
	public LyObjectList loadData(InputStream istream)
	{
		// Create the InputSource.
		InputSource source = new InputSource(istream);

		return loadData(source);
	}

	/**
	 * Load the XML data from the specified stream.
	 *
	 * @param  reader current XML Stream to read
	 * @return the list of imported objects or null if an error occured
	 */
	public LyObjectList loadData(Reader reader)
	{
		// Create the InputSource.
		InputSource source = new InputSource(reader);

		return loadData(source);
	}

	/**
	 * Load the XML data from the specified stream.
	 *
	 * @param  source current XML Stream to read
	 * @return the list of imported objects or null if an error occured
	 */
	public LyObjectList loadData(InputSource source)
	{
		if (source == null)
			return null; // input not found

		source = convertSourceVersion(source); // Convert model using XSLT

		LyEnvironment environment = getEnvironment();

		if (_parent instanceof LyImportTableController)
		{
            LyAction      action = ((LyImportTableController)_parent).getImportAction();
			String		  behaviorClassName = (action == null ? null
                : (String)action.getParameter("_importContentBehavior"));

			try
			{
				if (behaviorClassName != null)
				{
                    LyApplication   application = getApplication();
					Class<?>        c = application.loadClass(behaviorClassName);

					_behavior = (LyImportContentBehavior)c.newInstance();
				}
			}
			catch (Exception e)
			{
				if (environment.isDebugLevel(LyEnvironment.DEBUG_ERROR))
					environment.writeException(e);
			}
		}

		if (_behavior == null)
			_behavior = new LyImportContentBehavior();

		try
		{
			return parseXMLSource(source);
		}
		catch (Exception e)
		{
			environment.logError("Error during Import.");

			if (e instanceof SAXParseException)
			{
				String msg = "[LyImportContentXmlController] Line : "
					+ ((SAXParseException)e).getLineNumber() + ", column : "
					+ ((SAXParseException)e).getColumnNumber() + ".\n" + e.getMessage();

				setLastErrorMessage(msg);
				environment.logError(msg);
			}
			else
				setLastErrorMessage(e.getMessage());

			if (environment.isDebugLevel(LyEnvironment.DEBUG_ERROR))
				environment.writeException(e);

			return null;
		}
	} // end method loadData

	/**
	 * Utility method for XML Source parsing.
	 *
	 * @param  source current input stream for XML content
	 * @return the list of imported objects (object sdirectly imported). Internal objects used by
	 *         imported objects (reference to struct fieldinfos) are found using
	 *         getInternalObjectList()
	 * @throws SAXException
	 * @throws IOException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 */
	protected LyObjectList parseXMLSource(InputSource source)
		throws SAXException,
			IOException,
			InstantiationException,
			IllegalAccessException
	{
		// Create a new Parser.
		LyEnvironment environment = getEnvironment();

		setLastErrorMessage(null); // clear history

		if (!_behavior.startImport(this))
			return null;

		// ** Create Document builder factory **
		String	defaultFactory = environment.getEnv("_LY_XML_DOCUMENT_BUILDER_FACTORY");
		String	envFactoryProperty = System.getProperty("javax.xml.parsers.DocumentBuilderFactory");

		if (defaultFactory == null)
			defaultFactory = System.getProperty("javax.xml.parsers.DocumentBuilderFactory");

		if (defaultFactory != null)
			System.setProperty("javax.xml.parsers.DocumentBuilderFactory", defaultFactory);

		DocumentBuilderFactory factory = null;
		ClassLoader			   saveLoader = null;
		Thread				   thread = null;
		Document			   d = null;
		LyObjectList		   list = null;

		// First try in the current environment.
		try
		{
			factory = DocumentBuilderFactory.newInstance();
		}
		catch (Throwable e)
		{
			factory = null;
		}

		if (factory == null)
		{
			// Change the classLoader
			thread = Thread.currentThread();
			saveLoader = thread.getContextClassLoader();
			thread.setContextClassLoader(environment.getClassLoader());

			try
			{
				factory = DocumentBuilderFactory.newInstance();
			}
			catch (Throwable e)
			{
				factory = null;
			}
		}

		if (factory != null)
		{
			try
			{
				if (isValidating())
					factory.setValidating(true);

				// Use a validating parser. Comment this line to turn off
				// validation
				setParserFeatures(factory);
			}
			catch (SAXException e1)
			{
				environment.logInfo("BF AppEngine : Invalid parser feature : " + e1.getMessage());
			}

			try
			{
				DocumentBuilder parser = factory.newDocumentBuilder();
				// Parse in a DOM tree the import.
				d = parser.parse(source);

				// Just to be sure there is no several adjacent text nodes
				d.normalize();

				// now we have them all, try to build the LyObjects from the dom
				// Structure
				list = parseDocument(d);
				list = _behavior.completeImport(this, (LySimpleObjectList)list);
				if (!_behavior.endImport(this))
					list = null;
			}
			catch (Exception e)
			{
				setLastErrorMessage(e.getMessage());
				environment.writeException(e);
				list = null;
			}
		}

		// restore classLoader
		if (saveLoader != null)
			thread.setContextClassLoader(saveLoader);

		// Restore property
		if (envFactoryProperty == null)
		{
			// We must change the Properties attribute of System class because
			// there is no method removeProperty and we can't call setProperty method with
			// a null value.
			// The value must return to null because somebody could use another XML Parser.
			Properties props = System.getProperties();

			props.remove("javax.xml.parsers.DocumentBuilderFactory");
		}
		else
			System.setProperty("javax.xml.parsers.DocumentBuilderFactory", envFactoryProperty);

		return list;

	} // end method parseXMLSource

	/**
	 * Utility method for XML Source parsing.
	 *
	 * @param  d root DOM object after parsing
	 * @return the list of imported objects (object sdirectly imported).
	 * @throws SAXException
	 * @throws IOException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 */
	protected LyObjectList parseDocument(Document d)
		throws SAXException,
			IOException,
			InstantiationException,
			IllegalAccessException
	{
		LySimpleObjectList list = new LySimpleObjectList();

		_internalObjectTable = new Hashtable<LyObject, LySimpleObjectList>();
		Element rootElement = d.getDocumentElement();

		// Ensure that the root tag is what we expect : DATA

		if (rootElement.getTagName().equals("DATA"))
		{
			// Everything is fine so far, lets parse content
			NodeList childNodes = rootElement.getChildNodes();

			for (int i = 0; i < childNodes.getLength(); i++)
			{
				Node node = childNodes.item(i);

				if (node instanceof Element)
					parseClassElement((Element)node, list);
			}
		}
		else
		{
			String message = "The root tag should be <DATA> in import.";

			setLastErrorMessage(message);
			getEnvironment().logError(message);
		}
		return list;
	} // end method parseDocument

	/**
	 * Utility method for XML Source parsing.
	 *
	 * @param classElement DOM node representing a classinfo
	 * @param list         the list of imported objects (object sdirectly imported).
	 * @throws SAXException
	 * @throws IOException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 */
	protected void parseClassElement(Element classElement, LySimpleObjectList list)
		throws SAXException,
			IOException,
			InstantiationException,
			IllegalAccessException
	{
		// Try to find the related classInfo
		LyClassInfo classInfo = getClassInfo(classElement.getTagName());

		if ((getClassInfo() != null) && (getClassInfo() != classInfo))
			return; // This is not the classInfo we expect : skip it.

		if (classInfo == null)
		{
			String message = "ClassInfo " + classElement.getTagName() + " is not found.";

			setLastErrorMessage(message);
			getEnvironment().logError(message);
		}
		// Lets find some objects in this classinfo
		NodeList childNodes = classElement.getChildNodes();

		for (int i = 0; i < childNodes.getLength(); i++)
		{
			Node node = childNodes.item(i);

			if (node instanceof Element)
			{
				if (((Element)node).getTagName().equals("OBJECT"))
				{
					LyObject object = parseObjectElement((Element)node, classInfo, false);

					if (object != null)
						list.addObject(object);
				}
			}
		}
	} // end method parseClassElement

	/**
	 * Utility method for XML Source parsing.
	 *
	 * @param  objectElement   DOM node representing an object
	 * @param  classInfo      the classInfo of the imported object
	 * @param  isStructObject indicates if this is a 'real' classInfo or an internal struct
	 *                        classInfo.
	 * @return current object if imported or null if not found. NOTE : the Lyobject is set but
	 *         created/set for real yet.
	 * @throws SAXException
	 * @throws IOException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 */
	protected LyObject parseObjectElement(Element objectElement, LyClassInfo classInfo,
		boolean isStructObject)
		throws SAXException,
			IOException,
			InstantiationException,
			IllegalAccessException
	{
		LyObject object = (LyObject)classInfo.newInstance();
		LyValue  value, lastValue = null;

		object.setTemporary(true);

		// Go through the attributes of the object
		// Lets find some objects in this classinfo
		NodeList    childNodes = objectElement.getChildNodes();
		LyFieldInfo fieldInfo, lastFieldInfo = null;

		for (int i = 0; i < childNodes.getLength(); i++)
		{
			Node node = childNodes.item(i);

			if (node instanceof Element)
			{
				// For struct objects , the classId tag is *not* a fieldInfo
				if ((isStructObject) && (((Element)node).getTagName().equals("classId")))
					continue; // Skip this one

				fieldInfo = classInfo.getFieldInfo(((Element)node).getTagName());
				if (fieldInfo != null)
				{
					// Get current value
					value = parseFieldInfo((Element)node, object, fieldInfo, classInfo);

					if (lastFieldInfo == fieldInfo)
					{
						// We have a multiple value, gather previous
						// information and current
						if ((value instanceof LyRelationValue)
								&& (lastValue instanceof LyRelationValue))
						{
							LyObjectId[] values = null;

							LyObjectId[] valuesBefore = (LyObjectId[])lastValue.getValue();
							LyObjectId[] valuesToAdd = (LyObjectId[])value.getValue();

							values = new LyObjectId[valuesBefore.length + valuesToAdd.length];
							for (int i1 = 0; i1 < valuesBefore.length; i1++)
								values[i1] = valuesBefore[i1];
							for (int i1 = 0; i1 < valuesToAdd.length; i1++)
								values[valuesBefore.length + i1] = valuesToAdd[i1];

							value = fieldInfo.buildFieldValue(values);
						}
						else
							getEnvironment().logWarning(
								"[LyImportContentXmlController] rceived multiple value for the fieldInfo "
								+ fieldInfo.getId()
								+ " which is allowed only for relations. ClassInfo : "
								+ classInfo.getId());
					}
					else
					{
						// Commit lastValue (this is over now)
						if (lastValue != null)
						{
							object.skipEncoding(true);
							object.setFieldValue(lastValue);
							object.skipEncoding(false);
							lastValue = null; // Cleared
						}

						lastFieldInfo = fieldInfo;
					}
					lastValue = value;
				}
				else
					getEnvironment().logWarning(" Import XML : fieldinfo "
						+ ((Element)node).getTagName()
						+ " is not found for classInfo " + classInfo.getId()
						+ " : skipped.");
			}
		}

		// Commit lastValue (this is over now)
		if (lastValue != null)
		{
			object.skipEncoding(true);
			object.setFieldValue(lastValue);
			object.skipEncoding(false);
		}

		return object;
	} // end method parseObjectElement

	/**
	 * Utility method for XML Source parsing of a fieldInfo value.
	 *
	 * @param  fieldElement DOM node representing a fieldInfo and its value
	 * @param  object       the imported object
	 * @param  fieldInfo    the fieldInfo of the imported object
	 * @param  classInfo    the classInfo of the imported object
	 * @return current value to set the imported object.
	 * @throws SAXException
	 * @throws IOException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 */
	protected LyValue parseFieldInfo(Element fieldElement, LyObject object,
		LyFieldInfo fieldInfo, LyClassInfo classInfo)
		throws SAXException,
			IOException,
			InstantiationException,
			IllegalAccessException
	{
		// we can have here a value, a Struct or an Array
		if (fieldInfo instanceof LyStructFieldInfo)
			return parseStructFieldInfo(fieldElement, object, (LyStructFieldInfo)fieldInfo,
				classInfo);
		else if (fieldInfo instanceof LyArrayFieldInfo)
			return parseArrayFieldInfo(fieldElement, object, (LyArrayFieldInfo)fieldInfo,
				classInfo);
		else if (fieldInfo instanceof LyRelationFieldInfo)
			return parseRelationFieldInfo(fieldElement, object, (LyRelationFieldInfo)fieldInfo,
				classInfo);
		// General case
		Node   childNode = fieldElement.getFirstChild();
		String value = null;

		// Get ready to build a value
		if (childNode instanceof Text)
		{
			value = ((Text)childNode).getData();
			if ((value != null) && (value.trim().equals("")))
				value = null;
			else
				value = decodeString(value);
			// Specific behavior may alter this value... (ie ids)
			value = _behavior.getFieldValue(this, classInfo, fieldInfo, value);

			// Specific encoding for dateFieldInfo
			if (fieldInfo instanceof LyDateFieldInfo)
				return decodeDateField((LyDateFieldInfo)fieldInfo, value);

			return fieldInfo.buildFieldValue(value);
		}

		return null;
	} // end method parseFieldInfo

	/**
	 * Decode a date field in the xml export format.
	 *
	 * @param  dateField The date field to export.
	 * @param  value     The value of the date field.
	 * @return The formatted date field.
	 */
	protected LyValue decodeDateField(LyDateFieldInfo dateField, String value)
	{
		Date				 date = null;
		LyDateFieldInfo.Type type = dateField.getType();
		SimpleDateFormat     dateFormat;

		try
		{
			if (type == LyDateFieldInfo.Type.DATE)
			{
				dateFormat = new SimpleDateFormat("yyyy-MM-dd");
				date = dateFormat.parse(value);
			}
			else if ((type == LyDateFieldInfo.Type.TIME)
					|| (type == LyDateFieldInfo.Type.HOUR_MIN))
			{
				dateFormat = new SimpleDateFormat("HH:mm:ss");
				date = dateFormat.parse(value);
			}
			else
			{
				dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
				date = dateFormat.parse(value);
			}
		}
		catch (ParseException e)
		{
			String msg = "[LyImportContentXmlController] Parse format error for LyDateFieldInfo : "
				+ dateField.getId() + " and value = " + value;

			setLastErrorMessage(msg);
			getEnvironment().logError(msg);
		}
		return dateField.buildFieldValue(date);
	} // end method decodeDateField

	/**
	 * Utility method for XML Source parsing of a relation fieldInfo value.
	 *
	 * @param  fieldElement DOM node representing a fieldInfo and its value
	 * @param  object       the imported object
	 * @param  fieldInfo    the fieldInfo of the imported object
	 * @param  classInfo    the classInfo of the imported object
	 * @return current value to set the imported object.
	 * @throws SAXException
	 * @throws IOException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 */
	protected LyValue parseRelationFieldInfo(Element fieldElement, LyObject object,
		LyRelationFieldInfo fieldInfo, LyClassInfo classInfo)
		throws SAXException,
			IOException,
			InstantiationException,
			IllegalAccessException
	{
		// whe have <classId>xxx</classId><objectId>yyy</objectId>
		// Go through the subelements of the field tag
		NodeList childNodes = fieldElement.getChildNodes();
		String   classInfoName = null, objectIdValue = null;

		for (int i = 0; i < childNodes.getLength(); i++)
		{
			Node node = childNodes.item(i);

			if (node instanceof Element)
			{
				// For struct objects , the classId tag is *not* a fieldInfo
				if (((Element)node).getTagName().equals("classId"))
				{
					Node subNode = node.getFirstChild();

					if (subNode instanceof Text)
						classInfoName = ((Text)subNode).getData();
				}
				// For struct objects , the classId tag is *not* a fieldInfo
				if (((Element)node).getTagName().equals("objectId"))
				{
					Node subNode = node.getFirstChild();

					if (subNode instanceof Text)
						objectIdValue = ((Text)subNode).getData();
				}
			}
		}
		if ((classInfoName != null) && (objectIdValue != null))
		{
			LyClassInfo relationClassInfo = getClassInfo(classInfoName);

			objectIdValue = decodeString(objectIdValue);
			LyObjectId objectId = decodeObjectIdValue(relationClassInfo, objectIdValue);

			if (relationClassInfo != null)
				return fieldInfo.buildFieldValue(objectId);
		}
		return null;
	} // end method parseRelationFieldInfo

	/**
	 * Utility method for XML Source parsing.
	 *
	 * @param  fieldElement DOM node representing a fieldInfo and its value
	 * @param  object       the imported object
	 * @param  fieldInfo    the fieldInfo of the imported object
	 * @param  classInfo    the classInfo of the imported object
	 * @return current value to set the imported object.
	 * @throws SAXException
	 * @throws IOException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 */
	protected LyValue parseStructFieldInfo(Element fieldElement, LyObject object,
		LyStructFieldInfo fieldInfo, LyClassInfo classInfo)
		throws SAXException,
			IOException,
			InstantiationException,
			IllegalAccessException
	{
		// Extract the classInfo and build the struct object
		NodeList childNodes = fieldElement.getChildNodes();
		// Step 1 : get the classInfo for the struct object
		String   structClassInfoName = null;

		for (int i = 0; (i < childNodes.getLength()) && (structClassInfoName == null); i++)
		{
			Node node = childNodes.item(i);

			if ((node instanceof Element)
					&& (((Element)node).getTagName().equals("classId")))
			{
				Node subNode = node.getFirstChild();

				if (subNode instanceof Text)
					structClassInfoName = ((Text)subNode).getData();
			}
		}
		LyClassInfo structClassInfo = getClassInfo(structClassInfoName);

		if (structClassInfo == null)
		{
			String msg = "[LyImportContentXmlController] Struct ClassInfo "
				+ structClassInfoName + " is not found.";

			setLastErrorMessage(msg);
			getEnvironment().logError(msg);
		}

		// Step 2 : build the structObject (recurse on object creation)
		LyObject structObject = parseObjectElement(fieldElement, structClassInfo, true);
		// Keep this referenced object as a related object
		if ((object != null) && (structObject != null))
			addInternalObject(object, structObject);

		return fieldInfo.buildFieldValue(structObject);
	} // end method parseStructFieldInfo

	/**
	 * Utility method for XML Source parsing.
	 *
	 * @param  fieldElement DOM node representing a fieldInfo and its value
	 * @param  object       the imported object
	 * @param  fieldInfo    the fieldInfo of the imported object
	 * @param  classInfo    the classInfo of the imported object
	 * @return current value to set the imported object.
	 * @throws SAXException
	 * @throws IOException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 */
	protected LyValue parseArrayFieldInfo(Element fieldElement, LyObject object,
		LyArrayFieldInfo fieldInfo, LyClassInfo classInfo)
		throws SAXException,
			IOException,
			InstantiationException,
			IllegalAccessException
	{
		// Extract the classInfo and build the struct object
		NodeList	    childNodes = fieldElement.getChildNodes();
		Vector<LyValue> values = new Vector<LyValue>();
		LyFieldInfo     innerFieldInfo = fieldInfo.getFieldInfo();

		for (int i = 0; i < childNodes.getLength(); i++)
		{
			Node node = childNodes.item(i); // This should be a sequence of the
			// same fieldInfo

			if (node instanceof Element)
				// Ok, we have one instance of the Array item
				values.addElement(parseFieldInfo((Element)node, object, innerFieldInfo, classInfo));
		}
		Object[] val = null;

		if (values.size() > 0)
		{
			LyValue value;

			val = new Object[values.size()];

			for (int i = 0; i < values.size(); i++)
			{
				value = values.elementAt(i);
				val[i] = value.getValue();
			}
		}

		return fieldInfo.buildFieldValue(val);
	} // end method parseArrayFieldInfo

	/**
	 * Method to decode a string.
	 *
	 * @param  string The string to decode.
	 * @return The string decoded.
	 */
	protected String decodeString(String string)
	{
		string = LyString.replace(string, "&lt;", "<");
		string = LyString.replace(string, "&gt;", ">");
		string = LyString.replace(string, "&apos;", "'");
		string = LyString.replace(string, "&amp;", "&");
		string = LyString.replace(string, "&quot;", "\"");

		//string = string.replace(LyTools.charFromString(_valueSep),
		//      LyTools.ARRAY_SEP);

		return asciiDecode(string);
	}

	/**
	 * Encode String in a qutoted printable like manner Avoid conflics in file.
	 * @param str the string to decode
	 * @return the decoded string
	 */
	protected String asciiDecode(String str)
	{
		int			 i, n = str.length();
		char		 c;
		StringBuffer sb = new StringBuffer(n);

		for (i = 0; i < n; i++)
		{
			c = str.charAt(i);

			if (c == '=')
			{
				try
				{
					String val = str.substring(i + 1, i + 3);
					int    skip = 2; // Skip next 2 chars

					if (val.equals("\\u"))
					{
						val = str.substring(i + 3, i + 7); // using unicode 4
						// letters in hexa
						// \uA1FE
						skip = 6; // Skip next 6 chars
					}
					String  hexStr = "0x" + val;
					Integer asciiCode = Integer.decode(hexStr);

					sb.append((char)asciiCode.intValue());

					i += skip;
				}
				catch (NumberFormatException e)
				{
					sb.append(c);
					getEnvironment().writeException(e);
				}

			}
			else
				sb.append(c);
		}

		return sb.toString();
	} // end method asciiDecode

	/**
	 * Method to decode an objectId value.
	 *
	 * @param  classInfo     Class of the objectId.
	 * @param  objectIdValue Value of the objectId. The specific XML decoding has been done.
	 * @return An id representing the decode value for the specified classInfo.
	 */
	protected LyObjectId decodeObjectIdValue(LyClassInfo classInfo, String objectIdValue)
	{
		return new LyObjectId(classInfo, objectIdValue);
	}

	/**
	 * Accessor to the Internal objects (Struct objects) generated during the import.
	 *
	 * @param importedObject real imported object which caused the creation of these internal
	 *                       objects
	 * @param internalObject the new internal object
	 */
	public void addInternalObject(LyObject importedObject, LyObject internalObject)
	{
		LySimpleObjectList list = _internalObjectTable.get(importedObject);

		if (list == null)
		{
			list = new LySimpleObjectList();
			_internalObjectTable.put(importedObject, list);
		}
		list.addObject(internalObject);
	}

	/**
	 * Accessor to the Internal objects (Struct objects) generated during the import This objects
	 * are used by imported objects in relations.
	 *
	 * @param  importedObject real imported object which caused the creation of these internal
	 *                        objects
	 * @return internal objects
	 */
	public LyObjectList getInternalObjectList(LyObject importedObject)
	{
		return _internalObjectTable.get(importedObject);
	}

	/**
	 * Accessor to the view controller (Frame or dialog controller) which is the container of this
	 * view.
	 *
	 * @return the reference to a view controller, default returns null since no view is related to
	 *         this controller.
	 */
	public LyViewController getViewController()
	{
		return null;
	}

	/**
	 * Factory of the default builder for a controller.
	 *
	 * @return null since no view is related to this controller
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return null;
	}

	/**
	 * Checks whether this parser is validating or not.
	 *
	 * @return True if validating, false otherwise.
	 */
	public boolean isValidating()
	{
		return _validating;
	}

	/**
	 * Sets whether this parser is validating or not.
	 *
	 * @param validating True if validating, false otherwise.
	 */
	public void setValidating(boolean validating)
	{
		_validating = validating;
	}

	/**
	 * Set the features of the parser.
	 *
	 * @param factory the xml parser
	 * @throws SAXException
	 */
	protected void setParserFeatures(DocumentBuilderFactory factory)
		throws SAXException
	{
		// Use a validating parser. Comment this line to turn off validation
		factory.setAttribute("http://xml.org/sax/features/validation", Boolean.valueOf(_validating));
		factory.setAttribute("http://apache.org/xml/features/validation/schema",
				Boolean.valueOf(_validating));
	}

	/**
	 * Apply XSLT Stylesheets to convert the source version of the XML File to latest version.
	 *
	 * @param  source source for parsing (input version)
	 * @return source the nw source after parsing (output version)
	 */
	public InputSource convertSourceVersion(InputSource source)
	{
		return source;
	}
} // end class LyImportContentXmlController
