/**
 * LyUnicodeDetectReader.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyUnicodeDetectReader.java 28422 2010-11-03 12:03:38Z ehelmste $
 */
package leon.control.content;

import java.io.*;


/**
 * Generic detect unicode textreader, which will use BOM mark to identify the encoding to be used.
 */
public class LyUnicodeDetectReader extends Reader
{
	/**
	 * File version
	 */
	public static final String    __VERSION =
		"$Id: LyUnicodeDetectReader.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Pushback input stream */
	protected PushbackInputStream internalIn;
	
	/** Input Stream reader */
	protected InputStreamReader   internalIn2 = null;
	
	/** Default encoding */
	protected String			  defaultEnc;

	private static final int	  BOM_SIZE = 4;

	/**
	 * Constructor.
	 * @param in         the inpustream to test
	 * @param defaultEnc may be null, default encoding to use if no detection could be done.
	 */
	protected LyUnicodeDetectReader(InputStream in, String defaultEnc)
	{
		internalIn = new PushbackInputStream(in, BOM_SIZE);
		this.defaultEnc = defaultEnc;
	}

	/**
	 * Get the default encoding
	 * @return the default encoding
	 */
	public String getDefaultEncoding()
	{
		return defaultEnc;
	}

	/**
	 * Get encoding
	 * @return the current encoding
	 */
	public String getEncoding()
	{
		if (internalIn2 == null)
			return null;
		return internalIn2.getEncoding();
	}

	/**
	 * Read-ahead four bytes and check for BOM marks. Extra bytes are unread back to the stream,
	 * only BOM bytes are skipped.
	 * @throws IOException 
	 */
	protected void init()
		throws IOException
	{
		if (internalIn2 != null)
			return;

		String encoding;
		byte[] bom = new byte[BOM_SIZE];
		int    n, unread;

		n = internalIn.read(bom, 0, bom.length);

		if ((bom[0] == (byte)0xEF) && (bom[1] == (byte)0xBB)
				&& (bom[2] == (byte)0xBF))
		{
			encoding = "UTF-8";
			unread = n - 3;
		}
		else if ((bom[0] == (byte)0xFE) && (bom[1] == (byte)0xFF))
		{
			encoding = "UTF-16BE";
			unread = n - 2;
		}
		else if ((bom[0] == (byte)0xFF) && (bom[1] == (byte)0xFE))
		{
			encoding = "UTF-16LE";
			unread = n - 2;
		}
		else if ((bom[0] == (byte)0x00) && (bom[1] == (byte)0x00)
				&& (bom[2] == (byte)0xFE) && (bom[3] == (byte)0xFF))
		{
			encoding = "UTF-32BE";
			unread = n - 4;
		}
		else if ((bom[0] == (byte)0xFF) && (bom[1] == (byte)0xFE)
				&& (bom[2] == (byte)0x00) && (bom[3] == (byte)0x00))
		{
			encoding = "UTF-32LE";
			unread = n - 4;
		}
		else
		{
			// Unicode BOM mark not found, unread all bytes
			encoding = defaultEnc;
			unread = n;
		}
		if (unread > 0)
			internalIn.unread(bom, (n - unread), unread);
		else if (unread < -1)
			internalIn.unread(bom, 0, 0);
		// Use given encoding
		if (encoding == null)
			internalIn2 = new InputStreamReader(internalIn);
		else
			internalIn2 = new InputStreamReader(internalIn, encoding);
	} // end method init

	/**
	 * Close this reader
	 * @throws IOException 
	 * 
	 */
	public void close()
		throws IOException
	{
		init();
		internalIn2.close();
	}

	/**
	 * @see java.io.Reader#read(char[], int, int)
	 */
	public int read(char[] cbuf, int off, int len)
		throws IOException
	{
		init();
		return internalIn2.read(cbuf, off, len);
	}
} // end class LyUnicodeDetectReader
