/**
 * LyChoiceFieldConsultController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyChoiceFieldConsultController.java 34940 2012-03-22 11:29:34Z jlsaouli $
 */

package leon.control.form;

import java.util.*;

import leon.app.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.setinterface.*;


/**
 * Field controller for consultation of LyChoiceFieldInfo.
 * @see LyConsultController
 */
public class LyChoiceFieldConsultController extends LyFieldConsultController
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyChoiceFieldConsultController.java 34940 2012-03-22 11:29:34Z jlsaouli $";

	/**
	 * Constructor.
	 * @param parent parent controller
	 * @param fieldInfo choice field handled by this controller
	 */
	public LyChoiceFieldConsultController(LyFormController parent, LyChoiceFieldInfo fieldInfo)
	{
		super(parent, fieldInfo);
	}

	/**
	 * Update the value of a fieldInfo shown in the view.
	 * @param newValue the new value of the field.
	 * @param oldValue former value of the field,or null if unknown.
	 */
	public void updateValue(Object newValue, Object oldValue)
	{
		LyChoiceFieldInfo choiceField = (LyChoiceFieldInfo)_fieldInfo;
		LyFieldItem		  fieldItem = getFieldItem();

		if (!choiceField.isMultiple()
			|| (fieldItem instanceof LyLabel))
		{
			super.updateValue(newValue, oldValue);
			return;
		}

		if (!(fieldItem instanceof LyText))
			return;

		String   labelStr;
		LyObject object = getObject();

		if (object != null)
		{
			labelStr = _formController.getLabel(object, choiceField, newValue);
			String toName = choiceField.toName(newValue);

			if ((toName != null) && toName.equals(labelStr)
					&& (labelStr.indexOf(LyApplication.VISIBLE_FIELD_SEP) >= 0))
				// We need to build manually the content...
				labelStr = buildChoiceValue(newValue);
		}
		else
		{
			if ((choiceField.getFieldUnit() != null)
					&& (choiceField.getFieldUnit().getFormat() != null))
				labelStr = choiceField.toName(newValue);
			else
				labelStr = buildChoiceValue(newValue);

		}
		((LyText)fieldItem).setLabel(labelStr);
	} // end method updateValue

	/**
	 * Build an equivalent choice string.
	 * @param  newValue the value to display
	 * @return the formatted string
	 */
	private String buildChoiceValue(Object newValue)
	{
		LyChoiceFieldInfo choiceField = (LyChoiceFieldInfo)_fieldInfo;
		String			  labelStr = null;
		LyChoiceValue     value = (LyChoiceValue)choiceField.buildFieldValue(newValue);
		int				  n = value.getValueCount();

		for (int i = 0; i < n; i++)
		{
			if (labelStr == null)
				labelStr = value.getName(i);
			else
				labelStr += "\n" + value.getName(i);
		}
		return labelStr;
	}

	/**
	 * Impact of a value change of a fieldInfo. The method invokes LyClassBehavior.propagate which
	 * may be used to implement specific impact on a value changed. This method is for internal use,
	 * application should use setFieldValue instead.
	 * @param value     new Value of the field.
	 * @see   leon.app.behavior.LyClassBehavior#propagate
	 */
	public void propagate(Object value)
	{
		// Execute rules on choice
		LyChoiceFieldInfo choice = (LyChoiceFieldInfo)_fieldInfo;
		LyApplication     application = getApplication();
		LyObject		  object = getObject();
		int				  priority = application.getPriorityLevel(object);
		LyRule			  rule, registeredRule;
		boolean			  checked;
		short			  i, nr;
		short			  j, n = choice.getChoiceOptionCount();
		LyChoiceOption    option;
		short			  p;
		LyEnvironment     environment = getEnvironment();
		LyInfoDataList<LyRule, Boolean>    rules = new LyInfoDataList<LyRule, Boolean>();
        int               k;

		// Two steps : first apply rules on options that are unchecked
		// and then on checked options
		for (p = 0; p < 2; p++)
		{
			for (i = 0; i < n; i++)
			{
				option = choice.getChoiceOption(i);
				nr = option.getRuleCount();

				if (nr == 0)
					continue;

				if (value instanceof BitSet)
					checked = ((BitSet)value).get(i);
				else if (value instanceof Short)
					checked = (((Short)value).shortValue() == i);
				else if (value instanceof Boolean)
					checked = (((Boolean)value).booleanValue() && (i == 1))
							|| (!((Boolean)value).booleanValue() && (i == 0));
				else
					checked = false;

				if (((p == 0) && checked)
						|| ((p == 1) && !checked))
					continue;

                if (p == 0)
                    environment.logTrace(">>> Apply reverse rule for " + _fieldInfo);
                else
                    environment.logTrace(">>> Apply rule for " + _fieldInfo);

				for (j = 0; j < nr; j++)
				{
					rule = option.getRule(j);

					if (rule.getPriority() > priority)
						continue;

					if (!(rule.getObjectInfo() instanceof LyFieldInfo))
						continue;

					// Check if rules on the same field and with the same operation already exist.
                    // If so, remove them from rules table because this new rule will cancel them.
                    for (k = rules.getSize() - 1; k >= 0; k--)
					{
						registeredRule = rules.getInfo(k);
						if ((rule.getObjectInfo() == registeredRule.getObjectInfo())
								&& (rule.getOperation() == registeredRule.getOperation()))
							rules.removeInfo(registeredRule);
					}

					rules.addInfo(rule, Boolean.valueOf(checked));
				}
			}
		}

		// Execute rules stored in the rules table
        boolean           hasVisibleRule = false;

		for (i = 0, nr = (short)rules.getSize(); i < nr; i++)
		{
			rule = rules.getInfo(i);
            checked = rules.getData(i).booleanValue();
            if (rule.getOperation() == LyRule.Operation.SET_VISIBLE)
                hasVisibleRule = true;

			((LyConsultController)_formController).execRule(rule, checked);
		}

		// If at least one rule shows or hide a field, reset form to compute
		// its new content and size.
		if (hasVisibleRule)
		{
			LyForm form = ((_formField.getParent() instanceof LyForm)
					? (LyForm)_formField.getParent() : null);

			if (form != null)
			{
				LyViewComponent internal = form.getInternalView();

				if (internal instanceof LyResetContentInterface)
					((LyResetContentInterface)internal).resetContent();
			}
		}

		super.propagate(value);
	} // end method propagate
} // end class LyChoiceFieldConsultController
