/**
 * LyChoiceFieldSetController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyChoiceFieldSetController.java 40861 2012-09-19 09:02:50Z jlsaouli $
 */

package leon.control.form;

import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.control.form.builder.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.setinterface.*;


/**
 * Field controller for modification of LyChoiceFieldInfo.
 * @see LySetController
 */
public class LyChoiceFieldSetController extends LyFieldSetController
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyChoiceFieldSetController.java 40861 2012-09-19 09:02:50Z jlsaouli $";

	/**
	 * Constructor.
	 * @param parent parent set controller
	 * @param fieldInfo choice field handled by this controller
	 */
	public LyChoiceFieldSetController(LySetController parent, LyChoiceFieldInfo fieldInfo)
	{
		super(parent, fieldInfo);
	}

	/**
	 * Extract current value for the specified fieldInfo in the form.
	 * @param  unit      Unit of the value.
	 * @return current value, or null if not found or field is disabled.
	 */
	public LyValue getFieldValue(LyUnit unit)
	{
		LyFieldItem fieldItem = getFieldItem();

		if (fieldItem instanceof LyBoolean)
		{
			LyBoolean        booleanField = (LyBoolean)fieldItem;
			LyBoolean.Status status = booleanField.getStatus();
			Boolean          value;

			if (status == LyBoolean.Status.EMPTY)
				value = null;
			else if (status == LyBoolean.Status.CHECKED)
				value = Boolean.TRUE;
			else
				value = Boolean.FALSE;

			return _fieldInfo.buildFieldValue(value);
		}
		else if (fieldItem instanceof LyChoice)
		{
			LyChoice            choice = (LyChoice)fieldItem;
			int		            n = choice.getOptionCount();

			if (choice.isMultiple()
		        && (n == 1)
		        && !_fieldInfo.isMultiple()
				&& !hasMark(_fieldInfo, LyFieldInfo.MARK_OPTIONAL))
			{
				Short    value;
				LyOption option = choice.getOption(0);

				value = new Short((short)(option.isChecked() ? 1 : 0));
				return _fieldInfo.buildFieldValue(value);
			}
		}

		return super.getFieldValue(unit);
	} // end method getFieldValue

	/**
	 * @see leon.control.form.LyFieldSetController#getStringValue()
	 */
	protected String getStringValue()
	{
		LyFieldItem   fieldItem = getFieldItem();
		String		  strValue = null;
		LyApplication application = getApplication();

		if (fieldItem instanceof LyDictionary)
		{
			// Scan the dictionnary tree and add the selected items...
			LyDictionary dico = (LyDictionary)fieldItem;
			LyEntry		 entry;
			StringBuffer sb = null;
			Vector<LyEntry>		 result = new Vector<LyEntry>();
			int			 i, n;
			boolean		 skipGroups = !((LyChoiceFieldInfo)_fieldInfo).doGroupsHaveValue();

			dico.getComponents(LyEntry.class, result);
			n = result.size();

			for (i = 0; i < n; i++)
			{
				entry = result.elementAt(i);
				if (skipGroups
						&& (entry.getChildCount() > 0)) // skip groups
					continue;

				if (!entry.isChecked())
					continue;
				if (sb == null)
					sb = new StringBuffer(entry.getId());
				else
					sb.append(application.VALUE_SEP).append(entry.getId());
			}
			if (sb != null)
				strValue = sb.toString();
		}
		else if (fieldItem instanceof LyChoice)
		{
			LyChoice     choice = (LyChoice)fieldItem;
			int			 i, n = choice.getOptionCount();
			LyOption     option;
			StringBuffer sb = null;

			for (i = 0; i < n; i++)
			{
				option = choice.getOption(i);
				if (option.isChecked()
						&& !option.getId().equals("LY_NO_CHOICE"))
				{
					if (sb == null)
						sb = new StringBuffer(option.getId());
					else
						sb.append(application.VALUE_SEP).append(option.getId());
				}

				if (sb != null)
					strValue = sb.toString();
			}
		}
		else if (fieldItem instanceof LyRelation)
		{
			LyRelation   relation = (LyRelation)fieldItem;
			int			 i, n = relation.getLinkCount();
			LyLink		 link;
			StringBuffer sb = null;

			for (i = 0; i < n; i++)
			{
				link = relation.getLink(i);

				if (link.isSelected()
						&& !link.getId().equals("LY_NO_CHOICE"))
				{
					if (sb == null)
						sb = new StringBuffer(link.getId());
					else
						sb.append(application.VALUE_SEP).append(link.getId());
				}
			}

			if (!_setController.isCreation()
					&& _fieldInfo.isMultiple())
			{
				// Add values that were "selected" but not shown
				LyObject object = getObject();

				if (object != null)
				{
					LyChoiceFieldInfo choice = (LyChoiceFieldInfo)_fieldInfo;
					int				  no = choice.getChoiceOptionCount();
					LyChoiceValue     choiceValue = object.getChoiceValue(choice);

					if (((no > n) || ((no == n) && !_fieldInfo.isMultiple()))
							&& (choiceValue.getValueCount() > 0))
					{
						LyChoiceOption option;
						short		   o;

						for (o = 0; o < no; o++)
						{
							option = choice.getChoiceOption(o);
							if (choiceValue.isSelected(o) // selected
									&& (relation.getComponent(option.getId()) == null)) // but not found
							{
								if (sb == null)
									sb = new StringBuffer(option.getId());
								else
									sb.append(application.VALUE_SEP).append(option.getId());
							}
						}
					}
				}
			}

			if (sb != null)
				strValue = sb.toString();
		}

		return strValue;
	} // end method getStringValue

	/**
	 * Internal use : update the value of a choice fieldInfo in the form. <B>Application should use
	 * setFieldValue instead.</B> This method is called from updateFieldItem. (Specialization)
	 * @param newValue     the new value of the field. This value may be null, a LyValue instance or
	 *                  the internal value associated to the fieldInfo. See LyValue.getValue()
	 * @param oldValue  the latest value of the field. This value may be null, a LyValue instance or
	 *                  the internal value associated to the fieldInfo. See LyValue.getValue()
	 * @see   #setFieldValue
	 * @see   #updateFieldItem
	 */
	protected void updateValue(Object newValue, Object oldValue)
	{
	    if (_setController.isConsult(_fieldInfo))
	    {
	        super.updateValue(newValue, oldValue);
	        return;
	    }

		LyFormField formField = getFormField();

		if (formField == null)
			return;

		LyChoiceFieldInfo	   choiceField = (LyChoiceFieldInfo)_fieldInfo;
		LyFieldItem			   fieldItem = getFieldItem();
		LyClassInfo			   classInfo = _setController.getClassInfo();
		boolean				   multiple = choiceField.isMultiple();
		short				   i, n = choiceField.getChoiceOptionCount();
		LyChoiceOption		   choiceOption;
		boolean				   checked, wasChecked;
		LySetBehaviorInterface behavior = _setController.getSetBehavior();
		Vector<LyChoiceOption> domain = behavior.getChoiceDomain(_setController, choiceField);
		boolean				   hasRule = (classInfo.getRuleCount() > 0);
		short				   nChecked = 0;
		boolean				   wasEnabled = true;

		if (fieldItem instanceof LyChoice)
		{
			LyChoice choice = (LyChoice)fieldItem;
			LyOption option;

			if (!multiple && choice.isMultiple() && (choice.getOptionCount() == 1))
			{
				option = choice.getOption(0);
				wasEnabled = option.isEnabled();
				wasChecked = option.isChecked();

				checked = (newValue != null) && (((Short)newValue).shortValue() == (n - 1));

				if (wasEnabled
						&& ((domain == null) || (domain.size() != n)))
					choice.enableOption(0, false);
				else if (!wasEnabled)
					choice.enableOption(0, true);

				if ((checked != wasChecked) || !wasEnabled)
					choice.checkOption(0, checked);
			}
			else
			{
				LyOption firstEnabledOption = null;

				for (i = 0; i < n; i++)
				{
					choiceOption = choiceField.getChoiceOption(i);
					option = choice.getOption(choiceOption.getId());
					if (option == null)
						continue;

					hasRule = hasRule || (choiceOption.getRuleCount() > 0);
					wasEnabled = option.isEnabled();
					wasChecked = option.isChecked();

					if (!multiple)
						checked = (newValue != null) && (((Short)newValue).shortValue() == i);
					else
						checked = (newValue != null) && ((BitSet)newValue).get(i);

					if ((domain == null)
							|| !domain.contains(choiceOption))
						choice.enableOption(option, false);
					else
					{
						choice.enableOption(option, true);
						if (firstEnabledOption == null)
							firstEnabledOption = option;
					}

					if ((checked != wasChecked) || !wasEnabled)
						// Check option
						choice.checkOption(option, checked);

					if (option.isChecked()
							&& !option.isEnabled()
							&& (choiceField.getContextField() != null))
						choice.checkOption(option, false);

					if (option.isChecked())
						nChecked++;
				}

				if (!multiple && (nChecked == 0))
				{
					if ((choice.getOptionCount() > n)
							&& choice.getOption(n).isEnabled())
						choice.checkOption(n, true);
					else if (!_formField.isOptional() || (newValue != null)) // Accept null if formfield is set optional
					{
						Short def = (Short)choiceField.getDefaultValue();

						if (def != null)
						{
							choiceOption = choiceField.getChoiceOption(def.shortValue());
							option = choice.getOption(choiceOption.getId());
						}
						else
							option = null; // Pas de slection , aucune valeur trouve
						if ((option != null)
								&& option.isEnabled())
							choice.checkOption(option, true);
					}
				}
				else if (!multiple && (nChecked == 1))
				{
					if (choice.getOptionCount() > n)
						choice.checkOption(n, false);
				}
			}
		}
		else if (fieldItem instanceof LyRelation)
		{
			// FFT00191 : optimizing relation updates
			LyRelation relation = (LyRelation)fieldItem;
			int		   j, nRelation = relation.getLinkCount();
			String     optionId;
			LyLink     link;
			boolean    emptyLink = !multiple && (nRelation > 0)
					&& relation.getLink(0).getId().equals("LY_NO_CHOICE");

			// Step 1 : parse the relation and remove options that are no
			// longer in the domain.
			for (j = nRelation - 1; j >= 0; j--)
			{
				link = relation.getLink(j);
				optionId = link.getId();

				if (link.getId().equals("LY_NO_CHOICE"))
					continue;

				choiceOption = choiceField.getChoiceOption(optionId);

				if ((domain == null) || !domain.contains(choiceOption))
					relation.removeLink(link);
			}

			// Step 2 : add options that are in the domain and not in the relation
			if (domain != null)
			{
				int     offset = (emptyLink ? 1 : 0);
				int     nDomain = domain.size();
				boolean found;

				for (i = 0; i < nDomain; i++)
				{
					nRelation = relation.getLinkCount();
					choiceOption = domain.elementAt(i);
					optionId = choiceOption.getId();

					for (j = offset, found = false; j < nRelation; j++)
					{
						link = relation.getLink(j);
						found = optionId.equals(link.getId());

						if (found)
							break;
					}

					if (!found)
					{
						LyInfoLabel infoLabel = choiceOption.getLabel(LyInfoLabel.Type.SEMANTIC);
						LyObject    object = _formController.getObject();
						String	    name;
						String	    label = null;

						if (_parent instanceof LyEditTableSetController)
							name = choiceOption.getAlias();
						else
						{
							name = choiceOption.getName();	
							if (infoLabel != null)
								label = infoLabel.getString(object);
						}
						
						link = new LyLink(optionId,
								(label == null ? name : (name + " : " + label)));
						if ((i + offset) >= nRelation)
							relation.addLink(link);
						else
							relation.insertLink(link, i + offset);
					}
				}
			}

			// Step 3 : update links selection
			short index;

			// Add an empty link when needed
            boolean       optional = _setController.isOptional(choiceField);

			if (!multiple && !emptyLink)
			{
				LyEnvironment environment = getEnvironment();
				String		  empty_value = "LY_CHOICE_EMPTY_VALUE";
				String		  val = environment.translate(empty_value);

				if (optional
			        || ((val != null)
						&& !val.equals("_null")))
				{
					if (optional
							&& (val != null)
							&& val.equals("_null"))
						empty_value = null;

					link = new LyLink("LY_NO_CHOICE", empty_value);
					relation.insertLink(link, 0);
					emptyLink = true;
				}
			}

			nRelation = relation.getLinkCount();
			for (j = nRelation - 1; j >= 0; j--)
			{
				link = relation.getLink(j);
				optionId = link.getId();
				choiceOption = choiceField.getChoiceOption(optionId);
				wasEnabled = formField.isEnabled()
						&& ((!(formField.getParent() instanceof LyForm))
							|| ((LyForm)formField.getParent()).isEditable());

				if (link.getId().equals("LY_NO_CHOICE"))
					continue;

				index = choiceField.getChoiceOptionIndex(optionId);

				checked = ((newValue != null)
							&& ((multiple && ((BitSet)newValue).get(index))
								|| (!multiple && (((Short)newValue).shortValue() == index))));
				if (checked)
                    nChecked++;
				wasChecked = link.isSelected();

				if ((checked != wasChecked) || !wasEnabled)
					relation.selectLink(j, checked);
			}

			if (!emptyLink
                    && !optional
                    && (nChecked == 0))
			    relation.selectLink(0, true);
		}
		else if (fieldItem instanceof LyBoolean)
		{
			if (n != 2) // XXX Throw illegalStateException
				return;

			LyBoolean booleanField = (LyBoolean)fieldItem;

			checked = (newValue != null) && (((Boolean)newValue).booleanValue());

			// We have two options

			if (newValue == null)
				booleanField.setStatus(LyBoolean.Status.EMPTY);
			else if (checked)
				booleanField.setStatus(LyBoolean.Status.CHECKED);
			else
				booleanField.setStatus(LyBoolean.Status.UNCHECKED);
		}
		else if (fieldItem instanceof LyDictionary)
		{
			LyDictionary	  dico = (LyDictionary)fieldItem;
			LyEntry			  entry;
			LyChoiceFieldInfo contextField = choiceField.getContextField();
			LyChoiceItem	  choiceItem;
			short			  ni = choiceField.getChoiceItemCount();

			if (contextField != null)
			{
				entry = ((LyFieldSetBuilder)getBuilder()).buildChoiceOptionTree(choiceField,
						domain);
				dico.setEntry(entry);
			}

			for (i = 0; i < ni; i++)
			{
				choiceItem = choiceField.getChoiceItem(i);
				entry = dico.getEntry(choiceItem.getId());

				if (entry == null)
					continue;

				hasRule = hasRule
						|| ((choiceItem instanceof LyChoiceOption)
							&& (((LyChoiceOption)choiceItem).getRuleCount() > 0));
				wasChecked = entry.isChecked();

				if (!multiple)
					checked = (newValue != null) && (((Short)newValue).shortValue() == i);
				else
					checked = (newValue != null) && ((BitSet)newValue).get(i);

				if (checked != wasChecked)
					// Check option
					entry.setChecked(checked);
			}
		}
	} // end method updateValue

	/**
	 * Impact of a value change of a fieldInfo. The method invokes LyClassBehavior.propagate which
	 * may be used to implement specific impact on a value changed. This method is for internal use,
	 * application should use setFieldValue instead.
	 * @param value     new Value of the field.
	 * @see   leon.app.behavior.LyClassBehavior#propagate
	 */
	public void propagate(Object value)
	{
		if (!isEnabled())
			return;

		// Execute rules on choice
		LyChoiceFieldInfo choice = (LyChoiceFieldInfo)_fieldInfo;
		LyApplication     application = getApplication();
		LyObject		  object = _setController.getObject();
		int				  priority = application.getPriorityLevel(object);
		LyRule			  rule, registeredRule;
		boolean			  checked;
		short			  i, nr;
		short			  j, n = choice.getChoiceOptionCount();
		LyChoiceOption    option;
		short			  p;
		LyEnvironment     environment = getEnvironment();
		LyInfoDataList<LyRule, Boolean>    rules = new LyInfoDataList<LyRule, Boolean>();
        int               k;

		// Two steps : first apply rules on options that are unchecked
		// and then on checked options
		for (p = 0; p < 2; p++)
		{
			for (i = 0; i < n; i++)
			{
				option = choice.getChoiceOption(i);
				nr = option.getRuleCount();

				if (nr == 0)
					continue;

				if (value instanceof BitSet)
					checked = ((BitSet)value).get(i);
				else if (value instanceof Short)
					checked = (((Short)value).shortValue() == i);
				else if (value instanceof Boolean)
					checked = (((Boolean)value).booleanValue() && (i == 1))
							|| (!((Boolean)value).booleanValue() && (i == 0));
				else
					checked = false;

				if (((p == 0) && checked)
						|| ((p == 1) && !checked))
					continue;

                if (p == 0)
                    environment.logTrace(">>> Apply reverse rule for " + _fieldInfo);
                else
                    environment.logTrace(">>> Apply rule for " + _fieldInfo);

				for (j = 0; j < nr; j++)
				{
					rule = option.getRule(j);

					if (rule.getPriority() > priority)
						continue;

					if (!(rule.getObjectInfo() instanceof LyFieldInfo))
						continue;

					// Check if rules on the same field and with the same operation already exist.
                    // If so, remove them from rules table because this new rule will cancel them.
                    for (k = rules.getSize() - 1; k >= 0; k--)
					{
						registeredRule = rules.getInfo(k);
						if ((rule.getObjectInfo() == registeredRule.getObjectInfo())
								&& (rule.getOperation() == registeredRule.getOperation()))
							rules.removeInfo(registeredRule);
					}

					rules.addInfo(rule, Boolean.valueOf(checked));
				}
			}
		}

		// Execute rules stored in the rules table
        boolean           hasVisibleRule = false;

		for (i = 0, nr = (short)rules.getSize(); i < nr; i++)
		{
			rule = rules.getInfo(i);
            checked = rules.getData(i).booleanValue();
            if (rule.getOperation() == LyRule.Operation.SET_VISIBLE)
                hasVisibleRule = true;

			_setController.execRule(rule, checked);
		}

		// If at least one rule shows or hide a field, reset form to compute
		// its new content and size.
		if (hasVisibleRule)
		{
			LyForm form = ((_formField.getParent() instanceof LyForm)
					? (LyForm)_formField.getParent() : null);

			if (form != null)
			{
				LyViewComponent internal = form.getInternalView();

				if (internal instanceof LyResetContentInterface)
					((LyResetContentInterface)internal).resetContent();
			}
		}

		super.propagate(value);
	} // end method propagate

	/**
	 * Changes current field status to (non-)optional. An optional field may accpt an empty (null)
	 * value and its label is not shown as bold.
	 * @param optional new fieldInfo status. (default is from the view description)
	 */
	public void setOptional(boolean optional)
	{
		if ((_formField != null)
				&& (_formField.isOptional() != optional))
		{
			if (_formField.getState() == LyFormField.STATE_ERROR)
				_formField.setState(LyFormField.STATE_NORMAL);

			_formField.setOptional(optional);

			// If it is a choice, then disable or enable the "no choice" option.
			if ((_fieldInfo instanceof LyChoiceFieldInfo)
					&& (_formField.getFieldItem() instanceof LyChoice)
					&& (((LyChoiceFieldInfo)_fieldInfo).getType()
						!= LyChoiceFieldInfo.Type.BOOLEAN))
			{
				LyChoice choice = (LyChoice)_formField.getFieldItem();
				LyOption option = null;
				int		 index = -1, i, n = choice.getOptionCount();

				for (i = 0; i < n; i++)
				{
					option = choice.getOption(i);
					if (option.getId().equals("LY_NO_CHOICE"))
					{
						index = i;
						break;
					}
				}

				// if the "no choice" option is found
				if (index > 0)
				{
					// if it was previously checked and it becomes mandatory then uncheck.
					if (!optional && option.isChecked())
					{
						short newChoice = (short)((index == 0) ? 1 : 0);

						choice.checkOption(index, false); // deselect no choice value
						setFieldValue(new LyChoiceValue((LyChoiceFieldInfo)_fieldInfo, newChoice),
							true);
					}

					// enable or disable the option
					choice.enableOption(index, optional);
				}
			}

			// If field is not optional and the value is null : set the default value
			Object oldValue = getValue();

			if (!optional && (oldValue == null))
			{
				Object defaultValue = getDefaultValue();

				if (defaultValue != null)
					setFieldValue(defaultValue, true);
			}
		}
	} // end method setOptional
} // end class LyChoiceFieldSetController
