/**
 * LyCompositionFieldSetController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyCompositionFieldSetController.java 34302 2012-02-29 10:59:38Z jlsaouli $
 */

package leon.control.form;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.control.builder.*;
import leon.data.*;
import leon.info.*;
import leon.info.save.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * Field controller for a modification of a simple composition relation.
 * @see LySetController
 */
public class LyCompositionFieldSetController extends LyFieldSetController
	implements LyFormResultInterface
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyCompositionFieldSetController.java 34302 2012-02-29 10:59:38Z jlsaouli $";

	/** Internal area controller */
	protected LyAreaController _areaController;
	
	/** indicate if a type change occurs */
	protected boolean		   _updateTypeChange;
	/** original object */
	protected LyObject		   _origin;
	
	/** first ? */
	protected boolean		   _first;

	/**
	 * Constructor.
	 * @param parent parent controller
	 * @param fieldInfo relation field handled by this controller
	 */
	public LyCompositionFieldSetController(LySetController parent,
		LyRelationFieldInfo fieldInfo)
	{
		super(parent, fieldInfo);

		_areaController = null;
		_updateTypeChange = false;
		_origin = null;
		_first = true;
	}

	/**
	 * Free this controller.
	 */
	public void free()
	{
		if (_areaController != null)
		{
			_areaController.free(this);
			_areaController = null;
		}

		super.free();
	}

	/**
	 * Clears all cached data for this controller (especially the view and the component).
	 */
	public void clearViewCache()
	{
		if (_areaController != null)
			_areaController.clearViewCache();
		super.clearViewCache();
	}

	/**
	 * Invoked when an action occurs. At this level, the implementation manages<BR>
	 * DEMAND_VALIDATE demands. (demand.getType = LyActionDemand.TYPE_ACTION AND demand.getSource()
	 * = LyController.DEMAND_VALIDATE)<BR>
	 * DEMAND_APPLY demands. (demand.getType = LyActionDemand.TYPE_ACTION AND demand.getSource() =
	 * LyController.DEMAND_APPLY)<BR>
	 * Default action (double click)<BR>
	 * Default action (double click)<BR>
	 * VALUE_CHANGED demands. (demand.getType = LyActionDemand.TYPE_VALUE_CHANGED
	 * @param demand the action demand
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_TYPE_CHANGED)
		{
			typeChanged();
			return;
		}

		super.actionDemand(demand);
	}

	/**
	 * Internal method to invoke on a relation a transverse action for a given classInfo.
	 * @param action    Transverse action.
	 * @param classInfo Class info on which the action is applied.
	 */
	public void invokeTransverseAction(LyAction action, LyClassInfo classInfo)
	{
		LyObjectList list = _areaController.getContent();
		LyAction     root = action.getRoot();
		String		 rootActionId = root.getId();

		if ((rootActionId == LyAction.ID_CREATE)
				&& (list != null)
				&& (list.getSize() != 0))
		{
			//      Switch action on set
			LyObject    object = list.getObject(0);
			LyClassInfo objectClassInfo = object.getClassInfo();
			LyAction    objectAction = objectClassInfo.getActionWithId(LyAction.ID_SET);

			if (objectAction != null)
				action = objectAction;
		}

		if (list != null)
			// Duplicate the list (maybe it is a dynamic list).
			list = new LySimpleObjectList(list);
		LyRelationFieldInfo relation = (LyRelationFieldInfo)_fieldInfo;
		LyAreaController    controller = (LyAreaController)invokeAction(action, classInfo, list,
				false);

		if (controller == null)
			return;

		if (controller instanceof LyTransverseInterface)
			((LyTransverseInterface)controller).setRelationParent(this, relation);

		if (relation.ignoresContext())
		{
			if (controller.getContext() != null)
				controller.setContext(null);
		}

		controller.showView();
	} // end method invokeTransverseAction

	/**
	 * Extract current value for the specified fieldInfo in the form.
	 * @return current value, or null if not found or field is disabled.
	 */
	public LyValue getFieldValue()
	{
		LyUnit unit = getUnit();

		if ((_areaController instanceof LyFormInterface)
				&& ((unit == null) || !unit.hasEditionFormat()))
		{
			LyObjectList relations = _areaController.getContent();

			// Structis/FFT00056
			LyFormField  formField = getFormField();

			if ((formField == null) || !formField.isEnabled())
				return null;

			return new LyRelationValue((LyRelationFieldInfo)_fieldInfo, relations);
		}
		else
			return super.getFieldValue();
	}

	/**
	 * Extract current value for the specified fieldInfo in the form.
	 * @param  unit Unit of the value.
	 * @return current value, or null if not found or field is disabled.
	 */
	public LyValue getFieldValue(LyUnit unit)
	{
		if ((_areaController instanceof LyFormInterface)
				&& ((unit == null) || !unit.hasEditionFormat()))
			return getFieldValue();
		else
			return super.getFieldValue(unit);
	}

	/**
	 * Internal use : update the value of a fieldInfo in the form.
	 * @param newValue the new value of the field. This value may be null, a LyValue instance or the
	 *                 internal value associated to the fieldInfo. See LyValue.getValue()
	 * @param oldValue the latest value of the field. This value may be null, a LyValue instance or
	 *                 the internal value associated to the fieldInfo. See LyValue.getValue()
	 */
	protected void updateValue(Object newValue, Object oldValue)
	{
		LyAreaController	   areaController = getAreaController();
        
        if (areaController == null)
            return;
        
		LyRelationFieldInfo    relationField = (LyRelationFieldInfo)_fieldInfo;
		LyObjectList		   context;
		LyObjectList		   previousContext = areaController.getContext();
		LySetBehaviorInterface behavior = _setController.getSetBehavior();
		LySetController        setController = getSetController();
		
		if ((_first) && (!relationField.isMultiple()))
		{
			LyObject   parentObject = getObject();
			LyObjectId objectId = null;

			if (parentObject != null)
			{
				objectId = (LyObjectId)parentObject.getValue(_fieldInfo);
				if (objectId != null)
					_origin = objectId.getLogicalObject();
			}
			_first = false;
		}

		context = behavior.getRelationContext(_setController, relationField);
		if ((context == null) && (getObject() != null))
			context = new LySimpleObjectList(getObject());

		if (((context == null) && (previousContext != null))
				|| ((context != null) && (previousContext == null))
				|| ((context != null) && (previousContext != null)
					&& !context.equals(previousContext)))
			areaController.setContext(context);
		else if ((setController.isApplied()
		    || !relationField.equals(newValue, oldValue)))
		{
			if (!relationField.isMultiple())
				_areaController = changeAreaFromTypeChange(newValue);

			areaController.reset();								
		}
        
	} // end method updateValue

	/**
	 * Method called at the begining of the parent form validation to check if everything is ok
	 * before completing the edition.
	 * @return One of the status codes indicating if the validation was done (STATUS_OK), if an
	 *         error occured (STATUS_KO), or if nothing was done by the user (STATUS_IGNORE).
	 */
	public short startValidate()
	{
		if (!isEnabled())
			return LyFormInterface.STATUS_IGNORE;

		if (_areaController instanceof LyFormInterface)
		{
			// If the field is optional, only create on need (if something changed)
			if (_areaController instanceof LyCreateController)
			{
				// If field is hidden, dont create if no value
				// Field may be optional by rule.
				boolean optional = isOptional() || isHidden();

				((LyCreateController)_areaController).setCreateIfModified(optional);
			}
			return ((LyFormInterface)_areaController).startValidate();
		}
		else
			return super.startValidate();
	}

	/**
	 * Method called at the end of the parent form validation to complete the edition.
	 * @return One of the status codes indicating if the validation was complete (STATUS_OK), if an
	 *         error occured (STATUS_KO), or if nothing was done by the user (STATUS_IGNORE).
	 */
	public short endValidate()
	{
		if (_areaController instanceof LyFormInterface)
			return ((LyFormInterface)_areaController).endValidate();
		else
			return super.endValidate();
	}

	/**
	 * Internal method for propagating a type change. This methods is called when the type field has
	 * been changed by the user.
	 * <P>The LySetController additionaly impact the view using the setState method.</P>
	 */
	protected void typeChanged()
	{
		LyRelationValue value = (LyRelationValue)getFieldValue();
		LyApplication   application = getApplication();
		LyRelation	    typeRelation = _formField.getRelation();
		LyClassInfo     newClassInfo, oldClassInfo;

		if ((value == null))
			// An error occured when getting the value.
			return;

		LyAreaController oldAreaController = _areaController;

		newClassInfo = application.getClassInfo(typeRelation.getValue());
		oldClassInfo = application.getClassInfo(oldAreaController.getClassInfo().getId());

		if (newClassInfo == oldClassInfo)
			return;

		_fieldType = newClassInfo;

		// Getting the new area
		LyArea			 oldArea = (LyArea)oldAreaController.getComponent();
		LyArea			 parentArea = (LyArea)oldArea.getParent();
		LyAreaController newAreaController = getAreaController();
		LyArea			 newArea = (LyArea)newAreaController.getComponent();

		newArea.setScrolling(LyArea.Scrolling.NO);

		LyComponent areaChild = newArea.getChild(0);

		setAreaNoScroll(areaChild);
		newArea.setLabel("");
		newAreaController.initialize();

		parentArea.removePanel(oldArea);
		parentArea.addPanel(newArea);
		typeRelation.setFocus();
	} // end method typeChanged

	/**
	 * Internal use. When an object is created in a popup creation form for a relation, the child
	 * controller invokes this method to set the newly created object as the current value of the
	 * relation (if a new object is created we automatically select this value). This method is
	 * never used by the application : use setFieldValue instead
	 * @param object        newly created object. The new value of the relation fieldInfo.
	 * @param relation the relation
	 */
	public void notifyFormResult(LyObject object, LyRelationFieldInfo relation)
	{
		if (_fieldInfo instanceof LyStructFieldInfo)
		{
			LyFieldItem fieldItem = _formField.getFieldItem();

			if (fieldItem instanceof LyText)
			{
				LyRelationValue value = new LyRelationValue((LyStructFieldInfo)_fieldInfo, object);

				updateFieldItem(value.getValue(), null, true);
			}
		}
		_setController.setActiveField(_fieldInfo);
	}

	/**
	 * Method used to propagate the type change from other controller.
	 * @param  newValue       The new value of the field.
	 * @return The areaController expected.
	 */
	protected LyAreaController changeAreaFromTypeChange(Object newValue)
	{
		// if the type has changed
		LyRelationFieldInfo relationField = (LyRelationFieldInfo)_fieldInfo;

		if ((relationField.fromFieldTag()
					&& (newValue != null)
					&& ((((LyObjectId)newValue).getClassInfo()) != _areaController.getClassInfo()))
				|| ((_areaController instanceof LyCreateController)
					&& (newValue != null)
					&& (((LyObjectId)newValue).getObject()
						!= ((LyCreateController)_areaController).getObject()))
				|| ((_areaController instanceof LySetController)
					&& !((LySetController)_areaController).isCreation()
					&& (newValue == null)))
		{
			LyArea oldArea = (LyArea)_areaController.getComponent();
			LyArea parentArea = oldArea;

			if ((getId() == null) || !getId().equals(parentArea.getId()))
			{
				if (!(oldArea.getParent() instanceof LyArea))
					return _areaController;

				parentArea = (LyArea)oldArea.getParent();
			}
			else
				return _areaController;
			if (newValue != null)
				_fieldType = ((LyObjectId)newValue).getClassInfo();
			else
				_fieldType = _areaController.getClassInfo();

			_updateTypeChange = true;
			LyAreaController areaController = getAreaController(true);

			_updateTypeChange = false;

			LyArea	    newArea = (LyArea)areaController.getComponent();
			LyComponent areaChild = newArea.getChild(0);

			newArea.setScrolling(LyArea.Scrolling.NO);
			setAreaNoScroll(areaChild);

			newArea.setLabel("");
			areaController.initialize();

			parentArea.removePanel(oldArea);
			parentArea.addPanel(newArea);
			LyAction action = getAction();

			// update of the relation type
			if ((newValue != null)
					&& !action.getRoot().getId().equals(LyAction.ID_CONSULT))
			{
				LyComponent component = parentArea.getParent();

				if ((component != null) && (component instanceof LyFormField))
				{
					LyRelation typeRelation = ((LyFormField)component).getRelation();

					if (typeRelation == null)
						return areaController;

					int    i, nl = typeRelation.getLinkCount();
					String newClassInfoId;

					for (i = 0; i < nl; i++)
					{
						newClassInfoId = ((LyObjectId)newValue).getClassInfo().getId();
						if (typeRelation.getLink(i).getId().equals(newClassInfoId))
						{
							typeRelation.selectLink(i, true);
							break;
						}
					}
				}
			}

			return areaController;
		}
		else
			return _areaController;
	} // end method changeAreaFromTypeChange

	/**
	 * Gets the specific areaController for the given field info.
	 * @return Area controller corresponding to the given field info, null in most cases.
	 */
	public LyAreaController getAreaController()
	{
		return getAreaController(false);
	}

	/**
	 * Gets the specific areaController.
	 * @param change indicate if a change occurs
	 * @return Area controller corresponding to the given field info, null in most cases.
	 */
	protected LyAreaController getAreaController(boolean change)
	{
		boolean typeChanged = false;

		if (_areaController != null)
		{
			if (_areaController.isIdle())
				return _areaController;

			// If the type has changed, removing the old areaController
			if (change
					|| (_areaController.getClassInfo() != _fieldType))
			{
				_areaController.setIdle(true);
				_areaController.free(_formController);
				_formController.removeAreaController(_areaController);
				_areaController = null;
				typeChanged = true;
			}
			else
				return _areaController;
		}

		_areaController = buildAreaController(typeChanged);

		LyBuilder builder = _parent.getBuilder();

		if ((builder instanceof LyFileBuilder)
				&& (_parentFieldController == null))
		{
			// Check if the view was built with a file builder
			LyFormField formField = getFormField();

			if ((formField != null) && !_fieldInfo.fromFieldTag())
			{
				LyComponent component = formField.getFieldItem();

				if ((component instanceof LyArea)
						&& (((LyArea)component).getPanelCount() > 0))
				{
					LyEnvironment environment = getEnvironment();
					LyAction	  action = _areaController.getAction();

					environment.logWarning("BF AppEngine : action " + action
						+ " should have its own specific view if needed");
				}

				// Replace existing area by sub controller area
				LyArea area = (LyArea)_areaController.getComponent();

				formField.setFieldItem(area);
				_areaController.initialize();
			}
		}

		return _areaController;
	} // end method getAreaController

	/**
	 * Build the area controller
	 * @param typeChanged true if the type has changed
	 * @return the area controller
	 */
	protected LyAreaController buildAreaController(boolean typeChanged)
	{
		if (_fieldType == null)
			_fieldType = getFieldType();

		LyClassInfo  classInfo = getClassInfo();
		LyObjectList list = null;
		LyObjectList context = null;
		LyObject     object = getModel();

		if ((object != null)
				&& classInfo.contains(_fieldInfo))
		{
			LyRelationValue value = object.getRelationValue((LyRelationFieldInfo)_fieldInfo);

			if (value != null)
				list = value.getObjectList();

			context = new LySimpleObjectList(object);
		}

		LyAction action = getAreaControllerAction(list, typeChanged);

		if (action == null)
			_areaController = null;
		else
		{
			if ((list == null) && (action.getMin() > 0))
				list = new LySimpleObjectList();
	
			LySession session = getSession();
	
			// We must create the controller with this as parent to manage
			// correctly the internal context (see LySetController.getControllerContext).
			// The parent is set to the form when we call addAreaController.
			_areaController = (LyAreaController)session.invokeAction(this, action, _fieldType, list, false);
	
			if (_areaController != null)
			{
				if (_areaController instanceof LyFormController)
					((LyFormController)_areaController).setCompositionRelation((LyRelationFieldInfo)
						_fieldInfo);
				if (_areaController instanceof LySetController)
					((LySetController)_areaController).setContextInternal();
				_areaController.setViewController(_formController.getViewController());
				if (context != null)
					_areaController.setContext(context);
	
				_formController.addAreaController(_areaController);
				
                if (_areaController instanceof LySetController)
                {
                    LyContext.Position position = getInsertPosition();
                    
                    if (position != null)
                    {
                        LyContext   controllerContext = ((LySetController)_areaController).getControllerContext();
                        
                        controllerContext.setKeepOrder(true);
                        controllerContext.setInsertPosition(position);
                    }
                }
			}
		}
		return _areaController;
	} // end method buildAreaController

	/**
	 * Internal use only.
	 * @param list 
	 * @param typeChanged 
	 * @return an action
	 */
	protected LyAction getAreaControllerAction(LyObjectList list, boolean typeChanged)
	{
		LySetBehaviorInterface behavior = (LySetBehaviorInterface)getBehavior();
		LyAction			   action = behavior.getAction(_setController, (LyRelationFieldInfo)_fieldInfo, _fieldType);

		if (action == null)
		{
			LyAction         formAction = getAction();
			String           actionId = formAction.getRoot().getId();
			LyApplication    application = getApplication();

	        if (!application.isEdited())
	        {
    			if ((list == null)
    			        && (actionId != LyAction.ID_CONSULT))
    				actionId = LyAction.ID_CREATE;
    
    			// type has changed --> CREATION
    			if ((typeChanged && !_updateTypeChange)
    					|| (list == null)
    					|| (list.getSize() == 0))
    			{
    				actionId = LyAction.ID_CREATE;
    				list = null;
    			}
    			
                if (_fieldType != null)
                    action = _fieldType.getActionWithId(actionId);
	        }
	        else if (_fieldType != null)
			{   
	        	LyAction	parentAction = _formController.getAction();
	        		
                if (_fieldInfo.hasMark("__" + parentAction.getTemplate().getId() + ".create"))
                	actionId = LyAction.ID_CREATE;
                else if (_fieldInfo.hasMark("__" + parentAction.getTemplate().getId() + ".set"))
                	actionId = LyAction.ID_SET;
                
			    LyAction     realAction = _fieldType.getActionWithId(actionId);
			    
                if (realAction != null)
                    actionId = "__eu.w4.ac" + realAction.getId() + "." + _fieldType.getId();
                action = application.getAction(actionId);
                if (action == null)
                    action = new LyAction(actionId, realAction, _fieldType, null);
			}
		}

		return action;
	} // end method getAreaControllerAction

	/**
	 * Gets the classInfo of the object pointed by the relation in the current form.
	 * @return The classInfo of the object pointed by the relation in the current form.
	 */
	public LyClassInfo getFieldType()
	{
		LyClassInfo		    classInfo = getClassInfo();
		LyRelationFieldInfo relation = (LyRelationFieldInfo)_fieldInfo;
		LyClassInfo		    fieldType = null;
		LyObject		    object = getObject();

		if ((object != null)
				&& classInfo.contains(relation))
		{
			LyRelationValue relationValue = object.getRelationValue(relation);

			if ((relationValue != null) && (relation != null))
			{
				LyObjectId  objectId = (LyObjectId)relationValue.getValue();
				LyClassInfo newClassInfo;

				if (objectId != null)
				{
					newClassInfo = objectId.getClassInfo();
					fieldType = newClassInfo;
				}
			}
		}

		if (fieldType == null)
		{
			LyInfoList<LyClassInfo> fieldTypes = relation.getClassInfos();
			int		   index = 0, nClassInfos = fieldTypes.getSize();

			fieldType = fieldTypes.getInfo(index++);
			while ((fieldType != null)
                    && fieldType.isAbstract()
                    && (index < nClassInfos))
				fieldType = fieldTypes.getInfo(index++);
		}

		return fieldType;
	} // end method getFieldType

	/**
	 * Changes current field status to (non-)editable. If the field is not editable, the value is
	 * read-only but however submited with all other values. A disabled field is not submited with
	 * all other values and is just ignored.
	 * @param editable new fieldInfo status. (default is true or obtained from the view
	 *                  description)
	 */
	public void setEditable(boolean editable)
	{
		if ((_formField != null)
				&& (_formField.isEditable() != editable))
		{
			super.setEditable(editable);
			if ((_formField.isEditable() == editable)
					&& (_areaController instanceof LyFormInterface))
				((LyFormInterface)_areaController).setEditable(editable);
		}
	}

	/**
	 * Changes current field status to (non-)enabled. If the field is not editable, and not submited
	 * with all other values.
	 * @param enabled  new fieldInfo status. (default is true or obtained from the view
	 *                  description)
	 */
	public void setEnabled(boolean enabled)
	{
		if ((_formField != null)
				&& (_formField.isEnabled() != enabled))
		{
			super.setEnabled(enabled);
			if (_areaController instanceof LyFormInterface)
				((LyFormInterface)_areaController).setEnabled(enabled);
		}
	}

	/**
	 * Method called when all modifications are committed.
	 */
	public void commit()
	{
		LyObject object = null;
		if (!isEnabled())
		{
			if (_areaController instanceof LyFormInterface)
			{
				((LyFormInterface)_areaController).cancel();
			}
		}
		else if (_areaController instanceof LyFormInterface)
		{
			((LyFormInterface)_areaController).commit();

			if (_parent instanceof LyFormController)
			{
				LyObject parent = ((LyFormController)_parent).getObject();

				// Update the value in the relation for auto-increment
				LyValue  value = getFieldValue();

				if ((parent != null) && (value != null))
				{

					if (_parent instanceof LyFormInterface)
					{
						LyContext context = ((LyFormInterface)_parent).getControllerContext();

						context.setValue(parent, value);
					}
					else
					{
						parent.setFieldValue(value);
						parent.set(false);
					}
				}
			}

			if (_areaController instanceof LyFormController)
				object = ((LyFormController)_areaController).getObject();
		}

		// Suppress oldObject
		if (isCreation()
            && (object != _origin)
            && (_origin != null))
		{
			LyContext context = ((LyFormInterface)_parent).getControllerContext();

			context.deleteObject(_origin);
		}
	} // end method commit

	/**
	 * Method called when editions is cancelled.
	 */
	public void cancel()
	{
		if (_areaController instanceof LyFormInterface)
			((LyFormInterface)_areaController).cancel();
	}

	/**
	 * Checks if the value for the fieldInfo has been modified.
	 * @return true if the value has been modified, false otherwise.
	 */
	public boolean checkIfModified()
	{
		if (_areaController instanceof LyFormInterface)
		{
			boolean optional = isOptional();

            if (_areaController instanceof LyCreateController)
            {
    			if (_setController instanceof LyCreateController)
    				optional |= ((LyCreateController)_setController).isCreateIfModified();
    			((LyCreateController)_areaController).setCreateIfModified(optional);
            }
            
			return ((LyFormInterface)_areaController).checkIfModified();
		}

		return super.checkIfModified();
	}

	/**
	 * Controls current edition.
	 * @return The error status.
	 */
	public short control()
	{
		if (_areaController instanceof LyFormInterface)
			return ((LyFormInterface)_areaController).control();
		else
			return super.control();
	}

	/**
	 * Sets the field content from the specified value Any value modified by the user is the form is
	 * lost. Default implementation just call updateFieldItem but this may be overriden in
	 * subclasses.
	 * @param value          new value to set in the field.
	 * @param freeOldValue   indicates if the old value should be freed (deleted) when it is an
	 *                       object
	 * @param createNewValue indicates if a new empty value should created when object is null and
	 *                       when the value is an object
	 */
	public void resetFromValue(Object value, boolean freeOldValue, boolean createNewValue)
	{
		LyAreaController areaController = getAreaController();

		if ((areaController instanceof LyFormController) && (value == null))
			((LyFormController)areaController).resetFromObject(null, freeOldValue, createNewValue);
		else if ((areaController instanceof LyFormController) && (value instanceof LyObjectId))
			((LyFormController)areaController).resetFromObject(((LyObjectId)value)
					.getLogicalObject(), freeOldValue, createNewValue);

		super.resetFromValue(value, freeOldValue, createNewValue);
	}
	   
    /**
     * Save field input.
     */
    public LyFieldSave save(boolean full)
    {
        LyFieldSave      fieldSave = super.save(full);
        LyAreaController areaController = getAreaController();
        LyFormSave  formSave = null;
        
        if (areaController instanceof LySetController)
            formSave = ((LySetController)areaController).save(full);
        
        if ((fieldSave == null) && (formSave == null))
            return null;
        else if (fieldSave == null)
            return new LyCompositionFieldSave(_fieldInfo, formSave);
        else
            return new LyCompositionFieldSave(fieldSave, formSave);
    }
    
    /**
     * Restore field input.
     */
    public boolean restore(LyFieldSave fieldSave)
    {
        if (!super.restore(fieldSave))
            return false;
        
        if (fieldSave instanceof LyCompositionFieldSave)
        {        
            LyAreaController areaController = getAreaController();
            LyFormSave  formSave = ((LyCompositionFieldSave)fieldSave)._formSave;
            
            if (areaController instanceof LySetController)
                ((LySetController)areaController).restore(formSave);
        }
        
        return true;
    }
    
	/**
	 * Resets specified fieldInfo value : the default value is fetched from the associated behavior.
	 * LyClassBehavior.getValue (LySetBehaviorInterface) The field is restored to its original value
	 * when the view was opened.
	 */
	public void reset()
	{
		super.reset();
		      
        LyAreaController areaController = getAreaController();
        
        if (areaController instanceof LySetController)
            ((LySetController)areaController).cancelChanges();
	}
} // end class LyCompositionFieldSetController
