/**
 * LyCreateOrSetTableController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyCreateOrSetController.java 34717 2012-03-15 14:53:42Z jlsaouli $
 */

package leon.control.form;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;


/**
 * This class implements a set controller used by the LyEditTableController This class is for
 * internal use only.
 * @see LyEditTableController
 */
public class LyCreateOrSetController extends LyCreateController
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyCreateOrSetController.java 34717 2012-03-15 14:53:42Z jlsaouli $";

	/** Boolean indicating if the controlled form is a creation form or a set form. */
	protected boolean		   _creation;
	

	/**
	 * Constructor. Creation or modification of one object. ClassInfo parameter is ignored, kept for
	 * compatibility and introspection capabilities.
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller).
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param classInfo this parameter is kept for introspection and compatibility. not used.
	 * @param objects   list of objects currently being modified.
	 */
	public LyCreateOrSetController(LySession session, LyController parent, LyAction action,
		LyClassInfo classInfo, LyObjectList objects)
	{
		super(session, parent, action, classInfo, objects);
		init();
	}

	/**
	 * May be overriden for initialization of variables.
	 */
	protected void init()
	{
		setCreation(true);
		_initialized = true;
	}

	/**
	 * Sets current model for the creation.
	 * <P>Reset default values of the form.</P>
	 * @param  object new object being modified.
	 * @return true if the object change was successful, false otherwise.
	 */
	public boolean setObject(LyObject object)
	{
		// Do not change the modified object when validating() a SET demand.
		// On creation, _changingObject is used to restart with a blank object.
		if ((!isCreation()) && (_changingObject) && (object == null))
			return false;
		return super.setObject(object);
	}

	/**
	 * End of the current action. The modifications are completed, and the application may take
	 * control and continue or give up action. If false, the process od closing the window aborts.
	 * <P>The method returns LyClassBehavior.endSet() (LySetBehaviorInterface)</P>
	 * @param  values Values returned by the controlled form.
	 * @return True if the action may end, false otherwise.
	 */
	protected boolean end(LyValueSet values)
	{
		if (!isCreation())
		{
			LySetBehaviorInterface behavior = getSetBehavior();
			boolean				   result;

			_ending = true;
			result = behavior.endSet(this, values);
			_ending = false;
			return result;
		}
		else
			return super.end(values);
	}

	/**
	 * Returns a boolean indicating if the action is or not a creation.
	 * @return true if in creation mode
	 */
	public boolean isCreation()
	{
		return _creation;
	}

	/**
	 * Internal use : sets if the action is or not a creation.
	 * @param creation the creation mode 
	 */
	public void setCreation(boolean creation)
	{
		_creation = creation;
		setCreateIfModified(creation);
	}

	/**
	 * Accessor to the action associated to this controller.
	 * @return The reference to associated action
	 */
	public LyAction getAction()
	{
		return (_creation ? _action : _classInfo.getActionWithId(LyAction.ID_SET));
	}

	/**
	 * Check idle status of the field controller.
	 * @return boolean indicating if the idle mode is on or off.
	 */
	public boolean isIdle()
	{
		if (_parent instanceof LyFieldSetController)
			return super.isIdle() || _parent.isIdle();
		else
			return super.isIdle();
	}
} // end class LyCreateOrSetController
