/**
 * LyDateFieldSetController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyDateFieldSetController.java 41425 2012-10-31 14:50:01Z jlsaouli $
 */

package leon.control.form;

import leon.app.*;
import leon.app.behavior.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * Default field controller for a modification.
 * @see LySetController
 */
public class LyDateFieldSetController extends LyFieldSetController
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyDateFieldSetController.java 41425 2012-10-31 14:50:01Z jlsaouli $";

	/** Name of the tranverse date tool. */
	public static final String DATE_TOOL = "Date_Tool";

	protected String _currentTimeZone = null;
	/**
	 * Constructor.
	 * @param parent parent set controller
	 * @param fieldInfo date field handled by this controller
	 */
	public LyDateFieldSetController(LySetController parent, LyFieldInfo fieldInfo)
	{
		super(parent, fieldInfo);
	}

	/**
	 * Invoked when an action occurs. At this level, the implementation manages<BR>
	 * DEMAND_VALIDATE demands. (demand.getType = LyActionDemand.TYPE_ACTION AND demand.getSource()
	 * = LyController.DEMAND_VALIDATE)<BR>
	 * DEMAND_APPLY demands. (demand.getType = LyActionDemand.TYPE_ACTION AND demand.getSource() =
	 * LyController.DEMAND_APPLY)<BR>
	 * Default action (double click)<BR>
	 * Default action (double click)<BR>
	 * VALUE_CHANGED demands. (demand.getType = LyActionDemand.TYPE_VALUE_CHANGED
	 * @param demand the action demand
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			String   sourceId = demand.getSource().intern();
			String[] parameters = LyString.stringToList(sourceId, ID_SEP);

			if ((parameters.length > 1) && parameters[1].equals(DATE_TOOL))
			{
				chooseDate();
				return;
			}
		}

		super.actionDemand(demand);
	}

	/**
	 * @see leon.control.form.LyFieldSetController#setEditable(boolean)
	 */
	public void setEditable(boolean editable)
	{
		super.setEditable(editable);

		if (_formField != null)
		{
			LyToolBar	toolbar = _formField.getToolBar();

			if (toolbar != null)
			{
				LyTool	tool = (LyTool)toolbar.getComponent("__dateTool" + LyController.ID_SEP +
						LyDateFieldSetController.DATE_TOOL + LyController.ID_SEP + _fieldInfo.getId());

				if (tool != null)
					tool.setEnabled(editable);
			}
		}
	}

	/**
	 * Choose date
	 */
	protected void chooseDate()
	{
		LyDateFieldInfo		  dateField = (LyDateFieldInfo)_fieldInfo;
		LyApplication		  application = getApplication();
		LyFormController      formController = getFormController();
		LyClassInfo           classInfo = formController.getClassInfo();
		LyAction			  action = application.getActionWithId(LyAction.ID_DATE_TOOL, classInfo);
		LyUserInputController dateInput = new LyUserInputController(getSession(), this, action);
		Object				  value = getValue();
		_LyDateToolBehavior	  behavior = new _LyDateToolBehavior();

		dateInput.setBehavior(behavior);
		behavior.init(dateInput);

		String	    label = null;
		LyInfoLabel infoLabel = getInfoLabel(dateField, LyInfoLabel.Type.NAME);

		if (infoLabel != null)
		{
			LyObject object = getObject();

			label = infoLabel.getString(object);
		}
		if (label == null)
			label = dateField.getName();

		String title = "LY_DATE_TOOL" + " : " + label;
		LyDate date;

		dateInput.setTitleBar(title);
		dateInput.setImage("LY_DATE_TOOL");
		date = dateInput.addDate("__dateField", label, null, dateField.getFormat(), false);

		date.setLabel(dateField.toString(value, true));
		date.setType(LyDate.getType(dateField.getType().ordinal()));
		dateInput.showView();
	} // end method chooseDate

	/**
	 * Internal method for propagating a unit change. This methods is called when the field unit has
	 * been changed by the user.
	 * <P>The LySetController additionaly impact the view using the setValid method.</P>
	 * @param newTimeZone The new fieldinfo unit.
	 * @param oldTimeZone The old fieldinfo unit.
	 */
	public void unitChanged(String newTimeZone, String oldTimeZone)
	{
		LyDateFieldInfo dateField = (LyDateFieldInfo)_fieldInfo;

		if (!dateField.isWithTimeZone())
		{
			super.unitChanged(newTimeZone, oldTimeZone);
			return;
		}

		// The units are the same
		if (newTimeZone.equals(oldTimeZone))
			return;

		// Update the view layer
		LyFieldUnit formFieldUnit = _formField.getFieldUnit();
		LyRelation  relation = (LyRelation)formFieldUnit.getFieldItem();

		relation.selectLink(oldTimeZone, false);
		relation.selectLink(newTimeZone, true);

		// the field value
		LyFieldInfoValue oldFieldValue = (LyFieldInfoValue)getFieldValue(oldTimeZone);

		LyFieldInfoValue fieldValue = (LyFieldInfoValue)getFieldValue(newTimeZone);
		if (fieldValue == null)
			// An error occured when getting the value.
			return;

		_currentValue = fieldValue;
		_currentTimeZone = newTimeZone;

		Object oldValue = (oldFieldValue == null ? null : oldFieldValue.getValue());
		Object newValue = (fieldValue == null ? null : fieldValue.getValue());

		updateFieldItem(newValue, oldValue, false);

		// The unit has changed : we must control the value with the new unit.
		valueChanged();
	}
	/**
	 * Internal use : update the value of a fieldInfo in the form.
	 * @param newValue the new value of the field. This value may be null, a LyValue instance or the
	 *                 internal value associated to thge fieldInfo. See LyValue.getValue()
	 * @param oldValue the latest value of the field. This value may be null, a LyValue instance or
	 *                 the internal value associated to thge fieldInfo. See LyValue.getValue()
	 */
	protected void updateValue(Object newValue, Object oldValue)
	{
		LyDateFieldInfo dateField = (LyDateFieldInfo)_fieldInfo;
		if (_setController.isConsult(_fieldInfo)
				|| !dateField.isWithTimeZone())
		{
			super.updateValue(newValue, oldValue);
			return;
		}


		// Default method consists to update label
		LyFieldItem fieldItem = getFieldItem();
		String	    label = dateField.toString(newValue, false, _currentTimeZone);

		if (fieldItem != null)
		{
			fieldItem.setLabel(label);
			fieldItem.setFieldValue(newValue);
		}
	}
	public LyValue getFieldValue(String timeZone)
	{
		LyValue fieldValue;
		Object  objectValue;
		LyDateFieldInfo dateField = (LyDateFieldInfo)_fieldInfo;
		String strValue = getStringValue();

		try
		{
			objectValue = dateField.parse(strValue, timeZone);
			fieldValue = dateField.buildFieldValue(objectValue);
		}
		catch (LyFieldInfoFormatException e)
		{
			if (getEnvironment().isDebugLevel(LyEnvironment.DEBUG_TRACE))
				getEnvironment().writeException(e);

			//FFT00047
			// We display the error message.
			LyError  error;
			LyObject object = _setController.getObject();

			String message = e.getMessage();
			if ((message != null) && message.startsWith("BF AppEngine : "))
			{
				message = message.substring(15);
			}
			error = new LyError(LyObject.LEVEL_ERROR, message, null, _fieldInfo, object, _fieldInfo);
			showError(object, error);

			// Setting the form field to invalid to stop the process.
			setState(LyFormInterface.STATUS_KO);

			return null;
		}

		return fieldValue;
	}
	/**
	 * Extract current value for the specified fieldInfo in the form.
	 * @return current value, or null if not found or field is disabled.
	 */
	public LyValue getFieldValue()
	{
		if (_setController.isConsult(_fieldInfo))
			return super.getFieldValue();

		LyFormField formField = getFormField();

		if ((formField == null) || !formField.isEnabled())
			return null;

		return getFieldValue(_currentTimeZone);
	}
	
	class _LyDateToolBehavior extends LyUserInputBehavior
	{
		/**
		 * Invoked when the user validates the content of the input.
		 *
		 * @param controller the current LyUserInputController instance
		 * @param demand     demand parameters sent by the view component.
		 */
		public void userInputDemand(LyController controller, LyDemand demand)
		{
			// Close dialog
			String id = demand.getSource();

			if (id.equals(LyFormController.DEMAND_VALIDATE))
			{
				LyParameterValues paramValues = demand.getParameterValues();

				if (paramValues != null)
				{
					String  dateValue = paramValues.getParameter("__dateField");
					LyValue value = _fieldInfo.buildFieldValue(dateValue);

					setFieldValue(value, true);
					// SIR/FFT02443
					valueChanged();
				}
			}

			super.userInputDemand(controller, demand);
		}
	} // end class _LyDateToolBehavior
} // end class LyDateFieldSetController
