/**
 * LyEditTableChoiceFieldSetController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyEditTableChoiceFieldSetController.java 41729 2012-12-06 09:47:58Z bangiole $
 */

package leon.control.form;

import java.util.*;

import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.control.builder.*;
import leon.control.form.builder.*;
import leon.info.*;
import leon.view.*;


/**
 * Field controler for a modification of a LyChoiceFieldInfo in an editable table.
 * @see LyEditTableSetController
 */
public class LyEditTableChoiceFieldSetController extends LyChoiceFieldSetController
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyEditTableChoiceFieldSetController.java 41729 2012-12-06 09:47:58Z bangiole $";

	/** Form used for a choice field info. */
	LyForm					   _form;

	/** Table controller. */
	LyEditTableController	   _editTableController;

	/** List of internal form fields stored by their ids. */
	Hashtable<String, LyFormField>				   _formFields;

	/**
	 * Constructor.
	 * @param parent parent controller
	 * @param fieldInfo choice field handled by this controller
	 */
	public LyEditTableChoiceFieldSetController(LyEditTableSetController parent,
		LyChoiceFieldInfo fieldInfo)
	{
		super(parent, fieldInfo);

		_editTableController = (LyEditTableController)parent.getParent();
	}

	/**
	 * @see leon.control.form.LyFieldSetController#getDefaultBuilder()
	 */
	public LyBuilder getDefaultBuilder()
	{
		return new LyEditTableMultipleChoiceSetBuilder(this);
	}

	/**
	 * Utility method to build the field id.
	 * @return the formField corresponding to the controlled fieldInfo
	 */
	String getFieldId()
	{
		String id = getId();
		String sep = LyFormController.FORM_FIELD_SEP;

		id += sep + _fieldInfo.getId();
		return id;
	}

	/**
	 * Accessor to the formField (View component) associated to the fieldInfo. The formField
	 * contains a caption (label of the fieldInfo), a fieldItem (value of the fieldInfo) and the
	 * constraints definition (how the component are shown in the view).
	 * @param  fieldInfo the fieldInfo which is associated to the view component.
	 * @return the formField corresponding to the given fieldInfo
	 */
	public LyFormField getFormField(LyFieldInfo fieldInfo)
	{
		LyChoiceFieldInfo choiceFieldInfo = (LyChoiceFieldInfo)_fieldInfo;
		// Specific case for CHOICE multiple : rebuild form field
		if (choiceFieldInfo.isMultiple()
				&& !choiceFieldInfo.isDictionary()
				&& (_editTableController.canBeEditable(_fieldInfo)
					|| (_editTableController.isEditable()
							&& !_editTableController.isEnabled())))
		{
			LyEditTableMultipleChoiceSetBuilder builder;

			builder = (LyEditTableMultipleChoiceSetBuilder)getBuilder();
			if (_form == null)
				_form = (LyForm)builder.buildComponent();

			// Build the aggregate formField with the accurate id
			LyFormField optionField;
			LyChoice    optionChoice, newChoice = new _LyPointerChoice();
			LyOption    option;
			String	    id;
			short	    i, n = choiceFieldInfo.getChoiceOptionCount();

			for (i = 0; i < n; i++)
			{
				// Get the table field id
				id = builder.getFieldId(this, choiceFieldInfo, i);
				optionField = getOptionField(id);

				if (optionField == null)
					// The marks was not set so the component is *NOT added*
					// create a dummy one (not shown)
					option = new LyOption(choiceFieldInfo.getChoiceOption(i).getId(), "");
				else
				{
					optionChoice = (LyChoice)optionField.getFieldItem();
					option = optionChoice.getOption(0);
				}

				// NewChoice is a specific choice which is not modifying the parent
				// of the option. Just a simple container.
				newChoice.addOption(option);
			}

			String	    fieldId = getFieldId();
			LyFormField newField = new LyFormField(fieldId);

			newField.setFieldItem(newChoice);
			return newField;
		}
		else
			return super.getFormField(fieldInfo);
	} // end method getFormField

	/**
	 * Internal use : update the value of a choice fieldInfo in the form. <B>Application should use
	 * setFieldValue instead.</B> This method is called from updateFieldItem. (Specialization)
	 * @param newValue     the new value of the field. This value may be null, a LyValue instance or
	 *                  the internal value associated to the fieldInfo. See LyValue.getValue()
	 * @param oldValue  the latest value of the field. This value may be null, a LyValue instance or
	 *                  the internal value associated to the fieldInfo. See LyValue.getValue()
	 * @see   #setFieldValue
	 * @see   #updateFieldItem
	 */
	protected void updateValue(Object newValue, Object oldValue)
	{
		LyChoiceFieldInfo choiceFieldInfo = (LyChoiceFieldInfo)_fieldInfo;
		// Specific case for CHOICE multiple : rebuild form field
		if (choiceFieldInfo.isMultiple()
				&& !choiceFieldInfo.isDictionary()
				&& _editTableController.canBeEditable(_fieldInfo))
		{
			LyEditTableMultipleChoiceSetBuilder builder;

			builder = (LyEditTableMultipleChoiceSetBuilder)getBuilder();
			if (_form == null)
				_form = (LyForm)builder.buildComponent();

			// Build the aggregate formField with the accurate id
			LyFormField			   optionField;
			LyChoice			   optionChoice;
			LyOption			   optionOption;
			LyChoiceOption		   option;
			String				   id;
			short				   i, n = choiceFieldInfo.getChoiceOptionCount();
			boolean				   checked, wasChecked;
			boolean				   enabled, wasEnabled;
			LySetBehaviorInterface behavior = _setController.getSetBehavior();
			Vector<LyChoiceOption> domain = behavior.getChoiceDomain(_setController,
					choiceFieldInfo);

			LyEditTableSetController	editTableSetController = (LyEditTableSetController)_setController;
			boolean						wasUpdating = editTableSetController.isUpdating();

			editTableSetController.setUpdating(true);

			for (i = 0; i < n; i++)
			{
				// Get the table field id
				id = builder.getFieldId(this, choiceFieldInfo, i);
				optionField = getOptionField(id);
				option = choiceFieldInfo.getChoiceOption(i);

				if (optionField != null)
				{
					optionChoice = (LyChoice)optionField.getFieldItem();
					optionOption = optionChoice.getOption(0);
					wasEnabled = optionOption.isEnabled();
					wasChecked = optionOption.isChecked();
					checked = (newValue != null) && ((BitSet)newValue).get(i);
					enabled = ((domain != null) && (domain.contains(option)));

					optionChoice.enableOption(optionOption, enabled);

					if ((checked != wasChecked) || !wasEnabled)
						// Check option
						optionChoice.checkOption(optionOption, checked);

					if (optionOption.isChecked()
							&& !optionOption.isEnabled()
							&& (choiceFieldInfo.getContextField() != null))
						optionChoice.checkOption(optionOption, false);
				}
			}

			editTableSetController.setUpdating(wasUpdating);
		}
		else
			super.updateValue(newValue, oldValue);
	} // end method updateValue

	/**
	 * Retrieve a form field from its identifier.
	 *
	 * @param id the identifier of the form field.
	 * @return the form field if found, null otherwise.
	 */
	protected LyFormField getOptionField(String id)
	{
		if ((_formFields != null) && (_formFields.get(id) != null))
			return _formFields.get(id);

		LyComponent component = _editTableController.getComponent();
		LyFormField optionField = null;

		if (component != null)
			optionField = (LyFormField)component.getComponent(id);
		if ((optionField == null) && (_form != null))
			optionField = (LyFormField)_form.getComponent(id);

		if (optionField != null)
		{
			if (_formFields == null)
				_formFields = new Hashtable<String, LyFormField>();

			_formFields.put(id, optionField);
		}

		return optionField;
	}

	/**
	 * Accessor to the optional status for a field in the form.
	 * @return current field status
	 */
	public boolean isOptional()
	{
		LyChoiceFieldInfo choiceFieldInfo = (LyChoiceFieldInfo)_fieldInfo;

		// Specific case for CHOICE multiple : rebuild form field
		if (choiceFieldInfo.isMultiple() && !choiceFieldInfo.isDictionary())
		{
			LySetBehaviorInterface	behavior = _setController.getSetBehavior();

			return behavior.isOptional(_setController, _fieldInfo);
		}

		return super.isOptional();
	}

	/**
	 * Changes current field status to (non-)editable.
	 * If the field is not editable, the value is read-only but
	 * however submited with all other values. A disabled field is
	 * not submited with all other values and is just ignored.
     * @param editable new fieldInfo status. (default is true or
     * obtained from the view description)
   	 */
	public void setEditable(boolean editable)
	{
		LyChoiceFieldInfo choiceFieldInfo = (LyChoiceFieldInfo)_fieldInfo;

		if (choiceFieldInfo.isMultiple() && !choiceFieldInfo.isDictionary())
		{
			LyEditTableMultipleChoiceSetBuilder	builder = (LyEditTableMultipleChoiceSetBuilder)getBuilder();
			LyFormField		 					optionField;
			String  							id;
			short								i, n = choiceFieldInfo.getChoiceOptionCount();

			for (i = 0; i < n; i++)
			{
				// Get the table field id
				id = builder.getFieldId(this, choiceFieldInfo, i);
				optionField = getOptionField(id);

				if (optionField != null)
					optionField.setEditable(editable);
			}
		}

		super.setEditable(editable);
	}

	/**
	 * Changes current field status to (non-)enabled.
	 * If the field is not editable, and not submited with all
	 * other values.
     * @param enabled new fieldInfo status. (default is true or
     * obtained from the view description)
   	 */
	public void setEnabled(boolean enabled)
	{
		LyChoiceFieldInfo choiceFieldInfo = (LyChoiceFieldInfo)_fieldInfo;

		if (choiceFieldInfo.isMultiple() && !choiceFieldInfo.isDictionary())
		{
			LyEditTableMultipleChoiceSetBuilder	builder = (LyEditTableMultipleChoiceSetBuilder)getBuilder();
			LyFormField		 					optionField;
			String  							id;
			short								i, n = choiceFieldInfo.getChoiceOptionCount();

			for (i = 0; i < n; i++)
			{
				// Get the table field id
				id = builder.getFieldId(this, choiceFieldInfo, i);
				optionField = getOptionField(id);

				if (optionField != null)
				{
					if (!enabled)
					{
						clear();
						optionField.setEnabled(false);
					}
					else
					{
						optionField.setEnabled(true);
						reset();
					}
				}
			}
		}

		super.setEnabled(enabled);
	}

	/**
     * Accessor to the enabled/disabled status for a field in the form.
     * @return current field status
     */
    public boolean isEnabled()
    {
        LyChoiceFieldInfo choiceFieldInfo = (LyChoiceFieldInfo)_fieldInfo;

        if (choiceFieldInfo.isMultiple()
            && !choiceFieldInfo.isDictionary())
        {
            LyEditTableMultipleChoiceSetBuilder builder = (LyEditTableMultipleChoiceSetBuilder)getBuilder();
            LyFormField                         optionField;
            String                              id;
            short                               i, n = choiceFieldInfo.getChoiceOptionCount();

            for (i = 0; (i < n); i++)
            {
                // Get the table field id
                id = builder.getFieldId(this, choiceFieldInfo, i);
                optionField = getOptionField(id);

                if ((optionField != null)
                        && optionField.isEnabled())
                    return true;
            }
        }

        return super.isEnabled();
    }

	/**
	 * Internal class used for a LyChoice pointer.
	 * This Choice may contains option but does not set the parent field of these options.
	 * Therefore, the options may belong to their real LyChoice container and virtually to this
	 * temporary LyChoicePointer
	 */
	class _LyPointerChoice extends LyChoice
	{
		/**
		 * Inserts a sub-component option at given position.
		 *
		 * @param option The option to add to options list.
		 * @param pos    Position where the sub-component is inserted.
		 */
		public void insertOption(LyOption option, int pos)
		{
			if (getInternalOptions() == null)
			{
				_options = new ArrayList<LyOption>(1);
				_children = _options;
			}
			else if (getInternalOptions().contains(option))
				return;

			if (pos < 0)
				getInternalOptions().add(option);
			else
				getInternalOptions().add(pos, option);

			// do no change parent
		}
	}
} // end class LyEditTableChoiceFieldSetController
