/**
 * LyConsultController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyFieldConsultController.java 34170 2012-02-15 14:17:29Z jlsaouli $
 */

package leon.control.form;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.control.builder.*;
import leon.control.form.builder.*;
import leon.data.*;
import leon.info.*;
import leon.view.*;


/**
 * Default field controller for a consultation.
 * @see LyConsultController
 */
public class LyFieldConsultController extends LyFieldController
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyFieldConsultController.java 34170 2012-02-15 14:17:29Z jlsaouli $";

	/** Consult value if changed when this controller is used in a set controller. */
	protected LyValue		   _consultValue;

	/**
	 * Constructor. The controller automatically registers on the specified session.
	 * @param parent parent controller
	 * @param fieldInfo the field handled by this controller
	 */
	public LyFieldConsultController(LyFormController parent, LyFieldInfo fieldInfo)
	{
		super(parent, fieldInfo);
	}

	/**
	 * @see leon.control.form.LyFieldController#getDefaultBuilder()
	 */
	public LyBuilder getDefaultBuilder()
	{
		return new LyFieldConsultBuilder(this);
	}

	/**
	 * Update the value of a fieldInfo shown in the view. The method is overriden for relation
	 * fieldInfo (updateRelation) and file fieldInfo (updateFile)
	 * @param newValue  the new value of the field.
	 * @param oldValue  former value of the field,or null if unknown. This parameter is not used
	 *                  within generic code.
	 * @param propagate Indicates whether the change is notified (propagated) to other fields or
	 *                  not. Parameter not used at this level
	 */
	public void updateFieldItem(Object newValue, Object oldValue, boolean propagate)
	{
		// Keep value
		_consultValue = _fieldInfo.buildFieldValue(newValue);

		// Update field item
		LyFieldItem fieldItem = getFieldItem();

		fieldItem.setFieldValue(newValue);

		if ((getObject() != null)
				&& (_fieldInfo instanceof LyStructFieldInfo)
				&& ((LyStructFieldInfo)_fieldInfo).fromFieldTag()
				&& !((LyStructFieldInfo)_fieldInfo).isMultiple())
		{
			LyRelationValue value = getObject().getRelationValue((LyStructFieldInfo)_fieldInfo);

			if (value.getValueCount() > 0)
			{
				LyObject object = value.getObject(0);

				if (object != null) {
				}
			}
		}

		updateValue(newValue, oldValue);

        if (propagate
                && !_fieldInfo.equals(newValue, oldValue))
            // Propagate value change if the value really changed
            propagate(newValue);

        setLastModified(System.currentTimeMillis());
    } // end method updateFieldItem

	/**
	 * Utility method (internal use for behavior) to calculate formulas of fieldInfos.
	 * Application should use setFieldValue instead.
	 * @param origin    the field that propagates the modification
	 * @param propagate Indicates whether the change is propagated to other fields.
	 * @param valueChange indicates whether the change occus due to a value changed event
	 * @see   #setFieldValue
	 */
	public void update(LyFieldInfo origin, boolean propagate, boolean valueChange)
	{
		// Reset field and propagate modification
		LyValue fieldValue;

		if (_fieldInfo.isImported()
				&& (_fieldInfo.isLocal()
					|| (_formController.getObject() == null)))
			fieldValue = computeImportValue();
		else
		    fieldValue = getFieldValue();

		// Update field item
		Object value = ((fieldValue == null) ? null : fieldValue.getValue());

		updateFieldItem(value, null, true);
	}
	
    /**
     * Impact of a value change of a fieldInfo. The method invokes LyClassBehavior.propagate which
     * may be used to implement specific impact on a value changed. This method is for internal use,
     * application should use setFieldValue instead.
     * @param value     new Value of the field.
     * @see   leon.app.behavior.LyClassBehavior#propagate
     */
    public void propagate(Object value)
    {
        if (!(_formController instanceof LySetController))
            return;
        
        // Generic processing : invoke behavior
        LySetController        setController = (LySetController)_formController;
        LySetBehaviorInterface behavior = setController.getSetBehavior();
        LyValue                fieldValue = _fieldInfo.buildFieldValue(value);

        behavior.propagate(setController, (LyFieldInfoValue)fieldValue);
    }

	/**
	 * Extract current value for the specified fieldInfo in the form.
	 * @return current value, or null if not found or field is disabled.
	 */
	public LyValue getFieldValue()
	{
		if (_consultValue != null)
			return _consultValue;
		else
		{
			LyObject object = getObject();

			return ((object == null) ? null : object.getFieldValue(_fieldInfo));
		}
	}

	/**
	 * Internal method to consult an object.
	 * @param object The object to consult.
	 */
	protected void consult(LyObject object)
	{
		LyClassInfo objectClassInfo = object.getClassInfo();
		LyAction    consult = objectClassInfo.getDefaultAction();

		if (consult == null)
		{
			// Consult not found, try with default action
			consult = objectClassInfo.getActionWithId(LyAction.ID_CONSULT);
			if (consult == null)
				return;
		}

		// Check if action is enabled, it won't be done in LySession
		// because this controller is a MultiAreaController
		LySession session = getSession();

		if (session.enableAction(consult, this, new LySimpleObjectList(object), true))
		{
			LyController	controller = invokeAction(consult, object.getClassInfo(), new LySimpleObjectList(object), false);
			
			if (controller instanceof LyAreaController)
				_formController.openSubAreaController((LyAreaController)controller, true);

			session.showView(controller);
		}
	}

	/**
	 * Changes current field status to (non-)hidden. If the field is hidden, it is not submited with
	 * all other values. When the field is set hidden, it is also set enabled.
	 * @param hidden   new fieldInfo visibility. (default is true)
	 */
	public void setHidden(boolean hidden)
	{
		if ((_formField != null)
				&& (_formField.isHidden() != hidden))
			_formField.setHidden(hidden);
	}
	
	/**
	 * Accessor to the hidden status of a field in the form.
	 * @return current field visibility
	 */
	public boolean isHidden()
	{
		return ((_formField == null) ? false : _formField.isHidden());
	}
	
	/**
	 * Changes the label for the specified fieldinfo.
	 * @param label the new label
	 * @param labelSet true to use this label in edition
	 */
	public void setLabel(LyInfoLabel label, boolean labelSet)
	{
		LyFormField formField = getFormField();

		if (formField == null)
			return;

		String   			str = null;
        LyInfoLabel.Type    type = label.getType();
		boolean  			reset = false;
		LyObject 			object = getObject();

		if (type == LyInfoLabel.Type.NAME)
		{
			LyCaption caption = formField.getCaption();
			String    oldStr = caption.getLabel();

			str = (labelSet ? label.getString(object) : _fieldInfo.getName());

			if ((str != null)
					&& ((oldStr == null)
						|| ((oldStr != null) && (!oldStr.equals(str)))))
			{
				// Something has changed
				caption.setLabel(str);
				reset = true;
			}
			else if (str == null)
			{
				caption.setLabel(null);
				reset = true;
			}
		}
		else if (type == LyInfoLabel.Type.GRAMMATICAL)
		{
			LyTip	    tip = formField.getTip();
			String	    overviewTip = ((tip == null) ? null : tip.getOverview());
			String	    oldStr = ((tip == null) ? null : tip.getLabel());
			LyInfoLabel tipLabel = getInfoLabel(_fieldInfo, LyInfoLabel.Type.GRAMMATICAL);

			if (labelSet)
				str = label.getString(object);
			else if (tipLabel != null)
				str = tipLabel.getString(object);

			if (str == null)
				str = _fieldInfo.getTip();

			String translatedStr = getEnvironment().translate(str);

			if ((translatedStr != null)
					&& (translatedStr.length() > 0)
					&& ((oldStr == null)
						|| ((oldStr != null) && (!oldStr.equals(str)))))
			{
				// Something has changed
				tip = new LyTip();
				tip.setLabel(str);
				tip.setOverview(overviewTip);
				formField.setTip(tip);
				reset = true;
			}
			else if (((translatedStr == null) || (translatedStr.length() == 0))
					&& (oldStr != null))
			{
				formField.setTip(null);
				reset = true;
			}
		}
		else if (type == LyInfoLabel.Type.SEMANTIC)
		{
			LyTip	    tip = formField.getTip();
			String	    labelTip = ((tip == null) ? null : tip.getLabel());
			String	    oldStr = ((tip == null) ? null : tip.getOverview());
			LyInfoLabel overviewLabel = getInfoLabel(_fieldInfo, LyInfoLabel.Type.SEMANTIC);

			if (labelSet)
				str = label.getString(object);
			else if (overviewLabel != null)
				str = overviewLabel.getString(object);
			else
				str = null;

			String translatedStr = getEnvironment().translate(str);

			if ((translatedStr != null)
					&& (translatedStr.length() > 0)
					&& ((oldStr == null)
						|| ((oldStr != null) && (!oldStr.equals(str)))))
			{
				// Something has changed
				tip = new LyTip();
				tip.setLabel(labelTip);
				tip.setOverview(str);
				formField.setTip(tip);
				reset = true;
			}
			else if (((translatedStr == null) || (translatedStr.length() == 0))
					&& (oldStr != null))
			{
				formField.setTip(null);
				reset = true;
			}
		}

		if (_formController.isInitialized() && reset)
		{
			// Reset the formfield and its parent form
			formField.resetContent();
			_formController.resetFormContent();
		}
	} // end method setLabel
	
	/**
	 * Get the transverse action associated to this field controller
	 * @return a list of transverse actions.
	 */
	public LyInfoDataList<LyAction, LyClassInfo> getTransverseActions()
	{
		// apply on relations (see subclasses LyRelationxxxController)
		return null;
	}
} // end class LyFieldConsultController
