/**
 * LyFieldController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyFieldController.java 34664 2012-03-15 07:28:54Z jlsaouli $
 */

package leon.control.form;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.control.builder.*;
import leon.control.form.builder.*;
import leon.data.*;
import leon.info.*;
import leon.info.misc.*;
import leon.info.save.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * Default field controller.
 * @see LyFormController
 */
public class LyFieldController extends LyController
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyFieldController.java 34664 2012-03-15 07:28:54Z jlsaouli $";

	/** Separator used to build form field identifiers. */
	public static final String FORM_FIELD_SEP = LyFormController.FORM_FIELD_SEP;
	/**Old form field sep */
	public static final String OLD_FORM_FIELD_SEP = ".";

	/**  */
	protected LyFormController _formController;

	/**  */
	protected LyFieldInfo	   _fieldInfo;

	/**  */
	protected LyFormField	   _formField;

	/**  */
	protected LyClassInfo	   _fieldType;

	/**
	 * Constructor.
	 * @param parent  parent controller (or null for 'root' controller).
	 * @param fieldInfo the associated field handled by this controller
	 */
	public LyFieldController(LyAreaController parent, LyFieldInfo fieldInfo)
	{
		super(parent.getSession(), parent,
			parent.getId() + LyController.CONTROLLER_ID_SEP + "field_" + fieldInfo.getId());

		_formController = (LyFormController)parent;
		_fieldInfo = fieldInfo;
		_formField = null;
		_fieldType = null;
	}

	/**
	 * Free this controller.
	 */
	public void free()
	{
		super.free();
		_formController = null;
		_fieldInfo = null;
		_formField = null;
		_fieldType = null;
	}

    /**
     * Clears all cached data for this controller (especially the view and the component).
     *
     * <P>This method may be invoked for memory optimization, but the controller is still valid and
     * may be used. (Unlike free) Further usage, needs to rebuild internal component & view which
     * may be time consuming.</P>
     */
    public void clearViewCache()
    {
        super.clearViewCache();
        _formField = null;
    }
    
	/**
	 * @see leon.control.LyController#getDefaultBuilder()
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return _parent.getBuilder(); // XXX
	}

	/**
	 * @see leon.control.LyController#getBehavior()
	 */
	public LyBehaviorInterface getBehavior()
	{
		return _parent.getBehavior(); // XXX
	}

    /**
     * @see leon.control.LyController#getViewController()
     */
    public LyViewController getViewController()
    {
		return _parent.getViewController(); // XXX
    }
    
	/**
	 * Get the field associated with this controller
	 * @return the associated field
	 */
	public LyFieldInfo getFieldInfo()
	{
		return _fieldInfo;
	}

	/**
	 * Get the classInfo associated with this controller
	 * @return the classInfo
	 */
	public LyClassInfo getClassInfo()
	{
		return _formController.getClassInfo();
	}

	/**
	 * Get the edited object
	 * @return the edited object
	 */
	public LyObject getObject()
	{
		return _formController.getObject();
	}

	/**
	 * Get the current action
	 * @return the action
	 */
	public LyAction getAction()
	{
		return _formController.getAction();
	}

	/**
	 * Get the current context
	 * @return the current context
	 */
	protected LyObjectList getContext()
	{
		return _formController.getContext();
	}

	/**
	 * Get the environment of the parent form controller
	 * @see leon.control.LyController#getEnvironment()
	 */
    public final LyEnvironment getEnvironment()
    {       
        return (_formController == null ? null : _formController.getEnvironment());
    }
    
    /**
     * Get the form controller
     * @return the form controller
     */
    public LyFormController getFormController()
    {
        return _formController;
    }

	/**
	 * Get the name of the image (key or path) representing an object instance.<br>
	 * Call the parent form controller method.
	 * @param  object source object for image association.
	 * @return the image path or key code.
	 */
	public String getImage(LyObject object)
	{
		return _formController.getImage(object);
	}

	/**
	 * Gets info label of the given type for specified class info.
	 * @param classInfo the classInfo
	 * @param type the type of label
	 * @return The info label if found, null otherwise.
	 */
	public LyInfoLabel getInfoLabel(LyClassInfo classInfo, LyInfoLabel.Type type)
	{
		return _formController.getInfoLabel(classInfo, type);
	}

	/**
	 * Gets info label of the given type for specified field info.
	 * @param fieldInfo a fieldInfo
	 * @param type the type of label
	 * @return The info label if found, null otherwise.
	 */
	public LyInfoLabel getInfoLabel(LyFieldInfo fieldInfo, LyInfoLabel.Type type)
	{
		return _formController.getInfoLabel(fieldInfo, type);
	}

	/**
	 * Gets label of an object, for specified field info and value.
	 * @param  object     Object for which a label is required.
	 * @param  fieldInfo  field info.
	 * @param  fieldValue value of the field.
	 * @return Label as a string.
	 */
	public String getLabel(LyObject object, LyFieldInfo fieldInfo, Object fieldValue)
	{
		return _formController.getLabel(object, fieldInfo, fieldValue);
	}

	/**
	 * Invoked when an action occurs. At this level, the implementation manages<BR>
	 * DEMAND_VALIDATE demands. (demand.getType = LyActionDemand.TYPE_ACTION AND demand.getSource()
	 * = LyController.DEMAND_VALIDATE)<BR>
	 * DEMAND_APPLY demands. (demand.getType = LyActionDemand.TYPE_ACTION AND demand.getSource() =
	 * LyController.DEMAND_APPLY)<BR>
	 * Default action (double click)<BR>
	 * Default action (double click)<BR>
	 * VALUE_CHANGED demands. (demand.getType = LyActionDemand.TYPE_VALUE_CHANGED
	 * @param demand the action demand
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			// Transverse action ?
			String		  sourceId = demand.getSource().intern();
			String[]	  parameters = LyString.stringToList(sourceId, ID_SEP);
			LyApplication application = getApplication();
			LyInfo		  info = application.getInfo(parameters[0]);

			if (info instanceof LyAction)
			{
				LyAction action = (LyAction)info;

				if (action != null)
				{
					String classInfoId = null;

					if (parameters.length > 2)
						classInfoId = parameters[1];

					if (classInfoId != null)
					{
						// Transverse action on a Relation
						LyClassInfo classInfo = application.getClassInfo(classInfoId);

						invokeTransverseAction(action, classInfo);
					}
					return;
				}
			}
		}
		_formController.actionDemand(demand);
	} // end method actionDemand
	
	/**
	 * Invoked when an menu item notification is sent to this listener.
	 *
	 * @param demand demand parameters sent by the view component.
	 * @see   #demandPerformed
	 */
	public void menuDemand(LyMenuDemand demand)
	{
		_formController.menuDemand(demand);
	}

	/**
	 * Internal method to invoke on a relation a transverse action for a given classInfo.
	 * @param action    Transverse action.
	 * @param classInfo Class info on which the action is applied.
	 */
	public void invokeTransverseAction(LyAction action, LyClassInfo classInfo)
	{
		LyAreaController controller = (LyAreaController)invokeAction(action, classInfo, null, false);

		if (controller != null)
		{
			_formController.openSubAreaController(controller, true);
			controller.showView();
		}
	}

	/**
	 * Get object value to put it in the form.
	 * @return Value of the reqested field info.
	 */
	protected Object getFormValue()
	{
		LyFormBehaviorInterface behavior = (LyFormBehaviorInterface)getBehavior();

		return behavior.getValue(_formController, _fieldInfo);
	}

	/**
	 * Impact of a value change of a fieldInfo.
	 * @param value     new Value of the field.
	 */
	public void propagate(Object value)
	{
	}
	
	/**
	 * Extract current value for the specified fieldInfo in the form.
	 * @return current value, or null if not found or field is disabled.
	 */
	public LyValue getFieldValue()
	{
		LyObject object = _formController.getObject();

		return ((object == null) ? null : object.getFieldValue(_fieldInfo));
	}

	/**
	 * Update the value of a fieldInfo shown in the view. The method sets the value of the
	 * associated fieldItem.
	 * @param newValue  the new value of the field.
	 * @param oldValue  former value of the field,or null if unknown.
	 * @param propagate Indicates whether the change is notified (propagated) to other fields or
	 *                  not.
	 * @see   #getFieldItem
	 */
	public void updateFieldItem(Object newValue, Object oldValue, boolean propagate)
	{
		updateValue(newValue, oldValue);
	}

	/**
	 * Update the value of a fieldInfo shown in the view. The method sets the value of the
	 * associated fieldItem.
	 * @param newValue the new value of the field.
	 * @param oldValue former value of the field,or null if unknown.
	 */
	protected void updateValue(Object newValue, Object oldValue)
	{
		// Default method consists of updating label
		LyFormField formField = getFormField();
		
		if (formField == null)
			return;
		
		LyFieldItem fieldItem = formField.getFieldItem();
		String	    label;
		LyObject    object = getObject();

		if (object != null)
			label = getLabel(object, _fieldInfo, newValue);
		else
			label = _fieldInfo.toName(newValue);

		if ((label != null)
		    && _fieldInfo.hasMark(LyFieldInfo.MARK_TRANSLATE)
			&& _fieldInfo.isMultiple())
		{
			LyEnvironment environment = getEnvironment();
			String[]	  labels = LyString.stringToList(label, LyApplication.VISIBLE_FIELD_SEP);
			int			  i, n = labels.length;

			for (i = 0, label = ""; i < n; i++)
			{
				if (i > 0)
					label += ", ";
				label += environment.translate(labels[i]);
			}
		}

		fieldItem.setFieldValue(newValue);
		if (label != null)
		{
			boolean	addPrefix = (!_fieldInfo.hasMark(LyFieldInfo.MARK_TRANSLATE)
		            && (fieldItem.getBean() == null)
		            && !_fieldInfo.isMultiple()
		            && (_fieldInfo.getFieldUnit() == null));
			
			if ((_fieldInfo instanceof LyTextFieldInfo)
					&& ((((LyTextFieldInfo)_fieldInfo).getType() == LyTextFieldInfo.Type.RICH)
						|| ((LyTextFieldInfo)_fieldInfo).isMultilingual()))
				addPrefix = false;
			
			if (addPrefix)
		        fieldItem.setLabel("$$" + label);
			else
		        fieldItem.setLabel(label);
		}	
		else
			fieldItem.setLabel(null);
		
		// Update the last modification time.
		setLastModified(System.currentTimeMillis());
	} // end method updateValue

	/**
	 * Accessor to the formField (View component) associated to the fieldInfo. The formField
	 * contains a caption (label of the fieldInfo), a fieldItem (value of the fieldInfo) and the
	 * constraints definition (how the component are shown in the view).
	 * @return the formField corresponding to the given fieldInfo
	 */
	public LyFormField getFormField()
	{
		return (_fieldInfo == null ? null : getFormField(_fieldInfo));
	}

	/**
	 * Accessor to the formField (View component) associated to the fieldInfo. The formField
	 * contains a caption (label of the fieldInfo), a fieldItem (value of the fieldInfo) and the
	 * constraints definition (how the component are shown in the view).
	 * @param  fieldInfo the fieldInfo which is associated to the view component.
	 * @return the formField corresponding to the given fieldInfo
	 */
	public LyFormField getFormField(LyFieldInfo fieldInfo)
	{
		if (_formField == null)
			setFormField(buildFormField(fieldInfo));

		return _formField;
	}

	/**
	 * Sets formField. Called by LyFormFieldBuilder to avoid loops. Reserved for internal use.
	 * @param formField the form field to set
	 */
	public void setFormField(LyFormField formField)
	{
		if (_formField == formField)
			return;
		else if (_formField == null)
			_formField = formField;
		else
			throw new IllegalStateException("BF AppEngine : formField already set on field controller "
				+ this);

		// Register this controller to receive user demands
		LyViewManager viewManager = getViewManager();

		// If the view was cleared, this method may be invoked several times
		viewManager.removeDemandListener(this);
		viewManager.addDemandListener(this);

		formField.setController(getId());
	}

	/**
	 * Internal method to build the form field or tou bfind it in current form if it already exists.
	 * @param  fieldInfo current fieldInfo to associate with a view form object
	 * @return the new or existing form field instance
	 */
	protected LyFormField buildFormField(LyFieldInfo fieldInfo)
	{
		LyBuilder   builder = getBuilder();
		LyFormField formField = null;

		if (_formController.getBuilder() instanceof LyFileBuilder)
		{
			// Should have already built the field
			// Retrieve formField corresponding to fieldInfo
			LyComponent root = _formController.getBuilder().getComponent();

			if (root != null)
			{
				String	    id = _formController.getClassInfo().getId()
					+ FORM_FIELD_SEP + fieldInfo.getId();
				LyComponent component = root.getComponent(id);

				if (component == null)
				{
					// Try with OLD_FORM_FIELD_SEP
					id = _formController.getClassInfo().getId()
						+ OLD_FORM_FIELD_SEP + fieldInfo.getId();
					component = root.getComponent(id);
				}

				if (component instanceof LyFormField)
				{
					formField = (LyFormField)component; // Found it
					_component = component;
				}
			}
		}

		if (formField == null)
		{
			if (builder instanceof LyFormFieldBuilder)
				formField = (LyFormField)getComponent();
			else if (builder instanceof LyFormBuilder)
				formField = ((LyFormBuilder)builder).buildFormField(fieldInfo, false);
			else
				return null;
		}

		if (formField != null)
			formField.setController(getId());
			
		if (_formController.hasMark(fieldInfo, LyFieldInfo.MARK_HIDDEN)
                && !fieldInfo.hasMark(LyFieldInfo.MARK_HIDDEN)) // XXX Needed for Studio
            // Hidden fields for current action are not present in forms but their values are posted on validation
		    formField.setHidden(true);
			
		return formField;
	} // end method buildFormField

	/**
	 * Gets field item used for values input of th controlled field info.
	 * @return the associated field item
	 */
	protected LyFieldItem getFieldItem()
	{
		LyFormField formField = getFormField();
		LyFieldItem fieldItem = ((formField == null) ? null : formField.getFieldItem());
		LyUnit	    unit = getDefaultFieldUnit();

		if (unit != null)
		{
			LyFormat format = unit.getFormat();

			if (format != null)
			{
				// Here fieldItem should be an area
				LyClassInfo classInfo = getClassInfo();

				formField = (LyFormField)formField.getComponent(classInfo.getId()
						+ LyFormController.FORM_FIELD_SEP + _fieldInfo.getId(), true);
				if (formField != null)
					fieldItem = formField.getFieldItem();
			}
		}

		return fieldItem;
	}

	/**
	 * Gets the specific areaController for the given field info.
	 * @return Area controller corresponding to the given field info, null in most cases.
	 */
	public LyAreaController getAreaController()
	{
		return null;
	}

	/**
	 * @see leon.control.LyController#getComponent()
	 */
	public LyComponent getComponent()
	{
		// If component is ready, return it
		if (_component != null)
			return _component;

		_component = getBuilder().buildComponent();

		return _component;
	}

	/**
	 * Method used to set no scrolling to the component in type change case if it's a LyArea
	 * component or a LyTabs component. If the component is a LyTabs, recall this method for the
	 * child component of each tab.
	 * @param component The component which we want set no scrollbar.
	 */
	public void setAreaNoScroll(LyComponent component)
	{
		if (component == null)
			return;

		if (component instanceof LyArea)
			((LyArea)component).setScrolling(LyArea.Scrolling.NO);
		else if (component instanceof LyTabs)
		{
			int   i, tabCount = ((LyTabs)component).getTabCount();
			LyTab tab;

			for (i = 0; i < tabCount; i++)
			{
				tab = ((LyTabs)component).getTab(i);
				setAreaNoScroll(tab.getChild(0));
			}
		}
	}

	/**
	 * Get the field controller at the given index.
	 * @param index an index
	 * @return null
	 */
	public LyFieldController getFieldController(int index)
	{
		return null;
	}

	/**
	 * Get this controller as a string.
	 * @see java.lang.Object#toString()
	 */
	public String toString()
	{
		return getClass().toString() + " : " + _fieldInfo;
	}

	/**
	 * Sets the valid status of a field in the form.
	 * @param status    current validity status
	 */
	public void setState(short status)
	{
		if ((_formField != null)
				&& (_formField.getState() != status))
		{
			_formField.setState(status);
			if (status == LyFormField.STATE_ERROR)
				_formController.setActiveField(_fieldInfo);
		}
	}

	/**
	 * Accessor to the state for a field in the form.
	 * @return current field state
	 */
	public short getState()
	{
		return ((_formField == null) ? -1 : _formField.getState());
	}

	/**
	 * Sets the field content from the specified value Any value modified by the user is the form is
	 * lost. Default implementation just call updateFieldItem but this may be overriden in
	 * subclasses.
	 * @param value          new value to set in the field.
	 * @param freeOldValue   indicates if the old value should be freed (deleted) when it is an
	 *                       object
	 * @param createNewValue indicates if a new empty value should created when object is null and
	 *                       when the value is an object
	 */
	public void resetFromValue(Object value, boolean freeOldValue, boolean createNewValue)
	{
		updateFieldItem(value, null, false);
	}

	/**
	 * Check idle status of the field controller.
	 * @return boolean indicating if the idle mode is on or off.
	 */
	public boolean isIdle()
	{
		return super.isIdle() || _formController.isIdle();
	}

	/**
	 * Sets the value of a fieldInfo.
	 * @param value     the new value of the field. This value may be null, a LyValue instance or
	 *                  the internal value associated to the fieldInfo. See LyValue.getValue()
	 * @param propagate Indicates whether the change is notified (propagated) to other fields or
	 *                  not. This is useless if you reset all fields for instance.
	 * @see   #getFieldValue
	 * @see   leon.data.LyValue#getValue
	 */
	public void setFieldValue(Object value, boolean propagate)
	{
		LyValue fieldValue = null;

		if (value instanceof LyValue)
			fieldValue = (LyValue)value;
		else if (value != null)
			fieldValue = _fieldInfo.buildFieldValue(value);

		setFieldValue(fieldValue, propagate);
	}

	/**
	 * Sets the value of a fieldInfo This methods extracts the value and calls updateFieldItem.
	 * @param newValue     the new value of the field. This value may be null to set the value to blank
	 *                  (empty).
	 * @param propagate Indicates whether the change is notified (propagated) to other fields or
	 *                  not. This is useless if you reset all fields for instance.
	 * @see   #getFieldValue
	 */
	public void setFieldValue(LyValue newValue, boolean propagate)
	{
		// Update field item
		LyValue oldValue = getFieldValue();
		Object  value1 = ((oldValue == null) ? null : oldValue.getValue());
		Object  value2 = ((newValue == null) ? null : newValue.getValue());

		updateFieldItem(value2, value1, propagate);
	}

	/**
	 * Internal method to compute the field value of a number field.
	 * @return Computed field value.
	 */
	protected LyValue computeImportValue()
	{
		LyFieldInfoImport   fieldImport = _fieldInfo.getImport();
		LyRelationFieldInfo importRelation = fieldImport.getRelation();
		LyRelationValue     relationValue = (LyRelationValue)_formController.getFieldValue(
				importRelation);
		Object			    fieldValue = fieldImport.getImportValue(_fieldInfo, relationValue);

		return _fieldInfo.buildFieldValue(fieldValue);
	}
	
    /**
     * Checks if a mark is set for the given fieldInfo in the context of the form action.
     * @param fieldInfo Field to be checked.
     * @param mark Mark to be checked.
     * @return True if the given field has the given mark, false otherwise.
     */
	public boolean hasMark(LyFieldInfo fieldInfo, int mark)
	{
	    return _formController.hasMark(fieldInfo, mark);
	}
	 
    /**
     * Save field input.
     */
    public LyFieldSave save(boolean full)
    {
        return null;
    }
    
    /**
     * Restore field input.
     */
    public boolean restore(LyFieldSave fieldSave)
    {
        return false;
    }

    /**
     * Get the default unit for the current field
     * @return the default unit (can be null)
     */
    public LyUnit getDefaultFieldUnit()
    {
    	return getDefaultFieldUnit(getFieldInfo());
    }

    /**
     * Get the default unit for the given field
     * @param fieldInfo a field
     * @return a default unit
     */
    public LyUnit getDefaultFieldUnit(LyFieldInfo fieldInfo)
    {
    	return _formController.getDefaultFieldUnit(fieldInfo);
    }
} // end class LyFieldController
