/**
 * LyFieldSetController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyFieldSetController.java 41590 2012-11-23 14:51:16Z bangiole $
 */

package leon.control.form;

import leon.app.behavior.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.control.builder.*;
import leon.control.form.builder.*;
import leon.data.*;
import leon.info.*;
import leon.info.misc.*;
import leon.info.save.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;


/**
 * Default field controller for a modification.
 * @see LySetController
 */
public class LyFieldSetController extends LyFieldConsultController
	implements LyControllerContextInterface
{
	/** File version */
	public static final String     __VERSION =
		"$Id: LyFieldSetController.java 41590 2012-11-23 14:51:16Z bangiole $";

	/**
	 * Name of a standard classInfo used in creation when type of the object is dependant of an
	 * other fieldinfo (for example: a group of radio bouton, each one defining a different type so
	 * we have to create the good form when an action occurs).
	 */
	public static final String     STANDARD_CLASSINFO = "Relation_Class";

	/** Name of the tranverse tool like STANDARD_CLASSINFO but for internal update in an array. */
	public static final String     STANDARD_ARRAY_TOOL = "Array_Class";

	/**
	 * Separator used for labels of transverse action tools (inserted between action name and class
	 * name).
	 */
	public static final String     TRANSVERSE_ACTION_SEP = " : ";

	/**  */
	protected LySetController	   _setController;

	/**  */
	protected LyFieldSetController _parentFieldController;

	/** Initial reference value. */
    protected LyValue              _initialValue;

	/** Initial reference value. */
	protected LyValue			   _lastValue;

	/** Last reference value. */
	protected LyValue			   _currentValue;

	/** Last reference unit. */
	protected LyUnit			   _currentUnit;

	/**
	 * Internal use to indicate that the controller is an internal hidden controller with no GUI
	 * (like in array).
	 */
	protected boolean			   _internal;

	/**
	 * Store the transverse action to avoid recomputation
	 */
	protected LyInfoDataList<LyAction, LyClassInfo> _transverseActions;

	/**
	 * Constructor.
	 * @param parent parent controller
	 * @param fieldInfo fieldInfo handled by this controller
	 */
	public LyFieldSetController(LySetController parent, LyFieldInfo fieldInfo)
	{
		super(parent, fieldInfo);

		_setController = parent;
		_parentFieldController = null;
		_currentValue = null;
		_currentUnit = null;
		_internal = false;
	}

	/**
	 * @see leon.control.form.LyFieldConsultController#getDefaultBuilder()
	 */
	public LyBuilder getDefaultBuilder()
	{
		return new LyFieldSetBuilder(this);
	}

	/**
	 * Get the transverse action associated to this field controller
	 * @return a list of transverse actions.
	 */
	public LyInfoDataList<LyAction, LyClassInfo> getTransverseActions()
	{
		if (_transverseActions != null)
			return _transverseActions;

		LySetController			setController = (LySetController)getParent();
		LySetBehaviorInterface	behavior = setController.getSetBehavior();

		_transverseActions = behavior.getTransverseActions(setController, _fieldInfo);
		return _transverseActions;
	}

	/**
	 * Return the mode of the controller
	 * @return true if in creation mode
	 */
	public boolean isCreation()
	{
		return _setController.isCreation(); // XXX
	}

	/**
	 * get the model object
	 * @return the model object
	 */
	public LyObject getModel()
	{
		return _setController.getModel(); // XXX
	}

	/**
	 * get the associated Set controller
	 * @return the set controller
	 */
	public LySetController getSetController()
	{
		return _setController;
	}

	/**
	 * Update the view of the controller with the specified values from the demand. This will force
	 * the graphical objects to use current demand values. This demand is performed after a user
	 * interaction. Unlike updateViewContent, the controller is only taking the values from the
	 * demand, on 'poling' the possible values.
	 * @see   #updateViewContent
	 * @param demand demand parameters sent by the view component.
	 */
	public void updateViewFromValues(LyDemand demand)
	{
		super.updateViewFromValues(demand);

		// Also propagate on the set controller .
		_setController.updateViewFromValues(demand);
	}

	/**
	 * Check (when it makes sence) the graphical component to see if they have a pending value typed
	 * by the user, and not yet refreshed in the LyView components. Like a textfield, which value
	 * may have changed in the graphical component and this modification has not been notified. This
	 * method may do nothing, depending on the nature of the graphical component.
	 */
	public void updateViewContent()
	{
		LyFieldItem fieldItem = getFieldItem();

		// Be careful : maybe the user was typing a value which is not validated yet ?
		fieldItem.updateViewContent();
	}

	/**
	 * Invoked when an action occurs. At this level, the implementation manages<BR>
	 * DEMAND_VALIDATE demands. (demand.getType = LyActionDemand.TYPE_ACTION AND demand.getSource()
	 * = LyController.DEMAND_VALIDATE)<BR>
	 * DEMAND_APPLY demands. (demand.getType = LyActionDemand.TYPE_ACTION AND demand.getSource() =
	 * LyController.DEMAND_APPLY)<BR>
	 * Default action (double click)<BR>
	 * Default action (double click)<BR>
	 * VALUE_CHANGED demands. (demand.getType = LyActionDemand.TYPE_VALUE_CHANGED
	 * @param demand the action demand
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_VALUE_CHANGED)
		{
			LyParameterValues values = demand.getParameterValues();

			// Update the view and send valueChange events for other fields if needed
			if (values != null)
			{
				_setController.updateViewFromValues(demand);
				_setController.setLastModified(System.currentTimeMillis());

				// If field is contained in a composition, call setLastModified() on the
				// composition controller to correctly update the web view
				LyRelationFieldInfo relation = _setController.getCompositionRelation();
				LyController	    parent = _setController.getParent();

				if ((relation != null) && (parent instanceof LyFormController))
				{
					LyFormController  formController = (LyFormController)parent;
					LyFieldController compositionController = formController.getFieldController(
							relation);

					if (compositionController != null)
						compositionController.setLastModified(System.currentTimeMillis());
				}
			}

			valueChanged();
			return;
		}
		else if (demand.getType() == LyActionDemand.TYPE_UNIT_CHANGED)
		{
			unitChanged(demand.getValue(), demand.getAdditionalValue());
			return;
		}
		else if (demand.getType() == LyActionDemand.TYPE_COMPLETE_VALUE)
		{
			completeValue(demand.getValue());
			return;
		}
		else if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			// Transverse action ON ARRAY?
			String   sourceId = demand.getSource().intern();
			String[] parameters = LyString.stringToList(sourceId, ID_SEP);

			if (parameters.length > 1)
			{
				if (parameters[1].equals(LyFieldSetController.STANDARD_ARRAY_TOOL))
				{
					// Transverse action on a Array : must create / set an entry in the table
					if (_parentFieldController instanceof LyArrayFieldSetController)
						_parentFieldController.actionDemand(demand);
					return;
				}
			}
		}

		// A validate demand may be sent before sending the new field value (when using the
		// CR key in an edit table for example).
		// Ensure value is up-to-date by calling the valueChanged() method first.
		if ((_formController instanceof LyEditTableSetController)
				&& (LyFormController.DEMAND_VALIDATE == demand.getSource()))
			valueChanged();

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Sets the value of a fieldInfo This methods extracts the value and calls updateFieldItem.
	 * @param newValue     the new value of the field. This value may be null to set the value to blank
	 *                  (empty).
	 * @param propagate Indicates whether the change is notified (propagated) to other fields or
	 *                  not. This is useless if you reset all fields for instance.
	 * @see   #getFieldValue
	 */
	public void setFieldValue(LyValue newValue, boolean propagate)
	{
		getEnvironment().logTrace("setValue " + _fieldInfo + " = " + newValue);

		// Update field item
		LyValue oldValue = getFieldValue();

		// Transform value to chosen unit if needed
		LyUnit  unit = getUnit();

		if ((unit != null)
				&& (newValue != null)
				&& !newValue.isNull())
		{
			Object value = _fieldInfo.convertToFieldUnit(newValue.getValue(), unit);

			newValue.setValue(value);
		}

		Object value1 = ((oldValue == null) ? null : oldValue.getValue());
		Object value2 = ((newValue == null) ? null : newValue.getValue());

		updateFieldItem(value2, value1, propagate);
	}

	/**
	 * Get current unit for the specified fieldInfo in the form.
	 * @return current unit, or null if not found or field is disabled.
	 */
	public leon.info.LyUnit getUnit()
	{
		LyUnits units = _fieldInfo.getUnits();

		if (units == null)
			return null;

		if ((_formField == null) || !_formField.isEnabled())
			return getDefaultFieldUnit();

		LyFieldUnit formFieldUnit = _formField.getFieldUnit();

		if (formFieldUnit == null)
			return null;

		LyFieldItem unitItem = formFieldUnit.getFieldItem();
		String	    unitId = unitItem.getValue();

		if ((unitId == null) && (unitItem instanceof LyRelation))
		{
			LyRelation relation = (LyRelation)unitItem;
			int		   nLink = relation.getLinkCount();
			LyLink     link;

			for (int j = 0; (j < nLink) && (unitId == null); j++)
			{
				link = relation.getLink(j);
				if (link.isSelected())
					unitId = link.getId();
			}
		}

		LyUnit unit = units.getUnit(unitId);

		if (unit == null)
			unit = getDefaultFieldUnit();//_fieldInfo.getFieldUnit();

		return unit;
	} // end method getUnit

	/**
	 * Extract final value for the specified fieldInfo in the form.
	 * @return Final value, or null if not found or field is disabled.
	 */
	public LyValue getFinalValue()
	{
		return getFieldValue();
	}

	/**
	 * Extract current value for the specified fieldInfo in the form.
	 * @return current value, or null if not found or field is disabled.
	 */
	public LyValue getFieldValue()
	{
		if (_setController.isConsult(_fieldInfo))
			return super.getFieldValue();

		LyFormField formField = getFormField();

		if ((formField == null) || !formField.isEnabled())
			return null;

		LyUnit unit = getUnit();

		return getFieldValue(unit);
	}

	/**
	 * Extract current value for the specified fieldInfo in the form.
	 * @param  unit Unit of the value.
	 * @return current value, or null if not found or field is disabled.
	 */
	public LyValue getFieldValue(LyUnit unit)
	{
		LyFieldItem fieldItem = getFieldItem();
		String	    strValue = null;

		if ((unit != null)
				&& unit.hasEditionFormat())
		{
			LyFormat format = unit.getFormat();
			Object   value = null;

			strValue = ((LyGetLabelInterface)fieldItem).getLabel();
			if (strValue != null)
			{
				value = format.parse(_fieldInfo, strValue, unit);
				value = _fieldInfo.convertFromFieldUnit(value, unit);
			}
			return _fieldInfo.buildFieldValue(value);
		}

		// Build value and return it
		LyValue fieldValue;
		Object  objectValue;

		strValue = getStringValue();

		try
		{
			objectValue = _fieldInfo.parse(strValue);
			fieldValue = _fieldInfo.buildFieldValue(objectValue);
		}
		catch (LyFieldInfoFormatException e)
		{
			if (getEnvironment().isDebugLevel(LyEnvironment.DEBUG_TRACE))
				getEnvironment().writeException(e);

			//FFT00047
			// We display the error message.
			LyError  error;
			LyObject object = _setController.getObject();

			String message = e.getMessage();
			if ((message != null) && message.startsWith("BF AppEngine : "))
			{
				message = message.substring(15);
			}
			error = new LyError(LyObject.LEVEL_ERROR, message, null, _fieldInfo, object, _fieldInfo);
			showError(object, error);

			// Setting the form field to invalid to stop the process.
			setState(LyFormInterface.STATUS_KO);

			return null;
		}

		LyUnits units = _fieldInfo.getUnits();

		if ((unit != null)
				&& (units != null)
				&& !fieldValue.isNull())
		{
			Object value = _fieldInfo.convertFromFieldUnit(fieldValue.getValue(), unit);

			fieldValue.setValue(value);
		}

		return fieldValue;
	} // end method getFieldValue

	/**
	 * Get the value for this controller as String
	 * @return a string value
	 */
	protected String getStringValue()
	{
		LyFieldItem fieldItem = getFieldItem();

		if (fieldItem instanceof LyGetLabelInterface)
		{
			// Be careful : maybe the user was typing a value which is not validated yet ?
			fieldItem.updateViewContent();

			return ((LyGetLabelInterface)fieldItem).getLabel();
		}
		else
			return null;
	}

	/**
	 * Update the field with the given value
	 * @param newValue the new value
	 * @param oldValue the old value
	 * @return return the real new value
	 */
	protected Object updateFieldItem(Object newValue, Object oldValue)
	{
        boolean     userChange = (_currentValue != null)
            && _fieldInfo.equals(newValue, _currentValue.getValue());
		LyUnit	    unit = getUnit();
		LyUnit	    fieldUnit = getDefaultFieldUnit();
		LyFieldItem fieldItem = getFieldItem();

		if (((unit != null)
					&& (unit.hasEditionFormat()))
				|| ((unit == null)
					&& (fieldUnit != null)
					&& fieldUnit.hasEditionFormat()))
		{
			LyFormat format = (((unit != null) && unit.hasEditionFormat()) ? unit.getFormat()
																		   : fieldUnit.getFormat());

			String   label = format.toString(_fieldInfo, newValue, unit);

			fieldItem.setLabel(label);
			fieldItem.setFieldValue(newValue);
		}
		else
		{
			updateValue(newValue, oldValue);
			// New value may have changed when updating
			newValue = getValue();
		}

        if (!userChange)
        {
            // Change not performed by end-user
            _currentValue = null;
            _currentUnit = null;
        }

		return newValue;
	} // end method updateFieldItem

	/**
	 * Internal use : update the value of a fieldInfo in the form. <B>Application should use
	 * setFieldValue instead.</B> For relation fieldInfo, this method invokes updateRelation
	 * (specialization). For choice fieldInfo, this method invokes updateChoice (specialization).
	 * @param newValue  the new value of the field. This value may be null, a LyValue instance or
	 *                  the internal value associated to thge fieldInfo. See LyValue.getValue()
	 * @param oldValue  the latest value of the field. This value may be null, a LyValue instance or
	 *                  the internal value associated to thge fieldInfo. See LyValue.getValue()
	 * @param propagate Indicates whether the change is notified (propagated) to other fields or
	 *                  not. This is useless if you reset all fields for instance.
	 * @see   #setFieldValue
	 */
	public void updateFieldItem(Object newValue, Object oldValue, boolean propagate)
	{
	    Object     initialValue = newValue;

		if (_setController.isConsult(_fieldInfo))
			super.updateFieldItem(newValue, oldValue, propagate);
		else
			newValue = updateFieldItem(newValue, oldValue);

		if (propagate)
		{
			// Propagate value change if the value really changed
			if (!_fieldInfo.equals(newValue, oldValue))
				valueChanged();
			else
				propagate = false;
		}

		// As soon as the field info is updated, reset valid state
		setState(LyFormInterface.STATUS_OK);

		if (propagate)
		{
			LyFieldItem fieldItem = getFieldItem();

			// Sets field value on field item
			fieldItem.setFieldValue(newValue);
		}

        _lastValue = _fieldInfo.buildFieldValue(newValue);

        LySetController   setController = getSetController();

        if (!setController.isInitialized())
            _initialValue = _fieldInfo.buildFieldValue(initialValue);

		// Update the last modification time.
		setLastModified(System.currentTimeMillis());
	} // end method updateFieldItem

	/**
	 * Internal use : update the value of a fieldInfo in the form.
	 * @param newValue the new value of the field. This value may be null, a LyValue instance or the
	 *                 internal value associated to thge fieldInfo. See LyValue.getValue()
	 * @param oldValue the latest value of the field. This value may be null, a LyValue instance or
	 *                 the internal value associated to thge fieldInfo. See LyValue.getValue()
	 */
	protected void updateValue(Object newValue, Object oldValue)
	{
		if (_setController.isConsult(_fieldInfo))
		{
			super.updateValue(newValue, oldValue);
			return;
		}

		// Default method consists to update label
		LyFieldItem fieldItem = getFieldItem();
		String	    label = _fieldInfo.toString(newValue);

		if (fieldItem != null)
		{
			fieldItem.setLabel(label);
			fieldItem.setFieldValue(newValue);
		}
	}

	/**
	 * Internal method for control and propagate a value change. This methods is called when the
	 * field has been changed by the user and checks current value validity.
	 * <P>This validity is called for possible specific application implementation on the
	 * LyClassBehavior instance.</P>
	 * <P>The LySetController additionally impact the view using the setValid method.</P>
	 * @see   leon.app.behavior.LyClassBehavior#controlNewValue
	 */
	protected void valueChanged()
	{
		valueChanged(false);
	}

	/**
	 * Internal method for control and propagate a value change. This methods is called when the
	 * field has been changed by the user and checks current value validity.
	 * <P>This validity is called for possible specific application implementation on the
	 * LyClassBehavior instance.</P>
	 * <P>The LySetController additionally impact the view using the setValid method.</P>
	 * @param force indicate if propage must be performed even if values has not changed
	 * @see   leon.app.behavior.LyClassBehavior#controlNewValue
	 */
	protected void valueChanged(boolean force)
	{
		if (_idle)
			return;

		Object  value = getValue();

        if (!force && (_currentValue != null)
                && _fieldInfo.equals(value, _currentValue.getValue()))
            return;

        if (!force && (_currentValue == null)
                && (_lastValue != null)
                && _fieldInfo.equals(value, _lastValue.getValue()))
        	return;

        // MAZ IPANEMA FFT00214
		if (!force && ((_currentValue != null)
				&& _currentValue.equals(_lastValue))
			&& ((_lastValue != null)
				&& _fieldInfo.equals(value, _lastValue.getValue())))
		{
			_currentValue = null;
			return;
		}
		_currentValue = _fieldInfo.buildFieldValue(value);

		_currentUnit = getUnit();
		LyUnits units = _fieldInfo.getUnits();

		if ((units != null) && (!units.isConvert()))
		{
			Object	newValue = _fieldInfo.convertToFieldUnit(_currentValue == null ? null : _currentValue.getValue(), _currentUnit);

			_currentValue = _fieldInfo.buildFieldValue(newValue);
		}

		if (_parentFieldController != null)
			_parentFieldController.valueChanged(force);
		else
			_setController.valueChanged(_fieldInfo);
	}

	/**
	 * Internal method for completing a formatted value (for a unit). This methods is called when
	 * the field unit has been changed by the user.
	 * @param value the current fieldValue (not completed).
	 */
	public void completeValue(String value)
	{
		if (_setController.isConsult(_fieldInfo))
			return;

		LyFormField formField = getFormField();

		if ((formField == null) || !formField.isEnabled())
			return;

		LyUnit unit = getUnit();

		if (unit == null)
			return;

		if (unit.getFormat() == null)
			return;

		String completedValue = unit.getFormat().completeValue(_fieldInfo, value, unit);

		if (completedValue != null)
		{
			LyFieldItem fieldItem = getFieldItem();

			fieldItem.setLabel(null);
			fieldItem.setLabel(completedValue);
		}
	} // end method completeValue

	/**
	 * Internal method for propagating a unit change. This methods is called when the field unit has
	 * been changed by the user.
	 * <P>The LySetController additionaly impact the view using the setValid method.</P>
	 * @param newUnitId The new fieldinfo unit.
	 * @param oldUnitId The old fieldinfo unit.
	 */
	public void unitChanged(String newUnitId, String oldUnitId)
	{
		if ((_formField == null) || !_formField.isEnabled())
			return;

		// the new unit id
		LyUnits units = _fieldInfo.getUnits();
		LyUnit  oldUnit = ((oldUnitId == null) ? getUnit() : units.getUnit(oldUnitId));
		LyUnit  newUnit = ((newUnitId == null) ? getUnit() : units.getUnit(newUnitId));

		// The units are the same
		if (newUnit == oldUnit)
			return;
		boolean forceUpdate = false;

		if ((newUnit.getFactor() != 1.0d)
				|| (oldUnit.getFactor() != 1.0d)
				|| (newUnit.getOffset() != 0.0d)
				|| (oldUnit.getOffset() != 0.0d))
			forceUpdate = true;
		// Update the view layer
		LyFieldUnit formFieldUnit = _formField.getFieldUnit();
		LyRelation  relation = (LyRelation)formFieldUnit.getFieldItem();

		relation.selectLink(oldUnit.getId(), false);
		relation.selectLink(newUnit.getId(), true);

		// the field value
		LyFieldInfoValue fieldValue = (LyFieldInfoValue)getFieldValue(oldUnit);

		if (fieldValue == null)
			// An error occured when getting the value.
			return;

		if (_currentUnit == null)
		{
			_currentValue = fieldValue;
			_currentUnit = oldUnit;
		}

		Object  oldValue = (_currentValue == null ? null : _currentValue.getValue()), newValue;
		boolean convert = units.isConvert();

		newValue = oldValue;

		// Convert the value from the reference unit to the new unit.
		if (convert)
			newValue = _fieldInfo.convertToFieldUnit(oldValue, newUnit);

		if (oldUnit.hasEditionFormat()
				|| newUnit.hasEditionFormat())
		{
			// Rebuild field item
			LyFieldSetBuilder builder = (LyFieldSetBuilder)getBuilder();
			LyFieldItem		  fieldItem = builder.buildFieldItem(_fieldInfo, newUnit);
			LyToolBar		  toolBar = null;

			if (getAreaController() == null)
				toolBar = builder.updateToolBar(_fieldInfo);

			if (_parentFieldController instanceof LyArrayFieldSetController)
			{
				LyTool tool = ((LyArrayFieldSetController)_parentFieldController).buildInnerTool(
						builder);

				if (tool != null)
					toolBar.addTool(tool);
			}

			_formField.setToolBar(toolBar);
			_formField.setFieldItem(fieldItem);
			updateFieldItem(newValue, oldValue, false);
			_formField.resetContent();

			// Reset the form
			_setController.resetFormContent();
		}
		else
		{
			if (convert
				&& (_fieldInfo instanceof LyNumberFieldInfo))
			{
				_idle = true;

				// Convert min, max, pitch values
				LyNumberFieldInfo numberField = (LyNumberFieldInfo)_fieldInfo;
				LyNumber		  number = (LyNumber)_formField.getFieldItem();
				double			  vmin, vmax, vpitch;

				vmin = number.getVMin();
				vmax = number.getVMax();
				vpitch = number.getPitch();

				Number min = numberField.getMin();
				Number max = numberField.getMax();
				Number pitch = numberField.getPitch();

				if (min != null)
				{
					min = new Double(vmin);
					min = (Number)_fieldInfo.convertFromFieldUnit(min, oldUnit);
					min = (Number)_fieldInfo.convertToFieldUnit(min, newUnit);
					number.setVMin(min.doubleValue());
				}
				if (max != null)
				{
					max = new Double(vmax);
					max = (Number)_fieldInfo.convertFromFieldUnit(max, oldUnit);
					max = (Number)_fieldInfo.convertToFieldUnit(max, newUnit);
					number.setVMax(max.doubleValue());
				}
				if (pitch != null)
				{
					pitch = new Double(vpitch);
					pitch = (Number)_fieldInfo.convertFromFieldUnit(pitch, oldUnit);
					pitch = (Number)_fieldInfo.convertToFieldUnit(pitch, newUnit);
					number.setVMin(pitch.doubleValue());
				}

				_idle = false;
			}

			updateFieldItem(newValue, oldValue, false);
		}

		// The unit has changed : we must control the value with the new unit.
		valueChanged(forceUpdate);
	} // end method unitChanged

	/**
	 * The form is closing and we check the values typed in by the user.
	 * @param free indicates if the form is closed and freed after a successful validation.
	 * @return the validation status
	 * @see LyFormInterface
	 */
	protected short validate(boolean free)
	{
		if (_setController.isCreation())
			return LyFormInterface.STATUS_IGNORE;
		else
			return LyFormInterface.STATUS_OK;
	}

	/**
	 * Start validation of the form.
	 * @return the start validate status
	 * @see LyFormInterface
	 */
	public short startValidate()
	{
		return LyFormInterface.STATUS_OK;
	}

	/**
	 * End validation of the form : commit edition.
	 * @return the end validate status
	 * @see LyFormInterface
	 */
	public short endValidate()
	{
		return LyFormInterface.STATUS_OK;
	}

	/**
	 * Method called when all modifications are committed.
	 */
	public void commit()
	{
	    _lastValue = _currentValue;
        _currentValue = null;
        _currentUnit = null;
	}

	/**
	 * Method called when the form is cancelled.
	 */
	public void cancel()
	{
        _currentValue = null;
        _currentUnit = null;
	}

	/**
	 * Accessor to the internal value associated to a filedInfo. This is a shortcut to
	 * getFieldValue(fieldInfo).getValue()
	 * @return the value if found and set, null otherwise.
	 */
	public Object getValue()
	{
		LyValue value = getFieldValue();

		return ((value == null) ? null : value.getValue());
	}

	/* Utility methods */

	/**
	 * Clears specified fieldInfo value : setting value to null.
	 */
	public void clear()
	{
		setFieldValue(null, true);
	}

	/**
	 * Resets specified fieldInfo value : the default value is fetched from the associated behavior.
	 * LyClassBehavior.getValue (LySetBehaviorInterface) The field is restored to its original value
	 * when the view was opened.
	 */
	public void reset()
	{
		LyObject object = _setController.getObject();

		if (object != null)
			object.removeNewValue(_fieldInfo);

		LySetBehaviorInterface behavior = _setController.getSetBehavior();
		Object				   value = behavior.getValue(_setController, _fieldInfo);

		if (value == null)
			value = getDefaultValue();

		setFieldValue(value, true);
	}

	/**
	 * Utility method (internal use for behavior) to calculate formulas of fieldInfos.
	 * Application should use setFieldValue instead.
	 * @param origin    the field that propagates the modification
	 * @param propagate Indicates whether the change is notified (propagated) to other fields or
	 *                  not.
	 * @param valueChange indicates if thie event occurs after a value changed
	 * @see   #setFieldValue
	 */
	public void update(LyFieldInfo origin, boolean propagate, boolean valueChange)
	{
        if (isEnabled()
            && (_fieldInfo.getFormula() != null)
            && (_setController.isConsult(_fieldInfo) || valueChange))
        {
            LyInfoList<LyFieldInfo>       formulaFields = _fieldInfo.getFormulaFields();

            if ((formulaFields != null)
                 && formulaFields.contains(origin))
            {
                LyValue value = _setController.computeFieldValue(_fieldInfo);

                setFieldValue(value, propagate);
            }
        }
        else
        {
    		// Reset field and propagate modification
    		LyValue value = getFieldValue();

    		if (_fieldInfo.isImported()
    				&& isEnabled()
    				&& (_fieldInfo.isLocal() || _setController.isCreation()))
    			value = computeImportValue();

    		setFieldValue(value, propagate);
        }
	}

	/**
	 * Changes current field status to (non-)optional. An optional field may accpt an empty (null)
	 * value and its label is not shown as bold.
	 * @param optional new fieldInfo status. (default is from the view description)
	 */
	public void setOptional(boolean optional)
	{
		if ((_formField != null)
				&& (_formField.isOptional() != optional))
		{
			if (_formField.getState() == LyFormField.STATE_ERROR)
				_formField.setState(LyFormField.STATE_NORMAL);

			_formField.setOptional(optional);

			// If field is not optional and the value is null : set the default value
			Object oldValue = getValue();

			if (!optional && (oldValue == null))
			{
				Object defaultValue = getDefaultValue();

				if (defaultValue != null)
					setFieldValue(defaultValue, true);
			}
		}
	}

	/**
	 * Accessor to the optional status for a field in the form.
	 * @return current field status
	 */
	public boolean isOptional()
	{
		return ((_formField != null) && _formField.isOptional());
	}

	/**
	 * Changes current field status to (non-)editable. If the field is not editable, the value is
	 * read-only but however submited with all other values. A disabled field is not submited with
	 * all other values and is just ignored.
	 * @param editable new fieldInfo status. (default is true or obtained from the view
	 *                  description)
	 */
	public void setEditable(boolean editable)
	{
		if ((_formField != null)
				&& (_formField.isEditable() != editable))
		{
			if (editable)
			{
				LySetController	setController = getSetController();

				if (!setController.isEditable())
					return;
			}

			if (_formField.getState() == LyFormField.STATE_ERROR)
				_formField.setState(LyFormField.STATE_NORMAL);
			_formField.setEditable(editable);
		}
	} // end method setEditable

	/**
	 * Accessor to the editable/ non editable status for a field in the form.
	 * @return current field status
	 */
	public boolean isEditable()
	{
		return ((_formField != null) && _formField.isEditable());
	}

	/**
	 * Changes current field status to (non-)enabled. If the field is not editable, and not submited
	 * with all other values.
	 * @param enabled  new fieldInfo status. (default is true or obtained from the view
	 *                  description)
	 */
	public void setEnabled(boolean enabled)
	{
		if ((_formField != null)
				&& (_formField.isEnabled() != enabled))
		{
			if (!enabled)
			{
				// if the field change from enable to not enable, clear its value
				clear(); // clear
				_formField.setEnabled(false);
			}
			else
			{
				_formField.setEnabled(true);
				reset(); // reset
			}

            LyToolBar toolBar = _formField.getToolBar();

            if (toolBar != null)
            {
                int         i, n = toolBar.getComponentCount();
                LyComponent component;

                for (i = 0; i < n; i++)
                {
                    component = toolBar.getComponent(i);
                    if (component instanceof LyTool)
                        ((LyTool)component).setEnabled(enabled);
                }
            }
		}
	}

	/**
	 * Accessor to the enabled/disabled status for a field in the form.
	 * @return current field status
	 */
	public boolean isEnabled()
	{
		return ((_formField != null) && _formField.isEnabled());
	}

	/**
	 * Changes current field status to (non-)hidden. If the field is hidden, it is not submited with
	 * all other values. When the field is set hidden, it is also set enabled.
	 * @param hidden   new fieldInfo visibility. (default is true)
	 */
	public void setHidden(boolean hidden)
	{
		if ((_formField != null)
				&& (_formField.isHidden() != hidden))
		{
			if (hidden)
			{
				// if the field is hidden, change it to not enabled
				setEnabled(false);
				_formField.setHidden(true);
			}
			else
			{
				_formField.setHidden(false);
				// set the field enabled
				setEnabled(true);
			}
		}
	}

	/**
	 * Get object value to put it in the form.
	 * @return Value of the reqested field info.
	 */
	protected Object getFormValue()
	{
		LyFormBehaviorInterface behavior = (LyFormBehaviorInterface)getBehavior();
		Object				    value = behavior.getValue(_setController, _fieldInfo);

		if (value == null)
			value = getDefaultValue();

		return value;
	}

	/**
	 * Internal method to get default value for a field info.
	 * @return The default value for the given field.
	 */
	protected Object getDefaultValue()
	{
		if ((_formField == null) || !_formField.isEnabled())
			return null;

		if (isOptional())
			return null;

		LySetBehaviorInterface behavior = (LySetBehaviorInterface)getBehavior();

		return behavior.getDefaultValue(_setController, _fieldInfo);
	}

	/**
     * Gets last value of this field controller.
     * @return The last value of this controller.
     */
    public final LyValue getLastValue()
    {
        return _lastValue;
    }

	/**
	 * Checks if the value for the fieldInfo has been modified.
	 * @return true if the value has been modified, false otherwise.
	 */
	public boolean checkIfModified()
	{
        // Check if value or unit was changed by end-user
		return (isEnabled()
		  && ((_currentValue != null) || (_currentUnit != null)));
	}

	/**
	 * Gets the controller context.
	 * @return The controller context.
	 */
	public LyContext getControllerContext()
	{
		return _setController.getControllerContext();
	}

	/**
	 * Controls current edition.
	 * @return The error status.
	 */
	public short control()
	{
		return LyFormInterface.STATUS_OK;
	}

	/**
	 * Identify this controller as being hidden.
	 * @param status the internal status
	 */
	public void setInternal(boolean status)
	{
		_internal = status;
	}

	/**
	 * Sets the field content from the specified value Any value modified by the user is the form is
	 * lost. Default implementation just call updateFieldItem but this may be overriden in
	 * subclasses.
	 * @param value          new value to set in the field.
	 * @param freeOldValue   indicates if the old value should be freed (deleted) when it is an
	 *                       object
	 * @param createNewValue indicates if a new empty value should created when object is null and
	 *                       when the value is an object
	 */
	public void resetFromValue(Object value, boolean freeOldValue, boolean createNewValue)
	{
		setFieldValue(value, false);
	}

	/**
     * Indicates if the operations of the controller context must be inserted at the beginning of the parent
     * operations or at the end.
     * @return Position where operations must be inserted, may be null.
	 */
	protected LyContext.Position getInsertPosition()
	{
        if (!(_fieldInfo instanceof LyRelationFieldInfo))
            return null;

	    LySetBehaviorInterface behavior = _setController.getSetBehavior();
	    LyContext.Position     position;

	    if (behavior != null)
	    {
	        position = behavior.getInsertPosition(_setController, (LyRelationFieldInfo)_fieldInfo);
	        if (position != null)
	            return position;
	    }

	    LyEnvironment          environment = getEnvironment();
	    String                 str = environment.getResource("LY_CONTEXT_INSERT_RELATION");

	    if (str != null)
	    {
	        if (str.equalsIgnoreCase("FIRST"))
	            return LyContext.Position.FIRST;
	        else if (str.equalsIgnoreCase("END"))
	            return LyContext.Position.END;
	    }

        boolean     optional = hasMark(_fieldInfo, LyFieldInfo.MARK_OPTIONAL);

        position = (optional ? LyContext.Position.END : LyContext.Position.FIRST);
        return position;
	}

	/**
	 * Add a file to the delete on free list
	 * @param path the file path to add to the delete list
	 */
	public void addTempFilePath(String path)
	{
		// we cannot delete file with a full cache (otherwise file won't be found)
		if (getFormController() instanceof LySetController)
			((LySetController)getFormController()).addTempFilePath(path);
	}

    /**
     * Save field input.
     */
    public LyFieldSave save(boolean full)
    {
        LyValue newFieldValue = getFieldValue();
        Object  newValue = (newFieldValue == null ? null : newFieldValue.getValue());

        if (full)
            return new LyFieldSave(_fieldInfo, newValue);

        LyValue oldFieldValue = _initialValue;
        Object  oldValue = (oldFieldValue == null ? null : oldFieldValue.getValue());

        if (!_fieldInfo.equals(newValue, oldValue))
            return new LyFieldSave(_fieldInfo, newValue, oldValue);
        else
            return null;
    }

    /**
     * Restore field input.
     */
    public boolean restore(LyFieldSave fieldSave)
    {
        if (fieldSave == null)
            return false;
        if (fieldSave._fieldInfo != _fieldInfo)
            return false;
        if (fieldSave._newValue == fieldSave._oldValue)
            // Could happen on compositions
            return true;

        // Gets current value
        LyValue fieldValue = _initialValue;

        // Compare old and current value before allowing restore for field info.
        if (!_setController.isCreation()
            && (fieldSave._oldValue != LyFieldSave.UNKNOWN_VALUE)
            && (fieldValue != null)
            && !_fieldInfo.equals(fieldValue.getValue(), fieldSave._oldValue)
            && !_fieldInfo.equals(fieldValue.getValue(), fieldSave._newValue))
        {
            // Field value can't be restored because it was changed after save
            LyObject object = _setController.getObject();
            LyError  error = new LyError(LyObject.LEVEL_ERROR, "LY_ER_RESTORE", new Object[] { _fieldInfo },
                  _fieldInfo, object, _fieldInfo);

            _setController.showError(object, error);
            setState(LyClassBehavior.STATUS_KO);
            return false;
        }
        else if ((fieldValue != null)
        		&& _fieldInfo.equals(fieldValue.getValue(), fieldSave._newValue))
            return true; // no need to restore
        else
        {
            // Update field item
            setFieldValue(fieldSave._newValue, true);
            return true;
        }
    }

    public final LyFieldSetController getParentFieldController()
    {
    	return _parentFieldController;
    }
} // end class LyFieldSetController
