/**
 * LyMultiCompositionArrayFieldSetController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyMultiCompositionArrayFieldSetController.java 28997 2010-12-23 10:17:15Z bmichel $
 */

package leon.control.form;

import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.view.*;


/**
 * Field controller for a modification of an array of composition relation in edit table mode.
 * @see LyFieldSetController
 */
public class LyMultiCompositionArrayFieldSetController extends LyMultiCompositionFieldSetController
	implements LyGetArrayInterface
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyMultiCompositionArrayFieldSetController.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** Array field info of this controller */
	protected LyArrayFieldInfo _array;

	/**
	 * Constructor.
	 * @param parent parent controller
	 * @param array array field handled by this controller
	 * @param relation relation field of the array
	 */
	public LyMultiCompositionArrayFieldSetController(LySetController parent,
		LyArrayFieldInfo array, LyRelationFieldInfo relation)
	{
		super(parent, relation);
		_array = array;
	}

	/**
	 * @see leon.control.form.LyMultiCompositionFieldSetController#buildAreaController(boolean)
	 */
	protected LyAreaController buildAreaController(boolean typeChanged)
	{
		if (_fieldType == null)
			_fieldType = getFieldType();

		LySetBehaviorInterface behavior = (LySetBehaviorInterface)getBehavior();
		LyAction			   action = behavior.getAction(_setController,
				(LyRelationFieldInfo)_fieldInfo, _fieldType);

		if (action == null)
		{
            LyApplication application = getApplication();
            String        actionId = LyAction.ID_SET_RELATION_TABLE;

            action = _fieldType.getActionWithId(actionId);
			if (action == null)
			{
                String          fieldTypeActionId = _fieldType.getId() + "." + actionId;
                
                action = application.getAction(fieldTypeActionId);
                if (action == null)           
                    action = new LyAction(fieldTypeActionId,
						application.getActionWithId(actionId, null), _fieldType, null);
			}
		}

		_areaController = new LyRelationTableController(getSession(), this, action,
				(LyRelationFieldInfo)_fieldInfo, _array, _fieldType, null);

		// if the classInfo is abstract, we can't create an object for it
		boolean editable = true;

		if (_fieldType.isAbstract())
		{
			editable = false;
			((LyRelationTableController)_areaController).forceActionsEnabled();
		}

		((LyFormInterface)_areaController).setEditable(editable);

		_areaController.initialize();

		return _areaController;
	} // end method buildAreaController

	/**
	 * Extract current value for the specified fieldInfo in the form.
	 * @return current value, or null if not found or field is disabled.
	 */
	public LyValue getFieldValue()
	{
		LyObjectList relations = _areaController.getContent();
		LyObjectId[] objectIds = null;
		int			 i, n = ((relations == null) ? 0 : relations.getSize());

		if (n > 0)
		{
			objectIds = new LyObjectId[n];
			for (i = 0; i < n; i++)
				objectIds[i] = relations.getObject(i).getObjectId();
		}

		return new LyArrayValue(_array, objectIds);
	}

	/**
	 * Internal method for control and propagate a value change.
	 */
	protected void valueChanged()
	{
		_setController.valueChanged(_array);
	}

	/**
	 * Impact of a value change of a fieldInfo.
	 * @param value new Value of the field.
	 */
	public void propagate(Object value)
	{
		// Generic processing : invoke behavior
		LySetBehaviorInterface behavior = _setController.getSetBehavior();
		LyValue				   fieldValue = _array.buildFieldValue(value);

		behavior.propagate(_setController, (LyFieldInfoValue)fieldValue);
	}

	/**
	 * Internal use : update the value of a fieldInfo in the form.
	 * @param newValue the new value of the field. This value may be null, a LyValue instance or the
	 *                 internal value associated to thge fieldInfo. See LyValue.getValue()
	 * @param oldValue the latest value of the field. This value may be null, a LyValue instance or
	 *                 the internal value associated to thge fieldInfo. See LyValue.getValue()
	 */
	protected void updateValue(Object newValue, Object oldValue)
	{
		LyAreaController	   areaController = getAreaController();
		LyRelationFieldInfo    relationField = (LyRelationFieldInfo)_fieldInfo;
		LyObjectList		   context;
		LyObjectList		   previousContext = areaController.getContext();
		LySetBehaviorInterface behavior = _setController.getSetBehavior();

		context = behavior.getRelationContext(_setController, relationField);
		if ((context == null) && (getObject() != null))
			context = new LySimpleObjectList(getObject());

		if (((context == null) && (previousContext != null))
				|| ((context != null) && (previousContext == null))
				|| ((context != null) && (previousContext != null)
					&& !context.equals(previousContext)))
			areaController.setContext(context);
		else
			_areaController.reset();
	}

	/**
	 * Method used to propagate the type change from other controller.
	 * @param  newValue       The new value of the field.
	 * @return The areaController expected.
	 */
	protected LyAreaController changeAreaFromTypeChange(Object newValue)
	{
		// TODO if we want to manage array of structs or relations with
		// inherited classes we must change this method.
		return _areaController;
	}

	/**
	 * Checks if the value for the fieldInfo has been modified.
	 * @return true if the value has been modified, false otherwise.
	 */
	public boolean checkIfModified()
	{
		if ((_areaController instanceof LyFormInterface)
				&& (((LyFormInterface)_areaController).startValidate()
					!= LyFormInterface.STATUS_IGNORE))
			return true;

		boolean  modified;
		LyObject object, model;
		Object   oldValue, newValue;

		object = _setController.getObject();
		model = _setController.getModel();

		// Get the value from the user form
		newValue = ((object == null) ? null : object.getValue(_array));

		// Get the default (opening) value
		oldValue = ((model == null) ? null : model.getValue(_array));

		modified = !_array.equals(oldValue, newValue);

		return modified;
	}

	/**
	 * Get the array fieldInfo associated to the controller.
	 * @return The array fieldInfo associated to the controller.
	 */
	public LyArrayFieldInfo getArray()
	{
		return _array;
	}

	/**
	 * Internal use : update the value of a fieldInfo in the form. <B>Application should use
	 * setFieldValue instead.</B> For relation fieldInfo, this method invokes updateRelation
	 * (specialization). For choice fieldInfo, this method invokes updateChoice (specialization).
	 * @param newValue  the new value of the field. This value may be null, a LyValue instance or
	 *                  the internal value associated to thge fieldInfo. See LyValue.getValue()
	 * @param oldValue  the latest value of the field. This value may be null, a LyValue instance or
	 *                  the internal value associated to thge fieldInfo. See LyValue.getValue()
	 * @param propagate Indicates whether the change is notified (propagated) to other fields or
	 *                  not. This is useless if you reset all fields for instance.
	 * @see   #setFieldValue
	 */
	public void updateFieldItem(Object newValue, Object oldValue, boolean propagate)
	{
		if (_setController.isConsult(_array))
			super.updateFieldItem(newValue, oldValue, propagate);
		else
			newValue = updateFieldItem(newValue, oldValue);

		if (propagate)
		{
			// Propagate value change if the value really changed
			if (!_array.equals(newValue, oldValue))
				propagate(newValue);
			else
				propagate = false;
		}

		// As soon as the field info is updated, reset valid state
		setState(LyFormInterface.STATUS_OK);

		if (propagate)
		{
			LyFieldItem fieldItem = getFieldItem();

			// Sets field value on field item
			fieldItem.setFieldValue(newValue);
		}

		// Update the last modification time.
		setLastModified(new Date().getTime());
	} // end method updateFieldItem

	/**
	 * Sets the value of a fieldInfo.
	 * @param value     the new value of the field. This value may be null, a LyValue instance or
	 *                  the internal value associated to the fieldInfo. See LyValue.getValue()
	 * @param propagate Indicates whether the change is notified (propagated) to other fields or
	 *                  not. This is useless if you reset all fields for instance.
	 * @see   #getFieldValue
	 * @see   leon.data.LyValue#getValue
	 */
	public void setFieldValue(Object value, boolean propagate)
	{
		LyValue fieldValue = null;

		if (value instanceof LyValue)
			fieldValue = (LyValue)value;
		else if (value != null)
			fieldValue = _array.buildFieldValue(value);

		setFieldValue(fieldValue, propagate);
	}
} // end class LyMultiCompositionArrayFieldSetController
