/**
 * LyRelationFieldConsultController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyRelationFieldConsultController.java 31611 2011-07-07 16:14:16Z nhamon $
 */

package leon.control.form;

import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * Field controller for consultation of a simple relation.
 * @see LyConsultController
 */
public class LyRelationFieldConsultController extends LyFieldConsultController
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyRelationFieldConsultController.java 31611 2011-07-07 16:14:16Z nhamon $";

	/**
	 * Constructor. The controller automatically registers on the specified session.
	 * @param parent parent controller
	 * @param fieldInfo relation field handled by this controller
	 */
	public LyRelationFieldConsultController(LyFormController parent,
		LyRelationFieldInfo fieldInfo)
	{
		super(parent, fieldInfo);
	}

	/**
	 * Invoked when an action occurs. At this level, the implementation manages hypertext navigation
	 * on file fieldInfo. Default invokes showLocation on the view manager to open a popup view.
	 * @param demand demand parameters sent by the view component.
	 * @see   LyFormController#actionDemand
	 * @see   LyController#demandPerformed
	 */
	public void actionDemand(LyActionDemand demand)
	{
		String demandId = demand.getValue();

		if ((demand.getType() == LyActionDemand.TYPE_HREF)
				&& (demandId != null))
		{
			LyObjectId oid = getObjectId(demandId);
			LyObject   object = oid.getLogicalObject();

			if (object != null)
				consult(object);
		}
		else
			super.actionDemand(demand);
	}

	/**
	 * Update the value of a fieldInfo shown in the view. The method is overriden for relation
	 * fieldInfo (updateRelation) and file fieldInfo (updateFile)
	 * @param newValue  the new value of the field.
	 * @param oldValue  former value of the field,or null if unknown. This parameter is not used
	 *                  within generic code.
	 * @see   #updateFieldItem(Object, Object, boolean)
	 */
	public void updateValue(Object newValue, Object oldValue)
	{
		LyRelationFieldInfo relationField = (LyRelationFieldInfo)_fieldInfo;
		LyFieldItem		    fieldItem = getFieldItem();
		boolean			    multiple = relationField.isMultiple();

		if (!multiple && !(fieldItem instanceof LyHRef))
			return;
		if (multiple && !(fieldItem instanceof LyHRefs))
			return;

		LyObject	  object = null;
		LyClassInfo   classInfo = null;
		LyHRef		  href = null;
		LyEnvironment environment = getEnvironment();
		LyAction	  action = null;
		boolean       addImage = (getEnvironment().getIntResource("LY_SHOW_RELATION_IMAGES") != 0);
		LyConsultController           controller = (LyConsultController)getFormController();
        LyConsultBehaviorInterface    behavior = controller.getConsultBehavior();
		
        if (!multiple)
		{
			LyObjectId objectId = (LyObjectId)newValue;

			href = (LyHRef)fieldItem;

			if ((objectId == null)
					|| (environment.testResource("LY_HIDE_UNKNOWN_OBJECTS") && (objectId.getLogicalObject() == null)))
			{
				href.setId(null);
				href.setLabel(null);
				href.setImage(null);
				href.setSensitive(false);
			}
			else
			{
				object = objectId.getLogicalObject();
				if (object == null)
					classInfo = objectId.getClassInfo();
				else
					classInfo = object.getClassInfo();

				href.setId((object == null) ? objectId.getFullString() : object.getId());
				href.setLabel((object == null) ? objectId.toString()
											   : _formController.getName(object));
				if (addImage)
					href.setImage((object == null) ? getImage(classInfo) : getImage(object));

				if (object != null)
				{
					action = classInfo.getActionWithId(LyAction.ID_CONSULT);
					if (action == null)
						action = classInfo.getDefaultAction();
				}

				if ((object == null)
						|| (action == null)
                        || !behavior.enableRelationHRef(controller, relationField, object)
						|| !_session.enableAction(action, this, new LySimpleObjectList(object), false)
						|| !environment.testResource("LY_FORM_TRANSLATE_RELATION_TO_LINK"))
					href.setSensitive(false);
				else
					href.setSensitive(true);
			}
		}
		else
		{
			LyObjectId[] objectIds = (LyObjectId[])newValue;
			LyHRefs		 hrefs = (LyHRefs)fieldItem;
			int			 i, n, pos = 0;

			objectIds = relationField.getOrderedIds(objectIds);

			// Remove objects figuring in old list and not in the new one
			n = hrefs.getHRefCount();
			for (i = n - 1; i >= 0; i--)
			{
				href = hrefs.getHRef(i);
				hrefs.removeHRef(href);
			}

			// Add remaining objects still figuring in new objects list
			n = ((objectIds == null) ? 0 : objectIds.length);
			
			for (i = 0; i < n; i++)
			{
				object = objectIds[i].getLogicalObject();

				if ((object == null) && environment.testResource("LY_HIDE_UNKNOWN_OBJECTS"))
					continue;
				
				if (object == null)
					classInfo = objectIds[i].getClassInfo();
				else
					classInfo = object.getClassInfo();

				href = new LyHRef();
				href.setId((object == null) ? objectIds[i].toString() : object.getId());
				href.setLabel((object == null) ? objectIds[i].toString()
											   : _formController.getName(object));
				if (addImage)
					href.setImage((object == null) ? getImage(classInfo) : getImage(object));

				if (object != null)
				{
					action = classInfo.getActionWithId(LyAction.ID_CONSULT);
					if (action == null)
						action = classInfo.getDefaultAction();
				}

				if ((object == null)
						|| (action == null)
                        || !behavior.enableRelationHRef(controller, relationField, object)
						|| !_session.enableAction(action, this, new LySimpleObjectList(object), false)
						|| !environment.testResource("LY_FORM_TRANSLATE_RELATION_TO_LINK"))
					href.setSensitive(false);
				else
					href.setSensitive(true);
				
				hrefs.insertHRef(href, pos);
				pos++;
			}
		}
		checkTransverseToolbar();
		// Update the last modification time of the controller.
		setLastModified(new Date().getTime());
	} // end method updateValue
	
	
	/**
	 * Get objects in the relation field to invoke actions for instance
	 * @return the objects displayed in the relation if any
	 */
	public  LyObjectList getObjects()
	{
		LyRelationFieldInfo relationField = (LyRelationFieldInfo)_fieldInfo;
		LyFieldItem		    fieldItem = getFieldItem();
		boolean			    multiple = relationField.isMultiple();

		if (!multiple && !(fieldItem instanceof LyHRef))
			return null;
		if (multiple && !(fieldItem instanceof LyHRefs))
			return null;

		LyObject	  object = null;
		LyHRef		  href = null;
		LySimpleObjectList list = null;
		LyObjectId oid;
		
		if (!multiple)
		{
			href = (LyHRef)fieldItem;

			if (href.getId() != null)
			{
				oid = getObjectId(href.getId());
				object = oid.getLogicalObject();
				if (object != null)
					list = new LySimpleObjectList(object);
			}
		}
		else
		{
			LyHRefs		 hrefs = (LyHRefs)fieldItem;
			int			 i, n;

			// Remove objects figuring in old list and not in the new one
			n = hrefs.getHRefCount();
			for (i = n - 1; i >= 0; i--)
			{
				href = hrefs.getHRef(i);
				
				if (href.getId() != null)
				{
					oid = getObjectId(href.getId());
					object = oid.getLogicalObject();
					if (object != null)
					{
						if (list == null)
							list = new LySimpleObjectList();
						list.addObject(object);
					}
				}
			}
		}
		return list;
	}
	/**
	 * Internal method to invoke on a relation a transverse action for a given classInfo.
	 * @param action    Transverse action.
	 * @param classInfo Class info on which the action is applied.
	 */
	public void invokeTransverseAction(LyAction action, LyClassInfo classInfo)
	{
		LyAreaController controller = (LyAreaController)invokeAction(action, classInfo, getObjects(), false);

		if (controller != null)
		{
			_formController.openSubAreaController(controller, true);
			controller.showView();
		}
		checkTransverseToolbar();
	}
	
	/**
	 * Get the transverse action associated to this field controller
	 * @return a list of transverse actions.
	 */
	public LyInfoDataList<LyAction, LyClassInfo> getTransverseActions()
	{
		LyRelationFieldInfo relation = (LyRelationFieldInfo)getFieldInfo();
		// Add toolbar for transverse actions if needed
		LyInfoDataList<LyAction, LyClassInfo>	actions = null;
		LyClassInfo classInfo;
		
		// Return the list of classes specified in the relation fieldInfo
		int n = relation.getClassInfoCount();
	
		for (int i = 0; i < n; i++)
		{
			classInfo = relation.getClassInfo(i);
	
	
			int na = classInfo.getActionCount();
	
			for (int j = 0; j < na; j++)
			{
				LyAction action = classInfo.getAction(j);
	
				// Action must have transverse mark
				if (!action.hasMark(LyAction.MARK_TRANSVERSE))
					continue;
				if (action.getRoot().getId().equals("_create"))
					continue; // within consult form
						
				// At last, action must validated and enabled
				if (getSession().validateAction(action, this))
				{
					if (actions == null)
						actions = new LyInfoDataList<LyAction, LyClassInfo>();
	
					actions.addInfo(action, classInfo);
				}
			}
		}
		return actions;
	}
	
	/**
	 * Check if toolbar should enable/disable some tools.
	 */
	protected void checkTransverseToolbar()
	{
		// Update transverse relation toolbar (enable / disable tools)
		LyFormField			   formField = getFormField();

		LyToolBar			   toolbar = formField.getToolBar();

		if (toolbar != null)
		{
			// Add toolbar for transverse actions if needed
			LyInfoDataList<LyAction, LyClassInfo> actions = getTransverseActions();

			if (actions == null)
				return;

			HashMap<String, LyAction>  actionsTable = new HashMap<String, LyAction>();

			// Get the related action (table for fast search)
			LyAction action;
			int		 i, na = actions.getSize();

			for (i = 0; i < na; i++)
			{
				action = actions.getInfo(i);
				actionsTable.put(action.getId(), action);
			}

			// Go through the toolbar to check if enabled
			LyObjectList	    objects = null;
			int				    nc = toolbar.getChildCount(), p;
			LyTool			    tool;
			String			    key;
			boolean			    enable;
			LySession		    session = getSession();
			LyClassInfo		    classInfo;

			for (i = 0; i < nc; i++)
			{
				if (!(toolbar.getComponent(i) instanceof LyTool))
					continue;

				tool = (LyTool)toolbar.getComponent(i);
                
                p = tool.getId().indexOf(LyController.ID_SEP);
                key = ((p > 0) ? tool.getId().substring(0, p) : tool.getId());
                action = actionsTable.get(key);
                
				// Check if still enabled
				if (action != null)
				{
					classInfo = actions.getData(action);
					enable = true;

					if (action.getMin() > 0)
					{
						// action applies on objects in the relation
						objects = getObjects();
					}
					else
						objects = null;

					enable = session.enableTransverseAction(action, classInfo, this,
							objects, false);

					tool.setEnabled(enable);
				}
			}
		}
	} // end method checkTransverseToolbar

} // end class LyRelationFieldConsultController
