/**
 * LyRelationTableController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyRelationTableController.java 41464 2012-11-08 17:02:12Z jlsaouli $
 */

package leon.control.form;

import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.control.builder.*;
import leon.control.form.builder.*;
import leon.data.*;
import leon.data.daemon.*;
import leon.info.*;
import leon.info.save.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;


/**
 * The function of this controller is to manage consultation and set of a relation of type
 * COMPOSITION. This controller is always included in a form controller. The consultation and
 * setting is done in a table.
 *
 * @see LyRelationFieldInfo
 * @see LyMultiCompositionFieldConsultController
 * @see LyMultiCompositionFieldSetController
 */
public class LyRelationTableController extends LyEditTableController
	implements LyFormInterface,
		LyFormResultInterface,
		LyControllerContextInterface
{
	/** File version */
	public static final String		  __VERSION =
		"$Id: LyRelationTableController.java 41464 2012-11-08 17:02:12Z jlsaouli $";

	/** Demand id : Move selection up. */
	public static final String		  DEMAND_MOVE_UP = "_moveUp".intern();

	/** Demand id : Move selection down. */
	public static final String		  DEMAND_MOVE_DOWN = "_moveDown".intern();

	/** Array controlled by this controller. */
	protected LyArrayFieldInfo		  _array;

	/** Relation controlled by this controller. */
	protected LyRelationFieldInfo     _relation;

	/** List of newly created objects. */
	protected LySimpleWorkSpace		  _newObjects;

	/** Boolean indicating that validation is being performed. */
	protected boolean				  _validating;

	/** Indicates if actions are enabled. */
	protected boolean				  _forceActionsEnabled;

	/** List of all objects */
	protected LySimpleObjectList	  _oldObjects;

	/** List of classInfo used by this controller */
	protected LyInfoList<LyClassInfo> _classInfos;

	/** Datasource listener */
	protected LyDataSourceListener    _listener;

	/** Controller context */
	protected LyContext				  _controllerContext;

	/** New objects listener */
	protected _LyNewObjectsListener   _newObjectsListener;

	/**
	 * Constructor.
	 *
	 * @param session   current session associated to the controller. The controller automatically
	 *                  registers on the specified session.
	 * @param parent    parent controller (or null for 'root' controller). Controller hierarchy
	 *                  indicates a child controller (for example a dialog created from a main view)
	 * @param action    related action. If an action is specified, the controller uses action
	 *                  parameters (action name, action specific XML view description ...)
	 * @param relation  relation controlled by this controller.
	 * @param array     array field controlled by this controller.
	 * @param classInfo class info of objects presented in the table.
	 * @param context   list of objects in the context. At most, this context contains one object
	 *                  (the object consulted or set).
	 */
	protected LyRelationTableController(LySession session, LyFieldController parent,
		LyAction action, LyRelationFieldInfo relation, LyArrayFieldInfo array,
		LyClassInfo classInfo, LyObjectList context)
	{
		super(session, parent, action, classInfo, context);

		_relation = relation;
		_array = array;
		_forceActionsEnabled = false;
		_newObjectsListener = null;
	}

	/**
	 * Accessor to set the last modification date of this controller. The modification date is
	 * updated when anything in the view has changed. The value should be (new Date()).getTime()
	 *
	 * @param time the new modificaion time
	 */
	public void setLastModified(long time)
	{
		if (_parent instanceof LyFieldSetController)
			_parent.getParent().setLastModified(time);
		super.setLastModified(time);
	}

	/**
	 * Free this controller.
	 *
	 * @param controller parent controller
	 */
	public void free(LyController controller)
	{
		if (_parent != controller)
			return;

		if (_newObjects != null)
		{
			_newObjects.free(_newObjectsListener);
			_newObjects = null;
		}
		_newObjectsListener = null;

		if (_workSpace != null)
		{
			try
			{
				_workSpace.free(this);
			}
			catch (Exception e)
			{
				getEnvironment().writeException(e);
			}

			_workSpace = null;
		}

		if (_classInfos != null)
		{
			LyClassInfo    classInfo;
			LyObjectSource os;
			int			   i, n = _classInfos.getSize();
			LySession	   session = getSession();

			for (i = 0; i < n; i++)
			{
				classInfo = _classInfos.getInfo(i);
				os = session.getObjectSource(classInfo);
				os.removeListener(_listener);
			}
			_classInfos = null;
			_oldObjects = null;
		}

		_relation = null;
		_array = null;

		// Rollback modifications
		if (_controllerContext != null)
		{
			_controllerContext.rollback();
			_controllerContext = null;
		}

		super.free(controller);
	} // end method free

	/**
	 * Retunrs relation controlled by this controller.
	 * @return the relation field
	 */
	public final LyRelationFieldInfo getRelation()
	{
		return _relation;
	}

	/**
	 * Retunrs array controlled by this controller.
	 * @return the array field
	 */
	public final LyArrayFieldInfo getArray()
	{
		return _array;
	}

	/**
	 * Checks if this controller manages an ordered relation or not.
	 *
	 * @return True if the controlled relation is ordered, false otherwise.
	 */
	public final boolean isOrdered()
	{
		return ((_relation instanceof LyStructFieldInfo)
				|| _relation.isOrdered());
	}

	/**
	 * Utility method (factory) to create setControllers related to objects being modified This
	 * method is overriden to mark the controller as local.
	 *
	 * @param  object current object being modified
	 * @return current setController. If no set controller is currently related, a new one is
	 *         created.
	 */
	public LyEditTableSetController setControllerFactory(LyObject object)
	{
		LyEditTableSetController result = super.setControllerFactory(object);

		if (!isConsult())
		{
			result.setRelationParent(this, null);
			result.setCompositionRelation(_relation);
	        result.setContextInternal();

            LyContext.Position position = ((LyFieldSetController)_parent).getInsertPosition();

            if (position != null)
            {
                getControllerContext().setKeepOrder(true);
                ((LyFieldSetController)_parent).getControllerContext().setKeepOrder(true);

                getControllerContext().setInsertPosition(position);
                //((LyFieldSetController)_parent).getControllerContext().setInsertPosition(position);
            }
		}
		return result;
	} // end method setControllerFactory

	/**
	 * Overriden to always return parent view controller.
	 * @return the parent view controller
	 */
	public LyViewController getViewController()
	{
		return _parent.getViewController();
	}

	/**
	 * Accessor to the current Environment instance This method returns the environment of the
	 * current Action or, if the previous one is null, the parent form environment.
	 *
	 * @return current environment instance.
	 */
	public LyEnvironment getEnvironment()
	{
		if (_environment != null)
			return _environment;

		LyEnvironment parentEnvironment = getParent().getEnvironment();

		_environment = _action.getEnvironment(parentEnvironment);
		return _environment;
	}

	/**
	 * Return the list of newly created objects.
	 *
	 * @return the list of newly created objects
	 */
	public LyObjectList getNewObjects()
	{
		return _newObjects;
	}

	/**
     * Return the list of deleted objects.
     *
     * @return the list of deleted objects
     */
    public LyObjectList getOldObjects()
    {
        return _oldObjects;
    }

	/**
	 * Internal use. Method called when an object is created in the table. This method is never used
	 * by the application : use setFieldValue instead
	 *
	 * @param object        newly created object. The new value of the relation fieldInfo.
	 * @param relationField the relation (null in this case)
	 */
	public void notifyFormResult(LyObject object, LyRelationFieldInfo relationField)
	{
	    // check if objects is in relation with the content
        if (object == null)
            return;
        LyClassInfo objectCls = object.getClassInfo();

        if ((objectCls != _classInfo) && ! objectCls.inheritsFrom(_classInfo))
            return;

		if (_newObjects == null)
		{
			_newObjects = new LySimpleWorkSpace(getSession(), object, true);
	        if (_newObjectsListener == null)
	            _newObjectsListener = new _LyNewObjectsListener();
			_newObjects.addListener(_newObjectsListener);
			if (isOrdered())
				_newObjects.setSorted(false);
		}
		else
			_newObjects.addObject(object);

		((LySimpleWorkSpace)_workSpace).addObject(object);

		if (isIdle()
				&& (_parent instanceof LyFieldSetController))
			((LyFieldSetController)_parent).valueChanged();

		if (_parent instanceof LyFieldSetController)
		{
			LySetController setController = ((LyFieldSetController)_parent).getSetController();

			setController.setActiveField(((LyFieldSetController)_parent).getFieldInfo());
		}
	} // end method notifyFormResult

	/**
	 * Internal use. Called when the parent form is starting validation.
	 * @return the status of start validate
	 * @see LyFormInterface
	 */
	public short startValidate()
	{
		_validating = true;

        enableShowMessage(true); // May raise only one error

		short result = super.startValidate();

		if ((result == LyFormInterface.STATUS_IGNORE) && (_newObjects != null))
			result = LyFormInterface.STATUS_OK;

		_validating = false;
		return result;
	}

	/**
	 * Internal use. Called when the parent form is ending validation.
	 * @return the status of end validate
	 * @see LyFormInterface
	 */
	public short endValidate()
	{
		// Do not set the flag _validating to true because
		// we could need to have a reset when an automatic validation
		// of the createController occurs.

		if (super.endValidate() == LyFormInterface.STATUS_KO)
			return LyFormInterface.STATUS_KO;

		if (_newObjects != null)
		{
			_newObjects.free(_newObjectsListener);
			_newObjects = null;
		}

		return LyFormInterface.STATUS_OK;
	}

	/**
	 * Search for the reverse relation of _relation in the specified classInfo.
	 *
	 * @param  classInfo
	 * @return The reverse relation of _relation if exists, null otherwise.
	 */
	protected LyRelationFieldInfo getReverseRelation(LyClassInfo classInfo)
	{
		// Maybe the relation is managed by a daemon ?
		LyApplication       application = getApplication();
		LyRelationDaemon    daemon = (LyRelationDaemon)application.getDaemon(_relation);
	    LyRelationFieldInfo relation;

		if (daemon != null)
		{
			if (daemon instanceof LyReverseRelationDaemon)
			{
				relation = ((LyReverseRelationDaemon)daemon).getReverseRelation();
				return relation;
			}
			return null;
		}

		LyInfoList<LyFieldInfo> fieldInfos = _classInfo.getFieldInfos();
		LyFieldInfo			    fieldInfo;
		int					    i, n = fieldInfos.getSize();

		for (i = 0; i < n; i++)
		{
			fieldInfo = fieldInfos.getInfo(i);
			if (!(fieldInfo instanceof LyRelationFieldInfo))
				continue;

			relation = (LyRelationFieldInfo)fieldInfo;
			if (relation.isMultiple())
				continue;
			if (relation.relatedWith(classInfo))
				return relation;
		}

		return null;
	} // end method getReverseRelation

	/**
	 * Internal use. Called when modifications are committed.
	 */
	public void commit()
	{
		_validating = true;

		if (_setControllers != null)
		{
            LyEditTableSetController                 set;
            Enumeration<LyEditTableSetController>    e = _setControllers.elements();

			while (e.hasMoreElements())
			{
				set = e.nextElement();
				set.commit();
			}
		}

		if (_createController != null)
		{
			if (_createController.wasChanged())
				_createController.commit();
			else
				_createController.cancel();
		}

		if (_controllerContext != null)
		{
			_controllerContext.commit();
			_controllerContext = null;
		}
		_validating = false;
	} // end method commit

	/**
	 * Internal use. Called when the parent form is cancelled.
	 */
	public void cancel()
	{
		// FFT01280: Avoid reset during cancel process
		_validating = true;

		super.cancel();

		if (_controllerContext != null)
		{
			// Rollback the context only if it has not been yet rollbacked
			// or commited
			if (!_controllerContext.isEnded())
				_controllerContext.rollback();
			_controllerContext = null;
		}
		_validating = false;
	}

	/**
	 * Check if modified.
	 * @return true if field was modified
	 */
	public boolean checkIfModified()
	{
		if ((_newObjects != null)
		        && (_newObjects.getSize() > 0))
			return true;
		if ((_oldObjects != null)
		        && (_oldObjects.getSize() > 0))
			return true;
	    if ((_controllerContext != null)
	            && !_controllerContext.isEmpty())
	        return true;

		boolean modified = false;

		if (_setControllers != null)
		{
            LyEditTableSetController                 set;
            Enumeration<LyEditTableSetController>    e = _setControllers.elements();

			while (e.hasMoreElements())
			{
				set = e.nextElement();
				modified |= set.checkIfModified();
			}
		}

		if (_createController != null)
			modified |= _createController.checkIfModified();

		return modified;
	}

	/**
	 * Extract current value for the specified fieldInfo in the form.
	 *
	 * @param  fieldInfo fieldInfo associated to the requested value
	 * @return current value, or null if not found or field is disabled.
	 */
	public LyValue getFieldValue(LyFieldInfo fieldInfo)
	{
		if (fieldInfo != _relation)
			return null;
		if (!_enabled)
			return null;

		if (_array != null)
			return new LyArrayValue(_array, getContent());
		else
			return new LyRelationValue(_relation, getContent());
	}

	/**
	 * Set the current context for the controller. If the context is unchanged, this method just
	 * call reset(). If there is a new context, this method free the current one and set the new
	 * one.
	 *
	 * @param  context Current context, or null if no context is set for the view
	 * @return True if controlled view was resetted, false otherwise.
	 */
	public boolean setContext(LyWorkSpace context)
	{
		if (_validating)
			return false;

		LyObject model = null;

		if (isCreation())
			model = ((LyFieldSetController)_parent).getModel();

		if ((_newObjects == null) && (model != null))
		{
			// We must clone the objects pointed by _relation and
			// add them to _newObjects list.
			int			 i, n;
			LyObject     object;
			LyClassInfo  classInfo;
			Object		 value;
			LyObjectList relations = null;

			classInfo = model.getClassInfo();
			if (classInfo.contains(_relation))
			{
				value = model.getValue(_relation);
				relations = _relation.getObjects(value);
			}

			n = ((relations == null) ? 0 : relations.getSize());
			for (i = 0; i < n; i++)
			{
				object = relations.getObject(i);
				object = cloneObject(object);
				if (object != null)
				{
					if (_newObjects == null)
					{
						_newObjects = new LySimpleWorkSpace(getSession(), object, true);
			            if (_newObjectsListener == null)
			                _newObjectsListener = new _LyNewObjectsListener();
						_newObjects.addListener(_newObjectsListener);
						if (isOrdered())
							_newObjects.setSorted(false);
					}
					else
						_newObjects.addObject(object);
				}
			}
		}

		if (_initialized
				&& !super.setContext(context))
			reset();

		return true;
	} // end method setContext

	/**
	 * Clone the specified object as it was created in the table.
	 * @param object the model object
	 * @return the resulting object
	 */
	protected LyObject cloneObject(LyObject object)
	{
		if (object == null)
			return null;

		LyObject			    clonedObject = null;
		LyClassInfo			    classInfo = object.getClassInfo();
		LyFieldInfo			    fieldInfo;
		LyInfoList<LyFieldInfo> fields = getFields();
		int					    i, n = fields.getSize();
		Object				    value;

		clonedObject = classInfo.createInstance();

		for (i = 0; i < n; i++)
		{
			fieldInfo = fields.getInfo(i);
			if (!isEditable(null, fieldInfo))
				continue;
			if (!fieldInfo.hasMark(LyFieldInfo.MARK_PRIVATE)
					&& !fieldInfo.hasMark(LyFieldInfo.MARK_ID))
			{
				value = object.getValue(fieldInfo);
				clonedObject.setValue(fieldInfo, value);
			}
		}

		LyContext context = getControllerContext();

		context.createObject(clonedObject);

		return clonedObject;
	} // end method cloneObject

	/**
	 * Reset all the content of the view. This method may be called several times.
	 *
	 * @param full Boolean indicating if a full reset (including reset of datas) is requiered or not
	 *             (only reset graphical area).
	 */
	public void reset(boolean full)
	{
		if (_validating)
			return;

		if (!full)
		{
			super.reset(false);
			return;
		}

		LySession    session = getSession();
		LyObjectList objects = null;
		LyObject     object = ((LyFieldController)_parent).getObject();

		if ((object != null) && !(object.isDeleted()))
		{
			LyClassInfo classInfo;
			Object	    value;

			// Don't get the classInfo for the object but from the parent FormController
			// This is necessary to handle virtual classes
			LyFormController  formController = null;
			
			if (_parent instanceof LyFieldController)
				formController = ((LyFieldController)_parent).getFormController();
			
			if (formController != null)
				classInfo = formController.getClassInfo();
			else
				classInfo = object.getClassInfo();

			if (classInfo.contains(_relation))
			{
				if (formController != null)
					value = ((LyFormBehaviorInterface)formController.getBehavior()).getValue(formController, _relation);
				else
					value = object.getValue(_relation);
				
				objects = _relation.getObjects(value);
			}
			else if ((_array != null) && classInfo.contains(_array))
			{
				if (formController != null)
					value = ((LyFormBehaviorInterface)formController.getBehavior()).getValue(formController, _array);
				else
					value = object.getValue(_array);
				
				if (value != null)
				{
					LyObjectId[] objectIds = (LyObjectId[])value;
					LyObject     obj;
					int			 i, n = objectIds.length;

					for (i = 0; i < n; i++)
					{
						if (objectIds[i] == null)
							continue;

						obj = objectIds[i].getLogicalObject();
						if (obj != null)
						{
							if (objects == null)
							{
								objects = new LySimpleObjectList();
								objects.setSorted(false);
							}
							((LySimpleObjectList)objects).addObject(obj);
						}
					}
				}
			}
		}

		if (_workSpace != null)
		{
			// Remove unused controllers
            Vector<LyObject>	oldObjects = _workSpace.getObjects();
            int             i, n = (oldObjects == null ? 0 : oldObjects.size());
            LySetController controller;

            for (i = 0; i < n; i++)
            {
                object = oldObjects.elementAt(i);
				controller = getSetController(object);
				if (((objects == null) || !objects.contains(object))
						&& (controller != null))
				{
					if (_setControllers != null)
						_setControllers.remove(object);
					controller.free(this);
				}
			}

			// Update create controller
			if (_createController != null)
			{
				if (_createController.wasChanged())
					_createController.setContext(_context);
				else
				{
					object = _createController.getObject();
					if ((object == null) || (objects == null)
							|| (!objects.contains(object)))
					{
						_createController.free(this);
						_createController = null;
					}
				}
			}

			// Free previous workspace
			_workSpace.free(this);
		}

		_workSpace = new LySimpleWorkSpace(session, true);

		if ((isOrdered()) || (_array != null))
			_workSpace.setSorted(false);
		else if (_sort != null)
			_workSpace.setSort(_sort);
		else
		{
			// FFT 1402 Check if we have a default sort ?
			LySort sort = (LySort)getParameter("_sort");

			if (sort == null)
				sort = getDefaultSort();
			else
				sort = sort.cloneSort();
			if (sort != null)
				_workSpace.setSort(sort);

			_sort = sort;
		}

		_initialWorkSpace = _workSpace;

		if (objects != null)
			_workSpace.addList(objects);

		if (_newObjects != null)
		{
			int i, n = _newObjects.getSize();

			for (i = 0; i < n; i++)
			{
				object = _newObjects.getObject(i);
				if (object.checkObject())
					((LySimpleWorkSpace)_workSpace).addObject(object);
			}

			if (objects != null)
			{
				// Update new objects
				for (i = n - 1; i >= 0; i--)
				{
					object = _newObjects.getObject(i);
					if ((objects.contains(object) && (object.getPeer() != null))
							|| !object.checkObject())
						_newObjects.removeObject(object);
				}
			}
		}

		// Listen to workspace events
		_workSpace.addListener(this);

		// Empty selections
		_selections = null;

		// Reset
		super.reset(false);
	} // end method reset

	/**
	 * Factory of the default builder for a controller. Overriden to create an instance of
	 * LyRelationTableBuilder
	 *
	 * @return the builder which creates the XML description of the view
	 */
	protected LyBuilder getDefaultBuilder()
	{
		return new LyRelationTableBuilder(this);
	}

	/**
	 * Updates the toolbar.
	 */
	protected void updateTools()
	{
		if (_component == null)
			return;

		super.updateTools();

		if (!isOrdered())
			return;

		// Update ToolBar
		LyToolBar toolBar = (LyToolBar)_component.getComponent("_toolBar");

		if (toolBar == null)
			return;

		int     no = _workSpace.getSize();
		int     ns = ((_selections == null) ? 0 : _selections.getSize());
		boolean moveUpEnabled = _editable && _enabled && (ns == 1) && (no > 1);
		boolean moveDownEnabled = moveUpEnabled;
		LyTool  tool;

		tool = (LyTool)toolBar.getComponent(DEMAND_MOVE_UP);
		if (moveUpEnabled)
		{
			LyObject object = _selections.getObject(0);

			moveUpEnabled = (_workSpace.getIndex(object) > 0);
		}
		tool.setEnabled(moveUpEnabled);

		tool = (LyTool)toolBar.getComponent(DEMAND_MOVE_DOWN);
		if (moveDownEnabled)
		{
			LyObject object = _selections.getObject(0);

			moveDownEnabled = (_workSpace.getIndex(object) < (no - 1));
		}
		tool.setEnabled(moveDownEnabled);
	} // end method updateTools

	/**
	 * Check if the given action is enabled in the context of this controller.
	 *
	 * @param action the action to check
	 * @return True if the action is enabled, false if disabled.
	 */
	protected boolean enableAction(LyAction action)
	{
		// If not editable, do not authorize create actions nor delete actions.
		LyAction create = getApplication().getAction(LyAction.ID_CREATE);
		LyAction delete = getApplication().getAction(LyAction.ID_DELETE);

		if ((_editable && _enabled)
				|| _forceActionsEnabled
				|| ((action != create)
					&& (action != delete)
					&& !action.hasTemplate(create)
					&& !action.hasTemplate(delete)))
		{
			boolean result = super.enableAction(action);

			if (result
					&& (action == _createAction))
			{
				int maxRows = getMaxObjects();
				int no = ((_workSpace == null) ? 0 : _workSpace.getSize());

				if ((maxRows > 0) && (no >= maxRows))
					result = false;
			}

			return result;
		}
		else
			return false;
	} // end method enableAction

	/**
	 * Set or create a new fast Filter.<br/>
	 * Overridden to disable fast filter on relation tables.
	 *
	 * @param fieldInfo     field info used to build the new fast filter
	 * @param x             x position where fast filter window must be displayed
	 * @param y             y position where fast filter window must be displayed
	 * @param extend        indicates if current filter must be extended with new one
	 */
	protected void buildFastFilter(LyFieldInfo fieldInfo, int x, int y, boolean extend)
	{
	}

	/**
	 * Force the actions tools to be enabled.
	 */
	protected void forceActionsEnabled()
	{
		_forceActionsEnabled = true;
	}

	/**
	 * Internal use. Invoked when an action occurs.
	 *
	 * @param demand Action event created from the view.
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			String source = demand.getSource();
			String id = source.intern();
			int    no = ((_workSpace == null) ? 0 : _workSpace.getSize());
			int    ns = ((_selections == null) ? 0 : _selections.getSize());

			if ((ns == 1) && (no > 1))
			{
				LySimpleWorkSpace content = (LySimpleWorkSpace)_workSpace;
				LyObject		  selectedObject = _selections.getObject(0);
				int				  indexSelectedObject = content.getIndex(selectedObject);

				if (id == DEMAND_MOVE_UP)
					moveObjectTo(selectedObject, indexSelectedObject - 1);
				else if (id == DEMAND_MOVE_DOWN)
					moveObjectTo(selectedObject, indexSelectedObject + 1);

			}
			// Forward key pressed in fields to parent
			if ((source.equals(LyFormController.DEMAND_VALIDATE))
					|| source.equals(LyFormController.DEMAND_CANCEL))
			{
				// Flandres/FFT00090
				if (_parent instanceof LyFieldSetController)
					((LyFieldSetController)_parent).actionDemand(demand);
				else if (_parent instanceof LyAreaController)
					((LyAreaController)_parent).actionDemand(demand);
				return;
			}
		}

		if ((demand.getType() == LyActionDemand.TYPE_DELETE)
				|| (demand.getType() == LyActionDemand.TYPE_INSERT))
		{
			if (!_editable || !_enabled)
				return; // Skip delete
		}

		// Update parent values if needed (forms)
		LyParameterValues values = demand.getParameterValues();

		if (values != null)
			values.updateViewFromValues(getParent().getComponent());

		super.actionDemand(demand);
	} // end method actionDemand

	/**
	 * Gets selected objects. Highlighted in the table.
	 * @return the selection
	 */
	public LyObjectList getSelection()
	{
		return getSelection(false);
	}

	/**
	 * Invoked when a selection on an object that no longer exists is called.
	 */
	protected void showObjectNotInViewError()
	{
		// Should never happen
	}

	/**
	 * Called when an object is added in the dynamic object list.
	 * @param event LyWorkSpace event
	 * @return true if event is handled
	 * @see LyWorkSpaceListener#objectAdded
	 */
	public boolean objectAdded(LyWorkSpaceEvent event)
	{
		if (super.objectAdded(event))
		{
			int maxRows = getMaxObjects();

			if (maxRows > 0)
			{
				int no = _workSpace.getSize();

				if (no == maxRows)
					removeRow(null, no);
			}

			if (_parent instanceof LyFieldSetController)
				((LyFieldSetController)_parent).valueChanged();
			return true;
		}
		else
			return false;
	}

	/**
	 * Called when an object is modified in the dynamic object list.
	 *
	 * @param event LyWorkSpace event
	 * @return true if event is handled
	 * @see LyWorkSpaceListener#objectAdded
	 */
	public boolean objectChanged(LyWorkSpaceEvent event)
	{
		if (_validating)
			return false;

		if (super.objectChanged(event))
		{
			// Check if an update is necessary or not
			LyValueSet  values = event.getValues();
			int		    i, n = ((values == null) ? 0 : values.getSize());
			LyValue     value;
			LyFieldInfo fieldInfo;

			for (i = 0; i < n; i++)
			{
				value = values.getFieldValue(i);
				fieldInfo = (LyFieldInfo)value.getField();
				if (fieldInfo.hasMark(LyFieldInfo.MARK_ID))
					return true;
			}

			// Force propagation if something changed on objects relation
			if (_parent instanceof LyFieldSetController)
				((LyFieldSetController)_parent).propagate(((LyFieldSetController)_parent)._currentValue);
			return true;
		}
		else
			return false;
	}

	/**
	 * Called when an object is removed from the dynamic object list.
	 *
	 * @param event LyWorkSpace event
	 * @return true if event is handled
	 * @see LyWorkSpaceListener#objectRemoved
	 */
	public boolean objectRemoved(LyWorkSpaceEvent event)
	{
		if (super.objectRemoved(event))
		{
			LyObject object = event.getObject();

			if ((_newObjects != null) && _newObjects.contains(object))
				_newObjects.removeObject(object);
			if (_oldObjects == null)
			{
				_oldObjects = new LySimpleObjectList();
				_classInfos = new LyInfoList<LyClassInfo>(false);
				_listener = new _LyDataSourceListener();
			}
			LyClassInfo    classInfo = object.getClassInfo();
			LySession	   session = getSession();
			LyObjectSource os = session.getObjectSource(classInfo);

			os.addListener(_listener, LyDataSourceEvent.TYPE_CREATE);
			_oldObjects.addObject(object);
			_classInfos.addInfo(classInfo);

			int maxRows = getMaxObjects();

			if (maxRows > 0)
			{
				int no = _workSpace.getSize();

				if (no == (maxRows - 1))
					insertRow(null, -1, false);
			}

			if (_parent instanceof LyFieldSetController)
				((LyFieldSetController)_parent).valueChanged();
			return true;
		}
		else
			return false;
	} // end method objectRemoved

	/**
	 * Called when grouped events are sent.
	 *
	 * @param  events Sorted list of events.
	 * @return True if the events are used by the listener, false otherwise.
	 */
	public boolean workSpaceEventsGrouped(ArrayList<LyWorkSpaceEvent> events)
	{
		int					  i, n = ((events == null) ? 0 : events.size());
		LyWorkSpaceEvent	  event;
		LyWorkSpaceEvent.Type type;
		boolean				  used = false;

		for (i = 0; i < n; i++)
		{
			event = events.get(i);
			type = event.getType();
			switch (type)
			{
			case ADD:
				used |= objectAdded(event);
				break;
			case CHANGE:
				used |= objectChanged(event);
				break;
			case REMOVE:
				used |= objectRemoved(event);
				break;
			default:
				break;
			}
		}
		return used;
	} // end method workSpaceEventsGrouped

	/**
	 * Invokes an action on the specified class and object list.
	 *
	 * <P>The action process is implemented on the session. This is a shortcut to
	 * getSession().invokeAction All other invokeAction method invoke this method.</P>
	 *
	 * @param action    the action which is executed.
	 * @param classInfo the target class of the action.
	 * @param objects   The list of objects on which the action applies. They must belong to the
	 *                  specified classInfo. Otherwise, use invokeAction(LyAction,LyObjectList)
	 * @param show      indicates if the resulting action must be shown
	 * @return the resulting controller
	 * @see   #invokeAction(LyAction, LyClassInfo, LyObjectList, boolean)
	 * @see   leon.app.LySession#invokeAction
	 */
	protected LyController invokeAction(LyAction action, LyClassInfo classInfo,
		LyObjectList objects, boolean show)
	{
		// We must verify if we can have the lock on the field.
		LyAction rootAction = action.getRoot();
		String   rootId = rootAction.getId();

		if (rootId != LyAction.ID_CONSULT)
		{
			LyContext context = getControllerContext();

			if (context != null)
			{
				LyObject object = ((LyFieldController)_parent).getObject();

				if (object != null)
				{
					LyClassInfo objectClassInfo;
					LyValue     value = null;

					objectClassInfo = object.getClassInfo();
					if (objectClassInfo.contains(_relation))
						value = object.getFieldValue(_relation);
					else if ((_array != null) && (objectClassInfo.contains(_array)))
						value = object.getFieldValue(_array);
					if (value != null)
					{
						LyValueSet values = new LyValueSet();

						values.addValue(value);
					}
				}
			}
		}

		// We must setting local attribute on set and delete controllers.
		LySession    session = getSession();
		LyController controller = session.invokeAction(this, action, classInfo, objects, false);

		if (controller instanceof LyAreaController)
		{
			// Retrieve first form controller having its own area stack controller,
			// otherwise retrieve top form controller.
			LyController     parentController = _parent;
			LyFormController topFormController = null;

			while (parentController != null)
			{
				parentController = parentController.getParent();

				if (parentController instanceof LyFormController)
				{
					topFormController = (LyFormController)parentController;

					if (topFormController.getAreaStackController() != null)
						break;
				}
				else if (parentController instanceof LyViewController)
					break;
			}

			if (topFormController != null)
				topFormController.openSubAreaController((LyAreaController)controller, true);
		}

		if (controller instanceof LyTransverseInterface)
			((LyTransverseInterface)controller).setRelationParent(this, _relation);

		// Compound action with transverse mark
		else if ((controller instanceof LyCompoundAreaController)
				&& (((LyCompoundAreaController)controller).getAreaController(0) instanceof LyTransverseInterface))
			((LyTransverseInterface)((LyCompoundAreaController)controller).getAreaController(0)).setRelationParent(this, _relation);

	    if ((controller instanceof LySetController) && (_parent instanceof LyFieldSetController))
        {
            LyContext.Position position = ((LyFieldSetController)_parent).getInsertPosition();

            if (position != null)
            {
                LyContext   controllerContext = ((LySetController)controller).getControllerContext();

                controllerContext.setKeepOrder(true);
                controllerContext.setInsertPosition(position);
            }
        }

        if (show)
            session.showView(controller);

		return controller;
	} // end method invokeAction

	/**
	 * Is this table in a form creation or not ?
	 * @return true if in creation mode
	 */
	protected boolean isCreation()
	{
		if (_parent instanceof LyFieldSetController)
			return ((LyFieldSetController)_parent).isCreation();
		else
			return false;
	}

	/**
	 * Gets the maximum number of objects allowed in this table.
	 *
	 * @return the maximum number of objects allowed in this table.
	 */
	protected int getMaxObjects()
	{
		if (_array != null)
			return _array.getMaxCardinality();
		else
			return _relation.getMax();
	}

	/**
	 * Gets the controller context.
	 *
	 * @return The controller context.
	 */
	public LyContext getControllerContext()
	{
		if (((_controllerContext == null) || (_controllerContext.isEnded()))
				&& (!isConsult()))
		{
			LySession		 session = getSession();
			LyContextManager contextManager = session.getContextManager();

			_controllerContext = contextManager.newContext(session,
					LyContext.Type.USER_CANCELLABLE);
			_controllerContext.setInternal();
			if (_parent instanceof LyControllerContextInterface)
			{
				LyContext context = ((LyControllerContextInterface)_parent).getControllerContext();

				context.addSon(_controllerContext);
			}
		}
		return _controllerContext;
	}

	/**
	 * Indicates if this table notifies the outline.
	 *
	 * @return A boolean which indicates if this table notifies the outline. Always false here.
	 */
	protected boolean notifyOutline()
	{
		return false;
	}

	/**
	 * Check if the fieldInfo can be edited
	 * @param fieldInfo the fieldInfo
	 * @return true if modification is valid
	 */
	public boolean canChange(LyFieldInfo fieldInfo)
	{
		// We must verify if we can have the lock on the field.
		LyContext context = getControllerContext();

		if (context != null)
		{
			LyObject object = ((LyFieldController)_parent).getObject();

			if (object != null)
			{
				LyClassInfo objectClassInfo;
				LyValue     value = null;

				objectClassInfo = object.getClassInfo();
				if (objectClassInfo.contains(_relation))
					value = object.getFieldValue(_relation);
				else if ((_array != null) && (objectClassInfo.contains(_array)))
					value = object.getFieldValue(_array);
				if (value != null)
				{
					LyValueSet values = new LyValueSet();

					values.addValue(value);
				}
			}
		}
		return true;
	} // end method canChange

	/**
	 * Controls current edition.
	 *
	 * @return The error status.
	 */
	public short control()
	{
		short status = STATUS_OK, setStatus;

		if (_setControllers != null)
		{
			// Check all controllers
		    LyEditTableSetController                 set;
			Enumeration<LyEditTableSetController>    e = _setControllers.elements();

			while (e.hasMoreElements())
			{
				set = e.nextElement();
				if (!set.isInitialized())
				    continue;

				setStatus = set.control();
				if ((setStatus != STATUS_OK)
						&& ((status == STATUS_OK) || (setStatus < status)))
					status = setStatus;
			}
		}

		if ((_createController != null) && (_createController.checkIfModified()))
		{
			setStatus = _createController.control();
			if ((setStatus != STATUS_OK)
					&& ((status == STATUS_OK) || (setStatus < status)))
				status = setStatus;
		}

		return status;
	} // end method control

	class _LyDataSourceListener
		implements LyDataSourceListener,
			LyGuiListener
	{

		/**
		 * Method for receiving an object creation.
		 *
		 * @param  event Event notifying the creation of an object.
		 * @return True indicating if the event was used, false otherwise.
		 */
		public boolean objectCreated(LyDataSourceEvent event)
		{
			if (_oldObjects == null)
				return false;
			LyObject object = (LyObject)event.getObject();

			if (!_oldObjects.contains(object))
				return false;

			_oldObjects.removeObject(object);
			if (_newObjects == null)
			{
				_newObjects = new LySimpleWorkSpace(getSession(), object, true);
	            if (_newObjectsListener == null)
	                _newObjectsListener = new _LyNewObjectsListener();
				_newObjects.addListener(_newObjectsListener);
				if (isOrdered())
					_newObjects.setSorted(false);
			}
			else
				_newObjects.addObject(object);

			((LySimpleWorkSpace)_workSpace).addObject(object);

			return true;
		}

		/**
		 * Method for receiving an object modification.
		 *
		 * @param  event Event notifying the modification of an object.
		 * @return True indicating if the event was used, false otherwise.
		 */
		public boolean objectModified(LyDataSourceEvent event)
		{
			return false;
		}

		/**
		 * Method for receiving an object deletion.
		 *
		 * @param  event Event notifying the deletion of an object.
		 * @return True indicating if the event was used, false otherwise.
		 */
		public boolean objectDeleted(LyDataSourceEvent event)
		{
			return false;
		}

		/**
		 * Check idle status of the field controller.
		 *
		 * @return boolean indicating if the idle mode is on or off.
		 */
		public boolean isIdle()
		{
			return LyRelationTableController.this.isIdle();
		}
	} // end class _LyDataSourceListener

	/**
	 * Check idle status of the field controller.
	 *
	 * @return boolean indicating if the idle mode is on or off.
	 */
	public boolean isIdle()
	{
		if (_parent instanceof LyFieldSetController)
			return super.isIdle() || _parent.isIdle();
		else
			return super.isIdle();
	}

	/**
	 * Internal method to get the classinfo related to the action.
	 *
	 * @param  action current action
	 * @return action.getClassInfo() but also takes into account a class field on the relation
	 */
	protected LyClassInfo getClassInfoForAction(LyAction action)
	{
		if (_relation.getClassField() != null)
		{
			LySetController set = null;

			if ((_parent != null) && (_parent.getParent() instanceof LySetController))
				set = (LySetController)_parent.getParent();

			if ((set != null) && (set.getBehavior() != null))
			{
				LyInfoList<LyClassInfo> list = ((LySetBehaviorInterface)set.getBehavior())
						.getRelationClasses(set, _relation);

				if (list.getSize() == 1)
					return list.getInfo(0);
			}
		}
		return action.getClassInfo();
	}

	/**
	 * Is this table in a form consult or not.
	 * @return true if in consult mode
	 */
	public boolean isConsult()
	{
		if (_parent instanceof LyFieldSetController)
			return false;
		else
			return true;
	}

	/**
	 * Check if the table has a menubar to update : this table is not associated to any menubar.
	 *
	 * @return false
	 */
	protected boolean hasMenu()
	{
		return false;
	}

	/**
	 * This method does nothing : relation table has no message fields.
	 */
	protected void updateMessages()
	{

	}

	/**
	 * This class is added just to avoid messages about workspace without listeners. XXX : Does the
	 * newObjects workspace really need to be dynamic ?
	 */
	class _LyNewObjectsListener
		implements LyWorkSpaceListener
	{
		public boolean objectAdded(LyWorkSpaceEvent event)
		{
			return false;
		}

		public boolean objectRemoved(LyWorkSpaceEvent event)
		{
			return false;
		}

		public boolean objectChanged(LyWorkSpaceEvent event)
		{
			return false;
		}
	}

	/**
	 * Move object to given index.
	 * @param object object to move
	 * @param position the new position
	 */
	protected void moveObjectTo(LyObject object, int position)
	{
		LySimpleWorkSpace content = (LySimpleWorkSpace)_workSpace;
		int				  no = content.getSize();
		int				  index = content.getIndex(object);

		if (index < 0)
			return;
		if (position < 0)
			position = 0;
		if (position >= no)
			position = no - 1;
		if (position == index)
			return;

		LyContext context = getControllerContext();
		boolean   isNew = ((_newObjects != null) && _newObjects.contains(object));

		context.setLinkUndo(true);

		content.removeObject(object);
		content.addObject(object, position);

		// When we do _workspace.removeObject, the object is also remove from _newObject list
		// So we restore it at the end.
		if (isNew)
		{
			LyObject newObject;
			int		 i, n = _newObjects.getSize(), p;

			for (i = 0; i < n; i++)
			{
				newObject = _newObjects.getObject(i);
				p = content.getIndex(newObject);
				if (p > position)
					break;
			}
			_newObjects.addObject(object, i);
		}

		selectObject(object, true);
		reset(false);
		if (_parent instanceof LyFieldSetController)
			((LyFieldSetController)_parent).valueChanged();

		context.setLinkUndo(false);
	} // end method moveObjectTo

	/**
	 * Invoked when a Drag & Drop occurs. (LyDemandListener interface) This method handles: opening
	 * and closing line, moving an interval, action
	 *
	 * @param demand drag-drop demand
	 */
	public void dragDropDemand(LyDragDropDemand demand)
	{
		if (demand.getType() == LyDragDropDemand.TYPE_DROP)
		{
			int ns = ((_selections == null) ? 0 : _selections.getSize());

			if (ns == 1)
			{
				LyObject selectedObject = _selections.getObject(0);

				moveObjectTo(selectedObject, demand.getIndex(0));
			}
		}
		else
			super.dragDropDemand(demand);
	}

	/**
	 * Show an error for the specified object. If object is null or if the resource
	 * LY_SHOWN_ERRORS_ON_OBJECTS is set to true, the error message is shown in a popup window else
	 * the specified error is added to the object.
	 *
	 * @param object The object in error
	 * @param error  The error o show
	 * @see   #showError(String)
	 */
	public void showError(LyObject object, LyError error)
	{
		if (!enableShowMessage())
			return;

		LyEnvironment environment = getEnvironment();

		if (error != null)
		{
			if (_parent instanceof LyFieldController)
			{
				String			  message = error.getMessage();
				Object[]		  args = error.getArgs();
				LyFieldController fieldController = (LyFieldController)_parent;
				LyFormController  formController = fieldController.getFormController();
				LyFieldInfo		  fieldInfo = fieldController.getFieldInfo();

				fieldController.setState(LyFormInterface.STATUS_KO);

				message = environment.getMessage(message, args);
				if (object == null)
					message = environment.getMessage("LY_ER_RELATION_TABLE_CREATE",
							new Object[] { fieldInfo.getName() })
						+ '\n' + message;
				else
					message = environment.getMessage("LY_ER_RELATION_TABLE",
							new Object[] { fieldInfo.getName(), getName(object) })
						+ '\n' + message;

				_errorTime = new Date().getTime();
				formController.showError(message);
				return;
			}
		}

		super.showError(object, error);
	} // end method showError

	/**
	 * Cancel edition of given controller.
	 * @param setController controller to cancel
	 * @param update true to update the view
	 */
	protected void cancel(LyEditTableSetController setController, boolean update)
	{
        if (update && (_parent instanceof LyFieldSetController))
            update = ((LyFieldSetController)_parent).getSetController().needUpdate();

		super.cancel(setController, update);

		if (update
				&& (_parent instanceof LyFieldController))
			((LyFieldController)_parent).setState(LyFormInterface.STATUS_OK);
	}

	/**
	 * Method for allowing fast filter.
	 *
	 * @return True if fast filter is enabled, false otherwise.
	 */
	protected boolean enableFastFilter()
	{
		return false;
	}


    /**
     * Save field input.
     */
    protected Vector<LyFormSave> save(boolean full)
    {
        Vector<LyFormSave>  formSaves = new Vector<LyFormSave>(1);
        LyFormSave          formSave;
        LyObject            object;

        if ((_setControllers != null)
            && !full)
        {
            LyEditTableSetController    setController;
            boolean                     newObject;
            int                         i, n = _workSpace.getSize();

            for (i = 0; i < n; i++)
            {
                object = _workSpace.getObject(i);
                newObject = (_newObjects != null) && _newObjects.contains(object);

                setController = getSetController(object);
                if (!setController.isInitialized())
                    setController.initialize();

                formSave = setController.save(newObject);
                if (formSave != null)
                    formSaves.add(formSave);
            }
        }
        else if (_workSpace != null)
        {
            int     i, n = _workSpace.getSize();

            for (i = 0; i < n; i++)
            {
                object = _workSpace.getObject(i);
                formSave = save(object);
                if (formSave != null)
                    formSaves.add(formSave);
            }
        }

        if (_createController != null)
        {
            formSave = _createController.save(false);
            if (formSave != null)
                formSaves.add(formSave);
        }

        if (_oldObjects != null)
        {
            int         i, n = _oldObjects.getSize();

            for (i = 0; i < n; i++)
            {
                object = _oldObjects.getObject(i);
                formSave = new LyFormSave(object);
                formSaves.add(formSave);
            }
        }

        if (formSaves.size() == 0)
            return null;

        return formSaves;
    }

    /**
     * Save new and pending object values.
     */
    protected LyFormSave save(LyObject object)
    {
        LyValueSet  newValues  = object.getNewValues();

        if (newValues == null)
        {
            newValues = object.getPendingValues();
            if (newValues == null)
                return null;
        }
        else
            newValues.addValues(object.getPendingValues());

        int         nv = newValues.getSize();

        if (nv == 0)
            return null;

        Vector<LyFieldSave>     fieldSaves = new Vector<LyFieldSave>(nv);
        LyFieldSave             fieldSave;
        int                     i, j, no;
        LyFieldInfo             fieldInfo;
        LyValue                 newFieldValue;
        Object                  oldValue;
        LyObject                subObject;
        LyFormSave              formSave;
        Vector<LyFormSave>      formSaves;
        LyObjectList            newObjects;
        LyObjectId              oldObjectId;

        for (i = 0; i < nv; i++)
        {
            newFieldValue = newValues.getFieldValue(i);
            fieldInfo = (LyFieldInfo)newFieldValue.getField();
            oldValue = (object.getPeer() == null ? null : object.getOldValue(fieldInfo));
            fieldSave = new LyFieldSave(fieldInfo, newFieldValue.getValue(), oldValue);

            if (!newFieldValue.isNull()
                && (fieldInfo instanceof LyRelationFieldInfo)
                && (((LyRelationFieldInfo)fieldInfo).getType() == LyRelationFieldInfo.Type.COMPOSITION))
            {
                if (!fieldInfo.isMultiple())
                {
                    subObject = ((LyRelationValue)newFieldValue).getObject(0);
                    formSave = save(subObject);
                    fieldSave = new LyCompositionFieldSave(fieldSave, formSave);
                }
                else
                {
                    // Save created and set objects
                    newObjects = ((LyRelationValue)newFieldValue).getObjectList();
                    no = (newObjects == null ? 0 : newObjects.getSize());
                    formSaves = new Vector<LyFormSave>(no);

                    for (j = 0; j < no; j++)
                    {
                        subObject = newObjects.getObject(j);
                        formSave = save(subObject);
                        if (formSave != null)
                            formSaves.add(formSave);
                    }

                    // Save deleted objects
                    if (oldValue != null)
                    {
                        for (j = 0, no = ((LyObjectId[])oldValue).length; j < no; j++)
                        {
                            oldObjectId = ((LyObjectId[])oldValue)[j];
                            subObject = ((LyObjectId[])oldValue)[j].getLogicalObject();
                            if ((subObject == null)
                                || (newObjects == null)
                                || !newObjects.contains(subObject))
                            {
                                formSave = new LyFormSave(oldObjectId.getClassInfo(), oldObjectId);
                                formSaves.add(formSave);
                            }
                        }
                    }

                    fieldSave = new LyMultiCompositionFieldSave(fieldSave, formSaves);
                }
            }

            fieldSaves.add(fieldSave);
        }

        formSave = new LyFormSave(object, fieldSaves);
        return formSave;
    }

    /**
     * Restore field input.
     */
    protected boolean restore(Vector<LyFormSave> formSaves)
    {
        if (formSaves == null)
            return false;

        int                         i, n = formSaves.size();
        LyFormSave                  formSave;
        LyObject                    object;
        LyEditTableSetController    editTableSetController;
        LyContext                   context = getControllerContext(), deleteContext = null;
        LySession                   session = getSession();
        LyContextManager            contextManager = session.getContextManager();
        LyCreateController          createController;
        LySetController             setController;
        LyClassInfo                 classInfo;
        LyAction                    createAction, setAction;

        for (i = 0; i < n; i++)
        {
            formSave = formSaves.elementAt(i);
            if (formSave._type == LyFormSave.Type.CREATE)
            {
                if (_createController != null)
                    // Create line
                    _createController.restore(formSave);
            }
            else
            {
                object = formSave._objectId.getLogicalObject();
                if (formSave._type == LyFormSave.Type.SET)
                {
                    if (object == null)
                    {
                        if (_createController != null)
                        {
                            // New object
                    		object = _createController.createObject();
                            _createController.restore(formSave);
                            validate(_createController);

                            n += updateFormSaves(formSaves, formSave, object);
                        }
                        else
                        {
                            classInfo = formSave._classInfo;
                            createAction = classInfo.getActionWithId(LyAction.ID_CREATE);
                            if (createAction != null)
                            {
                                createController = (LyCreateController)invokeAction(createAction, classInfo, null, false);
                                createController.getComponent(); // Force construction of component
                                createController.initialize();
                                createController.restore(formSave);
                                createController.validate(false);
                                createController.free(this);
                                object = createController.getObject();
                                if (object != null)
                                    notifyFormResult(object, null);
                            }
                        }
                    }
                    else
                    {
                        // Modification
                        editTableSetController = getSetController(object);
                        if (editTableSetController != null)
                        {
                            if (!editTableSetController.isInitialized())
                                editTableSetController.initialize();
                            editTableSetController.restore(formSave);
                            validate(editTableSetController);
                        }
                        else
                        {
                            classInfo = formSave._classInfo;
                            setAction = classInfo.getActionWithId(LyAction.ID_SET);
                            if ((setAction != null) && session.enableAction(setAction, this, new LySimpleObjectList(object), false))
                            {
                                setController = (LySetController)invokeAction(setAction, classInfo, new LySimpleObjectList(object), false);
                                if (setController != null)
                                {
                                    setController.getComponent(); // Force construction of component
                                    setController.initialize();
                                    setController.restore(formSave);
                                    setController.apply();
                                    setController.free(this);
                                }
                            }

                            if (object != null)
                                notifyFormResult(object, null);
                        }
                    }
                }
                else if (object != null)
                {
                    // Delete
                    if (deleteContext == null)
                    {
                        deleteContext = contextManager.newContext(session);
                        context.addSon(deleteContext);
                    }

                    deleteContext.deleteObject(object);
                }
            }
        }

        if (deleteContext != null)
            deleteContext.commit();
        return true;
    }

    /**
     * Update form saves with object id of created object.
     * @param formSaves List of form saves.
     * @param currentFormSave Current form save that created the object.
     * @param object New object.
     * @return Number of updates performed.
     */
    protected int updateFormSaves(Vector<LyFormSave> formSaves, LyFormSave currentFormSave, LyObject object)
    {
    	if ((formSaves == null)
    			|| (currentFormSave == null)
    			|| (object == null))
    		return 0;

    	LyObjectId		oldObjectId = currentFormSave._objectId;
    	LyObjectId		newObjectId = object.getObjectId();

    	if (newObjectId.equals(oldObjectId))
    		return 0;

    	int				i, j, k, ni, nj, nk, nu = 0;
    	LyFormSave		formSave;
    	LyFieldSave		fieldSave;
    	LyObjectId[]	objectIds, newObjectIds;
    	boolean			before = true, updated = false;

        for (i = 0, ni = formSaves.size(); i < ni; i++)
        {
            formSave = formSaves.elementAt(i);
            if (formSave == currentFormSave)
            {
            	currentFormSave._objectId = newObjectId;
            	before = true;
            }

            if (formSave._fieldSaves == null)
            	continue;

            for (j = 0, nj = formSave._fieldSaves.size(); j < nj; j++)
            {
            	fieldSave = formSave._fieldSaves.elementAt(j);
            	if ((fieldSave._newValue == null)
            			|| !(fieldSave._fieldInfo instanceof LyRelationFieldInfo))
            		continue;

            	if (fieldSave._fieldInfo.equals(fieldSave._newValue, oldObjectId))
            	{
            		fieldSave._newValue = newObjectId;
            		updated = true;
            	}
            	else if (fieldSave._fieldInfo.contains(fieldSave._newValue, oldObjectId))
            	{
            		objectIds = (LyObjectId[])fieldSave._newValue;
            		nk = objectIds.length;
            		newObjectIds = new LyObjectId[nk];

            		for (k = 0; k < nk; k++)
            			if (objectIds[k].equals(oldObjectId))
            			{
            				newObjectIds[k] = newObjectId;
            				updated = true;
            			}
            			else
            				newObjectIds[k] = objectIds[k];
            	}

            	if (before && updated)
            	{
            		formSaves.add(formSave);
            		nu++;
            	}
            	updated = false;
            }
        }

        return nu;
    }
} // end class LyRelationTableController
