/**
 * LyStructFieldSetController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyStructFieldSetController.java 33186 2011-11-17 11:02:46Z jlsaouli $
 */

package leon.control.form;

import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.view.*;
import leon.view.getinterface.*;


/**
 * Field controller for modification of a LyStructFieldInfo.
 * @see leon.control.LySetController
 */
public class LyStructFieldSetController extends LyFieldSetController
	implements LyFormResultInterface,
		LyDataSourceListener
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyStructFieldSetController.java 33186 2011-11-17 11:02:46Z jlsaouli $";

	/** The object of the struct class containing all the values. */
	protected LyObject		   _structObject;

	/** Indicate if we are in creation mode or not. */
	protected boolean		   _creation;

	/**
	 * Constructor.
	 * @param parent parent controller
	 * @param fieldInfo the struct fieldInfo handled by this controller
	 */
	public LyStructFieldSetController(LySetController parent,
		LyStructFieldInfo fieldInfo)
	{
		super(parent, fieldInfo);

		_creation = super.isCreation(); // Default mode
		initStructObject();

		// Listen to object source of struct object
		LyClassInfo    classInfo = fieldInfo.getClassInfo();
		LyObjectSource source = classInfo.getObjectSource();

        if (source != null)
            source.addListener(this, LyDataSourceEvent.TYPE_SET);
	}

	/**
	 * Free this controller.
	 */
	public void free()
	{
		// Stop listening to object source of struct object
		LyClassInfo    classInfo = ((LyStructFieldInfo)_fieldInfo).getClassInfo();
		LyObjectSource source = classInfo.getObjectSource();

        if (source != null)
            source.removeListener(this);

		super.free();
	}

	/**
	 * Initialize the struct object
	 */
	protected void initStructObject()
	{
		LyObject object = getModel();
		LyObject structObject = null;

		if ((object != null)
				&& object.getClassInfo().contains(_fieldInfo))
		{
			LyObjectId objectId = (LyObjectId)object.getValue(_fieldInfo);

			if (objectId != null)
			{
				structObject = objectId.getLogicalObject();
				setStructObject(structObject);
			}
		}
	}

	/**
	 * Set the struct object
	 * @param structObject the object to edit
	 */
	protected void setStructObject(LyObject structObject)
	{
		if (_structObject != null)
		{
			if (structObject == null)
			{
				// Delete struct object
				LyContext context = getControllerContext();

				context.deleteObject(_structObject);
				_structObject = null;
			}

			return;
		}

		if (isCreation())
		{
			// We must clone the struct object.
			LyValueSet  values = structObject.getValues();
			LyValueSet  newValues = new LyValueSet();
			LyValue     value;
			LyFieldInfo fieldInfo;
			int		    i, n = values.getSize();

			for (i = 0; i < n; i++)
			{
				value = values.getFieldValue(i);
				fieldInfo = (LyFieldInfo)value.getField();
				if (fieldInfo.hasMark(LyFieldInfo.MARK_ID))
					continue;
				newValues.addValue(fieldInfo, value.getValue());
			}

			buildStructObject(newValues);
		}
		else
			_structObject = structObject;
	} // end method setStructObject

	/**
	 * Internal method to invoke on a relation a transverse action for a given classInfo.
	 * @param action    Transverse action.
	 * @param classInfo Class info on which the action is applied.
	 */
	public void invokeTransverseAction(LyAction action, LyClassInfo classInfo)
	{
		LySimpleObjectList list = null;
		LyAction		   root = action.getRoot();
		String			   rootActionId = root.getId();

		if ((rootActionId == LyAction.ID_CREATE)
				&& (_structObject != null))
		{
			//      Switch action on set
			LyClassInfo structClassInfo = _structObject.getClassInfo();
			LyAction    structAction = structClassInfo.getActionWithId(LyAction.ID_SET);

			if (structAction != null)
				action = structAction;
		}

		if (_structObject != null)
			list = new LySimpleObjectList(_structObject);

		LyRelationFieldInfo relation = (LyRelationFieldInfo)_fieldInfo;
		LyAreaController    controller = (LyAreaController)invokeAction(action, classInfo, list,
				false);

		if (controller == null)
			return;

		if (controller instanceof LyTransverseInterface)
			((LyTransverseInterface)controller).setRelationParent(this, relation);

		if (relation.ignoresContext())
		{
			if (controller.getContext() != null)
				controller.setContext(null);

			controller.showView();
			return;
		}

		controller.showView();

		// XXX we can change title parameters after the initArea
		// Set the good image and the good title
		LyArea     area = (LyArea)controller.getComponent();
		LyTitleBar titleBar = area.getTitleBar();

		if (titleBar != null)
		{
			if (getEnvironment().testResource("LY_TITLEBAR_LEFT_IMAGE"))
			{
				// Set the good image
				if (rootActionId == LyAction.ID_CREATE)
					titleBar.setImage("LY_ACTION_CREATE");
				else if (rootActionId == LyAction.ID_SET)
					titleBar.setImage("LY_ACTION_SET");
			}
	
			// Set the good title
			String title = _fieldInfo.getName();

			titleBar.setLeftLabel(title);
			titleBar.setRightLabel(null);
		}
	} // end method invokeTransverseAction

	/**
	 * Internal use : update the value of a fieldInfo in the form.
	 * @param newValue the new value of the field. This value may be null, a LyValue instance or the
	 *                 internal value associated to thge fieldInfo. See LyValue.getValue()
	 * @param oldValue the latest value of the field. This value may be null, a LyValue instance or
	 *                 the internal value associated to thge fieldInfo. See LyValue.getValue()
	 */
	protected void updateValue(Object newValue, Object oldValue)
	{	       
		if (_structObject == null)
		{
			LyObjectId objectId = (LyObjectId)newValue;

			if (objectId != null)
				_structObject = objectId.getLogicalObject();
		}
		else if (newValue instanceof LyObjectId)
		{
			LyObjectId objectId = (LyObjectId)newValue;
			// The id we set must be the id of the struct being edited
			if (!_structObject.getObjectId().equals(objectId))
				_structObject = objectId.getLogicalObject();
		}

		LyObject    struct = ((newValue == null) ? null : _structObject);

		// Here there is no editable format.
		LyUnit	    unit = getUnit();
		LyFormat    format = ((unit != null) ? unit.getFormat() : null);
		LyFieldItem fieldItem = getFieldItem();
		String	    label;

		if (format == null)
		{
			// Search for a value label
			LyInfoLabel valueLabel = getInfoLabel(_fieldInfo, LyInfoLabel.Type.VALUE);
			LyObject    object = (isCreation() ? null : getObject());

			if (object != null)
			{
				LyClassInfo classInfo = object.getClassInfo();

				if (!classInfo.contains(_fieldInfo))
					object = null;
			}

			if ((valueLabel != null) && (object != null))
				label = getLabel(object, _fieldInfo, newValue);
			else if ((valueLabel != null) && (struct != null))
			{
				// We haven't any source object (case of creation)
				// Can we compute label even so ?
				LyUnit	    pathUnit;
				LyPath	    path;
				LyFieldInfo target;
				LyClassInfo structClass = struct.getClassInfo();
				int		    i, n = valueLabel.getPartsCount();
				String[]    args = new String[n];
				Object	    value;

				for (i = 0; i < n; i++)
				{
					path = valueLabel.getPath(i);
					target = path.getTargetFieldInfo();
					if (structClass.contains(target))
					{
						pathUnit = valueLabel.getUnit(i);
						value = struct.getValue(target);
						value = target.convertFromFieldUnit(value, pathUnit);
						args[i] = target.toName(value);
					}
					else
						break;
				}

				if (i == n)
				{
					String		  str = valueLabel.getValue();
					LyEnvironment env = getEnvironment();

					label = env.getString(str, args);
				}
				else
					label = _fieldInfo.toString(newValue);
			}
			else
				label = _fieldInfo.toString(newValue);
		}
		else
			label = format.toString(_fieldInfo, newValue, unit);

		fieldItem.setLabel(label);
		fieldItem.setFieldValue(newValue);
	} // end method updateValue

	/**
	 * Internal method for propagating a unit change. This methods is called when the field unit has
	 * been changed by the user.
	 * <P>The LySetController additionaly impact the view using the setValid method.</P>
	 * @param newUnitId The new fieldinfo unit.
	 * @param oldUnitId The old fieldinfo unit.
	 */
	public void unitChanged(String newUnitId, String oldUnitId)
	{
		super.unitChanged(newUnitId, oldUnitId);

		LyUnit	    newUnit = getUnit();
		LyFieldItem fieldItem = getFieldItem();

		if (fieldItem instanceof LyText)
		{
			LyText  text = (LyText)fieldItem;
			boolean editable = newUnit.hasEditionFormat();

			text.setEditable(editable);
		}
	}

	/**
	 * Extract current value for the specified fieldInfo in the form.
	 * @param  unit      Unit of the value.
	 * @return current value, or null if not found or field is disabled.
	 */
	public LyValue getFieldValue(LyUnit unit)
	{
		if (_structObject == null)
		{
			LyObject object = getObject();

			if (object != null)
			{
				LyClassInfo classInfo = object.getClassInfo();

				if (classInfo.contains(_fieldInfo))
				{
					LyValue value = object.getFieldValue(_fieldInfo);

					if ((value != null) && (!value.isNull()))
					{
						LyObjectId objectId = (LyObjectId)value.getValue();

						_structObject = objectId.getLogicalObject();
					}
				}
			}
		}

		LyFieldItem fieldItem = getFieldItem();
		String	    strValue = null;
		LyObjectId  objectId;

		if (unit != null)
		{
			LyFormat format = unit.getFormat();

			if (format != null)
			{
				LyValueSet values = null;

				strValue = ((LyGetLabelInterface)fieldItem).getLabel();
				if (strValue != null)
					values = (LyValueSet)format.parse(_fieldInfo, strValue, unit);
				objectId = buildStructObject(values);

				return _fieldInfo.buildFieldValue(objectId);
			}
		}

		if (_structObject != null)
		{
			objectId = _structObject.getObjectId();
			return _fieldInfo.buildFieldValue(objectId);
		}
		else
			return null;
	} // end method getFieldValue

	/**
	 * Build the struct object with the given values
	 * @param values a value set
	 * @return the object id of the built object
	 */
	protected LyObjectId buildStructObject(LyValueSet values)
	{
		if (values == null)
			return null;

		if (_fieldType == null)
			_fieldType = getFieldType();

		LyContext context = getControllerContext();

		if (_structObject == null)
		{
			_structObject = _fieldType.createInstance();
			_structObject.setValues(values);
			context.setIgnoreUndo(true);
			context.createObject(_structObject);
			context.setIgnoreUndo(false);
		}
		else
		{
			// Set only modified values.
			int		    i, n;
			LyValue     fieldValue;
			LyFieldInfo fieldInfo;
			Object	    oldValue, newValue;
			LyValueSet  newValues = null;

			n = values.getSize();
			for (i = 0; i < n; i++)
			{
				fieldValue = values.getFieldValue(i);
				fieldInfo = (LyFieldInfo)fieldValue.getField();
				oldValue = _structObject.getValue(fieldInfo);
				newValue = fieldValue.getValue();
				if (!fieldInfo.equals(oldValue, newValue))
				{
					if (newValues == null)
						newValues = new LyValueSet();
					newValues.addValue(fieldInfo, newValue);
				}
			}
			if (newValues != null)
			{
				if (_parentFieldController != null)
					context.setLinkUndo(true);
				context.setValues(_structObject, newValues);
				if (_parentFieldController != null)
					context.setLinkUndo(false);
			}
		}

		return ((_structObject == null) ? null : _structObject.getObjectId());
	} // end method buildStructObject

	/**
	 * Internal use. When an object is created in a popup creation form for a relation, the child
	 * controller invokes this method to set the newly created object as the current value of the
	 * relation (if a new object is created we automatically select this value). This method is
	 * never used by the application : use setFieldValue instead
	 * @param object        newly created object. The new value of the relation fieldInfo.
	 * @param relation      the relation
	 */
	public void notifyFormResult(LyObject object, LyRelationFieldInfo relation)
	{
		LyFieldItem fieldItem = getFieldItem();

		if (_structObject == null)
			_structObject = object;

		if (fieldItem instanceof LyText)
		{
			LyRelationValue value = new LyRelationValue((LyStructFieldInfo)_fieldInfo, _structObject);

			updateFieldItem(value.getValue(), null, true);
		}

		if (_parentFieldController instanceof LyArrayTableStructFieldSetController)
			((LyArrayTableStructFieldSetController)_parentFieldController).notifyFormResult(this);
		_setController.setActiveField(_fieldInfo);
	} // end method notifyFormResult

	/**
	 * Gets the classInfo of the object pointed by the relation in the current form.
	 * @return The classInfo of the object pointed by the relation in the current form.
	 */
	public LyClassInfo getFieldType()
	{
		LyClassInfo		    classInfo = getClassInfo();
		LyRelationFieldInfo relation = (LyRelationFieldInfo)_fieldInfo;
		LyClassInfo		    fieldType = null;
		LyObject		    object = getObject();

		if ((object != null)
				&& classInfo.contains(relation))
		{
			LyRelationValue relationValue = object.getRelationValue(relation);

			if ((relationValue != null) && (relation != null))
			{
				LyObjectId  objectId = (LyObjectId)relationValue.getValue();
				LyClassInfo newClassInfo;

				if (objectId != null)
				{
					newClassInfo = objectId.getClassInfo();
					fieldType = newClassInfo;
				}
			}
		}

		if (fieldType == null)
		{
			LyInfoList<LyClassInfo> fieldTypes = relation.getClassInfos();
			int		   index = 0, nClassInfos = fieldTypes.getSize();

			fieldType = fieldTypes.getInfo(index++);
			while ((fieldType.isAbstract()) && (index < nClassInfos))
				fieldType = fieldTypes.getInfo(index++);
		}

		return fieldType;
	} // end method getFieldType

	/**
	 * Indicate if thie field is in creation or modification mode of the struct object (use by array
	 * of structs).
	 * @return creation mode if true
	 */
	public boolean isCreation()
	{
		return _creation;
	}

	/**
	 * Indicate if thie field is in creation or modification mode of the struct object (use by array
	 * of structs).
	 * @param status new creation status
	 */
	public void setCreation(boolean status)
	{
		_creation = status;
	}

	/**
	 * Sets the field content from the specified value Any value modified by the user is the form is
	 * lost. Default implementation just call updateFieldItem but this may be overriden in
	 * subclasses.
	 * @param value          new value to set in the field.
	 * @param freeOldValue   indicates if the old value should be freed (deleted) when it is an
	 *                       object
	 * @param createNewValue indicates if a new empty value should created when object is null and
	 *                       when the value is an object
	 */
	public void resetFromValue(Object value, boolean freeOldValue, boolean createNewValue)
	{
		if (value == null)
		{
			_structObject = null;
			initStructObject();
		}
		else if (value instanceof LyObjectId)
		{
			LyObjectId oldValue = ((_structObject == null) ? null : _structObject.getObjectId());

			if (freeOldValue)
				_structObject = null;

			LyObject newObj = ((LyObjectId)value).getLogicalObject();

			setStructObject(newObj);
			updateFieldItem(value, oldValue);
		}

		super.resetFromValue(value, freeOldValue, createNewValue);
	}

	/**
	 * Method called when editions is cancelled.
	 */
	public void cancel()
	{
		_structObject = null;
	}

	/**
	 * Method for receiving an object creation.
	 * @param  event Event notifying the creation of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectCreated(LyDataSourceEvent event)
	{
		return false;
	}

	/**
	 * Method for receiving an object modification. Implemented to update value on object set
	 * notification.
	 * @param  event Event notifying the modification of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectModified(LyDataSourceEvent event)
	{
		LyObject object = (LyObject)event.getObject();

		if (object != _structObject)
			return false;

		LyFieldItem fieldItem = getFieldItem();

		if (fieldItem instanceof LyText)
		{
			LyRelationValue value = new LyRelationValue((LyStructFieldInfo)_fieldInfo,
					_structObject);

			// Do not consider operations for undo/redo when we receive
			// an update event.
			LyContext	    context = getControllerContext();

			context.setLinkUndo(true);
			updateFieldItem(value.getValue(), null, true);
			context.setLinkUndo(false);
		}
		return true;
	}

	/**
	 * Method for receiving an object deletion.
	 * @param  event Event notifying the deletion of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectDeleted(LyDataSourceEvent event)
	{
		return false;
	}

	/**
	 * Get the associated object
	 * @param create true to force creation of object
	 * @return the object
	 */
	public LyObject getObject(boolean create)
	{
		if ((_structObject == null) && create)
			buildStructObject(new LyValueSet());
		return _structObject;
	}
} // end class LyStructFieldSetController
