/**
 * LyTextFieldSetController.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyTextFieldSetController.java 35196 2012-04-11 07:11:24Z jlsaouli $
 */

package leon.control.form;

import java.lang.reflect.*;
import java.util.*;

import leon.app.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;



/**
 * Default text field controller for a modification.
 * @see LySetController
 */
public class LyTextFieldSetController extends LyFieldSetController
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyTextFieldSetController.java 35196 2012-04-11 07:11:24Z jlsaouli $";

	/** Id of the transverse tool for multilinguism. */
	public static final String     		LANG_TOOL = "_langTool";
	
	protected String					_currentLang;
	
	protected HashMap<String, String>	_langValues;
	
	
	/**
	 * Constructor.
	 * @param parent the parent set controller
	 * @param fieldInfo the fieldinfo handled by this controller
	 */
	public LyTextFieldSetController(LySetController parent, LyFieldInfo fieldInfo)
	{
		super(parent, fieldInfo);
		
		if (((LyTextFieldInfo)_fieldInfo).isMultilingual())
		{
			_currentLang = getSession().getEnvironment().getLanguage();
			_langValues = new HashMap<String, String>();
		}
	}
	
	/**
	 * Invoked when an action occurs.
	 * At this level, the implementation manages multilingual inputs.
	 */
	public void actionDemand(LyActionDemand demand)
	{
		if (demand.getType() == LyActionDemand.TYPE_ACTION)
		{
			String   sourceId = demand.getSource().intern();
			String[] parameters = LyString.stringToList(sourceId, ID_SEP);
			
			if ((parameters.length > 1)
					&& parameters[0].equals(LANG_TOOL))
			{
				changeLang(parameters[1]);
				return;
			}
		}
		
		super.actionDemand(demand);
	}
	
	/**
	 * Change lang to the given one.
	 */
	protected void changeLang(String lang)
	{
		String	value = (String)getValue();
		
		_langValues.put(_currentLang, value);
		
		// Update field and associated tool
		LyFormField	formField = getFormField();
		LyToolBar	toolBar = formField.getToolBar();
		
		if (toolBar != null)
		{
			LyTool	tool = (LyTool)toolBar.getComponent(LANG_TOOL);
			
			if (tool != null)
				tool.setLabel(lang.toUpperCase());
		}
		
		String	newValue;
		
		if (_langValues.containsKey(lang))
			newValue = _langValues.get(lang);
		else
		{
			String	defaultLang = getSession().getEnvironment().getLanguage();
			
			newValue = _langValues.get(defaultLang);
			if (newValue != null)
				newValue = translate(newValue, defaultLang, lang);
		}
		
		_currentLang = lang;
		updateValue(newValue, value);
	}
	
	/**
	 * Translate a string value from a lang to another.
	 */
	protected String translate(String value, String fromLang, String toLang)
	{
		if (value == null)
			return null;
		
		String			newValue = fromLang + ": " + value;
		LyEnvironment	environment = _parent.getEnvironment();
		
		try
		{
			String		referrer = environment.getResource("LY_GOOGLE_HTTP_REFERRER");
			String		key = environment.getResource("LY_GOOGLE_API_KEY");
			
			if ((referrer != null)
				&& (key != null))
			{
				LyApplication	application = _parent.getApplication();
				Class<?>		googleAPIClass = application.loadClass("com.google.api.GoogleAPI");
				
				// Set the HTTP referrer to the website address.
				Method			method;
				
				method = googleAPIClass.getMethod("setHttpReferrer", new Class[] { String.class });
				method.invoke(null, referrer);
	
			    // Set the Google Translate API key
			    // See: http://code.google.com/apis/language/translate/v2/getting_started.html
				method = googleAPIClass.getMethod("setKey", new Class[] { String.class });
				method.invoke(null, key);
	
				Class<?>		languageClass = application.loadClass("com.google.api.translate.Language");
				Object			fromLanguage, toLanguage;
				
				method = languageClass.getMethod("fromString", new Class[] { String.class });
				fromLanguage = method.invoke(null, fromLang);
				toLanguage = method.invoke(null, toLang);
				
				Class<?>	translateClass = application.loadClass("com.google.api.translate.Translate");
				Field		translatorField = translateClass.getDeclaredField("DEFAULT");
				Object		defaultTranslator = translatorField.get(null);

				method = translateClass.getMethod("execute",
					new Class[] { String.class, languageClass, languageClass });
				newValue = (String)method.invoke(defaultTranslator, value, fromLanguage, toLanguage);
			}
		}
		catch (Exception e)
		{
			environment.writeException(e);
		}

		return newValue;
	}
	
	/**
	 * Extract final value for the specified fieldInfo in the form.
	 * @return Final value, or null if not found or field is disabled.
	 */
	public LyValue getFinalValue()
	{
		if (!((LyTextFieldInfo)_fieldInfo).isMultilingual())
			return super.getFinalValue();
		
		String			value = (String)getValue();
		
		_langValues.put(_currentLang, value);
		return _fieldInfo.buildFieldValue(_langValues);
	}
	
	/**
	 * Internal use : update the value of a fieldInfo in the form.
	 * @param newValue the new value of the field. This value may be null, a LyValue instance or the
	 *                 internal value associated to the fieldInfo. See LyValue.getValue()
	 * @param oldValue the latest value of the field. This value may be null, a LyValue instance or
	 *                 the internal value associated to the fieldInfo. See LyValue.getValue()
	 */
	protected void updateValue(Object newValue, Object oldValue)
	{
		if (!(newValue instanceof String)
			|| !((LyTextFieldInfo)_fieldInfo).isMultilingual())
		{
			super.updateValue(newValue, oldValue);
			return;
		}
		
		LyTextFieldInfo	textFieldInfo = (LyTextFieldInfo)_fieldInfo;
		LyApplication	application = getApplication();
		String[][]		languages = application.getSupportedLanguages();
		int				i, nl = (languages == null ? 0 : languages.length);
		String			langValues = (String)newValue;
		String			langValue;
		
		for (i = 0; i < nl; i++)
		{
			langValue = textFieldInfo.getLangValue(languages[i][0], langValues);
			if (langValue != newValue)
			{
				_langValues.put(languages[i][0], langValue);
				if (languages[i][0].equals(_currentLang))
					newValue = langValue;
			}
		}
		
		super.updateValue(newValue, oldValue);
	}
} // end class LyTextFieldSetController
