/**
 * LyFieldConsultBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyFieldConsultBuilder.java 35981 2012-06-21 07:11:06Z jlsaouli $
 */

package leon.control.form.builder;

import leon.app.*;
import leon.control.*;
import leon.control.form.*;
import leon.info.*;
import leon.info.misc.*;
import leon.view.*;


/**
 * This builder construct a consult form field. The associated controller is a field consult
 * controller
 *
 * @see LyFieldConsultController
 */
public class LyFieldConsultBuilder extends LyFormFieldBuilder
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyFieldConsultBuilder.java 35981 2012-06-21 07:11:06Z jlsaouli $";

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyFieldConsultBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Build a text component for the given text field.
	 *
	 * @return the text component
	 */
	public LyFieldItem buildText()
	{
		LyFieldController fieldController = (LyFieldController)getParent();
		LyTextFieldInfo   textField = (LyTextFieldInfo)fieldController.getFieldInfo();

		String			  labelStr = textField.toName(null);
		boolean			  multiple = textField.isMultiple();
		boolean			  multilingual = textField.isMultilingual();
		boolean			  rich = (textField.getType() == LyTextFieldInfo.Type.RICH);

		if (multiple || rich || multilingual)
		{
			int ncols = textField.getNCols();
			int nrows = textField.getNRows();

			if (nrows <= 0)
			{
				if (multiple)
					nrows = 5;
				else
					nrows = 1;
			}

			LyText text = new LyText();

			text.setLabel(labelStr);
			text.setNCols(ncols);
			text.setNRows(nrows);

			if (rich)
				text.setType(LyText.Type.RICH);
			
			return text;
		}
		else
		{
			LyLabel label = new LyLabel();

	        if (labelStr != null)
	        {
	            LyFieldBean bean = textField.getFieldBean(LyAction.ID_CONSULT);
	            
	            if ((bean == null)
	                    && !textField.hasMark(LyFieldInfo.MARK_TRANSLATE))
	                label.setLabel("$$" + labelStr);
	            else
	                label.setLabel(labelStr);
	        }
			return label;
		}
	} // end method buildText

	/**
	 * Build a date component for the given date field.
	 *
	 * @return the date component component
	 */
	public LyFieldItem buildDate()
	{
		LyFieldController fieldController = (LyFieldController)getParent();
		LyDateFieldInfo   dateField = (LyDateFieldInfo)fieldController.getFieldInfo();

		String			  labelStr = dateField.toName(null);
		LyLabel			  label = new LyLabel();

		label.setLabel(labelStr);
		return label;
	}

	/**
	 * Build a number component for the given number field.
	 *
	 * @return the number component component
	 */
	public LyFieldItem buildNumber()
	{
		LyFieldController fieldController = (LyFieldController)getParent();
		LyNumberFieldInfo numberField = (LyNumberFieldInfo)fieldController.getFieldInfo();

		String			  labelStr = numberField.toName(null);
		LyLabel			  label = new LyLabel();

		label.setLabel(labelStr);
		return label;
	}

	/**
	 * Build a relation component for the given relation field.
	 *
	 * @return the relation component component
	 */
	public LyFieldItem buildStruct()
	{
		LyFieldController fieldController = (LyFieldController)getParent();
		LyStructFieldInfo structField = (LyStructFieldInfo)fieldController.getFieldInfo();

		String			  labelStr = structField.toName(null);

		LyLabel			  label = new LyLabel();

		label.setLabel(labelStr);
		return label;
	}

	/**
	 * Build a relation component for the given relation field.
	 *
	 * @return the relation component component
	 */
	public LyFieldItem buildRelation()
	{
		LyFieldController   fieldController = (LyFieldController)getParent();
		LyRelationFieldInfo relationField = (LyRelationFieldInfo)fieldController.getFieldInfo();
		boolean			    multiple = relationField.isMultiple();

		if (!multiple)
		{
			LyHRef href = new LyHRef();

			return href;
		}
		else
		{
			LyHRefs hrefs = new LyHRefs();
			int     nrows = relationField.getNRows();

			if (nrows > 0)
				hrefs.setNRows(nrows);

			return hrefs;
		}
	} // end method buildRelation

	/**
	 * Build a choice component for the given choice field.
	 *
	 * @return the choice component component
	 */
	public LyFieldItem buildChoice()
	{
		LyFieldController fieldController = (LyFieldController)getParent();
		LyChoiceFieldInfo choiceField = (LyChoiceFieldInfo)fieldController.getFieldInfo();
		String			  labelStr = choiceField.toName(null);

		if (choiceField.isMultiple())
		{
			LyText text = new LyText();
			int    no = choiceField.getChoiceOptionCount();
			int    nc = choiceField.getNCols();
			int    nRows = no, nCols = 0;

			// Try to guess what is the best height
			if (nc > 0)
				nRows = no / nc;
			// Limit height between 2 and 5 lines
			if (nRows <= 2)
			{
				if (no > 2)
					nRows = 3; // need to display Swing scrollbar. FFT00984
				else
					nRows = 2;
			}
			if (nRows > 5)
				nRows = 5;

			// Fix nCols to maximum label width
			LyChoiceOption option;
			short		   i;
			int			   l;
			String		   name;

			for (i = 0; i < no; i++)
			{
				option = choiceField.getChoiceOption(i);
				name = option.getName();
				l = name.length();
				if (l > nCols)
					nCols = l;
			}

			text.setNRows(nRows);
			text.setNCols(nCols + 5);

			if (labelStr != null)
				labelStr = labelStr.replace(LyApplication.VISIBLE_FIELD_SEP, '\n');
			text.setLabel(labelStr);

			return text;
		}
		else
		{
			LyLabel label = new LyLabel();

			label.setLabel(labelStr);
			return label;
		}
	} // end method buildChoice

	/**
	 * Build a file component for the given file field.
	 *
	 * @return the file component component
	 */
	public LyFieldItem buildFile()
	{
		LyFieldController fieldController = (LyFieldController)getParent();
		LyFileFieldInfo   fileField = (LyFileFieldInfo)fieldController.getFieldInfo();

		String			  directory = fileField.getDirectory();
		String			  label = fileField.toName(null);
		String			  mimeTypes = fileField.getMimeTypes();
		String			  id;
		boolean			  sensitive = true;

		if (directory == null)
			id = "file://" + label;
		else
		{
			String directoryLc = directory.toLowerCase();

			if (directoryLc.startsWith("http://")
			        || directoryLc.startsWith("https://")
					|| directoryLc.startsWith("file://")
					|| directoryLc.startsWith("ftp://"))
				id = directory + label;
			else
				id = "file://" + directory + label;
		}

		// !!! : Be careful : content is ALSO updated in the consult controller
		if (mimeTypes != null)
		{
			// XX May use various output depending on mime type : textarea, image...
			mimeTypes = mimeTypes.toLowerCase();

			if (mimeTypes.startsWith("image"))
			{
				LyIcon icon = new LyIcon(id, label);
				int    ncols = fileField.getNCols();
				int    nrows = fileField.getNRows();

				if (label != null)
					icon.setImage(id);

				icon.setWidth(ncols);
				icon.setHeight(nrows);
				icon.setBorder(LyIcon.Border.SINGLE);
				return icon;
			}
			else if (mimeTypes.equals("dir"))
				sensitive = false;
		}

		LyHRef href = new LyHRef();

		href.setId(id);
		href.setLabel(label);
		href.setSensitive(sensitive);
		return href;
	} // end method buildFile
	
	/**
	 * Build a toolbar for a given fieldInfo. The built toolbar is added to the form field
	 * corresponding to the given fieldInfo
	 *
	 * @return the toolbar for the given fieldInfo
	 */
	public LyToolBar buildToolBar()
	{
		LyFieldConsultController fieldController = (LyFieldConsultController)_parent;
		LyFieldInfo		  fieldInfo = fieldController.getFieldInfo();
		
		if ((fieldInfo instanceof LyRelationFieldInfo)
				&& (!((LyRelationFieldInfo)fieldInfo).isToolbarDisabled()))
		{
			LyInfoDataList<LyAction, LyClassInfo>	actions = fieldController.getTransverseActions();
			
			if ((actions != null) && (actions.getSize() > 0)) 
			{
				LyToolBar toolBar = buildToolBar(fieldInfo, actions);
			
				return toolBar;
			}
		}
		
		LyToolBar toolBar = super.buildToolBar();
		
		return completeToolBar(toolBar, fieldInfo);
	}

	/**
	 * Complete toolBar for given field info.
	 */
	protected LyToolBar completeToolBar(LyToolBar toolBar, LyFieldInfo fieldInfo)
	{		
		if ((fieldInfo instanceof LyTextFieldInfo)
			&& ((LyTextFieldInfo)fieldInfo).isMultilingual())
		{
			LyFieldController	fieldController = (LyFieldController)_parent;
			LyApplication		application = fieldController.getApplication();
			String[][]			languages = application.getSupportedLanguages();
			int					nl = (languages == null ? 0 : languages.length);
			
			if (nl > 1)
			{
				LySession		session = _parent.getSession();
				String			defaultLang = session.getEnvironment().getLanguage();
				int				i;
				LyTool			langTool;
				LyToolBar		subToolBar = new LyToolBar();
				LyArea			area = new LyArea();
	
				subToolBar.setController(_parent.getId());
				area.setToolBar(subToolBar);
				subToolBar.setFill(LyToolBar.Fill.HORIZONTAL);

				langTool = new LyTool(LyTextFieldSetController.LANG_TOOL, defaultLang.toUpperCase());
				langTool.setComponent(area);
				toolBar = addTool(toolBar, langTool);
				toolBar.setFill(LyToolBar.Fill.HORIZONTAL);
	
				for (i = 0; i < nl; i++)
				{
					langTool = new LyTool(LyTextFieldSetController.LANG_TOOL + LyController.ID_SEP  + languages[i][0], languages[i][0].toUpperCase());
					langTool.setTooltip(languages[i][1]);
					subToolBar.addTool(langTool);
				}
			}
		}
		
		return toolBar;
	} // end method completeToolBar
	
	/**
	 * Utility method a add a tool to a given toolbar and create it if it's null.
	 */
	protected LyToolBar addTool(LyToolBar toolBar, LyTool tool)
	{
		if (tool != null)
		{
			if (toolBar == null)
			{
				toolBar = new LyToolBar();
				toolBar.addTool(tool);
				toolBar.setController(_parent.getId());
			}
			else
				toolBar.insertTool(tool, 0);
		}
		
		return toolBar;
	}
} // end class LyFieldConsultBuilder
