/**
 * LyFieldSetBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyFieldSetBuilder.java 41811 2012-12-19 08:10:08Z bangiole $
 */

package leon.control.form.builder;

import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.control.*;
import leon.control.builder.*;
import leon.control.form.*;
import leon.data.*;
import leon.data.daemon.*;
import leon.info.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.view.*;


/**
 * This builder creates a modification form field.
 *
 * @see LyFieldSetController
 */
public class LyFieldSetBuilder extends LyFieldConsultBuilder
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyFieldSetBuilder.java 41811 2012-12-19 08:10:08Z bangiole $";

	/** Indicates if we show a label when we displayed a boolean choice field. */
	public static boolean		BOOLEAN_SHOW_LABEL = false;

	/**
	 * Indicates maximum number of options displayed as checkboxes or radioboxes.
	 * Beyond this number, options are displayed as relations (combo or multi-selector).
	 */
	public static int			MAX_CHECKBOX_OPTIONS = 9;


	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyFieldSetBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Build a view component for the given field.
	 *
	 * @param  editable  state for the component (true = editable)
	 * @return the view component for the field
	 */
	public LyFormField buildFormField(boolean editable)
	{
		LyFieldSetController fieldController = (LyFieldSetController)_parent;
		LyFieldInfo			 fieldInfo = fieldController.getFieldInfo();

		if (isConsult(fieldInfo))
			return super.buildFormField(editable);

		boolean				   editableForm = fieldController.isEditable();
		LyFormField			   formField = super.buildFormField(editableForm);
		LySetController		   set = (LySetController)fieldController.getParent();
		LySetBehaviorInterface behavior = set.getSetBehavior();

		if (behavior.needCheck(set, fieldInfo))
			formField.setNeedCheck(true);

		if (formField != null)
		{
			LyFieldInfo displayedFieldInfo = fieldInfo;

			if (fieldController instanceof LyGetArrayInterface)
				displayedFieldInfo = ((LyGetArrayInterface)fieldController).getArray();

			boolean isEditable = set.isEditable() && behavior.isEditable(set, displayedFieldInfo);
			boolean isEnabled = behavior.isEnabled(set, displayedFieldInfo);

			formField.setOptional(behavior.isOptional(set, displayedFieldInfo));
			formField.setEditable(isEditable);
			formField.setEnabled(isEnabled);

			// Update sub-area controller is possible
			LyAreaController areaController = fieldController.getAreaController();

			if (areaController instanceof LyFormInterface)
			{
				((LyFormInterface)areaController).setEditable(isEditable);
				if (areaController instanceof LySetController)
				{
					if (!isEnabled && isEditable)
						((LyFormInterface)areaController).setEnabled(isEnabled);
					else if (isEnabled && !isEditable)
						((LyFormInterface)areaController).setEnabled(isEditable);
				}
				else
					((LyFormInterface)areaController).setEnabled(isEnabled);
			}
		}

		// Add units
		LyFieldUnit unitsList = buildUnitsList(fieldInfo);

		if (unitsList != null)
			formField.setFieldUnit(unitsList);

		// Add a combo to choose relation type
		LyRelation typesRelation = buildTypesRelation(fieldInfo);

		if (typesRelation != null)
			formField.setRelation(typesRelation);

		return formField;
	} // end method buildFormField

	/**
	 * Build a field item for a given fieldInfo.
	 *
	 * @param  fieldInfo the fieldInfo corresponding to the built field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildFieldItem(LyFieldInfo fieldInfo)
	{
		LyUnit unit = ((LyFieldController)_parent).getDefaultFieldUnit();

		if (_parent instanceof LyCompositionFieldUnitSetController)
		{
			LyCompositionFieldUnitSetController fieldSetController =
				(LyCompositionFieldUnitSetController)_parent;

			if ((fieldInfo instanceof LyStructFieldInfo)
					&& (((LyStructFieldInfo)fieldInfo).fromFieldTag())
					&& !(((LyStructFieldInfo)fieldInfo).isMultiple()))
			{
				LyClassInfo		  subStructClass = fieldSetController.getFieldType();
				LyStructFieldInfo subStruct = (LyStructFieldInfo)subStructClass.getParent();

				unit = fieldSetController.getDefaultFieldUnit(subStruct);
			}

		}
		return buildFieldItem(fieldInfo, unit);
	}

	/**
	 * Build a field item for a given fieldInfo and unit.
	 *
	 * @param  fieldInfo the fieldInfo corresponding to the built field item
	 * @param  unit      the unit corresponding to the built field item
	 * @return the field item corresponding to the fieldInfo and unit.
	 */
	public LyFieldItem buildFieldItem(LyFieldInfo fieldInfo, LyUnit unit)
	{
		LySetBuilder set = (LySetBuilder)getFormBuilder();

		if (set == null)
			return buildFieldItem(unit);

		return set.buildFieldItem(fieldInfo, getValue(), unit);
	}

	/**
	 * Build a field item for a given fieldInfo and unit.
	 *
	 * @param  unit      the unit corresponding to the built field item
	 * @return the field item corresponding to the fieldInfo and unit.
	 */
	public LyFieldItem buildFieldItem(LyUnit unit)
	{
		LyFieldController fieldController = (LyFieldController)_parent;
		LyFieldInfo		  fieldInfo = fieldController.getFieldInfo();

		if (unit == null)
			return super.buildFieldItem(fieldInfo);

		LyFormat format = unit.getFormat();

		if ((format == null) || (!format.isEditable()))
			return super.buildFieldItem(fieldInfo);

		if (fieldController instanceof LyCompositionFieldUnitSetController)
		{
			LyArea area = new LyArea();
			LyForm form = new LyForm(fieldController.getId());

			area.setBorder(LyArea.Border.NONE);
			form.setBorder(false);
			area.addPanel(form);

			LyClassInfo classInfo = fieldController.getClassInfo();
			LyFormField formField = new LyFormField(classInfo.getId()
					+ LyFormController.FORM_FIELD_SEP + fieldInfo.getId());

			// Handle tip for this formField...
			LyFieldInfo realField = fieldInfo;

			if (fieldController instanceof LyCompositionFieldUnitSetController)
			{
				LyCompositionFieldUnitSetController unitSetController =
					(LyCompositionFieldUnitSetController)fieldController;

				realField = (LyFieldInfo)unitSetController.getFieldType().getParent();
			}

			LyInfoLabel   infoLabel = fieldController.getInfoLabel(realField,
					LyInfoLabel.Type.SEMANTIC);
			LyLabelDaemon daemon = ((infoLabel == null)
					? null : fieldController.getApplication().getDaemon(infoLabel));
			LyObject	  object = fieldController.getObject();
			String		  overview = null;

			if (infoLabel != null)
				overview = (((daemon == null) || (object == null)) ? infoLabel.getString(object)
																   : daemon.getString(object));
			else
				overview = null;

			LyEnvironment environment = _parent.getEnvironment();
			String		  translatedOverview = environment.translate(overview);

			if ((translatedOverview != null) && (translatedOverview.length() > 0))
			{
				LyTip viewTip = new LyTip();

				viewTip.setOverview(overview);
				formField.setTip(viewTip);
			}

			LyText text = new LyText();
			String formatValue = format.getValue();
			int    nCols = format.getNCols();

			if (formatValue != null)
				text.setFormat(formatValue);

			if (!LyText.SHOW_ARROWS)
				text.showArrows(false);

			if (nCols > 0)
				text.setNCols(nCols);

			text.setEditable(format.isEditable());
			formField.setFieldItem(text);
			form.addFormField(formField);
			return area;
		}
		else
		{
			LyText text = new LyText();
			String formatValue = format.getValue();
			int    nCols = format.getNCols();

			if (formatValue != null)
				text.setFormat(formatValue);
			if (nCols > 0)
				text.setNCols(nCols);

			if (!LyText.SHOW_ARROWS)
				text.showArrows(false);

			text.setEditable(format.isEditable());
			return text;
		}
	} // end method buildFieldItem

	/**
	 * Build a relation that contains all units of the given field.
	 *
	 * @param  fieldInfo FieldInfo for which a units list is required.
	 * @return The relation containing all possible units for the given field info.
	 */
	protected LyFieldUnit buildUnitsList(LyFieldInfo fieldInfo)
	{
		LySetBuilder set = (LySetBuilder)getFormBuilder();

		if (set == null)
			return buildUnitsList();
		return set.buildUnitsList(fieldInfo);
	}

	/**
	 * Build a relation that contains all units of the given field.
	 *
	 * @return The relation containing all possible units for the given field info.
	 */
	public LyFieldUnit buildUnitsList()
	{
		LyFieldController fieldController = (LyFieldController)_parent;
		LyFieldInfo		  fieldInfo = fieldController.getFieldInfo();
		LyUnits			  units = fieldInfo.getUnits();
		LyUnit			  fieldInfoUnit = fieldController.getDefaultFieldUnit();

		if ((fieldInfo instanceof LyStructFieldInfo)
				&& (_parent instanceof LyCompositionFieldUnitSetController))
		{
			LyStructFieldInfo struct = (LyStructFieldInfo)fieldInfo;

			if ((struct.fromFieldTag())
					&& !(struct.isMultiple()))
			{
				LyClassInfo		  subStructClass = ((LyCompositionFieldUnitSetController)_parent)
						.getFieldType();
				LyStructFieldInfo subStruct = (LyStructFieldInfo)subStructClass.getParent();

				units = subStruct.getUnits();

				if (units == null)
					return null;

				fieldInfoUnit = fieldController.getDefaultFieldUnit(subStruct);
			}
		}

		if ((units == null)
		    && ((fieldInfo instanceof LyNumberFieldInfo)
		         || (fieldInfoUnit == null)))
			return null;

		int		    nbUnits = (units == null ? 0 : units.getUnitCount());
		LyFieldUnit fieldUnit = new LyFieldUnit();
		String	    name, unitName, unitAlias;

		if ((nbUnits <= 1)
            && (fieldInfoUnit != null))
		{
			LyLabel label = new LyLabel();

			unitAlias = fieldInfoUnit.getAlias();
			unitName = fieldInfoUnit.getName();
            if ((unitAlias != null) && (unitAlias.length() > 0))
                name = unitAlias;
            else if ((unitName != null) && (unitName.length() > 0))
				name = unitName;
            else
                name = null;

			if (name != null)
				label.setLabel(" " + name);
			fieldUnit.setFieldItem(label);
			return fieldUnit;
		}

		LyRelation relation = new LyRelation();
		int		   i;
		LyUnit     unit;
		LyLink     link;

		for (i = 0; i < nbUnits; i++)
		{
			unit = units.getUnit(i);

			unitAlias = unit.getAlias();
			unitName = unit.getName();
            if ((unitAlias != null) && (unitAlias.length() > 0))
                name = unitAlias;
            else if ((unitName != null) && (unitName.length() > 0))
                name = unitName;
            else
                name = null;

			link = new LyLink(unit.getId(), name);
			relation.addLink(link);
			if (unit == fieldInfoUnit)
				relation.selectLink(i, true);
		}

		fieldUnit.setFieldItem(relation);
		return fieldUnit;
	} // end method buildUnitsList

	/**
	 * Build a text component for the given text field.
	 *
	 * @return the text component
	 */
	public LyFieldItem buildText()
	{
		LyFieldController fieldController = (LyFieldController)getParent();
		LyTextFieldInfo   textField = (LyTextFieldInfo)fieldController.getFieldInfo();

		if (isConsult(textField))
			return super.buildText();

		String		  label = textField.toName(null);
		boolean		  secret = textField.hasMark(LyFieldInfo.MARK_SECRET);
		int			  ncols = textField.getNCols();
		int			  nrows = textField.getNRows();
		int			  nchars = textField.getMaxChars();
		String		  format = textField.getFormat();
		String		  translatedFormat;
		boolean		  multiple = textField.isMultiple();
		boolean		  rich = (textField.getType() == LyTextFieldInfo.Type.RICH);
		LyEnvironment environment = getParent().getEnvironment();

		// translate format
		translatedFormat = environment.getString(format);

		if ((translatedFormat == null) && (format != null))
			translatedFormat = format;

		// Check format
		if (!LyString.checkFormat(environment, translatedFormat))
		{
			environment.logWarning("BF AppEngine : the format \"" + translatedFormat
				+ "\" is not valid. It will be ignored");
			translatedFormat = null;
		}

		if (multiple && (nrows <= 0))
			nrows = 5;

		LyText text = new LyText();

		text.setLabel(label);
		text.setSecret(secret);
		text.setNCols(ncols);
		text.setNRows(nrows);
		text.setNChars(nchars);
		text.setFormat(translatedFormat);

		if (rich)
			text.setType(LyText.Type.RICH);

		if (!LyText.SHOW_ARROWS)
			text.showArrows(false);

		int			   i, n = textField.getCharTypesCount();
		LyInfoCharType infoCharType;
		LyCharType     charType;

		for (i = 0; i < n; i++)
		{
			infoCharType = textField.getCharType(i);
			charType = new LyCharType(infoCharType.getValue());
			text.addCharType(charType);
		}

		// Get auto-completion values
		LySetController		   formController = (LySetController)fieldController.getParent();
		LySetBehaviorInterface behavior = formController.getSetBehavior();
		int					   autoCompletionMax = behavior.getAutoCompletionMaximum(formController,
				textField);

		if ((autoCompletionMax != 0)
//				&& !(formController instanceof LyEditTableSetController)
				&& (text.getNRows() <= 1))
		{
			String[]     values = behavior.getAutoCompletionValues(formController, textField);
			StringBuffer buffer = new StringBuffer();

			n = ((values == null) ? 0 : values.length);

			for (i = 0; i < n; i++)
			{
				if (i == 0)
				{
					buffer.append(autoCompletionMax);
					buffer.append(LyTools.PIPE_SEP);
				}
				else
					buffer.append(LyTools.VALUE_SEP);

				buffer.append(values[i]);
			}

			text.setAutoCompletionValues(buffer.toString());
		}

		return text;
	} // end method buildText

	/**
	 * Build a date component for the given date field.
	 *
	 * @return the date component component
	 */
	public LyFieldItem buildDate()
	{
		LyFieldController fieldController = (LyFieldController)getParent();
		LyDateFieldInfo   dateField = (LyDateFieldInfo)fieldController.getFieldInfo();

		if (isConsult(dateField))
			return super.buildDate();

		String				 label = dateField.toName(null);
		LyDateFieldInfo.Type type = dateField.getType();
		String				 format = dateField.getFormat();
		LyDate				 date = new LyDate();

		if (!LyDate.SHOW_ARROWS)
			date.showArrows(false);

		date.setLabel(label);

		switch (type)
		{
		case DATE_TIME:
			date.setType(LyDate.Type.DATE_TIME);
			break;
		case DATE_HOUR_MIN:
			date.setType(LyDate.Type.DATE_HOUR_MIN);
			break;
		case DATE:
			date.setType(LyDate.Type.DATE);
			break;
		case TIME:
			date.setType(LyDate.Type.TIME);
			break;
		case HOUR_MIN:
			date.setType(LyDate.Type.HOUR_MIN);
			break;
		default:
			throw new IllegalArgumentException("BF AppEngine : Illegal date type : " + type);
		}

		date.setFormat(format);

		return date;
	} // end method buildDate

	/**
	 * Build a number component for the given number field.
	 *
	 * @return the number component component
	 */
	public LyFieldItem buildNumber()
	{
		LyFieldController fieldController = (LyFieldController)getParent();
		LyNumberFieldInfo numberField = (LyNumberFieldInfo)fieldController.getFieldInfo();

		if (isConsult(numberField))
			return super.buildNumber();

		String   label = numberField.toName(null);
		String   format = numberField.getFormat();
		Number   vmin = numberField.getMin();
		Number   vmax = numberField.getMax();
		Number   pitch = numberField.getPitch();
		int		 decimals = numberField.getDecimals();
		LyUnit   fieldUnit = fieldController.getDefaultFieldUnit();
		String   unit = ((fieldUnit == null) ? null : fieldUnit.getName());
		LyUnits  units = numberField.getUnits();
		boolean  modulo = numberField.isModulo();
		LyNumber number = new LyNumber();

		if (!LyNumber.SHOW_ARROWS)
			number.showArrows(false);

		if (label != null)
			number.setLabel(label);
		if (format != null)
			number.setFormat(format);

		if (vmin != null)
			number.setVMin(vmin.doubleValue());
		if (vmax != null)
			number.setVMax(vmax.doubleValue());

		if (pitch != null)
			number.setPitch(pitch.doubleValue());
		number.setDecimals(decimals);
		if ((unit != null) && (units == null))
			number.setUnit(unit);

		number.setModulo(modulo);

		return number;
	} // end method buildNumber

	/**
	 * Build a relation component for the given relation field.
	 *
	 * @return the relation component component
	 */
	public LyFieldItem buildStruct()
	{
		LyFieldController fieldController = (LyFieldController)getParent();
		LyStructFieldInfo structField = (LyStructFieldInfo)fieldController.getFieldInfo();

		if (isConsult(structField))
			return super.buildStruct();

		LyStructFieldInfo struct = structField;

		if ((structField.fromFieldTag())
				&& !(structField.isMultiple())
				&& (_parent instanceof LyCompositionFieldUnitSetController))
		{
			LyCompositionFieldUnitSetController ctl = ((LyCompositionFieldUnitSetController)
					_parent);
			LyClassInfo						    fieldType = ctl.getFieldType();

			struct = (LyStructFieldInfo)fieldType.getParent();
		}

		String label = struct.toName(null);
		int    ncols = LyText.DEFAULT_NCOLS;

		LyText text = new LyText();

		text.setLabel(label);
		text.setNCols(ncols);
		text.setEditable(false);

		return text;
	} // end method buildStruct

	/**
	 * Build a relation component for the given relation field.
	 *
	 * @return the relation component component
	 */
	public LyFieldItem buildRelation()
	{
		LyFieldController   fieldController = (LyFieldController)getParent();
		LyRelationFieldInfo relationField = (LyRelationFieldInfo)fieldController.getFieldInfo();

		if (isConsult(relationField)
                || relationField.getShowAs() == LyRelationFieldInfo.ShowAs.LINK)
			return super.buildRelation();

		boolean    multiple = relationField.isMultiple();
		boolean    ordered = relationField.isOrdered();
		LyRelation relation = new LyRelation();
		int		   nRows = relationField.getNRows();
		int		   nCols = relationField.getNCols();

		relation.setMultiple(multiple);
		relation.setOrdered(ordered);
		relation.setNRows(nRows);
		relation.setNCols(nCols);

		// Get auto-completion maximum value
		LySetController		   formController = (LySetController)fieldController.getParent();
		LySetBehaviorInterface behavior = formController.getSetBehavior();
		int					   autoCompletionMax = behavior.getAutoCompletionMaximum(formController, relationField);
		boolean				   localCompletion = behavior.useLocalCompletion(formController, relationField);

		if ((autoCompletionMax > 0) && localCompletion)
			autoCompletionMax = -1;

		relation.setAutoCompletionMax(autoCompletionMax);

		return relation;
	}

	/**
	 * Build a fieldItem for a fieldinfo to present its value.
	 *
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildValueLabel()
	{
		LyFieldController fieldController = (LyFieldController)_parent;
		LyFieldInfo		  fieldInfo = fieldController.getFieldInfo();

		if (isConsult(fieldInfo))
			return super.buildFieldItem();
		return null;
	}

	/**
	 * Builds a relation to change the relation type.
	 *
	 * @param fieldInfo the corresponding fieldInfo
	 * @return a relation field item.
	 */
	protected LyRelation buildTypesRelation(LyFieldInfo fieldInfo)
	{
		LySetBuilder set = (LySetBuilder)getFormBuilder();

		if (set == null)
			return buildTypesRelation();

		return set.buildTypesRelation(fieldInfo);
	}

	/**
	 * Builds a relation to change the relation type.
	 *
	 * @return a relation field item.
	 */
	public LyRelation buildTypesRelation()
	{
		LyFieldController fieldController = (LyFieldController)_parent;
		LyFieldInfo		  fieldInfo = fieldController.getFieldInfo();

		if (!(fieldInfo instanceof LyRelationFieldInfo))
			return null;

		LyRelationFieldInfo relationField = (LyRelationFieldInfo)fieldInfo;

		if ((relationField.getType() != LyRelationFieldInfo.Type.COMPOSITION)
				|| !relationField.fromFieldTag()
				|| relationField.isMultiple()
				|| (relationField.getParent() instanceof LyArrayFieldInfo))
			return null;

		LyRelation			    typeRelation = new LyRelation();
		LyInfoList<LyClassInfo> classInfos = relationField.getClassInfos();
		LyClassInfo			    type;
		String				    name;
		int					    i, n = classInfos.getSize();
		int					    structIndex;

		// Filling the combo with the classInfos
		for (i = 0; i < n; i++)
		{
			type = classInfos.getInfo(i);
			if (type.isAbstract())
				continue;
			name = type.getName();

			// Removing the "__" and ".structClass" for the struct classes
			structIndex = name.indexOf(".structClass");
			if (structIndex > 0)
				name = name.substring(2, structIndex);

			// Adding the classInfo
			typeRelation.addLink(new LyLink(type.getId(), name));
		}

		int nl = typeRelation.getLinkCount();

		if (nl < 2)
			return null;

		// Selecting the good classInfo
		LyAreaController areaController = fieldController.getAreaController();
		LyClassInfo		 areaClassInfo = ((areaController == null) ? null
																   : areaController.getClassInfo());

		if ((areaClassInfo == null)
				&& (_parent instanceof LyCompositionFieldUnitSetController))
			areaClassInfo = ((LyCompositionFieldUnitSetController)_parent).getFieldType();

		if ((areaClassInfo == null) || (areaClassInfo.isAbstract()))
		{
			LyInfoList<LyClassInfo> areaClassInfos = relationField.getClassInfos();
			int					    index = 0, nClassInfos = areaClassInfos.getSize();

			areaClassInfo = areaClassInfos.getInfo(index++);
			while ((areaClassInfo.isAbstract()) && (index < nClassInfos))
				areaClassInfo = areaClassInfos.getInfo(index);
		}

		String classInfoId = areaClassInfo.getId();

		for (i = 0; i < nl; i++)
		{
			if (typeRelation.getLink(i).getId().equals(classInfoId))
			{
				typeRelation.selectLink(i, true);
				break;
			}
		}

		return typeRelation;
	} // end method buildTypesRelation

	/**
	 * Create the view components associated with each choice option group.
	 *
	 * @param group      the choice option group
	 * @param groupEntry the entry associated with the given group
	 * @param validGroups a list of valid groups
	 * @param validOptions a list of valid options
	 */
	protected void fillEntries(LyChoiceOptionGroup group, LyEntry groupEntry,
		LyInfoList<LyChoiceOptionGroup> validGroups, Vector<LyChoiceOption> validOptions)
	{
		short			  i, n = group.getChoiceItemCount();
		LyChoiceItem	  item;
		LyEntry			  entry;

		for (i = 0; i < n; i++)
		{
			item = group.getChoiceItem(i);

			if (validOptions.contains(item)
					|| ((item instanceof LyChoiceOptionGroup)
						&& validGroups.contains((LyChoiceOptionGroup)item)))
			{
				entry = buildChoiceItem(item);
				groupEntry.addEntry(entry);

				if ((item instanceof LyChoiceOptionGroup)
						&& validGroups.contains((LyChoiceOptionGroup)item))
					fillEntries((LyChoiceOptionGroup)item, entry, validGroups, validOptions);
			}
		}
	} // end method fillEntries

	/**
	 * Build the tree of option(s) for the choice.
	 *
	 * @param  choice the choice
	 * @param domain a domain list
	 * @return the first entry of the tree
	 */
	public LyEntry buildChoiceOptionTree(LyChoiceFieldInfo choice, Vector<LyChoiceOption> domain)
	{
		LyFieldController			    fieldController = (LyFieldController)_parent;
		LySetController				    formController = (LySetController)fieldController.getParent();
		LySetBehaviorInterface		    behavior = ((formController == null) ? null : formController.getSetBehavior());
		LyChoiceOptionGroup			    group;
		LyChoiceOption				    option;
		LyInfo						    parent;
		LyEntry						    rootEntry = null;
		LyEntry						    entry = null;
		LyInfoList<LyChoiceOptionGroup> groups;
		LyInfoList<LyChoiceOptionGroup> parents;
		int							    i, n;
		int							    optionCount = choice.getChoiceOptionCount();
		LyInfoList<LyChoiceOptionGroup> topGroup = new LyInfoList<LyChoiceOptionGroup>();

		if ((domain == null)
				|| (domain.size() == optionCount))
			// If there is no domain or if domain contains all options we can get groups
			// from the choice fieldInfo to optimize the search.
			groups = choice.getChoiceOptionGroups(true);
		else
		{
			// The groups must be search from the available options that are in the domain.
			LyInfoList<LyChoiceOptionGroup>		 choiceGroups = choice.getChoiceOptionGroups(true);
			int									 j, p;
			HashMap<String, LyChoiceOptionGroup> groupsMap = new HashMap<String, LyChoiceOptionGroup>();
			String								 id;

			groups = new LyInfoList<LyChoiceOptionGroup>();
			n = domain.size();
			for (i = 0; i < n; i++)
			{
				option = domain.elementAt(i);
				parents = new LyInfoList<LyChoiceOptionGroup>();

				p = choiceGroups.getSize();
				for (j = 0; j < p; j++)
				{
					group = choiceGroups.getInfo(j);
					if (group.contains(option, false))
						parents.addInfo(group);
				}

				p = parents.getSize();
				for (j = 0; j < p; j++)
				{
					group = parents.getInfo(j);

					if (!behavior.isValidGroup(formController, choice, option, group))
						continue;

					parent = group;
					while ((parent != null)
							&& (parent instanceof LyChoiceOptionGroup))
					{
						id = parent.getId();
						if (!groupsMap.containsKey(id))
							groupsMap.put(id, (LyChoiceOptionGroup)parent);
						parent = parent.getParent();
					}
				}
			}
			Collection<LyChoiceOptionGroup> values = groupsMap.values();
			Iterator<LyChoiceOptionGroup>   it = values.iterator();

			while (it.hasNext())
			{
				group = it.next();
				groups.addInfo(group);
			}
		}

		n = groups.getSize();
		if (n == 0)
			return null;
		else if (n == 1)
		{
			group = groups.getInfo(0);
			rootEntry = buildChoiceItem(group);
			fillEntries(group, rootEntry, groups, domain);
		}
		else
		{
			for (i = 0; i < n; i++)
			{
				group = groups.getInfo(i);

				while ((group != null)
						&& !(group.getParent() instanceof LyChoiceFieldInfo))
					group = (LyChoiceOptionGroup)group.getParent();

				if (group != null && !topGroup.contains(group))
					topGroup.addInfo(group);
			}

			n = topGroup.getSize();
			if (n == 1)
			{
				group = topGroup.getInfo(0);
				rootEntry = buildChoiceItem(group);
				fillEntries(group, rootEntry, groups, domain);
			}
			else
			{
                rootEntry = new LyEntry(choice.getId(), null);

				for (i = 0; i < n; i++)
				{
					group = topGroup.getInfo(i);
					entry = buildChoiceItem(group);
					rootEntry.addEntry(entry);
					fillEntries(group, entry, groups, domain);
				}
			}
		}

		return rootEntry;
	} // end method buildChoiceOptionTree

	/**
	 * Build the dictionary entry associated to the given choice item.
	 *
	 * @param item 	the choice item
	 * @return 		the associated entry
	 */
	protected LyEntry buildChoiceItem(LyChoiceItem item)
	{
		LyFieldController	fieldController = (LyFieldController)_parent;
		LySetController 	parent = (LySetController)fieldController.getParent();
		String 				name = item.getName();
		String 				label = null;
		LyInfoLabel 		infoLabel = null;

		if (item instanceof LyChoiceOption)
			infoLabel = ((LyChoiceOption)item).getLabel(LyInfoLabel.Type.SEMANTIC);
		else
			infoLabel = ((LyChoiceOptionGroup)item).getLabel(LyInfoLabel.Type.SEMANTIC);

		if (infoLabel != null)
			label = infoLabel.getString(parent.getObject());
		else
			label = null;

		LyEntry entry = new LyEntry(item.getId(),
				((label == null) ? name : ((name == null) ? label : (name + " : " + label))));

        entry.setOpen(true);
		return entry;
	}

	/**
	 * Build a choice component for the given choice field.
	 *
	 * @return the choice component component
	 */
	public LyFieldItem buildChoice()
	{
		LyFieldController fieldController = (LyFieldController)getParent();
		LyChoiceFieldInfo choiceField = (LyChoiceFieldInfo)fieldController.getFieldInfo();

		if (isConsult(choiceField))
			return super.buildChoice();

		LySetController		   parent = (LySetController)fieldController.getParent();
		boolean				   multiple = choiceField.isMultiple();
		boolean				   isBoolean = choiceField.getType() == LyChoiceFieldInfo.Type.BOOLEAN;
		short				   i, n = choiceField.getChoiceOptionCount();
		LySetBehaviorInterface behavior = parent.getSetBehavior();
		Vector<LyChoiceOption> domain = behavior.getChoiceDomain(parent, choiceField);
		LyChoiceOption		   choiceOption;

		if (choiceField.isDictionary())
		{
			// build a new LyDictionary
			LyDictionary dico = new LyDictionary();

			dico.setNCols(choiceField.getNCols());
			dico.setMultiple(choiceField.isMultiple());
			if (choiceField.doGroupsHaveValue())
			{
				dico.setSelectionType(LyDictionary.SelectionType.ALL);
				LyChoiceOptionGroup root = choiceField.getChoiceOptionGroup();

				dico.setRootHidden((root.getValue() == null));
			}

			LyEntry rootEntry = buildChoiceOptionTree(choiceField, domain);

			dico.setEntry(rootEntry);

			return dico;
		}

		int	nCols = choiceField.getNCols();
		int	nMax;

		if (parent instanceof LyEditTableSetController)
			nMax = 0;
		else
			nMax = MAX_CHECKBOX_OPTIONS;

		if (isBoolean)
		{
			LyBoolean	   booleanChoice = new LyBoolean();

			if (BOOLEAN_SHOW_LABEL
				&& (nMax > 0))
			{
				if (n > 0)
				{
					choiceOption = choiceField.getChoiceOption((short)0);
					booleanChoice.setLabelUnchecked(choiceOption.getName());
				}
				if (n > 1)
				{
					choiceOption = choiceField.getChoiceOption((short)1);
					booleanChoice.setLabelChecked(choiceOption.getName());
				}
			}

			booleanChoice.setNot(choiceField.isNot());
			return booleanChoice;
		}
		else if (!multiple
			&& (nCols == 0)
			&& (((n == 2)
			        && !fieldController.hasMark(choiceField, LyFieldInfo.MARK_OPTIONAL))
				|| ((n == 1)
				    && fieldController.hasMark(choiceField, LyFieldInfo.MARK_OPTIONAL))))
		{

			LyChoice choice = new LyChoice();
			LyOption option;

			choice.setMultiple(true);
			choiceOption = choiceField.getChoiceOption((short)0);
			option = new LyOption(choiceOption.getId(), null);
			choice.addOption(option);
			return choice;
		}
		else if ((nMax == 0)
				|| (nCols == 0)
				|| ((nCols == -1) && (n > nMax)))
		{
			LyRelation relation = new LyRelation();

			relation.setMultiple(multiple);

			// Relation is filled in the controller
			return relation;
		}
		else
		{
			LyChoice choice = new LyChoice();

			choice.setMultiple(multiple);
			if (nCols < 0)
				nCols = 5;
			choice.setMaxSpan(nCols);

			LyInfoList<LyChoiceOption> options = choiceField.getChoiceOptions(true);
			LyOption				   option;
			for (i = 0; i < n; i++)
			{
				choiceOption = options.getInfo(i);
				option = new LyOption(choiceOption.getId(), choiceOption.getName());
                if (!behavior.validateChoiceOption(parent, choiceField, choiceOption))
                    continue;

				LyInfoLabel infoLabel = choiceOption.getLabel(LyInfoLabel.Type.SEMANTIC);

				if (infoLabel != null)
					option.setOverview(infoLabel.getString(parent.getObject()));

				if ((domain != null) && (!domain.contains(choiceOption)))
					option.setEnabled(false);

				choice.addOption(option);
			}

			return choice;
		}
	} // end method buildChoice

	/**
	 * Build a field item for an array fieldInfo.
	 *
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildArray()
	{
		LyFieldController	   fieldController = (LyFieldController)getParent();
		LySetController		   set = (LySetController)fieldController.getParent();
		LySetBehaviorInterface behavior = set.getSetBehavior();
		LyArrayFieldInfo	   arrayField = (LyArrayFieldInfo)fieldController.getFieldInfo();

		if (behavior.useEditTable(set, arrayField))
			return super.buildArray();

		LyEnvironment     environment = _parent.getEnvironment();
		LyForm			  form = new LyForm(arrayField.getId());
		LyFieldController subFieldController;
		LyFormField		  valueFormField = null;

		// Build a toolbar to create/set objects in the table
		LyToolBar     toolBar = null;
		LyTool		  tool;
		LyApplication application = fieldController.getApplication();
		LyClassInfo   classInfo = set.getClassInfo();
		LyAction	  action = application.getActionWithId(LyAction.ID_UPDATE_ARRAY_VALUE, classInfo);

		tool = buildTool(action, true,
				LyFieldSetController.STANDARD_ARRAY_TOOL + LyController.ID_SEP
				+ arrayField.getId(),
				environment.getImageSizeInfo(DEFAULT_TRANSVERSE_ICON_WIDTH,
					DEFAULT_TRANSVERSE_ICON_WIDTH));

		// Create an edition field for creating/setting content in the view
		subFieldController = fieldController.getFieldController(-1);
		if (subFieldController != null)
		    valueFormField = subFieldController.getFormField();
        if (valueFormField != null)
        {
			valueFormField.setCaption(null);
			valueFormField.setConstraint(null);

			// If a toolbar exists, complete it, otherwise add a new one
			toolBar = valueFormField.getToolBar();

			if (toolBar == null)
				toolBar = new LyToolBar();
			toolBar.addTool(tool);

			valueFormField.setToolBar(toolBar);
			form.addFormField(valueFormField);
        }

		LyTable table = new LyTable("fieldset.array.table." + arrayField.getId());
		LyRows  rows = new LyRows();
		int     nrows = arrayField.getNRows();

		if (nrows <= 0)
			nrows = 3;

		table.setSelectionMode(LyTable.SelectionMode.SINGLE);
		table.setRows(rows);
		table.setNRows(nrows);

		LyFieldInfo fieldInfo = arrayField.getFieldInfo();

		if (fieldInfo instanceof LyStructFieldInfo)
		{
			LyInfoLabel label = ((LyFieldController)_parent).getInfoLabel(fieldInfo,
					LyInfoLabel.Type.VALUE);

			if (label == null)
			{
				LyHeaders headers = new LyHeaders();

				headers.addHeader(new LyHeader());
				table.setHeaders(headers);
			}
		}

		LyFormField tableField = new LyFormField("_table"
				+ LyFormController.FORM_FIELD_SEP + arrayField.getId());

		tableField.setFieldItem(table);

		toolBar = new LyToolBar();
		toolBar.setFill(LyToolBar.Fill.VERTICAL);
		action = application.getActionWithId(LyAction.ID_MOVE_UP_TOOL, classInfo);
		tool = buildTool(action, true,
				LyArrayFieldSetController.MOVE_UP_TOOL + LyController.ID_SEP
				+ arrayField.getId(), null);
		toolBar.addTool(tool);
		action = application.getActionWithId(LyAction.ID_MOVE_DOWN_TOOL, classInfo);
		tool = buildTool(action, true,
				LyArrayFieldSetController.MOVE_DOWN_TOOL + LyController.ID_SEP
				+ arrayField.getId(), null);
		toolBar.addTool(tool);
		toolBar.setController(fieldController.getId());
		tableField.setToolBar(toolBar);

		LyConstraint constraint = new LyConstraint();

		constraint.setRemainder(true);
		constraint.setFill(LyConstraint.Fill.HORIZONTAL);

		tableField.setConstraint(constraint);
		//return table;
		form.addFormField(tableField);

		form.setBorder(false);
		return form;
	} // end method buildArray

	/**
	 * Build a file component for the given file field.
	 *
	 * @return the file component component
	 */
	public LyFieldItem buildFile()
	{
		LyFieldController fieldController = (LyFieldController)getParent();
		LyFileFieldInfo   fileField = (LyFileFieldInfo)fieldController.getFieldInfo();

		if (isConsult(fileField))
			return super.buildFile();

		String	    label = fileField.toName(null);
		LyFile.Type type = LyFile.Type.TEXT;
		String	    mimeTypes = fileField.getMimeTypes();
		boolean		isBlob = (fieldController.getViewManager().hasProperty("MANAGE_ATTACHMENT")
				&& (fileField.getBinding() != null)
				&& ("BLOB".equals(fileField.getBinding().getPeerType())));

		// If file is an url and not a blob, use a text input.
		// In web mode, since web browsers do not send full path from file inputs, use a text input if
		// no upload and no directory is given and not a blob.
		if (!isBlob
				&& (((mimeTypes != null) && (mimeTypes.equals("text/x-url")))
						|| (getViewManager().hasProperty("WEB_MANAGER")
								&& (fileField.getAttachmentHandler() == null)
								&& (fileField.getDirectory() == null))))
		{
			LyText text = new LyText();
			text.setWrap(false);
			text.setLabel(label);
			return text;
		}

		LyFile 		file = new LyFile();
		// Try to guess best mimeType
		if (mimeTypes != null)
		{
			mimeTypes = mimeTypes.toLowerCase();

			if (mimeTypes.startsWith("text"))
				type = LyFile.Type.TEXT;
			else if (mimeTypes.startsWith("image"))
			{
				type = LyFile.Type.IMAGE;

				// Web managers do not display file component as an icon but as a file selector.
				// Do not use the image size for this selector.
				if (!getViewManager().hasProperty("WEB_MANAGER"))
				{
					int		width = fileField.getNCols();
					int		height = fileField.getNRows();

					file.setNCols(width);
					file.setNRows(height);
					file.setBorder(LyFile.Border.SINGLE);
				}
			}
			else if (mimeTypes.startsWith("audio")) //XX Check if valid
				type = LyFile.Type.AUDIO;
			else if (mimeTypes.startsWith("video"))
				type = LyFile.Type.VIDEO;
			else if (mimeTypes.equals("dir"))
				type = LyFile.Type.DIRECTORY;
		}

		file.setLabel(label);
		file.setType(type);
		file.setAttachmentDirectory(fileField.getAttachmentDirectory());
		if (fileField.getAttachmentHandler() != null)
			file.setAttachmentHandler(fileField.getAttachmentHandler());
		else if (isBlob)
		{
			file.setAttachmentHandler("default");

			String directory = fieldController.getViewManager().getTemporaryDir();

			if (directory != null)
				file.setAttachmentDirectory(directory);
			else
				file.setAttachmentDirectory(fileField.getAttachmentDirectory());
		}
		else
			file.setAttachmentHandler(fileField.getAttachmentHandler());

		file.setDirectory(fileField.getDirectory());
		file.setFilter(fileField.getFilter());

		return file;
	} // end method buildFile

	/**
	 * Indicates if the current field is in consult mode.
	 *
	 * @param  fieldInfo the field
	 * @return true if the field is in consult mode
	 */
	protected boolean isConsult(LyFieldInfo fieldInfo)
	{
		LySetBuilder builder = (LySetBuilder)getFormBuilder();

		if (builder == null)
			return isConsult();
		return builder.isConsult(fieldInfo);
	}

	/**
	 * Indicates if the current field is in consult mode.
	 *
	 * @return true if the field is in consult mode
	 */
	public boolean isConsult()
	{
		LyFieldController fieldController = (LyFieldController)_parent;
		LyFieldInfo		  fieldInfo = fieldController.getFieldInfo();
		LySetController   set = (LySetController)fieldController.getParent();

		return set.isConsult(fieldInfo);
	}

	/**
	 * Indicates if the current field is enabled.
	 *
	 * @param  fieldInfo the field
	 * @return true if the field is enabled
	 */
	protected boolean isEnabled(LyFieldInfo fieldInfo)
	{
		LySetBuilder builder = (LySetBuilder)getFormBuilder();

		if (builder == null)
			return isEditable();
		return builder.isEditable(fieldInfo);
	}

	/**
	 * Indicates if the current field is enabled.
	 *
	 * @return true if the field is enabled
	 */
	public boolean isEnabled()
	{
		LyFieldController fieldController = (LyFieldController)_parent;
		LyFieldInfo		  fieldInfo = fieldController.getFieldInfo();
		LySetController   set = (LySetController)fieldController.getParent();

		return set.isEnabled(fieldInfo);
	}

	/**
	 * Indicates if the current field is editable.
	 *
	 * @param  fieldInfo the field
	 * @return true if the field is editable
	 */
	protected boolean isEditable(LyFieldInfo fieldInfo)
	{
		LySetBuilder builder = (LySetBuilder)getFormBuilder();

		if (builder == null)
			return isEditable();
		return builder.isEditable(fieldInfo);
	}

	/**
	 * Indicates if the current field is editable.
	 *
	 * @return true if the field is editable
	 */
	public boolean isEditable()
	{
		LyFieldController fieldController = (LyFieldController)_parent;
		LyFieldInfo		  fieldInfo = fieldController.getFieldInfo();
		LySetController   set = (LySetController)fieldController.getParent();

		return set.isEditable(fieldInfo);
	}

	/**
	 * Update the toolbar for the given field.
	 *
	 * @param  fieldInfo the field
	 * @return the toolbar for the given field
	 */
	public LyToolBar updateToolBar(LyFieldInfo fieldInfo)
	{
		LySetBuilder builder = (LySetBuilder)getFormBuilder();

		if (builder == null)
			return updateToolBar();

		return builder.updateToolBar(fieldInfo);
	}

	/**
	 * Update the toolbar for the given field.
	 *
	 * @return the toolbar for the given field
	 */
	public LyToolBar updateToolBar()
	{
		LyFieldController fieldController = (LyFieldController)_parent;
		LyFieldInfo		  fieldInfo = fieldController.getFieldInfo();

		return buildToolBar(fieldInfo);
	}

	/**
	 * Builds the toolbar for the given field.
	 *
	 * @return the toolbar for the given field
	 */
	public LyToolBar buildToolBar()
	{
		LyFieldSetController fieldController = (LyFieldSetController)_parent;
		LyFieldInfo			 fieldInfo;

		if (fieldController instanceof LyGetArrayInterface)
			fieldInfo = ((LyGetArrayInterface)fieldController).getArray();
		else
			fieldInfo = fieldController.getFieldInfo();

		if (isConsult(fieldInfo))
			return null;

		if ((fieldInfo instanceof LyRelationFieldInfo)
				&& ((LyRelationFieldInfo)fieldInfo).isToolbarDisabled())
			return null;

		// Add toolbar for transverse actions if needed
		LyInfoDataList<LyAction, LyClassInfo>	actions = fieldController.getTransverseActions();
		LyToolBar								toolBar = buildToolBar(fieldInfo, actions);

		return completeToolBar(toolBar, fieldInfo);
	}

	/**
	 * Complete toolBar for given field info.
	 */
	protected LyToolBar completeToolBar(LyToolBar toolBar, LyFieldInfo fieldInfo)
	{
		LyFieldSetController fieldController = (LyFieldSetController)_parent;

		if (fieldInfo instanceof LyDateFieldInfo)
		{
			LyUnit unit = fieldController.getDefaultFieldUnit();

			if (unit == null)
				return null;

			LyFormat format = unit.getFormat();

			if ((format == null) || !format.isEditable())
				return null;

		    LySetController  set = (LySetController)fieldController.getParent();
		    LyClassInfo      classInfo = set.getClassInfo();
			LyApplication    application = fieldController.getApplication();
			LyAction	     action = application.getActionWithId(LyAction.ID_DATE_TOOL, classInfo);
			LyEnvironment    environment = application.getEnvironment();
			LyTool			 dateTool = buildTool(action, true,
				LyDateFieldSetController.DATE_TOOL + LyController.ID_SEP + fieldInfo.getId(),
				environment.getImageSizeInfo(DEFAULT_TRANSVERSE_ICON_WIDTH,
					DEFAULT_TRANSVERSE_ICON_WIDTH));

			dateTool.setTooltip("LY_DATE_TOOL");
			toolBar = addTool(toolBar, dateTool);
		}

		return super.completeToolBar(toolBar, fieldInfo);
	} // end method completeToolBar
} // end class LyFieldSetBuilder
