/**
 * LyFormFieldBuilder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyFormFieldBuilder.java 41407 2012-10-30 16:17:49Z bangiole $
 */

package leon.control.form.builder;

import leon.app.*;
import leon.control.*;
import leon.control.builder.*;
import leon.control.form.*;
import leon.data.*;
import leon.data.daemon.*;
import leon.info.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.view.*;
import leon.view.mobile.*;


/**
 * This builder creates a form field.
 *
 * @see LyFormController
 */
public abstract class LyFormFieldBuilder extends LyBuilder
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyFormFieldBuilder.java 41407 2012-10-30 16:17:49Z bangiole $";

	static final String		   NO_TAB_MODE = "NO_TAB";

	/** Default size for Transverse icon (actions) */
	public static int		   DEFAULT_TRANSVERSE_ICON_WIDTH = 15;

	/** Separator used to build form field identifiers in composition fields. */
	public static final String FORM_FIELD_COMPOSITION_SEP = "_._";

	/**
	 * Default constructor.
	 *
	 * @param parent parent controller
	 */
	public LyFormFieldBuilder(LyController parent)
	{
		super(parent);
	}

	/**
	 * Builds the view component.
	 *
	 * @return the view Component
	 */
	public LyComponent build()
	{
		LyFieldController fieldController = (LyFieldController)_parent;
		LyFieldInfo		  fieldInfo = fieldController.getFieldInfo();

		return buildFormField(fieldInfo, false);
	}

	/**
	 * Build a form field for a given fieldInfo.
	 *
	 * @param  fieldInfo the fieldInfo corresponding to the built form field
	 * @param  editable  indicates whether the form field must be editable or not
	 * @return the form field corresponding to the fieldInfo
	 */
	public LyFormField buildFormField(LyFieldInfo fieldInfo, boolean editable)
	{
		LyFormBuilder builder = getFormBuilder();

		if (builder == null)
			return buildFormField(editable);
		return builder.buildFormField(fieldInfo, editable);
	}

	/**
	 * Build a form field for a given fieldInfo.
	 *
	 * @param  editable indicates whether the form field must be editable or not
	 * @return the form field corresponding to the fieldInfo
	 */
	public LyFormField buildFormField(boolean editable)
	{
		LyFormField		  formField;
		LyCaption		  caption = null;
		LyFieldItem		  fieldItem;
		LyFieldController fieldController = (LyFieldController)_parent;
		LyFieldInfo		  fieldInfo = fieldController.getFieldInfo();
		LyClassInfo		  classInfo = fieldController.getClassInfo();
		LyEnvironment     environment = _parent.getEnvironment();
		LyObject		  object = fieldController.getObject();
		LyInfoLabel		  infoLabel = null;
		String			  label = null;
		String			  id = classInfo.getId() + LyFormController.FORM_FIELD_SEP + fieldInfo.getId();

		if ((fieldController.getParent() instanceof LySetController)
				&& (((LySetController)fieldController.getParent()).getCompositionRelation() != null)
				&& !((LySetController)fieldController.getParent()).getCompositionRelation().isMultiple())
		{
			LyRelationFieldInfo	composition = ((LySetController)fieldController.getParent()).getCompositionRelation();

			id = composition.getId() + LyFormFieldBuilder.FORM_FIELD_COMPOSITION_SEP + id;
		}
		else if (fieldController instanceof LyFieldSetController)
		{
			LyFieldSetController	parentFieldController = ((LyFieldSetController)fieldController).getParentFieldController();

			if (parentFieldController instanceof LyArrayFieldSetController)
			{
				LyFieldInfo array = parentFieldController.getFieldInfo();

				id = array.getId() + LyFormFieldBuilder.FORM_FIELD_COMPOSITION_SEP + id;
			}
		}

		formField = new LyFormField(id);
		fieldController.setFormField(formField);

		formField.setEditable(editable);

		if (fieldController.hasMark(fieldInfo, LyFieldInfo.MARK_OPTIONAL)
				|| (fieldController.getParent() instanceof LyConsultController))
			formField.setOptional(true);
		else
			formField.setOptional(false);

		if (fieldController instanceof LyGetArrayInterface)
		{
			LyArrayFieldInfo array = ((LyGetArrayInterface)fieldController).getArray();

			infoLabel = fieldController.getInfoLabel(array, LyInfoLabel.Type.NAME);
			label = array.getName();

			if (fieldController.hasMark(array, LyFieldInfo.MARK_OPTIONAL)
					|| (fieldController.getParent() instanceof LyConsultController))
				formField.setOptional(true);
			else
				formField.setOptional(false);
		}

		if (infoLabel == null)
			infoLabel = fieldController.getInfoLabel(fieldInfo, LyInfoLabel.Type.NAME);
		if (infoLabel != null)
			label = infoLabel.getString(object);

		if (label == null)
		{
	        LyViewManager	viewManager = getViewManager();

	        if ((viewManager == null)
	                || !viewManager.hasProperty(LyMobileViewManager.MOBILE_ONLY_PROPERTY))
	        	label = fieldInfo.getName();
	        else
	        	label = fieldInfo.getAlias();
		}

		if (!fieldInfo.hasMark(LyFieldInfo.MARK_NO_CAPTION))
		{
			String	translatedLabel = environment.translate(label);

			if ((label != null)
				&& !label.equals("_null")
				&& !"_null".equals(translatedLabel))
			{
				caption = new LyCaption();
				caption.setLabel(label);
				formField.setCaption(caption);
			}
		}

		fieldItem = buildFieldItem(fieldInfo);
		formField.setFieldItem(fieldItem);

		// Set constraints
		if (caption != null)
		{
			LyConstraint captionConstraint = getCaptionConstraint(fieldInfo, formField);

			if (captionConstraint != null)
				caption.setConstraint(captionConstraint);
		}

		LyConstraint fieldConstraint = getFieldConstraint(fieldInfo, formField);

		if (fieldConstraint != null)
			formField.setConstraint(fieldConstraint);

		// Add toolbar when needed
		LyToolBar toolBar = null;

		if (fieldController instanceof LyGetArrayInterface)
		{
			LyArrayFieldInfo array = ((LyGetArrayInterface)fieldController).getArray();

			toolBar = buildToolBar(array);
		}
		else
			toolBar = buildToolBar(fieldInfo);

		if (toolBar != null)
			formField.setToolBar(toolBar);
		else
		{
			if ((fieldInfo instanceof LyRelationFieldInfo)
					&& !fieldInfo.isMultiple()
					&& (fieldItem instanceof LyArea))
				((LyArea)fieldItem).setScrolling(LyArea.Scrolling.NO);
		}

		// Set tip
		infoLabel = fieldController.getInfoLabel(fieldInfo, LyInfoLabel.Type.GRAMMATICAL);
		LyLabelDaemon daemon = ((infoLabel == null)
				? null : fieldController.getApplication().getDaemon(infoLabel));
		String		  tip, overview;

		if (infoLabel != null)
			tip = (((daemon == null) || (object == null)) ? infoLabel.getString(object)
														  : daemon.getString(object));
		else
			tip = fieldInfo.getTip();

		infoLabel = fieldController.getInfoLabel(fieldInfo, LyInfoLabel.Type.SEMANTIC);
		daemon = ((infoLabel == null) ? null
									  : fieldController.getApplication().getDaemon(infoLabel));
		if (infoLabel != null)
			overview = (((daemon == null) || (object == null)) ? infoLabel.getString(object)
															   : daemon.getString(object));
		else
			overview = null;

		String translatedTip = environment.translate(tip);
		String translatedOverview = environment.translate(overview);

		LyTip  viewTip = null;

		if ((translatedTip != null) && (translatedTip.length() > 0))
		{
			viewTip = new LyTip();
			viewTip.setLabel(tip);
		}

		if ((translatedOverview != null) && (translatedOverview.length() > 0))
		{
			if (viewTip == null)
				viewTip = new LyTip();

			viewTip.setOverview(overview);
		}

		if (viewTip != null)
			formField.setTip(viewTip);

		if ((fieldInfo instanceof LyDateFieldInfo)
				&& ((LyDateFieldInfo)fieldInfo).isWithTimeZone())
		{
			LyFieldUnit fieldUnit = new LyFieldUnit();

			LyApplication application = fieldInfo.getApplication();
			String timeZones[] = application.getSupportedTimeZones();
			String userTz = getParent().getEnvironment().getEnv("LY_TIMEZONE");
			LyRelation relation = new LyRelation();
			LyLink link;
			int i, n = (timeZones == null ? 0 : timeZones.length);
			for (i = 0; i < n; i++)
			{
				link = new LyLink(timeZones[i], timeZones[i]);
				if (timeZones[i].equals(userTz))
					link.setSelected(true);
				else
					link.setSelected(false);
				relation.addLink(link);
			}
			fieldUnit.setFieldItem(relation);
			formField.setFieldUnit(fieldUnit);//setRelation(relation);
		}
		return formField;
	} // end method buildFormField

	/**
	 * Get caption constraint.
	 *
	 * @param  fieldInfo the fieldInfo
	 * @param  formField the formfield associated with the fieldInfo
	 * @return the constraint for the caption
	 */
	protected LyConstraint getCaptionConstraint(LyFieldInfo fieldInfo, LyFormField formField)
	{
		LyFormBuilder builder = getFormBuilder();

		if (builder == null)
			return getCaptionConstraint(formField);
		return builder.getCaptionConstraint(fieldInfo, formField);
	}

	/**
	 * Get caption constraint.
	 *
	 * @param  formField the formfield associated with the fieldInfo
	 * @return the constraint for the caption
	 */
	public LyConstraint getCaptionConstraint(LyFormField formField)
	{
		LyFieldController fieldController = (LyFieldController)_parent;
		LyFormController  formController = fieldController.getFormController();
		LyFieldInfo		  fieldInfo = fieldController.getFieldInfo();
		LyConstraint	  constraint = new LyConstraint();
		LyFieldConstraint captionConstraint
		    = fieldInfo.getConstraint(LyFieldConstraint.Type.CAPTION, formController.getAction());

		if (captionConstraint != null)
		{
			constraint.setRowSpan(captionConstraint.getRowSpan());
			constraint.setColSpan(captionConstraint.getColSpan());
			constraint.setRemainder(captionConstraint.isRemainder());
			constraint.setFill(LyFormBuilder.getFill(captionConstraint.getFill()));
			constraint.setHAlign(LyFormBuilder.getHAlign(captionConstraint.getHAlign()));
			constraint.setVAlign(LyFormBuilder.getVAlign(captionConstraint.getVAlign()));
			constraint.setColWeight(captionConstraint.getColWeight());
			constraint.setRowWeight(captionConstraint.getRowWeight());
			// XX to do colWeight rowWeight
		}
		else
		{
			LyFieldItem fieldItem = formField.getFieldItem();

			constraint.setHAlign(LyConstraint.HAlign.RIGHT);
			if ((!(fieldInfo instanceof LyChoiceFieldInfo)
						&& fieldInfo.isMultiple())
					|| ((fieldInfo instanceof LyTextFieldInfo)
						&& (((LyTextFieldInfo)fieldInfo).getNRows() > 1))
					|| ((fieldInfo instanceof LyChoiceFieldInfo)
						&& (((LyChoiceFieldInfo)fieldInfo).getNCols() > 0))
					|| ((fieldInfo instanceof LyFileFieldInfo)
						&& (fieldItem instanceof LyIcon))
					|| (fieldInfo instanceof LyArrayFieldInfo))
				constraint.setVAlign(LyConstraint.VAlign.TOP);
			else
				constraint.setVAlign(LyConstraint.VAlign.MIDDLE);
		}

		return constraint;
	} // end method getCaptionConstraint

	/**
	 * Get getFieldConstraint constraint.
	 *
	 * @param  fieldInfo the fieldInfo
	 * @param  formField the formfield associated with the fieldInfo
	 * @return the constraint for the caption
	 */
	protected LyConstraint getFieldConstraint(LyFieldInfo fieldInfo, LyFormField formField)
	{
		LyFormBuilder builder = getFormBuilder();

		if (builder == null)
			return getFieldConstraint(formField);
		return builder.getFieldConstraint(fieldInfo, formField);
	}

	/**
	 * Get field constraint.
	 *
	 * @param  formField the formfield associated with the fieldInfo
	 * @return the constraint for the caption
	 */
	public LyConstraint getFieldConstraint(LyFormField formField)
	{
		LyFieldController fieldController = (LyFieldController)_parent;
	    LyFormController  formController = fieldController.getFormController();
		LyFieldInfo		  fieldInfo = fieldController.getFieldInfo();
        LyFieldConstraint fieldConstraint
            = fieldInfo.getConstraint(LyFieldConstraint.Type.FIELD, formController.getAction());
		LyConstraint	  constraint = new LyConstraint();

		if (fieldConstraint != null)
		{
			constraint.setRowSpan(fieldConstraint.getRowSpan());
			constraint.setColSpan(fieldConstraint.getColSpan());
			constraint.setColWeight(fieldConstraint.getColWeight());
			constraint.setRowWeight(fieldConstraint.getRowWeight());

			if (fieldConstraint.isRemainder())
				constraint.setRemainder(true);
			else if (fieldConstraint.getColSpan() == LyFieldConstraint.REMAINDER)
			{
				// Below if kept for former version compatibility.
				_parent.getEnvironment().logWarning(
					"LyFormBuilder : LyFieldConstraint.REMAINDER value for colspan is deprecated. \nUse remainder attribute on CONSTRAINT instead.");

				constraint.setRemainder(true);
			}
			else
				constraint.setRemainder(false);

			// XX to do colWeight rowWeight
			constraint.setFill(LyFormBuilder.getFill(fieldConstraint.getFill()));
			constraint.setHAlign(LyFormBuilder.getHAlign(fieldConstraint.getHAlign()));
			constraint.setVAlign(LyFormBuilder.getVAlign(fieldConstraint.getVAlign()));
		}
		else
		{
			LyCaption    caption = formField.getCaption();
			LyFieldItem	 item = formField.getFieldItem();
			LyConstraint captionConstraint = ((caption == null) ? null : caption.getConstraint());

			constraint.setHAlign(LyConstraint.HAlign.LEFT);
			if (captionConstraint != null)
				constraint.setVAlign(captionConstraint.getVAlign());

			constraint.setRemainder(true);

			if (fieldInfo.isMultiple())
			{
				if (fieldInfo instanceof LyTextFieldInfo)
					constraint.setFill(LyConstraint.Fill.BOTH);
				else if (fieldInfo instanceof LyRelationFieldInfo)
					constraint.setFill(LyConstraint.Fill.HORIZONTAL);
				else if ((fieldInfo instanceof LyChoiceFieldInfo)
						&& (item instanceof LyRelation))
					constraint.setFill(LyConstraint.Fill.HORIZONTAL);
			}
		}

		return constraint;
	} // end method getFieldConstraint

	/**
	 * Build a field item for a given fieldInfo.
	 *
	 * @param  fieldInfo the fieldInfo corresponding to the built field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildFieldItem(LyFieldInfo fieldInfo)
	{
		LyFormBuilder builder = getFormBuilder();

		if (builder == null)
			return buildFieldItem();
		return builder.buildFieldItem(fieldInfo);
	}

	/**
	 * Build a field item for a given fieldInfo.
	 *
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildFieldItem()
	{
		LyFieldController fieldController = (LyFieldController)_parent;
		LyFieldInfo		  fieldInfo = fieldController.getFieldInfo();
		LyAreaController  areaController = fieldController.getAreaController();

		if (areaController != null)
		{
			LyArea area = (LyArea)areaController.getComponent();

			if (_parent instanceof LyCompositionFieldUnitSetController)
			{
				area.setLabel("");
				area.setBorder(LyArea.Border.IN);
				return area;
			}
			else
			{
				LyEnvironment environment = _parent.getEnvironment();
				String		  name = environment.translate(fieldInfo.getName());
				boolean		  addBorder = ((name == null) ? false :
					(name.trim().length() != 0) && !name.equals("_null") && !fieldInfo.isMultiple());

				// From field tag --> adding a parent area to contain the area fieldItem
				if ((fieldInfo.fromFieldTag())
						&& !(fieldInfo.isMultiple()))
				{
					LyArea parentArea = new LyArea();

					parentArea.setEnvironment(area.getEnvironment());

					if (area.getBorder() == LyArea.Border.NONE)
					{
						if (addBorder)
						{
							parentArea.setLabel("");
							parentArea.setBorder(LyArea.Border.IN);
						}

						parentArea.setScrolling(LyArea.Scrolling.NO);

						if (addBorder)
						{
							area.setLabel("");
							area.setScrolling(LyArea.Scrolling.NO);
						}

						LyComponent areaChild = area.getChild(0);

						if (areaController instanceof LySetController)
							fieldController.setAreaNoScroll(areaChild);
					}
					parentArea.addPanel(area);
					return parentArea;
				}
				else
				{
					area.setId(fieldController.getId());

					if ((area.getBorder() == LyArea.Border.NONE)
							&& addBorder)
					{
						area.setLabel("");
						area.setBorder(LyArea.Border.IN);
						area.setScrolling(LyArea.Scrolling.NO);
					}

					return area;
				}
			}
		}

		LyFieldItem fieldItem = null;
		LyInfoLabel valueLabel = fieldController.getInfoLabel(fieldInfo, LyInfoLabel.Type.VALUE);

		if (valueLabel != null)
			fieldItem = buildValueLabel(fieldInfo);

		if (fieldItem == null)
			fieldItem = buildInternalFieldItem(fieldInfo);

		// Check if field info has a bean
		LyFormController formController = (LyFormController)fieldController.getParent();
		LyAction		 action = formController.getAction();
		LyAction		 rootAction = action.getRoot();
		String			 rootActionType = rootAction.getId();
		boolean			 editable = (((formController instanceof LySetController)
						&& !((LySetController)formController).isConsult(fieldInfo))
					|| (formController instanceof LyFilterController));

		// If field is not editable, use consult bean
		if (!editable)
			rootActionType = LyAction.ID_CONSULT;

		LyFieldBean bean = fieldInfo.getFieldBean(rootActionType);

		if (bean != null)
		{
			fieldItem.setBean(bean.getJavaName());
			if (bean.getDatas() != null)
				fieldItem.setDatas(bean.getDatas());
		}

		return fieldItem;
	} // end method buildFieldItem

	/**
	 * Build a fieldItem for the given fieldInfo
	 * @param fieldInfo a fieldInfo
	 * @return the resulting fieldItem
	 */
	public LyFieldItem buildInternalFieldItem(LyFieldInfo fieldInfo)
	{
		LyFieldItem fieldItem = null;

		if (fieldInfo instanceof LyTextFieldInfo)
			fieldItem = buildText((LyTextFieldInfo)fieldInfo);
		else if (fieldInfo instanceof LyNumberFieldInfo)
			fieldItem = buildNumber((LyNumberFieldInfo)fieldInfo);
		else if (fieldInfo instanceof LyDateFieldInfo)
			fieldItem = buildDate((LyDateFieldInfo)fieldInfo);
		else if (fieldInfo instanceof LyStructFieldInfo)
			fieldItem = buildStruct((LyStructFieldInfo)fieldInfo);
		else if (fieldInfo instanceof LyRelationFieldInfo)
			fieldItem = buildRelation((LyRelationFieldInfo)fieldInfo);
		else if (fieldInfo instanceof LyChoiceFieldInfo)
			fieldItem = buildChoice((LyChoiceFieldInfo)fieldInfo);
		else if (fieldInfo instanceof LyFileFieldInfo)
			fieldItem = buildFile((LyFileFieldInfo)fieldInfo);
		else if (fieldInfo instanceof LyArrayFieldInfo)
			fieldItem = buildArray((LyArrayFieldInfo)fieldInfo);
		else
			_parent.getEnvironment().logWarning("FormBuilder : Unable to build fieldInfo : "
				+ fieldInfo);
		return fieldItem;
	}

	/**
	 * Build a field item for a text fieldInfo.
	 *
	 * @param  textField the text fieldInfo corresponding to the built field item
	 * @return the field item corresponding to the fieldInfo
	 */
	protected LyFieldItem buildText(LyTextFieldInfo textField)
	{
		LyFormBuilder builder = getFormBuilder();

		if (builder == null)
			return buildText();
		return builder.buildText(textField, getValue());
	}

	/**
	 * Build a field item for a text fieldInfo.
	 *
	 * @return the field item corresponding to the fieldInfo
	 */
	public abstract LyFieldItem buildText();

	/**
	 * Build a field item for a date fieldInfo.
	 *
	 * @param  dateField the date fieldInfo corresponding to the built field item
	 * @return the field item corresponding to the fieldInfo
	 */
	protected LyFieldItem buildDate(LyDateFieldInfo dateField)
	{
		LyFormBuilder builder = getFormBuilder();

		if (builder == null)
			return buildDate();
		return builder.buildDate(dateField, getValue());
	}

	/**
	 * Build a field item for a date fieldInfo.
	 *
	 * @return the field item corresponding to the fieldInfo
	 */
	public abstract LyFieldItem buildDate();

	/**
	 * Build a field item for a number fieldInfo.
	 *
	 * @param  numberField the number fieldInfo corresponding to the built field item
	 * @return the field item corresponding to the fieldInfo
	 */
	protected LyFieldItem buildNumber(LyNumberFieldInfo numberField)
	{
		LyFormBuilder builder = getFormBuilder();

		if (builder == null)
			return buildNumber();

		return builder.buildNumber(numberField, getValue());
	}

	/**
	 * Build a field item for a number fieldInfo.
	 *
	 * @return the field item corresponding to the fieldInfo
	 */
	public abstract LyFieldItem buildNumber();

	/**
	 * Build a field item for a struct fieldInfo.
	 *
	 * @param  structField the struct fieldInfo corresponding to the built field item
	 * @return the field item corresponding to the fieldInfo
	 */
	protected LyFieldItem buildStruct(LyStructFieldInfo structField)
	{
		LyFormBuilder builder = getFormBuilder();

		if (builder == null)
			return buildStruct();
		return builder.buildStruct(structField, getValue());
	}

	/**
	 * Build a field item for a struct fieldInfo.
	 *
	 * @return the field item corresponding to the fieldInfo
	 */
	public abstract LyFieldItem buildStruct();

	/**
	 * Build a field item for a relation fieldInfo.
	 *
	 * @param  relationField the relation fieldInfo corresponding to the built field item
	 * @return the field item corresponding to the fieldInfo
	 */
	protected LyFieldItem buildRelation(LyRelationFieldInfo relationField)
	{
		LyFormBuilder builder = getFormBuilder();

		if (builder == null)
			return buildRelation();
		return builder.buildRelation(relationField, getValue());
	}

	/**
	 * Build a field item for a relation fieldInfo.
	 *
	 * @return the field item corresponding to the fieldInfo
	 */
	public abstract LyFieldItem buildRelation();

	/**
	 * Build a field item for a choice fieldInfo.
	 *
	 * @param  choiceField the choice fieldInfo corresponding to the built field item
	 * @return the field item corresponding to the fieldInfo
	 */
	protected LyFieldItem buildChoice(LyChoiceFieldInfo choiceField)
	{
		LyFormBuilder builder = getFormBuilder();

		if (builder == null)
			return buildChoice();
		return builder.buildChoice(choiceField, getValue());
	}

	/**
	 * Build a field item for a choice fieldInfo.
	 *
	 * @return the field item corresponding to the fieldInfo
	 */
	public abstract LyFieldItem buildChoice();

	/**
	 * Build a field item for a file fieldInfo.
	 *
	 * @param  fileField the file fieldInfo corresponding to the built field item
	 * @return the field item corresponding to the fieldInfo
	 */
	protected LyFieldItem buildFile(LyFileFieldInfo fileField)
	{
		LyFormBuilder builder = getFormBuilder();

		if (builder == null)
			return buildFile();
		return builder.buildFile(fileField, getValue());
	}

	/**
	 * Build a field item for a file fieldInfo.
	 *
	 * @return the field item corresponding to the fieldInfo
	 */
	public abstract LyFieldItem buildFile();

	/**
	 * Build a field item for an array fieldInfo.
	 *
	 * @param  arrayField the array fieldInfo corresponding to the built field item
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildArray(LyArrayFieldInfo arrayField)
	{
		LyFormBuilder builder = getFormBuilder();

		if (builder == null)
			return buildArray();
		return builder.buildArray(arrayField, getValue());
	}

	/**
	 * Build a field item for an array fieldInfo.
	 *
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildArray()
	{
		LyFieldController fieldController = (LyFieldController)getParent();
		LyArrayFieldInfo  arrayField = (LyArrayFieldInfo)fieldController.getFieldInfo();
		LyTable 		  table = new LyTable(arrayField.getId());
		LyRows  		  rows = new LyRows();
		int     		  nrows = arrayField.getNRows();

		if (nrows <= 0)
			nrows = 3;

		table.setSelectionMode(LyTable.SelectionMode.NONE);
		table.setRows(rows);
		table.setNRows(nrows);

		return table;
	} // end method buildArray

	/**
	 * Build a fieldItem for a fieldinfo to present its value.
	 *
	 * @param  fieldInfo the fieldInfo corresponding to the built field item
	 * @return the field item corresponding to the fieldInfo
	 */
	protected LyFieldItem buildValueLabel(LyFieldInfo fieldInfo)
	{
		LyFormBuilder builder = getFormBuilder();

		if (builder == null)
			return buildValueLabel();
		return builder.buildValueLabel(fieldInfo, getValue());
	}

	/**
	 * Build a fieldItem for a fieldinfo to present its value.
	 *
	 * @return the field item corresponding to the fieldInfo
	 */
	public LyFieldItem buildValueLabel()
	{
		LyLabel label = new LyLabel();

		return label;
	}

	/**
	 * Build a toolbar for a given fieldInfo. The built toolbar is added to the form field
	 * corresponding to the given fieldInfo
	 *
	 * @param  fieldInfo the fieldInfo corresponding to the built toolbar
	 * @return the toolbar for the given fieldInfo
	 */
	protected LyToolBar buildToolBar(LyFieldInfo fieldInfo)
	{
		LyFormBuilder builder = getFormBuilder();

		if (builder == null)
			return buildToolBar();
		return builder.buildToolBar(fieldInfo);
	}

	/**
	 * Build a toolbar for a given fieldInfo. The built toolbar is added to the form field
	 * corresponding to the given fieldInfo
	 *
	 * @return the toolbar for the given fieldInfo
	 */
	public LyToolBar buildToolBar()
	{
		return null;
	}

	/**
	 * Builds the toolbar for the given field and list of actions.
	 *
	 * @param  fieldInfo   the field
	 * @param  actions     List of pairs (action, classInfo)
	 * @return the toolbar for the given field
	 */
	public LyToolBar buildToolBar(LyFieldInfo fieldInfo,
		LyInfoDataList<LyAction, LyClassInfo> actions)
	{
		LyFormBuilder builder = getFormBuilder();

		if (builder == null)
			return buildToolBar(actions);
		return builder.buildToolBar(fieldInfo, actions);
	}

	/**
	 * Builds the toolbar for the given relation field and list of actions.
	 *
	 * @param  actions List of pairs (action, classInfo)
	 * @return the toolbar for the given relation field
	 */
	public LyToolBar buildToolBar(LyInfoDataList<LyAction, LyClassInfo> actions)
	{
		LyFieldController fieldController = (LyFieldController)_parent;
		LyFieldInfo		  fieldInfo = fieldController.getFieldInfo();

		if ((fieldInfo instanceof LyRelationFieldInfo)
				&& ((LyRelationFieldInfo)fieldInfo).isToolbarDisabled())
			return null;

		int n = ((actions == null) ? 0 : actions.getSize());

		if (n == 0)
			return null;

		int		  i, j;
		LyAction  action, action2;
		LySession session = _parent.getSession();

		// for each action, check if it is valid
		for (i = n - 1; i >= 0; i--)
		{
			action = actions.getInfo(i);
			if (!session.validateAction(action, _parent))
				// the current action is not valid, remove it from the list
				actions.removeInfo(action);
		}

		n = ((actions == null) ? 0 : actions.getSize());
		if (n == 0)
			return null;

		LyToolBar     toolBar = new LyToolBar();
		LyTool		  tool = null;
		// Keep subtoolbars table while building transverse actions
		LyToolBar     subToolBar;
		LyArea		  area;
		LyClassInfo   classInfo, classInfo2;
		LyEnvironment environment = _parent.getEnvironment();
		Object		  data;

		toolBar.setController(_parent.getId());
		if (fieldInfo.isMultiple())
			toolBar.setFill(LyToolBar.Fill.VERTICAL);
		else
			toolBar.setFill(LyToolBar.Fill.HORIZONTAL);

		for (i = 0; i < n; i++)
		{
			action = actions.getInfo(i);
			data = actions.getData(i);

			if (data instanceof LyClassInfo)
				classInfo = (LyClassInfo)data;
			else if (data instanceof LyStructFieldInfo)
				classInfo = ((LyStructFieldInfo)data).getClassInfo();
			else
				classInfo = null;

			// Check if we have multiple occurences of the same action image
			// if so, build a tool with a sub menu
			subToolBar = null;
			if (!(fieldInfo instanceof LyRelationFieldInfo)
					|| (((LyRelationFieldInfo)fieldInfo).getClassField() == null)
					|| ignoreRelationClassField())
			{
				for (j = i + 1; j < n; j++)
				{
					action2 = actions.getInfo(j);

					if (_parent.getImage(action).equals(_parent.getImage(action2))
							&& environment.testResource("LY_GROUP_ACTIONS_BY_IMAGE"))
					{
						if (subToolBar == null)
						{
							subToolBar = new LyToolBar();
							subToolBar.setController(_parent.getId());

							if (environment.testResource("LY_WRITE_TOOL_LABEL"))
								subToolBar.setFill(LyToolBar.Fill.VERTICAL);
							else
							{
								if (fieldInfo.isMultiple())
									subToolBar.setFill(LyToolBar.Fill.VERTICAL);
								else
									subToolBar.setFill(LyToolBar.Fill.HORIZONTAL);
							}

							if (classInfo != null)
							{
								tool = buildTool(action, true,
										classInfo.getId()
										+ LyController.ID_SEP + fieldInfo.getId(), null);
								tool.setLabel(action.getAlias() + " : " + classInfo.getAlias());
                                if (action.getTip() == null)
                                    tool.setTooltip(tool.getLabel());
								tool.setImage(_parent.getImage(classInfo));
							}
							else
								tool = buildTool(action, true,
										LyFieldSetController.STANDARD_CLASSINFO
										+ LyController.ID_SEP + fieldInfo.getId(), null);
							subToolBar.addTool(tool);
						}

						classInfo2 = actions.getData(j);
						if (classInfo2 != null)
						{
							tool = buildTool(action2, true,
									classInfo2.getId() + LyController.ID_SEP + fieldInfo.getId(), null);
							tool.setLabel(action2.getAlias() + " : " + classInfo2.getAlias());
                            if (action2.getTip() == null)
                                tool.setTooltip(tool.getLabel());
							tool.setImage(_parent.getImage(classInfo2));
						}
						else
							tool = buildTool(action2, true,
									LyFieldSetController.STANDARD_CLASSINFO
									+ LyController.ID_SEP + fieldInfo.getId(), null);
						subToolBar.addTool(tool);
						actions.removeInfo(action2, classInfo2);
						j--;
						n--;
					}
				}

				tool = buildTool(action, true,
						((classInfo != null)
							? (classInfo.getId() + LyController.ID_SEP + fieldInfo.getId())
							: (LyFieldSetController.STANDARD_CLASSINFO + LyController.ID_SEP
								+ fieldInfo.getId())),
						environment.getImageSizeInfo(DEFAULT_TRANSVERSE_ICON_WIDTH,
							DEFAULT_TRANSVERSE_ICON_WIDTH));
			}
			else
			{
				for (j = i + 1; j < n; j++)
				{
					action2 = actions.getInfo(j);
					if (_parent.getImage(action).equals(_parent.getImage(action2)))
					{
						actions.removeInfo(action2, actions.getData(j));
						j--;
						n--;
					}
				}

				tool = buildTool(action, true,
						LyFieldSetController.STANDARD_CLASSINFO + LyController.ID_SEP
						+ fieldInfo.getId(),
						environment.getImageSizeInfo(DEFAULT_TRANSVERSE_ICON_WIDTH,
							DEFAULT_TRANSVERSE_ICON_WIDTH));
			}

			toolBar.addTool(tool);
			if (subToolBar != null)
			{
				area = new LyArea();
				area.setToolBar(subToolBar);
				tool.setComponent(area);
			}
			else if (classInfo != null)
			{
				// Check if we have another action with the same label
				for (j = 0; j < n; j++)
				{
					action2 = actions.getInfo(j);
					if ((action2 != action)
							&& action2.getName().equals(action.getName()))
						break;
				}

				if (j < n)
					tool.setLabel(action.getAlias() + LyFieldSetController.TRANSVERSE_ACTION_SEP
						+ classInfo.getAlias());
			}
		}

		return toolBar;
	} // end method buildToolBar

	/**
	 * Builds a tool description for a specified action.
	 *
	 * @param  action       the action for the tool
	 * @param  enable       enabled state for the tool
	 * @param  parameter    parameters for the associated action
	 * @param  iconSizeInfo size of the associated icon
	 * @return the tool
	 */
	public LyTool buildTool(LyAction action, boolean enable, String parameter,
		String iconSizeInfo)
	{
		String toolId = action.getId();

		if (parameter != null)
			toolId += LyController.ID_SEP + parameter;

		LyEnvironment env = _parent.getEnvironment();
		String		  label = (env.testResource("LY_WRITE_FIELD_TOOL_LABEL") ? action.getAlias()
																			 : null);
		LyTool		  tool = new LyTool(toolId, label);
		String		  iconPath = _parent.getImage(action);

		if (iconSizeInfo != null)
		{
			// Remove path info if there was one
			iconPath = env.getPathWithoutSize(iconPath);
			iconPath = iconPath + iconSizeInfo;
		}

		if (action.getTip() != null)
			tool.setTooltip(action.getTip());
		else
			tool.setTooltip(action.getName());
		tool.setImage(iconPath);
		tool.setEnabled(enable);
		tool.setDialog(action.hasMark(LyAction.MARK_DIALOG));

		return tool;
	} // end method buildTool

	/**
	 * Tells if we must ignore the class fields of relation while building toolbars.<br>
	 * This is used only for filter builders.
	 *
	 * @return true;
	 */
	boolean ignoreRelationClassField()
	{
		return false;
	}

	/**
	 * Accessor to the main form builder.
	 *
	 * @return the main form builder
	 */
	protected LyFormBuilder getFormBuilder()
	{
		LyFieldController parent = (LyFieldController)_parent;
		LyBuilder		  builder = parent.getParent().getBuilder();

		if (!(builder instanceof LyFormBuilder))
			return null;

		LyFormController form = (LyFormController)((LyFormBuilder)builder).getParent();
		LyFieldInfo		 fieldInfo = parent.getFieldInfo();

		if (form.getFieldController(fieldInfo) != parent)
			return null;
		else
			return (LyFormBuilder)builder;
	}

	/**
	 * Accessor to the current value of the field.
	 *
	 * @return the value of the fieldinfo
	 */
	protected Object getValue()
	{
		return null;
	}
} // end class LyFormFieldBuilder
