/**
 * LyContentReplacer.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.control.print;

import java.io.*;
import java.util.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;


/**
 * This class Reads a file (Ascii, XML, RTF, HTML ...) and replace some keys inside the file pattern
 * like $KEY$ to produce an output file with content.
 *
 * @version $Id: LyContentReplacer.java 34749 2012-03-16 12:28:52Z jlsaouli $
 * @author  Lyria
 */
public class LyContentReplacer
{
	/**
	 * File version
	 */
	public static final String		    __VERSION =
		"$Id: LyContentReplacer.java 34749 2012-03-16 12:28:52Z jlsaouli $";

	/** Escape encode character */
	public static String			    ESCAPE_ENCODE_SEP = "\\\\";
	/** Fieldinfo separator */
	public static String			    FIELDINFO_SEP = "::".intern();
	/** Field name separator */
	public static String			    FIELDNAME_SEP = "FIELDNAME_".intern();
	/** REPEAT_OBJECT_START Tag*/
	public static String			    REPEAT_OBJECT_START = "REPEAT_OBJECT_START".intern();
	/** REPEAT_OBJECT_ENDt Tag*/
	public static String			    REPEAT_OBJECT_END = "REPEAT_OBJECT_END".intern();
	/** IF_LAST_OBJECT_START tag */
	public static String			    IF_LAST_OBJECT_START = "IF_LAST_OBJECT_START".intern();
	/** IF_LAST_OBJECT_END Tag */
	public static String			    IF_LAST_OBJECT_END = "IF_LAST_OBJECT_END".intern();
	/** IF_NOT_LAST_OBJECT_START Tag */
	public static String			    IF_NOT_LAST_OBJECT_START = "IF_NOT_LAST_OBJECT_START"
			.intern();
	/** IF_NOT_LAST_OBJECT_END Tag */
	public static String			    IF_NOT_LAST_OBJECT_END = "IF_NOT_LAST_OBJECT_END".intern();
	/** IF_FIRST_OBJECT_START Tag */
	public static String			    IF_FIRST_OBJECT_START = "IF_FIRST_OBJECT_START".intern();
	/** IF_FIRST_OBJECT_END Tag */
	public static String			    IF_FIRST_OBJECT_END = "IF_FIRST_OBJECT_END".intern();
	/** IF_NOT_FIRST_OBJECT_START Tag */
	public static String			    IF_NOT_FIRST_OBJECT_START = "IF_NOT_FIRST_OBJECT_START"
			.intern();
	/** IF_NOT_FIRST_OBJECT_END Tag */
	public static String			    IF_NOT_FIRST_OBJECT_END = "IF_NOT_FIRST_OBJECT_END"
			.intern();
	/**  */
	public static String			    SORT_ID = "(SORT_ID=".intern();

	/** Replacing mode for encoding : text (no encoding). */
	public static final int			    MODE_TEXT = 0;

	/** Replacing mode for encoding : HTML (html encoding). */
	public static final int			    MODE_HTML = 1;

	/** Replacing mode for encoding : RTF (rtf encoding). */
	public static final int			    MODE_RTF = 2;

	/** Replacing mode for encoding : XML (xml encoding). */
	public static final int			    MODE_XML = 3;

	/**
	 * Display parameter to indicate what to display in the value of a field : default is value name
	 * (fieldValue.getName()).
	 */
	public static final int			    DISPLAY_VALUE_NAME = 1;

	/**
	 * Display parameter to indicate what to display in the value of a field : short name value
	 * (fieldValue.getShortName()).
	 */
	public static final int			    DISPLAY_VALUE_SHORT_NAME = 2;

	/**
	 * Display parameter to indicate what to display in the value of a field : internal value
	 * (fieldValue.getString()).
	 */
	public static final int			    DISPLAY_VALUE_INTERNAL = 3;

	/** Current environment. */
	protected LyEnvironment			    _environment;

	/** Current model (pattern). */
	protected String				    _modelFile;

	/** Current running mode. */
	protected int					    _mode;

	/** Root dictionnary. */
	protected Hashtable<String, String> _rootDictionnary;

	/**
	 * Constructor.
	 * @param environment  the current environment
	 */
	public LyContentReplacer(LyEnvironment environment)
	{
		_environment = environment;
	}

	/**
	 * Get the content fromt the model.
	 *
	 * @param  object current model
	 * @return the result as a String
	 */
	public String getContentFromModel(LyObject object)
	{
		if (object == null)
		{
			_environment.logError(
				"[LyContentReplacer] cannot replace content : current object is not specified.");
			return null;
		}
		LySimpleObjectList list = new LySimpleObjectList(object);

		return getContentFromModel(list);
	}

	/**
	 * Get the result of the transformation as a String.
	 *
	 * @param  objects current value to build replacement dictionnary
	 * @return the result as a String
	 */
	public String getContentFromModel(LyObjectList objects)
	{
		if (_modelFile == null)
		{
			_environment.logError(
				"[LyContentReplacer] cannot replace content : model file is not specified.");
			return null;
		}
		if (objects == null)
		{
			_environment.logError(
				"[LyContentReplacer] cannot replace content : current object list is not specified.");
			return null;
		}
		if (objects.getSize() <= 0)
		{
			_environment.logError(
				"[LyContentReplacer] cannot replace content : object list is empty.");
			return null;
		}

		StringWriter writer = new StringWriter();

		try
		{
			generateContentFromModel(objects, new PrintWriter(writer));
		}
		catch (IOException e)
		{
			_environment.writeException(e);
		}

		return writer.getBuffer().toString();
	} // end method getContentFromModel

	/**
	 * accessor to the original dictionnary.
	 *
	 * @return default dictionnary
	 */
	public Hashtable<String, String> getRootDictionnary()
	{
		if (_rootDictionnary == null)
			_rootDictionnary = new Hashtable<String, String>();
		return _rootDictionnary;
	}

	/**
	 * Generate the content for the specified object and write it to the output.
	 *
	 * @param  objects current objects used for content replacement
	 * @param  writer  current writer for the output
	 * @return number of objects generated in the output
	 * @throws IOException 
	 */
	public int generateContentFromModel(LyObjectList objects, PrintWriter writer)
		throws IOException
	{
		// Build an internal dictionnary for pattern matching
		LyClassInfo				  classInfo = objects.getObject(0).getClassInfo();
		Hashtable<String, String> dict = getRootDictionnary();
		int						  n;
		int						  page = 0;
		int						  nb_pages = 0;
		char					  tag = '@';
		boolean					  by_list = false;

		// First step : read model file :
		InputStreamReader		  s = null;

		String					  fileName = _environment.getFile(_modelFile);

		File					  f = new File(fileName);

		if (f.exists())
		{
			s = new FileReader(f);

			if (getMode() == MODE_XML)
			{
				InputStream is = new FileInputStream(f);

				if (is != null)
					s = new InputStreamReader(is, "UTF-8");
			}
		}

		if (s == null)
		{
			InputStream is = _environment.getResourceAsStream(_modelFile, null);

			if (is == null)
				is = _environment.getResourceAsStream(fileName, null);
			if (is != null)
				if (getMode() == MODE_XML)
					s = new InputStreamReader(is, "UTF-8");
				else
					s = new InputStreamReader(is);
		}

		StringBuffer model = new StringBuffer();

		if (s == null)
		{
			_environment.logWarning("[LyContentReplacer] model ressource not found in classpath : "
				+ _modelFile);
			return 0;
		}

		do
		{
			n = s.read();
			if (n >= 0)
				model.append((char)n);
		}
		while (n >= 0);

		s.close();
		f = null; //free
		s = null; // free

		dict.put("CURRENT_DATE", LyDateFormatter.format(new Date(), "dd/MM/yyyy")); // DATE
		dict.put("P", Integer.toString(page)); // PAGE
		dict.put("N", Integer.toString(nb_pages)); // NBPAGES
		dict.put("CLS", classInfo.getName()); // CLASSE

		if (page == nb_pages)
		{
			nb_pages = 0;
			page = 0;
		}

		// We fetch all non hidden attributes and we use the import field.

		// Fill dictionnary.
		LyObject object;
		int		 i;

		for (i = 0, n = objects.getSize(); i < n; i++)
		{
			object = objects.getObject(i);
			// default tags for object name and identifiers
			if (by_list)
			{
				dict.put("OID" + i, object.getId());
				dict.put("OIFS" + i, object.getObjectId().getFullString());
				dict.put("ONAME" + i, object.getName());
			}
			else
			{
				dict.put("OID", object.getId());
				dict.put("OIFS", object.getObjectId().getFullString());
				dict.put("ONAME", object.getName());
			}
		}

		matchPatterns(objects, model.toString(), writer, dict, tag, null, 0);
		return i;
	} // end method generateContentFromModel

	/**
	 * Accessor to Current model file.
	 *
	 * @return current model file
	 */
	public String getModelFile()
	{
		return _modelFile;
	}

	/**
	 * Accessor to Current model file.
	 *
	 * @param modelFile current model file
	 */
	public void setModelFile(String modelFile)
	{
		_modelFile = modelFile;

		_mode = MODE_TEXT;
		if (modelFile != null)
		{
			String s = modelFile.toLowerCase();

			if (s.indexOf(".htm") >= 0)
				_mode = MODE_HTML;
			else if (s.indexOf(".rtf") >= 0)
				_mode = MODE_RTF;
			else if (s.indexOf(".xml") >= 0)
				_mode = MODE_XML;
		}
	}

	/**
	 * Utility method to know if a fieldInfo is displayed in this result or not. Returns false if
	 * the fieldInfo is marked with LyFieldInfo.MARK_SECRET
	 *
	 * <P>These fields should never be shown to the user.</P>
	 *
	 * @param  fieldInfo the fieldInfo which is tested for addition in the view.
	 * @return true if the field should be added, false if it should be skipped.
	 */
	public boolean isDisplayed(LyFieldInfo fieldInfo)
	{
		return ((!fieldInfo.hasMark(LyFieldInfo.MARK_SECRET))
				&& (!fieldInfo.hasMark(LyFieldInfo.MARK_HIDDEN)));
	}

	/**
	 * Format an object into a two columns table wth the specified width.
	 *
	 * @param  object       object to format
	 * @param  labels       column for labels.
	 * @param  values       column for values
	 * @param  labels_width with of the label column
	 * @param  values_width with of the value column
	 * @return the number of rows in the table
	 */
/*public int formatObject(LyObject object, StringBuffer labels, StringBuffer values,
                                                        int labels_width, int values_width)
        {
                if ((object == null) || (labels == null) || (values == null))
                        return 0;

                LyClassInfo                             classInfo = object.getClassInfo();
                LyFieldInfo                             fieldInfo;
                LyValue                                 value;
                int                                             i, na, n1, n2, nl = 0;
                String                                  str;
                String                                  name;

                for (i = 0, na = classInfo.getFieldInfoCount(); i < na; i++)
                {
                        fieldInfo = classInfo.getFieldInfo(i);

                        str = fieldInfo.getName() + " :\n";
                        str = format(str, labels_width);
                        n1 = getNbLines(str);
                        labels.append(str);

                        value = object.getFieldValue(fieldInfo);
                        if (value == null)
                                n2 = 0;
                        else
                        {
                                str = getString(value, "\n");
                                if ((str == null) || (str.length() == 0))
                                        n2 = 1;
                                else
                                {
                                        str = format(str, values_width);
                                        n2 = getNbLines(str);
                                        values.append(str);
                                }
                                values.append('\n');
                        }

                        if (n1 > n2)
                                for (; n2 < n1; n2++)
                                        values.append('\n');
                                else if (n2 > n1)
                                        for (; n1 < n2; n1++)
                                                labels.append('\n');

                        nl += n1;
                }

                return nl;
        }

*/
	/**
	 * Utility method to get a multipleValue as a String with a separator.
	 *
	 * @param  value     value to convert as a String
	 * @param  separator separator between values
	 * @return the result String
	 */
	public String getString(LyValue value, String separator)
	{
		if (value instanceof LyRelationValue)
		{
			LyRelationValue rValue = (LyRelationValue)value;
			StringBuffer    sb = new StringBuffer();

			if (rValue.getValueCount() < 0)
				return null;

			for (int i = 0; i < rValue.getValueCount(); i++)
			{
				if (i > 0)
					sb.append(separator);
				sb.append(rValue.getObject(i).getName());
			}
			return sb.toString();
		}
		else if (value instanceof LyChoiceValue)
		{
			LyChoiceValue cValue = (LyChoiceValue)value;
			StringBuffer  sb = new StringBuffer();

			if (cValue.getValueCount() < 0)
				return null;

			for (int i = 0; i < cValue.getValueCount(); i++)
			{
				if (i > 0)
					sb.append(separator);
				sb.append(cValue.getString(i));
			}
			return sb.toString();

		}
		else
			return value.getString();
	} // end method getString

	/**
	 * Format a string to a maximum width with new lines inserted every 'step' position.
	 *
	 * @param  text         char string to display
	 * @param  step         nmber of characters for new line insert
	 * @param  fill         indicates if the string is filled at the end with filling_char
	 * @param  filling_char filling char for the end of the String
	 * @param  replace_cr   indicate if \n are replaced by blanks or not.
	 * @return the formatted string or null if an error occured
	 */
	public String format(String text, int step, boolean fill,
		char filling_char, boolean replace_cr)
	{
		if (text == null)
			return null;

		if (text.length() == 0)
			return null;

		if (replace_cr)
			text = text.replace('\n', ' ');

		int			 text_lg = text.length();
		StringBuffer result = new StringBuffer(2 * (text_lg + (text_lg / step)));
		String		 line;
		String		 punctuations = ":?., ;!";
		int			 cursor;
		int			 ptr, l, i;
		char		 cut, next;

		for (cursor = 0; (cursor + step) < text_lg; cursor += step)
		{
			line = text.substring(cursor, cursor + step);
			ptr = line.indexOf('\n');

			if (ptr == -1) // No CR for this line
			{
				// We cut on a blanc if it is not followed by a punctuation mark.
				cut = text.charAt(cursor + step);
				next = (((cursor + step + 1) < text_lg) ? text.charAt(cursor + step + 1) : '\0');

				if (!Character.isWhitespace(cut)
						|| (punctuations.indexOf(next) >= 0))
				{
					l = step;
					while ((l > 0)
							&& (!Character.isWhitespace(line.charAt(l - 1))
								|| (punctuations.indexOf(cut) >= 0)))
					{
						cut = line.charAt(l - 1);
						cursor--;
						l--;
					}
					if (l == 0)
						// not possible to render the whole word with the specified step.
						return text;

					result.append(line.substring(0, l));
				}
				else
				{
					while (((cursor + step) < text_lg)
							&& Character.isWhitespace(text.charAt(cursor + step)))
						cursor++;
					result.append(line);
					l = step;
				}
			}
			else
			{
				result.append(line.substring(0, ptr));
				cursor += ptr + 1 - step;
				l = ptr;
			}

			if (fill)
			{
				for (i = step - l; i > 0; i--)
					result.append(filling_char);
			}
			result.append('\n');
		}

		// copy last token
		result.append(text.substring(cursor, text_lg));
		if (fill)
		{
			for (i = step - text_lg + cursor; i > 0; i--)
				result.append(filling_char);
		}

		return result.toString();
	} // end method format

	/**
	 * Format the given text
	 * @param text the text
	 * @param step a step
	 * @return the formatted text
	 */
	public String format(String text, int step)
	{
		return format(text, step, false, '\0', false);
	}

	/**
	 * Formatage d'une liste d'objets grs en un tableau de deux lignes labels et valeurs.
	 *
	 * @param  object     Objet gr  formater.
	 * @param  nb_objects Nombre d'objets  traiter.
	 * @param  attributes Liste des attributs  considrer.
	 * @param  labels     Colonne des labels.
	 * @param  values     Colonne des valeurs.
	 * @return Retourne le nombre de lignes du tableau
	 */
/*public void formatObjects(LyObjectList objects, int nb_objects,
                StringBuffer labels, StringBuffer values)
        {
                if ((objects == null) || (nb_objects <= 0)
                                 || (labels == null) || (values == null))
                        return;

                int                                             i, j, l;
                int                                             no = objects.getSize();
                LyClassInfo                             classInfo = objects.getObject(0).getClassInfo();
                LyFieldInfo                             fieldInfo;
                LyObject                                object;
                LyValue                                 value;
                String                                  str, name;
                int                                             nc= classInfo.getFieldInfoCount();

                int[]                                   column_widths = new int[nc];

                // Recalculer pour chaque colonne la largeur maximale.
                for (i = 0; i < nc; i++)
                {
                        fieldInfo = classInfo.getFieldInfo(i);

                        if (!isDisplayed(fieldInfo))
                                continue;

                        column_widths[i] = fieldInfo.getName().length() + 1;

                        for (j = 0; (j < nb_objects) && (j < no); j++)
                        {
                                object = objects.getObject(j);
                                value = object.getFieldValue(fieldInfo);
                                str = getString(value, " ");
                                l = (str == null ? 1 : str.length() + 1);
                                if (l > column_widths[i])
                                        column_widths[i] = l;
                        }
                }

                // Construire la chane des labels
                for (i = 0; i < nc; i++)
                {
                        fieldInfo = classInfo.getFieldInfo(i);

                        if (!isDisplayed(fieldInfo))
                                continue;

                        name = fieldInfo.getName();
                        if (fieldInfo instanceof LyNumberFieldInfo)
                        {
                                for (l = name.length(); l < column_widths[i] - 1; l++)
                                        labels.append(' ');
                                labels.append(name);
                                labels.append(' ');
                        }
                        else
                        {
                                labels.append(name);
                                for (l = name.length(); l < column_widths[i]; l++)
                                        labels.append(' ');
                        }
                }

                // Construire la chane des valeurs
                for (i = 0; (i < nb_objects) && (i < no); i++)
                {
                        object = objects.getObject(i);

                        for (j = 0; j < nc; j++)
                        {
                                fieldInfo = classInfo.getFieldInfo(i);

                                if (!isDisplayed(fieldInfo))
                                        continue;

                                value = object.getFieldValue(fieldInfo);

                                if ((value == null) || (value.isNull()))
                                {
                                        for (l = 0; l < column_widths[j]; l++)
                                                values.append(' ');
                                        continue;
                                }

                                str = getString(value, " ");

                                l = (str == null ? 0 : str.length());

                                if (fieldInfo instanceof LyNumberFieldInfo)
                                {
                                        for (; l < column_widths[j] - 1; l++)
                                                values.append(' ');
                                        if (str != null)
                                                values.append(str);
                                        values.append(' ');
                                }
                                else
                                {
                                        if (str != null)
                                                values.append(str);
                                        for (; l < column_widths[j]; l++)
                                                values.append(' ');
                                }
                        }

                        values.append('\n');
                }
        }
*/
	/**
	 * Cette fonction permet de trouver la portion de chane de longueur maximale dlimite par un
	 * retour-charriot  l'intrieur de la chane de caractres passe en argument.
	 *
	 * @param  text Chane de caractres  tudier.
	 * @return Cette fonction retourne la longueur maximale d'un segment de chane dlimit par un
	 *         retour-charriot.
	 */
/*public static int getNbLines(String text)
        {
                if (text == null)
                        return 0;

                int                     idx, idx2, n;
                int                     l = text.length();

                for (n = 0, idx = 0, idx2 = 0;
                        (idx < l) && ((idx2 = text.indexOf('\n', idx)) >= 0);
                        n++, idx = idx2 + 1);
                if (idx < l) // La chane n'est pas termine par un retour-charriot.
                        n++;

                return n;
        }
*/
	/**
	 * Enables pattern matching in a file. The output file is identical as the original file but
	 * with replacement of marks with the value found in a dictionnary Marks use the following
	 * syntax
	 *
	 * <UL>
	 *   <LI>begin_mark<pattern>|lg|+end_mark</LI>
	 *   <LI>begin_mark and end_mark are given as parameters</LI>
	 *   <LI>pattern is the key associated to a value in the dictionnary</LI>
	 *   <LI>lg and + are optionals. Indicate a maximum length to substitue.</LI>
	 * </UL>
	 *
	 * Substitution may keep on processing if : if string length is greater than lg, the string is
	 * cut.<BR>
	 * If + is found then the remaining chars in the string will be used in the next patter. If not,
	 * the string is cut.<BR>
	 * If lg is not specified, the entire string is used for substitution.
	 * @param objects              a list of objects
	 *
	 * @param  model               original content for substitution
	 * @param  writer              result stream
	 * @param  dict                pattern matching dictionnary
	 * @param  tag                 tag delimiter -it is *2 in the model, like '@@')
	 * @param  parentContextStack  this is the stack of context (null by default at root level) for
	 *                             loops.
	 * @param  writeOutputDisabled counters for write disable requesters : must be 0 at root level
	 * @return nb of char read
	 */
	public int matchPatterns(LyObjectList objects, String model, PrintWriter writer,
		Hashtable<String, String> dict, char tag, Vector<LyObjectList> parentContextStack, int writeOutputDisabled)
	{
		if ((model == null) || (writer == null))
			return 0;
		
		StringBuffer vstr;
		String[]     list, list2;
		String		 str, tmp, vstr_value, vstr_key;
		int			 lg_max, i;
		int			 ch = 1;
		char		 eol = '\n';
		int			 idx = 0;
		int			 n = model.length();
		boolean		 tagOpening = false;

		try
		{
			while (idx < n)
			{
				// Read until next tag
				tagOpening = false;
				while (idx < n)
				{
					ch = model.charAt(idx);
					idx++;
					if (ch == tag)
					{
						if (tagOpening)
							break; // Got '@@'
						else
							tagOpening = true;
					}
					else
					{
						if (tagOpening)
						{
							if (writeOutputDisabled <= 0)
								writer.write(tag); // One was kept in buffer
							tagOpening = false;
						}
						if (writeOutputDisabled <= 0)
							writer.write(ch);
					}
				}

				// Check end of file.
				if (idx >= n)
					break;

				// Fetch the string between the tags
				vstr = new StringBuffer();
				tagOpening = false;
				while (idx < n)
				{
					ch = model.charAt(idx);
					idx++;
					if (ch == tag)
					{
						if (tagOpening)
							break; // got '@@'
						else
							tagOpening = true;
					}
					else
					{
						if (tagOpening)
						{
							vstr.append(tag); // One was kept in buffer
							tagOpening = false;
						}
						vstr.append((char)ch);
					}
				}

				// Check end of file.
				if (idx >= n)
					break;

				// Check if the size is specified
				list = LyString.stringToList(vstr.toString(), '|');

				vstr_key = list[0];
				vstr_value = dict.get(vstr_key);

				int display = DISPLAY_VALUE_NAME;

				if (list.length > 2)
				{
					if ((list[2] != null) && (list[2].equals("SHORT_NAME")))
						display = DISPLAY_VALUE_SHORT_NAME;
					else if ((list[2] != null) && (list[2].equals("INTERNAL")))
						display = DISPLAY_VALUE_INTERNAL;
				}

				if (vstr_value == null)
				{
					String[] keys = LyString.stringToList(vstr_key, FIELDINFO_SEP);

					vstr_value = getValueFromInterpreter(keys, objects, parentContextStack, 0,
							display);

					if (vstr_value != null)
						vstr_value = vstr_value.intern();
				}

				if (vstr_value == REPEAT_OBJECT_START)
				{
					LySimpleObjectList		  subContext;
					LyObject				  object;
					Hashtable<String, String> subDict;
					int						  nbRead = 0;
					Vector<LyObjectList>	  subParentContextStack = new Vector<LyObjectList>();

					// Copy current parent stack :
					if (parentContextStack != null)
					{
						for (int i1 = 0; i1 < objects.getSize(); i1++)
							subParentContextStack.addElement(parentContextStack.elementAt(i1));
					}

					// add parent context to the stack
					subParentContextStack.addElement(objects);

					// Check if we have a relation specified to repeat (otherwise repeat on current context)
					String[]     keys = LyString.stringToList(vstr_key, FIELDINFO_SEP);
					LyObjectList repeatContext;

					if (keys.length == 1)
						repeatContext = objects;
					else
					{
						// This MUST be a relation value
						String[] subKeys = new String[keys.length - 1];

						for (int i1 = 0; i1 < (keys.length - 1); i1++)
							subKeys[i1] = keys[i1 + 1];
						repeatContext = getObjectListFromInterpreter(subKeys, objects, 0);
					}

					// No recurse in repeatable structure
					for (int i1 = 0; (repeatContext != null) && (i1 < repeatContext.getSize());
							i1++)
					{
						object = repeatContext.getObject(i1);
						subContext = new LySimpleObjectList();
						subContext.addObject(object);

						// Clone dictionnary context
						subDict = new Hashtable<String, String>(dict);
						subDict.put("ONAME", object.getName());

						if (i1 > 0)
							subDict.remove("LY_FIRST_OBJECT");
						else
							subDict.put("LY_FIRST_OBJECT", "true");
						if (i1 < (repeatContext.getSize() - 1))
							subDict.remove("LY_LAST_OBJECT");
						else
							subDict.put("LY_LAST_OBJECT", "true");

						nbRead = matchPatterns(subContext, model.substring(idx), writer, subDict,
								tag, subParentContextStack, writeOutputDisabled);
					}
					//  if nothing to replace, skip section
					if ((repeatContext == null) || (repeatContext.getSize() <= 0))
					{
						int		posRepeatObjectEnd = model.indexOf(REPEAT_OBJECT_END, idx);
						int		posRepeatObjectEndFinal = model.indexOf(tag, posRepeatObjectEnd);
						
						nbRead = posRepeatObjectEndFinal - idx + 2;
					}
					idx = idx + nbRead; // Skip repeatable section now
					continue;
				}
				if (vstr_value == REPEAT_OBJECT_END)
					// Exiting loop
					return idx;
				if (vstr_value == IF_LAST_OBJECT_START)
				{
					if (dict.get("LY_LAST_OBJECT") == null)
						writeOutputDisabled++;
					vstr_value = "";
				}
				if (vstr_value == IF_LAST_OBJECT_END)
				{
					if (dict.get("LY_LAST_OBJECT") == null)
						writeOutputDisabled--;
					vstr_value = "";
				}
				if (vstr_value == IF_NOT_LAST_OBJECT_START)
				{
					if (dict.get("LY_LAST_OBJECT") != null)
						writeOutputDisabled++;
					vstr_value = "";
				}
				if (vstr_value == IF_NOT_LAST_OBJECT_END)
				{
					if (dict.get("LY_LAST_OBJECT") != null)
						writeOutputDisabled--;
					vstr_value = "";
				}
				if (vstr_value == IF_FIRST_OBJECT_START)
				{
					if (dict.get("LY_FIRST_OBJECT") == null)
						writeOutputDisabled++;
					vstr_value = "";
				}
				if (vstr_value == IF_FIRST_OBJECT_END)
				{
					if (dict.get("LY_FIRST_OBJECT") == null)
						writeOutputDisabled--;
					vstr_value = "";
				}
				if (vstr_value == IF_NOT_FIRST_OBJECT_START)
				{
					if (dict.get("LY_FIRST_OBJECT") != null)
						writeOutputDisabled++;
					vstr_value = "";
				}
				if (vstr_value == IF_NOT_FIRST_OBJECT_END)
				{
					if (dict.get("LY_FIRST_OBJECT") == null)
						writeOutputDisabled--;
					vstr_value = "";
				}

				if (vstr_value == null)
				{
					if (writeOutputDisabled <= 0)
					{
						// Write the tag two times.
						writer.write(tag);
						writer.write(tag);

						// Write the key.
						for (i = 0; i < vstr_key.length(); i++)
							writer.write(vstr_key.charAt(i));

						// Write the tag two times.
						writer.write(tag);
						writer.write(tag);
					}
					continue;
				}

				if ((list.length == 1)
						|| ((list.length > 1) && (list[1].length() <= 0)))
				{
					// No max length specified : write all.
					if (writeOutputDisabled <= 0)
					{
						str = encodeString(vstr_value);
						writer.write(str);
					}
				}
				else if (list.length > 1)
				{
					// list[1] contains the maximum display length.
					lg_max = Integer.parseInt(list[1]);

					if (vstr_value.length() > lg_max)
					{
						// We keep only lg_max chars.
						tmp = format(vstr_value, lg_max, false, '\0', false);
						list2 = LyString.stringToList(tmp, eol);

						// Encode the string for display.
						if (writeOutputDisabled <= 0)
						{
							str = encodeString(list2[0]);
							for (i = 0; i < str.length(); i++)
								writer.write(str.charAt(i));
						}
						// rebuild the string for display.
						vstr_value = new String();

						for (i = 1; i < list2.length; i++)
						{
							if (i != 1)
								vstr_value += eol;
							vstr_value += list2[i];
						}
					}
					else
					{
						// Keep all.
						list2 = LyString.stringToList(vstr_value, eol);

						// Encode the string for display.
						if (writeOutputDisabled <= 0)
						{
							str = encodeString(list2[0]);
							for (i = 0; i < str.length(); i++)
								writer.write(str.charAt(i));
						}
						// rebuild the string for display.
						vstr_value = new String();
						for (i = 1; i < list2.length; i++)
						{
							if (i != 1)
								vstr_value += eol;
							vstr_value += list2[i];
						}
					}
				}
				//dict.remove(vstr_key);
				//dict.put(vstr_key, vstr_value);
			}
		}
		catch (Exception e)
		{
			_environment.writeException(e);
		}
		return idx;
	} // end method matchPatterns

	/**
	 * get the value from the interpreter within current context.
	 *
	 * @param  keys     the tags to extract from context
	 * @param  objects current context
	 * @param  parentContextStack the stack of all parent context (to go up with '..' in interpreter)
	 * @param  depth   current depth in analysis of keys
	 * @param  display indicates what to display:one of the DISPLAY_VALUE_xxx constants
	 * @return the value or the tag if it makes no sense or "" if the value is empty.
	 */
	public String getValueFromInterpreter(String[] keys, LyObjectList objects,
		Vector<LyObjectList> parentContextStack, int depth, int display)
	{
		if ((objects == null)
    	        || (objects.getSize() <= 0)
    	        || (keys == null)
    			|| (keys.length <= depth))
			return "";
		
		// Try to decode
		LyObject    currentObject;
		LyClassInfo classInfo;
		LyFieldInfo fieldInfo;
		LyValue     value = null;
		String	    subKey, subKeyNoIndex;
		int		    index;

		// Parse info from key
		subKey = keys[depth];
		index = getIndexFromKey(subKey);
		subKeyNoIndex = clearIndexFromKey(subKey);

		if (subKeyNoIndex.startsWith(REPEAT_OBJECT_START))
			return REPEAT_OBJECT_START; // Loop mark found
		if (subKeyNoIndex.startsWith(REPEAT_OBJECT_END))
			return REPEAT_OBJECT_END; // Loop mark found
		if (subKeyNoIndex.startsWith(IF_LAST_OBJECT_START))
			return IF_LAST_OBJECT_START; // If mark found
		if (subKeyNoIndex.startsWith(IF_LAST_OBJECT_END))
			return IF_LAST_OBJECT_END; // If mark found
		if (subKeyNoIndex.startsWith(IF_NOT_LAST_OBJECT_START))
			return IF_NOT_LAST_OBJECT_START; // If mark found
		if (subKeyNoIndex.startsWith(IF_NOT_LAST_OBJECT_END))
			return IF_NOT_LAST_OBJECT_END; // If mark found
		if (subKeyNoIndex.startsWith(IF_FIRST_OBJECT_START))
			return IF_FIRST_OBJECT_START; // If mark found
		if (subKeyNoIndex.startsWith(IF_FIRST_OBJECT_END))
			return IF_FIRST_OBJECT_END; // If mark found
		if (subKeyNoIndex.startsWith(IF_NOT_FIRST_OBJECT_START))
			return IF_NOT_FIRST_OBJECT_START; // If mark found
		if (subKeyNoIndex.startsWith(IF_NOT_FIRST_OBJECT_END))
			return IF_NOT_FIRST_OBJECT_END; // If mark found

		if (subKeyNoIndex.equals(".."))
		{
			// Go up in context stack
			if ((parentContextStack == null) || (parentContextStack.size() == 0))
				_environment.logError("Cant resolve key for parent '..' : " + subKeyNoIndex
					+ " context is empy !");
			LyObjectList parentContext = parentContextStack.lastElement();
			// Rebuild parentContextStack
			Vector<LyObjectList>		 newParentContextStack = new Vector<LyObjectList>(parentContextStack);

			newParentContextStack.removeElement(parentContext);
			// Recurse
			return getValueFromInterpreter(keys, parentContext, newParentContextStack, depth + 1,
				display);
		}

		if (index >= 0)
			currentObject = objects.getObject(index);
		else
			currentObject = objects.getObject(0); // in loop

		// try to fetch the fieldinfo
		classInfo = currentObject.getClassInfo();
		fieldInfo = classInfo.getFieldInfo(subKeyNoIndex, false);

		// Maybee this is the name (the label) of the field we ae looking for
		if (fieldInfo == null)
		{
			String fieldId;

			if (subKeyNoIndex.startsWith(FIELDNAME_SEP)) // Got it
			{
				fieldId = subKeyNoIndex.substring(FIELDNAME_SEP.length());

				fieldInfo = classInfo.getFieldInfo(fieldId, false);

				if (fieldInfo != null)
					return fieldInfo.getName();
			}
		}
		// Failed to field the fieldInfo : likely an error in our template file
		if (fieldInfo == null)
		{
			_environment.logError("Cant resolve key : " + subKeyNoIndex + " in classinfo "
				+ classInfo.getId() + " : " + classInfo.getName());
			return null;
		}

		value = currentObject.getFieldValue(fieldInfo);

		depth = depth + 1; // one level deeper
		if (depth == keys.length)
		{
			// last one ! We got it
			String resu;

			if (display == DISPLAY_VALUE_SHORT_NAME)
			{
				resu = ((LyFieldInfoValue)value).getShortName();
				resu = (resu != null) ? _environment.translate(resu) : resu;
			}
			else if (display == DISPLAY_VALUE_INTERNAL)
				resu = value.getString();
			else
			{
				resu = ((LyFieldInfoValue)value).getName(true);
				resu = (resu != null) ? _environment.translate(resu) : resu;
			}

			return ((resu == null) ? "" : resu);
		}
		else
		{
			// this MUST be a relationValue
			LyRelationValue		 relationValue = (LyRelationValue)value;
			LyObjectList		 subContext = relationValue.getObjectList();

			// Rebuild parentContextStack
			Vector<LyObjectList> newParentContextStack = (parentContextStack != null)
				? new Vector<LyObjectList>(parentContextStack) : new Vector<LyObjectList>();

			newParentContextStack.addElement(objects);

			return getValueFromInterpreter(keys, subContext, newParentContextStack, depth, display);
		}
	} // end method getValueFromInterpreter

	/**
	 * get the list of value from the interpreter within current context.
	 *
	 * @param  keys     the tag to extract from context
	 * @param  objects current context
	 * @param  depth   current depth in analysis of keys
	 * @return the value or the tag if it makes no sense or "" if the value is empty.
	 */
	public LyObjectList getObjectListFromInterpreter(String[] keys, LyObjectList objects, int depth)
	{
		if ((objects == null) || (objects.getSize() <= 0) || (keys == null)
				|| (keys.length <= depth))
			return (new LySimpleObjectList());
		// Try to decode
		LyObject	    currentObject;
		LyClassInfo     classInfo;
		LyFieldInfo     fieldInfo;
		LyRelationValue value = null;
		String		    subKey, subKeyNoIndex;
		int			    index;
		LyObjectList    result = null;
		LySort		    sort = null;

		// Parse info from key
		subKey = keys[depth];
		index = getIndexFromKey(subKey);
		subKeyNoIndex = clearIndexFromKey(subKey);

		if (index >= 0)
			currentObject = objects.getObject(index);
		else
			currentObject = objects.getObject(0); // in loop

		// Check if a sort is specified
		{
			int n1 = subKeyNoIndex.indexOf(SORT_ID), n2;

			if (n1 > 0)
			{
				n2 = subKeyNoIndex.indexOf(")", n1);

				if (n2 > 0)
				{
					String sortId = subKeyNoIndex.substring(n1 + SORT_ID.length(), n2);

					subKeyNoIndex = subKeyNoIndex.substring(0, n1);

					try
					{
						sort = (LySort)currentObject.getApplication().getInfo(sortId);
					}
					catch (Exception e)
					{
						_environment.logError("Can't resolve key to LySort : " + sortId
							+ " for relation " + subKeyNoIndex + "\n" + e.getMessage());
					}
				}
			}
		}
		// try to fetch the fieldinfo
		classInfo = currentObject.getClassInfo();
		fieldInfo = classInfo.getFieldInfo(subKeyNoIndex, false);

		// Failed to field the fieldInfo : likely an error in our template file
		if (!(fieldInfo instanceof LyRelationFieldInfo))
		{
			_environment.logError("Cant resolve key to LyRelationFieldInfo : " + subKeyNoIndex
				+ " in classinfo " + classInfo.getId() + " : " + classInfo.getName() + " found : "
				+ fieldInfo);
			return null;
		}

		value = (LyRelationValue)currentObject.getFieldValue(fieldInfo);

		depth = depth + 1; // one level deeper
		if (depth == keys.length)
			// last one ! We got it
			result = value.getObjectList();
		else
		{
			// Go deeper to find the list for the loop
			LyObjectList subContext = value.getObjectList();

			result = getObjectListFromInterpreter(keys, subContext, depth);
		}

		if ((result != null) && (sort != null))
			result.setSort(sort);

		return result;
	} // end method getObjectListFromInterpreter

	/**
	 * Get the specified index for the key index is like fieldInfoId[index].
	 *
	 * @param  key the key to parse
	 * @return 0 if first index, -1 if within loop
	 */
	public int getIndexFromKey(String key)
	{
		if (key == null)
			return 0;
		int n1 = key.indexOf('[');

		if (n1 > 0)
		{
			int n2 = key.indexOf(']', n1);

			if (n2 > 0)
			{
				String s = key.substring(n1 + 1, n2);

				if (s.equals("*"))
					return -1; // within a loop
				int index = LyTools.intFromString(s);

				if (index >= 0)
					return index;
				else
					_environment.logError(" Index is not understood : " + s);
			}
		}
		return 0;

	}

	/**
	 * Get the Key name without the index information index is like fieldInfoId[index].
	 *
	 * @param  key the key to parse
	 * @return the key
	 */
	public String clearIndexFromKey(String key)
	{
		if (key == null)
			return key;

		int n1 = key.indexOf('[');

		if (n1 > 0)
		{
			int n2 = key.indexOf(']', n1);

			if (n2 > 0)
				return key.substring(0, n1);
		}
		return key;

	}

	/**
	 * Encode the string for the target output.
	 *
	 * @param  value the value to encode
	 * @return the encoded string depending on current MODE_xxx which is set by setMode or
	 *         autodetect
	 */
	public String encodeString(String value)
	{
		if (value.startsWith(ESCAPE_ENCODE_SEP))
			return value.substring(ESCAPE_ENCODE_SEP.length());
		if (_mode == MODE_HTML)
			return LyString.getHtmlString(value);
		else if (_mode == MODE_RTF)
		{
			String resu;

			resu = LyString.replace(value, "\r\n", "\\par ");
			resu = LyString.replace(resu, "\n", "\\par ");

			return resu;
		}
		else if (_mode == MODE_XML)
		{
			String resu;

			resu = LyString.replace(value, "&", "&amp;");
			resu = LyString.replace(resu, "<", "&lt;");
			resu = LyString.replace(resu, ">", "&gt;");

			return resu;
		}
		return value;
	}

	/**
	 * Accessor to Current mode file.
	 *
	 * @return current mode file
	 */
	public int getMode()
	{
		return _mode;
	}

} // end class LyContentReplacer
