/*
 * LyArrayValue.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import leon.info.*;


/**
 * This class offers methods for constructing, accessing and modifying values corresponding to a
 * LyArrayFieldInfo.
 *
 * @see     LyArrayFieldInfo
 * @see     LyObject
 * @version $Id: LyArrayValue.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public final class LyArrayValue extends LyFieldInfoValue
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyArrayValue.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Constructs an empty array value for the given array field.
	 *
	 * @param arrayFieldInfo Array fieldInfo for which a value is created.
	 */
	public LyArrayValue(LyArrayFieldInfo arrayFieldInfo)
	{
		super(arrayFieldInfo);
	}

	/**
	 * Constructs an array value for the given array field from an array.
	 *
	 * @param arrayFieldInfo Array fieldInfo for which a value is created.
	 * @param value          Value given as array.
	 */
	public LyArrayValue(LyArrayFieldInfo arrayFieldInfo, Object[] value)
	{
		super(arrayFieldInfo);

		if (value == null)
			return;

		// Check null values
		int i, n = value.length, nNull = 0;

		for (i = 0; i < n; i++)
		{
			if (value[i] == null)
				nNull++;
		}

		if (n == nNull)
			return;

		// Set array
		Object[] array = arrayFieldInfo.getArray(n - nNull);
		int		 j;

		for (i = 0, j = 0; i < n; i++)
		{
			if (value[i] != null)
			{
				array[j] = value[i];
				j++;
			}
		}

		setValue(array);
	}

	/**
	 * Constructs an array value for the given array field from an object.
	 *
	 * @param arrayFieldInfo Array fieldInfo for which a value is created.
	 * @param value          Value given as an object.
	 */
	public LyArrayValue(LyArrayFieldInfo arrayFieldInfo, Object value)
	{
		this(arrayFieldInfo, (Object[])value);
	}

	/**
	 * Constructs an array value for the given array field from an objects list.
	 *
	 * @param arrayFieldInfo Array fieldInfo for which a value is created.
	 * @param list           Value given as an objects list.
	 */
	public LyArrayValue(LyArrayFieldInfo arrayFieldInfo, LyObjectList list)
	{
		super(arrayFieldInfo, null);

		LyObjectId[] objectIds = null;
		int			 i, n = ((list == null) ? 0 : list.getSize());
		LyObject     object;

		if (n > 0)
		{
			objectIds = new LyObjectId[n];
			for (i = 0; i < n; i++)
			{
				object = list.getObject(i);
				objectIds[i] = object.getObjectId();
			}
		}

		setValue(objectIds);
	}

	/**
	 * Gets value of this array value as an array.
	 *
	 * @return An array corresponding to this value.
	 */
	public Object[] getArrayValue()
	{
		return (Object[])_value;
	}

	/**
	 * Gets value of this array as an array of LyFieldInfoValue. Each value in this array is a value
	 * corresponding to the type of the array.
	 *
	 * @return An array of LyFieldInfoValue corresponding to this value.
	 */
	public LyFieldInfoValue[] getArrayFieldValue()
	{
		int n = getValueCount();

		if (n == 0)
			return null;

		Object[]		   value = getArrayValue();
		LyFieldInfo		   fieldInfo = ((LyArrayFieldInfo)_field).getFieldInfo();
		LyFieldInfoValue[] result = new LyFieldInfoValue[n];

		for (int i = 0; i < n; i++)
			result[i] = (LyFieldInfoValue)fieldInfo.buildFieldValue(value[i]);

		return result;
	}

	/**
	 * Get the value at given index from value table.
	 *
	 * @param  index Index of the desired value.
	 * @return The value at the specified index.
	 */
	public Object getValue(int index)
	{
		int count = getValueCount();

		if (index < 0)
			throw new IllegalArgumentException("BF AppEngine : index can't be negative");
		if (index >= count)
			throw new IllegalArgumentException("BF AppEngine : index must be lower than " + count);

		Object[] value = getArrayValue();

		return value[index];
	}

	/**
	 * Get the field info value at given index fron value table.
	 *
	 * @param  index Index of the desired field info value
	 * @return The value at the specified index.
	 */
	public LyFieldInfoValue getFieldValue(int index)
	{
		int count = getValueCount();

		if (index < 0)
			throw new IllegalArgumentException("BF AppEngine : index can't be negative");
		if (index >= count)
			throw new IllegalArgumentException("BF AppEngine : index must be lower than " + count);

		Object[]    value = getArrayValue();
		LyFieldInfo fieldInfo = ((LyArrayFieldInfo)_field).getFieldInfo();

		return (LyFieldInfoValue)fieldInfo.buildFieldValue(value[index]);
	}

	/**
	 * Get the number of values found in this array value.
	 *
	 * @return Number of values.
	 */
	public int getValueCount()
	{
		Object[] value = getArrayValue();

		return ((value == null) ? 0 : value.length);
	}
} // end class LyArrayValue
