/*
 * LyBooleanValue.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyBooleanValue.java 34904 2012-03-21 15:55:41Z bangiole $
 */
package leon.data;

import leon.info.*;


/**
 * This class offers methods for constructing, accessing and modifying values corresponding to a
 * LyChoiceFieldInfo with a boolean type.
 *
 * @see LyChoiceFieldInfo
 * @see LyObject
 */
public class LyBooleanValue extends LyFieldInfoValue
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyBooleanValue.java 34904 2012-03-21 15:55:41Z bangiole $";

	/**
	 * Constructs an empty boolean value for the given choice field.
	 *
	 * @param choiceFieldInfo Choice fieldInfo for which a value is created.
	 */
	public LyBooleanValue(LyChoiceFieldInfo choiceFieldInfo)
	{
		super(choiceFieldInfo);
		if (choiceFieldInfo.getType() != LyChoiceFieldInfo.Type.BOOLEAN)
			throw new IllegalStateException("BF AppEngine : choice value for fieldInfo "
				+ choiceFieldInfo
				+ " can't be a boolean because the type of this choice is not boolean");
	}

	/**
	 * Constructs a boolean choice value for the given choice field from a string.
	 *
	 * @param choiceFieldInfo Choice fieldInfo for which a value is created.
	 * @param value           String value.
	 * @see   LyChoiceFieldInfo#parse
	 */
	public LyBooleanValue(LyChoiceFieldInfo choiceFieldInfo, String value)
	{
		super(choiceFieldInfo, value);
		if (choiceFieldInfo.getType() != LyChoiceFieldInfo.Type.BOOLEAN)
			throw new IllegalStateException("BF AppEngine : choice value for fieldInfo "
				+ choiceFieldInfo
				+ " can't be a boolean because the type of this choice is not boolean");
	}

	/**
	 * Constructs a boolean value for the given choice field from a boolean.
	 *
	 * @param choiceFieldInfo Choice fieldInfo for which a value is created.
	 * @param value           Value passed as a boolean.
	 */
	public LyBooleanValue(LyChoiceFieldInfo choiceFieldInfo, boolean value)
	{
		this(choiceFieldInfo, Boolean.valueOf(value));
	}

	/**
	 * Constructs a boolean value for the given choice field from a Boolean.
	 *
	 * @param choiceFieldInfo Choice fieldInfo for which a value is created.
	 * @param value           Value passed as a Boolean.
	 */
	public LyBooleanValue(LyChoiceFieldInfo choiceFieldInfo, Boolean value)
	{
		super(choiceFieldInfo, value);
		if (choiceFieldInfo.getType() != LyChoiceFieldInfo.Type.BOOLEAN)
			throw new IllegalStateException("BF AppEngine : choice value for fieldInfo "
				+ choiceFieldInfo
				+ " can't be a boolean because the type of this choice is not boolean");
	}

	/**
	 * Gets value of this boolean value as a Boolean.
	 *
	 * @return    A Boolean corresponding to this value.
	 * @exception IllegalStateException If the choice field is not a boolean.
	 */
	public Boolean getBooleanValue()
	{
		Object value = getValue();

		if ((value != null) && !(value instanceof Boolean))
			throw new IllegalStateException("BF AppEngine : choice value for fieldInfo "
				+ getFieldInfo() + " can't be returned as a Boolean");

		return (Boolean)value;
	}

	/**
	 * Gets value of this boolean value as a boolean (primitive type).
	 *
	 * @return A short corresponding to this value. Returns false if the boolean value is null.
	 */
	public boolean booleanValue()
	{
		Boolean bool = getBooleanValue();

		return ((bool == null) ? false : bool.booleanValue());
	}

} // end class LyBooleanValue
