/*
 * LyCache.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;

import leon.app.*;
import leon.info.*;
import leon.misc.*;


/**
 * This class s an abstract class that implements the generic behavior of a memory cache for
 * instances of objects managed by a LyDataSource. Basically, it only keeps in memory used objects
 * (using strong references).
 *
 * <p>This class is both used for storing logical or physical objects.</p>
 *
 * <p>The cache has an associated class that gives information about stored objects. This associated
 * class implements LyClass interface (for logical objects, this is an instance of LyClassInfo).</p>
 *
 * <p>Caches keeps several tables :</p>
 *
 * <ul>
 * <LI>An hashtable that contains pairs of object identifiers ans objects (LyObjectId, Object) to
 * easily retrieve an object from its id.</LI>
 * <LI>When logical objects are stored, an extra hashtable maintaining associations beetween
 * physical object and logical object, is managed. This table is used to find a logical object from
 * a physical object faster than a sequential loop on objects list.</LI>
 * <LI>When objects are deleted, they are first marked as removed and stored in a table of removed
 * objects and then they are finally deleted.</LI>
 * <LI>Another hashtable may also keep tracks of object ids that don't correspond to objects of the
 * cache to avoid repetitive exceptions.
 *
 * <p>In the current version, basic cache implementations are provided. More sophisticated ones will
 * be implemented in future versions.</p>
 * </LI>
 * </ul>
 *
 * @see      LyDataSource
 * @see      LyClass
 * @see      LyClassInfo
 * @see      LyInfoCache
 * @see      LyObjectId
 * @version  $Id: LyCache.java 41712 2012-12-05 08:51:26Z bangiole $
 * @author   Lyria
 */
public abstract class LyCache
{
    /** File version. */
    public static final String __VERSION = "$Id: LyCache.java 41712 2012-12-05 08:51:26Z bangiole $";

    /** Result types. */
    public enum Result
    {
        /** result list contains all results.*/
        TOTAL,
        /** result list contains no result.*/
        NULL,
        /** result list contains some of the results.*/
        PARTIAL
    }

    /**
     * To represent an unknown object in this cache. Used in association with a LyObjectId in the
     * hashTable _unknowObjects.
     */
    static final String UNKNOWN_OBJECT = "__Unknown_Object__".intern();

    /** Number of unknown objects references kept. */
    public static int NB_UNKNOWN_OBJECTS = 50;

    /**
     * Policy of the cache. This object contains the name of the Java class for the implementation
     * of the cache. Thsi class must inherit from LyCache.
     */
    protected LyInfoCache _policy;

    /** Class description of objects stored in this cache. */
    protected LyClass _class;

    /**
     * Hashtable of unknown ids. This data structure should be changed since only keys are used in
     * this table
     */
    protected Hashtable<LyObjectId, Object> _unknownObjects;

    /**
     * List of unknown ids. This list is used to keep the chronological order of the unknown objects
     * to remove the older when NB_UNKNOWN_OBJECTS is reached.
     */
    protected Vector<LyObjectId> _unknownList;

    /** List of removed objects. Keeps objects before removing them definitely. */
    protected Vector<Object> _removedObjects;

    /** Boolean that indicates if the cache is loaded or not. */
    protected boolean _loaded;

    /** Boolean that indicates if the cache can be marked as loaded or not. */
    protected boolean _allowLoaded;

    /**
     * Constructor of a cache of objects related to a given class description.
     *
     * @param  cacheClass  Class description of objects stored in the cache.
     * @param  policy      Description of th epolicy of thsi cache.
     */
    public LyCache(LyClass cacheClass, LyInfoCache policy)
    {
        // Initialize class attribute
        _class = cacheClass;
        _policy = policy;

        // At the begining there is no objects in the cache
        _unknownObjects = null;
        _unknownList = null;
        _removedObjects = null;
        _loaded = false;
        _allowLoaded = true;
    }

    /**
     * Retrieves an object in this cache from the given objectId.
     *
     * @param   objectId  the object id
     *
     * @return  the associated object
     */
    protected abstract Object getFromCache(LyObjectId objectId);

    /**
     * Adds an object in this cache.
     *
     * @param  objectId  the object id
     * @param  object    the object toadd to cache
     */
    protected abstract void putToCache(LyObjectId objectId, Object object);

    /**
     * Removes an object in this cache corresponding to the given objectId.
     *
     * @param  objectId  the id of the object to remove
     */
    protected abstract void removeFromCache(LyObjectId objectId);

    /**
     * Gets all objects stored in this cache as an enumeration.
     *
     * @return  Enumeration of objects found in the cache.
     */
    protected abstract Enumeration<Object> getElements();

    /**
     * Clear cache content.
     */
    protected abstract void clearCache();

    /**
     * Gets an object from its physical representation (peer). This method is only valid if stored
     * objects are logical objects (instances of LyObject).
     *
     * @param      peer  Physical representation of the searched object.
     *
     * @return     Object with the specified physical representation if it exists, otherwise null.
     *
     * @exception  IllegalStateException  If this cache is not a cache that keep logical objects.
     */
    public abstract LyObject getObjectFromPeer(Object peer);

    /**
     * Adds a pair (peer, object) in this cache.
     *
     * @param  peer    the peer object
     * @param  object  the logical object
     */
    protected abstract void putPeerObject(Object peer, LyObject object);

    /**
     * Removes objects associated to the given peer in this cache.
     *
     * @param  peer  the peer object to remove
     */
    protected abstract void removePeerObject(Object peer);

    /**
     * Return the number of objects within this cache that match the given filter (all objects of
     * the cache if filter is null). This method gives objects that are in the cache, no other
     * objects are loaded.
     *
     * @param   filter  Filter used for selecting objects. If filter is null, the total is returned.
     *
     * @return  Number of objects in the cache that match the filter.
     */
    public abstract int getSize(LyFilter filter);

    /**
     * Tests if this cache is empty or not (contains no object).
     *
     * @return  True if the cache is empty, false otherwise.
     */
    public abstract boolean isEmpty();

	/**
     * Setter for the allow loaded attribute.
     *
     * @param  allowLoaded  the value of the allowLoaded attribute
     */
    public abstract void allowLoaded(boolean allowLoaded);

    /**
     * Gets cache policy of this cache.
     *
     * @return  the policy of this cache.
     */
    public LyInfoCache getPolicy()
    {
        return _policy;
    }

    /**
     * Check cache.
     */
    protected void check()
    {
    }

    /**
     * Check if an object with the given id is found in cache.
     *
     * @param   objectId  the id of the object to check
     *
     * @return  True if an object is found, false otherwise.
     */
    protected boolean checkObject(LyObjectId objectId)
    {
        Object object = getFromCache(objectId);

        return (object != null);
    }

    /**
     * Check if the given object is found in cache.
     *
     * @param   object  the object to check
     *
     * @return  True if the given object is found, false otherwise.
     */
    protected boolean checkObject(Object object)
    {
        LyObjectId objectId = _class.getObjectId(object);
        Object     cachedObject = getFromCache(objectId);

        return (object == cachedObject);
    }

    /**
     * Gets an object in this cache by its identifier.
     *
     * @param      id  Object identifier.
     *
     * @return     Object found in cache with the given identifier, or null if no object is found
     *             and if the id was registered in unknown ids list.
     *
     * @exception  LyNoSuchObjectException  If no object of this cache has this identifier and if
     *                                      the given id has not been registered in unknown ids
     *                                      list.
     */
    public Object getObject(LyObjectId id) throws LyNoSuchObjectException
    {
        if (id == null)
            return null;

        Object object = getFromCache(id);

        if (object != null)
            return object;

        // Exception only if the given id has not been registered in unknown ids list
        if (!isUnknown(id))
            throw new LyNoSuchObjectException("BF AppEngine : no object with id " + id
                + " found in cache");

        return null;
    }

    /**
     * Adds a new object to this cache. An application should not use this method directly.
     *
     * @param      object  Object added in the cache.
     *
     * @exception  IllegalArgumentException  If the given object doesn't match the class specified
     *                                       for this cache.
     * @exception  IllegalStateException     If the given object is already in this cache.
     */
    protected void putObject(Object object)
    {
        // Check object class
        if (!_class.getObjectClass().isAssignableFrom(object.getClass()))
            throw new IllegalArgumentException(
                "BF AppEngine : Attempt to add an item within a cache with invalid class :"
                + object.getClass() + " Class :" + _class.getObjectClass());

        LyEnvironment environment = getEnvironment();

        if ((_class instanceof LyClassInfo) && (((LyObject)object).isDeleted()))
        {
            environment.logWarning("BF AppEngine : Try to add a deleted object to the cache "
                + object);
            return;
        }

        if ((_class instanceof LyClassInfo) && (((LyObject)object).isLocallyDeleted()))
        {
            environment.logWarning("BF AppEngine : Try to add a locally deleted object to the cache "
                + object);
            return;
        }

        // Retrieve object id.
        LyObjectId objectId = _class.getObjectId(object);

        /* Check if object is not already in cache */
        Object cachedObject = (objectId == null ? null : getFromCache(objectId));

        if ((cachedObject != null)
                && (cachedObject != object))
            environment.logWarning("BF AppEngine : Object " + objectId
                + " is already in cache " + cachedObject);

        // Remove object from unknown objects if needed
        if (isRemoved(object))
            _removedObjects.removeElement(object);

        if (objectId != null)
        {
	        // Put the object in the cache.
	        putToCache(objectId, object);
	        objectId.setObject(object);
        }

        // Remove id from unkwown ids if requiered
        if (isUnknown(objectId))
        {
            _unknownObjects.remove(objectId);
            _unknownList.remove(objectId);
        }

        // If associated class is a non local classInfo,
        // stores pair (physical object, logical object)
        if (_class instanceof LyClassInfo)
        {
            Object peer = ((LyObject)object).getPeer();

            if (peer != null)
                putPeerObject(peer, (LyObject)object);
        }
    } // end method putObject

    /**
     * Removes an object from the cache. An application should not use this method directly.
     *
     * @param      object  Object to remove from the cache.
     *
     * @exception  IllegalStateException  If the given object is not in the cache.
     */
    protected void startRemoveObject(Object object)
    {
        if (object == null)
            return;

        // Get the object id and its locks count.
        LyObjectId objectId = _class.getObjectId(object);

        // Remove the object from the cache, throw excpetion if it fails
        // Do not get the object directly in _objects collection because
        // we maybe manage weak references.
        Object     cachedObject = null;

        try
        {
            cachedObject = getObject(objectId);
        }
        catch (Exception e)
        {
            cachedObject = null;
        }

        if (cachedObject == null)
        {
            LyEnvironment environment = getEnvironment();

            environment.logWarning("BF AppEngine : Attempting to remove an object that is not in the cache : "
                    + objectId.getFullString());

            if (environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
                Thread.dumpStack();

            return;
        }

        if ((_removedObjects != null) && _removedObjects.contains(object))
            return; // start remove was already done

        removeFromCache(objectId);

        // Register object in removed objects
        if (_removedObjects == null)
            _removedObjects = new Vector<Object>(1);

        _removedObjects.addElement(object);

        // Remove peer object if associated class is a classInfo
        if (_class instanceof LyClassInfo)
        {
            Object peer = ((LyObject)object).getPeer();

            if (peer != null)
                removePeerObject(peer);
        }

        return;
    } // end method startRemoveObject

    /**
     * Deletes the given object from cache. This object must have been previously marked as removed.
     *
     * @param      object  Object to delete from cache.
     *
     * @exception  IllegalStateException     If the object has not been marked as removed from this
     *                                       cache before trying to delete it. It's still in list of
     *                                       stored objects.
     * @exception  IllegalArgumentException  If the object was not found in removed list.
     */
    protected void removeObject(Object object)
    {
        if (object == null)
            return;

        // Check if object is still in objects list
        if (contains(object))
            throw new IllegalStateException("BF AppEngine : Object " + object
                + " must be marked as removed from cache before trying to delete it");

        synchronized (this)
        {
            // Remove object from removed objects list, throw an exception if it fails
            if ((_removedObjects != null)
                    && !_removedObjects.removeElement(object))
                throw new IllegalArgumentException("BF AppEngine : Object " + object
                    + " was not found in removed objects list (removed twice ?)");
        }

        // Retrieve object id.
        LyObjectId objectId = _class.getObjectId(object);

        // Put it in unknown ids
        registerUnknownId(objectId);
    }

    /**
     * Check if an object is in the cache.
     *
     * @param   object  Object to check.
     *
     * @return  True if the object is found, false otherwise.
     */
    public boolean contains(Object object)
    {
        LyObjectId id = _class.getObjectId(object);
        Object     cachedObject = getFromCache(id);

        return (cachedObject == object);
    }

    /**
     * Check if an object is in the cache.
     *
     * @param   objectId  Object to check.
     *
     * @return  True if the object is found, false otherwise.
     */
    public boolean contains(LyObjectId objectId)
    {
        Object cachedObject = getFromCache(objectId);

        return (cachedObject != null);
    }

    /**
     * Return the number of objects within this cache that match the given filter If needed,
     * depending on the cache policy, objects that are not in the cache when the method is called
     * may be loaded in it.
     *
     * @param   filter  Filter used for selecting objects. If filter is null, the total is returned.
     *
     * @return  Number of objects in the cache that match the filter after this call.
     */
    public int getCount(LyFilter filter)
    {
        if (isLoaded())
            return getSize(filter);

        return -1;
    }

    /**
     * Adds objects to the given object list that match the specified filter.
     *
     * @param   filter  Filter to select objects.
     * @param   list    List where results are added.
     *
     * @return  One of the result type Result.TOTAL : the list contains all the results, Result.NULL
     *          : the lists contains no result, Result.PARTIAL : the list contains partial results.
     *
     *          <p>Result.TOTAL can only be returned when the cache is loaded.</p>
     */
    public Result getObjects(LyFilter filter, LySimpleObjectList list)
    {
        if (_class instanceof LyClassInfo)
        {
            // It is useless to return  'partial' result
            // since the objectsource will have to send a request anyway.
            // XX to do : more clever cache policy to know if it has all for current request
            if (!isLoaded())
                return Result.NULL;

            if (!isGood(filter))
                return Result.NULL;

            Enumeration<Object> elements = getElements();

            if (filter == null)
            {
                LyObject object;

                while (elements.hasMoreElements())
                {
                    object = (LyObject)elements.nextElement();
                    if ((object != null)
                            && ((filter == null) || (filter.match(object))))
                        list.addObject(object);
                }
            }
            else
                filter.match(elements, list);

            return (isLoaded() ? Result.TOTAL : Result.PARTIAL);
        }

        return Result.NULL;
    } // end method getObjects

    boolean isGood(LyFilter filter)
    {
        if (filter == null)
            return true;

        int             i, n = filter.getSize();
        LyFilterElement e;
        LyFilter        f;
        boolean         b = true;
        LyFieldInfo     fieldInfo;
        LyApplication   application;

        for (i = 0; (i < n) && b; i++)
        {
            e = filter.getElement(i);
            if (e instanceof LyJoinFilter)
                b = false;
            else if (e instanceof LyFilterRelation)
            {
                f = ((LyFilterRelation)e).getFilter();
                b = isGood(f);
            }
            else if (e instanceof LyFilter)
            {
                f = (LyFilter)e;
                b = isGood(f);
            }
            else if (e instanceof LyFilterExpression)
            {
                fieldInfo = ((LyFilterExpression)e).getFieldInfo();
                application = fieldInfo.getApplication();
                if ((fieldInfo instanceof LyRelationFieldInfo)
                        && (fieldInfo.isLocal())
                        && (application.getDaemon(fieldInfo) != null))
                {
                    // check if target classes are fully loaded
                    LyInfoList<LyClassInfo> targets = ((LyRelationFieldInfo)fieldInfo).getClassInfos();
                    LyInfoList<LyClassInfo> realTargets = new LyInfoList<LyClassInfo>();
                    LyInfoList<LyClassInfo> subTargets;
                    LyClassInfo             classInfo;
                    LyObjectSource          os;

                    for (int j = 0; j < targets.getSize(); j++)
                    {
                        classInfo = targets.getInfo(j);
                        subTargets = classInfo.getSubClasses(true);
                        realTargets.addInfo(classInfo);

                        if (subTargets != null)
                            realTargets.addList(subTargets);
                    }

                    for (int j = 0; j < realTargets.getSize(); j++)
                    {
                        classInfo = realTargets.getInfo(j);
                        if (classInfo.isAbstract())
                            continue;

                        os = classInfo.getObjectSource();

                        if ((os != null) && !os.getCache().isLoaded())
                        {
                            b = false;
                            break;
                        }
                    }
                }
            }
        }

        return b;
    } // end method isGood

    /**
     * Adds objects to the given object list that have a relation value which contains the specified
     * object.
     *
     * @param   targetObject  target object which is pointed by the relation list
     * @param   relations     list of related relations (between our object and the target)
     * @param   list          List where results are added.
     *
     * @return  One of the result type Result.TOTAL : the list contains all the results, Result.NULL
     *          : the lists contains no result,
     */
    Result getObjects(LyObject targetObject, LyInfoList<LyFieldInfo> relations, LySimpleObjectList list)
    {
        if (_class instanceof LyClassInfo)
        {
            LyObject            object;
            LyObjectId          targetObjectId = targetObject.getObjectId();
            LyFieldInfo         relation;
            Enumeration<Object> elements = getElements();
            Object              value;
            int                 i, n = relations.getSize();

            while (elements.hasMoreElements())
            {
                object = (LyObject)elements.nextElement();

                if (object == null)
                    continue;

                for (i = 0; i < n; i++)
                {
                    relation = relations.getInfo(i);

                    // Do not compute values we only want from memory (cache)
                    value = object.getFromMemory(relation);

                    if ((value != null)
                            && relation.contains(value, targetObjectId)) // Got it ?
                    {
                        list.addObject(object);
                        break; // break 'for' not while. Stop seeking on relation, object is found
                    }
                }
            }

            return Result.TOTAL;
        }

        return Result.NULL;
    } // end method getObjects

    /**
     * Method for loading objects in the cache. This method should never be called directly from the
     * application.
     */
    protected void load()
    {
        if (_class instanceof LyClassInfo)
        {
            LyObjectSource objectSource;
            LyClassInfo    classInfo;
            LyEnvironment  environment;

            classInfo = (LyClassInfo)_class;
            objectSource = classInfo.getObjectSource();
            environment = getEnvironment();

            Vector<LyFieldInfo> fields;

            if (environment.testEnv("LY_SAVE_FILES"))
                fields = null;
            else
            {
                BitSet bitset = new BitSet();
                bitset.set(LyFieldInfo.MARK_TABLE);
                bitset.set(LyFieldInfo.MARK_LOAD);
                LyInfoList<LyFieldInfo> list = classInfo.getFieldInfos(bitset, false);

                fields = ((list == null) ? null : list.getInfos());
            }

            // Change _loaded value before loading to avoid loops
            setLoaded(true);

            objectSource.loadObjects(fields, null, null, -1);
        }
    } // end method load

    /**
     * Method for unloading objects from the cache. This method should never be called directly from
     * the application.
     */
    protected void unload()
    {
        if (_class instanceof LyClassInfo)
        {
            // We must set all the objects as deleted to avoid warnings
            // in finalization.
            Enumeration<Object> elements = getElements();
            LyObject            object;

            while (elements.hasMoreElements())
            {
                object = (LyObject)elements.nextElement();
                object.setDeleted(true);
            }
        }

        clearCache();

        _unknownObjects = null;
        _unknownList = null;

        setLoaded(false);
    }

    /**
     * Tests if an object id has been registered as corresponding to an unknown object for this
     * cache.
     *
     * @param   id  Tested object id.
     *
     * @return  True if the object id corresponds to an unknown object, false otherwise.
     */
    protected final boolean isUnknown(LyObjectId id)
    {
        if (_unknownObjects == null)
            return false;

        return _unknownObjects.containsKey(id);
    }

    /**
     * Tests whether an object is being removed.
     *
     * @param   object  Tested object.
     *
     * @return  True if the object in removed objects list, false otherwise.
     */
    protected final boolean isRemoved(Object object)
    {
        if (_removedObjects == null)
            return false;

        // Don't use contains that is based on equals method of given object, to be sure to remove the right object
        int index = _removedObjects.indexOf(object);

        return ((index >= 0) && (_removedObjects.get(index) == object));
    }

    /**
     * Tests whether an object is being removed.
     *
     * @param   objectId  Tested object id.
     *
     * @return  True if the object pointed by the object id is in removed objects list, false
     *          otherwise.
     */
    protected final boolean isRemoved(LyObjectId objectId)
    {
        if (_removedObjects == null)
            return false;

        synchronized (this)
        {
            Object     removedObject;
            LyObjectId removedObjectId;
            int        i, n = _removedObjects.size();

            for (i = 0; i < n; i++)
            {
                removedObject = _removedObjects.get(i);
                removedObjectId = _class.getObjectId(removedObject);
                if (removedObjectId.equals(objectId))
                    return true;
            }
        }

        return false;
    }

    /**
     * Registers an object id as corresponding to an unknown object. Once registered, the getObject
     * method won't throw exceptions for this id.
     *
     * @param  id  Object id that is registered.
     *
     * @see    #getObject
     */
    protected void registerUnknownId(LyObjectId id)
    {
        if (getFromCache(id) != null)
            throw new IllegalStateException("BF AppEngine : the given id " + id
                + " corresponds to a known object");

        if (_unknownObjects == null)
        {
            _unknownObjects = new Hashtable<LyObjectId, Object>(1);
            _unknownList = new Vector<LyObjectId>();
        }

        _unknownObjects.put(id, UNKNOWN_OBJECT);
        _unknownList.add(id);
        if (_unknownList.size() > NB_UNKNOWN_OBJECTS)
        {
            id = _unknownList.elementAt(0);
            _unknownList.removeElementAt(0);
            _unknownObjects.remove(id);
        }
    }

    /**
     * Retrieves the class associated to this cache. If the cache is used for storing logical
     * objects (LyObject), the returned class is an instance of LyClassInfo.
     *
     * @return  Class associated to this cache.
     */
    public final LyClass getCacheClass()
    {
        return _class;
    }

    /**
     * Get list of objects found in this cache.
     *
     * @return  Vector of all objects of this cache. An empty Vector if the cache is empty (never
     *          returns null).
     */
    public Vector<Object> getObjects()
    {
        Vector<Object>      list = new Vector<Object>(1);
        Enumeration<Object> elements = getElements();
        Object              object;

        while (elements.hasMoreElements())
        {
            object = elements.nextElement();
            list.addElement(object);
        }

        return list;
    }

    /**
     * Get list of objects found in this cache.
     *
     * @param   sorted  True if the requiered list must be sorted.
     *
     * @return  An objects list of all objects of this cache. An empty list if the cache is empty
     *          (never returns null).
     */
    public LyObjectList getObjectList(boolean sorted)
    {
        if (!(_class instanceof LyClassInfo))
            throw new IllegalStateException("BF AppEngine : attempt to get an object list on a non classInfo cache");

        LySimpleObjectList  list = new LySimpleObjectList();
        Enumeration<Object> elements = getElements();

        list.setSorted(sorted);

        LyObject object;

        while (elements.hasMoreElements())
        {
            object = (LyObject)elements.nextElement();
            list.addObject(object);
        }

        return list;
    }

    /**
     * Tests whether this cache is completely loaded or not.
     *
     * @return  True if the cache is loaded, false otherwise.
     */
    public boolean isLoaded()
    {
        return _loaded;
    }

    /**
     * Marks this cache as loaded or not. This is for internal use only : changing the loaded status
     * of the cache on wrong purpose may generate an inconsistent cache
     *
     * @param  loaded  True to mark the cache loaded, false otherwise.
     */
    public void setLoaded(boolean loaded)
    {
        _loaded = _allowLoaded && loaded;
    }

    /**
     * Gets environment of the cache.
     *
     * @return  the enviroment of this cache
     */
    public LyEnvironment getEnvironment()
    {
        LyDataSource  dataSource = _class.getDataSource();
        LyApplication application = dataSource.getApplication();

        return application.getEnvironment();
    }

    /**
     * This method recomputes the association between an object and its objectId.
     *
     * @param   object       The object
     * @param   oldObjectId  The old value of the objectId of the specified object
     *
     * @return  true if rehash was performed
     */
    public boolean rehash(Object object, LyObjectId oldObjectId)
    {
        if ((object != null)
                && (oldObjectId != null)
                && contains(oldObjectId))
        {
            LyObjectId objectId = _class.getObjectId(object);

            if (objectId.equals(oldObjectId))
                return false;

            removeFromCache(oldObjectId);
            putToCache(objectId, object);
            objectId.setObject(object);

            // Remove id from unkwown ids if requiered
            if (isUnknown(objectId))
            {
                _unknownObjects.remove(objectId);
                _unknownList.remove(objectId);
            }

            return true;
        }

        return false;
    }

    /**
     * Clear the cache from all its unknown ids.
     * Calling this method can affect performance of the application (in a wrong way).
     */
    public void clearUnknownIdsCache()
    {
        _unknownObjects = new Hashtable<LyObjectId, Object>(1);
        _unknownList = new Vector<LyObjectId>();
    }
} // end class LyCache
