/*
 * LyChoiceValue.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;

import leon.app.*;
import leon.info.*;


/**
 * This class offers methods for constructing, accessing and modifying values corresponding to a
 * LyChoiceFieldInfo.
 *
 * @see  LyChoiceFieldInfo
 * @see  LyObject
 */
public final class LyChoiceValue extends LyFieldInfoValue
{
    /** File version. */
    public static final String __VERSION = "$Id: LyChoiceValue.java 28980 2010-12-22 10:53:09Z jlsaouli $";

    /**
     * Constructs an empty choice value for the given choice field.
     *
     * @param  choiceFieldInfo  Choice fieldInfo for which a value is created.
     */
    public LyChoiceValue(LyChoiceFieldInfo choiceFieldInfo)
    {
        super(choiceFieldInfo);

        if (choiceFieldInfo.getType() == LyChoiceFieldInfo.Type.BOOLEAN)
            throw new IllegalStateException("BF AppEngine : the value for fieldInfo " + choiceFieldInfo
                + " can't be a choice because the type of this choice is boolean");
    }

    /**
     * Constructs a choice value for the given choice field from a string.
     *
     * @param  choiceFieldInfo  Choice fieldInfo for which a value is created.
     * @param  value            String value. If the choice field is multiple, several options may
     *                          be selected from the same string, in that case, the different
     *                          sub-strings corresponding to each option must be separated by
     *                          LyApplication.VALUE_SEP. For one option, the given string can be (in
     *                          this order) : the internal value of the option, its id or its the
     *                          name (untranslated).
     *
     * @see    LyChoiceFieldInfo#parse
     */
    public LyChoiceValue(LyChoiceFieldInfo choiceFieldInfo, String value)
    {
        super(choiceFieldInfo, value);

        if (choiceFieldInfo.getType() == LyChoiceFieldInfo.Type.BOOLEAN)
            throw new IllegalStateException("BF AppEngine : the value for fieldInfo " + choiceFieldInfo
                + " can't be a choice because the type of this choice is boolean");
    }

    /**
     * Constructs a choice value for the given choice field from a short.
     *
     * @param      choiceFieldInfo  Choice fieldInfo for which a value is created.
     * @param      value            Value passed as a short. This short value represents the
     *                              position of the chosen option in options list (0 is the firs
     *                              one).
     *
     * @exception  IllegalArgumentException  If the value is negative.
     * @exception  IllegalArgumentException  If the value is greater or equal to number of options
     *                                       in the given choice.
     */
    public LyChoiceValue(LyChoiceFieldInfo choiceFieldInfo, short value)
    {
        this(choiceFieldInfo, new Short(value));
    }

    /**
     * Constructs a choice value for the given choice field from an int.
     *
     * @param      choiceFieldInfo  Choice fieldInfo for which a value is created.
     * @param      value            Value passed as an int. This integer value represents the
     *                              position of the chosen option in options list (0 is the firs
     *                              one).
     *
     * @exception  IllegalArgumentException  If the value is negative.
     * @exception  IllegalArgumentException  If the value is greater or equal to number of options
     *                                       in the given choice.
     */
    public LyChoiceValue(LyChoiceFieldInfo choiceFieldInfo, int value)
    {
        this(choiceFieldInfo, new Short((short)value));
    }

    /**
     * Constructs a choice value for the given choice field from a Short object.
     *
     * @param      choiceFieldInfo  Choice fieldInfo for which a value is created.
     * @param      value            Value passed as a Short. This Short value represents the
     *                              position of the chosen option in options list (0 is the firs
     *                              one).
     *
     * @exception  IllegalArgumentException  If the value is negative.
     * @exception  IllegalArgumentException  If the value is greater or equal to number of options
     *                                       in the given choice.
     */
    public LyChoiceValue(LyChoiceFieldInfo choiceFieldInfo, Short value)
    {
        super(choiceFieldInfo, null);

        if (choiceFieldInfo.getType() == LyChoiceFieldInfo.Type.BOOLEAN)
            throw new IllegalStateException("BF AppEngine : the value for fieldInfo " + choiceFieldInfo
                + " can't be a choice because the type of this choice is boolean");

        if (value == null)
            return;

        short shortValue = value.shortValue();

        if (shortValue == -1) // XXX : means null. Is this still needed ?
            return;

        short ni = choiceFieldInfo.getChoiceItemCount();

        if (shortValue < 0)
            throw new IllegalArgumentException("BF AppEngine : choice value  for fieldInfo "
                + choiceFieldInfo + " can't be negative");

        if (shortValue >= ni)
            throw new IllegalStateException("BF AppEngine : choice value for fieldInfo "
                + choiceFieldInfo + " can't be greater than " + (ni - 1) + " (value given is " + value + ")");

        if (!choiceFieldInfo.isMultiple())
            setValue(value);
        else
        {
            BitSet bs = new BitSet();

            bs.set(shortValue);
            setValue(bs);
        }
    }

    /**
     * Constructs a choice value for the given choice field from a BitSet. Each bit of the given
     * BitSet indicates if the option at the same position in choice options list is selected.
     *
     * @param      choiceFieldInfo  Choice fieldInfo for which a value is created.
     * @param      value            Value passed as a BitSet.
     *
     * @exception  IllegalStateException  If the choice field is not multiple.
     */
    public LyChoiceValue(LyChoiceFieldInfo choiceFieldInfo, BitSet value)
    {
        super(choiceFieldInfo, value);

        if (choiceFieldInfo.getType() == LyChoiceFieldInfo.Type.BOOLEAN)
            throw new IllegalStateException("BF AppEngine : the value for fieldInfo "
                + choiceFieldInfo
                + " can't be a choice because the type of this choice is boolean");

        if (!choiceFieldInfo.isMultiple())
            throw new IllegalStateException("BF AppEngine : choice value for fieldInfo "
                + choiceFieldInfo + " can't be a bitset because this choice is not multiple");
    }

    /**
     * Method for checking if index passed to methods is valid or not.
     *
     * @param      index  index to check
     *
     * @exception  IllegalArgumentException  If the given index is negative.
     * @exception  IllegalArgumentException  If the given index is greater or equal to the number of
     *                                       options of the choice field.
     */
    public void checkIndex(int index)
    {
        LyChoiceFieldInfo choiceFieldInfo = (LyChoiceFieldInfo)getFieldInfo();
        int               n = choiceFieldInfo.getChoiceItemCount();

        if (index < 0)
            throw new IllegalArgumentException("BF AppEngine : the requested index " + index
                + " can't be negative");

        if (index >= n)
            throw new IllegalArgumentException("BF AppEngine : the requested index " + index
                + " can't be greater to the number of options of the choice field ("
                + n + " for choice field " + choiceFieldInfo + ")");
    }

    /**
     * Gets value of this choice value as a short.
     *
     * @return  A short corresponding to this value.
     */
    public short shortValue()
    {
        Short value = getShortValue();

        return ((value == null) ? -1 : value.shortValue());
    }

    /**
     * Gets value of this choice value as a Short.
     *
     * @return     A Short corresponding to this value.
     *
     * @exception  IllegalStateException  If the choice field is multiple.
     */
    public Short getShortValue()
    {
        Object value = getValue();

        if ((value != null) && !(value instanceof Short))
            throw new IllegalStateException("BF AppEngine : choice value for fieldInfo "
                + getFieldInfo()
                + " can't be returned as a Short because this choice is multiple");

        return (Short)value;
    }

    /**
     * Gets value of this choice value as a BitSet.
     *
     * @return     A BitSet corresponding to this value.
     *
     * @exception  IllegalStateException  If the choice field is not multiple.
     */
    public BitSet getBitSetValue()
    {
        Object value = getValue();

        if ((value != null) && !(value instanceof BitSet))
            throw new IllegalStateException("BF AppEngine : choice value for fieldInfo "
                + getFieldInfo()
                + " can't be returned as a BitSet because this choice is not multiple");

        return (BitSet)value;
    }

    /**
     * Gets number of selected options in this choice value.
     *
     * @return  Number of selected options (0 if no option is selected in this value, can't be
     *          greater than 1 if the corresponding choice field is not multiple).
     */
    public short getValueCount()
    {
        Object value = getValue();

        if (value == null)
            return 0;
        else if (value instanceof Short)
            return 1;
        else
            return (short)LyBitSet.getCount((BitSet)value);
    }

    /**
     * Gets the id of the specified selected option in this choice value.
     *
     * @param      index  Index of requested option id.
     *
     * @return     the id of the option
     *
     * @exception  IllegalArgumentException  If index is negative.
     * @exception  IllegalArgumentException  If index is greater than the number of options selected
     *                                       in this value.
     */
    public String getString(int index)
    {
        short count = getValueCount();

        if (index < 0)
            throw new IllegalArgumentException("BF AppEngine : index can't be negative");

        if (index >= count)
            throw new IllegalArgumentException("BF AppEngine : index must be lower than " + count);

        LyChoiceFieldInfo fieldInfo = (LyChoiceFieldInfo)getFieldInfo();

        Object value = getValue();

        if (value instanceof Short)
            return fieldInfo.toString(value);
        else
        {
            BitSet bs = (BitSet)value;
            BitSet bsIndex = new BitSet(count);
            short  i, found = 0, n = fieldInfo.getChoiceItemCount();

            for (i = 0; i < n; i++)
            {
                if (bs.get(i))
                {
                    // is it the requested index ?
                    if (found == index)
                    {
                        bsIndex.set(i);
                        return fieldInfo.toString(bsIndex);
                    }

                    found++;
                }
            }
        }
        // should never happen
        throw new IllegalArgumentException("BF AppEngine : can find value at requested index " + index);
    } // end method getString

    /**
     * Gets the name of the specified selected option in this choice value.
     *
     * @param      index  Index of requested option name.
     *
     * @return     the name of the option
     *
     * @exception  IllegalArgumentException  If index is negative.
     * @exception  IllegalArgumentException  If index is greater than the number of options selected
     *                                       in this value.
     */
    public String getName(int index)
    {
        short count = getValueCount();

        if (index < 0)
            throw new IllegalArgumentException("BF AppEngine : index can't be negative");

        if (index >= count)
            throw new IllegalArgumentException("BF AppEngine : index must be lower than " + count);

        LyChoiceFieldInfo fieldInfo = (LyChoiceFieldInfo)getFieldInfo();

        Object value = getValue();

        if (value instanceof Short)
            return fieldInfo.toName(value);
        else
        {
            BitSet bs = (BitSet)value;
            BitSet bsIndex = new BitSet(count);
            short  i, found = 0, n = fieldInfo.getChoiceItemCount();

            for (i = 0; i < n; i++)
            {
                if (bs.get(i))
                {
                    // is it the requested index ?
                    if (found == index)
                    {
                        bsIndex.set(i);
                        return fieldInfo.toName(bsIndex);
                    }

                    found++;
                }
            }
        }
        // should never happen
        throw new IllegalArgumentException("BF AppEngine : can find value at requested index " + index);
    } // end method getName

    /**
     * Return the value for a given level. If the choice is multiple, the result is the
     * concatenation of the ids of all selected option(s) for the given level
     *
     * @param   level  the requested level
     *
     * @return  the value
     */
    public String getValue(int level)
    {
        short        count = getValueCount();
        short        i;
        StringBuffer sb = null;

        for (i = 0; i < count; i++)
        {
            if (sb == null)
                sb = new StringBuffer();
            else
                sb.append(LyApplication.VISIBLE_VALUE_SEP);

            sb.append(getValue(level, i));
        }

        return ((sb == null) ? null : sb.toString());
    }

    /**
     * Compute the value of an option for a given level.
     *
     * @param   option  the option
     * @param   level   the level
     *
     * @return  the value at the given level
     */
    String getChoiceOptionValueAtLevel(LyChoiceOption option, int level)
    {
        Vector<LyInfo> result = new Vector<LyInfo>();

        result.addElement(option);

        LyInfo parent = option.getParent();

        if ((parent != null) && (parent instanceof LyChoiceOptionGroup))
            result.insertElementAt(parent, 0);

        while ((parent != null) && (parent.getParent() instanceof LyChoiceOptionGroup))
        {
            parent = parent.getParent();
            result.insertElementAt(parent, 0);
        }

        // Now we have all levels.
        String value = null;

        if (level < result.size())
        {
            LyChoiceOptionGroup group = (LyChoiceOptionGroup)result.elementAt(level);
            Object              groupValue = group.getValue();

            if (groupValue == null)
                value = option.getId();
            else
                value = groupValue.toString();
        }
        else
        {
            Object optionValue = option.getValue();

            if (optionValue == null)
                value = option.getId();
            else
                value = optionValue.toString();
        }

        return value;
    } // end method getChoiceOptionValueAtLevel

    /**
     * Compute the value at the given level for the given data.
     *
     * @param   fieldValue  the currentValue
     * @param   level       the level
     *
     * @return  a String
     */
    String computeValue(Object fieldValue, int level)
    {
        if (fieldValue == null)
            return null;

        LyChoiceFieldInfo fieldInfo = (LyChoiceFieldInfo)getFieldInfo();

        if (!fieldInfo.isMultiple())
        {
            LyChoiceOption choiceOption = null;

            if (fieldInfo.getType() == LyChoiceFieldInfo.Type.BOOLEAN)
            {
                Boolean b = (Boolean)fieldValue;

                if (b != null)
                {
                    if (b.booleanValue())
                        choiceOption = fieldInfo.getChoiceOption((short)1);
                    else
                        choiceOption = fieldInfo.getChoiceOption((short)0);
                }
            }
            else
            {
                Short s = (Short)fieldValue;

                choiceOption = fieldInfo.getChoiceOption(s.shortValue());
            }

            if (choiceOption != null)
                return getChoiceOptionValueAtLevel(choiceOption, level);
            else
                return null;
        }

        BitSet bs = (BitSet)fieldValue;
        short  i, n = (short)bs.size();
        String str = null;

        for (i = 0; i < n; i++)
        {
            if (bs.get(i))
            {
                if (str == null)
                    str = getChoiceOptionValueAtLevel(fieldInfo.getChoiceOption(i), level);
                else
                    str = str + fieldInfo.getApplication().VALUE_SEP
                        + getChoiceOptionValueAtLevel(fieldInfo.getChoiceOption(i), level);
            }
        }

        return str;
    } // end method computeValue

    /**
     * Return the value for a specific index at the given level.
     *
     * @param   level  the requested level
     * @param   index  the index of the requested option id
     *
     * @return  the resulting id
     */
    public String getValue(int level, int index)
    {
        short count = getValueCount();

        if (index < 0)
            throw new IllegalArgumentException("BF AppEngine : index can't be negative");

        if (index >= count)
            throw new IllegalArgumentException("BF AppEngine : index must be lower than " + count);

        LyChoiceFieldInfo fieldInfo = (LyChoiceFieldInfo)getFieldInfo();

        Object value = getValue();

        if (value instanceof Short)
            return computeValue(value, level);
        else
        {
            BitSet bs = (BitSet)value;
            BitSet bsIndex = new BitSet(count);
            short  i, found = 0, n = fieldInfo.getChoiceItemCount();

            for (i = 0; i < n; i++)
            {
                if (bs.get(i))
                {
                    // is it the requested index ?
                    if (found == index)
                    {
                        bsIndex.set(i);
                        return computeValue(bsIndex, level);
                    }

                    found++;
                }
            }
        }
        // should never happen
        throw new IllegalArgumentException("BF AppEngine : can find value at requested index " + index);
    } // end method getValue

    /**
     * Checks if the option at the given index is selected or not in this choice value. If the
     * choice field is multiple, several choices may be selected simultaneously.
     *
     * @param   index  Index of the tested option. The given index is checked by method checkIndex,
     *                 this method may throw exceptions.
     *
     * @return  True if selected, false otherwise.
     *
     * @see     LyChoiceValue#checkIndex
     */
    public boolean isSelected(short index)
    {
        Object value = getValue();

        if (value == null)
            return false;
        else if (value instanceof Short)
            return (((Short)value).shortValue() == index);
        else
            return ((BitSet)value).get(index);
    }

    /**
     * Checks an option with the given value, id or name is selected in this choice value. If the
     * choice field is multiple, several choices may be selected simultaneously.
     *
     * @param      value  Value, id or name of the tested option.
     *
     * @return     True if selected, false otherwise.
     *
     * @exception  IllegalArgumentException  if no option corresponds to the given value, id or
     *                                       name.
     *
     * @see        LyChoiceValue#checkIndex
     */
    public boolean isSelected(Object value)
    {
        LyChoiceFieldInfo choice = (LyChoiceFieldInfo)getFieldInfo();
        short             index = choice.getChoiceOptionIndex(value);

        return ((index < 0) ? false : isSelected(index));
    }

    /**
     * Sets an option in this choice value. Does nothing if the corresponding option was already
     * selected. If the choice field is not multiple, deselects the previous selected one.
     *
     * @param  index  Index of the set option. The given index is checked by method checkIndex, this
     *                method may throw exceptions.
     *
     * @see    LyChoiceValue#checkIndex
     */
    public void setOption(short index)
    {
        LyFieldInfo choiceFieldInfo = getFieldInfo();

        if (!choiceFieldInfo.isMultiple())
            setValue(new Short(index));
        else
        {
            BitSet bs = getBitSetValue();

            if (bs == null)
            {
                bs = new BitSet();
                bs.set(index);
                setValue(bs);
            }
            else
                bs.set(index);
        }
    }

    /**
     * Clears an option in this choice value. Does nothing if the corresponding option was not
     * selected.
     *
     * @param  index  Index of the cleared option. The given index is checked by method checkIndex,
     *                this method may throw exceptions.
     *
     * @see    LyChoiceValue#checkIndex
     */
    public void clearOption(short index)
    {
        Object value = getValue();

        if (value == null)
            return;

        if (value instanceof Short)
        {
            if (((Short)value).shortValue() == index)
                setValue(null);
        }
        else
            ((BitSet)value).clear(index);
    }

    /**
     * Gets the value of the specified selected option in this choice value.
     *
     * @param      index  Index of requested option id.
     *
     * @return     the value of the selected option
     *
     * @exception  IllegalArgumentException  If index is negative.
     * @exception  IllegalArgumentException  If index is greater than the number of options selected
     *                                       in this value.
     */
    public Object getOptionValue(int index)
    {
        short count = getValueCount();

        if (index < 0)
            throw new IllegalArgumentException("BF AppEngine : index can't be negative");

        if (index >= count)
            throw new IllegalArgumentException("BF AppEngine : index must be lower than " + count);

        LyChoiceFieldInfo fieldInfo = (LyChoiceFieldInfo)getFieldInfo();
        Object            value = getValue();

        if (value instanceof Short)
        {
            LyChoiceItem item = fieldInfo.getChoiceItem(((Short)value).shortValue());

            return item.getValue();
        }
        else
        {
            BitSet       bs = (BitSet)value;
            BitSet       bsIndex = new BitSet(count);
            short        i, found = 0, n = fieldInfo.getChoiceItemCount();
            LyChoiceItem item;

            for (i = 0; i < n; i++)
            {
                if (bs.get(i))
                {
                    // is it the requested index ?
                    if (found == index)
                    {
                        bsIndex.set(i);
                        item = fieldInfo.getChoiceItem(i);
                        return item.getValue();
                    }

                    found++;
                }
            }
        }
        // should never happen
        throw new IllegalArgumentException("BF AppEngine : can find value at requested index " + index);
    } // end method getString
} // end class LyChoiceValue
