/*
 * LyClass.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;


/**
 * This interface represents the description of physical classes : classes from which datas are
 * obtained to display logical objects presented to the end-user using Leonardi.
 *
 * <p>The enclosed methods represent the level of services requiered from these physical classes to
 * be able to interface various types of datas.</p>
 *
 * <p>Different kinds of methods are specified :</p>
 *
 * <ul>
 *   <LI>methods to get information about the underlying physical class : getId, getName, getFields,
 *     getField, getIdFields, getObjectClass,</LI>
 *   <LI>method for building new objects : newInstance,</LI>
 *   <LI>methods for introspecting objects : getObjectId, getValue, getValues,</LI>
 *   <LI>methods for modifying existing objects : setValue, copy,</LI>
 *   <LI>method to get data source : getDataSource.
 *
 *     <p>This interface is mainly implemented by physical classes on which logical classes are
 *     interfaced.</p>
 *
 *     <p>But it is also implemented by LyClassInfo, though this class represents logical classes.
 *     This allows to see a logical class as a physical class for another logical class.</p>
 *   </LI>
 * </ul>
 *
 * @see LyField
 * @see leon.peer.LyPeerClass
 * @see leon.info.LyClassInfo
 * @see LyDataSource
 */
public interface LyClass
{
	/** File version */
	String __VERSION = "$Id: LyClass.java 28980 2010-12-22 10:53:09Z jlsaouli $";

	/**
	 * Gets identifier of the class.
	 *
	 * @return Identifier of the class.
	 */
	String getId();

	/**
	 * Gets name of the class. The name of the class is the name in the information system. By
	 * default, if no name is specified, this method returns the id.
	 *
	 * @return Name of the class.
	 */
	String getName();

	/**
	 * Gets the physical representation of the instances : java class that really holds instances of
	 * this class.
	 *
	 * @return Java class of the "objects" of this class.
	 */
	Class<?> getObjectClass();

	/**
	 * Gets the data source that manages all instances of this class.
	 *
	 * @return Data source that manages the instances of this class.
	 */
	LyDataSource getDataSource();

	/**
	 * Gets the list of fields that identifies instances of this class.
	 *
	 * @return List of fields from which an identifier can be calculated.
	 */
	Vector<? extends LyField> getIdFields();

	/**
	 * Gets the list of fields of this class.
	 *
	 * @return List of all fields of this class.
	 */
	Vector<? extends LyField> getFields();

	/**
	 * Retrieves a field on this class by its identifier.
	 *
	 * @param  id Identifier of the searched field.
	 * @return The field of this class which has the specified identifier, null if no field is
	 *         found.
	 */
	LyField getField(String id);

	/**
	 * Gets identifier for an instance of this class.
	 *
	 * @param  object Instance of this class.
	 * @return Identifier of the given object.
	 */
	LyObjectId getObjectId(Object object);

	/**
	 * Gets the value corresponding to the given field for an instance of this class.
	 *
	 * @param  object Instance of this class.
	 * @param  field  Field for which a value is requested.
	 * @return Value of the given field on the specified instance.
	 */
	Object getValue(Object object, LyField field);

	/**
	 * Sets the value corresponding to the given field on an instance of this class.
	 *
	 * @param object Instance of this class on which a value is set.
	 * @param field  Field for which a value is set.
	 * @param value  Value set for the given field on the specified instance.
	 */
	void setValue(Object object, LyField field, Object value);

	/**
	 * Copies all field values of an instance of this class on another instance.
	 *
	 * @param  object1 Object on which values are copied.
	 * @param  object2 Object from which values are get.
	 * @return Field values that are copied from object2 to object1.
	 */
	LyValueSet copy(Object object1, Object object2);

	/**
	 * Gets all field values from an instance of this class.
	 *
	 * @param  object Instance of this class.
	 * @return All the field values of the given instance.
	 */
	LyValueSet getValues(Object object);

	/**
	 * Creates a new instance of this class.
	 *
	 * @return    New instance of this class.
	 * @exception IllegalAccessException If the class or initializer is not accessible.
	 * @exception InstantiationException If an application tries to instantiate an abstract class or
	 *                                   an interface, or if the instantiation fails for some other
	 *                                   reason.
	 */
	Object newInstance()
		throws InstantiationException,
			IllegalAccessException;

	/**
	 * Recompute the object id of the specified object.
	 * @param object object to rehash
	 */
	void rehash(Object object);
}