/*
 * LyClassWorkSpace.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;

import leon.app.*;
import leon.info.*;


/**
 * This class implements a class workspace : workspace for objects belonging to a given class info
 * and optionnally, matching a filter.
 *
 * @see LyClassInfo
 * @see LyFilter
 */
public class LyClassWorkSpace extends LyWorkSpace
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyClassWorkSpace.java 40913 2012-09-25 08:10:49Z bangiole $";

	/** Class info to which objects of this workspace belong. */
	LyClassInfo				   _classInfo;

	/** Filter applied for objects loaded in this workspace. */
	LyFilter				   _filter;

	/** List of fieldinfos to load. */
	LyInfoList<LyFieldInfo>    _fieldInfos;

	/**
	 * Constructs a class workspace within a given session for a class info.
	 *
	 * @param session   Session in which the workspace is constructed.
	 * @param classInfo Class info to which objects of this workspace belong.
	 * @param dynamic   Specifies if this workspace is dynamic or not.
	 */
	public LyClassWorkSpace(LyDataSession session, LyClassInfo classInfo,
		boolean dynamic)
	{
		this(session, classInfo, null, null, null, dynamic, true);
	}

	/**
	 * Constructs a class workspace within a given session for a class info with specified filter
	 * applied to objects of this class.
	 *
	 * @param session   Session in which the workspace is constructed.
	 * @param classInfo Class info to which objects of this workspace belong.
	 * @param filter    Filter applied to objects for being in this workspace.
	 * @param dynamic   Specifies if this workspace is dynamic or not.
	 */
	public LyClassWorkSpace(LyDataSession session, LyClassInfo classInfo,
		LyFilter filter, boolean dynamic)
	{
		this(session, classInfo, null, filter, null, dynamic, true);
	}

	/**
	 * Constructs a class workspace within a given session for a class info with specified filter
	 * applied to objects of this class and sorted by the given sort.
	 *
	 * @param session   Session in which the workspace is constructed.
	 * @param classInfo Class info to which objects of this workspace belong.
	 * @param filter    Filter applied to objects for being in this workspace.
	 * @param sort      Sort requested in this workspace.
	 * @param dynamic   Specifies if this workspace is dynamic or not.
	 */
	public LyClassWorkSpace(LyDataSession session, LyClassInfo classInfo,
		LyFilter filter, LySort sort, boolean dynamic)
	{
		this(session, classInfo, null, filter, sort, dynamic, true);
	}

	/**
	 * Internal constructor with all parameters for construction of a class workpace and an
	 * additionnal one that indicates if the workspace is loaded or not.
	 *
	 * @param session   Session in which the workspace is constructed.
	 * @param classInfo Class info to which objects of this workspace belong.
	 * @param filter    Filter applied to objects for being in this workspace.
	 * @param sort      Sort requested in this workspace.
	 * @param dynamic   Specifies if this workspace is dynamic or not.
	 * @param sorted    Indicates if this workspace is sorted or not.
	 */
	public LyClassWorkSpace(LyDataSession session, LyClassInfo classInfo,
		LyFilter filter, LySort sort, boolean dynamic, boolean sorted)
	{
		this(session, classInfo, null, filter, sort, dynamic, sorted);
	}

	/**
	 * Internal constructor with all parameters for construction of a class workpace and an
	 * additionnal one that indicates if the workspace is loaded or not.
	 *
	 * @param session    Session in which the workspace is constructed.
	 * @param classInfo  Class info to which objects of this workspace belong.
	 * @param fieldInfos List of the fieldinfos to load.
	 * @param filter     Filter applied to objects for being in this workspace.
	 * @param sort       Sort requested in this workspace.
	 * @param dynamic    Specifies if this workspace is dynamic or not.
	 * @param sorted     Indicates if this workspace must be sorted or not.
	 */
	public LyClassWorkSpace(LyDataSession session, LyClassInfo classInfo,
		LyInfoList<LyFieldInfo> fieldInfos, LyFilter filter, LySort sort, boolean dynamic,
		boolean sorted)
	{
		this(session, classInfo, fieldInfos, filter, sort, dynamic, sorted, false);
	}

	/**
	 * Internal constructor with all parameters for construction of a class workpace and an
	 * additionnal one that indicates if the workspace is loaded or not.
	 *
	 * @param session    Session in which the workspace is constructed.
	 * @param classInfo  Class info to which objects of this workspace belong.
	 * @param fieldInfos List of the fieldinfos to load.
	 * @param filter     Filter applied to objects for being in this workspace.
	 * @param sort       Sort requested in this workspace.
	 * @param dynamic    Specifies if this workspace is dynamic or not.
	 * @param sorted     Indicates if this workspace must be sorted or not.
	 * @param local      Specifies if this workspace is local or not.
	 */
	public LyClassWorkSpace(LyDataSession session, LyClassInfo classInfo,
		LyInfoList<LyFieldInfo> fieldInfos, LyFilter filter, LySort sort, boolean dynamic,
		boolean sorted, boolean local)
	{
		super(session, dynamic, local);
		setSorted(sorted);

		if (classInfo == null)
			throw new IllegalArgumentException("BF AppEngine : classInfo can't be null");

		_classInfo = classInfo;
		_fieldInfos = fieldInfos;
		_filter = filter;
		_sort = sort;

		// Initial loading
		if (getClass() == LyClassWorkSpace.class)
		{
			reset();

			if (dynamic)
			{
				// Add this as listener of classInfo objectSource
				getObjectSource().addListener(this);

				createFilterListeners(_filter);
			}
		}
	}

	/**
	 * Free this workspace. Remove this workspace as listener from object source of the class info
	 * specified for this workspace.
	 */
	protected void free()
	{
		if (isDynamic())
		{
		    LyObjectSource    source = getObjectSource();

		    if (source != null)
		        source.removeListener(this);
		}

		super.free();
	}

	/**
	 * Return a sort using 'real' fieldInfo as opposed to a sort using virtual field.
	 *
	 * @param  sort the original sort
	 * @return a sort using real fields instead of virtual one if possible.
	 */
	protected LySort getRealSort(LySort sort)
	{
		if (sort == null)
			return null;

		if (!_classInfo.isVirtual())
			return sort;

		LySort		    result = new LySort();
		LyClassInfo     realClass = _classInfo.getRealClass();
		LyFieldInfo     field;
		int			    i, n = sort.getCriterionCount();
		LySortCriterion criterion;

		for (i = 0; i < n; i++)
		{
			criterion = sort.getCriterion(i);
			field = realClass.getRealFieldInfo(criterion.getFieldInfo());
			result.addCriterion(field, criterion.isIncreasing());
		}
		return result;
	}

	/**
	 * Return a filter using 'real' fieldInfo as opposed to a filter using virtual field.
	 *
	 * @param  filter the original filter
	 * @return a filter using real fields instead of virtual one if possible.
	 */
	protected LyFilter getRealFilter(LyFilter filter)
	{
		if (filter == null)
			return null;

		if (!_classInfo.isVirtual())
			return filter;

		LyFilter		   result = new LyFilter(filter.getType(), filter.isNot());
		LyClassInfo		   realClass = _classInfo.getRealClass();
		LyFieldInfo		   field;
		int				   i, n = filter.getSize();
		LyFilterElement    element;
		LyFilterExpression filterExp;
		LyFilterRelation   filterRel;
		LyFilter		   subFilter;

		for (i = 0; i < n; i++)
		{
			element = filter.getElement(i);
			if (element instanceof LyFilterExpression)
			{
				filterExp = (LyFilterExpression)element;
				field = filterExp.getFieldInfo();
				element = new LyFilterExpression(realClass.getRealFieldInfo(field),
						filterExp.isInverted(), filterExp.getCondition(), filterExp.getValue(),
						filterExp.getModifiers());
			}
			else if (element instanceof LyFilterRelation)
			{
				filterRel = (LyFilterRelation)element;
				field = filterRel.getRelationFieldInfo();
				element = new LyFilterRelation((LyRelationFieldInfo)realClass.getRealFieldInfo(
							field), filterRel.getFilter());
			}
			else if (element instanceof LyFilter)
			{
				subFilter = (LyFilter)element;
				element = getRealFilter(subFilter);
			}

			result.addElement(element);
		}
		return result;
	} // end method getRealFilter

	/**
	 * Return the list of all class infos to which objects of this workspace belong. This method is
	 * overriden to optimize it since we known from start to which class the objects of this
	 * workspace belong.
	 *
	 * @return List of classes to which objects of this workspace belong. For this type of
	 *         workspace, the result list will always contain only one class.
	 */
	public LyInfoList<LyClassInfo> getClasses()
	{
		LyInfoList<LyClassInfo> classes = new LyInfoList<LyClassInfo>();

		classes.addInfo(_classInfo);
		return classes;
	}

	/**
	 * Gets class info to which objects of this workspace belong.
	 *
	 * @return Class info specified for this workspace.
	 */
	public LyClassInfo getClassInfo()
	{
		return _classInfo;
	}

	/**
	 * Gets the filter of this class workspace.
	 *
	 * @return Filter specified for this class workspace.
	 */
	public LyFilter getFilter()
	{
		return _filter;
	}

	/**
	 * Gets object source from which objects of this workspace are loaded.
	 *
	 * @return Object source of this workspace.
	 */
	public LyObjectSource getObjectSource()
	{
		return getObjectSource(_classInfo);
	}

	/**
	 * Gets list of fieldinfos to load.
	 *
	 * @return An info list containing fields to load.
	 */
	public LyInfoList<LyFieldInfo> getFieldInfos()
	{
		return _fieldInfos;
	}

	/**
	 * Method for resetting this workspace. Objects contained in this workspace are removed and
	 * reloaded from the object source of the class info specified for this workspace.
	 */
	public synchronized void reset()
	{
		// Remove all objects.
		super.reset();

		Vector<LyFieldInfo> fields = null;

		if (_fieldInfos != null)
		{
			fields = _fieldInfos.getInfos();
			if (fields == null)
				fields = new Vector<LyFieldInfo>();
		}

		LyResponse	response;

		if (!isSorted())
		{
			// Reload objects and avoid existence checking on loaded objects
			_checkExistence = false;
			response = getObjectSource().getObjects(fields, null, _filter, -1, _local, new _LyResponseListener());
			_checkExistence = true;
		}
		else
		{
			// Change _sorted value to avoid on-fly sorting
			_sorted = false;
			// and avoid existence checking on loaded objects
			_checkExistence = false;

			// Reload objects
			response = getObjectSource().getObjects(fields, getRealSort(getSort()),
					_filter, -1, _local, new _LyResponseListener());

			// Sort results if needed or simply restore _sorted value
			if (!response.isSorted())
				setSorted(true);
			else
				_sorted = true;
			// and restore existence checking
			_checkExistence = true;
		}

		if ((response != null)
				&& (response.getStatus() == LyResponse.Status.ERROR))
			throw new LyRequestException(response.getError());
	} // end method reset

	/**
	 * Sets list of fieldinfos to load.
	 *
	 * @param fields An info list containing fields to load.
	 */
	public void setFieldInfos(LyInfoList<LyFieldInfo> fields)
	{
		_fieldInfos = fields;
		reset();
	}

	/**
	 * Returns a string representation of this object.
	 *
	 * @return This workspace returned as a string.
	 */
	public String toString()
	{
		return super.toString() + "\nclassInfo=" + _classInfo + ", filter=" + getFilter();
	}

	/* Implementation of LyDataSourceListener */

	/**
	 * Method for receiving an object creation. This method is overriden to check if the object
	 * matchs the filter specified on this workspace if it has one.
	 *
	 * @param  event Event notifying the creation of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectCreated(LyDataSourceEvent event)
	{
		LyObject object = (LyObject)event.getObject();

		// If there is a filter, the object must match it.
		if ((_filter != null) && !matchFilter(object))
			return false;

		return super.objectCreated(event);
	}

	/**
	 * Method for receiving an object modification. If the object concerned by the event was not in
	 * this workspace, that means it didn't match its filter, so we now check if modifications done
	 * change this and in that case, we add the object in the workspace. On the contrary, if the
	 * object was in the workspace, we check if the object still matches the filter, and if not, we
	 * remove it from this workspace.
	 *
	 * @param  event Event notifying the modification of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectModified(LyDataSourceEvent event)
	{
		// If the object's event is added or removed from the workspace
		// we let the addObject or removeObject method do the job.
		if (useEvent(event))
			return true;

		// Notify the change, the list may need to be resorted
		return super.objectModified(event);
	}

	/* Implementation of LyDataAdvancedSourceListener */

	/**
	 * Called when an object state is changed.
	 *
	 * @param  event Event notifying object state change.
	 * @return True if the event is used by the listener, false otherwise.
	 */
	public boolean objectStateChanged(LyAdvancedDataSourceEvent event)
	{
		// If the object's event is added or removed from the workspace
		// we let the addObject or removeObject method do the job.
		if (useEvent(event))
			return true;

		// Notify the change, the list may need to be resorted
		return super.objectStateChanged(event);
	}

	/**
	 * Method for receiving user event on an object.
	 *
	 * @param  event Event notifying the user event of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectUserEvent(LyAdvancedDataSourceEvent event)
	{
		// If the object's event is added or removed from the workspace
		// we let the addObject or removeObject method do the job.
		if (useEvent(event))
			return true;

		// Notify the change, the list may need to be resorted
		return super.objectUserEvent(event);
	}

	/**
	 * Test if the specified event is used to add or remove an object from this workspace.
	 *
	 * @param  event Event to test
	 * @return true if the event adds or removes the object'event to or from the workspace, false
	 *         otherwise.
	 */
	protected boolean useEvent(LyDataSourceEvent event)
	{
		// If one of related objects is in the parent context, the new
		// object is added in this context.
		LyObject object = (LyObject)event.getObject();

		// If the object is not in the list
		if (indexOf(object) < 0)
		{
			if (object.isDeleted())
				return true;
			// If adding is enabled and if the object matches the filter,
			// add it to the list.
			if ((_filter == null) || matchFilter(object))
			{
				addObject(object);
				return true;
			}

			return false;
		}

		// If the object is in the list, but doesn't match the filter anymore,
		// remove it.
		if ((_filter != null) && !matchFilter(object))
		{
			removeObject(object);
			return true;
		}

		return false;
	} // end method useEvent

	/**
	 * Internal method used to check if the given object matches workspace filter.
	 * @param object the object to match
	 * @return true if the object match the filter
	 */
	protected boolean matchFilter(LyFilter filter, LyObject object)
	{
		if (filter == null)
			return true;

		if ((_session != null) && filter.isDynamic())
		{
			LyFilter  newFilter = filter.cloneFilter(_session);
			return newFilter.match(object);
		}
		else
			return filter.match(object);
	}

	/**
	 * Internal method used to check if the given values match workspace filter.
	 */
	protected boolean matchFilter(LyFilter filter, LyValueSet values)
	{
		if (filter == null)
			return true;

		if ((_session != null) && filter.isDynamic())
		{
		    LyFilter  newFilter = filter.cloneFilter(_session);
			return newFilter.match(values);
		}
		else
			return filter.match(values);
	}

	/**
	 * Internal method used to check if the given object matches workspace filter.
	 */
	protected boolean matchFilter(LyObject object)
	{
		return matchFilter(_filter, object);
	}

	/* Implementation of LyResponseListener */

	/**
	 * Internal class that implements This allows to load objects without sorting them and sort them
	 * at the end of the loading if requiered.
	 */
	class _LyResponseListener
		implements LyResponseListener
	{

		/**
		 * Method for receiving a new object in a response.
		 *
		 * @param response  Response for which a new object is received.
		 * @param object    The new object returned in the response.
		 * @param values List of values that were changed for the object.
		 */
		public void newResponse(LyResponse response, Object object, LyValueSet values)
		{
			// Add the loaded object in our workspace
			addObject((LyObject)object);
		}

		/**
		 * Method for ending a response.
		 *
		 * @param response Response which end is notified.
		 */
		public void endResponse(LyResponse response)
		{
		}
	}
} // end class LyClassWorkSpace
