/**
 * LyContext.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;

import leon.app.*;
import leon.data.datainterface.*;
import leon.info.*;
import leon.peer.*;


/**
 * This class represents a context managed by the context manager. A context is associated with a
 * user session (LyDataSession).
 *
 * @see     LyObject
 * @see     LyOperation
 * @see     LyContextManager
 * @author  Lyria
 */
public class LyContext
{
    /** File version. */
    public static final String __VERSION = "$Id: LyContext.java 34887 2012-03-21 09:54:53Z jlsaouli $";

    /** Static counter for _LyUpdated instances. */
    static long _counter = System.currentTimeMillis();

    /** Status of the context. */
    public enum Status
    {
        /** the context is started but there is no any registered operation. */
        STARTED,
        /** the context is started ad there is one or more registered operations. */
        RUNNING,
        /** the context is in error after a commit operation. */
        ERROR,
        /** the context is committed. */
        COMPLETED,
        /** the context is rollbacked. */
        ROLLBACKED,
        /** the context is committing. */
        COMMITTING
    }

    /** Type of the context. */
    public enum Type
    {
        /** cancellable user context.*/
        USER_CANCELLABLE,
        /** not cancellable user context.*/
        USER_NOT_CANCELLABLE,
        /** automatic context.*/
        AUTO
    }

    /**
     * Indicates the insertion positions of the operations of this context into the parent operation
     * list.
     */
    public enum Position
    {
        /** Insert position: First */
        FIRST,
        /** Insert position: End */
        END,
        /** Insert position: Default */
        DEFAULT
    }

    /** Reference to the user session. */
    protected LyDataSession _session;

    /**
     * Name of the context : this string is used to display informations about the undo/redo
     * operations.
     */
    protected String _name;

    /** Type of this context. */
    protected Type _type;

    /** Current status of the context. */
    protected Status _status;

    /** Registered operations on this context. Used to rollback operation. */
    protected ArrayList<LyOperation> _operations;

    /** Saved operations on this context. Used when saving the context. */
    protected ArrayList<LyOperation> _savedOperations;

    /**
     * Parent context in case of linked contexts. This attribute is set when there is a conflict
     * resolution.
     */
    protected LyContext _parent;

    /** List of contexts that have this context as parent. */
    protected ArrayList<LyContext> _sons;

    /** Context manager for this context. */
    protected LyContextManager _contextManager;

    /** Physical transaction for this context. */
    protected LyTransaction _transaction;

    /** The last sent operation. */
    protected LyOperation _lastOperation;

    /** The list of objects created by this context. */
    protected LySimpleObjectList _createdObjects;

    /** The list of objects modified by this context but not committed. */
    protected LySimpleObjectList _modifiedObjects;

    /** The list of relations modified by this context. */
    protected Vector<_LyUpdated> _modifiedRelations;

    /** List of undone operations that may be redone. */
    protected ArrayList<LyOperation> _undoneOperations;

    /** List of operations that can be undone. */
    protected ArrayList<LyOperation> _operationsToUndo;

    /** The current operation. */
    protected LyOperation _currentOperation;

    /** To mark operations as undo ignore. */
    protected boolean _ignoreUndo;

    /** To mark operations as linked for undo/redo operations. */
    protected boolean _linkUndo;

    /**
     * Indicates that this context is internal. Generally internal contexts are used in forms when
     * there is a sub form (for composition relation by example). Internal son contexts don't block
     * commit or rollback operations.
     */
    protected boolean _internal;

    /** Context listener. */
    protected LyContextListener _listener;

    /**
     * Indicates if this context may be kept in the undo/redo list of the context manager. If true,
     * the context is not kept.
     */
    protected boolean _undoDisabled;

    /**
     * Controls applied on objects in this context.
     *
     * @see  LyObject for possible values.
     */
    protected int _control;

    /** Indicates if an undo/redo is in progress. */
    protected boolean _undoRedoRunning;

    /**
     * Indicates if this context may update relations after commit or not. If true, relations are
     * not updated.
     */
    protected boolean _disableUpdateRelations;

    /** Indicates if this context may send events or not. If true, events are not sent. */
    protected boolean _disableEvents;

    /** Last error that occurred when sending operation. */
    protected String _errorMessage;

    /** Indicates if operations in this context should be only local or not. */
    protected boolean _onlyLocal;

    /** Indicates if we must take the order of the sons operations. */
    protected boolean _keepOrder;

    /**
     * Indicates that if this context has got a parent with keepOrder=true, its operation must be
     * executed before the parents operations or after.
     */
    protected LyContext.Position _insertionPosition;

    /** Store the sons operations to be executed first. */
    protected ArrayList<LyOperation> _firstOperations;

    /** Store the sons operations to be executed finally. */
    protected ArrayList<LyOperation> _lastOperations;

    /**
     * Generated id table : store the association between old ObjectId and new id generated by the
     * peer layer.
     */
    protected HashMap<String, LyObjectId> _generatedObjectId;

    /** Immediate update mode. */
    protected Boolean _immediateUpdate;

    /**
     * Constructs a context. Initialization is done with the init method.
     *
     * @param  session  User session associated with the context.
     * @param  type     Type of the context.
     */
    protected LyContext(LyDataSession session, Type type)
    {
        init();

        _session = session;
        _type = ((type == null) ? Type.USER_CANCELLABLE : type);
    }

    /**
     * Initialization of the data of the instance.
     */
    protected void init()
    {
        _session = null;
        _name = null;
        setStatus(Status.STARTED);
        _operations = new ArrayList<LyOperation>();
        _savedOperations = new ArrayList<LyOperation>();
        _undoneOperations = new ArrayList<LyOperation>();
        _operationsToUndo = new ArrayList<LyOperation>();
        _parent = null;
        _sons = null;
        _contextManager = null;
        _transaction = null;
        _lastOperation = null;
        _ignoreUndo = false;
        _linkUndo = false;
        _currentOperation = null;
        _internal = false;
        _listener = null;
        _undoDisabled = false;
        _control = LyObject.CONTROL_FULL;
        _undoRedoRunning = false;
        _disableUpdateRelations = false;
        _disableEvents = false;
        _errorMessage = null;
        _onlyLocal = false;
        _keepOrder = false;
        _insertionPosition = LyContext.Position.DEFAULT;
        _firstOperations = new ArrayList<LyOperation>();
        _lastOperations = new ArrayList<LyOperation>();
        _immediateUpdate = null;
    } // end method init

    /**
     * Free this context.
     */
    protected void free()
    {
        Iterator<LyOperation> it = _operations.iterator();
        LyOperation           operation;

        while (it.hasNext())
        {
            operation = it.next();
            it.remove();
            operation.free();
        }

        it = _savedOperations.iterator();
        while (it.hasNext())
        {
            operation = it.next();
            it.remove();
            operation.free();
        }
    }

    /**
     * Accessor to the current status.
     *
     * @return  the current status
     */
    public final Status getStatus()
    {
        return _status;
    }

    /**
     * Tests if this context has been committed or rollbacked.
     *
     * @return  true if this context has been committed or rollbacked, false otherwise.
     */
    public boolean isEnded()
    {
        return ((_status == Status.ROLLBACKED) || (_status == Status.COMPLETED));
    }

    /**
     * Accessor to the associated user session.
     *
     * @return  the associated session
     */
    public final LyDataSession getSession()
    {
        return _session;
    }

    /**
     * Accessor to the type of this context.
     *
     * @return  the type of this context
     */
    public final Type getType()
    {
        return _type;
    }

    /**
     * Accessor to the parent of this context.
     *
     * @return  The parent context of this context (may be null).
     */
    public LyContext getParent()
    {
        return _parent;
    }

    /**
     * To get the context manager of this context.
     *
     * @return  The context manager of this context.
     */
    protected LyContextManager getContextManager()
    {
        if (_contextManager == null)
            _contextManager = _session.getContextManager();

        return _contextManager;
    }

    /**
     * Modifier of the name of this context.
     *
     * @param  name  The new name of this context.
     */
    public void setName(String name)
    {
        _name = name;
    }

    /**
     * Accessor to the name of this context.
     *
     * @return  The name of this context.
     */
    protected String getName()
    {
        return _name;
    }

    /**
     * Internal method to get the associated transaction of this context.
     *
     * @return  The associated transaction of this context.
     */
    protected LyTransaction getTransaction()
    {
        if (_transaction == null)
        {
            LyApplication        application = _session.getApplication();
            LyTransactionManager tm = application.getTransactionManager();

            _transaction = tm.begin();
        }

        return _transaction;
    }

    /**
     * Tests if this context is an internal context or not.
     *
     * @return  true if the context is internal, false otherwise.
     */
    boolean isInternal()
    {
        return _internal;
    }

    /**
     * Tests if this context is ignored for undo/redo operations by the context manager.
     *
     * @return  true if the context is ignored, false otherwise.
     */
    boolean isUndoDisable()
    {
        return _undoDisabled;
    }

    /**
     * Accessor to the error message produced on last operation sent.
     *
     * @return  the error message
     */
    public final String getErrorMessage()
    {
        return _errorMessage;
    }

    /**
     * Marks this context as ignored for undo/redo by the context manager.
     */
    public void disableUndo()
    {
        if (_undoDisabled)
            return;

        _undoDisabled = true;

        if (_sons != null)
        {
            int       i, n = _sons.size();
            LyContext son;

            for (i = 0; i < n; i++)
            {
                son = _sons.get(i);
                son.disableUndo();
            }
        }
    }

    /**
     * Marks this context as internal. This method is for internal use only.
     */
    public void setInternal()
    {
        _internal = true;
        disableUndo();
    }

    /**
     * Sets listener of this context.
     *
     * @param  listener  a context listener
     */
    public void setListener(LyContextListener listener)
    {
        _listener = listener;
    }

    /**
     * Modifier of the current status of this context.
     *
     * @param  status  The new status of this context.
     */
    protected void setStatus(Status status)
    {
        _status = status;
    }

    /**
     * Sets the parent of this context.
     *
     * @param      parent  the parent of this context.
     *
     * @exception  IllegalStateException     if this context has already a parent and if the new
     *                                       parent is not null.
     * @exception  IllegalArgumentException  if the types of the current context and the specified
     *                                       context are not compatible.
     */
    protected void setParent(LyContext parent)
    {
        if ((parent == null) && (_parent != null))
            _parent.removeSon(this);
        else if ((_parent != null) && (_parent != parent))
            // A context can't change of parent.
            throw new IllegalStateException("BF AppEngine : can't change parent of a context");
        else if ((parent != null) && (_savedOperations.size() > 0))
            // A context with saved operations can't have a parent.
            throw new IllegalStateException("BF AppEngine : can't set a parent to a context with saved operations");

        if (_parent == parent)
            return;

        _parent = parent;

        if (_parent != null)
        {
            // FSC04.081 : we must test if the specified context can be the parent
            // of this context.
            Type type = getType();
            Type contextType = parent.getType();

            if ((contextType == Type.USER_NOT_CANCELLABLE)
                    && (type != Type.USER_NOT_CANCELLABLE))
                throw new IllegalArgumentException(
                    "BF AppEngine : a user not cancellable context can only be a parent of a user not cancellable context");

            if ((contextType == Type.AUTO)
                    && (type != Type.AUTO))
                throw new IllegalArgumentException(
                    "BF AppEngine : an automatic context can only be a parent of an automatic context");

            _parent.addSon(this);
        }
    } // end method setParent

    /**
     * Adds a son to the list of sons of this context.
     *
     * @param      son  The son to add
     *
     * @exception  IllegalStateException     if the son has another parent context.
     * @exception  IllegalArgumentException  if the types of the current context and the specified
     *                                       context are not compatible.
     */
    public void addSon(LyContext son)
    {
        if (isAncestor(son))
            throw new IllegalArgumentException(
                "BF AppEngine : can't add the given context as son because it's an ancestor of the current context");

        if (_savedOperations.size() > 0)
            // A context with saved operations can't have a son.
            throw new IllegalStateException("BF AppEngine : can't add a son to a context with saved operations");

        if (isEnded())
            // Can't add a son to an ended context.
            throw new IllegalStateException("BF AppEngine : can't add a son to an ended context");

        // FSC04.081 : we must test if the specified context can be a son
        // of this context.
        Type type = getType();
        Type contextType = son.getType();

        if ((type == Type.USER_NOT_CANCELLABLE)
                && (contextType != Type.USER_NOT_CANCELLABLE))
            throw new IllegalArgumentException(
                "BF AppEngine : a son of a user not cancellable context must be a user not cancellable context");

        if ((type == Type.AUTO)
                && (contextType != Type.AUTO))
            throw new IllegalArgumentException(
                "BF AppEngine : a son of an automatic context must be an automatic context");

        if (_sons == null)
            _sons = new ArrayList<LyContext>();

        // To avoid loop
        if (_sons.contains(son))
            return;

        LyContext parent = son.getParent();

        if ((parent != null) && (parent != this))
            // A context can't change of parent.
            throw new IllegalStateException("BF AppEngine : can't change parent of a context");

        _sons.add(son);
        try
        {
            son.setParent(this);
        }
        catch (RuntimeException e)
        {
            // If an excepton is raised when setting parent, remove son to remain consistent
            _sons.remove(son);
            throw e;
        }

        // Puts son in the same state as its parent
        son.enableControl(LyObject.CONTROL_FULL, false);
        son.enableControl(_control, true);

        if (_disableEvents)
            son.disableEvents();

        if (_undoDisabled)
            son.disableUndo();

        if (_disableUpdateRelations)
            son.disableUpdateRelations();

        if (_onlyLocal)
            son.setOnlyLocal(true);
    } // end method addSon

    /**
     * To get all the sons of this context.
     *
     * @return  A collection of the sons of this context or null if this context has no son.
     */
    public Collection<LyContext> getSons()
    {
        ArrayList<LyContext> result = null;
        int                  i, n = ((_sons == null) ? 0 : _sons.size());
        LyContext            son;

        for (i = 0; i < n; i++)
        {
            son = _sons.get(i);
            if (!son.isInternal())
            {
                if (result == null)
                    result = new ArrayList<LyContext>();

                result.add(son);
            }
        }

        return result;
    }

    /**
     * To get all the internal sons context of this context. getInternalSons() is for internal use,
     * applications should use getSons() instead.
     *
     * @return  A collection of the sons of this context or null if this context has no internal
     *          son.
     */
    public Collection<LyContext> getInternalSons()
    {
        return getInternalSons(false);
    }

    /**
     * To get all the internal sons context of this context. getInternalSons() is for internal use,
     * applications should use getSons() instead.
     *
     * @param   allInternalAndOthers  all sons whatever they are internal or not
     *
     * @return  A collection of the sons of this context or null if this context has no internal
     *          son.
     */
    public Collection<LyContext> getInternalSons(boolean allInternalAndOthers)
    {
        ArrayList<LyContext> result = null;
        int                  i, n = ((_sons == null) ? 0 : _sons.size());
        LyContext            son;

        for (i = 0; i < n; i++)
        {
            son = _sons.get(i);
            if (allInternalAndOthers || son.isInternal())
            {
                if (result == null)
                    result = new ArrayList<LyContext>();

                result.add(son);
            }
        }

        return result;
    }

    /**
     * Remove a son from the list of sons of this context.
     *
     * @param  son  The son context to remove.
     */
    protected void removeSon(LyContext son)
    {
        if (_sons == null)
            return;

        // To avoid loop
        if (!_sons.contains(son))
            return;

        _sons.remove(son);
        son.setParent(null);
    }

    /**
     * Tests if the specified context is a son of this context.
     *
     * @param   context  The context to test
     *
     * @return  true if the given context is a son of the current context, false otherwise.
     */
    public boolean isSon(LyContext context)
    {
        return ((_sons != null) && (_sons.contains(context)));
    }

    /**
     * Tests if the specified context is an ancestor of this context. A context C1 is an ancestor of
     * a context C2 if C1 is the parent of C2 or if a son of C1 is an ancestor of C2.
     *
     * @param   context  The context to test
     *
     * @return  true if the given context is an ancestor of the current context, false otherwise.
     */
    boolean isAncestor(LyContext context)
    {
        LyContext c = getParent();

        if (context == c)
            return true;
        else if (context._sons != null)
        {
            ArrayList<LyContext> sons = context._sons;
            Iterator<LyContext>  it = sons.iterator();

            while (it.hasNext())
            {
                c = it.next();
                if (c == this)
                    continue;

                if (isAncestor(c))
                    return true;
            }
        }

        return false;
    }

    /**
     * Gets the first common ancestor between this context an the specified context.
     *
     * @param   context  The context to search a common ancestor
     *
     * @return  The first common ancestor if exists, null otherwise.
     */
    LyContext getCommonAncestor(LyContext context)
    {
        LyContext c = getParent();

        if (c == null)
            return null;

        if (context.isAncestor(c))
            return c;

        return c.getCommonAncestor(context);
    }

    /**
     * Gets the root context of this context.
     *
     * @return  The root context of this context or this if this context has no parent.
     */
    public LyContext getRootContext()
    {
        LyContext c = getParent();

        if (c == null)
            return this;

        return c.getRootContext();
    }

    /**
     * Tests if this context has some not internal sons or not.
     *
     * @return  true if this context have some not internal sons, false otherwise.
     */
    protected boolean hasSons()
    {
        if (_sons != null)
        {
            Iterator<LyContext> it = _sons.iterator();
            LyContext           son;

            while (it.hasNext())
            {
                son = it.next();
                if (!son.isInternal())
                    return true;
            }
        }

        return false;
    }

    /**
     * To add an operation in the list of operations managed by this context.
     *
     * @param  operation  the operation to add.
     */
    protected void addOperation(LyOperation operation)
    {
        if (_undoRedoRunning)
            return;

        if (_currentOperation != null)
            _currentOperation.addOperation(operation);
        else
        {
            _operations.add(operation);
            addOperationToUndo(operation);
        }
    }

    /**
     * To add an operation in the list of operations to undo in this context.
     *
     * @param  operation  the operation to add.
     */
    protected void addOperationToUndo(LyOperation operation)
    {
        if (_undoRedoRunning)
            return;

        if ((isInternal()) && (_parent != null))
        {
            if (!operation.isUndoIgnored())
                operation.setUndoIgnored(_ignoreUndo);

            operation.setLinked(_linkUndo);
            _parent.addOperationToUndo(operation);
        }
        else
        {
            if (operation.getType() == LyOperation.Type.LIST)
            {
                // Remove all operations already in the list.
                Iterator<LyOperation> it = operation.iterator();
                LyOperation           op;

                while (it.hasNext())
                {
                    op = it.next();
                    if (_operationsToUndo.contains(op))
                        _operationsToUndo.remove(op);
                }
            }

            _operationsToUndo.add(operation);
            _undoneOperations.clear();
            if (!operation.isUndoIgnored())
                operation.setUndoIgnored(_ignoreUndo);

            operation.setLinked(_linkUndo);
        }
    } // end method addOperationToUndo

    /**
     * Internal method to reset all undo/redo operations.
     */
    protected void clearUndoRedo()
    {
        _operationsToUndo.clear();
        _undoneOperations.clear();
    }

    /**
     * Insert the operation in the first operation list.
     *
     * @param  operation  the operation to insert
     */
    private void addToFirstOperation(LyOperation operation)
    {
        _firstOperations.add(operation);
    }

    /**
     * Insert the operation in the last operation list.
     *
     * @param  operation  the operation to insert
     */
    private void addToLastOperation(LyOperation operation)
    {
        _lastOperations.add(operation);
    }

    /**
     * To add an operation in the list of operations managed by this context.
     *
     * @param  operations  the operations to add.
     */
    protected void addOperations(ArrayList<LyOperation> operations)
    {
        addOperations(operations, LyContext.Position.DEFAULT);
    }

    /**
     * To add an operation in the list of operations managed by this context.
     *
     * @param  operations      the operations to add.
     * @param  insertPosition  the insertion position
     */
    private void addOperations(ArrayList<LyOperation> operations, Position insertPosition)
    {
        if (_undoRedoRunning)
            return;

        if ((operations == null) || (operations.size() == 0))
            return;

        // Build a new operation with all operations and add it to the parent.
        LyOperation operation = new LyOperation(this, operations);

        if (_currentOperation != null)
            _currentOperation.addOperation(operation);
        else
        {
            // We must keep the order of sons operations
            // FLANDRES/FFT00036
            if (_keepOrder)
            {
                if (insertPosition == LyContext.Position.FIRST)
                    addToFirstOperation(operation);
                else if (insertPosition == LyContext.Position.END)
                    addToLastOperation(operation);
            }

            // Add the new operation at the right position
            LyOperation firstOperation = operations.get(0);
            long        firstDate = firstOperation.getCreationDate();
            LyOperation current;
            int         i = 0, n = _operations.size();

            for (i = 0; i < n; i++)
            {
                current = _operations.get(i);
                if (current.getCreationDate() > firstDate)
                    break;
            }

            if (i < n)
                _operations.add(i, operation);
            else
                _operations.add(operation);

            addOperationToUndo(operation);
        }
    } // end method addOperations

    /**
     * Get all operations of this context and of its sons sorted by date. Only not saved operations
     * are returned.
     *
     * @return  The list of all operations sorted by date.
     */
    ArrayList<LyOperation> getSortedOperations()
    {
        // Get all operations of this context and of its brothers.
        ArrayList<LyOperation> operations = new ArrayList<LyOperation>();
        int                    i, n;

        if (_operations != null)
        {
            if (!_keepOrder)
                operations.addAll(_operations);
            else
            {
                // Must keep the order of operation, so reoder them accrodingly
                ArrayList<LyOperation> clonedOperations = new ArrayList<LyOperation>(_operations);

                LyOperation           temp;
                Iterator<LyOperation> it;

                it = _firstOperations.iterator();
                while (it.hasNext())
                {
                    temp = it.next();
                    clonedOperations.remove(temp);
                    clonedOperations.add(0, temp);
                }

                it = _lastOperations.iterator();
                while (it.hasNext())
                {
                    temp = it.next();
                    clonedOperations.remove(temp);
                    clonedOperations.add(temp);
                }

                operations.addAll(clonedOperations);
            }
        }

        // Flaten the result : find LIST operations and add at root level
        // Need to sort itemized operations
        ArrayList<LyOperation>  flatOperations = new ArrayList<LyOperation>(operations.size());
        LyOperation             operation;
        Collection<LyOperation> operList;

        n = operations.size();
        for (i = 0; i < n; i++)
        {
            operation = operations.get(i);
            operList = operation.getOperations();
            flatOperations.addAll(operList);
        }

        if (_keepOrder)
        // Flandres/FFT00036
        // We want to keep operation order -> no sort.
        {
            partialSortInKeepOrder(flatOperations);
            return flatOperations;
        }
        // Sort the list by date.
        HashSet<LyOperation> set = new HashSet<LyOperation>(flatOperations);

        operations = new ArrayList<LyOperation>(set);
        Collections.sort(operations);

        return operations;
    } // end method getSortedOperations

    /**
     * Internal use : parial sorting set in keep order At least on the SAME object, the SET
     * operations must be executed in the right order FFT00458.
     *
     * @param  operations  operations to sort
     */
    protected void partialSortInKeepOrder(ArrayList<LyOperation> operations)
    {
        // At least on the SAME object, the SET operations must be executed in the right order
        if (operations == null)
            return;

        int n = operations.size();

        if (n <= 1)
            return; // nothing to do

        boolean                modified;
        LyObject               currentObject;
        long                   currentDate;
        int                    bestPosition;
        LyOperation            operation, oldOperation, operation2;
        ArrayList<LyOperation> sortedOperations = new ArrayList<LyOperation>();
        int                    i, j, start = 0;

        do
        {
            currentObject = null;
            currentDate = 0L;
            modified = false;

            for (i = start; (i < n) && (!modified); i++) // restart from 'start' when modified
            {
                operation = operations.get(i);

                if (operation.getType() == LyOperation.Type.SET)
                {
                    // First step, we do not work on any object yet, looking for a new one
                    if (currentObject == null)
                    {
                        if ((operation.getObject() != null)
                                && !sortedOperations.contains(operation))
                        {
                            currentObject = operation.getObject();
                            currentDate = operation.getCreationDate();
                            start = i; // We will start from this position afterwards
                        }
                    }
                    else
                    {
                        // Second step, we know the object and we want to compare to another one
                        if (operation.getObject() == currentObject)
                        {
                            // compare dates ?
                            if (operation.getCreationDate() < currentDate)
                            {
                                // got a wrong order, swap the operation found in step 1 with the best position (index is start)!

                                modified = true;
                                bestPosition = i;

                                for (j = i + 1; j < n; j++)
                                {
                                    operation2 = operations.get(j);
                                    if ((operation2.getObject() == currentObject)
                                            && (operation2.getCreationDate() < currentDate))
                                        bestPosition = j;
                                }

                                oldOperation = operations.remove(start);
                                operations.add(bestPosition, oldOperation);
                                sortedOperations.add(oldOperation);
                            }
                        }
                    }
                }
            }

            // we reached the end :
            if (!modified)
            {
                if (currentObject == null)
                    // nothing to sort
                    start = n; // Over
                else
                    start++;
            }
        }
        while (start < n);
    } // end method partialSortInKeepOrder

    /**
     * Get the list of operations of this context. Only not saved operatrions are returned. For
     * security and consistency, the list of returned operation is a copy of the original list.
     *
     * @return  The list of operations of this context.
     */
    public ArrayList<LyOperation> getOperations()
    {
        if (_operations == null)
            return null;
        else
            return new ArrayList<LyOperation>(_operations);
    }

    /**
     * Get the list of all (saved and not saved) operations of this context. For security and
     * consistency, the list of returned operation is a copy of the original list.
     *
     * @return  The list of operations of this context.
     */
    protected ArrayList<LyOperation> getAllOperations()
    {
        ArrayList<LyOperation> operations = null;

        if (_savedOperations != null)
            operations = new ArrayList<LyOperation>(_savedOperations);

        if (_operations != null)
        {
            if (operations == null)
                operations = new ArrayList<LyOperation>(_operations);
            else
                operations.addAll(_operations);
        }

        return operations;
    }

    /**
     * Internal method to add a list of created objects and updated relations to this context.
     *
     * @param  createdObjects    The created object to add.
     * @param  updatedRelations  The updated relations to add.
     */
    protected void addUpdatedRelations(LySimpleObjectList createdObjects, Vector<_LyUpdated> updatedRelations)
    {
        if (_undoRedoRunning)
            return;

        if (createdObjects != null)
        {
            if (_createdObjects == null)
            {
                _createdObjects = new LySimpleObjectList();
                _createdObjects.setSorted(false);
            }

            _createdObjects.addList(createdObjects);
        }

        if (updatedRelations != null)
        {
            if (_modifiedRelations == null)
                _modifiedRelations = new Vector<_LyUpdated>();

            _modifiedRelations.addAll(updatedRelations);
        }
    }

/*String getKey(LyObject source, LyRelationFieldInfo fieldInfo)
 *      {
 *              return source.hashCode() + fieldInfo.getId();
 *      }
 */
    /**
     * Add a list of created objects.
     *
     * @param  createdObjects  a list of created objects
     */
    protected void addCreatedObjectInternal(LySimpleObjectList createdObjects)
    {

    }

    /**
     * Internal method to update relations modified in this context. This method is used in the end
     * of the commit method to set correctly the object identifiers when these ids were changed in a
     * creation.
     *
     * @param  refresh  Indicated if events may be sent for these updates.
     * @param  target   the target object for the relations
     */
    protected void updateRelations(LyObject target)
    {
        if (_disableUpdateRelations)
            return;
        // No created objects : nothing to do
        if ((_createdObjects == null) || (_createdObjects.getSize() == 0))
            return;
        // No modified relations : nothing to do
        if ((_modifiedRelations == null) || (_modifiedRelations.size() == 0))
            return;

        int                i, n = _createdObjects.getSize();
        _LyUpdated         updated;
        LyObject           source;
        LySimpleObjectList sources = null;

        for (i = n - 1; i >= 0; i--)
        {
            source = _createdObjects.getObject(i);
            if (source.isDeleted())
            {
                _createdObjects.removeObject(source);
                source = source.getObjectId().getLogicalObject();
                if ((source != null)
                        && !source.isDeleted())
                    _createdObjects.addObject(source, i);
            }
        }

        // All created objects were deleted : nothing to do
        if (_createdObjects.getSize() == 0)
            return;

        // Update the relations.
        Collections.sort(_modifiedRelations);

        n = _modifiedRelations.size();
        for (i = 0; i < n; i++)
        {
            updated = _modifiedRelations.elementAt(i);
            source = updated.update(target);
            if (source == null)
            {
                if (target != null)
                    continue;

                source = updated._source;
                if (!source.hasNewValues()
                        && !source.hasPendingValues())
                    continue;
            }

            if (sources == null)
                sources = new LySimpleObjectList();

            if (!sources.contains(source))
                sources.addObject(source);
        }

        if ((sources != null)
                && (target == null))
        {
            int operFlags = LyObject.OPER_NONE;

            for (i = 0, n = sources.getSize(); i < n; i++)
            {
                source = sources.getObject(i);
                if ((source.getPeer() == null)
                        && !source.getClassInfo().isLocal())
                    operFlags |= LyObject.OPER_LOCAL;

                source.set(null, operFlags, LyObject.CONTROL_NONE);
            }
        }
    } // end method updateRelations

    /**
     * Internal method used to send an operation of this context.
     *
     * @param   operation  The operation to send.
     * @param   local      Indicates if the operation is local or not.
     * @param   refresh    Indicates if the operation may updated object or not.
     *
     * @return  the response of this operation.
     */
    protected LyResponse sendOperation(LyOperation operation)
    {
        if (_currentOperation == null)
            _currentOperation = operation;

        boolean control = (getType() == Type.USER_CANCELLABLE);

        LyResponse response = operation.send(true, false, false, (control ? _control : LyObject.CONTROL_NONE));

        if (_currentOperation == operation)
            _currentOperation = null;

        if ((response != null)
                && (response.getStatus() == LyResponse.Status.ERROR))
            _errorMessage = response.getError();
        else
            _errorMessage = null;

        return response;
    }

    /**
     * This method send all pending operations locally to update objects.
     */
    public void sendEvents()
    {
        sendEvents(true, false);
    }

    /**
     * Internal method to send events corresponding to the not saved operations of this context.
     *
     * @param  local       Indicates if the operation to consider is local or not.
     * @param  rollbacked  Indicated if the operation is rollbacked or not.
     */
    protected void sendEvents(boolean local, boolean rollbacked)
    {
        if (_disableEvents)
            return;

        LyOperation                  operation;
        int                          i, n = _operations.size();
        int                          idx, limit;
        ArrayList<LyDataSourceEvent> events = null, operEvents;

        if (_lastOperation != null)
            idx = _operations.indexOf(_lastOperation);
        else
            idx = -1;

        if (rollbacked)
        {
            i = n - 1;
            limit = idx;
        }
        else
        {
            i = idx + 1;
            limit = n;
        }

        while (i != limit)
        {
            operation = _operations.get(i);
            operEvents = operation.getEvents(local, rollbacked);
            if (operEvents != null)
            {
                if (events == null)
                    events = operEvents;
                else
                    events.addAll(operEvents);
            }

            if (rollbacked)
                i--;
            else
                i++;
        }

        sendEvents(events);
    } // end method sendEvents

    /**
     * Send grouped events specified in the parameter list.
     *
     * @param  events  List of events to send.
     */
    protected void sendEvents(ArrayList<LyDataSourceEvent> events)
    {
        if ((_disableEvents)
                || (events == null)
                || events.isEmpty())
            return;

        LyApplication  application = _session.getApplication();
        LyEventManager em = application.getEventManager();

        em.postEvents(events);
    }

    /**
     * To commit the context. All registered operations are committed.
     *
     * @return     the status of the context at the end of the commit operation. This status could
     *             be : Status.ERROR if there is one error in an object updated by this context,
     *             Status.COMPLETED if the commit operation succeed, Status.ROLLBACKED if the
     *             context has been rollbacked.
     *
     * @exception  IllegalStateException  if the context has some sons contexts.
     */
    public Status commit()
    {
        if ((_status != Status.RUNNING)
                && (_status != Status.STARTED)
                && (_status != Status.ERROR))
            return getStatus();

        if (hasSons())
            throw new IllegalStateException("BF AppEngine : can't commit a context with sons");

        if (_modifiedObjects != null)
        {
            // Commit modified objects that have not been committed.
            LyObject object;

            while (_modifiedObjects.getSize() > 0)
            {
                object = _modifiedObjects.getObject(0);
                if (object.isDeleted()
                        || !object.hasNewValues())
                    _modifiedObjects.removeObject(object);
                else
                {
                    setObject(object);

                    if ((_status != Status.RUNNING)
                            && (_status != Status.STARTED)
                            && (_status != Status.ERROR))
                        return getStatus();
                }
            }
        }

        setStatus(Status.COMMITTING);

        if ((_control != LyObject.CONTROL_NONE)
                && (getType() != Type.USER_CANCELLABLE))
        {
            // We must control all updated objects before committing.
            // Only modified and created objects may be controlled.
            HashMap<LyObject, LyInfoList<LyFieldInfo>> objects = getUpdatedObjects();
            Set<LyObject>                              keys = objects.keySet();
            Iterator<LyObject>                         it = keys.iterator();
            LyObject                                   object;
            LyInfoList<LyFieldInfo>                    fieldInfos;

            while (it.hasNext())
            {
                object = it.next();
                if (object.isDeleted())
                    continue;

                fieldInfos = objects.get(object);
                object.control(fieldInfos, _control);
                object.controlParent(_control & LyObject.CONTROL_FIELD_CONTROLS);
            }
        }

        // Check context before committing
        if ((_control != LyObject.CONTROL_NONE)
                && (getErrorLevel() == LyObject.LEVEL_ERROR))
        {
            // Can't commit, there are problems
            setStatus(Status.ERROR);
            return getStatus();
        }

        // Commit internal sons
        if ((_sons != null) && (!_sons.isEmpty()))
        {
            // We must work on a duplicate list to avoid
            // problems in the loop
            ArrayList<LyContext> sons = new ArrayList<LyContext>(_sons);
            int                  i, n = sons.size();
            LyContext            son;

            for (i = 0; i < n; i++)
            {
                son = sons.get(i);
                son.commit();
            }

            _sons = null;
        }

        Status newStatus = Status.COMPLETED;

        // If this context has a parent we must give all operations to it.
        if (_parent != null)
        {
            // If we have to keep operation order, our parent has to do it too.
            if (_keepOrder)
                _parent.setKeepOrder(_keepOrder);

            // Reorder operation if necessary.
            ArrayList<LyOperation> clonedOperations = new ArrayList<LyOperation>(_operations);

            LyOperation           temp;
            Iterator<LyOperation> it;

            it = _firstOperations.iterator();
            while (it.hasNext())
            {
                temp = it.next();
                clonedOperations.remove(temp);
                clonedOperations.add(0, temp);
            }

            it = _lastOperations.iterator();
            while (it.hasNext())
            {
                temp = it.next();
                clonedOperations.remove(temp);
                clonedOperations.add(temp);
            }

            // Add all operations to parent
            _parent.addOperations(clonedOperations, _insertionPosition);

            // Add updated relations and created objects to the parent
            _parent.addUpdatedRelations(_createdObjects, _modifiedRelations);

            _parent.addGeneratedObjectId(_generatedObjectId);

            // Send pending events :
            // events are sent only if this context is Type.USER_CANCELLABLE or
            // if _parent is Type.USER_CANCELLABLE.
            // TODO We must verify if events may be sent : we must respect chronological order.
            if ((getType() == Type.USER_CANCELLABLE)
                    || (_parent.getType() == Type.USER_CANCELLABLE))
            {
                if ((_session != null)
                        && (_session.getApplication().getCurrentContext() == this))
                {
                    // Flandres FFT
                    // Send event will cause an update in daemons which should use parent context if any
                    _session.getApplication().setCurrentContext(_parent);
                    sendEvents();
                    _session.getApplication().setCurrentContext(this);
                }
                else
                    sendEvents(); // Keep myself as current context (should not happen)
            }
            // Maybe we must send events ?
            _operations = new ArrayList<LyOperation>();

            // Remove the context from the parent.
            _parent.removeSon(this);
        }
        else
        {
            _transaction = getTransaction();

            // Get all my operations sorted by date and commit them in chronological order.
            ArrayList<LyOperation>       operations = getSortedOperations();
            LyOperation                  operation;
            ArrayList<LyDataSourceEvent> events = null, operEvents;
            boolean                      control = (_parent != null) && isInternal();
            LyResponse                   response;
            LyObjectId                   idBefore = null, idAfter = null;
            int                          i, n = operations.size();

            _errorMessage = null;

            for (i = 0; i < n; i++)
            {
                idBefore = null;
                idAfter = null;

                operation = operations.get(i);

                if (((operation.getType() == LyOperation.Type.CREATE)
            				|| (operation.getType() == LyOperation.Type.SET))
                        && (operation._object != null)
                        && (operation._values == null))
                    // Look if we need to swap generated ids by new ids in SET operations
                    idBefore = operation._object.getObjectId();

                if (_generatedObjectId != null)
                    updateOperationValues(operation);

                // Never send operations with refresh flags : if this context is
                // "immediate gui update" then the events have been sent during the
                // local operation. If this context is not "immediate gui update" then
                // the events will be grouped and sent later in the sendEvents call.
                response = operation.send(false, false, false, (control ? _control : LyObject.CONTROL_NONE));

                if ((response != null)
                        && (response.getStatus() == LyResponse.Status.ERROR))
                {
                    setStatus(Status.ERROR);

                    if (_errorMessage == null)
                        _errorMessage = response.getError();
                    else if (_errorMessage.indexOf(response.getError()) < 0)
                        _errorMessage += '\n' + response.getError();

                    break;
                }

                if (!_disableEvents)
                {
                    // XXX Get all events as non local
                    operEvents = operation.getEvents(false, false);
                    // operEvents = operation.getEvents(true, false);
                    if (operEvents != null)
                    {
                        if (events == null)
                            events = operEvents;
                        else
                            events.addAll(operEvents);
                    }
                }

                if ((idBefore != null) && (operation._object != null))
                {
                    idAfter = operation._object.getObjectId();

                    if (!idBefore.equals(idAfter))
                    {
                        if (_generatedObjectId == null)
                            _generatedObjectId = new HashMap<String, LyObjectId>();

                        _generatedObjectId.put(idBefore.getFullString(), idAfter);

                        // TODO : optimize by updating only relations that point on created object
                        if (((operation.getType() == LyOperation.Type.CREATE)
                				|| (operation.getType() == LyOperation.Type.SET))
                            && (operation._onlyNotLocal
                                || (!operation._onlyLocal && !operation._onlyNotLocal)))
                            updateRelations(operation._object);
                    }
                }
            }

            if (_status == Status.COMPLETED)
                // Update relations to have the good object ids
                updateRelations(null);

            // Commit the transaction.
            LyApplication        application = _session.getApplication();
            LyTransactionManager tm = application.getTransactionManager();

            tm.end(_transaction);

            if (_transaction.getStatus() == LyTransaction.Status.ERROR)
            {
                rollback(false);
                newStatus = Status.ROLLBACKED;
            }
            else
            {
                // Send events.
                sendEvents(events);
            }

            // We must set the list of operations with all committed operations including
            // brothers operations.
            _operations = operations;
        }

        setStatus(newStatus);

        // Remove this context from the context manager.
        LyContextManager cm = getContextManager();

        cm.removeContext(this);
        _transaction = null;

        return getStatus();
    } // end method commit

    /**
     * Update values stored in the given operation.
     *
     * @param  operation  the operation to update
     */
    protected void updateOperationValues(LyOperation operation)
    {
        if ((_generatedObjectId != null)
                && (operation.getType() == LyOperation.Type.SET)
                && (operation._object != null)
                && (operation._values != null))
        {
            // Look if we need to swap generated ids by new ids in SET operations
            int        j, m = operation._values.getSize();
            LyObjectId idBefore = null, idAfter = null;
            LyValue    value;

            ;
            boolean      modified;
            LyObjectId[] idList;
            int          k, nv;

            for (j = m - 1; j >= 0; j--)
            {
                value = operation._values.getFieldValue(j);
                modified = false;

                if (value instanceof LyRelationValue)
                {
                    nv = ((LyRelationValue)value).getValueCount();
                    idList = new LyObjectId[nv];

                    for (k = 0; k < nv; k++)
                    {
                        idBefore = ((LyRelationValue)value).getObjectId(k);
                        idList[k] = idBefore;

                        if (idBefore != null)
                        {
                            idAfter = _generatedObjectId.get(idBefore.getFullString());

                            if (idAfter != null)
                            {
                                idList[k] = idAfter;
                                modified = true;
                            }
                        }
                    }

                    if (modified)
                    {
                        operation._values.removeValue(value.getField());
                        operation._values.addValue(new LyRelationValue((LyRelationFieldInfo)value.getField(), idList));
                    }

                    idBefore = null;
                }
            }
        }

        if (operation._operations != null)
        {
            int         i, n = operation._operations.size();
            LyOperation subOperation;

            for (i = 0; i < n; i++)
            {
                subOperation = operation._operations.get(i);
                if (subOperation != operation)
                    updateOperationValues(subOperation);
            }
        }
    }

    /**
     * Internal method used after operations of a context have been commited to notify the context
     * it has been commited by another context.
     *
     * @param  brotherContext  Brother context that committed this context.
     */
    void committed(LyContext brotherContext)
    {
        // Get all operations and remove them.
        int i, n = _operations.size();

        for (i = n - 1; i >= 0; i--)
            _operations.remove(i);

        if (_parent != null)
            _parent.removeSon(this);

        if (_transaction != null)
        {
            // Commit the transaction.
            LyApplication        application = _session.getApplication();
            LyTransactionManager tm = application.getTransactionManager();

            tm.end(_transaction, true);
        }

        setStatus(Status.COMPLETED);
        if (_listener != null)
            _listener.contextEnded(this, brotherContext);

        // Remove this context from the context manager.
        LyContextManager cm = getContextManager();

        // Ignore undo/redo for a brother context.
        setIgnoreUndo(true);
        cm.removeContext(this);
        _transaction = null;
    } // end method committed

    /**
     * Internal method used before operations of a context have been commited to notify the context
     * it will be commited by another context.
     *
     * @param  brotherContext  Brother context that committed this context.
     */
    void commit(LyContext brotherContext)
    {
        if (hasSons())
            throw new IllegalStateException("BF AppEngine : can't commit a context with sons");

        // commit transient sons
        if ((_sons != null) && (!_sons.isEmpty()))
        {
            // We must work on a duplicate list to avoid
            // problems in the loop
            ArrayList<LyContext> sons = new ArrayList<LyContext>(_sons);
            int                  i, n = sons.size();
            LyContext            son;

            for (i = 0; i < n; i++)
            {
                son = sons.get(i);
                son.commit();
            }

            _sons = null;
        }

        // This context is committed by a brother so we must not
        // keep it in undo/redo list.
        disableUndo();

        brotherContext.addUpdatedRelations(_createdObjects, _modifiedRelations);
        brotherContext.addGeneratedObjectId(_generatedObjectId);
        _createdObjects = null;
        _modifiedRelations = null;
        _generatedObjectId = null;
    } // end method commit

    /**
     * Rollback the current context.
     */
    public void rollback()
    {
        rollback(true);
    }

    /**
     * To rollback the context. Return to the initial state before starting the context.
     *
     * @param      sendEvents  indicates if rollback lust be notified
     *
     * @exception  IllegalStateException  if the context has some sons context.
     */
    protected void rollback(boolean sendEvents)
    {
        if ((_status != Status.RUNNING)
                && (_status != Status.STARTED)
                && (_status != Status.ERROR)
                && (_status != Status.COMMITTING))
            return;

        if (hasSons())
            throw new IllegalStateException("BF AppEngine : can't rollback a context with sons");

        setStatus(Status.ROLLBACKED);

        // Rollback internal sons
        if ((_sons != null) && (!_sons.isEmpty()))
        {
            // We must work on a duplicate list to avoid
            // problems in the loop
            ArrayList<LyContext> sons = new ArrayList<LyContext>(_sons);
            int                  i, n = sons.size();
            LyContext            son;

            for (i = 0; i < n; i++)
            {
                son = sons.get(i);
                son.rollback();
            }

            _sons = null;
        }

        // Get all my operations and all operations of my brothers sorted by date
        // and rollback them in reverse order.
        ArrayList<LyOperation> operations = getSortedOperations();

        // Get all operations and rollback them.
        LyOperation                  operation;
        int                          i, n = operations.size();
        ArrayList<LyDataSourceEvent> events = null, operEvents;
        boolean                      hasError = false;
        for (i = n - 1; i >= 0; i--)
        {
            operation = operations.get(i);
            hasError = (operation.getErrorLevel() == LyObject.LEVEL_ERROR);
            operation.rollback(false, _control);

            if (sendEvents && !hasError)
            {
                operEvents = operation.getEvents(true, true);
                if (operEvents != null)
                {
                    if (events == null)
                        events = operEvents;
                    else
                        events.addAll(operEvents);
                }
            }
        }

        // Remove useless events (set events if we delete the object just after...)
        if (events != null)
        {
            operEvents = new ArrayList<LyDataSourceEvent>(events);

            LyDataSourceEvent event, event2;
            Object            object;

            for (i = 0; i < events.size(); i++)
            {
                event = events.get(i);
                if (event.getType() == LyDataSourceEvent.TYPE_DELETE)
                {
                    object = event.getObject();
                    for (int j = 0; j < i; j++)
                    {
                        event2 = events.get(j);
                        if (event2.getObject() == object)
                            operEvents.remove(event2);
                    }
                }
                // Flandres FFT 461 : Set event send on deleted objects
                // the test above is not enough : object might be delete by another controller context in a tab area (wizard)
                if ((event.getType() == LyDataSourceEvent.TYPE_SET)
                        && (event.getObject() instanceof LyObject))
                {
                    LyObject obj = (LyObject)event.getObject();

                    if (obj.isDeleted())
                        operEvents.remove(event);
                }
            }

            events = operEvents;
        }

        if (_parent != null)
            _parent.removeSon(this);

        if (_transaction != null)
        {
            // Rollback the transaction.
            LyApplication        application = _session.getApplication();
            LyTransactionManager tm = application.getTransactionManager();

            tm.end(_transaction, false);
        }

        // Events must be sent after controls to have the good state in objects.
        if (sendEvents
                && (events != null))
            // Send group events for rollback operations
            sendEvents(events);

        setStatus(Status.ROLLBACKED);

        // Remove this context from the context manager.
        LyContextManager cm = getContextManager();

        cm.removeContext(this);
        _transaction = null;
    } // end method rollback

    /**
     * Internal method used after operations of a context have been rollbacked to notify the context
     * it has been rollbacked by another context. This method is used when operations of brothers
     * context have been rollbacked (they must be rollbacked in reverse chronologial order).
     *
     * @param  brotherContext  Brother context that rollbacked this context.
     */
    void rollbacked(LyContext brotherContext)
    {
        if (hasSons())
            throw new IllegalStateException("BF AppEngine : can't rollback a context with sons");

        // Rollback transient sons
        if ((_sons != null) && (!_sons.isEmpty()))
        {
            // We must work on a duplicate list to avoid
            // problems in the loop
            ArrayList<LyContext> sons = new ArrayList<LyContext>(_sons);
            int                  i, n = sons.size();
            LyContext            son;

            for (i = 0; i < n; i++)
            {
                son = sons.get(i);
                son.rollback();
            }

            _sons = null;
        }

        // Get all operations and remove them.
        int         i, n = _operations.size();
        LyOperation operation;

        for (i = n - 1; i >= 0; i--)
        {
            operation = _operations.remove(i);
            operation.free();
        }

        if (_parent != null)
            _parent.removeSon(this);

        if (_transaction != null)
        {
            // Rollback the transaction.
            LyApplication        application = _session.getApplication();
            LyTransactionManager tm = application.getTransactionManager();

            tm.end(_transaction, false);
        }

        setStatus(Status.ROLLBACKED);
        if (_listener != null)
            _listener.contextEnded(this, brotherContext);

        // Remove this context from the context manager.
        LyContextManager cm = getContextManager();

        cm.removeContext(this);
        _transaction = null;
    } // end method rollbacked

    /**
     * Delete an object in this context.
     *
     * @param      object  Object to delete.
     *
     * @return     Response to the deletion.
     *
     * @see        LyObject#delete(boolean)
     *
     * @exception  IllegalStateException        if the context has the status Status.ROLLBACKED or
     *                                          STATUS_COMPLETED.
     * @exception  LyIllegalOperationException  if this context has not the rights to do the
     *                                          deletion
     */
    public LyResponse deleteObject(LyObject object)
    {
        if (object == null)
            throw new IllegalArgumentException("BF AppEngine : object parameter can't be null");

        if ((_status == Status.COMPLETED)
                || (_status == Status.ROLLBACKED))
            throw new IllegalStateException("BF AppEngine : can't send operations on "
                + "a context in error or completed status");

        if (object.isDeleted())
            throw new IllegalStateException("BF AppEngine : can't delete object twice " + object);

        if (!object.checkObject()
                && ((_createdObjects == null) || !_createdObjects.contains(object)))
            throw new IllegalStateException("BF AppEngine : can't delete unknown object " + object);

        if (_status == Status.ROLLBACKED)
            return null;

        LyOperation operation = null;

        // Create an operation instance
        operation = new LyOperation(this, LyOperation.Type.DELETE, object);
        addOperation(operation);
        setStatus(Status.RUNNING);

        return sendOperation(operation);
    } // end method deleteObject

    /**
     * Create a new object in this context.
     *
     * @param      object  Object to create.
     *
     * @return     Response to the creation.
     *
     * @see        LyObject#create(boolean)
     *
     * @exception  IllegalStateException        if the context has the status Status.ROLLBACKED or
     *                                          STATUS_COMPLETED.
     * @exception  LyIllegalOperationException  if this context has not the rights to do the
     *                                          creation
     */
    public LyResponse createObject(LyObject object)
    {
        if (object == null)
            throw new IllegalArgumentException("BF AppEngine : object parameter can't be null");

        if ((_status == Status.COMPLETED)
                || (_status == Status.ROLLBACKED))
            throw new IllegalStateException("BF AppEngine : can't send operations on "
                + "a context in error or completed status");

        LyOperation foundOperation = null;
        boolean     contained = (_createdObjects != null) && _createdObjects.contains(object);

        if (object.checkObject()
                || contained)
        {
            if (object.getPeer() != null)
                throw new IllegalStateException("BF AppEngine : can't create object twice " + object);

            if (contained)
            {
                // Search for a create operation on this object.
                LyOperation op;
                int         i, n = ((_operations == null) ? 0 : _operations.size());

                for (i = 0; i < n; i++)
                {
                    op = _operations.get(i);
                    if ((op.getType() == LyOperation.Type.CREATE)
                            && (op.getObject() == object))
                    {
                        if (foundOperation == null)
                            foundOperation = op;
                        else
                            throw new IllegalStateException("BF AppEngine : can't create object twice " + object);
                    }
                }

                if (foundOperation != null)
                    foundOperation.setOnlyLocal();
            }
        }

        if (_status == Status.ROLLBACKED)
            return null;

        // Create an operation instance
        LyOperation operation;

        operation = new LyOperation(this, LyOperation.Type.CREATE, object);
        if (foundOperation != null)
            operation.setOnlyNotLocal();

        addOperation(operation);
        setStatus(Status.RUNNING);

        // Maybe some relations have been set before the creation so we
        // must add them to the list.
        addUpdatedRelations(object, object.getNewValues());

        if (!contained)
        {
            // Add the object to the list of created objects
            if (_createdObjects == null)
            {
                _createdObjects = new LySimpleObjectList(true);
                _createdObjects.setSorted(false);
            }

            _createdObjects.addObject(object);
        }

        return sendOperation(operation);
    } // end method createObject

    /**
     * Set an object in this context.
     *
     * @param      object  Object to modify.
     *
     * @return     Response to the modification.
     *
     * @see        LyObject#set(boolean)
     *
     * @exception  IllegalStateException        if the context has the status Status.ROLLBACKED or
     *                                          STATUS_COMPLETED.
     * @exception  LyIllegalOperationException  if this context has not the rights to do the
     *                                          modification
     */
    public LyResponse setObject(LyObject object)
    {
        if (object == null)
            throw new IllegalArgumentException("BF AppEngine : object parameter can't be null");

        if ((_status == Status.COMPLETED)
                || (_status == Status.ROLLBACKED))
            throw new IllegalStateException("BF AppEngine : can't send operations on "
                + "a context in error or completed status");

        if (object.isDeleted())
            throw new IllegalStateException("BF AppEngine : can't set deleted object " + object);

        if (!object.checkObject()
                && ((_createdObjects == null) || !_createdObjects.contains(object)))
            throw new IllegalStateException("BF AppEngine : can't set not created object " + object);

        LyValueSet newValues = object.getNewValues();

        if (!object.hasNewValues()
                && !object.hasPendingValues())
        {
            // Nothing done
            LyResponse response = new LyResponse(LyResponse.Operation.SET, object.getClassInfo());

            response.setStatus(LyResponse.Status.COMPLETED);

            if ((_modifiedObjects != null) && (_modifiedObjects.contains(object)))
                _modifiedObjects.remove(object);

            return response;
        }

        if (_status == Status.ROLLBACKED)
            return null;

        // Create an operation instance
        LyOperation operation;

        operation = new LyOperation(this, LyOperation.Type.SET, object);
        addOperation(operation);
        setStatus(Status.RUNNING);

        // Add the updated relations to the list
        addUpdatedRelations(object, newValues);

        if ((_modifiedObjects != null) && (_modifiedObjects.contains(object)))
            _modifiedObjects.remove(object);

        return sendOperation(operation);
    } // end method setObject

    /**
     * Set a value for an object in this context.
     *
     * @param      object   Object to modify.
     * @param      fieldId  Name of the fieldInfo to set.
     * @param      value    The value for this fieldInfo.
     *
     * @exception  IllegalStateException        if the context has the status Status.ROLLBACKED or
     *                                          STATUS_COMPLETED.
     * @exception  LyIllegalOperationException  if this context has not the rights to do the
     *                                          modification
     */
    public void setValue(LyObject object, String fieldId, Object value)
    {
        if (object == null)
            throw new IllegalArgumentException("BF AppEngine : object parameter can't be null");

        LyClassInfo classInfo = object.getClassInfo();
        LyFieldInfo fieldInfo = classInfo.getFieldInfo(fieldId);

        if (fieldInfo == null)
            throw new IllegalArgumentException("BF AppEngine : no field info with id "
                + fieldId + " found on class info " + classInfo);

        setValue(object, fieldInfo, value);
    }

    /**
     * Set a value for an object in this context.
     *
     * @param      object     Object to modify.
     * @param      fieldInfo  The fieldInfo to set.
     * @param      value      The value for this fieldInfo.
     *
     * @exception  IllegalStateException        if the context has the status Status.ROLLBACKED or
     *                                          STATUS_COMPLETED.
     * @exception  LyIllegalOperationException  if this context has not the rights to do the
     *                                          modification
     */
    public void setValue(LyObject object, LyFieldInfo fieldInfo, Object value)
    {
        LyValue fieldValue = fieldInfo.buildFieldValue(value);

        setValue(object, fieldValue);
    }

    /**
     * Set a value for an object in this context.
     *
     * @param      object  Object to modify.
     * @param      value   The value to set
     *
     * @exception  IllegalStateException        if the context has the status Status.ROLLBACKED or
     *                                          STATUS_COMPLETED.
     * @exception  LyIllegalOperationException  if this context has not the rights to do the
     *                                          modification
     */
    public void setValue(LyObject object, LyValue value)
    {
        LyValueSet values = new LyValueSet();

        values.addValue(value);
        setValues(object, values);
    }

    /**
     * Set values for an object in this context.
     *
     * @param      object  Object to modify.
     * @param      values  The values to set
     *
     * @exception  IllegalStateException        if the context has the status Status.ROLLBACKED or
     *                                          STATUS_COMPLETED.
     * @exception  LyIllegalOperationException  if this context has not the rights to do the
     *                                          modification
     */
    public void setValues(LyObject object, LyValueSet values)
    {
        if (object == null)
            throw new IllegalArgumentException("BF AppEngine : object parameter can't be null");

        if ((_status == Status.COMPLETED)
                || (_status == Status.ROLLBACKED))
            throw new IllegalStateException("BF AppEngine : can't send operations on "
                + "a context in error or completed status");

        if (object.isDeleted())
            throw new IllegalStateException("BF AppEngine : can't set deleted object " + object);

        if (!object.checkObject()
                && ((_createdObjects == null) || !_createdObjects.contains(object)))
        {
            // Send operation directly on the object without any context.
            object.setValues(null, values, LyObject.CONTROL_FULL, LyObject.OPER_NONE);
            return;
        }

        if (_status == Status.ROLLBACKED)
            return;

        // Create an operation instance
        LyOperation operation;

        operation = new LyOperation(this, LyOperation.Type.SET, object, values);
        addOperation(operation);
        setStatus(Status.RUNNING);

        // Add updated relations to the list.
        addUpdatedRelations(object, values);

        if ((_createdObjects == null) || !_createdObjects.contains(object))
        {
            if (_modifiedObjects == null)
            {
                _modifiedObjects = new LySimpleObjectList(false);
                _modifiedObjects.setSorted(false);
            }
    
            if (!_modifiedObjects.contains(object))
                _modifiedObjects.addObject(object);
        }

        object.setContextType(getType()); // We store the context trype because some control like readonly
        // field may differ when done by code or by user

        sendOperation(operation);

        object.setContextType(null); // Cleared
    } // end method setValues

    /**
     * Set the selection state of an object in this context.
     *
     * @param      object  Object to modify.
     * @param      state   The new selection state
     *
     * @exception  IllegalStateException  if the context has the status Status.ROLLBACKED or
     *                                    STATUS_COMPLETED or if the object does'nt exist.
     */
    public void setSelected(LyObject object, boolean state)
    {
        if (object == null)
            throw new IllegalArgumentException("BF AppEngine : object parameter can't be null");

        if ((_status == Status.COMPLETED)
                || (_status == Status.ROLLBACKED))
            throw new IllegalStateException("BF AppEngine : can't send operations on "
                + "a context in error or completed status");

        if (object.isDeleted())
            throw new IllegalStateException("BF AppEngine : can't set deleted object " + object);

        if (!object.checkObject()
                && ((_createdObjects == null) || !_createdObjects.contains(object)))
            throw new IllegalStateException("BF AppEngine : can't set not created object " + object);

        // Create an operation instance
        LyOperation operation;

        if (state)
            operation = new LyOperation(this, LyOperation.Type.SELECT, object);
        else
            operation = new LyOperation(this, LyOperation.Type.UNSELECT, object);

        addOperation(operation);
        setStatus(Status.RUNNING);

        sendOperation(operation);
    } // end method setSelected

    /**
     * Set the hidden state of an object in this context.
     *
     * @param      object  Object to modify.
     * @param      state   The new hidden state
     *
     * @exception  IllegalStateException  if the context has the status Status.ROLLBACKED or
     *                                    STATUS_COMPLETED or if the object does'nt exist.
     */
    public void setHidden(LyObject object, boolean state)
    {
        if (object == null)
            throw new IllegalArgumentException("BF AppEngine : object parameter can't be null");

        if ((_status == Status.COMPLETED)
                || (_status == Status.ROLLBACKED))
            throw new IllegalStateException("BF AppEngine : can't send operations on "
                + "a context in error or completed status");

        if (object.isDeleted())
            throw new IllegalStateException("BF AppEngine : can't set deleted object " + object);

        if (!object.checkObject()
                && ((_createdObjects == null) || !_createdObjects.contains(object)))
            throw new IllegalStateException("BF AppEngine : can't set not created object " + object);

        // Create an operation instance
        LyOperation operation;

        if (state)
            operation = new LyOperation(this, LyOperation.Type.HIDE, object);
        else
            operation = new LyOperation(this, LyOperation.Type.SHOW, object);

        addOperation(operation);
        setStatus(Status.RUNNING);

        sendOperation(operation);
    } // end method setHidden

    /**
     * Send an user evetn in this context.
     *
     * @param      object     Object concerned by the event.
     * @param      userEvent  The event to send
     *
     * @exception  IllegalStateException  if the context has the status Status.ROLLBACKED or
     *                                    STATUS_COMPLETED or if the object does'nt exist.
     */
    public void sendUserEvent(LyObject object, LyAdvancedDataSourceEvent userEvent)
    {
        if (object == null)
            throw new IllegalArgumentException("BF AppEngine : object parameter can't be null");

        if ((_status == Status.COMPLETED)
                || (_status == Status.ROLLBACKED))
            throw new IllegalStateException("BF AppEngine : can't send operations on "
                + "a context in error or completed status");

        if (object.isDeleted())
            throw new IllegalStateException("BF AppEngine : can't set deleted object " + object);

        if (!object.checkObject()
                && ((_createdObjects == null) || !_createdObjects.contains(object)))
            throw new IllegalStateException("BF AppEngine : can't set not created object " + object);

        // Create an operation instance
        LyOperation operation;

        operation = new LyOperation(this, object, userEvent);
        addOperation(operation);
        setStatus(Status.RUNNING);

        sendOperation(operation);
    }

    /**
     * Internal method used to add updated object ids mapping.
     *
     * @param  map  current id mapping
     */
    protected void addGeneratedObjectId(HashMap<String, LyObjectId> map)
    {
        _generatedObjectId = map;
    }

    /**
     * Internal method used to add updated relations to the list.
     *
     * @param  object  The object owner of the update relations.
     * @param  values  List of values for searching updated relations.
     */
    protected void addUpdatedRelations(LyObject object, LyValueSet values)
    {
        if ((_disableUpdateRelations)
                || (values == null))
            return;

        // Add the updated relations to the list.
        LyValue                    value;
        LyValue[]                  rv;
        LyFieldInfo                fieldInfo, fi;
        int                        i, n = values.getSize();
        LyObjectList               targets;
        ArrayList<LyObjectId>      targetIds;
        LySimpleObjectList         sol;
        _LyUpdatedRelation         updatedRelation;
        _LyUpdatedSimpleRelation   updatedSimpleRelation;
        _LyUpdatedArray            updatedArray;
        _LyUpdated                 updated;
        int                        j, k, p, nv, nc;
        LyObjectId                 id, createdId, targetId;
        LyObject                   target, createdObject;
        Hashtable<String, Integer> modifiedObjects = new Hashtable<String, Integer>();
        int                        pos;
        Integer                    index;
        String                     key;

        for (i = 0; i < n; i++)
        {
            value = values.getFieldValue(i);
            fieldInfo = (LyFieldInfo)value.getField();
            updated = null;
            targetIds = null;

            if (fieldInfo instanceof LyRelationFieldInfo)
            {
                // Search for for pointed objects : these objects may be created
                // in this context and may not be in the cache.
                LyRelationValue relValue = ((LyRelationValue)value);

                nv = relValue.getValueCount();

                sol = new LySimpleObjectList();
                sol.setSorted(false);

                nc = ((_createdObjects == null) ? 0 : _createdObjects.getSize());

                for (j = 0; j < nv; j++)
                {
                    id = ((LyRelationValue)value).getObjectId(j);
                    target = id.getLogicalObject();
                    if (target == null)
                    {
                        for (k = 0; k < nc; k++)
                        {
                            createdObject = _createdObjects.getObject(k);
                            createdId = createdObject.getObjectId();
                            if (createdId.equals(id))
                            {
                                sol.addObject(createdObject);
                                break;
                            }
                        }
                    }
                    else
                        sol.addObject(target);
                }

                targets = sol;

                // Some objects are not found (real objects created later in the context)
                // Flandres/FFT00234 : May happen during import and with immediateUpdate mode.
                nc = ((targets == null) ? 0 : targets.getSize());

                if (nc != nv)
                {
                    targetIds = new ArrayList<LyObjectId>();

                    for (k = 0; k < nv; k++)
                    {
                        targetIds = new ArrayList<LyObjectId>();

                        for (k = 0; k < nv; k++)
                        {
                            targetId = relValue.getObjectId(k);
                            if (targetId.getLogicalObject() == null)
                                targetIds.add(targetId);
                        }
                    }
                }

                if (nv > 1)
                {
                    updatedRelation = new _LyUpdatedRelation();
                    updatedRelation._source = object;
                    updatedRelation._targets = targets;
                    updatedRelation._targetIds = targetIds;
                    updatedRelation._relation = (LyRelationFieldInfo)fieldInfo;
                    updated = updatedRelation;
                }
                else
                {
                    updatedSimpleRelation = new _LyUpdatedSimpleRelation();
                    updatedSimpleRelation._source = object;
                    updatedSimpleRelation._target = ((nc == 0) ? null : targets.getObject(0));
                    updatedSimpleRelation._targetId = ((nv == nc) ? null : targetIds.get(0));
                    updatedSimpleRelation._relation = (LyRelationFieldInfo)fieldInfo;
                    updated = updatedSimpleRelation;
                }
            }
            else if (fieldInfo instanceof LyArrayFieldInfo)
            {
                fi = ((LyArrayFieldInfo)fieldInfo).getFieldInfo();
                if ((fi instanceof LyRelationFieldInfo)
                        && (value instanceof LyArrayValue))
                {
                    p = ((LyArrayValue)value).getValueCount();
                    if (p > 0)
                    {
                        updatedArray = new _LyUpdatedArray();
                        updatedArray._targets = new LySimpleObjectList[p];

                        @SuppressWarnings("unchecked")
                        ArrayList<LyObjectId>[] temp = new ArrayList[p];

                        updatedArray._targetIds = temp;

                        rv = ((LyArrayValue)value).getArrayFieldValue();
                        for (j = 0; j < p; j++)
                        {
                            LyRelationValue relValue = ((LyRelationValue)rv[j]);

                            targets = relValue.getObjectList();
                            updatedArray._targets[j] = targets;
                            // Some objects are not found (real objects created later in the context)
                            nv = relValue.getValueCount();
                            nc = ((targets == null) ? 0 : targets.getSize());

                            if (nc != nv)
                            {
                                targetIds = new ArrayList<LyObjectId>();

                                for (k = 0; k < nv; k++)
                                {
                                    targetId = relValue.getObjectId(k);
                                    if (targetId.getLogicalObject() == null)
                                        targetIds.add(targetId);
                                }
                            }

                            updatedArray._targetIds[j] = targetIds;
                        }

                        updatedArray._source = object;
                        updatedArray._array = (LyArrayFieldInfo)fieldInfo;
                        updated = updatedArray;
                    }
                }
            }

            if (updated == null)
                continue;

            if (_modifiedRelations == null)
                _modifiedRelations = new Vector<_LyUpdated>();

            _modifiedRelations.addElement(updated);
            pos = _modifiedRelations.size() - 1;

            key = object.getId() + "," + fieldInfo.getId();
            index = modifiedObjects.get(key);
            if (index != null)
                _modifiedRelations.removeElementAt(index.intValue());

            modifiedObjects.put(key, new Integer(pos));
        }
    } // end method addUpdatedRelations

    /**
     * Internal method to get the list of fieldInfos from a value set.
     *
     * @param   values  List of values to extract fieldInfos from.
     *
     * @return  The list of fieldInfos of the specified value set.
     */
    protected LyInfoList<LyFieldInfo> getFieldInfos(LyValueSet values)
    {
        if (values == null)
            return null;

        int                     i, n = values.getSize();
        LyFieldInfo             fieldInfo;
        LyValue                 value;
        LyInfoList<LyFieldInfo> fieldInfos = new LyInfoList<LyFieldInfo>();

        for (i = 0; i < n; i++)
        {
            value = values.getFieldValue(i);
            fieldInfo = (LyFieldInfo)value.getField();
            fieldInfos.addInfo(fieldInfo);
        }

        return fieldInfos;
    }

    /**
     * To mark operation as undo ignore. For internal use only.
     *
     * @param  ignore  true to ignore operations for undo, false otherwise.
     */
    public void setIgnoreUndo(boolean ignore)
    {
        _ignoreUndo = ignore;
    }

    /**
     * To force the context to keep (or not) the order of its sons operations.
     *
     * @param  keepOrder  true to keep the order of the sons operation INTERNAL USE ONLY
     */
    public void setKeepOrder(boolean keepOrder)
    {
        _keepOrder = keepOrder;
    }

    /**
     * Indicates if the operations of this context must be inserted at the beginning of the parent
     * operations or at the end.
     *
     * @param  position  the insertion position. INTERNAL USE ONLY
     */
    public void setInsertPosition(LyContext.Position position)
    {
        _insertionPosition = position;
    }

    /**
     * To mark operations as linked for undo/redo. For internal use only.
     *
     * @param  link  true to link operations, false otherwise.
     */
    public void setLinkUndo(boolean link)
    {
        _linkUndo = link;
        if ((isInternal()) && (_parent != null))
            _parent.setLinkUndo(link);
    }

    /**
     * To test if an undo operation can be done.
     *
     * @return  true if an undo is available, false otherwise.
     */
    public boolean isUndoAvailable()
    {
        if (_operationsToUndo == null)
            return false;

        int         i, n = _operationsToUndo.size();
        LyOperation operation;

        for (i = n - 1; i >= 0; i--)
        {
            operation = _operationsToUndo.get(i);
            if (!((operation.isUndoIgnored()) || operation.isLinked()))
                return true;
        }

        return false;
    }

    /**
     * To test if a redo operation can be done.
     *
     * @return  true if a redo is available, false otherwise.
     */
    public boolean isRedoAvailable()
    {
        if (_undoneOperations == null)
            return false;

        int         i, n = _undoneOperations.size();
        LyOperation operation;

        for (i = n - 1; i >= 0; i--)
        {
            operation = _undoneOperations.get(i);
            if (!((operation.isUndoIgnored()) || operation.isLinked()))
                return true;
        }

        return false;
    }

    /**
     * Undo the last operation of this context.
     */
    public void undo()
    {
        _undoRedoRunning = true;
        LyOperation operation;
        int         idx;

        while (!_operationsToUndo.isEmpty())
        {
            idx = _operationsToUndo.size() - 1;
            operation = _operationsToUndo.get(idx);
            _operationsToUndo.remove(idx);
            _undoneOperations.add(operation);
            if (operation.isUndoIgnored())
                continue;

            operation.rollback(false, _control);
            if (!operation.isLinked())
                break;
        }

        _undoRedoRunning = false;
    } // end method undo

    /**
     * Redo the last operation undone in this context.
     */
    public void redo()
    {
        _undoRedoRunning = true;
        LyOperation operation;
        int         idx;
        boolean     stop;
        LyResponse  response;

        _errorMessage = null;

        while (!_undoneOperations.isEmpty())
        {
            idx = _undoneOperations.size() - 1;
            operation = _undoneOperations.get(idx);
            _undoneOperations.remove(idx);
            _operationsToUndo.add(operation);
            if (operation.isUndoIgnored())
                continue;

            response = operation.send(true, false, true, _control);

            if ((response != null)
                    && (response.getStatus() == LyResponse.Status.ERROR))
            {
                if (_errorMessage == null)
                    _errorMessage = response.getError();
                else if (_errorMessage.indexOf(response.getError()) < 0)
                    _errorMessage += '\n' + response.getError();
            }

            stop = false;
            while ((!_undoneOperations.isEmpty()) && (!stop))
            {
                idx = _undoneOperations.size() - 1;
                operation = _undoneOperations.get(idx);
                if (operation.isUndoIgnored())
                {
                    _undoneOperations.remove(idx);
                    _operationsToUndo.add(operation);
                    continue;
                }

                if (operation.isLinked())
                    break;
                else
                    stop = true;
            }

            if (stop)
                break;
        }

        _undoRedoRunning = false;
    } // end method redo

    /**
     * Common interface for classes storing relations to update.
     */
    abstract class _LyUpdated
        implements Comparable<_LyUpdated>
    {
        /** Creation date of the object. */
        long _creationDate;

        /** The object to update. */
        LyObject _source;

        _LyUpdated()
        {
            _creationDate = LyContext._counter++;
        }

        public abstract LyObject update(LyObject newTarget);

        /**
         * Implementation of the Comparable interface. This method is used to order operations by
         * date.
         *
         * @param   updated  The update to compare
         *
         * @return  0 if the two objects are equals, 1 if this operation is greater, -1 otherwise.
         */
        public int compareTo(_LyUpdated updated)
        {
            long thisDate = this._creationDate;
            long opDate = updated._creationDate;

            if (thisDate == opDate)
                return 0;
            else if (thisDate > opDate)
                return 1;
            else
                return -1;
        }
    } // end class _LyUpdated

    /**
     * Internal class to store relations to update.
     */
    class _LyUpdatedSimpleRelation extends _LyUpdated
    {
        // The pointed object Id.
        LyObjectId _targetId;

        // The pointed object.
        LyObject _target;

        // The relation to update.
        LyRelationFieldInfo _relation;

        public String toString()
        {
            StringBuffer sb = new StringBuffer();

            sb.append("LyUpdateRelation\n");
            sb.append("\tSource = " + _source + "\n");
            sb.append("\tRelation = " + _relation + "\n");
            sb.append("\tTarget = " + _target + "\n");

            return sb.toString();
        }

        public LyObject update(LyObject newTarget)
        {
            if (_source.isDeleted())
                return null;

            if ((newTarget != null)
                    && (_target != null)
                    && (_target != newTarget))
                return null; // The relation doesn't refer the given object

            LyValue  value;
            LyObject target = null;

            if ((_target != null)
                    && _target.isDeleted())
            {
                target = _target.getObjectId().getLogicalObject();
                if ((target == null)
                        || target.isDeleted())
                    target = null;
            }
            else
                target = _target;

            // Now we add the id of objects that where not in cache when update
            if ((target == null)
                    && (_targetId != null))
            {
                target = _targetId.getLogicalObject();
                if ((target == null)
                        || target.isDeleted())
                    target = null;
                else if ((newTarget != null)
                        && (newTarget != target))
                    return null; // The relation doesn't refer the given object
            }

            if (!_source.hasNewValue(_relation)
                    && (target != null))
            {
                // Flandres Check if the target was REALLY created in the persistance layer,
                // otherwise this is a temporary id and we should not update the relation (yet)
                // Maybe there is a better way to check that the target was not created in peer ?
                if ((target.getPeer() == null) && !target.getClassInfo().isLocal())
                    return null; // SHOULD UPDATE LATER : target objects not created yet
            }

            value = _relation.buildFieldValue(target);

            // Keep old value
            Object oldValue = _source.getFromMemory(_relation);
            Object val = value.getValue();

            // Extract the right value from the given one
            val = _relation.processValue(val);

            if (_relation.equals(val, oldValue))
                return null;

            _source.setValue(null, _relation, value, LyObject.CONTROL_NONE, LyObject.OPER_NONE);

            if (_relation.hasMark(LyFieldInfo.MARK_ID))
            {
                LyObjectSource os = _source.getObjectSource();

                os.rehash(_source);
            }

            return _source;
        } // end method update
    } // end class _LyUpdatedSimpleRelation

    /**
     * Internal class to store relations to update.
     */
    class _LyUpdatedRelation extends _LyUpdated
    {
        /** The list of pointed objects Ids. */
        ArrayList<LyObjectId> _targetIds;

        /** The list of pointed objects. */
        LyObjectList _targets;

        /** The reltion to update. */
        LyRelationFieldInfo _relation;

        public String toString()
        {
            StringBuffer sb = new StringBuffer();

            sb.append("LyUpdateRelation\n");
            sb.append("\tSource = " + _source + "\n");
            sb.append("\tRelation = " + _relation + "\n");
            sb.append("\tTargets = " + _targets + "\n");

            return sb.toString();
        }

        public LyObject update(LyObject newTarget)
        {
            if (_source.isDeleted())
                return null;

            if ((newTarget != null)
                    && (_targets != null)
                    && !_targets.contains(newTarget))
                return null; // The relation doesn't refer the given object

            LyValue            value;
            LyObject           target;
            LyObjectId         targetId;
            LySimpleObjectList list = new LySimpleObjectList();
            int                i, n = ((_targets == null) ? 0 : _targets.getSize());

            list.setSorted(false);

            for (i = 0; i < n; i++)
            {
                target = _targets.getObject(i);
                if (target.isDeleted())
                {
                    target = target.getObjectId().getLogicalObject();
                    if ((target == null)
                            || target.isDeleted())
                        continue;
                }

                list.addObject(target);
            }
            // Now we add the id of objects that where not in cache when update
            if (_targetIds != null)
            {
                for (i = 0, n = _targetIds.size(); i < n; i++)
                {
                    targetId = _targetIds.get(i);
                    target = targetId.getLogicalObject();

                    if ((target == null) || (target.isDeleted()) || (list.contains(target)))
                        continue;

                    list.addObject(target);
                }

                if ((newTarget != null)
                        && !list.contains(newTarget))
                    return null; // The relation doesn't refer the given object
            }

            n = list.getSize();

            if (!_source.hasNewValue(_relation))
            {
                for (i = n - 1; i >= 0; i--)
                {
                    target = list.getObject(i);
                    // Flandres Check if the target was REALLY created in the persistance layer,
                    // otherwise this is a temporary id and we should not update the relation (yet)
                    // Maybee there is a better way to check that the target was not created in peer ?
                    if ((target.getPeer() == null) && !target.getClassInfo().isLocal())
                        return null; // SHOULD UPDATE LATER : target objects not created yet
                }
            }

            value = _relation.buildFieldValue(list);

            // Keep old value
            Object oldValue = _source.getFromMemory(_relation);
            Object val = value.getValue();

            // Extract the right value from the given one
            val = _relation.processValue(val);

            if (_relation.equals(val, oldValue))
                return null;

            _source.setValue(null, _relation, value, LyObject.CONTROL_NONE, LyObject.OPER_NONE);

            if (_relation.hasMark(LyFieldInfo.MARK_ID))
            {
                LyObjectSource os = _source.getObjectSource();

                os.rehash(_source);
            }

            return _source;
        } // end method update
    } // end class _LyUpdatedRelation

    /**
     * Internal class to store arrays of relations to update.
     */
    class _LyUpdatedArray extends _LyUpdated
    {

        /** The list of lists of objects pointed by the values of the array. */
        LyObjectList[] _targets;

        /** The list of lists of pointed objects Ids. */
        ArrayList<LyObjectId>[] _targetIds;

        /** The array to update. */
        LyArrayFieldInfo _array;

        public LyObject update(LyObject newTarget)
        {
            if (_source.isDeleted())
                return null;

            LyRelationFieldInfo relation;
            int                 i, n;
            Object[]            values;
            LyValue             value;
            LyObject            target;
            LyObjectId          targetId;
            LySimpleObjectList  list;
            int                 j, p;

            relation = (LyRelationFieldInfo)_array.getFieldInfo();
            n = _targets.length;
            values = new Object[n];
            for (i = 0; i < n; i++)
            {
                p = ((_targets[i] == null) ? 0 : _targets[i].getSize());
                list = new LySimpleObjectList();
                for (j = 0; j < p; j++)
                {
                    target = _targets[i].getObject(j);
                    if (target.isDeleted())
                        continue;

                    list.addObject(target);
                }
                // Now we add the id of objects that where not in cache when update
                p = 0;
                if ((_targetIds != null) && (_targetIds.length > i) && (_targetIds[i] != null))
                    p = _targetIds[i].size();

                for (j = 0; j < p; j++)
                {
                    targetId = _targetIds[i].get(j);
                    target = targetId.getLogicalObject();

                    if ((target == null) || (target.isDeleted()) || (list.contains(target)))
                        continue;

                    list.addObject(target);
                }

                value = relation.buildFieldValue(list);
                values[i] = value.getValue();
            }

            value = _array.buildFieldValue(values);
            _source.setValue(null, _array, value, LyObject.CONTROL_NONE, LyObject.OPER_NONE);
            return _source;
        } // end method update
    } // end class _LyUpdatedArray

    /**
     * Checks if this context has errors.
     *
     * @return  the maximum error level of this context or its sons.
     */
    public short getErrorLevel()
    {
        short errorLevel = LyObject.LEVEL_OK, level;

        if (_operations != null)
        {
            Iterator<LyOperation> it;
            LyOperation           operation;

            it = _operations.iterator();
            while (it.hasNext()
                    && (errorLevel != LyObject.LEVEL_ERROR))
            {
                operation = it.next();
                level = operation.getErrorLevel();
                if ((level != LyObject.LEVEL_OK)
                        && ((errorLevel == LyObject.LEVEL_OK)
                            || (level < errorLevel)))
                    errorLevel = level;
            }
        }

        // Check sons contexts
        if ((errorLevel != LyObject.LEVEL_ERROR)
                && (_sons != null))
        {
            Iterator<LyContext> it;
            LyContext           son;

            it = _sons.iterator();
            while (it.hasNext()
                    && (errorLevel != LyObject.LEVEL_ERROR))
            {
                son = it.next();
                level = son.getErrorLevel();
                if ((level != LyObject.LEVEL_OK)
                        && ((errorLevel == LyObject.LEVEL_OK)
                            || (level < errorLevel)))
                    errorLevel = level;
            }
        }

        return errorLevel;
    } // end method getErrorLevel

    /**
     * Gets the list of objects updated by this context and its sons with LEVEL_ERROR errors.
     *
     * @return  The list of objects updated by this context and its sons with LEVEL_ERROR errors or
     *          null if there is no objects with LEVEL_ERROR errors.
     */
    public Vector<LyObject> getObjectsInError()
    {
        Vector<LyObject> objects = null;
        LyObject         object;
        int              i, n;
        Vector<LyObject> errObjects;

        if (_operations != null)
        {
            Iterator<LyOperation> iterator;
            LyOperation           operation;

            iterator = _operations.iterator();
            while (iterator.hasNext())
            {
                operation = iterator.next();
                errObjects = operation.getObjectsInError();
                if (errObjects != null)
                {
                    n = errObjects.size();
                    for (i = 0; i < n; i++)
                    {
                        object = errObjects.elementAt(i);
                        if (objects == null)
                            objects = new Vector<LyObject>();

                        if (!objects.contains(object))
                            objects.add(object);
                    }
                }
            }
        }

        // Check sons contexts
        if (_sons != null)
        {
            Iterator<LyContext> iterator;
            LyContext           son;

            iterator = _sons.iterator();
            while (iterator.hasNext())
            {
                son = iterator.next();
                errObjects = son.getObjectsInError();
                if (errObjects != null)
                {
                    n = errObjects.size();
                    for (i = 0; i < n; i++)
                    {
                        object = errObjects.elementAt(i);
                        if (objects == null)
                            objects = new Vector<LyObject>();

                        if (!objects.contains(object))
                            objects.add(object);
                    }
                }
            }
        }

        return objects;
    } // end method getObjectsInError

    /**
     * Gets the list of errors on objects with the specified level.
     *
     * @param   level  Level of the errors
     * @return  The list of errors with the specified level for this context or null if there is
     *          no errors.
     */
    public Vector<LyError> getErrors(short level)
    {
        Vector<LyError>      errors = null;
        int                  i, n;
        LyError              error;
        
        if (_operations != null)
        {
            Iterator<LyOperation> iterator;
            LyOperation           operation;
            Vector<LyError>       operErrors;

            iterator = _operations.iterator();
            while (iterator.hasNext())
            {
                operation = iterator.next();
                operErrors = operation.getErrors(level);
                if (operErrors != null)
                {
                    n = operErrors.size();
                    for (i = 0; i < n; i++)
                    {
                        error = operErrors.elementAt(i);
                        if (errors == null)
                            errors = new Vector<LyError>();

                        if (!errors.contains(error))
                            errors.add(error);
                    }
                }
            }
        }

        // Check sons contexts
        if (_sons != null)
        {
            Iterator<LyContext> iterator;
            LyContext           son;
            Vector<LyError>     sonErrors;

            iterator = _sons.iterator();
            while (iterator.hasNext())
            {
                son = iterator.next();
                sonErrors = son.getErrors(level);
                if (sonErrors != null)
                {
                    n = sonErrors.size();
                    for (i = 0; i < n; i++)
                    {
                        error = sonErrors.elementAt(i);
                        if (errors == null)
                            errors = new Vector<LyError>();

                        if (!errors.contains(error))
                            errors.add(error);
                    }
                }
            }
        }

        return errors;
    } // end method getErrors

    /**
     * Internal method to get all updated (modified or created) objects by this context with
     * associated fields.
     *
     * @return  A map with association between object and fieldInfos updated by this context.
     */
    protected HashMap<LyObject, LyInfoList<LyFieldInfo>> getUpdatedObjects()
    {
        HashMap<LyObject, LyInfoList<LyFieldInfo>> result = new HashMap<LyObject, LyInfoList<LyFieldInfo>>();

        if (_operations == null)
            return result;

        Iterator<LyOperation> it = _operations.iterator();
        LyOperation           operation;

        while (it.hasNext())
        {
            operation = it.next();
            getUpdatedObjects(result, operation);
        }

        return result;
    }

    /**
     * Internal method to get all updated (modified or created) objects by the specified operation.
     *
     * @param  result     HashMap to put the results.
     * @param  operation  The operation.
     */
    protected void getUpdatedObjects(HashMap<LyObject, LyInfoList<LyFieldInfo>> result, LyOperation operation)
    {
        if (operation == null)
            return;

        LyOperation.Type        type = operation.getType();
        LyObject                object;
        LyInfoList<LyFieldInfo> fieldInfos, list;
        LyValueSet              oldValues;
        LyFieldInfo             fieldInfo;
        LyValue                 fieldValue;

        object = operation.getObject();
        switch (type)
        {

        case CREATE:
            result.put(object, null);
            break;

        case SET:
            oldValues = operation._oldValues;
            if (oldValues != null)
            {
                fieldInfos = new LyInfoList<LyFieldInfo>(false);
                for (int i = 0; i < oldValues.getSize(); i++)
                {
                    fieldValue = oldValues.getFieldValue(i);
                    fieldInfo = (LyFieldInfo)fieldValue.getField();
                    fieldInfos.addInfo(fieldInfo);
                }

                if (result.containsKey(object))
                {
                    list = result.get(object);
                    if (list != null)
                    {
                        list.addList(fieldInfos);
                        result.put(object, list);
                    }
                }
                else
                {
                    list = new LyInfoList<LyFieldInfo>(false);
                    list.addList(fieldInfos);
                    result.put(object, list);
                }
            }

            break;

        default:
            break;
        }

        ArrayList<LyOperation> operations = operation._operations;

        if (operations != null)
        {
            Iterator<LyOperation> it = operations.iterator();
            LyOperation           subOperation;

            while (it.hasNext())
            {
                subOperation = it.next();
                if (subOperation != operation)
                    getUpdatedObjects(result, subOperation);
            }
        }
    } // end method getUpdatedObjects

    /**
     * Gets this object as a string.
     *
     * @return  A string representation of this context.
     */
    public String toString()
    {
        StringBuffer buffer = new StringBuffer();
        LyOperation  operation;
        int          i, no = ((_operations == null) ? 0 : _operations.size());
        LyContext    context;
        int          nc = ((_sons == null) ? 0 : _sons.size());

        buffer.append("[CONTEXT ");
        buffer.append(hashCode());
        if (no > 0)
        {
            buffer.append(", ");
            buffer.append(no);
            buffer.append(" operations");
        }

        if (nc > 0)
        {
            buffer.append(", ");
            buffer.append(nc);
            buffer.append(" sub-contexts");
        }

        buffer.append('\n');

        for (i = 0; i < no; i++)
        {
            operation = _operations.get(i);
            buffer.append(operation.toString());
            buffer.append('\n');
        }

        for (i = 0; i < nc; i++)
        {
            context = _sons.get(i);
            buffer.append(context.toString());
            buffer.append('\n');
        }

        buffer.append("CONTEXT]");

        return buffer.toString();
    } // end method toString

    /**
     * Enable/Disable the given control on objects in this context.
     *
     * @param  control  Control that is enabled or disabled (see LyObject.CONTROL_XXX).
     * @param  enable   Boolean indicating if the control is enabled or disabled.
     */
    public void enableControl(int control, boolean enable)
    {
        if (enable)
            _control |= control;
        else
            _control &= ~control;

        if (_sons != null)
        {
            int       i, n = _sons.size();
            LyContext son;

            for (i = 0; i < n; i++)
            {
                son = _sons.get(i);
                son.enableControl(control, enable);
            }
        }
    }

    /**
     * Check if the specified object has already been created in this context or in another comitted
     * context.
     *
     * @param   object  The object ot test.
     *
     * @return  true if the specified object has already been created, false otherwise.
     */
    public boolean isAlreadyCreated(LyObject object)
    {
        LyOperation foundOperation = null;

        if (object.checkObject()
                || ((_createdObjects != null) && _createdObjects.contains(object)))
        {
            if (object.getPeer() != null)
                return true;

            if ((_createdObjects != null) && (_createdObjects.contains(object)))
            {
                // Search for a create operation on this object.
                LyOperation op;
                int         i, n = ((_operations == null) ? 0 : _operations.size());

                for (i = 0; i < n; i++)
                {
                    op = _operations.get(i);
                    if ((op.getType() == LyOperation.Type.CREATE)
                            && (op.getObject() == object))
                    {
                        if (foundOperation == null)
                            foundOperation = op;
                        else
                            return true;
                    }
                }
            }
        }

        return false;
    } // end method isAlreadyCreated

    /**
     * To disable the relations update.
     */
    public void disableUpdateRelations()
    {
        if (_disableUpdateRelations)
            return;

        _disableUpdateRelations = true;

        if (_sons != null)
        {
            int       i, n = _sons.size();
            LyContext son;

            for (i = 0; i < n; i++)
            {
                son = _sons.get(i);
                son.disableUpdateRelations();
            }
        }
    }

    /**
     * To disable the events sending.
     */
    public final void disableEvents()
    {
        enableEvents(false);
    }

    /**
     * To enable/disable the events sending.
     *
     * @param  enable  Boolean that indicates if events are enabled or disabled.
     */
    public void enableEvents(boolean enable)
    {
        if (_disableEvents == !enable)
            return;

        _disableEvents = !enable;

        if (_sons != null)
        {
            int       i, n = _sons.size();
            LyContext son;

            for (i = 0; i < n; i++)
            {
                son = _sons.get(i);
                son.enableEvents(enable);
            }
        }
    }

    /**
     * To rehash a colection of LyObjects.
     *
     * @param  objects  The collection of LyObject to rehash.
     */
    public void rehash(Collection<LyObject> objects)
    {
        if (objects == null)
            return;

        Iterator<LyObject> it = objects.iterator();
        LyObject           object;

        while (it.hasNext())
        {
            object = it.next();
            rehash(object);
        }
    }

    /**
     * To rehash a LyObject.
     *
     * @param  object  The object to rehash.
     */
    public void rehash(LyObject object)
    {
        if (object == null)
            return;

        LyObjectSource source;

        source = object.getObjectSource();
        source.rehash(object, !_disableEvents);
    }

    /**
     * To force or not all operations in this context to be only local.
     *
     * @param  onlyLocal  true to force operations to be local, false otherwise.
     */
    public void setOnlyLocal(boolean onlyLocal)
    {
        _onlyLocal = onlyLocal;
    }

    /**
     * Getter of the _onlyLocal attribute.
     *
     * @return  true if all operations of this context must be local.
     */
    public boolean isOnlyLocal()
    {
        return _onlyLocal;
    }

    /**
     * Proceted accessor to undone operations.
     *
     * @return  the list of undone operations (internal use only)
     */
    protected List<LyOperation> getUndoneOperations()
    {
        return _undoneOperations;
    }

    /**
     * Checks if this context is empty or not (i.e. has operations or not).
     *
     * @return  True if this context has no operation, false otherwise.
     */
    public boolean isEmpty()
    {
        return ((_operations == null) || (_operations.size() == 0));
    }

    public void addUserOperation(LyUserOperationInterface callback, LyObject object, Object userData)
    {
        LyOperation operation;

        operation = new LyOperation(this, object, callback, userData);
        addOperation(operation);
        setStatus(Status.RUNNING);
    }
} // end class LyContext
