/*
 * LyContextManager.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;

import leon.app.*;
import leon.info.*;
import leon.misc.*;


/**
 * This class manages the instances of LyContext. There is only one instance of this class for a
 * Leonardi application (i.e. for a primary module).
 *
 * @see  LyContext
 * @see  leon.app.LyApplication
 */
public class LyContextManager
    implements LyDataSourceListener
{
    /** File version. */
    public static final String __VERSION = "$Id: LyContextManager.java 34926 2012-03-22 10:10:46Z jlsaouli $";

    /** Reference to the session instance. */
    protected LyDataSession _session;

    /** List of contexts managed by the context manager. */
    protected Collection<LyContext> _contexts;

    /** List of contexts to undo. */
    protected ArrayList<_LyUndoRedoOperation> _contextsToUndo;

    /** List of undone contexts. */
    protected ArrayList<_LyUndoRedoOperation> _undoneContexts;

    /** List of objects used in undo/redo contexts. */

    protected LySimpleObjectList _objects;

    /** Number of managed undo/redo operations. */
    protected int _nbUndoRedo;

    /**
     * Constructs a context manager. Initialization is done with the init method.
     *
     * @param  session  Current session instance
     */
    public LyContextManager(LyDataSession session)
    {
        init();

        _session = session;

        LyEnvironment environment = session.getEnvironment();
        String		  undoRedoStr = environment.getResource("LY_NB_UNDO_REDO");
        
        if ((undoRedoStr != null)
        	&& (undoRedoStr.length() > 0))
        {
	        _nbUndoRedo = environment.getIntResource("LY_NB_UNDO_REDO");
	
	        if (_nbUndoRedo != 0)
	        {
	            LyApplication  application = session.getApplication();
	            LyEventManager em = application.getEventManager();
	
	            em.addListener(null, this, LyDataSourceEvent.TYPE_SET);
	            em.addListener(null, this, LyDataSourceEvent.TYPE_DELETE);
	        }
        }
    }

    /**
     * Initialisation of the data of the instance.
     */
    protected void init()
    {
        _session = null;
        _contexts = Collections.synchronizedCollection(new ArrayList<LyContext>());
        _objects = new LySimpleObjectList(false);
    }

    /**
     * Free this context manager.
     */
    public void free()
    {
        if (_nbUndoRedo != 0)
        {
            LyApplication  application = _session.getApplication();
            LyEventManager em = application.getEventManager();

            em.removeListener(null, this);
        }

        _session = null;
        _contexts = null;
        _objects = null;
        _contextsToUndo = null;
        _undoneContexts = null;
        _objects = null;
    }

    /**
     * To begin a new context. The new context is associated with the current user session.
     *
     * @param   session  the associated user session.
     * @return  the new created context.
     */
    public final LyContext newContext(LyDataSession session)
    {
        return newContext(session, LyContext.Type.AUTO);
    }
    
    /**
     * To begin a new context. The new context is associated with the current user session.
     *
     * @param   session  the associated user session.
     * @param   type     the type of the new context.
     *
     * @return  the new created context.
     */
    public LyContext newContext(LyDataSession session, LyContext.Type type)
    {
        LyContext context;

        context = new LyContext(session, type);
        addContext(context);
        return context;
    }

    /**
     * Accessor to the contexts list.
     *
     * @return  Collection of current contexts managed by the context manager.
     */
    public Collection<LyContext> getContexts()
    {
        return _contexts;
    }

    /**
     * Gets number of current contexts.
     *
     * @return  Number of contexts registered on this context managaer.
     */
    public int getContextCount()
    {
        return ((_contexts == null) ? 0 : _contexts.size());
    }

    /**
     * Gets context at the given index in contexts list.
     *
     * @param   index  Index of the requested context.
     *
     * @return  Context at the specified poistion if it exists, false otherwise.
     */
    public LyContext getContext(int index)
    {
        return ((_contexts == null) ? null : (LyContext)(_contexts.toArray()[index]));
    }

    /**
     * Adds a context to the list of contexts managed by this context manager.
     *
     * @param  context  the context to add.
     */
    protected void addContext(LyContext context)
    {
        if (_contexts == null)
            _contexts = Collections.synchronizedCollection(new ArrayList<LyContext>());

        _contexts.add(context);
    }

    /**
     * Removes the specified context from the list of contexts managed by this context manager.
     *
     * @param  context  the context to remove.
     */
    protected void removeContext(LyContext context)
    {
        _contexts.remove(context);
        if ((_nbUndoRedo != 0)
                && (context.getStatus() == LyContext.Status.COMPLETED)
                && (!context.isUndoDisable()))
            // The context was committed : we keep it in the
            // list of context to be undone.
            addContextToUndoList(context);
        else
            // We can free the context.
            context.free();
    }

    /**
     * Cancel all contexts. This method only unlocks the objects locked by all contexts.
     */
    public void cancelAll()
    {
        Iterator<LyContext> it;
        LyContext           context = null;

        // Unlock all objects locked by all contexts.
        synchronized (_contexts)
        {
            it = _contexts.iterator();
            context = null;
            while (it.hasNext())
            {
                context = (LyContext)it.next();
                context.free();
            }
        }

        _contexts = Collections.synchronizedCollection(new ArrayList<LyContext>());
    }

    /**
     * To add a context to the undo list.
     *
     * @param  context  The context to add.
     */
    protected void addContextToUndoList(LyContext context)
    {
        if (_nbUndoRedo == 0)
            return;

        ArrayList<LyOperation> operations = context.getAllOperations();

        if (operations.size() != 0)
        {
            _LyUndoRedoOperation undoRedo = new _LyUndoRedoOperation();

            undoRedo._context = context;
            undoRedo._operations = operations;

            if (_contextsToUndo == null)
                _contextsToUndo = new ArrayList<_LyUndoRedoOperation>();

            _contextsToUndo.add(undoRedo);
            resetUndoRedo(true);
        }
    }

    /**
     * Internal method to reset the lists of contexts for undo/redo operations.
     *
     * @param  freeUndone  To free the undone context list.
     */
    void resetUndoRedo(boolean freeUndone)
    {
        if (_nbUndoRedo == 0)
            return;

        int                  i, n, nbToUndo, nbUndone;
        _LyUndoRedoOperation undoRedo;

        if (freeUndone && (_undoneContexts != null))
        {
            n = _undoneContexts.size();
            for (i = 0; i < n; i++)
            {
                undoRedo = _undoneContexts.get(i);
                undoRedo._context.free();
            }

            _undoneContexts = null;
        }

        // Rebuild workSpace with objects of undo/redo contexts.
        _objects = new LySimpleObjectList(false);
        _objects.setSorted(false);

        n = ((_undoneContexts == null) ? 0 : _undoneContexts.size());
        for (i = 0; i < n; i++)
        {
            undoRedo = _undoneContexts.get(i);
            _objects.addList(undoRedo.getObjects());
        }

        nbUndone = n;

        n = ((_contextsToUndo == null) ? 0 : _contextsToUndo.size());
        for (i = 0; i < n; i++)
        {
            undoRedo = _contextsToUndo.get(i);
            _objects.addList(undoRedo.getObjects());
        }

        nbToUndo = n;

        n = nbToUndo + nbUndone;
        if ((_nbUndoRedo > 0) && (n > _nbUndoRedo))
            _contextsToUndo.remove(0);

        _objects.setSorted(true);

        _session.resetUndoRedo();
    } // end method resetUndoRedo

    /**
     * Method for clearing the lists of contexts for undo/redo operations.
     */
    public void clearUndoRedo()
    {
        if (_nbUndoRedo == 0)
            return;

        int                  i, n;
        _LyUndoRedoOperation undoRedo;

        if (_undoneContexts != null)
        {
            for (i = 0, n = _undoneContexts.size(); i < n; i++)
            {
                undoRedo = _undoneContexts.get(i);
                undoRedo._context.free();
            }

            _undoneContexts = null;
        }

        if (_contextsToUndo != null)
        {
            for (i = 0, n = _contextsToUndo.size(); i < n; i++)
            {
                undoRedo = _contextsToUndo.get(i);
                undoRedo._context.free();
            }

            _contextsToUndo = null;
        }

        // Rebuild objects list
        _objects = new LySimpleObjectList(false);

        _session.resetUndoRedo();
    } // end method clearUndoRedo

    /**
     * To test if an undo operation can be done.
     *
     * @return  true if an undo is available, false otherwise.
     */
    public boolean isUndoAvailable()
    {
        return ((_contextsToUndo != null) && (!_contextsToUndo.isEmpty()));
    }

    /**
     * To get the name of the next available undo operation.
     *
     * @return  The anme of the next available undo operation or null if no undo operation is
     *          available.
     */
    public String getUndoName()
    {
        if ((_contextsToUndo != null)
                && (!_contextsToUndo.isEmpty()))
        {
            int       idx = _contextsToUndo.size() - 1;
            LyContext context = _contextsToUndo.get(idx)._context;

            return context.getName();
        }

        return null;
    }

    /**
     * To test if a redo operation can be done.
     *
     * @return  true if a redo is available, false otherwise.
     */
    public boolean isRedoAvailable()
    {
        return ((_undoneContexts != null) && (!_undoneContexts.isEmpty()));
    }

    /**
     * To get the name of the next available redo operation.
     *
     * @return  The anme of the next available redo operation or null if no redo operation is
     *          available.
     */
    public String getRedoName()
    {
        if ((_undoneContexts != null)
                && (!_undoneContexts.isEmpty()))
        {
            int       idx = _undoneContexts.size() - 1;
            LyContext context = _undoneContexts.get(idx)._context;

            return context.getName();
        }

        return null;
    }

    /**
     * Undo all operations done by the last comitted context.
     *
     * @param   session  the current session
     *
     * @return  true if successfull
     */
    public boolean undo(LyDataSession session)
    {
        if (!isUndoAvailable())
            return false;

        _LyUndoRedoOperation undoRedo;
        int                  idx;

        LyApplication application = session.getApplication();
        LyContext     oldContext = application.getCurrentContext();
        // clear the current context
        application.setCurrentContext(null);

        idx = _contextsToUndo.size() - 1;
        undoRedo = _contextsToUndo.get(idx);

        // undo must be done in the root session.
        session = getRootSession(session);

        // Undo all operations in an automatic context.
        int                    i, n;
        LyOperation            operation;
        ArrayList<LyOperation> operations = undoRedo._operations;
        LyContext              context = newContext(session);
        boolean                result;

        context.setInternal();

        try
        {
            n = operations.size();
            for (i = n - 1; i >= 0; i--)
            {
                operation = (LyOperation)operations.get(i);
                undoOperation(context, operation);
            }

            context.commit();
        }
        catch (Exception e)
        {
            // Unexpected exception
            _session.getEnvironment().writeException(e);
            context.rollback();
        }

        if (_undoneContexts == null)
            _undoneContexts = new ArrayList<_LyUndoRedoOperation>();

        _contextsToUndo.remove(undoRedo);

        if (context.getStatus() == LyContext.Status.COMPLETED)
        {
            _undoneContexts.add(undoRedo);
            result = true;
        }
        else
        {
            // An error occured : we rollback the context.
            LyEnvironment environment = _session.getEnvironment();

            environment.logError("An error occured while undoing context");

            Vector<LyError> v = context.getErrors(LyObject.LEVEL_ERROR);
            LyError         error;
            String          message;

            n = ((v == null) ? 0 : v.size());
            for (i = 0; i < n; i++)
            {
                error = v.elementAt(i);
                message = error.getMessage(environment);
                environment.logError(message);
            }

            context.rollback();
            result = false;
        }

        resetUndoRedo(false);

        // Restore old context...
        application.setCurrentContext(oldContext);
        return result;
    } // end method undo

    /**
     * Undo the specified operation in the specified context.
     *
     * @param  context    the context in which the operation must be undone.
     * @param  operation  The operation to undo.
     */
    protected void undoOperation(LyContext context, LyOperation operation)
    {
        int                    i, n;
        LyOperation            subOperation;
        LyOperation.Type       operType;
        LyObject               object;
        ArrayList<LyOperation> operations;
        LyValueSet             oldValues;

        operType = operation.getType();
        object = operation.getObject();
        operations = operation._operations;
        oldValues = operation._oldValues;

        // Get all sub operations in reverse order and rollback them.
        if ((operations != null)
                && (operType != LyOperation.Type.DELETE))
        {
            n = operations.size();
            for (i = n - 1; i >= 0; i--)
            {
                subOperation = (LyOperation)operations.get(i);
                if (subOperation != operation)
                    undoOperation(context, subOperation);
            }
        }

        switch (operType)
        {

        case CREATE:
            if (object.isDeleted())
                break;

            context.deleteObject(object);
            break;

        case DELETE:
            n = operations.size();
            for (i = n - 1; i >= 0; i--)
            {
                subOperation = (LyOperation)operations.get(i);
                if (subOperation == operation)
                {
                    LyClassInfo objectClassInfo = object.getClassInfo();
                    LyInfo      parent = objectClassInfo.getParent();
                    // SIR/FFT02288
                    if ((!(parent instanceof LyArrayFieldInfo)
                                && !(parent instanceof LyRelationFieldInfo))
                            || (parent instanceof LyStructFieldInfo))
                    {
                        object.reset(context, true);
                        context.createObject(object);
                    }
                }
                else
                    undoOperation(context, subOperation);
            }

            break;

        case SET:
            if (object.isDeleted())
                break;

            if ((oldValues != null) && (oldValues.getSize() > 0))
            {
                // Set back the old values to the object for local values
                // Remove id values : this values may have been set only in case of creation
                // So an undo operation may delete the object and removing values
                // is not important.
                LyValueSet  valuesToSet = new LyValueSet();
                LyValue     value;
                LyFieldInfo fieldInfo;

                n = oldValues.getSize();
                for (i = 0; i < n; i++)
                {
                    value = oldValues.getFieldValue(i);
                    fieldInfo = (LyFieldInfo)value.getField();
                    if (!fieldInfo.hasMark(LyFieldInfo.MARK_ID))
                        valuesToSet.addValue(value);
                }

                if (valuesToSet.getSize() > 0)
                {
                    object.skipEncoding(true);
                    context.setValues(object, valuesToSet);
                    object.skipEncoding(false);
                }
            }

            break;

        case DELETES:
            break;

        default:
            break;
        }
    } // end method undoOperation

    /**
     * Redo all operations undo by the last undo call.
     *
     * @param   session  the current session
     *
     * @return  True if last operation was successfully redone.
     */
    public boolean redo(LyDataSession session)
    {
        if (!isRedoAvailable())
            return false;

        _LyUndoRedoOperation undoRedo;
        int                  idx;

        idx = _undoneContexts.size() - 1;
        undoRedo = _undoneContexts.get(idx);

        // redo must be done in the root session.
        session = getRootSession(session);

        // Redo all operations in an automatic context.
        int         i, n;
        LyOperation operation;
        ArrayList<LyOperation>   operations = undoRedo._operations;
        LyContext   context = newContext(session);
        boolean     result;

        context.setInternal();

        try
        {
            n = operations.size();
            for (i = 0; i < n; i++)
            {
                operation = (LyOperation)operations.get(i);
                redoOperation(context, operation);
            }

            context.commit();
        }
        catch (Exception e)
        {
            // Unexpected exception
            _session.getEnvironment().writeException(e);
            context.rollback();
        }

        if (_contextsToUndo == null)
            _contextsToUndo = new ArrayList<_LyUndoRedoOperation>();

        _undoneContexts.remove(undoRedo);

        if (context.getStatus() == LyContext.Status.COMPLETED)
        {
            _contextsToUndo.add(undoRedo);
            result = true;
        }
        else
        {
            // An error occured : we rollback the context.
            LyEnvironment environment = _session.getEnvironment();

            environment.logError("An error occured while redoing context");

            Vector<LyError> v = context.getErrors(LyObject.LEVEL_ERROR);
            LyError         error;
            String          message;

            n = ((v == null) ? 0 : v.size());
            for (i = 0; i < n; i++)
            {
                error = v.elementAt(i);
                message = error.getMessage(environment);
                environment.logError(message);
            }

            context.rollback();
            result = false;
        }

        resetUndoRedo(false);
        return result;
    } // end method redo

    /**
     * Redo the specified operation in the specified context.
     *
     * @param  context    the context in which the operation must be redone.
     * @param  operation  The operation to redo.
     */
    protected void redoOperation(LyContext context, LyOperation operation)
    {
        LyOperation            subOperation;
        LyOperation.Type       operType;
        LyObject               object;
        Iterator<LyOperation>  it;
        ArrayList<LyOperation> operations;
        LyValueSet             values;

        operType = operation.getType();
        object = operation.getObject();
        operations = operation._operations;
        values = operation._values;

        switch (operType)
        {

        case CREATE:
            LyClassInfo objectClassInfo = object.getClassInfo();
            LyInfo      parent = objectClassInfo.getParent();
// SIR/FFT02288

            if ((!(parent instanceof LyArrayFieldInfo)
                        && !(parent instanceof LyRelationFieldInfo))
                    || (parent instanceof LyStructFieldInfo))
            {
                object.reset(context, !operation._onlyNotLocal);
                context.createObject(object);
            }

            break;

        case DELETE:
            context.deleteObject(object);
            return;

        case SET:
            if (!object.isDeleted())
            {
                if (values == null)
                    context.setObject(object);
                else
                {
                    object.skipEncoding(true);
                    context.setValues(object, values);
                    object.skipEncoding(false);
                }
            }

            break;

        case DELETES:
            break;

        default:
            break;
        }

        if ((operations != null)
                && (operType != LyOperation.Type.DELETE))
        {
            // Get all sub operations and send them.
            it = operations.iterator();
            while (it.hasNext())
            {
                subOperation = (LyOperation)it.next();
                if (subOperation != operation)
                    redoOperation(context, subOperation);
            }
        }
    } // end method redoOperation

    /**
     * Internal method to get the root session of a session.
     *
     * @param   session  The session
     *
     * @return  The root session of the specified session.
     */
    protected LyDataSession getRootSession(LyDataSession session)
    {
        LyDataSession parent = session;

        while (parent != null)
        {
            session = parent;
            parent = session.getParent();
        }

        return session;
    }

    /* Implementation of LyDatasourceListener */

    /**
     * Method for receiving an object creation.
     *
     * @param   event  Event notifying the creation of an object.
     *
     * @return  True indicating if the event was used, false otherwise.
     */
    public boolean objectCreated(LyDataSourceEvent event)
    {
        // Nothing to do.
        return false;
    }

    /**
     * Method for receiving an object deletion.
     *
     * @param   event  Event notifying the deletion of an object.
     *
     * @return  True indicating if the event was used, false otherwise.
     */
    public boolean objectDeleted(LyDataSourceEvent event)
    {
        return receiveEvent(event);
    }

    /**
     * Method for receiving an object modification.
     *
     * @param   event  Event notifying the modification of an object.
     *
     * @return  True indicating if the event was used, false otherwise.
     */
    public boolean objectModified(LyDataSourceEvent event)
    {
        return receiveEvent(event);
    }

    /**
     * Internal method used to treat events.
     *
     * @param   event  The event to treat.
     *
     * @return  true if the event was used, false otherwise.
     */
    protected boolean receiveEvent(LyDataSourceEvent event)
    {
        if (_nbUndoRedo == 0)
            return false;

        if (!event.isNotified())
            return false;

        LyObject object = (LyObject)event.getObject();

        if (!_objects.contains(object))
            return false;

        // Remove all undo/redo context using this object.
        int                  i, n;
        _LyUndoRedoOperation undoRedo;
        LyObjectList         objects;

        n = ((_contextsToUndo == null) ? 0 : _contextsToUndo.size());
        for (i = 0; i < n; i++)
        {
            undoRedo = _contextsToUndo.get(i);
            objects = undoRedo.getObjects();
            if (objects.contains(object))
                break;
        }

        if (i != n)
        {
            // Remove all remaining context
            for (int j = n - 1; j >= i; j--)
            {
                undoRedo = _contextsToUndo.remove(j);
                undoRedo._context.free();
            }

            resetUndoRedo(true);
            return true;
        }

        n = ((_undoneContexts == null) ? 0 : _undoneContexts.size());
        for (i = 0; i < n; i++)
        {
            undoRedo = _undoneContexts.get(i);
            objects = undoRedo.getObjects();
            if (objects.contains(object))
                break;
        }

        if (i != n)
        {
            // Remove all remaining context
            for (int j = n - 1; j >= i; j--)
            {
                undoRedo = _undoneContexts.remove(j);
                undoRedo._context.free();
            }

            resetUndoRedo(false);
            return true;
        }

        return false;
    } // end method receiveEvent

    /**
     * Internal class to store informations about an undo/redo operation.
     */
    class _LyUndoRedoOperation
    {
        LyContext              _context;
        ArrayList<LyOperation> _operations;

        LySimpleObjectList getObjects()
        {
            LySimpleObjectList objects = new LySimpleObjectList(false);
            int                i, n = _operations.size();
            LyOperation        operation;
            LyObjectList       opObjects;

            objects.setSorted(false);
            for (i = 0; i < n; i++)
            {
                operation = _operations.get(i);
                opObjects = getObjects(operation);
                objects.addList(opObjects);
            }

            return objects;
        }

        /**
         * Internal method to get the list of objects impacted by an operation.
         *
         * @param   operation  The operation.
         *
         * @return  The list of objects impacted by the specified operation (may be empty).
         */
        LyObjectList getObjects(LyOperation operation)
        {
            LySimpleObjectList objects = new LySimpleObjectList(false);
            LyObject           object = operation.getObject();

            objects.setSorted(false);
            if ((object != null) && (!object.isDeleted()))
                objects.add(object);

            LyObjectList operObjects = operation.getObjects();
            int          i, n = ((operObjects == null) ? 0 : operObjects.getSize());

            for (i = 0; i < n; i++)
            {
                object = operObjects.getObject(i);
                if (!object.isDeleted())
                    objects.add(object);
            }

            Iterator<LyOperation> it = operation.iterator();
            LyOperation           subOperation;
            LyObjectList          subObjects;

            if (it != null)
            {
                while (it.hasNext())
                {
                    subOperation = (LyOperation)it.next();
                    if (subOperation != operation)
                    {
                        subObjects = getObjects(subOperation);
                        objects.addList(subObjects);
                    }
                }
            }

            return objects;
        } // end method getObjects
    } // end class _LyUndoRedoOperation
} // end class LyContextManager
