/*
 * LyDataSource.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.lang.reflect.*;
import java.util.*;

import leon.app.*;
import leon.data.cache.*;
import leon.info.*;
import leon.info.binding.*;
import leon.misc.*;
import leon.notifier.*;
import leon.peer.*;
import leon.peer.util.*;


/**
 * This class implements a collection of data of the same class. It provides methods for accessing,
 * creating, deleting, modifying and counting datas.
 *
 * <p>Both physical class and logical class are managed by data sources. The only difference is that
 * physical classes use direct instances of this class, and logical classes are handled by instances
 * of an inherited class, LyObjectSource.</p>
 *
 * <p>Each data source is connected to a data provider which is responsible to manage real objects :
 * encoding requests and decoding results.</p>
 *
 * <p>Data sources are valid within an application.</p>
 *
 * <p>Objects returned via data sources are kept in caches.</p>
 *
 * @see LyClass
 * @see LyObjectSource
 * @see LyCache
 */
public abstract class LyDataSource
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyDataSource.java 41712 2012-12-05 08:51:26Z bangiole $";

	/** Application owner of this data source. */
	protected LyApplication    _application;

	/** Provider of the datas managed by the data source. */
	protected LyDataProvider   _provider;

	/** Class of data handled by this data source. */
	protected LyClass		   _class;

	/**
	 * Cache associated to this data source that holds instances of the class managed by this data
	 * source.
	 */
	protected LyCache		   _cache;

	/**
	 * Last modification date. (Any change on objects contained within this data source shall update
	 * this date).
	 */
	protected long			   _lastModified;

	/**
	 * Indicates if this data source send requests synchronously or not. If true, when sent,
	 * requests are blocked until results are obtained.
	 */
	protected boolean		   _synchronous;

	/**
	 * Indicates if this data source is read only or writable. If true, only consultation requests
	 * are accepted, others are rejected.
	 */
	protected boolean		   _readOnly;

	/**
	 * Indicates that we are performing a rollBack on this datasource (see LyJoinRelationDaemon).
	 */
	protected boolean		   _rollBack;

	/**
	 * If an error occured during the load of an object by its id, the object
	 * is not registered in the unknown object list if this parameter is set to true.
	 */
	protected boolean		   _noUnknownIds;

	/**
	 * Default constructor of a data source. Only used by inherited classes.
	 */
	protected LyDataSource()
	{
		// sets the last modification date to current date
		_lastModified = (new Date()).getTime();

		// read/write and asynchrounous by default
		_readOnly = false;
		_synchronous = false;
		_noUnknownIds = false;

		// No rollback now
		_rollBack = false;
	}

	/**
	 * Constructs a data source within the given application that manages the specified class.
	 *
	 * @param provider  Provider on which this data source is connected.
	 * @param dataClass Class of data handled by this new data source.
	 */
	public LyDataSource(LyDataProvider provider, LyClass dataClass)
	{
		this();

		// Initialize class attributes.
		_application = provider.getApplication();
		_provider = provider;
		_class = dataClass;

		// The cache is specified by requestCachePolicy.
		_cache = null;

		_application.putDataSource(this);
	}

	/**
	 * Internal method to set cache policy. Default cache policy is Full cache and it can be
	 * downgraded to Simple cache if a request is issued for Simple and No cache.
	 *
	 * @param     cachePolicy The cache policy to set.
	 * @exception IllegalStateException If the data source is already associated to a cache with a
	 *                                  different policy. If we can't build the specified cache.
	 */
    protected void requestCachePolicy(LyInfoCache cachePolicy)
	{
		if ((_cache != null) && (!_cache.getPolicy().equals(cachePolicy)))
		{
			_application.getEnvironment().logWarning("BF AppEngine : data source for class "
				+ _class.getName()
				+ " is already associated to a cache with a different policy (current policy = "
				+ _cache.getPolicy().getCachePolicyName()
				+ ", requested = " + cachePolicy.getCachePolicyName() + "), keeping current one.");
			return; // Keep existing policy
		}

		if ((_cache != null) && (_cache.getPolicy().equals(cachePolicy)))
			return; // Nothing to do

		Class<?>[]  types;
		Object[] parameters;

		// Build cache
		types = new Class[2];
		parameters = new Object[2];

		types[0] = LyClass.class;
		parameters[0] = _class;
		types[1] = LyInfoCache.class;
		parameters[1] = cachePolicy;

		@SuppressWarnings("unchecked")
		Class<LyCache> javaClass = (Class<LyCache>)cachePolicy.getCachePolicy();

		try
		{
			Constructor<LyCache> constructor = javaClass.getConstructor(types);

			_cache = constructor.newInstance(parameters);
		}
		catch (InvocationTargetException e)
		{
			_application.getEnvironment().writeException(e.getTargetException());
			_cache = null;
		}
		catch (Exception e)
		{
			_application.getEnvironment().writeException(e);
			_cache = null;
		}

		if (_cache == null)
			throw new IllegalStateException("BF AppEngine : can't create cache " + javaClass);

		if (_provider.canBeFullyLoaded(_class))
			_cache.allowLoaded(true);
		else
			_cache.allowLoaded(false);
	} // end method requestCachePolicy

	/**
	 * Sets the cache for this data source.
	 *
	 * @param     cachePolicy Informations for the cache to associate to this object source.
	 * @exception IllegalArgumentException If the given cache policy is null.
	 * @exception IllegalStateException    If the data source is already associated to a cache that
	 *                                     contains objects.
	 */
	public abstract void setCache(LyInfoCache cachePolicy);

	/**
	 * Gets the provider of the data in this data source.
	 *
	 * @return Provider of the data managed in this data source.
	 */
	public final LyDataProvider getDataProvider()
	{
		return _provider;
	}

	/**
	 * Retrieves the class of data handled by this data source.
	 *
	 * @return Class of the data managed in this data source.
	 */
	public final LyClass getDataClass()
	{
		return _class;
	}

	/**
	 * Gets the cache associated to this data source. Depends on the cache policy chosen for the
	 * data class.
	 *
	 * @return Cache instance associated to this data source.
	 */
	public final LyCache getCache()
	{
		return _cache;
	}

	/**
	 * Gets the application of this data source.
	 *
	 * @return Application of this data source.
	 */
	public final LyApplication getApplication()
	{
		return _application;
	}

	/**
	 * Returns the last time a modification occured within this data source.
	 *
	 * @return Last modification date of this data source.
	 */
	public long getLastModified()
	{
		return _lastModified;
	}

	/**
	 * Checks if the data source is read only or not.
	 *
	 * @return True if the data source is read only, false if it is writable.
	 */
	public final boolean isReadOnly()
	{
		return _readOnly;
	}

	/**
	 * Checks if the data source is synchronous or not.
	 *
	 * @return True if the data source sends request in synchronous mode, false if they are sent
	 *         asynchronously.
	 */
	public final boolean isSynchronous()
	{
		return _synchronous;
	}

	/**
	 * Sets the last time a modification occured within this dataSource.
	 *
	 * @param time Last modification date.
	 */
	protected void setLastModified(long time)
	{
		_lastModified = time;
	}

	/**
	 * Sets the write mode of this data source. When turned to read only, modification requests are
	 * not accepted anymore.
	 *
	 * @param readOnly True if read only, false if writable.
	 */
	public void setReadOnly(boolean readOnly)
	{
		_readOnly = readOnly;
	}

	/**
	 * Sets the ignore error on load for this datasource.
	 * If an error occured during the load of an object by its id, the object
	 * is not registered in the unknown object list if this parameter is set to true.
	 * - INTERNAL USE ONLY -
	 * @param ignoreErrorOnLoad True to enable the mode, false if writable.
	 */
	public void setIgnoreErrorOnLoad(boolean ignoreErrorOnLoad)
	{
		_noUnknownIds = ignoreErrorOnLoad;
	}

	/**
	 * Sets the synchronous mode of this data source.
	 *
	 * @param synchronous True if the data source sends request in synchronous mode, false if they
	 *                    are sent asynchronously.
	 */
	public void setSynchronous(boolean synchronous)
	{
		_synchronous = synchronous;
	}

	/**
	 * Registers a listener to receive all events on objects contained within this data source.
	 *
	 * @param listener New listener. To receive these events, the listener have to implement the
	 *                 LyDataSourceListener interface.
	 */
	public void addListener(LyDataSourceListener listener)
	{
		addListener(listener, LyDataSourceEvent.TYPE_ALL);
	}

	/**
	 * Registers a listener to receive events of the given type on objects contained within this
	 * data source.
	 *
	 * @param     listener New listener. To receive these events, the listener have to implement the
	 *                     LyDataSourceListener interface.
	 * @param     type     Indicates which type of event the listener wants to receive. Types are
	 *                     defined as constant in LyDataSourceEvent class : TYPE_CREATE, TYPE_SET,
	 *                     TYPE_DELETE
	 * @exception IllegalArgumentException If the listener is null or if the given type is unknown.
	 */
	public abstract void addListener(LyDataSourceListener listener, short type);

	/**
	 * Registers a listener to receive modification events concerning given fields on objects
	 * contained within this data source. Warning ! This method only registers the listener for
	 * modification events.
	 *
	 * @param     listener New listener. To receive these events, the listener have to implement the
	 *                     LyDataSourceListener interface.
	 * @param     fields   List of fields on which the listener wants to be notified when they are
	 *                     modified.
	 * @exception IllegalArgumentException If the listener is null.
	 */
	public abstract void addListener(LyDataSourceListener listener, Vector<LyField> fields);

	/**
	 * Unregisters a listener for notifications from this data source. The given listener is
	 * unregistered for all types of events he has been registered for.
	 *
	 * @param     listener The listener which shall not receive further events.
	 * @exception IllegalArgumentException If the listener is null or if it is not known as listener
	 *                                     on this data source.
	 */
	public abstract void removeListener(LyDataSourceListener listener);

	/**
	 * Unloads completely the data of this datasource. Empties associated cache. This methods
	 * notifies virews for objects deletions.
	 */
	public void unload()
	{
		unload(true);
	}

	/**
	 * Unloads completely the data of this datasource. Empties associated cache.
	 *
	 * @param notify indicates wether the unload action should notify objects deletion. Notifying
	 *               deletions makes sense when views are opened and when application is not
	 *               exiting.
	 */
	public void unload(boolean notify)
	{
		if (_cache == null)
			return;
		// Find objects of the cache
		Vector<Object> objects = _cache.getObjects();

		// Do as if objects were deleted
		if (notify
				&& (objects != null)
				&& (objects.size() > 0))
		{
			// Find objects of the cache
			LyResponse response = new LyResponse(LyResponse.Operation.DELETE, _class);
			int		   i, n = ((objects == null) ? 0 : objects.size());
			Object     object;

			for (i = 0; i < n; i++)
			{
				object = objects.elementAt(i);

				// Set the status to the response.
				if (i == (n - 1))
					response.setStatus(LyResponse.Status.COMPLETED);
				else
					response.setStatus(LyResponse.Status.RUNNING);

				// Add the object in the response.
				response.addResponse(object, null);

				LyDataSourceEvent event;

				event = new LyDataSourceEvent(this, LyDataSourceEvent.TYPE_DELETE, object, null);
				event.setLoading(true);
				event.setLocal(response.isLocal());
				postEvent(event, null);
			}
		}

		// Unload cache
		_cache.unload();
	} // end method unload

	/**
	 * Internal method to load an object from the associated data provider.
	 *
	 * @param  id           Object identifier of the searched object.
     * @param  fields       Fields to load.
	 * @param  classBinding Binding between logical class and physical class. This binding may
	 *                      contain extra information required for encoding the request.
	 * @return The requested object if it was found, or null if it could not be found.
	 */
	protected Object loadObjectFromProvider(LyObjectId id, Vector<LyField> fields, LyClassBinding classBinding)
	{
		// Try to get the object from data provider.
		boolean		      synchronous;
		LyResponse	      response;
        LyResponse.Status status;
		Object		      object = null;

		// The request must be synchronous.
		synchronous = isSynchronous();
		setSynchronous(true);

		if (fields == null)
        {
            if  (_class instanceof LyClassInfo)
    		{
    			LyInfoList<LyFieldInfo> list = ((LyClassInfo)_class).getFieldInfos(LyFieldInfo.MARK_TABLE);

    			if ((list != null) && (list.getSize() > 0))
    				fields = new Vector<LyField>(list.getInfos());
                else
                    fields = new Vector<LyField>();
    		}
        }

		// Load the object.
		response = loadObject(id, fields, null, classBinding);

		// Restore the previous synchronous mode.
		setSynchronous(synchronous);

		// Check the status of the response.
		status = response.getStatus();

		if ((status == LyResponse.Status.COMPLETED)
				&& (response.getSize() == 1))
		{
			object = response.getResponse(0);

			// Don't put the object in the cache because the answer of the loading
			// request has already done the job !
		}
		else
		{
			// Register this id as unknown on the cache to avoid future errors.
			if (!_noUnknownIds && ((classBinding == null) || (classBinding.getFilter() == null)))
				_cache.registerUnknownId(id);
			object = null;
		}

		return object;
	} // end method loadObjectFromProvider

	/**
	 * Gets an object from this data source by the specified identifier.
	 *
	 * @param  id           Object identifier of the searched object.
	 * @param fields        A list of fields that need to be fetched. Null for all.
	 * @param  classBinding Binding between logical class and physical class. This binding may
	 *                      contain extra information required for encoding the request.
	 * @return The requested object if it was found, or null if it could not be found.
	 */
	public Object getObject(LyObjectId id, Vector<LyField> fields, LyClassBinding classBinding)
	{
		if (id == null)
			throw new IllegalArgumentException("BF AppEngine : id parameter can't be null");

		// Check if first
		if (!id.check(_application))
			return null;

		Object object = null;

		try
		{
			// Try first to get the object from the cache.
			if (_cache.checkObject(id))
				object = _cache.getObject(id);
			else if ((_cache.isLoaded()) || (_cache.isUnknown(id)) || (_cache.isRemoved(id)))
				object = null;
			else
				object = loadObjectFromProvider(id, fields, classBinding);
		}
		catch (LyNoSuchObjectException e)
		{
			if (_cache.isLoaded())
				return null;

			object = loadObjectFromProvider(id, fields, classBinding);
		}

		return object;
	} // end method getObject

	/**
	 * Gets an object from this data source by the specified identifier.
	 *
	 * @param  id Object identifier of the searched object.
	 * @return The requested object if it was found, or null if it could not be found.
	 */
	public abstract Object getObject(LyObjectId id);

	/**
	 * Return the number of objects within this data source.
	 *
	 * @return Number of objects found.
	 */
	public int getSize()
	{
		return getSize(null, null);
	}

	/**
	 * Return the number of objects within this data source matching the given filter.
	 *
	 * @param  filter Filter applied on objects for counting.
	 * @return Number of objects matching the given filter in this data source.
	 */
	public int getSize(LyFilter filter)
	{
		return getSize(filter, null);
	}

	/**
	 * Return the number of objects within this data source matching the given filter.
	 *
	 * @param  filter       Filter applied on objects for counting.
	 * @param  classBinding Binding between logical class and physical class. This binding may
	 *                      contain extra information required for encoding the request (such as
	 *                      extra filter).
	 * @return Number of objects matching the given filter in this data source.
	 */
	public int getSize(LyFilter filter, LyClassBinding classBinding)
	{
		if (_cache != null)
		{
			int n = _cache.getCount(filter);

			if (n >= 0)
				return n;
		}

		// The cache can't say how many objects are in the data source.
		// We must send a request to the data provider to know.
		// The request must be synchronous.
		boolean			  synchronous;
		LyResponse		  response;
        LyResponse.Status status;
		LyRequestEncoder  encoder;

		synchronous = isSynchronous();
		setSynchronous(true);

		// Send the request.
		encoder = _provider.getRequestEncoder(this);
		response = new LyResponse(LyResponse.Operation.COUNT, _class);
		encoder.setResponse(response);
		encoder.setClassBinding(classBinding);
		encoder.getSize(filter);

		// Set the older synchronous mode.
		setSynchronous(synchronous);

		// Check the status of the response.
		status = response.getStatus();

		if ((status == LyResponse.Status.COMPLETED)
				&& (response.getSize() == 1))
		{
			Integer size = (Integer)response.getResponse(0);

			return size.intValue();
		}
		else
			return -1;
	} // end method getSize

	/**
	 * Internal method to post an event to listeners.
	 *
	 * @param  event Posted event.
	 * @return True if the vent was sent and used by at least one listener, false otherwise.
	 */
	protected final boolean postEvent(LyDataSourceEvent event)
	{
		return postEvent(event, null);
	}

	/**
	 * Internal method to post an event to listeners.
	 *
	 * @param     event         Posted event.
	 * @param     skipListeners Listeners that were already notified and that must be skipped by
	 *                          this method to avoid a double notification of the same event.
	 * @return    True if the vent was sent and used by at least one listener, false otherwise.
	 * @exception IllegalArgumentException If the given event is null.
	 */
	protected abstract boolean postEvent(LyDataSourceEvent event,
		Vector<LyDataSourceListener> skipListeners);

	/**
	 * Tests if an object corresponding to a given object id is found in the cache of this data
	 * source.
	 *
	 * @param  objectId Object id for which an object is searched.
	 * @return True if an object with the given object id is found in the cache, else false.
	 */
	public boolean checkObject(LyObjectId objectId)
	{
		try
		{
			return (_cache.getObject(objectId) != null);
		}
		catch (LyNoSuchObjectException e)
		{
			return false;
		}
	}

	/**
	 * Internal method to check validy of an object given as parameter.
	 *
	 * @param  object Checked object.
	 * @return True if the given object is found in the cache, else false.
	 */
	public boolean checkObject(Object object)
	{
		try
		{
			LyObjectId objectId = _class.getObjectId(object);

			return (_cache.getObject(objectId) != null);
		}
		catch (LyNoSuchObjectException e)
		{
			return false;
		}
	}

	/**
	 * Internal method to check validy of an object given as parameter.
	 *
	 * @param     object  Checked object.
	 * @param     inCache Flag indicated if the result must be or not in cache.
	 * @exception IllegalArgumentException If one of the above checks fails.
	 */

	protected void checkObject(Object object, boolean inCache)
	{
		// Check the object.
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		// Check if the object has a good class.
		if (!_class.getObjectClass().isAssignableFrom(object.getClass()))
			throw new IllegalArgumentException("BF AppEngine : object has bad classInfo");

		// Check if the object is in the cache
		boolean resultCache = checkObject(object);

		if (inCache && !resultCache)
			throw new IllegalArgumentException("BF AppEngine : object "
				+ _class.getObjectId(object) + " is unknown " + object);
		else if (!inCache && resultCache)
			throw new IllegalArgumentException("BF AppEngine : object "
				+ _class.getObjectId(object) + " is already existing");
	}

	/**
	 * Put the specified object into the cache.
	 *
	 * @param object The object.
	 */
	protected void putObject(Object object)
	{
		_cache.putObject(object);
	}

	/**
	 * Start removing the specified object from the cache.
	 *
	 * @param object The object.
	 */
	protected void startRemoveObject(Object object)
	{
		_cache.startRemoveObject(object);
	}

	/**
	 * Remove the specified object from the cache.
	 *
	 * @param object The object.
	 */
	protected void removeObject(Object object)
	{
		_cache.removeObject(object);
	}

	/* Methods for creating an object */

	/**
	 * Create a new object on the associated data provider. This method commits local creation by
	 * creating the object on the physical layer.
	 *
	 * @param     object       Object to create.
	 * @param     listener     Listener of the response.
	 * @param     classBinding Binding between logical class and physical class. This binding may
	 *                         contain extra information required for encoding the request.
	 * @return    Response to the creation.
	 * @exception IllegalStateException If the data source is in read only mode.
	 */
	public LyResponse createObject(Object object, LyResponseListener listener,
		LyClassBinding classBinding)
	{
		return createObject(object, listener, classBinding, true);
	}

	/**
	 * Create a new object on the associated data provider. This method commits local creation by
	 * creating the object on the physical layer.
	 *
	 * @param     object       Object to create.
	 * @param     listener     Listener of the response.
	 * @param     classBinding Binding between logical class and physical class.
	 * @param     refresh      Indicates if events for this operation may bre sent or not. This
	 *                         binding may contain extra information required for encoding the
	 *                         request.
	 * @return    Response to the creation.
	 * @exception IllegalStateException If the data source is in read only mode.
	 */
	protected LyResponse createObject(Object object, LyResponseListener listener,
		LyClassBinding classBinding, boolean refresh)
	{
		// If the data source is read only we can't create an object.
		if (isReadOnly())
			throw new IllegalStateException(
				"BF AppEngine : Attempt to create an object on a read only data source");

		// Check the object.
		// checkObject(object, false);

		// Now we can send the request to the data provider.
		LyResponse		 response;
		LyRequestEncoder encoder;

		encoder = _provider.getRequestEncoder(this);
		response = new LyResponse(LyResponse.Operation.CREATE, _class);
		if (listener != null)
			response.setListener(listener);
		response.setRefresh(refresh);

		encoder.setResponse(response);
		encoder.setClassBinding(classBinding);
		if (encoder.createObject(object))
            _provider.setModified(true, _class, object);

		return response;
	}

	/* Methods for modifying objects */

	/**
	 * Set an object on the associated data provider. This method commits local modification by
	 * setting the object on the physical layer.
	 *
	 * @param     object       Object to modify.
	 * @param     listener     Listener of the response.
	 * @param     classBinding Binding between logical class and physical class This binding may
	 *                         contain extra information required for encoding the request.
	 * @return    Response to the modification.
	 * @exception IllegalStateException    If the data source is in read only mode.
	 * @exception IllegalArgumentException If the given object is invalid.
	 */
	public LyResponse setObject(Object object, LyResponseListener listener,
		LyClassBinding classBinding)
	{
		return setObject(object, listener, classBinding, true);
	}

	/**
	 * Set an object on the associated data provider. This method commits local modification by
	 * setting the object on the physical layer.
	 *
	 * @param     object       Object to modify.
	 * @param     listener     Listener of the response.
	 * @param     classBinding Binding between logical class and physical class
	 * @param     refresh      Indicates if events for this operation may bre sent or not. This
	 *                         binding may contain extra information required for encoding the
	 *                         request.
	 * @return    Response to the modification.
	 * @exception IllegalStateException    If the data source is in read only mode.
	 * @exception IllegalArgumentException If the given object is invalid.
	 */
	protected LyResponse setObject(Object object, LyResponseListener listener,
		LyClassBinding classBinding, boolean refresh)
	{
		return setObject(object, listener, classBinding, refresh, true);
	}

	/**
	 * Propaget change on the given object
	 * @param object the object to set
	 * @param listener a response listener
	 * @param classBinding physical binding of the object
	 * @param refresh true to refresh the object
	 * @param needCheck to check
	 * @return a response
	 */
	protected LyResponse setObject(Object object, LyResponseListener listener,
		LyClassBinding classBinding, boolean refresh, boolean needCheck)
	{
		// If the data source is read only we can't set an object.
		if (isReadOnly())
			throw new IllegalStateException(
				"BF AppEngine : Attempt to set an object on a read only data source");

		// Check the object.
		if (needCheck)
			checkObject(object, true);

		// Now we can send the request to the data provider.
		LyResponse		 response;
		LyRequestEncoder encoder;

		encoder = _provider.getRequestEncoder(this);
		response = new LyResponse(LyResponse.Operation.SET, _class);
		if (listener != null)
			response.setListener(listener);
		response.setRefresh(refresh);

		encoder.setResponse(response);
		encoder.setClassBinding(classBinding);
        if (encoder.setObject(object))
            _provider.setModified(true, _class, object);

		return response;
	}

	/**
	 * Set a list of objects on the associated data provider. This method commits local
	 * modifications by setting the objects on the physical layer.
	 *
	 * @param     objects      List of objects to modify.
	 * @param     listener     Listener of the response.
	 * @param     classBinding Binding between logical class and physical class This binding may
	 *                         contain extra information required for encoding the request.
	 * @return    Response to the modification.
	 * @exception IllegalStateException    If the data source is in read only mode.
	 * @exception IllegalArgumentException If given objects list is null or if one of these objects
	 *                                     is invalid.
	 */
	public LyResponse setObjects(Vector<? extends Object> objects, LyResponseListener listener,
		LyClassBinding classBinding)
	{
		// If the data source is read only we can't set objects.
		if (isReadOnly())
			throw new IllegalStateException(
				"BF AppEngine : Attempt to set objects on a read only data source");

		if (objects == null)
			throw new IllegalArgumentException("BF AppEngine : objects can't be null");

		// Check all objects
		int    i, n = objects.size();
		Object object;

		for (i = 0; i < n; i++)
		{
			object = objects.elementAt(i);
			checkObject(object, true);
		}

		// Now we can send the request to the data provider.
		LyResponse		 response;
		LyRequestEncoder encoder;

		encoder = _provider.getRequestEncoder(this);
		response = new LyResponse(LyResponse.Operation.SET, _class);
		if (listener != null)
			response.setListener(listener);

		encoder.setResponse(response);
		encoder.setClassBinding(classBinding);
        if (encoder.setObjects(new Vector<Object>(objects)))
            _provider.setModified(true, _class, objects);

		return response;
	} // end method setObjects

	/* Methods for deleting objects */

	/**
	 * Delete an object on the associated data provider. This method commits local deletion by
	 * deleting the object on the physical layer.
	 *
	 * @param     object       Object to delete.
	 * @param     listener     Listener of the response.
	 * @param     classBinding Binding between logical class and physical class This binding may
	 *                         contain extra information required for encoding the request.
	 * @return    Response to the deletion.
	 * @exception IllegalStateException    If the data source is in read only mode.
	 * @exception IllegalArgumentException If the given object is invalid.
	 */
	public LyResponse deleteObject(Object object, LyResponseListener listener,
		LyClassBinding classBinding)
	{
		return deleteObject(object, listener, classBinding, true, true);
	}

	/**
	 * Delete an object on the associated data provider. This method commits local deletion by
	 * deleting the object on the physical layer.
	 *
	 * @param     object       Object to delete.
	 * @param     listener     Listener of the response.
	 * @param     classBinding Binding between logical class and physical class
	 * @param     refresh      Indicates if events for this operation may be sent or not. This
	 *                         binding may contain extra information required for encoding the
	 *                         request.
	 * @param     check        Indicates if the checking in the cache may be done or not.
	 * @return    Response to the deletion.
	 * @exception IllegalStateException    If the data source is in read only mode.
	 * @exception IllegalArgumentException If the given object is invalid.
	 */
	protected LyResponse deleteObject(Object object, LyResponseListener listener,
		LyClassBinding classBinding, boolean check, boolean refresh)
	{
		// If the data source is read only we can't delete an object.
		if (isReadOnly())
			throw new IllegalStateException(
				"BF AppEngine : Attempt to delete an object on a read only data source");

		// Check the object.
		if (check)
			checkObject(object, true);

		// Now we can send the request to the data provider.
		LyResponse		 response;
		LyRequestEncoder encoder;

		encoder = _provider.getRequestEncoder(this);
		response = new LyResponse(LyResponse.Operation.DELETE, _class);
		if (listener != null)
			response.setListener(listener);
		response.setRefresh(refresh);

		encoder.setResponse(response);
		encoder.setClassBinding(classBinding);
		if (encoder.deleteObject(object))
            _provider.setModified(true, _class, object);

		return response;
	}

	/**
	 * Delete a list of objects on the associated data provider. This method commits local deletions
	 * by deleting the objects on the physical layer.
	 *
	 * @param     objects      List of objects to delete.
	 * @param     listener     Listener of the response.
	 * @param     classBinding Binding between logical class and physical class This binding may
	 *                         contain extra information required for encoding the request.
	 * @return    Response to the deletion.
	 * @exception IllegalStateException    If the data source is in read only mode.
	 * @exception IllegalArgumentException If given objects list is null or if one of these objects
	 *                                     is invalid.
	 */
	public LyResponse deleteObjects(Vector<? extends Object> objects, LyResponseListener listener,
		LyClassBinding classBinding)
	{
		return deleteObjects(objects, listener, classBinding, true, true);
	}

	/**
	 * Delete a list of objects on the associated data provider. This method commits local deletions
	 * by deleting the objects on the physical layer.
	 *
	 * @param     objects      List of objects to delete.
	 * @param     listener     Listener of the response.
	 * @param     classBinding Binding between logical class and physical class This binding may
	 *                         contain extra information required for encoding the request.
	 * @param     check        Indicates if the checking in the cache may be done or not.
	 * @param     refresh      Indicates if events for this operation may bre sent or not.
	 * @return    Response to the deletion.
	 * @exception IllegalStateException    If the data source is in read only mode.
	 * @exception IllegalArgumentException If given objects list is null or if one of these objects
	 *                                     is invalid.
	 */
	protected LyResponse deleteObjects(Vector<? extends Object> objects,
		LyResponseListener listener,
		LyClassBinding classBinding, boolean check, boolean refresh)
	{
		// If the data source is read only we can't delete objects.
		if (isReadOnly())
			throw new IllegalStateException(
				"BF AppEngine : Attempt to delete objects on a read only data source");

		// Check all objects
		if (check)
		{
			int    i, n;
			Object object;

			n = objects.size();
			for (i = 0; i < n; i++)
			{
				object = objects.elementAt(i);
				checkObject(object, true);
			}
		}

		// Now we can send the request to the data provider.
		LyResponse		 response;
		LyRequestEncoder encoder;

		encoder = _provider.getRequestEncoder(this);
		response = new LyResponse(LyResponse.Operation.DELETE, _class);
		if (listener != null)
			response.setListener(listener);
		response.setRefresh(refresh);

		encoder.setResponse(response);
		encoder.setClassBinding(classBinding);
		if (encoder.deleteObjects(new Vector<Object>(objects)))
		    _provider.setModified(true, _class, objects);

		return response;
	} // end method deleteObjects

	/* Methods for loading objects */

	/**
	 * Method to reload an object from its id on the associated data provider. Load fields of an
	 * object from a given object id.
	 *
	 * @param  id           Object identifier corresponding to the object to reload.
	 * @param  fields       Fields to load (null means all fields).
	 * @param  listener     Listener of the response.
	 * @param  classBinding binding between logical class and physical class This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the loading.
	 */
	public LyResponse reloadObject(LyObjectId id, Vector<? extends LyField> fields,
		LyResponseListener listener, LyClassBinding classBinding)
	{
		LyResponse		 response;
		LyRequestEncoder encoder;

		encoder = _provider.getRequestEncoder(this);
		response = new LyResponse(LyResponse.Operation.LOAD, _class);
		if (listener != null)
			response.setListener(listener);

		encoder.setResponse(response);
		encoder.setClassBinding(classBinding);

        if (fields == null)
            encoder.loadObject(id, null);
        else
            encoder.loadObject(id, new Vector<LyField>(fields));

		// not found in data provider : may be still in cache ?
		// FFT Flandres 357, if not found in response, simulate a DELETE notification
		// mark the object as deleted and remove from cache.
		if ((response.getStatus() == LyResponse.Status.COMPLETED)
				&& ((response.getObjects() == null)
					|| (response.getObjects().getSize() == 0)))
		{
			try
			{
				Object object;

				if (_cache != null)
				{
					object = _cache.getObject(id);

					if (object instanceof LyObject)
					{
						((LyObject)object).setDeleted(true); // no longer found
						// Remove the object from the cache
						_cache.startRemoveObject(object);

						try
						{
							// Notify listeners
							LyDataSourceEvent event = new LyDataSourceEvent(this,
									LyDataSourceEvent.TYPE_DELETE, object, null);

							// This event is the result of a reload.
							// Deletion must remain local.
							event.setLocal(true);

							postEvent(event);

							// Notif may add the object back in the cache
							if (_cache.contains(object))
								_cache.startRemoveObject(object);
						}
						finally
						{
							_cache.removeObject(object);
						}
					}
				}
			}
			catch (LyNoSuchObjectException e)
			{
			}
		}

		return response;
	} // end method reloadObject

	/**
	 * Method to load an object from its id on the associated data provider. Load fields of an
	 * object from a given object id.
	 *
	 * @param  id           Object identifier corresponding to the object to load.
	 * @param  fields       Fields to load (null means all fields).
	 * @param  listener     Listener of the response.
	 * @param  classBinding binding between logical class and physical class This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the loading.
	 */
	public LyResponse loadObject(LyObjectId id, Vector<? extends LyField> fields,
		LyResponseListener listener, LyClassBinding classBinding)
	{
		LyResponse		 response;
		LyRequestEncoder encoder;

		encoder = _provider.getRequestEncoder(this);
		response = new LyResponse(LyResponse.Operation.LOAD, _class);
		if (listener != null)
			response.setListener(listener);

		encoder.setResponse(response);
		encoder.setClassBinding(classBinding);

		Vector<LyField> additionnalFields = null;
		if ( (_cache instanceof LyNoCache)
				&& (fields != null))
		{
			Object object;
			try
			{
				object = _cache.getObject(id);
			}
			catch (LyNoSuchObjectException e)
			{
				object = null;
			}
			if ((object != null) && (object instanceof LyRowObject))
			{
				LyRowObject rowObject = (LyRowObject)object;
				LyValueSet values = rowObject.getValues();
				int i, nbValues = (values == null ? 0 : values.getSize());
				LyField field;

				for (i = 0; i < nbValues; i++)
				{
					field = values.getFieldValue(i).getField();
					if (!fields.contains(field))
					{
						if (additionnalFields == null)
							additionnalFields = new Vector<LyField>();
						additionnalFields.add(field);
					}
				}
			}
		}
		if (fields == null)
			encoder.getObject(id, null);
		else
		{
			Vector<LyField> finalFields = new Vector<LyField>(fields);
			if (additionnalFields != null)
				finalFields.addAll(additionnalFields);
			encoder.getObject(id, finalFields);
		}

		return response;
	}

	/**
	 * Method for loading objects on the associated data provider. Load fields of a specified number
	 * of objects sorted by a given sort and that match a given filter.
	 *
	 * @param  fields       Fields to load (null means all fields).
	 * @param  sort         Gives sort criteria, or null if there is no sort (the order is the one
	 *                      of the physical layer).
	 * @param  filter       Gives filter criteria, or null if there is no filter (all objects).
	 * @param  n            Indicates the number of objects to load (-1 means no limit).
	 * @param  listener     Listener of the response.
	 * @param  classBinding binding between logical class and physical class This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the loading.
	 */
	public LyResponse loadObjects(Vector<? extends LyField> fields, LySort sort, LyFilter filter,
		int n, LyResponseListener listener, LyClassBinding classBinding)
	{
		LyResponse		 response;
		LyRequestEncoder encoder;

		encoder = _provider.getRequestEncoder(this);
		response = new LyResponse(LyResponse.Operation.LOAD, _class);
		if (listener != null)
			response.setListener(listener);

		encoder.setResponse(response);
		encoder.setClassBinding(classBinding);
        if (fields == null)
            encoder.getObjects(null, sort, filter, n);
        else
            encoder.getObjects(new Vector<LyField>(fields), sort, filter, n);

		return response;
	}

	/* Interface between LyDataSource and LyRequestEncoder */

	/**
	 * Method used to receive a new response from data provider to a request sent by this data
	 * source. This method may be called several times if several objects are concerned.
	 *
	 * @param     response  Response where there is a new object concerned.
	 * @param     status    Status of the new response. Valid status are defined in LyResponse.
	 * @param     object    Object concerned by this response.
	 * @param     oldValues Values that have changed for the object with this response (null when
	 *                      the operation is not a load or a modifcation).
	 * @exception IllegalArgumentException If response parameter is null.
	 * @see       LyResponse
	 * @see       #createObject
	 * @see       #setObject
	 * @see       #setObjects
	 * @see       #deleteObject
	 * @see       #deleteObjects
	 * @see       #loadObject
	 * @see       #loadObjects
	 */
	public void receiveResponse(LyResponse response, LyResponse.Status status, Object object,
		LyValueSet oldValues)
	{
		if (response == null)
			throw new IllegalArgumentException("BF AppEngine : response can't be null");

		LyResponse.Operation operation = response.getOperation();

		if ((status == LyResponse.Status.COMPLETED)
				&& (operation == LyResponse.Operation.LOAD)
				&& response.isFullLoad())
			setLoaded(true);

		if ((status == LyResponse.Status.ERROR)
            || (object == null))
		{
			response.setStatus(status);
			response.addResponse(null, null);
			return;
		}

		switch (operation)
		{
		case CREATE:
			receiveCreate(response, status, object);
			break;

		case SET:
			receiveSet(response, status, object, oldValues);
			break;

		case DELETE:
			receiveDelete(response, status, object);
			break;

		case LOAD:
			receiveLoad(response, status, object, oldValues);
			break;

		case COUNT:
			receiveCount(response, status, object);
			break;
		default:
			break;
		}
	} // end method receiveResponse

	/**
	 * Internal method used to receive a new response from data provider to a creation request sent
	 * on createObject. Object is added in the cache of this data source and a creation event is
	 * sent to listeners. If the data provider has a notifier, the creation is notified.
	 *
	 * @param response Response to the create operation.
	 * @param status   Status of the new response. Valid status are defined in LyResponse.
	 * @param object   Newly created object.
	 * @see   #createObject
	 */
	protected abstract void receiveCreate(LyResponse response, LyResponse.Status status, Object object);

	/**
	 * Internal method used to receive a new response from data provider to a set request sent on
	 * setObject or setObjects. Object is modified and a set event is sent to listeners. If the data
	 * provider has a notifier, the set is notified.
	 *
	 * @param response  Response to the set operation.
	 * @param status    Status of the new response. Valid status are defined in LyResponse.
	 * @param object    Modified object.
	 * @param oldValues Old values of the object before the set operation.
	 * @see   #setObject
	 * @see   #setObjects
	 */
	protected abstract void receiveSet(LyResponse response, LyResponse.Status status, Object object,
		LyValueSet oldValues);

	/**
	 * Internal method used to receive a new response from data provider to a deletion request sent
	 * on deleteObject or deleteObjects. A deletion event is sent to listeners and the object is
	 * deleted from the cache If the data provider has a notifier, the deletion is notified.
	 *
	 * @param response Response to the deletion operation.
	 * @param status   Status of the new response. Valid status are defined in LyResponse.
	 * @param object   Deleted object.
	 * @see   #deleteObject
	 * @see   #deleteObjects
	 */
	protected abstract void receiveDelete(LyResponse response, LyResponse.Status status, Object object);

	/**
	 * Internal method used to receive a new response from data provider to a load request sent on
	 * loadObject or loadObjects. If needed, the object is added in the cache or it is updated, and
	 * then, a load event is sent to listeners.
	 *
	 * @param response  Response to the load operation.
	 * @param status    Status of the new response. Valid status are defined in LyResponse.
	 * @param object    Loaded object.
	 * @param oldValues Old values of the object before the load operation.
	 * @see   #loadObject
	 * @see   #loadObjects
	 */
	protected abstract void receiveLoad(LyResponse response, LyResponse.Status status, Object object,
		LyValueSet oldValues);

	/**
	 * Internal method used to receive a new response from data provider to a count request sent on
	 * one of the getSize methods.
	 *
	 * @param response Response to the load operation.
	 * @param status   Status of the new response. Valid status are defined in LyResponse.
	 * @param object   Loaded object.
	 * @see   #loadObject
	 * @see   #loadObjects
	 */
	protected void receiveCount(LyResponse response, LyResponse.Status status, Object object)
	{
		// Set the status to the response.
		response.setStatus(status);

		// Add the object in the response.
		response.addResponse(object, null);
	}

	/* Interface between LyDataSource and LyEventEncoder */

	/**
	 * This method is used by the event notifier to notify the data source for data changes made by
	 * other applications using the same datas.
	 *
	 * @param  groupId Id of the group of events. May be null for non grouped events.
	 * @param  type    Type of the event. Valid types are defined in LyEventNotifier.
	 * @param  object  Object concerned by the event
	 * @param  values  New values.
	 * @return True if the event was treated with success, false otherwise.
	 */
	public boolean receiveEvent(String groupId, LyEventNotifier.EventType type,
        Object object, LyValueSet values)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		LyEnvironment environment = _application.getEnvironment();

		if (environment.isDebugLevel(LyEnvironment.DEBUG_BUS))
		{
			environment.logTrace("LyDataSource : receiveEvent");
			environment.logTrace("\tgroup = " + groupId);
			environment.logTrace("\tclass = " + _class.getId());
			environment.logTrace("\ttype = " + type);
			environment.logTrace("\tobject = " + _class.getObjectId(object));
		}

		LyDataSourceEvent event;
		boolean			  result = false;

		switch (type)
		{
		case CREATE:
			// Check if object is already in cache (may occur when using paging)
			LyObjectId objectId = _class.getObjectId(object);
			Object     cachedObject;

			try
			{
				cachedObject = _cache.getObject(objectId);
			}
			catch (LyNoSuchObjectException e)
			{
				cachedObject = null;
			}

			if (cachedObject == null)
				// Put object in cache
				_cache.putObject(object);
			else
			{
				// Copy values from object to cachedObject.
				_class.copy(cachedObject, object);
				object = cachedObject;

			}

			// XXX Could we send event even the object already exists ?

			// Notify listeners
			event = new LyDataSourceEvent(this, LyDataSourceEvent.TYPE_CREATE, object, null);
			event.setNotified(true);
			event.setLoading(true); // FFT00171
			event.setGroupId(groupId);
			result = postEvent(event);
			break;

		case UPDATE:
			// Notify listeners
			event = new LyDataSourceEvent(this, LyDataSourceEvent.TYPE_SET, object, values);
			event.setNotified(true);
			event.setGroupId(groupId);
			result = postEvent(event);
			break;

		case DELETE:
			// Remove the object from the cache
			_cache.startRemoveObject(object);

			try
			{
				// Notify listeners
				event = new LyDataSourceEvent(this, LyDataSourceEvent.TYPE_DELETE, object, null);
				event.setNotified(true);
				event.setGroupId(groupId);
				result = postEvent(event);

				// Notif may add the object back in the cache
				if (_cache.contains(object))
					_cache.startRemoveObject(object);
			}
			finally
			{
				_cache.removeObject(object);
			}
			break;
		default:
			break;
		}

		return result;
	} // end method receiveEvent

	/**
	 * Set the rollBack status of this datasource.
	 *
	 * @param rollBack true indicates that a rollback is on the way
	 */
	public void setRollBack(boolean rollBack)
	{
		_rollBack = rollBack;
	}

	/**
	 * The rollBack status of this datasource.
	 *
	 * @return true during a rollback
	 */
	public boolean isRollBack()
	{
		return _rollBack;
	}

	/**
	 * Internal method used to build an event to send.
	 *
	 * @param  type     The type of the event
	 * @param  object   The object of the event
	 * @param  values   The values of the event
	 * @param  loading  The flag loading of the event
	 * @param  notified The flag notified of the event
	 * @param  local    The flag local of the event
	 * @return The build event.
	 */
	protected LyDataSourceEvent eventFactory(short type, Object object, LyValueSet values,
		boolean loading, boolean notified, boolean local)
	{
		LyDataSourceEvent event = new LyDataSourceEvent(this, type, object, values);

		event.setLoading(loading);
		event.setNotified(notified);
		event.setLocal(local);
		return event;
	}

	/**
     * Marks the associated cache as loaded or not.
     *
     * @param loaded True to mark the cache loaded, false otherwise.
     */
    protected void setLoaded(boolean loaded)
    {
        if (_cache != null)
            _cache.setLoaded(loaded);
    }
} // end class LyDataSource
