/*
 * LyDataSourceEvent.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

/**
 * This class implements an event of creation/modification/deletion for an object within a data
 * source. The event is sent to 'registered listeners' : classes which implements the
 * LyDataSourceListener interface and which were added as listener of the dataSource.
 *
 * @see LyDataSource
 * @see LyDataSourceListener
 */
public class LyDataSourceEvent
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyDataSourceEvent.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Event type : ALL, all the types of events. */
	public static final short  TYPE_ALL = -1;

	/** Event type : CREATE, creation of one object. creating a new object */
	public static final short  TYPE_CREATE = 0;

	/** Event type : SET, modification of one object. */
	public static final short  TYPE_SET = 1;

	/** Event type : DELETE, deletion of one object. */
	public static final short  TYPE_DELETE = 2;

	/** Event type : SELECTION_CHANGED, selection change of one object. */
	public static final short  TYPE_SELECTION = 3;

	/** Event type : VISIBILITY_CHANGED, visibility change of one object. */
	public static final short  TYPE_VISIBILITY = 4;

	/** Event type : ERROR_CHANGED, state change of one object. */
	public static final short  TYPE_ERROR = 5;

	/** Event type : ERROR_ADDED, error added to one object. */
	public static final short  TYPE_ERROR_ADDED = 6;

	/** Event type : ERROR_REMOVED, error removed from one object. */
	public static final short  TYPE_ERROR_REMOVED = 7;

	/**
	 * Constant : LAST_TYPE, the last number type used by Leonardi. User types must be greater than
	 * this value.
	 */
	public static final short  LAST_TYPE = 100;

	/** The data source which has thrown this event. */
	LyDataSource			   _source;

	/** The type of event which occured. One of the TYPE_XXX constants */
	short					   _type;

	/** The object concerned by this notification. */
	Object					   _object;

	/**
	 * The set of values associated to this event. May be null. Only modification events have values
	 * : they are old values of the modified object (values of the object before the modification
	 * for fields that were modified).
	 */
	LyValueSet				   _values;

	/** This boolean indicates that the event comes from a loading or an unloading request. */
	boolean					   _loading;

	/** This boolean indicates that the event comes a notification. */
	boolean					   _notified;

	/** This boolean indicates that the event comes from a local operation. */
	boolean					   _local;

	/**
	 * Error added or removed on object concerned by this event. Used for event types : TYPE_ERROR,
	 * TYPE_ERROR_ADDED, TYPE_ERROR_REMOVED.
	 */
	LyError					   _error;

	/** Id of the group of this event. This id is used when a notifier receives a grouped event. */
	String					   _groupId;

	/** Additional informations for this event. */
	LyEventInfo				   _eventInfo;

	/**
	 * Constructs a data source event of the given type with the specified values.
	 *
	 * @param source Transmitter which created the event.
	 * @param type   Type of event which occured. One of the TYPE_XXX constants.
	 * @param object Object concerned by this notification.
	 * @param values Set of values associated to this event. May be null.
	 */
	public LyDataSourceEvent(LyDataSource source, short type, Object object,
		LyValueSet values)
	{
		_type = type;
		_object = object;
		_source = source;
		_values = values;
		_error = null;
		_loading = false;
		_notified = false;
		_local = false;
		_groupId = null;
	}

	/**
	 * Constructs a data source event of the given type with no associated values.
	 *
	 * @param source Transmitter which created the event.
	 * @param type   Type of event which occured. One of the TYPE_XXX constants.
	 * @param object Object concerned by this notification.
	 */
	public LyDataSourceEvent(LyDataSource source, short type, Object object)
	{
		this(source, type, object, null);
	}

	/**
	 * Gets type of this event.
	 *
	 * @return Type of this event. One of the TYPE_XXX constants
	 */
	public short getType()
	{
		return _type;
	}

	/**
	 * Gets data source which created and threw this event.
	 *
	 * @return Data source which transmits this event.
	 */
	public LyDataSource getSource()
	{
		return _source;
	}

	/**
	 * Gets object concerned by this notification.
	 *
	 * @return Object concerned by this event.
	 */
	public Object getObject()
	{
		return _object;
	}

	/**
	 * Gets the set of values associated to this event.
	 *
	 * @return Set of values of this event (may be null).
	 */
	public LyValueSet getValues()
	{
		return _values;
	}

	/**
	 * Gets the error raised or removed o the object concerned by this event. Used for event types :
	 * TYPE_ERROR, TYPE_ERROR_ADDED, TYPE_ERROR_REMOVED.
	 *
	 * @return The error (may be null).
	 */
	public LyError getError()
	{
		return _error;
	}

	/**
	 * Access to the 'loading' status. An event has loading status set to true if it was created
	 * from a load request. Internal use only (daemons)
	 *
	 * @return True if this event corresponds to a loading, false otherwise.
	 */
	public boolean isLoading()
	{
		return _loading;
	}

	/**
	 * Access to the 'notified' status. An event has notified status set to true if it was created
	 * from a notification. Internal use only (daemons)
	 *
	 * @return True if this event corresponds to a notification, false otherwise.
	 */
	public boolean isNotified()
	{
		return _notified;
	}

	/**
	 * Access to the 'local' status. An event has local status set to true if it was created from a
	 * local operation. Internal use only (daemons)
	 *
	 * @return True if this event is local, false otherwise.
	 */
	public boolean isLocal()
	{
		return _local;
	}

	/**
	 * Access to the group identifier of this event.
	 *
	 * @return The group identifier of this event. May be null if the event is not grouped.
	 */
	public String getGroupId()
	{
		return _groupId;
	}

	/**
	 * Modifier of the 'loading' status. An event is marked loaded if it was created from a load
	 * request. Internal use only (daemons)
	 *
	 * @param status True if this event corresponds to a loading, false otherwise.
	 */
	void setLoading(boolean loading)
	{
		_loading = loading;
	}

	/**
	 * Modifier of the 'notified' status. An event is marked notified if it was created from a
	 * notification. Internal use only (daemons)
	 *
	 * @param status True if this event corresponds to a notification, false otherwise.
	 */
	void setNotified(boolean notified)
	{
		_notified = notified;
	}

	/**
	 * Modifier of the 'local' status. An event is marked local if it was created from a local
	 * operation. Internal use only (daemons)
	 *
	 * @param status True if this event is local, false otherwise.
	 */
	void setLocal(boolean local)
	{
		_local = local;
	}

	/**
	 * Sets the error raised or removed o the object concerned by this event. Used for event types :
	 * TYPE_ERROR, TYPE_ERROR_ADDED, TYPE_ERROR_REMOVED.
	 *
	 * @param error The error.
	 */
	void setError(LyError error)
	{
		_error = error;
	}

	/**
	 * Sets the group identifier of this event.
	 *
	 * @param the group identifier.
	 */
	void setGroupId(String groupId)
	{
		_groupId = groupId;
	}

	/**
	 * @return Returns the eventInfo.
	 */
	public LyEventInfo getEventInfo()
	{
		return _eventInfo;
	}

	/**
	 * @param eventInfo The eventInfo to set.
	 */
	public void setEventInfo(LyEventInfo eventInfo)
	{
		_eventInfo = eventInfo;
	}
} // end class LyDataSourceEvent
