/*
 * LyDateValue.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;
import leon.info.*;


/**
 * This class offers methods for constructing, accessing and modifying values corresponding to a
 * LyDateFieldInfo.
 *
 * @see LyDateFieldInfo
 * @see LyObject
 */
public final class LyDateValue extends LyFieldInfoValue
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyDateValue.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Constructs an empty date value for the given date field.
	 *
	 * @param dateFieldInfo Date fieldInfo for which a value is created.
	 */
	public LyDateValue(LyDateFieldInfo dateFieldInfo)
	{
		super(dateFieldInfo);
	}

	/**
	 * Constructs a date value for the given date field from a string.
	 *
	 * @param dateFieldInfo Date fieldInfo for which a value is created.
	 * @param value         String value as expected by parse method.
	 * @see   LyDateFieldInfo#parse
	 */
	public LyDateValue(LyDateFieldInfo dateFieldInfo, String value)
	{
		super(dateFieldInfo, value);
	}

	/**
	 * Constructs a date value for the given date field from a Date. Depending on date type, this
	 * method extracts from the given Date requiered elements (day, month, year, hour, minutes,
	 * seconds, week, day_of_week).
	 *
	 * @param dateFieldInfo Date fieldInfo for which a value is created.
	 * @param value         Date value.
	 */
	public LyDateValue(LyDateFieldInfo dateFieldInfo, Date value)
	{
		super(dateFieldInfo, value);
	}


	/**
	 * Constructs a date value for the given date field from a Long representing the number of
	 * milliseconds since January 1, 1970, 00:00:00 GMT.
	 *
	 * @param dateFieldInfo Date fieldInfo for which a value is created.
	 * @param value         The milliseconds since January 1, 1970, 00:00:00 GMT.
	 */
	public LyDateValue(LyDateFieldInfo dateFieldInfo, Long value)
	{
		this(dateFieldInfo, ((value == null) ? (Date)null : new Date(value.longValue())));
	}
	
	/**
	 * Constructs a date value for the given date field from a long representing the number of
	 * milliseconds since January 1, 1970, 00:00:00 GMT.
	 *
	 * @param dateFieldInfo Date fieldInfo for which a value is created.
	 * @param value         The milliseconds since January 1, 1970, 00:00:00 GMT.
	 */
	public LyDateValue(LyDateFieldInfo dateFieldInfo, long value)
	{
		this(dateFieldInfo, new Date(value));
	}

	/**
	 * Gets value of this date value as a Date.
	 *
	 * @return    A Date corresponding to this value.
	 * @exception IllegalStateException If the value can't be returned as a Date (the internal type
	 *                                  of the date field is not a Date).
	 */
	public Date getDateValue()
	{
		Object value = getValue();

		if ((value != null) && !(value instanceof Date))
			throw new IllegalStateException("BF AppEngine : value " + value
				+ " for fieldInfo " + getFieldInfo() + " can't be casted to Date");

		return (Date)value;
	}

	/**
	 * Gets value of this date value as a long representing the number of milliseconds since January
	 * 1, 1970, 00:00:00 GMT.
	 *
	 * @return    A Long corresponding to this value.
	 * @exception IllegalStateException If the value can't be returned as a Long.
	 */
	public Long getLongValue()
	{
		Date date = getDateValue();

		return ((date == null) ? null : new Long(date.getTime()));
	}

	/**
	 * Gets value of this date value as a long representing the number of milliseconds since January
	 * 1, 1970, 00:00:00 GMT.
	 *
	 * @return    A long corresponding to this value.
	 * @exception IllegalStateException If the value can't be returned as a long.
	 */
	public long longValue()
	{
		Date date = getDateValue();

		return ((date == null) ? -1 : date.getTime());
	}
} // end class LyDateValue
