/*
 * LyError.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import leon.info.*;
import leon.misc.*;


/**
 * This class defines an error on an instance of LyObject.
 *
 * @see     LyObject
 * @version $Id: LyError.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyError
{
	/** File version */
	public static final String		  __VERSION =
		"$Id: LyError.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Type of error: Error on field */
	public static short				  TYPE_FIELD = 0;
	/** Type of error: Error on class */
	public static short				  TYPE_CLASS = 1;

	/** Id of the error. */
	protected String				  _id;

	/** Level of this error. */
	protected short					  _level;

	/** Message associated to this error. */
	protected String				  _message;

	/** Arguments of the message. This array must match the dynamic parts of the message. */
	protected Object[]				  _arguments;
	/** Extern Arguments of the message. */
	protected Object[]				  _externArguments;

	/** Origin of the error : this attribute may be null for a class error. */
	protected LyFieldInfo			  _origin;

	/** Object part of the cause of the error. */
	protected LyObject				  _objectCause;

	/**
	 * Field info part of the cause of the error. This attribute may be null if the cause is global
	 * to the object cause.
	 */
	protected LyInfoList<LyFieldInfo> _fieldInfosCause;

	/** Type of the error. The value must be one of the constant TYPE_FIELD or TYPE_CLASS. */
	protected short					  _type;

	/** Subtype of the error. */
	protected short					  _subType;

	/** Transmitter of the error (may be null). */
	protected Object				  _transmitter;

	/**
	 * Constructs an error. The type of the error is TYPE_FIELD is fiedInfoCause parameter is not
	 * null, TYPE_CLASS if null.
	 *
	 * @param level          Level of the error
	 * @param message        Message associated to this error
	 * @param args           List of args to the message
	 * @param origin         Origin of the error. May be null for a class error.
	 * @param objectCause    Object part cause of the error
	 * @param fieldInfoCause FieldInfo part cause of the error. May be null if tha cause is global
	 *                       to the object cause.
	 */
	public LyError(short level, String message, Object[] args,
		LyFieldInfo origin, LyObject objectCause, LyFieldInfo fieldInfoCause)
	{
		this(level, ((fieldInfoCause == null) ? TYPE_CLASS : TYPE_FIELD), LyObject.CONTROL_NONE,
			message, args, origin, objectCause,
			((fieldInfoCause == null) ? null : new LyInfoList<LyFieldInfo>(fieldInfoCause)));
	}

	/**
	 * Constructs an error.
	 *
	 * @param level           Level of the error
	 * @param type            Type of the error
	 * @param message         Message associated to this error
	 * @param args            List of args to the message
	 * @param origin          Origin of the error. May be null for a class error.
	 * @param objectCause     Object part cause of the error
	 * @param fieldInfosCause List of fieldInfos cause of the error.
	 */
	public LyError(short level, short type, String message, Object[] args,
		LyFieldInfo origin, LyObject objectCause, LyInfoList<LyFieldInfo> fieldInfosCause)
	{
		this(level, type, LyObject.CONTROL_NONE, message, args, origin, objectCause,
			fieldInfosCause);
	}

	/**
	 * Constructs an empty error.
	 *
	 * @param level          Level of the error
	 * @param type           Type of the error
	 * @param subType        Sub-type of the error
	 * @param message        Message associated to this error
	 * @param args           List of args to the message
	 * @param origin         Origin of the error. May be null for a class error.
	 * @param objectCause    Object part cause of the error
	 * @param fieldInfoCause FieldInfo part cause of the error. May be null if tha cause is global
	 *                       to the object cause.
	 */
	LyError(short level, short type, short subType, String message, Object[] args,
		LyFieldInfo origin, LyObject objectCause, LyFieldInfo fieldInfoCause)
	{
		this(level, type, subType, message, args, origin, objectCause,
			((fieldInfoCause == null) ? null : new LyInfoList<LyFieldInfo>(fieldInfoCause)));
	}

	/**
	 * Constructs an empty error.
	 *
	 * @param level           Level of the error
	 * @param type            Type of the error
	 * @param subType         Sub-type of the error
	 * @param message         Message associated to this error
	 * @param args            List of args to the message
	 * @param origin          Origin of the error. May be null for a class error.
	 * @param objectCause     Object part cause of the error
	 * @param fieldInfosCause List of fieldInfo cause of the error. May be null if the cause is
	 *                        global to the object cause.
	 */
	LyError(short level, short type, short subType, String message, Object[] args,
		LyFieldInfo origin, LyObject objectCause, LyInfoList<LyFieldInfo> fieldInfosCause)
	{
		_id = Integer.toString(hashCode()).intern();
		_level = level;
		_message = message;
		_arguments = args;
		if (args == null)
			_externArguments = null;
		else
		{
			int i, n = args.length;

			_externArguments = new Object[n];
			for (i = 0; i < n; i++)
				_externArguments[i] = args[i];
		}
		_origin = origin;
		_objectCause = objectCause;
		_fieldInfosCause = fieldInfosCause;
		_type = type;
		_subType = subType;
		_transmitter = null;
	}

	/**
	 * Gets error id.
	 * @return an id
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Returns the type of the error.
	 * @return the type of error
	 */
	public short getType()
	{
		return _type;
	}

	/**
	 * Returns the sub-type of the error.
	 * @return the sub type of this error
	 */
	public short getSubType()
	{
		return _subType;
	}

	/**
	 * Returns the level of this error.
	 * @return the error level
	 */
	public short getLevel()
	{
		return _level;
	}

	/**
	 * Returns the message of this error.
	 * @return the message associated to this error
	 */
	public String getMessage()
	{
		return _message;
	}

	/**
	 * Returns the arguments of the error message.
	 * @return an array of arguments
	 */
	public Object[] getArgs()
	{
		return _externArguments;
	}

	/**
	 * Return the cause object of this error.
	 * @return the cause object
	 */
	public LyObject getObjectCause()
	{
		return _objectCause;
	}

	/**
	 * Return the fieldInfo object of this error.
	 * @return a list of fields
	 */
	public LyInfoList<LyFieldInfo> getFieldInfosCause()
	{
		return _fieldInfosCause;
	}

	/**
	 * Return the origin of this error.
	 * @return a field
	 */
	public LyFieldInfo getOrigin()
	{
		return _origin;
	}

	/**
	 * Gets error message expanded with its arguments for the given environment.
	 * @param environment the environment
	 * @return the error message ready to be displayed
	 */
	public String getMessage(LyEnvironment environment)
	{
		if (environment == null)
		{
			if (_origin == null)
				return getMessage();
			else
				environment = _origin.getEnvironment();
		}

		return environment.getMessage(_message, _externArguments);
	}

	/**
	 * Checks equality between this error and the given one.
	 * @param o the object to compare
	 * @return true if this object and the given object are similar
	 */
	public boolean equals(Object o)
	{
		if (o == this)
			return true;
		if (!(o instanceof LyError))
			return false;

		LyError error = (LyError)o;

		if (_level != error._level)
			return false;
		if (_origin != error._origin)
			return false;
		if (_objectCause != error._objectCause)
			return false;
		if ((_fieldInfosCause != null)
				&& (error._fieldInfosCause != null))
		{
			int n, ne;

			n = _fieldInfosCause.getSize();
			ne = error._fieldInfosCause.getSize();
			if (n != ne)
				return false;
			if (!_fieldInfosCause.contains(error._fieldInfosCause))
				return false;
		}
		if (((_fieldInfosCause == null) && (error._fieldInfosCause != null))
				|| ((_fieldInfosCause != null) && (error._fieldInfosCause == null)))
			return false;
		if ((_message != null)
				&& (error._message != null)
				&& (!_message.equals(error._message)))
			return false;
		if (((_message == null) && (error._message != null))
				|| ((_message != null) && (error._message == null)))
			return false;

		int na = ((_arguments == null) ? 0 : _arguments.length);
		int nae = ((error._arguments == null) ? 0 : error._arguments.length);

		if (na != nae)
			return false;

		Object arg, argError;
		int    i;

		for (i = 0; i < na; i++)
		{
			arg = _arguments[i];
			argError = error._arguments[i];
			if ((arg != null)
					&& (argError != null)
					&& (!arg.equals(argError)))
				return false;
			if (((arg == null) && (argError != null))
					|| ((arg != null) && (argError == null)))
				return false;
		}

		return true;
	} // end method equals

	/**
	 * Dumps this object as a string.
	 * @return a string describing this object
	 */
	public String toString()
	{
		return getMessage(null);
	}

	/**
	 * Gets transmitter of the error.
	 * @return the transmitter
	 */
	public Object getTransmitter()
	{
		return _transmitter;
	}

	/**
	 * Sets transmitter of the error.
	 * @param transmitter the transmiter to use
	 */
	public void setTransmitter(Object transmitter)
	{
		_transmitter = transmitter;
	}
} // end class LyError
