/*
 * LyEventManager.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.data;

import leon.app.*;
import leon.info.*;
import leon.misc.*;

import java.util.*;


/**
 * This class is used to manage events in an application.
 *
 * @see     LyDataSource
 * @see     LyObjectSource
 * @see     LyDataSourceEvent
 * @author  Lyria
 */
public class LyEventManager
{
	/** File version */
	public static final String __VERSION = "$Id: LyEventManager.java 35628 2012-05-30 14:46:55Z bangiole $";

	public static final String									   NO_SESSION_LISTENERS = "NO_SESSION_LISTENERS".intern();


	/** Reference to the application instance. */
	protected LyApplication _application;

	/** Association between classInfos and listeners of the object of these classInfos. */
	protected HashMap<LyObjectSource, Vector<_LyDataSourceClient>> _classInfoListeners;

	/** Listeners of al classInfos. */
	protected Vector<_LyDataSourceClient> _allClassInfoListeners;

	/** Listeners of global events. */
	protected Vector<_LyGlobalClient> _globalListeners;

	/** Listeners of error events. */
	protected Vector<LyErrorListener> _errorListeners;

	/** List of association between an id of grouped events and a list of events to group. */
	protected HashMap<String, ArrayList<LyDataSourceEvent>> _groupedEvents;

	/**
	 * Constructs a transaction manager. Initialization is done with the init method.
	 *
	 * @param  application  Current application instance
	 */
	public LyEventManager(LyApplication application)
	{
		init();

		_application = application;
	}

	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		_application = null;
		_globalListeners = null;
		_classInfoListeners = new HashMap<LyObjectSource, Vector<_LyDataSourceClient>>();
		_allClassInfoListeners = new Vector<_LyDataSourceClient>();
		_groupedEvents = new HashMap<String, ArrayList<LyDataSourceEvent>>();
	}

	/**
	 * Registers a listener to receive events of the given type on objects contained within the
	 * specified object source.
	 *
	 * @param      objectSource  The object source. If this parameter is null, the listener listens
	 *                           to all type of objects.
	 * @param      listener      New listener. To receive these events, the listener have to
	 *                           implement the LyDataSourceListener interface.
	 * @param      type          Indicates which type of event the listener wants to receive. Types
	 *                           are defined as constant in LyDataSourceEvent class : TYPE_CREATE,
	 *                           TYPE_SET, TYPE_DELETE
	 *
	 * @exception  IllegalArgumentException  If the listener is null or if the given type is
	 *                                       unknown.
	 */
	public void addListener(LyObjectSource objectSource, LyDataSourceListener listener, short type)
	{
		if (listener == null)
			throw new IllegalArgumentException("BF AppEngine : listener can't be null");

		if ((type != LyDataSourceEvent.TYPE_ALL)
				&& (type < LyDataSourceEvent.TYPE_CREATE))
			throw new IllegalArgumentException("BF AppEngine : Attempt to add a listener with an unknown type of event");

		if (objectSource == null)
		{
			_LyDataSourceClient client = new _LyDataSourceClient(listener, type, null);

			_allClassInfoListeners.addElement(client);
		}
		else
		{
			Vector<_LyDataSourceClient> listeners = _classInfoListeners.get(objectSource);

			if (listeners == null)
				listeners = new Vector<_LyDataSourceClient>(1);

			synchronized (listeners)
			{
				_LyDataSourceClient client = new _LyDataSourceClient(listener, type, null);

				listeners.addElement(client);
			}

			_classInfoListeners.put(objectSource, listeners);
		}
	} // end method addListener

	/**
	 * Internal use only : An objectsource is replaced by another one. Swich the listeners to the
	 * new source
	 *
	 * @param  oldSource  the old object source
	 * @param  newSource  the new object source
	 */
	public void switchObjectSource(LyObjectSource oldSource, LyObjectSource newSource)
	{
		if ((oldSource == null) || (newSource == null))
			return;

		Vector<_LyDataSourceClient> newListeners = _classInfoListeners.get(newSource);

		if (newListeners == null)
		{
			newListeners = new Vector<_LyDataSourceClient>(1);
			_classInfoListeners.put(newSource, newListeners);
		}

		Vector<_LyDataSourceClient> oldListeners = _classInfoListeners.get(oldSource);

		synchronized (oldListeners)
		{
			for (int i = oldListeners.size() - 1; i >= 0; i--)
			{
				_LyDataSourceClient client = oldListeners.get(i);

				// To avoid loop : we must not change the registration treee of the LyExtendedObjectSource
				if ((client._listener != null)
						&& (client._listener
							instanceof LyExtendObjectSource._LyAdvancedDataSourceListener))
					continue;

				oldListeners.remove(client);
				newListeners.add(client);
			}
		}
	} // end method switchObjectSource

	/**
	 * Registers a listener to receive modification events concerning given fields on objects
	 * contained within the specified object source. Warning ! This method only registers the
	 * listener for modification events.
	 *
	 * @param      objectSource  The object source.
	 * @param      listener      New listener. To receive these events, the listener have to
	 *                           implement the LyDataSourceListener interface.
	 * @param      fields        List of fields on which the listener wants to be notified when they
	 *                           are modified.
	 *
	 * @exception  IllegalArgumentException  If the listener is null.
	 */
	public void addListener(LyObjectSource objectSource, LyDataSourceListener listener, Vector<LyField> fields)
	{
		if (listener == null)
			throw new IllegalArgumentException("BF AppEngine : listener can't be null");

		Vector<_LyDataSourceClient> listeners = _classInfoListeners.get(objectSource);

		if (listeners == null)
			listeners = new Vector<_LyDataSourceClient>(1);

		synchronized (listeners)
		{
			_LyDataSourceClient client = new _LyDataSourceClient(listener, LyDataSourceEvent.TYPE_SET, fields);

			listeners.addElement(client);
		}

		_classInfoListeners.put(objectSource, listeners);
	}

	/**
	 * Unregisters all listeners for an objectSource.
	 *
	 * @param  objectSource  The objectSource which shall not post further events.
	 */
	protected void removeListeners(LyObjectSource objectSource)
	{
		_classInfoListeners.remove(objectSource);
	}

	/**
	 * Unregisters a listener for notifications from an object source. The given listener is
	 * unregistered for all types of events he has been registered for.
	 *
	 * @param      objectSource  The object source
	 * @param      listener      The listener which shall not receive further events.
	 *
	 * @exception  IllegalArgumentException  If the listener is null or if it is not known as
	 *                                       listener on this data source.
	 */
	public void removeListener(LyObjectSource objectSource, LyDataSourceListener listener)
	{
		removeListener(objectSource, listener, LyDataSourceEvent.TYPE_ALL);
	}

	/**
	 * Unregisters a listener for notifications from an object source. The given listener is
	 * unregistered for the specified type of events he has been registered for.
	 *
	 * @param      objectSource  The object source
	 * @param      listener      The listener which shall not receive further events.
	 * @param      type          The type of event to be unregistered.
	 *
	 * @exception  IllegalArgumentException  If the listener is null or if it is not known as
	 *                                       listener on this data source or if the given type is
	 *                                       unknown.
	 */
	public void removeListener(LyObjectSource objectSource, LyDataSourceListener listener, short type)
	{
		if (listener == null)
			throw new IllegalArgumentException("BF AppEngine : Listener can't be null");

		if ((type != LyDataSourceEvent.TYPE_ALL)
				&& (type < LyDataSourceEvent.TYPE_CREATE))
			throw new IllegalArgumentException("BF AppEngine : Attempt to remove a listener with an unknown type of event");

		if (objectSource == null)
		{
			synchronized (_allClassInfoListeners)
			{
				_LyDataSourceClient client;
				int				    i, n;

				for (n = _allClassInfoListeners.size(), i = n - 1; i >= 0; i--)
				{
					client = _allClassInfoListeners.elementAt(i);
					if ((client.getListener() == listener)
							&& ((type == LyDataSourceEvent.TYPE_ALL) || (client.getType() == type)))
						_allClassInfoListeners.removeElementAt(i);
				}

				if (_allClassInfoListeners.size() == n)
					throw new IllegalArgumentException("BF AppEngine : Attempt to remove a non registered listener");
			}
		}
		else
		{
			Vector<_LyDataSourceClient> listeners = _classInfoListeners.get(objectSource);

			if (listeners == null)
				throw new IllegalStateException("BF AppEngine : Attempt to remove a non registered listener");

			synchronized (listeners)
			{
				_LyDataSourceClient client;
				int				    i, n;

				for (n = listeners.size(), i = n - 1; i >= 0; i--)
				{
					client = listeners.elementAt(i);
					if ((client.getListener() == listener)
							&& ((type == LyDataSourceEvent.TYPE_ALL) || (client.getType() == type)))
						listeners.removeElementAt(i);
				}

				if (listeners.size() == n)
					throw new IllegalArgumentException("BF AppEngine : Attempt to remove a non registered listener");
			}

			if (listeners.size() == 0)
				_classInfoListeners.remove(objectSource);
			else
				_classInfoListeners.put(objectSource, listeners);
		}
	} // end method removeListener

	/**
	 * Internal method to post an event to listeners.
	 *
	 * @param      event          Posted event.
	 * @param      skipListeners  Listeners that were already notified and that must be skipped by
	 *                            this method to avoid a double notification of the same event.
	 *
	 * @return     True if the vent was sent and used by at least one listener, false otherwise.
	 *
	 * @exception  IllegalArgumentException  If the given event is null.
	 */
	protected boolean postEvent(LyDataSourceEvent event, Vector<LyDataSourceListener> skipListeners)
	{
		if (event == null)
			throw new IllegalArgumentException("BF AppEngine : event can't be null");

		String groupId = event.getGroupId();

		if ((groupId != null) && (_groupedEvents.containsKey(groupId)))
		{
			ArrayList<LyDataSourceEvent> events = _groupedEvents.get(groupId);

			events.add(event);
			return true;
		}

		short type = event.getType();

		if ((type == LyDataSourceEvent.TYPE_ERROR_ADDED)
				|| (type == LyDataSourceEvent.TYPE_ERROR_REMOVED))
			return postErrorEvent(event);

		if ((event instanceof LyAdvancedDataSourceEvent)
				&& (type == LyDataSourceEvent.TYPE_SELECTION))
			// We always post a global selection event.
			postGlobalSelectionEvent((LyAdvancedDataSourceEvent)event);

		LyObjectSource source = (LyObjectSource)event.getSource();

		// Update the last modified time of the object source.
		source.setLastModified(System.currentTimeMillis());

		Vector<_LyDataSourceClient> listeners = _classInfoListeners.get(source);

		if (((listeners == null) || (listeners.size() == 0))
				&& (_allClassInfoListeners.size() == 0))
			return false;

		_LyDataSourceClient[] clients = null;

		if ((listeners != null)
		    && (listeners.size() > 0))
		{
		    synchronized(listeners)
	        {
	            clients = new _LyDataSourceClient[listeners.size()];
	            listeners.toArray(clients);
	        }
		}

		if (_allClassInfoListeners.size() > 0)
		{
            synchronized(_allClassInfoListeners)
            {
                int     na = _allClassInfoListeners.size();

                if (clients == null)
                {
                    clients = new _LyDataSourceClient[na];
                    _allClassInfoListeners.toArray(clients);
                }
                else
                {
                    _LyDataSourceClient[]   allClients = new _LyDataSourceClient[na];
                    _LyDataSourceClient[]   oldClients = clients;
                    int                     no = oldClients.length;

                    clients = new _LyDataSourceClient[no + na];
                    System.arraycopy(oldClients, 0, clients, 0, no);
                    _allClassInfoListeners.toArray(allClients);
                    System.arraycopy(allClients, 0, clients, no, na);
                }
            }
		}

		// To avoid a test (listeners != null)
		if (listeners == null)
			listeners = new Vector<_LyDataSourceClient>(1);

		LyEnvironment         environment = _application.getEnvironment();
		boolean		          repostGuiEvents = environment.repostGuiEvents();
	    short                 clientType;
	    int                   i, nc = clients.length;
	    LyDataSourceListener  listener;
	    boolean               used = false;
		boolean		  useLocks = environment.testResource("LY_USE_LOCK_FOR_NOTIFIER_EVENTS");
		// First we need to sort clients by sessions...
		Map<Object, List<_LyDataSourceClient>> sessionClients =
			new HashMap<Object, List<_LyDataSourceClient>>();
		List<_LyDataSourceClient> listenerList;

		Object session;
		for (i = 0; i < nc; i++)
		{
			// Check if the client is interested by the event type
			clientType = clients[i].getType();
			if ((clientType != LyDataSourceEvent.TYPE_ALL) && (clientType != type))
				continue;

			if ((!listeners.contains(clients[i]))
					&& (!_allClassInfoListeners.contains(clients[i])))
				// Be careful, some clients may disappear during this loop.
				continue;

			// Check if the listener must be skipped
			listener = clients[i].getListener();
			if ((skipListeners != null)
					&& skipListeners.contains(listener))
				continue;

			if ((listener instanceof LyGuiListener)
					&& ((LyGuiListener)listener).isIdle())
				continue;

			session = null;
			if (listener instanceof LyGetSessionInterface)
			{
				session = ((LyGetSessionInterface)listener).getSession();
			}
			if (session == null)
				session = NO_SESSION_LISTENERS;

			listenerList = sessionClients.get(session);
			if (listenerList == null)
			{
				listenerList = new ArrayList<_LyDataSourceClient>();
				sessionClients.put(session, listenerList);
			}
			listenerList.add(clients[i]);
		}

		Set<Object>			keys = sessionClients.keySet();
		Iterator<Object>	it = keys.iterator();
		_LyDataSourceClient	client;

		while(it.hasNext())
		{
			session = it.next();
			listenerList = sessionClients.get(session);

			if (session instanceof LyDataSession)
				environment = ((LyDataSession)session).getEnvironment();
			else
				environment = _application.getEnvironment();

			if (event.isNotified() && useLocks)
			{
				environment.execGui(new _LyCallMultiClient(listenerList, session, event));
				used = true;
			}
			else
			{
				int nbListener = listenerList == null ? 0 : listenerList.size();
				for (i = 0; i < nbListener; i ++)
				{
					client = listenerList.get(i);
					listener = client.getListener();
					if (repostGuiEvents
						&& repostGuiEvents(listener))
					{
						environment.execGui(new _LyCallClient(client, event));
						used = true;
					}
					else
						used |= callClient(client, event);
				}
			}
		}
		return used;
	} // end method postEvent

	synchronized boolean callClient(_LyDataSourceClient client, LyDataSourceEvent event)
	{
		short				 type = event.getType();
		boolean				 advanced = (event instanceof LyAdvancedDataSourceEvent);
		boolean				 used = false;
		LyDataSourceListener listener = client.getListener();

		checkListener(listener);

		try
		{
			// We must test if listener has some LyFilteredObjectSource in its session.
			if (!testListener(listener, event))
				return false;

			if (advanced)
			{
				// Send advanced events only to advanced listeners
				if (listener instanceof LyAdvancedDataSourceListener)
				{
					switch (type)
					{

					case LyDataSourceEvent.TYPE_SELECTION:
					case LyDataSourceEvent.TYPE_VISIBILITY:
					case LyDataSourceEvent.TYPE_ERROR:
						used |= ((LyAdvancedDataSourceListener)listener).objectStateChanged((LyAdvancedDataSourceEvent)
								event);
						break;

					default:
						used |= ((LyAdvancedDataSourceListener)listener).objectUserEvent((LyAdvancedDataSourceEvent)
								event);
						break;
					}
				}
			}
			else
			{
				// Send error events only to error listeners
				if (listener instanceof LyErrorListener)
				{
					switch (type)
					{

					case LyDataSourceEvent.TYPE_ERROR_ADDED:
						used |= ((LyErrorListener)listener).errorAdded(event);
						break;

					case LyDataSourceEvent.TYPE_ERROR_REMOVED:
						if (client.matches(event))
							used |= ((LyErrorListener)listener).errorRemoved(event);

						break;

					default:
						break;
					}
				}

				switch (type)
				{

				case LyDataSourceEvent.TYPE_CREATE:
					used |= listener.objectCreated(event);
					break;

				case LyDataSourceEvent.TYPE_SET:
					if (client.matches(event))
						used |= listener.objectModified(event);

					break;

				case LyDataSourceEvent.TYPE_DELETE:
					used |= listener.objectDeleted(event);
					break;

				default:
					break;
				}
			}
		}
		catch (Exception e)
		{
			LyDataSession session = getSession(listener);
			if (session == null)
				session = (_application == null ? null : _application.getCurrentSession());

			if (session != null)
				session.handleException(e);
			else
			{
				LyEnvironment environment = _application.getEnvironment();

				environment.writeException(e);
			}
		}

		return used;
	} // end method callClient

	/**
	 * Method to post a global selection event from an object selection event.
	 *
	 * @param  event  the object event
	 */
	protected void postGlobalSelectionEvent(LyAdvancedDataSourceEvent event)
	{
		LyGlobalEvent globalEvent;
		LyObject	  object = (LyObject)event.getObject();

		globalEvent = new LyGlobalEvent(LyGlobalEvent.TYPE_SELECTION, object);
		postEvent(globalEvent);
	}

	/**
	 * Method to post a user event on an object.
	 *
	 * @param  event  The event to post
	 */
	public void postEvent(LyAdvancedDataSourceEvent event)
	{
		short type = event.getType();

		if (type <= LyDataSourceEvent.LAST_TYPE)
			throw new IllegalArgumentException("BF AppEngine : Attempt to post an event with a bad type (" + type + ")");

		postEvent(event, null);
	}

	/**
	 * Registers a listener to receive global events of the given type.
	 *
	 * @param      listener  New listener. To receive these events, the listener have to implement
	 *                       the LyGlobalEventListener interface.
	 * @param      type      Indicates which type of event the listener wants to receive. Types are
	 *                       defined as constant in LyGlobalEvent class : TYPE_ACTIVE_VIEW,
	 *                       TYPE_SELECTION
	 *
	 * @exception  IllegalArgumentException  If the listener is null or if the given type is
	 *                                       unknown.
	 */
	public void addListener(LyGlobalEventListener listener, short type)
	{
		if (listener == null)
			throw new IllegalArgumentException("BF AppEngine : listener can't be null");

		if ((type != LyGlobalEvent.TYPE_ALL) && (type < LyGlobalEvent.TYPE_ACTIVE_VIEW))
			throw new IllegalArgumentException("BF AppEngine : Attempt to add a listener with an unknown type of event");

		if (_globalListeners == null)
			_globalListeners = new Vector<_LyGlobalClient>(1);

		synchronized (_globalListeners)
		{
			_LyGlobalClient client = new _LyGlobalClient(listener, type);

			_globalListeners.addElement(client);
		}
	}

	/**
	 * Unregisters a listener for global events. The given listener is unregistered for all types of
	 * events he has been registered for.
	 *
	 * @param      listener  The listener which shall not receive further events.
	 *
	 * @exception  IllegalArgumentException  If the listener is null or if it is not known as
	 *                                       listener.
	 */
	public void removeListener(LyGlobalEventListener listener)
	{
		removeListener(listener, LyGlobalEvent.TYPE_ALL);
	}

	/**
	 * Unregisters a listener for global events for a given type. The given listener is unregistered
	 * for the specified type of events he has been registered for.
	 *
	 * @param      listener  The listener which shall not receive further events.
	 * @param      type      The type of events to be unregistered.
	 *
	 * @exception  IllegalArgumentException  If the listener is null or if it is not known as
	 *                                       listener or if the given type is unknown.
	 */
	public void removeListener(LyGlobalEventListener listener, short type)
	{
		if (listener == null)
			throw new IllegalArgumentException("BF AppEngine : Listener can't be null");

		if ((type != LyGlobalEvent.TYPE_ALL) && (type < LyGlobalEvent.TYPE_ACTIVE_VIEW))
			throw new IllegalArgumentException("BF AppEngine : Attempt to remove a listener with an unknown type of event");

		if (_globalListeners == null)
			throw new IllegalStateException("BF AppEngine : Attempt to remove a non registered listener");

		synchronized (_globalListeners)
		{
			_LyGlobalClient client;
			int			    i, n;

			for (n = _globalListeners.size(), i = n - 1; i >= 0; i--)
			{
				client = _globalListeners.elementAt(i);
				if ((client.getListener() == listener)
						&& ((type == LyGlobalEvent.TYPE_ALL) || (client.getType() == type)))
					_globalListeners.removeElementAt(i);
			}

			if (_globalListeners.size() == n)
				throw new IllegalArgumentException("BF AppEngine : Attempt to remove a non registered listener");
		}
	} // end method removeListener

	/**
	 * Method to post a global event.
	 *
	 * @param  event  The event to post
	 */
	public void postEvent(LyGlobalEvent event)
	{
		if (event == null)
			throw new IllegalArgumentException("BF AppEngine : event can't be null");

		if ((_globalListeners == null) || (_globalListeners.size() == 0))
			return;

		int				  i, n;
		_LyGlobalClient[] clients;

		synchronized (_globalListeners)
		{
			n = _globalListeners.size();
			clients = new _LyGlobalClient[n];

			for (i = 0; i < n; i++)
				clients[i] = _globalListeners.elementAt(i);
		}

		LyEnvironment		  environment = _application.getEnvironment();
		boolean				  repostGuiEvents = environment.repostGuiEvents();
		LyGlobalEventListener listener;

		for (i = 0; i < n; i++)
		{
			if (!_globalListeners.contains(clients[i]))
				// Be careful, some clients may disappear during this loop.
				continue;

			listener = clients[i].getListener();
			if ((listener instanceof LyGuiListener)
					&& ((LyGuiListener)listener).isIdle())
				continue;

			// Check if the client is interested by the event type
			if (clients[i].matches(event))
			{
				if (repostGuiEvents
						&& repostGuiEvents(listener))
					environment.execGui(new _LyCallClient(clients[i], event));
				else
					callClient(clients[i], event);
			}
		}
	} // end method postEvent

	synchronized void callClient(_LyGlobalClient client, LyGlobalEvent event)
	{
		short				  type = event.getType();
		LyGlobalEventListener listener = client.getListener();

		checkListener(listener);

		switch (type)
		{

		case LyGlobalEvent.TYPE_ACTIVE_VIEW:
			listener.activeViewChanged(event);
			break;

		case LyGlobalEvent.TYPE_SELECTION:
			listener.selectionChanged(event);
			break;

		default:
			listener.userEvent(event);
			break;
		}
	}

	/**
	 * To add an error listener to this event manager.
	 *
	 * @param  listener  The listener to add.
	 */
	public void addErrorListener(LyErrorListener listener)
	{
		if (_errorListeners == null)
			_errorListeners = new Vector<LyErrorListener>();

		_errorListeners.addElement(listener);
	}

	/**
	 * Remove the specified listener from the list of error listener.
	 *
	 * @param  listener  The listener to remove.
	 */
	public void removeErrorListener(LyErrorListener listener)
	{
		if (listener == null)
			throw new IllegalArgumentException("BF AppEngine : Listener can't be null");

		if (_errorListeners == null)
			throw new IllegalStateException("BF AppEngine : Attempt to remove a non registered listener");

		synchronized (_errorListeners)
		{
			if (!(_errorListeners.removeElement(listener)))
				throw new IllegalArgumentException("BF AppEngine : Attempt to remove a non registered listener");
		}
	}

	/**
	 * Post an error event.
	 *
	 * @param   event  The event to post
	 *
	 * @return  true if the event was used by a listener, false otherwise.
	 */
	protected boolean postErrorEvent(LyDataSourceEvent event)
	{
		if ((_errorListeners == null) || (_errorListeners.size() == 0))
			return false;

		LyErrorListener[] clients;
		int				  i, n;

		synchronized (_errorListeners)
		{
			n = _errorListeners.size();
			clients = new LyErrorListener[n];

			for (i = 0; i < n; i++)
				clients[i] = _errorListeners.elementAt(i);
		}

		boolean		    used = false;
		LyEnvironment   environment = _application.getEnvironment();
		boolean		    repostGuiEvents = environment.repostGuiEvents();
		LyErrorListener listener;

		for (i = 0; i < n; i++)
		{
			listener = clients[i];
			if (!_errorListeners.contains(listener))
				// Be careful, some clients may disappear during this loop.
				continue;

			if ((listener instanceof LyGuiListener)
					&& ((LyGuiListener)listener).isIdle())
				continue;

			if (repostGuiEvents
					&& repostGuiEvents(listener))
			{
				environment.execGui(new _LyCallClient(listener, event));
				used = true;
			}
			else
				callClient(listener, event);
		}

		return used;
	} // end method postErrorEvent

	synchronized boolean callClient(LyErrorListener client, LyDataSourceEvent event)
	{
		boolean used = false;
		short   type = event.getType();

		// We must test if the listener has some LyFilteredObjectSource in its session.
		if (!testListener(client, event))
			return false;

		checkListener(client);

		switch (type)
		{

		case LyDataSourceEvent.TYPE_ERROR_ADDED:
			used |= client.errorAdded(event);
			break;

		case LyDataSourceEvent.TYPE_ERROR_REMOVED:
			used |= client.errorRemoved(event);
			break;

		default:
			break;
		}

		return used;
	}

	/**
	 * Internal method to post grouped events.
	 *
	 * @param  events  List of events to post.
	 */
	protected void postEvents(ArrayList<LyDataSourceEvent> events)
	{
		if ((events == null) || events.isEmpty())
			return;

		int				  n = events.size();
		LyDataSourceEvent event;

		if (n == 1)
		{
			event = events.get(0);
			postEvent(event, null);
			return;
		}

		short															    type;
		LyObject														    object;
		LyObjectSource													    source;
		ArrayList<LyObject>												    objects = new ArrayList<LyObject>();
		Vector<_LyDataSourceClient>										    listeners;
		short															    clientType;
		LyDataSourceListener											    listener;
		_LyDataSourceClient[]											    clients;
		int																    i, j, nc;
		Vector<LyAdvancedDataSourceListener>							    advancedListeners = null;
		LyAdvancedDataSourceListener									    advancedListener;
		HashMap<LyAdvancedDataSourceListener, ArrayList<LyDataSourceEvent>> eventsForListeners = null;
		ArrayList<LyDataSourceEvent>									    listenerEvents;
		Date															    date;
		LyEnvironment													    environment = _application.getEnvironment();
		boolean															    repostGuiEvents = environment.repostGuiEvents();

		// For each event, search for its listeners and stock this event
		// in an association between the listener and a list of events.
		for (i = 0; i < n; i++)
		{
			event = events.get(i);
			type = event.getType();
			object = (LyObject)event.getObject();
			source = (LyObjectSource)event.getSource();

			// If the event is a selection event, we must post a global selection event.
			if (type == LyDataSourceEvent.TYPE_SELECTION)
				postGlobalSelectionEvent((LyAdvancedDataSourceEvent)event);

			// Conserve a list of objects concerned by events during the post
			// for locking during the sending of events.
			if ((object != null) && !objects.contains(object))
				objects.add(object);

			// Update the last modified time of the object source.
			date = new Date();
			source.setLastModified(date.getTime());

			listeners = _classInfoListeners.get(source);

			if (((listeners == null) || (listeners.size() == 0))
					&& (_allClassInfoListeners.size() == 0))
				continue;

	        if ((listeners != null)
	            && (listeners.size() > 0))
	        {
	            synchronized(listeners)
	            {
	                clients = new _LyDataSourceClient[listeners.size()];
	                listeners.toArray(clients);
	            }
	        }
	        else
	            clients = null;

	        if (_allClassInfoListeners.size() > 0)
	        {
	            synchronized(_allClassInfoListeners)
	            {
	                int     na = _allClassInfoListeners.size();

	                if (clients == null)
	                {
	                    clients = new _LyDataSourceClient[na];
	                    _allClassInfoListeners.toArray(clients);
	                }
	                else
	                {
	                    _LyDataSourceClient[]   allClients = new _LyDataSourceClient[na];
	                    _LyDataSourceClient[]   oldClients = clients;
	                    int                     no = oldClients.length;

	                    clients = new _LyDataSourceClient[no + na];
	                    System.arraycopy(oldClients, 0, clients, 0, no);
	                    _allClassInfoListeners.toArray(allClients);
	                    System.arraycopy(allClients, 0, clients, no, na);
	                }
	            }
	        }

			for (j = 0, nc = clients.length; j < nc; j++)
			{
				// Check if the client is interested by the event type
				clientType = clients[j].getType();
				if ((clientType != LyDataSourceEvent.TYPE_ALL) && (clientType != type))
					continue;

				if (((listeners == null) || !listeners.contains(clients[j]))
						&& (!_allClassInfoListeners.contains(clients[j])))
					// Be careful, some clients may disappear during this loop.
					continue;

				listener = clients[j].getListener();
				if ((listener instanceof LyGuiListener)
						&& ((LyGuiListener)listener).isIdle())
					continue;

				// Check if the listener must be skipped
				if (!(listener instanceof LyAdvancedDataSourceListener))
				{
					if (repostGuiEvents
							&& repostGuiEvents(listener))
						environment.execGui(new _LyCallClient(clients[j], event));
					else
						// Call appropriate method
						callClient(clients[j], event);
				}
				else
				{
					if ((type == LyDataSourceEvent.TYPE_SET)
							&& (!clients[j].matches(event)))
						continue;

					// Stock event.
					advancedListener = (LyAdvancedDataSourceListener)listener;

					if (advancedListeners == null)
						advancedListeners = new Vector<LyAdvancedDataSourceListener>();
					if (!advancedListeners.contains(advancedListener))
					    advancedListeners.add(advancedListener);

					if (eventsForListeners == null)
					{
						eventsForListeners = new HashMap<LyAdvancedDataSourceListener, ArrayList<LyDataSourceEvent>>();
						listenerEvents = null;
					}
					else
					    listenerEvents = eventsForListeners.get(advancedListener);

					if (listenerEvents == null)
						listenerEvents = new ArrayList<LyDataSourceEvent>();
					listenerEvents.add(event);

					eventsForListeners.put(advancedListener, listenerEvents);
				}
			}
		}

		if (advancedListeners == null)
			return;

		// Post all grouped events
		for (i = 0, n = advancedListeners.size(); i < n; i++)
		{
			advancedListener = advancedListeners.elementAt(i);
			if ((advancedListener instanceof LyGuiListener)
					&& ((LyGuiListener)advancedListener).isIdle())
				continue;

			listenerEvents = eventsForListeners.get(advancedListener);
			if (repostGuiEvents
					&& repostGuiEvents(advancedListener))
				environment.execGui(new _LyCallClient(advancedListener, listenerEvents));
			else
				callListener(advancedListener, listenerEvents);
		}
	} // end method postEvents

	synchronized void callListener(LyAdvancedDataSourceListener listener, ArrayList<LyDataSourceEvent> events)
	{
		try
		{
			listener.dataSourceEventsGrouped(events);
		}
		catch (Exception e)
		{
			LyDataSession session = getSession(listener);
			if (session == null)
				session = (_application == null ? null : _application.getCurrentSession());

			if (session != null)
				session.handleException(e);
			else
			{
				LyEnvironment environment = _application.getEnvironment();

				environment.writeException(e);
			}
		}
	}

	/**
	 * Tells the manager to group all events for the specified group identifier.
	 *
	 * @param  groupId  The group identifier.
	 */
	public void groupEvents(String groupId)
	{
		if (_groupedEvents == null)
			_groupedEvents = new HashMap<String, ArrayList<LyDataSourceEvent>>();

		ArrayList<LyDataSourceEvent> events = _groupedEvents.get(groupId);

		if (events == null)
			events = new ArrayList<LyDataSourceEvent>();

		_groupedEvents.put(groupId, events);
	}

	/**
	 * To post all events grouped for the specified group identifier.
	 *
	 * @param  groupId  The group identifier.
	 */
	public void postGroupEvents(String groupId)
	{
		if (_groupedEvents == null)
			throw new IllegalArgumentException("BF AppEngine : Attempt to post group events for an unknown group ("
				+ groupId + ")");

		ArrayList<LyDataSourceEvent> events = _groupedEvents.remove(groupId);

		postEvents(events);
	}

	/* Internal classes */

	/**
	 * This class is used to store listener parameters given when the listener registered on data
	 * source events.
	 */
	final class _LyDataSourceClient
	{
		// Listener of the data source
		LyDataSourceListener _listener;

		// Type of events the listener is interested in
		short _type;

		// List of fields on which the listener wants to receive modification events
		Vector<LyField> _fields;

		/**
		 * Constructs a client of the data source.
		 *
		 * @param  listener  Listener of the data source
		 * @param  type      Type of events the listener is interested in.
		 * @param  fields    List of fields on which the listener wants to receive modification
		 *                   events (may be null, null means all fields).
		 */
		_LyDataSourceClient(LyDataSourceListener listener, short type, Vector<LyField> fields)
		{
			_listener = listener;
			_type = type;
			_fields = fields;
		}

		/**
		 * Gets listener of this client.
		 *
		 * @return  The listener that registers these parameters.
		 */
		LyDataSourceListener getListener()
		{
			return _listener;
		}

		/**
		 * Gets type of events the listener is interested in.
		 *
		 * @return  The type of events this client wants to receive.
		 */
		short getType()
		{
			return _type;
		}

		/**
		 * Verifies if a field in the given event matches one of the fields on which the listener
		 * wants to receive modification events.
		 *
		 * @param   event  Checked event.
		 *
		 * @return  True if the listener needs to be notified, false otherwise.
		 */
		boolean matches(LyDataSourceEvent event)
		{
			if ((event == null)
					|| (event.getType() != LyDataSourceEvent.TYPE_SET)
					|| ((_type != LyDataSourceEvent.TYPE_ALL)
						&& (_type != LyDataSourceEvent.TYPE_SET)))
				return false;

			if (_fields == null)
				return true;

			LyValueSet values = event.getValues();

			if (values == null)
				return true;

			LyField field;
			LyValue value;
			int     i, n = values.getSize();

			for (i = 0; i < n; i++)
			{
				value = values.getFieldValue(i);
				field = value.getField();
				if (_fields.contains(field))
					return true;
			}

			return false;
		} // end method matches
	} // end class _LyDataSourceClient

	/**
	 * This class is used to store listener parameters given when the listener registered for global
	 * events.
	 */
	final class _LyGlobalClient
	{
		// Listener of the global events
		LyGlobalEventListener _listener;

		// Type of events the listener is interested in
		short _type;

		/**
		 * Constructs a client of global events.
		 *
		 * @param  listener  Listener of the events
		 * @param  type      Type of events the listener is interested in.
		 */
		_LyGlobalClient(LyGlobalEventListener listener, short type)
		{
			_listener = listener;
			_type = type;
		}

		/**
		 * Gets listener of this client.
		 *
		 * @return  The listener that registers these parameters.
		 */
		LyGlobalEventListener getListener()
		{
			return _listener;
		}

		/**
		 * Gets type of events the listener is interested in.
		 *
		 * @return  The type of events this client wants to receive.
		 */
		short getType()
		{
			return _type;
		}

		/**
		 * Verifies if a field in the given event matches one of the fields on which the listener
		 * wants to receive modification events.
		 *
		 * @param   event  Checked event.
		 *
		 * @return  True if the listener needs to be notified, false otherwise.
		 */
		boolean matches(LyGlobalEvent event)
		{
			if (event == null)
				return false;

			if (_type == LyGlobalEvent.TYPE_ALL)
				return true;

			return (_type == event.getType());
		}
	} // end class _LyGlobalClient

	class _LyCallMultiClient
		implements Runnable,
				   LyGetSessionInterface
	{
		List<_LyDataSourceClient> _clients;
		Object _session;
		Object _event;

		_LyCallMultiClient(List<_LyDataSourceClient> client, Object session, Object event)
		{
			_clients = client;
			_event = event;
			_session = session;
		}

		public void run()
		{
			try
			{
				if (_session instanceof LyDataSession)
					((LyDataSession)_session).getHardLock();

				int i, nbClients = _clients == null ? 0 : _clients.size();
				for (i = 0; i < nbClients; i++)
					callClient(_clients.get(i), (LyDataSourceEvent)_event);
			}
			finally
			{
				if (_session instanceof LyDataSession)
					((LyDataSession)_session).releaseHardLock();
			}
		}

		/**
		 * Gets the session of current object.
		 *
		 * @return the session associated to the object.
		 */
		public LyDataSession getSession()
		{
			if (_session instanceof LyDataSession)
				return (LyDataSession)_session;
			else
				return null;
		}
	} // end class _LyCallClient

	class _LyCallClient
		implements Runnable,
		   		   LyGetSessionInterface
	{
		Object _client;
		Object _listener;
		Object _event;

		_LyCallClient(Object client, Object event)
		{
			_client = client;
			_event = event;

			if (_client instanceof _LyDataSourceClient)
				_listener = ((_LyDataSourceClient)_client).getListener();
			else if (_client instanceof _LyGlobalClient)
				_listener = ((_LyGlobalClient)_client).getListener();
			else
				_listener = client;
		}

		@SuppressWarnings("unchecked")
		public void run()
		{
			if ((_listener instanceof LyGuiListener)
					&& ((LyGuiListener)_listener).isIdle())
				return;

			if (_client instanceof _LyDataSourceClient)
				callClient((_LyDataSourceClient)_client, (LyDataSourceEvent)_event);
			else if (_client instanceof LyErrorListener)
				callClient((LyErrorListener)_client, (LyDataSourceEvent)_event);
			else if (_client instanceof _LyGlobalClient)
				callClient((_LyGlobalClient)_client, (LyGlobalEvent)_event);
			else if (_client instanceof LyAdvancedDataSourceListener)
				callListener((LyAdvancedDataSourceListener)_client,
					(ArrayList<LyDataSourceEvent>)_event);
		}

		/**
		 * Gets the session of current object.
		 *
		 * @return the session associated to the object.
		 */
		public LyDataSession getSession()
		{
			if (_listener instanceof LyGetSessionInterface)
				return ((LyGetSessionInterface)_listener).getSession();
			else
				return null;
		}
	} // end class _LyCallClient

	/**
	 * Checks listener to try to prevent problems.
	 * @param listener a listener
	 */
	protected void checkListener(Object listener)
	{
		if (listener instanceof LyWorkSpace)
		{
			// Check if the workspace has listeners to avoid memory leaks
			LyWorkSpace workSpace = (LyWorkSpace)listener;

			if (workSpace.isDynamic())
			{
				if (workSpace.getListenerCount() == 0)
				{
					LyEnvironment environment = _application.getEnvironment();
					LyDataSession session = workSpace.getSession();
					boolean		  closed = (session != null)
							&& (session.getState() == LyDataSession.State.CLOSED);

					environment.logWarning("BF AppEngine : workspace " + workSpace
						+ " is dynamic but doesn't have any listeners"
						+ (closed ? " (session closed)." : " (session alive)."));

					if (closed)
						// Force freeing
						workSpace.free();
				}
			}
		}
	} // end method checkListener

	/**
	 * Tests if the listener has some LyFilteredObjectSource in its session.
	 *
	 * @param   listener  the listener to test
	 * @param   event     original event
	 *
	 * @return  true if there an objectsource.
	 */
	protected boolean testListener(LyDataSourceListener listener, LyDataSourceEvent event)
	{
		LyDataSession session = getSession(listener);

		if (session != null)
		{
			LyObject object = (LyObject)event.getObject();

			if (object == null)
				return true;

			LyClassInfo    classInfo = object.getClassInfo();
			LyObjectSource source = session.getObjectSource(classInfo);

			if (source instanceof LyFilteredObjectSource)
			{
				// All events from the filtered object source are OK.
				LyDataSource sourceEvent = event.getSource();

				if (sourceEvent != null)
				{
					if (sourceEvent == source)
						return true;
					else if (sourceEvent instanceof LyFilteredObjectSource)
						return false;
				}

				LyFilter filter = ((LyFilteredObjectSource)source).getFilter();

				if (!filter.match(object))
					return false;
			}
		}

		return true;
	} // end method testListener

	/**
	 * get the session of the given listener if possible
	 * @param listener a listener
	 * @return a data session
	 * @see LyGetSessionInterface
	 */
	protected LyDataSession getSession(LyDataSourceListener listener)
	{
		LyDataSession session = null;

		if (listener instanceof LyGetSessionInterface)
			session = ((LyGetSessionInterface)listener).getSession();

		return session;
	}

	/**
	 * Checks whether events must be reposted for specified listener.
	 * @param listener The listener.
	 * @return True if events must be reposted, false otherwise.
	 */
	protected boolean repostGuiEvents(Object listener)
	{
		if (listener instanceof LyGuiListener)
			return true;

		if ((listener instanceof LyWorkSpace)
				|| (listener instanceof LyGetWorkSpaceInterface))
		{
			LyWorkSpace	ws = null;

			if (listener instanceof LyGetWorkSpaceInterface)
				ws = ((LyGetWorkSpaceInterface)listener).getWorkSpace();
			else
				ws = (LyWorkSpace)listener;

			int			i, n = ws.getListenerCount();

			for (i = 0; i < n; i++)
			{
				if (repostGuiEvents(ws.getListener(i)))
					return true;
			}
		}

		return false;
	}
} // end class LyEventManager
