/*
 * LyExtendObjectSource.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;
import leon.app.behaviorinterface.*;
import leon.info.*;
import leon.info.binding.*;
import leon.peer.*;
import leon.peer.extend.*;


/**
 * An extends object source is associated to a logical class (LyClassInfo) which has derived
 * classes.
 *
 * <p>Objects sources are associated to data sources to dispatch demands on the underlying physical
 * layer when requests are made for logical objects.</p>
 *
 * @see LyObject
 * @see LyClassInfo
 */
public class LyExtendObjectSource extends LyObjectSource
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyExtendObjectSource.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Object source that contains objects of the classInfo when this classInfo is not abstract but
	 * have some derived classes.
	 */
	protected LyObjectSource   _objectSource;

	/**
	 * Default constructor of an extends object source. Only used by inherited classes.
	 */
	protected LyExtendObjectSource()
	{
		_objectSource = null;
		_cache = null;
		_listener = null;
	}

	/**
	 * Constructs an extends object source for a given class info.
	 * @param provider the data provider 
	 * @param classInfo Class info for which an extends object source is constructed.
	 */
	public LyExtendObjectSource(LyDataProvider provider, LyClassInfo classInfo)
	{
		super(provider, classInfo);

		if (!classInfo.isAbstract())
			throw new IllegalStateException(
				"BF AppEngine : can't create an extended object source for non abstract class info "
				+ classInfo);

		_objectSource = null;
		_cache = null;
		_listener = null;
	}

	/**
	 * Constructs an extends object source for a given class info.
	 * @param provider the data provider
	 *
	 * @param classInfo Class info for which an extends object source is constructed.
	 * @param objectSource the objectsource
	 */
	public LyExtendObjectSource(LyDataProvider provider, LyClassInfo classInfo,
		LyObjectSource objectSource)
	{
		super(provider, classInfo);

		_objectSource = objectSource;
		_cache = null;
		_listener = null;
	}

	/**
	 * Free this object source. Remove this object source as listener of the sub classes
	 * objectsources.
	 */
	public void free()
	{
		if (_listener != null)
		{
			// Listen to subclasses to resend events
			LyObjectSource		    objectSource;
			LyClassInfo			    classInfo;
			LyInfoList<LyClassInfo> subClasses = ((LyClassInfo)_class).getSubClasses();
			int					    i, n = ((subClasses == null) ? 0 : subClasses.getSize());

			for (i = 0; i < n; i++)
			{
				classInfo = subClasses.getInfo(i);
				objectSource = _application.getObjectSource(classInfo);
				objectSource.removeListener(_listener);
			}

			if (_objectSource != null)
				_objectSource.removeListener(_listener);
		}

		super.free();
	}

	/**
	 * Internal method to listen to sub-classes.
	 */
	public void listenToSubClasses()
	{
		if (_listener != null)
			return;

		_listener = new _LyAdvancedDataSourceListener();

		// Listen to subclasses to resend events
		LyObjectSource		    objectSource;
		LyClassInfo			    classInfo;
		LyInfoList<LyClassInfo> subClasses = ((LyClassInfo)_class).getSubClasses();
		int					    i, n = ((subClasses == null) ? 0 : subClasses.getSize());

		for (i = 0; i < n; i++)
		{
			classInfo = subClasses.getInfo(i);
			objectSource = _application.getObjectSource(classInfo);
			objectSource.addListener(_listener);
		}

		if (_objectSource != null)
			_objectSource.addListener(_listener);
	}

	/**
	 * Internal method to listen to a subclass events. For internal use only.
	 *
	 * @param subClass
	 */
	protected void listenToSubClass(LyClassInfo subClass)
	{
		if (_listener == null)
			return;

		LyInfoList<LyClassInfo> subClasses = ((LyClassInfo)_class).getSubClasses();
		LyObjectSource		    objectSource;

		if ((subClasses != null)
				&& (subClasses.contains(subClass)))
		{
			objectSource = _application.getObjectSource(subClass);
			objectSource.addListener(_listener);
		}
	}

	/**
	 * Method to notify the loading of a class to an object source.
	 * @param classInfo the classInfo
	 */
	public void notifyClassLoading(LyClassInfo classInfo)
	{
		super.notifyClassLoading(classInfo);

		listenToSubClass(classInfo);
	}

	/**
	 * To get the objectSource associated with the classInfo when this classInfo is not abstract.
	 *
	 * @return the object source that managed the objects of the classInfo.
	 */
	public LyObjectSource getObjectSource()
	{
		return _objectSource;
	}

	/**
	 * Internal method to get the proper object source of a given class info : ask to the current
	 * session, then to the current application.
	 *
	 * @param  classInfo the class for wich the object source is requested.
	 * @return an object source
	 */
	protected LyObjectSource getObjectSource(LyClassInfo classInfo)
	{
		LyObjectSource		   objectSource = null;
		leon.app.LyDataSession session = _application.getCurrentSession();

		if (session != null)
			objectSource = session.getObjectSource(classInfo);

		if (objectSource == null)
			objectSource = _application.getObjectSource(classInfo);

		return objectSource;
	}

	/**
	 * Sets the cache for this object source.
	 *
	 * @param     cachePolicy Informations for the cache to associate to this object source.
	 * @exception IllegalArgumentException If the given cache is null.
	 * @exception IllegalStateException    If the object source is already associated to a cache
	 *                                     that contains objects.
	 */
	public void setCache(LyInfoCache cachePolicy)
	{
		if (_objectSource != null)
			_objectSource.setCache(cachePolicy);
	}

	/**
	 * Recompute the object id of the specified object.
	 */
	protected void rehash(LyObject object, boolean sendEvent)
	{
		if (_objectSource != null)
			_objectSource.rehash(object, sendEvent);
	}

	/**
	 * Tests if the given object is found in the cache of this data source.
	 *
	 * @param  object Object searched.
	 * @return True if the given object is found in the cache, else false.
	 */
	public boolean checkObject(Object object)
	{
		if ((_objectSource != null) && (_objectSource.checkObject(object)))
			return true;

		LyClassInfo classInfo = ((LyObject)object).getClassInfo();

		if (classInfo.inheritsFrom((LyClassInfo)_class))
		{
			LyObjectSource objectSource = classInfo.getObjectSource();

			return objectSource.checkObject(object);
		}

		return false;
	}

	/**
	 * Tests if an object corresponding to a given object id is found in the cache of this data
	 * source.
	 *
	 * @param  objectId Object id for which an object is searched.
	 * @return True if an object with the given object id is found in the cache, else false.
	 */
	public boolean checkObject(LyObjectId objectId)
	{
		if ((_objectSource != null) && (_objectSource.checkObject(objectId)))
			return true;

		LyClassInfo classInfo = objectId.getClassInfo();

		if (classInfo == (LyClassInfo)_class)
		{
			LyInfoList<LyClassInfo>		   subClasses = ((LyClassInfo)_class).getSubClasses();
			int							   i, n = ((subClasses == null) ? 0 : subClasses.getSize());
			LyObjectSource				   objectSource;
			String						   str = objectId.toString();
			LyObjectId					   id;

			// Try on behavior
			LyApplicationBehaviorInterface behavior = _application.getApplicationBehavior();

			if (behavior != null)
				subClasses = behavior.getSubClasses(_application, objectId, null);

			// If not found try a generic search
			if (subClasses == null)
				subClasses = ((LyClassInfo)_class).getSubClasses();

			n = ((subClasses == null) ? 0 : subClasses.getSize());
			for (i = 0; i < n; i++)
			{
				classInfo = subClasses.getInfo(i);
				id = new LyObjectId(classInfo, str);
				objectSource = _application.getObjectSource(classInfo);
				if (objectSource.checkObject(id))
					return true;
			}

		}
		else if (classInfo.inheritsFrom((LyClassInfo)_class))
		{
			LyObjectSource objectSource = classInfo.getObjectSource();

			return objectSource.checkObject(objectId);
		}

		return false;
	} // end method checkObject

	/**
	 * Loads completely this object source : loads the associated cache according to the cache
	 * policy.
	 */
	public void load()
	{
		if (_objectSource != null)
			_objectSource.load();
	}

	/**
	 * Unloads completely this object source : unloads the associated cache and data source. This
	 * method does noting since no cache is supported on this source
	 *
	 * @param notify indicates wether the unload action should notify objects deletion. Notifying
	 *               deletions makes sense when views are opened and when application is not
	 *               exiting.
	 */
	public void unload(boolean notify)
	{
		if (_objectSource != null)
			_objectSource.unload(notify);
	}

	/**
	 * Registers a listener to receive events of the given type on objects contained within this
	 * data source.
	 *
	 * @param     listener New listener. To receive these events, the listener have to implement the
	 *                     LyDataSourceListener interface.
	 * @param     type     Indicates which type of event the listener wants to receive. Types are
	 *                     defined as constant in LyDataSourceEvent class : TYPE_CREATE, TYPE_SET,
	 *                     TYPE_DELETE
	 * @exception IllegalArgumentException If the listener is null or if the given type is unknown.
	 */
	public void addListener(LyDataSourceListener listener, short type)
	{
		if (_listener == null)
			listenToSubClasses();

		super.addListener(listener, type);
	}

	/**
	 * Registers a listener to receive modification events concerning given fields on objects
	 * contained within this data source. Warning ! This method only registers the listener for
	 * modification events.
	 *
	 * @param     listener New listener. To receive these events, the listener have to implement the
	 *                     LyDataSourceListener interface.
	 * @param     fields   List of fields on which the listener wants to be notified when they are
	 *                     modified.
	 * @exception IllegalArgumentException If the listener is null.
	 */
	public void addListener(LyDataSourceListener listener, Vector<LyField> fields)
	{
		if (_listener == null)
			listenToSubClasses();

		super.addListener(listener, fields);
	}

	/**
	 * Gets an object from this object source by the specified identifier.
	 *
	 * @param  id     Object identifier of the searched object.
	 * @param  godown indicates if the object is searched recursivly in inherited classes.
	 * @param  local  indicates if the request is local (from memory) or not.
	 * @return The requested object if it was found, or null if it could not be found. This object
	 *         is an instance of LyObject (or an inherited class).
	 */
	public LyObject getObject(LyObjectId id, boolean godown, boolean local)
	{
		LyObject   result = null;
        
        // First try locally
        if (!local)
        {
            result = getObject(id, godown, true);
            if (result != null)
                return result;
        }
        
		LyObjectId objectId;
		String     str = id.toString();

		if (_objectSource != null)
		{
			objectId = new LyObjectId(_class, str);
			result = (LyObject)_objectSource.getObject(objectId, local);
			if (result != null)
				return result;
		}
		if (!godown)
			return result;

		// Then, try on behavior
        LyInfoList<LyClassInfo>         subClasses = null;
		LyApplicationBehaviorInterface  behavior = _application.getApplicationBehavior();

		if (behavior != null)
			subClasses = behavior.getSubClasses(_application, id, null);

		// If not found try a generic search
		if (subClasses == null)
			subClasses = ((LyClassInfo)_class).getSubClasses();

        int                     i, n = ((subClasses == null) ? 0 : subClasses.getSize());
        LyClassInfo             classInfo;
        LyObjectSource          objectSource;
        
		for (i = 0; i < n; i++)
		{
			classInfo = subClasses.getInfo(i);
			objectId = new LyObjectId(classInfo, str);
			objectSource = _application.getObjectSource(classInfo);
			result = (LyObject)objectSource.getObject(objectId, local);
			if (result != null)
				return result;
		}

		return null;
	} // end method getObject

	/**
	 * Gets an object from this object source by the specified identifier.
	 *
	 * @param  id Object identifier of the searched object.
	 * @return The requested object if it was found, or null if it could not be found. This object
	 *         is an instance of LyObject (or an inherited class).
	 */
	public Object getObject(LyObjectId id)
	{
		return getObject(id, true, false);
	}

	/**
	 * Gets an object from this object source by the specified identifier.
	 *
	 * @param  id    Object identifier of the searched object.
	 * @param  local Indicates if the demand is local (gets object from memory) or not.
	 * @return The requested object if it was found, or null if it could not be found. This object
	 *         is an instance of LyObject (or an inherited class).
	 */
	public Object getObject(LyObjectId id, boolean local)
	{
		return getObject(id, true, local);
	}

	/**
	 * Return the number of objects within this object source matching the given filter.
	 *
	 * @param  filter Filter applied on objects for counting.
	 * @return Number of objects matching the given filter in this object source.
	 */
	public int getSize(LyFilter filter)
	{
		return getSize(filter, false);
	}

	/**
	 * Return the number of objects within this object source matching the given filter.
	 *
	 * @param  filter Filter applied on objects for counting.
	 * @param  local  Indicates if the request must be local or not.
	 * @return Number of objects matching the given filter in this object source.
	 */
	public int getSize(LyFilter filter, boolean local)
	{
		if (local)
		{
			// Create a subrequest for each sub classe.
			int					    i, nbClasses, size;
			LyClassInfo			    subClass;
			LyObjectSource		    objectSource;
			LyInfoList<LyClassInfo> subClasses = ((LyClassInfo)_class).getSubClasses();

			nbClasses = ((subClasses == null) ? 0 : subClasses.getSize());
			size = 0;

			// Maybe the class is not abstract.
			if (_objectSource != null)
				// Class is not abstract, we must send a request to its objectSource.
				size += _objectSource.getSize(filter, local);

			for (i = nbClasses - 1; i >= 0; i--)
			{
				subClass = subClasses.getInfo(i);
				objectSource = getObjectSource(subClass);
				size += objectSource.getSize(filter, local);
			}

			return size;
		}
		// Send the request.
		else if ((_objectSource == null) && (((LyClassInfo)_class).getBinding() != null))
		{
			// Send the using abstract class binding (no join or union).
			synchronized (_provider)
			{
				try
				{
					((LyExtendDataProvider)_provider).setUseClassBindingForAbstractClass(true);

					int size = super.getSize(filter, local);

					return size;
				}
				catch (RuntimeException e)
				{
					throw e;
				}
				finally
				{
					((LyExtendDataProvider)_provider).setUseClassBindingForAbstractClass(false);
				}
			}
		}
		else
		{
			// We must send a request to the data provider to know.
			// The request must be synchronous.
			boolean			  synchronous;
			LyResponse		  response;
            LyResponse.Status status;
			LyRequestEncoder  encoder;

			synchronous = isSynchronous();
			setSynchronous(true);

			encoder = _provider.getRequestEncoder(this);
			response = new LyResponse(LyResponse.Operation.COUNT, _class);
			encoder.setResponse(response);
			encoder.getSize(filter);

			// Set the older synchronous mode.
			setSynchronous(synchronous);

			// Check the status of the response.
			status = response.getStatus();

			if ((status == LyResponse.Status.COMPLETED) && (response.getSize() == 1))
			{
				Integer size = (Integer)response.getResponse(0);

				return size.intValue();
			}
			else
				return -1;
		}
	} // end method getSize

	/* Methods for creating an object */

	/**
	 * Given an object id corresponding to this extend object source, compute the real classinfo
	 * (corresponding to the correct sub class.
	 *
	 * @param  object the LyObject
	 * @return the correct classInfo for this object.
	 */
	public LyClassInfo getRealClassInfo(LyObject object)
	{
		LyClassInfo			    classInfo;
		LyClassBinding		    classBinding;
		LyFilter			    filter;
		LyInfoList<LyClassInfo> subClasses = ((LyClassInfo)_class).getSubClasses();
		LyClassInfo			    classInfoResult = null;

		for (int i = 0; i < subClasses.getSize(); i++)
		{
			classInfo = subClasses.getInfo(i);

			classInfoResult = classInfo.getObjectSource().getRealClassInfo(object);
			if (classInfoResult != null)
				return classInfoResult;

			classBinding = classInfo.getBinding();
			filter = classBinding.getFilter();
			if (filter == null)
				// class is valid
				return classInfo;

			if (filter.match(object))
				return classInfo;
		}

		classBinding = ((LyClassInfo)_class).getBinding();
		filter = classBinding.getFilter();

		if (filter == null)
			return (LyClassInfo)_class;

		if (filter.match(object))
			return (LyClassInfo)_class;

		return null;

	} // end method getRealClassInfo

	/**
	 * Create a new object in this object source. This method commits local creation by creating the
	 * corresponding physical object.
	 *
	 * @param     object   Object to create.
	 * @param     listener Listener of the response.
	 * @param     local    Indicates if the creation is local or not.
	 * @param     refresh  Indicates if the creation may be notified or not.
	 * @return    Response to the creation.
	 * @exception IllegalStateException If the object source is in read only mode.
	 */
	public LyResponse createObject(LyObject object, LyResponseListener listener, boolean local,
		boolean refresh)
	{
		if (_objectSource != null)
		{
			LyClassInfo classInfo = object.getClassInfo();

			if (classInfo == _class)
				return _objectSource.createObject(object, listener, local, refresh);
		}
		throw new IllegalStateException(
			"BF AppEngine : can't create an object from an extends object source with an abstract class");
	}

	/* Methods for setting objects */

	/**
	 * Set an object of this object source. This method commits local modification by setting the
	 * corresponding physical object.
	 *
	 * @param     object   Object to modify.
	 * @param     listener Listener of the response.
	 * @param     local    Indicates if the modification must be local or not.
	 * @param     refresh  Indicates if the modification may be notified or not.
	 * @return    Response to the modification.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If the given object is invalid.
	 */
	public LyResponse setObject(LyObject object, LyResponseListener listener, boolean local,
		boolean refresh)
	{
		return setObject(object, listener, local, refresh, true);
	}

	protected LyResponse setObject(LyObject object, LyResponseListener listener,
		boolean local, boolean refresh, boolean needCheck)
	{
		if (_objectSource != null)
		{
			LyClassInfo classInfo = object.getClassInfo();

			if (classInfo == _class)
				return _objectSource.setObject(object, listener, local, refresh, needCheck);
		}
		throw new IllegalStateException(
			"BF AppEngine : can't set an object from an extends object source with an abstract class");
	}

	/**
	 * Set a list of objects of this object source. This method commits local modifications by
	 * setting the corresponding physical objects.
	 *
	 * @param     object   List of objects to modify.
	 * @param     listener Listener of the response.
	 * @return    Response to the modification.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If one of the given objects is invalid.
	 */
	public LyResponse setObjects(Vector<? extends Object> objects, LyResponseListener listener)
	{
		// We can send the request to the data provider.
		LyResponse		 response;
		LyRequestEncoder encoder;

		encoder = _provider.getRequestEncoder(this);
		response = new LyResponse(LyResponse.Operation.SET, _class);
		if (listener != null)
			response.setListener(listener);

		encoder.setResponse(response);
		encoder.setObjects(new Vector<Object>(objects));

		return response;
	}

	/* Methods for deleting objects */

	/**
	 * Delete an object of this object source. This method commits local deletion by deleting the
	 * corresponding physical object.
	 *
	 * @param     object            Object to delete.
	 * @param     listener          Listener of the response.
	 * @param     local             Indicates if the operation is local or not.
	 * @param     refresh           Indicates if the operation may be notified or not.
	 * @return    Response to the deletion.
	 * @param     deleteLinkObjects Indicates if linked objects (composition relations,...)
	 *                              must be deleted or not.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If the given object is invalid.
	 */
	public LyResponse deleteObject(LyObject object, LyResponseListener listener, boolean local,
		boolean refresh, boolean deleteLinkedObjects)
	{
		if (_objectSource != null)
		{
			LyClassInfo classInfo = object.getClassInfo();

			if (classInfo == _class)
				return _objectSource.deleteObject(object, listener, local, refresh,
					deleteLinkedObjects);
		}

		throw new IllegalStateException(
			"BF AppEngine : can't delete an object from an extends object source with an abstract class");
	}

	/**
	 * Delete a list of objects of this object source. This method commits local deletions by
	 * deleting the corresponding physical objects.
	 *
	 * @param     objects           List of objects to delete.
	 * @param     listener          Listener of the response.
	 * @param     local             Indicates if the operation is local or not.
	 * @param     refresh           Indicates if the operation may be notified or not.
	 * @param     deleteLinkObjects Indicates if linked objects (composition relations, ...)
	 *                              must be deleted or not.
	 * @return    Response to the deletion.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If one of the given objects is invalid.
	 */
	public LyResponse deleteObjects(Vector<? extends Object> objects, LyResponseListener listener,
		boolean local,
		boolean refresh, boolean deleteLinkedObjects)
	{
		// We can send the request to the data provider.
		LyResponse		 response;
		LyRequestEncoder encoder;

		encoder = _provider.getRequestEncoder(this);
		response = new LyResponse(LyResponse.Operation.DELETE, _class);
		response.setLocal(local);
		response.setRefresh(refresh);
		response.deleteLinkedObjects(deleteLinkedObjects);
		if (listener != null)
			response.setListener(listener);

		encoder.setResponse(response);
		encoder.deleteObjects(new Vector<Object>(objects));

		return response;
	}

	/* Methods for accessing objects */

	/**
	 * Retrieve objects on this object source. Retrieve a specified number of objects sorted by a
	 * given sort and that match a given filter. These objects are first searched in the associated
	 * cache to avoid requests to the physical layer.
	 *
	 * @param  sort     Gives sort criteria, or null if there is no sort (the order is the one of
	 *                  the physical layer).
	 * @param  filter   Gives filter criteria, or null if there is no filter (all objects).
	 * @param  n        Indicates the number of objects to retrieve (-1 means no limit).
	 * @param  local    Indicates if the request is local (from memory) or not.
	 * @param  listener Listener of the response.
	 * @return Response to this search.
	 */
	public LyResponse getObjects(Vector<LyFieldInfo> fieldInfos, LySort sort, LyFilter filter,
		int n,
		boolean local, LyResponseListener listener)
	{
		if (local)
		{
			// Create a subrequest for each sub classe.
			int					    i, p, nbClasses;
			LyClassInfo			    subClass;
			LyObjectSource		    objectSource;
			LyInfoList<LyClassInfo> subClasses = ((LyClassInfo)_class).getSubClasses();
			LyResponse			    response;
			LySimpleObjectList	    list = new LySimpleObjectList();
			LyObject			    object;

			nbClasses = ((subClasses == null) ? 0 : subClasses.getSize());
			list.setSorted(false);
			list.setSort(sort);

			// Maybe the class is not abstract.
			if (_objectSource != null)
			{
				// Class is not abstract, we must send a request to its objectSource.
				response = _objectSource.getObjects(fieldInfos, sort, filter, n, local, null);
				for (i = 0; i < response.getSize(); i++)
				{
					object = (LyObject)response.getResponse(i);
					list.addObject(object);
				}
			}

			for (i = nbClasses - 1; i >= 0; i--)
			{
				subClass = subClasses.getInfo(i);
				objectSource = getObjectSource(subClass);
				response = objectSource.getObjects(fieldInfos, sort, filter, n, local, null);
				for (int j = 0; j < response.getSize(); j++)
				{
					object = (LyObject)response.getResponse(j);
					list.addObject(object);
				}
			}

			response = new LyResponse(LyResponse.Operation.LOAD, _class);
			if (listener != null)
				response.setListener(listener);
			response.setStatus(LyResponse.Status.RUNNING);
			response.setFiltered(true);
			response.setCounted(true);
			if (n == -1)
			{
				// There is no limit so we let the asker do the job.
				response.setSorted(false);
				p = list.getSize();
			}
			else
			{
				// There is a limit : we must sort the list to return the good result.
				response.setSorted(true);
				list.setSorted(true);
				p = n;
				if (p > list.getSize())
					// The result may have less results than the limit.
					p = list.getSize();
			}
			for (i = 0; i < p; i++)
			{
				object = list.getObject(i);
				response.addResponse(object, null);
			}
			response.setStatus(LyResponse.Status.COMPLETED);
			response.addResponse(null, null);
			return response;
		}
		else
			return loadObjects(fieldInfos, sort, filter, n, listener);
	} // end method getObjects

	/**
	 * Finds an object from its object peer (associated physical object). The object is searched in
	 * the cache and if it is not found, it is created.
	 *
	 * @param  peer Physical object.
	 * @return The logical object that was found or created, or null if the creation fails.
	 */
	public LyObject getObjectFromPeer(Object peer)
	{
		if (_objectSource != null)
			return _objectSource.getObjectFromPeer(peer);
		throw new IllegalStateException(
			"BF AppEngine : can't get an object from peer from an extends object source with an abstract class");
	}

	/**
	 * Create an object from a given peer object (associated physical object).
	 *
	 * @param  peer Peer object.
	 * @return Created object or null if the object doesn't match the class binding of the class
	 *         info.
	 */
	protected LyObject createObjectFromPeer(Object peer)
	{
		if (_objectSource != null)
			return _objectSource.createObjectFromPeer(peer);
		throw new IllegalStateException(
			"BF AppEngine : can't create an object from peer from an extends object source with an abstract class");
	}

	/**
	 * Internal method used to receive a new response from data provider to a creation request sent
	 * on createObject. Object is added in the cache of this data source and a creation event is
	 * sent to listeners. If the data provider has a notifier, the creation is notified.
	 *
	 * @param response Response to the create operation.
	 * @param status   Status of the new response. Valid status are defined in LyResponse.
	 * @param object   Newly created object.
	 * @see   #createObject
	 */
	protected void receiveCreate(LyResponse response, LyResponse.Status status, Object object)
	{
		throw new IllegalStateException(
			"BF AppEngine : can't receive create in an extends object source");
	}

	/**
	 * Internal method used to receive a new response from data provider to a set request sent on
	 * setObject or setObjects. This method is overriden to filter objects according to class
	 * binding.
	 *
	 * @param response  Response to the set operation.
	 * @param status    Status of the new response. Valid status are defined in LyResponse.
	 * @param object    Modified object.
	 * @param oldValues Old values of the object before the set operation.
	 * @see   #setObject
	 * @see   #setObjects
	 */
	protected void receiveSet(LyResponse response, LyResponse.Status status, Object object,
		LyValueSet oldPeerValues)
	{
		throw new IllegalStateException("BF AppEngine : can't receive set in an extends object source");
	}

	/**
	 * Internal method used to receive a new response from data provider to a deletion request sent
	 * on deleteObject or deleteObjects. A deletion event is sent to listeners and the object is
	 * deleted from the cache If the data provider has a notifier, the deletion is notified.
	 *
	 * @param response Response to the deletion operation.
	 * @param status   Status of the new response. Valid status are defined in LyResponse.
	 * @param object   Deleted object.
	 * @see   #deleteObject
	 * @see   #deleteObjects
	 */
	protected void receiveDelete(LyResponse response, LyResponse.Status status, Object object)
	{
		throw new IllegalStateException(
			"BF AppEngine : can't receive delete in an extends object source");
	}

	/**
	 * Internal method used to receive a new response from data provider to a load request sent on
	 * loadObject or loadObjects. This method is overriden to filter objects according to class
	 * binding.
	 *
	 * @param response Response to the load operation.
	 * @param status   Status of the new response. Valid status are defined in LyResponse.
	 * @param object   Loaded object.
	 * @see   #loadObject
	 * @see   #loadObjects
	 */
	protected void receiveLoad(LyResponse response, LyResponse.Status status, Object object,
		LyValueSet oldValues)
	{
		throw new IllegalStateException(
			"BF AppEngine : can't receive load in an extends object source");
	}

	/**
	 * To validate this objectSource. This method scans all fields of the associated class info and
	 * register this object source as listener on all object sources corresponding to classes that
	 * are pointed by relations. The purpose is to receive events on id. and name changes on related
	 * objects to transform them as modifications on objects managed by this objects source. As
	 * related objects are presented by their names, the goal is to ease the refreshing of the
	 * display when names change.
	 */
	public void validate()
	{
		// Nothing to do if the class is abstract otherwise validate local object source.
		if (_objectSource != null)
			_objectSource.validate();
	}

	/**
	 * To get the instance of LyDataSourceListener of this object source.
	 *
	 * @return the instance of LyDataSourceListener of this object source.
	 */
	public LyDataSourceListener getListener()
	{
		return null;
	}

	/* Implementation of LyDataSourceListener */

	final class _LyAdvancedDataSourceListener
		implements LyAdvancedDataSourceListener
	{

		/**
		 * Method to transfer events from sublcasses object sources to the listeners of this extend
		 * object source.
		 */
		protected boolean rePost(LyDataSourceEvent event)
		{
			if (event instanceof LyAdvancedDataSourceEvent)
			{
				LyAdvancedDataSourceEvent myEvent;

				myEvent = new LyAdvancedDataSourceEvent(LyExtendObjectSource.this, event.getType(),
						event.getObject(), ((LyAdvancedDataSourceEvent)event).getOldState());
				myEvent.setLoading(event.isLoading());
				myEvent.setNotified(event.isNotified());
				myEvent.setLocal(event.isLocal());
				return postEvent(myEvent);
			}
			else
			{
				LyDataSourceEvent myEvent;

				myEvent = new LyDataSourceEvent(LyExtendObjectSource.this, event.getType(),
						event.getObject(), event.getValues());
				myEvent.setLoading(event.isLoading());
				// FFT01455: Propagate notified status
				myEvent.setNotified(event.isNotified());
				myEvent.setLocal(event.isLocal());
				return postEvent(myEvent);
			}
		}

		/**
		 * Method for receiving an object creation.
		 *
		 * @param  event Event notifying the creation of an object.
		 * @return Boolean indicating if the event was used, false otherwise.
		 */
		public boolean objectCreated(LyDataSourceEvent event)
		{
			return rePost(event);
		}

		/**
		 * Method for receiving an object deletion.
		 *
		 * @param  event Event notifying the deletion of an object.
		 * @return Boolean indicating if the event was used, false otherwise.
		 */
		public boolean objectDeleted(LyDataSourceEvent event)
		{
			return rePost(event);
		}

		/**
		 * Method for receiving an object modification.
		 *
		 * @param  event Event notifying the modification of an object.
		 * @return Boolean indicating if the event was used, false otherwise.
		 */
		public boolean objectModified(LyDataSourceEvent event)
		{
			return rePost(event);
		}

		/**
		 * Method for receiving state changes on an object.
		 *
		 * @param  event Event notifying the state change of an object.
		 * @return True indicating if the event was used, false otherwise.
		 */
		public boolean objectStateChanged(LyAdvancedDataSourceEvent event)
		{
			return rePost(event);
		}

		/**
		 * Method for receiving user event on an object.
		 *
		 * @param  event Event notifying the user event of an object.
		 * @return True indicating if the event was used, false otherwise.
		 */
		public boolean objectUserEvent(LyAdvancedDataSourceEvent event)
		{
			return rePost(event);
		}

		/**
		 * Called when grouped events are sent.
		 *
		 * @param  events Sorted list of events.
		 * @return True if the events are used by the listener, false otherwise.
		 */
		public boolean dataSourceEventsGrouped(ArrayList<LyDataSourceEvent> events)
		{
			if ((events == null) || (events.isEmpty()))
				return false;

			int							 i, n = events.size();
			LyDataSourceEvent			 event, myEvent;
			ArrayList<LyDataSourceEvent> myEvents = new ArrayList<LyDataSourceEvent>();

			for (i = 0; i < n; i++)
			{
				event = (LyDataSourceEvent)events.get(i);

				if (event instanceof LyAdvancedDataSourceEvent)
					myEvent = new LyAdvancedDataSourceEvent(LyExtendObjectSource.this,
							event.getType(), event.getObject(),
							((LyAdvancedDataSourceEvent)event).getOldState());
				else
					myEvent = new LyDataSourceEvent(LyExtendObjectSource.this, event.getType(),
							event.getObject(), event.getValues());
				myEvent.setLoading(event.isLoading());
				// FFT01455: Propagate notified status
				myEvent.setNotified(event.isNotified());
				myEvent.setLocal(event.isLocal());
				myEvents.add(myEvent);
			}

			//      Call postEvent on the event manager
			LyEventManager em = _application.getEventManager();

			em.postEvents(myEvents);

			return true;
		} // end method dataSourceEventsGrouped
	} // end class _LyAdvancedDataSourceListener

	/**
     * Marks the associated cache as loaded or not.
     *
     * @param loaded True to mark the cache loaded, false otherwise.
     */
    public void setLoaded(boolean loaded)
    {
    	setLoaded(loaded, false);
    }
  /**
     * Marks the associated cache as loaded or not.
     *
     * @param loaded True to mark the cache loaded, false otherwise.
     */
    public void setLoaded(boolean loaded, boolean subClassesOnly)
    {
        LyInfoList<LyClassInfo> subClasses = ((LyClassInfo)_class).getSubClasses();
        int                     i, n = ((subClasses == null) ? 0 : subClasses.getSize());
        LyClassInfo             subClass;
        LyObjectSource          objectSource;

        if ((!subClassesOnly) && (_objectSource != null))
            _objectSource.setLoaded(loaded);

        for (i = 0; i < n; i++)
        {
            subClass = subClasses.getInfo(i);
            objectSource = getObjectSource(subClass);
            objectSource.setLoaded(loaded);
        }
    }
    
	/**
	 * Finalize method for an object, remove from cache
	 * @param object LyObject being finalized
	 */
	public void finalizeObject(LyObject object)
	{
		if (getObjectSource() != null)
			getObjectSource().finalizeObject(object);
	}

} // end class LyExtendObjectSource
