/*
 * LyFieldInfoValue.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import leon.info.*;


/**
 * This class is a value wrapper, it offers methods for constructing, accessing and modifying values
 * corresponding to a fieldInfo.
 *
 * <p>It offers methods that hides the internal representation of the different types of fieldInfos
 * for consulting, creating and modifying logical objects (LyObject).</p>
 *
 * <p>For existing objects, these "wrapped values" can be got by using getFieldValue method of
 * LyObject.</p>
 *
 * <p>This class is abstract. A class, adapted to each sub-type of fieldInfo, is provided (see
 * inherited classes).</p>
 *
 * @see LyFieldInfo
 * @see LyObject
 * @see LyObject#getFieldValue
 */
public class LyFieldInfoValue extends LyValue
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyFieldInfoValue.java 33634 2011-12-19 15:51:13Z bangiole $";

	/**
	 * Constructs an empty value for the given fieldInfo.
	 *
	 * @param fieldInfo FieldInfo for which a value is created.
	 */
	public LyFieldInfoValue(LyFieldInfo fieldInfo)
	{
		super(fieldInfo, null);
	}

	/**
	 * Constructs a value for the given fieldInfo from an untyped object.
	 *
	 * @param fieldInfo FieldInfo for which a value is created.
	 * @param value     Value given as an untyped object.
	 */
	public LyFieldInfoValue(LyFieldInfo fieldInfo, Object value)
	{
		super(fieldInfo, value);
		
		if ((value != null)
		        && (value.getClass() != fieldInfo.getValueClass()))
		    throw new IllegalArgumentException("BF AppEngine : value " + value + " of class "
		        + value.getClass() + ", is not valid for fieldinfo " + fieldInfo
		        + " which requires values of class " + fieldInfo.getValueClass());
	}

	/**
	 * Constructs a value for the given fieldInfo from a string. The precise content of this string
	 * depends on each type of fieldInfo (see similar constructors in inherited classes).
	 *
	 * @param fieldInfo FieldInfo for which a value is created.
	 * @param value     Value given as a string. This method uses the parse method of LyFieldInfo.
	 * @see   LyFieldInfo#parse
	 */
	public LyFieldInfoValue(LyFieldInfo fieldInfo, String value)
	{
		this(fieldInfo, fieldInfo.parse(value));
	}

	/**
	 * Gets the fieldInfo to which this value corresponds. Result of LyValue.getField casted as to
	 * LyFieldInfo.
	 *
	 * @return FieldInfo of this value.
	 */
	public LyFieldInfo getFieldInfo()
	{
		return (LyFieldInfo)getField();
	}

    /**
     * Returns a "displayable string" (string that can be shown to the end-user) for this value.
     *
     * @return The value presented as a displayable String. This method uses the toName method of
     *         LyFieldInfo. The result depends on each type of fieldInfo.
     * @see    LyFieldInfo#toName
     */
    public final String getName()
    {
        return getName(false);
    }
    
    /**
     * Returns a "displayable string" (string that can be shown to the end-user) for this value.
     *
     * @param  removeTags Remove rich-text tags.
     * @return The value presented as a displayable String. This method uses the toName method of
     *         LyFieldInfo. The result depends on each type of fieldInfo.
     * @see    LyFieldInfo#toName
     */
    public String getName(boolean removeTags)
    {
        return getFieldInfo().toName(getValue(), removeTags);
    }

	/**
	 * Returns this value as a short "displayable string" (short string that can be shown to the
	 * end-user).
	 *
	 * @return The value presented as a short displayable String. This method uses the toShortName
	 *         method of LyFieldInfo. The result depends on each type of fieldInfo.
	 * @see    LyFieldInfo#toShortName
	 */
	public String getShortName()
	{
		return getFieldInfo().toShortName(getValue());
	}

	/**
	 * Converts this value to an "internal" string (not shown to end-user, but that enables
	 * reconstruction through LyFieldInfo.parse or using the constructor with value passed as
	 * string). This string may not be significant to the end-user and usually should not be
	 * displayed.
	 *
	 * @return An internal String corresponding to this value. This method uses the toString method
	 *         of LyFieldInfo. The result depends on each type of fieldInfo.
	 * @see    LyFieldInfo#toString
	 * @see    LyFieldInfo#parse
	 */
	public String getString()
	{
		return getFieldInfo().toString(getValue());
	}

	/**
	 * Compares this value to the given value.
	 *
	 * @param  value Value to which this value is compared.
	 * @return 0 If the two values are identical, a positive number if this value1 is "greater" than
	 *         the given value. A negative number is returned, otherwise.
	 * @see    LyFieldInfo#compare
	 */
	public final int compare(LyFieldInfoValue value)
	{
		return getFieldInfo().compare(getValue(), ((value == null) ? null : value.getValue()));
	}

	/**
	 * Compares this value to the given value for equality.
	 *
	 * @param  value Value to which this value is compared.
	 * @return True if the values are equals, else false.
	 * @see    LyFieldInfo#compare
	 */
	public final boolean equals(LyFieldInfoValue value)
	{
		return getFieldInfo().equals(getValue(), ((value == null) ? null : value.getValue()));
	}

	/**
	 * Compares this value to the given value for equality.
	 *
	 * @param  value Value to which this value is compared.
	 * @return True if the values are equals, else false.
	 * @see    LyFieldInfo#compare
	 */
	public boolean equals(Object value)
	{
		if (value instanceof LyFieldInfoValue)
		{
			if (getFieldInfo() != ((LyFieldInfoValue)value).getFieldInfo())
				return false;
			return equals((LyFieldInfoValue)value);
		}
		else
			return super.equals(value);
	}

	/**
	 * Checks if the given value is contained in this value.
	 *
	 * @param  value Value to which this value is compared.
	 * @return True if the second value is contained in the this value, else returns false.
	 */
	public final boolean contains(LyFieldInfoValue value)
	{
		return getFieldInfo().contains(getValue(), ((value == null) ? null : value.getValue()));
	}
} // end class LyFieldInfoValue
