/*
 * LyFileValue.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import leon.info.*;
import leon.peer.*;
import leon.peer.union.*;
import leon.peer.util.*;

import java.io.*;


/**
 * This class offers methods for constructing, accessing and modifying values corresponding to a
 * LyFileFieldInfo.
 *
 * @see LyFileFieldInfo
 * @see LyObject
 */
public final class LyFileValue extends LyFieldInfoValue
{

    /** File version */
    public static final String __VERSION =
        "$Id: LyFileValue.java 35842 2012-06-12 16:11:25Z jlsaouli $";

    /** Peer locator to the file (value depends on the data provider implementation). */
    protected String _locator;

    /**
     * Constructs an empty file value for the given file field.
     *
     * @param fileFieldInfo File fieldInfo for which a value is created.
     */
    public LyFileValue(LyFileFieldInfo fileFieldInfo)
    {
        super(fileFieldInfo);
    }

    /**
     * Constructs a file value for the given file field from a string.
     *
     * @param fileFieldInfo File fieldInfo for which a value is created.
     * @param value         String value as expected by parse method.
     * @see   LyFileFieldInfo#parse
     */
    public LyFileValue(LyFileFieldInfo fileFieldInfo, String value)
    {
        super(fileFieldInfo, value);
    }

    /**
     * Internal method to check if the specified value is associated with a content Used only when
     * attachment handler is 'blob' for backward compatibility.
     */
    protected void checkForContent()
    {
        LyFileFieldInfo fileFieldInfo = (LyFileFieldInfo) getFieldInfo();

        // When we have an attachment handler : should use it to fetch content
        if (("blob".equals(fileFieldInfo.getAttachmentHandler())) && (getValue() != null)
                && (getString().length() > 0))
        {

            try
            {
                setInputStream(new FileInputStream(getString()));
            }
            catch (FileNotFoundException e)
            {
                e.printStackTrace();
            }
        }
    }

    /**
     * Faftory method to create an instance of a peer handler for attachment.
     *
     * @see leon.peer.util.LyBlobFileHandler
     */
    public LyBlobFileHandler handlerFactory()
    {
        return new LyBlobFileHandler();
    }

    /**
     * Gets the InputStream to the underlying value which is a 'File' The attachment handler MUST be
     * "blob", additional FILE fields types may be supported later. The application may be able to
     * read the content of the file. This inputStream is generated from the peer value (by the
     * datasource)
     *
     * @return the inputstream to read the file or null if it could not be found.
     */
    public InputStream getInputStream()
    {

        if (_locator != null)
        {
            LyBlobFileHandler handler = handlerFactory();

            return handler.getInputStream(_locator);
        }

        return null;
    }

    /**
     * Sets the the content of the value which is a 'file'. The content should be read from the
     * specified InputStream, and is stored in a file or in a blob or in a specific storage
     * depending on the data provider. When the mthod exits, the inputStream is no longer used and
     * may be closed. The attachment handler MUST be "blob", additional FILE fields types may be
     * supported later.
     *
     * @param iStream the inputstream used by the data provider to read the file.
     */
    public void setInputStream(InputStream iStream)
    {
        _locator = null;

        if (iStream == null)
            return;

        LyBlobFileHandler handler = handlerFactory();
        String            dir = ((LyFileFieldInfo) getFieldInfo()).getAttachmentDirectory();

        _locator = handler.setInputStream(dir, getFieldInfo().getId(), iStream);
    }

    /**
     * Gets the locator of the file when attachment handler is 'blob'. This is for internal use only
     * since the reference to this locator may change depending on the data provider and File types
     * : use getInputStream and setInputStream instead. This is a mean to fetch the file from the
     * data provider
     *
     * @see    getInputStream()
     * @see    setInputStream()
     * @return the locator (reference to the file or content in the data provider)
     */
    public String getLocator()
    {

        if (_locator == null)
            checkForContent();

        return _locator;
    }

    /**
     * Fetch attachment from the peer data provider. This methods actually download or get from the
     * blob the target attachment.
     *
     * @return true if attachment was found, false otherwise
     */
    public boolean fetchAttachment(LyObject object)
    {
        // Searching for a non local value.
        Object peer = object.getPeer();

        if (peer == null)
            return false;

        LyBlobFileHandler handler = handlerFactory();
        LyField           field = object.getPeerField(getFieldInfo());
        String            dir = ((LyFileFieldInfo) getFieldInfo()).getAttachmentDirectory();

        _locator = handler.fetchAttachment(object, field, dir, getFieldInfo().getId(),
                (String) getValue());

        return (_locator != null);
    }

    /**
     * Fetch blob content from the peer data provider. This methods actually download or get from the
     * blob the peer datasource.
     *
     * @return true if attachment was found, false otherwise
     */
    public boolean getBlobContent(LyObject object)
    {
        if (object instanceof LyBlobRequestInterface)
        {
            _locator = ((LyBlobRequestInterface) object).getBlobContent(object, getFieldInfo(),
                    (String) getValue());
            return (_locator != null);
        }

        // Searching for a non local value.
        Object peer = object.getPeer();

        if (peer == null)
            return false;

        if (peer instanceof LyUnionObject)
        {
        	LyUnionObject	unionObject = (LyUnionObject)peer;
        	
        	peer = unionObject.getObject1();
        	if (!(peer instanceof LyRowObject))
        		peer = unionObject.getObject2();
        }
        	
        if (!(peer instanceof LyRowObject))
            throw new IllegalArgumentException("[LyFileValue] getBlobContent is only implemented on LyRowObject yet.");

        final LyDataSource dataSource = ((LyRowObject) peer).getTableClass().getDataSource();

        if (dataSource == null)
            return false;

        final LyDataProvider   provider = dataSource.getDataProvider();
        final LyRequestEncoder encoder = provider.getRequestEncoder(dataSource);

        if (encoder instanceof LyBlobRequestEncoder)
            _locator = ((LyBlobRequestEncoder) encoder).getBlobContent(object, getFieldInfo(),
                    (String) getValue());

        return (_locator != null);
    }
} // end class LyFileValue
