/*
 * LyFilterRelationListener.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import leon.info.*;


/**
 * This class is the listener for receiving events on objects matched by a filter relation.
 *
 * @see LyDataSource
 * @see LyDataSourceEvent
 */
public class LyFilterRelationListener
	implements LyDataSourceListener
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyFilterRelationListener.java 29141 2011-01-12 11:22:10Z jlsaouli $";

	/** Client workspace. */
	LyWorkSpace				   _workSpace;

	/** Parent listener. */
	LyFilterRelationListener   _parent;

	/** Class that contains the relation concerned by the filter relation. */
	LyClassInfo				   _classInfo;

	/** Filter applied on objects. */
	LyFilter				   _filter;

	/** Filter relation managed by this listener. */
	LyFilterRelation		   _filterRelation;

	/**
	 * Constructor.
	 */
	public LyFilterRelationListener(LyWorkSpace workSpace, LyFilterRelationListener parent,
		LyClassInfo classInfo, LyFilter filter, LyFilterRelation filterRelation)
	{
		_workSpace = workSpace;
		_parent = parent;
		_classInfo = classInfo;
		_filter = filter;
		_filterRelation = filterRelation;

		// Listen on all possible classes
		LyInfoList<LyClassInfo> relationClasses = filterRelation.getPossibleClasses(false);
		int					    i, n = relationClasses.getSize();
		LyClassInfo			    relationClassInfo;
		LyObjectSource		    objectSource;

		for (i = 0; i < n; i++)
		{
			relationClassInfo = relationClasses.getInfo(i);
			objectSource = _workSpace.getObjectSource(relationClassInfo);
			objectSource.addListener(this);
		}
	}

	/**
	 * Free this listener.
	 */
	void free()
	{
		LyInfoList<LyClassInfo> classes = _filterRelation.getPossibleClasses(false);
		int					    i, n = classes.getSize();
		LyClassInfo			    classInfo;
		LyObjectSource		    objectSource;

		for (i = 0; i < n; i++)
		{
			classInfo = classes.getInfo(i);
			objectSource = _workSpace.getObjectSource(classInfo);
			objectSource.removeListener(this);
		}
	}

	/**
	 * Method for receiving an object creation.
	 *
	 * @param  event Event notifying the creation of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectCreated(LyDataSourceEvent event)
	{
		// Nothing to do
		return false;
	}

	/**
	 * Method for receiving an object modification.
	 *
	 * @param  event Event notifying the modification of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectModified(LyDataSourceEvent event)
	{
		LyValueSet values = event.getValues();
		LyFilter   filter = _filterRelation.getFilter();

		if ((values != null)
				&& !filter.isImpacted(values, false))
			return false;

		LyObject object = (LyObject)event.getObject();
		boolean  add = filter.match(object);

		checkSources(object, add);
		return true;
	}

	/**
	 * Method for receiving an object deletion.
	 *
	 * @param  event Event notifying the deletion of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectDeleted(LyDataSourceEvent event)
	{
		// Check if the object match the filter
		LyObject object = (LyObject)event.getObject();
		LyFilter filter = _filterRelation.getFilter();

		if (!filter.match(object))
			return false;

		checkSources(object, false);
		return true;
	}

	/**
	 * Check objects that point to the given object through the relation contained in the filter
	 * relation.
	 */
	protected void checkSources(LyObject target, boolean add)
	{
		LyObjectList sources = getImpactedObjects(target, add);

		if ((sources == null)
				|| (sources.getSize() == 0))
			return;

		if (_parent != null)
			_parent.checkFilter(sources, add);
		else
			_workSpace.checkFilter(sources, add);
	}

	/**
	 * Get impacted objects of parent work space.
	 */
	protected LyObjectList getImpactedObjects(LyObject object, boolean match)
	{
		// Find objects of the source classInfo that reference
		// the given object through the relation
		LyRelationFieldInfo          relation = _filterRelation.getRelationFieldInfo();
		LyFilter					 relFilter = new LyFilter();
		LyFilterExpression.Condition condition = (relation.isMultiple()
				? LyFilterExpression.Condition.CONTAINS : LyFilterExpression.Condition.EQUALS);
		LyFilterExpression			 filterExp;

		filterExp = new LyFilterExpression(relation, false, condition, object.getObjectId());
		filterExp.setModifier(LyFilterExpression.MOD_CASE_SENSITIVE);

		relFilter.addElement(filterExp);

		LyWorkSpace		   list = new LyClassWorkSpace(_workSpace.getSession(), _classInfo,
				new LyInfoList<LyFieldInfo>(), relFilter, null, false, false, false);
		int				   i, n = list.getSize();
		LyObject		   source;
		LySimpleObjectList objects = null;

		for (i = 0; i < n; i++)
		{
			source = list.getObject(i);
			if ((match && _filter.match(source))
					|| (!match && !_filter.match(source)))
			{
				if (objects == null)
					objects = new LySimpleObjectList();
				objects.addObject(source);
			}
		}

		return objects;
	} // end method getImpactedObjects

	/**
	 * Check objects.
	 */
	protected void checkFilter(LyObjectList objects, boolean add)
	{
		LyFilter filter = _filterRelation.getFilter();
		int		 i, n = objects.getSize();
		LyObject object;

		for (i = 0; i < n; i++)
		{
			object = objects.getObject(i);
			if (add && filter.match(object))
				checkSources(object, true);
			else if (!add && !filter.match(object))
				checkSources(object, false);
		}
	}
} // end class LyFilterRelationListener
