/*
 * LyFilteredObjectSource.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;

import leon.app.*;
import leon.info.*;
import leon.info.binding.*;


/**
 * This class implements an object source with filtered datas. It allows to have a partial view of
 * an object source by filtering its datas. For example, you can have only equipments of a network.
 *
 * @see LyObjectSource
 * @see LyFilter
 */
public final class LyFilteredObjectSource extends LyObjectSource
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyFilteredObjectSource.java 40675 2012-09-05 16:26:11Z jlsaouli $";

	/** The filter applied to the data. */
	protected LyFilter		   _filter;

	/**
	 * Constructs a filtered object source.
	 *
	 * @param session   Session owner of this object source.
	 * @param classInfo Class info associated with the object source.
	 * @param filter    Filter critera of this object source.
	 */
	public LyFilteredObjectSource(LyDataSession session, LyClassInfo classInfo,
		LyFilter filter)
	{
		// Initialize class attributes.
		_application = session.getApplication();
		_class = classInfo;

		// No cache at all
		_cache = null;

		// No datasource.
		_dataSource = null;

		_listener = new _LyAdvancedDataSourceListener();
		getObjectSource().addListener(_listener);
		_filter = filter;

		// Adding this object source to the session.
		session.putObjectSource(this);
	}

	/**
	 * Retrieves the class info of this object source.
	 *
	 * @return Class info of objects within this object source.
	 */
	public LyClassInfo getClassInfo()
	{
		return (LyClassInfo)_class;
	}

	/**
	 * Retrieves the object source (not filtered).
	 *
	 * @return the underlying object source.
	 */
	public LyObjectSource getObjectSource()
	{
		return getClassInfo().getObjectSource();
	}

	/**
	 * Retrieves the filter of this object source.
	 *
	 * @return Filter used for objects within this object source.
	 */
	public LyFilter getFilter()
	{
		return _filter;
	}

	/**
	 * Associates this object source to a data source representing the physical layer. This method
	 * throws an exception because a filtered object source can't listen to a data source. The
	 * events are treated by the instance of object source from where the datas are filtered.
	 *
	 * @param     dataSource Data source to which this object source is associated.
	 * @exception IllegalStateException If the given data source is not null.
	 */
	public void listenTo(LyDataSource dataSource)
	{
		if (dataSource == null)
			return;

		throw new IllegalStateException(
			"BF AppEngine : a LyFilteredObjectSource can't listen to a LyDataSource");
	}

	/**
	 * Sets the cache of this object source. For a filtered object source, this method always throws
	 * an exception : there is no cache for a filtered object source. The objects are managed by the
	 * instance of LyObjectSource from where the datas are filtered.
	 *
	 * @param     cache Cache to associate to this object source.
	 * @exception IllegalStateException If the given cache is not null.
	 */
	public void setCache(LyCache cache)
	{
		if (cache == null)
			return;

		throw new IllegalStateException("BF AppEngine : Can't set a cache to a LyFilteredObjectSource");
	}

	/**
	 * Free this object source. After calling this method, you must not attempt to access data with
	 * this object source.
	 */
	public void free()
	{
		getObjectSource().removeListener(_listener);
		_listener = null;
		super.free();
	}

	/**
	 * Internal method to merge the given filter with the filter of this object source.
	 *
	 * @param  filter Filter to merge.
	 * @return Filter result of the merge of the two filters with a AND condition.
	 */
	LyFilter getFilter(LyFilter filter)
	{
		if (filter == null)
			return _filter;

		LyFilter andFilter = new LyFilter(LyFilter.Type.AND);

		andFilter.addElement(_filter);
		andFilter.addElement(filter);

		return andFilter;
	}

	/**
	 * Loads completely this object source. For a filtering object source, this method does nothing
	 * because there is no associated cache.
	 */
	public void load()
	{
	}

	/**
	 * Gets an object from this object source by the specified identifier. The object is returned
	 * only if it matches the filter of this object source.
	 *
	 * @param  id Object identifier of the searched object.
	 * @return The requested object if it was found, or null if it could not be found or if the
	 *         found object doesn't match the filter of this object source.
	 */
	public Object getObject(LyObjectId id)
	{
		return getObject(id, false);
	}

	/**
	 * Gets an object from this object source by the specified identifier. The object is returned
	 * only if it matches the filter of this object source.
	 *
	 * @param  id    Object identifier of the searched object.
	 * @param  local Indicates if the request must be local or not.
	 * @return The requested object if it was found, or null if it could not be found or if the
	 *         found object doesn't match the filter of this object source.
	 */
	public Object getObject(LyObjectId id, boolean local)
	{
		LyObject object = (LyObject)getObjectSource().getObject(id, local);

		if ((object == null) || !_filter.match(object))
			return null;

		return object;
	}

	/**
	 * Return the number of objects within this object source matching the given filter. The
	 * specified filter is merged with the filter of this object source.
	 *
	 * @param  filter Filter applied on objects for counting.
	 * @param  local  Indicates if the request must be local or not.
	 * @return Number of objects matching the given filter in this object source.
	 */
	public int getSize(LyFilter filter, boolean local)
	{
		return getObjectSource().getSize(getFilter(filter), local);
	}

	/**
	 * Create a new object in this object source. Redirects creation on the object source on which
	 * this object source is plugged.
	 *
	 * @param     object   Object to create.
	 * @param     listener Listener of the response.
	 * @param     local    Indicates if the creation is local or not.
	 * @param     refresh  Indicates if the creation may be notified or not.
	 * @return    Response to the creation.
	 * @exception IllegalStateException If the object source is in read only mode.
	 */
	public LyResponse createObject(LyObject object, LyResponseListener listener,
		boolean local, boolean refresh)
	{
		return getObjectSource().createObject(object, listener, local, refresh);
	}

	/**
	 * Set an object of this object source. Redirects modification on the object source on which
	 * this object source is plugged.
	 *
	 * @param     object   Object to modify.
	 * @param     listener Listener of the response.
	 * @param     local    Indicates if the modification must be local or not.
	 * @param     refresh  Indicates if the modification may be notified or not.
	 * @return    Response to the modification.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If the given object is invalid.
	 */
	public LyResponse setObject(LyObject object, LyResponseListener listener,
		boolean local, boolean refresh)
	{
		return getObjectSource().setObject(object, listener, local, refresh);
	}

	/**
	 * Set a list of objects of this object source. Redirects modifications on the object source on
	 * which this object source is plugged.
	 *
	 * @param     object   List of objects to modify.
	 * @param     listener Listener of the response.
	 * @return    Response to the modification.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If one of the given objects is invalid.
	 */
	public LyResponse setObjects(Vector<? extends Object> objects, LyResponseListener listener)
	{
		return getObjectSource().setObjects(objects, listener);
	}

	/**
	 * Delete an object of this object source. Redirects deletion on the object source on which this
	 * object source is plugged.
	 *
	 * @param     object            Object to delete.
	 * @param     listener          Listener of the response.
	 * @param     local             Indicates if the operation must be local or not.
	 * @param     refresh           Indicates if the operaton may be notified or not.
	 * @param     deleteLinkObjects Indicates if linked objects (composition relations, ...)
	 *                              must be deleted or not.
	 * @return    Response to the deletion.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If the given object is invalid.
	 */
	public LyResponse deleteObject(LyObject object, LyResponseListener listener,
		boolean local, boolean refresh, boolean deleteLinkedObjects)
	{
		return getObjectSource().deleteObject(object, listener, local, refresh,
			deleteLinkedObjects);
	}

	/**
	 * Delete a list of objects of this object source. Redirects deletions on the object source on
	 * which this object source is plugged.
	 *
	 * @param     objects           List of objects to delete.
	 * @param     listener          Listener of the response.
	 * @param     local             Indicates if the operation must be local or not.
	 * @param     refresh           Indicates if the operaton may be notified or not.
	 * @param     deleteLinkObjects Indicates if linked objects (composition relations, ...)
	 *                              must be deleted or not.
	 * @return    Response to the deletion.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If one of the given objects is invalid.
	 */
	public LyResponse deleteObjects(Vector<? extends Object> objects, LyResponseListener listener,
		boolean local, boolean refresh, boolean deleteLinkedObjects)
	{
		return getObjectSource().deleteObjects(objects, listener, local, refresh,
			deleteLinkedObjects);
	}

	/**
	 * Method to load an object by its id from this object source. Redirects loading on the object
	 * source on which this object source is plugged.
	 *
	 * @param  id       Object identifier corresponding to the object to load.
	 * @param  listener Listener of the response.
	 * @return Response to the loading.
	 */
	public LyResponse loadObject(LyObjectId id, Vector<? extends LyField> fields,
		LyResponseListener listener)
	{
		return getObjectSource().loadObject(id, fields, listener);
	}

	/**
	 * Method for loading objects on this object source. Merges the given filter with the filter of
	 * this object source and redirects loadings on the object source on which this object source is
	 * plugged.
	 *
	 * @param  sort     Gives sort criteria, or null if there is no sort (the order is the one of
	 *                  the physical layer).
	 * @param  filter   Gives filter criteria, or null if there is no filter (all objects).
	 * @param  n        Indicates the number of objects to load (-1 means no limit).
	 * @param  listener Listener of the response.
	 * @return Response to the loading.
	 */
	public LyResponse loadObjects(Vector<? extends LyField> fieldInfos, LySort sort,
		LyFilter filter, int n,
		LyResponseListener listener)
	{
		return getObjectSource().loadObjects(fieldInfos, sort, getFilter(filter), n, listener);
	}

	/**
	 * Retrieve objects on this object source. Retrieve a specified number of objects sorted by a
	 * given sort and that match a given filter. Merges the given filter with the filter of this
	 * object source and redirects searchs on the object source on which this object source is
	 * plugged.
	 *
	 * @param  sort     Gives sort criteria, or null if there is no sort (the order is the one of
	 *                  the physical layer).
	 * @param  filter   Gives filter criteria, or null if there is no filter (all objects).
	 * @param  n        Indicates the number of objects to retrieve (-1 means no limit).
	 * @param  local    Indicates if the request is local (from memory) or not.
	 * @param  listener Listener of the response.
	 * @return Response to this search.
	 */
	public LyResponse getObjects(Vector<LyFieldInfo> fieldInfos, LySort sort, LyFilter filter,
		int n,
		boolean local, LyResponseListener listener)
	{
		return getObjectSource().getObjects(fieldInfos, sort, getFilter(filter), n, local,
			listener);
	}

	/**
	 * Tests if an object corresponding to a given object id is found in the cache of this data
	 * source.
	 *
	 * @param  objectId Object id for which an object is searched.
	 * @return True if an object with the given object id is found in the cache, else false.
	 */
	public boolean checkObject(LyObjectId objectId)
	{
		return getObjectSource().checkObject(objectId);
	}

	/**
	 * Internal method to check validy of an object given as parameter.
	 *
	 * @param  object Checked object.
	 * @return True if the given object is found in the cache, else false.
	 */
	public boolean checkObject(Object object)
	{
		return getObjectSource().checkObject(object);
	}

	/**
	 * Internal method to check validy of an object given as parameter.
	 *
	 * @param     object Checked object.
	 * @exception IllegalArgumentException If one of the above checks fails.
	 */
	protected void checkObject(Object object, boolean inCache)
	{
		getObjectSource().checkObject(object, inCache);
	}

	/* Implementation of LyDataSourceListener */

	final class _LyAdvancedDataSourceListener
		implements LyAdvancedDataSourceListener
	{

		/**
		 * Method for receiving an object creation. Checks whether the created object matchs the
		 * filter of this object source and if so, notify listeners of this object source.
		 *
		 * @param  event                 Event notifying the creation of an object.
		 * @return Boolean indicating if the event was used, false otherwise.
		 * @param  IllegalStateException If the received event was not thrown by the associated
		 *                               object source.
		 */
		public boolean objectCreated(LyDataSourceEvent event)
		{
			LyDataSourceEvent myEvent = testEvent(event);

			if (myEvent != null)
				return postEvent(myEvent);
			return false;
		}

		/**
		 * Method for receiving an object deletion. Checks whether the deleted object matchs the
		 * filter of this object source and if so, notify listeners of this object source.
		 *
		 * @param  event                 Event notifying the deletion of an object.
		 * @return Boolean indicating if the event was used, false otherwise.
		 * @param  IllegalStateException If the received event was not thrown by the associated
		 *                               object source.
		 */
		public boolean objectDeleted(LyDataSourceEvent event)
		{
			LyDataSourceEvent myEvent = testEvent(event);

			if (myEvent != null)
				return postEvent(myEvent);
			return false;
		}

		/**
		 * Method for receiving an object modification. Notifies a creation, modification or a
		 * deletion or ignore the event depending on the fact the object matchs or not the filter of
		 * the object source and matched it previously or not.
		 *
		 * @param  event                 Event notifying the modification of an object.
		 * @return Boolean indicating if the event was used, false otherwise.
		 * @param  IllegalStateException If the received event was not thrown by the associated
		 *                               object source.
		 */
		public boolean objectModified(LyDataSourceEvent event)
		{
			LyDataSourceEvent myEvent = testEvent(event);

			if (myEvent != null)
				return postEvent(myEvent);
			return false;
		}

		/**
		 * Method for receiving state changes on an object.
		 *
		 * @param  event Event notifying the state change of an object.
		 * @return True indicating if the event was used, false otherwise.
		 */
		public boolean objectStateChanged(LyAdvancedDataSourceEvent event)
		{
			return rePost(event);
		}

		/**
		 * Method for receiving user event on an object.
		 *
		 * @param  event Event notifying the user event of an object.
		 * @return True indicating if the event was used, false otherwise.
		 */
		public boolean objectUserEvent(LyAdvancedDataSourceEvent event)
		{
			return rePost(event);
		}

		/**
		 * Called when grouped events are sent.
		 *
		 * @param  events Sorted list of events.
		 * @return True if the events are used by the listener, false otherwise.
		 */
		public boolean dataSourceEventsGrouped(ArrayList<LyDataSourceEvent> events)
		{
			if ((events == null) || (events.isEmpty()))
				return false;

			int							 i, n = events.size();
			LyDataSourceEvent			 event, myEvent;
			ArrayList<LyDataSourceEvent> myEvents = new ArrayList<LyDataSourceEvent>();

			for (i = 0; i < n; i++)
			{
				event = (LyDataSourceEvent)events.get(i);
				myEvent = testEvent(event);
				if (myEvent != null)
					myEvents.add(myEvent);
			}

			if (myEvents.isEmpty())
				return false;
			else
			{
				// Call postEvent on the event manager
				LyEventManager em = _application.getEventManager();

				em.postEvents(myEvents);
				return true;
			}
		} // end method dataSourceEventsGrouped

		LyDataSourceEvent testEvent(LyDataSourceEvent event)
		{
			if (event.getSource() != getObjectSource())
				throw new IllegalStateException(
					"BF AppEngine : receives an event fron an unknown objectSource");

			short			  type = event.getType();
			LyObject		  object = (LyObject)event.getObject();
			LyDataSourceEvent myEvent = null;

			if (event instanceof LyAdvancedDataSourceEvent)
			{
				if (_filter.match(object))
				{
					myEvent = new LyAdvancedDataSourceEvent(LyFilteredObjectSource.this,
							event.getType(), object,
							((LyAdvancedDataSourceEvent)event).getOldState());
					myEvent.setNotified(event.isNotified());
					myEvent.setLocal(event.isLocal());
				}
			}
			else if (type == LyDataSourceEvent.TYPE_SET)
			{
			    if (event.isLocal()
			            && !getClassInfo().isLocal())
			        // Event comes from a rehash, object not yet created
			        return null;
			    
				// We must verify if the object changed will be a deletion, a creation
				// or a modification for this filtered object source.
				LyValueSet			    oldValues = event.getValues();
				LyValue				    value;
				int					    i, n;
				boolean				    wasIn, isIn;

				// Optimization: if no field of _filter is in oldValues, no call object.getValues()
				LyInfoList<LyFieldInfo> fieldsFiltered = _filter.getFieldInfos(false);
				LyValueSet			    oldFilteredValues, newFilteredValues;
				LyFieldInfo			    fieldInfo;

				oldFilteredValues = new LyValueSet();
				newFilteredValues = new LyValueSet();
				n = ((fieldsFiltered == null) ? 0 : fieldsFiltered.getSize());

				for (i = 0; i < n; i++)
				{
					fieldInfo = fieldsFiltered.getInfo(i);
					if ((oldValues != null) && (oldValues.contains(fieldInfo)))
						value = oldValues.getFieldValue(fieldInfo);
					else
						value = object.getFieldValue(fieldInfo);
					oldFilteredValues.addValue(value);

					value = object.getFieldValue(fieldInfo);
					newFilteredValues.addValue(value);
				}

				wasIn = _filter.match(oldFilteredValues);
				isIn = _filter.match(newFilteredValues);

				if (wasIn && isIn)
					// The object was in the filtered object source and is still in.
					myEvent = new LyDataSourceEvent(LyFilteredObjectSource.this,
							LyDataSourceEvent.TYPE_SET, object, event.getValues());
				else if (wasIn)
					// The object was in the filtered object source and is not in anymore.
					myEvent = new LyDataSourceEvent(LyFilteredObjectSource.this,
							LyDataSourceEvent.TYPE_DELETE, object, null);
				else if (isIn)
					// The object was not in the filtered object source and is now in.
					myEvent = new LyDataSourceEvent(LyFilteredObjectSource.this,
							LyDataSourceEvent.TYPE_CREATE, object, null);
				else
					// The object was not in the filtered object source and is still not in
					// so we ignore the event.
					return null;

				myEvent.setNotified(event.isNotified());
				myEvent.setLocal(event.isLocal());
			}
			else if (type == LyDataSourceEvent.TYPE_DELETE)
			{
				if (_filter.match(object))
				{
					// Notify listeners.
					myEvent = new LyDataSourceEvent(LyFilteredObjectSource.this,
							LyDataSourceEvent.TYPE_DELETE, object, null);
					myEvent.setNotified(event.isNotified());
					myEvent.setLocal(event.isLocal());
				}
			}
			else if (type == LyDataSourceEvent.TYPE_CREATE)
			{
				if (_filter.match(object))
				{
					// Notify listeners.
					myEvent = new LyDataSourceEvent(LyFilteredObjectSource.this,
							LyDataSourceEvent.TYPE_CREATE, object, null);

					myEvent.setLoading(event.isLoading());
					myEvent.setNotified(event.isNotified());
					myEvent.setLocal(event.isLocal());
				}
			}

			return myEvent;
		} // end method testEvent

		protected boolean rePost(LyAdvancedDataSourceEvent event)
		{
			LyObject object = (LyObject)event.getObject();

			if (_filter.match(object))
			{
				// Notify listeners.
				LyAdvancedDataSourceEvent myEvent;

				myEvent = new LyAdvancedDataSourceEvent(LyFilteredObjectSource.this,
						event.getType(), object, event.getOldState());
				myEvent.setNotified(event.isNotified());
				myEvent.setLocal(event.isLocal());
				return postEvent(myEvent);
			}

			return false;
		}
	} // end class _LyAdvancedDataSourceListener
	
	/**
     * Marks the associated cache as loaded or not.
     *
     * @param loaded True to mark the cache loaded, false otherwise.
     */
    protected void setLoaded(boolean loaded)
    {
        getObjectSource().setLoaded(loaded);
    }
    
    /**
     * Method to reload an object from its id on the associated data provider. Load fields of an
     * object from a given object id.
     *
     * @param  id           Object identifier corresponding to the object to reload.
     * @param  fields       Fields to load (null means all fields).
     * @param  listener     Listener of the response.
     * @param  classBinding binding between logical class and physical class This binding may
     *                      contain extra information required for encoding the request.
     * @return Response to the loading.
     */
    public LyResponse reloadObject(LyObjectId id, Vector<? extends LyField> fields,
        LyResponseListener listener, LyClassBinding classBinding)
    {
        return getObjectSource().reloadObject(id, fields, listener, classBinding);
    }
} // end class LyFilteredObjectSource
