/*
 * LyIntersectionWorkSpace.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import leon.app.*;


/**
 * This class implements an intersection workspace : workspace that is the intersection of other
 * workspaces.
 *
 * <p>Workspaces are added in a intersection-workspace usind addList method. The result workspace
 * can be dynamic only if all added workspaces are dynamic.</p>
 *
 * @see LyMultiWorkSpace
 */
public final class LyIntersectionWorkSpace extends LyMultiWorkSpace
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyIntersectionWorkSpace.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Constructs an empty intersection-workspace within a given session.
	 *
	 * @param session Session in which this workspace is constructed.
	 * @param dynamic Specifies if this workspace is dynamic or not.
	 */
	public LyIntersectionWorkSpace(LyDataSession session, boolean dynamic)
	{
		super(session, dynamic);
	}

	/**
	 * Adds an objects list to our intersection workspace.
	 *
	 * @param list                     Objects list that is added to our workspace.
	 * @throw IllegalStateException If this workspace is dynamic and the given list is not.
	 */
	public void addList(LyObjectList list)
	{
		if (isDynamic())
		{
			if (!(list instanceof LyWorkSpace) || !list.isDynamic())
				throw new IllegalStateException(
					"BF AppEngine : non dynamic objects lists can't be added to a dynamic intersection-workspace");

			// Add the given list to our objects lists
			_subObjectLists.addElement(list);

			// Listen to workspace events on this list
			((LyWorkSpace)list).addListener(_listener);
		}

		LyObjectList result = null;

		// For first list added, add all the objects
		if (_objects == null)
			result = list;
		// Else add only the intersection
		else
		{
			result = getIntersection(list);
			_objects = new _LyObjects();
		}

		if (result != null)
		{
			synchronized (result)
			{
				int i, n = result.getSize();

				for (i = 0; i < n; i++)
				{
					if (isDynamic())
						incrementCount(result.getObject(i));
					else
						addObject(result.getObject(i));
				}
			}
		}

		if ((result == null) || (result.getSize() == 0))
			return;

		_classes = result.getClasses();
	} // end method addList

	/* Management of the content of the list */

	/**
	 * Internal method to change some values of an object in this workspace. Overriden to check that
	 * the given object is contained in the current workspace.
	 *
	 * @param  object Object that is changed.
	 * @param  values Values changed
	 * @param info an event info
	 * @return True if the object position changed, false otherwise.
	 */
	protected synchronized boolean changeObject(LyObject object, LyValueSet values,
		LyEventInfo info)
	{
		if (indexOf(object) < 0)
			return false;

		return super.changeObject(object, values, info);
	}

	/**
	 * Removes an object from this workspace. Overriden to keep the counter of the given object.
	 *
	 * @param     object Object to remove from the workspace.
	 * @return    Index where the object was found in this workspace.
	 * @exception IllegalArgumentException If the object was not found in this workspace.
	 */
	protected int removeObject(LyObject object)
	{
		// If object is removed, its number of occurences may not be null.
		// Keep this number and set it after object deletion (count is removed
		// in the super method).
		int old_count = getCount(object);
		int index = super.removeObject(object);

		if (old_count > 1)
			_counters.put(object, new Integer(old_count - 1));

		return index;
	}

	/**
	 * Increments counter value for the given object. If the counter value equals the number of
	 * sub-workspaces minus one, the object is added in this workspace.
	 *
	 * @param object Object for which the counter value is incremented.
	 */
	void incrementCount(LyObject object)
	{
		synchronized (_counters)
		{
			int count = getCount(object);

			if (count == (getObjectListCount() - 1))
				addObject(object);

			_counters.put(object, new Integer(count + 1));
		}
	}

	/**
	 * Decrements counter value for the given object. If the counter value equals the number of
	 * sub-workspaces, the object is removed from this workspace.
	 *
	 * @param object Object for which the counter value is decremented.
	 */
	void decrementCount(LyObject object)
	{
		synchronized (_counters)
		{
			int count = getCount(object);

			if (count == getObjectListCount())
				removeObject(object);
			else
				_counters.put(object, new Integer(count - 1));
		}
	}
} // end class LyIntersectionWorkSpace
