/*
 * LyMetricsTools.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.data;

import java.util.*;
import leon.app.*;
import leon.info.*;
import leon.misc.*;


/**
 * Utility class to manage metrics.
 */
public class LyMetricsTools
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyXmlObjectTool.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** Possible values for metric types. */
	public enum MetricType
	{
		SESSION,
		ACTION,
		DATA_CREATION,
		PHYSICAL_REQUEST,
		EXCEPTION,
		APPLICATIVE;
	}

	/** Constant for metrics class id. */
	public static String METRIC_CLASS = "__leonMetric";

	/** Constant for metrics type field id. */
	public static String METRIC_TYPE = "__leonMetricType";

	/** Constant for metrics start date field id. */
	public static String METRIC_START_DATE = "__leonMetricStartDate";

	/** Constant for metrics start duration field id. */
	public static String METRIC_START_DURATION = "__leonMetricStartDuration";

	/** Constant for metrics end date field id. */
	public static String METRIC_END_DATE = "__leonMetricEndDate";

	/** Constant for metrics userId field id. */
	public static String METRIC_USER_ID = "__leonMetricUserId";

	/** Constant for metrics actionId field id. */
	public static String METRIC_ACTION_ID = "__leonMetricActionId";

	/** Constant for metrics classId field id. */
	public static String METRIC_CLASS_ID = "__leonMetricClassId";

	/** Constant for metrics value field id. */
	public static String METRIC_VALUE = "__leonMetricValue";


	/**
	 * Indicates whether metrics area activated for a given session.
	 * @param session current session.
	 * @return true if metrics are activated, false otherwise.
	 */
	public static boolean areMetricsActivated(LyDataSession session, MetricType type)
	{
		LyEnvironment	environment = (session != null ? session.getEnvironment() : null);

		if ((environment == null)
				|| !environment.testResource("LY_ACTIVATE_METRICS"))
			return false;

		LyApplication	application = session.getApplication();
		LyClassInfo		metricCls = application.getClassInfo(METRIC_CLASS);

		if (metricCls.getObjectSource() == null)
			return false;

		return true;
	}

	/**
	 * Creates a new metric and returns the LyObject instance.
	 */
	public static LyObject createMetric(LyDataSession session, MetricType type, Date startDate, Long startDuration,
			Date endDate, String actionId, String classId, String value)
	{
		if (areMetricsActivated(session, type) && (type != null))
		{
			LyApplication	application = session.getApplication();
			LyClassInfo		metricCls = application.getClassInfo(METRIC_CLASS);
			LyObject		metric = metricCls.createInstance();

			metric.setValue(METRIC_TYPE, type.toString());
			metric.setValue(METRIC_START_DATE, startDate);
			metric.setValue(METRIC_START_DURATION, startDuration);
			metric.setValue(METRIC_END_DATE, endDate);
			metric.setValue(METRIC_USER_ID, getUserId(session));
			metric.setValue(METRIC_ACTION_ID, actionId);
			metric.setValue(METRIC_CLASS_ID, classId);
			metric.setValue(METRIC_VALUE, value);
			metric.create();

			return metric;
		}

		return null;
	}

	/**
	 * Sets the end date of a given metric.
	 */
	public static void setMetricEndDate(LyDataSession session, MetricType type, LyObject metric, Date endDate)
	{
		if (areMetricsActivated(session, type) && (metric != null))
		{
			metric.setValue(METRIC_END_DATE, endDate);
			metric.set();
		}
	}

	/**
	 * Sets the userId of a given metric.
	 */
	public static void setMetricUserId(LyDataSession session, MetricType type, LyObject metric)
	{
		if (areMetricsActivated(session, type) && (metric != null))
		{
			metric.setValue(METRIC_USER_ID, getUserId(session));
			metric.set();
		}
	}

	/**
	 * Computes the userId for a given session.
	 */
	private static String getUserId(LyDataSession session)
	{
		LyObject	user = session.getUser();
		String		userId = null;

		if (user != null)
			userId = user.getObjectId().getFullString();

		return userId;
	}
} // end class LyMetricsTools
