/*
 * LyMultiClassWorkSpace.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import leon.app.*;
import leon.info.*;


/**
 * This class implements a multi-class workspace : workspace for objects belonging to a given list
 * of classes info and optionnally, matching a filter.
 *
 * @see LyClassInfo
 * @see LyFilter
 */
public class LyMultiClassWorkSpace extends LyWorkSpace
{
	/** File version */
	public static final String		  __VERSION =
		"$Id: LyMultiClassWorkSpace.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** List of class infos to which objects of this workspace belong. */
	protected LyInfoList<LyClassInfo> _classInfos;

	/** Filter applied for objects loaded in this workspace. */
	protected LyFilter				  _filter;

	/**
	 * Constructs a multi-class workspace within a given session for a list of class infos.
	 *
	 * @param session    Session in which the workspace is constructed.
	 * @param classInfos List of class infos to which objects of this workspace belong.
	 * @param dynamic    Specifies if this workspace is dynamic or not.
	 */
	public LyMultiClassWorkSpace(LyDataSession session, LyInfoList<LyClassInfo> classInfos,
		boolean dynamic)
	{
		this(session, classInfos, null, null, dynamic, true);
	}

	/**
	 * Constructs a multi-class workspace within a given session for a list of class infos with
	 * specified filter applied to objects of these classes.
	 *
	 * @param session    Session in which the workspace is constructed.
	 * @param classInfos List of class infos to which objects of this workspace belong.
	 * @param filter     Filter applied to objects for being in this workspace.
	 * @param dynamic    Specifies if this workspace is dynamic or not.
	 */
	public LyMultiClassWorkSpace(LyDataSession session, LyInfoList<LyClassInfo> classInfos,
		LyFilter filter,
		boolean dynamic)
	{
		this(session, classInfos, filter, null, dynamic, true);
	}

	/**
	 * Constructs a multi-class workspace within a given session for a list of class infos with
	 * specified filter applied to objects of these classes and sorted by the given sort.
	 *
	 * @param session    Session in which the workspace is constructed.
	 * @param classInfos List of class infos to which objects of this workspace belong.
	 * @param filter     Filter applied to objects for being in this workspace.
	 * @param sort       Sort requested in this workspace.
	 * @param dynamic    Specifies if this workspace is dynamic or not.
	 */
	public LyMultiClassWorkSpace(LyDataSession session, LyInfoList<LyClassInfo> classInfos,
		LyFilter filter,
		LySort sort, boolean dynamic)
	{
		this(session, classInfos, filter, sort, dynamic, true);
	}

	/**
	 * Internal constructor with all parameters for construction of a multi-class workpace and an
	 * additionnal one that indicates if the workspace is loaded or not.
	 *
	 * @param session    Session in which the workspace is constructed.
	 * @param classInfos List of class infos to which objects of this workspace belong.
	 * @param filter     Filter applied to objects for being in this workspace.
	 * @param sort       Sort requested in this workspace.
	 * @param dynamic    Specifies if this workspace is dynamic or not.
	 * @param sorted     Indicates if this workspace must be sorted or not.
	 */
	public LyMultiClassWorkSpace(LyDataSession session, LyInfoList<LyClassInfo> classInfos,
		LyFilter filter,
		LySort sort, boolean dynamic, boolean sorted)
	{
		this(session, classInfos, filter, sort, dynamic, sorted, false);
	}

	/**
	 * Internal constructor with all parameters for construction of a multi-class workpace and an
	 * additionnal one that indicates if the workspace is loaded or not.
	 *
	 * @param session    Session in which the workspace is constructed.
	 * @param classInfos List of class infos to which objects of this workspace belong.
	 * @param filter     Filter applied to objects for being in this workspace.
	 * @param sort       Sort requested in this workspace.
	 * @param dynamic    Specifies if this workspace is dynamic or not.
	 * @param sorted     Indicates if this workspace must be sorted or not.
	 * @param local      Specifies if this workspace is local or not.
	 */
	public LyMultiClassWorkSpace(LyDataSession session, LyInfoList<LyClassInfo> classInfos,
		LyFilter filter,
		LySort sort, boolean dynamic, boolean sorted, boolean local)
	{
		super(session, dynamic, local);
		setSorted(sorted);

		if ((classInfos == null) || (classInfos.getSize() == 0))
			throw new IllegalArgumentException(
				"BF AppEngine : list of classInfos can't be null nor empty");

		_classInfos = classInfos;
		_filter = filter;
		_sort = sort;

		// Initial loading
		reset();

		if (dynamic)
		{
			LyClassInfo    classInfo;
			LyObjectSource objectSource;
			int			   i, n = _classInfos.getSize();

			for (i = 0; i < n; i++)
			{
				classInfo = _classInfos.getInfo(i);
				objectSource = getObjectSource(classInfo);
				objectSource.addListener(this);
			}

			createFilterListeners(_filter);
		}
	}

	/**
	 * Free this workspace. Remove this workspace as listener from object sources of the class infos
	 * specified for this workspace.
	 */
	protected void free()
	{
		if (isDynamic())
		{
			LyClassInfo    classInfo;
			LyObjectSource objectSource;
			int			   i, n = _classInfos.getSize();

			for (i = 0; i < n; i++)
			{
				classInfo = _classInfos.getInfo(i);
				objectSource = getObjectSource(classInfo);
				objectSource.removeListener(this);
			}
		}

		super.free();
	}

	/**
	 * Return the list of all class infos to which objects of this workspace belong. This method is
	 * overriden to optimize it since we known from start to which classes the objects of this
	 * workspace belong.
	 *
	 * @return List of classes to which objects of this workspace belong.
	 */
	public LyInfoList<LyClassInfo> getClasses()
	{
		return new LyInfoList<LyClassInfo>(_classInfos);
	}

	/**
	 * Gets the filter of this class workspace.
	 *
	 * @return Filter specified for this class workspace.
	 */
	public LyFilter getFilter()
	{
		return _filter;
	}

	/**
	 * Method for resetting this workspace. Objects contained in this workspace are removed and
	 * reloaded from the object sources of the class infos specified for this workspace.
	 */
	public synchronized void reset()
	{
		// Remove all objects.
		super.reset();

		boolean wasSorted = isSorted();
		// Reload objects
		setSorted(false);

		LyClassInfo		    classInfo;
		LyObjectSource	    objectSource;
		int				    i, n = _classInfos.getSize();
		_LyResponseListener listener = new _LyResponseListener();

		for (i = 0; i < n; i++)
		{
			classInfo = _classInfos.getInfo(i);
			objectSource = getObjectSource(classInfo);
			// TODO : optimize the loaded fieldinfos.
			objectSource.getObjects(null, getSort(), _filter, -1, _local, listener);
		}

		setSorted(wasSorted);
	}

	/**
	 * Returns a string representation of this workspace.
	 *
	 * @return This workspace returned as a string.
	 */
	public String toString()
	{
		return super.toString() + "\nclassInfos=" + _classInfos.toString()
		+ ", filter=" + _filter;
	}

	/* Implementation of LyDataSourceListener */

	/**
	 * Method for receiving an object creation. This method is overriden to check if the object
	 * matchs the filter specified on this workspace if it has one.
	 *
	 * @param  event Event notifying the creation of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectCreated(LyDataSourceEvent event)
	{
		LyObject object = (LyObject)event.getObject();

		// If there is a filter, the object must match it.
		if ((_filter != null) && !_filter.match(object))
			return false;

		return super.objectCreated(event);
	}

	/**
	 * Method for receiving an object modification. If the object concerned by the event was not in
	 * this workspace, that means it didn't match its filter, so we now check if modifications done
	 * change this and in that case, we add the object in the workspace. On the contrary, if the
	 * object was in the workspace, we check if the object still matches the filter, and if not, we
	 * remove it from this workspace.
	 *
	 * @param  event Event notifying the modification of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectModified(LyDataSourceEvent event)
	{
		// If the object's event is added or removed from the workspace
		// we let the addObject or removeObject method do the job.
		if (useEvent(event))
			return true;

		// Notify the change, the list may need to be resorted
		return super.objectModified(event);
	}

	/* Implementation of LyDataAdvancedSourceListener */

	/**
	 * Called when an object state is changed.
	 *
	 * @param  event Event notifying object state change.
	 * @return True if the event is used by the listener, false otherwise.
	 */
	public boolean objectStateChanged(LyAdvancedDataSourceEvent event)
	{
		// If the object's event is added or removed from the workspace
		// we let the addObject or removeObject method do the job.
		if (useEvent(event))
			return true;

		// Notify the change, the list may need to be resorted
		return super.objectStateChanged(event);
	}

	/**
	 * Method for receiving user event on an object.
	 *
	 * @param  event Event notifying the user event of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectUserEvent(LyAdvancedDataSourceEvent event)
	{
		// If the object's event is added or removed from the workspace
		// we let the addObject or removeObject method do the job.
		if (useEvent(event))
			return true;

		// Notify the change, the list may need to be resorted
		return super.objectUserEvent(event);
	}

	/**
	 * Test if the specified event is used to add or remove an object from this workspace.
	 *
	 * @param  event Event to test
	 * @return true if the event adds or removes the object'event to or from the workspace, false
	 *         otherwise.
	 */
	protected boolean useEvent(LyDataSourceEvent event)
	{
		// If one of related objects is in the parent context, the new
		// object is added in this context.
		LyObject object = (LyObject)event.getObject();

		// If the object is not in the list
		if (indexOf(object) < 0)
		{
			// If adding is enabled and if the object matches the filter,
			// add it to the list.
			if ((_filter != null) && _filter.match(object))
			{
				addObject(object);
				return true;
			}

			return false;
		}

		// If the object is in the list, but doesn't match the filter anymore,
		// remove it.
		if ((_filter != null) && !_filter.match(object))
		{
			removeObject(object);
			return true;
		}

		return false;
	} // end method useEvent

	/* Implementation of LyResponseListener */

	/**
	 * Internal class that implements This allows to load objects without sorting them and sort them
	 * at the end of the loading if required.
	 */
	class _LyResponseListener
		implements LyResponseListener
	{

		/**
		 * Method for receiving a new object in a response.
		 *
		 * @param response  Response for which a new object is received.
		 * @param object    The new object returned in the response.
		 * @param values List of values that were changed for the object.
		 */
		public void newResponse(LyResponse response, Object object, LyValueSet values)
		{
			// Add the loaded object in our workspace
			addObject((LyObject)object);
		}

		/**
		 * Method for ending a response.
		 *
		 * @param response Response which end is notified.
		 */
		public void endResponse(LyResponse response)
		{
			// Sort responses if they are not
			if (_sorted && !response.isSorted())
				setSorted(true);
		}
	} // end class _LyResponseListener
} // end class LyMultiClassWorkSpace
