/*
 * LyMultiWorkSpace.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;
import leon.app.*;
import leon.info.*;


/**
 * This class implements a multiple workspace : workspace that is the result of merging other
 * workspaces.
 *
 * <p>Workspaces are added in a multi-workspace usind addList method. The result workspace can be
 * dynamic only if all added workspaces are dynamic.</p>
 *
 * <p>Since objects come from different workspaces and may be present in several of them, a counter
 * is associated to each object to count how many times they are present in these workspaces.</p>
 *
 * @see LyWorkSpaceListener
 */
public class LyMultiWorkSpace extends LyWorkSpace
{
	/** File version */
	public static final String   __VERSION =
		"$Id: LyMultiWorkSpace.java 35628 2012-05-30 14:46:55Z bangiole $";

	/** Objects lists added in this multi-workspace. */
	Vector<LyObjectList>		 _subObjectLists;

	/** Objects lists added in this multi-workspace Lock. */
	Object						 _subObjectListsLock = new Object();

	/** List of class infos represented in this list. */
	LyInfoList<LyClassInfo>		 _classes;

	/**
	 * Object counters table : only used when the workspace is dynamic. Associates a counter to each
	 * object of this workspace to be able to known when they have to be removed.
	 */
	Hashtable<LyObject, Integer> _counters;
	
	/** Counter Lock */
	Object						 _counterLock = new Object();

	/** Internal listener of sub-workspaces. */
	LyWorkSpaceListener			 _listener;

	/**
	 * Constructs an empty multi-workspace within a given session.
	 *
	 * @param session Session in which this workspace is constructed.
	 * @param dynamic Specifies if this workspace is dynamic or not.
	 */
	public LyMultiWorkSpace(LyDataSession session, boolean dynamic)
	{
		super(session, dynamic, false);

		// Initializing the attributes.
		_subObjectLists = new Vector<LyObjectList>(2);
		_classes = null;

		if (isDynamic())
		{
			_counters = new Hashtable<LyObject, Integer>();
			_listener = new _LyWorkSpaceListener();
		}
		else
		{
			_counters = null;
			_listener = null;
		}
	}

	/**
	 * Free this workspace. Overriden to unregister this workspace as listener of sub-workspaces if
	 * this workspace is dynamic.
	 */
	protected void free()
	{
		// Remove this as listener from sub-workspaces
		if (isDynamic())
		{
			synchronized (_subObjectListsLock)
			{
				LyObjectList list;
				int			 i, n = getObjectListCount();

				for (i = 0; i < n; i++)
				{
					list = getObjectList(i);
					if ((list instanceof LyWorkSpace) && list.isDynamic())
						((LyWorkSpace)list).free(_listener);
				}
			}

			_subObjectLists = null;
		}

		_classes = null;

		super.free();
	}

	/**
	 * Gets objects list at the given index.
	 *
	 * @param  index Index of the requested objects list.
	 * @return Objects list at the given index if found, null otherwise.
	 */
	LyObjectList getObjectList(int index)
	{
		if ((index < 0) || (index >= _subObjectLists.size()))
			return null;

		return _subObjectLists.elementAt(index);
	}

	/**
	 * Gets number of objects list managed in this multi-workspace.
	 *
	 * @return Number of objects list in this multi-workspace.
	 */
	int getObjectListCount()
	{
		return _subObjectLists.size();
	}

	/**
	 * Adds an objects list to our multiple workspace.
	 *
	 * @param list                     Objects list that is added to our workspace.
	 * @param IllegalArgumentException If this workspace is dynamic and the given list is not or if the list is null.
	 */
	public void addList(LyObjectList list)
	{
		if (list == null)
			throw new IllegalArgumentException("Leonardi : List can't be null");

		if (isDynamic())
		{
			if (!(list instanceof LyWorkSpace) || !list.isDynamic())
				throw new IllegalStateException(
					"BF AppEngine : non dynamic objects lists can't be added to a dynamic multi-workspace");

			if (_subObjectLists.contains(list))
				return;

			// Add the given list to our objects lists
			_subObjectLists.addElement(list);

			// Listen to workspace events on this list
			((LyWorkSpace)list).addListener(_listener);
		}

		// Add objects of the given list to our workspace
		synchronized (list)
		{
			int		 i, n;
			LyObject object;

			for (i = 0, n = list.getSize(); i < n; i++)
			{
				object = list.getObject(i);
				if (isDynamic())
					incrementCount(object);
				else
					addObject(object);
			}
		}

		// Add classes of the given list to our list of known classes
		LyInfoList<LyClassInfo> classes = list.getClasses();

		if ((classes == null) || (classes.getSize() == 0))
			return;

		if (_classes == null)
			_classes = new LyInfoList<LyClassInfo>();

		_classes.addList(classes);
	} // end method addList

	/**
	 * Removes an objects list from our multiple workspace.
	 *
	 * @param list                     Objects list that is removed from our workspace.
	 * @throw IllegalStateException If this workspace is dynamic and the given list is not found
	 *                                 in our workspace or if the list is null.
	 */
	public void removeList(LyObjectList list)
	{
		if (list == null)
			throw new IllegalArgumentException("Leonardi : List can't be null");

		if (isDynamic())
		{
			if (!_subObjectLists.contains(list))
				throw new IllegalStateException(
					"BF AppEngine : the given list was not found in this multi-workspace");

			// Remove the given list to our objects lists
			_subObjectLists.removeElement(list);

			// Stop listening to workspace events on this list
			((LyWorkSpace)list).removeListener(_listener);
		}

		// Remove objects of the given list to our workspace
		synchronized (list)
		{
			int		 i, n;
			LyObject object;

			for (i = 0, n = list.getSize(); i < n; i++)
			{
				object = list.getObject(i);
				if (isDynamic())
					decrementCount(object);
				else
					removeObject(object);
			}
		}

		// Recompute list of classes
		if (isDynamic())
		{
			LyObjectList		    subList;
			LyInfoList<LyClassInfo> classes;
			int					    i, n = getObjectListCount();

			_classes = new LyInfoList<LyClassInfo>();
			for (i = 0; i < n; i++)
			{
				subList = getObjectList(i);
				classes = subList.getClasses();
				if ((classes != null) && (classes.getSize() > 0))
					_classes.addList(classes);
			}
		}
		else
		{
			LyInfoList<LyClassInfo> classes = list.getClasses();
			int					    i, n = ((classes == null) ? 0 : classes.getSize());
			LyClassInfo			    classInfo;
			int					    j, no = getSize();
			LyObject			    object;

			for (i = 0; i < n; i++)
			{
				classInfo = classes.getInfo(i);
				for (j = 0; j < no; j++)
				{
					object = list.getObject(j);
					if (object.getClassInfo() == classInfo)
						break;
				}
				if (j == no)
					// classInfo not found, remove it from classes list
					classes.removeInfo(classInfo);
			}
		}
	} // end method removeList

	/**
	 * Return the list of all class infos to which objects of this workspace belong.
	 *
	 * @return List of classes to which objects of this workspace belong.
	 */
	public LyInfoList<LyClassInfo> getClasses()
	{
		return ((_classes == null) ? null : new LyInfoList<LyClassInfo>(_classes));
	}

	/* Management of the content of the list */

	/**
	 * Method for resetting a workspace. This method is overriden to reset all sub-workspaces.
	 */
	public synchronized void reset()
	{
		// Remove all objects.
		super.reset();

		if (isDynamic())
		{
			// Reset objects counters
			_counters = (isDynamic() ? new Hashtable<LyObject, Integer>() : null);

			// Reset all sub-workspaces
			synchronized (_subObjectListsLock)
			{
				LyObjectList list;
				int			 i, n = getObjectListCount();

				for (i = 0; i < n; i++)
				{
					list = getObjectList(i);
					if (list instanceof LyWorkSpace)
						((LyWorkSpace)list).reset();
				}
			}
		}
	}

	/**
	 * Removes an object from this workspace. Overriden to remove the counter of the given object.
	 *
	 * @param     object Object to remove from the workspace.
	 * @return    Index where the object was found in this workspace.
	 * @exception IllegalArgumentException If the object was not found in this workspace.
	 */
	protected int removeObject(LyObject object)
	{
		// Remove object
		int index = super.removeObject(object);

		if ((index >= 0) && (isDynamic()))
			// Remove counter for this object
			_counters.remove(object);

		return index;
	}

	/**
	 * Gets counter value for the given object (number of occurrences of the object in the different
	 * sub-workspaces).
	 *
	 * @param  object Object for which the counter value is requested.
	 * @return Counter value for the given object.
	 */
	int getCount(LyObject object)
	{
		if (_counters == null)
			return -1;

		Integer count = _counters.get(object);

		return ((count == null) ? 0 : count.intValue());
	}

	/**
	 * Increments counter value for the given object. If the counter value was 0, the object is
	 * added in this workspace.
	 *
	 * @param object Object for which the counter value is incremented.
	 */
	void incrementCount(LyObject object)
	{
		synchronized (_counterLock)
		{
			int count = getCount(object);

			if (count == 0)
				addObject(object);
			else
				counterChanged(object, true);

			_counters.put(object, new Integer(count + 1));
		}
	}

	/**
	 * Decrements counter value for the given object. If the counter value falls to 0, the object is
	 * removed from this workspace.
	 *
	 * @param object Object for which the counter value is decremented.
	 */
	void decrementCount(LyObject object)
	{
		synchronized (_counterLock)
		{
			int count = getCount(object);

			if (count == 1)
				removeObject(object);
			else
			{
				counterChanged(object, false);
				_counters.put(object, new Integer(count - 1));
			}
		}
	}

	void counterChanged(LyObject object, boolean incr)
	{
		LyWorkSpaceEvent event = new LyWorkSpaceEvent(this, LyWorkSpaceEvent.Type.COUNTER_CHANGE,
				object, getIndex(object));

		event.setIncremented(incr);
		notify(event);
	}

	/**
	 * Returns a string representation of this workspace.
	 *
	 * @return This workspace returned as a string.
	 */
	public String toString()
	{
		StringBuffer buffer = new StringBuffer(super.toString());

		synchronized (_subObjectListsLock)
		{
			LyObjectList list;
			int			 i, n = getObjectListCount();

			for (i = 0; i < n; i++)
			{
				list = getObjectList(i);
				buffer.append("\nparent[");
				buffer.append(i);
				buffer.append("] = ");
				buffer.append(list.toString());
			}
		}

		return buffer.toString();
	}

	/**
	 * Method for retrieving source objects that lead to given target object.
	 *
	 * @param  target Target object
	 * @return Source objects that leade to target object.
	 */
	public LySimpleObjectList getSources(LyObject target)
	{
		synchronized (_subObjectListsLock)
		{
			LySimpleObjectList results = null, listResults;
			LyObjectList	   list;
			int				   i, n = getObjectListCount();

			for (i = 0; i < n; i++)
			{
				list = getObjectList(i);
				if (list instanceof LyWorkSpace)
				{
					listResults = ((LyWorkSpace)list).getSources(target);
					if (listResults != null)
					{
						if (results == null)
							results = new LySimpleObjectList();
						results.addList(listResults);
					}
				}
			}

			return results;
		}
	}

	/* Implementation of LyWorkSpaceListener */

	/**
	 * Internal class that implements LyWorkSpaceListener.
	 */
	class _LyWorkSpaceListener
		implements LyWorkSpaceListener,
		LyGetWorkSpaceInterface
	{

		/**
		 * Events counters table : only used when the workspace is dynamic to prevent multiple
		 * notifications for object changes. Associates a counter to each object that count the
		 * number of calls on object Modified. The change is really notified to listeners when the
		 * events count is equal to the number of occurrences of the notified object.
		 */
		protected Hashtable<LyObject, Integer> _events;

		/**
		 * Called when an object is added in the workspace. Increments count of the object concerned
		 * by this event.
		 *
		 * @param  event Event notifying object add.
		 * @return True if the event is used by the listener, false otherwise.
		 * @see    LyMultiWorkSpace#incrementCount
		 */
		public boolean objectAdded(LyWorkSpaceEvent event)
		{
			LyObject object = event.getObject();

			incrementCount(object);
			return true;
		}

		/**
		 * Called when an object is removed from the workspace. Decrements count of the object
		 * concerned by this event.
		 *
		 * @param  event Event notifying object remove.
		 * @return True if the event is used by the listener, false otherwise.
		 * @see    LyMultiWorkSpace#decrementCount
		 */
		public boolean objectRemoved(LyWorkSpaceEvent event)
		{
			LyObject object = event.getObject();

			decrementCount(object);
			return true;
		}

		/**
		 * Called when an object has changed in the workspace. Wait for the last call of this event
		 * before notifying the change to our listeners.
		 *
		 * @param  event Event notifying object change.
		 * @return True if the event is used by the listener, false otherwise.
		 */
		public boolean objectChanged(LyWorkSpaceEvent event)
		{
			LyObject object = event.getObject();
			int		 count = getCount(object);

			if (count == 1)
				return changeObject(object, event.getValues(), event.getEventInfo());

			int eventCount;

			if (_events == null)
				eventCount = 0;
			else
			{
				Integer objectEventCount = _events.get(object);

				eventCount = ((objectEventCount == null) ? 0 : objectEventCount.intValue());
			}

			if ((eventCount + 1) == count)
			{
				_events.remove(object);

				return changeObject(object, event.getValues(), event.getEventInfo());
			}
			else
			{
				if (_events == null)
					_events = new Hashtable<LyObject, Integer>(1);

				_events.put(object, new Integer(eventCount + 1));
			}

			return true;
		} // end method objectChanged
		
		public LyWorkSpace getWorkSpace()
		{
			return LyMultiWorkSpace.this;
		}
	} // end class _LyWorkSpaceListener
} // end class LyMultiWorkSpace
