/*
 * LyNumberValue.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import leon.info.*;


/**
 * This class offers methods for constructing, accessing and modifying values corresponding to a
 * LyNumberFieldInfo.
 *
 * @see LyNumberFieldInfo
 * @see LyObject
 */
public final class LyNumberValue extends LyFieldInfoValue
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyNumberValue.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Constructs an empty number value for the given number field.
	 *
	 * @param numberFieldInfo Number fieldInfo for which a value is created.
	 */
	public LyNumberValue(LyNumberFieldInfo numberFieldInfo)
	{
		super(numberFieldInfo);
	}

	/**
	 * Constructs a number value for the given number field from a string.
	 *
	 * @param numberFieldInfo Number fieldInfo for which a value is created.
	 * @param value           Value given as Number.
	 */
	public LyNumberValue(LyNumberFieldInfo numberFieldInfo, Number value)
	{
		super(numberFieldInfo);

		if ((value != null)
				&& (value.getClass() != numberFieldInfo.getValueClass()))
		{
			switch (numberFieldInfo.getType())
			{
			case INT:
				value = new Integer(value.intValue());
				break;

			case SHORT:
				value = new Short(value.shortValue());
				break;

			case LONG:
				value = new Long(value.longValue());
				break;

			case FLOAT:
				value = new Float(value.floatValue());
				break;

			case DOUBLE:
				value = new Double(value.doubleValue());
				break;
			}
		}

		setValue(value);
	}

	/**
	 * Constructs a number value for the given number field from a short.
	 *
	 * @param numberFieldInfo Number fieldInfo for which a value is created.
	 * @param value           Value given as a short.
	 */
	public LyNumberValue(LyNumberFieldInfo numberFieldInfo, short value)
	{
		this(numberFieldInfo, (double)value);
	}

	/**
	 * Constructs a number value for the given number field from an int.
	 *
	 * @param numberFieldInfo Number fieldInfo for which a value is created.
	 * @param value           Value given as an int.
	 */
	public LyNumberValue(LyNumberFieldInfo numberFieldInfo, int value)
	{
		this(numberFieldInfo, (double)value);
	}

	/**
	 * Constructs a number value for the given number field from a long.
	 *
	 * @param numberFieldInfo Number fieldInfo for which a value is created.
	 * @param value           Value given as a long.
	 */
	public LyNumberValue(LyNumberFieldInfo numberFieldInfo, long value)
	{
		this(numberFieldInfo, (double)value);
	}

	/**
	 * Constructs a number value for the given number field from a float.
	 *
	 * @param numberFieldInfo Number fieldInfo for which a value is created.
	 * @param value           Value given as a float.
	 */
	public LyNumberValue(LyNumberFieldInfo numberFieldInfo, float value)
	{
		this(numberFieldInfo, (double)value);
	}

	/**
	 * Constructs a number value for the given number field from a double.
	 *
	 * @param numberFieldInfo Number fieldInfo for which a value is created.
	 * @param value           Value given as a double.
	 */
	public LyNumberValue(LyNumberFieldInfo numberFieldInfo, double value)
	{
		super(numberFieldInfo);

		switch (numberFieldInfo.getType())
		{
		case INT:
			setValue(new Integer((int)value));
			break;

		case SHORT:
			setValue(new Short((short)value));
			break;

		case LONG:
			setValue(new Long((long)value));
			break;

		case FLOAT:
			setValue(new Float((float)value));
			break;

		case DOUBLE:
			setValue(new Double(value));
			break;
		}
	}

	/**
	 * Constructs a number value for the given number field from a string.
	 *
	 * @param numberFieldInfo Number fieldInfo for which a value is created.
	 * @param value           String value (as expected by parse method).
	 * @see   LyNumberFieldInfo#parse
	 */
	public LyNumberValue(LyNumberFieldInfo numberFieldInfo, String value)
	{
		super(numberFieldInfo, value);
	}

	/**
	 * Gets value of this number value as a Number.
	 *
	 * @return A Number corresponding to this value.
	 */
	public Number getNumberValue()
	{
		return (Number)getValue();
	}

	/**
	 * Gets value of this number value as a Short.
	 *
	 * @return A Short corresponding to this value.
	 */
	public Short getShortValue()
	{
		Number number = getNumberValue();

		return ((number == null)
			? null : ((number instanceof Short) ? (Short)number : new Short(number.shortValue())));
	}

	/**
	 * Gets value of this number value as an Integer.
	 *
	 * @return An Integer corresponding to this value.
	 */
	public Integer getIntegerValue()
	{
		Number number = getNumberValue();

		return ((number == null)
			? null
			: ((number instanceof Integer) ? (Integer)number : new Integer(number.intValue())));
	}

	/**
	 * Gets value of this number value as a Long.
	 *
	 * @return A Long corresponding to this value.
	 */
	public Long getLongValue()
	{
		Number number = getNumberValue();

		return ((number == null)
			? null : ((number instanceof Long) ? (Long)number : new Long(number.longValue())));
	}

	/**
	 * Gets value of this number value as a Float.
	 *
	 * @return A Float corresponding to this value.
	 */
	public Float getFloatValue()
	{
		Number number = getNumberValue();

		return ((number == null)
			? null : ((number instanceof Float) ? (Float)number : new Float(number.floatValue())));
	}

	/**
	 * Gets value of this number value as a Double.
	 *
	 * @return A Double corresponding to this value.
	 */
	public Double getDoubleValue()
	{
		Number number = getNumberValue();

		return ((number == null)
			? null
			: ((number instanceof Double) ? (Double)number : new Double(number.doubleValue())));
	}

	/**
	 * Gets value of this number value as a short.
	 *
	 * @return A short corresponding to this value. Returns Short.MIN_VALUE if the number value is
	 *         null.
	 */
	public short shortValue()
	{
		Number number = getNumberValue();

		return ((number == null) ? Short.MIN_VALUE : number.shortValue());
	}

	/**
	 * Gets value of this number value as an int.
	 *
	 * @return An int corresponding to this value. Returns Integer.MIN_VALUE if the number value is
	 *         null.
	 */
	public int intValue()
	{
		Number number = getNumberValue();

		return ((number == null) ? Integer.MIN_VALUE : number.intValue());
	}

	/**
	 * Gets value of this number value as a long.
	 *
	 * @return A long corresponding to this value. Returns Long.MIN_VALUE if the number value is
	 *         null.
	 */
	public long longValue()
	{
		Number number = getNumberValue();

		return ((number == null) ? Long.MIN_VALUE : number.longValue());
	}

	/**
	 * Gets value of this number value as a float.
	 *
	 * @return A float corresponding to this value. Returns Float.MIN_VALUE if the number value is
	 *         null.
	 */
	public float floatValue()
	{
		Number number = getNumberValue();

		return ((number == null) ? Float.MIN_VALUE : number.floatValue());
	}

	/**
	 * Gets value of this number value as a double.
	 *
	 * @return A double corresponding to this value. Returns Double.MIN_VALUE if the number value is
	 *         null.
	 */
	public double doubleValue()
	{
		Number number = getNumberValue();

		return ((number == null) ? Double.MIN_VALUE : number.doubleValue());
	}
} // end class LyNumberValue
