/*
 * LyObject.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.lang.ref.*;
import java.lang.reflect.*;
import java.util.*;

import leon.app.*;
import leon.data.daemon.*;
import leon.data.datainterface.*;
import leon.info.*;
import leon.info.binding.*;
import leon.misc.*;


/**
 * This class implements a logical object.
 *
 * <p>A logical object belongs to a class info, has an object identifier (LyObjectId) that is unique
 * within an application and may have a name.</p>
 *
 * <p>Logical objects are mapped on physical objects (peer), that contain real values corresponding
 * to non local fields. Local values are managed on logical objects and are not persistent.</p>
 *
 * <p>Logical objects can be created, modified and deleted. These Operations are managed by objects
 * source (LyObjectSource).</p>
 *
 * <p>They can be given in objects list or workspaces (dynamic objects lists that may send events on
 * objects).</p>
 *
 * <p>This class may be inherited to offer non abstract interfaces to access or modify fields. Each
 * class info can have its own subclass.</p>
 *
 * @see LyClassInfo
 * @see LyObjectId
 * @see LyApplication
 * @see LyObjectList
 * @see LyObjectSource
 */
public class LyObject
{
	/** File version */
	public static final String		  __VERSION =
		"$Id: LyObject.java 41148 2012-10-11 14:58:26Z jlsaouli $";

	/** Error levels. Erros levels greater than LEVEL_WARNING are user specific levels. */
	public static final short		  LEVEL_OK = 0;
	public static final short		  LEVEL_ERROR = 1;
	public static final short		  LEVEL_WARNING = 2;

	/** States of the object. */
	// Default state
	protected static final short	  STATE_DEFAULT = 0;
	// Indicates if this object has been deleted.
	protected static final short	  STATE_DELETED = 1;
	// Indicates if this object has been locally deleted.
	protected static final short	  STATE_LOCALLY_DELETED = 2;
	// Selection state of this object.
	protected static final short	  STATE_SELECTED = 4;
	// Visibility state of this object.
	protected static final short	  STATE_HIDDEN = 8;
	// To avoid loops in control method.
	protected static final short	  STATE_IN_CONTROL = 16;
	// To avoid encoding value
	protected static final short	  STATE_SKIP_ENCODING = 32;
	// This object is not intended to be created. temporary only.
	protected static final short	  STATE_TEMPORARY = 64;
	// This object successfully passed class controls
	protected static final short	  STATE_CLASS_CONTROLLED = 128;
    // This object is completely loaded
    protected static final short      STATE_FULLY_LOADED = 256;

	/** Controls that may be applied on objects. */
	// No control
	public static final short		  CONTROL_NONE = 0;
	// Id control
	public static final short		  CONTROL_ID = 1;
	// AppId control
	public static final short		  CONTROL_APP_ID = 2;
	// Mandatory control
	public static final short		  CONTROL_MANDATORY = 4;
	// Unicity control
	public static final short		  CONTROL_UNICITY = 8;
	// Value control
	public static final short		  CONTROL_VALUE = 16;
	// Rules control
	public static final short		  CONTROL_RULES = 32;
	// Application field controls
	public static final short		  CONTROL_APP_FIELD_CONTROLS = 64;
	// Application field controls
	public static final short		  CONTROL_APP_CLASS_CONTROLS = 128;
	// Loading controls
	public static final short		  CONTROL_LOADING = 256;

	// Used to reset all controls of a type.
	private static final short		  ALL_CONTROLS = 1;

	// Class controls
	public static final short		  CONTROL_CLASS_CONTROLS = CONTROL_RULES
			| CONTROL_APP_CLASS_CONTROLS
			| CONTROL_MANDATORY | CONTROL_ID | CONTROL_APP_ID;
	// Field controls
	public static final short		  CONTROL_FIELD_CONTROLS = CONTROL_MANDATORY
			| CONTROL_UNICITY | CONTROL_VALUE | CONTROL_APP_FIELD_CONTROLS;
	// All controls
	public static final short		  CONTROL_FULL = CONTROL_CLASS_CONTROLS
			| CONTROL_FIELD_CONTROLS;
	// Controls disabled on loading
	public static final short		  CONTROL_DISABLED_ON_LOADING = CONTROL_MANDATORY | CONTROL_ID
			| CONTROL_APP_ID | CONTROL_UNICITY | CONTROL_VALUE;

    public static short               DEFAULT_CONTROL = CONTROL_FULL;

	/** Flags for operations. */
	// No flags
	public static final short		  OPER_NONE = 0;
	// Local operation
	public static final short		  OPER_LOCAL = 1;
	// Operation with refresh needed
	public static final short		  OPER_NEED_REFRESH = 2;
	// Operation with context needed
	public static final short		  OPER_NEED_CONTEXT = 4;
	// Operation with set needed
	public static final short		  OPER_NEED_SET = 8;

	// For upward compatibility; Check optional on data layer
	public static boolean 			  CHECK_OPTIONAL_ON_OBJECTS = false;

	/** Default object returned when the object corresponding to a key is not found. */
    private static final String       NOT_FOUND = "__NOT_FOUND__".intern();

    /** Default object returned when the object corresponding to a key is not loaded. */
    private static final Object       NOT_LOADED = "__NOT_LOADED__".intern();

	/** Number of objects for debug. */
	public static int				  _nbObjects;

	/** Class info to which this object belongs. */
	protected LyClassInfo			  _classInfo;

	/** Object identifier. */
	protected LyObjectId			  _id;

	/** Name of the object. */
	protected String				  _name;

	/** Instance supporting physical values (LyPeerFieldValue). */
	protected Object				  _peer;

	/** Values set containing all local values. */
	protected LyValueSet			  _localValues;

	/** Values set containing new values before they are committed. */
	protected LyValueSet			  _newValues;

	/** Values set containing values locally committed. */
	protected LyValueSet			  _pendingValues;

	/**
	 * Values corresponding to the last accessed fields. This value is kept to optimize getValue
	 * method.
	 */
	protected Object[] 			      _lastValues;

	/** List of error on this object. */
	protected ArrayList<LyError>	  _errors;

	/** The error level of this object. */
	protected short					  _errorLevel;

	/** Global state of the logical object. */
	protected short					  _state;

	/**
	 * Context type during modification state of this object : this information may be used by the
	 * LyContext to indicate the kind of modification (User or by code) for read only attributes The
	 * default value -1 means 'undefined' or not within a context.
	 */
	protected LyContext.Type		  _setContextType;

	/** Parent object (used for struct). */
	protected WeakReference<LyObject> _parentReference;

	/** Parent fieldInfo (used for struct). */
	protected LyFieldInfo			  _parentFieldInfo;

	/**
	 * Default Constructor. Note that in this case the class info is not set at first, before using
	 * this object this class info must be set using setClassInfo method.
	 */
	public LyObject()
	{
		_classInfo = null;
		_id = null;
		_name = null;
		_peer = null;
		_localValues = null;
		_newValues = null;
		_pendingValues = null;
		_errors = null;
		_errorLevel = LEVEL_OK;
		_state = STATE_DEFAULT;
		_setContextType = null; // Unknown
		_parentFieldInfo = null;
		_nbObjects++;
	}

	/**
	 * Constructs a logical object for a given class info.
	 *
	 * @param classInfo Class info of this object.
	 */
	public LyObject(LyClassInfo classInfo)
	{
		this();
		_classInfo = classInfo;

		if (this.getClass() != _classInfo.getObjectClass())
			throw new IllegalArgumentException(
				"BF AppEngine : Attempt to create a LyObject but classInfo :" + _classInfo
				+ " expects :" + _classInfo.getObjectClass());
	}

	/**
	 * Constructs a logical object for a given class info with the specified list of values.
	 *
	 * @param classInfo Class info of this object.
	 * @param values    List of values for this object.
	 */
	public LyObject(LyClassInfo classInfo, LyValueSet values)
	{
		this(classInfo);

		setValues(values);
	}

	/**
	 * Sets class info for this object. This method can only be used for an object constructed with
	 * default constructor (the one with no parameters).
	 *
	 * @param     classInfo Class info of this object.
	 * @exception IllegalStateException If the class info of this object was already set.
	 */
	public void setClassInfo(LyClassInfo classInfo)
	{
		if (classInfo == _classInfo)
			return;
		
		if ((_classInfo != null)
				&& (classInfo != null)
				&& (classInfo.getTemplate() != _classInfo))
			throw new IllegalStateException("BF AppEngine : changing classInfo from "
				+ _classInfo.getId() + " to " + ((classInfo != null) ? classInfo.getId() : "null")
				+ " for an object is not allowed");

		_classInfo = classInfo;
	}

	/**
	 * Finalize method.
	 */
	protected void finalize()
		throws Throwable
	{
		_nbObjects--;

		LyObjectSource objectSource = getObjectSource();

		if (objectSource != null)
		{
			objectSource.finalizeObject(this);
		}

		// Call free to notify object finalization
		free();

		// Clean object on finalize to remove it from daemons
		cleanObject();
	}

	/**
	 * Free object. Method called when the object is either removed from memory or deleted.
	 */
	protected void free()
	{
	}

	/**
	 * Free local values computed by daemon.
	 */
	public void freeDaemonValues()
	{
        LyApplication     application = _classInfo.getApplication();

        if (application == null)
            return;

        // Freeing calculated values.
        int               i, n = _classInfo.getFieldInfoCount();
        LyFieldInfo       fieldInfo;
        LyFieldInfoDaemon daemon;
        LyValueSet		  localValues;

        // Freeing calculated values.

        for (i = 0; i < n; i++)
        {
            fieldInfo = _classInfo.getFieldInfo(i);
            if (!fieldInfo.isLocal())
                continue;

            // Remove daemon value
            daemon = application.getDaemon(fieldInfo);
            if (daemon != null)
            {
                daemon.freeValue(this);

	            // Remove local value
                localValues = _localValues;
	            if (localValues != null)
	                localValues.removeValue(fieldInfo);
            }

            // Remove last value
            removeLastValue(fieldInfo);
        }
	}

	/**
	 * Clean the object when it is removed from the cache INTERNAL USE ONLY.
	 */
	private void cleanObject()
	{
	    freeDaemonValues();

	    synchronized (this)
	    {
	    	_localValues = null;
	    }

		removed();
		touch();
	} // end method cleanObject

	/**
	 * Clean all the local value computed by a daemon INTERNAL USE ONLY.
	 */
	void cleanDaemonValue()
	{
		cleanDaemonValue((LyInfoList<LyFieldInfo>)null);
	} // end method cleanDaemonValue

	/**
	 * Clean all the local values computed by a daemon for the given fields.
	 * @param fields a list of fields
	 */
	void cleanDaemonValue(List<LyFieldInfo> fields)
	{
		cleanDaemonValue(new LyInfoList<LyFieldInfo>(fields));
	} // end method cleanDaemonValue

	/**
	 * Clean all the local values computed by a daemon for the given fields.
	 * @param fields a list of fields
	 */
	void cleanDaemonValue(LyInfoList<LyFieldInfo> fields)
	{
		if (fields == null)
			fields = getClassInfo().getFieldInfos();

		int						i, n = fields.getSize();
		LyFieldInfo				fieldInfo;
		LyFieldInfoDaemon		daemon;
		LyApplication			application = getApplication();
		LyValueSet				localValues;

		// Freeing calculated values.

		for (i = 0; i < n; i++)
		{
			fieldInfo = fields.getInfo(i);
			daemon = application.getDaemon(fieldInfo);

			if (daemon != null)
			{
				daemon.freeValue(this);

				localValues = _localValues;
				if (localValues != null)
					localValues.removeValue(fieldInfo);
			}
		}
	} // end method cleanDaemonValue

	/**
	 * Called when the counter of locks reaches zero. This method must be overriden in interfaces
	 * classes when the application has to do some work (freeing resources or something else).
	 * Warning : the current object has been removed from the cache and local values have been freed
	 * when this method is called.
	 */
	protected void removed()
	{
	}

	/**
	 * Retrieves the class info of this object.
	 *
	 * @return Class info of this object.
	 */
	public final LyClassInfo getClassInfo()
	{
		return _classInfo;
	}

	/**
	 * Checks if this object is an instance of the given class info.
	 *
	 * @param  classInfo Class info to check.
	 * @return True if this object is an instance of the given class info, false otherwise.
	 */
	public final boolean instanceOf(LyClassInfo classInfo)
	{
		if (classInfo.isVirtual())
			classInfo = classInfo.getRealClass();

		if (classInfo == _classInfo)
			return true;
		else
			return _classInfo.inheritsFrom(classInfo);
	}

	/**
	 * Retrieves application of the class info of this object.
	 *
	 * @return Application of the class info of this object.
	 */
	public final LyApplication getApplication()
	{
		return _classInfo.getApplication();
	}

	/**
	 * Retrieves environment of the class info of this object.
	 *
	 * @return Environment of the class info of this object.
	 */
	public final LyEnvironment getEnvironment()
	{
		return _classInfo.getEnvironment();
	}

	/**
	 * Shortcut to get object source for this object. The object source is retrieved from the class
	 * info of this object.
	 *
	 * @return Object source corresponding to the class info of this object.
	 * @see    LyClassInfo#getObjectSource
	 */
	public final LyObjectSource getObjectSource()
	{
		return (_classInfo == null ? null : _classInfo.getObjectSource());
	}

	/**
	 * Returns the identifier of this object (LyObjectId).
	 *
	 * @return Identifier of the object.
	 */
	public LyObjectId getObjectId()
	{
		if (_id != null)
			return _id;

		LyInfoList<LyFieldInfo> fieldInfos = _classInfo.getIdFieldInfos();
		int					    i, n = fieldInfos.getSize();
		Object				    result;
		LyFieldInfo			    fieldInfo;

		if (n == 1)
		{
			// The identifier of this object is the value of the field.
			fieldInfo = fieldInfos.getInfo(0);
			result = getValue(fieldInfo);

			if (result == null)
				result = LyObjectId.getIdValue(_classInfo, fieldInfo);
		}
		else
		{
			// The identifier of object is an array of values of all id fields.
			result = new Object[n];

			for (i = 0; i < n; i++)
			{
				fieldInfo = fieldInfos.getInfo(i);
				((Object[])result)[i] = getValue(fieldInfo);
				if ((((Object[])result)[i] == null) && (_peer == null))
					((Object[])result)[i] = LyObjectId.getIdValue(_classInfo, fieldInfo);
			}
		}

		_id = new LyObjectId(_classInfo, result);
		_id.setObject(this);
		return _id;
	} // end method getObjectId

	/**
	 * Returns the identifier of this object (LyObjectId).
	 *
	 * @return Identifier of the object.
	 */
	public LyObjectId getObjectId(LyValueSet values)
	{
		if ((values == null)
				|| !values.hasMark(LyFieldInfo.MARK_ID))
			return getObjectId();

		LyInfoList<LyFieldInfo> fieldInfos = _classInfo.getIdFieldInfos();
		int					    i, n = fieldInfos.getSize();
		Object				    result;
		LyFieldInfo			    fieldInfo;
		LyValue				    fieldValue;

		if (n == 1)
		{
			// The identifier of this object is the value of the field.
			fieldInfo = fieldInfos.getInfo(0);
			if (values.contains(fieldInfo))
			{
				fieldValue = values.getFieldValue(fieldInfo);
				result = fieldValue.getValue();
			}
			else
				result = getValue(fieldInfo);
		}
		else
		{
			// The identifier of object is an array of values of all id fields.
			result = new Object[n];

			for (i = 0; i < n; i++)
			{
				fieldInfo = fieldInfos.getInfo(i);
				if (values.contains(fieldInfo))
				{
					fieldValue = values.getFieldValue(fieldInfo);
					((Object[])result)[i] = fieldValue.getValue();
				}
				else
					((Object[])result)[i] = getValue(fieldInfo);
			}
		}

		if (result == null)
			return null;

		return new LyObjectId(_classInfo, result);
	} // end method getObjectId

	/**
	 * Get object id as a string. This is an internal unique representation of the identifier
	 * <classInfoId>\<id> To get the value of the idField, use getValue(<idField>) or
	 * getObjectId().toString() instead.
	 *
	 * @return Object identifier returned as a string.
	 * @see    LyObjectId#getFullString
	 */
	public final String getId()
	{
		LyObjectId objectId = getObjectId();

		return ((objectId == null) ? null : objectId.getFullString());
	}

	/**
	 * Get the id of this object in a displayable form.
	 *
	 * @return A string representation of the id of this object computed with the LyInfoLabel of
	 *         type ID if any, else returns the result of the getId() method.
	 */
	public String getDisplayId()
	{
		LyInfoLabel label = _classInfo.getLabel(LyInfoLabel.Type.ID);

		if (label != null)
			return label.getString(this);

		char nameSep = getApplication().NAME_SEP;

		return getNameString(LyFieldInfo.MARK_ID, nameSep);
	}

	/**
	 * Gets values names for this object of all field infos that have the given mark, concatenated
	 * in a string and separated by the specified separator.
	 *
	 * @param  mark      Mark checked on field infos.
	 * @param  separator Separator used to separate values name if several field infos match the
	 *                   specified mark.
	 * @return String containing values names on this object for all field infos that have the given
	 *         mark.
	 */
	String getNameString(int mark, char separator)
	{
		LyInfoList<LyFieldInfo> fieldInfos = _classInfo.getFieldInfos(mark);
		int					    n = ((fieldInfos == null) ? 0 : fieldInfos.getSize());

		if (n == 0)
			return null;

		String	    str, result = null;
		Object	    value;
		LyFieldInfo fieldInfo;
		int		    i;

		for (i = 0; i < n; i++)
		{
			fieldInfo = fieldInfos.getInfo(i);
			value = getValue(fieldInfo);
			str = fieldInfo.toName(value, true);
			if (str == null)
				continue;

			if (result == null)
				result = str;
			else
				result = result + separator + str;
		}

		return result;
	} // end method getNameString

	/**
	 * Get name of this object.
	 *
	 * @return Values on this object for all fields that have MARK_NAME concatenated in a string and
	 *         separated by the name separator (NAME_SEP) of the application.
	 */
	public String getName()
	{
		if ((_lastValues != null)
				&& (_name != null))
			return _name;

		// To avoid loops, first calculate a name from id
		if ((_name == null) || (_name.length() == 0))
		{
			_name = getDisplayId();
			if ((_name == null) && (_id != null))
				_name = _id.toString();
		}

		// Basic method : use fields with NAME mark
		char   nameSep = getApplication().NAME_SEP;
		String name = getNameString(LyFieldInfo.MARK_NAME, nameSep);

		if (name != null)
			_name = name;

		// Use label for a more sophisticated name
		LyInfoLabel label = _classInfo.getLabel(LyInfoLabel.Type.NAME);

		if (label != null)
		{
			LyApplication application = getApplication();
			LyLabelDaemon daemon = application.getDaemon(label);

			if (daemon != null)
				name = daemon.getString(this);
			else
				name = label.getString(this);

			if (name != null)
				_name = name;
		}

		return _name;
	} // end method getName

	/**
	 * Fetchs the value from this object corresponding to the given field info.
	 *
	 * @param  fieldInfo Field info for which a value is requested. The fieldInfo must exist in the
	 *                   associated class info.
	 * @return Value of the given field info for this object. The class of the returned value
	 *         depends on the type of the fieldInfo (Integer, Date, String ...).
	 * @see    LyFieldInfo#getValueClass
	 */
	public Object getValue(LyFieldInfo fieldInfo)
	{
		// Search in the last values.
		Object value = getLastValue(fieldInfo);

		if ((value != NOT_FOUND)
		        && (value != NOT_LOADED))
			return value;

		// Look for get method for the given fieldInfo
		fieldInfo = _classInfo.getRealFieldInfo(fieldInfo);

		Method getMethod = _classInfo.getGetMethod(fieldInfo);

		// If not found, call generic method
		if (getMethod == null)
		{
			value = get(fieldInfo);
			setLastValue(fieldInfo, value);
			return value;
		}

		// Otherwise, call it
		try
		{
			// SIR/FFT02353
			boolean wasAccessible = getMethod.isAccessible();

			if (!wasAccessible)
				getMethod.setAccessible(true);

			value = getMethod.invoke(this, (Object[])null);

			if (!wasAccessible)
				getMethod.setAccessible(wasAccessible);

			setLastValue(fieldInfo, value);
			return value;
		}
		catch (InvocationTargetException e)
		{
			LyEnvironment environment = getEnvironment();

			if (fieldInfo.hasMark(LyFieldInfo.MARK_NAME)
					|| fieldInfo.hasMark(LyFieldInfo.MARK_ID))
				// To avoid loop
				environment.logError("BF AppEngine : exception caught on getValue "
					+ e + " with object " + hashCode() + " of class info \"" + _classInfo
					+ "\" while accessing to \"" + fieldInfo + "\"");
			else
				environment.logError("BF AppEngine : exception caught on getValue "
					+ e + " with object " + this + " of class info \"" + _classInfo
					+ "\" while accessing to \"" + fieldInfo + "\"");

			environment.writeException(e.getTargetException());
			setLastValue(fieldInfo, null);
			return null;
		}
		catch (Exception e)
		{
			LyEnvironment environment = getEnvironment();

			if (fieldInfo.hasMark(LyFieldInfo.MARK_NAME)
					|| fieldInfo.hasMark(LyFieldInfo.MARK_ID))
				// To avoid loop
				environment.logError("BF AppEngine : exception caught on getValue "
					+ e + " with object " + hashCode() + " of class info \"" + _classInfo
					+ "\" while accessing to \"" + fieldInfo + "\"");
			else
				environment.logError("BF AppEngine : exception caught on getValue "
					+ e + " with object " + this + " of class info \"" + _classInfo
					+ "\" while accessing to \"" + fieldInfo + "\"");

			environment.writeException(e);
			setLastValue(fieldInfo, null);
			return null;
		}
	} // end method getValue

	/**
	 * Fetchs the value from this object corresponding to the given fieldId.
	 *
	 * @param  fieldId Id of the associated fieldInfo. Must be a valid fieldInfo for the class info
	 *                 of this object.
	 * @return Value of the field info for this object. The class of the returned value depends on
	 *         the type of the fieldInfo (Integer, Date, String ...).
	 * @throws IllegalArgumentException If the field id doesn't correspond to a field info in the
	 *                                  class info of this object.
	 */
	public final Object getValue(String fieldId)
	{
		// Retrieve the field info that corresponds to the given field id
		LyFieldInfo fieldInfo = _classInfo.getFieldInfo(fieldId);

		if (fieldInfo == null)
			throw new IllegalArgumentException("BF AppEngine : no field info with id "
				+ fieldId + " found on class info " + _classInfo);

		return getValue(fieldInfo);
	}

	/**
	 * Gets the field value from this object for the given field info.
	 *
	 * @param  fieldId Id of the associated fieldInfo. Must be a valid fieldInfo for the class info
	 *                 of this object.
	 * @return Field value of the given field info for this object. The class of the returned value
	 *         depends on the type of the fieldInfo (Integer, Date, String ...).
	 */
	public LyValue getFieldValue(String fieldId)
	{
		// Retrieve the field info that corresponds to the given field id
		LyFieldInfo fieldInfo = _classInfo.getFieldInfo(fieldId);

		return getFieldValue(fieldInfo);
	}

	/**
	 * Gets the field value from this object for the given field info.
	 *
	 * @return Field value of the given field info for this object. The class of the returned value
	 *         depends on the type of the fieldInfo (Integer, Date, String ...).
	 */
	public LyValue getFieldValue(LyFieldInfo fieldInfo)
	{
		Object value;

		fieldInfo = _classInfo.getRealFieldInfo(fieldInfo);
		value = getValue(fieldInfo);

		return fieldInfo.buildFieldValue(value);
	}

	/**
	 * Get choice value from this object corresponding to the given choice field.
	 *
	 * @param  fieldId Id of the associated fieldInfo. Must be a valid fieldInfo for the class info
	 *                 of this object.
	 * @return Choice value corresponding to the given choice field for this object.
	 */
	public final LyChoiceValue getChoiceValue(String fieldId)
	{
		// Retrieve the field info that corresponds to the given field id
		LyChoiceFieldInfo fieldInfo = (LyChoiceFieldInfo)_classInfo.getFieldInfo(fieldId);

		return getChoiceValue(fieldInfo);
	}

	/**
	 * Get choice value from this object corresponding to the given choice field.
	 *
	 * @param  fieldInfo Choice field info for which a value is requested.
	 * @return Choice value corresponding to the given choice field for this object.
	 */
	public final LyChoiceValue getChoiceValue(LyChoiceFieldInfo fieldInfo)
	{
		return (LyChoiceValue)getFieldValue(fieldInfo);
	}

	/**
	 * Get choice value from this object corresponding to the given choice field.
	 *
	 * @param  fieldId Id of the associated fieldInfo. Must be a valid fieldInfo for the class info
	 *                 of this object.
	 * @return Choice value corresponding to the given choice field for this object.
	 */
	public final LyBooleanValue getBooleanValue(String fieldId)
	{
		// Retrieve the field info that corresponds to the given field id
		LyChoiceFieldInfo fieldInfo = (LyChoiceFieldInfo)_classInfo.getFieldInfo(fieldId);

		return getBooleanValue(fieldInfo);
	}

	/**
	 * Get choice value from this object corresponding to the given choice field.
	 *
	 * @param  fieldInfo Choice field info for which a value is requested.
	 * @return Choice value corresponding to the given choice field for this object.
	 */
	public final LyBooleanValue getBooleanValue(LyChoiceFieldInfo fieldInfo)
	{
		return (LyBooleanValue)getFieldValue(fieldInfo);
	}

	/**
	 * Get date value from this object corresponding to the given date field.
	 *
	 * @param  fieldId Id of the associated fieldInfo. Must be a valid fieldInfo for the class info
	 *                 of this object.
	 * @return Date value corresponding to the given date field for this object.
	 */
	public final LyDateValue getDateValue(String fieldId)
	{
		// Retrieve the field info that corresponds to the given field id
		LyDateFieldInfo fieldInfo = (LyDateFieldInfo)_classInfo.getFieldInfo(fieldId);

		return getDateValue(fieldInfo);
	}

	/**
	 * Get date value from this object corresponding to the given date field.
	 *
	 * @param  fieldInfo Date field info for which a value is requested.
	 * @return Date value corresponding to the given date field for this object.
	 */
	public final LyDateValue getDateValue(LyDateFieldInfo fieldInfo)
	{
		return (LyDateValue)getFieldValue(fieldInfo);
	}

	/**
	 * Get file value from this object corresponding to the given file field.
	 *
	 * @param  fieldId Id of the associated fieldInfo. Must be a valid fieldInfo for the class info
	 *                 of this object.
	 * @return File value corresponding to the given file field for this object.
	 */
	public final LyFileValue getFileValue(String fieldId)
	{
		// Retrieve the field info that corresponds to the given field id
		LyFileFieldInfo fieldInfo = (LyFileFieldInfo)_classInfo.getFieldInfo(fieldId);

		return getFileValue(fieldInfo);
	}

	/**
	 * Get file value from this object corresponding to the given file field.
	 *
	 * @param  fieldInfo File field info for which a value is requested.
	 * @return File value corresponding to the given file field for this object.
	 */
	public final LyFileValue getFileValue(LyFileFieldInfo fieldInfo)
	{
		return (LyFileValue)getFieldValue(fieldInfo);
	}

	/**
	 * Get number value from this object corresponding to the given number field.
	 *
	 * @param  fieldId Id of the associated fieldInfo. Must be a valid fieldInfo for the class info
	 *                 of this object.
	 * @return Number value corresponding to the given number field for this object.
	 */
	public final LyNumberValue getNumberValue(String fieldId)
	{
		// Retrieve the field info that corresponds to the given field id
		LyNumberFieldInfo fieldInfo = (LyNumberFieldInfo)_classInfo.getFieldInfo(fieldId);

		return getNumberValue(fieldInfo);
	}

	/**
	 * Get number value from this object corresponding to the given number field.
	 *
	 * @param  fieldInfo Number field info for which a value is requested.
	 * @return Number value corresponding to the given number field for this object.
	 */
	public final LyNumberValue getNumberValue(LyNumberFieldInfo fieldInfo)
	{
		return (LyNumberValue)getFieldValue(fieldInfo);
	}

	/**
	 * Get relation value from this object corresponding to the given relation field.
	 *
	 * @param  fieldId Id of the associated fieldInfo. Must be a valid fieldInfo for the class info
	 *                 of this object.
	 * @return Relation value corresponding to the given relation field for this object.
	 */
	public final LyRelationValue getRelationValue(String fieldId)
	{
		// Retrieve the field info that corresponds to the given field id
		LyRelationFieldInfo fieldInfo = (LyRelationFieldInfo)_classInfo.getFieldInfo(fieldId);

		return getRelationValue(fieldInfo);
	}

	/**
	 * Get relation value from this object corresponding to the given relation field.
	 *
	 * @param  fieldInfo Relation field info for which a value is requested.
	 * @return Relation value corresponding to the given relation field for this object.
	 */
	public final LyRelationValue getRelationValue(LyRelationFieldInfo fieldInfo)
	{
		return (LyRelationValue)getFieldValue(fieldInfo);
	}

	/**
	 * Get text value from this object corresponding to the given text field.
	 *
	 * @param  fieldId Id of the associated fieldInfo. Must be a valid fieldInfo for the class info
	 *                 of this object.
	 * @return Text value corresponding to the given text field for this object.
	 */
	public final LyTextValue getTextValue(String fieldId)
	{
		// Retrieve the field info that corresponds to the given field id
		LyTextFieldInfo fieldInfo = (LyTextFieldInfo)_classInfo.getFieldInfo(fieldId);

		return getTextValue(fieldInfo);
	}

	/**
	 * Get text value from this object corresponding to the given text field.
	 *
	 * @param  fieldInfo Text field info for which a value is requested.
	 * @return Text value corresponding to the given text field for this object.
	 */
	public final LyTextValue getTextValue(LyTextFieldInfo fieldInfo)
	{
		return (LyTextValue)getFieldValue(fieldInfo);
	}

	/**
	 * Access a value of an object, but do not compute value. This only seeks into memory and is
	 * reserved for internal use only application should use get() method instead.
	 *
	 * @param  fieldInfo                Field info on which a value is requested.
	 * @return Value for the given field info on this object. The value is an instance of the class
	 *         returned by getValueClass method of the given field info.
	 * @param  IllegalArgumentException If the given field info is not contained in the class info
	 *                                  of this object.
	 * @param  IllegalStateException    If the binding of the given field info is not supported.
	 * @see    LyFieldInfo#getValueClass
	 * @see    #get(LyFieldInfo)
	 */
	final public Object getFromMemory(LyFieldInfo fieldInfo)
	{
        // Search in the last values.
        Object value = getLastValue(fieldInfo);

        if (value != NOT_FOUND)
        {
            if (value == NOT_LOADED)
                return null;
            else
                return value;
        }

        value = fromMemory(fieldInfo);

        if (value == NOT_FOUND)
            return null;

        setLastValue(fieldInfo, value);

        if (value == NOT_LOADED)
            return null;
        else
            return value;
	} // end method getFromMemory

	/**
     * Access a value of an object, but do not compute value. This only seeks into memory and is
     * reserved for internal use only application should use get() method instead.
     *
     * @param  fieldInfo                Field info on which a value is requested.
     * @return Value for the given field info on this object. The value is an instance of the class
     *         returned by getValueClass method of the given field info.
     * @param  IllegalArgumentException If the given field info is not contained in the class info
     *                                  of this object.
     * @param  IllegalStateException    If the binding of the given field info is not supported.
     * @see    LyFieldInfo#getValueClass
     * @see    #get(LyFieldInfo)
     */
    protected Object fromMemory(LyFieldInfo fieldInfo)
    {
        // Check first that this fieldInfo is contained in our classInfo
        fieldInfo = _classInfo.getRealFieldInfo(fieldInfo);

        LyClass        peerClass = _classInfo.getPeerClass();
        LyFieldBinding binding = fieldInfo.getBinding();
        LyValue        value;
        LyValueSet	   values;

        // Searching in the new values.
        values = _newValues;
        value = ((values == null) ? null : values.getFieldValue(fieldInfo));
        if (value != null)
            // Found.
            return value.getValue();

        values = _pendingValues;
        value = ((values == null) ? null : values.getFieldValue(fieldInfo));
        if (value != null)
            // Found
            return value.getValue();

        if (fieldInfo.isLocal()
        	|| _classInfo.isLocal())
        {
            // Searching for a local value.
            Object object;

            values = _localValues;
            value = ((values == null) ? null : values.getFieldValue(fieldInfo));
            object = ((value == null) ? null : value.getValue());

            if (value != null)
                return object;

            if (fieldInfo.isComputed())
                // Do *not* Compute
                return NOT_FOUND;
            else if (fieldInfo.isImported())
                // Do *not* Compute
                return NOT_FOUND;
            else
                return null;
        }
        else
        {
            // Searching for a non local value.
            Object peer = getPeer();

            if (peer == null)
                return NOT_LOADED;

            try
            {
                if (binding instanceof LyFieldsBinding)
                {
                    LyField[] fields = getPeerFields(fieldInfo);
                    int       i, n = fields.length;
                    Object[]  objects = new Object[n];
                    int       nNull = 0;

                    for (i = 0; i < n; i++)
                    {
                        objects[i] = peerClass.getValue(peer, fields[i]);
                        if (objects[i] == null)
                            nNull++;
                    }

                    if (nNull == n)
                        return null;

                    return fieldInfo.getValueFromObject(objects, this);
                }
                else
                {
                    LyField field = getPeerField(fieldInfo);
                    Object  object = peerClass.getValue(peer, field);

                    return fieldInfo.getValueFromObject(object, this);
                }
            }
            catch (LyValueNotLoadedException e)
            {
                return NOT_LOADED;
            }
        }
    } // end method fromMemory

	/**
	 * Internal method to get a local value. This method compute the value if it is not in
	 * _localValues list.
	 *
	 * @param  fieldInfo The local fieldInfo
	 * @return The local value for the given field info.
	 */
	public final Object getLocalValue(LyFieldInfo fieldInfo)
	{
		return getLocalValue(fieldInfo, true);
	}

	/**
	 * Internal method to get a local value. This method compute the value if it is not in
	 * _localValues list.
	 *
	 * @param  fieldInfo The local fieldInfo
	 * @param  compute   Allow computing of computed values.
	 * @return The local value for the given field info.
	 */
	public Object getLocalValue(LyFieldInfo fieldInfo, boolean compute)
	{
		Object  	object;
		LyValue 	value;
		LyValueSet	localValues = _localValues;

		value = ((localValues == null) ? null : localValues.getFieldValue(fieldInfo));
		object = ((value == null) ? null : value.getValue());

		if (value != null)
			return object;

		if (fieldInfo.isComputed())
		{
			if (!compute)
				return null;

			object = fieldInfo.compute(this);
			return object;
		}
		else if (fieldInfo.isImported())
			return getImportValue(fieldInfo);
		else if (fieldInfo.hasMark(LyFieldInfo.MARK_NOT_NULL))
		{
			object = fieldInfo.getDefaultValue();
			if ((object != null)
					&& (object instanceof String)
					&& (((String)object).charAt(0) == '$'))
				object = null;
			else
				addLocalValue(fieldInfo, object);
			return object;
		}
		else
			return null;
	} // end method getLocalValue

	/**
	 * Adds a local value to this object.
	 * @param fieldInfo Field info for which a value is added.
	 * @param value	    Value of the field info.
	 */
	protected synchronized void addLocalValue(LyFieldInfo fieldInfo, Object value)
	{
		if (_localValues == null)
			_localValues = new LyValueSet();
		_localValues.addValue(fieldInfo, value);
	}

	/**
	 * Internal method to set a local value. Should *not be used* by the application (internal use
	 * only)
	 *
	 * @param fieldInfo The local fieldInfo
	 * @param value     The local value for the given field info.
	 */
	public void setLocalValue(LyFieldInfo fieldInfo, Object value)
	{
		if ((!fieldInfo.isLocal()) && (!_classInfo.isLocal()))
			return;

		fieldInfo = _classInfo.getRealFieldInfo(fieldInfo);

		// Keep old value
		LyValueSet	localValues = _localValues;
		Object 		oldValue = ((localValues == null) ? null : localValues.getFieldValue(fieldInfo));

		if (value instanceof LyValue)
			value = ((LyValue)value).getValue();
		if (oldValue instanceof LyValue)
			oldValue = ((LyValue)oldValue).getValue();

		// Extract the right value from the given one
		value = fieldInfo.processValue(value);

		if (fieldInfo.equals(value, oldValue))
			return;

		addLocalValue(fieldInfo, value);

		// Notify changes with old values
		LyValueSet oldValues = new LyValueSet(false);

		oldValues.addValue(fieldInfo, oldValue);
		refresh(oldValues);
	} // end method setLocalValue

    /**
     * Remove local value for given field info.
     * @param fieldInfo Local field info for which value is removed.
     * @return Old value, null if not found or if neither field nor class are local.
     */
    public Object removeLocalValue(LyFieldInfo fieldInfo)
    {
        if (!fieldInfo.isLocal()
                && !_classInfo.isLocal())
            return null;

        fieldInfo = _classInfo.getRealFieldInfo(fieldInfo);

		LyValueSet	localValues = _localValues;

        if (localValues != null)
            return localValues.removeValue(fieldInfo);
        else
            return null;
    }

	/**
	 * Default internal method to access a value on this object from a field info.
	 *
	 * @param  fieldInfo                Field info on which a value is requested.
	 * @return Value for the given field info on this object. The value is an instance of the class
	 *         returned by getValueClass method of the given field info.
	 * @param  IllegalArgumentException If the given field info is not contained in the class info
	 *                                  of this object.
	 * @param  IllegalStateException    If the binding of the given field info is not supported.
	 * @see    LyFieldInfo#getValueClass
	 */
	protected Object get(LyFieldInfo fieldInfo)
	{
		return get(fieldInfo, false, false);
	}

	private Object get(LyFieldInfo fieldInfo, boolean skipNewValue, boolean skipPendingValue)
	{
		// Check first that this fieldInfo is contained in our classInfo
		fieldInfo = _classInfo.getRealFieldInfo(fieldInfo);

		LyValue 	value;
		LyValueSet	values;

		if (!skipNewValue)
		{
			// Searching in the new values.
			values = _newValues;
			value = ((values == null) ? null : values.getFieldValue(fieldInfo));
			if (value != null)
				return value.getValue();
		}

	    if (!skipPendingValue)
	    {
			values = _pendingValues;
    		value = ((values == null) ? null : values.getFieldValue(fieldInfo));
    		if (value != null)
    			return value.getValue();
	    }

		if (fieldInfo.isLocal() || _classInfo.isLocal())
			return getLocalValue(fieldInfo);
		else
		{
			// Searching for a non local value.
			Object peer = getPeer();

			if (peer == null)
				return null;

			LyFieldBinding binding = fieldInfo.getBinding();
			LyClass		   peerClass = _classInfo.getPeerClass();

			try
			{
				if (binding instanceof LyFieldsBinding)
				{
					LyField[] fields = getPeerFields(fieldInfo);
					int		  i, n = fields.length;
					Object[]  objects = new Object[n];
					int		  nNull = 0;

					for (i = 0; i < n; i++)
					{
						objects[i] = peerClass.getValue(peer, fields[i]);
						if (objects[i] == null)
							nNull++;
					}

					if (nNull == n)
						return null;

					return fieldInfo.getValueFromObject(objects, this);
				}
				else
				{
					LyField field = getPeerField(fieldInfo);
					Object  object = peerClass.getValue(peer, field);
					Object  objectValue = fieldInfo.getValueFromObject(object, this);

					return objectValue;
				}
			}
			catch (LyValueNotLoadedException e)
			{
				LyEnvironment environment = getEnvironment();

				if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
					environment.logTrace("BF AppEngine : reposting a request for object "
						+ hashCode() + " of class info \"" + _classInfo
						+ "\" because field \"" + fieldInfo
						+ "\" is not loaded on this object, an improvement could consist to add 'load' mark to the indicated field.");

				if (fieldInfo.hasMark(LyFieldInfo.MARK_ID))
					throw new IllegalStateException("BF AppEngine : error found on object "
						+ hashCode() + " of class info \"" + _classInfo
						+ "\", field \"" + fieldInfo
						+ " has ID mark and is not loaded");

				LyObjectSource source = getObjectSource();
				boolean		   synchronous = source.isSynchronous();

				source.setSynchronous(true);

				// Load all fields except blobs if not necessary
				Vector<LyFieldInfo> fields = _classInfo.getFields();
				int					i, n = fields.size();
				LyFieldInfo			field;

				for (i = n -1; i >= 0; i--)
				{
					field = fields.get(i);
					if ((field != fieldInfo)
							&& (field instanceof LyFileFieldInfo)
							&& (field.getBinding() != null)
							&& field.getBinding().getPeerType().equals("BLOB"))
						fields.remove(i);
				}

				LyResponse  		response = source.loadObject(getObjectId(), fields, null, null);

				if (response.getStatus() == LyResponse.Status.COMPLETED)
				    setFullyLoaded(true);

				source.setSynchronous(synchronous);
				return getFromMemory(fieldInfo);
			}
		}
	} // end method get

	/**
	 * Internal method to get the value from an import.
	 */
	Object getImportValue(LyFieldInfo fieldInfo)
	{
		LyFieldInfoImport   fieldImport = fieldInfo.getImport();
		LyRelationFieldInfo relation = fieldImport.getRelation();
		LyRelationValue     relationValue = getRelationValue(relation);

		return fieldImport.getImportValue(fieldInfo, relationValue);
	}

	/**
	 * Internal method to access a value on this object from a field info id. Shortcut for inherited
	 * classes.
	 *
	 * @param  fieldId                  Field id corresponding to the field info on which a value is
	 *                                  requested.
	 * @return Value for the given field info on this object. The value is an instance of the class
	 *         returned by getValueClass method of the field info.
	 * @param  IllegalArgumentException If the field id doesn't correspond to a field info in the
	 *                                  class info of this object.
	 * @see    LyFieldInfo#getValueClass
	 * @see    LyObject#get(LyFieldInfo)
	 */
	protected final Object get(String fieldId)
	{
		LyFieldInfo fieldInfo = _classInfo.getFieldInfo(fieldId);

		if (fieldInfo == null)
		{
			LyClassInfo superClass = _classInfo.getSuperClass();

			if (superClass != null)
				fieldInfo = superClass.getFieldInfo(fieldId);
			if (fieldInfo == null)
				throw new IllegalArgumentException("BF AppEngine : no field info with id "
					+ fieldId + " found on class info " + _classInfo);
		}

		return get(fieldInfo);
	}

	/**
	 * Fetchs the complete set of values for this object.
	 * @fromMemory true to get all values from memory only
	 * @return Values set containing values for all fieldInfos of the class info to which this
	 *         object belongs.
	 */
	public LyValueSet getValues(boolean fromMemory)
	{
		LyValueSet  valueSet = new LyValueSet(false);
		LyFieldInfo fieldInfo;
		Object	    value;
		int		    i, n = _classInfo.getFieldInfoCount();

		// Go through the fieldInfos and build the valueSet
		for (i = 0; i < n; i++)
		{
			fieldInfo = _classInfo.getFieldInfo(i);
			if (fromMemory)
				value = getFromMemory(fieldInfo);
			else
				value = getValue(fieldInfo);
			valueSet.addValue(fieldInfo, value);
		}

		return valueSet;
	}

	/**
	 * Fetchs the complete set of values for this object.
	 *
	 * @return Values set containing values for all fieldInfos of the class info to which this
	 *         object belongs.
	 */
	public LyValueSet getValues()
	{
		return getValues(false);
	}

	/**
     * Fetchs the set of values for the given field infos on this object.
     *
     * @param  fieldInfos Field infos on which values are requested.
     * @return Values set containing values for all given fields of the class info to which this
     *         object belongs.
     */
    public LyValueSet getValues(LyInfoList<LyFieldInfo> fieldInfos)
    {
        return getValues(fieldInfos, false);
    }

	/**
	 * Fetchs the set of values for the given field infos on this object.
	 *
	 * @param  fieldInfos Field infos on which values are requested.
	 * @return Values set containing values for all given fields of the class info to which this
	 *         object belongs.
	 */
	protected LyValueSet getValues(LyInfoList<LyFieldInfo> fieldInfos, boolean fromMemory)
	{
		if (fieldInfos == null)
			return null;

		LyValueSet  valueSet = new LyValueSet(false);
		LyFieldInfo fieldInfo;
		Object	    value;
		int		    i, n = fieldInfos.getSize();

		// Go through the fieldInfos and build the valueSet
		for (i = 0; i < n; i++)
		{
			fieldInfo = fieldInfos.getInfo(i);
	        if (fromMemory)
                value = getFromMemory(fieldInfo);
            else
                value = getValue(fieldInfo);
			valueSet.addValue(fieldInfo, value);
		}

		return valueSet;
	}

	/**
	 * Fetchs the complete set of local values for this object.
	 *
	 * @return Values set containing all local values for this object.
	 */
	public LyValueSet getLocalValues()
	{
		LyValueSet	localValues = _localValues;

		return (localValues == null ? null : new LyValueSet(localValues));
	}

	/**
	 * Fetchs the complete set of new values for this object.
	 *
	 * @return Values set containing all new values for this object.
	 */
	public LyValueSet getNewValues()
	{
		LyValueSet	newValues = new LyValueSet();

		newValues.addValues(_newValues);
		return newValues;
	}

	/**
	 * Fetchs the complete set of pending values for this object.
	 *
	 * @return Values set containing all pending values for this object.
	 */
	public LyValueSet getPendingValues()
	{
		LyValueSet	pendingValues = new LyValueSet();

		pendingValues.addValues(_pendingValues);
		return pendingValues;
	}

	/**
	 * Fetchs the complete set of old values for this object (values for which a new value exists).
	 *
	 * @return Values set containing all old values fo this object.
	 */
	public LyValueSet getOldValues()
	{
		if (_newValues == null)
			return null;

		LyValueSet  result = getNewValues();
		LyValue     newValue;
		LyFieldInfo fieldInfo;
		Object	    value;
		int		    i, n = result.getSize();

		for (i = 0; i < n; i++)
		{
			newValue = result.getValue(i);
			fieldInfo = (LyFieldInfo)newValue.getField();

			value = get(fieldInfo, true, false);

			newValue = fieldInfo.buildFieldValue(value);
			result.setValueAt(newValue, i);
		}

		return result;
	}

	/**
     * Gets old value (excluding new value) on this object for given field info.
     * @param fieldInfo Field info for which the old value is requested.
     * @return The old value of the specified field.
     */
    public final Object getOldValue(LyFieldInfo fieldInfo)
    {
        return get(fieldInfo, true, true);
    }

	/**
	 * Retrieves the associated peer (physical object that really stores values for non local
	 * fields).
	 *
	 * @return Peer associated to this object, null if the object belongs to a local class info
	 *         (class with no persistence) or if the object was not yet created on the physical
	 *         layer.
	 */
	public Object getPeer()
	{
		return _peer;
	}

	/**
	 * Gets the physical field that corresponds to the given field info. This method should be used
	 * only when a non multiple binding is specified on that field info.
	 *
	 * @param  fieldInfo Field info for which the physical field is requested.
	 * @return Returns the physical field, null if an error occurs.
	 */
	protected final LyField getPeerField(LyFieldInfo fieldInfo)
	{
		LyClassBinding classBinding = _classInfo.getBinding();

		return ((classBinding == null) ? null : classBinding.getPeerField(fieldInfo));
	}

	/**
	 * Gets the array of physical fields that correspond to the given field info. This method should
	 * be used only when a multiple binding is specified on that field info.
	 *
	 * @param  fieldInfo Field info for which physical fields are requested.
	 * @return Returns an array of physical fields, null if an error occurs.
	 */
	protected final LyField[] getPeerFields(LyFieldInfo fieldInfo)
	{
		LyClassBinding classBinding = _classInfo.getBinding();

		return ((classBinding == null) ? null : classBinding.getPeerFields(fieldInfo));
	}

	/**
	 * Method to reset last accessed field and value when peer object is modified. This method is
	 * used by LyObjectSource when an object is modified.
	 *
	 * @see LyObjectSource#objectModified
	 */
	void touch()
	{
		Object[]	lastValues = _lastValues;

	    if (lastValues != null)
        {
            int      i, n = lastValues.length;

            for (i = 0; i < n; i++)
                lastValues[i] = NOT_FOUND;
        }

	    _lastValues = lastValues;
		_name = null;
	}

	/**
     * Internal method to get the last accessed value for a fieldinfo.
     *
     * @param fieldInfo Last accessed field info.
     * @return Value for given field info, NOT_FOUND if not found.
     */
    protected Object getLastValue(LyFieldInfo fieldInfo)
    {
        // Search in the last values.
        int     index = _classInfo.getIndex(fieldInfo);

        if (index < 0)
            return NOT_FOUND;

		Object[]	lastValues = _lastValues;

        if (lastValues == null)
            return NOT_FOUND;

        // Search in the last values.
        return lastValues[index];
    }

    /**
     * Internal method to remove the last accessed value for a field info.
     *
     * @param fieldInfo Last accessed field info.
     * @return Value for given field info, NOT_FOUND if not found.
     */
    protected Object removeLastValue(LyFieldInfo fieldInfo)
    {
        int     index = _classInfo.getIndex(fieldInfo);

        if (index < 0)
            return NOT_FOUND;

		Object[]	lastValues = _lastValues;

        if (lastValues == null)
            return NOT_FOUND;

        Object  value = lastValues[index];

        lastValues[index] = NOT_FOUND;
        return value;
    }

	/**
	 * Internal method to modify the last accessed value.
	 *
	 * @param fieldInfo Last accessed field info.
	 * @param value     Value returned on this access.
	 */
	protected void setLastValue(LyFieldInfo fieldInfo, Object value)
	{
		if (fieldInfo == null)
		{
			touch();
			return;
		}

		if (isVirtualFieldInfo(fieldInfo))
			return; // Field is computed by Java code : Do not keep result in cache

        if (fieldInfo.isLocal()
                || _classInfo.isLocal())
            return; // Field is computed by Java code : Do not keep result in cache

		Object[]	lastValues = _lastValues;

		if (lastValues == null)
		{
		    int      i, n = _classInfo.getFieldInfoCount();

			lastValues = new Object[n];
			for (i = 0; i < n; i++)
			    lastValues[i] = NOT_FOUND;
		}

        int     index = _classInfo.getIndex(fieldInfo);

        if (index >= 0)
            lastValues[index] = value;
		_lastValues = lastValues;
	}

	/**
	 * Sets a value for this object on the fieldInfo corresponding to the given id.
	 *
	 * @param fieldId                  Id of the set field info.
	 * @param value                    Value to set on the corresponding field info. This value
	 *                                 should be an instance of the class returned by getValueClass
	 *                                 of the field info.
	 * @param IllegalArgumentException If the field id doesn't correspond to a field info in the
	 *                                 class info of this object.
	 * @see   LyFieldInfo#getValueClass
	 */
	public void setValue(String fieldId, Object value)
	{
		LyFieldInfo fieldInfo = _classInfo.getFieldInfo(fieldId);

		if (fieldInfo == null)
			throw new IllegalArgumentException("BF AppEngine : no field info with id "
				+ fieldId + " found on class info " + _classInfo);

		setValue(fieldInfo, value);
	}

	/**
	 * Sets a value for this object on the given fieldInfo.
	 *
	 * @param fieldInfo Field info to set.
	 * @param value     Value to set on that field. This value should be an instance of the class
	 *                  returned by getValueClass of the field info.
	 * @see   LyFieldInfo#getValueClass
	 */
	public void setValue(LyFieldInfo fieldInfo, Object value)
	{
		setValue(null, fieldInfo, value, DEFAULT_CONTROL, OPER_NEED_CONTEXT);
	}

	/**
	 * Sets a value for this object on the given fieldInfo. For internal use only.
	 *
	 * @param context   The modification context that sets this value.
	 * @param fieldInfo Field info to set.
	 * @param value     Value to set on that field. This value should be an instance of the class
	 *                  returned by getValueClass of the field info.
	 * @param control
	 * @param refresh
	 * @see   LyFieldInfo#getValueClass
	 */
	protected void setValue(LyContext context, LyFieldInfo fieldInfo, Object value,
		int control, int operFlags)
	{
        fieldInfo = _classInfo.getRealFieldInfo(fieldInfo);

		// Keep old value
		Object oldValue = (control != CONTROL_NONE ? getFromMemory(fieldInfo) : null);

		if (value instanceof LyValue)
			value = ((LyValue)value).getValue();

		// Extract the right value from the given one
		value = fieldInfo.processValue(value);

		if ((control != CONTROL_NONE)
		        && fieldInfo.equals(value, oldValue)
				&& !isVirtualFieldInfo(fieldInfo)
				&& ! fieldInfo.isComputed()) // if fieldInfo is computed, getFromMemory always return null
			return;

		// Do not encode the value to check changes because undo/redo must have
		// a change operation with an encoded value.

		if ((context == null) && ((operFlags & OPER_NEED_CONTEXT) > 0))
		{
			LyApplication application = getApplication();

			context = application.getCurrentContext();

			if (canUseContext(context) && (!isVirtualFieldInfo(fieldInfo)))
			{
				context.setValue(this, fieldInfo, value);
				return;
			}
			context = null;
		}

		if ((control != CONTROL_NONE)
			&& !fieldInfo.isLocal()
			&& !_classInfo.isLocal()
	        && hasNewValue(fieldInfo)
			&& !hasPendingValue(fieldInfo)
			&& fieldInfo.equals(value, get(fieldInfo, true, true))
			&& !isVirtualFieldInfo(fieldInfo))
		{
			removeNewValues(context, new LyInfoList<LyFieldInfo>(fieldInfo),
				(control & CONTROL_FIELD_CONTROLS), (operFlags & OPER_NEED_REFRESH));
			return;
		}

		// Remove last accessed values (fieldInfo and depends)
		if (_lastValues != null)
			touch();

		// Reset "class controlled" state
		setClassControlled(false);

		// Look for set method for the given fieldInfo
		Method setMethod = _classInfo.getSetMethod(fieldInfo);

		// If not found, call generic method
		if (setMethod == null)
			set(fieldInfo, value, control);
		else
		{
			// Otherwise, call it
			Object[] args = new Object[1];

			args[0] = value;

			try
			{
				// SIR/FFT02353
				boolean wasAccessible = setMethod.isAccessible();

				if (!wasAccessible)
					setMethod.setAccessible(true);

				setMethod.invoke(this, args);

				if (!wasAccessible)
					setMethod.setAccessible(wasAccessible);

				// If a field composing name is modified, reset name
				if (fieldInfo.hasMark(LyFieldInfo.MARK_NAME)
						|| (!_classInfo.isNameDifferentOfId()
							&& fieldInfo.hasMark(LyFieldInfo.MARK_ID)))
					touch();
			}
			catch (InvocationTargetException e)
			{
				LyEnvironment environment = getEnvironment();

				environment.logError("BF AppEngine : exception caught on setValue "
					+ e + " for object " + this + " of class info \"" + _classInfo
					+ "\" while setting \"" + fieldInfo + "\" with value " + value + "\"");
				environment.writeException(e.getTargetException());
			}
			catch (Exception e)
			{
				LyEnvironment environment = getEnvironment();

				environment.logError("BF AppEngine : exception caught on setValue "
					+ e + " for object " + this + " of class info \"" + _classInfo
					+ "\" while setting \"" + fieldInfo + "\" with value " + value + "\"");
				environment.writeException(e);
			}
		}

		// If fieldInfo is a struct, try to set parent of struct
		if ((fieldInfo instanceof LyStructFieldInfo)
				&& !(_classInfo.getParent() instanceof LyArrayFieldInfo))
		{
			if (oldValue instanceof LyObjectId)
			{
				LyObject structObject = ((LyObjectId)oldValue).getLogicalObject();

				if (structObject != null)
					structObject.setParent(null, null);
			}
			if (value instanceof LyObjectId)
			{
				LyObject structObject = ((LyObjectId)value).getLogicalObject();

				if (structObject != null)
					structObject.setParent(this, fieldInfo);
			}
		}
		// If fieldInfo is an array of structs, try to set parent of structs
		else if ((fieldInfo instanceof LyArrayFieldInfo)
				&& (((LyArrayFieldInfo)fieldInfo).getFieldInfo() instanceof LyStructFieldInfo))
		{
			if (oldValue instanceof LyObjectId[])
			{
				LyObjectId[] structs = (LyObjectId[])oldValue;
				int			 i, n = structs.length;
				LyObject     structObject;

				for (i = 0; i < n; i++)
				{
					structObject = structs[i].getLogicalObject();
					if (structObject != null)
						structObject.setParent(null, null);
				}
			}
			if (value instanceof LyObjectId[])
			{
				LyObjectId[] structs = (LyObjectId[])value;
				int			 i, n = structs.length;
				LyObject     structObject;

				for (i = 0; i < n; i++)
				{
					structObject = structs[i].getLogicalObject();
					if (structObject != null)
						structObject.setParent(this, fieldInfo);
				}
			}
		}
	} // end method setValue

	/**
	 * For internal use only.
	 *
	 * @param  context
	 * @param  relation
	 * @param  object
	 * @return
	 */
	boolean addRelation(LyContext context, LyRelationFieldInfo relation, LyObject object,
		int control, int operFlags)
	{
		if (isDeleted())
			return false;

		if ((context == null)
				&& ((operFlags & OPER_NEED_CONTEXT) > 0))
		{
			LyApplication application = getApplication();

			context = application.getCurrentContext();
			if (!canUseContext(context))
				context = null;
		}

		// FFT00964.
		if ((context != null) && (!canUseContext(context)))
			return true;

		LySimpleObjectList objects = null;

		objects = relation.getObjects(this);

		if ((objects != null) && objects.contains(object))
			return false;

		if (relation.isMultiple())
		{
			if (objects == null)
				objects = new LySimpleObjectList();
			objects.addObject(object);
			if (context != null)
				context.setValue(this, relation, objects);
			else
				setValue(null, relation, objects, control, operFlags);
		}
		else if (context != null)
			context.setValue(this, relation, object.getObjectId());
		else
			setValue(null, relation, object.getObjectId(), control, operFlags);

		return true;
	} // end method addRelation

	/**
	 * For internal use only.
	 *
	 * @param context
	 * @param relation
	 * @param object
	 */
	boolean removeRelation(LyContext context, LyRelationFieldInfo relation, LyObject object,
		int control, int operFlags)
	{
		if (isDeleted())
			return false;

		if ((context == null)
				&& ((operFlags & OPER_NEED_CONTEXT) > 0))
		{
			LyApplication application = getApplication();

			context = application.getCurrentContext();
			if (!canUseContext(context))
				context = null;
		}

		if ((context != null) && (!canUseContext(context)))
			return true;

		LySimpleObjectList objects = null;

		objects = relation.getObjects(this);

		if (relation.isMultiple())
		{
			if ((objects != null) && objects.contains(object))
				objects.removeObject(object);

			if ((objects != null) && (objects.getSize() > 0))
			{
				if (context != null)
					context.setValue(this, relation, objects);
				else
					setValue(null, relation, objects, control, operFlags);
			}
			else if (context != null)
				context.setValue(this, relation, null);
			else
				setValue(null, relation, null, control, operFlags);
		}
		else
		{
			// If the relation points to an unknown object we let
			// the pending relation.
			// If the relation points to the object to remove we must
			// set the object.
			if ((objects != null) && objects.contains(object))
			{
				if (hasNewValue(relation)
						&& !hasPendingValue(relation))
					removeNewValue(context, relation, control, OPER_NONE);
				else if (context != null)
					context.setValue(this, relation, null);
				else
					setValue(null, relation, null, control, operFlags);
			}
			else
			{
				// We must verify if the relation points to the object id.
				LyObjectId id = (LyObjectId)getValue(relation);
				LyObjectId objectId = object.getObjectId();

				if ((id != null) && (id.equals(objectId)))
				{
					if (hasNewValue(relation)
							&& !hasPendingValue(relation))
						removeNewValue(context, relation, control, OPER_NONE);
					else if (context != null)
						context.setValue(this, relation, null);
					else
						setValue(null, relation, null, control, operFlags);
				}
				else
					return false;
			}
		}

		return true;
	} // end method removeRelation

	/**
	 * Sets a field value on this object.
	 *
	 * @param value Field value to set on this object.
	 */
	public void setFieldValue(LyValue value)
	{
		if (value == null)
			throw new IllegalArgumentException("BF AppEngine : value parameter can't be null");

		setValue((LyFieldInfo)value.getField(), value.getValue());
	}

	/**
	 * Sets a value on this object for the given fieldInfo.
	 *
	 * @param fieldInfo                Field info for which a value is set on this object.
	 * @param value                    Value assigned on this object for the given field info. This
	 *                                 value must be an instance of the class returned by
	 *                                 getValueClass method of the given field info.
	 * @see   LyFieldInfo#getValueClass
	 * @param IllegalArgumentException If the given field info is not contained in the class info of
	 *                                 this object.
	 */
	protected void set(LyFieldInfo fieldInfo, Object value)
	{
		fieldInfo = _classInfo.getRealFieldInfo(fieldInfo);

		// Keep old value
		Object oldValue = getFromMemory(fieldInfo);

		if (value instanceof LyValue)
			value = ((LyValue)value).getValue();

		// Extract the right value from the given one
		value = fieldInfo.processValue(value);

		if (fieldInfo.equals(value, oldValue)
				&& !isVirtualFieldInfo(fieldInfo)
				&& !fieldInfo.isComputed())
			return;

		set(fieldInfo, value, DEFAULT_CONTROL & CONTROL_FIELD_CONTROLS);
	}

	/**
	 * Sets a value on this object for the given fieldInfo.
	 *
	 * @param fieldInfo                Field info for which a value is set on this object.
	 * @param value                    Value assigned on this object for the given field info. This
	 *                                 value must be an instance of the class returned by
	 *                                 getValueClass method of the given field info.
	 * @see   LyFieldInfo#getValueClass
	 * @param IllegalArgumentException If the given field info is not contained in the class info of
	 *                                 this object.
	 */
	protected void set(LyFieldInfo fieldInfo, Object value, int control)
	{
		// Trigger controls on fieldInfo
		if ((control & CONTROL_FIELD_CONTROLS) > 0)
			callFieldControls(fieldInfo, value, control, true);

		// Encode the value if needed
		if (!skipEncoding())
			value = fieldInfo.encodeValue(value);

		synchronized (this)
		{
			if (_newValues == null)
				_newValues = new LyValueSet();
			_newValues.addValue(fieldInfo, value);
		}

		LyValueSet	pendingValues = _pendingValues;

		if (pendingValues != null) // no need to check if the fieldInfo has a value in the list
			pendingValues.removeValue(fieldInfo);

		setLastValue(fieldInfo, value);

		if (control != CONTROL_NONE)
			controlParent(control & CONTROL_FIELD_CONTROLS);

		// If a field composing name is modified, reset name
		if (fieldInfo.hasMark(LyFieldInfo.MARK_NAME)
				|| (!_classInfo.isNameDifferentOfId()
					&& fieldInfo.hasMark(LyFieldInfo.MARK_ID)))
			touch();
	}

	/**
	 * Sets a value on this object for the field info with the given id. This method is a shortcut
	 * for inherited classes.
	 *
	 * @param fieldId                  Id of the fieldInfo for which a value is set.
	 * @param value                    Value assigned on this object for the given field info. This
	 *                                 value must be an instance of the class returned by
	 *                                 getValueClass method of the given field info.
	 * @param IllegalArgumentException If the field id doesn't correspond to a field info in the
	 *                                 class info of this object.
	 * @see   LyFieldInfo#getValueClass
	 */
	protected final void set(String fieldId, Object value)
	{
		LyFieldInfo fieldInfo = _classInfo.getFieldInfo(fieldId);

		if (fieldInfo == null)
			throw new IllegalArgumentException("BF AppEngine : no field info with id "
				+ fieldId + " found on class info " + _classInfo);

		set(fieldInfo, value);
	}

	/**
	 * Set values on this object with the given values set.
	 *
	 * @param values Values set containing field values to set on this object.
	 */
	public void setValues(LyValueSet values)
	{
		setValues(null, values, DEFAULT_CONTROL, OPER_NEED_CONTEXT);
	}

	/**
	 * Set values on this object with the given values set. For internal use only.
	 *
	 * @param context The modification context doing this modification.
	 * @param values  Values set containing field values to set on this object.
	 * @param control
	 * @param refresh
	 */
	protected void setValues(LyContext context, LyValueSet values, int control, int operFlags)
	{
		if (values == null)
			return;

		if ((context == null) && ((operFlags & OPER_NEED_CONTEXT) > 0))
		{
			LyApplication application = getApplication();

			context = application.getCurrentContext();
			if (canUseContext(context))
			{
				context.setValues(this, values);
				return;
			}
			context = null;
		}

		LyFieldInfo fieldInfo;
		LyValue     value;
		int		    i, n = values.getSize();

		for (i = 0; i < n; i++)
		{
			value = values.getValue(i);
			fieldInfo = (LyFieldInfo)value.getField();
			setValue(context, fieldInfo, value.getValue(), control, operFlags);
		}
	} // end method setValues

	/**
	 * Sets the peer (physical layer) of this object. This method should normally not be used except
	 * by the object source when creating a logical object from its peer. Once associated, the peer
	 * can't be changed anymore.
	 *
	 * @exception IllegalStateException    If this object has already a peer.
	 * @exception IllegalArgumentException If the class of the given peer is not the expected
	 *                                     physical object class.
	 */
	public void setPeer(Object peer)
	{
		LyClass peerClass = _classInfo.getPeerClass();

		if (_peer != null)
			throw new IllegalStateException("BF AppEngine : changing peer for an object is not allowed");

		if (!peerClass.getObjectClass().isAssignableFrom(peer.getClass()))
			throw new IllegalArgumentException("BF AppEngine : invalid class peer "
				+ peer.getClass() + "(expected class " + peerClass.getObjectClass() + ")");

		_peer = peer;
	}

	void reset(boolean resetPeer)
	{
		reset(null, resetPeer);
	}

	void reset(LyContext context, boolean resetPeer)
	{
		// Clean up object
		setLastValue(null, null);
		_parentReference = null;
		_errors = null;
		_errorLevel = LEVEL_OK;

		// Get values before resetting peer
		LyValueSet values = getValues();

		// Reset peer
		if (resetPeer)
		{
			if (!isDeleted())
				throw new IllegalStateException(
					"BF AppEngine : reset for a not deleted object is not allowed");

			_peer = null;
		}

		setDeleted(false);
		setLocallyDeleted(false);

		skipEncoding(true);
		//setValues(values);
		setValues(context, values, DEFAULT_CONTROL, OPER_NEED_CONTEXT);
		skipEncoding(false);

	} // end method reset

	/**
	 * Creates an object : to commit its creation on the underlying data provider. Values for this
	 * object should have been given previously using setFieldValue or setValue methods.
	 *
	 * @return    Returns a response indicating status of this creation.
	 * @exception IllegalStateException If the class info for this object is abstract.
	 * @see       LyObject#setFieldValue
	 * @see       LyObject#setValue
	 */
	public LyResponse create()
	{
		if (_classInfo.isTemporary())
			return create(true);
		else
			return create(false);
	}

	/**
	 * Creates an object : to commit its creation on the underlying data provider. Values for this
	 * object should have been given previously using setFieldValue or setValue methods.
	 *
	 * @param     local Indicates if the creation must be local or not.
	 * @return    Returns a response indicating status of this creation.
	 * @exception IllegalStateException If the class info for this object is abstract.
	 * @see       LyObject#setFieldValue
	 * @see       LyObject#setValue
	 */
	public LyResponse create(boolean local)
	{
		if (!local)
		{
			LyApplication application = getApplication();
			LyContext     context = application.getCurrentContext();

			if (canUseContext(context))
				return context.createObject(this);
		}

		int operFlags = (local ? (OPER_LOCAL | OPER_NEED_CONTEXT | OPER_NEED_REFRESH)
							   : OPER_NEED_REFRESH);

		return create(null, operFlags, DEFAULT_CONTROL);
	}

	/**
	 * Creates an object : to commit its creation on the underlying data provider. Values for this
	 * object should have been given previously using setFieldValue or setValue methods.
	 *
	 * @param     context   The modification context doing the creation.
	 * @param     local     Indicates if the creation must be local or not.
	 * @param     refresh   Indicates if the creation may be notified or not.
	 * @param     control   Indicates if class controls are applied.
	 * @return    Returns a response indicating status of this creation.
	 * @exception IllegalStateException If the class info for this object is abstract.
	 * @see       LyObject#setFieldValue
	 * @see       LyObject#setValue
	 */
	protected LyResponse create(LyContext context, int operFlags, int control)
	{
		if (_classInfo.isAbstract())
			throw new IllegalStateException(
				"BF AppEngine : can't create an object for the abstract class info " + _classInfo);

		if (isTemporary())
			throw new IllegalStateException("BF AppEngine : can't create atemporary object " + getId());

		skipEncoding(false);

		boolean checkError = false;

		// Trigger controls attached to classInfo
		if ((((operFlags & OPER_LOCAL) <= 0) || (_newValues != null))
				&& ((control & CONTROL_CLASS_CONTROLS) > 0))
			callClassControls(_newValues, control);
		// SIR/FFT02367
		else if ((_newValues != null)
			&& ((control & CONTROL_ID) > 0))
		{
			// If we already have some values, we must force the ID control to ensure that the object can be created
			// If after the test, the object is in error, we don't add it to the cache and we return a response with an error.
			callClassControls(_newValues, CONTROL_ID);
			checkError = true;
		}

		if ((checkError || ((operFlags & OPER_LOCAL) <= 0))
				&& (getErrorLevel() == LEVEL_ERROR))
		{
			LyResponse response = new LyResponse(LyResponse.Operation.CREATE, _classInfo);
			LyError    error = getLastError(LEVEL_ERROR);

			if (error != null)
				response.setError(error.getMessage(_classInfo.getEnvironment()));
			response.setLocal(((operFlags & OPER_LOCAL) > 0));
			response.setStatus(LyResponse.Status.ERROR);
			return response;
		}

		// Create object on source
		LyObjectSource objectSource = getObjectSource();
		LyResponse     response = objectSource.createObject(this, null,
				((operFlags & OPER_LOCAL) > 0), ((operFlags & OPER_NEED_REFRESH) > 0));

		// We may have left some errors not notified during setValue() on an object which is not yet created
		if ((control != CONTROL_NONE)
				&& (getErrorsCount() > 0))
			control((LyValueSet)null, control);

		return response;
	} // end method create

	/**
	 * Sets an object : to commit its modification on the underlying data provider. Values for this
	 * object should have been given previously using setFieldValue or setValue methods.
	 *
	 * @return    Returns a response indicating status of this modification.
	 * @exception IllegalStateException If the peer (physical object) doesn't exist and the class
	 *                                  info for this object is not local.
	 * @see       LyObject#setFieldValue
	 * @see       LyObject#setValue
	 */
	public LyResponse set()
	{
		return set(false);
	}

	/**
	 * Sets an object : to commit its modification on the underlying data provider. Values for this
	 * object should have been given previously using setFieldValue or setValue methods.
	 *
	 * @param     local Indicates if the modification must be local or not.
	 * @return    Returns a response indicating status of this modification.
	 * @exception IllegalStateException If the peer (physical object) doesn't exist and the class
	 *                                  info for this object is not local.
	 * @see       LyObject#setFieldValue
	 * @see       LyObject#setValue
	 */
	public LyResponse set(boolean local)
	{
		if (!local)
		{
			LyApplication application = getApplication();
			LyContext     context = application.getCurrentContext();

			if (canUseContext(context))
				return context.setObject(this);
		}

		int operFlags = (local ? (OPER_LOCAL | OPER_NEED_CONTEXT | OPER_NEED_REFRESH)
							   : OPER_NEED_REFRESH);

		return set(null, operFlags, DEFAULT_CONTROL);
	}

	/**
	 * Sets an object : to commit its modification on the underlying data provider. Values for this
	 * object should have been given previously using setFieldValue or setValue methods. For
	 * internal use only.
	 *
	 * @param     context   The modification context doing this modification.
	 * @param     operFlags Indicates the various operation flags for this operation.
	 * @param     control   Indicates the control that should be executed during this operation.
	 * @return    Returns a response indicating status of this modification.
	 * @exception IllegalStateException If the peer (physical object) doesn't exist and the class
	 *                                  info for this object is not local.
	 * @see       LyObject#setFieldValue
	 * @see       LyObject#setValue
	 */
	protected LyResponse set(LyContext context, int operFlags, int control)
	{
		skipEncoding(false);

		if ((_newValues == null)
			&& ((_pendingValues == null) || ((operFlags & OPER_LOCAL) > 0)))
		{
			// Return a completed response
			LyResponse response = new LyResponse(LyResponse.Operation.SET, _classInfo);

			response.setLocal(((operFlags & OPER_LOCAL) > 0));
			response.setStatus(LyResponse.Status.COMPLETED);
			return response;
		}

		if ((_peer == null)
				&& !_classInfo.isLocal())
			operFlags |= OPER_LOCAL;

		// Trigger controls attached to classInfo
		if ((control & CONTROL_CLASS_CONTROLS) > 0)
			callClassControls(_newValues, control);

		if (((operFlags & OPER_LOCAL) <= 0) && (getErrorLevel() == LEVEL_ERROR))
		{
			LyResponse response = new LyResponse(LyResponse.Operation.SET, _classInfo);
			LyError    error = getLastError(LEVEL_ERROR);

			if (error != null)
				response.setError(error.getMessage(_classInfo.getEnvironment()));
			response.setLocal(((operFlags & OPER_LOCAL) > 0));
			response.setStatus(LyResponse.Status.ERROR);
			return response;
		}

		// Set object on source
		LyObjectSource objectSource = getObjectSource();
		LyResponse     response = objectSource.setObject(this, null, ((operFlags & OPER_LOCAL) > 0),
				((operFlags & OPER_NEED_REFRESH) > 0), false);

		if ((context == null)
				&& ((operFlags & OPER_NEED_CONTEXT) > 0))
		{
			LyApplication application = getApplication();

			context = application.getCurrentContext();
			if (!canUseContext(context))
				context = null;
		}

		return response;
	} // end method set

	/**
	 * Deletes an object : to commit its deletion on the underlying data provider.
	 *
	 * @return    Returns a response indicating status of this deletion.
	 * @exception IllegalStateException If the peer (physical object) doesn't exist and the class
	 *                                  info for this object is not local.
	 */
	public LyResponse delete()
	{
		if (getClassInfo().isTemporary())
			return delete(true);
		else
			return delete(false);
	}

	/**
	 * Deletes an object : to commit its deletion on the underlying data provider.
	 *
	 * @param     local Indicates if the deletion is local or not
	 * @return    Returns a response indicating status of this deletion.
	 * @exception IllegalStateException If the peer (physical object) doesn't exist and the class
	 *                                  info for this object is not local.
	 */
	public LyResponse delete(boolean local)
	{
		if (!local)
		{
			LyApplication application = getApplication();
			LyContext     context = application.getCurrentContext();

			if (canUseContext(context))
				return context.deleteObject(this);
		}

		int operFlags = (local ? (OPER_LOCAL | OPER_NEED_CONTEXT | OPER_NEED_REFRESH)
							   : OPER_NEED_REFRESH);

		return delete(null, operFlags, DEFAULT_CONTROL);
	}

	/**
	 * Deletes an object : to commit its deletion on the underlying data provider. For internal use
	 * only.
	 *
	 * @param     context   The modification context doing this deletion.
	 * @param     operFlags Indicates the various operation flag for this operation.
	 * @param     control   Indicates the control that should be executed during this operation.
	 * @return    Returns a response indicating status of this deletion.
	 * @exception IllegalStateException If the peer (physical object) doesn't exist and the class
	 *                                  info for this object is not local.
	 */
	protected LyResponse delete(LyContext context, int operFlags, int control)
	{
		if (_peer == null)
			operFlags |= OPER_LOCAL;

		LyObjectSource objectSource = getObjectSource();
		LyResponse     response;

		if ((context == null)
				&& ((operFlags & OPER_NEED_CONTEXT) > 0))
		{
			LyApplication application = getApplication();

			context = application.getCurrentContext();
			if (!canUseContext(context))
				context = null;
		}

		// Delete the object.
		if (((operFlags & OPER_LOCAL) > 0)
				&& (context != null))
			// Do not delete link objects because the context must done the job.
			response = objectSource.deleteObject(this, null, ((operFlags & OPER_LOCAL) > 0),
					((operFlags & OPER_NEED_REFRESH) > 0), false);
		else
			response = objectSource.deleteObject(this, null, ((operFlags & OPER_LOCAL) > 0),
					((operFlags & OPER_NEED_REFRESH) > 0));

		LyObject parent = getParent();

		if (parent != null)
		{
			int     i, n = getErrorsCount();
			LyError error;

			for (i = n - 1; i >= 0; i--)
			{
				error = getError(i);
				parent.removeError(error);
			}
		}

		return response;
	} // end method delete

	/**
	 * Checks if this object is found on its object source cache.
	 *
	 * @return True if it exists, false otherwise.
	 */
	public boolean checkObject()
	{
		LyObjectSource objectSource = getObjectSource();

		return ((_id == null) ? false : objectSource.checkObject(this));
	}

	/**
	 * Validates modifications done on this object. Transfer new local values into valid local
	 * values. This method is solely used by LyObjectSource after a creation or a modification has
	 * been successfully committed.
	 *
	 * @return Returns values set containing field values for modified local fields.
	 */
	protected synchronized final LyValueSet validate()
	{
		LyValueSet oldValues = null;
		LyValueSet pendingValues = getPendingValues();

		if (pendingValues != null)
		{
			oldValues = new LyValueSet(false);
			validateNewValues(oldValues, pendingValues);

			_pendingValues = null;
		}

		LyValueSet newValues = getNewValues();

		if (newValues != null)
		{
			if (oldValues == null)
				oldValues = new LyValueSet(false);
			validateNewValues(oldValues, newValues);

			_newValues = null;
		}

		// Reset "class controlled" state for this object and its parent
		setClassControlled(false);

		return oldValues;
	}

	/**
	 * Internal method to validate a list of newValues.
	 *
	 * @param oldValues Values to put old local values
	 * @param newValues Values to validate
	 */
	private void validateNewValues(LyValueSet oldValues, LyValueSet newValues)
	{
		if (newValues == null)
			return;

		int		    i, n = newValues.getSize();
		LyFieldInfo fieldInfo;
		LyValue     value;
		Object	    newValue;
		LyValueSet	localValues;

	    if (_classInfo.isLocal())
        {
	    	localValues = _localValues;

            for (i = 0; i < n; i++)
            {
                value = newValues.getValue(i);
                fieldInfo = (LyFieldInfo)value.getField();
                value = (localValues == null ? null : localValues.getFieldValue(fieldInfo));
                oldValues.addValue(fieldInfo, (value == null ? null : value.getValue()));
            }

            synchronized (this)
            {
	            if (_localValues == null)
	                _localValues = new LyValueSet(newValues);
	            else
	                _localValues.addValues(newValues);
            }
            return;
        }

		for (i = 0; i < n; i++)
		{
			value = newValues.getValue(i);
			fieldInfo = (LyFieldInfo)value.getField();

			if (fieldInfo.isLocal())
			{
				newValue = value.getValue();

		    	localValues = _localValues;
				value = ((localValues == null) ? null : localValues.getFieldValue(fieldInfo));
				oldValues.addValue(fieldInfo, ((value == null) ? null : value.getValue()));

				addLocalValue(fieldInfo, newValue);
			}
		}
	} // end method validateNewValues

	/**
	 * Method used only by LyObjectSource to move new values to pending values in case of a local
	 * operation.
	 */
	protected synchronized void moveNewToPendingValues()
	{
		if (_newValues == null)
			return;

		LyValueSet	pendingValues = _pendingValues;

		if (pendingValues == null)
			pendingValues = new LyValueSet();
		pendingValues.addValues(_newValues);
		_pendingValues = pendingValues;

		_newValues = null;
	}

	/**
	 * Method used only by LyObjectSource to move new values to pending values in case of a local
	 * operation.
	 */
	protected synchronized void movePendingToNewValues()
	{
		if (_pendingValues == null)
			return;

		LyValueSet	newValues = _newValues;

		if (newValues == null)
			newValues = new LyValueSet();
		newValues.addValues(_pendingValues);
		_newValues = newValues;

		_pendingValues = null;
	}

	/**
	 * Compares this object with another object using the given sort. This method can be overriden
	 * to provide a specific but more efficient implementation.
	 *
	 * @param  object Object to compare with this object.
	 * @param  sort   Sort used to compare the two objects.
	 * @param  byName Boolean indicating whether default sorting is done on ids or names when the
	 *                two objects are similar for the given sort. This is used to have a full order
	 *                when sorting an objects list to be sure that sorting the same list twice will
	 *                give the same result.
	 * @return 0 If the two objects are identical for the specified fieldInfo, a positive number if
	 *         this object is "greater" than object. If object is "greater" than this object, a
	 *         negative number is returned.
	 */
	public int compare(LyObject object, LySort sort, boolean byName)
	{
		if (object == null)
			return 1;
		if (object == this)
			return 0;
		if (sort == null)
			return compare(object, byName);

		// Compare objects on each criterion of the sort
		LySortCriterion criterion;
		int			    i, n = sort.getCriterionCount();
		int			    result;

		for (i = 0; i < n; i++)
		{
			criterion = sort.getCriterion(i);
			result = criterion.compare(this, object);
			if (result != 0)
				return result;
		}

		// If no difference was found, we compare objects with their ids
		// to have a complete order.
		return compare(object, byName);
	}

	/**
	 * Compares this object with another object using default comparison. This comparison is based
	 * on object ids or names depending on byName parameter.
	 *
	 * @param  object Object to compare with this object.
	 * @param  byName Boolean indicating whether sorting is done on ids or names.
	 * @return 0 If the two objects are identical for the specified fieldInfo, a positive number if
	 *         this object is "greater" than object. If object is "greater" than this object, a
	 *         negative number is returned.
	 */
	public int compare(LyObject object, boolean byName)
	{
		if (object == null)
			return 1;
		if (object == this)
			return 0;

		// If objects belong to the same class, we compare the values of these
		// objects on the name (or id) fields. We do it instead of using the
		// object id because the sorting result may be different (by example
		// with a number fieldInfo).
		if (object.getClassInfo() == _classInfo)
		{
			LyInfoList<LyFieldInfo> fieldInfos;

			if (byName)
			{
				LyEnvironment environment = getEnvironment();

				boolean ignoreAccent = ((environment != null)
							&& environment.testResource("LY_IGNORE_TEXT_ACCENT"));

				fieldInfos = _classInfo.getNameFieldInfos();
				if ((fieldInfos.getSize() == 1)
						&& (fieldInfos.getInfo(0) instanceof LyTextFieldInfo)
						&& !ignoreAccent)
					// Let compare by name
					fieldInfos = null;
			}
			else
				fieldInfos = _classInfo.getIdFieldInfos();

			if (fieldInfos != null)
			{
				int		    i, n = fieldInfos.getSize(), result;
				Object	    v1, v2;
				LyFieldInfo fieldInfo;

				for (i = 0; i < n; i++)
				{
					fieldInfo = fieldInfos.getInfo(i);
					v1 = getValue(fieldInfo);
					v2 = object.getValue(fieldInfo);

					if (!byName)
					{
						BitSet modifiers = new BitSet();

						modifiers.set(LyFilterExpression.MOD_NOT_BY_NAME);
						result = fieldInfo.compare(v1, v2, modifiers);
					}
					else
						result = fieldInfo.compare(v1, v2);

					if (result != 0)
						return result;
				}

				if (!byName)
				{
					// Id shouldn't be equals... use checkId
					String id1 = getId();
					String id2 = object.getId();

					if ((id1 == null) && (id2 == null))
						return 0; // ???
					if (id1 == null)
						return -1;
					if (id2 == null)
						return 1;

					return id1.compareTo(id2);
				}

				// If both objects have the same values for every fields, return 0
				return 0;
			}
		}

		if (byName)
		{
			String name = getName();
			String objectName = object.getName();
			int    test = name.compareTo(objectName);

			if (test != 0)
				return test;
		}

		LyObjectId id = getObjectId();
		LyObjectId objectId = object.getObjectId();

		return id.compare(objectId);
	} // end method compare

	/**
	 * Gets in the given session a list of objects of the specified target class related to this
	 * object. Depending on the dynamic parameter, the result list can be static or dynamic (may
	 * send events when its content is modified).
	 *
	 * @param  session     Session in which this search occurs. This parameter may be null if the
	 *                     request is not dynamic.
	 * @param  targetClass Target class.
	 * @param  dynamic     Boolean indicating if the result list has to be dynamic or not.
	 * @return Result list containing objects of the target class related to this object in the
	 *         given session.
	 */
	public LyObjectList getList(LyDataSession session, LyClassInfo targetClass,
		boolean dynamic, boolean sorted)
	{
		return getList(session, targetClass, dynamic, sorted, false);
	}

	/**
	 * Gets in the given session a list of objects of the specified target class related to this
	 * object. Depending on the dynamic parameter, the result list can be static or dynamic (may
	 * send events when its content is modified).
	 *
	 * @param  session     Session in which this search occurs. This parameter may be null if the
	 *                     request is not dynamic.
	 * @param  targetClass Target class.
	 * @param  dynamic     Boolean indicating if the result list has to be dynamic or not.
	 * @return Result list containing objects of the target class related to this object in the
	 *         given session.
	 */
	public LyObjectList getList(LyDataSession session, LyClassInfo targetClass,
		boolean dynamic, boolean sorted, boolean local)
	{
		LyObjectList list;

		if (dynamic)
			list = new LySimpleWorkSpace(session, this, dynamic);
		else
			list = new LySimpleObjectList(this);

		list.setSorted(sorted);
		return list.getList(session, targetClass, dynamic, local);
	}

	/**
	 * Gets in the given session a list of objects of the specified target class related to this
	 * object. Depending on the dynamic parameter, the result list can be static or dynamic (may
	 * send events when its content is modified).
	 *
	 * @param  session     Session in which this search occurs. This parameter may be null if the
	 *                     request is not dynamic.
	 * @param  targetClass Target class.
	 * @param  dynamic     Boolean indicating if the result list has to be dynamic or not.
	 * @return Result list containing objects of the target class related to this object in the
	 *         given session.
	 */
	public LyObjectList getList(LyDataSession session, LyClassInfo targetClass,
		boolean dynamic)
	{
		return getList(session, targetClass, dynamic, true, false);
	}

	/**
	 * Gets in the given session a list of objects of the specified target class related to this
	 * object. This list is <b>not</b> dynamic.
	 *
	 * @param  session     Session in which this search occurs. This parameter may be null.
	 * @param  targetClass Target class.
	 * @return Result list containing objects of the target class related to this object in the
	 *         given session.
	 */
	public LyObjectList getList(LyDataSession session, LyClassInfo targetClass)
	{
		return getList(session, targetClass, false, true, false);
	}

	/**
	 * Get list of array objects For internal use only.
	 *
	 * @return List of array objects
	 */
	public LyObjectList getArrayObjects()
	{
		LyFieldInfo			   fieldInfo;
		int					   i, n = _classInfo.getFieldInfoCount();
		LyArrayFieldInfo	   array;
		LyObjectList		   arrayObjects = null, objects;
		Object				   value;
		LyArrayFieldInfoDaemon daemon;
		LyApplication		   application;

		for (i = 0; i < n; i++)
		{
			fieldInfo = _classInfo.getFieldInfo(i);
			if (!(fieldInfo instanceof LyArrayFieldInfo))
				continue;

			array = (LyArrayFieldInfo)fieldInfo;
			application = fieldInfo.getApplication();
			daemon = (LyArrayFieldInfoDaemon)application.getDaemon(array);
			if (daemon != null)
			{
				value = getValue(array);
				if (value == null)
					continue;

				objects = daemon.getArrays(this);
				if (objects != null)
				{
					if (arrayObjects == null)
						arrayObjects = new LySimpleObjectList(objects);
					else
						arrayObjects.addList(objects);
				}
			}
		}

		LyObject arrayObject;

		n = ((arrayObjects == null) ? 0 : arrayObjects.getSize());
		for (i = n - 1; i >= 0; i--)
		{
			arrayObject = arrayObjects.getObject(i);
			if (arrayObject.isDeleted())
				arrayObjects.remove(arrayObject);
		}
		n = ((arrayObjects == null) ? 0 : arrayObjects.getSize());
		return ((n == 0) ? null : arrayObjects);
	} // end method getArrayObjects

	/**
	 * Get list of sub-objects (objects pointed by this object on composition relations).
	 *
	 * @return List of sub-objects.
	 */
	public LyObjectList getSubObjects(boolean local)
	{
		LyFieldInfo				 fieldInfo;
		int						 i, nf = _classInfo.getFieldInfoCount();
		int                      j, no;
		LyRelationFieldInfo		 relation;
		LySimpleObjectList		 subObjects = null, objects;
		Object					 value;
		LyRelationFieldInfo.Type type;
        LyObject                 subObject;

		for (i = 0; i < nf; i++)
		{
			fieldInfo = _classInfo.getFieldInfo(i);
			if (!(fieldInfo instanceof LyRelationFieldInfo)
					|| (fieldInfo.isLocal() != local))
				continue;

			relation = (LyRelationFieldInfo)fieldInfo;
			type = relation.getType();

			if ((type == LyRelationFieldInfo.Type.COMPOSITION)
			    || (type == LyRelationFieldInfo.Type.DEPENDENCY))
			{
				value = getValue(relation);
				if (value == null)
					continue;

				objects = relation.getObjects(value);
				no = (objects == null ? 0 : objects.getSize());
				if (no == 0)
				    continue;

				if (subObjects == null)
					subObjects = new LySimpleObjectList(objects);
				else
					subObjects.addList(objects);

		       if (!local
		           && (type == LyRelationFieldInfo.Type.COMPOSITION))
               {
                   // The current object is the parent of the subObjects.
                   for (j = 0; j < no; j++)
                   {
                       subObject = subObjects.getObject(j);
                       subObject.setParent(this, null);
                   }
               }
			}
		}

		return subObjects;
	} // end method getSubObjects

	/**
	 * Get list of join-objects (objects pointed by this object on join relations).
	 *
	 * @return List of join-objects.
	 */
	public LyObjectList getJoinObjects()
	{
		LyValueSet	 localValues = getLocalValues();

		if (localValues == null)
			return null;

		LyObjectList 	 joinObjects = null, objects;
		int				 i, n = localValues.getSize();
		LyValue			 value;
		LyFieldInfo		 fieldInfo;
		LyRelationDaemon daemon;
		LyApplication    application;

		for (i = 0; i < n; i++)
		{
			value = localValues.getValue(i);
			fieldInfo = (LyFieldInfo)value.getField();
			if (fieldInfo instanceof LyRelationFieldInfo)
			{
				application = fieldInfo.getApplication();
				daemon = (LyRelationDaemon)application.getDaemon(fieldInfo);
				if (daemon instanceof LyJoinRelationDaemon)
				{
					objects = ((LyJoinRelationDaemon)daemon).getJoins(this);
					if (objects != null)
					{
						if (joinObjects == null)
							joinObjects = new LySimpleObjectList(objects);
						else
							joinObjects.addList(objects);
					}
				}
			}
		}

		return joinObjects;
	} // end method getJoinObjects

	/**
	 * Returns a string representation of this object.
	 *
	 * @return This object returned as a string. This string contains the name of this object
	 *         followed by its id enclosed within parenthesis.
	 */
	public String toString()
	{
		return getName() + " (" + getId() + ")";
	} // end method update

	/**
	 * To mark this object as deleted.
	 */
	protected final void setDeleted(boolean deleted)
	{
		if (deleted)
			_state |= STATE_DELETED;
		else
			_state &= ~STATE_DELETED;
	}

	/**
	 * Indicated if this object has been deleted or not.
	 *
	 * @return true if the object has been deleted.
	 */
	public final boolean isDeleted()
	{
		return ((_state & STATE_DELETED) != 0);
	}

	/**
	 * To mark this object as deleted.
	 */
	protected final void setLocallyDeleted(boolean locallyDeleted)
	{
		if (locallyDeleted)
			_state |= STATE_LOCALLY_DELETED;
		else
			_state &= ~STATE_LOCALLY_DELETED;
	}

	/**
	 * Indicated if this object has been deleted or not.
	 *
	 * @return true if the object has been deleted.
	 */
	protected final boolean isLocallyDeleted()
	{
		return ((_state & STATE_LOCALLY_DELETED) != 0);
	}

	/**
	 * Remove the fieldInfo value from the new values list.
	 *
	 * @param fieldInfo The fieldInfo
	 */
	public void removeNewValue(LyFieldInfo fieldInfo)
	{
		removeNewValue(null, fieldInfo, CONTROL_FIELD_CONTROLS, OPER_NEED_REFRESH);
	}

	/**
	 * Remove the fieldInfo value from the new values list. For internal use only.
	 *
	 * @param context   The modification context removing the value.
	 * @param fieldInfo The fieldInfo
	 * @param refresh
	 */
	protected void removeNewValue(LyContext context, LyFieldInfo fieldInfo, int control,
		int operFlags)
	{
		removeNewValues(context, new LyInfoList<LyFieldInfo>(fieldInfo), control, operFlags);
	}

	/**
	 * Remove the list of fieldInfo value from the new values list.
	 *
	 * @param fieldInfos The list of fieldInfos
	 */
	public void removeNewValues(LyInfoList<LyFieldInfo> fieldInfos)
	{
		removeNewValues(null, fieldInfos, DEFAULT_CONTROL, OPER_NEED_REFRESH);
	}

	/**
	 * Remove the list of fieldInfo value from the new values list. For internal use only.
	 *
	 * @param context           The modification context removing the values.
	 * @param fieldInfo         The list of fieldInfos
	 * @param callClassControls Indicates if class controls may be called or not.
	 * @param control
	 * @param refresh
	 */
	protected void removeNewValues(LyContext context, LyInfoList<LyFieldInfo> fieldInfos,
		int control, int operFlags)
	{
		if (_newValues == null)
			return;

		if (fieldInfos == null)
			fieldInfos = _classInfo.getFieldInfos();

		LyFieldInfo fieldInfo;
		int		    i, n = fieldInfos.getSize();
		LyValueSet  values = new LyValueSet(false);
		LyValueSet  oldValues = new LyValueSet(false);
		LyValue     value, restoredValue;
		LyValueSet	newValues;

		for (i = 0; i < n; i++)
		{
			fieldInfo = fieldInfos.getInfo(i);
			fieldInfo = _classInfo.getRealFieldInfo(fieldInfo);

			newValues = _newValues;
			if ((newValues != null)
				&& newValues.contains(fieldInfo))
			{
				value = newValues.removeValue(fieldInfo);

				removeLastValue(fieldInfo);

				restoredValue = getFieldValue(fieldInfo);
				values.addValue(restoredValue);
				oldValues.addValue(value);

				// Values of the object have changed for this fieldInfo :
				// we must call the controls on this fieldInfo.
				if ((control & CONTROL_FIELD_CONTROLS) > 0)
					callFieldControls(fieldInfo, restoredValue.getValue(), control);
			}
		}

		synchronized (this)
		{
			if ((_newValues != null)
					&& (_newValues.getSize() == 0))
				_newValues = null;
		}

		// Reset "class controlled" state
		setClassControlled(false);

		if (values.getSize() > 0)
		{
			if ((operFlags & OPER_NEED_REFRESH) > 0)
				refresh(oldValues);

			// Call class controls to restore error state
			if (control != CONTROL_NONE)
			{
				if ((control & CONTROL_CLASS_CONTROLS) > 0)
					callClassControls(values, control);
				if ((control & CONTROL_LOADING) == 0)
					controlParent(control & CONTROL_FIELD_CONTROLS);
			}
		}
	} // end method removeNewValues

	/**
	 * Remove the fieldInfo value from the pending values list. For internal use only.
	 *
	 * @param fieldInfo
	 */
	public void removePendingValue(LyFieldInfo fieldInfo)
	{
		removePendingValues(new LyInfoList<LyFieldInfo>(fieldInfo));
	}

	/**
	 * Remove the list of fieldInfo value from the pending values list. For internal use only.
	 *
	 * @param fieldInfos The list for fieldInfos.
	 */
	public void removePendingValues(LyInfoList<LyFieldInfo> fieldInfos)
	{
		removePendingValues(null, fieldInfos, DEFAULT_CONTROL, OPER_NEED_REFRESH);
	}

	/**
	 * Remove the list of fieldInfo value from the pending values list. For internal use only.
	 *
	 * @param context    The modification context removing the values.
	 * @param fieldInfos The list fo fieldInfos.
	 */
	protected void removePendingValues(LyContext context, LyInfoList<LyFieldInfo> fieldInfos,
		int control, int operFlags)
	{
		if (_pendingValues == null)
			return;

		if (fieldInfos == null)
			fieldInfos = _classInfo.getFieldInfos();

		LyFieldInfo fieldInfo;
		int		    i, n = fieldInfos.getSize();
		LyValueSet  values = new LyValueSet(false);
		LyValue     value;
		LyValueSet	pendingValues;

		for (i = 0; i < n; i++)
		{
			fieldInfo = fieldInfos.getInfo(i);
			fieldInfo = _classInfo.getRealFieldInfo(fieldInfo);

			pendingValues = _pendingValues;
			if ((pendingValues != null)
				&& pendingValues.contains(fieldInfo))
			{
				value = pendingValues.removeValue(fieldInfo);

				removeLastValue(fieldInfo);

				values.addValue(value);

				// Values of the object have changed for this fieldInfo :
				// we must call the controls on this fieldInfo.
				if ((control & CONTROL_FIELD_CONTROLS) > 0)
					callFieldControls(fieldInfo, getValue(fieldInfo), control);
			}
		}

		synchronized (this)
		{
			if ((_pendingValues != null)
					&& (_pendingValues.getSize() == 0))
				_pendingValues = null;
		}

		// Reset "class controlled" state
		setClassControlled(false);

		if (values.getSize() > 0)
		{
			if ((operFlags & OPER_NEED_REFRESH) > 0)
				refresh(values);

			if (control != CONTROL_NONE)
			{
				if ((control & CONTROL_CLASS_CONTROLS) > 0)
					// Call class controls to restore error state
					callClassControls(values, control);
				if ((control & CONTROL_LOADING) == 0)
					controlParent(control & CONTROL_FIELD_CONTROLS);
			}
		}
	} // end method removePendingValues

	/**
     * Reset storing cache of the id. The next call of getObjectId will recompute the id of this
     * object.
     */
    protected void resetObjectId(LyObjectId id)
    {
        _id = id;
        touch();
    }

	/**
	 * Reset storing cache of the id. The next call of getObjectId will recompute the id of this
	 * object.
	 */
	public void resetObjectId()
	{
		resetObjectId(null);
	}

	/**
	 * Reset name of this obect.
	 */
	protected void resetName()
	{
		_name = null;

		LyInfoLabel label = _classInfo.getLabel(LyInfoLabel.Type.NAME);

		if (label != null)
		{
			LyApplication application = getApplication();
			LyLabelDaemon daemon = application.getDaemon(label);

			if (daemon != null)
				daemon.removeResult(this);
		}
	}

	/**
	 * To get the error level of this object.
	 *
	 * @return the error level of the object. Error levels greater than LEVEL_WARNING are user
	 *         specific levels.
	 */
	public final short getErrorLevel()
	{
		return _errorLevel;
	}

	/**
	 * Get error level on this object for the given fieldInfo.
	 *
	 * @param  fieldInfo FieldInfo on which error level is requested.
	 * @return Status of calling controls.
	 */
	public short getErrorLevel(LyFieldInfo fieldInfo)
	{
		short				    errorLevel, level = LEVEL_OK;
		int					    i, n = getErrorsCount();
		LyError				    error;
		LyInfoList<LyFieldInfo> infoList;

		for (i = 0; i < n; i++)
		{
			error = getError(i);
			infoList = error.getFieldInfosCause();
			if ((infoList != null)
					&& (infoList.contains(fieldInfo)))
			{
				errorLevel = error.getLevel();
				if ((errorLevel != LEVEL_OK)
						&& ((level == LEVEL_OK) || (errorLevel < level)))
					level = errorLevel;
			}
		}

		return level;
	}

	/**
	 * To add an error to this object.
	 *
	 * @param error the error to add.
	 */
	public void addError(LyError error)
	{
		if (_errors == null)
			_errors = new ArrayList<LyError>(1);
		else if (_errors.contains(error))
			return;

        LyEnvironment environment = getEnvironment();

        environment.logTrace("BF AppEngine : error added on object "
            + this + " : " + error.getMessage(environment));

		_errors.add(error);
		errorAdded(error);

		// If the error level has changed, we must post a state change.
		short errorLevel = error.getLevel();
		short objectErrorLevel = getErrorLevel();

		if (((objectErrorLevel == LEVEL_OK) && (errorLevel != LEVEL_OK)) // If Object Ok, always keep error
				|| ((errorLevel != LEVEL_OK) && (errorLevel < objectErrorLevel))) // If object Error; keep error if lower (more important)
		{
			_errorLevel = errorLevel;
			stateChanged(LyDataSourceEvent.TYPE_ERROR, objectErrorLevel);
		}

		LyObject parent = getParent();

		if (parent != null)
			parent.addError(error);
	}

	/**
	 * Remove the error at the specified index.
	 *
	 * @param  idx
	 * @return The removed error
	 */
	public LyError removeError(int idx)
	{
		if (idx < 0)
			throw new ArrayIndexOutOfBoundsException("BF AppEngine : " + idx + " < 0 ");

		int n = getErrorsCount();

		if (idx >= n)
			throw new ArrayIndexOutOfBoundsException("BF AppEngine : " + idx + " >= " + n);

		LyError error = _errors.remove(idx);

		if (_errors.size() == 0)
			_errors = null;

		errorRemoved(error);
		refreshErrorLevel();

		LyObject parent = getParent();

		if (parent != null)
			parent.removeError(error);

		return error;
	}

	/**
	 * Remove the specified error.
	 *
	 * @param  idx
	 * @return The removed error
	 */
	public LyError removeError(LyError error)
	{
		if (_errors == null)
			return null;

		// Find error index
		int     i, n = _errors.size();
		LyError err;

		for (i = 0; i < n; i++)
		{
			err = _errors.get(i);
			if (err.equals(error))
				return removeError(i);
		}

		return null;
	}

	/**
	 * To get the number of errors associated to this object.
	 *
	 * @return The number of errors.
	 */
	public int getErrorsCount()
	{
		return ((_errors == null) ? 0 : _errors.size());
	}

	/**
	 * To get the error at the specified index.
	 *
	 * @param  idx the index
	 * @return the error at the idx position.
	 */
	public LyError getError(int idx)
	{
		if (idx < 0)
			throw new ArrayIndexOutOfBoundsException("BF AppEngine : " + idx + " < 0 ");

		int n = getErrorsCount();

		if (idx >= n)
			throw new ArrayIndexOutOfBoundsException("BF AppEngine : " + idx + " >= " + n);

		return _errors.get(idx);
	}

	/**
	 * To get the error with the specified id.
	 *
	 * @param  errorId Id of the requested error.
	 * @return the error with the given id if it exists.
	 */
	public LyError getError(String errorId)
	{
		int     i, n = getErrorsCount();
		LyError error;

		errorId = errorId.intern();
		for (i = 0; i < n; i++)
		{
			error = getError(i);
			if (error.getId() == errorId)
				return error;
		}

		return null;
	}

	/**
	 * To get the last error on this object with the specified level.
	 *
	 * @param  level the level of the error
	 * @return the last error.
	 */
	public LyError getLastError(short level)
	{
		int     i, n = getErrorsCount();
		LyError error;

		for (i = n - 1; i >= 0; i--)
		{
			error = getError(i);
			if (error.getLevel() == level)
				return error;
		}

		return null;
	}

	/**
	 * Reset all errors caused by the specified fieldInfo.
	 *
	 * @param fieldInfo Cause of errors to reset.
	 */
	public void resetFieldInfoErrors(LyFieldInfo fieldInfo)
	{
		resetFieldInfoErrors(fieldInfo, CONTROL_NONE, true);
	}

	/**
	 * Reset all errors caused by the specified fieldInfo.
	 *
	 * @param fieldInfo Cause of errors to reset.
	 */
	private void resetFieldInfoErrors(LyFieldInfo fieldInfo, int subType,
		boolean resetClassErrors)
	{
		int					    i, n = getErrorsCount();
		LyError				    error;
		LyInfoList<LyFieldInfo> infoList;

		for (i = n - 1; i >= 0; i--)
		{
			error = getError(i);
			if (!resetClassErrors
					&& (error.getType() == LyError.TYPE_CLASS))
				continue;
			if ((error.getType() != LyError.TYPE_CLASS)
					&& (subType >= 0)
					&& (error.getSubType() != subType))
				continue;

			infoList = error.getFieldInfosCause();
			if ((infoList != null)
					&& (infoList.contains(fieldInfo)))
				removeError(i);
		}
	}

	/**
	 * Reset all errors caused by class controls.
	 */
	public void resetClassErrors()
	{
		resetClassErrors(CONTROL_NONE);
	}

	/**
	 * Reset all errors caused by class controls.
	 */
	private void resetClassErrors(int subType)
	{
		int     i, n = getErrorsCount();
		LyError error;

		for (i = n - 1; i >= 0; i--)
		{
			error = getError(i);
			if ((error.getType() == LyError.TYPE_CLASS)
					&& ((subType < 0)
						|| (error.getSubType() == subType)))
				removeError(i);
		}
	}

	/**
	 * Recomputes the error level of the object and post a state change event if necessary.
	 *
	 * @deprecated This method must'nt be used directly and will be not public in future releases.
	 */
	public void refreshErrorLevel()
	{
		short   errorLevel, oldErrorLevel, newErrorLevel = LEVEL_OK;
		int     i, n = getErrorsCount();
		LyError error;

		oldErrorLevel = getErrorLevel();
		for (i = 0; i < n; i++)
		{
			error = getError(i);
			errorLevel = error.getLevel();
			if (newErrorLevel == LEVEL_OK)
				newErrorLevel = errorLevel;
			else if ((errorLevel != LEVEL_OK) && (errorLevel < newErrorLevel))
				newErrorLevel = errorLevel;
		}

		if (newErrorLevel != oldErrorLevel)
		{
			_errorLevel = newErrorLevel;
			stateChanged(LyDataSourceEvent.TYPE_ERROR, oldErrorLevel);
		}
	}

	/**
	 * Refresh a list of fields on this object. Updates all views that contains this object.
	 *
	 * @param fieldInfos List of fields to refresh (may be null).
	 */
	public void refresh(LyInfoList<LyFieldInfo> fieldInfos)
	{
		if (fieldInfos == null)
			fieldInfos = _classInfo.getIdFieldInfos();

		LyValueSet  values = null;
		int		    i, n = fieldInfos.getSize();
		LyFieldInfo fieldInfo;

		values = new LyValueSet(false);
		for (i = 0; i < n; i++)
		{
			fieldInfo = fieldInfos.getInfo(i);
			values.addValue(fieldInfo, getFromMemory(fieldInfo));
		}

		refresh(values);
	}

	/**
	 * Refresh new values.
	 */
	public void refresh()
	{
		refresh(_newValues);
	}

	/**
	 * Send an update event for this object. Updates all views that contains this object.
	 *
	 * @param values List of oldValues updated (may be null).
	 */
	public void refresh(LyValueSet values)
	{
		if ((values == null) || (values.getSize() == 0))
			return;

		touch();

		LyObjectSource source = getObjectSource();

		if (source.checkObject(this))
		{
			LyDataSourceEvent event = new LyDataSourceEvent(source, LyDataSourceEvent.TYPE_SET,
					this, values);

			event.setLocal(true);
			source.postEvent(event);
		}
	}

	/**
	 * Get parent object of this sub object.
	 *
	 * @return The parent object if this object is a sub-object, null otherwise.
	 * @deprecated use getParents
	 */
	public LyObject getParent()
	{
		if (_parentReference != null)
		{
			LyObject parent = _parentReference.get();

			if (parent != null)
				return parent;
		}

		if (_classInfo.getParent() instanceof LyStructFieldInfo)
			return getParentOfStruct();

		return null;
	} // end method getParent

	/**
     * Get parents object of this object.
     *
     * @return The parent objects if this object is a sub-object, null otherwise.
     */
    public LyObjectList getParents()
    {
        LyFieldInfo         fieldInfo;
        int                 i, n = _classInfo.getFieldInfoCount();
        LyRelationFieldInfo relation;
        LySimpleObjectList  parents = null, objects;

        for (i = 0; i < n; i++)
        {
            fieldInfo = _classInfo.getFieldInfo(i);
            if (!(fieldInfo instanceof LyRelationFieldInfo))
                continue;

            relation = (LyRelationFieldInfo)fieldInfo;
            if (relation.getType() == LyRelationFieldInfo.Type.PARENT)
            {
                objects = relation.getObjects(this);
                if ((objects != null)
                    && (objects.getSize() >= 1))
                {
                    if (parents == null)
                        parents = new LySimpleObjectList(objects);
                    else
                        parents.addList(objects);
                }
            }
        }

        return parents;
    } // end method getParents

	/**
	 * Sets parent of a struct object.
	 */
	protected void setParent(LyObject parent, LyFieldInfo parentFieldInfo)
	{
		if (parent == null)
		{
			_parentReference = null;
			_parentFieldInfo = parentFieldInfo;
			return;
		}

		if (parent.getClassInfo().getParent() instanceof LyArrayFieldInfo)
			// An array object can't be parent
			return;
		if ((_parentReference != null)
				&& (_parentReference.get() == parent))
			return;

		_parentReference = new WeakReference<LyObject>(parent);

		_parentFieldInfo = parentFieldInfo;

		if (_errors == null)
			return;

		// Report errors on parent
		int     i, n = _errors.size();
		LyError error;

		for (i = 0; i < n; i++)
		{
			error = _errors.get(i);
			parent.addError(error);
		}
	} // end method setParent

	/**
	 * Tests if this object has been created and comitted to the physical layer.
	 */
	private boolean isCreated()
	{
		LyClassInfo classInfo = getClassInfo();
		boolean     result = true;

		if (!classInfo.isLocal())
			result = (getPeer() != null);
		else
		{
			LyObjectSource source = getObjectSource();

			result = source.checkObject(this);
		}

		return result;
	}

	/**
	 * Method to call controls corresponding to the given fieldInfo.
	 *
	 * @param  fieldInfo FieldInfo that holds on controls to apply.
	 * @param  value     Value assigned on this object for the given field info.
	 * @return Result status of calling controls.
	 */
	public short callFieldControls(LyFieldInfo fieldInfo, Object value)
	{
		return callFieldControls(fieldInfo, value,
			(CONTROL_FIELD_CONTROLS & ~CONTROL_DISABLED_ON_LOADING) | CONTROL_MANDATORY, false);
	}

	/**
	 * Method to call controls corresponding to the given fieldInfo.
	 *
	 * @param  fieldInfo      FieldInfo that holds on controls to apply.
	 * @param  value          Value assigned on this object for the given field info.
	 * @param  checkMandatory Indicates if the mandatory control may be called or not.
	 * @return Result status of calling controls.
	 */
	protected short callFieldControls(LyFieldInfo fieldInfo, Object value,
		int controls, boolean resetClassErrors)
	{
		if ((controls & CONTROL_FIELD_CONTROLS) == 0)
			return LyControlInterface.STATUS_OK;
		if (_classInfo.isTemporary())
			return LyControlInterface.STATUS_OK;

		LyApplication application = getApplication();
		int			  priority = application.getPriorityLevel(this);

		if (priority < 0)
		{
			resetFieldInfoErrors(fieldInfo, ALL_CONTROLS, resetClassErrors); // reset all errors
			return LyControlInterface.STATUS_OK;
		}

		short   result = LyControlInterface.STATUS_OK;
		LyError error;

		// Reset errors caused by the fieldInfo
		resetFieldInfoErrors(fieldInfo, CONTROL_NONE, resetClassErrors);

		if (((controls & CONTROL_LOADING) == 0) // Disabled on loading
				&& (((controls & CONTROL_MANDATORY) | CONTROL_VALUE | CONTROL_UNICITY) > 0))
		{
			if (value instanceof LyValueSet)
			{
				LyValueSet values = (LyValueSet)value;

				value = getValue(fieldInfo);
				values.addValue(fieldInfo, value);
			}

			// Check if mandatory
			if ((controls & CONTROL_MANDATORY) > 0)
				result = checkMandatory(fieldInfo, value);

			if ((controls & CONTROL_VALUE) > 0)
			{
				// Perform value controls
				LyEnvironment environment = getEnvironment();
				String		  errorMessage;

				resetFieldInfoErrors(fieldInfo, CONTROL_VALUE, false);

				errorMessage = fieldInfo.controlValue(environment, value);
				if (errorMessage != null)
				{
					// Create the appropriate error
					error = new LyError(LEVEL_ERROR, LyError.TYPE_FIELD, CONTROL_VALUE,
							errorMessage, null, fieldInfo, this, fieldInfo);
					addError(error);
					result = LyControlInterface.STATUS_ERROR;
				}
			}

			// Check unicity control
			if ((controls & CONTROL_UNICITY) > 0)
			{
				error = checkUnicity(fieldInfo, value);
				if (error != null)
				{
					// Add the error to the errors list
					addError(error);
					result = LyControlInterface.STATUS_ERROR;
				}
			}
		}

		// Call application controls
		if ((controls & CONTROL_APP_FIELD_CONTROLS) == 0)
			return result;

		LyControl		   control;
		LyFieldInfoControl instance;
		short			   controlResult;
		int				   i, n = fieldInfo.getControlCount();
		LyValueSet		   values;
		int				   errorCount;
		LyEnvironment	   environment = getEnvironment();

		// Reset control field errors
		resetFieldInfoErrors(fieldInfo, CONTROL_APP_FIELD_CONTROLS, false);

		for (i = 0; i < n; i++)
		{
			control = fieldInfo.getControl(i);
			// Check priority
			if (control.getPriority() > priority)
				continue;
			// Check control type
			if (!control.hasType(LyControl.Type.FIELD))
				continue;

			// Call control to validate values
			if (value instanceof LyValueSet)
			{
				values = (LyValueSet)value;
				value = getValue(fieldInfo);
				values.addValue(fieldInfo, value);
			}

			try
			{
				errorCount = getErrorsCount();
				error = ((errorCount > 0) ? getError(errorCount - 1) : null);

				instance = (LyFieldInfoControl)control.getInstance(this);
				controlResult = instance.checkNewValue(control, this,
						new LyFieldInfoValue(fieldInfo, value), control.getArguments());

				if (controlResult != LyControlInterface.STATUS_OK)
				{
					if (getLastError(controlResult) == error)
						environment.logError("BF AppEngine : control " + control
							+ " returned an error on object " + this
							+ " but didn't add a new error on this object");

					if ((result == LyControlInterface.STATUS_OK)
							|| (controlResult < result))
						result = controlResult;
				}
			}
			catch (Exception e)
			{
				environment.logError("BF AppEngine : exception caught on field control call on object "
					+ this);
				environment.writeException(e);
			}
		}

		return result;
	} // end method callFieldControls

	protected short callFieldControls(LyFieldInfo fieldInfo, Object value, int control)
	{
		return callFieldControls(fieldInfo, value, control, false);
	}

	/**
	 * Unicity check for given fieldInfo and value.
	 *
	 * @param  fieldInfo Field info to check.
	 * @param  value     Value associated to the field info.
	 * @return An error if fieldInfo is declared unique and its value is not, null if everything is
	 *         ok.
	 */
	protected LyError checkUnicity(LyFieldInfo fieldInfo, Object value)
	{
		if (fieldInfo.hasMark(LyFieldInfo.MARK_UNIQUE))
		{
			if (value == null)
				resetFieldInfoErrors(fieldInfo, CONTROL_UNICITY, false);
			else
			{
				// Retrieve class info in which the field info was defined
				LyClassInfo classInfo = getClassInfo();

				while (classInfo.isInherited(fieldInfo))
					classInfo = classInfo.getSuperClass();

				// Check unicity
				LyApplication  application = getApplication();
				LyDataSession  session = application.getCurrentSession();
				LyObjectSource source;

				if (session != null)
					source = session.getObjectSource(classInfo);
				else
					source = application.getObjectSource(classInfo);

				LyFilter		   filter = new LyFilter(LyFilter.Type.AND);
				LyFilterExpression filterExp = new LyFilterExpression(fieldInfo, value);
				LyResponse		   response;

				filter.addElement(filterExp);
				response = source.getObjects(filter);

				if ((response.getSize() > 1)
						|| ((response.getSize() == 1) && !response.contains(this)))
				{
					// Attribut n'autorisant que des valeurs uniques
					LyError error;

					error = new LyError(LEVEL_ERROR, LyError.TYPE_FIELD, CONTROL_UNICITY,
							"LY_ER_UNIQUE", new String[] { fieldInfo.getName() }, fieldInfo,
							(LyObject)response.getResponse(0), fieldInfo);
					return error;
				}
				else
					resetFieldInfoErrors(fieldInfo, CONTROL_UNICITY, false);
			}
		}

		return null;
	} // end method checkUnicity

	/**
	 * Internal method to call controls associated to class info of this object.
	 *
	 * @param  values List of controlled values.
	 * @return Result of calling controls.
	 */
	public short callClassControls(LyValueSet values)
	{
		return callClassControls(values, CONTROL_CLASS_CONTROLS);
	}

	/**
	 * Internal method to call controls associated to class info of this object.
	 *
	 * @param  values  List of controlled values.
	 * @param  control Class controls applied on this object.
	 * @return Result of calling controls.
	 */
	short callClassControls(LyValueSet values, int control)
	{
		if ((_state & STATE_CLASS_CONTROLLED) > 0)
			return LyControlInterface.STATUS_OK;
		if ((control & CONTROL_CLASS_CONTROLS) == 0)
			return LyControlInterface.STATUS_OK;
		if (_classInfo.isTemporary())
			return LyControlInterface.STATUS_OK;

		LyApplication application = getApplication();
		int			  priority = application.getPriorityLevel(this);

		if (priority < 0)
		{
			resetClassErrors(ALL_CONTROLS); // reset all errors
			return LyControlInterface.STATUS_OK;
		}

		// Reset class errors
		resetClassErrors(CONTROL_NONE);

		short result = LyControlInterface.STATUS_OK;
		short subResult;

		if ((control & CONTROL_APP_CLASS_CONTROLS) > 0)
		{
			subResult = checkControls(values);
			if ((subResult != LyControlInterface.STATUS_OK)
					&& ((result == LyControlInterface.STATUS_OK)
						|| (subResult < result)))
				result = subResult;
		}

		if ((control & CONTROL_RULES) > 0)
		{
			subResult = checkRules(values);
			if ((subResult != LyControlInterface.STATUS_OK)
					&& ((result == LyControlInterface.STATUS_OK)
						|| (subResult < result)))
				result = subResult;
		}

		if ((control & CONTROL_MANDATORY) > 0)
		{
			subResult = checkMandatoryFields();
			if ((subResult != LyControlInterface.STATUS_OK)
					&& ((result == LyControlInterface.STATUS_OK)
						|| (subResult < result)))
				result = subResult;
		}

		if ((control & CONTROL_ID) > 0)
		{
			subResult = checkId(values);
			if ((subResult != LyControlInterface.STATUS_OK)
					&& ((result == LyControlInterface.STATUS_OK)
						|| (subResult < result)))
				result = subResult;
		}

		if ((control & CONTROL_APP_ID) > 0)
		{
			subResult = checkAppId(values);
			if ((subResult != LyControlInterface.STATUS_OK)
					&& ((result == LyControlInterface.STATUS_OK)
						|| (subResult < result)))
				result = subResult;
		}

		if ((control & CONTROL_LOADING) == 0)
		{
			controlParent(control & CONTROL_CLASS_CONTROLS);

			if ((result == LyControlInterface.STATUS_OK)
					&& ((control & CONTROL_CLASS_CONTROLS) == CONTROL_CLASS_CONTROLS))
				// Controlled state can be changed only when all class controls are done
				setClassControlled(true);
		}

		return result;
	} // end method callClassControls

	/**
	 * Checks controls.
	 */
	protected short checkControls(LyValueSet values)
	{
		// Call controls
		short		   result = LyControlInterface.STATUS_OK;
		LyControl	   control;
		LyApplication  application = getApplication();
		int			   priority = application.getPriorityLevel(this);
		LyClassControl instance;
		short		   controlResult;
		short		   i, n = (short)_classInfo.getControlCount();
		LyError		   error;
		int			   errorCount;
		LyEnvironment  environment = getEnvironment();

		// Reset control class errors
		resetClassErrors(CONTROL_APP_CLASS_CONTROLS);

		for (i = 0; i < n; i++)
		{
			control = _classInfo.getControl(i);
			// Check control type
			if (!control.hasType(LyControl.Type.CLASS))
				continue;
			// Check priority
			if (control.getPriority() > priority)
				continue;

			// Call control to validate values
			try
			{
				errorCount = getErrorsCount();
				error = ((errorCount > 0) ? getError(errorCount - 1) : null);

				instance = (LyClassControl)control.getInstance(this);
				controlResult = instance.validate(control, this, values, control.getArguments());
				if (controlResult != LyControlInterface.STATUS_OK)
				{
					if (getLastError(controlResult) == error)
						environment.logError("BF AppEngine : control " + control
							+ " returned an error on object " + this
							+ " but didn't add a new error on this object");

					if ((result == LyControlInterface.STATUS_OK)
							|| (controlResult < result))
						result = controlResult;
				}
			}
			catch (Exception e)
			{

				environment.logError("BF AppEngine : exception caught on class control call on object "
					+ this);
				environment.writeException(e);
			}
		}

		return result;
	} // end method checkControls

	/**
	 * Checks rules.
	 */
	protected short checkRules(LyValueSet values)
	{
		if (!hasNewValues() && !hasPendingValues())
			return LyControlInterface.STATUS_OK;

		// Perform controls corresponding to rules
		short			 result = LyControlInterface.STATUS_OK;
		short			 i, n = _classInfo.getRuleCount();
		LyRule			 rule;
		LyFilter		 filter;
		LyRule.Operation operation;
		LyObjectInfo     objectInfo;
		LyFieldInfo		 fieldInfo;
		LyError			 error;
		Object[]		 args;
		String			 errorMessage;
		LyFieldInfo		 cause;
		LyApplication    application = getApplication();
		int				 priority = application.getPriorityLevel(this);
		short			 type, subType;
		boolean			 onCheck;

		for (i = 0; i < n; i++)
		{
			rule = _classInfo.getRule(i);
			// Check priority
			if (rule.getPriority() > priority)
				continue;

			operation = rule.getOperation();
			if ((operation != LyRule.Operation.SET_EDITABLE)
					&& (operation != LyRule.Operation.SET_OPTIONAL)
					&& (operation != LyRule.Operation.CLEAR)
					&& (operation != LyRule.Operation.SET_VALUE))
				continue;

			filter = rule.getFilter();
			onCheck = rule.isOnCheck();
			if (filter == null)
				continue;

			type = LyError.TYPE_CLASS;
			subType = CONTROL_RULES;
			objectInfo = rule.getObjectInfo();

			if (!(objectInfo instanceof LyFieldInfo))
				continue;

			fieldInfo = (LyFieldInfo)objectInfo;
			errorMessage = null;
			cause = null;

			switch (operation)
			{
			case SET_EDITABLE:
                if ((values != null)
                    && values.contains(fieldInfo))
                {
                    LyInfoList<LyFieldInfo> filterFields = filter.getFieldInfos(true);
                    int                     f, nf = (filterFields == null ? 0 : filterFields.getSize());
                    LyValueSet              checkedValues = new LyValueSet(false);
                    LyFieldInfo             filterField;

                    for (f = 0; f < nf; f++)
                    {
                        filterField = filterFields.getInfo(f);
                        checkedValues.addValue(filterField, get(filterField, true, false));
                    }

                    if ((checkedValues != null)
                        && (checkedValues.getSize() > 0)
                        && (filter.match(checkedValues) != onCheck))
                    {
                        if (filter.contains(fieldInfo) // Only oldValues are accepted
                                || (filter.match(values) != onCheck)) // New values are also checked
                            // Error fieldInfo is not editable
                            errorMessage = "LY_ER_FIELD_NOT_EDITABLE";
                    }
                }
				break;

			case SET_OPTIONAL:
				if ((getValue(fieldInfo) == null)
						&& (filter.match(this) != onCheck))
				{
					// Error fieldInfo is mandatory
					errorMessage = "LY_ER_FIELD_MANDATORY";
					cause = fieldInfo;
					type = LyError.TYPE_FIELD;
					subType = CONTROL_MANDATORY;
				}
				break;

			case CLEAR:
				if ((values != null)
						&& values.contains(fieldInfo)
						&& !values.getFieldValue(fieldInfo).isNull()
						&& filter.match(this))
					// Error fieldInfo should be set to null
					errorMessage = "LY_ER_FIELD_CLEAR";
				break;

			case SET_VALUE:
				if ((values != null)
						&& values.contains(fieldInfo)
						&& filter.match(this))
				{
					Object value = values.getFieldValue(fieldInfo).getValue();
					Object ruleValue = rule.getValue();

					if (!fieldInfo.equals(value, ruleValue))
						// Error fieldInfo should be different from rule value
						errorMessage = "LY_ER_FIELD_SET_VALUE";
				}
				break;

			default:
				break;
			}

			if (errorMessage != null)
			{
				// Create the appropriate error
				if (rule.getValue() != null)
				{
					args = new Object[2];
					args[0] = fieldInfo.getName();
					args[1] = rule.getValue();
				}
				else
				{
					args = new Object[1];
					args[0] = fieldInfo.getName();
				}
				error = new LyError(LEVEL_ERROR, type, subType, errorMessage, args, fieldInfo, this,
						cause);
				addError(error);
				result = LyControlInterface.STATUS_ERROR;
			}
			else if (type == LyError.TYPE_CLASS)
				resetClassErrors(subType);
			else
				resetFieldInfoErrors(fieldInfo, subType, false);
		}

		return result;
	} // end method checkRules

	/**
	 * Checks mandatory fields.
	 */
	protected short checkMandatoryFields()
	{
		short		  result = LyControlInterface.STATUS_OK;

		if (!CHECK_OPTIONAL_ON_OBJECTS)
			return result;

		short	    i, n = (short)_classInfo.getFieldInfoCount();
		LyFieldInfo fieldInfo;
		Object	    value;

		for (i = 0; i < n; i++)
		{
			fieldInfo = _classInfo.getFieldInfo(i);
			value = getValue(fieldInfo);
			if (checkMandatory(fieldInfo, value) == LyControlInterface.STATUS_ERROR)
				result = LyControlInterface.STATUS_ERROR;
		}

		return result;
	}

	/**
	 * Check mandatory.
	 */
	protected short checkMandatory(LyFieldInfo fieldInfo, Object value)
	{
		short result = LyControlInterface.STATUS_OK;

		if (value != null)
		{
			// A non optional value cannot be empty or just with space for text field
			if (!(fieldInfo instanceof LyTextFieldInfo)
					|| ( ((String)value).trim().length() != 0 ))
			{
			resetFieldInfoErrors(fieldInfo, CONTROL_MANDATORY, false);
			return result;
		}
		}

		LyEnvironment environment = getEnvironment();

		if (!environment.testResource("LY_CHECK_OPTIONAL_ON_OBJECTS"))
			return result;

		short			  j, nr = _classInfo.getRuleCount();
		boolean			  hasOptionalRule = false;
		LyRule			  rule;
		LyFieldInfo		  field;
		int				  i, n = _classInfo.getFieldInfoCount();
		short			  o, no;
		LyChoiceFieldInfo choice;
		LyChoiceOption    option;
		LyApplication     application = getApplication();
		int				  priority = application.getPriorityLevel(this);
		LyFilter		  filter;
		LyRule.Operation  operation;
		LyObjectInfo	  objectInfo;
		LyError			  error;
		Object[]		  args;
		boolean			  onCheck;

		for (j = 0; j < nr; j++)
		{
			rule = _classInfo.getRule(j);
			// Check priority
			if (rule.getPriority() > priority)
				continue;

			operation = rule.getOperation();
			if ((operation != LyRule.Operation.SET_EDITABLE)
					&& (operation != LyRule.Operation.SET_OPTIONAL)
					&& (operation != LyRule.Operation.CLEAR)
					&& (operation != LyRule.Operation.SET_VALUE))
				continue;

			filter = rule.getFilter();
			onCheck = rule.isOnCheck();
			if (filter == null)
				continue;

			objectInfo = rule.getObjectInfo();
			if ((objectInfo == fieldInfo)
					&& (operation == LyRule.Operation.SET_OPTIONAL))
			{
				hasOptionalRule = true;
				if (filter.match(this) != onCheck)
				{
					args = new Object[1];
					args[0] = fieldInfo.getName();
					error = new LyError(LEVEL_ERROR, LyError.TYPE_FIELD, CONTROL_MANDATORY,
							"LY_ER_FIELD_MANDATORY", args, fieldInfo, this, fieldInfo);
					addError(error);
					result = LyControlInterface.STATUS_ERROR;
				}
				else
					resetFieldInfoErrors(fieldInfo, CONTROL_MANDATORY, false);
			}
		}

		for (i = 0; (i < n) && !hasOptionalRule; i++)
		{
			field = _classInfo.getFieldInfo(i);
			if (field instanceof LyChoiceFieldInfo)
			{
				choice = (LyChoiceFieldInfo)field;
				no = choice.getChoiceOptionCount();
				for (o = 0; (o < no) && !hasOptionalRule; o++)
				{
					option = choice.getChoiceOption(o);
					nr = option.getRuleCount();
					for (j = 0; (j < nr) && !hasOptionalRule; j++)
					{
						rule = option.getRule(j);
						if ((rule.getOperation() == LyRule.Operation.SET_OPTIONAL)
								&& (rule.getObjectInfo() == fieldInfo))
							hasOptionalRule = true;
					}
				}
			}
		}

		if (!hasOptionalRule)
		{
			if (fieldInfo.hasMark(LyFieldInfo.MARK_OPTIONAL))
				return result;

			// Create the appropriate error
			args = new Object[1];
			args[0] = fieldInfo.getName();
			error = new LyError(LEVEL_ERROR, LyError.TYPE_FIELD, CONTROL_MANDATORY,
					"LY_ER_FIELD_MANDATORY", args, fieldInfo, this, fieldInfo);
			addError(error);
			result = LyControlInterface.STATUS_ERROR;
		}

		return result;
	} // end method checkMandatory

	/**
	 * Checks id.
	 */
	public short checkId(LyValueSet values)
	{
		// If one of fieldInfos in values has the mark MARK_ID we must
		// check the unicity of the combination of the values for all the
		// fieldInfos having this mark.
		short result = LyControlInterface.STATUS_OK;

		if ((values == null)
				|| (!values.hasMark(LyFieldInfo.MARK_ID)))
			return result;

		if (_classInfo.isInternal())
		    return result;

		LyInfoList<LyFieldInfo> fields = _classInfo.getIdFieldInfos();
		int					    i, nf = fields.getSize();
		LyFieldInfo			    fieldInfo;
		LyValue				    value;
		LyFieldInfo[]		    idFields = new LyFieldInfo[nf];
		Object[]			    idValues = new Object[nf];
		Object				    idValue;

		// Store id field and its value inside arrays
		for (i = 0; i < nf; i++)
		{
			fieldInfo = fields.getInfo(i);
			idFields[i] = fieldInfo;
			if (values.contains(fieldInfo))
			{
				value = values.getFieldValue(fieldInfo);
				idValue = value.getValue();
			}
			else
				idValue = getValue(fieldInfo);

			if (idValue == null) // Not yet valuated
				return result;

			idValues[i] = idValue;
		}

		// Refuse id correponding to an existing object

		// If the current class inherits from other non abstract classes,
		// we must verify that the id does not correpond to an existing
		// objet in this classes.
		LyClassInfo classInfo = _classInfo;
		LyClassInfo superClass = classInfo.getSuperClass();

		while ((superClass != null) && !superClass.isAbstract())
		{
			// We must test that all fieldInfos are in the class.
			for (i = 0; i < nf; i++)
			{
				if (!superClass.contains(idFields[i]))
					break;
			}
			if (i == nf)
			{
				classInfo = superClass;
				superClass = superClass.getSuperClass();
			}
			else
				superClass = null;
		}

		// Build a filter correponding to the values.
		LyFilterExpression expression;
		LyFilter		   filter = new LyFilter(LyFilter.Type.AND);

		for (i = 0; i < nf; i++)
		{
			expression = new LyFilterExpression(idFields[i], LyFilterExpression.Condition.EQUALS,
					idValues[i]);
			filter.addElement(expression);
		}

		LyObjectSource objectSource = classInfo.getObjectSource();
		LyResponse     response = objectSource.getObjects(filter);
		int			   nr = response.getSize();

		if ((nr > 1)
				|| ((nr == 1) && !response.contains(this)))
		{
			// The combination of values must be unique
			// Add an error to the object
			LyObject			    cause = null;
			StringBuffer		    buffer = new StringBuffer();
			Object[]			    args;
			LyInfoList<LyFieldInfo> causes = new LyInfoList<LyFieldInfo>();

			args = new Object[1];
			for (i = 0; i < idFields.length; i++)
			{
				if (i > 0)
					buffer.append('/');
				if (idValues[i] != null)
					buffer.append(idValues[i].toString());
				else
					buffer.append('-');
				causes.addInfo(idFields[i]);
			}
			args[0] = buffer.toString();

			for (i = 0; i < nr; i++)
			{
				cause = (LyObject)response.getResponse(i);
				if (cause != this)
					break;
			}

			LyError error = new LyError(LyObject.LEVEL_ERROR, LyError.TYPE_CLASS, CONTROL_ID,
					"LY_ER_BADID", args, null, cause, causes);

			addError(error);
			result = LyControlInterface.STATUS_ERROR;
		}
		else
			resetClassErrors(CONTROL_ID);

		return result;
	} // end method checkId

	/**
	 * To get the filter used to check the app id values.
	 *
	 * @param     appIdFields Array of the appId fields.
	 * @param     appIdValues Array of the appId values. This array must have the same number of
	 *                        elements than appIdFields and elements must match the elements of
	 *                        appIdFields.
	 * @return    The filter. This filter must not contain an exclusive condition for this object.
	 * @exception IllegalArgumentException if appIdFields and appIdValues do not match.
	 */
	protected LyFilter getCheckAppIdFilter(LyFieldInfo[] appIdFields, Object[] appIdValues)
	{
		if ((appIdFields == null)
				|| (appIdFields.length == 0)
				|| (appIdValues == null)
				|| (appIdValues.length == 0))
			return null;

		if (appIdFields.length != appIdValues.length)
			throw new IllegalArgumentException(
				"BF AppEngine : app id fields and app id values have not the same number of elements");

		// Build a filter correponding to the values.
		int				   i, nf = appIdFields.length;
		LyFilterExpression expression;
		LyFilter		   filter = new LyFilter(LyFilter.Type.AND);

		for (i = nf - 1; i >= 0; i--)
		{
			expression = new LyFilterExpression(appIdFields[i], LyFilterExpression.Condition.EQUALS,
					appIdValues[i]);
			filter.addElement(expression);
		}

		return filter;
	}

	/**
	 * Checks application id.
	 */
	public short checkAppId(LyValueSet values)
	{
		// If one of fieldInfos in values has the mark MARK_APP_ID we must
		// check the unicity of the combination of the values for all the
		// fieldInfos having this mark.
		short result = LyControlInterface.STATUS_OK;

		if ((values == null)
				|| (!values.hasMark(LyFieldInfo.MARK_APP_ID)))
			return result;

		LyInfoList<LyFieldInfo> fields = _classInfo.getFieldInfos(LyFieldInfo.MARK_APP_ID);
		int					    i, nf = fields.getSize();
		LyFieldInfo			    fieldInfo;
		LyValue				    value;
		LyFieldInfo[]		    appIdFields = new LyFieldInfo[nf];
		Object[]			    appIdValues = new Object[nf];
		Object				    appValue;
		LyInfoList<LyFieldInfo> causes = new LyInfoList<LyFieldInfo>();

		// Store id field and its value inside arrays
		for (i = 0; i < nf; i++)
		{
			fieldInfo = fields.getInfo(i);
			appIdFields[i] = fieldInfo;
			causes.addInfo(fieldInfo);
			if (values.contains(fieldInfo))
			{
				value = values.getFieldValue(fieldInfo);
				appValue = value.getValue();
			}
			else
				appValue = getValue(fieldInfo);
			appIdValues[i] = appValue;
		}

		// Refuse appId correponding to an existing object

		// If the current class inherits from other non abstract classes,
		// we must verify that the id does not correpond to an existing
		// objet in this classes.
		LyClassInfo classInfo = _classInfo;
		LyClassInfo superClass = classInfo.getSuperClass();

		while (superClass != null)
		{
			// We must test that all fieldInfos are in the class.
			for (i = 0; i < nf; i++)
			{
				if (!superClass.contains(appIdFields[i]))
					break;
			}
			if (i == nf)
			{
				classInfo = superClass;
				superClass = superClass.getSuperClass();
			}
			else
				superClass = null;
		}

		// Build a filter correponding to the values.
		LyFilter	   filter = getCheckAppIdFilter(appIdFields, appIdValues);
		LyObjectSource objectSource = classInfo.getObjectSource();
		LyResponse     response = objectSource.getObjects(filter);
		int			   nr = response.getSize();

		if ((nr > 1)
				|| ((nr == 1) && !response.contains(this)))
		{
			// Check that all response have the correct number of AppIdFields
			boolean		 hasError = false;
			LyObjectList tempList = response.getObjects();
			int			 listSize = ((tempList == null) ? 0 : tempList.getSize());
			LyObject     tempObject;

			for (i = 0; i < listSize; i++)
			{
				tempObject = tempList.getObject(i);
				if (tempObject == this)
					continue;

				classInfo = tempObject.getClassInfo();
				fields = classInfo.getFieldInfos(LyFieldInfo.MARK_APP_ID);

				if (fields.getSize() == nf)
				{
					hasError = true;
					break;
				}
			}

			if (hasError)
			{
				// The combination of values must be unique
				// Add an error to the object
				LyObject     cause = null;
				StringBuffer buffer = new StringBuffer();
				Object[]     args;

				args = new Object[1];
				for (i = 0; i < appIdFields.length; i++)
				{
					if (i > 0)
						buffer.append(",");
					buffer.append(appIdFields[i].getName());
				}
				args[0] = buffer.toString();

				for (i = 0; i < nr; i++)
				{
					cause = (LyObject)response.getResponse(i);
					if (cause != this)
						break;
				}

				LyError error = new LyError(LyObject.LEVEL_ERROR, LyError.TYPE_CLASS,
						CONTROL_APP_ID, "LY_ER_BAD_APP_ID", args, null, cause, causes);

				addError(error);
				result = LyControlInterface.STATUS_ERROR;
			}
		}
		else
			resetClassErrors(CONTROL_APP_ID);

		return result;
	} // end method checkAppId

	/**
	 * Modifier of the selection state.
	 */
	public void setSelected(boolean selected)
	{
		setSelected(selected, true);
	}

	/**
	 * Modifier of the selection state.
	 */
	protected void setSelected(boolean selected, boolean refresh)
	{
		boolean isSelected = isSelected();

		if (isSelected != selected)
		{
			// We manage the selection state only if LY_NOTIFY_SELECT resource
			// is set to true.
			LyEnvironment environment = getEnvironment();

			if (!environment.testResource("LY_NOTIFY_SELECTION"))
				return;

			short oldState = (isSelected ? (short)1 : (short)0);

			if (selected)
				_state |= STATE_SELECTED;
			else
				_state &= ~STATE_SELECTED;

			if (refresh)
				stateChanged(LyDataSourceEvent.TYPE_SELECTION, oldState);
		}
	}

	/**
	 * Accessor to the the selection state.
	 */
	public boolean isSelected()
	{
		return ((_state & STATE_SELECTED) != 0);
	}

	/**
	 * Modifier of the visibility state.
	 */
	public void setHidden(boolean hidden)
	{
		setHidden(hidden, true);
	}

	/**
	 * Modifier of the visibility state.
	 */
	protected void setHidden(boolean hidden, boolean refresh)
	{
		boolean isHidden = isHidden();

		if (isHidden != hidden)
		{
			short oldState = (isHidden ? (short)0 : (short)1);

			if (hidden)
				_state |= STATE_HIDDEN;
			else
				_state &= ~STATE_HIDDEN;
			if (refresh)
				stateChanged(LyDataSourceEvent.TYPE_VISIBILITY, oldState);
		}
	}

	/**
	 * Accessor to the the visibility state.
	 */
	public final boolean isHidden()
	{
		return ((_state & STATE_HIDDEN) != 0);
	}

	/**
	 * Internal method to post a state change event.
	 *
	 * @param type     Type of the state change.
	 * @param oldState Old value of the state.
	 */
	protected void stateChanged(short type, short oldState)
	{
		stateChanged(type, oldState, null);
	}

	/**
	 * Internal method to post a state change event.
	 *
	 * @param type     Type of the state change.
	 * @param oldState Old value of the state.
	 * @param info     optional additional information about the state change.
	 */
	protected void stateChanged(short type, short oldState, LyEventInfo info)
	{
		LyObjectSource source = getObjectSource();

		if (!source.checkObject(this))
			return;

		LyAdvancedDataSourceEvent event = new LyAdvancedDataSourceEvent(source, type, this,
				oldState);

		event.setEventInfo(info);
		source.postEvent(event);
	}

	/**
	 * Checks if this object has a new value registered for the given field info.
	 *
	 * @param  fieldInfo The field info to check.
	 * @return True if this object has a new value for the field info given as parameter, false
	 *         otherwise.
	 */
	public boolean hasNewValue(LyFieldInfo fieldInfo)
	{
		fieldInfo = _classInfo.getRealFieldInfo(fieldInfo);

		LyValueSet	newValues = _newValues;

		return ((newValues != null) && newValues.contains(fieldInfo));
	}

	/**
	 * Checks if this object has some new values.
	 *
	 * @return True if this object has some new values, false otherwise.
	 */
	public boolean hasNewValues()
	{
		LyValueSet	newValues = _newValues;

		return ((newValues != null) && (newValues.getSize() != 0));
	}

	/**
	 * Checks if this object has a pending value registered for the given field info.
	 *
	 * @param  fieldInfo The field info to check.
	 * @return True if this object has a pending value for the field info given as parameter, false
	 *         otherwise.
	 */
	public boolean hasPendingValue(LyFieldInfo fieldInfo)
	{
		fieldInfo = _classInfo.getRealFieldInfo(fieldInfo);

		LyValueSet	pendingValues = _pendingValues;

		return ((pendingValues != null) && pendingValues.contains(fieldInfo));
	}

	/**
	 * Checks if this object has some pending values.
	 *
	 * @return True if this object has some pending values, false otherwise.
	 */
	public boolean hasPendingValues()
	{
		LyValueSet	pendingValues = _pendingValues;

		return ((pendingValues != null) && (pendingValues.getSize() != 0));
	}

	/**
	 * Gets label for given field and value.
	 *
	 * @param  fieldInfo
	 * @param  fieldValue
	 * @return The label.
	 */
	public String getLabel(LyFieldInfo fieldInfo, Object fieldValue)
	{
		LyInfoLabel label = fieldInfo.getLabel(LyInfoLabel.Type.VALUE);

		if (label != null)
		{
			LyApplication application = getApplication();
			LyLabelDaemon daemon = application.getDaemon(label);

			if (daemon != null)
				return daemon.getString(this);
			else
				return label.getString(this);
		}
		else
			return fieldInfo.toName(fieldValue);
	}

	/**
	 * Gets short label for given field and value.
	 *
	 * @param  fieldInfo
	 * @param  fieldValue
	 * @return The short label.
	 */
	public String getShortLabel(LyFieldInfo fieldInfo, Object fieldValue)
	{
		LyInfoLabel label = fieldInfo.getLabel(LyInfoLabel.Type.SHORT_VALUE);

		if (label != null)
		{
			LyApplication application = getApplication();
			LyLabelDaemon daemon = application.getDaemon(label);

			if (daemon != null)
				return daemon.getString(this);
			else
				return label.getString(this);
		}
		else
			return fieldInfo.toShortName(fieldValue);
	}

	/**
	 * Method for controlling this object before its creation or set. Apply all class and field
	 * controls on values of the object.
	 *
	 * @return The global status returned by controls.
	 */
	public short control()
	{
		return control((LyValueSet)null, false);
	}

	/**
	 * Method changes made on this object following an external notification.
	 *
	 * @param oldValues Old values for all the changes made on this object.
	 */
	public short control(LyValueSet oldValues)
	{
		return control(oldValues, false);
	}

	/**
	 * Method changes made on this object following an external notification.
	 *
	 * @param oldValues Old values for all the changes made on this object.
	 * @param loading   Boolean indicating if controls occur on loading or not.
	 */
	public short control(LyValueSet oldValues, boolean loading)
	{
		return control(oldValues,
			(loading ? ((DEFAULT_CONTROL & ~CONTROL_DISABLED_ON_LOADING) | CONTROL_LOADING)
					 : DEFAULT_CONTROL));
	}

	/**
	 * Method changes made on this object following an external notification.
	 *
	 * @param oldValues Old values for all the changes made on this object.
	 * @param control   Controls performed on this object.
	 */
	protected short control(LyValueSet oldValues, int control)
	{
		LyInfoList<LyFieldInfo> fieldInfos = null;

		if (oldValues != null)
		{
			fieldInfos = new LyInfoList<LyFieldInfo>();

			int		    i, n = oldValues.getSize();
			LyFieldInfo fieldInfo;
			LyValue     fieldValue;

			for (i = 0; i < n; i++)
			{
				fieldValue = oldValues.getValue(i);
				fieldInfo = (LyFieldInfo)fieldValue.getField();
				fieldInfos.addInfo(fieldInfo);
			}
		}

		return control(fieldInfos, control);
	}

	/**
	 * Method changes made on this object following an external notification.
	 *
	 * @param fieldInfos List of fieldInfos to control.
	 * @param control    Controls performed on this object.
	 */
	protected short control(LyInfoList<LyFieldInfo> fieldInfos, int control)
	{
		if (control == CONTROL_NONE)
			return LyControlInterface.STATUS_OK;
		if ((_state & STATE_IN_CONTROL) != 0)
			return LyControlInterface.STATUS_OK;

		_state |= STATE_IN_CONTROL;

		if (fieldInfos == null)
			fieldInfos = _classInfo.getFieldInfos();

		int n = ((fieldInfos == null) ? 0 : fieldInfos.getSize());

		if (n == 0)
			return LyControlInterface.STATUS_OK;

		LyValueSet  newValues = new LyValueSet(false);
		LyFieldInfo fieldInfo;
		short	    status = LyControlInterface.STATUS_OK, controlStatus;
		int		    i;

		if ((control & CONTROL_FIELD_CONTROLS) > 0)
		{
			for (i = 0; i < n; i++)
			{
				fieldInfo = fieldInfos.getInfo(i);

				// Don't access to value now, it would compute it
				// and local or partial values must be computed only when necessary
				controlStatus = callFieldControls(fieldInfo, newValues,
						(control & ~CONTROL_MANDATORY), false);
				if ((controlStatus != LyControlInterface.STATUS_OK)
						&& ((status == LyControlInterface.STATUS_OK)
							|| (controlStatus < status)))
					status = controlStatus;
			}
		}

		if ((control & CONTROL_CLASS_CONTROLS) > 0)
		{
			controlStatus = callClassControls(newValues, control);
			if ((controlStatus != LyControlInterface.STATUS_OK)
					&& ((status == LyControlInterface.STATUS_OK)
						|| (controlStatus < status)))
				status = controlStatus;
		}

		_state &= ~STATE_IN_CONTROL;
		return status;
	} // end method control

	/**
	 * Internal method to post an error add event.
	 */
	protected void errorAdded(LyError error)
	{
		LyDataSourceEvent event;
		LyObjectSource    source = getObjectSource();

		// We do not send errors on objects before they are created
		if (_peer == null)
		{
			if (!source.checkObject(this))
				return;
		}

		event = new LyDataSourceEvent(source, LyDataSourceEvent.TYPE_ERROR_ADDED, this);
		event.setError(error);
		source.postEvent(event);
	}

	/**
	 * Internal method to post an error remove event.
	 */
	protected void errorRemoved(LyError error)
	{
		LyDataSourceEvent event;
		LyObjectSource    source = getObjectSource();

		event = new LyDataSourceEvent(source, LyDataSourceEvent.TYPE_ERROR_REMOVED, this);
		event.setError(error);
		source.postEvent(event);
	}

	/**
	 * Check skip encoding of value until next set() or create() call, or until this parameter is
	 * changed. Should be for internal use only.
	 *
	 * @return boolean encoding status. set() or create() clears this skip encoding status
	 */
	protected final boolean skipEncoding()
	{
		return ((_state & STATE_SKIP_ENCODING) != 0);
	}

	/**
	 * Check if this is a temporary object (never created in the system, just an holder for values).
	 *
	 * @return temporary state of the object.
	 */
	public final boolean isTemporary()
	{
		return ((_state & STATE_TEMPORARY) != 0);
	}

	/**
	 * Set this object as a temporary object (never created in the system, just an holder for
	 * values).
	 *
	 * @param temporary state of the object.
	 */
	public final void setTemporary(boolean temporary)
	{
		if (temporary)
			_state |= STATE_TEMPORARY;
		else
			_state &= ~STATE_TEMPORARY;
	}

	/**
	 * Skip encoding of value untill next set() or create() call, or untill this parameter is
	 * changed. Should be for internal use only.
	 *
	 * @param boolean encoding status. set() or create() clears this skip encoding status
	 */
	public final void skipEncoding(boolean encoding)
	{
		if (encoding)
			_state |= STATE_SKIP_ENCODING;
		else
			_state &= ~STATE_SKIP_ENCODING;
	}

	/**
	 * Method to force the reload of the object.
	 *
	 * @param fieldInfos List of fieldInfos to reload.
	 */
	public void reload(LyInfoList<LyFieldInfo> fieldInfos)
	{
		if (!isCreated())
			return;

		LyObjectSource	    source;
		boolean			    synchronous;
		Vector<LyFieldInfo> fields = ((fieldInfos == null) ? null : fieldInfos.getInfos());

		source = getObjectSource();
		synchronous = source.isSynchronous();

		source.setSynchronous(true);
		source.reloadObject(getObjectId(), fields, null);
		source.setSynchronous(synchronous);
	}

	/**
	 * Context type during modification state of this object : this information may be used by the
	 * LyContext to indicate the kind of modification (User or by code) for read only attributes The
	 * default value -1 means 'undefined' or not within a context.
	 *
	 * @param type one of the LyContext.TYPE_xxx or -1 if undefined.
	 */
	public void setContextType(LyContext.Type type)
	{
		_setContextType = type;
	}

	/**
	 * Controls parent of this object (used for structs).
	 */
	void controlParent(int control)
	{
		if ((control & CONTROL_LOADING) > 0)
			// Don't control parent on loading
			return;

		if (_classInfo.getParent() instanceof LyStructFieldInfo)
		{
			// Parent field is a struct or an array of structs
			LyFieldInfo fieldInfo = getParentFieldInfo();
			LyObject    parent = getParentOfStruct();

			if ((parent != null) && (fieldInfo != null))
			{
				Object value = parent.getValue(fieldInfo);

				if ((control & CONTROL_CLASS_CONTROLS) == CONTROL_CLASS_CONTROLS)
				{
					if (!parent.isModified())
					{
						LyValueSet values = new LyValueSet(false);

						values.addValue(fieldInfo, value);
						parent.callClassControls(values, control);
					}
				}

				if ((control & CONTROL_FIELD_CONTROLS) > 0)
					parent.callFieldControls(fieldInfo, value, control);
			}
		}
	} // end method controlParent

	/**
	 * Gets the LyObject that contains this LyObject as a struct value.
	 */
	LyObject getParentOfStruct()
	{
		if (!(_classInfo.getParent() instanceof LyStructFieldInfo))
			return null;

		LyObject		  parent = null;
		LyStructFieldInfo structFieldInfo = (LyStructFieldInfo)_classInfo.getParent();

		if (_parentReference != null)
		{
			parent = _parentReference.get();
			if (parent != null)
			    return parent;
		}

		// Don't try to access to parent if the object is just local
		if (getPeer() == null)
			return null;

		Vector<Object> parentInfos = structFieldInfo.getParentInfos(this);
		LyFieldInfo    parentFieldInfo;

		if (parentInfos != null)
		{
			parent = (LyObject)parentInfos.elementAt(0);
			parentFieldInfo = (LyFieldInfo)parentInfos.elementAt(1);
			setParent(parent, parentFieldInfo);
		}

		return parent;
	} // end method getParentOfStruct

	/**
	 * Gets the struct fieldInfo or the array field info that contains structs of the parent that
	 * points to this object.
	 */
	LyFieldInfo getParentFieldInfo()
	{
		if (!(_classInfo.getParent() instanceof LyStructFieldInfo))
			return null;

		if (_parentFieldInfo != null)
			return _parentFieldInfo;

		// Called just to compute the parent fieldInfo.
		getParentOfStruct();

		return _parentFieldInfo;
	}

	/**
	 * To know if this object has been modified.
	 */
	boolean isModified()
	{
		return ((_newValues != null) || (_pendingValues != null));
	}

	/**
	 * To know if the specified context may be used to send an operation.
	 *
	 * @param  context The context to test.
	 * @return true if the specified context may be used, false otherwise.
	 */
	boolean canUseContext(LyContext context)
	{
		return ((context != null)
				&& (!context.isEnded())
				&& (context.getStatus() != LyContext.Status.COMMITTING));
	}

	/**
	 * Tests if the fieldinfo is virtual : computed by the Java code of the generated interface. The
	 * overridden methods does not store the result in the underlying LyObject, and values from
	 * memory / pending / newValues have no meaning when this methods returns true. This method
	 * always returns false by default but may be overridden in the subclasses (java interface)
	 *
	 * @param  fieldInfo the fieldInfo to test
	 * @return true when the field is computed by the interface
	 */
	protected boolean isVirtualFieldInfo(LyFieldInfo fieldInfo)
	{
		return false;
	}

	/**
	 * Modifier of the class controlled state.
	 */
	protected void setClassControlled(boolean classControlled)
	{
		boolean isClassControlled = ((_state & STATE_CLASS_CONTROLLED) != 0);

		if (isClassControlled != classControlled)
		{
			if (classControlled)
				_state |= STATE_CLASS_CONTROLLED;
			else
			{
				_state &= ~STATE_CLASS_CONTROLLED;

				LyObject parent = getParent();

				if ((parent != null) && !parent.isModified())
					parent.setClassControlled(false);
			}
		}
	}

	/**
	 * Checks if this instance is found on its object source cache.
	 *
	 * @return True if it exists, false otherwise.
	 */
	public boolean checkInstance()
	{
		LyObjectSource objectSource = getObjectSource();

		return ((_id == null) ? false : (objectSource.getObject(getObjectId()) == this));
	}

	/**
	 * Method called when a modification notification has been received. This method should be
	 * overriden to copmpute local fields.
	 *
	 * @param values List of old values before the modification.
	 */
	public void modified(LyValueSet values)
	{
	}

	/**
	 * Method called when objectId of the object has been rehashed to update daemons.
	 *
	 * @param oldObjectId Old object id.
	 * @param newObjectId New object id.
	 */
	protected void idRehashed(LyObjectId oldObjectId, LyObjectId newObjectId)
	{
		LyFieldInfo			    fieldInfo;
		LyInfoList<LyFieldInfo> fieldInfos = _classInfo.getFieldInfos();
		int					    i, n = ((fieldInfos == null) ? 0 : fieldInfos.getSize());
		LyApplication		    application = getApplication();
		LyDaemon			    daemon;
		LyInfoLabel			    label;

		for (i = 0; i < n; i++)
		{
			fieldInfo = fieldInfos.getInfo(i);
			daemon = application.getDaemon(fieldInfo);
			if (daemon != null)
				daemon.idRehashed(oldObjectId, newObjectId);

			label = fieldInfo.getLabel(LyInfoLabel.Type.VALUE);
			if (label != null)
			{
				daemon = application.getDaemon(label);
				if (daemon != null)
					daemon.idRehashed(oldObjectId, newObjectId);
			}

			label = fieldInfo.getLabel(LyInfoLabel.Type.SHORT_VALUE);
			if (label != null)
			{
				daemon = application.getDaemon(label);
				if (daemon != null)
					daemon.idRehashed(oldObjectId, newObjectId);
			}

			label = fieldInfo.getLabel(LyInfoLabel.Type.GRAMMATICAL);
			if (label != null)
			{
				daemon = application.getDaemon(label);
				if (daemon != null)
					daemon.idRehashed(oldObjectId, newObjectId);
			}

			label = fieldInfo.getLabel(LyInfoLabel.Type.SEMANTIC);
			if (label != null)
			{
				daemon = application.getDaemon(label);
				if (daemon != null)
					daemon.idRehashed(oldObjectId, newObjectId);
			}
		}

		label = _classInfo.getLabel(LyInfoLabel.Type.NAME);
		if (label != null)
		{
			daemon = application.getDaemon(label);
			if (daemon != null)
				daemon.idRehashed(oldObjectId, newObjectId);
		}
	} // end method idRehashed

    /**
     * Modifier of the loaded state.
     */
    protected void setFullyLoaded(boolean loaded)
    {
        boolean isLoaded = isFullyLoaded();

        if (isLoaded != loaded)
        {
            if (loaded)
                _state |= STATE_FULLY_LOADED;
            else
                _state &= ~STATE_FULLY_LOADED;
        }
    }

    /**
     * Accessor to the the loaded state.
     */
    public final boolean isFullyLoaded()
    {
        return ((_state & STATE_FULLY_LOADED) != 0);
    }
} // end class LyObject
