/*
 * LyObjectId.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.lang.ref.*;
import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.info.*;
import leon.misc.*;


/**
 * This class implements object identifiers. An instance of this class identifies uniquely an object
 * within an application.
 *
 * <p>Object ids are used for identifying both logical or physical objects.</p>
 *
 * <p>Object ids contains :</p>
 *
 * <ul>
 * <LI>a class to which the identified object belongs, This is a LyClassInfo for a logical object
 * (LyObject) or a LyPeerClass for a physical object.</LI>
 * <LI>a string that identifies the object within the class. This string is the result of
 * concatenation of all values corresponding to id fields for the identified object.</LI>
 * </ul>
 *
 * @see  LyClass
 * @see  LyClassInfo
 * @see  leon.peer.LyPeerClass
 * @see  LyObject
 */
public final class LyObjectId
{
    /** File version. */
    public static final String __VERSION = "$Id: LyObjectId.java 28980 2010-12-22 10:53:09Z jlsaouli $";

    /** Separators used when object are identify by more than one field. */
    public static char FIELD_SEP = (char)3;
    public static char OLD_FIELD_SEP = LyApplication.FIELD_SEP;
    public static char VALUE_SEP = (char)4;

    /**
     * Displayable separators used when object are identify by more than one field and the
     * resquested string should not contain specific characters.
     */
    public static char VISIBLE_FIELD_SEP = LyApplication.VISIBLE_FIELD_SEP;
    public static char VISIBLE_VALUE_SEP = LyApplication.VISIBLE_VALUE_SEP;

    /** Separator used when to separate class id from object id within this class. */
    public static char CLASS_SEP = LyApplication.CLASS_SEP;

    /** Internal table used to speed up getFullString method. */
    private static char[] _charTable = new char[100];

    /**
     * Utility method to calculate an id string for an object.
     */
    static protected String toString(Object object)
    {
        if (object == null)
            return "";

        if (!object.getClass().isArray())
        {
            if (object instanceof Date)
            { // inconsistency on day of week in date.toString()
                String resu = LyDateFormatter.toString((Date)object);

                return resu;
            }
            else
                return object.toString();
        }

        Object[]     array = (Object[])object;
        int          i, n = array.length;
        StringBuffer buffer = new StringBuffer("[");
        String       str;

        for (i = 0; i < n; i++)
        {
            str = toString(array[i]);
            if (i > 0)
                buffer.append(VALUE_SEP);

            if (str != null)
                buffer.append(str);
        }

        buffer.append("]");
        return buffer.toString();
    }

    /**
     * Utility method to calculate an id string for an object.
     */
    static protected Object getIdValue(LyClassInfo classInfo, LyFieldInfo fieldInfo)
    {
        LyApplication                  application = classInfo.getApplication();
        LyApplicationBehaviorInterface behavior = application.getApplicationBehavior();
        Object                         result;

        if (fieldInfo instanceof LyNumberFieldInfo)
        {
            LyNumberFieldInfo.Type type = ((LyNumberFieldInfo)fieldInfo).getType();

            if (type == LyNumberFieldInfo.Type.INT)
                result = new Integer(behavior.getNextInt(application, classInfo));
            else if (type == LyNumberFieldInfo.Type.SHORT)
                result = new Integer(behavior.getNextShort(application, classInfo));
            else if (type == LyNumberFieldInfo.Type.FLOAT)
                result = new Float(behavior.getNextFloat(application, classInfo));
            else
                result = new Long(behavior.getNextLong(application, classInfo));
        }
        else if (fieldInfo instanceof LyTextFieldInfo)
        {
            String prefix = classInfo.getId();

            result = prefix + "#" + classInfo.getNextCount();
        }
        else if (fieldInfo instanceof LyDateFieldInfo)
            result = new Date();
        else if (fieldInfo instanceof LyRelationFieldInfo)
        {
            LyRelationFieldInfo     relation = (LyRelationFieldInfo)fieldInfo;
            LyClassInfo             relationClassInfo = relation.getClassInfo(0);
            LyInfoList<LyFieldInfo> idFieldInfos = relationClassInfo.getIdFieldInfos();
            int                     i, n = idFieldInfos.getSize();
            LyFieldInfo             idFieldInfo;

            if (n == 1)
            {
                idFieldInfo = idFieldInfos.getInfo(0);
                result = getIdValue(relationClassInfo, idFieldInfo);
            }
            else
            {
                result = new Object[n];
                for (i = 0; i < n; i++)
                {
                    idFieldInfo = idFieldInfos.getInfo(i);
                    ((Object[])result)[i] = getIdValue(relationClassInfo, idFieldInfo);
                }
            }

            result = new LyObjectId(relationClassInfo, result);
        }
        else if (fieldInfo instanceof LyChoiceFieldInfo)
            result = ((LyChoiceFieldInfo)fieldInfo).getDefaultValue();
        else
            result = new Integer(behavior.getNextInt(application, classInfo));

        return result;
    } // end method getIdValue

    /* Protected attibutes */

    /** Class of the object represented by this id. */
    LyClass _class;

    /** String that identifies the object within the class. */
    String _stringId;

    /**
     * The object identifier represented as a string from which the object id can be reconstructed
     * (and that is, in particular, enclosed in XML description of views built by Leonardi). It is
     * the result of concatenation of class id (LyClass.getId) and the above stringId.
     */
    String _fullString;

    /** Weak reference on object corresponding to this objectId. */
    WeakReference<Object> _objectReference;

    /**
     * Constructs an object id from an object class and an Object that identifies it within this
     * class.
     *
     * @param      objectClass  Class of the object that is represented by this object id.
     * @param      id           Object that identifies the object within the given class. This
     *                          object may be an array of Objects if the object class contains
     *                          several id fields.
     *
     * @exception  IllegalArgumentException  If the given object class is null.
     * @exception  IllegalArgumentException  If the given id is null.
     */
    public LyObjectId(LyClass objectClass, Object id)
    {
        if (objectClass == null)
            throw new IllegalArgumentException("BF AppEngine : objectClass can't be null (id : "
                + id + ")");

        if (id == null)
            throw new IllegalArgumentException("BF AppEngine : id can't be null (objectClass : "
                + objectClass + ")");

        _class = objectClass;

        // Build string id
        if (id.getClass().isArray())
        {
            if (id.getClass() != LyObjectId[].class) // FFT00744
            {
                Object[] array = (Object[])id;
                int      i, n = array.length;
                String   str;

                for (i = 0; i < n; i++)
                {
                    str = ((array[i] == null) ? "" : toString(array[i]));
                    if (_stringId == null)
                        _stringId = str;
                    else
                        _stringId = _stringId + FIELD_SEP + str;
                }
            }
            else
            {
                LyObjectId[] array = (LyObjectId[])id;

                if (array.length == 0)
                    throw new IllegalArgumentException("BF AppEngine : id can't be null (objectClass : "
                        + objectClass + ")");
                else if (array.length == 1)
                    _stringId = array[0]._stringId;
                else
                    throw new IllegalArgumentException(
                        "BF AppEngine : can not set object id from multiple ids (objectClass : "
                        + objectClass + ")");
            }
        }
        else if (id.getClass() == LyObjectId.class)
            _stringId = ((LyObjectId)id)._stringId;
        else
            _stringId = toString(id);

        // FEV00074 : if _stringId contains CLASS_SEP we remove the class part of the id.
        int idx = _stringId.indexOf(CLASS_SEP);

        if ((idx >= 0)
                && _class.getId().equals(_stringId.substring(0, idx)))
            _stringId = _stringId.substring(idx + 1);

        if (_stringId != null)
            _stringId = _stringId.intern();
    }

    /**
     * Builds an objectId in a given application from an encoded string.
     *
     * @param      application  Application in which this object id is constructed.
     * @param      objectId     Object id encoded as a string. This string must start with the class
     *                          id (LyClass.getId), followed by CLASS_SEP and by a string that
     *                          identifies the object within the class (this should be the result of
     *                          concatenation of all values corresponding to id fields of the
     *                          class).
     *
     * @exception  IllegalArgumentException  If the given string objectId is null.
     * @exception  IllegalArgumentException  If the class separator can't be find in the given
     *                                       string.
     * @exception  IllegalArgumentException  If class id extracted from the objectId can't be found
     *                                       in given application.
     */
    public LyObjectId(LyApplication application, String objectId)
    {
        if (objectId == null)
            throw new IllegalArgumentException("BF AppEngine : objectId can't be null");

        // Retrieve the class of the object
        int pos = objectId.indexOf(CLASS_SEP);

        if (pos < 0)
            throw new IllegalArgumentException("BF AppEngine : can't find class separator in given string objectId "
                + objectId);

        String classId = objectId.substring(0, pos);

        _class = application.getClassInfo(classId);

        if (_class == null)
        {
            LyDataSource dataSource = application.getDataSource(classId);

            if (dataSource != null)
                _class = dataSource.getDataClass();
        }

        if (_class == null)
            throw new IllegalArgumentException("BF AppEngine : no class found in application for the given classId "
                + classId);

        _stringId = objectId.substring(pos + 1);
        _stringId = _stringId.intern();
    }

    /**
     * Internal method to check if object id is valid.
     *
     * @param   application  Application in which this object id is checked.
     *
     * @return  True if object id is valid, false otherwise.
     */
    protected boolean check(LyApplication application)
    {
        if (_class == null)
            return false;

        if (!(_class instanceof LyClassInfo))
            return true; // Peer level is always valid

        LyEnvironment             environment = application.getEnvironment();
        Vector<? extends LyField> idFields = _class.getIdFields();

        if (idFields == null)
        {
            environment.logWarning("BF AppEngine : invalid object id " + this
                + " no id fields found on class " + _class);
            return false;
        }

        int      i, n = idFields.size();
        LyField  field;
        String[] ids = toStrings();
        int      nIds = ((ids == null) ? 0 : ids.length);

        if (n > nIds)
        {
            ids = LyString.stringToList(_stringId, OLD_FIELD_SEP);
            nIds = ((ids == null) ? 0 : ids.length);
            if (n > nIds)
            {
                environment.logWarning("BF AppEngine : invalid object id " + this
                    + " expecting " + n + " id fields, found " + nIds);
                return false;
            }

            _stringId = _stringId.replace(OLD_FIELD_SEP, FIELD_SEP);
            _stringId = _stringId.intern();
        }
        else if (n < nIds)
            return true;

        Class<?> idClass;

        for (i = 0; i < n; i++)
        {
            if ((ids[i] == null)
                    || (ids[i].length() == 0))
                continue;

            field = idFields.elementAt(i);
            idClass = field.getValueClass();
            if (idClass == String.class)
                continue;
            else if (Number.class.isAssignableFrom(idClass))
            {
                // Check if id is a valid Number
                try
                {
                    if (idClass == Integer.class)
                        Integer.valueOf(ids[i]);
                    else if (idClass == Short.class)
                        Short.valueOf(ids[i]);
                    else if (idClass == Long.class)
                        Long.valueOf(ids[i]);
                    else if (idClass == Double.class)
                        Double.valueOf(ids[i]);
                    else if (idClass == Float.class)
                        Float.valueOf(ids[i]);
                }
                catch (NumberFormatException e)
                {
                    environment.logWarning("BF AppEngine : invalid object id " + this
                        + ", invalid number format on " + ids[i]);
                    return false;
                }
            }
            else if (Date.class == idClass)
            {
                // Check if id is a valid Date
                if (LyDateFormatter.fromString(ids[i]) == null)
                {
                    environment.logWarning("BF AppEngine : invalid object id " + this
                        + ", invalid date format on " + ids[i]);
                    return false;
                }
            }
        }

        return true;
    } // end method check

    /**
     * Gets object class to which the identified object belongs.
     *
     * @return  Class of the identified object.
     */
    public LyClass getObjectClass()
    {
        return _class;
    }

    /**
     * Gets object class info to which the identified logical object belongs.
     *
     * @return  Class of the identified logical object.
     */
    public LyClassInfo getClassInfo()
    {
        return (LyClassInfo)_class;
    }

    /**
     * Gets string that identifies uniquely the object within the class. This string may contain non
     * displayable characters.
     *
     * @return  String that identifies the object within the class.
     */
    public String toString()
    {
        return _stringId;
    }

    /**
     * Gets displayable string that identifies uniquely the object within the class. This string
     * contains only displayable characters.
     *
     * @return  Displayable string that identifies the object within the class. Same string as the
     *          one returned by toString but with FIELD_SEP replaced by VISIBLE_FIELD_SEP.
     */
    public String toName()
    {
        if ((_class instanceof LyClassInfo)
                && (_objectReference != null))
        {
            LyObject object = (LyObject)_objectReference.get();

            if (object != null)
            {
                LyDataSource source = _class.getDataSource();

                if (source.checkObject(object))
                    return object.getName();
            }
        }

        String name = _stringId;

        name = name.replace(FIELD_SEP, VISIBLE_FIELD_SEP);
        name = name.replace(VALUE_SEP, VISIBLE_VALUE_SEP);
        return name;
    }

    /**
     * Gets an array of displayable strings that splits the object id using FIELD_SEP.
     *
     * @return  Array of strings corresponding to the different parts of object id (each substring
     *          should correspond to each id field of the object class).
     */
    public String[] toNames()
    {
        String name = _stringId;

        name = name.replace(VALUE_SEP, VISIBLE_VALUE_SEP);
        return LyString.stringToList(name, FIELD_SEP);
    }

    /**
     * Gets an array of strings that splits the object id using FIELD_SEP.
     *
     * @return  Array of strings corresponding to the different parts of object id (each substring
     *          should correspond to each id field of the object class).
     */
    public String[] toStrings()
    {
        return LyString.stringToList(_stringId, FIELD_SEP);
    }

    /**
     * Returns a hash code value corresponding this object id.
     *
     * @return  Hash code value for this object id.
     */
    public int hashCode()
    {
        return ((_stringId == null) ? 0 : _stringId.hashCode());
    }

    /**
     * Compares this object id with another object id using default comparison.
     *
     * @param   objectId  Object id to compare with this object id.
     *
     * @return  0 If the two object ids are identical, a positive number if this object id is
     *          "greater" than object id. Otherwise, a negative number is returned.
     */
    public int compare(LyObjectId objectId)
    {
        if (objectId == null)
            return 1;

        if (objectId == this)
            return 0;

        String strId = getFullString();
        String objectStrId = objectId.getFullString();

        return strId.compareTo(objectStrId);
    }

    /**
     * Compares the given object with this object id for equlity. This method is overriden to
     * compare class id and string id.
     *
     * @return  True if this object id is the same as the given object, false otherwise.
     */
    public boolean equals(Object object)
    {
        if (!(object instanceof LyObjectId))
            return false;

        if (object == this)
            return true;

        return equals((LyObjectId)object, true);
    }

    /**
     * Compares the given object with this object id for equlity. This method is overriden to
     * compare class id and string id.
     *
     * @param   object
     *
     * @return  True if this object id is the same as the given object, false otherwise.
     */
    public boolean equals(LyObjectId objectId, boolean allowSuperClass)
    {
        if (objectId == this)
            return true;

        if ((objectId == null)
                || (_stringId != objectId.toString()))
            return false;

        LyClass objectClass = objectId.getObjectClass();

        if (allowSuperClass
                && (_class instanceof LyClassInfo)
                && (objectClass instanceof LyClassInfo))
        {
            LyClassInfo classInfo = (LyClassInfo)_class;
            LyClassInfo objectClassInfo = (LyClassInfo)objectClass;

            return ((classInfo == objectClassInfo)
                    || (objectClassInfo.inheritsFrom(classInfo))
                    || (classInfo.inheritsFrom(objectClassInfo)));
        }
        else
            return (_class == objectClass);
    }

    /**
     * Gets a string representation of this object id. This string contains the class id concatened
     * to CLASS_SEP, followed by the string id.
     *
     * @return  String representation of this object id.
     */
    public String getFullString()
    {
        if (_fullString != null)
            return _fullString;

        // Optimized to build fullString
        String classId = _class.getId();
        int    nc = classId.length();
        int    ns = _stringId.length();

        if (_charTable.length < (nc + ns + 1))
            _charTable = new char[nc + ns + 1];

        classId.getChars(0, nc, _charTable, 0);
        _charTable[nc] = CLASS_SEP;
        _stringId.getChars(0, ns, _charTable, nc + 1);

        _fullString = new String(_charTable, 0, nc + ns + 1);
        return _fullString.intern();
    }

    /**
     * Retrieves object identified by this object id from the data source of object class.
     *
     * @return  Object if it's found, null otherwise.
     */
    public Object getObject()
    {
        Object        object;
        LyDataSource  source = _class.getDataSource();
        LyDataSession session = null;

        if (_class instanceof LyClassInfo)
        {
            LyApplication application = ((LyClassInfo)_class).getApplication();

            session = application.getCurrentSession();

            if (session != null)
                source = session.getDataSource(_class);
        }

        if (source == null)
            return null;

        if (_objectReference != null)
        {
            object = _objectReference.get();
            if (object != null)
            {
                try
                {
                    LyCache cache = source.getCache();

                    if ((cache != null)
                            && (cache.getObject(this) == object))
                        return object;
                }
                catch (LyNoSuchObjectException e)
                {
                }
            }

            setObject(null);
        }

        object = source.getObject(this);
        if (object != null)
            setObject(object);

        return object;
    } // end method getObject

    /**
     * Internal method to store given object as referenced object of this id.
     */
    void setObject(Object object)
    {
        _objectReference = new WeakReference<Object>(object);
    }

    /**
     * Retrieves logical object (LyObject) identified by this object id from the object source of
     * the class info. Same as getObject but casted to LyObject.
     *
     * @return  Logical object if it's found, null otherwise.
     */
    public LyObject getLogicalObject()
    {
        LyObject result = (LyObject)getObject();

        if ((result != null) && (result.getClassInfo() != _class))
        {
            _class = result.getClassInfo();
            _fullString = null;
        }

        return result;
    }

    /**
     * Retrieves logical object identified by this object id.
     *
     * @deprecated  Replaced by getLogicalObject.
     */
    public LyObject getLyObject()
    {
        return getLogicalObject();
    }

    /**
     * Checks if object corresponding to this object id exists on the data source of object class.
     *
     * @return  True if an object corresponding to this object id is found, false otherwise.
     */
    public boolean checkObject()
    {
        LyDataSource source = _class.getDataSource();

        return ((source == null) ? false : source.checkObject(this));
    }
} // end class LyObjectId
