/*
 * LyObjectList.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;
import leon.app.*;
import leon.info.*;
import leon.misc.*;


/**
 * This class implements a list of logical objects (LyObject).
 *
 * <p>This list can be sorted by a specified sort. If this sort is null or not enough
 * discriminating, a default sort is used to be sure that we have a full order (sorting the list
 * several times will always produce the same result). The default sort can be done either on
 * objects names or on objects ids.</p>
 *
 * <p>It is also possible to turn these lists to not sorted.</p>
 *
 * <p>Objects list can be dynamic (see classes that extend LyWorkSpace). This means that
 * modifications made on the content of the list can be notified to registered listeners.</p>
 *
 * @see LyObject
 */
public abstract class LyObjectList
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyObjectList.java 36457 2012-08-09 13:35:10Z jlsaouli $";

	/** List of objects. */
	protected _LyObjects	   _objects;

	/** Current sort. */
	protected LySort		   _sort;

	/** Boolean indicating if default sorting on objects is done on object names or ids. */
	protected boolean		   _sortByName;

	/** Boolean indicating if sorting is required or not. */
	protected boolean		   _sorted;

	/** For optimization : Check if objects are already contained before adding them in list. */
	protected boolean		   _checkExistence;

	/**
	 * Constructs an empty list.
	 */
	protected LyObjectList()
	{
		_objects = null;
		_sort = null;
		_sortByName = false;
		_sorted = true;
		_checkExistence = true;
	}

	/**
	 * Constructs a list with a single object.
	 *
	 * @param object Object added in the constructed list.
	 */
	protected LyObjectList(LyObject object)
	{
		this();

		_objects = new _LyObjects(1);
		add(object);
	}

	/**
	 * Return a vector of objects contained in this list.
	 *
	 * @return Vector containing all objects of this list. This list is a copy of the internal list
	 *         and can be freely modified.
	 */
	public final Vector<LyObject> getObjects()
	{
		if (_objects == null)
			return null;

		// Return a clone of objects list to avoid manipulation of internal list
		return new Vector<LyObject>(_objects);
	}

    /**
     * Return an array of object ids contained in this list.
     *
     * @return Array of object ids for objects of this list.
     */
    public final LyObjectId[] toObjectIds()
    {
        if (_objects == null)
            return null;

        int             i, n = _objects.size();
        LyObjectId[]    ids = new LyObjectId[n];
        
        for (i = 0; i < n; i++)
            ids[i] = _objects.getObject(i).getObjectId();
        return ids;
    }
    
	/**
	 * Gets the current sort of this list.
	 *
	 * @return Current sort of this list (may be null if default sorting is used).
	 */
	public LySort getSort()
	{
		return _sort;
	}

	/**
	 * Checks whether this list is dynamic or not. A dynamic list is a list that notifies changes
	 * made on its content.
	 *
	 * @return True if this list is dynamic, false if it's not.
	 */
	public boolean isDynamic()
	{
		return false;
	}

	/**
	 * Checks whether this list is sorted or not.
	 *
	 * @return True if this list is sorted, false if it's not.
	 */
	public boolean isSorted()
	{
		return _sorted;
	}

	/**
	 * Sets the current sort of this list. The list is immediatly resorted.
	 *
	 * @param sort New sort for this list.
	 */
	public void setSort(LySort sort)
	{
		_sort = sort;
		_sorted = true;
		resort();
	}

	/**
	 * Specifies if the defaut sorting is done on objects names or on objects ids. Default sorting
	 * if used when comparing two objects with the current sort don't discriminate them (or if it's
	 * null), to have a full order of our list.
	 *
	 * @param byName True if default sorting is done with objects names, false if objects ids are
	 *               used.
	 */
	public void setSortByName(boolean byName)
	{
		if (_sortByName == byName)
			return;

		_sortByName = byName;
		resort();
	}

	/**
	 * Sets boolean indicating if this list has to be sorted or not (even if no explicit sort is
	 * specified).
	 *
	 * @param sorted Boolean specifying if this list has to be sorted or not.
	 */
	public void setSorted(boolean sorted)
	{
		if (_sorted == sorted)
			return;

		_sorted = sorted;
		resort();
	}

	/**
	 * Internal method to resort this list according to the current sort parameters. Does nothing if
	 * this list is specified as not sorted.
	 */
	protected void resort()
	{
		if (_sorted && (getSize() > 1))
			_objects.resort();
	}

	/**
	 * Called by the garbage collector on an object when garbage collection determines that there
	 * are no more references to the object. Overriden to unlock objects contained in this list.
	 *
	 * @see LyObject#unlock
	 */
	protected void finalize()
		throws Throwable
	{
		if (isEmpty())
			return;

		LyObject	  object = _objects.getObject(0);
		LyApplication application = object.getApplication();

		if ((application == null) || !application.isValid())
			return;

		// We must unlock all the objects of this list.
		unlockObjects();
	}

	/**
	 * Unlock objects of this list.
	 *
	 * @see LyObject#unlock
	 */
	final void unlockObjects()
	{
		if (_objects == null)
			return;

		synchronized (this)
		{
			int		 i, n = getSize();
			LyObject object;

			for (i = 0; i < n; i++)
			{
				object = _objects.getObject(i);
				objectRemoved(object);
			}
		}
	}

	/**
	 * Return the list of all class infos to which objects of this list belong. This method is
	 * declared abstract to force sub-classes to give a fast implementation.
	 *
	 * @return List of classes to which objects of this list belong.
	 */
	public abstract LyInfoList<LyClassInfo> getClasses();

	/**
	 * Internal method to swap the two objects at the given indexes.
	 *
	 * @param lo First index.
	 * @param hi Second index.
	 */
	void swapObjects(int lo, int hi)
	{
	}

	/* Management of the content of the list */

	/**
	 * Resets this list. By default, this method clears all objects of the list.
	 */
	public void reset()
	{
		unlockObjects();
		_objects = null;
	}

	/**
	 * Checks if the list is empty or not.
	 *
	 * @return True is the list is empty, false otherwise.
	 */
	public boolean isEmpty()
	{
		return (getSize() == 0);
	}

    /**
     * Returns the number of objects of the given classInfo contained in this list.
     * @param classInfo Class info for which the number of objects is requested.
     * 
     * @return Number of objects in this list of the given class info.
     */
    public int getSize(LyClassInfo classInfo)
    {
        if (_objects == null)
            return 0;

        int  size = 0;

        synchronized (this)
        {
            LyObject object;
            int      i, n = getSize();

            for (i = 0; i < n; i++)
            {
                object = getObject(i);
                if (object.instanceOf(classInfo))
                    size++;
            }
        }

        return size;
    }
    
	/**
	 * Returns the number of objects contained in this list.
	 *
	 * @return Number of objects in this list.
	 */
	public int getSize()
	{
		return ((_objects == null) ? 0 : _objects.size());
	}

	/**
	 * Test if the specified object is contained in this list.
	 *
	 * @param     object Checked object.
	 * @return    True if the object is in this list, false otherwise.
	 * @exception IllegalArgumentException If the given object is null.
	 */
	public boolean contains(LyObject object)
	{
		return (getIndex(object) >= 0);
	}

	/**
	 * Register object in this list.
	 */
	protected void objectAdded(LyObject object)
	{
	}

	/**
	 * Register object in this list.
	 */
	protected void objectRemoved(LyObject object)
	{
	}

	/**
	 * Internal method to add an object in the list. The object is added in this list according to
	 * sort parameters.
	 *
	 * @param     object Object to add in this list.
	 * @return    If the object was not in this list, returns a positive index giving the position
	 *            where the object was added. Otherwise, returns a negative index giving the
	 *            position minus 1 where the same object was found.
	 * @exception IllegalArgumentException If the given object is null.
	 */
	synchronized int add(LyObject object)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		if (_objects == null)
			_objects = new _LyObjects(1);

		if (!_sorted)
		{
			if (_checkExistence)
			{
				int pos = getIndex(object);

				if (pos >= 0)
					return -1 - pos;
			}

			_objects.addElement(object);
			objectAdded(object);
			return getSize() - 1;
		}

		// Add the object at the right position
		int		 i, iMin = 0, iMax = getSize() - 1;
		LyObject cursor;

		while (iMax >= iMin)
		{
			i = (iMin + iMax) / 2;
			cursor = _objects.getObject(i);

			if (_checkExistence && object.equals(cursor))
				return -1 - i;

			// Compare objects with the current sort
			if (object.compare(cursor, _sort, _sortByName) >= 0)
				iMin = i + 1;
			else
				iMax = i - 1;
		}

		_objects.insertElementAt(object, iMin);
		objectAdded(object);
		return iMin;
	} // end method add

	/**
	 * Internal method to add an object in the list at a given position.
	 *
	 * @param     object Object to add in this list.
	 * @param     index  Position where the object is added. This parameter is ignored if the list
	 *                   is sorted. A negative index means at the end of the list.
	 * @return    If the object was not in this list, returns a positive index giving the position
	 *            where the object was added. Otherwise, returns a negative index giving the
	 *            position minus 1 where the same object was found.
	 * @exception IllegalArgumentException If the given object is null.
	 */
	synchronized int add(LyObject object, int index)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		if (_sorted)
			// The list becomes not sorted
			_sorted = false;

		if (_checkExistence)
		{
			int pos = getIndex(object);

			if (pos >= 0)
				return -1 - pos;
		}

		if (_objects == null)
			_objects = new _LyObjects(1);

		int n = _objects.size();

		if ((index >= 0) && (index < n))
			_objects.insertElementAt(object, index);
		else
		{
			_objects.addElement(object);
			index = n;
		}

		objectAdded(object);
		return index;
	} // end method add

	/**
	 * Removes an object from the list. When removed from the list, the object is unlocked.
	 *
	 * @param     object Object to remove from this list.
	 * @return    Index where the object was found in this list, -1 if object was not in the list.
	 * @see       LyObject#unlock
	 */
	synchronized int remove(LyObject object)
	{
		int index = indexOf(object);

		if (index < 0)
			return -1;

		_objects.removeElementAt(index);

		if (_objects.size() == 0)
			_objects = null;

		objectRemoved(object);
		return index;
	}
	
	/**
     * Removes an object from the list. When removed from the list, the object is unlocked.
     *
     * @param     object Object to remove from this list.
     * @return    Index where the object was found in this list, -1 if object was not in the list.
     * @see       LyObject#unlock
     */
    synchronized LyObject remove(int index)
    {
        if (index < 0)
            return null;
        if (index >= getSize())
            return null;

        LyObject    object = _objects.remove(index);

        if (_objects.size() == 0)
            _objects = null;

        objectRemoved(object);
        return object;
    }

	/**
	 * Internal method to change some values of an object in this list. This method removes the
	 * given object and readds it in this list to give it its right position regarding to sort
	 * parameters specified on this list.
	 *
	 * @param  object Object that is changed.
	 * @param  values Values changed
	 * @return True if the object position changed, false otherwise.
	 */
	protected boolean change(LyObject object, LyValueSet values)
	{
		if (!_sorted
		        || (getSize() == 1))
			return false;

		boolean moved = false;

		synchronized (this)
		{
		    int   index = indexOf(object);
		    
		    remove(index);
			moved = (add(object) != index);
		}

		return moved;
	}

	/**
	 * Returns the object at the specified index.
	 *
	 * @param     idx Index of the searched object.
	 * @return    Object at the given index.
	 * @exception IndexOutOfBoundsException If the given index is negative or greater or equal to
	 *                                      the number of objects in this list.
	 */
	public final synchronized LyObject getObject(int index)
	{
		return (_objects == null ? null : _objects.getObject(index));
	}

	/**
	 * Returns the object with the specified identifier.
	 *
	 * @param     id Id of the searched object.
	 * @return    Object with the specified identifier if it was found, else null.
	 * @exception IllegalArgumentException If the given id is null.
	 */
	public LyObject getObject(LyObjectId objectId)
	{
		if (objectId == null)
			throw new IllegalArgumentException("BF AppEngine : objectId can't be null");
		if (_objects == null)
			return null;

		synchronized (this)
		{
			int		 i, n = getSize();
			LyObject object;

			for (i = 0; i < n; i++)
			{
				object = _objects.getObject(i);
				if (object.getObjectId().equals(objectId))
					return object;
			}
		}

		return null;
	}

	/**
	 * Returns the object with the specified identifier given as a string within the specified
	 * application.
	 *
	 * @param  id Id of the search object given as a string.
	 * @return Object with the specified identifier if it was found, else null.
	 */
	public LyObject getObject(LyApplication application, String id)
	{
		LyObjectId objectId = new LyObjectId(application, id);

		return getObject(objectId);
	}

	/**
     * Returns the position of the specified object using
     * This method is used instead of getIndex when searched object is changed,
     * because the fields values on which the list is sorted could have been changed.
     *
     * @param     object     Object searched in this list.
     * @return    Index of the object in this list if it was found, -1 if it was not found.
     * @exception IllegalArgumentException If the given object is null.
     * @see getIndex
     */
    public int indexOf(LyObject object)
    {
        return (_objects == null ? -1 : _objects.indexOf(object));
    }
    
	/**
	 * Returns the position of the specified object.
	 *
	 * @param     object Object searched in this list.
	 * @return    Index of the object in this list if it was found, -1 if it was not found.
	 * @exception IllegalArgumentException If the given object is null.
	 * @see indexOf
	 */
	public int getIndex(LyObject object)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");
		
		if (_objects == null)
		    return -1;

        if (!_sorted
                || !isDynamic() // If list is not dynamic, we can't be sure that the order is still correct
                || object.isDeleted()) // LyContext discards set events when objects are deleted
    		// Slow search using indexOf method
    		return indexOf(object);
		
		if ((_sort != null)
		          && _sort.hasCalculatedFields(object.getClassInfo()))
	        // In that case, we can't be sure that the value is correctly calculated
            return indexOf(object);
		
		try
		{
    	    // Find the object position
            int      i, iMin = 0, iMax = getSize() - 1;
            LyObject cursor;
    
            while (iMax >= iMin)
            {
                i = (iMin + iMax) / 2;
                cursor = _objects.getObject(i);
    
                if (object.equals(cursor))
                    return i;
    
                // Compare objects with the current sort
                if (object.compare(cursor, _sort, _sortByName) >= 0)
                    iMin = i + 1;
                else
                    iMax = i - 1;
            }
		}
		catch (IllegalArgumentException e)
		{
		    // Happens when this list has a sort and given object doesn't belong to the same classes as list objects
		}
        
        return -1;
	}

	/**
	 * Adds objects of the specified list to this list.
	 *
	 * @param objectList Objects to add to this list.
	 */
	public void addList(LyObjectList objectList)
	{
		// Use internal method to add the given list to this list
		add(objectList);
	}

	/**
	 * Internal method to add an objects list to this list.
	 *
	 * @param  objectList Objects to add to this list.
	 * @return IllegalArgumentException If the given list is null.
	 */
	protected void add(LyObjectList objectList)
	{
		if (objectList == null)
			throw new IllegalArgumentException("BF AppEngine : objectList can't be null");

		// Even if the list is empty, create an empty list
		int n = objectList.getSize();

		if (_objects == null)
			_objects = new _LyObjects(n);

		if (n == 0)
			return;

		synchronized (objectList)
		{
			LyObjectList list = new _LyObjectList(objectList);

			if ((n > 1)
					&& _sorted
					&& (!list._sorted
						|| (list._sort != _sort)
						|| (list._sortByName != _sortByName)))
			{
				list._sorted = true;
				list._sort = _sort;
				list._sortByName = _sortByName;
				list.resort();
			}

			LyObject object;
			int		 i, size = getSize();

			if ((size == 0)
					|| (!_sorted && !_checkExistence))
			{
				_objects.addAll(list._objects);
				for (i = 0; i < n; i++)
				{
					object = list.getObject(i);
					objectAdded(object);
				}
				return;
			}
			else if (!_sorted)
			{
				for (i = 0; i < n; i++)
				{
					object = list.getObject(i);
					add(object);
				}
				return;
			}

			// Merge lists
			int		 j;
			LyObject cursor = _objects.getObject(0);

			for (i = 0, j = 1; i < n; i++)
			{
				object = list.getObject(i);

				while ((j < size)
						&& (object.compare(cursor, _sort, _sortByName) > 0))
				{
					cursor = _objects.getObject(j);
					j++;
				}
				if (_checkExistence && (object == cursor))
					continue;

				if ((j < size)
						|| (object.compare(cursor, _sort, _sortByName) < 0)) // At the end we have cursor (last object from orig. list) and object (from added list)
				{
					_objects.insertElementAt(object, j - 1);
					objectAdded(object);
					j++;
					size++;
				}
				else
				{
					while (i < n)
					{
						object = list._objects.getObject(i);
						_objects.add(object);
						objectAdded(object);
						i++;
					}
				}
			}
		}
	} // end method add

	/**
	 * Gets intersection of this list with the given list.
	 *
	 * @param  list List intersected with this list.
	 * @return IllegalArgumentException If the given list is null.
	 */
	public LyObjectList getIntersection(LyObjectList list)
	{
		if (_objects == null)
			return null;
		if ((list == null) || (list.getSize() == 0))
			return null;

		synchronized (this)
		{
			LySimpleObjectList result = null;
			LyObject		   object;
			int				   i, n = getSize();

			for (i = 0; i < n; i++)
			{
				object = _objects.getObject(i);
				if (!list.contains(object))
					continue;
				if (result == null)
				{
					result = new LySimpleObjectList(false);
					result.setSorted(_sorted);
				}
				result.add(object);
			}

			return result;
		}
	} // end method getIntersection

	/**
	 * Checks if the given list intersects with this list, this means at least one object is common
	 * to the two lists.
	 *
	 * @param  list List tested for intersection with this list.
	 * @return True if at least one object of the given list is present in this list, false
	 *         otherwise.
	 */
	public boolean intersects(LyObjectList list)
	{
		if (_objects == null)
			return false;
		if ((list == null) || (list.getSize() == 0))
			return false;

		synchronized (this)
		{
			LyObject object;
			int		 i, n = getSize();

			for (i = 0; i < n; i++)
			{
				object = _objects.getObject(i);
				if (list.contains(object))
					return true;
			}
		}

		return false;
	}

	/* Cross references */

	/**
	 * Calculate cross references. Gets in the given session, an objects list containing objects of
	 * the specified class that are related with objects of this list.
	 *
	 * @param     session     Session in which this search is made. This parameter may be null if
	 *                        the request is not dynamic.
	 * @param     targetClass Target class of objects to retrieve.
	 * @param     dynamic     Specifies if the result list has to be dynamic or not.
	 * @return    An objects list containing objects of the specified class that are related with
	 *            objects of this list in the given session.
	 * @exception IllegalArgumentException If target class is null.
	 */
	public LyObjectList getList(LyDataSession session, LyClassInfo targetClass, boolean dynamic)
	{
		return getList(session, targetClass, dynamic, false);
	}

	/**
	 * Calculate cross references. Gets in the given session, an objects list containing objects of
	 * the specified class that are related with objects of this list.
	 *
	 * @param     session     Session in which this search is made. This parameter may be null if
	 *                        the request is not dynamic.
	 * @param     targetClass Target class of objects to retrieve.
	 * @param     dynamic     Specifies if the result list has to be dynamic or not.
	 * @param     local       Specifies if the result list has to be local or not.
	 * @return    An objects list containing objects of the specified class that are related with
	 *            objects of this list in the given session.
	 * @exception IllegalArgumentException If target class is null.
	 */
	public LyObjectList getList(LyDataSession session, LyClassInfo targetClass, boolean dynamic,
		boolean local)
	{
		if (targetClass == null)
			throw new IllegalArgumentException("BF AppEngine : targetClass can't be null");

		// Differentiate  cross reference from null list and EMPTY list (size = 0)
		if (!dynamic && ((_objects == null) || (_objects.size() == 0)))
			return new LySimpleObjectList();

		LyInfoList<LyClassInfo> classes = getClasses();

		if (classes == null)
			return null;

		int n = classes.getSize();

		if (n == 0)
			return null;

		// For each classes of this list
		LyClassInfo		 classInfo;
		LyObjectList     result = null, list;
		LyMultiWorkSpace mws = null;
		int				 i;

		for (i = 0; i < n; i++)
		{
			classInfo = classes.getInfo(i);
			list = getList(session, classInfo, targetClass, dynamic, local);
			if (list == null)
				continue;

			if (result == null)
				result = list;
			else if (mws == null)
			{
				mws = new LyMultiWorkSpace(session, dynamic);
				mws.setSorted(_sorted);
				mws.addList(result);
				mws.addList(list);
				result = mws;
			}
			else
				result.addList(list);
		}

		return result;
	} // end method getList

	/**
	 * Calculate cross references. Gets in the given session, an objects list containing objects of
	 * the specified class that are related with objects of this list and that pass the given
	 * filter.
	 *
	 * @param     session     Session in which this search is made. This parameter may be null if
	 *                        the request is not dynamic.
	 * @param     targetClass Target class of objects to retrieve (may be null).
	 * @param     filter      a filter for the result list (may be null).
	 * @return    An objects list containing objects of the specified class that are related with
	 *            objects of this list in the given session and that pass the given filter.
	 * @exception IllegalArgumentException If targetClass and filter are both null.
	 */
	public final LyObjectList getList(LyDataSession session, LyClassInfo targetClass, LyFilter filter)
	{
		return getList(session, targetClass, filter, false, false);
	}

	/**
	 * Calculate cross references. Gets in the given session, an objects list containing objects of
	 * the specified class that are related with objects of this list and that pass the given
	 * filter.
	 *
	 * @param     session     Session in which this search is made. This parameter may be null if
	 *                        the request is not dynamic.
	 * @param     targetClass Target class of objects to retrieve (may be null).
	 * @param     filter      a filter for the result list (may be null).
	 * @param     local       Specifies if the result list must be local or not.
	 * @return    An objects list containing objects of the specified class that are related with
	 *            objects of this list in the given session and that pass the given filter.
	 * @exception IllegalArgumentException If targetClass and filter are both null.
	 */
	public final LyObjectList getList(LyDataSession session, LyClassInfo targetClass, LyFilter filter,
		boolean local)
	{
		return getList(session, targetClass, filter, false, local);
	}
	
	/**
	 * Calculate cross references. Gets in the given session, an objects list containing objects of
	 * the specified class that are related with objects of this list and that pass the given
	 * filter.
	 *
	 * @param     session     Session in which this search is made. This parameter may be null if
	 *                        the request is not dynamic.
	 * @param     targetClass Target class of objects to retrieve (may be null).
	 * @param     filter      a filter for the result list (may be null).
	 * @param     local       Specifies if the result list must be local or not.
	 * @return    An objects list containing objects of the specified class that are related with
	 *            objects of this list in the given session and that pass the given filter.
	 * @exception IllegalArgumentException If targetClass and filter are both null.
	 */
	public LyObjectList getList(LyDataSession session, LyClassInfo targetClass, LyFilter filter,
		boolean dynamic, boolean local)
	{
		if ((targetClass == null) && (filter == null))
			throw new IllegalStateException(
				"BF AppEngine : parameters targetClass and filter can't be both null");

		LyObjectList result = null;

		if (targetClass != null)
		{
			result = getList(session, targetClass, dynamic, local);
			if ((result == null) || (result.getSize() == 0) || (filter == null))
				return result;
		}
		else
			result = this;

		if (filter == null)
			return result;
		
		return new LySubWorkSpace(session, result, filter, dynamic);
	} // end method getList

	/**
	 * Calculate cross references. Gets in the given session, an objects list containing objects of
	 * the specified class that are related with objects of this list.
	 *
	 * @param     session     Session in which this search is made.
	 * @param     targetClass Target class of objects to retrieve.
	 * @return    An objects list containing objects of the specified class that are related with
	 *            objects of this list in the given session.
	 * @exception IllegalArgumentException If target class is null.
	 */
	public final LyObjectList getList(LyDataSession session, LyClassInfo targetClass)
	{
		return getList(session, targetClass, false);
	}

	/**
	 * Gets an objects list containing objects of the target class that are related to objects of
	 * this list that belong to the source class.
	 *
	 * @param     session     Session in which this search is made.
	 * @param     sourceClass Source class.
	 * @param     targetClass Target class.
	 * @param     dynamic     Specifies if the result list has to be dynamic or not.
	 * @param     local       Specifies if the result list has to be local or not.
	 * @exception IllegalStateException If no route beetween the source class and the target class
	 *                                  was found.
	 */
	protected LyObjectList getList(LyDataSession session, LyClassInfo sourceClass,
		LyClassInfo targetClass, boolean dynamic, boolean local)
	{
		if (targetClass.isVirtual())
			targetClass = targetClass.getRealClass();

		if ((sourceClass == targetClass)
				|| sourceClass.inheritsFrom(targetClass)
				|| targetClass.inheritsFrom(sourceClass))
			// Return a sub list of this list
			return new LySubWorkSpace(session, this, targetClass, (LyFilter)null, dynamic, local);

		LyInfoList<LyRoute> routes = sourceClass.getRoutes(targetClass);

		if (routes == null)
		{
			// No route to get related objects
			LyApplication application = sourceClass.getApplication();
			LyEnvironment environment = application.getEnvironment();

			environment.logWarning("BF AppEngine : can't find route from class "
				+ sourceClass + " to class " + targetClass);
			return null;
		}

		return getList(session, sourceClass, routes, dynamic, local);
	}

	/**
	 * Gets an objects list containing objects accessed through the route with the given id from
	 * objects of this list that belong to the specified source class.
	 *
	 * @param     session     Session in which this search is made.
	 * @param     sourceClass Source class for objects of this list. This may restrict objects of
	 *                        this list that are used to calculate the result list.
	 * @param     routeId     Identifier of the route used to reach taraget objects.
	 * @param     dynamic     Specifies if the result list has to be dynamic or not.
	 * @exception IllegalArgumentException If no route with the given id is found on the source
	 *                                     class.
	 */
	public LyObjectList getList(LyDataSession session, LyClassInfo sourceClass,
		String routeId, boolean dynamic)
	{
		LyRoute route = sourceClass.getRoute(routeId);

		if (route == null)
			throw new IllegalArgumentException("BF AppEngine : \"" + routeId
				+ "\" is not a valid route for class " + sourceClass);

		return getList(session, sourceClass, route, dynamic);
	}

	/**
	 * Gets an objects list containing objects accessed through the given routes from objects of
	 * this list that belong to the specified source class.
	 *
	 * @param     session     Session in which this search is made.
	 * @param     sourceClass Source class.
	 * @param     routes      List of routes used to reach target objects.
	 * @param     dynamic     Specifies if the result list has to be dynamic or not.
	 * @exception IllegalArgumentException If the given list of routes is null or empty.
	 */
	public final LyObjectList getList(LyDataSession session, LyClassInfo sourceClass,
		LyInfoList<? extends LyRoute> routes, boolean dynamic)
	{
		return getList(session, sourceClass, routes, dynamic, false);
	}

	/**
	 * Gets an objects list containing objects accessed through the given routes from objects of
	 * this list that belong to the specified source class.
	 *
	 * @param     session     Session in which this search is made.
	 * @param     sourceClass Source class.
	 * @param     routes      List of routes used to reach target objects.
	 * @param     dynamic     Specifies if the result list has to be dynamic or not.
	 * @param     local       Specifies if the result list has to be local or not.
	 * @exception IllegalArgumentException If the given list of routes is null or empty.
	 */
	public LyObjectList getList(LyDataSession session, LyClassInfo sourceClass,
		LyInfoList<? extends LyRoute> routes, boolean dynamic, boolean local)
	{
		if ((routes == null) || (routes.getSize() == 0))
			throw new IllegalArgumentException(
				"BF AppEngine : can't get objects from an empty list of routes");

		int			 i, n = routes.getSize();
		LyRoute		 route;
		LyObjectList result = null, ws, mws = null;

		for (i = 0; i < n; i++)
		{
			route = routes.getInfo(i);
			ws = getList(session, sourceClass, route, dynamic, local);
			if (ws == null)
				continue;

			if (result == null)
				result = ws;
			else if (mws == null)
			{
				mws = new LyMultiWorkSpace(session, dynamic);
				mws.setSorted(_sorted);
				mws.addList(result);
				mws.addList(ws);
				result = mws;
			}
			else
				mws.addList(ws);
		}

		return result;
	} // end method getList

	/**
	 * Gets an objects list containing objects accessed through the given route from objects of this
	 * list that belong to the specified source class.
	 *
	 * @param session     Session in which this search is made.
	 * @param sourceClass Source class for objects of this list.
	 * @param route       Route used to reach target objects.
	 * @param dynamic     Specifies if the result list has to be dynamic or not.
	 */
	public final LyObjectList getList(LyDataSession session, LyClassInfo sourceClass,
		LyRoute route, boolean dynamic)
	{
		return getList(session, sourceClass, route, dynamic, false);
	}

	/**
	 * Gets an objects list containing objects accessed through the given route from objects of this
	 * list that belong to the specified source class.
	 *
	 * @param session     Session in which this search is made.
	 * @param sourceClass Source class for objects of this list.
	 * @param route       Route used to reach target objects.
	 * @param dynamic     Specifies if the result list has to be dynamic or not.
	 * @param local       Specifies if the result list has to be local or not.
	 */
	public final LyObjectList getList(LyDataSession session, LyClassInfo sourceClass,
		LyRoute route, boolean dynamic, boolean local)
	{
		return getList(session, sourceClass, route, dynamic, local, false);
	}
	
	/**
	 * Gets an objects list containing objects accessed through the given route from objects of this
	 * list that belong to the specified source class.
	 *
	 * @param session     Session in which this search is made.
	 * @param sourceClass Source class for objects of this list.
	 * @param route       Route used to reach target objects.
	 * @param dynamic     Specifies if the result list has to be dynamic or not.
	 * @param local       Specifies if the result list has to be local or not.
	 * @param recurse     Activates recursion when sourceClass and targetClass are identical.
	 */
	public LyObjectList getList(LyDataSession session, LyClassInfo sourceClass,
		LyRoute route, boolean dynamic, boolean local, boolean recurse)
	{
		LyInfo parent = route.getParent();

		if (parent instanceof LyProject)
			// We must check the route.
			route.internalCheck(sourceClass);
		if (route.isReverse())
			return getReverseList(session, sourceClass, route, dynamic, local);

		Vector<LyStep> steps = route.getSteps();

		if (steps == null)
		{
			LyClassInfo targetClass = route.getTarget();

			return getList(session, sourceClass, targetClass, dynamic, local);
		}

		int			 i, ns = steps.size();
		LyStep		 step;
		LyClassInfo  target = null;
		LyObjectList result = null, current = this;

		for (i = 0; i < ns; i++)
		{
			step = steps.elementAt(i);
			target = step.getTarget();

			result = new LyStepWorkSpace(session, current, sourceClass, step, dynamic, local, recurse);
			sourceClass = target;
			current = result;
		}

		if ((target != null) && (target != route.getTarget()))
			return result.getList(session, target, route.getTarget(), dynamic, local);

		return result;
	} // end method getList

	/**
	 * Gets an objects list containing objects that point to objects of this list that belong to the
	 * specified source class, through the given route.
	 *
	 * @param session     Session in which this search is made.
	 * @param sourceClass Source class for objects of this list.
	 * @param route       Route used to reach source objects.
	 * @param dynamic     Specifies if the result list has to be dynamic or not.
	 * @param local       Specifies if the result list has to be local or not. Warning ! Dynamic
	 *                    reverse lists are not supported in this version.
	 */
	protected LyObjectList getReverseList(LyDataSession session, LyClassInfo sourceClass,
		LyRoute route, boolean dynamic, boolean local)
	{
		if (dynamic)
		{
			LyApplication application = sourceClass.getApplication();
			LyEnvironment environment = application.getEnvironment();

			environment.logWarning("Dynamic reverse lists are not supported in this version");
		}

		// XXX Reverse object list can't really be dynamic,
		// it requieres a new type of workspace
		LyClassInfo		  targetClass = route.getTarget();
		LyClassWorkSpace  ws = new LyClassWorkSpace(session, targetClass,
				new LyInfoList<LyFieldInfo>(), null, null, false, _sorted, local);
		LyObject		  target, source;
		int				  i, n = ws.getSize();
		int				  j, ns;
		LyObjectList	  sources;
		LySimpleWorkSpace result = new LySimpleWorkSpace(session, dynamic);

		result.setSorted(_sorted);

		for (i = 0; i < n; i++)
		{
			target = ws.getObject(i);
			sources = target.getList(session, sourceClass, false, true, local);

			for (j = 0, ns = sources.getSize(); j < ns; j++)
			{
				source = sources.getObject(j);
				if (contains(source))
				{
					result.addObject(target);
					break;
				}
			}
		}

		return result;
	} // end method getReverseList

	/**
	 * Gets list of objects of the given class of this list.
	 *
	 * @param classInfo Class info of the searched objects.
	 */
	protected Vector<LyObject> getObjects(LyClassInfo classInfo)
	{
		if (_objects == null)
			return null;

		Vector<LyObject> objects = new Vector<LyObject>();

		synchronized (this)
		{
			LyObject object;
			int		 i, n = getSize();

			for (i = 0; i < n; i++)
			{
				object = getObject(i);
				if (object.instanceOf(classInfo))
					objects.addElement(object);
			}
		}

		return objects;
	}

	/**
	 * Remove from the list inherited classes that have parent class in the list.
	 */
	protected LyInfoList<LyClassInfo> getParentClasses(LyInfoList<LyClassInfo> classes)
	{
		if (classes == null)
			return null;

		int		    i, j, n = classes.getSize();
		LyClassInfo classInfo, classInfo2;

		for (i = n - 1; i >= 0; i--)
		{
			classInfo = classes.getInfo(i);
			for (j = 0; j < (i + 1); j++)
			{
				classInfo2 = classes.getInfo(j);
				if (classInfo.inheritsFrom(classInfo2))
				{
					classes.removeInfo(classInfo);
					break;
				}
			}
		}

		return classes;
	}

	/**
	 * Methods for setting all objects of this list. Calls setObjects on each classes of the list.
	 */
	public void set(LyValueSet values)
	{
		if (_objects == null)
			return;

		LyInfoList<LyClassInfo> classes = getClasses();
		int					    nc = ((classes == null) ? 0 : classes.getSize());

		if (nc == 0)
			return;

		LyClassInfo		 classInfo;
		LyObjectSource   objectSource;
		Vector<LyObject> objects;
		int				 i, j, no;
		LyObject		 object;

		for (i = 0; i < nc; i++)
		{
			classInfo = classes.getInfo(i);
			objectSource = classInfo.getObjectSource();

			if (nc == 1)
				objects = getObjects();
			else
				objects = getObjects(classInfo);

			if (objects != null)
			{
				for (j = 0, no = objects.size(); j < no; j++)
				{
					object = objects.elementAt(j);
					object.setValues(values);
				}

				objectSource.setObjects(objects);
			}
		}
	} // end method set

	/**
	 * Methods for deleting all objects of this list. Calls deleteObjects on each classes of the
	 * list.
	 */
	public void delete()
	{
		delete(false, true);
	}

	/**
	 * Methods for deleting all objects of this list. Calls deleteObjects on each classes of the
	 * list.
	 *
	 * @param local   Indicates if deletions must be local or not
	 * @param refresh Indicates if the deletions may be notified or not.
	 */
	public void delete(boolean local, boolean refresh)
	{
		if (_objects == null)
			return;

		LyInfoList<LyClassInfo> classes = getClasses();

		classes = getParentClasses(classes);

		int nc = ((classes == null) ? 0 : classes.getSize());

		if (nc == 0)
			return;

		LyClassInfo		 classInfo;
		LyObjectSource   objectSource;
		Vector<LyObject> objects;
		int				 i;

		for (i = 0; i < nc; i++)
		{
			classInfo = classes.getInfo(i);
			objectSource = classInfo.getObjectSource();

			if (nc == 1)
				objects = getObjects();
			else
				objects = getObjects(classInfo);

			if (objects != null)
				objectSource.deleteObjects(objects, null, local, refresh);
		}
	} // end method delete

	/**
	 * Compare the content of the object lists Object lists are equals if they have exactlky the
	 * same LyObjects Sorting is *NOT* taken into account.
	 *
	 * @return true if the set of contained objects is identical.
	 */
	public boolean equals(Object targetList)
	{
		if (!(targetList instanceof LyObjectList))
			return false;
		return equals((LyObjectList)targetList);
	}
	
	/**
	 * Compare the content of the object lists Object lists are equals if they have exactlky the
	 * same LyObjects Sorting is *NOT* taken into account.
	 *
	 * @return true if the set of contained objects is identical.
	 */
	public boolean equals(LyObjectList targetList)
	{   
		if (targetList == null)
			return false;
		
        if (targetList == this)
            return true;

		if (_objects == null)
			return (targetList.isEmpty());

		// Go through the list of objects.
		synchronized (this)
		{
			int i, n = targetList.getSize();

			if (n != getSize())
				// Number of objects differs !
				return false;

			LyObject object, object2;

			for (i = 0; i < n; i++)
			{
				object = _objects.getObject(i);
                object2 = targetList._objects.getObject(i);

				if ((object != object2)
                        && !targetList.contains(object))
					return false; // At least on object differs
			}

			return true;
		}
	} // end method equals

	/**
	 * Returns a string representation of this object.
	 *
	 * @return This objects list returned as a string.
	 */
	public String toString()
	{
		if (_objects == null)
			return getClass() + "@" + hashCode() + " = {}";

		StringBuffer buffer = new StringBuffer();

		synchronized (this)
		{
			LyObject object;
			int		 i, n = getSize();

			buffer.append(getClass());
			buffer.append(' ');
			buffer.append(hashCode());
			buffer.append(" = {");
			buffer.append(n);
			buffer.append(": ");

			for (i = 0; (i < n) && (i < 5); i++) // 5 is enough
			{
				object = _objects.getObject(i);
				if (i > 0)
					buffer.append(", ");
				buffer.append(object);
			}
			if (n >= 5)
				buffer.append(", ...");
			buffer.append('}');
		}

		return buffer.toString();
	} // end method toString

	/**
	 * Gets a new list corresponding to this list sorted by the given label.
	 *
	 * @param  label Label used to sort this list.
	 * @return The new list sorted by the given label.
	 */
	public LyObjectList getSortedList(LyInfoLabel label)
	{
		if (label == null)
			return this;

		LySimpleObjectList list = new LySimpleObjectList();

		list.setSorted(false);

		if (getSize() == 0)
			return list;

		Vector<LyObject> objects = getObjects();
		int				 i, j, n = objects.size();
		LyObject		 obj1, obj2;
		String			 label1, label2;

		for (i = 0; i < n; i++)
		{
			obj1 = objects.elementAt(i);
			label1 = label.getString(obj1);

			for (j = i + 1; j < n; j++)
			{
				obj2 = objects.elementAt(j);
				label2 = label.getString(obj2);
				if (label1.compareTo(label2) > 0)
				{
					objects.setElementAt(obj1, j);
					objects.setElementAt(obj2, i);
					obj1 = obj2;
					label1 = label2;
				}
			}

			list.addObject(obj1);
		}

		return list;
	} // end method getSortedList

	/**
	 * Returns an implementation of the Iterator interface for this LyObjectList.
	 */
	public Iterator<LyObject> iterator()
	{
		return new _LyObjectListIterator(this);
	}

	/**
	 * Internal class used to speed up accesses to objects (especially sorting).
	 */
    final class _LyObjects extends Vector<LyObject>
	{
		/**
		 * 
		 */
		private static final long serialVersionUID = -1995491634294800577L;

		/**
		 * Constructs an empty vector.
		 */
		_LyObjects()
		{
		}

		/**
		 * Constructs a vector ofthe given size.
		 *
		 * @param n Size of the requested vector.
		 */
		_LyObjects(int n)
		{
			super(n);
		}

		/**
		 * Direct access to object at the given position (no control is made).
		 *
		 * @param  i Index of requested object.
		 * @return Object found.
		 */
		LyObject getObject(int i)
		{
			return (LyObject)elementData[i];
		}

		/**
		 * Direct setting of object at the given position (no control is made).
		 *
		 * @param object Object set.
		 * @param i      Index of the added object.
		 */
		void setObjectAt(LyObject object, int i)
		{
			elementData[i] = object;
		}
        
        /**
         * Sort objects list.
         */
        protected void resort()
        {
            _objects.quickSort(0, size() - 1);
        }
        
        /**
         * Utility method that implements quick sorting. Sorts objets beetween the given indexes.
         *
         * @param lo0 Min index.
         * @param hi0 Max index.
         */
        final void quickSort(int lo0, int hi0)
        {
            if (hi0 < lo0)
                return;

            int      lo = lo0;
            int      hi = hi0;
            LyObject mid, low, high;

            // Arbitrarily establishing partition element as the midpoint of the array.
            mid = (LyObject)elementData[(lo0 + hi0) / 2];

            // loop through the array until indices cross
            while (lo <= hi)
            {
                // Find the first element that is greater than or equal
                // to the partition element starting from the left Index.
                low = (LyObject)elementData[lo];

                while ((lo < hi0) && (low.compare(mid, _sort, _sortByName) < 0))
                {
                    ++lo;
                    low = (LyObject)elementData[lo];
                }

                // Find an element that is smaller than or equal to
                // the partition element starting from the right Index.
                high = (LyObject)elementData[hi];

                while ((hi > lo0) && (high.compare(mid, _sort, _sortByName) > 0))
                {
                    --hi;
                    high = (LyObject)elementData[hi];
                }

                // if the indexes have not crossed, swap
                if (lo <= hi)
                {
                    if (lo < hi)
                    {
                        elementData[lo] = high;
                        elementData[hi] = low;
                        swapObjects(lo, hi);
                    }
                    ++lo;
                    --hi;
                }
            }

            // If the right index has not reached the left side of array
            // must now sort the left partition.
            if (lo0 < hi)
                quickSort(lo0, hi);

            // If the left index has not reached the right side of array
            // must now sort the right partition.
            if (lo < hi0)
                quickSort(lo, hi0);
        } // end method quickSort
	} // end class _LyObjects


	/**
	 * Internal class used to speed up accesses adds of lists whithin list.
	 */
	final class _LyObjectList extends LyObjectList
	{
		/* insert_here:private LyObjectList outer_LyObjectList; */
		
		_LyObjectList(LyObjectList list)
		{
			_objects = new _LyObjects(list.getSize());
			if (list._objects != null)
				_objects.addAll(list._objects);
			_sorted = list._sorted;
			_sort = list._sort;
			_sortByName = list._sortByName;
		}

		protected void finalize()
			throws Throwable
		{
			// Don't unlock objects since they were not locked
		}

		public LyInfoList<LyClassInfo> getClasses()
		{
			return null;
		}
	}


	/**
	 * Internal class used for iteration over the list.
	 */
	final class _LyObjectListIterator
		implements Iterator<LyObject>
	{
		int			 _currentIndex;
		LyObjectList _list;
		boolean		 _removed;

		_LyObjectListIterator(LyObjectList list)
		{
			_currentIndex = 0;
			_list = list;
			_removed = false;
		}

		public boolean hasNext()
		{
			if (_currentIndex >= _list.getSize())
				return false;
			return true;
		}

		public void remove()
		{
			if ((!_removed) && (_currentIndex > 0))
			{
				LyObject object = _list.getObject(_currentIndex - 1);

				_list.remove(object);
				_currentIndex--;
				_removed = true;
			}
			else if (_removed)
				throw new IllegalStateException(
					"Current object has been already removed from this list");
			else if (_currentIndex <= 0)
				throw new IllegalStateException("There is no current object for this list");
		}

		public LyObject next()
		{
			try
			{
				LyObject object = _list.getObject(_currentIndex);

				_currentIndex++;
				_removed = false;
				return object;
			}
			catch (IndexOutOfBoundsException e)
			{
				throw new NoSuchElementException();
			}
		}
	} // end class _LyObjectListIterator
} // end class LyObjectList
