/*
 * LyObjectSource.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;

import leon.data.cache.*;
import leon.info.*;
import leon.info.binding.*;
import leon.info.location.*;
import leon.misc.*;
import leon.notifier.*;
import leon.peer.*;
import leon.peer.local.*;


/**
 * An object source is associated to each logical class (LyClassInfo). The object source is the
 * entry point for loading, creating, modifying and deleting logical objects (LyObject).
 *
 * <p>Objects sources are associated to data sources to dispatch demands on the underlying physical
 * layer when requests are made for logical objects.</p>
 *
 * @see LyObject
 * @see LyClassInfo
 */
public class LyObjectSource extends LyDataSource
{
	public static final String __VERSION =
		"$Id: LyObjectSource.java 41712 2012-12-05 08:51:26Z bangiole $";

	/** The dataSource for the physical representation of the objects in this objectSource. */
	LyDataSource			   _dataSource;

	/** Listener of the data source (adapter that implements LyDataSourceListener). */
	LyDataSourceListener	   _listener;

	/**
	 * Listener of the data sources of class infos pointed by relations of the class info managed by
	 * this object source (adapter that implements LyDataSourceListener).
	 */
	LyDataSourceListener	   _relationListener;

	/**
	 * Listener of the data sources of class infos pointed by structs of the class info managed by
	 * this object source (adapter that implements LyDataSourceListener).
	 */
	LyDataSourceListener	   _structListener;

	/** List of loaded fieldinfos for this source. */
	LyInfoList<LyFieldInfo>    _loadedFieldInfos;

	/** List of classInfos to listen to when loaded. */
	LyInfoList<LyClassInfo>    _waitForLoading;

	/** List of objects locally created. */
	ArrayList<LyObject>		   _locallyCreatedObjects;

	/** Local cache to speed up multiple getSize invocation (creating a form). */
	protected Object[]		   _getsizeCache;

	/** Local cache to speed up multiple getSize invocation with local request (creating a form). */
	protected Object[]		   _getsizelocalCache;

	/** Param max duration of _getSizeCache in ms. */
	public static long		   GET_SIZE_CACHE_DELAY = 250L;

	private boolean			_useGroupEvent;

	/**
	 * Default constructor of an object source. Only used by inherited classes.
	 */
	protected LyObjectSource()
	{
	}

	/**
	 * Constructs an object source for a given class info and a provider.
	 *
	 * @param provider  The provider of this data source.
	 * @param classInfo Class info for which an object source is constructed.
	 */
	public LyObjectSource(LyDataProvider provider, LyClassInfo classInfo)
	{
		super(provider, classInfo);

		_dataSource = null;
		_listener = null;
		_relationListener = null;
		_structListener = null;
		if ((!classInfo.isAbstract())
				&& (!classInfo.isLocal()))
			_locallyCreatedObjects = new ArrayList<LyObject>();

		resetLoadedFieldInfos();
		LyEnvironment environment = _application.getEnvironment();
		_useGroupEvent = environment.testResource("LY_GROUP_EVENTS_ON_LOADING");
	}

	/**
	 * Retrieves the class info for which this object source was constructed.
	 *
	 * @return Class info of this object source.
	 */
	public LyClassInfo getClassInfo()
	{
		return (LyClassInfo)super.getDataClass();
	}

	/**
	 * Associates this object source to a data source representing the physical layer.
	 *
	 * @param     dataSource Data source to which this object source is associated.
	 * @exception IllegalArgumentException If the given data source is null.
	 * @exception IllegalStateException    If this object source is already associated to a data
	 *                                     source.
	 */
	public void listenTo(LyDataSource dataSource)
	{
		if (_dataSource != null)
			throw new IllegalStateException(
				"BF AppEngine : the object source is already listening to a dataSource");

		if (dataSource == null)
			throw new IllegalArgumentException("BF AppEngine : dataSource can't be null");

		_dataSource = dataSource;
		if (_listener == null)
			_listener = new _LyDataSourceListener();
		_dataSource.addListener(_listener);

		// Requests cache policy of the data source
		//_dataSource.requestCachePolicy(_cache.getClass());
	}

	/**
	 * Sets the cache for this object source.
	 *
	 * @param     cachePolicy Informations for the cache to associate to this object source.
	 * @exception IllegalArgumentException If the given cache policy is null.
	 * @exception IllegalStateException    If the object source is already associated to a cache
	 *                                     that contains objects.
	 */
	public void setCache(LyInfoCache cachePolicy)
	{
		if ((_cache != null) && (!_cache.isEmpty()))
			return;

		if (cachePolicy == null)
			throw new IllegalArgumentException("BF AppEngine : cachePolicy parameter can't be null");

		Class<?> cacheClass = cachePolicy.getCachePolicy();

		if (cacheClass == LyFullCache.class)
			_cache = new LyFullCache(_class, cachePolicy);
		else if (cacheClass == LyNoCache.class)
			_cache = new LyNoCache(_class, cachePolicy);
		else
			_cache = new LySimpleCache(_class, cachePolicy);

		if (_cache != null)
		{
			if (_provider.canBeFullyLoaded(_class))
				_cache.allowLoaded(true);
			else
				_cache.allowLoaded(false);
		}
		// Requests cache policy of the data source
		if (_dataSource != null)
			_dataSource.requestCachePolicy(cachePolicy);
		_getsizeCache = null;
		_getsizelocalCache = null;
	}

	/**
	 * Free this object source. Remove this object source as listener of the associated data source.
	 */
	public void free()
	{
		LyEventManager em = _application.getEventManager();

		em.removeListeners(this);

		if (_listener != null)
		{
			_dataSource.removeListener(_listener);
			_listener = null;
		}
		_relationListener = null;
		_structListener = null;
		_dataSource = null;
		_getsizeCache = null;
		_getsizelocalCache = null;
	}

	/**
	 * Loads completely this object source : loads the associated cache according to the cache
	 * policy.
	 */
	public void load()
	{
		_cache.load();
		_getsizeCache = null;
		_getsizelocalCache = null;
	}

	/**
	 * Unloads completely this object source : unloads the associated cache and data source. Empties
	 * associated cache.
	 *
	 * @param notify indicates wether the unload action should notify objects deletion. Notifying
	 *               deletions makes sense when views are opened and when application is not
	 *               exiting.
	 */
	public void unload(boolean notify)
	{
		super.unload(notify);

		if (_dataSource != null)
			_dataSource.unload(notify);
		_getsizeCache = null;
		_getsizelocalCache = null;
//_locallyCreatedObjects=null;
		resetLoadedFieldInfos();
	}

	/**
	 * Sets the list of loaded fieldinfos for this object source.
	 *
	 * @param fieldInfos The list of loaded fieldInfos.
	 */
	public void addLoadedFieldInfos(Vector<LyFieldInfo> fieldInfos)
	{
		if (fieldInfos == null)
		{
			LyClassInfo classInfo = getClassInfo();

			_loadedFieldInfos = classInfo.getFieldInfos();
		}
		else
		{
			if (_loadedFieldInfos == null)
				_loadedFieldInfos = new LyInfoList<LyFieldInfo>(false);
			else if (!_loadedFieldInfos.isWritable())
				_loadedFieldInfos = new LyInfoList<LyFieldInfo>(_loadedFieldInfos);

			_loadedFieldInfos.addList(fieldInfos);
		}
	}

	/**
	 * Reset the list of loaded fieldInfos.
	 */
	public void resetLoadedFieldInfos()
	{
		LyClassInfo classInfo = getClassInfo();

		if (classInfo.isLocal())
			_loadedFieldInfos = classInfo.getFieldInfos();
		else
			_loadedFieldInfos = classInfo.getFieldInfos(LyFieldInfo.MARK_LOCAL);
	}

	/**
	 * Given an object, return the real class info of this object. For a normal object, return the
	 * object classInfo if it belongs to this objectsource.
	 *
	 * @param  object the LyObject
	 * @return the correct classInfo for this object.
	 */
	public LyClassInfo getRealClassInfo(LyObject object)
	{
		if (object.getObjectSource() == this)
			return object.getClassInfo();
		return null;
	}

	/**
	 * To know if the specified list of fieldInfos are loaded or not.
	 *
	 * @param  fieldInfos The list of fieldInfos
	 * @return true if the fieldInfos are loaded else false.
	 */
	public boolean areFieldsLoaded(LyInfoList<LyFieldInfo> fieldInfos)
	{
		if ((_loadedFieldInfos == null)
				|| (fieldInfos == null))
			return false;

        int              i, n = fieldInfos.getSize();
        LyFieldInfo      fieldInfo;

        for (i = 0; i < n; i++)
        {
            fieldInfo = fieldInfos.getInfo(i);
            if (fieldInfo.hasMark(LyFieldInfo.MARK_LOCAL))
                continue;
            if (!_loadedFieldInfos.contains(fieldInfo))
                return false;
        }

		return true;
	}

	/**
	 * Gets an object from this object source by the specified identifier.
	 *
	 * @param  id Object identifier of the searched object.
	 * @return The requested object if it was found, or null if it could not be found. This object
	 *         is an instance of LyObject (or an inherited class).
	 */
	public Object getObject(LyObjectId id)
	{
		LyClassInfo    classInfo = getClassInfo();
		LyClassBinding classBinding = classInfo.getBinding();
		Object		   result = super.getObject(id, null, classBinding);

		return result;
	}

	/**
	 * Gets an object from this object source by the specified identifier.
	 *
	 * @param  id    Object identifier of the searched object.
	 * @param  local Indicates if the demand is local (gets object from memory) or not.
	 * @return The requested object if it was found, or null if it could not be found. This object
	 *         is an instance of LyObject (or an inherited class).
	 */
	public Object getObject(LyObjectId id, boolean local)
	{
		if (id == null)
			throw new IllegalArgumentException("BF AppEngine : id parameter can't be null");

		Object result = null;

		if (local)
		{
			// Check if first
			if (!id.check(_application))
				return null;

			try
			{
				// Try to get the object from the cache.
				if (_cache.checkObject(id))
					result = _cache.getObject(id);
			}
			catch (LyNoSuchObjectException e)
			{
				result = null;
			}

			return result;
		}
		else
			result = getObject(id);

		return result;
	} // end method getObject

	/**
	 * Return the number of objects within this object source matching the given filter.
	 *
	 * @param  filter Filter applied on objects for counting.
	 * @return Number of objects matching the given filter in this object source.
	 */
	public int getSize(LyFilter filter)
	{
		return getSize(filter, false);
	}

	/**
	 * Return the number of objects within this object source matching the given filter.
	 *
	 * @param  filter Filter applied on objects for counting.
	 * @param  local  Indicates if the request must be local or not.
	 * @return Number of objects matching the given filter in this object source.
	 */
	public int getSize(LyFilter filter, boolean local)
	{
		if (local)
		{
			int		    size = 0;
			Object[] getsizeCache = _getsizelocalCache; // copy to prevent concurrent access

			if (getsizeCache != null)
			{
				// same request ?
				if ((getsizeCache[0] == filter)
						&& (((System.currentTimeMillis() - ((Long)getsizeCache[1]).longValue())
								- GET_SIZE_CACHE_DELAY) < 0)
						&& (getsizeCache[3] == Thread.currentThread()))
					size = ((Integer)getsizeCache[2]).intValue();
			}
			if (size == -1) // not found in cache
			{
				Enumeration<Object> elements = _cache.getElements();

				size = filter.match(elements);

				_getsizelocalCache = new Object[]
					{
						filter,
						new Long(System.currentTimeMillis()),
						new Integer(size),
						Thread.currentThread()
					};
			}

			return size;
		}
		else
		{
			int		 size = -1;
			Object[] getsizeCache = _getsizeCache; // copy to prevent concurrent access

			if (getsizeCache != null)
			{
				// same request ?
				if ((getsizeCache[0] == filter)
						&& (((System.currentTimeMillis() - ((Long)getsizeCache[1]).longValue())
								- GET_SIZE_CACHE_DELAY) < 0)
						&& (getsizeCache[3] == Thread.currentThread()))
					size = ((Integer)getsizeCache[2]).intValue();
			}
			if (size == -1) // not found in cache
			{
				LyClassInfo    classInfo = getClassInfo();
				LyClassBinding classBinding = classInfo.getBinding();

				size = super.getSize(filter, classBinding);
				_getsizeCache = new Object[]
					{
						filter,
						new Long(System.currentTimeMillis()),
						new Integer(size),
						Thread.currentThread()
					};
			}
			return size;
		}
	} // end method getSize

	/* Methods for creating an object */

	/**
	 * Create a new object in this object source. This method commits local creation by creating the
	 * corresponding physical object.
	 *
	 * @param     object Object to create.
	 * @return    Response to the creation.
	 * @exception IllegalStateException If the object source is in read only mode.
	 */
	public final LyResponse createObject(LyObject object)
	{
		return createObject(object, null);
	}

	/**
	 * Create a new object in this object source. This method commits local creation by creating the
	 * corresponding physical object.
	 *
	 * @param     object   Object to create.
	 * @param     listener Listener of the response.
	 * @return    Response to the creation.
	 * @exception IllegalStateException If the object source is in read only mode.
	 */
	public LyResponse createObject(LyObject object, LyResponseListener listener)
	{
		return createObject(object, listener, false, true);
	}

	/**
	 * Create a new object in this object source. This method commits local creation by creating the
	 * corresponding physical object.
	 *
	 * @param     object   Object to create.
	 * @param     listener Listener of the response.
	 * @param     local    Indicates if the creation must be local or not.
	 * @param     refresh  Indicates if the creation may be notified or not.
	 * @return    Response to the creation.
	 * @exception IllegalStateException If the object source is in read only mode.
	 */
	public LyResponse createObject(LyObject object, LyResponseListener listener, boolean local,
		boolean refresh)
	{
		// If the data source is read only we can't create an object.
		if (isReadOnly())
			throw new IllegalStateException(
				"BF AppEngine : Attempt to create an object on a read only object source");

		LyClassInfo    classInfo = getClassInfo();
		LyClassBinding classBinding = classInfo.getBinding();
		LyResponse     response;

		if (local)
		{
			// We send the request to the local data provider.
			LyRequestEncoder encoder;
			LyDataProvider   localDataProvider;

			localDataProvider = _application.getLocalDataProvider();
			encoder = localDataProvider.getRequestEncoder(this);
			response = new LyResponse(LyResponse.Operation.CREATE, _class);
			// Set the local and refresh status of the response.
			response.setLocal(true);
			response.setRefresh(refresh);
			if (listener != null)
				response.setListener(listener);

			encoder.setResponse(response);
			encoder.setClassBinding(classBinding);
			encoder.createObject(object);
		}
		else
			response = super.createObject(object, listener, classBinding, refresh);

		_getsizeCache = null;
		return response;
	} // end method createObject

	/* Methods for setting objects */

	/**
	 * Set an object of this object source. This method commits local modification by setting the
	 * corresponding physical object.
	 *
	 * @param     object Object to modify.
	 * @return    Response to the modification.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If the given object is invalid.
	 */
	public final LyResponse setObject(LyObject object)
	{
		return setObject(object, null);
	}

	/**
	 * Set an object of this object source. This method commits local modification by setting the
	 * corresponding physical object.
	 *
	 * @param     object   Object to modify.
	 * @param     listener Listener of the response.
	 * @return    Response to the modification.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If the given object is invalid.
	 */
	public LyResponse setObject(LyObject object, LyResponseListener listener)
	{
		return setObject(object, listener, false, true);
	}

	/**
	 * Set an object of this object source. This method commits local modification by setting the
	 * corresponding physical object.
	 *
	 * @param     object   Object to modify.
	 * @param     listener Listener of the response.
	 * @param     local    Indicates if the modification is local or not.
	 * @param     refresh  Indicates if the modification may be notified or not.
	 * @return    Response to the modification.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If the given object is invalid.
	 */
	public LyResponse setObject(LyObject object, LyResponseListener listener,
		boolean local, boolean refresh)
	{
		return setObject(object, listener, local, refresh, true);
	}

	protected LyResponse setObject(LyObject object, LyResponseListener listener,
		boolean local, boolean refresh, boolean needCheck)
	{
		// If the data source is read only we can't create an object.
		if (isReadOnly())
			throw new IllegalStateException(
				"BF AppEngine : Attempt to set an object on a read only object source");

		LyClassInfo    classInfo = getClassInfo();
		LyClassBinding classBinding = classInfo.getBinding();
		LyResponse     response;

		if (local)
		{
			// We send the request to the local data provider.
			LyRequestEncoder encoder;
			LyDataProvider   localDataProvider;

			localDataProvider = _application.getLocalDataProvider();
			encoder = localDataProvider.getRequestEncoder(this);
			response = new LyResponse(LyResponse.Operation.SET, _class);
			// Set the local and refresh status of the response.
			response.setLocal(local);
			response.setRefresh(refresh);
			if (listener != null)
				response.setListener(listener);

			encoder.setResponse(response);
			encoder.setClassBinding(classBinding);
			encoder.setObject(object);
		}
		else
			response = super.setObject(object, listener, classBinding, refresh, needCheck);

		_getsizeCache = null;
		return response;
	} // end method setObject

	/**
	 * Set a list of objects of this object source. This method commits local modifications by
	 * setting the corresponding physical objects.
	 *
	 * @param     objects List of objects to modify.
	 * @return    Response to the modification.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If one of the given objects is invalid.
	 */
	public final LyResponse setObjects(Vector<? extends Object> objects)
	{
		return setObjects(objects, null);
	}

	/**
	 * Set a list of objects of this object source. This method commits local modifications by
	 * setting the corresponding physical objects.
	 *
	 * @param     objects  List of objects to modify.
	 * @param     listener Listener of the response.
	 * @return    Response to the modification.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If one of the given objects is invalid.
	 */
	public LyResponse setObjects(Vector<? extends Object> objects, LyResponseListener listener)
	{
		LyClassInfo    classInfo = getClassInfo();
		LyClassBinding classBinding = classInfo.getBinding();

		return super.setObjects(objects, listener, classBinding);
	}

	/* Methods for deleting objects */

	/**
	 * Delete an object of this object source. This method commits local deletion by deleting the
	 * corresponding physical object.
	 *
	 * @param     object Object to delete.
	 * @return    Response to the deletion.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If one of the given objects is invalid.
	 */
	public final LyResponse deleteObject(LyObject object)
	{
		return deleteObject(object, null);
	}

	/**
	 * Delete an object of this object source. This method commits local deletion by deleting the
	 * corresponding physical object.
	 *
	 * @param     object   Object to delete.
	 * @param     listener Listener of the response.
	 * @return    Response to the deletion.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If the given object is invalid.
	 */
	public LyResponse deleteObject(LyObject object, LyResponseListener listener)
	{
		return deleteObject(object, listener, false, true);
	}

	/**
	 * Delete an object of this object source. This method commits local deletion by deleting the
	 * corresponding physical object.
	 *
	 * @param     object   Object to delete.
	 * @param     listener Listener of the response.
	 * @param     local    Indicates if the deletion is local or not.
	 * @param     refresh  Indicates if the deletion may be notified or not.
	 * @return    Response to the deletion.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If the given object is invalid.
	 */
	public LyResponse deleteObject(LyObject object, LyResponseListener listener,
		boolean local, boolean refresh)
	{
		return deleteObject(object, listener, local, refresh, true);
	}

	/**
	 * Delete an object of this object source. This method commits local deletion by deleting the
	 * corresponding physical object.
	 *
	 * @param     object            Object to delete.
	 * @param     listener          Listener of the response.
	 * @param     local             Indicates if the deletion is local or not.
	 * @param     refresh           Indicates if the deletion may be notified or not.
	 * @param     deleteLinkObjects Indicates if linked objects (composition relations, ...)
	 *                              must be deleted or not.
	 * @return    Response to the deletion.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If the given object is invalid.
	 */
	protected LyResponse deleteObject(LyObject object, LyResponseListener listener,
		boolean local, boolean refresh, boolean deleteLinkedObjects)
	{
		// If the data source is read only we can't delete an object.
		if (isReadOnly())
			throw new IllegalStateException(
				"BF AppEngine : Attempt to delete an object on a read only object source");

		LyClassInfo    classInfo = getClassInfo();
		LyClassBinding classBinding = classInfo.getBinding();
		LyResponse     response;

		if (local)
		{
			// We send the request to the local data provider.
			LyLocalRequestEncoder encoder;
			LyDataProvider		  localDataProvider;

			localDataProvider = _application.getLocalDataProvider();
			encoder = (LyLocalRequestEncoder)localDataProvider.getRequestEncoder(this);
			response = new LyResponse(LyResponse.Operation.DELETE, _class);
			// Set the local and refresh status of the response.
			response.setLocal(local);
			response.setRefresh(refresh);
			response.deleteLinkedObjects(deleteLinkedObjects);
			if (listener != null)
				response.setListener(listener);

			encoder.setResponse(response);
			encoder.setClassBinding(classBinding);
			encoder.deleteObject(object);
		}
		else
		{
			boolean check = !object.isLocallyDeleted();

			response = super.deleteObject(object, listener, classBinding, check, refresh);
		}
		_getsizeCache = null;

		return response;
	} // end method deleteObject

	/**
	 * Delete a list of objects of this object source. This method commits local deletions by
	 * deleting the corresponding physical objects.
	 *
	 * @param     objects List of objects to delete.
	 * @return    Response to the deletion.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If one of the given objects is invalid.
	 */
	public final LyResponse deleteObjects(Vector<? extends Object> objects)
	{
		return deleteObjects(objects, null);
	}

	/**
	 * Delete a list of objects of this object source. This method commits local deletions by
	 * deleting the corresponding physical objects.
	 *
	 * @param     objects  List of objects to delete.
	 * @param     listener Listener of the response.
	 * @return    Response to the deletion.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If one of the given objects is invalid.
	 */
	public LyResponse deleteObjects(Vector<? extends Object> objects, LyResponseListener listener)
	{
		return deleteObjects(objects, listener, false, true);
	}

	/**
	 * Delete a list of objects of this object source. This method commits local deletions by
	 * deleting the corresponding physical objects.
	 *
	 * @param     objects  List of objects to delete.
	 * @param     listener Listener of the response.
	 * @param     local    Indicates if the deletion is local or not.
	 * @param     refresh  Indicates if the deletion may be notified or not.
	 * @return    Response to the deletion.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If one of the given objects is invalid.
	 */
	public LyResponse deleteObjects(Vector<? extends Object> objects, LyResponseListener listener,
		boolean local, boolean refresh)
	{
		return deleteObjects(objects, listener, local, refresh, true);
	}

	/**
	 * Delete a list of objects of this object source. This method commits local deletions by
	 * deleting the corresponding physical objects.
	 *
	 * @param     objects           List of objects to delete.
	 * @param     listener          Listener of the response.
	 * @param     local             Indicates if the deletion is local or not.
	 * @param     refresh           Indicates if the deletion may be notified or not.
	 * @param     deleteLinkObjects Indicates if linked objects (composition relations, ...)
	 *                              must be deleted or not.
	 * @return    Response to the deletion.
	 * @exception IllegalStateException    If the object source is in read only mode.
	 * @exception IllegalArgumentException If one of the given objects is invalid.
	 */
	public LyResponse deleteObjects(Vector<? extends Object> objects, LyResponseListener listener,
		boolean local, boolean refresh, boolean deleteLinkedObjects)
	{
		// If the data source is read only we can't delete an object.
		if (isReadOnly())
			throw new IllegalStateException(
				"BF AppEngine : Attempt to delete objects on a read only object source");

		LyClassInfo    classInfo = getClassInfo();
		LyClassBinding classBinding = classInfo.getBinding();
		LyResponse     response;

		if (local)
		{
			// We send the request to the local data provider.
			LyLocalRequestEncoder encoder;
			LyDataProvider		  localDataProvider;

			localDataProvider = _application.getLocalDataProvider();
			encoder = (LyLocalRequestEncoder)localDataProvider.getRequestEncoder(this);
			response = new LyResponse(LyResponse.Operation.DELETE, _class);
			// Set the local and refresh status of the response.
			response.setLocal(local);
			response.setRefresh(refresh);
			response.deleteLinkedObjects(deleteLinkedObjects);
			if (listener != null)
				response.setListener(listener);

			encoder.setResponse(response);
			encoder.setClassBinding(classBinding);
			encoder.deleteObjects(new Vector<Object>(objects));
		}
		else
		{
			boolean  check = false;
			int		 i, n = objects.size();
			LyObject object;

			for (i = 0; (i < n) && !check; i++)
			{
				object = (LyObject)objects.elementAt(i);
				check = !object.isLocallyDeleted();
			}

			response = super.deleteObjects(objects, listener, classBinding, check, refresh);
		}

		_getsizeCache = null;
		return response;
	} // end method deleteObjects

	/* Methods for loading objects */

	/**
	 * Method to reload an object by its id from this object source.
	 *
	 * @param  id Object identifier corresponding to the object to reload.
	 * @return Response to the loading.
	 */
	public final LyResponse reloadObject(LyObjectId id)
	{
		return reloadObject(id, null, null);
	}

	/**
	 * Method to reload an object by its id from this object source.
	 *
	 * @param  id       Object identifier corresponding to the object to reload.
	 * @param  listener Listener of the response.
	 * @return Response to the loading.
	 */
	public LyResponse reloadObject(LyObjectId id, LyResponseListener listener)
	{
		return loadObject(id, null, listener);
	}

	/**
	 * Method to reload an object from its id on the associated data provider.
	 *
	 * @param  id       Object identifier corresponding to the object to reload.
	 * @param  fields   Fields to load (null means all fields).
	 * @param  listener Listener of the response.
	 * @return Response to the loading.
	 */
	public LyResponse reloadObject(LyObjectId id, Vector<LyFieldInfo> fields,
		LyResponseListener listener)
	{
		LyClassInfo    classInfo = getClassInfo();
		LyClassBinding classBinding = classInfo.getBinding();
		LyObject   	   object = id.getLogicalObject();

		if (!isPersistent(classBinding))
		{

			LyResponse response = new LyResponse(LyResponse.Operation.LOAD, classInfo);

			response.setStatus(LyResponse.Status.RUNNING);
			response.addResponse(object, null);
			response.setStatus(LyResponse.Status.COMPLETED);
			return response;
		}

		_getsizeCache = null;
		object.cleanDaemonValue(fields);
		object.touch();

		return reloadObject(id, fields, listener, classBinding);
	}

	/**
	 * To test if the sepcified class binding is persistent : a class binding is persistent if it is
	 * bind only with persitent providers.
	 *
	 * @param  classBinding the class binding to test.
	 * @return true if the specified class binding is persistent, false otherwise.
	 */
	protected boolean isPersistent(LyClassBinding classBinding)
	{
		LyEnvironment environment = _application.getEnvironment();

		if (environment.testEnv("LY_USE_FILES"))
			return false;

		if (classBinding instanceof LyUnionBinding)
		{
			LyClassBinding binding1, binding2;

			binding1 = ((LyUnionBinding)classBinding).getClassBinding1();
			binding2 = ((LyUnionBinding)classBinding).getClassBinding2();
			return (isPersistent(binding1) && isPersistent(binding2));
		}

		String locationId = null;

		if (classBinding != null)
			locationId = classBinding.getLocation();

		if (locationId == null)
			locationId = _application.getDefaultLocationId();
		LyInfo info = _application.getInfo(locationId);

		if (info instanceof LyLocation)
			return ((LyLocation)info).isPersistent();
		else
			return false;
	} // end method isPersistent

	/**
	 * Method to load an object by its id from this object source.
	 *
	 * @param  id Object identifier corresponding to the object to load.
	 * @return Response to the loading.
	 */
	public final LyResponse loadObject(LyObjectId id)
	{
		return loadObject(id, null, null);
	}

	/**
	 * Method to load an object by its id from this object source.
	 *
	 * @param  id       Object identifier corresponding to the object to load.
	 * @param  listener Listener of the response.
	 * @return Response to the loading.
	 */
	public final LyResponse loadObject(LyObjectId id, LyResponseListener listener)
	{
		return loadObject(id, null, listener);
	}

	/**
	 * Method to load an object from its id on the associated data provider.
	 *
	 * @param  id       Object identifier corresponding to the object to load.
	 * @param  fields   Fields to load (null means all fields).
	 * @param  listener Listener of the response.
	 * @return Response to the loading.
	 */
	public LyResponse loadObject(LyObjectId id, Vector<? extends LyField> fields,
		LyResponseListener listener)
	{
		LyClassInfo    classInfo = getClassInfo();
		LyClassBinding classBinding = classInfo.getBinding();

		if (!classInfo.isAbstract())
		{
			if (fields == null)
				fields = classInfo.getFields();

			// Only load the fields that are not already loaded.
			Vector<LyFieldInfo> fieldsToLoad = new Vector<LyFieldInfo>();
			LyFieldInfo		    fieldInfo;
			int				    i, nf = (fields == null ? 0 : fields.size());

			for (i = 0; i < nf; i++)
			{
				fieldInfo = (LyFieldInfo)fields.elementAt(i);
				if (!_loadedFieldInfos.contains(fieldInfo))
					fieldsToLoad.add(fieldInfo);
			}

			if (!fieldsToLoad.isEmpty())
				fields = fieldsToLoad;
		}

		return super.loadObject(id, fields, listener, classBinding);
	}

	/**
	 * Method for loading objects on this object source. Load a specified number of objects sorted
	 * by a given sort and that match a given filter.
	 *
	 * @param  fieldInfos The list of the fieldInfos to load if necessary. If this parameter is null
	 *                    all the field infos are loaded. If the list is empty, only fields with
	 *                    mark id or name and fields used by sort and filter are loaded.
	 * @param  sort       Gives sort criteria, or null if there is no sort (the order is the one of
	 *                    the physical layer).
	 * @param  filter     Gives filter criteria, or null if there is no filter (all objects).
	 * @param  n          Indicates the number of objects to load (-1 means no limit).
	 * @return Response to the loading.
	 */
	public final LyResponse loadObjects(Vector<LyFieldInfo> fieldInfos, LySort sort, LyFilter filter, int n)
	{
		return loadObjects(fieldInfos, sort, filter, n, null);
	}

	/**
	 * Method for loading objects on this object source. Load a specified number of objects sorted
	 * by a given sort and that match a given filter.
	 *
	 * @param  fieldInfos The list of the fieldInfos to load if necessary. If this parameter is null
	 *                    all the field infos are loaded. If the list is empty, only fields with
	 *                    mark id or name and fields used by sort and filter are loaded.
	 * @param  sort       Gives sort criteria, or null if there is no sort (the order is the one of
	 *                    the physical layer).
	 * @param  filter     Gives filter criteria, or null if there is no filter (all objects).
	 * @param  n          Indicates the number of objects to load (-1 means no limit).
	 * @param  listener   Listener of the response.
	 * @return Response to the loading.
	 */
	public LyResponse loadObjects(Vector<? extends LyField> fieldInfos, LySort sort,
		LyFilter filter, int n, LyResponseListener listener)
	{
		LyClassInfo		    classInfo = getClassInfo();

		if (!classInfo.isAbstract())
		{
			if (fieldInfos == null)
			{
				BitSet marks = new BitSet();

				marks.set(LyFieldInfo.MARK_ID);
				marks.set(LyFieldInfo.MARK_NAME);
				marks.set(LyFieldInfo.MARK_LOAD);
				fieldInfos = classInfo.getFieldInfos(marks).getInfos();
			}

			// Only load the fields that are not already loaded.
			Vector<LyFieldInfo> fieldsToLoad = new Vector<LyFieldInfo>();
			LyFieldInfo		    fieldInfo;
			int				    i, nf = (fieldInfos == null ? 0 : fieldInfos.size());

			for (i = 0; i < nf; i++)
			{
				fieldInfo = (LyFieldInfo)fieldInfos.elementAt(i);
				if (!_loadedFieldInfos.contains(fieldInfo))
					fieldsToLoad.add(fieldInfo);
			}

			if (!fieldsToLoad.isEmpty())
				fieldInfos = fieldsToLoad;
		}

		_getsizeCache = null;
		return super.loadObjects(fieldInfos, sort, filter, n, listener, classInfo.getBinding());
	}

	/* Methods for accessing objects */

	/**
	 * Retrieve objects on this object source. Retrieve a specified number of objects sorted by a
	 * given sort and that match a given filter. These objects are first searched in the associated
	 * cache to avoid requests to the physical layer.
	 *
	 * @param  fieldInfos The list of the fieldInfos to load if necessary. If this parameter is null
	 *                    all the field infos are loaded. If the list is empty, only fields with
	 *                    mark id or name and fields used by sort and filter are loaded.
	 * @param  sort       Gives sort criteria, or null if there is no sort (the order is the one of
	 *                    the physical layer).
	 * @param  filter     Gives filter criteria, or null if there is no filter (all objects).
	 * @param  n          Indicates the number of objects to retrieve (-1 means no limit).
	 * @param  local      Indicates if the request is local (from memory) or not.
	 * @param  listener   Listener of the response.
	 * @return Response to this search.
	 */
	public LyResponse getObjects(Vector<LyFieldInfo> fieldInfos, LySort sort, LyFilter filter,
		int n, boolean local, LyResponseListener listener)
	{
		LySimpleObjectList list = new LySimpleObjectList(true); // optim. : don't check existence
		LyCache.Result	   result;
		LyResponse		   response;
		LyObject		   object;
		int				   i, p;
		boolean			   fieldsAreLoaded;

		// Check cache : this method permits the cache to load or unload
		// some objects before searching some of them.
		if (!local)
		{
			_cache.check();

			if (fieldInfos == null)
			{
				LyClassInfo classInfo = getClassInfo();

				fieldsAreLoaded = areFieldsLoaded(classInfo.getFieldInfos());
			}
			else
				fieldsAreLoaded = areFieldsLoaded(new LyInfoList<LyFieldInfo>(fieldInfos));
		}
		else
			fieldsAreLoaded = true;

		if (!fieldsAreLoaded)
			result = LyCache.Result.NULL;
		else
		{
			list.setSort(sort);
			list.setSorted(false);

			if (!local)
				result = _cache.getObjects(filter, list);
			else
			{
				Enumeration<Object>     elements = _cache.getElements();
				LyInfoList<LyFieldInfo> fields = (filter == null ? null : filter.getFieldInfos());
				LyValueSet              values;

				while (elements.hasMoreElements())
				{
					object = (LyObject)elements.nextElement();
					if (filter == null)
						list.addObject(object);
					else if (local)
					{
						values = object.getValues(fields, true);
						if (filter.match(values))
							list.addObject(object);
					}
					else if (filter.match(object))
						list.addObject(object);
				}
				result = LyCache.Result.TOTAL;
			}
		}

		if (result == LyCache.Result.TOTAL)
		{
			response = new LyResponse(LyResponse.Operation.LOAD, _class);
			if (listener != null)
				response.setListener(listener);
			response.setStatus(LyResponse.Status.RUNNING);
			response.setFiltered(true);
			response.setCounted(true);
			response.setDistinctValues(true);
			if (n == -1)
			{
				// There is no limit so we let the asker do the job.
				response.setSorted(false);
				p = list.getSize();
			}
			else
			{
				// There is a limit : we must sort the list to return the good result.
				response.setSorted(true);
				list.setSorted(true);
				p = n;
				if (p > list.getSize())
					// The result may have less results than the limit.
					p = list.getSize();
			}
			for (i = 0; i < p; i++)
			{
				object = list.getObject(i);
				response.addResponse(object, null);
			}
			response.setStatus(LyResponse.Status.COMPLETED);
			response.addResponse(null, null);
		}
		else
		{
			// The response is partial or empty.
			if ((_locallyCreatedObjects != null)
					&& (_locallyCreatedObjects.size() != 0))
			{
				LyObjectList objects;

				response = loadObjects(fieldInfos, sort, filter, n, null);

				objects = response.getObjects();
				list.setSort(sort);
				list.setSorted(false);
				list.addList(objects);

				for (i = 0; i < _locallyCreatedObjects.size(); i++)
				{
					object = _locallyCreatedObjects.get(i);
					if ((filter == null) || (filter.match(object)))
						list.add(object);
				}

				response = new LyResponse(LyResponse.Operation.LOAD, _class);
				if (listener != null)
					response.setListener(listener);
				response.setStatus(LyResponse.Status.RUNNING);
				response.setFiltered(true);
				response.setCounted(true);
				if (n == -1)
				{
					// There is no limit so we let the asker do the job.
					response.setSorted(false);
					p = list.getSize();
				}
				else
				{
					// There is a limit : we must sort the list to return the good result.
					response.setSorted(true);
					list.setSorted(true);
					p = n;
					if (p > list.getSize())
						// The result may have less results than the limit.
						p = list.getSize();
				}
				for (i = 0; i < p; i++)
				{
					object = list.getObject(i);
					response.addResponse(object, null);
				}
				response.setStatus(LyResponse.Status.COMPLETED);
				response.addResponse(null, null);
			}
			else
				response = loadObjects(fieldInfos, sort, filter, n, listener);
		}

		_getsizeCache = null;

		return response;
	} // end method getObjects

	/**
	 * Retrieve objects on this object source. Retrieve a specified number of objects sorted by a
	 * given sort and that match a given filter. These objects are first searched in the associated
	 * cache to avoid requests to the physical layer.
	 *
	 * @param  sort   Gives sort criteria, or null if there is no sort (the order is the one of the
	 *                physical layer).
	 * @param  filter Gives filter criteria, or null if there is no filter (all objects).
	 * @param  n      Indicates the number of objects to retrieve (-1 means no limit).
	 * @return Response to this search.
	 */
	public final LyResponse getObjects(LySort sort, LyFilter filter, int n)
	{
		return getObjects(null, sort, filter, n, false, null);
	}

	/**
	 * Retrieve objects on this object source that match a given filter. These objects are first
	 * searched in the associated cache to avoid requests to the physical layer.
	 *
	 * @param  filter Gives filter criteria, or null if there is no filter (all objects).
	 * @return Response to this search.
	 */
	public final LyResponse getObjects(LyFilter filter)
	{
		return getObjects(null, null, filter, -1, false, null);
	}

	/**
	 * Retrieve objects on this object source sorted by a given sort. These objects are first
	 * searched in the associated cache to avoid requests to the physical layer.
	 *
	 * @param  sort Gives sort criteria, or null if there is no sort (the order is the one of the
	 *              physical layer).
	 * @return Response to this search.
	 */
	public final LyResponse getObjects(LySort sort)
	{
		return getObjects(null, sort, null, -1, false, null);
	}

	/**
	 * Recompute the object id of the specified object.
	 *
	 * @param object the object to recompute its id.
	 */
	public final void rehash(LyObject object)
	{
		rehash(object, true);
	}

	/**
	 * Recompute the object id of the specified object.
	 *
	 * @param object    the object to recompute its id.
	 * @param sendEvent Flag to indicate if an event may be sent or not.
	 */
	protected void rehash(LyObject object, boolean sendEvent)
	{
		LyObjectId oldObjectId = _class.getObjectId(object);
		LyObjectId newObjectId;

		object.resetObjectId();

		boolean isRehash = _cache.rehash(object, oldObjectId);

		newObjectId = object.getObjectId();

		if (!isRehash || (newObjectId.equals(oldObjectId)))
			// No rehash or no change so nothing to do.
			return;

		// Callback object to update daemons.
		object.idRehashed(oldObjectId, newObjectId);

		// Update id of new object on player for test mode
		LyEnvironment environment = _application.getEnvironment();
		LyPlayer	  player = environment.getPlayer();

		if (player != null)
		{
			String internalId = (String)player.getValue(oldObjectId.getFullString(), false);

			if (internalId == null)
				recordId(object);
			else
			{
				LyClassInfo classInfo = object.getClassInfo();

				player.addValue(object.getId(), internalId, false);
				if (!classInfo.isNameDifferentOfId())
				{
					internalId = internalId.substring(0, internalId.length() - 3);
					player.addValue(object.getName(), internalId + ".name", true);
				}
			}
		}

		// Rehash peer object
		Object peer = object.getPeer();

		if (peer != null)
		{
			LyClassInfo classInfo = object.getClassInfo();
			LyClass     peerClass = classInfo.getPeerClass();

			peerClass.rehash(peer);
		}

		if (!sendEvent)
			return;

		// Post an event to notify the change of id fields.
		LyClassInfo			    classInfo = object.getClassInfo();
		LyInfoList<LyFieldInfo> idFields = classInfo.getIdFieldInfos();
		String[]			    strings = oldObjectId.toStrings();

		if (strings.length != idFields.getSize())
			object.refresh(idFields);
		else
		{
		    try
		    {
    			LyValueSet  values = new LyValueSet();
    			int		    i, n = idFields.getSize();
    			LyFieldInfo idField;
    			LyValue     value;

    			for (i = 0; i < n; i++)
    			{
    				idField = idFields.getInfo(i);
    				value = idField.buildFieldValue(strings[i]);
    				values.addValue(value);
    			}

    			object.refresh(values);
		    }
		    catch (Exception e)
		    {
		        // May happen when constructing value for arrays or struct from string
		        object.refresh(idFields);
		    }
		}

		_getsizeCache = null;
	} // end method rehash

	/**
	 * Finds an object from its object peer (associated physical object). The object is searched in
	 * the cache and if it is not found, it is created.
	 *
	 * @param  peer Physical object.
	 * @return The logical object that was found or created, or null if the creation fails.
	 */
	public LyObject getObjectFromPeer(Object peer)
	{
		if (peer == null)
			// An object with a null peer is null.
			return null;

		if (!_dataSource.checkObject(peer))
			return null;

		LyObject object = _cache.getObjectFromPeer(peer);

		if (object != null)
			return object;

		// The object does not exist : we must create it.
		object = createObjectFromPeer(peer);
		if (object == null)
			return null; // Creation fails

		if (_cache.isRemoved(object.getObjectId()))
			// Object locally deleted, don't recreate it
			return null;

		// Putting the objet in the cache.
		_cache.putObject(object);

		// Notify all the listeners.
		LyDataSourceEvent event;

		event = new LyDataSourceEvent(this, LyDataSourceEvent.TYPE_CREATE, object, null);
		event.setLocal(true);
		postEvent(event);

		return object;
	} // end method getObjectFromPeer

	/**
	 * Create an object from a given peer object (associated physical object).
	 *
	 * @param  peer Peer object.
	 * @return Created object or null if the object doesn't match the class binding of the class
	 *         info.
	 */
	protected LyObject createObjectFromPeer(Object peer)
	{
		if (peer == null)
			// An object can't have a null peer.
			return null;

		LyObject    object = null;

		// Trying to create the object.
		LyClassInfo classInfo = getClassInfo();

		object = classInfo.createInstance();
		object.setPeer(peer);

		// We must verify if the new object matches the filter
		LyClassBinding classBinding = classInfo.getBinding();

		if (classBinding != null)
		{
			// XXX Is there a way to avoid this check when the result is already filtered ?
			LyFilter filter = classBinding.getFilter();

			if ((filter != null) && !filter.match(object))
				return null;
		}

		LyClassInfo realClass = classInfo.getObjectSource().getRealClassInfo(object);

		if (realClass == classInfo)
			return object;

		return null;
	} // end method createObjectFromPeer

	/**
	 * Get old values for the given object corresponding to the given set values changed on physical
	 * object.
	 *
	 * @param  object        Object for which old values are requested.
	 * @param  oldPeerValues Modifications made on physical object.
	 * @param  validate      Indicates if we must validate updates in memory.
	 * @return List of values for the given object corresponding to modifications made on the
	 *         physical layer.
	 */
	protected LyValueSet getOldValues(LyObject object, LyValueSet oldPeerValues, boolean validate)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		LyValueSet result = null;

		if (validate)
			result = object.validate();

		if (oldPeerValues == null)
			return result;

		if ((result == null) && (oldPeerValues.getSize() == 0))
			return new LyValueSet();

		LyClassInfo    classInfo = getClassInfo();
		LyClassBinding classBinding = classInfo.getBinding();
		int			   i, n = classInfo.getFieldInfoCount();
		LyFieldInfo    fieldInfo;
		LyFieldBinding binding;
		LyValue		   value;
		Object		   oldValue;
		LyField		   field;
		LyField[]	   fields;
		Object[]	   values;
		int			   j, nf;

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			binding = fieldInfo.getBinding();
			if (binding == null)
				continue;

			if (binding instanceof LyFieldsBinding)
			{
				fields = classBinding.getPeerFields(fieldInfo);
				values = null;

				for (j = 0, nf = fields.length; j < nf; j++)
				{
					value = oldPeerValues.getFieldValue(fields[j]);
					if (value != null)
					{
						if (values == null)
							values = new Object[nf];
						values[j] = value.getValue();
					}
				}

				if (values != null)
				{
					oldValue = fieldInfo.getValueFromObject(values, object);
					if (result == null)
						result = new LyValueSet();
					result.addValue(fieldInfo, oldValue);
				}
			}
			else
			{
				field = classBinding.getPeerField(fieldInfo);
				value = oldPeerValues.getFieldValue(field);
				if (value != null)
				{
					oldValue = fieldInfo.getValueFromObject(value.getValue(), object);
					if (result == null)
						result = new LyValueSet();
					result.addValue(fieldInfo, oldValue);
				}
			}
		}

		return result;
	} // end method getOldValues

	/**
	 * Internal method used to receive a new response from data provider to a creation request sent
	 * on createObject. Object is added in the cache of this data source and a creation event is
	 * sent to listeners. If the data provider has a notifier, the creation is notified.
	 *
	 * @param response Response to the create operation.
	 * @param status   Status of the new response. Valid status are defined in LyResponse.
	 * @param object   Newly created object.
	 * @see   #createObject
	 */
	protected void receiveCreate(LyResponse response, LyResponse.Status status, Object object)
	{
		LyObject lyObject = (LyObject)object;

		if (!response.isLocal())
		{
			lyObject.validate();

			// Add the object in the cache.
			// XXX How to avoid warnings ?
			// Maybe we could have some warnings when we put the object in the cache
			// twice because a local object can become no local with a second creation.
			// But we must do it to manage correctly the reverse relation between
			// the LyObject and its peer in the cache layer.
			_cache.putObject(object);

			// Recompute object id
            rehash(lyObject, response.isRefresh());

			if (_locallyCreatedObjects != null)
				_locallyCreatedObjects.remove(lyObject);
		}
		else
		{
			lyObject.moveNewToPendingValues();
			lyObject.resetObjectId();

			// Add the object in the cache.
			_cache.putObject(object);

			lyObject.setLocallyDeleted(false);
			lyObject.setDeleted(false);
			if (_locallyCreatedObjects != null)
				_locallyCreatedObjects.add(lyObject);
		}

		// Set the status of the response.
		response.setStatus(status);

		// Add the object in the response.
		response.addResponse(object, null);

		// Notify all the listeners.
		if (response.isRefresh())
		{
			Vector<LyDataSourceListener> skipListeners = response.getSkippedListeners();
			LyDataSourceEvent            event;

			event = new LyDataSourceEvent(this, LyDataSourceEvent.TYPE_CREATE, object, null);
			event.setLocal(response.isLocal());
			postEvent(event, skipListeners);
		}

		// Record id of new object on player for test mode
		recordId(lyObject);

		_getsizeCache = null;

	} // end method receiveCreate

	/**
	 * Method used to record object id for test mode.
	 */
	public void recordId(LyObject object)
	{
		// Record id of new object on player for test mode
		LyEnvironment environment = _application.getEnvironment();
		LyPlayer	  player = environment.getPlayer();
		String		  objectId = object.getId();
		LyClassInfo   classInfo = object.getClassInfo();

		if (player != null)
		{
			if (player.getValue(objectId, false) != null)
				return; // already recorded

			String internalId = LyPlayer.NEW_MARKER + _class.getId().toUpperCase();
			int    n = 1;

			while (player.getValue(internalId + "." + n + ".id", false) != null)
				n++;
			player.addValue(object.getId(), internalId + "." + n + ".id", false);
			if (!classInfo.isNameDifferentOfId())
				player.addValue(object.getName(), internalId + "." + n + ".name", true);
		}
	}

	/**
	 * Internal method used to receive a new response from data provider to a set request sent on
	 * setObject or setObjects. This method is overriden to filter objects according to class binding.
	 *
	 * @param response      Response to the set operation.
	 * @param status        Status of the new response. Valid status are defined in LyResponse.
	 * @param object        Modified object.
	 * @param oldPeerValues Old peer values of the object before the set operation.
	 * @see   #setObject
	 * @see   #setObjects
	 */
	protected void receiveSet(LyResponse response, LyResponse.Status status, Object object, LyValueSet oldPeerValues)
	{
		LyObject lyObject = (LyObject)object;

		// Set the status to the response.
		response.setStatus(status);

		// Test if the object matches the class filter.
		LyClassInfo    classInfo = getClassInfo();
		LyClassBinding classBinding = classInfo.getBinding();

		_getsizeCache = null;

		if ((classBinding != null)
		    && (lyObject.getPeer() != null))
		{
			LyFilter filter = classBinding.getFilter();

			if ((filter != null) && !filter.match(object))
			{
				// Remove object from the cache only if response is not local.
				// If response is local we must wait for the commit to avoid operations on deleted objects.
				if (!response.isLocal())
					_cache.startRemoveObject(object);

				try
				{
					// Notify all the listeners.
					if (response.isRefresh()
							|| !response.isLocal())
					{
						LyDataSourceEvent myEvent;

						myEvent = new LyDataSourceEvent(LyObjectSource.this, LyDataSourceEvent.TYPE_DELETE, object, null);
						myEvent.setLocal(response.isLocal());
						postEvent(myEvent);
					}
				}
				finally
				{
					if (!response.isLocal())
					{
						if (lyObject.isLocallyDeleted()
							|| !lyObject.isDeleted())
						{
							_cache.removeObject(object);
							lyObject.setDeleted(true);
						}
					}
					else
						lyObject.setLocallyDeleted(true);

					if (_locallyCreatedObjects != null)
						_locallyCreatedObjects.remove(object);

				}
				return;
			}
		}

		// Add the object in the response.
		// Retrieves the values that are changed.
		LyValueSet oldValues;
		boolean    isLocal = response.isLocal();

		if ((oldPeerValues == null) || (oldPeerValues.getSize() > 0))
		{
			if (isLocal)
			{
				oldValues = lyObject.getOldValues();
				lyObject.moveNewToPendingValues();
			}
			else
				oldValues = getOldValues(lyObject, oldPeerValues, true);

			oldValues = addOldLocalValues(oldValues, lyObject);
			lyObject.touch();
		}
		else
		{
			// We do not add local values if peer values was empty but we need
			// to validate the object and return an empty valueset to indicate that
			// changed values are not known.
			lyObject.validate();
			oldValues = new LyValueSet();
		}

		if (!response.isLocal()
		        && (oldValues != null)
			    && oldValues.hasMark(LyFieldInfo.MARK_ID))
			// Recompute object id
			rehash((LyObject)object, response.isRefresh());

		response.addResponse(object, oldValues);

		// Notify all the listeners.
		if (response.isRefresh())
		{
			Vector<LyDataSourceListener> skipListeners = response.getSkippedListeners();
			LyDataSourceEvent            event;

			event = new LyDataSourceEvent(LyObjectSource.this, LyDataSourceEvent.TYPE_SET, object, oldValues);
			event.setLocal(response.isLocal());
			postEvent(event, skipListeners);
		}
	} // end method receiveSet

	/**
	 * Method used to receive a new response from data provider to a request sent by this data
	 * source. This method may be called several times if several objects are concerned.
	 *
	 * @param     response  Response where there is a new object concerned.
	 * @param     status    Status of the new response. Valid status are defined in LyResponse.
	 * @param     object    Object concerned by this response.
	 * @param     oldValues Values that have changed for the object with this response (null when
	 *                      the operation is not a load or a modifcation).
	 * @exception IllegalArgumentException If response parameter is null.
	 * @see       LyResponse
	 * @see       #createObject
	 * @see       #setObject
	 * @see       #setObjects
	 * @see       #deleteObject
	 * @see       #deleteObjects
	 * @see       #loadObject
	 * @see       #loadObjects
	 */
	public void receiveResponse(LyResponse response, LyResponse.Status status, Object object,
		LyValueSet oldValues)
	{
		super.receiveResponse(response, status, object, oldValues);

		LyResponse.Operation operation = response.getOperation();

		if ((operation == LyResponse.Operation.LOAD)
				&& _useGroupEvent)
		{
			if ((status == LyResponse.Status.COMPLETED)
					|| (status == LyResponse.Status.ERROR)
					|| (object == null))
			{
				String		   groupId = "" + response.hashCode();
				LyEventManager em = _application.getEventManager();

				em.postGroupEvents(groupId);
			}
		}
	}

	/**
	 * Internal method used to receive a new response from data provider to a load request sent on
	 * loadObject or loadObjects. This method is overriden to filter objects according to class
	 * binding.
	 *
	 * @param response  Response to the load operation.
	 * @param status    Status of the new response. Valid status are defined in LyResponse.
	 * @param object    Loaded object.
	 * @param oldValues Old values of the object before the load operation.
	 * @see   #loadObject
	 * @see   #loadObjects
	 */
	protected void receiveLoad(LyResponse response, LyResponse.Status status, Object object,
		LyValueSet oldValues)
	{
        boolean     refresh = !_application.isIdle();
        String      groupId = null;

        if (refresh)
        {
    		if (_useGroupEvent)
    		{
    			LyEventManager em = _application.getEventManager();

    			groupId = "" + response.hashCode();
    			em.groupEvents(groupId);
    		}
        }

		LyDataSourceEvent event = null;

		// If the object is in the cache : it's a reload.
		// First we must verify if the object already matches the filter.
		// If not it's a deletion.
		// Else, it's a modification.
		Object			  peer = ((LyObject)object).getPeer();
		Object			  cachedObject = _cache.getObjectFromPeer(peer);

		if (cachedObject == null)
		{
			// Putting the objet in the cache.
			LyObjectId objectId = ((LyObject)object).getObjectId();

			if (_cache.isRemoved(objectId))
				return;
			_cache.putObject(object);

            if (refresh)
            {
    			// Prepare the event for the listeners.
    			event = new LyDataSourceEvent(this, LyDataSourceEvent.TYPE_CREATE, object, null);
    			event.setLocal(response.isLocal());
    			oldValues = new LyValueSet();
            }
		}
		else
		{
			// We must verify if the object matches the filter
			LyClassInfo    classInfo = getClassInfo();
			LyClassBinding classBinding = classInfo.getBinding();

			// case of temporary object
			if (object != cachedObject)
			{
				((LyObject)object).setDeleted(true);
				object = cachedObject;
			}

			if (classBinding != null)
			{
				LyFilter filter = classBinding.getFilter();

				if ((filter != null) && !filter.match(object))
				{
					// We must remove the object from the cache.
					_cache.startRemoveObject(object);

					try
					{
						if (refresh)
						{
							// Prepare the event for the listeners.
							event = new LyDataSourceEvent(this, LyDataSourceEvent.TYPE_DELETE, object, null);
							event.setLocal(response.isLocal());
							postEvent(event);
						}
					}
					finally
					{
						_cache.removeObject(object);
						((LyObject)object).setDeleted(true);
					}
					return;
				}
			}

			// The object has changed : prepare the event for the listeners.
			oldValues = getOldValues((LyObject)object, oldValues, false);
			oldValues = addOldLocalValues(oldValues, (LyObject)object);
			if (refresh && (oldValues != null))
			{
				event = new LyDataSourceEvent(this, LyDataSourceEvent.TYPE_SET, object, oldValues);
				event.setLocal(response.isLocal());
			}
		}

		// Reset informations stocked in the object.
		((LyObject)object).touch();

		// Set the status to the response.
		response.setStatus(status);

		// Add the object in the response.
		response.addResponse(object, oldValues);

		if (_locallyCreatedObjects != null)
			_locallyCreatedObjects.remove(object);
		// Notify all the listeners.
		if (event != null)
		{
			Vector<LyDataSourceListener> skipListeners = response.getSkippedListeners();

			// XXX workaround for join-table bug on creation
			event.setLoading(true);
			event.setGroupId(groupId);
			postEvent(event, skipListeners);
		}
	} // end method receiveLoad

	/**
	 * Internal method used to receive a new response from data provider to a deletion request sent
	 * on deleteObject or deleteObjects. A deletion event is sent to listeners and the object is
	 * deleted from the cache If the data provider has a notifier, the deletion is notified.
	 *
	 * @param response Response to the deletion operation.
	 * @param status   Status of the new response. Valid status are defined in LyResponse.
	 * @param object   Deleted object.
	 * @see   #deleteObject
	 * @see   #deleteObjects
	 */
	protected void receiveDelete(LyResponse response, LyResponse.Status status, Object object)
	{
		// Mark object as removed in the cache.
		if (!((LyObject)object).isDeleted())
			_cache.startRemoveObject(object);

		try
		{
			// Set the status to the response.
			response.setStatus(status);

			// Add the object in the response.
			response.addResponse(object, null);

			// Notify listeners.
			if (response.isRefresh())
			{
				Vector<LyDataSourceListener> skipListeners = response.getSkippedListeners();
				LyDataSourceEvent            event;

				event = new LyDataSourceEvent(this, LyDataSourceEvent.TYPE_DELETE, object, null);
				event.setLocal(response.isLocal());
				postEvent(event, skipListeners);
			}
		}
		finally
		{
			// Delete object from the cache
			if (((((LyObject)object).getPeer() == null)
					&& !((LyObject)object).isDeleted())
				|| (!response.isLocal()
					&& (((LyObject)object).isLocallyDeleted()
						|| !((LyObject)object).isDeleted())))
			{
				_cache.removeObject(object);
			}

			((LyObject)object).setDeleted(true);

			if (response.isLocal())
				((LyObject)object).setLocallyDeleted(true);

			if (_locallyCreatedObjects != null)
				_locallyCreatedObjects.remove(object);
			_getsizeCache = null;
		}
	} // end method receiveDelete

	/**
	 * This method is used by the event notifier to notify the data source for data changes made by
	 * other applications using the same datas.
	 *
	 * @param  groupId Id of the group of events. May be null for non grouped events.
	 * @param  type    Type of the event. Valid types are defined in LyEventNotifier.
	 * @param  object  Object concerned by the event
	 * @param  values  New values.
	 * @return True if the event was treated with success, false otherwise.
	 */
	public boolean receiveEvent(String groupId, LyEventNotifier.EventType type, Object object, LyValueSet values)
	{
		boolean result = super.receiveEvent(groupId, type, object, values);

		if (result && (type == LyEventNotifier.EventType.DELETE))
			((LyObject)object).setDeleted(true);

		return result;
	}

	/**
	 * To validate this objectSource. This method scans all fields of the associated class info and
	 * register this object source as listener on all object sources corresponding to classes that
	 * are pointed by relations. The purpose is to receive events on id. and name changes on related
	 * objects to transform them as modifications on objects managed by this objects source. As
	 * related objects are presented by their names, the goal is to ease the refreshing of the
	 * display when names change.
	 */
	public void validate()
	{
		if (((LyClassInfo)_class).getParent() instanceof LyArrayFieldInfo)
			return;

		LyClassInfo			      classInfo;
		LyObjectSource		      objectSource;
		LyFieldInfo			      fieldInfo;
		LyRelationFieldInfo       relation;
		LyStructFieldInfo	      struct;
		Vector<? extends LyField> fields = _class.getFields();
		int					      i, n = fields.size();
		int					      j, p;
		LyInfoList<LyClassInfo>   classInfoList = new LyInfoList<LyClassInfo>(false);

		// Building a list of all pointed classInfos and of imported fields.
		for (i = 0; i < n; i++)
		{
			fieldInfo = (LyFieldInfo)fields.elementAt(i);

			if (fieldInfo instanceof LyArrayFieldInfo)
				fieldInfo = ((LyArrayFieldInfo)fieldInfo).getFieldInfo();

			if (fieldInfo instanceof LyStructFieldInfo)
			{
				struct = (LyStructFieldInfo)fieldInfo;

				if (_structListener == null)
					_structListener = new _LyStructListener();

				for (j = 0, p = struct.getClassInfoCount(); j < p; j++)
				{
					classInfo = struct.getClassInfo(j);
					objectSource = _application.getObjectSource(classInfo);
					objectSource.addListener(_structListener, LyDataSourceEvent.TYPE_SET);
				}
			}
			else if (fieldInfo instanceof LyRelationFieldInfo)
			{
				relation = (LyRelationFieldInfo)fieldInfo;

				for (j = 0, p = relation.getClassInfoCount(); j < p; j++)
				{
					classInfo = relation.getClassInfo(j);
					classInfoList.addInfo(classInfo);
				}
			}
		}

		// For each classInfo, listen to modification of NAME attribute and to deletions.
		LyInfoList<LyFieldInfo> fieldInfos;
		int					    c, nc = classInfoList.getSize();
		int					    f, nf;
		LyFieldInfoImport	    fieldImport;
		Vector<LyFieldInfo>     sourceFields = null;
		Vector<LyFieldInfo>     idFields = null;
		LyFieldInfo			    idField, importedField;

		for (c = 0; c < nc; c++)
		{
			classInfo = classInfoList.getInfo(c);
			objectSource = _application.getObjectSource(classInfo);
			sourceFields = null;

			if ((objectSource == null)
					|| !classInfo.isLoaded())
			{
				// If classInfo is still not loaded, wait for its loading to listen to it
				if (_waitForLoading == null)
					_waitForLoading = new LyInfoList<LyClassInfo>();
				_waitForLoading.addInfo(classInfo);
				continue;
			}

			if (classInfo.isNameDifferentOfId())
			{
				// Get NAME fields
				fieldInfos = classInfo.getFieldInfos(LyFieldInfo.MARK_NAME);
				sourceFields = fieldInfos.getInfos();
				if (sourceFields == null)
					sourceFields = new Vector<LyFieldInfo>();

				// Add id fields
				idFields = classInfo.getIdFields();
				nf = ((idFields == null) ? 0 : idFields.size());
				for (f = 0; f < nf; f++)
				{
					idField = idFields.get(f);
					if (!sourceFields.contains(idField))
						sourceFields.add(idField);
				}
			}
            else
            {
                // Add modifiable id fields
                idFields = classInfo.getIdFields();
                nf = ((idFields == null) ? 0 : idFields.size());
                for (f = 0; f < nf; f++)
                {
					idField = idFields.get(f);
                    if (idField.hasMark(LyFieldInfo.MARK_CREATE)
                        || idField.hasMark(LyFieldInfo.MARK_SET))
                    {
                        if (sourceFields == null)
							sourceFields = new Vector<LyFieldInfo>();
                        sourceFields.add(idField);
                    }
                }
            }

			// Add local imported fields
			for (i = 0; i < n; i++)
			{
				fieldInfo = (LyFieldInfo)fields.elementAt(i);
				if (!fieldInfo.isLocal())
					continue;

				fieldImport = fieldInfo.getImport();
				if ((fieldImport != null)
					&& (fieldImport.getClassInfo() == classInfo)
					&& ((sourceFields == null)
						|| !sourceFields.contains(fieldImport.getFieldInfo())))
				{
					if (sourceFields == null)
						sourceFields = new Vector<LyFieldInfo>();
					sourceFields.addElement(fieldImport.getFieldInfo());
				}
			}

	        // Add fields imported in formulas
			LyInfoList<LyFieldInfo> formulaImports;

            for (i = 0; i < n; i++)
            {
                fieldInfo = (LyFieldInfo)fields.elementAt(i);
                if (!fieldInfo.isLocal())
                    continue;
                if (fieldInfo.getFormula() == null)
                    continue;

                formulaImports = fieldInfo.getFormulaImports(classInfo);
                if (formulaImports == null)
                    continue;

                for (f = 0, nf = formulaImports.getSize(); f < nf; f++)
                {
                    importedField = formulaImports.getInfo(f);
                    if ((sourceFields == null)
                        || !sourceFields.contains(importedField))
                    {
                        if (sourceFields == null)
                            sourceFields = new Vector<LyFieldInfo>();
                        sourceFields.addElement(importedField);
                    }
                }
            }

			_relationListener = new _LyRelationListener(classInfo);
			objectSource.addListener(_relationListener, LyDataSourceEvent.TYPE_DELETE);
			if (sourceFields != null)
				objectSource.addListener(_relationListener, new Vector<LyField>(sourceFields));
		}
	} // end method validate

	/**
	 * Method to notify the loading of a class to an object source.
	 *
	 * @param classInfo the loaded class
	 */
	public void notifyClassLoading(LyClassInfo classInfo)
	{
		if ((_waitForLoading == null)
				|| !_waitForLoading.contains(classInfo))
			return;

		BitSet				    marks = new BitSet();
		Vector<LyFieldInfo>     sourceFields = null;
		LyInfoList<LyFieldInfo> fieldInfos;
		LyObjectSource		    objectSource;

		marks.set(LyFieldInfo.MARK_ID);
		marks.set(LyFieldInfo.MARK_NAME);
		fieldInfos = classInfo.getFieldInfos(marks, false);
		sourceFields = fieldInfos.getInfos();

		objectSource = _application.getObjectSource(classInfo);
        _relationListener = new _LyRelationListener(classInfo);
		objectSource.addListener(_relationListener, LyDataSourceEvent.TYPE_DELETE);
		if (sourceFields != null)
			objectSource.addListener(_relationListener, new Vector<LyField>(sourceFields));

		if (_waitForLoading.getSize() == 1)
			_waitForLoading = null;
		else
			_waitForLoading.removeInfo(classInfo);
	} // end method notifyClassLoading

	/**
	 * To get the instance of LyDataSourceListener of this object source.
	 *
	 * @return the instance of LyDataSourceListener of this object source.
	 */
	public LyDataSourceListener getListener()
	{
		return _listener;
	}

	/* Implementation of LyDataSourceListener */

	/**
	 * Internal method used to add local values to a given values set for local fields that depend
	 * on fields represented in values list.
	 *
	 * @param  values the value set
	 * @param  object the object
	 * @return the value set
	 */
	protected LyValueSet addOldLocalValues(LyValueSet values, LyObject object)
	{
		if (values == null)
			return null;

		// Find local depends
		int					    i, n;
		LyValue				    value;
		LyFieldInfo			    fieldInfo;
		LyInfoList<LyFieldInfo> localDepends = null;

		for (i = 0, n = values.getSize(); i < n; i++)
		{
			value = values.getFieldValue(i);
			fieldInfo = (LyFieldInfo)value.getField();
			localDepends = getLocalDepends(fieldInfo, localDepends);
		}

		if (localDepends == null)
			return values;

		// Add local values in values set for changes
		Object  newValue, oldValue;

		boolean forceAddValue = false;
		boolean ignoreAddValue = false;

		for (i = 0, n = localDepends.getSize(); i < n; i++)
		{
			fieldInfo = localDepends.getInfo(i);
			if (values.contains(fieldInfo))
				continue;
			ignoreAddValue = false;
			oldValue = object.getFromMemory(fieldInfo);

			if (fieldInfo.isLocal())
			{
				if (fieldInfo.isComputed())
				{
					if (((fieldInfo instanceof LyRelationFieldInfo)
								|| (fieldInfo instanceof LyArrayFieldInfo))
							&& (getApplication().getDaemon(fieldInfo) != null))
					{
						ignoreAddValue = true;
						forceAddValue = false;
					}
					else
						forceAddValue = true;
				}
				else if (fieldInfo.isImported())
					forceAddValue = true;
			}
			else
				forceAddValue = false;

			if (!ignoreAddValue)
			{
				newValue = object.getValue(fieldInfo);

				if (forceAddValue || !fieldInfo.equals(newValue, oldValue))
					values.addValue(fieldInfo, oldValue);
			}
		}

		return values;
	} // end method addOldLocalValues

	/**
	 * Gets local depends recursively for the given field info.
	 *
	 * @param  fieldInfo    the fieldInfo
	 * @param  localDepends The local depends result
	 * @return the local depends
	 */
	protected LyInfoList<LyFieldInfo> getLocalDepends(LyFieldInfo fieldInfo,
		LyInfoList<LyFieldInfo> localDepends)
	{
		LyInfoList<LyFieldInfo> depends = fieldInfo.getDepends();

		if (depends == null)
			return localDepends;

		int		    i, n = depends.getSize();
		LyFieldInfo depend;
		LyClassInfo classInfo = getClassInfo();

		for (i = 0; i < n; i++)
		{
			depend = depends.getInfo(i);
			if (!depend.isLocal())
				continue;
			if (!classInfo.contains(depend))
				// May happen when virtual classes are used
				continue;

			if (localDepends == null)
				localDepends = new LyInfoList<LyFieldInfo>();
			localDepends.addInfo(depend);

			getLocalDepends(depend, localDepends);
		}

		return localDepends;
	} // end method getLocalDepends

	/**
	 * Internal class to listen on peer objects update.
	 */
	class _LyDataSourceListener
		implements LyDataSourceListener
	{
		/**
		 * Method for receiving an object creation. This callback is called when an object peer is created.
		 *
		 * @param  event                 Event notifying the creation of an object.
		 * @return Boolean indicating if the event was used, false otherwise.
		 * @param  IllegalStateException If the received event was not thrown by the associated data
		 *                               source.
		 */
		public boolean objectCreated(LyDataSourceEvent event)
		{
			if (event.getSource() != _dataSource)
				throw new IllegalStateException("BF AppEngine : receives an event from an unknown dataSource");

			LyObject object = null;
			Object   peer = event.getObject();

			_getsizeCache = null;

			if (peer != null)
			{
				object = _cache.getObjectFromPeer(peer);

				if (object != null)
					return false;

				// The object does not exist : we must create it.
				object = createObjectFromPeer(peer);
				if (object == null)
					return false;

				// Putting the objet in the cache.
				_cache.putObject(object);

				if (event.isNotified())
					object.control((LyValueSet)null, true);

				if (!event.isLoading())
				    // Record new object
				    recordId(object);

				// Notify all the listeners.
				LyDataSourceEvent myEvent;

				myEvent = new LyDataSourceEvent(LyObjectSource.this, LyDataSourceEvent.TYPE_CREATE, object, null);
				myEvent.setLoading(event.isLoading());
				myEvent.setNotified(event.isNotified());
				myEvent.setLocal(event.isLocal());
				myEvent.setGroupId(event.getGroupId());

				return postEvent(myEvent);
			}

			return false;
		} // end method objectCreated

		/**
		 * Method for receiving an object deletion.
		 *
		 * @param  event Event notifying the deletion of an object.
		 * @return Boolean indicating if the event was used, false otherwise.
		 */
		public boolean objectDeleted(LyDataSourceEvent event)
		{
			if (event.getSource() != _dataSource)
				throw new IllegalStateException("BF AppEngine : receives an event from an unknown dataSource");

			Object   peer = event.getObject();
			LyObject object = null;

			_getsizeCache = null;

			if (peer != null)
				object = _cache.getObjectFromPeer(peer);

			if (object == null)
				return false;

			// Removing the object from the cache.
			_cache.startRemoveObject(object);
			boolean result = false;
			try
			{
				// Notify all the listeners.
				LyDataSourceEvent myEvent;

				myEvent = new LyDataSourceEvent(LyObjectSource.this, LyDataSourceEvent.TYPE_DELETE, object, null);
				myEvent.setLocal(event.isLocal());
				myEvent.setNotified(event.isNotified());
				myEvent.setGroupId(event.getGroupId());
				myEvent.setLoading(event.isLoading());

				result = postEvent(myEvent);
			}
			finally
			{
				_cache.removeObject(object);
				object.setDeleted(true);
				object.setLocallyDeleted(true);
			}
			return result;
		} // end method objectDeleted

		/**
		 * Method for receiving an object modification.
		 *
		 * @param  event Event notifying the modification of an object.
		 * @return Boolean indicating if the event was used, false otherwise.
		 */
		public boolean objectModified(LyDataSourceEvent event)
		{
			if (event.getSource() != _dataSource)
				throw new IllegalStateException("BF AppEngine : receives an event from an unknown dataSource");

			Object	    peer = event.getObject();
			LyClassInfo classInfo = getClassInfo();
			LyObject    object = null;

			_getsizeCache = null;

			if (peer != null)
			{
				object = _cache.getObjectFromPeer(peer);

				LyClassBinding classBinding = classInfo.getBinding();
				LyFilter	   filter = ((classBinding == null) ? null : classBinding.getFilter());

				if (object == null)
				{
					// The object is not in the cache
					if (filter != null)
					{
						// Maybe now the object match the filter.
						object = createObjectFromPeer(peer);
						if (object != null)
						{
							// Putting the object in the cache.
							_cache.putObject(object);

							// Notify all the listeners.
							LyDataSourceEvent myEvent;

							myEvent = new LyDataSourceEvent(LyObjectSource.this,
									LyDataSourceEvent.TYPE_CREATE, object, null);
							myEvent.setLocal(event.isLocal());
							myEvent.setNotified(event.isNotified());
							myEvent.setGroupId(event.getGroupId());
							return postEvent(myEvent);
						}
					}
					return false;
				}

				// We must verify if the object matches the filter
				if ((filter != null) && (!filter.match(object)))
				{
					// We must remove the object from the cache.
					_cache.startRemoveObject(object);

					try
					{
						// Notify all the listeners.
						LyDataSourceEvent myEvent;

						myEvent = new LyDataSourceEvent(LyObjectSource.this,
								LyDataSourceEvent.TYPE_DELETE, object, null);
						myEvent.setLocal(event.isLocal());
						myEvent.setNotified(event.isNotified());
						myEvent.setGroupId(event.getGroupId());
						postEvent(myEvent);
					}
					finally
					{
						_cache.removeObject(object);
						object.setDeleted(true);
					}
					return true;
				}
			}
			else
				return false;

			// Retrieves the values that are changed.
			LyValueSet oldPeerValues, oldValues;

			oldPeerValues = event.getValues();
			oldValues = getOldValues(object, oldPeerValues, false);
			oldValues = addOldLocalValues(oldValues, object);

			// Notify all the listeners.
			LyDataSourceEvent myEvent;

			if (event.isNotified())
				object.control(oldValues, true);
			object.touch();

			LyClass peerClass = object.getClassInfo().getPeerClass();
			boolean isRollBack = false;
			if (peerClass != null)
			{
				LyDataSource ds = peerClass.getDataSource();

				isRollBack = ds.isRollBack();
			}

			if (isRollBack)
				object.removePendingValues(null);

			myEvent = new LyDataSourceEvent(LyObjectSource.this, LyDataSourceEvent.TYPE_SET, object, oldValues);
			myEvent.setLocal(event.isLocal());
			myEvent.setNotified(event.isNotified());
			myEvent.setGroupId(event.getGroupId());

			if (event.isNotified())
				object.modified(oldValues);
			return postEvent(myEvent);
		} // end method objectModified
	} // end class _LyDataSourceListener


	/**
	 * Internal class to listen for update of of objects pointed by relations of the class info of
	 * this source.
	 */
	class _LyRelationListener
		implements LyAdvancedDataSourceListener
	{
	    // Listened class info.
        LyClassInfo                 _classInfo;

        // Relations that points to the listened class info.
        LyInfoList<LyFieldInfo>     _relations;

		// To avoid loops.
		boolean                     _notifying;


        _LyRelationListener(LyClassInfo classInfo)
        {
            _classInfo = classInfo;
            _relations = getRelations(classInfo);
        }

		/**
         * Not used.
		 */
		public boolean objectCreated(LyDataSourceEvent event)
		{
			return false;
		}

		/**
		 * Internal Method called when a pointed object is deleted.
		 *
		 * @param  event Event notifying the deletion of an object.
		 * @return Boolean indicating if the event was used, false otherwise.
		 */
		public boolean objectDeleted(LyDataSourceEvent event)
		{
			if (event.isLoading())
				// Delete resulting from an unload, ignore it to avoid reloads during unload...
				return false;

	        LyObject     object = (LyObject)event.getObject();

			return objectIdModified(null, object.getObjectId());
		}

		/**
		 * Internal method called when a pointed object name or id is modified.
		 *
		 * @param  event Event notifying the modification of an object.
		 * @return Boolean indicating if the event was used, false otherwise.
		 */
		public boolean objectModified(LyDataSourceEvent event)
		{
			if (_notifying)
				return false; // Avoid loop on circular relations
			if (_cache.isEmpty())
				// If we don't have objects in the cache, there's nothing to do
				return false;

			LyObject	 object = (LyObject)event.getObject();
			LyValueSet	 values = event.getValues();

            if ((values != null)
                && values.hasMark(LyFieldInfo.MARK_ID))
			{
                LyObjectId objectId = object.getObjectId();
				LyObjectId oldObjectId = object.getObjectId(values);

				if (oldObjectId == null)
				    return false;
				if (objectIdModified(objectId, oldObjectId))
				    return true;
			}

			// Propagate the notification on related objects
			LyObjectList	  	relatedObjects = getRelatedObjects(object, _relations);
			int				  	i, j ,k, nRelObjects = ((relatedObjects == null) ? 0 : relatedObjects.getSize());
			int				  	nRelValues, nValue;
			LyObject		  	relObject;
			LyObjectId		  	valId;
            LyValueSet        	relValues;
            LyValueSet        	changedValues;
            LyRelationValue	  	value;
			LyDataSourceEvent 	newEvent;
			boolean			  	result = false;
            LyEventInfo       	eventInfo;
            LyObjectId			objectId = object.getObjectId();
            boolean				modified = false;

			for (i = 0; i < nRelObjects; i++)
			{
				relObject = relatedObjects.getObject(i);
				if ((relObject == object)
						|| !relObject.checkObject())
					// If object is not found in cache, no need to update in this case.
					continue;

				changedValues = new LyValueSet();
				relValues = relObject.getValues(_relations);
				relValues = addOldLocalValues(relValues, relObject);
				nRelValues = relValues.getSize();

				if (nRelValues == 0)
					continue; // no values, no modification -- Should not happen?

				for (j = 0; j < nRelValues; j++)
				{
					modified = false;
					if (!(relValues.getFieldValue(j) instanceof LyRelationValue))
						continue;

					value = ((LyRelationValue)relValues.getFieldValue(j));
					nValue = value.getValueCount();

					for (k = 0; k < nValue; k++)
					{
						valId = value.getObjectId(k);
						if (objectId.equals(valId))
						{
							modified = true;
							break;
						}
					}
					if (modified)
						changedValues.add(value);
				}

				if (changedValues.size() > 0)
				{
    				newEvent = new LyDataSourceEvent(LyObjectSource.this, LyDataSourceEvent.TYPE_SET,
    						relObject, changedValues);
    				newEvent.setNotified(event.isNotified());
    				newEvent.setLoading(event.isLoading());
    				newEvent.setLocal(event.isLocal());

                    eventInfo = new LyEventInfo();
                    eventInfo.setObject(object);
                    eventInfo.setValues(values);
                    eventInfo.setAdditionalEventInfo(event.getEventInfo());
                    newEvent.setEventInfo(eventInfo);

    				_notifying = true;
    				// Reset last values to force recomputing of computed values
    				relObject.touch();
    				result |= postEvent(newEvent);
    				_notifying = false;
				}
			}

			return result;
		} // end method objectModified

	    /**
         * Internal method called when a pointed object name or id is modified.
         * @param objectId     New object id, null when object is deleted.
         * @param oldObjectId  Previous object id.
         * @return True if object id was really changed, false otherwise.
         */
        protected boolean objectIdModified(LyObjectId objectId, LyObjectId oldObjectId)
        {
            if ((objectId != null)
                    && objectId.equals(oldObjectId))
                return false;

            LyObjectList relatedObjects = getRelatedObjects(oldObjectId, _relations);
            int          n = (relatedObjects == null ? 0 : relatedObjects.getSize());

            if (n == 0)
                return true;

            LyObject     relObject;
            int          i, j, nr = _relations.getSize(), k, no;
            LyFieldInfo  relation;
            LyObjectId[] relObjectIds;
            LyObjectId   relObjectId;
            boolean      setValue, modified = false;
            LyValueSet   changedValues;

            for (i = 0; i < n; i++)
            {
                relObject = relatedObjects.getObject(i);
                changedValues = new LyValueSet();

                for (j = 0; j < nr; j++)
                {
                    relation = _relations.getInfo(j);
                    if (relation.isMultiple())
                    {
                        relObjectIds = (LyObjectId[])relObject.getValue(relation);
                        no = ((relObjectIds == null) ? 0 : relObjectIds.length);
                        setValue = false;

                        for (k = 0; k < no; k++)
                        {
                            if ((relObjectIds[k] != null)
                                && relObjectIds[k].equals(oldObjectId))
                            {
                                relObjectIds = relObjectIds.clone();
                                relObjectIds[k] = objectId;
                                setValue = true;
                            }
                        }

                        if (setValue)
                        {
                            relObject.setValue(relation, relObjectIds);
                            changedValues.addValue(relation, relObjectIds);
                            modified = true;
                        }
                    }
                    else
                    {
                        relObjectId = (LyObjectId)relObject.getValue(relation);
                        if ((relObjectId != null)
                            && relObjectId.equals(oldObjectId))
                        {
                            relObject.setValue(relation, objectId);
                            changedValues.addValue(relation, objectId);
                            modified = true;
                        }
                    }
                }

                if (modified)
                {
                    relObject.set();
                    modified = false;
                }
            }

            return true;
        }

		/**
		 * Internal method to retrieve objects that point the given object.
		 *
		 * @param  object    Pointed object.
		 * @param  relations List of relations that potentially point to source.
		 * @return List of all objects that point the specified object.
		 */
		protected LyObjectList getRelatedObjects(LyObject object, LyInfoList<LyFieldInfo> relations)
		{
			LySimpleObjectList result = new LySimpleObjectList();

			_cache.getObjects(object, relations, result);
			return result;
		}

		/**
		 * Internal method to retrieve objects that point the given object.
		 *
		 * @param  object    Pointed object.
		 * @param  relations List of relations that potentially point to source.
		 * @return List of all objects that point the specified object.
		 */
		protected LyObjectList getRelatedObjects(LyObjectId oldObjectId,
			LyInfoList<LyFieldInfo> relations)
		{
			int				   i, nr = relations.getSize();
			LyFieldInfo		   relation;
			LyFilter		   filter = new LyFilter(LyFilter.Type.OR);
			LyFilterExpression filterExp;

			for (i = 0; i < nr; i++)
			{
				relation = relations.getInfo(i);
				filterExp = new LyFilterExpression(relation,
						(relation.isMultiple() ? LyFilterExpression.Condition.CONTAINS
											   : LyFilterExpression.Condition.EQUALS), oldObjectId);
				filter.addElement(filterExp);
			}

			LyResponse response = getObjects(null, null, filter, -1, true, null);

			return response.getObjects();
		}

		/**
		 * Internal method to retrieve relations that can point to the given class info.
		 *
		 * @param  classInfo Pointed class info.
		 * @return Result list of relations (or arrays of relations) that point to given class info.
		 */
		protected LyInfoList<LyFieldInfo> getRelations(LyClassInfo classInfo)
		{
			Vector<? extends LyField>    fields = _class.getFields();
			int					         i, n = fields.size();
			LyFieldInfo			         fieldInfo, subFieldInfo;
			LyInfoList<LyFieldInfo>      relations = new LyInfoList<LyFieldInfo>();

			for (i = 0; i < n; i++)
			{
				fieldInfo = (LyFieldInfo)fields.elementAt(i);

				if ((fieldInfo instanceof LyRelationFieldInfo)
						&& ((LyRelationFieldInfo)fieldInfo).relatedWith(classInfo))
					relations.addInfo(fieldInfo);
				else if (fieldInfo instanceof LyArrayFieldInfo)
				{
					subFieldInfo = ((LyArrayFieldInfo)fieldInfo).getFieldInfo();
					if ((subFieldInfo instanceof LyRelationFieldInfo)
							&& ((LyRelationFieldInfo)subFieldInfo).relatedWith(classInfo))
						relations.addInfo(fieldInfo);
				}
			}

			return relations;
		}

		/**
		 * Method for receiving state changes on an object.
		 *
		 * @param  event Event notifying the state change of an object.
		 * @return True indicating if the event was used, false otherwise.
		 */
		public boolean objectStateChanged(LyAdvancedDataSourceEvent event)
		{
			// Propagate the notification on related objects
			return false;
		}

		/**
		 * Method for receiving user event on an object.
		 *
		 * @param  event Event notifying the user event of an object.
		 * @return True indicating if the event was used, false otherwise.
		 */
		public boolean objectUserEvent(LyAdvancedDataSourceEvent event)
		{
			return false;
		}

		/**
		 * Called when grouped events are sent.
		 *
		 * @param  events Sorted list of events.
		 * @return True if the events are used by the listener, false otherwise.
		 */
		public boolean dataSourceEventsGrouped(ArrayList<LyDataSourceEvent> events)
		{
			if (_notifying)
				return false; // Avoid loop on circular relations

			// Sort events to get only interesting.
			if ((events == null) || (events.isEmpty()))
				return false;

			int							 e, nEvents = events.size();
			LyDataSourceEvent			 event;
			short						 type;
			ArrayList<LyDataSourceEvent> keepEvents = new ArrayList<LyDataSourceEvent>();
			LyObject					 object, relObject;
			LyObjectList				 relatedObjects;
			int							 i, j, k, nRelObjects, nValue, nRelValues;
			LyDataSourceEvent			 newEvent;
			LyValueSet					 values, relValues, changedValues;
			boolean                      modified;
			LyRelationValue              value;
			LyObjectId                   objectId, valId;
			LyEventInfo                  eventInfo;

			for (e = 0; e < nEvents; e++)
			{
				event = (LyDataSourceEvent)events.get(e);
				type = event.getType();

				if (type == LyDataSourceEvent.TYPE_CREATE)
					continue;

				if (type == LyDataSourceEvent.TYPE_DELETE)
				{
					objectDeleted(event);
					continue;
				}

				values = event.getValues();
				if ((values != null)
					&& values.hasMark(LyFieldInfo.MARK_ID))
				{
					objectModified(event);
					continue;
				}

				// Propagate the notification on related objects
				object = (LyObject)event.getObject();
				objectId = object.getObjectId();

				relatedObjects = getRelatedObjects(object, _relations);
				nRelObjects = ((relatedObjects == null) ? 0 : relatedObjects.getSize());

				for (i = 0; i < nRelObjects; i++)
				{
					relObject = relatedObjects.getObject(i);
					if ((relObject == object)
							|| !relObject.checkObject())
						// If object is not found in cache, no need to update in this case.
						continue;

	                changedValues = new LyValueSet();
	                relValues = relObject.getValues(_relations);
	                relValues = addOldLocalValues(relValues, relObject);
	                nRelValues = relValues.size();

	                if (nRelValues == 0)
	                    continue; // no values, no modification -- Should not happen?

	                for (j = 0; j < nRelValues; j++)
	                {
	                    modified = false;
	                    if (!(relValues.getFieldValue(j) instanceof LyRelationValue))
	                        continue;

	                    value = ((LyRelationValue)relValues.getFieldValue(j));
	                    nValue = value.getValueCount();

	                    for (k = 0; k < nValue; k++)
	                    {
	                        valId = value.getObjectId(k);
	                        if (objectId.equals(valId))
	                        {
	                            modified = true;
	                            break;
	                        }
	                    }
	                    if (modified)
	                        changedValues.add(value);
	                }

	                if (changedValues.size() > 0)
	                {
	                    newEvent = new LyDataSourceEvent(LyObjectSource.this, LyDataSourceEvent.TYPE_SET,
	                            relObject, changedValues);
	                    newEvent.setNotified(event.isNotified());
	                    newEvent.setLoading(event.isLoading());
	                    newEvent.setLocal(event.isLocal());

	                    eventInfo = new LyEventInfo();
	                    eventInfo.setObject(object);
	                    eventInfo.setValues(values);
	                    eventInfo.setAdditionalEventInfo(event.getEventInfo());
	                    newEvent.setEventInfo(eventInfo);

    					keepEvents.add(newEvent);
	                }
				}
			}

			if (keepEvents.isEmpty())
				return false;

			// Call postEvent on the event manager
			LyEventManager em = _application.getEventManager();

			_notifying = true;
			em.postEvents(keepEvents);
			_notifying = false;

			return true;
		} // end method dataSourceEventsGrouped
	} // end class _LyRelationListener


	/**
	 * Internal class to listen for update of of objects pointed by structs of the class info of
	 * this source.
	 */
	class _LyStructListener
		implements LyDataSourceListener
	{

		/** To avoid loops. */
		boolean _notifying;

		/**
		 * Not used.
		 */
		public boolean objectCreated(LyDataSourceEvent event)
		{
			return false;
		}

		/**
		 * Not used.
		 */
		public boolean objectDeleted(LyDataSourceEvent event)
		{
			return false;
		}

		/**
		 * Internal method called when a pointed object name or id is modified.
		 *
		 * @param  event Event notifying the modification of an object.
		 * @return Boolean indicating if the event was used, false otherwise.
		 */
		@SuppressWarnings("deprecation")
        public boolean objectModified(LyDataSourceEvent event)
		{
			if (_notifying)
				return false; // Avoid loop on circular relations

			// GTA en attente, correction  faire simultanment avec LyOperation
			// pour grer les vnements groups entre le struct et son parent.
			if (!event.isLoading())
				return false; // Other events are managed by LyContext and LyOperation.

			// Propagate the notification on related objects
			LyObject object = (LyObject)event.getObject();
			LyObject parent = object.getParent();

			if ((parent == null)
					|| (parent.getClassInfo() != getClassInfo()))
				return false;

			LyFieldInfo parentField = object.getParentFieldInfo();

			if (parentField == null)
				return false;

			LyDataSourceEvent newEvent;
			boolean			  result = false;
			LyValueSet		  values = new LyValueSet();

			values.addValue(parent.getFieldValue(parentField));
			newEvent = new LyDataSourceEvent(LyObjectSource.this, LyDataSourceEvent.TYPE_SET,
					parent, values);
			newEvent.setNotified(event.isNotified());
			newEvent.setLoading(event.isLoading());
			newEvent.setLocal(event.isLocal());

			LyEventInfo info = new LyEventInfo();

			info.setObject(object);
			info.setValues(event.getValues());
			info.setAdditionalEventInfo(event.getEventInfo());

			newEvent.setEventInfo(info);

			_notifying = true;
			result |= postEvent(newEvent);
			_notifying = false;

			return result;
		} // end method objectModified
	} // end class _LyStructListener

	/* Methods for using object source as data source */

	/**
	 * Gets an object from this data source by the specified identifier.
	 *
	 * @param  id           Object identifier of the searched object.
	 * @param  classBinding Binding between logical class and physical class.
	 * @return The requested object if it was found, or null if it could not be found.
	 */
	public Object getObject(LyObjectId id, LyClassBinding classBinding)
	{
		return getObject(id);
	}

	/**
	 * Return the number of objects within this data source matching the given filter.
	 *
	 * @param  filter       Filter applied on objects for counting.
	 * @param  classBinding Binding between logical class and physical class.
	 * @return Number of objects matching the given filter in this data source.
	 */

	public int getSize(LyFilter filter, LyClassBinding classBinding)
	{
		return getSize(filter);
	}

	/**
	 * Create a new object on the associated data provider.
	 *
	 * @param  object       Object to create.
	 * @param  listener     Listener of the response.
	 * @param  classBinding Binding between logical class and physical class.
	 * @return Response to the creation.
	 */
	public final LyResponse createObject(Object object, LyResponseListener listener,
		LyClassBinding classBinding)
	{
		return createObject((LyObject)object, listener);
	}

	/**
	 * Set an object on the associated data provider.
	 *
	 * @param  object       Object to modify.
	 * @param  listener     Listener of the response.
	 * @param  classBinding Binding between logical class and physical class
	 * @return Response to the modification.
	 */
	public final LyResponse setObject(Object object, LyResponseListener listener,
		LyClassBinding classBinding)
	{
		return setObject((LyObject)object, listener);
	}

	/**
	 * Set a list of objects on the associated data provider.
	 *
	 * @param  objects      List of objects to modify.
	 * @param  listener     Listener of the response.
	 * @param  classBinding Binding between logical class and physical class
	 * @return Response to the modification.
	 */
	public final LyResponse setObjects(Vector<? extends Object> objects, LyResponseListener listener,
		LyClassBinding classBinding)
	{
		return setObjects(objects, listener);
	}

	/**
	 * Delete an object on the associated data provider.
	 *
	 * @param  object       Object to delete.
	 * @param  listener     Listener of the response.
	 * @param  classBinding Binding between logical class and physical class
	 * @return Response to the deletion.
	 */
	public final LyResponse deleteObject(Object object, LyResponseListener listener,
		LyClassBinding classBinding)
	{
		return deleteObject((LyObject)object, listener);
	}

	/**
	 * Delete a list of objects on the associated data provider.
	 *
	 * @param  objects      List of objects to delete.
	 * @param  listener     Listener of the response.
	 * @param  classBinding Binding between logical class and physical class
	 * @return Response to the deletion.
	 */
	public final LyResponse deleteObjects(Vector<? extends Object> objects, LyResponseListener listener,
		LyClassBinding classBinding)
	{
		return deleteObjects(objects, listener);
	}

	/**
	 * Method to load an object from its id on the associated data provider.
	 *
	 * @param  id           Object identifier corresponding to the object to load.
	 * @param  fields       Fields to load (null means all fields).
	 * @param  listener     Listener of the response.
	 * @param  classBinding binding between logical class and physical class
	 * @return Response to the loading.
	 */
	public final LyResponse loadObject(LyObjectId id, Vector<? extends LyField> fields,
		LyResponseListener listener, LyClassBinding classBinding)
	{
		return loadObject(id, fields, listener);
	}

	/**
	 * Method for loading objects on the associated data provider.
	 *
	 * @param  fields       Fields to load (null means all fields).
	 * @param  sort         Gives sort criteria, or null if there is no sort.
	 * @param  filter       Gives filter criteria, or null if there is no filter.
	 * @param  n            Indicates the number of objects to load (-1 means no limit).
	 * @param  listener     Listener of the response.
	 * @param  classBinding binding between logical class and physical class
	 * @return Response to the loading.
	 */
	public final LyResponse loadObjects(Vector<? extends LyField> fields, LySort sort, LyFilter filter,
		int n, LyResponseListener listener, LyClassBinding classBinding)
	{
		return loadObjects(fields, sort, filter, n, listener);
	}

	/**
	 * Registers a listener to receive events of the given type on objects contained within this
	 * data source.
	 *
	 * @param     listener New listener. To receive these events, the listener have to implement the
	 *                     LyDataSourceListener interface.
	 * @param     type     Indicates which type of event the listener wants to receive. Types are
	 *                     defined as constant in LyDataSourceEvent class : TYPE_CREATE, TYPE_SET,
	 *                     TYPE_DELETE
	 * @exception IllegalArgumentException If the listener is null or if the given type is unknown.
	 */
	public void addListener(LyDataSourceListener listener, short type)
	{
		// Call addListener on event manager
		LyEventManager em = _application.getEventManager();

		em.addListener(this, listener, type);
	}

	/**
	 * Registers a listener to receive modification events concerning given fields on objects
	 * contained within this data source. Warning ! This method only registers the listener for
	 * modification events.
	 *
	 * @param     listener New listener. To receive these events, the listener have to implement the
	 *                     LyDataSourceListener interface.
	 * @param     fields   List of fields on which the listener wants to be notified when they are
	 *                     modified.
	 * @exception IllegalArgumentException If the listener is null.
	 */
	public void addListener(LyDataSourceListener listener, Vector<LyField> fields)
	{
		// Call addListener on event manager
		LyEventManager em = _application.getEventManager();

		em.addListener(this, listener, fields);
	}

	/**
	 * Unregisters a listener for notifications from this data source. The given listener is
	 * unregistered for all types of events he has been registered for.
	 *
	 * @param     listener The listener which shall not receive further events.
	 * @exception IllegalArgumentException If the listener is null or if it is not known as listener
	 *                                     on this data source.
	 */
	public void removeListener(LyDataSourceListener listener)
	{
		// Call addListener on event manager
		LyEventManager em = _application.getEventManager();

		em.removeListener(this, listener);
	}

	/**
	 * Internal method to post an event to listeners.
	 *
	 * @param     event         Posted event.
	 * @param     skipListeners Listeners that were already notified and that must be skipped by
	 *                          this method to avoid a double notification of the same event.
	 * @return    True if the vent was sent and used by at least one listener, false otherwise.
	 * @exception IllegalArgumentException If the given event is null.
	 */
	protected boolean postEvent(LyDataSourceEvent event, Vector<LyDataSourceListener> skipListeners)
	{
        if (_application.isIdle())
            return true;

		LyObject object = (LyObject)event.getObject();

		if ((object != null)
	            && !event.isLoading()
				&& (event.getType() != LyDataSourceEvent.TYPE_DELETE)
				&& !object.isDeleted()
				&& !checkObject(object))
			// Don't post event for an object that is not yet created
			return false;

		// Call postEvent on the event manager
		LyEventManager em = _application.getEventManager();
		boolean		   used = em.postEvent(event, skipListeners);

		return used;
	}

	/**
	 * Gets data source of physical layer associated to this object source.
	 *
	 * @return The data source of physical layer associated to this object source
	 */
	public LyDataSource getPeerDataSource()
	{
		return _dataSource;
	}

	/**
	 * Raises an error on this object source.
	 *
	 * @param error  The raised error
	 * @param object The object in error
	 */
	public void raiseError(LyError error, LyObject object)
	{
		LyDataSourceEvent event;

		event = new LyDataSourceEvent(this, LyDataSourceEvent.TYPE_ERROR_ADDED, object);
		event.setError(error);
		postEvent(event);
	}

	/**
	 * Removes an error on this object source.
	 *
	 * @param error  The removed error
	 * @param object The object in error
	 */
	public void removeError(LyError error, LyObject object)
	{
		LyDataSourceEvent event;

		event = new LyDataSourceEvent(this, LyDataSourceEvent.TYPE_ERROR_REMOVED, object);
		event.setError(error);
		postEvent(event);
	}

	/**
	 * Finalize method for an object, remove from cache
	 * @param object LyObject being finalized
	 */
	public void finalizeObject(LyObject object)
	{
		// Call checkObject for this object to avoid keeping references
		// in cache for objects that are finalized (see LySimpleCache).
		LyCache cache = getCache();

		if ((cache != null)
				&& ((!(cache instanceof LyFullCache)) || (cache.isLoaded())))
			cache.checkObject(object);
	}

} // end class LyObjectSource
