/*
 * LyOperation.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import leon.app.LyApplication;
import leon.app.LyDataSession;
import leon.data.datainterface.LyUserOperationInterface;
import leon.info.LyClassInfo;
import leon.info.LyFieldInfo;
import leon.info.LyInfoList;
import leon.info.LyStructFieldInfo;


/**
 * This class represents an operation in a context.
 *
 * @see     LyContext
 * @see     LyObject
 * @author  Lyria
 */
public class LyOperation
    implements Comparable<LyOperation>
{
    public static final String __VERSION = "$Id: LyOperation.java 34501 2012-03-13 08:51:28Z jlsaouli $";

    /** Counter for ordering operations. */
    private static long _counter = System.currentTimeMillis();

    /** Operation types. */
    public enum Type
    {
        // Creation type of an operation.
        CREATE,
        // Modification type of an operation.
        SET,
        // Deletion type of an operation.
        DELETE,
        // List of sub operations type of an operation.
        LIST,
        // Deletion of a set of objects type of an operation.
        DELETES,
        // Selection of an object.
        SELECT,
        // Unselection of an object.
        UNSELECT,
        // Hide an object.
        HIDE,
        // Show an object.
        SHOW,
        // Send an user event.
        USER,
        // User action
        USER_OPERATION

    }

    /** Type of the operation. */
    Type _type;

    /** Object impacted by the operation. */
    LyObject _object;

    /** Context of the operation. */
    LyContext _context;

    /** Set of values to modify an object. */
    LyValueSet _values;

    /** Old values for a SET operation. */
    LyValueSet _oldValues;

    /** List of operations managed as one operation. */
    ArrayList<LyOperation> _operations;

    /** List of objects to delete. */
    LyObjectList _objects;

    /** Indicates if this operation must be ignored in undo/redo actions. */
    boolean _undoIgnored;

    /** Indicates if this operation must be linked in undo/redo actions. */
    boolean _linked;

    /** Creation date of this operation. */
    long _creationDate;

    /** User event to send in case of Type.USER. */
    LyAdvancedDataSourceEvent _userEvent;

    boolean _onlyLocal;
    boolean _onlyNotLocal;

    LyUserOperationInterface _userOperation;

    Object _userData;

    /**
     * Constructs a new operation.
     *
     * @param  context  the context of the operation
     * @param  type     the type of the operation
     * @param  object   the object impacted by the operation
     */
    public LyOperation(LyContext context, Type type, LyObject object)
    {
        _context = context;
        _type = type;
        _object = object;
        _oldValues = null;
        _values = null;
        _operations = null;
        _objects = null;
        _creationDate = LyOperation._counter++;
        _userEvent = null;

        if ((_type == Type.SELECT)
                || (_type == Type.UNSELECT)
                || (_type == Type.HIDE)
                || (_type == Type.SHOW)
                || (_type == Type.USER))
            _undoIgnored = true;

        _onlyLocal = false;
        _onlyNotLocal = false;

        if (_context.isOnlyLocal())
            setOnlyLocal();
    }

    /**
     * Constructs a new operation.
     *
     * @param  context  the context of the operation
     * @param  type     the type of the operation
     * @param  object   the object impacted by the operation
     * @param  values   the set of values for modifying the given object.
     */
    public LyOperation(LyContext context, Type type, LyObject object, LyValueSet values)
    {
        this(context, type, object);
        _values = values;
    }

    /**
     * Construct a new operation.
     *
     * @param  context     the context of the operation.
     * @param  operations  List of sub operations to add in this operation.
     */
    public LyOperation(LyContext context, ArrayList<LyOperation> operations)
    {
        this(context, Type.LIST, null);
        _operations = operations;
    }

    /**
     * Construct a new operation to delete a list of objects.
     *
     * @param  context  The context owner of the operation.
     * @param  objects  The objects to delete
     */
    protected LyOperation(LyContext context, LyObjectList objects)
    {
        this(context, Type.DELETES, null);
        _objects = objects;
    }

    /**
     * Construct a new operation to send the specified user event.
     *
     * @param  context    The context owner of the operation.
     * @param  object     The object on which the event is sent.
     * @param  userEvent  The user event to send.
     */
    protected LyOperation(LyContext context, LyObject object, LyAdvancedDataSourceEvent userEvent)
    {
        this(context, Type.USER, object);
        _userEvent = userEvent;
    }

    /**
     * Construct a new operation to send the specified user event.
     *
     * @param  context    The context owner of the operation.
     * @param  object     The object on which the event is sent.
     * @param  userEvent  The user event to send.
     */
    protected LyOperation(LyContext context, LyObject object, LyUserOperationInterface callBack, Object userData)
    {
        this(context, Type.USER_OPERATION, object);
        _userOperation = callBack;
        _userData = userData;
    }

    /**
     * Gets the object impacted by this operation.
     *
     * @return  The object impacted by this operation (may be null).
     */
    public LyObject getObject()
    {
        return _object;
    }

    /**
     * Gets the objects impacted by this operation.
     *
     * @return  The objects impacted by this operation (may be null).
     */
    protected LyObjectList getObjects()
    {
        return _objects;
    }

    /**
     * To acces to the list of operations.
     *
     * @return  An iterator on the collection of operations.
     */
    protected Iterator<LyOperation> iterator()
    {
        if (_operations != null)
            return _operations.iterator();

        return null;
    }

    /**
     * To acces to the Flat internal list of operations.
     *
     * @return  current operation if simple, or a flat list of all suboperations
     */
    protected Collection<LyOperation> getOperations()
    {
        ArrayList<LyOperation> resu = new ArrayList<LyOperation>();
        Iterator<LyOperation>  iter = iterator();

        if ((_type != Type.LIST) || (iter == null) || (!iter.hasNext()))
            resu.add(this);
        else
        {
            // recurse and add sub-operations
            LyOperation subOper;

            while (iter.hasNext())
            {
                subOper = iter.next();
                resu.addAll(subOper.getOperations());
            }
        }

        return resu;
    }

    /**
     * Free this operation.
     */
    public void free()
    {
        if (_operations != null)
        {
            LyOperation           operation;
            Iterator<LyOperation> it;

            it = _operations.iterator();
            while (it.hasNext())
            {
                operation = it.next();
                if (operation != this)
                    operation.free();
            }

            _operations = null;
        }
    }

    /* Comparable implementation */

    /**
     * Implementation of the Comparable interface. This method is used to order operations by date.
     *
     * @param   o  The operation to compare
     *
     * @return  0 if the two objects are equals, 1 if this operation is greater, -1 otherwise.
     */
    public int compareTo(LyOperation operation)
    {
        long thisDate = getCreationDate();
        long opDate = operation.getCreationDate();

        if (thisDate == opDate)
            return 0;
        else if (thisDate > opDate)
            return 1;
        else
            return -1;
    }

    protected void setOnlyLocal()
    {
        _onlyLocal = true;
    }

    protected void setOnlyNotLocal()
    {
        _onlyNotLocal = true;
    }

    /**
     * Internal method to add an operation in the list of operations.
     *
     * @param  operation  the operation to add.
     */
    protected void addOperation(LyOperation operation)
    {
        if (_operations == null)
            _operations = new ArrayList<LyOperation>();

        _operations.add(operation);
    }

    /**
     * Method to execute this operation.
     *
     * @param   local    Indicates if the execution must be local or not.
     * @param   refresh  Indicates if the local operation may refresh the object.
     * @param   redo     Indicates if this operation is for a redo or not.
     * @param   control  Indicates which controls must be done on this operation.
     *
     * @return  the response of the operation or null for a values modification operation.
     */
    protected LyResponse send(boolean local, boolean refresh, boolean redo, int control)
    {
        LyResponse  response = null;
        LyOperation operation;

        switch (_type)
        {

        case CREATE:
            response = sendCreate(local, refresh, redo, control);
            break;

        case DELETE:
            response = sendDelete(local, refresh, redo, control);
            return response;

        case SET:
            response = sendSet(local, refresh, redo, control);
            if (local && !redo)
                return response;

            break;

        case DELETES:
            if (local)
            {
                // Create one operation for each object to delete.
                int      i, n = ((_objects == null) ? 0 : _objects.getSize());
                LyObject object;

                for (i = 0; i < n; i++)
                {
                    object = _objects.getObject(i);
                    operation = new LyOperation(_context, Type.DELETE, object);
                    addOperation(operation);
                }
            }

            break;

        case SELECT:
            _object.setSelected(true, refresh);
            return response;

        case UNSELECT:
            _object.setSelected(false, refresh);
            return response;

        case HIDE:
            _object.setHidden(true, refresh);
            return response;

        case SHOW:
            _object.setHidden(false, refresh);
            return response;

        case USER:
            if (refresh)
            {
                LyDataSession  session;
                LyApplication  application;
                LyEventManager em;

                session = _context.getSession();
                application = session.getApplication();
                em = application.getEventManager();
                em.postEvent(_userEvent);
            }

            return response;

        case USER_OPERATION:
            if (_userOperation != null)
                response = _userOperation.send(this, _userData, local, refresh, redo, control);

            return response;

        default:
            break;
        }

        if ((response != null)
                && (response.getStatus() == LyResponse.Status.ERROR))
            return response;

        if (_operations != null)
        {
            // Get all sub operations and send them.
            ArrayList<LyOperation> copyOfOperations = new ArrayList<LyOperation>(_operations);
            Iterator<LyOperation>  it = copyOfOperations.iterator();

            while (it.hasNext())
            {
                operation = it.next();
                if (operation != this)
                {
                    response = operation.send(local, refresh, redo, control);
                    if ((response != null)
                            && (response.getStatus() == LyResponse.Status.ERROR))
                        return response;
                }
            }
        }

        return response;
    } // end method send

    /**
     * Send a create operation.
     *
     * @param   local    Indicates if the execution must be local or not.
     * @param   refresh  Indicates if the local operation may refresh the object.
     * @param   redo     Indicates if this operation is for a redo or not.
     * @param   control  Indicates which controls must be done on this operation.
     *
     * @return  the response of the operation.
     */
    protected LyResponse sendCreate(boolean local, boolean refresh, boolean redo, int control)
    {
        if (_onlyLocal && !local)
        {
            // Flandres Restore the lyObject to original state after creation
            _object.removePendingValues(null, null, LyObject.CONTROL_NONE, LyObject.OPER_NONE);
            _object.removeNewValues(null, null, LyObject.CONTROL_NONE, LyObject.OPER_NONE);
            return null;
        }

        if (_onlyNotLocal && local)
        {
            LyObject    object = getObject();
            
            if ((object != null)
                    && object.hasNewValues())
                object.moveNewToPendingValues();
            return null;
        }

        LyResponse response = null;
        int        operFlags = LyObject.OPER_NONE;

        if (local)
            operFlags |= LyObject.OPER_LOCAL;

        if (refresh)
            operFlags |= LyObject.OPER_NEED_REFRESH;

        if (local && !redo)
            response = _object.create(_context, operFlags, control);
        else if (redo)
        {
            _object.setLocallyDeleted(false);
            _object.setDeleted(false);
            response = _object.create(null, operFlags, control);
        }
        else
        {
            if (!_object.isDeleted())
                response = _object.create(null, operFlags, control);
        }

        return response;
    } // end method sendCreate

    /**
     * Send a delete operation.
     *
     * @param   local    Indicates if the execution must be local or not.
     * @param   refresh  Indicates if the local operation may refresh the object.
     * @param   redo     Indicates if this operation is for a redo or not.
     * @param   control  Indicates which controls must be done on this operation.
     *
     * @return  the response of the operation.
     */
    protected LyResponse sendDelete(boolean local, boolean refresh, boolean redo, int control)
    {
        LyResponse            response = null;
        Iterator<LyOperation> it;
        LyOperation           operation;
        int                   operFlags = LyObject.OPER_NONE;

        if (local)
            operFlags |= LyObject.OPER_LOCAL;

        if (refresh)
            operFlags |= LyObject.OPER_NEED_REFRESH;

        if (local && !redo)
        {
            // Get sub objects and join objects
            LyObjectList localSubObjects, subObjects, joinObjects, arrayObjects;

            arrayObjects = _object.getArrayObjects();
            localSubObjects = _object.getSubObjects(true);
            subObjects = _object.getSubObjects(false);
            joinObjects = _object.getJoinObjects();

            // Delete local join objects, sub objects, peer and finally remaining sub objects

            // Delete join objects.
            if (joinObjects != null)
            {
                operation = new LyOperation(_context, joinObjects);
                addOperation(operation);
                operation.send(local, refresh, redo, control);
            }

            // Delete array objects
            if (arrayObjects != null)
            {
                operation = new LyOperation(_context, arrayObjects);
                addOperation(operation);
                operation.send(local, refresh, redo, control);
            }

            // Delete sub objects
            if (localSubObjects != null)
            {
                operation = new LyOperation(_context, localSubObjects);
                addOperation(operation);
                operation.send(local, refresh, redo, control);
            }

            addOperation(this);
            response = _object.delete(_context, operFlags, control);

            // Delete remaining sub objects
            if (subObjects != null)
            {
                operation = new LyOperation(_context, subObjects);
                addOperation(operation);
                operation.send(local, refresh, redo, control);
            }
        }
        else
        {
            // It is a commit of a local operation.
            if (_operations != null)
            {
                // Get all sub operations and send them.
                it = _operations.iterator();
                while (it.hasNext())
                {
                    operation = it.next();
                    if (operation == this)
                        response = _object.delete(null, operFlags, control);
                    else
                        operation.send(local, refresh, redo, control);
                }
            }
            else
                response = _object.delete(null, operFlags, control);
        }

        return response;
    } // end method sendDelete

    /**
     * Send a set operation..
     *
     * @param   local    Indicates if the execution must be local or not.
     * @param   refresh  Indicates if the local operation may refresh the object.
     * @param   redo     Indicates if this operation is for a redo or not.
     * @param   control  Indicates which controls must be done on this operation.
     *
     * @return  the response of the operation or null for a values modification operation.
     */
    protected LyResponse sendSet(boolean local, boolean refresh, boolean redo, int control)
    {
        LyResponse  response = null;
        LyClassInfo classInfo = _object.getClassInfo();

        if (_object.isDeleted())
        {
            response = new LyResponse(LyResponse.Operation.SET, classInfo);
            response.setLocal(local);
            response.setStatus(LyResponse.Status.COMPLETED);
            return response;
        }

        int operFlags = LyObject.OPER_NONE;

        if (local)
            operFlags |= LyObject.OPER_LOCAL;

        if (refresh)
            operFlags |= LyObject.OPER_NEED_REFRESH;

        if (_values == null)
        {
            if (local && !redo)
                response = _object.set(_context, operFlags, control);
            else
                response = _object.set(null, operFlags, control);

            if (response.getSize() > 0)
                _oldValues = response.getValues(0);
            else
                _oldValues = null;
        }
        else if (local)
        {
            LyValueSet  oldValues = new LyValueSet();
            int         i, n = _values.getSize();
            LyFieldInfo fieldInfo;
            LyValue     value, oldValue;
            Object      v, ov;

            for (i = 0; i < n; i++)
            {
                value = _values.getFieldValue(i);
                fieldInfo = (LyFieldInfo)value.getField();
                oldValue = _object.getFieldValue(fieldInfo);
                v = value.getValue();
                ov = oldValue.getValue();
                if (!fieldInfo.equals(v, ov)
                        || fieldInfo.isLocal())
                    oldValues.addValue(oldValue);
            }

            response = new LyResponse(LyResponse.Operation.SET, classInfo);
            response.setLocal(local);
            response.setStatus(LyResponse.Status.COMPLETED);

            if (oldValues.getSize() > 0)
            {
                // Don't send any change if there is'nt any !!!
                if (!redo)
                    _object.setValues(_context, _values, control, operFlags);
                else
                    _object.setValues(null, _values, control, operFlags);

                _oldValues = oldValues;

                if (refresh)
                {
                    ArrayList<LyDataSourceEvent> events = getUpdateEvents(local, false);

                    if (events != null)
                        _context.sendEvents(events);
                }
            }
        }
        else
            // Flandres
            // not local and not null, ensure the values are really the values we have for the operation
            _object.setValues(null, _values, control, operFlags);

        return response;
    } // end method sendSet

    /**
     * Method to rollback the operation. This method execute the inverse operation of the current
     * operation.
     *
     * @param  refresh  Indicates if the operation may refresh the objet or not.
     * @param  control  Indicates which controls must be done on this operation.
     */
    protected void rollback(boolean refresh, int control)
    {
        LyOperation           operation;
        Iterator<LyOperation> it;
        boolean               freeSubOperations = false;
        int                   i, n;

        // Get all sub operations in reverse order and rollback them.
        if (_operations != null)
        {
            n = _operations.size();
            for (i = n - 1; i >= 0; i--)
            {
                operation = _operations.get(i);
                if (operation != this)
                    operation.rollback(refresh, control);
            }
        }

        int operFlags = LyObject.OPER_LOCAL;

        if (refresh)
            operFlags |= LyObject.OPER_NEED_REFRESH;

        switch (_type)
        {

        case CREATE:
            if (!_onlyNotLocal)
                _object.delete(null, operFlags, control);

            break;

        case DELETE:
            if (_operations != null)
            {
                it = _operations.iterator();
                while (it.hasNext())
                {
                    operation = it.next();
                    if (operation == this)
                    {
                        _object.setDeleted(false);
                        _object.setLocallyDeleted(false);
                        _object.create(null, operFlags, control);
                    }
                }

                freeSubOperations = true;
            }
            else
            {
                _object.setDeleted(false);
                _object.setLocallyDeleted(false);
                _object.create(null, operFlags, control);
            }

            break;

        case SET:
            if (_values != null)
            {
                LyInfoList<LyFieldInfo> modifiedFields = new LyInfoList<LyFieldInfo>();

                for (i = 0; i < _values.getSize(); i++)
                    modifiedFields.addInfo((LyFieldInfo)_values.getFieldValue(i).getField());

                if ((_oldValues != null) && (_oldValues.getSize() > 0))
                {
                    LyInfoList<LyFieldInfo> oldFields = new LyInfoList<LyFieldInfo>();

                    for (i = 0; i < _oldValues.getSize(); i++)
                        oldFields.addInfo((LyFieldInfo)_oldValues.getFieldValue(i).getField());

                    // Set back the old values to the object for local values
                    // Old values are already encoded !!!
                    _object.skipEncoding(true);
                    _object.setValues(null, _oldValues, control, operFlags);
                    _object.skipEncoding(false);

                    if (refresh)
                    {
                        ArrayList<LyDataSourceEvent> events = getUpdateEvents(true, true);

                        if (events != null)
                            _context.sendEvents(events);
                    }
                }
                else
                    // Suppress new values.
                    _object.removeNewValues(null, modifiedFields, control, operFlags);
            }
            else if ((_oldValues != null) && (_oldValues.getSize() > 0))
                // Set back the old values to the object for local values
                _object.movePendingToNewValues();

            break;

        case DELETES:
            freeSubOperations = true;
            break;

        case SELECT:
            _object.setSelected(false, refresh);
            break;

        case UNSELECT:
            _object.setSelected(true, refresh);
            break;

        case HIDE:
            _object.setHidden(false, refresh);
            break;

        case SHOW:
            _object.setHidden(true, refresh);
            break;

        case USER:
            if (refresh)
            {
                LyDataSession  session;
                LyApplication  application;
                LyEventManager em;

                session = _context.getSession();
                application = session.getApplication();
                em = application.getEventManager();
                em.postEvent(_userEvent);
            }

            break;

        default:
            break;
        }

        // Get all sub operations and rollback them.
        if ((_operations != null) && (freeSubOperations))
        {
            it = _operations.iterator();
            while (it.hasNext())
            {
                operation = it.next();
                if (operation != this)
                    operation.free();
            }

            _operations = null;
        }
    } // end method rollback

    /**
     * Gets the list of LyDataSourceEvents correponding to this operation.
     *
     * @param   local       Indicates if events are local or not.
     * @param   rollbacked  Indicates if events correspond to rollback or not.
     *
     * @return  The list of events corresponding to this operation.
     */
    protected ArrayList<LyDataSourceEvent> getEvents(boolean local, boolean rollbacked)
    {
        ArrayList<LyDataSourceEvent> events = null;
        int                          i, n;
        LyOperation                  operation;
        ArrayList<LyDataSourceEvent> operEvents;

        if ((_operations != null) && rollbacked)
        {
            // Get all events of sub operations in reverse order.
            n = _operations.size();
            for (i = n - 1; i >= 0; i--)
            {
                operation = _operations.get(i);
                if (operation != this)
                {
                    operEvents = operation.getEvents(local, rollbacked);
                    if (operEvents != null)
                    {
                        if (events == null)
                            events = operEvents;
                        else
                            events.addAll(operEvents);
                    }
                }
            }
        }

        operEvents = null;
        switch (_type)
        {

        case CREATE:
            operEvents = getCreateEvents(local, rollbacked);
            break;

        case DELETE:
            operEvents = getDeleteEvents(local, rollbacked);
            if (operEvents != null)
            {
                if (events == null)
                    events = operEvents;
                else
                    events.addAll(operEvents);
            }

            return events;

        case SET:
            operEvents = getSetEvents(local, rollbacked);
            break;

        case DELETES:
            break;

        case SELECT:
            if (rollbacked)
                operEvents = getStateChangedEvent(LyAdvancedDataSourceEvent.TYPE_SELECTION, (short)1);
            else
                operEvents = getStateChangedEvent(LyAdvancedDataSourceEvent.TYPE_SELECTION, (short)0);

            break;

        case UNSELECT:
            if (rollbacked)
                operEvents = getStateChangedEvent(LyAdvancedDataSourceEvent.TYPE_SELECTION, (short)0);
            else
                operEvents = getStateChangedEvent(LyAdvancedDataSourceEvent.TYPE_SELECTION, (short)1);

            break;

        case HIDE:
            if (rollbacked)
                operEvents = getStateChangedEvent(LyAdvancedDataSourceEvent.TYPE_VISIBILITY, (short)0);
            else
                operEvents = getStateChangedEvent(LyAdvancedDataSourceEvent.TYPE_SELECTION, (short)1);

            break;

        case SHOW:
            if (rollbacked)
                operEvents = getStateChangedEvent(LyAdvancedDataSourceEvent.TYPE_VISIBILITY, (short)1);
            else
                operEvents = getStateChangedEvent(LyAdvancedDataSourceEvent.TYPE_SELECTION, (short)0);

            break;

        case USER:
            operEvents = new ArrayList<LyDataSourceEvent>();
            operEvents.add(_userEvent);
            break;

        default:
            break;
        }

        if (operEvents != null)
        {
            if (events == null)
                events = operEvents;
            else
                events.addAll(operEvents);
        }

        if ((_operations != null) && (!rollbacked))
        {
            // Get all sub operations and send them.
            Iterator<LyOperation> it = _operations.iterator();

            while (it.hasNext())
            {
                operation = it.next();
                if (operation != this)
                {
                    operEvents = operation.getEvents(local, rollbacked);
                    if (operEvents != null)
                    {
                        if (events == null)
                            events = operEvents;
                        else
                            events.addAll(operEvents);
                    }
                }
            }
        }

        return events;
    } // end method getEvents

    ArrayList<LyDataSourceEvent> getStateChangedEvent(short type, short oldState)
    {
        ArrayList<LyDataSourceEvent> events = new ArrayList<LyDataSourceEvent>();
        LyAdvancedDataSourceEvent    event;
        LyObjectSource               source;

        source = _object.getObjectSource();
        event = new LyAdvancedDataSourceEvent(source, type, _object, oldState);
        events.add(event);
        return events;
    }

    ArrayList<LyDataSourceEvent> getCreateEvents(boolean local, boolean rollbacked)
    {
        if (_onlyLocal && !local)
            return null;

        if (_onlyNotLocal && local)
            return null;

        ArrayList<LyDataSourceEvent> events = new ArrayList<LyDataSourceEvent>();
        LyObjectSource               source = _object.getObjectSource();
        LyDataSourceEvent            event;
        short                        type;

        if (rollbacked)
            type = LyDataSourceEvent.TYPE_DELETE;
        else
            type = LyDataSourceEvent.TYPE_CREATE;

        event = new LyDataSourceEvent(source, type, _object, null);
        event.setLocal(local);
        events.add(event);

        return events;
    }

    ArrayList<LyDataSourceEvent> getDeleteEvents(boolean local, boolean rollbacked)
    {
        ArrayList<LyDataSourceEvent> events = new ArrayList<LyDataSourceEvent>();
        LyObjectSource               source = _object.getObjectSource();
        LyDataSourceEvent            event;
        short                        type;

        if (rollbacked)
            type = LyDataSourceEvent.TYPE_CREATE;
        else
            type = LyDataSourceEvent.TYPE_DELETE;

        if (_operations != null)
        {
            // Get all sub operations and send them.
            // If rollbacked get operations in reverse order.
            int                          i, n = _operations.size();
            int                          limit;
            LyOperation                  operation;
            ArrayList<LyDataSourceEvent> operEvents;

            if (rollbacked)
            {
                i = n - 1;
                limit = -1;
            }
            else
            {
                i = 0;
                limit = n;
            }

            while (i != limit)
            {
                operation = _operations.get(i);
                if (operation == this)
                {
                    event = new LyDataSourceEvent(source, type, _object, null);
                    event.setLocal(local);
                    events.add(event);
                }
                else
                {
                    operEvents = operation.getEvents(local, rollbacked);
                    if (operEvents != null)
                        events.addAll(operEvents);
                }

                if (rollbacked)
                    i--;
                else
                    i++;
            }
        }
        else
        {
            event = new LyDataSourceEvent(source, type, _object, null);
            event.setLocal(local);
            events.add(event);
        }

        return events;
    } // end method getDeleteEvents

    /**
     * Internal metho to get a list of events corresponding to a set operation.
     *
     * @param   local
     * @param   rollbacked
     *
     * @return
     */
    ArrayList<LyDataSourceEvent> getSetEvents(boolean local, boolean rollbacked)
    {
        // Can't send an event without any old value (for a rollback, old values
        // are in _values)
        if ((rollbacked) && (_values == null))
            return null;

        // Can't send an event without any old value (for a commit, old values
        // are in _oldValues).
        if ((!rollbacked) && (_oldValues == null))
            return null;

        // if _values are not null, it is an operation due to a call to
        // LyObject.setValue(). We don't send this event because there is
        // also an operation due to a call to LyObject.set with all modified
        // values : with this we optimize the number of events (one event instead
        // of n with n the number of modified values).
        if ((!rollbacked) && (_values != null))
            return null;

        return getUpdateEvents(local, rollbacked);
    }

    @SuppressWarnings("deprecation")
    ArrayList<LyDataSourceEvent> getUpdateEvents(boolean local, boolean rollbacked)
    {
        ArrayList<LyDataSourceEvent> events = new ArrayList<LyDataSourceEvent>();
        LyObjectSource               source = _object.getObjectSource();
        LyDataSourceEvent            event;
        LyValueSet                   values = (rollbacked ? _values : _oldValues);

        event = new LyDataSourceEvent(source, LyDataSourceEvent.TYPE_SET, _object, values);
        event.setLocal(local);
        events.add(event);

        // GTA : en attente : correction  faire simultanment avec LyObjectSource pour
        // grer les vnements groups entre struct et son parent.
        LyClassInfo classInfo = _object.getClassInfo();
        LyObject    parent;
        LyFieldInfo parentField;
        LyObject    object = _object;
        LyEventInfo info = new LyEventInfo();

        info.setObject((LyObject)event.getObject());
        info.setValues(event.getValues());

        while (classInfo.getParent() instanceof LyStructFieldInfo)
        {
            parent = object.getParent();
            parentField = object.getParentFieldInfo();
            if ((parent != null) && (parentField != null))
            {
                values = new LyValueSet();

                values.addValue(parent.getFieldValue(parentField));
                event = new LyDataSourceEvent(parent.getObjectSource(), LyDataSourceEvent.TYPE_SET, parent, values);
                event.setEventInfo(info);
                event.setLoading(false);
                event.setLocal(local);
                events.add(event);
                object = parent;
                classInfo = parent.getClassInfo();
                info = new LyEventInfo();
                info.setObject(object);
                info.setValues(values);
                info.setAdditionalEventInfo(event.getEventInfo());
            }
            else
                break;
        }

        return events;
    } // end method getUpdateEvents

    /**
     * To get the type of the current operation.
     *
     * @return  The type of the current operation.
     */
    public Type getType()
    {
        return _type;
    }

    /**
     * To get the type of the current operation.
     *
     * @return      The type of the current operation.
     *
     * @deprecated  Use getType().
     */
    public Type getOperation()
    {
        return getType();
    }

    /**
     * To tag this operation as undo ignore.
     *
     * @param  ignore  true to tag this operation as undo ignore, false otherwise.
     */
    protected void setUndoIgnored(boolean ignore)
    {
        _undoIgnored = ignore;
    }

    /**
     * Test if this operation is tagged as undo ignore.
     *
     * @return  true if this operation is tagged undo ignore, false otherwise.
     */
    protected boolean isUndoIgnored()
    {
        return _undoIgnored;
    }

    /**
     * To tag this operation as linked with another for undo/redo operations.
     *
     * @param  linked  true to link this operation, false otherwise.
     */
    protected void setLinked(boolean linked)
    {
        _linked = linked;
    }

    /**
     * Test if this operation is linked or not.
     *
     * @return  true if this operation is linked, false otherwise.
     */
    protected boolean isLinked()
    {
        return _linked;
    }

    /**
     * Get date of creation of this operation.
     *
     * @return  A long representing the creation date of this operation.
     */
    protected long getCreationDate()
    {
        return _creationDate;
    }

    /**
     * Gets this object as a string.
     *
     * @return  A string represention of this operation.
     */
    public String toString()
    {
        String oper = "[OPERATION " + hashCode();
        int    no = ((_operations == null) ? 0 : _operations.size());

        if (no > 0)
            oper += ", " + no + " sub-operations";

        switch (_type)
        {

        case CREATE:
            oper += ", CREATE " + _object;
            break;

        case DELETE:
            oper += ", DELETE " + _object;
            break;

        case SET:
            if (_values == null)
                oper += ", SET " + _object;
            else
                oper += ", SET_VALUES " + _object + "\n\t" + _values;

            break;

        case LIST:
            oper += ", LIST ";
            break;

        case DELETES:
            oper += ", DELETES " + _objects.toString();
            break;

        case SELECT:
            oper += ", SELECT " + _object;
            break;

        case UNSELECT:
            oper += ", UNSELECT " + _object;
            break;

        case HIDE:
            oper += ", HIDE " + _object;
            break;

        case SHOW:
            oper += ", SHOW " + _object;
            break;

        case USER:
            oper += ", USER EVENT " + _object;
            break;

        default:
            break;
        }

        if (_operations != null)
        {
            StringBuffer buffer = new StringBuffer(oper);
            LyOperation  operation;
            int          i, n = _operations.size();

            buffer.append("\n");
            for (i = 0; i < n; i++)
            {
                operation = _operations.get(i);
                if (operation == this)
                    buffer.append("[OPERATION this " + hashCode() + "]");
                else
                    buffer.append(operation.toString());

                buffer.append("\n");
            }

            buffer.append("OPERATION]");
            return buffer.toString();
        }
        else
            return oper + ']';
    } // end method toString

    /**
     * Checks if this operation has errors.
     *
     * @return  the maximum error level of objects concerned by the operation.
     */
    protected short getErrorLevel()
    {
        short errorLevel = LyObject.LEVEL_OK, level;

        if (_object != null)
            errorLevel = _object.getErrorLevel();

        if ((_objects != null)
                && (errorLevel != LyObject.LEVEL_ERROR))
        {
            int      i, n = _objects.getSize();
            LyObject object;

            for (i = 0; (i < n) && (errorLevel != LyObject.LEVEL_ERROR); i++)
            {
                object = _objects.getObject(i);
                if (object.isDeleted())
                    continue;

                level = object.getErrorLevel();
                if ((level != LyObject.LEVEL_OK)
                        && ((errorLevel == LyObject.LEVEL_OK)
                            || (level < errorLevel)))
                    errorLevel = level;
            }
        }

        if ((_operations != null)
                && (errorLevel != LyObject.LEVEL_ERROR))
        {
            Iterator<LyOperation> it = _operations.iterator();
            LyOperation           operation;

            while (it.hasNext()
                    && (errorLevel != LyObject.LEVEL_ERROR))
            {
                operation = it.next();
                if (operation == this)
                    continue;

                level = operation.getErrorLevel();
                if ((level != LyObject.LEVEL_OK)
                        && ((errorLevel == LyObject.LEVEL_OK)
                            || (level < errorLevel)))
                    errorLevel = level;
            }
        }

        return errorLevel;
    } // end method getErrorLevel

    /**
     * Gets the list of objects locked by this result with LEVEL_ERROR errors.
     *
     * @return  The list of objects locked by this result and with LEVEL_ERROR errors or null if
     *          there is no objects with LEVEL_ERROR errors.
     */
    protected Vector<LyObject> getObjectsInError()
    {
        Vector<LyObject> objects = null;

        if ((_object != null)
                && (_object.getErrorLevel() == LyObject.LEVEL_ERROR))
        {
            objects = new Vector<LyObject>();
            objects.addElement(_object);
        }

        if (_objects != null)
        {
            int      i, n = _objects.getSize();
            LyObject object;

            for (i = 0; i < n; i++)
            {
                object = _objects.getObject(i);
                if (object.getErrorLevel() == LyObject.LEVEL_ERROR)
                {
                    if (objects == null)
                        objects = new Vector<LyObject>();

                    if (!objects.contains(object))
                        objects.addElement(object);
                }
            }
        }

        if (_operations != null)
        {
            Iterator<LyOperation> it = _operations.iterator();
            LyOperation           operation;
            Vector<LyObject>      errObjects;
            int                   i, n;
            LyObject              object;

            while (it.hasNext())
            {
                operation = it.next();
                if (operation == this)
                    continue;

                errObjects = operation.getObjectsInError();
                if (errObjects != null)
                {
                    n = errObjects.size();
                    for (i = 0; i < n; i++)
                    {
                        object = errObjects.elementAt(i);
                        if (objects == null)
                            objects = new Vector<LyObject>();

                        if (!objects.contains(object))
                            objects.add(object);
                    }
                }
            }
        }

        return objects;
    } // end method getObjectsInError

    /**
     * Gets the list of errors on locked objects with the specified level.
     *
     * @param   level  Level of the errors
     *
     * @return  The list of errors with the specified level for locked objects or null if there is
     *          no errors.
     */
    public Vector<LyError> getErrors(short level)
    {
        Vector<LyObject> objects = getObjectsInError();

        if (objects == null)
            return null;

        Vector<LyError> errors = null;
        LyObject        object;
        int             i, n = objects.size();
        LyError         error;
        int             j, ne;

        for (i = 0; i < n; i++)
        {
            object = objects.elementAt(i);
            ne = object.getErrorsCount();
            for (j = 0; j < ne; j++)
            {
                error = object.getError(j);
                if (error.getLevel() == level)
                {
                    if (errors == null)
                        errors = new Vector<LyError>();

                    errors.addElement(error);
                }
            }
        }

        return errors;
    } // end method getErrors

    /**
     * Accessor to internal values : Set of values to modify an object. Internal use only.
     *
     * @return  values or null if none
     */
    protected LyValueSet getValues()
    {
        return _values;
    }

    /**
     * Accessor to OLD internal values : Set of values to modify an object. Internal use only.
     *
     * @return  values or null if none
     */
    protected LyValueSet getOldValues()
    {
        return _oldValues;
    }

    /**
     * Context of the operation.
     */
    protected LyContext getContext()
    {
        return _context;
    }
} // end class LyOperation
