/*
 * LyPageWorkSpace.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;

import leon.app.*;
import leon.info.*;


/**
 * This class implements a page workspace : workspace where objects are loaded per "page" containing
 * a given number of objects of a class info and optionally. matching a filter.
 *
 * <p>This mode of loading requires to be able to order and filter objects. These parameters are
 * calculated on a field info that is given on construction of the workspace.</p>
 *
 * <p>Once loaded, it's possible on demand to load the next or the previous page.</p>
 */
public final class LyPageWorkSpace extends LyClassWorkSpace
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyPageWorkSpace.java 40675 2012-09-05 16:26:11Z jlsaouli $";

	public static boolean		FORCE_RELOAD_PAGE_FOR_DATES = false;
	
	public static int			DEFAULT_LIMIT = 10;
	
	public static int			MAX_LIMIT = 10000;
	
	public static int			MAX_PEER_LIMIT = 1000000;
	
	// Iterator limit must be < to MAX_LIMIT
	public static int			ITERATOR_LIMIT = 1000;
	
	
	/** Page sort. */
	LySort					   _pageSort;

	/** Page sort. */
	LySort					   _classSort;

	/** Number of objects per page in this page workspace. */
	int						   _limit;

	/** Temporary number of objects per page in this page workspace. */
	int						   _temporaryLimit;

	/** Object corresponding to the current min value for the limit field. */
	LyObject				   _min;

	/** Object corresponding to the current max value for the limit field. */
	LyObject				   _max;

	/** Boolean indicating if previous or next page was asked. */
	boolean					   _next;

	/** Page number. */
	short					   _page;

	/** Filter used internally for page workspace (avoid conflict with _filter). */
	LyFilter				   _pageFilter;

	/** The total number of objects for the class of this workspace (including filtering). */
	Integer					   _totalCount;

	/** Boolean indicating if peer can paginated results. */
	boolean                    _paginatedByPeer;

	/** Boolean indicating if peer can limit results. */
	boolean                    _partialNotificationByPeer;


	/**
	 * Constructs a class workspace within a given session for a class info.
	 *
	 * @param session    Session in which the workspace is constructed.
	 * @param classInfo  Class info to which objects of this workspace belong.
	 * @param filter     Filter applied to objects for being in this workspace.
	 * @param sort       Sort requested in this workspace.
	 * @param limit      Number of objects per page.
	 */
	public LyPageWorkSpace(LyDataSession session, LyClassInfo classInfo,
		LyFilter filter, LySort sort, int limit)
	{
		this(session, classInfo, null, filter, sort, limit, true, true);
	}

	/**
	 * Constructs a class workspace within a given session for a class info.
	 *
	 * @param session    Session in which the workspace is constructed.
	 * @param classInfo  Class info to which objects of this workspace belong.
	 * @param fieldInfos List of fieldinfos to load.
	 * @param filter     Filter applied to objects for being in this workspace.
	 * @param sort       Sort requested in this workspace.
	 * @param limit      Number of objects per page.
	 */
	public LyPageWorkSpace(LyDataSession session, LyClassInfo classInfo,
		LyInfoList<LyFieldInfo> fieldInfos, LyFilter filter, LySort sort, int limit)
	{
		this(session, classInfo, fieldInfos, filter, sort, limit, true, true);
	}

	/**
	 * Constructs a class workspace within a given session for a class info.
	 *
	 * @param session    Session in which the workspace is constructed.
	 * @param classInfo  Class info to which objects of this workspace belong.
	 * @param fieldInfos List of fieldinfos to load.
	 * @param filter     Filter applied to objects for being in this workspace.
	 * @param sort       Sort requested in this workspace.
	 * @param limit      Number of objects per page.
	 * @param dynamic    Specifies if this workspace is dynamic or not.
	 * @param sorted     Indicates if this workspace is sorted or not.
	 */
	public LyPageWorkSpace(LyDataSession session, LyClassInfo classInfo,
		LyInfoList<LyFieldInfo> fieldInfos, LyFilter filter, LySort sort, int limit,
		boolean dynamic, boolean sorted)
	{
		super(session, classInfo, fieldInfos, null, null, dynamic, sorted);

		_filter = filter;
		_pageSort = null;
		_limit = (limit <= 0 ? DEFAULT_LIMIT : limit);
		_temporaryLimit = -1;
		_min = null;
		_max = null;
		_page = 0;
		_pageFilter = filter;
		_totalCount = null;

		// Warning ! reset is called by setSort, don't call it again
		setSort(sort);

		if (dynamic)
		{
		    LyObjectSource    source = getObjectSource();

		    if (source != null)
    			// Add this as listener of classInfo objectSource
    			source.addListener(this);

			createFilterListeners(filter);
		}
	}

	/**
	 * Free this workspace.
	 */
	protected void free()
	{
		setMin(null);
		setMax(null);

		_pageFilter = null;

		super.free();
	}

	/**
	 * Gets the filter of this class workspace.
	 *
	 * @return Filter specified for this class workspace.
	 */
	public LyFilter getFilter()
	{
		return _pageFilter;
	}

	/**
	 * Sets the filter of this class workspace.
	 *
	 * @param filter Filter specified for this class workspace.
	 */
	public void setFilter(LyFilter filter)
	{
		setFilter(filter, true);
	}

	/**
	 * Sets the filter of this class workspace. The list is immediatly resorted and filtered if
	 * reset param is true.
	 *
	 * @param filter Filter specified for this class workspace.
	 * @param reset  indicates if we reset immediately or not
	 */
	public void setFilter(LyFilter filter, boolean reset)
	{
		_filter = getRealFilter(filter);
		_pageFilter = _filter;

		_pageSort = null;

		if (reset)
			reset();
	}

	/**
	 * Sets the current sort of this list. The list is immediatly resorted.
	 *
	 * @param sort New sort for this list.
	 */
	public void setSort(LySort sort)
	{
		setSort(sort, true);
	}

	/**
	 * Sets the current sort of this list. The list is immediatly resorted. The list is immediatly
	 * resorted and filtered if reset param is true.
	 *
	 * @param sort  New sort for this list.
	 * @param reset indicates if we reset immediately or not
	 */
	public void setSort(LySort sort, boolean reset)
	{
		if (reset)
			clear();

		_classSort = sort;

		if (_classInfo.isVirtual())
			_sort = getRealSort(sort);
		else
			_sort = sort;

		_sort = buildPageSort(true, false);

		_pageSort = null;
		_pageFilter = _filter;
		if (reset)
			reset();
	}

	/**
	 * Sets the at the same time the current sort and the filter of this list. The list is
	 * immediatly resorted and filtered if reset param is true.
	 *
	 * @param filter Filter specified for this class workspace.
	 * @param sort   New sort for this list.
	 * @param reset  indicates if we reset immediately or not
	 */
	public void setFilterAndSort(LyFilter filter, LySort sort, boolean reset)
	{
		setFilter(filter, false);
		setSort(sort, false);
		if (reset)
			reset();
	}

	/**
	 * Method for resetting this workspace. Objects contained in this workspace are removed and
	 * reloaded from the object source of the class info specified for this workspace.
	 */
	public synchronized void reset()
	{
		// Remove all objects.
		clear();

		// Reconstruct pageSort and pageFilter only if if they are null
		_next = true;
		if (_pageSort == null)
			_pageSort = buildPageSort(_next, false);

		if (_pageFilter == _filter)
			_pageFilter = buildPageFilter(_next, _pageSort, false);

		// Mark list as not sorted
		setSorted(false);

		// Load objects
		LyObjectSource	    objectSource = getObjectSource();

		if (objectSource == null) // May happen when in studio...
			return;

		Vector<LyFieldInfo> fields = null;

		if (_fieldInfos != null)
		{
			fields = _fieldInfos.getInfos();
			if (fields == null)
				fields = new Vector<LyFieldInfo>();
		}

		LyResponse    response;

		_page = 0;
		response = objectSource.getObjects(fields, _pageSort, _pageFilter, _limit, _local,
			new _LyResponseListener());

		if ((response != null)
		        && (response.getStatus() == LyResponse.Status.COMPLETED))
		    _paginatedByPeer = response.isPaginated();

		if ((response != null)
		        && (response.getStatus() == LyResponse.Status.COMPLETED))
		    _partialNotificationByPeer = response.isPartialNotification();

		// Reset total count
		_totalCount = null;
	} // end method reset

	/**
	 * Method for reloading this workspace. Objects contained in this workspace are removed and
	 * reloaded from the object source of the class info specified for this workspace.
	 */
	public synchronized void reload()
	{
		// Remove all objects.
		clear();

		// Mark list as not sorted
		setSorted(false);

		// Load objects
		LyObjectSource	    objectSource = getObjectSource();
		Vector<LyFieldInfo> fields = null;

		if (_fieldInfos != null)
		{
			fields = _fieldInfos.getInfos();
			if (fields == null)
				fields = new Vector<LyFieldInfo>();
		}

		objectSource.getObjects(fields, _pageSort, _pageFilter, _limit, _local,
			new _LyResponseListener());

		// Reset total count
		_totalCount = null;
	}

	/**
	 * Clears the content of the workspace.
	 */
	protected synchronized void clear()
	{
		// Reset min and max
		setMin(null);
		setMax(null);
		super.clear();
	}

	/**
	 * Load objects corresponding to the next or previous objects "page".
	 *
	 * @param next True if the next page is requested, false otherwise.
	 */
	public synchronized void loadPage(boolean next)
	{
		// Always reconstruct pageSort and pageFilter
		_pageSort = buildPageSort(next, false);
		_pageFilter = buildPageFilter(next, _pageSort, false);
		_next = next;

		if (!next && (_min != null))
		{
			// we do ask for a previous page
			// check if we have more previous than expected,
			// if yes, we are currently in page _page + n
			int expectedSize = _page * _limit;
	        int size = (_page == 1 && !_paginatedByPeer ? getPreviousPagesSize() : expectedSize);

			if (size > expectedSize)
				// if yes, increment page counter
				_page++;
			else if ((size <= _limit) || (_page <= 1))
			{
				// we have just/not enough object to build the 0 page,
				// rebuild completely the first page
				setSort(_classSort);
				return;
			}
		}

		// Remove all objects.
		clear();

		// Mark list as not sorted
		setSorted(false);

		// Load objects
		LyObjectSource	    objectSource = getObjectSource();
		Vector<LyFieldInfo> fields = null;

		if (_fieldInfos != null)
		{
			fields = _fieldInfos.getInfos();
			if (fields == null)
				fields = new Vector<LyFieldInfo>();
		}

		int      n = _limit;

		if (_paginatedByPeer)
		    // Encode in limit both number of objects and page number by assuming limit is always < MAX_LIMIT
    		n = (next ? (_page + 1) * MAX_LIMIT + _limit : (_page - 1) * MAX_LIMIT + _limit);

		objectSource.getObjects(fields, _pageSort, _pageFilter, n, _local, new _LyResponseListener());

		if (!checkIfEmpty())
		{
			if (next)
				_page++;
			else
				_page--;
		}
	} // end method loadPage

	/**
	 * Load objects corresponding to a given objects "page".
	 *
	 * @param page Index of requested page.
	 */
	public synchronized void loadPage(short page)
	{
		if (page == _page)
			return;

		int			offset = page - _page;
		boolean		next = (offset > 0);

		if (_paginatedByPeer)
		{
			// Always reconstruct pageSort and pageFilter
			_pageSort = buildPageSort(next, false);
			_pageFilter = buildPageFilter(next, _pageSort, false);
			_next = next;
			_page = page;

			// Remove all objects.
			clear();

			// Mark list as not sorted
			setSorted(false);

			// Load objects
			LyObjectSource	    objectSource = getObjectSource();
			Vector<LyFieldInfo> fields = null;

			if (_fieldInfos != null)
			{
				fields = _fieldInfos.getInfos();
				if (fields == null)
					fields = new Vector<LyFieldInfo>();
			}

			int      n = _limit;

			if (_paginatedByPeer)
			    // Encode in limit both number of objects and page number by assuming limit is always < MAX_LIMIT
	    		n = _page * MAX_LIMIT + _limit;

			objectSource.getObjects(fields, _pageSort, _pageFilter, n, _local, new _LyResponseListener());
			checkIfEmpty();
			return;
		}

		if (page == 0)
		{
			_pageSort = null;
			_pageFilter = _filter;
			reset();
		}
		else
		{
			int		totalCount = getTotalCount();
			int		pagesNumber = totalCount / _limit;
			int		lastPageCount = totalCount % _limit;
			int		realLimit = _limit;
			int		tmpLimit = realLimit * Math.abs(offset);
			boolean	startFromEnd = (page > (pagesNumber + _page) / 2);

			// Requested page is rather at the end : we reach the given page starting from the end
			if (startFromEnd)
				tmpLimit = lastPageCount + (pagesNumber - page) * realLimit;

			if (_partialNotificationByPeer)
			{
			    // Encode in limit both number of physical objects to load and number of objects to notify
				// by assuming total number of physical objects to select is always < MAX_PEER_LIMIT
				if (page == pagesNumber)
					tmpLimit = lastPageCount * MAX_PEER_LIMIT + tmpLimit;
				else
					tmpLimit = realLimit * MAX_PEER_LIMIT + tmpLimit;
			}

			_temporaryLimit = realLimit;
			_limit = tmpLimit;

			if (startFromEnd)
			{
				next = false;
				_min = null;
				_max = null;
				_pageSort = buildPageSort(next, false);
				_pageFilter = buildPageFilter(next, _pageSort, false);
			}

			loadPage(next);
			_page = page;
			_limit = realLimit;
			_temporaryLimit = -1;
		}
	}

    /**
     * Method for reloading current page.
     */
    protected synchronized void reloadCurrentPage()
    {
        short page = _page;

        _page = -1;
        loadPage(page);
    }

	/**
	 * Checks if workspace is empty and reset it if needed.

	 * @return true if workspace is empty, false otherwise
	 */
	protected boolean checkIfEmpty()
	{
		if (getSize() == 0)
		{
			if (_paginatedByPeer)
			{
				reset();
			}
			else if (_page > 0)
			{

				_page = 0;
				loadPage(true);
			}
			else
			{
				_pageSort = null;
				_pageFilter = _filter;
				reset();
			}

			return true;
		}

		return false;
	}

	/**
	 * Load bound object.
	 *
	 * @param min True if the min is requested, false otherwise.
	 */
	protected synchronized void loadBound(boolean min)
	{
		// Always reconstruct pageSort and pageFilter
		LySort			    pageSort = buildPageSort(!min, true);

        if (FORCE_RELOAD_PAGE_FOR_DATES
        	&& (pageSort != null))
        {
        	int				i, n = pageSort.getCriterionCount();
        	LySortCriterion	criterion;
        	LyFieldInfo		fieldInfo;

        	for (i = 0; i < n; i++)
        	{
        		criterion = pageSort.getCriterion(i);
        		fieldInfo = criterion.getFieldInfo();
        		if (fieldInfo instanceof LyDateFieldInfo)
        		{
                    sendReloadEvent(true);
                    reloadCurrentPage();
                    sendReloadEvent(false);
        			return;
        		}
        	}
        }

		LyFilter		    filter = buildPageFilter(!min, pageSort, true);

		// Load objects
		LyObjectSource	    objectSource = getObjectSource();
		Vector<LyFieldInfo> fields = null;

		if (_fieldInfos != null)
		{
			fields = _fieldInfos.getInfos();
			if (fields == null)
				fields = new Vector<LyFieldInfo>();
		}

		objectSource.getObjects(fields, pageSort, filter, 1, _local, new _LyResponseListener());
	}

	/**
	 * Internal method to build sort used for loading objects on object source.
	 *
	 * @param next True if the next page is requested, false otherwise.
	 * @param loadBound Set when reloading boundary object.
	 * @return Sort for loading another page.
	 */
	LySort buildPageSort(boolean next, boolean loadBound)
	{
		LySort	    sort = new LySort();
		LyFieldInfo fieldInfo;
		int		    i, n;

		if (_sort != null)
		{
			LySortCriterion criterion;

			for (i = 0, n = _sort.getCriterionCount(); i < n; i++)
			{
				criterion = _sort.getCriterion(i);
				if (next
				        || (!loadBound && _paginatedByPeer))
					sort.addCriterion(criterion);
				else
				{
					fieldInfo = criterion.getFieldInfo();
					sort.addCriterion(fieldInfo, !criterion.isIncreasing());
				}
			}
		}

		Vector<LyFieldInfo> idFields = _classInfo.getIdFields();

		if (idFields == null)
		{
			if (_session != null)
				_session.getEnvironment().logWarning("Trying to use a Lypageworkspace "
					+ "on a class (probably abstract) with no id fields");
			return sort;
		}

		for (i = 0, n = idFields.size(); i < n; i++)
		{
			fieldInfo = idFields.elementAt(i);
			if ((_sort == null) || (_sort.getCriterion(fieldInfo) == null))
				sort.addCriterion(fieldInfo, next);
		}

		return sort;
	} // end method buildPageSort

	/**
	 * Internal method to build filter used for loading objects on object source.
	 *
	 * @param next      True if the next page is requested, false otherwise.
	 * @param pageSort  Page sort.
	 * @param loadBound Set when reloading boundary object.
	 * @return Filter for loading another page. This filter is a merge of the filter requiered for
	 *         loading another page on the object source and the filter specified at start on this
	 *         workspace.
	 */
	LyFilter buildPageFilter(boolean next, LySort pageSort, boolean loadBound)
	{
        if ((_pageSort == null)
                || (!loadBound && _paginatedByPeer))
            return _filter;

		LyObject ref = (next ? _max : _min);

		if (ref == null)
			return _filter;

		// Create filter and sub-filters
		LySortCriterion criterion;
		LyFieldInfo     fieldInfo;
		int			    i, n = pageSort.getCriterionCount();
		LyFilter	    pageFilter = new LyFilter(LyFilter.Type.OR);
		LyFilter	    subFilter = null;
		LyFilter	    temp = null;
		LyFilterElement exp;
		int			    j;

		for (i = n - 1; i >= 0; i--)
		{
			if (n > 1)
			{
				subFilter = new LyFilter(LyFilter.Type.AND);

				for (j = 0; j < i; j++)
				{
					criterion = pageSort.getCriterion(j);
					fieldInfo = criterion.getFieldInfo();
					exp = null;

		            if (fieldInfo instanceof LyRelationFieldInfo)
		                exp = buildNameRelationFilterElement((LyRelationFieldInfo)fieldInfo, ref, null);

		            if (exp == null)
    					exp = new LyFilterExpression(fieldInfo, false, LyFilterExpression.Condition.EQUALS,
    							ref.getOldValue(fieldInfo));

					subFilter.addElement(exp);
				}
			}

			criterion = pageSort.getCriterion(i);
			fieldInfo = criterion.getFieldInfo();
			temp = new LyFilter(LyFilter.Type.OR);

			// We may add a relation condition by name
			exp = null;
			if (fieldInfo instanceof LyRelationFieldInfo)
				exp = buildNameRelationFilterElement((LyRelationFieldInfo)fieldInfo, ref,
						criterion.isIncreasing());

			if (exp == null)
				exp = new LyFilterExpression(fieldInfo, false,
						(criterion.isIncreasing() ? LyFilterExpression.Condition.GREATER_THAN
												  : LyFilterExpression.Condition.LOWER_THAN),
						ref.getOldValue(fieldInfo));

			temp.addElement(exp);
			if (!criterion.isIncreasing()
				&& (ref.getOldValue(fieldInfo) != null))
			{
				exp = new LyFilterExpression(fieldInfo, false, LyFilterExpression.Condition.EQUALS,
						null);

				temp.addElement(exp);
			}
			else if (criterion.isIncreasing()
				&& (ref.getOldValue(fieldInfo) == null))
			{
				exp = new LyFilterExpression(fieldInfo, true, LyFilterExpression.Condition.EQUALS, null);
				temp.addElement(exp);
			}

			if (n > 1)
			{
				subFilter.addElement(temp);
				pageFilter.addElement(subFilter);
			}
			else
				pageFilter.addElement(temp);

/* Do not remove : to use if null values are greater than other values.
                        if (criterion.isIncreasing() && (ref.getOldValue(fieldInfo) != null))
                        {
                                exp = new LyFilterExpression(fieldInfo, false,
                                        LyFilterExpression.Condition.EQUALS, null);
                                pageFilter.addElement(exp);
                        }
                        else if (!criterion.isIncreasing() && (ref.getOldValue(fieldInfo) == null))
                        {
                                exp = new LyFilterExpression(fieldInfo, true,
                                        LyFilterExpression.Condition.EQUALS, null);
                                pageFilter.addElement(exp);
                        }
*/
/*if (!criterion.isIncreasing() && (ref.getOldValue(fieldInfo) != null))
                        {
                                exp = new LyFilterExpression(fieldInfo, false,
                                        LyFilterExpression.Condition.EQUALS, null);

                                pageFilter.addElement(exp);
                        }
                        else if (criterion.isIncreasing() && (ref.getOldValue(fieldInfo) == null))
                        {
                                exp = new LyFilterExpression(fieldInfo, true,
                                        LyFilterExpression.Condition.EQUALS, null);
                                pageFilter.addElement(exp);
                        }
*/
			//
			if (i > 1)
			{
				criterion = pageSort.getCriterion(i - 1);
				fieldInfo = criterion.getFieldInfo();
				if (ref.getOldValue(fieldInfo) == null)
					i--;
			}

		}

		LyFilter filter = null;

		if ((_filter != null) && (pageFilter != null))
		{
			filter = new LyFilter(LyFilter.Type.AND);
			filter.addElement(_filter);
			filter.addElement(pageFilter);
		}
		else if (_filter != null)
			filter = _filter;
		else if (pageFilter != null)
			filter = pageFilter;

		return filter;
	} // end method buildPageFilter

	/**
	 * Internal method to build filter used for loading objects on object source.
	 *
	 * @param  next     True if the next page is requested, false otherwise.
	 * @param  pageSort Page sort.
	 * @param  ref      object used as the limit.
	 * @param  pageSort Page sort.
	 * @return Filters for loading another page. These filters are a merge of the filters required.
	 */
	LyInfoList<LyFilter> buildSizeFilters(LyObject ref, LySort pageSort)
	{
		LyInfoList<LyFilter> result = new LyInfoList<LyFilter>();

		if (_pageSort == null)
		{
			result.addInfo(_filter);
			return result;
		}

		if (ref == null)
		{
			result.addInfo(_filter);
			return result;
		}

		// Create filter and sub-filters
		LySortCriterion		    criterion;
		LyFieldInfo			    fieldInfo;
		int					    i, n = pageSort.getCriterionCount();
		LyFilter			    pageFilter = new LyFilter(LyFilter.Type.OR);
		LyFilter			    subFilter = null;
		LyFilter			    temp = null;
		LyFilterElement		    exp;
		int					    j;
		Vector<LyFilterElement> relationFilters = null;

		for (i = n - 1; i >= 0; i--)
		{
			if (n > 1)
			{
				subFilter = new LyFilter(LyFilter.Type.AND);

				for (j = 0; j < i; j++)
				{
					criterion = pageSort.getCriterion(j);
					fieldInfo = criterion.getFieldInfo();
                    exp = null;

                    if (fieldInfo instanceof LyRelationFieldInfo)
                        exp = buildNameRelationFilterElement((LyRelationFieldInfo)fieldInfo, ref, null);

                    if (exp == null)
                        exp = new LyFilterExpression(fieldInfo, false, LyFilterExpression.Condition.EQUALS,
							ref.getOldValue(fieldInfo));

					subFilter.addElement(exp);
				}
			}

			criterion = pageSort.getCriterion(i);
			fieldInfo = criterion.getFieldInfo();

			// We may add a relation condition by name
			exp = null;
			if (fieldInfo instanceof LyRelationFieldInfo)
				exp = buildNameRelationFilterElement((LyRelationFieldInfo)fieldInfo, ref,
						criterion.isIncreasing());

			if (exp == null)
			{
				temp = new LyFilter(LyFilter.Type.OR);

				// Do not build a filter for values lower or greater than null
				if (ref.getOldValue(fieldInfo) == null)
					continue;

				exp = new LyFilterExpression(fieldInfo, false,
						(criterion.isIncreasing() ? LyFilterExpression.Condition.GREATER_THAN
												  : LyFilterExpression.Condition.LOWER_THAN),
						ref.getOldValue(fieldInfo));
				temp.addElement(exp);
			}
			else
			{
				if (relationFilters == null)
					relationFilters = new Vector<LyFilterElement>();
				relationFilters.add(exp);
			}

			if (!criterion.isIncreasing() && (ref.getOldValue(fieldInfo) != null))
			{
				if (temp == null)
					temp = new LyFilter(LyFilter.Type.OR);
				exp = new LyFilterExpression(fieldInfo, false, LyFilterExpression.Condition.EQUALS,
						null);
				temp.addElement(exp);
			}
			else if (criterion.isIncreasing() && (ref.getOldValue(fieldInfo) == null))
			{
				if (temp == null)
					temp = new LyFilter(LyFilter.Type.OR);
				exp = new LyFilterExpression(fieldInfo, true, LyFilterExpression.Condition.EQUALS, null);
				temp.addElement(exp);
			}

			if (relationFilters != null)
			{
				// We have to set the type of the filter to 'AND' due to relation filter
				LyFilter supFilter = new LyFilter(LyFilter.Type.AND);

				for (j = 0; j < relationFilters.size(); j++)
					supFilter.addElement(relationFilters.get(j));

				temp = supFilter;
				relationFilters = null;
			}

			if (n > 1)
			{
				subFilter.addElement(temp);
				pageFilter.addElement(subFilter);
			}
			else
				pageFilter.addElement(temp);

			/* Do not remove : to use if null values are greater than other values.
			                if (criterion.isIncreasing() && (ref.getOldValue(fieldInfo) != null))
			                {
			                        exp = new LyFilterExpression(fieldInfo, false,
			                                LyFilterExpression.Condition.EQUALS, null);
			                        pageFilter.addElement(exp);
			                }
			                else if (!criterion.isIncreasing() && (ref.getOldValue(fieldInfo) == null))
			                {
			                        exp = new LyFilterExpression(fieldInfo, true,
			                                LyFilterExpression.Condition.EQUALS, null);
			                        pageFilter.addElement(exp);
			                }
			*/
			/*                      if (!criterion.isIncreasing()  && (ref.getOldValue(fieldInfo) != null))
			                {
			                        exp = new LyFilterExpression(fieldInfo, false,
			                                LyFilterExpression.Condition.EQUALS, null);
			                        pageFilter.addElement(exp);
			                }
			                else if (criterion.isIncreasing() && (ref.getOldValue(fieldInfo) == null))
			                {
			                        exp = new LyFilterExpression(fieldInfo, true,
			                                LyFilterExpression.Condition.EQUALS, null);
			                        pageFilter.addElement(exp);
			                }
			*/
			if (i > 1)
			{
				criterion = pageSort.getCriterion(i - 1);
				fieldInfo = criterion.getFieldInfo();
				if (ref.getOldValue(fieldInfo) == null)
					i--;
			}

			LyFilter filter = null;

			if ((_filter != null) && (pageFilter != null))
			{
				filter = new LyFilter(LyFilter.Type.AND);
				filter.addElement(_filter);
				filter.addElement(pageFilter);
			}
			else if (_filter != null)
				filter = _filter;
			else if (pageFilter != null)
				filter = pageFilter;

			result.addInfo(filter.cloneFilter());
			pageFilter = new LyFilter(LyFilter.Type.OR);
		}

		return result;
	} // end method buildSizeFilters

	/**
	 * Builds a filter element for the 'by name' filtering on relations.
	 *
	 * @param  relation   a logical relation field.
	 * @param  reference  the current reference object in this workspace
	 * @param  increasing tells if the related sort criterion is increasing.
	 * @return a filter element or null.
	 */
	protected LyFilterElement buildNameRelationFilterElement(LyRelationFieldInfo relation,
		LyObject reference, Boolean increasing)
	{
		if (relation.isMultiple() || (reference == null))
			return null;

		LyObjectId oid = (LyObjectId)reference.getValue(relation);
		LyObject   o = (oid == null) ? null : oid.getLogicalObject();

		if (o != null)
		{
			LyFilter    subSubFilter = new LyFilter();
			LyClassInfo refClass = o.getClassInfo();

			// if ID fields equals NAME fields : this is useless
			// it is also useless for local classes
			if (refClass.isNameDifferentOfId() && !refClass.isLocal())
			{
				LyInfoList<LyFieldInfo> nameFields = refClass.getFieldInfos(LyFieldInfo.MARK_NAME);
				LyFieldInfo			    nameField = nameFields.getInfo(0);

				subSubFilter.addElement(new LyFilterExpression(nameField, false,
						(increasing == null ? LyFilterExpression.Condition.EQUALS :
						    (increasing ? LyFilterExpression.Condition.GREATER_THAN
						        : LyFilterExpression.Condition.LOWER_THAN)),
						o.getValue(nameField)));

				return new LyFilterRelation(relation, subSubFilter);
			}
		}

		return null;
	} // end method buildNameRelationFilterElement

	/**
	 * Get the number of object before the current page.
	 *
	 * @return the number of object before the current page
	 */
	protected int getPreviousPagesSize()
	{
		LySort				 pageSort = buildPageSort(false, false);
		LyInfoList<LyFilter> filters = buildSizeFilters(_min, pageSort);
		LyObjectSource		 objectSource = getObjectSource();
		int					 i, n = filters.getSize(), count = 0;
		LyFilter			 filter;

		// Optimization : use several filters instead of a single one.
		// Resulting SQL requests may be more efficient than only one complex request.
		for (i = 0; i < n; i++)
		{
			filter = filters.getInfo(i);
			count += objectSource.getSize(filter);
		}

		return count;
	}

	/**
	 * Returns a string representation of this object.
	 *
	 * @return This workspace returned as a string.
	 */
	public String toString()
	{
		return super.toString() + ", limit=" + _limit;
	}

	/**
	 * Gets the current page number.
	 * @return the page number
	 */
	public short getPage()
	{
		return _page;
	}
    
    /**
     * Indicates whether peer can paginate results.
     */
    public boolean isPaginatedByPeer()
    {
    	return _paginatedByPeer;
    }

	/**
	 * Gets page count.
	 * @return the page count
	 */
	public short getPageCount()
	{
		return (short)((getTotalCount() + _limit - 1) / _limit);
	}

	/**
	 * @return The total number of objects for the class of this workspace including class
	 *         filtering.
	 */
	public int getTotalCount()
	{
		// compute total count if it needs an update
		if (_totalCount == null)
		{
			LyObjectSource objectSource = getObjectSource();

			if (objectSource != null)
			    _totalCount = objectSource.getSize(_filter);
		}

		// Check local number of objects to avoid having a total lower than the number
		// of objects displayed in the list
		int   localCount = getSize();

		if (_totalCount != null)
		    return Math.max(_totalCount, localCount);
		else
		    return localCount;
	}

	/* Implementation of LyDataSourceListener */

    /**
     * Method for receiving an object creation. This method is overridden to check if the object is
     * greater than min and lower than max, regarding current sort.
     *
     * @param  event Event notifying the creation of an object.
     * @return True indicating if the event was used, false otherwise.
     */
    @Override public boolean objectCreated(LyDataSourceEvent event)
    {
        LyObject object = (LyObject) event.getObject();

        // We must handle notification (for notif isLoading = true and isNotified=true)
        if (event.isLoading()
        		&& !event.isNotified()) // Loading
            return true;

        synchronized (this)
        {
            if (contains(object))
                // The object is already in our page, ignore event
                return true;

            if ((_filter != null)
            		&& !matchFilter(_filter, object))
                // The object is not in our list, ignore event
                return false;

            // Increment total count
            if (_totalCount != null)
                _totalCount = new Integer(_totalCount.intValue() + 1);

            if ((_max == null)
            	|| (object.compare(_max, _sort, false) > 0))
            {
                // Object is greater than max
                if (getSize() >= _limit) // and the page is full
                {
                    // Post a fake event to force update of views if needed (maybe we have one more page)
                    LyWorkSpaceEvent wsEvent = new LyWorkSpaceEvent(this, LyWorkSpaceEvent.Type.ADD, object, -1);

                    wsEvent.setEventInfo(event.getEventInfo());
                    notify(wsEvent);

                    _totalCount = null;
                    return false;
                }

                // Add object to complete page
                super.objectCreated(event);

                // Created object becomes max of the list
                setMax(object);

                if (_min == null)
                    // Object becomes min if list was empty
                    setMin(object);

                return true;
            }

            // Object is lower than max
            if (getSize() >= _limit)
                // Remove max to keep the same number of objects
                removeObject(_max);

            if (object.compare(_min, _sort, false) > 0)
                // Object is greater than min, add it to the page
                super.objectCreated(event);
            else
            {
                // Object is lower than min
                if (_page > 0) // and we're not on first page
                {
                    // Object min - 1 will become min
                    loadBound(true);
                    return true;
                }

                // First page, add object to our page
                super.objectCreated(event);

                // Created object becomes min of the list
                setMin(object);
            }

            // Last object becomes max
            setMax(getObject(getSize() - 1));
        }

        return true;
    } // end method objectCreated

    private void sendReloadEvent(boolean start)
    {
    	LyWorkSpaceEvent.Type type = (start ? LyWorkSpaceEvent.Type.ADD : LyWorkSpaceEvent.Type.REMOVE);
    	LyWorkSpaceEvent wsEvent = new LyWorkSpaceEvent(this, type);

        wsEvent.setEventInfo(null);
        notify(wsEvent);
    }

    /**
     * Method for receiving an object modification. This method is overridden to check if the object
     * is greater than min and lower than max, regarding current sort.
     *
     * @param  event Event notifying the modification of an object.
     * @return True indicating if the event was used, false otherwise.
     */
    @Override public synchronized boolean objectModified(LyDataSourceEvent event)
    {
        LyValueSet oldValues = event.getValues();

        if (oldValues != null)
        {
            // First check if changes concern filter and sort
            LyInfoList<LyFieldInfo> fieldInfos = null;

            if (_pageFilter != null)
                fieldInfos = _pageFilter.getFieldInfos();
            else if (_filter != null)
                fieldInfos = _filter.getFieldInfos();

            if (_sort != null)
            {
	            if (fieldInfos == null)
	                fieldInfos = _sort.getFieldInfos();
	            else
	                fieldInfos.addList(_sort.getFieldInfos());
            }

            if (fieldInfos != null)
            {
                // Check if at least on field is contained in values changes
                int     i, n = fieldInfos.getSize();
                boolean intersect = false;

                for (i = 0; (i < n) && !intersect; i++)
                    intersect = oldValues.contains(fieldInfos.getInfo(i));

                if (!intersect)
                    // Just handle event to notify listeners
                    return super.objectModified(event);
            }
        }

        LyObject object = (LyObject) event.getObject();

        if (indexOf(object) >= 0) // The object is in the page
        {
            if ((_filter != null) // But not in the list anymore
                    && !matchFilter(_filter, object))
                // Do as if the object was deleted, total count will be updated
                return objectDeleted(event);

            // Object is still in the list, total count is unchanged

            if ((object == _min)
            	|| (object == _max))
            {
                // We're not able to determine what to do, force reloading of page
                sendReloadEvent(true);
                reloadCurrentPage();
                sendReloadEvent(false);
                return true;
            }

            if (object.compare(_max, _sort, false) > 0) // Still in list but greater than max
            {
                // If object must be moved after the last object, remove it and load the
                // next object of the global list.
                removeObject(object);
                loadBound(false);
                return true;
            }

            if (object.compare(_min, _sort, false) < 0) // Still in list but lower than min
            {
                // If object must be moved before the first object, remove it and load the
                // previous object of the global list.
                removeObject(object);
                loadBound(true);
                return true;
            }

            // Handle modification event
            if (!super.objectModified(event))
                return false;

            // Update min and max
            setMin(getObject(0));
            setMax(getObject(getSize() - 1));
            return true;
        }
        else
        {
            // The object is not in the page
            boolean matchedClassFilterBefore = true;

            if (_filter != null)
            {
                // Check if it was in the list before change
                LyValueSet              values = new LyValueSet();
                LyInfoList<LyFieldInfo> fields = _filter.getFieldInfos(false);
                int                     i, n = fields.getSize();
                LyFieldInfo             fieldInfo;

                for (i = 0; i < n; i++)
                {
                    fieldInfo = fields.getInfo(i);

                    if ((oldValues != null)
                    		&& oldValues.contains(fieldInfo))
                        values.addValue(oldValues.getFieldValue(fieldInfo));
                    else
                        values.addValue(object.getFieldValue(fieldInfo));
                }

                matchedClassFilterBefore = matchFilter(_filter, values);
            }

            if ((_filter != null)
            	&& !matchFilter(_filter, object))
            {
                // The object is not in the list.
                if (!matchedClassFilterBefore) // And it was not before, ignore event
                    return false;

                // But it was in the list before, do as if the object was deleted
                return objectDeleted(event);
            }

            // The object is in the list.
            if (!matchedClassFilterBefore
            		|| (_min == null))
                // It wasn't before, do as if it was created
                return objectCreated(event);

            // And it was before but it may move, first find where it was in list
            boolean         wasBeforeMin = false;
            LySortCriterion criterion;
            LyFieldInfo     fieldInfo;
            int             i, n = (_sort == null ? 0 : _sort.getCriterionCount());
            int             result = 0;
            LyValue         oldFieldValue;
            Object          oldValue, minValue;

            for (i = 0; i < n; i++)
            {
                criterion = _sort.getCriterion(i);
                fieldInfo = criterion.getFieldInfo();
                oldFieldValue = (oldValues == null ? null : oldValues.getFieldValue(fieldInfo));

                if (oldFieldValue != null)
                    oldValue = oldFieldValue.getValue();
                else
                    oldValue = object.getValue(fieldInfo);

                minValue = _min.getValue(fieldInfo);

                result = criterion.compare(oldValue, minValue);

                if (result != 0)
                {
                    wasBeforeMin = (result < 0);
                    break;
                }
            }

            if (result == 0)
                // No change on sort criteria
                return false;

            if (object.compare(_min, _sort, false) < 0) // Still in list but lower than min
            {
                if (wasBeforeMin)
                    return false;

                // Remove max
                removeObject(_max);
                // Reload min
                loadBound(true);
            }
            else if (object.compare(_max, _sort, false) > 0) // Still in list but greater than max
            {
                if (!wasBeforeMin) // Then after max because not in page
                    return false;

                // Remove min
                removeObject(_min);
                // Reload max
                loadBound(false);
            }
            else
            {
                if (wasBeforeMin) // Object now in page
                {
                    // Remove min
                    removeObject(_min);
                    // Add object to page
                    addObject(object);
                }
                else // Object now in page
                {
                    // Remove max
                    removeObject(_max);
                    // Add object to page
                    addObject(object);
                }

                // Update min and max
                setMin(getObject(0));
                setMax(getObject(getSize() - 1));
            }

            return true;
        }
    } // end method objectModified

	/**
	 * Test if the specified event is used to add or remove an object from this workspace.
	 *
	 * @param  event Event to test
	 * @return true if the event adds or removes the object'event to or from the workspace, false
	 *         otherwise.
	 */
	protected boolean useEvent(LyDataSourceEvent event)
	{
		return false;
	}
	
    /**
     * Method for receiving an object deletion. This method is overridden to update min and max is if
     * required.
     *
     * @param  event Event notifying the deletion of an object.
     * @return True indicating if the event was used, false otherwise.
     */
    @Override public synchronized boolean objectDeleted(LyDataSourceEvent event)
    {
        LyObject object = (LyObject) event.getObject();

        if ((event.getType() == LyDataSourceEvent.TYPE_DELETE) // May be called by objectModified
        		&& (_filter != null)
        		&& !matchFilter(_filter, object))
            // Object was not in the list, ignore deletion event.
            return false;

        // Decrement total count
        if (_totalCount != null)
            _totalCount = new Integer(_totalCount.intValue() - 1);

        if (contains(object))
        {
            // Object is the page, handle event to remove it
            if (!super.objectDeleted(event))
                // Normally could not occur
                return false;

            if ((getSize() + 1) >= _limit)
                // Object max + 1 will become max
                loadBound(false);
            else if ((_page > 0)
            		&& (getSize() == 0))
                // Page became empty and is not the first page, reload previous page
                loadPage(!_next);
        }
        else if ((_max != null)
        	&& (object.compare(_max, _sort, false) > 0))
        {
            // Object was not in the page, and was greater than max
            if (getSize() >= _limit) // And page was full
            {
                // Post a fake event to force update of views if needed (maybe we have one page less)
                LyWorkSpaceEvent wsEvent = new LyWorkSpaceEvent(this, LyWorkSpaceEvent.Type.REMOVE, object, -1);

                wsEvent.setEventInfo(event.getEventInfo());
                notify(wsEvent);
            }
        }
        else if ((_min != null)
        	&& (object.compare(_min, _sort, false) < 0))
        {
            // Object was not in the page, and is lower than min
            if (getSize() == 1)
            {
                if (_page > 0)
                    // Object was alone in its page, reload previous page
                    loadPage(!_next);

                return true;
            }

            // Remove min
            removeObject(_min);

            // First object becomes min
            setMin(getObject(0));

            if ((getSize() + 1) >= _limit)
                // Object max + 1 will become max
                loadBound(false);
        }

        return true;
    } // end method objectDeleted

    /**
     * Called when grouped events are sent.
     *
     * @param  events Sorted list of events.
     * @return True if the events are used by the listener, false otherwise.
     */
    @Override public boolean dataSourceEventsGrouped(ArrayList<LyDataSourceEvent> events)
    {
        if (events.size() > 1)
        {
            // Reload current page
            _totalCount = null;
            reloadCurrentPage();

            // Post a fake event to force update of views if needed (number of pages may have changed)
            LyDataSourceEvent event = events.get(0);
            LyObject          object = (LyObject) event.getObject();
            LyWorkSpaceEvent  wsEvent = new LyWorkSpaceEvent(this, LyWorkSpaceEvent.Type.ADD, object, -1);

            wsEvent.setEventInfo(event.getEventInfo());
            notify(wsEvent);

            return true;
        }
        else
            return super.dataSourceEventsGrouped(events);
    }

	/* Implementation of LyResponseListener */

	/**
	 * Internal class that implements This allows to load objects without sorting them and sort them
	 * at the end of the loading if requiered.
	 */
	class _LyResponseListener
		implements LyResponseListener
	{
		LySimpleObjectList _list;

		/**
		 * Method for receiving a new object in a response.
		 *
		 * @param response  Response for which a new object is received.
		 * @param object    The new object returned in the response.
		 * @param values List of values that were changed for the object.
		 */
		public void newResponse(LyResponse response, Object object, LyValueSet values)
		{
			// Add the loaded object in our workspace
			if (_next
			        || _paginatedByPeer)
				addObject((LyObject)object);
			else
			{
				// Responses are in reverse order for the workspace so we stock them
				// to add them in right order in endResponse method.
				if (_list == null)
				{
					_list = new LySimpleObjectList();
					_list.setSorted(false);
				}
				_list.addObject((LyObject)object);
			}
		}

		/**
		 * Method for ending a response.
		 *
		 * @param response Response which end is notified.
		 */
		public void endResponse(LyResponse response)
		{
			int n;

			if (!_next)
			{
				// The responses are in the right order for the request but in
				// reverse order for the workspace. So we have stocked the responses
				// to add them in the right order for the workspace.
				LyObject object;

				n = ((_list == null) ? 0 : _list.getSize());
				for (int i = n - 1; i >= 0; i--)
				{
					object = _list.getObject(i);
					addObject(object);
				}
				_list = null;
			}
			// Get the min and max objects.
			n = getSize();

			// Sort results if needed or simply restore _sorted value
			if (!response.isSorted())
				setSorted(true);
			else
				_sorted = true;

			if (n == 0)
			{
				setMin(null);
				setMax(null);
			}
			else
			{
				if ((!_partialNotificationByPeer
						&& (_temporaryLimit > 0))
					|| ((!response.isCounted())
						&& (_limit >= 1)
						&& (n > _limit)))
				{
					boolean	isTemporaryLimit = (_temporaryLimit > 0);
					int		limit = (isTemporaryLimit ? _temporaryLimit : _limit);
					int		indexToRemove = (_next ? limit : 0);

					if (isTemporaryLimit)
						indexToRemove = (_next ? 0 : limit);

					// Specific for domino connector (maybe other later): not sorted and not counted : sort is done above, simulate count now
					while (getSize() > limit)
					{
						LyObject obj = getObject(indexToRemove);

						removeObject(obj);
					}
					n = getSize();
				}
				setMin(getObject(0));
				setMax(getObject(n - 1));
			}
		} // end method endResponse
	} // end class _LyResponseListener

	/**
	 * Sets the _min objects and locks it.
	 *
	 * @param obj new min object
	 */
	protected void setMin(LyObject obj)
	{
		// If same, do nothing
		if (_min == obj)
			return;

		_min = obj;
	}

	/**
	 * Sets the _max objects and locks it.
	 *
	 * @param obj new max object
	 */
	protected void setMax(LyObject obj)
	{
		// If same, do nothing
		if (_max == obj)
			return;

		_max = obj;
	}

	/**
	 * @see leon.data.LyObjectList#getSort()
	 */
	public LySort getSort()
	{
		return _classSort;
	}
	
	/**
	 * Returns an implementation of the Iterator interface for this LyObjectList.
	 */
	public Iterator<LyObject> iterator()
	{
		return new _LyPageWorkSpaceIterator(this);
	}
	
	/**
	 * Internal class used for iteration over the list.
	 */
	final class _LyPageWorkSpaceIterator
		implements Iterator<LyObject>
	{
		int			 	_currentIndex;
		LyPageWorkSpace _list;

		_LyPageWorkSpaceIterator(LyObjectList list)
		{
			_list = new LyPageWorkSpace(_session, _classInfo,
				_fieldInfos, _filter, _sort, ITERATOR_LIMIT, false, _sorted);
		}

		public boolean hasNext()
		{
			if (_currentIndex >= getTotalCount())
				return false;
			return true;
		}

		public void remove()
		{
			throw new IllegalStateException("Unsupported method");
		}

		public LyObject next()
		{
			if (_currentIndex > ((_list._page + 1) * _list._limit - 1))
				_list.loadPage(true);
			
			try
			{
				LyObject object = _list.getObject(_currentIndex % _list._limit);

				_currentIndex++;
				return object;
			}
			catch (IndexOutOfBoundsException e)
			{
				throw new NoSuchElementException();
			}
		}
	} // end class _LyObjectListIterator
} // end class LyPageWorkSpace
