/*
 * LyRelationValue.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import leon.info.*;


/**
 * This class offers methods for constructing, accessing and modifying values corresponding to a
 * LyRelationFieldInfo.
 *
 * @see LyRelationFieldInfo
 * @see LyObject
 */
public final class LyRelationValue extends LyFieldInfoValue
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyRelationValue.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Constructs an empty relation value for the given relation field.
	 *
	 * @param relationFieldInfo Relation fieldInfo for which a value is created.
	 */
	public LyRelationValue(LyRelationFieldInfo relationFieldInfo)
	{
		super(relationFieldInfo);
	}

	/**
	 * Constructs a relation value for the given relation field from a string.
	 *
	 * @param relationFieldInfo Relation fieldInfo for which a value is created.
	 * @param value             String value (as expected by parse method).
	 * @see   LyRelationFieldInfo#parse
	 */
	public LyRelationValue(LyRelationFieldInfo relationFieldInfo, String value)
	{
		super(relationFieldInfo, value);
	}

	/**
	 * Constructs a relation value for the given relation field from a LyObjectId.
	 *
	 * @param     relationFieldInfo Relation fieldInfo for which a value is created.
	 * @param     value             Value given as a LyObjectId.
	 * @exception IllegalArgumentException If the given object id doesn't concern a class to which
	 *                                     the relation field is related with.
	 */
	public LyRelationValue(LyRelationFieldInfo relationFieldInfo, LyObjectId value)
	{
		super(relationFieldInfo);

		if (value == null)
			return;

		LyClassInfo objectClass = (LyClassInfo)value.getObjectClass();

		if ((!relationFieldInfo.relatedWith(objectClass))
				&& ((!objectClass.isTemporary())
					|| (!relationFieldInfo.relatedWith((LyClassInfo)objectClass.getTemplate()))))
			throw new IllegalArgumentException("BF AppEngine : class of object id " + objectClass
				+ " is not supported by relation field " + relationFieldInfo);

		if (relationFieldInfo.isMultiple())
		{
			LyObjectId[] values = new LyObjectId[1];

			values[0] = value;
			setValue(values);
		}
		else
			setValue(value);
	}

	/**
	 * Constructs a relation value for the given relation field from an array of LyObjectId.
	 *
	 * @param     relationFieldInfo Relation fieldInfo for which a value is created.
	 * @param     value             Value given as an array of LyObjectId.
	 * @exception IllegalArgumentException If the relation field is not multiple and the array
	 *                                     length is greater than 1.
	 * @exception IllegalArgumentException If one of the given object ids doesn't concern a class to
	 *                                     which the relation field is related with.
	 */
	public LyRelationValue(LyRelationFieldInfo relationFieldInfo, LyObjectId[] value)
	{
		super(relationFieldInfo);

		if (value == null)
			return;

		int n = value.length;

		if (n == 0)
			return;

		if (!relationFieldInfo.isMultiple() && (n > 1))
			throw new IllegalArgumentException("BF AppEngine : fieldInfo "
				+ relationFieldInfo + " is not multiple and can't accept more than one value");

		// Check classes of object ids
		LyClassInfo objectClass;
		int		    i, nNull = 0;

		for (i = 0; i < n; i++)
		{
			if (value[i] == null)
				nNull++;
			else
			{
				objectClass = (LyClassInfo)value[i].getObjectClass();
				if (!relationFieldInfo.relatedWith(objectClass))
					throw new IllegalArgumentException("BF AppEngine : class of object id "
						+ objectClass
						+ " is not supported by relation field " + relationFieldInfo);
			}
		}

		if (nNull == n)
			return;
		else if (nNull > 0)
		{
			if (relationFieldInfo.isMultiple())
			{
				LyObjectId[] newValue = new LyObjectId[n - nNull];
				int			 nv = 0;

				for (i = 0; i < n; i++)
					if (value[i] != null)
					{
						newValue[nv] = value[i];
						nv++;
					}

				setValue(newValue);
			}
			else
			{
				for (i = 0; i < n; i++)
					if (value[i] != null)
					{
						setValue(value[i]);
						break;
					}
			}

		}
		else if (relationFieldInfo.isMultiple())
			setValue(value);
		else
			setValue(value[0]);
	}

	/**
	 * Constructs a relation value for the given relation field from a LyObject.
	 *
	 * @param     relationFieldInfo Relation fieldInfo for which a value is created.
	 * @param     value             Value given as a LyObject.
	 * @exception IllegalArgumentException If the given object doesn't belong to a class to which
	 *                                     the relation field is related with.
	 */
	public LyRelationValue(LyRelationFieldInfo relationFieldInfo, LyObject value)
	{
		super(relationFieldInfo);

		if (value == null)
			return;

		LyClassInfo classInfo = value.getClassInfo();

		if (!relationFieldInfo.relatedWith(classInfo))
			throw new IllegalArgumentException("BF AppEngine : class info " + classInfo
				+ " is not supported by relation field " + relationFieldInfo);

		if (relationFieldInfo.isMultiple())
		{
			LyObjectId[] values = new LyObjectId[1];

			values[0] = value.getObjectId();
			setValue(values);
		}
		else
			setValue(value.getObjectId());
	}

	/**
	 * Constructs a relation value for the given relation field from a LyObjectList.
	 *
	 * @param     relationFieldInfo Relation fieldInfo for which a value is created.
	 * @param     value             Value given as a LyObjectId.
	 * @exception IllegalArgumentException If the relation field is not multiple and the list size
	 *                                     is greater than 1.
	 * @exception IllegalArgumentException If one object of the given list doesn't belong to a class
	 *                                     to which the relation field is related with.
	 */
	public LyRelationValue(LyRelationFieldInfo relationFieldInfo, LyObjectList value)
	{
		super(relationFieldInfo);

		if ((value == null) || (value.getSize() == 0))
			return;

		// Check classes of object ids
		if (relationFieldInfo.isMultiple())
		{
			int			 i, n = value.getSize();
			LyObjectId[] values = new LyObjectId[n];
			LyObject     object;
			LyClassInfo  classInfo;

			for (i = 0; i < n; i++)
			{
				object = value.getObject(i);
				classInfo = object.getClassInfo();

				if (!relationFieldInfo.relatedWith(classInfo))
					throw new IllegalArgumentException("BF AppEngine : class info "
						+ classInfo + " is not supported by relation field " + relationFieldInfo);

				values[i] = object.getObjectId();
			}

			setValue(values);
		}
		else
		{
			if (value.getSize() > 1)
				throw new IllegalStateException("BF AppEngine : fieldInfo "
					+ relationFieldInfo + " is not multiple and can't accept more than one value");

			LyObject    object = value.getObject(0);
			LyClassInfo classInfo = object.getClassInfo();

			if (!relationFieldInfo.relatedWith(classInfo))
				throw new IllegalArgumentException("BF AppEngine : class info "
					+ classInfo + " is not supported by relation field " + relationFieldInfo);

			setValue(object.getObjectId());
		}
	}

	/**
	 * Gets value at the given index as a LyObjectId.
	 *
	 * @param     index Index of requested object id.
	 * @return    Value at the given index returned as LyObjectId.
	 * @exception IllegalArgumentException If index is negative.
	 * @exception IllegalArgumentException If index is greater than the number of objects selected
	 *                                     in this value.
	 */
	public LyObjectId getObjectId(int index)
	{
		short count = getValueCount();

		if (index < 0)
			throw new IllegalArgumentException("BF AppEngine : index can't be negative");
		if (index >= count)
			throw new IllegalArgumentException("BF AppEngine : index must be lower than " + count);

		Object value = getValue();

		if (value == null)
			return null;
		else if (value instanceof LyObjectId)
			return (LyObjectId)value;
		else
			return ((LyObjectId[])value)[index];
	}

	/**
	 * Gets value at the given index as a LyObject.
	 *
	 * @param     index Index of requested object.
	 * @return    Value at the given index returned as LyObject.
	 * @exception IllegalArgumentException If index is negative.
	 * @exception IllegalArgumentException If index is greater than the number of objects selected
	 *                                     in this value.
	 */
	public LyObject getObject(int index)
	{
		LyObjectId objectId = getObjectId(index);

		return objectId.getLogicalObject();
	}

	/**
	 * Gets value at the given index as a String.
	 *
	 * @param     index Index of requested string.
	 * @return    Value at the given index returned as String.
	 * @exception IllegalArgumentException If index is negative.
	 * @exception IllegalArgumentException If index is greater than the number of objects selected
	 *                                     in this value.
	 */
	public String getString(int index)
	{
		LyObjectId objectId = getObjectId(index);

		return objectId.toName();
	}

	/**
	 * Gets number of selected objects in this relation value.
	 *
	 * @return Number of selected objects (0 if no object is selected in this value, can't be
	 *         greater than 1 if the corresponding relation field is not multiple).
	 */
	public short getValueCount()
	{
		Object value = getValue();

		if (value == null)
			return 0;
		else if (value instanceof LyObjectId)
			return 1;
		else
			return (short)((LyObjectId[])value).length;
	}

	/**
	 * Gets value of this relation value as a LyObjectList.
	 *
	 * @return An objects list corresponding to this relation value, null if the number of values
	 *         contained in this relation value is 0.
	 */
	public LyObjectList getObjectList()
	{
		int n = getValueCount();

		if (n == 0)
			return null;

		LyRelationFieldInfo relation = (LyRelationFieldInfo)getFieldInfo();
		LySimpleObjectList  list = new LySimpleObjectList();
		LyObject		    object;
		int				    i;

		if (relation.isOrdered())
			list.setSorted(false);

		for (i = 0; i < n; i++)
		{
			object = getObject(i);
			if (object != null)
				list.addObject(object);
		}

		return list;
	}
} // end class LyRelationValue
