/*
 * LyResponse.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;


/**
 * This class implements a response to an operation requested on a data source.
 *
 * <p>Supported operations are : create, set, delete, load, count, save.</p>
 *
 * <p>The status of the response is used to know : if the operation has been completed (COMPLETED),
 * if it has not been sent yet (PENDING), if there is no result yet (RUNNING), or if an error
 * occured (ERROR).</p>
 *
 * <p>Depending on the requested operation, the result can be simple or multiple.</p>
 *
 * <p>If an error occurs, the response contains an error message which indicates the origin of the
 * error.</p>
 *
 * @see LyDataSource
 */
public final class LyResponse
{
	/** File version */
	public static final String   __VERSION =
		"$Id: LyResponse.java 29553 2011-02-10 11:11:34Z nhamon $";

    /** Response statuses. */
    public enum Status
    {
    	/** an error occured during the operation. */
    	ERROR,
    	/** the operation completed successfully. */
    	COMPLETED,
    	/** the operation is not sent yet. */
    	PENDING,
    	/** the operation is in progress. */
    	RUNNING;
    }

    /** Operation types. */
    public enum Operation
    {
    	/** Create operation type. */
    	CREATE,
    	/** Set operation type. */
    	SET,
    	/** Delete operation type. */
    	DELETE,
    	/** Load operation type. */
    	LOAD,
    	/** Count operation type. */
    	COUNT,
    	/** Save operation type. */
    	SAVE
    }

	/** The class on which is the operation. */
	protected LyClass			 _class;

	/** Type of the operation. */
	protected Operation  		 _operation;

	/** Status of the response. */
	protected Status    		 _status;

	/** Error message when an error occured. */
	protected String			 _error;

	/**
	 * The set of responses : it is a set of objects. These objects are data objects for all
	 * operations except for OPER_COUNT where the object is an instance of Integer with the result
	 * of the count.
	 */
	protected Vector<Object>     _responses;

	/**
	 * The set of values : it is a set of LyValueSet. These objects are old values of application
	 * data objects. This vector may contain null values.
	 */
	protected Vector<LyValueSet> _values;

	/**
	 * Listener of the response. This is the client which requested the operation on the data
	 * source.
	 */
	protected LyResponseListener _listener;

	/** Indicates if the response is sorted (for a load operation). */
	protected boolean			 _sorted;

	/** Indicates if the response is filtered (for a load operation). */
	protected boolean			 _filtered;

	/** Indicates if the response is counted (for a load operation). */
	protected boolean			 _counted;

	/** Indicates if the response has distinct values or not. */
	protected boolean			 _distinctValues;

	/** Indicates if the response is partial (for a load operation). */
	protected boolean			 _partial;

	/** Indicates if the response is full or not (for a load operation). */
	protected boolean			 _fullLoad;

	/** Indicates if the response is in a notification case. */
	protected boolean			 _notified;

	/** Indicates if the operation is local or not. */
	protected boolean			 _local;

	/** Indicates if the response may caused the sending of events or not. */
	protected boolean			 _refresh;

	/**
	 * Indicates if linked objects (composition relations, etc...) must be deleted in a
	 * delete operation.
	 */
	protected boolean			 _deleteLinkedObjects;

	/** Indicates if the response is paginated (for a load operation). */
    protected boolean            _paginated;

	/** Indicates if the response may manage partial notification of loaded objects. */
    protected boolean            _partialNotification;
    
    /** Indicates if the response enable reloading of existing objects or not. */
    protected boolean            _enableReloading;
    
	/**
	 * Constructs a response concerning the given operation type.
	 *
	 * @param operation Type of operation.
	 * @param cls       The class of the operation.
	 */
	public LyResponse(Operation operation, LyClass cls)
	{
		// Initialize attributes.
		_operation = operation;
		_class = cls;
		_status = Status.PENDING;
		_error = null;
		_responses = null;
		_values = null;
		_listener = null;
		_sorted = false;
		_filtered = false;
		_counted = false;
		_partial = false;
		_fullLoad = false;
		_distinctValues = false;
		_notified = false;
		_local = false;
		_refresh = true;
		_deleteLinkedObjects = true;
		_paginated = false;
		_enableReloading = true;
		_partialNotification = false;
	}

	/**
	 * Gets the operation type of the response.
	 *
	 * @return Operation type of this response.
	 */
	public Operation getOperation()
	{
		return _operation;
	}

	/**
	 * Gets the status of the response.
	 *
	 * @return Status of this response.
	 */
	public Status getStatus()
	{
		return _status;
	}

	/**
	 * Gets the listener which sent the request to the data source.
	 *
	 * @return Listener of this response.
	 */
	public LyResponseListener getListener()
	{
		return _listener;
	}

	/**
	 * Gets error message of the response.
	 *
	 * @return    Error message.
	 * @exception IllegalStateException If the response has not the error status.
	 */
	public String getError()
	{
		if (_status != Status.ERROR)
			throw new IllegalStateException(
				"BF AppEngine : Attempt to get an error message from a response that is not in error");

		return _error;
	}

	/**
	 * Checks if the response is sorted. For a load request, responses given take into account the
	 * sort that was requested.
	 *
	 * @return True if responses were sorted, false otherwise.
	 */
	public boolean isSorted()
	{
		return _sorted;
	}

	/**
	 * Checks if the response is filtered. For a load request, responses given take into account the
	 * filter that was specified.
	 *
	 * @return True if responses were filtered, false otherwise.
	 */
	public boolean isFiltered()
	{
		return _filtered;
	}

	/**
	 * Checks if the response is counted. For a load request, responses given take into account the
	 * number of responses that was requested.
	 *
	 * @return True if responses were counted, false otherwise.
	 */
	public boolean isCounted()
	{
		return _counted;
	}

	/**
	 * Checks if the response has distincted values or not. For a load request, responses may
	 * contain duplicate values depending of the capacity of the provider.
	 *
	 * @return True if responses contains distincted values , false otherwise.
	 */
	public boolean hasDistinctValues()
	{
		return _distinctValues;
	}

	/**
	 * Checks if the response is partial.
	 *
	 * @return True if responses is partial, false otherwise.
	 */
	public boolean isPartial()
	{
		return _partial;
	}

	/**
	 * Checks if the response is a ful load.
	 *
	 * @return True if responses is a full load, false otherwise.
	 */
	public boolean isFullLoad()
	{
		return _fullLoad;
	}

	/**
	 * Check if the response is a notification one.
	 *
	 * @return true if the response is a notification, false otherwise.
	 */
	public boolean isNotified()
	{
		return _notified;
	}

	/**
	 * Check if the operation is local or not.
	 *
	 * @return true if the operation is local, false otherwise.
	 */
	public boolean isLocal()
	{
		return _local;
	}

	/**
	 * Check if the operation is refreshed or not.
	 *
	 * @return true if the operation is refresh, false otherwise.
	 */
	public boolean isRefresh()
	{
		return _refresh;
	}

	/**
	 * Checks if linked objects are deleted or not.
	 *
	 * @return True if deleted, false otherwise.
	 */
	public boolean deleteLinkedObjects()
	{
		return _deleteLinkedObjects;
	}

	/**
	 * Gets the number of available responses.
	 *
	 * @return Number of available responses.
	 */
	public synchronized int getSize()
	{
		return ((_responses == null) ? 0 : _responses.size());
	}

	/**
	 * Get the object of this response at the given position.
	 *
	 * @param     idx Index of the requested object.
	 * @return    Object at the given index.
	 * @exception IllegalStateException     If the response is still pending (not sent).
	 * @exception IndexOutOfBoundsException If index is negative or greater or equals to the number
	 *                                      of available responses.
	 */
	public synchronized Object getResponse(int idx)
	{
		if (_status == Status.PENDING)
			throw new IllegalStateException(
				"BF AppEngine : Attempt to get a response on a pending operation");

		if (idx < 0)
			throw new IndexOutOfBoundsException(
				"BF AppEngine : Attempt to get an out of bounds response: " + idx + " < 0");

		int n = ((_responses == null) ? 0 : _responses.size());

		if (idx >= n)
			throw new IndexOutOfBoundsException(
				"BF AppEngine : Attempt to get an out of bounds response: " + idx + " >= " + n);

		return _responses.elementAt(idx);
	}

	/**
	 * Get the values set of this response at the given position.
	 *
	 * @param     idx Index of the requested values set.
	 * @return    Values set at the given index. May be null.
	 * @exception IllegalStateException     If the response is still pending (not sent).
	 * @exception IndexOutOfBoundsException If index is negative or greater or equals to the number
	 *                                      of available responses.
	 */
	public synchronized LyValueSet getValues(int idx)
	{
		if (_status == Status.PENDING)
			throw new IllegalStateException(
				"BF AppEngine : Attempt to get a response on a pending operation");

		if (idx < 0)
			throw new IndexOutOfBoundsException(
				"BF AppEngine : Attempt to get an out of bounds response: " + idx + " < 0");

		int n = ((_values == null) ? 0 : _values.size());

		if (idx >= n)
			throw new IndexOutOfBoundsException(
				"BF AppEngine : Attempt to get an out of bounds response: " + idx + " >= " + n);

		return _values.elementAt(idx);
	}

	/**
     * Gets a copy of the list of objects in the response. This methods makes a copy of the list an
     * calling this method may be slower than getResponse.
     *
     * @return    The object list
     * @exception IllegalStateException If the response is still pending (not sent).
     * @see       #getResponse
     */
    public synchronized LyObjectList getObjects()
    {
        return getObjects(_distinctValues);
    }
    
	/**
	 * Gets a copy of the list of objects in the response. This methods makes a copy of the list an
	 * calling this method may be slower than getResponse.
	 *
	 * @param distinctValues   Boolean indicating if distinct values are requested.
	 * @return    The object list
	 * @exception IllegalStateException If the response is still pending (not sent).
	 * @see       #getResponse
	 */
	public synchronized LyObjectList getObjects(boolean distinctValues)
	{
		int				   n = getSize();
		int				   i = 0;
		LyObject		   o;
		LySimpleObjectList sol = new LySimpleObjectList(!distinctValues);

		while (i < n)
		{
			o = (LyObject)getResponse(i++);
			sol.addObject(o);
		}

		return sol;
	}

	/**
	 * Checks if an object is contained in responses list.
	 *
	 * @param  object Checked object.
	 * @return True if the given object is contained in this response.
	 */
	public boolean contains(LyObject object)
	{
		return ((_responses != null) && _responses.contains(object));
	}

	/**
	 * Sets the listener of this response.
	 *
	 * @param     listener Listener of this response.
	 * @exception IllegalStateException If this response has already a listener.
	 */
	public void setListener(LyResponseListener listener)
	{
		if (_listener != null)
			throw new IllegalStateException("BF AppEngine : the response has already a listener");

		_listener = listener;
	}

	/**
	 * Call the listener to notify it of a new response concerning the given object (with an
	 * optional list of values given).
	 *
	 * @param object New object of the response.
	 * @param values List of values changed on the object (may be null).
	 */
	protected void callListener(Object object, LyValueSet values)
	{
		if (_listener == null)
			return;

		if (object != null)
			_listener.newResponse(this, object, values);
		else
			_listener.endResponse(this);
	}

	/**
	 * Sets whether this response is sorted or not. This method is used by data providers to
	 * indicate if the sort could be done on the physical layer.
	 *
	 * @param sorted True if the response is sorted, false otherwise.
	 */
	public void setSorted(boolean sorted)
	{
		_sorted = sorted;
	}

	/**
	 * Sets whether this response is filtered or not. This method is used by data providers to
	 * indicate if the filter could be applied on the physical layer.
	 *
	 * @param filtered True if the response is filtered, false otherwise.
	 */
	public void setFiltered(boolean filtered)
	{
		_filtered = filtered;
	}

	/**
	 * Set whether this response is a notification or not. This method is used by the event decoder
	 * when loading an object.
	 *
	 * @param notified true if the response is a notification.
	 */
	public void setNotified(boolean notified)
	{
		_notified = notified;
	}

	/**
	 * Set whether this operation is local or not. This method is used by the object source for
	 * local operations.
	 *
	 * @param local true if the operation is local.
	 */
	public void setLocal(boolean local)
	{
		_local = local;
	}

	/**
	 * Set whether this operation is refresh or not. This method is used by the object source for
	 * refresh operations.
	 *
	 * @param refresh true if the operation is refresh.
	 */
	public void setRefresh(boolean refresh)
	{
		_refresh = refresh;
	}

	/**
	 * Modified of the _deleteLinkedObjects property.
	 * @param deleteLinkedObjects a list of LyObject
	 */
	public void deleteLinkedObjects(boolean deleteLinkedObjects)
	{
		_deleteLinkedObjects = deleteLinkedObjects;
	}

	/**
	 * Sets whether this response is counted or not. This method is used by data providers to
	 * indicate if the count could be done on the physical layer.
	 *
	 * @param counted True if the response is counted, false otherwise.
	 */
	public void setCounted(boolean counted)
	{
		_counted = counted;
	}

	/**
	 * Sets whether this response has distinct values or not. This method is used by data providers
	 * to indicate if the physical layer can distinct values or not for a LOAD request.
	 *
	 * @param distinctValues True if the response has distinct values, false otherwise.
	 */
	public void setDistinctValues(boolean distinctValues)
	{
		_distinctValues = distinctValues;
	}

	/**
	 * Sets whether this response is partial or not.
	 *
	 * @param partial True if the response is partial, false otherwise.
	 */
	public void setPartial(boolean partial)
	{
		_partial = partial;
	}

	/**
	 * Sets whether this response is a full loading or not.
	 *
	 * @param fullLoad True if the response is a full loading, false otherwise.
	 */
	public void setFullLoad(boolean fullLoad)
	{
		_fullLoad = fullLoad;
	}

	/**
	 * Sets the status of the response.
	 *
	 * @param     status New status of this response.
	 * @exception IllegalArgumentException If the given status is unknown.
	 */
	public void setStatus(Status status)
	{
		_status = status;
	}

	/**
	 * Sets the error message of the response.
	 *
	 * @param error New error message.
	 */
	public void setError(String error)
	{
		_error = error;
	}

	/**
	 * Adds a new response for the given object in this response.
	 *
	 * @param     object Object concerned by the response.
	 * @param     values List of values changed on the given object (may be null).
	 * @exception IllegalStateException If the status of this response is pending.
	 * @exception IllegalStateException If the status of this response is error and the given object
	 *                                  is not null.
	 */
	public synchronized void addResponse(Object object, LyValueSet values)
	{
		if (((_status == Status.ERROR) && (object != null))
				|| (_status == Status.PENDING))
			throw new IllegalStateException(
				"BF AppEngine : Attempt to add a response with a bad status");

		if (_responses == null)
			_responses = new Vector<Object>(1);

		if (_values == null)
			_values = new Vector<LyValueSet>(1);

		if (object != null)
		{
			if ((!_distinctValues) && (_responses.contains(object)))
				return;

			_responses.addElement(object);
			_values.addElement(values);
		}

		callListener(object, values);
	}

	/**
	 * Method to get listeners that don't need to be notified of an event following this response.
	 *
	 * @return List of listeners that don't need to be notified.
	 */
	public Vector<LyDataSourceListener> getSkippedListeners()
	{
		if (_listener instanceof LyGetDataSourceListenersInterface)
			return ((LyGetDataSourceListenersInterface)_listener).getDataSourceListeners();
		else if (_listener instanceof LyDataSourceListener)
		{
			Vector<LyDataSourceListener> listeners = new Vector<LyDataSourceListener>(1);

			listeners.addElement((LyDataSourceListener)_listener);
			return listeners;
		}
		else
			return null;
	}
	
	/**
     * Checks if the response is paginated. This method is used to check if pagination
     * was done on the physical layer.
     *
     * @return True if responses were paginated, false otherwise.
     */
    public boolean isPaginated()
    {
        return _paginated;
    }
    
	/**
     * Sets whether this response is paginated or not. This method is used by data providers to
     * indicate if the pagination could be done on the physical layer.
     *
     * @param paginated True if the response is paginated, false otherwise.
     */
    public void setPaginated(boolean paginated)
    {
        _paginated = paginated;
    }
    
    /**
     * Checks if the response allows existing objects to be reloaded.
     *
     * @return True if reloading is enabled, false otherwise.
     */
    public boolean isReloadingEnabled()
    {
        return _enableReloading;
    }
    
    /**
     * Sets whether the response allows existing objects to be reloaded.
     */
    public void enableReloading(boolean reloading)
    {
        _enableReloading = reloading;
    }

	/**
     * Checks if the response manages partial notifications of loaded objects.
     *
     * @return True if responses manages partial notifications of loaded objects, false otherwise.
     */
    public boolean isPartialNotification()
    {
        return _partialNotification;
    }
    
	/**
     * Sets whether this response manages partial notifications of loaded objects.
     *
     * @param partialNotification True if the response manages partial notifications of loaded objects, false otherwise.
     */
    public void setPartialNotification(boolean partialNotification)
    {
    	_partialNotification = partialNotification;
    }
} // end class LyResponse
