/*
 * LySimpleObjectList.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;
import leon.info.*;
import leon.misc.*;


/**
 * This class implements a writable list of objects.
 *
 * <p>These lists offer methods to add and remove objects.</p>
 *
 * <p>This kind of list may allow multiple occurrences of the same object if this mode is chosen.
 * </p>
 */
public final class LySimpleObjectList extends LyObjectList
{
	/** File version */
	public static final String		   __VERSION =
		"$Id: LySimpleObjectList.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * List of classes of this object list. These classes are stored in an instance of LyMultiList
	 * to count number of objects corresponding to each of these classes.
	 */
	protected LyMultiList<LyClassInfo> _classes;

	/**
	 * Constructs an empty list with multiple occurrences not allowed.
	 */
	public LySimpleObjectList()
	{
		this(false);
	}

	/**
	 * Constructs a writable objects list with multiple occurrences allowed or not dependending on
	 * the given parameter.
	 *
	 * @param allowMultipleOccurences Boolean indicating if multiple occurences of the same object
	 *                                is allowed in this list.
	 */
	public LySimpleObjectList(boolean allowMultipleOccurences)
	{
		_classes = new LyMultiList<LyClassInfo>();
		if (allowMultipleOccurences)
			_checkExistence = false;
	}

	/**
	 * Constructs a simple objects list with a single object. This new list doesn't allow multiple
	 * occurrences.
	 *
	 * @param object Object to add to the constructed list.
	 */
	public LySimpleObjectList(LyObject object)
	{
		this(false);
		if (object != null)
			addObject(object);
	}

	/**
	 * Construct a simple objects list from an existing objects list. This new list doesn't allow
	 * multiple occurrences.
	 *
	 * @param list Objects to add to the constructed list.
	 */
	public LySimpleObjectList(LyObjectList list)
	{
		this(list, false);
	}

	/**
	 * Construct a simple objects list from an existing objects list. This new list doesn't allow
	 * multiple occurrences.
	 *
	 * @param list     Objects to add to the constructed list.
	 * @param sameSort Boolean indicating if sort of the list parameter should be applied to the new
	 *                 list.
	 */
	public LySimpleObjectList(LyObjectList list, boolean sameSort)
	{
		this(false);

		if (list != null)
		{
			_sorted = list._sorted;
			_sortByName = list._sortByName;
			if (sameSort)
				_sort = list._sort;

			addList(list);
		}
	}

	/**
	 * Return the list of all class infos to which objects of this workspace belong.
	 *
	 * @return List of classes to which objects of this workspace belong.
	 */
	public LyInfoList<LyClassInfo> getClasses()
	{
		Vector<LyClassInfo> classes = _classes.getObjects();

		return new LyInfoList<LyClassInfo>(classes);
	}

	/**
	 * Get count of objects for the given class info.
	 *
	 * @param  classInfo Class info for which objects count is requested.
	 * @return Number of objects of the given class info found in this workspace.
	 */
	public int getCount(LyClassInfo classInfo)
	{
		return _classes.getCount(classInfo);
	}

	/**
	 * Adds an object to this list.
	 *
	 * @param     object Object to add to this list.
	 * @exception IllegalArgumentException If the given object is null or if the list alreday
	 *                                     contains this object and doesn't allow multiple
	 *                                     occurrences.
	 */
	public void addObject(LyObject object)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		add(object);
	}

	/**
	 * Adds an object to this list at a given position.
	 *
	 * @param     object   Object to add to this list.
	 * @param     position Position where the objects is added.
	 * @exception IllegalArgumentException If the given object is null or if the list alreday
	 *                                     contains this object and doesn't allow multiple
	 *                                     occurrences.
	 */
	public void addObject(LyObject object, int position)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		add(object, position);
	}

	/**
	 * Removes an object from this list.
	 *
	 * @param     object Object to remove from this list.
	 * @exception IllegalArgumentException If the given object is null.
	 */
	public void removeObject(LyObject object)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		remove(object);
	}

	/**
	 * Register object in this list.
	 * @param object an object added to the list
	 */
	protected void objectAdded(LyObject object)
	{
		super.objectAdded(object);

		// Add one occurence on the class of the object
		LyClassInfo classInfo = object.getClassInfo();

		_classes.addObject(classInfo);
	}

	/**
	 * Register object in this list.
	 * @param object an object removed from the list
	 */
	protected void objectRemoved(LyObject object)
	{
		super.objectRemoved(object);

		// Remove one occurence on the class of the object
		_classes.removeObject(object.getClassInfo());
	}

	/**
	 * Gets the number of occurences count for the given object.
	 * @param object the object
	 *
	 * @return Number of occurences of the given object in this list.
	 */
	public int getCount(LyObject object)
	{
		int index = getIndex(object);

		if (index < 0)
			return 0;

		int iMin, iMax, n = getSize();

		for (iMin = index - 1; (iMin >= 0) && (_objects.elementAt(iMin) == object); iMin--)
			;
		for (iMax = index + 1; (iMax < n) && (_objects.elementAt(iMax) == object); iMax++)
			;

		return iMax - iMin - 1;
	}
} // end class LySimpleObjectList
