/*
 * LySimpleWorkSpace.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;
import leon.app.*;
import leon.info.*;
import leon.misc.*;


/**
 * This class implements a simple workspace : dynamic objects list where objects are added and
 * removed "by hand" (directly using addObject and removeObject methods) .
 *
 * <p>Objects of different classes can be managed in this type of workspace.</p>
 */
public final class LySimpleWorkSpace extends LyWorkSpace
{
	/** File version */
	public static final String		   __VERSION =
		"$Id: LySimpleWorkSpace.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * List of classes of this workSpace. These classes are stored in an instance of LyMultiList to
	 * count number of objects corresponding to each of these classes.
	 */
	protected LyMultiList<LyClassInfo> _classes;

	/** Allows listener to be notified on object creations for classes of this list. */
	protected boolean				   _notifyCreation;

	/**
	 * Allows listener to be notified on object changes for classes of this list when the modified
	 * object is not yet in the list.
	 */
	protected boolean				   _notifyChange;

	/**
	 * Constructs a simple workspace within a given session.
	 *
	 * @param session Session in which this workspace is constructed.
	 * @param dynamic Specifies if this workspace is dynamic or not.
	 */
	public LySimpleWorkSpace(LyDataSession session, boolean dynamic)
	{
		super(session, dynamic, false);

		_classes = new LyMultiList<LyClassInfo>();
		_notifyCreation = false;
		_notifyChange = false;
	}

	/**
	 * Constructs a simple workspace within a given session with a single object.
	 *
	 * @param session Session in which this workspace is constructed.
	 * @param object  Object added in the constructed workspace.
	 * @param dynamic Specifies if this workspace is dynamic or not.
	 */
	public LySimpleWorkSpace(LyDataSession session, LyObject object, boolean dynamic)
	{
		this(session, dynamic);

		addObject(object);
	}

	/**
	 * Constructs a simple workspace within a given session from an objects list.
	 *
	 * @param session Session in which this workspace is constructed.
	 * @param list    Objects list used to initialize this workspace.
	 * @param dynamic Specifies if this workspace is dynamic or not.
	 */
	public LySimpleWorkSpace(LyDataSession session, LyObjectList list, boolean dynamic)
	{
		this(session, dynamic);

		if (list != null)
		{
			setSorted(list.isSorted());
			addList(list);
		}
	}

	/**
	 * Free this workspace. Removes this workspace as listener from all object sources corresponding
	 * to object classes found in this workspace.
	 */
	protected void free()
	{
		if (isDynamic())
		{
			LyObjectSource source;
			LyClassInfo    classInfo;
			int			   i, n = _classes.getSize();

			for (i = 0; i < n; i++)
			{
				classInfo = _classes.getObject(i);
				source = getObjectSource(classInfo);
				source.removeListener(this);
			}
		}

		super.free();
	}

	/**
	 * Return the list of all class infos to which objects of this workspace belong.
	 *
	 * @return List of classes to which objects of this workspace belong.
	 */
	public LyInfoList<LyClassInfo> getClasses()
	{
		Vector<LyClassInfo> classes = _classes.getObjects();

		return new LyInfoList<LyClassInfo>(classes);
	}

	/**
	 * Get count of objects for the given class info.
	 *
	 * @param  classInfo Class info for which objects count is requested.
	 * @return Number of objects of the given class info found in this workspace.
	 */
	public int getCount(LyClassInfo classInfo)
	{
		return _classes.getCount(classInfo);
	}

	/**
	 * Allows listeners to be notified on object creations for classes of this workspace. This
	 * method has effect only if this workspace is dynamic.
	 *
	 * @param notifyCreation Boolean indicating if notifications for creations are requested.
	 */
	public void setNotifyCreation(boolean notifyCreation)
	{
		if (!isDynamic())
			return;
		if (_notifyCreation == notifyCreation)
			return;

		_notifyCreation = notifyCreation;

		LyClassInfo    classInfo;
		LyObjectSource source;
		int			   i, n = _classes.getSize();

		for (i = 0; i < n; i++)
		{
			classInfo = _classes.getObject(i);
			source = getObjectSource(classInfo);
			if (_notifyCreation)
				source.addListener(this, LyDataSourceEvent.TYPE_CREATE);
			else
			{
				source.removeListener(this);
				source.addListener(this, LyDataSourceEvent.TYPE_SET);
				source.addListener(this, LyDataSourceEvent.TYPE_DELETE);
				source.addListener(this, LyDataSourceEvent.TYPE_SELECTION);
				source.addListener(this, LyDataSourceEvent.TYPE_VISIBILITY);
				source.addListener(this, LyDataSourceEvent.TYPE_ERROR);
			}
		}
	} // end method setNotifyCreation

	/**
	 * Allows listeners to be notified on object changes for classes of this workspace concerning
	 * objects that are not in the workspace. This method has effect only if this workspace is
	 * dynamic.
	 *
	 * @param notifyChange Boolean indicating if notifications for all changes are requested.
	 */
	public void setNotifyChange(boolean notifyChange)
	{
		if (!isDynamic())
			return;
		if (_notifyChange == notifyChange)
			return;

		_notifyChange = notifyChange;
	}

	/**
	 * Add given class to known classes to be able to receive creation events. This method does make
	 * sense only if this workspace is dynamic. It's necessary to receive creation before any object
	 * of the given class was added.
	 *
	 * @param classInfo The class info to register.
	 */
	public void registerClassInfo(LyClassInfo classInfo)
	{
		// Add this class to known classes and add this as listener if necessary
		if (_classes.addObject(classInfo) == 1)
		{
			if (isDynamic())
			{
				LyObjectSource source = getObjectSource(classInfo);

				if (_notifyCreation)
                {
                    if (getSize(classInfo) > 0)
                        // If we already have an object of specified classInfo, increment classInfo count
                        // to avoid unregistering it when removing the last object of the classInfo.
                        _classes.addObject(classInfo);
					source.addListener(this, LyDataSourceEvent.TYPE_CREATE);
                }
				source.addListener(this, LyDataSourceEvent.TYPE_SET);
				source.addListener(this, LyDataSourceEvent.TYPE_DELETE);
				source.addListener(this, LyDataSourceEvent.TYPE_SELECTION);
				source.addListener(this, LyDataSourceEvent.TYPE_VISIBILITY);
				source.addListener(this, LyDataSourceEvent.TYPE_ERROR);
			}
		}
	}

	/**
	 * Remove given class to known classes to be able to receive creation events. This method does
	 * make sense only if this workspace is dynamic.
	 *
	 * @param classInfo The class info to unregister.
	 */
	protected void unregisterClassInfo(LyClassInfo classInfo)
	{
		// Remove one occurence on the class of the object and remove this
		// as listener when there is no more occurrence of the class
		if (_classes.removeObject(classInfo) == 0)
		{
			if (isDynamic())
			{
				LyObjectSource source = getObjectSource(classInfo);

				source.removeListener(this);
				if (_notifyCreation)
					source.addListener(this, LyDataSourceEvent.TYPE_CREATE);
				if (_notifyChange)
					source.addListener(this, LyDataSourceEvent.TYPE_SET);
			}
		}
	}

	/* Management of the content of the list */

	/**
	 * Register object in this list.
	 * @param object an object added to the list
	 */
	protected void objectAdded(LyObject object)
	{
		super.objectAdded(object);

		// Add this class to known classes and add this as listener if necessary
		LyClassInfo classInfo = object.getClassInfo();

		registerClassInfo(classInfo);
	}

	/**
	 * Register object in this list.
	 * @param object an object removed from the list
	 */
	protected void objectRemoved(LyObject object)
	{
		super.objectRemoved(object);

		// Add this class to known classes and add this as listener if necessary
		LyClassInfo classInfo = object.getClassInfo();

		unregisterClassInfo(classInfo);
	}

	/**
	 * Adds an object in this workspace. This method is overriden from LyWorkSpace to become public.
	 *
	 * @param  object Object to add in the workspace.
	 * @return If the object was not in this list, returns a positive index giving the position
	 *         where the object was added. Otherwise, returns a negative index giving the position
	 *         minus 1 where the same object was found.
	 */
	public int addObject(LyObject object)
	{
		return super.addObject(object);
	}

	/**
	 * Adds an object in this workspace at the given position. This method is overriden from
	 * LyWorkSpace to become public.
	 *
	 * @param  object   Object to add in the workspace.
	 * @param  position Position where the object is added in the list.
	 * @return If the object was not in this list, returns a positive index giving the position
	 *         where the object was added. Otherwise, returns a negative index giving the position
	 *         minus 1 where the same object was found.
	 */
	public synchronized int addObject(LyObject object, int position)
	{
		int index = add(object, position);

		if (index < 0)
			return index;

		// Notify this to listeners
		if ((_listeners == null) || (_listeners.size() == 0))
			return index;

		LyWorkSpaceEvent event = new LyWorkSpaceEvent(this, LyWorkSpaceEvent.Type.ADD, object,
				index);

		notify(event);
		return index;
	}

	/**
	 * Removes an object from this workspace. This method is overriden from LyWorkSpace to become
	 * public.
	 *
	 * @param     object Object to remove from the workspace.
	 * @return    Index where the object was found in this workspace.
	 * @exception IllegalArgumentException If the object was not found in this workspace.
	 */
	public int removeObject(LyObject object)
	{
		return super.removeObject(object);
	}

	/**
	 * Adds objects of the specified list to this list. This method is overriden from LyWorkSpace to
	 * re-enable it for this type of workspace.
	 *
	 * @param list Objects to add to this list.
	 */
	public void addList(LyObjectList list)
	{
		add(list);

		// If the workspace is not dynamic, do not send events...
		if (!isDynamic()
                || (_listeners == null)) // when created, the workspace may not have listeners yet
			return;

		// if this workspace is dynamic, we should notify an objectAdded for each object of the list.
		LyObject newObject;
		int		 index;

		for (int i = 0; i < list.getSize(); i++)
		{
			newObject = list.getObject(i);
			index = this.getIndex(newObject);

			if (index >= 0)
			{
				LyWorkSpaceEvent event = new LyWorkSpaceEvent(this, LyWorkSpaceEvent.Type.ADD,
						newObject, index);

				notify(event);
			}
		}
	}

	/**
	 * Removes objects of the specified list to this list.
	 *
	 * @param list Objects to remove to this list.
	 */
	public void removeList(LyObjectList list)
	{
		LyObject object;
		int		 i, n = list.getSize();

		for (i = 0; i < n; i++)
		{
			object = list.getObject(i);
			if (contains(object))
				removeObject(object);
		}
	}

	/* Implementation of LyDataSourceListener */

	/**
	 * Method for receiving an object creation. If the object is contained in this workspace, that
	 * means it was added with addObject method, so we send an event to our listeners.
	 *
	 * @param  event Event notifying the creation of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectCreated(LyDataSourceEvent event)
	{
		if (!_notifyCreation)
			return false;

		if ((_listeners == null) || (_listeners.size() == 0))
			return false;

		LyObject object = (LyObject)event.getObject();

		if (contains(object))
			return false;

		// In that case, this event is only notified, the object is not really
		// added in the list, listener will really add the object if needed
		LyWorkSpaceEvent workSpaceEvent = new LyWorkSpaceEvent(this, LyWorkSpaceEvent.Type.ADD,
				object, -1);

		workSpaceEvent.setEventInfo(event.getEventInfo());

		notify(workSpaceEvent);
		return true;
	}

	/**
	 * Method for receiving an object modification. Changes the object concerned by the event if it
	 * is found in this workspace.
	 *
	 * @param  event Event notifying the modification of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectModified(LyDataSourceEvent event)
	{
		LyObject object = (LyObject)event.getObject();

		if (!_notifyChange
				|| (indexOf(object) >= 0))
			return super.objectModified(event);

		// Notify the change even if the object is not in the list
		LyWorkSpaceEvent workSpaceEvent = new LyWorkSpaceEvent(this, LyWorkSpaceEvent.Type.CHANGE,
				object, -1, -1, event.getValues());

		workSpaceEvent.setEventInfo(event.getEventInfo());
		notify(workSpaceEvent);
		return true;
	}

	/**
	 * Allow multiple occurences of the same object.
	 */
	public void allowMultipleOccurences()
	{
		_checkExistence = false;
	}
} // end class LySimpleWorkSpace
