/*
 * LyStepWorkSpace.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;

import leon.app.*;
import leon.info.*;


/**
 * This workspace implements a step workspace : workspace for calculating an objects list obtained
 * from an initial objects list using a step.
 *
 * <p>A step contains several relations, a target class and a direction.</p>
 *
 * <p>Step workspaces can be chained to calculate more complicated cross references using routes
 * that contain several steps.</p>
 *
 * @see  LyStep
 * @see  LyRoute
 */
public final class LyStepWorkSpace extends LyWorkSpace
{
    /** File version. */
    public static final String __VERSION = "$Id: LyStepWorkSpace.java 35868 2012-06-14 08:40:31Z jlsaouli $";

    /** Parent objects list. */
    LyObjectList _parentList;

    /** Source classInfo. */
    LyClassInfo _parentClass;

    /** Step used to calculate result list. */
    LyStep _step;

    /**
     * Object counters table : only used when the workspace is dynamic. Associates a counter to each
     * object of this workspace to be able to known when they have to be removed.
     */
    Hashtable<LyObject, Integer> _counters;

    /** Internal listener to receive events on parent list. */
    LyWorkSpaceListener _listener;

	/**
	 * Boolean that indicates if reverse searches can be optimized (when all relations of the step
	 * have a reverse relation).
	 */
	boolean				_optimizeReverse;
	
	/**
	 * Boolean that indicates if recursion is on.
	 */
	boolean				_recurse;

    /**
     * Constructs a step workspace within a given session starting from an objects list restricted
     * to a class info using a specified step.
     *
     * @param      session      Session in which the workspace is constructed.
     * @param      parentList   Parent objects list of this step workspace.
     * @param      parentClass  Class info to which objects of this parent list have to belong.
     * @param      step         Step used for calculating result list.
     * @param      dynamic      Specifies if this workspace is dynamic or not.
     * @param      local        Specifies if this workspace is local or not.
     *
     * @exception  IllegalArgumentException  If parent list or parent class info is null.
     */
    protected LyStepWorkSpace(LyDataSession session, LyObjectList parentList, LyClassInfo parentClass, LyStep step,
        boolean dynamic, boolean local)
    {
    	this(session, parentList, parentClass, step, dynamic, local, false);
    }
    
    /**
     * Constructs a step workspace within a given session starting from an objects list restricted
     * to a class info using a specified step.
     *
     * @param      session      Session in which the workspace is constructed.
     * @param      parentList   Parent objects list of this step workspace.
     * @param      parentClass  Class info to which objects of this parent list have to belong.
     * @param      step         Step used for calculating result list.
     * @param      dynamic      Specifies if this workspace is dynamic or not.
     * @param      local        Specifies if this workspace is local or not.
     * @param 	   recurse      Activates recursion when sourceClass and targetClass are identical.
     *
     * @exception  IllegalArgumentException  If parent list or parent class info is null.
     */
    protected LyStepWorkSpace(LyDataSession session, LyObjectList parentList, LyClassInfo parentClass, LyStep step,
        boolean dynamic, boolean local, boolean recurse)
    {
        super(session, dynamic, local);
        
        _recurse = recurse;

        if (parentList == null)
            throw new IllegalArgumentException("BF AppEngine : parentList can't be null");

        if (parentClass == null)
            throw new IllegalArgumentException("BF AppEngine : parentClass can't be null");

        setSorted(parentList.isSorted());

        // Initializing the attributes.
        _parentList = parentList;
        _parentClass = parentClass;
        _step = step;
        _counters = null;

		// Check if reverse searches can be optimized
		LyInfoList<LyFieldInfo> relations = _step.getRelations();
		int					    i, n = relations.getSize();
		LyFieldInfo			    fieldInfo;

		_optimizeReverse = true;
		for (i = 0; (i < n) && _optimizeReverse; i++)
		{
			// Find objects referenced by this relation on this object
			fieldInfo = relations.getInfo(i);
			_optimizeReverse = ((fieldInfo instanceof LyRelationFieldInfo)
						&& (((LyRelationFieldInfo)fieldInfo).getReverseRelation() != null));
		}

        // Initial loading
        reset();

        if (dynamic)
        {
            // Add this as listener of classInfo objectSource
            LyClassInfo    target = _step.getTarget();
            LyObjectSource objectSource = getObjectSource(target);

            objectSource.addListener(this);

            if (_parentList.isDynamic()
                    && (_parentList instanceof LyWorkSpace))
            {
                // Listen to workspace events
                _listener = new _LyWorkSpaceListener();
                ((LyWorkSpace)_parentList).addListener(_listener);
            }

            createFilterListeners(_step.getFilter());
        }
        else
            // In that case, we don't need to keep reference
            _parentList = null;
    }

    /**
     * Free this workspace. Remove this workspace as listener of the parent workspace and of the
     * object source of the target class.
     */
    protected void free()
    {
        if (isDynamic())
        {
            _counters = null;

            LyClassInfo    target = _step.getTarget();
            LyObjectSource objectSource = getObjectSource(target);

            objectSource.removeListener(this);

            if (_parentList instanceof LyWorkSpace)
                // Remove this context from workspace events listeners
                ((LyWorkSpace)_parentList).free(_listener);
        }

        super.free();
    }

    /**
     * Return the list of all class infos to which objects of this workspace belong. Since the step
     * as a target class, we known from start to which class the objects of this workspace belong,
     * so this method is overriden to optimize it.
     *
     * @return  List of classes to which objects of this workspace belong. For this type of
     *          workspace, the result list will always contain only one class.
     */
    public LyInfoList<LyClassInfo> getClasses()
    {
        LyInfoList<LyClassInfo> classes = new LyInfoList<LyClassInfo>();

        classes.addInfo(_step.getTarget());
        return classes;
    }

    /**
     * Gets the objects list that is filtered by this sub-workspace.
     *
     * @return  Objects list filtered by this sub-workspace.
     */
    LyObjectList getParentList()
    {
        return _parentList;
    }

    /**
     * Gets the step used for calculating result list from parent list.
     *
     * @return  Step used by this step workspace.
     */
    LyStep getStep()
    {
        return _step;
    }

    /**
     * Gets counter value for the given object (number of occurrences of the object in the different
     * sub-workspaces).
     *
     * @param   object  Object for which the counter value is requested.
     *
     * @return  Counter value for the given object.
     */
    int getCount(LyObject object)
    {
        if (_counters == null)
            return -1;

        Integer count = _counters.get(object);

        return ((count == null) ? 0 : count.intValue());
    }

    /**
     * Sets counter value for the given object. If the counter value was 0, the object is added in
     * this workspace. If the counter value falls to 0, the object is removed from this workspace.
     *
     * @param  object  Object for which the counter value is changed.
     */
    void setCount(LyObject object, int newCount)
    {
        if (_counters == null)
            return;

        if (newCount <= 0)
        {
            if (getCount(object) > 0)
                removeObject(object);
        }
        else
        {
            Integer count = _counters.get(object);

            if (count == null)
                addObject(object);

            _counters.put(object, new Integer(newCount));
        }
    }

    /**
     * Increments counter value for the given object. If the counter value was 0, the object is
     * added in this workspace.
     *
     * @param  object  Object for which the counter value is incremented.
     */
    void incrementCount(LyObject object)
    {
        int counter = getCount(object) + 1;

        setCount(object, counter);

        if (counter > 1)
            counterChanged(object, true);
    }

    /**
     * Decrements counter value for the given object. If the counter value falls to 0, the object is
     * removed from this workspace.
     *
     * @param  object  Object for which the counter value is decremented.
     */
    void decrementCount(LyObject object)
    {
        int counter = getCount(object) - 1;

        setCount(object, counter);

        if (counter > 0)
            // Notify LyWorkSpaceCounterListener
            counterChanged(object, false);
    }

    void counterChanged(LyObject object, boolean incr)
    {
        LyWorkSpaceEvent event = new LyWorkSpaceEvent(this, LyWorkSpaceEvent.Type.COUNTER_CHANGE, object,
                getIndex(object));

        event.setIncremented(incr);
        notify(event);
    }

    /**
     * Method for resetting this workspace. Objects contained in this workspace are removed and
     * reloaded from parent list using the specified step.
     */
    public synchronized void reset()
    {
        // Remove all objects.
        super.reset();

        if (_parentList == null)
            return;

        // Clear objects list and dependencies
        _objects = null;
        _counters = (isDynamic() ? new Hashtable<LyObject, Integer>() : null);

        // Find objects referenced by this relation on this object
        LyObject         parent, object;
        Vector<LyObject> objects = _parentList.getObjects();
        
        if (objects == null)
        	return;
        
        int              i, np = objects.size();
        LyObjectList     list;
        int              j, no;
        LyClassInfo      parentClass;
        Vector<LyObject> subObjects;

        for (i = 0; i < np; i++)
        {
            parent = (LyObject)objects.elementAt(i);
            parentClass = parent.getClassInfo();
            if ((parentClass != _parentClass)
                    && !parentClass.inheritsFrom(_parentClass))
                continue;

            // Find related objects for this object
            list = getObjects(parent);       
            if (list == null)
                continue;
            
            if (_recurse
            	&& (_parentClass == _step.getTarget()))
            {
            	// Recurse if source and target classes are identical
            	subObjects = list.getObjects();
            	if (subObjects != null)
            	{
	            	objects.addAll(subObjects);
	            	np = objects.size();
            	}
            }

            // Add related objects to our result list
            for (j = 0, no = list.getSize(); j < no; j++)
            {
                object = list.getObject(j);
                if (_counters != null)
                    incrementCount(object);
                else
                    addObject(object);
            }
        }
    } // end method reset

    /**
     * Removes an object from this workspace.
     *
     * @param   object  Object to remove from the workspace.
     *
     * @return  Index where the object was found in this workspace.
     */
    protected int removeObject(LyObject object)
    {
        int index = getIndex(object);

        if ((index >= 0)
                && (_counters != null))
            // Remove counter for this object
            _counters.remove(object);

        // Remove object
        index = super.removeObject(object);

        if ((index >= 0)
                && (_counters != null))
            // Remove counter for this object
            _counters.remove(object);

        return index;
    }

    /**
     * Utility method to check if the given modifications impact relations involved in the current
     * step.
     *
     * @param   values  Values set that contains modifications done.
     *
     * @return  True if one of the relations of the step has a value found in the given values,
     *          false otherwise.
     */
    boolean isImpacted(LyValueSet values)
    {
        if (values == null)
            return true;

        LyInfoList<LyFieldInfo> relations = _step.getRelations();
        int                     i, n = relations.getSize();
        LyFieldInfo             relation;

        for (i = 0; i < n; i++)
        {
            // Find objects referenced by this relation on this object
            relation = relations.getInfo(i);
            if (values.getFieldValue(relation, true) != null)
                return true;
        }

        return false;
    }

    /**
     * Returns a string representation of this workspace.
     *
     * @return  This workspace returned as a string.
     */
    public String toString()
    {
        return super.toString() + "\nstep=" + _step + "\nparent=" + _parentList;
    }

    /* Internal methods to calculate cross references */

    /**
     * Method for retrieving source objects that lead to given target object.
     *
     * @param   target  Target object
     *
     * @return  Source objects that leade to target object.
     */
    public LySimpleObjectList getSources(LyObject target)
    {
        if (!contains(target))
            return null;

        LySimpleObjectList sources = getObjects(target, _parentList, true);

        if (sources == null)
            return null;

        if (_parentList instanceof LyStepWorkSpace)
        {
            int                i, n = sources.getSize();
            LyObject           source;
            LySimpleObjectList results = null, sourceResults;

            for (i = 0; i < n; i++)
            {
                source = sources.getObject(i);
                sourceResults = ((LyStepWorkSpace)_parentList).getSources(source);
                if (results == null)
                    results = sourceResults;
                else if (sourceResults != null)
                    results.addList(sourceResults);
            }

            return results;
        }
        else
            return sources;
    } // end method getSources

    /**
     * Utility method to caculate objects related to the given object.
     *
     * @param   object  Object for which related objects are requested. This object may belong to
     *                  source class or to target class.
     *
     * @return  A simple objects list that contains results, this list allows multiple occurrences
     *          of the same object.
     */
    LySimpleObjectList getObjects(LyObject object)
    {
        return getObjects(object, null);
    }

    /**
     * Utility method to caculate objects related to the given object.
     *
     * @param   object  Object for which related objects are requested. This object may belong to
     *                  source class or to target class.
     * @param   inList  If not null, this parameter limits the search to objects found in the given
     *                  list.
     *
     * @return  A simple objects list that contains results, this list allows multiple occurrences
     *          of the same object.
     */
    LySimpleObjectList getObjects(LyObject object, LyObjectList inList)
    {
        LySimpleObjectList result;

        if (object.instanceOf(_parentClass))
            result = getObjects(object, inList, false);
        else if (object.instanceOf(_step.getTarget()))
            result = getObjects(object, inList, true);
        else
        {
            result = new LySimpleObjectList(true);
            result.setSorted(_sorted);
        }

        return result;
    }

    /**
     * Utility method to caculate objects related to the given object.
     *
     * @param   object      Object for which related objects are requested. This object may belong
     *                      to source class or to target class.
     * @param   inList      If not null, this parameter limits the search to objects found in the
     *                      given list.
     * @param   fromTarget  Indicates if the specified object is from target class or parent class.
     *
     * @return  A simple objects list that contains results, this list allows multiple occurrences
     *          of the same object.
     */
    LySimpleObjectList getObjects(LyObject object, LyObjectList inList, boolean fromTarget)
    {
        LySimpleObjectList result = new LySimpleObjectList(true);

        result.setSorted(_sorted);

        if (!fromTarget)
        {
            if (!_step.isReverse())
                getRelatedObjects(result, object, _step.getTarget(), _step.getFilter());
            else
                getReversedObjects(result, object, _step.getTarget(), _step.getFilter(), inList);
        }
        else
        {
            if (_step.isReverse())
                getRelatedObjects(result, object, _parentClass, null);
            else
                getReversedObjects(result, object, _parentClass, null, inList);
        }

        return result;
    }

    /**
     * Utility method to get objects related to the given object using directly relations specified
     * in the step.
     *
     * @param  result       Result objects list.
     * @param  object       Object from which relation values are taken.
     * @param  targetClass  Class info to which result objects have to belong.
     * @param  filter       Filter applied on objects before adding them to result list if not null.
     */
    void getRelatedObjects(LySimpleObjectList result, LyObject object, LyClassInfo targetClass, LyFilter filter)
    {
        LyInfoList<LyFieldInfo> relations = _step.getRelations();
        int                     i, n = relations.getSize();
        LyFieldInfo             relation;
        Object                  value;

        for (i = 0; i < n; i++)
        {
            // Find objects referenced by this relation on this object
            relation = relations.getInfo(i);
            value = object.getValue(relation);
            if (value == null)
                continue;

            getRelatedObjects(result, value, targetClass, filter);
        }
    }

    /**
     * Gets objects found in values of the given values set or on the given object for relations
     * found in the value set corresponding to relations of the step, belonging to its target class
     * info and matching its filter, if not null.
     *
     * @param   values  Values set in which relations are selected and values are taken if parent is
     *                  null.
     * @param   parent  Object on which values are taken.
     *
     * @return  A simple objects list that contains results, this list allows multiple occurrences
     *          of the same object.
     */
    LyObjectList getRelatedObjects(LyValueSet valueSet, LyObject parent)
    {
        LySimpleObjectList result = new LySimpleObjectList(true);

        result.setSorted(_sorted);

        LyClassInfo             targetClass = _step.getTarget();
        LyFilter                filter = _step.getFilter();
        LyInfoList<LyFieldInfo> relations = _step.getRelations();
        int                     i, n = relations.getSize();
        LyFieldInfo             relation;
        LyValue                 value;

        for (i = 0; i < n; i++)
        {
            // Find objects referenced by this relation on this object
            relation = relations.getInfo(i);
            if (!valueSet.contains(relation, true))
                continue;

            if (parent == null)
                value = valueSet.getFieldValue(relation, true);
            else
                value = parent.getFieldValue(relation);

            getRelatedObjects(result, value.getValue(), targetClass, filter);
        }

        return result;
    } // end method getRelatedObjects

    /**
     * Add in result list objects found in the given relation value that belong to the specified
     * class info and match the filter (if not null).
     *
     * @param  result       Result objects list.
     * @param  value        Relation value in which objects are taken.
     * @param  targetClass  Class info to which result objects have to belong.
     * @param  filter       Filter applied on objects before adding them to result list if not null.
     */
    void getRelatedObjects(LySimpleObjectList result, Object value, LyClassInfo targetClass, LyFilter filter)
    {
        if (value == null)
            return;

        if (value instanceof LyValue) // To prevent errors
        {
            getRelatedObjects(result, ((LyValue)value).getValue(), targetClass, filter);
            return;
        }

        LyObject    rel;
        LyClassInfo relClassInfo;

        // Find objects referenced by this relation on this object
        if (value instanceof LyObjectId)
        {
            rel = getLogicalObject((LyObjectId)value);
            if (rel == null)
                return;

            relClassInfo = rel.getClassInfo();
            if ((relClassInfo != targetClass)
                    && (!rel.getClassInfo().inheritsFrom(targetClass)))
                return;

            if ((filter == null) || filter.match(rel))
                result.addObject(rel);
        }
        else if (value instanceof LyObjectId[])
        {
            LyObjectId[] list = (LyObjectId[])value;
            int          i, n = list.length;

            for (i = 0; i < n; i++)
            {
                if (list[i] == null)
                    continue;

                rel = getLogicalObject(list[i]);
                if (rel == null)
                    continue;

                relClassInfo = rel.getClassInfo();
                if ((relClassInfo != targetClass)
                        && (!rel.getClassInfo().inheritsFrom(targetClass)))
                    continue;

                if ((filter == null) || filter.match(rel))
                    result.addObject(rel);
            }
        }
        else
            throw new IllegalArgumentException("BF AppEngine : invalid values class "
                + value.getClass());
    } // end method getRelatedObjects

    /**
     * Utility method to get a logical object from its id. If this step workspace is local the
     * object is searched from memory.
     *
     * @param   objectId  The logical object Id
     *
     * @return  The logical object correponding to the specified id.
     */
    LyObject getLogicalObject(LyObjectId objectId)
    {
        if (_local)
        {
            LyClassInfo    classInfo = objectId.getClassInfo();
            LyObjectSource source = classInfo.getObjectSource();

            return (LyObject)source.getObject(objectId, _local);
        }
        else
            return objectId.getLogicalObject();
    }

    /**
     * Utility method to get objects of the target class that reference the given object through
     * relations of this step and match the specified filter (if not null).
     *
     * @param  result       Result objects list.
     * @param  object       Pointed object.
     * @param  targetClass  Class info to which result objects have to belong.
     * @param  filter       Filter applied on objects before adding them to result list if not null.
     * @param  inList       If not null, this parameter limits the search to objects found in the
     *                      given list.
     */
    void getReversedObjects(LySimpleObjectList result, LyObject object, LyClassInfo targetClass, LyFilter filter,
        LyObjectList inList)
    {
        LyInfoList<LyFieldInfo> relations = _step.getRelations();
        int                     i, n = relations.getSize();

		if (_optimizeReverse)
		{
			LyFieldInfo relation, reverseRelation;
			Object	    value;

			for (i = 0; i < n; i++)
			{
				// Find objects referenced by this relation on this object
				relation = relations.getInfo(i);
				if (!(relation instanceof LyRelationFieldInfo))
					continue;

				reverseRelation = ((LyRelationFieldInfo)relation).getReverseRelation();
				value = object.getValue(reverseRelation);
				if (value == null)
					continue;

				getRelatedObjects(result, value, targetClass, filter);
			}

			return;
		}

        // Find objects of the target classInfo that reference
        // the given object through relations of this step
        LyFilter                     relFilter = new LyFilter(LyFilter.Type.OR);
        LyFilterExpression           filterExp;
        LyFilterExpression.Condition condition;
        LyFieldInfo                  fieldInfo;

        for (i = 0; i < n; i++)
        {
            fieldInfo = relations.getInfo(i);
            if ((fieldInfo instanceof LyRelationFieldInfo)
                    && !((LyRelationFieldInfo)fieldInfo).isMultiple())
                condition = LyFilterExpression.Condition.EQUALS;
            else
                condition = LyFilterExpression.Condition.CONTAINS;

            filterExp = new LyFilterExpression(fieldInfo, false, condition, object.getObjectId());
            filterExp.setModifier(LyFilterExpression.MOD_CASE_SENSITIVE);

            relFilter.addElement(filterExp);
        }

        if (inList != null)
        {
            LyWorkSpace list = new LySubWorkSpace(_session, inList, targetClass, relFilter, false, _local);

            result.addList(list);
            return;
        }

        // Build full filter
        LyFilter fullFilter;

        if (filter == null)
            fullFilter = relFilter;
        else
        {
            fullFilter = new LyFilter(LyFilter.Type.AND);
            fullFilter.addElement(filter);
            fullFilter.addElement(relFilter);
        }

        // Get filtered list
        LyWorkSpace list = new LyClassWorkSpace(_session, targetClass, new LyInfoList<LyFieldInfo>(), fullFilter, null,
                false, _sorted, _local);

        result.addList(list);
    } // end method getReversedObjects

    /* Implementation of LyDataSourceListener */

    /**
     * Method for receiving an object creation.
     *
     * @param   event  Event notifying the creation of an object.
     *
     * @return  True indicating if the event was used, false otherwise.
     */
    public boolean objectCreated(LyDataSourceEvent event)
    {
        LyFilter filter = _step.getFilter();
        LyObject object = (LyObject)event.getObject();

        // If there is a filter, the object must match it.
        if ((filter != null) && !filter.match(object))
            return false;

        if (contains(object)) // This shouldn't append
            return false;

        // Get objects pointed by this object through this step and count objects
        // found in parent list, this will give us our object counter.
        LyObjectList list = getObjects(object);
        int          i, n = list.getSize();
        LyObject     parent;
        int          count = 0;

        for (i = 0; i < n; i++)
        {
            parent = list.getObject(i);
            if (_parentList.contains(parent))
                count++;
        }

        setCount(object, count);
        return true;
    } // end method objectCreated

    /**
     * Method for receiving an object modification.
     *
     * @param   event  Event notifying the modification of an object.
     *
     * @return  True indicating if the event was used, false otherwise.
     */
    public boolean objectModified(LyDataSourceEvent event)
    {
        LyFilter filter = _step.getFilter();
        LyObject object = (LyObject)event.getObject();
        int      count = getCount(object);

        // If the object doesn't match the filter
        if ((filter != null) && !filter.match(object))
        {
            // If it was not in our list, ignore this event
            if (count == 0)
                return false;

            // Otherwise, remove the object from our result list
            removeObject(object);
            return true;
        }
        else if (_step.isReverse() && (filter != null) && (count == 0))
            // Do as if the object was created
            return objectCreated(event);
        else if ((_step.isReverse() && !isImpacted(event.getValues()))
                || (!_step.isReverse() && (filter == null)))
        {
            // It it was not in our list, ignore this event
            if (count == 0)
                return false;

            // Otherwise, notify the modification
            changeObject(object, event.getValues(), event.getEventInfo());
            return true;
        }

        LySimpleObjectList list = getObjects(object);
        int                newCount = 0;
        int                i, n = list.getSize();
        LyObject           parent;

        for (i = 0; i < n; i++)
        {
            parent = list.getObject(i);
            if (_parentList.contains(parent))
                newCount++;
        }

        if ((count == 0) && (newCount == 0))
            // Not interesting in modification of unknown objects
            return false;

        if (newCount != count)
            setCount(object, newCount);
        else if (_step.isReverse())
        {
            counterChanged(object, false);
            counterChanged(object, true);
        }

        if ((count != 0)
                && (indexOf(object) >= 0))
            // Notify the change, the list may need to be resorted
            changeObject(object, event.getValues(), event.getEventInfo());

        return true;
    } // end method objectModified

    /* LyWorkSpaceListener */
    /**
     * Internal class that implements LyWorkSpaceListener used to receive events on parent
     * workspace.
     */
    class _LyWorkSpaceListener
        implements LyWorkSpaceListener,
            LyWorkSpaceCounterListener,
            LyGetWorkSpaceInterface
    {

        /**
         * Called when an object is added in the parent workspace.
         *
         * @param   event  Event notifying object add.
         *
         * @return  True if the event is used by the listener, false otherwise.
         */
        public boolean objectAdded(LyWorkSpaceEvent event)
        {
            LyObject parent = event.getObject();

            if (!isRelated(parent.getClassInfo()))
                return false;

            // Add related objects to our result list
            // Remove related objects from our result list
            LySimpleObjectList list;

            if (_step.isReverse())
            {
                list = new LySimpleObjectList(true);
                list.setSorted(_sorted);
                getReversedObjects(list, parent, _step.getTarget(), _step.getFilter(), null);
            }
            else
                list = getObjects(parent);

            LyObject object;
            int      i, n = list.getSize();

            for (i = 0, n = list.getSize(); i < n; i++)
            {
                object = list.getObject(i);
                incrementCount(object);
            }

            return true;
        } // end method objectAdded

        /**
         * Called when an object is removed from the parent workspace.
         *
         * @param   event  Event notifying object remove.
         *
         * @return  True if the event is used by the listener, false otherwise.
         */
        public boolean objectRemoved(LyWorkSpaceEvent event)
        {
            LyObject parent = event.getObject();

            if (!isRelated(parent.getClassInfo()))
                return false;

            // Remove related objects from our result list
            LySimpleObjectList list;

            if (_step.isReverse())
            {
                list = new LySimpleObjectList(true);
                list.setSorted(_sorted);
                getReversedObjects(list, parent, _step.getTarget(), null, LyStepWorkSpace.this);
            }
            else
                list = getObjects(parent);

            LyObject object;
            int      i, n = list.getSize();

            for (i = 0, n = list.getSize(); i < n; i++)
            {
                object = list.getObject(i);
                decrementCount(object);
            }

            return true;
        } // end method objectRemoved

        /**
         * Called when an object has changed in the parent workspace.
         *
         * @param   event  Event notifying object change.
         *
         * @return  True if the event is used by the listener, false otherwise.
         */
        public boolean objectChanged(LyWorkSpaceEvent event)
        {
            LyValueSet values = event.getValues();

            if (values == null)
                return false;

            if (_step.isReverse())
                return false;

            LyObject parent = event.getObject();

            if (!isRelated(parent.getClassInfo())
                    || !isImpacted(event.getValues()))
                return false;

            LyObjectList list = getRelatedObjects(values, null);
            LyObjectList newList = getRelatedObjects(values, parent);
            LyObject     object;
            int          i, n;

            // Remove objects of the old list not found in the new one
            if (list != null)
            {
                for (i = 0, n = list.getSize(); i < n; i++)
                {
                    object = list.getObject(i);
                    if ((newList == null)
                            || !newList.contains(object))
                        decrementCount(object);
                }
            }

            // Add objects of the new list not found in the old one
            if (newList != null)
            {
                for (i = 0, n = newList.getSize(); i < n; i++)
                {
                    object = newList.getObject(i);
                    if ((list == null)
                            || !list.contains(object)
                            || (indexOf(object) < 0)) // To detect some inconsistency
                        incrementCount(object);
                }
            }

            return true;
        } // end method objectChanged

        /**
         * Called when the counter associated with the specified object has been incremented or
         * decremented.
         *
         * @param   event  Event notifying counter change.
         *
         * @return  True if the event is used by the listener, false otherwise.
         */
        public boolean counterChanged(LyWorkSpaceEvent event)
        {
            boolean incremented = event.isIncremented();

            if (incremented)
                return objectAdded(event);
            else
                return objectRemoved(event);
        }

        private boolean isRelated(LyClassInfo classInfo)
        {
            if ((classInfo == null) || (_parentClass == null))
                return false;

            if ((_parentClass == classInfo)
                    || _parentClass.inheritsFrom(classInfo)
                    || classInfo.inheritsFrom(_parentClass))
                return true;

            return false;
        }

		public LyWorkSpace getWorkSpace()
		{
			return LyStepWorkSpace.this;
		}
	} // end class _LyWorkSpaceListener

} // end class LyStepWorkSpace
