/*
 * LySubWorkSpace.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;

import leon.app.*;
import leon.info.*;


/**
 * This class implements a sub-workSpace : workspace for calculating a sub-list from an initial
 * objects list. Filtering is either done on objects class info, or with a filter or both.
 *
 * @see LyClassInfo
 * @see LyFilter
 */
public final class LySubWorkSpace extends LyWorkSpace
{
	/** File version */
	public static final String __VERSION =
		"$Id: LySubWorkSpace.java 41187 2012-10-15 12:18:25Z bangiole $";

	/** Parent objects list that is filtered by this workspace. */
	LyObjectList			   _parentList;

	/** Class info to which objects of this workspace have to belong (may be null). */
	LyClassInfo				   _classInfo;

	/** Filter applied for objects loaded in this workspace (may be null). */
	LyFilter				   _filter;

	/** Internal listener to receive events on parent list. */
	LyWorkSpaceListener		   _listener;

	/**
	 * Constructs a sub-workspace within a given session filtered by a class info and/or a specified
	 * filter.
	 *
	 * @param     session    Session in which the workspace is constructed.
	 * @param     objectList Objects list that is filtered by this sub-workspace.
	 * @param     classInfo  Class info to which objects of this workspace have to belong (may be
	 *                       null).
	 * @param     filter     Filter applied for objects loaded in this workspace (may be null).
	 * @param     dynamic    Specifies if this workspace is dynamic or not.
	 * @param     local      Specifies if this workspace is local or not.
	 * @exception IllegalArgumentException If the constructed sub-workspace is dynamic and the given
	 *                                     objects list is not.
	 */
	public LySubWorkSpace(LyDataSession session, LyObjectList objectList,
		LyClassInfo classInfo, LyFilter filter, boolean dynamic, boolean local)
	{
		super(session, dynamic, local);

		if (objectList == null)
			throw new IllegalStateException("BF AppEngine : parentList can't be null");

		setSorted(objectList.isSorted());
		// Initializing the attributes.
		_parentList = objectList;
		_classInfo = classInfo;
		_filter = filter;

		if (dynamic)
		{
			if (!_parentList.isDynamic()
					|| !(_parentList instanceof LyWorkSpace))
				throw new IllegalArgumentException(
					"BF AppEngine : can't build a dynamic sub-workspace on a non dynamic objects list");

			// Listen to workspace events
			_listener = new _LyAdvancedWorkSpaceListener();
			((LyWorkSpace)_parentList).addListener(_listener);

			createFilterListeners(_filter);
		}

		// Initial loading
		reset();
	}

	/**
	 * Constructs a sub-workspace within a given session filtered by a class info and/or a specified
	 * filter.
	 *
	 * @param     session    Session in which the workspace is constructed.
	 * @param     objectList Objects list that is filtered by this sub-workspace.
	 * @param     classInfo  Class info to which objects of this workspace have to belong (may be
	 *                       null).
	 * @param     filter     Filter applied for objects loaded in this workspace (may be null).
	 * @param     dynamic    Specifies if this workspace is dynamic or not.
	 * @exception IllegalArgumentException If the constructed sub-workspace is dynamic and the given
	 *                                     objects list is not.
	 */
	public LySubWorkSpace(LyDataSession session, LyObjectList objectList,
		LyClassInfo classInfo, LyFilter filter, boolean dynamic)
	{
		this(session, objectList, classInfo, filter, dynamic, false);
	}

	/**
	 * Constructs a sub-workspace within a given session filtered by a class info.
	 *
	 * @param     session    Session in which the workspace is constructed.
	 * @param     objectList Objects list that is filtered by this sub-workspace.
	 * @param     classInfo  Class info to which objects of this workspace have to belong (may be
	 *                       null).
	 * @param     dynamic    Specifies if this workspace is dynamic or not.
	 * @exception IllegalArgumentException If the constructed sub-workspace is dynamic and the given
	 *                                     objects list is not.
	 */
	public LySubWorkSpace(LyDataSession session, LyObjectList objectList,
		LyClassInfo classInfo, boolean dynamic)
	{
		this(session, objectList, classInfo, null, dynamic);
	}

	/**
	 * Constructs a sub-workspace within a given session filtered by the specified filter.
	 *
	 * @param     session    Session in which the workspace is constructed.
	 * @param     objectList Objects list that is filtered by this sub-workspace.
	 * @param     filter     Filter applied for objects loaded in this workspace (may be null).
	 * @param     dynamic    Specifies if this workspace is dynamic or not.
	 * @exception IllegalArgumentException If the constructed sub-workspace is dynamic and the given
	 *                                     objects list is not.
	 */
	public LySubWorkSpace(LyDataSession session, LyObjectList objectList, LyFilter filter,
		boolean dynamic)
	{
		this(session, objectList, null, filter, dynamic);
	}

	/**
	 * Free this workspace. Remove this workspace as listener of the parent workspace.
	 */
	protected void free()
	{
		if (_parentList instanceof LyWorkSpace)
			((LyWorkSpace)_parentList).free(_listener);

		super.free();
	}

	/**
	 * Gets the objects list that is filtered by this sub-workspace.
	 *
	 * @return Objects list filtered by this sub-workspace.
	 */
	public LyObjectList getParentList()
	{
		return _parentList;
	}

	/**
	 * Gets class info to which objects of this workspace belong.
	 *
	 * @return Class info specified for this workspace (may be null).
	 */
	public LyClassInfo getClassInfo()
	{
		return _classInfo;
	}

	/**
	 * Gets the filter of this sub-workspace.
	 *
	 * @return Filter specified for this sub-workspace (may be null).
	 */
	public LyFilter getFilter()
	{
		return _filter;
	}

	/**
	 * Return the list of all class infos to which objects of this workspace belong.
	 *
	 * @return List of classes to which objects of this workspace belong.
	 */
	public LyInfoList<LyClassInfo> getClasses()
	{
		if (_classInfo == null)
			return _parentList.getClasses();

		LyInfoList<LyClassInfo> classes = new LyInfoList<LyClassInfo>();

		classes.addInfo(_classInfo);
		return classes;
	}

	/**
	 * Changes the filter of this sub-workspace. On this call, the workspace is resetted.
	 *
	 * @param filter New filter specified for this sub-workspace.
	 */
	public void setFilter(LyFilter filter)
	{
		_filter = filter;

		reset();
	}

	/**
	 * Method for resetting this workspace. Objects contained in this workspace are removed and
	 * reloaded from parent list.
	 */
	public synchronized void reset()
	{
		// Remove all objects.
		super.reset();

		// Get from parent list all objects of the given classInfo
		// matching the filter if any
		LyObjectList list = null;

		if (_filter != null)
		{
			if (_parentList instanceof LyClassWorkSpace)
			{
				// Optimisation : try to filter directly from source if its cache is not loaded
				LyClassWorkSpace	    parent = (LyClassWorkSpace)_parentList;
				LyClassInfo			    classInfo = ((_classInfo != null) ? _classInfo
																		  : parent.getClassInfo());
				LyInfoList<LyFieldInfo> fieldInfos = parent.getFieldInfos();
				LyFilter			    filter, parentFilter = parent.getFilter();

				if (fieldInfos != null)
					fieldInfos.addList(_filter.getFieldInfos());

				if (parentFilter == null)
					filter = _filter;
				else
				{
					filter = new LyFilter();
					filter.addElement(parentFilter);
					filter.addElement(_filter);
				}

				list = new LyClassWorkSpace(_session, classInfo, fieldInfos, filter,
						parent.getSort(), false, parent.isSorted(), parent.isLocal());
			}

			if (list == null)
			{
			    if (_classInfo != null)
			        list = _parentList.getList(_session, _classInfo);
			    else
			        list = _parentList;

				Vector<LyObject> objects = _parentList.getObjects();

				if (objects == null)
					return;

				list = _filter.match(objects.elements(), null);
			}
		}
		else
			list = _parentList;

		Vector<LyObject> oldObjects = getObjects();
		Vector<LyObject> objects = new Vector<LyObject>();
		Vector<LyObject> parentObjects =  (list == null ? null : list.getObjects());
		LyObject		 object;
		int				 i, n = (parentObjects == null ? 0 : parentObjects.size());

		for (i = 0; i < n; i++)
		{
			object = parentObjects.get(i);
			if ((_classInfo != null) && !object.instanceOf(_classInfo))
				continue;

			objects.addElement(object);
			if ((oldObjects == null) || !oldObjects.contains(object))
				addObject(object);
		}

		n = ((oldObjects == null) ? 0 : oldObjects.size());
		for (i = 0; i < n; i++)
		{
			object = oldObjects.elementAt(i);
			if (!objects.contains(object))
				removeObject(object);
		}
	} // end method reset

	/**
	 * Returns a string representation of this object.
	 *
	 * @return This workspace returned as a string.
	 */
	public String toString()
	{
		return super.toString() + "\nclassInfo=" + _classInfo + "\nparent=" + _parentList;
	}

	/* Implementation of LyWorkSpaceListener */

	class _LyAdvancedWorkSpaceListener
		implements LyAdvancedWorkSpaceListener, LyGetWorkSpaceInterface
	{

		/**
		 * Called when an object is added in the workspace.
		 *
		 * @param  event Event notifying object add.
		 * @return True if the event is used by the listener, false otherwise.
		 */
		public boolean objectAdded(LyWorkSpaceEvent event)
		{
			LyObject object = event.getObject();

			// If a class info is specified, the object must belong to it
			if ((_classInfo != null) && (!object.instanceOf(_classInfo)))
				return false;

			// If there is a filter, the object must match it.
			if ((_filter != null) && !_filter.match(object))
			{
				// Post an event to force update of views if needed
				LyWorkSpaceEvent wsEvent = new LyWorkSpaceEvent(LySubWorkSpace.this,
						LyWorkSpaceEvent.Type.ADD, object, -1);

				wsEvent.setEventInfo(event.getEventInfo());

				LySubWorkSpace.this.notify(wsEvent);
				return false;
			}

			addObject(object);
			return true;
		}

		/**
		 * Called when an object is removed from the workspace.
		 *
		 * @param  event Event notifying object remove.
		 * @return True if the event is used by the listener, false otherwise.
		 */
		public boolean objectRemoved(LyWorkSpaceEvent event)
		{
			LyObject object = event.getObject();

			if (indexOf(object) < 0)
			{
				// Post an event to force update of views if needed
				LyWorkSpaceEvent wsEvent = new LyWorkSpaceEvent(LySubWorkSpace.this,
						LyWorkSpaceEvent.Type.REMOVE, object, -1);

				wsEvent.setEventInfo(event.getEventInfo());

				LySubWorkSpace.this.notify(wsEvent);
				return false;
			}

			removeObject(object);
			return true;
		}

		/**
		 * Called when an object has changed in the workspace.
		 *
		 * @param  event Event notifying object change.
		 * @return True if the event is used by the listener, false otherwise.
		 */
		public boolean objectChanged(LyWorkSpaceEvent event)
		{
			if (event.getValues() == null)
				// It's a sort, don't check object
				return true;

			LyObject object = event.getObject();

			// Ignore objects that don't belong to the specified class info
			if ((_classInfo != null) && (!object.instanceOf(_classInfo)))
				return false;

			// If the object was not in the list and match the filter, add it
			if (indexOf(object) < 0)
			{
				if ((_filter != null) && _filter.match(object))
				{
					addObject(object);
					return true;
				}

				return false;
			}

			// If the object is in the list, but doesn't match the filter anymore,
			// remove it.
			if ((_filter != null) && !_filter.match(object))
			{
				removeObject(object);
				return true;
			}

			// Notify the change, the list may need to be resorted
			changeObject(object, event.getValues(), event.getEventInfo());
			return true;
		} // end method objectChanged

		/**
		 * Called when an object state is changed.
		 *
		 * @param  event Event notifying object state change.
		 * @return True if the event is used by the listener, false otherwise.
		 */
		public boolean objectStateChanged(LyAdvancedWorkSpaceEvent event)
		{
			// If the object's event is added or removed from the workspace
			// we let the addObject or removeObject method do the job.
			if (useEvent(event))
				return true;

			// Notify the change, the list may need to be resorted
			// Post an event to force update of views if needed
	         LyObject        object = event.getObject();
			LyWorkSpaceEvent wsEvent = new LyAdvancedWorkSpaceEvent(LySubWorkSpace.this,
					event.getType(), object, getIndex(object), event.getOldState());

			wsEvent.setEventInfo(event.getEventInfo());

			LySubWorkSpace.this.notify(wsEvent);
			return true;
		}

		/**
		 * Method for receiving user event on an object.
		 *
		 * @param  event Event notifying the user event of an object.
		 * @return True indicating if the event was used, false otherwise.
		 */
		public boolean objectUserEvent(LyAdvancedWorkSpaceEvent event)
		{
			// If the object's event is added or removed from the workspace
			// we let the addObject or removeObject method do the job.
			if (useEvent(event))
				return true;

			// Notify the change, the list may need to be resorted
			// Post an event to force update of views if needed
			LyWorkSpaceEvent wsEvent;
	        LyObject         object = event.getObject();

			wsEvent = new LyAdvancedWorkSpaceEvent(LySubWorkSpace.this, event.getType(), object,
					getIndex(object), event.getOrigin());

			wsEvent.setEventInfo(event.getEventInfo());

			LySubWorkSpace.this.notify(wsEvent);
			return true;
		}

		/**
		 * Called when grouped events are sent.
		 *
		 * @param  events Sorted list of events.
		 * @return True if the events are used by the listener, false otherwise.
		 */
        public boolean workSpaceEventsGrouped(ArrayList<LyWorkSpaceEvent> events)
		{
			if ((events == null) || (events.isEmpty()))
				return false;

			synchronized (this)
			{
				// Change LyDataSourceEvents for LyWorkSpaceEvents if events are useful and send them.
		        ArrayList<LyWorkSpaceEvent>   stockedEvents = new ArrayList<LyWorkSpaceEvent>();

		        if (_stockedEvents == null)
		            _stockedEvents = new Hashtable<Thread, ArrayList<LyWorkSpaceEvent>>();
		        _stockedEvents.put(Thread.currentThread(), stockedEvents);

				int				        i, n = events.size();
				LyWorkSpaceEvent        event;
                LyWorkSpaceEvent.Type   type;

				for (i = 0; i < n; i++)
				{
					event = (LyWorkSpaceEvent)events.get(i);
					type = event.getType();
					switch (type)
					{
					case ADD:
						objectAdded(event);
						break;
					case REMOVE:
						objectRemoved(event);
						break;
					case CHANGE:
						objectChanged(event);
						break;
					case SELECTION_CHANGE:
					case VISIBILITY_CHANGE:
						objectStateChanged((LyAdvancedWorkSpaceEvent)event);
						break;
					case USER_EVENT:
						objectUserEvent((LyAdvancedWorkSpaceEvent)event);
						break;
					default:
						break;
					}
				}

		        if (stockedEvents.isEmpty())
		        {
		            _stockedEvents.remove(Thread.currentThread());
		            return false;
		        }
		        else
		        {
		            ArrayList<LyWorkSpaceEvent> wsEvents = stockedEvents;

		            _stockedEvents.remove(Thread.currentThread());
					return LySubWorkSpace.this.notify(wsEvents);
				}
			}
		} // end method workSpaceEventsGrouped

		protected boolean useEvent(LyAdvancedWorkSpaceEvent event)
		{
			LyObject object = event.getObject();

			// Ignore objects that don't belong to the specified class info
			if ((_classInfo != null) && (!object.instanceOf(_classInfo)))
				return false;

			// If the object was not in the list and match the filter, add it
			if (indexOf(object) < 0)
			{
				if ((_filter != null) && _filter.match(object))
				{
					addObject(object);
					return true;
				}

				return false;
			}

			// If the object is in the list, but doesn't match the filter anymore,
			// remove it.
			if ((_filter != null) && !_filter.match(object))
			{
				removeObject(object);
				return true;
			}

			return false;
		} // end method useEvent

		/**
		 * Called when the counter associated with the specified object has been incremented or
		 * decremented.
		 *
		 * @param  event Event notifying counter change.
		 * @return True if the event is used by the listener, false otherwise.
		 */
		public boolean counterChanged(LyWorkSpaceEvent event)
		{
			LyObject object = event.getObject();

			if ((_filter != null) && _filter.match(object))
			{
				boolean			 incremented = event.isIncremented();
				LyWorkSpaceEvent counterEvent = new LyWorkSpaceEvent(LySubWorkSpace.this,
						LyWorkSpaceEvent.Type.COUNTER_CHANGE, object, getIndex(object));

				event.setIncremented(incremented);
				LySubWorkSpace.this.notify(counterEvent);
				return true;
			}
			else
				return false;
		}

		public LyWorkSpace getWorkSpace()
		{
			return LySubWorkSpace.this;
		}
	} // end class _LyAdvancedWorkSpaceListener

	/**
	 * Method for retrieving source objects that lead to given target object.
	 *
	 * @param  target Target object
	 * @return Source objects that leade to target object.
	 */
	public LySimpleObjectList getSources(LyObject target)
	{
		if (contains(target))
			return new LySimpleObjectList(target);
		else
			return null;
	}
} // end class LySubWorkSpace
