/*
 * LyValue.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

/**
 * This class is a field value, it implements a memory representation of a 'pair' [field, value].
 *
 * <P>This class is used when giving a set of values (LyValueSet) to create or modify an object.</P>
 *
 * <p>This method is abstract. It has two main inherited sub-classes :</p>
 *
 * <ul>
 *   <LI>LyFieldInfoValue, represents a "logical" field value (value corresponding to a field info,
 *     LyFieldInfo),</LI>
 *   <LI>LyPeerFieldValue, represents a "physical" field value (value corresponding to a peer field,
 *     LyPeerField).</LI>
 * </ul>
 *
 * @see LyField
 * @see leon.info.LyFieldInfo
 * @see LyFieldInfoValue
 * @see leon.peer.LyPeerField
 * @see leon.peer.LyPeerFieldValue
 * @see LyValueSet
 */
public abstract class LyValue
{
	/** File version */
	public static final String __VERSION = "$Id: LyValue.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Field to which this value corresponds to. */
	LyField					   _field;

	/** Value associated to the field. */
	Object					   _value;

	/**
	 * Constructs a field value.
	 *
	 * @param     field Field of this field value.
	 * @param     value Value of this field value.
	 */
	protected LyValue(LyField field, Object value)
	{
		_field = field;
		_value = value;
	}

	/**
	 * Gets the field for which this field value was created.
	 *
	 * @return Field of this field value.
	 */
	public final LyField getField()
	{
		return _field;
	}

	/**
	 * Gets the value that was given to the field.
	 *
	 * @return Value of this field value.
	 */
	public final Object getValue()
	{
		return _value;
	}

	/**
	 * Gets the value that was given to the field as a String. This String value is an internal
	 * value and is usually not shown to the end user. This method is overriden to be more accurate
	 * by subclass LyFieldInfoValue.
	 *
	 * @return Value of this field value as a String or null if no value is set.
	 */
	public String getString()
	{
		return ((_value == null) ? null : _value.toString());
	}

	/**
	 * Checks if this value associated to the field is null.
	 *
	 * @return Value of this field value.
	 */
	public final boolean isNull()
	{
		return (_value == null);
	}

	/**
	 * Sets the value of this field value.
	 *
	 * @param     value Value assigned to this field value.
	 * @exception IllegalArgumentException If the class of the field is not assignable from the
	 *                                     given value.
	 */
	public void setValue(Object value)
	{
		if ((value != null)
				&& !_field.getValueClass().isAssignableFrom(value.getClass()))
			throw new IllegalArgumentException("BF AppEngine : class " + _field.getValueClass()
				+ " of the field " + _field.getId()
				+ " is not assignable from class " + value.getClass()
				+ " of value " + value);

		_value = value;
	}

	/**
	 * Returns this value as a string representation for debug / dump. The format returned is
	 * LyValue(<field>=<value>). To get the String representation without text format, use
	 * getString() instead.
	 *
	 * @return String representing his field value.
	 * @see    #getString()
	 */
	public String toString()
	{
		if ((_value != null) && _value.getClass().isArray())
		{
			StringBuffer buffer = new StringBuffer();
			Object[]     array = (Object[])_value;
			int			 i, n = array.length;

			buffer.append('(');
			buffer.append(_field.getId());
			buffer.append("=[");
			buffer.append(n);
			buffer.append(": ");
			for (i = 0; i < n; i++)
			{
				if (i > 0)
					buffer.append(", ");
				buffer.append(array[i].toString());
			}
			buffer.append("])");
			return buffer.toString();
		}
		else
			return "(" + _field.getId() + "=" + _value + ")";
	}
} // end class LyValue
