/*
 * LyValueSet.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;

import leon.app.*;
import leon.info.*;


/**
 * This class is used to hold a list of values (LyValue).
 *
 * <p>It allows to give a list of 'values' when creating or modifying an object, or to get a list of
 * values from an object.</p>
 *
 * @see LyValue
 */
public class LyValueSet extends Vector<LyValue>
{
	/**
	 * 
	 */
	private static final long serialVersionUID = 2611763305982771238L;

	/** File version */
	public static final String __VERSION =
		"$Id: LyValueSet.java 35859 2012-06-14 07:43:02Z jlsaouli $";

	/** Class for which this value set is built. */
	LyClass        _class;
	
    /** Boolean indicating if checking is performed when adding values. */
	boolean        _checkFields;
    
	/**
	 * Constructs an empty values set.
	 */
	public LyValueSet()
	{
	    this(true);
	}
    
    /**
     * Constructs an empty values set.
     * @param checkFields value of the checkfield properties
     */
    public LyValueSet(boolean checkFields)
    {
        // Create a Vector for one element to allow synchronization on this instance
        super(1);
        
        _class = null;
        _checkFields = checkFields;
    }
    
    /**
     * Constructs a value set for a given class.
     * @param valuesClass a LyClass
     */
    public LyValueSet(LyClass valuesClass)
    {
        this(true);
        
        _class = valuesClass;
        _checkFields = true;
    }
    
    /**
     * Constructs a values set from an existing one.
     * @param values a list of values
     */
    public LyValueSet(LyValueSet values)
    {
        super(values);
        
        _class = null;
        _checkFields = values._checkFields;
    }
    
    /**
     * Direct access to value at the given position (no control is made).
     *
     * @param  i Index of requested value.
     * @return Value found.
     */
    final synchronized LyValue getValue(int i)
    {
        return (LyValue)elementData[i];
    }

    /**
     * Direct setting of value at the given position (no control is made).
     *
     * @param object Value set.
     * @param i      Index of the added value.
     */
    final synchronized void setValueAt(LyValue value, int i)
    {
        elementData[i] = value;
    }
    
	/**
	 * Retrieves a field value (LyValue) in this values set for a given field.
	 *
	 * @param  field Field for which a field value in searched in this values set.
	 * @return The field value corresponding to the given field if it's found, null otherwise.
	 */
	public synchronized LyValue getFieldValue(LyField field)
	{
		if (field == null)
			return null;

		int     i, n = size();
		LyValue value;

		for (i = 0; i < n; i++)
		{
			value = getValue(i);
			if (value.getField() == field)
				return value;
		}

		return null;
	}

	/**
	 * Retrieves a field value (LyValue) in this values set for a given field.
	 *
	 * @param  fieldInfo         Field info for which a field value in searched in this values set.
	 * @param  allowAccessByType Allow access by type. If true, value may be retrieved from
	 * a template field of given field info.
	 * @return The field value corresponding to the given field if it's found, null otherwise.
	 */
	public synchronized LyValue getFieldValue(LyFieldInfo fieldInfo, boolean allowAccessByType)
	{
		if (fieldInfo == null)
			return null;

		int     i, n = size();
		LyValue value;
		LyField field;

		for (i = 0; i < n; i++)
		{
			value = getValue(i);
			field = value.getField();
			if ((field == fieldInfo)
					|| (allowAccessByType
						&& (field instanceof LyFieldInfo)
						&& ((LyFieldInfo)field).hasTemplate(fieldInfo)))
				return value;
		}

		return null;
	}

	/**
	 * Retrieves a field value (LyValue) in this values set for a given field.
	 *
	 * @param  id Field id for which a field value in searched in this values set.
	 * @return The field value corresponding to the given field if it's found, null otherwise.
	 */
	public synchronized LyValue getFieldValue(String id)
	{
		if (id == null)
			return null;

		int     i, n = size();
		LyValue value;

		for (i = 0; i < n; i++)
		{
			value = getValue(i);
			if (value.getField().getId().equals(id))
				return value;
		}

		return null;
	}

	/**
	 * Retrieves a field value (LyValue) in this values set at the given index.
	 *
	 * @param     index Index of the searched field value.
	 * @return    Field value at the specified index.
	 * @exception ArrayIndexOutOfBoundsException If the given index is negative.
	 * @exception ArrayIndexOutOfBoundsException If the given index is greater or equals to the
	 *                                           number of values of the set.
	 */
	public synchronized LyValue getFieldValue(int index)
	{
		LyValue value;

		if (index < 0)
			throw new ArrayIndexOutOfBoundsException("BF AppEngine : index out of bounds "
				+ index + " < 0");
		if (index >= size())
			throw new ArrayIndexOutOfBoundsException("BF AppEngine : index out of bounds "
				+ index + " > " + size());

		value = getValue(index);
		return value;
	}

	/**
	 * Gets the number of field values in the values set.
	 *
	 * @return Number of field values.
	 */
	public int getSize()
	{
		return size();
	}

	/**
	 * Adds a new field value in the set. If a field value with the same field already exists in
	 * this values set the value of this field value is overriden with the given value. Otherwise, a
	 * new field value is built (using buildFieldValue method of LyField) and added to the values
	 * set.
	 *
	 * @param value Field value to add to this values set.
	 */
	public synchronized void addValue(LyValue value)
	{
        LyField    field = value.getField();
        
        if ((_class != null)
                && (_class.getField(field.getId()) == null))
            throw new IllegalArgumentException("BF AppEngine : invalid argument " + field
                + ", classValueSet can only contains value for class " + _class);
        
        if (_checkFields)
        {
    		// Do we already have a field value concerning the same field ?
    		LyValue fieldValue;
    		int     i, n = size();
    
    		for (i = 0; i < n; i++)
    		{
    			fieldValue = getValue(i);
    			if ((fieldValue != null)
                     && (fieldValue.getField() == value.getField()))
    			{
    				// Replace value of the field value
    				setValueAt(value, i);
    				return;
    			}
    		}
        }

		// Add value
		addElement(value);
	}

	/**
	 * Adds a new pair (field, value) in the values set.
	 *
	 * @param field Field for which a field value is added.
	 * @param value Value assigned to the given field.
	 */
	public void addValue(LyField field, Object value)
	{
		if (field == null)
			return;

		// Build field value
		LyValue fieldValue = field.buildFieldValue(value);

		// Add it to the values set
		addValue(fieldValue);
	}

	/**
	 * Adds a set of values in this set.
	 *
	 * @param values Set of values added in this set.
	 */
	public synchronized void addValues(LyValueSet values)
	{
		if (values == null)
			return;

		LyValue value;
		int     i, n = values.getSize();

		for (i = 0; i < n; i++)
		{
			value = values.getValue(i);
			addValue(value);
		}
	}

	/**
	 * Removes the first field value corresponding to a given field from the set.
	 *
	 * @param  field Field for which a field value has to be removed.
	 * @return The removed field value if one corresponding to the given field was found, null
	 *         otherwise.
	 */
	public synchronized LyValue removeValue(LyField field)
	{
		LyValue value;
		int     i, n;

		for (i = 0, n = size(); i < n; i++)
		{
			value = getValue(i);
			if (value.getField() == field)
			{
				removeElementAt(i);
				return value;
			}
		}

		return null;
	}

	/**
	 * Checks if a field value corresponding to the given field is present in this values set.
	 *
	 * @param  field Field for which an existing field value is searched.
	 * @return True if a field value for the given field was found, false otherwise.
	 */
	public synchronized boolean contains(LyField field)
	{
		int     i, n = size();
		LyValue value;

		for (i = 0; i < n; i++)
		{
			value = getValue(i);
			if (value.getField() == field)
				return true;
		}

		return false;
	}

	/**
	 * Checks if a field value corresponding to the given field is present in this values set.
	 *
	 * @param  fieldInfo         Field for which an existing field value is searched.
	 * @param  allowAccessByType Allow access by type.
	 * @return True if a field value for the given field was found, false otherwise.
	 */
	public synchronized boolean contains(LyFieldInfo fieldInfo, boolean allowAccessByType)
	{
		int     i, n = size();
		LyValue value;
		LyField field;

		for (i = 0; i < n; i++)
		{
			value = getValue(i);
			field = value.getField();
			if ((field == fieldInfo)
					|| (allowAccessByType
						&& (field instanceof LyFieldInfo)
						&& ((LyFieldInfo)field).hasTemplate(fieldInfo)))
				return true;
		}

		return false;
	}

	/**
	 * Checks if the value set contains at least one field value corresponding to the given field
	 * info that has the specified mark. This method does only make sense if the values set contains
	 * logical values (LyFieldInfoValue).
	 *
	 * @param  mark Searched mark.
	 * @return True if a field value was found, false otherwise.
	 */
	public synchronized boolean hasMark(int mark)
	{
		int     i, n = size();
		LyValue value;
		LyField field;

		for (i = 0; i < n; i++)
		{
			value = getValue(i);
			field = value.getField();

			if ((field instanceof LyFieldInfo)
					&& ((LyFieldInfo)field).hasMark(mark))
				return true;
		}

		return false;
	}

	/**
	 * Checks if the value set contains at least one field value corresponding to the given field
	 * info that has the specified marks. This method does only make sense if the values set
	 * contains logical values (LyFieldInfoValue).
	 *
	 * @param  marks    Searched marks.
	 * @param  matchAll Specifies if this instance have to match all specified marks or only one.
	 * @return True if a field value was found, false otherwise.
	 */
	public synchronized boolean hasMarks(BitSet marks, boolean matchAll)
	{
		if (marks == null)
			return true;

		int     i, n = size();
		LyValue value;
		LyField field;

		for (i = 0; i < n; i++)
		{
			value = getValue(i);
			field = value.getField();

			if ((field instanceof LyFieldInfo)
					&& ((LyFieldInfo)field).hasMarks(marks, matchAll))
				return true;
		}

		return false;
	}

	/**
	 * Dumps content of this value set in a string.
	 * @return a string
	 */
	public synchronized String toString()
	{
		StringBuffer buffer = new StringBuffer();

		buffer.append("{");

		int     i, n = size();
		LyValue value;

		buffer.append(n);
		buffer.append(": ");

		for (i = 0; i < n; i++)
		{
			value = getValue(i);
			if (i > 0)
				buffer.append("\n, ");
			buffer.append(value.toString());
		}

		buffer.append(" }");
		return buffer.toString();
	}
    
    /**
     *  Gets a map containing pairs of (field, value) of this values set.
     * @return a map
     */
    public synchronized Map<LyField, LyValue> getMap()
    {
        Map<LyField, LyValue>   map = new HashMap<LyField, LyValue>();
        LyValue                 value;
        int                     i, n = size();
        
        for (i = 0; i < n; i++)
        {
            value = getValue(i);
            map.put(value.getField(), value);
        }
            
        return map;
    }
    
    /**
     * Enable the check state during the addition of values
     */
    public void enableCheckField()
    {
    	_checkFields = true; 
    }
    
    /**
     * Retrieves application of the class info of this object.
     *
     * @return Application of the class info of this object.
     */
    public final LyApplication getApplication()
    {
        return (_class instanceof LyClassInfo ? ((LyClassInfo)_class).getApplication() : null);
    }
} // end class LyValueSet
