/*
 * LyWorkSpace.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

import java.util.*;

import leon.app.*;
import leon.info.*;
import leon.info.misc.*;
import leon.misc.*;


/**
 * This class implements workspaces. Workspaces are dynamic objects lists : lists that send events
 * to listeners when their content changes.
 *
 * <p>A workspace is constructed within a session and is only accessible in that session.</p>
 *
 * <p>Different types of workspaces are provided, they are implemented as subclasses of this
 * abstract class and correspond to different needs (see inherited classes).</p>
 *
 * <p>A workspace can be local, that is all objects are searched from memory.</p>
 *
 * <p>When workspaces are not used anymore, they should be freed via free method to release
 * resources they use to keep themselves up to date. This point is rather important because if they
 * are not freed, they can't be removed by garbage collector and may still use memory and CPU for
 * nothing.</p>
 *
 * @see LyWorkSpaceEvent
 * @see LyAdvancedWorkSpaceEvent
 * @see LyWorkSpaceListener
 * @see LyAdvancedWorkSpaceListener
 */
public abstract class LyWorkSpace extends LyObjectList
	implements LyAdvancedDataSourceListener,
		LyGetSessionInterface
{
	public static final String				   __VERSION =
		"$Id: LyWorkSpace.java 33730 2012-01-06 15:18:17Z bangiole $";

	/** Listeners of the workspace. */
	protected Vector<LyWorkSpaceListener>	   _listeners;

    /** Listeners of the workspace Lock. */
    protected Object						   _listenersLock = new Object();

	/** Boolean indicating if this workspace is dynamic or not. */
	protected boolean						   _dynamic;

	/** Session to which the workspace belongs. */
	protected LyDataSession					   _session;

	/** Counter for locks (can't be freed if greater than 0). */
	protected short							   _lockCount;

	/** Boolean indicating if this workspace is local or not. */
	protected boolean						   _local;

	/** Filter listeners of the workspace. */
	protected Vector<LyFilterRelationListener> _filterListeners;

    /** Grouped events by thread. */
    Hashtable<Thread, ArrayList<LyWorkSpaceEvent>>	_stockedEvents;

	/**
	 * Constructs a workspace within a session.
	 *
	 * @param session Session in which this workspace is constructed.
	 * @param dynamic Specifies if this workspace is dynamic or not.
	 * @param local   Specifies if this workspace is local or not.
	 */
	protected LyWorkSpace(LyDataSession session, boolean dynamic, boolean local)
	{
		_session = session;
		_listeners = (dynamic ? new Vector<LyWorkSpaceListener>(1) : null);
		_dynamic = dynamic;
		_lockCount = 0;
		_local = local;
		_stockedEvents = null;
	}

	/**
	 * Remove the given listener of this workspace and free workspace if this listener was the last.
	 *
	 * @param  listener Removed listener.
	 * @return Returns a boolean indicating if the workspace was freed or not.
	 */
	public final boolean free(LyWorkSpaceListener listener)
	{
		if (isDynamic())
		{
			// Remove given listener
			if (listener != null)
				removeListener(listener);

			if ((_listeners != null) && (_listeners.size() > 0))
				// We can't free a workspace that still have listeners.
				return false;
		}

		// If locked, don't free
		if (_lockCount > 0)
			return false;

		// Free resources
		free();

		return true;
	}

	/**
	 * Internal method to free this workspace.
	 */
	protected synchronized void free()
	{
		if (_filterListeners != null)
		{
			int						 i, n = _filterListeners.size();
			LyFilterRelationListener filterListener;

			for (i = 0; i < n; i++)
			{
				filterListener = _filterListeners.elementAt(i);
				filterListener.free();
			}

			_filterListeners = null;
		}

		if (isDynamic())
		{
			_listeners = null;
			_dynamic = false;
		}

		// Free allocated resources
		unlockObjects();
		_objects = null;
		_sort = null;
		_lockCount = -1;
	} // end method free

	/**
	 * Lock this workspace. A locked workspace can't be freed.
	 *
	 * @see #unlock
	 */
	public void lock()
	{
		_lockCount++;
	}

	/**
	 * Unlock this workspace.
	 *
	 * @see #lock
	 */
	public void unlock()
	{
		_lockCount--;

		if (_lockCount == 0)
		{
			if (!isDynamic()
					|| (_listeners == null)
					|| (_listeners.size() == 0))
				free();
		}
	}

	/**
	 * Method for resetting a workspace. By default, this method clears the content of the
	 * workspace. But depending on the type of workspace, this method may recompute the entire
	 * content of the workspace.
	 */
	public synchronized void reset()
	{
		clear();
	}

	/**
	 * Clears the content of the workspace.
	 */
	protected synchronized void clear()
	{
		Vector<LyObject>  objects = getObjects();
		int		          i, n = ((objects == null) ? 0 : objects.size());
		LyObject          object;

		for (i = 0; i < n; i++)
		{
			object = objects.elementAt(i);
			removeObject(object);
		}
	}

	/**
	 * Internal method to resort this list according to the current sort parameters. Does nothing if
	 * this list is specified as not sorted.
	 */
	protected synchronized void resort()
	{
		super.resort();
	}

	/**
	 * Checks whether this workspace is dynamic or not.
	 *
	 * @return True if the workspace is dynamic, false otherwise.
	 */
	public final boolean isDynamic()
	{
		return _dynamic;
	}

	/**
	 * Checks whether this workspace is local or not.
	 *
	 * @return True if the workspace is local, false otherwise.
	 */
	public final boolean isLocal()
	{
		return _local;
	}

	/**
	 * Gets session on which this workspace was built.
	 *
	 * @return The session of this workspace.
	 */
	public final LyDataSession getSession()
	{
		return _session;
	}

	/**
	 * Gets listener at the given position.
	 */
	public LyWorkSpaceListener getListener(int index)
	{
		if (_listeners == null)
			return null;

		return _listeners.elementAt(index);
	}

	/**
	 * Gets number of listeners of this workspace.
	 */
	public int getListenerCount()
	{
		return ((_listeners == null) ? 0 : _listeners.size());
	}

	/**
	 * Swap objects at the given indexes. This method is overriden from LyObjectList to notify this
	 * change to listeners.
	 *
	 * @param lo First index.
	 * @param hi Second index.
	 */
	protected synchronized void swapObjects(int lo, int hi)
	{
		// Swap objects
		super.swapObjects(lo, hi);

		if ((_listeners == null) || (_listeners.size() == 0))
			return;

		// Notify this to listeners
		LyWorkSpaceEvent event;

		event = new LyWorkSpaceEvent(this, getObject(hi), hi, lo);
		notify(event);
	}

	/**
	 * Utility method to get an object source for a given classInfo.
	 *
	 * @param  classInfo Class info for which an object source is requested.
	 * @return Object source for the given class info.
	 */
	protected LyObjectSource getObjectSource(LyClassInfo classInfo)
	{
		LyObjectSource objectSource;

		if (_session != null)
			objectSource = _session.getObjectSource(classInfo);
		else
			objectSource = classInfo.getObjectSource();

		return objectSource;
	}

	/* Management of listeners */

	/**
	 * Adds a listener on this workspace.
	 *
	 * @param listener                 Listener to register.
	 * @param IllegalStateException    If this workspace is not dynamic.
	 * @param IllegalArgumentException If the given listener is null.
	 */
	public void addListener(LyWorkSpaceListener listener)
	{
		if (!isDynamic())
			throw new IllegalStateException("BF AppEngine : non dynamic list can't have listeners");
		if (listener == null)
			throw new IllegalArgumentException("BF AppEngine : listener can't be null");

		synchronized (_listenersLock)
		{
			if (_listeners.contains(listener))
				return;


			_listeners.addElement(listener);
		}
	}

	/**
	 * Removes a listener on this workspace.
	 *
	 * @param listener                 Listener to unregister.
	 * @param IllegalStateException    If this workspace is not dynamic or if the given listener is
	 *                                 not registered as listener of this workspace.
	 * @param IllegalArgumentException If the given listener is null.
	 */
	public void removeListener(LyWorkSpaceListener listener)
	{
		if (!isDynamic())
			throw new IllegalStateException(
				"BF AppEngine : non dynamic workspace don't have listeners");
		if (listener == null)
			throw new IllegalArgumentException("BF AppEngine : listener can't be null");

		synchronized (_listenersLock)
		{
			if (!_listeners.removeElement(listener))
				throw new IllegalStateException(
					"BF AppEngine : the given listener is not registered as listener of the workspace");
		}
	}

	/**
	 * Notifies an event to all listeners of this workspace.
	 *
	 * @param event                 Event to notify to listeners.
	 * @param IllegalStateException If this workspace is not dynamic.
	 */
	protected final void notify(LyWorkSpaceEvent event)
	{
		if (_lockCount < 0)
			// Workspace is freed
			return;

		if (!isDynamic())
			throw new IllegalStateException(
				"BF AppEngine : non dynamic list can't notify events since they don't have listeners");

        if (_stockedEvents != null)
        {
            ArrayList<LyWorkSpaceEvent> stockedEvents = _stockedEvents.get(Thread.currentThread());

            if (stockedEvents != null)
            {
                stockedEvents.add(event);
                return;
            }
        }

		if (_listeners.size() == 0)
			return;

		Vector<LyWorkSpaceListener> listeners;

		synchronized (_listenersLock)
		{
			listeners = new Vector<LyWorkSpaceListener>(_listeners);
		}

		LyWorkSpaceListener listener;
		int				    i, n = listeners.size();
		LyEnvironment	    environment = ((_session != null) ? _session.getEnvironment() : null);
		boolean			    repostGuiEvents = (environment != null)
				&& environment.repostGuiEvents();

		for (i = 0; i < n; i++)
		{
			listener = listeners.elementAt(i);

			// Check if something changes during loop
			if (_listeners == null)
				break;
			if (!_listeners.contains(listener))
				continue;

			if ((listener instanceof LyGuiListener)
					&& ((LyGuiListener)listener).isIdle())
				continue;

			if (repostGuiEvents
					&& (listener instanceof LyGuiListener))
				environment.execGui(new _LyGuiNotifier(listener, event));
			else
				notifyListener(listener, event);
		}
	} // end method notify

	/**
	 * Internal method used to notify a listener.
	 */
	protected boolean notifyListener(LyWorkSpaceListener listener, LyWorkSpaceEvent event)
	{
        LyWorkSpaceEvent.Type type = event.getType();
		boolean               used = false;
		LyApplication         application = (_session != null ? _session.getApplication() : null);
		LyEvaluator           evaluator = null;

		try
		{
		    if ((application != null)
		            && (listener instanceof LyEvaluator))
		        evaluator = application.setCurrentEvaluator((LyEvaluator)listener);
		        
    		if (event instanceof LyAdvancedWorkSpaceEvent)
    		{
    			if (listener instanceof LyAdvancedWorkSpaceListener)
    			{
    				if (type == LyWorkSpaceEvent.Type.USER_EVENT)
    					used = ((LyAdvancedWorkSpaceListener)listener).objectUserEvent(
    							(LyAdvancedWorkSpaceEvent)event);
    				else
    					used = ((LyAdvancedWorkSpaceListener)listener).objectStateChanged(
    							(LyAdvancedWorkSpaceEvent)event);
    			}
    		}
    		else if (type == LyWorkSpaceEvent.Type.ADD)
    			used = listener.objectAdded(event);
    		else if (type == LyWorkSpaceEvent.Type.CHANGE)
    			used = listener.objectChanged(event);
    		else if (type == LyWorkSpaceEvent.Type.REMOVE)
    			used = listener.objectRemoved(event);
    		else if (type == LyWorkSpaceEvent.Type.COUNTER_CHANGE)
    		{
    			if (listener instanceof LyWorkSpaceCounterListener)
    				used = ((LyWorkSpaceCounterListener)listener).counterChanged(event);
    		}
		}
		finally
		{
	          if ((application != null)
	                  && (evaluator != null))
	              application.setCurrentEvaluator(evaluator);
		}

		return used;
	} // end method notifyListener

	/**
	 * Notifies an event to all listeners of this workspace.
	 *
	 * @param event                 Event to notify to listeners.
	 * @param IllegalStateException If this workspace is not dynamic.
	 */
	protected boolean notify(ArrayList<LyWorkSpaceEvent> events)
	{
		if (_lockCount < 0)
			// Workspace is freed
			return false;

		if (!isDynamic())
			throw new IllegalStateException(
				"BF AppEngine : non dynamic list can't notify events since they don't have listeners");

		if (_listeners.size() == 0)
			return false;

		Vector<LyWorkSpaceListener> listeners;

		synchronized (_listenersLock)
		{
			listeners = new Vector<LyWorkSpaceListener>(_listeners);
		}

		LyWorkSpaceListener listener;
		LyWorkSpaceEvent    event;
		int				    i, n = listeners.size();
		int				    j, p = events.size();
		boolean			    used = false;
		LyEnvironment	    environment = ((_session != null) ? _session.getEnvironment() : null);
		boolean			    repostGuiEvents = (environment != null)
				&& environment.repostGuiEvents();

		for (i = 0; i < n; i++)
		{
			listener = listeners.elementAt(i);

			// Check if something changes during loop
			if (_listeners == null)
				break;

			if (!_listeners.contains(listener))
				continue;
			if ((listener instanceof LyGuiListener)
					&& ((LyGuiListener)listener).isIdle())
				continue;

			if (!(listener instanceof LyAdvancedWorkSpaceListener))
			{
				// Send each event lonely
				for (j = 0; j < p; j++)
				{
					event = events.get(j);
					if (repostGuiEvents
							&& (listener instanceof LyGuiListener))
					{
						environment.execGui(new _LyGuiNotifier(listener, event));
						used = true;
					}
					else
						used |= notifyListener(listener, event);
				}
			}
			else
			{
				if (repostGuiEvents
						&& (listener instanceof LyGuiListener))
				{
					environment.execGui(new _LyGuiNotifier((LyAdvancedWorkSpaceListener)listener,
							events));
					used = true;
				}
				else
					used |= notifyListener((LyAdvancedWorkSpaceListener)listener, events);
			}
		}

		return used;
	} // end method notify

	/**
	 * Internal method used to notify a listener.
	 */
	protected boolean notifyListener(LyAdvancedWorkSpaceListener listener,
	     ArrayList<LyWorkSpaceEvent> events)
	{
	    LyApplication  application = (_session != null ? _session.getApplication() : null);
	    LyEvaluator    evaluator = null;

	    try
	    {
	        if ((application != null)
	                && (listener instanceof LyEvaluator))
	            evaluator = application.setCurrentEvaluator((LyEvaluator)listener);
        
	        return listener.workSpaceEventsGrouped(events);
	        }
	    finally
	    {
            if ((application != null)
                    && (evaluator != null))
                application.setCurrentEvaluator(evaluator);
	    }
	}

	/* Management of the content of the list */

	/**
	 * Adds an object in this workspace. Calls add method of LyObjectList and if the object was
	 * successfully added, notifies this adding to listeners.
	 *
	 * @param  object Object to add in the workspace.
	 * @return If the object was not in this workspace, returns a positive index giving the position
	 *         where the object was added. Otherwise, returns a negative index giving the position
	 *         minus 1 where the same object was found.
	 * @see    LyObjectList#add
	 */
	protected synchronized int addObject(LyObject object)
	{
		int index = add(object);

		if (index < 0)
			return index;

		// Notify this to listeners
		if ((_listeners == null) || (_listeners.size() == 0))
			return index;

		LyWorkSpaceEvent event = new LyWorkSpaceEvent(this, LyWorkSpaceEvent.Type.ADD, object,
				index);

		notify(event);
		return index;
	}

	/**
	 * Removes an object from this workspace. Calls remove method of LyObjectList and if the object
	 * was successfully removed, notifies this removal to listeners.
	 *
	 * @param     object Object to remove from the workspace.
	 * @return    Index where the object was found in this workspace.
	 * @exception IllegalArgumentException If the object was not found in this workspace.
	 */
	protected synchronized int removeObject(LyObject object)
	{
		int index = remove(object);

		// Notify this to listeners
		if ((_listeners == null) || (_listeners.size() == 0))
			return index;

		LyWorkSpaceEvent event = new LyWorkSpaceEvent(this, LyWorkSpaceEvent.Type.REMOVE, object,
				index);

		notify(event);
		return index;
	}

	/**
	 * Internal method to change some values of an object in this workspace. Calls change method of
	 * LyObjectList and notifies this change to listeners.
	 *
	 * @param  object Object that is changed.
	 * @param  values Values changed
	 * @return True if the object position changed, false otherwise.
	 */
	protected synchronized boolean changeObject(LyObject object, LyValueSet values, LyEventInfo info)
	{
		if (_objects == null)
			return false;

		boolean moved = false;
		int     index;
		synchronized (this)
		{
			index = indexOf(object);
			if (isSorted())
			{
				remove(index);
				moved = (add(object) != index);
			}
		}
		if ((values == null)
		        && !moved)
			return false;

		// Notify this to listeners
		if ((_listeners == null) || (_listeners.size() == 0))
			return moved;

		int				 newIndex = (moved ? getIndex(object) : index);
		LyWorkSpaceEvent event = new LyWorkSpaceEvent(this, LyWorkSpaceEvent.Type.CHANGE, object, newIndex, index, values);

		event.setEventInfo(info);

		notify(event);
		return moved;
	} // end method changeObject

	/**
	 * Adds objects of the specified list to this list. This method may be inherited by subclasses
	 * to change its behavior.
	 *
	 * <p>It's overriden from LyObjectList to throw an exception if this workspace is dynamic since
	 * we won't known how to recompute it anymore.</p>
	 *
	 * @param     objectList Objects to add to this list.
	 * @exception IllegalStateException If this workspace is dynamic.
	 */
	public void addList(LyObjectList list)
	{
		if (isDynamic())
			throw new IllegalStateException(
				"BF AppEngine : objects lists can't be added to a dynamic workspace");

		super.addList(list);
	}

	/* Implementation of LyAdvancedDataSourceListener */

	/**
	 * Method for receiving an object creation. Adds the object concerned by the event in this
	 * workspace if it was not already in.
	 *
	 * @param  event Event notifying the creation of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectCreated(LyDataSourceEvent event)
	{
		LyObject object = (LyObject)event.getObject();

		if (contains(object))
			return false;

		// Add the object to the list
		addObject(object);
		return true;
	}

	/**
	 * Method for receiving an object deletion. Removes the object concerned by the event if it is
	 * found in this workspace.
	 *
	 * @param  event Event notifying the deletion of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectDeleted(LyDataSourceEvent event)
	{
		LyObject object = (LyObject)event.getObject();

		// If the object is not in the list, ignore this event
		if (!contains(object))
			return false;

		// Remove the object from the list
		removeObject(object);
		return true;
	}

	/**
	 * Method for receiving an object modification. Changes the object concerned by the event if it
	 * is found in this workspace.
	 *
	 * @param  event Event notifying the modification of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectModified(LyDataSourceEvent event)
	{
		LyObject object = (LyObject)event.getObject();

		// If the object is not in the list
		if (indexOf(object) < 0)
			return false;

		// Notify the change, the list may need to be resorted
		changeObject(object, event.getValues(), event.getEventInfo());
		return true;
	}

	/**
	 * Called when an object state is changed.
	 *
	 * @param  event Event notifying object state change.
	 * @return True if the event is used by the listener, false otherwise.
	 */
	public boolean objectStateChanged(LyAdvancedDataSourceEvent event)
	{
		LyObject object = (LyObject)event.getObject();

		// If the object is not in the list
		if (indexOf(object) < 0)
			return false;

		// Notify this to listeners
		if ((_listeners == null) || (_listeners.size() == 0))
			return false;

		int index;

		synchronized (this)
		{
			index = getIndex(object);
		}

		short                   type = event.getType();
        LyWorkSpaceEvent.Type   eventType;

		switch (type)
		{
		case LyDataSourceEvent.TYPE_SELECTION:
			eventType = LyWorkSpaceEvent.Type.SELECTION_CHANGE;
			break;
		case LyDataSourceEvent.TYPE_VISIBILITY:
			eventType = LyWorkSpaceEvent.Type.VISIBILITY_CHANGE;
			break;
		case LyDataSourceEvent.TYPE_ERROR:
			eventType = LyWorkSpaceEvent.Type.ERROR_CHANGE;
			break;
		default:
			throw new IllegalArgumentException("BF AppEngine : Bad type of event (type = "
				+ type + ")");
		}

		short					 oldState = event.getOldState();
		LyAdvancedWorkSpaceEvent myEvent
            = new LyAdvancedWorkSpaceEvent(this, eventType, object, index, oldState);

		myEvent.setEventInfo(event.getEventInfo());
		notify(myEvent);
		return true;
	} // end method objectStateChanged

	/**
	 * Method for receiving user event on an object.
	 *
	 * @param  event Event notifying the user event of an object.
	 * @return True indicating if the event was used, false otherwise.
	 */
	public boolean objectUserEvent(LyAdvancedDataSourceEvent event)
	{
		LyObject object = (LyObject)event.getObject();

		// If the object is not in the list
		if (!contains(object))
			return false;

		// Notify this to listeners
		if ((_listeners == null) || (_listeners.size() == 0))
			return false;

		int index;

		synchronized (this)
		{
			index = getIndex(object);
		}

		LyWorkSpaceEvent.Type	 type = LyWorkSpaceEvent.Type.USER_EVENT;
		LyAdvancedWorkSpaceEvent myEvent = new LyAdvancedWorkSpaceEvent(this, type, object, index,
				event);

		myEvent.setEventInfo(event.getEventInfo());
		notify(myEvent);
		return true;
	}

	/**
	 * Called when grouped events are sent.
	 *
	 * @param  events Sorted list of events.
	 * @return True if the events are used by the listener, false otherwise.
	 */
	public synchronized boolean dataSourceEventsGrouped(ArrayList<LyDataSourceEvent> events)
	{
		if ((events == null) || (events.isEmpty()))
			return false;

        // Change LyDataSourceEvents for LyWorkSpaceEvents if events are useful and send them.
        ArrayList<LyWorkSpaceEvent> stockedEvents = new ArrayList<LyWorkSpaceEvent>();

        if (_stockedEvents == null)
            _stockedEvents = new Hashtable<Thread, ArrayList<LyWorkSpaceEvent>>();

        _stockedEvents.put(Thread.currentThread(), stockedEvents);

		int				  i, n = events.size();
		LyDataSourceEvent event;
		short			  type;
		
        if (n > 1)
            // Force resort
            resort();

		for (i = 0; i < n; i++)
		{
			event = (LyDataSourceEvent)events.get(i);
			type = event.getType();
			if (type > LyDataSourceEvent.LAST_TYPE)
				objectUserEvent((LyAdvancedDataSourceEvent)event);
			else
			{
				switch (type)
				{
				case LyDataSourceEvent.TYPE_CREATE:
					objectCreated(event);
					break;
				case LyDataSourceEvent.TYPE_DELETE:
					objectDeleted(event);
					break;
				case LyDataSourceEvent.TYPE_SET:
					objectModified(event);
					break;
				case LyDataSourceEvent.TYPE_SELECTION:
				case LyDataSourceEvent.TYPE_VISIBILITY:
					objectStateChanged((LyAdvancedDataSourceEvent)event);
					break;
				default:
					break;
				}
			}
		}

        if (stockedEvents.isEmpty())
        {
            _stockedEvents.remove(Thread.currentThread());
            return false;
        }
        else
        {
            ArrayList<LyWorkSpaceEvent> wsEvents = stockedEvents;

            _stockedEvents.remove(Thread.currentThread());
            return notify(wsEvents);
        }
	} // end method dataSourceEventsGrouped

	/**
	 * Internal method to create listeners on relation filters.
	 */
	protected void createFilterListeners(LyFilter filter)
	{
		if (filter == null)
			return;

		createFilterListeners(null, getClasses(), filter);
	}

	/**
	 * Internal method to create listeners on relation filters.
	 */
	protected void createFilterListeners(LyFilterRelation filterRelation,
		LyInfoList<LyClassInfo> classes, LyFilter filter)
	{
		int						 i, n = filter.getSize();
		LyFilterElement			 element;
		LyFilterRelationListener listener;
		int						 j, nc = ((classes == null) ? 0 : classes.getSize());
		LyClassInfo				 classInfo;
		LyFilterRelation		 subFilterRelation;
		LyInfoList<LyClassInfo>  subClasses;

		for (i = 0; i < n; i++)
		{
			element = filter.getElement(i);
			if (element instanceof LyFilterRelation)
			{
				subFilterRelation = (LyFilterRelation)element;
				for (j = 0; j < nc; j++)
				{
					classInfo = classes.getInfo(j);
					listener = new LyFilterRelationListener(this, null, classInfo, filter,
							subFilterRelation);
					if (_filterListeners == null)
						_filterListeners = new Vector<LyFilterRelationListener>();
					_filterListeners.add(listener);
					subClasses = subFilterRelation.getPossibleClasses(true);
					createFilterListeners(subFilterRelation, subClasses,
						subFilterRelation.getFilter());
				}
			}
			else if (element instanceof LyFilter)
				createFilterListeners(filterRelation, classes, (LyFilter)element);
		}
	} // end method createFilterListeners

	/**
	 * Check filter.
	 */
	protected void checkFilter(LyObjectList objects, boolean add)
	{
		LyDataSourceEvent event;
		int				  i, n = objects.getSize();
		LyObject		  object;

		for (i = 0; i < n; i++)
		{
			object = objects.getObject(i);
			if (add)
			{
				if (contains(object))
					continue;
				// Notify a "creation"
				event = new LyDataSourceEvent(object.getObjectSource(),
						LyDataSourceEvent.TYPE_CREATE, object, null);
				objectCreated(event);
			}
			else
			{
				if (!contains(object))
					continue;
				// Notify a "deletion"
				event = new LyDataSourceEvent(object.getObjectSource(),
						LyDataSourceEvent.TYPE_DELETE, object, null);
				objectDeleted(event);
			}
		}
	} // end method checkFilter

	/**
	 * Method for retrieving source objects that lead to given target object.
	 *
	 * @param  target Target object
	 * @return Source objects that leade to target object.
	 */
	public LySimpleObjectList getSources(LyObject target)
	{
		return null;
	}

	/**
	 * Internal class for reposting events in Gui queue.
	 */
	class _LyGuiNotifier
		implements Runnable
	{
		LyWorkSpaceListener           _listener;
		LyWorkSpaceEvent              _event;
		ArrayList<LyWorkSpaceEvent>   _events;

		_LyGuiNotifier(LyWorkSpaceListener listener, LyWorkSpaceEvent event)
		{
			_listener = listener;
			_event = event;
		}

		_LyGuiNotifier(LyAdvancedWorkSpaceListener listener, ArrayList<LyWorkSpaceEvent> events)
		{
			_listener = listener;
			_events = events;
		}

		public void run()
		{
			if ((_listener instanceof LyGuiListener)
					&& ((LyGuiListener)_listener).isIdle())
				return;

			if (_event != null)
				notifyListener(_listener, _event);
			else
				notifyListener((LyAdvancedWorkSpaceListener)_listener, _events);
		}
	} // end class _LyGuiNotifier
} // end class LyWorkSpace
