/*
 * LyWorkSpaceEvent.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data;

/**
 * This class implements an event of adding/removing/changing an object within a workspace (dynamic
 * objects list). These events are sent to 'registered listeners' : classes which implements the
 * LyWorkSpaceListener interface and that were added as listeners on a workspace.
 *
 * @see LyWorkSpace
 * @see LyWorkSpaceListener
 */
public class LyWorkSpaceEvent
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyWorkSpaceEvent.java 33332 2011-11-30 08:15:07Z bangiole $";

    /** Event types. */
    public enum Type
    {
        /** ADD, the object of this event was added in the workspace. */
    	ADD,
    	/** REMOVE, the object of this event was removed from the workspace. */
    	REMOVE,
    	/**
    	 * CHANGE, an object was changed. This notifies either a values change on the
    	 * concerned object or a move of the object within the objects list.
    	 * <p>Note that object position may also change if one of the modified fields is one of the
    	 * fields specified in the list sort.</p>
    	 */
    	CHANGE,
    	/** SELECTION_CHANGE, an object selection state was changed. */
    	SELECTION_CHANGE,
    	/** VISIBILITY_CHANGE, an object visibility state was changed. */
    	VISIBILITY_CHANGE,
    	/** ERROR_CHANGE, an object error state was changed. */
    	ERROR_CHANGE,
    	/** USER_EVENT, an object error state was changed. */
    	USER_EVENT,
    	/** COUNTER_CHANGE, a counter of an object was changed. */
    	COUNTER_CHANGE
    }

	/** The workspace which has thrown this event. */
	LyWorkSpace				   _source;

	/** The type of event that occured. One of the TYPE_XXX constants */
	Type					   _type;

	/** The object concerned by the notification. */
	LyObject				   _object;

	/** The set of values associated to this event. May be null. */
	LyValueSet				   _values;

	/** The new index position of the object in the workspace. */
	int						   _index;

	/**
	 * The old index position of the object in the dynamic object list. Only used for event type :
	 * TYPE_CHANGE.
	 */
	int						   _oldIndex;

	/**
	 * Indicates if the couonter has been incremented or decremented. Only used for event type :
	 * TYPE_COUNTER_CHANGE.
	 */
	boolean					   _incremented;

	/** Additional informations for this event. */
	LyEventInfo				   _eventInfo;

	/**
	 * Constructs an event of the given type concerning the specified object.
	 *
	 * @param     source   Workspace that creates this event.
	 * @param     type     Type of event that occured. One of the TYPE_XXX constants.
	 * @param     object   Object concerned by the notification.
	 * @param     index    Index of the object in the workspace.
	 * @param     oldIndex Old index of the object in the workspace.
	 * @param     values   Set of values associated to this event. May be null.
	 * @exception IllegalArgumentException If source parameter or object parameter is null.
	 * @exception IllegalArgumentException If the given type is not one of valid types.
	 */
	public LyWorkSpaceEvent(LyWorkSpace source, Type type, LyObject object,
		int index, int oldIndex, LyValueSet values)
	{
		if (source == null)
			throw new IllegalArgumentException("BF AppEngine : source parameter can't be null");
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object parameter can't be null");

		_source = source;
		_type = type;
		_object = object;
		_index = index;
		_oldIndex = oldIndex;
		_values = values;
		_incremented = false;
	}

	/**
	 * Constructs an event notifying that the given object was moved within the workspace from the
	 * old position to the new one (due to a new sort). The type of this event is TYPE_CHANGE.
	 *
	 * @param source   Workspace that creates this event.
	 * @param object   Object concerned by the notification.
	 * @param index    Index of the object in the workspace.
	 * @param oldIndex Old index of the object in the workspace.
	 */
	public LyWorkSpaceEvent(LyWorkSpace source, LyObject object, int index, int oldIndex)
	{
		this(source, Type.CHANGE, object, index, oldIndex, null);
	}

	/**
	 * Constructs an event of the given type concerning the specified object without old position
	 * and associated values.
	 *
	 * @param source Workspace that creates this event.
	 * @param type   Type of event that occured. One of the TYPE_XXX constants.
	 * @param object Object concerned by the notification.
	 * @param index  Index of the object in the workspace.
	 */
	public LyWorkSpaceEvent(LyWorkSpace source, Type type, LyObject object, int index)
	{
		this(source, type, object, index, -1, null);
	}

	/**
	 * Internal Use Only - Constructs an event of the given type concerning the specified object without old position
	 * and associated values.
	 * 
	 *
	 * @param source Workspace that creates this event.
	 * @param type   Type of event that occured. One of the TYPE_XXX constants.
	 */
	LyWorkSpaceEvent(LyWorkSpace source, Type type)
	{
		if (source == null)
			throw new IllegalArgumentException("Leonardi : source parameter can't be null");
		
		_source = source;
		_type = type;
		_object = null;
		_index = -10;
		_oldIndex = -1;
		_values = null;
		_incremented = false;
	}
	
	/**
	 * Gets the type of this event.
	 *
	 * @return Type of this event : one of the TYPE_XXX constants.
	 */
	public Type getType()
	{
		return _type;
	}

	/**
	 * Gets workspace that created and notified this event.
	 *
	 * @return Workspace origin of this event.
	 */
	public LyWorkSpace getSource()
	{
		return _source;
	}

	/**
	 * Gets the object concerned by this notification.
	 *
	 * @return The object concerned by this event.
	 */
	public LyObject getObject()
	{
		return _object;
	}

	/**
	 * Gets index of the object in the workspace which created this event.
	 *
	 * @return Index of the object in the workspace.
	 */
	public int getIndex()
	{
		return _index;
	}

	/**
	 * Gets previous index of the object in the workspace before this event was notified.
	 *
	 * @return Old index of the object in the workspace.
	 */
	public int getOldIndex()
	{
		return _oldIndex;
	}

	/**
	 * Gets the set of values associated to this event.
	 *
	 * @return Set of values if any specified, may be null.
	 */
	public LyValueSet getValues()
	{
		return _values;
	}

	/**
	 * @return Returns the eventInfo.
	 */
	public LyEventInfo getEventInfo()
	{
		return _eventInfo;
	}

	/**
	 * @param eventInfo The eventInfo to set.
	 */
	public void setEventInfo(LyEventInfo eventInfo)
	{
		_eventInfo = eventInfo;
	}

	/**
	 * @return The flag indicated if the counter has been incremented (true) or decremented (false).
	 */
	public boolean isIncremented()
	{
		return _incremented;
	}

	/**
	 * Sets the flag indicated if the counter has been incremented or decremented.
	 *
	 * @param incremented the new value of the _incremented flag.
	 */
	public void setIncremented(boolean incremented)
	{
		_incremented = incremented;
	}
} // end class LyWorkSpaceEvent
