/*
 * LyFullCache.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data.cache;

import leon.data.*;
import leon.info.*;


/**
 * This class implements a cache that loads completely the instances of the associated class on
 * first access to data.
 *
 * <p>Normally, this kind of cache is only used for storing logical objects (LyObject).</p>
 *
 * @see LyObject
 */
public class LyFullCache extends LyObjectCache
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyFullCache.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Constructor of a full cache.
	 *
	 * @param cacheClass Description of the class corresponding to objects stored in this cache.
	 * @param policy     Description of th epolicy of thsi cache.
	 */
	public LyFullCache(LyClass cacheClass, LyInfoCache policy)
	{
		super(cacheClass, policy);
	}

	/**
	 * Finds an object in this cache from the specified identifier. This methods is overriden to
	 * force loading if the cache is not loaded.
	 * @param objectId the object if of the wanted object
	 * @return an object
	 * @throws LyNoSuchObjectException if the object is not found 
	 */
	public Object getObject(LyObjectId objectId)
		throws LyNoSuchObjectException
	{
		if (objectId == null)
			return null;

		// Loads cache if needed
		if (!isLoaded())
			load();

		return super.getObject(objectId);
	}

	/**
	 */
	protected void check()
	{
		// Loads cache if needed
		if (!isLoaded())
			load();
	}

	/**
	 * Checks if an object in this cache with the specified identifier. This methods is overriden to
	 * force loading if the cache is not loaded.
	 * @param id an object id
	 * 
	 * @return true if there is an object with the given id in the cache
	 */
	protected boolean checkObject(LyObjectId id)
	{
		// Loads cache if needed
		if (!isLoaded())
			load();

		return super.checkObject(id);
	}

	/**
	 * Check if the given object is found in cache.
	 * @param object the object to check
	 *
	 * @return True if the given object is found, false otherwise.
	 */
	protected boolean checkObject(Object object)
	{
		// Loads cache if needed
		if (!isLoaded())
			load();

		return super.checkObject(object);
	}

	/**
	 * Return the number of objects matching the given filter within this cache. 
	 * This methods is overriden to force loading if the cache is not loaded.
	 * @param filter the filter to check
	 * @return a number of obejcts
	 */
	public int getCount(LyFilter filter)
	{
		// Loads cache if needed
		if (!isLoaded())
			load();

		return getSize(filter);
	}

	/**
	 * Adds objects to the given object list that match the specified filter. This methods is
	 * overriden to force loading if the cache is not loaded.
	 * @param filter a filter
	 * @param list a list of objects
	 * @return the status of the method
	 */
	public Result getObjects(LyFilter filter, LySimpleObjectList list)
	{
		// Loads cache if needed
		if (!isLoaded())
			load();

		return super.getObjects(filter, list);
	}

	/**
	 * Setter for the allow loaded attribute.
	 * For a full cache, allowLoaded is forced to be true;
	 * @param allowLoaded the value of the allowLoaded attribute
	 */
	public void allowLoaded(boolean allowLoaded)
	{
		_allowLoaded = true;		
	}
} // end class LyFullCache
