/*
 * LyLruCache.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data.cache;

import java.util.*;
import leon.data.*;
import leon.info.*;


/**
 * This class implements a LRU (Last recently Used) cache in which objects are stored and unloaded
 * on the fly, depending to dataSource needs, but n instances are keeped in memory after removing.
 */
public class LyLruCache extends LySimpleCache
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyLruCache.java 28980 2010-12-22 10:53:09Z jlsaouli $";

	/** List of last recently removed objects in this cache. */
	protected LinkedList<Object>	   _keepedObjects;

	/** Number of last recently objects kepped in memory. Default limit is 50. */
	protected int			           _limit;

	/**
	 * Constructor of a simple cache of objects related to a given class description.
	 *
	 * @param cacheClass Class description of objects stored in the cache.
	 * @param policy     Description of th epolicy of thsi cache.
	 */
	public LyLruCache(LyClass cacheClass, LyInfoCache policy)
	{
		super(cacheClass, policy);
        
		_keepedObjects = new LinkedList<Object>();
		String userData = policy.getUserData();

		try
		{
			_limit = Integer.parseInt(userData);
		}
		catch (Exception e)
		{
			_limit = 50;
		}
	}

	/**
	 * Internal method used to get an object from the last recently removed ones.
	 * @param id an object id
	 * @return an object
	 */
	protected Object getFromKeepingList(LyObjectId id)
	{
		Object     object = null;
		LyObjectId objectId;

		synchronized (this)
		{
			Iterator<Object> it = _keepedObjects.iterator();

			while (it.hasNext())
			{
				object = it.next();
				objectId = _class.getObjectId(object);

				if (objectId.equals(id))
					return object;
			}
		}

		return null;
	}

	/**
	 * Internal method used to add an object to the last recently removed ones. If limit is exeeded,
	 * the older object of this list is removed from.
	 * @param object the object to add
	 */
	protected void addToKeepingList(Object object)
	{
		synchronized (this)
		{
			// Adding the object in first position in the recent used objects.
			_keepedObjects.addFirst(object);
			if (_keepedObjects.size() > _limit)
				_keepedObjects.removeLast();
		}
	}

	/**
	 * Internal method used to remove an object from the last recently used ones.
	 * @param object the object to remove
	 */
	protected void removeFromKeepingList(Object object)
	{
		synchronized (this)
		{
			_keepedObjects.remove(object);
		}
	}

	/**
	 * Overriden method to check in the last recently removed object list.
	 * @param object the object to check
	 * @return true if the object is in the cache
	 */
	protected boolean checkObject(Object object)
	{
		boolean result = super.checkObject(object);

		if (!result)
		{
			LyObjectId id = _class.getObjectId(object);

			result = (getFromKeepingList(id) == object);
		}

		return result;
	}

	/**
	 * Overriden method to check in the last recently removed object list.
	 * @param id an object id
	 * @return true if an object with the given id is in the cache
	 */
	protected boolean checkObject(LyObjectId id)
	{
		boolean result = super.checkObject(id);

		if (!result)
			result = (getFromKeepingList(id) != null);
		return result;
	}

	/**
	 * Gets an object in this cache by its identifier.
	 *
	 * @param     id Object identifier.
	 * @return    Object found in cache with the given identifier, or null if no object is found and
	 *            if the id was registered in unknown ids list.
	 * @exception LyNoSuchObjectException If no object of this cache has this identifier and if the
	 *                                    given id has not been regsitered in unknown ids list.
	 */
	public Object getObject(LyObjectId id)
		throws LyNoSuchObjectException
	{
		if (id == null)
			return null;

		Object object = getFromCache(id);

		if (object == null)
		{
			// Try from the keeping list.
			object = getFromKeepingList(id);
			if (object != null)
				// Put the object back in the cache.
				putObject(object);
			else if (!isUnknown(id))
				// Throw exception only if the given id has not been registered
				// in unknown ids list
				throw new LyNoSuchObjectException("BF AppEngine : no object with id " + id
					+ " found in cache");
		}

		return object;
	}

	/**
	 * Adds a new object to this cache. This method is overriden to remove the object from the
	 * keeping list.
	 * @param object the object to add
	 */
	protected void putObject(Object object)
	{
		super.putObject(object);
		removeFromKeepingList(object);
	}

	/**
	 * Removes an object from the cache. This method is overriden to add the object in the keeping
	 * list.
	 * @param object the object to remove
	 */
	protected void removeObject(Object object)
	{
		super.removeObject(object);
		addToKeepingList(object);
	}

	/**
	 * To unload objects from the cache. This method is overriden to reset the keeping list.
	 */
	protected void unload()
	{
		synchronized (this)
		{
			super.unload();
			_keepedObjects = new LinkedList<Object>();
		}
	}
} // end class LyLruCache
