/*
 * LyObjectCache.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data.cache;

import java.util.*;
import leon.data.*;
import leon.info.*;


/**
 * This class implements a cache that stores objects in memory.
 */
public abstract class LyObjectCache extends LyCache
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyObjectCache.java 41320 2012-10-23 13:25:02Z bangiole $";

	/**
	 * Objects table of the cache. This table is an hashtable that stores pairs of (LyObjectId,
	 * LyObject).
	 */
	protected Hashtable<LyObjectId, Object>		   _objects;

	/**
	 * Hashtable of peers associated with objects in the cache. Used to optimize the method
	 * getObjectFromPeer. Stores pairs of physical and logical Objects (Object, LyObject).
	 */
	protected Hashtable<Object, LyObject>		   _peerObjects;

	/**
	 * Constructor of a full cache.
	 *
	 * @param cacheClass Description of the class corresponding to objects stored in this cache.
	 * @param policy     Description of th epolicy of thsi cache.
	 */
	public LyObjectCache(LyClass cacheClass, LyInfoCache policy)
	{
		super(cacheClass, policy);

		_objects = new Hashtable<LyObjectId, Object>(1);
	}

	/**
	 * Retrieve an object in this cache from the given objectId.
	 * @param objectId 
	 * @return the object corresponding to the given object id
	 */
	protected Object getFromCache(LyObjectId objectId)
	{
		synchronized (this)
		{
			return _objects.get(objectId);
		}
	}

	/**
	 * Adds an object in this cache.
	 * @param objectId the object id
	 * @param object  the object to add
	 */
	protected void putToCache(LyObjectId objectId, Object object)
	{
		synchronized (this)
		{
			_objects.put(objectId, object);
			if (object == null)
			{
				getEnvironment().logError("Registering a null object for id " + objectId);
				getEnvironment().writeException(new Exception());
			}
		}
	}

	/**
	 * Removes an object in this cache corresponding to the given objectId.
	 * @param objectId object id to remove from cache
	 */
	protected void removeFromCache(LyObjectId objectId)
	{
		synchronized (this)
		{
			_objects.remove(objectId);
		}
	}

	/**
	 * Gets an object from its physical representation (peer). This method is only valid if stored
	 * objects are logical objects (instances of LyObject).
	 *
	 * @param     peer Physical representation of the searched object.
	 * @return    Object with the specified physical representation if it exists, otherwise null.
	 * @exception IllegalStateException If this cache is not a cache that keep logical objects.
	 */
	public LyObject getObjectFromPeer(Object peer)
	{
		if (_peerObjects == null)
			return null;

		return _peerObjects.get(peer);
	}

	/**
	 * Adds a pair (peer, object) in this cache.
	 * @param peer the peer object
	 * @param object the associated LyObject
	 */
	protected void putPeerObject(Object peer, LyObject object)
	{
		if (_peerObjects == null)
			_peerObjects = new Hashtable<Object, LyObject>(1);
		_peerObjects.put(peer, object);
	}

	/**
	 * Removes objects associated to the given peer in this cache.
	 * @param peer the peer object to remove
	 */
	protected void removePeerObject(Object peer)
	{
		if (_peerObjects == null)
			return;
		_peerObjects.remove(peer);
	}

	/**
	 * Return the number of objects within this cache that match the given filter (all objects of
	 * the cache if filter is null). This method gives objects that are in the cache, no other
	 * objects are loaded.
	 *
	 * @param  filter Filter used for selecting objects. If filter is null, the total is returned.
	 * @return Number of objects in the cache that match the filter.
	 */
	public int getSize(LyFilter filter)
	{
		if (filter == null)
			return _objects.size();

		if (!(_class instanceof LyClassInfo))
			return -1;

		int		    n = 0;
		Enumeration<Object> e = getElements();
		Object	    o;

		while (e.hasMoreElements())
		{
			o = e.nextElement();
			if (filter.match(o))
				n++;
		}

		return n;
	}

	/**
	 * Tests if this cache is empty or not (contains no object).
	 *
	 * @return True if the cache is empty, false otherwise.
	 */
	public boolean isEmpty()
	{
		return _objects.isEmpty();
	}

	/**
	 * Gets all objects stored in this cache as an enumeration.
	 *
	 * @return Enumeration of objects found in the cache.
	 */
	protected Enumeration<Object> getElements()
	{
		Enumeration<Object> e = _objects.elements();

		return e;
	}

	/**
	 * Clear cache content.
	 */
	protected void clearCache()
	{
		_objects = new Hashtable<LyObjectId, Object>(1);
		_peerObjects = null;
	}
} // end class LyObjectCache
