/*
 * LySimpleCache.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data.cache;

import java.lang.ref.*;
import java.util.*;

import leon.data.*;
import leon.info.*;


/**
 * This class implements a simple cache in which objects are stored and unloaded on the fly,
 * depending to dataSource needs.
 *
 * <p>Each object kept in the cache has a number of locks (uses) associated to it, when this number
 * of locks become 0, the object is removed from the cache.</p>
 */
public class LySimpleCache extends LyCache
{
    /** File version. */
    public static final String __VERSION = "$Id: LySimpleCache.java 30705 2011-05-02 09:54:36Z bangiole $";

    /**
     * Objects table of the cache. This table is an hashtable that stores pairs of (LyObjectId,
     * WeakReference).
     */
    protected Hashtable<LyObjectId, WeakReference<Object>> _weakReferences;

    /**
     * Hashtable of peers associated with objects in the cache. Used to optimize the method
     * getObjectFromPeer. Stores pairs of physical and logical Objects (Object, WeakReference).
     */
    protected Hashtable<Object, WeakReference<LyObject>> _peerObjects;

    /**
     * Constructor of a simple cache of objects related to a given class description.
     *
     * @param  cacheClass  Class description of objects stored in the cache.
     * @param  policy      Description of th epolicy of thsi cache.
     */
    public LySimpleCache(LyClass cacheClass, LyInfoCache policy)
    {
        super(cacheClass, policy);

        _allowLoaded=false;
        _weakReferences = new Hashtable<LyObjectId, WeakReference<Object>>(1);
    }

    /**
     * Retrieve an object in this cache from the given objectId.
     *
     * @param   objectId  the object id to find
     *
     * @return  an object corresponding to the given object Id if found
     */
    protected Object getFromCache(LyObjectId objectId)
    {
        synchronized (this)
        {
            WeakReference<Object> reference = _weakReferences.get(objectId);
            Object                object = null;

            if (reference != null)
            {
                object = reference.get();
                if (object == null)
                {
                    removeFromCache(objectId);
                }
            }

            return object;
        }
    }

    /**
     * Adds an object in this cache.
     *
     * @param  objectId  the object id of the object to add
     * @param  object    the object to add
     */
    protected void putToCache(LyObjectId objectId, Object object)
    {
        synchronized (this)
        {
            _weakReferences.put(objectId, new WeakReference<Object>(object));
        }
    }

    /**
     * Removes an object in this cache corresponding to the given objectId.
     *
     * @param  objectId  the object id to remove
     */
    protected void removeFromCache(LyObjectId objectId)
    {
        synchronized (this)
        {
            _weakReferences.remove(objectId);
        }
    }

    /**
     * Clear cache.
     */
    protected void clearCache()
    {
        _weakReferences = new Hashtable<LyObjectId, WeakReference<Object>>(1);
        _peerObjects = null;
    }

    /**
     * Gets all objects stored in this cache as an enumeration.
     *
     * @return  Enumeration of objects found in the cache.
     */
    protected Enumeration<Object> getElements()
    {
        synchronized (this)
        {
            Vector<Object>                    c = new Vector<Object>();
            Collection<WeakReference<Object>> elements = _weakReferences.values();
            Iterator<WeakReference<Object>>   it = elements.iterator();
            WeakReference<Object>             reference;
            Object                            object;

            while (it.hasNext())
            {
                reference = it.next();
                object = reference.get();
                if (object != null)
                    c.add(object);
            }

            return c.elements();
        }
    }

    /**
     * Gets an object from its physical representation (peer). This method is only valid if stored
     * objects are logical objects (instances of LyObject).
     *
     * @param      peer  Physical representation of the searched object.
     *
     * @return     Object with the specified physical representation if it exists, otherwise null.
     *
     * @exception  IllegalStateException  If this cache is not a cache that keep logical objects.
     */
    public LyObject getObjectFromPeer(Object peer)
    {
        if (!(_class instanceof LyClassInfo))
            throw new IllegalStateException(
                "BF AppEngine : attempt to get an object from its peer on a non classInfo cache");

        if (_peerObjects == null)
            return null;

        if (peer == null)
            return null;

        WeakReference<LyObject> reference = _peerObjects.get(peer);

        if (reference == null)
            return null;

        LyObject object = reference.get();

        if (object != null)
            return object;
        else
            _peerObjects.remove(peer);

        return null;
    }

    /**
     * Adds a pair (peer, object) in this cache.
     *
     * @param  peer    the peer object
     * @param  object  the associated LyObject
     */
    protected void putPeerObject(Object peer, LyObject object)
    {
        if (_peerObjects == null)
            _peerObjects = new Hashtable<Object, WeakReference<LyObject>>(1);

        _peerObjects.put(peer, new WeakReference<LyObject>(object));
    }

    /**
     * Removes objects associated to the given peer in this cache.
     *
     * @param  peer  the peer object to remove
     */
    protected void removePeerObject(Object peer)
    {
        if (_peerObjects == null)
            return;

        _peerObjects.remove(peer);
    }

    /**
     * Return the number of objects within this cache that match the given filter (all objects of
     * the cache if filter is null). This method gives objects that are in the cache, no other
     * objects are loaded.
     *
     * @param   filter  Filter used for selecting objects. If filter is null, the total is returned.
     *
     * @return  Number of objects in the cache that match the filter.
     */
    public int getSize(LyFilter filter)
    {
        synchronized (this)
        {
            Collection<WeakReference<Object>> elements = _weakReferences.values();
            Iterator<WeakReference<Object>>   it = elements.iterator();
            WeakReference<Object>             reference;
            Object                            object;
            int                               n = 0;

            while (it.hasNext())
            {
                reference = it.next();
                object = reference.get();
                if ((object != null)
                        && ((filter == null) || (filter.match(object))))
                    n++;
            }

            return n;
        }
    }

    /**
     * Tests if this cache is empty or not (contains no object).
     *
     * @return  True if the cache is empty, false otherwise.
     */
    public boolean isEmpty()
    {
        if (_weakReferences.size() == 0)
            return true;

        synchronized (this)
        {
            Collection<WeakReference<Object>> elements = _weakReferences.values();
            Iterator<WeakReference<Object>>   it = elements.iterator();
            WeakReference<Object>             reference;
            Object                            object;

            while (it.hasNext())
            {
                reference = it.next();
                object = reference.get();
                if (object != null)
                    return false;
            }

            return true;
        }
    }

    /**
     * Check if the given object is found in cache.
     *
     * @param   object  the object to check
     *
     * @return  True if the given object is found, false otherwise.
     */
    protected boolean checkObject(Object object)
    {
        if (super.checkObject(object))
            return true;

        if (object instanceof LyObject)
        {
            Object peer = ((LyObject)object).getPeer();

            // Be sure not to remove peer if peer was not associated to given object in cache
            if (getObjectFromPeer(peer) == object)
                removePeerObject(peer);
        }

        return false;
    }

	/**
	 * Indicate if the cache can be marked as Loaded.
	 * For the simple cache, this implementation does nothing as marking this cache as loaded is a nonsense
	 */
	public void allowLoaded(boolean allowLoaded)
	{
	}
} // end class LySimpleCache
