/*
 * LyArrayFieldInfoDaemon.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data.daemon;

import java.lang.reflect.*;
import java.util.*;
import leon.data.*;
import leon.info.*;


/**
 * Daemon to manage array field info values.
 */
public class LyArrayFieldInfoDaemon extends LyFieldInfoDaemon
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyArrayFieldInfoDaemon.java 34503 2012-03-13 08:59:21Z jlsaouli $";

	/** Classinfo for value objects. */
	LyClassInfo				   _arrayClassInfo;

	/** The foreign key relation in the array class info. */
	LyRelationFieldInfo		   _fkRelation;

	/** The value fieldInfo in the array class. */
	LyFieldInfo				   _valueFieldInfo;

	/** The index fieldInfo in the array class. */
	LyNumberFieldInfo		   _indexFieldInfo;

	/** The index sort. */
	LySort					   _sort;

	/** Results table. */
	Hashtable<LyObjectId, Object>				   _results;

	/** Static object for storing an empty list in results hashtable. */
	static Object			   EMPTY_LIST = new Object();

	/**
	 * Default constructor.
	 * @param sourceClass the source class
	 * @param array the array field info
	 * @param arrayClassInfo the array classinfo
	 */
	public LyArrayFieldInfoDaemon(LyClassInfo sourceClass,
		LyArrayFieldInfo array, LyClassInfo arrayClassInfo)
	{
		super(sourceClass, array);

		_arrayClassInfo = arrayClassInfo;
		_results = new Hashtable<LyObjectId, Object>();

		// Listen to the arrayClassInfo data source
		LyObjectSource source;

		source = _arrayClassInfo.getObjectSource();
		source.addListener(new _LyArrayClassListener());

		// TODO Get a better way to do this !!!
		_fkRelation = (LyRelationFieldInfo)_arrayClassInfo.getFieldInfo(0);
		_valueFieldInfo = _arrayClassInfo.getFieldInfo(1);
		if (_arrayClassInfo.getFieldInfoCount() > 2)
		{
			_indexFieldInfo = (LyNumberFieldInfo)_arrayClassInfo.getFieldInfo(2);
			_sort = new LySort();
			_sort.addCriterion(_indexFieldInfo, true);
		}

		// Listen to events on data sources for classes pointed by the relation
		if (_valueFieldInfo instanceof LyRelationFieldInfo)
		{
			LyClassInfo		    classInfo;
			LyRelationFieldInfo relation = (LyRelationFieldInfo)_valueFieldInfo;
			int				    i, n = relation.getClassInfoCount();

			for (i = 0; i < n; i++)
			{
				classInfo = relation.getClassInfo(i);
				source = classInfo.getObjectSource();
				source.addListener(this, LyDataSourceEvent.TYPE_CREATE);
			}
		}
	}

	/**
	 * Checks whether a given object is contained in our objects list.
	 */
	boolean contains(LyObject object)
	{
		LyObjectId id = object.getObjectId();

		return _results.containsKey(id);
	}

	/**
	 * Get a result of the calculated value of the relation in a cache.
	 *
	 * @param  object the associated object of the calulated value.
	 * @return the calculated value.
	 */
	Object getResult(LyObject object)
	{
		LyObjectId id = object.getObjectId();

		return _results.get(id);
	}

	/**
	 * To put a result of the calculated value of the relation in a cache.
	 *
	 * @param object the associated object of the calulated value.
	 * @param result the calculated value.
	 */
	void putResult(LyObject object, Object result)
	{
		if (!object.checkInstance())
			// Never store an object that is not found in cache
			// This may occur when checking filter on peer object.
			return;

		LyObjectId id = object.getObjectId();

		_results.put(id, result);
	}

	/**
	 * To remove a result of the calculated value of the relation from the cache.
	 *
	 * @param object the associated object of the calulated value.
	 * @param result the calculated value.
	 */
	void removeResult(LyObject object)
	{
		LyObjectId id = object.getObjectId();

		_results.remove(id);
	}

	/**
	 * Find all array objects related to the source object.
	 *
	 * @param  source
	 * @return A LyObjectList containing all the array objects ordered by the index.
	 */
	LyObjectList findArrays(LyObject source)
	{
		// Find objects of the array classInfo that references
		// objects of source class through fk relation sorted by index
		LyFilter		   filter = new LyFilter(LyFilter.Type.OR);
		LyFilterExpression filterExp;
        LyFilterExpression.Condition			   condition;

		condition = LyFilterExpression.Condition.EQUALS;
		filterExp = new LyFilterExpression(_fkRelation, false, condition, source.getObjectId());
		filter.addElement(filterExp);

		// Get filtered list
		LyObjectList list;

		// GTA : maybe we must do this test to avoid unuseless requests
		//boolean                       local = (source.getPeer() == null);
		boolean		 local = false;

		list = new LyClassWorkSpace(null, _arrayClassInfo, null, filter, _sort, false, true, local);

		return ((list == null) ? new LySimpleObjectList() : new LySimpleObjectList(list, true));
	}

	/**
	 * Find array object related to the target object.
	 *
	 * @param  target
	 * @return The array object if axists, false otherwise.
	 */
	LyObject findArray(LyObject target, boolean local)
	{
		// Find objects of the array classInfo that references
		// objects of source class through fk relation sorted by index
		LyFilter		    filter = new LyFilter(LyFilter.Type.OR);
		LyFilterExpression  filterExp;
        LyFilterExpression.Condition			    condition;
		LyRelationFieldInfo relation = (LyRelationFieldInfo)_valueFieldInfo;

		condition = LyFilterExpression.Condition.EQUALS;
		filterExp = new LyFilterExpression(relation, false, condition, target.getObjectId());
		filter.addElement(filterExp);

		// Get filtered list
		LyObjectList list;

		// GTA : maybe we must do this test to avoid unuseless requests
		list = new LyClassWorkSpace(null, _arrayClassInfo, null, filter, _sort, false, true, local);

		if ((list == null) || (list.getSize() == 0))
			return null;
		else
			return list.getObject(0);
	}

	/**
	 * Gets references for an object.
	 * @param object object holding the array object 
	 * @return a list of Array Objects
	 */
	public LyObjectList getArrays(LyObject object)
	{
		Object result = getResult(object);

		if (result == null)
			return null;
		else if (result == EMPTY_LIST)
			return null;
		else if (result instanceof LyObject)
			return new LySimpleObjectList((LyObject)result);
		else
			return (LyObjectList)result;
	}

	/**
	 * To get the array value for the specified object.
	 *
	 * @param  object the object we want the value.
	 * @return the value of the array field for the specified object.
	 */
	public Object getValue(LyObject object)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		Object		 result = getResult(object);
		LyObjectList targets = null;

		if (result == null)
		{
			// We must compute the value.
			// In this case, nobody have ask for the value.
			targets = findArrays(object);

			// Put the result in the cache.
			if ((targets == null) || (targets.getSize() == 0))
				putResult(object, EMPTY_LIST);
			else if (targets.getSize() == 1)
				putResult(object, targets.getObject(0));
			else
				putResult(object, targets);
		}
		else if (result == EMPTY_LIST)
			targets = null;
		else if (result instanceof LyObject)
			targets = new LySimpleObjectList((LyObject)result);
		else
			targets = (LyObjectList)result;

		// Compute the value.
		return computeValue(targets);
	} // end method getValue

	/**
	 * Utility method to compute a value for the array from an objects list.
	 * @param objects the list of objects to compute
	 * @return the reult of the computation. In this case, an array of Object corresponding to the object in the list
	 */
	protected Object computeValue(LyObjectList objects)
	{
		int n = ((objects == null) ? 0 : objects.getSize());

		if (n == 0)
			return null;

		Class<?>  fieldClass = _valueFieldInfo.getValueClass();
		Object[]  values;

		try
		{
			values = (Object[])Array.newInstance(fieldClass, n);
		}
		catch (Exception e)
		{
			return null;
		}

		LyObject object;
		int		 i;

		if (_sort != null)
			objects.setSort(_sort);

		for (i = 0; i < n; i++)
		{
			object = objects.getObject(i);
			values[i] = object.getValue(_valueFieldInfo);
		}

		return values;
	} // end method computeValue

	/**
	 * To free a calculated value for the specified object.
	 *
	 * @param object the object we want to free the value.
	 */
	public void freeValue(LyObject object)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		// Remove the value from the results.
		removeResult(object);
	}

	/**
	 * Unload daemon.
	 */
	public void unload()
	{
		_results.clear();
	}

	/**
	 * Method to add a relation beetween a source object and an array object.
	 *
	 * @param object Source object.
	 * @param array  Array object.
	 */
	void addRelation(LyObject object, LyObject array)
	{
		addRelation(object, array, true, null);
	}

	/**
	 * Method to add a relation beetween a source object and an array object.
	 *
	 * @param object Source object.
	 * @param array  Array object.
	 * @param event  optional event indicating when notified for instance, may be null
	 */
	void addRelation(LyObject object, LyObject array, boolean set, LyDataSourceEvent event)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null.");
		if (array == null)
			throw new IllegalArgumentException("BF AppEngine : target can't be null.");

		if (!object.instanceOf(_sourceClass))
			return;

		Object result = getResult(object);

		if (result == null)
			result = EMPTY_LIST;

		LySimpleObjectList targets;

		if (result == EMPTY_LIST)
		{
			putResult(object, array);
			targets = new LySimpleObjectList(array);
		}
		else if (result instanceof LyObject)
		{
			targets = new LySimpleObjectList((LyObject)result);
			if (result != array)
			{
				if (_sort != null)
					targets.setSort(_sort);
				targets.addObject(array);
				putResult(object, targets);
			}
		}
		else
		{
			targets = (LySimpleObjectList)result;
			if (!targets.contains(array))
				targets.addObject(array);
		}

		if (set)
		{
			Object value = computeValue(targets);

			setValue(object, _fieldInfo, value, event);
		}
	} // end method addRelation

	/**
	 * Method to remove a relation beetween a source object and an array object.
	 *
	 * @param object Source object.
	 * @param array  Array object.
	 */
	void removeRelation(LyObject object, LyObject array)
	{
		removeRelation(object, array, true, null);
	}

	/**
	 * Method to remove a relation beetween a source object and an array object.
	 *
	 * @param object Source object.
	 * @param array  Array object.
	 * @param event  optional event indicating when notified for instance, may be null
	 */
	void removeRelation(LyObject object, LyObject array, boolean set, LyDataSourceEvent event)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null.");
		if (array == null)
			throw new IllegalArgumentException("BF AppEngine : target can't be null.");

		if (!object.instanceOf(_sourceClass))
			return;

		Object result = getResult(object);

		if (result == null)
			return;

		if (result == EMPTY_LIST)
			return;

		if (result instanceof LyObject)
		{
			if (result != array)
				return;

			putResult(object, EMPTY_LIST);

			if (set)
				// Modify the object.
				setValue(object, _fieldInfo, null, event);
		}
		else
		{
			LySimpleObjectList targets = (LySimpleObjectList)result;

			if (!targets.contains(array))
				return;

			targets.removeObject(array);

			// Modify the object.
			int n = targets.getSize();

			if (n == 0)
				putResult(object, EMPTY_LIST);
			else if (n == 1)
				putResult(object, targets.getObject(0));

			if (set)
			{
				Object value = ((n == 0) ? null : computeValue(targets));

				setValue(object, _fieldInfo, value, event);
			}
		}
	} // end method removeRelation

	/**
	 * Gets the source relative to the specified array object.
	 */
	LyObject getSource(LyObject array)
	{
		LyObjectId     sourceId = (LyObjectId)array.getValue(_fkRelation);
		LyObjectSource objectSource = _sourceClass.getObjectSource();

		return (LyObject)objectSource.getObject(sourceId, true);
	}

	/* Implementation of LyDataSourceListener */

	/**
	 * Method for object creation callback.
	 * 
	 * @param event the datasource event 
	 * @return true if this event was handled
	 * @see LyDataSourceListener#objectCreated
	 */
	public boolean objectCreated(LyDataSourceEvent event)
	{
		if (event.isLocal())
			return false;

		LyObject source = (LyObject)event.getObject();

		if (!source.instanceOf(_sourceClass))
		{
			if (!event.isNotified())
				return false;

			// We receive a creation on a pointed object : refresh the source.
			LyObject array = findArray(source, true);

			if (array != null)
			{
				source = getSource(array);
				if (source != null)
					source.refresh(new LyInfoList<LyFieldInfo>(_fieldInfo));
			}
			return true;
		}

		if (event.isLoading()) // XX Differentiate a LOAD and a real CREATE
			return false;

// GTA : in comments to avoid pb when immediate gui updates are allowed.
//              if (contains(source))
//                      return false;

		LyArrayValue	   newValue = (LyArrayValue)source.getFieldValue(_fieldInfo);
		int				   n = ((newValue == null) ? 0 : newValue.getValueCount());
		LyObject		   array;
		LySimpleObjectList arrays = new LySimpleObjectList();

		if (_sort != null)
			arrays.setSort(_sort);

		for (int i = 0; i < n; i++)
		{
			// There is no array object, we must create it.
			array = _arrayClassInfo.createInstance();
			array.setValue(_fkRelation, source.getObjectId());
			array.setValue(_valueFieldInfo, newValue.getValue(i));
			array.setValue(_indexFieldInfo, new Integer(i));
			addRelation(source, array, false, event);
			array.create();
			arrays.addObject(array);
		}

		// Do not forget to put the array list in the result table !
		n = arrays.getSize();
		if (n == 0)
			putResult(source, EMPTY_LIST);
		else if (n == 1)
			putResult(source, arrays.getObject(0));
		else
			putResult(source, arrays);

		Object value = ((n == 0) ? null : computeValue(arrays));

		setValue(source, _fieldInfo, value, event);
		return true;
	} // end method objectCreated

	/**
	 * Method for object deletion callback.
	 *
	 * @param event the datasource event 
	 * @return true if this event was handled
	 * @see LyDataSourceListener#objectDeleted
	 */
	public boolean objectDeleted(LyDataSourceEvent event)
	{
		LyObject source = (LyObject)event.getObject();

		// If the object is not in the list, ignore this event
		if (!contains(source))
			return false;

		// Remove object from results table
		removeResult(source);
		return true;
	}

	/**
	 * Method for object modification callback.
	 *
	 * @param event the datasource event 
	 * @return true if this event was handled
	 * @see LyDataSourceListener#objectModified
	 */
	public boolean objectModified(LyDataSourceEvent event)
	{
		if (event.isLocal())
			return false;

		LyObject source = (LyObject)event.getObject();

		// If the object is not in our list, ignore this event
		if (!contains(source))
			return false;

		LyValueSet values = event.getValues();

		if ((values != null) && (values.getFieldValue(_fieldInfo) == null))
			return false;

		// Find deltas between new value and old value
		LyArrayValue oldValue = ((values == null) ? null
									: (LyArrayValue)values.getFieldValue(_fieldInfo));

		if (oldValue == null)
			return false;

		LyArrayValue newValue = (LyArrayValue)source.getFieldValue(_fieldInfo);
		
		if ((newValue != null)
				&& _fieldInfo.equals(oldValue.getValue(), newValue.getValue()))
			return false;

		LyObjectList	   oldList = getArrays(source);
		LyObject		   array;
		int				   i, nn, no;
		LyFieldInfoValue   v;
		Object			   v1, v2;
		LySimpleObjectList newList = new LySimpleObjectList();

		newList.setSorted(false);
		if (oldList != null)
		{
			// Duplicate list to avoid conflict during add/remove operations
			oldList = new LySimpleObjectList(oldList);
			if (_sort != null)
				oldList.setSort(_sort);
		}

		no = ((oldList == null) ? 0 : oldList.getSize());
		nn = ((newValue == null) ? 0 : newValue.getValueCount());

		// Update existing objects
		for (i = ((no < nn) ? no : nn) - 1; i >= 0; i--)
		{
			array = oldList.getObject(i);
			if (array.isDeleted())
				continue;

			// Just sets the value if newValue and oldValue are not the same.
			v = newValue.getFieldValue(i);
			v1 = v.getValue();
			v2 = array.getValue(_valueFieldInfo);
			if (!_valueFieldInfo.equals(v1, v2))
			{
				if (array.getNumberValue(_indexFieldInfo).intValue() != i)
				{
					setValue(array, _valueFieldInfo, v1, event, false);
					setValue(array, _indexFieldInfo, new Integer(i), event);
				}
				else
					setValue(array, _valueFieldInfo, v1, event);
			}
			newList.addObject(array, 0);
		}

		if (nn > no)
		{
			// Create new objects
			for (i = no; i < nn; i++)
			{
				v = newValue.getFieldValue(i);
				v1 = v.getValue();

				array = _arrayClassInfo.createInstance();
				array.setValue(_fkRelation, source.getObjectId());
				array.setValue(_valueFieldInfo, v1);
				array.setValue(_indexFieldInfo, new Integer(i));
				addRelation(source, array, false, event);
				array.create();
				newList.addObject(array);
			}
		}
		else if (nn < no)
		{
			// Delete extra objects
			for (i = nn; i < no; i++)
			{
				array = oldList.getObject(i);
				if (newList.contains(array))
					continue;
				if (array.isDeleted())
					continue;

				array.setValue(_valueFieldInfo, null);
				removeRelation(source, array, false, event);
				array.delete();
			}
		}

		Object value = ((nn == 0) ? null : computeValue(newList));

		setValue(source, _fieldInfo, value, event);
		return true;
	} // end method objectModified

	/**
	 * Called when objectId of an object has been changed. This method must update tables with
	 * references to objectIds.
	 *
	 * @param oldObjectId
	 * @param newObjectId
	 */
	public void idRehashed(LyObjectId oldObjectId, LyObjectId newObjectId)
	{
		Object value;

		if ((_results != null)
				&& _results.containsKey(oldObjectId))
		{
			value = _results.remove(oldObjectId);
			_results.put(newObjectId, value);
		}
	}

	/**
	 * Internal class to listen to array class info events.
	 */
	class _LyArrayClassListener
		implements LyDataSourceListener
	{

		/**
		 * Method for object creation callback.
		 *
		 * @param event the datasource event 
		 * @return true if this event was handled
		 * @see LyDataSourceListener#objectCreated
		 */
		public boolean objectCreated(LyDataSourceEvent event)
		{
			if (event.isLocal())
				return false;
			if (!event.isNotified())
				return false;

			LyObject array = (LyObject)event.getObject();
			LyObject source = getSource(array);

			if (source != null)
				addRelation(source, array, true, event);
			return true;
		}

		/**
		 * Method for object deletion callback.
		 *
		 * @param event the datasource event 
		 * @return true if this event was handled
		 * @see LyDataSourceListener#objectDeleted
		 */
		public boolean objectDeleted(LyDataSourceEvent event)
		{
			if (event.isLocal())
				return false;

			LyObject array = (LyObject)event.getObject();
			LyObject source = getSource(array);

			if (source == null)
				return false;

			LyObjectList arrays = getArrays(source);

			if ((arrays == null)
					|| !arrays.contains(array))
				return false;

			if (!event.isNotified())
			{
				// Change existing items starting from the deleted one
				Object   value = array.getValue(_indexFieldInfo);
				int		 index = ((Number)value).intValue();
				int		 i, n = arrays.getSize();
				LyObject next;

				for (i = index + 1; i < n; i++)
				{
					next = arrays.getObject(i);
					value = next.getValue(_valueFieldInfo);
					setValue(array, _valueFieldInfo, value, event);
					array = next;
				}
			}

			// Delete last item
			removeRelation(source, array, true, event);
			return true;
		} // end method objectDeleted

		/**
		 * Method for object modification callback.
		 *
		 * @param event the datasource event 
		 * @return true if this event was handled
		 * @see LyDataSourceListener#objectModified
		 */
		public boolean objectModified(LyDataSourceEvent event)
		{
			if (event.isLocal() || !event.isNotified())
				return false;

			LyObject array = (LyObject)event.getObject();
			LyObject source = getSource(array);

			if (source != null)
				addRelation(source, array, true, event);
			return true;
		}
	} // end class _LyArrayClassListener
} // end class LyArrayFieldInfoDaemon
