/*
 * LyDaemon.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data.daemon;

import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;


/**
 * Daemon for managing objects.
 */
public abstract class LyDaemon
	implements LyDataSourceListener
{
	/** File version */
	public static final String __VERSION = "$Id: LyDaemon.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Class containing the info. */
	LyClassInfo				   _sourceClass;
    
    /** Info to which the daemon is attached to (could be null). */
    LyInfo                     _info;
    

    /**
     * Daemon constructor.
     * @param sourceClass the source class for this daemon
     */
    protected LyDaemon(LyClassInfo sourceClass)
    {
        this(sourceClass, null);
    }
    
	/**
	 * Default constructor.
	 * @param sourceClass the source class
	 * @param info the source info
	 */
	protected LyDaemon(LyClassInfo sourceClass, LyInfo info)
	{
		if (sourceClass == null)
			throw new IllegalStateException("BF AppEngine : sourceClass parameter can't be null");

		_sourceClass = sourceClass.getRealClass();
        _info = (info != null ? info : _sourceClass);

		// Listen to data source for class containing the relation
		LyObjectSource source;

		source = _sourceClass.getObjectSource();
		source.addListener(this);

		// Register the daemon
		LyApplication application = sourceClass.getApplication();

		application.addDaemon(this);
	}

	/**
	 * Unloads this daemon if it internally handles data. This method is called when the application
	 * is completely unloaded.
	 */
	public void unload()
	{
	}

	/**
	 * To get the current environment of this daemon.
	 *
	 * @return The current environment of this daemon.
	 */
	protected LyEnvironment getEnvironment()
	{
		LyApplication application = _sourceClass.getApplication();

		return application.getEnvironment();
	}

	/**
	 * Accessor to source class.
	 *
	 * @return The source class of this daemon.
	 */
	public LyClassInfo getSourceClass()
	{
		return _sourceClass;
	}

	/**
	 * Called when objectId of an object has been changed. This method must update tables with
	 * references to objectIds.
	 *
	 * @param oldObjectId
	 * @param newObjectId
	 */
	public abstract void idRehashed(LyObjectId oldObjectId, LyObjectId newObjectId);

	/* Implementation of LyDataSourceListener */

	/**
	 * Method for object creation callback.
	 *
	 * @param event the datasource event 
	 * @return true if this event was handled
	 * @see LyDataSourceListener#objectCreated
	 */
	public boolean objectCreated(LyDataSourceEvent event)
	{
		return false;
	}

	/**
	 * Method for object deletion callback.
	 *
	 * @param event the datasource event 
	 * @return true if this event was handled
	 * @see LyDataSourceListener#objectDeleted
	 */
	public boolean objectDeleted(LyDataSourceEvent event)
	{
		return false;
	}

	/**
	 * Method for object modification callback.
	 *
	 * @param event the datasource event 
	 * @return true if this event was handled
	 * @see LyDataSourceListener#objectModified
	 */
	public boolean objectModified(LyDataSourceEvent event)
	{
		return false;
	}
    
    /**
     * Gets info to which the daemon is associated.
     * @return the info associated to this daemon
     */
    public LyInfo getInfo()
    {
        return _info;
    }
} // end class LyDaemon
