/*
 * LyJoinRelationDaemon.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data.daemon;

import java.util.*;
import leon.data.*;
import leon.info.*;


/**
 * Daemon that handles multiple relations managed with join elements.
 */
public class LyJoinRelationDaemon extends LyRelationDaemon
{
	/** File version */
	public static final String    __VERSION =
		"$Id: LyJoinRelationDaemon.java 41520 2012-11-13 17:18:51Z jlsaouli $";

	/** Limit for optimizing request to find targets. */
	public static int			 OPTIM_REQUEST_LIMIT = 20;

	/** Join class. */
	protected LyClassInfo		  _joinClass;

	/** Join relations. Source relation*/
	protected LyRelationFieldInfo _sourceRelation;
	/** Join relations. Target relation*/
	protected LyRelationFieldInfo _targetRelation;

	/**
	 * Boolean indicating if value corresponding to sourceRelation must be put first when
	 * constructing objectId for joinClass.
	 */
	protected boolean			  _sourceFirst;

	/** Use to keep already computed objects */
	protected Hashtable<LyObjectId, LySimpleObjectList>			_computedList; // Force keeping computed in cache because of CACHE AUTO.


	/**
	 * Default constructor.
	 * @param sourceClass The source class
	 * @param relation  the relation handled by the daemon
	 * @param joinClass the join class
	 * @param sourceRelation the source relation in the join class
	 * @param targetRelation the target relation in the join class
	 */
	public LyJoinRelationDaemon(LyClassInfo sourceClass,
		LyRelationFieldInfo relation, LyClassInfo joinClass,
		LyRelationFieldInfo sourceRelation, LyRelationFieldInfo targetRelation)
	{
		super(sourceClass, relation, false);

		if (joinClass == null)
			throw new IllegalArgumentException("BF AppEngine : joinClass can't be null");
		if (sourceRelation == null)
			throw new IllegalArgumentException("BF AppEngine : sourceRelation can't be null");
		if (targetRelation == null)
			throw new IllegalArgumentException("BF AppEngine : targetRelation can't be null");

		_joinClass = joinClass;
		_sourceRelation = sourceRelation;
		_targetRelation = targetRelation;
		_sourceFirst = (_joinClass.getIndex(_sourceRelation)
					< _joinClass.getIndex(_targetRelation));

		if (!_sourceRelation.relatedWith(sourceClass))
			throw new IllegalStateException("BF AppEngine : invalid relation "
				+ _sourceRelation + " for this daemon : relation is not related with class "
				+ sourceClass.getId());

		// Listen to data source for join class and its peer class
		LyObjectSource source;

		source = _joinClass.getObjectSource();
		source.addListener(this);

		// Listen to deletion on data sources for classes pointed by the relation
		LyClassInfo classInfo;
		int		    i, n = relation.getClassInfoCount();

		for (i = 0; i < n; i++)
		{
			classInfo = relation.getClassInfo(i);
			source = classInfo.getObjectSource();
			source.addListener(this, LyDataSourceEvent.TYPE_DELETE);
		}
	}

	/**
	 * Get the join objects associated to the given object
	 * @param source the source object
	 * @return a list of join objects.
	 */
	public LyObjectList getJoins(LyObject source)
	{
		LyObjectList list = source.getList(null, _joinClass);

		return list;
	}

	/**
	 * To put a result of the calculated value of the relation in a cache.
	 *
	 * @param object the associated object of the calulated value.
	 * @param result the calculated value.
	 */
	protected void putResult(LyObject object, Object result)
	{
		if (object.getPeer() == null) // XXX Is it really useful
			// Don't register local objects
			return;

		super.putResult(object, result);
	}

	/**
	 * Method to add a relation between a source object and a target object.
	 *
	 * @param object Source object.
	 * @param target Target object.
	 * @param source Boolean indicating if this adding results on a modification
	 * @param event  optional event indicating when notified for instance, may be null on source
	 *               object or not.
	 */
	protected void addRelation(LyObject object, LyObject target, boolean source,
		LyDataSourceEvent event)
	{
		if ((object.getPeer() == null) || (target.getPeer() == null))
			return;

		LyObjectList targets = getTargets(object);

		if ((targets != null) && targets.contains(target))
			return;

		LyObject join = _joinClass.createInstance();

		join.setValue(_sourceRelation, object.getObjectId());
		join.setValue(_targetRelation, target.getObjectId());
		join.create(event.isNotified());
	}

	/**
	 * Method to remove a relation between a source object and a target object.
	 *
	 * @param object Source object.
	 * @param target Target object.
	 * @param source Boolean indicating if the removal results on a modification on source object or
	 *               not.
	 * @param event  optional event indicating when notified for instance, may be null
	 */
	protected void removeRelation(LyObject object, LyObject target, boolean source,
		LyDataSourceEvent event)
	{
		LyObjectList targets = getTargets(object);

		if ((targets == null) || !targets.contains(target))
			return;

        if (_sourceRelation.hasMark(LyFieldInfo.MARK_ID)
            && _targetRelation.hasMark(LyFieldInfo.MARK_ID))
        {
            String[] stringIds = new String[2];

            stringIds[_sourceFirst ? 0 : 1] = object.getObjectId().toString();
            stringIds[_sourceFirst ? 1 : 0] = target.getObjectId().toString();

            LyObjectId joinId = new LyObjectId(_joinClass, stringIds);
            LyObject   join = (LyObject)joinId.getObject();

            // if the join object is null, it has already been deleted,
            // then clean up the relation on source object with a call to the super method.
            if ((join == null) && (!object.isDeleted()))
                super.removeRelation(object, target, source, event);
            else if (join != null)
            {
            	join.delete(event.isNotified());
            }
        }
        else
        {
            LyFilter  filter = new LyFilter();

            filter.addElement(new LyFilterExpression(_sourceRelation, object.getObjectId()));
            filter.addElement(new LyFilterExpression(_targetRelation, target.getObjectId()));

            LyObjectSource    joinSource = _joinClass.getObjectSource();
            LyResponse        response = joinSource.getObjects(filter);
            LyObjectList      joins = response.getObjects();

            joins.delete();
        }
	}

	/**
	 * Abstract method for calculating objects related to the given object.
	 * @param target the target object
	 * @param fromResults
	 * @return a list of source objects
	 */
	protected LyObjectList findSources(LyObject target, boolean fromResults)
	{
		// Do not use classic cross-references computation to avoid putting
		// the target object in the cache if it is under deletion.
		// Then, look for references of the target through all the results already stored.
		LySimpleObjectList       result = new LySimpleObjectList();
		LyObjectId		         sourceId;
		Object			         value;
		LyObject		         source;
		Enumeration<LyObjectId>  e = _results.keys();

		while (e.hasMoreElements())
		{
			sourceId = e.nextElement();
			source = sourceId.getLogicalObject();
			if (source == null)
				continue;

			value = _results.get(sourceId);
			if ((value == target) // LyObject
					|| ((value instanceof LyObjectList)
						&& ((LyObjectList)value).contains(target)))
				result.addObject(source);
		}

		return result;
	}

	/**
	 * Abstract method for calculating objects related to the given object.
	 * @param source the source object
	 * @return a list of target objects
	 */
	protected LyObjectList findTargets(LyObject source)
	{
		if (source.getPeer() == null)
			return null;

		LyFilter     filter = computeFilter(source);
        LyObjectList list = new LyClassWorkSpace(null, _joinClass, filter, false);

		if (_computedList == null)
			_computedList = new Hashtable<LyObjectId, LySimpleObjectList>();

		// Keep reference of join objects to avoid finalization.
		_computedList.put(source.getObjectId(), new LySimpleObjectList(list));

        LyObject           join, target;
        LyRelationValue    rValue;
        int                i, n = list.getSize();

	    if ((n > 1)
		    && (n <= OPTIM_REQUEST_LIMIT)
	        && (_targetRelation.getClassInfoCount() == 1))
	    {
            LyClassInfo        targetClassInfo = _targetRelation.getClassInfo(0);
            LyObjectSource     targetObjectSource = targetClassInfo.getObjectSource();
            LyCache            cache = targetObjectSource.getCache();

    		if ((cache != null)
    		    && !(cache instanceof leon.data.cache.LyFullCache)
    		    && !cache.isLoaded())
    		{
    		    // Optimization, build a filter for loading objects with a single request
    		    // instead of loading them one by one by their id.
    		    LyInfoList<LyFieldInfo>  idFields = targetClassInfo.getIdFieldInfos();
    		    LyFieldInfo              idField;
    		    LyFilter                 targetFilter = new LyFilter(LyFilter.Type.OR), subFilter;
    		    int                      j, nIdFields = idFields.getSize();
    		    LyObjectId               targetId;
    		    String[]                 ids;

                for (i = 0; i < n; i++)
                {
                    join = list.getObject(i);
                    rValue = join.getRelationValue(_targetRelation);
                    targetId = rValue.getObjectId(0);

                    if (nIdFields == 1)
                    {
                        idField = idFields.getInfo(0);
                        targetFilter.addElement(new LyFilterExpression(idField, targetId.toString()));
                    }
                    else
                    {
                        ids = targetId.toStrings();
                        subFilter = new LyFilter();

            		    for (j = 0; j < nIdFields; j++)
            		    {
            		        idField = idFields.getInfo(j);
            		        subFilter.addElement(new LyFilterExpression(idField,
            		            (j < ids.length ? ids[j] : null)));
            		    }

            		    targetFilter.addElement(subFilter);
                    }
                }

                LyObjectList result = new LyClassWorkSpace(null, targetClassInfo, targetFilter, false);

                return new LySimpleObjectList(result);
    		}
	    }

        LySimpleObjectList result = null;

		for (i = 0; i < n; i++)
		{
			join = list.getObject(i);
			rValue = join.getRelationValue(_targetRelation);

			if (rValue.isNull() || (rValue.getValueCount() == (short)0))
				continue;
			target = rValue.getObject(0);
			if (target == null)
				continue;

			if (result == null)
				result = new LySimpleObjectList(target);
			else
				result.addObject(target);
		}

		return result;

	} // end method findTargets

	/**
	 * Internal method to compute the filter needed to get join objects form a source object.
	 *
	 * @param  source The source object.
	 * @return The filter.
	 */
	protected LyFilter computeFilter(LyObject source)
	{
		LyFilter		   filter = new LyFilter();
		LyFilterExpression expression;
		BitSet			   modifiers = new BitSet();

		modifiers.set(LyFilterExpression.MOD_CASE_SENSITIVE);
		expression = new LyFilterExpression(_sourceRelation, false, LyFilterExpression.Condition.EQUALS,
				source.getObjectId(), modifiers);
		filter.addElement(expression);

		return filter;
	}

	/* Implementation of LyDataSourceListener */

	/**
	 * Method for object creation callback.
	 *
	 * @param event the datasource event
	 * @return true if this event was handled
	 * @see LyDataSourceListener#objectCreated
	 */
	public boolean objectCreated(LyDataSourceEvent event)
	{
		if (event.isLocal())
			return false;

		LyObject object = (LyObject)event.getObject();

		if (!object.instanceOf(_joinClass))
			return super.objectCreated(event);

		// When a rollBack occurs, we receive a load event.
		// In this case, we must ignore the loading parameter
		// and refresh the daemon
		boolean isRollBack = false;
		LyClass peerClass = object.getClassInfo().getPeerClass();

		if (peerClass != null)
		{
			LyDataSource ds = peerClass.getDataSource();

			isRollBack = ds.isRollBack();
		}
		else
			isRollBack = true; // Force refresh of the daemon

		if (event.isLoading() && !isRollBack && (!event.isNotified()))
			return false;

		LyObjectId objectId;
		LyObject   source, target;

		objectId = (LyObjectId)object.getValue(_sourceRelation);
		source = (objectId == null ? null : (LyObject)objectId.getObject());

		if ((source != null) && !contains(source))
			return false;

		objectId = (LyObjectId)object.getValue(_targetRelation);
		target = (objectId == null ? null : (LyObject)objectId.getObject());

		if ((source != null) && (target != null))
		{
			LySimpleObjectList list = ((_computedList == null)
					? null : _computedList.get(source.getObjectId()));

			if (list == null)
			{
				list = new LySimpleObjectList(object);
				if (_computedList == null)
					_computedList = new Hashtable<LyObjectId, LySimpleObjectList>();
				_computedList.put(source.getObjectId(), list); // Keep reference to avoid finalize.
			}
			else
				list.addObject(object);

			super.addRelation(source, target, false, event);
		}
		return true;
	} // end method objectCreated

	/**
	 * Method for object deletion callback.
	 *
	 * @param event the datasource event
	 * @return true if this event was handled
	 * @see LyDataSourceListener#objectDeleted
	 */
	public boolean objectDeleted(LyDataSourceEvent event)
	{
		if (event.isLocal())
			return false;

		/*if (event.isLoading())
		{
		        // FFT00949
		        // A deletion event is 'loading' when it's related to an unload() request
		        // This may occur when going into sleepingAPP mode or on
		        // User request. Do not propagate on related objects to avoid loops.
		        return true;
		}*/

		LyObject object = (LyObject)event.getObject();

		if (!object.instanceOf(_joinClass))
		{
			if ((_computedList != null)
					&& object.instanceOf(_sourceClass))
				_computedList.remove(object.getObjectId());
			return super.objectDeleted(event);
		}

		LyObjectId objectId;
		LyObject   source, target;

		objectId = (LyObjectId)object.getValue(_sourceRelation);
		source = (objectId == null ? null : (LyObject)objectId.getObject());

		objectId = (LyObjectId)object.getValue(_targetRelation);
		target = (objectId == null ? null : (LyObject)objectId.getObject());

		if ((source != null) && (target != null))
		{
			LySimpleObjectList list = ((_computedList == null)
					? null : _computedList.get(source.getObjectId()));

			if (list == null)
			{
				LyFilter filter = computeFilter(source);

				list = new LySimpleObjectList(new LyClassWorkSpace(null, _joinClass, filter,
							false));
				if (_computedList == null)
					_computedList = new Hashtable<LyObjectId, LySimpleObjectList>();
				_computedList.put(source.getObjectId(), list); // Keep reference to avoid finalize.
			}
			else if (list.contains(object))
				list.removeObject(object);

			super.removeRelation(source, target, false, event);
		}

		return true;
	} // end method objectDeleted

	/**
	 * Method for object modification callback.
	 *
	 * @param event the datasource event
	 * @return true if this event was handled
	 * @see LyDataSourceListener#objectModified
	 */
	public boolean objectModified(LyDataSourceEvent event)
	{
		if (event.isLocal())
			return false;

		LyObject object = (LyObject)event.getObject();

		if (!object.instanceOf(_joinClass))
			return super.objectModified(event);

		// Normally, modifications on join classes are not supported
		// XXX To be discussed : some modification are valid when the name of
		// a target object has changed.
		// throw new IllegalStateException("BF AppEngine : receive modification on joinClass");

		return true;
	}

	/**
	 * To free a calculated value for the specified object.
	 *
	 * @param object the object we want to free the value.
	 */
	public void freeValue(LyObject object)
	{
		if ((_computedList != null) && (object != null))
			_computedList.remove(object.getObjectId());

		super.freeValue(object);
	}

	/**
	 * Unload daemon.
	 */
	public void unload()
	{
		super.unload();
		if (_computedList != null)
			_computedList.clear();
	}
} // end class LyJoinRelationDaemon
