/*
 * LyRoutedRelationDaemon.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data.daemon;

import java.util.*;

import leon.app.*;
import leon.data.*;
import leon.info.*;


/**
 * Daemon for updating dynamic labels.
 */
public class LyLabelDaemon extends LyDaemon
    implements LyWorkSpaceListener,
        LyWorkSpaceCounterListener,
        LyLabelDaemonListener
{
    /** File version. */
    public static final String __VERSION = "$Id: LyLabelDaemon.java 28980 2010-12-22 10:53:09Z jlsaouli $";

    /** Label managed by this daemon. */
    LyInfoLabel _label;

    /** Results tables. */
    Hashtable<LyObjectId, Object> _results;
    Hashtable<LyObjectId, String> _labels;

    /** Listeners of this daemon. */
    Vector<LyLabelDaemonListener> _listeners;

    /** Boolean indicating if label contains paths that have steps. */
    boolean _hasSteps;

    /** Global workspace with all sources. */
    LySimpleWorkSpace _sourcesWorkSpace;

    /** Global workspace with all targets. */
    LyWorkSpace _targetsWorkSpace;

    /** Boolean indicating if a refresh is being performed (to avoid loops). */
    boolean _refreshing;

    /** Static object for storing an empty list in results hashtable. */
    static Object EMPTY_LIST = new Object();

    /**
     * Default constructor.
     *
     * @param  sourceClass  Class that contains the label.
     * @param  label        The label to watch on.
     */
    public LyLabelDaemon(LyClassInfo sourceClass, LyInfoLabel label)
    {
        this(sourceClass, label, null);
    }

    /**
     * Constructor of a label daemon with a specific listener.
     *
     * @param  sourceClass  Class that contains the label.
     * @param  label        The label to watch on.
     * @param  listener     the label daemon listener
     */
    public LyLabelDaemon(LyClassInfo sourceClass, LyInfoLabel label, LyLabelDaemonListener listener)
    {
        super(sourceClass, label);

        if (label == null)
            throw new IllegalStateException("BF AppEngine : label parameter can't be null");

        _label = label;

        _results = new Hashtable<LyObjectId, Object>();
        _labels = new Hashtable<LyObjectId, String>();
        _hasSteps = false;

        if (listener != null)
            addListener(listener);

        // Register as listener of labels specified on paths
        LyInfoList<LyPath> paths = label.getPaths();
        int                i, n = ((paths == null) ? 0 : paths.getSize());
        LyPath             path;
        LyInfoLabel        pathLabel;
        LyApplication      application = sourceClass.getApplication();
        LyLabelDaemon      daemon;

        for (i = 0; i < n; i++)
        {
            path = paths.getInfo(i);
            if (!_hasSteps)
                _hasSteps = (path.getSteps() != null);

            pathLabel = path.getFieldLabel();
            if ((pathLabel != null)
                    && (pathLabel != label))
            {
                if (pathLabel == label)
                    throw new IllegalStateException("BF AppEngine : invalid label "
                        + label + ", it contains a path that references this label");

                daemon = application.getDaemon(pathLabel);
                if (daemon == null)
                    daemon = new LyLabelDaemon(path.getTarget(), pathLabel, this);
                else
                    daemon.addListener(this);
            }
        }

        _sourcesWorkSpace = null;
        _targetsWorkSpace = null;
    }

    /**
     * Accessor to the label managed by this daemon.
     *
     * @return  the label handled by this daemon
     */
    public LyInfoLabel getInfoLabel()
    {
        return _label;
    }

    /**
     * To get the relation value for the specified object.
     *
     * @param   object  the object to compute
     *
     * @return  the label value
     */
    public String getString(LyObject object)
    {
        if (object == null)
            return null;

        LyObjectId id = object.getObjectId();
        String     label = _labels.get(id);

        if (label == null)
        {
            label = computeString(object);
            if (label != null)
                _labels.put(id, label);
        }

        return label;
    }

    /**
     * Compute the label for the given object.
     *
     * @param   object  the object
     *
     * @return  the label
     */
    public String computeString(LyObject object)
    {
        if (object == null)
            throw new IllegalArgumentException("BF AppEngine : object can't be null");

        Object result = getResult(object);

        if (result == null)
        {
            // We must compute the value.
            // In this case, nobody have ask for the value.
            LyObjectList targets = findTargets(object);

            if (targets == null)
                putResult(object, EMPTY_LIST);
            else
                // Put the result in the cache.
                putResult(object, targets);
        }

        // Return the value.
        String label = _label.computeString(object);

        return label;
    }

    /**
     * Get a result of the calculated value of the relation in a cache.
     *
     * @param   object  the associated object of the calulated value.
     *
     * @return  the calculated value.
     */
    Object getResult(LyObject object)
    {
        LyObjectId id = object.getObjectId();

        return _results.get(id);
    }

    /**
     */
    void putResult(LyObject object, Object targets)
    {
        if (!object.checkInstance())
            // Never store an object that is not found in cache
            // This may occur when checking filter on peer object.
            return;

        LyObjectId id = object.getObjectId();

        _results.put(id, targets);
    }

    /**
     * Remove an object from our objects list.
     *
     * @param  object  the object to remove
     */
    public void removeResult(LyObject object)
    {
        LyObjectId id = object.getObjectId();

        _results.remove(id);
        _labels.remove(id);

        if (_sourcesWorkSpace != null)
            _sourcesWorkSpace.removeObject(object);
    }

    void updateResult(LyObject object, LyObject target, boolean add)
    {
        Object             result = getResult(object);
        LySimpleObjectList list;

        if ((result == null)
                || (result == EMPTY_LIST))
        {
            if (!add)
            {
                // May never occured ?
            }
            else
            {
                list = new LySimpleObjectList(target);
                putResult(object, list);
            }
        }
        else
        {
            list = new LySimpleObjectList((LyObjectList)result);
            if ((add) && (!list.contains(target)))
                list.addObject(target);
            else if (list.contains(target))
                list.removeObject(target);

            if (list.getSize() == 0)
            {
                LyObjectId id = object.getObjectId();

                _results.put(id, EMPTY_LIST);
            }
            else
                putResult(object, list);
        }
    } // end method updateResult

    /**
     * To retrieve all objects by a target.
     */
    LyObjectList findSources(LyObject target, boolean fromResults)
    {
        if ((_targetsWorkSpace == null) && (!fromResults))
            return null;
        else if ((_results == null) && (fromResults))
            return null;
        else if (fromResults)
        {
            LySimpleObjectList      result = null;
            Enumeration<LyObjectId> e = _results.keys();
            LyObjectId              sourceId;
            LyObject                source;
            Object                  r;

            while (e.hasMoreElements())
            {
                sourceId = e.nextElement();
                source = sourceId.getLogicalObject();
                if (source == null)
                    continue;

                r = getResult(source);
                if ((r instanceof LyObject)
                        && (r == target))
                {
                    if (result == null)
                        result = new LySimpleObjectList(source);
                    else
                        result.addObject(source);
                }
                else if ((r instanceof LyObjectList)
                        && ((LyObjectList)r).contains(target))
                {
                    if (result == null)
                        result = new LySimpleObjectList(source);
                    else
                        result.addObject(source);
                }
            }

            return result;
        }
        else
            return _targetsWorkSpace.getSources(target);
    } // end method findSources

    /**
     * Method for calculating objects related to the given object.
     */
    LyObjectList findTargets(LyObject source)
    {
        if (!_hasSteps)
            return null;

        LyInfoList<LyPath> paths = _label.getPaths();

        if (_sourcesWorkSpace == null)
        {
            _sourcesWorkSpace = new LySimpleWorkSpace(null, source, true);
            _targetsWorkSpace = (LyWorkSpace)_sourcesWorkSpace.getList(null, _sourceClass, paths, true);
            _targetsWorkSpace.addListener(this);

            if (_targetsWorkSpace.getSize() > 0)
                return new LySimpleObjectList(_targetsWorkSpace);
            else
                return null;
        }

        _sourcesWorkSpace.addObject(source);

        LyObjectList list = new LySimpleObjectList(source);
        LyObjectList targets = list.getList(null, _sourceClass, paths, false);

        if (targets == null)
            return null;

        return new LySimpleObjectList(targets);
    } // end method findTargets

    /**
     * Unload daemon.
     */
    public void unload()
    {
        if (_targetsWorkSpace != null)
            _targetsWorkSpace.free(this);

        if (_sourcesWorkSpace != null)
            _sourcesWorkSpace.free(null);

        _results.clear();
        _labels.clear();
    }

    /**
     * Checks whether a given object is contained in our objects list.
     */
    boolean contains(LyObject object)
    {
        LyObjectId id = object.getObjectId();

        return _results.containsKey(id);
    }

    /* Implementation of LyDataSourceListener */

    /**
     * Method for object creation callback.
     *
     * @param   event  the datasource event
     *
     * @return  true if this event was handled
     *
     * @see     LyDataSourceListener#objectCreated
     */
    public boolean objectCreated(LyDataSourceEvent event)
    {
        return false;
    }

    /**
     * Method for object deletion callback.
     *
     * @param   event  the datasource event
     *
     * @return  true if this event was handled
     *
     * @see     LyDataSourceListener#objectDeleted
     */
    public boolean objectDeleted(LyDataSourceEvent event)
    {
        LyObject object = (LyObject)event.getObject();

        // If the object is not in the list, ignore this event
        if (!contains(object))
            return false;

        // Remove object from results table
        removeResult(object);
        return true;
    }

    /**
     * Method for object modification callback.
     *
     * @param   event  the datasource event
     *
     * @return  true if this event was handled
     *
     * @see     LyDataSourceListener#objectModified
     */
    public boolean objectModified(LyDataSourceEvent event)
    {
        if (_refreshing)
            // Loop in labels definition
            return false;

        LyObject   object = (LyObject)event.getObject();
        LyObjectId id = object.getObjectId();

        // If the object is not in the list, ignore this event
        if (!contains(object))
            return false;

        LyValueSet values = event.getValues();
        String     oldLabel = _labels.get(id);
        String     newLabel = computeString(object);

        if ((oldLabel == newLabel)
                || ((oldLabel != null) && (oldLabel.equals(newLabel))))
            return false;

        _labels.put(id, newLabel);
        if (values == null)
        {
            refresh(object);
            return true;
        }

        LyValue     value;
        LyFieldInfo fieldInfo;
        int         i, n = values.getSize();

        for (i = 0; i < n; i++)
        {
            value = values.getFieldValue(i);
            fieldInfo = (LyFieldInfo)value.getField();
            if (_label.contains(fieldInfo))
            {
                refresh(object);
                return true;
            }
        }

        // XXX How optimize the case of a modified relation (a step in a path) ?
        refresh(object);
        return true;
    } // end method objectModified

    /* Implementation of LyWorkSpaceListener */

    /**
     * Called when an object is added in the dynamic object list.
     *
     * @param   event  event of adding object
     *
     * @return  the event is used by the listener
     */
    public boolean objectAdded(LyWorkSpaceEvent event)
    {
        return processEvent(event, true);
    }

    /**
     * Called when an object is removed from the dynamic object list.
     *
     * @param   event  event of removing object
     *
     * @return  the event is used by the listener
     */
    public boolean objectRemoved(LyWorkSpaceEvent event)
    {
        return processEvent(event, false);
    }

    /**
     * Called when the counter associated with the specified object has been incremented.
     *
     * @param   event  Object on which the was incremented.
     *
     * @return  True if the event is used by the listener, false otherwise.
     */
    public boolean counterChanged(LyWorkSpaceEvent event)
    {
        return processEvent(event, event.isIncremented());
    }

    /**
     * Internal method to process a workspace event for adding or removing a target.
     *
     * @param   event  the workspace event to process
     * @param   add    true to add a target
     *
     * @return  true if event was handled
     */
    protected boolean processEvent(LyWorkSpaceEvent event, boolean add)
    {
        LyObject     target = event.getObject();
        LyObjectList wsSources = findSources(target, false);
        LyObjectList sources = null;
        int          i, n;
        LyObject     object;

        if (add)
            sources = wsSources;
        else
        {
            LyObjectList resultSources = findSources(target, true);

            if (resultSources != null)
            {
                if (wsSources == null)
                    sources = new LySimpleObjectList(resultSources);
                else
                {
                    // Only sources in resultSources that are not in wsSources
                    // must be updated.
                    n = resultSources.getSize();

                    for (i = 0; i < n; i++)
                    {
                        object = resultSources.getObject(i);
                        if (!wsSources.contains(object))
                        {
                            if (sources == null)
                                sources = new LySimpleObjectList();

                            ((LySimpleObjectList)sources).addObject(object);
                        }
                    }
                }
            }
        }

        String     oldLabel, newLabel;
        boolean    refresh;
        LyObjectId id;

        n = ((sources == null) ? 0 : sources.getSize());
        for (i = 0; i < n; i++)
        {
            object = sources.getObject(i);
            id = object.getObjectId();
            refresh = (getResult(object) != null);
            oldLabel = _labels.get(id);
            newLabel = computeString(object);

            // Update target list.
            updateResult(object, target, add);

            if ((oldLabel == newLabel)
                    || ((oldLabel != null) && (oldLabel.equals(newLabel))))
                continue;

            _labels.put(id, newLabel);

            if (refresh)
                refresh(object);
        }

        return true;
    } // end method processEvent

    /**
     * Called when an object has changed in the dynamic object list.
     *
     * @param   event  event of changing object
     *
     * @return  the event is used by the listener
     */
    public boolean objectChanged(LyWorkSpaceEvent event)
    {
        LyObject     target = event.getObject();
        LyObjectList sources = findSources(target, false);
        LyObject     object;
        int          i, n = ((sources == null) ? 0 : sources.getSize());
        String       oldLabel, newLabel;
        LyValueSet   values = event.getValues();
        LyObjectId   id;

        for (i = 0; i < n; i++)
        {
            object = sources.getObject(i);
            id = object.getObjectId();
            oldLabel = _labels.get(id);
            newLabel = computeString(object);

            if ((oldLabel == newLabel)
                    || ((oldLabel != null) && (oldLabel.equals(newLabel))))
                continue;

            _labels.put(id, newLabel);
            if (values == null)
            {
                refresh(object);
                continue;
            }

            LyValue     value;
            LyFieldInfo fieldInfo;
            int         j, p = values.getSize();

            for (j = 0; j < p; j++)
            {
                value = values.getFieldValue(j);
                fieldInfo = (LyFieldInfo)value.getField();
                if (_label.contains(fieldInfo))
                {
                    refresh(object);
                    continue;
                }
            }

            // XXX How optimize the case of a modified relation (a step in a path) ?
            refresh(object);
        }

        return true;
    } // end method objectChanged

    /**
     * To set the relation value with new values.
     */
    void refresh(LyObject object)
    {
        LyInfoList<LyFieldInfo> fields = null;

        // XXX : optimize the refresh depending the label type
        LyInfoLabel.Type type = _label.getType();

        if ((_label.getParent() instanceof LyFieldInfo)
                && ((type == LyInfoLabel.Type.VALUE)
                    || (type == LyInfoLabel.Type.SHORT_VALUE)
                    || (type == LyInfoLabel.Type.GRAMMATICAL)
                    || (type == LyInfoLabel.Type.USER)))
        {
            LyFieldInfo fieldInfo = (LyFieldInfo)_label.getParent();
            LyClassInfo classInfo = object.getClassInfo();

            fieldInfo = classInfo.getRealFieldInfo(fieldInfo);
            if (fieldInfo != null)
            {
                fields = new LyInfoList<LyFieldInfo>();
                fields.addInfo(fieldInfo);
            }
        }

        if (_listeners != null)
        {
            LyLabelDaemonListener listener;
            int                   i, n = _listeners.size();

            for (i = 0; i < n; i++)
            {
                listener = _listeners.elementAt(i);
                listener.refresh(object, fields);
            }
        }

        if (type != LyInfoLabel.Type.USER)
        {
            // No refresh on objects on user labels
            _refreshing = true;
            object.refresh(fields);
            _refreshing = false;
        }
    } // end method refresh

    /**
     * Adds a listener of this daemon.
     *
     * @param  listener  The listener of this daemon.
     */
    public void addListener(LyLabelDaemonListener listener)
    {
        if (listener == null)
            throw new IllegalArgumentException("BF AppEngine : listener can't be null");

        if (_listeners == null)
            _listeners = new Vector<LyLabelDaemonListener>(1);

        _listeners.add(listener);
    }

    /**
     * Removes a listener from this daemon.
     *
     * @param  listener  The listener of this daemon.
     */
    public void removeListener(LyLabelDaemonListener listener)
    {
        if (listener == null)
            throw new IllegalArgumentException("BF AppEngine : listener can't be null");

        if ((_listeners == null)
                || (!_listeners.contains(listener)))
            throw new IllegalArgumentException("BF AppEngine : listener is unknown");

        _listeners.remove(listener);
    }

    /**
     * Get listeners of this daemon.
     *
     * @return  The listeners of this daemon.
     */
    public Vector<LyLabelDaemonListener> getListeners()
    {
        return ((_listeners == null) ? null : new Vector<LyLabelDaemonListener>(_listeners));
    }

    /**
     * Get listener of this daemon at the given position.
     *
     * @param   index  index of the requested listener.
     *
     * @return  The listeners of this daemon.
     */
    public LyLabelDaemonListener getListener(int index)
    {
        return ((_listeners == null) ? null : (LyLabelDaemonListener)_listeners.elementAt(index));
    }

    /**
     * Get listeners count of this daemon.
     *
     * @return  The number of listeners of this daemon.
     */
    public int getListenerCount()
    {
        return ((_listeners == null) ? 0 : _listeners.size());
    }

    /**
     * Method called when a refresh is requiered.
     *
     * @param  object  Object to refresh.
     * @param  fields  Impacted fields (may be null).
     */
    public void refresh(LyObject object, LyInfoList<LyFieldInfo> fields)
    {
        if (_results == null)
            return;

        LyObjectSource    source = object.getObjectSource();
        LyValueSet        values = object.getValues(fields);
        LyDataSourceEvent event = new LyDataSourceEvent(source, LyDataSourceEvent.TYPE_SET, object, values);

        objectModified(event);
    }

    /**
     * Called when objectId of an object has been changed. This method must update tables with
     * references to objectIds.
     *
     * @param  oldObjectId
     * @param  newObjectId
     */
    public void idRehashed(LyObjectId oldObjectId, LyObjectId newObjectId)
    {
        if ((_results != null)
                && _results.containsKey(oldObjectId))
        {
            Object value = _results.remove(oldObjectId);

            _results.put(newObjectId, value);
        }

        if ((_labels != null)
                && _labels.containsKey(oldObjectId))
        {
            String value = _labels.remove(oldObjectId);

            _labels.put(newObjectId, value);
        }
    }
} // end class LyLabelDaemon
