/*
 * LyPathDaemon.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data.daemon;

import java.util.*;

import leon.data.*;
import leon.info.*;


/**
 * Daemon for path on field info.
 */
public class LyPathDaemon extends LyFieldInfoDaemon
    implements LyWorkSpaceListener,
        LyWorkSpaceCounterListener
{
    /** File version. */
    public static final String __VERSION = "$Id: LyPathDaemon.java 28980 2010-12-22 10:53:09Z jlsaouli $";

    /** Path used for calculating the field info value. */
    LyPath _path;

    /** Results tables. */
    Hashtable<LyObjectId, Object> _results;
    Hashtable<LyObjectId, Object> _values;

    /** Global workspace with all sources. */
    LySimpleWorkSpace _sourcesWorkSpace;

    /** Global workspace with all targets. */
    LyWorkSpace _targetsWorkSpace;

    /** Static object for storing an empty list in results hashtable. */
    static Object EMPTY_LIST = new Object();
    static Object EMPTY_VALUE = new Object();

    /**
     * Default constructor.
     *
     * @param  sourceClass  the source class
     * @param  fieldInfo    the fieldInfo
     */
    public LyPathDaemon(LyClassInfo sourceClass, LyFieldInfo fieldInfo)
    {
        super(sourceClass, fieldInfo);

        _path = fieldInfo.getPath();
        _values = new Hashtable<LyObjectId, Object>();
        _results = new Hashtable<LyObjectId, Object>();
        _sourcesWorkSpace = null;
        _targetsWorkSpace = null;
    }

    /**
     * To get the value for the specified object.
     *
     * @param   object  the object we want the value.
     *
     * @return  the value of field info for the specified object.
     */
    public Object getValue(LyObject object)
    {
        if (object == null)
            throw new IllegalArgumentException("BF AppEngine : object can't be null");

        LyObjectId id = object.getObjectId();
        Object     value = _values.get(id);

        if (value == null)
        {
            value = computeValue(object);
            if (value == null)
                _values.put(id, EMPTY_VALUE);
            else
                _values.put(id, value);
        }
        else if (value == EMPTY_VALUE)
            return null;

        return value;
    }

    /**
     * To get the relation value for the specified object.
     *
     * @param   object  the object
     *
     * @return  the computed value
     */
    public Object computeValue(LyObject object)
    {
        if (object == null)
            throw new IllegalArgumentException("BF AppEngine : object can't be null");

        Object       result = getResult(object);
        LyObjectList targets = null;

        if (result == null)
        {
            // We must compute the value.
            // In this case, nobody have ask for the value.
            targets = findTargets(object);

            if (targets == null)
                putResult(object, EMPTY_LIST, true);
            else
                // Put the result in the cache.
                putResult(object, targets, true);

            result = getResult(object);
            if (result == null)
                return null;
        }

        if (result == EMPTY_LIST)
            return null;

        LyFieldInfo targetFieldInfo = _path.getTargetFieldInfo();

        if (((LyObjectList)result).getSize() == 1)
            return ((LyObjectList)result).getObject(0).getValue(targetFieldInfo);
        else
            return targetFieldInfo.compute((LyObjectList)result, true);
    } // end method computeValue

    /**
     * To free a calculated value for the specified object.
     *
     * @param  object  the object we want to free the value.
     */
    public void freeValue(LyObject object)
    {
        removeResult(object);
    }

    /**
     * Get a result of the calculated value of the relation in a cache.
     *
     * @param   object  the associated object of the calulated value.
     *
     * @return  the calculated value.
     */
    Object getResult(LyObject object)
    {
        LyObjectId id = object.getObjectId();

        return _results.get(id);
    }

    /**
     * Add a targets list for an object.
     */
    void putResult(LyObject object, Object targets, boolean unlock)
    {
        if (!object.checkInstance())
            // Never store an object that is not found in cache
            // This may occur when checking filter on peer object.
            return;

        LyObjectId id = object.getObjectId();

        _results.put(id, targets);

        Object value = computeValue(object);

        if (value == null)
            value = EMPTY_VALUE;

        _values.put(id, value);
    }

    /**
     * Remove an object from our objects list.
     */
    void removeResult(LyObject object)
    {
        LyObjectId id = object.getObjectId();

        _values.remove(id);
        _results.remove(id);

        if (_sourcesWorkSpace != null)
            _sourcesWorkSpace.removeObject(object);
    }

    /**
     * Gets references for an object.
     */
    LyObjectList getTargets(LyObject object)
    {
        Object result = getResult(object);

        if (result == null)
            return null;
        else if (result == EMPTY_LIST)
            return null;
        else
            return (LyObjectList)result;
    }

    /**
     * To retrieve all objects by a target.
     */
    LyObjectList findSources(LyObject target, boolean fromResults)
    {
        if ((_targetsWorkSpace == null) && (!fromResults))
            return null;
        else if ((_results == null) && fromResults)
            return null;
        else if (fromResults)
        {
            LySimpleObjectList      result = null;
            Enumeration<LyObjectId> e = _results.keys();
            LyObject                source;
            Object                  r;
            LyObjectId              sourceId;

            while (e.hasMoreElements())
            {
                sourceId = e.nextElement();
                source = sourceId.getLogicalObject();
                if (source == null)
                    continue;

                r = getResult(source);
                if ((r instanceof LyObject)
                        && (r == target))
                {
                    if (result == null)
                        result = new LySimpleObjectList(source);
                    else
                        result.addObject(source);
                }
                else if ((r instanceof LyObjectList)
                        && ((LyObjectList)r).contains(target))
                {
                    if (result == null)
                        result = new LySimpleObjectList(source);
                    else
                        result.addObject(source);
                }
            }

            return result;
        }
        else
            return _targetsWorkSpace.getSources(target);
    } // end method findSources

    /**
     * Method for calculating objects related to the given object.
     */
    LyObjectList findTargets(LyObject source)
    {
        if (_sourcesWorkSpace == null)
        {
            _sourcesWorkSpace = new LySimpleWorkSpace(null, source, true);
            _targetsWorkSpace = (LyWorkSpace)_sourcesWorkSpace.getList(null, _sourceClass, _path, true);
            _targetsWorkSpace.addListener(this);

            if (_targetsWorkSpace.getSize() > 0)
                return new LySimpleObjectList(_targetsWorkSpace);
            else
                return null;
        }

        _sourcesWorkSpace.addObject(source);

        LyObjectList list = new LySimpleObjectList(source);
        LyObjectList targets = list.getList(null, _sourceClass, _path, false);

        if (targets == null)
            return null;

        return new LySimpleObjectList(targets);
    }

    /**
     * Unload daemon.
     */
    public void unload()
    {
        if (_targetsWorkSpace != null)
            _targetsWorkSpace.free(this);

        if (_sourcesWorkSpace != null)
            _sourcesWorkSpace.free(null);

        _results.clear();
        _values.clear();
    }

    /* Implementation of LyDataSourceListener */

    /**
     * Method for object creation callback.
     *
     * @param   event  the datasource event
     *
     * @return  true if this event was handled
     *
     * @see     LyDataSourceListener#objectCreated
     */
    public boolean objectCreated(LyDataSourceEvent event)
    {
        return false;
    }

    /**
     * Method for object deletion callback.
     *
     * @param   event  the datasource event
     *
     * @return  true if this event was handled
     *
     * @see     LyDataSourceListener#objectDeleted
     */
    public boolean objectDeleted(LyDataSourceEvent event)
    {
        LyObject object = (LyObject)event.getObject();

        // If the object is not in the list, ignore this event
        if (!contains(object))
            return false;

        // Remove object from results table
        removeResult(object);
        return true;
    }

    /**
     * Method for object modification callback.
     *
     * @param   event  the datasource event
     *
     * @return  true if this event was handled
     *
     * @see     LyDataSourceListener#objectModified
     */
    public boolean objectModified(LyDataSourceEvent event)
    {
        // Done automatically by workSpaces
        return true;
    }

    /**
     * Checks whether a given object is contained in our objects list.
     */
    boolean contains(LyObject object)
    {
        LyObjectId id = object.getObjectId();

        return _results.containsKey(id);
    }

    /* Implementation of LyWorkSpaceListener */

    /**
     * Called when an object is added in the dynamic object list.
     *
     * @param   event  event of adding object
     *
     * @return  the event is used by the listener
     */
    public boolean objectAdded(LyWorkSpaceEvent event)
    {
        LyObject     target = event.getObject();
        LyObjectList sources;

        sources = findSources(target, false);

        refresh(sources, target, true);
        return true;
    }

    /**
     * Called when an object is removed from the dynamic object list.
     *
     * @param   event  event of removing object
     *
     * @return  the event is used by the listener
     */
    public boolean objectRemoved(LyWorkSpaceEvent event)
    {
        LyObject           target = event.getObject();
        LyObjectList       resultSources = findSources(target, true);
        LyObjectList       wsSources = findSources(target, false);
        LySimpleObjectList sources = null;

        if (resultSources != null)
        {
            if (wsSources == null)
                sources = new LySimpleObjectList(resultSources);
            else
            {
                // Only sources in resultSources that are not in wsSources
                // must be updated.
                int      i, n = resultSources.getSize();
                LyObject source;

                for (i = 0; i < n; i++)
                {
                    source = resultSources.getObject(i);
                    if (!wsSources.contains(source))
                    {
                        if (sources == null)
                            sources = new LySimpleObjectList();

                        sources.addObject(source);
                    }
                }
            }
        }

        refresh(sources, target, false);
        return true;
    } // end method objectRemoved

    /**
     * Called when the counter associated with the specified object has been incremented.
     *
     * @param   event  the workspace event.
     *
     * @return  True if the event is used by the listener, false otherwise.
     */
    public boolean counterChanged(LyWorkSpaceEvent event)
    {
        if (event.isIncremented())
            return objectAdded(event);
        else
            return objectRemoved(event);
    }

    /**
     * Called when an object has changed in the dynamic object list.
     *
     * @param   event  event of changing object
     *
     * @return  the event is used by the listener
     */
    public boolean objectChanged(LyWorkSpaceEvent event)
    {
        LyFieldInfo  targetFieldInfo = _path.getTargetFieldInfo();
        LyObject     target = event.getObject();
        LyObjectList sources = findSources(target, false);
        int          i, n = ((sources == null) ? 0 : sources.getSize());
        LyObject     object;
        Object       oldValue, newValue;
        LyValueSet   values = event.getValues();
        LyObjectId   id;

        for (i = 0; i < n; i++)
        {
            object = sources.getObject(i);
            id = object.getObjectId();

            oldValue = _values.get(id);
            newValue = computeValue(object);

            if (oldValue == EMPTY_VALUE)
                oldValue = null;

            if (_fieldInfo.equals(oldValue, newValue))
                continue;
            else if (newValue == null)
                _values.put(id, EMPTY_VALUE);
            else
                _values.put(id, newValue);

            if ((values == null)
                    || (values.getSize() == 0)
                    || values.contains(targetFieldInfo))
                refresh(object);
        }

        return true;
    } // end method objectChanged

    /**
     * Add an object to our objects list with given targets.
     */
    void addObject(LyObject object, LyObject target)
    {
        Object             r = getResult(object);
        LySimpleObjectList list;

        if ((r == null)
                || (r == EMPTY_LIST))
            list = new LySimpleObjectList();
        else if (r instanceof LyObject)
            list = new LySimpleObjectList((LyObject)r);
        else
            list = (LySimpleObjectList)r;

        list.addObject(target);
        putResult(object, list, false);
    }

    /**
     * Add an object to our objects list with given targets.
     */
    void removeObject(LyObject object, LyObject target)
    {
        Object             r = getResult(object);
        LySimpleObjectList list;

        if ((r != null)
                && (r != EMPTY_LIST))
        {
            list = (LySimpleObjectList)r;
            if (list.contains(target))
                list.removeObject(target);

            if (list.getSize() == 0)
                r = EMPTY_LIST;
            else
                r = list;
        }

        if (r != null)
            putResult(object, r, false);
    }

    void refresh(LyObjectList objects, LyObject target, boolean add)
    {
        int      i, n = ((objects == null) ? 0 : objects.getSize());
        LyObject object;

        for (i = 0; i < n; i++)
        {
            object = objects.getObject(i);
            if (add)
                addObject(object, target);
            else
                removeObject(object, target);

            Object oldValue = _values.get(object.getObjectId());
            Object newValue = computeValue(object);

            if (oldValue == EMPTY_VALUE)
                oldValue = null;

            if (!_fieldInfo.equals(oldValue, newValue))
            {
                if (newValue == null)
                    newValue = EMPTY_VALUE;

                _values.put(object.getObjectId(), newValue);
                refresh(object);
            }
        }
    }

    /**
     * To set the relation value with new values.
     */
    void refresh(LyObject object)
    {
        LyInfoList<LyFieldInfo> fields = new LyInfoList<LyFieldInfo>(_fieldInfo);

        object.refresh(fields);
    }

    /**
     * Called when objectId of an object has been changed. This method must update tables with
     * references to objectIds.
     *
     * @param  oldObjectId
     * @param  newObjectId
     */
    public void idRehashed(LyObjectId oldObjectId, LyObjectId newObjectId)
    {
        Object value;

        if ((_results != null)
                && _results.containsKey(oldObjectId))
        {
            value = _results.remove(oldObjectId);
            _results.put(newObjectId, value);
        }

        if ((_values != null)
                && _values.containsKey(oldObjectId))
        {
            value = _values.remove(oldObjectId);
            _values.put(newObjectId, value);
        }
    }
} // end class LyPathDaemon
