/*
 * LyRelationDaemon.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data.daemon;

import java.util.*;
import leon.data.*;
import leon.info.*;


/**
 * Daemon for calculated relations.
 */
public abstract class LyRelationDaemon extends LyFieldInfoDaemon
{
	/** File version */
	public static final String __VERSION =
		"$Id: LyRelationDaemon.java 41140 2012-10-11 09:52:22Z bangiole $";

	/** Boolean indicating if we should add new sourceClass objects when they are created. */
	protected boolean		   _addNewObjects;

	/** Results table. */
	protected Hashtable<LyObjectId, Object>		   _results;

	/** Static object for storing an empty list in results hashtable. */
	static protected Object    EMPTY_LIST = "__emptyList__";

	/**
	 * Default constructor.
	 * @param sourceClass the source class
	 * @param relation the relation field
	 * @param addNewObjects true to handle new objects
	 */
	protected LyRelationDaemon(LyClassInfo sourceClass, LyRelationFieldInfo relation,
		boolean addNewObjects)
	{
		super(sourceClass, relation);

		_addNewObjects = addNewObjects;
		_results = new Hashtable<LyObjectId, Object>();
	}

	/**
	 * Get a result of the calculated value of the relation in a cache.
	 *
	 * @param  object the associated object of the calculated value.
	 * @return the calculated value.
	 */
	protected Object getResult(LyObject object)
	{
		LyObjectId id = object.getObjectId();

		return _results.get(id);
	}

	/**
	 * To put a result of the calculated value of the relation in a cache.
	 *
	 * @param object the associated object of the calculated value.
	 * @param result the calculated value.
	 */
	protected void putResult(LyObject object, Object result)
	{
		if (!object.checkInstance())
			// Never store an object that is not found in cache
			// This may occur when checking filter on peer object.
			return;

		if ((result != EMPTY_LIST)
			&& !(result instanceof LyObject)
			&& !(result instanceof LySimpleObjectList))
		{
			if (result == null)
				result = EMPTY_LIST;
			else if (result instanceof LyObjectList)
				result = new LySimpleObjectList((LyObjectList)result);
			else
				throw new IllegalStateException("Leonardi : result " + result
					+ " for object " + object + " is invalid.");
		}

		LyObjectId id = object.getObjectId();

		_results.put(id, result);
	}

	/**
	 * To remove a result of the calculated value of the relation from the cache.
	 *
	 * @param object the associated object of the calculated value.
	 */
	protected void removeResult(LyObject object)
	{
		LyObjectId id = object.getObjectId();

		_results.remove(id);
	}

	/**
	 * Add an object to our objects list with given targets.
	 *
	 * @param object the objects to add
	 * @param targets a list of target
	 	 * @param event optional event indicating when notified for instance, may be null
	 */
	protected void addObject(LyObject object, LyObjectList targets, LyDataSourceEvent event)
	{
		putResult(object, EMPTY_LIST);

		int i, n = ((targets == null) ? 0 : targets.getSize());

		for (i = 0; i < n; i++)
			addRelation(object, targets.getObject(i), true, event);
	}

	/**
	 * Remove an object from our objects list.
	 *
	 * @param object the object to remove
	 * @param event optional event indicating when notified for instance, may be null
	 */
	protected void removeObject(LyObject object, LyDataSourceEvent event)
	{
		LyObjectList list = getTargets(object);
		LyObject     result;
		int			 i, n = ((list == null) ? 0 : list.getSize());

		for (i = n - 1; i >= 0; i--)
		{
			result = list.getObject(i);
			removeRelation(object, result, true, event);
		}

		removeResult(object);
	}

	/**
	 * Checks whether a given object is contained in our objects list.
	 * @param object the object to check
	 * @return true if the object is in the list
	 */
	protected boolean contains(LyObject object)
	{
		LyObjectId id = object.getObjectId();

		return _results.containsKey(id);
	}

	/**
	 * Method to add a relation between a source object and a target object.
	 *
	 * @param object Source object.
	 * @param target Target object.
	 * @param source Boolean indicating if this adding results on a modification on source object or
	 *               not.
	 * @param event  optional event indicating when notified for instance, may be null
	 */
	protected void addRelation(LyObject object, LyObject target, boolean source,
		LyDataSourceEvent event)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null.");
		if (target == null)
			throw new IllegalArgumentException("BF AppEngine : target can't be null.");

		if (!object.instanceOf(_sourceClass))
			return;

		Object result = getResult(object);

		if (result == null)
		{
			if (_addNewObjects)
				throw new IllegalStateException("BF AppEngine : result for object "
					+ object + " shouldn't be null");
			else
				result = EMPTY_LIST;
		}

		LySimpleObjectList targets;

		if (result == EMPTY_LIST)
		{
			putResult(object, target);
			targets = new LySimpleObjectList(target);
		}
		else if (result instanceof LyObject)
		{
			if (result == target)
				return;

			targets = new LySimpleObjectList((LyObject)result);
			targets.addObject(target);
			putResult(object, targets);
		}
		else
		{
            if (result instanceof LySimpleObjectList)
                targets = (LySimpleObjectList)result;
            else
                targets = new LySimpleObjectList((LyObjectList)result);

			if (targets.contains(target))
				return;

			targets.addObject(target);
			putResult(object, targets);
		}

		// Notify change on object
		Object value = computeValue(targets);

		setValue(object, _fieldInfo, value, event);
	} // end method addRelation

	/**
	 * Method to remove a relation between a source object and a target object.
	 *
	 * @param object Source object.
	 * @param target Target object.
	 * @param source Boolean indicating if the removal results on a modification on source object or
	 *               not.
	 * @param event  optional event indicating when notified for instance, may be null
	 */
	protected void removeRelation(LyObject object, LyObject target, boolean source,
		LyDataSourceEvent event)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null.");
		if (target == null)
			throw new IllegalArgumentException("BF AppEngine : target can't be null.");

		if (!object.instanceOf(_sourceClass))
			return;

        if (event.isLoading())
        {
            // Unload event : remove result and local value on object
            removeResult(object);
            object.removeLocalValue(_fieldInfo);
            return;
        }

		Object result = getResult(object);

		if ((result == null) && _addNewObjects)
			throw new IllegalStateException("BF AppEngine : result for object "
				+ object + " shouldn't be null");
		else if (result == null)
			return;

		if (result == EMPTY_LIST)
			return;

		LySimpleObjectList targets;

		if (result instanceof LyObject)
		{
			if (result != target)
				return;

			putResult(object, EMPTY_LIST);
			targets = null;
		}
		else
		{
            if (result instanceof LySimpleObjectList)
                targets = (LySimpleObjectList)result;
            else
                targets = new LySimpleObjectList((LyObjectList)result);

			if (!targets.contains(target))
				return;

			targets.removeObject(target);

			int n = targets.getSize();

			if (n == 0)
				putResult(object, EMPTY_LIST);
			else if (n == 1)
				putResult(object, targets.getObject(0));
		}

		// Notify change on object
		Object value = computeValue(targets);

		setValue(object, _fieldInfo, value, event);
	} // end method removeRelation

	/**
	 * Gets references for an object.
	 * @param object the source object
	 * @return a list of target objects
	 */
	protected LyObjectList getTargets(LyObject object)
	{
		Object result = getResult(object);

		if ((result == null) && _addNewObjects)
			throw new IllegalStateException("BF AppEngine : object " + object + " not found");
		else if (result == null)
			return null;
		else if (result == EMPTY_LIST)
			return null;
		else if (result instanceof LyObject)
			return new LySimpleObjectList((LyObject)result);
		else
			return (LyObjectList)result;
	}

	/**
	 * To get the relation value for the specified object.
	 *
	 * @param  object the object we want the value.
	 * @return the value of the relation field for the specified object.
	 */
	public Object getValue(LyObject object)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		Object		 result = getResult(object);
		LyObjectList targets = null;

		if ((result == null) && _addNewObjects)
			throw new IllegalStateException("BF AppEngine : object " + object + " not found");
		else if (result == null)
		{
			// We must compute the value.
			// In this case, nobody have ask for the value.
			targets = findTargets(object);

			// Put the result in the cache.
			if ((targets == null) || (targets.getSize() == 0))
				putResult(object, EMPTY_LIST);
			else if (targets.getSize() == 1)
				putResult(object, targets.getObject(0));
			else
				putResult(object, targets);
		}
		else if (result == EMPTY_LIST)
			targets = null;
		else if (result instanceof LyObject)
			targets = new LySimpleObjectList((LyObject)result);
		else
			targets = (LyObjectList)result;

		// Compute the value.
		return computeValue(targets);
	} // end method getValue

	/**
	 * Utility method to compute a value for the relation from an objects list.
	 * @param objects the list of object to compute
	 * @return the result
	 */
	protected Object computeValue(LyObjectList objects)
	{
		int n = ((objects == null) ? 0 : objects.getSize());

		if (n == 0)
			return null;

		LyObject o;

		if (_fieldInfo.isMultiple())
		{
			LyObjectId[] value = ((n == 0) ? null : new LyObjectId[n]);
			int			 i;

			for (i = 0; i < n; i++)
			{
				o = objects.getObject(i);
				value[i] = o.getObjectId();
			}

			return value;
		}
		else
		{
			if (n > 1)
				throw new IllegalStateException(
					"BF AppEngine : find more than 1 object for a non multiple routed relation ("
					+ _fieldInfo + ").");

			o = objects.getObject(0);
			return o.getObjectId();
		}
	} // end method computeValue

	/**
	 * To free a calculated value for the specified object.
	 *
	 * @param object the object we want to free the value.
	 */
	public void freeValue(LyObject object)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		// Remove the value from the results.
		removeResult(object);
	}

	/**
	 * Abstract method for calculating objects related to the given object.
	 * @param target the target object
	 * @param fromResults
	 * @return a list of source objects
	 */
	abstract protected LyObjectList findSources(LyObject target, boolean fromResults);

	/**
	 * Abstract method for calculating objects related to the given object.
	 * @param source a source object
	 * @return a list of all target objects
	 */
	abstract protected LyObjectList findTargets(LyObject source);

	/**
	 * Gets references for an object.
	 *
	 * @param object the object
	 * @param event optional event indicating when notified for instance, may be null
	 * @return a list of references
	 */
	public LyObjectList getObjectList(LyObject object, LyDataSourceEvent event)
	{
		if (!object.instanceOf(_sourceClass))
			throw new IllegalArgumentException("BF AppEngine : invalid classInfo for object " + object);

		if (contains(object))
			return getTargets(object);

		// Add the given object to our list
		LyObjectList list = findTargets(object);

		addObject(object, list, event);
		return list;
	}

	/**
	 * Unload daemon.
	 */
	public void unload()
	{
		_results.clear();
	}

	/* Implementation of LyDataSourceListener */

	/**
	 * Method for object creation callback.
	 *
	 * @param event the datasource event
	 * @return true if this event was handled
	 * @see LyDataSourceListener#objectCreated
	 */
	public boolean objectCreated(LyDataSourceEvent event)
	{
		LyObject object = (LyObject)event.getObject();

		if (!object.instanceOf(_sourceClass))
			return false;

		// FFT01525
		// When a rollBack occurs, we receive a load event.
		// In this case, we must ignore the loading parameter
		// and refresh the daemon
		boolean isRollBack = false;
		LyClass peerClass = object.getClassInfo().getPeerClass();

		if (peerClass != null)
		{
			LyDataSource ds = peerClass.getDataSource();

			isRollBack = ds.isRollBack();
		}
		else
			isRollBack = true; // Force refresh of the daemon

		if (!_addNewObjects
				&& (event.isLoading() || event.isNotified())
				&& !isRollBack) // XX Differentiate a LOAD and a real CREATE
			return false;

		LyObjectList list = ((LyRelationFieldInfo)_fieldInfo).getObjects(object);

		if (contains(object))
			updateObject(object, list, event);
		else
			addObject(object, list, event);
		return true;
	} // end method objectCreated

	/**
	 * Method for object deletion callback.
	 *
	 * @param event the datasource event
	 * @return true if this event was handled
	 * @see LyDataSourceListener#objectDeleted
	 */
	public boolean objectDeleted(LyDataSourceEvent event)
	{
		LyObject object = (LyObject)event.getObject();

		if (!object.instanceOf(_sourceClass))
		{
			LyObjectList list = findSources(object, true);
			int			 i, n = ((list == null) ? 0 : list.getSize());

			for (i = 0; i < n; i++)
				removeRelation(list.getObject(i), object, false, event);

			return true;
		}

		// If the object is not in the list, ignore this event
		if (!contains(object))
			return false;

		// Remove object from results table
		removeObject(object, event);
		object.removeLocalValue(_fieldInfo);
		return true;
	}

	/**
	 * Method for object modification callback.
	 *
	 * @param event the datasource event
	 * @return true if this event was handled
	 * @see LyDataSourceListener#objectModified
	 */
	public boolean objectModified(LyDataSourceEvent event)
	{
		LyObject object = (LyObject)event.getObject();

		if (!object.instanceOf(_sourceClass))
			return false;

		// If the object is not in our list, ignore this event
		if (!_addNewObjects && !contains(object))
			return false;

		LyValueSet values = event.getValues();

		if ((values != null) && (values.getFieldValue(_fieldInfo) == null))
			return false;

		LyObjectList newList = ((LyRelationFieldInfo)_fieldInfo).getObjects(object);

		updateObject(object, newList, event);
		return true;
	}

	/**
	 * Update an object from our objects list.
	 * @param object the object to update
	 * @param newList a list of target objects
	 * @param event the source event
	 */
	protected void updateObject(LyObject object, LyObjectList newList, LyDataSourceEvent event)
	{
		// Find deltas between new value and old value
		LySimpleObjectList   list = new LySimpleObjectList(getTargets(object));
		LyObject	         target;
		int			         i, n;
		LyObjectSource       objectSource;

		n = ((list == null) ? 0 : list.getSize());
		for (i = n - 1; i >= 0; i--)
		{
			target = list.getObject(i);

			objectSource = target.getObjectSource();
			if (!target.isDeleted()
                && (objectSource.getObject(target.getObjectId()) != target))
			{
				getEnvironment().logWarning(
					"BF AppEngine : An invalid object was detected in the daemon, removing it");
				list.removeObject(target);
				continue;
			}

			if ((newList == null) || !newList.contains(target))
				removeRelation(object, target, true, event);
		}

		n = ((newList == null) ? 0 : newList.getSize());
		for (i = 0; i < n; i++)
		{
			target = newList.getObject(i);
			if ((list == null)
			        || !list.contains(target)
			        // In creation, it occurs that the value set by the daemon was erased
			        // when recreating the object (see Eliops/FFT00013).
			        || (event.getType() == LyDataSourceEvent.TYPE_CREATE))
				addRelation(object, target, true, event);
		}
	} // end method updateObject

	/**
	 * Called when objectId of an object has been changed. This method must update tables with
	 * references to objectIds.
	 *
	 * @param oldObjectId
	 * @param newObjectId
	 */
	public void idRehashed(LyObjectId oldObjectId, LyObjectId newObjectId)
	{
		Object value;

		if ((_results != null)
				&& _results.containsKey(oldObjectId))
		{
			value = _results.remove(oldObjectId);
			_results.put(newObjectId, value);
		}
	}
} // end class LyRelationDaemon
