/*
 * LyReverseRelationDaemon.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data.daemon;

import java.util.*;

import leon.data.*;
import leon.info.*;


/**
 * Dynamic list of LyObject.
 */
public class LyReverseRelationDaemon extends LyRelationDaemon
{
	public static final int       MAX_SIZE_FOR_RELOAD = 5;

	public static final String    __VERSION =
		"$Id: LyReverseRelationDaemon.java 41618 2012-11-27 13:09:34Z bangiole $";

    /** Reverse relation. */
    protected LyRelationFieldInfo _reverseRelation;

	protected Hashtable<LyObjectId, Set<LyObjectId>> _resultsReverse = new Hashtable<LyObjectId, Set<LyObjectId>>();

    /**
     * Default constructor.
     *
     * @param  sourceClass      the source class
     * @param  relation         the handled relation
     * @param  reverseRelation  the reverse relation
     */
    public LyReverseRelationDaemon(LyClassInfo sourceClass, LyRelationFieldInfo relation,
        LyRelationFieldInfo reverseRelation)
    {
        super(sourceClass, relation, false);

        if (!reverseRelation.relatedWith(sourceClass))
            throw new IllegalStateException("BF AppEngine : invalid relation "
                + reverseRelation + " for this daemon ");

        _reverseRelation = reverseRelation;

        // Listen to data sources for classes pointed by the relation
        LyClassInfo       classInfo;
        LyObjectSource    source;
        int               i, n = relation.getClassInfoCount();
        _LyTargetListener listener = new _LyTargetListener();

        for (i = 0; i < n; i++)
        {
            classInfo = relation.getClassInfo(i);
            source = classInfo.getObjectSource();
            source.addListener(listener);
        }
    }

    /**
     * Accessor to the reverse relation.
     *
     * @return  The reverse relation.
     */
    public LyRelationFieldInfo getReverseRelation()
    {
        return _reverseRelation;
    }

    /**
     * Method to add a relation between a source object and a target object.
     *
     * @param  object  Source object.
     * @param  target  Target object.
     * @param  source  Boolean indicating if this adding results on a modification
     * @param  event   optional event indicating when notified for instance, may be null on source
     *                 object or not.
     */
    protected void addRelation(LyObject object, LyObject target, boolean source, LyDataSourceEvent event)
    {
        if (!source)
        {
            Object result = getResult(object);

            if (result == null)
                result = getValue(object);

            if (result != null)
            {
                LyObjectList targets = getTargets(object);

	            if (!_fieldInfo.isMultiple()
            		&& (targets != null)
            		&& (targets.getSize() == 1))
        		{
	            	// To avoid an exception, try first to reload existing target
	            	LyObject		oldTarget = targets.getObject(0);

	            	LyInfoList<LyFieldInfo> fields = new LyInfoList<LyFieldInfo>();
	            	fields.addInfo(_reverseRelation);
	            	oldTarget.reload(fields);
        		}

                if ((targets == null)
                		|| !targets.contains(target))
                    super.addRelation(object, target, source, event);
            }

            return;
        }

        if (!_reverseRelation.isMultiple())
            setValue(target, _reverseRelation, object, event);
        else
        {
            LySimpleObjectList list = _reverseRelation.getObjects(target);

            if (list == null)
                list = new LySimpleObjectList();

            list.addObject(object);
            setValue(target, _reverseRelation, list, event);
        }

        LyObjectList targets = getTargets(object);

        if ((targets == null) || !targets.contains(target))
            super.addRelation(object, target, source, event);
    }

    /**
     * Method to remove a relation between a source object and a target object.
     *
     * @param  object  Source object.
     * @param  target  Target object.
     * @param  source  Boolean indicating if the removal results on a modification on source object
     *                 or not.
     * @param  event   optional event indicating when notified for instance, may be null
     */
    protected void removeRelation(LyObject object, LyObject target, boolean source, LyDataSourceEvent event)
    {
        if (!source)
        {
            LyObjectList targets = getTargets(object);

            if ((targets != null) && targets.contains(target))
                super.removeRelation(object, target, source, event);

            return;
        }

        if (target.isDeleted())
            return;

        if (event.isLoading())
        {
            // Unload event
            removeResult(object);
            object.removeLocalValue(_fieldInfo);
            return;
        }

        if (!_reverseRelation.isMultiple())
        {
            if (!_reverseRelation.hasMark(LyFieldInfo.MARK_NOT_NULL))
                setValue(target, _reverseRelation, null, event);
        }
        else
        {
            LySimpleObjectList list = _reverseRelation.getObjects(target);

            if (list != null)
            {
                list.removeObject(object);
            }

            if ((list != null)
                    || !_reverseRelation.hasMark(LyFieldInfo.MARK_NOT_NULL))
            	setValue(target, _reverseRelation, list, event);
        }

        LyObjectList targets = getTargets(object);

        if ((targets != null) && targets.contains(target))
            super.removeRelation(object, target, source, event);
    }

    /**
     * Remove an object from our objects list.
     *
     * @param  object  the object to remove
     * @param  event   optional event indicating when notified for instance, may be null
     */
    protected void removeObject(LyObject object, LyDataSourceEvent event)
    {
        // We simply have to remove the object from our list.
        removeResult(object);
    }

    /**
	 * Method for calculating objects related to the given object.
	 */
	protected LyObjectList findSources(LyObject target, boolean fromResults)
	{

		if ((target != null)
				&& fromResults)
		{
			boolean            returnResult = true;
			LySimpleObjectList resultList = null;

			if (_resultsReverse.isEmpty())
			{
				return resultList;
			}

			LyObjectId      targetId = target.getObjectId();
			Set<LyObjectId> results = _resultsReverse.get(targetId);
			Object          resultObj;

			if (results == null)
				return resultList;
			for (LyObjectId resultId : results)
			{
				try
				{
					resultObj = resultId.getClassInfo().getObjectSource().getCache().getObject(resultId);

					if (resultObj instanceof LyObject)
					{
						LyObject resultLyObject = (LyObject) resultObj;
						if (resultList == null)
						{
							resultList = new LySimpleObjectList();
						}
						resultList.addObject(resultLyObject);
					}
				}
				catch (LyNoSuchObjectException exc)
				{
					// Object not in cache => must try to get it from datasource
					if (results.size() < MAX_SIZE_FOR_RELOAD)
					{
						// Try to retrieve it from data source
						LyObject resultLyObject = resultId.getLogicalObject();
						if (resultList == null)
						{
							resultList = new LySimpleObjectList();
						}
						if (resultLyObject != null)
						{
							resultList.addObject(resultLyObject);
						}
					}
					else
					{
						returnResult = false;
						break;
					}
				}
			}
			if (returnResult)
				return resultList;
		}

		return _reverseRelation.getObjects(target);
	}

	/**
	 * Overridden to take into account result cache modification
	 * @see leon.data.daemon.LyRelationDaemon#putResult(leon.data.LyObject, java.lang.Object)
	 */
	protected void putResult(LyObject object, Object result)
	{
		super.putResult(object, result);

		if ((object != null)
				&& (result != LyRelationDaemon.EMPTY_LIST))
		{
			LyObjectId objectId = object.getObjectId();

			if (result instanceof LyObject)
			{
				LyObjectId resultId = ((LyObject) result).getObjectId();
				Set<LyObjectId> reverseSet = _resultsReverse.get(resultId);
				if (reverseSet == null)
				{
					reverseSet = new HashSet<LyObjectId>();
					_resultsReverse.put(resultId, reverseSet);
				}
				reverseSet.add(objectId);
			}
			else if (result instanceof LyObjectList)
			{
				LyObjectList    resultList = (LyObjectList) result;
				LyObject        resultObj;
				LyObjectId      resultId;
				Set<LyObjectId> reverseSet;
				int             i, n = resultList.getSize();

				for (i = 0; i < n; i++)
				{
					resultObj = resultList.getObject(i);
					if (resultObj != null)
					{
						resultId = resultObj.getObjectId();
						reverseSet = _resultsReverse.get(resultId);
						if (reverseSet == null)
						{
							reverseSet = new HashSet<LyObjectId>();
							_resultsReverse.put(resultId, reverseSet);
						}
						reverseSet.add(objectId);
					}
				}
			}
		}
	}

	/**
	 * Remove reverse entry if isolated entry... then call super.
	 * @see leon.data.daemon.LyRelationDaemon#removeResult(leon.data.LyObject)
	 */
	protected void removeResult(LyObject object)
	{
		if ((object != null)
				&& !_resultsReverse.isEmpty())
		{
			LyObjectId id = object.getObjectId();
			Object     result = _results.get(id);

			if (result instanceof LyObject)
			{
				LyObjectId      resultId = ((LyObject) result).getObjectId();
				Set<LyObjectId> reverseSet = _resultsReverse.get(resultId);

				if (reverseSet != null)
				{
					reverseSet.remove(id);
					if (reverseSet.size() == 0)
					{
						_resultsReverse.remove(resultId);
					}
				}
			}
			else if (result instanceof LyObjectList)
			{
				LyObjectList    resultList = (LyObjectList) result;
				int             i, n = resultList.getSize();
				LyObject        resultObj;
				LyObjectId      resultId;
				Set<LyObjectId> reverseSet;

				for (i = 0; i < n; i++)
				{
					resultObj = resultList.getObject(i);
					if (resultObj != null)
					{
						resultId = resultObj.getObjectId();
						reverseSet = _resultsReverse.get(resultId);

						if (reverseSet != null)
						{
							reverseSet.remove(id);
							if (reverseSet.size() == 0)
							{
								_resultsReverse.remove(resultId);
							}
						}
					}
				}
			}
		}
		super.removeResult(object);
	}

    /**
     * Method for calculating objects related to the given object.
     *
     * @param   target  the target object
     *
     * @return  a list of known sources
     */
    protected LyObjectList findKnownSources(LyObject target)
    {
        if (target.isDeleted())
            return null;

        LyRelationValue    value = target.getRelationValue(_reverseRelation);
        int                i, n = value.getValueCount();
        LyObjectId         objectId;
        LyObject           object;
        LySimpleObjectList list = null;

        for (i = 0; i < n; i++)
        {
            objectId = value.getObjectId(i);
            if (objectId.checkObject())
            {
                object = objectId.getLogicalObject();
                if ((object != null)
                        && contains(object))
                {
                    if (list == null)
                        list = new LySimpleObjectList();

                    list.addObject(object);
                }
            }
        }

        return list;
    } // end method findKnownSources

    /**
     * Gets references for an object.
     *
     * @param   target  the target object
     *
     * @return  a list of source object
     */
    protected LyObjectList getSources(LyObject target)
    {
        LySimpleObjectList      result = new LySimpleObjectList();
        Enumeration<LyObjectId> keys = _results.keys();
        LyObject                source;
        Object                  value;
        LyObjectId              sourceId;

        while (keys.hasMoreElements())
        {
            sourceId = keys.nextElement();
            value = _results.get(sourceId);
            source = sourceId.getLogicalObject();
            if (source == null)
                continue;

            if ((value instanceof LyObject) && (value == target))
                result.addObject(source);
            else if ((value instanceof LyObjectList) && (((LyObjectList)value).contains(target)))
                result.addObject(source);
        }

        return result;
    }

    /**
     * Abstract method for calculating objects related to the given object.
     *
     * @param   source  a source object
     *
     * @return  a list of target object
     */
    protected LyObjectList findTargets(LyObject source)
    {
        boolean local = false;

        if ((source.getPeer() == null)
                && !source.getClassInfo().isLocal())
            local = true;

        // Find objects of the target classInfo that references
        // objects of source class through reverse relation
        LyFilter                     filter = new LyFilter(LyFilter.Type.OR);
        LyFilterExpression           filterExp;
        LyFilterExpression.Condition condition;

        condition = (_reverseRelation.isMultiple() ? LyFilterExpression.Condition.CONTAINS
                : LyFilterExpression.Condition.EQUALS);
        filterExp = new LyFilterExpression(_reverseRelation, false, condition, source.getObjectId());
        filter.addElement(filterExp);

        // Get filtered list
        LyObjectList            result = null, list;
        LyClassInfo             classInfo;
        LyInfoList<LyFieldInfo> fields = new LyInfoList<LyFieldInfo>();
        LyRelationFieldInfo     relation = (LyRelationFieldInfo)_fieldInfo;
        int                     i, n = relation.getClassInfoCount();


        fields.addInfo(_reverseRelation);
        for (i = 0; i < n; i++)
        {
            classInfo = relation.getClassInfo(i);
            list = new LyClassWorkSpace(null, classInfo, fields, filter, null, false, true, local);
            if (result == null)
                result = new LySimpleObjectList();

            result.addList(list);
        }

        return result;
    } // end method findTargets

	/**
	 * Call super then unload _resultsReverse.
	 * @see leon.data.daemon.LyRelationDaemon#unload()
	 */
	public void unload()
	{
		super.unload();
		if (_resultsReverse != null)
		{
			_resultsReverse.clear();
		}
	} // end method unload

	/**
	 * Call super and update _resulsReverse.
	 * @see leon.data.daemon.LyRelationDaemon#idRehashed(leon.data.LyObjectId, leon.data.LyObjectId)
	 */
	public void idRehashed(LyObjectId oldObjectId, LyObjectId newObjectId)
	{
		super.idRehashed(oldObjectId, newObjectId);

		if (_resultsReverse != null)
		{
			Set<LyObjectId> reverseSet = _resultsReverse.get(oldObjectId);
			if (reverseSet != null)
			{
				_resultsReverse.remove(oldObjectId);
				_resultsReverse.put(newObjectId, reverseSet);
			}

			for (Set<LyObjectId> ids : _resultsReverse.values())
			{
				if (ids.contains(oldObjectId))
				{
					ids.remove(oldObjectId);
					ids.add(newObjectId);
				}
			}
		}
	} // end method idRehashed

    /* Implementation of LyDataSourceListener */

    /**
     * Method for object creation callback.
     *
     * @param   event  the datasource event
     *
     * @return  true if this event was handled
     *
     * @see     LyDataSourceListener#objectCreated
     */
    public boolean objectCreated(LyDataSourceEvent event)
    {
        if (event.isNotified())
            return true;

        return super.objectCreated(event);
    } // end method objectCreated

    /**
     * Method for object modification callback.
     *
     * @param   event  the datasource event
     *
     * @return  true if this event was handled
     *
     * @see     LyDataSourceListener#objectModified
     */
    public boolean objectModified(LyDataSourceEvent event)
    {
        LyObject object = (LyObject)event.getObject();

        // If the object is not in our list, ignore this event
        if (!contains(object))
            return false;

        if (event.isLocal() && !_reverseRelation.isMultiple())
        {
            // If some ID fields have been changed we must try
            // to compute relations.
            LyValueSet values = event.getValues();

            if (values != null)
            {
                LyFieldInfo fieldInfo;
                LyValue     value;
                int         i, n = values.getSize();

                for (i = 0; i < n; i++)
                {
                    value = values.getFieldValue(i);
                    fieldInfo = (LyFieldInfo)value.getField();
                    if (fieldInfo.hasMark(LyFieldInfo.MARK_ID))
                        break;
                }

                if (i < n)
                {
                    // The id has changed : compute the relations.
                    LyObjectList list = ((LyRelationFieldInfo)_fieldInfo).getObjects(object);
                    LyObject     target;
                    LyObjectId   newSourceId, sourceId;

                    newSourceId = object.getObjectId();
                    n = ((list == null) ? 0 : list.getSize());
                    for (i = 0; i < n; i++)
                    {
                        target = list.getObject(i);
                        sourceId = (LyObjectId)target.getValue(_reverseRelation);
                        if ((sourceId != null)
                                && (newSourceId != null)
                                && !newSourceId.equals(sourceId))
                            setValue(target, _reverseRelation, newSourceId, event);
                    }
                }
            }
        }

        return super.objectModified(event);
    } // end method objectModified

    class _LyTargetListener
        implements LyDataSourceListener
    {

        /**
         * Method for object creation callback.
         *
         * @param   event  the datasource event
         *
         * @return  true if this event was handled
         *
         * @see     LyDataSourceListener#objectCreated
         */
        public boolean objectCreated(LyDataSourceEvent event)
        {
            if (_results.size() == 0)
                return false;

            LyObject     object = (LyObject)event.getObject();
            LyObjectList list = findKnownSources(object);
            int          i, n = ((list == null) ? 0 : list.getSize());
            LyObject     source;

            for (i = 0; i < n; i++)
            {
                source = list.getObject(i);
                addRelation(source, object, false, event);
            }

            return true;
        }

        /**
         * Method for object modification callback.
         *
         * @param   event  the datasource event
         *
         * @return  true if this event was handled
         *
         * @see     LyDataSourceListener#objectModified
         */
        public boolean objectModified(LyDataSourceEvent event)
        {
            // Find deltas between new value and old value
            LyValueSet values = event.getValues();

            if ((values != null) && !(values.contains(_reverseRelation)))
                return false;

            LyObject     object = (LyObject)event.getObject();
            LyObjectList newList = findSources(object, false);
            LyObjectList list = getSources(object);
            LyObject     source;
            int          i, n;

            n = ((list == null) ? 0 : list.getSize());
            for (i = 0; i < n; i++)
            {
                source = list.getObject(i);
                if ((newList == null) || !newList.contains(source))
                    removeRelation(source, object, false, event);
            }

            n = ((newList == null) ? 0 : newList.getSize());
            for (i = 0; i < n; i++)
            {
                source = newList.getObject(i);
                if ((list == null) || !list.contains(source))
                    addRelation(source, object, false, event);
            }

            return true;
        } // end method objectModified

        /**
         * Method for object deletion callback.
         *
         * @param   event  the datasource event
         *
         * @return  true if this event was handled
         *
         * @see     LyDataSourceListener#objectCreated
         */
        public boolean objectDeleted(LyDataSourceEvent event)
        {
            LyObject object = (LyObject)event.getObject();
            LyObjectList list = findSources(object, true);
            int          i, n = ((list == null) ? 0 : list.getSize());

            for (i = 0; i < n; i++)
                removeRelation(list.getObject(i), object, false, event);

            return true;
        }
    } // end class _LyTargetListener
} // end class LyReverseRelationDaemon
