/*
 * LyRoutedRelationDaemon.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.data.daemon;

import java.util.*;

import leon.data.*;
import leon.info.*;


/**
 * Daemon for calculated relations.
 */
public class LyRoutedRelationDaemon extends LyRelationDaemon
    implements LyWorkSpaceListener,
        LyWorkSpaceCounterListener
{
    /** File version. */
    public static final String __VERSION = "$Id: LyRoutedRelationDaemon.java 28980 2010-12-22 10:53:09Z jlsaouli $";

    /** The source class. */
    protected LyClassInfo _targetClass;

    /** Global workspace with all sources. */
    protected LySimpleWorkSpace _sourcesWorkSpace;

    /** Global workspace with all targets. */
    protected LyWorkSpace _targetsWorkSpace;

    /**
     * Default constructor.
     *
     * @param  sourceClass    the source class
     * @param  relation       the handled relation
     * @param  addNewObjects
     */
    public LyRoutedRelationDaemon(LyClassInfo sourceClass, LyRelationFieldInfo relation, boolean addNewObjects)
    {
        super(sourceClass, relation, addNewObjects);

        _targetClass = relation.getClassInfo(0);
    }

    /**
     * Remove an object from our objects list.
     *
     * @param  object  the object to remove
     */
    protected void removeResult(LyObject object)
    {
        super.removeResult(object);

        if ((_sourcesWorkSpace != null)
                && _sourcesWorkSpace.contains(object))
            _sourcesWorkSpace.removeObject(object);
    }

    /**
     * Add an object to our objects list with given targets.
     *
     * @param  object   the source object
     * @param  targets  a list of associated target objects
     * @param  event    optional event indicating when notified for instance, may be null
     */
    protected void setObject(LyObject object, LyObjectList targets, LyDataSourceEvent event)
    {
        if (targets == null)
            return;

        putResult(object, targets);

        // Notify change on object
        if (!object.isDeleted())
        {
            Object value = computeValue(targets);

            setValue(object, _fieldInfo, value, event);
        }
    }

    /**
     * To set the relation value with new values.
     *
     * @param  sources  a list of sources object
     * @param  target   a target of all source object
     * @param  add      indicates if the object must be added or removed
     */
    protected void notifyChanges(LyObjectList sources, LyObject target, boolean add)
    {
        int                i, n = ((sources == null) ? 0 : sources.getSize());
        LyObject           object;
        LyObjectList       targets;
        LySimpleObjectList list;

        for (i = 0; i < n; i++)
        {
            object = sources.getObject(i);

            // If there is no result for this object, it is not worth
            // to notify changes because nobody has asked for the value.
            if (getResult(object) == null)
                continue;

            targets = getTargets(object);
            if (targets == null)
            {
                if (add)
                    targets = new LySimpleObjectList(target);
            }
            else
            {
                if (add)
                {
                    if (!(targets.contains(target)))
                    {
                        list = new LySimpleObjectList(target);
                        list.addList(targets);
                        targets = list;
                    }
                    else
                        targets = null;
                }
                else
                {
                    if (targets.contains(target))
                    {
                        list = new LySimpleObjectList(targets);
                        list.removeObject(target);
                        targets = list;
                    }
                    else
                        targets = null;
                }
            }

            if (targets != null)
                setObject(object, targets, null);
        }
    } // end method notifyChanges

    /**
     * Method to add a relation beetween a source object and a target object.
     *
     * @param  object  Source object.
     * @param  target  Target object.
     * @param  source  Boolean indicating if this adding results on a modification
     * @param  event   optional event indicating when notified for instance, may be null on source
     *                 object or not.
     */
    protected void addRelation(LyObject object, LyObject target, boolean source, LyDataSourceEvent event)
    {
        // Nothing to do : we don't treat the value relation by relation.
    }

    /**
     * Method to remove a relation beetween a source object and a target object.
     *
     * @param  object  Source object.
     * @param  target  Target object.
     * @param  source  Boolean indicating if the removal results on a modification on source object
     *                 or not.
     * @param  event   optional event indicating when notified for instance, may be null
     */
    protected void removeRelation(LyObject object, LyObject target, boolean source, LyDataSourceEvent event)
    {
        // Nothing to do : we don't treat the value relation by relation.
    }

    /**
     * Gets references for an object.
     *
     * @param   object  a source object
     *
     * @return  a list of target objects
     */
    protected LyObjectList getTargets(LyObject object)
    {
        Object result = getResult(object);

        if ((result == null) && _addNewObjects)
            throw new IllegalStateException("BF AppEngine : object " + object + " not found");
        else if (result == null)
            return null;
        else if (result == EMPTY_LIST)
            return null;
        else if (result instanceof LyObject)
            throw new IllegalStateException("BF AppEngine : target for object " + object
                + " must be an object list");
        else
            return (LyObjectList)result;
    }

    /**
     * To get the relation value for the specified object.
     *
     * @param   object  a source object
     *
     * @return  the value of the relation for this object
     */
    public Object getValue(LyObject object)
    {
        if (object == null)
            throw new IllegalArgumentException("BF AppEngine : object can't be null");

        Object       result = getResult(object);
        LyObjectList targets = null;

        if ((result == null) && _addNewObjects)
            throw new IllegalStateException("BF AppEngine : object " + object + " not found");
        else if (result == null)
        {
            // We must compute the value.
            // In this case, nobody have ask for the value.
            targets = findTargets(object);

            // Put the result in the cache.
            if ((targets == null) || (targets.getSize() == 0))
                putResult(object, EMPTY_LIST);
            else
                putResult(object, targets);
        }
        else if (result == EMPTY_LIST)
            targets = null;
        else if (result instanceof LyObject)
            targets = new LySimpleObjectList((LyObject)result);
        else
            targets = (LyObjectList)result;

        // Compute the value.
        Object value = computeValue(targets);

        // Return the value.
        return value;
    } // end method getValue

    /**
     * To retrive an object by the list of the objects in the value.
     *
     * @param   target       a target object
     * @param   fromResults
     *
     * @return  a list of sources objects
     */
    protected LyObjectList findSources(LyObject target, boolean fromResults)
    {
        if ((_targetsWorkSpace == null) && (!fromResults))
            return null;
        else if ((_results == null) && (fromResults))
            return null;
        else if (fromResults)
        {
            LySimpleObjectList      result = null;
            Enumeration<LyObjectId> e = _results.keys();
            LyObject                source;
            Object                  r;
            LyObjectId              sourceId;

            while (e.hasMoreElements())
            {
                sourceId = e.nextElement();
                source = sourceId.getLogicalObject();
                if (source == null)
                    continue;

                r = getResult(source);
                if ((r instanceof LyObject)
                        && (r == target))
                {
                    if (result == null)
                        result = new LySimpleObjectList(source);
                    else
                        result.addObject(source);
                }
                else if ((r instanceof LyObjectList)
                        && ((LyObjectList)r).contains(target))
                {
                    if (result == null)
                        result = new LySimpleObjectList(source);
                    else
                        result.addObject(source);
                }
            }

            return result;
        }
        else
            return _targetsWorkSpace.getSources(target);
    } // end method findSources

    /**
     * Method for calculating objects related to the given object.
     *
     * @param   source  a source object
     *
     * @return  a list of target objects
     */
    protected LyObjectList findTargets(LyObject source)
    {
        LyInfoList<LyRoute> routes = ((LyRelationFieldInfo)_fieldInfo).getRoutes();

        if (_sourcesWorkSpace == null)
        {
            _sourcesWorkSpace = new LySimpleWorkSpace(null, source, true);
            _targetsWorkSpace = (LyWorkSpace)_sourcesWorkSpace.getList(null, _sourceClass, routes, true);
            _targetsWorkSpace.addListener(this);

            if (_targetsWorkSpace.getSize() > 0)
                return new LySimpleObjectList(_targetsWorkSpace);
            else
                return null;
        }

        _sourcesWorkSpace.addObject(source);

        LyObjectList list = new LySimpleObjectList(source);
        LyObjectList targets = list.getList(null, _sourceClass, routes, false);

        if (targets == null)
            return null;

        return new LySimpleObjectList(targets);
    }

    /**
     * Gets references for an object.
     *
     * @param   object  an object
     * @param   event   optional event indicating when notified for instance, may be null
     *
     * @return  a list of references
     */
    public LyObjectList getObjectList(LyObject object, LyDataSourceEvent event)
    {
        if (!object.instanceOf(_sourceClass))
            throw new IllegalArgumentException("BF AppEngine : invalid classInfo for object " + object);

        if (contains(object))
            return getTargets(object);

        // Add the given object to our list
        LyObjectList list = findTargets(object);

        addObject(object, list, event);
        return list;
    }

    /**
     * Unload daemon.
     */
    public void unload()
    {
        if (_targetsWorkSpace != null)
            _targetsWorkSpace.free(this);

        if (_sourcesWorkSpace != null)
            _sourcesWorkSpace.free(null);

        super.unload();
    }

    /* Implementation of LyDataSourceListener */

    /**
     * Method for object creation callback.
     *
     * @param   event  the datasource event
     *
     * @return  true if this event was handled
     *
     * @see     LyDataSourceListener#objectCreated
     */
    public boolean objectCreated(LyDataSourceEvent event)
    {
        LyObject object = (LyObject)event.getObject();

        if (!object.instanceOf(_sourceClass))
            return false;

        if (!_addNewObjects)
            return false;

        if (contains(object))
            return false;

        LyObjectList list = findTargets(object);

        addObject(object, list, event);
        return true;
    }

    /**
     * Method for object deletion callback.
     *
     * @param   event  the datasource event
     *
     * @return  true if this event was handled
     *
     * @see     LyDataSourceListener#objectDeleted
     */
    public boolean objectDeleted(LyDataSourceEvent event)
    {
        return super.objectDeleted(event);
    }

    /**
     * Method for object modification callback.
     *
     * @param   event  the datasource event
     *
     * @return  true if this event was handled
     *
     * @see     LyDataSourceListener#objectModified
     */
    public boolean objectModified(LyDataSourceEvent event)
    {
        // Done automatically by workSpaces
        return true;
    }

    /* Implementation of LyWorkSpaceListener */
    /**
     * Called when an object is added in the dynamic object list.
     *
     * @param   event  event of adding object
     *
     * @return  the event is used by the listener
     */
    public boolean objectAdded(LyWorkSpaceEvent event)
    {
        LyObject     target = event.getObject();
        LyObjectList sources;

        sources = findSources(target, false);

        notifyChanges(sources, target, true);
        return true;
    }

    /**
     * Called when an object is removed from the dynamic object list.
     *
     * @param   event  event of removing object
     *
     * @return  the event is used by the listener
     */
    public boolean objectRemoved(LyWorkSpaceEvent event)
    {
        LyObject           target = event.getObject();
        LyObjectList       resultSources = findSources(target, true);
        LyObjectList       wsSources = findSources(target, false);
        LySimpleObjectList sources = null;

        if (resultSources != null)
        {
            if (wsSources == null)
                sources = new LySimpleObjectList(resultSources);
            else
            {
                // Only sources in resultSources that are not in wsSources
                // must be updated.
                int      i, n = resultSources.getSize();
                LyObject source;

                for (i = 0; i < n; i++)
                {
                    source = resultSources.getObject(i);
                    if (!wsSources.contains(source))
                    {
                        if (sources == null)
                            sources = new LySimpleObjectList();

                        sources.addObject(source);
                    }
                }
            }
        }

        notifyChanges(sources, target, false);
        return true;
    } // end method objectRemoved

    /**
     * Called when an object has changed in the dynamic object list.
     *
     * @param   event  event of changing object
     *
     * @return  the event is used by the listener
     */
    public boolean objectChanged(LyWorkSpaceEvent event)
    {
        // Nothing to do
        return false;
    }

    /**
     * Called when the counter associated with the specified object has been incremented.
     *
     * @param   event  the workspace event.
     *
     * @return  True if the event is used by the listener, false otherwise.
     */
    public boolean counterChanged(LyWorkSpaceEvent event)
    {
        if (event.isIncremented())
            return objectAdded(event);
        else
            return objectRemoved(event);
    }
} // end class LyRoutedRelationDaemon
