/*
 * LyFieldInfo.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.info;

import leon.app.*;
import leon.data.*;
import leon.info.binding.*;
import leon.info.daemon.*;
import leon.misc.*;

import java.lang.reflect.*;
import java.util.*;


/**
 * This class represents an array for a given field info.
 */
public class LyArrayFieldInfo extends LyFieldInfo
{
	public static final String __version = "$Id: LyArrayFieldInfo.java 35260 2012-04-18 08:00:53Z jlsaouli $";

	/** Field info managed in this array. */
	protected LyFieldInfo _fieldInfo;

	/** Min cardinality. */
	protected short _minCard;

	/** Max cardinality. */
	protected short _maxCard;

	/** Java class used for values. */
	protected Class<? extends Object> _valueClass;

	/** Informations about the daemon to manage this array. */
	protected LyArrayDaemon _daemon;

	/**
	 * Number of rows that should be shown when displaying the array (-1 if no preference is given).
	 */
	int _nRows;

	/**
	 * Init method.
	 */
	protected void init()
	{
		super.init();

		_fieldInfo = null;
		_minCard = -1;
		_maxCard = -1;
		_daemon = null;
		_valueClass = null;
		_nRows = -1;
	}

	/**
	 * Constructor of an array field info.
	 *
	 * @param  id        The identifier of the field.
	 * @param  javaName  Name of the generated java field.
	 * @param  template  Field info element of the array.
	 * @param  minCard   Min. cardinality of the array.
	 * @param  maxCard   Max. cardinality of the array.
	 */
	protected LyArrayFieldInfo(String id, String javaName, LyArrayFieldInfo template, short minCard, short maxCard)
	{
		this(id);
		construct(javaName, template, minCard, maxCard);
	}

	/**
	 * Constructor of an array field info.
	 *
	 * @param  id  The identifier of the field.
	 */
	protected LyArrayFieldInfo(String id)
	{
		super(id);
	}

	/**
	 * Construction of an array field info.
	 *
	 * @param  javaName  Name of the generated java field.
	 * @param  template  Field info element of the array.
	 * @param  minCard   Min. cardinality of the array.
	 * @param  maxCard   Max. cardinality of the array.
	 */
	protected void construct(String javaName, LyArrayFieldInfo template, short minCard, short maxCard)
	{
		super.construct(javaName, template);

		if ((minCard < 0) && (template != null))
			_minCard = template._minCard;
		else
			_minCard = minCard;

		if ((maxCard == 0) && (template != null))
			_maxCard = template._maxCard;
		else
			_maxCard = maxCard;
	}

	/**
	 * Init this instance attributes from its template.
	 *
	 * @param  parts  Indicates which parts of the template are copied.
	 */
	void initFromTemplate(int parts)
	{
        parts &= ~_initParts;
        if (parts == 0)
            return;

		// Field contained in array is not specified in binding model,
		// we have to call initFromTemplate by hand
		if (parts == PART_BIND)
			_fieldInfo.initFromTemplate(PART_BIND);

		if (_template == null)
			return;

		super.initFromTemplate(parts);

		LyArrayFieldInfo template = (LyArrayFieldInfo)_template;

		if ((parts & PART_BIND) > 0)
			_daemon = template._daemon;

		if ((parts & PART_DATA) > 0)
			_fieldInfo = template._fieldInfo;

		if ((parts & PART_GUI) > 0)
			_nRows = template._nRows;
	}

	/**
	 * Final step of the parsing : check consistency.
	 *
	 * @param  parent  Object info in which this field is checked.
	 * @param  parts   Indicates which parts are checked.
	 */
	void check(LyObjectInfo parent, int parts)
	{
		super.check(parent, parts);

		if ((parts & PART_GUI) > 0)
		{
			if ((_minCard >= 1)
					&& hasMark(MARK_OPTIONAL))
				getEnvironment().logWarning("BF AppEngine : there is the mark OPTIONAL for fieldInfo "
					+ this + " and min cardinality is " + _minCard);
			else if ((_minCard == 0)
					&& !hasMark(MARK_OPTIONAL))
			{
				getEnvironment().logWarning("BF AppEngine : add mark OPTIONAL for fieldInfo "
					+ this + " which min cardinality is 0");
				if (_marks == null)
					_marks = new BitSet();

				_marks.set(MARK_OPTIONAL);
			}
			else if (_minCard < 0)
				_minCard = (short)(hasMark(MARK_OPTIONAL) ? 0 : 1);

			if (hasMark(MARK_CREATE)
					&& !_fieldInfo.hasMark(MARK_CREATE))
			{
				BitSet marks = _fieldInfo.getMarks();

				if (marks == null)
					marks = new BitSet();

				marks.set(MARK_CREATE);
				_fieldInfo.setMarks(marks);
				getEnvironment().logWarning("BF AppEngine : add mark CREATE for fieldInfo "
					+ _fieldInfo + " because array fieldInfo " + this + " has this mark");
			}

			if (hasMark(MARK_SET)
					&& !_fieldInfo.hasMark(MARK_SET))
			{
				BitSet marks = _fieldInfo.getMarks();

				if (marks == null)
					marks = new BitSet();

				marks.set(MARK_SET);
				_fieldInfo.setMarks(marks);
				getEnvironment().logWarning("BF AppEngine : add mark SET for fieldInfo "
					+ _fieldInfo + " because array fieldInfo " + this + " has this mark");
			}

			if (!hasMark(MARK_TRANSLATE)
					&& _fieldInfo.hasMark(MARK_TRANSLATE))
			{
				if (_marks == null)
					_marks = new BitSet();

				_marks.set(MARK_TRANSLATE);
			}
		}

		if ((parts & PART_DATA) > 0)
		{
			if (_fieldInfo instanceof LyArrayFieldInfo)
				throw new IllegalStateException("BF AppEngine : invalid array " + this
					+ ", array of array not supported in this version");
			else if ((_fieldInfo instanceof LyRelationFieldInfo)
					&& !(_fieldInfo instanceof LyStructFieldInfo))
			{
				LyRelationFieldInfo		 relation = (LyRelationFieldInfo)_fieldInfo;
				LyRelationFieldInfo.Type type = relation.getType();

				if (type == LyRelationFieldInfo.Type.COMPOSITION)
					throw new IllegalStateException("BF AppEngine : invalid array " + this
						+ ", only array of associations is supported in this version");

				// Add dependencies on contextField and classField
				if (relation.getContextField() != null)
					relation.getContextField().addDepend(this);

				if (relation.getClassField() != null)
					relation.getClassField().addDepend(this);
			}

			if (_maxCard == 0)
				_maxCard = -1;

			if ((_minCard == 0) && !hasMark(MARK_OPTIONAL))
			{
				getEnvironment().logWarning("BF AppEngine : add mark OPTIONAL for fieldInfo "
					+ this + " which min cardinality is 0");
				if (_marks == null)
					_marks = new BitSet();

				_marks.set(MARK_OPTIONAL);
			}

			if ((_minCard >= 1) && (_maxCard > 0) && (_minCard > _maxCard))
			{
				getEnvironment().logWarning("BF AppEngine : switch min and max cardinality for fieldInfo "
					+ this + " which min cardinality is " + _minCard + " and max cardinality is "
					+ _maxCard);
				_minCard = _maxCard;
				_maxCard = _minCard;
			}
		}
	} // end method check

	/**
	 * Internal method to get an untyped value as a values table.
	 */
	protected Object[] getValues(Object fieldValue)
	{
		if (fieldValue == null)
			return null;

		if (fieldValue.getClass() != getValueClass())
		{
			if (fieldValue.getClass() == _fieldInfo.getValueClass())
			{
				Object[] value = new Object[1];

				value[0] = fieldValue;
				return value;
			}

			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ fieldValue.getClass() + " expected " + getValueClass());
		}

		return (Object[])fieldValue;
	}

	/**
	 * Get the value at given index from value table.
	 *
	 * @param   fieldValue  The value table.
	 * @param   index       Index of the desired value.
	 *
	 * @return  The value if it exists, null otherwise.
	 */
	public Object getValue(Object fieldValue, int index)
	{
		if ((fieldValue == null) || (index < 0))
			return null;

		Object[] values = getValues(fieldValue);

		return ((index >= values.length) ? null : values[index]);
	}

	/**
	 * Get the number of values found in the given value table.
	 *
	 * @param   fieldValue  The value table.
	 *
	 * @return  Number of values.
	 */
	public int getValueCount(Object fieldValue)
	{
		Object[] values = getValues(fieldValue);

		return ((values == null) ? 0 : values.length);
	}

	/**
	 * Returns class of values for this field.
	 *
	 * @return  Effective class of values built by this fieldInfo.
	 *
	 * @see     LyObject#getValue
	 */
	public Class<?> getValueClass()
	{
		if (_valueClass != null)
			return _valueClass;

		if (_fieldInfo == null)
			return Object[].class;

		Class<?> fieldClass = _fieldInfo.getValueClass();

		_valueClass = (Array.newInstance(fieldClass, 0)).getClass();
		return _valueClass;
	}

	/**
	 * Returns an array of the specified length for building valid values.
	 *
	 * @param   length  The length of the desired array.
	 *
	 * @return  An array of the specified length.
	 */
	public Object[] getArray(int length)
	{
		Class<?> fieldClass = ((_fieldInfo == null) ? Object.class : _fieldInfo.getValueClass());

		return (Object[])Array.newInstance(fieldClass, length);
	}

	/**
	 * Builds the appropriate LyValue corresponding to this field from a value provided as an
	 * Object.
	 *
	 * @param   value  Value from which the LyValue is built.
	 *
	 * @return  An instance of a class that inherits from LyValue corresponding the given value.
	 */
	public LyValue buildFieldValue(Object value)
	{
		if (value instanceof LyValue)
			value = ((LyValue)value).getValue();

		if ((value == null)
				|| (value instanceof Object[]))
			return new LyArrayValue(this, (Object[])value);
		else
			return super.buildFieldValue(value);
	}

	/**
	 * This method allows to create a value from a string for this field. The class of the returned
	 * value is the one given by getValueClass method.
	 *
	 * @param      stringValue  The value presented as a string as produced by the toString method.
	 *
	 * @return     The value decoded from the given string.
	 *
	 * @exception  LyFieldInfoFormatException  If the string can't be parsed properly.
	 *
	 * @see        LyFieldInfo#toString
	 * @see        LyFieldInfo#getValueClass
	 */
	public Object parse(String stringValue) throws LyFieldInfoFormatException
	{
		if (stringValue == null)
			return null;

		String[] strings;

		if (stringValue.indexOf(LyApplication.ARRAY_SEP) >= 0)
			strings = LyString.tokenizer(stringValue, LyApplication.ARRAY_SEP);
		else
			strings = LyString.tokenizer(stringValue, LyApplication.VISIBLE_ARRAY_SEP);

		int		 i = 0, n = strings.length;
		Object[] values = null;

		if ((n != 1)
		    || (_fieldInfo.parse(strings[0]) != null))
		{
			if (_fieldInfo instanceof LyNumberFieldInfo)
			{
				switch (((LyNumberFieldInfo)_fieldInfo).getType())
				{
				case INT:
					values = new Integer[n];
					break;

				case SHORT:
					values = new Short[n];
					break;

				case LONG:
					values = new Long[n];
					break;

				case FLOAT:
					values = new Float[n];
					break;

				case DOUBLE:
					values = new Double[n];
					break;

				default:
					return null;
				}
			}
			else if ((_fieldInfo instanceof LyTextFieldInfo)
					 || (_fieldInfo instanceof LyFileFieldInfo))
				values = new String[n];
			else if (_fieldInfo instanceof LyDateFieldInfo)
				values = new Date[n];
			else if (_fieldInfo instanceof LyRelationFieldInfo)
				values = new LyObjectId[n];
			else
				values = new Object[n];

			for (i = 0; i < n; i++)
				values[i] = _fieldInfo.parse(strings[i]);
		}

		return values;
	}

	/**
	 * Returns a "displayable string" (string that can be shown to the end-user) of the specified
	 * value for this field.
	 *
	 * @param   value  The internal value to convert to a displayable string. The expected class of
	 *                 this value is the one returned from the getValueClass method.
	 *
	 * @return  The value presented as a displayable String.
	 *
	 * @see     LyFieldInfo#getValueClass
	 */
	public String toName(Object fieldValue)
	{
		return toName(fieldValue, null);
	}

	/**
	 * Returns a "displayable string" (string that can be shown to the end-user) of the specified
	 * value for this field.
	 *
	 * @param   value   The internal value to convert to a displayable string. The expected class of
	 *                  this value is the one returned from the getValueClass method.
	 * @param   String  separator : The value of separator between field values.
	 *
	 * @return  The value presented as a displayable String.
	 *
	 * @see     LyFieldInfo#getValueClass
	 */
	public String toName(Object fieldValue, String separator)
	{
		if (fieldValue == null)
			return null;

		String name = getFormattedName(fieldValue);

		if (name != null)
			return name;

		Object[]     values = getValues(fieldValue);
		int			 i, n = values.length;
		StringBuffer buffer = new StringBuffer();

		for (i = 0; i < n; i++)
		{
			if (i > 0)
			{
				if (separator != null)
					buffer.append(separator);
				else
					buffer.append(LyApplication.VISIBLE_ARRAY_SEP);
			}

			buffer.append(_fieldInfo.toName(values[i]));
		}

		return buffer.toString();
	}

	/**
	 * Returns the specified value for this field as a short "displayable" string (short string that
	 * can be shown to the end-user). Default implementation consists of returning same result as
	 * toName.
	 *
	 * @param   value  The internal value to convert to a short displayable string. The expected
	 *                 class of this value is the one returned from the getValueClass method.
	 *
	 * @return  The value presented as a short displayable String.
	 *
	 * @see     LyFieldInfo#getValueClass
	 * @see     LyFieldInfo#toName
	 */
	public String toShortName(Object fieldValue)
	{
		if (fieldValue == null)
			return null;

		String name = getFormattedShortName(fieldValue);

		if (name != null)
			return name;

		Object[]     values = getValues(fieldValue);
		int			 i, n = values.length;
		StringBuffer buffer = new StringBuffer();

		for (i = 0; i < n; i++)
		{
			if (i > 0)
				buffer.append(LyApplication.VISIBLE_ARRAY_SEP);

			buffer.append(_fieldInfo.toShortName(values[i]));
		}

		return buffer.toString();
	}

	/**
	 * Converts the specified value for this field to an "internal" string (string that enables
	 * reconstruction through parse, but not shown to the user). This string may not be significant
	 * to the end-user and usually should not be displayed.
	 *
	 * @param   value  The internal value to convert to an internal string. The expected class of
	 *                 this value is the one returned from the getValueClass method.
	 *
	 * @return  the value as an internal String.
	 *
	 * @see     LyFieldInfo#parse
	 * @see     LyFieldInfo#getValueClass
	 */
	public String toString(Object fieldValue)
	{
		if (fieldValue == null)
			return null;

		String string = getFormattedString(fieldValue);

		if (string != null)
			return string;

		Object[]     values = getValues(fieldValue);
		int			 i, n = values.length;
		StringBuffer buffer = new StringBuffer();

		for (i = 0; i < n; i++)
		{
			if (i > 0)
				buffer.append(LyApplication.ARRAY_SEP);

			buffer.append(_fieldInfo.toString(values[i]));
		}

		return buffer.toString();
	}

	/**
	 * Unitary control of the given value and produce a legible message to the end user if the value
	 * is incorrect. This method is called when validating a form containing this field info.
	 * Default implementation calls check value to prevent cases where it was not done on changes.
	 *
	 * @param   environment  Environment giving the user language if a message has to be produced.
	 * @param   value        Value to control. The expected class of this value is the one returned
	 *                       from the getValueClass method.
	 *
	 * @return  Returns an error message if the value is incorrect, null otherwise.
	 *
	 * @see     LyFieldInfo#getValueClass
	 */
	public String controlValue(LyEnvironment environment, Object fieldValue)
	{
		if (fieldValue == null)
			return null;

		Object[] values = getValues(fieldValue);
		int		 i, n = values.length;
		String   message = null;

		if (environment == null)
			environment = _application.getEnvironment();

		if ((_minCard > 0) && (n < _minCard) && !(hasMark(MARK_OPTIONAL) && (n == 0)))
			return environment.getMessage("LY_ER_ARRAY_MIN", getName(), new Integer(_minCard));

		if ((_maxCard > 0) && (n > _maxCard))
			return environment.getMessage("LY_ER_ARRAY_MAX", getName(), new Integer(_maxCard));

		for (i = 0; (i < n) && (message == null); i++)
			message = _fieldInfo.controlValue(environment, values[i]);

		return message;
	}

	/**
	 * Unitary check of the given value and produce a legible message to the end user if the value
	 * is incorrect. This method is called when changing in a form the value corresponding to this
	 * field info.
	 *
	 * @param   environment  Environment giving the user language if a message has to be produced.
	 * @param   value        Value to check. The expected class of this value is the one returned
	 *                       from the getValueClass method.
	 *
	 * @return  Returns an error message if the value is incorrect, null otherwise.
	 *
	 * @see     LyFieldInfo#getValueClass
	 */
	public String checkValue(LyEnvironment environment, Object fieldValue)
	{
		if (fieldValue == null)
			return null;

		Object[] values = getValues(fieldValue);
		int		 i, n = values.length;
		String   message = null;

		for (i = 0; (i < n) && (message == null); i++)
			message = _fieldInfo.checkValue(environment, values[i]);

		return message;
	}

	/**
	 * Compares two values for equality on this field. The excepted class of the values to compare
	 * is the one returned from the getValueClass method. The default implementation consits of
	 * comparing values using compare method and returning true when result is 0. Due to the fact
	 * that this method can be called often, this method is provided to speed up this compare when
	 * it's possible to do better.
	 *
	 * @param   value1     First value.
	 * @param   value2     Second value.
	 * @param   modifiers  Modifiers used for checking (see MOD_XXX constants of
	 *                     LyFilterExpression).
	 *
	 * @return  True if the values are equals, else false.
	 *
	 * @see     LyFieldInfo#compare
	 * @see     LyFieldInfo#getValueClass
	 */
	public boolean equals(Object value1, Object value2, BitSet modifiers)
	{
		if (value1 == value2)
			return true;
		else if ((value1 == null) || (value2 == null))
			return false;

		Object[] values1 = getValues(value1);
		Object[] values2 = getValues(value2);
		int		 n1 = values1.length, n2 = values2.length;

		if (n1 != n2)
			return false;

		int i;

		for (i = 0; i < n1; i++)
		{
			if (!(_fieldInfo.equals(values1[i], values2[i], modifiers)))
				return false;
		}

		return true;
	}

	/**
	 * Compares two values for the specified field. The expected class of the values to compare is
	 * the one returned from the getValueClass method.
	 *
	 * @param   value1     First value.
	 * @param   value2     Second value.
	 * @param   modifiers  Modifiers used for checking (see MOD_XXX constants of
	 *                     LyFilterExpression).
	 *
	 * @return  0 If the two values are identical for the specified fieldInfo, a positive number if
	 *          value1 is "greater" than value2. If value2 is "greater" than value1, a negative
	 *          number is returned.
	 *
	 * @see     LyFieldInfo#getValueClass
	 */
	public int compare(Object value1, Object value2, BitSet modifiers)
	{
		if (value1 == value2)
			return 0;

		if (value1 == null)
			return -1;

		if (value2 == null)
			return 1;

		Object[] values1 = getValues(value1);
		Object[] values2 = getValues(value2);
		int		 i, n1 = values1.length, n2 = values2.length;
		int		 test = 0;

		for (i = 0; (i < n1) && (i < n2); i++)
		{
			test = (_fieldInfo.compare(values1[i], values2[i], modifiers));
			if (test != 0)
				return test;
		}

		if (n1 == n2)
			return 0;
		else if (n1 < n2)
			return -1;
		else
			return 1;
	}

	/**
	 * Checks if the second value is contained in the first one, for this field. Warning ! This kind
	 * of check does not make sense for every type of field and an exception can be thrown in that
	 * case. The expected class of the values is the one returned from the getValueClass method.
	 *
	 * @param      value1     First value.
	 * @param      value2     Second value.
	 * @param      modifiers  Modifiers used for checking (see MOD_XXX constants of
	 *                        LyFilterExpression).
	 *
	 * @return     True if the second value is contained in the first one, else returns false.
	 *
	 * @exception  LyUnsupportedComparisonException  If it does not make sense to check if a value
	 *                                               is contained in an other one from this field
	 *                                               "point of view".
	 */
	public boolean contains(Object value1, Object value2, BitSet modifiers)
	{
		if ((value1 == null) || (value2 == null))
			return false;

		Object[] values1 = getValues(value1);
		Object[] values2 = getValues(value2);
		int		 i, j, n1 = values1.length, n2 = values2.length;
		boolean  contained;

		for (j = 0; j < n2; j++)
		{
			contained = false;
			for (i = 0; (i < n1) && !contained; i++)
			{
				if (_fieldInfo instanceof LyTextFieldInfo)
					contained = (_fieldInfo.contains(values1[i], values2[j], modifiers));
				else
					contained = (_fieldInfo.equals(values1[i], values2[j], modifiers));
			}

			if (!contained)
				return false;
		}

		return true;
	}

	/**
	 * Checks if the second value starts the first one, for this field. Warning ! This kind of check
	 * does not make sense for every type of field and an exception can be thrown in that case. The
	 * expected class of the values is the one returned from the getValueClass method.
	 *
	 * @param      value1     First value.
	 * @param      value2     Second value.
	 * @param      modifiers  Modifiers used for checking (see MOD_XXX constants of
	 *                        LyFilterExpression).
	 *
	 * @return     True if the second value starts the first one, else returns false.
	 *
	 * @exception  LyUnsupportedComparisonException  If it does not make sense to check if a value
	 *                                               starts the other one from this field "point of
	 *                                               view".
	 */
	public boolean startsWith(Object value1, Object value2, BitSet modifiers)
	{
		if ((value1 == null) || (value2 == null))
			return false;

		Object[] values1 = getValues(value1);
		Object[] values2 = getValues(value2);
		int		 i, n1 = values1.length, n2 = values2.length;

		if (n1 < n2)
			return false;

		for (i = 0; i < n2; i++)
		{
			if (!_fieldInfo.equals(values1[i], values2[i], modifiers))
				return false;
		}

		return true;
	}

	/**
	 * Checks if the second value ends the first one, for this field. Warning ! This kind of check
	 * does not make sense for every type of field and an exception can be thrown in that case. The
	 * expected class of the values is the one returned from the getValueClass method.
	 *
	 * @param      value1     First value.
	 * @param      value2     Second value.
	 * @param      modifiers  Modifiers used for checking (see MOD_XXX constants of
	 *                        LyFilterExpression).
	 *
	 * @return     True if the second value ends the first one, else returns false.
	 *
	 * @exception  LyUnsupportedComparisonException  If it does not make sense to check if a value
	 *                                               ends the other one from this field "point of
	 *                                               view".
	 */
	public boolean endsWith(Object value1, Object value2, BitSet modifiers)
	{
		if ((value1 == null) || (value2 == null))
			return false;

		Object[] values1 = getValues(value1);
		Object[] values2 = getValues(value2);
		int		 i, n1 = values1.length, n2 = values2.length;

		if (n1 < n2)
			return false;

		for (i = 0; i < n2; i++)
		{
			if (!_fieldInfo.equals(values1[n1 - i - 1], values2[n2 - i - 1], modifiers))
				return false;
		}

		return true;
	}

	/**
	 * Returns the filter conditions authorized on this field. The values must match the short
	 * constants defined in LyFilterExpression.
	 *
	 * @return  An array of conditions corresponding to the authorized conditions on this field.
	 *
	 * @see     leon.info.LyFilterExpression
	 */
	public LyFilterExpression.Condition[] getFilterConditions()
	{
		LyFilterExpression.Condition[] conditions = new LyFilterExpression.Condition[1];

		conditions[0] = LyFilterExpression.Condition.CONTAINS;

		return conditions;
	}

	/**
	 * Returns the filter modifiers authorized on this field. The values must match the short
	 * constants defined in LyFilterExpression.
	 *
	 * @return  An array of short corresponding to the authorized modifiers on this field.
	 *
	 * @see     leon.info.LyFilterExpression
	 */
	public short[] getFilterModifiers()
	{
		return _fieldInfo.getFilterModifiers();
	}

	/**
	 * Builds a logical value for this fieldInfo from a physical (peer) value. This method is used
	 * when the binding specified for this field is a LyFieldBinding.
	 *
	 * @param      object    The peer value.
	 * @param      instance  Instance containing the requiered value (could be null). This instance
	 *                       is provided to be able to construct a value for complex bindings
	 *                       (bindings involving other fields).
	 *
	 * @return     The logical value that was built (instance of the class returned from
	 *             getValueClass).
	 *
	 * @exception  LyUnsupportedFieldBindingException  Thrown if we don't known how to build a value
	 *                                                 from the given object.
	 *
	 * @see        LyFieldInfo#getPeerValue
	 * @see        leon.info.binding.LyFieldBinding
	 */
	
	public Object getValueFromObject(Object object, LyObject instance)
	{
        Object value = decodePeerValue(object);
        
        if (value == null)
            return null;

		if (value.getClass() == getValueClass())
			return value;

		// Build an array with given value(s)
		Object[] array = null;

		// If the object is a string, try to parse the string before...
		if (value instanceof String)
			value = parse((String)value);

		if (value instanceof Vector<?>)
		{
			// Gets an array from this vector
			Vector<?> vector = (Vector<?>)value;

			array = new Object[vector.size()];
			vector.copyInto(array);
		}
		else if (value.getClass().isArray())
		{
			int  i, n = Array.getLength(value);

			array = (Object[])Array.newInstance(Object.class, n);
			for (i = 0; i < n; i++)
				array[i] = Array.get(value, i);
		}
		else
		{
			array = new Object[1];
			array[0] = value;
		}

		if (array == null)
			return null;

		int i, n = array.length;

		if (n == 0)
			return null;

		// Build the array of values
		Object[] values = getArray(n);
		int		 nNull = 0;

		for (i = 0; i < n; i++)
		{
			values[i] = _fieldInfo.getValueFromObject(array[i], instance);
			if (values[i] == null)
				nNull++;
		}

		if (nNull == 0)
			return values;
		else if (n == nNull)
			return null;
		else
		{
			Object[] newValues = getArray(n - nNull);
			int		 j = 0;

			for (i = 0; i < n; i++)
				if (values[i] != null)
				{
					newValues[j] = values[i];
					j++;
				}

			return newValues;
		}
	} // end method getValueFromObject

	/**
	 * Returns an object of the given class from a value of this fieldInfo. The purpose of this
	 * method is to support transformations from logical to physical representation when simple
	 * bindings are used.
	 *
	 * @param      value        Value of this fieldInfo. This value is an instance of the class
	 *                          returned by getValueClass method.
	 * @param      objectClass  Expected class of the physical value.
	 * @param      instance     Instance containing the given value (could be null).
	 *
	 * @exception  LyUnsupportedFieldBindingException  Thrown if we don't known how to build a peer
	 *                                                 value from the given parameters.
	 *
	 * @see        LyFieldInfo#getValueClass
	 */
	public Object getPeerValue(Object value, Class<?> objectClass, LyObject instance)
	{
	    value = encodePeerValue(value, objectClass);
	    
		if (value == null)
			return null;

		if (value.getClass() == objectClass)
			return value;

		if (objectClass.isAssignableFrom(value.getClass()))
			return value;

		if (objectClass.isArray())
		{
			Object[] values = getValues(value);
			int		 i, n = values.length;
			Class<?>    fieldPeerClass = objectClass.getComponentType();
			Object[] array = (Object[])Array.newInstance(fieldPeerClass, n);

			for (i = 0; i < n; i++)
				array[i] = _fieldInfo.getPeerValue(values[i], fieldPeerClass, instance);

			return array;
		}
		else if (objectClass == String.class)
			return toString(value);

        return super.getPeerValue(value, objectClass, instance);
	} // end method getPeerValue

	/**
	 * Returns a table of objects for the given classes from a value of this fieldInfo.
	 *
	 * @param   value          Value of this fieldInfo.
	 * @param   objectClasses  Expected classes of the physical values.
	 * @param   instance       Instance containing the given value (could be null).
	 *
	 * @throws  LyUnsupportedFieldBindingException  If we don't known how to build the peer values
	 *                                              from the given parameters.
	 */
	public Object[] getPeerValues(Object value, Class<?>[] objectClasses, LyObject instance)
	{
	    Object[]   values = encodePeerValues(value, objectClasses);
	    
	    if (values != null)
	        return values;
	    
		if ((value == null)
				|| (objectClasses == null))
			return null;

		if (value instanceof Object[])
		{
		    int		 i, n = objectClasses.length;
		    Object[] result = new Object[n];

		    for (i = 0; i < n; i++)
		        result[i] = _fieldInfo.getPeerValue(((Object[])value)[i], objectClasses[i], null);
		    return result;
		}

		return super.getPeerValues(values, objectClasses, instance);
	}

	/**
	 * Gets the field info defined in this array.
	 */
	public LyFieldInfo getFieldInfo()
	{
		return _fieldInfo;
	}

	/**
	 * Gets minimum cardinality.
	 *
	 * @return  The minimum cardinality of this field.
	 */
	public short getMinCardinality()
	{
		load();
		return _minCard;
	}

	/**
	 * Gets maximum cardinality.
	 *
	 * @return  The maximum cardinality of this field.
	 */
	public short getMaxCardinality()
	{
		load();
		return _maxCard;
	}

	/**
	 * Gets the definition of the daemon managing this field.
	 *
	 * @return  The daemon managing this field.
	 */
	public LyArrayDaemon getDaemon()
	{
		return _daemon;
	}

	/**
	 * Gets the preferred number of rows shown to the user.
	 *
	 * @return  Number of rows, or -1 if no preference is given.
	 */
	public int getNRows()
	{
		return _nRows;
	}

	/**
	 * Sets field info defined in this array.
	 *
	 * @param  fieldInfo  the field info.
	 */
	protected void setFieldInfo(LyFieldInfo fieldInfo)
	{
		_fieldInfo = fieldInfo;
		initDaemon();
	}

	/**
	 * Sets the daemon information to manage this field.
	 *
	 * @param  daemon  The information about the daemon.
	 */
	void setDaemon(LyArrayDaemon daemon)
	{
		_daemon = daemon;
		initDaemon();
	}

	/**
	 * Internal method to initialize the internal structures managed by the daemon.
	 */
	void initDaemon()
	{
		if ((_daemon == null) || (_fieldInfo == null))
			return;

		boolean hasId = (_daemon.getIdBinding() != null);

		// Create internal class info with a field info per fieldBindng.
		String id = getId();

		LyClassInfo classInfo = new LyClassInfo("__" + id + ".arrayClass", null, null, null, null, false, false, null);

		// check if the binding is correct
		LyClassBinding daemonBinding = _daemon.getClassBinding();
		LyClassBinding parentBinding = ((_parent instanceof LyClassInfo) ? ((LyClassInfo)_parent).getBinding() : null);
		String		   currentLocation = daemonBinding.getLocation();
		String		   defaultLocation = _application.getDefaultLocationId();
		String		   parentLocation = ((parentBinding == null) ? defaultLocation : parentBinding.getLocation());

		if ((currentLocation == null)
				|| currentLocation.equals(defaultLocation))
		{
			// use parent location
			daemonBinding = new LyClassBinding(daemonBinding.getBinding(), parentLocation, daemonBinding.getFilter());
			_daemon.setClassBinding(daemonBinding);
		}

		classInfo.setBinding(_daemon.getClassBinding());
		classInfo.setCachePolicy(new LyInfoCache(LyInfoCache.CachePolicy.AUTO, null));

		// Create field info for the foreign key.
		LyRelationFieldInfo fkFieldInfo = new LyRelationFieldInfo("__" + id + ".fkField", null, null, null, false,
				false, LyRelationFieldInfo.Direction.NONE, LyRelationFieldInfo.Type.ASSOCIATION, null, false,
				false, LyRelationFieldInfo.ShowAs.DEFAULT);
		BitSet			    marks;

		marks = new BitSet();
		marks.set(LyFieldInfo.MARK_HIDDEN);
		marks.set(LyFieldInfo.MARK_NOT_NULL);
		marks.set(LyFieldInfo.MARK_OPTIONAL);
		if (!hasId)
			marks.set(LyFieldInfo.MARK_ID);

		fkFieldInfo.setMarks(marks);

		LyInfo parent = getParent();

		if (parent instanceof LyStructFieldInfo)
			fkFieldInfo.addClassInfo(((LyStructFieldInfo)parent).getClassInfo(0), null, -1);
		else
			fkFieldInfo.addClassInfo((LyClassInfo)getParent(), null, -1);

		classInfo.addFieldInfo(fkFieldInfo, -1);
		fkFieldInfo.setApplication(_application);
		fkFieldInfo.setParent(classInfo);
		fkFieldInfo.setBinding(_daemon.getFkBinding());

		// Create field info for the value
		LyFieldInfo valueFieldInfo = null;

		if (_fieldInfo instanceof LyChoiceFieldInfo)
		{
			LyChoiceFieldInfo choice = (LyChoiceFieldInfo)_fieldInfo;

			valueFieldInfo = new LyChoiceFieldInfo("__" + id + ".valueField", null, choice, choice.isMultiple(),
					choice.getSortBy(), choice.getType(), choice.isNot());
		}
		else if (_fieldInfo instanceof LyDateFieldInfo)
		{
			LyDateFieldInfo date = (LyDateFieldInfo)_fieldInfo;

			valueFieldInfo = new LyDateFieldInfo("__" + id + ".valueField", null, date, date.getType());
		}
		else if (_fieldInfo instanceof LyFileFieldInfo)
		{
			LyFileFieldInfo file = (LyFileFieldInfo)_fieldInfo;

			valueFieldInfo = new LyFileFieldInfo("__" + id + ".valueField", null, file, file.getMimeTypes(),
					file.getAttachmentHandler(), file.getAttachmentDirectory(), file.getDirectory(), file.getFilter());
		}
		else if (_fieldInfo instanceof LyNumberFieldInfo)
		{
			LyNumberFieldInfo number = (LyNumberFieldInfo)_fieldInfo;

			valueFieldInfo = new LyNumberFieldInfo("__" + id + ".valueField", null, number, number.getType(),
					number.isModulo());
		}
		else if (_fieldInfo instanceof LyTextFieldInfo)
		{
			LyTextFieldInfo text = (LyTextFieldInfo)_fieldInfo;

			valueFieldInfo = new LyTextFieldInfo("__" + id + ".valueField", null, text, text.isMultiple());
		}
		else if (_fieldInfo instanceof LyStructFieldInfo)
		{
			LyStructFieldInfo struct = (LyStructFieldInfo)_fieldInfo;

			valueFieldInfo = new LyStructFieldInfo("__" + id + ".valueField", null, null, struct, null,
					struct.isAbstract());
		}
		else if (_fieldInfo instanceof LyRelationFieldInfo)
		{
			LyRelationFieldInfo relation = (LyRelationFieldInfo)_fieldInfo;

			valueFieldInfo = new LyRelationFieldInfo("__" + id + ".valueField", null, null, relation,
					relation.isMultiple(), relation.isOrdered(), LyRelationFieldInfo.Direction.NONE, relation.getType(),
					null, false, false, LyRelationFieldInfo.ShowAs.DEFAULT);
		}

		if (valueFieldInfo != null)
		{
			valueFieldInfo.initFromTemplate(LyFieldInfo.PART_ALL);
			classInfo.addFieldInfo(valueFieldInfo, -1);
			valueFieldInfo.setApplication(_application);
			valueFieldInfo.setParent(classInfo);
			valueFieldInfo.setBinding(_daemon.getValueBinding());
		}

		if (_fieldInfo instanceof LyStructFieldInfo)
		{
			// Change class info of struct after initFromTemplate to avoid calls
			// of initFromTemplate on the corresponding class info
			LyStructFieldInfo	    struct = (LyStructFieldInfo)_fieldInfo;
			LyInfoList<LyClassInfo> classInfos = struct.getClassInfos();
			LyClassInfo			    structClassInfo;

			struct = (LyStructFieldInfo)valueFieldInfo;
			struct.resetClasses();

			for (int i = 0; i < classInfos.getSize(); i++)
			{
				structClassInfo = classInfos.getInfo(i);
				// Add class info to relation
				struct.addClassInfo(structClassInfo, null, -1);
			}
		}

		// Create field info for the index
		LyNumberFieldInfo indexFieldInfo = new LyNumberFieldInfo("__" + id + ".indexField", null, null,
				LyNumberFieldInfo.Type.INT, false);

		marks = new BitSet();
		marks.set(LyFieldInfo.MARK_HIDDEN);
		marks.set(LyFieldInfo.MARK_NOT_NULL);
		marks.set(LyFieldInfo.MARK_OPTIONAL);
		if (!hasId)
			marks.set(LyFieldInfo.MARK_ID);

		indexFieldInfo.setMarks(marks);
		classInfo.addFieldInfo(indexFieldInfo, -1);
		indexFieldInfo.setApplication(_application);
		indexFieldInfo.setParent(classInfo);
		indexFieldInfo.setBinding(_daemon.getIndexBinding());

		// create field info for the id
		if (hasId)
		{
			LyFieldInfo idFieldInfo = new LyNumberFieldInfo("__" + id + ".idField", null, null,
					LyNumberFieldInfo.Type.INT, false);

			marks = new BitSet();
			marks.set(LyFieldInfo.MARK_ID);
			marks.set(LyFieldInfo.MARK_HIDDEN);
			marks.set(LyFieldInfo.MARK_OPTIONAL);
			idFieldInfo.setMarks(marks);
			classInfo.addFieldInfo(idFieldInfo, -1);
			idFieldInfo.setApplication(_application);
			idFieldInfo.setParent(classInfo);
			idFieldInfo.setBinding(_daemon.getIdBinding());
		}

		// Add default actions for the array class.
		classInfo.addAction(_application.getAction("_consult"), -1);
		classInfo.addAction(_application.getAction("_create"), -1);
		classInfo.addAction(_application.getAction("_set"), -1);
		classInfo.addAction(_application.getAction("_delete"), -1);

		classInfo.setApplication(_application);
		classInfo.setParent(this);
		_daemon.setArrayClassInfo(classInfo);
	} // end method initDaemon

	/**
	 * Compares two objects for this fieldInfo.
	 */
	public boolean deepEquals(Object value1, Object value2, BitSet modifiers)
	{
		// We may compare relations first : if it's the same object then ok.
		if (super.equals(value1, value2, modifiers))
			return true;

		if ((value1 instanceof Object[]) && (value2 instanceof Object[]))
		{
			if (((Object[])value1).length != ((Object[])value2).length)
				return false;

			int j = 0;
			boolean equals = true;
			while (equals && j < ((Object[])value1).length)
			{
				Object v1 = ((Object[])value1)[j];
				Object v2 = ((Object[])value2)[j];

				// The object differs, we may compare all values
				LyObject object1 = null, object2 = null;

				if (v1 instanceof LyObject)
					object1 = (LyObject)v1;
				else if (v1 instanceof LyObjectId)
					object1 = ((LyObjectId)v1).getLogicalObject();

				if (v2 instanceof LyObject)
					object2 = (LyObject)v2;
				else if (v2 instanceof LyObjectId)
					object2 = ((LyObjectId)v2).getLogicalObject();

				if ((object1 != null) && (object2 != null))
				{
					// We may check better : look if values inside the array differ
					LyValueSet  valuesObj1 = object1.getValues();
					LyValue     valueObj1, valueObj2;
					LyFieldInfo fieldInfo;
					int		    n = valuesObj1.getSize();

					// go through values and stop on any difference
					if (n > 0)
					{
						if (getFieldInfo() instanceof LyStructFieldInfo)
							// Array of Struct
							equals = ((LyStructFieldInfo)getFieldInfo()).deepEquals(v1, v2, modifiers);
						else
						{
							int i;

							for (i = 0; i < n; i++)
							{
								valueObj1 = valuesObj1.getFieldValue(i);
								fieldInfo = (LyFieldInfo)valueObj1.getField();
								valueObj2 = object2.getFieldValue(fieldInfo.getId());

								equals = (fieldInfo.equals(valueObj1.getValue(), valueObj2.getValue(), modifiers));
							}
						}
					}
				}
				else 
					equals = (object1 == null && object2 == null);
				
				j++;
			}
			
			return equals;
		}

		return false;
	} // end method deepEquals

	/**
	 * Sets the preferred number of rows shown to the user.
	 *
	 * @param  nRows  Number of rows, or -1 if no preference.
	 */
	void setNRows(int nRows)
	{
		_nRows = nRows;
	}

	/**
	 * Sets minimum cardinality.
	 *
	 * @param  min  The minimum cardinality of this field.
	 */
	protected void setMinCardinality(short minCard)
	{
		_minCard = minCard;
	}

	/**
	 * Sets maximum cardinality.
	 *
	 * @param  max  The maximum cardinality of this field.
	 */
	protected void setMaxCardinality(short maxCard)
	{
		_maxCard = maxCard;
	}
} // end class LyArrayFieldInfo
