/*
 * LyBitSet.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;


/**
 * This class adds methods to be used on BitSets. It's not an extension of the standard class
 * java.util.BitSet because this class is final. Its only purpose is to add convenient methods for
 * testing properties and to hold further optimizations.
 * @see     java.util.BitSet
 * @version $Id: LyBitSet.java 26041 2010-07-06 06:54:12Z jlsaouli $
 * @author  Lyria
 */
public final class LyBitSet
	implements Cloneable
{
	public static final String __VERSION = "$Id: LyBitSet.java 26041 2010-07-06 06:54:12Z jlsaouli $";

    public enum Type
    {
    	// Action Mark type.
    	A_MARKS,
    	// Class Mark type.
    	C_MARKS,
    	// Field Mark type.
    	F_MARKS,
    	// Filter Mark type.
    	FL_MARKS
    }

	/** The bitset. */
	protected BitSet		   _marks;

	/**
	 * The bitset type.<br/>
	 * Possible values are:
	 * <ul>
	 *   <li>{@link #TYPE_A_MARKS}</li>
	 *   <li>{@link #TYPE_C_MARKS}</li>
	 *   <li>{@link #TYPE_F_MARKS}</li>
	 *   <li>{@link #TYPE_FL_MARKS}</li>
	 * </ul>
	 */
	protected Type			   _type;

	/**
	 * Checks if given marks are set in a bitset.
	 * @param     marks       The reference bitset.
	 * @param     testedMarks A set of bits to check with the reference bitset.
	 * @param     matchAll    Indicates if all marks must match or not.
	 * @return    true if matchAll is true and if tested marks matches all reference marks or if
	 *            matchAll is false and is at least one mark of the tested marks matches one mark of
	 *            the reference marks.
	 * @exception java.lang.IllegalArgumentException if the reference marks set is null or if the
	 *                                               testedMarks set is null or empty.
	 */
	public static boolean match(BitSet marks, BitSet testedMarks, boolean matchAll)
	{
		if (testedMarks == null)
			return false;

		if (matchAll)
		{
            BitSet  marks2 = (BitSet)testedMarks.clone();

            marks2.and(marks);
            return (marks2.cardinality() == testedMarks.cardinality());
		}
		else
            return marks.intersects(testedMarks);
	} // end method match

	/**
	 * Gives the number of bits set in a given bitset.
	 * @param  marks the bitset.
	 * @return number of bits set in the bitset.
	 */
	public static int getCount(BitSet marks)
	{
		int i, count, n = marks.size();

		for (i = 0, count = 0; i < n; i++)
		{
			if (marks.get(i))
				count++;
		}

		return count;
	}

	/**
	 * Constructor. Initialize {@link #_type} and {@link #_marks} with the given values.
	 * @param type   the BitSet type
	 * @param marks the bit set
	 */
	public LyBitSet(Type type, BitSet marks)
	{
		_type = type;
		_marks = marks;
	}

	/**
     * Constructor. Initialize {@link #_type} and {@link #_marks} with the given values.
     * @param type  the BitSet type
     * @param marks the bit set
     */
    public LyBitSet(Type type, String[] specificMarks)
    {
        _type = type;
        setSpecificMarks(specificMarks);
    }

	/**
	 * Gets the BitSet type. Type possible values are:
	 * <ul>
	 *   <li>{@link #Type.A_MARKS}</li>
	 *   <li>{@link #Type.C_MARKS}</li>
	 *   <li>{@link #Type.F_MARKS}</li>
	 *   <li>{@link #Type.FL_MARKS}</li>
	 * </ul>
	 * @return The type of the BitSet.
	 */
	public Type getType()
	{
		return _type;
	}

	/**
	 * Gets the java BitSet reference.
	 * @return The reference to the java BitSet.
	 */
	public BitSet getBitSet()
	{
		return ((_marks == null) ? null : (BitSet)(_marks.clone()));
	}

	/**
	 * Gets the marks size.
	 * @return Size of the marks.
	 */
	public int getSize()
	{
		return ((_marks == null) ? 0 : _marks.size());
	}

	/**
	 * Gets the value of the bit with the specified index.
	 * @param  mark the bit index
	 * @return the true if the bit with the specified index is set, false if it is not or if the
	 *         current BitSet is null
	 */
	public boolean get(int mark)
	{
		return ((_marks == null) ? false : _marks.get(mark));
	}

	/**
	 * Sets the bit at the specified index to true.
	 * @param mark the index of the bit to set
	 */
	public void set(int mark)
	{
		if (_marks == null)
			_marks = new BitSet();

		_marks.set(mark);
	}

	/**
	 * Checks if given marks are set in the current {@link #_marks bitset}.
	 * @param  testedMarks the set of bits to check.
	 * @param  matchAll    Indicates if all marks must match or not.
	 * @return true if matchAll is true and if all marks match or if matchAll is false and that at
	 *         least one mark match, false otherwise.
	 */
	public boolean match(BitSet testedMarks, boolean matchAll)
	{
		return match(_marks, testedMarks, matchAll);
	}

	/**
	 * Creates a new object of the same class as this object.
	 * @return    The new created object.
	 * @exception java.lang.InternalError if the super.clone() method throws a
	 *                                    CloneNotSupportedException.
	 */
	public synchronized Object clone()
	{
		try
		{
			LyBitSet newBitSet = (LyBitSet)super.clone();

			if (_marks != null)
				newBitSet._marks = (BitSet)(_marks.clone());

			return newBitSet;
		}
		catch (CloneNotSupportedException e)
		{
			// this shouldn't happen, since we are Cloneable
			throw new InternalError();
		}
	}

	/**
	 * Returns a string representation of this bit set. For every index for which this BitSet
	 * contains a bit in the set state, the decimal representation of that index is included in the
	 * result.<br />
	 * Example:
	 * <pre>
	   BitSet drPepper = new BitSet();
	 * </pre>
	 * <p>Now drPepper.toString() returns "{}".<br/>
	 * </p>
	 * <pre>
	   drPepper.set(2);
	 * </pre>
	 * <p>Now drPepper.toString() returns "{2}".</p>
	 * <pre>
	   drPepper.set(4);
	   drPepper.set(10);
	 * </pre>
	 * <p>Now drPepper.toString() returns "{2, 4, 10}".</p>
	 * @return a string representation of this bit set.
	 */
	public String toString()
	{
		if (_marks != null)
			return _marks.toString();
		else
			return super.toString();
	}

	/**
	 * Compares the current BitSet and the given BitSet values.
	 * @param  object object to compare to
	 * @return true if the two BitSets have exactly the same bits sets to true, false otherwise.
	 */
	public boolean equals(Object object)
	{
		if (this == object)
			return true;

		if (object instanceof LyBitSet)
		{
			LyBitSet compare = (LyBitSet)object;

			if ((_marks == compare._marks)
					|| ((_marks != null) && (_marks.equals(compare._marks))))
				return true;
		}

		return false;
	}

    /**
     * Gets mark index corresponding to the specific given mark name. Marks are represented as bits
     * in a bitset. Each mark has a bit position assigned to it. This mark allows to retrieve this
     * bit value by the name of the mark.
     *
     * @param  markName Name of the mark.
     * @param  add indicates if the.mark must be added if it doesn't exit
     * @return Returns mark value (bit position in a bits set)
     * @see    LyFieldInfo#getMarkName
     */
    public int getMark(String markName, boolean add)
    {
        return LyInfoTokens.getMarkIndex(_type, markName, add);
    }

    /**
     * Gets mark index corresponding to the specific given mark name. Marks are represented as bits
     * in a bitset. Each mark has a bit position assigned to it. This mark allows to retrieve this
     * bit value by the name of the mark.
     *
     * @param  markName Name of the mark.
     * @return Returns mark value (bit position in a bits set)
     * @see    LyFieldInfo#getMarkName
     */
    public int getMark(String markName)
    {
        return getMark(markName, false);
    }

    /**
     * Get mark name corresponding to a given mark value.
     *
     * @param  markIndex Mark value (index of a bit in a bits set).
     * @return Corresponding mark name.
     */
    public String getMarkName(int markIndex)
    {
        return LyInfoTokens.getMarkName(_type, markIndex);
    }

    /**
     * Accessor to the generic marks attached to this instance.
     */
    public BitSet getGenericMarks()
    {
        if (_marks == null)
            return null;

        int     nm = getGenericMarksNumber();

        return _marks.get(0, nm);
    }

    /**
     * Sets the specific marks attached to this instance (presented as an array of strings).
     */
    public BitSet setGenericMarks(BitSet marks)
    {
        int     nm = getGenericMarksNumber();

        if (_marks == null)
            _marks = (marks == null ? null : marks.get(0, nm));
        else
        {
        	_marks.clear(0, nm);
        	if (marks != null)
        	    _marks.or(marks.get(0, nm));
        }

        return _marks;
    }

    /**
     * Accessor to the specific marks attached to this instance (presented as an array of strings).
     */
    public String[] getSpecificMarks()
    {
        if (_marks == null)
            return null;

        int     nm = getGenericMarksNumber();

        if (_marks.size() < nm)
            return null;

        int                i, n = _marks.size();
        ArrayList<String>  list = null;

        for (i = nm; i < n; i++)
        {
            if (_marks.get(i))
            {
                if (list == null)
                    list = new ArrayList<String>();
                list.add(getMarkName(i));
            }
        }

        return (list == null ? null : list.toArray(new String[list.size()]));
    }

    /**
     * Sets the specific marks attached to this instance (presented as an array of strings).
     */
    public BitSet setSpecificMarks(String[] marks)
    {
        BitSet  bs = null;

        if (marks != null)
        {
            int     i, n = marks.length;
            int     index;

            bs = new BitSet();

            for (i = 0; i < n; i++)
            {
                index = getMark(marks[i], true);
                bs.set(index);
            }
        }

        return setSpecificMarks(bs);
    }

    /**
     * Sets the specific marks attached to this instance.
     */
    public BitSet setSpecificMarks(BitSet marks)
    {
        int     nm = getGenericMarksNumber();

        if (marks != null)
            marks.clear(0, nm);

        if (_marks == null)
            _marks = marks;
        else
        {
            if (_marks.length() >= nm)
                _marks.clear(nm, _marks.length());
            if (marks != null)
                _marks.or(marks);
        }

        return _marks;
    }

    /**
     * Gets number of generic marks according to bitset type.
     * @return Number of generic marks.
     */
    public int getGenericMarksNumber()
    {
        if (_type == Type.A_MARKS)
            return LyInfoTokens.actionMarks.length;
        else if (_type == Type.C_MARKS)
            return LyInfoTokens.classInfoMarks.length;
        else if (_type == Type.F_MARKS)
            return LyInfoTokens.fieldInfoMarks.length;
        else
            return -1;
    }
} // end class LyBitSet
