/*
 * LyChartInfo.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.io.*;
import java.util.*;
import leon.app.*;
import leon.misc.*;
import leon.info.infointerface.*;


/**
 * The LyChartInfo class implements the description of a chart applied on objects of the
 * application.
 *
 * @see leon.control.LyChartController
 */
public class LyChartInfo extends LyObjectInfo
	implements LyGetClassInfoInterface,
        LyExtraInfoInterface
{
	public static final String __version =
		"$Id: LyChartInfo.java 40733 2012-09-11 07:29:25Z jlsaouli $";

    public enum Type
    {
    	NUMBER,
        TEXT,
        DATE
    }

    public enum GraphType
    {
    	STACK_BAR,
        BAR,
        PIE,
        SCATTER_PLOT,
        PLOT,
        STACK_AREA
    }

    public enum ValType
    {
    	REAL,
        INTERVAL,
        SUM_REAL,
        COUNT_INTERVAL,
        COUNT_REAL
    }

	/** The Id of the LyClassInfo of the chart. Used only to save/load charts. */
	String					   _classInfoId = null;

	/** List of fields to analyse (Y). Elements are LyFieldInfo. */
	Vector<String>			   _selectedYFieldsInfo = null;

	/** The field on X axe (it can be null). */
	String					   _selectedXFieldInfo = null;

	/** The field for group (it can be null). */
	String					   _selectedGroupFieldInfo = null;

	/** Type of Y fields (from LyChartController). */
	Type					   _typeVarY;

	/** Type of X field (from LyChartController). */
	Type					   _typeVarX;

	/**
	 * Graphic type : PIE, SCATTER_PLOT, PLOT STACK_AREA, BAR, STACK_BAR, HI_LO. By default it is
	 * STACK_BAR.
	 */
	GraphType				   _selectedGraphType = GraphType.STACK_BAR;

	/** A String for the type of values for X (real, intervals). */
	ValType					   _xVal;

	/**
	 * A vector of values for the X intervals. Elements are the beginning, the width and an eventual
	 * time unity of interval.
	 */
	String[]           		   _xIntervals = null;

	/** A String for the type of values for Y (real, count or sum by interval). */
	ValType					   _yVal;

	/**
	 * A vector of values for the Y intervals. Elements are the beginning, the width and an eventual
	 * time unity of interval.
	 */
	String[]			        _yIntervals = null;

	/**
	 * A boolean used if the graphic type is Pie. True if there is one pie by Y var False if there
	 * is one pie by X value or interval
	 */
	boolean					   _pieBy = true;

	/** A String containing the title of the chart graphic. */
	String					   _title = null;

	/** A String containing the title of X axis. */
	String					   _xTitle = null;

	/** A String containing the title of Y axis. */
	String					   _yTitle = null;

	/**
	 * A vector of 3 values for the X axis. Pos0: min, pos1: max and pos2: distance between
	 * graduations.
	 */
	String[]       			   _xAxis = null;

	/**
	 * A vector of 3 values for the Y axis. Pos0: min, pos1: max and pos2: distance between
	 * graduations.
	 */
	String[]       			   _yAxis = null;

	/** A boolean used if the graphic type is not Pie,. */
	boolean					   _xVerticalLabels = false;

	/** A boolean used to show X Grid if the graphic type is not Pie,. */
	boolean					   _xGrid = true;

	/** A boolean used to show Y Grid if the graphic type is not Pie,. */
	boolean					   _yGrid = true;

	/** A boolean used to activate 3D effect (true = enabled),. */
	boolean					   _3dEffect = true;

	/** A boolean used to show/hide the legend (true = show),. */
	boolean					   _showLegend = true;

	/** A boolean used to show all the label tips (true = show),. */
	boolean					   _showTips = false;

	/** A boolean used to transform datas to percentiles (true = percentiles),. */
	boolean					   _ratios = false;

	/** Preference associated to the chart info. */
	LyPreference			   _preference = null;

	/**
	 * Initialisation of the data of the instance. This method is called by the constructor an may
	 * be overriden in subclasses.
	 */
	protected void init()
	{
		super.init();

		// Initialize class attributes
		_classInfoId = null;
		_selectedXFieldInfo = null;
		_selectedYFieldsInfo = new Vector<String>(1);
		_selectedGroupFieldInfo = null;
		_selectedGraphType = GraphType.STACK_BAR;
		_xIntervals = new String[3];
		_yIntervals = new String[3];
		_pieBy = true;
		_title = "";
		_xTitle = "";
		_yTitle = "";
		_xAxis = new String[3];
		_yAxis = new String[3];
		_xVerticalLabels = false;
		_xGrid = true;
		_yGrid = true;
		_3dEffect = true;
		_showLegend = true;
		_showTips = false;
		_ratios = false;
		_preference = null;
	} // end method init

	/**
	 * Constructor.
	 *
	 * @param id       The identifier of the object. May <b>not</b> be null.
	 * @param template Template used to construct this field.
	 */
	public LyChartInfo(String id, String name, LyChartInfo template)
	{
		super(id, template);

		_name = name;
	}

	/**
	 * Public constructor.
	 */
	public LyChartInfo()
	{
	}

	/**
	 * Constructor.
	 *
	 * @param name                   Name of the Chart.
	 * @param classInfoId            The Id of the LyClassInfo of the chart. Used only to save/load
	 *                               charts.
	 * @param selectedYFieldsInfo    List of fields to analyse (Y). Elements are LyFieldInfo.
	 * @param selectedXFieldInfo     The field id on X axe (it can be null).
	 * @param selectedGroupFieldInfo The field id for group (it can be null).
	 * @param typeVarY               Type of Y fields.
	 * @param typeVarX               Type of X field.
	 * @param selectedGraphType      Graphic type
	 * @param xVal                   type of values for X (real, intervals).
	 * @param xIntervals             values for the X intervals. Elements are the beginning, the
	 *                               width and an eventual time unity of interval.
	 * @param yVal                   type of values for Y (real, count or sum by interval).
	 * @param yIntervals             values for the Y intervals. Elements are the beginning, the
	 *                               width and an eventual time unity of interval.
	 * @param pieBy                  True if there is one pie by Y var, False if there is one pie by
	 *                               X value or interval
	 * @param title                  the title of the chart graphic.
	 * @param xTitle                 the title of X axis.
	 * @param yTitle                 the title of Y axis.
	 * @param xAxis                  A vector of 3 values for the X axis. Pos0: min, pos1: max and
	 *                               pos2: distance between graduations.
	 * @param yAxis                  A vector of 3 values for the Y axis. Pos0: min, pos1: max and
	 *                               pos2: distance between graduations.
	 * @param xVerticalLabels        used if the graphic type is not Pie,
	 * @param xGrid                  used to show X Grid if the graphic type is not Pie,
	 * @param yGrid                  used to show Y Grid if the graphic type is not Pie,
	 * @param 3dEffect               used to enable 3D Effect,
	 */
	public LyChartInfo(String id, String name, String classInfoId, String selectedXFieldInfo,
		String selectedGroupFieldInfo, Type typeVarY, Type typeVarX,
        GraphType selectedGraphType, ValType xVal, String[] xIntervals,
		ValType yVal, String[] yIntervals, boolean pieBy, String title,
		String xTitle, String yTitle, String[] xAxis, String[] yAxis,
		boolean xVerticalLabels, boolean xGrid, boolean yGrid, boolean d3Effect,
        boolean showLegend, boolean showTips, boolean ratios)
	{
		super(id);

		// Initialize class attributes
		setName(name);
		_classInfoId = classInfoId;
		_selectedXFieldInfo = selectedXFieldInfo;
		_selectedGroupFieldInfo = selectedGroupFieldInfo;
		_typeVarX = (typeVarX == null ? Type.TEXT : typeVarX);
		_typeVarY = (typeVarY == null ? Type.TEXT : typeVarY);
		_selectedGraphType = (selectedGraphType == null ? GraphType.STACK_BAR : selectedGraphType);
		_xIntervals = xIntervals;
		_xVal = (xVal == null ? ValType.REAL : xVal);
		_yIntervals = yIntervals;
		_yVal = (yVal == null ? ValType.REAL : yVal);
		_pieBy = pieBy;
		_title = title;
		_xTitle = xTitle;
		_yTitle = yTitle;
		_xAxis = xAxis;
		_yAxis = yAxis;
		_xVerticalLabels = xVerticalLabels;
		_xGrid = xGrid;
		_yGrid = yGrid;
		_3dEffect = d3Effect;
		_showLegend = showLegend;
		_showTips = showTips;
		_ratios = ratios;
	}

	/**
	 * Duplicate all the attributes of a LyChartInfo.
	 *
	 * @param  id the identifier of the new LyChartInfo
	 * @return an identical LyChartInfo with the same values but different objects and differnet
	 *         pointers
	 */
	public LyChartInfo cloneChartInfo(String id)
	{
		LyChartInfo newChartInfo = new LyChartInfo(id, null, this);

		if (null != getName())
			newChartInfo.setName(getName());
		if (null != getClassInfoId())
			newChartInfo.setClassInfoId(getClassInfoId());
		if (null != getSelectedXFieldInfo())
			newChartInfo.setSelectedXFieldInfo(getSelectedXFieldInfo());
		if (null != getSelectedGroupFieldInfo())
			newChartInfo.setSelectedGroupFieldInfo(getSelectedGroupFieldInfo());
		// y variables
		if (null != getSelectedYFieldsInfo())
		{
			for (int i = 0; i < getSelectedYFieldsInfo().size(); i++)
			{
				if (null != getSelectedYFieldsInfo().elementAt(i))
					newChartInfo.addYField(
							getSelectedYFieldsInfo().elementAt(i).toString());
				else
					newChartInfo.addYField(null);
			}
		}

		if (null != getTypeVarY())
			newChartInfo.setTypeVarY(getTypeVarY());
		if (null != getTypeVarX())
			newChartInfo.setTypeVarX(getTypeVarX());
		if (null != getXVal())
			newChartInfo.setXVal(getXVal());
		if (null != getYVal())
			newChartInfo.setYVal(getYVal());

		newChartInfo.setSelectedGraphType(getSelectedGraphType());

		// x intervals
		if (null != _xIntervals)
		{
			int	     i, n = _xIntervals.length;
			String[] xInt = new String[n];

			for (i = 0; i < n; i++)
			    xInt[i] = _xIntervals[i];
			newChartInfo.setXIntervals(xInt);
		}

		// y intervals
		if (null != _yIntervals)
		{
			int	     i, n = _yIntervals.length;
			String[] yInt = new String[n];

			for (i = 0; i < n; i++)
				yInt[i] = _yIntervals[i];
			newChartInfo.setYIntervals(yInt);
		}

		newChartInfo.setPieBy(Boolean.valueOf(getPieBy()).booleanValue());
		if (null != getTitle())
			newChartInfo.setTitle(getTitle());
		if (null != getXTitle())
			newChartInfo.setXTitle(getXTitle());
		if (null != getYTitle())
			newChartInfo.setYTitle(getYTitle());

		// x axis
		if (null != _xAxis)
		{
			int	     i, n = _xAxis.length;
			String[] xAxis = new String[n];

			for (i = 0; i < n; i++)
				xAxis[i] = _xAxis[i];
			newChartInfo.setXAxis(xAxis);
		}

		// y axis
        if (null != _yAxis)
        {
            int      i, n = _yAxis.length;
            String[] yAxis = new String[n];

            for (i = 0; i < n; i++)
                yAxis[i] = _yAxis[i];
            newChartInfo.setYAxis(yAxis);
        }

		newChartInfo.setXVerticalLabels(Boolean.valueOf(getXVerticalLabels()).booleanValue());
		newChartInfo.setXGrid(Boolean.valueOf(getXGrid()).booleanValue());
		newChartInfo.setYGrid(Boolean.valueOf(getYGrid()).booleanValue());
		newChartInfo.set3dEffect(Boolean.valueOf(get3dEffect()).booleanValue());
		newChartInfo.setShowLegend(Boolean.valueOf(getShowLegend()).booleanValue());
		newChartInfo.setShowTips(Boolean.valueOf(getShowTips()).booleanValue());
		newChartInfo.setRatios(Boolean.valueOf(getRatios()).booleanValue());
		newChartInfo.setFile(getFile());

		if (_preference != null)
			newChartInfo.setPreference(_preference.clonePreference());

        newChartInfo.setApplication(_application);
		return newChartInfo;
	} // end method cloneChartInfo

	/**
	 * Accessor to the Id of the LyClassInfo.
	 *
	 * @return the Id of the LyClassInfo.
	 */
	public String getClassInfoId()
	{
		return _classInfoId;
	}

	/**
	 * Accessor to the Y fields of the chart.
	 *
	 * @return the Y fields of the chart.
	 */
	public Vector<String> getSelectedYFieldsInfo()
	{
		return _selectedYFieldsInfo;
	}

	/**
	 * Accessor to the X field of the chart.
	 *
	 * @return the X field of the chart.
	 */
	public String getSelectedXFieldInfo()
	{
		return _selectedXFieldInfo;
	}

	/**
	 * Accessor to the group field of the chart.
	 *
	 * @return the group field of the chart.
	 */
	public String getSelectedGroupFieldInfo()
	{
		return _selectedGroupFieldInfo;
	}

	/**
	 * Accessor to the y type of variables.
	 *
	 * @return the y type of variables.
	 */
	public Type getTypeVarY()
	{
		return _typeVarY;
	}

	/**
	 * Accessor to the x type of variable.
	 *
	 * @return the x type of variable.
	 */
	public Type getTypeVarX()
	{
		return _typeVarX;
	}

	/**
	 * Accessor to the graphic type of the chart.
	 *
	 * @return the graphic type of the chart.
	 */
	public GraphType getSelectedGraphType()
	{
		return _selectedGraphType;
	}

	/**
	 * Accessor to the x type of values of the chart (real, intervals).
	 *
	 * @return the x type of values of the chart
	 */
	public ValType getXVal()
	{
		return _xVal;
	}

	/**
	 * Accessor to the y type of values of the chart (real, sum or count by interval).
	 *
	 * @return the y type of values of the chart
	 */
	public ValType getYVal()
	{
		return _yVal;
	}

	/**
	 * Accessor to the x intervals of the chart.
	 *
	 * @return the x intervals of the chart.
	 */
	public String[] getXIntervals()
	{
		return _xIntervals;
	}

	/**
	 * Accessor to the x interval begin.
	 *
	 * @return the x interval begin.
	 */
	public String getXIntervalBegin()
	{
	    return (_xIntervals == null ? null : _xIntervals[0]);
	}

	/**
	 * Accessor to the x interval length.
	 *
	 * @return the x interval length.
	 */
	public String getXIntervalLength()
	{
		return (_xIntervals == null ? null : _xIntervals[1]);
	}

	/**
	 * Accessor to the x interval unit.
	 *
	 * @return the x interval unit.
	 */
	public String getXIntervalUnit()
	{
	    return (_xIntervals == null ? null : _xIntervals[2]);
	}

	/**
	 * Accessor to the y intervals of the chart.
	 *
	 * @return the y intervals of the chart.
	 */
	public String[] getYIntervals()
	{
		return _yIntervals;
	}

	/**
	 * Accessor to the y interval begin.
	 *
	 * @return the y interval begin.
	 */
	public String getYIntervalBegin()
	{
		return (_yIntervals == null ? null : _yIntervals[0]);
	}

	/**
	 * Accessor to the y interval length.
	 *
	 * @return the y interval length.
	 */
	public String getYIntervalLength()
	{
	    return (_yIntervals == null ? null : _yIntervals[1]);
	}

	/**
	 * Accessor to the y interval unit.
	 *
	 * @return the y interval unit.
	 */
	public String getYIntervalUnit()
	{
	    return (_yIntervals == null ? null : _yIntervals[2]);
	}

	/**
	 * Accessor to the type of the pies, if there are any.
	 *
	 * @return the type of the pies, if there are any.
	 */
	public boolean getPieBy()
	{
		return _pieBy;
	}

	/**
	 * Accessor to the title of the chart graphic.
	 *
	 * @return the title of the chart graphic.
	 */
	public String getTitle()
	{
		return _title;
	}

	/**
	 * Accessor to the title of X axis.
	 *
	 * @return the title of X axis.
	 */
	public String getXTitle()
	{
		return _xTitle;
	}

	/**
	 * Accessor to the title of Y axis.
	 *
	 * @return the title of Y axis.
	 */
	public String getYTitle()
	{
		return _yTitle;
	}

	/**
	 * Accessor to the vector of 3 values for the X axis.
	 *
	 * @return the vector of 3 values for the X axis.
	 */
	public String[] getXAxis()
	{
		return _xAxis;
	}

	/**
	 * Accessor to the value for the X axis min.
	 *
	 * @return the X axis min.
	 */
	public String getXAxisMin()
	{
		return (_xAxis == null ? null : _xAxis[0]);
	}

	/**
	 * Accessor to the value for the X axis max.
	 *
	 * @return the X axis max.
	 */
	public String getXAxisMax()
	{
	    return (_xAxis == null ? null : _xAxis[1]);
	}

	/**
	 * Accessor to the value for the X axis width.
	 *
	 * @return the vector of 3 values for the X axis.
	 */
	public String getXAxisWidth()
	{
	    return (_xAxis == null ? null : _xAxis[2]);
	}

	/**
	 * Accessor to the value for the Y axis min.
	 *
	 * @return the X axis min.
	 */
	public String getYAxisMin()
	{
		return (_yAxis == null? null : _yAxis[0]);
	}

	/**
	 * Accessor to the value for the X axis max.
	 *
	 * @return the X axis max.
	 */
	public String getYAxisMax()
	{
	    return (_yAxis == null? null : _yAxis[1]);
	}

	/**
	 * Accessor to the value for the X axis width.
	 *
	 * @return the vector of 3 values for the X axis.
	 */
	public String getYAxisWidth()
	{
	    return (_yAxis == null? null : _yAxis[2]);
	}

	/**
	 * Accessor to the vector of 3 values for the Y axis.
	 *
	 * @return the vector of 3 values for the Y axis.
	 */
	public String[] getYAxis()
	{
		return _yAxis;
	}

	/**
	 * Accessor to the type of X labels : vertical if true, horizontal if false.
	 *
	 * @return vertical if true, horizontal if false.
	 */
	public boolean getXVerticalLabels()
	{
		return _xVerticalLabels;
	}

	/**
	 * Accessor to X Grid : enabled if true, disabled if false.
	 *
	 * @return enabled if true, disabled if false.
	 */
	public boolean getXGrid()
	{
		return _xGrid;
	}

	/**
	 * Accessor to Y Grid : enabled if true, disabled if false.
	 *
	 * @return enabled if true, disabled if false.
	 */
	public boolean getYGrid()
	{
		return _yGrid;
	}

	/**
	 * Accessor to 3D Effect : enabled if true, disabled if false.
	 *
	 * @return enabled if true, disabled if false.
	 */
	public boolean get3dEffect()
	{
		return _3dEffect;
	}

	/**
	 * Accessor to legend : show if true, hide if false.
	 *
	 * @return show if true, hide if false.
	 */
	public boolean getShowLegend()
	{
		return _showLegend;
	}

	/**
	 * Accessor to label tips : show if true, hide (or show only when the mouse is over the point)
	 * if false.
	 *
	 * @return label tips : show if true, hide (or show only when the mouse is over the point) if
	 *         false.
	 */
	public boolean getShowTips()
	{
		return _showTips;
	}

	/**
	 * Accessor to ratios : ratios if true, real values if false.
	 *
	 * @return the ratios : ratios if true, real values if false.
	 */
	public boolean getRatios()
	{
		return _ratios;
	}

	/**
	 * Accessor to the preference of this chart.
	 *
	 * @return the preference of this chart.
	 */
	public LyPreference getPreference()
	{
		return _preference;
	}

	/**
	 * Gets the ClassInfo of current controller.
	 *
	 * @return the classInfo associated to the controller.
	 */
	public LyClassInfo getClassInfo()
	{
		if (_classInfoId == null)
			return null;
		else
		{
			LyApplication application = getApplication();

			if (application == null)
				return null;
			return application.getClassInfo(_classInfoId);
		}
	}

	/**
	 * Method for retrieving the name of the image associated to the type of chart.
	 *
	 * @return Name of the image resource.
	 */
	public String getImage()
	{
		switch (_selectedGraphType)
		{
		case STACK_BAR:
			return "LY_CHART_STACK_BAR";
		case BAR:
			return "LY_CHART_BAR";
		case PIE:
			return "LY_CHART_PIE";
		case SCATTER_PLOT:
			return "LY_CHART_SCATTER_PLOT";
		case PLOT:
			return "LY_CHART_PLOT";
		case STACK_AREA:
			return "LY_CHART_AREA";
		default:
			return null;
		}
	}

	/**
	 * Sets the id of the instance.
	 *
	 * @param id Identifier of the instance.
	 */
	public void setId(String id)
	{
		super.setId(id);
	}

	/**
	 * Sets the name of the instance.
	 *
	 * @param name Name of the instance.
	 */
	public void setName(String name)
	{
		super.setName(name);
	}

	/**
	 * Sets the classInfo of the chart.
	 *
	 * @param classInfoId the classInfo of the chart.
	 */
	public void setClassInfoId(String classInfoId)
	{
		_classInfoId = classInfoId;
	}

	/**
	 * Sets the Y fields of the chart.
	 *
	 * @param selectedYFieldsInfo the Y fields of the chart.
	 */
	public void setSelectedYFieldsInfo(Vector<String> selectedYFieldsInfo)
	{
		_selectedYFieldsInfo = selectedYFieldsInfo;
	}

	/**
	 * Sets the X field of the chart.
	 *
	 * @param selectedXFieldInfo the X field of the chart.
	 */
	public void setSelectedXFieldInfo(String selectedXFieldInfo)
	{
		_selectedXFieldInfo = selectedXFieldInfo;
	}

	/**
	 * Sets the group field of the chart.
	 *
	 * @param selectedGroupFieldInfo the group field of the chart.
	 */
	public void setSelectedGroupFieldInfo(String selectedGroupFieldInfo)
	{
		_selectedGroupFieldInfo = selectedGroupFieldInfo;
	}

	/**
	 * Sets the description of the type of y variables.
	 *
	 * @param typeVarY the description of the type of y variables.
	 */
	public void setTypeVarY(Type typeVarY)
	{
		_typeVarY = typeVarY;
	}

	/**
	 * Sets the description of the type of x variable.
	 *
	 * @param typeVarX the description of the type of x variable.
	 */
	public void setTypeVarX(Type typeVarX)
	{
		_typeVarX = typeVarX;
	}

	/**
	 * Sets the graphic type of the chart.
	 *
	 * @param selectedGraphType the graphic type of the chart.
	 */
	public void setSelectedGraphType(GraphType selectedGraphType)
	{
		_selectedGraphType = selectedGraphType;
	}

	/**
	 * Sets the x type of values of the chart (real, intervals).
	 *
	 * @param xVal the x type of values of the chart (real, intervals).
	 */
	public void setXVal(ValType xVal)
	{
		_xVal = xVal;
	}

	/**
	 * Sets the y type of values of the chart (real, sum or count by interval).
	 *
	 * @param yVal the y type of values of the chart (real, sum or count by interval).
	 */
	public void setYVal(ValType yVal)
	{
		_yVal = yVal;
	}

	/**
     * Sets the x interval begin.
     */
    public void setXIntervalBegin(String xIntervalBegin)
    {
        if (_xIntervals == null)
            _xIntervals = new String[3];
        _xIntervals[0] = xIntervalBegin;
    }

    /**
     * Sets the x interval length.
     */
    public void setXIntervalLength(String xIntervalLength)
    {
        if (_xIntervals == null)
            _xIntervals = new String[3];
        _xIntervals[1] = xIntervalLength;
    }

    /**
     * Sets the x interval unit.
     */
    public void setXIntervalUnit(String xIntervalUnit)
    {
        if (_xIntervals == null)
            _xIntervals = new String[3];
        _xIntervals[2] = xIntervalUnit;
    }

	/**
	 * Sets the X intervals of the chart.
	 *
	 * @param xIntervals the X intervals of the chart.
	 */
	public void setXIntervals(String[] xIntervals)
	{
		_xIntervals = xIntervals;
	}

	/**
     * Sets the y interval begin.
     */
    public void setYIntervalBegin(String yIntervalBegin)
    {
        if (_yIntervals == null)
            _yIntervals = new String[3];
        _yIntervals[0] = yIntervalBegin;
    }

    /**
     * Sets the y interval length.
     */
    public void setYIntervalLength(String yIntervalLength)
    {
        if (_yIntervals == null)
            _yIntervals = new String[3];
        _yIntervals[1] = yIntervalLength;
    }

    /**
     * Sets the y interval unit.
     */
    public void setYIntervalUnit(String yIntervalUnit)
    {
        if (_yIntervals == null)
            _yIntervals = new String[3];
        _yIntervals[2] = yIntervalUnit;
    }

	/**
	 * Sets the Y intervals of the chart.
	 *
	 * @param yIntervals the Y intervals of the chart.
	 */
	public void setYIntervals(String[] yIntervals)
	{
		_yIntervals = yIntervals;
	}

	/**
	 * Sets the type of the pies, if there are any.
	 *
	 * @param pieBy the type of the pies, if there are any.
	 */
	public void setPieBy(boolean pieBy)
	{
		_pieBy = pieBy;
	}

	/**
	 * Sets the title of the chart graphic.
	 *
	 * @param title the title of the chart graphic.
	 */
	public void setTitle(String title)
	{
		_title = title;
	}

	/**
	 * Sets the title of X axis.
	 *
	 * @param xTitle the title of X axis.
	 */
	public void setXTitle(String xTitle)
	{
		_xTitle = xTitle;
	}

	/**
	 * Sets the title of Y axis.
	 *
	 * @param yTitle the title of Y axis.
	 */
	public void setYTitle(String yTitle)
	{
		_yTitle = yTitle;
	}

	/**
     * Sets the value for the Y axis min.
     */
    public void setYAxisMin(String yAxisMin)
    {
        if (_yAxis == null)
            _yAxis = new String[3];
        _yAxis[0] = yAxisMin;
    }

    /**
     * Sets the value for the Y axis max.
     */
    public void setYAxisMax(String yAxisMax)
    {
        if (_yAxis == null)
            _yAxis = new String[3];
        _yAxis[1] = yAxisMax;
    }

    /**
     * Sets the value for the Y axis width.
     */
    public void setYAxisWidth(String yAxixWidth)
    {
        if (_yAxis == null)
            _yAxis = new String[3];
        _yAxis[2] = yAxixWidth;
    }

    /**
     * Sets the value for the X axis min.
     */
    public void setXAxisMin(String xAxisMin)
    {
        if (_xAxis == null)
            _xAxis = new String[3];
        _xAxis[0] = xAxisMin;
    }

    /**
     * Sets the value for the X axis max.
     */
    public void setXAxisMax(String xAxisMax)
    {
        if (_xAxis == null)
            _xAxis = new String[3];
        _xAxis[1] = xAxisMax;
    }

    /**
     * Sets the value for the X axis width.
     */
    public void setXAxisWidth(String xAxixWidth)
    {
        if (_xAxis == null)
            _xAxis = new String[3];
        _xAxis[2] = xAxixWidth;
    }

	/**
	 * Sets the vector of 3 values for the X axis.
	 *
	 * @param xAxis the vector of 3 values for the X axis.
	 */
	public void setXAxis(String[] xAxis)
	{
		_xAxis = xAxis;
	}

	/**
	 * Sets the vector of 3 values for the Y axis.
	 *
	 * @param yAxis the vector of 3 values for the Y axis.
	 */
	public void setYAxis(String[] yAxis)
	{
		_yAxis = yAxis;
	}

	/**
	 * Sets the type of X labels : vertical if true, horizontal if false.
	 *
	 * @param xVerticalLabels vertical if true, horizontal if false.
	 */
	public void setXVerticalLabels(boolean xVerticalLabels)
	{
		_xVerticalLabels = xVerticalLabels;
	}

	/**
	 * Sets X Grid : enabled if true, disabled if false.
	 *
	 * @param xGrid enabled if true, disabled if false.
	 */
	public void setXGrid(boolean xGrid)
	{
		_xGrid = xGrid;
	}

	/**
	 * Sets Y Grid : enabled if true, disabled if false.
	 *
	 * @param yGrid enabled if true, disabled if false.
	 */
	public void setYGrid(boolean yGrid)
	{
		_yGrid = yGrid;
	}

	/**
	 * Sets 3D Effect : enabled if true, disabled if false.
	 *
	 * @param d3Effect enabled if true, disabled if false.
	 */
	public void set3dEffect(boolean d3Effect)
	{
		_3dEffect = d3Effect;
	}

	/**
	 * Sets the legend : show if true, hide if false.
	 *
	 * @param showLegend show if true, hide if false.
	 */
	public void setShowLegend(boolean showLegend)
	{
		_showLegend = showLegend;
	}

	/**
	 * Sets the label tips : show if true, hide if false.
	 *
	 * @param showTips show if true, hide if false.
	 */
	public void setShowTips(boolean showTips)
	{
		_showTips = showTips;
	}

	/**
	 * Sets the ratios : ratios if true, real values if false.
	 *
	 * @param ratios ratios if true, real values if false.
	 */
	public void setRatios(boolean ratios)
	{
		_ratios = ratios;
	}

	/**
	 * Sets the preference.
	 *
	 * @param the preference of this chart
	 */
	public void setPreference(LyPreference preference)
	{
		_preference = preference;
	}

	/**
	 * Add an element in Y fields id vector.
	 *
	 * @param yField an element in Y fields id vector.
	 */
	public void addYField(String yField)
	{
		if (_selectedYFieldsInfo == null)
			_selectedYFieldsInfo = new Vector<String>(1);
		_selectedYFieldsInfo.addElement(yField);
	}

    /**
     * Gets all fields that this chart info needs to be drawed.
     * @return List on needed fields.
     */
    public LyInfoList<LyFieldInfo> getFields()
    {
        LyInfoList<LyFieldInfo>     fields = new LyInfoList<LyFieldInfo>();
        LyFieldInfo                 field;

        if (_selectedXFieldInfo != null)
        {
            field = (LyFieldInfo)_application.getInfo(_selectedXFieldInfo);
            if (field != null)
                fields.addInfo(field);
        }

        if (_selectedYFieldsInfo != null)
        {
            int     i, n = _selectedYFieldsInfo.size();
            String  fieldId;

            for (i = 0; i < n; i++)
            {
                fieldId = _selectedYFieldsInfo.elementAt(i);
                field = (LyFieldInfo)_application.getInfo(fieldId);
                if (field != null)
                    fields.addInfo(field);
            }
        }

        if (_selectedGroupFieldInfo != null)
        {
            field = (LyFieldInfo)_application.getInfo(_selectedGroupFieldInfo);
            if (field != null)
                fields.addInfo(field);
        }

        return fields;
    }

	/**
	 * Make a string representing the object.
	 *
	 * @return a string representing the object.
	 */
	public String toString2()
	{
		StringBuffer sb = new StringBuffer();

		sb.append("xField : ").append(_selectedXFieldInfo).append('\n');
		sb.append("groupField : ").append(_selectedGroupFieldInfo).append('\n');
		sb.append("typeVarY    : ").append(_typeVarY).append('\n');
		sb.append("typeVarX : ").append(_typeVarX).append('\n');
		sb.append("graphType  : ").append(_selectedGraphType).append('\n');
		sb.append("xVal  : ").append(_xVal).append('\n');
		sb.append("yVal  : ").append(_yVal).append('\n');
		sb.append("xIntervals  : ").append(String.valueOf(_xIntervals)).append('\n');
		sb.append("yIntervals  : ").append(String.valueOf(_yIntervals)).append('\n');
		sb.append("pieBy  : ").append(_pieBy).append('\n');
		sb.append("titles  : ").append(_title).append(_xTitle).append(_yTitle).append('\n');
		sb.append("xAxis  : ").append(String.valueOf(_xAxis)).append('\n');
		sb.append("yAxis  : ").append(String.valueOf(_yAxis)).append('\n');
		sb.append("xVerticalLabels  : ").append(_xVerticalLabels).append('\n');
		sb.append("yFields : ");
		if (_selectedYFieldsInfo != null)
		{
			for (int i = 0; i < _selectedYFieldsInfo.size(); i++)
				sb.append(_selectedYFieldsInfo.elementAt(i)).append("  ");
		}

		return (sb.toString());
	}

	/* XML Writing */

    /**
     * Get owner of this chart.
     * @return The owner of this chart.
     */
    public String getUser()
    {
        return null;
    }

    /**
     * Method for saving preference file.
     */
    public boolean saveFile(LyXmlWriterInterface writer)
    {
        String      fileName;
        
        if (writer == null)
        	fileName = getFile();
        else
        	fileName = writer.getFile(this);

        if (fileName == null)
        {
            String   dir, userId = getUser();

            if (userId == null)
            {
                userId = "";
                dir = "LY_CHART_DIR";
            }
            else
            {
                userId = userId.substring(userId.indexOf(leon.data.LyObjectId.CLASS_SEP) + 1) + "_";
                dir = "LY_USER_CHART_DIR";
            }

            LyEnvironment environment = getEnvironment();
            String        dirPath = environment.getFile(dir);
            File          fileDirectory = null;

            if (dirPath != null)
                fileDirectory = new File(dirPath);

            // Try to create the directory if needed
            if ((fileDirectory != null) && (!fileDirectory.exists()))
            {
                try
                {
                    environment.logInfo("Creating directory :" + dirPath);
                    fileDirectory.mkdirs();
                }
                catch (SecurityException e)
                {
                    environment.logError("Unable to create directory :" + dirPath);
                    return false;
                }
            }

            File    file;
            int     i = 0;

            // find a file name which has not been affected
            do
            {
                fileName = dirPath + "/chart_" + userId + i + ".xml";
                file = new File(fileName);
                i++;
            }
            while (file.exists());

            setFile(fileName);
        }

        // Create xml file for the preference
        try
        {
            FileOutputStream 	output = new FileOutputStream(fileName);
            LyXmlPrintWriter    pw = new LyXmlPrintWriter(output);

            if (writer == null)
            	writeXml(pw, true, true);
            else
            	writer.writeXml(pw, this);

            pw.close();
            return true;
        }
        catch (Exception e)
        {
            LyEnvironment   environment = getEnvironment();

            environment.writeException(e);
            return false;
        }
    }

    /**
     * Method for freeing extra info (and sub-infos).
     */
    public void free()
    {
        if (_preference != null)
            _preference.free();
        super.free();
    }

    /**
     * Method for deleting preference file.
     */
    public boolean deleteFile()
    {
        String  filePath = getFile();

        if (filePath == null)
            return false;

        try
        {
            // Delete the file
            File    file = new File(filePath);

            return file.delete();
        }
        catch (SecurityException e)
        {
            LyEnvironment   environment = getEnvironment();

            environment.writeException(e);
            return false;
        }
    }

    /**
     * Method to rewrite XML corresponding to this component.
     *
     * @param pw The printWriter on which the XML is written.
     * @param header Boolean indicating if header must be written.
     * @param header Boolean indicating if body must be written.
     */
    public void writeXml(PrintWriter pw, boolean header, boolean body)
    {
        if (header)
        {
            pw.println("<?xml version='1.0' encoding='" + LyString.DEFAULT_CHARSET + "'?>");
            pw.println("<!DOCTYPE CHART SYSTEM 'leon/conf/dtd/leon_gui.dtd'>");
        }

        if (body)
        	writeXml(pw);
    }

	/**
	 * Method to rewrite XML corresponding to this component.
	 *
	 * @param pw the printWriter on which the XML is written.
	 */
	public void writeXml(PrintWriter pw)
	{
		// Write tag with its attributes
		pw.print("<CHART");

		pw.print(" id='" + LyString.getXmlString(_id) + "'");

		pw.print(" name='" + LyString.getXmlString(_name) + "'");

		pw.print(" classInfoId='" + LyString.getXmlString(_classInfoId) + "'");

		// first parameter window
		pw.print(" graphType='" + _selectedGraphType.toString() + "'");

		if (null != _selectedXFieldInfo)
			pw.print(" xField='" + LyString.getXmlString(_selectedXFieldInfo) + "'");

		if (null != _selectedGroupFieldInfo)
			pw.print(" groupField='" + LyString.getXmlString(_selectedGroupFieldInfo) + "'");

		pw.print(" typeVarX='" + _typeVarX.toString() + "'");

        pw.print(" typeVarY='" + _typeVarY.toString() + "'");

		pw.println();
		pw.print("      ");

		// second parameter window
        pw.print(" xVal='" + _xVal.toString() + "'");
        pw.print(" yVal='" + _yVal.toString() + "'");

		if (null != _xIntervals)
		{
			if ((_xIntervals.length > 0) && (null != _xIntervals[0]))
				pw.print("\n       xIntsBegin='" + _xIntervals[0] + "'");
			if ((_xIntervals.length > 1) && (null != _xIntervals[1]))
				pw.print(" xIntsLength='" + _xIntervals[1] + "'");
			if ((_xIntervals.length > 2) && (null != _xIntervals[2]))
				pw.print(" xIntsUnit='" + _xIntervals[2] + "'");
		}

		if (null != _yIntervals)
		{
			if ((_yIntervals.length > 0) && (null != _yIntervals[0]))
				pw.print("\n       yIntsBegin='" + _yIntervals[0] + "'");
			if ((_yIntervals.length > 1) && (null != _yIntervals[1]))
				pw.print(" yIntsLength='" + _yIntervals[1] + "'");
			if ((_yIntervals.length > 2) && (null != _yIntervals[2]))
				pw.print(" yIntsUnit='" + _yIntervals[2] + "'");
		}

		pw.print(" pieBy='" + _pieBy + "'");

		// third parameter window
		if (_title != null)
			pw.print("\n       title='" + LyString.getXmlString(_title) + "'");

		if (_xTitle != null)
			pw.print(" xTitle='" + LyString.getXmlString(_xTitle) + "'");
		if (_yTitle != null)
			pw.print(" yTitle='" + LyString.getXmlString(_yTitle) + "'");

		if (null != _xAxis)
		{
			if ((_xAxis.length > 0) && (null != _xAxis[0]))
				pw.print("\n       xAxisMin='" + _xAxis[0] + "'");
			if ((_xAxis.length > 1) && (null != _xAxis[1]))
				pw.print(" xAxisMax='" + _xAxis[1] + "'");
			if ((_xAxis.length > 2) && (null != _xAxis[2]))
				pw.print(" xAxisWidth='" + _xAxis[2] + "'");
		}

		if (null != _yAxis)
		{
			if ((_yAxis.length > 0) && (null != _yAxis[0]))
				pw.print("\n       yAxisMin='" + _yAxis[0] + "'");
			if ((_yAxis.length > 1) && (null != _yAxis[1]))
				pw.print(" yAxisMax='" + _yAxis[1] + "'");
			if ((_yAxis.length > 2) && (null != _yAxis[2]))
				pw.print(" yAxisWidth='" + _yAxis[2] + "'");
		}

		pw.println();
		pw.print("      ");

		if (_xVerticalLabels)
			pw.print(" xVerticalLabels='true'");
		if (!_xGrid)
			pw.print(" xGrid='false'");
		if (!_yGrid)
			pw.print(" yGrid='false'");
		if (!_3dEffect)
			pw.print(" d3Effect='false'");
		if (!_showLegend)
			pw.print(" showLegend='false'");
		if (_showTips)
			pw.print(" showTips='true'");
		if (_ratios)
			pw.print(" ratios='true'");

		pw.println(">");

		// YFIELD tags
		for (int i = 0; i < _selectedYFieldsInfo.size(); i++)
		{
			pw.print("    <YFIELD");
			pw.print(" value='" + LyString.getXmlString(_selectedYFieldsInfo.elementAt(i))
				+ "'");
			pw.println("/>");
		}

		if (_preference != null)
			_preference.writeXml(pw);

		pw.println("</CHART>");
	} // end method writeXml

	/**
     * Sets the classInfo of this chart.
     */
    void setClassInfo(LyClassInfo classInfo)
    {
        _classInfoId = (classInfo == null ? null : classInfo.getId());
    }
} // end class LyChartInfo
