/*
 * LyChoiceFieldInfo.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;

import leon.app.*;
import leon.data.*;
import leon.misc.*;


/**
 * This class implements the description of an enumerated field (choice with finite number of
 * options) of a class info. Choice fields are defined with a set of options (LyChoiceOption).
 *
 * <P>Options are pairs of displayable values and internal value.</P>
 *
 * <p>Choice fieldInfo may also accept "multiple" values : several options may be selected
 * simultaneously.</p>
 *
 * <p>See also tag <CHOICE> of leon.dtd</p>
 *
 * @see LyChoiceOption
 */
public final class LyChoiceFieldInfo extends LyFieldInfo
{
	public static final String	    __version =
		"$Id: LyChoiceFieldInfo.java 40918 2012-09-25 11:53:52Z jlsaouli $";

	/** "Sort by" values. */
    public enum SortBy
    {
        INDEX,
        ID,
        NAME,
        VALUE
    }

	/** Choice types. */
    public enum Type
    {
        ENUM,
        BOOLEAN
    }

    /** Type of the choice field. */
    Type                            _type;

	/** Set of options available on this choiceFieldInfo. */
	LyInfoList<LyChoiceOption>	    _options;

	/** Set of options groups available on this choiceFieldInfo. */
	LyInfoList<LyChoiceOptionGroup> _optionGroups;

	/** Set of items (options and groups with value) available on this choiceFieldInfo. */
	LyInfoList<LyInfo>              _items;

	/** Indicates if this field info is multiple or not. */
	boolean						    _multiple;

	/**
	 * Number of columns which should be shown when displaying the choice (-1 if no preference is
	 * given, 0 as list).
	 */
	int							    _nCols;

	/** Gives the type of sorting performed on this field info. */
	SortBy						    _sortBy;

	/** Indicates if during user input, this field info is inverted or. */
	boolean						    _not;

	/** Cache for the choiceOption group. */
	LyChoiceOptionGroup			    _choiceOptionGroup;

	/** Context field for this choice. */
	LyChoiceFieldInfo			    _contextField;


	/**
	 * Initialization of choice attributes.
	 */
	protected void init()
	{
		super.init();

		_options = new LyInfoList<LyChoiceOption>();
		_optionGroups = new LyInfoList<LyChoiceOptionGroup>();
		_items = new LyInfoList<LyInfo>();
		_multiple = false;
		_nCols = -1;
		_sortBy = SortBy.INDEX;
		_choiceOptionGroup = null;
		_contextField = null;
		_not = false;

		// Always add translate mark on choice fields
        _marks = new BitSet();
        _marks.set(MARK_TRANSLATE);
	}

	/**
	 * Init this instance attributes from its template.
	 *
	 * @param parts Indicates which parts of the template are copied.
	 */
	void initFromTemplate(int parts)
	{
        parts &= ~_initParts;
        if (parts == 0)
            return;
		if (_template == null)
			return;

		super.initFromTemplate(parts);

		LyChoiceFieldInfo template = (LyChoiceFieldInfo)_template;

		if ((parts & PART_DATA) > 0)
		{
			if (template._options != null)
				_options = new LyInfoList<LyChoiceOption>(template._options);
			if (template._optionGroups != null)
				_optionGroups = new LyInfoList<LyChoiceOptionGroup>(template._optionGroups);
			if (template._items != null)
				_items = new LyInfoList<LyInfo>(template._items);
			_contextField = template._contextField;
			_not = template._not;
		}
		if ((parts & PART_GUI) > 0)
		{
			_nCols = template._nCols;
			_sortBy = template._sortBy;
		}
	} // end method initFromTemplate

	/**
	 * Constructor of a choice field.
	 *
	 * @param id       Identifier of the field.
	 * @param javaName Name of the generated java field.
	 * @param template Template used to construct this field.
	 * @param multiple Boolean indicating if this choice is multiple or not.
	 * @param sortBy   Type of sorting performed for this choice field (one of INDEX, LABEL, VALUE).
	 * @param type     Type of the choice, one of ENUM and BOOLEAN
	 */
	LyChoiceFieldInfo(String id, String javaName, LyChoiceFieldInfo template,
		boolean multiple, SortBy sortBy, Type type, boolean not)
	{
		this(id);
		construct(javaName, template, multiple, sortBy, type, not);
	}

	/**
	 * Constructor of a choice field.
	 *
	 * @param id       Identifier of the field.
	 * @param javaName Name of the generated java field.
	 * @param template Template used to construct this field.
	 * @param multiple Boolean indicating if this choice is multiple or not.
	 * @param sortBy   Type of sorting performed for this choice field (one of INDEX, LABEL, VALUE).
	 * @param type     Type of the choice, one of ENUM and BOOLEAN
	 */
	LyChoiceFieldInfo(String id)
	{
		super(id);
	}

	/**
	 * Construction of a choice field.
	 *
	 * @param javaName Name of the generated java field.
	 * @param template Template used to construct this field.
	 * @param multiple Boolean indicating if this choice is multiple or not.
	 * @param sortBy   Type of sorting performed for this choice field (one of INDEX, LABEL, VALUE).
	 * @param type     Type of the choice, one of ENUM and BOOLEAN
	 */
	protected void construct(String javaName, LyChoiceFieldInfo template,
		boolean multiple, SortBy sortBy, Type type, boolean not)
	{
		super.construct(javaName, template);

		_multiple = multiple;
		_sortBy = (sortBy == null ? SortBy.INDEX : sortBy);
		_type = (type == null ? Type.ENUM : type);
		_not = not;
	}

	/**
	 * Final step of parsing : check consistency. Checks performed are the following :
	 *
	 * <ul>
	 *   <LI>options list shouldn't be empty,</LI>
	 *   <LI>check unicity of internal value options within the options list.</LI>
	 * </ul>
	 *
	 * @param     parent Object info in which this field is checked.
	 * @exception IllegalStateException If one of these checks fails.
	 */
	void check(LyObjectInfo parent, int parts)
	{
		if ((parts & PART_DATA) > 0)
		{
			// Check non emptyness of options list
			if ((_options == null) || (_options.getSize() == 0))
				throw new IllegalStateException("BF AppEngine : choice " + this
					+ " contains no option");

			// Check unicity of internal values
			short		   i, j, n = getChoiceItemCount();
			LyChoiceItem   item, item2;
			LyChoiceOption option;
			Object		   value, value2;
			LyRule		   rule;
			int			   ndef = 0;
			Object		   defaultValue = null;
			LyObjectInfo   objectInfo;

			for (i = 0; i < n; i++)
			{
				item = getChoiceItem(i);
				value = item.getValue();

				if ((item instanceof LyChoiceOption)
						&& ((LyChoiceOption)item).isDefault())
				{
					if (getType() == Type.BOOLEAN)
					{
						if (i == 0)
							defaultValue = Boolean.FALSE;
						else
							defaultValue = Boolean.TRUE;
					}
					else if (!isMultiple())
					{
						if (defaultValue == null)
							defaultValue = new Short(i);
					}
					else
					{
						if (defaultValue == null)
							defaultValue = new BitSet();
						((BitSet)defaultValue).set(i);
					}

					ndef++;
				}

				for (j = (short)(i + 1); j < n; j++)
				{
					item2 = getChoiceItem(j);
					value2 = item2.getValue();
					if ((value == value2)
							|| value2.equals(value))
						throw new IllegalStateException("BF AppEngine : option value " + value
							+ " should be unique within choice " + this);
				}

				// check the rules of the option and add dependencies
				if (item instanceof LyChoiceOption)
				{
					option = (LyChoiceOption)item;
					for (j = 0; j < option.getRuleCount(); j++)
					{
						rule = option.getRule(j);
						rule.check(this);
						objectInfo = rule.getObjectInfo();
						if (objectInfo instanceof LyFieldInfo)
							addDepend((LyFieldInfo)objectInfo, rule.hasStrongDepend());
					}
				}
			}

			if (!isMultiple())
			{
	            LyEnvironment    environment = getEnvironment();

				if (ndef > 1)
				    environment.logWarning("BF AppEngine : the non multiple choice " + this
						+ " should have only one default option (" + ndef + " found)");
			}

			if ((getType() == Type.BOOLEAN) && (n != 2))
				throw new IllegalStateException(
					"BF AppEngine : a boolean choice must have two options and only two options.");

			// Sets default value
			setDefaultValue(defaultValue);

			if (_contextField != null)
			{
				_contextField.addDepend(this);
				LyChoiceFieldInfo context = _contextField;

				while (context.getContextField() != null)
				{
					context = context.getContextField();
					context.addDepend(this);
				}
			}

			// Check that all or none groups have a value
			LyInfoList<LyChoiceOptionGroup> groups = getChoiceOptionGroups(true);
			LyChoiceOptionGroup			    group, root;
			boolean						    withValue = false;
			Object						    groupValue;
			int							    g, ng;

			root = getChoiceOptionGroup();
			groups.removeInfo(root);
			ng = groups.getSize();
			for (g = 0; g < ng; g++)
			{
				group = groups.getInfo(g);
				groupValue = group.getValue();
				if ((g == 0) && (groupValue != null))
					withValue = true;
				else if ((withValue && (groupValue == null))
						|| (!withValue && (groupValue != null)))
					throw new IllegalStateException("BF AppEngine : all options groups of choice field "
						+ this
						+ " should have a value or none of them (except the root option group)");
			}
		}

		if ((parts & PART_GUI) > 0)
		{
			// Check sortBy
			if ((_sortBy != SortBy.INDEX)
					&& isDictionary())
			{
				_application.getEnvironment().logWarning(
					"BF AppEngine : sortBy not supported for choice field "
					+ this + " which is a dictionary.");
				setSortBy(SortBy.INDEX);
			}
		}

		// Generic field check is done after because default value is set
		// within this method
		super.check(parent, parts);
	} // end method check

	/**
	 * Lock this choice field.
	 */
	protected void lock()
	{
		super.lock();

		// Turn options list to unwritable
		_options.lock();
		_optionGroups.lock();
	}

	/* Implementation of methods of LyFieldInfo */

	/**
	 * Returns class of values for this field. The value class corresponding to a choice field is :
	 *
	 * <ul>
	 *   <LI>java.lang.Short, if the field is not multiple,</LI>
	 *   <LI>java.util.BitSet if it is multiple.</LI>
	 * </ul>
	 */
	public Class<?> getValueClass()
	{
		if (getType() == Type.BOOLEAN)
			return Boolean.class;
		else
			return (isMultiple() ? BitSet.class : Short.class);
	}

	/**
	 * Builds the appropriate LyValue corresponding to this field for the given value object.
	 *
	 * @see LyChoiceValue
	 * @see LyBooleanValue
	 */
	public LyValue buildFieldValue(Object value)
	{
		if (value == null)
		{
			if (getType() == Type.BOOLEAN)
				return new LyBooleanValue(this);
			else
				return new LyChoiceValue(this);
		}
		else if (value instanceof Short)
		{
			if (getType() == Type.BOOLEAN)
				return new LyBooleanValue(this, value.toString());
			return new LyChoiceValue(this, (Short)value);
		}
		else if (value instanceof BitSet)
			return new LyChoiceValue(this, (BitSet)value);
		else if (value instanceof String)
		{
			if (getType() == Type.BOOLEAN)
				return new LyBooleanValue(this, (String)value);
			else
				return new LyChoiceValue(this, (String)value);
		}
		else if (value instanceof Boolean)
			return new LyBooleanValue(this, (Boolean)value);
        else if (value.getClass().isEnum())
            return new LyChoiceValue(this, ((Enum<?>)value).ordinal());
		else
			return super.buildFieldValue(value);
	}

	/**
	 * Constructs a value for this fieldInfo from an object for a given instance. Supported
	 * conversions are the following :
	 *
	 * <TABLE>
	 *   <TR>
	 *     <TD>Short, String</TD>
	 *     <TD>Short</TD>
	 *   </TR>
	 *   <TR>
	 *     <TD>BitSet, Number, boolean[], String[]</TD>
	 *     <TD>BitSet</TD>
	 *   </TR>
	 * </TABLE>
	 *
	 * @exception LyUnsupportedFieldBindingException If the requested conversion if none of the
	 *                                               supported conversions.
	 */
	public Object getValueFromObject(Object object, LyObject instance)
	{
        Object value = decodePeerValue(object);

        if (value == null)
            return null;

	    Class<?>  valueClass = getValueClass();

		if (valueClass == Boolean.class)
		{
			if (value instanceof String)
				value = parse((String)value);
			else if (value instanceof Boolean)
			    return value;
			else
			{
				short index = getChoiceOptionIndex(value);

				if (index == (short)0)
					value = Boolean.FALSE;
				else
					value = Boolean.TRUE;
			}

			return value;
		}
		else if (valueClass == Short.class)
		{
            if (value.getClass().isEnum())
                return new Short((short)((Enum<?>)value).ordinal());

			short index = getChoiceOptionIndex(value);

			if (index >= 0)
				return new Short(index);
		}
		else if (valueClass == BitSet.class)
		{
			// If the value is a string, try to parse it
			if (value instanceof String)
				value = parse((String)value);

			if ((value == null)
			        || (value.getClass() == BitSet.class))
				return value;

			if (value instanceof Number)
			{
				long		 l = ((Number)value).longValue();
				short		 i, n = getChoiceItemCount();
				BitSet		 bs = new BitSet();
				LyChoiceItem choiceItem;
				short		 s;

				for (i = 0; i < n; i++)
				{
					choiceItem = getChoiceItem(i);
					s = choiceItem.getShortValue();
					if (s < 0)
					{
						throw new IllegalArgumentException("BF AppEngine : invalid choice option value "+choiceItem.getValue()+" can not be saved as a long value in LyChoiceFieldInfo : "+getId());
					}
					if ((l & (1L << s)) != 0)
						bs.set(i);
				}

				return bs;
			}
			else if (value instanceof boolean[])
			{
				boolean[] table = (boolean[])value;
				int		  i, l = table.length, n = getChoiceItemCount();
				BitSet    bs = new BitSet();

				for (i = 0; (i < n) && (i < l); i++)
					if (table[i])
						bs.set(i);

				return bs;
			}
			else if (value.getClass().isArray())
			{
				Object[] table = (Object[])value;
				int		 i, l = table.length, n = getChoiceItemCount();
				BitSet   bs = new BitSet();
				short    index;

				for (i = 0; (i < n) && (i < l); i++)
				{
					index = getChoiceOptionIndex(table[i]);
					if (index >= 0)
						bs.set(index);
				}

				return bs;
			}
		}

		return super.getValueFromObject(value, instance);
	} // end method getValueFromObject

	/**
	 * Returns an object of the given class from a value of this fieldInfo. Supported conversions
	 * are the following :
	 *
	 * <TABLE>
	 *   <TR>
	 *     <TD>Short</TD>
	 *     <TD>Number, String</TD>
	 *   </TR>
	 *   <TR>
	 *     <TD>BitSet</TD>
	 *     <TD>BitSet, Long, Integer, boolean[], String[], Object[]</TD>
	 *   </TR>
	 * </TABLE>
	 *
	 * @exception LyUnsupportedFieldBindingException If the requested conversion if none of the
	 *                                               supported conversions.
	 */
	public Object getPeerValue(Object value, Class<?> objectClass, LyObject instance)
	{
	    value = encodePeerValue(value, objectClass);

		if (value == null)
			return null;

		if (value instanceof BitSet)
		{
			if (objectClass == BitSet.class)
				return value;
			else if ((objectClass == Long.class)
					|| (objectClass == Integer.class))
			{
				BitSet		 bs = (BitSet)value;
				short		 i, n = (short)bs.size();
				long		 number = 0;
				LyChoiceItem choiceItem;
				short		 s;

				for (i = 0; i < n; i++)
				{
					if (bs.get(i))
					{
						choiceItem = getChoiceItem(i);
						s = choiceItem.getShortValue();
						if (s < 0)
						{
							throw new IllegalArgumentException("BF AppEngine : invalid choice option value "+choiceItem.getValue()+" can not be saved as a long value in LyChoiceFieldInfo : "+getId());
						}
						number |= (1L << s);
					}
				}

				if (objectClass == Long.class)
					return new Long(number);
				else
					return new Integer((int)number);
			}
			else if (objectClass == String.class)
			{
				BitSet		 bs = (BitSet)value;
				short		 i, n = (short)bs.size();

				if (n == 0)
					return null;
				StringBuffer sb = new StringBuffer();
				LyChoiceItem choiceItem;

				for (i = 0; i < n; i++)
				{
					if (bs.get(i))
					{
						choiceItem = getChoiceItem(i);
						if (sb.length() > 0)
							sb.append(LyApplication.VISIBLE_VALUE_SEP);
						sb.append(choiceItem.getValue());
					}
				}
				return sb.toString();

			}
			else if (objectClass.isArray()
					&& (objectClass.getComponentType() == boolean.class))
			{
				BitSet    bs = (BitSet)value;
				short     i, n = (short)bs.size();
				boolean[] table = new boolean[n];

				for (i = 0; i < n; i++)
					table[i] = bs.get(i);

				return table;
			}
			else if (objectClass.isArray()
					&& (objectClass.getComponentType() == String.class))
			{
				BitSet		   bs = (BitSet)value;
				short		   i, n = (short)bs.size();
				Vector<String> objects = new Vector<String>(n);

				for (i = 0; i < n; i++)
					if (bs.get(i))
						objects.addElement(getChoiceItem(i).getValue().toString());

				String[] table = new String[objects.size()];

				objects.copyInto(table);
				return table;
			}
			else if (objectClass.isArray())
			{
				BitSet		   bs = (BitSet)value;
				short		   i, n = (short)bs.size();
				Vector<Object> objects = new Vector<Object>(n);

				for (i = 0; i < n; i++)
					if (bs.get(i))
						objects.addElement(getChoiceItem(i).getValue());

				Object[] table = new Object[objects.size()];

				objects.copyInto(table);
				return table;
			}
		}
		else
		{
			String str;

			if (value instanceof Boolean)
			{
	            Boolean b = (Boolean)value;
				short   index = 0;

				if (b.booleanValue())
					index = 1;

				str = getChoiceItem(index).getValue().toString();
			}
			else
			{
				if (value instanceof String)
				{
				    String s = (String)value;

				    str = getChoiceItem(s).getValue().toString();
				}
				else
				{
				    Short s = (Short)value;

    				if ((s.intValue() < 0) && hasMark(LyFieldInfo.MARK_OPTIONAL))
    					return null;

    				str = getChoiceItem(s.shortValue()).getValue().toString();
				}
			}

			if (objectClass == String.class)
				return str;

			try
			{
				if (objectClass == Short.class)
					return new Short(str);
				if (objectClass == Integer.class)
					return new Integer(str);
				if (objectClass == Long.class)
					return new Long(str);
				if (objectClass == Float.class)
					return new Float(str);
				if (objectClass == Double.class)
					return new Double(str);
				if ((objectClass == Boolean.class)
                        || (objectClass == boolean.class))
					return Boolean.valueOf(str.equals("1")
							|| str.equalsIgnoreCase("true"));
                if (objectClass.isEnum())
                    try
                    {
                    	@SuppressWarnings({ "unchecked", "rawtypes" })
                    	Class<Enum> temp = (Class<Enum>)objectClass;
                    	@SuppressWarnings("unchecked")
                    	Object o = Enum.valueOf(temp, str);
                    	return o;
                    }
                    catch (Exception e)
                    {
                        return objectClass.getEnumConstants()[new Short(str)];
                    }
			}
			catch (Exception e)
			{
				throw new LyUnsupportedFieldBindingException("BF AppEngine : " + e.getMessage());
			}
		}

		return super.getPeerValue(value, objectClass, instance);
	} // end method getPeerValue

	/**
	 * Builds internal value for an instance of LyChoiceFieldInfo from a string (previously
	 * constructed by toString).
	 *
	 * @see LyChoiceFieldInfo#toString
	 */
	public Object parse(String stringValue)
		throws LyFieldInfoFormatException
	{
		short index;

		if (getType() == Type.BOOLEAN)
		{
			if (stringValue == null)
				return null;

			index = getChoiceOptionIndex(stringValue);
			if (index == 0)
				return Boolean.FALSE;
			else if (index == 1)
				return Boolean.TRUE;
			else
				return null;
		}
		// Multiple ?
		else if (isMultiple())
		{
			if ((stringValue == null) || (stringValue.length() == 0))
				return null;

			String[] values;

			if (stringValue.indexOf(_application.VALUE_SEP) >= 0)
				values = LyString.tokenizer(stringValue, _application.VALUE_SEP);
			else
				values = LyString.tokenizer(stringValue, LyApplication.VISIBLE_VALUE_SEP);

			int     i, n = values.length;
			BitSet  bs = new BitSet();
			boolean set = false;

			for (i = 0; i < n; i++)
			{
				try
				{
					index = getChoiceOptionIndex(values[i]);
					if (index >= 0)
					{
						bs.set(index);
						set = true;
					}
				}
				catch (IllegalArgumentException e)
				{
					getEnvironment().writeException(e);
				}
			}

			// Have we found anything ?
			return (set ? bs : null);
		}

		if (stringValue == null)
			return null;

		// Check if the option does exists
		index = getChoiceOptionIndex(stringValue);
		if (index >= 0)
			return new Short(index);

		return null;
	} // end method parse

	/**
	 * Returns the specified value for this field as a "displayable" string.
	 *
	 * @param value the value to change as a displayable string.
	 * @param name  specifies if the value must be composed with name of the options (true) or
	 *              aliases (false)
	 */
	protected String toDisplayString(Object fieldValue, boolean toName)
	{
		return toDisplayString( fieldValue, toName, null);
	}

	/**
	 * Returns the specified value for this field as a "displayable" string.
	 * @param value the value to change as a displayable string.
	 * @param name  specifies if the value must be composed with name of the options (true) or
	 *              aliases (false)
	 */
	protected String toDisplayString(Object fieldValue, boolean toName, String separator)
	{
		if (fieldValue == null)
			return null;

		if (fieldValue.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ fieldValue.getClass() + " expected " + getValueClass());

		String str = null;

		if (getType() == Type.BOOLEAN)
		{
			Boolean b = (Boolean)fieldValue;
			short   index = 0;

			if (b.booleanValue())
				index = 1;
			if (!toName)
				str = getChoiceItem(index).getAlias();

			if (str == null)
				str = getChoiceItem(index).getName();

			return str;
		}
		else if (!isMultiple())
		{
			Short s = (Short)fieldValue;
			short ss = s.shortValue();

			if (ss < 0)
				return null;

			if (!toName)
				str = getChoiceItem(ss).getAlias();

			if (str == null)
				str = getChoiceItem(ss).getName();

			return str;
		}

		BitSet		  bs = (BitSet)fieldValue;
		short		  i, n = getChoiceItemCount();
		String		  name = null;

		// We must translate here the content before aggregate
		LyEnvironment env = getEnvironment();
		short[]		  order = getOrder();
		short		  position;

		for (i = 0; i < n; i++)
		{
			position = ((order == null) ? i : order[i]);
			if (bs.get(position))
			{
				str = null;
				if (!toName)
					str = getChoiceItem(position).getAlias();

				if (str == null)
					str = getChoiceItem(position).getName();

				if (name == null)
					name = env.translate(str);
				else
				{
					if(separator != null)
						name = name + separator + env.translate(str);
					else
						name = name + LyApplication.VISIBLE_FIELD_SEP + env.translate(str);
				}
			}
		}

		return name;
	} // end method toDisplayString

	/**
	 * Returns the specified value for this field as a "displayable" string.
	 */
	public String toName(Object fieldValue)
	{
		return toName(fieldValue, null);
	}

	/**
	 * Returns the specified value for this field as a "displayable" string.
	 *
	 * @param Object fieldValue : The fieldValue to display.
	 * @param String separator  : The value of separator between field values.
	 *
	 * @return String : Formated String.
	 */
	public String toName(Object fieldValue, String separator)
	{
		String name = getFormattedName(fieldValue);

		if (name != null)
			return name;

		return toDisplayString(fieldValue, true, separator);
	}

	/**
	 * Returns the specified value for this field as a short "displayable" string This method is
	 * used when presenting values in tables. In that case, values are presented using aliases of
	 * the options.
	 */
	public String toShortName(Object fieldValue)
	{
		String name = getFormattedShortName(fieldValue);

		if (name != null)
			return name;

		return toDisplayString(fieldValue, false);
	}

	/**
	 * Converts the specified value for this field to an "internal" string (string that enables
	 * reconstruction through parse method).
	 *
	 * @see LyChoiceFieldInfo#parse
	 */
	public String toString(Object fieldValue)
	{
		if (fieldValue == null)
			return null;

		if ((fieldValue.getClass() != getValueClass()) && !(fieldValue instanceof String))
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ fieldValue.getClass() + " expected " + getValueClass());

		String string = getFormattedString(fieldValue);

		if (string != null)
			return string;

		if (!isMultiple())
		{
			LyChoiceItem choiceItem = null;

			if (getType() == Type.BOOLEAN)
			{
				Boolean b = (Boolean)fieldValue;

				if (b != null)
				{
					if (b.booleanValue())
						choiceItem = getChoiceItem((short)1);
					else
						choiceItem = getChoiceItem((short)0);
				}
			}
			else if (fieldValue instanceof String)
				return fieldValue.toString();
			else
			{
				Short s = (Short)fieldValue;

				choiceItem = getChoiceItem(s.shortValue());
			}

			if (choiceItem != null)
				return choiceItem.getId();
			else
				return null;
		}

		BitSet bs = (BitSet)fieldValue;
		short  i, n = (short)bs.size();
		String str = null;

		for (i = 0; i < n; i++)
		{
			if (bs.get(i))
			{
				if (str == null)
					str = getChoiceItem(i).getId();
				else
					str = str + _application.VALUE_SEP + getChoiceItem(i).getId();
			}
		}

		return str;
	} // end method toString

	/**
	 * Checks the given value : the value must correspond to options of the choice.
	 */
	public String checkValue(LyEnvironment environment, Object fieldValue)
	{
		if (fieldValue == null)
			return super.checkValue(environment, fieldValue);

		if (fieldValue.getClass() != getValueClass())
			return environment.getMessage("LY_ER_FIELD_VALUE_CLASS", this, fieldValue.getClass(),
				getValueClass());

		if (!isMultiple())
		{
			if (getType() != Type.BOOLEAN)
			{
				Short s = (Short)fieldValue;
				short v = s.shortValue();

				if ((v < 0) || (v >= getChoiceItemCount()))
					return environment.getMessage("LY_ER_INVALID_OPTION_INDEX", this, s);
			}
		}
		else
		{
			BitSet bs = (BitSet)fieldValue;
			short  i, n = (short)bs.size();

			for (i = 0; i < n; i++)
			{
				if (bs.get(i)
						&& ((i < 0) || (i >= getChoiceItemCount())))
					return environment.getMessage("LY_ER_INVALID_OPTION_INDEX", this, new Short(i));
			}
		}

		return super.checkValue(environment, fieldValue);
	} // end method checkValue

	/**
	 * Compare two values of this fieldInfo.
	 */
	public int compare(Object value1, Object value2, BitSet modifiers)
	{
		if (value1 == value2)
			return 0; // identical.

		if (value1 == null)
		{
			// Null or empty bitset is the same
			if ((value2 instanceof BitSet) && (((BitSet)value2).length() <= 0))
				return 0;

			return -1;
		}
		if (value2 == null)
		{
			// Null or empty bitset is the same
			if ((value1 instanceof BitSet) && (((BitSet)value1).length() <= 0))
				return 0;

			return 1;
		}

		if (value1.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value1.getClass() + " expected " + getValueClass());
		if (value2.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value2.getClass() + " expected " + getValueClass());

		if (isMultiple() && !(getType() == Type.BOOLEAN))
		{
			short[] order = getOrder();
			short   i, j, position, position2;
			short   n = ((order == null) ? getChoiceItemCount() : (short)order.length);
			BitSet  bs1 = (BitSet)value1;
			BitSet  bs2 = (BitSet)value2;
			int     n1 = bs1.size(), n2 = bs2.size();
			boolean v1, v2;

			for (i = 0; i < n; i++)
			{
				position = ((order == null) ? i : order[i]);
				v1 = (position < n1) && bs1.get(position);
				v2 = (position < n2) && bs2.get(position);

				if (v1 && !v2)
				{
					for (j = (short)(i + 1); j < n; j++)
					{
						position2 = ((order == null) ? j : order[j]);
						if ((position2 < n2) && bs2.get(position2))
							return -(position + 1);
					}
					return (position + 1);
				}
				if (!v1 && v2)
				{
					for (j = (short)(i + 1); j < n; j++)
					{
						position2 = ((order == null) ? j : order[j]);
						if ((position2 < n1) && bs1.get(position2))
							return (position + 1);
					}
					return -(position + 1);
				}
			}

			return 0;
		}
		else
		{
			short index1, index2;

			if (getType() == Type.BOOLEAN)
			{
				Boolean boolean1 = (Boolean)value1;
				Boolean boolean2 = (Boolean)value2;

				index1 = (short)(boolean1.booleanValue() ? 1 : 0);
				index2 = (short)(boolean2.booleanValue() ? 1 : 0);
			}
			else
			{
				index1 = ((Short)value1).shortValue();
				index2 = ((Short)value2).shortValue();
			}
			if (_sortBy == SortBy.INDEX)
				return index1 - index2;

			LyChoiceItem item1 = getChoiceItem(index1);
			LyChoiceItem item2 = getChoiceItem(index2);
			int			 result = 0;

			if (_sortBy == SortBy.ID)
				result = item1.getId().compareTo(item2.getId());
			else if (_sortBy == SortBy.NAME)
			{
				LyEnvironment environment = getEnvironment();
				String		  name1 = environment.translate(item1.getName());
				String		  name2 = environment.translate(item2.getName());

				result = name1.compareTo(name2);
			}
			else if (_sortBy == SortBy.VALUE)
				result = item1.getValue().toString().compareTo(item2.getValue().toString());

			if (result == 0)
				return index1 - index2;
			else
				return result;
		}
	} // end method compare

	/**
	 * Returns the filter conditions authorized on this field. The values must match the short
	 * constants defined in LyFilterExpression.
	 *
	 * @return An array of conditions corresponding to the authorized conditions on this field.
	 * @see    leon.info.LyFilterExpression
	 */
	public LyFilterExpression.Condition[] getFilterConditions()
	{
		// If choice is multiple, only equals and contains are allowed.
		// If not, contains is not allowed.
		int nConditions = 1;

		if (getType() != Type.BOOLEAN)
			nConditions = (_multiple ? 2 : 5);

        LyFilterExpression.Condition[] conditions = new LyFilterExpression.Condition[nConditions];

		conditions[0] = LyFilterExpression.Condition.EQUALS;

		if (_multiple)
			conditions[1] = LyFilterExpression.Condition.CONTAINS;
		else if (getType() != Type.BOOLEAN)
		{
			conditions[1] = LyFilterExpression.Condition.LOWER_THAN;
			conditions[2] = LyFilterExpression.Condition.LOWER_OR_EQUALS;
			conditions[3] = LyFilterExpression.Condition.GREATER_THAN;
			conditions[4] = LyFilterExpression.Condition.GREATER_OR_EQUALS;
		}

		return conditions;
	}

	/**
	 * Checks if the second value is contained in the first one, for this field.
	 */
	public boolean contains(Object value1, Object value2, BitSet modifiers)
	{
		if (!_multiple)
			return super.contains(value1, value2, modifiers);

		if ((value1 == null) && (value2 == null))
			return true;
		else if (value1 == null)
			return false;
		else if (value2 == null)
			return true;

		if (value1.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value1.getClass() + " expected " + getValueClass());
		if (value2.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value2.getClass() + " expected " + getValueClass());

		BitSet bs1 = (BitSet)value1;
		BitSet bs2 = (BitSet)value2;
		int    i, n1 = bs1.size(), n2 = bs2.size();

		for (i = 0; i < n2; i++)
		{
			if (bs2.get(i)
					&& ((i >= n1) || !bs1.get(i)))
				return false;
		}

		return true;
	} // end method contains

    /**
     * Accessor to the choice type.
     *
     * @return Choice type.
     */
    public Type getType()
    {
        return _type;
    }

	/**
	 * Accessor to the multiple property.
	 *
	 * @return True if this field may have several values, false otherwise.
	 */
	public boolean isMultiple()
	{
		return _multiple;
	}

	/**
	 * Accessor to the not property.
	 *
	 * @return True if this field is inverted during user input (BOOLEAN only), false otherwise.
	 */
	public boolean isNot()
	{
		return _not;
	}

	/**
	 * Accessor to the type of sorting for this choice field.
	 *
	 * @return Type of sorting used for this choice field.
	 */
	public SortBy getSortBy()
	{
		return _sortBy;
	}

	/**
	 * Gets an option from the option list at a given position.
	 *
	 * @param  index Index of the searched option.
	 * @return The option found, null otherwise.
	 */
	public LyChoiceOption getChoiceOption(short index)
	{
		if (index < 0)
			return null;

		load();
		LyChoiceItem item = ((index >= 0) ? getChoiceItem(index) : null);

		if ((item == null) || !(item instanceof LyChoiceOption))
			return null;
		return (LyChoiceOption)item;
	}

	/**
	 * Gets index of given option in the options list.
	 *
	 * @param  option Searched option.
	 * @return Index of the given option if found, -1 otherwise.
	 */
	public short getIndex(LyChoiceOption option)
	{
		return (short)_options.getIndex(option);
	}

	/**
	 * Gets options count.
	 *
	 * @return Return the number of options in this choice.
	 */
	public short getChoiceOptionCount()
	{
		load();
		return (short)_options.getSize();
	}

	/**
	 * Returns the list of options available on this choice field.
	 *
	 * @return List of options of this choice field.
	 */
	public LyInfoList<LyChoiceOption> getChoiceOptions()
	{
		load();
		return _options;
	}

	/**
	 * Returns the sorted list of options available on this choice field.
	 *
	 * @param  sorted Boolean indicating if list must be sorted according to sortBy criterion.
	 * @return Sorted list of options of this choice field.
	 */
	public LyInfoList<LyChoiceOption> getChoiceOptions(boolean sorted)
	{
		if (!sorted
				|| (_sortBy == SortBy.INDEX))
			return getChoiceOptions();

		LyInfoList<LyChoiceOption> options = new LyInfoList<LyChoiceOption>();
		short[]					   order = getOrder();
		short					   i, n = (short)order.length;
		LyChoiceItem			   item;

		for (i = 0; i < n; i++)
		{
			item = getChoiceItem(order[i]);
			if (item instanceof LyChoiceOption)
				options.addInfo((LyChoiceOption)item);
		}

		return options;
	}

	/**
	 * Accessor to the choice option group associated with this choice field.
	 *
	 * @return the choiceOptionGroup
	 */
	public LyChoiceOptionGroup getChoiceOptionGroup()
	{
		if (_choiceOptionGroup != null)
			return _choiceOptionGroup;

		short			    n = getChoiceOptionCount();
		LyChoiceOption	    option;
		LyChoiceOptionGroup group;

		// first get the top level group
		if (n <= 0)
			return null;

		// Get the first option
		option = getChoiceOption((short)0);

		if (option.getParent() instanceof LyChoiceOptionGroup)
		{
    		// Find the top group for this option
    		group = (LyChoiceOptionGroup)option.getParent();
    		while (group.getParent() instanceof LyChoiceOptionGroup)
    			group = (LyChoiceOptionGroup)group.getParent();

    		_choiceOptionGroup = group;
    	    return group;
		}
		else
		    return null;
	}

	/**
	 * Get all groups where the given option is found.
	 */
	public LyInfoList<LyChoiceOptionGroup> getChoiceOptionGroups(LyChoiceOption option)
	{
		LyInfoList<LyChoiceOptionGroup> groups = new LyInfoList<LyChoiceOptionGroup>();
		LyInfoList<LyChoiceOptionGroup> subGroups = getChoiceOptionGroups(true);
		int							    i, n = ((subGroups == null) ? 0 : subGroups.getSize());
		LyChoiceOptionGroup			    group;

		for (i = 0; i < n; i++)
		{
			group = subGroups.getInfo(i);
			if (group.contains(option, false))
				groups.addInfo(group);
		}

		return groups;
	}

	// Order is kept for optimization when sortBy is not on index
	private short[] _order = null;

	/**
	 * Returns the list of options index sorted on sortBY criterion.
	 */
	protected short[] getOrder()
	{
		if (_sortBy == SortBy.INDEX)
			return null;

		if (_order != null)
			return _order;

		short		 i, j, n = getChoiceItemCount(), temp;
		LyChoiceItem item, item2;
		String		 str, str2;

		_order = new short[n];
		for (i = 0; i < n; i++)
			_order[i] = i;

		LyEnvironment environment = getEnvironment();

		for (i = 0; i < n; i++)
		{
			item = getChoiceItem(_order[i]);
			if (_sortBy == SortBy.ID)
				str = item.getId();
			else if (_sortBy == SortBy.NAME)
				str = environment.translate(item.getName());
			else // SORT_BY_VALUE
				str = item.getValue().toString();

			for (j = (short)(i + 1); j < n; j++)
			{
				item2 = getChoiceItem(_order[j]);
				if (_sortBy == SortBy.ID)
					str2 = item2.getId();
				else if (_sortBy == SortBy.NAME)
					str2 = environment.translate(item2.getName());
				else // SORT_BY_VALUE
					str2 = item2.getValue().toString();

				if (str.compareTo(str2) > 0)
				{
					temp = _order[i];
					_order[i] = _order[j];
					_order[j] = temp;
					str = str2;
				}
			}
		}

		return _order;
	} // end method getOrder

	/**
	 * Gets an item from the items list at a given position.
	 *
	 * @param  index Index of the searched item.
	 * @return The item found, null otherwise.
	 */
	public LyChoiceItem getChoiceItem(short index)
	{
		if (index < 0)
			return null;

		load();
		return (LyChoiceItem)_items.getInfo(index);
	}

	/**
	 * Gets index of given item in the items list.
	 *
	 * @param  item Searched item.
	 * @return Index of the given item if found, -1 otherwise.
	 */
	public short getIndex(LyChoiceItem item)
	{
		return (short)_items.getIndex((LyInfo)item);
	}

	/**
	 * Gets items count.
	 *
	 * @return Return the number of items in this choice.
	 */
	public short getChoiceItemCount()
	{
		load();
		return (short)_items.getSize();
	}

	/**
	 * Gets an item from the items list by its id, name or value. Search is made by item id, then by
	 * item name and finally using the internal value converted to a string.
	 *
	 * @param  string Searched string (id, name or value transformed in String).
	 * @return The item found, or null if it could not be found.
	 */
	public LyChoiceOption getChoiceOption(String string)
	{
		LyChoiceItem choiceItem = getChoiceItem(string);

		if (choiceItem instanceof LyChoiceOption)
			return (LyChoiceOption)choiceItem;
		else
			return null;
	}

	/**
	 * Gets an item from the items list by its id, name or value. Search is made by item id, then by
	 * item name and finally using the internal value converted to a string.
	 *
	 * @param  string Searched string (id, name or value transformed in String).
	 * @return The item found, or null if it could not be found.
	 */
	public LyChoiceItem getChoiceItem(String string)
	{
		if (string == null)
			throw new IllegalArgumentException("BF AppEngine : string parameter can't be null");

		// use internal value to speed up compare
		string = string.intern();

		LyChoiceItem item;
		int			 i, n = getChoiceItemCount();

		// Seek the string through the option list
		for (i = 0; i < n; i++)
		{
			item = (LyChoiceItem)_items.getInfo(i);
			if ((item.getId() == string)
					|| (item.getName() == string)
					|| (item.getValue().toString().equals(string)))
				return item;
		}

		// Not found
		return null;
	}

	/**
	 * Gets an option from the option list from the internal value.
	 *
	 * @param  value Internal value from which an option is searched.
	 * @return Option corresponding to the given value if found, or null if it could not be found.
	 */
	public LyChoiceOption getChoiceOptionByValue(Object value)
	{
		short		 index = getChoiceOptionIndex(value);

		LyChoiceItem item = ((index >= 0) ? getChoiceItem(index) : null);

		if ((item == null) || !(item instanceof LyChoiceOption))
			return null;

		return (LyChoiceOption)item;
	}

	/**
	 * Gets the internal index (position) of an option from its internal value, its id or its name.
	 *
	 * @param     value Value (or id, or name) from which an option is searched.
	 * @return    Index of the option corresponding to the given value (or id, or name).
	 * @exception IllegalArgumentException if no option corresponds to the given value (or id, or
	 *                                     name).
	 */
	public short getChoiceOptionIndex(Object value)
	{
		if (value instanceof LyChoiceOption)
			return getIndex((LyChoiceOption)value);
		if (value instanceof LyChoiceItem)
			return getIndex((LyChoiceItem)value);
		
		LyChoiceItem item;
		short		 i, n = getChoiceItemCount();

		if (value instanceof Number)
		{
			// Optimization when value is a number
			short v = ((Number)value).shortValue();

			for (i = 0; i < n; i++)
			{
				item = getChoiceItem(i);
				if (item.getShortValue() == v)
					return i;
			}
			if ((v > 0) && (v <= n))
			{
				// FFT0044 V3.3 : could be an index, but it is could also be an unknown value ...
				// Should try unknown values first but kept for compatibility
				if (value instanceof Short)
					return ((Short)value).shortValue();
			}
		}

		// Use internal value to speed up compare
		if (value != null)
			value = value.toString().intern();

		// Seek the value through the option values
		for (i = 0; i < n; i++)
		{
			item = getChoiceItem(i);
			if (item.getValue() == value)
				return i;
		}

		// Seek the value through the option ids
		for (i = 0; i < n; i++)
		{
			item = getChoiceItem(i);
			if (item.getId() == value)
				return i;
		}

		// Seek the value through the option names
		for (i = 0; i < n; i++)
		{
			item = getChoiceItem(i);
			if (item.getName() == value)
				return i;
		}

		// Not found : do we have an unknown option ?
		LyChoiceOption option;

		for (i = 0; i < n; i++)
		{
			option = getChoiceOption(i);
			if (option.isUnknown())
			{
				_application.getEnvironment().logWarning("BF AppEngine : unknown value or id or name \""
					+ value + "\" found on choice " + this);
				return i;
			}
		}

		if (value instanceof String)
		{
    		// Last try : check if given value could be the index
		    short   index = LyTools.shortFromString((String)value);

		    if ((index >= 0)
		            && (index < n))
		        return index;
		}

		throw new IllegalArgumentException("BF AppEngine : no option with value or id or name \""
			+ value + "\" found on choice " + this);
	} // end method getChoiceOptionIndex

	/**
	 * Gets the preferred number of columns shown to the user at the same time.
	 *
	 * @return Number of columns, or -1 if no preference is given.
	 */
	public int getNCols()
	{
		return _nCols;
	}

	/**
	 * Gets the context field for this choice field info.
	 *
	 * @return the context field of this choice field info, null if none
	 */
	public LyChoiceFieldInfo getContextField()
	{
		return _contextField;
	}

	/**
	 * Utility method to retrieve options with table attribute set.
	 *
	 * @return List of options shown in tables (all if no option is marked).
	 */
	public LyInfoList<LyChoiceOption> getTableOptions()
	{
		LyInfoList<LyChoiceOption> options = null;
		short					   i, n = getChoiceOptionCount();
		LyChoiceOption			   option;

		for (i = 0; i < n; i++)
		{
			option = getChoiceOption(i);
			if (option.hasTableMark())
			{
				if (options == null)
					options = new LyInfoList<LyChoiceOption>();

				options.addInfo(option);
			}
		}

		return ((options == null) ? _options : options);
	}


	void setChoiceOptionGroups(LyInfoList<LyChoiceOptionGroup> groups)
	{
		_optionGroups = new LyInfoList<LyChoiceOptionGroup>();
		int i, n = (groups == null ? 0 : groups.getSize());
		LyChoiceOptionGroup group;
		for (i = 0; i < n; i++)
		{
			group = groups.getInfo(i);
			addChoiceOptionGroup(group);
		}
	}

	public LyInfoList<LyChoiceOptionGroup> getChoiceOptionGroups()
	{
		return _optionGroups;
	}

	/**
	 * Utility method to retrieve options group of this choice field.
	 *
	 * @return List of options group.
	 */
	public LyInfoList<LyChoiceOptionGroup> getChoiceOptionGroups(boolean godown)
	{
		if (!godown)
			return _optionGroups;

		LyInfoList<LyChoiceOptionGroup> groups = new LyInfoList<LyChoiceOptionGroup>();
		int							    i, n = _optionGroups.getSize();
		LyChoiceOptionGroup			    group;

		for (i = 0; i < n; i++)
		{
			group = _optionGroups.getInfo(i);
			groups.addInfo(group);
			groups.addList(group.getChoiceOptionGroups(true));
		}

		return groups;
	}

	/**
	 * Compute if the choice is a dictionary or not.
	 *
	 * @param  choice the choice
	 * @return true if the choice is a dictionary
	 */
	public boolean isDictionary()
	{
		short			    i, n = getChoiceOptionCount();
		LyChoiceOption	    option;
		LyChoiceOptionGroup group;

		for (i = 0; i < n; i++)
		{
			option = getChoiceOption(i);
			if (option.getParent() instanceof LyChoiceOptionGroup)
			{
				group = (LyChoiceOptionGroup)option.getParent();
				if (group.getParent() instanceof LyChoiceOptionGroup)
					return true;
			}
		}

		return false;
	}

	/**
	 * Checks whether options groups have value or not.
	 */
	public boolean doGroupsHaveValue()
	{
		LyInfoList<LyChoiceOptionGroup> groups = getChoiceOptionGroups(true);
		LyChoiceOptionGroup			    root = getChoiceOptionGroup();

		if ((groups != null) && (groups.getSize() > 1))
		{
			groups.removeInfo(root);

			LyChoiceOptionGroup group = groups.getInfo(0);

			if (group.getValue() != null)
				return true;
		}

		return false;
	}

	/**
	 * Adds an option to this choice field at a given position.
	 *
	 * @param option   Option to add this choice.
	 * @param position Position where the option is added in the options list.
	 */
	void addChoiceOption(LyChoiceOption option, int position)
	{
		_options.addInfo(option, position);
		_items.addInfo(option, position);
	}

	/**
	 * Adds an options group to this choice field.
	 *
	 * @param group Options group to add this choice.
	 */
	void addChoiceOptionGroup(LyChoiceOptionGroup group)
	{
		_optionGroups.addInfo(group);
		addChoiceItems(group);
	}

	/**
	 * Adds items of the given options group to this choice field.
	 *
	 * @param group Options group to add this choice.
	 */
	void addChoiceItems(LyChoiceOptionGroup group)
	{
		if (group.getValue() != null)
			_items.addInfo(group);

		short		 i, n = group.getChoiceItemCount();
		LyChoiceItem item;

		for (i = 0; i < n; i++)
		{
			item = group.getChoiceItem(i);
			if (item instanceof LyChoiceOptionGroup)
				addChoiceItems((LyChoiceOptionGroup)item);
		}
	}

	/**
	 * Removes an option from this choice field.
	 *
	 * @param option Option to remove from this choice.
	 */
	void removeChoiceOption(LyChoiceOption option)
	{
		_options.removeInfo(option);
	}

	/**
	 * Resets options list (clear all options).
	 */
	void resetChoiceOptions()
	{
		_options.resetInfos();
		_optionGroups.resetInfos();
		_items.resetInfos();
	}

	/**
	 * Sets the prefered number of columns shown to the user at the same time.
	 *
	 * @param nCols Number of columns, or -1 if no preference.
	 */
	void setNCols(int nCols)
	{
		if(nCols < -1)
			_nCols = -1;
		else
			_nCols = nCols;
	}

	/**
	 * Sets type of sorting.
	 *
	 * @param sortBy Type of sorting performed for this choice field (one of INDEX, LABEL, VALUE).
	 */
	void setSortBy(SortBy sortBy)
	{
		_sortBy = sortBy;
	}

	/**
	 * Set the context field associated to this choice field info.
	 *
	 * @param contextField the context field associated to this choice field
	 */
	void setContextField(LyChoiceFieldInfo contextField)
	{
		_contextField = contextField;
	}

	/**
	 * Modifier for ot property.
	 *
	 * @param not True if this field is inverted during user input (BOOLEAN only), false otherwise.
	 */
	void setNot(boolean not)
	{
		_not = not;
	}

    /**
     * Sets the choice type.
     *
     * @param type Choice type.
     */
    protected void setType(Type type)
    {
        _type = type;
    }

    /**
     * Sets the multiple property.
     *
     * @param multiple True if this field may have several values, false otherwise.
     */
    protected void setMultiple(boolean multiple)
    {
        _multiple = multiple;
    }

    /**
     * Sets if the choice is a dictionary or not.
     *
     * @param multiple True if the choice is a dictionary, false otherwise.
     */
    protected void setDictionary(boolean dictionary)
    {
    }

    /**
     * Sets the list of options available on this choice field.
     *
     * @param options List of options of this choice field.
     */
    protected void setChoiceOptions(LyInfoList<LyChoiceOption> options)
    {
        _options = new LyInfoList<LyChoiceOption>();
        _items = new LyInfoList<LyInfo>();

        int     i, n = (options == null ? 0 : options.getSize());

        for (i = 0; i < n; i++)
            addChoiceOption(options.getInfo(i), -1);
    }

    /**
     * Sets the choice option group associated with this choice field.
     *
     * @param group the choiceOptionGroup
     */
    protected void setChoiceOptionGroup(LyChoiceOptionGroup group)
    {
         _choiceOptionGroup = group;
    }
} // end class LyChoiceFieldInfo
