/*
 * LyChoiceOption.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;


/**
 * This class implements the description of an option stored in an instance of LyChoiceFieldInfo.
 * <br>
 * One option is a possible choice in an enumerated fieldInfo.<br>
 * An option has the following attributes :
 *
 * <ul>
 *   <LI>a name used to represent this choice when displayed to the end-user,</LI>
 *   <LI>an internal value that is given to the data provider when this option is chosen,</LI>
 *   <LI>a boolean that specifies if the option is a default value or not That means : checked at
 *     creation. Several options can be checked in the same time if the option that contains them is
 *     multiple, otherwise only the first of the default options is checked.</LI>
 *   <LI>a list of rules.<br>
 *     Rules are triggered when options are checked or unchecked. They allow to change field states
 *     whithin the same view.<br>
 *     See also tag of leon.dtd</LI>
 * </ul>
 *
 * @see LyChoiceFieldInfo
 * @see LyRule
 */
public final class LyChoiceOption extends LyInfo
	implements LyChoiceItem
{
	public static final String __version =
		"$Id: LyChoiceOption.java 40970 2012-09-28 16:32:05Z jlsaouli $";

	/**
	 * The displayed value (may be shown to the user) of this option If multiple languages are used
	 * at the same time, this is the untranslated key.
	 */
	String					   _name;

	/** Short name used as column header in edit table mode. */
	String					   _alias;

	/** Internal value as transmitted to the data provider. */
	Object					   _value;

	/** To specify if this option is a default value or not. */
	boolean					   _default;

	/** Set if this option is disable. */
	boolean					   _disable;

	/** Set if this option used to redirect unknown values. */
	boolean					   _unknown;

	/** Indicates if this option is shown as column in edit tables. */
	boolean					   _table;

	/** Rules associated to this option. */
	LyInfoList<LyRule>		   _rules;

	/**
	 * Internal representation of the value used to speed up processings when internal values are
	 * shorts.
	 */
	short					   _shortValue;

	/** List of labels associated to this option. */
	LyInfoLabel[]			   _labels;

	/** List of user labels associated to this option. */
	Vector<LyInfoLabel>		   _userLabels;

	/**
	 * Initialisation of instance attributes.
	 */
	protected void init()
	{
		_name = null;
		_alias = null;
		_value = null;
		_default = false;
		_disable = false;
		_unknown = false;
		_table = false;
		_rules = null;
		_shortValue = -1;
		_labels = null;
		_userLabels = null;
	}

    /**
     * Constructor of an option.
     *
     * @param id      Id of this option. Used as identifier within data description.
     */
    LyChoiceOption(String id)
    {
        super(id);
    }
    
	/**
	 * Constructor of an option.
	 *
	 * @param id      Id of this option. Used as identifier within data description.
	 * @param name    Name of the option shown to the end-user.
	 * @param alias   Short name of the option shown to the end-user.
	 * @param value   Internal value transmitted to the data provider when this option is chosen.
	 * @param def     Boolean indicating if this option is a default option or not.
	 * @param disable Boolean indicating if this option is a disable or not.
	 * @param unknown Boolean indicating if this option is used to redirect unknown values.
	 * @param table   Boolean indicating if this option is shown as column in edit tables.
	 */
	LyChoiceOption(String id, String name, String alias, Object value,
		boolean def, boolean disable, boolean unknown, boolean table)
	{
		super(id);

		init();
		
		construct(name, alias, value, def, disable, unknown, table);
	}
	
	/**
	 * Construction of a choice option.
     * @param name    Name of the option shown to the end-user.
	 * @param alias   Short name of the option shown to the end-user.
	 * @param value   Internal value transmitted to the data provider when this option is chosen.
	 * @param def     Boolean indicating if this option is a default option or not.
	 * @param disable Boolean indicating if this option is a disable or not.
	 * @param unknown Boolean indicating if this option is used to redirect unknown values.
	 * @param table   Boolean indicating if this option is shown as column in edit tables.
	 */
	protected void construct(String name, String alias, Object value,
			boolean def, boolean disable, boolean unknown, boolean table)
	{
		_name = ((name == null) ? null : name.intern());
		_alias = ((alias == null) ? null : alias.intern());
		setValue(value);
		_default = def;
		_disable = disable;
		_unknown = unknown;
		_table = table;
	}
	
	/**
	 * Accessor to the name of this option.
	 *
	 * @return Name of this option.
	 */
	public String getName()
	{
		if (_name == null)
			return getId();

		return _name;
	}

	/**
	 * Accessor to the name of this option.
	 *
	 * @param name Name of this option.
	 */
	void setName(String name)
	{
		_name = name;
	}

    /**
     * Accessor for alias (shortname) of this instance.
     *
     * @return Alias of this instance. If the alias is null, this method returns the name.
     */
    public final String getAlias()
    {
        if (_alias == null)
            return getName();

        return _alias;
    }

    /**
     * Accessor for alias (shortname) of this instance. This method doesn't compute a value if it is
     * null
     *
     * @return Alias of this instance.
     */
    public final String getInternalAlias()
    {
        return _alias;
    }
    
	/**
     * Sets the alias of the instance.
     *
     * @param alias Alias of the instance.
     */
    final void setInternalAlias(String alias)
    {
        _alias = ((alias == null) ? null : alias.intern());
    }

	/**
	 * Accessor to the short name of this option.
	 *
	 * @param alias Short name of this option.
	 */
	void setAlias(String alias)
	{
		_alias = alias;
	}

	/**
	 * Gets the internal value (value transmitted to the data provider when this option is chosen).
	 *
	 * @return Internal value corresponding to this option.
	 */
	public Object getValue()
	{
		if (_value == null)
			return getId();

		return _value;
	}

	/**
	 * Tests is this option is a default one.
	 *
	 * @return True if this option is a default option, false otherwise.
	 */
	public boolean isDefault()
	{
		return _default;
	}

	/**
	 * Sets if this option is the default one.
	 *
	 * @param def True if this option is the default one, false otherwise.
	 */
	void setDefault(boolean def)
	{
		_default = def;
	}

	/**
	 * Sets if this option is disabled or not.
	 *
	 * @return True if this option is disable, false otherwise.
	 */
	public boolean isDisable()
	{
		return _disable;
	}

	/**
	 * Tests is this option is disabled or not.
	 *
	 * @param disable True if this option is disabled, false otherwise.
	 */
	void setDisable(boolean disable)
	{
		_disable = disable;
	}

	/**
	 * Tests is this option is used to redirect unknown values when reading datas.
	 *
	 * @return True if this option is used for unknown values, false otherwise.
	 */
	public boolean isUnknown()
	{
		return _unknown;
	}

	/**
	 * Use this option to redirect unknown values when reading datas.
	 *
	 * @param unknown True if this option is used for unknown values, false otherwise.
	 */
	void setUnknown(boolean unknown)
	{
		_unknown = unknown;
	}

	/**
	 * Tests is this option is shown as column in edit tables.
	 *
	 * @return True if this option is shown as column in edit tables, false otherwise.
	 */
	public boolean hasTableMark()
	{
		return _table;
	}

	/**
	 * Tests is this option is shown as column in edit tables.
	 *
	 * @param tableMark True if this option is shown as column in edit tables, false otherwise.
	 */
	void setTableMark(boolean tableMark)
	{
		_table = tableMark;
	}
	
	/**
	 * Do nothing
	 */
	void setRule(boolean rule)
	{
		// Nothing to do. This method has to exist, because the attribute 'option_rule' of class 'Option_cls' 
		// need a set method
	}
	
	
	/**
	 * Tests is this option has one or more rules
	 *
	 * @return True if this option has one or more rules, false otherwise.
	 */
	public boolean hasRule()
	{
		return _rules != null && _rules.getSize() > 0;
	}

	/**
	 * Gets a rule from the rules list at a given position.
	 *
	 * @param  index Position of the searched rule. If index is negative (-1) return null
	 * @return Rule at the given index if it exists.
	 */
	public LyRule getRule(short index)
	{
		if (_rules == null)
			return null;

		return _rules.getInfo(index);
	}

	/**
	 * Gets rules count.
	 *
	 * @return Return the number of rules in this option.
	 */
	public short getRuleCount()
	{
		return ((_rules == null) ? 0 : (short)_rules.getSize());
	}

	/**
	 * Returns the list of rules defined on this option;
	 *
	 * @return List of rules, or null if no rule was registered on this option.
	 */
	public LyInfoList<LyRule> getRules()
	{
		if (_rules == null)
			return null;

		if (_rules.isWritable())
			_rules.lock();

		return _rules;
	}

	/**
	 * Gets the short value corresponding to the internal value. Used to optimize processings when
	 * internal values are shorts.
	 *
	 * @return Short value corresponding to this option, 1 if value can't be casted to a short.
	 */
	public short getShortValue()
	{
		return _shortValue;
	}

	/**
	 * Adds a rule on this option.
	 *
	 * @param rule Rule to add this choice.
	 */
	void addRule(LyRule rule)
	{
		if (_rules == null)
			_rules = new LyInfoList<LyRule>(false);

		_rules.addInfo(rule);
	}

	/**
	 * Completes option with given parameter.
	 *
	 * @param name    Name of the option shown to the end-user.
	 * @param alias   Short name of the option shown to the end-user.
	 * @param disable Boolean indicating if this option is a disable or not.
	 * @param unknown Boolean indicating if this option is used to redirect unkown values.
	 * @param table   Boolean indicating if this option is shown as column in edit tables.
	 */
	void completeOption(String name, String alias, boolean disable, boolean table)
	{
		if (name != null)
			_name = name.intern();
		if (alias != null)
			_alias = alias.intern();
		_disable = disable;
		_table = table;
	}

	/**
	 * Sets option value.
	 *
	 * @param value Internal value tranmitted to the data provider when this option is chosen.
	 */
	void setValue(Object value)
	{
		if (value instanceof String)
			value = ((String)value).intern();

		_value = value;

		if (_value != null)
		{
			// Try to convert it to a short when possible to speed up future processings
			try
			{
				_shortValue = Short.parseShort(_value.toString());
			}
			catch (NumberFormatException e)
			{
				_shortValue = -1;
			}
		}
	}

	/**
	 * Add a new label to labels list of this option.
	 *
	 * @param label Label to add.
	 */
	void addLabel(LyInfoLabel label)
	{
		if (label == null)
			throw new IllegalArgumentException("BF AppEngine : label can't be null");

        LyInfoLabel.Type type = label.getType();

		if (type == LyInfoLabel.Type.USER)
		{
			if (_userLabels == null)
				_userLabels = new Vector<LyInfoLabel>(1);
			_userLabels.add(label);
			return;
		}

		if ((type != LyInfoLabel.Type.NAME)
				&& (type != LyInfoLabel.Type.GRAMMATICAL)
				&& (type != LyInfoLabel.Type.SEMANTIC)
				&& (type != LyInfoLabel.Type.VALUE)
				&& (type != LyInfoLabel.Type.SHORT_VALUE)
				&& (type != LyInfoLabel.Type.USER))
			throw new IllegalArgumentException("BF AppEngine : invalid type of label on fieldInfo "
				+ this);

		if (_labels == null)
			_labels = new LyInfoLabel[7];
		_labels[type.ordinal()] = label;
	} // end method addLabel

	/**
	 * Returns the number of labels of this option.
	 */
	public int getLabelCount()
	{
		return ((_labels == null) ? 0 : _labels.length);
	}

	/**
	 * Returns the the label of this option for the specified type.
	 */
	public LyInfoLabel getLabel(LyInfoLabel.Type type)
	{
		if (_labels == null)
			return null;

		return _labels[type.ordinal()];
	}

    /**
     * Returns the label of this option at the given index.
     */
    public LyInfoLabel getLabel(int index)
    {
        if ((_labels == null)
                || (index < 0)
                || (index > _labels.length))
            return null;

        return _labels[index];
    }
    
	/**
	 * Returns the number of labels of this option.
	 */
	public int getUserLabelCount()
	{
		return ((_userLabels == null) ? 0 : _userLabels.size());
	}

	/**
	 * Returns the user label at the given position.
	 */
	public LyInfoLabel getUserLabel(int i)
	{
		return ((_userLabels == null) ? null : (LyInfoLabel)_userLabels.elementAt(i));
	}

	/**
	 * Returns the user label for the given subType.
	 */
	public LyInfoLabel getUserLabel(String subType)
	{
		int		    i, n = getUserLabelCount();
		LyInfoLabel label;

		if (subType != null)
			subType = subType.intern();
		for (i = 0; i < n; i++)
		{
			label = _userLabels.elementAt(i);
			if (label.getSubType() != subType)
				return label;
		}

		return null;
	}
	
    /**
     * Sets the list of rules defined on this class info.
     */
    protected void setRules(LyInfoList<LyRule> rules)
    {
        int     i, n = (rules == null ? 0 : rules.getSize());
        
        _rules = null;
        for (i = 0; i < n; i++)
            addRule(rules.getInfo(i));
    }
} // end class LyChoiceOption
