/*
 * LyChoiceOptionGroup.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;


/**
 * This class implements the description of a group of choice options. A choice options group
 * contains choice options that are either choice options or other groups.
 */
public final class LyChoiceOptionGroup extends LyInfo
	implements LyChoiceItem
{
	public static final String __version =
		"$Id: LyChoiceOptionGroup.java 40970 2012-09-28 16:32:05Z jlsaouli $";

	/**
	 * The displayed value (may be shown to the user) of this options group If multiple languages
	 * are used at the same time, this is the untranslated key.
	 */
	String					   _name;

	/** Short name. */
	String					   _alias;

	/** Tip. */
	String					   _tip;

	/** List of options of this group. */
	LyInfoList<LyInfo>         _options;

	/** The context option for this group. */
	LyChoiceOption			   _contextOption;

	/** Value of this choice options group. */
	Object					   _value;

	/** List of labels associated to this option. */
	LyInfoLabel[]			   _labels;

	/** List of user labels associated to this option. */
	Vector<LyInfoLabel>		   _userLabels;

	/**
	 * Internal representation of the value used to speed up processings when internal values are
	 * shorts.
	 */
	short					   _shortValue;

	/**
	 * Initialisation of instance attributes.
	 */
	protected void init()
	{
		_name = null;
		_alias = null;
		_tip = null;
		_options = null;
		_contextOption = null;
		_value = null;
		_shortValue = -1;
	}

    /**
     * Constructor of an options group.
     *
     * @param id    Id of this options group. Used as identifier within data description.
     */
    LyChoiceOptionGroup(String id)
    {
        super(id);
    }
    
	/**
	 * Constructor of an options group.
	 *
	 * @param id    Id of this options group. Used as identifier within data description.
	 * @param name  Name of the options group shown to the end-user.
	 * @param alias Short name of the options group.
	 * @param Value Value of the options group.
	 */
	LyChoiceOptionGroup(String id, String name, String alias, String value)
	{
		super(id);

		init();

		_name = ((name == null) ? null : name.intern());
		_alias = ((alias == null) ? null : alias.intern());
		setValue(value);
		_options = new LyInfoList<LyInfo>();
	}

	/**
	 * Lock this options group.
	 */
	protected void lock()
	{
		super.lock();

		// Turn options list to unwritable
		_options.lock();
	}

	/**
	 * Accessor to the name of this options group.
	 *
	 * @return Name of this options group.
	 */
	public String getName()
	{
		return _name;
	}

	/**
	 * Accessor to the name of this option group.
	 *
	 * @param name Name of this option group.
	 */
	void setName(String name)
	{
		_name = name;
	}

	/**
	 * Accessor to the short name of this option group.
	 *
	 * @param alias Short name of this option group.
	 */
	void setAlias(String alias)
	{
		_alias = alias;
	}

	/**
	 * Accessor for alias (shortname) of this instance.
	 *
	 * @return Alias of this instance. If the alias is null, this method returns the name.
	 */
	public String getAlias()
	{
		if (_alias == null)
			return getName();

		return _alias;
	}

    /**
     * Accessor for alias (shortname) of this instance. This method doesn't compute a value if it is
     * null
     *
     * @return Alias of this instance.
     */
    public final String getInternalAlias()
    {
        return _alias;
    }
    
	/**
     * Sets the alias of the instance.
     *
     * @param alias Alias of the instance.
     */
    final void setInternalAlias(String alias)
    {
        _alias = ((alias == null) ? null : alias.intern());
    }
    
	/**
	 * Accessor to the tip of this options group.
	 *
	 * @return Tip of this options group.
	 */
	public String getTip()
	{
		return _tip;
	}

	/**
	 * Gets an item from the options group at a given position.
	 *
	 * @param  index Index of the searched item.
	 * @return The item found, null otherwise.
	 */
	public LyChoiceItem getChoiceItem(short index)
	{
		if (index < 0)
			return null;

		return (LyChoiceItem)_options.getInfo(index);
	}

	/**
	 * Gets options count.
	 *
	 * @return Return the number of options in this options group.
	 */
	public short getChoiceItemCount()
	{
		return (short)_options.getSize();
	}

	/**
	 * Returns the list of options available on this options group.
	 *
	 * @return List of options of this options group.
	 */
	public LyInfoList<LyInfo> getChoiceItems()
	{
		return _options;
	}

	/**
	 * Returns the list of options available on this options group.
	 *
	 * @return List of options of this options.
	 */
	public LyInfoList<LyChoiceOption> getChoiceOptions(boolean godown)
	{
		LyInfoList<LyChoiceOption> options = new LyInfoList<LyChoiceOption>();
		short					   i, n = getChoiceItemCount();
		LyChoiceItem			   choiceItem;

		for (i = 0; i < n; i++)
		{
			choiceItem = getChoiceItem(i);
			if (choiceItem instanceof LyChoiceOption)
				options.addInfo((LyChoiceOption)choiceItem);
			else if (godown)
				options.addList(((LyChoiceOptionGroup)choiceItem).getChoiceOptions(true));
		}

		return options;
	}

	/**
	 * Utility method to retrieve options group of this choice field.
	 *
	 * @return List of options group.
	 */
	public LyInfoList<LyChoiceOptionGroup> getChoiceOptionGroups(boolean godown)
	{
		LyInfoList<LyChoiceOptionGroup> groups = null;
		short						    i, n = getChoiceItemCount();
		LyChoiceItem				    choiceItem;
		LyChoiceOptionGroup			    group;

		for (i = 0; i < n; i++)
		{
			choiceItem = getChoiceItem(i);
			if (choiceItem instanceof LyChoiceOptionGroup)
			{
				group = (LyChoiceOptionGroup)choiceItem;
				if (groups == null)
					groups = new LyInfoList<LyChoiceOptionGroup>();
				groups.addInfo(group);
				if (godown)
					groups.addList(group.getChoiceOptionGroups(true));
			}
		}

		return groups;
	}

	/**
	 * Get the context option for this group.
	 *
	 * @return the context option for this group
	 */
	public LyChoiceOption getOption()
	{
		return _contextOption;
	}

	/**
	 * Utility method to check if this group contains the given choice item.
	 *
	 * @param  choiceItem searched choice item.
	 * @return True if contained, false otherwise.
	 */
	public boolean contains(LyChoiceItem item, boolean godown)
	{
		if (item == this)
			return true;

		short		 i, n = getChoiceItemCount();
		LyChoiceItem choiceItem;

		for (i = 0; i < n; i++)
		{
			choiceItem = getChoiceItem(i);
			if (choiceItem == item)
				return true;
			if (godown
					&& (choiceItem instanceof LyChoiceOptionGroup)
					&& ((LyChoiceOptionGroup)choiceItem).contains(item, true))
				return true;
		}

		return false;
	}

	/**
	 * Accessor to the value of this options group.
	 *
	 * @return Value of this options group.
	 */
	public Object getValue()
	{
		return _value;
	}

	/**
	 * Gets the short value corresponding to the internal value. Used to optimize processings when
	 * internal values are shorts.
	 *
	 * @return Short value corresponding to this group, 1 if value can't be casted to a short.
	 */
	public short getShortValue()
	{
		return _shortValue;
	}

	/**
	 * Adds an item to this options group at a given position.
	 *
	 * @param item Choice item to add this options group.
	 */
	void addChoiceItem(LyChoiceItem item)
	{
	    if (_options == null)
	        _options = new LyInfoList<LyInfo>();
		_options.addInfo((LyInfo)item);
	}

	/**
	 * Removes an item from this options group.
	 *
	 * @param item Choice item to remove from this options group
	 */
	void removeChoiceItem(LyChoiceItem item)
	{
		_options.removeInfo((LyInfo)item);
	}

	/**
	 * Resets options list (clear all options).
	 */
	void resetChoiceItems()
	{
		_options.resetInfos();
	}

	/**
	 * Sets the value of this options group.
	 *
	 * @param value Value of this options group.
	 */
	void setValue(Object value)
	{
		if (value instanceof String)
			value = ((String)value).intern();

		if ((_value != null) && (_value != value))
			throw new IllegalStateException("BF AppEngine : trying to change value of options group "
				+ this + " from " + _value + " to " + value);

		_value = value;

		if (_value != null)
		{
			// Try to convert it to a short when possible to speed up future processings
			try
			{
				_shortValue = Short.parseShort(_value.toString());
			}
			catch (NumberFormatException e)
			{
				_shortValue = -1;
			}
		}
	}

	/**
	 * Set the context option for this group.
	 *
	 * @param option the context option
	 */
	void setOption(LyChoiceOption option)
	{
		_contextOption = option;
	}

	/**
	 * Complete options group method used by Gui parser to fulfill this options group with gui
	 * attributes.
	 */
	void completeChoiceOptionGroup(String name, String alias)
	{
		_name = ((name == null) ? null : name.intern());
		_alias = ((alias == null) ? null : alias.intern());
	}

	/**
	 * Add a new label to labels list of this group.
	 *
	 * @param label Label to add.
	 */
	void addLabel(LyInfoLabel label)
	{
		if (label == null)
			throw new IllegalArgumentException("BF AppEngine : label can't be null");

        LyInfoLabel.Type type = label.getType();

		if (type == LyInfoLabel.Type.USER)
		{
			if (_userLabels == null)
				_userLabels = new Vector<LyInfoLabel>(1);
			_userLabels.add(label);
			return;
		}

		if ((type != LyInfoLabel.Type.NAME)
				&& (type != LyInfoLabel.Type.GRAMMATICAL)
				&& (type != LyInfoLabel.Type.SEMANTIC)
				&& (type != LyInfoLabel.Type.VALUE)
				&& (type != LyInfoLabel.Type.SHORT_VALUE)
				&& (type != LyInfoLabel.Type.USER))
			throw new IllegalArgumentException("BF AppEngine : invalid type of label on fieldInfo "
				+ this);

		if (_labels == null)
			_labels = new LyInfoLabel[7];
		_labels[type.ordinal()] = label;
	} // end method addLabel

	/**
	 * Returns the number of labels of this group.
	 */
	public int getLabelCount()
	{
		return ((_labels == null) ? 0 : _labels.length);
	}

	/**
	 * Returns the the label of this group for the specified type.
	 */
	public LyInfoLabel getLabel(LyInfoLabel.Type type)
	{
		if (_labels == null)
			return null;

		return _labels[type.ordinal()];
	}

    /**
     * Returns the label of this group at the given index.
     */
    public LyInfoLabel getLabel(int index)
    {
        if ((_labels == null)
                || (index < 0)
                || (index > _labels.length))
            return null;

        return _labels[index];
    }
    
	/**
	 * Returns the number of labels of this group.
	 */
	public int getUserLabelCount()
	{
		return ((_userLabels == null) ? 0 : _userLabels.size());
	}

	/**
	 * Returns the user label at the given position.
	 */
	public LyInfoLabel getUserLabel(int i)
	{
		return ((_userLabels == null) ? null : (LyInfoLabel)_userLabels.elementAt(i));
	}

	/**
	 * Returns the user label for the given subType.
	 */
	public LyInfoLabel getUserLabel(String subType)
	{
		int		    i, n = getUserLabelCount();
		LyInfoLabel label;

		if (subType != null)
			subType = subType.intern();
		for (i = 0; i < n; i++)
		{
			label = _userLabels.elementAt(i);
			if (label.getSubType() != subType)
				return label;
		}

		return null;
	}
	
	/**
     * Sets the list of options available on this options group.
     */
    protected void setChoiceItems(LyInfoList<LyChoiceOption> options)
    {
        int     i, n = (options == null ? 0 : options.getSize());
        
        _options = null;
        for (i = 0; i < n; i++)
            addChoiceItem(options.getInfo(i));
    }
} // end class LyChoiceOptionGroup
