/*
 * LyClassInfo.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.lang.reflect.*;
import java.util.*;

import leon.app.*;
import leon.data.*;
import leon.info.binding.*;
import leon.info.location.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.union.*;


/**
 * The instances of this class provide a description of an application class.
 *
 * <p>Instances of LyClassInfo are a memory representation of the information read from the data
 * description file.<br>
 * The data description file (XML) is read at startup : for each tag <CLASS> tag found, the
 * application creates a new instance of LyClassInfo to store the information from the XML file in a
 * memory representation. This information is used during execution by the control package to offer
 * some generic processing.</p>
 *
 * <p>Instances of LyClassInfo offer the following services :</p>
 *
 * <ul>
 *   <LI>General information about the class : identifier, name, help file, resource files (icon,
 *     data files ...)</LI>
 *   <LI>Access to the meta description of the application classes : mainly, fields (attributes) and
 *     actions,</LI>
 *   <LI>Information about the connection to the data source : reference to binding information,
 *   </LI>
 *   <LI>Set of marks (properties of the class) to offer generic processing such as automatic
 *     loading and unloading.</LI>
 *   <LI>Routes that define how to calculate cross-references between classes.</LI>
 *   <LI>A class info may also have a behavior which defines behaviors required for main actions :
 *     consult, create, clone, set, delete. This behavior is only used for these actions when no
 *     other behavior is set for the given action.</LI>
 *   <LI>A cache policy may be specified on the class info. This allows to tune how objects of this
 *     class are loaded/unloaded in memory.</LI>
 *   <LI>A default sort can be defined on a class info : this sort is used on table views when no
 *     sort is given for that view.<br>
 *     This class implements the LyClass interface. This allows to use a classInfo as a
 *     physical/peer class for an other classInfo (application classes can be built over other
 *     application classes).<br>
 *     See also tag <CLASS> of leon.dtd.</LI>
 * </ul>
 *
 * @see LyFieldInfo
 * @see LyAction
 * @see LyFilter
 * @see leon.app.behavior.LyClassBehavior
 * @see LyCache
 */
public class LyClassInfo extends LyObjectInfo
	implements LyClass
{
	public static final String __version =
		"$Id: LyClassInfo.java 40625 2012-08-31 13:57:36Z jlsaouli $";

	/** Mark : To specify that the class is local (no physical representation). */
	public static final int    MARK_LOCAL = 0;

	/** Mark : Load this class on launch. */
	public static final int    MARK_LOAD = 1;

	public static boolean      ALLOW_NON_ABSTRACT_INHERITED_CLASS = false;

	/**
	 * Gets mark index corresponding to the given mark name.
	 *
	 * @param  markName Mark name.
	 * @param  add indicates if the.mark must be added if it doesn't exit
	 * @return Mark value.
	 */
	public static int getMark(String markName, boolean add)
	{
		return LyInfoTokens.getMarkIndex(LyBitSet.Type.C_MARKS, markName, add);
	}

	/**
	 * Gets mark index corresponding to the given mark name.
	 *
	 * @param  markName Mark name.
	 * @return Mark value.
	 */
	public static int getMark(String markName)
	{
		return getMark(markName, false);
	}

	/**
	 * Get mark name for the given index of a mark.
	 *
	 * @param  markIndex Index of the mark.
	 * @return Mark name.
	 */
	public static String getMarkName(int markIndex)
	{
		return LyInfoTokens.getMarkName(LyBitSet.Type.C_MARKS, markIndex);
	}

	/** Super class info. */
	LyClassInfo					    _superClass;

	/** List of sub-class infos. */
	LyInfoList<LyClassInfo>		    _subClasses;

	/** Binding (mean to access information from a data provider). */
	LyClassBinding				    _binding;

	/** List of the fieldInfos (attributes) of the application class. */
	LyInfoList<LyFieldInfo>		    _fieldInfos;

	/** List of actions available on the application class. */
	LyInfoList<LyAction>		    _actions;

	/** All routes of the application class. */
	LyInfoList<LyRoute>			    _routes;

	/**
	 * Informations specifying which cache policy should be used for this class. The cache is built
	 * from the java class specified in this object by reflection.
	 */
	LyInfoCache					    _cachePolicy;

	/**
	 * Behavior class that defines behavior for main actions of a class : consult, create, clone,
	 * set, delete. The behavior is built from this java class by reflection.
	 */
	LyJavaName						_behaviorClassName;

	/** Default sort : this sort is used in table views if no sort is given for this view. */
	LySort						    _sort;

	/** Name of the generated java class (application class) if code generation is used. */
	LyJavaName					    _javaName;

	/** Physical representation for objects of this class. */

	Class<?>					    _physicalClass;

	/** Get methods corresponding to each fieldInfo. */
	Method[]					    _getMethods;

	/** Set methods corresponding to each fieldInfo. */
	Method[]					    _setMethods;

	/** Optimization : cache the list of ID fields. */
	LyInfoList<LyFieldInfo>		    _idFields;

	/** Optimization : cache the list of NAME fields. */
	LyInfoList<LyFieldInfo>		    _nameFields;

	/** Hashtable of indexes to speed up fieldInfo accesses. */
	Hashtable<LyFieldInfo, Integer> _indexes;

    /** Hashtable of fields by id to speed up fieldInfo searches. */
    Hashtable<String, LyFieldInfo>  _fieldsById;

	/** Object source of this classInfo. */
	LyObjectSource				    _objectSource;

	/** For optimization : keep information if name and id are the same. */
	boolean						    _nameDifferentOfId;

	/** For optimization : keep information if name is a sub-part of id. */
	boolean						    _namePartOfId;

	/**
	 * Boolean indicating if the class info is temporary or not (class info created for internal
	 * data processing).
	 */
	boolean						    _temporary;

	/** Rules associated to this class info. */
	LyInfoList<LyRule>			    _rules;

	/** List of controls associated to this class. */
	LyInfoList<LyControl>		    _controls;

	/** Boolean that specifies if this class is abstract (can't be instantiated). */
	boolean						    _abstract;

	/** Boolean that specifies if this class is virtual. */
	boolean						    _virtual;

	/** List of labels associated to this class. */
	LyInfoLabel[]				    _labels;

	/** List of user labels associated to this class. */
	Vector<LyInfoLabel>			    _userLabels;

	/** Internal counter to compute internal objects id. */
	int							    _counter;

	/** Allow inherited classes to be not abstract. */
	Boolean						    _allowNonAbstractInheritedClasses;

    /**
     * Indicates if field is local or not.
     */
    Boolean                         _local;

	/**
	 * Indicates whether we must check the existence of the template when inserting a field info
	 * in current class info.
	 */
	protected boolean				_checkFieldInsertion;

	/**
	 * Initialization of classInfo attributes.
	 */
	protected void init()
	{
		super.init();

		// Initialize class attributes.
		_superClass = null;
		_subClasses = null;
		_fieldInfos = new LyInfoList<LyFieldInfo>(false);
		_fieldsById = null;
		_actions = new LyInfoList<LyAction>(false);
		_routes = null;
		_sort = null;
		_binding = null;
		_physicalClass = null;
		_getMethods = null;
		_setMethods = null;
		_idFields = null;
		_nameFields = null;
		_indexes = null;
		_objectSource = null;
		_behaviorClassName = null;
		_nameDifferentOfId = true;
		_namePartOfId = false;
		_temporary = false;
		_cachePolicy = new LyInfoCache();
		_cachePolicy.setParent(this);
		_abstract = false;
		_virtual = false;
		_rules = null;
		_controls = null;
		_labels = null;
		_checkFieldInsertion = true;
	} // end method init

	/**
	 * Initializes this instance attributes from its template.
	 *
	 * @param parts Indicates which parts of the template are copied.
	 */
	void initFromTemplate(int parts)
	{
        parts &= ~_initParts;
        if (parts == 0)
            return;
		if (_template == null)
			return;

		LyClassInfo template = (LyClassInfo)_template;

		super.initFromTemplate(parts);

		if ((parts & PART_DATA) > 0)
		{
			if (_superClass == null)
				setSuperClass(template._superClass);

			_fieldInfos = new LyInfoList<LyFieldInfo>(template._fieldInfos);
			_idFields = null;
			_nameFields = null;
			if (template._routes != null)
			    _routes = new LyInfoList<LyRoute>(template._routes);
			if (template._controls != null)
				_controls = new LyInfoList<LyControl>(template._controls);
			if (template._rules != null)
				_rules = new LyInfoList<LyRule>(template._rules);

			int		    i, n = getFieldInfoCount();
			LyFieldInfo field, subField;

			for (i = 0; i < n; i++)
			{
				field = getFieldInfo(i);
				if (field instanceof LyStructFieldInfo)
					((LyStructFieldInfo)field).addParentClass(this);
				else if (field instanceof LyArrayFieldInfo)
				{
					subField = ((LyArrayFieldInfo)field).getFieldInfo();
					if (subField instanceof LyStructFieldInfo)
						((LyStructFieldInfo)subField).addParentClass(this, field);
				}
			}
		}
		if ((parts & PART_BIND) > 0)
		{
			if ((_binding == null)
					&& !isVirtual())
			{
				if (template._binding != null)
					_binding = (LyClassBinding)template._binding.clone();
			}
			_cachePolicy = (LyInfoCache)template._cachePolicy.clone();
		}
		if ((parts & PART_GUI) > 0)
		{
			_actions = new LyInfoList<LyAction>(template._actions);
			_sort = template._sort;
			if (_behaviorClassName == null)
				_behaviorClassName = template._behaviorClassName;
		}
		if (((parts & PART_GUI) > 0)
				|| ((parts & PART_LABEL) > 0))
		{
			LyInfoLabel label;
			int		    i, nl = template.getLabelCount();
			int		    nu = template.getUserLabelCount();

			for (i = 0; i < nl; i++)
			{
				label = template._labels[i];
				if ((label != null) && (getLabel(i) == null))
					addLabel(label);
			}
			for (i = 0; i < nu; i++)
			{
				label = template.getUserLabel(i);
				if ((_userLabels == null) || !_userLabels.contains(label))
					addLabel(label);
			}
		}
	} // end method initFromTemplate

	/**
	 * Final step of parsing : checks integrity of the classInfo and completes the initialization if
	 * necessary. Checks performed are the following :
	 *
	 * <ul>
	 *   <LI>the fields list must not be empty,</LI>
	 *   <LI>the class info should contain id fields,</LI>
	 *   <LI>all routes have to be valid.</LI>
	 * </ul>
	 *
	 * @param     parent Object info in which this class is checked.
	 * @param     parts  Indicates which parts are checked
	 * @exception parent                Parent object info (should be a project).
	 * @exception IllegalStateException If one of the checks fails
	 */
	void check(LyObjectInfo parent, int parts)
	{
		super.check(parent, parts);

		if (!(parent instanceof LyProject))
			// Check only once
			return;

		if ((parts & PART_DATA) > 0)
		{
			// The fields list must not be empty
			if ((_fieldInfos.getSize() == 0) && !isAbstract())
				throw new IllegalStateException("BF AppEngine : non abstract classInfo " + this
					+ " contains no field");

			// The class info should contain id fields
			LyInfoList<LyFieldInfo> idFields = getIdFieldInfos();

			if ((!isAbstract())
					&& ((idFields == null)
						|| (idFields.getSize() == 0)))
				throw new IllegalStateException("BF AppEngine : classInfo " + this
					+ " has no id field");
		}

		LyEnvironment environment = getEnvironment();

		if ((parts & PART_BIND) > 0)
		{
			// If the class info is declared local, the binding is ignored.
			if (hasMark(MARK_LOCAL) && (_binding != null))
			{
				_binding = null;
				environment.logWarning("BF AppEngine : binding for class " + this
					+ " will be ignored because this class info has the mark LOCAL.");
			}
			else if (_binding != null)
				_binding.check(this);

			// If the class info is declared local, the load mark is removed
			if (hasMark(MARK_LOCAL) && hasMark(MARK_LOAD))
			{
				_marks.clear(MARK_LOAD);
				environment.logWarning("BF AppEngine : remove mark LOAD for class " + this
					+ " because this class info has mark LOCAL.");
			}

			// Check that all subclasses of a superclass have correct bindings with filters
			LyInfoList<LyClassInfo>  subClasses = getSubClasses();
			int                      n = (subClasses == null ? 0 : subClasses.getSize());

			if ((n > 1)
			    || ((n == 1) && !isAbstract()))
			{
    			LyClassInfo      subClass, subClass2;
    			LyClassBinding   classBinding, classBinding2;
    			int              i, j;

    			for (i = 0; i < n; i++)
    			{
    			    subClass = subClasses.getInfo(i);
    			    if (subClass.isAbstract())
    			        continue;

    			    classBinding = subClass.getBinding();
    			    if (classBinding == null)
    			        continue;

    			    if (classBinding.getFilter() == null)
    			    {
        			    for (j = 0; j < n; j++)
        			    {
        			        if (j == i)
        			            continue;

        			        subClass2 = subClasses.getInfo(j);
        			        classBinding2 = subClass2.getBinding();

        	                 if (classBinding2 == null)
        	                        continue;

            			    if ((classBinding.getBinding() == classBinding2.getBinding())
            			            && (classBinding.getLocation() == classBinding2.getLocation()))
            		             environment.logError("BF AppEngine : subclass " + subClass + " of class " + this
            		                 + " has the same binding as class " + subClass2
            		                 + " with no filter on its binding, this may produce unpredictable results.");
        			    }
    			    }
    			}
			}
		}

		if ((parts & PART_DATA) > 0)
		{
			// Check all the routes.
			int     i;
			LyRoute route;

			for (i = getRoutesCount() - 1; i >= 0; i--)
			{
				route = getRoute(i);
				try
				{
					route.check(this);
				}
				catch (Exception e)
				{
					environment.logWarning(e.getMessage());
					removeRoute(route);
					environment.logWarning("BF AppEngine : route " + route
						+ " removed from classInfo " + this + " because check failed.");
				}
			}
		}

		if ((parts & PART_BIND) > 0)
		{
			LyClassBinding binding = getBinding();

			if (binding != null)
			{
				String		  locationId = binding.getLocation();
				LyApplication application = getApplication();
				LyLocation    location = (LyLocation)application.getInfo(locationId);

				if (location != null)
					location.check(this);
			}
		}

		if (((parts & PART_GUI) > 0)
				|| ((parts & PART_LABEL) > 0))
		{
			// Check labels.
			int		    i, n = ((_labels == null) ? 0 : _labels.length);
			LyInfoLabel label;

			for (i = 0; i < n; i++)
				if (_labels[i] != null)
					_labels[i].check(this);

			n = ((_userLabels == null) ? 0 : _userLabels.size());
			for (i = 0; i < n; i++)
			{
				label = _userLabels.elementAt(i);
				if (label != null)
					label.check(this);
			}
		}

		if (((parts & PART_DATA) > 0)
				|| ((parts & PART_GUI) > 0))
		{
			// Check rules.
			LyRule rule;
			int    i, n = ((_rules == null) ? 0 : _rules.getSize());

			for (i = 0; i < n; i++)
			{
				rule = _rules.getInfo(i);
				rule.check(this);
			}
		}

		if ((parts & PART_GUI) > 0)
		{
			// Add actions of virtual class to its real class
			if (_virtual && (_actions != null))
			{
				LyAction    action, template;
				LyClassInfo realClass;
				int		    i, n = _actions.getSize();

				for (i = 0; i < n; i++)
				{
					action = _actions.getInfo(i);
					template = (LyAction)action.getTemplate();
					realClass = getRealClass();

					if ((template == null)
							|| !realClass.contains(template))
						realClass.addAction(action, -1);
				}
			}
		}

		if ((parts & PART_BIND) > 0)
		{
			// We must verify there is no mandatory composition relation
			// to this classInfo.
			LyFieldInfo				 fieldInfo;
			LyRelationFieldInfo		 relation;
			LyRelationFieldInfo.Type type;
			BitSet					 marks;
			int						 i, n = getFieldInfoCount();

			for (i = 0; i < n; i++)
			{
				fieldInfo = getFieldInfo(i);
				if (fieldInfo instanceof LyRelationFieldInfo)
				{
					relation = (LyRelationFieldInfo)fieldInfo;
					type = relation.getType();
					if ((type == LyRelationFieldInfo.Type.COMPOSITION)
							&& (relation.relatedWith(this))
							&& (!relation.hasMark(LyFieldInfo.MARK_OPTIONAL)))
					{
						marks = relation.getMarks();

						if (marks == null)
							marks = new BitSet();

						marks.set(LyFieldInfo.MARK_OPTIONAL);
						relation.setMarks(marks);
						environment.logWarning("BF AppEngine : add mark OPTIONAL on the fieldInfo "
							+ relation
							+ " which is a composition relation related with its owner classInfo ("
							+ this + ").");
					}
				}
			}

			// Check that only local fields were added to virtual classes
			if (isVirtual())
			{
				LyFieldInfo template;

				for (i = 0; i < n; i++)
				{
					fieldInfo = getFieldInfo(i);
					if (_superClass.contains(fieldInfo))
						continue;

					template = (LyFieldInfo)fieldInfo.getTemplate();
					if (!fieldInfo.isLocal()
							&& ((template == null) || !_superClass.contains(template)))
						throw new IllegalStateException("BF AppEngine : non-local field info "
							+ fieldInfo + " can't be added to virtual class " + this);
				}
			}
		}
	} // end method check

	/**
	 * Lock this class info.
	 */
	protected void lock()
	{
		if (!isLoaded())
			return;

		super.lock();

		// Turn lists to unwritable
		// Routes stays writable because other routes are added when needed
		// Sub-classes stays also writable because other classes may
		// be added . example : import of data => creation of a temporary class info.
		// FieldInfos stays writable because another module can add a facet relation
		// to a classInfo or new actions to existing classes.
		if (_controls != null)
			_controls.lock();
		if (_rules != null)
			_rules.lock();
		if (_labels != null)
		{
			int i, n = ((_labels == null) ? 0 : _labels.length);

			for (i = 0; i < n; i++)
				if (_labels[i] != null)
					_labels[i].lock();
		}

		if (getLabel(LyInfoLabel.Type.NAME) != null)
		{
			_nameDifferentOfId = true;
			_namePartOfId = false;
			return;
		}

		// Do object names differ from object ids ?
		computeNameDifferentOfId();
	} // end method lock

	/**
	 * Compute if the name is different of id.
	 */
	protected void computeNameDifferentOfId()
	{
		LyFieldInfo fieldInfo;
		int		    i, n = getFieldInfoCount();
		boolean     hasNames = false, id, name;

		_nameDifferentOfId = false;
		_namePartOfId = true;

		for (i = 0; i < n; i++)
		{
			fieldInfo = getFieldInfo(i);
			name = fieldInfo.hasMark(LyFieldInfo.MARK_NAME);
			if (name)
				hasNames = true;
		}

		LyRelationFieldInfo       relation;
		LyInfoList<LyClassInfo>	  classes;
		LyClassInfo		          classInfo;
		int				          j, nc;

		for (i = 0; i < n; i++)
		{
			fieldInfo = getFieldInfo(i);
			id = fieldInfo.hasMark(LyFieldInfo.MARK_ID);
			name = fieldInfo.hasMark(LyFieldInfo.MARK_NAME);

			if (name && !id)
			{
				_nameDifferentOfId = true;
				_namePartOfId = false;
				break;
			}
			else if (id)
			{
				if (hasNames && !name)
					_nameDifferentOfId = true;
				else if (fieldInfo instanceof LyRelationFieldInfo)
				{
					// In this case, we can't really be sure that id and name differ
					// for related classes because it depends on the order of lock on classes list.
					relation = (LyRelationFieldInfo)fieldInfo;
					classes = relation.getClassInfos();
					nc = classes.getSize();

					for (j = 0; j < nc; j++)
					{
						classInfo = relation.getClassInfo(j);
						classInfo.load();

						if (classInfo.isNameDifferentOfId())
						{
							_nameDifferentOfId = true;
							_namePartOfId = false;
							break;
						}
					}
				}
			}
		}
	} // end method computeNameDifferentOfId

	/**
	 * Constructor of a class info.
	 *
	 * @param id Identifier of the class info.
	 */
	LyClassInfo(String id)
	{
		super(id, false);
	}

	/**
	 * Constructor of a class info.
	 *
	 * @param id                Identifier of the class info.
	 * @param superClass        Reference of the superClass.
	 * @param template          Reference of a class template used to initialize this new instance.
	 *                          This parameter is ignored if superClass is not null.
	 * @param javaName          Name of the java class that stores data for instances of this class
	 *                          (these instances are logical instances of the classInfo, since they
	 *                          are physical instances of the java class pointed by this name).
	 * @param behaviorClassName Java class name that defines default behaviors for this class on
	 *                          main actions : consult, create, clone, set, delete.
	 * @param abstractClass     Boolean that specifies if this class is abstract or not.
	 * @param virtual           Boolean that specifies if this class is virtual or not.
	 * @param userData          User data.
	 */
	LyClassInfo(String id, LyClassInfo superClass, LyClassInfo template,
		String javaName, String behaviorClassName, boolean abstractClass,
		boolean virtual, Object userData)
	{
		this(id);
		construct(superClass, template, javaName, behaviorClassName, abstractClass, virtual,
			userData);
	}

	/**
	 * Construction of a class info.
	 *
	 * @param superClass        Reference of the superClass.
	 * @param template          Reference of a class template used to initialize this new instance.
	 *                          This parameter is ignored if superClass is not null.
	 * @param javaName          Name of the java class that stores data for instances of this class
	 *                          (these instances are logical instances of the classInfo, since they
	 *                          are physical instances of the java class pointed by this name).
	 * @param behaviorClassName Java class name that defines default behaviors for this class on
	 *                          main actions : consult, create, clone, set, delete.
	 * @param abstractClass     Boolean that specifies if this class is abstract or not.
	 * @param virtual           Boolean that specifies if this class is virtual or not.
	 * @param userData          User data.
	 */
	protected void construct(LyClassInfo superClass, LyClassInfo template,
		String javaName, String behaviorClassName, boolean abstractClass,
		boolean virtual, Object userData)
	{
		if (superClass == null)
		{
			if (virtual)
				throw new IllegalStateException("BF AppEngine : virtual class " + this
					+ " must have a super-class");

			super.construct(template);
		}
		else
			super.construct(superClass);

		_abstract = abstractClass;
		_virtual = virtual;
		setUserData(userData);

		if (superClass != null)
			setSuperClass(superClass);

        setJavaName(javaName);

		// Sets behavior class
		setBehaviorClassName(behaviorClassName);

		// Warns user if template is not used
		if ((superClass != null)
				&& (template != null)
				&& (template != superClass))
			getEnvironment().logWarning("BF AppEngine : template " + template
				+ " ignored for class info " + this
				+ " because this class has a super class");
	} // end method construct

	/**
	 * Constructs a temporary class info for internal data processings from a given class info
	 * template.
	 *
	 * @param template Reference of a class template used to initialize this new instance.
	 * @param local    Boolean indicating if the class info should be local or not.
	 */
	public LyClassInfo(LyClassInfo template, boolean local)
	{
		this(template, null, null, local, PART_ALL);
	}

	/**
	 * Constructs a temporary class info for internal data processings from a given class info
	 * template.
	 *
	 * @param template Reference of a class template used to initialize this new instance.
	 * @param binding  Binding of this temporary class info.
	 * @param location Location of the binding.
	 * @param local    Boolean indicating if the class info should be local or not.
	 */
	public LyClassInfo(LyClassInfo template, String binding, String location, boolean local,
		int parts)
	{
		this(template.getId() + ':' + LyString.getRandomId(),
		    null, template, (template._javaName == null ? null : template._javaName.getJavaName()),
			null, template.isAbstract(), template.isVirtual(), template.getUserData());

		_temporary = true;
		initFromTemplate(parts);

		// Set application
		setApplication(template.getApplication());

		if (local)
		{
			if (_marks == null)
				_marks = new BitSet();
			_marks.set(MARK_LOCAL);
		}
		else
		{
			if (binding == null)
				binding = getId();

			if (location == null)
				location = getId();

			// Create a binding
			LyClassBinding classBinding = new LyClassBinding(binding, location, null);

			setBinding(classBinding);
		}
	}

	/**
	 * Constructs a temporary class info for internal data processings from a given class info
	 * template.
	 *
	 * @param template    Reference of a class template used to initialize this new instance.
	 * @param binding     Binding of this temporary class info.
	 * @param location    Location of the binding.
	 * @param local       Boolean indicating if the class info should be local or not.
	 * @param cloneStruct Boolean indicating if the structFields must be cloned.
	 */
	public LyClassInfo(LyClassInfo template, String binding, String location,
		boolean local, boolean cloneStruct, int parts)
	{
		this(template, binding, location, local, cloneStruct, true, true, parts);
	}

	/**
	 * Constructs a temporary class info for internal data processings from a given class info
	 * template.
	 *
	 * @param template     Reference of a class template used to initialize this new instance.
	 * @param binding      Binding of this temporary class info.
	 * @param location     Location of the binding.
	 * @param local        Boolean indicating if the class info should be local or not.
	 * @param cloneStruct  Boolean indicating if the structFields must be cloned.
	 * @param keepJavaName Boolean indicating if the javaname must be kept.
	 * @param keepBehavior Boolean indicating if the behavior must be kept.
	 */
	public LyClassInfo(LyClassInfo template, String binding, String location,
		boolean local, boolean cloneStruct, boolean keepJavaName, boolean keepBehavior, int parts)
	{
		this(template, binding, location, local, parts);

		if (!keepJavaName)
			_javaName = null;

		if (!keepBehavior)
			_behaviorClassName = null;

		_parent = template;

		if (cloneStruct)
		{
			LyFieldInfo		  fieldInfo;
			LyStructFieldInfo structFieldInfo;

			for (int i = 0; i < _fieldInfos.getSize(); i++)
			{
				fieldInfo = _fieldInfos.getInfo(i);

				if (fieldInfo instanceof LyStructFieldInfo)
				{
					structFieldInfo = (LyStructFieldInfo)fieldInfo;
					_fieldInfos.removeInfoAt(i);
					structFieldInfo.removeParentClass(this);
					if (fieldInfo.hasMark(LyFieldInfo.MARK_ID))
					{
						_idFields = null;
						_nameFields = null;
					}
					if (fieldInfo.hasMark(LyFieldInfo.MARK_NAME))
						_nameFields = null;

					String				    id = "__" + structFieldInfo.getId();
					String				    javaName = structFieldInfo.getJavaName();
					String				    className = structFieldInfo.getClassName();
					boolean				    isAbstract = template._abstract;
					LyStructFieldInfo	    extendedStruct = structFieldInfo.getExtendedStruct();
					LyInfoList<LyClassInfo> classes = structFieldInfo.getClassInfos();

					if (location == null)
						location = _id;

					LyStructFieldInfo newStructFieldInfo = new LyStructFieldInfo(id, javaName,
							extendedStruct, structFieldInfo, className, isAbstract);

					newStructFieldInfo.setMarks(structFieldInfo.getMarks());

					if (newStructFieldInfo.getName().equals(newStructFieldInfo.getId()))
						newStructFieldInfo.setName(fieldInfo.getName());

					LyClassInfo			    clonedClassInfo;
					LyInfoList<LyClassInfo> clonedClassInfos = new LyInfoList<LyClassInfo>();

					for (int j = 0; j < classes.getSize(); j++)
					{
						clonedClassInfo = new LyClassInfo(classes.getInfo(j), null, location, false,
								true, parts);
						clonedClassInfos.addInfo(clonedClassInfo);
					}

					newStructFieldInfo.replaceClassInfos(clonedClassInfos);
					structFieldInfo.getApplication().removeInfo(newStructFieldInfo);
					newStructFieldInfo.setApplication(structFieldInfo.getApplication());
					_fieldInfos.addInfo(newStructFieldInfo, i);
					newStructFieldInfo.addParentClass(this);

					// Force recomputing of indexes and methodTables
					resetFieldIndexes();
				}
			}
		}
	}

	/**
     * Returns extra tags.
     */
    protected LyTag[] getExtraTags()
    {
        return new LyTag[] { _behaviorClassName, _javaName };
    }

	/**
	 * Free a class info. Only a temporary class info can be freed.
	 *
	 * @exception IllegalStateException If this class info is not temporary.
	 */
	public void free()
	{
		/*if (!_temporary)
			throw new IllegalStateException("BF AppEngine : class info " + this
				+ " can't be freed because it's not temporary.");*/

		// remove the class from the super-class
		if (_superClass != null)
			_superClass._subClasses.removeInfo(this);

		// Also free cloned structs
		LyFieldInfo		  fieldInfo;
		LyStructFieldInfo structFieldInfo;

		for (int i = 0; i < _fieldInfos.getSize(); i++)
		{
			fieldInfo = _fieldInfos.getInfo(i);

			if (fieldInfo instanceof LyStructFieldInfo)
			{
				structFieldInfo = (LyStructFieldInfo)fieldInfo;

				if (structFieldInfo.getId().startsWith("__"))
				{
					LyInfoList<LyClassInfo> classes = structFieldInfo.getClassInfos();

					for (int j = 0; j < classes.getSize(); j++)
					{
						LyClassInfo structClassInfo = classes.getInfo(j);

						if (structClassInfo.isTemporary())
							structClassInfo.free();
					}

					structFieldInfo.free();
				}
			}
		}
		super.free();
	} // end method free

	/* Implementation of LyClass interface*/

	/**
	 * Gets physical class used to store physical instances of this application class.
	 *
	 * @return The physical Java class (as specified by javaName attribute or LyObject.class if
	 *         javaName is not set, or if the class corresponding to this name can't be found). For
	 *         a classInfo, getObjectClass should return a class that extends LyObject.
	 */
	public Class<?> getObjectClass()
	{
		load();

		// Use cached value if available
		if (_physicalClass != null)
			return _physicalClass;

		// Try to load application class from its name
		String javaName = getJavaName();

		if (javaName != null)
		{
			int pos = javaName.lastIndexOf('.');

			if (pos < 0)
				javaName = Character.toUpperCase(javaName.charAt(0)) + javaName.substring(1);
			else
				javaName = javaName.substring(0, pos + 1)
					+ Character.toUpperCase(javaName.charAt(pos + 1))
					+ javaName.substring(pos + 2);
		}

		try
		{
			LyApplication application = getApplication();

			_physicalClass = application.loadClass(javaName);
		}
		catch (Exception e)
		{
		}

		//When we have an inheritance in the classinfo, the interface class of the
		//superclass may be abstract so we can't instantiate it.
		LyClassInfo superClass = getSuperClass();

		if ((_physicalClass == null)
				&& (superClass != null)
				&& (!superClass.isAbstract()))
		{
			_physicalClass = superClass.getObjectClass();
			if ((_physicalClass != null)
					&& (_javaName != null))
				getEnvironment().logWarning("Lyria : Java class not found "
					+ getJavaName() + ". Using parent class ("
					+ _physicalClass.getName() + ").");
		}

		if (_physicalClass == null)
		{
			if (_javaName != null)
				getEnvironment().logWarning("Lyria : Java class not found "
					+ getJavaName() + ". Using generic one (LyObject).");

			// The default class.
			_physicalClass = LyObject.class;
		}
		else
			initMethodTables();

		return _physicalClass;
	} // end method getObjectClass

	/**
	 * Internal method used to initialize methods tables when physical class is known.
	 */
	void initMethodTables()
	{
		if ((_physicalClass == null)
				|| (_physicalClass == LyObject.class))
			return;

		// Fill up methods tables
		LyFieldInfo fieldInfo;
		int		    i, n = getFieldInfoCount();

		_getMethods = new Method[n];
		_setMethods = new Method[n];

		for (i = 0; i < n; i++)
		{
			fieldInfo = getFieldInfo(i);
			initMethods(fieldInfo, i);
		}
	}

	/**
	 * Internal method used to initialize methods tables when physical class is known.
	 */

	void initMethods(LyFieldInfo fieldInfo, int index)
	{
		if ((_getMethods == null)
				|| (index < 0)
				|| (index >= _getMethods.length))
			throw new IllegalStateException("BF AppEngine : invalid index " + index);

		String  fieldName, methodName;
		Class<?>[] setParameters = new Class[1];

		setParameters[0] = LyFieldInfo.class;

		fieldName = fieldInfo.getJavaName();
		fieldName = Character.toUpperCase(fieldName.charAt(0)) + fieldName.substring(1);
		methodName = "get" + fieldName;

		try
		{
			_getMethods[index] = _physicalClass.getMethod(methodName, (Class[])null);
		}
		catch (Exception e)
		{
			//getEnvironment().logTrace(e + " " + _physicalClass + "." + methodName);
			// The method was not found, no need to do anything
		}

		if ((fieldInfo instanceof LyRelationFieldInfo)
			&& ((_getMethods[index] == null)
				|| (_getMethods[index].getReturnType() != fieldInfo.getValueClass())))
		{
			if (((LyRelationFieldInfo)fieldInfo).isMultiple())
				methodName += "Ids";
			else
				methodName += "Id";
			try
			{
				_getMethods[index] = _physicalClass.getMethod(methodName, (Class[])null);
			}
			catch (Exception e)
			{
				//getEnvironment().logTrace(e + " " + _physicalClass + "." + methodName);
				// The method was not found, no need to do anything
			}
		}

		if ((_getMethods[index] != null)
				&& (_getMethods[index].getReturnType() != fieldInfo.getValueClass()))
		{
			getEnvironment().logError("BF AppEngine : accessor for field info "
				+ fieldInfo + " on classInfo " + this
				+ " is ignored because it doesn't return the expected type : "
				+ fieldInfo.getValueClass() + " but : "
				+ _getMethods[index].getReturnType());
			_getMethods[index] = null;
		}

		try
		{
			setParameters[0] = fieldInfo.getValueClass();

			_setMethods[index] = _physicalClass.getMethod("set" + fieldName, setParameters);
		}
		catch (Exception e)
		{
			getEnvironment().logTrace(e + " " + _physicalClass + ".set" + fieldName);
		}
	} // end method initMethods

	/**
	 * Gets data source for this class.
	 *
	 * @return The data source for this class. For a class info, the data source is in fact an
	 *         instance of LyObjetSource.
	 */
	public LyDataSource getDataSource()
	{
		return getObjectSource();
	}

	/**
	 * Gets id fields (field infos that identify instances of this class).
	 *
	 * @return Vector of id. fields. Normally, can't be null since a class with no id field is not
	 *         accepted.
	 */
	public Vector<LyFieldInfo> getIdFields()
	{
		LyInfoList<LyFieldInfo> infoList = getIdFieldInfos();

		return ((infoList == null) ? null : infoList.getInfos());
	}

	/**
	 * Gets list of fields (field infos) of this class.
	 *
	 * @return Vector of fields (field infos). Normally, can't be null since a class with no field
	 *         is not accepted.
	 */
	public Vector<LyFieldInfo> getFields()
	{
		LyInfoList<LyFieldInfo> infoList = getFieldInfos();

		return ((infoList == null) ? null : infoList.getInfos());
	}

	/**
	 * Retrieves a field (field info) by its id.
	 *
	 * @param  id Id of the searched field.
	 * @return Field with the given id if found, null otherwise.
	 */
	public LyField getField(String id)
	{
		return getFieldInfo(id);
	}

	/**
	 * Gets objectId of a given object.
	 *
	 * @param  object Object for which an objectId is requested.
	 * @return ObjectId of the given object.
	 */
	public LyObjectId getObjectId(Object object)
	{
		return ((LyObject)object).getObjectId();
	}

	/**
	 * Gets value of an object on a given field (field info).
	 *
	 * @param  object Object instance of this class.
	 * @param  field  Field on which a value is asked.
	 * @return The corresponding value of the object for the given field.
	 */
	public Object getValue(Object object, LyField field)
	{
		return ((LyObject)object).getValue((LyFieldInfo)field);
	}

	/**
	 * Sets value of an object on a given field (field info).
	 *
	 * @param object Object instance of this class.
	 * @param field  Field on which the value is set.
	 * @param value  Value to set on the object for the given field.
	 */
	public void setValue(Object object, LyField field, Object value)
	{
		((LyObject)object).setValue((LyFieldInfo)field, value);
	}

	/**
	 * Gets values set for the given object.
	 *
	 * @param  object Object on which values are requested.
	 * @return List of values for the given object.
	 */
	public LyValueSet getValues(Object object)
	{
		return ((LyObject)object).getValues();
	}

	/**
	 * This method does nothing on a classInfo but it has to be present and public because this
	 * method is defined in the LyClass interface and LyClassInfo implements this interface.
	 *
	 * @param  object1 Object that receives values of object2.
	 * @param  object2 Object to copy.
	 * @return This method always returns null.
	 * @see    leon.data.LyClass
	 */
	public LyValueSet copy(Object object1, Object object2)
	{
		if (object1 == null)
			throw new IllegalArgumentException("BF AppEngine : object1 can't be null");
		if (object2 == null)
			throw new IllegalArgumentException("BF AppEngine : object2 can't be null");

		// XXX
		return null;
	}

	/**
	 * Creates a new instance of this class.
	 *
	 * @return A logical instance of this class (physical instance of the class returned by
	 *         getObjectClass method, that should be a subclass of LyObject).
	 */

	public Object newInstance()
		throws InstantiationException,
			IllegalAccessException
	{
		Class<?>	    objectClass = getObjectClass();
		LyObject    object = (LyObject)objectClass.newInstance();
		LyClassInfo classInfo = getRealClass();

		if (object.getClassInfo() != classInfo)
			object.setClassInfo(classInfo);

		return object;
	}

	/**
	 * Creates a new instance of this class. Same as newInstance but catching exceptions and
	 * returning a LyObject.
	 *
	 * @return A logical instance of this class (physical instance of the class returned by
	 *         getObjectClass method, that should be a subclass of LyObject).
	 */
	public LyObject createInstance()
	{
		try
		{
			LyObject object = (LyObject)newInstance();

			return object;
		}
		catch (InstantiationException e)
		{
			throw new RuntimeException(e);
		}
		catch (IllegalAccessException e)
		{
			throw new RuntimeException(e);
		}
	}

	/**
	 * Checks if this class info inherits from the given one.
	 *
	 * @param  classInfo Checked class info
	 * @return True if this class info inherits from the given one, false otherwise.
	 */
	public boolean inheritsFrom(LyClassInfo classInfo)
	{
	    if (classInfo == null)
	        return false;
	    else if (_superClass == classInfo)
			return true;
		else if (_superClass == null)
			return false;
		else
			return _superClass.inheritsFrom(classInfo);
	}

	/**
	 * Checks if this class info is abstract or not. A class info is considered as abstract if it
	 * has at least one sub class.
	 *
	 * @return True if this class info is abstract, false otherwise.
	 */
	public boolean isAbstract()
	{
		if (_abstract)
			return true;
		else if (_subClasses != null)
		{
			if ((_allowNonAbstractInheritedClasses != null) && !_allowNonAbstractInheritedClasses.booleanValue())
			{
				_abstract = true;
			}
			else if (_allowNonAbstractInheritedClasses == null)
			{
				if (!ALLOW_NON_ABSTRACT_INHERITED_CLASS)
					return true;
			}
			return _abstract;
		}
		else
			return false;
	}

	/**
	 * Checks if this class info is virtual or not.
	 *
	 * @return True if this class info is virtual, false otherwise.
	 */
	public boolean isVirtual()
	{
		return _virtual;
	}

	/**
	 * Checks if the given field info is inherited : the field info was defined on the super-class.
	 *
	 * @param  fieldInfo Checked field info
	 * @return True if the field info is inherited, false otherwise.
	 */
	public boolean isInherited(LyFieldInfo fieldInfo)
	{
		return ((_superClass != null) && _superClass.contains(fieldInfo));
	}

    /**
     * Checks if the given field info is derivated : the field info was defined on the super-class or its template.
     *
     * @param  fieldInfo Checked field info
     * @return True if the field info is derivated, false otherwise.
     */
    public boolean isDerivated(LyFieldInfo fieldInfo)
    {
        return (isInherited(fieldInfo)
                || ((_template != null) && ((LyClassInfo)_template).contains(fieldInfo)));
    }

	/**
	 * Checks if this class is temporary (dynamic built).
	 *
	 * @retun True if the class is temporary, false otherwise.
	 */
	public boolean isTemporary()
	{
		return _temporary;
	}

	/**
	 * Returns the super class of this class info.
	 *
	 * @return Super class of this class info, null otherwise.
	 */
	public LyClassInfo getSuperClass()
	{
		return _superClass;
	}

	/**
	 * Gets the list of sub-classes.
	 *
	 * @return Info list that contains the sub-classes of this class info, null if this class info
	 *         is not abstract.
	 */
	public LyInfoList<LyClassInfo> getSubClasses()
	{
		return getSubClasses(false);
	}

	/**
	 * Gets the list of sub-classes.
	 *
	 * @param  godown indicates if the sub-classes are searched recursivly.
	 * @return Info list that contains the sub-classes of this class info, null if this class info
	 *         is not abstract.
	 */
	public LyInfoList<LyClassInfo> getSubClasses(boolean godown)
	{
	    if (_subClasses == null)
	        return null;
		if (!godown)
			return new LyInfoList<LyClassInfo>(_subClasses);

		LyInfoList<LyClassInfo> result = new LyInfoList<LyClassInfo>();
		int					    n = _subClasses.getSize();
		LyClassInfo			    subClass;

		for (int i = 0; i < n; i++)
		{
			subClass = _subClasses.getInfo(i);
			result.addList(subClass.getSubClasses(godown));
		}
		result.addList(_subClasses);

		return result;
	}

	/**
	 * Returns object source corresponding to this class info.
	 *
	 * @return The object source of this class info.
	 */
	public LyObjectSource getObjectSource()
	{
		if (_objectSource == null)
		{
			LyApplication application = getApplication();

			_objectSource = (LyObjectSource)application.getDataSource(_id);
		}

		return _objectSource;
	}

	/**
	 * Internal use to clear the objectsource cache when the object source has changed. The
	 * objectsource will be computed again from the application next time getObjectSource() is
	 * called.
	 *
	 * @see #getObjectSource()
	 */
	public void clearObjectSourceCache()
	{
		_objectSource = null;
	}

	/**
	 * Gets java class name corresponding to this class info when java code generation is used.
	 *
	 * @return Name of the java class corresponding to this classInfo. By default, if no java name
	 *         is set, this method returns the id of this class info.
	 */
	public String getJavaName()
	{
		if (_javaName != null)
			return _javaName.getJavaName();
		else
			return  getDefaultJavaName();
	}
	
	/**
	 * Gets default java class name corresponding to this class info.
	 *
	 * @return Name of the default java class corresponding to this classInfo.
	 */
	public final String getDefaultJavaName()
	{		
		String	javaName = getId();
		
		return checkJavaName(javaName);
	}
	
	/**
	 * check javaName attribute, replace points with underscore.
	 *
	 * @return new javaName
	 */
	public String checkJavaName(String javaName)
	{
		LyInfo	parent = getParent();
		
		if ((parent != null)
				&& javaName.startsWith(parent.getId()))
			javaName = javaName.substring(parent.getId().length());
				
		return javaName;
	}

	/**
	 * Gets behavior class name.
	 *
	 * @return Java name of behavior class (may be null).
	 */
	public String getBehaviorClassName()
	{
		return (_behaviorClassName == null ? null : _behaviorClassName.getJavaName());
	}

	/**
	 * Gets behavior class. Class that defines behavior required for this class info for main
	 * actions (consult, create, clone, set, delete).
	 *
	 * @return Behavior class for this class info (may be null).
	 */

	public Class<?> getBehaviorClass()
	{
	    if ((_behaviorClassName == null)
	        || (_application.isEdited()
	            && !_behaviorClassName.isGeneric()))
	        return null;

		// loading classes
		Class<?> classBehavior = _behaviorClassName.getJavaClass();

		if (classBehavior != null)
		    return classBehavior;

		LyEnvironment environment = getEnvironment();

		environment.logError("BF AppEngine : unknown behavior class " + _behaviorClassName
			+ " on class " + _id);
		return null;
	}

	/**
	 * Returns get method for the given fieldInfo. This method will be called by reflection.
	 *
	 * @param  fieldInfo Field info for which the get method is requested.
	 * @return The get method corresponding to the given field info (can be null).
	 */
	public Method getGetMethod(LyFieldInfo fieldInfo)
	{
		if (_getMethods == null)
			return null;

		int index = getIndex(fieldInfo);

		if (index < 0)
		{
			if (fieldInfo.isLocal())
				return null;
			else
				throw new IllegalArgumentException("BF AppEngine : no field info with id "
					+ fieldInfo.getId() + " found on class info " + this);
		}

		return _getMethods[index];
	}

	/**
	 * Returns set method for the given fieldInfo. This method will be called by reflection.
	 *
	 * @param  fieldInfo Field info for which the set method is requested.
	 * @return The set method corresponding to the given field info (can be null).
	 */
	public Method getSetMethod(LyFieldInfo fieldInfo)
	{
		if (_setMethods == null)
			return null;

		int index = getIndex(fieldInfo);

		if (index < 0)
		{
			if (fieldInfo.isLocal())
				return null;
			else
				throw new IllegalArgumentException("BF AppEngine : no field info with id "
					+ fieldInfo.getId() + " found on class info " + this);
		}

		return _setMethods[index];
	}

	/**
	 * Gets list of field infos of this class info.
	 *
	 * @return List of field infos, can't be null since a class info with no fields is not accepted.
	 */
	public LyInfoList<LyFieldInfo> getFieldInfos()
	{
		load();
		return (_fieldInfos == null ? null : new LyInfoList<LyFieldInfo>(_fieldInfos));
	}

	/**
	 * Gets id field infos (field infos that have MARK_ID).
	 *
	 * @return List of id field infos, can't be null since a class info with no id fields is not
	 *         accepted.
	 */
	public LyInfoList<LyFieldInfo> getIdFieldInfos()
	{
		if (_idFields == null)
			_idFields = getFieldInfos(LyFieldInfo.MARK_ID);

		return new LyInfoList<LyFieldInfo>(_idFields);
	}

	/**
	 * Gets name field infos (field infos that have MARK_NAME).
	 *
	 * @return List of name field infos, can't be null since a class info with no name fields use
	 *         the id fields as name fields.
	 */
	public LyInfoList<LyFieldInfo> getNameFieldInfos()
	{
		if (_nameFields == null)
		{
			_nameFields = getFieldInfos(LyFieldInfo.MARK_NAME);

			if ((_nameFields == null) || (_nameFields.getSize() == 0))
				_nameFields = getFieldInfos(LyFieldInfo.MARK_ID);
		}

		return new LyInfoList<LyFieldInfo>(_nameFields);
	}

	/**
	 * Search the fieldInfo with the given identifier.
	 *
	 * @param     id        Identifier of the fieldInfo
	 * @param     exception Boolean indicating if an exception is thrown if not found.
	 * @return    The fieldInfo corresponding to the given id if found, or null if not found and the
	 *            parameter exception is set to false.
	 * @exception IllegalStateException if the fieldInfo corresponding to the given id was not found
	 *                                  in this classInfo and if the parameter exception is set to
	 *                                  true.
	 */
	public LyFieldInfo getFieldInfo(String id, boolean exception)
	{
        LyFieldInfo fieldInfo;

        if (_fieldsById != null)
        {
            fieldInfo = _fieldsById.get(id);
            if (fieldInfo != null)
                return fieldInfo;
        }

		load();

		fieldInfo = _fieldInfos.getInfo(id);

		if (fieldInfo == null)
		{
		    // Access by type ?
	        LyInfo template = _application.getInfo(id);

	        if (template instanceof LyFieldInfo)
	            fieldInfo = getRealFieldInfo((LyFieldInfo)template, false);
		}

        if (fieldInfo != null)
        {
            if (_fieldsById == null)
                _fieldsById = new Hashtable<String, LyFieldInfo>(getFieldInfoCount());
            _fieldsById.put(id, fieldInfo);
        }
        else if (exception)
			throw new IllegalStateException("BF AppEngine : no field info with id " + id
				+ " found on class info " + this);

		return fieldInfo;
	}

	/**
	 * Search the fieldInfo with the given identifier. This method calls getFieldInfo(id, true).
	 *
	 * @param     id Identifier of the fieldInfo
	 * @return    The fieldInfo corresponding to the given id if found.
	 * @exception IllegalStateException if the fieldInfo corresponding to the given id was not found
	 *                                  in this classInfo.
	 * @see       getFieldInfo(String, boolean)
	 */
	public LyFieldInfo getFieldInfo(String id)
	{
		return getFieldInfo(id, true);
	}

	/**
	 * Gets a list of fieldInfos that match the given marks.
	 *
	 * @param  marks Requested marks for result fieldInfos.
	 * @return List of field infos that match the marks, an empty list if no field info has these
	 *         marks.
	 */
	public LyInfoList<LyFieldInfo> getFieldInfos(BitSet marks)
	{
		load();
		return _fieldInfos.getInfos(marks, false);
	}

	/**
	 * Gets a list of fieldInfos that match the given mark.
	 *
	 * @param  mark Requested mark for result fieldInfos.
	 * @return List of field infos that match the given mark, an empty list if no field info has
	 *         this mark.
	 */
	public LyInfoList<LyFieldInfo> getFieldInfos(int mark)
	{
		load();
		return _fieldInfos.getInfos(mark);
	}

	/**
	 * Gets list of fieldInfo that match one or all of the given marks.
	 *
	 * @param  marks    Requested marks, or null for the entire fieldInfo list.
	 * @param  matchAll Indicates if the fieldIndo must have at least one mark (false) or all of
	 *                  them (true). Any fieldInfo that match a mark from the mark list is returned
	 *                  if matchAll is false.
	 * @return The set of requested fieldInfos, an empty vector if no result is found.
	 */
	public LyInfoList<LyFieldInfo> getFieldInfos(BitSet marks, boolean matchAll)
	{
		load();
		return _fieldInfos.getInfos(marks, matchAll);
	}

	/**
	 * Fetches the action with the specified identifier.
	 *
	 * @param  id Identifier of the requested action
	 * @return The action identified by this id, or null if it could not be found.
	 */
	public LyAction getAction(String id)
	{
		load();
		return _actions.getInfo(id);
	}

	/**
	 * Fetches the list of actions which match (one or all of) the specified marks.
	 *
	 * @param  marks    The requested marks.
	 * @param  matchAll Indicates if the action must have at least one mark ('false') or all of them
	 *                  ('true').
	 * @return List of the requested actions, or an empty list if no action has these marks.
	 */
	public LyInfoList<LyAction> getActions(BitSet marks, boolean matchAll)
	{
		load();
		return _actions.getInfos(marks, matchAll);
	}

	/**
	 * Fetches the list of actions that match (or not match) the specified mark.
	 *
	 * @param  mark The requested mark.
	 * @param  not  To invert the search : if this parameter is set to 'true', the searched action
	 *              are those that don't have the given mark.
	 * @return List of the requested actions, or an empty list if no action have this mark (or if
	 *         all have it if not is set to 'true').
	 */
	public LyInfoList<LyAction> getActions(int mark, boolean not)
	{
		load();
		return _actions.getInfos(mark, not);
	}

	/**
	 * Fetches the list of actions that match the specified mark.
	 *
	 * @param  mark The requested mark.
	 * @return List of the requested actions, or an empty list if no action have this mark.
	 */
	public LyInfoList<LyAction> getActions(int mark)
	{
		load();
		return _actions.getInfos(mark, false);
	}

	/**
	 * Retrieve the default action for this classInfo. The default action is the first action
	 * specified on this class info with the DEFAULT mark or the first action of the classInfo if no
	 * action has this mark.
	 *
	 * @return Default action for this classInfo.
	 */
	public LyAction getDefaultAction()
	{
		LyInfoList<LyAction> defaultActions = _actions.getInfos(LyAction.MARK_DEFAULT);
		LyAction			 action;

		if ((defaultActions == null) || (defaultActions.getSize() == 0))
		{
			action = getActionWithId(LyAction.ID_CONSULT);

			if ((action != null) && action.hasMark(LyAction.MARK_HIDDEN))
				action = null;
		}
		else
			action = defaultActions.getInfo(0);

		return action;
	}

	/**
	 * Indicates if this class info has at least one action of the given "type". This means that an
	 * action of the class info has the given id or has a template action ancestor that has this id.
	 * This method is used to check if a type of action (for example, consultation) is enabled on
	 * this class.
	 *
	 * @param  actionId Action identifier that is tested on this class.
	 * @return The action that inherits an action whose id is the given parameter, null otherwise.
	 */
	public LyAction getActionWithId(String actionId)
	{
		// We get all action with id or template actionId
		LyInfoList<LyAction> actions = getActionsWithId(actionId);

		// we get the first action with a min equals to or lesser than zero
		int i = 0;
		LyAction action = null;
		while(action == null && i < actions.getSize())
		{
			if(actions.getInfo(i).getMin() <= 0)
				action = actions.getInfo(i);
			i++;
		}

		if(action != null)
			return action;

		if(actions.getSize() > 0)
			return actions.getInfo(0);

		return null;
	}

	/**
	 * Indicates if this class info has at least one action of the given "type". This means that an
	 * action of the class info has the given id or has a template action ancestor that has this id.
	 * This method is used to check if a type of action (for example, consultation) is enabled on
	 * this class.
	 *
	 * @param  actionId Action identifier that is tested on this class.
	 * @return The actions that inherits an action whose id is the given parameter, null otherwise.
	 */
	public LyInfoList<LyAction> getActionsWithId(String actionId)
	{
		if (actionId == null)
			return null;

		LyInfoList<LyAction> actions = getActions();
		LyInfoList<LyAction> result = new LyInfoList<LyAction>();
		int					 i, n = ((actions == null) ? 0 : actions.getSize());
		LyAction			 action, template;

		actionId = actionId.intern();

		for (i = 0; i < n; i++)
		{
			action = actions.getInfo(i);
			template = action;

			while ((template != null) && (template.getId() != actionId))
				template = (LyAction)template.getTemplate();

			if (template != null)
				result.addInfo(action);
		}

		return result;
	}

	/**
	 * Searchs on this class for an action that "inherits" of the given action.
	 *
	 * @param  action Searched action.
	 * @return The action that inherits the given action on this class, null if not found.
	 */
	public LyAction getRealAction(LyAction action)
	{
		if (action == null)
			return null;

		LyInfoList<LyAction> actions = getActions();
		int					 i, n = ((actions == null) ? 0 : actions.getSize());
		LyAction			 classAction, model;

		for (i = 0; i < n; i++)
		{
			classAction = actions.getInfo(i);
			if (classAction.isAbstract()
			        && isAbstract())
				continue;

			model = classAction;
			do
			{
				if (model == action)
					return classAction;
				else
					model = (LyAction)model.getTemplate();
			}
			while (model != null);
		}

		return null;
	} // end method getRealAction

	/**
	 * Gets the binding information. Generic or specific data providers use the binding information
	 * to fetch data from and to a data provider.
	 *
	 * @return Binding information for this class info.
	 */
	public LyClassBinding getBinding()
	{
		if (!isLoaded())
			return null;

		if (isVirtual())
			// Redirect on super-class binding
			return _superClass.getBinding();

		if ((_binding == null)
				&& !isLocal())
		{
			// Build a default binding.
			LyApplication application = getApplication();
			String		  defaultLocationId = application.getDefaultLocationId();

			_binding = new LyClassBinding(getId(), defaultLocationId, null);
            _binding.setDefaultBinding(true);
		}

		return _binding;
	}

	/**
	 * Gets cache policy for this class info. The cache policy determines how objects are loaded and
	 * unloaded in memory.
	 *
	 * @return Informations about the cache policy chosen for this class. Caches are built by
	 *         reflection from this informations.
	 */
	public LyInfoCache getCachePolicy()
	{
		return _cachePolicy;
	}

	/**
	 * Tests if the class is local (whether it has the mark MARK_LOCAL or not).
	 *
	 * @return True if this class is local, false otherwise.
	 */
	public boolean isLocal()
	{
        if (_local != null)
            return _local.booleanValue();

        // Check whether this field has LOCAL mark or not.
        boolean   local = (isAbstract() || hasMark(MARK_LOCAL));

        if (_locked)
            _local = Boolean.valueOf(local);
        return local;
	}

	/**
	 * Gets peer class of this class info. The peer class represents how "physical objects" are
	 * handled.
	 *
	 * @return Returns peer class of this class info.
	 */
	public final LyClass getPeerClass()
	{
		return (_binding == null ? null : _binding.getPeerClass());
	}

	/**
	 * Gets default sort for this classInfo.
	 *
	 * @return Default sort for this classInfo (may be null).
	 */
	public LySort getSort()
	{
		return _sort;
	}

	/**
	 * Returns the fieldInfo at the specified index in fields list.
	 *
	 * @param  index Index of the requested fieldInfo
	 * @return The searched fieldInfo if it exists, or null if not found.
	 */
	public LyFieldInfo getFieldInfo(int index)
	{
		return _fieldInfos.getInfo(index);
	}

	/**
	 * Gets index for the given fielInfo.
	 *
	 * @return Index of the field info in field infos list, 1 if this fieldInfo doesn't belong to
	 *         this classInfo.
	 */
	public int getIndex(LyFieldInfo fieldInfo)
	{
		if (_indexes == null)
		{
            load();

			int		    i, n = getFieldInfoCount();
			LyFieldInfo field;

			_indexes = new Hashtable<LyFieldInfo, Integer>();
			for (i = 0; i < n; i++)
			{
				field = getFieldInfo(i);
				_indexes.put(field, new Integer(i));
			}
		}

		Integer index = _indexes.get(fieldInfo);

		return ((index == null) ? -1 : index.intValue());
	}

	/**
	 * Reset the internal table of indexes used to speed up fieldInfo accesses.
	 */
	public void resetFieldIndexes()
	{
		_indexes = null;
		initMethodTables();
	}

	/**
	 * Checks if the given fieldInfo is contained in this classInfo.
	 *
	 * @param  fieldInfo Checked fieldInfo.
	 * @return True if the given fieldInfo is found, false otherwise.
	 */
	public boolean contains(LyFieldInfo fieldInfo)
	{
		return (getIndex(fieldInfo) >= 0);
	}

	/**
	 * Method to check the given field info and return in some cases the real field info.
	 */
	public LyFieldInfo getRealFieldInfo(LyFieldInfo fieldInfo)
	{
		return getRealFieldInfo(fieldInfo, true);
	}

	/**
	 * Internal method to check the given field info and return in some cases the real field info.
	 */
	public LyFieldInfo getRealFieldInfo(LyFieldInfo fieldInfo, boolean exception)
	{
		// Check first that this fieldInfo is contained in our classInfo
		int index = getIndex(fieldInfo);

		if (index >= 0)
			return getFieldInfo(index);

		if (fieldInfo instanceof LyRelationFieldInfo)
		{
			// Maybe this relation is an inherited relation in the classInfo
			int				    i, n = getFieldInfoCount();
			LyFieldInfo		    field;
			LyRelationFieldInfo relation, extendedRelation;

			for (i = 0; i < n; i++)
			{
				field = getFieldInfo(i);
				if ((field instanceof LyRelationFieldInfo)
						&& (field.getClass() == fieldInfo.getClass()))
				{
					relation = (LyRelationFieldInfo)field;
					extendedRelation = relation.getExtendedRelation();
					if ((fieldInfo == extendedRelation)
							|| relation.hasSubType((LyRelationFieldInfo)fieldInfo))
					{
						_indexes.put(fieldInfo, new Integer(getIndex(field)));
						return field;
					}
				}
			}
		}

		// Allow access by type
		int		    i, n = getFieldInfoCount();
		LyFieldInfo field, realField;

		for (i = 0; i < n; i++)
		{
			realField = getFieldInfo(i);
			if (realField.getClass() != fieldInfo.getClass())
				continue;

			field = realField;
			while ((field != fieldInfo)
					&& (field != null))
				field = (LyFieldInfo)field.getTemplate();

			if (field == fieldInfo)
			{
				_indexes.put(fieldInfo, new Integer(getIndex(realField)));
				return realField;
			}
		}

		if (!fieldInfo.isLocal())
		{
			if (exception)
				throw new IllegalArgumentException("BF AppEngine : field info "
					+ fieldInfo + " is not in the classInfo " + this);
			return null;
		}

		// Try to get template for virtual classes
		LyFieldInfo template = (LyFieldInfo)fieldInfo.getTemplate();

		if ((template != null) && contains(template))
			return template;
		else
			return fieldInfo;
	} // end method getRealFieldInfo

	/**
	 * Returns the number of fieldInfos for this classInfo.
	 *
	 * @return Number of fieldInfos for this class.
	 */
	public int getFieldInfoCount()
	{
		load();
		return _fieldInfos.getSize();
	}

	/**
	 * Fetches the action at the specified index in actions list.
	 *
	 * @param  index Index of the searched action.
	 * @return Action at the given index if it's found, null if it's not found.
	 */
	public LyAction getAction(int index)
	{
		load();
		return _actions.getInfo(index);
	}

	/**
	 * Returns the number of action supported on this classInfo.
	 *
	 * @return Number of actions on this classInfo.
	 */
	public int getActionCount()
	{
		load();
		return _actions.getSize();
	}

	/**
	 * Returns the set of actions available on this class info.
	 *
	 * @return Actions available on this class info, or an empty list.
	 */
	public LyInfoList<LyAction> getActions()
	{
		load();
		return (_actions == null ? null : new LyInfoList<LyAction>(_actions));
	}

	/**
	 * Checks if the given action is contained in this classInfo.
	 *
	 * @param  action Action that is searched on this classInfo.
	 * @return True if the given action is found, false otherwise.
	 */
	public boolean contains(LyAction action)
	{
		load();
		return _actions.contains(action);
	}

	/**
	 * Return the number of routes in the class.
	 *
	 * @return Number of routes in the class.
	 */
	public int getRoutesCount()
	{
		load();
		return (_routes == null ? 0 : _routes.getSize());
	}

	/**
	 * Fetches the route at the specified index.
	 *
	 * @param  index Index of the searched route.
	 * @return The requested route if found, or null if it could not be found.
	 */
	public LyRoute getRoute(int index)
	{
		load();
		return (_routes == null ? null : _routes.getInfo(index));
	}

	/**
	 * Fetches the route with the specified id.
	 *
	 * @param  id Id of the searched route.
	 * @return The requested route if found, or null if it could not be found.
	 */
	public LyRoute getRoute(String id)
	{
		load();
		return (_routes == null ? null : _routes.getInfo(id));
	}

	/**
	 * Returns the set of routes defined on this class info. This is a COPY of the internal list,
	 * therefore this list may be freely manipulated by the application.
	 *
	 * @return The set of routes, or an empty list.
	 */
	public LyInfoList<LyRoute> getRoutes()
	{
		// Since routes list can't be locked because routes are dynamically added,
		// a copy is made to avoid manipulation of the internal list.
		load();
		return (_routes == null ? null : new LyInfoList<LyRoute>(_routes));
	}

	/**
	 * Returns a set of routes between this class info and the specified class. Routes are used to
	 * calculate cross references beetween two classes.
	 *
	 * @param  classInfo Class info with which routes are searched.
	 * @return List of routes between this class info and the given class or null if there is no
	 *         available route.
	 */
	public LyInfoList<LyRoute> getRoutes(LyClassInfo classInfo)
	{
		if (classInfo == null)
			throw new IllegalArgumentException("BF AppEngine : classInfo can't be null");

		LyInfoList<LyRoute> result = null;
		int				    i, n = getRoutesCount();
		LyRoute			    route;

		for (i = 0; i < n; i++)
		{
			route = getRoute(i);
			if ((route.getTarget() == classInfo)
					&& !route.isHidden())
			{
				if (result == null)
					result = new LyInfoList<LyRoute>();
				result.addInfo(route);
			}
		}

		if (result == null)
		{
			// No route defined : try to build one with direct relations.
			LyFieldInfo			             field, subField;
			LyRelationFieldInfo              relation;
            LyRelationFieldInfo.Direction    direction;
			LyInfoList<LyFieldInfo>          relations = new LyInfoList<LyFieldInfo>();

			// First we try in the forward direction.
			for (i = 0, n = getFieldInfoCount(); i < n; i++)
			{
				field = getFieldInfo(i);
				if (field instanceof LyRelationFieldInfo)
					relation = (LyRelationFieldInfo)field;
				else if (field instanceof LyArrayFieldInfo)
				{
					subField = ((LyArrayFieldInfo)field).getFieldInfo();
					if (subField instanceof LyRelationFieldInfo)
						relation = (LyRelationFieldInfo)subField;
					else
						continue;
				}
				else
					continue;

				direction = relation.getDirection();
				if (((direction == LyRelationFieldInfo.Direction.FORWARD)
							|| (direction == LyRelationFieldInfo.Direction.BOTH))
						&& relation.relatedWith(classInfo))
					// Relation is ok to build a route.
					relations.addInfo(field);
			}

			if (relations.getSize() == 0)
			{
				// No relation found : we try from the target classInfo with
				// backward relations.
				n = classInfo.getFieldInfoCount();
				for (i = 0; i < n; i++)
				{
					field = classInfo.getFieldInfo(i);
					if (field instanceof LyRelationFieldInfo)
						relation = (LyRelationFieldInfo)field;
					else if (field instanceof LyArrayFieldInfo)
					{
						subField = ((LyArrayFieldInfo)field).getFieldInfo();
						if (subField instanceof LyRelationFieldInfo)
							relation = (LyRelationFieldInfo)subField;
						else
							continue;
					}
					else
						continue;

					direction = relation.getDirection();
					if (((direction == LyRelationFieldInfo.Direction.BACKWARD)
								|| (direction == LyRelationFieldInfo.Direction.BOTH))
							&& relation.relatedWith(this))
						// Relation is ok to build a route.
						relations.addInfo(field);
				}
			}

			if (relations.getSize() != 0)
			{
				// We can now build a route.
				LyApplication application = getApplication();
				LyStep		  step;

				step = new LyStep(classInfo, relations, null);
				step.setApplication(application);
				route = new LyRoute(null, classInfo, null);
				route.setGenerated(true);
				route.setApplication(application);
				route.addStep(step);
				route.check(this);
				addRoute(route, -1);

				if (result == null)
					result = new LyInfoList<LyRoute>();
				result.addInfo(route);
			}
		}

		return result;
	} // end method getRoutes

	/**
	 * Check if this class info has object names : if object name and object id are different for
	 * instances of this class info.
	 *
	 * @return True if they are different, false otherwise.
	 */
	public boolean isNameDifferentOfId()
	{
		// if it has not been locked then compute if the name is different of id
		if (!_locked)
			computeNameDifferentOfId();

		return _nameDifferentOfId;
	}

	/**
	 * Check if for class info, the name is sub-part of the id.
	 *
	 * @return True if name is included in id, false otherwise.
	 */
	public boolean isNamePartOfId()
	{
		// if it has not been locked then compute if the name is different of id
		if (!_locked)
			computeNameDifferentOfId();

		return _namePartOfId;
	}

	/**
	 * Return the number of controls associated to this class.
	 *
	 * @return Number of controls of this class.
	 */
	public int getControlCount()
	{
		load();
		return ((_controls == null) ? 0 : _controls.getSize());
	}

	/**
	 * Fetches the control at the specified index.
	 *
	 * @param  index Index of the searched control.
	 * @return The requested control if found, or null if it could not be found.
	 */
	public LyControl getControl(int index)
	{
		load();
		return ((_controls == null) ? null : _controls.getInfo(index));
	}

	/**
	 * Returns the set of controls defined on this class info. This is a COPY of the internal list,
	 * therefore this list may be freely manipulated by the application.
	 *
	 * @return The set of controls, or an empty list.
	 */
	public LyInfoList<LyControl> getControls()
	{
		return (_controls == null ? null : new LyInfoList<LyControl>(_controls));
	}

	/**
	 * Gets a rule from the rules list at a given position.
	 *
	 * @param  index Position of the searched rule. If index is negative (-1) return null
	 * @return Rule at the given index if it exists.
	 */
	public LyRule getRule(short index)
	{
		return ((_rules == null) ? null : _rules.getInfo(index));
	}

	/**
	 * Gets rules count.
	 *
	 * @return Return the number of rules in this class info.
	 */
	public short getRuleCount()
	{
		return ((_rules == null) ? 0 : (short)_rules.getSize());
	}

	/**
	 * Returns the list of rules defined on this class info.
	 *
	 * @return List of rules, or null if no rule was registered on this class info.
	 */
	public LyInfoList<LyRule> getRules()
	{
		return (_rules == null ? null : new LyInfoList<LyRule>(_rules));
	}

	/**
	 * Checks if this classInfo results from a struct definition.
	 *
	 * @return True if this classInfo was created from a struct definition.
	 */
	public boolean fromStruct()
	{
		return (getParent() instanceof LyStructFieldInfo);
	}

	/**
	 * Checks if this classInfo results from an array definition.
	 *
	 * @return True if this classInfo was created from an array definition.
	 */
	public boolean fromArray()
	{
		return (getParent() instanceof LyArrayFieldInfo);
	}

	/**
	 * Retuns the tip (short help displayed in tool tip) associated to this instance. Overridden to
	 * use labels instead of tip when defined.
	 *
	 * @return Tip of the instance, or null if no tip was defined.
	 */
	public String getTip()
	{
		LyInfoLabel label = getLabel(LyInfoLabel.Type.GRAMMATICAL);

		return ((label == null) ? super.getTip() : label.getString(this));
	}

	/**
	 * Returns the real class corresponding to this class info.
	 *
	 * @return The first non virtual ancestor of this class info.
	 */
	public LyClassInfo getRealClass()
	{
		return ((!_virtual) ? this : _superClass.getRealClass());
	}

	/**
	 * Returns the field info corresponding to given one in this class info.
	 *
	 * @param  fieldInfo Real field info.
	 * @return The first virtual field info that has the given field info as template.
	 */
	public LyFieldInfo getVirtualFieldInfo(LyFieldInfo fieldInfo)
	{
		if (!_virtual)
			return null;
		if (contains(fieldInfo))
			return fieldInfo;

		LyFieldInfo field;
		int		    i, n = _fieldInfos.getSize();

		for (i = 0; i < n; i++)
		{
			field = _fieldInfos.getInfo(i);
			if (field.isLocal()
					&& field.hasTemplate(fieldInfo))
				return field;
		}

		return null;
	}

	/**
	 * Sets the super class of this class info.
	 *
	 * @param superClass Super class of this class info (may be null.
	 */
	void setSuperClass(LyClassInfo superClass)
	{
		if ((superClass != null)
				&& ((isTemporary() && (!superClass.isTemporary()))
					|| (!isTemporary() && (superClass.isTemporary()))))
			return;
		_superClass = superClass;

		// Add this as subclass of its superClass
		if (superClass != null)
			superClass.addSubClass(this);
	}

	/**
	 * Add a subclass of this class info.
	 *
	 * @param superClass Super class of this class info (may be null.
	 */
	void addSubClass(LyClassInfo subClass)
	{
		if (subClass.isVirtual())
			return;

		if (_subClasses == null)
			_subClasses = new LyInfoList<LyClassInfo>(false);

		_subClasses.addInfo(subClass);
	}

	/**
	 * Sets the binding for this class info.
	 *
	 * @param binding Binding for this class info.
	 */
	void setBinding(LyClassBinding binding)
	{
		if (isVirtual())
			// Can't set binding on a virtual class
			return;

		_binding = binding;
		if (binding != null)
		    binding.setParent(this);
	}

    /**
     * Sets if this class info is virtual or not.
     *
     * @param virtual True if this class info is virtual, false otherwise.
     */
    protected void setVirtual(boolean virtual)
    {
        _virtual = virtual;
    }

	/**
	 * Sets cache policy.
	 *
	 * @param cache Informations about the cache policy for this class.
	 */
	public void setCachePolicy(LyInfoCache cache)
	{
		_cachePolicy = cache;
	}

	/**
	 * Sets the default sort for this classInfo.
	 *
	 * @param sort Default sort.
	 */
	void setSort(LySort sort)
	{
		_sort = sort;
	}

	/**
	 * Sets the <code>_checkFieldInsertion</code> parameter.
	 *
	 * @param checkFieldInsertion true if we must check the existence of the template when inserting
	 * a field info in current class info, false otherwise.
	 */
	public void setCheckFieldInsertion(boolean checkFieldInsertion)
	{
		_checkFieldInsertion = checkFieldInsertion;
	}

	/**
	 * Adds a fieldInfo to this classInfo at the last position.
	 *
	 * @param fieldInfo FieldInfo to add in this classInfo.
	 */
	void addFieldInfo(LyFieldInfo fieldInfo)
	{
		addFieldInfo(fieldInfo, -1);
	}

	/**
	 * Adds a fieldInfo to this classInfo at a given position.
	 *
	 * @param fieldInfo FieldInfo to add in this classInfo.
	 * @param position  Requested position in the fields list.
	 */
	void addFieldInfo(LyFieldInfo fieldInfo, int position)
	{
		if (_checkFieldInsertion
				&& (isVirtual()
						|| (_superClass != null)))
		{
			LyFieldInfo template = (LyFieldInfo)fieldInfo.getTemplate();

			if ((template != null)
		        && contains(template)
		        && _superClass.contains(template))
			{
				if (isVirtual() && (position < 0))
					position = _superClass.getIndex(template);
				removeFieldInfo(template, false);
			}
		}

		_fieldInfos.removeInfo(fieldInfo);
		_fieldInfos.addInfo(fieldInfo, position);
		if (fieldInfo.hasMark(LyFieldInfo.MARK_ID))
		{
			_idFields = null;
			_nameFields = null;
		}

		if (fieldInfo.hasMark(LyFieldInfo.MARK_NAME))
			_nameFields = null;

		if (fieldInfo instanceof LyRelationFieldInfo)
		{
			if (fieldInfo instanceof LyStructFieldInfo)
				// Add this class info as parent of the relation
				((LyStructFieldInfo)fieldInfo).addParentClass(this);

			LyRelationFieldInfo extendedRelation = ((LyRelationFieldInfo)fieldInfo)
					.getExtendedRelation();

			if (_checkFieldInsertion && (extendedRelation != null))
			{
				// Do not use removeFieldInfo method because it does a control
				// on inherited fields.
				if ((contains(extendedRelation))
						&& ((_superClass == null)
							|| ((_superClass != null)
								&& !(_superClass.contains(extendedRelation)))))
					throw new IllegalStateException("BF AppEngine : Can't have a relation "
						+ fieldInfo + " and its parent relation "
						+ extendedRelation + " in the same classInfo " + this);

				if (contains(extendedRelation))
				{
					_fieldInfos.removeInfo(extendedRelation);
					if (extendedRelation.hasMark(LyFieldInfo.MARK_ID))
					{
						_idFields = null;
						_nameFields = null;
					}

					if (extendedRelation.hasMark(LyFieldInfo.MARK_NAME))
						_nameFields = null;

					if (extendedRelation instanceof LyStructFieldInfo)
						((LyStructFieldInfo)extendedRelation).removeParentClass(this);
				}
			}
		}
		else if (fieldInfo instanceof LyArrayFieldInfo)
		{
			LyFieldInfo subField = ((LyArrayFieldInfo)fieldInfo).getFieldInfo();

			if (subField instanceof LyStructFieldInfo)
				((LyStructFieldInfo)subField).addParentClass(this, fieldInfo);
		}

		// Force recomputing of indexes and methodTables
		resetFieldIndexes();
	} // end method addFieldInfo

	/**
	 * Removes a fieldInfo from this classInfo.
	 *
	 * @param fieldInfo FieldInfo to remove from this classInfo.
	 */
	void removeFieldInfo(LyFieldInfo fieldInfo)
	{
		removeFieldInfo(fieldInfo, true);
	}

	/**
	 * Removes a fieldInfo from this classInfo.
	 *
	 * @param fieldInfo FieldInfo to remove from this classInfo.
	 */
	void removeFieldInfo(LyFieldInfo fieldInfo, boolean check)
	{
		if (check
				&& (_superClass != null)
				&& _superClass.contains(fieldInfo))
			throw new IllegalStateException("BF AppEngine : can't remove fieldInfo " + fieldInfo
				+ " on class info " + this + ", because it was defined on its super-class.");

		if ((_fieldInfos.removeInfo(fieldInfo) < 0)
			&& (fieldInfo.getTemplate() != null))
		{
			removeFieldInfo((LyFieldInfo)fieldInfo.getTemplate(), check);
			return;
		}

		if (fieldInfo.hasMark(LyFieldInfo.MARK_ID))
		{
			_idFields = null;
			_nameFields = null;
		}

		if (fieldInfo.hasMark(LyFieldInfo.MARK_NAME))
			_nameFields = null;

		// Force recomputing of indexes and methodTables
		resetFieldIndexes();

		if (fieldInfo instanceof LyStructFieldInfo)
			((LyStructFieldInfo)fieldInfo).removeParentClass(this);
		else if (fieldInfo instanceof LyArrayFieldInfo)
		{
			LyFieldInfo subField = ((LyArrayFieldInfo)fieldInfo).getFieldInfo();

			if (subField instanceof LyStructFieldInfo)
				((LyStructFieldInfo)subField).removeParentClass(this, fieldInfo);
		}
	} // end method removeFieldInfo

	/**
	 * Moves a fieldInfo to this classInfo at a given position.
	 *
	 * @param fieldInfo FieldInfo to move in this classInfo.
	 * @param position  Requested position in the fields list.
	 */
	void moveFieldInfo(LyFieldInfo fieldInfo, int position)
	{
		if (!contains(fieldInfo))
			throw new IllegalArgumentException("BF AppEngine : fieldInfo "
				+ fieldInfo + " was not declared in classInfo "
				+ this + " in your data model");

		_fieldInfos.removeInfo(fieldInfo);
		_fieldInfos.addInfo(fieldInfo, position);

		// Force recomputing of indexes and methodTables
		resetFieldIndexes();
	}

	/**
	 * Moves a fieldInfo to this classInfo at a given position.
	 *
	 * @param fieldInfo FieldInfo to move in this classInfo.
	 * @param position  Requested position in the fields list.
	 */
	void moveActionInfo(LyAction action, int position)
	{
		if (!contains(action))
			throw new IllegalArgumentException("BF AppEngine : action "
				+ action + " was not declared in classInfo "
				+ this + " in your data model");

		_actions.removeInfo(action);
		_actions.addInfo(action, position);
	}

	/**
	 * Resets field info list (clear all field infos).
	 */
	void resetFieldInfos()
	{
		if ((_superClass != null)
				&& !isVirtual())
			throw new IllegalStateException("BF AppEngine : can't reset fieldInfos on class info "
				+ this + ", because some of these fields were defined on its super-class.");

		int		    i, n = getFieldInfoCount();
		LyFieldInfo field, subField;

		for (i = 0; i < n; i++)
		{
			field = getFieldInfo(i);
			if (field instanceof LyStructFieldInfo)
				((LyStructFieldInfo)field).removeParentClass(this);
			else if (field instanceof LyArrayFieldInfo)
			{
				subField = ((LyArrayFieldInfo)field).getFieldInfo();
				if (subField instanceof LyStructFieldInfo)
					((LyStructFieldInfo)subField).removeParentClass(this, field);
			}
		}

		_fieldInfos.resetInfos();
		_idFields = null;
		_nameFields = null;
		_indexes = null;
	} // end method resetFieldInfos

	/**
	 * Adds a new action to the actions list available on this class info.
	 *
	 * @param action Action that should be added
	 */
	void addAction(LyAction action)
	{
		addAction(action, -1);
	}

	/**
	 * Adds a new action to the actions list available on this class info.
	 *
	 * @param action Action that should be added
	 */
	void addAction(LyAction action, int position)
	{
        LyAction    template = (LyAction)action.getTemplate();

		if (_virtual)
		{
			LyClassInfo realClass = getRealClass();

			if ((template != null)
					&& realClass.contains(template))
			{
				if (position < 0)
					position = _actions.getIndex(template);
				_actions.removeInfo(template);
			}

			if (action.isActionRef())
			{
				action.addParameterDecl(new LyParameterDecl(LyAction.PARAM_CLASS,
                    LyParameterDecl.Type.CLASS, true, null));
				action.addParameter(new LyParameter(LyAction.PARAM_CLASS, this));
			}
		}

        if ((template != null)
            && template.isAbstract()
            && contains(template))
        {
            // If template is abstract and herited for this class info, get rid of template action
            if (position < 0)
                position = _actions.getIndex(template);
            _actions.removeInfo(template);
        }

		_actions.addInfo(action, position);
	}

	/**
	 * Removes the specified action from actions list of this class info.
	 *
	 * @param action Action to remove from actions list.
	 */
	void removeAction(LyAction action)
	{
		_actions.removeInfo(action);
	}

	/**
	 * Resets actions list (clear all actions).
	 */
	void resetActions()
	{
		if ((_superClass != null)
				&& !isVirtual())
			throw new IllegalStateException("BF AppEngine : can't reset actions on class info "
				+ this + ", because some of these actions were defined on its super-class.");

		_actions.resetInfos();
	}

	/**
	 * Adds a new route to routes list of this class info.
	 *
	 * @param route    Route to add to routes list.
	 * @param position Position where to add the route in the list.
	 */
	void addRoute(LyRoute route, int position)
	{
	    if (_routes == null)
	        _routes = new LyInfoList<LyRoute>(false);
		_routes.addInfo(route, position);
	}

	/**
	 * Removes the specified route from routes list of this class info.
	 *
	 * @param route Route to remove from routes list.
	 */
	void removeRoute(LyRoute route)
	{
	    if (_routes != null)
	        _routes.removeInfo(route);
	}

	/**
	 * Resets routes list (clear all routes).
	 */
	void resetRoutes()
	{
	    if (_routes != null)
	        _routes.resetInfos();
	}

	/**
	 * Sets behavior class name.
	 *
	 * @param behaviorClassName Java class name that defines default behaviors for this class
	 */
	void setBehaviorClassName(String behaviorClassName)
	{
		_behaviorClassName = (behaviorClassName == null ? null
		    : new LyJavaName(this, behaviorClassName));
	}

	/**
	 * Adds a new control to controls list of this class info.
	 *
	 * @param control  Control to add.
	 * @param position Position where to add the route in the list.
	 */
	void addControl(LyControl control)
	{
		if (_controls == null)
			_controls = new LyInfoList<LyControl>(false);
		_controls.addInfo(control);
	}

	/**
	 * Adds a rule on this class info.
	 *
	 * @param rule Rule to add this class info.
	 */
	void addRule(LyRule rule)
	{
		if (_rules == null)
			_rules = new LyInfoList<LyRule>(false);
		_rules.addInfo(rule);
	}

	/**
	 * Add a new label to labels list of this class info.
	 *
	 * @param label Label to add.
	 */
	void addLabel(LyInfoLabel label)
	{
		if (label == null)
			throw new IllegalArgumentException("BF AppEngine : label can't be null");

        LyInfoLabel.Type type = label.getType();

		if (type == LyInfoLabel.Type.USER)
		{
			if (_userLabels == null)
				_userLabels = new Vector<LyInfoLabel>(1);
			_userLabels.add(label);
			return;
		}

		if ((type != LyInfoLabel.Type.ID)
				&& (type != LyInfoLabel.Type.NAME)
				&& (type != LyInfoLabel.Type.GRAMMATICAL)
				&& (type != LyInfoLabel.Type.SEMANTIC)
				&& (type != LyInfoLabel.Type.VALUE)
				&& (type != LyInfoLabel.Type.SHORT_VALUE)
				&& (type != LyInfoLabel.Type.USER))
			throw new IllegalArgumentException("BF AppEngine : invalid type of label on classInfo "
				+ this + ". Expecting USER, ID, NAME, GRAMMATICAL or SEMANTIC for value='"
				+ label.getValue() + "'.");

		if (_labels == null)
			_labels = new LyInfoLabel[6];
		_labels[type.ordinal()] = label;
	} // end method addLabel

	/**
	 * Returns the number of labels of this classInfo.
	 */
	public int getLabelCount()
	{
		return ((_labels == null) ? 0 : _labels.length);
	}

	/**
	 * Returns the label of this classInfo for the specified type.
	 */
	public LyInfoLabel getLabel(LyInfoLabel.Type type)
	{
		if (_labels == null)
			return null;

		return _labels[type.ordinal()];
	}

    /**
     * Returns the label of this classInfo at the given index.
     */
    public LyInfoLabel getLabel(int index)
    {
        if ((_labels == null)
                || (index < 0)
                || (index > _labels.length))
            return null;

        return _labels[index];
    }

	/**
	 * Returns the number of labels of this classInfo.
	 */
	public int getUserLabelCount()
	{
		return ((_userLabels == null) ? 0 : _userLabels.size());
	}

	/**
	 * Returns all labels of this classInfo.
	 */
	public LyInfoLabel[] getLabels()
	{
		LyInfoLabel[] result = _labels;
		int			  nu = (_userLabels != null ? _userLabels.size() : 0);

		if (nu > 0)
		{
			int		i, nl = (_labels != null ? _labels.length : 0);

			result = new LyInfoLabel[nl + nu];

			for (i = 0; i < nl; i++)
				result[i] = _labels[i];

			for (i = 0; i < nu; i++)
				result[i + nl] = _userLabels.get(i);
		}

		return result;
	}

	/**
	 * Returns the user label at the given position.
	 */
	public LyInfoLabel getUserLabel(int i)
	{
		return ((_userLabels == null) ? null : (LyInfoLabel)_userLabels.elementAt(i));
	}

	/**
	 * Returns the user label for the given subType.
	 */
	public LyInfoLabel getUserLabel(String subType)
	{
		int		    i, n = getUserLabelCount();
		LyInfoLabel label;

		if (subType != null)
			subType = subType.intern();
		for (i = 0; i < n; i++)
		{
			label = _userLabels.elementAt(i);
			if (label.getSubType() == subType)
				return label;
		}

		return null;
	}

	/**
	 * Recompute the object id of the specified object.
	 */
	public void rehash(Object object)
	{
		LyObjectSource source = getObjectSource();

		source.rehash((LyObject)object);
	}

	/**
	 * Sets the application for this info. Once assigned, the application can be changed.
	 *
	 * @param     application The application to which the info belongs.
	 * @exception IllegalStateException If the info already belongs to an application.
	 */
	public void setApplication(LyApplication application)
	{
		super.setApplication(application);

		LyEnvironment environment = getEnvironment();

		if (!environment.isResourceDefined("LY_ALLOW_NON_ABSTRACT_INHERITED_CLASSES"))
			_allowNonAbstractInheritedClasses = null;
		else
			_allowNonAbstractInheritedClasses = environment.testResource("LY_ALLOW_NON_ABSTRACT_INHERITED_CLASSES");
	}

	/**
	 * Sets the abstract property of this classInfo.
	 *
	 * @param abstractClass Abstract property.
	 */
	void setAbstract(boolean abstractClass)
	{
		_abstract = abstractClass;
	}

	/**
	 * Gets the next value of the internal counter.
	 *
	 * @return The next available value of the counter.
	 */
	public int getNextCount()
	{
		return _counter++;
	}

	/**
	 * Accessor to the marks attached to this instance (presented as a bitset).
	 *
	 * @return A copy of the bitset which represents the marks (properties) of this instance.
	 * @see    java.util.BitSet
	 */
	protected BitSet getMarks(int parts)
	{
		return getMarks(_marks, parts);
	}

	/**
	 * Accessor to the marks attached to this instance (presented as a bitset).
	 *
	 * @return A copy of the bitset which represents the marks (properties) of this instance.
	 * @see    java.util.BitSet
	 */
	public static BitSet getMarks(BitSet marks, int parts)
	{
		if (marks == null)
			return null;

		BitSet newMarks = (BitSet)marks.clone();

		if ((parts & PART_GUI) > 0)
		{
			// Only accept gui marks
			newMarks.clear(MARK_LOCAL);
			newMarks.clear(MARK_LOAD);
		}
		if ((parts & PART_DATA) > 0)
		{
			// Only accept data marks
			newMarks.clear(MARK_LOCAL);
			newMarks.clear(MARK_LOAD);
		}
		if ((parts & PART_BIND) > 0)
		{
			// Only accept bind marks
			BitSet validMarks = new BitSet();

			validMarks.set(MARK_LOCAL);
			validMarks.set(MARK_LOAD);
			newMarks.and(validMarks);
		}

		return newMarks;
	} // end method getMarks

	/**
	 * Find fieldinfos from the peer value (connector value defined in class binding).
	 * Since multiple fields may be bound to the same data source binding, this method may return a list.
	 * @param  binding the binding to search
	 * @return the requested fieldInfos, an empty list if none is found or null if parameter is null
	 */
	public LyInfoList<LyFieldInfo> getfieldInfoFromBinding(String binding)
	{
		if (binding == null)
			return null;

		LyFieldInfo 			fieldInfo;
		LyFieldBinding			fieldBinding;
		LyInfoList<LyFieldInfo> fieldInfoList = getFieldInfos();
		LyInfoList<LyFieldInfo> result = new LyInfoList<LyFieldInfo>();

		for (int i=0; (fieldInfoList != null) && (i<fieldInfoList.getSize()); i++)
		{
			fieldInfo = fieldInfoList.getInfo(i);
			fieldBinding = fieldInfo.getBinding();

			if ((fieldBinding != null) && (binding.equals(fieldBinding.getBinding())))
			{
				result.addInfo(fieldInfo);
			}
		}
		return result;
	}

	/**
	 * Find the connectors for a classinfo
	 * Usually there is only one connector but some classes may have multiples (with unions)
	 * or none if not bound to any datasource
	 * @return the list of peer providers
	 */
	public List<LyPeerDataProvider> getPeerDataProviders()
	{
		LyClass peerClass = getPeerClass();

		return getPeerDataProviders(peerClass);
	}

	/**
	 * Find the connectors for a classinfo
	 * Usually there is only one connector but some classes may have multiples (with unions)
	 * or none if not bound to any datasource
	 * @param peerClass current peerclass being searched. May be different from getPeerClass for unionClasses
	 * @return the list of peer providers
	 */
	protected List<LyPeerDataProvider> getPeerDataProviders(LyClass peerClass)
	{
		List<LyPeerDataProvider> result = new ArrayList<LyPeerDataProvider>();

		if (peerClass == null)
			return result;
		else if (peerClass instanceof LyUnionClass)
		{
			LyUnionClass unionClass = (LyUnionClass)peerClass;

			LyClassBinding classBinding1 = unionClass.getUnionBinding().getClassBinding1();
			LyClassBinding classBinding2 = unionClass.getUnionBinding().getClassBinding2();

			List<LyPeerDataProvider> result1 = getPeerDataProviders(classBinding1.getPeerClass());
			List<LyPeerDataProvider> result2 = getPeerDataProviders(classBinding2.getPeerClass());

			result.addAll(result1);
			for (LyPeerDataProvider res : result2)
			{
				if (!result.contains(res))
					result.add(res);
			}
		}
		else
		{
			LyDataSource	    dataSource = peerClass.getDataSource();

			if (dataSource == null)
				return result;

			LyDataProvider      dataProvider = dataSource.getDataProvider();

			if (dataProvider instanceof LyPeerDataProvider)
				result.add((LyPeerDataProvider)dataProvider);
		}
		return result;
	}

    /**
     * Sets list of field infos of this class info.
     */
    protected void setFieldInfos(LyInfoList<LyFieldInfo> fieldInfos)
    {
        LyInfoList<LyFieldInfo>     oldFieldInfos = _fieldInfos;
        LyFieldInfo                 fieldInfo;
        int                         i, n = (fieldInfos == null ? 0 : fieldInfos.getSize());
        List<LyPeerDataProvider>    providers = getPeerDataProviders();
        LyPeerDataProvider          provider;
        int                         j, np = (providers == null ? 0 : providers.size());
        boolean                     unload = false;

        _fieldInfos = new LyInfoList<LyFieldInfo>(false);

        for (i = 0; i < n; i++)
        {
            fieldInfo = fieldInfos.getInfo(i);
            addFieldInfo(fieldInfo, -1);

            if ((oldFieldInfos != null)
                && !oldFieldInfos.contains(fieldInfo))
            {
                for (j = 0; j < np; j++)
                {
                    provider = providers.get(j);
                    provider.validate(this, fieldInfo);
                    unload = true;
                }
            }
        }

        if (oldFieldInfos != null)
        {
            for (i = 0, n = oldFieldInfos.getSize(); i < n; i++)
            {
                fieldInfo = oldFieldInfos.getInfo(i);
                if ((fieldInfos != null)
                    && !fieldInfos.contains(fieldInfo))
                {
                    for (j = 0; j < np; j++)
                    {
                        provider = providers.get(j);
                        provider.invalidate(this, fieldInfo);
                        unload = true;
                    }
                }
            }
        }

        if (unload)
        {
            LyObjectSource  source = getObjectSource();

            if(source != null)
            	source.unload(false);
        }
        else if (np == 0)
        {
            if (_application.isValid())
                try
                {
                    _application.loadClasses(new LyInfoList<LyClassInfo>(this));
                }
                catch (Exception e)
                {
                }
        }

        _indexes = null;
    }

    /**
     * Sets the list of actions available on this class info.
     */
    protected void setActions(LyInfoList<LyAction> actions)
    {
        _actions = new LyInfoList<LyAction>(false);

        int     i, n = (actions == null ? 0 : actions.getSize());

        for (i = 0; i < n; i++)
            addAction(actions.getInfo(i), -1);
    }

    /**
     * Sets java class name corresponding to this class info when java code generation is used.
     */
    protected void setJavaName(String javaName)
    {
        if (_virtual
            && (_superClass != null))
        {
            if (javaName != null)
                getEnvironment().logWarning("BF AppEngine : javaName " + javaName
                    + " ignored on virtual class info " + this);

            _javaName = _superClass._javaName;
        }
        else
            _javaName = (javaName == null ? null : new LyJavaName(this, javaName));
    }

    /**
     * Sets the list of routes defined on this class info.
     */
    protected void setRoutes(LyInfoList<LyRoute> routes)
    {
        int     i, n = (routes == null ? 0 : routes.getSize());

        _routes = null;
        for (i = 0; i < n; i++)
            addRoute(routes.getInfo(i), -1);
    }

    /**
     * Accessor to the generic marks attached to this instance.
     */
    public BitSet getGenericMarks()
    {
        return (new LyBitSet(LyBitSet.Type.C_MARKS, _marks)).getGenericMarks();
    }

    /**
     * Sets the generic marks attached to this instance.
     */
    protected void setGenericMarks(BitSet marks)
    {
        _marks = (new LyBitSet(LyBitSet.Type.C_MARKS, _marks)).setGenericMarks(marks);
    }

    /**
     * Accessor to the marks attached to this instance (presented as an array of strings).
     */
    public String[] getSpecificMarks()
    {
        return (new LyBitSet(LyBitSet.Type.C_MARKS, _marks)).getSpecificMarks();
    }

    /**
     * Sets the specific marks attached to this instance (presented as an array of strings).
     */
    protected void setSpecificMarks(String[] marks)
    {
        _marks = (new LyBitSet(LyBitSet.Type.C_MARKS, _marks)).setSpecificMarks(marks);
    }

    /**
     * Sets the list of rules defined on this class info.
     */
    protected void setRules(LyInfoList<LyRule> rules)
    {
        int     i, n = (rules == null ? 0 : rules.getSize());

        _rules = null;
        for (i = 0; i < n; i++)
            addRule(rules.getInfo(i));
    }

    /**
     * Sets the list of controls defined on this class info.
     */
    protected void setControls(LyInfoList<LyControl> controls)
    {
        int     i, n = (controls == null ? 0 : controls.getSize());

        _controls = null;
        for (i = 0; i < n; i++)
            addControl(controls.getInfo(i));
    }

    /**
     * Sets the list of controls defined on this class info.
     */
    protected void setLabels(LyInfoLabel[] labels)
    {
        int     i, n = (labels == null ? 0 : labels.length);

        _labels = null;
        _userLabels = null;

        for (i = 0; i < n; i++)
            addLabel(labels[i]);
    }
} // end class LyClassInfo
