/*
 * LyCompoundAction.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;

import leon.app.*;
import leon.info.misc.*;
import leon.misc.*;


/**
 * This class implements the description of compound actions : action composed of 2 sub-actions.
 * Compound actions are used to allow building of composite views (views composed of sub-views).
 * Sub-actions of a compound action can also be compound actions this gives us the richness for
 * building complex presentations.<br>
 * This class extends LyAction : a compound action is an action and can be used everywhere actions
 * are used.<br>
 * The extra attributes on compound actions are the following :
 *
 * <ul>
 *   <LI>two sub-actions,</LI>
 *   <LI>a context index which values are : -1, 0, 1 This context index defines which action view is
 *     used as context for the other one (selections cause the other view to be refreshed)
 *
 *     <p>-1 : no action is context for the other one,</p>
 *
 *     <p>0 : first action is context for the second one,</p>
 *
 *     <p>1 : second action is context for the first one.</p>
 *   </LI>
 *   <LI>a presentation style which defines how sub-actions views are presented Supported values are
 *     defined as public static and named Style.XXX In the current version, two styles are supported
 *     (horizontal and vertical).</LI>
 *   <LI>ratios attributed to each sub-action view. This is a string contained numbers that give a
 *     weight on each of sub-panel views (for example "30 70" gives 30% to the first sub-action
 *     view, 70% to the second one).
 *
 *     <p>See also tag <COMPOUND_ACTION> of leon.dtd</p>
 *   </LI>
 * </ul>
 */
public final class LyCompoundAction extends LyAction
{
	public static final String __version =
		"$Id: LyCompoundAction.java 33190 2011-11-17 14:55:53Z kchenal $";

    /** Presentation styles . */
    public enum Style
    {
        // Presentation style : horizontal.
        HORIZONTAL,
        // Presentation style : vertical.
        VERTICAL,
        // Presentation style : grid.
        GRID
    }

    public enum Context
    {
        FIRST,
        SECOND,
        NONE
    }
    
    public enum ClientContext
    {
        FIRST,
        SECOND,
        BOTH,
        NONE
    }

	/** First action. */
	LyInfoList<LyAction>	   _subActions;

	/**
	 * Context index : values are -1, 0, 1. 1 = no context, 0 = context is given by first action, 1
	 * = context is given by second action.
	 */
	Context					   _context;

	/** Presentation style. Defines how sub-actions views are presented. */
	Style					   _style;

	/** Ratios attributed to each sub-action view. */
	String					   _sizes;

	/**
	 * Index of action that gives context to client action : values are -2, -1, 0, 1. 2 = both
	 * actions give context, 1 = no context, 0 = context is given by first action, 1 = context is
	 * given by second action.
	 */
	ClientContext      		   _clientContext;

	/**
	 * Initialization of compound action attributes.
	 */
	protected void init()
	{
		super.init();

		// Initialize class attributes.
		_subActions = null;
		_style = null;
		_sizes = null;
		_context = null;
		_clientContext = null;
	}

	/**
	 * Constructor of a new compound action.
	 *
	 * @param id                  Identifier of the compound action.
	 * @param template            Reference of the action template.
	 * @param controllerClassName Java class name of the action controller.
	 * @param behaviorClassName   Java class name of the action behavior.
	 * @param builderClassName    Java class name of the action builder.
	 * @param style               Presentation style.
	 * @param ratios              Ratios attributed to each sub-action.
	 * @param context        Context index. Accepted values are : -1 = no context, 0 = context
	 *                            is given by first action, 1 = context is given by second action.
	 * @param clientContext  Client context index. Accepted values are : -2 = both, -1 = no
	 *                            context, 0 = context is given by first action, 1 = context is
	 *                            given by second action.
	 * @param target              The target zone id for this action.
	 */
	protected LyCompoundAction(String id, LyAction template, String controllerClassName,
		String behaviorClassName, String builderClassName, Style style, String sizes,
        Context context, ClientContext clientContext, String target)
	{
		this(id);
		construct(template, controllerClassName, behaviorClassName, builderClassName, style, sizes,
			context, clientContext, target);
	}

	/**
	 * Constructor of an external compound action.
	 *
	 * @param id Identifier of the compound action.
	 */
	LyCompoundAction(String id)
	{
		super(id);
	}

	/**
	 * Constructor of a new compound action.
	 *
	 * @param template            Reference of the action template.
	 * @param controllerClassName Java class name of the action controller.
	 * @param behaviorClassName   Java class name of the action behavior.
	 * @param builderClassName    Java class name of the action builder.
	 * @param style               Presentation style.
	 * @param ratios              Ratios attributed to each sub-action.
	 * @param context        Context index. Accepted values are : -1 = no context, 0 = context
	 *                            is given by first action, 1 = context is given by second action.
	 * @param clientContext  Client context index. Accepted values are : -2 = both, -1 = no
	 *                            context, 0 = context is given by first action, 1 = context is
	 *                            given by second action.
	 * @param target              The target zone id for this action.
	 */
	protected void construct(LyAction template, String controllerClassName,
		String behaviorClassName, String builderClassName, Style style, String sizes,
		Context context, ClientContext clientContext, String target)
	{
		super.construct(template, controllerClassName, behaviorClassName, builderClassName, null, target);

		// Initialize mark noTitlebar='true' -> to keep compatibility with previous version
		// to force titleBar : add noTitlebar='true' + operation='CLEAR' in metamodel
		BitSet marks = getMarks();

		if (marks == null)
			marks = new BitSet();
		marks.set(LyAction.MARK_NO_TITLEBAR);
		marks.set(LyAction.MARK_FRAME_VIEW);
		setMarks(marks);

		if (style != null)
			setStyle(style);

		if (sizes != null)
			setSizes(sizes);
		
        setContext(context);
        setClientContext(clientContext);
	}

	/**
	 * Constructor of a compound action for a xref action.
	 *
	 * @param xrefAction XRef action.
	 */
	public LyCompoundAction(LyAction xrefAction)
	{
		this(xrefAction.getId(), xrefAction, null, null, null, Style.HORIZONTAL, "0 100",
			Context.FIRST, ClientContext.SECOND, null);

		initFromTemplate();

		if (_controllerClassName == null)
			_controllerClassName = new LyJavaName(this, "leon.control.LyXRefAreaController");

		LyApplication			  application = xrefAction.getApplication();
		LyAction				  contextAction = (LyAction)application.getInfo("__context");
		LyAction				  tableAction = (LyAction)application.getInfo("_table");
		LyAction				  classAction;
		LyClassInfo				  classInfo;
		LyFilter				  classFilter;
		Hashtable<String, Object> parameters = null;
		String					  actionId;
		String					  limit;

		classInfo = (LyClassInfo)xrefAction.getParameter("_classInfo");
		classFilter = (LyFilter)xrefAction.getParameter("_classFilter");
		limit = (String)xrefAction.getParameter("_limit");

		if (classFilter != null)
		{
			parameters = new Hashtable<String, Object>(1);
			parameters.put("_classFilter", classFilter);
		}
		if (LyTools.intFromString(limit) > 0)
		{
			if (parameters == null)
				parameters = new Hashtable<String, Object>(1);
			parameters.put("_limit", limit);
		}

		actionId = "_table" + LyAction.ACTION_ID_SEP + classInfo.getId();
		classAction = (LyAction)application.getInfo(actionId);

		if (classAction == null)
		{
			classAction = new LyAction(actionId, tableAction, classInfo, parameters);
			classAction.setApplication(application);
			classAction.setName(xrefAction.getName());
			classAction.setFilePrefix(xrefAction.getFilePrefix());
			classAction.setMin(1);
		}

		addSubAction(contextAction);
		addSubAction(classAction);
		_application = application;

		if (getFilePrefix() == null)
			setFilePrefix(classInfo.getFilePrefix());
		if (getName() == getId())
			setName(classInfo.getName());
	}

	/**
	 * Constructs a new compound action.
	 *
	 * @param template   Action template.
	 * @param parameters Action parameters.
	 */
	public LyCompoundAction(LyAction template, Hashtable<String, Object> parameters)
	{
		super(template, parameters);
	}

	/**
	 * Init this instance attributes from its template.
	 *
	 * @param parts Indicates which parts of the template are copied.
	 */
	void initFromTemplate(int parts)
	{
        parts &= ~_initParts;
        if (parts == 0)
        {
            checkDefaultValues();
            return;
        }

		if (_template == null)
		{
			checkDefaultValues();
			return;
		}
		
		super.initFromTemplate(parts);
		
		if (_template instanceof LyCompoundAction)
		{
			LyCompoundAction template = (LyCompoundAction)_template;
			
			if ((template._style != null) && (_style == null))
				_style = template._style;
			
			if ((template._sizes != null) && (_sizes == null))
				_sizes = template._sizes;
			
			if ((template._context != null) && (_context == null))
				_context = template._context;
			
			if ((template._clientContext != null) && (_clientContext == null))
				_clientContext = template._clientContext;

			if (template._subActions != null)
				_subActions = new LyInfoList<LyAction>(template._subActions);
			
			checkDefaultValues();
		}
	}
	
	/**
	 * Check the default values for some attributes and set them if needed.
	 */
	protected void checkDefaultValues()
	{
		if (_style == null)
			_style = Style.HORIZONTAL;
		
		if (_context == null)
			_context = Context.NONE;
		
		if (_clientContext == null)
			_clientContext = ClientContext.BOTH;
	}

	/**
	 * Final step at the end of the parsing.
	 *
	 * @exception IllegalArgumentException If the first or the second action is not viewable.
	 */
	void check(LyObjectInfo parent)
	{
		super.check(parent);

		
        // setting the default controller
        if (_controllerClassName == null)
            _controllerClassName = new LyJavaName(this, "leon.control.LyCompoundAreaController");
		
        //	adding the notformobile mark
      /*  if (_marks == null)
			_marks = new BitSet();
        _marks.set(LyAction.MARK_NOTFORMOBILE);*/
        
		int		 i, n = ((_subActions == null) ? 0 : _subActions.getSize());
		LyAction action;

		for (i = 0; i < n; i++)
		{
			action = _subActions.getInfo(i);
			if ((action == null) || (action.isLoaded() && !action.isViewable()))
				throw new IllegalArgumentException("BF AppEngine : invalid action " + action
					+ " used in compound action " + this
					+ ", only viewable actions can be used in a compound action");
		}
	}

	/**
	 * Lock this tab action.
	 */
	protected void lock()
	{
		if (!isLoaded())
			return;

		super.lock();

		// Turn tab actions list to unwritable
		if (_subActions != null)
			_subActions.lock();
	}

	/**
	 * Gets first sub-action of this compound action.
	 *
	 * @return The first action contained in this compound action.
	 */
	public LyAction getFirstAction()
	{
		return (((_subActions == null) || (_subActions.getSize() < 1)) ? null
																	   : _subActions.getInfo(0));
	}

	/**
	 * Gets second sub-action of this compound action.
	 *
	 * @return The second action contained in this compound action.
	 */
	public LyAction getSecondAction()
	{
		return (((_subActions == null) || (_subActions.getSize() < 2)) ? null
																	   : _subActions.getInfo(1));
	}

	/**
	 * Gets presentation style which defines how sub-actions views are presented.
	 *
	 * @return Presentation style of this compound action. Supported values are defined as public
	 *         static and named Style.XXX.
	 */
	public Style getStyle()
	{
		return _style;
	}

	/**
	 * Gets ratios attributed to each sub-action view. This is a string contained numbers that give
	 * a weight on each of sub-views (for example "30 70" gives 30% to the first sub-action view,
	 * 70% to the second one).
	 *
	 * @return Relative sizes of each sub-action view.
	 */
	public String getSizes()
	{
		return _sizes;
	}

	/**
	 * Gets ratios attributed to each sub-action view. This is a string contained numbers that give
	 * a weight on each of sub-views (for example "30 70" gives 30% to the first sub-action view,
	 * 70% to the second one).
	 *
	 * @return Relative sizes of each sub-action view.
	 */
	public int[] getIntSizes()
	{
		if (_sizes == null)
			return null;

		String[] list = LyString.stringToList(_sizes, ' ');
		int		 i, n = ((list == null) ? 0 : list.length);
		int[]    sizes = new int[n];

		for (i = 0; i < n; i++)
		{
			if (list[i].equals("|"))
				sizes[i] = 0;
			else if (list[i].equals("||"))
				sizes[i] = 1;
			else
				sizes[i] = LyTools.intFromString(list[i]);
		}

		return sizes;
	}

	/**
	 * Specifies which action is used as context for the other one (selections on the view of the
	 * context action cause the other view to be refreshed).
	 *
	 * @return The first action, the second action or null.
	 */
	public LyAction getContextAction()
	{
		if (_style == Style.GRID)
			return null;

        int     index = _context.ordinal();
        
		if ((index < 0)
				|| (index >= getSubActionCount()))
			return null;

		return getSubAction(index);
	}

    /**
     * Specifies if the context action is the first the second or none.
     *
     * @return the context index; NONE if compound action has style GRID
     */
    public short getContextIndex()
    {
       if (_style == Style.GRID)
           return -1;
       
       switch (_context)
       {
       case FIRST:
           return 0;
       case SECOND:
           return 1;
       default:
           return -1;
       }
    }
    
	/**
	 * Specifies if the context action is the first the second or none.
	 *
	 * @return the context index; NONE if compound action has style GRID
	 */
	public Context getContext()
	{
		return (_style == Style.GRID ? Context.NONE : _context);
	}

    /**
     * Specifies if the client context action is the first action, the second one, none or both of
     * them.
     *
     * @return the client context index
     */
    public short getClientContextIndex()
    {
        if (_style == Style.GRID)
            return -1;
        
        switch (_clientContext)
        {
        case FIRST:
            return 0;
        case SECOND:
            return 1;
        case BOTH:
            return -2;
        default:
            return -1;
        }
    }
    
	/**
	 * Specifies if the client context action is the first action, the second one, none or both of
	 * them.
	 *
	 * @return the client context index
	 */
	public ClientContext getClientContext()
	{
		return _clientContext;
	}

	/**
	 * Add an sub-action for this compound action.
	 *
	 * @param action Sub-action to add to this compound action.
	 */
	void addSubAction(LyAction action)
	{
		if (_subActions == null)
			_subActions = new LyInfoList<LyAction>();

		_subActions.addInfo(action);
		
		if ((_style != Style.GRID)
		        && (_subActions.getSize() > 2))
		    _subActions.removeInfoAt(0);
	}

	/**
	 * Gets sub-actions list.
	 *
	 * @return List of these actions.
	 */
	public LyInfoList<LyAction> getSubActions()
	{
		return _subActions;
	}

	/**
	 * Gets sub-action at the given index in sub-actions list.
	 *
	 * @param  index Index of the searched sub-action.
	 * @return Sub-action at the given index, null if no sub-action is defined.
	 */
	public LyAction getSubAction(int index)
	{
		return ((_subActions == null) ? null : _subActions.getInfo(index));
	}

	/**
	 * Gets number of sub-actions.
	 *
	 * @return Number of sub-actions of this action.
	 */
	public int getSubActionCount()
	{
		return ((_subActions == null) ? 0 : _subActions.getSize());
	}
     
    /**
     * Sets the context action is the first the second or none.
     *
     *@param context The context index; NONE if compound action has style GRID
     */
    protected void setContext(Context context)
    {
        _context = (context == null ? Context.NONE : context);
    }
    
    /**
     * Sets the client context action is the first action, the second one, none or both of them.
     *
     * @param clientContext the client context index
     */
    protected void setClientContext(ClientContext clientContext)
    {
        _clientContext = clientContext;
    }
    
    /**
     * Sets sub-actions list.
     *
     * @param actions List of these actions.
     */
    protected void setSubActions(LyInfoList<LyAction> actions)
    {
        _subActions = new LyInfoList<LyAction>(actions);
    }
    
    /**
     * Sets presentation style which defines how sub-actions views are presented.
     *
     * @param style Presentation style of this compound action. Supported values are defined as public
     *              static and named Style.XXX.
     */
    protected void setStyle(Style style)
    {
        _style = style;
    }

    /**
     * Sets ratios attributed to each sub-action view. This is a string contained numbers that give
     * a weight on each of sub-views (for example "30 70" gives 30% to the first sub-action view,
     * 70% to the second one).
     *
     * @param sizes Relative sizes of each sub-action view.
     */
    protected void setSizes(String sizes)
    {
        _sizes = sizes;
    }
    
    /**
     * Method used to check if this action is viewable or not.
     * @return True.
     */
    public boolean isViewable()
    {
        return true;
    }
} // end class LyCompoundAction
