/*
 * LyControl.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;
import leon.data.*;
import leon.data.datainterface.*;
import leon.info.misc.*;
import leon.misc.*;


/**
 * This class defines a control : Specific program associated to a class and a field and triggered
 * on objects change.
 */
public class LyControl extends LyInfo
{
	public static final String			  __version =
		"$Id: LyControl.java 40680 2012-09-06 06:52:08Z jlsaouli $";

    /** Level values. */
    public enum Level
    {
        CLASS,
        INSTANCE
    }

	/** Types of control. */
    public enum Type
    {
	   CLASS,
       FIELD
    }

	/** Java name of the control. */
	LyJavaName							  _javaName;

	/** Priority of the control. */
	int									  _priority;

	/** Level of the control. */
	Level								  _level;

	/** Hashtable of the arguments of the control. */
	Hashtable<String, LyArgument>		  _args;

	/** Hashtable of pair (classInfo, controlInterface) for persistent controls (LEVEL_CLASS). */
	Hashtable<Object, LyControlInterface> _interfaces;
	
    /** Parsed formula associated to this control : result of formula expression parsing. */
    LyFormula                             _formula;
    
    /** Error message */
    String                                _message;


	/* Constructor */

	/**
	 * Constructor of a control.
	 *
	 * @param id       Id of the control.
	 * @param javaName Java name of the control.
	 * @param priority Priority of the control.
	 * @param modeType Control type.
	 */
	LyControl(String id, String javaName, Level level, int priority, LyTag.Model modelType)
	{
		this(id);
		construct(javaName, level, priority, modelType);
	}

	/**
	 * Constructor of a control.
	 *
	 * @param id Id of the control.
	 */
	protected LyControl(String id)
	{
		super(id, false);
	}

	/**
     * Returns extra tags.
     */
    protected LyTag[] getExtraTags()
    {
        return new LyTag[] { _javaName };
    }
    
    protected void setJavaName(String javaName)
    {
        _javaName = (javaName == null ? null : new LyJavaName(_parent, javaName));
    }
    
    protected void setLevel(Level level)
    {
        _level = (level == null ? Level.CLASS : level);
    }
    
    protected void setPriority(int priority)
    {
        _priority = priority;
    }
    
	/**
	 * Construction of a control.
	 *
	 * @param javaName Java name of the control.
	 * @param priority Priority of the control.
	 * @param modeType Control type.
	 */
	void construct(String javaName, Level level, int priority, LyTag.Model modelType)
	{
		super.construct();

		setJavaName(javaName);
		setLevel(level);
		setPriority(priority);
		setModelType(modelType);
	}

	/**
	 * Gets java name of the control.
	 */
	public final String getJavaName()
	{
		return (_javaName == null ? null : _javaName.getJavaName());
	}

	/**
	 * Gets java class.
	 *
	 * @return Java class for this control.
	 */
	public Class<?> getJavaClass()
	{
		load();
		
		if (_javaName == null)
		    return null;
		
		// loading classes
		Class<?> javaClass = _javaName.getJavaClass();

		if (javaClass != null)
		    return javaClass;
		
		LyEnvironment environment = getEnvironment();
		
		environment.logError("BF AppEngine : unknown java class " + _javaName + " for control " + _id);
		return null;
	}

	/**
	 * Gets instance of the java class of this control.
	 *
	 * @param  classInfo Class info for which the instance is requested.
	 * @return An instance that implements the java class
	 */
	public LyControlInterface getInstance(LyObject object)
	{
		if (object == null)
			return null;

		if (getLevel() == Level.CLASS)
			return getInstance(object.getClassInfo());
		else
			return getInstance((Object)object);
	}

	/**
	 * Gets instance of the java class of this control.
	 *
	 * @param  classInfo Class info for which the instance is requested.
	 * @return An instance that implements the java class
	 */
	public LyControlInterface getInstance(LyClassInfo classInfo)
	{
		return getInstance((Object)classInfo);
	}

	/**
	 * Gets instance of the java class of this control.
	 *
	 * @param  classInfo Class info for which the instance is requested.
	 * @return An instance that implements the java class
	 */
	LyControlInterface getInstance(Object key)
	{
		LyControlInterface instance = null;

		if ((key != null)
				&& (_interfaces != null))
			instance = _interfaces.get(key);

		if (instance != null)
			return instance;

		Class<?> javaClass = getJavaClass();
		
		if (javaClass == null)
		    return _formula;

		try
		{
			instance = (LyControlInterface)javaClass.newInstance();
		}
		catch (Exception e)
		{
			LyEnvironment environment = getEnvironment();

			environment.writeException(e);
			environment.logError("BF AppEngine : instanciation of control interface "
				+ javaClass + " failed on control " + this);
		}

		if ((instance != null)
				&& (key != null))
		{
			if (_interfaces == null)
				_interfaces = new Hashtable<Object, LyControlInterface>();
			_interfaces.put(key, instance);
		}

		return instance;
	} // end method getInstance

	/**
	 * Gets level of the control.
	 */
	public final Level getLevel()
	{
		load();
		return _level;
	}

	/**
	 * Gets priority of the control.
	 */
	public final int getPriority()
	{
		load();
		return _priority;
	}

	/**
	 * Gets an argument by its name.
	 *
	 * @param     name Name for which a value is searched.
	 * @return    Found value if any, or null if no argument was stored with the given name.
	 * @exception IllegalArgumentException If the given name is null.
	 */
	public final Object getValue(Object name)
	{
		if (name == null)
			throw new IllegalArgumentException("BF AppEngine : name parameter can't be null");

		// Get value associated to that name
		load();
		return ((_args == null) ? null : _args.get(name));
	}

	/**
	 * Gets table of arguments associated to this instance.
	 *
	 * @return Returns the table of all the arguments associated to this control. This is a copy of
	 *         the internal hashtable that contains these arguments.
	 */
	public final Hashtable<String, LyArgument> getArgs()
	{
		load();
		return (_args == null ? null : new Hashtable<String, LyArgument>(_args));
	}

	/**
	 * Gets an array of arguments usable for calling the interface of the control.
	 *
	 * @return An array of all the arguments of this control, may be null.
	 */
	public LyArgument[] getArguments()
	{
		if (_args == null)
			return null;

		Collection<LyArgument> args = _args.values();
		LyArgument[]		   arguments = new LyArgument[args.size()];

		return args.toArray(arguments);
	}

	/**
	 * Checks whether this control performs the given type of control.
	 *
	 * @param  type Checked type.
	 * @return True if this control has the given type, false otherwise.
	 */
	public boolean hasType(Type type)
	{
		Class<?> javaClass = getJavaClass();

		if ((type == Type.CLASS)
		    && (((javaClass == null)
		            && (getParent() instanceof LyClassInfo))
				|| LyClassControl.class.isAssignableFrom(javaClass)))
			return true;
		
		if ((type == Type.FIELD)
		    && (((javaClass == null)
	                  && (getParent() instanceof LyFieldInfo))
				|| LyFieldInfoControl.class.isAssignableFrom(javaClass)))
			return true;

		return false;
	}

	/**
	 * Adds an argument to this control. An argument is a value associated to a name stored in a
	 * hashtable.
	 *
	 * @param     name  Name of the argument. If a value was already stored for that name, this
	 *                  value is overriden.
	 * @param     value Value of the argument.
	 * @return    The previous value associated to the given name if there was one, null otherwise.
	 * @exception IllegalArgumentException If the given name or value is null.
	 */
	final Object addArgument(LyArgument arg)
	{
		if (arg == null)
			throw new IllegalArgumentException("BF AppEngine : argument parameter can't be null");

		// Creates hashtable if null
		if (_args == null)
			_args = new Hashtable<String, LyArgument>(1);

		// Associates argument to the name
		return _args.put(arg.getName(), arg);
	}

	/**
	 * Check if this control is equal to the given one.
	 */
	public boolean equals(Object control)
	{
		return (control instanceof LyControl)
			&& (_javaName == ((LyControl)control)._javaName)
			&& (_priority == ((LyControl)control)._priority)
			&& ((_args == ((LyControl)control)._args)
				|| ((_args != null)
					&& (((LyControl)control)._args != null)
					&& _args.equals(((LyControl)control)._args)));
	}
    
    /**
     * Sets the array of arguments usable for calling the interface of the control.
     */
    protected void setArguments(LyArgument[] arguments)
    {
        int     i, n = (arguments == null ? 0 : arguments.length);
        
        _args = null;
        for (i = 0; i < n; i++)
            addArgument(arguments[i]);
    }
	
    /**
     * Gets the formula (as a string) associated to this control. The formula is provided to
     * calculate values of this field thanks to a simple mathematics expression using other numeric
     * fields in the same application class.
     *
     * @return Current formula or null if no formula is set.
     */
    public String getFormula()
    {
        return (_formula == null ? null : _formula.getExpression());
    }
    
    /**
     * Gets the formula associated to this control.
     *
     * @return Current formula or null if no formula is set.
     */
    public final LyFormula getLyFormula()
    {
        return _formula;
    }
    
    /**
     * Sets the formula associated to the field.
     *
     * @param     formula Current formula, or null if none is specified.
     * @exception ParseException If the given formula couldn't be parsed successfully.
     */
    void setFormula(String formula)
    {
        _formula = LyFormula.formulaFactory(this, formula);
    }
    
    /**
     * Gets message of the control.
     */
    public final String getMessage()
    {
        return _message;
    }
    
    /**
     * Sets message of the control.
     */
    protected void setMessage(String message)
    {
        _message = message;
    }
} // end class LyControl
