/*
 * LyDateFieldInfo.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current Version: $Id: LyDateFieldInfo.java 41407 2012-10-30 16:17:49Z bangiole $.
*/

package leon.info;

import java.text.*;
import java.util.*;

import leon.data.*;
import leon.misc.*;


/**
 * This class implements the description of a Date field.
 *
 * <P>A date fieldInfo is the implementation of a date attribute in a application class.</P>
 *
 * <P>Types for this fieldInfo indicate how much information is shown to the user (year, month, day,
 * hour, minute, seconds). Specific format may also be used. Supported types are declared as public
 * static named TYPE_XXX</P>
 *
 * <p>See also tag <DATE> of leon.dtd</p>
 *
 * @version $Id: LyDateFieldInfo.java 41407 2012-10-30 16:17:49Z bangiole $
 * @author  Lyria
 */
public final class LyDateFieldInfo extends LyFieldInfo
{
	public static final String __version =
		"$Id: LyDateFieldInfo.java 41407 2012-10-30 16:17:49Z bangiole $";

    /** Value : now. */
    public static String       VALUE_NOW = "_now";

    /** Value : today. */
    public static String       VALUE_TODAY = "_today";

    /** Value : day. */
    public static String	   VALUE_DAY = "_day";

    /** Date types. */
    public enum Type
    {
        // Date type : date (day+month+year) and time (hour+min+seconds).
        DATE_TIME,
        // Date type : date (day+month+year) and (hour+min).
        DATE_HOUR_MIN,
    	// Date type : date (day+month+year).
    	DATE,
    	// Date type : time (hour+min+seconds).
    	TIME,
    	// Date type : (hour+min).
    	HOUR_MIN
    }

    /** Type of the date field. */
    Type                        _type;

	/** Format to display date values for this field info. */
	String		               _displayDateFormat = null;

	/** Indicates if this date is multi-timezone. */
	boolean					   _withTimeZone;

	/** Optimization : keep default Date format (shared between DateFieldInfo). */
	static DateFormat		   _defaultFormat = new SimpleDateFormat();

	/** To optimize toName method, we keep a static calendar to format dates. */
	static GregorianCalendar   _calendar = null;

	/** Default peer date format when peer type is String. */
	public static String	   DEFAULT_PEER_FORMAT = "yyyy-MM-dd hh:mm:ss.SSS+ZZ";

	static
	{
		_calendar = new GregorianCalendar();
	}

	/**
	 * Init method.
	 *
	 * @see LyObjectInfo#init
	 */
	protected void init()
	{
		super.init();

        _type = Type.DATE_TIME;
		_displayDateFormat = null;
		_withTimeZone = false;
	}

	/**
	 * Init this instance attributes from its template.
	 *
	 * @param parts Indicates which parts of the template are copied.
	 */
	void initFromTemplate(int parts)
	{
        parts &= ~_initParts;
        if (parts == 0)
            return;
		if (_template == null)
			return;

		super.initFromTemplate(parts);

		if ((parts & PART_GUI) > 0)
		{
			LyDateFieldInfo template = (LyDateFieldInfo)_template;

			_displayDateFormat = template._displayDateFormat;
		}
	}

	/**
	 * Constructor of a date field.
	 *
	 * @param id       Identifier of the date field.
	 * @param javaName Name of the generated java field.
	 * @param template Template used to construct this field.
	 * @param type     Type of field info. One of the TYPE_ values
	 */
	LyDateFieldInfo(String id, String javaName, LyDateFieldInfo template, Type type)
	{
		this(id);
		construct(javaName, template, type, false);
	}

	/**
	 * Constructor of a date field.
	 *
	 * @param id           Identifier of the date field.
	 * @param javaName     Name of the generated java field.
	 * @param template     Template used to construct this field.
	 * @param type         Type of field info. One of the TYPE_ values
	 * @param withTimeZone Indicate if this field handles timeZones.
	 */
	LyDateFieldInfo(String id, String javaName, LyDateFieldInfo template, Type type, boolean withTimeZone)
	{
		this(id);
		construct(javaName, template, type, withTimeZone);
	}

	/**
	 * Constructor of a date field.
	 *
	 * @param id Identifier of the date field.
	 */
	LyDateFieldInfo(String id)
	{
		super(id);
	}

	/**
	 * Construction of a date field.
	 *
	 * @param javaName Name of the generated java field.
	 * @param template Template used to construct this field.
	 * @param type     Type of field info. One of the TYPE_ values
	 * @param withTimeZone Indicate if this field handles timeZones.
	 */
	protected void construct(String javaName, LyDateFieldInfo template, Type type, boolean withTimeZone)
	{
		super.construct(javaName, template);

        _type = (type == null ? Type.DATE_TIME : type);
        setWithTimeZone(withTimeZone);
	}

	/* Implementation of LyFieldInfo methods */

	/**
	 * Returns class of values for this field. The value class corresponding to a date field is
	 * java.util.Date.
	 */
	public Class<?> getValueClass()
	{
		return Date.class;
	}

	/**
	 * Builds the appropriate LyValue corresponding to this field for the given value object.
	 *
	 * @see LyDateValue
	 */
	public LyValue buildFieldValue(Object value)
	{
		if (value == null)
			return new LyDateValue(this);
		else if (value instanceof Date)
		{
			Date date = getRightDate((Date)value);
			return new LyDateValue(this, date);
		}
		else if (value instanceof Long)
		{
			Date date = new Date(((Long) value).longValue());
			date = getRightDate(date);
			return new LyDateValue(this, date);
		}
		else if (value instanceof String)
			return new LyDateValue(this, (String)value);
		else
			return super.buildFieldValue(value);
	}

	/**
	 * Gets a table containing the begin and end date of the current day
	 *
	 * @return  a table containing the begin and end date of the current day
	 */
	public static Date[] getTodayDate()
	{
		Date[] dates = new Date[2];
		Calendar todayCalendar = GregorianCalendar.getInstance(Locale.FRANCE);
    	Calendar calendar = new GregorianCalendar(
    			todayCalendar.get(GregorianCalendar.YEAR),
    			todayCalendar.get(GregorianCalendar.MONTH),
    			todayCalendar.get(GregorianCalendar.DAY_OF_MONTH),
    			0, 0, 0);
    	dates[0] = new Date(calendar.getTimeInMillis());

    	calendar = new GregorianCalendar(
    			todayCalendar.get(GregorianCalendar.YEAR),
    			todayCalendar.get(GregorianCalendar.MONTH),
    			todayCalendar.get(GregorianCalendar.DAY_OF_MONTH),
    			23, 59, 59);
    	dates[1] = new Date(calendar.getTimeInMillis());

    	return dates;
	}

	/**
	 *
	 * @param _value
	 * @return
	 */
	public static Date[] getDay(LyDateFieldInfo fieldInfo, String value)
	{
		String dateValue = value;
		dateValue = dateValue.replace("_day_", "");
		dateValue = dateValue.replace("_", "/");
		Date realDate = (Date)fieldInfo.parse(dateValue);
		Date[] dates = new Date[2];

		Calendar calendar = new GregorianCalendar();
		calendar.setTime(realDate);
		calendar.set(GregorianCalendar.HOUR_OF_DAY, 0);
		calendar.set(GregorianCalendar.MINUTE, 0);
		calendar.set(GregorianCalendar.SECOND, 0);
		dates[0] = new Date(calendar.getTimeInMillis());

		calendar = new GregorianCalendar();
		calendar.setTime(realDate);
		calendar.set(GregorianCalendar.HOUR_OF_DAY, 23);
		calendar.set(GregorianCalendar.MINUTE, 59);
		calendar.set(GregorianCalendar.SECOND, 59);
		dates[1] = new Date(calendar.getTimeInMillis());

		return dates;
	}

	/**
	 * Builds internal value for an instance of LyDateFieldInfo from a string (previously
	 * constructed by toString).
	 * @param stringValue the value to parse
	 * @param timeZone the reference timezone
	 * @see LyDateFieldInfo#toString
	 */
	public Object parse(String stringValue, String timeZone)
		throws LyFieldInfoFormatException
	{
		// FFT00139 : this method must be symetric to the parse method.
		if (stringValue == null)
			return null;

		LyEnvironment environment = getEnvironment();
		DateFormat    dateFormat = dateFormatFactory();

		if (stringValue.equals(VALUE_NOW))
			stringValue = dateFormat.format(new Date());
		else if(stringValue.equals(VALUE_TODAY))
		{
			return VALUE_TODAY;
		}
		else if(stringValue.startsWith(VALUE_DAY))
		{
			return stringValue;
		}

		// If the instance of dateFormat is not ready, create a new one, otherwise use cache
		Date					   date = null;
		LyFieldInfoFormatException e = null;

		try
		{
			date = parseDate(stringValue, dateFormat, timeZone);
		}
		catch (LyFieldInfoFormatException e1)
		{
			e = e1; // Try again
		}

		// If we can not parse with the default format, try display format
		if ((date == null) && (getFormat() != null))
		{
			DateFormat displayDateFormat = environment.getDisplayDateFormat(getFormat());

			try
			{
				date = parseDate(stringValue, displayDateFormat);
			}
			catch (LyFieldInfoFormatException e1)
			{
				e = e1; // Try again
			}
		}

		// If parse is still invalid, try with default format ?
		if (date == null)
		{
			try
			{
				date = parseDate(stringValue, _defaultFormat);
			}
			catch (LyFieldInfoFormatException e1)
			{
				e = e1; // Try again
			}

		}
		// Can not format date ...
		if ((date == null) && (e != null))
			throw e;

		return date;

	}
	/**
	 * Builds internal value for an instance of LyDateFieldInfo from a string (previously
	 * constructed by toString).
	 *
	 * @see LyDateFieldInfo#toString
	 */
	public Object parse(String stringValue)
		throws LyFieldInfoFormatException
	{
		return parse(stringValue, null);
	} // end method parse

	/**
	 * Compare two values of this fieldInfo.
	 */
	public int compare(Object value1, Object value2, BitSet modifiers)
	{
		if (value1 == value2)
			return 0; // identical.
		if (value1 == null)
			return -1;
		if (value2 == null)
			return 1;

		Date	d1 = getRightDate((Date)value1);
		Date	d2 = getRightDate((Date)value2);

		// Compare on values in seconds
		long l1 = d1.getTime();
		long l2 = d2.getTime();

		l1 = l1 / 1000L;
		l2 = l2 / 1000L;

		// NB : we can not return l1-l2 as int value : may have cast issues
		if (l1 == l2)
			return 0;

		return ((l1 > l2) ? 1 : -1);
	} // end method compare

	/**
	 * Returns the filter conditions authorized on this field. The values must match the short
	 * constants defined in LyFilterExpression.
	 *
	 * @return An array of conditions corresponding to the authorized conditions on this field.
	 * @see    leon.info.LyFilterExpression
	 */
	public LyFilterExpression.Condition[] getFilterConditions()
	{
        LyFilterExpression.Condition[] conditions = new LyFilterExpression.Condition[5];

		conditions[0] = LyFilterExpression.Condition.EQUALS;
		conditions[1] = LyFilterExpression.Condition.LOWER_THAN;
		conditions[2] = LyFilterExpression.Condition.LOWER_OR_EQUALS;
		conditions[3] = LyFilterExpression.Condition.GREATER_THAN;
		conditions[4] = LyFilterExpression.Condition.GREATER_OR_EQUALS;

		return conditions;
	}

	/**
	 * Returns the specified value for this field as a short "displayable" string (short string that
	 * can be shown to the end-user). Default implementation consists of returning same result as
	 * toName.
	 *
	 * @param  value The internal value to convert to a short displayable string. The excepted class
	 *               of this value is the one returned from the getValueClass method.
	 * @return The value presented as a short displayable String.
	 * @see    LyFieldInfo#getValueClass
	 * @see    LyFieldInfo#toName
	 */
	public String toShortName(Object fieldValue)
	{

		if (!getValueClass().isAssignableFrom(fieldValue.getClass()))
			return super.toShortName(fieldValue);
		Object convertedValue = fieldValue;

		return super.toShortName(convertedValue);
	}

	/**
	 * Returns the specified value for this field as a "displayable" string.
	 */
	public String toName(Object fieldValue)
	{
		if (fieldValue == null)
			return null;

        if (fieldValue instanceof String)
        {
            if (fieldValue.equals(VALUE_NOW))
                  return "LY_FILTER_NOW";
            else if (fieldValue.equals(VALUE_TODAY))
                  return "LY_FILTER_TODAY";
            else if (((String)fieldValue).startsWith(VALUE_DAY))
                return ((String)fieldValue).substring(VALUE_DAY.length() + 1).replace('_', '/');
        }

		if (!getValueClass().isAssignableFrom(fieldValue.getClass()))
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ fieldValue.getClass() + " expected " + getValueClass());

		String name = getFormattedName(fieldValue);

		if (name != null)
			return name;

		// Transforms the requested date to a visible string matching the date format.
		// First, try to get a display Format if any is specified.
		LyEnvironment environment = getEnvironment();

		if (_displayDateFormat != null)
		{
			DateFormat displayFormat = environment.getDisplayDateFormat(_displayDateFormat);

			return displayFormat.format((Date)fieldValue);
		}

		// Use 'internal' dateFormat based on fieldInfo type
		String format = environment.getStringDateFormat(getType().ordinal());

		return LyDateFormatter.format((Date)fieldValue, format, environment);
	} // end method toName

	/**
	 * Converts the specified value for this field to an "internal" string (string that enables
	 * reconstruction through parse method).
	 *
	 * @see LyDateFieldInfo#parse
	 */
	public String toString(Object fieldValue)
	{
		return toString(fieldValue, false);
	}

	/**
	 * Converts the specified value for this field to an "internal" string (string that enables
	 * reconstruction through parse method).
	 *
	 * @see LyDateFieldInfo#parse
	 */
	public String toString(Object fieldValue, boolean ignoreFormat)
	{
		return toString(fieldValue, ignoreFormat, null);
	}

	/**
	 * Converts the specified value for this field to an "internal" string (string that enables
	 * reconstruction through parse method).
	 *  @param timeZone the time zone to display
	 */
	public String toString(Object fieldValue, boolean ignoreFormat, String timeZone)
	{
		if (fieldValue == null)
			return null;

		if(fieldValue instanceof String &&
				(fieldValue.equals(VALUE_NOW)
						|| fieldValue.equals(VALUE_TODAY)
						|| ((String)fieldValue).startsWith(VALUE_DAY)))
			return (String)fieldValue;

		if (!getValueClass().isAssignableFrom(fieldValue.getClass()))
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ fieldValue.getClass() + " expected " + getValueClass());

		if (!ignoreFormat)
		{
			String string = getFormattedString(fieldValue);

			if (string != null)
				return string;
		}

		LyEnvironment environment = getEnvironment();
		DateFormat    dateFormat = dateFormatFactory();
		String		  str = LyDateFormatter.format((Date)fieldValue, dateFormat, getEnvironment(), timeZone);

		// If we can not format with the default format, try display format
		if ((str == null) && (_displayDateFormat != null))
		{
			DateFormat displayFormat = environment.getDisplayDateFormat(_displayDateFormat);

			str = LyDateFormatter.format((Date)fieldValue, displayFormat, getEnvironment(), timeZone);
		}

		// If format is still invalid, try with default format ?
		if (str == null)
		{
			synchronized (_defaultFormat)
			{
				str = LyDateFormatter.format((Date)fieldValue, _defaultFormat, getEnvironment(), timeZone);
			}
		}

		return str;
	} // end method toString

	/**
	 * Constructs a value for this fieldInfo from an object for a given instance. Supported
	 * conversions are the following :
	 *
	 * <TABLE>
	 *   <TR>
	 *     <TD>Date, Long, String</TD>
	 *     <TD>Date</TD>
	 *   </TR>
	 * </TABLE>
	 *
	 * @exception LyUnsupportedFieldBindingException If the requested conversion if none of the
	 *                                               supported conversions.
	 */
	public Object getValueFromObject(Object object, LyObject instance)
	{
        Object value = decodePeerValue(object);

		if (value == null)
			return null;

		Class<?> valueClass = getValueClass();

		if (valueClass.isAssignableFrom(value.getClass()))
			return value;

		if (value instanceof Long)
			return new Date(((Long)value).longValue());
		else if (value instanceof String)
		{
			String valueAsString = (String)value;

			if (valueAsString.trim().length() == 0)
				return null;
			else if (valueAsString.charAt(0) == '_')
				return parse(valueAsString);
			else
				return LyDateFormatter.parse(valueAsString, DEFAULT_PEER_FORMAT);
		}

		return super.getValueFromObject(value, instance);
	}

	/**
	 * Returns an object of the given class from a value of this fieldInfo. Supported conversions
	 * are the following :
	 *
	 * <TABLE>
	 *   <TR>
	 *     <TD>Date</TD>
	 *     <TD>Date, Long, String</TD>
	 *   </TR>
	 * </TABLE>
	 *
	 * @exception LyUnsupportedFieldBindingException If the requested conversion if none of the
	 *                                               supported conversions.
	 */
	public Object getPeerValue(Object value, Class<?> objectClass, LyObject instance)
	{
	    value = encodePeerValue(value, objectClass);

		if (value == null)
			return null;

		if (value.getClass() == objectClass)
			return value;

		if (objectClass == Long.class)
			return new Long(((Date)value).getTime());
		else if (objectClass == String.class)
			return toString(value);

        return super.getPeerValue(value, objectClass, instance);
	}

	/**
	 * Parses a dateString for the given dateFormat and raise a Warning message if needed.
	 *
	 * @param  dateString The date String we want to parse.
	 * @param  dateFormat The dateFormat instance used for parsing.
	 * @return Date if the parsing succeeded, null otherwise
	 */
	protected Date parseDate(String dateString, DateFormat dateFormat)
	{
		return parseDate(dateString, dateFormat, null);
	}

	/**
	 * Parses a dateString for the given dateFormat and raise a Warning message if needed.
	 *
	 * @param  dateString The date String we want to parse.
	 * @param  dateFormat The dateFormat instance used for parsing.
	 * @param  timeZone the timezone used for parsing
	 * @return Date if the parsing succeeded, null otherwise
	 */
	protected Date parseDate(String dateString, DateFormat dateFormat, String timeZone)
		throws LyFieldInfoFormatException
	{
		if ((dateString == null) || (dateString.trim().length() == 0) || (dateFormat == null))
			return null;

		Date		  result = null;
		boolean		  error = false;

		try
		{
			if (dateFormat == dateFormatFactory())
				result = LyDateFormatter.parse(dateString, getStringDateFormat(), getEnvironment(), timeZone);
			else
			{
				if (timeZone == null)
					timeZone = getEnvironment().getEnv("LY_TIMEZONE");
				if (timeZone != null)
				{

					dateFormat.setTimeZone(TimeZone.getTimeZone(timeZone));
				}
				result = (dateFormat.parse(dateString));
			}


			// Perform additional check because Java may Accept date as 31/31/2003
			if (dateFormat instanceof SimpleDateFormat)
			{
				if (checkDate(dateString, (SimpleDateFormat)dateFormat) == null)
				{
					error = true;
					result = null; // Not a valid date
				}
			}
		}
		catch (Exception e)
		{
			error = true;
		}
		finally
		{
			// Raises a Warning message
			//_application.getEnvironment().logWarning("BF AppEngine : Invalid date format in LyDateFieldInfo "
			//      + getId() +" (display format) :" + dateString +  " : " + e.getMessage());
			if (error)
			{
				String errorMessage = _application.getEnvironment().getMessage("LY_ER_MATCH_FORMAT",
						new String[] { dateString, getName() });

				throw new LyFieldInfoFormatException("BF AppEngine : " + errorMessage);
			}

			if (result != null)
			{
				// Normalize the value
				long time = result.getTime();

				time = time / 1000L;
				time = time * 1000L;
				result.setTime(time);
			}
		}

		return result;
	} // end method parseDate

    /**
     * Accessor to the date type.
     *
     * @return Date type.
     */
    public Type getType()
    {
        return _type;
    }

	/**
	 * Gets default value for this date field.
	 *
	 * @return Default value for this date field. The returned value is an instance of the class
	 *         returned by getValueClass.
	 */
	public Object getDefaultValue()
	{
		if ((_defaultValue != null)
				&& (_defaultValue instanceof String))
		{
			try
			{
				return parse((String)_defaultValue);
			}
			catch (Exception e)
			{
				LyEnvironment environment = getEnvironment();

				environment.writeException(e);
				return null;
			}
		}
		else
			return super.getDefaultValue();
	}

	/**
	 * Gets the raw default value as it is stored in the meta-model, without any translation or
	 * formatting. Method inserted for the needs of the meta-model editor.
	 *
	 * @return Default value for this date field as it was originally stored.
	 */
	public String getInternalDefaultValue()
	{
		if (_defaultValue == null)
			return null;
		else if (_defaultValue instanceof String)
			return (String)_defaultValue;
		else
			 return toString(_defaultValue);
	}

	/**
	 * Accessor for the display format property of this date field.
	 *
	 * @return The display format, or null if no format was set.
	 */
	public String getDisplayFormat()
	{
		return _displayDateFormat;
	}

	/**
	 * Gets the internal format as a string : the format used for selecting dates inside views. This
	 * format is different from the display format. The meaning of this format depends on the date
	 * type.
	 *
	 * @return the internal format and create it if it does not exists.
	 */
	public String getFormat()
	{
		DateFormat    dateFormat = dateFormatFactory();

		return ((SimpleDateFormat)dateFormat).toPattern();
	}

	/**
	 * Checks whether this date field is using time zone during input
	 *
	 * @return True if the date field is using time zone for input, false otherwise.
	 */
	public boolean isWithTimeZone()
	{
		return _withTimeZone;
	}

	/**
	 * Modifier for the display format property of this date field. Overriden to remove cached
	 * displayFormat when format is changed
	 *
	 * @param format Format value, or null if no format is set.
	 */
	void setDisplayFormat(String format)
	{
		if (format != null)
			_displayDateFormat = format.intern();
	}

	/**
	 * Sets this date field to use time zone during input.
	 *
	 * @param withTimeZone True if multilingual, false otherwise.
	 */
	void setWithTimeZone(boolean withTimeZone)
	{
		_withTimeZone = withTimeZone;
	}

	/**
	 * Overrides parent setType to check whether the given type is valid or not.
	 */
	void setType(Type type)
	{
	    _type = type;
	}

	/**
	 * Checks the date value according to a date format.
	 *
	 * @param  val               string to parse as a date
	 * @param  currentDateFormat expected for the date
	 * @return null if an error occured
	 */
	public static String checkDate(String val, SimpleDateFormat currentDateFormat)
	{
		String  dayString;
		String  monthString;
		String  yearString;
		int     day;
		int     month;
		int     year;
		int     indexDM;
		int     indexMY;
		int     indexYT;
		int     maxDayOfMonth;
		boolean bFrenchFormat = true;

		// Detecting date format: if first character is d: french format if M: us format
		// otherwise the format is unknown and it just return the string.
		String  formatPattern;

		formatPattern = currentDateFormat.toPattern();
		if ('d' == formatPattern.charAt(0))
			bFrenchFormat = true;
		else if ('M' == formatPattern.charAt(0))
			bFrenchFormat = false;
		else
			// Unknown format
			return val;

		// Find first separator between Day and Month
		// if it doesn't exist, the date is not valid
		indexDM = val.indexOf('/');
		if (indexDM != -1)
		{
			// Find second separator (between Month and Year)
			// if it doesn't exist, the date is not valid
			indexMY = val.indexOf('/', indexDM + 1);

			if (indexMY != -1)
			{
				indexYT = val.indexOf(' ', indexMY + 1); // We may hace TIME at the end...

				// get day, month and year as strings
				if (bFrenchFormat)
				{
					dayString = val.substring(0, indexDM);
					monthString = val.substring(indexDM + 1, indexMY);
				}
				else
				{
					monthString = val.substring(0, indexDM);
					dayString = val.substring(indexDM + 1, indexMY);
				}
				if (indexYT > 0)
					yearString = val.substring(indexMY + 1, indexYT);
				else
					yearString = val.substring(indexMY + 1);

				// Check if those strings have values
				if (dayString.length() == 0)
					return null;

				if (monthString.length() == 0)
					return null;

				if (yearString.length() == 0)
					return null;

				// Convert strings to int
				try
				{
					day = Integer.parseInt(dayString);
					month = Integer.parseInt(monthString);
					year = Integer.parseInt(yearString);
				}
				catch (NumberFormatException exc)
				{
					return null;
				}

				// Check if month value is correct, i.e. between 1 and 12
				if ((month > 12) || (month <= 0))
					return null;

				// Compute the maximum number of day for the given month
				// 31 for january, march, may, july, august, october and december
				// 28 or 29 for february
				// 30 for all other months
				switch (month)
				{
				case 1:
				case 3:
				case 5:
				case 7:
				case 8:
				case 10:
				case 12:
					maxDayOfMonth = 31;
					break;

				case 2:
					synchronized (_calendar)
					{
						if (_calendar.isLeapYear(year))
							maxDayOfMonth = 29;
						else
							maxDayOfMonth = 28;
					}
					break;
				default:
					maxDayOfMonth = 30;
					break;
				}

				// check if the day exists in the given month
				if (day > maxDayOfMonth)
					return null;

				// Compute year:
				// if year is less than 1 and if the length of the string contening the year
				// is not two, the year is not valid
				if ((yearString.length() != 2) && (year < 1))
					return null;

				// FFT0150 - If the year exceeds 9999, it is invalid: this behavior keep consitency
				// with the WEB views.
				if (yearString.length() > 4)
					return null;

				return (val); // confidence in java Parse method
			}
			else
				return val; // unknown format :confidence n java Parse method
		}
		else
			return val; // unknown format :confidence n java Parse method
	} // end method checkDate

    /**
     * Creates a new DateFormat instance of the given type.
     * @return a new DateFormat instance.
     * @see    LyEnvironment#dateFormatFactory(int)
     */
    public DateFormat dateFormatFactory()
    {
        LyEnvironment environment = getEnvironment();

        return environment.dateFormatFactory(_type.ordinal());
    }

    /**
     * Returns the string representing the date format of the given type.
     * @return a string representing the format.
     * @see    LyEnvironment#dateFormatFactory(int)
     */
    public String getStringDateFormat()
    {
        LyEnvironment environment = getEnvironment();

        return environment.getStringDateFormat(_type.ordinal());
    }

    /**
     * Gets the appropriate date corresponding to date field type.
     * @param date The date.
     * @return The date corresponding to date field type.
     */
	protected Date getRightDate(Date date)
    {
    	if (date == null)
    		return null;

    	synchronized (_calendar)
    	{
			_calendar.setTime(date);
			_calendar.set(Calendar.MILLISECOND, 0);

	    	if (_type == Type.DATE_HOUR_MIN)
	    		_calendar.set(Calendar.SECOND, 0);
	    	else if (_type == Type.DATE)
	        {
	    		_calendar.set(Calendar.HOUR, 0);
	    		_calendar.set(Calendar.MINUTE, 0);
	    		_calendar.set(Calendar.SECOND, 0);
	        }
	    	else if (_type == Type.TIME)
	    	{
	    		_calendar.set(Calendar.YEAR, 1970);
	    		_calendar.set(Calendar.MONTH, 0);
	    		_calendar.set(Calendar.DATE, 0);
	    	}
	    	else if (_type == Type.HOUR_MIN)
	    	{
	    		_calendar.set(Calendar.YEAR, 1970);
	    		_calendar.set(Calendar.MONTH, 0);
	    		_calendar.set(Calendar.DATE, 0);
	    		_calendar.set(Calendar.SECOND, 0);
	    	}

	    	return _calendar.getTime();
    	}
    }

    /**
     * Sets the raw default value as it is stored in the meta-model, without any translation or
     * formatting.
     */
    protected void setInternalDefaultValue(String defaultValue)
    {
        _defaultValue = defaultValue;
    }

	/**
	 * Convert a value from the field unit to the given unit.
	 *
	 * @param  value The value to convert.
	 * @param  unit  The new unit.
	 * @return The value converted from the reference unit.
	 */
	public Object convertToFieldUnit(Object value, LyUnit unit)
	{
		if (value == null)
			return null;

		LyUnit unitRef = getFieldUnit();

		if (unit == unitRef)
			return value;

		Object convert = super.convertToFieldUnit(value, unit);

		if ((convert != value) || (unit == null))
			return convert;

		Date 	dateValue = (Date)value;
		long	time = dateValue.getTime();
		double  offset = unit.getOffset();

		time += offset;
		if (unitRef != null)
			time -= unitRef.getOffset();

		Date date = new Date(time);

		return getRightDate(date);
	} // end method convertToFieldUnit

	/**
	 * Convert a value to the field unit from another unit.
	 *
	 * @param  value The value to convert.
	 * @param  unit  The old unit.
	 * @return The value converted to the field unit.
	 */
	public Object convertFromFieldUnit(Object value, LyUnit unit)
	{
		if (value == null)
			return null;

		LyUnit unitRef = getFieldUnit();

		if (unit == unitRef)
			return value;

		Object convert = super.convertFromFieldUnit(value, unit);

		if ((convert != value)
				|| (unit == null))
			return convert;

		Date 	dateValue = (Date)value;
		long	time = dateValue.getTime();
		double  offset = unit.getOffset();

		if (unitRef != null)
			time += unitRef.getOffset();
		time -= offset;

		Date date = new Date(time);

		return getRightDate(date);
	} // end method convertFromFieldUnit
} // end class LyDateFieldInfo
