/*
 * LyFieldConstraint.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current Version: $Id: LyFieldConstraint.java 34729 2012-03-16 07:48:27Z jlsaouli $.
 */

package leon.info;

/**
 * The LyFieldConstraint class implements constraint associated to a field.
 */
public class LyFieldConstraint extends LyTag
{
	public static final String __version =
		"$Id: LyFieldConstraint.java 34729 2012-03-16 07:48:27Z jlsaouli $";

	/** Default value for all the attributes of the constraint. */
	public static final int    DEFAULT_VALUE = -2;

	/** Remainder value for colspan and rowspan attributes. */
	public static final int    REMAINDER = -1;

    /** Fill values. */
    public enum Fill
    {
        HORIZONTAL,
        VERTICAL,
        BOTH,
        NONE,
        DEFAULT
    }

    /** Horizontal alignment values. */
    public enum HAlign
    {
        LEFT,
        CENTER,
        RIGHT,
        DEFAULT
    }
    
    /** Vertical alignment values. */
    public enum VAlign
    {
    	TOP,
        MIDDLE,
        BOTTOM,
        DEFAULT
    }

    /** Type values */
    public enum Type
    {
        // FIELD constraint : apply to the field edition part.
        FIELD,
        // CAPTION constraint : apply to the caption edition part.
        CAPTION
    }

	/** The constraint type which MUST be TYPE_FIELD or TYPE_CAPTION. Default is TYPE_FIELD */
	Type					   _type;
	
	/** Action where the constraint is used. */
	LyAction                   _action;

	/** Number of columns that the component occupies. */
	int						   _colSpan;

	/** Number of rows that the component occupies. */
	int						   _rowSpan;

	/** Weight (width) of the column that the component occupies. */
	int						   _colWeight;

	/** Weight (height) of the row that the component occupies. */
	int						   _rowWeight;

	/** Remainder indicates the end of a line. */
	boolean					   _remainder;

	/** Gives fill policy for that container. */
	Fill					   _fill;

	/** Horizontal alignment. */
	HAlign					   _hAlign;

	/** Vertical alignment. */
	VAlign					   _vAlign;

    /**
     * Default constructor.
     */
    protected LyFieldConstraint()
    {
    	_type = Type.FIELD;
		_action = null;
		_colSpan = DEFAULT_VALUE;
		_rowSpan = DEFAULT_VALUE;
		_colWeight = DEFAULT_VALUE;
		_rowWeight = DEFAULT_VALUE;
		_remainder = false;
		_fill = Fill.DEFAULT;
		_hAlign = HAlign.DEFAULT;
		_vAlign = VAlign.DEFAULT;
    }
    
    protected void setType(Type type)
    {
        _type = (type == null ? Type.FIELD : type);
    }
    
    protected void setColSpan(int colSpan)
    {
        _colSpan = colSpan;
    }
    
    protected void setRowSpan(int rowSpan)
    {
        _rowSpan = rowSpan;
    }
    
    protected void setRemainder(boolean remainder)
    {
        _remainder = remainder;
    }
    
    protected void setColWeight(int colWeight)
    {
        _colWeight = colWeight;
    }
    
    protected void setRowWeight(int rowWeight)
    {
        _rowWeight = rowWeight;
    }
   
    protected void setFill(Fill fill)
    {
        _fill = fill;
    }
  
    protected void setHAlign(HAlign hAlign)
    {
        _hAlign = hAlign;
    }
    
    protected void setVAlign(VAlign vAlign)
    {
        _vAlign = vAlign;
    }
    
	/**
	 * Constructor with all required attributes.
	 */
	public LyFieldConstraint(Type type, int colSpan, int rowSpan, boolean remainder,
		int colWeight, int rowWeight, Fill fill, HAlign hAlign, VAlign vAlign)
	{
		// Initialize attributes
		setType(type);
		_action = null;
		_colSpan = colSpan;
		_rowSpan = rowSpan;
		_colWeight = colWeight;
		_rowWeight = rowWeight;
		_remainder = remainder;

		_fill = (fill == null ? Fill.DEFAULT : fill);
		_hAlign = (hAlign == null ? HAlign.DEFAULT : hAlign);
		_vAlign = (vAlign == null ? VAlign.DEFAULT : vAlign);
	}

	/**
	 * Gets type.
	 *
	 * @return type of the constraint.
	 */
	public Type getType()
	{
		return _type;
	}

    /**
     * Gets action where the constraint is used.
     *
     * @return Action where the constraint is used (may be null when used in all forms,
     * excepted when explicitly specified for a given action).
     */
    public LyAction getAction()
    {
        return _action;
    }
    
	/**
	 * Gets value of colSpan.
	 *
	 * @return value of colSpan.
	 */
	public int getColSpan()
	{
		return _colSpan;
	}

	/**
	 * Gets value of rowSpan.
	 *
	 * @return value of rowSpan.
	 */
	public int getRowSpan()
	{
		return _rowSpan;
	}

	/**
	 * Gets value of colWeight. Weight (width) of the column that the component occupies
	 *
	 * @return value of colWeight.
	 */
	public int getColWeight()
	{
		return _colWeight;
	}

	/**
	 * Gets value of rowWeight. Weight (height) of the row that the component occupies
	 *
	 * @return value of rowWeight.
	 */
	public int getRowWeight()
	{
		return _rowWeight;
	}

	/**
	 * Gets value of remainder.
	 *
	 * @return true if remainder is set
	 */
	public boolean isRemainder()
	{
		return _remainder;
	}

	/**
	 * Gets value of fill.
	 *
	 * @return value of fill.
	 */
	public Fill getFill()
	{
		return _fill;
	}

	/**
	 * Gets value of hAlign.
	 *
	 * @return value of hAlign.
	 */
	public HAlign getHAlign()
	{
		return _hAlign;
	}

	/**
	 * Gets value of vAlign.
	 *
	 * @return value of vAlign.
	 */
	public VAlign getVAlign()
	{
		return _vAlign;
	}

	/**
	 * Check if this field constraint is equal to the given one.
	 */
	public boolean equals(Object constraint)
	{
		if (!(constraint instanceof LyFieldConstraint))
			return false;
		
		return equals((LyFieldConstraint)constraint);
	}
	
	/**
	 * Check if this field constraint is equal to the given one.
	 */
	public boolean equals(LyFieldConstraint constraint)
	{
		return (constraint != null)
			&& (_type == constraint._type)
			&& (_colSpan == constraint._colSpan)
			&& (_rowSpan == constraint._rowSpan)
			&& (_colWeight == constraint._colWeight)
			&& (_rowWeight == constraint._rowWeight)
			&& (_remainder == constraint._remainder)
			&& (_fill == constraint._fill)
			&& (_hAlign == constraint._hAlign)
			&& (_vAlign == constraint._vAlign);
	}
	
    /**
     * Sets action where the constraint is used.
     * @param action Action where the constraint is used (may be null when used in all forms,
     * excepted when explicitly specified for a given action).
     */
    void setAction(LyAction action)
    {
        _action = action;
    }
} // end class LyFieldConstraint
