/*
 * LyFieldInfo.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.lang.reflect.Method;
import java.util.*;

import leon.app.*;
import leon.data.*;
import leon.data.daemon.*;
import leon.info.binding.*;
import leon.info.infointerface.*;
import leon.info.misc.*;
import leon.misc.*;


/**
 * This class represents the information related to a field (information attached to an attribute of
 * an application class). It's merely a direct mapping of information as it is defined in the data
 * description file to describe a given field. This class is an abstract one, it contains common
 * attributes for all type of fields. One specific class is then defined fo each type of fields :
 *
 * <ul>
 *   <LI>ARRAY LyArrayFieldInfo</LI>
 *   <LI>CHOICE LyChoiceFieldInfo</LI>
 *   <LI>DATE LyDateFieldInfo</LI>
 *   <LI>FILE LyFileFieldInfo</LI>
 *   <LI>NUMBER LyNumberFieldInfo</LI>
 *   <LI>RELATION LyRelationFieldInfo</LI>
 *   <LI>STRUCT LyStructFieldInfo</LI>
 *   <LI>TEXT LyTextFieldInfo</LI>
 * </ul>
 *
 *  <p>These instances are built by the parser from the data description file and must not be
 *  modified after parsing.</p>
 *
 *  <p>This class extends LyObjectInfo and thus contains all attributes defined in that class
 *  (name, alias, marks, tip, ...).</p>
 *
 *  <p>Specific attributes are added to describe information related to a field. These specific
 *  attributes are :</p>
 *  <ul>
 *   <LI><b>binding</b>Information required to bind this field to a "physical field". This binding
 *     indicates where and how to find the data which corresponds to this field. There are different
 *     types of binding, each of them has a specific class that implements it. These bindings are
 *     located in package leon.app.info.binding.</LI>
 *   <LI><b>type</b>A field may have a type as defined by this attribute. This is a number that
 *     indicates the type of the field in a given field class (for example types of NUMBER are INT,
 *     SHORT, LONG, FLOAT, DOUBLE). Valid values are defined by each sub-class of this class, if
 *     needed.</LI>
 *   <LI><b>tab</b>. This is a name for the tab that contains this field when fields are organized
 *     as tabs (mainly, in forms). All the fields having the same tab name are located in the same
 *     tab.</LI>
 *   <LI><b>javaName</b>. This is the name of the java field generated for this field when java code
 *     corresponding to the data description is produced.</LI>
 *   <LI><b>format</b>. This is a string format used on some fields to indicates how field values
 *     are displayed.</LI>
 *   <LI><b>default value</b>. Default value is the value used when no value is given and this field
 *     is mandatory. The real type for this value depends of the type of the field to which ti's
 *     associated, this is the one used to internally represents data on the field.</LI>
 *   <LI><b>depends</b>. Depends is a list of other fields that depend on this field. This means
 *     that when this field is modified, values displayed on the depending fields are changed and
 *     must be recomputed.</LI>
 *   <LI><b>beans</b>. This is an hashtable of alternate graphical objects (beans) used to display
 *     values for this field. The key represents the action in which the bean is used.
 *
 *     <p>This class implements the LyField interface. This means that this class can be used as a
 *     physical class ("data provider") for an other field info.</p>
 *   </LI>
 * </ul>
 */
public abstract class LyFieldInfo extends LyObjectInfo
	implements LyField
{
	public static final String  __version =
		"$Id: LyFieldInfo.java 40943 2012-09-26 10:23:52Z jlsaouli $";

	/** Boolean used to store a strong dependency. */
	public static final Boolean STRONG_DEPENDENCY_VALUE = Boolean.TRUE;

	/** Boolean used to store a weak dependency. */
	public static final Boolean WEAK_DEPENDENCY_VALUE = Boolean.FALSE;

	/** Secret value. */
	public static final String	SECRET_VALUE = "*****";

	// WARNING ! : the value of the marks must be the same as the index of
	// this mark in the array fieldInfoMarks in the class LyInfoTokens.

	/**
	 * Mark : indicating that the value for the field having this mark is used as part of the
	 * identifier for objects corresponding to classes that contain this field. Ids are reserved for
	 * internal used and rarely displayed to the user. Several fields may have this mark whithin the
	 * same class. The value associated to an id field can not be modified.
	 */
	public static final short   MARK_ID = 0;

	/**
	 * Mark : indicating that the value for the field having this mark is used as part of the name
	 * for objects corresponding to classes that contain this field. Names are used when displaying
	 * objects to the user. Several fields may have this mark whithin the same class.
	 */
	public static final short   MARK_NAME = 1;

	/**
	 * Mark : indicating that the field having this mark must not be transmitted to data providers
	 * with a null value. This means that is no value is given when building a request to data
	 * providers, the default value is used.
	 */
	public static final short   MARK_NOT_NULL = 2;

	/** Mark : fields having this mark are present in creation form. */
	public static final short   MARK_CREATE = 3;

	/**
	 * Mark : Fields having this mark are present in modification form. Fields which have mark ID
	 * can't have this mark in the same time.
	 */
	public static final short   MARK_SET = 4;

	/**
	 * Mark : a field with this mark has an optional input when presented in forms (creation or
	 * modification).
	 */
	public static final short   MARK_OPTIONAL = 5;

	/**
	 * Mark : indicates that a field must have a unique value (value not present for other objects
	 * belonging to the same class).
	 */
	public static final short   MARK_UNIQUE = 6;

	/**
	 * Mark : indicates that a field is never shown directly to the end user (for example, when used
	 * on text fields, the associated text input in forms shows asterisks when user enters text).
	 * This mark is merely used for passwords.
	 */
	public static final short   MARK_SECRET = 7;

	/**
	 * Mark : indicates that a field is private and so can't be copied when cloning an object from
	 * an existing one.
	 */
	public static final short   MARK_PRIVATE = 8;

	/**
	 * Mark : this mark indicates that a field is present in views showing tables. This mark is put
	 * on main fields of a class.
	 */
	public static final short   MARK_TABLE = 9;

	/**
	 * Mark : indicating that a field represents the status of the object. Several fields can have
	 * this mark in the same class. For instance in a table, this status is shown as an icon to the
	 * left of each row.
	 */
	public static final short   MARK_STATUS = 10;

	/** Mark : indicating that table views are sortable on fields having this mark. */
	public static final short   MARK_SORT = 11;

	/**
	 * Mark : indicating that views presenting objects of classes may be filtered on fields having
	 * this mark.
	 */
	public static final short   MARK_FILTER = 12;

	/**
	 * Mark : indicates that the field is shown in the find object form. This form allows to
	 * retrieve objects by giving searched value on fields having this mark.
	 */
	public static final short   MARK_FIND = 13;

	/**
	 * Mark : indicates that the field is local (no physical representation). Local fields are never
	 * transmitted to data providers, and so can't be persistent. Depending on field type, local
	 * fields may be calculated.
	 */
	public static final short   MARK_LOCAL = 14;

	/**
	 * Mark : indicates that the field is present in editable forms but can't be modified (consult
	 * only).
	 */
	public static final short   MARK_CONSULT = 15;

	/**
	 * Mark : indicates that the field is present in creation forms but can't be modified (consult
	 * only).
	 */
	public static final short   MARK_CREATE_CONSULT = 16;

	/**
	 * Mark : indicates that the field is present in set forms but can't be modified (consult only).
	 */
	public static final short   MARK_SET_CONSULT = 17;

	/** Mark : indicates that the field must be posted when modified. */
	public static final short   MARK_NEED_POST = 18;

	/** Mark : indicates that the field is chartable on fields having this mark. */
	public static final short   MARK_CHART = 19;

	/** Mark : indicates that the field is read-only (not editable). */
	public static final short   MARK_READ_ONLY = 20;

	/** Mark : indicates that the field is disable (not enabled). */
	public static final short   MARK_DISABLE = 21;

	/** Mark : shortcut for TABLE, FILTER, SORT and FIND marks. */
	public static final short   MARK_MAIN = 22;

	/** Mark : indicates that the field in not shown in consultation form. */
	public static final short   MARK_HIDDEN = 23;

	/** Mark : indicates that values for this field must be loaded. */
	public static final short   MARK_LOAD = 24;

	/** Mark : indicates that values for this field is computed by data provider. */
	public static final short   MARK_PROVIDER_VALUE = 25;

	/**
	 * Mark : indicating that the value for the field having this mark is used as part of the
	 * application identifier for objects corresponding to classes that contain this field. Several
	 * fields may have this mark whithin the same class and the combination of values of these
	 * fields must be unique for a classInfo. The value associated to an appId field can be
	 * modified.
	 */
	public static final short   MARK_APP_ID = 26;

	/**
	 * Mark : indicates that values for this field is computed by data provider and that this mark
	 * is set b y the binding model. For internal use only.
	 */
	static final short		    MARK_PROVIDER_VALUE_BIND = 27;

	/** Mark : this mark indicates that a field is present in views showing complex tables. */
	public static final short   MARK_COMPLEX_TABLE = 28;

	/** Mark : this mark indicates that a field is used by text search action. */
	public static final short   MARK_SEARCH = 29;

	/** Mark : this mark activates auto-completion for this field. */
	public static final short   MARK_AUTO_COMPLETION = 30;

	/** Mark : this mark activates translation for this field. */
    public static final short   MARK_TRANSLATE = 31;
    
	/** Mark : this mark removes caption for this field in forms. */
    public static final short   MARK_NO_CAPTION = 32;


	// Field groups
	static Hashtable<String, LyFieldGroup>   _fieldGroups;

	// Field tabs
	static Hashtable<String, LyFieldTab>     _fieldTabs;


	/**
	 * Gets mark index corresponding to the specific given mark name. Marks are represented as bits
	 * in a bitset. Each mark has a bit position assigned to it. This mark allows to retrieve this
	 * bit value by the name of the mark.
	 *
	 * @param  markName Name of the mark.
	 * @param  add indicates if the.mark must be added if it doesn't exit
	 * @return Returns mark value (bit position in a bits set)
	 * @see    LyFieldInfo#getMarkName
	 */
	public static int getMark(String markName, boolean add)
	{
		return LyInfoTokens.getMarkIndex(LyBitSet.Type.F_MARKS, markName, add);
	}

	/**
	 * Gets mark index corresponding to the specific given mark name. Marks are represented as bits
	 * in a bitset. Each mark has a bit position assigned to it. This mark allows to retrieve this
	 * bit value by the name of the mark.
	 *
	 * @param  markName Name of the mark.
	 * @return Returns mark value (bit position in a bits set)
	 * @see    LyFieldInfo#getMarkName
	 */
	public static int getMark(String markName)
	{
		return getMark(markName, false);
	}

	/**
	 * Get mark name corresponding to a given mark value.
	 *
	 * @param  markIndex Mark value (index of a bit in a bits set).
	 * @return Corresponding mark name.
	 * @see    LyFieldInfo#getMark
	 */
	public static String getMarkName(int markIndex)
	{
		return LyInfoTokens.getMarkName(LyBitSet.Type.F_MARKS, markIndex);
	}

	/**
	 * Binding (mean to access information from a data provider) Information used to retrieve data
	 * corresponding to this field from the data provider. Local fields have no bindings.
	 */
	LyFieldBinding						 _binding;

	/**
	 * Tab : name of the tab associated to this field. The tab information is used by generic
	 * processing to build views organized by tabs. Tab info may be null (not in a tab).
	 */
	String								 _tab;

	/** Name of the generated java field. Used if code generation is used from data model. */
	String								 _javaName;

	/**
	 * Format as a string. The effective content of this attribute depend on the type of the field.
	 */
	String								 _format;

	/**
	 * The default value. Used in some views for non optional fields (don't have OPTIONAL mark) or
	 * when building creation requests to data providers to give a value for fields that don't allow
	 * null values (NOT_NULL mark).
	 */
	Object								 _defaultValue;

	/** List of fields that depend on this field. */
	LyInfoDataList<LyFieldInfo, Boolean> _depends;

	/**
	 * List of beans (names of the java class that must be used to build the control for a specific
	 * action).
	 */
	Hashtable<String, LyFieldBean>		 _beans;

	/** Import information. */
	LyFieldInfoImport					 _import;

	/** Path information. */
	LyPath								 _path;

	/** Constraint information used for form layout. */
	Vector<LyFieldConstraint>			_constraints;

	/** Constraint information used for form layout. */
	LyFieldConstraint					 _captionConstraint;

	/** List of controls associated to this field. */
	LyInfoList<LyControl>				 _controls;

	/** List of units of this field. */
	LyUnits								 _units;

	/** Current unit of this field. */
	LyUnit								 _unit;

	/** Behavior class name. */
	String								 _behaviorClassName;

	/** Indicates if this field was built from a FIELD tag. */
	boolean								 _fieldTag;

	/** List of labels associated to this field. */
	LyInfoLabel[]						 _labels;

	/** List of user labels associated to this field. */
	Vector<LyInfoLabel>					 _userLabels;

	/** Indicates if this field is calculated. */
	boolean								 _calculated;

    /**
     * The field tab.
     */
    LyFieldTab                            _fieldTab;

    /**
     * The field group.
     */
    LyFieldGroup                          _fieldGroup;

    /**
     * Indicates if field is local or not.
     */
    Boolean                               _local;

    /** Parsed formula associated to this field : result of formula expression parsing. */
    LyFormula                             _formula;


	/**
	 * Init method.
	 */
	protected void init()
	{
		super.init();

		_binding = null;
		_javaName = null;
		_tab = null;
		_beans = null;
		_format = null;
		_defaultValue = null;
		_depends = null;
		_import = null;
		_path = null;
		_constraints = null;
		_captionConstraint = null;
		_controls = null;
		_units = null;
		_unit = null;
		_behaviorClassName = null;
		_fieldTag = false;
		_labels = null;
		_calculated = false;
        _fieldTab = null;
        _fieldGroup = null;
        _local = null;
        _formula = null;
	}

	/**
	 * Init this instance attributes from its template.
	 *
	 * @param parts Indicates which parts of the template are copied.
	 */
	void initFromTemplate(int parts)
	{
        parts &= ~_initParts;
        if (parts == 0)
            return;
		if (_template == null)
			return;

		super.initFromTemplate(parts);

		LyFieldInfo template = (LyFieldInfo)_template;

		if ((parts & PART_DATA) > 0)
		{
			_format = template._format;
			_defaultValue = template._defaultValue;
			if (template._controls != null)
				_controls = new LyInfoList<LyControl>(template._controls);
			_units = template._units;
			_unit = template._unit;
			if (_behaviorClassName == null)
				_behaviorClassName = template._behaviorClassName;

            setFormula(template.getFormula());
		}
		if ((parts & PART_BIND) > 0)
		{
			if (template._binding != null)
				_binding = (LyFieldBinding)template._binding.clone();
		}
		if ((parts & PART_GUI) > 0)
		{
			if (template._beans != null)
				_beans = new Hashtable<String, LyFieldBean>(template._beans);
			if (template._constraints != null)
			    _constraints = new Vector<LyFieldConstraint>(template._constraints);
			_captionConstraint = template._captionConstraint;
		}
		if (((parts & PART_GUI) > 0)
				|| ((parts & PART_LABEL) > 0))
		{
			if (_tab == null)
				_tab = template._tab;
            if (_fieldTab == null)
                _fieldTab = template._fieldTab;
            if (_fieldGroup == null)
                _fieldGroup = template._fieldGroup;

			LyInfoLabel label;
			int		    nl = template.getLabelCount();
			int		    nu = template.getUserLabelCount();
			int		    i;

			for (i = 0; i < nl; i++)
			{
				label = template._labels[i];
				if ((label != null) && (getLabel(i) == null))
					addLabel(label);
			}
			for (i = 0; i < nu; i++)
			{
				label = template.getUserLabel(i);
				if ((_userLabels == null) || !_userLabels.contains(label))
					addLabel(label);
			}
		}
	} // end method initFromTemplate

	/**
	 * Constructor.
	 *
	 * @param id       Identifier of the field.
	 * @param javaName Name of the generated java field.
	 * @param template Template used to construct this field.
	 * @param type     Type of field info. One of the TYPE_ values protected LyFieldInfo(String id,
	 *                 String javaName, LyFieldInfo template, short type) { this(id, template);
	 *                 construct(javaName, template, type); }
	 */

	/**
	 * Constructor.
	 *
	 * @param id The identifier of the field.
	 */
	protected LyFieldInfo(String id)
	{
		super(id, false);
	}

	/**
	 * Construction of a field.
	 *
	 * @param javaName Name of the generated java field.
	 * @param template Template used to construct this field.
	 */
	protected void construct(String javaName, LyFieldInfo template)
	{
		super.construct(template);

		setJavaName(javaName);
	}

    /**
     * Sets name of the generated java field.
     * @param javaName Name of the generated java field.
     */
    protected void setJavaName(String javaName)
    {
        _javaName = javaName;
    }

	/**
	 * Final step of the parsing : check consistency. For a fieldInfo, checks are the following : -
	 * a field info can't have both ID and SET marks in the same time (in this case, the SET mark is
	 * removed).
	 *
	 * @param parent Object info in which this field is checked.
	 * @param parts  Indicates which parts are checked
	 */
	void check(LyObjectInfo parent, int parts)
	{
		super.check(parent, parts);

		if ((parts & PART_GUI) > 0)
		{
			// If the fieldInfo has CONSULT mark, force both CREATE_CONSULT and SET_CONSULT marks
			if (hasMark(MARK_CONSULT))
			{
				_marks.set(MARK_CREATE_CONSULT);
				_marks.set(MARK_SET_CONSULT);
			}

			// If the fieldInfo has MAIN mark, force TABLE, SORT, FILTER and FIND marks
			if (hasMark(MARK_MAIN))
			{
				_marks.set(MARK_TABLE);
				_marks.set(MARK_SORT);
				_marks.set(MARK_FILTER);
				_marks.set(MARK_FIND);
			}

			// If the fieldInfo has SECRET mark, clear SORT, FILTER and FIND marks
			if (hasMark(MARK_SECRET))
			{
				_marks.clear(MARK_SORT);
				_marks.clear(MARK_FILTER);
				_marks.clear(MARK_FIND);
			}
		}

		if ((parts & PART_BIND) > 0)
		{
			if (hasMark(MARK_NOT_NULL)
					&& !(this instanceof LyRelationFieldInfo) // Relations can be filled in from context
					&& (getDefaultValue() == null))
			{
				_marks.clear(MARK_NOT_NULL);
				getEnvironment().logWarning("BF AppEngine : remove mark NOT_NULL from the fieldInfo "
					+ this + " because this field doesn't have a default value.");
			}

			// If the field info is declared local, the binding is ignored.
			if (isLocal() && (_binding != null))
			{
				LyFieldInfo	template = (LyFieldInfo)getTemplate();
				
				if (((template == null)
						|| (template._binding == null))
					&& (!(parent instanceof LyClassInfo)
						|| !((LyClassInfo)parent).isVirtual()))
					getEnvironment().logWarning("BF AppEngine : binding for field " + this
						+ " will be ignored because this field info has the mark LOCAL.");
				
				_binding = null;
			}

			// If the field info is declared local, the load mark is removed
			if (hasMark(MARK_LOCAL) && hasMark(MARK_LOAD))
			{
				_marks.clear(MARK_LOAD);
				if (!(parent instanceof LyClassInfo)
						|| !((LyClassInfo)parent).isVirtual())
					getEnvironment().logWarning("BF AppEngine : remove mark LOAD for field " + this
						+ " because this field info has mark LOCAL");
			}

			// Check local mark for field info added to virtual class
			if (!hasMark(MARK_LOCAL)
					&& (parent instanceof LyClassInfo)
					&& ((LyClassInfo)parent).isVirtual())
			{
				LyClassInfo realClassInfo = ((LyClassInfo)parent).getRealClass();

				if (!realClassInfo.contains(this))
				{
					if (_marks == null)
						_marks = new BitSet();
					_marks.set(MARK_LOCAL);
					getEnvironment().logWarning("BF AppEngine : add mark LOCAL to the fieldInfo "
						+ this + " because it was added to a virtual class info.");
				}
			}
		}

		// Check import binding
		if ((_import != null) && (parent instanceof LyClassInfo))
		{
			if ((parts & PART_BIND) > 0)
				// Check import
				_import.check((LyClassInfo)parent, this);

			if ((parts & PART_GUI) > 0)
			{
				// Check marks
				if (isLocal() && (hasMark(MARK_CREATE) || hasMark(MARK_SET)))
				{
				    if (hasMark(MARK_CREATE))
				    {
                        _marks.clear(MARK_CREATE);
				        _marks.set(MARK_CREATE_CONSULT);
				    }
                    if (hasMark(MARK_SET))
                    {
                        _marks.clear(MARK_SET);
                        _marks.set(MARK_SET_CONSULT);
                    }

					getEnvironment().logWarning(
						"BF AppEngine : remove marks CREATE and SET from the fieldInfo "
						+ this + " which is a local imported field.");
				}

				// Add depend on relation of the import
				LyFieldInfo relation = _import.getRelation();

				relation.addDepend(this);
			}
		}

		if (((parts & PART_GUI) > 0)
				|| ((parts & PART_LABEL) > 0))
		{
			// Check labels.
			int		    i, n = ((_labels == null) ? 0 : _labels.length);
			LyInfoLabel label;

			for (i = 0; i < n; i++)
				if (_labels[i] != null)
					_labels[i].check(this);

			n = ((_userLabels == null) ? 0 : _userLabels.size());
			for (i = 0; i < n; i++)
			{
				label = _userLabels.elementAt(i);
				if (label != null)
					label.check(this);
			}
		}

		// Check path binding
		if ((_path != null) && (parent instanceof LyClassInfo))
		{
			if ((parts & PART_BIND) > 0)
			{
				// Check path
				_path.check((LyClassInfo)parent);

				// Check marks
				if (!hasMark(MARK_LOCAL))
				{
					if (_marks == null)
						_marks = new BitSet();
					_marks.set(MARK_LOCAL);
					getEnvironment().logWarning("BF AppEngine : add mark LOCAL to the fieldInfo "
						+ this + " which is a calculated field.");
				}
			}

			if ((parts & PART_GUI) > 0)
			{
				if (hasMark(MARK_CREATE) || hasMark(MARK_SET))
				{
					_marks.clear(MARK_CREATE);
					_marks.clear(MARK_SET);
					getEnvironment().logWarning(
						"BF AppEngine : remove marks CREATE and SET from the fieldInfo "
						+ this + " which is a local calculated field.");
				}
			}
		}

        if ((parts & PART_DATA) > 0)
        {
            if (isComputed()
                    && ((parent instanceof LyClassInfo)
                        || (parent instanceof LyStructFieldInfo)))
            {
                // Add dependencies on fieldInfos contained in formula
                LyClassInfo classInfo;

                if (parent instanceof LyClassInfo)
                    classInfo = (LyClassInfo)parent;
                else
                    classInfo = ((LyStructFieldInfo)parent).getClassInfo(0);

                LyInfoList<LyFieldInfo> dependencies = _formula.getDependencies(classInfo);
                LyFieldInfo             fieldInfo;

                if (dependencies == null)
                    return;
                int i, n = dependencies.getSize();

                for (i = 0; i < n; i++)
                {
                    fieldInfo = dependencies.getInfo(i);

                    // Detect circular dependency
                    if (fieldInfo == this)
                    {
                        _formula = null;
                        throw new IllegalStateException("BF AppEngine : invalid formula on field "
                            + this + ", circular dependency detected");
                    }

                    fieldInfo.addDepend(this);
                }
            }
        }

	} // end method check

	/**
	 * Lock this field info.
	 */
	protected void lock()
	{
		super.lock();

		// Lock lists
		if (_controls != null)
			_controls.lock();
		if (_labels != null)
		{
			int i, n = ((_labels == null) ? 0 : _labels.length);

			for (i = 0; i < n; i++)
				if (_labels[i] != null)
					_labels[i].lock();
		}
	}

	/**
	 * Returns class of values for this field. This class is the class of the value returned when
	 * getValue is done on an object. This is the class of the internal Leonardi representation for
	 * the values of this field whatever the real "physical class" to which this field binds to, is.
	 *
	 * @return Effective class of values built by this fieldInfo. This method is abstract, the real
	 *         class depends on each type of field.
	 * @see    LyObject#getValue
	 */
	public abstract Class<?> getValueClass();

	/**
	 * Builds the appropriate LyValue corresponding to this field from a value provided as an
	 * Object.
	 *
	 * @param  value Value from which the LyValue is built.
	 * @return An instance of a class that inherits from LyValue corresponding the given value.
	 *         Classes extending LyValue are given to ease data manipulations, they contain methods
	 *         that encapsulate data conversions from and to the internal representation. There is
	 *         one class for each type of field.
	 */
	public LyValue buildFieldValue(Object value)
	{
		if (value instanceof LyValue)
			value = ((LyValue)value).getValue();

		return new LyFieldInfoValue(this, value);
	}

	/**
	 * Processes the given value to get the right value before transmitting it to the peer layer.
	 *
	 * @param  value Value to process.
	 * @return The processed value.
	 */
	public Object processValue(Object value)
	{
		// Extract the right value from the given one
		LyValue fieldValue = buildFieldValue(value);

		return fieldValue.getValue();
	}

	/**
	 * Encodes the given value to get the right value before tranmitting it to the peer layer.
	 *
	 * @param  value Value to encode.
	 * @return The encoded value.
	 */
	public Object encodeValue(Object value)
	{
		// Nothing to do
		return value;
	}

	/**
	 * This method allows to create a value from a string for this field. The class of the returned
	 * value is the one given by getValueClass method.
	 *
	 * @param     stringValue The value presented as a string as produced by the toString method.
	 * @return    The value decoded from the given string.
	 * @exception LyFieldInfoFormatException If the string can't be parsed properly.
	 * @see       LyFieldInfo#toString
	 * @see       LyFieldInfo#getValueClass
	 */
	public abstract Object parse(String stringValue)
		throws LyFieldInfoFormatException;

	/**
	 * Returns a "displayable string" (string that can be shown to the end-user) of the specified
	 * value for this field.
	 *
	 * @param  value The internal value to convert to a displayable string. The excepted class of
	 *               this value is the one returned from the getValueClass method.
	 * @return The value presented as a displayable String.
	 * @see    LyFieldInfo#getValueClass
	 */
	public abstract String toName(Object fieldValue);

	/**
	 * Returns a "displayable string" (string that can be shown to the end-user) of the specified
	 * value for this field.
	 *
	 * @param  value The internal value to convert to a displayable string. The excepted class of
	 *               this value is the one returned from the getValueClass method.
	 * @param  removeTags Indicates whether HTML tags have to be removed from result.
	 * @return The value presented as a displayable String.
	 * @see    LyFieldInfo#getValueClass
	 */
	public String toName(Object fieldValue, boolean removeTags)
	{
		return toName(fieldValue);
	}

	/**
	 * Returns the specified value for this field as a short "displayable" string (short string that
	 * can be shown to the end-user). Default implementation consists of returning same result as
	 * toName.
	 *
	 * @param  value The internal value to convert to a short displayable string. The excepted class
	 *               of this value is the one returned from the getValueClass method.
	 * @return The value presented as a short displayable String.
	 * @see    LyFieldInfo#getValueClass
	 * @see    LyFieldInfo#toName
	 */
	public String toShortName(Object fieldValue)
	{
		String str = getFormattedShortName(fieldValue);

		return ((str == null) ? toName(fieldValue) : str);
	}

	/**
	 * Converts the specified value for this field to an "internal" string (string that enables
	 * reconstruction through parse, but not shown to the user). This string may not be significant
	 * to the end-user and usually should not be displayed.
	 *
	 * @param  value The internal value to convert to an internal string. The excepted class of this
	 *               value is the one returned from the getValueClass method.
	 * @return the value as an internal String.
	 * @see    LyFieldInfo#parse
	 * @see    LyFieldInfo#getValueClass
	 */
	public abstract String toString(Object fieldValue);

	/**
	 * Returns a formatted "edition string" of the specified value for this field formatted by field
	 * unit format if found.
	 *
	 * @param  value The internal value to convert to an edition string.
	 * @return The value presented as a formatted edition String.
	 */
	protected String getFormattedString(Object fieldValue)
	{
		if ((_unit == null) || (_unit.getFormat() == null))
			return null;

		return _unit.getFormat().toString(this, fieldValue, _unit);
	}

	/**
	 * Returns a formatted "displayable string" of the specified value for this field formatted by
	 * field unit format if found.
	 *
	 * @param  value The internal value to convert to a displayable string.
	 * @return The value presented as a formatted displayable String.
	 */
	protected String getFormattedName(Object fieldValue)
	{
		if (hasMark(MARK_SECRET))
			return SECRET_VALUE;

		if ((_unit == null) || (_unit.getFormat() == null))
			return null;

		return _unit.getFormat().toName(this, fieldValue, _unit);
	}

	/**
	 * Returns the specified value for this field as a short "displayable" string formatted by field
	 * unit format if found.
	 *
	 * @param  value The internal value to convert to a short displayable string.
	 * @return The value presented as a short formatted displayable String.
	 */
	protected String getFormattedShortName(Object fieldValue)
	{
		if (hasMark(MARK_SECRET))
			return SECRET_VALUE;

		if ((_unit == null) || (_unit.getFormat() == null))
			return null;

		return _unit.getFormat().toShortName(this, fieldValue, _unit);
	}

	/**
	 * Unitary control of the given value and produce a legible message to the end user if the value
	 * is incorrect. This method is called when validating a form containing this field info.
	 * Default implementation calls check value to prevent cases where it was not done on changes.
	 *
	 * @param  environment Environment giving the user language if a message has to be produced.
	 * @param  value       Value to control. The excepted class of this value is the one returned
	 *                     from the getValueClass method.
	 * @return Returns an error message if the value is incorrect, null otherwise.
	 * @see    LyFieldInfo#getValueClass
	 */
	public String controlValue(LyEnvironment environment, Object fieldValue)
	{
		return checkValue(environment, fieldValue);
	}

	/**
	 * Unitary check of the given value and produce a legible message to the end user if the value
	 * is incorrect. This method is called when changing in a form the value corresponding to this
	 * field info.
	 *
	 * @param  environment Environment giving the user language if a message has to be produced.
	 * @param  value       Value to check. The excepted class of this value is the one returned from
	 *                     the getValueClass method.
	 * @return Returns an error message if the value is incorrect, null otherwise.
	 * @see    LyFieldInfo#getValueClass
	 */
	public String checkValue(LyEnvironment environment, Object fieldValue)
	{
		// No generic check is done, so return null.
		return null;
	}

	/**
	 * Compares two values for the specified field. The excepted class of the values to compare is
	 * the one returned from the getValueClass method.
	 *
	 * @param  value1    First value.
	 * @param  value2    Second value.
	 * @param  modifiers Modifiers used for checking (see MOD_XXX constants of LyFilterExpression).
	 * @return 0 If the two values are identical for the specified fieldInfo, a positive number if
	 *         value1 is "greater" than value2. If value2 is "greater" than value1, a negative
	 *         number is returned.
	 * @see    LyFieldInfo#getValueClass
	 */
	public abstract int compare(Object value1, Object value2, BitSet modifiers);

	/**
	 * Compares two values for the specified field. The excepted class of the values to compare is
	 * the one returned from the getValueClass method.
	 *
	 * @param  value1 First value.
	 * @param  value2 Second value.
	 * @return 0 If the two values are identical for the specified fieldInfo, a positive number if
	 *         value1 is "greater" than value2. If value2 is "greater" than value1, a negative
	 *         number is returned.
	 * @see    LyFieldInfo#getValueClass
	 */
	public final int compare(Object value1, Object value2)
	{
		return compare(value1, value2, null);
	}

	/**
	 * Compares two values for equality on this field. The excepted class of the values to compare
	 * is the one returned from the getValueClass method. The default implementation consits of
	 * comparing values using compare method and returning true when result is 0. Due to the fact
	 * that this method can be called often, this method is provided to speed up this compare when
	 * it's possible to do better.
	 *
	 * @param  value1 First value.
	 * @param  value2 Second value.
	 * @return True if the values are equals, else false.
	 * @see    LyFieldInfo#compare
	 * @see    LyFieldInfo#getValueClass
	 */
	public final boolean equals(Object value1, Object value2)
	{
		return equals(value1, value2, null);
	}

	/**
	 * Compares two values for equality on this field. The excepted class of the values to compare
	 * is the one returned from the getValueClass method. The default implementation consits of
	 * comparing values using compare method and returning true when result is 0. Due to the fact
	 * that this method can be called often, this method is provided to speed up this compare when
	 * it's possible to do better.
	 *
	 * @param  value1    First value.
	 * @param  value2    Second value.
	 * @param  modifiers Modifiers used for checking (see MOD_XXX constants of LyFilterExpression).
	 * @return True if the values are equals, else false.
	 * @see    LyFieldInfo#compare
	 * @see    LyFieldInfo#getValueClass
	 */
	public boolean equals(Object value1, Object value2, BitSet modifiers)
	{
		return (compare(value1, value2, modifiers) == 0);
	}

	/**
	 * Checks if the second value is contained in the first one, for this field. Warning ! This kind
	 * of check does not make sense for every type of field and an exception can be thrown in that
	 * case. The excepted class of the values is the one returned from the getValueClass method.
	 *
	 * @param     value1 First value.
	 * @param     value2 Second value.
	 * @return    True if the second value is contained in the first one, else returns false.
	 * @exception LyUnsupportedComparisonException If it does not make sense to check if a value is
	 *                                             contained in an other one from this field "point
	 *                                             of view".
	 */
	public final boolean contains(Object value1, Object value2)
	{
		return contains(value1, value2, null);
	}

	/**
	 * Checks if the second value is contained in the first one, for this field. Warning ! This kind
	 * of check does not make sense for every type of field and an exception can be thrown in that
	 * case. The excepted class of the values is the one returned from the getValueClass method.
	 *
	 * @param     value1    First value.
	 * @param     value2    Second value.
	 * @param     modifiers Modifiers used for checking (see MOD_XXX constants of
	 *                      LyFilterExpression).
	 * @return    True if the second value is contained in the first one, else returns false.
	 * @exception LyUnsupportedComparisonException If it does not make sense to check if a value is
	 *                                             contained in an other one from this field "point
	 *                                             of view".
	 */
	public boolean contains(Object value1, Object value2, BitSet modifiers)
	{
		return equals(value1, value2, modifiers);
	}

	/**
	 * Checks if the second value starts the first one, for this field. Warning ! This kind of check
	 * does not make sense for every type of field and an exception can be thrown in that case. The
	 * excepted class of the values is the one returned from the getValueClass method.
	 *
	 * @param     value1    First value.
	 * @param     value2    Second value.
	 * @param     modifiers Modifiers used for checking (see MOD_XXX constants of
	 *                      LyFilterExpression).
	 * @return    True if the second value starts the first one, else returns false.
	 * @exception LyUnsupportedComparisonException If it does not make sense to check if a value
	 *                                             starts the other one from this field "point of
	 *                                             view".
	 */
	public boolean startsWith(Object value1, Object value2, BitSet modifiers)
	{
		LyEnvironment environment = getEnvironment();
		String		  message = environment.getMessage("LY_ER_UNSUPPORTED_COMPARISON", _id);

		throw new LyUnsupportedComparisonException("BF AppEngine : " + message);
	}

	/**
	 * Checks if the second value ends the first one, for this field. Warning ! This kind of check
	 * does not make sense for every type of field and an exception can be thrown in that case. The
	 * excepted class of the values is the one returned from the getValueClass method.
	 *
	 * @param     value1    First value.
	 * @param     value2    Second value.
	 * @param     modifiers Modifiers used for checking (see MOD_XXX constants of
	 *                      LyFilterExpression).
	 * @return    True if the second value ends the first one, else returns false.
	 * @exception LyUnsupportedComparisonException If it does not make sense to check if a value
	 *                                             ends the other one from this field "point of
	 *                                             view".
	 */
	public boolean endsWith(Object value1, Object value2, BitSet modifiers)
	{
		LyEnvironment environment = getEnvironment();
		String		  message = environment.getMessage("LY_ER_UNSUPPORTED_COMPARISON", _id);

		throw new LyUnsupportedComparisonException("BF AppEngine : " + message);
	}

	/**
	 * Returns the filter conditions authorized on this field. The values must match the short
	 * constants defined in LyFilterExpression.
	 *
	 * @return An array of short corresponding to the authorized conditions on this field.
	 * @see    leon.info.LyFilterExpression
	 */
	public LyFilterExpression.Condition[] getFilterConditions()
	{
        LyFilterExpression.Condition[] conditions = new  LyFilterExpression.Condition[9];

		// Default : return all the conditions.
		conditions[0] = LyFilterExpression.Condition.EQUALS;
		conditions[1] = LyFilterExpression.Condition.LOWER_THAN;
		conditions[2] = LyFilterExpression.Condition.LOWER_OR_EQUALS;
		conditions[3] = LyFilterExpression.Condition.GREATER_THAN;
		conditions[4] = LyFilterExpression.Condition.GREATER_OR_EQUALS;
		conditions[5] = LyFilterExpression.Condition.CONTAINS;
		conditions[6] = LyFilterExpression.Condition.STARTS_WITH;
		conditions[7] = LyFilterExpression.Condition.ENDS_WITH;

		return conditions;
	}

	/**
	 * Returns the filter modifiers authorized on this field. The values must match the short
	 * constants defined in LyFilterExpression.
	 *
	 * @return An array of short corresponding to the authorized modifiers on this field.
	 * @see    leon.info.LyFilterExpression
	 */
	public short[] getFilterModifiers()
	{
		// By default, no modifiers
		return null;
	}

	/**
	 * Decodes peer value.
	 */
	protected final Object decodePeerValue(Object object)
	{
	    Class<?>          valueClass = getValueClass();
	    LyFieldBinding    binding = getBinding();

	    if (binding != null)
	        object = binding.decodeValue(this, object, valueClass);
	    return object;
	}

	/**
	 * Builds a logical value for this fieldInfo from a physical (peer) value. This method is used
	 * when the binding specified for this field is a LyFieldBinding.
	 *
	 * @param     object   The peer value.
	 * @param     instance Instance containing the required value (could be null). This instance is
	 *                     provided to be able to construct a value for complex bindings (bindings
	 *                     involving other fields).
	 * @return    The logical value that was built (instance of the class returned from
	 *            getValueClass).
	 * @exception LyUnsupportedFieldBindingException Thrown if we don't known how to build a value
	 *                                               from the given object.
	 * @see       LyFieldInfo#getPeerValue
	 * @see       LyFieldBinding
	 */
	public Object getValueFromObject(Object object, LyObject instance)
	{
        if (object == null)
        {
            if ((_formula != null) && hasMark(LyFieldInfo.MARK_LOCAL))
                return _formula.getValue(instance);
            else
                return null;
        }

		if (object.getClass() == getValueClass())
			return object;

		throw new LyUnsupportedFieldBindingException(
			"BF AppEngine : don't known how to build a value for the field : "
			+ this + " from object " + object + "(" + object.getClass()
			+ ") on instance " + instance);
	}

	/**
     * Decodes value.
     */
    protected final Object encodePeerValue(Object value, Class<?> objectClass)
    {
        LyFieldBinding    binding = getBinding();

        if (binding != null)
            value = binding.encodeValue(this, value, objectClass);
        return value;
    }

	/**
	 * Returns an object of the given class from a value of this fieldInfo. The purpose of this
	 * method is to support transformations from logical to physical representation when simple
	 * bindings are used.
	 *
	 * @param     value       Value of this fieldInfo. This value is an instance of the class
	 *                        returned by getValueClass method.
	 * @param     objectClass Expected class of the physical value.
	 * @param     instance    Instance containing the given value (could be null).
	 * @exception LyUnsupportedFieldBindingException Thrown if we don't known how to build a peer
	 *                                               value from the given parameters.
	 * @see       LyFieldInfo#getValueClass
	 */
	public Object getPeerValue(Object value, Class<?> objectClass, LyObject instance)
	{
	    if (value == null)
	        return null;

        /* throw new LyUnsupportedFieldBindingException("BF AppEngine : can't create a peer value of " + objectClass
            + " for field info " + this + " from given value " + value + " (" + value.getClass()
            + ") for instance " + instance); */
	    return value;
	}

	/**
     * Decodes value.
     */
	protected final Object[] encodePeerValues(Object value, Class<?>[] objectClasses)
    {
        if ((value != null)
                && (value.getClass() != getValueClass()))
            throw new IllegalArgumentException("BF AppEngine : invalid value class " + value.getClass() + " expected "
                + getValueClass());

        LyFieldBinding    binding = getBinding();
        Object[]          values = null;

        if (binding instanceof LyFieldsBinding)
            values = ((LyFieldsBinding)binding).encodeValues(this, value, objectClasses);

        return values;
    }

	/**
	 * Returns a table of objects for the given classes from a value of this fieldInfo. The purpose
	 * of this method is to support transformations from logical to physical representation when
	 * multiple bindings are used.
	 *
	 * @param     value         Value of this fieldInfo. This value is an instance of the class
	 *                          returned by getValueClass method.
	 * @param     objectClasses Expected classes of the physical values.
	 * @param     instance      Instance containing the given value (could be null).
	 * @exception LyUnsupportedFieldBindingException Thrown if we don't known how to build the peer
	 *                                               values from the given parameters.
	 * @see       LyFieldInfo#getValueClass
	 */
	public Object[] getPeerValues(Object value, Class<?>[] objectClasses, LyObject instance)
	{
	    if (value == null)
	        return null;

        /* StringBuffer buffer = new StringBuffer();
        int          i, n = (objectClasses == null ? 0 : objectClasses.length);

        buffer.append("BF AppEngine : can't create a peer value of [");
        for (i = 0; i < n; i++)
        {
            if (i > 0)
                buffer.append(", ");
            buffer.append(objectClasses[i]);
        }

        buffer.append("] for field info ").append(this).append(" from given value ").append(value);
        buffer.append(" (").append(value.getClass()).append(") for instance ").append(instance);

        throw new LyUnsupportedFieldBindingException(buffer.toString()); */
	    if (value.getClass().isArray())
	        return (Object[])value;
	    else
	        return new Object[] { value };
	}

	/**
	 * Method for computing value for local fields.
	 *
	 * @param     object Object for which the value of this field is computed.
	 * @return    The computed value if it could be done. The returned value is an instance of the
	 *            class given by getValueClass().
	 * @exception IllegalStateException Thrown if values of this fieldInfo are not computed.
	 * @see       LyFieldInfo#getValueClass
	 */
	public Object compute(LyObject object)
	{
		if (!isComputed())
			return null;

		if (_formula != null)
		    return _formula.getValue(object);

		LyApplication     application = getApplication();
		LyFieldInfoDaemon daemon = application.getDaemon(this);

		return daemon.getValue(object);
	}

	/**
     * Method for evaluating the value corresponding to this field with the given values set. If
     * this number field has a formula, it's used to calculate the returned value.
     *
     * @param  values Values set that contains part or all of the values needed to compute the
     *                formula. If not all values are given, the formula uses the previous value
     * @return The value corresponding to this field using the given values. The returned value is
     *         an instance of the class given by getValueClass method.
     */
    public Object compute(LyValueSet values)
    {
        return (_formula == null ? null : _formula.getValue(values));
    }

	/**
	 * Returns a boolean indicating is the value of this fieldInfo is computed or not. Returns true
	 * if the field is local and controlled by a daemon.
	 *
	 * @return True if the relation is computed, false otherwise.
	 */
	public boolean isComputed()
	{
	    if (_formula != null)
	        return true;

		if (!isLocal())
			return false;

		// Check whether this relation is managed by a daemon.
		LyApplication     application = getApplication();
		LyFieldInfoDaemon daemon = application.getDaemon(this);

		return (daemon != null);
	}

	/**
	 * Returns a boolean indicating is the value of this fieldInfo is imported or not (default is
	 * false).
	 *
	 * @return True if values of this fieldInfo are imported, false otherwise.
	 */
	public boolean isImported()
	{
		return (_import != null);
	}

	/**
     * Method for computing a value using function associated to this field for a given objects
     * list. This method is called when this field is presented in a table that has a footer.
     *
     * @param  objectList List of objects for which a footer value is computed.
     * @return Value computed with field function for the given objects. Returns null if this field
     *         has no associated function.
     */
    public final Object compute(LyObjectList objectList)
    {
        return compute(objectList, false);
    }

	/**
	 * Method for computing a value using function associated to this field for a given objects
	 * list. This method is called when this field is presented in a table that has a footer.
	 *
	 * @param  objectList  List of objects for which a footer value is computed.
	 * @param  importField True if calling comes on computing of a field import, false otherwise.
	 * @return Value computed with field function for the given objects. Returns null if this field
	 *         has no associated function.
	 */
	public Object compute(LyObjectList objectList, boolean importField)
	{
		return null;
	}

	/**
	 * Convert a value to the reference unit from the given unit.
	 *
	 * @param  value The value to convert.
	 * @param  unit  The old unit.
	 * @return The value converted to the given unit from the reference unit.
	 */

	public Object convertFromFieldUnit(Object value, LyUnit unit)
	{
		if (value == null)
			return null;
		if (value.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value.getClass() + " expected " + getValueClass());
		if ((unit == null)
				|| (unit == _unit))
			return value;

		Class<?> converterClass = unit.getConverterClass();

		if (converterClass == null)
		{
			converterClass = getUnits().getConverterClass();
			if (converterClass == null)
				return value;
		}

		try
		{
			LyConverterInterface converter = (LyConverterInterface)converterClass.newInstance();

			value = converter.convertFromFieldUnit(this, value, unit);
		}
		catch (Exception e)
		{
			// Dump stack
			getEnvironment().writeException(e);
			return null;
		}

		if (value.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value.getClass() + " expected " + getValueClass());
		return value;
	} // end method convertFromFieldUnit

	/**
	 * Convert a value from the reference unit to the given one.
	 *
	 * @param  value The value to convert.
	 * @param  unit  The new unit.
	 * @return The value converted from the given unit to the reference unit.
	 */

	public Object convertToFieldUnit(Object value, LyUnit unit)
	{
		if (value == null)
			return null;
		if (value.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value.getClass() + " expected " + getValueClass());
		if ((unit == null)
				|| (unit == _unit))
			return value;

		Class<?> converterClass = unit.getConverterClass();

		if (converterClass == null)
		{
			LyUnits units = getUnits();

			if (units == null)
				// No converter found so no conversion made !!!
				return value;

			converterClass = getUnits().getConverterClass();

			if (converterClass == null)
				// No converter found so no conversion made !!!
				return value;
		}

		try
		{
			LyConverterInterface converter = (LyConverterInterface)converterClass.newInstance();

			value = converter.convertToFieldUnit(this, value, unit);
		}
		catch (Exception e)
		{
			// Dump stack
			getEnvironment().writeException(e);
			return null;
		}

		if (value.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value.getClass() + " expected " + getValueClass());
		return value;
	} // end method convertToFieldUnit

	/**
	 * Get field dependencies. A dependency means that a change on this field impacts the given
	 * field. If both fields are presented in a view, when a change occurs on this field graphical
	 * items representing both fields need to be refresh.
	 */
	public final LyInfoDataList<LyFieldInfo, Boolean> getDepends()
	{
		return getDepends(false);
	}

	/**
	 * Get field dependencies. A dependency means that a change on this field impacts the given
	 * field. If both fields are presented in a view, when a change occurs on this field graphical
	 * items representing both fields need to be refresh.
	 *
	 * @param  strong If true, gives only strong dependencies.
	 * @return List of field infos that depend on this field info (the associated data is a Boolean
	 *         indicating if the dependency is strong or not).
	 */
	public LyInfoDataList<LyFieldInfo, Boolean> getDepends(boolean strong)
	{
		if (_depends == null)
			return null;

		// If needed, lock depends list before returning it
		if (_depends.isWritable())
			_depends.lock();

		if (!strong)
			return _depends;

		LyInfoDataList<LyFieldInfo, Boolean> strongDepends =
			new LyInfoDataList<LyFieldInfo, Boolean>();
		int									 i, n = _depends.getSize();
		Boolean								 data;
		LyFieldInfo							 fieldInfo;

		for (i = 0; i < n; i++)
		{
			data = _depends.getData(i);
			if (data.booleanValue())
			{
				fieldInfo = _depends.getInfo(i);
				strongDepends.addInfo(fieldInfo, data);
			}
		}

		return strongDepends;
	} // end method getDepends

	/**
	 * Return the number of controls associated to this field.
	 *
	 * @return Number of controls of this field.
	 */
	public int getControlCount()
	{
		load();
		return ((_controls == null) ? 0 : _controls.getSize());
	}

	/**
	 * Fetches the control at the specified index.
	 *
	 * @param  index Index of the searched control.
	 * @return The requested control if found, or null if it could not be found.
	 */
	public LyControl getControl(int index)
	{
		load();
		return ((_controls == null) ? null : _controls.getInfo(index));
	}

	/**
	 * Get field controls.
	 */
	public LyInfoList<LyControl> getControls()
	{
		load();
		return _controls;
	}

	/**
	 * Get field units.
	 *
	 * @return Units list.
	 */
	public LyUnits getUnits()
	{
		load();
		return _units;
	}

	/**
	 * Get field unit.
	 *
	 * @return Unit of the field.
	 */
	public LyUnit getFieldUnit()
	{
		load();
		if ((_unit == null) && (_units != null))
			_unit = _units.getReferenceUnit();
		return _unit;
	}

	/**
	 * Gets behavior class name.
	 *
	 * @return Java name of behavior (may be null).
	 */
	public String getBehaviorClassName()
	{
		return _behaviorClassName;
	}

	/**
	 * Gets behavior class. Class that defines behavior for this field info for filtering and
	 * sorting.
	 *
	 * @return Behavior class for this field info (may be null).
	 */
	public Class<?> getBehaviorClass()
	{
		// loading classes
		Class<?> behaviorClass = null;

		if (_behaviorClassName != null)
			behaviorClass = _application.loadClass(_behaviorClassName,
					"BF AppEngine : unknown behavior class " + _behaviorClassName
					+ " on field " + _id);

		return behaviorClass;
	}

	/**
	 * Gets java field name corresponding to this field when java code generation is used.
	 *
	 * @return Name of the java field corresponding to this fieldInfo. By default, if no java name
	 *         is set, this method returns the id of this field.
	 */
	public String getJavaName()
	{
		if (_javaName != null)
			return _javaName;
		
		_javaName = getDefaultJavaName();				
		return _javaName;
	}
	
	/**
	 * Gets default java field name corresponding to this field info.
	 *
	 * @return Name of the default java field corresponding to this fieldInfo.
	 */
	public final String getDefaultJavaName()
	{
		String	javaName = getId();
		
		return checkJavaName(javaName);
	}

	/**
	 * check javaName attribute, replace points with underscore.
	 *
	 * @return new javaName
	 */
	public String checkJavaName(String javaName)
	{
		LyInfo	parent = getParent();
		
		if ((parent != null)
				&& javaName.startsWith(parent.getId()))
			javaName = javaName.substring(parent.getId().length());
		
		// Avoid conflict with LyObject access methods
		try
		{
			Method	method = LyObject.class.getMethod("get" + javaName);
			
			if (method != null)
			{
				// Conflict with a LyObject method
				javaName = getId();
	
				if ((_application != null)
						&& javaName.startsWith(_application.getId()))
					javaName = javaName.substring(_application.getId().length());
			}
		}
		catch (Exception e)
		{
		}
		
		String[]     old_char = new String[] { ".", "{", "}" };
		StringBuffer sb = new StringBuffer(javaName);

		for (int i = 0; i < old_char.length; i++)
		{
			sb = new StringBuffer(sb.toString());
			int start = 0;

			while (sb.indexOf(old_char[i], start) != -1)
			{
				start = sb.indexOf(old_char[i], start);
				sb.replace(start, start + 1, "__");
			}
		}
		
		return sb.toString();
	}

	/**
	 * Accessor to the tab name. The tab information is used by generic processing to build Windows
	 * (for example modification window), and will add the corresponding control in a tab. Tab info
	 * may be null (do not add to any tab)
	 *
	 * @return the tab or null if no tab is requested.
	 */
	public final String getTab()
	{
		return (_fieldTab == null ? _tab : _fieldTab.getName());
	}

    /**
     * Accessor to the field  tab.
     *
     * @return the field tab or null if no field tab was defined on this field.
     */
    public final LyFieldTab getFieldTab()
    {
        return _fieldTab;
    }

    /**
     * Accessor to the field  group.
     *
     * @return the field group or null if no field group was defined on this field.
     */
    public final LyFieldGroup getFieldGroup()
    {
        return _fieldGroup;
    }

	/**
	 * Accessor to the beans information. A bean is an interface for a graphical object that may be
	 * used instead of standard objects for a given action.
	 *
	 * @param      actionId Action id for which a bean is searched.
	 * @return     Bean associated to the given action id, null if no bean is specified on this
	 *             action.
	 * @deprecated Use getFieldBean
	 */
	public final String getBean(String actionId)
	{
		LyFieldBean bean = getFieldBean(actionId);

		return ((bean == null) ? null : bean.getJavaName());
	}

	/**
	 * Accessor to the field bean. A bean is an interface for a graphical object that may be used
	 * instead of standard objects for a given action.
	 *
	 * @param  actionId Action id for which a bean is searched.
	 * @return Bean associated to the given action id, null if no bean is specified on this action.
	 */
	public final LyFieldBean getFieldBean(String actionId)
	{
		if (_beans == null)
			return null;

		return _beans.get(actionId);
	}

	/**
	 * Gets list of beans defined for one field info.
	 *
	 * @return Vector containing all beans, null if no beans was registered for this field info.
	 */
	public final Vector<LyFieldBean> getBeans()
	{
		if (_beans == null)
			return null;

		Vector<LyFieldBean> beans = new Vector<LyFieldBean>(3);
		LyFieldBean		    bean;

		bean = getFieldBean(LyAction.ID_CONSULT);
		if (bean != null)
			beans.add(bean);
		bean = getFieldBean(LyAction.ID_SET);
		if ((bean != null) && !beans.contains(bean))
			beans.add(bean);
		bean = getFieldBean(LyAction.ID_NEW_FILTER);
		if ((bean != null) && !beans.contains(bean))
			beans.add(bean);

		return beans;
	}

	/**
	 * Reads the binding information. Generic or specific connection classes use the binding
	 * information to fetch data from and to a data provider. For instance the binding info may be a
	 * column name in a RDBMS.
	 *
	 * @return the binding information.
	 */
	public LyFieldBinding getBinding()
	{
		if ((_binding == null) && !isLocal())
        {
			// Build a default binding.
			_binding = new LyFieldBinding(getId(), null, null, false);
			_binding.setDefaultBinding(true);
        }

		return _binding;
	}

	/**
	 * Gets the format as a string. The meaning of this format depends on each type of field.
	 *
	 * @return Current format or null if no format is specified for this field.
	 */
	public String getFormat()
	{
		return _format;
	}

	/**
	 * Gets the default value. This default value if an instance of the class returned by
	 * getValueClass.
	 *
	 * @return Default value for this fieldInfo, null if no default value was set.
	 */
	public Object getDefaultValue()
	{
		return _defaultValue;
	}

	/**
	 * Tests if the field is calculated.
	 *
	 * @return True if the field is calculated, false otherwise
	 */
	public boolean isCalculated()
	{

		return _calculated
			|| ((_template != null) && (_template != this) && ((LyFieldInfo)_template).isCalculated());
	}

	/**
	 * Tests if the field is multiple.
	 *
	 * @return True if this field is multiple, false otherwise.
	 */
	public boolean isMultiple()
	{
		return false;
	}

	/**
	 * Tests if the field is local. This means that this field has no physical representation and
	 * thus is not persistent.
	 *
	 * @return True if this field is local, false otherwise.
	 */
	public final boolean isLocal()
	{
        if (_local != null)
            return _local.booleanValue();

		// Check whether this field has LOCAL mark or not.
		boolean   local = hasMark(MARK_LOCAL);

        if (_locked)
            _local = Boolean.valueOf(local);
        return local;
	}

	/**
	 * Gets import information for this field info. Import allows to use a relation of the same
	 * class info to get the value of a field of the object(s) pointed by the relation.
	 *
	 * @return Import information for this fieldInfo, null if no import was defined.
	 */
	public LyFieldInfoImport getImport()
	{
		return _import;
	}

	/**
	 * Gets path information for this field info. Path allows to get the value of a field of the
	 * objects pointed by this path.
	 *
	 * @return Path information for this fieldInfo, null if no path was defined.
	 */
	public LyPath getPath()
	{
		return _path;
	}

	/**
	 * Gets field constraint used in forms layouts. This applies to the content of the edition
	 * field, the related label is managed by a caption constraint (getCaptionConstraint).
	 *
	 * @return Field constraint of this fieldInfo, null if no constraint was defined.
	 */
	public final LyFieldConstraint getConstraint()
	{
	    return getConstraint(LyFieldConstraint.Type.FIELD, null);
	}

	/**
	 * Gets field caption (label) constraint used in forms layouts. This applies to the caption
	 * (label) of the edition field. The constraint for the field itself is managed by getConstraint
	 *
	 * @return Caption constraint of this fieldInfo, null if no constraint was defined.
	 */
	public final LyFieldConstraint getCaptionConstraint()
	{
		return getConstraint(LyFieldConstraint.Type.CAPTION, null);
	}

	/**
     * Gets field constraint used in forms layouts.
     * @param type Type of the requested constraint.
     * @param action Action where the constraint is used.
     * @return Constraint found, null if no constraint was defined.
     */
    public final LyFieldConstraint getConstraint(LyFieldConstraint.Type type, LyAction action)
    {
        if (_constraints == null)
            return null;

        int                 i, n = _constraints.size();
        LyFieldConstraint   constraint, defaultConstraint = null;

        for (i = 0; i < n; i++)
        {
            constraint = _constraints.elementAt(i);
            if (constraint.getType() == type)
            {
                if (constraint.getAction() == action)
                    return constraint;

                if ((defaultConstraint == null)
                		&& (action != null)
                		&& action.hasTemplate(constraint.getAction()))
                	defaultConstraint = constraint;

                if ((defaultConstraint == null)
                        && (constraint.getAction() == null)
                        && (action != null))
                    defaultConstraint = constraint;
            }
        }

        return defaultConstraint;
    }

    /**
     * Gets all constraints of this field info.
     */
    public final Vector<LyFieldConstraint> getConstraints()
    {
        return _constraints;
    }

	/**
	 * Tests if the field was built with a FIELD tag.
	 *
	 * @return True if this field was declared with a FIELD tag, false otherwise.
	 */
	public boolean fromFieldTag()
	{
		return _fieldTag;
	}

	/**
	 * The following modifiers are not public to avoid modification after parsing. They are provided
	 * to be used only by the data model parser. Visibility shouldn't be changed except for a VERY
	 * good reason.
	 */

	/**
	 * Sets the binding information. Generic or specific connection classes use the binding
	 * information to fetch data from and to a data provider. For instance the binding info may be a
	 * column name in a RDBMS.
	 *
	 * @param binding Binding information.
	 * @see   LyFieldBinding
	 */
	final void setBinding(LyFieldBinding binding)
	{
		_binding = binding;
	    if (binding != null)
	        binding.setParent(this);
	}

	/**
	 * Sets the tab information. The tab information is used by generic processing to build Windows
	 * (for example modification window), and will add the corresponding control in a tab. Tab info
	 * may be null (do not add to any tab)
	 *
	 * @param tab The tab name or null if no tab is requested.
	 */
	final void setTab(String tab)
	{
        if (tab == null)
        {
            _fieldTab = null;
            _tab = null;
            return;
        }

        if ((_fieldTab != null)
        		&& (_fieldTab.getParent() == this))
            _fieldTab.setName(tab);
        else
        {
            if (_fieldTabs == null)
                _fieldTabs = new Hashtable<String, LyFieldTab>();
            else
                _fieldTab = _fieldTabs.get(tab);

            if ((_fieldTab == null)
            	|| (_fieldTab.getParent() != this))
            {
                _fieldTab = new LyFieldTab(tab);
                _fieldTab.setParent(this);
                _fieldTabs.put(tab, _fieldTab);
            }
        }

		_tab = ((tab == null) ? null : tab.intern());
	}

    /**
     * Sets the field tab information.
     *
     * @param tab The field tab or null if no tab is requested.
     */
    final void setFieldTab(LyFieldTab fieldTab)
    {
        _fieldTab = fieldTab;
        _tab = (fieldTab == null ? null : fieldTab.getName());
    }

    /**
     * Sets the group name to which this instance belongs.
     *
     * @param group Group name.
     */
    void setGroup(String group)
    {
        if (group == null)
        {
            _fieldGroup = null;
            _group = null;
            return;
        }

        if ((_fieldGroup != null)
        		&& (_fieldGroup.getParent() == this))
            _fieldGroup.setName(group);
        else
        {
            if (isExtern()
                    && ((_template == null) || (_group != _template._group)))
                return; // may happen on extern infos

            if (_fieldGroups == null)
                _fieldGroups = new Hashtable<String, LyFieldGroup>();
            else
                _fieldGroup = _fieldGroups.get(group);

            if ((_fieldGroup == null)
                || (_fieldGroup.getParent() != this))
            {
                _fieldGroup = new LyFieldGroup(group);
                _fieldGroup.setParent(this);
                _fieldGroups.put(group, _fieldGroup);
            }
        }

        super.setGroup(group);
    }

    /**
     * Sets the field group information.
     *
     * @param tab The field group or null if no tab is requested.
     */
    final void setFieldGroup(LyFieldGroup fieldGroup)
    {
        _fieldGroup = fieldGroup;
        _group = (fieldGroup == null ? null : fieldGroup.getName());
    }

	/**
	 * Sets the display format for this field. The meaning of this format depends on each type of
	 * field.
	 *
	 * @param format The format or null if no format is specified
	 */
	void setFormat(String format)
	{
		_format = ((format == null) ? null : format.intern());
	}

	/**
	 * Add a dependency on the given fieldInfo.
	 *
	 * @param fieldInfo Field for which a dependency on this field is added.
	 */
	final public void addDepend(LyFieldInfo fieldInfo)
	{
		addDepend(fieldInfo, true);
	}

	/**
	 * Add a dependency on the given fieldInfo.
	 *
	 * @param fieldInfo Field for which a dependency on this field is added.
	 * @param strong    True if the dependency between this field and the given one is strong (they
	 *                  must be modified simultaneously).
	 */
	final public void addDepend(LyFieldInfo fieldInfo, boolean strong)
	{
		if (_depends == null)
			_depends = new LyInfoDataList<LyFieldInfo, Boolean>();

		int     index = _depends.getIndex(fieldInfo);
		Boolean dependencyValue = ((index >= 0) ? (Boolean)_depends.getData(fieldInfo) : null);

		// If dependency is not registered, add it with the strong parameter.
		// Otherwise, register it only if dependency was weak and becomes strong.
		if (dependencyValue == null)
			_depends.addInfo(fieldInfo, strong ? STRONG_DEPENDENCY_VALUE : WEAK_DEPENDENCY_VALUE);
		else if (strong && (dependencyValue == WEAK_DEPENDENCY_VALUE))
		{
			_depends.removeInfo(fieldInfo, WEAK_DEPENDENCY_VALUE);
			_depends.addInfo(fieldInfo, STRONG_DEPENDENCY_VALUE);
		}
	}

	/**
	 * Adds a bean to this field info.
	 *
	 * @param bean the field bean.
	 */
	final void addBean(LyFieldBean bean)
	{
		if (_beans == null)
			_beans = new Hashtable<String, LyFieldBean>(5);

        LyFieldBean.Type actionType = bean.getActionType();

		if (actionType == LyFieldBean.Type.ALL)
		{
			_beans.put(LyAction.ID_CONSULT, bean);
			_beans.put(LyAction.ID_CREATE, bean);
			_beans.put(LyAction.ID_SET, bean);
			_beans.put(LyAction.ID_CLONE, bean);
			_beans.put(LyAction.ID_NEW_FILTER, bean);
			_beans.put(LyAction.ID_SET_FILTER, bean);
		}
		else if (actionType == LyFieldBean.Type.CONSULT)
			_beans.put(LyAction.ID_CONSULT, bean);
		else if (actionType == LyFieldBean.Type.EDIT)
		{
			_beans.put(LyAction.ID_CREATE, bean);
			_beans.put(LyAction.ID_SET, bean);
			_beans.put(LyAction.ID_CLONE, bean);
		}
		else if (actionType == LyFieldBean.Type.FILTER)
		{
			_beans.put(LyAction.ID_NEW_FILTER, bean);
			_beans.put(LyAction.ID_SET_FILTER, bean);
		}
		bean.setParent(this);
	}

	/**
	 * Sets default value for this field info.
	 *
	 * @param defaultValue Default value.
	 */
	void setDefaultValue(Object defaultValue)
	{
		_defaultValue = defaultValue;
	}

	/**
	 * Sets the calculated state of this field info.
	 *
	 * @param isCalculated the calculated state
	 */
	void setCalculated(boolean isCalculated)
	{
		_calculated = isCalculated;
	}

	/**
	 * Builds the appropriate field value corresponding to this field from a value given as an
	 * Object.
	 *
	 * @deprecated Replaced by buildFieldValue
	 */
	public final LyValue buildLyValue(Object value)
	{
		return buildFieldValue(value);
	}

	/**
	 * Sets import information for this field info. Import allows to use a relation of the same
	 * class info to get the value of a field of the object(s) pointed by the relation.
	 *
	 * @param importInfo Import information.
	 */
	public void setImport(LyFieldInfoImport importInfo)
	{
		_import = importInfo;
	}

	/**
	 * Sets path information for this field info. Path allows to get the value of a field of the
	 * object(s) pointed this path.
	 *
	 * @param path Path information.
	 */
	public void setPath(LyPath path)
	{
		_path = path;
	}

	/**
	 * Adds a constraint used in forms layouts.
	 *
	 * @param constraint Constraint of this fieldInfo.
	 */
	public void addConstraint(LyFieldConstraint constraint)
	{
	    if (_constraints == null)
	        _constraints = new Vector<LyFieldConstraint>(2);
	    else
	    {
	        LyFieldConstraint   oldConstraint = getConstraint(constraint.getType(), constraint.getAction());

	        if (oldConstraint == constraint)
	            return;
	        else if (oldConstraint != null)
	            _constraints.remove(oldConstraint);
	    }

		_constraints.add(constraint);
	}

	/**
	 * Accessor to the marks attached to this instance (presented as a bitset).
	 *
	 * @return A copy of the bitset which represents the marks (properties) of this instance.
	 * @see    java.util.BitSet
	 */
	protected final BitSet getMarks(int parts)
	{
		return getMarks(_marks, parts);
	}

	/**
	 * Accessor to the marks attached to this instance (presented as a bitset).
	 *
	 * @return A copy of the bitset which represents the marks (properties) of this instance.
	 * @see    java.util.BitSet
	 */
	public static final BitSet getMarks(BitSet marks, int parts)
	{
		if (marks == null)
			return null;

		BitSet newMarks = (BitSet)marks.clone();

		if ((parts & PART_GUI) > 0)
		{
			// Clear data marks
			newMarks.clear(MARK_ID);
			newMarks.clear(MARK_SECRET);
			newMarks.clear(MARK_UNIQUE);
			newMarks.clear(MARK_READ_ONLY);
			newMarks.clear(MARK_APP_ID);
			// Clear bind "marks"
			newMarks.clear(MARK_LOCAL);
			newMarks.clear(MARK_LOAD);
			newMarks.clear(MARK_NOT_NULL);
			newMarks.clear(MARK_PROVIDER_VALUE);
			newMarks.clear(MARK_PROVIDER_VALUE_BIND);
		}
		if ((parts & PART_DATA) > 0)
		{
			// Only accept data marks
			BitSet validMarks = new BitSet();

			validMarks.set(MARK_ID);
			validMarks.set(MARK_OPTIONAL);
			validMarks.set(MARK_SECRET);
			validMarks.set(MARK_UNIQUE);
			validMarks.set(MARK_READ_ONLY);
			validMarks.set(MARK_APP_ID);
			newMarks.and(validMarks);
		}
		if ((parts & PART_BIND) > 0)
		{
			// Only accept bind marks
			BitSet validMarks = new BitSet();

			validMarks.set(MARK_LOCAL);
			validMarks.set(MARK_LOAD);
			validMarks.set(MARK_NOT_NULL);
			validMarks.set(MARK_PROVIDER_VALUE);
			newMarks.and(validMarks);
		}

		return newMarks;
	} // end method getMarks

	/**
	 * Adds a control on this field info.
	 *
	 * @param control Control to add to this field info.
	 */
	void addControl(LyControl control)
	{
		if (_controls == null)
			_controls = new LyInfoList<LyControl>();

		_controls.addInfo(control);
	}

	/**
	 * Set field units.
	 *
	 * @param units Units list.
	 */
	void setUnits(LyUnits units)
	{
		_units = units;
		if ((_units != null) && (_unit == null))
			_unit = units.getReferenceUnit();
	}

	/**
     * Set field unit.
     *
     * @param unit Field unit.
     */
    void setFieldUnit(LyUnit unit)
    {
        if ((unit != null)
                && (_units != null)
                && !_units.contains(unit))
            throw new IllegalStateException("BF AppEngine : can't set unit " + unit
                + " on field info " + this
                + " because this unit is not contained in field info units list");

        _unit = unit;
    }

	/**
	 * Set field unit.
	 *
	 * @deprecated see setFieldUnit.
	 */
	void setUnit(LyUnit unit)
	{
		setFieldUnit(unit);
	}

	/**
	 * Sets behavior class name.
	 *
	 * @param behaviorClassName Java name of behavior class (may be null).
	 */
	void setBehaviorClassName(String behaviorClassName)
	{
		_behaviorClassName = behaviorClassName;
	}

	/**
	 * Sets whether the field was built with a FIELD tag or not.
	 *
	 * @param fieldTag True if this field was declared with a FIELD tag, false otherwise.
	 */
	void setFieldTag(boolean fieldTag)
	{
		_fieldTag = fieldTag;
	}

	/**
	 * Add a new label to labels list of this field info.
	 *
	 * @param label Label to add.
	 */
	void addLabel(LyInfoLabel label)
	{
		if (label == null)
			throw new IllegalArgumentException("BF AppEngine : label can't be null");

		LyInfoLabel.Type type = label.getType();

		if (type == LyInfoLabel.Type.USER)
		{
			if (_userLabels == null)
				_userLabels = new Vector<LyInfoLabel>(1);
			_userLabels.add(label);
			return;
		}

		if ((type != LyInfoLabel.Type.NAME)
				&& (type != LyInfoLabel.Type.GRAMMATICAL)
				&& (type != LyInfoLabel.Type.SEMANTIC)
				&& (type != LyInfoLabel.Type.VALUE)
				&& (type != LyInfoLabel.Type.SHORT_VALUE)
				&& (type != LyInfoLabel.Type.USER))
			throw new IllegalArgumentException("BF AppEngine : invalid type of label on fieldInfo "
				+ this + ". Expecting USER, NAME, GRAMMATICAL, SEMANTIC or VALUE for label value '"
				+ label.getValue() + "'");

		if (_labels == null)
			_labels = new LyInfoLabel[7];
		_labels[type.ordinal()] = label;
	} // end method addLabel

	/**
	 * Returns the number of labels of this fieldInfo.
	 */
	public int getLabelCount()
	{
		return ((_labels == null) ? 0 : _labels.length);
	}

	/**
	 * Returns the the label of this fieldInfo for the specified type.
	 */
	public LyInfoLabel getLabel(LyInfoLabel.Type type)
	{
		if (_labels == null)
			return null;

		return _labels[type.ordinal()];
	}

    /**
     * Returns the label of this fieldInfo at the given index.
     */
    public LyInfoLabel getLabel(int index)
    {
        if ((_labels == null)
                || (index < 0)
                || (index > _labels.length))
            return null;

        return _labels[index];
    }

	/**
	 * Returns the number of labels of this fieldInfo.
	 */
	public int getUserLabelCount()
	{
		return ((_userLabels == null) ? 0 : _userLabels.size());
	}

	/**
	 * Returns the user label at the given position.
	 */
	public LyInfoLabel getUserLabel(int i)
	{
		return ((_userLabels == null) ? null : (LyInfoLabel)_userLabels.elementAt(i));
	}

	/**
	 * Returns the user label for the given subType.
	 */
	public LyInfoLabel getUserLabel(String subType)
	{
		int		    i, n = getUserLabelCount();
		LyInfoLabel label;

		if (subType != null)
			subType = subType.intern();
		for (i = 0; i < n; i++)
		{
			label = _userLabels.elementAt(i);
			if (label.getSubType() == subType)
				return label;
		}

		return null;
	}

	/**
	 * Returns all labels of this classInfo.
	 */
	public LyInfoLabel[] getLabels()
	{
		LyInfoLabel[] result = _labels;

		if ((_userLabels != null) && (_userLabels.size() > 0))
		{
			int labelsNb = (_labels == null ? 0 : _labels.length);
			result = new LyInfoLabel[labelsNb + _userLabels.size()];

			int beg = 0;

			for (int i = 0; i < labelsNb; i++)
				result[i] = _labels[i];

			beg = labelsNb;

			for (int i = 0; i < _userLabels.size(); i++)
				result[i + beg] = _userLabels.get(i);
		}

		return result;
	}

	/**
	 * Checks if a label belongs to this field info.
	 */
	protected boolean hasLabel(LyInfoLabel label)
	{
		if (_labels != null)
		{
			int i, n = _labels.length;

			for (i = 0; i < n; i++)
				if (_labels[i] == label)
					return true;
		}
		if ((_userLabels != null) && _userLabels.contains(label))
			return true;

		return false;
	}

	/**
	 * Check if a specific mark is set on this field info from its name.
	 *
	 * @param  markName Name of the mark tested for this field info.
	 * @return True if the mark is set, false otherwise.
	 */
	public final boolean hasMark(String markName)
	{
		return hasMark(getMark(markName));
	}

    /**
     * Sets field constraint used in forms layouts. This applies to the content of the edition
     * field, the related label is managed by a caption constraint (getCaptionConstraint).
     *
     * @param constraint Field constraint of this fieldInfo, null if no constraint was defined.
     */
    protected void setConstraint(LyFieldConstraint constraint)
    {
        LyFieldConstraint   oldConstraint = getConstraint();

        if (oldConstraint != constraint)
        {
            if (oldConstraint != null)
                _constraints.remove(oldConstraint);
            if (constraint != null)
                addConstraint(constraint);
        }
    }

    /**
     * Sets field caption (label) constraint used in forms layouts. This applies to the caption
     * (label) of the edition field. The constraint for the field itself is managed by getConstraint
     *
     * @param captionConstraint Caption constraint of this fieldInfo, null if no constraint was defined.
     */
    protected void setCaptionConstraint(LyFieldConstraint captionConstraint)
    {
        LyFieldConstraint   oldConstraint = getCaptionConstraint();

        if (oldConstraint != captionConstraint)
        {
            if (oldConstraint != null)
                _constraints.remove(oldConstraint);
            if (captionConstraint != null)
                addConstraint(captionConstraint);
        }
    }

    /**
     * Set field controls.
     */
    protected void setControls(LyInfoList<LyControl> controls)
    {
        int     i, n = (controls == null ? 0 : controls.getSize());

        _controls = null;
        for (i = 0; i < n; i++)
            addControl(controls.getInfo(i));
    }

    /**
     * Sets labels of this classInfo.
     */
    protected void setLabels(LyInfoLabel[] labels)
    {
        int     i, n = (labels == null ? 0 : labels.length);

        _labels = null;
        for (i = 0; i < n; i++)
            addLabel(labels[i]);
    }

    /**
     * Accessor to the generic marks attached to this instance.
     */
    public BitSet getGenericMarks()
    {
        return (new LyBitSet(LyBitSet.Type.F_MARKS, _marks)).getGenericMarks();
    }

    /**
     * Sets the generic marks attached to this instance.
     */
    protected void setGenericMarks(BitSet marks)
    {
        _marks = (new LyBitSet(LyBitSet.Type.F_MARKS, _marks)).setGenericMarks(marks);
    }

    /**
     * Accessor to the specific marks attached to this instance (presented as an array of strings).
     */
    public String[] getSpecificMarks()
    {
        return (new LyBitSet(LyBitSet.Type.F_MARKS, _marks)).getSpecificMarks();
    }

    /**
     * Sets the specific marks attached to this instance (presented as an array of strings).
     */
    protected void setSpecificMarks(String[] marks)
    {
        _marks = (new LyBitSet(LyBitSet.Type.F_MARKS, _marks)).setSpecificMarks(marks);
    }

    /**
     * Sets list of beans defined on this field info.
     */
    protected void setBeans(Vector<LyFieldBean> beans)
    {
        int     i, n = (beans == null ? 0 : beans.size());

        _beans = null;
        for (i = 0; i < n; i++)
            addBean(beans.elementAt(i));
    }

    /**
     * Gets sub-types (list of fields).
     *
     * @return Gets sub-types of this field.
     */
    protected LyInfoList<LyRelationFieldInfo> getSubTypes()
    {
        return null;
    }

    /**
     * Sets sub-types (list of fields).
     *
     * @param subTypes sub-types of this field.
     */
    protected void  setSubTypes(LyInfoList<LyRelationFieldInfo> subTypes)
    {
    }

    /**
     * Gets the formula (as a string) associated to this number field. The formula is provided to
     * calculate values of this field thanks to a simple mathematic expression using other numerc
     * fields in the same application class.
     *
     * @return Current formula or null if no formula is set.
     */
    public String getFormula()
    {
        return (_formula == null ? null : _formula.getExpression());
    }

    /**
     * Gets the formula associated to this control.
     *
     * @return Current formula or null if no formula is set.
     */
    public final LyFormula getLyFormula()
    {
        return _formula;
    }
    
    /**
     * Sets the formula associated to the field.
     *
     * @param     formula Current formula, or null if none is specified.
     * @exception ParseException If the given formula couldn't be parsed successfully.
     */
    void setFormula(String formula)
    {
        _formula = LyFormula.formulaFactory(this, formula);
    }

    /**
     * Gets field infos required for computing value of this field.
     *
     * @return The list of fields used in the field formula, null if this field doesn't have a
     *         formula.
     */
    public LyInfoList<LyFieldInfo> getFormulaFields()
    {
        return (_formula == null ? null : _formula.getFormulaFields());
    }

    /**
     * Gets field imports of the given class required for computing value of this field.
     * @param classInfo Class of the fields.
     *
     * @return The list of fields used in the field formula, null if this field doesn't have a
     *         formula.
     */
    public LyInfoList<LyFieldInfo> getFormulaImports(LyClassInfo classInfo)
    {
        return (_formula == null ? null : _formula.getImports(classInfo));
    }
} // end class LyFieldInfo
