/*
 * LyFieldInfoImport.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import leon.data.*;


/**
 * Container for import information.
 */
public class LyFieldInfoImport extends LyTag
{
	public static final String __VERSION =
		"$Id: LyFieldInfoImport.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Relation from which the value is imported. */
	LyRelationFieldInfo		   _relation;

	/** Import class. */
	LyClassInfo				   _classInfo;

	/** Filter applied on objects used for the import. */
	LyFilter				   _filter;

	/** Imported field. */
	LyFieldInfo				   _fieldInfo;

	/**
	 * Constructor.
	 */
	LyFieldInfoImport()
	{
		_relation = null;
		_classInfo = null;
		_filter = null;
		_fieldInfo = null;
	}

	/**
	 * Constructor.
	 */
	LyFieldInfoImport(LyRelationFieldInfo relation, LyClassInfo classInfo,
		LyFilter filter, LyFieldInfo fieldInfo)
	{
		_relation = relation;
		_classInfo = classInfo;
		_filter = filter;
		_fieldInfo = fieldInfo;
	}

	/**
	 * Returns the relation from which the value is imported.
	 */
	public LyRelationFieldInfo getRelation()
	{
		return _relation;
	}

	/**
	 * Returns the import class.
	 */
	public LyClassInfo getClassInfo()
	{
		if ((_classInfo == null)
		        && (_relation != null)
		        && (_relation.getClassInfoCount() == 1))
			_classInfo = _relation.getClassInfo(0);

		return _classInfo;
	}

	/**
	 * Returns the filter applied on objects used for the import.
	 */
	public LyFilter getFilter()
	{
		return _filter;
	}

	/**
	 * Returns the imported field.
	 */
	public LyFieldInfo getFieldInfo()
	{
		return _fieldInfo;
	}

	/**
	 * Sets the relation from which the value is imported.
	 */
	void setRelation(LyRelationFieldInfo relation)
	{
		_relation = relation;
	}

	/**
	 * Sets the import class.
	 */
	void setClassInfo(LyClassInfo classInfo)
	{
		_classInfo = classInfo;
	}

	/**
	 * Sets the import filter.
	 */
	void setFilter(LyFilter filter)
	{
		_filter = filter;
	}

	/**
	 * Sets the imported field.
	 */
	void setFieldInfo(LyFieldInfo fieldInfo)
	{
		_fieldInfo = fieldInfo;
	}

	/**
	 * Internal method to check validity for information of this import for the specified fieldInfo
	 * of a given class info.
	 *
	 * @exception IllegalStateException If fields of the import are not consistent.
	 */
	void check(LyClassInfo classInfo, LyFieldInfo fieldInfo)
	{
		// Relation must belong to the parent class
		if (!classInfo.contains(_relation))
			throw new IllegalStateException("BF AppEngine : invalid import relation " + _relation
				+ " for field " + fieldInfo + " in class " + classInfo
				+ ", field and relation must belong to the same class info");

		/* If the relation is multiple, this field must be a number
		if (relation.isMultiple()
		                && !(this instanceof LyNumberFieldInfo))
		        throw new IllegalStateException("BF AppEngine : invalid import relation " + _relation
		                + " for field " + this + " in class " + classInfo
		                + ", as the relation is multiple, the field must be a number");
		*/

		// Imported field must belong to the import class
		LyClassInfo importClassInfo = getClassInfo();

		if ((importClassInfo == null) || !importClassInfo.contains(_fieldInfo))
			throw new IllegalStateException("BF AppEngine : invalid imported field " + _fieldInfo
				+ " for field " + fieldInfo + " in class " + classInfo
				+ ", the field is not a field of class info " + importClassInfo);

		// Imported field must belong to the import class
		// except if fieldInfo is a number (count)
		if ((fieldInfo.getClass() != _fieldInfo.getClass())
				&& !(fieldInfo instanceof LyNumberFieldInfo))
			throw new IllegalStateException("BF AppEngine : invalid imported field " + _fieldInfo
				+ " for field " + fieldInfo + " in class " + classInfo
				+ ", the field and the imported field must be of the same type");
	} // end method check

	/**
	 * Utility method to compute import value for a fieldInfo from a relation value.
	 *
	 * @param     fieldInfo     Field info for which an imported value is requested.
	 * @param     relationValue Value from which the import is done. The relation of this relation
	 *                          value must match the one if this import.
	 * @return    Value imported from this field info import.
	 * @exception IllegalArgumentException If the given field info doesn't have this import as field
	 *                                     import.
	 * @exception IllegalArgumentException If the relation specified for the import doesn't
	 *                                     correspond to the one of the given relation value.
	 */
	public Object getImportValue(LyFieldInfo fieldInfo, LyRelationValue relationValue)
	{
		if (fieldInfo.getImport() != this)
			throw new IllegalArgumentException(
				"BF AppEngine : the current import is not the one of the given fieldInfo " + fieldInfo);

		if (relationValue == null)
			return null;

		LyRelationFieldInfo relation = getRelation();

		if ((relationValue.getFieldInfo() != relation) && !((LyRelationFieldInfo)relationValue.getFieldInfo()).inheritsFrom(relation))
			throw new IllegalArgumentException("BF AppEngine : the relation of the given value "
				+ relationValue.getFieldInfo() + " doesn't correspond to the one of the import "
				+ relation);

		Object value = relationValue.getValue();

		if (value == null)
			return null;

		LyObjectList objects = relation.getObjects(value);

		if ((objects == null) || (objects.getSize() == 0))
			return null;

		LyClassInfo importClass = getClassInfo();
		LyFilter    filter = getFilter();

		if ((importClass != null) || (filter != null))
		{
			objects = objects.getList(null, importClass, filter);
			if ((objects == null) || (objects.getSize() == 0))
				return null;
		}

		LyFieldInfo importedFieldInfo = getFieldInfo();

		if ((importedFieldInfo.getClass() != fieldInfo.getClass())
				&& (fieldInfo instanceof LyNumberFieldInfo))
		{
			LyNumberValue numberValue = new LyNumberValue((LyNumberFieldInfo)fieldInfo,
					objects.getSize());

			return numberValue.getValue();
		}

		Object importValue;

		if (objects.getSize() == 1)
			importValue = (objects.getObject(0)).getValue(importedFieldInfo);
		else
			importValue = importedFieldInfo.compute(objects, true);

		return importValue;
	} // end method getImportValue
} // end class LyFieldInfoImport
