/*
 * LyFileFieldInfo.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.io.*;
import java.util.*;
import leon.data.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.peer.util.*;


/**
 * This class implements the description of a file (path) field from an application class.
 * Displaying the content of this field will be up to the implementation of the graphical component.
 *
 * <p>Additionally, a fileFieldInfo may contain some extra information about the way the text should
 * be shown (number of rows/columns).</p>
 *
 * <p>See also tag <FILE> of leon.dtd</p>
 */
public final class LyFileFieldInfo extends LyFieldInfo
{
	public static final String __VERSION =
		"$Id: LyFileFieldInfo.java 32363 2011-09-16 15:44:58Z bangiole $";

	/** File types. */
	public enum Type
    {
        TEXT,
        IMAGE,
        URL,
        DIR
    }
    
	/** Value : absolute. */
	public static String	   VALUE_ABSOLUTE = "_absolute";

	
	/** List of supported types (mime types). */
	String					   _mimeTypes;

	/**
	 * Number of rows which should be shown when displaying the file (-1 if no preference is given).
	 */
	int						   _nRows;

	/**
	 * Number of columns which should be shown when displaying the file (-1 if no preference is
	 * given).
	 */
	int						   _nCols;

	/** Attachment directory where the attachment are saved on server. */
	String					   _attachmentDirectory;

	/**
	 * Attachment handler in charge or saving and fetching attachments on server. Default value is
	 * null.
	 */
	LyJavaName				   _attachmentHandler;

	/** Directory where file value is searched. */
	String					   _directory;

	/** Filter for file value. */
	String					   _filter;

	/**
	 * Init method.
	 *
	 * @see LyObjectInfo#init
	 */
	protected void init()
	{
		super.init();

		_mimeTypes = "text/plain";
		_nRows = -1;
		_nCols = -1;
		_attachmentDirectory = null;
		_attachmentHandler = null;
		_directory = null;
		_filter = null;
	}

	/**
	 * Init this instance attributes from its template.
	 *
	 * @param parts Indicates which parts of the template are copied.
	 */
	void initFromTemplate(int parts)
	{
        parts &= ~_initParts;
        if (parts == 0)
            return;
		if (_template == null)
			return;

		super.initFromTemplate(parts);

		LyFileFieldInfo template = (LyFileFieldInfo)_template;

		if ((parts & PART_DATA) > 0)
		{
			_mimeTypes = template._mimeTypes;
			_directory = template._directory;
			_filter = template._filter;
		}
		if ((parts & PART_GUI) > 0)
		{
			_nRows = template._nRows;
			_nCols = template._nCols;
			_attachmentDirectory = template._attachmentDirectory;
			_attachmentHandler = template._attachmentHandler;
		}
	}

	/**
	 * Constructor of a file field.
	 *
	 * @param id               Identifier of the file field.
	 * @param javaName         Name of the generated java field.
	 * @param template         Template used to construct this field.
	 * @param mimeTypes        Supported mime-types on this file field.
	 * @param attachmentHadler to store/retrieve files on server side. See available mime types :
	 *                         text/plain, image/gif, image/jpeg, text/html and so on... Depends on
	 *                         display capacity.
	 */
	LyFileFieldInfo(String id, String javaName, LyFileFieldInfo template,
		String mimeTypes, String attachmentHandler, String attachmentDirectory,
		String directory, String filter)
	{
		this(id);
		construct(javaName, template, mimeTypes, attachmentHandler, attachmentDirectory, directory,
			filter);
	}

	/**
	 * Constructor of a file field.
	 *
	 * @param id Identifier of the file field.
	 */
	LyFileFieldInfo(String id)
	{
		super(id);
	}

	/**
	 * Construction of a file field.
	 *
	 * @param javaName         Name of the generated java field.
	 * @param template         Template used to construct this field.
	 * @param mimeTypes        Supported mime-types on this file field.
	 * @param attachmentHadler to store/retrieve files on server side. See available mime types :
	 *                         text/plain, image/gif, image/jpeg, text/html and so on... Depends on
	 *                         display capacity.
	 */
	protected void construct(String javaName, LyFileFieldInfo template,
		String mimeTypes, String attachmentHandler, String attachmentDirectory,
		String directory, String filter)
	{
		super.construct(javaName, template);

		setMimeTypes(mimeTypes);
		setAttachmentHandler(attachmentHandler);
		setAttachmentDirectory(attachmentDirectory);
		setDirectory(directory);
		setFilter(filter);
	}

	/**
     * Returns extra tags.
     */
    protected LyTag[] getExtraTags()
    {
        return new LyTag[] { _attachmentHandler };
    }
    
	/* Implementation of LyFieldInfo methods */

	/**
	 * Returns class of values for this field. The value class corresponding to a file field is :
	 * String.class
	 */
	public Class<String> getValueClass()
	{
		return String.class;
	}

	/**
	 * Builds the appropriate LyValue corresponding to this field for the given value object.
	 *
	 * @see LyFileValue
	 */
	public LyValue buildFieldValue(Object value)
	{
		if (value == null)
			return new LyFileValue(this);
		else if (value instanceof String)
		{
			if (((String)value).length() == 0)
				return new LyFileValue(this);
			else
				return new LyFileValue(this, (String)value);
		}
		else
			return super.buildFieldValue(value);
	}

	/**
	 * Builds internal value for an instance of this fieldInfo from a string (previously constructed
	 * by toString).
	 *
	 * @see LyFileFieldInfo#toString
	 */
	public Object parse(String stringValue)
		throws LyFieldInfoFormatException
	{
		if ((stringValue == null) || (stringValue.length() == 0))
			return null;

		// Nothing to do : keep parameter as is.
		return stringValue;
	}

	/**
	 * Checks the given value : - the value must end with gif or jpg or jpeg if image type.
	 */
	public String checkValue(LyEnvironment environment, Object fieldValue)
	{
		if (fieldValue == null)
			return super.checkValue(environment, fieldValue);

		if (environment == null)
			environment = getEnvironment();

		// Don't accept files that are not images
		if ((_mimeTypes != null) && (_mimeTypes.startsWith("image")))
		{
			String value = fieldValue.toString();

			if (value.length() == 0)
				return super.checkValue(environment, fieldValue);

			String lowerValue = value.toLowerCase();

			if (!lowerValue.endsWith("gif")
					&& !lowerValue.endsWith("jpg")
					&& !lowerValue.endsWith("jpeg")
					&& !lowerValue.endsWith("png"))
				return environment.getMessage("LY_ER_IMAGE_FILE", 
						new Object[] { value, environment.translate(this.getName()) });
		}

		if (_filter != null)
		{
			String   value = fieldValue.toString();
			String[] filters = LyString.stringToList(_filter, " ");
			int		 pos, i, n = filters.length;
			boolean  match = false;

			for (i = 0; (i < n) && !match; i++)
			{
				pos = filters[i].indexOf(".");
				if ((pos >= 0) && value.endsWith(filters[i].substring(pos)))
					match = true;
			}

			if (!match)
				return environment.getMessage("LY_ER_FILE_FILTER",
					new Object[] { value, environment.translate(this.getName()), _filter });
		}

		return super.checkValue(environment, fieldValue);
	} // end method checkValue

	/**
	 * Return the specified value for this field as a "displayable" string.
	 */
	public String toName(Object fieldValue)
	{
		if (fieldValue == null)
			return null;

		if (fieldValue.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ fieldValue.getClass() + " expected " + getValueClass());

		String name = getFormattedName(fieldValue);

		if (name != null)
			return name;

        name = (String)fieldValue;
        if (name != null)
        {
            LyEnvironment     environment = getEnvironment();
            
            if ((environment != null)
                && (environment.getViewManager() != null))
            {
                String tempDir = environment.getViewManager().getTemporaryDir();
                
                if ((tempDir != null)
                        && (tempDir.length() > 0)
                        && (name.length() > tempDir.length())
                        && name.startsWith(tempDir))
                    return name.substring(tempDir.length() + 1);
            }
        }
        
        return name;
	}

	/**
	 * Converts the specified value for this field to an "internal" string (string that enables
	 * reconstruction through parse).
	 *
	 * @see LyFileFieldInfo#parse
	 */
	public String toString(Object fieldValue)
	{
		if (fieldValue == null)
			return null;

		if (fieldValue.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ fieldValue.getClass() + " expected " + getValueClass());

		String string = getFormattedString(fieldValue);

		if (string != null)
			return string;

		return (String)fieldValue;
	}

	/**
	 * Compare two objects for the specified fieldInfo.
	 */
	public int compare(Object value1, Object value2, BitSet modifiers)
	{
		if (value1 == value2)
			return 0;
		if (value1 == null)
			return -1;
		if (value2 == null)
			return 1;

		/*
		if (value1.getClass() != getValueClass())
		        throw new IllegalArgumentException("BF AppEngine : invalid value class "
		                + value1.getClass() + " expected " + getValueClass());
		if (value2.getClass() != getValueClass())
		        throw new IllegalArgumentException("BF AppEngine : invalid value class "
		                + value2.getClass() + " expected " + getValueClass());
		*/

		return ((String)value1).compareTo((String)value2);
	}

	/**
	 * Returns the filter conditions authorized on this field. The values must match the short
	 * constants defined in LyFilterExpression.
	 *
	 * @return An array of conditions corresponding to the authorized conditions on this field.
	 * @see    leon.info.LyFilterExpression
	 */
	public LyFilterExpression.Condition[] getFilterConditions()
	{
        LyFilterExpression.Condition[] conditions = new LyFilterExpression.Condition[2];

		conditions[0] = LyFilterExpression.Condition.EQUALS;
		conditions[1] = LyFilterExpression.Condition.CONTAINS;

		return conditions;
	}

	/**
	 * Checks if the second value is contained in the first one, for this field.
	 */
	public boolean contains(Object value1, Object value2, BitSet modifiers)
	{
		if ((value1 == null) && (value2 == null))
			return true;
		else if (value1 == null)
			return false;
		else if (value2 == null)
			return true;

		if (value1.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value1.getClass() + " expected " + getValueClass());
		if (value2.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value2.getClass() + " expected " + getValueClass());

		String str1 = (String)value1;
		String str2 = (String)value2;

		return (str1.indexOf(str2) != -1);
	}

	/**
	 * Constructs a value for this fieldInfo from an object for a given instance. Supported
	 * conversions are the following :
	 *
	 * <TABLE>
	 *   <TR>
	 *     <TD>String, String[], Number</TD>
	 *     <TD>String</TD>
	 *   </TR>
	 * </TABLE>
	 *
	 * @exception LyUnsupportedFieldBindingException If the requested conversion if none of the
	 *                                               supported conversions.
	 */
	public Object getValueFromObject(Object object, LyObject instance)
	{
        Object value = decodePeerValue(object);
        
        if (value == null)
            return null;
        
		if (value.getClass() == String.class)
			return value;
		
		if (value instanceof String[])
			return LyString.listToString((String[])value, '\n');
		
		if (value instanceof Number)
			return ((Number)value).toString();
		
		if (value instanceof LyPeerFileValue)
		{
			LyPeerFileValue fileValue = (LyPeerFileValue)value;

			return fileValue.getName();
		}

		return super.getValueFromObject(value, instance);
	}

	/**
	 * Returns an object of the given class from a value of this fieldInfo. Supported conversions
	 * are the following :
	 *
	 * <TABLE>
	 *   <TR>
	 *     <TD>String</TD>
	 *     <TD>String, String[], Number</TD>
	 *   </TR>
	 * </TABLE>
	 *
	 * @exception LyUnsupportedFieldBindingException If the requested conversion if none of the
	 *                                               supported conversions.
	 */
	public Object getPeerValue(Object value, Class<?> objectClass, LyObject instance)
	{
	    value = encodePeerValue(value, objectClass);
	    
		if (value == null)
			return null;

		if (value.getClass() == objectClass)
			return value;

		if (objectClass == LyPeerFileValue.class)
		{
			LyPeerFileValue fileValue = new LyPeerFileValue();
			LyFileValue     fv = instance.getFileValue(this);

			fileValue.setName((String)value);
			fileValue.setLocator(fv.getLocator());
			return fileValue;
		}

		if (objectClass == String[].class)
			return LyString.stringToList((String)value, '\n');
		try
		{
			if (objectClass == Integer.class)
				return Integer.valueOf((String)value);
			else if (objectClass == Short.class)
				return Short.valueOf((String)value);
			else if (objectClass == Long.class)
				return Long.valueOf((String)value);
			else if (objectClass == Double.class)
				return Double.valueOf((String)value);
			else if (objectClass == Float.class)
				return Float.valueOf((String)value);
			else if (objectClass == Byte.class)
				return Byte.valueOf((String)value);
		}
		catch (NumberFormatException e)
		{
			throw new LyUnsupportedFieldBindingException("BF AppEngine : exception caught on conversion "
				+ e.getMessage() + " on file field : " + this);
		}

        return super.getPeerValue(value, objectClass, instance);
	} // end method getPeerValue

	/**
	 * Gets the expected number of rows shown to the end-user.
	 *
	 * @return Number of rows specified for this file field when displaying content of the file.
	 */
	public int getNRows()
	{
		return _nRows;
	}

	/**
	 * Gets the expected number of columns shown to the end-user.
	 *
	 * @return Number of columns specified for this file field when displaying content of the file.
	 */
	public int getNCols()
	{
		return _nCols;
	}

	/**
	 * Get types of the file field. The type is found from the mime types value.
	 *
	 * @return Type of this fieldInfo, or TYPE_TEXT if type is not set. One of the constants
	 *         starting with TYPE_.
	 */
	public Type getType()
	{
		if (_mimeTypes == null)
			return Type.TEXT;
		else if (_mimeTypes.startsWith("image"))
			return Type.IMAGE;
		else if (_mimeTypes.equals("dir"))
			return Type.DIR;
		else if (_mimeTypes.startsWith("http://")
		        || _mimeTypes.startsWith("https://")
				|| _mimeTypes.startsWith("ftp://")
				|| _mimeTypes.startsWith("file://"))
			return Type.URL;
		else
			return Type.TEXT;
	}

	/**
	 * Get current mime-types supported by this file field.
	 *
	 * @return List of mime-types given in a string separated by blank.
	 */
	public String getMimeTypes()
	{
		return _mimeTypes;
	}

	/**
	 * Get Attachment handler used to encode/decode files submitted. one default implementation is
	 * leon.view.web.LyFileDecoder
	 *
	 * @return the attachment handler, or null to disable server upload (default).
	 */
	public String getAttachmentHandler()
	{
		return (_attachmentHandler == null ? null : _attachmentHandler.getJavaName());
	}

	/**
	 * Get Attachment directory used to create the path of files submitted.
	 *
	 * @return the path of the attachment directory.
	 */
	public String getAttachmentDirectory()
	{
		return _attachmentDirectory;
	}

	/**
	 * Get directory where value of this field is searched.
	 *
	 * @return the path of the directory (default is null)
	 */
	public String getDirectory()
	{
		return _directory;
	}

	/**
	 * Get filter applied to this file values.
	 *
	 * @return the filter (default is null : all files accepted)
	 */
	public String getFilter()
	{
		return _filter;
	}

	/**
	 * Sets supported mime types for this file field.
	 *
	 * @param mimeTypes List of mime-types supported by this file field. MIME-TYPES ARE NOT FULLY
	 *                  SUPPORTED.
	 */
	void setMimeTypes(String mimeTypes)
	{
		_mimeTypes = mimeTypes;
	}

	/**
	 * Sets default value for this field.
	 *
	 * @param defaultString Default path for this file field.
	 */
	void setDefault(String defaultString)
	{
		if (defaultString != null)
			defaultString = defaultString.intern();
		setDefaultValue(defaultString);
	}

	/**
	 * Sets the expected number of rows shown to the user.
	 *
	 * @param nRows Number of rows shown to the user.
	 */
	void setNRows(int nRows)
	{
		_nRows = nRows;
	}

	/**
	 * Sets the expected number of columns shown to the user.
	 *
	 * @param nCols Number of cols shown to the user.
	 */
	void setNCols(int nCols)
	{
		_nCols = nCols;
	}

	/**
	 * Sets Attachment handler used to encode/decode files submitted. one default implementation is
	 * leon.view.web.LyFileDecoder
	 *
	 * @param handler the attachment handler, or null to disable server upload (default).
	 */
	void setAttachmentHandler(String handler)
	{
		_attachmentHandler = (handler == null ? null : new LyJavaName(this, handler));
	}

	/**
	 * Sets Attachment directory used to located encoded/decoded files submitted.
	 *
	 * @param attachmentDirectory the attachment directory (default is files)
	 */
	void setAttachmentDirectory(String attachmentDirectory)
	{
		_attachmentDirectory = (attachmentDirectory == null ? null : attachmentDirectory.intern());
	}

	/**
	 * Utility method to get an URL from a value.
	 *
	 * @param  label the current label of the field
	 * @return the url to use in order to display this field.
	 */
	public String getUrl(LyEnvironment environment, String label)
	{
		// Is it a local file or an external location
		if ((label == null) || (label.trim().length() == 0))
			return null;

		// Is it already a URL ?
		if (label.startsWith("http://")
		        || label.startsWith("https://")
		        || label.startsWith("file://")
		        || label.startsWith("ftp://"))
			return label;

		String directory = getDirectory();
		String label2 = label;

		if ((directory != null)
				&& !label.toUpperCase().startsWith(directory.toUpperCase()))
		{
			File file = new File(label);

			if (!file.exists())
			{
				label2 = directory + label;
				label2 = environment.getString().expand(label2);
				file = new File(label2);
				if (!file.exists())
				{
					label2 = directory + "/" + label;
					label2 = environment.getString().expand(label2);
					file = new File(label2);
				}

				if (!file.exists())
				{
					label2 = directory + label;
					if (label2.indexOf("$LY_APP_DOC_DIR$") >= 0)
					{
						String appId = getApplication().getId();

						appId = appId.replace('.', '/');
						label2 = LyString.replace(label2, "$LY_APP_DOC_DIR$", appId);
						label2 = label2.replace('\\', '/');

						InputStream iStream = environment.getResourceAsStream(label2, null);

						if (iStream == null)
							return environment.getString().expand(label); // not found
						else
							return label2;
					}
					else
						return environment.getString().expand(label);
				}
			}
		}

		label2 = label2.replace('\\', '/');

		String docDir = environment.getEnv("LY_APP_DOC_DIR");
		String appDoc = environment.getEnv("LY_APP_DOC");

		if ((docDir != null) && (appDoc != null))
		{
			// USE URL starting with APP_DOC and not the file location APP_DOC_DIR
			// Avoid File.separator issues
			docDir = docDir.replace('\\', '/');
			appDoc = appDoc.replace('\\', '/');

			if (label2.startsWith(docDir))
				return (appDoc + label2.substring(docDir.length()));
			else if (label2.startsWith("http://")
			        || label2.startsWith("https://")
					|| label2.startsWith("file://")
					|| label2.startsWith("ftp://"))
				return (label2);
			else
				return ("file://" + label2);
		}
		else
			return ("file://" + label2);
	} // end method getUrl
    
    /**
     * Sets directory where value of this field is searched.
     *
     * @param directory The path of the directory (default is null)
     */
    protected void setDirectory(String directory)
    {
        if ((directory != null)
                && (directory.equals("absolute") || directory.equals("_absolute")))
            _directory = LyFileFieldInfo.VALUE_ABSOLUTE;
        else
            _directory = directory;
    }

    /**
     * Sets filter applied to this file values.
     *
     * @param filter The filter (default is null : all files accepted)
     */
    protected void setFilter(String filter)
    {
        _filter = filter;
    }
} // end class LyFileFieldInfo
