/*
 * LyFilter.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.io.*;
import java.util.*;

import leon.app.*;
import leon.data.*;
import leon.data.daemon.*;
import leon.info.misc.*;
import leon.misc.*;


/**
 * The LyFilter class implements the description of a filter used to filter objects list.<br>
 * A filter has a type which defines the type of filtering performed. Supported filter types are :
 * Type.AND (and filter), Type.OR (or filter).<br>
 * Filters contain filter elements which is an interface implemented buy both LyFilerExpression and
 * by LyFilter. This means that a filter may contain simple filter expressions and sub-filters. This
 * allows to build complex filters.
 *
 * <p>See also tag <FILTER> of leon.dtd</p>
 *
 * @see LyFilterExpression
 * @see LyFilterRelation
 */
public final class LyFilter extends LyObjectInfo
	implements LyFilterElement
{
	public static final String __VERSION =
		"$Id: LyFilter.java 41445 2012-11-07 10:15:37Z jlsaouli $";

	/** Filter types. */
	public enum Type
	{
		AND, OR;
	}

	/** Filter type : Type.AND , Type.OR. */
	Type			   _type;

	/**
	 * List of filter elements at current level. Elements are filter expressions and other filters.
	 */
	List<LyFilterElement> _filterElements;

	/** Indicates if the filter is a negation of the expression. */
	boolean			   _not;

	/** Indicates if the filter is dynamic (has an element that depends on a dynamic value
	 *  given by the session).
	 */
	Boolean            _dynamic;


	/**
	 * Initialisation of filter attributes.
	 */
	protected void init()
	{
		super.init();

		_type = Type.AND;
		_filterElements = new ArrayList<LyFilterElement>();
		_not = false;
		_dynamic = null;
	}

	/**
	 * Init filter attributes from its template.
	 */
	void initFromTemplate()
	{
		if ((PART_ALL & _initParts) > 0)
			return;
		if (_template == null)
			return;

		super.initFromTemplate();

		LyFilter template = (LyFilter)_template;

		if (template._filterElements != null)
			_filterElements = new ArrayList<LyFilterElement>(template._filterElements);
	}

	/**
	 * Constructor of an identified filter.
	 *
	 * @param id       Id of the filter.
	 * @param template Template model used to construct this filter.
	 * @param type     Type of the filter. Supported filter types are : Type.AND (and filter),
	 *                 Type.OR (or filter).
	 */
	public LyFilter(String id, LyFilter template, Type type)
	{
		this(id, template, type, false);
	}

	/**
	 * Constructor of an identified filter.
	 *
	 * @param id       Id of the filter.
	 * @param template Template model used to construct this filter.
	 * @param type     Type of the filter.
	 * @param not      Indicates if the filter is a negation of the expression. Supported filter
	 *                 types are : Type.AND (and filter), Type.OR (or filter).
	 */
	public LyFilter(String id, LyFilter template, Type type, boolean not)
	{
		this(id);
		construct(template, type, not);
		initFromTemplate();
	}

	/**
	 * Construction of a filter.
	 *
	 * @param template Template model used to construct this filter.
	 * @param type     Type of the filter.
	 */
	protected void construct(LyFilter template, Type type)
	{
		construct(template, type, false);
	}

	/**
	 * Construction of a filter.
	 *
	 * @param template Template model used to construct this filter.
	 * @param type     Type of the filter.
	 * @param not      Indicates if the filter is a negation of the expression.
	 */
	protected void construct(LyFilter template, Type type, boolean not)
	{
		super.construct(template);
		_type = (type == null ? Type.AND : type);
		_not = not;
	}

	/**
	 * Constructor of a filter with a given filter type.
	 *
	 * @param type Type of the filter. Supported filter types are : Type.AND (and filter), Type.OR
	 *             (or filter).
	 */
	public LyFilter(Type type)
	{
		this(type, false);
	}

	/**
	 * Constructor of a filter with a given filter type.
	 *
	 * @param type Type of the filter.
	 * @param not  Indicates if the filter is a negation of the expression. Supported filter types
	 *             are : Type.AND (and filter), Type.OR (or filter).
	 */
	public LyFilter(Type type, boolean not)
	{
		super();
		_type = type;
		_not = not;
	}

	/**
	 * Constructor of a default filter. Default type is Type.AND (and filter).
	 */
	public LyFilter()
	{
		this(Type.AND, false);
	}

	/**
	 * Constructor of an extern filter.
	 *
	 * @param id Identifier of the extern filter.
	 */
	LyFilter(String id)
	{
		this(id, false);
	}

	/**
	 * Constructor of an extern filter.
	 *
	 * @param id  Identifier of the extern filter.
	 * @param not Indicates if the filter is a negation of the expression.
	 */
	LyFilter(String id, boolean not)
	{
		super(id, false);
		_not = not;
	}

    /**
     * Sets id of this info.
     */
    public void setId(String id)
    {
		if (_id.startsWith("__")
		        && _id.endsWith("__")) // Random id
			super.setId(id);
    }

	/**
	 * Final step of the parsing : check consistency. For a filter, checks are the following : - the
	 * condition for filter expressions must be authorized on the fieldInfo of the expression.
	 */
	void check(LyObjectInfo parent)
	{
		int i, nElements = _filterElements.size();

		// Parse the filtering elements to retrieve the filter expressions
		for (i = 0; i < nElements; i++)
		{
			LyFilterElement filterElement = _filterElements.get(i);

			if (filterElement instanceof LyFilterExpression)
			{
				// For filter expressions, check if the condition is authorized
				// on the type of filtered field info.
				LyFilterExpression			   exp = (LyFilterExpression)filterElement;
				LyFilterExpression.Condition[] conditions = exp.getFieldInfo()
						.getFilterConditions();
				LyFilterExpression.Condition   condition = exp.getCondition();
				int							   j, n = conditions.length;
				boolean						   match = false;

				for (j = 0; j < n; j++)
				{
					if (condition == conditions[j])
					{
						match = true;
						break;
					}
				}

				if (!match)
				{
					String conditionExp = LyFilterExpression.getCondition(condition);

					throw new IllegalStateException("BF AppEngine : condition "
						+ conditionExp.substring(8, conditionExp.length())
						+ " not authorized on field info " + exp.getFieldInfo());
				}
			}
		}
	} // end method check

	/**
	 * Clone this filter.
	 */
	public LyFilter cloneFilter()
	{
		return cloneFilter(null);
	}

	/**
	 * Clone this filter.
	 */
	public LyFilter cloneFilter(LyDataSession session)
	{
		LyFilter filter = new LyFilter(null, null, getType(), _not);

		filter._name = getName();

		synchronized (_filterElements)
		{
			LyFilterElement element;
			int			    i, n = _filterElements.size();

			// Go through the element list
			for (i = 0; i < n; i++)
			{
				element = _filterElements.get(i);

				if (element instanceof LyFilter)
				{
					element = ((LyFilter)element).cloneFilter(session);
					((LyFilter)element)._parent = filter;
				}
				else if (element instanceof LyFilterExpression)
				{
					element = ((LyFilterExpression)element).cloneFilterExpression(session);
					((LyFilterExpression)element)._parent = filter;
				}

				filter.addElement(element);
			}
		}

		return filter;
	}

	/**
	 * Clones and inverts this filter.
	 *
	 * @return the cloned and inverted filter
	 */
	public LyFilter getInvertedFilter()
	{
		LyFilter filter = new LyFilter(getId() + LyString.getRandomId(), null,
				((getType() == Type.AND) ? Type.OR : Type.AND));

		filter._name = getName();

		synchronized (_filterElements)
		{
			LyFilterElement element;
			int			    i, n = _filterElements.size();

			// Go through the element list and invert them
			for (i = 0; i < n; i++)
			{
				element = _filterElements.get(i);

				if (element instanceof LyFilter)
					// For a filter, just invert it
					element = ((LyFilter)element).getInvertedFilter();
				else if (element instanceof LyFilterExpression)
				{
					// For an expression, clone it and invert it
					element = ((LyFilterExpression)element).cloneFilterExpression();
					((LyFilterExpression)element).setInverted(true);
				}
				else if (element instanceof LyFilterRelation)
				{
					// For a filter relation, clone it and invert its filter
					element = ((LyFilterRelation)element).cloneFilterRelation();

					LyFilter relationFilter = ((LyFilterRelation)element).getFilter();

					if (relationFilter != null)
						((LyFilterRelation)element).setFilter(relationFilter.getInvertedFilter());
				}
				else if (element instanceof LyFormula)
				{
					// For an expression, clone it and invert it
					element = ((LyFormula)element).cloneFormula();
					((LyFormula)element).setInverted(true);
				}

				filter.addElement(element);
			}
		}

		return filter;
	} // end method getInvertedFilter

	/**
	 * Check if the specified object matches current filter.
	 *
	 * @param  object Object to check.
	 * @return True if the object matches filter criteria, false otherwise.
	 * @see    LyFilterElement#match
	 */
	public boolean match(Object object)
	{
		load();

		boolean   typeAnd = (getType() == Type.AND);
		boolean   result;
        Object[]  elements = _filterElements.toArray();
        int       i, n = (elements == null ? 0 : elements.length);

		// Go through the element list
		for (i = 0; i < n; i++)
		{
			result = ((LyFilterElement)elements[i]).match(object);

			if (typeAnd && !result)
				// AND must match all conditions
				return _not;
			else if (!typeAnd && result)
				// OR must match at least one condition
				return !_not;
		}

		// FFT00312 : if filter has no expression, every object matches.
		if (n == 0)
			return !_not;

		// Finally :
		if (typeAnd)
			// All matches were true
			return !_not;
		else
			// All matches were false
			return _not;
	} // end method match

	/**
	 * Check if the specified values matches current filter.
	 *
	 * @param  values Values to check.
	 * @return True if the object matches filter criteria, false otherwise.
	 * @see    LyFilterElement#match
	 */
	public boolean match(LyValueSet values)
	{
		load();

		boolean   typeAnd = (getType() == Type.AND);
		boolean   result;
        Object[]  elements = _filterElements.toArray();
        int       i, n = (elements == null ? 0 :elements.length);

		// Go through the element list
		for (i = 0; i < n; i++)
		{
			result = ((LyFilterElement)elements[i]).match(values);

			if (typeAnd && !result)
				// AND must match all conditions
				return _not;
			else if (!typeAnd && result)
				// OR must match at least one condition
				return !_not;
		}
		// Finally :
		if (typeAnd)
			// All matches were true
			return !_not;
		else
			// All matches were false
			return _not;
	} // end method match

	/**
	 * Indicates if current filter may be applied to the specified class info. Verifies that every
	 * field info contained in current filter is contained in the specified class info.
	 *
	 * @param  classInfo Class info to check.
	 * @return True if the filter may be applied to specified class info, false otherwise.
	 */
	public boolean matchClass(LyClassInfo classInfo)
	{
		load();

		LyInfoList<LyFieldInfo> fieldInfos = getFieldInfos(false);
		LyFieldInfo			    field;
		int					    i, n = fieldInfos.getSize();

		for (i = 0; i < n; i++)
		{
			field = fieldInfos.getInfo(i);
			if (classInfo.getFieldInfo(field.getId(), false) == null)
				return _not;
		}

		return !_not;
	}

	/**
	 * Dumps current filter as a String.
	 *
	 * @return String representation of this filter.
	 */
	public String dump()
	{
		StringBuffer sb = new StringBuffer(500);

		synchronized (_filterElements)
		{
			LyFilterElement element;
			int			    n = _filterElements.size();
			boolean		    typeAnd = (getType() == Type.AND);

			sb.append('(');
			if (_not)
				sb.append("!(");
			// Go through the element list
			for (int i = 0; i < n; i++)
			{
				element = _filterElements.get(i);
				// Add operator
				if (i != 0)
				{
					if (typeAnd)
						sb.append(" AND ");
					else
						sb.append(" OR ");
				}
				sb.append(element.toString());
			}
			sb.append(')');
			if (_not)
				sb.append(')');
		}

		return sb.toString();
	} // end method toString

	/**
	 * Writes XML corresponding to this filter element.
	 *
	 * @param pw PrintWriter on which the XML is written.
	 */
	public void writeXml(PrintWriter pw)
	{
		// Write tag with its attributes
		pw.print("<FILTER");

		if (getId() != null)
			pw.print(" id='" + getId() + "'");
		else
			pw.print(" id='_filter" + LyString.getXmlString(LyString.getRandomId()) + "'");

		if (getType() == Type.OR)
			pw.print(" type='OR'");

		pw.println(">");

		if ((getName() != null) && !(getName().startsWith("_")))
		{
			pw.print("<NAME value='");
			pw.print(getName());
			pw.println("'/>");
		}

		synchronized (_filterElements)
		{
			LyFilterElement element;
			int			    i, n = _filterElements.size();

			// Go through the element list
			for (i = 0; i < n; i++)
			{
				element = _filterElements.get(i);
				element.writeXml(pw);
			}
		}

		pw.println("</FILTER>");
	} // end method writeXml

	/**
	 * Gets list of fieldInfos that this filter refers to.
	 *
	 * @return List of fieldInfos that the filter depends on.
	 */
	public LyInfoList<LyFieldInfo> getFieldInfos()
	{
		return getFieldInfos(true);
	}

	/**
	 * Gets list of fieldInfos that this filter refers to.
	 *
	 * @param  recurseFilterRelation indicates if search must be performed recursively on the
	 *                               relation filters elements. If false, only relation field info
	 *                               of the relation filters are added to the result list.
	 * @return List of fieldInfos that the filter depends on.
	 */
	public LyInfoList<LyFieldInfo> getFieldInfos(boolean recurseFilterRelation)
	{
		load();

		synchronized (_filterElements)
		{
			LyInfoList<LyFieldInfo> fieldInfos = new LyInfoList<LyFieldInfo>();
			LyFilterElement		    element;
			int					    i, n = _filterElements.size();

			// Go through the element list
			for (i = 0; i < n; i++)
			{
				element = _filterElements.get(i);

				if (element instanceof LyFilter)
					fieldInfos.addList(((LyFilter)element).getFieldInfos(recurseFilterRelation));
				else if (element instanceof LyFilterExpression)
					fieldInfos.addInfo(((LyFilterExpression)element).getFieldInfo());
				else if (element instanceof LyFilterRelation)
				{
					if (recurseFilterRelation)
						fieldInfos.addList(((LyFilterRelation)element).getFieldInfos());
					else
						fieldInfos.addInfo(((LyFilterRelation)element).getRelationFieldInfo());
				}
				else if (element instanceof LyFormula)
				    fieldInfos.addList(((LyFormula)element).getFormulaFields());
			}

			return fieldInfos;
		}
	} // end method getFieldInfos

	/**
	 * Gets all filter values corresponding to the given field info.
	 *
	 * @param fieldInfo A field info.
	 * @param result    result list.
	 */
	public void getValues(LyFieldInfo fieldInfo, Vector<Object> result)
	{
		load();

		LyFilterElement element;
		int			    i, n = _filterElements.size();

		for (i = 0; i < n; i++)
		{
			element = _filterElements.get(i);
			if ((element instanceof LyFilterExpression)
					&& (((LyFilterExpression)element).getFieldInfo() == fieldInfo))
				result.addElement(((LyFilterExpression)element).getValue());
			else if (element instanceof LyFilter)
				((LyFilter)element).getValues(fieldInfo, result);
		}
	}

	/**
	 * Adds to given list objects of the given list that match this filter. This method is provided
	 * for optimization purpose to avoid multiple accesses of filter elements.
	 *
	 * @param  objects List to filter.
	 * @param  list    Result list (may be null).
	 * @return An objects list containing all objects of the list given as parameter that match this
	 *         filter.
	 */
	public LyObjectList match(Enumeration<?> objects, LySimpleObjectList list)
	{
		if (objects == null)
			return null;

		if ((getSize() == 0) && !_not)
			return list;

		if (list == null)
			list = new LySimpleObjectList();

		LyObject  object;
		boolean   typeAnd = (getType() == Type.AND);
		boolean   result;
		Object[]  elements = _filterElements.toArray();
		int       i, n = (elements == null ? 0 :elements.length);

		while (objects.hasMoreElements())
		{
			object = (LyObject)objects.nextElement();

			for (i = 0; i < n; i++)
			{
				result = ((LyFilterElement)elements[i]).match(object);

				if (typeAnd && !result)
				{
					// AND must match all conditions
					if (_not)
						list.addObject(object);
					break;
				}
				else if (!typeAnd && result)
				{
					// OR must match at least one condition
					if (!_not)
						list.addObject(object);
					break;
				}
			}

			if ((i == n)
					&& (((n == 0) && !_not)
						|| (typeAnd && !_not)
						|| (!typeAnd && _not)))
				list.addObject(object);
		}

		return list;
	} // end method match

	/**
	 * Counts objects of the given list that match this filter. This method is provided
	 * for optimization purpose to avoid multiple accesses of filter elements.
	 *
	 * @param  objects List of objects to apply on filter.
	 * @return Number of objects matching this filter.
	 */
	public int match(Enumeration<Object> objects)
	{
		if (objects == null)
			return 0;

		if ((getSize() == 0) && !_not)
			return 0;

		LyObject  object;
		boolean   typeAnd = (getType() == Type.AND);
		boolean   result;
		Object[]  elements = _filterElements.toArray();
		int       i, n = (elements == null ? 0 :elements.length);
		int       nm = 0;

		while (objects.hasMoreElements())
		{
			object = (LyObject)objects.nextElement();

			for (i = 0; i < n; i++)
			{
				result = ((LyFilterElement)elements[i]).match(object);

				if (typeAnd && !result)
				{
					// AND must match all conditions
					if (_not)
						nm++;
					break;
				}
				else if (!typeAnd && result)
				{
					// OR must match at least one condition
					if (!_not)
						nm++;
					break;
				}
			}

			if ((i == n)
					&& (((n == 0) && !_not)
							|| (typeAnd && !_not)
							|| (!typeAnd && _not)))
				nm++;
		}

		return nm;
	} // end method match

	/**
	 * Get a filter element from this filter at the given index.
	 *
	 * @param  idx Index of the searched element.
	 * @return Filter element at the searched position which is either a new sub-filter or
	 *         filterExpression or null if no such element exists.
	 */
	public LyFilterElement getElement(int idx)
	{
		load();

		synchronized (_filterElements)
		{
			if ((idx < 0) || (idx >= _filterElements.size()))
				return null;

			return _filterElements.get(idx);
		}
	}

	/**
	 * Returns the list of filter elements defined on this filter This is a COPY of the internal
	 * list, therefore this list may be freely manipulated by the application.
	 *
	 * @return List of filter elements, or an empty Vector. Never returns null.
	 */
	public Vector<LyFilterElement> getElements()
	{
		load();

		return new Vector<LyFilterElement>(_filterElements);
	}

	/**
	 * Return the number of filter elements of this filter.
	 *
	 * @return Number of elements in the filter.
	 */
	public int getSize()
	{
		load();

		return (_filterElements == null ? 0 : _filterElements.size());
	}

	/**
	 * Accessor to the type of the filter.
	 *
	 * @return Filter type (Type.AND or Type.OR).
	 */
	public Type getType()
	{
		return _type;
	}

	/**
	 * Indicates whether the filter is simple or extended. A filter is extended if it contains
	 * another filter or if it contains filter expressions that does not apply on the same field
	 * info. If there is only one field info reached by many expressions, it should be a relation or
	 * choice field info (otherwise filter is considered as an extended one).
	 *
	 * @return true if the filter is an extended one, false if it is simple.
	 */
	public boolean isExtended()
	{
		int n = _filterElements.size();

		if (n > 0)
		{
			int			    i;
			LyFilterElement element;
			LyFieldInfo     field = null, fieldTmp = null;

			for (i = 0; i < n; i++)
			{
				element = getElement(i);

				if ((element instanceof LyFilter)
						|| (element instanceof LyFilterRelation))
					return true;

				if (element instanceof LyFilterCriterion)
					return true;

				LyFilterExpression expression = (LyFilterExpression)element;

				fieldTmp = expression.getFieldInfo();

				if (field == null)
					field = fieldTmp;
				// If more than one field info, filter is an extended one
				else if (field != fieldTmp)
					return true;
				else if ((n > 1) && !(fieldTmp instanceof LyRelationFieldInfo)
						&& !(fieldTmp instanceof LyChoiceFieldInfo))
					return true;
			}
		}

		return false;
	} // end method isExtended

	/**
	 * Checks if an expression is found concerning the given field info.
	 *
	 * @param  fieldInfo The fieldInfo.
	 * @return True if an expression with the specified is found, false otherwise.
	 */
	public boolean contains(LyFieldInfo fieldInfo)
	{
		load();

		synchronized (_filterElements)
		{
			LyFilterElement element;
			int			    i, n = _filterElements.size();

			for (i = 0; i < n; i++)
			{
				element = _filterElements.get(i);
				if ((element instanceof LyFilterExpression)
						&& (((LyFilterExpression)element).getFieldInfo() == fieldInfo))
					return true;
				else if ((element instanceof LyFilter)
						&& ((LyFilter)element).contains(fieldInfo))
					return true;
				else if ((element instanceof LyFilterRelation)
						&& ((LyFilterRelation)element).contains(fieldInfo))
					return true;
				else if ((element instanceof LyFormula)
						&& ((LyFormula)element).contains(fieldInfo))
					return true;
			}
		}

		return false;
	}

	/**
	 * Checks if a filter element is found in this filter.
	 *
	 * @param  element The filter element.
	 * @return True if the filter element is found, false otherwise.
	 */
	public boolean contains(LyFilterElement element)
	{
		return ((_filterElements != null) && _filterElements.contains(element));
	}

	/**
	 * Indicates if the filter is a negation of the expression.
	 *
	 * @return true if the filter is a negation of the expression, false otherwise.
	 */
	public boolean isNot()
	{
		return _not;
	}

	/**
	 * Indicate if this filter has a least one local fields.
	 * Local fields with join or reverse damon are not taken into account.
	 * @return true if the filter only use local fields.
	 */
	public boolean hasLocalFields()
	{
		LyInfoList<LyFieldInfo>		fields = getFieldInfos(true);
		int							i, n = (fields == null ? 0 : fields.getSize());
		LyFieldInfo					fieldInfo;
		LyFieldInfoDaemon			daemon;
		
		for (i = 0; i < n ;	i++)
		{
			fieldInfo = fields.getInfo(i);

			if ((fieldInfo.getParent() instanceof LyClassInfo)
					&& ((LyClassInfo)fieldInfo.getParent()).isLocal())
				return true; // field is local

			if (!fieldInfo.isLocal())
				continue;

			if (getApplication() != null)
			{
				daemon = getApplication().getDaemon(fieldInfo);
				if ((daemon != null)
					&& ((daemon instanceof LyJoinRelationDaemon)
						|| (daemon instanceof LyReverseRelationDaemon)))
					continue;
			}

			return true;
		}

		return false;
	}

	/**
	 * Sets filter type for this filter.
	 *
	 * @param     type New type for filter.
	 * @exception IllegalArgumentException If the given type is one of the supported types (Type.AND
	 *                                     or Type.OR).
	 */
	public void setType(Type type)
	{
		if (isLocked())
			throw new IllegalStateException("BF AppEngine : filter " + this
				+ " was locked and thus can't be modified anymore");

		if ((type != Type.AND) && (type != Type.OR))
			throw new IllegalArgumentException("BF AppEngine : invalid filter type " + type);

		_type = type;
	}

	/**
	 * Adds a new element to this filter.
	 *
	 * @param element Element to add to this filter. This element can be either a sub-filter or a
	 *                filter expression.
	 */
	public void addElement(LyFilterElement element)
	{
		if (isLocked())
			throw new IllegalStateException("BF AppEngine : filter " + this
				+ " was locked and thus can't be modified anymore");

		synchronized (_filterElements)
		{
			if (element != null)
				_filterElements.add(element);
		}

		_dynamic = null;
	}

	/**
	 * Inserts a new element to this filter at the given position.
	 * @param element  Element to add to this filter. This element can be either a sub-filter or a
	 *                 filter expression.
	 * @param position Position where the element is inserted.
	 */
	public void addElement(LyFilterElement element, int position)
	{
		if (isLocked())
			throw new IllegalStateException("BF AppEngine : filter " + this
					+ " was locked and thus can't be modified anymore");

		synchronized (_filterElements)
		{
			if (element != null)
				_filterElements.add(position, element);
		}

		_dynamic = null;
	}

	/**
	 * Removes an element from this filter.
	 *
	 * @param element Element to remove from the filter.
	 */
	public void removeElement(LyFilterElement element)
	{
		if (isLocked())
			throw new IllegalStateException("BF AppEngine : filter " + this
				+ " was locked and thus can't be modified anymore");

		synchronized (_filterElements)
		{
			if (element != null)
				_filterElements.remove(element);
		}
	}

	/**
	 * Gets index for the given element.
	 *
	 * @return Index of the element in the filter.
	 */
	public int getIndex(LyFilterElement element)
	{
		return (_filterElements == null ? -1
				: _filterElements.indexOf(element));
	}

	/**
	 * Checks if this filter if impacted by the changes contained in the value set.
	 */
	public boolean isImpacted(LyValueSet values, boolean recurse)
	{
		if (values == null)
			return true;

		LyInfoList<LyFieldInfo> fieldInfos = getFieldInfos(recurse);
		int					    i, n = fieldInfos.getSize();
		LyFieldInfo			    fieldInfo;

		for (i = 0; i < n; i++)
		{
			// Find objects referenced by this relation on this object
			fieldInfo = fieldInfos.getInfo(i);
			if (values.getFieldValue(fieldInfo) != null)
				return true;
		}

		return false;
	}

	/**
	 * Checks if this filter is dynamic or not.
	 * The filter is dynamic when it has an element that depends on a dynamic value
	 * given by the session.
	 */
	public boolean isDynamic()
	{
		if (_dynamic != null)
			return _dynamic.booleanValue();

		boolean      dynamic = false;

		synchronized (_filterElements)
		{
			LyFilterElement element;
			int             i, n = _filterElements.size();

			// Go through the element list
			for (i = 0; (i < n) && !dynamic; i++)
			{
				element = _filterElements.get(i);

				if (element instanceof LyFilter)
					dynamic |= ((LyFilter)element).isDynamic();
				else if (element instanceof LyFilterExpression)
					dynamic |= ((LyFilterExpression)element).isDynamic();
			}
		}

		_dynamic = Boolean.valueOf(dynamic);
		return dynamic;
	}

	/**
	 * Returns the list of filter elements defined on this filter This is a COPY of the internal
	 * list, therefore this list may be freely manipulated by the application.
	 *
	 * @return List of filter elements, or an empty Vector. Never returns null.
	 */
	protected void setElements(Vector<LyFilterElement> elements)
	{
		_filterElements = new Vector<LyFilterElement>();

		int     i, n = (elements == null ? 0 : elements.size());

		for (i = 0; i < n; i++)
			addElement(elements.get(i));
	}

	/**
	 * Sets the formula associated to the field.
	 *
	 * @param     expression Current formula expression, or null if none is specified.
	 */
	void addFormula(String expression)
	{
		LyFormula   formula = LyFormula.formulaFactory(this, expression);

		addElement(formula);
	}

	/**
	 * Get an peer equivalent of the current filter
	 *
	 * @return a peer equivalent of the current filter
	 */
	public LyFilter getPeerFilter()
	{
		if ((_filterElements == null)
				|| (_filterElements.size() < 1))
			return this;

		LyFilter			tempFilter = new LyFilter(_type, _not);
		LyFilterElement[]	peerElement;
		LyFilterElement		filterElem;
		int					i, n = _filterElements.size();

		for (i = 0; i < n; i++)
		{
			filterElem = _filterElements.get(i);

			if (filterElem instanceof LyFilter)
				tempFilter.addElement(((LyFilter)filterElem).getPeerFilter());
			else if (filterElem instanceof LyFilterExpression)
			{
				LyFilterExpression filterExpr = (LyFilterExpression)filterElem;
				Object             value = filterExpr.getValue();

				if ((value != null)
				    && (value.equals(LyDateFieldInfo.VALUE_TODAY)
						|| value.toString().startsWith(LyDateFieldInfo.VALUE_DAY)))
				{
					peerElement = filterExpr.getPeerExpression();

					if (peerElement.length > 1)
					{
						LyFilter subFilter = new LyFilter();

						if (filterExpr.isInverted())
							subFilter.setType(Type.OR);
						else
							subFilter.setType(Type.AND);

						for (int j = 0; j < peerElement.length; j++)
							subFilter.addElement(peerElement[j]);

						tempFilter.addElement(subFilter);
					}
					else
						tempFilter.addElement(peerElement[0]);
				}
				else
					tempFilter.addElement(filterElem);
			}
			else
				tempFilter.addElement(filterElem);
		}

		return tempFilter;
	}
} // end class LyFilter
