/*
 * LyFilterCriterion.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.io.*;
import leon.data.*;


/**
 * This class represents a filter criterion. It contains a filter element defined by a java class.
 *
 * @see LyFilter
 * @see LyFilterElement
 */
public final class LyFilterCriterion extends LyInfo
	implements LyFilterElement
{
	public static final String __VERSION =
		"$Id: LyFilterCriterion.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** Interface used to match values. */
	String					   _javaName;

	/** Class that implements the LyFilterElement interface. */
	
	Class<?>				   _javaClass;

	/** Instance of the LyFilterElement interface. */
	LyFilterElement			   _element;

    /**
     * Constructor for a filter criterion.
     */
    protected LyFilterCriterion()
    {
    }
    
	/**
	 * Constructor for a filter criterion.
	 *
	 * @param id       Id of the filter criterion.
	 * @param javaName Name of the java class corresponding to this filter criterion.
	 */
	public LyFilterCriterion(String id, String javaName)
	{
		super(id);

		_javaName = javaName;
		_javaClass = null;
		_element = null;
	}

	/**
	 * Constructor for a filter criterion with a Class that implements LyFilterElement.
	 *
	 * @param javaClass Class that implements the LyFilterElement interface.
	 */
	public LyFilterCriterion(Class<?> javaClass)
	{
		if (javaClass == null)
			throw new IllegalArgumentException("BF AppEngine : javaClass can't be null.");
		if (!LyFilterElement.class.isAssignableFrom(javaClass))
			throw new IllegalArgumentException("BF AppEngine : invalid class "
				+ javaClass
				+ ", the given class must implement leon.info.LyFilterElement interface");

		_javaClass = javaClass;
		_javaName = javaClass.getName();
		_element = null;
	}

	/**
	 * Constructor for a filter criterion with a given filter element.
	 *
	 * @param element Filter element.
	 */
	public LyFilterCriterion(LyFilterElement element)
	{
		if (element == null)
			throw new IllegalArgumentException("BF AppEngine : element can't be null.");

		_element = element;
		_javaClass = element.getClass();
		_javaName = _javaClass.getName();
	}

	/**
	 * Indicates whether the specified object matches the filter criterion.
	 *
	 * @param  object Object for which the filter criterion is checked.
	 * @return True if the objects matches criterion, false otherwise.
	 */
	public boolean match(Object object)
	{
		LyFilterElement element = getElement();

		if (element != null)
			return element.match(object);

		return false;
	}

	/**
	 * Indicates whether the specified values matches the filter criterion.
	 *
	 * @param  values List of values for which the filter criterion is checked.
	 * @return True if the values matches criterion, false otherwise.
	 */
	public boolean match(LyValueSet values)
	{
		LyFilterElement element = getElement();

		if (element != null)
			return element.match(values);

		return false;
	}

	/**
	 * Writes XML corresponding to this criterion element.
	 *
	 * @param pw PrintWriter on which the XML is written.
	 */
	public void writeXml(PrintWriter pw)
	{
		// Write tag with its attributes
		pw.print("<FILTER_CRITERIA");

		pw.print(" javaName=\"" + _javaName + "\"");
		pw.println("/>");
	}

	/**
	 * Returns a displayable string of this criterion.
	 *
	 * @return A displayable string representation of this criterion.
	 */
	public String getName()
	{
		LyFilterElement element = getElement();

		if (element != null)
			return element.getName();

		return null;
	}

	/**
	 * Clone this filter criterion.
	 */
	public LyFilterCriterion cloneFilterCriterion()
	{
		LyFilterCriterion criterion = new LyFilterCriterion(null, getJavaName());

		criterion.setApplication(getApplication());
		return criterion;
	}

	/**
	 * Gets comparator instance.
	 */
	protected LyFilterElement getElement()
	{
		if (_element != null)
			return _element;

		try
		{
			if (_javaClass == null)
				// Create comparator
				_javaClass = _application.loadClass(_javaName,
						"BF AppEngine : unknown element class " + _javaName
						+ " for filter criteria " + getId());

			_element = (LyFilterElement)_javaClass.newInstance();
		}
		catch (Exception e)
		{
			if (_javaClass == null)
				throw new IllegalStateException("BF AppEngine : invalid class name "
					+ _javaName + " for filter criteria " + getId());
			else
				throw new IllegalStateException("BF AppEngine : invalid class "
					+ _javaClass + " for filter criteria " + getId());
		}

		return _element;
	} // end method getElement

	/**
	 * Gets name of the java class corresponding to this filter criterion.
	 *
	 * @return Name of the java class corresponding to this filter criterion.
	 */
	public String getJavaName()
	{
		return _javaName;
	}

	/**
	 * Gets java class corresponding to this filter criterion.
	 *
	 * @return The java class corresponding to this filter criterion.
	 */
	public Class<?> getJavaClass()
	{
		return _javaClass;
	}

	/**
	 * Sets java class name corresponding to this filter criterion.
	 *
	 * @param javaName Name of the java class corresponding to this filter criterion.
	 */
	public void setJavaName(String javaName)
	{
		_javaName = javaName;
	}
} // end class LyFilterCriterion
