/*
 * LyFilterExpression.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.io.*;
import java.util.*;

import leon.app.*;
import leon.data.*;
import leon.misc.*;


/**
 * The LyFilterExpression class implements the description of an elementary expression applied on
 * objects of the application. A filter expression is an expression of one elementary criterion.
 * Filters are like a 'tree' of sub-filters AND/OR, terminated by one or several filter expressions.
 * A complex condition (a filter) is a logical association of elementary conditions.<br>
 * The structure of an elementary expression is :
 *
 * <ul>
 *   <LI>a <b>field</b> (<i>LyFieldInfo</i>) : when checking if an object matches this expression
 *     the value is read for the specified fieldInfo.</LI>
 *   <LI>a <b>condition</b> : equals, lower than, greater than, contains ... Conditions are defined
 *     as public static named Condition.XXX</LI>
 *   <LI>a <b>value</b> which should be compared to the value of the application object for the
 *     specified field info.</LI>
 *   <LI>a boolean <b>not</b> which indicates if thre result of the expression must be reversed or
 *     not.
 *
 *     <p>See also tag <EXP> of leon.dtd</p>
 *   </LI>
 * </ul>
 *
 * @see LyFilter
 */
public final class LyFilterExpression extends LyInfo
	implements LyFilterElement
{
	public static final String __VERSION =
		"$Id: LyFilterExpression.java 36600 2012-08-28 07:00:42Z jlsaouli $";

	/** Condition. */
    public enum Condition
    {
        EQUALS,
        LOWER_THAN,
        LOWER_OR_EQUALS,
        GREATER_THAN,
        GREATER_OR_EQUALS,
        CONTAINS,
        STARTS_WITH,
        ENDS_WITH
    }

	/** Key names corresponding to supported conditions. */
	static String[]			   conditionNames =
	{
		"LY_COND_EQUALS",
		"LY_COND_LOWER_THAN",
		"LY_COND_LOWER_OR_EQUALS",
		"LY_COND_GREATER_THAN",
		"LY_COND_GREATER_OR_EQUALS",
		"LY_COND_CONTAINS",
		"LY_COND_STARTS_WITH",
		"LY_COND_ENDS_WITH",
	};

	/** Key names corresponding to supported conditions. */
	static String[]			   notConditionNames =
	{
		"LY_COND_NOT_EQUALS",
		"LY_COND_GREATER_OR_EQUALS",
		"LY_COND_GREATER_THAN",
		"LY_COND_LOWER_OR_EQUALS",
		"LY_COND_LOWER",
		"LY_COND_NOT_CONTAINS",
		"LY_COND_NOT_STARTS_WITH",
		"LY_COND_NOT_ENDS_WITH",
	};

	/** Modifier case sensitive. */
	public static final short  MOD_CASE_SENSITIVE = 0;

	/** Modifier no accents. */
	public static final short  MOD_RAW = 1;

	/** Modifier not by name for relations. */
	public static final short  MOD_NOT_BY_NAME = 2;

	/** Modifier for a filter which applies only in peer connector. Match value always return true in GUI. Internal use only */
	public static final short  MOD_PEER_FILTER_ONLY = 3;

	/** Modifier for multilingual filtering. */
	public static final short  MOD_MULTILINGUAL = 4;

	/** Key names corresponding to supported modifiers. */
	static String[]			   modifierNames =
	{
		"LY_MOD_CASE_SENSITIVE",
		null,
		null,
		null,
		"LY_MOD_MULTILINGUAL"
	};

	/** Recompute the _now and $xxx in filter expression every 'n' seconds. -1 means never, 0 always. */
	public static int		RECOMPUTE_DATE_DELAY = -1;

	/**
	 * Return the name corresponding to the specified condition.
	 *
	 * @param  condition Condition (one of the Condition.XXX)
	 * @return Untranslated name corresponding to the given condition if found, null otherwise.
	 */
	public static String getCondition(Condition condition)
	{
		return conditionNames[condition.ordinal()];
	}

	/**
	 * Return the name corresponding to the specified not condition.
	 *
	 * @param  condition Condition (one of the Condition.XXX)
	 * @return Untranslated name corresponding to the given condition if found, null otherwise.
	 */
	public static String getNotCondition(Condition condition)
	{
		return notConditionNames[condition.ordinal()];
	}

	/**
	 * Return the condition corresponding to the given name.
	 *
	 * @param  conditionName Untranslated name.
	 * @return Condition if found, null otherwise.
	 */
	public static Condition getCondition(String conditionName)
	{
        try
        {
            return Condition.valueOf(conditionName);
        }
        catch (IllegalArgumentException e)
        {
            return null;
        }
	}

	/**
	 * Return the name corresponding to the specified modifier.
	 *
	 * @param  modifier Modifier (one of the MOD_XXX)
	 * @return Untranslated name corresponding to the given modifier if found, null otherwise.
	 */
	public static String getModifier(short modifier)
	{
		if ((modifier < 0) || (modifier >= modifierNames.length))
			return null;

		return modifierNames[modifier];
	}

	/**
	 * Return the modifier corresponding to the given name.
	 *
	 * @param  modifierName Untranslated name.
	 * @return Modifier if found, null otherwise.
	 */
	public static short getModifier(String modifierName)
	{
		if (modifierName == null)
			return -1;

		// Go through the list of known symbols
		short i, n = (short)modifierNames.length;

		for (i = 0; i < n; i++)
		{
			if (modifierName.equals(modifierNames[i]))
				return i;
		}

		return -1;
	}

	/** FieldInfo on which this filter expression applies. */
	LyFieldInfo _fieldInfo;

	/** Boolean indicating if the filter expression is inverted or not. */
	boolean     _not;

	/**
	 * Condition controlled on the fieldInfo value to match the expression. See the Condition.XXX
	 * constants, default is Condition.EQUALS.
	 */
	Condition   _condition;

	/** Value associated to this filter expression used for comparing to fieldInfo values. */
	Object	    _value;

	/**
	 * BitSet containing the modifiers of the expression. See the MOD_XXX constants, default is null
	 * (no modifier).
	 */
	BitSet	    _modifiers;

	/**
	 * Initial value associated to this filter expression used in constructor when this expression
	 * was built (such as _now which is replaced automatically).
	 */
	Object	    _initValue;

	/**
	 * Last computation date.
	 */
	long	    _recomputeDate;

	/**
	 * Computation delay in milliseconds.
	 */
	long	    _recomputeDateDelay;

	/**
	 * Internal use to check if the filter is checked in memory or not
	 */
	boolean	 	_checkedPeerFilterOnly = true;
	/**
	 * Initialisation of the filter expression attributes.
	 */
	protected void init()
	{
		_fieldInfo = null;
		_modifiers = new BitSet();
		_condition = Condition.EQUALS;
		_initValue = null;
		_value = null;
		_recomputeDate = 0;
	}

	/**
     * Constructor of a filter expression.
     *
     * @param id        Id of the filter expression.
     * @param fieldInfo FieldInfo on which this filter expression applies.
     * @param not       Inverts filter condition or not.
     * @param condition Filter condition.
     * @param value     Value of this expression (may be null).
     * @param modifiers List of modifiers (may be null). This value is an instance of the class
     *                  returned by the getValueClass method of LyFieldInfo.
     * @see   LyFieldInfo#getValueClass
     */
    public LyFilterExpression(String id, LyFieldInfo fieldInfo, boolean not, Condition condition,
        Object value, BitSet modifiers)
    {
        super(id);

		init();

		_fieldInfo = fieldInfo;
		_not = not;
		setCondition(condition);
		if (modifiers != null)
			_modifiers = (BitSet)modifiers.clone();

		// Set application for dynamic filters
		if (fieldInfo != null)
		{
			_application = fieldInfo.getApplication();

			// Do not know if needed to recompute
			_recomputeDateDelay = RECOMPUTE_DATE_DELAY;
			if (_recomputeDateDelay < 0)
				_recomputeDateDelay = -1; // never try again
			else
				_recomputeDateDelay = _recomputeDateDelay * 1000; // in ms
		}

        setInitValue(value);
    }

	/**
	 * Constructor of a filter expression.
	 *
	 * @param fieldInfo FieldInfo on which this filter expression applies.
	 * @param not       Inverts filter condition or not.
	 * @param condition Filter condition.
	 * @param value     Value of this expression (may be null).
	 * @param modifiers List of modifiers (may be null). This value is an instance of the class
	 *                  returned by the getValueClass method of LyFieldInfo.
	 * @see   LyFieldInfo#getValueClass
	 */
	public LyFilterExpression(LyFieldInfo fieldInfo, boolean not, Condition condition,
		Object value, BitSet modifiers)
	{
	    this(null, fieldInfo, not, condition, value, modifiers);
	}

    /**
     * Default constructor of a filter expression.
     */
    protected LyFilterExpression()
    {
    }

	/**
	 * Constructor of a filter expression.
	 *
	 * @param fieldInfo FieldInfo on which this filter expression applies.
	 * @param not       Inverts filter condition or not.
	 * @param condition Filter condition.
	 * @param value     Value of this expression (may be null). This value is an instance of the
	 *                  class returned by the getValueClass method of LyFieldInfo.
	 * @see   LyFieldInfo#getValueClass
	 */
	public LyFilterExpression(LyFieldInfo fieldInfo, boolean not, Condition condition,
		Object value)
	{
		this(null, fieldInfo, not, condition, value, null);
	}

	/**
	 * Constructor of a filter expression (shortcut of the previous constructor with <i>not</i> set
	 * to false).
	 *
	 * @param fieldInfo FieldInfo on which this filter expression applies.
	 * @param condition Filter condition.
	 * @param value     Value of this expression (may be null).
	 */
	public LyFilterExpression(LyFieldInfo fieldInfo, Condition condition, Object value)
	{
		this(null, fieldInfo, false, condition, value, null);
	}

	/**
	 * Constructor of a filter expression (shortcut of the main constructor with <i>not</i> set to
	 * false and condition set to Condition.EQUALS).
	 *
	 * @param fieldInfo FieldInfo on which this filter expression applies.
	 * @param value     Value of this expression (may be null).
	 */
	public LyFilterExpression(LyFieldInfo fieldInfo, Object value)
	{
		this(null, fieldInfo, false, Condition.EQUALS, value, null);
	}

	/**
	 * Clone this filter expression.
	 */
	public LyFilterExpression cloneFilterExpression()
	{
		return cloneFilterExpression(null);
	}

	/**
	 * Clone this filter expression within a session.
	 */
	protected LyFilterExpression cloneFilterExpression(LyDataSession session)
	{
		LyFilterExpression exp = (LyFilterExpression)clone();
		Object 			   value = getInitValue();

		if ((session != null)
				&& (value instanceof String)
				&& (((String)value).length() > 0)
				&& (((String)value).charAt(0) == '$'))
			value = session.getValue(value);

		exp.setInitValue(value);
		exp._value = null;
		return exp;
	}

	/**
	 * Indicates whether the specified object matches the filter expression.
	 *
	 * @param  object Object which should be checked.
	 * @return True if the given object matches this expression, false otherwise.
	 * @see    LyFilterElement#match
	 */
	public boolean match(Object object)
	{
		if (object == null)
			return false;

		// Find the ObjectWrapper
		LyObject lyObject = (LyObject)object;
		Object   value = lyObject.getValue(_fieldInfo);

		return matchValue(value);
	}

	/**
	 * Indicates whether the specified values matches the filter expression.
	 *
	 * @param  values Values which should be checked.
	 * @return True if the given object matches this expression, false otherwise.
	 * @see    LyFilterElement#match
	 */
	public boolean match(LyValueSet values)
	{
		if (values == null)
			return false;

		// Find the ObjectWrapper
		LyValue lyValue = values.getFieldValue(_fieldInfo);
		Object  value = ((lyValue == null) ? null : lyValue.getValue());

		return matchValue(value);
	}

	/**
	 * Checks if a given date fieldInfo value matches the filter expression.
	 *
	 * @param value      the fieldInfo
	 * @param expValue   the expression
	 * @return           true if matches, false otherwise
	 */
	boolean matchDateValue(Date value, Date[] expValue)
	{
		boolean match = false;

		switch (_condition)
		{
		case EQUALS:
			match = (_fieldInfo.compare(value, expValue[0], _modifiers) >= 0)
			           && (_fieldInfo.compare(value, expValue[1], _modifiers) <= 0);
			break;
		case LOWER_THAN:
			match = (_fieldInfo.compare(value, expValue[0], _modifiers) < 0);
			break;
		case LOWER_OR_EQUALS:
			match = (_fieldInfo.compare(value, expValue[1], _modifiers) <= 0);
			break;
		case GREATER_THAN:
			match = (_fieldInfo.compare(value, expValue[1], _modifiers) > 0);
			break;
		case GREATER_OR_EQUALS:
			match = (_fieldInfo.compare(value, expValue[0], _modifiers) >= 0);
			break;
		default:
			break;
		}

		return match;
	}

	/**
	 * Checks if a given Date value matches the filter expression.
	 *
	 * @param value      the fieldInfo
	 * @param expValue   the expression
	 * @return           true if matches, false otherwise
	 */
	boolean matchOtherValue(Object value, Object expValue)
	{
		boolean match = false;

		switch (_condition)
		{
		case EQUALS:
			match = _fieldInfo.equals(value, expValue, _modifiers);
			break;
		case LOWER_THAN:
			match = (_fieldInfo.compare(value, expValue, _modifiers) < 0);
			break;
		case LOWER_OR_EQUALS:
			match = (_fieldInfo.compare(value, expValue, _modifiers) <= 0);
			break;
		case GREATER_THAN:
			match = (_fieldInfo.compare(value, expValue, _modifiers) > 0);
			break;
		case GREATER_OR_EQUALS:
			match = (_fieldInfo.compare(value, expValue, _modifiers) >= 0);
			break;
		case CONTAINS:
			match = _fieldInfo.contains(value, expValue, _modifiers);
			break;
		case STARTS_WITH:
			match = _fieldInfo.startsWith(value, expValue, _modifiers);
			break;
		case ENDS_WITH:
			match = _fieldInfo.endsWith(value, expValue, _modifiers);
			break;
		}

		return match;
	}

	/**
	 * Checks if a given fieldInfo value matches the filter expression.
	 *
	 * @param  value Value to check. This value is an instance of the class returned by the
	 *               getValueClass method of LyFieldInfo.
	 * @return True if the given value matches this expression, false otherwise.
	 */
	boolean matchValue(Object value)
	{
		if (hasModifier(MOD_PEER_FILTER_ONLY))
			return true;

		boolean match = false;
		Object  expValue = getValue();

		if(expValue != null && expValue.equals(LyDateFieldInfo.VALUE_TODAY))
		{
			Date[] dates = LyDateFieldInfo.getTodayDate();
			match = matchDateValue((Date)value, dates);
		}
		else if(expValue != null && expValue.toString().startsWith(LyDateFieldInfo.VALUE_DAY))
		{
			Date[] dates = LyDateFieldInfo.getDay((LyDateFieldInfo)_fieldInfo, expValue.toString());
			match = matchDateValue((Date)value, dates);
		}
		else
		{
			match = matchOtherValue(value, expValue);
		}

		return (_not ? (!match) : match);
	} // end method matchValue

    /**
     * Dumps this expression as a String.
     * @return Expression written in a string (may be displayed to the user or console)
     */
    public String dump()
    {
        return toString(false);
    }

	/**
	 * Dumps this expression as a String.
	 *
	 * @param  name boolean indicating if names or strings are requested in result string.
	 * @return Expression written in a string (may be displayed to the user or console)
	 */
	public String toString(boolean name)
	{
		StringBuffer result = new StringBuffer();

		// add Attribute info
		result.append((_fieldInfo == null) ? "null" : _fieldInfo.getName());
		result.append(" ");

		// add Condition (as string)
		if (_not)
			result.append(getNotCondition(_condition));
		else
			result.append(getCondition(_condition));
		result.append(" ");

        if ((_fieldInfo instanceof LyDateFieldInfo)
            && (_initValue instanceof String)
            && (_initValue.equals("_now")
                || _initValue.equals("_today")
                || ((String)_initValue).startsWith("_day")))
        {
            if (name)
                result.append(_fieldInfo.toName(_initValue));
            else
                result.append(_fieldInfo.toString(_initValue));
        }
		else
		{
			Object value;

			try
			{
				value = getValue();
			}
			catch (Exception e)
			{
				value = (_value != null ? _value.toString() : null);
			}

			if (value == null)
				result.append("LY_FILTER_EMPTY_VALUE");
			else if (_fieldInfo == null)
				result.append(value.toString());
			else if (name)
				result.append(_fieldInfo.toName(value));
			else if ((_fieldInfo instanceof LyRelationFieldInfo)
					|| ((value != null) && (value.toString().length() > 0) && value.toString().charAt(0) == '$'))
				// Workaround to let user type in $USER-like values
				result.append(value.toString());
			else
				result.append(_fieldInfo.toString(value));
		}

		return result.toString();
	} // end method toString

	/**
	 * Writes XML corresponding to this filter element.
	 *
	 * @param pw PrintWriter on which the XML is written.
	 */
	public void writeXml(PrintWriter pw)
	{
		// Write tag with its attributes
		pw.print("<EXP");

		String fieldId = getFieldInfo().getId();

		pw.print(" idref='" + LyString.getXmlString(fieldId) + "'");

		if (_not)
			pw.print(" not='true'");

		if (hasModifier(MOD_CASE_SENSITIVE))
			pw.print(" caseSensitive='true'");

		String condition = getCondition(_condition);

		condition = condition.substring(8);

		pw.print(" condition='" + condition + "'");

		String stringValue;

		if ((_initValue instanceof String)
				&& ((LyDateFieldInfo.VALUE_NOW.equals(_initValue)
						|| (LyDateFieldInfo.VALUE_TODAY.equals(_initValue))
						|| (_initValue != null && _initValue.toString().startsWith(LyDateFieldInfo.VALUE_DAY)))))
			stringValue = (String)_initValue;
		else
		{
			Object value = getValue();

			if ((_fieldInfo instanceof LyDateFieldInfo)
					&& (value instanceof Date))
			{
				stringValue = LyDateFormatter.format((Date) value, LyDateFieldInfo.DEFAULT_PEER_FORMAT);
			}
			else
			{
				stringValue = ((value == null) ? ""
										   : LyString.getXmlString(_fieldInfo.toString(value)));
			}
			stringValue = LyString.replace(stringValue, "" + _fieldInfo.getApplication().VALUE_SEP,
					"" + LyApplication.VISIBLE_VALUE_SEP);

			stringValue = LyString.replace(stringValue, "" + LyTools.VECTOR_SEP,
					LyApplication.VISIBLE_VECTOR_SEP);
		}

		pw.println(" value='" + stringValue + "'/>");
	} // end method writeXml

	/**
	 * Returns a displayable string of this expression.
	 *
	 * @return A displayable string representation of this expression.
	 */
	public String getName()
	{
		return toString();
	}

	/**
	 * Gets the fieldInfo on which this expression applies.
	 *
	 * @return FieldInfo concerned by this filter expression.
	 */
	public LyFieldInfo getFieldInfo()
	{
		return _fieldInfo;
	}

	/**
	 * Gets the condition for the filter expression.
	 *
	 * @return The condition : one of the Condition.XXX constants.
	 */
	public Condition getCondition()
	{
		return _condition;
	}

	/**
	 * Check whether this expression is inverted or not.
	 *
	 * @return True is the expression is inverted, false otherwise.
	 */
	public boolean isInverted()
	{
		return _not;
	}

	/**
	 * Check whether this expression is case sensitive or not.
	 *
	 * @return True is the expression is case sensitive, false otherwise.
	 */
	public boolean isCaseSensitive()
	{
		return hasModifier(MOD_CASE_SENSITIVE);
	}

	/**
	 * Checks whether the expression has the given modifier set.
	 *
	 * @param  modifier Modifier to set on the expression. See the MOD_XXX constants, default is
	 *                  null (no modifier).
	 * @return True if the modifier is on, false it is off.
	 */
	public boolean hasModifier(short modifier)
	{
		checkPeerOnlyFilter();
		return ((_modifiers != null) && _modifiers.get(modifier));
	}

	/**
	 * Gets the value on which comparison occur.
	 *
	 * @return Value used for comparison. This value is an instance of the class returned by the
	 *         getValueClass method of the fieldInfo concerned by this filter expression.
	 */
	public synchronized Object getValue()
	{
		if (_value == null)
		{
			if (_initValue == null)
				return null;

			_value = _initValue;
		}
		else if ((_recomputeDateDelay >= 0)
				&& ((System.currentTimeMillis() - _recomputeDate) >= _recomputeDateDelay))
		{
			// May recompute filter
			if ((_initValue instanceof String)
					&& (_initValue.equals(LyDateFieldInfo.VALUE_NOW)
							|| ((String)_initValue).contains(LyDateFieldInfo.VALUE_TODAY)
							|| ((String)_initValue).contains(LyDateFieldInfo.VALUE_DAY)))
				_value = _initValue;
			else if ((_initValue instanceof String)
					&& (((String)_initValue).length() > 0)
					&& (((String)_initValue).charAt(0) == '$'))
				_value = _initValue;
			else
				_recomputeDateDelay = -1; // Not a dynamic value

			_recomputeDate = System.currentTimeMillis();
		}
		else
			return _value;

		if (!_application.isEdited()
		        && (_value instanceof String)
				&& (((String)_value).length() > 0)
				&& (((String)_value).charAt(0) == '$'))
		{
			LyApplication application = getApplication();

			if (application != null)
			{
				LyDataSession session = application.getCurrentSession();

				if (session != null)
					_value = session.getValue(_value);
			}
		}

		if ((_fieldInfo != null)
				&& (_value != null)
				&& (_value.getClass() != _fieldInfo.getValueClass()))
		{
			try
			{
				_value = _fieldInfo.getValueFromObject(_value, null);
			}
			catch (Exception e)
			{
				LyApplication application = getApplication();
				LyEnvironment environment = application.getEnvironment();

				environment.writeException(e);
				_value = null;
			}
		}

		return _value;
	} // end method getValue

	/**
	 * Gets the initial value of this filter expression The values such as "_now" are replaced using
	 * getValue() but not in getInitValue().
	 *
	 * @return Value init used when creating the exepression.
	 */
	public Object getInitValue()
	{
		return _initValue;
	}

	/**
	 * Gets wrapped value.
	 *
	 * @return Returns wrapped value.
	 */
	public LyValue getFieldValue()
	{
		return _fieldInfo.buildFieldValue(getValue());
	}

	/**
	 * Gets modifiers of the expression.
	 *
	 * @return A bitset of modifiers (constants prefixed by MOD_).
	 */
	public BitSet getModifiers()
	{
		return ((_modifiers == null) ? null : (BitSet)_modifiers.clone());
	}

	/**
	 * Sets a modifier on this expression.
	 *
	 * @param modifier Modifier to set on the expression. See the MOD_XXX constants, default is null
	 *                 (no modifier).
	 */
	public void setModifier(short modifier)
	{
		if (_modifiers == null)
			_modifiers = new BitSet();
		_modifiers.set(modifier);
	}

	/**
     * Unsets a modifier on this expression.
     *
     * @param modifier Modifier to unset on the expression. See the MOD_XXX constants, default is null
     *                 (no modifier).
     */
    public void unsetModifier(short modifier)
    {
        if (_modifiers != null)
            _modifiers.clear(modifier);
    }

    /**
     * Sets whether this expression is case sensitive or not.
     *
     * @param caseSensitive True is the expression is case sensitive, false otherwise.
     */
    public void setCaseSensitive(boolean caseSensitive)
    {
        if (caseSensitive)
            setModifier(MOD_CASE_SENSITIVE);
        else
            unsetModifier(MOD_CASE_SENSITIVE);
    }

	/**
	 * Sets the fieldInfo on which this expression applies.
	 *
	 * @param fieldInfo FieldInfo concerned by this filter expression.
	 */
	public void setFieldInfo(LyFieldInfo fieldInfo)
	{
		if (isLocked())
			throw new IllegalStateException("BF AppEngine : filter expression " + this
				+ " was locked and thus can't be modified anymore");

		_fieldInfo = fieldInfo;
		_checkedPeerFilterOnly = true; // check next time
	}

	/**
	 * Check if this field if peer only (not verified in the GUI)
	 */
	protected void checkPeerOnlyFilter()
	{
		if ((_fieldInfo != null) && (_checkedPeerFilterOnly))
		{
			int mark = LyFieldInfo.getMark("PEER_FILTER_ONLY");

			if ((mark > 0) && (_fieldInfo.hasMark(mark)))
				setModifier(MOD_PEER_FILTER_ONLY);
			_checkedPeerFilterOnly = false; // do not check again
		}
	}

    /**
     * Set the inverted mode of this filter expression.
     *
     * @param not indicates whether this filter expression should be inverted or not.
     */
    public void setInverted(boolean not)
    {
        if (isLocked())
            throw new IllegalStateException("BF AppEngine : filter expression " + this
                + " was locked and thus can't be modified anymore");

        _not = not;
    }

	/**
     * Checks if this filter expression is dynamic or not.
     * The filter expression is dynamic when it has an value that depends on a dynamic value
     * given by the session.
     */
    public boolean isDynamic()
    {
        Object value = getInitValue();

        if ((value instanceof String)
                && (((String)value).length() > 0)
                && (((String)value).charAt(0) == '$'))
            return true;
        else
            return false;
    }

    /**
     * Sets the condition for the filter expression.
     *
     * @param condition The condition : one of the Condition.XXX constants.
     */
    protected void setCondition(Condition condition)
    {
        _condition = (condition == null ? Condition.EQUALS : condition);
    }

    /**
     * Sets the initial value of this filter expression The values such as "_now" are replaced using
     * getValue() but not in getInitValue().
     *
     * @param value Value init used when creating the expression.
     */
    protected void setInitValue(Object value)
    {
        if (value instanceof LyValue)
            _initValue = ((LyValue)value).getValue();
        else
            // Conversion in the right type for the field info will be done
            // on first use by getValue method
            _initValue = value;
    }

    public LyFilterExpression[] getPeerExpression()
    {
    	LyFilterExpression[] filterExpr = null;

    	Date[] dates;
    	if(LyDateFieldInfo.VALUE_TODAY.equals(_value))
    		dates = LyDateFieldInfo.getTodayDate();
    	else if (_value != null && ((String)_value).startsWith(LyDateFieldInfo.VALUE_DAY))
    		dates = LyDateFieldInfo.getDay((LyDateFieldInfo)_fieldInfo, (String)_value);
    	else return new LyFilterExpression[] { this };

    	switch (_condition)
		{
		case EQUALS:
			filterExpr = new LyFilterExpression[2];

			if(_not)
			{
				filterExpr[0] = new LyFilterExpression(_fieldInfo, Condition.GREATER_THAN, dates[1]);
				filterExpr[1] = new LyFilterExpression(_fieldInfo, Condition.LOWER_THAN, dates[0]);
			}
			else
			{
				filterExpr[0] = new LyFilterExpression(_fieldInfo, Condition.GREATER_OR_EQUALS, dates[0]);
				filterExpr[1] = new LyFilterExpression(_fieldInfo, Condition.LOWER_OR_EQUALS, dates[1]);
			}
			break;
		case LOWER_THAN:
			filterExpr = new LyFilterExpression[1];
			filterExpr[0] = new LyFilterExpression(_fieldInfo, Condition.LOWER_THAN, dates[0]);
			break;
		case LOWER_OR_EQUALS:
			filterExpr = new LyFilterExpression[1];
			filterExpr[0] = new LyFilterExpression(_fieldInfo, Condition.LOWER_OR_EQUALS, dates[1]);
			break;
		case GREATER_THAN:
			filterExpr = new LyFilterExpression[1];
			filterExpr[0] = new LyFilterExpression(_fieldInfo, Condition.GREATER_THAN, dates[1]);
			break;
		case GREATER_OR_EQUALS:
			filterExpr = new LyFilterExpression[1];
			filterExpr[0] = new LyFilterExpression(_fieldInfo, Condition.GREATER_OR_EQUALS, dates[0]);
			break;
		default:
			break;
		}

    	return filterExpr;
    }
} // end class LyFilterExpression
