/*
 * LyFilterRelation.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.io.*;
import java.util.*;
import leon.data.*;
import leon.misc.*;


/**
 * The LyFilterRelation class implements the description of a filter used to filter objects list on
 * a relation field.<br>
 * A relation filter is defined by a relation field info and by a filter.
 *
 * <p>See also tag <FILTER> of leon.dtd</p>
 *
 * @see LyFilter
 * @see LyFilterExpression
 */
public final class LyFilterRelation extends LyInfo
	implements LyFilterElement
{
	public static final String __VERSION =
		"$Id: LyFilterRelation.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Relation FieldInfo on which this filter expression applies. */
	LyRelationFieldInfo		   _fieldInfo;

	/**
	 * ClassInfo on which this filter expression applies. This parameter is just indicative and may
	 * not be set. It is not persistent (not saved in the XML description of such a filter).
	 */
	LyClassInfo				   _classInfo;

	/** Filter used on the relation. */
	LyFilter				   _filter;

	/**
	 * Optional info : In order to OPTIMIZE database select, indicate which are the accepted target
	 * classes. Content is set by the application to avoid searching in useless possible classes.
	 */
	ArrayList<LyClassInfo>     _acceptedTargetClassInfos;

	/** Indicates a fast search in database without aliases. */
	boolean					   _ignoreAliases = false;

	/**
	 * Initialisation of filter attributes.
	 */
	protected void init()
	{
		_fieldInfo = null;
		_filter = null;
		_classInfo = null;
	}

	/**
	 * Constructor of an identified relation filter.
	 *
	 * @param id        The identifier of the object.
	 * @param fieldInfo Relation FieldInfo on which this filter expression applies.
	 * @param filter    Filter used for this relation filter.
	 */
	public LyFilterRelation(String id, LyRelationFieldInfo fieldInfo, LyFilter filter)
	{
		super(id);

		init();

		_fieldInfo = fieldInfo;
		_filter = filter;

		// Set application for dynamic filters
		if (fieldInfo != null)
			_application = fieldInfo.getApplication();
	}

	/**
	 * Constructor of an identified relation filter.
	 *
	 * @param fieldInfo Relation FieldInfo on which this filter expression applies.
	 * @param filter    Filter used for this relation filter.
	 */
	public LyFilterRelation(LyRelationFieldInfo fieldInfo, LyFilter filter)
	{
		this(null, fieldInfo, filter);
	}

	/**
	 * Constructor of a relation filter with a given field info.
	 *
	 * @param fieldInfo FieldInfo on which this filter expression applies.
	 */
	public LyFilterRelation(LyRelationFieldInfo fieldInfo)
	{
		this(null, fieldInfo, null);
	}

	/**
	 * Constructor of a default relation filter.
	 */
	public LyFilterRelation()
	{
		this(null, null, null);
	}

	/**
	 * Check if the specified object matches current filter.
	 *
	 * @param  object Object to check.
	 * @return True if the object matches filter criteria, false otherwise.
	 * @see    LyFilterElement#match
	 */
	public boolean match(Object object)
	{
		if (object == null)
			return false;
		if (_filter == null)
			return true;

		// Find the ObjectWrapper
		LyObject	    lyObject = (LyObject)object;
		LyRelationValue value = lyObject.getRelationValue(_fieldInfo);

		return matchRelationValue(value);
	}

	/**
	 * Check if the specified values matches current relation filter.
	 *
	 * @param  values Values to check.
	 * @return True if the object matches filter criteria, false otherwise.
	 * @see    LyFilterElement#match
	 */
	public boolean match(LyValueSet values)
	{
		if (values == null)
			return false;
		if (_filter == null)
			return true;

		LyRelationValue value = (LyRelationValue)values.getFieldValue(_fieldInfo);

		return matchRelationValue(value);
	}

	/**
	 * Checks if a given relation fieldInfo value matches the filter.
	 *
	 * @param  relationValue Value to check.
	 * @return True if at least one object of the relation matches filter, false otherwise.
	 */
	protected boolean matchRelationValue(LyRelationValue relationValue)
	{
		if (relationValue == null)
			return false;

		short		  i, count = relationValue.getValueCount();
		LyObject	  relationObject;
		LyClassInfo   classInfo;
		LyEnvironment environment = getEnvironment();

		for (i = 0; i < count; i++)
		{
			relationObject = relationValue.getObject(i);
			if (relationObject == null)
			{
				environment.logError("BF AppEngine : Unknown object found "
					+ relationValue.getName()
					+ " when filtering on relation " + _fieldInfo);
				continue;
			}

			// Check that related object contains field info of current filter
			classInfo = relationObject.getClassInfo();

			if (!_filter.matchClass(classInfo))
				continue;

			// If at least one object match the filter, return true
			if (_filter.match(relationObject))
				return true;
		}

		return false;
	} // end method matchRelationValue

	/**
	 * Lock this relation filter.
	 */
	public void lock()
	{
		super.lock();

		// Lock filter
		if (_filter != null)
			_filter.lock();
	}

	/**
	 * Clone this relation filter.
	 */
	public LyFilterRelation cloneFilterRelation()
	{
		LyFilterRelation filter = new LyFilterRelation(getRelationFieldInfo(),
				_filter.cloneFilter());

		filter.setClassInfo(_classInfo);

		return filter;
	}

	/**
	 * Dumps current relation filter as a String.
	 *
	 * @return String representation of this relation filter.
	 */
	public String dump()
	{
		StringBuffer sb = new StringBuffer(500);

		sb.append((_fieldInfo == null) ? "null" : _fieldInfo.getName());
		sb.append(" match ");
		sb.append((_filter == null) ? "null" : _filter.toString());

		return sb.toString();
	}

	/**
	 * Writes XML corresponding to this relation filter element.
	 *
	 * @param pw PrintWriter on which the XML is written.
	 */
	public void writeXml(PrintWriter pw)
	{
		// Write tag with its attributes
		pw.print("<FILTER_RELATION");

		String fieldId = getRelationFieldInfo().getId();

		pw.print(" idref='" + LyString.getXmlString(fieldId) + "'");
		pw.println(">");

		_filter.writeXml(pw);

		pw.println("</FILTER_RELATION>");
	}

	/**
	 * Returns a displayable string of this element.
	 *
	 * @return A displayable string representation of this element.
	 */
	public String getName()
	{
		return toString();
	}

	/**
	 * Checks if an expression is found concerning the given field info.
	 *
	 * @param  fieldInfo The fieldInfo.
	 * @return True if an expression with the specified is found, false otherwise.
	 */
	public boolean contains(LyFieldInfo fieldInfo)
	{
		if (_fieldInfo == fieldInfo)
			return true;

		return ((_filter != null) && (_filter.contains(fieldInfo)));
	}

	/**
	 * Gets the fieldInfo on which this relation filter applies.
	 *
	 * @return RelationFieldInfo concerned by this relation filter.
	 */
	public LyRelationFieldInfo getRelationFieldInfo()
	{
		return _fieldInfo;
	}

	/**
	 * Gets the classInfo on which this relation filter applies.
	 *
	 * @return ClassInfo concerned by this relation filter.
	 */
	public LyClassInfo getClassInfo()
	{
		if ((_classInfo == null) && (_filter != null) && (_fieldInfo != null))
		{
			// If classInfo was not set, retrieve a class that can be used for current filter.
			int		    i, n = _fieldInfo.getClassInfoCount();
			LyClassInfo classInfo;

			for (i = 0; i < n; i++)
			{
				classInfo = _fieldInfo.getClassInfo(i);

				if (_filter.matchClass(classInfo))
				{
					_classInfo = classInfo;
					break;
				}
			}
		}

		return _classInfo;
	}

	/**
	 * Get the filter from this relation filter.
	 *
	 * @return Filter of this relation filter.
	 */
	public LyFilter getFilter()
	{
		return _filter;
	}

	/**
	 * Gets list of fieldInfos that this filter refers to.
	 *
	 * @return List of fieldInfos that the filter depends on.
	 */
	public LyInfoList<LyFieldInfo> getFieldInfos()
	{
		LyInfoList<LyFieldInfo> fieldInfos = new LyInfoList<LyFieldInfo>();

		if (_fieldInfo != null)
			fieldInfos.addInfo(_fieldInfo);

		if (_filter != null)
			fieldInfos.addList(_filter.getFieldInfos());

		return fieldInfos;
	}

	/**
	 * Returns the list of class infos on which the relation filter may be applied.
	 *
	 * @param  notAbstract Specifies if the search classInfos must be only concrete classes or not.
	 * @return the list of class infos on which the relation filter may be applied
	 */
	public LyInfoList<LyClassInfo> getPossibleClasses(boolean notAbstract)
	{
		LyInfoList<LyClassInfo> result = new LyInfoList<LyClassInfo>();
		int					    i, nClass = getAcceptedTargetCount();

		if (nClass > 0)
		{
			// We take EXACTLY what the user asked regardless of what is available on the relation
			// in order to optimize the filter search using a parent abstract classInfo
			for (i = 0; i < nClass; i++)
				result.addInfo(getAcceptedTargetClassInfos(i));

			return result;
		}

		LyRelationFieldInfo     fieldInfo = getRelationFieldInfo();
		LyInfoList<LyClassInfo> classInfos = fieldInfo.getClassInfos();
		LyInfoList<LyClassInfo> subClasses;
		LyClassInfo			    classInfo;
		LyFilter			    filter = getFilter();
		int					    j, nSubClasses;
		boolean				    classMatchFilter;

		nClass = classInfos.getSize();
		for (i = 0; i < nClass; i++)
		{
			classInfo = classInfos.getInfo(i);
			classMatchFilter = filter.matchClass(classInfo);

			if (classInfo.isAbstract() && (notAbstract || (!classMatchFilter)))
			{
				subClasses = classInfo.getSubClasses(true);
				nSubClasses = ((subClasses == null) ? 0 : subClasses.getSize());
				for (j = 0; j < nSubClasses; j++)
				{
					classInfo = subClasses.getInfo(j);
					if ((!classInfo.isAbstract()) && filter.matchClass(classInfo))
						result.addInfo(classInfo);
				}
			}
			else if (classMatchFilter)
				result.addInfo(classInfo);
		}

		return result;
	} // end method getPossibleClasses

	/**
	 * Sets the fieldInfo on which this expression applies.
	 *
	 * @param fieldInfo FieldInfo concerned by this filter expression.
	 */
	public void setRelationFieldInfo(LyRelationFieldInfo fieldInfo)
	{
		if (isLocked())
			throw new IllegalStateException("BF AppEngine : relation filter " + this
				+ " was locked and thus can't be modified anymore");

		_fieldInfo = fieldInfo;
	}

	/**
	 * Sets the filter to this relation filter.
	 *
	 * @param filter Filter to set to this relation filter.
	 */
	public void setFilter(LyFilter filter)
	{
		if (isLocked())
			throw new IllegalStateException("BF AppEngine : relation filter " + this
				+ " was locked and thus can't be modified anymore");

		_filter = filter;
	}

	/**
	 * Sets the classInfo on which this expression applies.
	 *
	 * @param classInfo ClassInfo concerned by this filter expression.
	 */
	public void setClassInfo(LyClassInfo classInfo)
	{
		if (isLocked())
			throw new IllegalStateException("BF AppEngine : relation filter " + this
				+ " was locked and thus can't be modified anymore");

		_classInfo = classInfo;
	}

	/**
	 * (Optional, used by the application) Add a possible target classInfo to the list. If
	 * specified, the filter will apply ONLY in target classInfos
	 *
	 * @param classInfo target classInfo to look for filtering
	 */
	public void addAcceptedTargetClassInfo(LyClassInfo targetClassInfo)
	{
		if (targetClassInfo == null)
			return;
		if (_acceptedTargetClassInfos == null)
			_acceptedTargetClassInfos = new ArrayList<LyClassInfo>();
		if (!_acceptedTargetClassInfos.contains(targetClassInfo))
			_acceptedTargetClassInfos.add(targetClassInfo);
	}

	/**
	 * Optional info : In order to OPTIMIZE database select, get the accepted target class. Content
	 * is set by the application to avoid searching in useless possible classes.
	 *
	 * @param  index index of the classInfo
	 * @return target classInfos to look for filtering
	 */
	public LyClassInfo getAcceptedTargetClassInfos(int index)
	{
		if (_acceptedTargetClassInfos == null)
			return null;
		else
			return _acceptedTargetClassInfos.get(index);
	}

	/**
	 * Optional info : In order to OPTIMIZE database select, indicate the number of classes which
	 * are the accepted as target classes. Content is set by the application to avoid searching in
	 * useless possible classes.
	 *
	 * @return the count of targetClassInfos
	 */
	public int getAcceptedTargetCount()
	{
		if (_acceptedTargetClassInfos == null)
			return 0;
		else
			return _acceptedTargetClassInfos.size();
	}
} // end class LyFilterRelation
