/**
 * LyInclude.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;

import leon.app.*;
import leon.misc.*;


/**
 * This class represents an include of an external model. Its attributes are :
 *
 * <ul>
 *   <LI><b>id</b><br>
 *     Id of the external model.</LI>
 *   <LI><b>file</b> (identifier)<br>
 *     XML root file of the external model.</LI>
 * </ul>
 *
 * @version $Id: LyInclude.java 41139 2012-10-11 09:50:31Z bangiole $
 * @author  Lyria
 */
public class LyInclude extends LyInfo
{
	private static final String	W4_MODEL_ID		= "w4Model";
	private static final String	CMS_MODEL_ID	= "cmsModel";

	public static final String	W4_APPMODEL_APPLICATION_ID	= "eu.w4.leon.appmodel.appdefault";
	public static final String	CMS_APPMODEL_APPLICATION_ID	= "eu.w4.cms.appmodel.appdefault";
	public static final String __VERSION =
		"$Id: LyInclude.java 41139 2012-10-11 09:50:31Z bangiole $";

	/* Static attributes */

	/** The include must load its children. */
	public static boolean	   _allowLoad = true;

	/* Protected attibutes */

	/** the application id of the external model. */
	protected String		   _applicationId;

	/** Version of this include. */
	String					   _version;

	/** Boolean that indicates if this include is loaded on start or on its first use. */
	boolean					   _load;

	/** Root element of this include (null while not loaded). */
	Object					   _rootElement;

    /** XML file of this include. */
    String                     _xmlFile;

    List<String>			   _loadedLanguage = new ArrayList<String>();

    /**
     * Default constructor.
     */
    protected LyInclude()
    {
    }

	/**
	 * Constructor with a given id.
	 *
	 * @param applicationId The application id of the external model. Used to load properties file.
	 * @param file          The root file of the external model
     *                      (could be a ini file, a XML file or a compiled model).
	 * @param version       Version of loaded model.
	 * @param load          True if the model is loaded on start or on its first use.
	 */
	public LyInclude(String id, String applicationId, String file, String version, boolean load)
	{
		super(((id == null) ? ("_Include" + applicationId) : id), false);

		if (applicationId != null)
			_applicationId = applicationId.intern();
		else
			_applicationId = null;

        setFile(file);
		setXmlFile(file);
        setVersion(version);
		setLoadedOnStart(load);

		LyApplication application = null;

		if (applicationId != null)
			application = LyApplication.getApplication(applicationId);

		if (application != null)
		{
			_loaded = true;
			_application = application;
			if (_application.getInfo(_id) == null)
				_application.putInfo(this);
		}

		_rootElement = null;
	}

	/**
	 * Constructor with a given id.
	 *
	 * @param applicationId The application id of the external model. Used to load properties file.
	 * @param file          The root file of the external model.
	 * @param version       Version of loaded model.
	 * @param load          True if the model is loaded on declaration or on its first use.
	 */
	public LyInclude(String applicationId, String file, String version, boolean load)
	{
		this(null, applicationId, file, version, load);
	}

	/**
	 * Sets the application for this info. Overridden to load include if needed only when application
	 * has been set.
	 *
	 * @param     application The application to which the info belongs.
	 * @exception IllegalStateException If the info already belongs to an application.
	 */
	public void setApplication(LyApplication application)
	{
		if (_application == null)

			super.setApplication(application);

		if (_load)
			load(this);
	}

	/**
	 * Gets the application id of the external model.
	 *
	 * @return The application id.
	 */
	public final String getApplicationId()
	{
		return _applicationId;
	}

    /**
     * Sets the XML file for this include.
     *
     * @param the XML file for this include.
     */
    public void setXmlFile(String xmlFile)
    {
        _xmlFile = xmlFile;
    }

	/**
	 * Gets the XML root file of the external model.
	 *
	 * @return The XML root file.
	 */
	public final String getXmlFile()
	{
		return _xmlFile;
	}

	/**
	 * Gets version of this include.
	 *
	 * @return Version.
	 */
	public String getVersion()
	{
		return _version;
	}

	/**
	 * Checks if this include is loaded on start or on its first use.
	 *
	 * @return True if this include is loaded on start, false otherwise.
	 */
	public boolean isLoadedOnStart()
	{
		return _load;
	}

	/**
	 * Retrieve the root element of the include.
	 *
	 * @return Root element loaded by this include.
	 */
	public Object getRootElement()
	{
		return _rootElement;
	}

	/**
	 * Computes the default location id for the include application.
	 *
	 * @return the default location id of the include application.
	 */
	String computeDefaultLocationId()
	{
		return LyApplication.buildDefaultLocationId(getApplicationId());
	}

	/**
	 * Loads when needed a class info from the XML file that contains it.
	 *
	 * @param IllegalStateException If the root element of the XML file is neither a project no a
	 *                              classInfo.
	 * @param IllegalStateException If this class info was not found in its XML file.
	 */
	void load(LyInfo info)
	{
		if (_loaded)
			return;

		if (!_allowLoad)
			return;

		String infoId = ((info == this) ? getApplicationId() : info.getId());

		if (_file.endsWith(".ini"))
		{
			// Mark include as loaded
			_loaded = true;

			LyApplication application = new LyApplication(_application, getApplicationId(), _file,
					null, null);

			if (!application.isValid())
				throw new IllegalStateException("BF AppEngine : included application "
					+ getApplicationId() + " not valid");
			if (!info.isLoaded())
				throw new IllegalStateException("BF AppEngine : info with " + infoId
					+ " was not found in the specified XML file " + _file);

			_rootElement = application.getProject();
			((LyProject)_rootElement)._parent = this;

			checkInfos();
			return;
		}

		LyEnvironment environment = _application.getEnvironment();
		String		  defaultLocationId = _application.getDefaultLocationId();

		try
		{
    		_application.setDefaultLocationId(computeDefaultLocationId());

    		if (_applicationId != null)
    		{
    			if (W4_MODEL_ID.equalsIgnoreCase(_applicationId))
    				environment.load(W4_APPMODEL_APPLICATION_ID, null, false);
    			else if (CMS_MODEL_ID.equalsIgnoreCase(_applicationId))
    				environment.load(CMS_APPMODEL_APPLICATION_ID, null, false);
    			else
    				environment.load(_applicationId, null, false);
    		}
    		// Mark include as loaded
    		_loaded = true;

    		if (environment.getLanguage() != null)
    			_loadedLanguage.add(environment.getLanguage());

    		if (_file.endsWith(".xml"))
    		{
    			environment.logInfo("Loading XML meta-model : " + _file
    				+ " when accessing object info " + infoId);

    			LyInfoParser parser = new LyInfoParser(_application, _file);

    			_rootElement = parser.getRootElement();
    			if (_rootElement instanceof LyTag)
    				((LyTag)_rootElement)._parent = this;
    		}
    		else
    		{
    			environment.logInfo("Loading compiled meta-model : " + _file
    				+ " when accessing object info " + infoId);

    			try
    			{
    				// Create a new meta model.
    				Class<?>    infoClass = _application.loadClass(_file);
    				LyMetaModel metaModel = (LyMetaModel)infoClass.newInstance();

    				metaModel.load(_application);
    				_rootElement = metaModel.getRootElement();
        			((LyObjectInfo)_rootElement)._parent = this;
    				_loaded = true;
    			}
    			catch (Exception e)
    			{
    				environment.logError("Error loading class " + _file);
    				environment.writeException(e);
    			}
    		}

    		if (_rootElement instanceof LyClassInfo)
    		{
    			if (!_application.isEdited())
    			{
	    			LyInfoList<LyClassInfo> classes = new LyInfoList<LyClassInfo>();

	    			classes.addInfo((LyClassInfo)_rootElement);
	    			_application.loadClasses(classes);
    			}
    		}
    		else if (_rootElement instanceof LyProject)
    		{
    			LyProject project = ((LyProject)_rootElement);

    			if (_version != null)
    			{
    				String projectVersion = project.getVersion();

    				if ((projectVersion == null)
    						|| (!_version.equals(projectVersion)))
    					throw new IllegalStateException("BF AppEngine : invalid version for project "
    						+ project.getId() + " (found " + projectVersion
    						+ ", waiting for " + _version + ")");
    			}

    			if (!_application.isEdited())
    				_application.loadProject(project);
    		}
    		else
    			throw new IllegalStateException("BF AppEngine : invalid extern file " + _file
    				+ ", it should contain a project or a class info as root element");

    		if (!info.isLoaded())
    			throw new IllegalStateException("BF AppEngine : info with " + infoId
    				+ " was not found in the specified XML file " + _file);

    		checkInfos();
		}
		finally
		{
		    _application.setDefaultLocationId(defaultLocationId);
		}
	} // end method load

	protected void checkInfos()
	{
		if (!isLoaded())
			return;

		// Check all infos with this include as parent to detect invalid references
		LyInfoList<LyInfo> infos = _application.getInfoList(null);
		LyInfo	           info;
		int                i, n = (infos == null ? 0 : infos.getSize());

		for (i = 0; i < n; i++)
		{
			info = infos.getInfo(i);
			if ((info.getParent() == this)
					&& !info.isLoaded())
				getEnvironment().logWarning("BF AppEngine : info with " + info.getId()
					+ " was not found in the specified XML file " + _file);
		}
	}

    /**
     * Sets version of this include.
     *
     * @param version Version.
     */
    protected void setVersion(String version)
    {
        _version = version;
    }

    /**
     * Checks if this include is loaded on start or on its first use.
     *
     * @param loadedOnStart True if this include is loaded on start, false otherwise.
     */
    protected void setLoadedOnStart(boolean loadedOnStart)
    {
        _load = loadedOnStart;
    }

    /**
     * Sets the application id of the external model.
     *
     * @param applicationId The application id.
     */
    protected void setApplicationId(String applicationId)
    {
        _applicationId = applicationId;
    }

    public void loadExtraLanguage(LyEnvironment environment)
    {
    	if (environment.getLanguage() == null)
    		return;

    	if (_loadedLanguage.contains(environment.getLanguage()))
    			return;

    	if (_applicationId != null)
		{
			if (W4_MODEL_ID.equalsIgnoreCase(_applicationId))
				environment.loadInParentBundles(W4_APPMODEL_APPLICATION_ID, null);
			else if (CMS_MODEL_ID.equalsIgnoreCase(_applicationId))
				environment.loadInParentBundles(CMS_APPMODEL_APPLICATION_ID, null);
			else
				environment.loadInParentBundles(_applicationId, null);
		}
    	if (environment.getLanguage() != null)
			_loadedLanguage.add(environment.getLanguage());
    }
} // end class LyInclude
