/**
 * LyInfo.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import leon.app.*;
import leon.misc.*;
import leon.parser.*;


/**
 * This class implements a simple object with an id built when parsing the data model. Its
 * attributes are :
 *
 * <ul>
 *   <LI><b>application</b><br>
 *     Reference to the application to which this object belongs (several applications can be
 *     handled in the same process).</LI>
 *   <LI><b>id</b> (identifier)<br>
 *     The identifier should be unique within the application but no control enforces it. It's
 *     usually not displayed to the end-user and reserved for an internal use. If an id is not
 *     given, an automatic one is given on a random basis In that case, id starts with an underscore
 *     to indicates that it was internally assigned.<br>
 *     Leonardi pre-defined infos have an id starting with underscore (_). Applications should avoid
 *     to start infos ids by _.<br>
 *     An info can be retrieved within an application by its id.</LI>
 * </ul>
 */
public abstract class LyInfo extends LyTag
	implements LyCommentInterface
{
	public static final String __VERSION = "$Id: LyInfo.java 41259 2012-10-17 12:06:14Z bangiole $";

	/**
	 * Tests if this info is internal (created by Leonardi) or not.
	 */
	public static final boolean isInternal(String id)
	{
		return ((id != null)
				&& (id.startsWith("_leon")
					|| id.startsWith("__")
					|| ((id.charAt(0) == '_') && (id.charAt(id.length() - 1) == '_'))));
	}


	/** Application to which this info belongs. */
	LyApplication	   _application;

	/** Identifier of the info. */
	String			   _id;

	/** Boolean indicating if the info was loaded or not. */
	boolean			   _loaded;

	/** Boolean indicating if this info is locked or not. */
	boolean			   _locked;

	/** The location of the xml description of this info. */
	String			   _file;

	/** Number of sons. */
	int				   _sonsCount;


	/**
	 * Init method used to initialize attributes of a new object. This method is called by
	 * constructors and may be overriden in subclasses to complete initialization.
	 */
	protected void init()
	{
	}

	/**
	 * Default constructor with no parameter. Since no id is given, an id is automatically
	 * attributed to this instance.
	 */
	protected LyInfo()
	{
		this(null, true);
	}

	/**
	 * Constructor with a given id.
	 *
	 * @param id The identifier of the object. If null, a random id started with an underscore is
	 *           automatically given as the id of the info.
	 */
	protected LyInfo(String id)
	{
		this(id, true);
	}

	/**
	 * Constructor with a given id.
	 *
	 * @param id     The identifier of the object. If null, a random id started with an underscore
	 *               is automatically given as the id of the info.
	 * @param loaded Boolean indicating if this info was loaded or not.
	 */
	protected LyInfo(String id, boolean loaded)
	{
		_application = null;
		_id = null;
		_sonsCount = 0;
		_loaded = false;
		_locked = false;
		_file = null;

		init();

		// Automatically builds an id if the given one is null
		if (id == null)
			id = "__" + LyString.getRandomId() + "__";

		_id = id.intern();

		// Set loaded attribute
		setLoaded(loaded);
	}

	/**
	 * Construct this info.
	 */
	protected void construct()
	{
		if (_loaded)
			// Reset info content
			init();
		else
			// Mark object as loaded
			setLoaded(true);
	}

	/**
	 * Lock info to avoid modifications.
	 */
	protected void lock()
	{
		if (_loaded)
			// Info can't be locked if not yet loaded
			_locked = true;
	}

	/**
	 * Internal method to free an info : remove it from its application.
	 */
	protected void free()
	{
		if (_application == null)
			return;

		_application.removeInfo(this);
		_application = null;
	}

	/**
	 * Loads when needed a class info from the XML file that contains it.
	 *
	 * @param IllegalStateException If the root element of the XML file is neither a project no a
	 *                              classInfo.
	 * @param IllegalStateException If this class info was not found in its XML file.
	 */
	public void load()
	{
		if (!_loaded
				&& (_parent instanceof LyInclude))
			((LyInclude)_parent).load(this);
	}

	/**
	 * Tests if the class is loaded or not.
	 *
	 * @return True if this class is loaded, false otherwise.
	 */
	public boolean isLoaded()
	{
		return _loaded;
	}

	/**
     * Returns extra tags.
     */
    protected LyTag[] getExtraTags()
    {
        return null;
    }

	/**
	 * Accessor for the id of the info.
	 *
	 * @return Idenfifier of the info.
	 */
	public final String getId()
	{
		return _id;
	}

	/**
	 * Accessor for the name of the info.
	 * To be overridden, default implementation returns null.
	 *
	 * @return Name of the info.
	 */
	public String getName()
	{
		return null;
	}

	/**
	 * Accessor for the alias of the info.
	 * To be overridden, default implementation calls getName().
	 *
	 * @return Alias of the info.
	 */
	public String getAlias()
	{
		return getName();
	}

	/**
	 * Accessor for the application to which the info belongs.
	 *
	 * @return Application of the info.
	 */
	public final LyApplication getApplication()
	{
		if ((_application == null) && (_parent != null))
			_application = _parent.getApplication();

		return _application;
	}

	/**
	 * Checks if this info is locked or not.
	 */
	public final boolean isLocked()
	{
		return _locked;
	}

	/**
	 * Accessor to the file of this info.
	 *
	 * @return the file of this info.
	 */
	public final String getFile()
	{
		return _file;
	}

	/**
	 * Converts this info to a string (returns the id of the info).
	 */
	public final String toString()
	{
		return getId();
	}

	/**
	 * Checks if an info is included or not (not declared as extern but loaded by an include).
	 *
	 * @return True if the info is included, false otherwise.
	 */
	public final boolean isIncluded()
	{
		return isExtern()
		    || ((_application != null)
				&& (_application.getProject() != getRootProject()));
	}

	/**
	 * Gets include from which is info comes from.
	 * @return Include if this info belongs to an include, false otherwise.
	 */
	public LyInclude getInclude()
	{
		if (isExtern())
		{
			LyInfo	parent = getParent();

			while ((parent != null)
				   && !(parent instanceof LyInclude))
				parent = parent.getParent();

			return (LyInclude)parent;
		}
		else if (_application != null)
		{
			LyProject	rootProject = getRootProject();

			if (_application.getProject() == getRootProject())
				return null;

			LyInfoList<LyInclude>	includes = _application.getInfoList(LyInclude.class);
			int						i, n = (includes == null ? 0 : includes.getSize());
			LyInclude				include;

			for (i = 0; i < n; i++)
			{
				include = includes.getInfo(i);
				if (include.getRootElement() == rootProject)
					return include;
			}
		}

		return null;
	}

	/**
	 * Tests if this info is internal (created by Leonardi) or not.
	 */
	public final boolean isInternal()
	{
		return isInternal(_id);
	}

    /**
     * Sets id of this info.
     */
    protected void setId(String id)
    {
        if ((id != null) && (_id != null) && id.equals(_id))
            return;

        // Replace random id by an id based on parent id
        if ((_application != null)
                && (_id != null))
            _application.removeInfo(this);

        // Automatically builds an id if the given one is null
        if (id == null)
        {
            if (_parent != null)
            {
                String className;

                className = getClass().getName();
                className = className.substring(className.lastIndexOf(".Ly") + 3);

                id = '_' + _parent.getId() + '.' + className + _parent._sonsCount + '_';
            }
            else
                id = "__" + LyString.getRandomId() + "__";
        }

        _id = id.intern();

        if (_application != null)
            _application.putInfo(this);
    }

	/**
	 * Sets the application for this info. Once assigned, the application can be changed.
	 *
	 * @param     application The application to which the info belongs.
	 * @exception IllegalStateException If the info already belongs to an application.
	 */
	public void setApplication(LyApplication application)
	{
		if (application == null)
		{
			if (_application != null)
			{
				_application.removeInfo(this);
				_application = null;
			}

			return;
		}

		if ((_application != null)
				&& (application != _application))
			// Application can't be changed once set
			return;

		_application = application;

		// Add this info to the parent application if not already done
		LyInfo	oldInfo = _application.getInfo(getId());

		if (oldInfo == null)
			_application.putInfo(this);
//		else if (oldInfo != this)
//			throw new IllegalStateException("Leonardi : an info with id " + getId()
//				+ " is already registered on application " + this);

		// Register extra tags
		LyTag[]   extraTags = getExtraTags();
		int       i, n = (extraTags == null ? 0 : extraTags.length);

		for (i = 0; i < n; i++)
		{
		    if (extraTags[i] != null)
		    {
		        _application.putTag(extraTags[i]);
		        if (extraTags[i].getParent() == null)
		            extraTags[i].setParent(this);
		    }
		}
	}

	/**
	 * Sets the parent of this info.
	 *
	 * @param the parent of this info
	 */
	protected void setParent(LyInfo parent)
	{
		if (parent == _parent)
			return;

		if ((_parent != null)
		        && (parent != null))
			throw new IllegalStateException("BF AppEngine : can't set parent " + parent + " for info "
				+ this + ", it has already one parent " + _parent);

		if ((parent instanceof LyInclude)
				&& parent.isLoaded()
				&& !isLoaded())
			getEnvironment().logWarning("BF AppEngine : info " + this
				+ " declared on " + parent + " is not known in this model.");

		_parent = parent;

		if (_parent != null)
		{
		    _parent._sonsCount++;

    		if (_id.startsWith("__")
    		        && _id.endsWith("__")) // Random id
    			setId(null);
		}
	} // end method setParent

	/**
	 * Add this info to its parent project.
	 * @param project
	 */
	protected void addToParentProject()
	{
	    if (_parent instanceof LyProject)
	    {
	        LyProject  parentProject = (LyProject)_parent;

	        if (!parentProject.hasInfo(this))
	            parentProject.addInfo(this);
	    }
	}

	/**
	 * Sets the file for this info.
	 *
	 * @param the file of this info
	 */
	public void setFile(String file)
	{
		_file = file;
	}

	/**
     * Sets the info loaded state.
     */
    protected void setLoaded(boolean loaded)
    {
        _loaded = loaded;
    }
} // end class LyInfo
