/**
 * LyInfoAppFactory.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import org.xml.sax.*;
import leon.app.*;
import leon.info.model.*;
import leon.misc.*;


/**
 * This class builds instances corresponding to XML tags read from the data description file. Almost
 * all objects built during parsing inherit from LyInfo (except very elementary objects).
 *
 * @see     LyProject
 * @see     LyInclude
 * @version $Id: LyInfoAppFactory.java 32430 2011-09-20 16:18:10Z jlsaouli $
 * @author  Lyria
 */
public final class LyInfoAppFactory extends LyInfoFactory
{
	public static final String __VERSION =
		"$Id: LyInfoAppFactory.java 32430 2011-09-20 16:18:10Z jlsaouli $";

	/**
	 * Constructor of the factory for an application. All objects built by this factory will belong
	 * to the given application.
	 *
	 * @param parser      Parser that builts this factory.
	 * @param application Application for which the factory is built.
	 */
	public LyInfoAppFactory(LyInfoParser parser, LyApplication application)
	{
		super(parser, application);
	}

	/**
	 * Builds node corresponding to an element read from the XML data description file. The result
	 * associated to the built node depends on the node tag. This method is invoked by SAXParser at
	 * the beginning of each new tag.
	 *
	 * @param     tag  Current tag.
	 * @param     node Current node.
	 * @return    The object built corresponding to the given tag and attributes.
	 * @exception SAXException if the result could not have been built.
	 */
	public Object buildElement(LyInfoNode node, String tag)
		throws SAXException
	{
		// Init result
		Object result = null;

		// Get some common information
		String id = node.getId();
		String type = node.getString(LyInfoTokens.ATT_TYPE);
		String templateId = node.getString(LyInfoTokens.ATT_MODEL);
		LyInfo template = getInfo(templateId);

		// Build the appropriate result corresponding to each node.
		if (tag == LyInfoTokens.TAG_INCLUDE)
		{
			String applicationId = node.getString(LyInfoTokens.ATT_APP_ID);
			LyInfo info = _application.getInfo((id == null) ? ("_Include" + applicationId) : id);

			if (info instanceof LyInclude)
				result = info;
			else if (info != null)
				throw new SAXException("BF AppEngine : invalid include \"" + id
					+ "\", an info with the same id was found but with a different type.");
			else
			{
				String  xmlFile = node.getString(LyInfoTokens.ATT_XML_FILE);
				String  version = node.getString(LyInfoTokens.ATT__VERSION);
				boolean load = node.getBoolean(LyInfoTokens.ATT_LOAD);

				result = new LyInclude(applicationId, xmlFile, version, load);
			}
		}
		else if (tag == LyInfoTokens.TAG_INFO)
		{
			LyInfo info = _application.getInfo(id);

			if (info == null)
			{
				LyInfo parent = (LyInfo)node.getParent().getResult();

				if ((parent instanceof LyInclude)
						&& ((LyInclude)parent).isLoaded())
				{
					LyEnvironment env = _application.getEnvironment();

					env.logWarning("BF AppEngine : info with " + id
						+ " was not found in the specified include file " + parent);
					type = null;
				}
			}

			if (type == null)
				result = null;
			else if (type.equals("CLASS"))
			{
				if (info instanceof LyClassInfo)
					result = info;
				else if (info == null)
					result = new LyClassInfo(id);
			}
			else if (type.equals("COMPOUND_ACTION"))
			{
				if (info instanceof LyCompoundAction)
					result = info;
				else if (info == null)
					result = new LyCompoundAction(id);
			}
			else if (type.equals("TAB_ACTION"))
			{
				if (info instanceof LyTabAction)
					result = info;
				else if (info == null)
					result = new LyTabAction(id);
			}
			else if (type.equals("ACTION"))
			{
				if (info instanceof LyAction)
					result = info;
				else if (info == null)
					result = new LyAction(id);
			}
			else if (type.equals("ARRAY"))
			{
				if (info instanceof LyArrayFieldInfo)
					result = info;
				else if (info == null)
					result = new LyArrayFieldInfo(id);
			}
			else if (type.equals("CHOICE"))
			{
				if (info instanceof LyChoiceFieldInfo)
					result = info;
				else if (info == null)
					result = new LyChoiceFieldInfo(id);
			}
			else if (type.equals("DATE"))
			{
				if (info instanceof LyDateFieldInfo)
					result = info;
				else if (info == null)
					result = new LyDateFieldInfo(id);
			}
			else if (type.equals("FILE"))
			{
				if (info instanceof LyFileFieldInfo)
					result = info;
				else if (info == null)
					result = new LyFileFieldInfo(id);
			}
			else if (type.equals("NUMBER"))
			{
				if (info instanceof LyNumberFieldInfo)
					result = info;
				else if (info == null)
					result = new LyNumberFieldInfo(id);
			}
			else if (type.equals("RELATION"))
			{
				if (info instanceof LyRelationFieldInfo)
					result = info;
				else if (info == null)
					result = new LyRelationFieldInfo(id);
			}
			else if (type.equals("STRUCT"))
			{
				if (info instanceof LyStructFieldInfo)
					result = info;
				else if (info == null)
					result = new LyStructFieldInfo(id);
			}
			else if (type.equals("TEXT"))
			{
				if (info instanceof LyTextFieldInfo)
					result = info;
				else if (info == null)
					result = new LyTextFieldInfo(id);
			}
			else if (type.equals("FILTER"))
			{
				if (info instanceof LyFilter)
					result = info;
				else if (info == null)
					result = new LyFilter(id);
			}
			else if (type.equals("CHARTYPE"))
			{
				if (info instanceof LyInfoCharType)
					result = info;
				else if (info == null)
					result = new LyInfoCharType(id);
			}
			else if (type.equals("CONTROL"))
			{
				if (info instanceof LyControl)
					result = info;
				else if (info == null)
					result = new LyControl(id);
			}
			else if (type.equals("REGEX"))
			{
				if (info instanceof LyRegExp)
					result = info;
				else if (info == null)
					result = new LyRegExp(id);
			}
			else if (type.equals("UNIT"))
			{
				if (info instanceof LyUnit)
					result = info;
				else if (info == null)
					result = new LyUnit(id);
			}
			else if (type.equals("UNITS"))
			{
				if (info instanceof LyUnits)
					result = info;
				else if (info == null)
					result = new LyUnits(id);
			}

			if ((result == null) && (info != null))
				throw new SAXException("BF AppEngine : invalid extern info \"" + id
					+ "\" (type " + type
					+ "), an info with the same id was found but with a different type.");
		}
		else if (tag == LyInfoTokens.TAG_PROJECT)
		{
			String    version = node.getString(LyInfoTokens.ATT__VERSION);
			String    appBehavior = node.getString(LyInfoTokens.ATT_APP_BEHAVIOR);
			String    sessionBehavior = node.getString(LyInfoTokens.ATT_SESSION_BEHAVIOR);
			String    name = node.getString(LyInfoTokens.ATT_NAME);
			String    tip = node.getString(LyInfoTokens.ATT_TIP);
			String    prefix = node.getString(LyInfoTokens.ATT_PREFIX);

			LyProject project = new LyProject(id, (LyProject)template, version, appBehavior,
					sessionBehavior);

			project.setName(name);
			project.setTip(tip);
			project.setFilePrefix(prefix);
			result = project;
		}
		else if (tag == LyInfoTokens.TAG_BIND_MODEL)
		{
			// Build the model
			String xmlFile = node.getString(LyInfoTokens.ATT_XML_FILE);
			String version = node.getString(LyInfoTokens.ATT__VERSION);

			result = new LyModel(LyModel.TYPE_BIND, xmlFile, version, null);

			// Parse bind model
			LyInfoParser parser = new LyInfoParser(_application, xmlFile, LyInfoParser.Type.BIND,
					_parser.isValidating());

			_parser.addResults(parser);
		}
		else if (tag == LyInfoTokens.TAG_DATA_MODEL)
		{
			// Build the model
			String xmlFile = node.getString(LyInfoTokens.ATT_XML_FILE);
			String version = node.getString(LyInfoTokens.ATT__VERSION);
			String schema = node.getString(LyInfoTokens.ATT_SCHEMA);

			result = new LyModel(LyModel.TYPE_DATA, xmlFile, version, schema);

			// Parse data model
			LyInfoParser parser = new LyInfoParser(_application, xmlFile, LyInfoParser.Type.DATA,
					_parser.isValidating());

			_parser.addResults(parser);
		}
		else if (tag == LyInfoTokens.TAG_GUI_MODEL)
		{
			// Build the model
			String xmlFile = node.getString(LyInfoTokens.ATT_XML_FILE);
			String version = node.getString(LyInfoTokens.ATT__VERSION);

			result = new LyModel(LyModel.TYPE_GUI, xmlFile, version, null);

			// Parse gui model
			LyInfoParser parser = new LyInfoParser(_application, xmlFile, LyInfoParser.Type.GUI,
					_parser.isValidating());

			_parser.addResults(parser);
		}
		else
			result = super.buildElement(node, tag);

		return result;
	} // end method buildElement

	/**
	 * Completes parent result with the result of the current node.
	 *
	 * @param     parentResult Result of the parent node.
	 * @param     node         Current node.
	 * @param     result       Result of current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeParentResult(Object parentResult, LyInfoNode node, Object result)
		throws SAXException
	{
		if (parentResult instanceof LyModel)
			completeModel((LyModel)parentResult, node);
		else
			super.completeParentResult(parentResult, node, result);
	}

	/**
	 * Completes an objectInfo with the result of the current node.
	 *
	 * @param     objectInfo ObjectInfo to be completed.
	 * @param     node       Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeObjectInfo(LyObjectInfo objectInfo, LyInfoNode node)
		throws SAXException
	{
		if (objectInfo instanceof LyProject)
			completeProject((LyProject)objectInfo, node);
		else
			super.completeObjectInfo(objectInfo, node);
	}

	/**
	 * Completes a project with the result of the current node.
	 *
	 * @param project Project to be completed.
	 * @param node    Current node.
	 */
	void completeProject(LyProject project, LyInfoNode node)
	{
		String tag = node.getTag();

		if (tag == LyInfoTokens.TAG_INFO)
			return;

		// Add the current result to project by using the right method
		Object result = node.getResult();

		if (result instanceof LyProject)
			project.addProject((LyProject)result);
		else if (result instanceof LyModel)
			project.addModel((LyModel)result);
		else if (result instanceof LyInclude)
			project.addInfo((LyInfo)result);
	}

	/**
	 * Completes a model with the result of the current node.
	 *
	 * @param model Model to be completed.
	 * @param node  Current node.
	 */
	void completeModel(LyModel model, LyInfoNode node)
	{
	}

	/**
	 * Returns the model type of this factory.
	 */
	protected LyTag.Model getModelType()
	{
		return LyTag.Model.APP;
	}

	void endParse(LyInfoNode node)
		throws SAXException
	{
		super.endParse(node);

		if (node != _rootNode)
			return;

		// Complete parts that have not been fully initialized
		LyInfoList<LyInfo> results = _parser.getResults();
		int				   i, n = results.getSize();
		LyInfo			   info;
		LyObjectInfo	   objectInfo;
		int				   initParts, remainingParts;

		for (i = 0; i < n; i++)
		{
			info = results.getInfo(i);
			if (info instanceof LyObjectInfo)
			{
				objectInfo = (LyObjectInfo)info;
				if (objectInfo.getTemplate() != null)
				{
					initParts = objectInfo._initParts;
					if (initParts > 0)
					{
						remainingParts = LyObjectInfo.PART_ALL & ~initParts;
						if (remainingParts > 0)
						{
							objectInfo.initFromTemplate(remainingParts);
							objectInfo.check(null, remainingParts);
						}
					}
				}
			}
		}
	} // end method endParse
} // end class LyInfoAppFactory
