/**
 * LyInfoBindFactory.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;

import leon.app.*;
import leon.info.binding.*;
import leon.info.daemon.*;
import leon.info.location.*;
import leon.info.notifier.*;
import leon.misc.*;

import org.xml.sax.*;


/**
 * This class builds instances corresponding to XML tags read from the data description file. Almost
 * all objects built during parsing inherit from LyInfo (except very elementary objects).
 *
 * @version $Id: LyInfoBindFactory.java 40928 2012-09-25 17:14:19Z jlsaouli $
 * @author  Lyria
 */
public final class LyInfoBindFactory extends LyInfoFactory
{
	public static final String		  __VERSION =
		"$Id: LyInfoBindFactory.java 40928 2012-09-25 17:14:19Z jlsaouli $";

	// Keeping information during validation stage (check case sensitivity of bindings)
	Hashtable<String, LyClassBinding> _bindingsValidationTable;

	/**
	 * Constructor of the factory for an application. All objects built by this factory will belong
	 * to the given application.
	 *
	 * @param parser      Parser that builts this factory.
	 * @param application Application for which the factory is built.
	 */
	public LyInfoBindFactory(LyInfoParser parser, LyApplication application)
	{
		super(parser, application);
	}

	/**
	 * Builds node corresponding to an element read from the XML data description file. The result
	 * associated to the built node depends on the node tag. This method is invoked by SAXParser at
	 * the beginning of each new tag.
	 *
	 * @param     tag  Current tag.
	 * @param     node Current node.
	 * @return    The object built corresponding to the given tag and attributes.
	 * @exception SAXException if the result could not have been built.
	 */
	@Override
  public Object buildElement(LyInfoNode node, String tag)
		throws SAXException
	{
		// Init result
		Object result = null;

		// Get some common information
		String id = node.getId();
		String type = node.getString(LyInfoTokens.ATT_TYPE);

		if (tag == LyInfoTokens.TAG_ARRAY)
		{
			result = getInfo(id);
			if (!(result instanceof LyArrayFieldInfo))
				throw new SAXException("BF AppEngine : info " + id + " is not an array field info");
		}
		else if (tag == LyInfoTokens.TAG_ARRAY_DAEMON)
			result = new LyArrayDaemon();
		else if (tag == LyInfoTokens.TAG_C_BIND)
		{
			String value = node.getString(LyInfoTokens.ATT_VALUE);
			String location = node.getString(LyInfoTokens.ATT_LOCATION);

		    if (location == null)
		        location = _application.getDefaultLocationId();

			result = new LyClassBinding(value, location, null);
		}
		else if (tag == LyInfoTokens.TAG_C_BINDS)
		{
			boolean strict = node.getBoolean(LyInfoTokens.ATT_STRICT);
			
			result = new LyUnionBinding(strict);
		}
		else if (tag == LyInfoTokens.TAG_C_BIND_INFO)
		{
			String value = node.getString(LyInfoTokens.ATT_IDREF);

			result = new LyClassInfoBinding(value);
		}
		else if (tag == LyInfoTokens.TAG_C_BIND_LDAP)
		{
			String value = node.getString(LyInfoTokens.ATT_VALUE);
			String location = node.getString(LyInfoTokens.ATT_LOCATION);
			String fdnFormat = node.getString(LyInfoTokens.ATT_FDN_FORMAT);

			if (location == null)
				location = _application.getDefaultLocationId();

			result = new LyLdapClassBinding(value, location, null);
			((LyLdapClassBinding)result).setFdnFormat(fdnFormat);
		}
		else if (tag == LyInfoTokens.TAG_CACHE)
		{
            String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
            String userData = node.getString(LyInfoTokens.ATT_USER_DATA);

            if (javaName != null)
                result = new LyInfoCache(javaName, userData);
            else
                result = new LyInfoCache(LyInfoTokens.getClassInfoCachePolicy(type), userData);
		}
		else if (tag == LyInfoTokens.TAG_CHOICE)
		{
			result = getInfo(id);
			if (result == null)
			{
				BitSet marks = new BitSet();

				result = new LyTextFieldInfo(id);
				marks.set(LyFieldInfo.MARK_HIDDEN);
				marks.set(LyFieldInfo.MARK_OPTIONAL);
				((LyFieldInfo)result).setMarks(marks);
			}
			else if (!(result instanceof LyChoiceFieldInfo))
				throw new SAXException("BF AppEngine : info " + id + " is not a choice field info");
		}
		else if (tag == LyInfoTokens.TAG_CLASS)
		{
			result = getInfo(id);
			if (!(result instanceof LyClassInfo))
				throw new SAXException("BF AppEngine : info " + id + " is not a class info");

			boolean local = node.getBoolean(LyInfoTokens.classInfoMarks[LyClassInfo.MARK_LOCAL]);
			boolean load = node.getBoolean(LyInfoTokens.classInfoMarks[LyClassInfo.MARK_LOAD]);

			if (local || load)
			{
				LyClassInfo classInfo = (LyClassInfo)result;
				BitSet	    marks = classInfo.getMarks();

				if (marks == null)
					marks = new BitSet();
				if (local)
					marks.set(LyClassInfo.MARK_LOCAL);
				if (load)
					marks.set(LyClassInfo.MARK_LOAD);
				classInfo.setMarks(marks);
			}
		}
		else if (tag == LyInfoTokens.TAG_DATE)
		{
			result = getInfo(id);
			if (result == null)
			{
				BitSet marks = new BitSet();

				result = new LyDateFieldInfo(id);
				marks.set(LyFieldInfo.MARK_HIDDEN);
				marks.set(LyFieldInfo.MARK_OPTIONAL);
				((LyFieldInfo)result).setMarks(marks);
			}
			else if (!(result instanceof LyDateFieldInfo))
				throw new SAXException("BF AppEngine : info " + id + " is not a date field info");
		}
		else if (tag == LyInfoTokens.TAG_F_BIND)
		{
			String  value = node.getString(LyInfoTokens.ATT_VALUE);
			String  def = node.getString(LyInfoTokens.ATT_DEFAULT);
			String  cbind = node.getString(LyInfoTokens.ATT_CBIND);
			boolean isId = node.getBoolean(LyInfoTokens.ATT_ID);
			String  javaType = node.getString(LyInfoTokens.ATT_JAVA_TYPE);
			String  javaEncoder = node.getString(LyInfoTokens.ATT_JAVA_ENCODER);
			String  peerType = node.getString(LyInfoTokens.ATT_PEER_TYPE);

			result = new LyFieldBinding(value, def, cbind, isId, javaType, javaEncoder, peerType);
		}
		else if (tag == LyInfoTokens.TAG_F_BINDS)
			result = new LyFieldsBinding();
		else if (tag == LyInfoTokens.TAG_F_BIND_INFO)
		{
			String value = node.getString(LyInfoTokens.ATT_IDREF);

			result = new LyFieldInfoBinding(value);
		}
		else if (tag == LyInfoTokens.TAG_F_BIND_LDAP)
		{
			String  value = node.getString(LyInfoTokens.ATT_VALUE);
			boolean parent = node.getBoolean(LyInfoTokens.ATT_PARENT);
      		boolean binary = node.getBoolean(LyInfoTokens.ATT_BINARY);
			result = new LyLdapFieldBinding(value, parent,binary);
		}
		else if (tag == LyInfoTokens.TAG_FIELD)
		{
			result = getInfo(id);
			if (result == null)
			{
				BitSet marks = new BitSet();

				result = new LyTextFieldInfo(id);
				marks.set(LyFieldInfo.MARK_HIDDEN);
				marks.set(LyFieldInfo.MARK_OPTIONAL);
				((LyFieldInfo)result).setMarks(marks);
			}
			else if (!(result instanceof LyFieldInfo))
				throw new SAXException("BF AppEngine : info " + id + " is not a text field info");
		}
		else if (tag == LyInfoTokens.TAG_FILE)
		{
			result = getInfo(id);
			if (result == null)
			{
				BitSet marks = new BitSet();

				result = new LyTextFieldInfo(id);
				marks.set(LyFieldInfo.MARK_HIDDEN);
				marks.set(LyFieldInfo.MARK_OPTIONAL);
				((LyFieldInfo)result).setMarks(marks);
			}
			else if (!(result instanceof LyFileFieldInfo))
				throw new SAXException("BF AppEngine : info " + id + " is not a file field info");
		}
		else if (tag == LyInfoTokens.TAG_IMPORT)
			result = new LyFieldInfoImport();
		else if (tag == LyInfoTokens.TAG_JOIN)
			result = new LyJoinDaemon();
		else if (tag == LyInfoTokens.TAG_JOIN_DAEMON)
			result = new LyJoinDaemon();
		else if (tag == LyInfoTokens.TAG_L_FILE)
		{
			LyLocation location = (LyLocation)_application.getInfo(id);

			if ((location != null) && (location.getClass() == LyFileLocation.class))
				return null;

			String url = node.getString(LyInfoTokens.ATT_URL);
			String valueSep = node.getString(LyInfoTokens.ATT_VALUE_SEP);
			String fieldSep = node.getString(LyInfoTokens.ATT_FIELD_SEP);

			if (id == null)
				id = _application.getDefaultLocationId();

			location = new LyFileLocation(id, url, LyInfoTokens.getFileLocationType(type),
				fieldSep, valueSep);
			result = location;
		}
		else if (tag == LyInfoTokens.TAG_L_LDAP)
		{
			LyLocation location = (LyLocation)_application.getInfo(id);

			if ((location != null) && (location.getClass() == LyLdapLocation.class))
				return null;

			String  url = node.getString(LyInfoTokens.ATT_URL);
			String  baseObjectName = node.getString(LyInfoTokens.ATT_BASE_OBJECT_NAME);
			String  user = node.getString(LyInfoTokens.ATT_USER);
			String  password = node.getString(LyInfoTokens.ATT_PASSWORD);
			String listen = node.getString(LyInfoTokens.ATT_LISTEN);

			if (id == null)
				id = _application.getDefaultLocationId();

			location = new LyLdapLocation(id, url, baseObjectName, user, password, listen);
			result = location;
		}
		else if (tag == LyInfoTokens.TAG_L_SQL)
		{
			LyLocation location = (LyLocation)_application.getInfo(id);

			if ((location != null) && (location.getClass() == LySqlLocation.class))
				return null;

			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			String  url = node.getString(LyInfoTokens.ATT_URL);
			String  driver = node.getString(LyInfoTokens.ATT_DRIVER);
			String  user = node.getString(LyInfoTokens.ATT_USER);
			String  password = node.getString(LyInfoTokens.ATT_PASSWORD);
			String  dbName = node.getString(LyInfoTokens.ATT_DBNAME);
			String  nbConnectionStr = node.getString(LyInfoTokens.ATT_NB_CONNECTION);
			String  nbConnection = "1";
			String  valueSep = node.getString(LyInfoTokens.ATT_VALUE_SEP);
			String trim = node.getString(LyInfoTokens.ATT_TRIM);
			String	metadata = node.getString(LyInfoTokens.ATT_METADATA);

			if (nbConnectionStr != null)
			{
				LyEnvironment env = _application.getEnvironment();
				String		  str = env.getEnv(nbConnectionStr);

				if (str != null)
					nbConnection = str;
				else
					nbConnection = node.getString(LyInfoTokens.ATT_NB_CONNECTION);
			}

			if (id == null)
				id = _application.getDefaultLocationId();

			location = new LySqlLocation(id, javaName, url, driver, user, password, dbName,
					nbConnection, valueSep, trim, metadata);
			result = location;
		}
		else if (tag == LyInfoTokens.TAG_L_GEN)
		{
			LyLocation location = (LyLocation)_application.getInfo(id);

			if ((location != null) && (location.getClass() == LyGenLocation.class))
				return null;

			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			boolean persistent = node.getBoolean(LyInfoTokens.ATT_PERSISTENT);

			location = new LyGenLocation(id, javaName, persistent);
			result = location;
		}
		else if (tag == LyInfoTokens.TAG_NOTIFIER)
		{
			LyNotifier notifier = (LyNotifier)_application.getInfo(id);

			if ((notifier != null) && (notifier.getClass() == LyNotifier.class))
				return notifier;

			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			String properties = node.getString(LyInfoTokens.ATT_PROPERTIES);

			if (id == null)
				id = _application.getId();

			result = new LyNotifier(id, javaName, properties);
		}
		else if (tag == LyInfoTokens.TAG_NUMBER)
		{
			result = getInfo(id);
			if (result == null)
			{
				BitSet marks = new BitSet();

				result = new LyNumberFieldInfo(id);
				marks.set(LyFieldInfo.MARK_HIDDEN);
				marks.set(LyFieldInfo.MARK_OPTIONAL);
				((LyFieldInfo)result).setMarks(marks);
			}
			else if (!(result instanceof LyNumberFieldInfo))
				throw new SAXException("BF AppEngine : info " + id + " is not a number field info");
		}
		else if (tag == LyInfoTokens.TAG_OPTION)
		{
			result = getInfo(id);
			if (!(result instanceof LyChoiceOption))
				throw new SAXException("BF AppEngine : info " + id + " is not a choice option");

			String  value = node.getString(LyInfoTokens.ATT_VALUE);
			boolean unknown = node.getBoolean(LyInfoTokens.ATT_UNKNOWN);

			if (value != null)
				((LyChoiceOption)result).setValue(value);
			((LyChoiceOption)result).setUnknown(unknown);
		}
		else if (tag == LyInfoTokens.TAG_OPTIONS)
		{
			result = getInfo(id);
			if (result != null)
			{
				if (!(result instanceof LyChoiceOptionGroup))
					throw new SAXException("BF AppEngine : info " + id
						+ " is not a choice options group");

				String value = node.getString(LyInfoTokens.ATT_VALUE);

				if (value != null)
					((LyChoiceOptionGroup)result).setValue(value);
			}

			result = new Vector<LyChoiceOption>();
		}
		else if (tag == LyInfoTokens.TAG_PROJECT)
		{
			result = getInfo(id);
			if (!(result instanceof LyProject))
				throw new SAXException("BF AppEngine : info " + id + " is not a project");
		}
		else if (tag == LyInfoTokens.TAG_RELATION)
		{
			result = getInfo(id);
			if (result == null)
			{
				BitSet marks = new BitSet();

				result = new LyTextFieldInfo(id);
				marks.set(LyFieldInfo.MARK_HIDDEN);
				marks.set(LyFieldInfo.MARK_OPTIONAL);
				((LyFieldInfo)result).setMarks(marks);
			}
			else if (!(result instanceof LyRelationFieldInfo))
				throw new SAXException("BF AppEngine : info " + id + " is not a relation field info");
		}
		else if (tag == LyInfoTokens.TAG_REVERSE)
			result = new LyReverseDaemon();
		else if (tag == LyInfoTokens.TAG_STRUCT)
		{
			result = getInfo(id);
			if (!(result instanceof LyStructFieldInfo))
				throw new SAXException("BF AppEngine : info " + id + " is not a struct field info");
		}
		else if (tag == LyInfoTokens.TAG_TEXT)
		{
			result = getInfo(id);
			if (result == null)
			{
				BitSet marks = new BitSet();

				result = new LyTextFieldInfo(id);
				marks.set(LyFieldInfo.MARK_HIDDEN);
				marks.set(LyFieldInfo.MARK_OPTIONAL);
				((LyFieldInfo)result).setMarks(marks);
			}
			else if (!(result instanceof LyTextFieldInfo))
				throw new SAXException("BF AppEngine : info " + id + " is not a text field info");
		}
		else if (tag == LyInfoTokens.TAG_UNION)
		{
			String value = node.getString(LyInfoTokens.ATT_VALUE);
			String value2 = node.getString(LyInfoTokens.ATT_VALUE2);

			result = new LyUnion(value, value2);
		}
		else
			result = super.buildElement(node, tag);

		if (result instanceof LyFieldInfo)
		{
			boolean local = node.getBoolean(LyInfoTokens.fieldInfoMarks[LyFieldInfo.MARK_LOCAL]);
			boolean load = node.getBoolean(LyInfoTokens.fieldInfoMarks[LyFieldInfo.MARK_LOAD]);
			boolean notNull = node.getBoolean(
					LyInfoTokens.fieldInfoMarks[LyFieldInfo.MARK_NOT_NULL]);
			boolean providerValue = node.getBoolean(
					LyInfoTokens.fieldInfoMarks[LyFieldInfo.MARK_PROVIDER_VALUE]);

			if (local || load || notNull || providerValue)
			{
				LyFieldInfo fieldInfo = (LyFieldInfo)result;
				BitSet	    marks = fieldInfo.getMarks();

				if (marks == null)
					marks = new BitSet();
				if (local)
					marks.set(LyFieldInfo.MARK_LOCAL);
				if (load)
					marks.set(LyFieldInfo.MARK_LOAD);
				if (notNull)
					marks.set(LyFieldInfo.MARK_NOT_NULL);
				if (providerValue)
				{
					marks.set(LyFieldInfo.MARK_PROVIDER_VALUE);
					marks.set(LyFieldInfo.MARK_PROVIDER_VALUE_BIND);
				}
				fieldInfo.setMarks(marks);
			}
		}

		return result;
	} // end method buildElement

	/**
	 * Init object info from its template.
	 *
	 * @param objectInfo Object info to initialize.
	 */
	@Override
  protected void initFromTemplate(LyObjectInfo objectInfo)
	{
		if ((objectInfo instanceof LyClassInfo) || (objectInfo instanceof LyFieldInfo))
			objectInfo.initFromTemplate(LyObjectInfo.PART_BIND);
		else
			super.initFromTemplate(objectInfo);
	}

	@Override
  protected void check(LyObjectInfo objectInfo, LyObjectInfo parentInfo)
	{
		if ((objectInfo instanceof LyClassInfo) || (objectInfo instanceof LyFieldInfo))
			objectInfo.check(parentInfo, LyObjectInfo.PART_BIND);
		else
			super.check(objectInfo, parentInfo);
	}

	/**
	 * The purpose of this method is to resolve object references now that all valid object ids are
	 * known.
	 *
	 * @param     tag    Current tag.
	 * @param     result Current result.
	 * @return    New result.
	 * @exception SAXExpection If the object found doesn't belong to the appropriate class.
	 */
	@Override
  Object resolveObjectInfoRef(LyInfoNode node, String tag, Object result)
		throws SAXException
	{
		// Compute idrefs on tags ACTIONS, FIELDS ans OPTIONS
		if ((tag == LyInfoTokens.TAG_FIELDS) || (tag == LyInfoTokens.TAG_OPTIONS))
		{
			if (!(result instanceof Vector<?>)) // Nothing to do
				return null;

			@SuppressWarnings("unchecked")
			Vector<LyInfo> list = (Vector<LyInfo>)result;

			String		   idrefsAtt = node.getString(LyInfoTokens.ATT_IDREFS);
			String		   operation = node.getString(LyInfoTokens.ATT_OPERATION);
			String[]	   idrefs = null;
			int			   i, n = 0;
			LyInfo		   info;

			if (idrefsAtt != null)
			{
				idrefs = LyString.stringToList(idrefsAtt, ' ');
				n = ((idrefs == null) ? 0 : idrefs.length);
			}

			for (i = 0; i < n; i++)
			{
				// Check out class reference
				info = getInfo(idrefs[i]);
				if ((tag == LyInfoTokens.TAG_FIELDS) && !(info instanceof LyFieldInfo))
					raiseError("BF AppEngine : invalid fieldInfo ref : " + idrefs[i]);
				else if ((tag == LyInfoTokens.TAG_OPTIONS) && !(info instanceof LyChoiceOption))
				    raiseError("BF AppEngine : invalid option ref : " + idrefs[i]);

				// Add reference to the result list
				list.addElement(info);

				if (operation != LyInfoTokens.OPER_LIST_REMOVE)
					node.addExtraInfo(info);
			}
		}
		// Fill up join informations
		else if (tag == LyInfoTokens.TAG_JOIN)
		{
			String str;
			LyInfo info;

			// joinClass
			str = node.getString(LyInfoTokens.ATT_JOIN_CLASS);
			info = getInfo(str);
			if (!(info instanceof LyClassInfo))
				throw new SAXException("BF AppEngine : invalid classInfo ref : " + str);
			((LyJoinDaemon)result).setJoinClass((LyClassInfo)info);

			// sourceRelation
			str = node.getString(LyInfoTokens.ATT_SOURCE_RELATION);
			info = getInfo(str);
			if (!(info instanceof LyRelationFieldInfo))
				throw new SAXException("BF AppEngine : invalid relationfieldInfo ref : " + str);
			else if (((LyRelationFieldInfo)info).isMultiple())
				throw new SAXException("BF AppEngine : relation fieldInfo must no be multiple : " + str);
			((LyJoinDaemon)result).setSourceRelation((LyRelationFieldInfo)info);

			// targetRelation
			str = node.getString(LyInfoTokens.ATT_TARGET_RELATION);
			info = getInfo(str);
			if (!(info instanceof LyRelationFieldInfo))
				throw new SAXException("BF AppEngine : invalid relationFieldInfo ref : " + str);
			else if (((LyRelationFieldInfo)info).isMultiple())
				throw new SAXException("BF AppEngine : relation fieldInfo must no be multiple : " + str);
			((LyJoinDaemon)result).setTargetRelation((LyRelationFieldInfo)info);
		}
		// Fill up reverse informations
		else if (tag == LyInfoTokens.TAG_REVERSE)
		{
			String str;
			LyInfo info;

			// reverseRelation
			str = node.getString(LyInfoTokens.ATT_REVERSE_RELATION);
			info = getInfo(str);
			if (!(info instanceof LyRelationFieldInfo))
				throw new SAXException("BF AppEngine : invalid relationfieldInfo ref : " + str);
			((LyReverseDaemon)result).setReverseRelation((LyRelationFieldInfo)info);
		}
		// Set filter on classBinding
		else if ((tag == LyInfoTokens.TAG_C_BIND) || (tag == LyInfoTokens.TAG_C_BIND_LDAP))
		{
			// Is the location ref a LyLocation ?
			String idref = node.getString(LyInfoTokens.ATT_LOCATION);

			if ((idref != null)
				&& !idref.startsWith("LY_"))
			{
				LyInfo location = getInfo(idref);

				if (!(location instanceof LyLocation))
					throw new SAXException("BF AppEngine : invalid location ref : " + idref);
			}
		}
		else if (tag == LyInfoTokens.TAG_C_BIND_INFO)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);
			LyInfo info = getInfo(idref);

			if (!(info instanceof LyClassInfo))
				throw new SAXException("BF AppEngine : invalid classInfo binding : " + idref);
		}
		else if (tag == LyInfoTokens.TAG_F_BIND_INFO)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);
			LyInfo info = getInfo(idref);

			if (!(info instanceof LyFieldInfo))
				throw new SAXException("BF AppEngine : invalid fieldInfo binding : " + idref);
		}
		else if (tag == LyInfoTokens.TAG_IMPORT)
		{
			String idref = node.getString(LyInfoTokens.ATT_RELATION);
			LyInfo info = getInfo(idref);

			if (!(info instanceof LyRelationFieldInfo))
				throw new SAXException("BF AppEngine : invalid relation on import : " + idref);
			((LyFieldInfoImport)result).setRelation((LyRelationFieldInfo)info);

			idref = node.getString(LyInfoTokens.ATT_CLASS);
			info = getInfo(idref);

			if ((info != null) && !(info instanceof LyClassInfo))
				throw new SAXException("BF AppEngine : invalid class on import : " + idref);
			((LyFieldInfoImport)result).setClassInfo((LyClassInfo)info);

			idref = node.getString(LyInfoTokens.ATT_IDREF);
			info = getInfo(idref);

			if (!(info instanceof LyFieldInfo))
				throw new SAXException("BF AppEngine : invalid field on import : " + idref);
			((LyFieldInfoImport)result).setFieldInfo((LyFieldInfo)info);
		}
		// Reference to a notifier
		else if (tag == LyInfoTokens.TAG_NOTIFIER_REF)
		{
			String idref;

			idref = node.getString(LyInfoTokens.ATT_IDREF);
			result = getInfo(idref);

			if (!(result instanceof LyNotifier))
				throw new SAXException("BF AppEngine : invalid notifier ref : " + idref);
		}

		return result;
	} // end method resolveObjectInfoRef

	/**
	 * Completes parent result with the result of the current node.
	 *
	 * @param result Result of current node.
	 * @param node   Current node.
	 * @param result Result of current node.
	 */
	@Override
  void completeParentResult(Object parentResult, LyInfoNode node, Object result)
		throws SAXException
	{
		if (parentResult instanceof LyUnionBinding)
			// If parentResult is an union binding, we complete it with this result
			completeUnionBinding((LyUnionBinding)parentResult, node);
		else if ((parentResult instanceof LyClassBinding)
				|| (parentResult instanceof LyLdapClassBinding))
			// If parentResult is a class binding, we complete it with this result
			completeClassBinding((LyClassBinding)parentResult, node);
		else if (parentResult instanceof LyFieldsBinding)
			// If parentResult is a fields binding, we complete it with this result
			completeFieldsBinding((LyFieldsBinding)parentResult, node);
		else if (parentResult instanceof LyFieldInfoImport)
			// If parentResult is a field info import, we complete it with this result
			completeFieldInfoImport((LyFieldInfoImport)parentResult, node);
		else if (parentResult instanceof LyArrayDaemon)
			// If parentResult is an array daemon, we complete it with this result
			completeArrayDaemon((LyArrayDaemon)parentResult, node);
		else if (parentResult instanceof LyJoinDaemon)
			// If parentResult is a join daemon, we complete it with this result
			completeJoinDaemon((LyJoinDaemon)parentResult, node);
	}

	/**
	 * Completes an info with the result of the current node.
	 *
	 * @param     info Info to be completed.
	 * @param     node Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	@Override
  void completeInfo(LyInfo info, LyInfoNode node)
		throws SAXException
	{
		// Call the appropriate method to complete the info
		if (info instanceof LyLocation)
			completeLocation((LyLocation)info, node);
		else
			super.completeInfo(info, node);
	}

	/**
	 * Completes a project with the result of the current node.
	 *
	 * @param project Project to be completed.
	 * @param node    Current node.
	 */
	@Override
  void completeProject(LyProject project, LyInfoNode node)
	{
		// Add the current result to project by using the right method
		Object result = node.getResult();

		if (result instanceof LyLocation)
			project.addLocation((LyLocation)result);
	    else if (result instanceof LyNotifier)
	        project.addNotifier((LyNotifier)result);
	}

	/**
	 * Completes an fieldInfo with the result of the current node.
	 *
	 * @param     fieldInfo FieldInfo to be completed.
	 * @param     node      Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	@Override
  void completeFieldInfo(LyFieldInfo fieldInfo, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags corresponding to objectInfo attributes
		// and complete the given fieldInfo.
		if (result instanceof LyFieldBinding)
		{
			fieldInfo.setBinding((LyFieldBinding)result);

			boolean notNull = node.getBoolean(
					LyInfoTokens.fieldInfoMarks[LyFieldInfo.MARK_NOT_NULL]);

			if (notNull)
			{
				BitSet marks = fieldInfo.getMarks();

				if (marks == null)
					marks = new BitSet();
				marks.set(LyFieldInfo.MARK_NOT_NULL);
				fieldInfo.setMarks(marks);
			}
		}
		else if (result instanceof LyFieldInfoImport)
			fieldInfo.setImport((LyFieldInfoImport)result);
		else if (result instanceof LyPath)
			fieldInfo.setPath((LyPath)result);

		// For other tags, call the appropriate method to complete the fieldInfo
		else if (fieldInfo instanceof LyArrayFieldInfo)
			completeArrayFieldInfo((LyArrayFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyStructFieldInfo)
			completeStructFieldInfo((LyStructFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyRelationFieldInfo)
			completeRelationFieldInfo((LyRelationFieldInfo)fieldInfo, node);
	} // end method completeFieldInfo

	/**
	 * Completes an array field info with the result of the current node.
	 *
	 * @param     arrayFieldInfo ArrayFieldInfo to be completed.
	 * @param     node           Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeArrayFieldInfo(LyArrayFieldInfo arrayFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to arrayFieldInfo attributes
		// and complete the given arrayFieldInfo.
		if (tag == LyInfoTokens.TAG_ARRAY_DAEMON)
			arrayFieldInfo.setDaemon((LyArrayDaemon)result);
	}

	/**
	 * Completes a relation field info with the result of the current node.
	 *
	 * @param     relationFieldInfo RelationFieldInfo to be completed.
	 * @param     node              Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeRelationFieldInfo(LyRelationFieldInfo relationFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to relationFieldInfo attributes
		// and complete the given relationFieldInfo.
		if (tag == LyInfoTokens.TAG_ROUTES)
		{
			String           operation = node.getString(LyInfoTokens.ATT_OPERATION);
			int              pos = node.getInt(LyInfoTokens.ATT_POSITION);

			@SuppressWarnings("unchecked")
			Vector<LyRoute>  routes = (Vector<LyRoute>)result;

			int              i, n = routes.size();

			if (operation != null)
				operation = operation.intern();

			if (operation == LyInfoTokens.OPER_LIST_RESET)
			{
				relationFieldInfo.resetRoutes();
				pos = 0;
			}

			if ((operation == LyInfoTokens.OPER_LIST_ADD)
					|| (operation == LyInfoTokens.OPER_LIST_RESET))
			{
				for (i = 0; i < n; i++)
					relationFieldInfo.addRoute(routes.elementAt(i),
						((pos < 0) ? -1 : (pos + i)));
			}
			else if (operation == LyInfoTokens.OPER_LIST_REMOVE)
			{
				for (i = 0; i < n; i++)
					relationFieldInfo.removeRoute(routes.elementAt(i));
				node.setResult(null);
			}
		}
		else if ((tag == LyInfoTokens.TAG_JOIN) || (tag == LyInfoTokens.TAG_JOIN_DAEMON)
				|| (tag == LyInfoTokens.TAG_REVERSE))
			relationFieldInfo.setDaemon((LyInfoDaemon)result);
		else if (result instanceof LyClassBinding)
		{
			LyClassInfo classInfo = relationFieldInfo.getClassInfo(0);

			completeClassInfo(classInfo, node);
		}
	} // end method completeRelationFieldInfo

	/**
	 * Completes a choice field info with the result of the current node.
	 *
	 * @param     choiceFieldInfo ChoiceFieldInfo to be completed.
	 * @param     node            Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeChoiceFieldInfo(LyChoiceFieldInfo choiceFieldInfo, LyInfoNode node)
	{
		String tag = node.getTag();
		Object result = node.getResult();

		if (tag == LyInfoTokens.TAG_OPTIONS)
		{
			LyChoiceOptionGroup group = choiceFieldInfo.getChoiceOptionGroup();
			Vector<?>			    items = (Vector<?>)result;
			int				    i, n = items.size();
			Object			    item;

			for (i = 0; i < n; i++)
			{
				item = items.elementAt(i);
				if ((item instanceof LyChoiceItem) && !group.contains((LyChoiceItem)item, false))
					throw new IllegalStateException("BF AppEngine : choice item " + item
						+ " was not declared in choice field info " + choiceFieldInfo
						+ " in your data model");
			}
		}
	}

	/**
	 * Completes a struct field info with the result of the current node.
	 *
	 * @param     relationFieldInfo structFieldInfo to be completed.
	 * @param     node              Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeStructFieldInfo(LyStructFieldInfo structFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String	    tag = node.getTag();
		Object	    result = node.getResult();
		LyClassInfo classInfo = structFieldInfo.getClassInfo(0);

		if (result instanceof LyClassBinding)
			classInfo.setBinding((LyClassBinding)result);
		else if (tag == LyInfoTokens.TAG_CACHE)
			classInfo.setCachePolicy((LyInfoCache)result);
		else if (tag == LyInfoTokens.TAG_FIELDS)
		{
			// Check that all fields really belong to the given classInfo
			Vector<?>			    fields = (Vector<?>)result;
			int					    i, n = fields.size();
			LyFieldInfo			    fieldInfo;
			int					    j, ns;
			LyClassInfo			    subClass;
			LyEnvironment		    env = _application.getEnvironment();
			String				    message;
			LyInfoList<LyClassInfo> subClasses;

			for (i = 0; i < n; i++)
			{
				fieldInfo = (LyFieldInfo)fields.elementAt(i);
				if (classInfo.contains(fieldInfo))
					continue;

				if (classInfo.getSuperClass() != null)
				{
					classInfo.addFieldInfo(fieldInfo, 0);
					message = env.getMessage("LY_UNION_CREATING_FIELD_SON", fieldInfo.getId(),
							classInfo.getId(), classInfo.getSuperClass().getId());
					env.logWarning(message);
				}
				else
				{
					subClasses = classInfo.getSubClasses();

					if ((classInfo.isAbstract()) && (subClasses != null)
							&& (subClasses.getSize() > 0))
					{
						classInfo.addFieldInfo(fieldInfo, 0);
						for (j = 0, ns = subClasses.getSize(); j < ns; j++)
						{
							subClass = subClasses.getInfo(j);
							subClass.addFieldInfo(fieldInfo, 0);
						}

						message = env.getMessage("LY_UNION_CREATING_FIELD_PARENT",
								fieldInfo.getId(), classInfo.getId());
						env.logWarning(message);
					}
					else
					{
						String operation = node.getString(LyInfoTokens.ATT_OPERATION);

						if (operation != LyInfoTokens.OPER_LIST_REMOVE)
						{
							// Check that all fields really belong to the given classInfo
							for (i = 0; i < n; i++)
							{
								fieldInfo = (LyFieldInfo)fields.elementAt(i);
								classInfo.moveFieldInfo(fieldInfo, -1);
							}
						}
					}
				}
			}
		}
	} // end method completeStructFieldInfo

	/**
	 * Completes a class info with the result of the current node.
	 *
	 * @param classInfo ClassInfo to be completed.
	 * @param node      Current node.
	 */
	@Override
  void completeClassInfo(LyClassInfo classInfo, LyInfoNode node)
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to classInfo attributes
		// and complete the given classInfo.
		if (result instanceof LyClassBinding)
			classInfo.setBinding((LyClassBinding)result);
		else if (tag == LyInfoTokens.TAG_CACHE)
			classInfo.setCachePolicy((LyInfoCache)result);
		else if (tag == LyInfoTokens.TAG_FIELDS)
		{
			// Check that all fields really belong to the given classInfo
			Vector<?>			    fields = (Vector<?>)result;
			int					    i, n = fields.size();
			LyFieldInfo			    fieldInfo;
			LyEnvironment		    env = _application.getEnvironment();
			String				    message;
			LyInfoList<LyClassInfo> subClasses;
			LyClassInfo			    subClass;
			int					    j, ns;

			for (i = 0; i < n; i++)
			{
				fieldInfo = (LyFieldInfo)fields.elementAt(i);
				if (classInfo.contains(fieldInfo))
					continue;

				if (classInfo.getSuperClass() != null)
				{
					classInfo.addFieldInfo(fieldInfo, 0);
					message = env.getMessage("LY_UNION_CREATING_FIELD_SON", fieldInfo.getId(),
							classInfo.getId(), classInfo.getSuperClass().getId());
					env.logWarning(message);
				}
				else
				{
					subClasses = classInfo.getSubClasses();

					if ((classInfo.isAbstract()) && (subClasses != null)
							&& (subClasses.getSize() > 0))
					{
						classInfo.addFieldInfo(fieldInfo, 0);
						for (j = 0, ns = subClasses.getSize(); j < ns; j++)
						{
							subClass = subClasses.getInfo(j);
							subClass.addFieldInfo(fieldInfo, 0);
						}

						message = env.getMessage("LY_UNION_CREATING_FIELD_PARENT",
								fieldInfo.getId(), classInfo.getId());
						env.logWarning(message);
					}
					else
					{
						String operation = node.getString(LyInfoTokens.ATT_OPERATION);

						if (operation != LyInfoTokens.OPER_LIST_REMOVE)
						{
							// Check that all fields really belong to the given classInfo
							for (i = 0; i < n; i++)
							{
								fieldInfo = (LyFieldInfo)fields.elementAt(i);
								classInfo.moveFieldInfo(fieldInfo, -1);
							}
						}
					}
				}
			}
		}
	} // end method completeClassInfo

	/**
	 * Completes a generic location with the result of the current node.
	 *
	 * @param     location Location to be completed.
	 * @param     node     Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeLocation(LyLocation location, LyInfoNode node)
	{
		Object result = node.getResult();
		String tag = node.getTag();

		// Check for tags or results corresponding to route attributes
		// and complete the given route.
		if ((location instanceof LyGenLocation) && (tag == LyInfoTokens.TAG_DATA))
			((LyGenLocation)location).putData(node.getString(LyInfoTokens.ATT_NAME), result);
		else if (result instanceof LyNotifier)
			location.setNotifier((LyNotifier)result);
	}

	/**
	 * Completes an union binding with the result of the current node.
	 *
	 * @param     unionBinding Union binding to be completed.
	 * @param     node         Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeUnionBinding(LyUnionBinding unionBinding, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to union binding attributes
		// and complete the given class binding.
		if (result instanceof LyClassBinding)
			unionBinding.addClassBinding((LyClassBinding)result);
		else if (result instanceof LyUnion)
			unionBinding.addUnion((LyUnion)result);;
	}

	/**
	 * Completes a class binding with the result of the current node.
	 *
	 * @param     classBinding Class binding to be completed.
	 * @param     node         Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeClassBinding(LyClassBinding classBinding, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to class binding attributes
		// and complete the given class binding.
		if (result instanceof LyFilter)
			classBinding.setFilter((LyFilter)result);
	}

	/**
	 * Completes a fields binding with the result of the current node.
	 *
	 * @param     fieldsBinding Fields binding to be completed.
	 * @param     node          Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeFieldsBinding(LyFieldsBinding fieldsBinding, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to fields binding attributes
		// and complete the given fields binding.
		if (result instanceof LyFieldBinding)
			fieldsBinding.addFieldBinding((LyFieldBinding)result);
	}

	/**
	 * Completes a field info import with the result of the current node.
	 *
	 * @param     fieldInfoImport Field info import to be completed.
	 * @param     node            Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeFieldInfoImport(LyFieldInfoImport fieldInfoImport, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to field info import attributes
		// and complete the given field info import
		if (result instanceof LyFilter)
			fieldInfoImport.setFilter((LyFilter)result);
	}

	/**
	 * Completes an array daemon with the result of the current node.
	 *
	 * @param     arrayDaemon Array daemon to be completed.
	 * @param     node        Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeArrayDaemon(LyArrayDaemon arrayDaemon, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to field info import attributes
		// and complete the given field info import
		if (result instanceof LyClassBinding)
			arrayDaemon.setClassBinding((LyClassBinding)result);
		else if (result instanceof LyFieldBinding)
			arrayDaemon.addFieldBinding((LyFieldBinding)result);
	}

	/**
	 * Completes a join daemon with the result of the current node.
	 *
	 * @param     joinDaemon Join daemon to be completed.
	 * @param     node       Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeJoinDaemon(LyJoinDaemon joinDaemon, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to field info import attributes
		// and complete the given field info import
		if (result instanceof LyClassBinding)
			joinDaemon.setClassBinding((LyClassBinding)result);
		else if (result instanceof LyFieldBinding)
			joinDaemon.addFieldBinding((LyFieldBinding)result);
	}

	/**
	 * Extra checks for result. Throws an exception if check fails.
	 *
	 * @param result Checked result.
	 */
	@Override
  protected void checkResult(Object result)
	{
		// Check bindings : should not have a different case value
		if ((result instanceof LyClassBinding)
				&& !(result instanceof LyClassInfoBinding))
		{
			String value = ((LyClassBinding)result).getBinding();

			if (value != null)
			{
				String lcValue = value.toLowerCase();

				if (_bindingsValidationTable == null)
					_bindingsValidationTable = new Hashtable<String, LyClassBinding>();

				// Get lower case related binding
				LyClassBinding oldBinding = _bindingsValidationTable.get(lcValue);

				if (oldBinding == null)
					_bindingsValidationTable.put(lcValue, (LyClassBinding)result);
				else if (!oldBinding.getBinding().equals(value))
					throw new IllegalArgumentException("BF AppEngine : The binding '" + value
						+ "' was already defined as '" + oldBinding.getBinding()
						+ "' : bindings must not differ only by case sensitivity.");
			}
		}
	}

	/**
	 * Returns the model type of this factory.
	 */
	@Override
  protected LyTag.Model getModelType()
	{
		return LyTag.Model.BIND;
	}

} // end class LyInfoBindFactory
