/*
 * LyInfoCache.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import leon.data.cache.*;
import leon.info.misc.*;
import leon.misc.*;


/**
 * This class is used to store informations on cache.
 */
public class LyInfoCache extends LyTag
	implements Cloneable
{
	public static final String __VERSION =
		"$Id: LyInfoCache.java 34766 2012-03-16 15:04:22Z jlsaouli $";

    /** Cache policies. */
    public enum CachePolicy
    {
        FULL(LyFullCache.class),
        AUTO(LySimpleCache.class),
        NO(LyNoCache.class);

		Class<?>   _policy;

		CachePolicy(Class<?> policy)
        {
            _policy = policy;
        }

		Class<?> getPolicyClass()
        {
            return _policy;
        }

        String getName()
        {
            return _policy.getName();
        }
    }

    /**
     * Type of cache policy, null means custom.
     */
    protected CachePolicy       _type;

	/**
	 * The name of the java class representing the implementation of the cache policy. This class
	 * must inherit from leon.data.LyCache.
	 */
	protected LyJavaName        _cachePolicyName;

	/** User data associated with the cache policy. */
	protected String		   _userData;

	/** Indicates if this cache policy is the default one. */
	protected boolean          _default;


    /**
     * Default constructor.
     */
    public LyInfoCache()
    {
        this(CachePolicy.FULL, null);
        _default = true;
    }

    /**
     * Constructor.
     */
    public LyInfoCache(String cachePolicyName)
    {
        this(cachePolicyName, null);
    }

	/**
	 * Constructor.
	 */
	public LyInfoCache(String cachePolicyName, String userData)
	{
        _type = null;
		setCachePolicyName(cachePolicyName);
		_userData = userData;
	}

	/**
	 * Constructor.
	 */
	public LyInfoCache(CachePolicy cachePolicy, String userData)
	{
	    if (cachePolicy == null)
	        cachePolicy = CachePolicy.FULL;

	    _type = cachePolicy;
        setCachePolicyName(cachePolicy.getName());
		_userData = userData;
	}

    /**
     * Returns extra tags.
     */
    protected LyTag[] getExtraTags()
    {
        return new LyTag[] { _cachePolicyName };
    }

	/**
	 * Accessor for cache policy.
	 */
	public Class<?> getCachePolicy()
	{
		if (_type != null)
		    return _type.getPolicyClass();

	    if (_cachePolicyName == null)
	        return null;

        // loading classes
        Class<?>   cachePolicy = _cachePolicyName.getJavaClass();

        if (cachePolicy != null)
            return cachePolicy;

        LyEnvironment environment = getEnvironment();

        environment.logError("BF AppEngine : unknown cache policy " + _cachePolicyName
            + ". Setting cache to FULL.");

        _type = CachePolicy.FULL;
		return _type.getPolicyClass();
	}

    /**
     * Accessor for type of cache policy.
     */
    public CachePolicy getType()
    {
        return _type;
    }

	/**
	 * Accessor for cache policy name.
	 */
	public String getCachePolicyName()
	{
		return (_cachePolicyName == null ? null : _cachePolicyName.getJavaName());
	}

	/**
	 * Accessor for user data.
	 */
	public String getUserData()
	{
		return _userData;
	}

	/**
     * Accessor for default property.
     */
    public boolean isDefault()
    {
        return _default;
    }

	/**
	 * Overriden to test on policy and user data.
	 */
	public boolean equals(Object object)
	{
		if (object == null)
			return false;
		
		LyInfoCache cache = (LyInfoCache)object;

		if (cache._cachePolicyName != _cachePolicyName)
			return false;

		String data = cache.getUserData();

		if ((data == null) && (_userData == null))
			return true;

		if ((data != null) && (_userData != null) && (data.equals(_userData)))
			return true;

		return false;
	}

    /**
     * Sets for user data.
     */
    public void setUserData(String userData)
    {
        _userData = userData;
    }

    /**
     * Internal method used to set cache policy name.
     */
    protected void setCachePolicyName(String cachePolicyName)
    {
        _cachePolicyName = (cachePolicyName == null ? null
            : new LyJavaName(_parent, cachePolicyName));
    }

	@Override
	protected void setParent(LyInfo parent)
	{
		super.setParent(parent);
		if (_cachePolicyName != null)
			_cachePolicyName.setParent(_parent);
	}


} // end class LyInfoCache
