/**
 * LyInfoDataFactory.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import org.xml.sax.*;

import java.util.*;

import leon.app.*;
import leon.misc.*;
import leon.parser.*;
import leon.info.misc.*;


/**
 * This class builds instances corresponding to XML tags read from the data description file. Almost
 * all objects built during parsing inherit from LyInfo (except very elementary objects).
 *
 * @see     LyInfoParser
 * @see     LyInfo
 * @see     LyObjectInfo
 * @see     LyFieldInfo
 * @see     LyChoiceFieldInfo
 * @see     LyChoiceOption
 * @see     LyDateFieldInfo
 * @see     LyFileFieldInfo
 * @see     LyNumberFieldInfo
 * @see     LyRelationFieldInfo
 * @see     LyTextFieldInfo
 * @see     LyClassInfo
 * @see     LyProject
 * @see     LyFilter
 * @see     LyFilterExpression
 * @see     LyFilterRelation
 * @see     LySort
 * @see     LySortCriterion
 * @see     LyRoute
 * @see     LyStep
 * @version $Id: LyInfoDataFactory.java 41407 2012-10-30 16:17:49Z bangiole $
 * @author  Lyria
 */
public final class LyInfoDataFactory extends LyInfoFactory
	implements LyXMLElementFactory
{
	public static final String __VERSION =
		"$Id: LyInfoDataFactory.java 41407 2012-10-30 16:17:49Z bangiole $";

	/**
	 * Constructor of the factory for an application. All objects built by this factory will belong
	 * to the given application.
	 *
	 * @param parser      Parser that builts this factory.
	 * @param application Application for which the factory is built.
	 */
	public LyInfoDataFactory(LyInfoParser parser, LyApplication application)
	{
		super(parser, application);
	}

	/**
	 * Builds node corresponding to an element read from the XML data description file. The result
	 * associated to the built node depends on the node tag. This method is invoked by SAXParser at
	 * the beginning of each new tag.
	 *
	 * @param     tag  Current tag.
	 * @param     node Current node.
	 * @return    The object built corresponding to the given tag and attributes.
	 * @exception SAXException if the result could not have been built.
	 */
	public Object buildElement(LyInfoNode node, String tag)
		throws SAXException
	{
		// Init result
		Object result = null;

		// Get some common information
		String id = node.getId();
		String type = node.getString(LyInfoTokens.ATT_TYPE);
		String templateId = node.getString(LyInfoTokens.ATT_MODEL);
		LyInfo template = getInfo(templateId);

		if (tag == LyInfoTokens.TAG_ARRAY)
		{
			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			short  cMin = (short)node.getInt(LyInfoTokens.ATT_CMIN);
			short  cMax = (short)node.getInt(LyInfoTokens.ATT_CMAX, 0);
			String behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			LyInfo fieldInfo = getInfo(id);

			if (fieldInfo == null)
				result = new LyArrayFieldInfo(id, javaName, (LyArrayFieldInfo)template, cMin, cMax);
			else
			{
				if (!(fieldInfo instanceof LyArrayFieldInfo))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not an array field info");
				if (fieldInfo.isLoaded())
					throw new SAXException("BF AppEngine : array field info " + id
						+ " already exists");

				((LyArrayFieldInfo)fieldInfo).construct(javaName, (LyArrayFieldInfo)template, cMin,
					cMax);
				result = fieldInfo;
			}

			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);
		}
		else if (tag == LyInfoTokens.TAG_CHARTYPE)
		{
			String value = node.getString(LyInfoTokens.ATT_VALUE);
			LyInfo charType = getInfo(id);

			if (charType == null)
				result = new LyInfoCharType(id, value);
			else
			{
				if (!(charType instanceof LyInfoCharType))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not a char type");
				if (charType.isLoaded())
					throw new SAXException("BF AppEngine : char type " + id
						+ " already exists");

				((LyInfoCharType)charType).construct(value);
				result = charType;
			}
		}
		else if (tag == LyInfoTokens.TAG_CHOICE)
		{
			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			boolean multiple = node.getBoolean(LyInfoTokens.ATT_MULTIPLE);
			String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			LyInfo  fieldInfo = getInfo(id);

			if (fieldInfo == null)
				result = new LyChoiceFieldInfo(id, javaName, (LyChoiceFieldInfo)template, multiple,
						LyChoiceFieldInfo.SortBy.INDEX, LyInfoTokens.getChoiceFieldInfoType(type),
						false);
			else
			{
				if (!(fieldInfo instanceof LyChoiceFieldInfo))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not a choice field info");
				if (fieldInfo.isLoaded())
					throw new SAXException("BF AppEngine : choice field info " + id
						+ " already exists");

				((LyChoiceFieldInfo)fieldInfo).construct(javaName, (LyChoiceFieldInfo)template,
					multiple, LyChoiceFieldInfo.SortBy.INDEX,
					LyInfoTokens.getChoiceFieldInfoType(type), false);
				result = fieldInfo;
			}

			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);
		}
		else if (tag == LyInfoTokens.TAG_CLASS)
		{
			String superClassId = node.getString(LyInfoTokens.ATT_EXTENDS);
			LyInfo superClass = getInfo(superClassId);

			// Forbid forward reference on super-class
			if ((superClassId != null)
					&& (superClassId.length() > 0)
					&& (superClass == null))
				throw new IllegalStateException(
					"BF AppEngine : Forward reference to super-class is not supported. Class '" + id
					+ "' with super-class : '" + superClassId + "'.");

			if (id != null)
			{
				String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
				boolean abstractClass = node.getBoolean(LyInfoTokens.ATT_ABSTRACT);
				boolean virtual = node.getBoolean(LyInfoTokens.ATT_VIRTUAL);
				String  userData = node.getString(LyInfoTokens.ATT_USER_DATA);
				LyInfo  classInfo = getInfo(id);

				if (virtual && (superClass == null))
					throw new IllegalStateException("BF AppEngine : declaration of the virtual class "
						+ classInfo + " requires a super-class");

				if (classInfo == null)
					result = new LyClassInfo(id, (LyClassInfo)superClass, (LyClassInfo)template,
							javaName, null, abstractClass, virtual, userData);
				else
				{
					if (!(classInfo instanceof LyClassInfo))
						throw new SAXException("BF AppEngine : extern info " + id
							+ " is not a class info");
					if (classInfo.isLoaded())
						throw new SAXException("BF AppEngine : class info " + id
							+ " already exists");

					((LyClassInfo)classInfo).construct((LyClassInfo)superClass,
						(LyClassInfo)template, javaName, null, abstractClass, virtual, userData);
					result = classInfo;
				}
			}
			else if (superClass != null)
				result = superClass;
			else
				throw new IllegalStateException(
					"BF AppEngine : class declaration requires either an 'id' to create a new class or an 'extends' to extend an existing class");
		}
		else if (tag == LyInfoTokens.TAG_CLASSES)
			result = new Vector<LyClassInfo>();
		else if (tag == LyInfoTokens.TAG_CONTROL)
		{
			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			int    priority = node.getInt(LyInfoTokens.ATT_PRIORITY);
			String level = node.getString(LyInfoTokens.ATT_LEVEL);
			LyInfo control = getInfo(id);

			if (control == null)
				result = new LyControl(id, javaName, LyInfoTokens.getControlLevel(level), priority,
				    LyTag.Model.DATA);
			else
			{
				if (!(control instanceof LyControl))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not a control");
				if (control.isLoaded())
					throw new SAXException("BF AppEngine : control " + id
						+ " already exists");

				((LyControl)control).construct(javaName, LyInfoTokens.getControlLevel(level),
					priority, LyTag.Model.DATA);
				result = control;
			}

            String message = node.getString(LyInfoTokens.ATT_MESSAGE);

            ((LyControl)result).setMessage(message);
		}
		else if (tag == LyInfoTokens.TAG_DATE)
		{
			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			String behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			boolean withTimeZone = node.getBoolean(LyInfoTokens.ATT_WITHTIMEZONE);
			LyInfo fieldInfo = getInfo(id);


			if (fieldInfo == null)
				result = new LyDateFieldInfo(id, javaName, (LyDateFieldInfo)template,
						LyInfoTokens.getDateFieldInfoType(type));
			else
			{
				if (!(fieldInfo instanceof LyDateFieldInfo))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not a date field info");
				if (fieldInfo.isLoaded())
					throw new SAXException("BF AppEngine : date field info " + id
						+ " already exists");

				((LyDateFieldInfo)fieldInfo).construct(javaName, (LyDateFieldInfo)template,
					LyInfoTokens.getDateFieldInfoType(type), withTimeZone);
				result = fieldInfo;
			}

			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);
		}
		else if (tag == LyInfoTokens.TAG_ENCODING)
			result = node.getString(LyInfoTokens.ATT_JAVA_NAME);
		else if ((tag == LyInfoTokens.TAG_C_MARKS)
				|| (tag == LyInfoTokens.TAG_F_MARKS))
			result = buildMarks(tag, node);
		else if (tag == LyInfoTokens.TAG_FIELD)
		{
			template = getInfo(type);
			if (template == null)
				throw new SAXException("BF AppEngine : type " + type
					+ " is not a valid field type for field " + id);

			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);

			template.load();

			if (template instanceof LyArrayFieldInfo)
			{
				LyArrayFieldInfo model = (LyArrayFieldInfo)template;

				result = new LyArrayFieldInfo(id, javaName, model, model.getMinCardinality(),
						model.getMaxCardinality());
			}
			else if (template instanceof LyChoiceFieldInfo)
			{
				LyChoiceFieldInfo model = (LyChoiceFieldInfo)template;

				result = new LyChoiceFieldInfo(id, javaName, model, model.isMultiple(),
						model.getSortBy(), model.getType(), model.isNot());
			}
			else if (template instanceof LyDateFieldInfo)
			{
				LyDateFieldInfo model = (LyDateFieldInfo)template;

				result = new LyDateFieldInfo(id, javaName, model, model.getType());
			}
			else if (template instanceof LyFileFieldInfo)
			{
				LyFileFieldInfo model = (LyFileFieldInfo)template;

				result = new LyFileFieldInfo(id, javaName, model, model.getMimeTypes(),
						model.getAttachmentDirectory(), model.getAttachmentHandler(),
						model.getDirectory(), model.getFilter());
			}
			else if (template instanceof LyNumberFieldInfo)
			{
				LyNumberFieldInfo model = (LyNumberFieldInfo)template;

				result = new LyNumberFieldInfo(id, javaName, model, model.getType(),
						model.isModulo());
			}
			else if (template instanceof LyStructFieldInfo)
			{
				LyStructFieldInfo struct, model = (LyStructFieldInfo)template;

				struct = new LyStructFieldInfo(id);
				struct.setFieldTag(true);
				struct.construct(javaName, null, model, model.getClassName(), model.isAbstract());
				result = struct;
			}
			else if (template instanceof LyRelationFieldInfo)
			{
				LyRelationFieldInfo model = (LyRelationFieldInfo)template;

				result = new LyRelationFieldInfo(id, javaName, null, model, model.isMultiple(),
						model.isOrdered(), model.getDirection(), model.getType(),
						model.getUserData(), model.ignoresContext(),
						model.isToolbarDisabled(), model.getShowAs());
			}
			else if (template instanceof LyTextFieldInfo)
			{
				LyTextFieldInfo model = (LyTextFieldInfo)template;

				result = new LyTextFieldInfo(id, javaName, model, model.isMultiple());
			}
			else
				throw new SAXException("BF AppEngine : type " + type
					+ " is not a valid field type for field " + id
					+ " (not a field info)");

			String behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);

			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);

			boolean calculated = node.getBoolean(LyInfoTokens.ATT_CALCULATED);

			((LyFieldInfo)result).setCalculated(calculated);
			((LyFieldInfo)result).setFieldTag(true);
		}
		else if (tag == LyInfoTokens.TAG_FILE)
		{
			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			String mimeTypes = node.getString(LyInfoTokens.ATT_MIME_TYPES);
			String directory = node.getString(LyInfoTokens.ATT_DIRECTORY);
			String filter = node.getString(LyInfoTokens.ATT_FILTER);
			String behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			LyInfo fieldInfo = getInfo(id);

			if (fieldInfo == null)
				result = new LyFileFieldInfo(id, javaName, (LyFileFieldInfo)template, mimeTypes,
						null, null, directory, filter);
			else
			{
				if (!(fieldInfo instanceof LyFileFieldInfo))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not a file field info");
				if (fieldInfo.isLoaded())
					throw new SAXException("BF AppEngine : file field info " + id
						+ " already exists");

				((LyFileFieldInfo)fieldInfo).construct(javaName, (LyFileFieldInfo)template,
					mimeTypes, null, null, directory, filter);
				result = fieldInfo;
			}

			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);
		}
		else if (tag == LyInfoTokens.TAG_MARK)
			result = node.getString(LyInfoTokens.ATT_NAME);
		else if (tag == LyInfoTokens.TAG_NUMBER)
		{
			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			boolean modulo = node.getBoolean(LyInfoTokens.ATT_MODULO);
			LyInfo  fieldInfo = getInfo(id);

			if (fieldInfo == null)
				result = new LyNumberFieldInfo(id, javaName, (LyNumberFieldInfo)template,
						LyInfoTokens.getNumberFieldInfoType(type), modulo);
			else
			{
				if (!(fieldInfo instanceof LyNumberFieldInfo))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not a number field info");
				if (fieldInfo.isLoaded())
					throw new SAXException("BF AppEngine : number field info " + id
						+ " already exists");

				((LyNumberFieldInfo)fieldInfo).construct(javaName, (LyNumberFieldInfo)template,
					LyInfoTokens.getNumberFieldInfoType(type), modulo);
				result = fieldInfo;
			}

			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);
		}
		else if (tag == LyInfoTokens.TAG_OPTION)
		{
			String  value = node.getString(LyInfoTokens.ATT_VALUE);
			boolean def = node.getBoolean(LyInfoTokens.ATT_DEFAULT);

			result = new LyChoiceOption(id, null, null, value, def, false, false, false);
		}
		else if (tag == LyInfoTokens.TAG_OPTIONS)
		{
			String operation = node.getString(LyInfoTokens.ATT_OPERATION);

			if ((operation == LyInfoTokens.OPER_LIST_RESET)
					|| (operation == LyInfoTokens.OPER_LIST_ADD))
			{
				String value = node.getString(LyInfoTokens.ATT_VALUE);

				result = new LyChoiceOptionGroup(id, null, null, value);
			}
			else
				result = new Vector<LyChoiceOption>();
		}
		else if (tag == LyInfoTokens.TAG_PROJECT)
		{
			result = getInfo(id);
			if (result == null)
				result = new LyProject(id, null, null, null, null);
			else if (!(result instanceof LyProject))
				throw new SAXException("BF AppEngine : info " + id + " is not a project");
		}
		else if (tag == LyInfoTokens.TAG_REGEX)
		{
			String value = node.getString(LyInfoTokens.ATT_VALUE);
			LyInfo regex = getInfo(id);

			if (regex == null)
				result = new LyRegExp(id, value);
			else
			{
				if (!(regex instanceof LyRegExp))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not a regexp");
				if (regex.isLoaded())
					throw new SAXException("BF AppEngine : regex " + id
						+ " already exists");

				((LyRegExp)regex).construct(value);
				result = regex;
			}
		}
		else if (tag == LyInfoTokens.TAG_RELATION)
		{
			String extendsId = node.getString(LyInfoTokens.ATT_EXTENDS);
			LyInfo extendedRelation = getInfo(extendsId);

			// Forbid forward reference on extends
			if ((extendsId != null)
					&& (extendsId.length() > 0)
					&& (extendedRelation == null))
				throw new IllegalStateException(
					"BF AppEngine : Forward reference to super-relation is not supported. Relation '"
					+ id + "' with super-relation : '" + extendsId + "'.");

			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			boolean abstractRelation = node.getBoolean(LyInfoTokens.ATT_ABSTRACT);
			boolean multiple = node.getBoolean(LyInfoTokens.ATT_MULTIPLE);
			boolean ordered = node.getBoolean(LyInfoTokens.ATT_ORDERED);
			boolean calculated = node.getBoolean(LyInfoTokens.ATT_CALCULATED);
			String  direction = node.getString(LyInfoTokens.ATT_DIRECTION);
			String  userData = node.getString(LyInfoTokens.ATT_USER_DATA);
			String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			LyInfo  fieldInfo = getInfo(id);

			if (fieldInfo == null)
				result = new LyRelationFieldInfo(id, javaName,
						(LyRelationFieldInfo)extendedRelation, (LyRelationFieldInfo)template,
						multiple, ordered, LyInfoTokens.getRelationFieldInfoDirection(direction),
						LyInfoTokens.getRelationFieldInfoType(type), userData, false, false,
						LyRelationFieldInfo.ShowAs.DEFAULT);
			else
			{
				if (!(fieldInfo instanceof LyRelationFieldInfo))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not a relation field info");
				if (fieldInfo.isLoaded())
					throw new SAXException("BF AppEngine : relation field info " + id
						+ " already exists");

				((LyRelationFieldInfo)fieldInfo).construct(javaName,
					(LyRelationFieldInfo)extendedRelation, (LyRelationFieldInfo)template, multiple,
					ordered, LyInfoTokens.getRelationFieldInfoDirection(direction),
					LyInfoTokens.getRelationFieldInfoType(type), userData, false, false,
					LyRelationFieldInfo.ShowAs.DEFAULT);
				result = fieldInfo;
			}

			((LyFieldInfo)result).setCalculated(calculated);
			((LyRelationFieldInfo)result).setAbstract(abstractRelation);
			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);
		}
		else if (tag == LyInfoTokens.TAG_RULE)
		{
			String			 str = node.getString(LyInfoTokens.ATT_OPERATION);
			int				 priority = node.getInt(LyInfoTokens.ATT_PRIORITY);
			LyRule.Operation operation = LyInfoTokens.getRuleOperation(str);
			boolean			 onCheck = node.getBoolean(LyInfoTokens.ATT_ON_CHECK);
			String			 value = node.getString(LyInfoTokens.ATT_VALUE);

			result = new LyRule(id, null, priority, operation, onCheck, value,
			        LyTag.Model.DATA);
		}
		else if (tag == LyInfoTokens.TAG_STRUCT)
		{
			String extendsId = node.getString(LyInfoTokens.ATT_EXTENDS);
			LyInfo extendedStruct = getInfo(extendsId);

			// Forbid forward reference on extends
			if ((extendsId != null)
					&& (extendsId.length() > 0)
					&& (extendedStruct == null))
				throw new IllegalStateException(
					"BF AppEngine : Forward reference to super-struct is not supported. Struct '" + id
					+ "' with super-struct : '" + extendsId + "'.");

			// Create relation field info
			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			String  className = node.getString(LyInfoTokens.ATT_CLASS_NAME);
			boolean abstractClass = node.getBoolean(LyInfoTokens.ATT_ABSTRACT);
			String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			LyInfo  fieldInfo = getInfo(id);

			if (fieldInfo == null)
				result = new LyStructFieldInfo(id, javaName, (LyStructFieldInfo)extendedStruct,
						(LyStructFieldInfo)template, className, abstractClass);
			else
			{
				if (!(fieldInfo instanceof LyStructFieldInfo))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not a struct field info");
				if (fieldInfo.isLoaded())
					throw new SAXException("BF AppEngine : struct field info " + id
						+ " already exists");

				((LyStructFieldInfo)fieldInfo).construct(javaName,
					(LyStructFieldInfo)extendedStruct, (LyStructFieldInfo)template, className,
					abstractClass);
				result = fieldInfo;
			}

			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);
		}
		else if (tag == LyInfoTokens.TAG_TEXT)
		{
			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			boolean multiple = node.getBoolean(LyInfoTokens.ATT_MULTIPLE);
			boolean multilingual = node.getBoolean(LyInfoTokens.ATT_MULTILINGUAL);
			String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			LyInfo  fieldInfo = getInfo(id);

			if (fieldInfo == null)
				result = new LyTextFieldInfo(id, javaName, (LyTextFieldInfo)template,
				    multiple, LyInfoTokens.getTextFieldInfoType(type), multilingual);
			else
			{
				if (!(fieldInfo instanceof LyTextFieldInfo))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not a text field info");
				if (fieldInfo.isLoaded())
					throw new SAXException("BF AppEngine : text field info " + id
						+ " already exists");

				((LyTextFieldInfo)fieldInfo).construct(javaName, (LyTextFieldInfo)template,
					multiple, LyInfoTokens.getTextFieldInfoType(type), multilingual);
				result = fieldInfo;
			}

			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);
		}
		else if (tag == LyInfoTokens.TAG_UNIT)
		{
			String factor = node.getString(LyInfoTokens.ATT_FACTOR);
			String offset = node.getString(LyInfoTokens.ATT_OFFSET);
			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			LyInfo unit = getInfo(id);

			if ((factor == null) || (factor.length() == 0))
				factor = "1.0";

			if (unit == null)
				result = new LyUnit(id, LyTools.doubleFromString(factor, 1.0),
						LyTools.doubleFromString(offset), javaName);
			else
			{
				if (!(unit instanceof LyUnit))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not a unit");
				if (unit.isLoaded())
					throw new SAXException("BF AppEngine : unit " + id
						+ " already exists");

				((LyUnit)unit).construct(LyTools.doubleFromString(factor, 1.0),
					LyTools.doubleFromString(offset), javaName);
				result = unit;
			}
		}
		else if (tag == LyInfoTokens.TAG_UNITS)
		{
			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			LyInfo units = getInfo(id);

			if (units == null)
				result = new LyUnits(id, javaName);
			else
			{
				if (!(units instanceof LyUnits))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not a units list");
				if (units.isLoaded())
					throw new SAXException("BF AppEngine : units list " + id
						+ " already exists");

				((LyUnits)units).construct(javaName);
				result = units;
			}
		}
		else if (tag == LyInfoTokens.TAG_UNITS_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyUnits))
				throw new SAXException("BF AppEngine : invalid units ref : " + idref);

			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);

			if (javaName != null)
				((LyUnits)result).setJavaName(javaName);
		}
		else
			result = super.buildElement(node, tag);

		return result;
	} // end method buildElement

	/**
	 * Init object info from its template.
	 *
	 * @param objectInfo Object info to initialize.
	 */
	protected void initFromTemplate(LyObjectInfo objectInfo)
	{
		if ((objectInfo instanceof LyClassInfo)
				|| (objectInfo instanceof LyFieldInfo))
			objectInfo.initFromTemplate(LyObjectInfo.PART_DATA);
		else
			super.initFromTemplate(objectInfo);
	}

	protected void check(LyObjectInfo objectInfo, LyObjectInfo parentInfo)
	{
		if ((objectInfo instanceof LyClassInfo)
				|| (objectInfo instanceof LyFieldInfo))
			objectInfo.check(parentInfo, LyObjectInfo.PART_DATA);
		else
			super.check(objectInfo, parentInfo);
	}

	/**
	 * The purpose of this method is to resolve object references now that all valid object ids are
	 * known.
	 *
	 * @param     tag    Current tag.
	 * @param     result Current result.
	 * @return    New result.
	 * @exception SAXExpection If the object found doesn't belong to the appropriate class.
	 */
	Object resolveObjectInfoRef(LyInfoNode node, String tag, Object result)
		throws SAXException
	{
		// Compute idrefs on tags ACTIONS, CLASSES, FIELDS, OPTIONS
		if ((tag == LyInfoTokens.TAG_CLASSES)
				|| (tag == LyInfoTokens.TAG_FIELDS)
				|| (tag == LyInfoTokens.TAG_OPTIONS))
		{
			if (result == null) // Nothing to do
				return null;

			String   idrefsAtt = node.getString(LyInfoTokens.ATT_IDREFS);
			String   operation = node.getString(LyInfoTokens.ATT_OPERATION);
			String[] idrefs = null;
			int		 i, n = 0;
			LyInfo   info;

			if (idrefsAtt != null)
			{
				idrefs = LyString.stringToList(idrefsAtt, ' ');
				n = ((idrefs == null) ? 0 : idrefs.length);
			}

			for (i = 0; i < n; i++)
			{
				// Check out class reference
				info = getInfo(idrefs[i]);
				if ((tag == LyInfoTokens.TAG_CLASSES)
						&& !(info instanceof LyClassInfo))
					raiseError("BF AppEngine : invalid class ref : " + idrefs[i]);
				else if ((tag == LyInfoTokens.TAG_FIELDS)
						&& !(info instanceof LyFieldInfo))
				    raiseError("BF AppEngine : invalid fieldInfo ref : " + idrefs[i]);
				else if ((tag == LyInfoTokens.TAG_OPTIONS)
						&& !(info instanceof LyChoiceOption))
				    raiseError("BF AppEngine : invalid option ref : " + idrefs[i]);

				// Add reference to the result list
				if (result instanceof LyChoiceOptionGroup)
					((LyChoiceOptionGroup)result).addChoiceItem((LyChoiceOption)info);
				else
				{
					@SuppressWarnings("unchecked")
					Vector<Object> temp = (Vector<Object>)result;
					temp.addElement(info);
				}

				if (operation != LyInfoTokens.OPER_LIST_REMOVE)
					node.addExtraInfo(info);
			}

			if (result instanceof LyChoiceOptionGroup)
			{
				String option = node.getString(LyInfoTokens.ATT_OPTION);

				if (option != null)
				{
					info = getInfo(option);
					if (info instanceof LyChoiceOption)
						((LyChoiceOptionGroup)result).setOption((LyChoiceOption)info);
					else
						throw new SAXException("BF AppEngine : info " + option
							+ " is not a choice option");
				}
			}
		}
		// Reference to a fieldInfo
		else if ((tag == LyInfoTokens.TAG_FIELD_REF)
				|| (tag == LyInfoTokens.TAG_CLASS_FIELD)
				|| (tag == LyInfoTokens.TAG_CONTEXT_FIELD))
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyFieldInfo))
				throw new SAXException("BF AppEngine : invalid fieldInfo ref : " + idref);
		}
		// Reference to a classInfo
		else if (tag == LyInfoTokens.TAG_CLASS_REF)
		{
			String idref;
			LyInfo classInfo, filter;

			idref = node.getString(LyInfoTokens.ATT_IDREF);
			classInfo = getInfo(idref);
			if (!(classInfo instanceof LyClassInfo))
				throw new SAXException("BF AppEngine : invalid classInfo ref : " + idref);

			idref = node.getString(LyInfoTokens.ATT_FILTER);
			filter = getInfo(idref);
			if ((filter != null) && !(filter instanceof LyFilter))
				throw new SAXException("BF AppEngine : invalid filter ref : " + idref);

			// build classInfoRef
			result = new LyClassRef((LyClassInfo)classInfo, (LyFilter)filter);
		}
		// Reference to a unit
		else if (tag == LyInfoTokens.TAG_UNIT_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyUnit))
				throw new SAXException("BF AppEngine : invalid unit ref : " + idref);
		}
		// subTypes on FIELD
		else if (tag == LyInfoTokens.TAG_FIELD)
		{
			String subTypes = node.getString(LyInfoTokens.ATT_SUB_TYPES);

			if ((subTypes != null) && !(result instanceof LyRelationFieldInfo))
			{
				_application.getEnvironment().logWarning(
					"BF AppEngine : subTypes attribute ignored on field "
					+ result + ", because this field is not a relation");
				return result;
			}

			if (subTypes == null)
/*
                                if (result instanceof LyRelationFieldInfo)
                                        throw new SAXException("BF AppEngine : missing subTypes attribute on field "
                                                + result + ", because this field is a relation");
                                else
*/
				return result; // Nothing to do

			LyRelationFieldInfo relation = (LyRelationFieldInfo)result;
			String[]		    idrefs = LyString.stringToList(subTypes, ' ');
			int				    i, n = ((idrefs == null) ? 0 : idrefs.length);
			LyInfo			    info;

			for (i = 0; i < n; i++)
			{
				// Check out subType reference
				info = getInfo(idrefs[i]);
				if (!(info instanceof LyRelationFieldInfo))
					throw new SAXException("BF AppEngine : invalid field ref : " + idrefs[i]);

				if (((LyRelationFieldInfo)info).isAbstract())
				{
					String relationStr = ((relation instanceof LyStructFieldInfo) ? "struct"
																				  : "relation");

					throw new IllegalStateException("BF AppEngine : " + relationStr + " \""
						+ relation.getId() + "\" can't have an abstract subType : \""
						+ info.getId() + "\"");
				}
				relation.addSubType((LyRelationFieldInfo)info);
			}
		}
		// Reference to a charType
		else if (tag == LyInfoTokens.TAG_CHARTYPE_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyInfoCharType))
				throw new SAXException("BF AppEngine : invalid charType ref : " + idref);
		}
		// Reference to a regular expression
		else if (tag == LyInfoTokens.TAG_REGEX_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyRegExp))
				throw new SAXException("BF AppEngine : invalid reg exp ref : " + idref);
		}
		// Rule : Reference to a fieldInfo
		else if (tag == LyInfoTokens.TAG_RULE)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);
			LyInfo info = getInfo(idref);

			if (!(info instanceof LyFieldInfo))
				throw new SAXException("BF AppEngine : invalid fieldInfo ref : " + idref
					+ " found on " + result);

			((LyRule)result).setObjectInfo((LyFieldInfo)info);
		}

		return result;
	} // end method resolveObjectInfoRef

	/**
	 * Completes parent result with the result of the current node.
	 *
	 * @param parentResult Result of the parent node.
	 * @param node         Current node.
	 * @param result       Result of current node.
	 */
	void completeParentResult(Object parentResult, LyInfoNode node, Object result)
		throws SAXException
	{
		if ((parentResult instanceof LyClassRef)
				&& (result instanceof LyFilter))
			// If parentResult is a classInfoRef, we complete it with this result
			((LyClassRef)parentResult).setFilter((LyFilter)result);
	}

	/**
	 * Completes an info with the result of the current node.
	 *
	 * @param     info Info to be completed.
	 * @param     node Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeInfo(LyInfo info, LyInfoNode node)
		throws SAXException
	{
		// Call the appropriate method to complete the info
		if (info instanceof LyChoiceOptionGroup)
			completeChoiceOptionGroup((LyChoiceOptionGroup)info, node);
		else if (info instanceof LyUnits)
			completeUnits((LyUnits)info, node);
		else if (info instanceof LyRule)
			completeRule((LyRule)info, node);
		else
			super.completeInfo(info, node);
	}

	/**
	 * Completes an objectInfo with the result of the current node.
	 *
	 * @param     objectInfo ObjectInfo to be completed.
	 * @param     node       Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeObjectInfo(LyObjectInfo objectInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to objectInfo attributes
		// and complete the given objectInfo.
		if (tag == LyInfoTokens.TAG_DATA)
			objectInfo.putData(node.getString(LyInfoTokens.ATT_NAME), result,
			    LyTag.Model.DATA);

		// For other tags, call the appropriate method to complete the objectInfo
		else
			super.completeObjectInfo(objectInfo, node);
	}

	/**
	 * Completes a project with the result of the current node.
	 *
	 * @param project Project to be completed.
	 * @param node    Current node.
	 */
	void completeProject(LyProject project, LyInfoNode node)
	{
		// Add the current result to project by using the right method
		Object result = node.getResult();

		if (result instanceof LyInfo)
			project.addInfo((LyInfo)result);
	}

	/**
	 * Completes an fieldInfo with the result of the current node.
	 *
	 * @param     fieldInfo FieldInfo to be completed.
	 * @param     node      Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeFieldInfo(LyFieldInfo fieldInfo, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();
		String tag = node.getTag();

		// Check for tags corresponding to objectInfo attributes
		// and complete the given fieldInfo.
		if (tag == LyInfoTokens.TAG_F_MARKS)
			mergeMarks(fieldInfo, (LyBitSet)result, node);
		else if ((tag == LyInfoTokens.TAG_CONTROL)
				|| (tag == LyInfoTokens.TAG_CONTROL_REF))
			fieldInfo.addControl((LyControl)result);
		else if ((tag == LyInfoTokens.TAG_UNIT)
				|| (tag == LyInfoTokens.TAG_UNIT_REF))
			fieldInfo.setFieldUnit((LyUnit)result);
		else if ((tag == LyInfoTokens.TAG_UNITS)
				|| (tag == LyInfoTokens.TAG_UNITS_REF))
			fieldInfo.setUnits((LyUnits)result);
        else if (tag == LyInfoTokens.TAG_FORMULA)
            fieldInfo.setFormula((String)result);

		// For other tags, call the appropriate method to complete the fieldInfo
		else if (fieldInfo instanceof LyArrayFieldInfo)
			completeArrayFieldInfo((LyArrayFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyChoiceFieldInfo)
			completeChoiceFieldInfo((LyChoiceFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyDateFieldInfo)
			completeDateFieldInfo((LyDateFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyNumberFieldInfo)
			completeNumberFieldInfo((LyNumberFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyStructFieldInfo)
			completeStructFieldInfo((LyStructFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyRelationFieldInfo)
			completeRelationFieldInfo((LyRelationFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyTextFieldInfo)
			completeTextFieldInfo((LyTextFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyFileFieldInfo)
			completeFileFieldInfo((LyFileFieldInfo)fieldInfo, node);
	} // end method completeFieldInfo

	/**
	 * Completes an array field info with the result of the current node.
	 *
	 * @param     arrayFieldInfo ArrayFieldInfo to be completed.
	 * @param     node           Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeArrayFieldInfo(LyArrayFieldInfo arrayFieldInfo, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		if (result instanceof LyFieldInfo)
			arrayFieldInfo.setFieldInfo((LyFieldInfo)result);
	}

	/**
	 * Completes a date field info with the result of the current node.
	 *
	 * @param     dateFieldInfo DateFieldInfo to be completed.
	 * @param     node          Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeDateFieldInfo(LyDateFieldInfo dateFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to dateFieldInfo attributes
		// and complete the given dateFieldInfo.
		if (tag == LyInfoTokens.TAG_FORMAT)
			dateFieldInfo.setDisplayFormat((String)result);
		else if (tag == LyInfoTokens.TAG_DEFAULT)
		{
			Object value;

			if (((String)result).equals("now")
					|| ((String)result).equals("_now"))
				value = LyDateFieldInfo.VALUE_NOW;
			else
				value = dateFieldInfo.parse((String)result);

			dateFieldInfo.setDefaultValue(value);
		}
	}

	/**
	 * Completes a number field info with the result of the current node.
	 *
	 * @param     numberFieldInfo NumberFieldInfo to be completed.
	 * @param     node            Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeNumberFieldInfo(LyNumberFieldInfo numberFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String				   tag = node.getTag();
		Object				   result = node.getResult();

		// Check for tags corresponding to numberFieldInfo attributes
		// and complete the given numberFieldInfo.
		LyNumberFieldInfo.Type type = numberFieldInfo.getType();

		if (tag == LyInfoTokens.TAG_FORMAT)
			numberFieldInfo.setFormat((String)result);
		else if (tag == LyInfoTokens.TAG_FUNCTION)
			numberFieldInfo.setFunction(LyInfoTokens.getNumberFieldInfoFunction((String)result));
		else if (tag == LyInfoTokens.TAG_FORMULA)
			numberFieldInfo.setFormula((String)result);
		else if (tag == LyInfoTokens.TAG_MIN)
		{
			Number number = LyInfoTokens.parseNumber(type, (String)result);

			numberFieldInfo.setMin(number);
		}
		else if (tag == LyInfoTokens.TAG_MAX)
		{
			Number number = LyInfoTokens.parseNumber(type, (String)result);

			numberFieldInfo.setMax(number);
		}
		else if (tag == LyInfoTokens.TAG_N_DEFAULT)
		{
			Number number = LyInfoTokens.parseNumber(type, (String)result);

			numberFieldInfo.setDefault(number);
		}
		else if (tag == LyInfoTokens.TAG_PRECISION)
			numberFieldInfo.setPrecision(LyInfoTokens.parseNumber(type, (String)result));
	} // end method completeNumberFieldInfo

	/**
	 * Completes a text field info with the result of the current node.
	 *
	 * @param     textFieldInfo TextFieldInfo to be completed.
	 * @param     node          Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeTextFieldInfo(LyTextFieldInfo textFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to textFieldInfo attributes
		// and complete the given textFieldInfo.
		if (tag == LyInfoTokens.TAG_FORMAT)
			textFieldInfo.setFormat((String)result);
		else if (tag == LyInfoTokens.TAG_ENCODING)
			textFieldInfo.setEncoding((String)result);
		else if ((tag == LyInfoTokens.TAG_CHARTYPE)
				|| (tag == LyInfoTokens.TAG_CHARTYPE_REF))
			textFieldInfo.addCharType((LyInfoCharType)result);
		else if ((tag == LyInfoTokens.TAG_REGEX)
				|| (tag == LyInfoTokens.TAG_REGEX_REF))
			textFieldInfo.setRegExp((LyRegExp)result);
		else if (tag == LyInfoTokens.TAG_MIN)
			textFieldInfo.setMinChars(LyInfoTokens.parseInt((String)result));
		else if ((tag == LyInfoTokens.TAG_NCHARS)
				|| (tag == LyInfoTokens.TAG_MAX))
			textFieldInfo.setMaxChars(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_DEFAULT)
			textFieldInfo.setDefault((String)result);
	}

	/**
	 * Completes a file field info with the result of the current node.
	 *
	 * @param     fileFieldInfo FileFieldInfo to be completed.
	 * @param     node          Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeFileFieldInfo(LyFileFieldInfo fileFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to fileFieldInfo attributes
		// and complete the given fileFieldInfo.
		if (tag == LyInfoTokens.TAG_DEFAULT)
			fileFieldInfo.setDefault((String)result);
	}

	/**
	 * Completes a choice field info with the result of the current node.
	 *
	 * @param     choiceFieldInfo ChoiceFieldInfo to be completed.
	 * @param     node            Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeChoiceFieldInfo(LyChoiceFieldInfo choiceFieldInfo, LyInfoNode node)
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to choiceFieldInfo attributes
		// and complete the given choiceFieldInfo.
		if (tag == LyInfoTokens.TAG_OPTIONS)
		{
			String		   operation = node.getString(LyInfoTokens.ATT_OPERATION);
			int			   pos = node.getInt(LyInfoTokens.ATT_POSITION);
			LyChoiceOption option;

			if (operation != null)
				operation = operation.intern();

			if (operation == LyInfoTokens.OPER_LIST_RESET)
			{
				choiceFieldInfo.resetChoiceOptions();
				pos = 0;
			}

			if ((operation == LyInfoTokens.OPER_LIST_ADD)
					|| (operation == LyInfoTokens.OPER_LIST_RESET))
			{
				LyChoiceOptionGroup		   optionGroup = (LyChoiceOptionGroup)result;
				LyInfoList<LyChoiceOption> options = optionGroup.getChoiceOptions(true);
				int						   i, n = options.getSize();

				for (i = 0; i < n; i++)
				{
					option = options.getInfo(i);
					choiceFieldInfo.addChoiceOption(option, ((pos < 0) ? -1 : (pos + i)));
				}

				choiceFieldInfo.addChoiceOptionGroup(optionGroup);
			}
			else if (operation == LyInfoTokens.OPER_LIST_REMOVE)
			{
				@SuppressWarnings("unchecked")
				Vector<LyChoiceOption> options = (Vector<LyChoiceOption>)result;
				int					   i, n = options.size();

				for (i = 0; i < n; i++)
				{
					option = options.elementAt(i);
					choiceFieldInfo.removeChoiceOption(option);
				}

				node.setResult(null);
			}
		}
		else if (tag == LyInfoTokens.TAG_CONTEXT_FIELD)
		{
			LyFieldInfo fieldInfo = (LyFieldInfo)result;

			if (!(fieldInfo instanceof LyChoiceFieldInfo))
				throw new IllegalStateException("BF AppEngine : context field " + fieldInfo
					+ " is not a CHOICE field. Only choice field can be used as a context field"
					+ "for another choice.");

			choiceFieldInfo.setContextField((LyChoiceFieldInfo)fieldInfo);
		}

	} // end method completeChoiceFieldInfo

	/**
	 * Completes a choice option info with the result of the current node.
	 *
	 * @param     choiceOption ChoiceOption to be completed.
	 * @param     node         Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeChoiceOption(LyChoiceOption choiceOption, LyInfoNode node)
	{
		// Check for tags corresponding to choiceOption attributes
		// and complete the given option.
	}

	/**
	 * Completes a choice options group info with the result of the current node.
	 *
	 * @param     choiceOptionGroup ChoiceOptionGroup to be completed.
	 * @param     node              Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeChoiceOptionGroup(LyChoiceOptionGroup choiceOptionGroup, LyInfoNode node)
	{
		// Check for tags corresponding to choiceOption attributes
		// and complete the given option.
		Object result = node.getResult();

		if (result instanceof LyChoiceItem)
			choiceOptionGroup.addChoiceItem((LyChoiceItem)result);
	}

	/**
	 * Completes a relation field info with the result of the current node.
	 *
	 * @param     relationFieldInfo RelationFieldInfo to be completed.
	 * @param     node              Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeRelationFieldInfo(LyRelationFieldInfo relationFieldInfo,
		LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to relationFieldInfo attributes
		// and complete the given relationFieldInfo.
		if (tag == LyInfoTokens.TAG_DEFAULT)
			relationFieldInfo.setDefault((String)result);
		else if (tag == LyInfoTokens.TAG_CLASSES)
		{
			String			    operation = node.getString(LyInfoTokens.ATT_OPERATION);
			int				    pos = node.getInt(LyInfoTokens.ATT_POSITION);

			Vector<?>		    classes = (Vector<?>)result;
			int				    i, j, n = classes.size();
			Object			    info;
			LyClassInfo		    classInfo, extendedClassInfo;
			LyFilter		    filter;
			LyRelationFieldInfo extendedRelation = relationFieldInfo.getExtendedRelation();

			if (extendedRelation != null)
			{
				relationFieldInfo.resetClasses();
				for (i = 0; i < n; i++)
				{
					info = classes.elementAt(i);
					filter = null;
					if (info instanceof LyClassInfo)
						classInfo = (LyClassInfo)info;
					else
					{
						classInfo = ((LyClassRef)info).getClassInfo();
						filter = ((LyClassRef)info).getFilter();
					}

					for (j = 0; j < extendedRelation.getClassInfoCount(); j++)
					{
						extendedClassInfo = extendedRelation.getClassInfo(j);
						if (classInfo.inheritsFrom(extendedClassInfo))
							break;
					}
					if (j == extendedRelation.getClassInfoCount())
						throw new SAXException("BF AppEngine : classInfo " + classInfo.getId()
							+ " does not extend any class from the extended relation "
							+ extendedRelation.getId());

					relationFieldInfo.addClassInfo(classInfo, filter, i);
				}
			}
			else
			{
				if (operation != null)
					operation = operation.intern();

				if (operation == LyInfoTokens.OPER_LIST_RESET)
				{
					relationFieldInfo.resetClasses();
					pos = 0;
				}

				if ((operation == LyInfoTokens.OPER_LIST_ADD)
						|| (operation == LyInfoTokens.OPER_LIST_RESET))
				{
					for (i = 0; i < n; i++)
					{
						info = classes.elementAt(i);
						if (info instanceof LyClassInfo)
							relationFieldInfo.addClassInfo((LyClassInfo)info, null,
								((pos < 0) ? -1 : (pos + i)));
						else
							relationFieldInfo.addClassInfo(((LyClassRef)info).getClassInfo(),
								((LyClassRef)info).getFilter(), ((pos < 0) ? -1 : (pos + i)));
					}
				}
				else if (operation == LyInfoTokens.OPER_LIST_REMOVE)
				{
					for (i = 0; i < n; i++)
					{
						info = classes.elementAt(i);
						if (info instanceof LyClassInfo)
							relationFieldInfo.removeClassInfo((LyClassInfo)info, null);
						else
							relationFieldInfo.removeClassInfo(((LyClassRef)info)
									.getClassInfo(), ((LyClassRef)info).getFilter());
					}
					node.setResult(null);
				}
			}
		}
		else if (tag == LyInfoTokens.TAG_MIN)
			relationFieldInfo.setMin(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_MAX)
			relationFieldInfo.setMax(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_CLASS_FIELD)
			relationFieldInfo.setClassField((LyFieldInfo)result);
		else if (tag == LyInfoTokens.TAG_CONTEXT_FIELD)
			relationFieldInfo.setContextField((LyRelationFieldInfo)result);
	} // end method completeRelationFieldInfo

	/**
	 * Completes a struct field info with the result of the current node.
	 *
	 * @param     relationFieldInfo structFieldInfo to be completed.
	 * @param     node              Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeStructFieldInfo(LyStructFieldInfo structFieldInfo,
		LyInfoNode node)
		throws SAXException
	{
		String	    tag = node.getTag();
		Object	    result = node.getResult();
		LyClassInfo classInfo = structFieldInfo.getClassInfo(0);

		if (tag == LyInfoTokens.TAG_FIELDS)
			completeClassInfo(classInfo, node);
		else if ((tag == LyInfoTokens.TAG_RULE)
				|| (tag == LyInfoTokens.TAG_RULE_REF))
			classInfo.addRule((LyRule)result);
	}

	/**
	 * Completes a rule with the result of the current node.
	 *
	 * @param     rule Rule to be completed.
	 * @param     node Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeRule(LyRule rule, LyInfoNode node)
		throws SAXException
	{
	    String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags or results corresponding to rule attributes
		// and complete the given label.
		if (result instanceof LyFilter)
			rule.setFilter((LyFilter)result);
	    else if (tag == LyInfoTokens.TAG_FORMULA)
	        rule.setFormula((String)result);
	}

	/**
	 * Completes a units list with the result of the current node.
	 *
	 * @param units Units list to be completed.
	 * @param node  Current node.
	 */
	void completeUnits(LyUnits units, LyInfoNode node)
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to units list attributes
		// and complete the given units list.
		if (result instanceof LyUnit)
			units.addUnit((LyUnit)result);
	}

	/**
	 * Completes a class info with the result of the current node.
	 *
	 * @param classInfo ClassInfo to be completed.
	 * @param node      Current node.
	 */
	void completeClassInfo(LyClassInfo classInfo, LyInfoNode node)
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to classInfo attributes
		// and complete the given classInfo.
		if (tag == LyInfoTokens.TAG_C_MARKS)
			mergeMarks(classInfo, (LyBitSet)result, node);
		else if (tag == LyInfoTokens.TAG_FIELDS)
		{
			String			    operation = node.getString(LyInfoTokens.ATT_OPERATION);
			int				    pos = node.getInt(LyInfoTokens.ATT_POSITION);

			@SuppressWarnings("unchecked")
			Vector<LyFieldInfo> fields = (Vector<LyFieldInfo>)result;
			int				    i, n = fields.size();

			if (operation != null)
				operation = operation.intern();

			if (operation == LyInfoTokens.OPER_LIST_RESET)
			{
				classInfo.resetFieldInfos();
				pos = 0;
			}

			if ((operation == LyInfoTokens.OPER_LIST_ADD)
					|| (operation == LyInfoTokens.OPER_LIST_RESET))
			{
				for (i = 0; i < n; i++)
					classInfo.addFieldInfo(fields.elementAt(i), ((pos < 0) ? -1 : (pos + i)));
			}
			else if (operation == LyInfoTokens.OPER_LIST_REMOVE)
			{
				for (i = 0; i < n; i++)
					classInfo.removeFieldInfo(fields.elementAt(i));
				node.setResult(null);
			}
		}
		else if (tag == LyInfoTokens.TAG_ROUTES)
		{
			String		    operation = node.getString(LyInfoTokens.ATT_OPERATION);
			int			    pos = node.getInt(LyInfoTokens.ATT_POSITION);

			@SuppressWarnings("unchecked")
			Vector<LyRoute> routes = (Vector<LyRoute>)result;
			int			    i, n = routes.size();

			if (operation != null)
				operation = operation.intern();

			if (operation == LyInfoTokens.OPER_LIST_RESET)
			{
				classInfo.resetRoutes();
				pos = 0;
			}

			if ((operation == LyInfoTokens.OPER_LIST_ADD)
					|| (operation == LyInfoTokens.OPER_LIST_RESET))
			{
				for (i = 0; i < n; i++)
					classInfo.addRoute(routes.elementAt(i), ((pos < 0) ? -1 : (pos + i)));
			}
			else if (operation == LyInfoTokens.OPER_LIST_REMOVE)
			{
				for (i = 0; i < n; i++)
					classInfo.removeRoute(routes.elementAt(i));
				node.setResult(null);
			}
		}
		else if ((tag == LyInfoTokens.TAG_CONTROL)
				|| (tag == LyInfoTokens.TAG_CONTROL_REF))
			classInfo.addControl((LyControl)result);
		else if ((tag == LyInfoTokens.TAG_RULE)
				|| (tag == LyInfoTokens.TAG_RULE_REF))
			classInfo.addRule((LyRule)result);
	} // end method completeClassInfo

	/**
	 * Reset marks of the given.
	 */
	protected BitSet resetMarks(LyBitSet.Type type, BitSet marks)
	{
		if (marks != null)
		{
			// Only reset data marks
			if (type == LyBitSet.Type.C_MARKS)
			{
				BitSet dataMarks = LyClassInfo.getMarks(marks, LyObjectInfo.PART_DATA);

				marks.andNot(dataMarks);
				return marks;
			}
			else if (type == LyBitSet.Type.F_MARKS)
			{
				BitSet dataMarks = LyFieldInfo.getMarks(marks, LyObjectInfo.PART_DATA);

				marks.andNot(dataMarks);
				return marks;
			}
		}

		return super.resetMarks(type, marks);
	}

	/**
	 * Returns the model type of this factory.
	 */
	protected LyTag.Model getModelType()
	{
		return LyTag.Model.DATA;
	}
} // end class LyInfoDataFactory
