/*
 * LyInfoDataList.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;


/**
 * This class is an utility class to manage a list of Info with an associated data. This data is
 * untyped to allow different usages.<br>
 * By default, this list doesn't accept multiple occurrences of the same pair (info, data) but this
 * mode can be set on to enable them (this mode can only be chosen at beginning).<br>
 * At start, these lists are writable, but can't be modified anymore, once turned to unwritable.
 *
 * @see LyInfo
 * @see LyObjectInfo
 */
public final class LyInfoDataList<Info extends LyInfo, Data>
    extends LyInfoList<Info>
{
	public static final String __VERSION =
		"$Id: LyInfoDataList.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** List of data. */
	Vector<Data>			   _datas;

	/**
	 * Boolean indicating if multiple occurrences of the same pair (info, data) is allowed in the
	 * list. The same fieldInfo may be found several times but with different data associated to
	 * it.
	 */
	boolean					   _multiple;

	/* Constructors */

	/**
	 * Default constructor with no parameters. Multiple occurrences are not allowed.
	 */
	public LyInfoDataList()
	{
		this(false);
	}

	/**
	 * Constructor with choice of multiple occurrences or not.
	 *
	 * @param multiple Boolean indicating if multiple occurrences are allowed or not.
	 */
	public LyInfoDataList(boolean multiple)
	{
		super(true);

		_datas = null;
		_multiple = multiple;
	}
    
    /**
     * Constructor with choice of multiple occurrences or not.
     *
     * @param multiple Boolean indicating if multiple occurrences are allowed or not.
     */
    public LyInfoDataList(LyInfoDataList<Info, Data> infoDataList)
    {
        super(true);

        if (infoDataList == null)
            return;

        int i, n = infoDataList.getSize();

        for (i = 0; i < n; i++)
            addInfo(infoDataList.getInfo(i), infoDataList.getData(i));
        _multiple = infoDataList._multiple; 
    }

	/**
	 * Gets data at the specified index.
	 *
	 * @param     index Index of the requested data.
	 * @return    Data at the given index (may be null if no data specified at this position).
	 * @exception ArrayIndexOutOfBoundsException If index is negative, or greater than or equal to
	 *                                           the number of infos in the list.
	 */
	public Data getData(int index)
	{
		if (index < 0)
			throw new ArrayIndexOutOfBoundsException("BF AppEngine : " + index + " < 0 ");

		int n = getSize();

		if (index >= n)
			throw new ArrayIndexOutOfBoundsException("BF AppEngine : " + index + " >= " + n);

		return _datas.elementAt(index);
	}

	/**
	 * Gets data associated to the given info.
	 *
	 * @param     info Info for which an associated data is searched.
	 * @return    Data associated to this info (may be null if no data specified for this info).
	 * @exception IllegalArgumentException If the given info is not in this list.
	 */
	public Data getData(Info info)
	{
		int index = getIndex(info);

		if (index < 0)
			throw new IllegalArgumentException("BF AppEngine : info " + info + "is not in this list");

		return getData(index);
	}

	/**
	 * Adds a new pair (info, data) to this list at the given position.
	 *
	 * @param     info     Info that should be added.
	 * @param     data     Data associated to the info.
	 * @param     position Position where this pair has to be added. A negative index or an index
	 *                     greater than or equal to the number of infos in this list, means insert
	 *                     at the end of the list.
	 * @return    Effective index where the pair (info, data) was inserted. -1 if this pair was
	 *            already in this list and this list doesn't allow multiple occurrences.
	 * @exception IllegalStateException If this list is not writable.
	 */
	public int addInfo(Info info, Data data, int position)
	{
		if (!isWritable())
			throw new IllegalStateException("BF AppEngine : this info list is not writable");

		// Adds the new pair (info, data) if not already added
		if (!_multiple
				&& contains(info, data))
			return -1;

		int index = super.addInfo(info, position);

		if (index < 0)
			return -1;

		if (_datas == null)
			_datas = new Vector<Data>(1);

		_datas.insertElementAt(data, index);
		return index;
	}

	/**
	 * Adds a new pair (info, data) to this list at the end of the list.
	 *
	 * @param     info Info that should be added.
	 * @param     data Data associated to the info.
	 * @return    Effective index where the pair (info, data) was inserted. -1 if this pair was
	 *            already in this list and this list doesn't allow multiple occurrences.
	 * @exception IllegalStateException If this list is not writable.
	 */
	public int addInfo(Info info, Data data)
	{
		return addInfo(info, data, -1);
	}

	/**
	 * Adds a new info with no associated data to this list at the given position.
	 *
	 * @param     info     Info that should be added.
	 * @param     position Position where this info must be added. A negative index or an index
	 *                     greater than or equal to the number of infos in this list, means insert
	 *                     at the end of the list.
	 * @return    Effective index where the info was inserted. -1 if this info associated with no
	 *            data was already in this list and this list doesn't allow multiple occurrences.
	 * @exception IllegalStateException If this list is not writable.
	 */
	public int addInfo(Info info, int position)
	{
		return addInfo(info, null, position);
	}

	/**
	 * Removes the specified info from the info list. This method is overridden to remove data
	 * associated to this info. If several pairs are registered with the same field info, the first
	 * one is removed.
	 *
	 * @param  info Info to remove.
	 * @return Index of the removed pair, -1 if the info was not in this list.
	 */
	public int removeInfo(Info info)
	{
		if (!isWritable())
			throw new IllegalStateException("BF AppEngine : this info list is not writable");

		// data is removed by removeInfoAt called by removeInfo
		return super.removeInfo(info);
	}

	/**
	 * Removes the specified pair (info, data) from this list.
	 *
	 * @param     info Info to remove.
	 * @param     data Data associated to the info.
	 * @return    Index of the removed pair, -1 if the given pair couldn't be found.
	 * @exception IllegalStateException If this list is not writable.
	 */
	public int removeInfo(Info info, Data data)
	{
		if (!isWritable())
			throw new IllegalStateException("BF AppEngine : this info list is not writable");

		int index = getIndex(info, data);

		removeInfoAt(index);
		return index;
	}

	/**
	 * Remove info at the given position.
	 *
	 * @param     index Index of the info to remove.
	 * @exception ArrayIndexOutOfBoundsException If index is negative, or greater than or equal to
	 *                                           the number of infos in the list.
	 */
	void removeInfoAt(int index)
	{
		// Remove info
		super.removeInfoAt(index);

		// Remove associated data
		_datas.removeElementAt(index);
	}

	/**
	 * Adds all infos of the given list to this list. If the given list is a LyInfoDataList,
	 * associated data are also copied. Otherwise, infos are added with no associated data.
	 *
	 * @param     list List of infos to add to this list.
	 * @exception IllegalStateException If this list is not writable.
	 */
	public void addList(LyInfoDataList<Info, Data> list)
	{
		if (!isWritable())
			throw new IllegalStateException("BF AppEngine : this info list is not writable");

		if (list == null)
			return;

		int  i, n = list.getSize();
		Info info;
		Data data;

		for (i = 0; i < n; i++)
		{
			info = list.getInfo(i);
			data = list.getData(i);
			addInfo(info, data);
		}
	}

	/**
	 * Resets this list (remove all pairs from this list).
	 *
	 * @exception IllegalStateException If this list is not writable.
	 */
	public void resetInfos()
	{
		if (!isWritable())
			throw new IllegalStateException("BF AppEngine : this info list is not writable");

		super.resetInfos();
		_datas = null;
	}

	/**
	 * Get index of a pair (info, data) in this list.
	 *
	 * @param  info The searched info.
	 * @param  data Associated data (null means no data).
	 * @return Index if this pair in the list if it was found, -1 otherwise.
	 */
	public int getIndex(Info info, Data data)
	{
		int  i, n = getSize();
		Info current;

		for (i = 0; i < n; i++)
		{
			current = getInfo(i);
			if ((current == info)
					&& (getData(i) == data))
				return i;
		}

		return -1;
	}

	/**
	 * Checks if the given pair (info, data) is contained in this list.
	 *
	 * @param  info The searched info.
	 * @param  info Its associated data.
	 * @return True if the given pair is found in this list, false otherwise.
	 */
	public boolean contains(Info info, Data data)
	{
		return ((getIndex(info, data) < 0) ? false : true);
	}
} // end class LyInfoDataList
