/**
 * LyInfoFactory.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import org.xml.sax.*;

import java.util.*;
import leon.app.*;
import leon.info.misc.*;
import leon.misc.*;
import leon.parser.*;


/**
 * This class builds instances corresponding to XML tags read from the data description file. Almost
 * all objects built during parsing inherit from LyInfo (except very elementary objects).
 *
 * @see     LyInfoParser
 * @version $Id: LyInfoFactory.java 32662 2011-10-07 08:44:54Z bangiole $
 * @author  Lyria
 */
abstract public class LyInfoFactory
	implements LyXMLElementFactory
{
	public static final String __VERSION =
		"$Id: LyInfoFactory.java 32662 2011-10-07 08:44:54Z bangiole $";

	public static final String NULL_VALUE = "_null";
	public static final String OLD_NULL_VALUE = "NULL";

	/** Info parser that built this factory. */
	LyInfoParser			   _parser;

	/** Application for which this factory was built. */
	LyApplication			   _application;

	/** Kept for final post processing. */
	LyInfoNode				   _rootNode;

	/** Kept for creating tree structure (to be ready for final processing). */
	LyInfoNode				   _node;

	/** Cut node : node where an invalid condition was encountered. */
	LyInfoNode				   _cutNode;

	/** String that store parsing error(s). */
	String					   _parsingErrors;

	/**
	 * Constructor of the factory for an application. All objects built by this factory will belong
	 * to the given application.
	 *
	 * @param parser      Parser that builts this factory.
	 * @param application Application for which the factory is built.
	 */
	public LyInfoFactory(LyInfoParser parser, LyApplication application)
	{
		_parser = parser;
		_application = application;
		_rootNode = null;
		_node = null;
		_cutNode = null;
		_parsingErrors = null;
	}

	/**
	 * Gets application for which this factory was built.
	 *
	 * @return Application of this factory.
	 */
	protected LyApplication getApplication()
	{
		return _application;
	}

	/**
	 * Retrieves an info from its id.
	 *
	 * @param  id Id of searched info.
	 * @return The info if it was found, null otherwise.
	 */
	public LyInfo getInfo(String id)
	{
		LyApplication application = getApplication();

		return application.getInfo(id);
	}

	/**
	 * Builds node corresponding to an element read from the XML data description file. The result
	 * associated to the built node depends on the node tag. This method is invoked by SAXParser at
	 * the beginning of each new tag.
	 *
	 * @param     uri        The Namespace URI.
	 * @param     localName  The Namespace local name.
	 * @param     qName      The qualified (prefixed) name.
	 * @param     attributes The SAX2 attributes.
     * @param     parent     The parent node (created by {@link #buildElement(String, String, String,
     *                       Attributes)})
     * @return    The object built corresponding to the given tag and attributes.
     * @exception SAXException if the result could not have been built.
     */
    public final Object buildElement(String uri, String localName, String qName, Attributes attributes, Object parent)
		throws SAXException
	{
		// Create node instance
		String     tag = qName.intern();
		LyInfoNode node = new LyInfoNode(tag, attributes);
		String     id = node.getId();

		// Set tree structure
		if (_rootNode == null)
			_rootNode = node;

		if (_node != null)
			node.setParent(_node);

		_node = node;

		if (_cutNode != null)
			return null;

		if ((id != null)
			&& id.startsWith("_leon")
			&& (_application.getInfo(id) != null))
		{
			_cutNode = node;
			return null;
		}

		// Check condition before building result
		LyEnvironment env = _application.getEnvironment();

		if (tag != LyInfoTokens.TAG_EXP)
		{
			String condition = node.getString(LyInfoTokens.ATT_CONDITION);

			if (condition != null)
			{
				boolean not = (condition.charAt(0) == '!');

				if (not)
					condition = condition.substring(1);

				if ((!not && !env.testEnv(condition))
						|| (not && env.testEnv(condition)))
				{
					_cutNode = node;
					return null;
				}
			}
		}

		// Init result
		Object result = null;

		// Get some common information
		String type = node.getString(LyInfoTokens.ATT_TYPE);
		String templateId = node.getString(LyInfoTokens.ATT_MODEL);
		LyInfo template = getInfo(templateId);

		// Forbid forward reference on template
		if ((templateId != null)
		        && (templateId.length() > 0)
		        && (template == null))
			raiseError("BF AppEngine : Forward references to template objects are not supported. Element '"
			    + id + "' with template : '" + templateId + "'.");

		// Build the appropriate result corresponding to each node.
		try
		{
			if (tag == LyInfoTokens.TAG_ARG)
				result = new LyArgument(node.getString(LyInfoTokens.ATT_NAME),
						node.getString(LyInfoTokens.ATT_VALUE));
			else if (tag == LyInfoTokens.TAG_EXP)
			{
				boolean not = node.getBoolean(LyInfoTokens.ATT_NOT);
				boolean caseSensitive = node.getBoolean(LyInfoTokens.ATT_CASE_SENSITIVE);
				String  condition = node.getString(LyInfoTokens.ATT_CONDITION);
				String  value = node.getString(LyInfoTokens.ATT_VALUE);

				if ((value != null)
						&& (value.equalsIgnoreCase(NULL_VALUE)
							|| value.equalsIgnoreCase(OLD_NULL_VALUE))) // Kept for compatibility
					value = null;

				value = LyString.replace(value, LyApplication.VISIBLE_VECTOR_SEP, "" + LyTools.VECTOR_SEP);

				LyFilterExpression exp = new LyFilterExpression(null, not,
						LyInfoTokens.getFilterExpCondition(condition), value);

				if (caseSensitive)
					exp.setModifier(LyFilterExpression.MOD_CASE_SENSITIVE);

				result = exp;
			}
			else if (tag == LyInfoTokens.TAG_FILTER)
			{
				LyInfo         filter = getInfo(id);
				LyFilter.Type  filterType = LyInfoTokens.getFilterType(type);

				if (filter == null)
				{
					result = new LyFilter(id, (LyFilter)template, filterType);
					((LyFilter)result).setModelType(getModelType());
				}
				else if (filter == template)
					result = ((LyFilter)filter).cloneAsTemplate();
				else
				{
					if (!(filter instanceof LyFilter))
						throw new SAXException("BF AppEngine : extern info " + id + " is not a filter");
					if (((LyFilter)filter).isLoaded())
						throw new SAXException("BF AppEngine : filter " + id + " already exists");

					((LyFilter)filter).construct((LyFilter)template, filterType);
					result = filter;
				}
			}
			else if (tag == LyInfoTokens.TAG_FILTER_CRITERIA)
			{
				String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);

				result = new LyFilterCriterion(id, javaName);
			}
			else if (tag == LyInfoTokens.TAG_FILTER_RELATION)
				result = new LyFilterRelation();
			else if (tag == LyInfoTokens.TAG_PATH)
			{
				result = new LyPath(id, null, null, null, null);
				((LyPath)result).setModelType(getModelType());
			}
			else if (tag == LyInfoTokens.TAG_ROUTE)
			{
				boolean reverse = node.getBoolean(LyInfoTokens.ATT_REVERSE);
				boolean bidirectional = node.getBoolean(LyInfoTokens.ATT_BIDIRECTIONAL);
				boolean hidden = node.getBoolean(LyInfoTokens.ATT_HIDDEN);

				result = new LyRoute(id, null, null, reverse, bidirectional, hidden);
			}
			else if (tag == LyInfoTokens.TAG_SORT)
			{
				LyInfo	sort = getInfo(id);
				
				if (sort == null)
					result = new LySort(id, (LySort)template);
				else if (sort == template)
					result = ((LySort)sort).cloneAsTemplate();
				else
				{
					if (!(sort instanceof LySort))
						throw new SAXException("BF AppEngine : extern info " + id + " is not a sort");
					if (((LySort)sort).isLoaded())
					{
						// throw new SAXException("BF AppEngine : sort " + id + " already exists");
						// Appmodel : old preferences have sorts with the same id (FFT03201)
						result = new LySort(id, (LySort)template);
					}
					else
					{
						((LySort)sort).construct((LySort)template);
						result = sort;
					}
				}
			}
			else if (tag == LyInfoTokens.TAG_SORT_CRITERIA)
			{
				boolean increasing = node.getBoolean(LyInfoTokens.ATT_INCREASING);
				String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
                boolean groupBy = node.getBoolean(LyInfoTokens.ATT_GROUP_BY);

				result = new LySortCriterion(id, null, increasing, javaName);
                ((LySortCriterion)result).setGroupBy(groupBy);
			}
			else if (tag == LyInfoTokens.TAG_ROUTES)
				result = new Vector<LyRoute>();
			else if (tag == LyInfoTokens.TAG_FIELDS)
				result = new Vector<LyFieldInfo>();
			else
				result = buildElement(node, tag);
		}
		catch (Exception e)
		{
			raiseError(e.getMessage());
			
			// Ignore element
	        result = null;
		}

		completeResult(_node, result);
		_node.setResult(result);

		return result;
	} // end method buildElement

	/**
	 * Builds node corresponding to an element read from the XML data description file. The result
	 * associated to the built node depends on the node tag. This method is invoked by SAXParser at
	 * the beginning of each new tag.
	 *
	 * @param     node Current node.
	 * @param     tag  Current tag.
	 * @return    The object built corresponding to the given tag and attributes.
	 * @exception SAXException if the result could not have been built.
	 */
	public Object buildElement(LyInfoNode node, String tag)
		throws SAXException
	{
		String result = node.getString(LyInfoTokens.ATT_VALUE);

		return result;
	}

	/**
	 * Complete result of the given node.
	 */
	void completeResult(LyInfoNode node, Object result)
	{
		if (result instanceof LyInfo)
		    completeInfoResult(node, (LyInfo)result);
		else if (result instanceof LyTag)
		    completeTagResult(node, (LyTag)result);
	} // end method completeResult

	/**
     * Complete result of the given node.
     */
    void completeInfoResult(LyInfoNode node, LyInfo info)
    {
        // Register this info on parser results
        boolean   extern = ((node.getTag() == LyInfoTokens.TAG_INFO)
            || ((node.getTag() == LyInfoTokens.TAG_CLASS)
                && (node.getId() == null)));
        
        _parser.addResult(info, extern);

        // Set info parent
        completeTagResult(node, info);

        // Set the application to this info.
        info.setApplication(_application);
    } // end method completeResult
    
    /**
     * Complete result of the given node.
     */
    void completeTagResult(LyInfoNode node, LyTag tag)
    {
        // Set info parent
        if (tag.getParent() == null)
        {
            LyInfoNode parentNode = node.getParent();
            Object     parentResult;

            while (parentNode != null)
            {
                parentResult = parentNode.getResult();
                if (parentResult instanceof LyInfo)
                {
                    tag.setParent((LyInfo)parentResult);
                    break;
                }
                else
                    parentNode = parentNode.getParent();
            }
        }
    } // end method completeResult
    
	/**
	 * End of building a node. Starts last step of parsing - resolve references to other objects, -
	 * check result consistency. The result depends on the node tag. This method is invoked by
	 * SAXParser at the end of each new tag.
	 *
	 * @param     uri       The Namespace URI.
	 * @param     localName The Namespace local name.
	 * @param     qName     The qualified (prefixed) name.
	 * @param     parent    - The parent node (created by buildElement)
	 * @param     current   - The current node (created by buildElement)
	 * @exception SAXException if the result could not have been built. LyInfoDataFactory#endParse
	 *                         LyInfoDataFactory#checkResult
	 */
	public void endElement(String uri, String localName, String qName,
		Object parent, Object current)
		throws SAXException
	{
		if (_node == null)
			return;

		if (_cutNode == _node)
			_cutNode = null;

		LyInfoNode parentNode = _node.getParent();

		if ((parentNode == null)
				|| ((_node.getId() != null)
					&& _node.getId().equals("_leonStandardInclude")))
			// Final step : Parsing ends.
			endParse(_node);

		if (parentNode != null)
		{
			// Set son hierarchy
			parentNode.addChild(_node);
			_node = parentNode;
		}
		else
		{
			boolean ok = checkResult(null, _rootNode);

			_rootNode.free();
			_rootNode = null;
			_node = null;

			if (!ok)
			{
				String errors = getParsingErrors();

				if ((errors != null) && (errors.trim().length() > 0))
					throw new Error(errors);
				else
					throw new Error();
			}
		}
	} // end method endElement

	/**
	 * Last step in objects construction : finalize objectInfos. We need an extra step because info
	 * references can't be resolved before the document has been fully parsed, because our grammar
	 * enables forward references. We take advantage of this step to complete our objects.
	 *
	 * @param     parentResult Result of the parent node.
	 * @exception SAXException If an error is encountered during this final step.
	 */
	void endParse(LyInfoNode node)
		throws SAXException
	{
		if (node == null)
			return;
		if (node._ended)
			return;

		Object result = node.getResult();

		if ((result instanceof LyObjectInfo)
				&& (((LyInfo)result).getApplication() == _application)
				&& (node.getTag() != LyInfoTokens.TAG_INFO))
		{
			String id = node.getId();
			String superClassId = node.getString(LyInfoTokens.ATT_EXTENDS);

			// Don't call initFromTemplate when extending a class
			if ((id != null) || (superClassId == null))
				// At this time, we can copy template information.
				// That's why, we don't authorize forward reference for templates
				// because template objects must be fully completed before this copy.
				initFromTemplate((LyObjectInfo)result);
		}

		try
		{
		    // It's time now to check all references
		    resolveObjectInfoRef(node);
		    
	        // ObjectIdRef may have updated result content
	        result = node.getResult();
		}
		catch (SAXException e)
		{
		    raiseError(e.getMessage());
		    
		    // Ignore result
		    if (result instanceof LyInfo)
		        result = null; // Ignore result
		}

		if (result == null)
		{
			node._ended = true;
			return;
		}

		// Iterate on each child node
		int		   i, n = node.getChildCount();
		LyInfoNode child;

		for (i = 0; i < n; i++)
		{
			child = node.getChild(i);
			endParse(child);
		}

		// Consolidate parent result with result of the current node
		LyInfoNode parent = node.getParent();

		if (parent != null)
		{
			Object parentResult = parent.getResult();

			if (parentResult != null)
				completeParentResult(parentResult, node);
		}

		// Do not need parsing info any longer
		// node.free();
		node._ended = true;
	} // end method endParse

	/**
	 * Init object info from its template.
	 *
	 * @param objectInfo Object info to initialize.
	 */
	protected void initFromTemplate(LyObjectInfo objectInfo)
	{
		objectInfo.initFromTemplate();
	}

	/**
	 * The purpose of this method is to resolve object references now that all valid object ids are
	 * known.
	 *
	 * @exception SAXExpection If the object found doesn't belong to the appropriate class.
	 */
	protected final void resolveObjectInfoRef(LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Reference to a filter
		if (tag == LyInfoTokens.TAG_FILTER_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyFilter))
				throw new SAXException("BF AppEngine : invalid filter ref : " + idref);
		}
		// Reference to a filter criteria
		else if (tag == LyInfoTokens.TAG_FILTER_CRITERIA_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyFilterCriterion))
				throw new SAXException("BF AppEngine : invalid filter criteria ref : " + idref);
		}
		// Reference to a filter expression
		else if (tag == LyInfoTokens.TAG_EXP)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);
			LyInfo info = getInfo(idref);

			if (!(info instanceof LyFieldInfo))
				throw new SAXException("BF AppEngine : invalid fieldInfo ref : " + idref + " found on " + result);

			((LyFilterExpression)result).setFieldInfo((LyFieldInfo)info);
		}
		// Reference to a filter relation
		else if (tag == LyInfoTokens.TAG_FILTER_RELATION)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);
			LyInfo info = getInfo(idref);

			if (!(info instanceof LyRelationFieldInfo))
				throw new SAXException("BF AppEngine : invalid relation fieldInfo ref : " + idref + " found on " + result);

			((LyFilterRelation)result).setRelationFieldInfo((LyRelationFieldInfo)info);
		}
		// Fill up path informations
		else if (tag == LyInfoTokens.TAG_PATH)
		{
			// target classInfo
			String target = node.getString(LyInfoTokens.ATT_TARGET);
			LyInfo targetClassInfo = getInfo(target);

			if (targetClassInfo != null)
			{
				if (targetClassInfo instanceof LyStructFieldInfo)
					targetClassInfo = ((LyStructFieldInfo)targetClassInfo).getClassInfo(0);
				if (!(targetClassInfo instanceof LyClassInfo))
					throw new SAXException("BF AppEngine : invalid classInfo ref : " + target);
				((LyPath)result).setTarget((LyClassInfo)targetClassInfo);
			}

			// source classInfo
			String source = node.getString(LyInfoTokens.ATT_SOURCE);
			LyInfo sourceClassInfo = getInfo(source);

			if (sourceClassInfo != null)
			{
				if (!(sourceClassInfo instanceof LyClassInfo))
					throw new SAXException("BF AppEngine : invalid classInfo ref : " + source);

				((LyPath)result).setSourceClassInfo((LyClassInfo)sourceClassInfo);
			}

			// target fieldInfo
			String field = node.getString(LyInfoTokens.ATT_FIELD);
			LyInfo fieldInfo = getInfo(field);

			if ((fieldInfo != null) && (!(fieldInfo instanceof LyFieldInfo)))
				throw new SAXException("BF AppEngine : invalid fieldInfo ref : " + field);

			((LyPath)result).setTargetFieldInfo((LyFieldInfo)fieldInfo);

			// field label
			String label = node.getString(LyInfoTokens.ATT_LABEL);
			LyInfo fieldLabel = getInfo(label);

			if ((fieldLabel != null) && (!(fieldLabel instanceof LyInfoLabel)))
				throw new SAXException("BF AppEngine : invalid label ref : " + fieldLabel);

			((LyPath)result).setFieldLabel((LyInfoLabel)fieldLabel);
		}
		// Fill up route informations
		else if (tag == LyInfoTokens.TAG_ROUTE)
		{
			// classInfo
			String target = node.getString(LyInfoTokens.ATT_TARGET);
			LyInfo classInfo = getInfo(target);

			if (!(classInfo instanceof LyClassInfo))
				throw new SAXException("BF AppEngine : invalid classInfo ref : " + target);

			((LyRoute)result).setTarget((LyClassInfo)classInfo);
		}
		// Reference to a route
		else if (tag == LyInfoTokens.TAG_ROUTE_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyRoute))
				throw new SAXException("BF AppEngine : invalid route ref : " + idref);
		}
		// Step : References to a classInfo, to relation fields and filter
		else if (tag == LyInfoTokens.TAG_STEP)
		{
			// classInfo
			String target = node.getString(LyInfoTokens.ATT_TARGET);
			LyInfo classInfo = getInfo(target);

			if (classInfo instanceof LyStructFieldInfo)
				classInfo = ((LyStructFieldInfo)classInfo).getClassInfo(0);
			if (!(classInfo instanceof LyClassInfo))
				throw new SAXException("BF AppEngine : invalid classInfo ref : " + target);

			// relations
			String				    relationRefs = node.getString(LyInfoTokens.ATT_IDREFS);
			LyInfoList<LyFieldInfo> relations = new LyInfoList<LyFieldInfo>();
			String[]			    idrefs = null;
			int					    i, n = 0;
			LyInfo				    relation;

			if (relationRefs != null)
			{
				idrefs = LyString.stringToList(relationRefs, ' ');
				n = idrefs.length;
			}

			for (i = 0; i < n; i++)
			{
				// Check out relation reference
				relation = getInfo(idrefs[i]);
				if ((relation instanceof LyRelationFieldInfo)
						|| ((relation instanceof LyArrayFieldInfo)
							&& (((LyArrayFieldInfo)relation).getFieldInfo()
								instanceof LyRelationFieldInfo)))
					relations.addInfo((LyFieldInfo)relation);
				else
					throw new SAXException("BF AppEngine : invalid relation ref : " + idrefs[i]);
			}

			// filter
			String filterRef = node.getString(LyInfoTokens.ATT_FILTER);
			LyInfo filter = null;

			if (filterRef != null)
			{
				filter = getInfo(filterRef);

				if (!(filter instanceof LyFilter))
					throw new SAXException("BF AppEngine : invalid filter ref : " + filterRef);
			}

			// build LyStep
			result = new LyStep((LyClassInfo)classInfo, relations, (LyFilter)filter);
			completeResult(node, result);
		}
		// Reference to a sort
		else if (tag == LyInfoTokens.TAG_SORT_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LySort))
				throw new SAXException("BF AppEngine : invalid sort ref : " + idref);
		}
		// FielInfo of a sort criteria
		else if (tag == LyInfoTokens.TAG_SORT_CRITERIA)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);
			LyInfo info = getInfo(idref);

			if (!(info instanceof LyFieldInfo))
				throw new SAXException("BF AppEngine : invalid fieldInfo ref : " + idref + " found on " + result);

			((LySortCriterion)result).setFieldInfo((LyFieldInfo)info);
		}
		// Reference to a sort criteria
		else if (tag == LyInfoTokens.TAG_SORT_CRITERIA_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LySortCriterion))
				throw new SAXException("BF AppEngine : invalid sort criteria ref : " + idref);
		}
		// Reference to a control
		else if (tag == LyInfoTokens.TAG_CONTROL_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyControl))
				throw new SAXException("BF AppEngine : invalid control ref : " + idref);
		}
		// Argument
		else if (tag == LyInfoTokens.TAG_ARG)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			if (idref != null)
			{
				LyInfo info = getInfo(idref);

				if (info == null)
					throw new SAXException("BF AppEngine : invalid info ref : " + idref);

				// Add reference to the result list
				((LyArgument)result).setValue(info);
			}
		}
		else
			result = resolveObjectInfoRef(node, tag, result);

		// Update result with actual reference.
		if (result != null)
			node.setResult(result);
	} // end method resolveObjectInfoRef

	/**
	 * The purpose of this method is to resolve object references now that all valid object ids are
	 * known.
	 *
	 * @param     tag    Current tag.
	 * @param     result Current result.
	 * @return    New result.
	 * @exception SAXExpection If the object found doesn't belong to the appropriate class.
	 */
	Object resolveObjectInfoRef(LyInfoNode node, String tag, Object result)
		throws SAXException
	{
		return null;
	}

	/**
	 * Completes parent result with the result of the current node.
	 *
	 * @param     parentResult Result of the parent node.
	 * @param     node         Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	final void completeParentResult(Object parentResult, LyInfoNode node)
		throws SAXException
	{
		// If we have no result, there's nothing to do
		Object result = node.getResult();

		if (result == null)
			return;

		try
		{
			if (parentResult instanceof Vector<?>){
				// When parentResult is a Vector, we simply add this result to it
				@SuppressWarnings("unchecked")
				Vector<Object> temp = (Vector<Object>)parentResult;
				temp.addElement(result);
			}
			else if (parentResult instanceof LyBitSet)
			{
				// When parentResult is a BitSet, we add the tag corresponding
				// to this result into the BitSet
				LyInfoNode      parent = node.getParent();
				LyBitSet.Type   type = null;
				int		        index;
				
				if (parent.getTag() == LyInfoTokens.TAG_A_MARKS)
				    type = LyBitSet.Type.A_MARKS;
				else if (parent.getTag() == LyInfoTokens.TAG_C_MARKS)
	                type = LyBitSet.Type.C_MARKS;
                else if (parent.getTag() == LyInfoTokens.TAG_F_MARKS)
                    type = LyBitSet.Type.F_MARKS;
				
				index = LyInfoTokens.getMarkIndex(type, (String)result, true);
				((LyBitSet)parentResult).set(index);
			}
			else if (parentResult instanceof LyInfo)
				// If parentResult is an Info, we complete it with this result
				completeInfo((LyInfo)parentResult, node);
			else
				completeParentResult(parentResult, node, result);
		}
		catch (Exception e)
		{
			raiseError(e.getMessage());
		}
	} // end method completeParentResult

	/**
	 * Completes parent result with the result of the current node.
	 *
	 * @param     parentResult Result of the parent node.
	 * @param     node         Current node.
	 * @param     result       Result of current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeParentResult(Object parentResult, LyInfoNode node, Object result)
		throws SAXException
	{
	}

	/**
	 * Completes an info with the result of the current node.
	 *
	 * @param     info Info to be completed.
	 * @param     node Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeInfo(LyInfo info, LyInfoNode node)
		throws SAXException
	{
		// Call the appropriate method to complete the info
		if (info instanceof LyObjectInfo)
			completeObjectInfo((LyObjectInfo)info, node);
		else if (info instanceof LyChoiceOption)
			completeChoiceOption((LyChoiceOption)info, node);
		else if (info instanceof LyFilterRelation)
			completeFilterRelation((LyFilterRelation)info, node);
		else if (info instanceof LyPath)
			// A path is a route with a target fieldinfo.
			completeRoute((LyPath)info, node);
		else if (info instanceof LyRoute)
			completeRoute((LyRoute)info, node);
		else if (info instanceof LyStep)
			completeStep((LyStep)info, node);
		else if (info instanceof LyControl)
			completeControl((LyControl)info, node);
	}

	/**
	 * Completes an objectInfo with the result of the current node.
	 *
	 * @param     objectInfo ObjectInfo to be completed.
	 * @param     node       Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeObjectInfo(LyObjectInfo objectInfo, LyInfoNode node)
		throws SAXException
	{
		// Call the appropriate method to complete the objectInfo
		if (objectInfo instanceof LyClassInfo)
			completeClassInfo((LyClassInfo)objectInfo, node);
		else if (objectInfo instanceof LyFieldInfo)
			completeFieldInfo((LyFieldInfo)objectInfo, node);
		else if (objectInfo instanceof LyProject)
			completeProject((LyProject)objectInfo, node);
		else if (objectInfo instanceof LyFilter)
			completeFilter((LyFilter)objectInfo, node);
		else if (objectInfo instanceof LySort)
			completeSort((LySort)objectInfo, node);
	}

	/**
	 * Completes a project with the result of the current node.
	 *
	 * @param project Project to be completed.
	 * @param node    Current node.
	 */
	void completeProject(LyProject project, LyInfoNode node)
	{
	}

	/**
	 * Completes a class info with the result of the current node.
	 *
	 * @param classInfo ClassInfo to be completed.
	 * @param node      Current node.
	 */
	void completeClassInfo(LyClassInfo classInfo, LyInfoNode node)
	{
	}

	/**
	 * Completes an fieldInfo with the result of the current node.
	 *
	 * @param     fieldInfo FieldInfo to be completed.
	 * @param     node      Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeFieldInfo(LyFieldInfo fieldInfo, LyInfoNode node)
		throws SAXException
	{
	}

	/**
	 * Completes a choice option info with the result of the current node.
	 *
	 * @param     choiceOption ChoiceOption to be completed.
	 * @param     node         Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeChoiceOption(LyChoiceOption choiceOption, LyInfoNode node)
	{
	}

	/**
	 * Completes a filter with the result of the current node.
	 *
	 * @param filter Filter to be completed.
	 * @param node   Current node.
	 */
	void completeFilter(LyFilter filter, LyInfoNode node)
	{
	    String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags or results corresponding to filter attributes
		// and complete the given filter.
		if (result instanceof LyFilterExpression)
			filter.addElement((LyFilterExpression)result);
		else if (result instanceof LyFilter)
			filter.addElement((LyFilter)result);
		else if (result instanceof LyFilterRelation)
			filter.addElement((LyFilterRelation)result);
		else if (result instanceof LyFilterCriterion)
			filter.addElement((LyFilterCriterion)result);
		else if (tag == LyInfoTokens.TAG_FORMULA)
		    filter.addFormula((String)result);
	}

	/**
	 * Completes a filter relation with the result of the current node.
	 *
	 * @param filter Filter relation to be completed.
	 * @param node   Current node.
	 */
	void completeFilterRelation(LyFilterRelation filter, LyInfoNode node)
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to filter attributes
		// and complete the given filter.
		if (result instanceof LyFilter)
			filter.setFilter((LyFilter)result);
	}

	/**
	 * Completes a route with the result of the current node.
	 *
	 * @param     route Route to be completed.
	 * @param     node  Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeRoute(LyRoute route, LyInfoNode node)
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to route attributes
		// and complete the given route.
		if (result instanceof LyStep)
			route.addStep((LyStep)result);
		else if (result instanceof LyRoute)
			route.addRoute((LyRoute)result);
	}

	/**
	 * Completes a step with the result of the current node.
	 *
	 * @param     step Step to be completed.
	 * @param     node Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeStep(LyStep step, LyInfoNode node)
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to step attributes
		// and complete the given step.
		if (result instanceof LyFilter)
			step.setFilter((LyFilter)result);
	}

	/**
	 * Completes a sort with the result of the current node.
	 *
	 * @param     sort Sort to be completed.
	 * @param     node Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeSort(LySort sort, LyInfoNode node)
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to sort attributes
		// and complete the given sort.
		if (result instanceof LySortCriterion)
			sort.addCriterion((LySortCriterion)result);
	}

	/**
	 * Completes a control with the result of the current node.
	 *
	 * @param     control Control to be completed.
	 * @param     node    Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeControl(LyControl control, LyInfoNode node)
	{
	    String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags or results corresponding to control attributes
		// and complete the given control.
		if (result instanceof LyArgument)
			control.addArgument((LyArgument)result);
		else if (tag == LyInfoTokens.TAG_FORMULA)
            control.setFormula((String)result);
	}

	/**
	 * Builds a BitSet of a given type. This type defines the set of supported bits.
	 *
	 * @param     type Type of the BitSet (A_MARKS, C_MARKS, F_MARKS).
	 * @param     node Current node.
	 * @return    A bitSet with the standard bits set according to the given type.
	 * @exception IllegalArgumentException If the type is invalid.
	 */
	LyBitSet buildMarks(String type, LyInfoNode node)
	{
		// Check if we have a valid type
		String[]         marksTable;
        LyBitSet.Type    marksType;

		if (type == LyInfoTokens.TAG_A_MARKS)
		{
			marksTable = LyInfoTokens.actionMarks;
			marksType = LyBitSet.Type.A_MARKS;
		}
		else if (type == LyInfoTokens.TAG_C_MARKS)
		{
			marksTable = LyInfoTokens.classInfoMarks;
			marksType = LyBitSet.Type.C_MARKS;
		}
		else if (type == LyInfoTokens.TAG_F_MARKS)
		{
			marksTable = LyInfoTokens.fieldInfoMarks;
			marksType = LyBitSet.Type.F_MARKS;
		}
		else
			throw new IllegalArgumentException("BF AppEngine : invalid tag type " + type);

		// Set bits specified on this node
        int    na = node.getNbAttributes() -1; // marks have always an "operation" attribute that is not a mark
        int    ns = 0;
        int    i, n = marksTable.length;
        BitSet marks = new BitSet(n);

		for (i = 0; (i < n) && (ns < na); i++)
		{
			if (node.getBoolean(marksTable[i]))
            {
				marks.set(i);
                ns++;
            }
		}

		return new LyBitSet(marksType, marks);
	} // end method buildMarks

	/**
	 * Merges objectInfo marks with the given marks depending on the operation specified on this
	 * node. Supported operations are : RESET : replace objectInfo marks by the given marks, OR :
	 * add the given marks to objectInfo marks, AND : AND the given marks with objectInfo marks, XOR
	 * : XOR the given marks with objectInfo marks, CLEAR : clear the given marks on objectInfo
	 * marks.
	 *
	 * @param info  objectInfo for which we complete marks.
	 * @param marks The marks to merge with objectInfo marks.
	 * @param node  Current node.
	 */
	void mergeMarks(LyObjectInfo info, LyBitSet newMarks, LyInfoNode node)
	{
		if (newMarks == null)
			return;

		BitSet        marks = newMarks.getBitSet();
        LyBitSet.Type type = newMarks.getType();

		if (marks == null)
			return;

		BitSet infoMarks = info.getMarks();
		String operation = node.getString(LyInfoTokens.ATT_OPERATION);

		if (operation != null)
			operation = operation.intern();

		if (operation == LyInfoTokens.OPER_BS_RESET)
		{
			infoMarks = resetMarks(type, infoMarks);
			infoMarks.or(marks);
		}
		else if (operation == LyInfoTokens.OPER_BS_OR)
		{
			if (infoMarks == null)
				infoMarks = marks;
			else
				infoMarks.or(marks);
		}
		else if ((operation == LyInfoTokens.OPER_BS_AND) && (infoMarks != null))
			infoMarks.and(marks);
		else if ((operation == LyInfoTokens.OPER_BS_XOR) && (infoMarks != null))
			infoMarks.xor(marks);
		else if ((operation == LyInfoTokens.OPER_BS_CLEAR) && (infoMarks != null))
		{
			if (info instanceof LyFieldInfo)
			{
				if (marks.get(LyFieldInfo.MARK_MAIN))
				{
					marks.set(LyFieldInfo.MARK_TABLE);
					marks.set(LyFieldInfo.MARK_SORT);
					marks.set(LyFieldInfo.MARK_FILTER);
					marks.set(LyFieldInfo.MARK_FIND);
				}
	
				if (marks.get(LyFieldInfo.MARK_CONSULT))
				{
					marks.set(LyFieldInfo.MARK_CREATE_CONSULT);
					marks.set(LyFieldInfo.MARK_SET_CONSULT);
				}
			}
			
			int i, n = marks.size();

			for (i = 0; i < n; i++)
			{
				if (marks.get(i))
					infoMarks.clear(i);
			}
		}
		else if ((operation == LyInfoTokens.OPER_BS_AND) && (infoMarks == null))
			infoMarks = marks; // Add means reset when empty

		info.setMarks(infoMarks);
	} // end method mergeMarks

	/**
	 * Reset marks of the given.
	 */
	protected BitSet resetMarks(LyBitSet.Type type, BitSet marks)
	{
		return new BitSet();
	}

	/**
	 * Final step to check if everything is ok. Calls recursively check method on each result that
	 * is an objectInfo.
	 *
	 * @param  node Current node.
	 * @return False if check fails, true if it succeeds.
	 * @see    LyObjectInfo#check
	 */
	protected final boolean checkResult(Object parent, LyInfoNode node)
	{
		Object result = node.getResult();

		if ((result == null)
				|| (result instanceof LyInclude))
			return true;

		// Check objectInfo associated to this node
		LyObjectInfo	   parentInfo = (LyObjectInfo)parent;
		boolean			   ok = true;
		LyObjectInfo	   objectInfo = null;
		LyInfoList<LyInfo> infos = node.getExtraInfos();

		try
		{
			// Check result
			if (result instanceof LyObjectInfo)
			{
				objectInfo = (LyObjectInfo)result;
				
				if (objectInfo.isLoaded())
					try
					{
						check(objectInfo, parentInfo);
					}
					catch (Exception e)
					{
						getApplication().getEnvironment().writeException(e);
						raiseError(e.getMessage());
					}

				parentInfo = objectInfo;
			}

			checkResult(result);

			// Check extra infos that may be found on the current node
			if (infos != null)
			{
				// Iterate on each info
				int    i, n = infos.getSize();
				LyInfo info;

				for (i = 0; i < n; i++)
				{
					info = infos.getInfo(i);
					if (info instanceof LyObjectInfo)
					{
						objectInfo = (LyObjectInfo)info;
						if (_parser.isValidating() && objectInfo.isLoaded())
							check(objectInfo, parentInfo);
					}
				}
			}
		}
		catch (Exception e)
		{
            LyApplication application = getApplication();
            LyEnvironment env = application.getEnvironment();
            String        message = "Parsing error: '" + e.getMessage()
                + "' (Object : " + objectInfo + ", Parent : " + parent + ")";
            
		    if (_parser.isValidating())
		    {
    			env.err.println("\nLyria checking error  : Fatal error (not recoverable).");
    			env.err.println("     info = " + objectInfo);
    			env.err.println("   parent = " + parent);
    			env.err.println("     Type = " + e.getClass());
    			env.err.println("  Message = " + e.getMessage());
    			env.err.println("");
    
    			if (_parsingErrors == null)
    				_parsingErrors = "";
    
    			_parsingErrors = _parsingErrors + message + "\n";
    			
    			ok = false;
		    }
		    else
		        env.logError(message);
		}

		// Iterate on each child node
		int		   i, n = node.getChildCount();
		LyInfoNode child;

		for (i = 0; i < n; i++)
		{
			child = node.getChild(i);
			ok = ok && checkResult(parentInfo, child);
		}

		return ok;
	} // end method checkResult

	/**
	 * Gets parsing errors if any.
	 *
	 * @return parsing errors if any, null otherwise.
	 */
	public String getParsingErrors()
	{
		if ((_parsingErrors == null) || (_parsingErrors.trim().length() < 0))
			return null;

		return _parsingErrors;
	}

	protected void check(LyObjectInfo objectInfo, LyObjectInfo parentInfo)
	{
		objectInfo.check(parentInfo);
	}

	/**
	 * Extra checks for result. Throws an exception if check fails.
	 *
	 * @param result Checked result.
	 */
	protected void checkResult(Object result)
	{
	}

	/**
	 * Returns the model type of this factory.
	 */
	protected abstract LyTag.Model getModelType();
	
	/**
	 * Raises parsing error.
	 */
	protected void raiseError(String error)
	    throws SAXException
	{
	    if (_parser.isValidating())
	        throw new SAXException(error);
	    else
	        _application.getEnvironment().logError(error);
	}
} // end class LyInfoFactory
