/**
 * LyInfoGuiFactory.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import org.xml.sax.*;

import java.util.*;
import leon.app.*;
import leon.data.*;
import leon.info.misc.*;
import leon.info.save.*;
import leon.misc.*;


/**
 * This class builds instances corresponding to XML tags read from the data description file. Almost
 * all objects built during parsing inherit from LyInfo (except very elementary objects).
 *
 * @see     LyAction
 * @see     LyCompoundAction
 * @see     LyTabAction
 * @see     LyRule
 * @see     LyChartInfo
 * @see     LyPreference
 * @version $Id: LyInfoGuiFactory.java 35875 2012-06-14 09:51:35Z nhamon $
 * @author  Lyria
 */
public class LyInfoGuiFactory extends LyInfoFactory
{
	public static final String __VERSION =
		"$Id: LyInfoGuiFactory.java 35875 2012-06-14 09:51:35Z nhamon $";

    LyInfoList<LyFieldTab>              _applicationTabs;
    LyInfoList<LyFieldGroup>            _applicationGroups;
    
	/**
	 * Constructor of the factory for an application. All objects built by this factory will belong
	 * to the given application.
	 *
	 * @param parser      Parser that builts this factory.
	 * @param application Application for which the factory is built.
	 */
	public LyInfoGuiFactory(LyInfoParser parser, LyApplication application)
	{
		super(parser, application);
        
        _applicationTabs = new LyInfoList<LyFieldTab>();
        _applicationGroups = new LyInfoList<LyFieldGroup>();
	}

	/**
	 * Builds node corresponding to an element read from the XML data description file. The result
	 * associated to the built node depends on the node tag. This method is invoked by SAXParser at
	 * the beginning of each new tag.
	 *
	 * @param     tag  Current tag.
	 * @param     node Current node.
	 * @return    The object built corresponding to the given tag and attributes.
	 * @exception SAXException if the result could not have been built.
	 */
	public Object buildElement(LyInfoNode node, String tag)
		throws SAXException
	{
		// Init result
		Object result = null;

		// Get some common information
		String id = node.getId();
		String type = node.getString(LyInfoTokens.ATT_TYPE);
		String templateId = node.getString(LyInfoTokens.ATT_MODEL);
		LyInfo template = getInfo(templateId);

		// Specific case of forward reference on ACTION_REF
		if (tag == LyInfoTokens.TAG_ACTION_REF)
		{
			String classId = node.getString(LyInfoTokens.ATT_CLASS);

			if (classId != null)
			{
				// The action referenced in ACTION_REF is also used as template
				templateId = node.getString(LyInfoTokens.ATT_IDREF);
				template = getInfo(templateId);

				if ((templateId != null)
						&& (templateId.length() > 0)
						&& (template == null))
					throw new IllegalStateException(
						"BF AppEngine : Forward references to action model are not supported on ACTION_REF. Action ref. with action : '"
						+ templateId
						+ "' and class '" + classId + "'.");
			}
		}

		// Build the appropriate result corresponding to each node.
		if ((tag == LyInfoTokens.TAG_A_MARKS)
				|| (tag == LyInfoTokens.TAG_C_MARKS)
				|| (tag == LyInfoTokens.TAG_F_MARKS))
			result = buildMarks(tag, node);
		else if (tag == LyInfoTokens.TAG_ACTION)
		{
			String  controller = node.getString(LyInfoTokens.ATT_CONTROLLER);
			String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			String  builder = node.getString(LyInfoTokens.ATT_BUILDER);
			String  view = node.getString(LyInfoTokens.ATT_VIEW);
			String  target = node.getString(LyInfoTokens.ATT_TARGET);
			LyInfo  action = getInfo(id);
			boolean abstractAction = node.getBoolean(LyInfoTokens.ATT_ABSTRACT);

			if (action == null)
				result = new LyAction(id, (LyAction)template, controller, behavior, builder, view,
						target);
			else
			{
				if (!(action instanceof LyAction)
						|| (action instanceof LyTabAction)
						|| (action instanceof LyCompoundAction))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not a simple action");
				if (((LyAction)action).isLoaded())
					throw new SAXException("BF AppEngine : action " + id
						+ " already exists");

				((LyAction)action).construct((LyAction)template, controller, behavior, builder,
					view, target);
				result = action;
			}
			if (abstractAction)
				((LyAction)result).setAbstract(true);
		}
		else if (tag == LyInfoTokens.TAG_ACTIONS)
			result = new Vector<LyAction>();
		else if (tag == LyInfoTokens.TAG_ARRAY)
		{
			result = getInfo(id);
			if (!(result instanceof LyArrayFieldInfo))
				throw new SAXException("BF AppEngine : info " + id
					+ " is not an array field info");
		}
		else if (tag == LyInfoTokens.TAG_BEAN)
		{
			String actionType = node.getString(LyInfoTokens.ATT_ACTION);
			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);

			result = new LyFieldBean(javaName, LyInfoTokens.getFieldBeanActionType(actionType));
		}
		else if (tag == LyInfoTokens.TAG_CHART)
		{
			String				  name = node.getString(LyInfoTokens.ATT_NAME);
			String				  classInfoId = node.getString(LyInfoTokens.ATT_CLASS_INFO_ID);
			String				  xField = node.getString(LyInfoTokens.ATT_X_FIELD);
			String				  groupField = node.getString(LyInfoTokens.ATT_GROUP_FIELD);
			LyChartInfo.Type	  typeVarX = LyInfoTokens.getChartInfoTypeVar(node.getString(
						LyInfoTokens.ATT_TYPE_X));
			LyChartInfo.Type	  typeVarY = LyInfoTokens.getChartInfoTypeVar(node.getString(
						LyInfoTokens.ATT_TYPE_Y));
			LyChartInfo.GraphType graphType = LyInfoTokens.getChartInfoGraphType(node.getString(
						LyInfoTokens.ATT_GRAPH_TYPE));
			LyChartInfo.ValType   xVal = LyInfoTokens.getChartInfoValType(node.getString(
						LyInfoTokens.ATT_X_VAL));
			LyChartInfo.ValType   yVal = LyInfoTokens.getChartInfoValType(node.getString(
						LyInfoTokens.ATT_Y_VAL));
			String[]     		  xIntervals = new String[3];

			xIntervals[0] = node.getString(LyInfoTokens.ATT_X_INTS_BEGIN);
			xIntervals[1] = node.getString(LyInfoTokens.ATT_X_INTS_LENGTH);
			xIntervals[2] = node.getString(LyInfoTokens.ATT_X_INTS_UNIT);

			String[]              yIntervals = new String[3];

			yIntervals[0] = node.getString(LyInfoTokens.ATT_Y_INTS_BEGIN);
			yIntervals[1] = node.getString(LyInfoTokens.ATT_Y_INTS_LENGTH);
			yIntervals[2] = node.getString(LyInfoTokens.ATT_Y_INTS_UNIT);

			boolean		   pieBy = node.getBoolean(LyInfoTokens.ATT_PIE_BY);
			String		   title = node.getString(LyInfoTokens.ATT_TITLE);
			String		   xTitle = node.getString(LyInfoTokens.ATT_X_TITLE);
			String		   yTitle = node.getString(LyInfoTokens.ATT_Y_TITLE);

			String[] xAxis = new String[3];

			xAxis[0] = node.getString(LyInfoTokens.ATT_X_AXIS_MIN);
			xAxis[1] = node.getString(LyInfoTokens.ATT_X_AXIS_MAX);
			xAxis[2] = node.getString(LyInfoTokens.ATT_X_AXIS_WIDTH);

			String[] yAxis = new String[3];

			yAxis[0] = node.getString(LyInfoTokens.ATT_Y_AXIS_MIN);
			yAxis[1] = node.getString(LyInfoTokens.ATT_Y_AXIS_MAX);
			yAxis[2] = node.getString(LyInfoTokens.ATT_Y_AXIS_WIDTH);

			boolean xVerticalLabels = node.getBoolean(LyInfoTokens.ATT_X_VERT_LABELS);
			boolean xGrid = node.getBoolean(LyInfoTokens.ATT_X_GRID);
			boolean yGrid = node.getBoolean(LyInfoTokens.ATT_Y_GRID);
			boolean d3Effect = node.getBoolean(LyInfoTokens.ATT_3D_EFFECT);
			boolean showLegend = node.getBoolean(LyInfoTokens.ATT_SHOW_LEGEND);
			boolean showTips = node.getBoolean(LyInfoTokens.ATT_SHOW_TIPS);
			boolean ratios = node.getBoolean(LyInfoTokens.ATT_RATIOS);

			result = new LyChartInfo(id, name, classInfoId, xField, groupField, typeVarY, typeVarX,
					graphType, xVal, xIntervals, yVal, yIntervals, pieBy, title, xTitle, yTitle,
					xAxis, yAxis, xVerticalLabels, xGrid, yGrid, d3Effect, showLegend, showTips,
					ratios);
		}
		else if (tag == LyInfoTokens.TAG_CHOICE)
		{
			result = getInfo(id);
			if (!(result instanceof LyChoiceFieldInfo))
				throw new SAXException("BF AppEngine : info " + id
					+ " is not a choice field info");

			LyChoiceFieldInfo		 choice = (LyChoiceFieldInfo)result;
			String					 sortByString = node.getString(LyInfoTokens.ATT_SORT_BY);
			LyChoiceFieldInfo.SortBy sortBy = LyInfoTokens.getChoiceSortBy(sortByString);
			boolean					 not = node.getBoolean(LyInfoTokens.ATT_NOT);

			if (sortBy != LyChoiceFieldInfo.SortBy.INDEX)
				choice.setSortBy(sortBy);
			if (not)
				choice.setNot(true);
		}
		else if (tag == LyInfoTokens.TAG_CLASS)
		{
			result = getInfo(id);
			if (!(result instanceof LyClassInfo))
				throw new SAXException("BF AppEngine : info " + id
					+ " is not a class info");

			String behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);

			if (behavior != null)
				((LyClassInfo)result).setBehaviorClassName(behavior);
		}
        else if (tag == LyInfoTokens.TAG_COMPOSITION_SAVE)
            result = new LyCompositionFieldSave();
		else if (tag == LyInfoTokens.TAG_COMPOUND_ACTION)
		{
			String  controller = node.getString(LyInfoTokens.ATT_CONTROLLER);
			String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			String  builder = node.getString(LyInfoTokens.ATT_BUILDER);
			String  style = node.getString(LyInfoTokens.ATT_STYLE);
			String  context = node.getString(LyInfoTokens.ATT_CONTEXT);
			String  sizes = node.getString(LyInfoTokens.ATT_SIZES);
			String  clientContext = node.getString(LyInfoTokens.ATT_CLIENT_CONTEXT);
			String  target = node.getString(LyInfoTokens.ATT_TARGET);
			boolean abstractAction = node.getBoolean(LyInfoTokens.ATT_ABSTRACT);
			LyInfo  compoundAction = getInfo(id);

			if (compoundAction == null)
				result = new LyCompoundAction(id, (LyAction)template, controller, behavior, builder,
						LyInfoTokens.getCompoundActionStyle(style), sizes,
						LyInfoTokens.getCompoundActionContext(context),
						LyInfoTokens.getCompoundActionClientContext(clientContext), target);
			else
			{
				if (!(compoundAction instanceof LyCompoundAction))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not a compound action");
				if (((LyCompoundAction)compoundAction).isLoaded())
					throw new SAXException("BF AppEngine : compound action " + id
						+ " already exists");

				((LyCompoundAction)compoundAction).construct((LyAction)template, controller,
					behavior, builder, LyInfoTokens.getCompoundActionStyle(style), sizes,
					LyInfoTokens.getCompoundActionContext(context),
					LyInfoTokens.getCompoundActionClientContext(clientContext), target);
				result = compoundAction;
			}

			if (abstractAction)
				((LyAction)result).setAbstract(true);
		}
		else if (tag == LyInfoTokens.TAG_CONSTRAINT)
		{
			String  typeC = node.getString(LyInfoTokens.ATT_TYPE);
			String  colSpanStr = node.getString(LyInfoTokens.ATT_COL_SPAN);
			int     colSpan = ((colSpanStr == null) ? LyFieldConstraint.DEFAULT_VALUE
													: node.getInt(LyInfoTokens.ATT_COL_SPAN));
			String  rowSpanStr = node.getString(LyInfoTokens.ATT_ROW_SPAN);
			int     rowSpan = ((rowSpanStr == null) ? LyFieldConstraint.DEFAULT_VALUE
													: node.getInt(LyInfoTokens.ATT_ROW_SPAN));
			boolean remainder = node.getBoolean(LyInfoTokens.ATT_REMAINDER);
			String  colWeightStr = node.getString(LyInfoTokens.ATT_COL_WEIGHT);
			int     colWeight = ((colWeightStr == null) ? LyFieldConstraint.DEFAULT_VALUE
														: node.getInt(LyInfoTokens.ATT_COL_WEIGHT));
			String  rowWeightStr = node.getString(LyInfoTokens.ATT_ROW_WEIGHT);
			int     rowWeight = ((rowWeightStr == null) ? LyFieldConstraint.DEFAULT_VALUE
														: node.getInt(LyInfoTokens.ATT_ROW_WEIGHT));
			String  fill = node.getString(LyInfoTokens.ATT_FILL);
			String  hAlign = node.getString(LyInfoTokens.ATT_H_ALIGN);
			String  vAlign = node.getString(LyInfoTokens.ATT_V_ALIGN);

			result = new LyFieldConstraint(LyInfoTokens.getConstraintType(typeC), colSpan, rowSpan,
					remainder, colWeight, rowWeight, LyInfoTokens.getConstraintFill(fill),
					LyInfoTokens.getConstraintHAlign(hAlign),
					LyInfoTokens.getConstraintVAlign(vAlign));
		}
		else if (tag == LyInfoTokens.TAG_CONTROL)
		{
			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			int    priority = node.getInt(LyInfoTokens.ATT_PRIORITY);
			String level = node.getString(LyInfoTokens.ATT_LEVEL);
			LyInfo control = getInfo(id);

			if (control == null)
				result = new LyControl(id, javaName, LyInfoTokens.getControlLevel(level), priority,
						getModelType());
			else
			{
				if (!(control instanceof LyControl))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not a control");
				if (control.isLoaded())
					throw new SAXException("BF AppEngine : control " + id
						+ " already exists");

				((LyControl)control).construct(javaName, LyInfoTokens.getControlLevel(level),
					priority, getModelType());
				result = control;
			}
		}
		else if (tag == LyInfoTokens.TAG_DATE)
		{
			result = getInfo(id);
			if (!(result instanceof LyDateFieldInfo))
				throw new SAXException("BF AppEngine : info " + id
					+ " is not a date field info");
		}
		else if (tag == LyInfoTokens.TAG_FIELD)
		{
			result = getInfo(id);
			if (!(result instanceof LyFieldInfo))
				throw new SAXException("BF AppEngine : info " + id
					+ " is not a field info");
		}
        else if (tag == LyInfoTokens.TAG_FIELD_SAVE)
        {
            String      idref = node.getString(LyInfoTokens.ATT_FIELD);
            LyFieldInfo fieldInfo = (LyFieldInfo)_application.getInfo(idref);
            String      newValue = node.getString(LyInfoTokens.ATT_VALUE);
            String      oldValue = node.getString(LyInfoTokens.ATT_OLD_VALUE);

            result = new LyFieldSave(fieldInfo, newValue, oldValue);
        }
		else if (tag == LyInfoTokens.TAG_FILE)
		{
			result = getInfo(id);
			if (!(result instanceof LyFileFieldInfo))
				throw new SAXException("BF AppEngine : info " + id
					+ " is not a file field info");

			String attachmentHandler = node.getString(LyInfoTokens.ATT_ATTACHMENT_HANDLER);
			String attachmentDirectory = node.getString(LyInfoTokens.ATT_ATTACHMENT_DIRECTORY);

			((LyFileFieldInfo)result).setAttachmentHandler(attachmentHandler);
			((LyFileFieldInfo)result).setAttachmentDirectory(attachmentDirectory);
		}
        else if (tag == LyInfoTokens.TAG_FORM_SAVE)
        {
            String      classId = node.getString(LyInfoTokens.ATT_CLASS);
            LyClassInfo classInfo = _application.getClassInfo(classId);
            String      str = node.getString(LyInfoTokens.ATT_OBJECT_ID);
            LyObjectId  objectId = (str == null ? null : new LyObjectId(classInfo, str));

            result = new LyFormSave(LyInfoTokens.getFormSaveType(type), classInfo, objectId, null);
        }
		else if (tag == LyInfoTokens.TAG_FORMAT)
		{
			String  value = node.getString(LyInfoTokens.ATT_VALUE);
			boolean editable = node.getBoolean(LyInfoTokens.ATT_EDITABLE);
			int     nCols = node.getInt(LyInfoTokens.ATT_NCOLS);
			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);

			result = new LyFormat(value, editable, nCols, javaName);
		}
		else if (tag == LyInfoTokens.TAG_GEOMETRY)
		{
			int x = node.getInt(LyInfoTokens.ATT_X);
			int y = node.getInt(LyInfoTokens.ATT_Y);
			int width = node.getInt(LyInfoTokens.ATT_WIDTH);
			int height = node.getInt(LyInfoTokens.ATT_HEIGHT);
			int minWidth = node.getInt(LyInfoTokens.ATT_MIN_WIDTH);
			int minHeight = node.getInt(LyInfoTokens.ATT_MIN_HEIGHT);

			result = new LyGeometry(x, y, width, height, minWidth, minHeight);
		}
		else if (tag == LyInfoTokens.TAG_GROUP)
		{
			String value = node.getString(LyInfoTokens.ATT_VALUE);
			String state = node.getString(LyInfoTokens.ATT_STATE);

			if (id != null)
				result = new LyFieldGroup(id, value, LyInfoTokens.getFieldGroupState(state));
			else
				result = value;
		}
		else if (tag == LyInfoTokens.TAG_INFO)
		{
			result = getInfo(id);
			if (!(result instanceof LyInfo))
				throw new SAXException("BF AppEngine : unknown info " + id);
		}
		else if (tag == LyInfoTokens.TAG_LABEL)
		{
			String value = node.getString(LyInfoTokens.ATT_VALUE);
			String subType = node.getString(LyInfoTokens.ATT_SUB_TYPE);

			result = new LyInfoLabel(id, value, LyInfoTokens.getInfoLabelType(type), subType);
			((LyInfo)result).setModelType(getModelType());
		}
		else if (tag == LyInfoTokens.TAG_MARK)
			result = node.getString(LyInfoTokens.ATT_NAME);
        else if (tag == LyInfoTokens.TAG_MULTI_COMPOSITION_SAVE)
            result = new LyMultiCompositionFieldSave();
		else if (tag == LyInfoTokens.TAG_NUMBER)
		{
			result = getInfo(id);
			if (!(result instanceof LyNumberFieldInfo))
				throw new SAXException("BF AppEngine : info " + id
					+ " is not a number field info");
		}
		else if (tag == LyInfoTokens.TAG_OPTION)
		{
			result = getInfo(id);
			if (!(result instanceof LyChoiceOption))
				throw new SAXException("BF AppEngine : info " + id
					+ " is not a choice option");

			String  name = node.getString(LyInfoTokens.ATT_NAME);
			String  alias = node.getString(LyInfoTokens.ATT_ALIAS);
			boolean disable = node.getBoolean(LyInfoTokens.ATT_DISABLE);
			boolean table = node.getBoolean(LyInfoTokens.ATT_TABLE);

			((LyChoiceOption)result).completeOption(name, alias, disable, table);
		}
		else if (tag == LyInfoTokens.TAG_OPTIONS)
		{
			if (id == null)
				result = new Vector<LyChoiceOption>();
			else
			{
				result = getInfo(id);
				if (!(result instanceof LyChoiceOptionGroup))
					throw new SAXException("BF AppEngine : info " + id
						+ " is not a choice options group");

				String name = node.getString(LyInfoTokens.ATT_NAME);
				String alias = node.getString(LyInfoTokens.ATT_ALIAS);

				((LyChoiceOptionGroup)result).completeChoiceOptionGroup(name, alias);
			}
		}
		else if (tag == LyInfoTokens.TAG_PARAMETER_DECL)
		{
			String  name = node.getString(LyInfoTokens.ATT_NAME);
			boolean required = node.getBoolean(LyInfoTokens.ATT_REQUIRED);
			String  domain = node.getString(LyInfoTokens.ATT_DOMAIN);

			result = new LyParameterDecl(name, LyInfoTokens.getParameterDeclType(type), required,
					domain);
		}
		else if (tag == LyInfoTokens.TAG_PARAMETER)
			result = new LyParameter(node.getString(LyInfoTokens.ATT_NAME),
					node.getString(LyInfoTokens.ATT_VALUE));
		else if (tag == LyInfoTokens.TAG_PREFERENCE)
		{
	        String  idref = node.getString(LyInfoTokens.ATT_IDREF);
			String  name = node.getString(LyInfoTokens.ATT_NAME);
			String  user = node.getString(LyInfoTokens.ATT_USER);
			int     nbObjects = node.getInt(LyInfoTokens.ATT_NB_OBJECTS);
			boolean defaultPref = node.getBoolean(LyInfoTokens.ATT_DEFAULT);
			boolean defaultForMobilePref = node.getBoolean(LyInfoTokens.ATT_DEFAULT_FOR_MOBILE);
			boolean forPortletPref = node.getBoolean(LyInfoTokens.ATT_FOR_PORTLET);
			boolean readOnly = node.getBoolean(LyInfoTokens.ATT_READ_ONLY);
			String  lastFileName = node.getString(LyInfoTokens.ATT_LAST_FILE_NAME);
			String  columnSizes = node.getString(LyInfoTokens.ATT_COLUMN_SIZES);
			String	showAs = node.getString(LyInfoTokens.ATT_SHOW_AS);

			result = new LyPreference(id, idref, name, null, nbObjects, user, defaultPref, columnSizes);
			((LyPreference)result).setDefaultForMobile(defaultForMobilePref);
			((LyPreference)result).setForPortlet(forPortletPref);
			((LyPreference)result).setReadOnly(readOnly);
			((LyPreference)result).setLastFileName(lastFileName);
			((LyPreference)result).setShowAs(LyInfoTokens.getPreferenceShowAs(showAs));
		}
		else if (tag == LyInfoTokens.TAG_PROJECT)
		{
			if (id.startsWith("_leon")
					|| id.startsWith("__leon"))
			{
				String version = node.getString(LyInfoTokens.ATT__VERSION);

				result = new LyProject(id, (LyProject)template, version, null, null);
			}
			else
			{
				result = getInfo(id);
				if (!(result instanceof LyProject))
					throw new SAXException("BF AppEngine : info " + id + " is not a project");
			}
		}
		else if (tag == LyInfoTokens.TAG_RELATION)
		{
			result = getInfo(id);
			if (!(result instanceof LyRelationFieldInfo))
				throw new SAXException("BF AppEngine : info " + id
					+ " is not a relation field info");

			LyRelationFieldInfo relation = (LyRelationFieldInfo)result;
			boolean			    noContext = node.getBoolean(LyInfoTokens.ATT_NOCONTEXT);
			boolean			    noToolbar = node.getBoolean(LyInfoTokens.ATT_NOTOOLBAR);
			String			    showAs = node.getString(LyInfoTokens.ATT_SHOW_AS);

			relation.completeRelation(noContext, noToolbar,
				LyInfoTokens.getRelationFieldInfoShowAs(showAs));
		}
		else if (tag == LyInfoTokens.TAG_RESOURCE)
			result = new LyActionResource(LyInfoTokens.getResourceType(type),
					node.getString(LyInfoTokens.ATT_NAME), node.getString(LyInfoTokens.ATT_VALUE));
		else if (tag == LyInfoTokens.TAG_RULE)
		{
			String			 str = node.getString(LyInfoTokens.ATT_OPERATION);
			int				 priority = node.getInt(LyInfoTokens.ATT_PRIORITY);
			LyRule.Operation operation = LyInfoTokens.getRuleOperation(str);
			boolean			 onCheck = node.getBoolean(LyInfoTokens.ATT_ON_CHECK);
			String			 value = node.getString(LyInfoTokens.ATT_VALUE);

			result = new LyRule(id, null, priority, operation, onCheck, value, getModelType());
		}
		else if (tag == LyInfoTokens.TAG_STICKER)
		{
			String  user = node.getString(LyInfoTokens.ATT_USER);
			String  objectId = node.getString(LyInfoTokens.ATT_OBJECT_ID);
			String  content = node.getString(LyInfoTokens.ATT_CONTENT);
			int		x = node.getInt(LyInfoTokens.ATT_X);
			int		y = node.getInt(LyInfoTokens.ATT_Y);
			int		width = node.getInt(LyInfoTokens.ATT_WIDTH);
			int		height = node.getInt(LyInfoTokens.ATT_HEIGHT);

			result = new LySticker(id, user, null, objectId, content, x, y, width, height);
		}
		else if (tag == LyInfoTokens.TAG_STRUCT)
		{
			result = getInfo(id);
			if (!(result instanceof LyStructFieldInfo))
				throw new SAXException("BF AppEngine : info " + id
					+ " is not a struct field info");

			String classBehavior = node.getString(LyInfoTokens.ATT_CLASS_BEHAVIOR);

			if (classBehavior != null)
			{
				LyClassInfo classInfo = ((LyStructFieldInfo)result).getClassInfo();

				classInfo.setBehaviorClassName(classBehavior);
			}
		}
		else if (tag == LyInfoTokens.TAG_TAB)
		{
			String value = node.getString(LyInfoTokens.ATT_VALUE);

			if (id != null)
				result = new LyFieldTab(id, value);
			else
				result = value;
		}
		else if (tag == LyInfoTokens.TAG_TAB_ACTION)
		{
			String  controller = node.getString(LyInfoTokens.ATT_CONTROLLER);
			String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			String  builder = node.getString(LyInfoTokens.ATT_BUILDER);
			String  target = node.getString(LyInfoTokens.ATT_TARGET);
			short   clientContext = (short)node.getInt(LyInfoTokens.ATT_CLIENT_CONTEXT, LyTabAction.CONTEXT_ALL);
			boolean abstractAction = node.getBoolean(LyInfoTokens.ATT_ABSTRACT);
			LyInfo  tabAction = getInfo(id);

			if (tabAction == null)
				result = new LyTabAction(id, (LyAction)template, controller, behavior, builder,
						clientContext, target, LyInfoTokens.getTabActionType(type));
			else
			{
				if (!(tabAction instanceof LyTabAction))
					throw new SAXException("BF AppEngine : extern info " + id
						+ " is not a tab action");
				if (((LyTabAction)tabAction).isLoaded())
					throw new SAXException("BF AppEngine : tab action " + id
						+ " already exists");

				((LyTabAction)tabAction).construct((LyAction)template, controller, behavior,
					builder, clientContext, target, LyInfoTokens.getTabActionType(type));
				result = tabAction;
			}

			if (abstractAction)
				((LyAction)result).setAbstract(true);
		}
		else if (tag == LyInfoTokens.TAG_TEXT)
		{
			result = getInfo(id);
			if (!(result instanceof LyTextFieldInfo))
				throw new SAXException("BF AppEngine : info " + id
					+ " is not a text field info");
		}
		else if (tag == LyInfoTokens.TAG_UNIT)
		{
			if (id == null)
			{
				String value = node.getString(LyInfoTokens.ATT_VALUE);

				result = new LyUnit(null, value);
			}
			else
			{
				result = getInfo(id);
				if (!(result instanceof LyUnit))
					throw new SAXException("BF AppEngine : invalid unit " + id);

				String  value = node.getString(LyInfoTokens.ATT_VALUE);
				String  name = node.getString(LyInfoTokens.ATT_NAME);
				String  alias = node.getString(LyInfoTokens.ATT_ALIAS);
				boolean print = node.getBoolean(LyInfoTokens.ATT_PRINT);

				((LyUnit)result).completeUnit(value, name, alias, print);
			}
		}
		else if (tag == LyInfoTokens.TAG_UNITS)
		{
			result = getInfo(id);
			if (!(result instanceof LyUnits))
				throw new SAXException("BF AppEngine : invalid units list " + id);

			String  name = node.getString(LyInfoTokens.ATT_NAME);
			boolean convert = node.getBoolean(LyInfoTokens.ATT_CONVERT);

			((LyUnits)result).setName(name);
			((LyUnits)result).setConvert(convert);
		}
		else if (tag == LyInfoTokens.TAG_PATH_REF)
		{
			id = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(id);

			if (!(result instanceof LyPath))
				throw new SAXException("BF AppEngine : info " + id + " is not a valid path");
		}
		else
			result = super.buildElement(node, tag);

		return result;
	} // end method buildElement

	/**
	 * Init object info from its template.
	 *
	 * @param objectInfo Object info to initialize.
	 */
	protected void initFromTemplate(LyObjectInfo objectInfo)
	{
		if ((objectInfo instanceof LyClassInfo)
				|| (objectInfo instanceof LyFieldInfo))
			objectInfo.initFromTemplate(LyObjectInfo.PART_GUI);
		else
			super.initFromTemplate(objectInfo);
	}

	protected void check(LyObjectInfo objectInfo, LyObjectInfo parentInfo)
	{
		if ((objectInfo instanceof LyClassInfo)
				|| (objectInfo instanceof LyFieldInfo))
			objectInfo.check(parentInfo, LyObjectInfo.PART_GUI);
		else
			super.check(objectInfo, parentInfo);
	}

	/**
	 * The purpose of this method is to resolve object references now that all valid object ids are
	 * known.
	 *
	 * @param     tag    Current tag.
	 * @param     result Current result.
	 * @return    New result.
	 * @exception SAXExpection If the object found doesn't belong to the appropriate class.
	 */
	Object resolveObjectInfoRef(LyInfoNode node, String tag, Object result)
		throws SAXException
	{
		// Compute idrefs on tags ACTIONS, FIELDS ans OPTIONS
		if ((tag == LyInfoTokens.TAG_ACTIONS)
				|| (tag == LyInfoTokens.TAG_FIELDS)
				|| (tag == LyInfoTokens.TAG_OPTIONS))
		{
			if (!(result instanceof Vector<?>)) // Nothing to do
				return null;

			@SuppressWarnings("unchecked")
			Vector<LyInfo> list = (Vector<LyInfo>)result;
			String		   idrefsAtt = node.getString(LyInfoTokens.ATT_IDREFS);
			String		   operation = node.getString(LyInfoTokens.ATT_OPERATION);
			String[]	   idrefs = null;
			int			   i, n = 0;
			LyInfo		   info;

			if (idrefsAtt != null)
			{
				idrefs = LyString.stringToList(idrefsAtt, ' ');
				n = ((idrefs == null) ? 0 : idrefs.length);
			}

			for (i = 0; i < n; i++)
			{
				// Check out class reference
				info = getInfo(idrefs[i]);
				if ((tag == LyInfoTokens.TAG_ACTIONS)
						&& !(info instanceof LyAction))
				    raiseError("BF AppEngine : invalid action ref : " + idrefs[i]);
				else if ((tag == LyInfoTokens.TAG_FIELDS)
						&& !(info instanceof LyFieldInfo))
				    raiseError("BF AppEngine : invalid fieldInfo ref : " + idrefs[i]);
				else if ((tag == LyInfoTokens.TAG_OPTIONS)
						&& !(info instanceof LyChoiceOption))
				    raiseError("BF AppEngine : invalid option ref : " + idrefs[i]);

				// Add reference to the result list
				list.addElement(info);
				if (operation != LyInfoTokens.OPER_LIST_REMOVE)
					node.addExtraInfo(info);
			}
		}
		else if (tag == LyInfoTokens.TAG_PARAMETER)
		{
			if (result == null) // Nothing to do
				return null;

			String idrefsAtt = node.getString(LyInfoTokens.ATT_IDREFS);
			LyInfo info;

			if (idrefsAtt != null)
			{
				String[]		   idrefs = LyString.stringToList(idrefsAtt, ' ');
				int				   i, n = idrefs.length;
				LyInfoList<LyInfo> list = new LyInfoList<LyInfo>();

				for (i = 0; i < n; i++)
				{
					// Check out class reference
					info = getInfo(idrefs[i]);
					if (info == null)
						throw new SAXException("BF AppEngine : invalid info ref : " + idrefs[i]);

					// Add reference to the result list
					list.addInfo(info);
				}

				((LyParameter)result).setValue(list);
			}
			else
			{
				String idref = node.getString(LyInfoTokens.ATT_IDREF);

				if (idref != null)
				{
					info = getInfo(idref);
					if (info == null)
						throw new SAXException("BF AppEngine : invalid info ref : " + idref);

					// Add reference to the result list
					((LyParameter)result).setValue(info);
				}
			}
		}
		// Reference to a fieldInfo
		else if (tag == LyInfoTokens.TAG_FIELD_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyFieldInfo))
				throw new SAXException("BF AppEngine : invalid fieldInfo ref : " + idref);
		}
		// Rule : Reference to a fieldInfo
		else if (tag == LyInfoTokens.TAG_RULE)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);
			LyInfo info = getInfo(idref);

			if ((info instanceof LyFieldInfo)
					|| (info instanceof LyClassInfo))
				((LyRule)result).setObjectInfo((LyObjectInfo)info);
			else
				throw new SAXException("BF AppEngine : invalid info ref : " + idref
					+ " found on " + result);
		}
		// Reference to a rule
		else if (tag == LyInfoTokens.TAG_RULE_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyRule))
				throw new SAXException("BF AppEngine : invalid rule ref : " + idref);
		}
		// Fill up rootAction on Project
		else if (tag == LyInfoTokens.TAG_PROJECT)
		{
			String idref = node.getString(LyInfoTokens.ATT_ROOT_ACTION);

			if (idref != null)
			{
				LyInfo action = getInfo(idref);

				if (!(action instanceof LyAction))
					throw new SAXException("BF AppEngine : invalid action ref : " + idref);

				((LyProject)result).setRootAction((LyAction)action);
			}
		}
		// Reference to an action
		else if (tag == LyInfoTokens.TAG_ACTION_REF)
		{
			String id;
			String idref;
			LyInfo info;

			id = node.getString(LyInfoTokens.ATT_ID);
			idref = node.getString(LyInfoTokens.ATT_IDREF);

			info = getInfo(idref);
			if (!(info instanceof LyAction))
				throw new SAXException("BF AppEngine : invalid action ref : " + idref);

			LyAction action = (LyAction)info;

			idref = node.getString(LyInfoTokens.ATT_CLASS);
			info = null;

			if (idref != null)
			{
				info = getInfo(idref);
				if (!(info instanceof LyClassInfo))
					throw new SAXException("BF AppEngine : invalid class ref : " + idref);
			}

			if (info == null)
				result = action;
			else if (info instanceof LyClassInfo)
			{
				result = new LyAction(id, action, (LyClassInfo)info, null);
				completeResult(node, result);
			}
		}
		else if (tag == LyInfoTokens.TAG_ACTION)
		{
			String idref = node.getString(LyInfoTokens.ATT_DEFAULT_ACTION);

			if (idref != null)
			{
				LyInfo action = getInfo(idref);

				if (!(action instanceof LyAction))
					throw new SAXException("BF AppEngine : invalid action ref : " + idref);

				((LyAction)result).setDefaultAction((LyAction)action);
			}
		}
		else if (tag == LyInfoTokens.TAG_PREFERENCE)
		{
		    String   idref = node.getString(LyInfoTokens.ATT_FIXED_COLUMN);
			LyInfo   info = getInfo(idref);

			if ((info != null) && !(info instanceof LyFieldInfo))
				throw new SAXException("BF AppEngine : invalid field info on preference : " + idref);
			((LyPreference)result).setFixedColumn((LyFieldInfo)info);
		}
		else if (tag == LyInfoTokens.TAG_STICKER)
		{
		    String   idref = node.getString(LyInfoTokens.ATT_ACTION);
			LyInfo   info = getInfo(idref);

			if ((info != null) && !(info instanceof LyAction))
				throw new SAXException("Leonardi : invalid action on sticker : " + idref);
			((LySticker)result).setAction((LyAction)info);
		}
		// Reference to a field group
		else if (tag == LyInfoTokens.TAG_GROUP_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyFieldGroup))
				throw new SAXException("BF AppEngine : invalid field group ref : " + idref);
		}
		// Reference to a field tab
		else if (tag == LyInfoTokens.TAG_TAB_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyFieldTab))
				throw new SAXException("BF AppEngine : invalid field tab ref : " + idref);
		}
        else if (tag == LyInfoTokens.TAG_CONSTRAINT)
        {
            String  idref = node.getString(LyInfoTokens.ATT_ACTION);
           
            if (idref != null)
            {
		        LyInfo  info = getInfo(idref);
		        
		        if (info != null && !(info instanceof LyAction))
		        	throw new SAXException("BF AppEngine : invalid action on constraint : " + idref);

		        ((LyFieldConstraint)result).setAction((LyAction)info);
            }
        }

		return result;
	} // end method resolveObjectInfoRef

	/**
	 * Completes parent result with the result of the current node.
	 *
	 * @param     parentResult Result of the parent node.
	 * @param     node         Current node.
	 * @param     result       Result of current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeParentResult(Object parentResult, LyInfoNode node, Object result)
		throws SAXException
	{
		if (parentResult instanceof LyParameter)
			// If parentResult is a parameter, we complete it with this result
			((LyParameter)parentResult).setValue(result);
		else if (parentResult instanceof LyFieldBean)
			// If parentResult is a field bean, we complete it with this result
			completeFieldBean((LyFieldBean)parentResult, node);
	    else if (parentResult instanceof LyFormSave)
            completeFormSave((LyFormSave)parentResult, node);
        else if (parentResult instanceof LyCompositionFieldSave)
            completeCompositionSave((LyCompositionFieldSave)parentResult, node);
        else if (parentResult instanceof LyMultiCompositionFieldSave)
            completeMultiCompositionSave((LyMultiCompositionFieldSave)parentResult, node);
	}

	/**
	 * Completes an info with the result of the current node.
	 *
	 * @param     info Info to be completed.
	 * @param     node Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeInfo(LyInfo info, LyInfoNode node)
		throws SAXException
	{
		// Call the appropriate method to complete the info
		if (info instanceof LyChoiceOptionGroup)
			completeChoiceOptionGroup((LyChoiceOptionGroup)info, node);
		else if (info instanceof LyInfoLabel)
			completeInfoLabel((LyInfoLabel)info, node);
		else if (info instanceof LyRule)
			completeRule((LyRule)info, node);
		else if (info instanceof LyUnit)
			completeUnit((LyUnit)info, node);
		else if (info instanceof LyFieldGroup)
			completeFieldGroup((LyFieldGroup)info, node);
		else if (info instanceof LyFieldTab)
			completeFieldTab((LyFieldTab)info, node);
		else
			super.completeInfo(info, node);
	}

	/**
	 * Completes an objectInfo with the result of the current node.
	 *
	 * @param     objectInfo ObjectInfo to be completed.
	 * @param     node       Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeObjectInfo(LyObjectInfo objectInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to objectInfo attributes
		// and complete the given objectInfo.
		if (tag == LyInfoTokens.TAG_NAME)
			objectInfo.setName((String)result);
		else if (tag == LyInfoTokens.TAG_ALIAS)
			objectInfo.setAlias((String)result);
		else if ((tag == LyInfoTokens.TAG_GROUP)
				&& (result instanceof String))
			objectInfo.setGroup((String)result);
		else if (tag == LyInfoTokens.TAG_PREFIX)
			objectInfo.setFilePrefix((String)result);
		else if (tag == LyInfoTokens.TAG_TIP)
			objectInfo.setTip((String)result);
		else if (tag == LyInfoTokens.TAG_DATA)
			objectInfo.putData(node.getString(LyInfoTokens.ATT_NAME), result, getModelType());

		// For other tags, call the appropriate method to complete the objectInfo
		else if (objectInfo instanceof LyAction)
			completeAction((LyAction)objectInfo, node);
		else if (objectInfo instanceof LyChartInfo)
			completeChartInfo((LyChartInfo)objectInfo, node);
		else if (objectInfo instanceof LyPreference)
			completePreference((LyPreference)objectInfo, node);
		else
			super.completeObjectInfo(objectInfo, node);
	} // end method completeObjectInfo

	/**
	 * Completes a project with the result of the current node.
	 *
	 * @param project Project to be completed.
	 * @param node    Current node.
	 */
	void completeProject(LyProject project, LyInfoNode node)
	{
		// Add the current result to project by using the right method
		Object result = node.getResult();

		if (result instanceof LyInfo)
			project.addInfo((LyInfo)result);
	}

	/**
	 * Completes an action with the result of the current node.
	 *
	 * @param     action Action to be completed.
	 * @param     node   Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeAction(LyAction action, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();
		String tag = node.getTag();

		// Check for tags corresponding to action attributes
		// and complete the given action.
		if (tag == LyInfoTokens.TAG_A_MARKS)
			mergeMarks(action, (LyBitSet)result, node);
		else if (tag == LyInfoTokens.TAG_MENU)
			action.setMenu((String)result);
		else if (tag == LyInfoTokens.TAG_MIN)
			action.setMin(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_MAX)
			action.setMax(LyInfoTokens.parseInt((String)result));
		else if (result instanceof LyFilter)
			action.setFilter((LyFilter)result);
		else if (tag == LyInfoTokens.TAG_PARAMETER_DECL)
			action.addParameterDecl((LyParameterDecl)result);
		else if (tag == LyInfoTokens.TAG_PARAMETER)
			action.addParameter((LyParameter)result);
		else if (tag == LyInfoTokens.TAG_RESOURCE)
			action.addResource((LyActionResource)result);
		else if (tag == LyInfoTokens.TAG_SHORTCUT)
			action.setShortcut((String)result);
		else if (tag == LyInfoTokens.TAG_GEOMETRY)
			action.setGeometry((LyGeometry)result);
		else if (tag == LyInfoTokens.TAG_ACTIONS)
		{
			String			 operation = node.getString(LyInfoTokens.ATT_OPERATION);
			int				 pos = node.getInt(LyInfoTokens.ATT_POSITION);
			
			@SuppressWarnings("unchecked")
			Vector<LyAction> actions = (Vector<LyAction>)result;
			int				 i, n = actions.size();

			if (operation != null)
				operation = operation.intern();

			if (operation == LyInfoTokens.OPER_LIST_RESET)
			{
				action.resetActions();
				pos = 0;
			}

			if ((operation == LyInfoTokens.OPER_LIST_ADD)
					|| (operation == LyInfoTokens.OPER_LIST_RESET))
			{
				for (i = 0; i < n; i++)
					action.addAction(actions.elementAt(i), ((pos < 0) ? -1 : (pos + i)));
			}
			else if (operation == LyInfoTokens.OPER_LIST_REMOVE)
			{
				for (i = 0; i < n; i++)
					action.removeAction(actions.elementAt(i));
				node.setResult(null);
			}
		}
		else if (action instanceof LyCompoundAction)
			completeCompoundAction(tag, (LyCompoundAction)action, result);
		else if (action instanceof LyTabAction)
			completeTabAction(tag, (LyTabAction)action, result);
	} // end method completeAction

	/**
	 * Completes a compound action with the result of the current node.
	 *
	 * @param     compoundAction Compound action to be completed.
	 * @param     node           Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeCompoundAction(String tag, LyCompoundAction compoundAction, Object result)
		throws SAXException
	{
		// Check for tags corresponding to compound action attributes
		if (result instanceof LyAction)
			compoundAction.addSubAction((LyAction)result);
	}

	/**
	 * Completes a tab action with the result of the current node.
	 *
	 * @param     tabAction Tab action to be completed.
	 * @param     node      Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeTabAction(String tag, LyTabAction tabAction, Object result)
		throws SAXException
	{
		// Check for tags corresponding to tab action attributes
		if (result instanceof LyAction)
			tabAction.addTabAction((LyAction)result);
	}

	/**
	 * Completes an fieldInfo with the result of the current node.
	 *
	 * @param     fieldInfo FieldInfo to be completed.
	 * @param     node      Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeFieldInfo(LyFieldInfo fieldInfo, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();
		String tag = node.getTag();

		// Check for tags corresponding to objectInfo attributes
		// and complete the given fieldInfo.
		if (tag == LyInfoTokens.TAG_F_MARKS)
			mergeMarks(fieldInfo, (LyBitSet)result, node);
		else if ((tag == LyInfoTokens.TAG_CONTROL)
				|| (tag == LyInfoTokens.TAG_CONTROL_REF))
			fieldInfo.addControl((LyControl)result);
        else if ((tag == LyInfoTokens.TAG_TAB)
                || (tag == LyInfoTokens.TAG_TAB_REF))
        {
            if (result instanceof LyFieldTab)
            {
                fieldInfo.setFieldTab((LyFieldTab)result);
                _applicationTabs.addInfo((LyFieldTab)result);
            }
            else
            {
                // Try to get the same tab 
                LyFieldTab tab = _applicationTabs.getInfo((String) result, true);
                
                // Build new tab and add it to the application
                if (tab == null)
                {
                    tab = new LyFieldTab(null, (String) result);
                    tab.setApplication(_application);
                    tab.setParent(fieldInfo);
                    _applicationTabs.addInfo(tab);
                }
                
                fieldInfo.setFieldTab(tab);
            }
        }
        else if ((tag == LyInfoTokens.TAG_GROUP)
                || (tag == LyInfoTokens.TAG_GROUP_REF))
        {
            if (result instanceof LyFieldGroup)
            {
                fieldInfo.setFieldGroup((LyFieldGroup)result);
                _applicationGroups.addInfo((LyFieldGroup)result);
            }
            else
            {
                // Try to get the same group
                LyFieldGroup group = _applicationGroups.getInfo((String) result, true);
                
                // Build new group and add it to the application
                if (group == null)
                {
                    group = new LyFieldGroup(null, (String) result, null);
                    group.setApplication(_application);
                    group.setParent(fieldInfo);
                    _applicationGroups.addInfo(group);
                }
                
                fieldInfo.setFieldGroup(group);
            }
        }
		else if (tag == LyInfoTokens.TAG_BEAN)
			fieldInfo.addBean((LyFieldBean)result);
		else if (tag == LyInfoTokens.TAG_LABEL)
			fieldInfo.addLabel((LyInfoLabel)result);
		else if (result instanceof LyFieldConstraint)
			fieldInfo.addConstraint((LyFieldConstraint)result);
		// For other tags, call the appropriate method to complete the fieldInfo
		else if (fieldInfo instanceof LyChoiceFieldInfo)
			completeChoiceFieldInfo((LyChoiceFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyDateFieldInfo)
			completeDateFieldInfo((LyDateFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyNumberFieldInfo)
			completeNumberFieldInfo((LyNumberFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyStructFieldInfo)
			completeStructFieldInfo((LyStructFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyRelationFieldInfo)
			completeRelationFieldInfo((LyRelationFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyTextFieldInfo)
			completeTextFieldInfo((LyTextFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyFileFieldInfo)
			completeFileFieldInfo((LyFileFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyArrayFieldInfo)
			completeArrayFieldInfo((LyArrayFieldInfo)fieldInfo, node);
	} // end method completeFieldInfo

	/**
	 * Completes a date field info with the result of the current node.
	 *
	 * @param     dateFieldInfo DateFieldInfo to be completed.
	 * @param     node          Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeDateFieldInfo(LyDateFieldInfo dateFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to dateFieldInfo attributes
		// and complete the given dateFieldInfo.
		if (tag == LyInfoTokens.TAG_FORMAT)
			dateFieldInfo.setDisplayFormat(((LyFormat)result).getValue());
	}

	/**
	 * Completes a number field info with the result of the current node.
	 *
	 * @param     numberFieldInfo NumberFieldInfo to be completed.
	 * @param     node            Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeNumberFieldInfo(LyNumberFieldInfo numberFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String				   tag = node.getTag();
		Object				   result = node.getResult();

		// Check for tags corresponding to numberFieldInfo attributes
		// and complete the given numberFieldInfo.
		LyNumberFieldInfo.Type type = numberFieldInfo.getType();

		if (tag == LyInfoTokens.TAG_FUNCTION)
			numberFieldInfo.setFunction(LyInfoTokens.getNumberFieldInfoFunction((String)result));
		else if (tag == LyInfoTokens.TAG_PITCH)
		{
			Number number = LyInfoTokens.parseNumber(type, (String)result);

			numberFieldInfo.setPitch(number);
		}
		else if (tag == LyInfoTokens.TAG_DECIMALS)
			numberFieldInfo.setDecimals(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_UNIT)
			numberFieldInfo.setFieldUnit((LyUnit)result);
	}

	/**
	 * Completes a text field info with the result of the current node.
	 *
	 * @param     textFieldInfo TextFieldInfo to be completed.
	 * @param     node          Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeTextFieldInfo(LyTextFieldInfo textFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to textFieldInfo attributes
		// and complete the given textFieldInfo.
		if (tag == LyInfoTokens.TAG_NROWS)
			textFieldInfo.setNRows(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_NCOLS)
			textFieldInfo.setNCols(LyInfoTokens.parseInt((String)result));
	}

	/**
	 * Completes a file field info with the result of the current node.
	 *
	 * @param     fileFieldInfo FileFieldInfo to be completed.
	 * @param     node          Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeFileFieldInfo(LyFileFieldInfo fileFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to fileFieldInfo attributes
		// and complete the given fileFieldInfo.
		if (tag == LyInfoTokens.TAG_NROWS)
			fileFieldInfo.setNRows(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_NCOLS)
			fileFieldInfo.setNCols(LyInfoTokens.parseInt((String)result));
	}

	/**
	 * Completes a choice field info with the result of the current node.
	 *
	 * @param     choiceFieldInfo ChoiceFieldInfo to be completed.
	 * @param     node            Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeChoiceFieldInfo(LyChoiceFieldInfo choiceFieldInfo, LyInfoNode node)
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to choiceFieldInfo attributes
		// and complete the given choiceFieldInfo.
		if (tag == LyInfoTokens.TAG_NCOLS)
			choiceFieldInfo.setNCols(LyInfoTokens.parseInt((String)result));
		else if ((tag == LyInfoTokens.TAG_OPTIONS)
				&& (result instanceof Vector<?>))
		{
			LyChoiceOptionGroup group = choiceFieldInfo.getChoiceOptionGroup();
			
			Vector<?>		    items = (Vector<?>)result;
			int				    i, n = items.size();
			Object			    item;

			for (i = 0; i < n; i++)
			{
				item = items.elementAt(i);
				if ((item instanceof LyChoiceItem)
						&& !group.contains((LyChoiceItem)item, false))
					throw new IllegalStateException("BF AppEngine : choice item "
						+ item + " was not declared in choice field info "
						+ choiceFieldInfo + " in your data model");
			}
		}
	} // end method completeChoiceFieldInfo

	/**
	 * Completes a choice option info with the result of the current node.
	 *
	 * @param     choiceOption ChoiceOption to be completed.
	 * @param     node         Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeChoiceOption(LyChoiceOption choiceOption, LyInfoNode node)
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to choiceOption attributes
		// and complete the given option.
		if ((tag == LyInfoTokens.TAG_RULE)
				|| (tag == LyInfoTokens.TAG_RULE_REF))
			choiceOption.addRule((LyRule)result);
		else if (tag == LyInfoTokens.TAG_LABEL)
			choiceOption.addLabel((LyInfoLabel)result);
	}

	/**
	 * Completes a choice options group info with the result of the current node.
	 *
	 * @param     choiceOptionGroup ChoiceOptionGroup to be completed.
	 * @param     node              Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeChoiceOptionGroup(LyChoiceOptionGroup choiceOptionGroup, LyInfoNode node)
	{
		// Check for tags corresponding to choiceOption attributes
		// and complete the given option.
		Object result = node.getResult();

		if ((result instanceof LyChoiceItem)
				&& !choiceOptionGroup.contains((LyChoiceItem)result, true))
			throw new IllegalStateException("BF AppEngine : choice item "
				+ result + " was not declared in choice options group "
				+ choiceOptionGroup + " in your data model");
	}

	/**
	 * Completes a relation field info with the result of the current node.
	 *
	 * @param     relationFieldInfo RelationFieldInfo to be completed.
	 * @param     node              Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeRelationFieldInfo(LyRelationFieldInfo relationFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to relationFieldInfo attributes
		// and complete the given relationFieldInfo.
		if (tag == LyInfoTokens.TAG_NROWS)
			relationFieldInfo.setNRows(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_NCOLS)
            relationFieldInfo.setNCols(LyInfoTokens.parseInt((String)result));
	}

	/**
	 * Completes an array field info with the result of the current node.
	 *
	 * @param     arrayFieldInfo ArrayFieldInfo to be completed.
	 * @param     node           Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeArrayFieldInfo(LyArrayFieldInfo arrayFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to arrayFieldInfo attributes
		// and complete the given arrayFieldInfo.
		if (tag == LyInfoTokens.TAG_NROWS)
			arrayFieldInfo.setNRows(LyInfoTokens.parseInt((String)result));
	}

	/**
	 * Completes a struct field info with the result of the current node.
	 *
	 * @param     relationFieldInfo structFieldInfo to be completed.
	 * @param     node              Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeStructFieldInfo(LyStructFieldInfo structFieldInfo,
		LyInfoNode node)
		throws SAXException
	{
		String	    tag = node.getTag();
		Object	    result = node.getResult();
		LyClassInfo classInfo = structFieldInfo.getClassInfo(0);

		if (tag == LyInfoTokens.TAG_FIELDS)
		{
			String operation = node.getString(LyInfoTokens.ATT_OPERATION);

			if ((getModelType() == LyTag.Model.GUI)
			    && (operation != LyInfoTokens.OPER_LIST_REMOVE))
			{
				// Check that all fields really belong to the given classInfo
				Vector<?>	    fields = (Vector<?>)result;
				int		    i, n = fields.size();
				LyFieldInfo fieldInfo;

				for (i = 0; i < n; i++)
				{
					fieldInfo = (LyFieldInfo)fields.elementAt(i);
					if (!classInfo.contains(fieldInfo))
						throw new IllegalStateException("BF AppEngine : fieldInfo "
							+ fieldInfo + " was not declared in struct "
							+ structFieldInfo + " in your data model");
					classInfo.removeFieldInfo(fieldInfo);
				}

				// Add fields in the right order.
				for (i = 0; i < n; i++)
				{
					fieldInfo = (LyFieldInfo)fields.elementAt(i);
					classInfo.addFieldInfo(fieldInfo, -1);
				}
			}
		}
		else if ((tag == LyInfoTokens.TAG_RULE)
				|| (tag == LyInfoTokens.TAG_RULE_REF))
			classInfo.addRule((LyRule)result);
	} // end method completeStructFieldInfo

	/**
	 * Completes a class info with the result of the current node.
	 *
	 * @param classInfo ClassInfo to be completed.
	 * @param node      Current node.
	 */
	void completeClassInfo(LyClassInfo classInfo, LyInfoNode node)
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to classInfo attributes
		// and complete the given classInfo.
		if (tag == LyInfoTokens.TAG_C_MARKS)
			mergeMarks(classInfo, (LyBitSet)result, node);
		else if (tag == LyInfoTokens.TAG_ACTIONS)
		{
			String			 operation = node.getString(LyInfoTokens.ATT_OPERATION);
			int				 pos = node.getInt(LyInfoTokens.ATT_POSITION);
			
			@SuppressWarnings("unchecked")
			Vector<LyAction> actions = (Vector<LyAction>)result;
			int				 i, n = actions.size();

			if (operation != null)
				operation = operation.intern();

			if (operation == LyInfoTokens.OPER_LIST_RESET)
			{
				classInfo.resetActions();
				pos = 0;
			}

			if ((operation == LyInfoTokens.OPER_LIST_ADD)
					|| (operation == LyInfoTokens.OPER_LIST_RESET))
			{
				for (i = 0; i < n; i++)
					classInfo.addAction(actions.elementAt(i), ((pos < 0) ? -1 : (pos + i)));
			}
			else if (operation == LyInfoTokens.OPER_LIST_REMOVE)
			{
				for (i = 0; i < n; i++)
					classInfo.removeAction(actions.elementAt(i));
				node.setResult(null);
			}
		}
		else if ((tag == LyInfoTokens.TAG_SORT)
				|| (tag == LyInfoTokens.TAG_SORT_REF))
			classInfo.setSort((LySort)result);
		else if (tag == LyInfoTokens.TAG_FIELDS)
		{
			String operation = node.getString(LyInfoTokens.ATT_OPERATION);

			if ((getModelType() == LyTag.Model.GUI)
					&& (operation != LyInfoTokens.OPER_LIST_REMOVE))
			{
				// Check that all fields really belong to the given classInfo
				Vector<?>	fields = (Vector<?>)result;
				int		    i, n = fields.size();
				LyFieldInfo fieldInfo;

				for (i = 0; i < n; i++)
				{
					fieldInfo = (LyFieldInfo)fields.elementAt(i);
					classInfo.moveFieldInfo(fieldInfo, -1);
				}
			}
		}
		else if ((tag == LyInfoTokens.TAG_CONTROL)
				|| (tag == LyInfoTokens.TAG_CONTROL_REF))
			classInfo.addControl((LyControl)result);
		else if ((tag == LyInfoTokens.TAG_RULE)
				|| (tag == LyInfoTokens.TAG_RULE_REF))
			classInfo.addRule((LyRule)result);
		else if (tag == LyInfoTokens.TAG_LABEL)
			classInfo.addLabel((LyInfoLabel)result);
	} // end method completeClassInfo

	/**
	 * Completes a chartInfo with the result of the current node.
	 *
	 * @param     chartInfo ChartInfo to be completed.
	 * @param     node      Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeChartInfo(LyChartInfo chartInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags or results corresponding to chartInfo attributes
		// and complete the given chartInfo.
		if (tag == LyInfoTokens.TAG_YFIELD)
			chartInfo.addYField((String)result);
		else if (tag == LyInfoTokens.TAG_PREFERENCE)
			chartInfo.setPreference((LyPreference)result);
	}

	/**
	 * Completes a preference with the result of the current node.
	 *
	 * @param     preference preference to be completed.
	 * @param     node       Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completePreference(LyPreference preference, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to preference attributes
		// and complete the given preference.
		if (result instanceof LyFilter)
			preference.setFilter((LyFilter)result);
		else if (result instanceof LySort)
			preference.setSort((LySort)result);
		else if (result instanceof LyFieldInfo)
			preference.addFieldInfo((LyFieldInfo)result);
	}

	/**
	 * Completes a field bean with the result of the current node.
	 *
	 * @param     fieldBean Field bean to be completed.
	 * @param     node      Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeFieldBean(LyFieldBean fieldBean, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();
		String tag = node.getTag();

		// Check for tags or results corresponding to class binding attributes
		// and complete the given class binding.
		if (tag == LyInfoTokens.TAG_DATA)
			fieldBean.putData(node.getString(LyInfoTokens.ATT_NAME),
				((result == null) ? null : result.toString()));
	}

	/**
	 * Completes a field group with the result of the current node.
	 *
	 * @param     fieldGroup Field group to be completed.
	 * @param     node       Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeFieldGroup(LyFieldGroup fieldGroup, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();
		String tag = node.getTag();

		// Check for tags or results corresponding to field group attributes
		// and complete the given field group.
		if (tag == LyInfoTokens.TAG_PREFIX)
			fieldGroup.setPrefix((String)result);
		else if (result instanceof LyFieldConstraint)
			fieldGroup.setConstraint((LyFieldConstraint)result);
	}

	/**
	 * Completes a field tab with the result of the current node.
	 *
	 * @param     fieldTab Field tab to be completed.
	 * @param     node     Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeFieldTab(LyFieldTab fieldTab, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();
		String tag = node.getTag();

		// Check for tags or results corresponding to field tab attributes
		// and complete the given field tab.
		if (tag == LyInfoTokens.TAG_PREFIX)
			fieldTab.setPrefix((String)result);
	}

	/**
	 * Completes a label with the result of the current node.
	 *
	 * @param     label Label to be completed.
	 * @param     node  Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeInfoLabel(LyInfoLabel label, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to label attributes
		// and complete the given label.
		if (result instanceof LyPath)
			label.addPath((LyPath)result);
		else if (result instanceof LyUnit)
			label.addUnit((LyUnit)result);
	}

	/**
	 * Completes a rule with the result of the current node.
	 *
	 * @param     rule Rule to be completed.
	 * @param     node Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeRule(LyRule rule, LyInfoNode node)
		throws SAXException
	{
	    String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags or results corresponding to rule attributes
		// and complete the given label.
		if (result instanceof LyFilter)
			rule.setFilter((LyFilter)result);
		else if (result instanceof LyInfoLabel)
			rule.setLabel((LyInfoLabel)result);
	    else if (tag == LyInfoTokens.TAG_FORMULA)
	        rule.setFormula((String)result);
	}

	/**
	 * Completes a unit with the result of the current node.
	 *
	 * @param unit Unit to be completed.
	 * @param node Current node.
	 */
	void completeUnit(LyUnit unit, LyInfoNode node)
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags or results corresponding to unit attributes
		// and complete the given unit.
		if (tag == LyInfoTokens.TAG_FORMAT)
			unit.setFormat((LyFormat)result);
	}

	/**
     * Completes a form save with the result of the current node.
     *
     * @param formSave Form save to be completed.
     * @param node Current node.
     */
    void completeFormSave(LyFormSave formSave, LyInfoNode node)
    {
        String tag = node.getTag();
        Object result = node.getResult();

        // Complete the given form save with field saves.
        if (result instanceof LyFieldSave)
            formSave.addFieldSave((LyFieldSave)result);
        else if (tag == LyInfoTokens.TAG_DATA)
            formSave.putData(node.getString(LyInfoTokens.ATT_NAME),
                (result == null ? null : result.toString()));
    }
    
    /**
     * Completes a composition save with the result of the current node.
     *
     * @param formSave Composition save to be completed.
     * @param node Current node.
     */
    void completeCompositionSave(LyCompositionFieldSave compositionSave, LyInfoNode node)
    {
        String tag = node.getTag();
        Object result = node.getResult();

        // Check for tags or results corresponding to unit attributes
        // and complete the given unit.
        if (tag == LyInfoTokens.TAG_FIELD_SAVE)
        {
            LyFieldSave    fieldSave = (LyFieldSave)result;
            
            compositionSave._fieldInfo = fieldSave._fieldInfo;
            compositionSave._newValue = fieldSave._newValue;
            compositionSave._oldValue = fieldSave._oldValue;
        }
        else if (tag == LyInfoTokens.TAG_FORM_SAVE)
            compositionSave._formSave = (LyFormSave)result;
    }
    
    /**
     * Completes a multi-composition save with the result of the current node.
     *
     * @param multiCompositionSave Multi-composition save save to be completed.
     * @param node Current node.
     */
    void completeMultiCompositionSave(LyMultiCompositionFieldSave multiCompositionSave, LyInfoNode node)
    {
        String tag = node.getTag();
        Object result = node.getResult();

        // Check for tags or results corresponding to unit attributes
        // and complete the given unit.
        if (tag == LyInfoTokens.TAG_FIELD_SAVE)
        {
            LyFieldSave    fieldSave = (LyFieldSave)result;
            
            multiCompositionSave._fieldInfo = fieldSave._fieldInfo;
            multiCompositionSave._newValue = fieldSave._newValue;
            multiCompositionSave._oldValue = fieldSave._oldValue;
        }
        else if (tag == LyInfoTokens.TAG_FORM_SAVE)
            multiCompositionSave.addFormSave((LyFormSave)result);
    }

	/**
	 * Reset marks of the given.
	 */
	protected BitSet resetMarks(LyBitSet.Type type, BitSet marks)
	{
		if (marks != null)
		{
			// Only reset Gui marks
			if (type == LyBitSet.Type.C_MARKS)
			{
				BitSet guiMarks = LyClassInfo.getMarks(marks, LyObjectInfo.PART_GUI);

				marks.andNot(guiMarks);
				return marks;
			}
			else if (type == LyBitSet.Type.F_MARKS)
			{
				BitSet guiMarks = LyFieldInfo.getMarks(marks, LyObjectInfo.PART_GUI);

				marks.andNot(guiMarks);
				return marks;
			}
		}

		return super.resetMarks(type, marks);
	}

	/**
	 * Returns the model type of this factory.
	 */
	protected LyTag.Model getModelType()
	{
		return LyTag.Model.GUI;
	}
} // end class LyInfoGuiFactory
