/*
 * LyInfoLabel.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;
import leon.app.*;
import leon.data.*;
import leon.data.daemon.*;
import leon.misc.*;


/**
 * This class defines a dynamic label.
 */
public class LyInfoLabel extends LyInfo
{
	public static final String     __VERSION =
		"$Id: LyInfoLabel.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Label types. */
    public enum Type
    {
    	ID,
        NAME,
        GRAMMATICAL,
        SEMANTIC,
        VALUE,
        SHORT_VALUE,
        USER
    }

	/** Value of the label. */
	String						   _value;

	/** Type of the label. */
	Type						   _type;

	/** Ordered list of paths of the label. */
	Vector<LyPath>				   _paths;

	/** Ordered list of units of the label. */
	Vector<LyUnit>				   _units;

	/**
	 * List of instance of LyPath (each par of the label) of this label. A unit may be associated to
	 * each path.
	 */
	LyInfoDataList<LyPath, LyUnit> _parts;

	/** Sub-type of the label. */
	String						   _subType;

	/* Constructor */

    /**
     * Default constructor of a label.
     */
    LyInfoLabel()
    {
    }
    
	/**
	 * Constructor of a label.
	 *
	 * @param value   Value of the label.
	 * @param type    Type of the label
	 * @param subType Sub-type of the label
	 */
	LyInfoLabel(String id, String value, Type type, String subType)
	{
		super(id);

		setValue(value);
		setType(type);
		_paths = null;
		_units = null;
		setSubType(subType);
	}

	/**
	 * Final step of the parsing : check consistency.
	 */
	void check(LyObjectInfo source)
	{
		if (_paths == null)
			return;

		int nbp = _paths.size();
		int nbu = ((_units == null) ? 0 : _units.size());

		if (nbp > nbu)
			addUnit(null);

		LyPath	    path;
		LyUnit	    unit;
		LyFieldInfo fieldInfo;
		LyUnits     units;
		LyInfo	    info;
		int		    i;

		_parts = new LyInfoDataList<LyPath, LyUnit>();

		for (i = 0; i < nbp; i++)
		{
			path = _paths.elementAt(i);
			if ((_units == null) || (i >= _units.size()))
			    addUnit(null);

			unit = _units.elementAt(i);

			// Check consistency between target fieldinfo and unit
			fieldInfo = path.getTargetFieldInfo();
			if ((fieldInfo == null) && (unit != null))
				throw new IllegalArgumentException(
					"BF AppEngine : a path can't have a unit with a null target fieldinfo, " + this);
			if ((fieldInfo != null) && (unit != null))
			{
				units = fieldInfo.getUnits();
				if ((units == null) || !units.contains(unit))
					throw new IllegalArgumentException(
						"BF AppEngine : invalid unit reference in path for target fieldinfo "
						+ fieldInfo + ", " + this);
			}

			// Check the validity of the path.
			if ((source != null) && (source instanceof LyFieldInfo))
			{
				info = source.getParent();
				if (info instanceof LyClassInfo)
					source = (LyClassInfo)info;
				else
					source = null;
			}
			else if (!(source instanceof LyClassInfo))
				source = null;
			else if (_type == Type.ID)
			{
				// Only ID fields of the classInfo are allowed
				LyClassInfo classInfo = (LyClassInfo)source;

				if (!classInfo.contains(fieldInfo))
					throw new IllegalArgumentException(
						"BF AppEngine : an ID label can't have a path with a target fieldinfo ("
						+ fieldInfo + ") not owned by the classInfo, " + this);
				else if (!fieldInfo.hasMark(LyFieldInfo.MARK_ID))
					throw new IllegalArgumentException(
						"BF AppEngine : an ID label can't have a path with a target fieldinfo ("
						+ fieldInfo + ") without mark ID, " + this);
			}

			path.check((LyClassInfo)source);

			_parts.addInfo(path, unit);
		}
	} // end method check

	/**
	 * Lock info to avoid modifications.
	 */
	protected void lock()
	{
		super.lock();

		if (_parts != null)
			_parts.lock();
	}

	/**
	 * Gets value of the label.
	 */
	public final String getValue()
	{
		return _value;
	}

	/**
	 * Gets type of the label.
	 */
	public final Type getType()
	{
		return _type;
	}

	/**
	 * Gets paths of this label.
	 */
	public final LyInfoList<LyPath> getPaths()
	{
		return ((_paths == null) ? null : new LyInfoList<LyPath>(_paths));
	}

	/**
	 * Gets units of this label.
	 */
	public final LyInfoList<LyUnit> getUnits()
	{
		return ((_units == null) ? null : new LyInfoList<LyUnit>(_units));
	}

	/**
	 * Gets sub-type of the label.
	 */
	public final String getSubType()
	{
		return _subType;
	}

	/**
	 * Gets string corresponding to this label for the given object.
	 */
	public String getString(LyObject object)
	{
		if ((object == null)
				&& (_paths == null))
		{
			LyDataSession session = _application.getCurrentSession();
			LyEnvironment environment;

			if (session != null)
				environment = session.getEnvironment();
			else
				environment = getEnvironment();

			return environment.getString(_value);
		}

		LyLabelDaemon daemon = _application.getDaemon(this);

		if (daemon != null)
			return daemon.getString(object);
		else
			return computeString(object);
	}

	/**
	 * Gets string corresponding to this label for the given object.
	 */
	public String computeString(LyObject object)
	{
		LyEnvironment environment = getEnvironment();

		try
		{
			int		 n = getPartsCount();
			String[] arguments = null;

			if ((object != null)
					&& (n > 0))
			{
				LyPath path;
				LyUnit unit;
				int    i;

				arguments = new String[n];
				for (i = 0; i < n; i++)
				{
					path = getPath(i);
					unit = getUnit(i);
					arguments[i] = path.getString(object, unit);
				}
			}

			return environment.getString(_value, arguments);
		}
		catch (Exception e)
		{
			environment.logError("BF AppEngine : an exception occured while computing label "
				+ this + " on object " + object.getId());
			environment.writeException(e);
			return null;
		}
	} // end method computeString

	/**
	 * Gets string corresponding to this label for the given class info.
	 */
	public String getString(LyClassInfo classInfo)
	{
		LyDataSession session = _application.getCurrentSession();
		LyEnvironment environment;

		if (session != null)
			environment = session.getEnvironment();
		else
			environment = getEnvironment();

		int		 i, n = getPartsCount();
		String[] arguments = new String[n];
		LyPath   path;
		LyUnit   unit;

		for (i = 0; i < n; i++)
		{
			path = getPath(i);
			unit = getUnit(i);
			arguments[i] = path.getString(classInfo, unit);
		}

		return environment.getString(_value, arguments);
	}

	/**
	 * Checks if this label contains the given field info.
	 */
	public boolean contains(LyFieldInfo fieldInfo)
	{
		int    i, n = getPartsCount();
		LyPath path;

		for (i = 0; i < n; i++)
		{
			path = getPath(i);
			if (path.getTargetFieldInfo() == fieldInfo)
				return true;
		}

		return false;
	}

	/**
	 * Adds a path to this label.
	 *
	 * @param     path Path to add to this label.
	 * @exception IllegalArgumentException If path parameter is null.
	 */
	void addPath(LyPath path)
	{
		if (path == null)
			throw new IllegalArgumentException("BF AppEngine : path can't be null");

		if (_paths == null)
			_paths = new Vector<LyPath>(1);

		int nbp = _paths.size();
		int nbu = ((_units == null) ? 0 : _units.size());

		if (nbp != nbu)
			addUnit(null);

		_paths.addElement(path);
	}

	/**
	 * Adds a unit to this path.
	 *
	 * @param unit Unit to add to this path.
	 */
	void addUnit(LyUnit unit)
	{
		if (_units == null)
			_units = new Vector<LyUnit>(1);

		_units.addElement(unit);
	}

	/**
	 * Adds a part to this label.
	 *
	 * @param path Path to add to this label.
	 * @param unit Unit to add to this label.
	 */
	void addPart(LyPath path, LyUnit unit)
	{
		if (_parts == null)
			_parts = new LyInfoDataList<LyPath, LyUnit>();

		addPath(path);
		addUnit(unit);
		_parts.addInfo(path, unit);
	}

	/**
	 * Returns the number of parts of this label.
	 */
	public int getPartsCount()
	{
		return ((_parts == null) ? 0 : _parts.getSize());
	}

	/**
	 * Returns the path of this label at the specified index.
	 */
	public LyPath getPath(int idx)
	{
		if (_parts == null)
			throw new IllegalArgumentException("BF AppEngine : can't get path of an empty label");
		return _parts.getInfo(idx);
	}

	/**
	 * Returns the unit of this label at the specified index.
	 */
	public LyUnit getUnit(int idx)
	{
		if (_parts == null)
			throw new IllegalArgumentException("BF AppEngine : can't get unit of an empty label");
		return _parts.getData(idx);
	}
    
    /**
     * Sets value of the label.
     */
    protected void setValue(String value)
    {
        _value = ((value == null) ? null : value.intern());
    }

    /**
     * Sets type of the label.
     */
    protected void setType(Type type)
    {
        _type = (type == null ? Type.NAME : type);
    }
    
    /**
     * Sets sub-type of the label.
     */
    protected void setSubType(String subType)
    {
        _subType = ((subType == null) ? null : subType.intern());
    }
    
    /**
     * Sets paths of this label.
     */
    protected void setPaths(LyInfoList<LyPath> paths)
    {
        int     i, np = (paths == null ? 0 : paths.getSize());
        int     nu = (_units == null ? 0 : _units.size());
        
        _parts = null;
        _paths = null;
        
        for (i = 0; i < np; i++)
            addPart(paths.getInfo(i), (i < nu ? _units.elementAt(i) : null));
    }

    /**
     * Sets units of this label.
     */
    protected void setUnits(LyInfoList<LyUnit> units)
    {
        int     i, nu = (units == null ? 0 : units.getSize());
        int     np = (_paths == null ? 0 : _paths.size());
        
        _parts = null;
        _units = null;
        
        for (i = 0; i < np; i++)
            addPart(_paths.elementAt(i), (i < nu ? units.getInfo(i) : null));
    }
} // end class LyInfoLabel
