/*
 * LyInfoList.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import leon.misc.*;

import java.util.*;


/**
 * This class is an utility class to manage a list of LyInfo. Its only purpose is to avoid duplicate
 * code between all classes that extend LyInfo.<br>
 * By default, this list doesn't accept multiple occurrences of the same info, but this mode can be
 * set on to enable them (this mode can only be chosen at construction).<br>
 * After construction, these lists are writable, but once turned to unwritable by lock method, they
 * can't be modified anymore (nor returned to a writable state).
 *
 * @see LyInfo
 * @see LyObjectInfo
 */
public class LyInfoList<Info extends LyInfo> implements Cloneable
{
    public static final String __VERSION =
        "$Id: LyInfoList.java 36486 2012-08-14 11:35:09Z jlsaouli $";

    // Optimization : Last index used by getInfo from String
    private int _lastIndex = 0;

    /** Boolean indicating if multiple occurrences of the same info are allowed in the list. */
    boolean _multiple;

    /**
     * Table of infos of this list. We don't use a Vector to save memory and speed up processing.
     */
    LyInfo[] _infos;

    /**
     * Boolean indicating if this list is writable or not. Once turned to unwritable, these lists
     * can't become writable anymore.
     */
    boolean _writable;

    /* Internal methods */

    /**
     * Default constructor with no parameters. This new list won't allow multiple occurrences of the
     * same info.
     */
    public LyInfoList()
    {
        this(true);
    }

    /**
     * Constructor with choice of multiple occurrences or not.
     *
     * @param multiple Boolean indicating if multiple occurrences are allowed or not.
     */
    public LyInfoList(boolean multiple)
    {
        _multiple = multiple;
        _infos = null;
        _writable = true;
    }

    /**
     * Constructs a new list from an existing one. The new list allows multiple occurrences if the
     * copied one allows this.
     *
     * @param infoList Copied list.
     */
    public LyInfoList(LyInfoList<Info> infoList)
    {
        this(((infoList != null) && infoList._multiple) ? true : false);

        if (infoList == null)
            return;

        int n = infoList.getSize();

        _lastIndex = 0;

        if (n > 0)
        {
            _infos = new LyInfo[n];
            System.arraycopy(infoList._infos, 0, _infos, 0, n);
        }
        else
            _infos = null;
    }

    /**
     * Constructs a new list from a List containing infos. The new list doesn't multiple
     * occurrences even if the copied vector contains multiple occurrences of the same objects.
     *
     * @param list Vector from which infos are copied.
     */
    public LyInfoList(List<Info> list)
    {
        this(false);

        if (list == null)
            return;

        int i, n = list.size();

        for (i = 0; i < n; i++)
            addInfo(list.get(i));
    }

    /**
    * Constructs a new list from an array of infos
    *
    * @param infos Infos array.
    */
    public LyInfoList(Info[] infos)
    {
        this(false);

        if (infos == null)
            return;

        int i, n = infos.length;

        for (i = 0; i < n; i++)
            addInfo(infos[i]);
    }

    /**
     * Constructs a new list and add the given info to it. Other infos can be added after if needed
     * before locking the list.
     *
     * @param info Info added in our new list.
     */
    public LyInfoList(Info info)
    {
        this(false);

        if (info != null)
            add(info);
    }

    /**
     * Internal method to add an info at the end of this list.
     *
     * @param  info  Info to add in this list.
     * @return Effective index where the info was inserted.
     */
    int add(Info info)
    {
        return insertInfoAt(info, -1);
    }

    /**
     * Internal method to insert an info at the given index in this list.
     *
     * @param  info  Info to add in this list.
     * @param  index Requested index. A negative index or an index greater than or equal to the
     *               number of infos in this list, means insert at the end of the list.
     * @return Effective index where the info was inserted.
     */
    int insertInfoAt(Info info, int index)
    {

        if (_infos == null)
        {
            _infos = new LyInfo[1];
            _infos[0] = info;
            return 0;
        }

        Object[] oldInfos = _infos;
        int      n = _infos.length;

        _infos = new LyInfo[n + 1];

        if ((index < 0) || (index > n))
            index = n;

        if (index > 0)
            System.arraycopy(oldInfos, 0, _infos, 0, index);

        _infos[index] = info;

        if (index < n)
        {
            System.arraycopy(oldInfos, index, _infos, index + 1, n - index);
            return index;
        }
        else
            return n;
    } // end method insertInfoAt

    /**
     * Remove info at the given position.
     *
     * @param     index Index of the info to remove.
     * @exception ArrayIndexOutOfBoundsException If index is negative, or greater than or equal to
     *                                           the number of infos in the list.
     */
    void removeInfoAt(int index)
    {

        if (index < 0)
            throw new ArrayIndexOutOfBoundsException("BF AppEngine : " + index + " < 0 ");

        int n = getSize();

        if (index >= n)
            throw new ArrayIndexOutOfBoundsException("BF AppEngine : " + index + " >= " + n);

        _lastIndex = 0;

        if (n == 1)
            _infos = null;
        else
        {
            LyInfo[] oldInfos = _infos;

            _infos = new LyInfo[n - 1];

            if (index > 0)
                System.arraycopy(oldInfos, 0, _infos, 0, index);

            if (index < (n - 1))
                System.arraycopy(oldInfos, index + 1, _infos, index, n - index - 1);
        }
    }

    public Class<?> getComponentType()
    {
        return LyInfo.class;
    }

    /**
     * Checks writable mode on this list.
     *
     * @return True if the list is writable, false otherwise.
     */
    public final boolean isWritable()
    {
        return _writable;
    }

    /**
     * Fetches the info with the specified identifier or name (only works for stored infos that are
     * instances of LyObjectInfo). The first search is made by id, and if it fails a search is then
     * done by name.
     *
     * @param  id Identifier (or name) of the requested info
     * @return The reference to the requested info, or null if it could not be found.
     */
    public final Info getInfo(String id)
    {
        // Search first by id
        Info info = getInfo(id, false);

        // If it fails, search by name
        return ((info != null) ? info : getInfo(id, true));
    }

    /**
     * Fetches the info with the specified identifier or name (only works for stored infos that are
     * instances of LyObjectInfo).
     *
     * @param  id     Identifier or name of the requested info, as specified by the following
     *                parameter.
     * @param  byName Boolean indicating if the search is done on id or on name.
     * @return Reference to the requested info, or null if it could not be found.
     */
    public final synchronized Info getInfo(String id, boolean byName)
    {

        if (id == null)
            return null;

        int n = getSize();

        if (n == 0)
            return null;

        Info info;

        // Intern id to allow fast compares by ==
        id = id.intern();

        // Browse the set of infos to try to meet the requested id or name
        // Restart at the index of the last object found to accelerate next calls
        // with same parameters.
        int i = _lastIndex;

        do
        {
            @SuppressWarnings("unchecked")
            Info temp = (Info) _infos[i];
            info = temp;

            // Checks by id or by name
            if ((!byName && (((LyInfo) info).getId() == id))
                    || (byName
                        && ((info instanceof LyObjectInfo)
                            && (((LyObjectInfo) info).getName() == id))))
            {
                _lastIndex = i;
                return info;
            }

            i = (i + 1) % n;
        }
        while (i != _lastIndex);

        _lastIndex = 0;
        return null;
    } // end method getItem

    /**
     * Gets infos that have the given marks (all or one of them). This method is only available for
     * a list of LyObjectInfo.
     *
     * @param  marks    The requested marks
     * @param  matchAll Indicates if the info must have at least one mark (false) or all of them
     *                  (true).
     * @return A list of objectInfos from this list that have the given marks (may be empty).
     */
    public final LyInfoList<Info> getInfos(BitSet marks, boolean matchAll)
    {

        if (_infos == null)
            return null;

        LyInfoList<Info> result = new LyInfoList<Info>(_multiple);
        int              i, n = getSize();
        Info             item;
        LyObjectInfo     objectInfo;

        // Browse the set of available info to try to meet the requested id
        for (i = 0; i < n; i++)
        {
            @SuppressWarnings("unchecked")
            Info temp = (Info) _infos[i];
            item = temp;

            if (item instanceof LyObjectInfo)
            {
                objectInfo = (LyObjectInfo) item;

                // Checks if the info match requested marks/condition.
                if ((marks == null) || objectInfo.hasMarks(marks, matchAll))
                    result.add(item);
            }
        }

        return result;
    }

    /**
     * Shortcut for the previous method with 'matchAll' set to true (infos matching all marks
     * requested).
     *
     * @param  marks The requested marks.
     * @return List of infos of this list matching all given marks. (may be empty).
     * @see    LyInfoList#getInfos
     */
    public final LyInfoList<Info> getInfos(BitSet marks)
    {
        return getInfos(marks, true);
    }

    /**
     * Gets the list of infos that match or that don't match a given mark. This method is provided
     * to allow optimization if only of mark is searched on infos of this list.
     *
     * @param  mark The requested mark
     * @param  not  To invert the search : the result are the infos that have not the mark.
     * @return List of infos of this list matching (or not matching) the given mark (may be empty).
     */
    public final LyInfoList<Info> getInfos(int mark, boolean not)
    {

        if ((mark < 0) || (_infos == null))
            return null;

        LyInfoList<Info> result = new LyInfoList<Info>(_multiple);
        int              i, n = getSize();
        LyObjectInfo     objectInfo;
        Object           item;
        boolean          hasMark;

        // Browse the set of available info to try to meet the requested id
        for (i = 0; i < n; i++)
        {
            item = _infos[i];

            if (item instanceof LyObjectInfo)
            {
                objectInfo = (LyObjectInfo) item;
                hasMark = objectInfo.hasMark(mark);

                // Checks if the info match requested marks/condition.
                if ((hasMark && !not) || (!hasMark && not))
                {
                    @SuppressWarnings("unchecked")
                    Info temp = (Info) objectInfo;
                    result.add(temp);
                }
            }
        }

        return result;
    } // end method getInfos

    /**
     * Shortcut for the previous method with 'not' set to false (matching required).
     *
     * @param  mark The requested mark.
     * @return List of infos of this list matching the given mark (may be empty).
     * @see    LyInfoList#getInfos
     */
    public final LyInfoList<Info> getInfos(int mark)
    {
        return getInfos(mark, false);
    }

    /**
     * Returns the number of infos in this list.
     *
     * @return Number of infos of this list.
     */
    public final int getSize()
    {
        return ((_infos == null) ? 0 : _infos.length);
    }

    /**
     * Gets info at the specified index. This method is kept for upward compatibility.
     *
     * @param     index Index of the requested info.
     * @return    Info at the given index.
     * @exception ArrayIndexOutOfBoundsException If index is negative, or greater than or equal to
     *                                           the number of infos in the list.
     */
    @SuppressWarnings("unchecked")
    public final Info getInfo(int index)
    {

        if (_infos == null)
            return null;

        return (Info) _infos[index];
    }

    /**
     * Returns a vector containing all infos of this list. This is a COPY of the internal list,
     * therefore this list may be freely manipulated by the application.
     *
     * @return Set of infos of this list, or an empty Vector.
     */
    public final Vector<Info> getInfos()
    {

        if (_infos == null)
            return null;

        // Return a copy to avoid manipulation of the internal list
        int          i, n = getSize();
        Vector<Info> result = new Vector<Info>(n);

        for (i = 0; i < n; i++)
        {
            @SuppressWarnings("unchecked")
            Info temp = (Info) _infos[i];
            result.addElement(temp);
        }

        return result;
    }

    /**
     * Returns a sub-list of this list of all infos of a given java class.
     *
     * @param  infoClass The java class.
     * @return The sub-list of all infos of the given java class.
     */
    public <T extends Info> LyInfoList<T> getInfos(Class<T> infoClass)
    {
        LyInfoList<T> result = new LyInfoList<T>(_multiple);
        int           i, n = getSize();

        for (i = 0; i < n; i++)
        {
            @SuppressWarnings("unchecked")
            Info info = (Info) _infos[i];

            if ((info != null)
                    && ((infoClass == null) || infoClass.isAssignableFrom(info.getClass())))
            {
                @SuppressWarnings("unchecked")
                T temp = (T) info;
                result.add(temp);
            }
        }

        return result;
    }

    /**
     * Returns a list containing all object infos of this list sorted by their id.
     *
     * @return Vector of infos of this list.
     */
    public final LyInfoList<Info> getInfosById()
    {

        if (_infos == null)
            return null;

        LyInfoList<Info> result = new LyInfoList<Info>(this);
        LyInfo           info, info2;
        int              i, j, n = getSize();

        for (i = 0; i < n; i++)
        {
            info = result._infos[i];

            for (j = i + 1; j < n; j++)
            {
                info2 = result._infos[j];

                if (info.getId().compareTo(info2.getId()) > 0)
                {
                    result._infos[i] = info2;
                    result._infos[j] = info;
                    info = info2;
                }
            }
        }

        return result;
    } // end method getInfosById

    /**
     * Returns a list containing all object infos of this list sorted by their names.
     * @deprecated Use getInfosByName(LyEnvironment).
     */
    public final LyInfoList<Info> getObjectInfosByName()
    {
        return getInfosByName();
    }
    
    /**
     * Returns a list containing all infos of this list sorted by their names. The sort does
     * not translate the names from the environment
     *
     * @return List of infos of this list.
     */
    public final LyInfoList<Info> getInfosByName()
    {
        return getInfosByName(null);
    }

    /**
     * Returns a list containing all object infos of this list sorted by their names.
     * @deprecated Use getInfosByName(LyEnvironment).
     */
    public final LyInfoList<Info> getObjectInfosByName(LyEnvironment translationEnv)
    {
    	return getInfosByName(translationEnv);
    }
    
    /**
     * Returns a list containing all infos of this list sorted by their names.
     *
     * @param  translationEnv optional environment to translate strings
     * @return List of infos of this list.
     */
    public final LyInfoList<Info> getInfosByName(LyEnvironment translationEnv)
    {
        if (_infos == null)
            return null;

        int              i, j, n = getSize();
        LyInfoList<Info> result = new LyInfoList<Info>(this);
        LyInfo 			 info, info2;
        String       	 name, name2;

        for (i = 0; i < n; i++)
        {
            info = result._infos[i];
            name = info.getName();

            if (translationEnv != null)
                name = translationEnv.translate(name);

            for (j = i + 1; j < n; j++)
            {
                info2 = result._infos[j];
                name2 = info2.getName();

                if (translationEnv != null)
                    name2 = translationEnv.translate(name2);

                if (name.compareTo(name2) > 0)
                {
                    result._infos[i] = info2;
                    result._infos[j] = info;
                    info = info2;
                    name = name2;
                }
            }
        }

        return result;
    } // end method getObjectInfosByName

    /**
     * Locks this list (turn it to unwritable). Once locked, this list can't become writable anymore
     * and won't accept any modification of its content.
     */
    public final void lock()
    {

        if (!_writable)
            return;

        _writable = false;

        // Lock infos
        int i, n = getSize();

        for (i = 0; i < n; i++)
            _infos[i].lock();
    }

    /**
     * Adds a new info to this info list at the specified position.
     *
     * @param     info     Info that should be added.
     * @param     position Requested position. A negative index or an index greater than or equal to
     *                     the number of infos in this list, means insert at the end of the list.
     * @return    Effective index where the info was inserted. -1 if the info was already in this
     *            list and this list doesn't allow multiple occurrences.
     * @exception IllegalStateException If this list is not writable.
     */
    public int addInfo(Info info, int position)
    {

        if (info == null)
            return -1;

        // Adds the new info if not already added
        if (!_multiple && contains(info))
            return -1;

        if (!_writable)
            throw new IllegalStateException("BF AppEngine : this info list is not writable");

        return insertInfoAt(info, position);
    }

    /**
     * Adds a new info at the end of this info list.
     *
     * @param     info Info that should be added
     * @return    Effective index where the info was inserted. -1 if the info was already in this
     *            list and this list doesn't allow multiple occurrences.
     * @exception IllegalStateException If this list is not writable.
     */
    public final int addInfo(Info info)
    {
        return addInfo(info, -1);
    }

    /**
     * Removes the specified info from this info list.
     *
     * @param     info Info to remove.
     * @return    Index of the removed info, -1 if the info was not in this list.
     * @exception IllegalStateException If this list is not writable.
     */
    public int removeInfo(Info info)
    {
        int index = getIndex(info);

        if (index < 0)
            return -1;

        if (!_writable)
            throw new IllegalStateException("BF AppEngine : this info list is not writable");

        removeInfoAt(index);
        return index;
    }

    /**
    * Removes all info of the specified class from this info list.
    *
    * @param     c.
    * @return    Index of the removed info, -1 if the info was not in this list.
    * @exception IllegalStateException If this list is not writable.
    */
    protected <T extends Info> LyInfoList<T> removeInfos(Class<T> infoClass)
    {
        LyInfoList<T> result = new LyInfoList<T>(_multiple);
        int           i, n = getSize();

        for (i = n - 1; i >= 0; i--)
        {
            @SuppressWarnings("unchecked")
            Info info = (T) _infos[i];

            if ((infoClass == null) || infoClass.isAssignableFrom(info.getClass()))
            {
                @SuppressWarnings("unchecked")
                T temp = (T) info;
                result.add(temp);
                removeInfoAt(i);
            }
        }

        return result;
    }

    /**
     * Adds all infos of the given list to this list.
     *
     * @param     list List of infos to add to this list.
     * @exception IllegalStateException If this list is not writable.
     */
    public void addList(LyInfoList<Info> list)
    {

        if (list == null)
            return;

        if (!_writable)
            throw new IllegalStateException("BF AppEngine : this info list is not writable");

        int  i, n = list.getSize();
        Info info;

        for (i = 0; i < n; i++)
        {
            info = list.getInfo(i);
            addInfo(info);
        }
    }

    /**
     * Adds all infos of the given list to this list.
     *
     * @param     list List of infos to add to this list.
     * @exception IllegalStateException If this list is not writable.
     */
    public void addList(List<Info> list)
    {

        if (list == null)
            return;

        if (!_writable)
            throw new IllegalStateException("BF AppEngine : this info list is not writable");

        int  i, n = list.size();
        Info info;

        for (i = 0; i < n; i++)
        {
            info = list.get(i);
            addInfo(info);
        }
    }

    /**
     * Resets info list (remove all infos from this list).
     *
     * @exception IllegalStateException If this list is not writable.
     */
    public void resetInfos()
    {

        if (!_writable)
            throw new IllegalStateException("BF AppEngine : this info list is not writable");

        _infos = null;
    }

    /**
     * Get index of a given info.
     *
     * @param  info The searched info.
     * @return Index if the info in this list if it was found, -1 otherwise.
     */
    public final int getIndex(Info info)
    {

        if (_infos == null)
            return -1;

        int i, n = _infos.length;

        for (i = 0; i < n; i++)
            if (_infos[i] == info)
                return i;

        return -1;
    }

    /**
     * Checks if the given info is contained in this list.
     *
     * @param  info The searched info.
     * @return True if the given info is found in this list, false otherwise.
     */
    public final boolean contains(Info info)
    {
        return (getIndex(info) >= 0);
    }

    /**
     * Checks if the given list of infos are contained in this list.
     *
     * @param  infos the searched list of infos.
     * @return True if all the infos of the list are found in this list, false otherwise.
     */
    public final boolean contains(LyInfoList<Info> infos)
    {
        return contains(infos, true);
    }

    /**
     * Checks if the given list of infos are contained in this list.
     *
     * @param  infos the searched list of infos.
     * @return True if all the infos of the list are found in this list, false otherwise.
     */
    public final boolean contains(LyInfoList<Info> infos, boolean all)
    {

        if ((infos == null) || (infos == this))
            return true;

        int  i, n = infos.getSize();
        Info info;

        for (i = 0; i < n; i++)
        {
            info = infos.getInfo(i);

            if (!all && contains(info))
                return true;

            if (all && !contains(info))
                return false;
        }

        return all;
    }

    /**
     * Returns a clone of this list. Redefined to properly clone internal table of infos.
     *
     * @return Cloned list.
     */
    public synchronized Object clone()
    {
        LyInfoList<Info> clone = new LyInfoList<Info>();
        int              n = getSize();

        try
        {
            super.clone();
        }
        catch (CloneNotSupportedException e)
        {
        }

        clone._lastIndex = 0;
        clone._multiple = _multiple;

        if (n > 0)
        {
            clone._infos = new LyInfo[n];
            System.arraycopy(_infos, 0, clone._infos, 0, n);
        }
        else
            clone._infos = null;

        clone._writable = true;
        return clone;
    }

    /**
     * Compare infolist content.
     *
     * @param obj another list to compare content
     */
    public boolean equals(Object obj)
    {
        if (obj == this)
            return true;

        if (!(obj instanceof LyInfoList<?>))
            return false;

        @SuppressWarnings("unchecked")
        LyInfoList<Info> list = (LyInfoList<Info>) obj;

        if (getSize() != list.getSize())
            return false;

        int i, n = getSize();

        for (i = 0; i < n; i++)
            if (getInfo(i) != list.getInfo(i))
                return false;

        return true;
    } // end method equals

    /**
     * Dumps content of this list.
     */
    public String toString()
    {
        StringBuffer buffer = new StringBuffer();
        int          i, n = getSize();
        Info         info;

        buffer.append(hashCode());
        buffer.append(" = {");
        buffer.append(n);
        buffer.append(": ");

        for (i = 0; i < n; i++)
        {
            info = getInfo(i);

            if (i > 0)
                buffer.append(", ");

            buffer.append(info.toString());
        }

        buffer.append('}');
        return buffer.toString();
    }

    /**
     * Returns an array containing all of the elements in this collection.
     * @return an array containing the elements of this collection
     */
    public LyInfo[] toArray()
    {

        if (_infos == null)
            return null;

        // Return a copy to avoid manipulation of the internal list
        int      size = getSize();
        LyInfo[] result = new LyInfo[size];

        System.arraycopy(_infos, 0, result, 0, size);
        return result;
    }
} // end class LyInfoList
