/**
 * LyInfoNode.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import org.xml.sax.*;
import java.util.*;
import leon.parser.*;


/**
 * Class that stores information corresponding to a node during parsing :
 *
 * <ul>
 *   <LI>tag name,</LI>
 *   <LI>attributes list,</LI>
 *   <LI>attributes values,</LI>
 *   <LI>parent node,</LI>
 *   <LI>child nodes,</LI>
 *   <LI>Result of the parsing. Depending on the tag node, the result can have different forms
 *     (classes that inherit from LyInfo, BitSet, String, ...).</LI>
 * </ul>
 */
public final class LyInfoNode extends LyXmlNode
{
	public static final String __VERSION =
		"$Id: LyInfoNode.java 28422 2010-11-03 12:03:38Z ehelmste $";

	// Tag name.
	String					   _tag;

    // Identifier
    String                     _id;

	// Attributes values
	HashMap<String, String>    _attributesValues;

	// Result of the parsing
	Object					   _result;

	// Parent node
	LyInfoNode				   _parent;

	// Child nodes
	List<LyInfoNode>		   _children;

	// Extra infos
	LyInfoList<LyInfo>		   _infos;

	// Ended status
	boolean					   _ended;

	/**
	 * Constructor.
	 */
	LyInfoNode(String tag, Attributes attributes)
	{
		_tag = tag;
		copyAttributeValues(attributes);
        _id = getString(LyInfoTokens.ATT_ID);
	}

	/**
	 * Copy attribute values for post processing use.
	 */
	void copyAttributeValues(Attributes attributes)
	{
		int       i, n = attributes.getLength();
		String    name, value;
        boolean   filterFalseValues = _tag.endsWith("_MARKS");

		_attributesValues = new HashMap<String, String>(n);

		for (i = 0; i < n; i++)
		{
			value = attributes.getValue(i);
            if (!filterFalseValues
                || !value.equals("false"))
            {
                name = attributes.getQName(i);
                _attributesValues.put(name, value);
            }
		}
	}

	/**
	 * Returns result corresponding to this node. This result is the object built during parsing.
	 *
	 * @return The result corresponding to this node.
	 */
	Object getResult()
	{
		return _result;
	}

	/**
	 * Sets result corresponding to this node. This result is the object built during parsing.
	 *
	 * @param result The result corresponding to this node.
	 */
	void setResult(Object result)
	{
		_result = result;
	}

	/**
	 * Utility method to return the "interned" tag name.
	 *
	 * @return Tag name for this node.
	 */
	String getTag()
	{
		return _tag;
	}

	/**
	 * Adds a child to this node to child list.
	 *
	 * @param node Child node.
	 */
	void addChild(LyInfoNode node)
	{
		if (_children == null)
			_children = new ArrayList<LyInfoNode>(20);

		_children.add(node);
	}

	/**
	 * Gets child at the given index in child list.
	 *
	 * @param  idx Index of the searched child.
	 * @return The child node if it exists, null otherwise.
	 */
	LyInfoNode getChild(int idx)
	{
		if (_children == null)
			return null;

		return _children.get(idx);
	}

	/**
	 * Gets children count.
	 *
	 * @return Number of child nodes for this node.
	 */
	int getChildCount()
	{
		if (_children == null)
			return 0;

		return (_children.size());
	}

	/**
	 * Sets parent node.
	 *
	 * @param parent The parent node.
	 */
	void setParent(LyInfoNode node)
	{
		_parent = node;
	}

	/**
	 * Gets parent node.
	 *
	 * @return Parent node (null for root node).
	 */
	LyInfoNode getParent()
	{
		return _parent;
	}

	/**
	 * Gets value for an attribute.
	 *
	 * @param  attributeName Attribute name.
	 * @return Value of this attribute, null if no value is found.
	 */
	public String getAttribute(String attributeName)
	{
		return _attributesValues.get(attributeName);
	}

	/**
	 * Fetch the value of the ID attribute.
	 *
	 * @return ID attribute value, or null if it could not be found
	 */
	String getId()
	{
        return _id;
	}

	/**
	 * Clean up at the end of parsing. Get rid of parsed tree structure and all references. Method
	 * also free child nodes recursively.
	 */
	void free()
	{
		if (_children != null)
		{
			int i, n = _children.size();

			for (i = 0; i < n; i++)
				_children.get(i).free();
		}
		_attributesValues = null;
		_result = null;
		_parent = null;
		_children = null;
		_tag = null;
		_infos = null;
	}

	/**
	 * Add an info to the node.
	 */
	void addExtraInfo(LyInfo info)
	{
		if (_infos == null)
			_infos = new LyInfoList<LyInfo>();
		_infos.addInfo(info);
	}

	/**
	 * Get infos list associated to this node.
	 */
	LyInfoList<LyInfo> getExtraInfos()
	{
		return _infos;
	}
    
    /**
     * Get number of attributes.
     */
    int getNbAttributes()
    {
        return _attributesValues.size();
    }
} // end class LyInfoNode
