/**
 * LyInfoNodeFactory.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;

import leon.misc.LyEnvironment;

import leon.app.*;

import leon.info.binding.*;
import leon.info.daemon.*;
import leon.info.location.*;
import leon.info.misc.*;
import leon.info.model.*;
import leon.info.notifier.*;
import leon.misc.*;

import org.xml.sax.*;


/**
 * This class builds instances corresponding to XML tags read from the data description file. Almost
 * all objects built during parsing inherit from LyInfo (except very elementary objects).
 *
 * @see     LyInfoParser
 * @see     LyInfo
 * @see     LyObjectInfo
 * @see     LyFieldInfo
 * @see     LyChoiceFieldInfo
 * @see     LyChoiceOption
 * @see     LyDateFieldInfo
 * @see     LyFileFieldInfo
 * @see     LyNumberFieldInfo
 * @see     LyRelationFieldInfo
 * @see     LyTextFieldInfo
 * @see     LyAction
 * @see     LyCompoundAction
 * @see     LyTabAction
 * @see     LyClassInfo
 * @see     LyProject
 * @see     LyFilter
 * @see     LyFilterExpression
 * @see     LyFilterRelation
 * @see     LySort
 * @see     LySortCriterion
 * @see     LyRoute
 * @see     LyStep
 * @see     LyRule
 * @see     LyChartInfo
 * @see     LyPreference
 * @see     LyTreemapConfiguration
 * @version $Id: LyInfoNodeFactory.java 41407 2012-10-30 16:17:49Z bangiole $
 * @author  Lyria
 */
public final class LyInfoNodeFactory extends LyInfoFactory
{
	public static final String		  __VERSION =
		"$Id: LyInfoNodeFactory.java 41407 2012-10-30 16:17:49Z bangiole $";

	// Keeping information during validation stage (check case sensitivity of bindings)
	Hashtable<String, LyClassBinding> 	_bindingsValidationTable;

	LyInfoList<LyFieldTab> 				_applicationTabs;
	LyInfoList<LyFieldGroup>			_applicationGroups;

	/**
	 * Constructor of the factory for an application. All objects built by this factory will belong
	 * to the given application.
	 *
	 * @param parser      Parser that builts this factory.
	 * @param application Application for which the factory is built.
	 */
	public LyInfoNodeFactory(LyInfoParser parser, LyApplication application)
	{
		super(parser, application);

		_applicationTabs = new LyInfoList<LyFieldTab>();
		_applicationGroups = new LyInfoList<LyFieldGroup>();
	}

	/**
	 * Builds node corresponding to an element read from the XML data description file. The result
	 * associated to the built node depends on the node tag. This method is invoked by SAXParser at
	 * the beginning of each new tag.
	 *
	 * @param     tag  Current tag.
	 * @param     node Current node.
	 * @return    The object built corresponding to the given tag and attributes.
	 * @exception SAXException if the result could not have been built.
	 */
	@Override
  public Object buildElement(LyInfoNode node, String tag)
		throws SAXException
	{
		// Init result
		Object result = null;

		// Get some common information
		String id = node.getId();
		String type = node.getString(LyInfoTokens.ATT_TYPE);
		String templateId = node.getString(LyInfoTokens.ATT_MODEL);
		LyInfo template = getInfo(templateId);

		// Specific case of forward reference on ACTION_REF
		if (tag == LyInfoTokens.TAG_ACTION_REF)
		{
			String classId = node.getString(LyInfoTokens.ATT_CLASS);

			if (classId != null)
			{
				// The action referenced in ACTION_REF is also used as template
				templateId = node.getString(LyInfoTokens.ATT_IDREF);
				template = getInfo(templateId);

				if ((templateId != null)
						&& (templateId.length() > 0)
						&& (template == null))
					raiseError( "BF AppEngine : Forward references to action model are not supported on ACTION_REF. Action ref. with action : '"
						+ templateId + "' and class '" + classId + "'.");
			}
		}

		// Build the appropriate result corresponding to each node.
		if (tag == LyInfoTokens.TAG_ACTION)
		{
			String  controller = node.getString(LyInfoTokens.ATT_CONTROLLER);
			String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			String  builder = node.getString(LyInfoTokens.ATT_BUILDER);
			String  view = node.getString(LyInfoTokens.ATT_VIEW);
			LyInfo  action = getInfo(id);
			String  target = node.getString(LyInfoTokens.ATT_TARGET);
			boolean abstractAction = node.getBoolean(LyInfoTokens.ATT_ABSTRACT);

			if (action == null)
				result = new LyAction(id, (LyAction)template, controller, behavior, builder, view, target);
			else if (action == template)
				result = ((LyAction)action).cloneAsTemplate();
			else
			{
				if (!(action instanceof LyAction)
						|| (action instanceof LyTabAction)
						|| (action instanceof LyCompoundAction))
					throw new SAXException("BF AppEngine : extern info " + id + " is not a simple action");
				if (((LyAction)action).isLoaded())
					throw new SAXException("BF AppEngine : action " + id + " already exists");

				((LyAction)action).construct((LyAction)template, controller, behavior, builder,
					view, target);
				result = action;
			}
			if (abstractAction)
				((LyAction)result).setAbstract(true);
		}
		else if (tag == LyInfoTokens.TAG_ARRAY)
		{
			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			short  cMin = (short)node.getInt(LyInfoTokens.ATT_CMIN);
			short  cMax = (short)node.getInt(LyInfoTokens.ATT_CMAX, 0);
			String behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			LyInfo fieldInfo = getInfo(id);

			if (fieldInfo == null)
				result = new LyArrayFieldInfo(id, javaName, (LyArrayFieldInfo)template, cMin, cMax);
			else if (fieldInfo == template)
				result = ((LyFieldInfo)fieldInfo).cloneAsTemplate();
			else
			{
				if (!(fieldInfo instanceof LyArrayFieldInfo))
					throw new SAXException("BF AppEngine : extern info " + id + " is not an array field info");
				if (fieldInfo.isLoaded())
					throw new SAXException("BF AppEngine : array field info " + id + " already exists");

				((LyArrayFieldInfo)fieldInfo).construct(javaName, (LyArrayFieldInfo)template, cMin, cMax);
				result = fieldInfo;
			}

			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);
		}
		else if (tag == LyInfoTokens.TAG_ARRAY_DAEMON)
			result = new LyArrayDaemon();
		else if (tag == LyInfoTokens.TAG_BEAN)
		{
			String actionType = node.getString(LyInfoTokens.ATT_ACTION);
			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);

			result = new LyFieldBean(javaName, LyInfoTokens.getFieldBeanActionType(actionType));
		}
		else if (tag == LyInfoTokens.TAG_C_BIND)
		{
			String value = node.getString(LyInfoTokens.ATT_VALUE);
			String location = node.getString(LyInfoTokens.ATT_LOCATION);

			if (location == null)
				location = _application.getDefaultLocationId();

			result = new LyClassBinding(value, location, null);
		}
		else if (tag == LyInfoTokens.TAG_C_BINDS)
		{
			boolean strict = node.getBoolean(LyInfoTokens.ATT_STRICT);

			result = new LyUnionBinding(strict);
		}
		else if (tag == LyInfoTokens.TAG_C_BIND_INFO)
		{
			String value = node.getString(LyInfoTokens.ATT_IDREF);

			result = new LyClassInfoBinding(value);
		}
		else if (tag == LyInfoTokens.TAG_C_BIND_LDAP)
		{
			String value = node.getString(LyInfoTokens.ATT_VALUE);
			String location = node.getString(LyInfoTokens.ATT_LOCATION);
			String fdnFormat = node.getString(LyInfoTokens.ATT_FDN_FORMAT);
			String rdnName = node.getString(LyInfoTokens.ATT_RDN);
			String classFilter = node.getString(LyInfoTokens.ATT_LDAP_CLASS_FILTER);

			if (location == null)
				location = _application.getDefaultLocationId();

			result = new LyLdapClassBinding(value, location, null);
			((LyLdapClassBinding)result).setFdnFormat(fdnFormat);
			((LyLdapClassBinding)result).setRdn(rdnName);
			((LyLdapClassBinding)result).setClassFilter(classFilter);
		}
		else if (tag == LyInfoTokens.TAG_CACHE)
		{
			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			String userData = node.getString(LyInfoTokens.ATT_USER_DATA);

			if (javaName != null)
				result = new LyInfoCache(javaName, userData);
			else
				result = new LyInfoCache(LyInfoTokens.getClassInfoCachePolicy(type), userData);
		}
		else if (tag == LyInfoTokens.TAG_CHART)
		{
			String				  name = node.getString(LyInfoTokens.ATT_NAME);
			String				  classInfoId = node.getString(LyInfoTokens.ATT_CLASS_INFO_ID);
			String				  xField = node.getString(LyInfoTokens.ATT_X_FIELD);
			String				  groupField = node.getString(LyInfoTokens.ATT_GROUP_FIELD);
			LyChartInfo.Type	  typeVarX = LyInfoTokens.getChartInfoTypeVar(node.getString(
						LyInfoTokens.ATT_TYPE_X));
			LyChartInfo.Type	  typeVarY = LyInfoTokens.getChartInfoTypeVar(node.getString(
						LyInfoTokens.ATT_TYPE_Y));
			LyChartInfo.GraphType graphType = LyInfoTokens.getChartInfoGraphType(node.getString(
						LyInfoTokens.ATT_GRAPH_TYPE));
			LyChartInfo.ValType   xVal = LyInfoTokens.getChartInfoValType(node.getString(
						LyInfoTokens.ATT_X_VAL));
			LyChartInfo.ValType   yVal = LyInfoTokens.getChartInfoValType(node.getString(
						LyInfoTokens.ATT_Y_VAL));
            String[]              xIntervals = new String[3];

            xIntervals[0] = node.getString(LyInfoTokens.ATT_X_INTS_BEGIN);
            xIntervals[1] = node.getString(LyInfoTokens.ATT_X_INTS_LENGTH);
            xIntervals[2] = node.getString(LyInfoTokens.ATT_X_INTS_UNIT);

            String[]              yIntervals = new String[3];

            yIntervals[0] = node.getString(LyInfoTokens.ATT_Y_INTS_BEGIN);
            yIntervals[1] = node.getString(LyInfoTokens.ATT_Y_INTS_LENGTH);
            yIntervals[2] = node.getString(LyInfoTokens.ATT_Y_INTS_UNIT);

            boolean        pieBy = node.getBoolean(LyInfoTokens.ATT_PIE_BY);
            String         title = node.getString(LyInfoTokens.ATT_TITLE);
            String         xTitle = node.getString(LyInfoTokens.ATT_X_TITLE);
            String         yTitle = node.getString(LyInfoTokens.ATT_Y_TITLE);

            String[] xAxis = new String[3];

            xAxis[0] = node.getString(LyInfoTokens.ATT_X_AXIS_MIN);
            xAxis[1] = node.getString(LyInfoTokens.ATT_X_AXIS_MAX);
            xAxis[2] = node.getString(LyInfoTokens.ATT_X_AXIS_WIDTH);

            String[] yAxis = new String[3];

            yAxis[0] = node.getString(LyInfoTokens.ATT_Y_AXIS_MIN);
            yAxis[1] = node.getString(LyInfoTokens.ATT_Y_AXIS_MAX);
            yAxis[2] = node.getString(LyInfoTokens.ATT_Y_AXIS_WIDTH);

			boolean xVerticalLabels = node.getBoolean(LyInfoTokens.ATT_X_VERT_LABELS);
			boolean xGrid = node.getBoolean(LyInfoTokens.ATT_X_GRID);
			boolean yGrid = node.getBoolean(LyInfoTokens.ATT_Y_GRID);
			boolean d3Effect = node.getBoolean(LyInfoTokens.ATT_3D_EFFECT);
			boolean showLegend = node.getBoolean(LyInfoTokens.ATT_SHOW_LEGEND);
			boolean showTips = node.getBoolean(LyInfoTokens.ATT_SHOW_TIPS);
			boolean ratios = node.getBoolean(LyInfoTokens.ATT_RATIOS);

			result = new LyChartInfo(id, name, classInfoId, xField, groupField, typeVarY, typeVarX,
					graphType, xVal, xIntervals, yVal, yIntervals, pieBy, title, xTitle, yTitle,
					xAxis, yAxis, xVerticalLabels, xGrid, yGrid, d3Effect, showLegend, showTips,
					ratios);
		}
		else if (tag == LyInfoTokens.TAG_CHARTYPE)
		{
			String value = node.getString(LyInfoTokens.ATT_VALUE);
			LyInfo charType = getInfo(id);

			if (charType == null)
				result = new LyInfoCharType(id, value);
			else
			{
				if (!(charType instanceof LyInfoCharType))
					throw new SAXException("BF AppEngine : extern info " + id + " is not a char type");
				if (charType.isLoaded())
					throw new SAXException("BF AppEngine : char type " + id + " already exists");

				((LyInfoCharType)charType).construct(value);
				result = charType;
			}
		}
		else if (tag == LyInfoTokens.TAG_CHOICE)
		{
			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			boolean multiple = node.getBoolean(LyInfoTokens.ATT_MULTIPLE);
			String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			String  sortBy = node.getString(LyInfoTokens.ATT_SORT_BY);
			boolean not = node.getBoolean(LyInfoTokens.ATT_NOT);
			LyInfo  fieldInfo = getInfo(id);

			if (fieldInfo == null)
				result = new LyChoiceFieldInfo(id, javaName, (LyChoiceFieldInfo)template, multiple,
						LyInfoTokens.getChoiceSortBy(sortBy),
						LyInfoTokens.getChoiceFieldInfoType(type), not);
			else if (fieldInfo == template)
				result = ((LyFieldInfo)fieldInfo).cloneAsTemplate();
			else
			{
				if (!(fieldInfo instanceof LyChoiceFieldInfo))
					throw new SAXException("BF AppEngine : extern info " + id + " is not a choice field info");
				if (fieldInfo.isLoaded())
					throw new SAXException("BF AppEngine : choice field info " + id + " already exists");

				((LyChoiceFieldInfo)fieldInfo).construct(javaName, (LyChoiceFieldInfo)template,
					multiple, LyInfoTokens.getChoiceSortBy(sortBy),
					LyInfoTokens.getChoiceFieldInfoType(type), not);
				result = fieldInfo;
			}

			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);
		}
		else if (tag == LyInfoTokens.TAG_CLASS)
		{
			String superClassId = node.getString(LyInfoTokens.ATT_EXTENDS);
			LyInfo superClass = getInfo(superClassId);

			// Forbid forward reference on super-class
			if ((superClassId != null)
					&& (superClassId.length() > 0)
					&& (superClass == null))
				raiseError("BF AppEngine : Forward reference to super-class is not supported. Class '"
					+ id + "' with super-class : '" + superClassId + "'.");

			if (id != null)
			{
				String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
				boolean abstractClass = node.getBoolean(LyInfoTokens.ATT_ABSTRACT);
				boolean virtual = node.getBoolean(LyInfoTokens.ATT_VIRTUAL);
				String  userData = node.getString(LyInfoTokens.ATT_USER_DATA);
				String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
				LyInfo  classInfo = getInfo(id);

				if (virtual && (superClass == null))
					raiseError("BF AppEngine : declaration of the virtual class "
						+ classInfo + " requires a super-class");

				if (classInfo == null)
					result = new LyClassInfo(id, (LyClassInfo)superClass, (LyClassInfo)template,
						javaName, behavior, abstractClass, virtual, userData);
				else if (classInfo == template)
					result = ((LyClassInfo)classInfo).cloneAsTemplate();
				else
				{
					if (!(classInfo instanceof LyClassInfo))
						throw new SAXException("BF AppEngine : extern info " + id + " is not a class info");
					if (classInfo.isLoaded())
						throw new SAXException("BF AppEngine : class info " + id + " already exists");

					((LyClassInfo)classInfo).construct((LyClassInfo)superClass,
						(LyClassInfo)template, javaName, behavior, abstractClass, virtual,
						userData);
					result = classInfo;
				}
			}
			else if (superClass != null)
				result = superClass;
			else
				throw new IllegalStateException(
					"BF AppEngine : class declaration requires either an 'id' to create a new class or an 'extends' to extend an existing class");
		}
		else if (tag == LyInfoTokens.TAG_CLASSES)
			result = new Vector<LyClassInfo>();
		else if (tag == LyInfoTokens.TAG_COMPOUND_ACTION)
		{
			String  controller = node.getString(LyInfoTokens.ATT_CONTROLLER);
			String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			String  builder = node.getString(LyInfoTokens.ATT_BUILDER);
			String  style = node.getString(LyInfoTokens.ATT_STYLE);
			String  context = node.getString(LyInfoTokens.ATT_CONTEXT);
			String  sizes = node.getString(LyInfoTokens.ATT_SIZES);
			String  clientContext = node.getString(LyInfoTokens.ATT_CLIENT_CONTEXT);
			String  target = node.getString(LyInfoTokens.ATT_TARGET);
			boolean abstractAction = node.getBoolean(LyInfoTokens.ATT_ABSTRACT);
			LyInfo  compoundAction = getInfo(id);

			if (compoundAction == null)
				result = new LyCompoundAction(id, (LyAction)template, controller, behavior, builder,
						LyInfoTokens.getCompoundActionStyle(style), sizes,
						LyInfoTokens.getCompoundActionContext(context),
						LyInfoTokens.getCompoundActionClientContext(clientContext), target);
			else if (compoundAction == template)
				result = ((LyCompoundAction)compoundAction).cloneAsTemplate();
			else
			{
				if (!(compoundAction instanceof LyCompoundAction))
					throw new SAXException("BF AppEngine : extern info " + id + " is not a compound action");
				if (((LyCompoundAction)compoundAction).isLoaded())
					throw new SAXException("BF AppEngine : compound action " + id + " already exists");

				((LyCompoundAction)compoundAction).construct((LyAction)template, controller,
					behavior, builder, LyInfoTokens.getCompoundActionStyle(style), sizes,
					LyInfoTokens.getCompoundActionContext(context),
					LyInfoTokens.getCompoundActionClientContext(clientContext), target);
				result = compoundAction;
			}

			if (abstractAction)
				((LyAction)result).setAbstract(true);
		}
		else if (tag == LyInfoTokens.TAG_CONSTRAINT)
		{
			String  typeC = node.getString(LyInfoTokens.ATT_TYPE);
			String  colSpanStr = node.getString(LyInfoTokens.ATT_COL_SPAN);
			int     colSpan = ((colSpanStr == null) ? LyFieldConstraint.DEFAULT_VALUE
													: node.getInt(LyInfoTokens.ATT_COL_SPAN));
			String  rowSpanStr = node.getString(LyInfoTokens.ATT_ROW_SPAN);
			int     rowSpan = ((rowSpanStr == null) ? LyFieldConstraint.DEFAULT_VALUE
													: node.getInt(LyInfoTokens.ATT_ROW_SPAN));
			boolean remainder = node.getBoolean(LyInfoTokens.ATT_REMAINDER);
			String  colWeightStr = node.getString(LyInfoTokens.ATT_COL_WEIGHT);
			int     colWeight = ((colWeightStr == null) ? LyFieldConstraint.DEFAULT_VALUE
														: node.getInt(LyInfoTokens.ATT_COL_WEIGHT));
			String  rowWeightStr = node.getString(LyInfoTokens.ATT_ROW_WEIGHT);
			int     rowWeight = ((rowWeightStr == null) ? LyFieldConstraint.DEFAULT_VALUE
														: node.getInt(LyInfoTokens.ATT_ROW_WEIGHT));
			String  fill = node.getString(LyInfoTokens.ATT_FILL);
			String  hAlign = node.getString(LyInfoTokens.ATT_H_ALIGN);
			String  vAlign = node.getString(LyInfoTokens.ATT_V_ALIGN);

			result = new LyFieldConstraint(LyInfoTokens.getConstraintType(typeC), colSpan, rowSpan,
					remainder, colWeight, rowWeight, LyInfoTokens.getConstraintFill(fill),
					LyInfoTokens.getConstraintHAlign(hAlign),
					LyInfoTokens.getConstraintVAlign(vAlign));
		}
		else if (tag == LyInfoTokens.TAG_CONTROL)
		{
			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			int    priority = node.getInt(LyInfoTokens.ATT_PRIORITY);
			String level = node.getString(LyInfoTokens.ATT_LEVEL);
			LyInfo control = getInfo(id);

			if (control == null)
				result = new LyControl(id, javaName, LyInfoTokens.getControlLevel(level), priority, LyTag.Model.FULL);
			else
			{
				if (!(control instanceof LyControl))
					throw new SAXException("BF AppEngine : extern info " + id + " is not a control");
				if (control.isLoaded())
					throw new SAXException("BF AppEngine : control " + id + " already exists");

				((LyControl)control).construct(javaName, LyInfoTokens.getControlLevel(level),
					priority, LyTag.Model.FULL);
				result = control;
			}

	        String message = node.getString(LyInfoTokens.ATT_MESSAGE);

			((LyControl)result).setMessage(message);
		}
		else if (tag == LyInfoTokens.TAG_MODEL)
		{
			// Build the model
			String version = node.getString(LyInfoTokens.ATT__VERSION);
			String schema = node.getString(LyInfoTokens.ATT_SCHEMA);

			result = new LyModel(LyModel.TYPE_ALL, null, version, schema);
		}
		else if (tag == LyInfoTokens.TAG_DATE)
		{
			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			String behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			boolean withTimeZone = node.getBoolean(LyInfoTokens.ATT_WITHTIMEZONE);
			LyInfo fieldInfo = getInfo(id);

			if (fieldInfo == null)
				result = new LyDateFieldInfo(id, javaName, (LyDateFieldInfo)template,
					LyInfoTokens.getDateFieldInfoType(type), withTimeZone);
			else if (fieldInfo == template)
				result = ((LyFieldInfo)fieldInfo).cloneAsTemplate();
			else
			{
				if (!(fieldInfo instanceof LyDateFieldInfo))
					throw new SAXException("BF AppEngine : extern info " + id + " is not a date field info");
				if (fieldInfo.isLoaded())
					throw new SAXException("BF AppEngine : date field info " + id + " already exists");

				((LyDateFieldInfo)fieldInfo).construct(javaName, (LyDateFieldInfo)template,
					LyInfoTokens.getDateFieldInfoType(type), withTimeZone);
				result = fieldInfo;
			}

			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);
		}
		else if (tag == LyInfoTokens.TAG_ENCODING)
			result = node.getString(LyInfoTokens.ATT_JAVA_NAME);
		else if (tag == LyInfoTokens.TAG_F_BIND)
		{
			String  value = node.getString(LyInfoTokens.ATT_VALUE);
			String  def = node.getString(LyInfoTokens.ATT_DEFAULT);
			String  cbind = node.getString(LyInfoTokens.ATT_CBIND);
			boolean isId = node.getBoolean(LyInfoTokens.ATT_ID);
			String  javaType = node.getString(LyInfoTokens.ATT_JAVA_TYPE);
			String  javaEncoder = node.getString(LyInfoTokens.ATT_JAVA_ENCODER);
			String  peerType = node.getString(LyInfoTokens.ATT_PEER_TYPE);

			result = new LyFieldBinding(value, def, cbind, isId, javaType, javaEncoder, peerType);
		}
		else if (tag == LyInfoTokens.TAG_F_BINDS)
			result = new LyFieldsBinding();
		else if (tag == LyInfoTokens.TAG_F_BIND_INFO)
		{
			String value = node.getString(LyInfoTokens.ATT_IDREF);

			result = new LyFieldInfoBinding(value);
		}
		else if (tag == LyInfoTokens.TAG_F_BIND_LDAP)
		{
			String  value = node.getString(LyInfoTokens.ATT_VALUE);
			boolean parent = node.getBoolean(LyInfoTokens.ATT_PARENT);
			boolean binary = node.getBoolean(LyInfoTokens.ATT_BINARY);
			result = new LyLdapFieldBinding(value, parent,binary);
		}
		else if (tag == LyInfoTokens.TAG_F_MARKS)
			result = buildMarks(tag, node);
		else if (tag == LyInfoTokens.TAG_FIELD)
		{
			template = getInfo(type);
			if (template == null)
				throw new SAXException("BF AppEngine : type " + type + " is not a valid field type for field " + id);

			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);

			template.load();

			if (template instanceof LyArrayFieldInfo)
			{
				LyArrayFieldInfo model = (LyArrayFieldInfo)template;

				result = new LyArrayFieldInfo(id, javaName, model, model.getMinCardinality(),
						model.getMaxCardinality());
			}
			else if (template instanceof LyChoiceFieldInfo)
			{
				LyChoiceFieldInfo model = (LyChoiceFieldInfo)template;

				result = new LyChoiceFieldInfo(id, javaName, model, model.isMultiple(),
						model.getSortBy(), model.getType(), model.isNot());
			}
			else if (template instanceof LyDateFieldInfo)
			{
				LyDateFieldInfo model = (LyDateFieldInfo)template;

				result = new LyDateFieldInfo(id, javaName, model, model.getType());
			}
			else if (template instanceof LyFileFieldInfo)
			{
				LyFileFieldInfo model = (LyFileFieldInfo)template;

				result = new LyFileFieldInfo(id, javaName, model, model.getMimeTypes(),
						model.getAttachmentDirectory(), model.getAttachmentHandler(),
						model.getDirectory(), model.getFilter());
			}
			else if (template instanceof LyNumberFieldInfo)
			{
				LyNumberFieldInfo model = (LyNumberFieldInfo)template;

				result = new LyNumberFieldInfo(id, javaName, model, model.getType(),
						model.isModulo());
			}
			else if (template instanceof LyStructFieldInfo)
			{
				LyStructFieldInfo struct, model = (LyStructFieldInfo)template;

				struct = new LyStructFieldInfo(id);
				struct.setFieldTag(true);
				struct.construct(javaName, null, model, model.getClassName(), model.isAbstract());
				result = struct;
			}
			else if (template instanceof LyRelationFieldInfo)
			{
				LyRelationFieldInfo model = (LyRelationFieldInfo)template;

				result = new LyRelationFieldInfo(id, javaName, null, model, model.isMultiple(),
						model.isOrdered(), model.getDirection(), model.getType(),
						model.getUserData(), model.ignoresContext(),
						model.isToolbarDisabled(), model.getShowAs());
			}
			else if (template instanceof LyTextFieldInfo)
			{
				LyTextFieldInfo model = (LyTextFieldInfo)template;

				result = new LyTextFieldInfo(id, javaName, model, model.isMultiple());
			}

			String behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);

			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);

			boolean calculated = node.getBoolean(LyInfoTokens.ATT_CALCULATED);

			((LyFieldInfo)result).setCalculated(calculated);
			((LyFieldInfo)result).setFieldTag(true);
			completeFieldInfo((LyFieldInfo)result, node);
		}
		else if (tag == LyInfoTokens.TAG_FILE)
		{
			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			String mimeTypes = node.getString(LyInfoTokens.ATT_MIME_TYPES);
			String attachmentHandler = node.getString(LyInfoTokens.ATT_ATTACHMENT_HANDLER);
			String attachmentDirectory = node.getString(LyInfoTokens.ATT_ATTACHMENT_DIRECTORY);
			String directory = node.getString(LyInfoTokens.ATT_DIRECTORY);
			String filter = node.getString(LyInfoTokens.ATT_FILTER);
			String behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			LyInfo fieldInfo = getInfo(id);

			if (fieldInfo == null)
				result = new LyFileFieldInfo(id, javaName, (LyFileFieldInfo)template, mimeTypes,
						attachmentHandler, attachmentDirectory, directory, filter);
			else if (fieldInfo == template)
				result = ((LyFieldInfo)fieldInfo).cloneAsTemplate();
			else
			{
				if (!(fieldInfo instanceof LyFileFieldInfo))
					throw new SAXException("BF AppEngine : info " + id + " is not a file field info");
				if (fieldInfo.isLoaded())
					throw new SAXException("BF AppEngine : file field info " + id + " already exists");

				((LyFileFieldInfo)fieldInfo).construct(javaName, (LyFileFieldInfo)template,
					mimeTypes, attachmentHandler, attachmentDirectory, directory, filter);
				result = fieldInfo;
			}

			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);
		}
		else if (tag == LyInfoTokens.TAG_FORMAT)
		{
			String  value = node.getString(LyInfoTokens.ATT_VALUE);
			boolean editable = node.getBoolean(LyInfoTokens.ATT_EDITABLE);
			int     nCols = node.getInt(LyInfoTokens.ATT_NCOLS);
			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);

			result = new LyFormat(value, editable, nCols, javaName);
		}
		else if (tag == LyInfoTokens.TAG_GEOMETRY)
		{
			int x = node.getInt(LyInfoTokens.ATT_X);
			int y = node.getInt(LyInfoTokens.ATT_Y);
			int width = node.getInt(LyInfoTokens.ATT_WIDTH);
			int height = node.getInt(LyInfoTokens.ATT_HEIGHT);
			int minWidth = node.getInt(LyInfoTokens.ATT_MIN_WIDTH);
			int minHeight = node.getInt(LyInfoTokens.ATT_MIN_HEIGHT);

			result = new LyGeometry(x, y, width, height, minWidth, minHeight);
		}
		else if (tag == LyInfoTokens.TAG_GROUP)
		{
			String value = node.getString(LyInfoTokens.ATT_VALUE);
			String state = node.getString(LyInfoTokens.ATT_STATE);

			if (id != null)
				result = new LyFieldGroup(id, value, LyInfoTokens.getFieldGroupState(state));
			else
				result = value;
		}
		else if (tag == LyInfoTokens.TAG_INCLUDE)
		{
			String applicationId = node.getString(LyInfoTokens.ATT_APP_ID);
			LyInfo info = _application.getInfo((id == null) ? ("_Include" + applicationId) : id);

			if (info instanceof LyInclude)
				result = info;
			else if (info != null)
				throw new SAXException("BF AppEngine : invalid include \"" + id
					+ "\", an info with the same id was found but with a different type.");
			else
			{
				String  xmlFile = node.getString(LyInfoTokens.ATT_XML_FILE);
				String  version = node.getString(LyInfoTokens.ATT__VERSION);
				boolean load = node.getBoolean(LyInfoTokens.ATT_LOAD);

				result = new LyInclude(applicationId, xmlFile, version, load);
			}
		}
		else if (tag == LyInfoTokens.TAG_INFO)
		{
			LyInfo info = _application.getInfo(id);

			if (info == null)
			{
				LyInfo parent = (LyInfo)node.getParent().getResult();

				if ((parent instanceof LyInclude)
						&& ((LyInclude)parent).isLoaded())
				{
					LyEnvironment env = _application.getEnvironment();

					env.logWarning("BF AppEngine : info with " + id
						+ " was not found in the specified include file " + parent);
					type = null;
				}
			}

			if (type == null)
				result = null;
			else if (type.equals("CLASS"))
			{
				if (info instanceof LyClassInfo)
					result = info;
				else if (info == null)
					result = new LyClassInfo(id);
			}
			else if (type.equals("COMPOUND_ACTION"))
			{
				if (info instanceof LyCompoundAction)
					result = info;
				else if (info == null)
					result = new LyCompoundAction(id);
			}
			else if (type.equals("TAB_ACTION"))
			{
				if (info instanceof LyTabAction)
					result = info;
				else if (info == null)
					result = new LyTabAction(id);
			}
			else if (type.equals("ACTION"))
			{
				if (info instanceof LyAction)
					result = info;
				else if (info == null)
					result = new LyAction(id);
			}
			else if (type.equals("FILTER"))
			{
				if (info instanceof LyFilter)
					result = info;
				else if (info == null)
					result = new LyFilter(id);
			}
			else if (type.equals("ARRAY"))
			{
				if (info instanceof LyArrayFieldInfo)
					result = info;
				else if (info == null)
					result = new LyArrayFieldInfo(id);
			}
			else if (type.equals("CHOICE"))
			{
				if (info instanceof LyChoiceFieldInfo)
					result = info;
				else if (info == null)
					result = new LyChoiceFieldInfo(id);
			}
			else if (type.equals("DATE"))
			{
				if (info instanceof LyDateFieldInfo)
					result = info;
				else if (info == null)
					result = new LyDateFieldInfo(id);
			}
			else if (type.equals("FILE"))
			{
				if (info instanceof LyFileFieldInfo)
					result = info;
				else if (info == null)
					result = new LyFileFieldInfo(id);
			}
			else if (type.equals("NUMBER"))
			{
				if (info instanceof LyNumberFieldInfo)
					result = info;
				else if (info == null)
					result = new LyNumberFieldInfo(id);
			}
			else if (type.equals("RELATION"))
			{
				if (info instanceof LyRelationFieldInfo)
					result = info;
				else if (info == null)
					result = new LyRelationFieldInfo(id);
			}
			else if (type.equals("STRUCT"))
			{
				if (info instanceof LyStructFieldInfo)
					result = info;
				else if (info == null)
					result = new LyStructFieldInfo(id);
			}
			else if (type.equals("TEXT"))
			{
				if (info instanceof LyTextFieldInfo)
					result = info;
				else if (info == null)
					result = new LyTextFieldInfo(id);
			}
			else if (type.equals("FILTER"))
			{
				if (info instanceof LyFilter)
					result = info;
				else if (info == null)
					result = new LyFilter(id);
			}
			else if (type.equals("CHARTYPE"))
			{
				if (info instanceof LyInfoCharType)
					result = info;
				else if (info == null)
					result = new LyInfoCharType(id);
			}
			else if (type.equals("CONTROL"))
			{
				if (info instanceof LyControl)
					result = info;
				else if (info == null)
					result = new LyControl(id);
			}
			else if (type.equals("REGEX"))
			{
				if (info instanceof LyRegExp)
					result = info;
				else if (info == null)
					result = new LyRegExp(id);
			}
			else if (type.equals("UNIT"))
			{
				if (info instanceof LyUnit)
					result = info;
				else if (info == null)
					result = new LyUnit(id);
			}
			else if (type.equals("UNITS"))
			{
				if (info instanceof LyUnits)
					result = info;
				else if (info == null)
					result = new LyUnits(id);
			}
			else if (type.equals("L_GEN"))
			{
				if (info instanceof LyGenLocation)
					result = info;
				else if (info == null)
					result = new LyGenLocation(id);
			}
			else if (type.equals("L_FILE"))
			{
				if (info instanceof LyFileLocation)
					result = info;
				else if (info == null)
					result = new LyFileLocation(id);
			}
			else if (type.equals("L_SQL"))
			{
				if (info instanceof LySqlLocation)
					result = info;
				else if (info == null)
					result = new LySqlLocation(id);
			}
			else if (type.equals("L_LDAP"))
			{
				if (info instanceof LyLdapLocation)
					result = info;
				else if (info == null)
					result = new LyLdapLocation(id);
			}

			if ((result == null) && (info != null))
				throw new SAXException("BF AppEngine : invalid extern info \"" + id
					+ "\" (type " + type + "), an info with the same id was found but with a different type.");
		}
		else if (tag == LyInfoTokens.TAG_LABEL)
		{
			String value = node.getString(LyInfoTokens.ATT_VALUE);
			String subType = node.getString(LyInfoTokens.ATT_SUB_TYPE);

			result = new LyInfoLabel(id, value, LyInfoTokens.getInfoLabelType(type), subType);
			((LyInfo)result).setModelType(LyTag.Model.FULL);
		}
		else if (tag == LyInfoTokens.TAG_IMPORT)
			result = new LyFieldInfoImport();
		else if (tag == LyInfoTokens.TAG_JOIN)
			result = new LyJoinDaemon();
		else if (tag == LyInfoTokens.TAG_JOIN_DAEMON)
			result = new LyJoinDaemon();
		else if (tag == LyInfoTokens.TAG_L_FILE)
		{
			LyLocation location = (LyLocation)_application.getInfo(id);

			if ((location != null)
					&& (location.getClass() == LyFileLocation.class))
				return null;

			String url = node.getString(LyInfoTokens.ATT_URL);
			String valueSep = node.getString(LyInfoTokens.ATT_VALUE_SEP);
			String fieldSep = node.getString(LyInfoTokens.ATT_FIELD_SEP);

			if (id == null)
				id = _application.getDefaultLocationId();

			location = new LyFileLocation(id, url, LyInfoTokens.getFileLocationType(type),
				fieldSep, valueSep);
			result = location;
		}
		else if (tag == LyInfoTokens.TAG_L_LDAP)
		{
			LyLocation location = (LyLocation)_application.getInfo(id);

			if ((location != null)
					&& (location.getClass() == LyLdapLocation.class))
				return null;

			String  url = node.getString(LyInfoTokens.ATT_URL);
			String  baseObjectName = node.getString(LyInfoTokens.ATT_BASE_OBJECT_NAME);
			String  user = node.getString(LyInfoTokens.ATT_USER);
			String  password = node.getString(LyInfoTokens.ATT_PASSWORD);
			String securityType = node.getString(LyInfoTokens.ATT_SECURITY_TYPE);

			if (id == null)
				id = _application.getDefaultLocationId();

			location = new LyLdapLocation(id, url, baseObjectName, user, password, securityType);
			result = location;
		}
		else if (tag == LyInfoTokens.TAG_L_SQL)
		{
			LyLocation location = (LyLocation)_application.getInfo(id);

			if ((location != null)
					&& (location.getClass() == LySqlLocation.class))
				return null;

			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			String  url = node.getString(LyInfoTokens.ATT_URL);
			String  driver = node.getString(LyInfoTokens.ATT_DRIVER);
			String  user = node.getString(LyInfoTokens.ATT_USER);
			String  password = node.getString(LyInfoTokens.ATT_PASSWORD);
			String  dbName = node.getString(LyInfoTokens.ATT_DBNAME);
			String  nbConnectionStr = node.getString(LyInfoTokens.ATT_NB_CONNECTION);
			String  nbConnection = "1";
			String  valueSep = node.getString(LyInfoTokens.ATT_VALUE_SEP);
			String trim = node.getString(LyInfoTokens.ATT_TRIM);
			String	metadata = node.getString(LyInfoTokens.ATT_METADATA);

			if (nbConnectionStr != null)
			{
				LyEnvironment env = _application.getEnvironment();
				String		  str = env.getEnv(nbConnectionStr);

				if (str != null)
					nbConnection = str;
				else
					nbConnection = node.getString(LyInfoTokens.ATT_NB_CONNECTION);
			}

			if (id == null)
				id = _application.getDefaultLocationId();

			location = new LySqlLocation(id, javaName, url, driver, user, password, dbName,
					nbConnection, valueSep, trim, metadata);
			result = location;
		}
		else if (tag == LyInfoTokens.TAG_L_GEN)
		{
			LyLocation location = (LyLocation)_application.getInfo(id);

			if ((location != null)
					&& (location.getClass() == LyGenLocation.class))
				return null;

			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			boolean persistent = node.getBoolean(LyInfoTokens.ATT_PERSISTENT);

			location = new LyGenLocation(id, javaName, persistent);
			result = location;
		}
		else if (tag == LyInfoTokens.TAG_NOTIFIER)
		{
			LyNotifier notifier = (LyNotifier)_application.getInfo(id);

			if ((notifier != null)
					&& (notifier.getClass() == LyNotifier.class))
				return notifier;

			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			String properties = node.getString(LyInfoTokens.ATT_PROPERTIES);

			if (id == null)
				id = _application.getId();

			result = new LyNotifier(id, javaName, properties);
		}
		else if (tag == LyInfoTokens.TAG_NUMBER)
		{
			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			boolean modulo = node.getBoolean(LyInfoTokens.ATT_MODULO);
			LyInfo  fieldInfo = getInfo(id);

			if (fieldInfo == null)
				result = new LyNumberFieldInfo(id, javaName, (LyNumberFieldInfo)template,
						LyInfoTokens.getNumberFieldInfoType(type), modulo);
			else if (fieldInfo == template)
				result = ((LyFieldInfo)fieldInfo).cloneAsTemplate();
			else
			{
				if (!(fieldInfo instanceof LyNumberFieldInfo))
					throw new SAXException("BF AppEngine : info " + id + " is not a number field info");
				if (fieldInfo.isLoaded())
					throw new SAXException("BF AppEngine : number field info " + id + " already exists");

				((LyNumberFieldInfo)fieldInfo).construct(javaName, (LyNumberFieldInfo)template,
					LyInfoTokens.getNumberFieldInfoType(type), modulo);
				result = fieldInfo;
			}

			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);
		}
		else if (tag == LyInfoTokens.TAG_OPTION)
		{
			String  name = node.getString(LyInfoTokens.ATT_NAME);
			String  alias = node.getString(LyInfoTokens.ATT_ALIAS);
			String  value = node.getString(LyInfoTokens.ATT_VALUE);
			boolean def = node.getBoolean(LyInfoTokens.ATT_DEFAULT);
			boolean disable = node.getBoolean(LyInfoTokens.ATT_DISABLE);
			boolean unknown = node.getBoolean(LyInfoTokens.ATT_UNKNOWN);
			boolean table = node.getBoolean(LyInfoTokens.ATT_TABLE);

			result = new LyChoiceOption(id, name, alias, value, def, disable, unknown, table);
		}
		else if (tag == LyInfoTokens.TAG_OPTIONS)
		{
			String operation = node.getString(LyInfoTokens.ATT_OPERATION);

			if ((operation == LyInfoTokens.OPER_LIST_RESET)
					|| (operation == LyInfoTokens.OPER_LIST_ADD))
			{
				String name = node.getString(LyInfoTokens.ATT_NAME);
				String alias = node.getString(LyInfoTokens.ATT_ALIAS);
				String value = node.getString(LyInfoTokens.ATT_VALUE);

				result = new LyChoiceOptionGroup(id, name, alias, value);
			}
			else
				result = new Vector<LyChoiceOption>();
		}
		else if (tag == LyInfoTokens.TAG_PARAMETER)
			result = new LyParameter(node.getString(LyInfoTokens.ATT_NAME),
					node.getString(LyInfoTokens.ATT_VALUE));
		else if (tag == LyInfoTokens.TAG_PARAMETER_DECL)
		{
			String  name = node.getString(LyInfoTokens.ATT_NAME);
			boolean required = node.getBoolean(LyInfoTokens.ATT_REQUIRED);
			String  domain = node.getString(LyInfoTokens.ATT_DOMAIN);

			result = new LyParameterDecl(name, LyInfoTokens.getParameterDeclType(type), required,
					domain);
		}
		else if (tag == LyInfoTokens.TAG_PATH_REF)
		{
			id = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(id);

			if (!(result instanceof LyPath))
				throw new SAXException("BF AppEngine : info " + id + " is not a valid path");
		}
		else if (tag == LyInfoTokens.TAG_PREFERENCE)
		{
	        String  idref = node.getString(LyInfoTokens.ATT_IDREF);
			String  name = node.getString(LyInfoTokens.ATT_NAME);
			String  user = node.getString(LyInfoTokens.ATT_USER);
			int     nbObjects = node.getInt(LyInfoTokens.ATT_NB_OBJECTS);
			boolean defaultPref = node.getBoolean(LyInfoTokens.ATT_DEFAULT);
			boolean defaultForMobilePref = node.getBoolean(LyInfoTokens.ATT_DEFAULT_FOR_MOBILE);
			boolean forPortletPref = node.getBoolean(LyInfoTokens.ATT_FOR_PORTLET);
			boolean readOnly = node.getBoolean(LyInfoTokens.ATT_READ_ONLY);
			String  lastFileName = node.getString(LyInfoTokens.ATT_LAST_FILE_NAME);
			String  columnSizes = node.getString(LyInfoTokens.ATT_COLUMN_SIZES);
			String	showAs = node.getString(LyInfoTokens.ATT_SHOW_AS);

			result = new LyPreference(id, idref, name, null, nbObjects, user, defaultPref, columnSizes);
			((LyPreference)result).setDefaultForMobile(defaultForMobilePref);
			((LyPreference)result).setForPortlet(forPortletPref);
			((LyPreference)result).setReadOnly(readOnly);
			((LyPreference)result).setLastFileName(lastFileName);
			((LyPreference)result).setShowAs(LyInfoTokens.getPreferenceShowAs(showAs));
		}
		else if (tag == LyInfoTokens.TAG_TREEMAP_CONFIGURATION)
		{
			String  name = node.getString(LyInfoTokens.ATT_NAME);
			String  user = node.getString(LyInfoTokens.ATT_USER);
			int     levels = node.getInt(LyInfoTokens.ATT_HIERARCHY_LEVELS);
			String  aggregationType = node.getString(LyInfoTokens.ATT_AGGREGATION_TYPE);
			boolean filterActive = node.getBoolean(LyInfoTokens.ATT_FILTER_ACTIVE);
			boolean filterShowAll = node.getBoolean(LyInfoTokens.ATT_FILTER_SHOW_ALL);
			String  minColor = node.getString(LyInfoTokens.ATT_MIN_COLOR);
			String  maxColor = node.getString(LyInfoTokens.ATT_MAX_COLOR);
			String  thresholdColors = node.getString(LyInfoTokens.ATT_THRESHOLD_COLORS);
			String  borderSizes = node.getString(LyInfoTokens.ATT_BORDER_SIZES);
			String  labelStyles = node.getString(LyInfoTokens.ATT_LABEL_STYLES);

			result = new LyTreemapConfiguration(id, null, name, levels,
					LyInfoTokens.getTreemapConfigurationAggregationType(aggregationType),
					filterActive, filterShowAll, minColor, maxColor,
					LyInfoTokens.getTreemapConfigurationThresholdColors(thresholdColors),
					borderSizes, labelStyles, user);
		}
		else if (tag == LyInfoTokens.TAG_PROJECT)
		{
			String    version = node.getString(LyInfoTokens.ATT__VERSION);
			String    appBehavior = node.getString(LyInfoTokens.ATT_APP_BEHAVIOR);
			String    sessionBehavior = node.getString(LyInfoTokens.ATT_SESSION_BEHAVIOR);
			String    name = node.getString(LyInfoTokens.ATT_NAME);
			String    tip = node.getString(LyInfoTokens.ATT_TIP);
			String    prefix = node.getString(LyInfoTokens.ATT_PREFIX);

			LyProject project = new LyProject(id, (LyProject)template, version, appBehavior,
					sessionBehavior);

			project.setName(name);
			project.setTip(tip);
			project.setFilePrefix(prefix);
			result = project;
		}
		else if (tag == LyInfoTokens.TAG_REGEX)
		{
			String value = node.getString(LyInfoTokens.ATT_VALUE);
			LyInfo regex = getInfo(id);

			if (regex == null)
				result = new LyRegExp(id, value);
			else
			{
				if (!(regex instanceof LyRegExp))
					throw new SAXException("BF AppEngine : extern info " + id + " is not a regexp");
				if (regex.isLoaded())
					throw new SAXException("BF AppEngine : regex " + id + " already exists");

				((LyRegExp)regex).construct(value);
				result = regex;
			}
		}
		else if (tag == LyInfoTokens.TAG_RELATION)
		{
			String extendsId = node.getString(LyInfoTokens.ATT_EXTENDS);
			LyInfo extendedRelation = getInfo(extendsId);

			// Forbid forward reference on extends
			if ((extendsId != null)
					&& (extendsId.length() > 0)
					&& (extendedRelation == null))
				throw new IllegalStateException(
					"BF AppEngine : Forward reference to super-relation is not supported. Relation '"
					+ id + "' with super-relation : '" + extendsId + "'.");

			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			boolean abstractRelation = node.getBoolean(LyInfoTokens.ATT_ABSTRACT);
			boolean multiple = node.getBoolean(LyInfoTokens.ATT_MULTIPLE);
			boolean ordered = node.getBoolean(LyInfoTokens.ATT_ORDERED);
			boolean calculated = node.getBoolean(LyInfoTokens.ATT_CALCULATED);
			String  direction = node.getString(LyInfoTokens.ATT_DIRECTION);
			String  userData = node.getString(LyInfoTokens.ATT_USER_DATA);
			String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			boolean noContext = node.getBoolean(LyInfoTokens.ATT_NOCONTEXT);
			boolean noToolbar = node.getBoolean(LyInfoTokens.ATT_NOTOOLBAR);
			String  showAs = node.getString(LyInfoTokens.ATT_SHOW_AS);
			LyInfo  fieldInfo = getInfo(id);

			if (fieldInfo == null)
				result = new LyRelationFieldInfo(id, javaName,
						(LyRelationFieldInfo)extendedRelation, (LyRelationFieldInfo)template,
						multiple, ordered, LyInfoTokens.getRelationFieldInfoDirection(direction),
						LyInfoTokens.getRelationFieldInfoType(type), userData, noContext,
						noToolbar, LyInfoTokens.getRelationFieldInfoShowAs(showAs));
			else if (fieldInfo == template)
				result = ((LyFieldInfo)fieldInfo).cloneAsTemplate();
			else
			{
				if (!(fieldInfo instanceof LyRelationFieldInfo))
					throw new SAXException("BF AppEngine : extern info " + id + " is not a relation field info");
				if (fieldInfo.isLoaded())
					throw new SAXException("BF AppEngine : relation field info " + id + " already exists");

				((LyRelationFieldInfo)fieldInfo).construct(javaName,
					(LyRelationFieldInfo)extendedRelation, (LyRelationFieldInfo)template, multiple,
					ordered, LyInfoTokens.getRelationFieldInfoDirection(direction),
					LyInfoTokens.getRelationFieldInfoType(type), userData, noContext,
					noToolbar, LyInfoTokens.getRelationFieldInfoShowAs(showAs));
				result = fieldInfo;
			}

			((LyFieldInfo)result).setCalculated(calculated);
			((LyRelationFieldInfo)result).setAbstract(abstractRelation);
			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);
		}
		else if (tag == LyInfoTokens.TAG_RESOURCE)
			result = new LyActionResource(LyInfoTokens.getResourceType(type),
					node.getString(LyInfoTokens.ATT_NAME), node.getString(LyInfoTokens.ATT_VALUE));
		else if (tag == LyInfoTokens.TAG_REVERSE)
			result = new LyReverseDaemon();
		else if (tag == LyInfoTokens.TAG_RULE)
		{
			String			 str = node.getString(LyInfoTokens.ATT_OPERATION);
			int				 priority = node.getInt(LyInfoTokens.ATT_PRIORITY);
			LyRule.Operation operation = LyInfoTokens.getRuleOperation(str);
			boolean			 onCheck = node.getBoolean(LyInfoTokens.ATT_ON_CHECK);
			String			 value = node.getString(LyInfoTokens.ATT_VALUE);

			result = new LyRule(id, null, priority, operation, onCheck, value,
			        LyTag.Model.FULL);
		}
		else if (tag == LyInfoTokens.TAG_STRUCT)
		{
			String extendsId = node.getString(LyInfoTokens.ATT_EXTENDS);
			LyInfo extendedStruct = getInfo(extendsId);

			// Forbid forward reference on extends
			if ((extendsId != null)
					&& (extendsId.length() > 0)
					&& (extendedStruct == null))
				throw new IllegalStateException(
					"BF AppEngine : Forward reference to super-struct is not supported. Struct '"
					+ id + "' with super-struct : '" + extendsId + "'.");

			// Create relation field info
			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			String  className = node.getString(LyInfoTokens.ATT_CLASS_NAME);
			boolean abstractClass = node.getBoolean(LyInfoTokens.ATT_ABSTRACT);
			String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			String  classBehavior = node.getString(LyInfoTokens.ATT_CLASS_BEHAVIOR);
			LyInfo  fieldInfo = getInfo(id);

			if (fieldInfo == null)
				result = new LyStructFieldInfo(id, javaName, (LyStructFieldInfo)extendedStruct,
						(LyStructFieldInfo)template, className, abstractClass);
			else if (fieldInfo == template)
				result = ((LyFieldInfo)fieldInfo).cloneAsTemplate();
			else
			{
				if (!(fieldInfo instanceof LyStructFieldInfo))
					throw new SAXException("BF AppEngine : extern info " + id + " is not a struct field info");
				if (fieldInfo.isLoaded())
					throw new SAXException("BF AppEngine : struct field info " + id + " already exists");

				((LyStructFieldInfo)fieldInfo).construct(javaName,
					(LyStructFieldInfo)extendedStruct, (LyStructFieldInfo)template, className,
					abstractClass);
				result = fieldInfo;
			}

			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);
			if (classBehavior != null)
			{
				LyClassInfo classInfo = ((LyStructFieldInfo)result).getClassInfo();

				classInfo.setBehaviorClassName(classBehavior);
			}
		}
		else if (tag == LyInfoTokens.TAG_TAB)
		{
			String value = node.getString(LyInfoTokens.ATT_VALUE);

			if (id != null)
				result = new LyFieldTab(id, value);
			else
				result = value;
		}
		else if (tag == LyInfoTokens.TAG_TAB_ACTION)
		{
			String  controller = node.getString(LyInfoTokens.ATT_CONTROLLER);
			String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			String  builder = node.getString(LyInfoTokens.ATT_BUILDER);
			String  target = node.getString(LyInfoTokens.ATT_TARGET);
			short   clientContext = (short)node.getInt(LyInfoTokens.ATT_CLIENT_CONTEXT, LyTabAction.CONTEXT_ALL);
			boolean abstractAction = node.getBoolean(LyInfoTokens.ATT_ABSTRACT);
			LyInfo  tabAction = getInfo(id);

			if (tabAction == null)
				result = new LyTabAction(id, (LyAction)template, controller, behavior, builder,
						clientContext, target, LyInfoTokens.getTabActionType(type));
			else if (tabAction == template)
				result = ((LyTabAction)tabAction).cloneAsTemplate();
			else
			{
				if (!(tabAction instanceof LyTabAction))
					throw new SAXException("BF AppEngine : extern info " + id + " is not a tab action");
				if (((LyTabAction)tabAction).isLoaded())
					throw new SAXException("BF AppEngine : tab action " + id + " already exists");

				((LyTabAction)tabAction).construct((LyAction)template, controller, behavior,
					builder, clientContext, target, LyInfoTokens.getTabActionType(type));
				result = tabAction;
			}

			if (abstractAction)
				((LyAction)result).setAbstract(true);
		}
		else if (tag == LyInfoTokens.TAG_TEXT)
		{
			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			boolean multiple = node.getBoolean(LyInfoTokens.ATT_MULTIPLE);
			boolean multilingual = node.getBoolean(LyInfoTokens.ATT_MULTILINGUAL);
			String  behavior = node.getString(LyInfoTokens.ATT_BEHAVIOR);
			LyInfo  fieldInfo = getInfo(id);

			if (fieldInfo == null)
				result = new LyTextFieldInfo(id, javaName, (LyTextFieldInfo)template,
				    multiple, LyInfoTokens.getTextFieldInfoType(type), multilingual);
			else if (fieldInfo == template)
				result = ((LyFieldInfo)fieldInfo).cloneAsTemplate();
			else
			{
				if (!(fieldInfo instanceof LyTextFieldInfo))
					throw new SAXException("BF AppEngine : extern info " + id + " is not a text field info");
				if (fieldInfo.isLoaded())
					throw new SAXException("BF AppEngine : text field info " + id + " already exists");

				((LyTextFieldInfo)fieldInfo).construct(javaName, (LyTextFieldInfo)template,
					multiple, LyInfoTokens.getTextFieldInfoType(type), multilingual);
				result = fieldInfo;
			}

			if (behavior != null)
				((LyFieldInfo)result).setBehaviorClassName(behavior);
		}
		else if (tag == LyInfoTokens.TAG_UNION)
		{
			String value = node.getString(LyInfoTokens.ATT_VALUE);
			String value2 = node.getString(LyInfoTokens.ATT_VALUE2);

			result = new LyUnion(value, value2);
		}
		else if (tag == LyInfoTokens.TAG_UNIT)
		{
			String  factor = node.getString(LyInfoTokens.ATT_FACTOR);
			String  offset = node.getString(LyInfoTokens.ATT_OFFSET);
			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			String  value = node.getString(LyInfoTokens.ATT_VALUE);
			String  name = node.getString(LyInfoTokens.ATT_NAME);
			String  alias = node.getString(LyInfoTokens.ATT_ALIAS);
			boolean print = node.getBoolean(LyInfoTokens.ATT_PRINT);
			LyInfo  unit = getInfo(id);

			if ((factor == null) || (factor.length() == 0))
				factor = "1.0";

			if (unit == null)
			{
				result = new LyUnit(id, LyTools.doubleFromString(factor, 1.0),
						LyTools.doubleFromString(offset), javaName);
				((LyUnit)result).completeUnit(value, name, alias, print);
			}
			else
			{
				if (!(unit instanceof LyUnit))
					throw new SAXException("BF AppEngine : extern info " + id + " is not a unit");
				if (unit.isLoaded())
					throw new SAXException("BF AppEngine : unit " + id + " already exists");

				((LyUnit)unit).construct(LyTools.doubleFromString(factor, 1.0),
					LyTools.doubleFromString(offset), javaName);
				((LyUnit)unit).completeUnit(value, name, alias, print);
				result = unit;
			}
		}
		else if (tag == LyInfoTokens.TAG_UNITS)
		{
			String  javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);
			String  name = node.getString(LyInfoTokens.ATT_NAME);
			boolean convert = node.getBoolean(LyInfoTokens.ATT_CONVERT);
			LyInfo  units = getInfo(id);

			if (units == null)
			{
				result = new LyUnits(id, javaName);
				((LyUnits)result).setName(name);
				((LyUnits)result).setConvert(convert);
			}
			else
			{
				if (!(units instanceof LyUnits))
					throw new SAXException("BF AppEngine : extern info " + id + " is not a units list");
				if (units.isLoaded())
					throw new SAXException("BF AppEngine : units list " + id + " already exists");

				((LyUnits)units).setName(name);
				((LyUnits)units).construct(javaName);
				((LyUnits)units).setConvert(convert);
				result = units;
			}
		}
		else if (tag == LyInfoTokens.TAG_UNITS_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyUnits))
				throw new SAXException("BF AppEngine : invalid units ref : " + idref);

			String javaName = node.getString(LyInfoTokens.ATT_JAVA_NAME);

			if (javaName != null)
				((LyUnits)result).setJavaName(javaName);
		}
		else if (tag == LyInfoTokens.TAG_ACTIONS)
			result = new Vector<LyAction>();
		else if (tag == LyInfoTokens.TAG_CLASSES)
			result = new Vector<LyClassInfo>();
		else if (tag == LyInfoTokens.TAG_OPTIONS)
			result = new Vector<LyChoiceOption>();
		else if ((tag == LyInfoTokens.TAG_A_MARKS)
				|| (tag == LyInfoTokens.TAG_C_MARKS)
				|| (tag == LyInfoTokens.TAG_F_MARKS))
			result = buildMarks(tag, node);
		else if (tag == LyInfoTokens.TAG_MARK)
			result = node.getString(LyInfoTokens.ATT_NAME);
		else if (tag == LyInfoTokens.TAG_F_MARK)
		{
		    boolean   value = node.getBoolean(LyInfoTokens.ATT_VALUE);

		    result = "__" + node.getString(LyInfoTokens.ATT_ACTION)
		        + "." + node.getString(LyInfoTokens.ATT_MARK);
		    if (!value)
		        result = (String)result + "=false";
		}
		else
			result = super.buildElement(node, tag);

		return result;
	} // end method buildElement

	/**
	 * Last step in objects construction : finalize objectInfos. We need an extra step because info
	 * references can't be resolved before the document has been fully parsed, because our grammar
	 * enables forward references. We take advantage of this step to complete our objects.
	 *
	 * @param     parentResult Result of the parent node.
	 * @exception SAXException If an error is encountered during this final step.
	 */
	@Override
  void endParse(LyInfoNode node)
		throws SAXException
	{
		if (node == null)
			return;

		String tag = node.getTag();

		if (tag == LyInfoTokens.TAG_INFO)
		{
			// Iterate on each child node
			int		   i, n = node.getChildCount();
			LyInfoNode child;

			for (i = 0; i < n; i++)
			{
				child = node.getChild(i);
				endParse(child);
			}
			return;
		}

		super.endParse(node);
	}

	/**
	 * The purpose of this method is to resolve object references now that all valid object ids are
	 * known.
	 *
	 * @param     tag    Current tag.
	 * @param     result Current result.
	 * @return    New result.
	 * @exception SAXExpection If the object found doesn't belong to the appropriate class.
	 */
	@Override
  Object resolveObjectInfoRef(LyInfoNode node, String tag, Object result)
		throws SAXException
	{
		// Compute idrefs on tags ACTIONS, CLASSES, FIELDS, OPTIONS
		if ((tag == LyInfoTokens.TAG_ACTIONS)
			|| (tag == LyInfoTokens.TAG_CLASSES)
			|| (tag == LyInfoTokens.TAG_FIELDS)
			|| (tag == LyInfoTokens.TAG_OPTIONS))
		{
			if (result == null) // Nothing to do
				return null;

			String           idrefsAtt = node.getString(LyInfoTokens.ATT_IDREFS);
			String           operation = node.getString(LyInfoTokens.ATT_OPERATION);
			String[]         idrefs = null;
			int		         i, n = 0;
			LyInfo           info;

			if (idrefsAtt != null)
			{
				idrefs = LyString.stringToList(idrefsAtt, ' ');
				n = ((idrefs == null) ? 0 : idrefs.length);
			}

			for (i = 0; i < n; i++)
			{
				// Check out class reference
				info = getInfo(idrefs[i]);

				if ((tag == LyInfoTokens.TAG_ACTIONS)
						&& !(info instanceof LyAction))
				    raiseError("BF AppEngine : invalid action ref : " + idrefs[i]);
				else if ((tag == LyInfoTokens.TAG_CLASSES)
						&& !(info instanceof LyClassInfo))
				    raiseError("BF AppEngine : invalid class ref : " + idrefs[i]);
				else if ((tag == LyInfoTokens.TAG_FIELDS)
						&& !(info instanceof LyFieldInfo))
				    raiseError("BF AppEngine : invalid fieldInfo ref : " + idrefs[i]);
				else if ((tag == LyInfoTokens.TAG_OPTIONS)
						&& !(info instanceof LyChoiceOption))
				    raiseError("BF AppEngine : invalid option ref : " + idrefs[i]);
				else
				{
    				// Add reference to the result list
    				if (result instanceof LyChoiceOptionGroup)
    					((LyChoiceOptionGroup)result).addChoiceItem((LyChoiceOption)info);
    				else
    				{
    					@SuppressWarnings("unchecked")
    					Vector<Object> temp = (Vector<Object>)result;
    					temp.addElement(info);
    				}
    				if (operation != LyInfoTokens.OPER_LIST_REMOVE)
    					node.addExtraInfo(info);
				}
			}

			if (result instanceof LyChoiceOptionGroup)
			{
				String option = node.getString(LyInfoTokens.ATT_OPTION);

				if (option != null)
				{
					info = getInfo(option);
					if (info instanceof LyChoiceOption)
						((LyChoiceOptionGroup)result).setOption((LyChoiceOption)info);
					else
						throw new SAXException("BF AppEngine : info " + option + " is not a choice option");
				}
			}
		}
		// Fill up join informations
		else if (tag == LyInfoTokens.TAG_JOIN)
		{
			String str;
			LyInfo info;

			// joinClass
			str = node.getString(LyInfoTokens.ATT_JOIN_CLASS);
			info = getInfo(str);
			if (!(info instanceof LyClassInfo))
				throw new SAXException("BF AppEngine : invalid classInfo ref : " + str);
			((LyJoinDaemon)result).setJoinClass((LyClassInfo)info);

			// sourceRelation
			str = node.getString(LyInfoTokens.ATT_SOURCE_RELATION);
			info = getInfo(str);
			if (!(info instanceof LyRelationFieldInfo))
				throw new SAXException("BF AppEngine : invalid relationfieldInfo ref : " + str);
			else if (((LyRelationFieldInfo)info).isMultiple())
				throw new SAXException("BF AppEngine : relation fieldInfo must no be multiple : " + str);
			((LyJoinDaemon)result).setSourceRelation((LyRelationFieldInfo)info);

			// targetRelation
			str = node.getString(LyInfoTokens.ATT_TARGET_RELATION);
			info = getInfo(str);
			if (!(info instanceof LyRelationFieldInfo))
				throw new SAXException("BF AppEngine : invalid relationFieldInfo ref : " + str);
			else if (((LyRelationFieldInfo)info).isMultiple())
				throw new SAXException("BF AppEngine : relation fieldInfo must no be multiple : " + str);
			((LyJoinDaemon)result).setTargetRelation((LyRelationFieldInfo)info);
		}
		// Fill up reverse informations
		else if (tag == LyInfoTokens.TAG_REVERSE)
		{
			String str;
			LyInfo info;

			// reverseRelation
			str = node.getString(LyInfoTokens.ATT_REVERSE_RELATION);
			info = getInfo(str);
			if (!(info instanceof LyRelationFieldInfo))
				throw new SAXException("BF AppEngine : invalid relationfieldInfo ref : " + str);
			((LyReverseDaemon)result).setReverseRelation((LyRelationFieldInfo)info);
		}
		// Set filter on classBinding
		else if ((tag == LyInfoTokens.TAG_C_BIND)
				|| (tag == LyInfoTokens.TAG_C_BIND_LDAP))
		{
			// Is the location ref a LyLocation ?
			String idref = node.getString(LyInfoTokens.ATT_LOCATION);

			if ((idref != null)
				&& !idref.startsWith("LY_")) // Resolve
			{
				LyInfo location = getInfo(idref);

				if (!(location instanceof LyLocation))
					throw new SAXException("BF AppEngine : invalid location ref : " + idref);
			}
		}
		else if (tag == LyInfoTokens.TAG_C_BIND_INFO)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);
			LyInfo info = getInfo(idref);

			if (!(info instanceof LyClassInfo))
				throw new SAXException("BF AppEngine : invalid classInfo binding : " + idref);
		}
		else if (tag == LyInfoTokens.TAG_F_BIND_INFO)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);
			LyInfo info = getInfo(idref);

			if (!(info instanceof LyFieldInfo))
				throw new SAXException("BF AppEngine : invalid fieldInfo binding : " + idref);
		}
		else if (tag == LyInfoTokens.TAG_IMPORT)
		{
			String idref = node.getString(LyInfoTokens.ATT_RELATION);
			LyInfo info = getInfo(idref);

			if (!(info instanceof LyRelationFieldInfo))
				throw new SAXException("BF AppEngine : invalid relation on import : " + idref);
			((LyFieldInfoImport)result).setRelation((LyRelationFieldInfo)info);

			idref = node.getString(LyInfoTokens.ATT_CLASS);
			info = getInfo(idref);

			if ((info != null) && !(info instanceof LyClassInfo))
				throw new SAXException("BF AppEngine : invalid class on import : " + idref);
			((LyFieldInfoImport)result).setClassInfo((LyClassInfo)info);

			idref = node.getString(LyInfoTokens.ATT_IDREF);
			info = getInfo(idref);

			if (!(info instanceof LyFieldInfo))
				throw new SAXException("BF AppEngine : invalid field on import : " + idref);
			((LyFieldInfoImport)result).setFieldInfo((LyFieldInfo)info);
		}
		else if (tag == LyInfoTokens.TAG_PARAMETER)
		{
			if (result == null) // Nothing to do
				return null;

			String idrefsAtt = node.getString(LyInfoTokens.ATT_IDREFS);
			LyInfo info;

			if (idrefsAtt != null)
			{
				String[]		   idrefs = LyString.stringToList(idrefsAtt, ' ');
				int				   i, n = idrefs.length;
				LyInfoList<LyInfo> list = new LyInfoList<LyInfo>();

				for (i = 0; i < n; i++)
				{
					// Check out class reference
					info = getInfo(idrefs[i]);
					if (info == null)
						throw new SAXException("BF AppEngine : invalid info ref : " + idrefs[i]);

					// Add reference to the result list
					list.addInfo(info);
				}

				((LyParameter)result).setValue(list);
			}
			else
			{
				String idref = node.getString(LyInfoTokens.ATT_IDREF);

				if (idref != null)
				{
					info = getInfo(idref);
					if (info == null)
						throw new SAXException("BF AppEngine : invalid info ref : " + idref);

					// Add reference to the result list
					((LyParameter)result).setValue(info);
				}
			}
		}
		// Reference to a fieldInfo
		else if ((tag == LyInfoTokens.TAG_FIELD_REF)
				|| (tag == LyInfoTokens.TAG_CLASS_FIELD)
				|| (tag == LyInfoTokens.TAG_CONTEXT_FIELD))
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyFieldInfo))
				throw new SAXException("BF AppEngine : invalid fieldInfo ref : " + idref);
		}
		// Reference to a classInfo
		else if (tag == LyInfoTokens.TAG_CLASS_REF)
		{
			String idref;
			LyInfo classInfo, filter;

			idref = node.getString(LyInfoTokens.ATT_IDREF);
			classInfo = getInfo(idref);
			if (!(classInfo instanceof LyClassInfo))
				throw new SAXException("BF AppEngine : invalid classInfo ref : " + idref);

			idref = node.getString(LyInfoTokens.ATT_FILTER);
			filter = getInfo(idref);
			if ((filter != null) && !(filter instanceof LyFilter))
				throw new SAXException("BF AppEngine : invalid filter ref : " + idref);

			// build classInfoRef
			result = new LyClassRef((LyClassInfo)classInfo, (LyFilter)filter);
		}
		// Reference to a unit
		else if (tag == LyInfoTokens.TAG_UNIT_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyUnit))
				throw new SAXException("BF AppEngine : invalid unit ref : " + idref);
		}
		else if (tag == LyInfoTokens.TAG_FIELD)
		{
			String subTypes = node.getString(LyInfoTokens.ATT_SUB_TYPES);

			if ((subTypes != null) && !(result instanceof LyRelationFieldInfo))
			{
				_application.getEnvironment().logWarning(
					"BF AppEngine : subTypes attribute ignored on field "
					+ result + ", because this field is not a relation");
				return result;
			}

			if (subTypes == null)
/*
                if (result instanceof LyRelationFieldInfo)
                        throw new SAXException("BF AppEngine : missing subTypes attribute on field "
                                + result + ", because this field is a relation");
                else
*/
				return result; // Nothing to do

			LyRelationFieldInfo relation = (LyRelationFieldInfo)result;
			String[]		    idrefs = LyString.stringToList(subTypes, ' ');
			int				    i, n = ((idrefs == null) ? 0 : idrefs.length);
			LyInfo			    info;

			for (i = 0; i < n; i++)
			{
				// Check out subType reference
				info = getInfo(idrefs[i]);
				if (!(info instanceof LyRelationFieldInfo))
					throw new SAXException("BF AppEngine : invalid field ref : " + idrefs[i]);

				if (((LyRelationFieldInfo)info).isAbstract())
				{
					String relationStr = ((relation instanceof LyStructFieldInfo) ? "struct" : "relation");

					throw new IllegalStateException("BF AppEngine : " + relationStr + " \"" + relation.getId()
					    + "\" can't have an abstract subType : \"" + info.getId() + "\"");
				}

				relation.addSubType((LyRelationFieldInfo)info);
			}
		}
		else if (tag == LyInfoTokens.TAG_CHARTYPE_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyInfoCharType))
				throw new SAXException("BF AppEngine : invalid charType ref : " + idref);
		}
		// Reference to a regular expression
		else if (tag == LyInfoTokens.TAG_REGEX_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyRegExp))
				throw new SAXException("BF AppEngine : invalid reg exp ref : " + idref);
		}
		// Rule : Reference to a fieldInfo
		else if (tag == LyInfoTokens.TAG_RULE)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);
			LyInfo info = getInfo(idref);

			if ((info instanceof LyFieldInfo)
					|| (info instanceof LyClassInfo))
				((LyRule)result).setObjectInfo((LyObjectInfo)info);
			else
				throw new SAXException("BF AppEngine : invalid info ref : " + idref + " found on " + result);
		}
		// Reference to a rule
		else if (tag == LyInfoTokens.TAG_RULE_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyRule))
				throw new SAXException("BF AppEngine : invalid rule ref : " + idref);
		}
		// Fill up rootAction on Project
		else if (tag == LyInfoTokens.TAG_PROJECT)
		{
			String idref = node.getString(LyInfoTokens.ATT_ROOT_ACTION);

			if (idref != null)
			{
				LyInfo action = getInfo(idref);

				if (!(action instanceof LyAction))
					throw new SAXException("BF AppEngine : invalid action ref : " + idref);

				((LyProject)result).setRootAction((LyAction)action);
			}
		}
		// Reference to an action
		else if (tag == LyInfoTokens.TAG_ACTION_REF)
		{
			String id;
			String idref;
			LyInfo info;

			id = node.getString(LyInfoTokens.ATT_ID);
			idref = node.getString(LyInfoTokens.ATT_IDREF);

			info = getInfo(idref);
			if (!(info instanceof LyAction))
				throw new SAXException("BF AppEngine : invalid action ref : " + idref);

			LyAction action = (LyAction)info;

			idref = node.getString(LyInfoTokens.ATT_CLASS);
			info = null;

			if (idref != null)
			{
				info = getInfo(idref);
				if (!(info instanceof LyClassInfo))
					throw new SAXException("BF AppEngine : invalid class ref : " + idref);
			}

			if ((id == null)
                    && (info == null))
				result = action;
			else
			{
				result = new LyAction(id, action, (LyClassInfo)info, null);
				completeResult(node, result);
			}
		}
		else if (tag == LyInfoTokens.TAG_ACTION)
		{
			String idref = node.getString(LyInfoTokens.ATT_DEFAULT_ACTION);

			if (idref != null)
			{
				LyInfo action = getInfo(idref);

				if (!(action instanceof LyAction))
					throw new SAXException("BF AppEngine : invalid action ref : " + idref);

				((LyAction)result).setDefaultAction((LyAction)action);
			}
		}
		// Reference to an action
		else if (tag == LyInfoTokens.TAG_NOTIFIER_REF)
		{
			String idref;

			idref = node.getString(LyInfoTokens.ATT_IDREF);
			result = getInfo(idref);

			if (!(result instanceof LyNotifier))
				throw new SAXException("BF AppEngine : invalid notifier ref : " + idref);
		}
		else if (tag == LyInfoTokens.TAG_PREFERENCE)
		{
            String   idref = node.getString(LyInfoTokens.ATT_FIXED_COLUMN);
            LyInfo   info = getInfo(idref);

            if ((info != null) && !(info instanceof LyFieldInfo))
                throw new SAXException("BF AppEngine : invalid field info on preference : " + idref);
            ((LyPreference)result).setFixedColumn((LyFieldInfo)info);
		}
		else if (tag == LyInfoTokens.TAG_TREEMAP_CONFIGURATION)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);
			LyInfo info = getInfo(idref);

			if (!(info instanceof LyObjectInfo))
				throw new SAXException("BF AppEngine : invalid idref on treemap configuration : "
					+ idref);
			((LyTreemapConfiguration)result).setObjectInfo((LyObjectInfo)info);

			idref = node.getString(LyInfoTokens.ATT_COLOR_FIELD);
			info = getInfo(idref);

			if ((info != null) && !(info instanceof LyFieldInfo))
				throw new SAXException(
					"BF AppEngine : invalid color field info on treemap configuration : " + idref);
			((LyTreemapConfiguration)result).setColorFieldInfo((LyFieldInfo)info);

			idref = node.getString(LyInfoTokens.ATT_SIZE_FIELD);
			info = getInfo(idref);

			if ((info != null) && !(info instanceof LyFieldInfo))
				throw new SAXException(
					"BF AppEngine : invalid size field info on treemap configuration : " + idref);
			((LyTreemapConfiguration)result).setSizeFieldInfo((LyFieldInfo)info);

			idref = node.getString(LyInfoTokens.ATT_AGGREGATION_FIELD);
			info = getInfo(idref);

			if ((info != null) && !(info instanceof LyFieldInfo))
				throw new SAXException(
					"BF AppEngine : invalid aggregation field info on treemap configuration : "
					+ idref);
			((LyTreemapConfiguration)result).setAggregationFieldInfo((LyFieldInfo)info);

			idref = node.getString(LyInfoTokens.ATT_HIERARCHY_FIELDS);

			if (idref != null)
			{
				LyInfoList<LyFieldInfo> fields = new LyInfoList<LyFieldInfo>();
				int					    i, n = 0;
				String[]			    idrefs = LyString.stringToList(idref, ' ');

				n = ((idrefs == null) ? 0 : idrefs.length);

				for (i = 0; i < n; i++)
				{
					info = getInfo(idrefs[i]);

					if ((info != null) && !(info instanceof LyFieldInfo))
						throw new SAXException("BF AppEngine : invalid hierarchy field info on treemap configuration : " + idref);

					fields.addInfo((LyFieldInfo)info);
				}

				((LyTreemapConfiguration)result).setHierarchyFieldInfos(fields);
			}
		}
		// Reference to a field group
		else if (tag == LyInfoTokens.TAG_GROUP_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyFieldGroup))
				throw new SAXException("BF AppEngine : invalid field group ref : " + idref);
		}
		// Reference to a field tab
		else if (tag == LyInfoTokens.TAG_TAB_REF)
		{
			String idref = node.getString(LyInfoTokens.ATT_IDREF);

			result = getInfo(idref);
			if (!(result instanceof LyFieldTab))
				throw new SAXException("BF AppEngine : invalid field tab ref : " + idref);
		}
		else if (tag == LyInfoTokens.TAG_CONSTRAINT)
		{
            String  idref = node.getString(LyInfoTokens.ATT_ACTION);

            if (idref != null)
            {
		        LyInfo  info = getInfo(idref);

		        if (info != null && !(info instanceof LyAction))
		            throw new SAXException("BF AppEngine : invalid action on constraint : " + idref);

		        ((LyFieldConstraint)result).setAction((LyAction)info);
            }
		}
        else if (tag == LyInfoTokens.TAG_F_MARK)
        {
            String  idref = node.getString(LyInfoTokens.ATT_ACTION);

            if (idref != null)
            {
                LyInfo  info = getInfo(idref);

                if (info != null && !(info instanceof LyAction))
                    throw new SAXException("BF AppEngine : invalid action on field mark : " + idref);
            }
        }

		return result;
	} // end method resolveObjectInfoRef

	/**
	 * Completes parent result with the result of the current node.
	 *
	 * @param     parentResult Result of the parent node.
	 * @param     node         Current node.
	 * @param     result       Result of current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	@Override
  void completeParentResult(Object parentResult, LyInfoNode node, Object result)
		throws SAXException
	{
		if (parentResult instanceof LyUnionBinding)
			// If parentResult is an union binding, we complete it with this result
			completeUnionBinding((LyUnionBinding)parentResult, node);
		else if ((parentResult instanceof LyClassBinding)
				|| (parentResult instanceof LyLdapClassBinding))
			// If parentResult is a class binding, we complete it with this result
			completeClassBinding((LyClassBinding)parentResult, node);
		else if (parentResult instanceof LyFieldsBinding)
			// If parentResult is a fields binding, we complete it with this result
			completeFieldsBinding((LyFieldsBinding)parentResult, node);
		else if (parentResult instanceof LyFieldInfoImport)
			// If parentResult is a field info import, we complete it with this result
			completeFieldInfoImport((LyFieldInfoImport)parentResult, node);
		else if (parentResult instanceof LyParameter)
			// If parentResult is a parameter, we complete it with this result
			((LyParameter)parentResult).setValue(result);
		else if ((parentResult instanceof LyClassRef)
				&& (result instanceof LyFilter))
			// If parentResult is a classInfoRef, we complete it with this result
			((LyClassRef)parentResult).setFilter((LyFilter)result);
		else if (parentResult instanceof LyFieldBean)
			// If parentResult is a field bean, we complete it with this result
			completeFieldBean((LyFieldBean)parentResult, node);
		else if (parentResult instanceof LyArrayDaemon)
			// If parentResult is an array daemon, we complete it with this result
			completeArrayDaemon((LyArrayDaemon)parentResult, node);
		else if (parentResult instanceof LyJoinDaemon)
			// If parentResult is a join daemon, we complete it with this result
			completeJoinDaemon((LyJoinDaemon)parentResult, node);
		else if (parentResult instanceof LyModel)
			completeModel((LyModel)parentResult, node);
		else
			super.completeParentResult(parentResult, node, result);
	} // end method completeParentResult

	/**
	 * Completes an info with the result of the current node.
	 *
	 * @param     info Info to be completed.
	 * @param     node Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	@Override
  void completeInfo(LyInfo info, LyInfoNode node)
		throws SAXException
	{
		// For other tags, call the appropriate method to complete the info
		if (info instanceof LyLocation)
			completeLocation((LyLocation)info, node);
		else if (info instanceof LyChoiceOptionGroup)
			completeChoiceOptionGroup((LyChoiceOptionGroup)info, node);
		else if (info instanceof LyUnits)
			completeUnits((LyUnits)info, node);
		else if (info instanceof LyRule)
			completeRule((LyRule)info, node);
		else if (info instanceof LyInfoLabel)
			completeInfoLabel((LyInfoLabel)info, node);
		else if (info instanceof LyUnit)
			completeUnit((LyUnit)info, node);
		else if (info instanceof LyFieldGroup)
			completeFieldGroup((LyFieldGroup)info, node);
		else if (info instanceof LyFieldTab)
			completeFieldTab((LyFieldTab)info, node);
		else
			super.completeInfo(info, node);
	}

	/**
	 * Completes an objectInfo with the result of the current node.
	 *
	 * @param     objectInfo ObjectInfo to be completed.
	 * @param     node       Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	@Override
  void completeObjectInfo(LyObjectInfo objectInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to objectInfo attributes
		// and complete the given objectInfo.
		if (tag == LyInfoTokens.TAG_NAME)
			objectInfo.setName((String)result);
		else if (tag == LyInfoTokens.TAG_ALIAS)
			objectInfo.setAlias((String)result);
		else if ((tag == LyInfoTokens.TAG_GROUP)
				&& (objectInfo instanceof LyFieldInfo))
			completeFieldInfo((LyFieldInfo)objectInfo, node); // Field group management
		else if ((tag == LyInfoTokens.TAG_GROUP)
				&& (result instanceof String))
			objectInfo.setGroup((String)result);
		else if (tag == LyInfoTokens.TAG_PREFIX)
			objectInfo.setFilePrefix((String)result);
		else if (tag == LyInfoTokens.TAG_TIP)
			objectInfo.setTip((String)result);
		else if (tag == LyInfoTokens.TAG_DATA)
			objectInfo.putData(node.getString(LyInfoTokens.ATT_NAME), result);

		// For other tags, call the appropriate method to complete the objectInfo
		else if (objectInfo instanceof LyAction)
			completeAction((LyAction)objectInfo, node);
		else if (objectInfo instanceof LyChartInfo)
			completeChartInfo((LyChartInfo)objectInfo, node);
		else if (objectInfo instanceof LyPreference)
			completePreference((LyPreference)objectInfo, node);
		else if (objectInfo instanceof LyTreemapConfiguration)
			completeTreemapConfiguration((LyTreemapConfiguration)objectInfo, node);
		else if (objectInfo instanceof LyProject)
			completeProject((LyProject)objectInfo, node);
		else if (objectInfo instanceof LyFieldInfo)
			completeFieldInfo((LyFieldInfo)objectInfo, node);
		else
			super.completeObjectInfo(objectInfo, node);
	} // end method completeObjectInfo

	/**
	 * Completes a project with the result of the current node.
	 *
	 * @param project Project to be completed.
	 * @param node    Current node.
	 */
	@Override
  void completeProject(LyProject project, LyInfoNode node)
	{
		String tag = node.getTag();

		if (tag == LyInfoTokens.TAG_INFO)
			return;

		// Add the current result to project by using the right method
		Object result = node.getResult();

		if (result instanceof LyLocation)
			project.addLocation((LyLocation)result);
        else if (result instanceof LyNotifier)
            project.addNotifier((LyNotifier)result);
		else if (result instanceof LyProject)
			project.addProject((LyProject)result);
		else if (result instanceof LyModel)
			project.addModel((LyModel)result);
		else if (result instanceof LyInfo)
			project.addInfo((LyInfo)result);
	}

	/**
	 * Completes a model with the result of the current node.
	 *
	 * @param model Model to be completed.
	 * @param node  Current node.
	 */
	void completeModel(LyModel model, LyInfoNode node)
	{
	}

	/**
	 * Completes an action with the result of the current node.
	 *
	 * @param     action Action to be completed.
	 * @param     node   Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeAction(LyAction action, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();
		String tag = node.getTag();

		// Check for tags corresponding to action attributes
		// and complete the given action.
		if (tag == LyInfoTokens.TAG_A_MARKS)
			mergeMarks(action, (LyBitSet)result, node);
		else if (tag == LyInfoTokens.TAG_MENU)
			action.setMenu((String)result);
		else if (tag == LyInfoTokens.TAG_MIN)
			action.setMin(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_MAX)
			action.setMax(LyInfoTokens.parseInt((String)result));
		else if (result instanceof LyFilter)
			action.setFilter((LyFilter)result);
		else if (tag == LyInfoTokens.TAG_PARAMETER_DECL)
			action.addParameterDecl((LyParameterDecl)result);
		else if (tag == LyInfoTokens.TAG_PARAMETER)
			action.addParameter((LyParameter)result);
		else if (tag == LyInfoTokens.TAG_RESOURCE)
			action.addResource((LyActionResource)result);
		else if (tag == LyInfoTokens.TAG_SHORTCUT)
			action.setShortcut((String)result);
		else if (tag == LyInfoTokens.TAG_GEOMETRY)
			action.setGeometry((LyGeometry)result);
		else if (tag == LyInfoTokens.TAG_ACTIONS)
		{
			String			 operation = node.getString(LyInfoTokens.ATT_OPERATION);
			int				 pos = node.getInt(LyInfoTokens.ATT_POSITION);

			@SuppressWarnings("unchecked")
			Vector<LyAction> actions = (Vector<LyAction>)result;

			int				 i, n = actions.size();

			if (operation != null)
				operation = operation.intern();

			if (operation == LyInfoTokens.OPER_LIST_RESET)
			{
				action.resetActions();
				pos = 0;
			}

			if ((operation == LyInfoTokens.OPER_LIST_ADD)
					|| (operation == LyInfoTokens.OPER_LIST_RESET))
			{
				for (i = 0; i < n; i++)
					action.addAction(actions.elementAt(i), ((pos < 0) ? -1 : (pos + i)));
			}
			else if (operation == LyInfoTokens.OPER_LIST_REMOVE)
			{
				for (i = 0; i < n; i++)
					action.removeAction(actions.elementAt(i));
				node.setResult(null);
			}
		}
		else if (action instanceof LyCompoundAction)
			completeCompoundAction(tag, (LyCompoundAction)action, result);
		else if (action instanceof LyTabAction)
			completeTabAction(tag, (LyTabAction)action, result);
	} // end method completeAction

	/**
	 * Completes a compound action with the result of the current node.
	 *
	 * @param     compoundAction Compound action to be completed.
	 * @param     node           Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeCompoundAction(String tag, LyCompoundAction compoundAction, Object result)
		throws SAXException
	{
		// Check for tags corresponding to compound action attributes
		if (result instanceof LyAction)
			compoundAction.addSubAction((LyAction)result);
	}

	/**
	 * Completes a tab action with the result of the current node.
	 *
	 * @param     tabAction Tab action to be completed.
	 * @param     node      Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeTabAction(String tag, LyTabAction tabAction, Object result)
		throws SAXException
	{
		// Check for tags corresponding to tab action attributes
		if (result instanceof LyAction)
			tabAction.addTabAction((LyAction)result);
	}

	/**
	 * Completes an fieldInfo with the result of the current node.
	 *
	 * @param     fieldInfo FieldInfo to be completed.
	 * @param     node      Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	@Override
  void completeFieldInfo(LyFieldInfo fieldInfo, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();
		String tag = node.getTag();

		// Check for tags corresponding to objectInfo attributes
		// and complete the given fieldInfo.
		if (tag == LyInfoTokens.TAG_F_MARKS)
			mergeMarks(fieldInfo, (LyBitSet)result, node);
		else if ((tag == LyInfoTokens.TAG_TAB)
				|| (tag == LyInfoTokens.TAG_TAB_REF))
		{
			if (result instanceof LyFieldTab)
			{
				fieldInfo.setFieldTab((LyFieldTab)result);
				_applicationTabs.addInfo((LyFieldTab)result);
			}
			else
			{
				// Try to get the same tab
				LyFieldTab tab = _applicationTabs.getInfo((String) result, true);

				// Build new tab and add it to the application
				if (tab == null)
				{
					tab = new LyFieldTab(null, (String) result);
					tab.setApplication(_application);
                    tab.setParent(fieldInfo);
					_applicationTabs.addInfo(tab);
				}

				fieldInfo.setFieldTab(tab);
			}
		}
		else if ((tag == LyInfoTokens.TAG_GROUP)
				|| (tag == LyInfoTokens.TAG_GROUP_REF))
		{
			if (result instanceof LyFieldGroup)
			{
				fieldInfo.setFieldGroup((LyFieldGroup)result);
				_applicationGroups.addInfo((LyFieldGroup)result);
			}
			else
			{
				// Try to get the same group
				LyFieldGroup group = _applicationGroups.getInfo((String) result, true);

				// Build new group and add it to the application
				if (group == null)
				{
					group = new LyFieldGroup(null, (String) result, null);
					group.setApplication(_application);
                    group.setParent(fieldInfo);
					_applicationGroups.addInfo(group);
				}

				fieldInfo.setFieldGroup(group);
			}
		}
		else if ((tag == LyInfoTokens.TAG_CONTROL)
				|| (tag == LyInfoTokens.TAG_CONTROL_REF))
			fieldInfo.addControl((LyControl)result);
		else if ((tag == LyInfoTokens.TAG_UNIT)
				|| (tag == LyInfoTokens.TAG_UNIT_REF))
			fieldInfo.setFieldUnit((LyUnit)result);
		else if ((tag == LyInfoTokens.TAG_UNITS)
				|| (tag == LyInfoTokens.TAG_UNITS_REF))
			fieldInfo.setUnits((LyUnits)result);
		else if (tag == LyInfoTokens.TAG_LABEL)
			fieldInfo.addLabel((LyInfoLabel)result);
		else if (tag == LyInfoTokens.TAG_BEAN)
			fieldInfo.addBean((LyFieldBean)result);
	    else if (tag == LyInfoTokens.TAG_FORMULA)
	        fieldInfo.setFormula((String)result);
		else if (result instanceof LyFieldBinding)
			fieldInfo.setBinding((LyFieldBinding)result);
		else if (result instanceof LyFieldInfoImport)
			fieldInfo.setImport((LyFieldInfoImport)result);
		else if (result instanceof LyFieldConstraint)
			fieldInfo.addConstraint((LyFieldConstraint)result);
		else if (result instanceof LyPath)
			fieldInfo.setPath((LyPath)result);

		// For other tags, call the appropriate method to complete the fieldInfo
		else if (fieldInfo instanceof LyArrayFieldInfo)
			completeArrayFieldInfo((LyArrayFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyChoiceFieldInfo)
			completeChoiceFieldInfo((LyChoiceFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyDateFieldInfo)
			completeDateFieldInfo((LyDateFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyNumberFieldInfo)
			completeNumberFieldInfo((LyNumberFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyStructFieldInfo)
			completeStructFieldInfo((LyStructFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyRelationFieldInfo)
			completeRelationFieldInfo((LyRelationFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyTextFieldInfo)
			completeTextFieldInfo((LyTextFieldInfo)fieldInfo, node);
		else if (fieldInfo instanceof LyFileFieldInfo)
			completeFileFieldInfo((LyFileFieldInfo)fieldInfo, node);
	} // end method completeFieldInfo

	/**
	 * Completes an array field info with the result of the current node.
	 *
	 * @param     arrayFieldInfo ArrayFieldInfo to be completed.
	 * @param     node           Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeArrayFieldInfo(LyArrayFieldInfo arrayFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		if (result instanceof LyFieldInfo)
			arrayFieldInfo.setFieldInfo((LyFieldInfo)result);
		else if (tag == LyInfoTokens.TAG_ARRAY_DAEMON)
			arrayFieldInfo.setDaemon((LyArrayDaemon)result);
		else if (tag == LyInfoTokens.TAG_NROWS)
			arrayFieldInfo.setNRows(LyInfoTokens.parseInt((String)result));
	}

	/**
	 * Completes a date field info with the result of the current node.
	 *
	 * @param     dateFieldInfo DateFieldInfo to be completed.
	 * @param     node          Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeDateFieldInfo(LyDateFieldInfo dateFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to dateFieldInfo attributes
		// and complete the given dateFieldInfo.
		if (tag == LyInfoTokens.TAG_FORMAT)
			dateFieldInfo.setDisplayFormat(((LyFormat)result).getValue());
		else if (tag == LyInfoTokens.TAG_DEFAULT)
		{
			Object value;

			if (((String)result).equals("now")
					|| ((String)result).equals("_now"))
				value = LyDateFieldInfo.VALUE_NOW;
			else
				value = dateFieldInfo.parse((String)result);

			dateFieldInfo.setDefaultValue(value);
		}
	}

	/**
	 * Completes a number field info with the result of the current node.
	 *
	 * @param     numberFieldInfo NumberFieldInfo to be completed.
	 * @param     node            Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeNumberFieldInfo(LyNumberFieldInfo numberFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String				   tag = node.getTag();
		Object				   result = node.getResult();

		// Check for tags corresponding to numberFieldInfo attributes
		// and complete the given numberFieldInfo.
		LyNumberFieldInfo.Type type = numberFieldInfo.getType();

		if (tag == LyInfoTokens.TAG_FORMAT)
			numberFieldInfo.setFormat(((LyFormat)result).getValue());
		else if (tag == LyInfoTokens.TAG_FUNCTION)
			numberFieldInfo.setFunction(LyInfoTokens.getNumberFieldInfoFunction((String)result));
		else if (tag == LyInfoTokens.TAG_MIN)
		{
			Number number = LyInfoTokens.parseNumber(type, (String)result);

			numberFieldInfo.setMin(number);
		}
		else if (tag == LyInfoTokens.TAG_MAX)
		{
			Number number = LyInfoTokens.parseNumber(type, (String)result);

			numberFieldInfo.setMax(number);
		}
		else if (tag == LyInfoTokens.TAG_PITCH)
		{
			Number number = LyInfoTokens.parseNumber(type, (String)result);

			numberFieldInfo.setPitch(number);
		}
		else if (tag == LyInfoTokens.TAG_UNIT)
			numberFieldInfo.setFieldUnit((LyUnit)result);
		else if (tag == LyInfoTokens.TAG_DECIMALS)
			numberFieldInfo.setDecimals(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_N_DEFAULT)
		{
			Number number = LyInfoTokens.parseNumber(type, (String)result);

			numberFieldInfo.setDefault(number);
		}
		else if (tag == LyInfoTokens.TAG_PRECISION)
			numberFieldInfo.setPrecision(LyInfoTokens.parseNumber(type, (String)result));
	} // end method completeNumberFieldInfo

	/**
	 * Completes a text field info with the result of the current node.
	 *
	 * @param     textFieldInfo TextFieldInfo to be completed.
	 * @param     node          Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeTextFieldInfo(LyTextFieldInfo textFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to textFieldInfo attributes
		// and complete the given textFieldInfo.
		if (tag == LyInfoTokens.TAG_FORMAT)
			textFieldInfo.setFormat(((LyFormat)result).getValue());
		else if (tag == LyInfoTokens.TAG_ENCODING)
			textFieldInfo.setEncoding((String)result);
		else if ((tag == LyInfoTokens.TAG_CHARTYPE)
				|| (tag == LyInfoTokens.TAG_CHARTYPE_REF))
			textFieldInfo.addCharType((LyInfoCharType)result);
		else if ((tag == LyInfoTokens.TAG_REGEX)
				|| (tag == LyInfoTokens.TAG_REGEX_REF))
			textFieldInfo.setRegExp((LyRegExp)result);
		else if ((tag == LyInfoTokens.TAG_NCHARS)
				|| (tag == LyInfoTokens.TAG_MAX))
			textFieldInfo.setMaxChars(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_NROWS)
			textFieldInfo.setNRows(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_MIN)
			textFieldInfo.setMinChars(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_NCOLS)
			textFieldInfo.setNCols(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_DEFAULT)
			textFieldInfo.setDefault((String)result);
	} // end method completeTextFieldInfo

	/**
	 * Completes a file field info with the result of the current node.
	 *
	 * @param     fileFieldInfo FileFieldInfo to be completed.
	 * @param     node          Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeFileFieldInfo(LyFileFieldInfo fileFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to fileFieldInfo attributes
		// and complete the given fileFieldInfo.
		if (tag == LyInfoTokens.TAG_DEFAULT)
			fileFieldInfo.setDefault((String)result);
		else if (tag == LyInfoTokens.TAG_NROWS)
			fileFieldInfo.setNRows(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_NCOLS)
			fileFieldInfo.setNCols(LyInfoTokens.parseInt((String)result));
	}

	/**
	 * Completes a choice field info with the result of the current node.
	 *
	 * @param     choiceFieldInfo ChoiceFieldInfo to be completed.
	 * @param     node            Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeChoiceFieldInfo(LyChoiceFieldInfo choiceFieldInfo, LyInfoNode node)
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to choiceFieldInfo attributes
		// and complete the given choiceFieldInfo.
		if (tag == LyInfoTokens.TAG_NCOLS)
			choiceFieldInfo.setNCols(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_OPTIONS)
		{
			String		   operation = node.getString(LyInfoTokens.ATT_OPERATION);
			int			   pos = node.getInt(LyInfoTokens.ATT_POSITION);
			LyChoiceOption option;

			if (operation != null)
				operation = operation.intern();

			if (operation == LyInfoTokens.OPER_LIST_RESET)
			{
				choiceFieldInfo.resetChoiceOptions();
				pos = 0;
			}

			if ((operation == LyInfoTokens.OPER_LIST_ADD)
					|| (operation == LyInfoTokens.OPER_LIST_RESET))
			{
				LyChoiceOptionGroup		   optionGroup = (LyChoiceOptionGroup)result;
				LyInfoList<LyChoiceOption> options = optionGroup.getChoiceOptions(true);
				int						   i, n = options.getSize();

				for (i = 0; i < n; i++)
				{
					option = options.getInfo(i);
					choiceFieldInfo.addChoiceOption(option, ((pos < 0) ? -1 : (pos + i)));
				}

				choiceFieldInfo.addChoiceOptionGroup(optionGroup);
			}
			else if (operation == LyInfoTokens.OPER_LIST_REMOVE)
			{
				@SuppressWarnings("unchecked")
				Vector<LyChoiceOption> options = (Vector<LyChoiceOption>)result;
				int					   i, n = options.size();

				for (i = 0; i < n; i++)
				{
					option = options.elementAt(i);
					choiceFieldInfo.removeChoiceOption(option);
				}
				node.setResult(null);
			}
		}
		else if (tag == LyInfoTokens.TAG_NCOLS)
			choiceFieldInfo.setNCols(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_CONTEXT_FIELD)
		{
			LyFieldInfo fieldInfo = (LyFieldInfo)result;

			if (!(fieldInfo instanceof LyChoiceFieldInfo))
				throw new IllegalStateException("BF AppEngine : context field " + fieldInfo
					+ " is not a CHOICE field. Only choice field can be used as a context field"
					+ " for another choice. Check context of choiceFieldInfo '"
					+ choiceFieldInfo.getId() + "'.");

			choiceFieldInfo.setContextField((LyChoiceFieldInfo)fieldInfo);
		}
	} // end method completeChoiceFieldInfo

	/**
	 * Completes a choice option info with the result of the current node.
	 *
	 * @param     choiceOption ChoiceOption to be completed.
	 * @param     node         Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	@Override
  void completeChoiceOption(LyChoiceOption choiceOption, LyInfoNode node)
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to choiceOption attributes
		// and complete the given option.
		if ((tag == LyInfoTokens.TAG_RULE)
				|| (tag == LyInfoTokens.TAG_RULE_REF))
			choiceOption.addRule((LyRule)result);
		else if (tag == LyInfoTokens.TAG_LABEL)
			choiceOption.addLabel((LyInfoLabel)result);
	}

	/**
	 * Completes a choice options group info with the result of the current node.
	 *
	 * @param     choiceOptionGroup ChoiceOptionGroup to be completed.
	 * @param     node              Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeChoiceOptionGroup(LyChoiceOptionGroup choiceOptionGroup, LyInfoNode node)
	{
		// Check for tags corresponding to choiceOption attributes
		// and complete the given option.
		String tag = node.getTag();
		Object result = node.getResult();

		if (tag == LyInfoTokens.TAG_LABEL)
			choiceOptionGroup.addLabel((LyInfoLabel)result);
		else if (result instanceof LyChoiceItem)
			choiceOptionGroup.addChoiceItem((LyChoiceItem)result);
	}

	/**
	 * Completes a relation field info with the result of the current node.
	 *
	 * @param     relationFieldInfo RelationFieldInfo to be completed.
	 * @param     node              Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeRelationFieldInfo(LyRelationFieldInfo relationFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to relationFieldInfo attributes
		// and complete the given relationFieldInfo.
		if (tag == LyInfoTokens.TAG_CLASS_FIELD)
			relationFieldInfo.setClassField((LyFieldInfo)result);
		else if (tag == LyInfoTokens.TAG_CONTEXT_FIELD)
		{
			if (!(result instanceof LyRelationFieldInfo))
				throw new SAXException("BF AppEngine : Info " + result.toString()
					+ " is not a valid context field for the relation fieldInfo '"
					+ relationFieldInfo.getId() + "'"
					+ " . It should be a relation Field.");

			relationFieldInfo.setContextField((LyRelationFieldInfo)result);
		}
		else if (tag == LyInfoTokens.TAG_DEFAULT)
			relationFieldInfo.setDefault((String)result);
		else if (tag == LyInfoTokens.TAG_CLASSES)
		{
			String			    operation = node.getString(LyInfoTokens.ATT_OPERATION);
			int				    pos = node.getInt(LyInfoTokens.ATT_POSITION);

			Vector<?>		    classes = (Vector<?>)result;

			int				    i, j, n = classes.size();
			Object			    info;
			LyClassInfo		    classInfo, extendedClassInfo;
			LyFilter		    filter;
			LyRelationFieldInfo extendedRelation = relationFieldInfo.getExtendedRelation();

			if (extendedRelation != null)
			{
				relationFieldInfo.resetClasses();
				for (i = 0; i < n; i++)
				{
					info = classes.elementAt(i);
					filter = null;

					if (info instanceof LyClassInfo)
						classInfo = (LyClassInfo)info;
					else
					{
						classInfo = ((LyClassRef)info).getClassInfo();
						filter = ((LyClassRef)info).getFilter();
					}

					for (j = 0; j < extendedRelation.getClassInfoCount(); j++)
					{
						extendedClassInfo = extendedRelation.getClassInfo(j);
						if ((classInfo == extendedClassInfo)
								|| classInfo.inheritsFrom(extendedClassInfo))
							break;
					}
					if (j == extendedRelation.getClassInfoCount())
						throw new SAXException("BF AppEngine : classInfo " + classInfo.getId()
							+ " does not extend any class from the extended relation "
							+ extendedRelation.getId());

					relationFieldInfo.addClassInfo(classInfo, filter, i);
				}
			}
			else
			{
				if (operation != null)
					operation = operation.intern();

				if (operation == LyInfoTokens.OPER_LIST_RESET)
				{
					relationFieldInfo.resetClasses();
					pos = 0;
				}

				if ((operation == LyInfoTokens.OPER_LIST_ADD)
						|| (operation == LyInfoTokens.OPER_LIST_RESET))
				{
					for (i = 0; i < n; i++)
					{
						info = classes.elementAt(i);
						if (info instanceof LyClassInfo)
							relationFieldInfo.addClassInfo((LyClassInfo)info, null,
								((pos < 0) ? -1 : (pos + i)));
						else
							relationFieldInfo.addClassInfo(((LyClassRef)info).getClassInfo(),
								((LyClassRef)info).getFilter(), ((pos < 0) ? -1 : (pos + i)));
					}
				}
				else if (operation == LyInfoTokens.OPER_LIST_REMOVE)
				{
					for (i = 0; i < n; i++)
					{
						info = classes.elementAt(i);
						if (info instanceof LyClassInfo)
							relationFieldInfo.removeClassInfo((LyClassInfo)info, null);
						else
							relationFieldInfo.removeClassInfo(((LyClassRef)info).getClassInfo(),
								((LyClassRef)info).getFilter());
					}
					node.setResult(null);
				}
			}
		}
		else if (tag == LyInfoTokens.TAG_ROUTES)
		{
			String		    operation = node.getString(LyInfoTokens.ATT_OPERATION);
			int			    pos = node.getInt(LyInfoTokens.ATT_POSITION);

			@SuppressWarnings("unchecked")
			Vector<LyRoute> routes = (Vector<LyRoute>)result;

			int			    i, n = routes.size();

			if (operation != null)
				operation = operation.intern();

			if (operation == LyInfoTokens.OPER_LIST_RESET)
			{
				relationFieldInfo.resetRoutes();
				pos = 0;
			}

			if ((operation == LyInfoTokens.OPER_LIST_ADD)
					|| (operation == LyInfoTokens.OPER_LIST_RESET))
			{
				for (i = 0; i < n; i++)
					relationFieldInfo.addRoute(routes.elementAt(i), ((pos < 0) ? -1 : (pos + i)));
			}
			else if (operation == LyInfoTokens.OPER_LIST_REMOVE)
			{
				for (i = 0; i < n; i++)
					relationFieldInfo.removeRoute(routes.elementAt(i));
				node.setResult(null);
			}
		}
		else if ((tag == LyInfoTokens.TAG_JOIN)
				|| (tag == LyInfoTokens.TAG_JOIN_DAEMON)
				|| (tag == LyInfoTokens.TAG_REVERSE))
			relationFieldInfo.setDaemon((LyInfoDaemon)result);
		else if (tag == LyInfoTokens.TAG_NROWS)
			relationFieldInfo.setNRows(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_NCOLS)
            relationFieldInfo.setNCols(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_MIN)
			relationFieldInfo.setMin(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_MAX)
			relationFieldInfo.setMax(LyInfoTokens.parseInt((String)result));
		else if (tag == LyInfoTokens.TAG_CLASS_FIELD)
			relationFieldInfo.setClassField((LyFieldInfo)result);
		else if (tag == LyInfoTokens.TAG_CONTEXT_FIELD)
			relationFieldInfo.setContextField((LyRelationFieldInfo)result);
		else if (result instanceof LyClassBinding)
		{
			LyClassInfo classInfo = relationFieldInfo.getClassInfo(0);

			completeClassInfo(classInfo, node);
		}
	} // end method completeRelationFieldInfo

	/**
	 * Completes a struct field info with the result of the current node.
	 *
	 * @param     relationFieldInfo structFieldInfo to be completed.
	 * @param     node              Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeStructFieldInfo(LyStructFieldInfo structFieldInfo, LyInfoNode node)
		throws SAXException
	{
		String	    tag = node.getTag();
		Object	    result = node.getResult();
		LyClassInfo classInfo = structFieldInfo.getClassInfo(0);

		if (result instanceof LyClassBinding)
			classInfo.setBinding((LyClassBinding)result);
		else if (tag == LyInfoTokens.TAG_FIELDS)
			completeClassInfo(classInfo, node);
		else if (tag == LyInfoTokens.TAG_CACHE)
			classInfo.setCachePolicy((LyInfoCache)result);
		else if ((tag == LyInfoTokens.TAG_RULE)
				|| (tag == LyInfoTokens.TAG_RULE_REF))
			classInfo.addRule((LyRule)result);
	}

	/**
	 * Completes a rule with the result of the current node.
	 *
	 * @param     rule Rule to be completed.
	 * @param     node Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeRule(LyRule rule, LyInfoNode node)
		throws SAXException
	{
	    String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags or results corresponding to rule attributes
		// and complete the given label.
		if (result instanceof LyFilter)
			rule.setFilter((LyFilter)result);
		else if (result instanceof LyInfoLabel)
			rule.setLabel((LyInfoLabel)result);
		else if (tag == LyInfoTokens.TAG_FORMULA)
            rule.setFormula((String)result);
	}

	/**
	 * Completes a unit with the result of the current node.
	 *
	 * @param     unit Unit to be completed.
	 * @param     node Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeUnit(LyUnit unit, LyInfoNode node)
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags or results corresponding to unit attributes
		// and complete the given unit.
		if (tag == LyInfoTokens.TAG_FORMAT)
			unit.setFormat((LyFormat)result);
	}

	/**
	 * Completes a units list with the result of the current node.
	 *
	 * @param units Units list to be completed.
	 * @param node  Current node.
	 */
	void completeUnits(LyUnits units, LyInfoNode node)
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to units list attributes
		// and complete the given units list.
		if (result instanceof LyUnit)
			units.addUnit((LyUnit)result);
	}

	/**
	 * Completes a class info with the result of the current node.
	 *
	 * @param classInfo ClassInfo to be completed.
	 * @param node      Current node.
	 */
	@Override
  void completeClassInfo(LyClassInfo classInfo, LyInfoNode node)
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags corresponding to classInfo attributes
		// and complete the given classInfo.
		if (tag == LyInfoTokens.TAG_C_MARKS)
			mergeMarks(classInfo, (LyBitSet)result, node);
		else if (result instanceof LyClassBinding)
			classInfo.setBinding((LyClassBinding)result);
		else if (tag == LyInfoTokens.TAG_CACHE)
			classInfo.setCachePolicy((LyInfoCache)result);
		else if (tag == LyInfoTokens.TAG_FIELDS)
		{
			String			    operation = node.getString(LyInfoTokens.ATT_OPERATION);
			int				    pos = node.getInt(LyInfoTokens.ATT_POSITION);

			@SuppressWarnings("unchecked")
			Vector<LyFieldInfo> fields = (Vector<LyFieldInfo>)result;

			int				    i, n = fields.size();

			if (operation != null)
				operation = operation.intern();

			if (operation == LyInfoTokens.OPER_LIST_RESET)
			{
				classInfo.resetFieldInfos();
				pos = 0;
			}

			if ((operation == LyInfoTokens.OPER_LIST_ADD)
					|| (operation == LyInfoTokens.OPER_LIST_RESET))
			{
				for (i = 0; i < n; i++)
					classInfo.addFieldInfo(fields.elementAt(i), ((pos < 0) ? -1 : (pos + i)));
			}
			else if (operation == LyInfoTokens.OPER_LIST_REMOVE)
			{
				for (i = 0; i < n; i++)
					classInfo.removeFieldInfo(fields.elementAt(i));
				node.setResult(null);
			}
		}
		else if (tag == LyInfoTokens.TAG_ACTIONS)
		{
			String			 operation = node.getString(LyInfoTokens.ATT_OPERATION);
			int				 pos = node.getInt(LyInfoTokens.ATT_POSITION);

			@SuppressWarnings("unchecked")
			Vector<LyAction> actions = (Vector<LyAction>)result;

			int				 i, n = actions.size();

			if (operation != null)
				operation = operation.intern();

			if (operation == LyInfoTokens.OPER_LIST_RESET)
			{
				classInfo.resetActions();
				pos = 0;
			}

			if ((operation == LyInfoTokens.OPER_LIST_ADD)
					|| (operation == LyInfoTokens.OPER_LIST_RESET))
			{
				for (i = 0; i < n; i++)
					classInfo.addAction(actions.elementAt(i), ((pos < 0) ? -1 : (pos + i)));
			}
			else if (operation == LyInfoTokens.OPER_LIST_REMOVE)
			{
				for (i = 0; i < n; i++)
					classInfo.removeAction(actions.elementAt(i));
				node.setResult(null);
			}
		}
		else if (tag == LyInfoTokens.TAG_ROUTES)
		{
			String		    operation = node.getString(LyInfoTokens.ATT_OPERATION);
			int			    pos = node.getInt(LyInfoTokens.ATT_POSITION);

			@SuppressWarnings("unchecked")
			Vector<LyRoute> routes = (Vector<LyRoute>)result;

			int			    i, n = routes.size();

			if (operation != null)
				operation = operation.intern();

			if (operation == LyInfoTokens.OPER_LIST_RESET)
			{
				classInfo.resetRoutes();
				pos = 0;
			}

			if ((operation == LyInfoTokens.OPER_LIST_ADD)
					|| (operation == LyInfoTokens.OPER_LIST_RESET))
			{
				for (i = 0; i < n; i++)
					classInfo.addRoute(routes.elementAt(i), ((pos < 0) ? -1 : (pos + i)));
			}
			else if (operation == LyInfoTokens.OPER_LIST_REMOVE)
			{
				for (i = 0; i < n; i++)
					classInfo.removeRoute(routes.elementAt(i));
				node.setResult(null);
			}
		}
		else if ((tag == LyInfoTokens.TAG_SORT)
				|| (tag == LyInfoTokens.TAG_SORT_REF))
			classInfo.setSort((LySort)result);
		else if ((tag == LyInfoTokens.TAG_CONTROL)
				|| (tag == LyInfoTokens.TAG_CONTROL_REF))
			classInfo.addControl((LyControl)result);
		else if ((tag == LyInfoTokens.TAG_RULE)
				|| (tag == LyInfoTokens.TAG_RULE_REF))
			classInfo.addRule((LyRule)result);
		else if (tag == LyInfoTokens.TAG_LABEL)
			classInfo.addLabel((LyInfoLabel)result);
	} // end method completeClassInfo

	/**
	 * Completes a label with the result of the current node.
	 *
	 * @param     label Label to be completed.
	 * @param     node  Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeInfoLabel(LyInfoLabel label, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to label attributes
		// and complete the given label.
		if (result instanceof LyPath)
			label.addPath((LyPath)result);
		else if (result instanceof LyUnit)
			label.addUnit((LyUnit)result);
	}

	/**
	 * Completes a generic location with the result of the current node.
	 *
	 * @param     location Location to be completed.
	 * @param     node     Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeLocation(LyLocation location, LyInfoNode node)
	{
		Object result = node.getResult();
		String tag = node.getTag();

		// Check for tags or results corresponding to route attributes
		// and complete the given route.
		if ((location instanceof LyGenLocation) && (tag == LyInfoTokens.TAG_DATA))
			((LyGenLocation)location).putData(node.getString(LyInfoTokens.ATT_NAME), result);
		else if (result instanceof LyNotifier)
			location.setNotifier((LyNotifier)result);
	}

	/**
	 * Completes an union binding with the result of the current node.
	 *
	 * @param     unionBinding Union binding to be completed.
	 * @param     node         Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeUnionBinding(LyUnionBinding unionBinding, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to union binding attributes
		// and complete the given class binding.
		if (result instanceof LyClassBinding)
			unionBinding.addClassBinding((LyClassBinding)result);
		else if (result instanceof LyUnion)
			unionBinding.addUnion((LyUnion)result);
	}

	/**
	 * Completes a class binding with the result of the current node.
	 *
	 * @param     classBinding Class binding to be completed.
	 * @param     node         Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeClassBinding(LyClassBinding classBinding, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to class binding attributes
		// and complete the given class binding.
		if (result instanceof LyFilter)
			classBinding.setFilter((LyFilter)result);
	}

	/**
	 * Completes a fields binding with the result of the current node.
	 *
	 * @param     fieldsBinding Fields binding to be completed.
	 * @param     node          Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeFieldsBinding(LyFieldsBinding fieldsBinding, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to fields binding attributes
		// and complete the given fields binding.
		if (result instanceof LyFieldBinding)
			fieldsBinding.addFieldBinding((LyFieldBinding)result);
	}

	/**
	 * Completes a chartInfo with the result of the current node.
	 *
	 * @param     chartInfo ChartInfo to be completed.
	 * @param     node      Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeChartInfo(LyChartInfo chartInfo, LyInfoNode node)
		throws SAXException
	{
		String tag = node.getTag();
		Object result = node.getResult();

		// Check for tags or results corresponding to chartInfo attributes
		// and complete the given chartInfo.
		if (tag == LyInfoTokens.TAG_YFIELD)
			chartInfo.addYField((String)result);
		else if (tag == LyInfoTokens.TAG_PREFERENCE)
			chartInfo.setPreference((LyPreference)result);
	}

	/**
	 * Completes a preference with the result of the current node.
	 *
	 * @param     preference preference to be completed.
	 * @param     node       Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completePreference(LyPreference preference, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to preference attributes
		// and complete the given preference.
		if (result instanceof LyFilter)
			preference.setFilter((LyFilter)result);
		else if (result instanceof LySort)
			preference.setSort((LySort)result);
		else if (result instanceof LyFieldInfo)
			preference.addFieldInfo((LyFieldInfo)result);
	}

	/**
	 * Completes a treemap configuration with the result of the current node.
	 *
	 * @param     configuration configuration to be completed.
	 * @param     node          Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeTreemapConfiguration(LyTreemapConfiguration configuration, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to preference attributes
		// and complete the given preference.
		if (result instanceof LyFilter)
			configuration.setFilter((LyFilter)result);
	}

	/**
	 * Completes a field info import with the result of the current node.
	 *
	 * @param     fieldInfoImport Field info import to be completed.
	 * @param     node            Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeFieldInfoImport(LyFieldInfoImport fieldInfoImport, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to field info import attributes
		// and complete the given field info import
		if (result instanceof LyFilter)
			fieldInfoImport.setFilter((LyFilter)result);
	}

	/**
	 * Completes a field bean with the result of the current node.
	 *
	 * @param     fieldBean Field bean to be completed.
	 * @param     node      Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeFieldBean(LyFieldBean fieldBean, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();
		String tag = node.getTag();

		// Check for tags or results corresponding to class binding attributes
		// and complete the given class binding.
		if (tag == LyInfoTokens.TAG_DATA)
			fieldBean.putData(node.getString(LyInfoTokens.ATT_NAME),
				((result == null) ? null : result.toString()));
	}

	/**
	 * Completes a field group with the result of the current node.
	 *
	 * @param     fieldGroup Field group to be completed.
	 * @param     node       Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeFieldGroup(LyFieldGroup fieldGroup, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();
		String tag = node.getTag();

		// Check for tags or results corresponding to field group attributes
		// and complete the given field group.
		if (tag == LyInfoTokens.TAG_PREFIX)
			fieldGroup.setPrefix((String)result);
		else if (result instanceof LyFieldConstraint)
			fieldGroup.setConstraint((LyFieldConstraint)result);
	}

	/**
	 * Completes a field tab with the result of the current node.
	 *
	 * @param     fieldTab Field tab to be completed.
	 * @param     node     Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeFieldTab(LyFieldTab fieldTab, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();
		String tag = node.getTag();

		// Check for tags or results corresponding to field tab attributes
		// and complete the given field tab.
		if (tag == LyInfoTokens.TAG_PREFIX)
			fieldTab.setPrefix((String)result);
	}

	/**
	 * Completes an array daemon with the result of the current node.
	 *
	 * @param     arrayDaemon Array daemon to be completed.
	 * @param     node        Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeArrayDaemon(LyArrayDaemon arrayDaemon, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to field info import attributes
		// and complete the given field info import
		if (result instanceof LyClassBinding)
			arrayDaemon.setClassBinding((LyClassBinding)result);
		else if (result instanceof LyFieldBinding)
			arrayDaemon.addFieldBinding((LyFieldBinding)result);
	}

	/**
	 * Completes a join daemon with the result of the current node.
	 *
	 * @param     joinDaemon Join daemon to be completed.
	 * @param     node       Current node.
	 * @exception SAXExpection If an error occurs during this completion.
	 */
	void completeJoinDaemon(LyJoinDaemon joinDaemon, LyInfoNode node)
		throws SAXException
	{
		Object result = node.getResult();

		// Check for tags or results corresponding to field info import attributes
		// and complete the given field info import
		if (result instanceof LyClassBinding)
			joinDaemon.setClassBinding((LyClassBinding)result);
		else if (result instanceof LyFieldBinding)
			joinDaemon.addFieldBinding((LyFieldBinding)result);
	}

	/**
	 * Extra checks for result. Throws an exception if check fails.
	 *
	 * @param result Checked result.
	 */
	@Override
  protected void checkResult(Object result)
	{
		// Check bindings : should not have a different case value
		if ((result instanceof LyClassBinding)
				&& !(result instanceof LyClassInfoBinding))
		{
			String value = ((LyClassBinding)result).getBinding();

			if (value != null)
			{
				String lcValue = value.toLowerCase();

				if (_bindingsValidationTable == null)
					_bindingsValidationTable = new Hashtable<String, LyClassBinding>();

				// Get lower case related binding
				LyClassBinding oldBinding = _bindingsValidationTable.get(lcValue);

				if (oldBinding == null)
					_bindingsValidationTable.put(lcValue, (LyClassBinding)result);
				else if (!oldBinding.getBinding().equals(value))
					throw new IllegalArgumentException("BF AppEngine : The binding '" + value
						+ "' was already defined as '" + oldBinding.getBinding()
						+ "' : bindings must not differ only by case sensitivity.");
			}
		}
	}

	/**
	 * Returns the model type of this factory.
	 */
	@Override
  protected LyTag.Model getModelType()
	{
		return LyTag.Model.FULL;
	}
} // end class LyInfoNodeFactory