/**
 * LyInfoParser.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;
import java.util.*;
import leon.app.*;
import leon.misc.*;
import leon.parser.*;


/**
 * This class represents our XML data info parser. It's the entry point to parse XML data info. Our
 * parser is configured to build LyInfoNode nodes instead of standard ElementNode nodes.<br>
 * A parser belongs to an application. It is used to build the tree of info objects as described in
 * the XML file. The root of this tree is an instance of LyProject. All built objects belong to the
 * application for which the parser was created.
 *
 * @see     LyApplication
 * @see     LyProject
 * @version $Id: LyInfoParser.java 29006 2010-12-23 17:44:46Z bmichel $
 * @author  Lyria
 */
public final class LyInfoParser extends LySAXParser
{
	public static final String   __VERSION =
		"$Id: LyInfoParser.java 29006 2010-12-23 17:44:46Z bmichel $";

	// Parser types
    public enum Type
    {
        DEFAULT(null),
    	FULL("leon"),
    	APP("leon_app"),
    	BIND("leon_bind"),
    	DATA("leon_data"),
    	GUI("leon_gui");
        
        String      _grammarName;
        
        Type(String grammarName)
        {
            _grammarName = grammarName;
        }
        
        String getGrammarName()
        {
            return _grammarName;
        }
    }

	/** Application for which the parser was built. */
	protected LyApplication		 _application;

	/** List of parsed infos. */
	protected LyInfoList<LyInfo> _results;
	
	/** List of parsed infos qualified as externs. */
    protected LyInfoList<LyInfo> _externResults;

	/** Parser type. */
	protected Type  			 _type;

	/**
	 * Constructs a validating XML data info parser.
	 *
	 * @param     application Application for which a parser is built.
	 * @param     xmlFile     Path of the file that contains the Xml info.
	 * @exception IllegalargumentException If application is null.
	 * @exception SAXException             If an error occurs during parsing.
	 */
	public LyInfoParser(LyApplication application, String xmlFile)
	{
		this(application, xmlFile, Type.DEFAULT, true);
	}

	/**
	 * Constructs a XML data info parser.
	 *
	 * @param     application Application for which a parser is built.
	 * @param     xmlFile     Path of the file that contains the Xml info.
	 * @param     validating  Boolean indicating if the parser if validating or not.
	 * @exception IllegalargumentException If application is null.
	 * @exception SAXException             If an error occurs during parsing.
	 */
	public LyInfoParser(LyApplication application, String xmlFile,
		boolean validating)
	{
		this(application, xmlFile, Type.DEFAULT, validating);
	}

	/**
	 * Constructs a XML data info parser.
	 *
	 * @param     application Application for which a parser is built.
	 * @param     xmlFile     Path of the file that contains the Xml info.
	 * @param     validating  Boolean indicating if the parser if validating or not.
	 * @exception IllegalargumentException If application is null.
	 * @exception SAXException             If an error occurs during parsing.
	 */
	public LyInfoParser(LyApplication application, String xmlFile,
		Type type, boolean validating)
	{
		super(validating);

		if (application == null)
			throw new IllegalArgumentException("BF AppEngine : application can't be null");

		_application = application;
		_type = (type == null ? Type.DEFAULT : type);
		_results = new LyInfoList<LyInfo>(false);
		_externResults = new LyInfoList<LyInfo>(false);

		// Build factory
		LyInfoFactory factory = buildFactory(_type.getGrammarName());

		// Load the document
		LyEnvironment environment = _application.getEnvironment();
		Date		  d1 = new Date();

		loadDocument(xmlFile, factory, _application.getClassLoader(), environment);

		Date d2 = new Date();

		environment.logInfo("BF AppEngine : parsing XML data description file "
			+ xmlFile + " took " + (d2.getTime() - d1.getTime()) + " ms");
	}

	/**
	 * Builds default factory.
	 */
	protected LyInfoFactory buildFactory(String type)
	{
		if (type == null)
			return null;

		LyInfoFactory factory = null;

		if (type.equalsIgnoreCase(Type.FULL.getGrammarName()))
			factory = new LyInfoNodeFactory(this, _application);
		else if (type.equalsIgnoreCase(Type.APP.getGrammarName()))
			factory = new LyInfoAppFactory(this, _application);
		else if (type.equalsIgnoreCase(Type.BIND.getGrammarName()))
			factory = new LyInfoBindFactory(this, _application);
		else if (type.equalsIgnoreCase(Type.DATA.getGrammarName()))
			factory = new LyInfoDataFactory(this, _application);
		else if (type.equalsIgnoreCase(Type.GUI.getGrammarName()))
			factory = new LyInfoGuiFactory(this, _application);
		else if (type != null)
			throw new IllegalArgumentException("BF AppEngine : invalid parser type " + type);

		return factory;
	}

	/**
	 * Builds default factory.
	 */
	protected LyXMLElementFactory buildDefaultFactory()
	{
		LyEnvironment environment = _application.getEnvironment();
		String		  xmlGrammar = environment.getEnv("_LY_XML_GRAMMAR");
		LyInfoFactory factory = buildFactory(xmlGrammar);

		return ((factory != null) ? factory : super.buildDefaultFactory());
	}

	/**
	 * Gets application.
	 *
	 * @return Application for which the parser was built.
	 */
	public LyApplication getApplication()
	{
		return _application;
	}

	/**
	 * Returns root project found in data description XML file.
	 *
	 * @return    Root project built by this parser.
	 * @exception IllegalStateException If root result is null (this means that a fatal error
	 *                                  occurred during parsing).
	 * @exception IllegalStateException If root result is not a project.
	 */
	public LyProject getRootProject()
	{
		Object rootResult = getRootElement();

		if (rootResult == null)
			throw new IllegalStateException("BF AppEngine : fatal error during parsing");

		if (!(rootResult instanceof LyProject))
			throw new IllegalStateException("BF AppEngine : root result is not a project");

		return (LyProject)rootResult;
	}

	/**
	 * Gets list of parsed infos.
	 */
	public LyInfoList<LyInfo> getResults()
	{
		return _results;
	}

	/**
     * Gets list of parsed infos.
     */
    public LyInfoList<LyInfo> getExternResults()
    {
        return _externResults;
    }
    
	/**
	 * Add a result to the parser.
	 *
	 * @param info Info created by the parser.
	 */
	void addResult(LyInfo info, boolean extern)
	{
		_results.addInfo(info);
		if (extern)
		    _externResults.addInfo(info);
	}

	/**
	 * Add a results list of the given parser to this parser.
	 * @param parser Parser that adds its results.
	 */
	void addResults(LyInfoParser parser)
	{
		_results.addList(parser._results);
		_externResults.addList(parser._externResults);
	}
	
	/**
	 * Checks if the given info is considered as external or not.
	 * The info could either be unknown by parser or registered as external.
	 * @param info Info to be checked.
	 * @return True if external, false otherwise.
	 */
	public boolean isExtern(LyInfo info)
	{
	    return !_results.contains(info)
	        || _externResults.contains(info);
	}
} // end class LyInfoParser
