/*
 * LyJoinFilter.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.io.*;
import leon.data.*;


/**
 * The LyJoinFilter class implements the description of a filter used to filter objects with a join
 * condition.<br>
 * A join filter is defined by a source class info (i.e the class to make the join), a source field
 * info, a target field info and by a source filter.
 *
 * <p>See also tag <FILTER> of leon.dtd</p>
 *
 * @see LyFilter
 * @see LyFilterExpression
 */
public final class LyJoinFilter extends LyInfo
	implements LyFilterElement
{
	public static final String __VERSION =
		"$Id: LyJoinFilter.java 35476 2012-05-14 08:05:16Z bangiole $";

	/** Source class info on wich join will be computed. */
	LyClassInfo				   _sourceClassInfo;

	/** Filter used on source classInfo objects. */
	LyFilter				   _sourceFilter;

	/** Source fieldInfo on which join will be done. */
	LyFieldInfo				   _sourceFieldInfo;

	/** Target fieldInfo on which join will be done. */
	LyFieldInfo				   _targetFieldInfo;

	/**
	 * Initialisation of filter attributes.
	 */
	protected void init()
	{
		_sourceClassInfo = null;
		_sourceFilter = null;
		_sourceFieldInfo = null;
		_targetFieldInfo = null;
	}

	/**
	 * Constructor of an identified join filter.
	 *
	 * @param id        The identifier of the object.
	 * @param fieldInfo Relation FieldInfo on which this filter expression applies.
	 * @param filter    Filter used for this relation filter.
	 */
	public LyJoinFilter(String id, LyClassInfo classInfo, LyFilter filter,
		LyFieldInfo source, LyFieldInfo target)
	{
		super(id);

		init();

		_sourceClassInfo = classInfo;
		_sourceFilter = filter;
		_sourceFieldInfo = source;
		_targetFieldInfo = target;

		// Set application for dynamic filters
		if (_sourceClassInfo != null)
			_application = _sourceClassInfo.getApplication();
	}

	/**
	 * Constructor of an identified relation filter.
	 *
	 * @param fieldInfo Relation FieldInfo on which this filter expression applies.
	 * @param filter    Filter used for this relation filter.
	 */
	public LyJoinFilter(LyClassInfo classInfo, LyFilter filter,
		LyFieldInfo source, LyFieldInfo target)
	{
		this(null, classInfo, filter, source, target);
	}

	/**
	 * Check if the specified object matches current filter.
	 *
	 * @param  object Object to check.
	 * @return True if the object matches filter criteria, false otherwise.
	 * @see    LyFilterElement#match
	 */
	public boolean match(Object object)
	{
		if (object == null)
			return false;

		// Find the ObjectWrapper
		LyObject lyObject = (LyObject)object;
		LyValue  value = lyObject.getFieldValue(_targetFieldInfo);

		return matchValue(value);
	}

	/**
	 * Check if the specified values matches current relation filter.
	 *
	 * @param  values Values to check.
	 * @return True if the object matches filter criteria, false otherwise.
	 * @see    LyFilterElement#match
	 */
	public boolean match(LyValueSet values)
	{
		if (values == null)
			return false;

		LyValue value = values.getFieldValue(_targetFieldInfo);

		return matchValue(value);
	}

	/**
	 * Check if the specfified target value matches the filter.
	 *
	 * @param  value
	 * @return True if the value matches criteria, false otherwise.
	 */
	protected boolean matchValue(LyValue value)
	{
		LyFilter globalFilter = new LyFilter(LyFilter.Type.AND);
		globalFilter.addElement(new LyFilterExpression(_sourceFieldInfo, value));
		if (_sourceFilter != null)
			globalFilter.addElement(_sourceFilter);

		// Search for all source objects.
		LyObjectList list = new LyClassWorkSpace(null, _sourceClassInfo, null, globalFilter, null,
				false, false, false);

		LyObject     source;
		Object		 s;
		boolean		 match = false;
		Object		 v;

		int			 i, n = list.getSize();

		v = (value == null ? null : value.getValue());
		for (i = 0; (i < n) && !match; i++)
		{
			source = list.getObject(i);
			s = source.getValue(_sourceFieldInfo);
			match = matchValues(s, v);
		}
		return match;
	}

	protected boolean matchValues(Object source, Object target)
	{
		if ((source == null) && (target == null))
			return true;

		if ((source == null) && (target != null))
			return false;

		if ((source != null) && (target == null))
			return false;

		Class<? extends LyFieldInfo> s, t;

		s = _sourceFieldInfo.getClass();
		t = _targetFieldInfo.getClass();

		if (s == t)
			return ((_sourceFieldInfo.equals(source, target))
					|| (_targetFieldInfo.equals(target, source)));

		String s1 = source.toString();
		String s2 = target.toString();

		return (s1.equals(s2));
	}

	/**
	 * Lock this relation filter.
	 */
	public void lock()
	{
		super.lock();

		// Lock filter
		if (_sourceFilter != null)
			_sourceFilter.lock();
	}

	/**
	 * Dumps current relation filter as a String.
	 *
	 * @return String representation of this relation filter.
	 */
	public String dump()
	{
		StringBuffer sb = new StringBuffer(500);

		sb.append((_sourceClassInfo == null) ? "null" : _sourceClassInfo.getName());
		sb.append(" match ");
		sb.append((_sourceFilter == null) ? "null" : _sourceFilter.toString());
		sb.append(" and ");
		sb.append("(");
		sb.append((_sourceFieldInfo == null) ? "null" : _sourceFieldInfo.getName());
		sb.append(" == ");
		sb.append((_targetFieldInfo == null) ? "null" : _targetFieldInfo.getName());
		sb.append(")");

		return sb.toString();
	}

	/**
	 * Writes XML corresponding to this relation filter element.
	 *
	 * @param pw PrintWriter on which the XML is written.
	 */
	public void writeXml(PrintWriter pw)
	{
		// Write tag with its attributes
	}

	/**
	 * Returns a displayable string of this element.
	 *
	 * @return A displayable string representation of this element.
	 */
	public String getName()
	{
		return toString();
	}

	/**
	 * Checks if an expression is found concerning the given field info.
	 *
	 * @param  fieldInfo The fieldInfo.
	 * @return True if an expression with the specified is found, false otherwise.
	 */
	public boolean contains(LyFieldInfo fieldInfo)
	{
		if (_sourceFieldInfo == fieldInfo)
			return true;
		if (_targetFieldInfo == fieldInfo)
			return true;

		return ((_sourceFilter != null) && (_sourceFilter.contains(fieldInfo)));
	}

	/**
	 * Gets the source class info of this join filter.
	 */
	public LyClassInfo getSourceClassInfo()
	{
		return _sourceClassInfo;
	}

	/**
	 * Gets the source fieldInfo on which the join applies.
	 */
	public LyFieldInfo getSourceFieldInfo()
	{
		return _sourceFieldInfo;
	}

	/**
	 * Gets the target fieldInfo on which the join applies.
	 */
	public LyFieldInfo getTargetFieldInfo()
	{
		return _targetFieldInfo;
	}

	/**
	 * Gets the source filter of this join filter.
	 *
	 * @return Filter of this relation filter.
	 */
	public LyFilter getSourceFilter()
	{
		return _sourceFilter;
	}

	/**
	 * Gets list of fieldInfos that this filter refers to.
	 *
	 * @return List of fieldInfos that the filter depends on.
	 */
	public LyInfoList<LyFieldInfo> getFieldInfos()
	{
		LyInfoList<LyFieldInfo> fieldInfos = new LyInfoList<LyFieldInfo>();

		if (_sourceFieldInfo != null)
			fieldInfos.addInfo(_sourceFieldInfo);

		if (_targetFieldInfo != null)
			fieldInfos.addInfo(_targetFieldInfo);

		if (_sourceFilter != null)
			fieldInfos.addList(_sourceFilter.getFieldInfos());

		return fieldInfos;
	}

	/**
	 * Sets the source classInfo.
	 *
	 * @param classInfo source classInfo.
	 */
	public void setSourceClassInfo(LyClassInfo classInfo)
	{
		if (isLocked())
			throw new IllegalStateException("BF AppEngine : join filter " + this
				+ " was locked and thus can't be modified anymore");

		_sourceClassInfo = classInfo;
	}

	/**
	 * Sets the source fieldInfo.
	 *
	 * @param fieldInfo Source fieldInfo.
	 */
	public void setSourceFieldInfo(LyFieldInfo fieldInfo)
	{
		if (isLocked())
			throw new IllegalStateException("BF AppEngine : join filter " + this
				+ " was locked and thus can't be modified anymore");

		_sourceFieldInfo = fieldInfo;
	}

	/**
	 * Sets the target fieldInfo.
	 *
	 * @param fieldInfo Target fieldInfo.
	 */
	public void setTargetFieldInfo(LyFieldInfo fieldInfo)
	{
		if (isLocked())
			throw new IllegalStateException("BF AppEngine : join filter " + this
				+ " was locked and thus can't be modified anymore");

		_targetFieldInfo = fieldInfo;
	}

	/**
	 * Sets the source filter.
	 *
	 * @param filter Source filter.
	 */
	public void setSourceFilter(LyFilter filter)
	{
		if (isLocked())
			throw new IllegalStateException("BF AppEngine : relation filter " + this
				+ " was locked and thus can't be modified anymore");

		_sourceFilter = filter;
	}
} // end class LyJoinFilter
