/**
 * LyMetaModel.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;

import leon.app.*;
import leon.info.LyChartInfo.*;
import leon.info.binding.*;
import leon.info.daemon.*;
import leon.info.location.*;
import leon.info.misc.*;
import leon.info.misc.LyFieldGroup.*;
import leon.info.model.*;
import leon.info.notifier.*;
import leon.misc.*;


/**
 */
public abstract class LyMetaModel
{
	public static final String __VERSION =
		"$Id: LyMetaModel.java 41460 2012-11-08 16:34:27Z nhamon $";

	public static final String 	DEFAULT_ID = "id";

	public static final char 	DEFAULT_CHAR = '_';

	/** new Current application. */
	protected LyApplication    _application;

	/** Root element of the meta-model. */
	protected LyObjectInfo     _rootElement;

	/**
	 * Default constructor with no parameter.
	 */
	protected LyMetaModel()
	{
		_application = null;
		_rootElement = null;
	}

	/**
	 * Method for loading model.
	 */
	public void load(LyApplication application)
	{
		if (_application != null)
			throw new IllegalStateException("BF AppEngine : this model was already loaded");

		_application = application;

		createInfos();
		completeInfos();
	}

	protected abstract void createInfos();

	protected abstract void completeInfos();

	/**
	 * Init given info from its template.
	 */
	protected void initFromTemplate(LyObjectInfo info)
	{
		info.initFromTemplate();
	}

	/**
	 * Init given info from its template for specified parts of th model (BIND, GUI, DATA, LABEL).
	 */
	protected void initFromTemplate(LyObjectInfo info, int parts)
	{
		info.initFromTemplate(parts);
	}

	/**
	 * Access to root element of this meta-model.
	 */
	public LyObjectInfo getRootElement()
	{
		return _rootElement;
	}

	/**
	 * Access to root project of this meta-model.
	 */
	public LyProject getRootProject()
	{
		if (!(_rootElement instanceof LyProject))
			throw new IllegalStateException(
				"BF AppEngine : root element of meta-model is not a project");

		return (LyProject)_rootElement;
	}

	/**
	 * Gets an info from its identifier.
	 */
	protected LyInfo getInfo(String id, boolean raiseError)
	{
		LyInfo info = _application.getInfo(id);

		if (raiseError && (info == null))
			throw new IllegalStateException("BF AppEngine : Info with id \"" + id + "\" not found");

		return info;
	}

	protected final LyInfo getInfo(String id)
	{
		return getInfo(id, true);
	}

	/**
	 * Set root element of this meta-model.
	 */
	protected final void setRootElement(LyObjectInfo rootElement)
	{
		_rootElement = rootElement;
	}

	/**
	 * Set root project of this meta-model.
	 *
	 * @deprecated replaced by setRootElement
	 */
	protected final void setRootProject(LyProject rootProject)
	{
		_rootElement = rootProject;
	}

	/**
	 * Sets application of an info.
	 */
	protected final void setApplication(LyInfo info, LyApplication application)
	{
		info.setApplication(application);
	}

    /**
     * Sets parent of an info.
     */
    protected void setParent(LyInfo info, LyInfo parent)
    {
        if (info.getParent() instanceof LyInclude)
            return;

        // Sets parent to null to avoid exceptions when using Studio/MetaConverter
        info._parent = null;
        info.setParent(parent);
    }

	/* Infos constructors */

	protected BitSet newBitSet(int[] bits)
	{
		if (bits == null)
			return null;

		BitSet bs = new BitSet();
		int    i, n = bits.length;

		for (i = 0; i < n; i++)
			bs.set(bits[i]);
		return bs;
	}

	protected LyBitSet newLyBitSet(LyBitSet.Type type, int[] bits)
	{
		return new LyBitSet(type, newBitSet(bits));
	}

	protected LyInfoList<LyFieldInfo> newInfoList(String[] infoIds)
	{
		if (infoIds == null)
			return null;

		int				   i, n = infoIds.length;
		LyInfoList<LyFieldInfo> list = new LyInfoList<LyFieldInfo>();
		LyInfo			   info;

		for (i = 0; i < n; i++)
		{
			info = getInfo(infoIds[i]);
			list.addInfo((LyFieldInfo)info);
		}

		return list;
	}

	protected LyAction newAction(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyAction action = new LyAction(id);

			action.setApplication(_application);
			return action;
		}
		else if ((info instanceof LyAction)
				&& !(info instanceof LyTabAction)
				&& !(info instanceof LyCompoundAction))
			return (LyAction)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a simple action");
	}

	protected LyAction newAction(String id, LyAction model, String controller,
		String behavior, String builder, String view, String target)
	{
		LyInfo   info = getInfo(id, false);
		LyAction action;

		if (info == null)
			action = new LyAction(id, model, controller, behavior, builder, view, target);
		else if ((info instanceof LyAction)
				&& !(info instanceof LyTabAction)
				&& !(info instanceof LyCompoundAction))
		{
			action = (LyAction)info;
			action.construct(model, controller, behavior, builder, view, target);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a simple action");

		action.setApplication(_application);
		return action;
	}

	protected LyAction newAction(LyAction model, LyClassInfo classInfo, Hashtable<String, Object> parameters)
	{
		LyAction action = new LyAction(model, classInfo, parameters);

		action.setApplication(_application);
		return action;
	}

	protected LyArgument newArgument(String name, Object value)
	{
		LyArgument argument = new LyArgument(name, value);

		return argument;
	}

	protected void setDefaultAction(LyAction action, LyAction defaultAction)
	{
		action.setDefaultAction(defaultAction);
	}

	protected void setActionRef(LyAction action, boolean actionRef)
	{
		action.setActionRef(actionRef);
	}

	protected void setAbstract(LyAction action, boolean abstractAction)
	{
		action.setAbstract(abstractAction);
	}

	protected void setTarget(LyAction action, String target)
	{
		action.setTarget(target);
	}

	protected void setBehavior(LyAction action, String behavior)
	{
		action.setBehavior(behavior);
	}

	protected void setMenu(LyAction action, String menu)
	{
		action.setMenu(menu);
	}

	protected void setMin(LyAction action, int min)
	{
		action.setMin(min);
	}

	protected void setMax(LyAction action, int max)
	{
		action.setMax(max);
	}

	protected void addParameterDecl(LyAction action, String name,
        LyParameterDecl.Type type, boolean required, String domain)
	{
		LyAction	    model = (LyAction)action.getTemplate();
		LyParameterDecl parameterDecl = null;

		if (model != null)
		{
			// Avoid creation of a new declaration if model has the same one
			parameterDecl = model.getParameterDecl(name);
			if ((parameterDecl != null)
					&& parameterDecl.getType().equals(type)
					&& (parameterDecl.isRequired() == required)
					&& ((parameterDecl.getDomain() == domain)
						|| ((domain != null)
							&& (parameterDecl.getDomain() != null)
							&& parameterDecl.getDomain().equals(domain))))
			{
				action.addParameterDecl(parameterDecl);
				return;
			}
		}

		parameterDecl = new LyParameterDecl(name, type, required, domain);
		action.addParameterDecl(parameterDecl);
	}

	protected void addParameterDecl(LyAction action, LyParameterDecl parameterDecl)
	{
		action.addParameterDecl(parameterDecl);
	}

	protected void addParameter(LyAction action, String name, Object value)
	{
		LyAction    model = (LyAction)action.getTemplate();
		LyParameter parameter = null;

		if (model != null)
		{
			// Avoid creation of a parameter if model has the same one
			parameter = model.getLyParameter(name, true);
			if ((parameter != null)
					&& (((value == null)
							&& (parameter.getValue() == null))
						|| ((value != null)
							&& (parameter.getValue() != null)
							&& parameter.getValue().equals(value))))
			{
				action.addParameter(parameter);
				return;
			}
		}

		parameter = new LyParameter(name, value);
		action.addParameter(parameter);
	}

	protected void addParameter(LyAction action, LyParameter parameter)
	{
		action.addParameter(parameter);
	}

	protected void addResource(LyAction action, LyEnvironment.ResourceType type,
         String key, String value)
	{
		action.addResource(new LyActionResource(type, key, value));
	}

	protected void addResource(LyAction action, LyActionResource resource)
	{
		action.addResource(resource);
	}

	protected void setGeometry(LyAction action, LyGeometry geometry)
	{
		action.setGeometry(geometry);
	}

	protected void setShortcut(LyAction action, String shortcut)
	{
		action.setShortcut(shortcut);
	}

	protected void addAction(LyAction action, LyAction subAction)
	{
		action.addAction(subAction, -1);
	}

	protected void setFilter(LyAction action, LyFilter filter)
	{
		action.setFilter(filter);
	}

	protected void addPath(LyInfoLabel label, LyPath path)
	{
		label.addPath(path);
	}

	protected LyChartInfo newChartInfo(String id, String name, String classInfoId, String selectedXFieldInfo,
            String selectedGroupFieldInfo, Type typeVarY, Type typeVarX,
            GraphType selectedGraphType, ValType xVal, String[] xIntervals,
            ValType yVal, String[] yIntervals, boolean pieBy, String title,
            String xTitle, String yTitle, String[] xAxis, String[] yAxis,
            boolean xVerticalLabels, boolean xGrid, boolean yGrid, boolean d3Effect,
            boolean showLegend, boolean showTips, boolean ratios)
	{
		LyChartInfo chartInfo = new LyChartInfo(id, name, classInfoId, selectedXFieldInfo,
				selectedGroupFieldInfo, typeVarY, typeVarX, selectedGraphType, xVal, xIntervals,
				yVal, yIntervals, pieBy, title, xTitle, yTitle, xAxis, yAxis, xVerticalLabels,
				xGrid, yGrid, d3Effect, showLegend, showTips, ratios);

		chartInfo.setApplication(_application);
		return chartInfo;
	}

	protected void addYField(LyChartInfo chart, String yf)
	{
		chart.addYField(yf);
	}

	protected void setPreference(LyChartInfo chart, LyPreference pref)
	{
		chart.setPreference(pref);
	}

	protected LyArrayFieldInfo newArrayFieldInfo(String id, String javaName,
		LyArrayFieldInfo model, int cMin, int cMax)
	{
		LyInfo			 info = getInfo(id, false);
		LyArrayFieldInfo array;

		if (info == null)
			array = new LyArrayFieldInfo(id, javaName, model, (short)cMin, (short)cMax);
		else if (info instanceof LyArrayFieldInfo)
		{
			array = (LyArrayFieldInfo)info;
			array.construct(javaName, model, (short)cMin, (short)cMax);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not an array field info");

		array.setApplication(_application);
		return array;
	}

	protected LyArrayFieldInfo newArrayFieldInfo(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyArrayFieldInfo array = new LyArrayFieldInfo(id);

			array.setApplication(_application);
			return array;
		}
		else if (info instanceof LyArrayFieldInfo)
			return (LyArrayFieldInfo)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not an array field info");
	}

	protected void setFieldInfo(LyArrayFieldInfo array, LyFieldInfo fieldInfo)
	{
		array.setFieldInfo(fieldInfo);
	}

	protected void setDaemon(LyArrayFieldInfo array, LyArrayDaemon daemon)
	{
		array.setDaemon(daemon);
	}

	protected void setNRows(LyArrayFieldInfo array, int nrows)
	{
		array.setNRows(nrows);
	}

	protected LyChoiceFieldInfo newChoiceFieldInfo(String id, String javaName,
		LyChoiceFieldInfo model, boolean multiple, LyChoiceFieldInfo.SortBy sortBy)
	{
		return newChoiceFieldInfo(id, javaName, model, multiple, sortBy,
			LyChoiceFieldInfo.Type.ENUM, false);
	}

	protected LyChoiceFieldInfo newChoiceFieldInfo(String id, String javaName,
		LyChoiceFieldInfo model, boolean multiple, LyChoiceFieldInfo.SortBy sortBy, LyChoiceFieldInfo.Type type, boolean not)
	{
		LyInfo			  info = getInfo(id, false);
		LyChoiceFieldInfo choice;

		if (info == null)
			choice = new LyChoiceFieldInfo(id, javaName, model, multiple, sortBy, type, not);
		else if (info instanceof LyChoiceFieldInfo)
		{
			choice = (LyChoiceFieldInfo)info;
			choice.construct(javaName, model, multiple, sortBy, type, not);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a choice field info");

		choice.setApplication(_application);
		return choice;
	}

	protected LyChoiceFieldInfo newChoiceFieldInfo(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyChoiceFieldInfo choice = new LyChoiceFieldInfo(id);

			choice.setApplication(_application);
			return choice;
		}
		else if (info instanceof LyChoiceFieldInfo)
			return (LyChoiceFieldInfo)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a choice field info");
	}

	protected void addChoiceOption(LyChoiceFieldInfo choice, LyChoiceOption option)
	{
		choice.addChoiceOption(option, -1);
	}

	protected void setSortBy(LyChoiceFieldInfo choice, int sortBy)
	{
		choice.setSortBy(LyChoiceFieldInfo.SortBy.class.getEnumConstants()[sortBy]);
	}

	protected void setNot(LyChoiceFieldInfo choice, boolean not)
	{
		choice.setNot(not);
	}

	protected void setContextField(LyChoiceFieldInfo choice, LyChoiceFieldInfo contextField)
	{
		choice.setContextField(contextField);
	}

	protected void addChoiceOptionGroup(LyChoiceFieldInfo choice,
		LyChoiceOptionGroup group)
	{
		choice.addChoiceOptionGroup(group);
	}

	protected void setNCols(LyChoiceFieldInfo choice, int ncols)
	{
		choice.setNCols(ncols);
	}

	protected void addLabel(LyChoiceOptionGroup group, LyInfoLabel label)
	{
		group.addLabel(label);
	}

	protected LyChoiceOption newChoiceOption(String id, String name, String alias, Object value,
		boolean def, boolean disable, boolean unknown, boolean table)
	{
		LyInfo info = getInfo(id, false);
		LyChoiceOption option;

		if (info == null)
		{
			option = new LyChoiceOption(id, name, alias, value, def, disable, unknown,	table);
		}
		else if (info instanceof LyChoiceOption)
		{
			option = (LyChoiceOption)info;
			option.construct(name, alias, value, def, disable, unknown,	table);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a c");

		option.setApplication(_application);
		return option;
	}

	protected void addRule(LyChoiceOption option, LyRule rule)
	{
		option.addRule(rule);
	}

	protected void addLabel(LyChoiceOption option, LyInfoLabel label)
	{
		option.addLabel(label);
	}

	protected void setName(LyChoiceOption option, String name)
	{
		option.setName(name);
	}

	protected void setAlias(LyChoiceOption option, String alias)
	{
		option.setAlias(alias);
	}

	protected void setValue(LyChoiceOption option, String value)
	{
		option.setValue(value);
	}

	protected void setDisable(LyChoiceOption option, boolean disable)
	{
		option.setDisable(disable);
	}

	protected void setUnknown(LyChoiceOption option, boolean unknown)
	{
		option.setUnknown(unknown);
	}

	protected void setDefault(LyChoiceOption option, boolean def)
	{
		option.setDefault(def);
	}

	protected void setTableMark(LyChoiceOption option, boolean tableMark)
	{
		option.setTableMark(tableMark);
	}

	protected LyChoiceOptionGroup newChoiceOptionGroup(String id, String name,
		String alias, String value)
	{
		LyChoiceOptionGroup group = new LyChoiceOptionGroup(id, name, alias, value);

		group.setApplication(_application);
		return group;
	}

	protected void setName(LyChoiceOptionGroup group, String name)
	{
		group.setName(name);
	}

	protected void setAlias(LyChoiceOptionGroup group, String alias)
	{
		group.setAlias(alias);
	}

	protected void setValue(LyChoiceOptionGroup group, String value)
	{
		group.setValue(value);
	}

	protected void setOption(LyChoiceOptionGroup group, LyChoiceOption option)
	{
		group.setOption(option);
	}

	protected void addChoiceItem(LyChoiceOptionGroup group, LyChoiceItem item)
	{
		group.addChoiceItem(item);
	}

	protected LyClassInfo newClassInfo(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyClassInfo classInfo = new LyClassInfo(id);

			classInfo.setApplication(_application);
			return classInfo;
		}
		else if (info instanceof LyClassInfo)
			return (LyClassInfo)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a class info");
	}

	protected LyClassInfo newClassInfo(String id, LyClassInfo superClass, LyClassInfo model,
		String javaName, String behavior, boolean abstractClass, boolean virtual,
		String userData)
	{
		LyInfo	    info = getInfo(id, false);
		LyClassInfo classInfo;

		if (info == null)
			classInfo = new LyClassInfo(id, superClass, model, javaName, behavior, abstractClass,
					virtual, userData);
		else if (info instanceof LyClassInfo)
		{
			classInfo = (LyClassInfo)info;
			classInfo.construct(superClass, model, javaName, behavior, abstractClass, virtual,
				userData);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a class info");

		classInfo.setApplication(_application);
		return classInfo;
	}

	protected void setBinding(LyClassInfo classInfo, LyClassBinding binding)
	{
		classInfo.setBinding(binding);
	}

	protected void setBehavior(LyClassInfo classInfo, String behavior)
	{
		classInfo.setBehaviorClassName(behavior);
	}

	protected void setCachePolicy(LyClassInfo classInfo, LyInfoCache cachePolicy)
	{
		classInfo.setCachePolicy(cachePolicy);
	}

	protected void addFieldInfo(LyClassInfo classInfo, LyFieldInfo fieldInfo)
	{
		classInfo.addFieldInfo(fieldInfo, -1);
	}

	protected void setSort(LyClassInfo classInfo, LySort sort)
	{
		classInfo.setSort(sort);
	}

	protected void addAction(LyClassInfo classInfo, LyAction action)
	{
		classInfo.addAction(action, -1);
	}

	protected void addRoute(LyClassInfo classInfo, LyRoute route)
	{
		classInfo.addRoute(route, -1);
	}

	protected void addControl(LyClassInfo classInfo, LyControl control)
	{
		classInfo.addControl(control);
	}

	protected LyControl newControl(String id, String javaName, LyControl.Level level,
		int priority, String formula, String message, LyTag.Model type)
	{
		LyInfo    info = getInfo(id, false);
		LyControl control;

		if (info == null)
			control = new LyControl(id, javaName, level, priority, type);
		else if (info instanceof LyControl)
		{
			control = (LyControl)info;
			control.construct(javaName, level, priority, type);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a control");

        control.setFormula(formula);
        control.setMessage(message);
		control.setApplication(_application);
		return control;
	}

	protected LyControl newControl(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyControl control = new LyControl(id);

			control.setApplication(_application);
			return control;
		}
		else if (info instanceof LyControl)
			return (LyControl)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a control");
	}

	protected void addInfoArgument(LyControl control, String name, String value)
	{
		LyInfo info = getInfo(value);

		control.addArgument(new LyArgument(name, info));
	}

	protected void addArgument(LyControl control, String name, Object value)
	{
		control.addArgument(new LyArgument(name, value));
	}

	protected void addArgument(LyControl control, LyArgument arg)
	{
		control.addArgument(arg);
	}

	protected void addRule(LyClassInfo classInfo, LyRule rule)
	{
		classInfo.addRule(rule);
	}

	protected void addLabel(LyClassInfo classInfo, LyInfoLabel label)
	{
		classInfo.addLabel(label);
	}

	protected LyCompoundAction newCompoundAction(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyCompoundAction action = new LyCompoundAction(id);

			action.setApplication(_application);
			return action;
		}
		else if (info instanceof LyCompoundAction)
			return (LyCompoundAction)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a compound action");
	}

	protected LyCompoundAction newCompoundAction(String id, LyAction model,
		String controller, String behavior, String builder,
        LyCompoundAction.Style style, String sizes,
        LyCompoundAction.Context context, LyCompoundAction.ClientContext clientContext,
        String target)
	{
		LyInfo			 info = getInfo(id, false);
		LyCompoundAction action;

		if (info == null)
			action = new LyCompoundAction(id, model, controller, behavior, builder, style, sizes,
					context, clientContext, target);
		else if (info instanceof LyCompoundAction)
		{
			action = (LyCompoundAction)info;
			action.construct(model, controller, behavior, builder, style, sizes, context,
				clientContext, target);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a compound action");

		action.setApplication(_application);
		return action;
	}

	protected void addSubAction(LyCompoundAction compoundAction, LyAction action)
	{
		compoundAction.addSubAction(action);
	}

    protected LyFieldGroup newFieldGroup(String id, String name, State state)
    {
        LyFieldGroup group = new LyFieldGroup(id, name, state);
        group.setApplication(_application);
        return group;
    }

    protected LyFieldTab newFieldTab(String id, String name)
    {
        LyFieldTab tab = new LyFieldTab(id, name);
        tab.setApplication(_application);
        return tab;
    }

	protected final LyDateFieldInfo newDateFieldInfo(String id, String javaName,
		LyDateFieldInfo model, LyDateFieldInfo.Type type)
	{
		return newDateFieldInfo(id, javaName, model, type, false);
	}

	protected LyDateFieldInfo newDateFieldInfo(String id, String javaName,
			LyDateFieldInfo model, LyDateFieldInfo.Type type, boolean withTimeZone)
	{
		LyInfo		    info = getInfo(id, false);
		LyDateFieldInfo date;

		if (info == null)
			date = new LyDateFieldInfo(id, javaName, model, type);
		else if (info instanceof LyDateFieldInfo)
		{
			date = (LyDateFieldInfo)info;
			date.construct(javaName, model, type, withTimeZone);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
					+ " is not a date field info");

		date.setApplication(_application);
		return date;
	}

	protected LyDateFieldInfo newDateFieldInfo(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyDateFieldInfo date = new LyDateFieldInfo(id);

			date.setApplication(_application);
			return date;
		}
		else if (info instanceof LyDateFieldInfo)
			return (LyDateFieldInfo)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a date field info");
	}

	protected void setDisplayFormat(LyDateFieldInfo date, String format)
	{
		date.setDisplayFormat(format);
	}

	protected void setDefaultValue(LyDateFieldInfo date, Object defaultValue)
	{
		date.setDefaultValue(defaultValue);
	}

	protected void addConstraint(LyFieldInfo fieldInfo, LyFieldConstraint constraint)
	{
		fieldInfo.addConstraint(constraint);
	}

	protected LyFieldConstraint newFieldConstraint(LyFieldConstraint.Type type,
		int colSpan, int rowSpan, boolean remainder, int colWeight, int rowWeight,
        LyFieldConstraint.Fill fill, LyFieldConstraint.HAlign hAlign,
        LyFieldConstraint.VAlign vAlign, LyAction action)
	{
		return new LyFieldConstraint(type, colSpan, rowSpan, remainder, colWeight, rowWeight,
			fill, hAlign, vAlign);
	}

	protected void setFieldTag(LyFieldInfo fieldInfo, boolean fieldTag)
	{
		fieldInfo.setFieldTag(fieldTag);
	}

	protected void setBehavior(LyFieldInfo fieldInfo, String behavior)
	{
		fieldInfo.setBehaviorClassName(behavior);
	}

	protected void setBinding(LyFieldInfo fieldInfo, LyFieldBinding binding)
	{
		fieldInfo.setBinding(binding);
	}

	protected void setTab(LyFieldInfo fieldInfo, String tab)
	{
		fieldInfo.setTab(tab);
	}

    protected void setFieldTab(LyFieldInfo fieldInfo, LyFieldTab fieldTab)
    {
        fieldInfo.setFieldTab(fieldTab);
    }

    protected void setFieldGroup(LyFieldInfo fieldInfo, LyFieldGroup fieldGroup)
    {
        fieldInfo.setFieldGroup(fieldGroup);
    }

	protected void setCalculated(LyFieldInfo fieldInfo, boolean calculated)
	{
		fieldInfo.setCalculated(calculated);
	}

	protected void addBean(LyFieldInfo fieldInfo, LyFieldBean bean)
	{
		fieldInfo.addBean(bean);
	}

	protected void setImport(LyFieldInfo fieldInfo, LyFieldInfoImport fieldImport)
	{
		fieldInfo.setImport(fieldImport);
	}

	protected void addDepend(LyFieldInfo fieldInfo, LyFieldInfo depend, boolean strong)
	{
		fieldInfo.addDepend(depend, strong);
	}

	protected void addControl(LyFieldInfo fieldInfo, LyControl control)
	{
		fieldInfo.addControl(control);
	}

	protected void addLabel(LyFieldInfo fieldInfo, LyInfoLabel label)
	{
		fieldInfo.addLabel(label);
	}

	protected void setUnits(LyFieldInfo fieldInfo, LyUnits units)
	{
		fieldInfo.setUnits(units);
	}

	@SuppressWarnings("deprecation")
	protected void setUnit(LyFieldInfo fieldInfo, LyUnit unit)
	{
		fieldInfo.setUnit(unit);
	}

	protected void setPath(LyFieldInfo fieldInfo, LyPath path)
	{
		fieldInfo.setPath(path);
	}

    protected void setFormula(LyFieldInfo fieldInfo, String formula)
    {
        fieldInfo.setFormula(formula);
    }

	protected LyUnits newUnits(String id, String javaName, String name, boolean convert)
	{
		LyInfo  info = getInfo(id, false);
		LyUnits units;

		if (info == null)
			units = new LyUnits(id, javaName);
		else if (info instanceof LyUnits)
		{
			units = (LyUnits)info;
			units.construct(javaName);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a units list");

		units.setName(name);
		units.setConvert(convert);
		units.setApplication(_application);
		return units;
	}

	protected void setName(LyUnits units, String name)
	{
		units.setName(name);
	}

	protected void setConvert(LyUnits units, boolean convert)
	{
		units.setConvert(convert);
	}

	protected LyUnits newUnits(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyUnits units = new LyUnits(id);

			units.setApplication(_application);
			return units;
		}
		else if (info instanceof LyUnits)
			return (LyUnits)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a units");
	}

	protected void addUnit(LyUnits units, LyUnit unit)
	{
		units.addUnit(unit);
	}

	protected LyUnit newUnit(String id, double factor, double offset,
		String javaName, String value, String name, String alias, boolean print)
	{
		LyInfo info = getInfo(id, false);
		LyUnit unit;

		if (info == null)
			unit = new LyUnit(id, factor, offset, javaName);
		else if (info instanceof LyUnit)
		{
			unit = (LyUnit)info;
			unit.construct(factor, offset, javaName);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not an unit");

		unit.completeUnit(value, name, alias, print);
		unit.setApplication(_application);
		return unit;
	}

	protected void completeUnit(LyUnit unit, String value, String name, String alias, boolean print)
	{
		unit.completeUnit(value, name, alias, print);
	}

	protected LyUnit newUnit(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyUnit unit = new LyUnit(id);

			unit.setApplication(_application);
			return unit;
		}
		else if (info instanceof LyUnit)
			return (LyUnit)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a unit");
	}

	protected LyPreference newPreference(String id, String objectInfoId, String name,
	        LyFieldInfo fixedColumn, int nbObjects, String user, boolean defaultPref,
	        String columnSizes)
    {
        LyInfo info = getInfo(id, false);

        if (info == null)
        {
            LyPreference preference = new LyPreference(id, objectInfoId, name, fixedColumn,
                    nbObjects, user, defaultPref, columnSizes);

            preference.setApplication(_application);
            return preference;
        }
        else if (info instanceof LyPreference)
            return (LyPreference)info;
        else
            throw new IllegalStateException("BF AppEngine : extern info " + id
                + " is not a preference");
    }

	protected LyUnit newUnit(String id, String value)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyUnit unit = new LyUnit(null, value);

			unit.setApplication(_application);
			return unit;
		}
		else if (info instanceof LyUnit)
			return (LyUnit)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a unit");
	}

	protected void setFormat(LyUnit unit, LyFormat format)
	{
		unit.setFormat(format);
	}

	protected LyFieldInfoImport newFieldInfoImport(LyRelationFieldInfo relation,
		LyClassInfo classInfo, LyFilter filter, LyFieldInfo fieldInfo)
	{
		return new LyFieldInfoImport(relation, classInfo, filter, fieldInfo);
	}

	protected LyFileFieldInfo newFileFieldInfo(String id, String javaName, LyFileFieldInfo model,
		String mimeTypes, String attachmentHandler, String attachmentDirectory,
		String directory, String filter)
	{
		LyInfo		    info = getInfo(id, false);
		LyFileFieldInfo file;

		if (info == null)
			file = new LyFileFieldInfo(id, javaName, model, mimeTypes, attachmentHandler,
					attachmentDirectory, directory, filter);
		else if (info instanceof LyFileFieldInfo)
		{
			file = (LyFileFieldInfo)info;
			file.construct(javaName, model, mimeTypes, attachmentHandler, attachmentDirectory,
				directory, filter);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a file field info");

		file.setApplication(_application);
		return file;
	}

	protected LyFileFieldInfo newFileFieldInfo(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyFileFieldInfo file = new LyFileFieldInfo(id);

			file.setApplication(_application);
			return file;
		}
		else if (info instanceof LyFileFieldInfo)
			return (LyFileFieldInfo)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a file field info");
	}

	protected void setAttachmentHandler(LyFileFieldInfo file, String value)
	{
		file.setAttachmentHandler(value);
	}

	protected void setAttachmentDirectory(LyFileFieldInfo file, String value)
	{
		file.setAttachmentDirectory(value);
	}

	protected void setNRows(LyFileFieldInfo file, int nrows)
	{
		file.setNRows(nrows);
	}

	protected void setNCols(LyFileFieldInfo file, int ncols)
	{
		file.setNCols(ncols);
	}

	protected void setDefault(LyFileFieldInfo file, String defaultValue)
	{
		file.setDefault(defaultValue);
	}

	protected LyNotifier newNotifier(String id, String javaName, String properties)
	{
		LyNotifier notifier = new LyNotifier(id, javaName, properties);

		notifier.setApplication(_application);
		return notifier;
	}

	protected LyNumberFieldInfo newNumberFieldInfo(String id, String javaName,
		LyNumberFieldInfo model, LyNumberFieldInfo.Type type, boolean modulo)
	{
		LyInfo			  info = getInfo(id, false);
		LyNumberFieldInfo number;

		if (info == null)
			number = new LyNumberFieldInfo(id, javaName, model, type, modulo);
		else if (info instanceof LyNumberFieldInfo)
		{
			number = (LyNumberFieldInfo)info;
			number.construct(javaName, model, type, modulo);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a number field info");

		number.setApplication(_application);
		return number;
	}

	protected LyNumberFieldInfo newNumberFieldInfo(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyNumberFieldInfo number = new LyNumberFieldInfo(id);

			number.setApplication(_application);
			return number;
		}
		else if (info instanceof LyNumberFieldInfo)
			return (LyNumberFieldInfo)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a number field info");
	}

	protected void setFunction(LyNumberFieldInfo number, LyNumberFieldInfo.Function function)
	{
		number.setFunction(function);
	}

	protected void setMin(LyNumberFieldInfo number, Number min)
	{
		number.setMin(min);
	}

	protected void setMax(LyNumberFieldInfo number, Number max)
	{
		number.setMax(max);
	}

	protected void setPitch(LyNumberFieldInfo number, Number pitch)
	{
		number.setPitch(pitch);
	}

	protected void setDecimals(LyNumberFieldInfo number, int decimals)
	{
		number.setDecimals(decimals);
	}

	protected void setPrecision(LyNumberFieldInfo number, Number precision)
	{
		number.setPrecision(precision);
	}

	protected void setDefault(LyNumberFieldInfo number, Number defaultValue)
	{
		number.setDefault(defaultValue);
	}

	protected void setFormat(LyNumberFieldInfo number, String format)
	{
		number.setFormat(format);
	}

	protected void setName(LyObjectInfo objectInfo, String name)
	{
		objectInfo.setName(name);
	}

	protected void setAlias(LyObjectInfo objectInfo, String alias)
	{
		objectInfo.setAlias(alias);
	}

	protected void setGroup(LyObjectInfo objectInfo, String group)
	{
		objectInfo.setGroup(group);
	}

	protected void setFilePrefix(LyObjectInfo objectInfo, String prefix)
	{
		objectInfo.setFilePrefix(prefix);
	}

	protected void setTip(LyObjectInfo objectInfo, String tip)
	{
		objectInfo.setTip(tip);
	}

	protected void putData(LyObjectInfo objectInfo, String name, String data,
	    LyTag.Model modelType)
	{
		objectInfo.putData(name, data, modelType);
	}

	protected void setMarks(LyObjectInfo objectInfo, int[] marksTable)
	{
		objectInfo.setMarks(newBitSet(marksTable));
	}

	protected void setMarks(LyObjectInfo objectInfo, BitSet marks)
	{
		objectInfo.setMarks(marks);
	}

	protected void check(LyObjectInfo objectInfo, LyObjectInfo parent)
	{
		objectInfo.check(parent);
	}

	protected void check(LyObjectInfo objectInfo, LyObjectInfo parent, int parts)
	{
		objectInfo.check(parent, parts);
	}

	protected LyProject newProject(String id, LyProject model, String version,
		String applicationBehavior, String sessionBehavior)
	{
		LyProject project = new LyProject(id, model, version, applicationBehavior, sessionBehavior);

		project.setApplication(_application);
		return project;
	}

	protected void setRootAction(LyProject project, LyAction action)
	{
		project.setRootAction(action);
	}

	protected void addInfo(LyProject project, LyInfo info)
	{
		project.addInfo(info);
	}

	protected void addAction(LyProject project, LyAction action)
	{
		project.addAction(action);
	}

	protected void addFieldInfo(LyProject project, LyFieldInfo fieldInfo)
	{
		project.addFieldInfo(fieldInfo);
	}

	protected void addClassInfo(LyProject project, LyClassInfo classInfo)
	{
		project.addClassInfo(classInfo);
	}

	protected void addProject(LyProject project, LyProject subProject)
	{
		project.addProject(subProject);
	}

	protected void addFilter(LyProject project, LyFilter filter)
	{
		project.addFilter(filter);
	}

	protected void addSort(LyProject project, LySort sort)
	{
		project.addSort(sort);
	}

	protected void addLocation(LyProject project, LyLocation location)
	{
		project.addLocation(location);
	}

	protected void addNotifier(LyProject project, LyNotifier notifier)
	{
		project.addNotifier(notifier);
	}

	protected void addInclude(LyProject project, LyInclude include)
	{
		include.setApplication(_application);
	}
	
	protected void setApplicationBehaviorClassName(LyProject project, String applicationBehaviorClassName)
	{
		project.setApplicationBehaviorClassName(applicationBehaviorClassName);
	}
	
	protected void setSessionBehaviorClassName(LyProject project, String sessionBehaviorClassName)
	{
		project.setSessionBehaviorClassName(sessionBehaviorClassName);
	}

	protected LyRelationFieldInfo newRelationFieldInfo(String id, String javaName,
		LyRelationFieldInfo extendedRelation, LyRelationFieldInfo model,
		boolean multiple, boolean ordered, LyRelationFieldInfo.Direction direction,
        LyRelationFieldInfo.Type type, Object userData, boolean noContext, boolean noToolbar,
        LyRelationFieldInfo.ShowAs showAs)
	{
		LyInfo			    info = getInfo(id, false);
		LyRelationFieldInfo relation;

		if (info == null)
			relation = new LyRelationFieldInfo(id, javaName, extendedRelation, model, multiple,
					ordered, direction, type, userData, noContext, noToolbar, showAs);
		else if (info instanceof LyRelationFieldInfo)
		{
			relation = (LyRelationFieldInfo)info;
			relation.construct(javaName, extendedRelation, model, multiple, ordered, direction,
				type, userData, noContext, noToolbar, showAs);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a relation field info");

		relation.setApplication(_application);
		return relation;
	}

	protected LyRelationFieldInfo newRelationFieldInfo(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyRelationFieldInfo relation = new LyRelationFieldInfo(id);

			relation.setApplication(_application);
			return relation;
		}
		else if (info instanceof LyRelationFieldInfo)
			return (LyRelationFieldInfo)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a relation field info");
	}

	protected void setClassField(LyRelationFieldInfo relation, LyFieldInfo classField)
	{
		relation.setClassField(classField);
	}

	/**
	 * Accessor to the noContext property.
	 *
	 * @param noContext True if this relation field does not use the context of the views to compute
	 *                  the list of logical objects that may be used as values.
	 */
	protected void setIgnoresContext(LyRelationFieldInfo relation, boolean noContext)
	{
		relation.setIgnoresContext(noContext);
	}

	protected void setShowAs(LyRelationFieldInfo relation, LyRelationFieldInfo.ShowAs showAs)
	{
		relation.setShowAs(showAs);
	}

	/**
	 * Accessor to the noToolbar property.
	 *
	 * @param noToolbar True if this relation doesn't have a toolbar associated to it in edit forms.
	 */
	protected void setToolbarDisabled(LyRelationFieldInfo relation, boolean noToolbar)
	{
		relation.setToolbarDisabled(noToolbar);
	}

	protected void setContextField(LyRelationFieldInfo relation, LyRelationFieldInfo contextField)
	{
		relation.setContextField(contextField);
	}

	protected void addClassInfo(LyRelationFieldInfo relation, LyClassInfo classInfo,
		LyFilter filter)
	{
		relation.addClassInfo(classInfo, filter, -1);
	}

	protected void setDaemon(LyRelationFieldInfo relation, LyInfoDaemon daemon)
	{
		relation.setDaemon(daemon);
	}

	protected void setNRows(LyRelationFieldInfo relation, int nrows)
	{
		relation.setNRows(nrows);
	}

	protected void setNCols(LyRelationFieldInfo relation, int ncols)
	{
		relation.setNCols(ncols);
	}

	/**
	 * Handle cardinality for the given relation.
	 * If min value <= 0 : the field is optional
	 * If max value < 0 : the field is multiple and unlimited
	 * @param relation the relation
	 * @param min the min number of objects
	 * @param max the max number of objects
	 */
	protected void setCardinality(LyRelationFieldInfo relation, int min, int max)
	{
		// handle multiplicity
		relation.setMultiple(max != 1);

		// handle optionality
		BitSet marks = relation.getMarks();

		if (min <= 0)
		{
			if (marks == null)
				marks = new BitSet();
			marks.set(LyFieldInfo.MARK_OPTIONAL);
		}
		else if (marks != null)
			marks.clear(LyFieldInfo.MARK_OPTIONAL);

		relation.setMarks(marks);

		if (min > 0)
			relation.setMin(min);
		else
			relation.setMin(-1);

		if (max > 1)
			relation.setMax(max);
		else
			relation.setMax(-1);
	}

	protected void setMin(LyRelationFieldInfo relation, int min)
	{
		relation.setMin(min);
	}

	protected void setMax(LyRelationFieldInfo relation, int max)
	{
		relation.setMax(max);
	}

	protected void setDefault(LyRelationFieldInfo relation, String defaultValue)
	{
		relation.setDefault(defaultValue);
	}

	protected void addRoute(LyRelationFieldInfo relation, LyRoute route)
	{
		relation.addRoute(route, -1);
	}

	protected void setAbstract(LyRelationFieldInfo relation, boolean abstractRelation)
	{
		relation.setAbstract(abstractRelation);
	}

	protected void addSubType(LyRelationFieldInfo relation, LyRelationFieldInfo subType)
	{
		relation.addSubType(subType);
	}

	protected LyStructFieldInfo newStructFieldInfo(String id, String javaName,
		LyStructFieldInfo extendedRelation, LyStructFieldInfo model,
		String className, boolean abstractClass, boolean fieldTag)
	{
		LyInfo			  info = getInfo(id, false);
		LyStructFieldInfo struct;

		if (info == null)
		{
			struct = new LyStructFieldInfo(id);
			struct.setFieldTag(fieldTag);
			struct.construct(javaName, extendedRelation, model, className, abstractClass);
		}
		else if (info instanceof LyStructFieldInfo)
		{
			struct = (LyStructFieldInfo)info;
			struct.setFieldTag(fieldTag);
			struct.construct(javaName, extendedRelation, model, className, abstractClass);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a struct field info");

		struct.setApplication(_application);
		return struct;
	}

	protected LyStructFieldInfo newStructFieldInfo(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyStructFieldInfo struct = new LyStructFieldInfo(id);

			struct.setApplication(_application);
			return struct;
		}
		else if (info instanceof LyStructFieldInfo)
			return (LyStructFieldInfo)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a struct field info");
	}

	protected void addFieldInfo(LyStructFieldInfo struct, LyFieldInfo fieldInfo)
	{
		LyClassInfo classInfo = struct.getClassInfo(0);

		classInfo.addFieldInfo(fieldInfo, -1);
	}

	protected void setBinding(LyStructFieldInfo struct, LyClassBinding binding)
	{
		LyClassInfo classInfo = struct.getClassInfo(0);

		classInfo.setBinding(binding);
	}

	protected void setCachePolicy(LyStructFieldInfo struct, LyInfoCache cachePolicy)
	{
		LyClassInfo classInfo = struct.getClassInfo(0);

		classInfo.setCachePolicy(cachePolicy);
	}

	protected void addRule(LyStructFieldInfo struct, LyRule rule)
	{
		LyClassInfo classInfo = struct.getClassInfo(0);

		classInfo.addRule(rule);
	}

	protected void setClassBehavior(LyStructFieldInfo struct, String classBehavior)
	{
		LyClassInfo classInfo = struct.getClassInfo(0);

		classInfo.setBehaviorClassName(classBehavior);
	}

	protected void addParentClass(LyStructFieldInfo struct, LyClassInfo classInfo,
		LyFieldInfo fieldInfo)
	{
		struct.addParentClass(classInfo, fieldInfo);
	}

	protected LyRoute newRoute(String id, LyClassInfo target, Vector<LyStep> steps, boolean reverse,
		boolean bidirectional, boolean hidden)
	{
		LyRoute route = new LyRoute(id, target, steps, reverse, bidirectional, hidden);

		route.setApplication(_application);
		return route;
	}

	protected void setTarget(LyRoute route, LyClassInfo classInfo)
	{
		route.setTarget(classInfo);
	}

	protected void addStep(LyRoute route, LyStep step)
	{
		route.addStep(step);
	}

    protected void check(LyRoute route, LyClassInfo parent)
    {
        route.check(parent);
    }

	protected LyPath newPath(String id, LyClassInfo target, LyClassInfo classInfo,
		LyFieldInfo fieldInfo, LyInfoLabel fieldLabel, LyTag.Model modelType)
	{
		LyPath path = new LyPath(id, target, classInfo, fieldInfo, fieldLabel);

		path.setModelType(modelType);
		path.setApplication(_application);
		return path;
	}

	protected void setSourceClassInfo(LyPath path, LyClassInfo classInfo)
	{
		path.setSourceClassInfo(classInfo);
	}

	protected void setTargetFieldInfo(LyPath path, LyFieldInfo fieldInfo)
	{
		path.setTargetFieldInfo(fieldInfo);
	}

	protected void setFieldLabel(LyPath path, LyInfoLabel label)
	{
		path.setFieldLabel(label);
	}

	protected LyRule newRule(String id, LyObjectInfo objectInfo, int priority,
        LyRule.Operation operation, boolean set, Object value, LyFilter filter,
		LyInfoLabel label, String formula, LyTag.Model modelType)
	{
		LyRule rule = new LyRule(id, objectInfo, priority, operation, set, value, modelType);

		rule.setFilter(filter);
		rule.setLabel(label);
		rule.setFormula(formula);
		rule.setApplication(_application);
		return rule;
	}

	protected void setFilter(LyRule rule, LyFilter filter)
	{
		rule.setFilter(filter);
	}

	protected LySort newSort(String id, LyInfo model)
	{
		LySort sort = new LySort(id, (LySort)model);

		sort.setApplication(_application);
		return sort;
	}

	protected LySortCriterion newSortCriterion(LyFieldInfo fieldInfo, boolean increasing)
	{
		LySortCriterion criterion = new LySortCriterion(fieldInfo, increasing);

		criterion.setApplication(_application);
		return criterion;
	}

	protected LyStep newStep(LyClassInfo target, LyInfoList<LyFieldInfo> relations, LyFilter filter,
		boolean reverse)
	{
		LyStep step = new LyStep(target, relations, filter, reverse);

		step.setApplication(_application);
		return step;
	}

	protected LyStep newStep(String id, LyClassInfo target, LyInfoList<LyFieldInfo> relations, LyFilter filter,
		boolean reverse)
	{
		LyStep step = new LyStep(id, target, relations, filter, reverse);

		step.setApplication(_application);
		return step;
	}

	protected LyTabAction newTabAction(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyTabAction action = new LyTabAction(id);

			action.setApplication(_application);
			return action;
		}
		else if (info instanceof LyTabAction)
			return (LyTabAction)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a tab action");
	}

	protected LyTabAction newTabAction(String id, LyAction model, String controller,
		String behavior, String builder, int clientContext, String target,
        LyTabAction.Type type)
	{
		LyInfo	    info = getInfo(id, false);
		LyTabAction action;

		if (info == null)
			action = new LyTabAction(id, model, controller, behavior, builder, (short)clientContext,
					target, type);
		else if (info instanceof LyTabAction)
		{
			action = (LyTabAction)info;
			action.construct(model, controller, behavior, builder, (short)clientContext, target, type);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a tab action");

		action.setApplication(_application);
		return action;
	}

	protected void addTabAction(LyTabAction tabAction, LyAction action)
	{
		tabAction.addTabAction(action);
	}

	protected final LyTextFieldInfo newTextFieldInfo(String id, String javaName,
		LyTextFieldInfo model, boolean multiple, LyTextFieldInfo.Type type)
	{
		return this.newTextFieldInfo(id, javaName, model, multiple, type, false);
	}

	protected LyTextFieldInfo newTextFieldInfo(String id, String javaName,
		LyTextFieldInfo model, boolean multiple, LyTextFieldInfo.Type type,
		boolean multilingual)
	{
		LyInfo		    info = getInfo(id, false);
		LyTextFieldInfo text;

		if (info == null)
			text = new LyTextFieldInfo(id, javaName, model, multiple, type, multilingual);
		else if (info instanceof LyTextFieldInfo)
		{
			text = (LyTextFieldInfo)info;
			text.construct(javaName, model, multiple, type, multilingual);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a text field info");

		text.setApplication(_application);
		return text;
	}

	protected LyTextFieldInfo newTextFieldInfo(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyTextFieldInfo text = new LyTextFieldInfo(id);

			text.setApplication(_application);
			return text;
		}
		else if (info instanceof LyTextFieldInfo)
			return (LyTextFieldInfo)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a text field info");
	}

	protected void setNRows(LyTextFieldInfo text, int nrows)
	{
		text.setNRows(nrows);
	}

	protected void setNCols(LyTextFieldInfo text, int ncols)
	{
		text.setNCols(ncols);
	}

	protected void setMinChars(LyTextFieldInfo text, int minChars)
	{
		text.setMinChars(minChars);
	}

	protected void setMaxChars(LyTextFieldInfo text, int maxChars)
	{
		text.setMaxChars(maxChars);
	}

	protected void setFormat(LyTextFieldInfo text, String format)
	{
		text.setFormat(format);
	}

	protected void setDefault(LyTextFieldInfo text, String defaultValue)
	{
		text.setDefault(defaultValue);
	}

	protected void setRegExp(LyTextFieldInfo text, LyRegExp regExp)
	{
		text.setRegExp(regExp);
	}

	protected void addCharType(LyTextFieldInfo text, LyInfoCharType charType)
	{
		text.addCharType(charType);
	}

	protected void setEncoding(LyTextFieldInfo text, String encoding)
	{
		text.setEncoding(encoding);
	}

	protected LyRegExp newRegExp(String id, String value)
	{
		LyInfo   info = getInfo(id, false);
		LyRegExp regExp;

		if (info == null)
			regExp = new LyRegExp(id, value);
		else if (info instanceof LyRegExp)
		{
			regExp = (LyRegExp)info;
			regExp.construct(value);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a reg exp");

		regExp.setApplication(_application);
		return regExp;
	}

	protected LyRegExp newRegExp(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyRegExp regExp = new LyRegExp(id);

			regExp.setApplication(_application);
			return regExp;
		}
		else if (info instanceof LyRegExp)
			return (LyRegExp)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a regexp");
	}


	protected LyFieldTab newFieldTab(String id, String name, String prefix)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyFieldTab fieldtab = new LyFieldTab(id, name);
			fieldtab.setPrefix(prefix);
			fieldtab.setApplication(_application);
			return fieldtab;
		}
		else if (info instanceof LyFieldTab)
			return (LyFieldTab)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a fieldtab");
	}

	protected LyFieldGroup newFieldGroup(String id, String name, String prefix,
        LyFieldGroup.State state)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyFieldGroup fieldgroup = new LyFieldGroup(id, name, state);
			fieldgroup.setPrefix(prefix);
			fieldgroup.setApplication(_application);
			return fieldgroup;
		}
		else if (info instanceof LyFieldGroup)
			return (LyFieldGroup)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a fieldgroup");
	}

	protected LyInfoCharType newCharType(String id, String value)
	{
		LyInfo		   info = getInfo(id, false);
		LyInfoCharType charType;

		if (info == null)
			charType = new LyInfoCharType(id, value);
		else if (info instanceof LyInfoCharType)
		{
			charType = (LyInfoCharType)info;
			charType.construct(value);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a char type");

		charType.setApplication(_application);
		return charType;
	}

	protected LyInfoCharType newCharType(String id)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyInfoCharType charType = new LyInfoCharType(id);

			charType.setApplication(_application);
			return charType;
		}
		else if (info instanceof LyInfoCharType)
			return (LyInfoCharType)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a char type");
	}

	protected LyParameterDecl newParameterDecl(String name, LyParameterDecl.Type type,
        boolean required, String domain)
	{
		return new LyParameterDecl(name, type, required, domain);
	}

	protected LyFilter newFilter(String id, LyTag.Model modelType)
	{
		LyInfo info = getInfo(id, false);

		if (info == null)
		{
			LyFilter filter = new LyFilter(id);

			filter.setModelType(modelType);
			filter.setApplication(_application);
			return filter;
		}
		else if (info instanceof LyFilter)
			return (LyFilter)info;
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a filter");
	}

	protected LyFilter newFilter(String id, LyFilter template, LyFilter.Type type,
	    LyTag.Model modelType)
	{
		LyInfo   info = getInfo(id, false);
		LyFilter filter;

		if (info == null)
		{
			filter = new LyFilter(id, template, type);
			filter.setModelType(modelType);
		}
		else if (info instanceof LyFilter)
		{
			filter = (LyFilter)info;
			filter.construct(template, type);
		}
		else
			throw new IllegalStateException("BF AppEngine : extern info " + id
				+ " is not a filter");

		filter.setApplication(_application);
		return filter;
	}

	protected LyInfoLabel newLabel(String id, String value, LyInfoLabel.Type type,
		String subType, LyTag.Model modelType)
	{
		LyInfoLabel label = new LyInfoLabel(id, value, type, subType);

		label.setModelType(modelType);
		label.setApplication(_application);
		return label;
	}

	protected void addPart(LyInfoLabel label, LyPath path, LyUnit unit)
	{
		label.addPart(path, unit);
	}

	protected void addModel(LyProject project, LyModel model)
	{
		project.addModel(model);
	}

	protected void addFieldBinding(LyFieldsBinding fbinds, LyFieldBinding fbind)
	{
		fbinds.addFieldBinding(fbind);
	}

    /**
     * Gets a valid XML id from the given name (remove spaces, accents, strange chars, etc.).
     *
     * @param name the name
     * @return a valid id
     */
    public static String getValidId(String id)
    {
        if (id == null || id.length() == 0)
            id = DEFAULT_ID;

        int             i = 0, n = id.length();
        StringBuffer    buffer = new StringBuffer(n);
        String          regExpPonct1 = "\\.+-/=<>!?:;,|%`~";
        String          regExpPonct2 = "\\*\\#&$\\(\\){}\\]\'\"";
        char            c;

        for (i = 0; i < n; i++)
        {
            c = id.charAt(i);
            if ((c >= 'A') && (c <= 'Z'))
                buffer.append(c);
            else
            {
                if ((c == '') || (c == '') || (c == ''))
                    c = 'a';
                else if ((c == '') || (c == '') || (c == '') || (c == ''))
                    c = 'e';
                else if ((c == '') || (c == '') || (c == ''))
                    c = 'o';
                else if ((c == '') || (c == ''))
                    c = 'i';
                else if ((c == '') || (c == ''))
                    c = 'u';
                else if (c == '')
                    c = 'c';
                else if (c == ' ')
                    c = DEFAULT_CHAR;
                else if (regExpPonct1.indexOf(c) >= 0)
                    c = DEFAULT_CHAR;
                else if (regExpPonct2.indexOf(c) >= 0)
                    c = DEFAULT_CHAR;

                if ((c != '_')
                    || ((i > 0)
                        && (i < n - 1)
                        && ((buffer.length() == 0)
                            || (buffer.charAt(buffer.length() - 1) != '_'))))
                    buffer.append(c);
            }
        }

        // Check if id is really valid
        String	newId = buffer.toString();

        c = newId.charAt(0);
        if (((c < 'A') || (c > 'Z'))
	        	&& ((c < 'a') || (c > 'z'))
	        	&& (c != '_'))
        	newId = DEFAULT_ID + DEFAULT_CHAR + newId;

        return newId;
    }
} // end class LyMetaModel
