/*
 * LyNumberFieldInfo.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.text.*;
import java.util.*;
import leon.data.*;
import leon.misc.*;


/**
 * This class implements the description of numeric fields such as integers, floats, percent ...
 *
 * <p>The different types of numbers supported are represented as public, static attributes, named
 * TYPE_XXX. Supported types are INT, SHORT, LONG, FLOAT, DOUBLE.</p>
 *
 * <p>A NumbeFieldInfo is the implementation of a numeric attribute in an application class. In
 * addition to standard fieldInfo properties, instances of LyNumberFieldInfo also implements a
 * minimum and a maximum value. Moreover, it is possible to specify a pitch increment (or decrement)
 * which may be used when editing the field.</p>
 *
 * <p>When real numbers (FLOAT, DOUBLE) are used, the number of decimals can be specified.</p>
 *
 * <p>A unit may also be associated to this kind of field, this unit is only used to be displayed to
 * the end-user when values of number fields are shown.</p>
 *
 * <p>A function may be associated to a number field. This function defines the method to calculate
 * the value corresponding to a column of values when this field is shown in a table. Supported
 * functions are given has public, static attributes, named Function.XXX. Accepted values are : SUM,
 * AVERAGE, NONE (no function).</p>
 *
 * <p>Finally, number fields may also have a formula. This formula is a simple mathematic expression
 * involving other fields of the same application class, used to calculate values for the number
 * field (for example : suppose that this field represents a tax, and that this tax is ration of a
 * price, the formula could be "(price * 10)/100").</p>
 *
 * <p>See also tag <NUMBER> of leon.dtd</p>
 */
public final class LyNumberFieldInfo extends LyFieldInfo
{
	public static final String __VERSION =
		"$Id: LyNumberFieldInfo.java 41050 2012-10-05 15:38:27Z smaupin $";

	/** Number types. */
    public enum Type
    {
        INT,
        SHORT,
        LONG,
        FLOAT,
        DOUBLE
    }

    /** Functions. */
    public enum Function
    {
    	// Function none : no function associated with this field.
    	NONE,
    	// Function sum : function sum associated with this field.
    	SUM,
    	// Function average : function average associated with this field.
    	AVERAGE,
    	// Function average : other function associated with this field.
    	OTHER
    }

	/**  */
	public static final double LY_DOUBLE_MIN_VALUE = -Double.MAX_VALUE;

	/**  */
	public static final float  LY_FLOAT_MIN_VALUE = -Float.MAX_VALUE;

	/** Indicates whether precision on numbers must be used to compare two values. */
	public static boolean	   USE_PRECISION = false;

    /** Number type. */
    Type                        _type;

	/** Minimum value for this fieldInfo or null if none is specified. */
	Number					   _minValue;

	/** Maximum value for this fieldInfo or null if none is specified. */
	Number					   _maxValue;

	/** Indicates if this number is managed as a modulo. */
	boolean					   _modulo;

	/** Pitch (base increment in graphical interface). */
	Number					   _pitch;

	/** Number of decimals shown when displaying reals. */
	int						   _decimals;

	/** Precision used for comparing 2 values for this field info. */
	Number					   _precision;

	/** Function associated with this field : valid values are defined as constants Function.XXX. */
	Function				   _function;

	/** Indicates whether format of this number field may be changed automatically. */
	private boolean			   _allowFormatAutomaticUpdate;


	/**
	 * Init method.
	 */
	protected void init()
	{
		super.init();

		_type = Type.INT;
		_minValue = null;
		_maxValue = null;
		_modulo = false;
		_pitch = null;
		_decimals = -1;
		_precision = null;
		_function = Function.NONE;
		_allowFormatAutomaticUpdate = true;
	}

	/**
	 * Init this instance attributes from its template.
	 *
	 * @param parts Indicates which parts of the template are copied.
	 */
	void initFromTemplate(int parts)
	{
        parts &= ~_initParts;
        if (parts == 0)
            return;
		if (_template == null)
			return;

		super.initFromTemplate(parts);

		LyNumberFieldInfo template = (LyNumberFieldInfo)_template;

		if ((parts & PART_DATA) > 0)
		{
			if (template.getType() == getType())
			{
				_defaultValue = template._defaultValue;
				_minValue = template._minValue;
				_maxValue = template._maxValue;
				_precision = template._precision;
			}
			else
			{
				_defaultValue = getValueFromObject(template._defaultValue, null);
				_minValue = (Number)getValueFromObject(template._minValue, null);
				_maxValue = (Number)getValueFromObject(template._maxValue, null);
				_precision = (Number)getValueFromObject(template._precision, null);
			}
		}
		if ((parts & PART_GUI) > 0)
		{
			if (template.getType() == getType())
				_pitch = template._pitch;
			else
				_pitch = (Number)getValueFromObject(template._pitch, null);
			_decimals = template._decimals;
			_function = template._function;
		}
	} // end method initFromTemplate

	/**
	 * Constructor of a number field.
	 *
	 * @param id       Identifier of the field.
	 * @param javaName Name of the generated java field.
	 * @param template Template used to construct this field.
	 * @param type     Type of the number field info. One of the TYPE_XXX values
	 * @param modulo   Indicates if this number is a modulo.
	 */
	LyNumberFieldInfo(String id, String javaName, LyNumberFieldInfo template,
		Type type, boolean modulo)
	{
		this(id);
		construct(javaName, template, type, modulo);
	}

	/**
	 * Constructor of a number field.
	 *
	 * @param id Identifier of the field.
	 */
	LyNumberFieldInfo(String id)
	{
		super(id);
	}

	/**
	 * Construction of a number field.
	 *
	 * @param javaName Name of the generated java field.
	 * @param template Template used to construct this field.
	 * @param type     Type of the number field info. One of the TYPE_XXX values
	 * @param modulo   Indicates if this number is a modulo.
	 */
	protected void construct(String javaName, LyNumberFieldInfo template,
		Type type, boolean modulo)
	{
		super.construct(javaName, template);

        _type = (type == null ? Type.INT : type);
		if ((template == null)
				&& ((type == Type.INT)
					|| (type == Type.SHORT)
					|| (type == Type.LONG)))
			_decimals = 0;
		_modulo = modulo;
	}

	/**
	 * Final step of parsing : check consistency. Checks performed are the following :
	 *
	 * <ul>
	 *   <LI>min value should be lower or equal to max value,</LI>
	 *   <LI>default value should be beetween min and max,</LI>
	 *   <LI>if the field has a formula, detects circular dependency.</LI>
	 * </ul>
	 *
	 * @param     parent Object info in which this field is checked.
	 * @param     parts  Indicates which parts are checked.
	 * @exception IllegalStateException If one of these checks fails.
	 */
	void check(LyObjectInfo parent, int parts)
	{
		super.check(parent, parts);

		if ((parts & PART_DATA) > 0)
		{
			// Check if min <= max
			if ((_minValue != null)
					&& (_maxValue != null)
					&& (compare(_minValue, _maxValue) > 0))
				throw new IllegalStateException(
					"BF AppEngine : min should be lower or equal to max on number field " + this);

			// Default value has to be beetween min and max
			if ((_defaultValue != null)
					&& (((_minValue != null) && (compare(_defaultValue, _minValue) < 0))
						|| ((_maxValue != null) && (compare(_defaultValue, _maxValue) > 0))))
				throw new IllegalStateException(
					"BF AppEngine : default value should be greater or equal to min and lower or equal to max on number field "
					+ this);

			// Modulo requires min and max
			if (_modulo && ((_minValue == null) || (_maxValue == null)))
				throw new IllegalStateException("BF AppEngine : number field " + this
					+ " can't be a modulo because modulo requires a min and max for this number");

			// If min value is null, force its to min value
			if (_minValue == null)
			{
				switch (getType())
				{
				case INT:
					_minValue = new Integer(Integer.MIN_VALUE);
					break;
				case SHORT:
					_minValue = new Short(Short.MIN_VALUE);
					break;
				case LONG:
					_minValue = new Long(Long.MIN_VALUE);
					break;
				case DOUBLE:
					_minValue = new Double(LY_DOUBLE_MIN_VALUE);
					break;
				case FLOAT:
					_minValue = new Float(LY_FLOAT_MIN_VALUE);
					break;
				}
			}

			// If max value is null, force its to max value
			if (_maxValue == null)
			{
				switch (getType())
				{
				case INT:
					_maxValue = new Integer(Integer.MAX_VALUE);
					break;
				case SHORT:
					_maxValue = new Short(Short.MAX_VALUE);
					break;
				case LONG:
					_maxValue = new Long(Long.MAX_VALUE);
					break;
				case DOUBLE:
					_maxValue = new Double(Double.MAX_VALUE);
					break;
				case FLOAT:
					_maxValue = new Float(Float.MAX_VALUE);
					break;
				}
			}

			if (_decimals != 0)
			{
				switch (getType())
				{
				case INT:
				case SHORT:
				case LONG:
					getEnvironment().logWarning("BF AppEngine : number field " + this
						+ " shouldn't have decimals");
					_decimals = 0;
					break;
				default:
					break;
				}
			}
		}
	} // end method check

	/* Implementation of LyFieldInfo methods */

	/**
	 * Accessor to the type of the field. Meaning of this property depends on the implementation in
	 * subclasses
	 *
	 * @return Type of this fieldInfo, or TYPE_DEFAULT if type is not set.
	 */
	public Type getType()
	{
		if (!isLoaded())
			load();
		return _type;
	}

	/**
	 * Returns class of values for this field. The value class corresponding to a number field is :
	 *
	 * <ul>
	 *   <LI>java.lang.Short, if number type is SHORT,</LI>
	 *   <LI>java.lang.Integer, if number type is INT,</LI>
	 *   <LI>java.lang.Long, if number type is LONG,</LI>
	 *   <LI>java.lang.Float, if number type is FLOAT,</LI>
	 *   <LI>java.lang.Double, if number type is DOUBLE.</LI>
	 * </ul>
	 */
	public Class<?> getValueClass()
	{
		switch (getType())
		{
		case INT:
			return Integer.class;
		case SHORT:
			return Short.class;
		case LONG:
			return Long.class;
		case DOUBLE:
			return Double.class;
		case FLOAT:
			return Float.class;
		default:
			return null;
		}
	}

	/**
	 * Builds the appropriate LyValue corresponding to this field for the given value object.
	 *
	 * @see LyNumberValue
	 */
	public LyValue buildFieldValue(Object value)
	{
		if (value == null)
			return new LyNumberValue(this);
		else if (value instanceof Number)
			return new LyNumberValue(this, (Number)value);
		else if (value instanceof String)
			return new LyNumberValue(this, (String)value);
		else
			return super.buildFieldValue(value);
	}

	/**
	 * Processes the given value to get the right value before transmitting it to the peer layer.
	 *
	 * @param  value Value to process.
	 * @return The processed value.
	 */
	public Object processValue(Object value)
	{
		// Extract the right value from the given one
		Number number = (Number)super.processValue(value);

		if (!_modulo || (number == null))
			return number;

		// Use modulo to transform the given number
		double d = number.doubleValue();
		double max = _maxValue.doubleValue();
		double min = _minValue.doubleValue();

		if ((d >= min) && (d <= max))
			return number;

		double modulo = max - min;

		if (d > max)
		{
			d -= Math.ceil(Math.abs(d - max) / modulo) * modulo;
			while (d > max)
				d -= modulo;
		}
		else
		{
			d += Math.ceil(Math.abs(d - min) / modulo) * modulo;
			while (d < min)
				d += modulo;
		}

		// Now, get the right type of number
		LyNumberValue numberValue = new LyNumberValue(this, new Double(d));

		return numberValue.getValue();
	} // end method processValue

	/**
	 * Builds internal value for an instance of LyNumberFieldInfo from a string (previously
	 * constructed by toString).
	 *
	 * @see LyNumberFieldInfo#toString
	 */
	public Object parse(String stringValue)
		throws LyFieldInfoFormatException
	{
		if ((stringValue == null) || (stringValue.length() <= 0))
			return null;

		LyEnvironment environment = getEnvironment();

		try
		{
			switch (getType())
			{
			case INT:
				return Integer.valueOf(stringValue);
			case SHORT:
				return Short.valueOf(stringValue);
			case LONG:
				return Long.valueOf(stringValue);
			case FLOAT:
			{
				stringValue = stringValue.replace(',', '.'); // Pb Web
				return Float.valueOf(stringValue);
			}
			case DOUBLE:
			{
				stringValue = stringValue.replace(',', '.'); // Pb Web
				return Double.valueOf(stringValue);
			}
			default:
				return null;
			}
		}
		catch (NumberFormatException e)
		{
			// Raise an exception

			//FFT00237
			String message = environment.getMessage("LY_ER_NUMBER_FORMAT", getName());

			throw new LyFieldInfoFormatException("BF AppEngine : " + message);
		}
	} // end method parse

	/**
	 * Return the specified value for this field as a "displayable" string.
	 */
	public String toName(Object fieldValue)
	{
		if (fieldValue == null)
			return null;

		if (fieldValue.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ fieldValue.getClass() + " expected " + getValueClass());

		String name = getFormattedName(fieldValue);

		if (name != null)
			return name;

		String result = toShortName(fieldValue);

		// If any unit is specified, append it at the end
		if ((result != null)
				&& (result.length() != 0)
				&& (getUnit() != null))
		{
			result += " ";
			result += getUnit();
		}

		return result;
	}

	/**
	 * Return the specified value for this field as a short "displayable" string. This method is
	 * used when presenting values in tables. In that case, values are presented whithout unit.
	 *
	 * @see LyNumberFieldInfo#toName
	 */
	public String toShortName(Object fieldValue)
	{
		if (fieldValue == null)
			return null;

		if (fieldValue.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ fieldValue.getClass() + " expected " + getValueClass());

		String name = getFormattedShortName(fieldValue);

		if (name != null)
			return name;

		LyEnvironment environment = getEnvironment();
		NumberFormat  numberFormat = null;

		if (getFormat() != null)
			numberFormat = environment.getNumberFormat(getFormat());
		else if (!hasMark(LyFieldInfo.MARK_ID))
		{
			int digits = getDecimals();

            if (digits >= 0)
			{
				// We may have a format for digits
				StringBuffer buildFormat = new StringBuffer();

				buildFormat.append("###,###,###,##0");
				if (digits > 0)
					buildFormat.append('.');
				for (int i = 0; i < digits; i++)
					buildFormat.append('0');
				numberFormat = environment.getNumberFormat(buildFormat.toString());

				if (_allowFormatAutomaticUpdate)
					setFormat(buildFormat.toString());
			}
			else
			{
				numberFormat = environment.getNumberFormat(null);

				if (_allowFormatAutomaticUpdate)
					setFormat(((DecimalFormat)numberFormat).toPattern());
			}
		}

		// Format the result depending on the Object type
		StringBuffer buffer = new StringBuffer();

		if (numberFormat != null)
		{
			//FFT00173
			switch (getType())
			{
			case INT:
				buffer.append(numberFormat.format(((Number)fieldValue).intValue()));
				break;
			case SHORT:
				buffer.append(numberFormat.format(((Number)fieldValue).shortValue()));
				break;
			case LONG:
				buffer.append(numberFormat.format(((Number)fieldValue).longValue()));
				break;
			case DOUBLE:
				if (((Double)fieldValue).isNaN())
					buffer.append("NaN");
				else if (((Double)fieldValue).isInfinite()
						|| (((Double)fieldValue).doubleValue() == Double.MAX_VALUE)
						|| (((Double)fieldValue).doubleValue() == LY_DOUBLE_MIN_VALUE))
					buffer.append(fieldValue.toString());
				else
					buffer.append(numberFormat.format(((Number)fieldValue).doubleValue()));
				break;
			case FLOAT:
				buffer.append(numberFormat.format(((Number)fieldValue).floatValue()));
				break;
			default:
				buffer.append(numberFormat.format(((Number)fieldValue).doubleValue()));
				break;
			}
		}
		else
			buffer.append(fieldValue.toString());

		return buffer.toString();
	} // end method toShortName

	/**
	 * Converts the specified value for this field to an "internal" string (string that enables
	 * reconstruction through parse).
	 *
	 * @see LyNumberFieldInfo#parse
	 */
	public String toString(Object fieldValue)
	{
		if (fieldValue == null)
			return null;

		if (fieldValue.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ fieldValue.getClass() + " expected " + getValueClass());

		String string = getFormattedString(fieldValue);

		if (string != null)
			return string;

		int decimals = getDecimals();

		if (decimals > 0)
		{
			double value = Double.parseDouble(((Number)fieldValue).toString());

			value = LyTools.checkDecimals(value, decimals);
			return String.valueOf(value);
		}
		else // Decimals <= 0
		{
			String value = fieldValue.toString();

			if (value == null)
				return value;

			value = LyTools.checkDecimals(value, decimals);
			return value;
		}
	} // end method toString

	/**
	 * Sets the parameter indicating whether the format may be changed automatically.
	 * If true (default value), format of this number field will be computed automatically when the
	 * <code>toShortName(Object)</code> is called if it has not been set yet. Otherwise, format will
	 * not be set.
	 *
	 * @param update true if format must be changed automatically, false otherwise.
	 */
	public void allowFormatAutomaticUpdate(boolean update)
	{
		_allowFormatAutomaticUpdate = update;
	}

	/**
	 * Checks the given value : - the value must be greater or equal to min value, - the value must
	 * be lower or equal to max value.
	 */
	public String checkValue(LyEnvironment environment, Object fieldValue)
	{
		if (fieldValue == null)
			return super.checkValue(environment, fieldValue);

		if (fieldValue.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ fieldValue.getClass() + " expected " + getValueClass());

		if (environment == null)
			environment = getEnvironment();

		if ((_minValue != null) && (compare(fieldValue, _minValue) < 0))
			return environment.getMessage("LY_ER_MIN", getName());
		if ((_maxValue != null) && (compare(fieldValue, _maxValue) > 0))
			return environment.getMessage("LY_ER_MAX", getName());

		return super.checkValue(environment, fieldValue);
	}

	/**
	 * Compare two numeric values for the specified fieldInfo. If a precision is specified on the
	 * fieldInfo, it is used for equality comparison.
	 */
	public int compare(Object value1, Object value2, BitSet modifiers)
	{
		double d1;
		double d2;

		if (value1 == value2)
			return 0; // identical.
		if (value1 == null)
			return -1;
		if (value2 == null)
			return 1;

		Class<?> valueClass = getValueClass();

		if (value1.getClass() != valueClass)
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value1.getClass() + " expected " + valueClass);
		if (value2.getClass() != valueClass)
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value2.getClass() + " expected " + valueClass);
		if (valueClass == Double.class)
		{
			Double db1 = (Double)value1;
			Double db2 = (Double)value2;

			// Specific case for double to take into account Double.NaN value
			if (db1.isNaN() || db1.isInfinite()
					|| db2.isNaN() || db2.isInfinite())
				return db1.compareTo(db2);

			d1 = db1.doubleValue();
			d2 = db2.doubleValue();
		}
		else if (_precision == null)
		{
			if (valueClass == Short.class)
			{
				Short s1 = (Short)value1;
				Short s2 = (Short)value2;

				return s1.compareTo(s2);

			}
			else if (valueClass == Integer.class)
			{
				Integer i1 = (Integer)value1;
				Integer i2 = (Integer)value2;

				return i1.compareTo(i2);
			}
			else if (valueClass == Long.class)
			{
				Long l1 = (Long)value1;
				Long l2 = (Long)value2;

				return l1.compareTo(l2);
			}
			else
			{
				d1 = Double.parseDouble(((Number)value1).toString());
				d2 = Double.parseDouble(((Number)value2).toString());
			}
		}
		else
		{
			d1 = Double.parseDouble(((Number)value1).toString());
			d2 = Double.parseDouble(((Number)value2).toString());
		}

		if (_precision != null)
		{
			if (USE_PRECISION)
			{
				double precision = Double.parseDouble(_precision.toString());
				double d1min = d1 - precision;
				double d1max = d1 + precision;

				if ((d2 > d1min) && (d2 < d1max))
					return 0;
			}
			else
			{
				getEnvironment().logWarning("Precision not used on field " + getId() +
						". Set LyNumberFieldInfo.USE_PRECISION=true if you want to use precision.");

				if (d1 == d2)
					return 0;
			}
		}
		else if (d1 == d2)
			return 0;

		return ((d1 > d2) ? 1 : -1);
	} // end method compare

	/**
	 * Returns the filter conditions authorized on this field. The values must match the short
	 * constants defined in LyFilterExpression.
	 *
	 * @return An array of conditions corresponding to the authorized conditions on this field.
	 * @see    leon.info.LyFilterExpression
	 */
	public LyFilterExpression.Condition[] getFilterConditions()
	{
        LyFilterExpression.Condition[] conditions = new LyFilterExpression.Condition[5];

		conditions[0] = LyFilterExpression.Condition.EQUALS;
		conditions[1] = LyFilterExpression.Condition.LOWER_THAN;
		conditions[2] = LyFilterExpression.Condition.LOWER_OR_EQUALS;
		conditions[3] = LyFilterExpression.Condition.GREATER_THAN;
		conditions[4] = LyFilterExpression.Condition.GREATER_OR_EQUALS;

		return conditions;
	}

	/**
	 * Constructs a value for this fieldInfo from an object for a given instance. Supported
	 * conversions are the following :
	 *
	 * <TABLE>
	 *   <TR>
	 *     <TD>Number, String, Character</TD>
	 *     <TD>Short, Integer</TD>
	 *   </TR>
	 *   <TR>
	 *     <TD>Number, String</TD>
	 *     <TD>Long, Double, Float</TD>
	 *   </TR>
	 * </TABLE>
	 *
	 * @exception LyUnsupportedFieldBindingException If the given object can't be converted to a
	 *                                               value of this field.
	 */
	public Object getValueFromObject(Object object, LyObject instance)
	{
        Object value = decodePeerValue(object);

		if (value == null)
		{
			if ((_formula != null) && hasMark(LyFieldInfo.MARK_LOCAL))
				return _formula.getValue(instance);
			else
				return null;
		}

	    if (object == null)
	        return super.getValueFromObject(object, instance);

		Class<?> valueClass = getValueClass();

		if (value.getClass() == valueClass)
		{
			if (value instanceof Double)
			{
				double v = ((Number)value).doubleValue();

				if (v == Double.POSITIVE_INFINITY)
					value = new Double(Double.MAX_VALUE);
				else if (v == Double.NEGATIVE_INFINITY)
					value = new Double(LY_DOUBLE_MIN_VALUE);
			}
			else if (value instanceof Float)
			{
				float v = ((Number)value).floatValue();

				if (v == Float.POSITIVE_INFINITY)
				    value = new Float(Float.MAX_VALUE);
				else if (v == Float.NEGATIVE_INFINITY)
				    value = new Float(LY_FLOAT_MIN_VALUE);
			}

			return value;
		}

		if (value instanceof Number)
		{
			if (valueClass == Integer.class)
				return new Integer(((Number)value).intValue());
			else if (valueClass == Short.class)
				return new Short(((Number)value).shortValue());
			else if (valueClass == Long.class)
				return new Long(((Number)value).longValue());
			else if (valueClass == Double.class)
				return new Double(((Number)value).doubleValue());
			else if (valueClass == Float.class)
				return new Float(((Number)value).floatValue());
		}
		else if (value instanceof String)
		{
			try
			{
				if (valueClass == Integer.class)
					return Integer.valueOf((String)value);
				else if (valueClass == Short.class)
					return Short.valueOf((String)value);
				else if (valueClass == Long.class)
					return Long.valueOf((String)value);
				else if (valueClass == Double.class)
					return Double.valueOf((String)value);
				else if (valueClass == Float.class)
					return Float.valueOf((String)value);
			}
			catch (NumberFormatException e)
			{
				throw new LyUnsupportedFieldBindingException("BF AppEngine : " + e.getMessage());
			}
		}
		else if (value instanceof Character)
		{
			if (valueClass == Integer.class)
				return new Integer(((Character)value).charValue());
			else if (valueClass == Short.class)
				return new Short((short)((Character)value).charValue());
		}

		return super.getValueFromObject(value, instance);
	} // end method getValueFromObject

	/**
	 * Returns an object of the given class from a value of this fieldInfo. Supported conversions
	 * are the following :
	 *
	 * <TABLE>
	 *   <TR>
	 *     <TD>Short, Integer, Long, Double, Float</TD>
	 *     <TD>Number, String, Character</TD>
	 *   </TR>
	 * </TABLE>
	 *
	 * @exception LyUnsupportedFieldBindingException If the requested conversion if none of the
	 *                                               supported conversions.
	 */
	public Object getPeerValue(Object value, Class<?> objectClass, LyObject instance)
	{
        value = encodePeerValue(value, objectClass);

        if (value == null)
            return null;

		if (value.getClass() == objectClass)
		{
			if (objectClass == Double.class)
			{
				double v = ((Number)value).doubleValue();

				if (v == Double.POSITIVE_INFINITY)
					value = new Double(Double.MAX_VALUE);
				else if (v == Double.NEGATIVE_INFINITY)
					value = new Double(LY_DOUBLE_MIN_VALUE);
			}
			else if (objectClass == Float.class)
			{
				float v = ((Number)value).floatValue();

				if (v == Float.POSITIVE_INFINITY)
					value = new Float(Float.MAX_VALUE);
				else if (v == Float.NEGATIVE_INFINITY)
					value = new Float(LY_FLOAT_MIN_VALUE);
			}

			return value;
		}

        if (objectClass == Number.class)
            return value;
        else if ((objectClass == Integer.class)
                 || (objectClass == int.class))
			return new Integer(((Number)value).intValue());
		else if ((objectClass == Short.class)
                 || (objectClass == short.class))
			return new Short(((Number)value).shortValue());
		else if ((objectClass == Long.class)
                 || (objectClass == long.class))
			return new Long(((Number)value).longValue());
		else if ((objectClass == Double.class)
                 || (objectClass == double.class))
			return new Double(((Number)value).doubleValue());
		else if ((objectClass == Float.class)
                 || (objectClass == float.class))
			return new Float(((Number)value).floatValue());
		else if (objectClass == String.class)
			return value.toString();
		else if (objectClass == Character.class)
			return new Character((char)((Number)value).intValue());
		else if (objectClass == Object.class)
		    return value;

        return super.getPeerValue(value, objectClass, instance);
	} // end method getPeerValue

	/**
	 * Method for computing a value using function associated to this field for a given objects
	 * list. This method is called when this field is presented in a table that has a footer.
	 *
	 * @param  objectList List of objects for which a footer value is computed.
	 * @param  importField True if calling comes on computing of a field import, false otherwise.
	 * @return Value computed with field function for the given objects. Returns null if this field
	 *         has no associated function.
	 */
	public Object compute(LyObjectList objectList, boolean importField)
	{
		if (objectList == null)
			return null;

		if (!importField
		    && ((_function == Function.NONE)
		        || (_function == Function.OTHER)))
		    return null;

		// Workaround for java bug : values are multiplied by precision to avoid errors on operations
		int i, precision = 1;

		if ((getType() == Type.FLOAT) || (getType() == Type.DOUBLE))
		{
			int decimals = getDecimals();

			if (decimals <= 0)
				decimals = 3;

			for (i = 0; i < decimals; i++)
				precision = precision * 10;
		}

		Number   value;
		double   doubleValue;
		long     result = 0;
		int		 n = ((objectList == null) ? 0 : objectList.getSize());
		int		 nr = 0;
		LyObject object;
		boolean  noValue = true;

		for (i = 0; i < n; i++)
		{
			object = objectList.getObject(i);
			value = (Number)object.getValue(this);
			if (value != null)
			{
				doubleValue = value.doubleValue();
				if (!Double.isNaN(doubleValue))
				{
					result += Math.round(doubleValue * precision);
					nr++;
					noValue = false;
				}
			}
		}

		if (noValue)
			return null;

		if ((n != 0) && (_function == Function.AVERAGE))
			result = result / nr;

		value = new Double((double)result / precision);

		switch (getType())
		{
		case INT:
			value = new Integer(value.intValue());
			break;
		case SHORT:
			value = new Short(value.shortValue());
			break;
		case LONG:
			value = new Long(value.longValue());
			break;
		case DOUBLE:
			// value is already computed as a Double
			break;
		case FLOAT:
			value = new Float(value.floatValue());
			break;
		default:
			value = null;
		}

		return value;
	} // end method compute

	/**
	 * Convert a value to the field unit from another unit.
	 *
	 * @param  value The value to convert.
	 * @param  unit  The old unit.
	 * @return The value converted to the field unit.
	 */
	public Object convertFromFieldUnit(Object value, LyUnit unit)
	{
		if (value == null)
			return null;

		LyUnit unitRef = getFieldUnit();

		if (unit == unitRef)
			return value;

		Object convert = super.convertFromFieldUnit(value, unit);

		if ((convert != value) || (unit == null))
			return convert;

		// no javaname specified
		Number numberValue = (Number)value;
		double doubleValue = numberValue.doubleValue();
		double factor = unit.getFactor();
		double offset = unit.getOffset();

		if (unitRef != null)
		{
			doubleValue += unitRef.getOffset();
			doubleValue *= unitRef.getFactor();
		}
		if (factor != 0)
			doubleValue /= factor;
		doubleValue -= offset;

		switch (getType())
		{
		case SHORT:
			return new Short((short)Math.round(doubleValue));
		case LONG:
			return new Long(Math.round(doubleValue));
		case DOUBLE:
			return new Double(doubleValue);
		case FLOAT:
			return new Float((float)doubleValue);
		default:
			return new Integer((int)Math.round(doubleValue));
		}
	} // end method convertFromFieldUnit

	/**
	 * Convert a value from the field unit to the given unit.
	 *
	 * @param  value The value to convert.
	 * @param  unit  The new unit.
	 * @return The value converted from the reference unit.
	 */
	public Object convertToFieldUnit(Object value, LyUnit unit)
	{
		if (value == null)
			return null;

		LyUnit unitRef = getFieldUnit();

		if (unit == unitRef)
			return value;

		Object convert = super.convertToFieldUnit(value, unit);

		if ((convert != value) || (unit == null))
			return convert;

		Number numberValue = (Number)value;
		double doubleValue = Double.parseDouble(numberValue.toString());
		double factor = unit.getFactor();
		double offset = unit.getOffset();

		doubleValue += offset;
		doubleValue *= factor;

		if (unitRef != null)
		{
			factor = unitRef.getFactor();
			offset = unitRef.getOffset();

			if (factor != 0)
				doubleValue /= factor;
			doubleValue -= offset;
		}

		switch (getType())
		{
		case SHORT:
			return new Short((short)Math.round(doubleValue));
		case LONG:
			return new Long(Math.round(doubleValue));
		case DOUBLE:
			return new Double(doubleValue);
		case FLOAT:
			return new Float((float)doubleValue);
		default:
			return new Integer((int)Math.round(doubleValue));
		}
	} // end method convertToFieldUnit

	/**
	 * Gets the minimum value for this field, or null if no minimum value is specified.
	 *
	 * @return Minimum value of this number field;
	 */
	public Number getMin()
	{
		return _minValue;
	}

	/**
	 * Gets the maximum value for this field, or null if no maximum value is specified.
	 *
	 * @return Maximum value of this number field;
	 */
	public Number getMax()
	{
		return _maxValue;
	}

	/**
	 * Checks if this number is a modulo.
	 *
	 * @return True if this number is a modulo, false otherwise.
	 */
	public boolean isModulo()
	{
		return _modulo;
	}

	/**
	 * Gets the pitch (increment).
	 *
	 * @return Current pitch or null if no pitch is specified.
	 */
	public Number getPitch()
	{
		return _pitch;
	}

	/**
	 * Gets the number of digits displayed when shown to the user.
	 *
	 * @return Number of digits or -1 if no limit is specified.
	 */
	public int getDecimals()
	{
		return ((_type == Type.INT) || (_type == Type.LONG) || (_type == Type.SHORT) ? 0 : _decimals);
	}

	/**
	 * Gets the precision used for computing.
	 *
	 * @return Precision used.
	 */
	public Number getPrecision()
	{
		return _precision;
	}

	/**
     * Gets the default value. This default value if an instance of the class returned by
     * getValueClass.
     *
     * @return Default value for this fieldInfo, null if no default value was set.
     */
    public Number getDefault()
    {
        return (Number)_defaultValue;
    }

	/**
	 * Gets the unit string. If the unit is not null, this symbol is shown after the value when
	 * displayed to the user.
	 *
	 * @return     Current unit, or null if none is specified.
	 * @deprecated see getFieldUnit()
	 */
	public String getUnit()
	{
		return (_unit == null ? null : _unit.getName());
	}

	/**
	 * Gets the function associated to this number field. The function is used for computing the
	 * value displayed to the end-user when presenting a table view that contains a column with a
	 * footer showing values corresponding to this number field.
	 *
	 * @return Current function or null if no function is set.
	 */
	public Function getFunction()
	{
		return _function;
	}

	/**
	 * Sets the default value for this number field.
	 *
	 * @param defaultValue Default value (may be null)
	 */
	void setDefault(Number defaultValue)
	{
	    setDefaultValue(LyInfoTokens.parseNumber(_type, defaultValue));
	}

	/**
	 * Sets the minimum value for this field, or null if no minimum value is specified.
	 *
	 * @param minValue Minimum value to set.
	 */
	void setMin(Number minValue)
	{
	    _minValue = LyInfoTokens.parseNumber(_type, minValue);
	}

	/**
	 * Sets the maximum value for this field, or null if no maximum value is specified.
	 *
	 * @param maxValue Maximum value to set.
	 */
	void setMax(Number maxValue)
	{
		_maxValue = LyInfoTokens.parseNumber(_type, maxValue);
	}

	/**
	 * Sets the pitch (increment).
	 *
	 * @param pitch Current pitch or null if no pitch is specified.
	 */
	void setPitch(Number pitch)
	{
		_pitch = LyInfoTokens.parseNumber(_type, pitch);
	}

	/**
	 * Sets format of this field. The format determines how values of this fields are presented to
	 * the end-user. Overriden to remove cached numberFormat when format is changed.
	 *
	 * @param format Format encoded as a string, or null if no format is set.
	 */
	void setFormat(String format)
	{
		super.setFormat(format);
	}

	/**
	 * Sets the number of digits displayed when the user enters values.
	 *
	 * @param decimals Number of digits or -1 if nothing specified.
	 */
	void setDecimals(int decimals)
	{
		_decimals = (_type == Type.INT || _type == Type.LONG || _type == Type.SHORT ? 0 : decimals);
	}

	/**
	 * Sets the precision used for computing.
	 *
	 * @param precision Precision of this field or null if no precision is specified.
	 */
	void setPrecision(Number precision)
	{
		_precision = LyInfoTokens.parseNumber(_type, precision);
	}

	/**
	 * Sets the function associated to this field.
	 *
	 * @param     function Current function.
	 * @exception IllegalArgumentException If the given function is unknown.
	 */
	void setFunction(Function function)
	{
		_function = (function == null ? Function.NONE : function);
	}

	/**
	 * Sets number type.
	 *
	 * @param     type Number type.
	 * @exception IllegalArgumentException If the given type is not supported.
	 */
	void setType(Type type)
	{
	    _type = type;
	}

    /**
     * Sets whether this number is a modulo.
     *
     * @param modulo True if this number is a modulo, false otherwise.
     */
    void setModulo(boolean modulo)
    {
        _modulo = modulo;
    }
} // end class LyNumberFieldInfo
