/**
 * LyObjectInfo.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;

import leon.info.misc.*;


/**
 * This class is an abstract class that contains generic behavior of the classes which represent the
 * data information read from the data model description file. These classes are : LyClassInfo,
 * LyFieldInfo, LyAction, LyProject, LyFilter and LySort. This class inherits from LyInfo class :
 * this means that instances of this class have an id and belong to an application. The information
 * added on this class is the following :
 *
 * <ul>
 *   <LI><b>template</b>. Reference of another instance of the same class that is used as template.
 *     This allows to create a new object by copying an existing one giving only what is different
 *     from the template object.</LI>
 *   <LI><b>name</b>. Name of the instance. This name can be displayed to the user and is translated
 *     in different languages.</LI>
 *   <LI><b>alias</b>. This is a "short name" for the instance used when a small width is given to
 *     display the name. If no alias is specified (null), the name is used instead.</LI>
 *   <LI><b>group name</b>. Name of the "logical group" to which this instance belongs. Instances
 *     with the same value of the group attributes are considered as a group and are grouped
 *     together when building some views of an application.</LI>
 *   <LI><b>filePrefix</b>. Prefix used for building file names corresponding to this instance. File
 *     names are built by concatenating this prefix to a suffix (.gif, .hlp, ...). For example :
 *     <i>object</i>.gif could be the icon representation of the instance that has <i>object</i> as
 *     prefix).</LI>
 *   <LI><b>marks</b>. Marks are used to attach properties to instances of this class. Marks are
 *     simply a bitset where each bit represents a property. There are two types of marks : generic
 *     marks that implied a generic processing or behavior and specific marks (application defined
 *     marks) that could be used by specific coding. The generic marks are defined in each of the
 *     classes that inherit from this class. They are defined as public static short attributes
 *     which names start with MARK_.</LI>
 *   <LI><b>tip</b>. Tip is a short string help that can be displayed on graphic reprsentations of
 *     instance of this class. For example tool tip presented on the icon representing this
 *     instance.</LI>
 *   <LI><b>datas</b>. This a free hashtable that can store pairs of key, value read from the data
 *     description or set in the application code. Developers may freely use this data to store any
 *     piece of information.
 *
 *     <p>Accessors are given to access all these attributes. But they can only be set by the parser
 *     when reading the data description file.</p>
 *
 *     <P>Additionally this class implements the Cloneable interface : this means that all
 *     subclasses may need to override the clone() method (from Object) to duplicated some of the
 *     information they have in reference.</P>
 *   </LI>
 * </ul>
 *
 * @version $Id: LyObjectInfo.java 41313 2012-10-23 08:48:10Z bangiole $
 * @author  Lyria
 */
public abstract class LyObjectInfo extends LyInfo
	implements Cloneable
{
	public static final String __VERSION =
		"$Id: LyObjectInfo.java 41313 2012-10-23 08:48:10Z bangiole $";

	public static final int    PART_DATA = 1;
	public static final int    PART_BIND = 2;
	public static final int    PART_GUI = 4;
	public static final int    PART_LABEL = 8;
	public static final int    PART_ALL = PART_DATA | PART_BIND | PART_GUI | PART_LABEL;

	public static final String NULL = "_null".intern();

	/** Template from which this instance was initialized. */
	LyObjectInfo			   _template;

	/** Initialized parts. */
	int						   _initParts;

	/** Name of the instance. */
	String					   _name;

	/** Alias of the instance. */
	String					   _alias;

	/** Name of group to which this instance belongs. */
	String					   _group;

	/** The prefix of the instance. */
	String					   _filePrefix;

	/** Marks attached to this instance. */
	BitSet					   _marks;

	/** Tip info. */
	String					   _tip;

	/** Hashtable of the specific data information. */
	Hashtable<Object, LyData>  _datas;

	/** User data. */
	Object					   _userData;

	/**
	 * Init method.
	 */
	protected void init()
	{
		super.init();

		_template = null;
		_name = null;
		_alias = null;
		_group = null;
		_filePrefix = null;
		_marks = null;
		_tip = null;
		_datas = null;
		_userData = null;
	}

	/**
	 * Init this instance attributes from its template.
	 */
	void initFromTemplate()
	{
	    int    parts = PART_ALL & ~_initParts;

	    if (parts > 0)
	        initFromTemplate(parts);
	}

	/**
	 * Init this instance attributes from its template.
	 *
	 * @param parts Indicates which parts of the template are copied.
	 */
	void initFromTemplate(int parts)
	{
        parts &= ~_initParts;
        if (parts == 0)
            return;
		if (_template == null)
			return;

		// Load template if needed
		_template.load();

		if (_template._marks != null)
		{
			if (_marks == null)
				_marks = new BitSet();
			if (parts == PART_ALL)
				_marks.or(_template.getMarks());
			else
				_marks.or(_template.getMarks(parts));
		}

		if ((parts & PART_GUI) > 0)
		{
			if (_filePrefix == null)
				_filePrefix = _template._filePrefix;

			// Clone datas
			if (_template._datas != null)
				_datas = new Hashtable<Object, LyData>(_template._datas);
		}
		if (((parts & PART_GUI) > 0)
				|| ((parts & PART_LABEL) > 0))
		{
			if (_name == null)
				_name = _template._name;
			if (_alias == null)
				_alias = _template._alias;
			if (_group == null)
				_group = _template._group;
			if (_tip == null)
				_tip = _template._tip;
		}

		_initParts |= parts;
	} // end method initFromTemplate

	/**
	 * Default constructor with no parameters.
	 */
	protected LyObjectInfo()
	{
		this(null, true);
	}

	/**
	 * Constructor of an object info with an id.
	 *
	 * @param id Identifier of the object (can't be null).
	 */
	protected LyObjectInfo(String id)
	{
		this(id, true);
	}

	/**
	 * Constructor of an external object info.
	 *
	 * @param id     Identifier of the object info.
	 * @param loaded Boolean that indicates if this object info is being loaded.
	 */
	LyObjectInfo(String id, boolean loaded)
	{
		super(id, loaded);
	}

	/**
	 * Constructor of an object info with an id and an optional template.
	 *
	 * @param     id       Identifier of the object (can't be null).
	 * @param     template Template from which this new instance is initialized.
	 * @exception IllegalArgumentException Thrown if the template is not null and not of the same
	 *                                     class
	 */
	protected LyObjectInfo(String id, LyObjectInfo template)
	{
		this(id, false);
		construct(template);
	}

	/**
	 * Construction of an object info with an id and an optional template.
	 *
	 * @param     id       Identifier of the object (can't be null).
	 * @param     template Template from which this new instance is initialized.
	 * @exception IllegalArgumentException Thrown if the template is not null and not of the same
	 *                                     class
	 */
	protected void construct(LyObjectInfo template)
	{
		super.construct();

		if (template != _template)
		{
			if (_template != null)
				throw new IllegalStateException("BF AppEngine : template was already set for object "
					+ _id);

			// This instance and its template must be of the same class
			if ((template != null)
					&& (template.getClass() != getClass())
					&& !template.getClass().isAssignableFrom(getClass()))
				throw new IllegalArgumentException("BF AppEngine : the constructed object " + _id
					+ " and its template " + template + " must be of the same class");

			// Sets template
			_template = template;
		}
	}

	/**
	 * Final step at the end of the parsing. Checks consistency of the information contained in this
	 * instance. This method is called after the object is constructed and initialized.
	 *
	 * @param     parent Parent of this object info for this check : this means context of use of
	 *                   this instance (classInfo which declares the field, project that contains
	 *                   the classInfo...).
	 * @exception SAXException Exceptions can be thrown from this method is inconsistency are found
	 *                         in this last step of parsing.
	 */
	void check(LyObjectInfo parent)
	{
		check(parent, PART_ALL);
	}

	/**
	 * Final step at the end of the parsing. Checks consistency of the information contained in this
	 * instance. This method is called after the object is constructed and initialized.
	 *
	 * @param     parent Parent of this object info for this check : this means context of use of
	 *                   this instance (classInfo which declares the field, project that contains
	 *                   the classInfo...).
	 * @param     parts  Indicates which parts are checked.
	 * @exception SAXException Exceptions can be thrown from this method is inconsistency are found
	 *                         in this last step of parsing.
	 */
	void check(LyObjectInfo parent, int parts)
	{
	}

    /**
     * Converts this objectInfo to a string.
     *
     * @return Id of the object concatenated with its name enclosed in parenthesis.
     */
    public String toName()
    {
        String name = getName();

		return ((name == null) ? getId() : (getId() + " (" + getName() + ")"));
	}

	/**
	 * Check if a mark is set on this instance. Generic mark are declared in classes that inherit
	 * from this class. Specific marks can be found from their name by using getMarkIndex methods
	 * declared in these classes.
	 *
	 * @param  mark Mark tested for this object.
	 * @return True if the mark is set, false otherwise.
	 */
	public final boolean hasMark(int mark)
	{
		return ((_marks != null)
		        && (mark >= 0)
		        && _marks.get(mark));
	}

	/**
	 * Check if a group of marks are set on this instance.
	 *
	 * @param  testMarks Tested marks.
	 * @param  matchAll  Specifies if this instance have to match all specified marks or only one.
	 * @return True if the test succeeds, false otherwise.
	 */
	public final boolean hasMarks(BitSet testMarks, boolean matchAll)
	{
		return ((_marks != null)
				&& (testMarks != null)
				&& LyBitSet.match(_marks, testMarks, matchAll));
	}

	/**
	 * Adds user data on this instance. A user data is a value associated to a name stored in a
	 * hashtable.
	 *
	 * @param     name  Name of the user data. If a value was already stored for that name, this
	 *                  value is overridden.
	 * @param     value Value of the user data.
	 * @return    The previous value associated to the given name if there was one, null otherwise.
	 * @exception IllegalArgumentException If the given name or value is null.
	 */
	public final Object putData(Object name, Object value)
	{
		return putData(name, value, LyTag.Model.FULL);
	}

	/**
	 * Adds user data on this instance. A user data is a value associated to a name stored in a
	 * hashtable.
	 *
	 * @param     name      Name of the user data. If a value was already stored for that name, this
	 *                      value is overridden.
	 * @param     value     Value of the user data.
	 * @param     modelType Model type where the data is declared.
	 * @return    The previous value associated to the given name if there was one, null otherwise.
	 * @exception IllegalArgumentException If the given name or value is null.
	 */
	public final Object putData(Object name, Object value, LyTag.Model modelType)
	{
		if (name == null)
			throw new IllegalArgumentException("BF AppEngine : name parameter can't be null");
		if (value == null)
			throw new IllegalArgumentException("BF AppEngine : value parameter can't be null");

		if (_datas == null)
			// Creates hashtable if null
			_datas = new Hashtable<Object, LyData>(1);
		else
		{
			// Check if data already exists
			Object data = getData(name);

			if (data != null)
				getEnvironment().logInfo("BF AppEngine : a data with name \"" + name
					+ "\" was already registered with value \"" + data
					+ "\" on object info " + this
					+ ", replacing the value with the new one \"" + value + "\".");
		}

		// Associates value to the name
		LyData data = new LyData(name, value, modelType);

		data.setParent(this);
		return _datas.put(name, data);
	}

	/**
	 * Removes a user data on this instance.
	 *
	 * @param     name Name of the data to be removed.
	 * @return    The value that was associated to the given name if there was one, null otherwise.
	 * @exception IllegalArgumentException If the given name is null.
	 * @exception IllegalStateException    If this method is called before any data has been added
	 *                                     (data hashtable is null).
	 */
	public final Object removeData(Object name)
	{
		if (name == null)
			throw new IllegalArgumentException("BF AppEngine : name parameter can't be null");
		if (_datas == null)
			throw new IllegalStateException("BF AppEngine : no data associated to this object");

		// Removes data associated to that name
		return _datas.remove(name);
	}

	/**
	 * Implementation of the Cloneable interface. Overridden to clone the property table properly
	 */
	public synchronized Object clone()
	{
		LyObjectInfo clone = (LyObjectInfo)super.clone();

		// Duplicates the data table if not null.
		if (_datas != null)
			clone._datas = new Hashtable<Object, LyData>(_datas);

		return clone;
	}

	/**
	 * Accessor for name of this instance.
	 *
	 * @return Name of this instance. If the name is null, this method returns the id.
	 */
	public final String getName()
	{
		if (_name == null)
			return getId();

		return _name;
	}

	/**
	 * Accessor for name of this instance.
	 *
	 * @return Name of this instance. If the name is null, this method returns null.
	 */
	public final String getInternalName()
	{
		return _name;
	}

	/**
	 * Accessor for alias (shortname) of this instance.
	 *
	 * @return Alias of this instance. If the alias is null, this method returns the name.
	 */
	public final String getAlias()
	{
		if (_alias == null)
			return getName();

		return _alias;
	}

	/**
	 * Accessor for alias (shortname) of this instance. This method doesn't compute a value if it is
	 * null
	 *
	 * @return Alias of this instance.
	 */
	public final String getInternalAlias()
	{
		return _alias;
	}

	/**
	 * Accessor to the file prefix for this instance.
	 *
	 * @return File prefix of the instance. If the file prefix is null, this method returns the id.
	 */
	public final String getFilePrefix()
	{
		if ((_filePrefix == null)
		        || (_filePrefix == NULL))
			return _id;

		return _filePrefix;
	}

	/**
	 * Accessor to the file prefix for this instance.
	 *
	 * @return File prefix of the instance. If the file prefix is null, this method returns null.
	 */
	public final String getInternalFilePrefix()
	{
		return _filePrefix;
	}

	/**
	 * Returns the template used to initialize this instance.
	 *
	 * @return The template is one was used when creating this instance, null otherwise.
	 */
	public final LyObjectInfo getTemplate()
	{
		return _template;
	}

	/**
	 * Returns name of the group to which this instance belongs. This group is simply a string used
	 * to gather instances logically in the same named panel in some views.
	 *
	 * @return Name of the group, or null if no group was specified.
	 */
	public final String getGroup()
	{
		return ((_group == null) || (_group == NULL) ? null : _group);
	}

	/**
	 * Accessor to the marks attached to this instance (presented as a bitset).
	 *
	 * @return A copy of the bitset which represents the marks (properties) of this instance.
	 * @see    java.util.BitSet
	 */
	protected BitSet getMarks(int parts)
	{
		return getMarks();
	}

	/**
	 * Accessor to the marks attached to this instance (presented as a bitset).
	 *
	 * @return A copy of the bitset which represents the marks (properties) of this instance.
	 * @see    java.util.BitSet
	 */
	public final BitSet getMarks()
	{
		// Return a copy of our bitset to avoid modifications
		return ((_marks == null) ? null : (BitSet)_marks.clone());
	}

	/**
	 * Returns the tip (short help displayed in tool tip) associated to this instance.
	 *
	 * @return Tip of the instance, or null if no tip was defined.
	 */
	public String getTip()
	{
		return _tip;
	}

	/**
	 * Gets a user data by its name.
	 *
	 * @param     name Name for which a value is searched.
	 * @return    Found value if any, or null if no data was stored with the given name.
	 * @exception IllegalArgumentException If the given name is null.
	 */
	public final Object getData(Object name)
	{
		if (name == null)
			throw new IllegalArgumentException("BF AppEngine : name parameter can't be null");
		if (_datas == null)
			return null;

		// Gets data associated to that name
		LyData data = _datas.get(name);

		return ((data == null) ? null : data.getValue());
	}

	/**
     * Gets list of data associated to this instance.
     *
     * @return Returns the list of all the data associated to this object info.
     */
    public final Vector<LyData> getDataList()
    {
        if (_datas == null)
            return null;

        Enumeration<LyData>     datas = _datas.elements();
        Vector<LyData>          objectDatas = new Vector<LyData>();
        LyData                  data;

        while (datas.hasMoreElements())
        {
            data = datas.nextElement();
            objectDatas.add(data);
        }

        return objectDatas;
    }

	/**
	 * Gets table of data associated to this instance.
	 *
	 * @return Returns the table of all the data associated to this object info. This is a copy of
	 *         the internal hashtable that contains these data.
	 */
	public final Hashtable<Object, LyData> getDatas()
	{
		return (_datas == null ? null : new Hashtable<Object, LyData>(_datas));
	}

	/**
	 * Gets user data associated to this class info.
	 *
	 * @return User data (may be null).
	 */
	public Object getUserData()
	{
		return _userData;
	}

	/**
	 * Checks recursively if this object has the given object as template.
	 *
	 * @return True if success, false otherwise.
	 */
	public boolean hasTemplate(LyObjectInfo objectInfo)
	{
		if (_template == null)
			return false;
		if (_template == objectInfo)
			return true;
		return _template.hasTemplate(objectInfo);
	}

	/**
	 * Checks if this object info was used as template.
	 */
	public boolean isTemplate()
	{
	    LyInfoList<? extends LyObjectInfo>	objectInfos = _application.getInfoList(getClass());
	    LyObjectInfo objectInfo;
	    int          i, n = objectInfos.getSize();

	    for (i = 0; i < n; i++)
	    {
	        objectInfo = (LyObjectInfo)objectInfos.getInfo(i);
	        if (objectInfo.getTemplate() == this)
	            return true;
	    }

	    return false;
	}

	/**
	 * The following modifiers are not public to avoid modification after parsing. They are provided
	 * to be used only by the data model parser. Visibility shouldn't be changed except for a very
	 * good reason.
	 */

    /**
     * Sets template of this object info.
     *
     * @param     template Template of this object info.
     * @exception IllegalArgumentException Thrown if the template is not null and not of the same
     *                                     class
     */
    protected void setTemplate(LyObjectInfo template)
    {
        if (template == _template)
            return;

        if ((_template != null)
        		&& (_application != null)
        		&& !_application.isEdited())
            throw new IllegalStateException("BF AppEngine : template was already set for object " + _id);

        // This instance and its template must be of the same class
        if ((template != null)
                && (template.getClass() != getClass())
                && !template.getClass().isAssignableFrom(getClass()))
            throw new IllegalArgumentException("BF AppEngine : the constructed object " + _id
                + " and its template " + template + " must be of the same class");

        // Sets template
        _template = template;
    }

	/**
	 * Sets the name of the instance.
	 *
	 * @param name Name of the instance.
	 */
	void setName(String name)
	{
		if (isExtern()
				&& (_name != null)
				&& ((_template == null) || (_name != _template._name)))
			return; // may happen on extern infos

		_name = ((name == null) ? null : name.intern());
	}

	/**
	 * Sets the alias of the instance.
	 *
	 * @param alias Alias of the instance.
	 */
	final void setAlias(String alias)
	{
		if (isExtern()
				&& (_alias != null)
				&& ((_template == null) || (_alias != _template._alias)))
			return; // may happen on extern infos

		setInternalAlias(alias);
	}

	/**
     * Sets the alias of the instance.
     *
     * @param alias Alias of the instance.
     */
    final void setInternalAlias(String alias)
    {
        _alias = ((alias == null) ? null : alias.intern());
    }

	/**
	 * Sets the group name to which this instance belongs.
	 *
	 * @param group Group name.
	 */
	void setGroup(String group)
	{
		if (isExtern()
				&& (_group != null)
				&& ((_template == null) || (_group != _template._group)))
			return; // may happen on extern infos

		_group = ((group == null) ? null : group.intern());
	}

	/**
	 * Sets the file prefix corresponding to this instance.
	 *
	 * @param prefix File prefix for this instance.
	 */
	final void setFilePrefix(String prefix)
	{
		if (isExtern()
				&& (_filePrefix != null)
				&& ((_template == null) || (_filePrefix != _template._filePrefix)))
			return; // may happen on extern infos

		_filePrefix = ((prefix == null) ? null : prefix.intern());
	}

	/**
     * Sets the file prefix corresponding to this instance.
     *
     * @param prefix File prefix for this instance.
     */
	final void setInternalFilePrefix(String prefix)
	{
	    setFilePrefix(prefix);
	}

	/**
	 * Sets the tip on this instance.
	 *
	 * @param tip Tip associated to this instance.
	 */
	final void setTip(String tip)
	{
		_tip = ((tip == null) ? null : tip.intern());
	}

	/**
	 * Sets marks on this instance.
	 *
	 * @param marks Bitset that represent the marks attached to this instance. Marks must not be
	 *              modified after attributed to this instance since it is used directly without
	 *              copying them.
	 */
	final void setMarks(BitSet marks)
	{
		_marks = marks;
	}

	/**
	 * Sets user data for this class info.
	 *
	 * @param userData User data (may be null).
	 */
	void setUserData(Object userData)
	{
		_userData = userData;
	}

	/**
     * Sets list of data associated to this instance.
     */
    void setDataList(Vector<LyData> datas)
    {
        if ((datas == null)
                || datas.isEmpty())
            _datas = null;
        else
        {
            int     i, n = datas.size();
            LyData  data;

            _datas = new Hashtable<Object, LyData>(n);

            for (i = 0; i < n; i++)
            {
                data = datas.elementAt(i);
                _datas.put(data.getName(), data);
            }
        }
    }

    /**
     * Clone this object and make sets clone as template of this object info.
     */
    protected LyObjectInfo cloneAsTemplate()
    {
    	if (!isIncluded())
    		throw new IllegalStateException("Application Engine : can't override internal info " + this);
    	if (!isLoaded())
    		load();

    	if (_application.isEdited())
    	{
    		// Only keep template when under edition, otherwise we just override existing infos
	    	_template = (LyObjectInfo)cloneTag();
	    	_initParts = PART_ALL; // Mark as initialized

	    	// Reparent existing tags to template
	    	ArrayList<LyTag>	tags =_application.getTags(LyTag.class);
	    	LyTag				tag, tagTemplate;
	    	int					i, n = (tags == null ? 0 : tags.size());

	    	for (i = 0; i < n; i++)
	    	{
	    		tag = tags.get(i);
	    		if (tag.getParent() == this)
	    			tag._parent = _template;

	    		if (tag instanceof LyObjectInfo)
	    		{
	    			tagTemplate = ((LyObjectInfo)tag).getTemplate();
	    			if ((tagTemplate != null)
	    					&& (tagTemplate.getParent() == this))
	    				tagTemplate._parent = _template;
	    		}
	    	}

			if ((_parent instanceof LyClassInfo)
				&& (this instanceof LyFieldInfo))
			{
				LyClassInfo	classInfo = (LyClassInfo)_parent;
				LyFieldInfo	fieldInfo = (LyFieldInfo)this;
				int			index = classInfo.getIndex(fieldInfo);

				classInfo.removeFieldInfo(fieldInfo);
				classInfo.addFieldInfo((LyFieldInfo)_template, index);
			}
    	}

        _parent = null;
        return this;
    }
} // end class LyObjectInfo
