/**
 * LyPath.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import leon.app.*;
import leon.data.*;
import leon.misc.*;


/**
 * This class defines a path.
 */
public class LyPath extends LyRoute
{
	public static final String __VERSION = "$Id: LyPath.java 33186 2011-11-17 11:02:46Z jlsaouli $";

	/** Fieldinfo target of the path (may be null). */
	LyFieldInfo				   _targetFieldInfo;

	/** ClassInfo source of the path (may be null). */
	LyClassInfo				   _sourceClassInfo;

	/** Label to use on target field info. */
	LyInfoLabel				   _fieldLabel;

	/* Constructor */

    /**
     * Constructor of a path.
     *
     * @param id         Id of the path.
     */
    protected LyPath(String id)
    {
        this(id, null, null, null, null);
    }
    
	/**
	 * Constructor of a path.
	 *
	 * @param id         Id of the path.
	 * @param target     Target classinfo of the path.
	 * @param classInfo  Source classinfo of the path.
	 * @param fieldinfo  Target fieldinfo of the path.
	 * @param fieldLabel Label to use on target field.
	 */
	public LyPath(String id, LyClassInfo target, LyClassInfo classInfo,
		LyFieldInfo fieldInfo)
	{
		this(id, target, classInfo, fieldInfo, null);
	}

	/**
	 * Constructor of a path.
	 *
	 * @param id         Id of the path.
	 * @param target     Target classinfo of the path.
	 * @param classInfo  Source classinfo of the path.
	 * @param fieldinfo  Target fieldinfo of the path.
	 * @param fieldLabel Label to use on target field.
	 */
	public LyPath(String id, LyClassInfo target, LyClassInfo classInfo,
		LyFieldInfo fieldInfo, LyInfoLabel fieldLabel)
	{
		super(id, target, null);
		_sourceClassInfo = classInfo;
		_targetFieldInfo = fieldInfo;
		_fieldLabel = fieldLabel;
	}

	/**
	 * Check consistency of this path.
	 */
	void check(LyClassInfo source)
	{
		if (_sourceClassInfo != null)
			source = _sourceClassInfo;

		if ((_target == null) && (_steps != null))
		{
			int    n = _steps.size();
			LyStep step = _steps.elementAt(n - 1);

			_target = step.getTarget();
		}

		if ((_target == null) && (source != null))
			_target = source;

		if (source != null)
			super.check(source);

		if (_target == null)
			throw new IllegalArgumentException("BF AppEngine : target must be declared for path "
				+ this);

		if ((_targetFieldInfo != null)
				&& !_target.contains(_targetFieldInfo))
			throw new IllegalArgumentException("BF AppEngine : target fieldInfo "
				+ _targetFieldInfo + " must be declared on target classInfo "
				+ _target + " on path " + this);

		if ((_targetFieldInfo != null)
				&& (_fieldLabel != null)
				&& !_targetFieldInfo.hasLabel(_fieldLabel))
			throw new IllegalArgumentException("BF AppEngine : invalid label "
				+ _fieldLabel + ", you can only use a label declared on target field "
				+ _targetFieldInfo);
	} // end method check

	/**
	 * Check consistency of this path.
	 */
	public void check()
	{
		check(null);
	}

	/**
	 * Gets target fieldinfo of the path.
	 */
	public final LyFieldInfo getTargetFieldInfo()
	{
		return _targetFieldInfo;
	}

	/**
	 * Gets source classInfo of the path.
	 */
	public final LyClassInfo getSourceClassInfo()
	{
		return _sourceClassInfo;
	}

	/**
	 * Gets label to use on target field.
	 */
	public final LyInfoLabel getFieldLabel()
	{
		return _fieldLabel;
	}

	/**
	 * Gets string corresponding to this path for the given object.
	 */
	public String getString(LyObject object, LyUnit unit)
	{
		LySimpleObjectList list = new LySimpleObjectList(object);

		return getString(list, unit);
	}

	/**
	 * Gets string corresponding to this path for the given class info.
	 */
	public String getString(LyClassInfo classInfo, LyUnit unit)
	{
		LyDataSession session = _application.getCurrentSession();
		LyObjectList  list = new LyClassWorkSpace(session, classInfo, false);

		return getString(list, unit);
	}

	/**
	 * Gets string corresponding to this path for the given objects list.
	 */
	String getString(LyObjectList list, LyUnit unit)
	{
		if ((list == null) || (list.getSize() == 0))
			return null;

		LyDataSession session = _application.getCurrentSession();
		LyObjectList  targets = null;

		if ((_steps == null) || (_steps.size() == 0))
			targets = list.getList(session, _target);
		else
		{
			LyClassInfo classInfo = _sourceClassInfo;

			if (classInfo == null)
			{
				classInfo = list.getClasses().getInfo(0);
				if (classInfo == _target)
					targets = list.getList(session, classInfo);
			}

			if (targets == null)
				targets = list.getList(session, classInfo, this, false);
		}

		int			  i, n = ((targets == null) ? 0 : targets.getSize());
		LyObject	  target;
		Object		  value;
		StringBuffer  buffer = new StringBuffer();
		String		  name;
		LyEnvironment environment = getEnvironment();

		for (i = 0; i < n; i++)
		{
			if (i > 0)
				buffer.append(' ');

			target = targets.getObject(i);

			if (_fieldLabel != null)
				buffer.append(_fieldLabel.getString(target));
			else if (_targetFieldInfo != null)
			{
				value = target.getValue(_targetFieldInfo);
				if (value != null)
				{
					value = _targetFieldInfo.convertFromFieldUnit(value, unit);
					name = _targetFieldInfo.toName(value);

					if ((_targetFieldInfo instanceof LyChoiceFieldInfo)
							|| ((_targetFieldInfo instanceof LyNumberFieldInfo)
									&& ((LyNumberFieldInfo)_targetFieldInfo).getFieldUnit() != null))
						name = environment.translate(name);
					
					buffer.append(name);
				}
			}
			else
				// Default is to use the name of the target object
				buffer.append(target.getName());
		}

		return buffer.toString();
	} // end method getString

	/**
	 * Sets the source classinfo of the path.
	 *
	 * @param classInfo the source classinfo to set.
	 */
	void setSourceClassInfo(LyClassInfo classInfo)
	{
		_sourceClassInfo = classInfo;
	}

	/**
	 * Sets the target fieldinfo of the path.
	 *
	 * @param fieldinfo the target fieldinfo to set
	 */
	void setTargetFieldInfo(LyFieldInfo fieldInfo)
	{
		_targetFieldInfo = fieldInfo;
	}

	/**
	 * Sets label to use on target field.
	 *
	 * @param fieldLabel label used on target field of this path.
	 */
	void setFieldLabel(LyInfoLabel fieldLabel)
	{
		_fieldLabel = fieldLabel;
	}

	/**
	 * Adds a step to this path.
	 *
	 * @param     target   Target class of the added step.
	 * @param     relation List of relations for the added step.
	 * @param     filter   Filter of the added step (can be null)
	 * @param     reverse  Boolean indicating the direction of the added step.
	 * @exception IllegalArgumentException If target parameter is null.
	 * @exception IllegalArgumentException If relations parameter is null or if the list is empty.
	 */
	public void addStep(LyClassInfo target, LyInfoList<LyFieldInfo> relations, LyFilter filter,
		boolean reverse)
	{
		LyStep step = new LyStep(target, relations, filter, reverse);

		addStep(step);
	}
} // end class LyPath
