/*
 * LyPreference.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.io.*;

import leon.misc.*;
import leon.info.infointerface.*;


/**
 * This class represents a preference. It contains the following attributes : - the action on which
 * the preference is performed, - the filter used by the preference, - the sort used by the
 * preference, - a name that may be displayed, - the fielInfo corresponding to the fixed column, -
 * the owner of this preference.
 *
 * @see LyObjectInfo
 * @see LyFilter
 * @see LySort
 */
public final class LyPreference extends LyObjectInfo
    implements LyExtraInfoInterface
{
	public static final String __VERSION =
		"$Id: LyPreference.java 40742 2012-09-11 11:29:29Z jlsaouli $";

	public enum ShowAs {
		TABLE,
		LIST
	}

	/** Object info concerned by the preference. */
    String                     _objectInfoId;

	/** Filter used by the preference. */
	LyFilter				   _filter;

	/** Sort used by the preference. */
	LySort					   _sort;

	/** The field info corresponding to the fixed column of the preference. */
	LyFieldInfo				   _fixedColumn;

	/** Number of displayed objects. */
	int						   _nbObjects;

	/** List of fields of the preference. */
	LyInfoList<LyFieldInfo>    _fieldInfos;

	/** Id of the user that owns the preference. */
	String					   _user;

	/** Boolean indicating if the preference is the default one. */
	boolean					   _default;

	/** Boolean indicating if the preference is the default one for mobile view. */
	boolean					   _defaultForMobile;

	/** Boolean indicating if the preference is for portlet only. */
	boolean					   _forPortlet;

	/** Boolean indicating if the preference can be deleted or not. */
	boolean					   _readOnly;

	/** String indicating the last file name used in file dialog (only for import action) */
	String                     _lastFileName;

	/** String indicating the size of the columns. */
	String					   _columnSizes;

	/** Display mode. */
	ShowAs					   _showAs;

	/**
     * Constructor for a preference.
     *
     * @param id          the identifier of the preference.
     * @param classInfo   the concerned object info.
     * @param name        the name of the preference.
     * @param fixedColumn field info for which its column is fixed for the preference.
     * @param nbObjects   number of displayed objects.
     * @param user        the id of the user that owns the preference.
     * @param defaultPref indicates if this preference is the default one.
     */
    public LyPreference(String id, LyObjectInfo objectInfo, String name,
        LyFieldInfo fixedColumn, int nbObjects, String user, boolean defaultPref)
    {
        this(id, (String)null, name, fixedColumn, nbObjects, user, defaultPref, null);

        setObjectInfo(objectInfo);
    }

	/**
	 * Constructor for a preference.
	 *
	 * @param id          the identifier of the preference.
	 * @param classInfo   the concerned object info.
	 * @param name        the name of the preference.
	 * @param fixedColumn field info for which its column is fixed for the preference.
	 * @param nbObjects   number of displayed objects.
	 * @param user        the id of the user that owns the preference.
	 * @param defaultPref indicates if this preference is the default one.
	 */
	protected LyPreference(String id, String objectInfoId, String name,
		LyFieldInfo fixedColumn, int nbObjects, String user, boolean defaultPref,
		String columnSizes)
	{
		super(id);

		_objectInfoId = objectInfoId;
		_filter = null;
		_sort = null;
		_fixedColumn = fixedColumn;
		_nbObjects = nbObjects;
		_user = user;
		_default = defaultPref;
		_defaultForMobile = false;
		_forPortlet = false;
		_columnSizes = columnSizes;
		_lastFileName = null;

		setName(name);
	}

	/**
	 * Clone a preference.
	 */
	public LyPreference clonePreference()
	{
		return clonePreference(_id + LyString.getRandomId());
	}

	/**
	 * Clone a preference.
	 *
	 * @param id identifier of the cloned preference.
	 */
	public LyPreference clonePreference(String id)
	{
		LyPreference pref = new LyPreference(id, _objectInfoId, _name,
				_fixedColumn, _nbObjects, _user, _default, _columnSizes);

		pref._defaultForMobile = _defaultForMobile;
		pref._forPortlet = _forPortlet;

		if (_fieldInfos != null)
			pref.setFieldInfos(_fieldInfos); // list cloned by set method

		if (_filter != null)
			pref.setFilter(_filter.cloneFilter());

		if (_sort != null)
			pref.setSort(_sort.cloneSort());

		return pref;
	}

    /**
	 * Method for freeing extra info (and sub-infos).
	 */
    public void free()
    {
        if (_filter != null)
            _filter.free();
        if (_sort != null)
            _sort.free();
        super.free();
    }

    /**
     * Method for saving preference file.
     */
    public boolean saveFile(LyXmlWriterInterface writer)
    {
        String      fileName;
        
        if (writer == null)
        	fileName = getFile();
        else
        	fileName = writer.getFile(this);

        if (fileName == null)
        {
            String   dir, userId = getUser();

            if (userId == null)
            {
                userId = "";
                dir = "LY_PREFERENCE_DIR";
            }
            else
            {
                userId = userId.substring(userId.indexOf(leon.data.LyObjectId.CLASS_SEP) + 1) + "_";
                dir = "LY_USER_PREFERENCE_DIR";
            }

            // retrieve the name of the data directory
            LyEnvironment environment = getEnvironment();
            String        dirPath = environment.getFile(dir);
            File          fileDirectory = null;

            if (dirPath != null)
                fileDirectory = new File(dirPath);

            // Try to create the directory if needed
            if ((fileDirectory != null)
                && !fileDirectory.exists())
            {
                try
                {
                    environment.logInfo("Creating directory :" + dirPath);
                    fileDirectory.mkdirs();
                }
                catch (SecurityException e)
                {
                    environment.logError("Unable to create directory :" + dirPath);
                    return false;
                }
            }

            File    file;
            int     i = 0;

            // find a file name which has not been affected
            do
            {
                fileName = dirPath + "/preference_" + userId + i + ".xml";
                file = new File(fileName);
                i++;
            }
            while (file.exists());

            setFile(fileName);
        }

        // Create xml file for the preference
        try
        {
            FileOutputStream 	output = new FileOutputStream(fileName);
            LyXmlPrintWriter    pw = new LyXmlPrintWriter(output);

            if (writer == null)
            	writeXml(pw, true, true);
            else
            	writer.writeXml(pw, this);
            
            pw.close();
            return true;
        }
        catch (Exception e)
        {
            LyEnvironment   environment = getEnvironment();

            environment.writeException(e);
            return false;
        }
    }

    /**
     * Method for deleting preference file.
     */
    public boolean deleteFile()
    {
        String  filePath = getFile();

        if (filePath == null)
            return false;

        try
        {
            // Delete the file
            File    file = new File(filePath);

            return file.delete();
        }
        catch (SecurityException e)
        {
            LyEnvironment   environment = getEnvironment();

            environment.writeException(e);
            return false;
        }
    }

    /**
     * Method to rewrite XML corresponding to this component.
     *
     * @param pw The printWriter on which the XML is written.
     * @param header Boolean indicating if header must be written.
     * @param header Boolean indicating if body must be written.
     */
    public void writeXml(PrintWriter pw, boolean header, boolean body)
    {
        if (header)
        {
            pw.println("<?xml version='1.0' encoding='" + LyString.DEFAULT_CHARSET + "'?>");
            pw.println("<!DOCTYPE PREFERENCE SYSTEM 'leon/conf/dtd/leon_gui.dtd'>");
        }

        if (body)
        	writeXml(pw);
    }

	/**
	 * Writes XML corresponding to this filter element.
	 *
	 * @param pw PrintWriter on which the XML is written.
	 */
	public void writeXml(PrintWriter pw)
	{
		// Write tag with its attributes
		pw.print("<PREFERENCE");
		pw.print(" id='" + LyString.getXmlString(getId()) + "'");
		pw.print(" idref='" + LyString.getXmlString(_objectInfoId) + "'");
		pw.print(" name='" + LyString.getXmlString(_name) + "'");
		if (_fixedColumn != null)
			pw.print(" fixedColumn='" + LyString.getXmlString(_fixedColumn.getId()) + "'");
		if (_nbObjects > 0)
			pw.print(" nbObjects='" + _nbObjects + "'");
		if (_user != null)
			pw.print(" user='" + LyString.getXmlString(_user) + "'");
		if (_columnSizes != null)
			pw.print(" columnSizes='" + LyString.getXmlString(_columnSizes) + "'");
		if (_default)
			pw.print(" default='true'");
		if (_defaultForMobile)
			pw.print(" defaultForMobile='true'");
		if (_forPortlet)
			pw.print(" forPortlet='true'");
		if( _lastFileName != null)
			pw.print(" lastFileName='" + LyString.getXmlString(_lastFileName) + "'");
		if (_showAs != null)
			pw.print(" showAs='" + _showAs + "'");
		pw.println(">");

		if (_filter != null)
			_filter.writeXml(pw);

		if (_sort != null)
			_sort.writeXml(pw);

		LyInfo info;
		int    i, n;

		if (_fieldInfos != null)
		{
			for (i = 0, n = _fieldInfos.getSize(); i < n; i++)
			{
				info = _fieldInfos.getInfo(i);
				pw.print("<FIELD_REF idref='" + LyString.getXmlString(info.getId()) + "'/>\n");
			}
		}

		// Write external dependencies
		LyInfoList<LyFieldInfo> externs = new LyInfoList<LyFieldInfo>(false);

		writeExternXml(pw, _objectInfoId);

		if (_fixedColumn != null)
			externs.addInfo(_fixedColumn);

		if (_filter != null)
			externs.addList(_filter.getFieldInfos());

		if (_sort != null)
			externs.addList(_sort.getFieldInfos());

		if (_fieldInfos != null)
			externs.addList(_fieldInfos);

		for (i = 0, n = externs.getSize(); i < n; i++)
		{
			info = externs.getInfo(i);
			writeExternXml(pw, info.getId());
		}


		pw.println("</PREFERENCE>");
	} // end method writeXml

	/**
	 * Writes XML for an external idref.
	 *
	 * @param pw PrintWriter on which the XML is written.
	 * @param id id that must be written.
	 */
	protected void writeExternXml(PrintWriter pw, String id)
	{
		pw.println("<EXTERN id='" + LyString.getXmlString(id) + "'/>");
	}

	/**
     * Get object info referred by this preference.
     *
     * @return The concerned object info id (action or classInfo).
     */
    public LyObjectInfo getObjectInfo()
    {
        return (LyObjectInfo)_application.getInfo(_objectInfoId);
    }

	/**
	 * Get object info referred by this preference.
	 *
	 * @return The concerned object info id (action or classInfo).
	 */
	public String getObjectInfoId()
	{
		return _objectInfoId;
	}

	/**
	 * Get filter used by this preference.
	 *
	 * @return The concerned filter.
	 */
	public LyFilter getFilter()
	{
		return _filter;
	}

	/**
	 * Get sort used by this preference.
	 *
	 * @return The concerned sort.
	 */
	public LySort getSort()
	{
		return _sort;
	}

	/**
	 * Get the field info for which its column is fixed in this preference.
	 *
	 * @return Field info for which its column is fixed in this preference.
	 */
	public LyFieldInfo getFixedColumn()
	{
		return _fixedColumn;
	}

	/**
	 * Get the number of displayed objects.
	 *
	 * @return number of displayed objects, 0 if not set.
	 */
	public int getNbObjects()
	{
		return _nbObjects;
	}

	/**
	 * Checks if this preference is the default one or not.
	 *
	 * @return True if default preference, false otherwise.
	 */
	public boolean isDefault()
	{
		return _default;
	}

	/**
	 * Checks if this preference is the default one or not for mobile view.
	 *
	 * @return True if default preference for mobile, false otherwise.
	 */
	public boolean isDefaultForMobile()
	{
		return _defaultForMobile;
	}

	/**
	 * Checks if this preference is for portlet only.
	 *
	 * @return True if preference for portlet, false otherwise.
	 */
	public boolean isForPortlet()
	{
		return _forPortlet;
	}

	/**
	 * Checks if this preference can be changed or not.
	 *
	 * @return True if this preference can't be changed, false otherwise.
	 */
	public boolean isReadOnly()
	{
		return _readOnly;
	}

	/**
	 * Gets list of field infos of this preference.
	 *
	 * @return List of field infos, may be null.
	 */
	public LyInfoList<LyFieldInfo> getFieldInfos()
	{
		if (_fieldInfos != null)
			_fieldInfos.lock();
		return _fieldInfos;
	}

	/**
	 * Get owner of this preference.
	 *
	 * @return The owner of this preference.
	 */
	public String getUser()
	{
		return _user;
	}

	/**
	 * Get size of columns of this preference.
	 *
	 * @return The sizes of columns of this preference.
	 */
	public String getColumnSizes()
	{
		return _columnSizes;
	}

	/**
	 * Get the display mode of this preference.
	 *
	 * @return Display mode chose for this preference.
	 */
	public ShowAs getShowAs()
	{
		return _showAs;
	}

	/**
     * Sets the object info referred by this preference.
     *
     * @param action Object info concerned by this preference.
     */
    public void setObjectInfo(LyObjectInfo objectInfo)
    {
        if (!(objectInfo instanceof LyAction)
                && !(objectInfo instanceof LyClassInfo))
            throw new IllegalStateException("BF AppEngine : invalid objectInfo for preference : "
                + objectInfo + ", objectInfo should be an action or a class info");

        _objectInfoId = objectInfo.getId();
    }

	/**
	 * set the field info for which its column is fixed in this preference.
	 *
	 * @param fixedColumn Field info for which its column is fixed in this preference.
	 */
	public void setFixedColumn(LyFieldInfo fixedColumn)
	{
		_fixedColumn = fixedColumn;
	}

	/**
	 * Set the number of displayed objects.
	 *
	 * @param nbObjects number of displayed objects, 0 if not set.
	 */
	public void setNbObjects(int nbObjects)
	{
		_nbObjects = nbObjects;
	}

	/**
	 * Sets whether this preference is the default one or not.
	 *
	 * @param defaultPref True if default preference, false otherwise.
	 */
	public void setDefault(boolean defaultPref)
	{
		_default = defaultPref;
	}

	/**
	 * Sets whether this preference is the default one or not for mobile view.
	 *
	 * @param defaultPref True if default preference for mobile, false otherwise.
	 */
	public void setDefaultForMobile(boolean defaultForMobile)
	{
		_defaultForMobile = defaultForMobile;
	}

	/**
	 * Sets whether this preference is for portlet only.
	 *
	 * @param forPortlet True if preference for portlet only, false otherwise.
	 */
	public void setForPortlet(boolean forPortlet)
	{
		_forPortlet = forPortlet;
	}

	/**
	 * Sets whether this preference can be changed or not.
	 *
	 * @param readOnly True if this preference can't be nor changed neither deleted, false
	 *                 otherwise.
	 */
	public void setReadOnly(boolean readOnly)
	{
		_readOnly = readOnly;
	}

	/**
	 * Sets the size of the columns for this preference.
	 *
	 * @param columnSizes the size of the columns for this preference.
	 */
	public void setColumnSizes(String columnSizes)
	{
		_columnSizes = columnSizes;
	}

	/**
	 * Get filter used by this preference.
	 *
	 * @return The concerned filter.
	 */
	public void setFilter(LyFilter filter)
	{
		_filter = filter;
		if (_filter.getParent() == null)
			_filter.setParent(this);
		if (getApplication() != null)
			_filter.setApplication(getApplication());
	}

	/**
	 * Get sort used by this preference.
	 *
	 * @return The concerned sort.
	 */
	public void setSort(LySort sort)
	{
		_sort = sort;
		if (_sort.getParent() == null)
			_sort.setParent(this);
		if (getApplication() != null)
			_sort.setApplication(getApplication());
	}

	/**
	 * Sets list of field infos of this preference.
	 *
	 * @param fieldInfos List of field infos, may be null.
	 */
	public void setFieldInfos(LyInfoList<LyFieldInfo> fieldInfos)
	{
		if (fieldInfos == null)
			_fieldInfos = null;
		else
			_fieldInfos = new LyInfoList<LyFieldInfo>(fieldInfos);
	}

	/**
	 * Adds a fieldInfo to this preference.
	 *
	 * @param fieldInfo FieldInfo to add in this preference.
	 */
	void addFieldInfo(LyFieldInfo fieldInfo)
	{
		if (_fieldInfos == null)
			_fieldInfos = new LyInfoList<LyFieldInfo>();
		_fieldInfos.addInfo(fieldInfo);
	}

	/**
	 * This method return the name of the last open file name (only for import).
	 *
	 * @return _lastFileName
	 */
	public String getLastFileName()
	{
		return _lastFileName;
	}

	/**
	 * This method set the name of the last open file name (only for import).
	 *
	 * @param fileName
	 */
	public void setLastFileName(String fileName)
	{
		_lastFileName = fileName;
	}

	/**
	 * Sets the display mode of this preference.
	 *
	 * @param showAs Display mode chose for this preference.
	 */
	public void setShowAs(ShowAs showAs)
	{
		_showAs = showAs;
	}

} // end class LyPreference
