/*
 * LyProject.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;
import leon.info.daemon.*;
import leon.info.location.*;
import leon.info.misc.*;
import leon.info.model.*;
import leon.info.notifier.*;
import leon.misc.*;


/**
 * This class is the description of a project. Project is the root object built from parsing the
 * data description file. A project is a simple set of objectInfo instances. It can contain :
 * sub-projects, classes, fields, actions, sorts, filters, locations.
 *
 * <p>A (root) action can be defined on a project, this action is executed on starting after
 * parsing.</p>
 *
 * <p>A (root) project may have an application behavior class and a session behavior class. These
 * classes are used to customize the behavior of an application and of a session. They are built by
 * reflection.</p>
 *
 * <p>Root action and behaviors are not used on sub-projects of the root project.</p>
 *
 * <p>See also tag <PROJECT> of leon.dtd</p>
 *
 * @see LyClassInfo
 * @see LyFieldInfo
 * @see LyAction
 * @see LyFilter
 * @see LySort
 * @see LyLocation
 * @see leon.app.LyApplication
 * @see leon.app.LyDataSession
 */
public final class LyProject extends LyObjectInfo
{
	public static final String __VERSION =
		"$Id: LyProject.java 34882 2012-03-20 16:51:09Z nhamon $";

	/** Version of this project. */
	String					   _version;

	/** List of all info contained in this project. */
	LyInfoList<LyInfo>		   _contents;

	/** Root action. */
	LyAction				   _rootAction;

	/** Application behavior class. */
	LyJavaName				   _applicationBehavior;

	/** Session behavior class. */
	LyJavaName				   _sessionBehavior;

	/** List of models of this project. */
	Vector<LyModel>			   _models;

	/**
	 * Initialisation of instance attributes.
	 */
	protected void init()
	{
		super.init();

		// Initialize project attributes.
		_version = null;
		_contents = new LyInfoList<LyInfo>(false);
		_rootAction = null;
		_models = null;
	}

	/**
	 * Init project attributes from its template.
	 */
	void initFromTemplate()
	{
		if ((PART_ALL & _initParts) > 0)
			return;
		if (_template == null)
			return;

		super.initFromTemplate();

		LyProject template = (LyProject)_template;

		if (_applicationBehavior == null)
			_applicationBehavior = template._applicationBehavior;
		if (_sessionBehavior == null)
			_sessionBehavior = template._sessionBehavior;

		if (template._contents != null)
			_contents = new LyInfoList<LyInfo>(template._contents);

		_rootAction = template._rootAction;
	}

    /**
     * Constructor of a project.
     *
     * @param id                           Identifier of the project.
     */
    LyProject(String id)
    {
        super(id);
    }
            
	/**
	 * Constructor of a project.
	 *
	 * @param id                           Identifier of the project.
	 * @param template                     Project template.
	 * @param version                      Version of the project.
	 * @param name                         Name of the project.
	 * @param tip                          Tip of the project.
	 * @param applicationBehaviorClassName Class name of the behavior of the application.
	 * @param sessionBehaviorClassName     Class name of the behavior of the session.
	 */
	LyProject(String id, LyProject template, String version,
		String applicationBehaviorClassName, String sessionBehaviorClassName)
	{
		super(id, template);

		// Sets version and behavior classes
		_version = version;
		setApplicationBehaviorClassName(applicationBehaviorClassName);
		setSessionBehaviorClassName(sessionBehaviorClassName);
	}

    /**
     * Returns extra tags.
     */
    protected LyTag[] getExtraTags()
    {
        return new LyTag[] { _applicationBehavior, _sessionBehavior };
    }
    
    /**
     * Checks project
     */
    public void check(LyObjectInfo parent)
    {
        check(parent, PART_ALL);
    }
     
	/**
	 * Lock this project.
	 */
	public void lock()
	{
		super.lock();

		// Lock all info lists to avoid modifications
		_contents.lock();
	}

	/**
	 * Gets application behavior class name.
	 *
	 * @return Java name of application behavior class (may be null).
	 */
	public final String getApplicationBehaviorClassName()
	{
		return (_applicationBehavior == null ? null : _applicationBehavior.getJavaName());
	}

	/**
	 * Returns application behavior class.
	 *
	 * @return Returns java class that is used to customize behavior of an application.
	 */
	public Class<?> getApplicationBehaviorClass()
	{
	    if ((_applicationBehavior == null)
	            || _application.isEdited())
	        return null;
	    
		Class<?> applicationBehaviorClass = _applicationBehavior.getJavaClass();

		if (applicationBehaviorClass != null)
		    return applicationBehaviorClass;
		
	    LyEnvironment     environment = getEnvironment();
	    
	    environment.logError("BF AppEngine : unknown application behavior class "
	        + _applicationBehavior + " on project " + _id);
		return null;
	}

	/**
	 * Gets session behavior class name.
	 *
	 * @return Java name of session behavior class (may be null).
	 */
	public final String getSessionBehaviorClassName()
	{
	    return (_sessionBehavior == null ? null : _sessionBehavior.getJavaName());
	}

	/**
	 * Return session behavior class.
	 *
	 * @return Returns java class that is used to customize behavior of a session.
	 */
	public Class<?> getSessionBehaviorClass()
	{
       if ((_sessionBehavior == null)
                || _application.isEdited())
            return null;
        
        Class<?> sessionBehaviorClass = _sessionBehavior.getJavaClass();

        if (sessionBehaviorClass != null)
            return sessionBehaviorClass;
        
        LyEnvironment     environment = getEnvironment();
        
        environment.logError("BF AppEngine : unknown application behavior class "
            + _applicationBehavior + " on project " + _id);
        return null;
	}

	/**
	 * Gets version of this project.
	 *
	 * @return Version.
	 */
	public String getVersion()
	{
		return _version;
	}

	/**
	 * Returns the sub-project at the given index.
	 *
	 * @param  index Index of the requested project.
	 * @return The project if it's found, null otherwise.
	 */
	public LyProject getProject(int index)
	{
		LyInfoList<LyProject> projects = getProjects();

		return projects.getInfo(index);
	}

	/**
	 * Returns the number of sub-projects in this project.
	 *
	 * @return Number of classInfos in this project.
	 */
	public int getProjectCount()
	{
		return getInfoCount(LyProject.class);
	}

	/**
	 * Gets projects list.
	 *
	 * @return List of sub-projects of this project.
	 */
	public LyInfoList<LyProject> getProjects()
	{
		return getInfos(LyProject.class);
	}
	
	/**
	 * Returns the project with the given identifier.
	 * 
	 * @param projectId the identifier of the project that must be retrieved
	 * @return the sub-project with the given identifier (may be the project itself)
	 */
	public LyProject getProject(String projectId)
	{
		if (projectId == null)
			return null;
		
		if (projectId.equals(getId()))
			return this;
		
		LyInfoList<LyProject>	projects = getProjects();
		int						i, np = projects.getSize();
		LyProject			    subProject, resultProject;

		for (i = 0; i < np; i++)
		{
			subProject = projects.getInfo(i);
			resultProject = subProject.getProject(projectId);
			
			if (resultProject != null)
				return resultProject;
		}

		return null;
	}

	/**
	 * Returns the classInfo at the given index.
	 *
	 * @param  index Index of the requested classInfo.
	 * @return The classInfo if it's found, null otherwise.
	 */
	public LyClassInfo getClassInfo(int index)
	{
		LyInfoList<LyClassInfo> classes = getClasses();

		return classes.getInfo(index);
	}

	/**
	 * Returns the number of classInfos in this project.
	 *
	 * @return Number of classInfos in this project.
	 */
	public int getClassInfoCount()
	{
		return getInfoCount(LyClassInfo.class);
	}

	/**
	 * Gets classes list.
	 *
	 * @return List of classes of this project.
	 */
	public LyInfoList<LyClassInfo> getClasses()
	{
		return getInfos(LyClassInfo.class);
	}

	/**
	 * Gets list of loaded classes of this project.
	 *
	 * @return List of loaded classes of this project.
	 */
	public LyInfoList<LyClassInfo> getLoadedClasses()
	{
		LyInfoList<LyClassInfo> loadedClasses = new LyInfoList<LyClassInfo>(false);
		LyClassInfo			    classInfo;
		LyFieldInfo			    fieldInfo;
		LyInfoList<LyClassInfo> classes = getClasses();
		int					    i, nc = classes.getSize();

		for (i = 0; i < nc; i++)
		{
			classInfo = classes.getInfo(i);
			if (classInfo.isLoaded())
			{
				loadedClasses.addInfo(classInfo);
				loadedClasses.addList(getSubClasses(classInfo));
			}
		}

		LyInfoList<LyFieldInfo> fields = getFields();
		int					    nf = getFieldInfoCount();

		for (i = 0; i < nf; i++)
		{
			fieldInfo = fields.getInfo(i);
			if (fieldInfo.isLoaded())
				loadedClasses.addList(getSubClasses(fieldInfo));
		}

		LyInfoList<LyProject> projects = getProjects();
		int					  np = projects.getSize();

		if (np == 0)
			return loadedClasses;

		LyProject			    subProject;
		LyInfoList<LyClassInfo> subClasses;

		for (i = 0; i < np; i++)
		{
			subProject = projects.getInfo(i);
			if (subProject.isLoaded())
			{
				subClasses = subProject.getLoadedClasses();
				loadedClasses.addList(subClasses);
			}
		}

		return loadedClasses;
	} // end method getLoadedClasses

	/**
	 * Gets list of loaded subclasses of this class info.
	 *
	 * @return List of subclasses of the given class info.
	 */
	LyInfoList<LyClassInfo> getSubClasses(LyClassInfo classInfo)
	{
		LyInfoList<LyClassInfo> subClasses = null;
		LyFieldInfo			    fieldInfo;
		int					    j, nf;

		for (j = 0, nf = classInfo.getFieldInfoCount(); j < nf; j++)
		{
			fieldInfo = classInfo.getFieldInfo(j);
			if (fieldInfo.isLoaded())
			{
				if (subClasses == null)
					subClasses = getSubClasses(fieldInfo);
				else
					subClasses.addList(getSubClasses(fieldInfo));
			}
		}

		return subClasses;
	}

	/**
	 * Gets list of loaded subclasses of this field info.
	 *
	 * @return List of subclasses of the given field info.
	 */
	LyInfoList<LyClassInfo> getSubClasses(LyFieldInfo fieldInfo)
	{
		LyInfoList<LyClassInfo> subClasses = null;
		LyClassInfo			    relationClassInfo;

		if (fieldInfo instanceof LyRelationFieldInfo)
		{
			relationClassInfo = ((LyRelationFieldInfo)fieldInfo).getClassInfo(0);
			if (relationClassInfo.getParent() == fieldInfo)
			{
				if (subClasses == null)
					subClasses = new LyInfoList<LyClassInfo>();
				subClasses.addInfo(relationClassInfo);
				subClasses.addList(getSubClasses(relationClassInfo));
			}
			else
			{
				LyInfoDaemon daemon = ((LyRelationFieldInfo)fieldInfo).getDaemon();

				if (daemon instanceof LyJoinDaemon)
				{
					relationClassInfo = ((LyJoinDaemon)daemon).getJoinClass();
					if ((relationClassInfo != null)
					    && (relationClassInfo.getParent() == fieldInfo))
					{
						if (subClasses == null)
							subClasses = new LyInfoList<LyClassInfo>();
						subClasses.addInfo(relationClassInfo);
						subClasses.addList(getSubClasses(relationClassInfo));
					}
				}
			}
		}
		else if (fieldInfo instanceof LyArrayFieldInfo)
		{
			LyArrayDaemon daemon = ((LyArrayFieldInfo)fieldInfo).getDaemon();

			if (daemon != null)
			{
				relationClassInfo = daemon.getArrayClassInfo();
				if (relationClassInfo.getParent() == fieldInfo)
				{
					if (subClasses == null)
						subClasses = new LyInfoList<LyClassInfo>();
					subClasses.addInfo(relationClassInfo);
					subClasses.addList(getSubClasses(relationClassInfo));
				}
			}

			LyFieldInfo			    subFieldInfo = ((LyArrayFieldInfo)fieldInfo).getFieldInfo();
			LyInfoList<LyClassInfo> subSubClasses = getSubClasses(subFieldInfo);

			if (subSubClasses != null)
			{
				if (subClasses == null)
					subClasses = new LyInfoList<LyClassInfo>();
				subClasses.addList(subSubClasses);
			}
		}

		return subClasses;
	} // end method getSubClasses

	/**
	 * Returns the action at the given index.
	 *
	 * @param  index Index of the requested action.
	 * @return The action if it's found, null otherwise.
	 */
	public LyAction getAction(int index)
	{
		LyInfoList<LyAction> actions = getActions();

		return actions.getInfo(index);
	}

	/**
	 * Returns the number of actions in this project.
	 *
	 * @return Number of actions in this project.
	 */
	public int getActionCount()
	{
		return getInfoCount(LyAction.class);
	}

	/**
	 * Gets actions list.
	 *
	 * @return List of actions of this project.
	 */
	public LyInfoList<LyAction> getActions()
	{
		return getInfos(LyAction.class);
	}

	/**
	 * Returns the fieldInfo at the given index.
	 *
	 * @param  index Index of the requested fieldInfo.
	 * @return The fieldInfo if it's found, null otherwise.
	 */
	public LyFieldInfo getFieldInfo(int index)
	{
		LyInfoList<LyFieldInfo> fields = getFields();

		return fields.getInfo(index);
	}

	/**
	 * Returns the number of fieldInfos in this project.
	 *
	 * @return Number of fieldInfos in this project.
	 */
	public int getFieldInfoCount()
	{
		return getInfoCount(LyFieldInfo.class);
	}

	/**
	 * Gets fields list.
	 *
	 * @return List of fields registered on this project.
	 */
	public LyInfoList<LyFieldInfo> getFields()
	{
		return getInfos(LyFieldInfo.class);
	}

	/**
	 * Returns the filter at the given index.
	 *
	 * @param  index Index of the requested filter.
	 * @return The filter if it's found, null otherwise.
	 */
	public LyFilter getFilter(int index)
	{
		LyInfoList<LyFilter> filters = getFilters();

		return filters.getInfo(index);
	}

	/**
	 * Returns the number of filters in this project.
	 *
	 * @return Number of filters in this project.
	 */
	public int getFilterCount()
	{
		return getInfoCount(LyFilter.class);
	}

	/**
	 * Gets filters list.
	 *
	 * @return List of filters registered on this project.
	 */
	public LyInfoList<LyFilter> getFilters()
	{
		return getInfos(LyFilter.class);
	}

	/**
	 * Returns the sort at the given index.
	 *
	 * @param  index Index of the requested sort.
	 * @return The sort if it's found, null otherwise.
	 */
	public LySort getSort(int index)
	{
		LyInfoList<LySort> sorts = getSorts();

		return sorts.getInfo(index);
	}

	/**
	 * Returns the number of sorts in this project.
	 *
	 * @return Number of sorts in this project.
	 */
	public int getSortCount()
	{
		return getInfoCount(LySort.class);
	}

	/**
	 * Gets sorts list.
	 *
	 * @return List of sorts registered on this project.
	 */
	public LyInfoList<LySort> getSorts()
	{
		return getInfos(LySort.class);
	}

	/**
	 * Returns the location at the given index.
	 *
	 * @param  index Index of the requested location.
	 * @return The location if it's found, null otherwise.
	 */
	public LyLocation getLocation(int index)
	{
		LyInfoList<LyLocation> locations = getLocations();

		return locations.getInfo(index);
	}

	/**
	 * Returns the number of locations in this project.
	 *
	 * @return Number of locations in this project.
	 */
	public int getLocationCount()
	{
		return getInfoCount(LyLocation.class);
	}

	/**
	 * Gets locations list.
	 *
	 * @return List of locations registered on this project.
	 */
	public LyInfoList<LyLocation> getLocations()
	{
		return getInfos(LyLocation.class);
	}

	/**
	 * Returns the notifier at the given index.
	 *
	 * @param  index Index of the requested notifier.
	 * @return The notifier if it's found, null otherwise.
	 */
	public LyNotifier getNotifier(int index)
	{
		LyInfoList<LyNotifier> notifiers = getNotifiers();

		return notifiers.getInfo(index);
	}

	/**
	 * Returns the number of notifiers in this project.
	 *
	 * @return Number of notifiers in this project.
	 */
	public int getNotifierCount()
	{
		return getInfoCount(LyNotifier.class);
	}

	/**
	 * Gets notifiers of this project.
	 *
	 * @return notifiers.
	 */
	public LyInfoList<LyNotifier> getNotifiers()
	{
		return getInfos(LyNotifier.class);
	}

	/**
	 * Get the contents of this project.
	 *
	 * @return the contents of this project
	 */
	public LyInfoList<LyInfo> getInfos()
	{
		return _contents;
	}

	/**
	 * Get the LyInfo at the given index.
	 *
	 * @param  index index of the requested LyInfo
	 * @return the info if it's found, null otherwise
	 */
	public LyInfo getInfo(int index)
	{
		return _contents.getInfo(index);
	}

	/**
	 * Return the number of infos in this project.
	 *
	 * @return Number of infos in this project.
	 */
	public int getInfoCount()
	{
		return _contents.getSize();
	}
	
	/**
     * Checks if given info is registered in this project.
     *
     * @return True if info is known, false otherwise.
     */
    public boolean hasInfo(LyInfo info)
    {
        return _contents.contains(info);
    }

	/**
	 * Get the list of infos for the given class.
	 *
	 * @param  infoClass the class for the info
	 * @return the list of info of the given class
	 */
	public <T  extends LyInfo> LyInfoList<T> getInfos(Class<T> infoClass)
	{
		return _contents.getInfos(infoClass);
	}

	/**
	 * Get the number of infos of the given class in this project.
	 *
	 * @param  infoClass the class for the info
	 * @return the number of infos of the given class in this project
	 */
	public int getInfoCount(Class<? extends LyInfo> infoClass)
	{
		LyInfoList<? extends LyInfo> list = getInfos(infoClass);

		return ((list == null) ? 0 : list.getSize());
	}

	/**
	 * Returns root action of this project.
	 *
	 * @return Root action if any, null otherwise.
	 */
	public LyAction getRootAction()
	{
		return _rootAction;
	}

	/**
	 * To get a model of this project by its type.
	 *
	 * @param type Type of the searched model
	 */
	public LyModel getModel(String type)
	{
		if ((type == null) || (_models == null))
			return null;

		int     i, n = _models.size();
		LyModel model;

		for (i = 0; i < n; i++)
		{
			model = _models.elementAt(i);
			if (type.equals(model.getType()))
				return model;
		}

		return null;
	}

	/**
	 * Returns the number of models in this project.
	 *
	 * @return Number of models in this project.
	 */
	public int getModelsCount()
	{
		return ((_models == null) ? 0 : _models.size());
	}

	/**
	 * To get the models of this project.
	 */
	public Vector<LyModel> getModels()
	{
		return ((_models == null) ? null : new Vector<LyModel>(_models));
	}

	/**
	 * The following modifiers are not public to avoid modification after parsing. They are provided
	 * to be used only by the data model parser. Visibility shouldn't be changed except for a VERY
	 * good reason.
	 */

	/**
	 * Adds a sub-project to this project.
	 *
	 * @param project Sub-project to add to this project.
	 */
	void addProject(LyProject project)
	{
		_contents.addInfo(project);
	}

	/**
	 * Adds a classInfo to this project.
	 *
	 * @param classInfo ClassInfo to add to this project.
	 */
	void addClassInfo(LyClassInfo classInfo)
	{
		_contents.addInfo(classInfo);
	}

	/**
	 * Adds a fieldInfo to this project.
	 *
	 * @param fieldInfo FieldInfo to add to this project.
	 */
	void addFieldInfo(LyFieldInfo fieldInfo)
	{
		_contents.addInfo(fieldInfo);
	}

	/**
	 * Adds an action to this project.
	 *
	 * @param action Action to add to this project.
	 */
	void addAction(LyAction action)
	{
		_contents.addInfo(action);
	}

	/**
	 * Adds a filter to this project.
	 *
	 * @param filter Filter to add to this project.
	 */
	void addFilter(LyFilter filter)
	{
		_contents.addInfo(filter);
	}

	/**
	 * Adds a sort to this project.
	 *
	 * @param sort Sort to add to this project.
	 */
	void addSort(LySort sort)
	{
		_contents.addInfo(sort);
	}

	/**
	 * Adds a location to this project.
	 *
	 * @param location Location to add to this project.
	 */
	void addLocation(LyLocation location)
	{
		_contents.addInfo(location);
	}

	/**
	 * Adds a notifier to this project.
	 *
	 * @param notifier notifier to add to this project.
	 */
	void addNotifier(LyNotifier notifier)
	{
		_contents.addInfo(notifier);
	}

	/**
	 * Adds a LyInfo to this project.
	 *
	 * @param info LyInfo to add to this project
	 */
	void addInfo(LyInfo info)
	{
		_contents.addInfo(info);
	}

	/**
	 * Sets root action of this project.
	 *
	 * @param action Root action.
	 */
	void setRootAction(LyAction action)
	{
		_rootAction = action;
	}

	/**
	 * Sets application behavior class name.
	 *
	 * @param applicationBehaviorClassName Java name of application behavior class (may be null).
	 */
	void setApplicationBehaviorClassName(String applicationBehaviorClassName)
	{
		_applicationBehavior = (applicationBehaviorClassName == null ? null
		     : new LyJavaName(this, applicationBehaviorClassName));
	}

	/**
	 * Sets session behavior class name.
	 *
	 * @param sessionBehaviorClassName Java name of session behavior class (may be null).
	 */
	void setSessionBehaviorClassName(String sessionBehaviorClassName)
	{
		_sessionBehavior = (sessionBehaviorClassName == null ? null
		     : new LyJavaName(this, sessionBehaviorClassName));
	}

	/**
	 * Adds a LyModel to this project.
	 *
	 * @param model LyModel to add to this project
	 */
	void addModel(LyModel model)
	{
		if (_models == null)
			_models = new Vector<LyModel>(1);
		_models.addElement(model);
	}
    
    /**
     * Sets version of this project.
     *
     * @param version Version.
     */
    protected void setVersion(String version)
    {
        _version = version;
    }
    
    /**
     * Sets locations list.
     */
	protected void setLocations(LyInfoList<LyInfo> locations)
    {
        _contents.removeInfos(LyLocation.class);
        _contents.addList(locations);
    }
    
    /**
     * Sets projects list.
     */
	protected void setProjects(LyInfoList<LyInfo> projects)
    {
        _contents.removeInfos(LyProject.class);
        _contents.addList(projects);
    }
    
    /**
     * Sets classes list.
     */
	protected void setClasses(LyInfoList<LyInfo> classes)
    {
        _contents.removeInfos(LyClassInfo.class);
        _contents.addList(classes);
    }
    
    /**
     * Sets fields list.
     */
	protected void setFields(LyInfoList<LyInfo> fields)
    {
        _contents.removeInfos(LyFieldInfo.class);
        _contents.addList(fields);
    }
    
    /**
     * Sets actions list.
     */
	protected void setActions(LyInfoList<LyInfo> actions)
    {
        _contents.removeInfos(LyAction.class);
        _contents.addList(actions);
    }
    
    /**
     * Sets filters list.
     */
	protected void setFilters(LyInfoList<LyInfo> filters)
    {
        _contents.removeInfos(LyFilter.class);
        _contents.addList(filters);
    }
    
    /**
     * Sets sorts list.
     */
	protected void setSorts(LyInfoList<LyInfo> sorts)
    {
        _contents.removeInfos(LySort.class);
        _contents.addList(sorts);
    }
    
    /**
     * Sets infos list.
     */
	protected void setInfos(LyInfoList<LyInfo> infos)
    {
        _contents = new LyInfoList<LyInfo>(infos);
    }
    
    /**
     * Sets models list.
     */
    protected void setModels(Vector<LyModel> models)
    {
        int     i, n = (models == null ? 0 : models.size());
        
        _models = null;
        for (i = 0; i < n; i++)
            addModel(models.elementAt(i));
    }
    
} // end class LyProject
