/**
 * LyRelationFieldInfo.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.lang.reflect.*;
import java.util.*;

import leon.app.*;
import leon.data.*;
import leon.info.binding.*;
import leon.info.daemon.*;
import leon.info.misc.*;
import leon.misc.*;


/**
 * This class implements the description of relation fields. These fields represent relations
 * between objects of an application.
 *
 * <p>A relation may be multiple, this means an object is related to several objects in the same
 * time.</p>
 *
 * <p>In that case, relations can be ordered when the order of values for a given relation field has
 * a meaning (the same list with a different order is not the same value).</p>
 *
 * <p>For a relation field, related objects are taken into the specified list of classes (instances
 * of LyClassInfo) authorized on this field. An optional additional filter may be associated to each
 * of these classes to filter objects proposed to end_user when he has to enter a value on a
 * relation field.</p>
 *
 * <p>Relations can be typed. The type determines the graphical representation of the relation.</p>
 *
 * <p>Relation may also have a direction. This property is used to indicate under which conditions
 * the relation is usable when calculating cross-references between objects of different classes.
 * The supported values are defined as constants, named Direction.XXX.</p>
 *
 * <ul>
 *   <LI>FORWARD : usable only for direct crossing,</LI>
 *   <LI>BACKWARD : usable only for reverse crossing,</LI>
 *   <LI>BOTH : usable only for both direct and reverse crossing,</LI>
 *   <LI>NONE : not usable for cross-references.
 *
 *     <p>Relation can have a default value. This default value can be either an object id given as
 *     a string or the name of a variable registered on the user session, in that last case, the
 *     value starts with a $.</p>
 *
 *     <p>Relations may also contains references to a "class field" and to a "context field". They
 *     are both used to restrict the number of objects proposed to the end-user when he has to enter
 *     values for the corresponding relation field. The class field refers to an other field of the
 *     same class that gives at one time the enabled class(es) for this relation. The context field
 *     refers to an other field of the same class that gives at one time a set of objects that is
 *     used to calculate via cross references the list of enabled objects for this relation.</p>
 *
 *     <p>Finally, local relations can be specified by a given set of routes that are used when
 *     calculating values for this field.</p>
 *
 *     <p>See also tag <RELATION> of leon.dtd</p>
 *   </LI>
 * </ul>
 *
 * @see LyClassInfo
 * @see LyFilter
 * @see LyRoute
 * @see leon.data.daemon.LyRelationDaemon
 */
public class LyRelationFieldInfo extends LyFieldInfo
{
	public static final String __VERSION = "$Id: LyRelationFieldInfo.java 41283 2012-10-19 15:03:28Z bangiole $";

	/** Directions. */
	public enum Direction
	{
		/**
		 * Direction forward : the relation can be traversed when calculating cross references beetween
		 * the class that contains this field and classes accepted on this relation.
		 */
		FORWARD,
		/**
		 * Direction backward : the relation can be traversed when calculating cross references beetween
		 * the classes enabled on this relation and the class that contains this field.
		 */
		BACKWARD,
		/**
		 * Direction both : the relation can be traversed when calculating cross references in both
		 * forward and backward directions.
		 */
		BOTH,
		/** Direction none : the relation can't be traversed to calculate cross references. */
		NONE
	}

	/** Relation types. */
	public enum Type
	{
		// Association : the relation represents a simple association between two classes.
		ASSOCIATION,
		// Composition : the relation represents a composition of related classes.
		COMPOSITION,
	    // Parent : the relation represents a parent of the related classes.
        PARENT,
	    // Dependency : the relation represents a dependency of the containing class.
        DEPENDENCY
	}

	/** Show as values. */
	public enum ShowAs
	{
		// Show as : default representation.
		DEFAULT,
		// Show as : table representation.
		TABLE,
		//Show as : hyper text link representation.
		LINK,
        // Show as : form.
        FORM
	}

	/** Type of the relation field. */
	Type _type;

	/** Relation that is extended by this relation if it exists. */
	LyRelationFieldInfo _extendedRelation;

	/** Abstract property : indicates that this field is abstract. */
	boolean _abstract;

	/** Multiple property : indicates that this field may have several values. */
	boolean _multiple;

	/** Indicates if this field is ordered or not (default is not). */
	boolean _ordered;

	/**
	 * Indicates if this field is ignoring the context of the views to compute the list of available
	 * values.
	 */
	boolean _noContext;

	/** Indicates if this field doesn't accept a toolbar. */
	boolean _noToolbar;

	/**
	 * Set of instance of LyClassInfo accepted on this relation. A filter may be associated to each
	 * class.
	 */
	LyInfoDataList<LyClassInfo, LyClassRef> _classes;

	/**
	 * Direction of the relation : used to enable or disable crossing when calculating cross
	 * references.
	 */
	Direction _direction;

	/**
	 * Default value. Either a string corresponding to an objectId or a string starting by '$' that
	 * gives the name of the variable registered on user session.
	 */
	String _default;

	/** Reference to the fieldInfo that gives information to this relation on objects classes. */
	LyFieldInfo _classField;

	/** Reference to the fieldInfo that gives information to this relation on objects contexts. */
	LyRelationFieldInfo _contextField;

	/** List of routes defining this relation. */
	LyInfoList<LyRoute> _routes;

	/** Informations about the daemon to manage this relation. */
	LyInfoDaemon _daemon;

	/**
	 * Number of rows which should be shown when displaying the relation (-1 if no preference is
	 * given).
	 */
	int _nRows;

	/**
	* Number of cols which should be shown when displaying the relation (-1 if no preference is
	* given).
	*/
	int _nCols;

	/** Number min. of values required for this relation. */
	int _min;

	/** Number max. of values accepted for this relation. */
	int _max;

	/** Gives the type of presentation used for this relation. */
	ShowAs _showAs;

	/** Sub-types. */
	LyInfoList<LyRelationFieldInfo> _subTypes;

	/** Indicates if daemons have been initialized. */
	boolean _daemonInitialized;

	/** Boolean indicating if compares should be case sensitive or not. */
	Boolean _caseSensitive;

	/**
	 * indicate if already checked target classinfo for CMS type detection
	 */
	protected boolean _cmsTypeChecked = false;

	/**
	 * Constructor of a relation.
	 *
	 * @param id               Identifier of the field.
	 * @param javaName         Name of the generated java field.
	 * @param extendedRelation Relation extended by this relation.
	 * @param template         Template used to construct this field.
	 * @param multiple         Indicates is the relation field is multiple or not.
	 * @param ordered          Indicates is relation field is ordered or not.
	 * @param direction        Direction of the relation.
	 * @param type             Type of the relation.
	 * @param reverseRelation  Reverse relation for this relation.
	 * @param userData         User data associated to this relation.
	 * @param noContext        Indicates if the relation field ignores the context of the views to
	 *                         compute available values.
	 * @param noToolbar        Indicates if this field doesn't accept a toolbar.
	 * @param showAs           Gives the type of presentation used for this relation. One of the
	 *                         SHOW_AS_XXX value.
	 */
	LyRelationFieldInfo(String id, String javaName, LyRelationFieldInfo extendedRelation, LyRelationFieldInfo template,
		boolean multiple, boolean ordered, Direction direction, Type type, Object userData,
		boolean noContext, boolean noToolbar, ShowAs showAs)
	{
		this(id);
		construct(javaName, extendedRelation, template, multiple, ordered, direction, type, userData,
			noContext, noToolbar, showAs);
	}

	/**
	 * Constructor of a relation.
	 *
	 * @param id Identifier of the field.
	 */
	LyRelationFieldInfo(String id)
	{
		super(id);
	}

	/**
	 * Init method.
	 */
	protected void init()
	{
		super.init();

		_extendedRelation = null;
		_abstract = false;
		_multiple = false;
		_ordered = false;
		_noContext = false;
		_noToolbar = false;
		_default = null;
		_classes = new LyInfoDataList<LyClassInfo, LyClassRef>();
		_classField = null;
		_contextField = null;
		_direction = Direction.BOTH;
		_routes = null;
		_daemon = null;
		_nRows = -1;
		_nCols = -1;
		_min = -1;
		_max = -1;
		_showAs = ShowAs.DEFAULT;
		_subTypes = null;
		_daemonInitialized = false;
		_caseSensitive = null;
		_cmsTypeChecked = false;
	}

	/**
	 * Init this instance attributes from its template.
	 *
	 * @param parts Indicates which parts of the template are copied.
	 */
	void initFromTemplate(int parts)
	{
        parts &= ~_initParts;
        if (parts == 0)
            return;

		if (_template == null)
			return;

		super.initFromTemplate(parts);

		LyRelationFieldInfo template = (LyRelationFieldInfo)_template;

		if ((parts & PART_DATA) > 0)
		{
			if (_extendedRelation == null)
				_extendedRelation = template._extendedRelation;

			_default = template._default;
			_classes = new LyInfoDataList<LyClassInfo, LyClassRef>(template._classes);
			if (_multiple)
			{
				_min = template._min;
				_max = template._max;
			}

			_classField = template._classField;
			_contextField = template._contextField;
			if (template._subTypes != null)
				_subTypes = new LyInfoList<LyRelationFieldInfo>(template._subTypes);
		}

		if ((parts & PART_BIND) > 0)
		{
			if (template._routes != null)
				_routes = new LyInfoList<LyRoute>(template._routes);

			_daemon = template._daemon;
		}

		if ((parts & PART_GUI) > 0)
		{
			if (_multiple)
				_nRows = template._nRows;

			_nCols = template._nCols;
		}
	} // end method initFromTemplate

	/**
	 * Construction of a relation.
	 *
	 * @param javaName         Name of the generated java field.
	 * @param extendedRelation Relation extended by this relation.
	 * @param template         Template used to construct this field.
	 * @param multiple         Indicates is the relation field is multiple or not.
	 * @param ordered          Indicates is relation field is ordered or not.
	 * @param direction        Direction of the relation.
	 * @param type             Type of the relation.
	 * @param reverseRelation  Reverse relation for this relation.
	 * @param userData         User data associated to this relation.
	 * @param noContext        Indicates if the relation field ignores the context of the views to
	 *                         compute available values.
	 * @param noToolbar        Indicates if this field doesn't accept a toolbar.
	 * @param showAs           Gives the type of presentation used for this relation. One of the
	 *                         SHOW_AS_XXX value.
	 */
	protected void construct(String javaName, LyRelationFieldInfo extendedRelation, LyRelationFieldInfo template,
		boolean multiple, boolean ordered, Direction direction, Type type, Object userData, boolean noContext,
		boolean noToolbar, ShowAs showAs)
	{
		super.construct(javaName, ((extendedRelation != null) ? extendedRelation : template));

		_extendedRelation = extendedRelation;
		_multiple = multiple;
		_ordered = ordered;
		_direction = ((direction == null) ? Direction.BOTH : direction);
		_type = ((type == null) ? Type.ASSOCIATION : type);
		setUserData(userData);
		_noContext = noContext;
		_noToolbar = noToolbar;
		_showAs = ((showAs == null) ? ShowAs.DEFAULT : showAs);
	}

	/**
	 * Final step of parsing : check consistency. Checks performed are the following :
	 *
	 * <ul>
	 *   <LI>the relation must refer at least one class info,</LI>
	 *   <LI>if not null, classField and contextField should belong to parent class,</LI>
	 *   <LI>check routes.</LI>
	 * </ul>
	 *
	 * @param     parent Object info in which this field is checked.
	 * @param     parts  Indicates which parts are checked.
	 * @exception IllegalStateException If one of these checks fails.
	 */
	void check(LyObjectInfo parent, int parts)
	{
		super.check(parent, parts);

		if ((parts & PART_DATA) > 0)
			// The relation must refer at least one class
			if (getClassInfoCount() == 0)
				throw new IllegalStateException("BF AppEngine : relation " + this + " doesn't refer to any class");

	    LyEnvironment environment = getEnvironment();

		if ((parts & PART_BIND) > 0)
        {
			if (!_daemonInitialized)
				initDaemon(parent);

            if (_daemon instanceof LyReverseDaemon)
            {
                if (_direction != Direction.NONE)
                    environment.logWarning("BF AppEngine : relation " + this
                        + " computed by a reverse daemon may be used for cross-references, it could more efficient"
                        + " to avoid using it in such cases, by setting its direction attribute to NONE.");
            }
        }

		if (parent instanceof LyClassInfo)
		{
			LyClassInfo classInfo = (LyClassInfo)parent;

			if ((parts & PART_GUI) > 0)
			{
				// ClassField should belong to parent class
				if ((_classField != null) && !classInfo.contains(_classField))
					throw new IllegalStateException("BF AppEngine : invalid class field " + _classField + " for relation "
						+ this + " in class " + classInfo);

				// If _classField is a choice field info, check that the names of
				// the options are class names and that these class are among
				// the possible class for this relation.
				if (_classField instanceof LyChoiceFieldInfo)
				{
					LyChoiceFieldInfo	    choiceField = (LyChoiceFieldInfo)_classField;
					LyChoiceOption		    option;
					LyInfoList<LyClassInfo> appClasses = _application.getClassInfos();
					LyClassInfo			    linkClass;
					String				    optionName;
					short				    i, n = choiceField.getChoiceOptionCount();

					for (i = 0; i < n; i++)
					{
						option = choiceField.getChoiceOption(i);
						optionName = option.getName();

						linkClass = appClasses.getInfo(optionName, true);

						if ((linkClass == null) || !relatedWith(linkClass))
							throw new IllegalStateException("BF AppEngine : invalid option name " + optionName
								+ " for choice " + choiceField + " in class " + classInfo
								+ ". This choice is a CLASS_FIELD for relation " + this
								+ ". All options names must match the name of a class declared on this relation, but "
								+ linkClass + " was not found.");
					}
				}
			}

			if ((parts & PART_DATA) > 0)
			{
				// ContextField should belong to parent class
				if ((_contextField != null) && !classInfo.contains(_contextField))
					throw new IllegalStateException("BF AppEngine : invalid context field " + _contextField
						+ " for relation " + this + " in class " + classInfo);
			}

			if ((parts & PART_BIND) > 0)
			{
				int     i, n = getRoutesCount();
				LyRoute route;

				// Check all the routes.
				for (i = 0; i < n; i++)
				{
					route = getRoute(i);
					try
					{
						route.check(classInfo);
					}
					catch (Exception e)
					{
						environment.logError("BF AppEngine : route check failed for route " + route + ", "
							+ e.getMessage());
					}
				}
			}

			if ((parts & PART_GUI) > 0)
			{
				if (getRoutesCount() > 0)
				{
					// This relation has some routes : We remove the marks CREATE and SET.
					if (hasMark(MARK_CREATE))
					{
						_marks.clear(MARK_CREATE);
						environment.logWarning("BF AppEngine : remove mark CREATE from the fieldInfo " + this
							+ " which has routes.");
					}

					if (hasMark(MARK_SET))
					{
						_marks.clear(MARK_SET);
						environment.logWarning("BF AppEngine : remove mark SET from the fieldInfo " + this
							+ " which has routes.");
					}
				}
			}
		}

		// Relations of type composition must be private
		if (getType() == Type.COMPOSITION)
		{
			if ((parts & PART_DATA) > 0)
			{
				if (getClassInfoCount() > 1)
					environment.logWarning("BF AppEngine : relation with type composition " + this
						+ " can't accept more than one class.");

				if (!isMultiple() && (getClassInfo(0) == parent))
					throw new IllegalStateException("BF AppEngine : simple composition relation " + this
						+ " can't be contained in class " + parent + " to which it is related to");

				if (hasMark(MARK_UNIQUE))
				{
					_marks.clear(MARK_UNIQUE);
					environment.logWarning("BF AppEngine : remove mark UNIQUE from the fieldInfo " + this
						+ " which is a relation with type composition");
				}
			}

			if ((parts & PART_GUI) > 0)
			{
				if (!hasMark(MARK_PRIVATE))
				{
					if (_marks == null)
						_marks = new BitSet();

					_marks.set(MARK_PRIVATE);
					environment.logWarning("BF AppEngine : add mark PRIVATE for the fieldInfo " + this
						+ " which is a relation with type composition");
				}
			}
		}

		if ((parts & PART_DATA) > 0)
		{
			// Add dependencies on contextField and classField
			if (_contextField != null)
				_contextField.addDepend(this);

			if (_classField != null)
				_classField.addDepend(this);

			if (!isMultiple())
			{
				if (_min >= 0)
				{
					environment.logWarning("BF AppEngine : MIN not used for non multiple relation " + this);
					_min = -1;
				}

				if (_max >= 0)
				{
					environment.logWarning("BF AppEngine : MAX not used for non multiple relation " + this);
					_max = -1;
				}
			}
			else
			{
				if (_max == 0)
				{
					environment.logWarning("BF AppEngine : MAX should be greater than 0 for relation field " + this);
					_max = -1;
				}

				if ((_min >= 0) && (_max > 0) && (_min > _max))
					throw new IllegalStateException("BF AppEngine : min should be lower or equal to max for relation field "
						+ this);
			}
		}

		if ((parts & PART_BIND) > 0)
		{
			if (_daemon != null)
			{
				if (!hasMark(MARK_LOCAL))
				{
					if (_marks == null)
						_marks = new BitSet();

					_marks.set(MARK_LOCAL);
					_application.getEnvironment().logWarning("BF AppEngine : add mark LOCAL to relation " + this
						+ ", it should be local because it's calculated by a daemon.");
				}

				_daemon.check(this);

				if (_daemon instanceof LyReverseDaemon)
				{
					// Add load mark on reverse relation if needed
					LyRelationFieldInfo reverse = ((LyReverseDaemon)_daemon).getReverseRelation();

					if ((hasMark(MARK_LOAD)
					        || hasMark(MARK_TABLE)
					        || hasMark(MARK_NAME))
					    && !reverse.hasMark(MARK_LOCAL)
					    && !reverse.hasMark(MARK_LOAD))
					{
						if (reverse._marks == null)
							reverse._marks = new BitSet();

						reverse._marks.set(MARK_LOAD);
						getEnvironment().logWarning("BF AppEngine : add mark LOAD for field " + reverse
							+ " for optimizing the computing of relation " + this + " by reverse daemon");
					}
				}
			}
		}

        if ((parts & PART_GUI) > 0)
        {
            // Check if this relation needs translate mark
            if (!hasMark(MARK_TRANSLATE)
                && needTranslateMark())
            {
                if (_marks == null)
                    _marks = new BitSet();
                _marks.set(MARK_TRANSLATE);
            }
        }
	} // end method check

	/**
	 * Checks if translate mark is needed on this relation.
	 * @return True if needed, false otherwise.
	 */
	protected boolean needTranslateMark()
	{
	    int                        i, nc = getClassInfoCount();
	    LyClassInfo                classInfo;
	    int                        j, nf;
	    LyFieldInfo                fieldInfo;
	    LyInfoList<LyFieldInfo>    nameFields;

	    for (i = 0; i < nc; i++)
	    {
	        classInfo = getClassInfo(i);
	        nameFields = classInfo.getNameFieldInfos();

	        for (j = 0, nf = nameFields.getSize(); j < nf; j++)
	        {
	            fieldInfo = nameFields.getInfo(j);
	            if (fieldInfo.hasMark(LyFieldInfo.MARK_TRANSLATE))
	                return true;
	        }
	    }

	    return false;
	}

	/**
	 * Lock this relation field.
	 */
	protected void lock()
	{
		super.lock();
		if (!_locked)
			return;

		// Lock internal lists
		_classes.lock();
		if (_subTypes != null)
			_subTypes.lock();
	}

	/* Implementation of LyFieldInfo methods */

	/**
	 * Returns class of values for this field. The value class corresponding to a number field is :
	 *
	 * <ul>
	 *   <LI>leon.data.LyObjectId, if the relation is not multiple,</LI>
	 *   <LI>leon.data.LyObjectId[], if the relation is multiple.</LI>
	 * </ul>
	 */
	public Class<?> getValueClass()
	{
		if (isMultiple())
			return LyObjectId[].class;
		else
			return LyObjectId.class;
	}

	/**
	 * Builds the appropriate LyValue corresponding to this field for the given value object.
	 *
	 * @see LyRelationValue
	 */
	public LyValue buildFieldValue(Object value)
	{
		if (value == null)
			return new LyRelationValue(this);
		else if (value instanceof LyObjectId)
			return new LyRelationValue(this, (LyObjectId)value);
		else if (value instanceof LyObjectId[])
			return new LyRelationValue(this, (LyObjectId[])value);
		else if (value instanceof LyObject)
			return new LyRelationValue(this, (LyObject)value);
		else if (value instanceof LyObjectList)
			return new LyRelationValue(this, (LyObjectList)value);
		else if (value instanceof String)
			return new LyRelationValue(this, (String)value);
		else
			return super.buildFieldValue(value);
	}

	/**
	 * Builds internal value for this field from a string (previously constructed by toString).
	 *
	 * @see LyRelationFieldInfo#toString
	 */
	public Object parse(String stringValue) throws LyFieldInfoFormatException
	{
		if (stringValue == null)
			return null;

		LyObjectId  objectId;
		LyClassInfo classInfo;

		// Multiple ?
		if (isMultiple())
		{
			// Try Visible separator
			String[] values;

			if (stringValue.indexOf(_application.VALUE_SEP) >= 0)
				values = LyString.tokenizer(stringValue, _application.VALUE_SEP);
			else if (stringValue.indexOf(LyApplication.VISIBLE_ARRAY_SEP) >= 0)
				values = LyString.tokenizer(stringValue, LyApplication.VISIBLE_ARRAY_SEP);
			else
				values = LyString.tokenizer(stringValue, LyApplication.VISIBLE_VALUE_SEP);

			int				   i, n = values.length;
			Vector<LyObjectId> objectIds = new Vector<LyObjectId>(n);

			for (i = 0; i < n; i++)
			{
				if ((_classes.getSize() == 1) && (values[i].indexOf(LyObjectId.CLASS_SEP) == -1))
				{
					classInfo = getClassInfo(0);
					objectId = new LyObjectId(classInfo, values[i]);
				}
				//FFT00400
				else if (values[i].indexOf(LyObjectId.CLASS_SEP) != -1)
				{
					String value = parseClass(values[i]);
					String classId = value.substring(0, value.indexOf(LyObjectId.CLASS_SEP));
					if ((_classes.getSize() == 1) && (getApplication().getClassInfo(classId) == null))
					{
						objectId = new LyObjectId(getClassInfo(0), value);
					}
					else
						objectId = new LyObjectId(_application, value);
				}
				else
					objectId = new LyObjectId(_application, values[i]);

				objectIds.addElement(objectId);
			}

			LyObjectId[] list = new LyObjectId[objectIds.size()];

			objectIds.copyInto(list);
			return list;
		}
		else
		{
			if ((_classes.getSize() == 1) && (stringValue.indexOf(LyObjectId.CLASS_SEP) == -1))
			{

				// There is no separator but only one classInfo : we can build
				// the objectId.
				classInfo = getClassInfo(0);
				objectId = new LyObjectId(classInfo, stringValue);
			}

			//FFT00400
			else if (stringValue.indexOf(LyObjectId.CLASS_SEP) != -1)
				objectId = new LyObjectId(_application, parseClass(stringValue));
			else
				objectId = new LyObjectId(_application, stringValue);

			return objectId;
		}
	} // end method parse

	/**
	 * This method remplace the class specified in the given string by the current classInfo, if the
	 * class is the peer class of the classInfo.
	 *
	 * @param  stringValue the object id as a string
	 * @return the object id as a string for the field
	 */
	protected String parseClass(String stringValue)
	{
		//FFT00400

		// We have a class Separator
		String className = stringValue.substring(0, stringValue.indexOf(LyObjectId.CLASS_SEP));

		LyClassInfo classInfo;

		int			   n = _classes.getSize();
		Vector<String> names = new Vector<String>(n);

		// Fill in the names vector
		for (int i = 0; i < n; i++)
			names.add(getClassInfo(i).getId());

		if (names.contains(className))
			return stringValue;

		// for each classInfo, get is peerClass, if the peer class is a LyClassInfo,
		// check if it is equals to the className
		// if yes -> change it

		for (int i = 0; i < n; i++)
		{
			classInfo = getClassInfo(i);
			LyClass peerClass = classInfo.getPeerClass();

			if (peerClass instanceof LyClassInfo)
			{

				// check if the peer class is the class in the string
				if (peerClass.getId().equals(className))
				{
					String value = stringValue.substring(stringValue.indexOf(LyObjectId.CLASS_SEP));

					stringValue = classInfo.getId() + value;

					return stringValue;
				}
			}
		}

		return stringValue;
	} // end method parseClass

	/**
	 * Returns the specified value for this field as a "displayable" string.
	 */
	public String toName(Object fieldValue)
	{
		return toName(fieldValue, null);
	}

	/**
	 * Returns the specified value for this field as a "displayable" string.
	 *
	 * @param Object fieldValue : The fieldValue to display.
	 * @param String separator  : The value of separator between field values.
	 *
	 * @return String : Formated String.
	 */
	public String toName(Object fieldValue, String separator)
	{
		if (fieldValue == null)
			return null;

		if (fieldValue.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class " + fieldValue.getClass() + " expected "
				+ getValueClass());

		String name = getFormattedName(fieldValue);

		if (name != null)
			return name;

		if (!isMultiple())

			// Return the name of the related object
			return getName((LyObjectId)fieldValue);

		// Return the concatenation of names of related objects
		// Display name for a multiple relation with name order.
		LyObjectId[] objectIds = getOrderedIds((LyObjectId[])fieldValue);
		StringBuffer buffer = new StringBuffer();
		int			 i, n = objectIds.length;
		String		 objectName;

		for (i = 0; i < n; i++)
		{
			objectName = getName(objectIds[i]);
			if (objectName == null)
				continue;

			if (buffer.length() > 0)
			{
				if (separator != null)
					buffer.append(separator);
				else
					buffer.append(LyApplication.VISIBLE_FIELD_SEP);
			}

			buffer.append(objectName);
		}

		return buffer.toString();
	} // end method toName

	/**
	 * Sort list of object identifiers.<br />
	 * Sort is made on the name of corresponding objects. If relation is ordered, current method
	 * does not sort the list which is returned as it is.
	 *
	 * @param  objectIds list of object identifiers to be sorted.
	 * @return the list of object identifiers sorted by the names of the corresponding objects.
	 */
	public LyObjectId[] getOrderedIds(LyObjectId[] objectIds)
	{
		// If relation is ordered, do not sort the list
		if (isOrdered())
			return objectIds;

		if ((objectIds == null) || (objectIds.length < 2))
			return objectIds;

		int			 i, j, n = objectIds.length;
		LyObjectId[] newObjectIds = new LyObjectId[n];
		String[]     objectNames = new String[n];
		LyObjectId   id;
		String		 name;

		for (i = 0; i < n; i++)
		{
			newObjectIds[i] = objectIds[i];
			objectNames[i] = getName(objectIds[i]);
		}

		// Order the object identifiers by object names
		for (i = 0; i < n; i++)
		{
			for (j = i + 1; j < n; j++)
			{
				name = objectNames[j];
				id = newObjectIds[j];

				if ((objectNames[i] != null) && (name != null) && (objectNames[i].compareTo(name) > 0))
				{
					objectNames[j] = objectNames[i];
					objectNames[i] = name;
					newObjectIds[j] = newObjectIds[i];
					newObjectIds[i] = id;
				}
			}
		}

		return newObjectIds;
	} // end method getOrderedIds

	/**
	 * Returns the specified value for this field as a "displayable" string.
	 */
	protected String getName(LyObjectId objectId)
	{
		LyClassInfo classInfo = objectId.getClassInfo();

		if (getEnvironment().testResource("LY_HIDE_UNKNOWN_OBJECTS")
				&& (objectId.getLogicalObject() == null))
			return null;

		// Optimization : avoid loading if id and name are the same
		if (!classInfo.isNameDifferentOfId())
			return objectId.toName();

		// and if name is a sub-part of id
		else if (classInfo.isNamePartOfId())
		{
			LyInfoList<LyFieldInfo> idFieldInfos = classInfo.getIdFieldInfos();
			int					    i, n = idFieldInfos.getSize();
			LyFieldInfo			    fieldInfo;
			String[]			    names = objectId.toNames();
			int					    nn = ((names == null) ? 0 : names.length);
			StringBuffer		    name = new StringBuffer();
			char				    nameSep = getApplication().NAME_SEP;

			for (i = 0; i < n; i++)
			{
				fieldInfo = idFieldInfos.getInfo(i);
				if (fieldInfo.getBinding() instanceof LyFieldsBinding)
					break; // if fbinds the id is a composition of multiple ids from the fbinds : skip optimizatiuon

				if (fieldInfo.hasMark(LyFieldInfo.MARK_NAME) && (i < nn))
				{
					if (name.length() > 0)
						name.append(nameSep);

					name.append(names[i]);
				}
			}

			if (name.length() > 0)
				return name.toString();
		}

		LyObject object = objectId.getLogicalObject();

		return ((object != null) ? object.getName() : objectId.toName());
	} // end method getName

	/**
	 * Converts the specified value for this field to an "internal" string (string that enables
	 * reconstruction through parse).
	 *
	 * @see LyRelationFieldInfo#parse
	 */
	public String toString(Object fieldValue)
	{
		if (fieldValue == null)
			return null;

		if (fieldValue.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class " + fieldValue.getClass() + " expected "
				+ getValueClass());

		String string = getFormattedString(fieldValue);

		if (string != null)
			return string;

		if (!isMultiple())
		{
			if (getEnvironment().testResource("LY_HIDE_UNKNOWN_OBJECTS")
					&& (((LyObjectId)fieldValue).getLogicalObject() == null))
				return null;

			// Return the name of the related object
			return ((LyObjectId)fieldValue).getFullString();
		}

		// Return the concatenation of names of related objects
		LyObjectId[] objectIds = (LyObjectId[])fieldValue;
		int			 i, n = objectIds.length;
		String		 id = null;

		for (i = 0; i < n; i++)
		{
			if (objectIds[i] == null)
				continue;

			if (getEnvironment().testResource("LY_HIDE_UNKNOWN_OBJECTS")
					&& (objectIds[i].getLogicalObject() == null))
				continue;

			if (id == null)
				id = objectIds[i].getFullString();
			else
				id = id + _application.VALUE_SEP + objectIds[i].getFullString();
		}

		return ((id == null) ? "" : id);
	} // end method toString

	/**
	 * Controls the given value : - the number of values must be greater or equal to min, - the
	 * number of values must be lower or equal to max.
	 */
	public String controlValue(LyEnvironment environment, Object fieldValue)
	{
		if (fieldValue == null)
			return super.controlValue(environment, fieldValue);

		if (fieldValue.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class " + fieldValue.getClass() + " expected "
				+ getValueClass());

		String message = super.controlValue(environment, fieldValue);

		if (message != null)
			return message;

		if (isMultiple())
		{
			int n = ((LyObjectId[])fieldValue).length;

			if (environment == null)
				environment = _application.getEnvironment();

			if ((_min > 0) && (n < _min))
				return environment.getMessage("LY_ER_RELATION_MIN", getName(), new Integer(_min));

			if ((_max > 0) && (n > _max))
				return environment.getMessage("LY_ER_RELATION_MAX", getName(), new Integer(_max));
		}

		return null;
	} // end method controlValue

	/**
	 * Checks the given value : - the value must correspond to options of the choice.
	 */
	public String checkValue(LyEnvironment environment, Object fieldValue)
	{
		if (fieldValue == null)
			return super.checkValue(environment, fieldValue);

		if (fieldValue.getClass() != getValueClass())
			return environment.getMessage("LY_ER_FIELD_VALUE_CLASS", this, fieldValue.getClass(), getValueClass());

		// Check class info of the values
		if (!isMultiple())
		{
			LyClassInfo classInfo = ((LyObjectId)fieldValue).getClassInfo();

			if (!relatedWith(classInfo))
				return environment.getMessage("LY_ER_RELATION_CLASS_INFO", this, classInfo);
		}
		else
		{
			LyObjectId[] objectIds = (LyObjectId[])fieldValue;
			int			 i, n = objectIds.length;
			LyClassInfo  classInfo;

			for (i = 0; i < n; i++)
			{
				if (objectIds[i] == null)
					continue;

				classInfo = objectIds[i].getClassInfo();
				if (!relatedWith(classInfo))
					return environment.getMessage("LY_ER_RELATION_CLASS_INFO", this, classInfo);
			}
		}

		return super.checkValue(environment, fieldValue);
	} // end method checkValue

	/**
	 * Compares two objects for this fieldInfo.
	 */
	public int compare(Object value1, Object value2, BitSet modifiers)
	{

		/*
		if (value1 == value2)
		        return 0;
		if (value1 == null)
		        return -1;
		if (value2 == null)
		        return 1;
		if (!isMultiple()
		                && ((LyObjectId)value1).equals((LyObjectId)value2))
		        return 0;
		*/
		if (equals(value1, value2, modifiers))
			return 0;

		if (value1 == null)
			return -1;

		if (value2 == null)
			return 1;

		boolean caseSensitive = false;
		boolean byName = true;

		if (modifiers != null)
		{
			caseSensitive = modifiers.get(LyFilterExpression.MOD_CASE_SENSITIVE);
			byName = !modifiers.get(LyFilterExpression.MOD_NOT_BY_NAME);
		}
		else
		{
			if (_caseSensitive != null)
				caseSensitive = _caseSensitive.booleanValue();
			else
			{
				int					    i, n = getClassInfoCount();
				LyClassInfo			    classInfo;
				LyInfoList<LyFieldInfo> nameFields;
				int					    j, nf;
				LyFieldInfo			    fieldInfo;

				for (i = 0; (i < n) && !caseSensitive; i++)
				{
					classInfo = getClassInfo(i);
					if (classInfo.isNameDifferentOfId())
						nameFields = classInfo.getFieldInfos(LyFieldInfo.MARK_NAME);
					else
						nameFields = classInfo.getIdFieldInfos();

					for (j = 0, nf = nameFields.getSize(); (j < nf) && !caseSensitive; j++)
					{
						fieldInfo = nameFields.getInfo(j);
						caseSensitive = !(fieldInfo instanceof LyTextFieldInfo);
					}
				}

				_caseSensitive = Boolean.valueOf(caseSensitive);
			}
		}

		// If there is only one target class for this relation, check if name fields for this class
		// is a number field. In this case, use the compare method on the number field in order to
		// use the correct comparison algorithm.
		if (byName && (getClassInfoCount() == 1))
		{
			LyClassInfo classInfo = getClassInfo(0);
			LyInfoList<LyFieldInfo>  nameFields;

			if (classInfo.isNameDifferentOfId())
				nameFields = classInfo.getFieldInfos(LyFieldInfo.MARK_NAME);
			else
				nameFields = classInfo.getIdFieldInfos();

			if ((nameFields.getSize() == 1) && (nameFields.getInfo(0) instanceof LyNumberFieldInfo)
				&& (value1 instanceof LyObjectId) && (value2 instanceof LyObjectId))
			{
				LyNumberFieldInfo numberFieldInfo = (LyNumberFieldInfo)nameFields.getInfo(0);
				LyObject		  object1 = ((LyObjectId)value1).getLogicalObject();
				LyObject		  object2 = ((LyObjectId)value2).getLogicalObject();

				if ((object1 != null) && (object2 != null))
					return numberFieldInfo.compare(object1.getValue(numberFieldInfo), object2.getValue(numberFieldInfo),
						null);
			}
		}

		int comp = 0;

		if (byName)
		{
			String name1 = toName(value1);
			String name2 = toName(value2);

			if (name1 == null)
				return -1;

			if (name2 == null)
				return 1;

			if (caseSensitive)
				comp = name1.compareTo(name2);
			else
				comp = name1.compareToIgnoreCase(name2);
		}

		if (comp == 0)
		{
			String id1 = toString(value1);
			String id2 = toString(value2);

			if (id1 == null)
				return -1;

			if (id2 == null)
				return 1;

			comp = id1.compareTo(id2);
		}

		return comp;
	} // end method compare

	/**
	 * Returns the filter conditions authorized on this field. The values must match the short
	 * constants defined in LyFilterExpression.
	 *
	 * @return An array of conditions corresponding to the authorized conditions on this field.
	 * @see    leon.info.LyFilterExpression
	 */
	public LyFilterExpression.Condition[] getFilterConditions()
	{

		// If relation is multiple, only equals and contains are allowed.
		// If not, only equals is allowed.
		int							   nConditions = (_multiple ? 2 : 1);
		LyFilterExpression.Condition[] conditions = new LyFilterExpression.Condition[nConditions];

		conditions[0] = LyFilterExpression.Condition.EQUALS;

		if (_multiple)
			conditions[1] = LyFilterExpression.Condition.CONTAINS;

		return conditions;
	}

	/**
	 * Compares two objects for equality.
	 */
	public boolean equals(Object value1, Object value2, BitSet modifiers)
	{
		if (value1 == value2)
			return true;
		else if ((value1 == null) || (value2 == null))
			return false;

		if (!isMultiple())
		{
			LyObjectId id1, id2;

			id1 = (LyObjectId)value1;
			id2 = (LyObjectId)value2;

			return id1.equals(id2);
		}
		else
		{
			LyObjectId[] ids1, ids2;
			int			 i, n, p;

			ids1 = (LyObjectId[])value1;
			ids2 = (LyObjectId[])value2;

			n = ids1.length;
			p = ids2.length;

			if (n != p)
				return false;

			// case of an ordered relation
			if (isOrdered())
			{
				for (i = 0; i < n; i++)
				{
					if (((ids1[i] == null) && (ids2[i] != null))
						|| ((ids1[i] != null) && ((ids2[i] == null) || !ids1[i].equals(ids2[i]))))
						return false;
				}
			}
			else
			{ // non-ordered relation : check if  each arrays contains the same elements.

				// copy the 2nd array into this list
				ArrayList<LyObjectId> list2 = new ArrayList<LyObjectId>();

				for (i = 0; i < n; i++)
					list2.add(ids2[i]);

				for (i = 0; i < n; i++)
				{
					if (list2.contains(ids1[i]))
						list2.remove(list2.indexOf(ids1[i])); // remove the object found
					else
						return false;
				}
			}

			return true;
		}
	} // end method equals

	/**
	 * Checks if the second value is contained in the first one, for this field.
	 */
	public boolean contains(Object value1, Object value2, BitSet modifiers)
	{
		//if (!_multiple)
		//      return super.contains(value1, value2);

		if ((value1 == null) && (value2 == null))
			return true;
		else if (value1 == null)
			return false;
		else if (value2 == null)
			return false;

		Class<?> valueClass = getValueClass();

		if (value1.getClass() != valueClass)
			throw new IllegalArgumentException("BF AppEngine : invalid value class " + value1.getClass() + " expected "
				+ valueClass);

		if ((value2.getClass() != valueClass) && (valueClass.getComponentType() != value2.getClass()))
			throw new IllegalArgumentException("BF AppEngine : invalid value class " + value2.getClass() + " expected "
				+ valueClass);

		if (isMultiple())
		{
			if (value2 instanceof LyObjectId)
			{
				LyObjectId[] list1 = (LyObjectId[])value1;
				int			 i, n = list1.length;
				LyObjectId[] val1, val2;

				for (i = 0; i < n; i++)
				{
					val1 = new LyObjectId[1];
					val1[0] = list1[i];
					val2 = new LyObjectId[1];
					val2[0] = (LyObjectId)value2;
					if (equals(val1, val2))
						return true;
				}

				return false;
			}
			else if (value2 instanceof LyObjectId[])
			{
				LyObjectId[] list2 = (LyObjectId[])value2;
				int			 i, n = list2.length;

				for (i = 0; i < n; i++)
				{
					if (!contains(value1, list2[i]))
						return false;
				}
			}
			else
				return false;
		}
		else
		{

			// not multiple
			if (value2 instanceof LyObjectId)
				return equals(value1, value2);
			else if (value2 instanceof LyObjectId[])
			{
				LyObjectId[] list2 = (LyObjectId[])value2;
				int			 i, n = list2.length;

				for (i = 0; i < n; i++)
				{
					if (!contains(value1, list2[i]))
						return false;
				}
			}
			else
				return false;
		}

		return true;
	} // end method contains

	/**
	 * Check if the given classInfo is a valid classInfo for the current relation.
	 *
	 * @param  classInfo the classInfo to check
	 * @return true if the given classInfo is valid for the relation
	 */
	protected boolean validateClassInfo(LyClassInfo classInfo)
	{
		if (classInfo.isVirtual())
			classInfo = classInfo.getRealClass();

		LyClassInfo currentClassInfo;

		// First check if the given classInfo is valid for the current fieldInfo
		if (_classes.contains(classInfo))
			return true;

		// We must check if the given class extends one of the classInfo in the list
		int i, n = _classes.getSize();

		for (i = 0; i < n; i++)
		{
			currentClassInfo = _classes.getInfo(i);
			if (currentClassInfo.isVirtual())
				currentClassInfo = currentClassInfo.getRealClass();

			if ((classInfo == currentClassInfo) || classInfo.inheritsFrom(currentClassInfo))
				return true;
		}

		return false;
	}

	/**
	 * Constructs a value for this fieldInfo from an object for a given instance. Supported
	 * conversions are the following :
	 *
	 * <TABLE>
	 *   <TR>
	 *     <TD>LyObjectId, String, Object</TD>
	 *     <TD>LyObjectId</TD>
	 *   </TR>
	 *   <TR>
	 *     <TD>LyObjectId[], Object[], String[], Vector</TD>
	 *     <TD>LyObjectId[]</TD>
	 *   </TR>
	 * </TABLE>
	 *
	 * @exception LyUnsupportedFieldBindingException If the given object can't be converted to a
	 *                                               value of this field.
	 */
	public Object getValueFromObject(Object object, LyObject instance)
	{
        Object value = decodePeerValue(object);

        if (value == null)
            return null;

		if (!isMultiple())
			return getObjectId(value, instance);

		// Build an array with given value(s)
		Object[] array;

		// If the object is a string, try to parse the string before...
		if (value instanceof String)
			value = parse((String)value);

		if (value instanceof Collection<?>)
		{
			// Gets an array from the collection
			Collection<?> temp = (Collection<?>)value;
			array = temp.toArray();
		}
		else if (value.getClass().isArray())
			array = (Object[])value;
		else
		{
			array = new Object[1];
			array[0] = value;
		}

		int   i, n = (array == null ? 0 : array.length);

		if (n == 0)
			return null;

		// Build the array of objectIds
		LyObjectId[] objectIds = new LyObjectId[n];
		int			 nNull = 0;

		for (i = 0; i < n; i++)
		{
			objectIds[i] = getObjectId(array[i], instance);
			if (objectIds[i] == null)
				nNull++;
		}

		if (nNull == 0)
			return objectIds;
		else if (n == nNull)
			return null;
		else
		{
			LyObjectId[] newObjectIds = new LyObjectId[n - nNull];
			int			 j = 0;

			for (i = 0; i < n; i++)
				if (objectIds[i] != null)
				{
					newObjectIds[j] = objectIds[i];
					j++;
				}

			return newObjectIds;
		}

	} // end method getValueFromObject

	/**
	 * Internal method to build a LyObjectId from another LyObjectId.
	 *
	 * @param  objectId ObjectId to convert to a valid LyObjectId.
	 * @param  instance Instance of LyObject for which the value of this field is requested.
	 * @return The object id if it could be built, null otherwise.
	 */
	LyObjectId getObjectId(LyObjectId objectId, LyObject instance)
	{
		// If object is a LyObjectId, we have to check that it can match our classInfo
		LyClassInfo objectClassInfo = objectId.getClassInfo();

		// A class can accept our LyObjectId, we return it
		if (validateClassInfo(objectClassInfo))
			return objectId;

		// We check if a peer class of one of the ClassInfo
		// is equals to the object ClassInfo, if true, we change the classInfo.
		int		    i, n = _classes.getSize();
		LyClassInfo classInfo;
		LyClass     peerClass;
		String	    id;

		for (i = 0; i < n; i++)
		{
			classInfo = getClassInfo(i);
			peerClass = classInfo.getPeerClass();

			if (peerClass instanceof LyClassInfo)
			{
				if (objectClassInfo.equals(peerClass))
				{
					id = classInfo.getId() + LyObjectId.CLASS_SEP + objectId.toString();
					return new LyObjectId(_application, id);
				}
			}
		}

		// We failed to get a valid value for this relation
		return null;
	} // end method getObjectId

	/**
	 * Internal method to build a LyObjectId from an Object that could be either a reference or an
	 * id.
	 *
	 * @param  object   Value object to convert to a LyObjectId.
	 * @param  instance Instance of LyObject for which the value of this field is requested.
	 * @return The object id if it could be built, null otherwise.
	 */
	LyObjectId getObjectId(Object object, LyObject instance)
	{
		if (object == null)
			return null;

		if (object instanceof LyObjectId)
			return getObjectId((LyObjectId)object, instance);

		if ((object instanceof String)
		        && (((String)object).indexOf(LyObjectId.CLASS_SEP) >= 0))
		{
			LyObjectId result = null;
			try
			{
				result = new LyObjectId(_application, (String)object);
			}
			catch (Exception e)
			{
				// Ok, not really a class sep
				result = null;
			}
			if (result != null)
				return result;
		}

		if (_classes.getSize() == 1)
		{
			LyClassInfo ci = getClassInfo(0);
			LyClass     pc = ci.getPeerClass();

			if ((pc == null) || (pc.getObjectClass() != object.getClass()))
				return new LyObjectId(getClassInfo(0), object);
		}

		LyClassInfo[] classes = null;
		int			  i, n = getClassInfoCount();

		if ((_classField != null) && (instance != null))
		{
			// Get valid classes from classField value
			Object	    classValue = instance.getValue(_classField);
			String	    str = _classField.toName(classValue);
			String[]    list = LyString.stringToList(str, _application.VALUE_SEP);
			int		    c, nc = ((list == null) ? 0 : list.length);
			LyClassInfo classInfo;

			for (i = 0; i < n; i++)
			{
				classInfo = getClassInfo(i);

				for (c = 0; c < nc; c++)
				{
					if (classInfo.getName().equals(list[c]) || classInfo.getId().equals(list[c]))
					{
						if (classes == null)
							classes = new LyClassInfo[nc];

						classes[c] = classInfo;
					}
				}
			}

			if ((classes != null) && (classes.length == 1))
				return new LyObjectId(classes[0], object);
		}

		if (classes == null)
		{
			classes = new LyClassInfo[n];
			for (i = 0; i < n; i++)
				classes[i] = getClassInfo(i);
		}

		LyObjectSource source;
		LyObjectId     objectId;
		LyObject	   relObject = null;
		LyClass		   peerClass;

		// Is this object a reference to another physical object ?
		for (i = 0, n = classes.length; i < n; i++)
		{
			if (classes[i] == null)
				continue;

			peerClass = classes[i].getPeerClass();
			if (peerClass == null)
			{
				// The classInfo is local : we test directly the object.
				if (object.getClass() == classes[i].getObjectClass())
					return ((LyObject)object).getObjectId();
			}
			else if (object.getClass() == peerClass.getObjectClass())
			{
				source = (LyObjectSource)classes[i].getDataSource();
				relObject = source.getObjectFromPeer(object);
				if (relObject != null)
					return relObject.getObjectId();
				else
					return new LyObjectId(classes[i], object);
			}
		}

		// The object has to be an object id
		Object peer;
		int    p = 0;

		for (p = 0; p < 2; p++)
		{
			for (i = 0; i < n; i++)
			{
				if (classes[i] == null)
					continue;

				source = classes[i].getObjectSource();
				peerClass = classes[i].getPeerClass();

				if (peerClass == null)
				{
					// The class is local : object is a logical id
					objectId = new LyObjectId(classes[i], object);
					if (((p == 0) && objectId.checkObject()) || (p == 1))
						relObject = objectId.getLogicalObject();
				}
				else
				{
					// The class is not local : object is a physical id
					objectId = new LyObjectId(classes[i].getPeerClass(), object);
					if (((p == 0) && objectId.checkObject()) || (p == 1))
					{
						peer = objectId.getObject();
						relObject = source.getObjectFromPeer(peer);

						if ((relObject == null) && (p == 1))
						{
							// Last attempt on logical level
							objectId = new LyObjectId(classes[i], object);
							if (objectId.checkObject())
								relObject = objectId.getLogicalObject();
						}
					}
				}

				if (relObject != null)
					return relObject.getObjectId();
			}
		}

		return null;
	} // end method getObjectId

	/**
	 * Returns an object of the given class from a value of this fieldInfo. Supported conversions
	 * are the following :
	 *
	 * <TABLE>
	 *   <TR>
	 *     <TD>LyObjectId</TD>
	 *     <TD>LyObjectId, String</TD>
	 *   </TR>
	 *   <TR>
	 *     <TD>LyObjectId[]</TD>
	 *     <TD>LyObjectId[], String[], Object[], Vector</TD>
	 *   </TR>
	 * </TABLE>
	 *
	 * @exception LyUnsupportedFieldBindingException If the requested conversion if none of the
	 *                                               supported conversions.
	 */
	public Object getPeerValue(Object value, Class<?> objectClass, LyObject instance)
	{
	    value = encodePeerValue(value, objectClass);

		if (value == null)
			return null;

		if (objectClass.isAssignableFrom(value.getClass()))
			return value;

		if (value.getClass() == LyObjectId.class)
		{
			Object object = getObject((LyObjectId)value, objectClass, instance);

			return object;
		}
		else if (value.getClass() == (LyObjectId[].class))
		{
			if (objectClass.isArray())
			{
				LyObjectId[] list = (LyObjectId[])value;
				int			 i, n = list.length;
				Object[]     array;

				if (objectClass == (String[].class))
					array = new String[n];
				else if (objectClass == (Integer[].class))
					array = new Integer[n];
				else if (objectClass == (LyObjectId[].class))
					array = new LyObjectId[n];
				else
					array = (Object[])Array.newInstance(objectClass.getComponentType(), n);

				for (i = 0; i < n; i++)
					array[i] = getObject(list[i], objectClass.getComponentType(), instance);

				return array;
			}
			else if (objectClass == String.class)
			{
	            LyObjectId[] list = (LyObjectId[])value;
	            int          i, n = list.length;
	            StringBuffer buffer = new StringBuffer();
	            Object       object;

	            for (i = 0; i < n; i++)
	            {
	                if (i > 0)
	                    buffer.append(LyApplication.VISIBLE_ARRAY_SEP);
	                object = getObject(list[i], objectClass.getComponentType(), instance);
	                buffer.append(object);
	            }

	            return buffer.toString();
			}
	        else // if (objectClass.isAssignableFrom(Vector.class))
            {
                Collection<Object>      res = new Vector<Object>();
                LyObjectId[]            list = (LyObjectId[])value;
                int                     i, j, n = list.length;
                Object                  temp = null;
                LyInfoList<LyClassInfo> classInfoList;
                LyClassInfo             classInfo;

                for (i = 0; i < n; i++)
                {
                    if (objectClass.getComponentType() != null)
                        temp = getObject(list[i], objectClass.getComponentType(), instance);
                    else
                    {
                        classInfoList = getClassInfos();
                        for (j = 0; j < classInfoList.getSize(); j++)
                        {
                            classInfo = classInfoList.getInfo(j);
                            if (classInfo.getPeerClass() != null)
                                temp = getObject(list[i], classInfo.getPeerClass().getObjectClass(), instance);
                            else
                                temp = getObject(list[i], null, instance);
                            if (temp != null)
                                break;
                        }
                    }

                    if (temp != null)
                        res.add(temp);
                    /*else
                        throw new LyUnsupportedFieldBindingException("BF AppEngine : can't create a peer value of "
                            + objectClass + " for relation " + this + " from given value " + value + " ("
                            + value.getClass() + ") of instance " + instance); */
                }

                return res;
            }
		}

        return super.getPeerValue(value, objectClass, instance);
	} // end method getPeerValue

	/**
	 * Internal method to build a peer value from a LyObject.
	 *
	 * @param  objectId    Object id to convert to a peer value.
	 * @param  objectClass Requested object class for the reconstructed object.
	 * @param  instance    Instance of LyObject from which the value is built.
	 * @return The value object if it could be built, null otherwise.
	 */
	Object getObject(LyObjectId objectId, Class<?> objectClass, LyObject instance)
	{
		if (objectId == null)
			return null;

		if (objectClass == LyObjectId.class)
			return objectId;

		LyObject object = objectId.getLogicalObject();
		String   id;

		if (object != null)
		{
			if (objectClass == null)
			    return object.getPeer();
			else
			{
		        LyClassInfo classInfo = object.getClassInfo();
		        LyClass     peerClass = classInfo.getPeerClass();
				Object      peer = object.getPeer();

				if ((peerClass != null)
				        && objectClass.isAssignableFrom(peerClass.getObjectClass()))
					return peer;
				else if (peer != null)
				{
					id = peerClass.getObjectId(peer).toString();

					Object resu = getObjectFromId(id, objectClass, instance);

					if (resu != null)
						return resu;
				}
			}
		}
		else if (objectClass == null)
		    return null;

		// object or peer class is null. The classinfo associated to object (LyObject)
		// has no binding and is a local class
		// Try to do our best
		id = objectId.toString();
		return getObjectFromId(id, objectClass, instance);
	} // end method getObject

	/**
	 * Internal method to build a peer value from an object id.
	 *
	 * @param  id          Id to convert to a peer value.
	 * @param  objectClass Requested object class for the reconstructed object.
	 * @param  instance    Instance of LyObject from which the value is built.
	 * @return The value object if it could be built, null otherwise.
	 */
	Object getObjectFromId(String id, Class<?> objectClass, LyObject instance)
	{
		if (id == null)
			return null;

		if (objectClass == String.class)
			return id;

		if (Number.class.isAssignableFrom(objectClass))
		{
			// In some case, the ID of an object is a Number.
			Number result;

			try
			{
				if (objectClass == Integer.class)
					result = Integer.decode(id);
				else if (objectClass == Short.class)
					result = Short.decode(id);
				else if (objectClass == Long.class)
					result = Long.decode(id);
				else if (objectClass == Float.class)
					result = Float.valueOf(id);
				else if (objectClass == Double.class)
					result = Double.valueOf(id);
				else
					result = null;
			}
			catch (NumberFormatException e)
			{
				return null;
			}

			return result;
		}
		else if (objectClass == Date.class)

			// In some case, the ID of an object is a Date.
			return LyDateFormatter.fromString(id);

		// XXX : What else can we do ?
		return null;
	} // end method getObjectFromId

	/**
	 * Returns a table of objects for the given classes from a value of this fieldInfo.
	 *
	 * @param     value         Value of this fieldInfo.
	 * @param     objectClasses Expected classes of the physical values.
	 * @param     instance      Instance containing the given value (could be null).
	 * @exception LyUnsupportedFieldBindingException Thrown if we don't known how to build the peer
	 *                                               values from the given parameters.
	 */
	public Object[] getPeerValues(Object value, Class<?>[] objectClasses, LyObject instance)
	{
        Object[]   values = encodePeerValues(value, objectClasses);

        if (values != null)
            return values;

		if ((value == null)
		        || (objectClasses == null))
			return null;

		int   i, n = objectClasses.length;

		if (value.getClass() == LyObjectId.class)
		{
			String[] ids = ((LyObjectId)value).toStrings();

			if (ids.length == n)
			{
				Object[] objects = new Object[n];

				for (i = 0; i < n; i++)
					objects[i] = getObjectFromId(ids[i], objectClasses[i], instance);

				if (objects != null)
					return objects;
			}
		}

        return super.getPeerValues(value, objectClasses, instance);
	} // end method getPeerValues

	/**
	 * Accessor to the relation type.
	 *
	 * @return Choice type.
	 */
	public Type getType()
	{
		return _type;
	}

	/**
	 * Returns the relation that is extended by this relation.
	 *
	 * @return Relation extended by this relation, null otherwise.
	 */
	public final LyRelationFieldInfo getExtendedRelation()
	{
		return _extendedRelation;
	}

	/**
	 * Accessor to the abstract property.
	 *
	 * @return True if this field is abstract.
	 */
	public boolean isAbstract()
	{
		return _abstract;
	}

	/**
	 * Accessor to the multiple property.
	 *
	 * @return True if this field may have several values.
	 */
	public final boolean isMultiple()
	{
		return _multiple;
	}

	/**
	 * Accessor to the ordered property.
	 *
	 * @return True if this field is ordered.
	 */
	public final boolean isOrdered()
	{
		return _ordered;
	}

	/**
	 * Accessor to the noContext property.
	 *
	 * @return True if this relation field does not use the context of the views to compute the list
	 *         of logical objects that may be used as values.
	 */
	public final boolean ignoresContext()
	{
		return _noContext;
	}

	/**
	 * Accessor to the noContext property.
	 *
	 * @param noContext True if this relation field does not use the context of the views to compute
	 *                  the list of logical objects that may be used as values.
	 */
	void setIgnoresContext(boolean noContext)
	{
		_noContext = noContext;
	}

	/**
	 * Accessor to the noToolbar property.
	 *
	 * @return True if this relation doesn't have a toolbar associated to it in edit forms.
	 */
	public final boolean isToolbarDisabled()
	{
		return _noToolbar;
	}

	/**
	 * Accessor to the noToolbar property.
	 *
	 * @param noToolbar True if this relation doesn't have a toolbar associated to it in edit forms.
	 */
	void setToolbarDisabled(boolean noToolbar)
	{
		_noToolbar = noToolbar;
	}

	/**
	 * Return the default value or null if no default is specified.
	 *
	 * @return Default value. If the default value specified for this field in data description file
	 *         starts with $, the returned value is the name of the variable following this
	 *         character. Otherwise, the returned value is an instance of the class given by
	 *         getValueClass.
	 * @see    #getValueClass
	 */
	public Object getDefaultValue()
	{
		if (_defaultValue != null)
			return _defaultValue;

		if ((_default != null) && (_default.startsWith("$")))
			return _default;

		try
		{
			_defaultValue = parse(_default);
			return _defaultValue;
		}
		catch (Exception e)
		{
			_application.getEnvironment().writeException(e);
		}

		return null;
	}

	/**
	 * Return a string representation of the default value or null if no default is specified.
	 *
	 * @return Default value.
	 */
	public String getDefault()
	{
		return _default;
	}

	/**
	 * Gets the direction of the relation.
	 *
	 * @return Direction of this relation field. One of the constants named Direction.XXX.
	 */
	public final Direction getDirection()
	{
		return _direction;
	}

	/**
	 * Gets the class field of this relation. The class field is used to restrict the number of
	 * objects proposed to the end-user when he has to enter values corresponding to this field. The
	 * class field refers to an other field that gives at one time the enabled class(es) for this
	 * relation.
	 *
	 * @return Class field of this relation (may be null)
	 */
	public final LyFieldInfo getClassField()
	{
		return _classField;
	}

	/**
	 * Gets the context field of this relation. The context field is used to restrict the number of
	 * objects proposed to the end-user when he has to enter values corresponding to this field. The
	 * context field refers to an other field of the same class that gives at one time a set of
	 * objects that is used to calculate via cross references the list of enabled objects for this
	 * relation.
	 *
	 * @return Context field of this relation (may be null)
	 */
	public final LyRelationFieldInfo getContextField()
	{
		return _contextField;
	}

	/**
	 * Test if this relation has the specified classInfo in its related classes.
	 *
	 * @param  classInfo Class info to test.
	 * @return True if the relation has the specified class in its related classes, false otherwise.
	 */
	public boolean relatedWith(LyClassInfo classInfo)
	{
		if (_classes.contains(classInfo))
			return true;

		// Now we must tests if the classInfo inherits from one class in the list or
		// if one class of the list inherits from the classInfo.
		LyClassInfo cls;
		int		    i, n = _classes.getSize();

		for (i = 0; i < n; i++)
		{
			cls = _classes.getInfo(i);
			if ((classInfo.inheritsFrom(cls)) || (cls.inheritsFrom(classInfo)))
				return true;
		}

		return false;
	}

	/**
	 * Retrieves the classInfo at the specified index associated to this relation field.
	 *
	 * @param  idx index of the requested classInfo.
	 * @return the reference to the instance of classInfo, or null if there is no classInfo for the
	 *         specified index.
	 */
	public final LyClassInfo getClassInfo(int idx)
	{
		load();
		//checkCMSType();
		return _classes.getInfo(idx);
	}

//	public void checkCMSType()
//	{
//		if ((_cmsTypeChecked) || (getApplication() == null) || (!getApplication().isValid())
//			|| getApplication().isClosing())
//			return;
//
//		// CMS Type detection
//		_cmsTypeChecked = true;
//
//		String cmsType = (String)this.getData("CMS_TYPE");
//
//		if (cmsType != null)
//		{
//			LyClassInfo parentClassInfo = _classes.getInfo(0);
//
//			if (parentClassInfo.getPeerClass() == null)
//				return;
//
//			LyDataSource   dataSource = parentClassInfo.getPeerClass().getDataSource();
//			LyDataProvider provider = dataSource.getDataProvider();
//
//			if (provider instanceof CmsDataProvider)
//			{
//				final CmsRequestEncoder encoder = (CmsRequestEncoder)provider.getRequestEncoder(dataSource);
//
//				LyClassInfo resultClassInfo = encoder.createClassInfoFromType(cmsType);
//
//				if ((resultClassInfo != null) && (resultClassInfo != parentClassInfo))
//				{
//
//					_classes = new LyInfoDataList<LyClassInfo, LyFilter>();
//					_classes.addInfo(resultClassInfo);
//					_classes.lock();
//				}
//			}
//
//		}
//	}

	/**
	 * Get filter associated to the given classInfo in this field.
	 *
	 * @param classInfo
	 *            ClassInfo for which we look for an associated filter.
	 * @return The filter associated to the given classInfo, or null if no
	 *         filter has been specified for this classInfo in this relation.
	 */
	public LyFilter getFilter(LyClassInfo classInfo)
	{
		if (!_classes.contains(classInfo))
		{
			// check if the parent is in this list
			LyClassInfo parent = null;

			for (int i = 0; i < _classes.getSize(); i++)
			{
				parent = _classes.getInfo(i);

				if (classInfo.inheritsFrom(parent))
					return _classes.getData(parent).getFilter();
			}
		}

		return _classes.getData(classInfo).getFilter();
	}

	/**
	 * Retrieves the filter at the specified index.
	 *
	 * @param  idx Index of the requested filter.
	 * @return Filter at the given position, or null if there is no filter for the specified index.
	 */
	public final LyFilter getFilter(int idx)
	{
	    LyClassRef     classRef = _classes.getData(idx);

	    return (classRef == null ? null : classRef.getFilter());
	}

	/**
	 * Gets number of class infos allowed for this relation.
	 *
	 * @return Number of allowed classes.
	 */
	public final int getClassInfoCount()
	{
		load();
		return _classes.getSize();
	}

	/**
     * Returns the list of target classes available on this relation field.
     *
     * @return List of target classes of this relation (never returns null).
     */
    public final Vector<LyClassRef> getClasses()
    {
        load();
        //checkCMSType();

        int                 i, n = _classes.getSize();
        Vector<LyClassRef>  classes = new Vector<LyClassRef>(n);

        for (i = 0; i < n; i++)
            classes.add(_classes.getData(i));

        return classes;
    }

	/**
	 * Returns the list of target classes available on this relation field.
	 *
	 * @return List of target classes of this relation (never returns null).
	 */
	public final LyInfoDataList<LyClassInfo, LyFilter> getClassInfos()
	{
		load();

        int                                    i, n = _classes.getSize();
        LyInfoDataList<LyClassInfo, LyFilter>  classes = new LyInfoDataList<LyClassInfo, LyFilter>();

        for (i = 0; i < n; i++)
            classes.addInfo(_classes.getInfo(i), _classes.getData(i).getFilter());

		return classes;
	}

	/**
	 * Retrieves all the filters associated to the classes of this relation, this is a COPY of the
	 * internal list.
	 *
	 * @return List of filters associated to target classes of this relation.
	 */
	public LyInfoList<LyFilter> getFilters()
	{
		LyInfoList<LyFilter> filters = new LyInfoList<LyFilter>();
		LyFilter			 filter;
		int					 i, n = getClassInfoCount();

		for (i = 0; i < n; i++)
		{
			filter = getFilter(i);
			if (filter != null)
				filters.addInfo(filter);
		}

		return filters;
	}

	/**
	 * Return the number of routes used to calculate values of this relation.
	 *
	 * @return Number of routes of this relation.
	 */
	public final int getRoutesCount()
	{
		return ((_routes == null) ? -1 : _routes.getSize());
	}

	/**
	 * Gets the route at the specified index in routes list.
	 *
	 * @param  index Index of the searched route.
	 * @return The requested route, or null if it could not be found.
	 */
	public final LyRoute getRoute(int index)
	{
		return _routes.getInfo(index);
	}

	/**
	 * Gets the route with the specified id in routes list of this relation.
	 *
	 * @param  id Id of the searched route.
	 * @return The requested route, or null if it could not be found.
	 */
	public final LyRoute getRoute(String id)
	{
		return _routes.getInfo(id);
	}

	/**
	 * Returns the set of routes available of this relation.
	 *
	 * @return Set of routes of this relation. Never returns null.
	 */
	public final LyInfoList<LyRoute> getRoutes()
	{
		return _routes;
	}

	/**
	 * Adds a new route to the routes list at a given position.
	 *
	 * @param route    Route that should be added
	 * @param position Position of the added route in the list.
	 */
	void addRoute(LyRoute route, int position)
	{
		if (_routes == null)
			_routes = new LyInfoList<LyRoute>();

		_routes.addInfo(route, position);
	}

	/**
	 * Accessor of informations about the daemon to manage this relation.
	 *
	 * @return The object that contains informations about the daemon to manage this relation.
	 */
	public final LyInfoDaemon getDaemon()
	{
		return _daemon;
	}

	/**
	 * Gets the preferred number of rows shown to the user at the same time. Not used for non
	 * multiple relations.
	 *
	 * @return Number of rows, or -1 if no preference is given.
	 */
	public final int getNRows()
	{
		return _nRows;
	}

	/**
	* Gets the preferred number of columns shown to the user at the same time.
	* @return Number of cols, or -1 if no preference is given.
	*/
	public final int getNCols()
	{
		return _nCols;
	}

	/**
	 * Gets the min number of required values for this field info. Not used for non multiple
	 * relations.
	 *
	 * @return Number of requiered values.
	 */
	public final int getMin()
	{
		return _min;
	}

	/**
	 * Gets the max number of accepted values for this field info. Not used for non multiple
	 * relations.
	 *
	 * @return Number of accepted values.
	 */
	public final int getMax()
	{
		return _max;
	}

	/**
	 * Gets sub-types (list of relations).
	 *
	 * @return Gets sub-types of this relation.
	 */
	public final LyInfoList<LyRelationFieldInfo> getSubTypes()
	{
		return _subTypes;
	}

	/**
	 * Gets value of this field as an objects list for the given instance.
	 *
	 * @param  object Instance for which the value is requested as an objects list.
	 * @return Objects list corresponding to the value of this field in the given instance (may be
	 *         null).
	 */
	public LySimpleObjectList getObjects(LyObject object)
	{
		if (object == null)
			return null;

		Object value = object.getValue(this);

		return getObjects(value);
	}

	/**
	 * Converts the given value of this field to an objects list.
	 *
	 * @param  value Value to convert to an objects list. This value must be an instance of the
	 *               class given by getValueClass.
	 * @return Objects list corresponding to the given value of this field (may be null).
	 */
	public LySimpleObjectList getObjects(Object value)
	{
		if (value == null)
			return null;

		if (value.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class " + value.getClass() + " expected "
				+ getValueClass());

		if (value instanceof LyObjectId)
		{
			LyObject object = ((LyObjectId)value).getLogicalObject();

			return ((object == null) ? null : new LySimpleObjectList(object));
		}
		else if (value instanceof LyObjectId[])
		{
			LyObjectId[]	   objectIds = (LyObjectId[])value;
			int				   i, n = objectIds.length;
			LyObject		   object;
			LySimpleObjectList list = new LySimpleObjectList();

			// List is not sorted to improve performance
			// Ordered relattions need objects in the same order as in value
			// and Non ordered relations don't care about order
			list.setSorted(false);

			for (i = 0; i < n; i++)
			{
				if (objectIds[i] == null)
					continue;

				object = objectIds[i].getLogicalObject();
				if (object != null)
					list.addObject(object);
			}

			return list;
		}

		return null;
	} // end method getObjects

	/**
	 * Accessor for the showAs property.
	 *
	 * @return The type of presentation used for this relation. One of the SHOW_AS_XXX value.
	 */
	public ShowAs getShowAs()
	{
		return _showAs;
	}

	/**
	 * Accessor for the showAs property.
	 *
	 * @param showAs The type of presentation used for this relation. One of the SHOW_AS_XXX value.
	 */
	void setShowAs(ShowAs showAs)
	{
		_showAs = showAs;
	}

	/**
	 * Checks if this relation inherits from the given one.
	 *
	 * @param  relation Checked relation.
	 * @return True if this relation inherits from the given one, false otherwise.
	 */
	public boolean inheritsFrom(LyRelationFieldInfo relation)
	{
		if (_extendedRelation == relation)
			return true;
		else if (_extendedRelation == null)
			return false;
		else
			return _extendedRelation.inheritsFrom(relation);
	}

	/**
	 * This a convenient method for the parser to reset the classInfo list of this package. This
	 * method must not become public.
	 */
	void resetClasses()
	{
		_classes.resetInfos();
	}

	/**
	 * Adds a (classInfo, filter) to this relation.
	 *
	 * @param classInfo ClassInfo to add to this relation.
	 * @param filter    Filter used to reduce values offered to the end user on this classInfo
	 * @param pos       Position in the list of classes.
	 */
	void addClassInfo(LyClassInfo classInfo, LyFilter filter, int pos)
	{
	    if (_classes.contains(classInfo)
	            && (_classes.getData(classInfo).getFilter() == filter))
	        return;

	    LyClassRef classRef = new LyClassRef(classInfo, filter);

	    classRef.setParent(this);
		_classes.addInfo(classInfo, classRef, pos);
	}

	/**
	 * Removes a pair (classInfo, filter) from this relation.
	 *
	 * @param classInfo ClassInfo to remove from this relation.
	 * @param filter    Filter associated to the classInfo.
	 */
	void removeClassInfo(LyClassInfo classInfo, LyFilter filter)
	{
        int     i, n = _classes.getSize();

        for (i = n - 1; i >= 0; i--)
        {
            if ((_classes.getInfo(i) == classInfo)
                && (_classes.getData(i).getFilter() == filter))
            {
                _classes.removeInfoAt(i);
                return;
            }
        }
	}

	/**
	 * Sets the abstract property.
	 *
	 * @param abstractRelation True if this field is abstract.
	 */
	void setAbstract(boolean abstractRelation)
	{
		_abstract = abstractRelation;
	}

	/**
	 * Sets the multiple property.
	 *
	 * @param multiple True if this field may have several values.
	 */
	void setMultiple(boolean multiple)
	{
		_multiple = multiple;
	}

	/**
	 * Sets the ordered property.
	 *
	 * @param ordered True if this field is ordered.
	 */
	void setOrdered(boolean ordered)
	{
		_ordered = ordered;
	}

	/**
	 * Sets the default value as a string.
	 *
	 * @param defaultValue New default value (may be null) for this field.
	 */
	void setDefault(String def)
	{

		// For multiple default values
		if (def != null)
		{
			if (isMultiple())
				def = def.replace(LyApplication.VISIBLE_VALUE_SEP, _application.VALUE_SEP);

			_default = def.intern();
		}
		else
			_default = null;
	}

	/**
	 * Sets the enabled direction for cross references computing.
	 *
	 * @param direction Enabled direction on this field.
	 */
	void setDirection(Direction direction)
	{
		_direction = direction;
	}

	/**
	 * Sets the type of the relation.
	 *
	 * @param type Type of this field.
	 */
	void setType(Type type)
	{
		_type = type;
	}

	/**
	 * Sets the field that gives information about the available classes.
	 *
	 * @param fieldInfo Class fieldInfo.
	 */
	void setClassField(LyFieldInfo classField)
	{
		_classField = classField;
	}

	/**
	 * Sets the field that gives information about the context.
	 *
	 * @param fieldInfo Context fieldInfo.
	 */
	void setContextField(LyRelationFieldInfo contextField)
	{
		_contextField = contextField;
	}

	/**
	 * Removes the specified route from the routes list.
	 *
	 * @param route Route to remove.
	 */
	void removeRoute(LyRoute route)
	{
		_routes.removeInfo(route);
	}

	/**
	 * Resets routes list.
	 */
	void resetRoutes()
	{
		if (_routes != null)
			_routes.resetInfos();
	}

	/**
	 * Sets the field that gives information about the daemon to manage this relation.
	 *
	 * @param daemon The information about the daemon.
	 */
	void setDaemon(LyInfoDaemon daemon)
	{
		_daemon = daemon;
		_binding = null;

		initDaemon(null);
	}

	/**
	 * Sets the prefered number of rows shown to the user at the same time.
	 *
	 * @param nRows Number of rows, or -1 if no preference.
	 */
	void setNRows(int nRows)
	{
		_nRows = nRows;
	}

	/**
	* Sets the preferred number of columns shown to the user at the same time.
	* @param nCols Number of columns, or -1 if no preference.
	*/
	void setNCols(int nCols)
	{
		_nCols = nCols;
	}

	/**
	 * Sets the min number of required values for this field info. Not used for non multiple
	 * relations.
	 *
	 * @param min Min. number of requiered values.
	 */
	void setMin(int min)
	{
		_min = min;

		if ((min == 0) && (!hasMark(MARK_OPTIONAL)))
		{
			if (_marks == null)
				_marks = new BitSet();

			_marks.set(MARK_OPTIONAL);
			getEnvironment().logWarning("BF AppEngine : add mark OPTIONAL for the fieldInfo " + this
				+ " which has its MIN element set to 0.");
		}
	}

	/**
	 * Sets the max number of accepted values for this field info. Not used for non multiple
	 * relations.
	 *
	 * @param max Max number of accepted values.
	 */
	void setMax(int max)
	{
		_max = max;
	}

	/**
	 * Complete relation.
	 *
	 * @param noContext Indicates if the relation field ignores the context of the views to compute
	 *                  available values.
	 * @param noToolbar Indicates if this field doesn't accept a toolbar.
	 * @param showAs    Gives the type of presentation used for this relation. One of the
	 *                  SHOW_AS_XXX value.
	 */
	void completeRelation(boolean noContext, boolean noToolbar, ShowAs showAs)
	{
		_noToolbar = noToolbar;
		_noContext = noContext;
		_showAs = showAs;
	}

	/**
	 * Internal method to initialize the internal structures managed by the daemon. Only unitialized
	 * join daemons are concerned by this method.
	* @param parent Object info in which the field daemon is intialized.
	 */
	void initDaemon(LyObjectInfo parent)
	{
		if (_daemon == null)
			return;
		if (!(_daemon instanceof LyJoinDaemon))
			return;
		if (_daemonInitialized)
			return;

		int n = getClassInfoCount();

		if (n <= 0)
			return; // Maybee not fully initialized when parsing meta model

		_daemonInitialized = true;

		LyJoinDaemon joinDaemon = (LyJoinDaemon)_daemon;

		if (joinDaemon.getJoinClass() != null)
			return;

		// Create internal class info with a field info per fieldBindng.
		String id = getId();

		LyClassInfo classInfo = new LyClassInfo("__" + id + ".joinClass", null, null, null, null, false, false, null);

		classInfo.setBinding(joinDaemon.getClassBinding());
		classInfo.setCachePolicy(new LyInfoCache(LyInfoCache.CachePolicy.AUTO, null));

		// Create field info for the source foreign key.
		LyRelationFieldInfo sourceFieldInfo = new LyRelationFieldInfo("__" + id + ".sourceFk", null, null, null, false,
				false, LyRelationFieldInfo.Direction.BOTH, LyRelationFieldInfo.Type.ASSOCIATION, null, false,
				false, LyRelationFieldInfo.ShowAs.DEFAULT);
		BitSet			    marks;

		marks = new BitSet();
		marks.set(LyFieldInfo.MARK_HIDDEN);
		marks.set(LyFieldInfo.MARK_NOT_NULL);
		marks.set(LyFieldInfo.MARK_ID);
		sourceFieldInfo.setMarks(marks);

		LyClassInfo parentClassInfo = null;

		if (parent instanceof LyClassInfo)
			parentClassInfo = (LyClassInfo)parent;
		else if (getParent() instanceof LyStructFieldInfo)
			parentClassInfo = ((LyStructFieldInfo)getParent()).getClassInfo();
		else if (getParent() instanceof LyClassInfo)
			parentClassInfo = (LyClassInfo)getParent();

		if (parentClassInfo != null)
			sourceFieldInfo.addClassInfo(parentClassInfo, null, -1);

		classInfo.addFieldInfo(sourceFieldInfo, -1);
		sourceFieldInfo.setApplication(_application);
		sourceFieldInfo.setParent(classInfo);
		sourceFieldInfo.setBinding(joinDaemon.getSourceBinding());

		// Create field info for the target foreign key.
		LyRelationFieldInfo targetFieldInfo = new LyRelationFieldInfo("__" + id + ".targetFk", null, null, null, false,
				false, LyRelationFieldInfo.Direction.BOTH, LyRelationFieldInfo.Type.ASSOCIATION, null, false,
				false, LyRelationFieldInfo.ShowAs.DEFAULT);

		marks = new BitSet();
		marks.set(LyFieldInfo.MARK_HIDDEN);
		marks.set(LyFieldInfo.MARK_NOT_NULL);
		marks.set(LyFieldInfo.MARK_ID);
		targetFieldInfo.setMarks(marks);

		LyClassInfo targetClassInfo;
		LyFilter    filter;

		for (int i = 0; i < n; i++)
		{
			targetClassInfo = getClassInfo(i);
			filter = getFilter(targetClassInfo);
			targetFieldInfo.addClassInfo(targetClassInfo, filter, -1);
		}

		classInfo.addFieldInfo(targetFieldInfo, -1);
		targetFieldInfo.setApplication(_application);
		targetFieldInfo.setParent(classInfo);
		targetFieldInfo.setBinding(joinDaemon.getTargetBinding());

		classInfo.setApplication(_application);
		classInfo.setParent(this);
		joinDaemon.setJoinClass(classInfo);
		joinDaemon.setSourceRelation(sourceFieldInfo);
		joinDaemon.setTargetRelation(targetFieldInfo);
	} // end method initDaemon

	/**
	 * Adds a sub-type to this relation.
	 *
	 * @param relation Sub type.
	 */
	void addSubType(LyRelationFieldInfo relation)
	{
		if (relation.isAbstract())
			throw new IllegalStateException("BF AppEngine : "
				+ ((this instanceof LyStructFieldInfo) ? "struct" : "relation") + " \"" + getId()
				+ "\" can't have an abstract subType : \"" + relation.getId() + "\"");

		if (_subTypes == null)
		{
			_subTypes = new LyInfoList<LyRelationFieldInfo>();
			resetClasses();
		}

		_subTypes.addInfo(relation);

		LyInfoList<LyClassInfo> classInfos = getClassInfos();
		LyInfoList<LyClassInfo> relationClassInfos = relation.getClassInfos();
		int					    i, n = relationClassInfos.getSize();
		LyRelationFieldInfo     model = (LyRelationFieldInfo)_template;
		LyInfoList<LyClassInfo> subClassInfos;
		LyClassInfo			    classInfo;

		for (i = 0; i < n; i++)
		{
			classInfo = relationClassInfos.getInfo(i);
			if ((model == null) || !model.relatedWith(classInfo))
				throw new IllegalStateException("BF AppEngine : can't sub-type relation field " + this + " with class "
					+ classInfo + " because its model refers neither to this class nor to one of its super-class");

			if (!classInfos.contains(classInfo))
			{
				subClassInfos = classInfo.getSubClasses(true);

				// Add class info to relation
				if ((subClassInfos != null) && (subClassInfos.contains(classInfos)))
					addClassInfo(classInfo, null, 0);
				else
					addClassInfo(classInfo, null, -1);
			}
		}
	} // end method addSubType

	/**
	 * Checks if the given relation is a sub-type of this relation.
	 *
	 * @param  relation Sub type.
	 * @return True if the given relation is a sub-type of this relation, false otherwise.
	 */
	public boolean hasSubType(LyRelationFieldInfo relation)
	{
		return ((_subTypes == null) ? false : _subTypes.contains(relation));
	}

    /**
     * Sets the relation that is extended by this relation.
     *
     * @param extendedRelation Relation extended by this relation, null otherwise.
     */
    protected void setExtendedRelation(LyRelationFieldInfo extendedRelation)
    {
        _extendedRelation = extendedRelation;
    }

    /**
     * Sets the list of target classes available on this relation field.
     *
     * @param classInfos List of target classes of this relation.
     */
    protected void setClasses(Vector<LyClassRef> classes)
    {
        _classes = new LyInfoDataList<LyClassInfo, LyClassRef>();

        int         i, n = (classes == null ? 0 : classes.size());
        LyClassRef  classRef;

        for (i = 0; i < n; i++)
        {
            classRef = classes.get(i);
            _classes.addInfo(classRef.getClassInfo(), classRef);
        }
    }

    /**
     * Sets the list of routes available on this relation field.
     *
     * @param routes List of target classes of this relation.
     */
    protected void setRoutes(LyInfoList<LyRoute> routes)
    {
        _routes = new LyInfoList<LyRoute>(routes);
    }

    /**
     * Sets sub-types (list of relations).
     *
     * @param subTypes sub-types of this relation.
     */
    protected void  setSubTypes(LyInfoList<LyRelationFieldInfo> subTypes)
    {
        int     i, n = (subTypes == null ? 0 : subTypes.getSize());

        _subTypes = null;
        for (i = 0; i < n; i++)
            addSubType(subTypes.getInfo(i));
    }

	/**
	 * Returns the reverse relation of this relation.
	 *
	 * @return Reverse relation of this relation.
	 */
	public final LyRelationFieldInfo getReverseRelation()
	{
		if (_daemon instanceof LyReverseDaemon)
			return ((LyReverseDaemon)_daemon).getReverseRelation();
		else
			return null;
	}
} // end class LyRelationFieldInfo
