/*
 * LyRoute.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;
import leon.app.*;


/**
 * This class implements a route between two classes. A route defines how to get objects of a class
 * from a list of objects belonging to an other class (cross references). A route is defined in a
 * source class, has a sequence of steps and a target class. A route can be either explicit (defined
 * by programmer in data description file) or implicit (calculated automatically using relations and
 * enabled crossings).
 *
 * <p>Routes are said reverse if computing should start from target class and returning to source
 * class (this kind of computing is used for optimization purpose). A reverse route contains no
 * steps.</p>
 *
 * <p>Routes are said bidirectional if they can be used in both directions (from source to target or
 * from target to source).<br>
 * A route can be complete if the sequence of steps leads to the target class or partial if it gives
 * just a part of the path from to source class to the target class. In this case, the route will be
 * completed by other routes starting from the last class reached by the steps and leading to the
 * target class.</p>
 *
 * <p>Only complete routes can be bidirectional.<br>
 * See also tag <ROUTE> of leon.dtd.</p>
 *
 * @see LyStep
 * @see LyClassInfo
 */
public class LyRoute extends LyInfo
{
	public static final String __VERSION = "$Id: LyRoute.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The target class. */
	protected LyClassInfo	   _target;

	/** The set of steps. */
	protected Vector<LyStep>   _steps;

	/** Indicates if this route is reverse or not. */
	protected boolean		   _reverse;

	/** Indicates if this route is bi-directional or not. */
	protected boolean		   _bidirectional;

	/** Indicates if this route is used for computing X-Refs or not. */
	protected boolean		   _hidden;

	/** Indicates if this route was generated automatically by Leonardi. */
	protected boolean		   _generated;

	/**
	 * Init method.
	 */
	protected void init()
	{
		super.init();

		_target = null;
		_steps = null;
		_reverse = false;
		_bidirectional = false;
		_hidden = false;
		_generated = false;
	}

	/**
	 * Constructor of a route.
	 *
	 * @param id      Identifier of the route.
	 * @param target  Target class.
	 * @param steps   Sequence of steps that define this route.
	 * @param reverse Boolean indicating if this route is reverse or not.
	 * @param reverse Boolean indicating if this route is bidirectional or not.
	 */
	LyRoute(String id, LyClassInfo target, Vector<LyStep> steps, boolean reverse,
		boolean bidirectional, boolean hidden)
	{
		super(id);

		_target = target;
		_steps = steps;
		_reverse = reverse;
		_bidirectional = bidirectional;
		_hidden = hidden;
	}

    /**
     * Constructs a route with given id.
     *
     * @param id     Identifier of the route.
     */
    LyRoute(String id)
    {
        super(id);
    }
            
	/**
	 * Constructs a route that if neither reverse, nor bidirectional. This method is a shortcut of
	 * the previous one, with both these parameters set to false.
	 *
	 * @param id     Identifier of the route.
	 * @param target Target class.
	 * @param steps  Sequence of steps that define this route.
	 */
	LyRoute(String id, LyClassInfo target, Vector<LyStep> steps)
	{
		this(id, target, steps, false, false, false);
	}

	/**
	 * Final step of parsing : checks integrity of the route and completes the initialization if
	 * necessary. Checks performed are the following :
	 *
	 * <ul>
	 *   <LI>source class can't be null,</LI>
	 *   <LI>target class can't be null,</LI>
	 *   <LI>if the route is reverse, checks if routes from target class to source class exist,</LI>
	 *   <LI>if the route is not reverse, checks if the sequence of steps if not null, nor empty,
	 *   </LI>
	 *   <LI>if the route is not complete, checks if routes starting from the last class reached by
	 *     steps to the target class, exist,</LI>
	 *   <LI>checks if the route goes to far (contains extra steps after target class has been
	 *     reached),</LI>
	 *   <LI>If the route is bidirectional, checks if all steps are bidirectional and if the route
	 *     is complete.</LI>
	 * </ul>
	 *
	 * @param     source Source class that contains this route.
	 * @exception IllegalStateException If one of these checks fails.
	 */
	void check(LyClassInfo source)
	{
		if (source == null)
			throw new IllegalArgumentException("BF AppEngine : source can't be null");
		if (_target == null)
			throw new IllegalStateException("BF AppEngine : target can't be null "
				+ source + ", " + this);
		if (_reverse)
		{
			LyInfoList<LyRoute> routes = _target.getRoutes(source);

			if ((routes == null) || (routes.getSize() == 0))
				throw new IllegalStateException("BF AppEngine : invalid reversed route "
					+ source + ", " + this);
			return;
		}
		if ((_steps == null) || (_steps.size() == 0))
		{
			if (getClass() == LyRoute.class)
				throw new IllegalStateException("BF AppEngine : steps can't be null "
					+ source + ", " + this);
			// Allowed for paths
			return;
		}

		LyClassInfo current = source;
		LyStep	    step;
		int		    i, n = _steps.size();
		boolean     allowBidirect = true;

		for (i = 0; i < n; i++)
		{
			step = _steps.elementAt(i);

			// Check the step.
			step.check(current);
			allowBidirect |= step.isBidirectional();

			// Continue with the target class.
			current = step.getTarget();
		}

		if ((current != _target) && !current.inheritsFrom(_target))
		{
			LyInfoList<LyRoute> routes = current.getRoutes(_target);

			if ((routes == null) || (routes.getSize() == 0))
				throw new IllegalStateException(
					"BF AppEngine : no valid route defined by steps for route "
					+ source + ", " + this);
		}

		if (_bidirectional)
		{
			if (!allowBidirect
					|| ((current != _target) && !current.inheritsFrom(_target)))
				throw new IllegalStateException("BF AppEngine : route " + this
					+ " can't be bidirectional because it contains a relation that is not bidirectional or because the route is not complete");

			// Create inverse route
			Vector<LyStep> steps = new Vector<LyStep>(n);
			LyStep		   newStep;
			LyApplication  application = getApplication();

			for (i = 0, current = source; i < n; i++)
			{
				step = _steps.elementAt(i);
				newStep = new LyStep(current, step.getRelations(), null);
				newStep.setApplication(application);
				newStep.check(step.getTarget());
				steps.insertElementAt(newStep, 0);
				current = step.getTarget();
			}

			LyRoute route = new LyRoute(null, source, steps, _reverse, false, _hidden);

			route.setApplication(application);
			route.setGenerated(true);
			_target.addRoute(route, -1);
		}
	} // end method check

	/**
	 * Internal method used only by LyObjectList class to check a route defined in a project before
	 * beeing used.
	 *
	 * @param source Source class a the route.
	 */
	public void internalCheck(LyClassInfo source)
	{
		check(source);
	}

	/**
	 * Lock this route.
	 */
	protected void lock()
	{
		super.lock();

		// Lock steps
		LyStep step;
		int    i, n = ((_steps == null) ? 0 : _steps.size());

		for (i = 0; i < n; i++)
		{
			step = _steps.elementAt(i);
			step.lock();
		}
	}

	/**
	 * Accessor to the target class of this route.
	 *
	 * @return Target class of this route.
	 */
	public LyClassInfo getTarget()
	{
		return _target;
	}

	/**
	 * Accessor to the sequence of steps of this route. This is a COPY of the internal list,
	 * therefore this list may be freely manipulated by the application.
	 *
	 * @return Sequence of steps, or null.
	 */
	public Vector<LyStep> getSteps()
	{
		if (_steps == null)
			return null;

		// Return a copy to avoid manipulation of the internal list
		return new Vector<LyStep>(_steps);
	}

	/**
	 * Indicates if this route is reverse or not.
	 *
	 * @return True if it's reverse, false otherwise.
	 */
	public boolean isReverse()
	{
		return _reverse;
	}

	/**
	 * Indicates if this route is bidirectional or not.
	 *
	 * @return True if the route is bidirectional, false otherwise.
	 */
	public boolean isBidirectional()
	{
		return _bidirectional;
	}

	/**
	 * Indicates if this route is used for computing X-Refs or not.
	 *
	 * @return True if the route is bidirectional, false otherwise.
	 */
	public boolean isHidden()
	{
		return _hidden;
	}

	/**
	 * Sets the target of the route.
	 *
	 * @param     classInfo the target to set
	 * @exception IllegalArgumentException If classInfo parameter is null.
	 */
	void setTarget(LyClassInfo classInfo)
	{
		if (classInfo == null)
			throw new IllegalArgumentException("BF AppEngine : classInfo can't be null");

		_target = classInfo;
	}

	/**
	 * Adds a step to this route.
	 *
	 * @param     step Step to add to this route.
	 * @exception IllegalArgumentException If step parameter is null.
	 */
	void addStep(LyStep step)
	{
		if (step == null)
			throw new IllegalArgumentException("BF AppEngine : step can't be null");

		if (_steps == null)
			_steps = new Vector<LyStep>(3);

		_steps.addElement(step);
	}

	/**
	 * Adds the given route to this route. The result consists of adding all steps of the added
	 * route to this route.
	 *
	 * @param     route Route to add to this route.
	 * @exception IllegalArgumentException If route parameter is null or if the route contains no
	 *                                     step.
	 */
	void addRoute(LyRoute route)
	{
		if (route == null)
			throw new IllegalArgumentException("BF AppEngine : route can't be null");
		if (route._steps.size() == 0)
			throw new IllegalArgumentException("BF AppEngine : empty route");

		int    i, n = route._steps.size();
		LyStep step;

		for (i = 0; i < n; i++)
		{
			step = route._steps.elementAt(i);
			addStep(step);
		}
	}

	/**
	 * Returns a string representation of this object.
	 */
	public String dump()
	{
		StringBuffer buffer = new StringBuffer();

		buffer.append("Route ");
		buffer.append(getId());
		buffer.append(", target = ");
		buffer.append(_target);
		buffer.append(" :");

		int    i, n = ((_steps == null) ? 0 : _steps.size());
		LyStep step;

		for (i = 0; i < n; i++)
		{
			step = _steps.elementAt(i);
			buffer.append("\n\t");
			buffer.append(step.toString());
		}

		return buffer.toString();
	}

	/**
	 * Return the generated state of the current route.
	 *
	 * @return true if the route was generated by Leonardi
	 */
	public boolean isGenerated()
	{
		return _generated;
	}

	/**
	 * Set the generated state for the current route*.
	 *
	 * @param generated true if the route was generated by Leonardi
	 */
	protected void setGenerated(boolean generated)
	{
		_generated = generated;
	}

    /**
     * Sets if this route is used for computing X-Refs or not.
     *
     * @param hidden True if the route is bidirectional, false otherwise.
     */
    protected void setHidden(boolean hidden)
    {
        _hidden = hidden;
    }
    
    /**
     * Sets if this route is reverse or not.
     */
    protected void setReverse(boolean reverse)
    {
        _reverse = reverse;
    }

    /**
     * Sets if this route is bidirectional or not.
     */
    protected void setBidirectional(boolean bidirectional)
    {
        _bidirectional = bidirectional;
    }
    
    /**
     * Sets steps of this route.
     */
    protected void setSteps(Vector<LyStep> steps)
    {
        int     i, n = (steps == null ? 0 : steps.size());
        
        _steps = null;
        for (i = 0; i < n; i++)
            addStep(steps.elementAt(i));
    }
} // end class LyRoute
