/*
 * LyRule.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import leon.info.misc.*;

/**
 * This class implements a rule triggered upon choice option modification. Rules are attached to
 * choice options and are triggered when options are checked ou unchecked. They define an operation
 * to perform on a fieldInfo located in the same edition view (creation or modification form). This
 * class extends LyInfo : this means that rules can be identified within an application.<br>
 * A rule contains extra attributes :
 *
 * <ul>
 *   <LI>a fieldInfo, that is impacted by the given rule,</LI>
 *   <LI>an operation, which defines the nature of the impact done on the fieldInfo, Supported
 *     operations are defined as public static named Operation.XXX.</LI>
 *   <LI>a modifier, which allow to invert the operation done (only for operations :
 *     Operation.SET_ENABLED, Operation.SET_EDITABLE, Operation.SET_OPTIONAL).</LI>
 * </ul>
 *
 * @see LyChoiceOption
 * @see LyChoiceFieldInfo
 */
public final class LyRule extends LyInfo
{
	public static final String __VERSION = "$Id: LyRule.java 40680 2012-09-06 06:52:08Z jlsaouli $";

    /** Operations. */
    public enum Operation
    {
    	// Operation performed : set enable/disable fieldInfo input.
    	SET_ENABLED,
    	// Operation performed : set editable/not editable fieldInfo input.
    	SET_EDITABLE,
    	// Operation performed : set optional/mandatory fieldInfo input.
    	SET_OPTIONAL,
    	// Operation performed : clear fieldInfo input.
    	CLEAR,
    	// Operation performed : reset fieldInfo input.
    	RESET,
    	// Operation performed : set fieldInfo with the given value.
    	SET_VALUE,
    	// Operation performed : show the specified field.
    	SHOW,
    	// Operation performed : set the specified field visible.
    	SET_VISIBLE,
    	// Operation performed : set the label of specified field.
    	SET_LABEL
    }

	/**
	 * ObjectInfo impacted by this rule : must be a LyFieldInfo or a LyClassInfo (only for SET_LABEL
	 * operation).
	 */
	LyObjectInfo			   _objectInfo;

	/** Field operation. */
	Operation				   _operation;

	/** Modifier : To specify if rule is triggered on check or on uncheck. */
	boolean					   _onCheck;

	/** Value assigned to the fieldInfo on SET_VALUE operation. */
	Object					   _value;

	/** Filter that triggers rule. */
	LyFilter				   _filter;

	/** Priority of the control. */
	int						   _priority;

	/** Label assigned to the fieldInfo on SET_LABEL operation. */
	LyInfoLabel				   _label;

    /** Parsed formula associated to this field : result of formula expression parsing. */
    LyFormula                  _formula;
    
    /**
     * Constructor of a rule.
     *
     * @param id         Id of the rule (optional).
     */
    LyRule(String id)
    {
        super(id);
    }
    
	/**
	 * Constructor of a rule.
	 *
	 * @param id         Id of the rule (optional).
	 * @param objectInfo FieldInfo impacted by this rule.
	 * @param priority   Priority of the rule.
	 * @param operation  Operation done when the rule is triggered.
	 * @param onCheck    Modifier which defines if the rule is triggered when the option is checked
	 *                   or unchecked.
	 * @param value      Value assigned to the fieldInfo on SET_VALUE operation. Only used for
	 *                   operations named Operation.SET_XXX.
	 * @param modelType  Type of the model where the rule was defined.
	 */
	LyRule(String id, LyObjectInfo objectInfo, int priority,
		Operation operation, boolean onCheck, Object value, LyTag.Model modelType)
	{
		super(id);

		setObjectInfo(objectInfo);
		setPriority(priority);
		setOperation(operation);
		setOnCheck(onCheck);
		// Conversion in the right type for the field info will be done
		// on first use by getValue method
		setInitValue(value);
		setFilter(null);
		setModelType(modelType);
	}

	/**
	 * Final step of parsing : checks integrity of the rule and completes the initialization if
	 * necessary. Checks performed are the following :
	 *
	 * <ul>
	 *   <LI>if the rule is SET_LABEL, the label can't be null,</LI>
	 *   <LI>if the rule is not SET_LABEL and if there is a label, sets the label to null and
	 *     performs a warning,</LI>
	 *   <LI>if the rule is SET_LABEL and source is a LyClassInfo, the label type must be
	 *     GRAMMATICAL,</LI>
	 *   <LI>if the rule is SET_LABEL ans source is a LyChoiceFieldInfo, the label type must be
	 *     GRAMMATICAL, SEMANTIC or NAME.</LI>
	 * </ul>
	 *
	 * @param     source Source that contains this rule : this source is a LyClassInfo or a
	 *                   LyChoiceFieldInfo.
	 * @exception IllegalStateException If checks fails.
	 */
	void check(LyObjectInfo source)
	{
		if ((_operation == Operation.SET_LABEL) && (_label == null))
			throw new IllegalStateException("BF AppEngine : invalid rule on source "
				+ source + " : a SET_LABEL rule must have a label");
		else if ((_operation != Operation.SET_LABEL) && (_label != null))
		{
			_label = null;
			getEnvironment().logWarning("BF AppEngine : label for rule " + this
				+ " on source " + source
				+ " will be ignored because this rule is not a SET_LABEL rule.");
		}
		else if (_operation == Operation.SET_LABEL)
		{
			// We must verify than the label type is compatible with the rule.
            LyInfoLabel.Type type = _label.getType();

			if ((_objectInfo instanceof LyClassInfo)
					&& (type != LyInfoLabel.Type.GRAMMATICAL))
				throw new IllegalStateException("BF AppEngine : invalid label type on rule "
					+ this + " for source " + source);
			else if ((_objectInfo instanceof LyFieldInfo)
					&& (type != LyInfoLabel.Type.GRAMMATICAL)
					&& (type != LyInfoLabel.Type.SEMANTIC)
					&& (type != LyInfoLabel.Type.NAME))
				throw new IllegalStateException("BF AppEngine : invalid label type on rule "
					+ this + " for source " + source);
		}

		if (_label != null)
			_label.check(source);
	} // end method check

	/**
	 * Accessor of the fieldInfo impacted by this rule.
	 *
	 * @return     The fieldInfo concerned by this rule (may be null).
	 * @deprecated Use getObjectInfo instead.
	 */
	public LyFieldInfo getFieldInfo()
	{
		return ((_objectInfo instanceof LyFieldInfo) ? (LyFieldInfo)_objectInfo : null);
	}

	/**
	 * Accessor of the fieldInfo impacted by this rule.
	 *
	 * @return The fieldInfo concerned by this rule.
	 */
	public LyObjectInfo getObjectInfo()
	{
		return _objectInfo;
	}

	/**
	 * Gets priority of the rule.
	 */
	public int getPriority()
	{
		return _priority;
	}

	/**
	 * Gets the operation done on this rule.
	 *
	 * @return Operation defined for this rule.
	 */
	public Operation getOperation()
	{
		return _operation;
	}

	/**
	 * Specifies if the rule is triggered when checking or on unchecking the associated option.
	 *
	 * @return True if set, false if unset.
	 */
	public boolean isOnCheck()
	{
		return _onCheck;
	}

	/**
	 * Accessor of the value used by this rule on SET_VALUE operation.
	 *
	 * @return The value used by this rule on SET_VALUE operation.
	 */
	public Object getValue()
	{
		if ((_value != null)
				&& (_objectInfo instanceof LyFieldInfo))
			return ((LyFieldInfo)_objectInfo).getValueFromObject(_value, null);

		return _value;
	}
	
	/**
     * Accessor of the value used by this rule on SET_VALUE operation.
     *
     * @return The value used by this rule on SET_VALUE operation.
     */
    public final Object getInitValue()
    {
        return _value;
    }
    
	/**
     * Accessor of the value used by this rule on SET_VALUE operation.
     *
     * @return The value used by this rule on SET_VALUE operation.
     */
    public String getStringValue()
    {
        return (_value != null ? _value.toString() : null);
    }

	/**
	 * Accessor for the filter that triggers this rule.
	 */
	public LyFilter getFilter()
	{
		return _filter;
	}

	/**
	 * Accessor of the label used by this rule on SET_LABEL operation.
	 *
	 * @return The label used by this rule on SET_LABEL operation.
	 */
	public LyInfoLabel getLabel()
	{
		return _label;
	}

	/**
	 * Sets the fieldInfo impacted by this rule.
	 *
	 * @param fieldInfo FieldInfo concerned by this rule.
	 */
	void setObjectInfo(LyObjectInfo objectInfo)
	{
		_objectInfo = objectInfo;
	}

	/**
	 * Set the filter that triggers this rule.
	 */
	void setFilter(LyFilter filter)
	{
		_filter = filter;
	}

	/**
	 * Sets the label to use by this rule on SET_LABEL operation.
	 *
	 * @param label The label to use by this rule on SET_LABEL operation.
	 */
	void setLabel(LyInfoLabel label)
	{
		_label = label;
	}

	/**
	 * Specifies if the rule leads to a strong dependency with the field declared on the rule. i.e.
	 * : the modification of the field that defined the rule implies a change on the fieldInfo
	 * concerned by the rule.
	 *
	 * @return True if the dependency is strong, false otherwise.
	 */
	boolean hasStrongDepend()
	{
		return (((_objectInfo instanceof LyFieldInfo)
					&& ((_operation == Operation.SET_ENABLED)
						|| ((_operation == Operation.SET_OPTIONAL) && !_onCheck)))
				|| (_operation == Operation.CLEAR)
				|| (_operation == Operation.SET_VALUE)
				|| ((_operation == Operation.SET_VISIBLE)
					&& !_objectInfo.hasMark(LyFieldInfo.MARK_OPTIONAL)));
	}
    
    /**
     * Sets priority of the rule.
     */
    protected void setPriority(int priority)
    {
        _priority = priority;
    }

    /**
     * Sets the operation done on this rule.
     */
    protected void setOperation(Operation operation)
    {
        _operation = operation;
    }

    /**
     * Specifies if the rule is triggered when checking or on unchecking the associated option.
     */
    protected void setOnCheck(boolean onCheck)
    {
        _onCheck = onCheck;
    }

    /**
     * Sets the value used by this rule on SET_VALUE operation.
     */
    protected void setInitValue(Object value)
    {
        _value = value;
    }
	
    /**
     * Gets the formula (as a string) associated to this number field. The formula is provided to
     * calculate values of this field thanks to a simple mathematic expression using other numerc
     * fields in the same application class.
     *
     * @return Current formula or null if no formula is set.
     */
    public String getFormula()
    {
        return (_formula == null ? null : _formula.getExpression());
    }
    
    /**
     * Gets the formula associated to this control.
     *
     * @return Current formula or null if no formula is set.
     */
    public final LyFormula getLyFormula()
    {
        return _formula;
    }
    
    /**
     * Sets the formula associated to the field.
     *
     * @param     formula Current formula, or null if none is specified.
     * @exception ParseException If the given formula couldn't be parsed successfully.
     */
    void setFormula(String formula)
    {
        _formula = LyFormula.formulaFactory(this, formula);
    }
} // end class LyRule
