/*
 * LySort.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.io.PrintWriter;
import java.util.Vector;

import leon.misc.LyString;


/**
 * This class represents a sort used to compare 2 objects in order to be able to sort objects list.
 * It contains a list of criteria.<br>
 * See also tag <SORT> of leon.dtd
 *
 * @see LySortCriterion
 * @see leon.data.LyObject#compare
 * @see leon.data.LyObjectList
 */
public final class LySort extends LyObjectInfo
{
	public static final String  __VERSION = "$Id: LySort.java 33165 2011-11-16 08:07:09Z nhamon $";

	/** List of criteria of this sort. */
	LyInfoList<LySortCriterion> _criteria;

	/**
	 * Init method.
	 */
	protected void init()
	{
		_criteria = new LyInfoList<LySortCriterion>();
	}

	/**
	 * Init objectInfo attributes from its template.
	 */
	void initFromTemplate()
	{
		if (_template == null)
			return;

		super.initFromTemplate();

		// Copy criteria from template
		LySort template = (LySort)_template;

		_criteria = new LyInfoList<LySortCriterion>(template._criteria);
	}

	/**
	 * Default constructor with to parameters to build a sort.
	 */
	public LySort()
	{
		super();
	}

	/**
	 * Constructs an identified sort from a template.
	 *
	 * @param id       Id of the new sort.
	 * @param template Sort from which criteria are copied.
	 */
	public LySort(String id, LySort template)
	{
		super(id, template);
		//initFromTemplate();
	}

	/**
	 * Clone this sort.
	 */
	public LySort cloneSort()
	{
		LySort sort = new LySort("__" + LyString.getRandomId() + "__", null);

		synchronized (_criteria)
		{
			LySortCriterion criterion;
			int			    i, n = getCriterionCount();

			// Go through the element list
			for (i = 0; i < n; i++)
			{
				criterion = getCriterion(i);
				criterion = criterion.cloneSortCriterion();
				sort.addCriterion(criterion);
				criterion._parent = sort;
			}
		}

		return sort;
	}

	/**
	 * Lock this sort.
	 */
	public void lock()
	{
		super.lock();

		// Lock criteria
		_criteria.lock();
	}
	
	/**
	 * Gets the criterion at the given index in this sort.
	 *
	 * @param     index Index of the searched criterion.
	 * @return    The criterion found at this position in criteria list of this sort.
	 * @exception ArrayIndexOutOfBoundsException If index is negative of greater or equal to the
	 *                                           number of criteria.
	 */
	public LySortCriterion getCriterion(int index)
	{
		return _criteria.getInfo(index);
	}

	/**
	 * Gets number of criteria of this sort.
	 *
	 * @return Number of criteria.
	 */
	public int getCriterionCount()
	{
		return _criteria.getSize();
	}

	/**
	 * Gets a copy of criteria list.
	 *
	 * @return Copy of this sort citeria. Since a copy is made, the returned list can freely be
	 *         modified.
	 */
	public Vector<LySortCriterion> getCriteria()
	{
		return _criteria.getInfos();
	}

	/**
	 * Checks if the sort contains the given field info.
	 *
	 * @param  fieldInfo FieldInfo that is searched.
	 * @return True if the sort contains the given fieldInfo, false otherwise.
	 */
	public boolean contains(LyFieldInfo fieldInfo)
	{
		return (getIndex(fieldInfo) >= 0);
	}
	
	/**
	 * Checks if a sort criterion is found in this sort.
	 *
	 * @param criterion The sort criterion.
	 * @return True if the sort criterion is found, false otherwise.
	 */
	public boolean contains(LySortCriterion criterion)
	{
		return ((_criteria != null) && _criteria.contains(criterion));
	}

	/**
	 * Gets sort criterion corresponding to the given fieldInfo.
	 *
	 * @param  fieldInfo FieldInfo for which a criterion is searched.
	 * @return The criterion found for the given fieldInfo, null if no criterion in this sort
	 *         concerns this fieldInfo.
	 */
	public LySortCriterion getCriterion(LyFieldInfo fieldInfo)
	{
		int index = getIndex(fieldInfo);

		return ((index < 0) ? null : getCriterion(index));
	}

	/**
	 * Gets position of the criterion corresponding to the given fieldInfo.
	 *
	 * @param  fieldInfo FieldInfo for which a criterion is searched.
	 * @return Index of the criterion found for the given fieldInfo, -1 if no criterion in this sort
	 *         concerns this fieldInfo.
	 */
	public int getIndex(LyFieldInfo fieldInfo)
	{
		LySortCriterion criterion;
		int			    i, n = getCriterionCount();

		for (i = 0; i < n; i++)
		{
			criterion = getCriterion(i);
			if (criterion.getFieldInfo() == fieldInfo)
				return i;
		}

		return -1;
	}

	/**
	 * Gets list of fieldInfos that this sort refers to.
	 *
	 * @return List of fieldInfos that the sort depends on.
	 */
	public LyInfoList<LyFieldInfo> getFieldInfos()
	{
		synchronized (_criteria)
		{
			LyInfoList<LyFieldInfo> fieldInfos = new LyInfoList<LyFieldInfo>();
			LySortCriterion		    criterion;
			int					    i, n = getCriterionCount();

			// Go through the element list
			for (i = 0; i < n; i++)
			{
				criterion = getCriterion(i);
				fieldInfos.addInfo(criterion.getFieldInfo());
			}

			return fieldInfos;
		}
	}

    /**
     * Gets list of fieldInfos corresponding to groups.
     *
     * @return List of fieldInfos corresponding to groups, maybe null.
     */
    public LyInfoList<LyFieldInfo> getGroupByFieldInfos()
    {
        synchronized (_criteria)
        {
            LyInfoList<LyFieldInfo> fieldInfos = null;
            LySortCriterion         criterion;
            int                     i, n = getCriterionCount();

            // Go through the element list
            for (i = 0; i < n; i++)
            {
                criterion = getCriterion(i);
                if (criterion.isGroupBy())
                {
                    if (fieldInfos == null)
                        fieldInfos = new LyInfoList<LyFieldInfo>();
                    fieldInfos.addInfo(criterion.getFieldInfo());
                }
                else
                    // Stop as soon as a criterion has not groupBy feature
                    break;
            }

            return fieldInfos;
        }
    }
    
    /**
     * Checks if this sort defines grouping.
     *
     * @return True if groups are defined, false otherwise.
     */
    public boolean hasGroups()
    {
        synchronized (_criteria)
        {
            return (getCriterionCount() <= 0 ? false :  getCriterion(0).isGroupBy());
        }
    }
    
    /**
     * Checks if this sort has fields that are calculated.
     *
     * @return True the sort contains calculated fields, false otherwise.
     */
    public boolean hasCalculatedFields(LyClassInfo classInfo)
    {
        synchronized (_criteria)
        {
            LySortCriterion criterion;
            int             i, n = getCriterionCount();
            LyFieldInfo     fieldInfo;

            // Go through the element list
            for (i = 0; i < n; i++)
            {
                criterion = getCriterion(i);
                fieldInfo = criterion.getFieldInfo();
                if (fieldInfo.isLocal()
                         && (classInfo.getGetMethod(fieldInfo) != null))
                    return true;
            }

            return false;
        }
    }
    
    /**
     * Checks if this sort has at least one criterion with a comparator.
     *
     * @return True the sort contains at least one criterion with a comparator, false otherwise.
     */
    public boolean hasCriterionWithComparator()
    {
        synchronized (_criteria)
        {
            LySortCriterion criterion;
            int             i, n = getCriterionCount();

            // Go through the element list
            for (i = 0; i < n; i++)
            {
                criterion = getCriterion(i);
                
                if (criterion.getJavaName() != null)
                    return true;
            }

            return false;
        }
    }
    
	/**
	 * Writes XML corresponding to this sort element.
	 *
	 * @param pw PrintWriter on which the XML is written.
	 */
	public void writeXml(PrintWriter pw)
	{
		// Write tag with its attributes
		pw.print("<SORT");
		if (getId() != null)
			pw.print(" id='" + getId() + "'");
		else
			pw.print(" id='_sort" + LyString.getXmlString(LyString.getRandomId()) + "'");
		pw.println(">");

		synchronized (_criteria)
		{
			LySortCriterion criterion;
			int			    i, n = getCriterionCount();

			// Go through the element list
			for (i = 0; i < n; i++)
			{
				criterion = getCriterion(i);
				criterion.writeXml(pw);
			}
		}

		pw.println("</SORT>");
	}

	/**
	 * Adds a criterion to this sort.
	 *
	 * @param criterion Criterion that is added to this sort.
	 */
	public void addCriterion(LySortCriterion criterion)
	{
		if (isLocked())
			throw new IllegalStateException("BF AppEngine : sort " + this
				+ " was locked and thus can't be modified anymore");
		if (criterion == null)
			throw new IllegalArgumentException("BF AppEngine : criterion can't be null");

		_criteria.addInfo(criterion);
	}

	/**
	 * Compare sort content (all criterion are equals).
	 *
	 * @param obj another sort to compare to
	 */
	public boolean equals(Object obj)
	{
		if (obj == this)
			return true;
		if (!(obj instanceof LySort))
			return false;
		LySort sort2 = (LySort)obj;

		if (getCriterionCount() != sort2.getCriterionCount())
			return false;

		int			    i, j, n = getCriterionCount();
		LySortCriterion crit1, crit2;

		for (i = 0; i < n; i++)
		{
			crit1 = getCriterion(i);

			for (j = 0; j < n; j++)
			{
				crit2 = sort2.getCriterion(j);

				if (crit1.equals(crit2))
					break;
			}
			if (j >= n)
				return false;
		}
		return true;
	} // end method equals

	/**
	 * Adds a criterion to this sort by giving a field info and a boolean that indicates if sort on
	 * this field is increasing or not. This method is a shortcut of the previous one.
	 *
	 * @param fieldInfo  Field info for which a sort criterion is added to this sort.
	 * @param increasing Sort criterion order.
     * @return The new sort criterion.
	 */
	public LySortCriterion addCriterion(LyFieldInfo fieldInfo, boolean increasing)
	{
		LySortCriterion criterion = new LySortCriterion(fieldInfo, increasing);

		addCriterion(criterion);
        return criterion;
	}
    
    /**
     * Adds all criteria of the given sort to this sort.
     * @param sort Sort added to this sort.
     */
    public void addSort(LySort sort)
    {
        synchronized (_criteria)
        {
            LySortCriterion criterion;
            int             i, n = sort.getCriterionCount();

            // Go through the element list
            for (i = 0; i < n; i++)
            {
                criterion = sort.getCriterion(i);
                addCriterion(criterion.cloneSortCriterion());
            }
        }
    }
    
    /**
     * Sets criteria list.
     */
    protected void setCriteria(Vector<LySortCriterion> criteria)
    {
        _criteria = new LyInfoList<LySortCriterion>();
        
        int     i, n = (criteria == null ? 0 : criteria.size());
        
        for (i = 0; i < n; i++)
            addCriterion(criteria.elementAt(i));
    }
} // end class LySort
