/*
 * LySortCriterion.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.io.*;
import java.util.*;
import leon.data.*;
import leon.info.infointerface.*;
import leon.misc.*;


/**
 * This class represents a sort criterion. It contains the following attributes : - the fieldInfo on
 * which the sort is performed, - a boolean indicating if the sort is increasing or not.<br>
 * A sort contains a list of sort criteria.<br>
 * See also : tag <SORT_CRITERIA> of leon.dtd
 *
 * @see LySort
 * @see LyFieldInfo
 */
public final class LySortCriterion extends LyInfo
{
	public static final String __VERSION =
		"$Id: LySortCriterion.java 32111 2011-08-29 07:47:52Z jlsaouli $";

	/** Field info concerned by the sort criterion. */
	LyFieldInfo				   _fieldInfo;

	/** Boolean indicating if the sort is increasing or not. */
	boolean					   _increasing;

	/** Interface used to compare values for a field info. */
	String					   _javaName;

	/** Comparator instance. */
	LyComparatorInterface	   _comparator;

	/**
	 * BitSet containing the modifiers of the expression. See the MOD_XXX constants, default is null
	 * (no modifier).
	 */
	BitSet					   _modifiers;
    
    /** Boolean indicating if a groupBy is requested on criterion. */
    boolean                    _groupBy;
    

    /**
     * Default constructor.
     */
    protected LySortCriterion()
    {
    }
    
	/**
	 * Constructor for a sort criterion.
	 *
	 * @param fieldInfo The concerned field info.
	 * @param incresing True if increasing, false if decreasing.
	 */
	public LySortCriterion(LyFieldInfo fieldInfo, boolean increasing)
	{
		this(null, fieldInfo, increasing, null, null);
	}

	/**
	 * Constructor for a sort criterion.
	 *
	 * @param id        Id of the sort criterion.
	 * @param fieldInfo The concerned field info.
	 * @param incresing True if increasing, false if decreasing.
	 * @param javaName  Name of the java class corresponding to this sort criterion.
	 * @param modifiers modifiers uses for sort the elements.
	 */
	public LySortCriterion(String id, LyFieldInfo fieldInfo, boolean increasing,
		String javaName, BitSet modifiers)
	{
		super(id);

		_fieldInfo = fieldInfo;
		_increasing = increasing;
		_javaName = javaName;
        _groupBy = false;

		if (modifiers != null)
			_modifiers = (BitSet)modifiers.clone();

		// Set application for dynamic sorts
		if ((id == null) && (fieldInfo != null))
			_application = fieldInfo.getApplication();
	}

	/**
	 * Constructor for a sort criterion.
	 *
	 * @param fieldInfo The concerned field info.
	 * @param incresing True if increasing, false if decreasing.
	 * @param javaName  Name of the java class corresponding to this sort criterion.
	 */
	public LySortCriterion(LyFieldInfo fieldInfo, boolean increasing, String javaName)
	{
		this(null, fieldInfo, increasing, javaName, null);
	}

	/**
	 * Constructor for a sort criterion.
	 *
	 * @param id        Id of the sort criterion.
	 * @param fieldInfo The concerned field info.
	 * @param incresing True if increasing, false if decreasing.
	 * @param javaName  Name of the java class corresponding to this sort criterion.
	 */
	public LySortCriterion(String id, LyFieldInfo fieldInfo, boolean increasing,
		String javaName)
	{
		this(id, fieldInfo, increasing, javaName, null);
	}

	/**
	 * Compares two objects according to sort criterion parameters.
	 *
	 * @param  object1 first object.
	 * @param  object2 first object.
	 * @return 0 is objects are equals according to this criterion, a negative value if first object
	 *         is lower than second object, a positive value if first object is greater than second
	 *         object.
	 */
	public int compare(LyObject object1, LyObject object2)
	{
		if (object1 == object2)
			return 0;
		if (object1 == null)
			return -1;
		if (object2 == null)
			return 1;

		LyComparatorInterface comparator = getComparator();

		if (comparator != null)
			return comparator.compare(this, object1, object2);

		Object value1 = object1.getValue(_fieldInfo);
		Object value2 = object2.getValue(_fieldInfo);
		return compare(value1, value2);
	}
	
	/**
	 * Compares two values according to sort criterion parameters.
	 *
	 * @param  value1 first value.
	 * @param  value2 first object.
	 * @return 0 is values are equals according to this criterion, a negative result if first value
	 *         is lower than second one, a positive result otherwise.
	 */
	public int compare(Object value1, Object value2)
	{
		int    result;

		if (_modifiers == null)
			result = _fieldInfo.compare(value1, value2);
		else
			result = _fieldInfo.compare(value1, value2, _modifiers);

		return (_increasing ? result : -result);
	}

	/**
	 * Writes XML corresponding to this criterion element.
	 *
	 * @param pw PrintWriter on which the XML is written.
	 */
	public void writeXml(PrintWriter pw)
	{
		// Write tag with its attributes
		pw.print("<SORT_CRITERIA");

		String fieldId = getFieldInfo().getId();

		pw.print(" idref='" + LyString.getXmlString(fieldId) + "'");
		if (!_increasing)
			pw.print(" increasing='false'");
        if (_groupBy)
            pw.print(" groupBy='true'");
		if (_javaName != null)
			pw.print(" javaName=\"" + _javaName + "\"");
		pw.println("/>");
	}

	/**
	 * Gets comparator instance.
	 */
	protected LyComparatorInterface getComparator()
	{
		if (_comparator != null)
			return _comparator;
		if (_javaName == null)
			return null;

		try
		{
			// Create comparator
			Class<?> comparatorClass = _application.loadClass(_javaName,
					"BF AppEngine : unknown comparator class " + _javaName
					+ " for field info " + _fieldInfo);

			_comparator = (LyComparatorInterface)comparatorClass.newInstance();
		}
		catch (Exception e)
		{
			throw new IllegalStateException("BF AppEngine : invalid comparator class "
				+ _javaName + " for field info " + _fieldInfo);
		}

		return _comparator;
	}

	/**
	 * Get field info referred by this sort criterion.
	 *
	 * @return The concerned field info.
	 */
	public LyFieldInfo getFieldInfo()
	{
		return _fieldInfo;
	}

	/**
	 * Tests if sort criterion order is increasing.
	 *
	 * @return True if the sort is increasing, false otherwise.
	 */
	public boolean isIncreasing()
	{
		return _increasing;
	}
    
    /**
     * Tests if sort criterion order groupBy mode is set.
     *
     * @return True if the groupBy mode is on, false otherwise.
     */
    public boolean isGroupBy()
    {
        return _groupBy;
    }

	/**
	 * Checks whether the expression has the given modifier set.
	 *
	 * @param  modifier Modifier to set on the expression. See the MOD_XXX constants, default is
	 *                  null (no modifier).
	 * @return True if the modifier is on, false it is off.
	 */
	public boolean hasModifier(short modifier)
	{
		return ((_modifiers != null) && _modifiers.get(modifier));
	}

	/**
	 * Checks whether the expression has the given modifier set.
	 *
	 * @param  modifier Modifier to set on the expression. See the MOD_XXX constants, default is
	 *                  null (no modifier).
	 * @return True if the modifier is on, false it is off.
	 */
	public BitSet getModifiers()
	{
		return _modifiers;
	}

	/**
	 * Gets java class name corresponding to this sort criterion.
	 *
	 * @return Name of the java class corresponding to this sort criterion.
	 */
	public String getJavaName()
	{
		return _javaName;
	}

	/**
	 * Sets the fieldInfo referred by this sort criterion.
	 *
	 * @param fieldInfo FieldInfo concerned by this sort criterion.
	 */
	public void setFieldInfo(LyFieldInfo fieldInfo)
	{
		if (isLocked())
			throw new IllegalStateException("BF AppEngine : criterion " + this
				+ " was locked and thus can't be modified anymore");

		_fieldInfo = fieldInfo;
	}

	/**
	 * Sets criterion order.
	 *
	 * @param increasing True if the sort order become increasing, false otherwise.
	 */
	public void setIncreasing(boolean increasing)
	{
		if (isLocked())
			throw new IllegalStateException("BF AppEngine : criterion " + this
				+ " was locked and thus can't be modified anymore");

		_increasing = increasing;
	}

    /**
     * Sets criterion groupBy mode.
     *
     * @param groupBy True if the groupBy mode is set, false otherwise.
     */
    public void setGroupBy(boolean groupBy)
    {
        if (isLocked())
            throw new IllegalStateException("BF AppEngine : criterion " + this
                + " was locked and thus can't be modified anymore");

        _groupBy = groupBy;
    }
    
	/**
	 * Sets java class name corresponding to this sort criterion.
	 *
	 * @param javaName Name of the java class corresponding to this sort criterion.
	 */
	public void setJavaName(String javaName)
	{
		_javaName = javaName;
	}

	/**
	 * Sets a modifier on this expression.
	 *
	 * @param modifier Modifier to set on the expression. See the MOD_XXX constants, default is null
	 *                 (no modifier).
	 */
	public void setModifier(short modifier)
	{
		if (_modifiers == null)
		    _modifiers = new BitSet();

		_modifiers.set(modifier);
	}

	/**
	 * Compare sort criterion content (all parameters are equals).
	 *
	 * @param obj another sort criterion to compare to
	 */
	public boolean equals(Object obj)
	{
		if (obj == this)
			return true;
		if (!(obj instanceof LySortCriterion))
			return false;

		LySortCriterion crit2 = (LySortCriterion)obj;

		if (_fieldInfo != crit2._fieldInfo)
		{
			// must have same id
			if ((_fieldInfo == null) || (crit2._fieldInfo == null))
				return false;
			if (!_fieldInfo.getId().equals(crit2._fieldInfo.getId()))
				return false;
		}
		if (_increasing != crit2._increasing)
			return false;
        if (_groupBy != crit2._groupBy) // Is it significant ?
            return false;

		if (_javaName != crit2._javaName)
		{
			if ((_javaName == null) || (crit2._javaName == null))
				return false;
			if (!_javaName.equals(crit2._javaName))
				return false;
		}
		if (_modifiers != crit2._modifiers)
		{
			if ((_modifiers == null) || (crit2._modifiers != null))
				return false;
			if (!_modifiers.equals(crit2._modifiers))
				return true;
		}
		return true;
	} // end method equals
    
    /**
     * Clone this criterion.
     */
    public LySortCriterion cloneSortCriterion()
    {
        LySortCriterion criterion = (LySortCriterion)clone();

        return criterion;
    }
} // end class LySortCriterion
