/*
 * LyStep.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

/**
 * This class implements a step between two classes. A step defines how to get objects of a class
 * from objects of an other class using a relation. This step is defined in a source class and has a
 * target class, relation fields between this two classes, a direction used to known how to use
 * those relations and an optional filter for filtering results. A step is defined within a route.
 *
 * <p>See also tag <STEP> of leon.dtd.</p>
 *
 * @see LyRoute
 */
public final class LyStep extends LyInfo
{
	public static final String		  __VERSION =
		"$Id: LyStep.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The target class of this step. */
	protected LyClassInfo			  _target;

	/** Relations or arrays of relation to be used to reach the target. */
	protected LyInfoList<LyFieldInfo> _relations;

	/**
	 * Boolean indicating the direction of the step : true if the relations are owned by the target
	 * class (relations are inverted : they belong to target class and reference source class). This
	 * attribute value is calculated internally.
	 */
	protected boolean				  _reverse;

	/** Filter for filtering results. */
	protected LyFilter				  _filter;

    /**
     * Default constructor for a step.
     */
    LyStep()
    {
    }
    
	/**
	 * Constructor of a step.
	 *
	 * @param     id       identifier of the step.
	 * @param     target   Target class.
	 * @param     relation List of relations for this step.
	 * @param     filter   Filter of the step (can be null)
	 * @param     reverse  Boolean indicating the direction of the step.
	 * @exception IllegalArgumentException If target parameter is null.
	 * @exception IllegalArgumentException If relations parameter is null or if the list is empty.
	 */
	LyStep(String id, LyClassInfo target, LyInfoList<LyFieldInfo> relations,
		LyFilter filter, boolean reverse)
	{
		super(id);

		if (target == null)
			throw new IllegalArgumentException("BF AppEngine : target of a step can't be null.");
		if ((relations == null) || (relations.getSize() == 0))
			throw new IllegalArgumentException(
				"BF AppEngine : relations of a step can't be null, nor empty.");

		_target = target;
		_relations = relations;
		_filter = filter;
		_reverse = reverse;
	}

	/**
	 * Constructor of a step.
	 *
	 * @param     target   Target class.
	 * @param     relation List of relations for this step.
	 * @param     filter   Filter of the step (can be null)
	 * @param     reverse  Boolean indicating the direction of the step.
	 * @exception IllegalArgumentException If target parameter is null.
	 * @exception IllegalArgumentException If relations parameter is null or if the list is empty.
	 */
	LyStep(LyClassInfo target, LyInfoList<LyFieldInfo> relations,
		LyFilter filter, boolean reverse)
	{
		this(null, target, relations, filter, reverse);
	}

	/**
	 * Constructor of a step.
	 *
	 * @param     target   Target class.
	 * @param     relation List of relations for this step.
	 * @param     filter   Filter of the step (can be null)
	 * @exception IllegalArgumentException If target parameter is null.
	 * @exception IllegalArgumentException If relations parameter is null or if the list is empty.
	 */
	LyStep(LyClassInfo target, LyInfoList<LyFieldInfo> relations, LyFilter filter)
	{
		this(null, target, relations, filter, false);
	}

	/**
	 * This method checks for the integrity of the step and completes the initialization if
	 * necessary. Checks performed are the following : - all relations should refer to source class
	 * or target class, - all relations should belong to source class or target class, - all
	 * relations should belong to the same class, - steps shouldn't use relations that have no
	 * direction, - steps should only use relations that will enable their traversal according to
	 * the step needs (if the step is not reverse, forward direction must be enabled and if the step
	 * is reverse, backward direction must be enabled).
	 *
	 * @param     source Source class for this step (starting point of the step).
	 * @exception IllegalStateException If one of these checks fails.
	 */
	void check(LyClassInfo source)
	{
		int				    i, n = _relations.getSize();
		LyFieldInfo		    fieldInfo;
		LyRelationFieldInfo relation;
		LyArrayFieldInfo    array;
		boolean			    reverse;

		for (i = 0; i < n; i++)
		{
			fieldInfo = _relations.getInfo(i);
			if (fieldInfo instanceof LyRelationFieldInfo)
				relation = (LyRelationFieldInfo)fieldInfo;
			else if (fieldInfo instanceof LyArrayFieldInfo)
			{
				array = (LyArrayFieldInfo)fieldInfo;
				relation = (LyRelationFieldInfo)array.getFieldInfo();
			}
			else
				continue;

			if (!relation.relatedWith(source)
					&& !relation.relatedWith(_target))
				throw new IllegalStateException("BF AppEngine : invalid step " + this
					+ ", relation neither refers to source class (" + source
					+ ") nor to target class (" + _target + ")");
			if ((source.getRealFieldInfo(fieldInfo, false) == null)
					&& (_target.getRealFieldInfo(fieldInfo, false) == null))
				throw new IllegalStateException("BF AppEngine : invalid step " + this
					+ ", relation is neither contained in source class (" + source
					+ ") nor in target class (" + _target + ")");

			// Computing the direction of the step.
			if (i == 0)
				_reverse = !(source.contains(fieldInfo) && relation.relatedWith(_target));
			else
			{
				reverse = !(source.contains(fieldInfo) && relation.relatedWith(_target));
				if (reverse != _reverse)
					throw new IllegalStateException("BF AppEngine : invalid step " + this
						+ ", all relations of a step must be owned by the same class");
			}
		}
	} // end method check

	/**
	 * Lock this step.
	 */
	protected void lock()
	{
		// Lock relations to avoid further modifications
		_relations.lock();
	}

	/**
	 * Accessor to the target class of this step.
	 *
	 * @return Target class of this step.
	 */
	public LyClassInfo getTarget()
	{
		return _target;
	}

	/**
	 * Accessor to relations of this step. Be careful ! It can contain relations or arrays of
	 * relation.
	 *
	 * @return List of relation fields of this step (never null, nor empty).
	 */
	public LyInfoList<LyFieldInfo> getRelations()
	{
		return _relations;
	}

	/**
	 * Accessor to the filter of this step. This filter is used to filter results.
	 *
	 * @return Filter of this step (may be null).
	 */
	public LyFilter getFilter()
	{
		return _filter;
	}

	/**
	 * Returns a boolean indicating if relations belong to target class or not (relations are
	 * inverted). In that case, the step is said reverse.
	 *
	 * @return True is the step is reverse, else false.
	 */
	public boolean isReverse()
	{
		return _reverse;
	}

	/**
	 * Checks if this step is bidirectional or not. A step is said bidirectional if it has no filter
	 * and all relations of the step are bidirectional.
	 *
	 * @return True is the step is bidirectional, else false.
	 */
	boolean isBidirectional()
	{
		// A step with a filter can't be bidirectional
		if (getFilter() != null)
			return false;

		// A step is bidirectional if all relations within this step are bidirectional
		LyFieldInfo		    info;
		LyRelationFieldInfo relation;
		LyArrayFieldInfo    array;
		int				    i, n = _relations.getSize();

		for (i = 0; i < n; i++)
		{
			info = _relations.getInfo(i);
			if (info instanceof LyRelationFieldInfo)
				relation = (LyRelationFieldInfo)info;
			else if (info instanceof LyArrayFieldInfo)
			{
				array = (LyArrayFieldInfo)info;
				relation = (LyRelationFieldInfo)array.getFieldInfo();
			}
			else
				continue;

			if (relation.getDirection() != LyRelationFieldInfo.Direction.BOTH)
				return false;
		}

		return true;
	} // end method isBidirectional

	/**
	 * Returns a string representation of this object.
	 */
	public String dump()
	{
		StringBuffer buffer = new StringBuffer();

		buffer.append("{ ");
		buffer.append(_target);
		if (_reverse)
			buffer.append(", reverse");
		if (_filter != null)
		{
			buffer.append(", ");
			buffer.append(_filter);
		}
		buffer.append(" : ");

		int				    i, n = _relations.getSize();
		LyFieldInfo		    info;
		LyRelationFieldInfo relation;
		LyArrayFieldInfo    array;

		for (i = 0; i < n; i++)
		{
			info = _relations.getInfo(i);
			if (i > 0)
				buffer.append(", ");

			if (info instanceof LyRelationFieldInfo)
			{
				array = null;
				relation = (LyRelationFieldInfo)info;
			}
			else
			{
				array = (LyArrayFieldInfo)info;
				relation = (LyRelationFieldInfo)array.getFieldInfo();
			}

			buffer.append(relation.toString());
			if (array != null)
				buffer.append("[]");
		}

		buffer.append(" }");

		return buffer.toString();
	} // end method toString

	/**
	 * Sets the filter of the step.
	 *
	 * @param filter the filter.
	 */
	protected void setFilter(LyFilter filter)
	{
		_filter = filter;
	}
    
    /**
     * Sets the target class of this step.
     *
     * @param target Target class of this step.
     */
    protected void setTarget(LyClassInfo target)
    {
        _target = target;
    }
    
    /**
     * Sets relations of this step.
     */
    protected void setRelations(LyInfoList<LyFieldInfo> relations)
    {
        if (relations == null)
            _relations = null;
        else
            _relations = new LyInfoList<LyFieldInfo>(relations);
    }
} // end class LyStep
