/*
 * LySticker.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.io.*;

import leon.misc.*;
import leon.info.infointerface.*;


/**
 * This class represents a sticker. It contains the following attributes : - the action on which
 * the sticker is displayed, - the owner of this sticker, - the object on which the sticker is
 * displayed, - the content of the sticker, - the x position of the sticker, - the y position of
 * the sticker, - the width of the sticker, - the height of the sticker.
 *
 * @see LyObjectInfo
 */
public final class LySticker extends LyObjectInfo
    implements LyExtraInfoInterface
{
	public static final String __VERSION =
		"$Id: LySticker.java 25079 2010-04-07 09:24:00Z bangiole $";

	/** Id of the user that owns the sticker. */
	String					   _user;

	/** Action on which the sticker is displayed. */
	LyAction				   _action;

	/** Id of the object on which the sticker is displayed. */
	String					   _objectId;
	
	/** Content of the sticker. **/
	String                     _content;
	
	/** X position of the sticker. **/
	int  	                   _x;
	
	/** Y position of the sticker. **/
	int  	                   _y;
	
	/** Width of the sticker. **/
	int  	                   _width;
	
	/** Height of the sticker. **/
	int  	                   _height;
    
	/**
     * Constructor for a sticker.
     *
     * @param id          the identifier of the sticker.
     * @param user        id of the user that owns the sticker.
     * @param action      action on which the sticker is displayed.
     * @param objectId    id of the object on which the sticker is displayed.
     * @param content     content of the sticker.
     * @param x           x position of the sticker.
     * @param y           y position of the sticker.
     * @param width       width of the sticker.
     * @param height      height of the sticker.
     */
    public LySticker(String id, String user, LyAction action, String objectId, String content, int x, int y, int width, int height)
    {
		super(id);

		_user = user;
		_action = action;
		_objectId = objectId;
		_content = content;
		_x = x;
		_y = y;
		_width = width;
		_height = height;
    }

	/**
	 * Clone a sticker.
	 */
	public LySticker cloneSticker()
	{
		LySticker sticker = new LySticker(_id + LyString.getRandomId(), _user, _action,
				_objectId, _content, _x, _y, _width, _height);

		return sticker;
	}
    
    /**
     * Method for freeing extra info (and sub-infos).
     */
    public void free()
    {
        super.free();
    }

    /**
     * Method for saving sticker file.
     */
    public boolean saveFile(LyXmlWriterInterface writer)
    {
        String      fileName;
        
        if (writer == null)
        	fileName = getFile();
        else
        	fileName = writer.getFile(this);
        
        if (fileName == null)
        {
            String   dir, userId = getUser();
            
            if (userId == null)
            {
                userId = "";
                dir = "LY_STICKER_DIR";
            }
            else
            {
                userId = userId.substring(userId.indexOf(leon.data.LyObjectId.CLASS_SEP) + 1) + "_";
                dir = "LY_USER_STICKER_DIR";
            }
    
            // retrieve the name of the data directory
            LyEnvironment environment = getEnvironment();
            String        dirPath = environment.getFile(dir);
            File          fileDirectory = null;

            if (dirPath != null)
                fileDirectory = new File(dirPath);

            // Try to create the directory if needed
            if ((fileDirectory != null)
                && !fileDirectory.exists())
            {
                try
                {
                    environment.logInfo("Creating directory :" + dirPath);
                    fileDirectory.mkdirs();
                }
                catch (SecurityException e)
                {
                    environment.logError("Unable to create directory :" + dirPath);
                    return false;
                }
            }

            File    file;
            int     i = 0;

            // find a file name which has not been affected
            do
            {
                fileName = dirPath + "/sticker_" + userId + i + ".xml";
                file = new File(fileName);
                i++;
            }
            while (file.exists());
            
            setFile(fileName);
        }
        
        // Create xml file for the sticker
        try
        {
            FileOutputStream 	output = new FileOutputStream(fileName);
            LyXmlPrintWriter    pw = new LyXmlPrintWriter(output);

            if (writer == null)
            	writeXml(pw, true, true);
            else
            	writer.writeXml(pw, this);
            
            pw.close();
            return true;
        }
        catch (Exception e)
        {
            LyEnvironment   environment = getEnvironment();
            
            environment.writeException(e);
            return false;
        }
    }
    
    /**
     * Method for deleting sticker file.
     */
    public boolean deleteFile()
    {
        String  filePath = getFile();

        if (filePath == null)
            return false;
        
        try
        {
            // Delete the file
            File    file = new File(filePath);
            
            return file.delete();
        }
        catch (SecurityException e)
        {
            LyEnvironment   environment = getEnvironment();
            
            environment.writeException(e);
            return false;
        }
    }
    
    /**
     * Method to rewrite XML corresponding to this component.
     *
     * @param pw The printWriter on which the XML is written.
     * @param header Boolean indicating if header must be written.
     * @param header Boolean indicating if body must be written.
     */
    public void writeXml(PrintWriter pw, boolean header, boolean body)
    {
        if (header)
        {
            pw.println("<?xml version='1.0' encoding='" + LyString.DEFAULT_CHARSET + "'?>");
            pw.println("<!DOCTYPE STICKER SYSTEM 'leon/conf/dtd/leon_gui.dtd'>");
        }
        
        if (body)
        	writeXml(pw);
    }
    
	/**
	 * Writes XML corresponding to this sticker element.
	 *
	 * @param pw PrintWriter on which the XML is written.
	 */
	public void writeXml(PrintWriter pw)
	{
		// Write tag with its attributes
		pw.print("<STICKER");
		pw.print(" id='" + LyString.getXmlString(getId()) + "'");
		if (_user != null)
			pw.print(" user='" + LyString.getXmlString(_user) + "'");
		if (_action != null)
			pw.print(" action='" + LyString.getXmlString(_action.getId()) + "'");
		if (_objectId != null)
			pw.print(" objectId='" + LyString.getXmlString(_objectId) + "'");
		if (_content != null)
			pw.print(" content='" + LyString.getXmlString(_content) + "'");

		pw.print(" x='" + _x + "'");
		pw.print(" y='" + _y + "'");
		pw.print(" width='" + _width + "'");
		pw.print(" height='" + _height + "'");
		pw.println(">");

		if (_action != null)
			writeExternXml(pw, _action.getId());
		
		pw.println("</STICKER>");
	} // end method writeXml

	/**
	 * Writes XML for an external idref.
	 *
	 * @param pw PrintWriter on which the XML is written.
	 * @param id id that must be written.
	 */
	protected void writeExternXml(PrintWriter pw, String id)
	{
		pw.println("<EXTERN id='" + LyString.getXmlString(id) + "'/>");
	}

	/**
     * Get the action referred by this sticker.
     *
     * @return The concerned action.
     */
    public LyAction getAction()
    {
        return _action;
    }

	/**
	 * Get the object id referred by this sticker.
	 *
	 * @return the object id referred by this sticker.
	 */
	public String getObjectId()
	{
		return _objectId;
	}

	/**
	 * Get the content of this sticker.
	 *
	 * @return the content of this sticker.
	 */
	public String getContent()
	{
		return _content;
	}

	/**
	 * Get the x position of this sticker.
	 *
	 * @return the x position of this sticker.
	 */
	public int getX()
	{
		return _x;
	}

	/**
	 * Get the y position of this sticker.
	 *
	 * @return the y position of this sticker.
	 */
	public int getY()
	{
		return _y;
	}

	/**
	 * Get the width of this sticker.
	 *
	 * @return the width of this sticker.
	 */
	public int getWidth()
	{
		return _width;
	}

	/**
	 * Get the height of this sticker.
	 *
	 * @return the height of this sticker.
	 */
	public int getHeight()
	{
		return _height;
	}

	/**
	 * Get owner of this sticker.
	 *
	 * @return The owner of this sticker.
	 */
	public String getUser()
	{
		return _user;
	}

	/**
     * Sets the action referred by this sticker.
     *
     * @param action action concerned by this sticker.
     */
    public void setAction(LyAction action)
    {
        _action = action;
    }

	/**
	 * Set the content of the sticker.
	 *
	 * @param content the content of the sticker.
	 */
	public void setContent(String content)
	{
		_content = content;
	}

	/**
	 * Set the x position of the sticker.
	 *
	 * @param x the x position of the sticker.
	 */
	public void setX(int x)
	{
		_x = x;
	}

	/**
	 * Set the y position of the sticker.
	 *
	 * @param y the y position of the sticker.
	 */
	public void setY(int y)
	{
		_y = y;
	}

	/**
	 * Set the width of the sticker.
	 *
	 * @param width the width of the sticker.
	 */
	public void setWidth(int width)
	{
		_width = width;
	}

	/**
	 * Set the height of the sticker.
	 *
	 * @param height the height of the sticker.
	 */
	public void setHeight(int height)
	{
		_height = height;
	}
} // end class LySticker
