/**
 * LyStructFieldInfo.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;
import leon.app.*;
import leon.data.*;
import leon.info.binding.*;
import leon.info.misc.*;


/**
 * This class implements the description of struct field info.
 *
 * @see LyRelationFieldInfo
 */
public final class LyStructFieldInfo extends LyRelationFieldInfo
{
	public static final String						   __VERSION =
		"$Id: LyStructFieldInfo.java 34640 2012-03-14 10:55:58Z jlsaouli $";

	/** List of classes that contains this structFieldInfo and real structs used. */
	protected LyInfoDataList<LyClassInfo, LyFieldInfo> _parentClasses;

	/**
	 * Constructor of a struct.
	 *
	 * @param id             Identifier of the field.
	 * @param javaName       Name of the generated java field.
	 * @param extendedStruct Struct extended by this struct.
	 * @param template       Template used to construct this field.
	 * @param javaName       Java name of the associated class.
	 * @param abstractClass  Indicates if the class that represents the struct is abstract or not.
	 */
	LyStructFieldInfo(String id, String javaName, LyStructFieldInfo extendedStruct,
		LyStructFieldInfo template, String className, boolean abstractClass)
	{
		this(id);
		construct(javaName, extendedStruct, template, className, abstractClass);
	}

	/**
	 * Constructor of a struct.
	 *
	 * @param id Identifier of the field.
	 */
	LyStructFieldInfo(String id)
	{
		super(id);
	}

	/**
	 * Construction of a struct.
	 *
	 * @param javaName       Name of the generated java field.
	 * @param extendedStruct Struct extended by this struct.
	 * @param template       Template used to construct this field.
	 * @param javaName       Java name of the associated class.
	 * @param abstractClass  Indicates if the class that represents the struct is abstract or not.
	 */
	protected void construct(String javaName, LyStructFieldInfo extendedStruct,
		LyStructFieldInfo template, String className, boolean abstractClass)
	{
		// Create relation field info
        boolean     fieldTag = _fieldTag;
        
		super.construct(javaName, extendedStruct, template, false, false,
			LyRelationFieldInfo.Direction.BOTH, LyRelationFieldInfo.Type.COMPOSITION, null,
			false, false, LyRelationFieldInfo.ShowAs.DEFAULT);
        _fieldTag = fieldTag;

		if (fromFieldTag())
		{
			// A new class is not needed in this case
			LyClassInfo classInfo = template.getClassInfo();

			addClassInfo(classInfo, null, -1);
			return;
		}

		// Create "internal" class info
		LyClassInfo superClassInfo = null;
		LyClassInfo templateClassInfo = null;

		if (extendedStruct != null)
			superClassInfo = extendedStruct.getClassInfo();
		else if (template != null)
			templateClassInfo = template.getClassInfo();

		LyStructClassInfo classInfo = new LyStructClassInfo(this, superClassInfo,
				templateClassInfo, className, abstractClass);

		// Add class to relation
		addClassInfo(classInfo, null, -1);
	} // end method construct

	/**
	 * Init this instance attributes from its template.
	 *
	 * @param parts Indicates which parts of the template are copied.
	 */
	void initFromTemplate(int parts)
	{
        parts &= ~_initParts;
        if (parts == 0)
            return;
		if (_template == null)
			return;

		LyClassInfo classInfo = getClassInfo();

		super.initFromTemplate(parts);

		if ((classInfo != null)
			&& (classInfo.getParent() == this))
		{
	        _classes = new LyInfoDataList<LyClassInfo, LyClassRef>();
			classInfo.initFromTemplate(parts);
			addClassInfo(classInfo, null, -1);
		}
	}

	/**
	 * Final step of parsing : check consistency. Checks performed are the following :
	 *
	 * <ul>
	 *   <LI>the relation must refer at least one class info,</LI>
	 *   <LI>if not null, classField and contextField should belong to parent class,</LI>
	 *   <LI>check routes.</LI>
	 * </ul>
	 *
	 * @param     parent Object info in which this field is checked.
	 * @param     parts  Indicates which parts are checked.
	 * @exception IllegalStateException If one of these checks fails.
	 */
	void check(LyObjectInfo parent, int parts)
	{
		if ((parts & PART_DATA) > 0)
		{
			// To have optional structs we must clear the mark before the super.check
			// and readd it after.
			boolean addOptional = !isMultiple() && hasMark(MARK_OPTIONAL);

			if (addOptional)
				_marks.clear(MARK_OPTIONAL);

			super.check(parent, parts);

			if (addOptional)
				_marks.set(MARK_OPTIONAL);
		}
		else
			super.check(parent, parts);

		LyInfo info;

		info = parent;
		while ((info != null) && !(info instanceof LyProject))
			info = info.getParent();

		if (fromFieldTag())
		{
			LyInfoList<LyClassInfo> subTypes = getClassInfos();
			LyClassInfo			    subStructClass;
			LyStructFieldInfo	    subStruct;

			for (int i = 0; i < subTypes.getSize(); i++)
			{
				subStructClass = subTypes.getInfo(i);
				if (subStructClass.isTemporary()
						&& (subStructClass.getParent() instanceof LyClassInfo))
					subStruct = (LyStructFieldInfo)subStructClass.getParent().getParent();
				else
					subStruct = (LyStructFieldInfo)subStructClass.getParent();
				subStruct.check((LyObjectInfo)subStruct.getParent(), parts);
			}
		}

		LyClassInfo classInfo = getClassInfo();

		if (((parts & PART_DATA) > 0)
				&& (classInfo.getParent() == this))
		{
			// If classInfo has no id field we must add one
			if (!classInfo.isAbstract())
			{
				LyInfoList<LyFieldInfo> idFieldInfos = classInfo.getIdFieldInfos();

				if ((idFieldInfos == null) || (idFieldInfos.getSize() == 0))
				{
					if (_template != null)
					{
						LyClassInfo templateClassInfo = ((LyStructFieldInfo)_template)
								.getClassInfo();

						idFieldInfos = templateClassInfo.getIdFieldInfos();
					}

					if ((idFieldInfos != null) && (idFieldInfos.getSize() > 0))
					{
						int i, n = idFieldInfos.getSize();

						for (i = 0; i < n; i++)
							classInfo.addFieldInfo(idFieldInfos.getInfo(i), i);
					}
					else
					{
						String			  id = classInfo.getId() + ".id";
						LyNumberFieldInfo number = new LyNumberFieldInfo(id, null, null,
								LyNumberFieldInfo.Type.INT, false);
						BitSet			  marks = new BitSet();

						marks.set(LyFieldInfo.MARK_ID);
						marks.set(LyFieldInfo.MARK_HIDDEN);
						marks.set(LyFieldInfo.MARK_PROVIDER_VALUE);
						marks.set(LyFieldInfo.MARK_OPTIONAL);
						number.setMarks(marks);
						number.setApplication(_application);
						number.setParent(classInfo);

						classInfo.addFieldInfo(number, 0);

						number.check(classInfo, parts);
					}
				}
			}
		}

		if (((parts & PART_GUI) > 0)
				&& (classInfo.getParent() == this))
		{
			classInfo.addAction(_application.getAction("_consult"), -1);
			classInfo.addAction(_application.getAction("_create"), -1);
			classInfo.addAction(_application.getAction("_set"), -1);
			classInfo.addAction(_application.getAction("_delete"), -1);

			classInfo.setName(_name);
		}

		if (((parts & PART_BIND) > 0)
				&& (classInfo.getParent() == this))
		{
			if (!isLocal()
					&& (classInfo._binding == null))
			{
				// Use a default binding because id is not compliant with db names
				String		   defaultLocationId = _application.getDefaultLocationId();
				LyClassBinding binding = new LyClassBinding("__" + _id + "_structClass",
						defaultLocationId, null);

                binding.setDefaultBinding(true);
				classInfo.setBinding(binding);
			}
			else if (isLocal())
			{
				BitSet marks = classInfo.getMarks();

				if (marks == null)
					marks = new BitSet();
				marks.set(LyClassInfo.MARK_LOCAL);
				classInfo.setMarks(marks);
			}

			String	    id = classInfo.getId() + ".id";
			LyFieldInfo fieldInfo = classInfo.getFieldInfo(id, false);

			if (!isLocal() && (fieldInfo != null)
					&& (fieldInfo._binding == null)
					&& (fieldInfo instanceof LyNumberFieldInfo))
			{
				// This field was created, add the binding
				// Use a default binding because id is not compliant with db names
				LyFieldBinding binding = new LyFieldBinding(classInfo.getBinding().getBinding()
						+ "_id", null, null, true);

                binding.setDefaultBinding(true);
				fieldInfo.setBinding(binding);
			}

			if ((fieldInfo != null)
					&& (fieldInfo._binding != null)
					&& (!fieldInfo.hasMark(LyFieldInfo.MARK_PROVIDER_VALUE_BIND)))
			{
				BitSet marks = fieldInfo.getMarks();

				if (marks == null)
					marks = new BitSet();

				marks.clear(LyFieldInfo.MARK_PROVIDER_VALUE);
				marks.clear(LyFieldInfo.MARK_OPTIONAL);
				fieldInfo.setMarks(marks);
			}
		}

		// Maybe we have to change the labels of the template by new labels
		// if they reference the template.
		// We must do this job before the call to super.initFromTemplate().
		if ((_template != null)
				&& (((parts & PART_GUI) > 0)
					|| ((parts & PART_LABEL) > 0)))
		{
			LyStructFieldInfo template = (LyStructFieldInfo)_template;
			LyInfoLabel		  label, tLabel;
			int				  nl = template.getLabelCount();
			int				  nu = template.getUserLabelCount();
			int				  i;

			for (i = 0; i < nl; i++)
			{
				tLabel = template._labels[i];
				if (tLabel != null)
				{
					label = getLabel(i);
					if ((label == null) || (label == tLabel))
					{
						label = cloneLabel(tLabel, template);
						addLabel(label);
					}
				}
			}
			for (i = 0; i < nu; i++)
			{
				tLabel = template.getUserLabel(i);
				if (tLabel != null)
				{
					label = getUserLabel(i);
					if ((label == null) || (label == tLabel))
					{
						label = cloneLabel(label, template);
						addLabel(label);
					}
				}
			}
		}

		classInfo.check((LyProject)info, parts);
	} // end method check

	/**
	 * This a convenient method for the parser to reset the classInfo list of this package. This
	 * method must not become public.
	 */
	void resetClasses()
	{
		LyClassInfo classInfo = getClassInfo();

		_classes.resetInfos();
		if ((classInfo != null)
				&& (classInfo.getParent() == this))
		{
			classInfo._temporary = true;
			classInfo.free();
		}
	}

	/**
	 * Internal method to clone a LyInfoLabel when this struct has a template. This method searches
	 * for a step in the specified label with a reference on the template to replace it with its own
	 * reference.
	 *
	 * @param  label    The label to clone
	 * @param  template The template of this struct.
	 * @return a new info label if the specified label has a reference on the template, the
	 *         specified label otherwise.
	 */
	LyInfoLabel cloneLabel(LyInfoLabel label, LyStructFieldInfo template)
	{
		if (label == null)
			return null;
		
		int					    i, j, k, n, p, q, us;
		LyInfoLabel			    newLabel = null;
		LyPath				    path, newPath;
		Vector<LyStep>		    steps;
		LyStep				    step, newStep;
		LyInfoList<LyPath>	    paths;
		LyInfoList<LyUnit>	    units;
		LyInfoList<LyFieldInfo> relations;
		LyInfoList<LyFieldInfo> newRelations;
		LyUnit				    unit;
		boolean				    change = false;
		LyApplication		    application = getApplication();
		LyFieldInfo			    relation;

		// We build the new label on the fly but we use it only if we detect
		// a change.
		paths = label.getPaths();
		units = label.getUnits();
		n = ((paths == null) ? 0 : paths.getSize());
		us = ((units == null) ? 0 : units.getSize());

		newLabel = new LyInfoLabel(null, label.getValue(), label.getType(), label.getSubType());
		newLabel.setModelType(label.getModelType());
		newLabel.setApplication(application);
		newLabel.setParent(this);

		// Search in all paths
		for (i = 0; i < n; i++)
		{
			path = paths.getInfo(i);
			unit = ((i < us) ? units.getInfo(i) : null);

			newPath = new LyPath(null, path.getTarget(), path.getSourceClassInfo(),
					path.getTargetFieldInfo(), path.getFieldLabel());
			newPath.setModelType(path.getModelType());
			newPath.setApplication(application);
			newPath.setParent(newLabel);

			steps = path.getSteps();
			p = ((steps == null) ? 0 : steps.size());

			// Search in all steps
			for (j = 0; j < p; j++)
			{
				step = steps.elementAt(j);
				relations = step.getRelations();
				q = relations.getSize();
				newRelations = new LyInfoList<LyFieldInfo>();

				// Search in all relations
				for (k = 0; k < q; k++)
				{
					relation = relations.getInfo(k);
					if (relation == template)
					{
						relation = this;
						// Mark this label as changed
						change = true;
					}
					newRelations.addInfo(relation);
				}

				newStep = new LyStep(step.getTarget(), newRelations, step.getFilter());
				newStep.setModelType(step.getModelType());
				newStep.setApplication(application);
				newStep.setParent(newPath);

				newPath.addStep(newStep);
			}
			newLabel.addPart(newPath, unit);
		}

		if (change)
			return newLabel;
		else
			return label;
	} // end method cloneLabel

	/**
	 * Sets the application for this info. Once assigned, the application can be changed.
	 *
	 * @param     application The application to which the info belongs.
	 * @exception IllegalStateException If the info already belongs to an application.
	 */
	public void setApplication(LyApplication application)
	{
		super.setApplication(application);

		LyClassInfo classInfo = getClassInfo();

		if ((classInfo != null)
				&& (classInfo.getParent() == this))
			// Sets the application to its classInfo
			classInfo.setApplication(application);
	}

	/**
	 * Gets parent class info corresponding to this struct. If this struct doesn't extend any other,
	 * gets the class info of this struct.
	 *
	 * @return Parent class info for this struct if exist else return class info for this struct.
	 */
	public LyClassInfo getStructClassInfo()
	{
		LyStructFieldInfo struct = getExtendedStruct();

		if ((struct == null) && (_template != null))
			return ((LyStructFieldInfo)_template).getClassInfo();
		if ((_template != null) && (struct != null) && (_template != struct))
			return ((LyStructFieldInfo)_template).getClassInfo();
		else
			return getClassInfo();
	}

	/**
	 * Checks if the class corresponding to this struct is abstract.
	 *
	 * @return True if abstract, false otherwise.
	 */
	public boolean isAbstract()
	{
		load();
        
        LyClassInfo     classInfo = getClassInfo();
        
        return (classInfo == null ? false : classInfo.isAbstract());
	}

	/**
	 * Gets java name of the associated class.
	 *
	 * @return Java name of the associated class.
	 */
	public String getClassName()
	{
		load();
        
        LyClassInfo     classInfo = getClassInfo();
        
        return (classInfo == null ? null : classInfo.getJavaName());
	}

	/**
	 * Gets rules from the associated class info.
	 *
	 * @return rules of the associated class.
	 */
	public LyInfoList<LyRule> getRules()
	{
		LyClassInfo classInfo = this.getClassInfo();

		return (classInfo == null ? null : classInfo.getRules());
	}

	/**
	 * Gets cache policy from the associated class info.
	 *
	 * @return rules of the associated class.
	 */
	public LyInfoCache getCachePolicy()
	{
		LyClassInfo classInfo = this.getClassInfo();

		return (classInfo == null ? null : classInfo.getCachePolicy());
	}

	/**
	 * Gets binding from the associated class info.
	 *
	 * @return binding of the associated class.
	 */
	public LyClassBinding getClassBinding()
	{
		LyClassInfo classInfo = this.getClassInfo();

		return (classInfo == null ? null : classInfo.getBinding());
	}

	/**
	 * Gets behavior from the associated class info.
	 *
	 * @return rules of the associated class.
	 */
	public String getClassBehaviorClassName()
	{
		LyClassInfo classInfo = this.getClassInfo();

		return (classInfo == null ? null : classInfo.getBehaviorClassName());
	}

	/**
	 * Gets list of field infos of this struct.
	 *
	 * @return List of field infos, can't be null since a class info with no fields is not accepted.
	 */
	public LyInfoList<LyFieldInfo> getFieldInfos()
	{
		load();
        
        LyClassInfo     classInfo = getClassInfo();
        
		return (classInfo == null ? null : classInfo.getFieldInfos());
	}

	/**
	 * Returns the struct that is extended by this struct.
	 *
	 * @return Struct extended by this struct, null otherwise.
	 */
	public LyStructFieldInfo getExtendedStruct()
	{
		return (LyStructFieldInfo)getExtendedRelation();
	}

	/**
	 * Gets class info corresponding to this struct.
	 *
	 * @return Class info for this struct.
	 */
	public LyClassInfo getClassInfo()
	{
		if (getClassInfoCount() > 0)
		    return getClassInfo(0);
		
	    if (fromFieldTag()
	        && (_template != null))
        {
            // A new class is not needed in this case
            LyClassInfo classInfo = ((LyStructFieldInfo)_template).getClassInfo();

            addClassInfo(classInfo, null, -1);
            return classInfo;
        }

        // Create "internal" class info
        LyClassInfo templateClassInfo = null;

        if (_template != null)
            templateClassInfo = ((LyStructFieldInfo)_template).getClassInfo();

        LyStructClassInfo classInfo = new LyStructClassInfo(this, null,
                templateClassInfo, null, false);

        // Add class to relation
        addClassInfo(classInfo, null, -1);
        return classInfo;
	}

	/**
	 * Returns a formatted "edition string" of the specified value for this field formatted by field
	 * unit format if found.
	 *
	 * @param  value The internal value to convert to an edition string.
	 * @return The value presented as a formatted edition String.
	 */
	protected String getFormattedString(Object fieldValue)
	{
		if ((_fieldTag) && (!_multiple))
		{
			LyObjectId		  objectId = ((fieldValue == null) ? null : (LyObjectId)fieldValue);
			LyObject		  object = ((objectId == null) ? null : objectId.getLogicalObject());
			LyClassInfo		  subStructClass = ((object == null) ? getClassInfo()
																 : object.getClassInfo());
			LyStructFieldInfo subStruct;

			if (subStructClass.isTemporary() && (subStructClass.getParent() instanceof LyClassInfo))
				subStruct = (LyStructFieldInfo)subStructClass.getParent().getParent();
			else
				subStruct = (LyStructFieldInfo)subStructClass.getParent();
			if ((subStruct._unit == null) || (subStruct._unit.getFormat() == null))
				return null;
			return subStruct._unit.getFormat().toString(subStruct, fieldValue, subStruct._unit);
		}
		else
			return super.getFormattedString(fieldValue);
	}

	/**
	 * Returns a formatted "displayable string" of the specified value for this field formatted by
	 * field unit format if found.
	 *
	 * @param  value The internal value to convert to a displayable string.
	 * @return The value presented as a formatted displayable String.
	 */
	protected String getFormattedName(Object fieldValue)
	{
		if (hasMark(MARK_SECRET))
			return SECRET_VALUE;
		
		if ((_fieldTag) && (!_multiple))
		{
			LyObjectId		  objectId = ((fieldValue == null) ? null : (LyObjectId)fieldValue);
			LyObject		  object = ((objectId == null) ? null : objectId.getLogicalObject());
			LyClassInfo		  subStructClass = ((object == null) ? getClassInfo()
																 : object.getClassInfo());
			LyStructFieldInfo subStruct;

			if (subStructClass.isTemporary() && (subStructClass.getParent() instanceof LyClassInfo))
				subStruct = (LyStructFieldInfo)subStructClass.getParent().getParent();
			else
				subStruct = (LyStructFieldInfo)subStructClass.getParent();

			if ((subStruct._unit == null) || (subStruct._unit.getFormat() == null))
				return null;
			
			return subStruct._unit.getFormat().toName(subStruct, fieldValue, subStruct._unit);
		}
		else
			return super.getFormattedName(fieldValue);
	}

	/**
	 * Returns the specified value for this field as a short "displayable" string formatted by field
	 * unit format if found.
	 *
	 * @param  value The internal value to convert to a short displayable string.
	 * @return The value presented as a short formatted displayable String.
	 */
	protected String getFormattedShortName(Object fieldValue)
	{
		if (hasMark(MARK_SECRET))
			return SECRET_VALUE;
		
		if ((_fieldTag) && (!_multiple))
		{
			LyObjectId		  objectId = ((fieldValue == null) ? null : (LyObjectId)fieldValue);
			LyObject		  object = ((objectId == null) ? null : objectId.getLogicalObject());
			LyClassInfo		  subStructClass = ((object == null) ? getClassInfo()
																 : object.getClassInfo());
			LyStructFieldInfo subStruct;

			if (subStructClass.isTemporary() && (subStructClass.getParent() instanceof LyClassInfo))
				subStruct = (LyStructFieldInfo)subStructClass.getParent().getParent();
			else
				subStruct = (LyStructFieldInfo)subStructClass.getParent();
			if ((subStruct._unit == null) || (subStruct._unit.getFormat() == null))
				return null;
			
			return subStruct._unit.getFormat().toName(subStruct, fieldValue, subStruct._unit);
		}
		else
			return super.getFormattedShortName(fieldValue);
	}

	/**
	 * Compares two objects for this fieldInfo.
	 */
	public boolean deepEquals(Object value1, Object value2, BitSet modifiers)
	{
		// We may compare relations first : if it's the same object then ok.
		if (equals(value1, value2, modifiers))
			return true;

		// The object differs, we may compare all values
		LyObject object1 = null, object2 = null;

		if (value1 instanceof LyObject)
			object1 = (LyObject)value1;
		else if (value1 instanceof LyObjectId)
			object1 = ((LyObjectId)value1).getLogicalObject();

		if (value2 instanceof LyObject)
			object2 = (LyObject)value2;
		else if (value2 instanceof LyObjectId)
			object2 = ((LyObjectId)value2).getLogicalObject();

		if ((object1 != null) && (object2 != null))
		{
			// We may check better : look if values inside the stuct differ
			LyValueSet  valuesObj1 = object1.getValues();
			LyValue     valueObj1, valueObj2;
			LyFieldInfo fieldInfo;
			int		    n = valuesObj1.getSize();

			// go through values and stop on any difference
			if (n > 0)
			{
				int i;

				for (i = 0; i < n; i++)
				{
					valueObj1 = valuesObj1.getFieldValue(i);
					fieldInfo = (LyFieldInfo)valueObj1.getField();
					valueObj2 = object2.getFieldValue(fieldInfo.getId());

					if (!fieldInfo.equals(valueObj1.getValue(), valueObj2.getValue(), modifiers))
					{
						// A difference on id only is not taken into account
						if (fieldInfo.hasMark(LyFieldInfo.MARK_ID))
							continue;
						if (fieldInfo instanceof LyArrayFieldInfo)
							// Struct of Array ?
							return ((LyArrayFieldInfo)fieldInfo).deepEquals(valueObj1.getValue(),
								valueObj2.getValue(), modifiers);

						return false;
					}
				}
			}

			return true;
		}

		return false;
	} // end method deepEquals

	/**
	 * Compares two objects for this fieldInfo.
	 */
	public int deepCompare(Object value1, Object value2, BitSet modifiers)
	{
		// We may compare relations first : if it's the same object then ok.
		int comp = super.compare(value1, value2, modifiers);

		if (comp == 0)
			return comp;
		// The object differs, we may compare all values
		LyObject object1 = null, object2 = null;

		if (value1 instanceof LyObject)
			object1 = (LyObject)value1;
		else if (value1 instanceof LyObjectId)
			object1 = ((LyObjectId)value1).getLogicalObject();
		if (value2 instanceof LyObject)
			object2 = (LyObject)value2;
		else if (value2 instanceof LyObjectId)
			object2 = ((LyObjectId)value2).getLogicalObject();

		if ((object1 != null) && (object2 != null))
		{
			// We may check better : look if values inside the stuct differ
			LyValueSet  valuesObj1 = object1.getValues();
			LyValue     valueObj1, valueObj2;
			LyFieldInfo fieldInfo;
			// go through values and stop on any difference
			if (valuesObj1.getSize() > 0)
			{
				comp = 0; // For the moment, equals...
				for (int i = 0; (i < valuesObj1.getSize()) && (comp == 0); i++)
				{
					valueObj1 = valuesObj1.getFieldValue(i);
					fieldInfo = (LyFieldInfo)valueObj1.getField();
					valueObj2 = object2.getFieldValue(fieldInfo.getId());

					comp = fieldInfo.compare(valueObj1.getValue(), valueObj2.getValue(), modifiers);
				}
			}
		}
		return comp;
	} // end method deepCompare

	/**
	 * Add a new label to labels list of this field info.
	 *
	 * @param label Label to add.
	 */
	void addLabel(LyInfoLabel label)
	{
        LyInfoLabel.Type type = label.getType();

		if (type != LyInfoLabel.Type.ID)
			super.addLabel(label);

		LyClassInfo classInfo = getClassInfo();

		if ((classInfo != null)
				&& (classInfo.getParent() == this))
		{
			if ((type == LyInfoLabel.Type.USER)
					|| (type == LyInfoLabel.Type.ID))
				classInfo.addLabel(label);
			else if ((type == LyInfoLabel.Type.VALUE)
					|| (type == LyInfoLabel.Type.SHORT_VALUE))
			{
				// Add VALUE label as NAME label for the classInfo corresponding to this struct
				// after removing first step
				LyInfoLabel		   newLabel;
				LyApplication	   application = getApplication();
				LyInfoList<LyPath> paths = label.getPaths();
				int				   i, n = ((paths == null) ? 0 : paths.getSize());
				LyPath			   path, newPath;
				LyInfoList<LyUnit> units = label.getUnits();
				int				   nu = ((units == null) ? 0 : units.getSize());
				LyUnit			   unit;
				Vector<LyStep>	   steps;
				LyStep			   step;
				int				   j, p;

				newLabel = new LyInfoLabel(null, label.getValue(), LyInfoLabel.Type.NAME,
						label.getSubType());
				newLabel.setModelType(label.getModelType());
				newLabel.setApplication(application);
				newLabel.setParent(this);

				// Search in all paths
				for (i = 0; i < n; i++)
				{
					path = paths.getInfo(i);
					unit = ((i < nu) ? units.getInfo(i) : null);

					newPath = new LyPath(null, path.getTarget(), path.getSourceClassInfo(),
							path.getTargetFieldInfo(), path.getFieldLabel());
					newPath.setModelType(path.getModelType());
					newPath.setApplication(application);
					newPath.setParent(newLabel);

					steps = path.getSteps();
					p = ((steps == null) ? 0 : steps.size());

					// Add all steps except first one
					for (j = 1; j < p; j++)
					{
						step = steps.elementAt(j);
						newPath.addStep(step);
					}
					newLabel.addPart(newPath, unit);
				}

				classInfo.addLabel(newLabel);
			}
		}
	} // end method addLabel

	/**
	 * Adds a control on this field info.
	 *
	 * @param control Control to add to this field info.
	 */
	void addControl(LyControl control)
	{
		super.addControl(control);

		LyClassInfo classInfo = getClassInfo();

		classInfo.addControl(control);
	}

	/**
	 * Replace the classInfos pointed by the structField.
	 *
	 * @param classInfos The new pointed classInfos.
	 */
	public void replaceClassInfos(LyInfoList<LyClassInfo> classInfos)
	{
		_classes = new LyInfoDataList<LyClassInfo, LyClassRef>();
		_classes.addList(classInfos);
	}

	/**
	 * Add parent class for this struct.
	 */
	protected void addParentClass(LyClassInfo classInfo)
	{
		addParentClass(classInfo, this);
	}

	/**
	 * Add parent class for this struct.
	 */
	void addParentClass(LyClassInfo classInfo, LyFieldInfo fieldInfo)
	{
		if (!(fieldInfo instanceof LyStructFieldInfo)
				&& (!(fieldInfo instanceof LyArrayFieldInfo)
					|| !(((LyArrayFieldInfo)fieldInfo).getFieldInfo() instanceof LyStructFieldInfo)))
			return;

		// We must add the parent class for all struct fieldInfos of the pointed
		// classInfos : if we have a struct with a list of subtypes, all subtypes
		// may have the classInfo as parent.
		int		    i, n = getClassInfoCount();
		LyClassInfo structClassInfo;
		LyInfo	    parent;

		for (i = 0; i < n; i++)
		{
			structClassInfo = getClassInfo(i);
			parent = structClassInfo.getParent();

			if ((parent instanceof LyStructFieldInfo) && (parent != this))
				((LyStructFieldInfo)parent).addParentClass(classInfo, fieldInfo);
		}

		if (_parentClasses == null)
			_parentClasses = new LyInfoDataList<LyClassInfo, LyFieldInfo>(false);
		_parentClasses.addInfo(classInfo, fieldInfo);

	} // end method addParentClass

	/**
	 * Remove parent class for this struct.
	 */
	protected void removeParentClass(LyClassInfo classInfo)
	{
		removeParentClass(classInfo, this);
	}

	/**
	 * Remove parent class for this struct.
	 */
	void removeParentClass(LyClassInfo classInfo, LyFieldInfo fieldInfo)
	{
		LyClassInfo structClassInfo = getClassInfo(0);

		if (structClassInfo.getParent() != this)
			((LyStructFieldInfo)getTemplate()).addParentClass(classInfo, fieldInfo);
		else if (_parentClasses != null)
			_parentClasses.removeInfo(classInfo, fieldInfo);
	}

	/**
	 * Gets parent informations for the specified struct object.
	 *
	 * @param  object The specified struct object.
	 * @return null or a vector with parent informations : the first element is the parent object,
	 *         the second element is the parent fieldInfo.
	 */
	public Vector<Object> getParentInfos(LyObject object)
	{
		if (_parentClasses == null)
			return null;

		// Search through parent classes
		Vector<Object>     result = null;
		LyDataSession	   session = _application.getCurrentSession();
		int				   i, n;
		LyClassInfo		   parentClass;
		LyObjectList	   parents;
		LyFilter		   relFilter;
		LyFilterExpression filterExp;
		LyFieldInfo		   fieldInfo;
		boolean			   local;
		int				   p;

		for (p = 0; p < 2; p++)
		{
			local = (p == 0); // First pass is only local
			n = _parentClasses.getSize();
			for (i = n - 1; i >= 0; i--)
			{
				parentClass = _parentClasses.getInfo(i);
				if (parentClass.isTemporary())
				{
					_parentClasses.removeInfoAt(i);
					continue;
				}

				fieldInfo = _parentClasses.getData(i);
				if (fieldInfo instanceof LyStructFieldInfo)
					filterExp = new LyFilterExpression(fieldInfo, false,
							LyFilterExpression.Condition.EQUALS, object.getObjectId());
				else
					filterExp = new LyFilterExpression(fieldInfo, false,
							LyFilterExpression.Condition.CONTAINS, object.getObjectId());

				relFilter = new LyFilter();
				filterExp.setModifier(LyFilterExpression.MOD_CASE_SENSITIVE);
				relFilter.addElement(filterExp);

				// Get filtered list
				parents = new LyClassWorkSpace(session, parentClass, null, relFilter, null, false,
						false, local);

				if ((parents != null) && !parents.isEmpty())
				{
					if (parents.getSize() == 1)
					{
						LyObject parent = parents.getObject(0);

						if (!local && !relFilter.match(parent))
							// Could occur when parent has a new or pending value for fieldInfo
							continue;

						result = new Vector<Object>(2);
						result.addElement(parent);
						result.addElement(fieldInfo);
						return result;
					}
					else
						throw new IllegalStateException(
							"BF AppEngine : more than one parent object found on object "
							+ object + " with parent class " + parentClass);
				}
			}
		}

		return null;
	} // end method getParentInfos

	/**
	 * Get parent of the given object.
	 */
	public LyObject getParent(LyObject object)
	{
		Vector<Object> parentInfos = getParentInfos(object);

		if (parentInfos == null)
			return null;

		return (LyObject)parentInfos.elementAt(0);
	}

	/**
	 * Gets list of parent classes.
	 */
	public LyInfoDataList<LyClassInfo, LyFieldInfo> getParentClasses()
	{
		return ((_parentClasses == null) ? null
            : new LyInfoDataList<LyClassInfo, LyFieldInfo>(_parentClasses));
	}
    
    /**
     * Sets list of field infos of this struct.
     *
     * @param fieldInfos List of field infos, can't be null since a class info with no fields is not accepted.
     */
    protected void setFieldInfos(LyInfoList<LyFieldInfo> fieldInfos)
    {
        getClassInfo().setFieldInfos(fieldInfos);
    }
    
    /**
     * Sets the struct that is extended by this struct.
     *
     * @param extendedStruct Struct extended by this struct.
     */
    protected void setExtendedStruct(LyStructFieldInfo extendedStruct)
    {
        setExtendedRelation(extendedStruct);
    }
    
    /**
     * Sets class info corresponding to this struct.
     */
    protected void setClassInfo(LyClassInfo classInfo)
    {
        addClassInfo(classInfo, null, -1);
    }
    
    /**
     * Sets cache policy to the associated class info.
     */
    protected void setCachePolicy(LyInfoCache cachePolicy)
    {
        getClassInfo().setCachePolicy(cachePolicy);
    }
    
    /**
     * Sets rules to the associated class info.
     */
    protected void setRules(LyInfoList<LyRule> rules)
    {
        //getClassInfo().setRules(rules);
    }
    
    /**
     * Sets behavior from the associated class info.
     */
    protected void setClassBehaviorClassName(String behaviorClassName)
    {
        getClassInfo().setBehaviorClassName(behaviorClassName);
    }
    
    /**
     * Sets binding from the associated class info.
     */
    protected void setClassBinding(LyClassBinding classBinding)
    {
        getClassInfo().setBinding(classBinding);
    }
} // end class LyStructFieldInfo
