/*
 * LyTabAction.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;

import leon.info.misc.*;


/**
 * This class implements the description of tab actions : action containing a list of sub-actions.
 * Tab actions are used to allow building of views presented as tabs. Sub-actions of a tab action
 * can also be tab or tab actions this gives us the richness for building complex presentations.<br>
 * This class extends LyAction : a tab action is an action and can be used everywhere actions are
 * used.<br>
 * The extra attributes on tab actions are the following :
 *
 * <ul>
 *   <LI>a list of actions.
 *
 *     <p>See also tag <TAB_ACTION> of leon.dtd</p>
 *   </LI>
 * </ul>
 */
public final class LyTabAction extends LyAction
{
	public static final String __VERSION =
		"$Id: LyTabAction.java 33087 2011-11-08 14:33:04Z kchenal $";

	/** Context : none. */
	public static short		   CONTEXT_NONE = -1;

	/** Context : both. */
	public static short		   CONTEXT_ALL = -2;

    /** Tab action types. */
    public enum Type
    {
    	// Type : tabs.
    	TABS,
    	// Type : wizard.
    	WIZARD
    }

	/** First action. */
	LyInfoList<LyAction>	   _tabActions;

	/**
	 * Index of action that gives context to client action, accepted values are : -2 : all actions
	 * give context, -1 : no context, i >= 0 : context is given by action at the specified index.
	 */
	short					   _clientContextIndex;

	/** Type of tab action. */
	Type					   _type;

	/**
	 * Initialization of action attributes.
	 */
	protected void init()
	{
		super.init();

		// Initialize class attributes.
		_tabActions = null;
		_clientContextIndex = CONTEXT_ALL;
		_type = Type.TABS;
	}

	/**
	 * Constructor of a new tab action.
	 *
	 * @param id                  Identifier of the tab action.
	 * @param template            Reference of the action template.
	 * @param controllerClassName Java class name of the action controller.
	 * @param behaviorClassName   Java class name of the action behavior.
	 * @param builderClassName    Java class name of the action builder.
	 * @param clientContextIndex  Client context index. Accepted values are : -2 = both, -1 = no
	 *                            context, 0 = context is given by first action, 1 = context is
	 *                            given by second action.
	 * @param target              The target zone id for this action.
	 * @param type                Type of the tab action.
	 */
	public LyTabAction(String id, LyAction template, String controllerClassName,
		String behaviorClassName, String builderClassName, short clientContextIndex,
		String target, Type type)
	{
		this(id);
		construct(template, controllerClassName, behaviorClassName, builderClassName,
			clientContextIndex, target, type);
	}

	/**
	 * Constructor of an external tab action.
	 *
	 * @param id Identifier of the tab action.
	 */
	LyTabAction(String id)
	{
		super(id);
	}

    protected void setType(Type type)
    {
        if (type == Type.TABS)
        {
            // Initialize mark noTitlebar='true' -> to keep compatibility with previous version
            // to force titleBar , add noTitlebar='false' + operation='RESET' in metamodel
            BitSet marks = getMarks();

            if (marks == null)
                marks = new BitSet();

            marks.set(LyAction.MARK_NO_TITLEBAR);
            marks.set(LyAction.MARK_FRAME_VIEW);

            setMarks(marks);
        }

        _type = (type == null ? Type.TABS : type);
    }
    
    /**
     * Specifies if the client context action is the action at the given index, none or all of them.
     */
    protected void setClientContextIndex(short clientContextIndex)
    {
        _clientContextIndex = clientContextIndex;
    }
    
	/**
	 * Construction of a new tab action.
	 *
	 * @param template            Reference of the action template.
	 * @param controllerClassName Java class name of the action controller.
	 * @param behaviorClassName   Java class name of the action behavior.
	 * @param builderClassName    Java class name of the action builder.
	 * @param clientContextIndex  Client context index. Accepted values are : -2 = both, -1 = no
	 *                            context, 0 = context is given by first action, 1 = context is
	 *                            given by second action.
	 * @param target              The target zone id for this action.
	 */
	protected void construct(LyAction template, String controllerClassName,
		String behaviorClassName, String builderClassName, short clientContextIndex,
		String target, Type type)
	{
		super.construct(template, controllerClassName, behaviorClassName, builderClassName, null,
			target);

		setType(type);
		_clientContextIndex = clientContextIndex;
	}

	/**
	 * Init this instance attributes from its template.
	 *
	 * @param parts Indicates which parts of the template are copied.
	 */
	void initFromTemplate(int parts)
	{
        parts &= ~_initParts;
        if (parts == 0)
            return;
		if (_template == null)
			return;

		super.initFromTemplate(parts);

		if (_template instanceof LyTabAction)
		{
			LyTabAction template = (LyTabAction)_template;

			if (template._tabActions != null)
				_tabActions = new LyInfoList<LyAction>(template._tabActions);
		}
	}

	/**
	 * Final step at the end of the parsing.
	 *
	 * @exception IllegalArgumentException If the one of the action in tab action is not viewable.
	 */
	void check(LyObjectInfo parent)
	{
		super.check(parent);

        // setting the default controller
        if (_controllerClassName == null)
            _controllerClassName = new LyJavaName(this, "leon.control.LyTabAreaController");
        
        //adding the notformobile mark
        if (_marks == null)
			_marks = new BitSet();
        //_marks.set(LyAction.MARK_NOTFORMOBILE);
        
		int		 i, n = ((_tabActions == null) ? 0 : _tabActions.getSize());
		LyAction action;

		for (i = 0; i < n; i++)
		{
			action = _tabActions.getInfo(i);
			if ((action == null) || (action.isLoaded() && !action.isViewable()))
				throw new IllegalArgumentException("BF AppEngine : invalid action " + action
					+ " used in tab action " + this
					+ ", only viewable actions can be used in a tab action");
		}
	}

	/**
	 * Lock this tab action.
	 */
	protected void lock()
	{
		if (!isLoaded())
			return;

		super.lock();

		// Turn tab actions list to unwritable
		if (_tabActions != null)
			_tabActions.lock();
	}

	/**
	 * Gets sub-actions list.
	 *
	 * @return List of these actions.
	 */
	public LyInfoList<LyAction> getTabActions()
	{
		return _tabActions;
	}

	/**
	 * Gets sub-action at the given index in sub-actions list.
	 *
	 * @param  index Index of the searched sub-action.
	 * @return Sub-action at the given index, null if no sub-action is defined.
	 */
	public LyAction getTabAction(int index)
	{
		return ((_tabActions == null) ? null : _tabActions.getInfo(index));
	}

	/**
	 * Gets number of sub-actions.
	 *
	 * @return Number of sub-actions of this action.
	 */
	public int getTabActionCount()
	{
		return ((_tabActions == null) ? 0 : _tabActions.getSize());
	}

	/**
	 * Specifies if the client context action is the action at the given index, none or all of them.
	 *
	 * @return the client context index
	 */
	public short getClientContextIndex()
	{
		return _clientContextIndex;
	}

	/**
	 * Gets type of the tab action.
	 *
	 * @return The tab action type.
	 */
	public Type getType()
	{
		return _type;
	}

	/**
	 * Add an sub-action for this action at the given position.
	 *
	 * @param action Sub-action to add to this action.
	 */
	void addTabAction(LyAction action)
	{
		if (_tabActions == null)
			_tabActions = new LyInfoList<LyAction>(false);

		_tabActions.addInfo(action);
	}
	
	/**
     * Sets sub-actions list.
     *
     * @param tabActions List of these actions.
     */
    void setTabActions(LyInfoList<LyAction> tabActions)
    {
        _tabActions = (tabActions == null ? null : new LyInfoList<LyAction>(tabActions));
    }
    
    /**
     * Method used to check if this action is viewable or not.
     * @return True.
     */
    public boolean isViewable()
    {
        return true;
    }
    
    /**
     * Gets behavior class name.
     *
     * @return Java name of behavior class (may be null).
     */
    public String getBehaviorClassName()
    {
    	String className = super.getBehaviorClassName();
    	if (className == null)
    		className="leon.app.behavior.LyTabActionBehavior";
    	return className;
    }
} // end class LyTabAction
