/**
 * LyInfo.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.info;

import leon.app.*;
import leon.misc.*;

import java.lang.reflect.Field;
import java.util.*;


/**
 * This class implements a simple object built when parsing the leonardi model.
 */
public abstract class LyTag
	implements Cloneable
{
	public static final String __VERSION = "$Id: LyInfo.java 16530 2009-02-24 12:44:18Z jls $";

	/** List of all models. */
	public enum Model
	{
		FULL, APP, BIND, DATA, GUI, LABEL
	}

	/** Info parent. */
	protected LyInfo   _parent;

	/** Type of the model that defines this info. */
	protected Model    _modelType;

	/** Comments read just before this tag. */
	protected String   _commentsStart;

	/** Comments read just after this tag. */
	protected String   _commentsEnd;
	
	/** Cloned indicator.*/
	protected boolean  _cloned;

	/**
	 * Default constructor with no parameter. Since no id is given, an id is automatically
	 * attributed to this instance.
	 */
	protected LyTag()
	{
		_parent = null;
		_modelType = Model.FULL;
		_commentsStart = null;
		_commentsEnd = null;
		_cloned = false;
	}

	/**
	 * Internal method to free an info : remove it from its application.
	 */
	protected void free()
	{
		LyApplication application = getApplication();

		if (application != null)
			application.removeTag(this);

		_parent = null;
	}

	/**
	 * Accessor for the id of the info.
	 *
	 * @return  Idenfifier of the info.
	 */
	public String getId()
	{
		return Integer.toString(hashCode());
	}

	/**
	 * Accessor for the application to which the info belongs.
	 *
	 * @return  Application of the info.
	 */
	public LyApplication getApplication()
	{
		return ((_parent == null) ? null : _parent.getApplication());
	}

	/**
	 * Gets parent of this info.
	 *
	 * @return  Parent of this info.
	 */
	public final LyInfo getParent()
	{
		return _parent;
	}
	
    /**
     * Checks if this info has given parent as ancestor.
     * @param parent    Parent that is checked as ancestor.
     * @return True if given parent is ancestor of this info.
     */
    public final boolean hasAncestor(LyInfo parent)
    {
        if (_parent == parent)
            return true;
        else if (_parent == null)
            return false;
        else
            return _parent.hasAncestor(parent);
    }

	/**
	 * Gets model type that defines this info.
	 *
	 * @return  TYPE_DATA, TYPE_GUI or TYPE_BIND.
	 */
	public final Model getModelType()
	{
		return _modelType;
	}

	/**
	 * Gets project to which this info belongs to.
	 *
	 * @return  Project of this info.
	 */
	public final LyProject getRootProject()
	{
		if ((_parent == null)
			|| (_parent instanceof LyInclude))
		{
			if (this instanceof LyProject)
				return (LyProject)this;
			else
				return null;
		}
		else
			return _parent.getRootProject();
	}

	/**
	 * To get the current environment when using the info.
	 *
	 * @return  The current environment.
	 */
	public final LyEnvironment getEnvironment()
	{
		LyApplication application = getApplication();

		// Application may not be initialized when creating a new LyInfo
		if (application == null)
		{
			// Try anyway
			Vector<LyApplication> apps = LyApplication.getApplications();
			int					  na = ((apps == null) ? 0 : apps.size());

			if (na > 0)
				application = apps.firstElement();

			if (application == null)
				return null;
		}

		LyEnvironment environment = null;
		LyDataSession currentSession = application.getCurrentSession();

		if (currentSession != null)
			environment = currentSession.getEnvironment();

		if (environment == null)
			environment = application.getEnvironment();

		return environment;
	}

	/**
	 * Checks if an info is declared as extern or not (comes from the include of an extern model).
	 *
	 * @return  True if the info is extern, false otherwise.
	 */
	public final boolean isExtern()
	{
		if (_parent == null)
			return false;
		else if (_parent instanceof LyInclude)
			return true;
		else
			return _parent.isExtern();
	}

	/**
	 * Sets the parent of this info.
	 *
	 * @param  the  parent of this info
	 */
	protected void setParent(LyInfo parent)
	{
		if (_parent == parent)
			return;

		if ((_parent != null)
				&& (parent != null))
			// Tag has already a parent
			return;

		if (parent != null)
		{
			// Register tag on new parent application
		    LyApplication application = parent.getApplication();
		    
		    if (application != null)
		        application.putTag(this);
		}
		else
		{
			// Unregister tag on previous parent application
	         LyApplication application = _parent.getApplication();
	            
	        if (application != null)
	            application.removeTag(this);
		}
		
	    _parent = parent;
	} // end method setParent

	/**
	 * Sets model type that defines this model.
	 *
	 * @param  modelType  MODEL_DATA, MODEL_GUI, MODEL_LABEL or MODEL_BIND.
	 */
	protected final void setModelType(Model modelType)
	{
		_modelType = modelType;
	}

	/**
	 * Store comments before the tag (comments at Start).
	 *
	 * @param  comments  comments to store
	 */
	public final void setCommentsStart(String comments)
	{
		_commentsStart = comments;
	}

	/**
	 * Store comments after the tag (comments at END).
	 *
	 * @param  comments  comments to store
	 */
	public final void setCommentsEnd(String comments)
	{
		_commentsEnd = comments;
	}

	/**
	 * Accessor to Stored comments before the tag (comments at Start).
	 *
	 * @return  starts comments on tag
	 */
	public final String getCommentsStart()
	{
		return _commentsStart;
	}

	/**
	 * Accessor to Stored comments after the tag (comments at END).
	 *
	 * @return  comments comments to store
	 */
	public final String getCommentsEnd()
	{
		return _commentsEnd;
	}
	
	/**
	 * Checks if this tag was created by clone method (used by AC).
	 * @return True if this tag is a result of clone method.
	 */
	public final boolean wasCloned()
	{
	    return _cloned;
	}
	
	/**
	 * Implementation of the Cloneable interface.
	 */
	public synchronized Object clone()
	{
		try
		{
			LyTag clone = (LyTag)super.clone();

			return clone;
		}
		catch (CloneNotSupportedException e)
		{
			// This shouldn't happen, since we are Cloneable.
			throw new InternalError();
		}
	}
	
	/**
     * Clone tag.
     * @return Cloned tag.
     */
    @SuppressWarnings({ "unchecked", "rawtypes" })
	protected LyTag cloneTag()
    {
        // Clone tag
        LyTag		newTag = (LyTag)clone();
        Class<?>	tagClass = newTag.getClass();
        Field[]		fields;
        int			i, nf;
        Object		value, newValue;
        
        try
        {
        	do
        	{
                fields = tagClass.getDeclaredFields();
                nf = fields.length;

		        for (i = 0; i < nf; i++)
		        {
		        	fields[i].setAccessible(true);
		        	
		        	value = fields[i].get(this);
		        	
		        	if (value instanceof BitSet)
		        		newValue = ((BitSet)value).clone();
		        	else if (value instanceof LyInfoDataList)
		        		newValue = new LyInfoDataList((LyInfoDataList)value);
		        	else if (value instanceof LyInfoList)
		        		newValue = new LyInfoList((LyInfoList)value);
		        	else if (value instanceof Vector)
			        	newValue = ((Vector)value).clone();
		        	else if (value instanceof Hashtable)
		        		newValue = ((Hashtable)value).clone();
		        	else if (value != null && (tagClass.isArray()))
		        		newValue = ((Object[])value).clone();
		        	else
		        		newValue = value;
		        		
		        	if (newValue != value)
		        		fields[i].set(newTag, newValue);
		        	
		        	fields[i].setAccessible(false);
		        }
		        
		        tagClass = tagClass.getSuperclass();
        	}
        	while (tagClass != null);
        	
        }
        catch (Exception e)
        {
        	getEnvironment().logError(e.getMessage());
        }
        
        newTag._commentsStart = null;
        newTag._commentsEnd = null;
        newTag._cloned = true;
        return newTag;
    }
} // end class LyInfo
