/*
 * LyTextFieldInfo.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.util.*;

import leon.app.*;
import leon.data.*;
import leon.info.infointerface.*;
import leon.misc.*;


/**
 * This class implements the description of text fields. Text fields may contain characters as
 * simple text (ASCII/Unicode) or will use other text formating protocols in future versions : HTML,
 * RTF. Displaying the content of this field will be up to the implementation of the graphical
 * component.
 *
 * <p>In addition to standard fieldInfo properties, a textFieldInfo may contain some extra
 * information about the way the text should be shown (number of rows/columns) or structured
 * (format/number of characters allowed).</p>
 *
 * <p>A text field may also be simple or multiple. This means that user inputs can be mono or
 * multiline.</p>
 *
 * <p>See also tag <TEXT> of leon.dtd</p>
 */
public final class LyTextFieldInfo extends LyFieldInfo
{
	public static final String __VERSION =
		"$Id: LyTextFieldInfo.java 41345 2012-10-25 15:47:42Z jlsaouli $";

	public static final String	OPEN_START_LANG = "<-";
	public static final String	CLOSE_START_LANG = "->";
	public static final String	OPEN_END_LANG = "</-";
	public static final String	CLOSE_END_LANG = "->";

	/** Text types. */
    public enum Type
    {
        // Standard text
        STANDARD,
        // Rich text.
        RICH
    }

	/** Indicates if this text field is multiple or not. */
	boolean					   _multiple;

	/**
	 * Number of rows which should be shown when displaying the text (-1 if no preference is given).
	 */
	int						   _nRows;

	/**
	 * Number of columns which should be shown when displaying the text (-1 if no preference is
	 * given).
	 */
	int						   _nCols;

	/** Minimum number of chars that may be set in this textInfo (-1 if no limit is given). */
	int						   _minChars;

	/** Maximum number of chars that may be set in this textInfo (-1 if no limit is given). */
	int						   _maxChars;

	/** Regular expression. */
	LyRegExp				   _regExp;

	/** Characters types. */
	Vector<LyInfoCharType>     _charTypes;

	/** Name of the java class used for encoding. */
	String					   _encoding;

	/** Ignore accent mode. */
	Boolean					   _ignoreAccent;

	/** Text type. */
	Type                       _type;

	/** Indicates if this text field is multilingual or not. */
	boolean					   _multilingual;


	/**
	 * Init method.
	 */
	protected void init()
	{
		super.init();

		_multiple = false;
		_nRows = -1;
		_nCols = -1;
		_minChars = -1;
		_maxChars = -1;
		_regExp = null;
		_charTypes = null;
		_encoding = null;
		_ignoreAccent = null;
		_type = Type.STANDARD;
		_multilingual = false;
	}

	/**
	 * Init this instance attributes from its template.
	 *
	 * @param parts Indicates which parts of the template are copied.
	 */
	void initFromTemplate(int parts)
	{
        parts &= ~_initParts;
        if (parts == 0)
            return;
		if (_template == null)
			return;

		super.initFromTemplate(parts);

		LyTextFieldInfo template = (LyTextFieldInfo)_template;

		if ((parts & PART_DATA) > 0)
		{
			_minChars = template._minChars;
			_maxChars = template._maxChars;
			_regExp = template._regExp;
			if (template._charTypes != null)
				_charTypes = new Vector<LyInfoCharType>(template._charTypes);
			_encoding = template._encoding;
		}
		if ((parts & PART_GUI) > 0)
		{
			_nRows = template._nRows;
			_nCols = template._nCols;
		}
	}

	/**
     * Constructor of a text field.
     *
     * @param id       Identifier of the field.
     * @param javaName Name of the generated java field.
     * @param template Template used to construct this field.
     * @param multiple Indicates is the text field is multiple or not.
     */
    LyTextFieldInfo(String id, String javaName, LyTextFieldInfo template,
        boolean multiple)
    {
        this(id, javaName, template, multiple, Type.STANDARD, false);
    }

	/**
	 * Constructor of a text field.
	 *
	 * @param id       Identifier of the field.
	 * @param javaName Name of the generated java field.
	 * @param template Template used to construct this field.
	 * @param multiple Indicates is the text field is multiple or not.
	 * @param type     Text type.
	 */
	LyTextFieldInfo(String id, String javaName, LyTextFieldInfo template,
		boolean multiple, Type type)
	{
		this(id, javaName, template, multiple, type, false);
	}

	/**
	 * Constructor of a text field.
	 *
	 * @param id       	   Identifier of the field.
	 * @param javaName     Name of the generated java field.
	 * @param template	   Template used to construct this field.
	 * @param multiple	   Indicates is the text field is multiple or not.
	 * @param type     	   Text type.
	 * @param multilingual Indicates is the text field is multilingual or not.
	 */
	LyTextFieldInfo(String id, String javaName, LyTextFieldInfo template,
		boolean multiple, Type type, boolean multilingual)
	{
		this(id);
		construct(javaName, template, multiple, type, multilingual);
	}

	/**
	 * Constructor of a text field.
	 *
	 * @param id Identifier of the field.
	 */
	LyTextFieldInfo(String id)
	{
		super(id);
	}

	/**
	 * Construction of a text field.
	 *
	 * @param javaName Name of the generated java field.
	 * @param template Template used to construct this field.
	 * @param multiple Indicates is the text field is multiple or not.
	 * @param multilingual Indicates is the text field is multilingual or not.
	 */
	protected void construct(String javaName, LyTextFieldInfo template,
		boolean multiple, Type type, boolean multilingual)
	{
		super.construct(javaName, template);

		setMultiple(multiple);
		setType(type);
		setMultilingual(multilingual);
	}

	/**
	 * Final step of parsing : check consistency. Checks performed are the following :
	 *
	 * <ul>
	 *   <LI>if text is id or name it can not be multiple.</LI>
	 * </ul>
	 *
	 * @param     parent Object info in which this field is checked.
	 * @param     parts  Indicates which parts are checked
	 * @exception IllegalStateException If one of these checks fails.
	 */
	void check(LyObjectInfo parent, int parts)
	{
		super.check(parent, parts);

		if ((parts & PART_DATA) > 0)
		{
			if (isMultiple()
					&& hasMark(MARK_ID))
				throw new IllegalStateException("BF AppEngine : text fieldInfo " + this
					+ " can not be multiple since it has the ID mark.");

			if (_maxChars == 0)
				throw new IllegalStateException("BF AppEngine : text fieldInfo " + this
					+ " can't have a maximal length of 0.");

			if ((_minChars != -1)
					&& (_maxChars != -1)
					&& (_minChars > _maxChars))
				throw new IllegalStateException("BF AppEngine : text fieldInfo " + this
					+ " can't have a minimum length (" + _minChars
					+ ") greater than the maximum lenght (" + _maxChars + ").");
		}

		if ((parts & PART_GUI) > 0)
		{
			if (isMultiple()
					&& hasMark(MARK_NAME))
			{
				setMultiple(false);
				_application.getEnvironment().logWarning("BF AppEngine : remove multiple attribute from"
					+ " the text fieldInfo " + this + " which has the NAME mark.");
			}

			if ((_encoding != null)
					&& !hasMark(MARK_SECRET))
			{
				getEnvironment().logWarning("BF AppEngine : add mark SECRET for fieldInfo "
					+ this + " which has an encoding");
				if (_marks == null)
					_marks = new BitSet();
				_marks.set(MARK_SECRET);
			}
		}
	} // end method check

	/* Implementation of LyFieldInfo methods */

	/**
	 * Returns class of values for this field. The unique value class corresponding to a text field
	 * is : String.class
	 */
	public Class<String> getValueClass()
	{
		return String.class;
	}

	/**
	 * Builds the appropriate LyValue corresponding to this field for the given value object.
	 *
	 * @see LyTextValue
	 */
	public LyValue buildFieldValue(Object value)
	{
		if (value == null)
			return new LyTextValue(this);
		else if (value instanceof String)
			return new LyTextValue(this, (String)value);
		else if (value instanceof HashMap)
		{
			@SuppressWarnings("rawtypes")
			HashMap			langValues = (HashMap)value;
			LyApplication	application = getApplication();
			String[][]		languages = application.getSupportedLanguages();
			int				i, nl = (languages == null ? 0 : languages.length);
			StringBuffer	buffer = new StringBuffer();
			String			langValue;

			for (i = 0; i < nl; i++)
			{
				langValue = (String)langValues.get(languages[i][0]);
				if (langValue != null)
				{
					buffer.append(OPEN_START_LANG);
					buffer.append(languages[i][0]);
					buffer.append(CLOSE_START_LANG);

					if (langValue != null)
						buffer.append(langValue);

					buffer.append(OPEN_END_LANG);
					buffer.append(languages[i][0]);
					buffer.append(CLOSE_END_LANG);
				}
			}

			return new LyTextValue(this, buffer.toString());
		}
		else
			return new LyTextValue(this, value);
	}

	/**
	 * Encodes the given value to get the right value before transmitting it to the peer layer.
	 *
	 * @param  value Value to encode.
	 * @return The encoded value.
	 */
	public Object encodeValue(Object value)
	{
		// Extract the right value from the given one
		String text = (String)super.encodeValue(value);

		if (_encoding == null)
			return text;

		// Create encoder
		LyEncoderInterface encoder = null;

		try
		{
			Class<?> javaClass = _application.loadClass(_encoding,
					"BF AppEngine : unknown encoder class " + _encoding + " for text field " + this);

			encoder = (LyEncoderInterface)javaClass.newInstance();
		}
		catch (Exception e)
		{
			LyEnvironment environment = getEnvironment();

			environment.logError("BF AppEngine : instanciation of encoder interface "
				+ _encoding + " failed on text field " + this);
			environment.writeException(e);
			return null;
		}

		// Call encoder
		try
		{
			LyTextValue textValue = encoder.encode(new LyTextValue(this, text));

			return textValue.getValue();
		}
		catch (Exception e)
		{
			LyEnvironment environment = getEnvironment();

			environment.logError("BF AppEngine : call of encoder interface "
				+ _encoding + " failed on text field " + this);
			environment.writeException(e);
			return null;
		}
	} // end method encodeValue

	/**
	 * Builds internal value for an instance of this fieldInfo from a string (previously constructed
	 * by toString).
	 *
	 * @see LyTextFieldInfo#toString
	 */
	public Object parse(String stringValue)
		throws LyFieldInfoFormatException
	{
		if ((stringValue == null) || (stringValue.length() == 0))
			return null;

		// Nothing to do : keep parameter as is.
		return stringValue;
	}

	/**
	 * Return the specified value for this field as a short "displayable" string. This method is
	 * used when presenting values in tables. In that case, values are presented whithout unit.
	 *
	 * @see LyNumberFieldInfo#toName
	 */
	public String toShortName(Object fieldValue)
	{
		if (fieldValue == null)
			return null;

		if (fieldValue.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ fieldValue.getClass() + " expected " + getValueClass());

		String name = getFormattedShortName(fieldValue);

		if (name != null)
			return name;
		else
			return formatStringValue((String)fieldValue, false);
	}

	/**
	 * Return the specified value for this field as a "displayable" string.
	 */
	public String toName(Object fieldValue)
	{
		return toName(fieldValue, false);
	}

	/**
	 * Returns a "displayable string" (string that can be shown to the end-user) of the specified
	 * value for this field.
	 *
	 * @param  value The internal value to convert to a displayable string. The excepted class of
	 *               this value is the one returned from the getValueClass method.
	 * @param  removeTags Indicates whether HTML tags have to be removed from result.
	 * @return The value presented as a displayable String.
	 * @see    LyFieldInfo#getValueClass
	 */
	public String toName(Object fieldValue, boolean removeTags)
	{
		if (fieldValue == null)
			return null;

		if (fieldValue.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ fieldValue.getClass() + " expected " + getValueClass());

		String name = getFormattedName(fieldValue);

		if (name != null)
			return name;

		String result = formatStringValue((String)fieldValue, removeTags);

		// If any unit is specified, append it at the end (see LyFieldSetBuilder to be consistent with set forms)
		if ((result != null)
			&& (result.length() != 0)
			&& (getFieldUnit() != null))
		{
			StringBuffer  sb = new StringBuffer(result);
			String		  unitAlias, unitName;
			LyUnit		  fieldInfoUnit = getFieldUnit();
			LyEnvironment environment = getEnvironment();

			unitAlias = fieldInfoUnit.getAlias();
			unitName = fieldInfoUnit.getName();

            if ((unitAlias != null) && (unitAlias.length() > 0))
                name = unitAlias;
            else if ((unitName != null) && (unitName.length() > 0))
                name = unitName;
            else
                name = null;

			if (name != null)
			{
				if (environment != null)
					name = environment.translate(name);
				sb.append(' ').append(name);
				result = sb.toString();
			}
		}

		return result;
	} // end method toName

	/**
	 * Converts the specified value for this field to an "internal" string (string that enables
	 * reconstruction through parse).
	 *
	 * @see LyTextFieldInfo#parse
	 */
	public String toString(Object fieldValue)
	{
		if (fieldValue == null)
			return null;

		if (fieldValue.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ fieldValue.getClass() + " expected " + getValueClass());

		String	string = getFormattedString(fieldValue);

		if (string != null)
			return string;
		else
			return formatStringValue((String)fieldValue, false);
	}

	/**
	 * Format string value.
	 */
	protected String formatStringValue(String value, boolean removeTags)
	{
		String	result = value;

		if (isMultilingual()
				&& (value != null))
			result = getLangValue(null, value);

		// Remove HTML tags from rich texts if view manager does not display rich texts
		if (getType() == Type.RICH)
		{
			if (removeTags)
				result = LyTools.removeHtmlTags(value);
			else
			{
				LyDataSession			session = _application.getCurrentSession();
				LyEnvironment			environment
					= (session == null ? _application.getEnvironment() : session.getEnvironment());
				LyViewManagerInterface	viewManager = environment.getViewManager();

				if (viewManager != null)
				{
	                if (!viewManager.hasProperty("DISPLAY_RICH_TEXT"))
	                	result = LyTools.removeHtmlTags(result);
	                else
	                	result = LyTools.encodeXmlTags(result);
				}
			}
		}

		return result;
	}

	/**
	 * Gets the value for specified language from given value.
	 * @param lang	Language (null for current lang).
	 * @param value	Initial value.
	 * @return The value for given language, the initial value if not found.
	 */
	public String getLangValue(String lang, String value)
	{
		if (lang == null)
		{
			LyDataSession	session = _application.getCurrentSession();
			LyEnvironment	environment
				= (session == null ? _application.getEnvironment() : session.getEnvironment());
			
			lang = environment.getLanguage();
		}

		String	result = value;
		int		p1 = value.indexOf(OPEN_START_LANG + lang + CLOSE_START_LANG);

		if (p1 >= 0)
		{
			int		p2 = value.indexOf(OPEN_END_LANG + lang + CLOSE_END_LANG);

			if (p2 > p1)
				result = value.substring(p1 + OPEN_START_LANG.length() + lang.length() + CLOSE_START_LANG.length(), p2);
		}

		return result;
	}

	/**
	 * Adds/Replaces the value for specified language in given value.
	 * @param value	Initial value.
	 * @param lang	Language.
	 * @param value	Lang value.
	 * @return The completed value.
	 */
	public String addLangValue(String value, String lang, String langValue)
	{
		StringBuffer	buffer = new StringBuffer();

		if (value != null)
		{
			int		p1 = value.indexOf(OPEN_START_LANG + lang + CLOSE_START_LANG);

			if (p1 >= 0)
			{
				int		p2 = value.indexOf(OPEN_END_LANG + lang + CLOSE_END_LANG);

				if (p2 > p1)
				{
					buffer.append(value.substring(0, p1 + OPEN_START_LANG.length() + lang.length() + CLOSE_START_LANG.length()));
					buffer.append(langValue);
					buffer.append(value.substring(p2));
				}

				return buffer.toString();
			}
		}

		if (value != null)
			buffer.append(value);

		buffer.append(OPEN_START_LANG);
		buffer.append(lang);
		buffer.append(CLOSE_START_LANG);

		if (langValue != null)
			buffer.append(langValue);

		buffer.append(OPEN_END_LANG);
		buffer.append(lang);
		buffer.append(CLOSE_END_LANG);

		return buffer.toString();
	}

	/**
	 * Returns the filter conditions authorized on this field. The values must match the short
	 * constants defined in LyFilterExpression.
	 *
	 * @return An array of conditions corresponding to the authorized conditions on this field.
	 * @see    leon.info.LyFilterExpression
	 */
	public LyFilterExpression.Condition[] getFilterConditions()
	{
        LyFilterExpression.Condition[] conditions = new LyFilterExpression.Condition[4];

		// Default : return all the conditions.
		conditions[0] = LyFilterExpression.Condition.EQUALS;
		conditions[1] = LyFilterExpression.Condition.CONTAINS;
		conditions[2] = LyFilterExpression.Condition.STARTS_WITH;
		conditions[3] = LyFilterExpression.Condition.ENDS_WITH;

		return conditions;
	}

	/**
	 * Returns the filter modifiers authorized on this field. The values must match the short
	 * constants defined in LyFilterExpression.
	 *
	 * @return An array of short corresponding to the authorized modifiers on this field.
	 * @see    leon.info.LyFilterExpression
	 */
	public short[] getFilterModifiers()
	{
		short[]		modifiers = new short[_multilingual ? 2 : 1];

		// Default : return all the modifiers.
		modifiers[0] = LyFilterExpression.MOD_CASE_SENSITIVE;
		if (_multilingual)
			modifiers[1] = LyFilterExpression.MOD_MULTILINGUAL;

		return modifiers;
	}

	/**
	 * Compares two values for equality on this field. The excepted class of the values to compare
	 * is the one returned from the getValueClass method. The default implementation consits of
	 * comparing values using compare method and returning true when result is 0. Due to the fact
	 * that this method can be called often, this method is provided to speed up this compare when
	 * it's possible to do better.
	 *
	 * @param  value1    First value.
	 * @param  value2    Second value.
	 * @param  modifiers Modifiers used for checking (see MOD_XXX constants of LyFilterExpression).
	 * @return True if the values are equals, else false.
	 * @see    LyFieldInfo#compare
	 * @see    LyFieldInfo#getValueClass
	 */
	public boolean equals(Object value1, Object value2, BitSet modifiers)
	{
		if (value1 == value2)
			return true;

		if ((value1 == null) || (value2 == null))
			return false;

		boolean multilingual = ((modifiers == null)
				|| modifiers.get(LyFilterExpression.MOD_MULTILINGUAL));
		String	str1, str2;

		if (multilingual
			|| !isMultilingual())
		{
			str1 = (String)value1;
			str2 = (String)value2;
		}
		else
		{
			str1 = getLangValue(null, (String)value1);
			str2 = getLangValue(null, (String)value2);
		}

		// By default equals is caseSensitive, unlike compare
		boolean caseSensitive = ((modifiers == null)
					|| modifiers.get(LyFilterExpression.MOD_CASE_SENSITIVE));
		boolean ignoreAccent;

		if (_ignoreAccent == null)
		{
			LyEnvironment environment = getEnvironment();

			ignoreAccent = ((environment != null)
						&& environment.testResource("LY_IGNORE_TEXT_ACCENT"));
			_ignoreAccent = Boolean.valueOf(ignoreAccent);
		}
		else
			ignoreAccent = _ignoreAccent.booleanValue();

		if (ignoreAccent)
		{
			str1 = LyString.getSimpleString(str1);
			str2 = LyString.getSimpleString(str2);
		}

		boolean		result;

		if (caseSensitive)
			result = str1.equals(str2);
		else
			result = str1.equalsIgnoreCase(str2);

		if (result
			&& ((str1 != value1)
				|| (str2 != value2)))
		{
			if (ignoreAccent)
			{
				str1 = LyString.getSimpleString((String)value1);
				str2 = LyString.getSimpleString((String)value2);
			}
			else
			{
				str1 = (String)value1;
				str2 = (String)value2;
			}

			if (caseSensitive)
				result = str1.equals(str2);
			else
				result = str1.equalsIgnoreCase(str2);
		}

		return result;
	} // end method equals

	/**
	 * Compare two value objects for this field info.
	 */
	public int compare(Object value1, Object value2, BitSet modifiers)
	{
		if (value1 == value2)
			return 0;
		if (value1 == null)
			return -1;
		if (value2 == null)
			return 1;

		boolean multilingual = ((modifiers == null)
				|| modifiers.get(LyFilterExpression.MOD_MULTILINGUAL));
		String	str1, str2;

		if (multilingual
			|| !isMultilingual())
		{
			str1 = (String)value1;
			str2 = (String)value2;
		}
		else
		{
			str1 = getLangValue(null, (String)value1);
			str2 = getLangValue(null, (String)value2);
		}

		boolean caseSensitive = ((modifiers != null)
					&& modifiers.get(LyFilterExpression.MOD_CASE_SENSITIVE));
		boolean raw = ((modifiers != null)
					&& modifiers.get(LyFilterExpression.MOD_RAW));
		boolean ignoreAccent;

		if (_ignoreAccent == null)
		{
			LyEnvironment environment = getEnvironment();

			ignoreAccent = ((environment != null)
						&& environment.testResource("LY_IGNORE_TEXT_ACCENT"));
			_ignoreAccent = Boolean.valueOf(ignoreAccent);
		}
		else
			ignoreAccent = _ignoreAccent.booleanValue();

		if (raw
			|| ignoreAccent)
		{
			str1 = LyString.getSimpleString(str1);
			str2 = LyString.getSimpleString(str2);
		}

		int		result;

		if (caseSensitive)
			result = str1.compareTo(str2);
		else
			result = str1.compareToIgnoreCase(str2);

		if ((result == 0)
			&& ((str1 != value1)
				|| (str2 != value2)))
		{
			if (raw
				|| ignoreAccent)
			{
				str1 = LyString.getSimpleString((String)value1);
				str2 = LyString.getSimpleString((String)value2);
			}
			else
			{
				str1 = (String)value1;
				str2 = (String)value2;
			}

			if (caseSensitive)
				result = str1.compareTo(str2);
			else
				result = str1.compareToIgnoreCase(str2);
		}

		return result;
	} // end method compare

	/**
	 * Checks if the second value is contained in the first one, for this field.
	 */
	public boolean contains(Object value1, Object value2, BitSet modifiers)
	{
		if ((value1 == null) && (value2 == null))
			return true;
		else if (value1 == null)
			return false;
		else if (value2 == null)
			return true;

		if (value1.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value1.getClass() + " expected " + getValueClass());
		if (value2.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value2.getClass() + " expected " + getValueClass());

		boolean multilingual = ((modifiers == null)
				|| modifiers.get(LyFilterExpression.MOD_MULTILINGUAL));
		String	str1, str2;

		if (multilingual
			|| !isMultilingual())
		{
			str1 = (String)value1;
			str2 = (String)value2;
		}
		else
		{
			str1 = getLangValue(null, (String)value1);
			str2 = getLangValue(null, (String)value2);
		}

		boolean		  caseSensitive = ((modifiers != null)
					&& modifiers.get(LyFilterExpression.MOD_CASE_SENSITIVE));
		LyEnvironment environment = getEnvironment();
		boolean		  ignoreAccent = ((environment != null)
					&& environment.testResource("LY_IGNORE_TEXT_ACCENT"));

		if (ignoreAccent)
		{
			str1 = LyString.getSimpleString(str1);
			str2 = LyString.getSimpleString(str2);
		}

		if (caseSensitive)
			return (str1.indexOf(str2) != -1);
		else
			return (str1.toLowerCase().indexOf(str2.toLowerCase()) != -1);
	} // end method contains

	/**
	 * Checks if the second value starts the first one, for this field.
	 */
	public boolean startsWith(Object value1, Object value2, BitSet modifiers)
	{
		if ((value1 == null) && (value2 == null))
			return true;
		else if (value1 == null)
			return false;
		else if (value2 == null)
			return true;

		if (value1.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value1.getClass() + " expected " + getValueClass());
		if (value2.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value2.getClass() + " expected " + getValueClass());

		boolean multilingual = ((modifiers == null)
				|| modifiers.get(LyFilterExpression.MOD_MULTILINGUAL));
		String	str1, str2;

		if (multilingual
			|| !isMultilingual())
		{
			str1 = (String)value1;
			str2 = (String)value2;
		}
		else
		{
			str1 = getLangValue(null, (String)value1);
			str2 = getLangValue(null, (String)value2);
		}

		boolean		  caseSensitive = ((modifiers != null)
					&& modifiers.get(LyFilterExpression.MOD_CASE_SENSITIVE));
		LyEnvironment environment = getEnvironment();
		boolean		  ignoreAccent = ((environment != null)
					&& environment.testResource("LY_IGNORE_TEXT_ACCENT"));

		if (ignoreAccent)
		{
			str1 = LyString.getSimpleString(str1);
			str2 = LyString.getSimpleString(str2);
		}

		if (caseSensitive)
			return str1.startsWith(str2);
		else
			return str1.toUpperCase().startsWith(str2.toUpperCase());
	} // end method startsWith

	/**
	 * Checks if the second value ends the first one, for this field.
	 */
	public boolean endsWith(Object value1, Object value2, BitSet modifiers)
	{
		if ((value1 == null) && (value2 == null))
			return true;
		else if (value1 == null)
			return false;
		else if (value2 == null)
			return true;

		if (value1.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value1.getClass() + " expected " + getValueClass());
		if (value2.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ value2.getClass() + " expected " + getValueClass());

		boolean multilingual = ((modifiers == null)
				|| modifiers.get(LyFilterExpression.MOD_MULTILINGUAL));
		String	str1, str2;

		if (multilingual
			|| !isMultilingual())
		{
			str1 = (String)value1;
			str2 = (String)value2;
		}
		else
		{
			str1 = getLangValue(null, (String)value1);
			str2 = getLangValue(null, (String)value2);
		}

		boolean		  caseSensitive = ((modifiers != null)
					&& modifiers.get(LyFilterExpression.MOD_CASE_SENSITIVE));
		LyEnvironment environment = getEnvironment();
		boolean		  ignoreAccent = ((environment != null)
					&& environment.testResource("LY_IGNORE_TEXT_ACCENT"));

		if (ignoreAccent)
		{
			str1 = LyString.getSimpleString(str1);
			str2 = LyString.getSimpleString(str2);
		}

		if (caseSensitive)
			return str1.endsWith(str2);
		else
			return str1.toUpperCase().endsWith(str2.toUpperCase());
	} // end method endsWith

	/**
	 * Constructs a value for this field info from an object for a given instance. Supported
	 * conversions are the following :
	 *
	 * <TABLE>
	 *   <TR>
	 *     <TD>String, String[], Number</TD>
	 *     <TD>String</TD>
	 *   </TR>
	 * </TABLE>
	 *
	 * @exception LyUnsupportedFieldBindingException If the requested conversion if none of the
	 *                                               supported conversions.
	 */
	public Object getValueFromObject(Object object, LyObject instance)
	{
        Object value = decodePeerValue(object);

        if (value == null)
            return null;

		if (value instanceof String[])
			return LyString.listToString((String[])value, '\n');
        else
            return value.toString();
	}

	/**
	 * Returns an object of the given class from a value of this fieldInfo. Supported conversions
	 * are the following :
	 *
	 * <TABLE>
	 *   <TR>
	 *     <TD>String</TD>
	 *     <TD>String, String[], Number</TD>
	 *   </TR>
	 * </TABLE>
	 *
	 * @exception LyUnsupportedFieldBindingException If the requested conversion if none of the
	 *                                               supported conversions.
	 */
	public Object getPeerValue(Object value, Class<?> objectClass, LyObject instance)
	{
        value = encodePeerValue(value, objectClass);

        if (value == null)
            return null;

		if ((value.getClass() == objectClass)
                || (objectClass == Object.class))
			return value;

		if (objectClass == String[].class)
			return LyString.stringToList((String)value, '\n');
		try
		{
			if (objectClass == Integer.class)
				return Integer.valueOf((String)value);
			else if (objectClass == Short.class)
				return Short.valueOf((String)value);
			else if (objectClass == Long.class)
				return Long.valueOf((String)value);
			else if (objectClass == Double.class)
				return Double.valueOf((String)value);
			else if (objectClass == Float.class)
				return Float.valueOf((String)value);
			else if (objectClass == Byte.class)
				return Byte.valueOf((String)value);
		}
		catch (NumberFormatException e)
		{
			throw new LyUnsupportedFieldBindingException(
				"BF AppEngine : exception caught on conversion "
				+ e.getMessage() + " on text field : " + this);
		}

        return super.getPeerValue(value, objectClass, instance);
	} // end method getPeerValue

	/**
	 * Checks the given value : - if the text field has a regex format (format starting with
	 * "regex:"), the given value must be conform to this format, - if a maximum number of
	 * characters is given on this text field, the value must not contain more characters than
	 * specified.
	 */
	public String checkValue(LyEnvironment environment, Object fieldValue)
	{
		if (fieldValue == null)
			return super.checkValue(environment, fieldValue);

		if (fieldValue.getClass() != getValueClass())
			throw new IllegalArgumentException("BF AppEngine : invalid value class "
				+ fieldValue.getClass() + " expected " + getValueClass());

		if (environment == null)
			environment = getEnvironment();

		String stringValue;

		if (isMultilingual())
			stringValue = getLangValue(null, (String)fieldValue);
		else
			stringValue = (String)fieldValue;

		String userStringValue = stringValue;
		String format = getFormat();

		if (hasMark(MARK_SECRET)
				&& (userStringValue != null))
		{
			int			 i, n = userStringValue.length();
			StringBuffer buffer = new StringBuffer(n);

			for (i = 0; i < n; i++)
				buffer.append('*');
			userStringValue = buffer.toString();
		}

		if (format != null)
		{
			// try to translate format
			String newFormat = environment.getString(format);

			if (newFormat != null)
				format = newFormat;
		}

		int nc = getCharTypesCount();

		if (nc > 0)
		{
            String  charType, c;
            boolean ok = true;
            int     i, j, n = stringValue.length();

            // Each character of the value must match at least one charType
            for (i = 0; (i < n) && ok; i++)
            {
                c = "" + stringValue.charAt(i);

                for (j = 0, ok = false; (j < nc) && !ok; j++)
                {
                    charType = getCharTypeValue(j);
                    ok |= LyString.matchFormat(environment, charType, c);
                }
            }

            if (!ok)
            {
                String   charTypes = null;

                for (j = 0; (j < nc) && !ok; j++)
                {
                    charType = getCharTypeValue(j);
                    if (charTypes == null)
                        charTypes = charType;
                    else
                        charTypes += ", " + charType;
                }

                return environment.getMessage("LY_ER_CHARTYPE_FORMAT", getName(), userStringValue, charTypes);
            }
		}

		String regex = getRegex();

		if (regex != null)
		{
			regex = environment.getString(regex);
			if (!LyString.matchFormat(environment, regex, stringValue))
				return environment.getMessage("LY_ER_REGEX_FORMAT", getName(), userStringValue, regex);
		}

		if ((format != null)
				&& !LyString.matchStringFormat(environment, format, stringValue))
			return environment.getMessage("LY_ER_STRING_FORMAT", getName(), userStringValue, format);

		// Check if the length of the String is less than the maximum number of chars
		int minChars = getMinChars();
		int maxChars = getMaxChars();

		if ((minChars >= 0) && (stringValue.length() < minChars))
		{
			String[] args = new String[4];

			args[0] = getName();
			args[1] = userStringValue;
			args[2] = String.valueOf(minChars);
			args[3] = String.valueOf(stringValue.length());
			return environment.getMessage("LY_ER_TEXT_MIN_CHARS", args);
		}
		if ((maxChars >= 0) && (stringValue.length() > maxChars))
		{
			if (getType() == Type.RICH)
			{
				String[] args = new String[4];

				args[0] = getName();
				args[1] = String.valueOf(maxChars);
				args[2] = String.valueOf(stringValue.length());
				return environment.getMessage("LY_ER_RICH_TEXT_MAX_CHARS", args);
			}
			else
			{
				String[] args = new String[4];

				args[0] = getName();
				args[1] = userStringValue;
				args[2] = String.valueOf(maxChars);
				args[3] = String.valueOf(stringValue.length());
				return environment.getMessage("LY_ER_TEXT_MAX_CHARS", args);
			}
		}

		return super.checkValue(environment, fieldValue);
	} // end method checkValue

	/**
	 * Checks whether this text field is multiple.
	 *
	 * @return True if the text field is multiple, false otherwise.
	 */
	public boolean isMultiple()
	{
		return _multiple;
	}

	/**
	 * Gets the prefered number of rows shown to the user at the same time. Not used for non
	 * multiple text fields.
	 *
	 * @return Number of rows, or -1 if no preference is given.
	 */
	public int getNRows()
	{
		return _nRows;
	}

	/**
	 * Gets the prefered number of columns shown to the user at the same time. For simple text
	 * fields, if no preference on the number of columns is given, this method will return the
	 * maximum number of chars allowed on this field.
	 *
	 * @return Number of columns, or -1 if no preference is given.
	 */
	public int getNCols()
	{
		if (!_multiple && (_nCols < 0))
		{
			if (_maxChars > 0)
				_nCols = _maxChars;
			else if (_format != null)
			{
				// Try to guess what is the best size for our format
				int  i, n = _format.length();
				int  p1, p2;
				char c;

				for (i = 0, _nCols = 0; i < n; i++)
				{
					c = _format.charAt(i);
					if (c == '[')
					{
						p1 = _format.indexOf('-', i + 1);
						p2 = _format.indexOf(']', p1 + 1);
						if ((p1 < 0) || (p2 < 0))
						{
							_nCols += n - i;
							break;
						}
						_nCols += Math.max(p1 - i - 1, p2 - p1 - 1);
						i = p2;
					}
					else
						_nCols++;
				}
			}
		}

		return _nCols;
	} // end method getNCols

	/**
	 * Gets the minimum number of characters allowed on values of this text field.
	 *
	 * @return Minimum number of characters, or -1 if non minimum is specified.
	 */
	public int getMinChars()
	{
		return _minChars;
	}

	/**
	 * Gets the maximum number of characters allowed on values of this text field.
	 *
	 * @return Maximum number of characters, or -1 if non maximum is specified.
	 */
	public int getMaxChars()
	{
		return _maxChars;
	}

	/**
	 * Gets the maximum number of characters allowed on values of this text field.
	 *
	 * @return     Maximum number of characters, or -1 if non maximum is specified.
	 * @deprecated see getMaxChars
	 */
	public int getNChars()
	{
		return getMaxChars();
	}

	/**
	 * Gets the regular expression for matching format of this text field.
	 *
	 * @return     the regular expression, or null if non regular expression is specified.
	 * @deprecated see getRegExp
	 */
	public String getRegex()
	{
		if (_regExp == null)
			return null;

		String value = _regExp.getValue();

		return value;
	}

	/**
	 * Gets the regular expression for matching format of this text field.
	 *
	 * @return the regular expression, or null if non regular expression is specified.
	 */
	public LyRegExp getRegExp()
	{
		return _regExp;
	}

	/**
	 * Returns the number of characters type of this text fieldinfo.
	 */
	public int getCharTypesCount()
	{
		return (_charTypes == null ? 0 : _charTypes.size());
	}

	/**
	 * Gets the characters type enabled by this text field at the specified index.
	 *
	 * @param  idx Index of the characters type.
	 * @return Characters enabled by this text field.
	 */
	public LyInfoCharType getCharType(int idx)
	{
		return (_charTypes == null ? null : _charTypes.elementAt(idx));
	}

	/**
	 * Gets the characters type enabled by this text field.
	 *
	 * @return the text field.
	 */
	public Vector<LyInfoCharType> getCharTypes()
	{
		return _charTypes;
	}

	/**
	 * Gets the characters type enabled by this text field at the specified index.
	 *
	 * @param  idx Index of the characters type.
	 * @return Characters enabled by this text field.
	 */
	public String getCharTypeValue(int idx)
	{
		LyInfoCharType charType = getCharType(idx);

		String		   value = charType.getValue();

		return getEnvironment().getString(value);
	}

	/**
	 * Gets the name of the encoding class for this text field.
	 *
	 * @return Name of the java class used to encode values for this text field if this text is
	 *         secret.
	 */
	public String getEncoding()
	{
		return _encoding;
	}

	/**
     * Sets the type of this text field.
     *
     * @return Type of this text.
     */
    public Type getType()
    {
        return _type;
    }

	/**
	 * Checks whether this text field is multilingual.
	 *
	 * @return True if the text field is multilingual, false otherwise.
	 */
	public boolean isMultilingual()
	{
		return _multilingual;
	}

	/**
	 * Sets this text field to multiple.
	 *
	 * @param multiple True if multiple, false otherwise.
	 */
	void setMultiple(boolean multiple)
	{
		_multiple = multiple;
	}

	/**
	 * Sets the preferred number of rows shown to the user at the same time.
	 *
	 * @param nRows Number of rows, or -1 if no preference.
	 */
	void setNRows(int nRows)
	{
		_nRows = nRows;
	}

	/**
	 * Sets the preferred number of columns shown to the user at the same time.
	 *
	 * @param nCols Number of columns, or -1 if no preference.
	 */
	void setNCols(int nCols)
	{
		_nCols = nCols;
	}

	/**
	 * Sets the minimum number of characters allowed for values on this text field.
	 *
	 * @param minChars Minimum number of characters, or -1 if non minimum is specified.
	 */
	void setMinChars(int minChars)
	{
		_minChars = minChars;
	}

	/**
	 * Sets the maximum number of characters allowed for values on this text field.
	 *
	 * @param maxChars Maximum number of characters, or -1 if non maximum is specified.
	 */
	void setMaxChars(int maxChars)
	{
		_maxChars = maxChars;
	}

	/**
	 * Sets default value for this text field.
	 *
	 * @param defaultString Default value.
	 */
	void setDefault(String defaultString)
	{
		if (defaultString != null)
			defaultString = defaultString.intern();
		setDefaultValue(defaultString);
	}

	/**
	 * Sets the display format for this field. The meaning of this format depends on each type of
	 * field.
	 *
	 * @param format The format or null if no format is specified
	 */
	void setFormat(String format)
	{
		super.setFormat(format);

		// Can't have a format and a regexp in the same time
		if (format != null)
		{
			setRegExp(null);
			_charTypes = null;
			setMaxChars(-1);
		}
	}

	/**
	 * Sets a regular expression for this text field.
	 *
	 * @param regExp regular expression or null if non expression is specified.
	 */
	void setRegExp(LyRegExp regExp)
	{
		_regExp = regExp;

		// Can't have a format and a regexp in the same time
		if (regExp != null)
			setFormat(null);
	}

	/**
	 * Sets the characters type enabled by this text field.
	 *
	 * @param charType Characters enabled by this text field.
	 */
	void addCharType(LyInfoCharType charType)
	{
        if (_charTypes == null)
            _charTypes = new Vector<LyInfoCharType>(1);
		_charTypes.addElement(charType);
	}

	/**
	 * Sets the name of the encoding class for this text field.
	 *
	 * @param regExp Name of the java class used to encode values for this text field if this text
	 *               is secret.
	 */
	void setEncoding(String encoding)
	{
		_encoding = encoding;
	}

    /**
     * Sets the characters type enabled by this text field.
     */
    protected void setCharTypes(Vector<LyInfoCharType> charTypes)
    {
        _charTypes = (charTypes == null ? null : new Vector<LyInfoCharType>(charTypes));
    }

	/**
     * Sets the type of this text field.
     *
     * @param type Type of this text.
     */
    void setType(Type type)
    {
        _type = type;
    }

	/**
	 * Sets this text field to multilingual.
	 *
	 * @param multilingual True if multilingual, false otherwise.
	 */
	void setMultilingual(boolean multilingual)
	{
		_multilingual = multilingual;
	}

    /**
     * Method for computing a value using function associated to this field for a given objects
     * list. This method is called when this field is presented in a table that has a footer.
     *
     * @param  objectList List of objects for which a footer value is computed.
     * @param  importField True if calling comes on computing of a field import, false otherwise.
     * @return Value computed with field function for the given objects. Returns null if this field
     *         has no associated function.
     */
    public Object compute(LyObjectList objectList, boolean importField)
    {
        if (objectList == null)
            return null;

        if (!importField)
            return null;

        String          value;
        StringBuffer    buffer = null;
        int             i, n = ((objectList == null) ? 0 : objectList.getSize());
        LyObject        object;

        for (i = 0; i < n; i++)
        {
            object = objectList.getObject(i);
            value = (String)object.getValue(this);
            if (value != null)
            {
                if (buffer == null)
                    buffer = new StringBuffer();
                else
                    buffer.append('\n');
                buffer.append(value);
            }
        }

        return (buffer == null ? null : buffer.toString());
    }
} // end class LyTextFieldInfo
