/*
 * LyTreemapConfiguration.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import java.io.*;

import leon.info.infointerface.*;
import leon.misc.*;


/**
 * This class represents a treemap configuration. It contains all the parameters that may be defined
 * by the user.
 *
 * @see LyObjectInfo
 * @see LyFilter
 */
public final class LyTreemapConfiguration extends LyObjectInfo
    implements LyExtraInfoInterface
{
	public static final String __VERSION =
		"$Id: LyTreemapConfiguration.java 40742 2012-09-11 11:29:29Z jlsaouli $";

	/** Algorithms for aggregation. */
    public enum Aggregation
    {
	   MINIMUM,
       MAXIMUM,
       AVERAGE,
       BALANCED
    }

	/** Object info concerned by the treemap configuration. */
	LyObjectInfo			   _objectInfo;

	/** Filter used by the treemap configuration. */
	LyFilter				   _filter;

	/** List of fields used to compute the hierarchy of the treemap. */
	LyInfoList<LyFieldInfo>    _hierarchyFields;

	/** Number of levels displayed in the treemap. */
	int						   _hierarchyLevels;

	/** The field info used to compute the sizes of the treemap elements. */
	LyFieldInfo				   _sizeField;

	/** The field info used to compute the colors of the treemap elements. */
	LyFieldInfo				   _colorField;

	/** Aggregation type of the treemap. */
    Aggregation				   _aggregationType;

	/** The field info used to compute aggregation in the treemap. */
	LyFieldInfo				   _aggregationField;

	/** Indicates if the filter is active or not. */
	boolean					   _filterActive;

	/** Indicates if all the objects must be shown (even if not match filter). */
	boolean					   _filterShowAll;

	/** Minimum color of the treemap. */
	String					   _minColor;

	/** Maximum color of the treemap. */
	String					   _maxColor;

	/** Threshold colors of the treemap. */
	String[]				   _thresholdColors;

	/** Sizes of the borders. */
	String					   _borderSizes;

	/** Styles of the labels. */
	String					   _labelStyles;

	/** Id of the user that owns the treemap configuration. */
	String					   _user;

	/**
	 * Constructor for a treemap configuration.
	 *
	 * @param id              the identifier of the treemap configuration.
	 * @param objectInfo      The concerned object info.
	 * @param name            the name of the treemap configuration.
	 * @param hierarchyLevels hierarchy levels of the treemap configuration.
	 * @param aggregationType aggregation type of the treemap configuration.
	 * @param filterActive    indicates if filter is active in the treemap configuration.
	 * @param filterShowAll   indicates if filter shows all the objects in the treemap
	 *                        configuration.
	 * @param minColor        min color of the treemap configuration.
	 * @param maxColor        max color of the treemap configuration.
	 * @param thresholdColors list of threshold colors of the treemap configuration.
	 * @param borderSizes     sizes of the borders of the treemap configuration.
	 * @param labelStyles     styles of the labels of the treemap configuration.
	 * @param user            the id of the user that owns the treemap configuration.
	 */
	public LyTreemapConfiguration(String id, LyObjectInfo objectInfo, String name,
		int hierarchyLevels, Aggregation aggregationType, boolean filterActive,
        boolean filterShowAll, String minColor, String maxColor,
        String[] thresholdColors, String borderSizes, String labelStyles, String user)
	{
		super(id);

		if (objectInfo != null)
			setObjectInfo(objectInfo);

		_user = user;
		_filter = null;
		_hierarchyLevels = hierarchyLevels;
		_sizeField = null;
		_colorField = null;
		_aggregationType = aggregationType;
		_aggregationField = null;
		_filterActive = filterActive;
		_filterShowAll = filterShowAll;
		_minColor = minColor;
		_maxColor = maxColor;
		_thresholdColors = thresholdColors;
		_borderSizes = borderSizes;
		_labelStyles = labelStyles;

		setName(name);
	}

	/**
	 * Clone a treemap configuration.
	 */
	public LyTreemapConfiguration cloneTreemapConfiguration()
	{
		LyTreemapConfiguration config = new LyTreemapConfiguration(_id + LyString.getRandomId(),
				_objectInfo, _name, _hierarchyLevels, _aggregationType, _filterActive,
				_filterShowAll, _minColor, _maxColor, _thresholdColors, _borderSizes, _labelStyles,
				_user);

		if (_hierarchyFields != null)
			config.setHierarchyFieldInfos(_hierarchyFields); // list cloned by set method

		if (_filter != null)
			config.setFilter(_filter.cloneFilter());

		return config;
	}

    /**
     * Method for saving preference file.
     */
    public boolean saveFile(LyXmlWriterInterface writer)
    {
        String      fileName;
        
        if (writer == null)
        	fileName = getFile();
        else
        	fileName = writer.getFile(this);
        
        if (fileName == null)
        {
            String   dir, userId = getUser();
            
            if (userId == null)
            {
                userId = "";
                dir = "LY_TREEMAP_DIR";
            }
            else
            {
                userId = userId.substring(userId.indexOf(leon.data.LyObjectId.CLASS_SEP) + 1) + "_";
                dir = "LY_USER_TREEMAP_DIR";
            }
            
            // retrieve the name of the data directory
            LyEnvironment environment = getEnvironment();
            String        dirPath = environment.getFile(dir);
            File          fileDirectory = null;

            if (dirPath != null)
                fileDirectory = new File(dirPath);

            // Try to create the directory if needed
            if ((fileDirectory != null)
                 && !fileDirectory.exists())
            {
                try
                {
                    environment.logInfo("Creating directory :" + dirPath);
                    fileDirectory.mkdirs();
                }
                catch (SecurityException e)
                {
                    environment.logError("Unable to create directory :" + dirPath);
                    return false;
                }
            }

            File    file;
            int     i = 0;
            
            // find a file name which has not been affected
            do
            {
                fileName = dirPath + "/treemap_configuration_" + userId + i + ".xml";
                file = new File(fileName);
                i++;
            }
            while (file.exists());
            
            setFile(fileName);
        }
        
        // Create xml file
        try
        {
            FileOutputStream 	output = new FileOutputStream(fileName);
            LyXmlPrintWriter    pw = new LyXmlPrintWriter(output);

            if (writer == null)
            	writeXml(pw, true, true);
            else
            	writer.writeXml(pw, this);
            
            pw.close();
            return true;
        }
        catch (Exception e)
        {
            LyEnvironment   environment = getEnvironment();
            
            environment.writeException(e);
            return false;
        }
    }
    
    /**
     * Method for freeing extra info (and sub-infos).
     */
    public void free()
    {
        if (_filter != null)
            _filter.free();
        super.free();
    }
    
    /**
     * Method for deleting preference file.
     */
    public boolean deleteFile()
    {
        String  filePath = getFile();

        if (filePath == null)
            return false;
        
        try
        {
            // Delete the file
            File    file = new File(filePath);
            
            return file.delete();
        }
        catch (SecurityException e)
        {
            LyEnvironment   environment = getEnvironment();
            
            environment.writeException(e);
            return false;
        }
    }
    
    /**
     * Method to rewrite XML corresponding to this component.
     *
     * @param pw The printWriter on which the XML is written.
     * @param header Boolean indicating if header must be written.
     * @param header Boolean indicating if body must be written.
     */
    public void writeXml(PrintWriter pw, boolean header, boolean body)
    {
        if (header)
        {
            pw.println("<?xml version='1.0' encoding='" + LyString.DEFAULT_CHARSET + "'?>");
            pw.println("<!DOCTYPE TREEMAP_CONFIGURATION SYSTEM 'leon/conf/dtd/leon.dtd'>");
        }
        
        if (body)
        	writeXml(pw);
    }
    
	/**
	 * Writes XML corresponding to this filter element.
	 *
	 * @param pw PrintWriter on which the XML is written.
	 */
	public void writeXml(PrintWriter pw)
	{
		// Write tag with its attributes
		pw.print("<TREEMAP_CONFIGURATION");
		pw.print(" id='" + LyString.getXmlString(getId()) + "'\n");
		pw.print(" idref='" + LyString.getXmlString(_objectInfo.getId()) + "'\n");

		if (_hierarchyFields != null)
		{
			int			 i, n = _hierarchyFields.getSize();
			LyFieldInfo  field;
			StringBuffer hierarchyString = new StringBuffer();

			for (i = 0; i < n; i++)
			{
				if (i != 0)
					hierarchyString.append(" ");

				field = _hierarchyFields.getInfo(i);
				hierarchyString.append(LyString.getXmlString(field.getId()));
			}

			pw.print(" hierarchyFields='" + hierarchyString + "'\n");
		}

		if ((_thresholdColors != null) && (_thresholdColors.length > 0))
		{
			int			 i, n = _thresholdColors.length;
			StringBuffer colorsString = new StringBuffer();

			for (i = 0; i < n; i++)
			{
				if (_thresholdColors[i] == null)
					continue;

				if (i != 0)
					colorsString.append(" ");

				colorsString.append(_thresholdColors[i]);
			}

			pw.print(" thresholdColors='" + colorsString + "'\n");
		}

		if (_hierarchyLevels >= 0)
			pw.print(" hierarchyLevels='" + _hierarchyLevels + "'\n");
		if (_sizeField != null)
			pw.print(" sizeField='" + LyString.getXmlString(_sizeField.getId()) + "'\n");
		if (_colorField != null)
			pw.print(" colorField='" + LyString.getXmlString(_colorField.getId()) + "'\n");
		if (_aggregationField != null)
			pw.print(" aggregationField='" + LyString.getXmlString(_aggregationField.getId())
				+ "'\n");

		String aggregationType = "BALANCED";

		if (_aggregationType == Aggregation.MINIMUM)
			aggregationType = "MINIMUM";
		else if (_aggregationType == Aggregation.MAXIMUM)
			aggregationType = "MAXIMUM";
		else if (_aggregationType == Aggregation.AVERAGE)
			aggregationType = "AVERAGE";

		if (aggregationType != null)
			pw.print(" aggregationType='" + aggregationType + "'\n");
		pw.print(" filterActive='" + (_filterActive ? "true" : "false") + "'\n");
		pw.print(" filterShowAll='" + (_filterShowAll ? "true" : "false") + "'\n");
		if (_minColor != null)
			pw.print(" minColor='" + LyString.getXmlString(_minColor) + "'\n");
		if (_maxColor != null)
			pw.print(" maxColor='" + LyString.getXmlString(_maxColor) + "'\n");
		if (_borderSizes != null)
			pw.print(" borderSizes='" + LyString.getXmlString(_borderSizes) + "'\n");
		if (_labelStyles != null)
			pw.print(" labelStyles='" + LyString.getXmlString(_labelStyles) + "'\n");
		if (_name != null)
			pw.print(" name='" + LyString.getXmlString(_name) + "'");
		if (_user != null)
			pw.print(" user='" + LyString.getXmlString(_user) + "'");
		pw.println(">");

		if (_filter != null)
			_filter.writeXml(pw);

		LyFieldInfo			    info;
		int					    i, n;

		// Write external dependencies
		LyInfoList<LyFieldInfo> externs = new LyInfoList<LyFieldInfo>(false);

		writeExternXml(pw, _objectInfo.getId());

		if (_filter != null)
			externs.addList(_filter.getFieldInfos());

		if (_hierarchyFields != null)
			externs.addList(_hierarchyFields);

		if (_sizeField != null)
			externs.addInfo(_sizeField);

		if (_colorField != null)
			externs.addInfo(_colorField);

		if (_aggregationField != null)
			externs.addInfo(_aggregationField);

		for (i = 0, n = externs.getSize(); i < n; i++)
		{
			info = externs.getInfo(i);
			writeExternXml(pw, info.getId());
		}

		pw.println("</TREEMAP_CONFIGURATION>");
	} // end method writeXml

	/**
	 * Writes XML for an external idref.
	 *
	 * @param pw PrintWriter on which the XML is written.
	 * @param id id that must be written.
	 */
	protected void writeExternXml(PrintWriter pw, String id)
	{
		pw.println("<EXTERN id='" + LyString.getXmlString(id) + "'/>");
	}

	/**
	 * Get object info referred by this treemap configuration.
	 *
	 * @return The concerned object info (action or classInfo).
	 */
	public LyObjectInfo getObjectInfo()
	{
		return _objectInfo;
	}

	/**
	 * Get filter used by this treemap configuration.
	 *
	 * @return The concerned filter.
	 */
	public LyFilter getFilter()
	{
		return _filter;
	}

	/**
	 * Get the field info used to compute the size of this treemap configuration.
	 *
	 * @return Field info used to compute the size in this treemap configuration.
	 */
	public LyFieldInfo getSizeFieldInfo()
	{
		return _sizeField;
	}

	/**
	 * Get the field info used to compute the color of this treemap configuration.
	 *
	 * @return Field info used to compute the color in this treemap configuration.
	 */
	public LyFieldInfo getColorFieldInfo()
	{
		return _colorField;
	}

	/**
	 * Get the aggregation type.
	 *
	 * @param aggregation type.
	 */
	public Aggregation getAggregationType()
	{
		return _aggregationType;
	}

	/**
	 * Get the field info used to compute aggregations in this treemap configuration.
	 *
	 * @return Field info used to compute aggregations in this treemap configuration.
	 */
	public LyFieldInfo getAggregationFieldInfo()
	{
		return _aggregationField;
	}

	/**
	 * Gets list of field infos of this treemap configuration.
	 *
	 * @return List of hierarchy field infos, may be null.
	 */
	public LyInfoList<LyFieldInfo> getHierarchyFieldInfos()
	{
		if (_hierarchyFields != null)
			_hierarchyFields.lock();
		return _hierarchyFields;
	}

	/**
	 * Return the hierarchy levels for this treemap configuration.
	 *
	 * @return the hierarchy levels for this treemap configuration
	 */
	public int getHierarchyLevels()
	{
		return _hierarchyLevels;
	}

	/**
	 * Indicates whether filter is active or not in this treemap configuration.
	 *
	 * @return true if filter is active, false otherwise
	 */
	public boolean isFilterActive()
	{
		return _filterActive;
	}

	/**
	 * Indicates whether filter shows all object or not in this treemap configuration.
	 *
	 * @return true if filter shows all objects, false otherwise
	 */
	public boolean isFilterShowAll()
	{
		return _filterShowAll;
	}

	/**
	 * Get minimum color of this treemap configuration.
	 *
	 * @return The minimum color of this treemap configuration.
	 */
	public String getMinColor()
	{
		return _minColor;
	}

	/**
	 * Get maximum color of this treemap configuration.
	 *
	 * @return The maximum color of this treemap configuration.
	 */
	public String getMaxColor()
	{
		return _maxColor;
	}

	/**
	 * Get the list of threshold colors of this treemap configuration.
	 *
	 * @return The list of threshold colors of this treemap configuration.
	 */
	public String[] getThresholdColors()
	{
		return _thresholdColors;
	}

	/**
	 * Get border sizes of this treemap configuration.
	 *
	 * @return The border sizes of this treemap configuration.
	 */
	public String getBorderSizes()
	{
		return _borderSizes;
	}

	/**
	 * Get label styles of this treemap configuration.
	 *
	 * @return The label styles of this treemap configuration.
	 */
	public String getLabelStyles()
	{
		return _labelStyles;
	}

	/**
	 * Get owner of this treemap configuration.
	 *
	 * @return The owner of this treemap configuration.
	 */
	public String getUser()
	{
		return _user;
	}

	/**
	 * Sets the object info referred by this treemap configuration.
	 *
	 * @param action Object info concerned by this treemap configuration.
	 */
	void setObjectInfo(LyObjectInfo objectInfo)
	{
		if (!(objectInfo instanceof LyAction)
				&& !(objectInfo instanceof LyClassInfo))
			throw new IllegalStateException("BF AppEngine : invalid objectInfo for "
				+ "treemap configuration : " + objectInfo
				+ ", objectInfo should be an action or a class info");

		_objectInfo = objectInfo;
	}

	/**
	 * Set the field info used to compute the sizes.
	 *
	 * @param sizeField Field info used to compute the sizes.
	 */
	public void setSizeFieldInfo(LyFieldInfo sizeField)
	{
		_sizeField = sizeField;
	}

	/**
	 * Set the hierarchy levels.
	 *
	 * @param hierarchyLevels number of hierarchy levels.
	 */
	public void setHierarchyLevels(int hierarchyLevels)
	{
		_hierarchyLevels = hierarchyLevels;
	}

	/**
	 * Set the activated state for filter.
	 *
	 * @param active indicates whether filter is active or not.
	 */
	public void setFilterActive(boolean active)
	{
		_filterActive = active;
	}

	/**
	 * Set the show all state for filter.
	 *
	 * @param active indicates whether filter shows all objects or not.
	 */
	public void setFilterShowAll(boolean showAll)
	{
		_filterShowAll = showAll;
	}

	/**
	 * Set the minimum color.
	 *
	 * @param minColor the new minimum color.
	 */
	public void setMinColor(String minColor)
	{
		_minColor = minColor;
	}

	/**
	 * Set the maximum color.
	 *
	 * @param minColor the new maximum color.
	 */
	public void setMaxColor(String maxColor)
	{
		_maxColor = maxColor;
	}

	/**
	 * Set the field info used to compute the colors.
	 *
	 * @param colorField Field info used to compute the colors.
	 */
	public void setColorFieldInfo(LyFieldInfo colorField)
	{
		_colorField = colorField;
	}

	/**
	 * Set the aggregation type.
	 *
	 * @param aggregationType aggregation type.
	 */
	public void setAggregationType(Aggregation aggregationType)
	{
		_aggregationType = aggregationType;
	}

	/**
	 * Set the field info used to compute aggregations.
	 *
	 * @param aggregationField Field info used to compute aggregations.
	 */
	public void setAggregationFieldInfo(LyFieldInfo aggregationField)
	{
		_aggregationField = aggregationField;
	}

	/**
	 * Sets the filter used by this treemap configuration.
	 *
	 * @param filter the new filter.
	 */
	public void setFilter(LyFilter filter)
	{
		if (filter == null)
		{
			_filter = null;
			return;
		}

		_filter = filter.cloneFilter();
		_filter.setParent(this);

		if (getApplication() != null)
			_filter.setApplication(getApplication());
	}

	/**
	 * Set the list of threshold colors.
	 *
	 * @param colors the list of threshold colors.
	 */
	public void setThresholdColors(String[] colors)
	{
		_thresholdColors = colors;
	}

	/**
	 * Set the border sizes.
	 *
	 * @param borderSizes the new border sizes.
	 */
	public void setBorderSizes(String borderSizes)
	{
		_borderSizes = borderSizes;
	}

	/**
	 * Set the label styles.
	 *
	 * @param labelStyles the new label styles.
	 */
	public void setLabelStyles(String labelStyles)
	{
		_labelStyles = labelStyles;
	}

	/**
	 * Sets list of hierarchy field infos of this treemap configuration.
	 *
	 * @param fieldInfos List of hierarchy field infos, may be null.
	 */
	public void setHierarchyFieldInfos(LyInfoList<LyFieldInfo> fieldInfos)
	{
		if (fieldInfos == null)
			_hierarchyFields = null;
		else
			_hierarchyFields = new LyInfoList<LyFieldInfo>(fieldInfos);
	}
} // end class LyTreemapConfiguration
