/*
 * LyUnit.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

import leon.info.misc.*;


/**
 * This class defines a unit associated to a text, a number or a date field.
 */
public class LyUnit extends LyInfo
{
	public static final String __VERSION = "$Id: LyUnit.java 40970 2012-09-28 16:32:05Z jlsaouli $";

	/** Value of the unit. */
	String					   _value;

	/** Factor used to make conversions between units. */
	double					   _factor;

	/** Offset used to make conversions between units. */
	double					   _offset;

	/** Display format of this unit. */
	LyFormat				   _format;

	/** Name of the unit. */
	String					   _name;

	/** Alias of the unit. */
	String					   _alias;

	/** Indicates if this unit is used for printing. */
	boolean					   _print;

	/** Interface used to convert this unit to the reference unit. */
	String					   _javaName;

	/** Converter class of the unit. */
	Class<?>				   _converterClass;

	/* Constructor */

	/**
	 * Constructor of a unit.
	 *
	 * @param id       Id of the unit.
	 * @param factor   Factor used for conversion
	 * @param offset   Factor used for conversion
	 * @param javaName Name of the class used to convert a value to the reference unit.
	 */
	LyUnit(String id, double factor, double offset, String javaName)
	{
		this(id);
		construct(factor, offset, javaName);
	}

	/**
	 * Constructor of a unit.
	 *
	 * @param id    Id of the unit.
	 * @param value Value of the unit.
	 */
	LyUnit(String id, String value)
	{
		this(id, 1.0, 0.0, null);

		setValue(value);
	}

	/**
	 * Constructor of a unit.
	 *
	 * @param id Id of the unit.
	 */
	LyUnit(String id)
	{
		super(id, false);
	}

	/**
	 * Construction of a unit.
	 *
	 * @param factor   Factor used for conversion
	 * @param offset   Factor used for conversion
	 * @param javaName Name of the class used to convert a value to the reference unit.
	 */
	void construct(double factor, double offset, String javaName)
	{
		super.construct();

		setFactor(factor);
		setOffset(offset);
		setJavaName(javaName);
	}

	/**
	 * Gets the factor used when converting this unit to the reference one.
	 *
	 * @return Factor between this unit and the reference (1.0 is this unit is the reference).
	 */
	public double getFactor()
	{
		load();
		return _factor;
	}

	/**
	 * Gets the offset used when converting this unit to the reference one.
	 *
	 * @return Offset between this unit and the reference (0.0 is this unit is the reference).
	 */
	public double getOffset()
	{
		load();
		return _offset;
	}

	/**
	 * Gets the display format used for presenting this unit.
	 *
	 * @return Display format of this unit.
	 */
	public LyFormat getFormat()
	{
		return _format;
	}

	/**
	 * Accessor for name of this unit.
	 *
	 * @return Name of this instance.
	 */
	public final String getName()
	{
		if (_name == null)
			return getValue();

		return _name;
	}

	/**
	 * Accessor for alias (shortname) of this unit.
	 *
	 * @return Alias of this unit. If the alias is null, this method returns the name.
	 */
	public final String getAlias()
	{
		if (_alias == null)
			return getName();

		return _alias;
	}

    /**
     * Accessor for alias (shortname) of this instance.
     * This method doesn't compute a value if it is null.
     *
     * @return Alias of this instance.
     */
    public final String getInternalAlias()
    {
        return _alias;
    }
    
	/**
     * Sets the alias of the instance.
     *
     * @param alias Alias of the instance.
     */
    final void setInternalAlias(String alias)
    {
        _alias = ((alias == null) ? null : alias.intern());
    }
    
	/**
	 * Checks whether this unit is used for printing.
	 *
	 * @return True if this unit is used for printing, false otherwise.
	 */
	public final boolean isPrint()
	{
		return _print;
	}

	/**
	 * Gets value of the unit. Kept for upward compatibility.
	 * @deprecated Use getName()
	 */
	public final String getValue()
	{
		if (_value == null)
			return getId();

		return _value;
	}

	/**
	 * Gets java class name for this unit. This is the name of java class used to convert this unit
	 * to the reference when the user selects this unit.
	 *
	 * @return Name of the java class for this unit.
	 */
	public String getJavaName()
	{
		return _javaName;
	}

	/**
	 * Gets the class to convert.
	 *
	 * @return The class of the converter.
	 */
	public Class<?> getConverterClass()
	{
		load();
		if ((_javaName != null) && (_converterClass == null))
			_converterClass = _application.loadClass(_javaName);

		return _converterClass;
	}

	/**
	 * Checks if this unit has an edition format. Format used to edit value from a simple string.
	 */
	public boolean hasEditionFormat()
	{
		return ((_format != null) && _format.isEditable());
	}

	/**
	 * Sets the display format used for presenting this unit.
	 *
	 * @param format Display format of this unit.
	 */
	final void setFormat(LyFormat format)
	{
		_format = format;
	}

	/**
	 * Complete unit, method used by Gui parser to fulfill unit with gui attributes.
	 */
	final void completeUnit(String value, String name, String alias, boolean print)
	{
		setValue(value);
		setName(name);
		setAlias(alias);
		setPrint(print);
	}
    
    /**
     * Sets name of this unit.
     *
     * @param name Name of this instance.
     */
    protected void setName(String name)
    {
        _name = ((name == null) ? null : name.intern());
    }
    
    /**
     * Sets the factor used when converting this unit to the reference one.
     *
     * @param facthor Factor between this unit and the reference (1.0 is this unit is the reference).
     */
    protected void setFactor(double factor)
    {
        _factor = factor;
    }

    /**
     * Sets the offset used when converting this unit to the reference one.
     *
     * @param offset Offset between this unit and the reference (0.0 is this unit is the reference).
     */
    protected void setOffset(double offset)
    {
        _offset = offset;
    }
    
    /**
     * Sets alias (shortname) of this unit.
     *
     * @param alias Alias of this unit. If the alias is null, this method returns the name.
     */
    protected void setAlias(String alias)
    {
        _alias = ((alias == null) ? null : alias.intern());
    }

    /**
     * Sets whether this unit is used for printing.
     *
     * @param print True if this unit is used for printing, false otherwise.
     */
    protected void setPrint(boolean print)
    {
        _print = print;
    }

    /**
     * Sets value of the unit. Kept for upward compatilibility.
     */
    protected void setValue(String value)
    {
        _value = ((value == null) ? null : value.intern());
    }

    /**
     * Sets java class name for this unit. This is the name of java class used to convert this unit
     * to the reference when the user selects this unit.
     *
     * @param javaName Name of the java class for this unit.
     */
    protected void setJavaName(String javaName)
    {
        _javaName = javaName;
    }

} // end class LyUnit
