/*
 * LyUnits.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info;

/**
 * This class defines a list of units.
 */
public class LyUnits extends LyInfo
{
	public static final String   __VERSION =
		"$Id: LyUnits.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** List of units. */
	protected LyInfoList<LyUnit> _units;

	/** Name of the unit. */
	String						 _name;

	/** Interface used to convert values between units of this units list. */
	String						 _javaName;

	/** Converter class of the units list. */
	Class<?>					_converterClass;

	/** Convert flag of the units list. If set to false, there is no conversion made in GUI. */
	boolean						 _convert;

	/* Constructor */

	/**
	 * Constructor of units list.
	 *
	 * @param id       Id of the units list.
	 * @param javaName Name of the class used to convert values between units.
	 */
	LyUnits(String id, String javaName)
	{
		this(id);
		construct(javaName);
	}

	/**
	 * Constructor of units list.
	 *
	 * @param id Id of the units list.
	 */
	LyUnits(String id)
	{
		super(id, false);
	}

	/**
	 * Construct of a units list.
	 *
	 * @param javaName Name of the class used to convert values between units.
	 */
	void construct(String javaName)
	{
		super.construct();

		_javaName = javaName;
	}

	/**
	 * Return the number of units associated to this field.
	 *
	 * @return Number of units of this field.
	 */
	public int getUnitCount()
	{
		load();
		return ((_units == null) ? 0 : _units.getSize());
	}

	/**
	 * Fetches the unit at the specified index.
	 *
	 * @param  index Index of the searched unit.
	 * @return The requested unit if found, or null if it could not be found.
	 */
	public LyUnit getUnit(int index)
	{
		load();
		return ((_units == null) ? null : _units.getInfo(index));
	}

	/**
	 * Get field units.
	 */
	public LyInfoList<LyUnit> getUnits()
	{
		return _units;
	}

	/**
	 * Accessor for name of this units list.
	 *
	 * @return Name of this units list.
	 */
	public final String getName()
	{
		if (_name == null)
			return getId();

		return _name;
	}

	/**
	 * Gets java class name corresponding to this units list. Name of the class used to convert
	 * values between units of this list.
	 *
	 * @return Name of the java class corresponding to this units list.
	 */
	public String getJavaName()
	{
		return _javaName;
	}

	/**
	 * Gets the class to convert.
	 *
	 * @return The class of the converter.
	 */
	public Class<?> getConverterClass()
	{
		load();
		if ((_javaName != null) && (_converterClass == null))
			_converterClass = _application.loadClass(_javaName);

		return _converterClass;
	}

	/**
	 * Gets reference unit (unit with factor=1 and offset=0).
	 *
	 * @return The reference unit if found, the first one otherwise.
	 */
	public LyUnit getReferenceUnit()
	{
		int    i, n = getUnitCount();
		LyUnit unit;

		for (i = 0; i < n; i++)
		{
			unit = getUnit(i);
			if ((unit.getFactor() == 1) && (unit.getOffset() == 0))
				return unit;
		}

		return ((n == 0) ? null : getUnit(0));
	}

	/**
	 * Gets the convert flag for this units list.
	 *
	 * @return The convert flag of this units list.
	 */
	public boolean isConvert()
	{
		return _convert;
	}

	/**
	 * Adds an unit to this field info.
	 *
	 * @param unit Unit to add to this field info.
	 */
	final void addUnit(LyUnit unit)
	{
		if (_units == null)
			_units = new LyInfoList<LyUnit>();
		_units.addInfo(unit);
	}

	/**
	 * Sets the name of the units list.
	 *
	 * @param name Name of the units list.
	 */
	void setName(String name)
	{
		_name = ((name == null) ? null : name.intern());
	}

	/**
	 * Sets java class name corresponding to this units list.
	 *
	 * @param javaName Name of the java class corresponding to this units list.
	 */
	public void setJavaName(String javaName)
	{
		_javaName = javaName;
	}

	/**
	 * Set the convert flag of this units list.
	 *
	 * @param convert Convert flag of the units list.
	 */
	void setConvert(boolean convert)
	{
		_convert = convert;
	}

	/**
	 * Checks if the given unit is contained in this units list.
	 *
	 * @param  unitChecked fieldInfo.
	 * @return True if the given unit is found, false otherwise.
	 */
	boolean contains(LyUnit unit)
	{
		if (_units == null)
			return false;
		return _units.contains(unit);
	}

	/**
	 * Retrieve a unit in units list from its id.
	 *
	 * @param  unitId Id of the searched unit.
	 * @return Unit corresponding to the id if found, null otherwise.
	 */
	public LyUnit getUnit(String unitId)
	{
		if (_units == null)
			return null;
		return _units.getInfo(unitId);
	}
    
    /**
     * Set field units.
     */
    protected void setUnits(LyInfoList<LyUnit> units)
    {
        int     i, n = (units == null ? 0 : units.getSize());
        
        _units = null;
        for (i = 0; i < n; i++)
            addUnit(units.getInfo(i));
    }
} // end class LyUnits
