/*
 * LyClassBinding.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info.binding;

import java.util.*;

import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.location.*;
import leon.misc.*;


/**
 * Container for the binding information found in the data description for classInfo.
 */
public class LyClassBinding extends LyBinding
	implements Cloneable
{
	public static final String __VERSION =
		"$Id: LyClassBinding.java 35781 2012-06-08 12:54:14Z jlsaouli $";

	/** Binding location (identifier of the data provider, when several providers exists). */
	String					   _location;

	/**
	 * Filter used when loading objects. The filter is used when loading objects when a logical
	 * class match partially a physical one.
	 */
	LyFilter				   _filter;

	/** Class corresponding to this binding. */
	LyClass					   _peerClass;

	/**
	 * Definition of all fields binding : list of association between a fieldInfo and a peerField.
	 */
	Hashtable<LyFieldInfo, Object>	   _peerFields;

	/**
	 * Definition of all fields binding : list of association between a peerField and a fieldInfo.
	 */
	Hashtable<LyField, LyFieldInfo>	   _fieldInfos;


    /**
     * Constructor.
     * @param binding  content of the binding information
     */
    public LyClassBinding(String binding)
    {
        this(binding, null, null);
    }

	/**
	 * Constructor.
	 * @param binding  content of the binding information
	 * @param location identify which data provider should be used.
	 * @param filter   filter used when loading objects.
	 */
	public LyClassBinding(String binding, String location, LyFilter filter)
	{
		super(binding);

		setLocation(location);
		_filter = filter;
		_peerClass = null;
		_peerFields = null;
		_fieldInfos = null;
	}

	/**
	 * Final step of parsing : checks integrity of the classBinding and completes the initialization if
	 * necessary.
	 * @exception IllegalStateException If one of the checks fails
	 */
	public void check(LyClassInfo parent)
	{
		LyApplication	application = parent.getApplication();

		if (_location == null)
		{
			if (_binding != null)
				_location = application.getDefaultLocationId();
		}
		else if (_location.startsWith("LY_"))
		{
			LyEnvironment	environment = parent.getEnvironment();
			String			realLocationId = environment.getResource(_location);

			if ((realLocationId != null)
				&& !"".equals(realLocationId))
			{
				LyInfo location = application.getInfo(realLocationId);

				if (!(location instanceof LyLocation))
					throw new IllegalStateException("BF AppEngine : invalid location ref : " + realLocationId);

				_location = realLocationId;
			}
			else
				_location = application.getDefaultLocationId();
		}
	}

	/**
	 * Reads the binding location information. this optional info identifies which data provider is
	 * used (when several exists within the application) to fetch the data with the binding info.
	 * @return The binding location which identifies a data provider, or null if multiple data
	 *         provider is not implemented in the application.
	 */
	public String getLocation()
	{
		return _location;
	}

	/**
	 * Gets the filter of this binding.
	 * @return The binding filter or null if no filter is set on this binding.
	 */
	public LyFilter getFilter()
	{
		return _filter;
	}

	/**
	 * Returns the class corresponding to this binding.
	 */
	public LyClass getPeerClass()
	{
		return _peerClass;
	}

	/**
	 * Sets the filter of this binding.
	 * @param filter the filter.
	 */
	public void setFilter(LyFilter filter)
	{
		_filter = filter;
	}

	/**
	 * Sets the class corresponding to this binding.
	 */
	public void setPeerClass(LyClass peerClass)
	{
		_peerClass = peerClass;
	}

	/**
	 */
	public void addPeerField(LyFieldInfo fieldInfo, LyField peerField)
	{
        if (peerField == null)
        {
            removePeerField(fieldInfo);
            return;
        }

		if (_peerFields == null)
			_peerFields = new Hashtable<LyFieldInfo, Object>();
		if (_fieldInfos == null)
			_fieldInfos = new Hashtable<LyField, LyFieldInfo>();

	    _peerFields.put(fieldInfo, peerField);
        _fieldInfos.put(peerField, fieldInfo);
	}

	/**
     */
    public Object removePeerField(LyFieldInfo fieldInfo)
    {
        if (_peerFields == null)
            return null;

        Object  value = _peerFields.remove(fieldInfo);

        if (value instanceof LyField)
            _fieldInfos.remove(value);
        else if (value instanceof LyField[])
        {
            LyField[] fields = (LyField[])value;
            int       i, n = fields.length;

            for (i = 0; i < n; i++)
                _fieldInfos.remove(fields[i]);
        }

        return value;
    }

	/**
	 */
	public LyField getPeerField(LyFieldInfo fieldInfo)
	{
		if (_peerFields == null)
			return null;

		return (LyField)_peerFields.get(fieldInfo);
	}

	public LyFieldBinding getFieldBinding(LyField field)
	{
		if (field instanceof LyFieldInfo)
			return ((LyFieldInfo)field).getBinding();

		if (!_fieldInfos.containsKey(field))
			return null;

		LyFieldInfo fieldInfo = _fieldInfos.get(field);

		return fieldInfo.getBinding();
	}

	/**
	 */
	public void addPeerFields(LyFieldInfo fieldInfo, LyField[] peerFields)
	{
	    if (peerFields == null)
	    {
	        removePeerField(fieldInfo);
	        return;
	    }

		if (_peerFields == null)
			_peerFields = new Hashtable<LyFieldInfo, Object>();
		if (_fieldInfos == null)
			_fieldInfos = new Hashtable<LyField, LyFieldInfo>();

		_peerFields.put(fieldInfo, peerFields);

		for (int i = 0; i < peerFields.length; i++)
			_fieldInfos.put(peerFields[i], fieldInfo);
	}

	/**
	 */
	public LyField[] getPeerFields(LyFieldInfo fieldInfo)
	{
		if (_peerFields == null)
			return null;

		Object value = _peerFields.get(fieldInfo);

		if (value instanceof LyField)
		{
			LyField[] fields = new LyField[1];

			fields[0] = (LyField)value;
			return fields;
		}
		else
			return (LyField[])value;
	}

	public boolean equals(Object o)
	{
		if (o == null)
			return false;
		if (o.getClass() != LyClassBinding.class)
			return false;

		LyClassBinding classBinding = (LyClassBinding)o;
		String		   binding = classBinding.getBinding();

		if ((binding != null)
				&& (_binding != null)
				&& (!binding.equals(_binding)))
			return false;
		if (((binding != null) && (_binding == null))
				|| ((binding == null) && (_binding != null)))
			return false;

		String		   location = classBinding.getLocation();

		if ((location != null)
				&& (_location != null)
				&& (!location.equals(_location)))
			return false;
		if (((location != null) && (_location == null))
				|| ((location == null) && (_location != null)))
			return false;

		LyFilter	   filter = classBinding.getFilter();

		if ((filter != null)
				&& (_filter != null)
				&& (!filter.toString().equals(_filter.toString())))
			return false;
		if (((filter != null) && (_filter == null))
				|| ((filter == null) && (_filter != null)))
			return false;

		return true;
	} // end method equals

    /**
     * Internal method used to set location.
     */
    protected void setLocation(String location)
    {
        _location = ((location == null) ? null : location.intern());
    }
} // end class LyClassBinding
