/*
 * LyFieldBinding.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info.binding;

import leon.app.*;
import leon.info.*;
import leon.misc.*;


/**
 * Container for the binding information found in the data description for fieldInfo.
 */
public class LyFieldBinding extends LyBinding
	implements Cloneable
{
	public static final String __VERSION =
		"$Id: LyFieldBinding.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Default value for an id field. */
	String					   _default;

	/** Class binding to which the binding belongs to. */
	String					   _cbind;

	/** Indicates if this bind is the id or a part of the id for the physical layer. */
	boolean					   _isId;

	/** Name of the java class for this binding. */
	String					   _javaType;

	/** Name of the java class used for encoding/decoding values. */
	String					   _javaEncoder;
	
	/** Encoder class. */
	Class<?>                   _javaEncoderClass;

	/** Peer type (used by RDBMS which accepts BLOBS and/or CLOBS) otherwise value is 'DEFAULT'. */
	String					   _peerType;
	
	
	/**
     * Default constructor.
     */
    protected LyFieldBinding()
    {
        super(null);
    }

	/**
	 * Constructor.
	 * @param binding content of the binding information
	 * @param def
	 * @param cbind
	 * @param isId
	 */
	public LyFieldBinding(String binding, String def, String cbind, boolean isId)
	{
		this(binding, def, cbind, isId, null, null, null);
	}
	
	/**
	 * Constructor kept for backward compatibility.
	 * @deprecated replaced by LyFieldBinding(String binding, String def, String cbind, boolean isId,
		String javaType, String javaEncoder, String peerType)
	 */
	public LyFieldBinding(String binding, String def, String cbind, boolean isId,
		String javaType, String javaEncoder)
	{
		this(binding, def, cbind, isId, javaType, javaEncoder, "DEFAULT");
	}
	
	/**
	 * Constructor.
	 * @param binding     content of the binding information
	 * @param def
	 * @param cbind
	 * @param isId
	 * @param javaType
	 * @param javaEncoder
	 * @param peerType Peer type (used by RDBMS which accepts BLOBS and/or CLOBS) otherwise value is 'DEFAULT'
	 */
	public LyFieldBinding(String binding, String def, String cbind, boolean isId,
		String javaType, String javaEncoder, String peerType)
	{
		super(binding);

		_default = def;
		_cbind = cbind;
		_isId = isId;
		_javaType = javaType;
		_javaEncoder = javaEncoder;
		_peerType = peerType;
	}

	/**
	 * Returns the default value of this binding.
	 */
	public String getDefault()
	{
		return _default;
	}

	/**
	 * Returns the cbind of this binding.
	 */
	public String getCBind()
	{
		return _cbind;
	}

	/**
	 * Indicates if this binding is id for the physical layer.
	 */
	public boolean isId()
	{
		return _isId;
	}

	/**
	 * Returns the name of the java class for this binding.
	 */
	public String getJavaType()
	{
		return _javaType;
	}

    /**
     * Returns the java class for this binding corresponding to java type.
     */
    public Class<?> getJavaTypeClass(LyApplication application)
    {
        if (_javaType == null)
            return null;
        
        ClassLoader appClassLoader = application.getClassLoader();
        Class<?>    typeClass;
        
        // If we have a classLoader on the application, use it.
        try
        {
            if (appClassLoader != null)
                typeClass = Class.forName(_javaType, false, appClassLoader);
            else
                typeClass = Class.forName(_javaType);
        }
        catch (ClassNotFoundException e)
        {
            application.getEnvironment().logError("BF AppEngine : The java type for the binding " + getBinding()
                + ") is invalid. Ignoring it.");
            typeClass = null;
        }
        
        return typeClass;
    }
    
	/**
	 * Peer type (used by RDBMS which accepts BLOBS and/or CLOBS) otherwise value is 'DEFAULT'
	 * @return Possible values are 'DEFAULT', 'BLOB' and 'CLOB'
	 */
	public String getPeerType()
	{
		return _peerType;
	}

	/**
	 * Returns the name of the java class used for encoding/decoding values for this binding.
	 */
	public String getJavaEncoder()
	{
		return _javaEncoder;
	}
    
	/**
     * Gets encoder class.
     */
    protected Class<?> getJavaEncoderClass(LyApplication application)
    {
        if (_javaEncoderClass != null)
            return _javaEncoderClass;
        
        if (_javaEncoder == null)
            return null;
        
        _javaEncoderClass = application.loadClass(_javaEncoder,
             "BF AppEngine : unknown encoder class " + _javaEncoder + " on binding " + _binding);
        
        if ((_javaEncoderClass != null)
                && !LyEncoderInterface.class.isAssignableFrom(_javaEncoderClass))
            throw new IllegalStateException("BF AppEngine : invalid encoder " + _javaEncoder
                + " specified on binding " + _binding + ", it must implement LyEncoderInterface.");
        
        return _javaEncoderClass;
    }
    
    /**
     * Sets the cbind of this binding.
     */
    protected void setCBind(String cbind)
    {
        _cbind = cbind;
    }
    
    /**
     * Sets the default value of this binding.
     */
    protected void setDefault(String def)
    {
        _default = def;
    }
    
    /**
     * Sets the name of the java class for this binding.
     */
    protected void setJavaType(String javaType)
    {
        _javaType = javaType;
    }
    
    /**
     * Sets peer type (used by RDBMS which accepts BLOBS and/or CLOBS) otherwise value is 'DEFAULT'
     * Possible values are 'DEFAULT', 'BLOB' and 'CLOB'
     */
    protected void setPeerType(String peerType)
    {
        _peerType = peerType;
    }
    
    /**
     * Sets the name of the java class used for encoding/decoding values for this binding.
     */
    protected void setJavaEncoder(String javaEncoder)
    {
        _javaEncoder = javaEncoder;
    }
    
    /**
     * Sets whether if this binding is id for the physical layer.
     */
    protected void setId(boolean isId)
    {
        _isId = isId;
    }
    
    /**
     * Builds a logical value for the given fieldInfo from a physical (peer) value.
     *
     * @param     fieldInfo  Field info for which the peer value is decoded.
     * @param     object     The peer value.
     * @param     valueClass Expected class of the logical value.
     * @return    The logical value that was built (instance of the valueClass).
     */
    public Object decodeValue(LyFieldInfo fieldInfo, Object value, Class<?> valueClass)
    {
        LyApplication   application = fieldInfo.getApplication();
        Class<?>        encoderClass = getJavaEncoderClass(application);
        
        if (encoderClass != null)
        {
            try
            {
                LyEncoderInterface encoder = (LyEncoderInterface)encoderClass.newInstance();

                return encoder.decodeValue(fieldInfo, value, valueClass);
            }
            catch (Exception e)
            {
                LyEnvironment   environment = application.getEnvironment();
                
                environment.writeException(e);
            }
        }
        
        return value;
    }
    
    /**
     * Returns an object of the given class from a value of the given fieldInfo.
     * The purpose of this method is to support transformations from logical to physical representation.
     *
     * @param     fieldInfo   Field info for which the logical value is encoded.
     * @param     value       Value of the fieldInfo. This value is an instance of the class
     *                        returned by getValueClass method.
     * @param     objectClass Expected class of the physical value.
     */
    public Object encodeValue(LyFieldInfo fieldInfo, Object value, Class<?> objectClass)
    {
        LyApplication   application = fieldInfo.getApplication();
        Class<?>        encoderClass = getJavaEncoderClass(application);
        
        if (encoderClass != null)
        {
            try
            {
                LyEncoderInterface encoder = (LyEncoderInterface)encoderClass.newInstance();

                return encoder.encodeValue(fieldInfo, value, objectClass);
            }
            catch (Exception e)
            {
                LyEnvironment   environment = application.getEnvironment();
                
                environment.writeException(e);
            }
        }
        
        return value;
    }
} // end class LyFieldBinding
