/*
 * LyUnionBinding.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info.binding;

import java.util.*;

import leon.data.*;
import leon.info.*;


/**
 * Container for the binding information found in the data description for classInfo.
 */
public class LyUnionBinding extends LyClassBinding
{
	public static final String __VERSION =
		"$Id: LyUnionBinding.java 40928 2012-09-25 17:14:19Z jlsaouli $";

	/**
	 * Class binding for class 1 of union
	 */
	LyClassBinding	_classBinding1;

	/**
	 * Class binding for class 2 of union
	 */
	LyClassBinding	_classBinding2;

	/**
	 * Vector of union fields for class 1
	 */
	Vector<LyUnion>	_unions;
	
	/**
	 * Indicates if union is strict or not.
	 */
	boolean			_strict;

	/**
	 * Constructor.
	 */
	public LyUnionBinding()
	{
		this(true);
	}

	/**
	 * Constructor.
	 */
	public LyUnionBinding(boolean strict)
	{
		super(null, null, null);

		_classBinding1 = null;
		_classBinding2 = null;
		_unions = new Vector<LyUnion>(1);
		_strict = strict;
	}

	/**
	 * Method to get the class binding of class 1
	 * @return class binding for class 1
	 */
	public LyClassBinding getClassBinding1()
	{
		return _classBinding1;
	}

	/**
	 * Method to get the class binding of class 2
	 * @return class binding for class 2
	 */
	public LyClassBinding getClassBinding2()
	{
		return _classBinding2;
	}

	/**
	 * Return union filter
	 * @return And of filters of the class bindings.
	 */
	public LyFilter getFilter()
	{
		LyFilter filter1 = ((_classBinding1 == null) ? null : _classBinding1.getFilter());
		LyFilter filter2 = ((_classBinding2 == null) ? null : _classBinding2.getFilter());

		if ((filter1 == null) && (filter2 == null))
			return null;
		else if (filter1 == null)
			return filter2;
		else if (filter2 == null)
			return filter1;
		else
		{
			LyFilter filter = new LyFilter(LyFilter.Type.AND);

			filter.addElement(filter1);
			filter.addElement(filter2);
			return filter;
		}
	}

	/**
	 * Return the union field of class 1 for the given index
	 * @param index index of field
	 * @return Id of field
	 */
	public String getValue1(int index)
	{
		return _unions.elementAt(index).getValue1();
	}

	/**
	 * Return the union field of class 1 for the given index
	 * @param index index of field
	 * @return Id of field
	 */
	public String getValue2(int index)
	{
		return _unions.elementAt(index).getValue2();
	}

	/**
	 * Return all union fields of class 1
	 * @return a vector of fields id
	 */
	public Vector<String> getValues1()
	{
	    int            i, n = _unions.size();
	    Vector<String> values1 = new Vector<String>(n);
	    
	    for (i = 0; i < n; i++)
	        values1.add(_unions.get(i).getValue1());
		return values1;
	}

	/**
	 * Return all union fields of class 2
	 * @return a vector of fields id
	 */
	public Vector<String> getValues2()
	{
        int            i, n = _unions.size();
        Vector<String> values2 = new Vector<String>(n);
        
        for (i = 0; i < n; i++)
            values2.add(_unions.get(i).getValue2());
        return values2;
	}

	/**
	 * Add a class binding
	 * @param classBinding class binding to add
	 */
	public void addClassBinding(LyClassBinding classBinding)
	{
		if (_classBinding1 == null)
			_classBinding1 = classBinding;
		else if (_classBinding2 == null)
			_classBinding2 = classBinding;
	}

    /**
     * Add an union to this binding.
     */
    public void addUnion(LyUnion union)
    {
        _unions.add(union);
    }
    
    /**
     * Clear all union.
     */
    public void clearUnionList()
    {
        _unions.removeAllElements();
    }
    
	/**
	 * set a pair of union fields for class 1 and class 2
	 * @deprecated Use addUnion
	 */
	public void setValues(String value1, String value2)
	{
	    _unions.add(new LyUnion(value1, value2));
	}
	
	/**
     * Return the unions list of this binding.
     */
    public Vector<LyUnion> getUnions()
    {
        return new Vector<LyUnion>(_unions);
    }

    /**
     * Return the union at the given index.
     */
    public LyUnion getUnion(int index)
    {
        return _unions.elementAt(index);
    }
    
	/**
     * Return the number of union fields pair
     * @return the number of union fields pair
     */
    public int getUnionCount()
    {
        return _unions.size();
    }
    
	/**
	 * Return the number of union fields pair
	 * @deprecated Use getUnionCount
	 */
	public int getValueCount()
	{
		return _unions.size();
	}

	/**
	 * Accessor to the peer field related to a given field info.
	 * @param  fieldInfo the reference of a field info.
	 * @return the related field info.
	 */
	public LyField getPeerField(LyFieldInfo fieldInfo)
	{
		LyField peerField = _classBinding1.getPeerField(fieldInfo);

		if (peerField == null)
			peerField = _classBinding2.getPeerField(fieldInfo);

		return peerField;
	}

	/**
	 * Accessor to the peer fields related to a given field info.
	 * @param  fieldInfo the reference of a field info.
	 * @return the associated peer fields.
	 */
	public LyField[] getPeerFields(LyFieldInfo fieldInfo)
	{
		LyField[] fields1 = _classBinding1.getPeerFields(fieldInfo);
		LyField[] fields2 = _classBinding2.getPeerFields(fieldInfo);
		int		  n1 = ((fields1 == null) ? 0 : fields1.length);
		int		  n2 = ((fields2 == null) ? 0 : fields2.length);
		int		  n = n1 + n2;

		if (n == 0)
			return null;

		LyField[] fields = new LyField[n];
		int		  i;

		for (i = 0; i < n1; i++)
			fields[i] = fields1[i];

		for (i = 0; i < n2; i++)
			fields[i + n1] = fields2[i];

		return fields;
	}

	/**
	 * Compare current object with another.
	 * @param o object to compare
	 */
	public boolean equals(Object o)
	{
		if (!(o instanceof LyUnionBinding))
			return false;
		return equals((LyUnionBinding)o);
	}
	
	/**
	 * Compare current object with another.
	 * @param o object to compare
	 */
	public boolean equals(LyUnionBinding o)
	{
		if (o == null)
			return false;
		if (o == this)
			return true;

		LyUnionBinding o1 = (LyUnionBinding)o;

		// Compare sub bindings
		if (getClassBinding1() == null)
		{
			if (o1.getClassBinding1() != null)
				return false;
		}
		else if (!getClassBinding1().equals(o1.getClassBinding1()))
			return false;

		if (getClassBinding2() == null)
		{
			if (o1.getClassBinding2() != null)
				return false;
		}
		else if (!getClassBinding2().equals(o1.getClassBinding2()))
			return false;

		// Compare values
		if (getValues1() == null)
		{
			if (o1.getValues1() != null)
				return false;
		}
		else if (getValues1().equals(o1.getValues1()))
			return false;

		if (getValues2() == null)
		{
			if (o1.getValues2() != null)
				return false;
		}
		else if (getValues2().equals(o1.getValues2()))
			return false;

		return true;
	} // end method equals
	
	/**
	 * Find all bindings (class bindings) for this union binding.
	 * Internal use only
	 * @return the name of physical bindings
	 */
	public List<String> getAllBindings()
	{
		LyClassBinding b1 = getClassBinding1();
		LyClassBinding b2 = getClassBinding2();
		
		List<String> resu = new ArrayList<String>();
		
		if (b1 instanceof LyUnionBinding)
		{
			List<String> resu1  = ((LyUnionBinding)b1).getAllBindings();
			
			resu.addAll(resu1);
		} else if ((b1 != null) && (b1.getBinding() != null))
			resu.add(b1.getBinding());
		
		if (b2 instanceof LyUnionBinding)
		{
			List<String> resu2  = ((LyUnionBinding)b2).getAllBindings();
			
			for(String binding : resu2)
			{
				if ((binding != null) && (!resu.contains(binding)))
					resu.add(binding);
			}
		} else if ((b2 != null) && (b2.getBinding() != null) && (!resu.contains(b2.getBinding())))
			resu.add(b2.getBinding());
		
		return resu;
	}
    
    /**
     * Method to set the class binding of class 1
     * @param classBinding class binding for class 1
     */
    protected void setClassBinding1(LyClassBinding classBinding)
    {
        _classBinding1 = classBinding;
    }

    /**
     * Method to set the class binding of class 2
     * @param classBinding class binding for class 2
     */
    protected void setClassBinding2(LyClassBinding classBinding)
    {
        _classBinding2 = classBinding;
    }
    
    /**
     * Sets the unions list of this binding.
     */
    protected void setUnions(Vector<LyUnion> unions)
    {
        _unions = (unions == null ? null : new Vector<LyUnion>(unions));
    }
    
    /**
     * Checks if the union is strict or not.
     * @return True if strict, false otherwise.
     */
    public final boolean isStrict()
    {
    	return _strict;
    }
    
    /**
     * Sets the union as strict or not.
     * @param strict True if strict, false otherwise.
     */
    protected void setStrict(boolean strict)
    {
    	_strict = strict;
    }
} // end class LyUnionBinding
