/*
 * LyJoinDaemon.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info.daemon;

import leon.info.*;
import leon.info.binding.*;


/**
 * This class holds informations for local multiple relations managed by a join class. This join is
 * defined by a join class, a source relation and a target relation in the join class. The source
 * relation must pointed to the class of the local relation managed by the join class. The target
 * relation must pointed to the class(es) pointed by the local relation managed by the join class.
 * <p>See also tag <JOIN> of leon.dtd.</p>
 * @see leon.data.daemon.LyJoinRelationDaemon
 */
public final class LyJoinDaemon extends LyInfoDaemon
{
	public static final String    __VERSION =
		"$Id: LyJoinDaemon.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The join class of this join. */
	protected LyClassInfo		  _joinClass;

	/** The source relation of this join. */
	protected LyRelationFieldInfo _sourceRelation;

	/** The target relation of this join. */
	protected LyRelationFieldInfo _targetRelation;

	/** The cbind info of this daemon. */
	protected LyClassBinding	  _classBinding;

	/** The fbind info of this daemon for the source foreign key. */
	protected LyFieldBinding	  _sourceBinding;

	/** The fbind info of this daemon for the source foreign key. */
	protected LyFieldBinding	  _targetBinding;

	/**
	 * Default constructor.
	 */
	public LyJoinDaemon()
	{
		super();

		_joinClass = null;
		_sourceRelation = null;
		_targetRelation = null;
	}

	/**
	 * Constructor with all parameters.
	 */
	public LyJoinDaemon(LyClassInfo joinClass, LyRelationFieldInfo sourceRelation,
		LyRelationFieldInfo targetRelation)
	{
		_joinClass = joinClass;
		_sourceRelation = sourceRelation;
		_targetRelation = targetRelation;
	}

	/**
	 * Final step of parsing : check consistency of the daemon.
	 * @param     parent Object info for which this daemon is checked.
	 * @exception IllegalStateException If the source or target relation doesn't belong to join
	 *                                  class.
	 */
	public void check(LyObjectInfo parent)
	{
		if (_joinClass == null)
			return;
		if (!_joinClass.contains(_sourceRelation))
			throw new IllegalStateException("BF AppEngine : sourceRelation " + _sourceRelation
				+ " must be defined in joinClass " + _joinClass);
		if (!_joinClass.contains(_targetRelation))
			throw new IllegalStateException("BF AppEngine : targetRelation " + _targetRelation
				+ " must be defined in joinClass " + _joinClass);
	}

	/**
	 * Accessor to the join class of this join.
	 * @return Join class of this join.
	 */
	public LyClassInfo getJoinClass()
	{
		return _joinClass;
	}

	/**
	 * Accessor to the source relation of this join.
	 * @return The source relation of this join.
	 */
	public LyRelationFieldInfo getSourceRelation()
	{
		return _sourceRelation;
	}

	/**
	 * Accessor to the target relation of this join.
	 * @return The target relation of this join.
	 */
	public LyRelationFieldInfo getTargetRelation()
	{
		return _targetRelation;
	}

	/**
	 * Accessor to the class binding of this join.
	 * @return The class binding of this join
	 */
	public LyClassBinding getClassBinding()
	{
		return _classBinding;
	}

	/**
	 * Accessor to the field binding of the source of this join.
	 * @return The field binding of the source of this join
	 */
	public LyFieldBinding getSourceBinding()
	{
		return _sourceBinding;
	}

	/**
	 * Accessor to the field binding of the targer of this join.
	 * @return The field binding of the target of this join
	 */
	public LyFieldBinding getTargetBinding()
	{
		return _targetBinding;
	}

	/**
	 * Modifier of the join class of this join.
	 * @param joinClass The class info reprenting the join class.
	 */
	public void setJoinClass(LyClassInfo joinClass)
	{
		_joinClass = joinClass;
	}

	/**
	 * Modifier of the source relation of this join.
	 * @param sourceRelation The relation field info reprenting the source relation of this join.
	 */
	public void setSourceRelation(LyRelationFieldInfo sourceRelation)
	{
		_sourceRelation = sourceRelation;
	}

	/**
	 * Modifier of the target relation of this join.
	 * @param targetRelation The relation field info reprenting the target relation of this join.
	 */
	public void setTargetRelation(LyRelationFieldInfo targetRelation)
	{
		_targetRelation = targetRelation;
	}

	/**
	 * Modifier of the class binding of this join.
	 * @param classBinding The class binding of this join
	 */
	public void setClassBinding(LyClassBinding classBinding)
	{
		_classBinding = classBinding;
	}

	/**
	 * Modifier of the field binding of the source of this join.
	 * @param fieldBinding The field binding of the source of this join
	 */
	public void setSourceBinding(LyFieldBinding fieldBinding)
	{
		_sourceBinding = fieldBinding;
	}

	/**
	 * Modifier of the field binding of the target of this join.
	 * @param fieldBinding The field binding of the target of this join
	 */
	public void setTargetBinding(LyFieldBinding fieldBinding)
	{
		_targetBinding = fieldBinding;
	}

	/**
	 * To add a field binding definition to this daemon. Fields binding must be added in the right
	 * order (source, target).
	 * @param fieldBinding the field binding to add.
	 */
	public void addFieldBinding(LyFieldBinding fieldBinding)
	{
		if (_sourceBinding == null)
			setSourceBinding(fieldBinding);
		else if (_targetBinding == null)
			setTargetBinding(fieldBinding);
	}
} // end class LyJoinDaemon
