/*
 * LyGenLocation.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info.location;

import java.util.*;

import leon.info.*;
import leon.info.misc.*;


/**
 * This class implements a generic data location.
 */
public class LyGenLocation extends LyLocation
{
	public static final String __VERSION =
		"$Id: LyGenLocation.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Parameter : NULL. */
	public static final Object PARAM_NULL = new Object();

	/** To specify the base object name of this location. */
	protected LyJavaName      _javaName;

	/**
	 * Indicates if this location can save objects for each update request. If this flag is false, a
	 * LyFullCache is associated with all LyDataSource of this provider.
	 */
	protected boolean		   _persistent;

	/** Hashtable of the specific data information. */
	protected Hashtable<Object, LyData>   _datas;

	/**
	 * Init method.
	 */
	protected void init()
	{
		super.init();

		_javaName = null;
		_datas = null;
		_persistent = true;
	}

    /**
     * Constructor.
     * @param id         identifier of the location
     */
    public LyGenLocation(String id)
    {
        this(id, null, true);
    }
    
	/**
	 * Constructor.
	 * @param id         identifier of the location
	 * @param javaName   Name of the java class of the provider.
	 * @param persistent Boolean that indicates if location is persistent or not.
	 */
	public LyGenLocation(String id, String javaName, boolean persistent)
	{
		super(id, null);

		setJavaName(javaName);
		_persistent = persistent;
	}

	/**
	 * Returns extra tags.
	 */
	protected LyTag[] getExtraTags()
	{
	    return new  LyTag[] { _javaName };
	}
	
	/**
	 * Accessor to the base object name of this location.
	 * @return the base object name of this location
	 */
	public String getJavaName()
	{
		return (_javaName == null ? null : _javaName.getJavaName());
	}

	/**
	 * Accessor of the persistent mode of this location.
	 * @return persistent mode.
	 */
	public boolean isPersistent()
	{
		return _persistent;
	}

	/**
	 * Adds user data on this instance. A user data is a value associated to a name stored in a
	 * hashtable.
	 * @param     name  Name of the user data. If a value was already stored for that name, this
	 *                  value is overriden.
	 * @param     value Value of the user data.
	 * @return    The previous value associated to the given name if there was one, null otherwise.
	 * @exception IllegalArgumentException If the given name or value is null.
	 */
	public final Object putData(Object name, Object value)
	{
		if (name == null)
			throw new IllegalArgumentException("BF AppEngine : data name can't be null");
		if (value == null)
			throw new IllegalArgumentException("BF AppEngine : data value can't be null (cf. " + name + ")");

		// Creates hashtable if null
		if (_datas == null)
			_datas = new Hashtable<Object, LyData>(1);

		// Associates value to the name
        LyData data = new LyData(name, value);

        data.setParent(this);
        return _datas.put(name, data);
	}

	/**
	 * Removes a user data on this instance.
	 * @param     name Name of the data to be removed.
	 * @return    The value that was associated to the given name if there was one, null otherwise.
	 * @exception IllegalArgumentException If the given name is null.
	 * @exception IllegalStateException    If this method is called before any data has been added
	 *                                     (datas hashtable is null).
	 */
	public final Object removeData(Object name)
	{
		if (name == null)
			throw new IllegalArgumentException("BF AppEngine : name parameter can't be null");
		if (_datas == null)
			throw new IllegalStateException("BF AppEngine : no data associated to this object");

		// Removes data associated to that name
		return _datas.remove(name);
	}

	/**
	 * Gets a user data by its name.
	 * @param     name Name for which a value is searched.
	 * @return    Found value if any, or null if no data was stored with the given name.
	 * @exception IllegalArgumentException If the given name is null.
	 */
	public final Object getData(Object name)
	{
		if (name == null)
			throw new IllegalArgumentException("BF AppEngine : name parameter can't be null");

		// Removes data associated to that name
		LyData data = ((_datas == null) ? null : _datas.get(name));
		
		return (data == null ? null : data.getValue());
	}

	/**
     * Gets list of datas associated to this instance.
     *
     * @return Returns the list of all the datas associated to this object info.
     */
    public final Vector<LyData> getDataList()
    {
        if (_datas == null)
            return null;
        
        Enumeration<LyData>     datas = _datas.elements();
        Vector<LyData>          objectDatas = new Vector<LyData>();
        LyData                  data;
        
        while (datas.hasMoreElements())
        {
            data = datas.nextElement();
            objectDatas.add(data);
        }
        
        return objectDatas;
    }
    
	/**
	 * Gets table of datas associated to this instance.
	 * @return Returns the table of all the datas associated to this object info. This is a copy of
	 *         the internal hashtable that contains these datas.
	 */
	public final Hashtable<Object, LyData> getDatas()
	{
		return (_datas == null ? null : new Hashtable<Object, LyData>(_datas));
	}
    
    /**
     * Sets the persistent mode of this location.
     * @param persistent persistent mode.
     */
    protected void setPersistent(boolean persistent)
    {
        _persistent = persistent;
    }
    
    /**
     * Sets datas associated to field bean.
     */
    protected void setDataList(Vector<LyData> datas)
    {
        if ((datas == null)
                || datas.isEmpty())
            _datas = null;
        else
        {
            int     i, n = datas.size();
            LyData  data;
            
            _datas = new Hashtable<Object, LyData>(n);
            
            for (i = 0; i < n; i++)
            {
                data = datas.elementAt(i);
                _datas.put(data.getName(), data);
            }
        }
    }
    
    /**
     * Sets the base object name of this location.
     * @param javaName the base object name of this location
     */
    protected void setJavaName(String javaName)
    {
        _javaName = (javaName == null ? null : new LyJavaName(this, javaName));
    }
} // end class LyGenLocation
