/*
 * LyLdapLocation.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info.location;

import leon.info.*;
import leon.info.binding.*;


/**
 * This class implements a LDAP data location.
 */
public class LyLdapLocation extends LyLocation
{
    public static final String __VERSION =
        "$Id: LyLdapLocation.java 35251 2012-04-17 15:32:36Z jlsaouli $";

    /** To specify the base object name of this location. */
    protected String _baseObjectName;

    /** To specify the user of this location. */
    protected String _user;

    /** To specify the password of this location. */
    protected String _password;

    /** To specify the security type (simple or ssl) of this location. */
    protected String _securityType;

    /**
     * Constructor.
     *
     * @param id  identifier of the location
     */
    public LyLdapLocation(String id)
    {
        this(id, null, null, null, null, "SIMPLE");
    }

    /**
     * Constructor.
     *
     * @param id  identifier of the location
     * @param url url of the provider
     */
    public LyLdapLocation(String id, String url, String baseObjectName, String user,
        String password, String securityType)
    {
        super(id, url);

        setBaseObjectName(baseObjectName);
        setUser(user);
        setPassword(password);
    }

    /**
     * Init method.
     */
    protected void init()
    {
        super.init();

        _baseObjectName = null;
        _user = null;
        _password = null;
    }

    /**
     * Accessor to the base object name of this location.
     *
     * @return the base object name of this location
     */
    public String getBaseObjectName()
    {
        return _baseObjectName;
    }

    /**
     * Accessor to the security type of this location.
     *
     * @return the security type of this location
     */
    public String getSecurityType()
    {
        return _securityType;
    }

    /**
     * Accessor to the user of this location.
     *
     * @return the user of this location
     */
    public String getUser()
    {
        return _user;
    }

    /**
     * Accessor to the password of this location.
     *
     * @return the password of this location
     */
    public String getPassword()
    {
        return _password;
    }

    /**
     * This method allows to test the configuration of a classInfo for the given location This
     * method is called during the check phase of the parsing.
     *
     * @param classInfo the classInfo to check
     */
    public void check(LyClassInfo classInfo)
    {
        // Check Field Binding for LDAP
        LyClassBinding classBinding = classInfo.getBinding();

        if (classBinding instanceof LyLdapClassBinding)
        {
            LyLdapFieldBinding      binding;
            LyFieldInfo             fieldInfo;
            boolean                 parentAttribute = false;
            LyInfoList<LyFieldInfo> fieldInfos = classInfo.getFieldInfos();
            int                     i;

            for (i = 0; i < fieldInfos.getSize(); i++)
            {
                fieldInfo = fieldInfos.getInfo(i);

                if (!(fieldInfo.getBinding() instanceof LyLdapFieldBinding))
                    continue;

                binding = (LyLdapFieldBinding) fieldInfo.getBinding();

                if (binding.isParent() && parentAttribute)
                    throw new IllegalStateException("BF AppEngine : In " + classInfo
                        + ", two fields (or more) have a LDAP binding with the parent attribute set");

                if (binding.isParent())
                    parentAttribute = true;

                if (binding.isParent() && !(fieldInfo instanceof LyRelationFieldInfo))
                    throw new IllegalStateException(
                        "BF AppEngine : Attribut parent must not be set for the LDAP binding because "
                        + fieldInfo + " is not a relation field");
            }
        }
    } // end method check

    /**
     * Sets the base object name of this location.
     *
     * @param baseObjectName the base object name of this location
     */
    protected void setBaseObjectName(String baseObjectName)
    {
        _baseObjectName = baseObjectName;
    }

    /**
     * Sets the user of this location.
     *
     * @param user the user of this location
     */
    protected void setUser(String user)
    {
        _user = user;
    }

    /**
     * Sets the Security Type of this location.
     *
     * @param user the Security Type of this location
     */
    protected void setSecurityType(String SecurityType)
    {
        _securityType = SecurityType;
    }

    /**
     * Sets the password of this location.
     *
     * @param password the password of this location
     */
    protected void setPassword(String password)
    {
        _password = password;
    }
} // end class LyLdapLocation
