/*
 * LySqlLocation.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info.location;

import leon.info.*;
import leon.info.misc.*;

/**
 * This class implements a SQL data location.
 */
public class LySqlLocation extends LyLocation
{
	public static final String __VERSION =
		"$Id: LySqlLocation.java 30280 2011-03-30 09:34:49Z bmichel $";

	/** Default separator used between multiple values. */
	public static final String VALUE_SEP = "\n".intern();

	/** To specify the JDBC driver of the SQL location. */
	protected String		   _driver;

	/** To specify the base object name of this location. */
	protected LyJavaName      _javaName;

	/** To specify the user of the SQL location. */
	protected String		   _user;

	/** To specify the password of the SQL location. */
	protected String		   _password;

	/** To specify the database name of the SQL location. */
	protected String		   _dbName;

	/** To specify the number of connections of the SQL location. */
	protected String		   _nbConnection;

	/** Value separator for multiple values. */
	protected String		   _valueSep;

	/** To know if values may be trimed. */
	protected String		   _trim;

	/** To know if the database metadata must be used to create the physical layer. */
	protected String		   _metadata;

	/**
	 * Init method. LyInfo#init
	 */
	protected void init()
	{
		super.init();

		_driver = null;
		_javaName = null;
		_user = null;
		_password = null;
		_dbName = null;
		_nbConnection = null;
		_valueSep = VALUE_SEP;
		_trim = null;
		_metadata = null;
	}
    
    /**
     * Constructor.
     * @param id           identifier of the location
     */
    public LySqlLocation(String id)
    {
        this(id, null, null, null, null, null, null, null, null, null, null);
    }

	/**
	 * Constructor.
	 * @param id           identifier of the location
	 * @param javaName     Name of the java class of the provider. May be null to use default.
	 * @param url          url of the provider
	 * @param driver       JDBC driver
	 * @param user         User name to connect to database
	 * @param password     User password to connect to database
	 * @param dbName       Database name
	 * @param nbConnection Number of connections
	 * @param valueSep     Value separator
	 * @param trim         Trim values or not
	 * @param metadata     Use database metadata
	 */
	public LySqlLocation(String id, String javaName, String url, String driver, String user,
		String password, String dbName, String nbConnection, String valueSep, String trim,
		String metadata)
	{
		super(id, url);

		_driver = driver;
		setJavaName(javaName);
		_user = user;
		_password = password;
		_dbName = dbName;
		if(nbConnection != null)
			_nbConnection = nbConnection;
		if (valueSep != null)
			_valueSep = valueSep;
		_trim = trim;
		_metadata = metadata;
	}

	/**
     * Returns extra tags.
     */
    protected LyTag[] getExtraTags()
    {
        return new  LyTag[] { _javaName };
    }
    
	/**
	 * Accessor to the driver of this location.
	 * @return the driver of this location
	 */
	public String getDriver()
	{
		return _driver;
	}

	/**
	 * Accessor to the base object name of this location May be null.
	 * @return the base object name of this location
	 */
	public String getJavaName()
	{
		return (_javaName == null ? null : _javaName.getJavaName());
	}

	/**
	 * Accessor to the user of this location.
	 * @return the user of this location
	 */
	public String getUser()
	{
		return _user;
	}

	/**
	 * Accessor to the password of this location.
	 * @return the password of this location
	 */
	public String getPassword()
	{
		return _password;
	}

	/**
	 * Accessor to the database name of this location.
	 * @return the database name of this location
	 */
	public String getDbName()
	{
		return _dbName;
	}

	/**
	 * Accessor to the number of connection of this location.
	 * @return the number of connection of this location
	 */
	public String getNbConnection()
	{
		return _nbConnection;
	}

	/**
	 * Accessor to the value separator of this location.
	 * @return the value separator of this location
	 */
	public String getValueSep()
	{
		return _valueSep;
	}

	/**
	 * To know if values may be trimmed.
	 * @return true if the values may be trimmed else false.
	 */
	public String isTrim()
	{
		return _trim;
	}

	/**
	 * To know if metadata must be used.
	 * @return true if metadata must be used else false.
	 */
	public String useMetadata()
	{
		return _metadata;
	}
    
    /**
     * Sets the driver of this location.
     * @param driver the driver of this location
     */
    protected void setDriver(String driver)
    {
        _driver = driver;
    }

    /**
     * Sets the base object name of this location May be null.
     * @param javaName the base object name of this location
     */
    protected void setJavaName(String javaName)
    {
        _javaName = (javaName == null ? null : new LyJavaName(this, javaName));
    }

    /**
     * Sets the user of this location.
     * @param user the user of this location
     */
    protected void setUser(String user)
    {
        _user = user;
    }

    /**
     * Sets the password of this location.
     * @param password the password of this location
     */
    protected void setPassword(String  password)
    {
        _password = password;
    }

    /**
     * Sets the database name of this location.
     * @param dbName the database name of this location
     */
    protected void setDbName(String dbName)
    {
        _dbName = dbName;
    }

    /**
     * Accessor to the number of connection of this location.
     * @param nbConnection the number of connection of this location
     */
    protected void setNbConnection(String nbConnection)
    {
        _nbConnection = nbConnection;
    }

    /**
     * Accessor to the value separator of this location.
     * @param valueSep the value separator of this location
     */
    protected void setValueSep(String valueSep)
    {
        _valueSep = valueSep;
    }

    /**
     * Sets if values may be trimmed.
     * @param trim true if the values may be trimmed else false.
     */
    protected void setTrim(String trim)
    {
        _trim = trim;
    }

    /**
     * Sets if metadata must be used.
     * @param metadata true if metadata must be used else false.
     */
    protected void setUseMetadata(String metadata)
    {
        _metadata = metadata;
    }
} // end class LySqlLocation
