/*
 * LyFieldBean.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info.misc;

import java.util.*;
import leon.info.*;

/**
 * This class defines a field bean : Graphical object used for interacting with a field in some
 * circumstances (consultation, edition, filtering).
 */
public class LyFieldBean extends LyTag
{
	public static final String __VERSION =
		"$Id: LyFieldBean.java 28422 2010-11-03 12:03:38Z ehelmste $";

    /** Bean types. */
    public enum Type
    {
    	ALL,
        CONSULT,
        EDIT,
        FILTER
    }

	/** Java name of the field bean. */
	LyJavaName				   _javaName;

	/** Type of action where the bean is used. */
	Type					   _actionType;

	/** Hashtable of the specific data information. */
	Hashtable<String, LyData>  _datas;

    /**
     * Constructor of a field bean.
     * @param javaName Java name of the field bean.
     */
    public LyFieldBean(String javaName)
    {
        this(javaName, null);
    }
    
	/**
	 * Constructor of a field bean.
	 * @param javaName Java name of the field bean.
	 * @param actionId Type of action where the bean is used.
	 */
	public LyFieldBean(String javaName, Type actionType)
	{
		setJavaName(javaName);
		setActionType(actionType);
		setDataList(null);
	}

	/**
     * Returns extra tags.
     */
    protected LyTag[] getExtraTags()
    {
        return new LyTag[] { _javaName };
    }
    
	/**
	 * Gets java name of the field bean.
	 */
	public final String getJavaName()
	{
		return (_javaName == null ? null : _javaName.getJavaName());
	}

	/**
	 * Gets type of action where the bean is used.
	 */
	public final Type getActionType()
	{
		return _actionType;
	}

	/**
	 * Gets a user data by its name.
	 * @param     name Name for which a value is searched.
	 * @return    Found value if any, or null if no data was stored with the given name.
	 * @exception IllegalArgumentException If the given name is null.
	 */
	public final String getData(String name)
	{
        if (name == null)
            throw new IllegalArgumentException("BF AppEngine : name parameter can't be null");
        if (_datas == null)
            return null;

        // Gets data associated to that name
        LyData data = _datas.get(name);

        return ((data == null) ? null : (String)data.getValue());
	}

	/**
	 * Gets table of datas associated to this instance.
	 * @return Returns the table of all the datas associated to this object info. This is a copy of
	 *         the internal hashtable that contains these datas.
	 */
	public final Hashtable<String, String> getDatas()
	{
		if (_datas == null) 
		   return null;
		

        Enumeration<LyData>       datas = _datas.elements();
        Hashtable<String, String> beanDatas = new Hashtable<String, String>();
        LyData                    data;
        
        while (datas.hasMoreElements())
        {
            data = datas.nextElement();
            beanDatas.put((String)data.getName(), (String)data.getValue());
        }
		
		return beanDatas;
	}

	/**
     * Gets list of data associated to this instance.
     *
     * @return Returns the list of all the data associated to this object info.
     */
    public final Vector<LyData> getDataList()
    {
        if (_datas == null)
            return null;
        
        Enumeration<LyData>     datas = _datas.elements();
        Vector<LyData>          beanDatas = new Vector<LyData>();
        LyData                  data;
        
        while (datas.hasMoreElements())
        {
            data = datas.nextElement();
            beanDatas.add(data);
        }
        
        return beanDatas;
    }
    
	/**
	 * Adds user data on this instance. A user data is a value associated to a name stored in a
	 * hashtable.
	 * @param     name  Name of the user data. If a value was already stored for that name, this
	 *                  value is overriden.
	 * @param     value Value of the user data.
	 * @return    The previous value associated to the given name if there was one, null otherwise.
	 * @exception IllegalArgumentException If the given name or value is null.
	 */
	public final String putData(String name, String value)
	{
		if (name == null)
			throw new IllegalArgumentException("BF AppEngine : name of a data can't be null");
		if (value == null)
			throw new IllegalArgumentException("BF AppEngine : value of a data can't be null");

		// Creates hashtable if null
		if (_datas == null)
			_datas = new Hashtable<String, LyData>(1);

		// Associates value to the name
		LyData    data = new LyData(name, value);

		data.setParent(getParent());
		_datas.put(name, data);
		return value;
	}
    
    /**
     * Sets java name of the field bean.
     * @param javaName Java name of the field bean.
     */
    protected void setJavaName(String javaName)
    {
        _javaName = (javaName == null ? null : new LyJavaName(_parent, javaName));
    }
    
    /**
     * Sets action type of the field bean.
     * @param actionId Type of action where the bean is used.
     */
    protected void setActionType(Type actionType)
    {
        _actionType = (actionType == null ? Type.ALL : actionType);
    }
    
    /**
     * Sets datas associated to field bean.
     */
    protected void setDataList(Vector<LyData> datas)
    {
        if ((datas == null)
                || datas.isEmpty())
            _datas = null;
        else
        {
            int     i, n = datas.size();
            LyData  data;
            
            _datas = new Hashtable<String, LyData>(n);
            
            for (i = 0; i < n; i++)
            {
                data = datas.elementAt(i);
                _datas.put((String)data.getName(), data);
            }
        }
    }
} // end class LyFieldBean
