/*
 * LyParameter.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info.misc;

import leon.app.*;
import leon.info.*;
import leon.info.infointerface.*;


/**
 * This class is used to store a format on a unit. A parameter is composed of a value, an editable
 * property, a size and a java name (name of a java class that implements LyFormatterInterface).
 */
public class LyFormat extends LyTag
{
	public static final String __VERSION = "$Id: LyFormat.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Value of the format. */
	protected String		   _value;

	/** Indicates if this format is used for edition. */
	protected boolean		   _editable;

	/** Size of the text field corresponding use to edit values according to this format. */
	protected int			   _nCols;

	/** Name of the class that implements LyFormatterInterface. */
	protected String		   _javaName;

	/** Formatter class. */
	protected Class<?>		   _formatterClass;

    /**
     * Constructor.
     */
    protected LyFormat()
    {
        setValue("*");
    }
    
	/**
	 * Constructor.
	 */
	public LyFormat(String value, boolean editable, int nCols, String javaName)
	{
		setValue(value);
		setEditable(editable);
		setNCols(nCols);
		setJavaName(javaName);
	}

	/**
	 * Accessor for value.
	 */
	public String getValue()
	{
		return _value;
	}

	/**
	 * Accessor for editable property.
	 */
	public boolean isEditable()
	{
		return _editable;
	}

	/**
	 * Accessor for nCols property.
	 */
	public int getNCols()
	{
		return _nCols;
	}

	/**
	 * Accessor for javaName property.
	 */
	public String getJavaName()
	{
		return _javaName;
	}

	/**
	 * Get formatter instance.
	 */
	protected LyFormatterInterface getFormatter(LyApplication application)
	{
		if (_javaName == null)
			return null;

		try
		{
			if (_formatterClass == null)
				_formatterClass = application.loadClass(_javaName);

			return (LyFormatterInterface)_formatterClass.newInstance();
		}
		catch (Exception e)
		{
			application.getEnvironment().writeException(e);
			return null;
		}
	}

	/**
	 * Method to format a value to a string with the given unit.
	 * @param  fieldInfo the fieldInfo for which a value is formatted
	 * @param  value     the value to format
	 * @param  unit      the destination unit of the string value
	 * @return The string corresponding to the given value and unit.
	 */
	public String toString(LyFieldInfo fieldInfo, Object value, LyUnit unit)
	{
		try
		{
			LyApplication		 application = fieldInfo.getApplication();
			LyFormatterInterface formatter = getFormatter(application);

			if (formatter != null)
				return formatter.toString(fieldInfo, value, unit);
		}
		catch (Exception e)
		{
			fieldInfo.getEnvironment().writeException(e);
		}

		return null;
	}

	/**
	 * Method to format a value to a displayable string with the given unit.
	 * @param  fieldInfo the fieldInfo for which a value is formatted
	 * @param  value     the value to format
	 * @param  unit      the destination unit of the string value
	 * @return The displayable string corresponding to the given value and unit.
	 */
	public String toName(LyFieldInfo fieldInfo, Object value, LyUnit unit)
	{
		try
		{
			LyApplication		 application = fieldInfo.getApplication();
			LyFormatterInterface formatter = getFormatter(application);

			if (formatter != null)
				return formatter.toName(fieldInfo, value, unit);
		}
		catch (Exception e)
		{
			fieldInfo.getEnvironment().writeException(e);
		}

		return null;
	}

	/**
	 * Method to format a value to a short displayable string with the given unit.
	 * @param  fieldInfo the fieldInfo for which a value is formatted
	 * @param  value     the value to format
	 * @param  unit      the destination unit of the string value
	 * @return The short displayable string corresponding to the given value and unit.
	 */
	public String toShortName(LyFieldInfo fieldInfo, Object value, LyUnit unit)
	{
		try
		{
			LyApplication		 application = fieldInfo.getApplication();
			LyFormatterInterface formatter = getFormatter(application);

			if (formatter != null)
				return formatter.toShortName(fieldInfo, value, unit);
		}
		catch (Exception e)
		{
			fieldInfo.getEnvironment().writeException(e);
		}

		return null;
	}

	/**
	 * Method to parse a string value with a given unit.
	 * @param  fieldInfo the fieldInfo for which the string value is parsed.
	 * @param  value     the string value to parse
	 * @param  source    the unit of the string value
	 * @return The value corresponding to the given string and unit.
	 */
	public Object parse(LyFieldInfo fieldInfo, String value, LyUnit unit)
	{
		try
		{
			LyApplication		 application = fieldInfo.getApplication();
			LyFormatterInterface formatter = getFormatter(application);

			if (formatter != null)
				return formatter.parse(fieldInfo, value, unit);
		}
		catch (Exception e)
		{
			fieldInfo.getEnvironment().writeException(e);
		}

		return null;
	}

	/**
	 * Methode to complete the given partial value.
	 * @param  fieldInfo the fieldInfo for which the string value is completed.
	 * @param  value     the string value to complete
	 * @param  unit      the unit of the string value
	 * @return the completed String or null if the generic completion has to be done.
	 */
	public String completeValue(LyFieldInfo fieldInfo, String value, LyUnit unit)
	{
		try
		{
			LyApplication		 application = fieldInfo.getApplication();
			LyFormatterInterface formatter = getFormatter(application);

			if ((formatter != null) && (formatter instanceof LyCompleteValueFormatterInterface))
				return ((LyCompleteValueFormatterInterface)formatter).completeValue(fieldInfo,
					value, unit);
		}
		catch (Exception e)
		{
			fieldInfo.getEnvironment().writeException(e);
		}

		return null;
	}
    
    /**
     * Sets java name of this format.
     * @param javaName Java name of this format.
     */
    protected void setJavaName(String javaName)
    {
        _javaName = javaName;
    }
   
    /**
     * Sets value of this format.
     * @param value Value of this format.
     */
    protected void setValue(String value)
    {
        _value = ((value == null) ? null : value.intern());
    }
    
    /**
     * Sets the editable property.
     */
    protected void setEditable(boolean editable)
    {
        _editable = editable;
    }

    /**
     * Sets the nCols property.
     */
    protected void setNCols(int nCols)
    {
        _nCols = nCols;
    }
} // end class LyFormat
