/*
 * LyFormula.java
 * Copyright (c) 2000-2010, W4.
 * All rights reserved.
 */

package leon.info.misc;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;

import leon.app.*;
import leon.data.*;
import leon.data.datainterface.*;
import leon.info.*;
import leon.misc.*;


/**
 * Class used to compute a formula on a LyFieldInfo.
 */
public abstract class LyFormula
    implements LyFilterElement,
        LyClassControl,
        LyFieldInfoControl
{   
    /** FieldInfo that used the formula. */
    LyInfo                    _info;
    
    /** The string expression. */
    String                    _expression;
    
    /** The compiled expression. */
    Object                    _compiledExpression;
    
    /** Field names. */
    String[]                  _formulaNames;
        
    /** Contains all fieldInfos necessary to compute value. */
    LyInfoList<LyFieldInfo>   _formulaFields;
    
    /** Boolean indicating if the filter expression is inverted or not. */
    boolean                  _not;
    
    /**
     * Default constructor.
     * Invoked by newInstance when checking expressions from studio.
     */
    protected LyFormula()
    { 
    }
    
    /**
     * Constructs a formula for a given fieldInfo.
     */
    protected LyFormula(LyInfo info, String expression)
    {
        List<String>    fieldNames = new ArrayList<String>();
            
        _info = info;
        _expression = expression;
        _compiledExpression = compileExpression(_expression, fieldNames);
        _not = false;

        if (fieldNames != null)
            _formulaNames = fieldNames.toArray(new String[fieldNames.size()]);
    }
    
    /**
     * Clones this formula.
     */
    public LyFormula cloneFormula()
    {
        return formulaFactory(_info, _expression);
    }
    
    /*
     * Return, the info related to this formula.
     */
    protected LyInfo getInfo()
    {
        return _info;
    }
    
    /**
     * Factory that builds the formula with chosen expression language.
     */
    public static LyFormula formulaFactory(LyInfo info, String expression)
    {
        if (expression == null)
            return null;
        
        LyFormula       formula = null;
        LyEnvironment   environment = info.getEnvironment();
        String          className = environment.getEnv("_LY_FORMULA_CLASS");
        
        if (className == null)
        {
            // Default expression language is MVEL
            if (expression != null && expression.startsWith("webservice://"))
            	className = "leon.info.misc.LyWSFormula";
            else
            	className = "leon.info.misc.LyMvelFormula";
        }
        try
        {
            // Create a logger of given class
            ClassLoader classLoader = environment.getClassLoader();
            Class<?>    formulaClass = ((classLoader != null) ?
                classLoader.loadClass(className) : Class.forName(className));
            Class<?>[]  types = new Class[2];
            Object[]    parameters = new Object[2];

            types[0] = LyInfo.class;
            parameters[0] = info;
            types[1] = String.class;
            parameters[1] = expression;

            Constructor<?> constructor = formulaClass.getConstructor(types);

            formula = (LyFormula)constructor.newInstance(parameters);
        }
        catch (InvocationTargetException e)
        {
            environment.writeException(e.getTargetException());
        }
        catch (Exception e)
        {
            environment.writeException(e);
        }
        
        return formula;
    }

    /**
     * Factory that checks the formula with chosen expression language.
     */
    public static String checkExpression(LyEnvironment environment, String expression)
    {
        LyFormula       formula = null;
        String          className = environment.getEnv("_LY_FORMULA_CLASS");
        List<String>    fieldNames = new ArrayList<String>();
        
        if (className == null)
            // Default expression language is MVEL
            className = "leon.info.misc.LyMvelFormula";
        
        try
        {
            // Create a logger of given class
            ClassLoader classLoader = environment.getClassLoader();
            Class<?>    formulaClass = ((classLoader != null) ?
                classLoader.loadClass(className) : Class.forName(className));

            formula = (LyFormula)formulaClass.newInstance();
            formula.compileExpression(expression, fieldNames);
        }
        catch (Exception e)
        {
            environment.logError("BF AppEngine : invalid expression : " + expression);
            return e.getMessage();
        }
        
        return null;
    }
    
    /**
     * Compile expression.
     */
    protected abstract Object compileExpression(String expression, List<String> fieldNames);
    
    /**
     * Evaluate the formula
     * Return depends on the fieldInto type
     * All values are contained in the LyObject
     */
    public abstract Object getValue(LyObject instance);

    /**
     * Evaluate the formula
     * Return depends on the fieldInto type
     * The value set must contains all the values necessary to compute the formula
     */
    public abstract Object getValue(LyValueSet values);

    /**
     * Gets expression of this formula.
     */
    public String getExpression()
    {
        return _expression;
    }
    
    /*
     * Return all the dependencies for this formula
     */
    public LyInfoList<LyFieldInfo> getDependencies(LyClassInfo classInfo)
    {
        if (_formulaNames == null)
            return null;

        LyInfoList<LyFieldInfo> dependencies = new LyInfoList<LyFieldInfo>();
        LyFieldInfo             fieldInfo;
        int                     i, n = _formulaNames.length;

        for (i = 0; i < n; i++)
        {
            fieldInfo = getFieldInfo(classInfo, _formulaNames[i]);
            if (fieldInfo != null)
                dependencies.addInfo(fieldInfo);
        }

        return dependencies;
    } // end method getDependencies
    
    /**
     * Returns a vector of all fieldInfo necessary to compute this formula.
     */
    public final String[] getFormulaNames()
    {
    	return _formulaNames;
    }
    
    /**
     * Returns a list of all fieldInfo necessary to compute this formula.
     */
    public LyInfoList<LyFieldInfo> getFormulaFields()
    {
        // if we already have fill the vector, send it
        if (_formulaFields != null)
            return _formulaFields;
        
        if (_formulaNames == null)
            return null;

        // else we create it and we fill it by asking to all subFormula and all values
        _formulaFields = new LyInfoList<LyFieldInfo>();

        LyFieldInfo     fieldInfo;
        int             i, n = _formulaNames.length;

        for (i = 0; i < n; i++)
        {
            fieldInfo = getFieldInfo(null, _formulaNames[i]);
            if (fieldInfo != null)
                _formulaFields.addInfo(fieldInfo);
        }

        return _formulaFields;
    } // end method getFormulaFields
    
    /**
     * Checks if a formula contains the given field info.
     *
     * @param  fieldInfo The fieldInfo.
     * @return True if the field info is found, false otherwise.
     */
    public boolean contains(LyFieldInfo fieldInfo)
    {
        LyInfoList<LyFieldInfo> formulaFields = getFormulaFields();
        
        return (formulaFields == null ? false : formulaFields.contains(fieldInfo));
    }
    
    /*
     * Return, if any, the field info used for this value.
     */
    protected LyFieldInfo getFieldInfo(LyClassInfo classInfo, String name)
    {
        if (name == null)
            return null;
        
        LyApplication application = _info.getApplication();
        
        if (name.startsWith("$"))
        {
            if (classInfo != null)
                // Fixed variable, not considered as dependency
                return null;
            
            name = name.substring(1);
            if (!(application.getInfo(name) instanceof LyFieldInfo)) // Fixed field
                // Session variable
                return null;
        }
        
        LyFieldInfo   field = null;

        if (classInfo != null)
            field = classInfo.getFieldInfo(name, false);
        else
        {
            LyInfo  info = application.getInfo(name);
            
            if (info instanceof LyFieldInfo)
                field = (LyFieldInfo)info;
        }

        return field;
    } // end method getDependency
    
    /**
     * Gets value for a given variable name and a context object.
     * @param name Name of the variable, either a field name or a session variable.
     * @param contextObj Context object, either a LyObject, LyValueSet, LyObjectId, LyObjectId[].
     * @return The found value.
     */
    protected Object getValue(String name, Object contextObj)
    {
        LyApplication application = _info.getApplication();
        LyInfo        info;
        
        // May be a session variable or a fixed attribute
        if (name.charAt(0) == '$')
        {
            // Try first to find value on session
            name = name.substring(1);

            LyEvaluator evaluator = application.getCurrentEvaluator();
            Object      value = (evaluator != null ? evaluator.getKeyValue(name) : null);
            
            return (value == null ? getNullValue() : value);
        }
        
        if (contextObj instanceof LyEvaluator)
        {
            Object      value = ((LyEvaluator)contextObj).getKeyValue(name);
        
            return (value == null ? getNullValue()
            	: (value instanceof LyValue ? ((LyValue)value).getValue() : value));
        }
        
        LyFieldInfo     fieldInfo = null;
        LyClassInfo		classInfo = null;
        
        info = application.getInfo(name);
        
        if (info instanceof LyClassInfo)
        	classInfo = (LyClassInfo)info;
        else if (info instanceof LyFieldInfo)
            fieldInfo = (LyFieldInfo)info;
        else if (info instanceof LyChoiceOption)
            return ((LyChoiceOption)info).getValue();
        else
        	return null;
        
        // ContextObj is a ValueSet : extract value from valueSet
        if ((contextObj instanceof LyValueSet)
        	&& (fieldInfo != null))
        {
            LyValue     fieldValue = ((LyValueSet)contextObj).getFieldValue(fieldInfo);
            Object      value = (fieldValue == null ? null : fieldValue.getValue());
            
            return (value == null ? getNullValue() : value);
        }
        
        LyObject		object = null;
        
        if (contextObj instanceof LyObject)
        	object = (LyObject)contextObj;
        else if (contextObj instanceof LyObjectId)
            // ContextObj is a object id : retrieve corresponding object
            object = ((LyObjectId)contextObj).getLogicalObject();
        
        if (object != null)
        {
        	Object		value = null;
        	
        	if (fieldInfo != null)
        	{
	        	value = object.getValue(fieldInfo);
	            return (value == null ? getNullValue() : value);
        	}
        	else if (classInfo != null)
        	{		
        		LyObjectList	list = object.getList(classInfo.getApplication().getCurrentSession(), classInfo);
        		int				n = (list == null ? 0 : list.getSize());
        		
        		if (n == 0)
        			return null;
        		else if (n == 1)
        			return list.getObject(0);
        		else
        			return list;
        	}
        }
       
        LyObjectList	objects = null;
        
        if (contextObj instanceof LyObjectList)
            objects = (LyObjectList)contextObj;
        else if (contextObj instanceof LyObjectId[])
        {
        	// ContextObj is an array of object ids : compute value from objects list
            LyObjectId[]  objectIds = (LyObjectId[])contextObj;
            int           n = objectIds.length;
            
            if (n == 0)
                return getNullValue();
            if (n == 1)
                contextObj = objectIds[0].getLogicalObject();
            else
            {
            	LySimpleObjectList	list = null;

                for (int i = 0; i < n; i++)
                {
                    if (objectIds[i] == null)
                        continue;
                    
                    object = objectIds[i].getLogicalObject();
                    if (object != null)
                    {
                        if (list == null)
                        	list = new LySimpleObjectList();
                        list.addObject(object);
                    }
                }
                
                objects = list;
            } 
        }

        if (objects != null)
        {
            if (fieldInfo != null)
            	return (objects == null? getNullValue() : fieldInfo.compute(objects));
            else if (classInfo != null)
            	return (objects == null? getNullValue() : objects.getList(null, classInfo));
        }
        
        // Not found
        return getNullValue();
    }
    
    /**
     * Value to be used in formula when object value is null.
     */
    protected Object getNullValue()
    {
        if (_info instanceof LyNumberFieldInfo)
            return 0;
        else if ((_info instanceof LyTextFieldInfo)
                 || (_info instanceof LyFileFieldInfo))
            return "";
        else
            return null;
    }
    
    /**
     * Returns fields imported through relations for the given class info.
     */
    public LyInfoList<LyFieldInfo> getImports(LyClassInfo classInfo)
    {
        LyInfoList<LyFieldInfo> formulaFields = getFormulaFields();
        
        if (formulaFields == null)
            return null;
        
        int                         i, n = formulaFields.getSize();
        LyFieldInfo                 fieldInfo, importedField;
        int                         p, p2, l = _expression.length();
        char                        c;
        String                      fieldId;
        LyInfoList<LyFieldInfo>     importedFields = null;
        
        for (i = 0; i < n; i++)
        {
            fieldInfo = formulaFields.getInfo(i);
            if ((fieldInfo instanceof LyRelationFieldInfo)
                && ((LyRelationFieldInfo)fieldInfo).relatedWith(classInfo))
            {
                p = 0;
                while ((p >= 0)
                   && (p < l)
                   && (p = _expression.indexOf(fieldInfo.getId(), p)) >= 0)
                {
                    p = _expression.indexOf('.', p);
                    if (p >= 0)
                    {
                        for (p2 = p + 1; p2 < l; p2++)
                        {
                            c = _expression.charAt(p2);
                            if (!Character.isDigit(c)
                                    && !Character.isLetter(c)
                                    && (c != '_'))
                                break;
                        }
                        
                        fieldId = _expression.substring(p + 1, p2);
                        importedField = classInfo.getFieldInfo(fieldId, false);
                        if (importedField != null)
                        {
                            if (importedFields == null)
                                importedFields = new LyInfoList<LyFieldInfo>();
                            importedFields.addInfo(importedField);
                        }
                        
                        p = p2;
                    }
                }
            }
        }
        
        return importedFields;
    }
    
    /**
     * Indicates whether the specified object matches the filter criterion.
     *
     * @param  object Object for which the filter criterion is checked.
     * @return True if the objects matches criterion, false otherwise.
     */
    public boolean match(Object object)
    {
        if (object instanceof LyObject)
            return (getValue((LyObject)object).equals(!_not));
        else
            return false;
    }

    /**
     * Indicates whether the specified values matches the filter criterion.
     *
     * @param  values List of values for which the filter criterion is checked.
     * @return True if the values matches criterion, false otherwise.
     */
    public boolean match(LyValueSet values)
    {
        return (getValue(values).equals(!_not));
    }

    /**
     * Writes XML corresponding to this filter element.
     *
     * @param pw PrintWriter on which the XML is written.
     */
    public void writeXml(PrintWriter pw)
    {
    }

    /**
     * Returns a displayable string of this element.
     *
     * @return A displayable string representation of this element.
     */
    public String getName()
    {
        return _expression;
    }
    
    /**
     * Set the inverted mode of this formula.
     */
    public void setInverted(boolean not)
    {        
        _not = not;
    }
    
    /**
     * Global control of values before creating or modifying an object. If needed, extra values may
     * be added during this control.
     *
     * @param  control   the description of the control.
     * @param  object    the modified object.
     * @param  newValues Set of values to control.
     * @param  args      the list of arguments specified in the XML description of the control
     * @return STATUS_ERROR if the control raises an error STATUS_WARNING if the control raises a
     *         warning STATUS_OK if the control is correct All other positive short values for user
     *         specific warning level.
     */
    public short validate(LyControl control, LyObject object, LyValueSet newValues, LyArgument[] args)
    {
        LyValueSet  values = new LyValueSet();
        LyClassInfo classInfo = object.getClassInfo();
        int         i, n = classInfo.getFieldInfoCount();
        LyFieldInfo fieldInfo;
        
        values.addValues(newValues);
        for (i = 0; i < n; i++)
        {
            fieldInfo = classInfo.getFieldInfo(i);
            if (!values.contains(fieldInfo))
                values.addValue(object.getFieldValue(fieldInfo));
        }
        
        boolean     result = getValue(values).equals(!_not);
        
        if (result)
            return LyControlInterface.STATUS_OK;
        
        String      message = control.getMessage();
        
        if (message == null)
            message = "LY_ER_CONTROL_FORMULA";
        
        LyError     error = new LyError(LyObject.LEVEL_ERROR, message,
                new String[] { _expression }, null, object, null);
        
        object.addError(error);
        return LyControlInterface.STATUS_ERROR;
    }
    
    /**
     * Dynamic check the value for the given field. This method is called when there is a
     * modification of the field in the user interface (Value changed, focus lost...).
     *
     * @param  control  the description of the control.
     * @param  object   the modified object.
     * @param  newValue field info value to check. This parameter cannot be null.
     * @param  args     the list of arguments specified in the XML description of the control
     * @return STATUS_ERROR if the control raises an error STATUS_WARNING if the control raises a
     *         warning STATUS_OK if the control is correct All other positive short values for user
     *         specific warning level.
     */
    public short checkNewValue(LyControl control, LyObject object, LyFieldInfoValue newValue, LyArgument[] args)
    {
        LyValueSet  values = new LyValueSet();
        
        values.addValues(object.getValues());
        values.addValue(newValue);

        boolean     result = getValue(values).equals(!_not);
        
        if (result)
            return LyControlInterface.STATUS_OK;
        
        String      message = control.getMessage();
        
        if (message == null)
            message = "LY_ER_CONTROL_FORMULA";
        
        LyFieldInfo fieldInfo = newValue.getFieldInfo();
        LyError     error = new LyError(LyObject.LEVEL_ERROR, message,
                new String[] { _expression }, fieldInfo, object, fieldInfo);
        
        object.addError(error);
        return LyControlInterface.STATUS_ERROR;
    }
} // end class LyFormula
