/*
 * LyMvelFormula.java
 * Copyright (c) 2000-2010, W4.
 * All rights reserved.
 */

package leon.info.misc;

import java.util.*;

import leon.data.*;
import leon.info.*;
import leon.misc.*;

import org.mvel2.*;
import org.mvel2.integration.*;
import org.mvel2.optimizers.dynamic.*;


/**
 * Class used to compute a formula on a LyFieldInfo.
 */
public class LyMvelFormula extends LyFormula
{
    /** Property handler is used to compute values of variables used in formula. */
    static _LyPropertyHandler    _propertyHandler;
    
    /**
     * Default constructor.
     * Invoked by newInstance when checking expressions from studio.
     */
    protected LyMvelFormula()
    { 
    }
    
    /**
     * Constructs a formula for a given info.
     */
    public LyMvelFormula(LyInfo info, String expression)
    {
        super(info, expression);
        
        if (_propertyHandler == null)
        {
        	_LyPropertyHandler	propertyHandler = new _LyPropertyHandler();
            
            PropertyHandlerFactory.registerPropertyHandler(LyObject.class, propertyHandler);
            PropertyHandlerFactory.registerPropertyHandler(LyValueSet.class, propertyHandler);
            PropertyHandlerFactory.registerPropertyHandler(LyObjectId.class, propertyHandler);
            PropertyHandlerFactory.registerPropertyHandler(LyObjectList.class, propertyHandler);
            PropertyHandlerFactory.registerPropertyHandler(LyEvaluator.class, propertyHandler);
            _propertyHandler = propertyHandler;
        }
    }

    /**
     * Compile expression.
     */
	protected Object compileExpression(String expression, List<String> fieldNames)
    {
        // Create an expression object
        ParserContext        ctx = new ParserContext();
        java.io.Serializable compiledExpression = MVEL.compileExpression(expression, ctx);

        Object temp = ctx.getInputs();
        @SuppressWarnings("unchecked")
    	Map<String, Class<?>>   inputs = (Map<String, Class<?>>)temp;
    	
        Set<String>          keys = inputs.keySet();
        
        fieldNames.addAll(keys);
        return compiledExpression;
    }
    
    /**
     * Evaluate the formula
     * Return depends on the fieldInto type
     * All values are contained in the LyObject
     */
    public Object getValue(LyObject instance)
    {
        LyFormula currentFormula = _propertyHandler.getCurrentFormula();
        
        try
        {
            _propertyHandler.setCurrentFormula(this);
            
            @SuppressWarnings("rawtypes")
			Object  expValue = MVEL.executeExpression(_compiledExpression, instance, new HashMap());
            
            if (_info instanceof LyFieldInfo)
            {
                try
                {
                    LyValue value = ((LyFieldInfo)_info).buildFieldValue(expValue);
                
                    return value.getValue();
                }
                catch (Exception e)
                {
                    LyEnvironment   environment = _info.getEnvironment();
                    
                    environment.logError("BF AppEngine : can't compute value for formula \"" + _expression
                        + "\" set on fieldinfo " + _info + " for given object " + instance);
                    return null;
                }
            }
            else
                return expValue;
        }
        catch (Exception e)
        {
        	if (_info.getEnvironment().isDebugLevel(LyEnvironment.DEBUG_ALL))
        		_info.getEnvironment().writeException(e);
            return null;
        }
        finally
        {
            _propertyHandler.setCurrentFormula(currentFormula);
        }
    }

    /**
     * Evaluate the formula
     * Return depends on the fieldInto type
     * The value set must contains all the values necessary to compute the formula
     */
    public Object getValue(LyValueSet values)
    {
        LyFormula currentFormula = _propertyHandler.getCurrentFormula();
        
        try
        {
            _propertyHandler.setCurrentFormula(this);
            
            @SuppressWarnings("rawtypes")
			Object  expValue = MVEL.executeExpression(_compiledExpression, values, new HashMap());
            
            if (_info instanceof LyFieldInfo)
            {
                try
                {
                    LyValue value = ((LyFieldInfo)_info).buildFieldValue(expValue);
                
                    return value.getValue();
                }
                catch (Exception e)
                {
                    LyEnvironment   environment = _info.getEnvironment();
                    
                    environment.logError("BF AppEngine : can't compute value for formula \"" + _expression
                        + "\" set on fieldinfo " + _info + " for given values " + values);
                    return null;
                }
            }
            else
                return expValue;
        }
        catch (Exception e)
        {
            _info.getEnvironment().writeException(e);
            return null;
        }
        finally
        {
            _propertyHandler.setCurrentFormula(currentFormula);
        }
    } // end method getValue
    
} // end class LyFormula

/**
 * Internal class used to compute values of variables used in formula.
 */
class _LyPropertyHandler
    implements PropertyHandler
{
    static final String        CURRENT_FORMULA = "_currentFormula";
    
    LyThreadData<LyFormula>    _formulas;
    
    static
    {
        // Desactivates JIT optimization and avoid messages for not implementing ProducesBytecode
        DynamicOptimizer.timeSpan = 0;
    }
    
    _LyPropertyHandler()
    {
        _formulas = new LyThreadData<LyFormula>();
    }
    
    public void setCurrentFormula(LyFormula formula)
    {
        _formulas.putData(CURRENT_FORMULA, formula);
    }
    
    public LyFormula getCurrentFormula()
    {
        return _formulas.getData(CURRENT_FORMULA);
    }
    
    public Object getProperty(String name, Object contextObj, VariableResolverFactory variableFactory)
    {    
        LyFormula   formula = getCurrentFormula();
        
        return formula.getValue(name, contextObj);
    }
    
    public Object setProperty(String name, Object contextObj, VariableResolverFactory variableFactory, Object value)
    {
        // Not implemented
        LyFormula   formula = getCurrentFormula();
        LyInfo      info = formula.getInfo();
        
        throw new IllegalStateException("BF AppEngine : invalid formula on field " + info
            + ", setProperty is not allowed.");
    }
}
