/*
 * LyNotifier.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.info.notifier;

import java.util.*;
import leon.info.*;
import leon.info.location.*;
import leon.info.misc.*;


/**
 * This class implements a Notifier description. A notifier is used to send messages from a process
 * such as creation, modification, deletion. The notifier also receives messages and keep the cached
 * data consistent in memory. Examples of notifiers : JMS, Bus...
 */
public class LyNotifier extends LyInfo
{
	public static final String __VERSION =
		"$Id: LyNotifier.java 28618 2010-11-18 10:35:09Z bangiole $";

	/** To specify a full path to a property file containing the parameters of the notifier. */
	String					   _properties;

	/**
	 * To specify the java class name used to create the instance. Possible values are
	 * eu.w4.ae.notifier.jms.LyJmsNotifier (shortcut alias: jms) 
	 * The class must be found in the classPath and extend
	 * leon.notifier.LyNotifier
	 */
	LyJavaName				   _javaName;

	/** The location to which the notifier belongs to. */
	LyLocation				   _location;

	/** Set of properties containing parameters values for this notifier. */
	Properties				   _propertiesSet;

	/**
	 * Init method.
	 */
	protected void init()
	{
		super.init();

		_properties = null;
		_javaName = null;
		_location = null;
		_propertiesSet = null;
	}

    /**
     * Constructs a notifier with given properties table.
     * @param id  Identifier of the location.
     */
    protected LyNotifier(String id)
    {
        this(id, null, (String)null);
    }
    
	/**
	 * Constructs a notifier with a properties file.
	 * @param id  Identifier of the location.
	 * @param url Url of the provider. @
	 */
	public LyNotifier(String id, String javaName, String properties)
	{
		super(id);

		setJavaName(javaName);
		setProperties(properties);
	}

	/**
	 * Constructs a notifier with given properties table.
	 * @param id  Identifier of the location.
	 * @param url Url of the provider.
	 */
	public LyNotifier(String id, String javaName, Properties propertiesSet)
	{
		super(id);

		setJavaName(javaName);
		_propertiesSet = propertiesSet;
	}
	
	/**
     * Returns extra tags.
     */
    protected LyTag[] getExtraTags()
    {
        return new LyTag[] { _javaName };
    }

	/**
	 * Accessor the java class name used to create the instance.
	 * @return the java class must be found in the classPath and extend leon.notifier.LyNotifier.
	 *         Possible values are eu.w4.ae.notifier.jms.LyJmsNotifier (shortcut alias: jms)
	 */
	public String getJavaName()
	{
		return (_javaName == null ? null : _javaName.getJavaName());
	}

	/**
	 * Accessor the full path to the properties file.
	 * @return the full path to the properties file.
	 */
	public String getProperties()
	{
		return _properties;
	}

	/**
	 * Accessor for the properties set.
	 * @return the set of properties that contains contains parameters for connecting this notifier.
	 */
	public Properties getPropertiesSet()
	{
		return _propertiesSet;
	}

	/**
	 * To set the location of this notifier.
	 * @param location the location of the notifier.
	 */
	public void setLocation(LyLocation location)
	{
		_location = location;
	}

	/**
	 * Accessor to the location of this notifier.
	 * @return the location of the notifier.
	 */
	public LyLocation getLocation()
	{
		return _location;
	}
    
    /**
     * Sets the full path of the properties file.
     */
    protected void setProperties(String properties)
    {
        _properties = properties;
    }
    
    /**
     * Sets the java class name used to create the instance.
     */
    protected void setJavaName(String javaName)
    {
        _javaName = (javaName == null ? null : new LyJavaName(this, javaName));
    }
} // end class LyNotifier
