/*
 * LyFormSave.java
 */

package leon.info.save;

import leon.app.*;
import leon.info.*;
import leon.info.infointerface.*;
import leon.misc.*;
import leon.data.*;

import java.io.*;
import java.util.*;

/**
 * Save of a field controller
 */
public class LyFormSaveFile extends LyFormSave
    implements LyExtraInfoInterface
{
    public static final String  PREFIX = "_form";
    
    public static String getId(LyAction action, LyClassInfo classInfo, LyObjectId objectId)
    {
        String  saveId = PREFIX + action.getId() + "_" + classInfo.getId();
        
        if (objectId != null)
        {
            String  id = objectId.toString();
            
            id = id.replace('/', '_');
            id = id.replace('\\', '_');
            id = id.replace(LyObjectId.FIELD_SEP, '_');
            id = id.replace(LyObjectId.VALUE_SEP, '_');
            saveId += '_' + id;
        }

        return saveId;
    }
    
    public static String getFileName(LyAction action, LyClassInfo classInfo, LyObjectId objectId, String userId)
    {
        LyEnvironment   environment = action.getEnvironment();
        String          dir = environment.getFile("LY_SAVE_DIR");
        String          saveId = getId(action, classInfo, objectId);
        String          fileName = dir + File.separator + saveId;
        
        if (userId != null)
            fileName += '_' + userId;
        fileName += ".xml";
        return fileName;
    }
    
    public static LyFormSaveFile loadSave(LyAction action, LyClassInfo classInfo, LyObject object, String userId)
    {
        if (action == null)
            return null;
        
        String                       saveId = getId(action, classInfo, (object == null ? null : object.getObjectId()));
        LyApplication                application = action.getApplication();
        Vector<LyExtraInfoInterface> saves = application.loadExtraInfosFromDB(saveId, userId);

        if ((saves != null)
            && (saves.size() == 1))
        {
            LyFormSaveFile    save = (LyFormSaveFile)saves.get(0);
            
            save._action = action;
            return save;
        }
        
        String  fileName = getFileName(action, classInfo, (object == null ? null : object.getObjectId()), userId);
        File    file = new File(fileName);
        
        if (!file.exists())
            return null;
        
        try
        {
            return new LyFormSaveFile(action, userId, file.getAbsolutePath());
        }
        catch (Exception e)
        {
            LyEnvironment   environment = action.getEnvironment();
            
            environment.writeException(e);
            return null;
        }
    }
    
    /** Id of the form save. */
    String            _id;
    
    /** Action that created the save. */
    public LyAction   _action;
    
    /** File containing the form save. */
    public String     _file;
    
    /** Id of the user that owns the save. */
    String            _user;
	
    public LyFormSaveFile(LyAction action, String user, LyFormSave save)
    {
        super(save._type, save._classInfo, save._objectId, save._fieldSaves);
        
        _action = action;
        _user = user;
        _file = null;
        _datas = save._datas;
    }
    
    public LyFormSaveFile(LyAction action, String user, String path)
    {
        LyApplication   application = action.getApplication();
        LyInfoParser    infoParser = new LyInfoParser(application, path, LyInfoParser.Type.GUI, false);
        Object          root = infoParser.getRootElement();
        
        if (!(root instanceof LyFormSave))
            throw new IllegalStateException("BF AppEngine : Not a valid form save file : " + path);
        
        LyFormSave save = (LyFormSave)root;
        
        _type = save._type;
        _classInfo = save._classInfo;
        _objectId = save._objectId;
        _fieldSaves = save._fieldSaves;
        _datas = save._datas;
        
        _action = action;
        _user = user;
        _file = path;
    }

    public void free()
    {   
    }

    /**
     * Accessor to the id of this info.
     * @return the id of this info.
     */
    public final String getId()
    {
        if (_id != null)
            return _id;
        
        _id = getId(_action, _classInfo, _objectId);
        _id = _id.intern();
        return _id;
    }
    
    public final String getFile()
    {
        return _file;
    }

    public final String getUser()
    {
        return _user;
    }

    public boolean saveFile(LyXmlWriterInterface writer)
    {
        String      fileName;
        
        if (writer == null)
        	fileName = getFile();
        else
        	fileName = writer.getFile(this);
        
        if (fileName == null)
        {
            String   dir, userId = getUser();
            
            if (userId == null)
                dir = "LY_SAVE_DIR";
            else
                dir = "LY_USER_SAVE_DIR";
    
            // retrieve the name of the data directory
            LyEnvironment environment = _action.getEnvironment();
            String        dirPath = environment.getFile(dir);
            File          fileDirectory = null;

            if (dirPath != null)
                fileDirectory = new File(dirPath);

            // Try to create the directory if needed
            if ((fileDirectory != null)
                && !fileDirectory.exists())
            {
                try
                {
                    environment.logInfo("Creating directory :" + dirPath);
                    fileDirectory.mkdirs();
                }
                catch (SecurityException e)
                {
                    environment.logError("Unable to create directory :" + dirPath);
                    return false;
                }
            }
            
            fileName = getFileName(_action, _classInfo, _objectId, userId);
            setFile(fileName);
        }
        
        // Create xml file for the preference
        try
        {
            FileOutputStream 	output = new FileOutputStream(fileName);
            LyXmlPrintWriter    pw = new LyXmlPrintWriter(output);

            if (writer == null)
            	writeXml(pw, true, true);
            else
            	writer.writeXml(pw, this);
            
            pw.close();
            return true;
        }
        catch (Exception e)
        {
            LyEnvironment   environment = _action.getEnvironment();
            
            environment.writeException(e);
            return false;
        }
    }

    public void setFile(String file)
    {
        _file = file;
    }
    
    public boolean deleteFile()
    {
        String  filePath = getFile();

        if (filePath == null)
            return false;
        
        try
        {
            // Delete the file
            File    file = new File(filePath);
            
            return file.delete();
        }
        catch (SecurityException e)
        {
            LyEnvironment   environment = _action.getEnvironment();
            
            environment.writeException(e);
            return false;
        }
    }

    public void writeXml(PrintWriter pw)
    {
        String  str = saveAsString();
        
        pw.println(str);
    }

    public void writeXml(PrintWriter pw, boolean header, boolean body)
    {
        if (header)
        {
            pw.println("<?xml version='1.0' encoding='" + LyString.DEFAULT_CHARSET + "'?>");
            pw.println("<!DOCTYPE FORM_SAVE SYSTEM 'leon/conf/dtd/leon_gui.dtd'>");
        }
        
        if (body)
        	writeXml(pw);
    }
}