/*
 * LyClassLoader.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.misc;

import java.io.*;
import java.util.*;
import java.util.zip.*;


/**
 * This class defines a specific class loader that enable to load classes located on the computer's
 * disk but not defined inside the CLASSPATH properties.<br />
 * <b>How to use:</b> in order to load a specific class, create a LyClassLoader with a classpath
 * that contains its file definition. Each path of the classpath are separated with the system path
 * separator (<code>File.pathSeparatorChar</code>). The classpath can contains directories, ZIP or
 * JAR archives.<br />
 * Example :<br />
 *
 * <ul>
 * <li>Creating a specific classpath: the classes that are defined inside the C:/dev/classes
 * directory and inside the leo.zip and nardi.jar could be loaded.</li>
 * <li>
 * <pre>
   String       myClassPath     = &quot;C:/dev/classes;C:/lib/leo.zip;C:/lib/nardi.jar&quot;;
 * </pre>
 * </li>
 * <li>Creating the class loader</li>
 * <li>
 * <pre>
   LyClassLoader        loader  = new LyClassLoader(myClassPath);
 * </pre>
 * </li>
 * <li>Loading a class</li>
 * <li>
 * <pre>
   Class        classToLoad     = loader.loadClass(&quot;leon.app.LySession&quot;);
 * </pre>
 * </li>
 * </ul>
 * <b>Remark:</b> if the class was previously loaded by the primordial class loader, the loadClass
 * method returns the reference of the existing class that corresponds to the given name. Once a
 * class is loaded, it is stored and won't be reloaded anymore.
 *
 * @version  $Id: LyClassLoader.java 28985 2010-12-22 16:03:13Z jlsaouli $
 * @author   Lyria
 */
public class LyClassLoader extends ClassLoader
{

    public static final String __VERSION = "$Id: LyClassLoader.java 28985 2010-12-22 16:03:13Z jlsaouli $";

    /** The table that contains loaded classes. */
    private Hashtable<String, Class<?>> _classes;

    /**
     * The class path that contains classes to be loaded, each member of the array defines a
     * directory or an archive.
     */
    protected String[] _classpath;

    /**
     * Class constructor, initialize the classpath with the given value.
     *
     * @param  classpath  The classpath that defines were classes files are located : this is a
     *                    string that defines several path separated with the system path separator
     *                    <code> {@link File#pathSeparatorChar}</code> .
     */
    public LyClassLoader(String classpath)
    {
        _classes = new Hashtable<String, Class<?>>();
        _classpath = LyString.stringToList(classpath, File.pathSeparatorChar);
    }

    /**
     * This method search the file corresponding to the given class name from the classpath. Then it
     * opens the class file, read it and return contained data.
     *
     * @param   className  The name of the class to read.
     *
     * @return  The array of bytes that defines the class or null if the class path is null or
     *          empty.
     */
    protected byte[] readClassFromClassPath(String className)
    {
        if ((_classpath == null) || (_classpath.length == 0))
            return null;

        // creating the pattern for the file of the class :
        // each '.' is replaced with a file separator and the ".class" extension is added
        String pattern = className.replace('.', File.separatorChar) + ".class";

        // the array of data
        byte[] data = null;

        // looking for the "className".class file inside each path of the classpath
        File           file, fileClass;
        ZipEntry       entry;
        String         entryName;
        InputStream    in = null;
        ZipInputStream zin = null;

        for (int i = 0; (i < _classpath.length) && (data == null); i++)
        {
            file = new File(_classpath[i]);
            if ((file.exists()) && (file.canRead()))
                // case of a directory : read the class file
                if (file.isDirectory())
                {
                    fileClass = new File(_classpath[i] + File.separator + pattern);
                    if ((fileClass.exists()) && (fileClass.canRead()))
                        try
                        {
                            in = new FileInputStream(fileClass);
                            data = new byte[in.available()];
                            in.read(data);
                            in.close();
                        }
                        catch (FileNotFoundException ex1)
                        {
                            ex1.printStackTrace();
                        } // should never occur!!
                        catch (IOException ex2)
                        { // should never occur
                            ex2.printStackTrace();
                        }
                }
                // case of ZIP and JAR archives : read the class file inside
                // the zip file.
                else
                {
                    try
                    {
                        // check if this is a ZIP file: a format error will be
                        // thrown if it is not a valid archive.
                        new ZipFile(_classpath[i]);
                        // creating an input stream for reading
                        zin = new ZipInputStream(new FileInputStream(_classpath[i]));
                        entry = zin.getNextEntry();
                        // parsing each entry of the zip file
                        while (entry != null)
                        {
                            // for the name the unix-style file separator is
                            // replaced
                            // with the system file separator.
                            entryName = entry.getName().replace('/', File.separatorChar);
                            if (pattern.equals(entryName)) // we have the good
                            // entry
                            {
                                // Reading the entry with decompressing if necessary
                                // bytes are read in several times because of the time
                                // spent into decompressing datas.
                                // The size of the data array is increased by 1
                                // to avoid infinite loop in the while loop.
                                data = new byte[(int)entry.getSize() + 1];
                                int totalBytes = 0;
                                int packetSize = 0;

                                while (packetSize != -1)
                                {
                                    packetSize = zin.read(data, totalBytes, data.length
                                            - totalBytes);
                                    totalBytes += packetSize;
                                }

                                entry = null;
                            }
                            else
                            { // closing current entry and going to next entry
                                zin.closeEntry();
                                entry = zin.getNextEntry();
                            }
                        }

                        zin.close(); // closing the stream
                    }
                    catch (ZipException ex1)
                    {

                        /** bad format : continue */
                    }
                    catch (IOException ex2)
                    { // should never occur
                        ex2.printStackTrace();
                    }
                }
        }
        // return read data
        return data;
    } // end method readClassFromClassPath

    /**
     * Loads the given class. Invoking this method is equivalent to invoking
     * {@link #loadClass(String, boolean) <tt>loadClass(name, true)</tt>}.
     *
     * @param   className  The fully qualified name of the desired class.
     *
     * @return  The loaded Class.
     *
     * @see     #loadClass(String, boolean)
     *
     * @throws  ClassNotFoundException  if the class is not found
     */
    public Class<?> loadClass(String className) throws ClassNotFoundException
    {
        return (loadClass(className, true));
    }

    /**
     * Specific implementation of loadClass : loads the class from the file inside the classpath and
     * then creates a instance of Class. If the class has been already loaded by the primordial
     * class loader, then this reference is returned. *
     *
     * @param   className  the complete class name with the package definition (ex:
     *                     leon.misc.LyClassLoader)
     * @param   resolveIt  indicates if the class has to be resolved or not.
     *
     * @return  the found Class object.
     *
     * @throws  ClassNotFoundException  if the class is not found
     * @throws  ClassFormatError        if the retrieved class does not matches the class format.
     */
    public synchronized Class<?> loadClass(String className, boolean resolveIt) throws ClassNotFoundException
    {
        Class<?> c;
        byte[]   data;

        // Check into the cache
        c = _classes.get(className);
        if (c != null)
            return c;

        // Check with the primordial class loader
        try
        {
            c = super.findSystemClass(className);
            return c;
        }
        catch (ClassNotFoundException e)
        {
            // the class is not a system class
            // it will be loaded from its bytecode file
        }

        // try to load it from our directory
        data = readClassFromClassPath(className);
        if (data == null)
            throw new ClassNotFoundException("BF AppEngine : Could not load class : " + className);

        // defines the class
        c = defineClass(null, data, 0, data.length, null);
        if (c == null)
            throw new ClassFormatError("BF AppEngine : Wrong class format : " + className);

        if (resolveIt)
            resolveClass(c);

        // store the class
        _classes.put(className, c);

        // return the Class
        return c;
    } // end method loadClass

} // end class LyClassLoader
