/*
 * LyColors.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

/**
 * The <code>LyColor</code> class is used encapsulate colors as RGB format.
 * @version $Id: LyColor.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public final class LyColor
{
	public static final String  __VERSION = "$Id: LyColor.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The system color. */
	public final static LyColor SYSTEM = new LyColor();

	/** The <i>white</i> color. */
	public final static LyColor WHITE = new LyColor(255, 255, 255);

	/** The <i>light gray</i> color. */
	public final static LyColor LIGHT_GRAY = new LyColor(192, 192, 192);

	/** The <i>gray</i> color. */
	public final static LyColor GRAY = new LyColor(128, 128, 128);

	/** The <i>dark gray</i> color. */
	public final static LyColor DARK_GRAY = new LyColor(64, 64, 64);

	/** The <i>black</i> color. */
	public final static LyColor BLACK = new LyColor(0, 0, 0);

	/** The <i>red</i> color. */
	public final static LyColor RED = new LyColor(255, 0, 0);

	/** The <i>pink</i> color. */
	public final static LyColor PINK = new LyColor(255, 175, 175);

	/** The <i>orange</i> color. */
	public final static LyColor ORANGE = new LyColor(255, 200, 0);

	/** The <i>yellow</i> color. */
	public final static LyColor YELLOW = new LyColor(255, 255, 0);

	/** The <i>green</i> color. */
	public final static LyColor GREEN = new LyColor(0, 255, 0);

	/** The <i>magenta</i> color. */
	public final static LyColor MAGENTA = new LyColor(255, 0, 255);

	/** The <i>cyan</i> color. */
	public final static LyColor CYAN = new LyColor(0, 255, 255);

	/** The <i>blue</i> color. */
	public final static LyColor BLUE = new LyColor(0, 0, 255);

	/** No color. */
	public final static LyColor NO_COLOR = new LyColor(-1, -1, -1);


	/** The color value. */
	int						    _value;

	/**
	 * Indicates if this color is a real color or if the component using it should use a system
	 * color.
	 */
	boolean					    _systemColor;

	/**
	 * Creates an opaque RGB color with the specified red, green, and blue values in the range (0 -
	 * 255).
	 * @param r the red component
	 * @param g the green component
	 * @param b the blue component
	 * @see   #getRed()
	 * @see   #getGreen()
	 * @see   #getBlue()
	 */
	public LyColor(int r, int g, int b)
	{
		int a = 255;

		_systemColor = false;
		_value = ((a & 0xFF) << 24) | ((r & 0xFF) << 16) | ((g & 0xFF) << 8) | ((b & 0xFF) << 0);
	}

	/**
	 * Creates an opaque RGB color. Inilializes {@link #_value} with the specified value and {@link
	 * #_systemColor} to false.
	 * @param value the RGB value
	 * @see   #getRGB()
	 */
	public LyColor(int value)
	{
		_value = value;
		_systemColor = false;
	}

	/**
	 * Creates a default system color. Initializes {@link #_value} to 0 and {@link #_systemColor} to
	 * true.
	 */
	public LyColor()
	{
		_value = 0;
		_systemColor = true;
	}

	/**
	 * Gets the RGB value representing the color (Bits 16-23 are red, 8-15 are green, 0-7 are blue).
	 * @return the RGB value of the color.
	 * @see    #getRed()
	 * @see    #getGreen()
	 * @see    #getBlue()
	 */
	public int getRGB()
	{
		return _value;
	}

	/**
	 * Gets the red component in the range 0-255.
	 * @return the red component.
	 * @see    #getRGB()
	 */
	public int getRed()
	{
		return (_value >> 16) & 0xFF;
	}

	/**
	 * Gets the green component in the range 0-255.
	 * @return the green component.
	 * @see    #getRGB()
	 */
	public int getGreen()
	{
		return (_value >> 8) & 0xFF;
	}

	/**
	 * Gets the blue component in the range 0-255.
	 * @return the blue component.
	 * @see    #getRGB()
	 */
	public int getBlue()
	{
		return (_value >> 0) & 0xFF;
	}

	/**
	 * This method is used by components that need to know if they should use the system color.
	 * @return true if the asker of the color should use a system color
	 */
	public boolean useSystemColor()
	{
		return _systemColor;
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object obj)
	{
		if ((obj == null) || !(obj instanceof LyColor))
			return false;
		
		LyColor		color = (LyColor)obj;
		
		boolean	result = ((color._value == _value) && (color._systemColor == _systemColor));
		
		return result;
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode()
	{
		return (_value + (_systemColor ? 1 : 0));
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString()
	{
		return "LyColor (" + getRed() + ", " + getGreen() + ", " + getBlue() + ")";
	}
} // end class LyColor
