/*
 * LyColors.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.io.*;
import java.util.*;


/**
 * This class is used to manage color resources of an application. For each instance a mapping is
 * done between color key and color name (e.g. LY_MENU_COLOR and Gray75) using the LyResourceBundle
 * mechanism. A static Hashtable of color codes is kept in memory: the table contains color name as
 * a key and the RGB components of the color (eg Gray75 and 191 191 191).
 * @version $Id: LyColors.java 36510 2012-08-17 15:43:06Z jlsaouli $
 * @author  Lyria
 */
public class LyColors extends LyResourceBundle
{
	public static final String __VERSION = "$Id: LyColors.java 36510 2012-08-17 15:43:06Z jlsaouli $";


	/** System color key. */
	public static final String SYSTEM_COLOR = "System";

	/** Default color key. */
	public static String	   DEFAULT_COLOR;

	/** Default color codes file. */
	public static String	   COLOR_CODES_FILE = "leon/properties/color_codes.properties";


	/** Table of Colors (mapping between color name and RGB components). */
	protected Hashtable<String, LyColor>		   _colors;


	static
	{
		DEFAULT_COLOR = "LY_DEFAULT_COLOR".intern();
	}

	/**
	 * Gets the hexadecimal string representation of the given RGB components.
	 * @param  r The red component of the color.
	 * @param  g The green component of the color.
	 * @param  b The blue component of the color.
	 * @return The hexadecimal code for these RGB components.
	 */
	public static String getHexString(int r, int g, int b)
	{
		String red = Integer.toHexString(r);
		String green = Integer.toHexString(g);
		String blue = Integer.toHexString(b);

		if (red.length() == 1)
			red = "0" + red;
		if (green.length() == 1)
			green = "0" + green;
		if (blue.length() == 1)
			blue = "0" + blue;

		return new String(red + green + blue);
	}

	/**
	 * Gets the hexadecimal string representation of the given color.
	 * @param  color The color to convert.
	 * @return The hexadecimal code for the color.
	 */
	public static String getHexString(LyColor color)
	{
		if (color == null)
			return "000000";

		return getHexString(color.getRed(), color.getGreen(), color.getBlue());
	}

	/**
	 * Gets the Color representation of the given hexedecimal color code.
	 * @param  hexColor the hexadecimal code for the color.
	 * @return The java Color for the hexadecimal code.
	 */
	public static LyColor getColorFromHexString(String hexColor)
	{
		if (hexColor == null)
			return LyColor.BLACK;

		if (hexColor.charAt(0) == '#')
			hexColor = hexColor.substring(1);

		if (hexColor.length() != 6)
			return LyColor.BLACK;

		int red = Integer.parseInt(hexColor.substring(0, 2), 16);
		int green = Integer.parseInt(hexColor.substring(2, 4), 16);
		int blue = Integer.parseInt(hexColor.substring(4, 6), 16);

		return new LyColor(red, green, blue);
	}

	/**
	 * Gets theColor computed with two colors and a ratio.
	 * @param  minColor the first color
	 * @param  maxColor the second color
	 * @param  ratio    the ratio that must be used to compute the color. If ratio equals 0, the
	 *                  first color is returned. If ratio equals 1, the second color is returned. If
	 *                  ratio is between 0 and 1, use a barycentric algorithm to compute the color.
	 * @return the color computed with the two given colors and the given ratio
	 */
	public static LyColor getBarycentricColor(LyColor minColor, LyColor maxColor, float ratio)
	{
		if (ratio == 0)
			return minColor;
		else if (ratio == 1)
			return maxColor;

		int redMin = minColor.getRed();
		int greenMin = minColor.getGreen();
		int blueMin = minColor.getBlue();
		int redMax = maxColor.getRed();
		int greenMax = maxColor.getGreen();
		int blueMax = maxColor.getBlue();
		int red = redMin + (int)(ratio * (redMax - redMin));
		int green = greenMin + (int)(ratio * (greenMax - greenMin));
		int blue = blueMin + (int)(ratio * (blueMax - blueMin));

		return new LyColor(red, green, blue);
	}


	/**
	 * Constructor with a given environment.
	 * @param environment The environment of the colors bundle.
	 */
	public LyColors(LyEnvironment environment)
	{
		this(environment, null);
	}

	/**
	 * Constructor with a given environment and name.
	 * @param environment The environment of the colors bundle.
	 * @param name        The name of the colors bundle.
	 */
	public LyColors(LyEnvironment environment, String name)
	{
		this(environment, name, environment.getLocale(), environment.getSkin());
	}

	/**
	 * Constructor with a given environment and name.
	 * @param environment The environment of the colors bundle.
	 * @param name        The name of the colors bundle.
	 * @param skin        The name of the skin
	 */
	public LyColors(LyEnvironment environment, String name, String skin)
	{
		this(environment, name, environment.getLocale(), skin);
	}

	/**
	 * Constructor with a given environment, name and locale. Calls the super implementation.
	 * @param environment The environment of the strings bundle.
	 * @param name        The name of the strings bundle.
	 * @param locale      The locale used by the strings bundle.
	 */
	protected LyColors(LyEnvironment environment, String name, Locale locale)
	{
		this(environment, name, locale, environment.getSkin());
	}
	
	/**
	 * Constructor with a given environment, name and locale.
	 * @param environment The environment of the colors bundle.
	 * @param name        The name of the colors bundle.
	 * @param locale      The locale used by the colors bundle.
	 * @param skin        The name of the skin
	 */
	protected LyColors(LyEnvironment environment, String name, Locale locale, String skin)
	{
		super(environment, (name == null ? "colors" : name), locale, skin);
		
		_colors = null;
	}

	/**
	 * Builds a parent colors bundle.
	 * @return The parent colors bundle for this colors bundle.
	 */
	protected LyResourceBundle buildParentBundle()
	{
		if (_environment == null)
			return null;

		return new LyColors(_environment.getParent(), _name, _locale, _skin);
	}

	/**
	 * Searches the given color into the {@link #COLOR_CODES_FILE color_codes} properties file.
	 * @param  key the color name
	 * @return the found color, null otherwise
	 */
	protected LyColor getFromCodeFile(String key)
	{
		if ((key == null) || (key.length() == 0))
			return null;

		_environment.logTrace("BF AppEngine : loading color " + key + " in colors table");

		InputStream input = _environment.getResourceAsStream(COLOR_CODES_FILE, null);

		if (input == null)
			return null;

		String   line, colorName;
		String[] tab;
		int		 red = 0, green = 0, blue = 0;
		int		 idx;
		LyColor  color = null;

		while ((line = LyString.getLine(input)) != null)
		{
			// Skip comments and empty lines
			if ((line.length() == 0) || (line.charAt(0) == '#'))
				continue;

			idx = line.indexOf('=');
			if (idx < 0)
				continue;

			colorName = line.substring(0, idx);
			if (!key.equalsIgnoreCase(colorName))
				continue;

			line = line.substring(idx + 1);
			tab = LyString.stringToList(line, ' ');
			if (tab.length != 3)
				continue;

			try
			{
				red = Integer.parseInt(tab[0]);
				green = Integer.parseInt(tab[1]);
				blue = Integer.parseInt(tab[2]);
			}
			catch (NumberFormatException e)
			{
			}

			color = new LyColor(red, green, blue);
			break;
		}

		LyString.closeInput(input);
		return color;
	} // end method getFromCodeFile

	/**
	 * Gets the list of color keys of the {@link #COLOR_CODES_FILE color_codes} properties file.
	 * @return the keys list as a Vector
	 */
	public Vector<String> getColorCodes()
	{
		InputStream input = _environment.getResourceAsStream(COLOR_CODES_FILE, null);

		if (input == null)
			return null;

		String line, colorName;
		Vector<String> result = new Vector<String>();
		int    idx;

		while ((line = LyString.getLine(input)) != null)
		{
			// Skip comments and empty lines
			if ((line.length() == 0) || (line.charAt(0) == '#'))
				continue;

			idx = line.indexOf('=');

			if (idx < 0)
				continue;

			colorName = line.substring(0, idx);
			line = line.substring(idx + 1);
			result.addElement(colorName);
		}

		LyString.closeInput(input);

		return result;
	} // end method getColorCodes

	/**
	 * Gets the color corresponding to the given key.
	 * @param  key the color key.
	 * @return the color corresponding to the given key if it is found, null otherwise
	 */
	public LyColor getColor(String key)
	{
		if (key == null)
			return null;

		// Retrieve the value associated with the key
		String value = get(key);

		// If the value was not found, return the default color
		if (value == null)
			value = key;
		else if (value.length() == 0)
		    return (_parent == null ? null : ((LyColors)_parent).getColor(key));
		else if (value.equals("_null"))
		    return LyColor.NO_COLOR;
		else if ((value.length() > 0)
				&& (value.charAt(0) == '$'))
			return getColor(value.substring(1));
		
		if (value.equalsIgnoreCase(SYSTEM_COLOR))
			return LyColor.SYSTEM;

		LyColor color = null;

		if (_colors != null)
		{
			// Try to retrieve the color code from the Hashtable of colors
			color = _colors.get(value);
			if (color != null)
				return color;
		}

		if (value.indexOf(' ') >= 0)
		{
			// Hardcoded color
			String[] tab = LyString.stringToList(value, ' ');

			if (tab.length == 3)
			{
				try
				{
					int red = Integer.parseInt(tab[0]);
					int green = Integer.parseInt(tab[1]);
					int blue = Integer.parseInt(tab[2]);

					color = new LyColor(red, green, blue);
				}
				catch (NumberFormatException e)
				{
				}
			}
		}
		else
			// If the color was not found, load the color from the colors
			// definition file
			color = getFromCodeFile(value);

		if (color == null)
			color = LyColor.NO_COLOR;

		// Register the color.
		if (_colors == null)
			_colors = new Hashtable<String, LyColor>();
		_colors.put(value, color);

		return color;
	} // end method getColor
} // end class LyColors
