/*
 * LyDateFormatter.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.text.*;
import java.util.*;


/**
 * This class implements some static methods to format and parse dates using a particular date
 * format. Format passed to these methods must use :
 * <UL>
 *   <LI><b>dd</b> for days</li>
 *   <LI><b>MM</b> for months</li>
 *   <LI><b>yy</b> or <b>yyyy</b> for years</li>
 *   <LI><b>hh</b> for a 12-hour clock and <b>HH</b> for a 24-hour clock</li>
 *   <LI><b>mm</b> for minutes</li>
 *   <LI><b>ss</b> for seconds</li>
 * </UL>
 * <p>Date formats examples: <code>MM/dd/yyyy hh:mm:ss</code> or <code>dd/MM/yy</code> or <code>
 * dd/MM/yyyy HH:mm</code>.</p>
 */
public final class LyDateFormatter
{
	public static final String __VERSION =
		"$Id: LyDateFormatter.java 41407 2012-10-30 16:17:49Z bangiole $";

	/** Static calendar used to format dates. */
	private static Calendar    _calendar;

	static
	{
		_calendar = new GregorianCalendar();
		_calendar.set(Calendar.MILLISECOND, 0);
		// Force calendar to be a pure gregorian calendar
		((GregorianCalendar)_calendar).setGregorianChange(new Date(Long.MIN_VALUE));
	}

	/** Internal date format to save ids, or peer values */
	public static DateFormat   _internalDateFormat;

	/**
	 * Formats a date with the given format.
	 * @param  date   The date that must be formatted.
	 * @param  format The format of the result.
	 * @param  env    Current environment (used to retrieve current date locale).
	 * @param  timeZone the time zone to use
	 * @return A string that represents the formatted date.
	 */
	public static String format(Date date, DateFormat format, LyEnvironment env, String timeZone)
	{
		if (format == null)
			return null;


		String				tzId = timeZone;
		if (tzId == null)
			tzId = (env == null ? null : env.getEnv("LY_TIMEZONE"));
		TimeZone			tz = (tzId == null ? TimeZone.getDefault() : TimeZone.getTimeZone(tzId));

		format.setTimeZone(tz);


		return format.format(date);

	}
	/**
	 * Formats a date with the given format.
	 * @param  date   The date that must be formatted.
	 * @param  format The format of the result.
	 * @param  env    Current environment (used to retrieve current date locale).
	 * @return A string that represents the formatted date.
	 */
	public static String format(Date date, DateFormat format, LyEnvironment env)
	{
		return format(date, format, env, null);
	}
	/**
	 * Formats a date with the given format.
	 * @param  date   The date that must be formatted.
	 * @param  format The format of the result.
	 * @return A string that represents the formatted date.
	 */
	public static String format(Date date, String format)
	{
		return format(date, format, null);
	} // end method format

	/**
	 * Formats a date with the given format.
	 * @param  date   The date that must be formatted.
	 * @param  format The format of the result.
	 * @param  env    Current environment (used to retrieve current date locale).
	 * @return A string that represents the formatted date.
	 */
	public static String format(Date date, String format, LyEnvironment env)
	{
		if (format == null)
			return null;

		Locale				locale = (env == null ? Locale.getDefault() : env.getDateLocale());
		String				tzId = (env == null ? null : env.getEnv("LY_TIMEZONE"));
		TimeZone			tz = (tzId == null ? TimeZone.getDefault() : TimeZone.getTimeZone(tzId));
		SimpleDateFormat	sdf = new SimpleDateFormat(format, locale);

		sdf.setTimeZone(tz);
		return sdf.format(date);
	} // end method format

	/**
	 * Parses a string to retrieve a date with the given format.
	 * @param  dateString The string that must be parsed.
	 * @param  format     The format of the string.
	 * @return A Date that represents the parsed string.
	 */
	public static Date parse(String dateString, String format)
	{
		return parse(dateString, format, null, null);
	}

	/**
	 * Parses a string to retrieve a date with the given format.
	 * @param  dateString The string that must be parsed.
	 * @param  format     The format of the string.
	 * @param  env        Current environment (used to retrieve current date locale).
	 * @return A Date that represents the parsed string.
	 */
	public static Date parse(String dateString, String format, LyEnvironment env)
	{


		return parse(dateString, format, env, null);
	}

	/**
	 * Parses a string to retrieve a date with the given format.
	 * @param  dateString The string that must be parsed.
	 * @param  format     The format of the string.
	 * @param  env        Current environment (used to retrieve current date locale).
	 * @param  timeZone   The time zone to use
	 * @return A Date that represents the parsed string.
	 */
	public static Date parse(String dateString, String format, LyEnvironment env, String timeZone)
	{
		if ((dateString == null) || (dateString.trim().length() == 0))
			return null;

		if (format == null)
			return null;

		if (timeZone == null)
			timeZone = (env == null ? null : env.getEnv("LY_TIMEZONE"));

		SimpleDateFormat sdf = new SimpleDateFormat(format);

		TimeZone			tz = (timeZone == null ? TimeZone.getDefault() : TimeZone.getTimeZone(timeZone));
		Date result = null;

		sdf.setTimeZone(tz);

		try
		{
			result = sdf.parse(dateString);
		}
		catch (ParseException e)
		{
			String str;
			int    pos, length = dateString.length();
			TimeZone oldTz;
			synchronized (_calendar)
			{
				oldTz = _calendar.getTimeZone();
				_calendar.setTimeZone(tz);
				if (((pos = format.indexOf("dd")) >= 0) && (length > (pos + 1)))
				{
					str = dateString.substring(pos, pos + 2);
					_calendar.set(Calendar.DAY_OF_MONTH, Integer.parseInt(str));
				}
				else
					_calendar.set(Calendar.DAY_OF_MONTH, 1);

				if (((pos = format.indexOf("MM")) >= 0) && (length > (pos + 1)))
				{
					str = dateString.substring(pos, pos + 2);
					_calendar.set(Calendar.MONTH, Integer.parseInt(str) - 1);
				}
				else
					_calendar.set(Calendar.MONTH, 0);

				if (((pos = format.indexOf("yyyy")) >= 0) && (length > (pos + 3)))
				{
					str = dateString.substring(pos, pos + 4);
					_calendar.set(Calendar.YEAR, Integer.parseInt(str));
				}
				else if (((pos = format.indexOf("yy")) >= 0) && (length > (pos + 1)))
				{
					int year;

					str = dateString.substring(pos, pos + 2);
					year = Integer.parseInt(str);
					year += ((year < 70) ? 2000 : 1900);
					_calendar.set(Calendar.YEAR, year);
				}
				else
					_calendar.set(Calendar.YEAR, 1970);

				if (((pos = format.indexOf("HH")) >= 0) && (length > (pos + 1)))
				{
					str = dateString.substring(pos, pos + 2);
					_calendar.set(Calendar.HOUR_OF_DAY, Integer.parseInt(str));
				}
				else if (((pos = format.indexOf("hh")) >= 0) && (length > (pos + 1)))
				{
					int hour;

					str = dateString.substring(pos, pos + 2);
					hour = Integer.parseInt(str);

					String	amLabel = "AM", pmLabel = "PM";

					if (env != null)
					{
						Locale				locale = env.getDateLocale();
						DateFormatSymbols	dfs = new DateFormatSymbols(locale);
						String[]			amPmStrings = dfs.getAmPmStrings();

						amLabel = amPmStrings[0];
						pmLabel = amPmStrings[1];
					}

					if ((dateString.indexOf(amLabel) >= 0) || (dateString.indexOf(pmLabel) >= 0))
					{
						if (dateString.indexOf(pmLabel) >= 0)
							hour += 12;

						if (hour == 0)
							hour = 12;
					}

					_calendar.set(Calendar.HOUR_OF_DAY, hour);
				}
				else
					_calendar.set(Calendar.HOUR_OF_DAY, 0);

				if (((pos = format.indexOf("mm")) >= 0) && (length > (pos + 1)))
				{
					str = dateString.substring(pos, pos + 2);
					_calendar.set(Calendar.MINUTE, Integer.parseInt(str));
				}
				else
					_calendar.set(Calendar.MINUTE, 0);

				if (((pos = format.indexOf("ss")) >= 0) && (length > (pos + 1)))
				{
					str = dateString.substring(pos, pos + 2);
					_calendar.set(Calendar.SECOND, Integer.parseInt(str));
				}
				else
					_calendar.set(Calendar.SECOND, 0);

				if (((pos = format.indexOf("SSS")) >= 0) && (length > (pos + 2)))
				{
					str = dateString.substring(pos, pos + 3);
					_calendar.set(Calendar.MILLISECOND, Integer.parseInt(str));
				}
				else
					_calendar.set(Calendar.MILLISECOND, 0);

				if (((pos = format.indexOf("ZZ")) >= 0) && (length > (pos + 1)))
				{
					str = dateString.substring(pos, pos + 2);
					int hour = Integer.parseInt(str);
					str = dateString.substring(pos + 2, pos + 4);
					int min = Integer.parseInt(str);
					int ms = hour * 3600000 + min * 60000;
					_calendar.set(Calendar.ZONE_OFFSET, ms);
				}
				Date parsedDate = _calendar.getTime();
				_calendar.setTimeZone(oldTz);
				return parsedDate;
			}
		}

		return result;
	} // end method parse

	/**
	 * Parses a string to retrieve a date from the {@link Date#toString()} method. In all cases, we
	 * must have : <code>dateA</code> equals <code>fromString(dateA.toString())</code>
	 * @param  dateString The string that must be parsed.
	 * @return A Date that represents the parsed string.
	 */
	public static Date fromString(String dateString)
	{
		Date			 date = null;

		try
		{
			date = getInternalDateFormat().parse(dateString);
		}
		catch (ParseException e)
		{
			return null;
		}
		return date;
	}

	/**
	 * Parses a string to retrieve a date from the {@link Date#toString()} method. In all cases, we
	 * must have : <code>dateA</code> equals <code>fromString(dateA.toString())</code>
	 * @param  dateString The string that must be parsed.
	 * @return A Date that represents the parsed string.
	 */
	public static String toString(Date date)
	{
		if (date == null)
			return null;
		else
			return getInternalDateFormat().format(date);
	}
	/**
	 * Gets the day of the month represented by the specified Date object. The returned value is
	 * between 1 and 31 and represents day of the month that contains (or begins with) the given
	 * Date time, as interpreted in the local time zone.
	 * @param  date the date
	 * @return the day of the month represented by the specified date.
	 */
	public static int getDate(Date date)
	{
		synchronized (_calendar)
		{
			_calendar.setTime(date);
			return _calendar.get(Calendar.DAY_OF_MONTH);
		}
	}

	/**
	 * Gets the day of the week represented by the specified date.<br />
	 * Possible returned values are:
	 * <ul>
	 *   <li>0 = Sunday</li>
	 *   <li>1 = Monday</li>
	 *   <li>2 = Tuesday</li>
	 *   <li>3 = Wednesday</li>
	 *   <li>4 = Thursday</li>
	 *   <li>5 = Friday</li>
	 *   <li>6 = Saturday</li>
	 * </ul>
	 * <p>Those values represent the day of the week that contains (or begins with) the instant in
	 * time represented by the Date object, as interpreted in the local time zone.</p>
	 * @param  date the date
	 * @return the day of the week represented by this date.
	 */
	public static int getDay(Date date)
	{
		synchronized (_calendar)
		{
			_calendar.setTime(date);
			return _calendar.get(Calendar.DAY_OF_WEEK);
		}
	}

	/**
	 * Gets the hour represented by the specified Date object. The returned value is a number (0
	 * through 23) representing the hour within the day that contains (or begins with) the instant
	 * in time represented by the Date object, as interpreted in the local time zone.
	 * @param  date the date
	 * @return the hour represented by this date.
	 */
	public static int getHours(Date date)
	{
		synchronized (_calendar)
		{
			_calendar.setTime(date);
			return _calendar.get(Calendar.HOUR_OF_DAY);
		}
	}

	/**
	 * Gets the number of minutes past the hour represented by the specified date, as interpreted in
	 * the local time zone. The value returned is between 0 and 59.
	 * @param  date the date
	 * @return the number of minutes past the hour represented by this date.
	 */
	public static int getMinutes(Date date)
	{
		synchronized (_calendar)
		{
			_calendar.setTime(date);
			return _calendar.get(Calendar.MINUTE);
		}
	}

	/**
	 * Gets a number representing the month that contains (or begins with) the instant in time
	 * represented by the specified Date object. The value returned is between 0 and 11, with the
	 * value 0 representing January.
	 * @param  date the date
	 * @return the month represented by this date.
	 */
	public static int getMonth(Date date)
	{
		synchronized (_calendar)
		{
			_calendar.setTime(date);
			return _calendar.get(Calendar.MONTH);
		}
	}

	/**
	 * Gets the number of seconds past the minute represented by the specified date. The value
	 * returned is between 0 and 61. The values 60 and 61 can only occur on those Java Virtual
	 * Machines that take leap seconds into account.
	 * @param  date the date
	 * @return the number of seconds past the minute represented by this date.
	 */
	public static int getSeconds(Date date)
	{
		synchronized (_calendar)
		{
			_calendar.setTime(date);
			return _calendar.get(Calendar.SECOND);
		}
	}

	/**
	 * Gets a value that is the result of subtracting 1900 from the year that contains (or begins
	 * with) the instant in time represented by the specifie Date object, as interpreted in the
	 * local time zone.
	 * @param  date the date
	 * @return the year represented by this date, minus 1900.
	 */
	public static int getYear(Date date)
	{
		synchronized (_calendar)
		{
			_calendar.setTime(date);
			return _calendar.get(Calendar.YEAR);
		}
	}

	/**
	 * Sets the day of the month of the specified Date object to the specified value. This Date
	 * object is modified so that it represents a point in time within the specified day of the
	 * month, with the year, month, hour, minute, and second the same as before, as interpreted in
	 * the local time zone. If the date was April 30, for example, and the date is set to 31, then
	 * it will be treated as if it were on May 1, because April has only 30 days.
	 * @param  date the date to modify
	 * @param  d    the day of the month value between 1-31.
	 * @return the new date object
	 */
	public static Date setDate(Date date, int d)
	{
		synchronized (_calendar)
		{
			_calendar.setTime(date);
			_calendar.set(Calendar.DAY_OF_MONTH, d);
			return _calendar.getTime();
		}
	}

	/**
	 * Sets the hour of the Date object to the specified value. The Date object is modified so that
	 * it represents a point in time within the specified hour of the day, with the year, month,
	 * date, minute, and second the same as before, as interpreted in the local time zone.
	 * @param  date the date to modify
	 * @param  h    the hour value.
	 * @return the new date object
	 */
	public static Date setHours(Date date, int h)
	{
		synchronized (_calendar)
		{
			_calendar.setTime(date);
			_calendar.set(Calendar.HOUR_OF_DAY, h);
			return _calendar.getTime();
		}
	}

	/**
	 * Sets the minutes of the Date object to the specified value. The Date object is modified so
	 * that it represents a point in time within the specified minute of the hour, with the year,
	 * month, date, hour, and second the same as before, as interpreted in the local time zone.
	 * @param  date the date to modify
	 * @param  m    the value of the minutes.
	 * @return the new date object
	 */
	public static Date setMinutes(Date date, int m)
	{
		synchronized (_calendar)
		{
			_calendar.setTime(date);
			_calendar.set(Calendar.MINUTE, m);
			return _calendar.getTime();
		}
	}

	/**
	 * Sets the month of the date to the specified value. The Date object is modified so that it
	 * represents a point in time within the specified month, with the year, date, hour, minute, and
	 * second the same as before, as interpreted in the local time zone. If the date was October 31,
	 * for example, and the month is set to June, then the new date will be treated as if it were on
	 * July 1, because June has only 30 days.
	 * @param  date the date to modify
	 * @param  m    the month value between 0-11.
	 * @return the new date object
	 */
	public static Date setMonth(Date date, int m)
	{
		synchronized (_calendar)
		{
			_calendar.setTime(date);
			_calendar.set(Calendar.MONTH, m);
			return _calendar.getTime();
		}
	}

	/**
	 * Sets the seconds of the Date to the specified value. The Date object is modified so that it
	 * represents a point in time within the specified second of the minute, with the year, month,
	 * date, hour, and minute the same as before, as interpreted in the local time zone.
	 * @param  date the date to modify
	 * @param  s    the seconds value.
	 * @return the new date object
	 */
	public static Date setSeconds(Date date, int s)
	{
		synchronized (_calendar)
		{
			_calendar.setTime(date);
			_calendar.set(Calendar.SECOND, s);
			return _calendar.getTime();
		}
	}

	/**
	 * Sets the year of the Date object to be the specified value plus 1900. The Date object is
	 * modified so that it represents a point in time within the specified year, with the month,
	 * date, hour, minute, and second the same as before, as interpreted in the local time zone. (Of
	 * course, if the date was February 29, for example, and the year is set to a non-leap year,
	 * then the new date will be treated as if it were on March 1.)
	 * @param  date the date to modify
	 * @param  y    the year value.
	 * @return the new date object
	 */
	public static Date setYear(Date date, int y)
	{
		synchronized (_calendar)
		{
			_calendar.setTime(date);
			_calendar.set(Calendar.YEAR, y);
			return _calendar.getTime();
		}
	}

	/**
	 * Allocates a Date object that represents the given date in milliseconds.
	 * @param dateInMillis  the date as a long.
	 * @return the new date object
	 */
	public static Date buildDate(long dateInMillis, String format)
	{
		synchronized (_calendar)
		{
			_calendar.setTimeInMillis(dateInMillis);
			_calendar.set(Calendar.MILLISECOND, 0);

			if (format == null)
				return _calendar.getTime();

			if (format.indexOf("dd") < 0)
				_calendar.set(Calendar.DAY_OF_MONTH, 0);

			if (format.indexOf("MM") < 0)
				_calendar.set(Calendar.MONTH, 0);

			if (format.indexOf("yy") < 0)
				_calendar.set(Calendar.YEAR, 1970);

			if ((format.indexOf("HH") < 0)
					&& (format.indexOf("hh") < 0))
				_calendar.set(Calendar.HOUR_OF_DAY, 0);

			if (format.indexOf("mm") < 0)
				_calendar.set(Calendar.MINUTE, 0);

			if (format.indexOf("ss") < 0)
				_calendar.set(Calendar.SECOND, 0);

			return _calendar.getTime();
		}
	}

	/**
	 * Allocates a Date object and initializes it so that it represents the instant at the start of
	 * the second specified by the year, month, date, hrs, min, and sec arguments, in the local time
	 * zone.
	 * @param  year  the year minus 1900.
	 * @param  month the month between 0-11.
	 * @param  date  the day of the month between 1-31.
	 * @param  hrs   the hours between 0-23.
	 * @param  min   the minutes between 0-59.
	 * @param  sec   the seconds between 0-59.
	 * @return the new date object
	 */
	public static Date buildDate(int year, int month, int date, int hrs, int min, int sec)
	{
		synchronized (_calendar)
		{
			_calendar.set(year + 1900, month, date, hrs, min, sec);
			_calendar.set(Calendar.MILLISECOND, 0);
			return _calendar.getTime();
		}
	}

	/**
	 * Checks whether the given date string is correct according to format.
	 * @param  val        date string
	 * @param  dateFormat date format
	 * @return The corrected date string, or null if the given string is not correct.
	 */
	public static String checkDate(String val, DateFormat dateFormat)
	{
		String			  dayString;
		String			  monthString;
		String			  yearString;
		int				  day;
		int				  month;
		int				  year;
		int				  indexDM;
		int				  indexMY;
		int				  maxDayOfMonth;
		boolean			  bFrenchFormat = true;

		GregorianCalendar calendar = new GregorianCalendar();
		// Detecting date format: if first character is d: french format if M:
		// us format
		// otherwise the format is unknown and it just return the string.

		SimpleDateFormat  currentDateFormat; // = new
											 // SimpleDateFormat("dd/MM/yyyy");
		String			  formatPattern;

		currentDateFormat = (SimpleDateFormat)(dateFormat);
		formatPattern = currentDateFormat.toPattern();
		if ('d' == formatPattern.charAt(0))
			bFrenchFormat = true;
		else if ('M' == formatPattern.charAt(0))
			bFrenchFormat = false;
		else
			// Unknown format
			return val;

		// Find first separator between Day and Month
		// if it doesn't exist, the date is not valid
		indexDM = val.indexOf('/');
		if (indexDM != -1)
		{
			// Find second separator (between Month and Year)
			// if it doesn't exist, the date is not valid
			indexMY = val.indexOf('/', indexDM + 1);

			if (indexMY != -1)
			{
				// get day, month and year as strings
				if (bFrenchFormat)
				{
					dayString = val.substring(0, indexDM);
					monthString = val.substring(indexDM + 1, indexMY);
				}
				else
				{
					monthString = val.substring(0, indexDM);
					dayString = val.substring(indexDM + 1, indexMY);
				}
				yearString = val.substring(indexMY + 1);

				// Check if those strings have values
				if (dayString.length() == 0)
					return null;

				if (monthString.length() == 0)
					return null;

				if (yearString.length() == 0)
					return null;

				// Convert strings to int
				try
				{
					day = Integer.parseInt(dayString);
					month = Integer.parseInt(monthString);
					year = Integer.parseInt(yearString);
				}
				catch (NumberFormatException exc)
				{
					return null;
				}

				// Check if month value is correct, i.e. between 1 and 12
				if ((month > 12) || (month <= 0))
					return null;

				// Compute the maximum number of day for the given month
				// 31 for january, march, may, july, august, october and
				// december
				// 28 or 29 for february
				// 30 for all other months
				switch (month)
				{
				case 1:
				case 3:
				case 5:
				case 7:
				case 8:
				case 10:
				case 12:
					maxDayOfMonth = 31;
					break;

				case 2:
				{
					if (calendar.isLeapYear(year))
						maxDayOfMonth = 29;
					else
						maxDayOfMonth = 28;
				}
				break;
				default:
					maxDayOfMonth = 30;
					break;
				}

				// check if the day exists in the given month
				if (day > maxDayOfMonth)
					return null;

				// Compute year:
				// if year is less than 1 and if the length of the string
				// contening the year
				// is not two, the year is not valid
				if ((yearString.length() != 2) && (year < 1))
					return null;

				// FFT0150 - If the year exceeds 9999, it is invalid: this
				// behavior keep consitency
				// with the WEB views.
				if (yearString.length() > 4)
					return null;

				// Here, the user is using quick date format (two digit)
				// if the value is between 00 and 49 the final year will be 2000
				// -> 2049
				// if the value is between 50 and 99 the final year will be 1950
				// -> 1999
				if (yearString.length() == 2)
				{
					if (year < 50)
						year = 2000 + year;
					else
						year = 1900 + year;
				}

				calendar.set(year, month - 1, day);

				return (dateFormat.format(calendar.getTime()));
			}
			else
				return null;
		}
		else
			return null;
	} // end method checkDate

	/**
	 * Gets the day of the year represented by the specified Date object. The returned value is
	 * between 1 and 366 and represents day of the year.
	 * @param  date the date
	 * @return the day of the year represented by the specified date.
	 */
	public static int getDayOfYear(Date date)
	{
		synchronized (_calendar)
		{
			_calendar.setTime(date);
			return _calendar.get(Calendar.DAY_OF_YEAR);
		}
	}

	/**
	 * Sets the day of the year of the specified Date object to the specified value. This Date
	 * object is modified so that it represents a point in time at the specified day. *
	 * @param  date the date to modify
	 * @param  d    the day of the year value between 0-366.
	 * @return the new date object
	 */
	public static Date setDayOfYear(Date date, int d)
	{
		synchronized (_calendar)
		{
			_calendar.setTime(date);
			_calendar.set(Calendar.DAY_OF_YEAR, d);
			return _calendar.getTime();
		}
	}

	/**
	 * Internal date format to save ids, or peer values
	 * @param format
	 */
	public static synchronized void setInternalDateFormat(String format)
	{
		if (_internalDateFormat == null)
			_internalDateFormat = new SimpleDateFormat(format);
//		else
//			throw new IllegalArgumentException("Internal format is already initialized : "+_internalDateFormat);
	}
	/**
	 * Internal date format to save ids, or peer values
	 * @return the formatter
	 */
	public static synchronized DateFormat getInternalDateFormat()
	{
		if (_internalDateFormat == null)
			_internalDateFormat = new SimpleDateFormat("EEE MMM dd HH:mm:ss zzz yyyy", Locale.US); // default value for compatibility

		return _internalDateFormat;
	}

	/**
	 * Utility method that returns the age from a given date.
	 * @param date The date from which the age is requested.
	 * @return The age corresponding to the date.
	 */
	public static int getAge(Date date)
	{
		if (date == null)
			return 0;

		Calendar	currentCal = Calendar.getInstance();
		Calendar	dateCal = Calendar.getInstance();

		dateCal.setTime(date);

		int		yearDiff = currentCal.get(Calendar.YEAR) - dateCal.get(Calendar.YEAR);

		currentCal.add(Calendar.YEAR,-yearDiff);
		if(dateCal.after(currentCal))
			yearDiff--;

		return yearDiff;
	}

	/**
	 * Utility method that returns the age from a given date as a float.
	 * @param date The date from which the age is requested.
	 * @return The age corresponding to the date.
	 */
	public static double getRealAge(Date date)
	{
		if (date == null)
			return 0;

		Calendar	currentCal = Calendar.getInstance();
		Calendar	dateCal = Calendar.getInstance();

		dateCal.setTime(date);

		long		timeDiff = currentCal.getTimeInMillis() - dateCal.getTimeInMillis();
		double		dayDiff = timeDiff /(1000 * 3600 * 24);
		double		yearDiff = dayDiff / 365;

		return yearDiff;
	}
} // end class LyDateFormatter
