/*
 * LyDepends.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.util.*;


/**
 * This class is a utility class that manages dependencies between objects. An object may be
 * associated with a list of objects. These dependencies could be represented by symmetrical links
 * between these objects.
 * @version $Id: LyDependencies.java 29006 2010-12-23 17:44:46Z bmichel $
 * @author  Lyria
 */
public final class LyDependencies
{

	public static final String __VERSION =
		"$Id: LyDependencies.java 29006 2010-12-23 17:44:46Z bmichel $";

	/** List of dependencies. */
	Hashtable<Object, Object>		   _depends;

	/**
	 * Constructor with no parameter. Initializes {@link #_depends} to null.
	 */
	public LyDependencies()
	{
		_depends = null;
	}

	/**
	 * Checks if the given object is contained in the dependencies.
	 * @param  object The object to find.
	 * @return true if the given object is found, false otherwise.
	 */
	public boolean contains(Object object)
	{
		return ((_depends == null) ? false : _depends.contains(object));
	}

	/**
	 * Adds a dependency between the two given objects.
	 * @param object1 The first object of the dependency.
	 * @param object2 The second object of the dependency.
	 */
	public void addDepend(Object object1, Object object2)
	{
		addPair(object1, object2);
		addPair(object2, object1);
	}

	/**
	 * Removes all occurences of the given object in the dependencies.
	 * @param object The object that must be removed.
	 */

	public void removeObject(Object object)
	{
		if ((object == null) || (_depends == null))
			return;

		// Get depends on this object
		Object value = _depends.get(object);

		if (value == null)
			return;

		// Removereverse depends
		if (value instanceof Vector<?>)
		{
			@SuppressWarnings("unchecked")
			Vector<Object> list = (Vector<Object>)value;
			int    i, n = list.size();
			Object depend;

			for (i = 0; i < n; i++)
			{
				depend = list.elementAt(i);
				removePair(depend, object);
			}
		}
		else
			removePair(value, object);

		// Remove object
		_depends.remove(object);
	} // end method removeObject

	/**
	 * Gets objects linked to the given object. This is a COPY of the internal list, therefore this
	 * list may be freely manipulated by the application.
	 * @param  object The object whose dependencies must be retrieved.
	 * @return List of objects linked to the given object, null if object is null or if it has no
	 *         dependendy.
	 */
    public Vector<Object> getDepends(Object object)
	{
		if ((object == null) || (_depends == null))
			return null;

		Object value = _depends.get(object);

		if (value == null)
			return null;

		if (value instanceof Vector<?>)
		{
			@SuppressWarnings("unchecked")
			Vector<Object> temp = (Vector<Object>)value;
			return temp;
		}

		Vector<Object> depends = new Vector<Object>(1);

		depends.addElement(value);
		return depends;
	}

	/**
	 * Gets the keys list of dependencies defined in this object.
	 * @return The enumeration of all the objects that are contained in these dependencies.
	 */
	public Enumeration<Object> keys()
	{
		// XXX Improve : keys and values are both returned, not only keys
		return ((_depends == null) ? (new Hashtable<Object, Object>()).keys() : _depends.keys());
	}

	/**
	 * Gets the number of dependencies between objects contained in this instance of LyDependencies.
	 * @return the number of dependencies in this list
	 */
	public int getSize()
	{
		return ((_depends == null) ? -1 : _depends.size());
	}

	/**
	 * Removes a dependency between the two given objects.
	 * @param object1 The first object of the dependency.
	 * @param object2 The second object of the dependency.
	 */
	protected void removeDepend(Object object1, Object object2)
	{
		removePair(object1, object2);
		removePair(object2, object1);
	}

	/**
	 * Adds a one way dependency beetween the two specified objects.
	 * @param  object The main object of the dependency.
	 * @param  depend The object linked to the first object.
	 * @return the number of remaining dependencies for the object.
	 */
    protected int addPair(Object object, Object depend)
	{
		if (_depends == null)
			_depends = new Hashtable<Object, Object>();

		Object depends = _depends.get(object);

		if (depends == null)
		{
			_depends.put(object, depend);
			return 1;
		}

		Vector<Object> list;

		if (depends instanceof Vector<?>)
		{
			@SuppressWarnings("unchecked")
			Vector<Object> temp = (Vector<Object>)depends;
			list = temp;
		}
		else
		{
			list = new Vector<Object>(2);
			list.addElement(depends);
			_depends.put(object, list);
		}

		list.addElement(depend);
		return list.size();
	}

	/**
	 * Removes a one way dependency beetween the two specified objects.
	 * @param     object The main object of the dependency.
	 * @param     depend The object linked to the first object.
	 * @return    Returns the number of remaining dependencies for the object.
	 * @exception java.lang.IllegalStateException if there is no dependency or if the depend object
	 *                                            is not registered.
	 */
	protected int removePair(Object object, Object depend)
	{
		if (_depends == null)
			return -1;

		Vector<Object> depends = getDepends(object);

		if (depends == null) // For debug
			throw new IllegalStateException("BF AppEngine : object " + object + " is not in this list");
		if (!depends.contains(depend)) // For debug
			throw new IllegalStateException("BF AppEngine : invalid depend " + depend + " for object "
				+ object);

		int n = depends.size();

		if (n == 1)
			_depends.remove(object);
		else
			depends.removeElement(depend);

		return (n - 1);
	}
} // end class LyDependencies
