/*
 * LyEnvironment.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.io.*;
import java.lang.reflect.*;
import java.text.*;
import java.util.*;

import javax.naming.*;

import leon.*;


/**
 * This class describes an execution environment. It gives an access to all properties like
 * language, debug mode, strings, colors, fonts and also to the type of view manager used. An
 * environment is linked to a session or an application. This means that several applications may be
 * launched in the same Java Virtual Machine using different resources. It also means that each user
 * session may have its own environment: this allows multi-lingualism in servlet mode. Each time a
 * new http client connects itself on the servlet, a new session is created using a new environment
 * if needed (to change the language for example).
 *
 * <P>An environment is described by:</P>
 *
 * <UL>
 *   <LI><B>identifier (_id)</B><BR>
 *     used to retrieve resources corresponding to this environment. If it is not null we use it to
 *     search the properties files (_"id") else we search for "_app" files.</LI>
 *   <LI><B>debug level (_debugLevel)</B><BR>
 *     indicates types of traces that should be displayed.</LI>
 *   <LI><B>language (_language)</B><BR>
 *     the language used in this environment (fr, en, de...).</LI>
 *   <LI><B>parent environment (_environment)</B><BR>
 *     used if a resource was not found in the current environment.</LI>
 * </UL>
 *
 * @author  Lyria
 * @see     leon.app.LyApplication
 * @see     leon.app.LySession
 */
public final class LyEnvironment
{

	public static final String __VERSION =
		"$Id: LyEnvironment.java 41344 2012-10-25 15:45:02Z nhamon $";

	/** Debug level: None = No debug. */
	public static final int    DEBUG_NONE = 0;
	/**
	 * Debug level: All = Writes information of the following levels:
	 *
	 * <ul>
	 *   <li>{@link #DEBUG_ERROR}</li>
	 *   <li>{@link #DEBUG_WARNING}</li>
	 *   <li>{@link #DEBUG_INFO} or {@link #DEBUG_LOG}</li>
	 *   <li>{@link #DEBUG_TRACE}</li>
	 *   <li>{@link #DEBUG_DBMS}</li>
	 *   <li>{@link #DEBUG_BUS}</li>
	 *   <li>{@link #DEBUG_XML}</li>
	 * </ul>
	 */
	public static final int    DEBUG_ALL = 1;

	/**
	 * Debug level: Error = writes only occurred errors.<br />
	 * An error means that an operation did not complete successfully and/or that the current result
	 * might be inaccurate.
	 */
	public static final int    DEBUG_ERROR = 2;

	/**
	 * Debug level: Warning = writes only occurred warning.<br />
	 * A warning occurres when an error or an unexpected result was found but the current operation
	 * may continue without any trouble.
	 */
	public static final int    DEBUG_WARNING = 3;

	/**
	 * Debug level: Info = writes only information, i.e. normal operations such as saving a file.
	 */
	public static final int    DEBUG_INFO = 4;

	/** Debug level: Log = Debug level: {@link #DEBUG_INFO Info}.<br />
	 */
	public static final int    DEBUG_LOG = DEBUG_INFO;

	/**
	 * Debug level: Trace = writes only Traces.<br />
	 * A trace is an information about method calls or stacks traces.
	 */
	public static final int    DEBUG_TRACE = 5;

	/** Debug level: DBMS = Writes information about DBMS connection and requests. */
	public static final int    DEBUG_DBMS = 6;

	/**
	 * Debug level: BUS = Writes information about the logical bus (messages between applications).
	 */
	public static final int    DEBUG_BUS = 7;

	/** Debug level: XML = Write information about XML (generation of views XML files). */
	public static final int    DEBUG_XML = 8;

	/** Maximum level visible with DEBUG_ALL. */
	static final int		   MAX_DEBUG = DEBUG_XML;

	/** Debug level: PERF = Writes informations about performance. */
	public static final int    DEBUG_PERF = 9;

	/** Debug level: GRAPHICS = Shows debug information about graphics (e.g. clipping area). */
	public static final int    DEBUG_GRAPHICS = 10;

	/** Debug level: PEER = Writes information about peer layer. */
	public static final int    DEBUG_PEER = 11;

	/** Debug level: DUMP HTML = Dump HTML pages to log directory. */
	public static final int    DEBUG_DUMP_HTML = 12;

	/** Debug level: HELP FILES = Dump Help Files that the application is seeking. */
	public static final int    DEBUG_HELP_FILES = 13;

	/** Debug level: TRANSACTION = Writes information about generic hooks on transactions. */
	public static final int    DEBUG_TRANSACTION = 14;

	/** Debug level: PRINT PDF = Writes information about generation of PDF files for printing. */
	public static final int    DEBUG_PRINT_PDF = 15;

	/** Debug level: LOCKS = Writes information about session locks. */
	public static final int    DEBUG_LOCKS = 16;

	/** Windows OS. */
	public static final short  OS_WINDOWS = 1;

	/** Unix Os : so far OS_UNIX = OS_SOLARIS || OS_LINUX || OS_HPUX || OS_AIX. */
	public static final short  OS_UNIX = 2;

	/** Solaris Os. */
	public static final short  OS_SOLARIS = 3;

	/** Linux Os. */
	public static final short  OS_LINUX = 4;

	/** HP-UX Os. */
	public static final short  OS_HPUX = 5;

	/** AIX Os. */
	public static final short  OS_AIX = 6;

	/** Mac OS. */
	public static final short  OS_MACOS = 7;

	/**
	 * Internal date format to store dates in ids or peer values
	 * default is in LyDateFormatter and is EEE MMM dd HH:mm:ss zzz yyyy.
	 */
	public static String	  INTERNAL_DATE_FORMAT = null;

	/** Maximum number of bak files. */
    static final int           BAK_LOG_FILES_COUNT = 1;

	/** String representing the Operating System on which the application runs. */
	static String			   _operatingSystem;

	/** Root environment (parent of all the environments). */
	static LyEnvironment	   _rootEnvironment = null;

	/** Decimal separator. */
	static Character		   _decimalSeparator;

	/** Thousands separator. */
	static Character		   _groupingSeparator;

	/** Resource types. */
	public enum ResourceType
    {
        // Resource type : STRING.
    	STRING,
    	// Resource type : MESSAGE.
    	MESSAGE,
    	// Resource type : FILE.
    	FILE,
    	// Resource type : IMAGE.
    	IMAGE,
    	// Resource type : FONT.
    	FONT,
    	// Resource type : COLOR.
    	COLOR,
    	// Resource type : RESOURCE.
    	RESOURCE
    }

	/**
	 * Delimiters : separators that may be encountered when translating keys used to split given key
	 * in sub-keys.
	 */
	public static final String				  DELIMITERS = " ,;:=/\\\n{}";

	/** Boolean indicating if the application is in test mode (record or replay). */
	public static boolean					  TEST_MODE = false;

	/** Boolean indicating if the application is in automatic test mode (record or replay). */
	public static boolean					  TEST_MODE_AUTO = false;

	/** Date type : date (day+month+year) and time (hour+min+seconds). */
	public static final short				  TYPE_DATE_TIME = 0;

	/** Date type : date (day+month+year) and hour (hour+min). */
	public static final short				  TYPE_DATE_HOUR_MIN = 1;

	/** Date type : date (day+month+year). */
	public static final short				  TYPE_DATE = 2;

	/** Date type : time (hour+min+seconds). */
	public static final short				  TYPE_TIME = 3;

	/** Date type : hour (hour+min). */
	public static final short				  TYPE_HOUR_MIN = 4;

	/** Context key for JNDI Resources **/
	public static final String				  NAMING_CONTEXT = "NAMING_CONTEXT";

	/** Context key for JNDI temp dir **/
	public static final String				  HOSTING_TEMPSPACE = "HOSTING_TEMPSPACE";

	/** Default skin name. */
	public static String					  DEFAULT_SKIN = "aqualight";

    /** Key used for current environment. */
    protected static final String             CURRENT_ENVIRONMENT = "_currentEnvironment";

    /** Collection used to store current environments. */
    protected static LyThreadData<LyEnvironment> _currentEnvironments;

    /**
     * Gets environment for current thread.
     * @return Current environment if set, null otherwise.
     */
    public static synchronized LyEnvironment getCurrentEnvironment()
    {
        LyEnvironment   environment = null;

        if (_currentEnvironments != null)
            environment = _currentEnvironments.getData(CURRENT_ENVIRONMENT);
        if (environment == null)
            environment = _rootEnvironment;

        return environment;
    }

    /**
     * Sets environment for current thread.
     * @param environment Current environment, null to unset.
     */
    public static synchronized LyEnvironment setCurrentEnvironment(LyEnvironment environment)
    {
        if (environment == null)
        {
            if (_currentEnvironments != null)
                return _currentEnvironments.removeData(CURRENT_ENVIRONMENT);
            else
                return null;
        }
        else
        {
            if (_currentEnvironments == null)
                _currentEnvironments = new LyThreadData<LyEnvironment>();
            return _currentEnvironments.putData(CURRENT_ENVIRONMENT, environment);
        }
    }


	/** Output stream to display informations. Default is System.out. */
	public PrintStream						  out;

	/** Output stream to display errors. Default is System.err. */
	public PrintStream						  err;

	/** Reference to a parent environment. */
	protected LyEnvironment					  _environment;

	/** The id of this environment. */
	protected String						  _id;

	/**
	 * Properties table for current application context. Properties may be added using the {@link
	 * #putEnv(String, Object)} method or may be read in the ".ini" file of the application.
	 */
	protected Properties					  _properties;

	/** Debug table info (contains all debug levels that are set). */
	protected BitSet						  _debugLevel;

	/** Language used in this environment. */
	protected String						  _language;

	/** Language used for dates in this environment. */
	protected String						  _dateLanguage;

	/** Dictionary (translated Strings). */
	protected LyStrings						  _strings;

	/** Application images. */
	protected LyImages						  _images;

	/** Application colors. */
	protected LyColors						  _colors;

	/** Application fonts. */
	protected LyFonts						  _fonts;

	/** Application files (path). */
	protected LyFiles						  _files;

	/** Internal messages. */
	protected LyStrings						  _messages;

	/** Internal resources (used to customize an application). */
	protected LyResources					  _resources;

    /** Attribute values */
    protected LyResources                     _viewAttValues;

	/** List of resources created dynamically. */
	protected Vector<_LyResource>			  _dynamicResources;

	/**
	 * Attribute used to specify if this environment may build a view manager or not. For example an
	 * environment for a new http client should not create a new view manager.
	 */
	protected boolean						  _hasViewManager;

	/** Reference to the current view manager. */
	protected LyViewManagerInterface		  _viewManager;

	/** Instance of LyString used for utilities methods. */
	protected LyString						  _string;

	/** Optimization : keep a Calendar reference. */
	protected Calendar						  _calendar;

	/** Instance of LyHelpFiles to garant access to help files resources. */
	protected LyHelpFiles					  _help;

	/** Association between displayed date formats (patterns) and DateFormat. */
	protected Hashtable<String, DateFormat>   _displayDateFormatTable;

	/** Association between number formats (patterns) and NumberFormat. */
	protected Hashtable<String, NumberFormat> _decimalFormatTable;

	/** ClassLoader used to load resources. */
	protected ClassLoader					  _classLoader;

	/** The logger used to log application messages. */
	protected LyLogger						  _logger;

	/** Extra models. */
	protected Vector<String>				  _extraModels;

	/** Current demand player recorded on this transmitter. */
	protected LyPlayer						  _player;

	/** Current skin for this environment. */
	protected String						  _skin;

	/** Store the perf logger if any */
	protected LyPerfLogger					  _perfLogger;

	/**
	 * Optimization: keep dateFormat array association between TYPE_xxx (e.g. TYPE_DATE_TIME) and
	 * format. May be set by the application to specify its own date formats.
	 */
	public String[]							  _stringDateFormatTable;

	/**
	 * Optimization: keep dateFormat array association between TYPE_xxx and format. May be set by
	 * the application to specify its own date formats.
	 */
	public DateFormat[]						  _dateFormatTable;

	/** When the environment is closed, may restore former System.out. */
	public PrintStream						  _oldSystemOut;

	/** When the environment is closed, may restore former System.err. */
	public PrintStream						  _oldSystemErr;

	/** Keep Image handler interface if found (internal use only). */
	private LyImageHandlerInterface			  _imageHandler;

	public Throwable						  _throwable;

	private Boolean								_isMobile = false;

	/**
	 * Constructor with a given identifier, parent environment and indication about view manager
	 * building.
	 *
	 * @param id             The environment identifier.
	 * @param parent         The parent environment.
	 * @param hasViewManager Indicates whether this environment may build a view manager or not.
	 */
	public LyEnvironment(String id, LyEnvironment parent, boolean hasViewManager)
	{
		init();

		if ((id == null) && (parent != null))
			id = parent.getId();
		_id = id;

		if (parent == null)
			_environment = _rootEnvironment;
		else
		{
			_environment = parent;
			out = parent.out;
			err = parent.err;
			//_isMobile = parent._isMobile;
		}

		_hasViewManager = hasViewManager;

		// Gets some parts of the parent environment.
		Properties prop = _environment.getProperties();

		if (prop != null)
			_properties = (Properties)prop.clone();

		BitSet debug = _environment.getDebugLevel();

		if (debug != null)
			_debugLevel = (BitSet)debug.clone();

		// Deactivate for studio
		if (hasViewManager
			&& (id != null)
		    && !id.equals("eu.w4.ac"))
		{
			try
			{
				Class<?> c = Class.forName("leon.misc.LyImageHandler");

				_imageHandler = (LyImageHandlerInterface)c.newInstance();
			}
			catch (Exception e)
			{
				_imageHandler = null;
			}
		}

		_skin = _environment._skin;
		_language = _environment._language;
	}

	/**
	 * Constructor with a given identifier, parent environment and indications about view manager
	 * building and type of platform used.
	 *
	 * @param id             The environment identifier.
	 * @param parent         The parent environment.
	 * @param hasViewManager Indicates whether this environment may build a view manager or not.
	 * @param isMobile		 Indicates if the platform used is a mobile of not.
	 */
	public LyEnvironment(String id, LyEnvironment parent, boolean hasViewManager, boolean isMobile)
	{
		this(id, parent, hasViewManager);
		_isMobile = isMobile;
	}

	/**
	 * Constructor with a given parent environment.
	 *
	 * @param parent The parent environment.
	 */
	public LyEnvironment(LyEnvironment parent)
	{
		this(null, parent, false);
	}

	/**
	 * Internal constructor (used only for rootEnvironment). Initializes this environment id to
	 * "leon"
	 */
	protected LyEnvironment()
	{
		this((String)null, null);
	}

	/**
	 * Internal constructor (used only for rootEnvironment). Initializes this environment id to
	 * "leon"
	 */
	protected LyEnvironment(String skin, String language)
	{
		_id = "leon";
		init();

		initResources();
	}

	/**
	 * Free root environment.
	 */
	public static synchronized void freeRootEnvironment()
	{
		if (_rootEnvironment != null)
		{
			_rootEnvironment.free();
			_rootEnvironment = null;

			// Remove remaining parent bundles
			LyResourceBundle.removeBundles(null);
		}
	}

	/**
	 * Checks the current Operating System.
	 *
	 * @param  osType One of the OS_XXX defines.
	 * @return true If os characteristics are found.
	 */
	public static boolean isOperatingSystem(short osType)
	{
		if (_operatingSystem == null)
			// Initialize
			_operatingSystem = System.getProperty("os.name");

		// ** Detect OS Name **
		if (osType == LyEnvironment.OS_WINDOWS)
			return (_operatingSystem.indexOf("Windows") >= 0);

		if (osType == LyEnvironment.OS_UNIX)
			return (isOperatingSystem(LyEnvironment.OS_SOLARIS)
					|| isOperatingSystem(LyEnvironment.OS_LINUX)
					|| isOperatingSystem(LyEnvironment.OS_HPUX)
					|| isOperatingSystem(LyEnvironment.OS_AIX));

		if (osType == LyEnvironment.OS_SOLARIS)
			return ((_operatingSystem.indexOf("Solaris") >= 0)
					|| (_operatingSystem.indexOf("SunOS") >= 0));
		if (osType == LyEnvironment.OS_LINUX)
			return (_operatingSystem.indexOf("Linux") >= 0);
		if (osType == LyEnvironment.OS_HPUX)
			return (_operatingSystem.indexOf("HP-UX") >= 0);
		if (osType == LyEnvironment.OS_AIX)
			return (_operatingSystem.indexOf("AIX") >= 0);
		if (osType == LyEnvironment.OS_MACOS)
			return (_operatingSystem.indexOf("Mac OS") >= 0);

		return false;
	} // end method isOperatingSystem

	/**
	 * Sets the decimal separator for all environment.
	 *
	 * @param decimalSeparator the new decimal separator.
	 */
	public static void setDecimalSeparator(Character decimalSeparator)
	{
		LyEnvironment._decimalSeparator = decimalSeparator;
	}

	/**
	 * Set the grouping separator for all environment.
	 *
	 * @param groupingSeparator the new grouping separator.
	 */
	public static void setGroupingSeparator(Character groupingSeparator)
	{
		LyEnvironment._groupingSeparator = groupingSeparator;
	}

	/**
	 * Gets the decimal separator for all environment.
	 *
	 * @return the decimal separator, null if it was not surcharged.
	 */
	public static Character getDecimalSeparator()
	{
		return LyEnvironment._decimalSeparator;
	}

	/**
	 * Gets the grouping separator for all environment.
	 *
	 * @return the grouping separator, null if it was not surcharged.
	 */
	public static Character getGroupingSeparator()
	{
		return LyEnvironment._groupingSeparator;
	}

	/**
	 * Initialization method. This method is called by constructors to initialize class attributes.
	 */
	protected void init()
	{
		synchronized (this)
		{
			if ((_rootEnvironment == null)
					&& ((_id == null) || !_id.equals("leon")))
				_rootEnvironment = new LyEnvironment();
		}

		_environment = null;
		_properties = null;
		_debugLevel = null;
		_language = null;
		_skin = DEFAULT_SKIN;
		_dateLanguage = null;
		_strings = null;
		_images = null;
		_colors = null;
		_fonts = null;
		_files = null;
		_messages = null;
		_resources = null;
        _viewAttValues = null;
		_dynamicResources = new Vector<_LyResource>(1);
		_hasViewManager = false;
		_viewManager = null;
		_string = null;
		_calendar = null;
		_help = new LyHelpFiles(this);
		_displayDateFormatTable = null;
		_decimalFormatTable = null;
		_classLoader = null;
		_logger = null;
		_extraModels = null;
		out = System.out;
		err = System.err;
	} // end method init

	/**
	 * Gets the properties table for current application context.
	 *
	 * @return The properties table for current application context.
	 * @see    LyEnvironment#setDefaultProperties()
	 */
	protected Properties getProperties()
	{
		return _properties;
	}

	/**
	 * This method return the absolute path of the OS home directory.
	 *
	 * @return the absolute path of the OS home directory
	 */
	protected String getHomePath()
	{
		return System.getProperty("user.home");
	}

	/**
	 * This method sets default properties values for the environment. For instance LY_PROJECT_FILE
	 * should be $LY_APP_DIR$/info/project.xml. It prevents from writing some basic resources in the
	 * configuration file of the application since they have a default value (such as LY_LEON_DIR,
	 * LY_APP_DIR, LY_PROJECT_FILE...).
	 *
	 * @see LyEnvironment#getProperties()
	 */
	protected void setDefaultProperties()
	{
		// ** Sets default values **

		putEnv("LY_TEMP_DIR", getHomePath());

		if (getEnv("LY_LEON_DIR") == null)
			putEnv("LY_LEON_DIR", System.getProperty("user.dir"));

		if (getEnv("LY_APP_DIR") == null)
			putEnv("LY_APP_DIR", System.getProperty("user.dir"));

		if (getEnv("LY_PROJECT_FILE") == null)
			putEnv("LY_PROJECT_FILE", getString().expand("$LY_APP_DIR$/info/project.xml"));

		if (getEnv("LY_DATA_DIR") == null)
			putEnv("LY_DATA_DIR", getString().expand("$LY_APP_DIR$/data"));
		else
			// LY_DATA_DIR may need to be expanded after LY_APP_DIR has been set
			putEnv("LY_DATA_DIR", getString().expand(getEnv("LY_DATA_DIR")));

		if (getEnv("LY_LEON_DOC") == null)
			putEnv("LY_LEON_DOC", getString().expand("$LY_LEON_DIR$"));

		if (getEnv("LY_APP_DOC") == null)
			putEnv("LY_APP_DOC", getString().expand("$LY_APP_DIR$"));

		if (getEnv("LY_LEON_DOC_DIR") == null)
			putEnv("LY_LEON_DOC_DIR", getString().expand("$LY_LEON_DOC$"));

		if (getEnv("LY_APP_DOC_DIR") == null)
			putEnv("LY_APP_DOC_DIR", getString().expand("$LY_APP_DOC$"));

		if (getEnv("LY_LANGUAGE") == null)
			putEnv("LY_LANGUAGE", Locale.getDefault().getLanguage());

		if (getEnv("LY_DATE_LANGUAGE") == null)
			putEnv("LY_DATE_LANGUAGE", getEnv("LY_LANGUAGE"));

		if (getEnv("LY_TEST_DIR") == null)
			putEnv("LY_TEST_DIR", getString().expand("$LY_APP_DIR$/test"));

		if (testEnv("LY_RECORD") || testEnv("LY_REPLAY") || TEST_MODE_AUTO)
			TEST_MODE = true;

		if (getEnv("LY_SKIN") == null)
			putEnv("LY_SKIN", DEFAULT_SKIN);
	} // end method setDefaultProperties

	/**
	 * This method is called after loading ".ini" files (environment settings). It ends the data
	 * initialization of this environment.
	 */
	protected void endInitialize()
	{
		setDefaultProperties();

        // First initialize output and error files.
        initOutFiles();

		// Check app. defined language
		String lang = getEnv("LY_LANGUAGE");

		// If we have a skin, load the corresponding bundles.
		String skin = getEnv("LY_SKIN");

		if (((skin != null) && (skin.length() > 0)) || ((lang != null) && (lang.length() > 0)))
			setSkinAndLanguage(skin, lang);

		String dateLang = getEnv("LY_DATE_LANGUAGE");

		if (dateLang != null)
			setDateLanguage(dateLang);
		else if (lang != null)
			setDateLanguage(lang);

		// force default date format if any
		if (INTERNAL_DATE_FORMAT != null)
			LyDateFormatter.setInternalDateFormat(INTERNAL_DATE_FORMAT);

		// Check if _LY_SAVE_DATA is set
	    if (getEnv("_LY_SAVE_DATA") == null)
	        putEnv("_LY_SAVE_DATA", "true");

		if (isDebugLevel(LyEnvironment.DEBUG_PERF))
		{
			String	perfFile = getEnv("LY_PERF_FILE");
			if (perfFile != null)
				_perfLogger = new LyPerfLogger(getString().expand(perfFile));
		}

		// Get naming context if specified
		String contextKey = getEnv(NAMING_CONTEXT);

		if (!NAMING_CONTEXT.equals(contextKey) && (contextKey != null))
			initializeFromJNDI(contextKey);

        // Init static attributes if needed
        initStaticAttributes();
	} // end method endInitialize

	/**
     * Iniitalizes outuput and error files.
     */
    protected void initOutFiles()
    {
        // Check output and error files.
        String           outputFile = getEnv("LY_OUTPUT_FILE");
        PrintStream      outPrint = null;
        FileOutputStream outFile;
        File             file, bakFile;
        int              nBakLogFiles = BAK_LOG_FILES_COUNT;

        if (outputFile != null)
            outputFile = getString().expand(outputFile);

        if ((outputFile != null)
            && (!outputFile.equals("NULL")))
        {
            // Set the standard and error outputs
            try
            {
                int     i;

                for (i = nBakLogFiles - 1; i >= 0; i--)
                {
                    if (i == 0)
                        file =  new File(outputFile);
                    else if (i == 1)
                        file = new File(outputFile + ".bak");
                    else
                        file = new File(outputFile + ".bak" + i);

                    if (file.exists())
                    {
                        // Move existing filename to filename.bak
                        if (i == 0)
                            bakFile = new File(outputFile + ".bak");
                        else
                            bakFile = new File(outputFile + ".bak" + (i + 1));

                        if (bakFile.exists())
                            bakFile.delete();

                        if (file.canWrite())
                            file.renameTo(bakFile);
                    }
                }

                outFile = new FileOutputStream(outputFile, true);

                outPrint = new PrintStream(outFile);
                setOutputStream(outPrint);
            }
            catch (IOException e)
            {
                writeException(e);
            }
        }

        String           errorFile = getEnv("LY_ERROR_FILE");
        PrintStream      errPrint = null;
        FileOutputStream errFile;

        if (errorFile != null)
            errorFile = getString().expand(errorFile);

        if ((errorFile != null)
            && ((outputFile == null)
                || (!errorFile.equals(outputFile)))
            && (!errorFile.equals("NULL")))
        {
            // Set the standard and error outputs
            try
            {
                int     i;

                for (i = nBakLogFiles - 1; i >= 0; i--)
                {
                    if (i == 0)
                        file =  new File(errorFile);
                    else if  (i == 1)
                        file = new File(errorFile + ".bak");
                    else
                        file = new File(errorFile + ".bak" + i);

                    if (file.exists())
                    {
                        // Move existing filename to filename.bak
                        if (i == 0)
                            bakFile = new File(errorFile + ".bak");
                        else
                            bakFile = new File(errorFile + ".bak" + (i + 1));

                        if (bakFile.exists())
                            bakFile.delete();

                        if (file.canWrite())
                            file.renameTo(bakFile);
                    }
                }

                errFile = new FileOutputStream(errorFile, true);

                errPrint = new PrintStream(errFile);
                setErrorStream(errPrint);

            }
            catch (IOException e)
            {
                writeException(e);
            }
        }
        // output is defined, but not error or error is same as output
        else if ((outPrint != null)
                && ((errorFile == null) || (errorFile.equals(outputFile))))
            setErrorStream(outPrint);

        // Set the standard and error outputs
        try
        {
            String systemOutFile = getEnv("_LY_SYSTEM_OUTPUT_FILE");
            String systemErrFile = getEnv("_LY_SYSTEM_ERROR_FILE");

            if (systemOutFile != null)
                systemOutFile = getString().expand(systemOutFile);

            if (systemErrFile != null)
                systemErrFile = getString().expand(systemErrFile);

            if ((systemOutFile != null)
                    && ((outputFile == null) || (!systemOutFile.equals(outputFile)))
                    && (!systemOutFile.equals("NULL")))
            {
                FileOutputStream systemOutStream = new FileOutputStream(systemOutFile, true);

                PrintStream      systemOutPrint = null;

                systemOutPrint = new PrintStream(systemOutStream);

                setSystemOut(systemOutPrint, false);
            }
            else if ((outPrint != null)
                    && ((systemOutFile == null) || (systemOutFile.equals(outputFile))))
                setSystemOut(outPrint, false);

            if ((systemErrFile != null)
                    && ((outputFile == null) || (!systemErrFile.equals(outputFile)))
                    && (!systemErrFile.equals("NULL")))
            {
                FileOutputStream systemErrStream = new FileOutputStream(systemErrFile, true);
                PrintStream      systemErrPrint = new PrintStream(systemErrStream);

                setSystemErr(systemErrPrint, false);
            }
            else if ((outPrint != null)
                    && ((systemErrFile == null) || (systemErrFile.equals(outputFile))))
                setSystemErr(outPrint, false);
        }
        catch (IOException e)
        {
            writeException(e);
        }
    }

	/**
	 * This method is called after loading ".ini" files (environment settings). It ends the data
	 * initialization of this environment and try to find some values from JEE JNDI Context.
	 * @param name of initial context to lookup, or null/empty for InitialContext
	 */
	protected void initializeFromJNDI(String contextName)
	{
		try
		{
			Context context;
			if ((contextName != null) && (contextName.length() > 0))
				context = (Context) new InitialContext().lookup(contextName);
			else
				context = new InitialContext();
			// Get root temp folder
			String tempspace = getEnv(HOSTING_TEMPSPACE);
			String rootTempDirectory = (tempspace == null ? null : (String)context.lookup(tempspace));

			if (rootTempDirectory != null)
			{
				putFile("LY_TEMP_DOC_DIR", rootTempDirectory);
				putFile("LY_TEMP_DOC", rootTempDirectory);

			}

		}
		catch (Exception e)
		{
			logWarning(e.getMessage());
		}
	}
    /**
     * Sets static attributes found in ini file.
     */
    protected void initStaticAttributes()
    {
        Enumeration<Object> keys = _properties.keys();
        String              key, className, varName;
        int                 p;
        Class<?>            lyClass = null;
        Field               field;
        String              value;
        ClassLoader         classLoader = getClassLoader();
        boolean             accessible;

        while (keys.hasMoreElements())
        {
            key = (String)keys.nextElement();
            p = key.lastIndexOf('.');
            if  (p > 0)
            {
                className = key.substring(0, p);
                varName = key.substring(p + 1);

                try
                {
                    if (classLoader == null)
                        lyClass = Class.forName(className);
                    else
                        lyClass = classLoader.loadClass(className);

                    field = lyClass.getDeclaredField(varName);
                    if ((field != null)
                        && Modifier.isStatic(field.getModifiers())
                        && !Modifier.isPrivate(field.getModifiers())
                        && !Modifier.isFinal(field.getModifiers()))
                    {
                        accessible = field.isAccessible();
                        if (!accessible)
                            field.setAccessible(true);

                        value = _properties.getProperty(key);
                        if  (!field.getType().isPrimitive())
                            field.set(null, value);
                        else if (field.getType() == boolean.class)
                            field.setBoolean(null, LyTools.booleanFromString(value));
                        else if (field.getType() == int.class)
                            field.setInt(null, LyTools.intFromString(value));
                        else if (field.getType() == short.class)
                            field.setShort(null, LyTools.shortFromString(value));
                        else if (field.getType() == long.class)
                            field.setLong(null, LyTools.longFromString(value));
                        else if (field.getType() == double.class)
                            field.setDouble(null, LyTools.doubleFromString(value));
                        else if (field.getType() == float.class)
                            field.setFloat(null, LyTools.floatFromString(value));
                        else if (field.getType() == char.class && value != null)
                            field.setChar(null, value.charAt(0));

                        if (!accessible)
                            field.setAccessible(false);
                    }
                 }
                catch (Exception e)
                {
                	writeException(e);
                }
            }
        }
    }

	/**
	 * Frees all resources used by this environment.
	 */
	public void free()
	{
		// Free all bundles associated to this environment
		LyResourceBundle.removeBundles(this);

		_id = null;
		_properties = null;
		_debugLevel = null;
		_language = null;
		_dateLanguage = null;
		_strings = null;
		_images = null;
		_colors = null;
		_fonts = null;
		_files = null;
		_messages = null;
		_resources = null;
        _viewAttValues = null;
		_dynamicResources = null;
		setViewManager(null);
        _hasViewManager = false;
		_string = null;
		_calendar = null;
		_help = null;
		_displayDateFormatTable = null;
		_decimalFormatTable = null;
		_classLoader = null;
		_logger = null;
		_extraModels = null;

		if (_perfLogger != null)
		{
			_perfLogger.close();
			_perfLogger = null;
		}

		// Closing output streams if needed
		try
		{
			if (((_environment == null) || (out != _environment.out)) && (out != null))
			{
				if ((out == System.out) && (_oldSystemOut != null))
					setSystemOut(_oldSystemOut, true);

				if ((out != System.out) && (out != err))
					out.close();
			}
		}
		catch (Exception e)
		{
			writeException(e);
		}

		out = null;
		_oldSystemOut = null;

		try
		{
			if (((_environment == null) || (err != _environment.err)) && (err != null))
			{
				if ((err == System.err) && (_oldSystemErr != null))
					setSystemErr(_oldSystemErr, true);

				if (err != System.err)
					err.close();
			}
		}
		catch (Exception e)
		{
			writeException(e);
		}

		err = null;
		_oldSystemErr = null;
		_environment = null; // Must be at the end (after closing out and err)
		// at least !)

	} // end method free

	/**
	 * Checks if this environment is the root environment.
	 *
	 * @return true if this environment is the root environment, false otherwise
	 */
	public boolean isRoot()
	{
		return (_environment == null);
	}

	/**
	 * Sets the stream used to display information.
	 *
	 * @param output The new output stream
	 */
	public void setOutputStream(PrintStream output)
	{
		if (output == null)
			return;

		if (((_environment == null) || (out != _environment.out)) && (out != System.out)
				&& (out != System.err) && (out != err))
			out.close();
		out = output;

		if ((_logger == null) || ((_environment != null) && (_logger == _environment.getLogger())))
			_logger = buildLogger(true);
	}

	/**
	 * Set the stream used to display errors.
	 *
	 * @param error The new error stream
	 */
	public void setErrorStream(PrintStream error)
	{
		if (error == null)
			return;

		if (((_environment == null) || (err != _environment.err)) && (err != System.out)
				&& (err != System.err) && (err != out))
			err.close();
		err = error;

		if ((_logger == null) || ((_environment != null) && (_logger == _environment.getLogger())))
			_logger = buildLogger(true);
	}

	/**
	 * Adds a resource with the given type, key and value in this environment.
	 *
	 * @param type         Type of the resource.
	 * @param key          Key of the resource.
	 * @param value        Value of the resource.
	 * @param persistent   Indicates that resource survives to environment reload (language or skin changes).
	 */
	public void putResource(ResourceType type, String key, String value, boolean persistent)
	{
		switch (type)
		{
		case STRING:
			putString(key, value);
			break;
		case MESSAGE:
			putMessage(key, value);
			break;
		case FILE:
			putFile(key, value);
			break;
		case IMAGE:
			putImage(key, value);
			break;
		case FONT:
			putFont(key, value);
			break;
		case COLOR:
			putColor(key, value);
			break;
		case RESOURCE:
			putResource(key, value);
			break;
		default:
			throw new IllegalArgumentException("BF AppEngine : unknown resource type, " + type);
		}

	    if (persistent)
	        _dynamicResources.addElement(new _LyResource(type, key, value));
	} // end method putResource

	/**
     * Adds a non-persistent resource with the given type, key and value in this environment.
     */
	public final void putResource(ResourceType type, String key, String value)
	{
	    putResource(type, key, value, false);
	}

	/**
     * Removes a resource with the given type and key from this environment.
     *
     * @param type  Type of the removed resource.
     * @param key   Key of the removed resource.
     */
    public void removeResource(ResourceType type, String key)
    {
        switch (type)
        {
        case STRING:
            if (_strings != null)
                _strings.remove(key);
            break;
        case MESSAGE:
            if (_messages != null)
                _messages.remove(key);
            break;
        case FILE:
            if (_files != null)
                _files.remove(key);
            break;
        case IMAGE:
            if (_images != null)
                _images.remove(key);
            break;
        case FONT:
            if (_fonts != null)
                _fonts.remove(key);
            break;
        case COLOR:
            if (_colors != null)
                _colors.remove(key);
            break;
        case RESOURCE:
            if (_resources != null)
                _resources.remove(key);
            break;
        default:
            throw new IllegalArgumentException("BF AppEngine : unknown resource type, " + type);
        }
    } // end method removeResource

    /**
     * Check the existence of a resource in the resources bundles
     * @param type type of the resource
     * @param key the key
     * @param checkAll if true, we check the existence in the application and in the AE properties. If false we only
     *                 look in the application
     * @return true if we found the resource
     */
    public boolean checkResource(ResourceType type, String key, boolean checkAll)
    {
    	switch (type)
		{
		case STRING:
			if (_strings != null)
				return _strings.checkKey(key, checkAll);
			break;
		case MESSAGE:
			if (_messages != null)
				return _messages.checkKey(key, checkAll);
			break;
		case FILE:
			if (_files != null)
				return _files.checkKey(key, checkAll);
			break;
		case IMAGE:
			if (_images != null)
				return _images.checkKey(key, checkAll);
			break;
		case FONT:
			if (_fonts != null)
				return _fonts.checkKey(key, checkAll);
			break;
		case COLOR:
			if (_colors != null)
				return _colors.checkKey(key, checkAll);
			break;
		case RESOURCE:
			if (_resources != null)
				return _resources.checkKey(key, checkAll);
			break;
		default:
			throw new IllegalArgumentException("BF AppEngine : unknown resource type, " + type);
		}

    	boolean goUp = true;
    	if (!checkAll )
    	{
    		if (_environment == null)
    			goUp = false;
    		else if (_environment._environment == null) // parent env is root env
    			goUp = false;
    	}
		if (goUp && (_environment != null))
			return _environment.checkResource(type, key, checkAll);

		return false;
    }
	/**
	 * Gets the resource identified with the given key and type.
	 *
	 * @param  type Type of the resource.
	 * @param  key  Key of the resource.
	 * @return Value of the resource if it exists, null otherwise.
	 */
	public String getResource(ResourceType type, String key)
	{
		switch (type)
		{
		case STRING:
			if (_strings != null)
				return _strings.get(key);
			break;
		case MESSAGE:
			if (_messages != null)
				return _messages.get(key);
			break;
		case FILE:
			if (_files != null)
				return _files.get(key);
			break;
		case IMAGE:
			if (_images != null)
				return _images.get(key);
			break;
		case FONT:
			if (_fonts != null)
				return _fonts.get(key);
			break;
		case COLOR:
			if (_colors != null)
				return _colors.get(key);
			break;
		case RESOURCE:
			if (_resources != null)
				return _resources.get(key);
			break;
		default:
			throw new IllegalArgumentException("BF AppEngine : unknown resource type, " + type);
		}

		if (_environment != null)
			return _environment.getResource(type, key);

		return null;
	} // end method getResource

	/**
	 * Gets resources identified with the given type.
	 *
	 * @param  type Type of the resources.
	 * @return Value of the resources if it exists, null otherwise.
	 */
	public HashMap <String,String> getResources(ResourceType type)
	{
		switch (type)
		{
		case STRING:
			if (_strings != null)
			{
				return getResourceElements(_strings);
			}
			break;
		case MESSAGE:
			if (_messages != null)
			{
				return getResourceElements(_messages);
			}
			break;
		case FILE:
			if (_files != null)
			{
				return getResourceElements(_files);
			}
			break;
		case IMAGE:
			if (_images != null)
			{
				return getResourceElements(_images);
			}
			break;
		case FONT:
			if (_fonts != null)
			{
				return getResourceElements(_fonts);
			}
			break;
		case COLOR:
			if (_colors != null)
			{
				return getResourceElements(_colors);
			}
			break;
		case RESOURCE:
			if (_resources != null)
			{
				return getResourceElements(_resources);
			}
			break;
		default:
			throw new IllegalArgumentException("BF AppEngine : unknown resource type, " + type);
		}

		if (_environment != null)
			return _environment.getResources(type);

		return null;
	} // end method getResource


	private HashMap<String, String> getResourceElements(LyResourceBundle source)
	{
		HashMap<String, String> result = new HashMap<String, String>();
		Vector<String> keys = source.getKeyList(false);
		String key = null;
		String value = null;

		for(int i = 0; i < keys.size(); i++)
		{
			key = keys.get(i);
			value = source.get(key);
			result.put(key, value);
		}
		return result;
	}

	/**
	 * Translates the given key to a String that may be shown to the user. Strings are found in the
	 * strings_XXX.properties files (see LyResourceBundle for more details). This method only call
	 * {@link #translate(String, boolean) translate(key, true)}
	 *
	 * @param  key The key to translate.
	 * @return The translated key if found in local dictionary, the key is returned 'as is'
	 *         otherwise.
	 * @see    leon.misc.LyStrings
	 * @see    #translate(String, boolean)
	 */
	public String translate(String key)
	{
		return translate(key, true);
	}

	/**
	 * Translates the given key to a String that may be shown to the user. Strings are found in the
	 * strings_XXX.properties files (see LyResourceBundle for more details).
	 *
	 * @param  key     the key to translate
	 * @param  subKeys indicates if a sub string should be analyzed if the given key can't be
	 *                 translated.
	 * @return The translated key if found in local dictionary, the key is returned 'as is'
	 *         otherwise.
	 */
	protected String translate(String key, boolean subKeys)
	{
		if (key == null)
			return null;

		if (key.startsWith("$$"))
			return key.substring(2);

		// Check if we have a valid key before searches
		if (!LyResourceBundle.UPPER_CASE_KEYS_ONLY
		    || key.toUpperCase().equals(key))
		{
    		// Look in strings table
    		String result = getString(key);

    		if (result != key) // found
    			return result;

    		// Second chance : look in messages table
    		result = getMessage(key);

    	    if (result != key) // found
    	        return result;
		}

		if (subKeys)
		{
			// Look for sub-keys
			StringBuffer buffer = new StringBuffer();
			String		 subKey;
			int			 i, l = key.length();
			int			 idx;
			char		 c;

			for (i = 0, idx = 0; i < l; i++)
			{
				c = key.charAt(i);
				if (DELIMITERS.indexOf(c) < 0)
					continue;

				if (idx < i)
				{
					subKey = key.substring(idx, i);
					if (subKey.startsWith("$$"))
						buffer.append(subKey.substring(2));
					else
						buffer.append(translate(subKey, false));
				}

				buffer.append(c);
				idx = i + 1;
			}

			if (idx > 0)
			{
				if (idx < l)
				{
					subKey = key.substring(idx, l);
					if (subKey.startsWith("$$"))
						buffer.append(subKey.substring(2));
					else
						buffer.append(translate(subKey, false));
				}

				return buffer.toString();
			}
		}

		if (key.startsWith("$$"))
			return key.substring(2);

		return key;
	} // end method translate

	/**
	 * Convenient routine to get a formated string without argument.
	 *
	 * @param  key the string to find
	 * @return the found message
	 * @see    LyEnvironment#getString(String, Object[])
	 */
	public String getString(String key)
	{
		return getString(key, (Object[])null);
	}

	/**
	 * Convenient routine to get a formated string with one argument.
	 *
	 * @param  key      the string key
	 * @param  argument the message argument
	 * @return the found message
	 * @see    LyEnvironment#getString(String, Object[])
	 */
	public String getString(String key, Object argument)
	{
		Object[] arguments = new Object[1];

		arguments[0] = argument;
		return getString(key, arguments);
	}

	/**
	 * Constructs a string from a key and an array of arguments. Used instead of the translate
	 * method when arguments are needed.
	 *
	 * @param  key       the string key.
	 * @param  arguments the array of arguments used to build the string.
	 * @return the string formated with the arguments. If not found, the key is returned followed by
	 *         the string representation of the arguments.
	 * @see    leon.misc.LyStrings#getString(String,Object[])
	 */
	public String getString(String key, Object[] arguments)
	{
		String result = null;

		if (_strings != null)
		{
			result = _strings.getString(key, arguments);

		    if (result == null)
		        result = key;
		}
		else if (_environment != null)
		    result = _environment.getString(key, arguments);

		return result;
	}

	/**
	 * Gets the mnemonic for the string corresponding to the given key.
	 *
	 * @param  key The key of the string.
	 * @return The mnemonic for the string corresponding to the given key or null if not found.
	 */
	public Character getMnemonic(String key)
	{
		Character result = null;

		if (_strings != null)
			result = _strings.getMnemonic(key);
		else if (_environment != null)
			result = _environment.getMnemonic(key);

		return result;
	}

	/**
	 * Adds a string to the table of specific strings defined for one session. Allows adding a
	 * specific string during the session execution. If the key defines an existing string in the
	 * session environment, it is overidden.
	 *
	 * @param key    The key that defines the string.
	 * @param string The translated string as it appears in views.
	 */
	public void putString(String key, String string)
	{
		if (_strings == null)
			_strings = new LyStrings(this);

		_strings.put(key, string);
	}

	/**
	 * Convenient routine to get a formated message without argument.
	 *
	 * @param  key the message key
	 * @return The message formated. If not found, the key is returned.
	 * @see    LyEnvironment#getMessage(String, Object[])
	 */
	public String getMessage(String key)
	{
		return getMessage(key, (Object[])null);
	}

	/**
	 * Convenient routine to get a formated message with only one argument.
	 *
	 * @param  key      the message key
	 * @param  argument the message argument
	 * @return The message formated with the argument. If not found, the key is returned followed by
	 *         the string representation of the arguments.
	 * @see    LyEnvironment#getMessage(String, Object[])
	 */
	public String getMessage(String key, Object argument)
	{
		Object[] arguments = new Object[1];

		arguments[0] = argument;
		return getMessage(key, arguments);
	}

	/**
	 * Convenient routine to get a formated message with two arguments.
	 *
	 * @param  key       the message key
	 * @param  argument1 the first argument
	 * @param  argument2 the second argument
	 * @return The message formated with the arguments. If not found, the key is returned followed
	 *         by the string representation of the arguments.
	 * @see    LyEnvironment#getMessage(String, Object[])
	 */
	public String getMessage(String key, Object argument1, Object argument2)
	{
		Object[] arguments = new Object[2];

		arguments[0] = argument1;
		arguments[1] = argument2;
		return getMessage(key, arguments);
	}

	/**
	 * Convenient routine to get a formated message with three arguments.
	 *
	 * @param  key       the message key
	 * @param  argument1 the first argument
	 * @param  argument2 the second argument
	 * @param  argument3 the third argument
	 * @return The message formated with the arguments. If not found, the key is returned followed
	 *         by the string representation of the arguments.
	 * @see    LyEnvironment#getMessage(String, Object[])
	 */
	public String getMessage(String key, Object argument1, Object argument2, Object argument3)
	{
		Object[] arguments = new Object[3];

		arguments[0] = argument1;
		arguments[1] = argument2;
		arguments[2] = argument3;

		return getMessage(key, arguments);
	}

	/**
	 * Constructs a message from a key and an array of arguments. Strings are found in the
	 * messages_XXX.properties files (see LyResourceBundle for more details).
	 *
	 * @param  key       The key of the message.
	 * @param  arguments The array of arguments used to construct the message.
	 * @return The message formated with the arguments. If not found, the key is returned followed
	 *         by the string representation of the arguments.
	 * @see    leon.misc.LyStrings#getString(String,Object[])
	 * @see    leon.misc.LyStrings
	 */
	public String getMessage(String key, Object[] arguments)
	{
		String result = null;

		if (_messages != null)
		{
	        if (arguments != null)
	        {
	            int i, n;

	            n = arguments.length;
	            for (i = 0; i < n; i++)
	            {
	                if (arguments[i] instanceof String)
	                    arguments[i] = translate((String)arguments[i]);
	            }
	        }

			result = _messages.getString(key, arguments);

		     if (result == null)
		            result = key;

		     if (_imageHandler != null)
		         result = _imageHandler.checkMessage(this, key, arguments, result);
		}
		else if (_environment != null)
			result = _environment.getMessage(key, arguments);

		return result;
	} // end method getMessage

	/**
	 * Adds a specific message in this environment.
	 *
	 * @param key    The key that defines the message.
	 * @param string The message content
	 */
	public void putMessage(String key, String string)
	{
		if (_messages == null)
			_messages = new LyStrings(this, "messages");

		_messages.put(key, string);
	}

	/**
     * Gets the image file name corresponding to the given key.
     * @param  key the image resource key
     * @return The image file name corresponding to the given key if found, null otherwise
     */
    public String getImageFileName(String key)
    {
        String result = null;

        if (_images != null)
            result = _images.getImageFileName(key);
        else if (_environment != null)
            result = _environment.getImageFileName(key);

        return result;
    }

	/**
	 * Retrieves an image path from the application environment.
	 *
	 * @param  key The resource name (key) of the image.
	 * @return The image path, or null if it could not be found.
	 * @see    leon.misc.LyImages#getImagePath(String)
	 */
	public String getImagePath(String key)
	{
		String result = null;

		if (_images != null)
			result = _images.getImagePath(key);
		else if (_environment != null)
			result = _environment.getImagePath(key);

		return result;
	}

	/**
	 * Retrieves an image url from the application environment.
	 *
	 * @param  key The resource name (key) of the image.
	 * @return The image url, or the key if it could not be found.
	 * @see    leon.misc.LyImages#getImageUrl(String)
	 */
	public final String getImageUrl(String key)
	{
        return getImageUrl(key, false, true);
	}

	/**
	 * Retrieves a disabled image url from the application environment.
	 *
	 * @param  key The resource name (key) of the image.
	 * @return The disabled image url, or the key if it could not be found.
	 * @see    leon.misc.LyImages#getDisabledImageUrl(String)
	 */
	public final String getDisabledImageUrl(String key)
	{
		return getImageUrl(key, true, true);
	}

	/**
     * Gets the image url corresponding to the given key.
     * @param  key          the image resource key
     * @param  disabled     Indicates if disabled image should be searched.
     * @param  defaultImage If true, return default image if the URL is not found.
     * @return The image url corresponding to the given key, the default image url if not found and
     *         defaultImage parameter is set to true, null otherwise.
     */
    public String getImageUrl(String key, boolean disabled, boolean defaultImage)
    {
        String result = null;

        if (_images != null)
            result = _images.getImageUrl(key, disabled, defaultImage);
        else if (_environment != null)
            result = _environment.getImageUrl(key, disabled, defaultImage);

        if ((result != null) && (result.charAt(0) == '$'))
            return getImageUrl(result.substring(1), disabled, defaultImage);

        return result;
    }

	/**
	 * Retrieves an image from the Application environment.
	 *
	 * @param  key The resource name (key) of the image.
	 * @return A reference to the image object, or null if it could not be found.
	 * @see    LyEnvironment#getImage(String,String)
	 * @see    leon.misc.LyImages
	 */
	public Object getImage(String key)
	{
		return getImage(key, null, true);
	}

	/**
	 * Retrieves an image from the Application environment.
	 *
	 * @param  key      The resource name (key) of the image
	 * @param  sizeCode The size code of the image (see LyImages.loadImage())
	 * @return a reference to the image object, or null if it could not be found.
	 * @see    LyEnvironment#getImage(String,String,boolean)
	 */
	public Object getImage(String key, String sizeCode)
	{
		return getImage(key, sizeCode, true);
	}

	/**
	 * Retrieves an image from the Application environment.
	 *
	 * @param  key          The resource name (key) of the image
	 * @param  sizeCode     The size code of the image (see LyImages.loadImage())
	 * @param  defaultImage If true, return default image if not found.
	 * @return a reference to the image object, or null if it could not be found.
	 * @see    leon.misc.LyImages#loadImage(Object,String)
	 * @see    LyEnvironment#getImageSizeInfo(int,int)
	 */
	public Object getImage(String key, String sizeCode, boolean defaultImage)
	{
        Object result = null;

        if (_images != null)
            result = _images.getImage(key, sizeCode, defaultImage);
        else if (_environment != null)
            result = _environment.getImage(key, sizeCode, defaultImage);

        return result;
	} // end method getImage

	/**
	 * Checks if given key image corresponds to a known image.
	 *
	 * @param  key          The resource name (key) of the image
	 * @return True if found, false otherwise.
	 */
	public boolean checkImage(String key)
	{
        boolean result = false;

        if (_images != null)
            result = _images.checkImage(key);
        else if (_environment != null)
            result = _environment.checkImage(key);

        return result;
	} // end method checkImage

	/**
	 * Indicates whether given image may be cached.
	 * @param image The image that may be cached.
	 * @return true if image may be cached, false otherwise
	 */
	public boolean isImageCached(Object image)
	{
        boolean result = false;

        if (_images != null)
            result = _images.isInCache(image);
        else if (_environment != null)
            result = _environment.isImageCached(image);

        return result;
	} // end method isImageCached
	
	/**
	 * Indicates if an image is animated.
	 *
	 * @param  key The resource name (key) of the image
	 * @return True if the image is animated, false otherwise.
	 */
	public boolean isImageAnimated(String key)
	{
		if (key == null)
			return false;

		String path = getImagePath(key);

		if (_images != null)
			return _images.isImageAnimated(path);
		else if (_environment != null)
			return _environment.isImageAnimated(key);

		return false;
	}

	/**
	 * Builds an image size description info (as a string). The syntax is : [VISIBLE_SEP3 + width +
	 * VISIBLE_SEP2 + height + VISIBLE_SEP3_BIS].
	 *
	 * @param  width  The desired width for the image.
	 * @param  height The desired height for the image.
	 * @return The image info.
	 * @see    leon.misc.LyImages#getImageSizeInfo(int,int)
	 */
	public String getImageSizeInfo(int width, int height)
	{
		String result = null;

		if (_images != null)
			result = _images.getImageSizeInfo(width, height);
		else if (_environment != null)
			result = _environment.getImageSizeInfo(width, height);

		return result;
	}

	/**
	 * Checks if a size is given in the path. This method does *not* attempt to load the image.
	 *
	 * @param  path The path to check.
	 * @return An array of two integers (width, height) describing the size of the image or null if
	 *         no size is given.
	 * @see    leon.misc.LyImages#getImageSize(String, boolean)
	 */
	public int[] getImageSize(String path)
	{
		return getImageSize(path, false);
	}

	/**
	 * Checks if a size is given in the path.
	 *
	 * @param  path            The path to check.
	 * @param  enableLoadImage indicates wether the application is allowed to load the image
	 * @return An array of two integers (width, height) describing the size of the image or null if
	 *         no size is given.
	 * @see    leon.misc.LyImages#getImageSize(String, boolean)
	 */
	public int[] getImageSize(String path, boolean enableLoadImage)
	{
		int[] result = null;

		if (_images != null)
			result = _images.getImageSize(path, enableLoadImage);
		else if (_environment != null)
			result = _environment.getImageSize(path, enableLoadImage);

		return result;
	}

	/**
	 * Retrieves the path of an image without the information about its size.
	 *
	 * @param  path The path of the image.
	 * @return The path of the image without the size information.
	 * @see    leon.misc.LyImages#getPathWithoutSize(String)
	 */
	public String getPathWithoutSize(String path)
	{
		String result = null;

		if (_images != null)
			result = _images.getPathWithoutSize(path);
		else if (_environment != null)
			result = _environment.getPathWithoutSize(path);

		return result;
	}

	/**
	 * Adds a specific image in this environment.
	 *
	 * @param key    The key that defines the image.
	 * @param string The specific image path.
	 */
	public void putImage(String key, String string)
	{
		if (_images == null)
			_images = new LyImages(this);

		_images.put(key, string);
	}

	/**
	 * Gets the path of the external leonardi documentation if exists (LY_LEON_EXTERNAL_URL), the
	 * Leonardi dicumentation otherwise (LY_LEON_DOC).
	 *
	 * @return the path to the documentation
	 * @see    #getEnv(String)
	 */
	public String getLeonDocExternalUrl()
	{
		String baseUrl = getEnv("LY_LEON_DOC_EXTERNAL");

		if (baseUrl == null)
			baseUrl = getEnv("LY_LEON_DOC");

		return baseUrl;
	}

	/**
	 * Retrieves a Font from the application environment.
	 *
	 * @param  key The resource name (key) of the font.
	 * @return A reference to the requested Font, or a default font if it could not be found.
	 * @see    leon.misc.LyFonts
	 */
	public LyFont getLyFont(String key)
	{
		LyFont font = null;

		if (_fonts != null)
			font = _fonts.getFont(key);
		else if (_environment != null)
			font = _environment.getLyFont(key);

		return font;
	}

	/**
	 * Retrieves a Font from the application environment.
	 *
	 * @param  key The resource name (key) of the font.
	 * @return A reference to the requested Font, or a default font if it could not be found.
	 * @see    leon.misc.LyFonts
	 */
	public Object getFont(String key)
	{
		LyFont				   font = getLyFont(key);
		LyViewManagerInterface viewManager = getViewManager();

		return (viewManager == null ? null : viewManager.getFont(font));
	}

	/**
	 * Adds a specific font in this environment.
	 *
	 * @param key    The key that defines the font.
	 * @param string The specific font.
	 */
	public void putFont(String key, String string)
	{
		if (_fonts == null)
			_fonts = new LyFonts(this);

		_fonts.put(key, string);
	}

	/**
	 * Retrieves a Color from the application environment.
	 *
	 * @param  key The resource name (key) of the Color.
	 * @return A reference to the requested Color, or null if it could not be found.
	 * @see    leon.misc.LyColors
	 */
	public LyColor getLyColor(String key)
	{
	    LyColor color = null;

		if (_colors != null)
		{
			color = _colors.getColor(key);

			if (color == LyColor.NO_COLOR)
				return null;
		}
		else if (_environment != null)
			color = _environment.getLyColor(key);

		return color;
	}

	/**
	 * Retrieves a Color from the application environment.
	 *
	 * @param  key The resource name (key) of the Color.
	 * @return A reference to the requested Color, or null if it could not be found.
	 * @see    leon.misc.LyColors
	 */
	public Object getColor(String key)
	{
		LyColor color = getLyColor(key);

		if ((color != null) && (color.useSystemColor()))
			return null;

		LyViewManagerInterface viewManager = getViewManager();

		return (viewManager == null ? null : viewManager.getColor(color));
	}

	/**
	 * Gets the color codes list of the color_codes properties file.
	 *
	 * @return the color codes list as a Vector
	 */
	public Vector<String> getColorCodes()
	{
		if (_colors != null)
			return _colors.getColorCodes();

		if (_environment != null)
			return _environment.getColorCodes();

		return null;
	}

	/**
	 * Adds a specific color in this environment.
	 *
	 * @param key    The key that defines the color.
	 * @param string The specific color.
	 */
	public void putColor(String key, String string)
	{
		if (_colors == null)
			_colors = new LyColors(this);

		_colors.put(key, string);
	}

	/**
	 * Retrieves a File path from the application environment.
	 *
	 * @param  key The resource name (key) of the File
	 * @return A reference to the requested Path, or null if it could not be found.
	 * @see    leon.misc.LyFiles
	 */
	public String getFile(String key)
	{
		String result = null;

		if (_files != null)
			result = _files.getPath(key);
		else if (_environment != null)
			result = _environment.getFile(key);

		return result;
	}

	/**
	 * Retrieves an url from the application environment.
	 *
	 * @param  key The resource name (key) of the url
	 * @return A reference to the requested url, or null if it could not be found.
	 * @see    leon.misc.LyFiles
	 */
	public String getUrl(String key)
	{
		String result = null;

		if (_files != null)
			result = _files.getUrl(key);
		else if (_environment != null)
			result = _environment.getUrl(key);

		return result;
	}

	/**
	 * Adds a specific file in this environment.
	 *
	 * @param key    The key that defines the file.
	 * @param string The specific file.
	 */
	public void putFile(String key, String string)
	{
		if (_files == null)
			_files = new LyFiles(this);

		_files.put(key, string);
	}

	/**
	 * Test if the given key is declared in a resource file
	 * @param key a key
	 * @return true if the key is declared in a resource file.
	 */
	public boolean isResourceDefined(String key)
	{
		boolean result = false;
		if (_resources != null)
		    result = _resources.hasKey(key);
		else if (_environment != null)
			result = _environment.isResourceDefined(key);

		return result;
	}

	/**
	 * Retrieves a resource from the application environment.
	 *
	 * @param  key The resource name (key).
	 * @return The value of the requested resource, or null if it could not be found.
	 * @see    leon.misc.LyResources
	 */
	public String getResource(String key)
	{
	    String     result = null;

		if (_resources != null)
		    result = _resources.get(key);
		else if (_environment != null)
			result = _environment.getResource(key);

		return result;
	}

	/**
	 * Checks if a resource if found for given key.
	 *
	 * @param  key Resource key.
	 * @return True if found, false otherwise.
	 */
	public boolean hasResource(ResourceType type, String key)
	{
		switch (type)
		{
		case STRING:
			if (_strings != null)
				return _strings.hasKey(key);
			break;
		case MESSAGE:
			if (_messages != null)
				_messages.hasKey(key);
			break;
		case FILE:
			if (_files != null)
				_files.hasKey(key);
			break;
		case IMAGE:
			if (_images != null)
				_images.hasKey(key);
			break;
		case FONT:
			if (_fonts != null)
				_fonts.hasKey(key);
			break;
		case COLOR:
			if (_colors != null)
				_colors.hasKey(key);
			break;
		case RESOURCE:
			if (_resources != null)
				_resources.hasKey(key);
			break;
		default:
			throw new IllegalArgumentException("BF AppEngine : unknown resource type, " + type);
		}

		return false;
	} // end method hasResource

	/**
	 * Checks if a resource is set to true or 1.
	 *
	 * @param  key The resource name (key).
	 * @return True if the resource exits and is set to ttrue or 1, false otherwise.
	 * @see    LyEnvironment#getResource(String)
	 */
	public boolean testResource(String key)
	{
		String value = getResource(key);

		if (value == null)
			return false;

		// Get rid of possible spaces in the value
		value = value.trim();

		return (value.equalsIgnoreCase("true") || value.equals("1"));
	}

	/**
	 * Adds a specific resource in this environment.
	 *
	 * @param key    The key that defines the resource.
	 * @param string The specific resource.
	 */
	public void putResource(String key, String string)
	{
		if (_resources == null)
			_resources = new LyResources(this);

		_resources.put(key, string);
	}

	/**
	 * Retrieves an integer resource from the application environment.
	 *
	 * @param  key The resource name (key).
	 * @return The value of the requested resource, or -1 if it could not be found.
	 * @see    leon.misc.LyResources
	 */
	public int getIntResource(String key)
	{
		String str = getResource(key);

		return ((str == null) ? -1 : LyTools.intFromString(str));
	}

	/**
	 * Retrieves a value from the contextual environment. Seeks in the local property table
	 * (_properties), and if not found a second search is done in the global environment (System or
	 * applet). Properties may be added using the putEnv method or may be read in the ini file of
	 * the application.
	 *
	 * @param  key The key associated with the environment parameter.
	 * @return The environment variable or null if the value could not be found.
	 * @see    LyEnvironment#putEnv(String,Object)
	 */
	public String getEnv(String key)
	{
		if ((key == null)
				|| (key.length() == 0))
			return null;

		if (key.charAt(0) == '$')
		{
			// Try first without $
			String value = getEnv(key.substring(1));

			if (value != null)
				return value;
		}

		if (_properties != null)
		{
	        // Check in contextual property table
	        Object value = _properties.get(key);

	        if (value != null) // Found
	            return value.toString();
		}

		// Search in the local or global environment.
		String	value;

		if (_environment != null)
			value = _environment.getEnv(key);
		else
			value = System.getProperty(key);

		if ((value == null)
				&& key.startsWith("_LY_"))
			// For upward compatibility, check also without _
			value = getEnv(key.substring(1));

		return value;
	} // end method getEnv

	/**
	 * Adds a value in the contextual environment. Both key and values should be not null.
	 *
	 * @param  key   The Id of the added property.
	 * @param  value The value added in the environnment.
	 * @return the previous value of the specified key in the environnment, or null if it did not
	 *         have one.
	 * @see    LyEnvironment#getEnv(String)
	 */
	public Object putEnv(String key, Object value)
	{
		if (key == null)
			return null;
		if (value == null)
		{
			if (_properties != null)
				_properties.remove(key);
			return null;
		}

		if (_properties == null)
			_properties = new Properties();

		if (key.equals("LY_SKIN"))
		    setSkin((String)value);
		else if (key.equals("LY_LANGUAGE"))
		    setLanguage((String)value);

		// Add a new entry in the property table
		return _properties.put(key, value);
	}

	/**
	 * Checks if an environment parameter is set. A parameter is set when the key is in the
	 * application environment properties and when the value is true or 1.
	 *
	 * @param  key The key that defines the parameter.
	 * @return true if the environment is set.
	 * @see    LyEnvironment#putEnv(String,Object)
	 */
	public boolean testEnv(String key)
	{
		String value = getEnv(key);

		if (value == null)
			return false;

		// Get rid of possible spaces in the value
		value = value.trim();

		return (value.equalsIgnoreCase("true") || value.equals("1"));
	}

    /**
     * Retrieves a value from this environment.
     *
     * @param  key The resource name (key).
     * @return The value corresponding to the given key, or null if it could not be found.
     * @see    leon.misc.LyResources
     */
    public String getViewAttValue(String key)
    {
        String result = null;

        if (_viewAttValues != null)
        {
            result = _viewAttValues.get(key);

            if ((result == null) || (result.length() == 0))
                return key;
        }
        else if (_environment != null)
            result = _environment.getViewAttValue(key);

        return result;
    }

	/**
	 * Initialize the environment.
	 */
	public void initResources()
	{
		// XXXX This is called twice by the LyApplication constructor:
		// - once for the LyEnvironment constructor
		// - once when the ini file is loaded.
		// XXX Check if necessary !

		// Free resource bundles associated to this environment
		ArrayList<LyEnvironment> sonList = LyResourceBundle.getSonEnvironments(this);

		if (_strings != null)
			_strings.free();
		if (_images != null)
			_images.free();
		if (_colors != null)
			_colors.free();
		if (_fonts != null)
			_fonts.free();
		if (_files != null)
			_files.free();
		if (_messages != null)
			_messages.free();
		if (_resources != null)
			_resources.free();
        if  (_viewAttValues != null)
            _viewAttValues.free();

		// Creating new resource bundles for this environment.
		_strings = new LyStrings(this);
		_images = new LyImages(this);
		_colors = new LyColors(this);
		_fonts = new LyFonts(this);
		_files = new LyFiles(this);
		_messages = new LyStrings(this, "messages");
		_string = new LyString(this);
		_resources = new LyResources(this);
        _viewAttValues = new LyResources(this, "views_values");

		LyEnvironment son;
		int           i, n = (sonList == null ? 0 : sonList.size());

		for (i = 0; i < n; i++)
		{
			son = sonList.get(i);
			son.initResources();
		}

		if (getEnv("LY_PROJECT_FILE") != null)
			putFile("LY_PROJECT_FILE", getEnv("LY_PROJECT_FILE"));
	} // end method initResources

	/**
	 * Loads the environment from the ".ini" file. Default file is found in system property
	 * LY_INI_FILE.
	 *
	 * @param  path The full path to the environment file.
	 * @return true if the file was successfully read, false otherwise.
	 * @see    #loadEnv(String, ClassLoader, boolean)
	 */
	public boolean loadEnv(String path)
	{
		return loadEnv(path, null, true);
	}

	/**
	 * Loads the environment from ".ini" file. Default file is found in system property LY_INI_FILE.
	 *
	 * @param      path        The full path to the environment file.
	 * @param      classLoader optional classloader to look in CLASSPATH
	 * @return     true if the file was successfully read, false otherwise.
	 * @deprecated use loadEnv(String path, ClassLoader classLoader, true) instead of this one
	 */
	public boolean loadEnv(String path, ClassLoader classLoader)
	{
		return loadEnv(path, classLoader, true);
	}

	/**
	 * Loads the environment from ini file. Default file is found in system property LY_INI_FILE.
	 *
	 * @param  path        The full path to the environment file.
	 * @param  classLoader optional classloader to look in CLASSPATH
	 * @param  initialize  indicates if we call the endInitialize method (true) or just read ini
	 *                     file (false)
	 * @return true if the file was successfully read, false otherwise.
	 */
	public boolean loadEnv(String path, ClassLoader classLoader, boolean initialize)
	{
		return loadEnv(path, classLoader, initialize, null);
	}


	/**
	 * Loads the environment from ini file. Default file is found in system property LY_INI_FILE.
	 *
	 * @param  path        The full path to the environment file.
	 * @param  classLoader optional classloader to look in CLASSPATH
	 * @param  initialize  indicates if we call the endInitialize method (true) or just read ini
	 *                     file (false)
	 * @param lyAppDir 	   new value of the LY_APP_DIR variable
	 * @return true if the file was successfully read, false otherwise.
	 */
	public boolean loadEnv(String path, ClassLoader classLoader, boolean initialize, String lyAppDir)
	{
		if (path == null)
			path = getEnv("LY_INI_FILE");

		if (classLoader == null)
			classLoader = getClassLoader();

		boolean initialized = false;

		if (path != null)
		{
			path = getString().expand(path);

			File	    f = new File(path);
			InputStream istream = null;

			try
			{
				// Check if .ini file exists
				if (f.exists())
					istream = new FileInputStream(f);
                else
					// Try from the classpath
					istream = getResourceAsStream(path, classLoader);

				if (istream != null)
				{
					// Load properties
					if (_properties == null)
						_properties = new Properties();

					_properties.load(istream);

					// Add extra environment variables to parent environment
					if (_environment != _rootEnvironment)
					{
			            Enumeration<Object> properties = _properties.keys();
			            String				key;
			            Object				value;

			            while (properties.hasMoreElements())
			            {
			                key = (String)properties.nextElement();
			                if (_environment.getEnv(key) == null)
			                {
			                	value = _properties.get(key);
			                	_environment.putEnv(key, value);
			                }
			            }
			        }

					if (lyAppDir != null)
						_properties.put("LY_APP_DIR", lyAppDir);

					putEnv("_LY_ACTUAL_INI_FILE", path);
					parseEnv();

					// The initialization process successfully completed
					initialized = true;
				}
				else
					out.println("BF AppEngine : .ini file \"" + path + "\" not found.");
			}
			catch (IOException e)
			{
				writeException(e);
			}
		}
		else
		{
			parseEnv();
			initialized = true;
		}

		if (initialize)
			endInitialize();

		return initialized;
	} // end method loadEnv

	/**
	 * Checks all initialization environment parameters (data directories, images, files...) This
	 * method shall generate warning messages if some resources are missing.
	 *
	 * @return true
	 * @see    #checkPath(String, String, boolean, boolean)
	 */
	public synchronized boolean checkEnv()
	{
		String  viewerType = getEnv("LY_VIEWER_TYPE");
		boolean web = ("HTML".equals(viewerType)
					|| "DHTML".equals(viewerType)
					|| "WEB".equals(viewerType)
					|| "WEB2".equals(viewerType)
					|| "JQUERY".equals(viewerType)
					|| "STRUTS".equals(viewerType));

		// Directories : key, fatalError, required
		checkPath("LY_LEON_DIR", null, true, true);
		checkPath("LY_APP_DIR", null, true, true);

		if (web)
		{
		    // Checks required in web mode only
			checkPath("LY_LEON_DOC_DIR", null, true, true);
			checkPath("LY_APP_DOC_DIR", null, true, true);
			checkPath("LY_APP_DOC_DIR", "/images", false, false);
			checkPath("LY_LEON_DOC_DIR", "/images", true, true);
			checkPath("LY_LEON_DOC_DIR", "/scripts", true, true);
		}

		checkPath("LY_DATA_DIR", null, false, false);

		// Files
		// The XML project file is checked only if we don't use a compiled meta-model.
		if (!web)
			checkFile("LY_EXTERNAL_VIEWER", false, false);

		return true;
	} // end method checkEnv

	/**
	 * Checks if the specified directory exists. If the file does not exists and error message is
	 * raised. The fatalError parameter indicates if the application can start after this error or
	 * not. FatalError throws an exception, otherwise a warning is printed to the console.
	 *
	 * @param  key        the key associated to the path we want to check
	 * @param  extraPath  additional path checked at the end of specified path
	 * @param  fatalError if an error occured, shall we log the error to the console or stop the
	 *                    execution of the application (Exception).
	 * @param  required   indicates that the resource MUST be set.
	 * @return current status (true if everything is fine), false otherwise.
	 */
	public boolean checkPath(String key, String extraPath, boolean fatalError, boolean required)
	{
		File   f = null;
		String path = getEnv(key);

		// Optional field
		if (path == null)
		{
			if (!required)
				return true;
		}
		else
		{
			if (extraPath != null)
				path = path + extraPath;

			path = getString().expand(path);

			f = new File(path);
		}

		if ((f == null) || (!f.exists()) || (!f.isDirectory()))
		{
			if (fatalError)
				throw new IllegalStateException("BF AppEngine : "
					+ getMessage("LY_ER_CHECK_PATH", key, f));
			else
			{
				logWarning("BF AppEngine : " + getMessage("LY_ER_CHECK_PATH", key, f));
				return false; // continue
			}
		}
		return true;
	} // end method checkPath

	/**
	 * Checks if the specified path exists. If the file does not exists and error message is raised.
	 * The fatalError parameter indicates if the application can start after this error or not.
	 * FatalError throws an exception, otherwise a warning is printed to the console.
	 *
	 * @param  key        the key associated to the path we want to check
	 * @param  fatalError if an error occured, shall we log the error to the console or stop the
	 *                    execution of the application (Exception).
	 * @param  required   indicates that the resource MUST be set.
	 * @return current status (true if everything is fine), false otherwise.
	 */
	public boolean checkFile(String key, boolean fatalError, boolean required)
	{
		File   f = null;
		String path = getFile(key);

		// Optional field
		if (((path == null) || path.equals(key))
				&& !required)
			return true;

		if (path != null)
			f = new File(path);

		if ((f == null) || (!f.exists()) || (!f.isFile()))
		{
			if (fatalError)
				throw new IllegalStateException("BF AppEngine : "
					+ getMessage("LY_ER_CHECK_FILE", key, f));
			else
			{
				logWarning(getMessage("LY_ER_CHECK_FILE", key, f));
				return false; // continue
			}
		}
		return true;
	}

	/**
	 * Gets the default timezone for current context and current environment.
	 *
	 * @return Default timezone.
	 */
	public TimeZone getTimeZone()
	{
		String timeZone;

		timeZone = getEnv("user.timezone");
		if (timeZone == null)
			timeZone = getEnv("LY_TIMEZONE");

		if (timeZone != null)
			return TimeZone.getTimeZone(timeZone);
		else
			return TimeZone.getDefault();
	}

	/**
	 * Retrieves an inputStream from a local address.
	 *
	 * @param  name        the address of the input stream to retrieve
	 * @param  classLoader the class loader that may retrieves the InputStream
	 * @return The inputStream if found, null otherwise
	 */
	public InputStream getResourceAsStream(String name, ClassLoader classLoader)
	{
		// Try from the classpath
		InputStream istream = null;

		if (classLoader == null)
			classLoader = getClassLoader();

		if (classLoader == null)
		{
			try
			{
				// try to get current class loader.
				// / May not be system classloader with .war archives
				classLoader = getClass().getClassLoader();
			}
			catch (Exception e)
			{
				writeException(e);
			}
		}

		if (classLoader == null)
			classLoader = ClassLoader.getSystemClassLoader();

		// For the method getResourceAsStream on the class loader, the file
		// separator must be '/' (see javadoc)
		name = LyString.replace(name, "\\", "/");

		try
		{
			istream = classLoader.getResourceAsStream(name);
		}
		catch (Exception e)
		{
		}

		return istream;
	} // end method getResourceAsStream

	/**
	 * Retrieves an inputStream from a local address.
	 *
	 * @param  path The path may be the access to a file or an URL.
	 * @return The inputStream or null if the stream does not exists.
	 */
	public InputStream getInputStream(String path)
	{
		return getInputStream(path, true);
	}

	/**
	 * Retrieves an inputStream from a local address.
	 *
	 * @param  path      The path may be the access to a file or an URL.
	 * @param  showError indicates if an error may be shown to the user
	 * @return The inputStream or null if the stream does not exists.
	 */
	@SuppressWarnings("resource")
	public InputStream getInputStream(String path, boolean showError)
	{
		InputStream stream = null;

		if ((_string == null) && (getId() != null))
			_string = new LyString(this);

		if (_string != null)
		{
			String expandedPath = _string.expand(path);

			try
			{
				stream = new FileInputStream(expandedPath);
			}
			catch (Exception e)
			{
				// Do not trace if another attempt is done from CLASSPATH
				if (showError)
				{
					writeException(e);
					String error = getMessage("LY_ER_FILE", expandedPath);

					if (getViewManager() != null)
						getViewManager().raiseError(error);
				}

				stream = null;
			}
		}

		if ((stream == null) && (_environment != null))
			stream = _environment.getInputStream(path, showError);

		return stream;
	} // end method getInputStream

	/**
	 * Retrieves an outputStream from a local address.
	 *
	 * @param  path      The path may be the access to a file or an URL.
	 * @param  showError indicates if an error may be shown to the user
	 * @return The outputStream or null if the stream does not exists.
	 */
	@SuppressWarnings("resource")
	public OutputStream getOutputStream(String path, boolean showError)
	{
		OutputStream stream = null;

		if ((_string == null) && (getId() != null))
			_string = new LyString(this);

		if (_string != null)
		{
			String expandedPath = _string.expand(path);

			try
			{
				stream = new FileOutputStream(expandedPath);
			}
			catch (Exception e)
			{
				// Do not trace if another attemp is done from CLASSPATH
				if (showError)
				{
					writeException(e);
					String error = getMessage("LY_ER_FILE", expandedPath);

					if (getViewManager() != null)
						getViewManager().raiseError(error);
				}

				stream = null;
			}
		}

		if ((stream == null) && (_environment != null))
			stream = _environment.getOutputStream(path, showError);

		return stream;
	} // end method getOutputStream

	/**
	 * Gets general configuration parameters from property file (".ini" file). They represents Lyria
	 * "kernel" properties such as LY_DEBUG_LEVEL.
	 *
	 * @see LyEnvironment#loadEnv(String)
	 */
	public void parseEnv()
	{
		// Debug level
		String debugLevel = getEnv("LY_DEBUG_LEVEL");

		if (debugLevel != null)
		{
			debugLevel = debugLevel.toUpperCase();

			if (debugLevel.indexOf("DEBUG_ALL") >= 0)
				setDebugLevel(LyEnvironment.DEBUG_ALL);

			if (debugLevel.indexOf("DEBUG_ERROR") >= 0)
				setDebugLevel(LyEnvironment.DEBUG_ERROR);

			if (debugLevel.indexOf("DEBUG_WARNING") >= 0)
				setDebugLevel(LyEnvironment.DEBUG_WARNING);

			if ((debugLevel.indexOf("DEBUG_INFO") >= 0) || (debugLevel.indexOf("DEBUG_LOG") >= 0))
				setDebugLevel(LyEnvironment.DEBUG_INFO);

			if (debugLevel.indexOf("DEBUG_TRACE") >= 0)
				setDebugLevel(LyEnvironment.DEBUG_TRACE);

			if (debugLevel.indexOf("DEBUG_DBMS") >= 0)
				setDebugLevel(LyEnvironment.DEBUG_DBMS);

			if (debugLevel.indexOf("DEBUG_BUS") >= 0)
				setDebugLevel(LyEnvironment.DEBUG_BUS);

			if (debugLevel.indexOf("DEBUG_PERF") >= 0)
				setDebugLevel(LyEnvironment.DEBUG_PERF);

			if (debugLevel.indexOf("DEBUG_XML") >= 0)
				setDebugLevel(LyEnvironment.DEBUG_XML);

			if (debugLevel.indexOf("DEBUG_GRAPHICS") >= 0)
				setDebugLevel(LyEnvironment.DEBUG_GRAPHICS);

			if (debugLevel.indexOf("DEBUG_PEER") >= 0)
				setDebugLevel(LyEnvironment.DEBUG_PEER);

			if (debugLevel.indexOf("DEBUG_DUMP_HTML") >= 0)
				setDebugLevel(LyEnvironment.DEBUG_DUMP_HTML);

			if (debugLevel.indexOf("DEBUG_HELP_FILES") >= 0)
				setDebugLevel(LyEnvironment.DEBUG_HELP_FILES);

			if (debugLevel.indexOf("DEBUG_TRANSACTION") >= 0)
				setDebugLevel(LyEnvironment.DEBUG_TRANSACTION);

			if (debugLevel.indexOf("DEBUG_PRINT_PDF") >= 0)
				setDebugLevel(LyEnvironment.DEBUG_PRINT_PDF);

			if (debugLevel.indexOf("DEBUG_LOCKS") >= 0)
				setDebugLevel(LyEnvironment.DEBUG_LOCKS);
		}

		if (getEnv("LY_APP_DIR") != null)
		{
			putEnv("LY_APP_DIR", getString().expand(getEnv("LY_APP_DIR")));
		}
		if (getEnv("LY_PROJECT_FILE") != null)
			putEnv("LY_PROJECT_FILE", getString().expand(getEnv("LY_PROJECT_FILE")));
		if (getEnv("LY_DATA_DIR") != null)
		{
			// If $LY_APP_DIR$ found in LY_DATA_DIR and LY_APP_DIR not defined,
			// do not set LY_APP_DIR now but in the setDefaultProperties method
			String dataDir = getEnv("LY_DATA_DIR");

			if ((getEnv("LY_APP_DIR") != null) || (dataDir.indexOf("$LY_APP_DIR$") < 0))
				putEnv("LY_DATA_DIR", getString().expand(getEnv("LY_DATA_DIR")));
		}
		if (getEnv("LY_LEON_DOC_DIR") != null)
			putEnv("LY_LEON_DOC_DIR", getString().expand(getEnv("LY_LEON_DOC_DIR")));
		if (getEnv("LY_APP_DOC_DIR") != null)
			putEnv("LY_APP_DOC_DIR", getString().expand(getEnv("LY_APP_DOC_DIR")));

		// Specific for Easy install
		if (getEnv("_LY_EXTERNAL_RESOURCES_FILE") != null
				&& !"true".equals(getEnv("_LY_EXTERNAL_RESOURCES_FILE_LOADED")))
		{
			String	path = getEnv("_LY_EXTERNAL_RESOURCES_FILE");
			File	f = new File(path);

			if (!f.exists()
				&& (getEnv("LY_APP_DIR") != null))
			{
				path = getEnv("LY_APP_DIR")+File.separator+path;
				f = new File(path);
				if(!f.exists())
					out.println("BF AppEngine : _LY_EXTERNAL_RESOURCES_FILE file \"" + path + "\" not found.");
			}

			putEnv("_LY_EXTERNAL_RESOURCES_FILE_LOADED", "true");

			String	actualIniFile = getEnv("_LY_ACTUAL_INI_FILE");

			loadEnv(path, null, false);
			putEnv("_LY_ACTUAL_INI_FILE", actualIniFile);
		}
	} // end method parseEnv

	/**
	 * Creates a new Calendar instance for the specified or default timeZone Set the time zone
	 * parameter in .ini file with the user.timezone key. If no timeZone is set, the default
	 * timeZone is ECT.
	 *
	 * @return A new Calendar instance for the environment timezone.
	 */
	public Calendar calendarFactory()
	{
		Calendar calendar;
		TimeZone tz = getTimeZone();

		if (tz != TimeZone.getDefault())
			TimeZone.setDefault(tz);

		calendar = Calendar.getInstance(tz);
		return calendar;
	}

	/**
	 * Returns the string representing the date format of the given type. Strings format table is
	 * set as association table type-format in _stringDateFormatTable.
	 *
	 * @param  type The date type (use LyDateFieldInfo types constants).
	 * @return a string representing the format.
	 * @see    leon.info.LyDateFieldInfo
	 * @see    LyEnvironment#dateFormatFactory(int)
	 */
	public String getStringDateFormat(int type)
	{
		// Check if table is ready
		if ((_stringDateFormatTable == null) && (_environment != null))
		{
			Locale locale = getDateLocale();
			Locale parentLocale = _environment.getDateLocale();

			if (locale.equals(parentLocale))
				return _environment.getStringDateFormat(type);
		}

		if (_stringDateFormatTable == null)
		{
			ResourceBundle rb = null;
			String		   dateFormat = "dd/MM/yyyy";
			String		   timeFormat = "HH:mm:ss";
			String		   hourMinFormat = "HH:mm";

			try
			{
				// JDK 1.3
				rb = ResourceBundle.getBundle("java.text.resources.LocaleElements",
						getDateLocale());
			}
			catch (Exception e)
			{
			}

			if (rb == null)
			{
				try
				{
					// JDK 1.4 (class is moved)
					rb = ResourceBundle.getBundle("sun.text.resources.LocaleElements",
							getDateLocale());
				}
				catch (Exception e)
				{
				}
			}

			if (rb == null)
			{
				try
				{
					// JDK 1.6 (class is moved again)
					rb = ResourceBundle.getBundle("sun.text.resources.FormatData", getDateLocale());
				}
				catch (Exception e)
				{
					writeException(e);
				}
			}

			if (rb != null)
			{
				String[] dateTimePatterns = rb.getStringArray("DateTimePatterns");

				if (dateTimePatterns != null)
				{
					// Retrieve short date format and complete format in order to get 4 digits for years,
					// two digits for months and two digits for days.
					String	datePattern = dateTimePatterns[DateFormat.SHORT + 4];

					if (datePattern != null)
					{
						datePattern = completeStringDateFormat(datePattern, 'y', 4);
						datePattern = completeStringDateFormat(datePattern, 'M', 2);
						datePattern = completeStringDateFormat(datePattern, 'd', 2);
						dateFormat = datePattern;
					}

					timeFormat = dateTimePatterns[DateFormat.MEDIUM];
					hourMinFormat = dateTimePatterns[DateFormat.SHORT];
				}
			}

			_stringDateFormatTable = new String[TYPE_HOUR_MIN + 1];

			_stringDateFormatTable[TYPE_DATE_TIME] = dateFormat + " " + timeFormat;
			_stringDateFormatTable[TYPE_DATE_HOUR_MIN] = dateFormat + " " + hourMinFormat;
			_stringDateFormatTable[TYPE_DATE] = dateFormat;
			_stringDateFormatTable[TYPE_TIME] = timeFormat;
			_stringDateFormatTable[TYPE_HOUR_MIN] = hourMinFormat;
		}

		return _stringDateFormatTable[type];
	} // end method getStringDateFormat

	/**
	 * Completes a given Java date format to be used in Leonardi applications.
	 */
	private String completeStringDateFormat(String dateFormat, char formatChar, int occurences)
	{
		String	datePattern = dateFormat;
		int		i, pos1 = datePattern.indexOf(formatChar), pos2 = datePattern.lastIndexOf(formatChar);

		if (pos1 < 0)
			return datePattern;

		for (i = 0; i < pos1 - pos2 + occurences - 1; i++)
			datePattern = datePattern.substring(0, pos1) + formatChar + datePattern.substring(pos1);

		return datePattern;
	}

	/**
	 * Creates a new DateFormat instance of the given type. Format table is set as association table
	 * type-pattern in _dateFormatTable.
	 *
	 * @param  type The type of date (use LyDateFieldInfo types constants).
	 * @return a new DateFormat instance.
	 * @see    leon.info.LyDateFieldInfo
	 * @see    LyEnvironment#getStringDateFormat(int)
	 */
	public DateFormat dateFormatFactory(int type)
	{
		// If parent environment has same locale than current environment
		// we use the parent environment.
		if ((_dateFormatTable == null) && (_environment != null))
		{
			Locale locale = getDateLocale();
			Locale parentLocale = _environment.getDateLocale();

			if (locale.equals(parentLocale))
				return _environment.dateFormatFactory(type);
		}

		if (_dateFormatTable == null)
		{
			int    t, n = TYPE_HOUR_MIN;
			String format;

			_dateFormatTable = new DateFormat[n + 1];

			for (t = 0; t < (n + 1); t++)
			{
				format = getStringDateFormat(t);
				_dateFormatTable[t] = new SimpleDateFormat(format, getDateLocale());
			}
		}

		return _dateFormatTable[type];
	} // end method dateFormatFactory

	/**
	 * Gets the date format associated to the given pattern.
	 *
	 * @param  pattern The pattern used for the requested format.
	 * @return The date format associated to the given pattern.
	 */
	public DateFormat getDisplayDateFormat(String pattern)
	{
		// Check first if pattern is internationalized
		pattern = getString(pattern);

		// If parent environment has same locale than current environment
		// we use the parent environment.
		if ((_displayDateFormatTable == null) && (_environment != null))
		{
			Locale locale = getDateLocale();
			Locale parentLocale = _environment.getDateLocale();

			if (locale.equals(parentLocale))
				return _environment.getDisplayDateFormat(pattern);
		}

		if (_displayDateFormatTable == null)
			_displayDateFormatTable = new Hashtable<String, DateFormat>();

		DateFormat result = null;

		result = _displayDateFormatTable.get(pattern);

		if (result == null)
		{
			result = new SimpleDateFormat(pattern, getDateLocale());
			_displayDateFormatTable.put(pattern, result);
		}

		return result;
	} // end method getDisplayDateFormat

	/**
	 * Gets the deciaml format associated to the given pattern.
	 *
	 * @param  pattern The pattern used for the requested format.
	 * @return The decimal format associated to the given pattern.
	 */
	public NumberFormat getNumberFormat(String pattern)
	{
		// Check first if pattern is internationalized
		pattern = getString(pattern);

		// If parent environment has same locale than current environment
		// we use the parent environment.
		if ((_decimalFormatTable == null) && (_environment != null))
		{
			Locale locale = getLocale();
			Locale parentLocale = _environment.getLocale();

			if (locale.equals(parentLocale))
				return _environment.getNumberFormat(pattern);
		}

		if (_decimalFormatTable == null)
			_decimalFormatTable = new Hashtable<String, NumberFormat>();

		NumberFormat result = null;

		if (pattern != null)
			result = _decimalFormatTable.get(pattern);

		if (result == null)
		{
			if (pattern == null)
			{
				result = NumberFormat.getInstance(getLocale());
				pattern = ((DecimalFormat)result).toPattern();

				if ((LyEnvironment._decimalSeparator != null)
						|| (LyEnvironment._groupingSeparator != null))
				{
					DecimalFormatSymbols symbols = new DecimalFormatSymbols(getLocale());

					if (LyEnvironment._decimalSeparator != null)
						symbols.setDecimalSeparator(LyEnvironment._decimalSeparator.charValue());

					if (LyEnvironment._groupingSeparator != null)
						symbols.setGroupingSeparator(LyEnvironment._groupingSeparator.charValue());

					((DecimalFormat)result).setDecimalFormatSymbols(symbols);
				}
			}
			else
			{
				DecimalFormatSymbols symbols = new DecimalFormatSymbols(getLocale());

				if (LyEnvironment._decimalSeparator != null)
					symbols.setDecimalSeparator(LyEnvironment._decimalSeparator.charValue());

				if (LyEnvironment._groupingSeparator != null)
					symbols.setGroupingSeparator(LyEnvironment._groupingSeparator.charValue());

				result = new DecimalFormat(pattern, symbols);
			}

			_decimalFormatTable.put(pattern, result);
		}

		return result;
	} // end method getNumberFormat

	/**
	 * Gets the help file for the specified context, or null if help file could not be found.
	 * Programs attempts to find in help directory : help/classId/actionId_XX.html then
	 * help/classId/rootActionId_XX.html then help/actionId_XX.html then help/rootActionId_XX.html
	 * which is the default leonardi file.
	 *
	 * @param  actionId     Current action identifier (attempt to get help for the specific action)
	 * @param  rootActionId Root action identifier
	 * @param  classId      Class id identifier.
	 * @return The help file path if it is found, null otherwise.
	 * @see    leon.misc.LyHelpFiles
	 */
	public String getHelpFile(String actionId, String rootActionId, String classId)
	{
		String result = null;

		if (_help != null)
			result = _help.getHelpFile(actionId, rootActionId, classId);
		else if (_environment != null)
			result = _environment.getHelpFile(actionId, rootActionId, classId);

		return result;
	}

	/**
	 * Retrieves the screen size.
	 *
	 * @param  component The component used to retrieve the screen size.
	 * @return The dimension of the screen used to display views.
	 * @see    leon.view.LyViewManager#getScreenSize(Object)
	 */
	public int[] getScreenSize(Object component)
	{
        LyViewManagerInterface viewManager = getViewManager();

        return ((viewManager == null) ? null : viewManager.getScreenSize(component));
	}

	/**
	 * Retrieves the viewport size available on the computer.
	 * @return The addition of all screen sizes.
	 */
	public int[] getMaxScreenSize()
	{
        LyViewManagerInterface viewManager = getViewManager();

        return ((viewManager == null) ? null : viewManager.getMaxScreenSize());
	}

	/**
	 * Gets the environment identifier.
	 *
	 * @return the environment identifier
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets the parent environment of this environment.
	 *
	 * @return the parent environment of this environment.
	 */
	protected LyEnvironment getParent()
	{
		return _environment;
	}

	/**
	 * Gets the environment language. Default values are defined in LANGUAGE_xxx constants.
	 *
	 * @return Current environment language if has been defined, null otherwise
	 * @see    LyEnvironment#setLanguage(String)
	 */
	public String getLanguage()
	{
		if (_language != null)
			return _language;

		if (_environment != null)
			return _environment.getLanguage();

		return null;
	}

	/**
	 * Gest the environment skin.
	 *
	 * @return Current environment skin if it has been defined, otherwise returns the parent one or
	 *         null if no parent.
	 * @see    LyEnvironment#setSkin(String)
	 */
	public String getSkin()
	{
		if (_skin != null)
			return _skin;

		if (_environment != null)
			return _environment.getSkin();

		return DEFAULT_SKIN;
	}

	/**
	 * Gets the environment date language. Default values are defined in LANGUAGE_xxx constants.
	 *
	 * @return Current environment date language if has been defined, null otherwise
	 * @see    LyEnvironment#setDateLanguage(String)
	 */
	public String getDateLanguage()
	{
		if (_dateLanguage != null)
			return _dateLanguage;

		if (_environment != null)
			return _environment.getDateLanguage();

		return null;
	}

	/**
	 * Gets the environment debug level. Default values are defined in DEBUG_xxx constants.
	 *
	 * @return A bitset containing all debug levels that are set.
	 * @see    LyEnvironment#isDebugLevel(int)
	 * @see    LyEnvironment#setDebugLevel(int,boolean)
	 */
	protected BitSet getDebugLevel()
	{
		return _debugLevel;
	}

	/**
	 * Checks if a debug level is set or not. Default values are defined in DEBUG_xxx constants.
	 *
	 * @param  level The level that must be checked.
	 * @return true if the level is set, false otherwise.
	 * @see    LyEnvironment#setDebugLevel(int,boolean)
	 */
	public boolean isDebugLevel(int level)
	{
		if (level < 0)
			return false;

		if ((level == DEBUG_NONE) && ((_debugLevel == null) || (_debugLevel.isEmpty())))
			return true;

		if (_debugLevel != null)
			// checks if current debug level is set.
			return _debugLevel.get(level);

		return false;
	}

	/**
	 * Retrieves the environment view manager. If the environment does not have its own view manager
	 * (_hasViewManager is false) the view manager of the parent is returned.
	 *
	 * @return A reference to the LyViewManager instance or to the view manager of the parent.
	 * @see    leon.view.LyViewManager
	 */
	public LyViewManagerInterface getViewManager()
	{
		return getViewManager(false);
	}

	/**
	 * Retrieves the environment view manager. If the environment does not have its own view manager
	 * (_hasViewManager is false) the view manager of the parent is returned.
	 *
	 * @return A reference to the LyViewManager instance or to the view manager of the parent.
	 * @see    leon.view.LyViewManager
	 */
	public LyViewManagerInterface getViewManager(boolean isMobile)
	{
		_isMobile = isMobile;
		if (_viewManager != null)
            return _viewManager;

		if (_hasViewManager)
		{
			if (testEnv("_LY_DATA_IN_TEMP_FILE"))
			{
				String tempDir = getString().expand("$LY_TEMP_DIR$");

				if ((tempDir != null) && !tempDir.equals(""))
					putEnv("LY_DATA_DIR", getString().expand("$LY_TEMP_DIR$/data"));
			}

			String viewer = getEnv("LY_VIEWER_TYPE");

			if (viewer == null)
				viewer = "SWING";
			else if (viewer.equals("NONE"))
			{
				_hasViewManager = false;
				return null;
			}

			String viewManagerClassname = getEnv("_LY_VIEWMANAGER_CLASS");
			String viewerClassName;

			if (viewer.indexOf("DHTML") >= 0)
			{
				out.println(
					"Use LY_VIEWER_TYPE=STRUTS in .ini file instead of LY_VIEWER_TYPE=DHTML. (deprecated)");
				viewer = "STRUTS";
			}
			if ((viewer.indexOf("WEB2") >= 0) || (viewer.indexOf("JQUERY") >= 0))
				viewerClassName = "leon.view.web.jquery.LyJQViewManager";
			else if ((viewer.indexOf("WEB") >= 0) || (viewer.indexOf("STRUTS") >= 0))
				viewerClassName = "leon.view.web.struts.LyStrutsViewManager";
			else if (viewer.indexOf("SWT") >= 0)
				viewerClassName = "eu.w4.ae.view.swt.LySwtViewManager";
			else if (viewer.indexOf("ECLIPSE") >= 0)
				viewerClassName = "eu.w4.ae.view.swt.eclipse.LyEclipseViewManager";
			else
				viewerClassName = "leon.view.win.swing.LySwingViewManager";

			if ((viewManagerClassname != null) && (viewManagerClassname.length() != 0))
				viewerClassName = viewManagerClassname;

			if(_isMobile == true)
				viewerClassName = "leon.view.mobile.LyMobileViewManager";

			Class<?>[]  types = new Class<?>[1];
			Object[]    parameters = new Object[1];

			types[0] = LyEnvironment.class;
			parameters[0] = this;

			try
			{
				ClassLoader classLoader = getClassLoader();
				Class<?>	viewerClass = ((classLoader != null)
						? classLoader.loadClass(viewerClassName)
						: Class.forName(viewerClassName));
				Constructor<?> constructor = viewerClass.getConstructor(types);

				setViewManager((LyViewManagerInterface)constructor.newInstance(parameters));
			}
			catch (Exception e)
			{
				writeException(e);
				throw new IllegalArgumentException("BF AppEngine : can't instantiate view manager "
					+ viewerClassName + " (specified viewer type : " + viewer + ").");
			}
		}
		else if (_environment != null)
		{
			String viewer = getEnv("LY_VIEWER_TYPE");
			if ((viewer == null) || !viewer.equals("NONE"))
				_viewManager = _environment.getViewManager();
		}

		return _viewManager;
	} // end method getViewManager

	/**
	 * Sets the application view manager. This method should be used with caution and only if the
	 * application needs to have the control on the view component instances that are created.
	 *
	 * @param viewManager a reference to the LyViewManager instance or to the view manager of the
	 *                    parent.
	 * @see   leon.view.LyViewManager
	 */
	public void setViewManager(LyViewManagerInterface viewManager)
	{
        if (_viewManager == viewManager)
            return;

        if ((viewManager != null)
                && (_viewManager != null))
            throw new IllegalStateException("BF AppEngine : environment " + getId()
                    + " has already a view manager");

		if ((viewManager != null)
                && (_environment != null)
		        && (_environment.getViewManager() == null))
            // If parent environment has no view manager, set the view manager
			_environment.setViewManager(viewManager);

        if ((viewManager == null)
            && _hasViewManager
            && ((_environment == null)
                || (_environment.getViewManager() != _viewManager)))
            // Free view manager if environment is the last to reference it
            _viewManager.free();

        _viewManager = viewManager;
	}

	/**
	 * Gets the default locale for current context and current environment.
	 *
	 * @return The locale corresponding to the language of the environment if it is set, the locale
	 *         corresponding to the LY_LANGUAGE configuration parameter otherwise.
	 */
	public Locale getLocale()
	{
		return getLocale(getLanguage());
	}

	/**
	 * Gets the default date locale for current context and current environment.
	 *
	 * @return The locale corresponding to the date language of the environment if it is set, the
	 *         locale corresponding to the LY_LANGUAGE configuration parameter otherwise.
	 */
	public Locale getDateLocale()
	{
		return getLocale(getDateLanguage());
	}

	/**
	 * Gets the locale of the given language.
	 *
	 * @param  lang the language
	 * @return The locale of the given language
	 */
	protected Locale getLocale(String lang)
	{
		if (lang != null)
		{
			Locale locale;

			try
			{
				lang = lang.toUpperCase();

				// Check if we have a 2 caracter language code.
				// These codes are the lower-case two-letter codes as defined
				// by ISO-639
				if (lang.length() == 2)
					locale = new Locale(lang, lang);
				// Check if we have a 2 caracter country code.
				// These codes are the upper-case two-letter codes as defined
				// by ISO-3166
				else if (lang.length() == 5)
					locale = new Locale(lang.substring(0, 2), lang.substring(3, 5));
				else
				{
					// Try to get locale with the name of static fields defined
					// in
					// class java.util.Locale.
					Field field = Locale.class.getField(lang);

					locale = (Locale)field.get(null);
				}
			}
			catch (Exception e)
			{
				if (isDebugLevel(LyEnvironment.DEBUG_ERROR))
					out.println("[ERROR] Locale :" + getLanguage() + " was not found.");

				locale = Locale.getDefault();
			}
			return locale;
		}
		else
			// TO DO : read in .ini file or in session context.
			return (Locale.getDefault());
	} // end method getLocale

	// FFT 00037 / V3.3
	/**
	 * Gets the charset for the given language. Language values are the language abreviations (such
	 * as fr, en , etc.) as defined in the ISO 639-1 languages codes. If the resource with the given
	 * language is not found, searches in the resource without language.
	 *
	 * @param  lang           the language whose charset is required
	 * @param  nullIfNotFound if true, return null if the charset is not found, else return the
	 *                        default charset
	 * @return the charset if found, or null or the default charset (depending on parameter
	 *         nullIfNotFound)
	 */
	public String getCharset(String lang, boolean nullIfNotFound)
	{
		String charset = null;
		String langSuffix = null;

		//search in the language property
		if (lang != null)
		{
			langSuffix = "_" + lang.toUpperCase();
			String envVarKey = "CHARSET_ENCODING" + langSuffix;

			charset = getEnv(envVarKey);

			if (charset != null)
				return charset;

		}

		//if not found or no language defined, search in the generic property
		String envVarKey = "CHARSET_ENCODING";

		charset = getEnv(envVarKey);

		if (charset != null)
			return charset;
		if (nullIfNotFound)
			return null;
		else
			return LyString.DEFAULT_CHARSET;
	} // end method getCharset

	/**
	 * Retrieves the instance of LyString associated to this environment.
	 *
	 * @return A reference to the LyString instance.
	 * @see    leon.misc.LyString
	 */
	public LyString getString()
	{
		if (_string == null)
			_string = new LyString(this);

		return _string;
	}

	/**
	 * Emits a brief sound signal.
	 */
	public void beep()
	{
        LyViewManagerInterface viewManager = getViewManager();

        if (viewManager != null)
            viewManager.beep();
	}

	/**
	 * Gets the Images bundle associated to this environment.
	 *
	 * @return A reference to the LyImages instance.
	 * @see    leon.misc.LyImages
	 */
	public LyImages getImages()
	{
		if (_images == null)
			_images = new LyImages(this);

		return _images;
	}

	/**
	 * Clears all the resource bundle caches for this environment.
	 */
	public void clearResourceBundleCaches()
	{
		if (_strings != null)
			_strings.clearCache();

		if (_colors != null)
			_colors.clearCache();

		if (_images != null)
			_images.clearCache();

		if (_fonts != null)
			_fonts.clearCache();

		if (_files != null)
			_files.clearCache();

		if (_messages != null)
			_messages.clearCache();

		if (_resources != null)
			_resources.clearCache();

        if (_viewAttValues != null)
            _viewAttValues.clearCache();
	}

	/**
	 * Gets demand player.
	 *
	 * @return the demand player
	 */
	public LyPlayer getPlayer()
	{
		return _player;
	}

	/**
	 * Sets the demand player.
	 *
	 * @param player new demand player value
	 */
	public void setPlayer(LyPlayer player)
	{
		if (player == _player)
			return;

		if ((player != null) && (_player != null))
		    _player.free();

		_player = player;
		if (_environment != null)
			_environment.setPlayer(player);
	}

	/**
	 * Gets the default class loader of this environment.
	 *
	 * @return The default class loader for this environment or null if the system class loader is
	 *         used.
	 */
	public ClassLoader getClassLoader()
	{
		if ((_classLoader == null) && (_environment != null))
			return _environment.getClassLoader();

		return _classLoader;
	}

	/**
	 * Gets the environment logger used to log application messages.
	 *
	 * @return the environment logger.
	 */
	public LyLogger getLogger()
	{
		if (_logger != null)
			return _logger;

		_logger = buildLogger(false);

		if (_logger != null)
			return _logger;

		if ((_environment != null) && (_environment.out == out) && (_environment.err == err))
		{
			_logger = _environment.getLogger();
			return _logger;
		}

		// Create a default logger
		_logger = new LyLogger(this);
		return _logger;
	}

	/**
	 * Gets the perfomance logger used to log perf messages.
	 *
	 * @return the performance logge if any.
	 */
	public LyPerfLogger getPerfLogger()
	{
		if (_perfLogger != null)
			return _perfLogger;

		if (_environment != null)
			return _environment.getPerfLogger();

		return null;
	}

	/**
	 * Builds a logger for this environment.
	 *
	 * @param  useDefault To build a default logger if no specific class is specified.
	 * @return the new logger
	 */
	private LyLogger buildLogger(boolean useDefault)
	{
		LyLogger logger = null;
		String   loggerClassName = getEnv("_LY_LOGGER_CLASS");

		if (loggerClassName != null)
		{
			try
			{
				// Create a logger of given class
				ClassLoader classLoader = getClassLoader();
				Class<?>	loggerClass = ((classLoader != null) ?
				    classLoader.loadClass(loggerClassName) : Class.forName(loggerClassName));
				Class<?>[]  types = new Class[1];
				Object[]    parameters = new Object[1];

				types[0] = LyEnvironment.class;
				parameters[0] = this;

				Constructor<?> constructor = loggerClass.getConstructor(types);

				logger = (LyLogger)constructor.newInstance(parameters);
			}
			catch (InvocationTargetException e)
			{
				e.getTargetException().printStackTrace(err);
			}
			catch (Exception e)
			{
				e.printStackTrace(err);
			}
		}

		if ((logger == null)
				&& useDefault)
			logger = new LyLogger(this);

		if (logger != null)
		{
		    String    rootVersion = getString("LY_ROOT_VERSION");
		    String    leonVersion = ProductProperties.getFullName()+"\nV"+getString("LY_PRODUCT_VERSION");

		    if ((rootVersion != null)
		            && !rootVersion.equals(leonVersion))
		        logger.info("Application Version : " + getString(rootVersion));
			logger.info(ProductProperties.getFullName()+" Version : " + getString(leonVersion));
			logger.info("Java Version : " + System.getProperty("java.version"));
			logger.info("Operating system : " + System.getProperty("os.name"));
			logger.info("Ini. file : " + getEnv("_LY_ACTUAL_INI_FILE"));
		}

		return logger;
	} // end method buildLogger

	/**
	 * Sets the default class loader for this environment. All resources loaded before changing the
	 * class loader are NOT unload.
	 *
	 * @param loader The new default class loader.
	 */
	public void setClassLoader(ClassLoader loader)
	{
		_classLoader = loader;
	}

	/**
	 * Sets environment logger.
	 *
	 * @param logger The logger used to log application messages.
	 */
	public void setLogger(LyLogger logger)
	{
		_logger = logger;
	}

	/**
	 * Sets the skin and language for this environment.
	 *
	 * @param skin     The new skin.
	 * @param language The new language.
	 * @see   LyEnvironment#setLanguage()
	 * @see   LyEnvironment#setSkin(String)
	 */
	public void setSkinAndLanguage(String skin, String language)
	{
		String newLanguage = null;

		if (language != null)
			newLanguage = language.toLowerCase().intern();

		boolean needChange = false;

		if ((newLanguage != null) && (_language != newLanguage))
		{
			needChange = true;
			_language = newLanguage;
		}
		String newSkin = null;

		if (skin != null)
			newSkin = skin.toLowerCase().intern();

		if ((newSkin != null) && (newSkin.length() > 0) && (_skin != newSkin))
		{
			needChange = true;
			_skin = newSkin;
		}

		if (!needChange)
			return;

		// If the language changes we must reinitialize properties files.
		initResources();

		if ((_dynamicResources != null) &&_dynamicResources.size() > 0)
		{
			// Reload resources created dynamically
			Vector<_LyResource> dynamicResources = _dynamicResources;
			int				    i, n = dynamicResources.size();
			_LyResource		    resource;

			_dynamicResources = new Vector<_LyResource>(n);

			for (i = 0; i < n; i++)
			{
				resource = dynamicResources.elementAt(i);
				putResource(resource._type, resource._key, resource._value);
			}
		}

		if (_extraModels != null)
		{
			Vector<String> extraModels = _extraModels;
			String		   model;
			int			   i, n = extraModels.size();

			_extraModels = new Vector<String>(n);

			for (i = 0; i < n; i++)
			{
				model = extraModels.elementAt(i);
				load(model, null, false);
			}
		}
	} // end method setSkinAndLanguage

	/**
	 * Sets the language for this environment. Default values are defined in LANGUAGE_xxx constants.
	 *
	 * @param language The new language.
	 * @see   LyEnvironment#getLanguage()
	 */
	public void setLanguage(String language)
	{
		setSkinAndLanguage(_skin, language);
		setDateLanguage(language);
	} // end method setLanguage

	/**
	 * Set the skin for current environment.
	 *
	 * @param skin           the name of the skin to be used.
	 * @param initialization indicates if we are in the initialization process.
	 */
	public void setSkin(String skin)
	{
		setSkinAndLanguage(skin, _language);
	} // end method setSkin

	/**
	 * Sets the date language for this environment. Default values are defined in LANGUAGE_xxx
	 * constants.
	 *
	 * @param language The new date language.
	 * @see   LyEnvironment#getLanguage()
	 */
	public void setDateLanguage(String language)
	{
		_dateLanguage = language;
		_stringDateFormatTable = null;
		_dateFormatTable = null;
	}

	/**
	 * Sets current debug level for this environment. Default values are defined in DEBUG_xxx
	 * constants.
	 *
	 * @param level The debug level that must be set.
	 * @see   LyEnvironment#setDebugLevel(int,boolean)
	 */
	public void setDebugLevel(int level)
	{
		setDebugLevel(level, true);
	}

	/**
	 * Sets current debug level for this environment. Default values are defined in DEBUG_xxx
	 * constants.
	 *
	 * @param level  The debug level that must be set.
	 * @param status true if the debugLevel is added, false if removed.
	 * @see   LyEnvironment#setDebugLevel(int)
	 */
	public void setDebugLevel(int level, boolean status)
	{
		if (level < 0)
			return;

		if (level == DEBUG_NONE)
			// clear debug level
			_debugLevel = null;
		else
		{
			if (_debugLevel == null)
				_debugLevel = new BitSet();

			// checks if current debug level is set.
			if (status)
				_debugLevel.set(level);
			else
				_debugLevel.clear(level);

			if (status)
			{
				if (level == LyEnvironment.DEBUG_ALL)
				{
					for (int i = 1; i <= MAX_DEBUG; i++)
						_debugLevel.set(i);
				}
				else if (level == DEBUG_WARNING)
					_debugLevel.set(LyEnvironment.DEBUG_ERROR);
				else if (level == DEBUG_INFO)
				{
					_debugLevel.set(LyEnvironment.DEBUG_ERROR);
					_debugLevel.set(LyEnvironment.DEBUG_WARNING);
				}
				else if (level == DEBUG_TRACE)
				{
					_debugLevel.set(LyEnvironment.DEBUG_ERROR);
					_debugLevel.set(LyEnvironment.DEBUG_WARNING);
					_debugLevel.set(LyEnvironment.DEBUG_INFO);
				}
			}
		}

		if ((_logger != null) && (_logger.getEnvironment() == this))
			_logger.setDebugLevel(_debugLevel);
	} // end method setDebugLevel

	/**
	 * Execute the given command in a separate process, capture its output and redirect it to the
	 * standard output.
	 *
	 * @param command system script that should be executed
	 * @param waiting indicates whether the call is waiting for process completion
	 * @see   LyProcessViewer
	 */
	public void executeCommand(String command, boolean waiting)
	{
		new LyProcessViewer(this, command, waiting);
	}

	/**
	 * Execute the given command in a separate process, capture its output and redirect it to the
	 * standard output.
	 *
	 * @param commands    command list which should be executed.
	 * @param waiting indicates whether the call is waiting for process completion
	 * @see   LyProcessViewer
	 */
	public void executeCommand(String []command, boolean waiting)
	{
		new LyProcessViewer(this, command, waiting);
	}

	/**
	 * This method logs fatal errors.
	 *
	 * @param message Message to log.
	 */
	public void logFatal(String message)
	{
		getLogger().fatal(message);
	}

	/**
	 * This method logs errors.
	 *
	 * @param message Message to log.
	 */
	public void logError(String message)
	{
		if (isDebugLevel(LyEnvironment.DEBUG_ERROR))
			getLogger().error(message);
	}

	/**
	 * This method logs warnings.
	 *
	 * @param message Message to log.
	 */
	public void logWarning(String message)
	{
		if (isDebugLevel(LyEnvironment.DEBUG_WARNING))
			getLogger().warning(message);
	}

	/**
	 * This method logs info.
	 *
	 * @param message Message to log.
	 */
	public void logInfo(String message)
	{
		if (isDebugLevel(LyEnvironment.DEBUG_INFO))
			getLogger().info(message);
	}

	/**
	 * This method logs traces.
	 *
	 * @param message Message to log.
	 */
	public void logTrace(String message)
	{
		if (isDebugLevel(LyEnvironment.DEBUG_TRACE))
			getLogger().trace(message);
	}

	/**
	 * This method logs error messages.
	 *
	 * @param message Message to log.
	 */
	public void logErr(String message)
	{
		getLogger().logErr(message);
	}

	/**
	 * This method logs output messages.
	 *
	 * @param message Message to log.
	 */
	public void logOut(String message)
	{
		getLogger().logOut(message);
	}

	/**
	 * Writes an exception to the error stream. DEBUG_ERROR level must be set to produce the stack
	 * trace.
	 */
	public void writeException(Throwable t)
	{
		if (isDebugLevel(DEBUG_ERROR))
			getLogger().writeException(t);
	}

	/**
	 * Adds to the current environment resources of properties files corresponding to the given
	 * application id.
	 *
	 * @param applicationId The application id.
	 */
	public void load(String applicationId)
	{
		load(applicationId, null, true);
	}

	/**
	 * Adds to the current environment resources of properties files corresponding to the given
	 * application id.
	 *
	 * @param applicationId The application id.
	 * @param classLoader   The environment classLoader.
	 */
	public void load(String applicationId, ClassLoader classLoader)
	{
		load(applicationId, classLoader, true);
	}

	/**
	 * Adds to the current environment resources of properties files corresponding to the given
	 * application id.
	 *
	 * @param applicationId The application id.
	 * @param classLoader   The environment classLoader.
	 * @param checkParentForBundle indicates if parent bundles must be used during properties merge
	 */
	public void load(String applicationId, ClassLoader classLoader, boolean checkParentForBundle)
	{
		LyEnvironment							  env = new LyEnvironment(applicationId, this, false);
		Hashtable<ResourceType, LyResourceBundle> bundles = new Hashtable<ResourceType, LyResourceBundle>();
		String									  key, value;

		if (classLoader != null)
			env.setClassLoader(classLoader);
        if (_language != null)
            env._language = _language;
        if (_skin != null)
            env._skin = _skin;

        // Add extra resources
		env.initResources();
		bundles.put(ResourceType.STRING, env._strings);
		bundles.put(ResourceType.MESSAGE, env._messages);
		bundles.put(ResourceType.FILE, env._files);
		bundles.put(ResourceType.IMAGE, env._images);
		bundles.put(ResourceType.FONT, env._fonts);
		bundles.put(ResourceType.COLOR, env._colors);
		bundles.put(ResourceType.RESOURCE, env._resources);

        Enumeration<ResourceType>   types = bundles.keys();
        ResourceType                t;
        LyResourceBundle            bundle;
        Enumeration<String>         keys;

        while (types.hasMoreElements())
		{
            t = types.nextElement();
            bundle = bundles.get(t);
			if (bundle == null)
				continue;

			keys = bundle.getKeys(false);
			while (keys.hasMoreElements())
			{
				key = keys.nextElement();

				if (!checkResource(t, key, checkParentForBundle))
				//if (value == null)
				{
					value = env.getResource(t, key);
					if (value != null)
						putResource(t, key, value);
				}
			}
		}

        // Add applicationId as extra model
		if (_extraModels == null)
			_extraModels = new Vector<String>(1);
		_extraModels.add(applicationId);

		env.free();
	} // end method load

	/**
	 * Adds to the current environment resources of properties files corresponding to the given
	 * application id.
	 *
	 * @param applicationId The application id.
	 * @param classLoader   The environment classLoader.
	 */
	public void loadInParentBundles(String applicationId, ClassLoader classLoader)
	{
		LyEnvironment							  env = new LyEnvironment(applicationId, this, false);
		Hashtable<ResourceType, LyResourceBundle> bundles = new Hashtable<ResourceType, LyResourceBundle>();
		String									  key, value;

		if (classLoader != null)
			env.setClassLoader(classLoader);
        if (_language != null)
            env._language = _language;
        if (_skin != null)
            env._skin = _skin;

		env.initResources();
		bundles.put(ResourceType.STRING, env._strings);
		bundles.put(ResourceType.MESSAGE, env._messages);
		bundles.put(ResourceType.FILE, env._files);
		bundles.put(ResourceType.IMAGE, env._images);
		bundles.put(ResourceType.FONT, env._fonts);
		bundles.put(ResourceType.COLOR, env._colors);
		bundles.put(ResourceType.RESOURCE, env._resources);

        Enumeration<ResourceType>   types = bundles.keys();
        ResourceType                t;
        LyResourceBundle            bundle;
        Enumeration<String>         keys;

        while (types.hasMoreElements())
		{
            t = types.nextElement();
            bundle = bundles.get(t);
			if (bundle == null)
				continue;

			keys = bundle.getKeys(false);
			while (keys.hasMoreElements())
			{
				key = keys.nextElement();

				if (!checkResource(t, key, false))
				{
					value = env.getResource(t, key);
					if (value != null)
					{
						switch (t)
						{
						case STRING:
						{
							if ((_strings != null) && (_strings._parent != null))
									_strings._parent.put(key, value);
						}
						break;
						case MESSAGE:
						{
							if ((_messages != null) && (_messages._parent != null))
								_messages._parent.put(key, value);
						}
							break;
						case FILE:
						{
							if ((_files != null) && (_files._parent != null))
								_files._parent.put(key, value);
						}
							break;
						case IMAGE:
						{
							if ((_images != null) && (_images._parent != null))
								_images._parent.put(key, value);
						}
							break;
						case FONT:
						{
							if ((_fonts != null) && (_fonts._parent != null))
								_fonts._parent.put(key, value);
						}
							break;
						case COLOR:
						{
							if ((_colors != null) && (_colors._parent != null))
								_colors._parent.put(key, value);
						}
							break;
						case RESOURCE:
						{
							if ((_resources != null) && (_resources._parent != null))
								_resources._parent.put(key, value);
						}
						break;
						default:
							throw new IllegalArgumentException("BF AppEngine : unknown resource type, " + t);
						}
					}
				}
			}
		}

		if (_extraModels == null)
			_extraModels = new Vector<String>(1);
		_extraModels.add(applicationId);

		env.free();
	} // end method load

	/**
	 * Checks if events on data need to be reposted in GUI event queue.
	 *
	 * @return True if the event need to be reposted, false otherwise.
	 */
	public boolean repostGuiEvents()
	{
		LyViewManagerInterface viewManager = getViewManager();

		return ((viewManager == null) ? false : viewManager.repostGuiEvents());
	}

	/**
	 * Raises a warning message to the end-user.
	 *
	 * @param message error message
	 */
	public void raiseError(String message)
	{
		if (getViewManager() != null)
			getViewManager().raiseError(message);
		else
			getLogger().error(message);
	}

	/**
	 * Calls the {@link #execGui(Runnable, boolean) #execGui(runnable, false)} method of the given
	 * runnable to be invoked by the user-interface thread asynchronously. The caller of this method
	 * continues to run in parallel, and is not notified when the runnable has completed.
	 *
	 * @param runnable Code to run on the user-interface thread.
	 */
	public void execGui(Runnable runnable)
	{
		execGui(runnable, false);
	}

	/**
	 * If the view manager exist calls the view manager {@link
	 * LyViewManagerInterface#execGui(Runnable, boolean) execGui(runnable, synchronous)} method.
	 * Else, calls the given runnable {@link Runnable#run()} method to be invoked by the
	 * user-interface thread either synchronously or asynchronously. If asynchronous, the caller of
	 * this method continues to run in parallel, and is not notified when the runnable has
	 * completed. Otherwise, the thread which calls this method is suspended until the runnable
	 * completes.
	 *
	 * @param runnable    Code to run on the user-interface thread.
	 * @param synchronous True is synchronous, false if asynchronous.
	 */
	public void execGui(Runnable runnable, boolean synchronous)
	{
		if (_viewManager != null)
			_viewManager.execGui(runnable, synchronous);
		else
			runnable.run();
	}

	/**
	 * Change the default System.out if it was not already set before.
	 *
	 * @param out     new output stream
	 * @param restore if true, force restore to tghe specified System.out
	 */
	protected void setSystemOut(PrintStream outStream, boolean restore)
	{
		if ((restore) || (!"true".equals(System.getProperty("leon.system.out.modified"))))
		{
			if (!restore)
			{
				_oldSystemOut = System.out;
				System.setProperty("leon.system.out.modified", "true");
			}
			else
				System.setProperty("leon.system.out.modified", "false");
			System.setOut(outStream);
		}
	}

	protected void setSystemErr(PrintStream errStream, boolean restore)
	{
		if ((restore) || (!"true".equals(System.getProperty("leon.system.err.modified"))))
		{
			if (!restore)
			{
				_oldSystemErr = System.err;
				System.setProperty("leon.system.err.modified", "true");
			}
			else
				System.setProperty("leon.system.err.modified", "false");
			System.setErr(errStream);
		}
	}

	/**
	 * Accessor to the properties file or parameter if specified in constructor.
	 *
	 * @param prefix    Prefix used for properties files name.
	 * @return the properties.
	 */
	public Properties getPropertiesFromContext(String prefix)
	{
		LyViewManagerInterface viewManager = getViewManager();

		return (viewManager == null ? null : viewManager.getPropertiesFromContext(prefix));
	}

	/**
	 * Gets image handler.
	 */
	protected LyImageHandlerInterface getImageHandler()
	{
        if (_imageHandler  != null)
            return _imageHandler;
        else if (_environment != null)
            return _environment.getImageHandler();
        else
            return null;
	}

	/**
	 * Internal class to store a dynamic resource.
	 */
	class _LyResource
	{
		/** Resource type. */
		ResourceType  _type;

		/** Dynamic resource is stored with this key. */
		String _key;

		/** Value of the dynamic resource to store. */
		String _value;

		/**
		 * Creates a new resource to store with the given type, key and value.
		 *
		 * @param type
		 * @param key
		 * @param value
		 */
		_LyResource(ResourceType type, String key, String value)
		{
			_type = type;
			_key = key;
			_value = value;
		}
	}

	public void setMobile(boolean isMobile)
	{
		_isMobile = isMobile;
	}
} // end class LyEnvironment
