/*
 * LyFileTools.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.util.*;


/**
 * This class offers general useful methods to manipulate files content.
 *
 * @version  $Id: LyFileTools.java 41849 2012-12-26 14:29:55Z bangiole $
 * @author   Lyria
 */
public class LyFileTools
{
    public static final String __VERSION = "$Id: LyFileTools.java 41849 2012-12-26 14:29:55Z bangiole $";

    /** Current environment. */
    protected LyEnvironment _environment;

    protected static Map<String, String> _mimeTypesFromExt;
    protected static Map<String, String> _extFromMimeType;

    /**
     * Constructor.Only initializes the current environment with the given one.
     *
     * @param  environment  The environment that must be used.
     */
    public LyFileTools(LyEnvironment environment)
    {
        _environment = environment;
    }

    /**
     * Replaces all keys found in the <code>in</code> Buffer with the equivalent String value found
     * in the <code>equivalent</code> Hashtable.<br />
     * Keys must be of this format: $KEY$. Other keys can be used in the replacement String. Keys
     * cannot be cut between two lines. If a key is find in a line but not in the hashtable, it is
     * replaced with an empty string.
     *
     * @param   environment  The environment within the operation occured.
     * @param   out          The output file
     * @param   in           The input file
     * @param   equivalent   The hastable Key->Replacement String.
     *
     * @return  true if everything was ok, false otherwise.
     */
    public static boolean replaceKey(LyEnvironment environment, Writer out, BufferedReader in,
        Hashtable<String, String> equivalent)
    {
        String line;
        int    first, next;
        String subString;
        String replaceString;

        if ((out == null) || (in == null) || (equivalent == null))
            return false;

        try
        {
            line = in.readLine();

            while (line != null)
            {
                // find a '$'
                first = line.indexOf('$');

                while (first >= 0)
                {

                    next = line.indexOf('$', first + 1);

                    if (next >= 0)
                    {
                        // we have something to replace
                        subString = line.substring(first, next + 1);

                        replaceString = equivalent.get(subString);

                        if (replaceString == null)
                            replaceString = "";

                        line = LyString.replace(line, subString, replaceString);
                    }

                    first = line.indexOf('$');
                }

                out.write(line + "\n");

                line = in.readLine();
            }
        }
        catch (IOException e)
        {
            environment.writeException(e);
            return false;
        }

        return true;
    } // end method replaceKey

    /**
     * Extracts the filename from the specified PATH.
     *
     * @param   path  input path
     *
     * @return  the filename only
     */
    public static String getFileName(String path)
    {
        if (path == null)
            return null;

        path = path.replace('\\', '/');

        int n1 = path.lastIndexOf('/');

        if (n1 >= 0)
            return path.substring(n1 + 1);
        else
            return path;
    }

    /**
     * Extracts the suffix (like .gif .doc) from the specified PATH.
     *
     * @param   path  input path
     *
     * @return  the suffix only or null if it could not be found
     */
    public static String getFileSuffix(String path)
    {
        String fileName = getFileName(path);

        int n1 = fileName.lastIndexOf('.');

        if (n1 >= 0)
            return fileName.substring(n1);
        else
            return null;
    }

    /**
     * Extracts the suffix (like .gif .doc) from the specified PATH
     * @param path input path
     * @return the suffix only or null if it could not be found
     */
    public static String removeFileCounter(String path)
    {
    	String fileName = getFileName(path);

    	int n1 = fileName.lastIndexOf('.');
    	if (n1 >=0)
    	{
    		int n2 = fileName.indexOf('_');

    		if (n2 < 0)
    			return fileName;
    		else
    		{
    			StringBuffer sb = new StringBuffer();

    			sb.append(fileName.substring(0, n2)).append(fileName.substring(n1));

    			return sb.toString();
    		}
    	}
    	return fileName;
    }

// FFT00037 / V3.3
    /**
     * Encode the input file in the given output file with the given charset.
     *
     * @param   environment      the application environment
     * @param   encodingCharset  the charset used to encode the input file
     * @param   inputFileName    the input file name
     * @param   outputFileName   the output file name. If null the result of encoding is redirect to
     *                           the current output stream.
     *
     * @return  an error message if any, null otherwise
     */
    public static String encodeFile(LyEnvironment environment, String encodingCharset, String inputFileName,
        String outputFileName)
    {
        return native2ascii(environment, false, encodingCharset, inputFileName, outputFileName);
    }

    // FFT00037 / V3.3
    /**
     * Decode the input file in the given output file with the given charset.
     *
     * @param   environment      the application environment
     * @param   encodingCharset  the input file charset used to decode it
     * @param   inputFileName    the input file name
     * @param   outputFileName   the output file name. If null the result of encoding is redirect to
     *                           the current output stream.
     *
     * @return  an error message if any, null otherwise
     */
    public static String decodeFile(LyEnvironment environment, String encodingCharset, String inputFileName,
        String outputFileName)
    {
        return native2ascii(environment, true, encodingCharset, inputFileName, outputFileName);
    }

    /**
     * Method to use the native2ascii Java tool.<br />
     * A <b>JDK 1.5</b> or more must be specified in the variable: INI_JDK_HOME of the application
     * ini file.<br/>
     *
     * <blockquote>Java tool synopsis :<br />
     * <code>native2ascii [-reverse] [-encoding encoding_name] [inputfile [outputfile]]</code><br />
     *
     * <blockquote><code>-reverse</code>:<br />
     *
     * <blockquote>Perform the reverse operation: convert a file with Latin-1 and/ or Unicode
     * encoded characters to one with native-encoded characters.<br />
     * </blockquote>
     * <code>-encoding encoding_name</code> :<br />
     *
     * <blockquote>Specify the encoding name which is used by the conversion procedure. The default
     * encoding is taken from System property file.encoding. The encoding_name string must be taken
     * from the first column of the table of supported encodings in the Supported Encodings
     * document.<br />
     * </blockquote>
     * <code>inputfile</code>:<br />
     *
     * <blockquote>The input file name. If, in addition, inputfile is omitted, standard input is
     * used for input.<br />
     * </blockquote>
     * <code>outputfile</code>:<br />
     *
     * <blockquote>The output file name. If it is omitted, standard output is used for output.<br />
     * </blockquote>
     * </blockquote>
     * </blockquote>
     *
     * @param   environment      the application environment
     * @param   reverse          indicates if the reverse operation has to be preformend or not
     * @param   encodingCharset  the encoding charset
     * @param   inputFileName    the input file name. If null, this method does nothing
     * @param   outputFileName   the input file name. If null, write the result in the standard
     *                           output.
     *
     * @return  an error message if any, null otherwise
     */
    @SuppressWarnings("deprecation")
	public static String native2ascii(LyEnvironment environment, boolean reverse, String encodingCharset,
        String inputFileName, String outputFileName)
    {
        // if there is no input file to convert, return false
        if ((inputFileName == null) || (inputFileName.trim().length() <= 0))
            return null;

        String givenInputFileName = inputFileName;

        // For logs :
        String encodeOrDecode = null;

        if (reverse)
            encodeOrDecode = "decode";
        else
            encodeOrDecode = "encode";

        // if the inputfile is the same as the outputfile, create a tmp input
        File tmpOutFile = null;
        File tmpInFile = null;

        if ((outputFileName != null) && outputFileName.equals(inputFileName))
        {
            try
            {
                tmpInFile = File.createTempFile("tmpInFile", ".tmp");
                File inFile = new File(givenInputFileName);

                // copy file into tmp file
                tmpInFile.delete();
                inFile.renameTo(tmpInFile);
                inputFileName = tmpInFile.getAbsolutePath();

                tmpOutFile = File.createTempFile("tmpOutFile", ".tmp");
                outputFileName = tmpOutFile.getAbsolutePath();
            }
            catch (IOException e)
            {
                e.printStackTrace();
            }
        }

        try
        {
            // Try to get Java home
            String	jdkHome = null;
            String	propJdkHome = environment.getEnv("JAVA_JDK_HOME");

            if (propJdkHome != null)
            {
                if (propJdkHome.endsWith("/") || propJdkHome.endsWith("\\"))
                    jdkHome = propJdkHome + "jre";
                else
                    jdkHome = propJdkHome + File.separator + "jre";
            }

            if ((jdkHome == null)
            		|| (jdkHome.trim().length() == 0))
                jdkHome = System.getProperty("java.home");

            if (jdkHome == null)
                throw new ClassNotFoundException();

            environment.logTrace("Native2ascii found jdk home:" + jdkHome);

            File		   java_tools = new File(jdkHome + "/../lib/tools.jar");
            URL  		   url_java_toolsJar = java_tools.toURL();
            URL[]          urls = new URL[] { url_java_toolsJar };
            URLClassLoader newClassLoader = new URLClassLoader(urls);
            Class<?>   	   n2aMain = newClassLoader.loadClass("sun.tools.native2ascii.Main");
            Class<?>[]     paramTypes = new Class[] { String[].class };

            // Build arguments
            Vector<String> arguments = new Vector<String>();
            String         commandLine = "native2ascii";

            if (reverse == true)
            {
                arguments.add("-reverse");
                commandLine = commandLine + " -reverse";
            }

            if (encodingCharset != null)
            {
                arguments.add("-encoding");
                arguments.add(encodingCharset);

                commandLine = commandLine + " -encoding " + encodingCharset;
            }

            arguments.add(inputFileName);
            commandLine = commandLine + " " + inputFileName;

            if (outputFileName != null)
            {
                arguments.add(outputFileName);
                commandLine = commandLine + " " + outputFileName;
            }

            String[] args = new String[arguments.size()];

            arguments.toArray(args);

            Method convert = n2aMain.getMethod("convert", paramTypes);

            Object  o = n2aMain.newInstance();
            Boolean resultB = (Boolean)convert.invoke(o, new Object[] { args });
            boolean result = resultB.booleanValue();

            o = null;
            convert = null;
            n2aMain = null;
            args = null;

            if (!result)
                return "Could not " + encodeOrDecode + " file " + givenInputFileName;

            if (tmpOutFile != null)
            {
                System.gc();
                System.gc();

                if ((tmpInFile != null) && tmpInFile.exists())
                {
                    boolean deleteOk = tmpInFile.delete();

                    if (!deleteOk)
                        tmpInFile.deleteOnExit();
                }

                File givenInFile = new File(givenInputFileName);

                if (givenInFile.exists())
                    givenInFile.delete();

                tmpOutFile.renameTo(givenInFile);
            }

            environment.logTrace("Native2ascii " + encodeOrDecode + " file '" + givenInputFileName + "', process='"
                + commandLine + "'.");

            return null;
        }
        catch (ClassNotFoundException e)
        {
            if (tmpOutFile != null)
            {
                System.gc();
                System.gc();

                boolean deleteOk = tmpOutFile.delete();

                if (!deleteOk)
                    tmpOutFile.deleteOnExit();

                if ((tmpInFile != null) && tmpInFile.exists())
                {
                    deleteOk = tmpInFile.delete();
                    if (!deleteOk)
                        tmpInFile.deleteOnExit();
                }
            }

            String message = "Could not " + encodeOrDecode + " file " + givenInputFileName + "\n"
                + "Property 'INI_JDK_HOME=JDK_Home_path' must be specified in application .ini file";

            return message;
        }
        catch (Exception e)
        {
            if (tmpOutFile != null)
            {

                System.gc();
                System.gc();

                boolean deleteOk = tmpOutFile.delete();

                if (!deleteOk)
                    tmpOutFile.deleteOnExit();

                if ((tmpInFile != null) && tmpInFile.exists())
                {
                    deleteOk = tmpInFile.delete();
                    if (!deleteOk)
                        tmpInFile.deleteOnExit();
                }
            }

            return "Could not " + encodeOrDecode + " file " + givenInputFileName;
        }
    }

    /**
     * find a probable mimetype from file extension.
     *
     * @param   fileName  fine path
     *
     * @return  the mimetype if found or null if non valid arguments.
     */
    public static String getMimeType(String fileName)
    {
        if (fileName == null)
            return null;

        String ext = getFileSuffix(fileName);

        String mimeType = null;

        if (ext != null)
        {
            if (ext.startsWith("."))
                ext = ext.substring(1);

            initMimeTypes();

            mimeType = _mimeTypesFromExt.get(ext.toLowerCase());
        }

        if (mimeType == null)
            return "application/octet-stream";
        else
            return mimeType;
    }

    /**
     * find a probable from file extension from given mimetype.
     *
     * @param   the  mimetype if found or null if non valid arguments.
     *
     * @return  the file ext if any
     */
    public static String getSuffixFromMimeType(String mimeType)
    {
        if (mimeType == null)
            return null;

        initMimeTypes();

        String ext = _extFromMimeType.get(mimeType);

        if ((ext == null) && (mimeType.startsWith("application/x-")))
            ext = mimeType.substring("application/x-".length());

        return ext;
    }

    /**
     * <p> Creates a new empty file in the specified directory, using the
     * given prefix and suffix strings to generate its name.
     * @param  prefix     The prefix string to be used in generating the file's
     *                    name; must be at least three characters long
     *
     * @param  suffix     The suffix string to be used in generating the file's
     *                    name; may be <code>null</code>, in which case the
     *                    suffix <code>".tmp"</code> will be used
     *
     * @param  directory  The directory in which the file is to be created, or
     *                    <code>null</code> if the default temporary-file
     *                    directory is to be used
     * @param  keepFileName  Try to keep filename if possible, or
     *
     * @return  An abstract pathname denoting a newly-created empty file
     *
     * @throws  IllegalArgumentException
     *          If the <code>prefix</code> argument contains fewer than three
     *          characters
     *
     * @throws  IOException  If a file could not be created
     *
     */
    public File createTempFile(String prefix, String suffix, File directory,
    		boolean keepFileName) throws IOException
    {
    	File f = null;

    	if (keepFileName)
    	{
    		// if not provided try to detect current temporary directory
    		if (directory == null)
    		{
    			try
    			{
    				String tempDir = (_environment == null ? null : _environment.getFile("LY_TEMP_DOC_DIR"));

    				if (tempDir == null || tempDir.equals("LY_TEMP_DOC_DIR") || tempDir.length() <= 0)
    					tempDir = System.getProperty("java.io.tmpdir");

    				if (tempDir != null && _environment != null)
    					tempDir = _environment.getString().expand(tempDir);
    				if (tempDir != null)
    					directory = new File(tempDir);

    			}
    			catch (Throwable e)
    			{
    			}
    		}
    		if (directory == null)
    		{
    			try
    			{
    				File f1 = File.createTempFile(prefix, suffix, directory);

    				if (f1 != null)
    					directory = f1.getParentFile();
    			}
    			catch (Throwable e1)
    			{
    			}
    		}
    		// find out how long we should keep a temp file
    		for (int counter=0; (f == null) && (directory != null) && (counter < 50); counter++)
    		{
    			// build the temporary filePath
    			StringBuffer path = new StringBuffer();
    			path.append(directory.getAbsolutePath()).append(File.separator).append(prefix);

    			if (counter > 0)
    				path.append('_').append(counter); // if a filename already exists, increment a filename counter
    			path.append(suffix);

    			// Build the file and check if it exists
    			File f1 = new File(path.toString());

    			if (f1.exists() && f1.canRead() && f1.canWrite())
    			{
    				// maybe the file is old
    				try
    				{
    					f1.delete();
    				}
    				catch (Throwable t)
    				{
    				}
    			}
    			// if the file is new we can keep it, otherwise next try with counter
    			if (!f1.exists())
    			{
    				f = f1;
    				f.deleteOnExit();
    			}
    		}
    	}

    	if (f == null)
    	{
    		f = File.createTempFile(prefix, suffix, directory);
    	}
    	if (f != null && _environment != null)
    		_environment.logTrace("Leonardi : create temp file "+f.getAbsolutePath());

    	return f;
    }

    /**
     * <p> compare if tempFiles could assumed to be the same
     * @param  String fileName1 first file
     * @param  String fileName2 second file
     *
     * @return  true if files are equivalent
     */
    public static boolean compareTempFiles(String fileName1, String fileName2)
    {
    	if (fileName1 == fileName2)
    		return true;
    	if (fileName1 == null || fileName2 == null)
    		return false;
    	if (fileName1.equals(fileName2))
    		return true;
    	// check if the filenames could fit
    	String f1 = removeFileCounter(fileName1);
    	String f2 = removeFileCounter(fileName2);

    	if (f1.equals(f2))
    	{
    		// same root file name
    		File file1 = new File(fileName1);
    		File file2 = new File(fileName2);

    		if (file1.exists() && file2.exists())
    		{
    			if (f1.length() == f2.length())
    				return true;
    		}
    	}


    	return false;
    }

    /**
     * Internal initialization of mime types (defaults).
     */
    protected static void initMimeTypes()
    {
        if (_mimeTypesFromExt == null)
        {
        	HashMap<String, String>	mimeTypesFromExt = new HashMap<String, String>();

            mimeTypesFromExt.put("jar", "application/java-archive");
            mimeTypesFromExt.put("xls", "application/excel");
            mimeTypesFromExt.put("ppt", "application/mspowerpoint");
            mimeTypesFromExt.put("pps", "application/mspowerpoint");
            mimeTypesFromExt.put("doc", "application/msword");

			// Mimetypes OPENXML -> do not remove should be restore one day
//			mimeTypesFromExt.put("xlsx", "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet");
//			mimeTypesFromExt.put("pptx", "application/vnd.openxmlformats-officedocument.presentationml.presentation");
//			mimeTypesFromExt.put("ppsx", "application/vnd.openxmlformats-officedocument.presentationml.slideshow");
//			mimeTypesFromExt.put("docx", "application/vnd.openxmlformats-officedocument.wordprocessingml.document");
			mimeTypesFromExt.put("xlsx", "application/excel");
			mimeTypesFromExt.put("pptx", "application/mspowerpoint");
			mimeTypesFromExt.put("ppsx", "application/mspowerpoint");
			mimeTypesFromExt.put("docx", "application/msword");

			mimeTypesFromExt.put("pdf", "application/pdf");
            mimeTypesFromExt.put("ps", "application/postscript");
            mimeTypesFromExt.put("rtf", "application/rtf");
            mimeTypesFromExt.put("gz", "application/x-gzip");
            mimeTypesFromExt.put("js", "application/x-javascript");
            mimeTypesFromExt.put("tar", "application/x-tar");
            mimeTypesFromExt.put("rar", "application/x-rar");
            mimeTypesFromExt.put("zip", "application/zip");
            mimeTypesFromExt.put("mp3", "audio/mpeg");
            mimeTypesFromExt.put("gif", "image/gif");
            mimeTypesFromExt.put("jpg", "image/jpeg");
            mimeTypesFromExt.put("jpeg", "image/jpeg");
            mimeTypesFromExt.put("png", "image/png");
            mimeTypesFromExt.put("tiff", "image/tiff");
            mimeTypesFromExt.put("css", "text/css");
            mimeTypesFromExt.put("html", "text/html");
            mimeTypesFromExt.put("txt", "text/plain");
            mimeTypesFromExt.put("xml", "text/xml");

            mimeTypesFromExt.put("mpeg", "video/mpeg");
            mimeTypesFromExt.put("mov", "video/quicktime");
            mimeTypesFromExt.put("avi", "video/x-msvideo");
            _mimeTypesFromExt = mimeTypesFromExt;
        }

        if (_extFromMimeType == null)
        {
        	HashMap<String, String>	extFromMimeType = new HashMap<String, String>();

            for (Iterator<String> iter = _mimeTypesFromExt.keySet().iterator(); iter.hasNext();)
            {
                String key = iter.next();

                extFromMimeType.put(_mimeTypesFromExt.get(key), key);
            }

            _extFromMimeType = extFromMimeType;
        }
    }

} // end class LyFileTools
