/*
 * LyFonts.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

/**
 * The <code>LyFont</code> class is used encapsulate fonts used by Leonardi applications.
 */
public final class LyFont
{

	public static final String __VERSION = "$Id: LyFont.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Font name. */
	String					   _name;

	/** Font size. */
	int						   _size;

	/** Boolean indicating if the font is in bold style or not. */
	boolean					   _bold;

	/** Boolean indicating if the font is in italic style or not. */
	boolean					   _italic;

	/**
	 * Constructor. Creates a font from a resource string.
	 * @param fontString the String representation of the font. The format of this string is:
	 *                   fontName[-(bold|italic)][-(bold|italic)][-size]<br />
	 *                   For example:
	 *                   <ul>
	 *                     <li>Courier-bold-10</li>
	 *                     <li>Arial-bold-italic whitch is equivalent to Arial-italic-bold</li>
	 *                     <li>Times</li>
	 *                   </ul>
	 *                   <p>Default name is "Courier", default size is "10".</p>
	 */
	public LyFont(String fontString)
	{
		String[] list = LyString.stringToList(fontString, '-');
		int		 n = ((list == null) ? 0 : list.length);

		_name = ((n > 0) ? list[0] : "Courier");

		_bold = (((n > 1) && list[1].equals("bold")) || ((n > 2) && list[2].equals("bold")));

		_italic = (((n > 1) && list[1].equals("italic")) || ((n > 2) && list[2].equals("italic")));

		int p = 1;

		if (_bold)
			p++;
		if (_italic)
			p++;

		try
		{
			_size = ((n > p) ? Integer.parseInt(list[p]) : 10);
		}
		catch (NumberFormatException e)
		{
			_size = 10;
		}
	}

	/**
	 * Gets font name.
	 * @return the font name
	 */
	public String getFontName()
	{
		return _name;
	}

	/**
	 * Gets font family name.
	 * @return the font family name
	 */
	public String getFamily()
	{
		return getFontName(); // XXX
	}

	/**
	 * Gets the font size.
	 * @return the font size
	 */
	public int getSize()
	{
		return _size;
	}

	/**
	 * Checks if the font is bold or not.
	 * @return true is the font is bold, false otherwise
	 */
	public boolean isBold()
	{
		return _bold;
	}

	/**
	 * Checks if the font is italic or not.
	 * @return true is the font is italic, false otherwise
	 */
	public boolean isItalic()
	{
		return _italic;
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object obj)
	{
		if ((obj == null) || !(obj instanceof LyFont))
			return false;
		
		LyFont		font = (LyFont)obj;
		
		boolean	result = ((font._size == _size)
				&& (font._bold == _bold)
				&& (font._italic == _italic)
				&& (font._name != null)
				&& (font._name.equals(_name)));
		
		return result;
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode()
	{
		return toString().hashCode();
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString()
	{
		String	result = "LyFont (" + getValue() + ")";
		return result;
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String getValue()
	{
		String	result =  _name ;
		
		if (_bold)
			result += "-bold";
		
		if (_size>0)
			result += "-" + _size;
		
		
		if (_italic)
			result += "-italic";
		
		return result;
	}
	
} // end class LyFont
