/*
 * LyFonts.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.util.*;


/**
 * This class is used to manage the font resources of an application. It is based on the mechanism
 * of LyResourceBundle.
 * @version $Id: LyFonts.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyFonts extends LyResourceBundle
{
	public static final String __VERSION = "$Id: LyFonts.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Default font key. */
	public static String	   DEFAULT_FONT;

	static
	{
		DEFAULT_FONT = "LY_DEFAULT_FONT".intern();
	}

	/** Table of fonts used as a cache. */
	protected Hashtable<String, LyFont> _fonts;

	/**
	 * Constructor with a given environment. Calls {@link #LyFonts(LyEnvironment, String)
	 * LyFonts(environment, null)} constructor.
	 * @param environment The environment of the fonts bundle.
	 */
	public LyFonts(LyEnvironment environment)
	{
		this(environment, null);
	}

	/**
	 * Constructor with a given environment and name. Calls {@link #LyFonts(LyEnvironment, String,
	 * Locale) LyFonts(environment, name, environment.getLocale())}.
	 * @param environment The environment of the fonts bundle.
	 * @param name        The name of the fonts bundle.
	 */
	public LyFonts(LyEnvironment environment, String name)
	{
		this(environment, name, environment.getLocale(), environment.getSkin());
	}

	/**
	 * Constructor with a given environment and name. Calls {@link #LyFonts(LyEnvironment, String,
	 * Locale) LyFonts(environment, name, environment.getLocale())}.
	 * @param environment The environment of the fonts bundle.
	 * @param name        The name of the fonts bundle.
	 * @param skin        The name of the skin
	 */
	public LyFonts(LyEnvironment environment, String name, String skin)
	{
		this(environment, name, environment.getLocale(), skin);
	}

	/**
	 * Constructor with a given environment, name and locale. Calls the super implementation.
	 * @param environment The environment of the strings bundle.
	 * @param name        The name of the strings bundle.
	 * @param locale      The locale used by the strings bundle.
	 */
	protected LyFonts(LyEnvironment environment, String name, Locale locale)
	{
		this(environment, name, locale, environment.getSkin());
	}
	
	/**
	 * Constructor with a given environment, name and locale. Calls the super implementation.
	 * @param environment The environment of the fonts bundle.
	 * @param name        The name of the fonts bundle.
	 * @param locale      The locale used by the fonts bundle.
	 * @param skin        The name of the skin
	 */
	protected LyFonts(LyEnvironment environment, String name, Locale locale, String skin)
	{
		super(environment, (name == null ? "fonts" : name), locale, skin);
		
	      _fonts = null;
	}

	/**
	 * If the bundle environment is not null, builds a parent fonts bundle.
	 * @return The parent fonts bundle for this fonts bundle, or null if the environment is null.
	 */
	protected LyResourceBundle buildParentBundle()
	{
		if (_environment == null)
			return null;

		return new LyFonts(_environment.getParent(), _name, _locale, _skin);
	}

	/**
	 * Gets the Font corresponding to the given key.
	 * @param  key The font key.
	 * @return The Font corresponding to the given key if found, otherwise creates a new default
	 *         font, adds it to the {@link #_fonts} list and return it.
	 */
	public LyFont getFont(String key)
	{
		if (key == null)
			key = DEFAULT_FONT;
		
		LyFont result = null;

		if (_fonts != null)
		{
			result = _fonts.get(key);
			if (result != null)
				return result;
		}

		// Retrieve the value associated with the key
		String value = get(key);

		if (value == null)
			value = key;
		else if ((value.length() > 0) && (value.charAt(0) == '$'))
			return getFont(value.substring(1));

		// Create Java Font
		result = new LyFont(value);

		// Cache the result.
		if (_fonts == null)
			_fonts = new Hashtable<String, LyFont>(10);
		_fonts.put(key, result);

		return result;
	} // end method getFont
} // end class LyFonts
