/*
 * LyHelpFiles.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.io.*;
import java.util.*;


/**
 * This class is used to manage access to application help files. Programs attempts to find in help
 * directory : help/lang/classId/actionId.html then help/lang/classId/rootActionId.html then
 * help/lang/actionId.html then help/lang/rootActionId_XX.html which is the default leonardi file.
 * If file could not be found, file are searched with no lang information.
 * @version $Id: LyHelpFiles.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyHelpFiles
{
	public static final String __VERSION =
		"$Id: LyHelpFiles.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Dummy String which indicates that Help file could not be found. */
	public static final String FILE_DOES_NOT_EXIST = "File does dot exists...".intern();

	/** Dummy String which indicates the help dir hashkey. */
	public static final String HELP_DIR = "HELP_DIR_HASHKEY".intern();

	/** Default help files extension. */
	public static String	   HELP_EXTENSION = ".html";

	/** Table of help files paths (keys are parameters concatenation). */
	protected Hashtable<String, String>		   _helpFiles;

	/** Current environment. */
	protected LyEnvironment    _environment;

	/**
	 * Constructor with a given environment. Creates a new {@link #_helpFiles} and sets the
	 * environment with the given value.
	 * @param environment Current environment.
	 */
	public LyHelpFiles(LyEnvironment environment)
	{
		_helpFiles = new Hashtable<String, String>();
		_environment = environment;
	}

	/**
	 * Gets the help file for the specified context, or null if help file could not be found.
	 * @param  actionId     Current action identifier (attempt to get help for the specific action)
	 * @param  rootActionId Root action identifier
	 * @param  classId      Class id identifier.
	 * @return The help file path, or null if it could not be found.
	 */
	public String getHelpFile(String actionId, String rootActionId, String classId)
	{
		// Check if the help directory may exist (otherwise, no need to go
		// further)
		String path = _helpFiles.get(HELP_DIR);
        
		if (path == null)
		{
			path = getHelpFile("/help");

			if (path == null) // Does not exists !
			{
				path = FILE_DOES_NOT_EXIST;
				_helpFiles.put(HELP_DIR, FILE_DOES_NOT_EXIST);
			}
			else
				_helpFiles.put(HELP_DIR, path);
		}
		if (path == FILE_DOES_NOT_EXIST)
			return null;

		// Attempt to fetch from cache
		StringBuffer buf = new StringBuffer();
		String		 hashKey;

		if (actionId != null)
			buf.append(actionId).append('/');

		if (rootActionId != null)
			buf.append(rootActionId).append('/');

		if (classId != null)
			buf.append(classId);

		// Can not build key : parameters are all null !
		if (buf.length() <= 0)
			return null;

		hashKey = buf.toString().intern();
		path = _helpFiles.get(hashKey);

		if (path != null)
		{
			// We have the result in cache.
			if (path == FILE_DOES_NOT_EXIST)
				return null;
			else
				return path;
		}

		// This is the first attempt to find the help file.
		// Build all possible filenames and try...

		String resu;
		Locale locale = _environment.getLocale();

		// Try with current language
		resu = getHelpFile(actionId, rootActionId, classId, locale.getCountry());
		// Retry without language
		if (resu == null)
			resu = getHelpFile(actionId, rootActionId, classId, null);

		// Not found whatever...
		if (resu == null)
		{
			// Keep failure in cache : shall not try next time
			_helpFiles.put(hashKey, FILE_DOES_NOT_EXIST);

			return null;
		}
		else
		{
			// Keep result in cache : shall not try next time
			_helpFiles.put(hashKey, resu);

			return resu;
		}
	} // end method getHelpFile

	/**
	 * Gets the help file for the specified context, or null if help file could not be found.<br />
	 * Program attempts to find files in help directory in this order:
	 * <ol>
	 *   <li>help/lang/classId/actionId.html</li>
	 *   <li>help/lang/classId/rootActionId.html</li>
	 *   <li>help/lang/actionId.html</li>
	 *   <li>help/lang/rootActionId_XX.html which is the default leonardi file.</li>
	 * </ol>
	 * @param  actionId     Current action identifier (attempt to get help for the specific action)
	 * @param  rootActionId Root action identifier
	 * @param  classId      Class id identifier.
	 * @param  lang         Current language, may be null
	 * @return The help file path,or null if it could not be found.
	 */
	protected String getHelpFile(String actionId, String rootActionId, String classId, String lang)
	{
		String rootPath, path;
		String resu = null;

		if ((lang == null) || (lang.length() == 0))
			rootPath = "/help/";
		else
			rootPath = "/help/" + lang.toLowerCase() + "/";

		// Try /help/lang/classId/index
		if ((actionId == null) && (classId != null))
		{
			path = rootPath + classId + "/index" + HELP_EXTENSION;
			resu = getHelpFile(path);

			if (resu != null)
				return resu;
		}

		// Try /help/lang/classId/actionId
		if ((actionId != null) && (classId != null))
		{
			path = rootPath + classId + "/" + actionId + HELP_EXTENSION;
			resu = getHelpFile(path);

			if (resu != null)
				return resu;
		}
		// Try /help/lang/classId/rootActionId
		if ((rootActionId != null) && (classId != null) && (!rootActionId.equals(actionId)))
		{
			path = rootPath + classId + "/" + rootActionId + HELP_EXTENSION;

			resu = getHelpFile(path);

			if (resu != null)
				return resu;
		}

		// Try /help/lang/index
		if (actionId == null)
		{
			path = rootPath + "index" + HELP_EXTENSION;

			resu = getHelpFile(path);

			if (resu != null)
				return resu;
		}

		// Try /help/lang/actionId
		if (actionId != null)
		{
			path = rootPath + actionId + HELP_EXTENSION;

			resu = getHelpFile(path);

			if (resu != null)
				return resu;
		}

		// Try /help/lang/rootActionId
		if ((rootActionId != null) && (!rootActionId.equals(actionId)))
		{
			path = rootPath + rootActionId + HELP_EXTENSION;

			resu = getHelpFile(path);

			if (resu != null)
				return resu;
		}

		// Help not found
		return null;
	} // end method getHelpFile

	/**
	 * Gets the relative path to the help file for the given file name. Help files are searched into
	 * LY_APP_DOC_DIR. Given path must be a relative path for this directory (eg
	 * /help/en/consult.html).
	 * @param  path The relative path in the LY_APP_DOC_DIR directory.
	 * @return The relative path of the help file or null if not found.
	 */
	protected String getHelpFile(String path)
	{
		// Try to check if file does exist in LY_APP_DOC_DIR in the application
		// path
		String qualifiedPath;
		String resultPath;

		// Check possible location for help files
		qualifiedPath = "$LY_APP_DOC_DIR$" + '/' + _environment.getId().replace('.', '/') + path;
		qualifiedPath = _environment.getString().expand(qualifiedPath);

		File    f = new File(qualifiedPath);
		boolean found = f.exists();

		if (!found)
		{
			qualifiedPath = "$LY_APP_DOC_DIR$" + path;
			qualifiedPath = _environment.getString().expand(qualifiedPath);

			f = new File(qualifiedPath);
			found = f.exists();

			resultPath = _environment.getString().expand("$LY_APP_DOC$" + path);
		}
		else
		{
			resultPath = "$LY_APP_DOC$" + '/' + _environment.getId().replace('.', '/') + path;
			resultPath = _environment.getString().expand(resultPath);
		}

		if (found)
		{
			if (_environment.isDebugLevel(LyEnvironment.DEBUG_HELP_FILES))
				_environment.logOut(" HELP FILE (found) : " + qualifiedPath);

			// Try to give path in LY_APP_DOC (client side)
			return resultPath;
		}
		else if (_environment.isDebugLevel(LyEnvironment.DEBUG_HELP_FILES))
			_environment.logOut(" HELP FILE (not found) : " + qualifiedPath);
		return null;
	} // end method getHelpFile
} // end class LyHelpFiles
