/*
 * LyImages.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.io.*;
import java.net.*;
import java.util.*;

import javax.imageio.*;
import javax.imageio.stream.*;


/**
 * This class is used to manage images of an application. It is based on the mechanism of
 * LyResourceBundle. Images are put in a cache for memory optimization.
 * @version $Id: LyImages.java 41344 2012-10-25 15:45:02Z nhamon $
 * @author  Lyria
 */
public class LyImages extends LyResourceBundle
{
	public static final String __VERSION = "$Id: LyImages.java 41344 2012-10-25 15:45:02Z nhamon $";

	/** Default image key. */
	public static String	   DEFAULT_IMAGE;

	/** Default disabled image key. */
	public static String	   DEFAULT_DISABLED_IMAGE;

	/** Not found key (internal use). */
	public static final int[]  NOT_FOUND = new int[] { -1, -1 };

	/** The prefix used for disabled images. */
	public static String   	   DISABLE_PREFIX = "disabled_";


	/** Association between paths and images (used for optimization). */
	protected Hashtable<String, Object>		   _images;

	/** Association between paths and images sizes (used for optimization). */
	protected Hashtable<String, int[]>		   _imagesSize;

	/** Association between images key and images path (used for optimization). */
	protected Hashtable<String, String>		   _imagesPath;

	/** Association between images key and images url (used for optimization). */
	protected Hashtable<String, String>		   _imagesUrl;

	/**
	 * Association between paths and boolean indicating if images are animated (used for
	 * optimization).
	 */
	protected Hashtable<String, Boolean>	   _imagesAnimated;

	static
	{
		DEFAULT_IMAGE = "LY_DEFAULT_IMAGE".intern();
		DEFAULT_DISABLED_IMAGE = "LY_DEFAULT_DISABLED_IMAGE".intern();
	}

	/**
	 * Constructor with a given environment. Calls {@link #LyImages(LyEnvironment, String)
	 * LyImages(environment, null)}.
	 * @param environment The environment of the images bundle.
	 */
	public LyImages(LyEnvironment environment)
	{
		this(environment, null);
	}

	/**
	 * Constructor with a given environment and name. Calls {@link #LyImages(LyEnvironment, String,
	 * Locale) LyImages(environment, name, environment.getLocale())}.
	 * @param environment The environment of the images bundle.
	 * @param name        The name of the images bundle.
	 */
	public LyImages(LyEnvironment environment, String name)
	{
		this(environment, name, environment.getLocale(), environment.getSkin());
	}

	/**
	 * Constructor with a given environment and name. Calls {@link #LyImages(LyEnvironment, String,
	 * Locale) LyImages(environment, name, environment.getLocale())}.
	 * @param environment The environment of the images bundle.
	 * @param name        The name of the images bundle.
	 * @param skin        The name of the skin
	 */
	public LyImages(LyEnvironment environment, String name, String skin)
	{
		this(environment, name, environment.getLocale(), skin);
	}

	/**
	 * Constructor with a given environment, name and locale. Calls the super implementation.
	 * @param environment The environment of the strings bundle.
	 * @param name        The name of the strings bundle.
	 * @param locale      The locale used by the strings bundle.
	 */
	protected LyImages(LyEnvironment environment, String name, Locale locale)
	{
		this(environment, name, locale, environment.getSkin());
	}

	/**
	 * Constructor with a given environment, name and locale. Calls the super implementation.
	 * @param environment The environment of the images bundle.
	 * @param name        The name of the images bundle.
	 * @param locale      The locale used by the images bundle.
	 * @param skin        The name of the skin
	 */
	protected LyImages(LyEnvironment environment, String name, Locale locale, String skin)
	{
		super(environment, (name == null ? "images" : name), locale, skin);

        _images = null;
        _imagesSize = null;
        _imagesPath = null;
        _imagesUrl = null;
        _imagesAnimated = null;
	}

	/**
	 * If the bundle environment is not null, builds a parent images bundle.
	 * @return The parent images bundle for this images bundle, or null if the environment is null.
	 */
	protected LyResourceBundle buildParentBundle()
	{
		if (_environment == null)
			return null;

		return new LyImages(_environment.getParent(), _name, _locale, _skin);
	}

	/**
	 * Clears all resources of this bundle. Overriden to free all parameters used for optimizations
	 * in this class.
	 */
	public void free()
	{
		if (_images != null)
		{
			Enumeration<Object> elements = _images.elements();
			Object	            image;

			while (elements.hasMoreElements())
			{
				image = elements.nextElement();
				if (image != NOT_FOUND)
					freeImage(image);
			}
	        _images = null;
		}

		_imagesPath = null;
		_imagesSize = null;
		_imagesUrl = null;
		_imagesAnimated = null;

		super.free();
	}

	/**
	 * Gets the current environment view manager default image.
	 * @return the default image
	 */
	public Object getDefaultImage()
	{
		LyViewManagerInterface viewManager = _environment.getViewManager();

		return (viewManager == null ? null : viewManager.getDefaultImage());
	}



	/**
     * Retrieves an image corresponding to a given key and size.
     *
     * @param  key          The resource name (key) of the image
     * @param  sizeCode     The size code of the image (see LyImages.loadImage())
     * @param  defaultImage If true, return default image if not found.
     * @return a reference to the image object, or null if it could not be found.
     * @see    leon.misc.LyImages#loadImage(Object,String)
     * @see    LyEnvironment#getImageSizeInfo(int,int)
     */
    public Object getImage(String key, String sizeCode, boolean defaultImage)
    {
        if (key == null)
            return null;

        if (!hasKey(key)
            && ((_environment.getId() == null)
                 || ((_parent != null)
                     && (_environment.getId() == _parent._environment.getId()))))
            return ((LyImages)_parent).getImage(key, sizeCode, defaultImage);

        if ((_images != null)
            && _images.containsKey(key))
        {
            Object cache = _images.get(key);

            if (cache == NOT_FOUND)
                return null;
            if (cache != null)
                return cache;
        }

        Object result = null;
        String path = getImagePath(key);

        if ((path != null)
            && (path.length() > 0))
        {
            if (path.charAt(0) == '$')
                return getImage(path.substring(1), sizeCode, defaultImage);

            if (sizeCode != null)
                path += sizeCode;

            result = loadImage(null, path);
        }

        if ((result == null)
        	&& (_environment != null)
        	&& (_environment.getViewManager() != null))
        {
        	path = _environment.getViewManager().getTemporaryDir()+File.separator+key;

            if (sizeCode != null)
                path += sizeCode;

            result = loadImage(null, path);
        }

        if ((result == null)
                && (_parent instanceof LyImages))
            result = ((LyImages)_parent).getImage(key, sizeCode, defaultImage);

        LyImageHandlerInterface imageHandler = _environment.getImageHandler();

        if (imageHandler != null)
            result = imageHandler.checkImage(_environment, key, path, result);

        if (!defaultImage)
            return result;

        // use default image
        if (result == null)
        {
            if (key.equals(DEFAULT_IMAGE))
                // Even the default image was not found, create one.
                // To do : use sizecode !
                result = getDefaultImage();
            else
                result = getImage(DEFAULT_IMAGE, sizeCode, true);

            // Always keep result in cache even if this is default image
            // Otherwise computation may be very slow if the application is testing images keys (like studio)
            putCache(path, result);
        }

        return result;
    } // end method getImage

	/**
	 * Gets sizes of the given image.
	 * @param  image the image to measure
	 * @return Image size (table with 2 entries).
	 * @see    LyViewManagerInterface#getSizes(Object)
	 */
	protected int[] getSizes(Object image)
	{
		LyViewManagerInterface viewManager = _environment.getViewManager();

		return (viewManager == null ? null : viewManager.getSizes(image));
	}

	/**
	 * Gets stream of given image
	 */
	protected InputStream getInputStream(String path)
	{
		path = getPathWithoutSize(path);

		InputStream 		resource = _environment.getResourceAsStream(path, null);

		if (resource == null)
		{
			if(path.startsWith("http://")
					|| path.startsWith("https://")
					|| path.startsWith("file://")
					|| path.startsWith("ftp://"))
			{
				// Try from URL
				try
				{
					URL url = new URL(path);

					resource = url.openStream();
				}
				catch (IOException e)
				{
					// Failed. Do not matter.
				}
			}
			else
			{
				// try with a basic file
				File file = new File(path);

				if (file.exists())
				{
					try
					{
						URI uri = file.toURI();

						resource = uri.toURL().openStream();
					}
					catch (Exception e)
					{
						// Failed. Do not matter.
					}
				}
			}
		}

		return resource;
	}

	/**
	 * Gets sizes of the given image.
	 * @param  image the image to measure
	 * @return Image size (table with 2 entries).
	 * @see    LyViewManagerInterface#getSizes(Object)
	 */
	protected int[] getSizes(String path)
	{
		try
		{
			InputStream			resource = getInputStream(path);

			if (resource == null)
				return null;

			ImageInputStream	in = ImageIO.createImageInputStream(resource);

			if (in == null)
				return null;

			try
			{
		        final Iterator<ImageReader>		readers = ImageIO.getImageReaders(in);

		        if (readers.hasNext())
		        {
	                ImageReader	reader = (ImageReader)readers.next();

	                try
	                {
	                    reader.setInput(in);
	                    return new int[] {	reader.getWidth(0), reader.getHeight(0) };
	                }
	                finally
	                {
	                	reader.dispose();
	                }
		        }

		        return null;
			}
			finally
			{
		        if (in != null)
		        	in.close();
			}
		}
        catch (IOException e)
        {
        	return null;
        }
	}

	/**
	 * Creates a scaled version of the given image.
	 * @param  image  The image to scale.
	 * @param  width  The width to which to scale the image.
	 * @param  height The height to which to scale the image.
	 * @return A scaled version of the image.
	 * @see    LyViewManagerInterface#getScaledInstance(Object, int, int)
	 */
	public Object getScaledInstance(Object image, int width, int height)
	{
		LyViewManagerInterface viewManager = _environment.getViewManager();

		return (viewManager == null ? null : viewManager.getScaledInstance(image, width, height));
	}

	/**
	 * Builds the image size description [{@link LyString#VISIBLE_SEP3 VISIBLE_SEP3} + width +
	 * {@link LyString#VISIBLE_SEP2 VISIBLE_SEP2} + height + {@link LyString#VISIBLE_SEP3_BIS
	 * VISIBLE_SEP3_BIS}].
	 * @param  width  requested width or -1 if only the size is specified (keeping aspect ratio)
	 * @param  height requested height (mandatory)
	 * @return The image size description.
	 */
	public String getImageSizeInfo(int width, int height)
	{
		StringBuffer sb = new StringBuffer(20);

		sb.append(_environment.getString().VISIBLE_SEP3);
		if (width >= 0)
		{
			sb.append(width);
			sb.append(_environment.getString().VISIBLE_SEP2);
		}
		sb.append(height);
		sb.append(_environment.getString().VISIBLE_SEP3_BIS);

		return sb.toString();
	}

	/**
	 * Indicates if an image is animated or not.
	 * @param  path The relative path of the image file.
	 * @return True if the image is animated, false otherwise.
	 */
	public boolean isImageAnimated(String path)
	{
		if (path == null)
			return false;

		boolean isAnimated = false;

		if (_imagesAnimated != null)
		{
			Object cache = _imagesAnimated.get(path);

			if (cache != null)
				return ((Boolean)cache).booleanValue();
		}

		if (_imagesAnimated == null)
			_imagesAnimated = new Hashtable<String, Boolean>();

		isAnimated = (path.indexOf("/animated_") >= 0) || (path.indexOf("\\animated_") >= 0);

		_imagesAnimated.put(path, Boolean.valueOf(isAnimated));
		return isAnimated;
	}

	/**
	 * Loads an image. Path may contain information about the image size. The syntax is : [path +
	 * {@link LyString#VISIBLE_SEP3 VISIBLE_SEP3} + width + {@link LyString#VISIBLE_SEP2
	 * VISIBLE_SEP2} + height + {@link LyString#VISIBLE_SEP3_BIS VISIBLE_SEP3_BIS}]. The path of the
	 * image must be relative : the image is searched in the classpath of the application.
	 * @param  path      The relative path of the image file.
	 * @param  component The client of the image creation.
	 * @param  newImage  Indicates whether a new Image must be created.
	 * @param  newImage  Indicates whether an error message should be displayed if an exception
	 * was thrown while loading image.
	 * @return The created or found in cache image, null otherwise.
	 * @see    LyImages#getImageSizeInfo
	 */
	public Object loadImage(Object component, String path, boolean newImage, boolean raiseError)
	{
		if (path == null)
			return null;
		if (_environment == null)
			return null;

		if (!newImage && (_images != null))
		{
			Object cache = _images.get(path);

			if (cache == NOT_FOUND)
				return null;
			if (cache != null)
				return cache;
		}

		try
		{
			// A size may be specified
			int     s = path.lastIndexOf(_environment.getString().VISIBLE_SEP3);
			Object  image = null;

			if (s >= 0)
			{
				int[]   size = getImageSize(path, false);

				if (size != null)
				{
					// Call the method without size information
	                String  subPath = path.substring(0, s);
	                boolean	remove = false;

	                if (_images != null)
	                    image = _images.get(subPath);
	                if ((image == null)
	                    || (image == NOT_FOUND))
	                {
	                    image = loadImage(component, subPath, newImage);
	                    remove = true;
	                }

					if (image != null)
					{
	                    Object  oldImage = image;
						int[]   imageSize = getSizes(image);

						// if there is two size information;
						// we assume that they are width and height
						if (size.length == 2)
						{
							if ((imageSize != null)
									&& ((imageSize.length != 2)
										|| (imageSize[0] != size[0])
										|| (imageSize[1] != size[1])))
								image = getScaledInstance(image, size[0], size[1]);
						}
						else
						{
							int    w, h;
							double ratio;
							// we assume that there is just a dimension (height),
							// so we keep aspect ratio
							int[]  sizes = getSizes(image);

							if ((sizes[0] < 0) || (sizes[1] < 0))
								ratio = 1.752;
							else
								ratio = ((double)sizes[0]) / ((double)sizes[1]);

							h = size[0];
							w = (int)(h * ratio);

							// Don't create a new image if the specified size is the
							// right one
							if ((w != sizes[0]) || (h != sizes[1]))
								image = getScaledInstance(image, w, h);
						}

	                    if (remove
	                    		&& (image != oldImage))
	                        // Don't keep old image
	                        removeCache(subPath);
					}
				}
			}

			// Load the image if necessary
			if (image == null)
			{
				// We must use the ClassLoader of the environment if it exists.
				InputStream resource = _environment.getResourceAsStream(path, null);

				// Try from URL stream
				if ((resource == null)
					&& (path.startsWith("http://")
					    || path.startsWith("https://")
					    || path.startsWith("file://")
						|| path.startsWith("ftp://")))
				{
					// Try from URL
					try
					{
						URL url = new URL(path);

						resource = url.openStream();
					}
					catch (IOException e)
					{
						// Failed. Do not matter.
					}
				}

				if (resource != null)
					image = loadImageFromStream(component, resource);
				else
				{
					// try to load from a complete path
					try
					{
						File f;

						f = new File(_environment.getString().expand(path));
						if (!f.exists() && _environment.isRoot())
						{
							String subPath;
							int    p = path.indexOf("/images/");

							if (p >= 0)
								subPath = path.substring(p + 8);
							else
								subPath = path;

                            String      docDir = _environment.getEnv("LY_APP_DOC_DIR");

                            if (docDir != null)
                                f = new File(_environment.getString().expand(docDir + "/images/" + subPath));
                            else
                                f = new File(_environment.getString().expand("/images/" + subPath));

							if (!f.exists())
                            {
								// The images are in LY_LEON_DOC/leon/images but they should be found
                                // in the CLASSPATH in leon.jar before
                                docDir = _environment.getEnv("LY_LEON_DOC_DIR");
                                if (docDir != null)
                                    f = new File(_environment.getString().expand(docDir + "/images/" + subPath));
                                else
                                    f = new File(_environment.getString().expand("/images/" + subPath));
                            }
						}

						resource = new FileInputStream(f);

						image = loadImageFromStream(component, resource);
					}
					catch (IOException e)
					{
						// _environment.logTrace(e.getMessage());
						putCache(path, null);
						return null;
					}
				}
			}

			// Wait for image to be entirely loaded :
			waitForLoading(component, image);

			// Put the image in the images table
			if (!newImage)
				putCache(path, image);

			return image;
		}
		catch (Exception e)
		{
			if (raiseError)
			{
				if (_environment == null)
					e.printStackTrace();
				else
				{
					_environment.writeException(e);

					String error = _environment.getMessage("LY_ER_FILE", path);

					_environment.raiseError(error);
				}
			}
			return null;
		}
	} // end method loadImage

	/**
	 * Loads an image. This methods only calls {@link #loadImage(Object, String, boolean)
	 * loadImage(component, path, false)}
	 * @param  component The client of the image creation.
	 * @param  path      The relative path of the image file.
	 * @return The image created.
	 * @see    LyImages#loadImage(Object, String, boolean)
	 */
	public final Object loadImage(Object component, String path)
	{
		return loadImage(component, path, false);
	}

	/**
	 * Loads an image. Path may contain information about the image size. The syntax is : [path +
	 * {@link LyString#VISIBLE_SEP3 VISIBLE_SEP3} + width + {@link LyString#VISIBLE_SEP2
	 * VISIBLE_SEP2} + height + {@link LyString#VISIBLE_SEP3_BIS VISIBLE_SEP3_BIS}]. The path of the
	 * image must be relative : the image is searched in the classpath of the application.
	 * @param  path      The relative path of the image file.
	 * @param  component The client of the image creation.
	 * @param  newImage  Indicates wheter a new Image must be created.
	 * @return The created or found in cache image, null otherwise.
	 * @see    LyImages#getImageSizeInfo
	 */
	public Object loadImage(Object component, String path, boolean newImage)
	{
		return loadImage(component, path, false, true);
	}

	/**
	 * Waits for the image loading.
	 * @param  component The client of the image loading
	 * @param  image     the image to load
	 * @throws InterruptedException if the wait is interruptued
	 * @see    LyViewManagerInterface#waitForLoading(Object, Object)
	 */
	protected void waitForLoading(Object component, Object image)
		throws InterruptedException
	{
		if (_environment != null)
		{
			LyViewManagerInterface viewManager = _environment.getViewManager();

			if (viewManager != null)
				viewManager.waitForLoading(component, image);
		}
	}

	/**
	 * Retreives an image from a Stream.
	 * @param  component The client of the image creation.
	 * @param  resource  input stream to the image data
	 * @return the image or null if an error occurs
	 * @throws IOException if something is wrong with the input stream
	 * @see    LyViewManagerInterface#loadImageFromStream(Object, InputStream)
	 */
	protected Object loadImageFromStream(Object component, InputStream resource)
		throws IOException
	{
		LyViewManagerInterface viewManager = _environment.getViewManager();

		return (viewManager == null ? null : viewManager.loadImageFromStream(component, resource));
	}

	/**
	 * Free the image resources if it is not in the cache, otherwise does nothing.
	 * @param image The image that must be freed.
	 */
	protected void freeImage(Object image)
	{
		LyViewManagerInterface viewManager = _environment.getViewManager();

		if ((viewManager != null)
                && (image != null)
                && ((_images == null) || !_images.contains(image)))
			viewManager.freeImage(image);
	}

	/**
	 * Clears application images (reset the cache).
	 */
	public void clearCache()
	{
		super.clearCache();
		_images = null;
		_imagesUrl = null;
		_imagesPath = null;
		_imagesAnimated = null;
		_imagesSize = null;
	}
	
	/**
	 * Indicates whether given image is contained in the images cache.
	 * @param image The image that may be cached.
	 * @return true if image is cached, false otherwise
	 */
	public boolean isInCache(Object image)
	{
		if ((_images != null) && _images.contains(image))
			return true;

		if (((_environment.getId() == null)
                 || ((_parent != null)
                     && (_environment.getId() == _parent._environment.getId()))))
			return ((LyImages)_parent).isInCache(image);
		
		return false;
	}
	
	/**
	 * Adds the specified pair (path/image) to the image cache.
	 * @param path  The image path that must be added to the cache.
	 * @param image The image that must be added to the cache.
	 */
	public void putCache(String path, Object image)
	{
		// First check if image may be cached
		LyViewManagerInterface	viewManager = _environment.getViewManager();
		boolean					useCache = viewManager.useCacheForImage(path, image);
		
		if (!useCache)
			return;
		
		if (_images == null)
			_images = new Hashtable<String, Object>(10);

		path = path.intern();

        Object  oldImage = _images.get(path);

        if (image == oldImage)
            return;

        if (oldImage != null)
            freeImage(oldImage);

		if (image == null)
			_images.put(path, NOT_FOUND);
		else
			_images.put(path, image);
	}

    /**
     * Removes the image associated to given key from the cache.
     * @param path  The image path that must be removed from the cache.
     */
    public void removeCache(String path)
    {
        if (_images == null)
            return;

        path = path.intern();

        Object image = _images.get(path);

        _images.remove(path);

        if (image != null)
            freeImage(image);
    }

	/**
	 * This method retrieves information about an image size. If this information is given in the
	 * path the result is an array of two integers describing the image dimension.
	 * @param  path            The path to check.
	 * @param  enableLoadImage indicates wether the application is allowed to load the image
	 * @return An array of two integers (width, height) describing the size of the image or null if
	 *         no size is given in the path.
	 * @see    LyImages#getImageSizeInfo(int, int)
	 */
	public int[] getImageSize(String path, boolean enableLoadImage)
	{
		if (path == null)
			return null;

        if (!hasKey(path)
            && ((_environment.getId() == null)
                 || ((_parent != null)
                     && (_environment.getId() == _parent._environment.getId()))))
            return ((LyImages)_parent).getImageSize(path, enableLoadImage);

		// Check if a size is specified
		int		n1 = path.lastIndexOf(_environment.getString().VISIBLE_SEP3);

		if (n1 <= 0)
		{
			// Check in cache if the image size is found
			if (_imagesSize != null)
			{
				Object sizeInfo = _imagesSize.get(path);

				if (sizeInfo == NOT_FOUND)
					return null;
				else if (sizeInfo instanceof int[])
					return ((int[])sizeInfo);
			}

			if (enableLoadImage)
			{
				String	imagePath = getImagePath(path);
				Object	image = (_images != null ? _images.get(path) : null);
				int[]   sizes;

				if ((image == NOT_FOUND)
						|| (image ==  null))
					sizes = getSizes(imagePath);
				else
					sizes = getSizes(image);

				// Keep in cache for further use
				if (_imagesSize == null)
					_imagesSize = new Hashtable<String, int[]>();

				if (sizes == null)
					_imagesSize.put(path, NOT_FOUND);
				else
					_imagesSize.put(path, sizes);

				return sizes;
			}

			return null;
		}
		else
		{
			int n2 = path.lastIndexOf(LyTools.toString(_environment.getString().VISIBLE_SEP2));
			int n3 = path.lastIndexOf(_environment.getString().VISIBLE_SEP3_BIS);

			if ((n2 > n1)
				&& (n3 > n2))
			{
				int w, h;

				w = LyTools.intFromString(path.substring(
							n1 + _environment.getString().VISIBLE_SEP3.length(), n2));
				h = LyTools.intFromString(path.substring(n2 + 1, n3));

				return new int[] { w, h };
			}
			else if ((n2 <= 0)
				&& (n3 > n1))
			{
				// There is just the information of height for the image
				// we have to keep the aspect ratio
				int h;

				h = LyTools.intFromString(path.substring(
							n1 + _environment.getString().VISIBLE_SEP3.length(), n3));

				return new int[] { h };
			}
			else
				return null;
		}
	} // end method getImageSize

	/**
	 * Gets the image path corresponding to the given key.
	 * @param  key the image resource key
	 * @return The image path corresponding to the given key if found, null otherwise
	 */
	public String getImagePath(String key)
	{
		if ((key == null)
		        || (key.length() == 0))
			return DEFAULT_IMAGE;

        if (!hasKey(key)
            && ((_environment.getId() == null)
                 || ((_parent != null)
                     && (_environment.getId() == _parent._environment.getId()))))
            return ((LyImages)_parent).getImagePath(key);

		String result = null;

		// Check in local cache
		if (_imagesPath != null)
		{
			result = _imagesPath.get(key);
			if (result != null)
				return result;
		}

        // Try first to find an extra resource with that key
		String    value = get(key);

		// If the value was not found, use the key
		if ((value == null)
		        || (value.trim().length() == 0))
			value = key;
		else if ((value.length() > 0)
		        && (value.charAt(0) == '$'))
			return value;

		result = getPath(value);

        // Keep in cache for further access
        if (_imagesPath == null)
            _imagesPath = new Hashtable<String, String>(10);

        _imagesPath.put(key, result);
        return result;
	} // end method getImagePath

	/**
	 * Gets path corresponding to given image name for associated environment.
	 */
	public String getPath(String imageName)
	{
	    // Build the path
	    String     result, dir = ".";

        if (_environment == null)
            dir = "leon";
        else if (_environment.getId() != null)
            dir = _environment.getId().replace('.', '/');

        // Expand the value to replace keys
        if (_environment != null)
            imageName = _environment.getString().expand(imageName, ".");

        imageName = imageName.replace('\\', '/');

        if (imageName.startsWith(dir + '/')
                || (imageName.indexOf("/images/") >= 0))
            result = imageName;
        else if (imageName.startsWith("http://")
                || imageName.startsWith("https://")
                || imageName.startsWith("file://")
                || imageName.startsWith("ftp://"))
            result = imageName;
        // If file exists keep given value (for example an absolute path)
        else if (new File(getPathWithoutSize(imageName)).exists())
            result = imageName;
        else
            result = dir + "/images/" + imageName;

        return result;
	}

    /**
     * Retrieves the path of an image without the information about its size.
     * @param  path The path of the image.
     * @return The path of the image without the size information.
     */
    public String getPathWithoutSize(String path)
    {
        // Check if a size is specified
    	// We may have a problem with path containing parenthesis.
    	// so first check if the path ends with a closing parenthesis.
    	if (!path.endsWith(_environment.getString().VISIBLE_SEP3_BIS))
    		return path;
        int n1 = (_environment == null ? -1 : path.indexOf(_environment.getString().VISIBLE_SEP3));

        return (n1 <= 0 ? path : path.substring(0, n1));
    }

	/**
	 * Gets the image URL corresponding to the given key. Calls {@link #getImageUrl(String, boolean)
	 * getImageUrl(key, false)}.
	 * @param  key the image resource key
	 * @return The image url corresponding to the given key if found, null otherwise
	 */
	public final String getImageUrl(String key)
	{
		return getImageUrl(key, false);
	}

	/**
	 * Gets the disable image URL corresponding to the given key. Calls {@link #getImageUrl(String,
	 * boolean) getImageUrl(key, true)}.
	 * @param  key the image resource key
	 * @return The disable image URL corresponding to the given key if found, null otherwise
	 */
	public final String getDisabledImageUrl(String key)
	{
		return getImageUrl(key, true);
	}

	/**
	 * Return the image url corresponding to the key passed. Calls {@link #getImageUrl(String,
	 * boolean, boolean) getImageUrl(key, disabled, true)}.
	 * @param  key      the image resource key
	 * @param  disabled Indicates if disabled image should be searched.
	 * @return The image url corresponding to the key passed or the default image url if not found.
	 */
	public final String getImageUrl(String key, boolean disabled)
	{
		return getImageUrl(key, disabled, true);
	}

	/**
	 * Gets the image url corresponding to the given key.
	 * @param  key          the image resource key
	 * @param  disabled     Indicates if disabled image should be searched.
	 * @param  defaultImage If true, return default image if the URL is not found.
	 * @return The image url corresponding to the given key, the default image url if not found and
	 *         defaultImage parameter is set to true, null otherwise.
	 */
	public String getImageUrl(String key, boolean disabled, boolean defaultImage)
	{
		// Check in local cache
		if (key == null)
			return (_environment.getLeonDocExternalUrl() + "/images/" + DEFAULT_IMAGE);

        if (!hasKey(key)
            && ((_environment.getId() == null)
                 || ((_parent != null)
                     && (_environment.getId() == _parent._environment.getId()))))
            return ((LyImages)_parent).getImageUrl(key, disabled, defaultImage);

	    String   resu = null;

		if (_imagesUrl != null)
		{
			if (!disabled)
				resu = _imagesUrl.get(key);
			else
				resu = _imagesUrl.get(DISABLE_PREFIX + key);

			if (resu != null)
				return resu;
		}

        String	value = get(key);

		// If the value was not found, return the key
		if (value == null)
			value = key;
		else if ((value.length() > 0) && (value.charAt(0) == '$'))
			return value;

		String	path = _environment.getPathWithoutSize(value);

		// If path starts with http, we assume path is already an url
		if ((path != null)
			&& (path.startsWith("http://")
		        || path.startsWith("https://")
				|| path.startsWith("file://")
				|| path.startsWith("ftp://")))
			return path;

		// If disabled image, add the DISABLE_PREFIX before file name
		if (disabled && (path != null))
			path = getDisableImageFileName(path) ;

		// If file exists keep given value (for example an absolute path)
		if ((path != null)
				&& (new File(path).exists()))
			resu = path;
		else
		{
			// try to find this image in the app_doc_dir
	        LyString  string = _environment.getString();
	        String	  appDoc = string.expand(_environment.getEnv("LY_APP_DOC"));
	        String	  appDocDir = string.expand(_environment.getEnv("LY_APP_DOC_DIR"));
	        String    fullImagePrefix = (path.contains("/images/") ? "/" : "/images/");
	        String	  testPath;

	        appDoc = appDoc.replace(File.separatorChar, '/');
	        appDocDir = appDocDir.replace(File.separatorChar, '/');

	        if (('/' + path).startsWith(appDoc + '/'))
	        	path = path.substring(appDoc.length() - 1);
	        testPath = appDocDir + fullImagePrefix + path;

			File      testFile = new File(testPath);

			if (!testFile.exists())
			{
				// Check if can find images directly in images dir
				int		p = path.indexOf("/images/");

				if (p >= 0)
				{
					path = path.substring(p);
					testPath = appDocDir + fullImagePrefix + path;
					testFile = new File(testPath);
				}
			}

			// If file exists keep given value (for example an absolute path)
			if (testFile.exists())
			{
				String	baseUrl = _environment.getEnv("_LY_APP_DOC_EXTERNAL");

				if (baseUrl == null)
					baseUrl = appDoc;
				resu = baseUrl + fullImagePrefix + path;
			}
			else
			{
				// try to find this image in the leon_doc_dir
				testPath = string.expand(_environment.getEnv("LY_LEON_DOC_DIR") + fullImagePrefix + path);
				testFile = new File(testPath);

				String baseUrl = _environment.getLeonDocExternalUrl();

				if (testFile.exists())
					resu = baseUrl + fullImagePrefix + path;
				else
				{
					if (!defaultImage)
						return null;

					_environment.logWarning("LyImages: image not found : " + testPath);
					if (!disabled)
					{
						if (DEFAULT_IMAGE.equals(key))
							return null;
						resu = getImageUrl(DEFAULT_IMAGE);
					}
					else
					{
						if (DEFAULT_DISABLED_IMAGE.equals(key))
							return null;
						resu = getImageUrl(DEFAULT_DISABLED_IMAGE);
					}

				}
			}
		}

		// Keep in cache for further access
		if (_imagesUrl == null)
			_imagesUrl = new Hashtable<String, String>(10);

		if (!disabled)
			_imagesUrl.put(key, resu);
		else
			_imagesUrl.put(DISABLE_PREFIX + key, resu);

		return resu;
	} // end method getImageUrl

	/**
	 * Gets the image file name corresponding to the given key.
	 * @param  key the image resource key
	 * @return The image file name corresponding to the given key if found, null otherwise
	 */
	public String getImageFileName(String key)
	{
		if (key == null)
			return key;

		String	filename = get(key);

		if ((filename != null)
				&& (filename.trim().length() != 0))
			return filename;

        if (_parent instanceof LyImages)
        	return ((LyImages)_parent).getImageFileName(key);

		return null;
	}

	/**
	 * Gets the disable image file name from the given image file name.
	 *
	 * @param  imageFileName the image file name whose associated disable image file name nned to be retreive
	 * @return the disable image file name
	 */
	public static String getDisableImageFileName(String imageFileName)
	{
		String disableImageFileName = imageFileName;

		// Try to find the last '\' or '/' path separator to get the file name.
		int sepPos = -1;
		int sep1Pos = disableImageFileName.lastIndexOf("\\");
		int sep2Pos = disableImageFileName.lastIndexOf("/");
		if (sep1Pos < sep2Pos)
			sepPos = sep2Pos;
		else
			sepPos = sep1Pos;

		int pointPos = disableImageFileName.lastIndexOf(".");

		if ((sepPos >= 0) && (sepPos < pointPos))
			disableImageFileName = disableImageFileName.substring(0, sepPos + 1) + LyImages.DISABLE_PREFIX + disableImageFileName.substring(sepPos + 1);
		else
			disableImageFileName = LyImages.DISABLE_PREFIX + disableImageFileName;

		return disableImageFileName;
	} // end method getDisableImageFileName

	/**
	 * Adds a resource in this resource bundle. Key and value must not be null.
	 * @param key   Key of the resource.
	 * @param value Value of this new resource.
	 */
	public void put(String key, String value)
	{
		super.put(key, value);

		if (_imagesPath != null)
		{
			String	path = _imagesPath.get(key);

			_imagesPath.remove(key);
			if (path != null)
			{
				// Remove image from cache and from size cache
				removeCache(path);

				if (_imagesSize != null)
					_imagesSize.remove(path);
				if (_imagesAnimated != null)
					_imagesAnimated.remove(path);
			}
		}

		if (_imagesUrl != null)
			_imagesUrl.remove(key);
	}

	/**
	 * Checks if given key image corresponds to a known image.
	 *
	 * @param  key          The resource name (key) of the image
	 * @return True if found, false otherwise.
	 */
	public boolean checkImage(String key)
	{
        if (key == null)
            return false;

        if (!hasKey(key)
            && ((_environment.getId() == null)
                 || ((_parent != null)
                     && (_environment.getId() == _parent._environment.getId()))))
            return ((LyImages)_parent).checkImage(key);

        if ((_images != null)
            && _images.containsKey(key))
        {
            Object cache = _images.get(key);

            if (cache == NOT_FOUND)
                return false;
            if (cache != null)
                return true;
        }

        boolean result = false;
        String 	path = getImagePath(key);

		path = getPathWithoutSize(path);

        if ((path != null)
            && (path.length() > 0))
        {
            if (path.charAt(0) == '$')
                return checkImage(path.substring(1));

            result = checkPath(path);
        }

        if (!result
        	&& (_environment != null)
        	&& (_environment.getViewManager() != null))
        {
        	path = _environment.getViewManager().getTemporaryDir()+File.separator+key;
            result = checkPath(path);
        }

        if (!result
                && (_parent instanceof LyImages))
            result = ((LyImages)_parent).checkImage(key);

        return result;
	} // end method checkImage

	/**
	 * Checks if given path image corresponds to a known file.
	 *
	 * @param  path File path.
	 * @return True if found, false otherwise.
	 */
	protected boolean checkPath(String path)
	{
		InputStream		resource = getInputStream(path);

		if (resource != null)
		{
			try
			{
				resource.close();
			}
			catch (IOException e)
			{
			}

			return true;
		}

		return false;
	}
} // end class LyImages
