/*
 * LyLastValues.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

/**
 * Utility class used to keep a list of last accessed values corresponding to keys.
 */
public class LyLastValues
{

	public static final String __VERSION =
		"$Id: LyLastValues.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Default number of elements stored into the list. */
	public static int		   DEFAULT_SIZE = 5;

	/** Default object returned when the object corresponding to a key is not found. */
	public static final Object NOT_FOUND = new Object();

	/** Last accessed keys. */
	protected Object[]		   _keys;

	/** Last accessed values. */
	protected Object[]		   _values;

	/**
	 * Constructor. Calls {@link #LyLastValues(int) LyLastValues(DEFAULT_SIZE)}
	 */
	public LyLastValues()
	{
		this(DEFAULT_SIZE);
	}

	/**
	 * Constructor. Creates the keys list and the values list of the given size.
	 * @param size the number of objects to store
	 */
	public LyLastValues(int size)
	{
		_keys = new Object[size];
		_values = new Object[size];
	}

	/**
	 * Adds a pair (key, value) to last accessed values list.
	 * @param key   the key to add
	 * @param value the value to add
	 */
	public synchronized void addValue(Object key, Object value)
	{
		if (key == null)
			throw new IllegalArgumentException("BF AppEngine : parameter key can't be null");

		int i, j, n = _keys.length;

		for (i = 0; (i < n) && (_keys[i] != key); i++)
			;

		if (i == 0)
		{
			_values[0] = value;
			return;
		}

		for (j = ((i < n) ? i : (n - 1)); j > 0; j--)
		{
			_keys[j] = _keys[j - 1];
			_values[j] = _values[j - 1];
		}

		_keys[0] = key;
		_values[0] = value;
	}

	/**
	 * Finds the value corresponding to the given key.
	 * @param  key the key
	 * @return The value if found, {@link #NOT_FOUND} otherwise.
	 */
	public synchronized Object getValue(Object key)
	{
		if (_keys == null)
			return NOT_FOUND;

		int i, n = _keys.length;

		for (i = 0; (i < n) && (_keys[i] != key); i++)
			;
		if (i == n)
			return NOT_FOUND;

		Object value = _values[i];

		if (i > 0)
		{
			int j;

			for (j = i; j > 0; j--)
			{
				_keys[j] = _keys[j - 1];
				_values[j] = _values[j - 1];
			}

			_keys[0] = key;
			_values[0] = value;
		}

		return value;
	} // end method getValue

	/**
	 * Removes value corresponding to given key.
	 * @param  key the key
	 * @return The removed value if found, {@link #NOT_FOUND} otherwise.
	 */
	public synchronized Object removeValue(Object key)
	{
		if (_keys == null)
			return NOT_FOUND;

		int i, n = _keys.length;

		for (i = 0; (i < n) && (_keys[i] != key); i++)
			;
		if (i == n)
			return NOT_FOUND;

		Object value = _values[i];
		int    j;

		for (j = i; j < (n - 1); j++)
		{
			_keys[j] = _keys[j + 1];
			_values[j] = _values[j + 1];
		}

		_keys[n - 1] = null;
		_values[n - 1] = null;
		return value;
	}
} // end class LyLastValues
