/*
 * LyLogger.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.util.*;


/**
 * The <code>LyLogger</code> class is used to log messages.
 */
public class LyLogger
{

	public static final String __VERSION = "$Id: LyLogger.java 32750 2011-10-17 15:13:35Z bangiole $";

	/** The logger environment. */
	protected LyEnvironment    _environment;

	/**
	 * Constructor for a logger. Sets the current environment with the given value.
	 * @param environment the logger environment
	 */
	public LyLogger(LyEnvironment environment)
	{
		_environment = environment;
	}

	/**
	 * Utility method to get the current date as a string.
	 * @return current date as a string.
	 */
	protected String getDateString()
	{
		if (_environment.isDebugLevel(LyEnvironment.DEBUG_ALL))
			return "[" + LyDateFormatter.format(new Date(), "MM/dd HH:mm:ss.SSS") + ", thread  "
			+ Thread.currentThread() + "]\n";
		else
			return "[" + LyDateFormatter.format(new Date(), "MM/dd HH:mm:ss") + "]";
	}

	/**
	 * Logs the given message on the standard output stream.
	 * @param message Message to log.
	 */
	public void logOut(String message)
	{
		if ((_environment == null) || (_environment.out == null))
			return;
		
		_environment.out.println(getDateString() + message);
	}

	/**
	 * Logs the given message on the standard error stream.
	 * @param message Message to log.
	 */
	public void logErr(String message)
	{
		if ((_environment == null) || (_environment.err == null))
			return;
		
		_environment.err.println(getDateString() + message);
	}

	/**
	 * Logs the given fatal error.
	 * @param message fatal error to log.
	 * @see   #logErr(String)
	 */
	public void fatal(String message)
	{
		logErr("[FATAL] " + message);
	}

	/**
	 * Logs the given error.
	 * @param message error to log.
	 */
	public void error(String message)
	{
		logErr("[ERROR] " + message);
	}

	/**
	 * Logs the given warning.
	 * @param message warning to log.
	 */
	public void warning(String message)
	{
		logErr("[WARNING] " + message);
	}

	/**
	 * Logs the given information.
	 * @param message information to log.
	 */
	public void info(String message)
	{
		logOut("[INFO] " + message);
	}

	/**
	 * Logs the given trace.
	 * @param message trace to log.
	 */
	public void trace(String message)
	{
		logOut("[TRACE] " + message);
	}

	/**
	 * Writes an exception and its stack trace to the error stream.
	 * @param t The exception to write.
	 */
	public void writeException(Throwable t)
	{
	    if ((t instanceof Exception)
			|| (t instanceof Error))
	    {
	        // Real exception
			Throwable		    t2 = new Throwable();
			StackTraceElement[] stackTrace = t2.getStackTrace();
			int				    ns = stackTrace.length;
			int				    i, n = ((ns > 5) ? 5 : ns);
	
			logErr("[STACK] Leonardi : exception caught in ");
			for (i = 2; i < n; i++)
				_environment.err.println("\t" + stackTrace[i]);
			if (i < ns)
				_environment.err.println("\t...");

			t.printStackTrace(_environment.err);
	    }
	    else
	    {
	        // Stack trace
	        logOut("[STACK] Leonardi : debug trace (not an exception) ");
            if (t == null)
                t = new Throwable();
	        t.printStackTrace(_environment.out);
	    }
	}

	/**
	 * Gets this logger environment.
	 * @return The environment by which this logger was created.
	 */
	public LyEnvironment getEnvironment()
	{
		return _environment;
	}

	/**
	 * Changes the debug level of this logger. This method does nothing.
	 * @param level The new debug level that must be set on the environment. The debug level is a
	 *              BitSet with values defined as LyEnvironment.DEBUG_xxx constants.
	 */
	protected void setDebugLevel(BitSet level)
	{
	}
    
    /**
     * Internal method used to retrieve current logger name.
     * @return Logger name (name of package calling log).
     */
    protected String getLoggerName()
    {
        Throwable           t = new Throwable();
        StackTraceElement[] stackTrace = t.getStackTrace();
        int                 i, n = stackTrace.length;
        String              name = null;
        
        for (i = 2; i < n; i++)
        {
            name = stackTrace[i].getClassName();
            if (!name.equals("leon.misc.LyLogger")
                && !(name.equals("leon.misc.LyEnvironment")
                     && stackTrace[i].getMethodName().startsWith("log")))
            {
                name = name.substring(0, name.lastIndexOf('.'));
                return name;
            }
        }
        
        return null;
    }
} // end class LyLogger
