/*
 * LyMultiList.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.util.*;


/**
 * This class is an utility class to manage a list of objects with a counter of occurences for each
 * object. It is like a java Hashtable containing these objects as keys and an integer value as
 * values.
 * @see     java.util.Hashtable
 * @see     java.util.Vector
 * @version $Id: LyMultiList.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public final class LyMultiList<T>
{

	public static final String __VERSION =
		"$Id: LyMultiList.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The list of objects (of type {@link _Entry}). */
	Vector<_Entry>			_objects;

	/**
	 * Default constructor. Sets the list of objects to null.
	 */
	public LyMultiList()
	{
		_objects = null;
	}

	/**
	 * Internal method used to get entry of the given object. An {@link _Entry entry} is the
	 * association between an object and its number of occurence.
	 * @param  object the object to find
	 * @return The entry of the object if it is found in the list, null otherwise.
	 */
	_Entry getEntry(T object)
	{
		_Entry entry;
		int    i, n = getSize();

		for (i = 0; i < n; i++)
		{
			entry = _objects.elementAt(i);
			if (entry._object == object)
				return entry;
		}

		return null; // Not found
	}

	/**
	 * Checks if the given object is contained in this list.
	 * @param  object The object to find.
	 * @return true if the given object is found, false otherwise.
	 */
	public boolean contains(T object)
	{
		return (getEntry(object) != null);
	}

	/**
	 * Gets the index of the given object.
	 * @param  object The object.
	 * @return The index if the object if it is found in this list, -1 otherwise.
	 */
	public int getIndex(T object)
	{
		_Entry entry;
		int    i, n = getSize();

		for (i = 0; i < n; i++)
		{
			entry = _objects.elementAt(i);
			if (entry._object == object)
				return i;
		}

		return -1; // Not found
	}

	/**
	 * Gets the number of occurences of the given object.
	 * @param  object The object.
	 * @return The number of occurences of the object if found in this list, 0 otherwise.
	 */
	public int getCount(T object)
	{
		_Entry entry = getEntry(object);

		return ((entry == null) ? 0 : entry._count);
	}

	/**
	 * Gets the number of objects in this list.
	 * @return the number of objects in this list or -1 if no object was added yet.
	 */
	public int getSize()
	{
		return ((_objects == null) ? -1 : _objects.size());
	}

	/**
	 * Gets the object at the specified index.
	 * @param  index The index of the requested object.
	 * @return The reference to the requested object, or null if it could not be found.
	 */
	public T getObject(int index)
	{
		if ((_objects == null) || (index < 0) || (index >= _objects.size()))
			return null;

		return _objects.elementAt(index)._object;
	}

	/**
	 * Adds the given object to the objects list.<br />
	 * <b>If</b> the object is ever in the list, increment its number of occurence<br />
	 * <b>Else</b> builds adds the object to the list.
	 * @param  object the object to add
	 * @return The number of occurences of the object.
	 */
	public int addObject(T object)
	{
		if (object == null)
			return -1;

		// Checks if the object vector is ready to use
		if (_objects == null)
			_objects = new Vector<_Entry>(1);

		// Adds the new object if not already added
		_Entry entry = getEntry(object);

		if (entry == null)
		{
			entry = new _Entry(object, 1);
			_objects.addElement(entry);
			return 1;
		}

		entry._count++;
		return entry._count;
	}

	/**
	 * Removes an occurence of the specified object from this list.<br />
	 * <b>If</b> the object occurence number equals 1, the object is removed from the list,<br />
	 * <b>Else</b>, the number of occurences is decremented.
	 * @param  object Object to remove.
	 * @return The number of remaining occurences of the object or -1 if the object was not in the
	 *         list.
	 */
	public int removeObject(T object)
	{
		if ((object == null) || (_objects == null))
			return -1;

		_Entry entry = getEntry(object);

		if (entry == null)
			return -1;

		if (entry._count == 1)
		{
			_objects.removeElement(entry);
			return 0;
		}

		entry._count--;
		return entry._count;
	}

	/**
	 * Removes all occurences of an object in this list.
	 * @param object The object to remove from the list.
	 */
	public void clearObject(T object)
	{
		if ((object == null) || (_objects == null))
			return;

		int index = getIndex(object);

		if (index < 0)
			return;
		else
			_objects.removeElementAt(index);
	}

	/**
	 * Gets the set of objects. This is a COPY of the internal list, therefore this list may be
	 * freely manipulated by the application.
	 * @return The set of objects if any, an empty Vector if the set is empty.
	 */
	public Vector<T> getObjects()
	{
		if (_objects == null)
			return null;

		// Return a copy to avoid manipulation of the internal list
		int          i, n = getSize();
		_Entry       entry;
		Vector<T>    objects = new Vector<T>(n);

		for (i = 0; i < n; i++)
		{
			entry = _objects.elementAt(i);
			objects.addElement(entry._object);
		}

		return objects;
	}

	/**
	 * Internal Class.
	 */

	/**
	 * Internal class to attach a counter to an object.
	 */
	private class _Entry
	{
		/** The object. */
		T      _object;

		/** The number of occurences of the object. */
		int     _count;

		/**
		 * Constructor. Builds a new entry that associate the given object to the given number of
		 * occurence of the object.
		 *
		 * @param object the object
		 * @param count  the number of occurences of the object
		 */
		_Entry(T object, int count)
		{
			_object = object;
			_count = count;
		}
	}
} // end class LyMultiList
