package leon.misc;

import java.io.*;
import java.util.*;

import leon.app.*;
import leon.data.*;

/**
 * Implements a performance logger.
 * 
 * A perf event consist of a start event and a end event.
 * Each basic event has a timestamp which is used to compute the duration.
 * 
 * @author bangiole
 *
 */
public class LyPerfLogger
{
	public static String ENTRY_SEP = ";";
	public static String VALUE_SEP = ":";
	
	public static String ERROR_STRING = "Error";
	
	private static final String	DEFAULT_SESSION_ID	= "default";
	enum PerfLoggerStatus
	{
		START,
		END
	};
	
	Map<String, _logInstance> _sessionMap = new HashMap<String, _logInstance>();
	_logWriter _writer;

	/**
	 * Constructor
	 * @param outputPath: the full path of the output file
	 */
	public LyPerfLogger(String outputPath)
	{
		_writer = new _logWriter(outputPath);
		Thread tr = new Thread(_writer);
		tr.setDaemon(true);
		tr.setName("Perf logger daemon");
		tr.start();
	}
	
	/**
	 * Stop the perf logger
	 */
	public void close()
	{
		_writer.stop();
		_writer = null; 
	}
	
	protected String getLogId(LyDataSession session)
	{
		String logId = null;
		if (session == null)
			logId = DEFAULT_SESSION_ID + "|" + Thread.currentThread().getId();
		else
			logId = session.getId() + "|" + Thread.currentThread().getId();
		return logId;
	}
	
	/**
	 * Add a start event
	 * @param application the application
	 * @param actionId the action id
	 * @param advancedParameters a user defined string
	 */
	public void addStartEvent(LyApplication application, String actionId, String advancedParameters)
	{
		addStartEvent(application.getCurrentSession(), actionId, advancedParameters);
	}

	/**
	 * Add a start event
	 * @param session the current session
	 * @param actionId the action id
	 * @param advancedParameters a user defined string
	 */
	public void addStartEvent(LyDataSession session, String actionId, String advancedParameters)
	{
		// first try to get an existing log
		String logId = getLogId(session);
		_logInstance log = _sessionMap.get(logId);
		if (log == null)
		{
			log = new _logInstance(session);
			_sessionMap.put(logId, log);
		}
		log.addStartEvent(actionId, advancedParameters);
	}
	
	/**
	 * Add an end event
	 * @param application the application
	 * @param actionId the action id (must be the same than the start event)
	 * @param advancedParameters a user defined string
	 */
	public void addEndEvent(LyApplication application, String actionId, String advancedParameters)
	{
		addEndEvent(application.getCurrentSession(), actionId, advancedParameters);
	}
	
	/**
	 * Add an end event
	 * @param session the current session
	 * @param actionId the action id (must be the same than the start event)
	 * @param advancedParameters a user defined string
	 */
	public void addEndEvent(LyDataSession session, String actionId, String advancedParameters)
	{
		
		String logId = getLogId(session);
		_logInstance log = _sessionMap.get(logId);
		if (log == null)
			throw new IllegalStateException("Cannot add end perf event before starting it...");
		
		boolean result = log.addEndEvent(actionId, advancedParameters);
		
		if (result)
		{
			// it was an end event, dump it...
			_writer.addToQueue(log);
			_sessionMap.remove(logId);
		}
		return;
	}

	/**
	 * An instance of a log for the current session
	 * This class will hold a list of "lines" for the current session for a given thread.
	 * A log is created when a start event is added for the first time.
	 * It will be keeped until the equivalent end event is added.
	 * 
	 */
	class _logInstance
	{
		List<_line>		_lines;
		long			_depth;
		LyDataSession	_session;
		
		_logInstance(LyDataSession session)
		{
			_depth = -1;
			_session = session;
			_lines = new ArrayList<_line>();
		}
		
		void addStartEvent(String actionId, String advancedParam)
		{
			_depth ++;
			_line line = new _line(_session, _depth, PerfLoggerStatus.START, actionId, advancedParam);
			_lines.add(line);
		}
		
		boolean addEndEvent(String actionId, String advancedParam)
		{
			_line line = new _line(_session, _depth, PerfLoggerStatus.END, actionId, advancedParam);
			_lines.add(line);
			_depth --;
			if (_depth < 0)
				return true;
			
			return false;
		}
		
		void free()
		{
			_lines = null;
			_session = null;
		}

		String dump()
		{
			// first, we must compute duration for each line
			
			StringBuffer sb = new StringBuffer();
			_line line, startLine;
			int i, j, n = (_lines == null ? 0 : _lines.size());
			for (i = 0; i < n ; i++)
			{
				line = _lines.get(i);
				if (line._status == PerfLoggerStatus.START)
					line.setDuration(0);
				else
				{
					// we have to find the start event
					for (j = 0; j < i; j++)
					{
						startLine = _lines.get(j);
						if (startLine._status != PerfLoggerStatus.START)
							continue;
						if (line._depth != startLine._depth)
							continue;
						if (!startLine._actionId.equals(line._actionId))
							continue;
						
						// Ok start found, compute duration
						line.setDuration(line._timeStamp - startLine._timeStamp);
						break;
					}
				}
				sb.append(line.toString());
				sb.append("\n");
			}
			return sb.toString();
		}
		
	}
	
	/**
	 * This class represents a line in the log file.
	 *
	 */
	class _line
	{
		
		long				_timeStamp;
		String				_sessionId;
		String				_userId;
		long				_threadId;
		String				_actionId;
		String				_advancedParameters;
		PerfLoggerStatus	_status;
		long				_depth;
		long				_duration;
		
		_line(LyDataSession session, long depth, PerfLoggerStatus status, String actionId, String advancedParam)
		{
			_timeStamp = System.currentTimeMillis();
			_sessionId = (session == null ? DEFAULT_SESSION_ID : session.getId());
			_threadId = Thread.currentThread().getId();
			LyObjectId userId = (session == null ? null : (LyObjectId)session.getValue("$USER"));
			_userId = (userId == null ? "" : userId.toName());
			_status = status;
			_depth = depth;
			_actionId = actionId;
			_advancedParameters = advancedParam;
			_duration = 0;
		}
		
		void setDuration(long duration)
		{
			_duration = duration;
		}
		public String toString()
		{
			StringBuffer sb = new StringBuffer();
			
			sb.append(_timeStamp);
			sb.append(ENTRY_SEP);
			sb.append(_depth);
			sb.append(ENTRY_SEP);
			sb.append(_sessionId);
			sb.append(ENTRY_SEP);
			sb.append(_threadId);
			sb.append(ENTRY_SEP);
			sb.append(_userId);
			sb.append(ENTRY_SEP);
			sb.append(_actionId);
			sb.append(ENTRY_SEP);
			sb.append(_status.name());
			sb.append(ENTRY_SEP);
			sb.append(_duration);
			sb.append(ENTRY_SEP);
			sb.append(_advancedParameters == null ? "" : _advancedParameters);
			sb.append(ENTRY_SEP);
			
			return sb.toString();
		}
	}
	
	/**
	 * This class implements a perf log writer in a separate daemon thread
	 * to avoid too many perf impacts
	 *
	 */
	class _logWriter implements Runnable
	{
		String				 _outputPath;
		FileWriter			 _fw;
		Queue<_logInstance>	 _fileQueue;
		boolean				 _quit;
		
		_logWriter(String outputPath)
		{
			_outputPath = outputPath;
			File outputFile = new File(_outputPath);
			_quit = false;
			if (outputFile.exists())
			{
				File bakFile = new File(outputFile.getName() + ".bak");
				boolean backup = true;
				if (bakFile.exists())
				{
					backup = bakFile.delete();
				}
				if (backup)
					outputFile.renameTo(bakFile);
			}
			
			try
			{
				_fw = new FileWriter(outputFile);
				
				StringBuffer sb = new StringBuffer();
				
				sb.append("timeStamp");
				sb.append(ENTRY_SEP);
				sb.append("depth");
				sb.append(ENTRY_SEP);
				sb.append("sessionId");
				sb.append(ENTRY_SEP);
				sb.append("threadId");
				sb.append(ENTRY_SEP);
				sb.append("user");
				sb.append(ENTRY_SEP);
				sb.append("action");
				sb.append(ENTRY_SEP);
				sb.append("Status");
				sb.append(ENTRY_SEP);
				sb.append("duration");
				sb.append(ENTRY_SEP);
				sb.append("advanced\n");
				
				_fw.append(sb.toString());
				_fw.flush();
			}
			catch (IOException e)
			{
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			_fileQueue = new LinkedList<_logInstance>() ;
		}
		
		void addToQueue(_logInstance log)
		{
			_fileQueue.offer(log);
		}
		
		void stop()
		{
			_quit = true;
		}
		
		void dump()
		{
			_logInstance log;
			
			log = _fileQueue.poll();
			if (log != null)
			{
				try
				{
					_fw.append(log.dump());
					_fw.flush();
					log.free();
				}
				catch (IOException e)
				{
				}
			}
		}
		
		public void run()
		{
			while (!_quit)
			{
				
				dump();
				try
				{
					Thread.sleep(100);
				}
				catch (InterruptedException e)
				{
				}
			}
			
			// stopping...
			// finish to write log
			while (!_fileQueue.isEmpty())
				dump();
			
			try
			{
				_fw.flush();
				_fw.close();
			}
			catch (IOException e)
			{
				e.printStackTrace();
			}
		}
		
	}
}
