/*
 * LyPrintWriter.java
 * Copyright (c) 2000-2011, W4.
 * All rights reserved.
 */

package leon.misc;

import java.io.*;


/**
 * LyPrintWriter contains utility methods to write files.
 */
public class LyPrintWriter extends PrintWriter
{
	public static final String __VERSION = "$Id$";

	/** Indentation level. */
	protected int			   _indent = 0;

	/** Indicates if the next line indentation must be print. */
	protected boolean		   _indentNext = false;

	/** Indicates if the current tag has changed. */
	protected boolean		   _changed = false;
	
	/** Indent string. */
	protected String		   _indentString = "\t";
	
	/** Number of lines. */
	protected int			   _nl = 0;

	/**
	 * Constructor. Calls the super implementation with a buffered writer builds from an output
	 * stream writer itself built with the given parameters. Sets {@link #_closeTag} and {@link
	 * #_changed} to false, then {@link #start() starts} the print.
	 * @param  output  the output stream where the file has to be wrote
	 * @param  charset the name of a supported charset
	 * @throws UnsupportedEncodingException if the named encoding is not supported
	 */
	protected LyPrintWriter(FileOutputStream output, String charset)
		throws UnsupportedEncodingException
	{
		super(new BufferedWriter(new OutputStreamWriter(output, charset)), false);
	}

	/**
	 * Constructor. Calls the super implementation, sets {@link #_closeTag} and {@link #_changed} to
	 * false, then {@link #start() starts} the print.
	 * @param output the file where the file will be written
	 */
	protected LyPrintWriter(FileOutputStream output)
	{
		super(output);
	}
	
	/**
	 * Gets the file path the given filename with the given extension at the given destination exits.
	 * @param  filename    the name of the file where the file will be written
	 * @param  destination the path of the file
	 * @param  type        the file extension type
	 * @return The file path.
	 */
	public static String getFilePath(String filename, String destination, String type)
	{
		String	path = LyString.replace(filename, "/", File.separator);

		if (destination != null)
			path = destination + File.separator + path;
		if ((type != null) && !path.endsWith(type))
			path += type;
		return path;
	}
	
	/**
	 * Checks if the given filename with the given extension at the given destination exits.
	 * @param  filename    the name of the file where the file will be written
	 * @param  destination the path of the file
	 * @param  type        the file extension type
	 * @return True if it exists, false otherwise.
	 */
	public static boolean checkFile(String filename, String destination, String type)
	{
		String	path = getFilePath(filename, destination, type);
		File	file = new File(path);
		
		return file.exists();
	}

	/**
	 * Gets the print writer associated to the given filename with the given extension at the given
	 * destination. If the file ever exist, a backup is made into the filename + type + ".bak" file.
	 * @param  filename    the name of the file where the file will be written
	 * @param  destination the path of the file
	 * @param  type        the file extension type
	 * @return the print writer
	 */
	public static LyPrintWriter getPrintWriter(String filename, String destination,
		String type)
	{
		FileOutputStream	output = getFileOutputStream(filename, destination, type);
		LyPrintWriter 		pw;

		try
		{
			pw = new LyPrintWriter(output, LyString.DEFAULT_CHARSET);
		}
		catch (UnsupportedEncodingException e)
		{
			pw = new LyPrintWriter(output);
		}

		return pw;
	} // end method getPrintWriter

	/**
	 * Gets the output stream corresponding to the given filename with the given extension at the given
	 * destination. If the file ever exist, a backup is made into the filename + type + ".bak" file.
	 * @param  filename    the name of the file where the file will be written
	 * @param  destination the path of the file
	 * @param  type        the file extension type
	 * @return the output stream
	 */
	protected static FileOutputStream getFileOutputStream(String filename, String destination,
		String type)
	{
		String path = getFilePath(filename, destination, type);
		int    n = path.lastIndexOf(File.separator);
		String dirname = ((n >= 0) ? path.substring(0, n) : ".");
		File   dir = new File(dirname);
		File   file = new File(path);
		File   bakFile = null;

		if (!dir.exists())
			dir.mkdirs();

		if (file.exists())
		{
			// Move existing file to .bak
			bakFile = new File(path + ".bak");
			if (bakFile.exists())
				bakFile.delete();

			// if we can not write the file, write the bak file
			if (file.canWrite())
				file.renameTo(bakFile);
			else
				file = bakFile;
		}

		// Open existing file or create a new one
		FileOutputStream	output = null;

		try
		{
			output = new FileOutputStream(file);
		}
		catch (IOException e)
		{
			e.printStackTrace();
			return null;
		}
		
		return output;
	}
	
	/**
	 * Utility methods to print, avoid ^M at the end of lines and to indent result.<br />
	 * If {@link #_closeTag} is set to <code>true</code> then print '&gt;' character and set the
	 * attribute to <code>false</code>.<br />
	 * If {@link #_indentNext} is set to true, print the number of tabulation of the {@link
	 * #_indent} parameter.<br />
	 * Calls the super implementation and set {@link #_changed} to <code>true</code>.
	 */
	public void print(String str)
	{
		if (_indentNext)
		{
			for (int i = 0; i < _indent; i++)
				super.print(_indentString);
			_indentNext = false;
		}

		super.print(str);

		_changed = true;
	}

	/**
	 * Prints the given string and the newline character ('\n'). Sets {@link #_indentNext} to <code>
	 * true</code>.
	 */
	public void println(String str)
	{
		print(str + '\n');
		_indentNext = true;
		_nl++;
	}

	/**
	 * Print the given String after calling {@link #println()} and printing two tabulations.
	 * @param str the string to print
	 */
	public void printnl(String str)
	{
		println();
		print(_indentString + _indentString + str);
	}

	/**
	 * Prints a newline character ('\n') and sets the {@link #_indentNext} parameter to <code>
	 * true</code>.
	 */
	public void println()
	{
		print('\n');
		_indentNext = true;
		_nl++;
	}

	/**
	 * Adds an indentation level.
	 */
	public void addIndent()
	{
		_indent++;
	}

	/**
	 * Removes an indentation level.
	 */
	public void removeIndent()
	{
		_indent--;
	}

	/**
	 * Indicates if the tag has changed or not.
	 * @return true if ths tag has changed, false otherwise
	 */
	public boolean wasChanged()
	{
		return _changed;
	}
	
	/**
	 * Gets indent string.
	 * @return The string used for indentation.
	 */
	public String getIndentString()
	{
		return _indentString;
	}
	
	/**
	 * Gets number of lines written on this writer.
	 * @return Number of lines written.
	 */
	public int getLinesCount()
	{
		return _nl;
	}
	
} // end class LyPrintWriter
