/*
 * LyProcessViewer.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.io.*;


/**
 * This class is used to execute an external process, capture its output and redirect it to the
 * standard console. If a script is used, all commands that require a user input (pause, etc.) will
 * block the process.
 * @version $Id: LyProcessViewer.java 34645 2012-03-14 11:07:05Z jlsaouli $
 * @author  Lyria
 */
public final class LyProcessViewer extends Thread
{
	public static final String __VERSION =
		"$Id: LyProcessViewer.java 34645 2012-03-14 11:07:05Z jlsaouli $";

	/** The running process. */
	protected Process		   _process;

	/** Return code of the process. */
	protected int			   _returnCode;

	/** Indicates if the process execution is a synchronous or asynchronous run. */
	protected boolean		   _waiting;

	/** Current environment. */
	protected LyEnvironment    _environment;

	/** Indicates if the current process is running or not. */
	protected boolean		   _isRunning;

	/** Refresh rate. */
	protected int			   _refreshRate;

	/**
	 * Constructor. Sets the current environment, waiting status and refreshRate with the given
	 * values. Creates the current process from the given Script then run() or start() it depentding
	 * on the waiting status. The result is sent to the standard output.
	 * @param environment the environment of the command
	 * @param script      system script which should be executed.
	 * @param waiting     indicates whether the call is waiting for process completion
	 * @param refreshRate refresh rate.
	 */
	public LyProcessViewer(LyEnvironment environment, String script, boolean waiting,
		int refreshRate)
	{
		this(environment, script, waiting, refreshRate, "LY_ERROR_CANNOT_SHOW_PDF");
		
	}

	/**
	 * Constructor. Sets the current environment, waiting status and refreshRate with the given
	 * values. Creates the current process from the given Script then run() or start() it depentding
	 * on the waiting status. The result is sent to the standard output.
	 * @param environment the environment of the command
	 * @param commands    command list which should be executed.
	 * @param waiting     indicates whether the call is waiting for process completion
	 * @param refreshRate refresh rate.
	 */
	public LyProcessViewer(LyEnvironment environment, String[] commands, boolean waiting,
			int refreshRate)
	{
		this(environment, commands, waiting, refreshRate, "LY_ERROR_CANNOT_SHOW_PDF");
	}
	
	/**
	 * Constructor. Sets the current environment, waiting status and refreshRate with the given
	 * values. Creates the current process from the given Script then run() or start() it depentding
	 * on the waiting status. The result is sent to the standard output.
	 * @param environment the environment of the command
	 * @param script      system script which should be executed.
	 * @param waiting     indicates whether the call is waiting for process completion
	 * @param refreshRate refresh rate.
	 * @param errorMessage The error message to show if the process fails.
	 */
	public LyProcessViewer(LyEnvironment environment, String script, boolean waiting,
		int refreshRate, String errorMessage)
	{
		this(environment, script, waiting, refreshRate, errorMessage, true);
	}

	/**
	 * Constructor. Sets the current environment, waiting status and refreshRate with the given
	 * values. Creates the current process from the given Script then run() or start() it depentding
	 * on the waiting status. The result is sent to the standard output.
	 * @param environment the environment of the command
	 * @param commands    command list which should be executed.
	 * @param waiting     indicates whether the call is waiting for process completion
	 * @param refreshRate refresh rate.
	 * @param errorMessage The error message to show if the process fails.
	 */
	public LyProcessViewer(LyEnvironment environment, String[] commands, boolean waiting,
			int refreshRate, String errorMessage)
	{
		this(environment, commands, waiting, refreshRate, errorMessage, true);
	}

	/**
	 * Constructor. Sets the current environment, waiting status and refreshRate with the given
	 * values. Creates the current process from the given Script then run() or start() it depentding
	 * on the waiting status. The result is sent to the standard output.
	 * 
	 * @param environment 	the environment of the command
	 * @param commands     	command list which should be executed.
	 * @param waiting     	indicates whether the call is waiting for process completion
	 * @param refreshRate 	refresh rate.
	 * @param errorMessage 	The error message to show if the process fails.
	 * @param raiseError 	if true, raise errors to the user
	 */
	public LyProcessViewer(LyEnvironment environment, String[] commands, boolean waiting,
			int refreshRate, String errorMessage, boolean raiseError)
	{
		_environment = environment;

		if (commands != null)
		{
			setName("BF AppEngine : script output viewer thread " + commands[0]);
			
			LyString string = environment.getString();

			for (int i = 0; i < commands.length; i++)
				if(!commands[i].equals("/c"))
					commands[i] = string.expand(commands[i]);

			try
			{
				_process = Runtime.getRuntime().exec(commands);
				StringBuffer messageLog = new StringBuffer("execute script :");
				for (int i = 0; i < commands.length; i++)
					messageLog.append(" ").append(commands[i]);
				environment.logTrace(messageLog.toString());
	
				_returnCode = 0;
				_waiting = waiting;
				_refreshRate = refreshRate;
				_isRunning = true;
	
				if (_waiting)
					run(); // Run Synchronous
				else
					start();
			}
			catch (IOException e)
			{
				_environment.writeException(e);
				if (raiseError)
				{
					if (errorMessage == null)
						errorMessage = "LY_ERROR_CANNOT_SHOW_PDF";
				
					_environment.raiseError(errorMessage);
				}
				
				_returnCode = -1;
			}
		}
	}
	
	/**
	 * Constructor. Sets the current environment, waiting status and refreshRate with the given
	 * values. Creates the current process from the given Script then run() or start() it depentding
	 * on the waiting status. The result is sent to the standard output.
	 * 
	 * @param environment 	the environment of the command
	 * @param script      	system script which should be executed.
	 * @param waiting     	indicates whether the call is waiting for process completion
	 * @param refreshRate 	refresh rate.
	 * @param errorMessage 	The error message to show if the process fails.
	 * @param raiseError 	if true, raise errors to the user
	 */
	public LyProcessViewer(LyEnvironment environment, String script, boolean waiting,
			int refreshRate, String errorMessage, boolean raiseError)
		{
			_environment = environment;

			setName("BF AppEngine : script output viewer thread " + script);

			if (script != null)
			{
				LyString string = environment.getString();

				script = string.expand(script);

				try
				{
					_process = Runtime.getRuntime().exec(script);
					_environment.logTrace("execute script : " + script);
	
					_returnCode = 0;
					_waiting = waiting;
					_refreshRate = refreshRate;
					_isRunning = true;
	
					if (_waiting)
						run(); // Run Synchronous
					else
						start();
				}
				catch (IOException e)
				{
	//				_environment.writeException(e);
	//				if (raiseError)
	//				{
	//					if (errorMessage == null)
	//						errorMessage = "LY_ERROR_CANNOT_SHOW_PDF";
	//				
	//					_environment.raiseError(errorMessage);
	//				}
	//				
					_returnCode = -1;
				}
			}
		}
	
	/**
	 * Constructor.<br />
	 * Calls {@link #LyProcessViewer(LyEnvironment, String, boolean, int)
	 * LyProcessViewer(environment, script, waiting, 500)} in order to set the refresh rate to
	 * 500ms.
	 * @param environment the environment of the command
	 * @param script      system script which should be executed.
	 * @param waiting     indicates whether the call is waiting for process completion
	 */
	public LyProcessViewer(LyEnvironment environment, String script, boolean waiting)
	{
		this(environment, script, waiting, 500);
	}

	/**
	 * Constructor.<br />
	 * Calls {@link #LyProcessViewer(LyEnvironment, String, boolean, int)
	 * LyProcessViewer(environment, script, waiting, 500)} in order to set the refresh rate to
	 * 500ms.
	 * @param environment the environment of the command
	 * @param commands    command list which should be executed.
	 * @param waiting     indicates whether the call is waiting for process completion
	 */
	public LyProcessViewer(LyEnvironment environment, String[] commands, boolean waiting)
	{
		this(environment, commands, waiting, 500);
	}

	/**
	 * Waits for process completion and redirect its outputs to standard output.
	 */
	public void run()
	{
		try
		{
			boolean running = true;

			while (running)
			{
				try
				{
					// Active waiting to check process output
					Thread.sleep(_refreshRate);
					printConsole();
					_returnCode = _process.exitValue();

					// If the process is still running, an exception occurs !
					running = false;
				}
				catch (IllegalThreadStateException e1)
				{
				}
			}

			printConsole();
			_isRunning = false;
		}
		catch (InterruptedException e)
		{
			_environment.writeException(e);
		}
	} // end method run

	/**
	 * Displays process output to standard output.
	 */
	public void printConsole()
	{
		if (_process == null)
			return;

		DataInputStream din;
		InputStream     in;
		StringBuffer    sb;

		// ** standard output queue
		try
		{
			in = _process.getInputStream();
			if ((in != null) && (_environment != null) && (_environment.out != null))
			{
				sb = new StringBuffer();
				din = new DataInputStream(in);
				while (din.available() > 0)
					sb.append((char)din.readByte());
				if (sb.length() > 0)
					_environment.out.print(sb.toString());
				_environment.out.flush();
			}
		}
		catch (IOException e)
		{
			_environment.writeException(e);
		}

		// ** Error output
		try
		{
			in = _process.getErrorStream();
			if ((in != null) && (_environment != null) && (_environment.err != null))
			{
				sb = new StringBuffer();
				din = new DataInputStream(in);
				while (din.available() > 0)
					sb.append((char)din.readByte());
				if (sb.length() > 0)
					_environment.err.print(sb.toString());
				_environment.err.flush();
			}
		}
		catch (IOException e)
		{
			_environment.writeException(e);
		}
	} // end method printConsole

	/**
	 * Abort the current process.
	 */
	public void abort()
	{
		if (!isRunning())
			return;

		_process.destroy();
	}

	/**
	 * Gets the process returnCode.
	 * @return 0 if process still running
	 */
	public int getReturnCode()
	{
		return _returnCode;
	}

	/**
	 * Checks if the process viewer is still running or not.
	 * @return true if the process is still running, false otherwise
	 */
	public boolean isRunning()
	{
		return _isRunning;
	}

	/**
	 * Gets process viewer environment.
	 * @return the environment
	 */
	public LyEnvironment getEnvironment()
	{
		return _environment;
	}

	/**
	 * Gets the process corresponding to command execution when used as console.
	 * @return the process
	 */
	public Process getProcess()
	{
		return _process;
	}
} // end class LyProcessViewer
