/*
 * LyResourceBundle.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;


/**
 * This class is used to manage the resources of an application. It is based on the resource bundle
 * mechanism used in the Java language. The resources must be contained in properties files (such as
 * strings_fr.properties). See the javadoc of ResourceBundle for more details on how to declare
 * these files). Its attributes are :
 * <UL>
 *   <LI>name<br>
 *     The name used to retrieve the properties file.</LI>
 *   <LI>application bundle<br>
 *     The resource bundle that contains the specific resources of the application.</LI>
 *   <LI>parent bundle<br>
 *     The resource bundle inherited by this resource bundle. Resources of the parent bundle are
 *     used only if the resource is not defined in the application bundle.</LI>
 *   <LI>environment<br>
 *     The environment of the resource bundle.</LI>
 *   <LI>locale<br>
 *     Indicates the language of the resource bundle.</LI>
 * </UL>
 * <p>This class must be inherited in order to manage a given type of resources (files, colors, strings...).</p>
 * @see     ResourceBundle
 */
public abstract class LyResourceBundle
{
	public static final String __VERSION =
		"$Id: LyResourceBundle.java 41151 2012-10-11 15:28:06Z bangiole $";

	/** List of all leonardi resource bundles (LyResourceBundle). */
	static private Vector<LyResourceBundle>    _lyBundles = new Vector<LyResourceBundle>(2);

	public static String                       PROPERTIES_DIR = "properties";

	public static String                       PACKAGE_DIR = null;

    public static boolean                      UPPER_CASE_KEYS_ONLY = true;


    /**
     * Finds the parent bundle for this bundle.
     * @return The parent resource bundle.
     */
    protected static LyResourceBundle findBundle(LyEnvironment environment, String name, Locale locale, String skin)
    {
        synchronized (_lyBundles)
        {
            LyResourceBundle bundle;
            int              i, n = _lyBundles.size();

            for (i = 0; i < n; i++)
            {
                bundle = _lyBundles.elementAt(i);
                if ((bundle._environment == environment)
                    && (bundle._name == name)
                    && (bundle._skin == skin)
                    && ((bundle._locale == locale)
                        || ((bundle._locale != null)
                            && (locale != null)
                            && bundle._locale.equals(locale))))
                    return bundle;
            }
        }

        return null;
    }

    /**
     * Find the son environments from specified parent
     * @param parent parent environment
     * @return son environments (which parent is the specified arg)
     */
    public static ArrayList<LyEnvironment> getSonEnvironments(LyEnvironment parent)
    {
        ArrayList<LyEnvironment> result = null;

        synchronized (_lyBundles)
        {
            // Duplicate the static Vector to avoid errors.
            int         i, n = _lyBundles.size();

            for (i = 0, n = _lyBundles.size(); i < n; i++)
            {
                LyResourceBundle bundle = _lyBundles.elementAt(i);

                if ((bundle._environment != null)
                    && (bundle._environment.getParent() == parent))
                {
                    if (result == null)
                        result = new ArrayList<LyEnvironment>();
                    else if (result.contains(bundle._environment))
                        continue;
                    result.add(bundle._environment);
                }
            }
        }

        return result;
    }

    /**
     * Free all resource bundles of the given environment.
     * @param environment the environment whose resource bundles must be freed
     */
    public static void removeBundles(LyEnvironment environment)
    {
        synchronized (_lyBundles)
        {
            // Duplicate the static Vector to avoid errors.
            Vector<LyResourceBundle>     bundles = new Vector<LyResourceBundle>(_lyBundles);
            LyResourceBundle             bundle;
            int                          i, n = bundles.size();

            for (i = 0; i < n; i++)
            {
                bundle = bundles.elementAt(i);
                if (bundle._environment == environment)
                    bundle.free();
            }
        }
    }

	/** The name used to retrieve the properties file. */
	protected String		                   _name;

	/** The name of the skin used to retrieve the properties files. */
	protected String                          _skin;

	/**
	 * The resource bundle inherited by this resource bundle. Resources of the parent bundle are
	 * used only if the resource is not defined in the application bundle.
	 */
	protected LyResourceBundle                _parent;

	/** The resource bundle current environment. */
	protected LyEnvironment                   _environment;

	/** Indicates the language of the resource bundle. */
	protected Locale		                  _locale;

	/** Map of values. */
    protected HashMap<String, String>         _values;

	/** Extra values. */
	protected HashMap<String, String>         _extraValues;


	/**
	 * Constructor with a given environment. Calls {@link #LyResourceBundle(LyEnvironment, String,
	 * Locale) LyResourceBundle(environment, null, null)}
	 * @param environment The environment of the bundle.
	 */
	public LyResourceBundle(LyEnvironment environment)
	{
		this(environment, null, null, null);
	}

	/**
	 * Constructor with a given environment and name. Calls {@link #LyResourceBundle(LyEnvironment,
	 * String, Locale) LyResourceBundle(environment, name, null)}
	 * @param environment The environment of the bundle.
	 * @param name        The name of the resource bundle.
	 */
	public LyResourceBundle(LyEnvironment environment, String name)
	{
		this(environment, name, null, null);
	}

	/**
	 * Constructor with a given environment, name and locale. Initializes the environment, name and
	 * locale with the given values then calls the {@link #init()} method.
	 * @param environment The environment of the bundle.
	 * @param name        The name of the resource bundle.
	 * @param locale      The locale used by the resource bundle.
	 */
	protected LyResourceBundle(LyEnvironment environment, String name, Locale locale, String skin)
	{
		_environment = environment;
		_name = (name == null ? null : name.intern());
		_locale = locale;
		_skin = (skin == null ? null : skin.intern());

        // Find the locale to be used
        if ((_locale == null)
                && (_environment != null))
            _locale = _environment.getLocale();

        if ((_skin == null)
                && (_environment != null))
            _skin = _environment.getSkin();

		load();

        // Retrieve parent bundle
        _parent = findParentBundle();

        _lyBundles.addElement(this);
	}

	/**
	 * Clear all resources of this bundle.
	 */
	public void free()
	{
		// Free all the sons of this bundle.
		synchronized (_lyBundles)
		{
			// Duplicate the static Vector to avoid errors.
			int                        i, n = _lyBundles.size();
			Vector<LyResourceBundle>   bundles = new Vector<LyResourceBundle>(n);

			for (i = 0; i < n; i++)
				bundles.addElement(_lyBundles.elementAt(i));

			LyResourceBundle bundle;

			for (i = 0, n = bundles.size(); i < n; i++)
			{
				bundle = bundles.elementAt(i);
				if (bundle._parent == this)
					bundle.free();
			}

			// Now we can remove this bundle from the list.
			_lyBundles.remove(this);
		}

		_values = null;
	    _extraValues = null;
		_parent = null;
		_environment = null;
		_locale = null;
	} // end method free

	/**
	 * Clears cache for this resource bundle.
	 * Reset the internal _lastKey and _lastValue attribute.
	 */
	public void clearCache()
	{
	}

	/**
     * Get a resource bundle corresponding to the given parameter.
     */
    private final ResourceBundle loadResourceBundle(String dir, String extension)
    {
        if (_skin != null)
            // Load first resources without skin
            loadResourceBundle(dir, extension, null);

        // Then load with skin
        return loadResourceBundle(dir, extension, _skin);
    }

	/**
	 * get a resource bundle corresponding to the given parameter.
	 * @param  dir       the classpath relative directory containing the property file
	 * @param  extension the extension of the property file
	 * @return a resource bundle if one was found.
	 */
	@SuppressWarnings("rawtypes")
	private ResourceBundle loadResourceBundle(String dir, String extension, String skin)
	{
		String    baseName = dir + "/" + PROPERTIES_DIR + "/" + (skin == null ? _name : skin + "/" + _name);

		if (extension != null)
		    baseName += extension;

		String            name = baseName + "_" + _locale.getLanguage();
		ResourceBundle    rb = null;
		ClassLoader       loader = _environment.getClassLoader();

		// Clear resource bundle cache
		try
		{
			// JDK1.6
			// clearCache method is invoked by reflection to allow compilation in java 1.5
			Method	method;

			if (loader == null)
			{
				method = ResourceBundle.class.getMethod("clearCache");
				method.invoke(null);
			}
			else
			{
				method = ResourceBundle.class.getMethod("clearCache", ClassLoader.class);
				method.invoke(null, loader);
			}
		}
		catch (Exception e)
		{
			// JDK1.5
			try
			{
				Field field = ResourceBundle.class.getDeclaredField("cacheList");

				field.setAccessible(true);
				((Map)field.get(ResourceBundle.class)).clear();
			}
			catch (Exception e2)
			{
			}
		}

		// Load resource bundle
		try
		{
		    if ("UTF-8".equals(LyString.DEFAULT_CHARSET))
		    {
		        // FFT03820 : manage UTF-8 properties files
    			if (loader != null)
    				rb = Utf8ResourceBundle.getBundle(baseName, _locale, loader);
    			else
    				rb = Utf8ResourceBundle.getBundle(baseName, _locale);
		    }
		    else
		    {
		        if (loader != null)
		            rb = ResourceBundle.getBundle(baseName, _locale, loader);
		        else
		            rb = ResourceBundle.getBundle(baseName, _locale);
		    }
		}
		catch (MissingResourceException e)
		{
			_environment.logTrace("BF AppEngine : Resource bundle file " + name + " was not found.");
			return null;
		}

		// Load all keys
		_environment.logTrace("BF AppEngine : Loading bundle file " + name + ".");

		if (_values == null)
		    _values = new HashMap<String, String>();

		Enumeration<String>   keys = rb.getKeys();
		String                key, value;

		while (keys.hasMoreElements())
		{
		    key = keys.nextElement();
		    value = rb.getString(key);
		    _values.put(key, value);
		}

		return rb;
	} // end method getResourceBundle

	/**
	 * Load method used to load resources of this bundle.
	 */
	protected void load()
	{
	    String    id = _environment.getId();

	    if (id == null)
	        return;

	    if ((_environment.getParent() != null)
	            && (id == _environment.getParent().getId()))
	        return;

		// Load specific resources (if existing)
		if (id.equals("leon"))
        {
		    ResourceBundle rb = null;
            // Loading leonardi resource bundle
		    if ((_skin == LyEnvironment.DEFAULT_SKIN)
		            || (findBundle(_environment, _name, _locale, LyEnvironment.DEFAULT_SKIN) == null))
		        rb = loadResourceBundle(id, null);
		    else
		        // Avoid to load twice resources without skin
		        rb = loadResourceBundle(id, null, _skin);

		    if (rb == null)
		    {
		        // maybe it's a eu.w4.ae module, try with this id...
		        if ((_skin == LyEnvironment.DEFAULT_SKIN)
	                    || (findBundle(_environment, _name, _locale, LyEnvironment.DEFAULT_SKIN) == null))
	                rb = loadResourceBundle("eu/w4/ae", null);
	            else
	                // Avoid to load twice resources without skin
	                rb = loadResourceBundle("eu/w4/ae", null, _skin);
		    }
            return;
        }

		String          parentId = _environment.getParent().getId();
	    String          dir = null, extension = null;
		String[]        tokens = LyString.stringToList(id, '.');
		int             i, n = tokens.length;
		ResourceBundle  rb = null;

		for (i = 0; i < n; i++)
		{
		    if (i == 0)
				dir = tokens[i];
			else
				dir += '/' + tokens[i];

            if (tokens[i].equals("leon")
                    && (i < n - 1)) // dtddoc uses leon.tools.dtddoc.leon as id
                continue;
            if (parentId.startsWith(dir.replace('/', '.')))
                continue;

            extension = "_" + tokens[i];
            rb = loadResourceBundle(dir, extension);

			if ((rb == null)
    			    && (PACKAGE_DIR != null)
    			    && !id.equals("leon")
    			    && !id.equals("eu.w4.ac"))
			    rb = loadResourceBundle(dir + "/" + PACKAGE_DIR, extension);
		}
	} // end method init

	/**
	 * Return the String corresponding to the key passed.
	 * @param  key The key to search
	 * @return The corresponding value or null if not found.
	 */
	public String get(String key)
	{
		if (key == null)
			return null;

        if (UPPER_CASE_KEYS_ONLY
            && !(this instanceof LyImages))
        {
            // Best guess: not a valid key if one of the 2 first characters is lower-case.
            if (key.length() > 1)
            {
                char  c0 = key.charAt(0), c1 = key.charAt(1);

                if (((c0 >= 'a') && (c0 <= 'z'))
                        || ((c1 >= 'a') && (c1 <= 'z')))
                    // Not a valid key
                    return null;
            }
        }

		String value = null;

		// Look in values
        if (_values != null)
			value = _values.get(key);

        if ((value != null)
            && (_extraValues != null)
            && (value.length() > 0)
            && !value.equals(key)
            && _extraValues.containsKey(key))
        {
            // The value may be a key of another resource
            String redirectedValue = get(value);

            if (redirectedValue != null)
                value = redirectedValue;
            return value;
        }

		// If the search was unsuccessful, try the parent bundle.
		if ((value == null)
		        && (_parent != null))
			value = _parent.get(key);

		return value;
	} // end method get

	/**
     * Check if the given key is found in this resource bundle.
     * @param key the key to search
     * @param checkAll if true, we check the existence in the application and in the AE properties. If false we only
     *                 look in the application
     * @return true if the key was found.
     */
    public boolean checkKey(String key, boolean checkAll)
    {
    	boolean goUp = true;
    	if (!checkAll)
    	{
    		if ((_parent == null) || _parent._environment._environment == null) // root env
    		{
    			goUp = false;
    		}
    	}
    	if ((_values != null) && _values.containsKey(key))
    		return true;
    	if (goUp && (_parent != null))
    		return _parent.checkKey(key, checkAll);
    	return false;
    } // end method hasKey

    /**
     * Check if the given key is found in this resource bundle.
     * @param key the key to search
     * @return true if the key was found.
     */
    public boolean hasKey(String key)
    {
       return (get(key) != null);
    } // end method hasKey

	/**
	 * Adds a resource in this resource bundle. Key and value must not be null.
	 * @param key   Key of the resource.
	 * @param value Value of this new resource.
	 */
	public void put(String key, String value)
	{
		if (key == null)
			throw new IllegalArgumentException("BF AppEngine : key parameter can't be null.");

	    if (_extraValues == null)
	        _extraValues = new HashMap<String, String>(1);
	    _extraValues.put(key, value);

		if (_values == null)
			_values = new HashMap<String, String>();
		_values.put(key, value);
	}

	/**
     * Removes a resource from this resource bundle.
     * @param key   Key of the removed resource.
     * @return Value of the removed resource.
     */
    public void remove(String key)
    {
        if (key == null)
            throw new IllegalArgumentException("BF AppEngine : key parameter can't be null.");

        if (_extraValues != null)
            _extraValues.remove(key);

        if (_values != null)
            _values.remove(key);
    }

	/**
	 * Gets the list of keys from extra resources, generic ResourceBundle and application
	 * ResourceBundle.
     * @param addParent Boolean indicating if parent keys are added or not in result.
	 * @return The enumeration of all this bundle keys.
	 */
	public Vector<String> getKeyList(boolean addParent)
	{
	    Vector<String> result = new Vector<String>();

		if (_values != null)
		    result.addAll(_values.keySet());
		if (addParent
		        && (_parent != null))
		    result.addAll(_parent.getKeyList(addParent));

		return result;
	}

	/**
	 * Gets the list of keys from extra resources, generic ResourceBundle and application
	 * ResourceBundle.
     * @param addParents Boolean indicating if parent keys are added or not in result.
	 * @return The enumeration of all this bundle keys.
	 */
	public Enumeration<String> getKeys(boolean addParent)
	{
		Vector<String> result = getKeyList(addParent);

		return result.elements();
	} // end method getKeys

	/**
     * Finds the parent bundle for this bundle.
     * @return The parent resource bundle.
     */
    protected LyResourceBundle findParentBundle()
    {
        if (_environment == null)
            return null;

        if (_environment.getParent() == null)
        {
            if ((_skin == LyEnvironment.DEFAULT_SKIN)
                    && (_locale == Locale.getDefault()))
                return null;

            return findBundle(_environment, _name, Locale.getDefault(), LyEnvironment.DEFAULT_SKIN);
        }

        LyResourceBundle    parent = findBundle(_environment.getParent(), _name, _locale, _skin);

        if (parent == null)
            parent = buildParentBundle();

        return parent;
    }


	/**
	 * Builds the parent bundle for this bundle. Must be overwritten in subclasses.
	 * @return The parent resource bundle.
	 */
	protected abstract LyResourceBundle buildParentBundle();

	/**
	 * Checks if this resource bundle contains any resource.
	 */
	protected boolean isEmpty()
	{
	    return (_values == null);
	}

	/**
	 * Save resource bundle.
	 */
	public void saveTo(String dir)
	{
	    if ((_values == null)
	           || (_values.size() == 0))
	       return;

	    if (dir == null)
	       dir = ".";

	    String  id = _environment.getId();
    	String  path = dir + '/' + id.replace('.', '/') + '/' + PROPERTIES_DIR + '/'
    	    + _name + id.substring(id.lastIndexOf('.'))+ "_" + _locale.getLanguage();

    	path = _environment.getString().expand(path);

    	String  dirPath = path.substring(0, path.lastIndexOf(File.separator));
        File    dirFile = null;

        dirFile = new File(dirPath);
        if (!dirFile.exists())
            dirFile.mkdirs();

        File    file = new File(path + ".properties");

    	if (file.exists())
        {
            // Create bak file
            File     bakFile = new File(path + ".bak");

            if (bakFile.exists())
                bakFile.delete();

            if (file.renameTo(bakFile))
                _environment.logInfo("File :" + path + " renamed to .bak");
        }

        // Open the file
        OutputStream output;
        PrintWriter  pw;

        try
        {
            output = new FileOutputStream(file);
            pw = new PrintWriter(output);
        }
        catch (FileNotFoundException e)
        {
            _environment.logError("Unable to open file :" + path + ".properties");
            return;
        }

        Vector<String>      keys = getKeyList(false);

        Collections.sort(keys);

        Iterator<String>    iterator = keys.iterator();
        String              key, value;

        while (iterator.hasNext())
        {
            key = iterator.next();
            value = _values.get(key);
            pw.println(key + '=' + value);
        }

        pw.close();
	}
} // end class LyResourceBundle

/**
 * UTF-8 ResourceBundle
 **/
abstract class Utf8ResourceBundle
{
    /**
     * Gets the unicode friendly resource bundle
     *
     * @param baseName
     * @see ResourceBundle#getBundle(String)
     * @return Unicode friendly resource bundle
     */
    public static final ResourceBundle getBundle(final String baseName)
    {
        return createUtf8PropertyResourceBundle(ResourceBundle
                .getBundle(baseName));
    }

    /**
     * Gets the unicode friendly resource bundle
     *
     * @param baseName
     *            the base name of the resource bundle, a fully qualified class
     *            name
     * @param locale
     *            the locale for which a resource bundle is desired
     * @param loader
     *            the class loader from which to load the resource bundle
     **/
    public static ResourceBundle getBundle(String baseName, Locale locale,
            ClassLoader loader)
    {
        return createUtf8PropertyResourceBundle(ResourceBundle.getBundle(
                baseName, locale, loader));
    }

    /**
     * Gets the unicode friendly resource bundle
     *
     * @param baseName
     *            the base name of the resource bundle, a fully qualified class
     *            name
     * @param locale
     *            the locale for which a resource bundle is desired
     **/
    public static ResourceBundle getBundle(String baseName, Locale locale)
    {
        return createUtf8PropertyResourceBundle(ResourceBundle.getBundle(
                baseName, locale));
    }

    /**
     * Creates unicode friendly {@link PropertyResourceBundle} if possible.
     *
     * @param bundle
     * @return Unicode friendly property resource bundle
     */
    private static ResourceBundle createUtf8PropertyResourceBundle(
            final ResourceBundle bundle)
    {
        if (!(bundle instanceof PropertyResourceBundle))
        {
            return bundle;
        }
        return new Utf8PropertyResourceBundle((PropertyResourceBundle) bundle);
    }

    /**
     * Resource Bundle that does the hard work
     */
    private static class Utf8PropertyResourceBundle extends ResourceBundle
    {
        /**
         * Bundle with unicode data
         */
        private final PropertyResourceBundle bundle;

        /**
         * Initializing constructor
         *
         * @param bundle
         */
        private Utf8PropertyResourceBundle(final PropertyResourceBundle bundle)
        {
            this.bundle = bundle;
        }

        @Override
        public Enumeration<String> getKeys()
        {
            return bundle.getKeys();
        }

        @Override
        protected Object handleGetObject(final String key)
        {
            final String value = bundle.getString(key);
            if (value == null)
                return null;
            else if (value.length() == 0)
                return value;
            try
            {
                String decodedVal = new String(value.getBytes("ISO-8859-1"),
                        "UTF-8");

                if ((!LyString.testUTF8String(decodedVal))
                        && (LyString.testUTF8String(value)))
                    return value;
                else
                    return decodedVal;
            } catch (final UnsupportedEncodingException e)
            {
                throw new RuntimeException("Encoding not supported", e);
            }
        }
    }
}