/*
 * LyString.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.io.*;
import java.net.*;
import java.security.*;
import java.util.*;

import leon.app.LyApplication;
import sun.security.action.*;


/**
 * This class offers general useful methods to manipulate Strings.
 * @version $Id: LyString.java 35682 2012-06-01 13:44:14Z kchenal $
 * @author  Lyria
 */
public class LyString
{
	public static final String    __VERSION =
		"$Id: LyString.java 35682 2012-06-01 13:44:14Z kchenal $";

	/** Hexadecimal table (used for hexadecimal conversion). */
	public static final char[]    HEXA_TABLE =
	{
		'0', '1', '2', '3', '4', '5', '6', '7',
		'8', '9', 'a', 'b', 'c', 'd', 'e', 'f',
	};

	/** String used to display choices in format. */
	protected static final String CHOICE_FORMAT_CHAR = ".";

	/** Character used to unspecialize delimiter. */
	public static char			  MODIFIER = '\\';

	/** String used to define a temporary delimiter. */
	public static String		  MODIFIER_STRING = "\1\2";

	/** Default charset used to read input streams. */
	public static String		  DEFAULT_CHARSET = "UTF-8";

	/** Default http protocol scheme. */
	public static String		  HTTP_PROTOCOL = "http://";

	/** Default https protocol scheme. */
    public static String          HTTPS_PROTOCOL = "https://";

	/** Non displayable separator. */
	public char					  SEP = (char)1;

	/** Displayable delimiter. */
	public char					  VISIBLE_SEP = '|';

	/** Displayable delimiter. */
	public char					  VISIBLE_SEP2 = ',';

	/** Displayable open delimiter associated to {@link #VISIBLE_SEP3_BIS}. */
	public String				  VISIBLE_SEP3 = " (";

	/** Displayable close delimiter associated to {@link #VISIBLE_SEP3}. */
	public String				  VISIBLE_SEP3_BIS = ")";

	/** Delimiter used when exporting data (between columns). */
	public char					  EXPORT_COLUMN_SEP = ';';

	/** First char used when exporting data to identify header line. */
	public String				  EXPORT_HEADER_MARK = "@";

	/** Line separator. */
	public String				  LINE_SEP;

	/** File separator. */
	public String				  FILE_SEPARATOR = File.separator;

	/** Current environment. */
	protected LyEnvironment		  _environment;

	/**
	 * Constructor with a given environment. Sets the current environment with the given value the
	 * sets the {@link #LINE_SEP} attribute with the environment resource.
	 * @param environment The environment of the LyString instance.
	 */
	public LyString(LyEnvironment environment)
	{
		_environment = environment;

		LINE_SEP = _environment.getEnv("line.separator");
	}

	/**
	 * Tokenize a given string into an array of tokens. Tokens are separated by the given delimiter.
	 * For example, the result of stringToList("token1/token2",'/') will be an array containing the
	 * "token1" and "token2" strings.
	 * @param  text      The string which contains tokens separated by a delimiter.
	 * @param  delimiter Used to separate tokens in the string.
	 * @return An array of tokens strings.
	 * @see    LyString#listToString(String[], String)
	 * @see    LyString#tokenizer(java.lang.String, char)
	 */
	public static String[] stringToList(String text, char delimiter)
	{
		if (text == null)
			return null;

		String sep = Character.toString(MODIFIER) + delimiter, sep2 = null;

		if (text.indexOf(sep) >= 0)
		{
			sep2 = MODIFIER_STRING;
			text = text.replace(sep, sep2);
		}

		String[] list;
		int		 l = text.length();
		int		 i, n, idx, idx2;

		if (l == 0)
		{
			list = new String[1];
			list[0] = new String("");

			return list;
		}

		// Count tokens
		for (n = 1, idx = 0, idx2 = 0; (idx < l) && ((idx2 = text.indexOf(delimiter, idx)) >= 0);
				n++, idx = idx2 + 1)
			;

		// create an array for n values
		list = new String[n];

		// fill tokens
		for (i = 0, idx = 0, idx2 = 0; (idx < l) && ((idx2 = text.indexOf(delimiter, idx)) >= 0);
				i++, idx = idx2 + 1)
			list[i] = text.substring(idx, idx2);
		if (i < n)
			list[n - 1] = text.substring(idx, l);

		if (sep2 != null)
		{
			for (i = 0; i < n; i++)
				list[i] = list[i].replace(sep2, sep);
		}
		return list;
	} // end method stringToList

	/**
	 * Tokenize a given string into an array of tokens. The tokens are separated by the given string
	 * delimiter. For example, the result of stringToList("token1::token2","::") will be an array
	 * containing the "token1" and "token2" strings.
	 * @param  text      The string which contains tokens separated by a delimiter.
	 * @param  delimiter Used to separate tokens in the string.
	 * @return An array of tokens strings.
	 * @see    LyString#listToString(String[], char)
	 * @see    LyString#tokenizer(java.lang.String, char)
	 */
	public static String[] stringToList(String text, String delimiter)
	{
		if (text == null)
			return null;

		String[] list;
		int		 l = text.length();
		int		 i, n, idx, idx2;

		if (l == 0)
		{
			list = new String[1];
			list[0] = new String("");

			return list;
		}

		// Count tokens
		for (n = 1, idx = 0, idx2 = 0; (idx < l) && ((idx2 = text.indexOf(delimiter, idx)) >= 0);
				n++, idx = idx2 + delimiter.length())
			;

		// create an array for n values
		list = new String[n];

		// fill tokens
		for (i = 0, idx = 0, idx2 = 0; (idx < l) && ((idx2 = text.indexOf(delimiter, idx)) >= 0);
				i++, idx = idx2 + delimiter.length())
			list[i] = text.substring(idx, idx2);
		if (i < n)
			list[n - 1] = text.substring(idx, l);

		return list;
	} // end method stringToList

	/**
	 * Concatenate the given tokens list as a String with the specified delimiter. For example, the
	 * result of listToString({"token1","token2",'/') will be the string "token1/token2".
	 * @param  list      The array of tokens.
	 * @param  separator The separator that must be used between the tokens in the result.
	 * @return The concatenation of tokens.
	 * @see    LyString#stringToList(String, String)
	 */
	public static String listToString(String[] list, char separator)
	{
		if ((list == null) || (list.length == 0))
			return null;

		int			 i, n = list.length,
			// Size is only an estimation
			size = n * ((list[0] == null) ? 10 : list[0].length());

		// Go through the table and concatenate tokens
		StringBuffer buffer = new StringBuffer(size);
		String		 str;
		String		 sep = Character.toString(separator);
		String		 sep2 = Character.toString(MODIFIER) + separator;

		for (i = 0; i < n; i++)
		{
			str = list[i];
			if (i > 0)
				buffer.append(separator);
			if (str != null)
			{
				if (str.indexOf(sep2) >= 0)
					str = str.replace(sep, sep2);
				buffer.append(str);
			}
		}

		return buffer.toString();
	} // end method listToString

	/**
	 * Concatenate the given tokens list as a String with the specified delimiter. For example, the
	 * result of listToString({"token1","token2",'::') will be the string "token1::token2".
	 * @param  list      The array of tokens.
	 * @param  separator The separator that must be used between the tokens in the result.
	 * @return The concatenation of tokens.
	 * @see    LyString#stringToList(String, char)
	 */
	public static String listToString(String[] list, String separator)
	{
		if ((list == null) || (list.length == 0))
			return null;

		int			 i, n = list.length,
			// Size is only an estimation
			size = n * ((list[0] == null) ? 10 : list[0].length());

		// Go through the table and concatenate tokens
		StringBuffer buffer = new StringBuffer(size);

		for (i = 0; i < n; i++)
		{
			if (list[i] != null)
				buffer.append(list[i]);

			if (i < (n - 1))
				buffer.append(separator);
		}

		return buffer.toString();
	}

	/**
	 * Gets the maximum length of a line in a text. Lines are separated by a carriage return
	 * (<code>\n</code>).
	 * @param  text The text that must be parsed.
	 * @return maximum The length of the greatest line.
	 */
	public static int getMaxLineWidth(String text)
	{
		if (text == null)
			return 0;

		int idx, idx2;
		int l = text.length();
		int max = 0;

		for (idx = 0, idx2 = 0; (idx < l) && ((idx2 = text.indexOf('\n', idx)) >= 0);
				idx = idx2 + 1)
		{
			if ((idx2 - idx) > max)
				max = idx2 - idx;

		}
		if ((l - idx) > max)
			max = l - idx;

		return max;
	}

	/**
	 * Replaces all occurrences of the given find token by the given replace token in the given text.
	 * @param  text    The original String that must be parsed and modified.
	 * @param  find    The token that should be replaced in the original string.
	 * @param  replace The token that should be used to replace the first token in the original
	 *                 string.
	 * @return The result of the replacement action.
	 */
	public static String replace(String text, String find, String replace)
	{
		if (text == null)
			return null;
		if (find == null)
			return text;
		if (replace == null)
			return text;

		return text.replace(find, replace);
	}

    /**
     * Replaces all occurrences of the given find token by the given replace token in the given text.
     * @param  text    The original String that must be parsed and modified.
     * @param  find    The tokens that should be replaced in the original string.
     * @param  replace The tokens that should be used to replace the tokens in the original string.
     * @return The result of the replacement action.
     */
    public static String replace(String text, char[] finds, char[] replaces)
    {
        if (text == null)
            return null;

        if (finds == null)
            return text;

        int       i, j, l = text.length();
        int       nf = finds.length;
        int       nr = (replaces == null ? 0 : replaces.length);
        char[]    buffer = text.toCharArray();
        char      c;

        for (i = 0; i < l; i++)
        {
            c = text.charAt(i);
            if (c == 0)
                continue;

            for (j = 0; (j < nf) && (c != finds[j]); j++);
            if ((j < nf) && (j < nr))
                buffer[i] = replaces[j];
        }

        return String.valueOf(buffer);
    }

	/**
     * Replaces all occurrences of the given find token by the given replace token in the given text.
     * @param  text    The original String that must be parsed and modified.
     * @param  find    The tokens that should be replaced in the original string.
     * @param  replace The tokens that should be used to replace the tokens in the original string.
     * @return The result of the replacement action.
     */
    public static String replace(String text, char[] finds, String[] replaces)
    {
        if (text == null)
            return null;

        if (finds == null)
            return text;

        int             i, j, l = text.length();
        int             nf = finds.length;
        int             nr = (replaces == null ? 0 : replaces.length);
        StringBuffer    buffer = new StringBuffer(l);
        char            c;

        for (i = 0; i < l; i++)
        {
            c = text.charAt(i);
            if  (c == 0)
                continue;

            for (j = 0; (j < nf) && (c != finds[j]); j++);
            if (j == nf)
                buffer.append(c);
            else if (j < nr)
                buffer.append(replaces[j]);
        }

        return buffer.toString();
    }

    static final char[]   psSearches = { '\\', '', '', '', '', '', '', '', '(', ')' };
	static final String[] psReplaces = { "\\\\", "\\340", "\\351", "\\350", "\\347",
	    "\\371", "\\352", "\\364", "\\(", "\\)" };

    /**
	 * Replaces non postscript chars with printable postscript chars in a string.
	 * @param  text The original string.
	 * @return The postScript printable string.
	 */
	public static String getPostscriptString(String text)
	{
	    return replace(text, psSearches, psReplaces);
	}

	/**
	 * Encodes the given string as an HTML string. Replace the non-authorized characters by their
	 * HTML code. For example, '&' will be replaced by '&amp;amp;', '\n' by '&lt;BR&gt;', etc.
	 * @param  text The original string.
	 * @return The HTML string (return "" if text was null).
	 * @see    #getHtmlString(String)
	 * @see    #getHtmlEntitiesString(String)
	 */
	public static String getHtmlString(String text)
	{
		// If text is null return "" (HTML representation of no value)
		if (text == null)
			return "";

		int n1, n2 = -1;

		n1 = text.indexOf("<![CDATA[");

		if (n1 >= 0)
		{
			n2 = text.indexOf("]]>", n1);

			if (n2 >= 0)
			{
				StringBuffer sb = new StringBuffer(text.length());

				sb.append(getHtmlString(text.substring(0, n1)));
				sb.append(text.substring(n1 + 9, n2)); // Disable encoding
				sb.append(getHtmlString(text.substring(n2 + 3))); // May
																  // recurse
				text = sb.toString();
			}
		}
		else
		{
			// First we replace all special character by the corresponding
			// entity
			text = getHtmlEntitiesString(text);

			// then we replace tab, CR by HTML Tag
			text = text.replace("\r\n", "<br />");
			text = replace(text,
			    new char[] { '\r', '\n', '\t' },
			    new String[] { "<br />", "<br />", "&nbsp;&nbsp;&nbsp;" } );
		}
		return text;
	} // end method getHtmlString

	/**
	 * Encodes the given string by encoding all ampersand characters. '&' is replaced by "&amp;amp;"
	 * except from the following sequences:
	 * <UL>
	 *   <LI>"&#D;" where D is a decimal number.</LI>
	 *   <LI>"&#xH;" or "&#XH;" where H is a hexadecimal number.</LI>
	 * </UL>
	 * @param  text The original string.
	 * @return the encoded string.
	 */
	public static String encodeAmpersand(String text)
	{
		if (text == null)
			return null;

		char		 c;
		String		 encodedAmpersand = "&amp;", sequence;
		StringBuffer result = new StringBuffer();
		int			 idx, idx2, idx3, idx4, l = text.length();

		for (idx = 0, idx2 = 0; (idx < l) && ((idx2 = text.indexOf('&', idx)) >= 0); idx = idx2 + 1)
		{
			result.append(text.substring(idx, idx2));
			// test the case of numeric character references.
			idx3 = idx2 + 1;
			if ((idx3 < (l - 2)) && (text.charAt(idx3) == '#')
					&& ((idx4 = text.indexOf(';', idx3)) >= 0))
			{
				// try to catch the sequence between &# and ;
				c = text.charAt(idx3 + 1);
				if ((c == 'x') || (c == 'X'))
					sequence = '#' + text.substring(idx3 + 2, idx4);
				else
					sequence = text.substring(idx3 + 1, idx4);

				// test if the sequence is a long int
				try
				{
					Long.decode(sequence);
					result.append('&');
				}
				catch (Exception e)
				{
					result.append(encodedAmpersand);
				}
			}
			else
				result.append(encodedAmpersand);
		}

		if (idx < l)
			result.append(text.substring(idx, l));
		// return result
		return result.toString();
	} // end method encodeAmpersand

    // XXX To be completed with cases of other languages
    static final char[] accentSearches = { '', '', '', '', '', '', '', '', '', '', '', '' };
    static final char[] accentReplaces = { 'a', 'a', 'e', 'e', 'e', 'e', 'i', 'i', 'u', 'o', 'o', 'c' };

    /**
     * Gets a string without accent from the given string.
     * @param  str String with accents.
     * @return String without accents.
     */
    public static String getSimpleString(String str)
    {
        return replace(str, accentSearches, accentReplaces);
    }

    static final char[]   heSearches = { '<', '>', '"', '\'' };
    static final String[] heReplaces = { "&lt;", "&gt;", "&quot;", "&#39;" };

	/**
	 * Encodes the given string as an HTML string. Replace the non-authorized characters by their
	 * HTML code. For example, '<' will be replaced by '&lt;'. This method will not replace
	 * '\n', '\t', etc.
	 * @param  text The original string.
	 * @return The HTML string (return "" if text was null).
	 */
	public static String getHtmlEntitiesString(String text)
	{
		// If text is null return "" (HTML representation of no value)
		if (text == null)
			return "";

		text = encodeAmpersand(text);

		text = replace(text, heSearches, heReplaces);

		// Specific case for the euro character (may be character 128)
		text = text.replace((char)128, '\u20AC');
		text = text.replace("\u20AC", "&euro;");

		// Specific case for "Microsoft Word quote character".
		text = text.replace("\u2019", "&#39;");

		return text;
	}

    static final char[]   jsSearches = { '\\', '<', '>', '"', '\'', '\n' };
    static final String[] jsReplaces = { "\\\\", "&lt;", "&gt;", "&quot;", "&#39;", "\\n"};

    /**
     * Encodes the given string as a javascript string. Replace non-authorized characters by their
     * HTML code. For example, '<' will be replaced by '&amp;lt;'. It will handle correctly &#39;
     * and &quot; in strings. All javascript string must be between double quote.
     * @param  text The original string.
     * @return The JavaScript string (return "" if text was null).
     */
    public static String getJavascriptString(String text)
    {
        // If text is null return "" (HTML representation of no value)
        if (text == null)
            return "";

        text = encodeAmpersand(text);
        text = replace(text, jsSearches, jsReplaces);
        return text;
    }

	/**
	 * encode a string to add into a javascript value like <input type="xxx" value="zzz"/>
	 * @param value input value
	 * @return encoded value
	 */
	public static String encodeValueInJavascript(String value)
	{
	    return replace(value,
            new char[] { '\\', '"' },
            new String[] { "\\\\", "\\\"" });
	}

	/**
	 * Decode a string to add into a javascript value like <input type="xxx" value="zzz"/>
	 * @param value input value
	 * @return encoded value
	 */
	public static String decodeValueFromJavascript(String value)
	{
        if (value == null)
            return null;

        String result;

        result = value.replace("\\\\", "\\");
        result = result.replace("\\\"", "\"");
        return result;
	}

	/**
	 * Encodes the given string as a javascript variable string. Replaces '.' by '_'.s
	 * @param  text The original string.
	 * @return The JavaScript variable string (return "" if text was null).
	 */
	public static String getJavascriptStringVar(String text)
	{
		// If text is null return "" (HTML representation of no value)
		if (text == null)
			return "";

		text = text.replace('.', '_');
		return text;
	}

	/**
	 * Encodes the given string as an XML string. Calls {@link #getXmlString(String, boolean,
	 * boolean) getXmlString(str, true, true)}
	 * @param  str The original string.
	 * @return The XML string (return "" if text was null).
	 * @see    #getXmlString(String, boolean, boolean)
	 */
	public static String getXmlString(String str)
	{
		return getXmlString(str, true, true);
	}

    static final char[]   xmlSearches = { '&', '<', '>', '\n', '"', '\'', '%', '', (char)3 };
    static final String[] xmlReplaces = { "&amp;", "&lt;", "&gt;", "&#10;", "&quot;", "&apos;", "&#37;", "&#8364;",
    	"" + LyApplication.FIELD_SEP };

	/**
	 * Encodes the given string as an XML string. Replace non-authorized characters by their XML
	 * code. For example, '&' will be replaced by '&amp;amp;', '%' by '&37;', etc.
	 * @param  str     The original string.
	 * @param  quotes  indicates if " and ' are encoded (quotes) : this is required only for
	 *                 attributes
	 * @param  percent indicates if % is encoded : this is required only for DTD
	 * @return The XML string (return "" if text was null).
	 */
	public static String getXmlString(String str, boolean quotes, boolean percent)
	{
        if (str == null)
            return "null";

        if (quotes && percent)
           return replace(str, xmlSearches, xmlReplaces);

        char[]   searches = new char[xmlSearches.length];
        String[] replaces = new String[xmlReplaces.length];
        int      i, n = 4;

        for (i = 0; i < n; i++)
        {
            searches[i] = xmlSearches[i];
            replaces[i] = xmlReplaces[i];
        }

        if (quotes)
        {
            searches[i] = xmlSearches[n];
            replaces[i] = xmlReplaces[n];
            searches[i + 1] = xmlSearches[n + 1];
            replaces[i + 1] = xmlReplaces[n + 1];
            i += 2;
        }

        if (percent)
        {
            searches[i] = xmlSearches[n + 2];
            replaces[i] = xmlReplaces[n + 2];
        }

        return replace(str, searches, replaces);
	}

    static final char[]   surrogateSearches =
    {(char) 1, (char) 2, (char) 4, (char) 5, (char) 6, (char) 7, (char) 8, (char) 11, (char) 12,
    	(char) 14, (char) 15, (char) 16, (char) 17, (char) 18, (char) 19, (char) 20, (char) 21,
    	(char) 22, (char) 23, (char) 24, (char) 25, (char) 26, (char) 27, (char) 28, (char) 29,
    	(char) 30, (char) 31};
    static final String[] surrogateReplaces =
    {"", "", "", "", "", "", "", "", "","", "", "", "", "", "", "", "",	"", "", "", "",
    	"", "", "", "",	"", ""};

    /**
	 * Encodes the given string as an XML string and delete all surrogate characters.
	 * Calls {@link #getXmlString(String, boolean,
	 * boolean) getXmlString(str, true, true)}
	 * @param  str The original string.
	 * @return The XML string (return "" if text was null).
	 * @see    #getXmlString(String, boolean, boolean)
	 */
	public static String getXmlStringStrict(String str)
	{
		String res = getXmlString(str, true, true);
		return replace(res, jsonSearches, jsonReplaces);
	}


	static final char[]   jsonSearches = 
		{'"', '\\', '/', '\b', '\f', '\n', '\r', '\t', (char)3};
	static final String[] jsonReplaces = 
		{"\\\"", "\\\\", "\\/", "\\b", "\\f", "\\n", "\\r", "\\t", "" + LyApplication.FIELD_SEP};

	/**
	 * Encodes the given string as an XML string and delete all surrogate characters.
	 * Calls {@link #getXmlString(String, boolean,
	 * boolean) getXmlString(str, true, true)}
	 * @param  str The original string.
	 * @return The XML string (return "" if text was null).
	 * @see    #getXmlString(String, boolean, boolean)
	 */
	public static String getJsonString(String str)
	{
		if(str == null)
		{
			return null;
		}
		str = str.replaceAll("\\\\u\\p{XDigit}\\p{XDigit}\\p{XDigit}\\p{XDigit}",
				"\\\\\\\\u\\p{XDigit}\\p{XDigit}\\p{XDigit}\\p{XDigit}");
		return replace(str, jsonSearches, jsonReplaces);
	}



	/**
	 * Format the given String depending on given parameters.
	 * @param  text        The original String to be formatted.
	 * @param  step        The index (number of chars) for carriage return (newLine). It also
	 *                     represents the length of each line in the generated string.
	 * @param  fill        Indicates if lines must be filled until their maximum length (step). If
	 *                     true, we fill the string with the specified character (fillingChar).
	 * @param  fillingChar The character used to fill the string if fill is true.
	 * @param  replaceCr   Indicates whether '\n' should be replaced with a space in the original
	 *                     String.
	 * @return the formatted string or null if an error occured or if original string is null or if
	 *         original string only contains spaces.
	 */
	public static String format(String text, int step, boolean fill, char fillingChar,
		boolean replaceCr)
	{
		if (text == null)
			return null;

		if (text.trim().equals(""))
			return null;

		if (replaceCr)
			text = text.replace('\n', ' ');

		int			 text_lg = text.length();
		StringBuffer result = new StringBuffer(2 * (text_lg + (text_lg / step)));
		String		 line;
		String		 punctuations = ":?., ;!";
		int			 cursor;
		int			 ptr, l, i;
		char		 cut, next;

		for (cursor = 0; (cursor + step) < text_lg; cursor += step)
		{
			line = text.substring(cursor, cursor + step);
			ptr = line.indexOf('\n');

			if (ptr == -1) // No Carriage Return for this line
			{
				// Cut on a white space, if it is not immediately
				// followed by a ponctuation mark.
				cut = text.charAt(cursor + step);
				next = (((cursor + step + 1) < text_lg) ? text.charAt(cursor + step + 1) : '\0');

				if (!Character.isWhitespace(cut) || (punctuations.indexOf(next) >= 0))
				{
					l = step;
					while ((l > 0)
							&& (!Character.isWhitespace(line.charAt(l - 1))
								|| (punctuations.indexOf(cut) >= 0)))
					{
						cut = line.charAt(l - 1);
						cursor--;
						l--;
					}
					if (l == 0)
						// step limit does not allow writing this word as
						// a whole
						return text;

					result.append(line.substring(0, l));
				}
				else
				{
					while (((cursor + step) < text_lg)
							&& Character.isWhitespace(text.charAt(cursor + step)))
						cursor++;
					result.append(line);
					l = step;
				}
			}
			else
			{
				result.append(line.substring(0, ptr));
				cursor += ptr + 1 - step;
				l = ptr;
			}

			if (fill)
			{
				for (i = step - l; i > 0; i--)
					result.append(fillingChar);
			}
			result.append('\n');
		}

		// Copy last block
		result.append(text.substring(cursor, text_lg));
		if (fill)
		{
			for (i = step - text_lg + cursor; i > 0; i--)
				result.append(fillingChar);
		}

		return result.toString();
	} // end method format

	/**
	 * Reads one line from an inputStream (until the end of line is found).
	 * @param  input The input stream.
	 * @return The next line or null if the end of file is detected.
	 */
	public static String getLine(InputStream input)
	{
	    // FFT 00037 / V3.3
		return getLine(input, LyString.DEFAULT_CHARSET);
	}

	// FFT 00037 / V3.3
	/**
	 * Reads one line from an inputStream (until the end of line is found).
	 * @param  input The input stream.
	 * @param charset expected charset to decode the inputstream
	 * @return The next line or null if the end of file is detected.
	 */
	public static String getLine(InputStream input, String charset)
	{
		_LyInput in = _LyInput.getInput(input, charset);

		if (in == null)
			return null;

		return in.getLine();
	}

	/**
	 * Closes the given InputStream. This method must be used if the {@link LyInput#getLine()} was
	 * called and the file was not readed till this end.
	 * @param input The input stream.
	 */
	public static void closeInput(InputStream input)
	{
		_LyInput in = _LyInput.getInput(input);

		if (in != null)
			in.close();
	}

	/**
	 * Sorts an array of strings. Uses the {@link String#compareTo(java.lang.String)} method to
	 * compare two strings. This method compares two strings lexicographically. The comparison is
	 * based on the Unicode value of each character in the strings.
	 * @param  strings The original array of strings.
	 * @return The sorted array of strings.
	 */
	public static String[] sortStrings(String[] strings)
	{
		// Basic checks
		if (strings == null)
			return null;
		int		 dim = strings.length;
		String[] sortedStrings = new String[dim];

		if (dim == 0)
			return sortedStrings;

		// Using vector instance to sort the array
		Vector<String> strs = new Vector<String>(dim);

		strs.addElement(strings[0]);

		boolean insert;
		int     k;

		for (int i = 1; i < dim; i++)
		{
			k = 0;
			insert = false;
			while ((k < strs.size()) && !insert)
			{
				if (strs.elementAt(k).compareTo(strings[i]) >= 0)
					insert = true;
				else
					k++;
			}
			strs.insertElementAt(strings[i], k);
		}

		// Re-build an array from the Vector
		for (int i = 0; i < dim; i++)
			sortedStrings[i] = strs.elementAt(i);
		return sortedStrings;
	} // end method sortStrings

	/**
	 * Gets the number of lines (separated by '\n') in a String.
	 * @param  text The string to analyse.
	 * @return The number of lines of the given string.
	 */
	public static int getNbLines(String text)
	{
		if (text == null)
			return 0;

		int idx, idx2, n;
		int l = text.length();

		for (n = 0, idx = 0, idx2 = 0; (idx < l) && ((idx2 = text.indexOf('\n', idx)) >= 0);
				n++, idx = idx2 + 1)
			;
		if (idx < l) // String is not terminated by \n
			n++;

		return n;
	}

	/**
	 * Encryption algorithm using SHA-1 from java.security. The original string is converted to a 20
	 * bytes array. The array of bytes is returned as its hexadecimal representation (String).
	 * @param  environment The environment whitin the encryption occured.
	 * @param  str         The original string.
	 * @return The encrypted string or null if an error occured.
	 */
	public static final String encrypt(LyEnvironment environment, String str)
	{
		if (str == null)
			return null;

		MessageDigest algorithm = null;
		byte[]		  str_b = null, result_b = null;
		String		  result;

		str_b = str.getBytes();
		try
		{
			algorithm = MessageDigest.getInstance("SHA-1");
			algorithm.reset();
			algorithm.update(str_b);
			result_b = algorithm.digest();
		}
		catch (NoSuchAlgorithmException e)
		{
			environment.logError(e.toString());
		}

		if (result_b != null)
			result = hexString(result_b);
		else
			result = null;

		return result;
	} // end method encrypt

	/**
	 * Converts an array of bytes to an hexadecimal String.
	 * @param  buf array of bytes to convert
	 * @return the hexadecimal String.
	 */
	protected static final String hexString(byte[] buf)
	{
		StringBuffer sb = new StringBuffer(buf.length * 2);
		int			 i;

		for (i = 0; i < buf.length; i++)
		{
			sb.append(HEXA_TABLE[(buf[i] >>> 4) & 15]);
			sb.append(HEXA_TABLE[buf[i] & 15]);
		}

		return sb.toString();
	}

	/**
	 * Creates randomly an identifier as a string (based on java.lang.Math.random() algorithm).
	 * @return A unique identifier.
	 */
	public static String getRandomId()
	{
		return Long.toHexString(Double.doubleToLongBits(Math.random()));
	}

	/**
	 * Tests if a string matches a regular expression.
	 * @param  environment environment whithin the test occured.
	 * @param  format      The regular expression.
	 * @param  value       The string that must be tested.
	 * @return true if the given string match the regular expression, false otherwise.
	 */
	public static boolean matchFormat(LyEnvironment environment, String format, String value)
	{
		if ((format == null) || (value == null))
			return true;

		return LyRegExpUtils.matchFormat(environment, format, value);
	}

	/**
	 * Tests if a format is valid.
	 * @param  environment environment whithin the test occured (parameter not used).
	 * @param  format      The format.
	 * @return true if the given format is valid, false otherwise.
	 */
	public static boolean checkStringFormat(LyEnvironment environment, String format)
	{
		if (format == null)
			return true;

		int     n0, n1, n2;
		int     minValue, maxValue;
		String  formatPart;
		String  min, max;
		boolean checked = true;

		do
		{
			n0 = format.indexOf('[');
			n1 = format.indexOf(']');

			if (n0 > 0)
			{
				// There is a fixed string
				format = format.substring(n0);
				continue;
			}
			else if (n0 < 0)
				// No more value -> fixed string
				break;

			formatPart = format.substring(n0 + 1, n1);
			n2 = formatPart.indexOf('-');
			min = formatPart.substring(0, n2);
			max = formatPart.substring(n2 + 1);

			try
			{
				minValue = Integer.parseInt(min);
				maxValue = Integer.parseInt(max);
			}
			catch (NumberFormatException e)
			{
				checked = false;
				break;
			}

			// If the min is greater than the max, stop and check is false...
			if (minValue > maxValue)
			{
				checked = false;
				break;
			}

			format = format.substring(n1 + 1);
		}
		while (format.length() != 0);

		return checked;
	} // end method checkStringFormat

	/**
	 * Strip given string : removes modifier from given string.
	 * @param  str      the String to strip
	 * @param  modifier the modifier to remove
	 * @return the given String without the given modifier
	 */
	public static String strip(String str, char modifier)
	{
		if (str == null)
			return null;

		int			 i, n = str.length();
		StringBuffer buffer = new StringBuffer();
		boolean		 joker = false;
		char		 c;

		for (i = 0; i < n; i++)
		{
			c = str.charAt(i);
			if ((c == modifier) && !joker)
				joker = true;
			else
			{
				buffer.append(c);
				joker = false;
			}
		}

		return buffer.toString();
	}

	/**
	 * Tests if a string match a simple format.
	 * @param  environment environment whithin the test occured (parameter not used).
	 * @param  format      The format.
	 * @param  value       The string that must be tested.
	 * @return true if the given string matches the format, false otherwise.
	 */
	public static boolean matchStringFormat(LyEnvironment environment, String format, String value)
	{
		if ((format == null) || (value == null))
			return true;

		// Check if the string is valid)
		int			 n0, n1, n2, n3, n4;
		int			 partLength;
		int			 minValue, maxValue, currentValue;
		int			 i, pos;
		String		 formatPart;
		String		 min, max;
		String		 valuePart;
		String		 fixedString;
		String		 currentFormat;
		String[]     formats;
		boolean		 checked = true;
		boolean		 isNumeric = true;
		boolean		 matching = false;
		char		 minChar;
		char		 maxChar;
		char		 currentChar;
		StringBuffer buffer;

		do
		{
			buffer = new StringBuffer(format);
			n0 = indexOf(buffer, '[', 0);
			n1 = indexOf(buffer, ']', n0);
			format = buffer.toString();

			if (n0 > 0)
			{
				// There is a fixed string
				fixedString = format.substring(0, n0);
				fixedString = strip(fixedString, MODIFIER);
				format = format.substring(n0);
				if (value.indexOf(fixedString) != 0)
				{
					// Fixed string was not found at the beginning of the
					// current value
					checked = false;
					break;
				}
				else
					value = value.substring(fixedString.length());

				continue;
			}

			if (n0 < 0)
			{
				// No more value -> fixed string
				if (!value.equals(format))
					checked = false;
				else
					value = ""; // FFT00792
				break;
			}

			buffer = new StringBuffer(format.substring(n0 + 1, n1));
			n2 = indexOf(buffer, '-', 0);
			n3 = indexOf(buffer, '|', 0);
			n4 = indexOf(buffer, ';', 0);
			formatPart = buffer.toString();

			if (n3 >= 0)
			{
				// it's a choice

				// Parse format
				formats = LyString.stringToList(formatPart, '|');
				currentFormat = "";

				for (i = 0; i < formats.length; i++)
				{
					// ignore empty choice format value
					if ((formats[i] == null) || (formats[i].trim().length() == 0))
						continue;

					if (formats[i].indexOf(CHOICE_FORMAT_CHAR) >= 0)
						formats[i] = formats[i].replace(CHOICE_FORMAT_CHAR, "");

					if (value.startsWith(formats[i]))
					{
						if (formats[i].length() > currentFormat.length())
							currentFormat = formats[i];
					}
				}

				if (currentFormat.length() == 0)
				{
					checked = false;
					break;
				}
				else
					valuePart = currentFormat;
			}
			else if (n4 >= 0)
			{
				// alphabetic intervals
				formats = LyString.stringToList(formatPart, ';');
				valuePart = value.substring(0, 1);
				currentChar = valuePart.charAt(0);
				matching = false;
				currentFormat = "";

				for (i = 0; i < formats.length; i++)
				{
					currentFormat = formats[i];
					pos = currentFormat.indexOf('-');
					min = currentFormat.substring(0, pos);
					max = currentFormat.substring(pos + 1);
					minChar = min.charAt(0);
					maxChar = max.charAt(0);

					if ((currentChar >= minChar) && (currentChar <= maxChar))
					{
						matching = true;
						break;
					}
				}

				if (!matching)
				{
					checked = false;
					break;
				}
			}
			else if (n2 < 0) // no dash '-' or dash out of range
			{
				formatPart = "[" + formatPart + "]";
				valuePart = value.substring(0, formatPart.length());

				// No more value -> fixed string which contains []
				if (!valuePart.equals(formatPart))
					checked = false;
			}
			else
			// if (n2 >= 0) // numeric or letter
			{
				min = formatPart.substring(0, n2);
				max = formatPart.substring(n2 + 1);

				if (max.length() >= min.length())
					partLength = max.length();
				else
					partLength = min.length();

				if (partLength > value.length())
				{
					checked = false;
					break;
				}

				valuePart = value.substring(0, partLength);
				try
				{
					minValue = Integer.parseInt(min);
					maxValue = Integer.parseInt(max);
					isNumeric = true;
				}
				catch (NumberFormatException e)
				{
					isNumeric = false;
					minValue = -1;
					maxValue = -1;
				}

				if (isNumeric)
				{
					try
					{
						currentValue = Integer.parseInt(valuePart);
					}
					catch (NumberFormatException e)
					{
						checked = false;
						break;
					}

					// If the current value is not in range
					// stop and check is false...
					if ((currentValue < minValue) || (currentValue > maxValue))
					{
						checked = false;
						break;
					}
				}
				else
				// alphaNumeric
				{
					if (partLength != 1)
					{
						checked = false;
						break;
					}

					minChar = min.charAt(0);
					maxChar = max.charAt(0);
					currentChar = valuePart.charAt(0);

					if (minChar > maxChar)
						return false;
					if (!Character.isLetter(currentChar))
					{
						checked = false;
						break;
					}
					if ((currentChar < minChar) && (currentChar > maxChar))
					{
						checked = false;
						break;
					}
				}
			}

			format = format.substring(n1 + 1);
			value = value.substring(valuePart.length());
		}
		while (format.length() != 0);

		// If value is longer than given format, return false
		if (value.length() > 0)
			checked = false;

		return checked;
	} // end method matchStringFormat

	/**
	 * Gets the index within the given string of the StringBuffer of the first occurrence of the
	 * specified delimiter that is not precede by the {@link #MODIFIER} character, starting the
	 * search at the specified index.
	 * @param  buffer    the buffer containing the string to analyse
	 * @param  delimiter the delimiter to find
	 * @param  fromIndex the position where to start the search
	 * @return the index of the first occurence of the delimiter not precede by {@link #MODIFIER}
	 *         from the given index
	 */
	public static int indexOf(StringBuffer buffer, char delimiter, int fromIndex)
	{
		if (fromIndex < 0)
			return -1;

		String str = buffer.toString();
		int    n = fromIndex;

		do
		{
			n = str.indexOf(delimiter, n);
			if ((n > 0) && (str.charAt(n - 1) == MODIFIER))
			{
				buffer.deleteCharAt(n - 1);
				str = buffer.toString();
			}
			else
				break;
		}
		while (n > 0);

		return n;
	}

	/**
	 * Checks the validity of the given format.
	 * @param  environment the environment (not used)
	 * @param  format      the format to check
	 * @return true if the format is valid, false otherwise
	 */
	public static boolean checkFormat(LyEnvironment environment, String format)
	{
		if ((format == null) || (format.length() == 0))
			return true;

		// Check if the string is valid)
		int			 n0, n1, n2, n3, n4;
		int			 minValue, maxValue;
		int			 i, pos;
		String		 formatPart;
		String		 min, max;
		String		 currentFormat;
		String[]     formats;
		boolean		 isNumeric = true;
		char		 minChar;
		char		 maxChar;
		StringBuffer buffer;

		do
		{
			buffer = new StringBuffer(format);
			n0 = indexOf(buffer, '[', 0);
			format = buffer.toString();

			if (n0 > 0)
			{
				// There is a fixed string
				format = format.substring(n0);
				continue;
			}

			n1 = indexOf(buffer, ']', n0);
			if ((n0 < 0) || (n1 < 0))
				// No more value
				break;

			buffer = new StringBuffer(format.substring(n0 + 1, n1));
			n2 = indexOf(buffer, '-', 0);
			n3 = indexOf(buffer, '|', 0);
			n4 = indexOf(buffer, ';', 0);
			formatPart = buffer.toString();

			if (n3 >= 0)
			{
				// it's a choice
				// Nothing to check... for now
			}
			else if (n4 >= 0)
			{
				// alphabetic intervals
				formats = LyString.stringToList(formatPart, ';');
				currentFormat = "";

				for (i = 0; i < formats.length; i++)
				{
					currentFormat = formats[i];
					pos = currentFormat.indexOf('-');
					min = currentFormat.substring(0, pos);
					max = currentFormat.substring(pos + 1);
					minChar = min.charAt(0);
					maxChar = max.charAt(0);

					if (minChar > maxChar) // min must be lower than max
						return false;
				}
			}
			else if (n2 < 0)
			{
				// There is a fixed string with '[' and ']'
				format = format.substring(n1 + 1);
				continue;
			}
			else
			// if (n2 >= 0) // numeric or letter
			{
				min = formatPart.substring(0, n2);
				max = formatPart.substring(n2 + 1);

				try
				{
					minValue = Integer.parseInt(min);
					maxValue = Integer.parseInt(max);
					isNumeric = true;
				}
				catch (NumberFormatException e)
				{
					isNumeric = false;
					minValue = -1;
					maxValue = -1;
				}

				if (isNumeric)
				{
					if (minValue > maxValue)
						return false; // min must me lower than max
				}
				else
				// alphaNumeric
				{
					minChar = min.charAt(0);
					maxChar = max.charAt(0);
					if (!Character.isLetter(minChar))
						return false; // min must be a letter

					if (!Character.isLetter(maxChar))
						return false; // min must be a letter

					if (minChar > maxChar) // min must be lower than max
						return false;
				}
			}

			format = format.substring(n1 + 1);
		}
		while (format.length() != 0);

		return true;
	} // end method checkFormat

	/**
	 * Decode a String which has multiple tokens separated by a character. This method only calls
	 * {@link #tokenizer(String, String) tokenizer(str, (new Character(delim)).toString())}
	 * @param  str   The string to parse into an array of tokens.
	 * @param  delim The delimiter between tokens.
	 * @return An array of strings which contains the tokens.
	 * @see    LyString#tokenizer(java.lang.String, java.lang.String)
	 */
	public static String[] tokenizer(String str, char delim)
	{
		return tokenizer(str, (new Character(delim)).toString());
	}

	/**
	 * Decode a String which has multiple tokens separated by a string. <b>Empty tokens</b> are <i>
	 * returned</i> as empty strings but not skipped like the java.util.StringTokenizer.
	 * @param  str   The string to parse into an array of tokens.
	 * @param  delim The delimiter between tokens.
	 * @return An array of strings which contains the tokens.
	 * @see    LyString#tokenizer(java.lang.String, char)
	 * @see    LyString#stringToList(String, String)
	 */
	public static String[] tokenizer(String str, String delim)
	{
		if ((str == null) || (delim == null) || (delim.length() == 0))
			return null;

		String[] tokenList;

		if (str.length() == 0)
		{
			tokenList = new String[1];
			tokenList[0] = str;

			return (tokenList);
		}
		StringTokenizer tokenizer = new StringTokenizer(str, delim, true);
		Vector<String>	resu = new Vector<String>(50);
		String		    token;
		// Indicates that the last item was a delimiter : check
		// for double delimiters
		boolean		    lastIsDelimiter = false;

		while (tokenizer.hasMoreTokens())
		{
			token = tokenizer.nextToken();

			// Check if the token is a 'real' token or a delimiter.
			if (token.equals(delim))
			{
				if (lastIsDelimiter) // two delimiters consecutively found !
					resu.addElement(""); // Add empty token.
				else
					lastIsDelimiter = true;
			}
			else
			{
				// a 'real' token is found.
				lastIsDelimiter = false;
				resu.addElement(token);
			}
		}

		// return the result as an array
		int n = resu.size();

		// Empty token list : return result 'as is'
		if (n == 0)
		{
			tokenList = new String[1];
			tokenList[0] = str;
		}
		else
		{
			// Copy result into the result.
			tokenList = new String[n];
			resu.copyInto(tokenList);
		}

		return (tokenList);
	} // end method tokenizer

	/**
	 * This utility method converts an URL defining the path of a file into a string that may be
	 * used in any WEB browser. This is used to display local images or files into HTML pages
	 * generated by Leonardi applications.<BR />
	 * <BR />
	 * <B>Example:</B><br />
	 * The incoming URL may be defined like this:
	 * <PRE>
	   C:\Documents and Settings\ma\My documents\leonardi.gif
	 * </PRE>
	 * <p>And the formatted URL will be:</p>
	 * <PRE>
	   file:///C|/Documents and Settings/%E9ma/My documents/leonardi.gif
	 * </PRE>
	 * <p>It can also be used for files retrieved from the network. The incoming URL may be defined
	 * like this:</p>
	 * <PRE>
	   \\ForeignPC\docs\notes.txt
	 * </PRE>
	 * <p>And the formatted URL will be:</p>
	 * <PRE>
	   file:////ForeignPC/docs/notes.txt
	 * </PRE>
	 * @param  url the incoming URL.
	 * @return the formatted URL.
	 */
    public static String encodeFileURL(String url)
	{
		StringBuffer encodedString = new StringBuffer("file:///");

		if (url == null)
			return encodedString.toString();

		if (url.startsWith("file://"))
			url = url.substring(7);

		if (url.startsWith("\\\\")) // case of documenst selected through the
									// network
		{
			encodedString.append('/');
			url = url.substring(2);
		}
		else
		// search for the hard drive characters : C:\
		{
			int sepPos = url.indexOf(':');

			if (sepPos >= 0)
			{
				encodedString.append(url.substring(0, sepPos)).append('|');
				if ((url.charAt(sepPos + 1) == '\\') || (url.charAt(sepPos + 1) == '/'))
				{
					encodedString.append('/');
					url = url.substring(sepPos + 2);
				}
				else
					url = url.substring(sepPos + 1);
			}
		}
		// replace '\' characters by '/'
		url = url.replace('\\', '/');

		// encode the url
		try
		{
			String dfltEncName
                = AccessController.doPrivileged(new GetPropertyAction("file.encoding"));

			url = URLEncoder.encode(url.toString(), dfltEncName);
		}
		catch (UnsupportedEncodingException e)
		{
		}

		// replace + by spaces
		url = url.replace('+', (char)32);
		// retrieves slashes
		int slashIndex = url.indexOf("%2F");

		while (slashIndex >= 0)
		{
			encodedString.append(url.substring(0, slashIndex)).append('/');
			url = url.substring(slashIndex + 3);
			slashIndex = url.indexOf("%2F");
		}
		encodedString.append(url);

		// return result
		return encodedString.toString();
	} // end method encodeFileURL

	/**
	 * Replaces all properties key found in the given String by their value found in the current
	 * environment.<br />
	 * Keys format must be $KEYS$. If the key is not found, it is replaced be the value specified in
	 * the <code>defaultStr</code>.
	 * @param  text       The original string, before any substitution.
	 * @param  defaultStr Default value for the property. If the property is not found, this value
	 *                    is used for the substitution.
	 * @return The string after all substitutions.
	 */
	protected String baseExpand(String text, String defaultStr)
	{
		if (text == null)
			return null;

		StringBuffer result = new StringBuffer();
		String		 name, value;
		int			 idx, idx2, idx3;
		int			 l = text.length();

		for (idx = 0, idx2 = 0, idx3 = 0; (idx < l) && ((idx2 = text.indexOf('$', idx)) >= 0);
				idx = idx3 + 1)
		{
			// Search next '$'
			idx3 = text.indexOf('$', idx2 + 1);
			if (idx3 == (idx2 + 1)) // replace '$$' with '$'
			{
				result.append(text.substring(idx, idx2 + 1));
				continue;
			}
			if (idx3 < 0)
				break;

			// Find the substitution in the property table
			name = text.substring(idx2 + 1, idx3);
			value = _environment.getEnv(name);

			if (idx < idx2)
				result.append(text.substring(idx, idx2));

			if ((value == null) || (value.length() == 0))
            {
				// try in resources
				value = _environment.getResource(name);

				// last chance : try in files
                if ((value == null) || (value.length() == 0))
                    value = _environment.getFile(name);
            }

			if ((value == null) || (value.length() == 0) || value.equals(name))
			{
				if (defaultStr != null)
					result.append(defaultStr);
			}
			else
				result.append(baseExpand(value, defaultStr));
		}
		if (idx < l)
			result.append(text.substring(idx, l));

		return result.toString();
	} // end method baseExpand

	/**
	 * Translates the given string into <code>application/x-www-form-urlencoded</code> format using
	 * a specific encoding scheme. This method uses the supplied encoding scheme to obtain the bytes
	 * for unsafe characters.
	 * @param  str String to be translated.
	 * @return the translated String.
	 * @see    URLEncoder#encode(String, String)
	 */
    public static String encodeURL(String str)
	{
		if (str == null)
			return null;

		String url = null;

		try
		{
			String dfltEncName = AccessController.doPrivileged(new GetPropertyAction("file.encoding"));

			url = URLEncoder.encode(str.toString(), dfltEncName);
		}
		catch (UnsupportedEncodingException e)
		{
		}

		return url;
	}

	/**
	 * Replaces all properties key found in the given String by their value found in the current
	 * environment.<br />
	 * Keys format must be $KEYS$. If the key is not found, it is replaced be the value specified in
	 * the <code>defaultStr</code>.<br />
	 * Also replaces '/' with File.separator mark for current OS.
	 * @param  text       The original string, before any substitution.
	 * @param  defaultStr Default value for the property. If the property is not found, this value
	 *                    is used for the substitution.
	 * @return The string after all substitutions.
	 * @see    #baseExpand(String, String)
	 */
	public String expand(String text, String defaultStr)
	{
		if (text == null)
			return null;

		if (text.startsWith(HTTP_PROTOCOL)
		        || text.startsWith(HTTPS_PROTOCOL))
			return expandUrl(text, defaultStr); // Detect URL pattern instead of File pattern

		StringBuffer result = new StringBuffer(baseExpand(text, defaultStr));

		if ((File.separatorChar != '/')
		    && !result.toString().startsWith(HTTP_PROTOCOL)
		    && !result.toString().startsWith(HTTPS_PROTOCOL))
		{
			String str = result.toString().replace("//", "\1\1");

			str = str.replace('/', File.separatorChar);
			str = str.replace("\1\1", "/");
			return str;
		}

		// For linux, change window pathes !
		if (File.separatorChar == '/')
		{
			String str = result.toString();
			str = str.replace('\\', File.separatorChar);

			return str;
		}

		return result.toString();
	}

	/**
	 * Replaces all properties key found in the given String by their value found in the current
	 * environment.<br />
	 * This method only calls {@link #expand(String, String) expand(text, null)}.
	 * @param  text The original string, before any substitution.
	 * @return The string after all substitutions.
	 */
	public String expand(String text)
	{
		return expand(text, null);
	}

	/**
	 * Replaces all properties key found in the given String by their value found in the current
	 * environment.<br />
	 * This method only calls {@link #expandUrl(String, String) expandUrl(text, null)}.
	 * @param  text The original string, before any substitution.
	 * @return The string after all substitutions.
	 */
	public String expandUrl(String text)
	{
		return expandUrl(text, null);
	}

	/**
	 * Replaces all properties key found in the given String by their value found in the current
	 * environment.<br />
	 * This method only calls {@link #baseExpand(String, String) baseExpand(text, defaultStr)} if
	 * the text is not null.
	 * @param  text       The original string, before any substitution.
	 * @param  defaultStr Default value for the property. If the property is not found, this value
	 *                    is used for the substitution.
	 * @return The string after all substitutions.
	 */
	public String expandUrl(String text, String defaultStr)
	{
		if (text == null)
			return null;

		StringBuffer result = new StringBuffer(baseExpand(text, defaultStr));

		if (File.separatorChar != '/')
		{
			String			str = result.toString();
			String			sep = File.separator;

			if (sep != null)
				str = str.replace(sep.charAt(0), '/');
			return str;
		}

		return result.toString();
	}

	/**
	 * In a string, enables replacement of patterns '%n' such as %1,%2, etc. The number indicates
	 * the position of the parameter in the array. If the position is not valid, the pattern is not
	 * replaced.
	 * @param  text   The original String or the key to find it.
	 * @param  params Parameters that should replace the associated patterns. The toString method is
	 *                called on these parameters (if a parameter is null "null" is used to replace
	 *                the corresponding pattern).
	 * @return A new String, after substitution of patterns.
	 * @see    LyString#compose(java.lang.String,java.lang.Object)
	 * @see    LyString#compose(java.lang.String,java.lang.Object, java.lang.Object)
	 * @see    LyString#compose(java.lang.String,java.lang.Object,
	 *         java.lang.Object,java.lang.Object)
	 */
	public String compose(String text, Object[] params)
	{
		if (text == null)
			return null;

		String str = _environment.translate(text);

		if (str == null)
			str = text;

		if ((params == null) || (params.length == 0))
			return str;

		StringBuffer result = new StringBuffer();
		String		 index;
		int			 idx, idx2, idx3, n;
		int			 l = str.length();
		char		 ch = '\0';
		String		 param;

		for (idx = 0, idx2 = 0, idx3 = 0, n = 0; (idx < l) && ((idx2 = str.indexOf('%', idx)) >= 0);
				idx = idx3)
		{
			for (idx3 = idx2 + 1; idx3 < l; idx3++)
			{
				ch = str.charAt(idx3);
				if (!Character.isDigit(ch))
					break;
			}

			if (idx3 == (idx2 + 1))
			{
				if (idx3 == l)
					break;

				if ((ch == 's') || (ch == 'd'))
					// Still allow %s and %d (%n should be used instead)
					idx3++;
				else
				{
					result.append(str.substring(idx, idx3));
					continue;
				}
			}
			else
			{
				// Get the index in the array
				index = str.substring(idx2 + 1, idx3);
				try
				{
					n = Integer.parseInt(index) - 1;
				}
				catch (NumberFormatException e)
				{
					// Should never happen
					n = -1;
				}
			}

			if ((n < 0) || (n >= params.length))
				result.append(str.substring(idx, idx3));
			else
			{
				result.append(str.substring(idx, idx2));
				if (params[n] == null)
					result.append("null");
				else
				{
					param = params[n].toString();
					param = _environment.translate(param);
					result.append(param);
				}
				n++;
			}
		}

		if (idx < l)
			result.append(str.substring(idx, l));

		return result.toString();
	} // end method compose

	/**
	 * In a string, enables replacement of one pattern (%1). This method calls {@link
	 * #compose(String, Object[]) compose(text, [param])}.
	 * @param  text  The original String.
	 * @param  param The parameter which should replace the associated pattern. The toString method
	 *               is called on this parameter (if the parameter is null "null" is used to replace
	 *               the pattern).
	 * @return A new String, after substitution of the pattern.
	 * @see    #compose(String, Object[])
	 */
	public String compose(String text, Object param)
	{
		Object[] params = new Object[1];

		params[0] = param;
		return compose(text, params);
	}

	/**
	 * In a string, enables replacement of two patterns (%1,%2). The number indicates the position
	 * of the parameter in the array. This method calls {@link #compose(String, Object[])
	 * compose(text, [param, param2])}.
	 * @param  text   The original String.
	 * @param  param  The first parameter that should replace the associated patterns.
	 * @param  param2 The second parameter that should replace the associated patterns.
	 * @return A new String, after substitution of patterns.
	 * @see    #compose(String, Object[])
	 */
	public String compose(String text, Object param, Object param2)
	{
		Object[] params = new Object[2];

		params[0] = param;
		params[1] = param2;

		return compose(text, params);
	}

	/**
	 * In a string, enables replacement of three patterns (%1,%2,%3). The number indicates the
	 * position of the parameter in the array. This method calls {@link #compose(String, Object[])
	 * compose(text, [param, param2, param3])}.
	 * @param  text   The original String.
	 * @param  param  The first parameter that should replace the associated patterns.
	 * @param  param2 The second parameter that should replace the associated patterns.
	 * @param  param3 The third parameter that should replace the associated patterns.
	 * @return A new String, after substitution of patterns.
	 * @see    #compose(String, Object[])
	 */
	public String compose(String text, Object param, Object param2, Object param3)
	{
		Object[] params = new Object[3];

		params[0] = param;
		params[1] = param2;
		params[2] = param3;

		return compose(text, params);
	}

	/**
	 * Gets the string that is the common part of all strings of the given list.
	 * @param  values a list of string
	 * @return if the list contain only one element, the single element, the common part of all
	 *         string of the list if any, null otherwise.
	 */
	static public String getCommonPart(Vector<String> values)
	{
		// if no string, return null
		if ((values == null) || (values.size() == 0))
			return null;

		// if only on string, return it
		if (values.size() == 1)
			return values.elementAt(0);

		// First get the two first strings
		String  first = values.elementAt(0);
		String  second = values.elementAt(1);

		int     minLength = Math.min(first.length(), second.length());
		boolean match = false;

		while ((minLength > 0) && !match)
		{
			match = first.regionMatches(0, second, 0, minLength);
			if (!match)
				minLength--;
		}

		if (!match) // nothing found
			return null;
		String matchString = first.substring(0, minLength);

		String temp;

		for (int i = 2; i < values.size(); i++)
		{
			temp = values.elementAt(i);
			if (temp.startsWith(matchString))
				continue;
			match = false;
			minLength = matchString.length();
			while ((minLength > 0) && !match)
			{
				minLength--;
				matchString = matchString.substring(0, minLength);

				match = temp.startsWith(matchString);
			}

			if (matchString.length() == 0)
				return null;
		}

		if (match)
			return matchString;

		return null;
	} // end method getCommonPart

    /**
     * Check if the String is not properly encoded and contains likely UTF-8
     * bytes
     *
     * @param value
     *            String to test
     * @return true if the String seems OK.
     */
    public static boolean testUTF8String(String value)
            throws java.io.UnsupportedEncodingException
    {
        for (int i = 0; value != null && i < value.length(); i++)
        {
            // char c = value.charAt(i);
            int val = value.charAt(i);
             /*if ((!Character.isLetterOrDigit(c)) && (c != ' ') && (c != '&')
                    && (c != '_') && (c != '"') && (c != '#') && (c != '\'')
                    && (c != '{') && (c != '(') && (c != '[') && (c != '-')
                    && (c != '|') && (c != '_') && (c != '\\') && (c != '@')
                    && (c != '') && (c != ')') && (c != ']') && (c != '+')
                    && (c != '=') && (c != '}') && (c != '') && (c != '$')
                    && (c != '') && (c != '*') && (c != '%') && (c != '<')
                    && (c != '>') && (c != '?') && (c != ',') && (c != '.')
                    && (c != ';') && (c != '/') && (c != ':') && (c != '!')
                    && (c != '<') && (c != '\r') && (c != '\n') && (c != '\t')
                    && (c != '~'))*/
                if (val== 65533) // last valid character
                   return false;
        }
        return true;
    }

    /**
     * Convert the Strings to bytes and analyse this using another charset
     * @param value string to check
     */
    public static String replaceCharSet(String value, String newCharSet)
    {
        String resu = value;
        if (value != null && value.length() > 0)
        {
                try
                {
                    byte[] b = new byte[value.length()];
                    for (int i=0; i<b.length; i++)
                        b[i] = (byte)value.charAt(i);
                    resu = new String(b,newCharSet);
                } catch (java.io.UnsupportedEncodingException uee)
                {
                //wrong encoding!
                }
        }
        return resu;
    }

    /**
     * Replace common misread characters to their UTF-8 equivalent
     *
     * @param s
     *            String to replace
     * @return the result string
     */
    public static String revertToUTF8(String s)
    {
        if (s != null)
        {
            s = replace(s, "é", "");
            s = replace(s, "à", "");
            s = replace(s, "è", "");
            s = replace(s, "ê", "");
            s = replace(s, "î", "");
            s = replace(s, "î", "");
            s = replace(s, "ù", "");
            s = replace(s, "ç", "");
            s = replace(s, "ô", "");
            s = replace(s, "ï", "");

        }
        return s;
    }
} // end class _LyString


/**
 * Internal class used to read files.
 */
class _LyInput
{

	public static final String __VERSION = "$Id: LyString.java 35682 2012-06-01 13:44:14Z kchenal $";

	/** List of LyInput instances. */
	static Vector<_LyInput>	   _inputs = new Vector<_LyInput>(5);

	/** The input stream associated to this LyInput. */
	InputStream				   _input = null;

	/** The input stream reader associated to this LyInput. */
	InputStreamReader		   _reader = null;

	/** The buffer reader associated to this LyInput. */
	BufferedReader			   _buff_reader = null;

	/**
	 * Constructor with a given InputStream. Calls
	 * {@link #LyString(InputStream, String)} with the given input and the
	 * default charset {@link LyString#DEFAULT_CHARSET}.
	 *
	 * @param input The InputStream of the LyInput instance.
	 */
	_LyInput(InputStream input)
	{
		// FFT 00037 / V3.3
		this(input, LyString.DEFAULT_CHARSET);
	}

	// FFT 00037 / V3.3
	/**
	 * Constructor with a given InputStream. Initializes the input stram with
	 * the given value, try to create the input stream reader from it and with
	 * the given charset. Then creates the buffer reader from the input stream
	 * reader and adds this new LyInput to the list of intputs.
	 *
	 * @param input The InputStream of the LyInput instance.
	 * @param charset the inputstream charset encoding
	 */
	_LyInput(InputStream input, String charset)
	{
		_input = input;
		try
		{
			if (charset != null)
				_reader = new InputStreamReader(_input, charset);
			else
				_reader = new InputStreamReader(_input, LyString.DEFAULT_CHARSET);
		}
		catch (UnsupportedEncodingException e)
		{
			_reader = new InputStreamReader(_input);
		}

		_buff_reader = new BufferedReader(_reader);
		synchronized (_inputs)
		{
			_inputs.addElement(this);
		}
	}

	/**
	 * Reads one line from the inputStream (until the end of line is found). If the end of stream is
	 * detected or if an IO exception occurs, calls the {@link #close()} method.
	 * @return The next line, or null if end of stream is detected.
	 */
	String getLine()
	{
		try
		{
			String str = _buff_reader.readLine();

			if (str == null)
				close();
			return str;
		}
		catch (IOException e)
		{
			close();
			return null;
		}
	}

	/**
	 * Closes the InputStream and removes the LyInput instance from the static list of LyInput
	 * instances.
	 */
	void close()
	{
		try
		{
			_buff_reader = null;
			_reader.close();
			_input.close();
		}
		catch (IOException e)
		{
		}

		synchronized (_inputs)
		{
			_inputs.removeElement(this);
		}
	}

	/**
	 * Retrieves the LyInput associated to the given InputStream.
	 * @param  input the input stream
	 * @return The LyInput instance associated to the given InputStream.
	 */
	static _LyInput getInput(InputStream input)
	{
		// FFT 00037 / V3.3
		return getInput(input, LyString.DEFAULT_CHARSET);
	}

	// FFT 00037 / V3.3
	/**
	 * Retrieves the LyInput associated to the given InputStream.
	 * @param  input the input stream
	 * @return The LyInput instance associated to the given InputStream.
	 */
	static _LyInput getInput(InputStream input, String charset)
	{
		synchronized (_inputs)
		{
			_LyInput in;
			int      i, n = _inputs.size();

			for (i = n - 1; i >= 0; i--)
			{
				in = _inputs.elementAt(i);
				if (in._input == input)
					return in;
			}

			// FFT 00037 / V3.3
			return new _LyInput(input, charset);
		}
	}

	/**
	 * Gets the string that is the common part of all strings of the given list.
	 * @param  values a list of string
	 * @return if the list contain only one element, the single element, the common part of all
	 *         string of the list if any, null otherwise.
	 */
	static public String getCommonPart(Vector<String> values)
	{
		// if no string, return null
		if ((values == null) || (values.size() == 0))
			return null;

		// if only on string, return it
		if (values.size() == 1)
			return values.elementAt(0);

		// First get the two first strings
		String  first = values.elementAt(0);
		String  second = values.elementAt(1);

		int     minLength = Math.min(first.length(), second.length());
		boolean match = false;

		while ((minLength > 0) && !match)
		{
			match = first.regionMatches(0, second, 0, minLength);
			if (!match)
				minLength--;
		}

		if (!match) // nothing found
			return null;
		String matchString = first.substring(0, minLength);

		String temp;

		for (int i = 2; i < values.size(); i++)
		{
			temp = values.elementAt(i);
			if (temp.startsWith(matchString))
				continue;
			match = false;
			minLength = matchString.length();
			while ((minLength > 0) && !match)
			{
				minLength--;
				matchString = matchString.substring(0, minLength);

				match = temp.startsWith(matchString);
			}

			if (matchString.length() == 0)
				return null;
		}

		if (match)
			return matchString;

		return null;
	} // end method getCommonPart
} // end class LyInput
