/*
 * LyStrings.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.text.*;
import java.util.*;


/**
 * This class is used to manage the strings resources of an application. It is based on the
 * mechanism of LyResourceBundle.
 * @version $Id: LyStrings.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyStrings extends LyResourceBundle
{

	public static final String __VERSION =
		"$Id: LyStrings.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Constructor with a given environment. Calls {@link #LyStrings(LyEnvironment, String)
	 * LyStrings(environment, null)} constructor.
	 * @param environment The environment of the strings bundle.
	 */
	public LyStrings(LyEnvironment environment)
	{
		this(environment, null);
	}

	/**
	 * Constructor with a given environment and name. Calls the {@link #LyStrings(LyEnvironment,
	 * String, Locale) LyStrings(environment, name, environment.getLocale())} constructor.
	 * @param environment The environment of the strings bundle.
	 * @param name        The name of the strings bundle.
	 */
	public LyStrings(LyEnvironment environment, String name)
	{
		this(environment, name, environment.getLocale(), environment.getSkin());
	}
	
	/**
	 * Constructor with a given environment and name. Calls the {@link #LyStrings(LyEnvironment,
	 * String, Locale) LyStrings(environment, name, environment.getLocale())} constructor.
	 * @param environment The environment of the strings bundle.
	 * @param name        The name of the strings bundle.
	 * @param skin        The name of the skin
	 */
	public LyStrings(LyEnvironment environment, String name, String skin)
	{
		this(environment, name, environment.getLocale(), skin);
	}

	/**
	 * Constructor with a given environment, name and locale. Calls the super implementation.
	 * @param environment The environment of the strings bundle.
	 * @param name        The name of the strings bundle.
	 * @param locale      The locale used by the strings bundle.
	 */
	protected LyStrings(LyEnvironment environment, String name, Locale locale)
	{
		this(environment, name, locale, environment.getSkin());
	}
	
	/**
	 * Constructor with a given environment, name and locale. Calls the super implementation.
	 * @param environment The environment of the strings bundle.
	 * @param name        The name of the strings bundle.
	 * @param locale      The locale used by the strings bundle.
	 * @param skin        The name of the skin
	 */
	protected LyStrings(LyEnvironment environment, String name, Locale locale, String skin)
	{
		super(environment, (name == null ? "strings" : name), locale, skin);
	}

	/**
	 * Builds the parent strings bundle.
	 * @return The parent strings bundle for this strings bundle.
	 */
	protected LyResourceBundle buildParentBundle()
	{
		if (_environment == null)
			return null;

		return new LyStrings(_environment.getParent(), _name, _locale, _skin);
	}

	/**
	 * Gets the string corresponding to the given key.
	 * @param  key The key of the string.
	 * @return The string corresponding to the key if it is found, the key otherwise
	 */
	public String getString(String key)
	{
		String value = get(key);

		// If not found return the key
		if (value == null)
			return key;

		if ((value.length() > 0) && (value.charAt(0) == '$'))
			value = getString(value.substring(1));

		int m = value.indexOf('&');

		if ((m >= 0) && ((m + 1) < value.length()) && (value.charAt(m + 1) != ' '))
			// remove mnemonic
			value = value.substring(0, m) + value.substring(m + 1);

		return value;
	}

	/**
	 * Gets the mnemonic of the string corresponding to the given key.
	 * @param  key The key of the string.
	 * @return The mnemonic of the string corresponding to the given key if it is found, null
	 *         otherwise
	 */
	public Character getMnemonic(String key)
	{
		String value = get(key);

		if (value == null)
			return null;

		if ((value.length() > 0) && (value.charAt(0) == '$'))
			value = get(value.substring(1));

		int m = value.indexOf('&');

		return (((m < 0) || ((m + 1) == value.length()) || (value.charAt(m + 1) == ' ') || (value.charAt(m + 1) == '&')) ? null
														 : new Character(value.charAt(m + 1)));
	}

	/**
	 * Gets the integer corresponding to the given key.
	 * @param  key The key of the integer.
	 * @return The integer corresponding to the key if it is found and is an integer value, 0 if it
	 *         is found and is not an integer value, -1 otherwise
	 */
	public int getInt(String key)
	{
		String value = get(key);
		int    result;

		// If not found return -1
		if (value == null)
			return -1;

		// Parse the result
		result = LyTools.intFromString(value);
		return result;
	}

	/**
	 * Returns the long corresponding to the key passed.
	 * @param  key The key of the long.
	 * @return The long corresponding to the key if it is found and is a long value, 0 if it is
	 *         found and is not a long value, -1 otherwise
	 */
	public long getLong(String key)
	{
		String value = get(key);
		long   result;

		// If not found return -1
		if (value == null)
			return -1;

		// Parse the result
		result = LyTools.longFromString(value);
		return result;
	}

	/**
	 * Gets a message formated with the specified arguments corresponding to the given key. See
	 * {@link java.text.MessageFormat} for syntax.
	 * @param  key       the string key
	 * @param  arguments the list of message argments
	 * @return the formatted string, the key if it is not found, null if the key is null.
	 */
	public String getString(String key, Object[] arguments)
	{
		if (key == null)
			return null;

		String message = getString(key);

		if (message == null)
		    return null;
		if (arguments == null)
			return message;

		// Double quotes to avoid their removal
		message = LyString.replace(message, "'", "''");

		// Format message with the given arguments
		MessageFormat format = new MessageFormat(message);

		// Ensure that args are not null
		int			  i, n = ((arguments == null) ? 0 : arguments.length);

		for (i = 0; i < n; i++)
		{
			if (arguments[i] == null)
				arguments[i] = "";
		}

		return format.format(arguments);
	}
} // end class LyStrings
