/*
 * LyTimer.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

/**
 * This class implements a simple timer which notify a client when timeout expires. The client can
 * stop the timer with the stop() method.
 */
public class LyTimer
	implements Runnable
{

	public static final String		 __VERSION =
		"$Id: LyTimer.java 34507 2012-03-13 09:11:03Z jlsaouli $";

	/** Internal status: STARTING. */
	private final static int		 STARTING = 1;

	/** Internal status: STOPPED. */
	private final static int		 STOPPED = 2;
	
	/** Internal status: STOPPED. */
	private final static int		 STOPPING = 3;

	/** Internal status: RUNNING. */
	private final static int		 RUNNING = 4;

	/** Timeout duration before client notification in ms. */
	protected long					 _duration;

	/** Timer name. */
	protected String				 _name;

	/** Client notified when timeout occurs. */
	protected LyTimerClientInterface _client;

	/** Indicates that timer is started but not stopped. */
	protected boolean				 _running;

	/** Client data : freely set by Timer user. */
	protected Object				 _clientData;

	/** Internal use only (lock) current Thread. */
	protected Thread				 _currentThread;

	/**
	 * Timer Status. Possible values are:
	 * <ul>
	 *   <li>{@link #STARTING}</li>
	 *   <li>{@link #STOPPED}</li>
	 *   <li>{@link #STOPPING}</li>
	 *   <li>{@link #RUNNING}</li>
	 * </ul>
	 */
	protected int					 _status;

	/** Indicates if the timer needs to restart or not. */
	protected boolean				 _needRestart;

	/** Indicates if the timer needs to stop or not. */
	protected boolean				 _needStop;

	/**
	 * Default constructor. Calls the constructor {@link #LyTimer(LyTimerClientInterface, long,
	 * String) LyTimer(client, duration, "BF AppEngine : timer thread")}
	 * @param client   Thread listener client
	 * @param duration expected timer duration
	 */
	public LyTimer(LyTimerClientInterface client, long duration)
	{
		this(client, duration, "BF AppEngine : timer thread");
	}

	/**
	 * Default constructor. Initializes duration, client and name with the given values, then
	 * initializes the status to {@link #STOPPED}, needRestart and need Stop to false.
	 * @param client   Thread listener client
	 * @param duration expected timer duration
	 * @param name     Name of the timer
	 */
	public LyTimer(LyTimerClientInterface client, long duration, String name)
	{
		_duration = duration;
		_client = client;
		_name = name;
		_status = LyTimer.STOPPED;
		_needRestart = false;
		_needStop = false;
	}

	/**
	 * Change the timer timeout duration. This modification will only happen after the current time
	 * interval is finished.
	 * @param duration the new timer duration
	 */
	public void setDuration(int duration)
	{
		_duration = duration;
	}

	/**
	 * Run the timer.
	 * @see java.lang.Thread#run()
	 */
	public void run()
	{
		if ((_client == null) || (_duration < 0))
			return;

		long    iteration = _duration / 100; // get the number of 100 ms
		boolean notify = true;

		while (iteration >= 0)
		{
			_status = LyTimer.RUNNING;
			_running = true;

			iteration--;

			try
			{
				Thread.sleep(100);
			}
			catch (InterruptedException e)
			{
				e.printStackTrace();
			}

			if (_needRestart)
			{
				_needRestart = false;
				_needStop = false;
				iteration = _duration / 100;
			}
			if (_needStop)
			{
				// don't change needStop
				notify = false;
				break;
			}
		}

		// Notify clients
		if (notify) // (_running) && (_client != null) && (_currentThread != null))
			_client.endTimer(this);
		
		synchronized (_currentThread)
		{
			_running = false;
			_status = LyTimer.STOPPED;
			_currentThread = null;
		}
	} // end method run

	/**
	 * Start this timer.
	 */
	public void start()
	{
		if (_status == LyTimer.STARTING) // already starting
			return;

		_status = LyTimer.STARTING;

		if (_currentThread != null)
			throw (new IllegalStateException("Timer is already running !"));

		_currentThread = new Thread(this, _name);
		_currentThread.start();
	}

	/**
	 * Stop this timer. Will not notify client.
	 */
	public void stop()
	{
	    _status = LyTimer.STOPPING;
		_needStop = true;
		_needRestart = false;
	}

	/**
	 * Restart this timer. Will not notify client.
	 */
	public void restart()
	{
		if (_status == LyTimer.STOPPED)
			start();
		_needRestart = true;
		_needStop = false;
	}

	/**
	 * Gets the client data.
	 * @return user client data freely set by the user
	 */
	public Object getClientData()
	{
		return _clientData;
	}

	/**
	 * Sets the client data.
	 * @param clientData user client data freely set by the user
	 */
	public void setClientData(Object clientData)
	{
		_clientData = clientData;
	}
} // end class LyTimer
