/*
 * LyTools.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.lang.reflect.*;
import java.text.*;
import java.util.*;
import java.util.regex.*;


/**
 * This class implements some basic conversion and utilities methods. Part of them are toString
 * methods that convert objects to their string representation and fromString methods that build
 * objects from a given string.
 * @version $Id: LyTools.java 36498 2012-08-16 08:52:28Z nhamon $
 * @author  Lyria
 */
public class LyTools
{
	public static final String __VERSION = "$Id: LyTools.java 36498 2012-08-16 08:52:28Z nhamon $";

	/** Separator used for applicative objects. */
	public static final char   OBJECT_SEP = (char)1;

	/** Separator used between values of applicative objects. */
	public static final char   VALUE_SEP = (char)2;

	/** Separator used for Vectors. */
	public static final char   VECTOR_SEP = (char)3;

	/** Separator used for Arrays. */
	public static final char   ARRAY_SEP = (char)4;

	/** Separator used for BitSets. */
	public static final char   BITSET_SEP = (char)5;

	/** Separator used between class name and id. */
	public static final char   CLASS_SEP = (char)6;

	/** Separator used for strings in web. */
	public static final char   PIPE_SEP = '|';

	/** Identifier of the last request. */
	protected static int	   _lastIdentifier = 0;

	/** Identifier which will automatically increment after each request. */
	protected static long	   _counterL;

	/** Internal randomizer. */
	private static Random	   _random;

	static
	{
		try
		{
			SimpleDateFormat fomat = new SimpleDateFormat("dd/MM/yyyy HH:mm:ss");
			_counterL = ((System.currentTimeMillis()
						- (fomat.parse("07/01/2004 00:00:00")).getTime()));
		}
		catch (ParseException e)
		{
		}

		_random = new Random(System.currentTimeMillis());
	}

	/**
	 * toString methods.
	 */

	/**
	 * Converts a character to a string.
	 * @param  c the character to convert
	 * @return the string that represents this character.
	 * @see    Character#toString(char)
	 */
	public static String toString(char c)
	{
		return Character.toString(c);
	}

	/**
	 * Convert a date to a string. Calls the method {@link LyDateFormatter#format(Date, String)
	 * LyDateFormatter.format(date, "dd/MM/yyyy HH:mm:ss")}
	 * @param  date The date that must be converted.
	 * @return The string that represents the date.
	 */
	public static String toString(Date date)
	{
		return LyDateFormatter.format(date, "dd/MM/yyyy HH:mm:ss");
	}

	/**
	 * Converts a BitSet to a string.
	 * @param  bs The BitSet that must be converted.
	 * @return The string that represents the BitSet.
	 */
	public static String toString(BitSet bs)
	{
		StringBuffer buffer = new StringBuffer();

		boolean		 first = true;

		int			 limit = bs.size();

		buffer.append(BITSET_SEP);

		for (int i = 0; i < limit; i++)
		{
			if (bs.get(i))
			{
				if (!first)
					buffer.append(BITSET_SEP);
				else
					first = false;

				buffer.append(i);
			}
		}

		buffer.append(BITSET_SEP);

		return buffer.toString();
	}

	/**
	 * fromString methods.
	 */

	/**
	 * Converts a string to an int.
	 * @param  value The string that must be parsed.
	 * @return The integer that represents the parsed string or 0 if not an int representation.
	 */
	public static int intFromString(String value)
	{
		try
		{
			if ((value == null) || (value.length() == 0))
				return 0;
			int i = Integer.parseInt(value);

			return i;
		}
		catch (NumberFormatException e)
		{
			// return 0 if not an integer
			return 0;
		}
	}

	/**
	 * Converts a string to an Integer.
	 * @param  value The string that must be parsed.
	 * @return The Integer that represents the parsed string or null if not an Integer
	 *         representation.
	 */
	public static Integer IntegerFromString(String value)
	{
		try
		{
			Integer i = Integer.valueOf(value);

			return i;
		}
		catch (NumberFormatException e)
		{
			return null;
		}
	}

	/**
	 * Converts a string to a float.
	 * @param  value The string that must be parsed.
	 * @return The float that represents the parsed string or 0F if not a float representation.
	 */
	public static float floatFromString(String value)
	{
		Float f = FloatFromString(value);

		if (f == null)
			// return 0 if not a float
			return 0F;
		else
			return f.floatValue();
	}

	/**
	 * Converts a string to a Float.
	 * @param  value The string that must be parsed.
	 * @return The Float that represents the parsed string or null if not a Float representation.
	 */
	public static Float FloatFromString(String value)
	{
		try
		{
			Float f = Float.valueOf(value);

			return f;
		}
		catch (NumberFormatException e)
		{
			return null;
		}
	}

	/**
	 * Converts a string to a double. Calls {@link #doubleFromString(String, double)
	 * doubleFromString(value, 0.0)}
	 * @param  value The string that must be parsed.
	 * @return The double that represents the parsed string or 0.0 if not a double representation.
	 * @see    #doubleFromString(String, double)
	 */
	public static double doubleFromString(String value)
	{
		return doubleFromString(value, 0.0);
	}

	/**
	 * Converts a string to a double.
	 * @param  value        The string that must be parsed.
	 * @param  defaultValue The default value if it is not a double value.
	 * @return The double that represents the parsed string or the default value if not a double
	 *         representation.
	 * @see    #DoubleFromString(String)
	 */
	public static double doubleFromString(String value, double defaultValue)
	{
		Double d = DoubleFromString(value);

		if (d == null)
			// return default value if not a double
			return defaultValue;
		else
			return d.doubleValue();
	}

	/**
	 * Converts a string to a Double.
	 * @param  value The string that must be parsed.
	 * @return The Double that represents the parsed string or null if not a Double representation.
	 */
	public static Double DoubleFromString(String value)
	{
		if (value == null)
			return null;

		// If value has an 'e' as the last char, ignore it
		if ((value.length() > 0)
				&& ((value.charAt(value.length() - 1) == 'e')
					|| (value.charAt(value.length() - 1) == 'E')))
			value = value.substring(0, value.length() - 1);

		try
		{
			Double d = Double.valueOf(value);

			return d;
		}
		catch (NumberFormatException e)
		{
			return null;
		}
	}

	/**
	 * Converts a string to a short.
	 * @param  value The string that must be parsed.
	 * @return The short that represents the parsed string or 0 if not a short representation.
	 */
	public static short shortFromString(String value)
	{
		try
		{
			short s = Short.parseShort(value);

			return s;
		}
		catch (NumberFormatException e)
		{
			// return 0 if not a short
			return 0;
		}
	}

	/**
	 * Converts a string to a Short.
	 * @param  value The string that must be parsed.
	 * @return The Short that represents the parsed string or null if not a Short representation.
	 */
	public static Short ShortFromString(String value)
	{
		try
		{
			Short s = Short.valueOf(value);

			return s;
		}
		catch (NumberFormatException e)
		{
			return null;
		}
	}

	/**
	 * Converts a string to a long.
	 * @param  value The string that must be parsed.
	 * @return The long that represents the parsed string or 0L if not a long representation.
	 */
	public static long longFromString(String value)
	{
		if (value == null)
			return 0L;

		try
		{
			double dbValue = Double.parseDouble(value);
			Double temp = new Double(dbValue);
			long   l = temp.longValue();

			return l;
		}
		catch (NumberFormatException e)
		{
			// return 0 if not a long
			return 0L;
		}
	}

	/**
	 * Converts a string to a Long.
	 * @param  value The string that must be parsed.
	 * @return The Long that represents the parsed string or null if not a Long representation.
	 */
	public static Long LongFromString(String value)
	{
		try
		{
			Long l = Long.valueOf(value);

			return l;
		}
		catch (NumberFormatException e)
		{
			return null;
		}
	}

	/**
	 * Converts a string to a byte.
	 * @param  value The string that must be parsed.
	 * @return The byte that represents the parsed string or 0 if not a byte representation.
	 */
	public static byte byteFromString(String value)
	{
		try
		{
			byte b = Byte.parseByte(value);

			return b;
		}
		catch (NumberFormatException e)
		{
			// return 0 if not a byte
			return 0;
		}
	}

	/**
	 * Converts a string to a Byte.
	 * @param  value The string that must be parsed.
	 * @return The Byte that represents the parsed string or null if not a Byte representation.
	 */
	public static Byte ByteFromString(String value)
	{
		try
		{
			Byte b = Byte.valueOf(value);

			return b;
		}
		catch (NumberFormatException e)
		{
			return null;
		}
	}

	/**
	 * Converts a string to a boolean.
	 * @param  value The string that must be parsed.
	 * @return true if the string value equals "true",
	 * @see    #BooleanFromString(String)
	 */
	public static boolean booleanFromString(String value)
	{
		Boolean b = BooleanFromString(value);

		return (b.booleanValue());
	}

	/**
	 * Converts a string to a Boolean.
	 * @param  value The string that must be parsed.
	 * @return The true Boolean object if the string value equals "true",
	 */
	public static Boolean BooleanFromString(String value)
	{
		Boolean b = Boolean.valueOf(value);

		return b;
	}

	/**
	 * Converts a string to a char.
	 * @param  value The string that must be parsed.
	 * @return The char that represents the parsed string or OBJECT_SEP if not a char
	 *         representation.
	 */
	public static char charFromString(String value)
	{
		if ((value == null) || (value.length() != 1))
			// Return a default character
			return OBJECT_SEP;
		else
			return (value.charAt(0));
	}

	/**
	 * Converts a string to a Character.
	 * @param  value The string that must be parsed.
	 * @return The Character that represents the parsed string or null if not a Character
	 *         representation.
	 */
	public static Character CharacterFromString(String value)
	{
		if ((value == null) || (value.length() != 1))
			return null;
		else
		{
			char c = value.charAt(0);

			return (new Character(c));
		}
	}

	/**
	 * Converts a string to a Date.
	 * @param  value The string that must be parsed. It must respect the date format 'dd/MM/yyyy
	 *               HH:mm:ss'.
	 * @return The Date that represents the parsed string or null if not a Date representation.
	 * @see    LyDateFormatter#parse(String, String)
	 */
	public static Date DateFromString(String value)
	{
		Date date = null;

		try
		{
			date = LyDateFormatter.parse(value, "dd/MM/yyyy HH:mm:ss");
		}
		catch (NumberFormatException e)
		{
		}

		return date;
	}

	/**
	 * Converts a string to an Array of objects. The string must contain the objects representation
	 * separated by ARRAY_SEP.
	 * @param  value    The string that must be converted.
	 * @param  classRef The class of the objects to create.
	 * @return The array of objects.
	 * @see    #VectorFromString(String, Class, char)
	 */
	public static Object[] ArrayFromString(String value, Class<?> classRef)
	{
		Vector<Object>   vector = VectorFromString(value, classRef, ARRAY_SEP);
		Object[]         result = new Object[vector.size()];

		vector.copyInto(result);
		return result;
	}

	/**
	 * Converts a string to a vector. The string must contain the objects representation separated
	 * by VECTOR_SEP. Only calls {@link #VectorFromString(String, Class, char)
	 * #VectorFromString(value, classRef, VECTOR_SEP)}
	 * @param  value    The string that must be converted.
	 * @param  classRef The class of the objects to create.
	 * @return The vector of objects.
	 */
	public static Vector<Object> VectorFromString(String value, Class<?> classRef)
	{
		return VectorFromString(value, classRef, VECTOR_SEP);
	}

	/**
	 * Converts a string to a vector. The string must contain the objects representation separated
	 * by SEP parameter.
	 * @param  value    The string that must be converted.
	 * @param  classRef The class of the objects to create.
	 * @param  SEP      the separator between values.
	 * @return The vector of objects.
	 */
	public static Vector<Object> VectorFromString(String value, Class<?> classRef, char SEP)
	{
		// Ignore first and last character if equal VECTOR_SEP
		int size = value.length();
		int first = 0, last = size;

		if (value.startsWith(toString(SEP)))
			first++;
		if (value.endsWith(toString(SEP)))
			last--;

		value = value.substring(first, last);

		Method method = null;

		if (classRef != java.lang.String.class)
		{
			// Retrieve the method to call to convert each object.
			// If there is no method corresponding to classRef (with a String
			// argument), we call the ObjectFromString method. It means the
			// current method is not implemented for vectors containing vectors,
			// arrays or any other Java class which is not Integer, Float,
			// Double, Long, Short, BitSet, Date, Boolean, String, Byte,
			// Character or applicative class.
			try
			{
				String   methodName = classRef.getName() + "FromString";
				String[] fullMethodName = LyString.stringToList(methodName, '.');

				methodName = fullMethodName[fullMethodName.length - 1];
				method = LyTools.class.getMethod(methodName, new Class[] { String.class });
			}
			catch (Exception e)
			{
				e.printStackTrace();
			}
		}

		String[] strings = LyString.stringToList(value, SEP);
		int		 i, n = strings.length;
		Vector<Object>   result = new Vector<Object>(n);
		Object   object = null;

		for (i = 0; i < n; i++)
		{
			if (method == null)
				result.addElement(strings[i]);
			else
			{
				try
				{
					object = method.invoke(null, new Object[] { strings[i] });
				}
				catch (Exception e)
				{
					e.printStackTrace();
				}

				if (object != null)
					result.addElement(object);
			}
		}

		return result;
	} // end method VectorFromString

	/**
	 * Converts a string to a BitSet. The string must contain the bits separated by BITSET_SEP.
	 * @param  value The string that must be parsed.
	 * @return The BitSet that represents the parsed string or null if not a BitSet representation.
	 */
	public static BitSet BitSetFromString(String value)
	{
		// Ignore first and last character if equal BITSET_SEP
		int size = value.length();
		int first = 0, last = size;

		if (value.startsWith(toString(BITSET_SEP)))
			first++;
		if (value.endsWith(toString(BITSET_SEP)))
			last--;

		value = value.substring(first, last);

		String[] bits = LyString.stringToList(value, BITSET_SEP);
		BitSet   bs = new BitSet();
		int		 i, n = bits.length;
		Integer  bit;

		for (i = 0; i < n; i++)
		{
			// Retrieve the value at position i
			bit = IntegerFromString(bits[i]);

			if (bit != null)
				bs.set(bit.intValue());
		}

		return bs;
	} // end method BitSetFromString

	/**
	 * Utilities Methods.
	 */

	/**
	 * Gets next identifier and increments internal counter.
	 * @return The next identifier.
	 */
	public static synchronized int getNextIdentifier()
	{
		return (LyTools._lastIdentifier++);
	}

	/**
	 * Gets next counter value and increments internal counter.
	 * @return The next counter value.
	 */
	public static synchronized long nextLong()
	{
		return (LyTools._counterL++);
	}

	/**
	 * Gets the next random value as an int.
	 * @return the next random value
	 * @see    Random#nextInt()
	 */
	public static synchronized int nextInt()
	{
		return (LyTools._random.nextInt());
	}

	/**
	 * Gets the next random value as a short. Calls {@link Random#nextInt(int)
	 * _random.nextInt(Short.MAX_VALUE)}
	 * @return the next random value
	 * @see    Random#nextInt(int)
	 */
	public static synchronized short nextShort()
	{
		return (short)(LyTools._random.nextInt(Short.MAX_VALUE));
	}

	/**
	 * Gets the next random value as a float.
	 * @return the next random value
	 * @see    Random#nextFloat()
	 */
	public static synchronized float nextFloat()
	{
		return (LyTools._random.nextFloat());
	}

	/**
	 * Gets the next random value as a double.
	 * @return the next random value
	 * @see    Random#nextDouble()
	 */
	public static synchronized double nextDouble()
	{
		return (LyTools._random.nextDouble());
	}

	/**
	 * Checks the number of decimals of the given double.
	 * @param  value    the double value to check
	 * @param  decimals The number of decimals allowed. A negative value means no limit on number of
	 *                  decimals.
	 * @return A double value corresponding to the given value with the right number of decimals.
	 *         For example the result of checkDecimals (3.454, 1) will be 3.4.
	 */
	public static double checkDecimals(double value, int decimals)
	{
		if (decimals < 0)
			return value;
		if (Double.isNaN(value))
			return value;

		int i, precision = 1;

		for (i = 0; i < decimals; i++)
			precision = precision * 10;

		return (double)Math.round(value * precision) / precision;
	}

	/**
	 * Checks the number of decimals of the given string representation of a number.
	 * @param  value    the string representation of the number to check
	 * @param  decimals The number of decimals. A negative value means no limit on number of
	 *                  decimals.
	 * @return A substring of the given string with the right number of decimals. For example the
	 *         result of checkDecimals (3.454, 1) will be 3.4.
	 */
	public static String checkDecimals(String value, int decimals)
	{
		if ((value == null) || (value.length() == 0))
			return null;

		if (decimals < 0)
			return value;

		value = removeExponent(value);

		int decimalPos = value.indexOf('.');

		if (decimalPos < 0)
			return value;

		if (decimals == 0)
			value = value.substring(0, decimalPos);
		else if ((value.length() - decimalPos - 1) > decimals)
			value = value.substring(0, decimalPos + decimals + 1);

		return value;
	}
	
	/**
	 * Remove all occurrence of a specific character in a string
	 * @param str the source string
	 * @param character the character to remove
	 * @return a string without the given character
	 */
	public static String removeChar(String str, char character)
	{
		if (str == null)
			return str;
		
		boolean hasGroup = str.indexOf(character) >= 0;
		if (!hasGroup)
			return str;
		
		// remove the group separator
		char[] charArray = str.toCharArray();
		StringBuffer sb = new StringBuffer();
		int i, length = charArray.length;
		
		for (i = 0; i < length; i++)
		{
			if (charArray[i] == character)
				continue;
			sb.append(charArray[i]);
		}
		return sb.toString();
	}

	/**
	 * Removes exponent from the given string representation of a double.
	 * @param  value Initial value result of Double.toString
	 * @return Result without exponent.
	 */
	public static String removeExponent(String value)
	{
		if (value == null)
			return null;

		value = value.replace('e', 'E');

		int expPos = value.indexOf('E');

		if ((expPos < 0) || (expPos == (value.length() - 1)))
			return value.replace(',', '.');

		String expString = value.substring(expPos + 1);
		int    exp = 1;
		
		try
		{
			exp = Integer.parseInt(expString);
		}
		catch (NumberFormatException e)
		{
			return value.replace(',', '.');
		}
		
		String newValue = value.substring(0, expPos);
		int    decimalPos = newValue.replace(',', '.').indexOf('.');
		int    i, length = newValue.length();

		if (decimalPos >= 0)
		{
			newValue = newValue.substring(0, decimalPos) + newValue.substring(decimalPos + 1);
			length--;
		}
		else
			decimalPos = length;

		if (exp > 0)
		{
			if ((exp + decimalPos) > length)
			{
				for (i = exp - length + decimalPos; i > 0; i--)
					newValue += '0';
			}
			else if ((exp + decimalPos) < length)
				newValue = newValue.substring(0, exp + decimalPos) + '.'
					+ newValue.substring(exp + decimalPos);
		}
		else
		{
			boolean negative = (newValue.charAt(0) == '-');

			if (negative)
			{
				newValue = newValue.substring(1);
				decimalPos--;
			}

			if (-exp >= decimalPos)
			{
				for (i = -exp - decimalPos; i > 0; i--)
					newValue = '0' + newValue;
				newValue = "0." + newValue;
			}
			else
				newValue = newValue.substring(0, exp + decimalPos) + '.'
					+ newValue.substring(exp + decimalPos);

			if (negative)
				newValue = '-' + newValue;
		}

		return newValue;
	} // end method removeExponent

	/**
	 * Calculate a pitch according to a given number of decimals.
	 * @param  pitch    The desired pitch.
	 * @param  decimals The maximum number of decimals. For example the result of getPitch(0.04, 1)
	 *                  will be 0.1.
	 * @return the computed pitch
	 */
	public static double getPitch(double pitch, int decimals)
	{
		double newPitch = pitch;

		// If integer, set the pitch to one if not at least an integer
		if (((decimals == 0) && (Math.abs(pitch) < 1)) || (pitch == 0))
			newPitch = 1;
		// If more than one decimal, check that the pitch is big enough
		else if ((decimals > 0) && (Math.abs(pitch) < (1 / (Math.pow(10, decimals)))))
			newPitch = (1 / (Math.pow(10, decimals)));

		// Check the number of decimals
		newPitch = checkDecimals(newPitch, decimals);

		return newPitch;
	}

	/**
	 * Gets a separator depending on object type. Here are the correspondencies:
	 * <ul>
	 *   <li>Array -> ARRAY_SEP</li>
	 *   <li>BitSet -> BITSET_SEP</li>
	 *   <li>Vector -> VECTOR_SEP</li>
	 *   <li>else -> OBJECT_SEP</li>
	 * </ul>
	 * @param  object The object whose type we want to retrieve separator.
	 * @return The character that is used as separator for the given object type.
	 */
	public static char getSeparator(Object object)
	{
		if (object.getClass().isArray())
			return ARRAY_SEP;
		else if (object instanceof BitSet)
			return BITSET_SEP;
		else if (object instanceof Vector<?>)
			return VECTOR_SEP;
		else
			return OBJECT_SEP;
	}

	/**
	 * Used in swing component to replace multilined String by its equivalent in HTML.
	 * @param  tip            the original String to replace.
	 * @param  maxCharPerLine maximum number of character in a line
	 * @return the equivalent String in HTML with "&lt;p&gt;" instead of "\n".
	 */
	public static String formatToHTML(String tip, int maxCharPerLine)
	{
		if (tip == null)
			return null;

		tip = breakString(tip, maxCharPerLine);

		String newTip = new String("");
		int    begin = 0, nEnd = tip.indexOf("\n"), lsEnd = tip.indexOf(Character.LINE_SEPARATOR);
		int	   end = ((nEnd >= 0) && (lsEnd > 0) ? Math.min(nEnd, lsEnd) : Math.max(nEnd, lsEnd));

		newTip = "<html><body marginwidth=4>";
		String s;

		while (end != -1)
		{
			s = tip.substring(begin, end);
			s = LyString.getHtmlEntitiesString(s);

			newTip += s + "<p>";
			begin = end + 1;
			nEnd = tip.indexOf("\n", begin);
			lsEnd = tip.indexOf(Character.LINE_SEPARATOR, begin);
			end = ((nEnd >= 0) && (lsEnd > 0) ? Math.min(nEnd, lsEnd) : Math.max(nEnd, lsEnd));
		}
		
		newTip += tip.substring(begin, tip.length()) + "</body></html>";
		return newTip;
	}

	/**
	 * Break the line in label that exceed maxChars number.
	 * @param  label    the label that must not exceed maxChar length
	 * @param  maxChars the maximum number of characters of the label
	 * @return a breaked version of the label
	 */
	public static String breakString(String label, int maxChars)
	{
		if (maxChars <= 0)
			return label;

		Vector<String> lines;

		if (label == null)
			label = "";

		StringTokenizer t = new StringTokenizer(label, "\n");

		lines = new Vector<String>();
		while (t.hasMoreElements())
		{
			String line = t.nextToken();

			if ((maxChars == -1) || (line.length() <= maxChars))
				lines.addElement(line);
			else
			{
				while (line.length() > maxChars)
				{
					int offset = line.lastIndexOf(' ', maxChars);

					if (offset == -1)
					{
						// didn't get one within maxChars!
						offset = line.indexOf(' ');
						if (offset == -1)
							break;
					}
					lines.addElement(line.substring(0, offset));
					line = line.substring(offset + 1);
				}
				lines.addElement(line);
			}
		}

		String breakedLabel = "";

		for (int i = 0; i < lines.size(); i++)
			breakedLabel += lines.get(i) + "\n";

		if (breakedLabel.length() > 0)
			breakedLabel = breakedLabel.substring(0, breakedLabel.length() - 1);

		return breakedLabel;
	} // end method breakString

	/**
	 * Returns a string representation of the contents of the specified array. If the array contains
	 * other arrays as elements, they are converted to
	 * @param  a the array whose string representation to return
	 * @return a string representation of <tt>a</tt>
	 */
	public static String toString(Object[] a)
	{
		// Should use Arrays.toString() in JDK 1.5 or >
		if (a == null)
			return "null";
		if (a.length == 0)
			return "[]";

		StringBuffer buf = new StringBuffer();

		for (int i = 0; i < a.length; i++)
		{
			if (i == 0)
				buf.append('[');
			else
				buf.append(", ");

			buf.append(String.valueOf(a[i]));
		}

		buf.append("]");
		return buf.toString();
	}
    
    /**
     * Removes accents from given string.
     * @param str Initial string.
     * @return String without accents.
     */
    public static String removeAccents(String s)
    {
        if (s == null)
            return null;
          
        // Java 6
        // return Normalizer.normalize(str, Normalizer.Form.NFD);
        
        s = s.replaceAll("[]","e"); 
        s = s.replaceAll("[]","u"); 
        s = s.replaceAll("[]","i"); 
        s = s.replaceAll("[]","a"); 
        s = s.replaceAll("","o"); 
        s = s.replaceAll("","c"); 
        s = s.replaceAll("[]","E"); 
        s = s.replaceAll("[]","U"); 
        s = s.replaceAll("[]","I"); 
        s = s.replaceAll("[]","A"); 
        s = s.replaceAll("","O"); 
        s = s.replaceAll("","C"); 
        return s;
    }
    
    public static String getIndentString()
	{
		LyEnvironment env = LyEnvironment.getCurrentEnvironment();
		String result = "\t";
		String spacesForTabs = env.getEnv("_LY_SPACES_FOR_TABS");
		
		if(Boolean.valueOf(spacesForTabs))
		{
			result = "  ";
			String strNbSpacesForTabs = env.getEnv("_LY_NB_SPACES_FOR_TABS");
			
			try
			{
				StringBuilder str = new StringBuilder();
				int nbSpacesForTabs = Integer.parseInt(strNbSpacesForTabs);
				for(int i=0; i< nbSpacesForTabs; i++)
					str.append(" ");
				
				if(str.length() > 1)
					result = str.toString();
			}
			catch (NumberFormatException nfe)
			{
				env.logWarning("_LY_NB_SPACES_FOR_TABS is not set or has not the right format." +
						" Two spaces are set for one tab");
			}
		}
		
		return result;
	}
	
	/**
	 * Encode Xml tags from a string to display it as Html.
	 * 
	 * @param value the string that must be parsed
	 * @return the encoded string
	 */
	public static String encodeXmlTags(String value)
	{
		if (value == null)
			return null;
		
		int				i, j, n = value.length();
		StringBuffer	buffer = new StringBuffer(n);
		char			c, c2;
		boolean			replace = false;
		
		for (i = 0; i < n; i++)
		{
			c = value.charAt(i);
			if ((c == '\n')
				&& !value.contains("<br/>")
				&& !value.contains("<br />"))
			{
				buffer.append("<br/>");
				continue;
			}
			
			if (c == '<')
			{
				replace = true;
				for (j = i + 1; (j < n) && replace; j++)
				{
					c2 = value.charAt(j);
					if ((c2 == ' ')
							|| (c2 == '>'))
						break;

					if ((c2 != '/')
							&& (c2 != '_')
							&& (c2 != '-')
							&& !Character.isUpperCase(c2))
						replace = false;
				}
					
				if (replace)
					buffer.append("&lt;");
				else
					buffer.append(c);
			}
			else if (c == '>')
			{
				if (replace)
					buffer.append("&gt;");
				else
					buffer.append(c);
				replace = false;
			}
			else
				buffer.append(c);
		}
		
	    return buffer.toString();
	}
	
	/**
	 * Remove tags from a string.
	 * 
	 * @param value the string that must be parsed
	 * @return the string without tags parts
	 */
	public static String removeHtmlTags(String value)
	{
		if (value == null)
			return null;
		
	    value = value.replaceAll("\\<(br|BR)( |\t)*>", String.valueOf((char)10));
	    value = value.replaceAll("\\<(br|BR)( |\t)*/>", String.valueOf((char)10));
	    value = value.replaceAll("\\<[a-z!].*?>", "");
	    value = value.replaceAll("\\</[a-z].*?>", "");
        value = value.replaceAll("(\n)", String.valueOf((char)10));
        value = value.replaceAll("(\t)", " ");
	    value = value.replaceAll(String.valueOf((char)10) + String.valueOf((char)10), String.valueOf((char)10)); // Avoid multiple CR
	    
		Hashtable<String, String> tagsTable = new Hashtable<String, String>();
		
		tagsTable.put("&nbsp;", String.valueOf((char)160));
		tagsTable.put("&iexcl;", String.valueOf((char)161));
		tagsTable.put("&cent;", String.valueOf((char)162));
		tagsTable.put("&pound;", String.valueOf((char)163));
		tagsTable.put("&curren;", String.valueOf((char)164));
		tagsTable.put("&yen;", String.valueOf((char)165));
		tagsTable.put("&brvbar;", String.valueOf((char)166));
		tagsTable.put("&sect;", String.valueOf((char)167));
		tagsTable.put("&uml;", String.valueOf((char)168));
		tagsTable.put("&copy;", String.valueOf((char)169));
		tagsTable.put("&ordf;", String.valueOf((char)170));
		tagsTable.put("&laquo;", String.valueOf((char)171));
		tagsTable.put("&not;", String.valueOf((char)172));
		tagsTable.put("&shy;", String.valueOf((char)173));
		tagsTable.put("&reg;", String.valueOf((char)174));
		tagsTable.put("&macr;", String.valueOf((char)175));
		tagsTable.put("&deg;", String.valueOf((char)176));
		tagsTable.put("&plusmn;", String.valueOf((char)177));
		tagsTable.put("&sup2;", String.valueOf((char)178));
		tagsTable.put("&sup3;", String.valueOf((char)179));
		tagsTable.put("&acute;", String.valueOf((char)180));
		tagsTable.put("&micro;", String.valueOf((char)181));
		tagsTable.put("&para;", String.valueOf((char)182));
		tagsTable.put("&middot;", String.valueOf((char)183));
		tagsTable.put("&cedil;", String.valueOf((char)184));
		tagsTable.put("&sup1;", String.valueOf((char)185));
		tagsTable.put("&ordm;", String.valueOf((char)186));
		tagsTable.put("&raquo;", String.valueOf((char)187));
		tagsTable.put("&frac14;", String.valueOf((char)188));
		tagsTable.put("&frac12;", String.valueOf((char)189));
		tagsTable.put("&frac34;", String.valueOf((char)190));
		tagsTable.put("&iquest;", String.valueOf((char)191));
		tagsTable.put("&Agrave;", String.valueOf((char)192));
		tagsTable.put("&Aacute;", String.valueOf((char)193));
		tagsTable.put("&Acirc;", String.valueOf((char)194));
		tagsTable.put("&Atilde;", String.valueOf((char)195));
		tagsTable.put("&Auml;", String.valueOf((char)196));
		tagsTable.put("&Aring;", String.valueOf((char)197));
		tagsTable.put("&AElig;", String.valueOf((char)198));
		tagsTable.put("&Ccedil;", String.valueOf((char)199));
		tagsTable.put("&Egrave;", String.valueOf((char)200));
		tagsTable.put("&Eacute;", String.valueOf((char)201));
		tagsTable.put("&Ecirc;", String.valueOf((char)202));
		tagsTable.put("&Euml;", String.valueOf((char)203));
		tagsTable.put("&Igrave;", String.valueOf((char)204));
		tagsTable.put("&Iacute;", String.valueOf((char)205));
		tagsTable.put("&Icirc;", String.valueOf((char)206));
		tagsTable.put("&Iuml;", String.valueOf((char)207));
		tagsTable.put("&ETH;", String.valueOf((char)208));
		tagsTable.put("&Ntilde;", String.valueOf((char)209));
		tagsTable.put("&Ograve;", String.valueOf((char)210));
		tagsTable.put("&Oacute;", String.valueOf((char)211));
		tagsTable.put("&Ocirc;", String.valueOf((char)212));
		tagsTable.put("&Otilde;", String.valueOf((char)213));
		tagsTable.put("&Ouml;", String.valueOf((char)214));
		tagsTable.put("&times;", String.valueOf((char)215));
		tagsTable.put("&Oslash;", String.valueOf((char)216));
		tagsTable.put("&Ugrave;", String.valueOf((char)217));
		tagsTable.put("&Uacute;", String.valueOf((char)218));
		tagsTable.put("&Ucirc;", String.valueOf((char)219));
		tagsTable.put("&Uuml;", String.valueOf((char)220));
		tagsTable.put("&Yacute;", String.valueOf((char)221));
		tagsTable.put("&THORN;", String.valueOf((char)222));
		tagsTable.put("&szlig;", String.valueOf((char)223));
		tagsTable.put("&agrave;", String.valueOf((char)224));
		tagsTable.put("&aacute;", String.valueOf((char)225));
		tagsTable.put("&acirc;", String.valueOf((char)226));
		tagsTable.put("&atilde;", String.valueOf((char)227));
		tagsTable.put("&auml;", String.valueOf((char)228));
		tagsTable.put("&aring;", String.valueOf((char)229));
		tagsTable.put("&aelig;", String.valueOf((char)230));
		tagsTable.put("&ccedil;", String.valueOf((char)231));
		tagsTable.put("&egrave;", String.valueOf((char)232));
		tagsTable.put("&eacute;", String.valueOf((char)233));
		tagsTable.put("&ecirc;", String.valueOf((char)234));
		tagsTable.put("&euml;", String.valueOf((char)235));
		tagsTable.put("&igrave;", String.valueOf((char)236));
		tagsTable.put("&iacute;", String.valueOf((char)237));
		tagsTable.put("&icirc;", String.valueOf((char)238));
		tagsTable.put("&iuml;", String.valueOf((char)239));
		tagsTable.put("&eth;", String.valueOf((char)240));
		tagsTable.put("&ntilde;", String.valueOf((char)241));
		tagsTable.put("&ograve;", String.valueOf((char)242));
		tagsTable.put("&oacute;", String.valueOf((char)243));
		tagsTable.put("&ocirc;", String.valueOf((char)244));
		tagsTable.put("&otilde;", String.valueOf((char)245));
		tagsTable.put("&ouml;", String.valueOf((char)246));
		tagsTable.put("&divide;", String.valueOf((char)247));
		tagsTable.put("&oslash;", String.valueOf((char)248));
		tagsTable.put("&ugrave;", String.valueOf((char)249));
		tagsTable.put("&uacute;", String.valueOf((char)250));
		tagsTable.put("&ucirc;", String.valueOf((char)251));
		tagsTable.put("&uuml;", String.valueOf((char)252));
		tagsTable.put("&yacute;", String.valueOf((char)253));
		tagsTable.put("&thorn;", String.valueOf((char)254));
		tagsTable.put("&yuml;", String.valueOf((char)255));
		tagsTable.put("&gt;", ">");
		tagsTable.put("&lt;", "<");
		tagsTable.put("&amp;", "&");
		tagsTable.put("&quot;", "'");
		tagsTable.put("&rsquo;", "'");
		tagsTable.put("&#39;", "'");
		
		StringBuilder strResult = new StringBuilder(value);
        Pattern pattern = Pattern.compile("&[#a-zA-z0-9]{2,6};");
		Matcher matcher = pattern.matcher(strResult);
		
		int start = 0;
		int end = 0;
		
		while(matcher.find(end))
		{
			MatchResult result = matcher.toMatchResult();
			
			start = result.start();
			end = result.end();
			
			String toReplace = strResult.substring(start, end);
			String replacement = tagsTable.get(toReplace);
			
			if(replacement != null)
			{
				strResult.replace(start, end, replacement);
				end = start + 1;
			}
			
			matcher = pattern.matcher(strResult);
		}
		
		value = strResult.toString();
		return value;
	}
	
	/**
	 * Check if the given string can be parse as float or double
	 * 
	 * @param str the string to check
	 * @return	true if the given string is a float or double, false otherwise
	 */
	public static boolean isDouble(String str)
	{
		final String Digits     = "(\\p{Digit}+)";
		final String HexDigits  = "(\\p{XDigit}+)";
		final String Exp        = "[eE][+-]?"+Digits;
		final String fpRegex    = ("[\\x00-\\x20]*[+-]?(NaN|Infinity|((("+Digits+"(\\.)?("+Digits+"?)("+Exp+")?)|(\\.("+Digits+")("+Exp+")?)|((" +
    		"(0[xX]" + HexDigits + "(\\.)?)|(0[xX]" + HexDigits + "?(\\.)" + HexDigits + "))[pP][+-]?" + Digits + "))" +
    		"[fFdD]?))[\\x00-\\x20]*");
		
		return Pattern.matches(fpRegex, str);
	}
} // end class LyTools
