/*
 * LyViewManagerInterface.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.io.*;
import java.util.*;


/**
 * Interface for view managers.
 */
public interface LyViewManagerInterface
{
	String __VERSION = "$Id: LyViewManagerInterface.java 41344 2012-10-25 15:45:02Z nhamon $";

	/**
	 * Frees the view manager. Clears all.
	 */
	void free();

	/**
	 * Prepares GUI to start.
	 */
	void prepareGui();

	/**
	 * Starts GUI.
	 * @param client the GUI client
	 */
	void startGui(LyGuiClientInterface client);

	/**
	 * Stops GUI.
	 * @param client the GUI client
	 */
	void stopGui(LyGuiClientInterface client);

	/**
	 * Causes the run() method of the given runnable to be invoked by the user-interface thread
	 * either synchronously or asynchronously. If asynchronous, the caller of this method continues
	 * to run in parallel, and is not notified when the runnable has completed. Otherwise, the
	 * thread which calls this method is suspended until the runnable completes.
	 * @param runnable    Code to run on the user-interface thread.
	 * @param synchronous True is synchronous, false if asynchronous.
	 */
	void execGui(Runnable runnable, boolean synchronous);

	/**
	 * Causes the run() method of the given runnable to be invoked by the user-interface thread
	 * asynchronously. The caller of this method continues to run in parallel, and is not notified
	 * when the runnable has completed.
	 * @param runnable Code to run on the user-interface thread.
	 */
	void execGui(Runnable runnable);

	/**
	 * Checks if the view manager has the given property.
	 * @param  property the property to find
	 * @return true is the view manager has the property, false otherwise
	 */
	boolean hasProperty(String property);

	/**
	 * Gets the current environment.
	 * @return the environment
	 */
	LyEnvironment getEnvironment();

	/**
	 * Clears all temporary resources used by the view manager.<br />
	 * This method must be implemented by sub classes.
	 */
	void clearTemporaryResources();

	/**
	 * Gets the temporary directory used to store temporary resources.
	 * @return the directory
	 */
	String getTemporaryDir();

	/**
	 * Gets the temporary location used to store temporary resources dependently of the viewer type
	 * (e.g. URL or relative URL for web).
	 * @return the temporary directory
	 */
	String getTemporaryLocation();

	/**
	 * To exit an application.
	 * @param code exit code of the application.
	 */
	void exit(int code);

	/**
	 * Checks if events on data need to be reposted in GUI event queue.
	 * @return True if the event need to be reposted, false otherwise.
	 */
	boolean repostGuiEvents();

	/**
	 * Indicates if the view manager is multi session or not.
	 * @return true if the view manager is multi session, false otherwise
	 */
	boolean isMultiSession();

	/**
	 * Retrieves a Color object from a LyColor.
	 * @param  color The leonardi color.
	 * @return A reference to the requested color if it is found, null otherwise
	 * @see    leon.misc.LyColor
	 */
	Object getColor(LyColor color);

	/**
	 * Retrieves a Font object from a LyFont.
	 * @param  font The leonardi font.
	 * @return A reference to the requested font if it is found, null otherwise
	 * @see    leon.misc.LyFont
	 */
	Object getFont(LyFont font);

	/**
	 * Retrieves the width of a string for a given LyFont.
	 * @param  font The leonardi font.
	 * @param  text The string value.
	 * @return The width of given string for given font.
	 */
	int getStringWidth(LyFont font, String text);

	/**
	 * Builds a default image.
	 * @return the built default image
	 */
	Object getDefaultImage();

	/**
	 * Gets images sizes. The result is given in an array where position 0 is the width and at
	 * position 1 the height.
	 * @param  image the image
	 * @return Image size (table with 2 entries).
	 */
	int[] getSizes(Object image);

	/**
	 * Creates a scaled version of the given image.
	 * @param  image  the image to scale.
	 * @param  width  the width to scale the image.
	 * @param  height the height to scale the image.
	 * @return A scaled version of the image.
	 */
	Object getScaledInstance(Object image, int width, int height);

	/**
	 * Retrieves the screen size.
	 * @param  component the component used to retrieve the screen size
	 * @return the dimension of the screen used to display views
	 */
	int[] getScreenSize(Object component);

	/**
	 * Retrieves the viewport size available on the computer.
	 * @return The addition of all screen sizes.
	 */
	int[] getMaxScreenSize();

	/**
	 * Emits a brief sound signal.
	 */
	void beep();

	/**
	 * Shows an error message to the end-user.
	 * @param message error message
	 */
	void raiseError(String message);

	/**
	 * Waits for the image loading.
	 * @param  component The image loading client
	 * @param  image     the image to load
	 * @throws InterruptedException if the wait is interruptued
	 */
	void waitForLoading(Object component, Object image)
		throws InterruptedException;

	/**
	 * Retrieves an image from a Stream.
	 * @param  component The image creation client.
	 * @param  resource  stream inputStream to the image data
	 * @return the image or null if an error occurs
	 * @throws IOException if something is wrong during the process
	 */
	Object loadImageFromStream(Object component, InputStream resource)
		throws IOException;

	/**
	 * Free resources of the image if not in the cache else does nothing.
	 * @param image The image that must be freed.
	 */
	void freeImage(Object image);
	
	/**
	 * Indicates whether given image may be cached.
	 * @param path  The path of image that may be cached.
	 * @param image The image that may be cached.
	 * @return true if image may be cached, false otherwise
	 */
	boolean useCacheForImage(String path, Object image);

	/**
	 * Gets the plugin id.
	 * @return The plugin id.
	 */
	String getPluginId();

	/**
	 * Sets the plugin id.
	 * @param pluginId the new plugin id
	 */
	void setPluginId(String pluginId);
	
	/**
     * Gets web root URL.
     */
    String getRootURL();
    
    /**
     * Accessor to the properties file.
     * @param prefix    Prefix used for properties files name.
     * @return the properties.
     */
    Properties getPropertiesFromContext(String prefix);
    
    /**
     * Indicates whether RTL (right to left) mode is activated
     * @return true is RTL mode is activated, false otherwise.
     */
    boolean isRtlMode();
}