/*
 * LyXmlPrintWriter.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.misc;

import java.io.*;


/**
 * LyXmlPrintWriter contains utility methods to write XML files.
 */
public class LyXmlPrintWriter extends LyPrintWriter
{
	public static final String __VERSION =
		"$Id: LyXmlPrintWriter.java 40733 2012-09-11 07:29:25Z jlsaouli $";

	/** Indicates if the current tag has to be closed or not. */
	protected boolean		   _closeTag;

	/**
	 * Constructor. Calls the super implementation with a buffered writer builds from an output
	 * stream writer itself built with the given parameters. Sets {@link #_closeTag} and {@link
	 * #_changed} to false, then {@link #start() starts} the print.
	 * @param  output  the output stream where the XML has to be wrote
	 * @param  charset the name of a supported charset
	 * @throws UnsupportedEncodingException if the named encoding is not supported
	 */
	public LyXmlPrintWriter(FileOutputStream output, String charset)
		throws UnsupportedEncodingException
	{
		super(output, charset);
	}

	/**
	 * Constructor. Calls the super implementation, sets {@link #_closeTag} and {@link #_changed} to
	 * false, then {@link #start() starts} the print.
	 * @param output the file where the XML will be wrote
	 */
	public LyXmlPrintWriter(FileOutputStream output)
	{
		super(output);
	}

	/**
	 * Writes the first line of the XML file where xml version and encoding type are defined.
	 */
	protected void start()
	{
		_closeTag = false;
		_indentString = LyTools.getIndentString();
		
		println("<?xml version=\"1.0\" encoding=\"" + LyString.DEFAULT_CHARSET + "\"?>");
	}

	/**
	 * Close the writer. Only calls the super implementation.
	 */
	public void close()
	{
		super.close();
	}
	
	/**
	 * Checks if the given filename with the xml extension at the given destination exits.
	 * @param  filename    the name of the file where the XML will be wrote
	 * @param  destination the path of the file
	 * @return True if it exists, false otherwise.
	 */
	public static final boolean checkFile(String filename, String destination)
	{
		return checkFile(filename, destination, ".xml");
	}
	
	/**
	 * Gets the print writer. This method only calls {@link #getPrintWriter(String, String, String)
	 * getPrintWriter(filename, destination, ".xml")}
	 * @param  filename    the name of the file where the XML will be wrote
	 * @param  destination the path of the file
	 * @return the XML print writer
	 * @see    #getPrintWriter(String, String, String)
	 */
	public static final LyXmlPrintWriter getPrintWriter(String filename, String destination)
	{
		return getPrintWriter(filename, destination, ".xml");
	}

	/**
	 * Gets the print writer associated to the given filename with the given extension at the given
	 * destination. If the file ever exist, a backup is made into the filename + type + ".bak" file.
	 * @param  filename    the name of the file where the XML will be wrote
	 * @param  destination the path of the file
	 * @param  type        the file type (mostly ".xml")
	 * @return the XML print writer
	 */
	public static LyXmlPrintWriter getPrintWriter(String filename, String destination,
		String type)
	{
		FileOutputStream	output = getFileOutputStream(filename, destination, type);
		LyXmlPrintWriter 	pw;

		try
		{
			pw = new LyXmlPrintWriter(output, LyString.DEFAULT_CHARSET);
		}
		catch (UnsupportedEncodingException e)
		{
			pw = new LyXmlPrintWriter(output);
		}

		pw.start();
		return pw;
	} // end method getPrintWriter

	/**
	 * Utility methods to print, avoid ^M at the end of lines and to indent result.<br />
	 * If {@link #_closeTag} is set to <code>true</code> then print '&gt;' character and set the
	 * attribute to <code>false</code>.<br />
	 * If {@link #_indentNext} is set to true, print the number of tabulation of the {@link
	 * #_indent} parameter.<br />
	 * Calls the super implementation and set {@link #_changed} to <code>true</code>.
	 */
	public void print(String str)
	{
		if (_closeTag)
		{
			println('>');
			_closeTag = false;
		}

		super.print(str);
	}

	/**
	 * Sets the {@link #_closeTag} parameter with the given value and calls. {@link #addIndent()}
	 * @param closeTag the new close tag value
	 */
	public void addIndent(boolean closeTag)
	{
		_closeTag = closeTag;
		addIndent();
	}

	/**
	 * Removes an indentation and if the given value is true and if {@link #_closeTag} is true,
	 * prints '/&gt;'n sets _closeTag to false and return true, else return false.
	 * @param  closeTag indicates if the end of tag must be print
	 * @return true if '/&gt;' is printed, false otherwise.
	 */
	public boolean removeIndent(boolean closeTag)
	{
		_indent--;

		if (closeTag && _closeTag)
		{
			_closeTag = false;
			println("/>");
			return true;
		}
		else
			return false;
	}
} // end class LyXmlPrintWriter
