/*
 * LyEventDecoder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.notifier;

import java.util.*;

import leon.app.*;
import leon.data.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.util.*;


/**
 * This class is used to decode events received from the notification service and update the
 * DataSource.
 *
 * @version  $Id: LyEventDecoder.java 36436 2012-08-08 09:02:05Z jlsaouli $
 * @author   Lyria
 * @see      LyEventConnection
 */
public class LyEventDecoder
{
    public static final String __VERSION = "$Id: LyEventDecoder.java 36436 2012-08-08 09:02:05Z jlsaouli $";

    /** Notifier instance. */
    protected LyEventNotifier _notifier;

    /** Datasource instance. */
    protected LyDataSource _dataSource;

    /** Separator within ids. */
    protected String _idSep = "\\";

    /** Id of the group for grouped events. */
    protected String _groupId;

    /**
     * Default constructor Initialization is done with the init method.
     *
     * @param  notifier    The associated notifier.
     * @param  dataSource  The dataSource for which an event is decoded.
     */
    public LyEventDecoder(LyEventNotifier notifier, LyDataSource dataSource)
    {
        init();

        _notifier = notifier;
        _dataSource = dataSource;
    }

    /**
     * Initialisation of the data of the instance.
     */
    protected void init()
    {
        _notifier = null;
        _dataSource = null;
        _groupId = null;
    }

    /**
     * Sets group id for this decoder.
     *
     * @param  groupId  Id of the group of messages received by this decoder.
     */
    public void setGroupId(String groupId)
    {
        _groupId = groupId;
    }

    /**
     * Accessor to the environment instance.
     *
     * @return  current environment instance
     */
    protected LyEnvironment getEnvironment()
    {
        return _notifier.getApplication().getEnvironment();
    }

    /**
     * Invoked when an object is modified.
     *
     * @param   id      object identifier
     * @param   values  list of modified values of the object; May be null if not specified.
     *
     * @return  true id all is OK.
     */
    public boolean receiveModification(String id, LyValueSet values)
    {
        if (id == null)
            throw new IllegalArgumentException("BF AppEngine : "
                + getEnvironment().getMessage("LY_ER_CANT_BE_NULL", "id"));

        Object     decodedId;
        LyObjectId objectId;
        LyClass    dataClass = _dataSource.getDataClass();

        decodedId = LyString.stringToList(id, LyTools.charFromString(_idSep));
        objectId = new LyObjectId(dataClass, decodedId);

        return receiveModification(objectId, values);
    }

    /**
     * Invoked when an object is modified.
     *
     * @param   objectId  object identifier
     * @param   values    list of modified values of the object; May be null if not specified.
     *
     * @return  true if all is OK.
     */
    public boolean receiveModification(LyObjectId objectId, LyValueSet values)
    {
        if (objectId == null)
            throw new IllegalArgumentException("BF AppEngine : "
                + getEnvironment().getMessage("LY_ER_CANT_BE_NULL", "objectId"));

        if (getEnvironment().isDebugLevel(LyEnvironment.DEBUG_BUS))
            getEnvironment().logOut("receiveModification for " + objectId);

        if (skipModification(objectId, values))
        {
            if (getEnvironment().isDebugLevel(LyEnvironment.DEBUG_BUS))
                getEnvironment().logOut(" - Modification of " + objectId + " is skipped.");

            return true;
        }

        Object     object = null;
        LyValueSet oldValues = null;

        // Check if the object is in the cache
        if (_dataSource.checkObject(objectId))
        {
            // Check if we have the modified values
            if (values != null)
            {
                // The notification has the values of the object. We try to set it.
                try
                {
                    object = _dataSource.getObject(objectId);

                    oldValues = commitChanges(object, values);

                    // Ok : now we notify the dataSource.
                    return notifyDataSource(LyEventNotifier.EventType.UPDATE, object, values, oldValues);
                }
                catch (Exception e)
                {
                    if (getEnvironment().isDebugLevel(LyEnvironment.DEBUG_BUS))
                        getEnvironment().writeException(e);

                    object = null;
                }

                if (object == null)
                    // There is an error : we must reload the object.
                    // That's all : the job is made with the reload.
                    object = loadObject(objectId);
            }
            else
                // No values in the notification : we must reload the object.
                // That's all : the job is made with the reload.
                object = loadObject(objectId);
        }
        else
            // The object is not in the cache, we load it to test if the changes
            // have some interest.
            // That's all : the job is made with the reload.
            object = loadObject(objectId);

        return true;
    } // end method receiveModification

    /**
     * Check if the notification is useful or not. By default, when cache is empty we ignore the
     * notification.
     *
     * @param   objectId  current object being notified of a modification
     * @param   values    current values (changes)
     *
     * @return  true if the event is skipped, false otherwise
     */
    protected boolean skipModification(LyObjectId objectId, LyValueSet values)
    {
        if ((_dataSource.getCache() != null)
        		&& _dataSource.getCache().isEmpty())
            return true; // XXX Avoid loading of caches on such events
        else
        	return false;
    }

    /**
     * Utility method to get and commit the modifications in an object (peer). Current
     * implementation is valid for instances of LyRowObject, otherwise this method should be
     * overriden.
     *
     * @param   object  current object being modified
     * @param   values  The changed values.
     *
     * @return  the modifications, or null if the modifications are unkwnown
     */
    public LyValueSet commitChanges(Object object, LyValueSet values)
    {
        if (values == null)
            return null;

        LyClass    dataClass = _dataSource.getDataClass();
        LyValueSet oldValues = null;
        LyValue    value;
        int        i, n = values.getSize();
        LyField    field;
        Object     val;

        for (i = 0; i < n; i++)
        {
            value = values.getFieldValue(i);
            field = value.getField();
            val = value.getValue();
            if ((field instanceof LyColumnField)
            	&& (("BLOB".equals(((LyColumnField)field).getPeerType()))
            		 || ("CLOB".equals(((LyColumnField)field).getPeerType()))))
                dataClass.setValue(object, field, LyPeerClass.VALUE_NOT_LOADED);                  	
            else
            	dataClass.setValue(object, field, val);
        }
        // Can get modifications ONLY if the peer objects are instances of LyRowObject
        // Otherwise override this method.
        if (object instanceof LyRowObject)
            oldValues = ((LyRowObject)object).validate(null);

        return oldValues;
    } // end method commitChanges

    /**
     * Invoked when an object is created.
     *
     * @param   id      object identifier
     * @param   values  list of values of the created object; May be null if not specified.
     *
     * @return  true if all is OK.
     */
    public boolean receiveCreation(String id, LyValueSet values)
    {
        Object     decodedId;
        LyObjectId objectId;
        LyClass    dataClass = _dataSource.getDataClass();

        decodedId = LyString.stringToList(id, LyTools.charFromString(_idSep));
        objectId = new LyObjectId(dataClass, decodedId);

        return receiveCreation(objectId, values);
    }

    /**
     * Invoked when an object is created.
     *
     * @param   objectId  object identifier
     * @param   values    list of values of the created object; May be null if not specified.
     *
     * @return  true if all is OK.
     */
    public boolean receiveCreation(LyObjectId objectId, LyValueSet values)
    {
        if (getEnvironment().isDebugLevel(LyEnvironment.DEBUG_BUS))
            getEnvironment().logOut("receiveCreation for " + objectId);

        LyClass dataClass = _dataSource.getDataClass();
        Object  object = null;

        // We do have the set of values : do not load from data provider
        if (values != null)
        {
            // The notification has the values of the object. We try to create it.
            try
            {
                object = dataClass.newInstance();
            }
            catch (InstantiationException e)
            {
                throw new RuntimeException(e);
            }
            catch (IllegalAccessException e)
            {
                throw new RuntimeException(e);
            }

            commitChanges(object, values);
        }

		if (object == null)
		{
			// Rely on loading to notify creation/loading
			object = loadObject(objectId);
			return (object != null);
		}
		else
			// Notify creation
			return notifyDataSource(LyEventNotifier.EventType.CREATE, object, values, null);
    } // end method receiveCreation

    /**
     * Invoked when an object is deleted. This method builds the LyObjectId and calls
     * receiveDeletion(LyObjectId)
     *
     * @param   id  deleted object id
     *
     * @see     #receiveDeletion(LyObjectId)
     *
     * @return  true if the deletion was successful
     */
    public boolean receiveDeletion(String id)
    {
        if (id == null)
            throw new IllegalArgumentException("BF AppEngine : "
                + getEnvironment().getMessage("LY_ER_CANT_BE_NULL", "id"));

        if (getEnvironment().isDebugLevel(LyEnvironment.DEBUG_BUS))
            getEnvironment().logOut("receiveDeletion for " + id);

        LyClass    dataClass = _dataSource.getDataClass();
        LyObjectId objectId;
        Object     decodedId;

        decodedId = LyString.stringToList(id, LyTools.charFromString(_idSep));
        objectId = new LyObjectId(dataClass, decodedId);

        return receiveDeletion(objectId);
    }

    /**
     * Invoked when an object is deleted. This method implements the impact on the data layer.
     *
     * @param   objectId  deleted object id
     *
     * @return  true if the deletion was successful
     */
    public boolean receiveDeletion(LyObjectId objectId)
    {
        if ((_dataSource.getCache() != null)
                && _dataSource.getCache().isEmpty())
            return true; // XXX Avoid loading of caches on such events

        if (!_dataSource.checkObject(objectId))
            return true;

        LyCache cache = _dataSource.getCache();
        Object  object = null;

        try
        {
            // Take object directly in cache
            object = cache.getObject(objectId);
        }
        catch (LyNoSuchObjectException e)
        {
            object = null;
        }

        if (object == null)
            return true;

        return notifyDataSource(LyEventNotifier.EventType.DELETE, object, null, null);
    }

    /**
     * Load an object from the specified object id.
     *
     * @param   objectId  the object id.
     *
     * @return  the loaded object or null if there is no object with the specfied object id.
     */
    protected Object loadObject(LyObjectId objectId)
    {
        // Try to get the object from data provider.
        boolean           synchronous;
        LyResponse        response;
        LyResponse.Status status;
        Object            object;

        // The request must be synchronous.
        synchronous = _dataSource.isSynchronous();
        _dataSource.setSynchronous(true);

        // Load the object.
        LyTransactionManager tm = _dataSource.getApplication().getTransactionManager();
        LyTransaction        transaction = tm.begin();

        response = _dataSource.loadObject(objectId, null, new _LyEventDecoderLoadResponseListener(tm, transaction),
                null);

        // Set the older synchronous mode.
        _dataSource.setSynchronous(synchronous);

        // Check the status of the response.
        status = response.getStatus();

        if ((status == LyResponse.Status.COMPLETED)
                && (response.getSize() == 1))
            object = response.getResponse(0);
        else
            object = null;

        return object;
    } // end method loadObject

    /**
     * To notify the data source for a new event.
     *
     * @param   type    The type of the event.
     * @param   object  The object of the event.
     * @param   values  The values of the event.
     * @param   values  The old values.
     *
     * @return  true if the event was used by the data source.
     */
    protected boolean notifyDataSource(LyEventNotifier.EventType type, Object object, LyValueSet values,
        LyValueSet oldValues)
    {
        Date    start = new Date();
        boolean result = _dataSource.receiveEvent(_groupId, type, object, oldValues);
        long    delay = (new Date()).getTime() - start.getTime();

        // Used in record mode to register events
        LyEnvironment environment = _notifier.getEnvironment();
        LyPlayer      player = environment.getPlayer();

        if (player instanceof LyEventRegister)
        {
            LyApplication           application = _notifier.getApplication();
            LyClass                 dataClass = _dataSource.getDataClass();
            HashMap<String, Object> valueMap = null;

            if (values != null)
            {
                int     i, n = values.getSize();
                LyValue value;
                LyField field;

                valueMap = new HashMap<String, Object>();
                for (i = 0; i < n; i++)
                {
                    value = values.getFieldValue(i);
                    field = value.getField();
                    valueMap.put(field.getId(), value.getValue());
                }
            }

            ((LyEventRegister)player).registerEvent(application.getId(), _notifier.getId(), _groupId, type,
                dataClass.getId(), dataClass.getObjectId(object).toString(), valueMap, delay);
        }

        return result;
    } // end method notifyDataSource

    /**
     * Internal class to listen for responses for a load operation.
     */
    class _LyEventDecoderLoadResponseListener
        implements LyResponseListener
    {
        LyTransactionManager _manager;
        LyTransaction        _transaction;

        _LyEventDecoderLoadResponseListener(LyTransactionManager manager, LyTransaction transaction)
        {
            _manager = manager;
            _transaction = transaction;
        }

        /**
         * Method for receiving a new object in a response.
         */
        public void newResponse(LyResponse response, Object object, LyValueSet oldValues)
        {
            if (response.getStatus() == LyResponse.Status.COMPLETED)
            {
                response.setNotified(true);
                _manager.end(_transaction);
                _transaction = null;
            }
        }

        /**
         * Method for ending a response.
         */
        public void endResponse(LyResponse response)
        {
            if (_transaction != null)
                _manager.end(_transaction);
        }
    } // end class _LyEventDecoderLoadResponseListener
} // end class LyEventDecoder
