/*
 * LyEventEncoder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.notifier;

import leon.data.*;
import leon.misc.*;


/**
 * This class is used by DataSources to transmit events to an instance of LyEventConnection.
 *
 * @version $Id: LyEventEncoder.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     LyEventConnection
 */
public class LyEventEncoder
{
	public static final String __VERSION =
		"$Id: LyEventEncoder.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Name of the type MODIFICATION. */
	public static final String TYPE_MODIFICATION = "MODIFICATION";

	/** Name of the type CREATION. */
	public static final String TYPE_CREATION = "CREATION";

	/** Name of the type DELETION. */
	public static final String TYPE_DELETION = "DELETION";

	/** Name of the type GROUP. */
	public static final String TYPE_GROUP = "GROUP";

	/** Notifier instance. */
	protected LyEventNotifier  _notifier;

	/** Datasource instance. */
	protected LyDataSource     _dataSource;

	/**
	 * Default constructor Initialization is done with the init method.
	 *
	 * @param notifier   The associated notifier
	 * @param dataSource The datasource for which an event is encoded.
	 */
	public LyEventEncoder(LyEventNotifier notifier, LyDataSource dataSource)
	{
		init();

		_notifier = notifier;
		_dataSource = dataSource;
	}

	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		_notifier = null;
		_dataSource = null;
	}

	/**
	 * Invoked when an object is deleted.
	 *
	 * @param  object object which is deleted.
	 * @return true if all is OK.
	 */
	public boolean postDeletion(Object object)
	{
		return true;
	}

	/**
	 * Invoked when an object is created. The notification is sent with values.
	 *
	 * @param  object object which has been created.
	 * @return true if the event was sent successfully
	 */
	public boolean postCreation(Object object)
	{
		return true;
	}

	/**
	 * Invoked when an object is modified. The notification is sent with values if oldValues is not
	 * set to null. If specified, old values must contain all modified fields.
	 *
	 * @param  object    object which has been modified.
	 * @param  oldValues set of values which have been modified
	 * @return true if the event was sent successfully
	 */
	public boolean postModification(Object object, LyValueSet oldValues)
	{
		return true;
	}

	/**
	 * Invoked to notify grouped events.
	 *
	 * @param  groupId   Id of the group of events. This id is used to group the notifications.
	 * @param  type      Type of the event. One of the NOTIF_XXX const in LyEventNotifier.
	 * @param  object    object impacted by the event.
	 * @param  oldValues set of values which have been modified in case of NOTIF_UPDATE.
	 * @return true if the event was sent successfully
	 */
	public boolean postGroupedEvent(String groupId, LyEventNotifier.EventType type,
        Object object, LyValueSet oldValues)
	{
		return true;
	}

	/**
	 * Accessor to the environment instance.
	 *
	 * @return the environment instance
	 */
	protected LyEnvironment getEnvironment()
	{
		return _notifier.getApplication().getEnvironment();
	}
} // end class LyEventEncoder
