/*
 * LyEventNotifier.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.notifier;

import java.io.*;
import java.util.*;

import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.notifier.*;
import leon.misc.*;


/**
 * This class is used by DataSources to transmit their events to an instance of EventEncoder.
 *
 * @version  $Id: LyEventNotifier.java 28982 2010-12-22 11:06:21Z jlsaouli $
 * @author   Lyria
 * @see      LyEventEncoder
 * @see      LyEventConnection
 */
public class LyEventNotifier
{
    public static final String __VERSION = "$Id: LyEventNotifier.java 28982 2010-12-22 11:06:21Z jlsaouli $";

    /** Notification types. */
    public enum EventType
    {
        CREATE, DELETE, UPDATE
    }

    /** Prefix for properties in ServletContext.* */
    public static String CONTEXT_PREFIX = "JMS_";

    /** The application instance. */
    protected LyApplication _application;

    /** Current notifier (contains all notifier parameters from the meta-model). */
    protected LyNotifier _notifier;

    /** Last error message. */
    protected String _error;

    /** Initialization properties. */
    protected Properties _initProperties;

    /** The location id. */
    protected String _locationId;

    /** Current notifying in group status. */
    protected boolean _groupedEvent;

    /** Indicates if notifier is ready or not. */
    protected boolean _ready;

    /**
     * Constructor Initialization is done with the init method.
     *
     * @param       application  the application instance
     * @param       notifier     contains the description of notification service parameters.
     *
     * @deprecated  replaced by LyEventNotifier(LyApplication application, LyNotifier notifier,
     *              Properties properties)
     */
    public LyEventNotifier(LyApplication application, LyNotifier notifier)
    {
        this(application, notifier, null);
    }

    /**
     * Constructor Initialization is done with the init method.
     *
     * @param  application  the application instance
     * @param  notifier     contains the description of notification service parameters.
     * @param  properties   the initialization properties or null if found in a properties file
     */
    public LyEventNotifier(LyApplication application, LyNotifier notifier, Properties properties)
    {
        init();
        _application = application;
        _notifier = notifier;
        _initProperties = properties;
    }

    /**
     * Initialisation of the data of the instance.
     */
    protected void init()
    {
        _notifier = null;
        _error = null;
        _initProperties = null;
        _locationId = null;
        _ready = false;
    }

    /**
     * Accessor to the properties file or parameter if specified in constructor.
     *
     * @return  the properties.
     */
    public Properties getProperties()
    {
        if (_initProperties != null)
            return _initProperties;

        // try from web.xml first
        LyEnvironment environment = getEnvironment();
        Properties    properties = environment.getPropertiesFromContext(CONTEXT_PREFIX);

        if (properties != null)
            return properties;

        properties = _notifier.getPropertiesSet();

        if (properties != null)
            return properties;

        String filename = _notifier.getProperties();

        if (filename == null)
        {
            environment.logError("BF AppEngine : JMS properties filename is null");
            return null;
        }

        LyApplication application = _notifier.getApplication();
        ClassLoader   loader = application.getClassLoader();

        if (loader == null)
        {
            application = getApplication();
            loader = application.getClassLoader();
        }

        InputStream stream = environment.getResourceAsStream(filename, loader);

        if (stream == null)
        {
            environment.logError("BF AppEngine : JMS properties filename is invalid : " + filename);
            return null;
        }

        try
        {
            properties = new Properties();
            properties.load(stream);
        }
        catch (Exception e)
        {
            if (environment.isDebugLevel(LyEnvironment.DEBUG_ERROR))
                environment.writeException(e);

            properties = new Properties();
        }

        return properties;
    } // end method getProperties

    /**
     * To get the id of the associated location.
     *
     * @return  The id of the associated location.
     */
    public String getLocationId()
    {
        if (_locationId == null)
            _locationId = _notifier.getLocation().getId();

        return _locationId;
    }

    /**
     * Set an error message : an error occured. and the error description is accessible through this
     * accessor
     *
     * @param  message  the error message
     */
    public void setError(String message)
    {
        _error = message;
        _application.setError(message);
    }

    /**
     * This method is used to build all the class peers for all the classInfo. It must be
     * specialized by subclasses. Basically, this method associates the dataProvider with local
     * classes. A class is local if there is no binding.
     *
     * @param   classInfos  Classes to validate.
     *
     * @return  True if the validation was successful, false otherwise.
     */
    public boolean validate(LyInfoList<LyClassInfo> classInfos)
    {
        return true;
    }

    /**
     * Close current connection.
     *
     * @return  true if the operation is successful
     */
    public boolean close()
    {
        _ready = false;
        return true;
    }

    /**
     * Create and open a new connection.
     *
     * @return  true if the operation is successful
     */
    public boolean createConnection()
    {
        _ready = true;
        return true;
    }

    /**
     * Invoked when the application is going on sleeping mode.
     */
    public void sleep()
    {
        close();
    }

    /**
     * Invoked when the application is going on sleeping mode.
     */
    public boolean wakeUp()
    {
        LyEnvironment env = getEnvironment();

        try
        {
            if (!createConnection())
            {
                getEnvironment().raiseError(env.getMessage("LY_ER_NOTIFIER_NOT_READY"));
                return false;
            }
        }
        catch (Exception e)
        {
            env.logError(env.getMessage("LY_ER_NOTIFIER_NOT_READY"));
            return false;
        }

        return true;
    }

    /**
     * Accessor to the environment.
     *
     * @return  the application instance
     */
    public LyApplication getApplication()
    {
        return _application;
    }

    /**
     * Accessor to the environment.
     *
     * @return  the environment instance
     */
    public LyEnvironment getEnvironment()
    {
        return _application.getEnvironment();
    }

    /**
     * Accessor to the notifier description (contains parameters read from the meta model).
     *
     * @return  the notifier instance
     */
    public LyNotifier getNotifier()
    {
        return _notifier;
    }

    /**
     * Shortcut to the notifier id : getNotifier().getId().
     *
     * @return  the notifier is
     */
    public String getId()
    {
        return ((_notifier == null) ? null : _notifier.getId());
    }

    /**
     * Accessor to the connection. Should be implemented by subclasses.
     *
     * @return  the connection instance.
     */
    public LyEventConnection getConnection()
    {
        return null;
    }

    /**
     * Accessor to a valid encoder instance for the specified datasource.
     *
     * @param   dataSource  the dataSource which creates the event
     *
     * @return  current encoder instance
     */
    public LyEventEncoder getEventEncoder(LyDataSource dataSource)
    {
        return new LyEventEncoder(this, dataSource);
    }

    /**
     * Accessor to a valid decoder instance for the specified datasource.
     *
     * @param   dataSource  the dataSource which will handle the event
     *
     * @return  current decoder instance
     */
    public LyEventDecoder getEventDecoder(LyDataSource dataSource)
    {
        return new LyEventDecoder(this, dataSource);
    }

    /**
     * Tests if this notifier supports the XA protocol or not. Subclasses may override this method.
     * The default implementation is to return false.
     *
     * @return  true if this notifier supports the XA protocol, false otherwise.
     */
    public boolean isXASupported()
    {
        return false;
    }

    /**
     * Indicates if the notifier is currently notifying a grouped event message : current returns
     * false. This is used internally to distinguish a load from the notifier and from the
     * application (UnionClass make the distinction for performance issues)
     *
     * @return  notifying status
     */
    public boolean isGroupedEvent()
    {
        return _groupedEvent;
    }

    /**
     * Indicates if the notifying is currently notifying a grouped event message : internal use
     * only. This is used internally to distinguish a load from the notifier and from the
     * application (UnionClass make the distinction for performance issues)
     *
     * @param  status  notifying status
     */
    public void setGroupedEvent(boolean status)
    {
        _groupedEvent = status;
    }

    /**
     * Internal method to set the locationId for this notifier.
     *
     * @param  locationId  The new locationId of this notifier.
     */
    public void setLocationId(String locationId)
    {
        _locationId = locationId;
    }

    /**
     * To know if this notifier supports the grouped events.
     *
     * @return  true if this notifier supports the grouped events.
     */
    public boolean supportsGroupedEvents()
    {
        return false;
    }

    /**
     * To send events for a specified group identifier.
     *
     * @param  groupId  the group identifier.
     */
    public void sendGroupedEvents(String groupId)
    {
    }

    /**
     * Checks if notifier is ready or not.
     */
    public boolean isReady()
    {
        return _ready;
    }
} // end class LyEventNotifier
