/*
 * LySaxParser.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.parser;

import org.xml.sax.*;
import org.xml.sax.helpers.*;
import java.io.*;

import javax.xml.parsers.*;
import leon.misc.*;


/**
 * This class represents our abstract Xml Sax parser.
 * @see     LyXmlNode
 * @version $Id: LySAXParser.java 35266 2012-04-18 13:26:55Z jlsaouli $
 * @author  Lyria
 */
public abstract class LySAXParser
{
	public static final String __VERSION =
		"$Id: LySAXParser.java 35266 2012-04-18 13:26:55Z jlsaouli $";

	/**
	 * Indicates that the XML parser should search in the CLASSPATH for entities.
	 */
	public static boolean	RESOLVE_XML_ENTITIES = true;
	
	/** When parsing the meta model, comments are usually ignored. If this value is
	 *  set, comments containing this String are kept on next LyInfo and may be used by code. Studio keeps them.
	 */
	public static String	COMMENTS_START_KEY = null;

	/** When parsing the meta model, comments are usually ignored. If this value is
	 *  set, comments containing this String are kept on next LyInfo and may be used by code. Studio keeps them.
	 */
	public static String	COMMENTS_END_KEY = null;

	/** Instance associated with the root element of the XML tree. */
	protected Object		_rootElement;

	/** Indicates if the parser is validating or not. */
	protected boolean		_validating;


	/**
	 * Constructs a validating SAX parser.
	 */
	public LySAXParser()
	{
		this(true);
	}

	/**
	 * Constructs a SAX parser.
	 * @param validating indicates if the parser is validating or not.
	 */
	public LySAXParser(boolean validating)
	{
		_validating = validating;
	}


	/**
	 * Set the features of the parser.
	 * @param  parser the xml parser
	 * @throws SAXException if something is wrong during feature sets
	 */
	protected void setParserFeatures(XMLReader parser)
		throws SAXException
	{
		// Use a validating parser. Comment this line to turn off validation
		parser.setFeature("http://xml.org/sax/features/validation", _validating);
		parser.setFeature("http://apache.org/xml/features/validation/schema", _validating);
		parser.setFeature("http://xml.org/sax/features/namespaces", false);
	}

	/**
	 * Load an XML Document with a given property file to build appropriate nodes.
	 * @param xmlFile     Path of the file that contains the XML info.
	 * @param factory     used to build the default handler used by the parser to load the document
	 * @param classLoader used to build the entity resolver used by the parser to load the document
	 * @param environment the environment
	 */
	protected void loadDocument(InputStream is, LyXMLElementFactory factory,
		ClassLoader classLoader, LyEnvironment environment)
	{
		XMLReader		  parser = null;
		_LyDefaultHandler handler = null;
		boolean			  parsingOk = false;

		
		if (is == null)
			throw new IllegalStateException("BF AppEngine : input stream can't be null");
		
		BufferedInputStream stream = new BufferedInputStream(is);
		
		stream.mark(-1);
		try
		{
			// Load the document.
			LySimpleEntityResolver entityResolver = null;

			// ** Create Sax Parser **
			String				   defaultParser = environment.getEnv("_LY_XML_PARSER");
			boolean				   forceParser = false;

			if (defaultParser == null)
			{
				defaultParser = System.getProperty("org.xml.sax.driver");

				if (defaultParser == null)
					defaultParser = "org.apache.xerces.parsers.SAXParser";
			}
			else
				forceParser = true;

			System.setProperty("org.xml.sax.driver", defaultParser);

			try
			{
				// We use a SAX2 parser API... (name space support for future use)
				// Attempt to use default parser
				// parser = XMLReaderFactory.createXMLReader();
				if (!forceParser)
				{
					SAXParserFactory factorySAX = SAXParserFactory.newInstance();

					if (factorySAX != null)
					{
						factorySAX.setNamespaceAware(true);
						SAXParser saxParser = factorySAX.newSAXParser();

						if (saxParser != null)
							parser = saxParser.getXMLReader();
					}
				}

				// May use apache's xerces parser as default
				if (parser == null)
					parser = XMLReaderFactory.createXMLReader(defaultParser);
			}
			catch (SAXException e)
			{
				// attempt with xerces (default : should work)
				parser = XMLReaderFactory.createXMLReader(defaultParser);
			}
			catch (ParserConfigurationException e)
			{
				// attempt with xerces (default : should work)
				parser = XMLReaderFactory.createXMLReader(defaultParser);
			}

			if (parser == null)
				environment.logInfo("Unable to load SAX parser : "
					+ defaultParser
					+ ".\n Try to set the org.xml.sax.driver system property to a valid sax 2 parser.");
			else
				environment.logInfo("Using SAX parser " + parser.getClass().getName());

			try
			{
				// Use a validating parser. Comment this line to turn off validation
				setParserFeatures(parser);
			}
			catch (SAXException e)
			{
				environment.logInfo("Invalid parser feature : " + e.getMessage());
			}
			
			if (RESOLVE_XML_ENTITIES)
			{
				environment.logTrace("Using entity Resolver ...");

				entityResolver = new LyEntityResolver(environment, classLoader);
			}
			else
				entityResolver = new LySimpleEntityResolver(environment);

			parser.setEntityResolver(entityResolver);

			// Try to autodetect XML Schema
			if (_validating)
			{
				try
				{
					StringBuffer sb = new StringBuffer();
					int			 n = 0;
					char		 c;

					while ((n >= 0) && (stream.available() > 0))
					{
						n = stream.read();
						if (n > 0)
						{
							c = (char)n;
							sb.append(c);
							if ((c == '[') || (sb.length() > 1000))
								break;
						}
					}
					if (sb.toString().toLowerCase().indexOf(".dtd") < 0)
					{
						environment.logInfo("Using XML Schema grammar : http://www.w3.org/2001/XMLSchema");
						parser.setProperty("http://java.sun.com/xml/jaxp/properties/schemaLanguage",
							"http://www.w3.org/2001/XMLSchema");
						parser.setFeature("http://xml.org/sax/features/namespaces", true);
					}
					else
						environment.logInfo("Using DTD grammar.");
				}
				catch (Exception e1)
				{
					environment.writeException(e1);
				}
				finally
				{
					// Restore the stream to the start...
					stream.reset();
				}
			}

			handler = new _LyDefaultHandler(factory, environment);
			parser.setContentHandler(handler);
			parser.setErrorHandler(handler);
			
			// Try to handle comments stored on infos (when needed)
			try
			{
				if (((COMMENTS_START_KEY != null)
							&& (COMMENTS_START_KEY.length() > 0))
						|| ((COMMENTS_END_KEY != null)
							&& (COMMENTS_END_KEY.length() > 0)))
					parser.setProperty("http://xml.org/sax/properties/lexical-handler", handler);
			}
			catch (Exception e1)
			{
				environment.writeException(e1);
			}
			
			parser.parse(new InputSource(stream));
				parsingOk = true;
		}
		catch (SAXException t)
		{
			// Display errors on the standard error
			if (t.getMessage() != null)
				environment.logFatal(t.getMessage());
			if ((t.getException() != null) && (environment.isDebugLevel(LyEnvironment.DEBUG_TRACE)))
				environment.writeException(t.getException());
		}
		catch (IOException t)
		{
			if (t.getMessage() != null)
				environment.logFatal(t.getMessage());
			// Display errors on the standard error
			environment.writeException(t);
		}
		finally
		{
			try
			{
				stream.close();
			}
			catch (IOException t)
			{
				if (t.getMessage() != null)
					environment.logFatal(t.getMessage());
				// Display errors on the standard error
				environment.writeException(t);
			}
		}

		// Clean up
		if (handler != null)
		{
			_rootElement = handler.getRootElement();
			handler.free();

			if (parsingOk)
			{
				short errorCount = handler.getErrorCount();
				short warningCount = handler.getWarningCount();

				if (errorCount > 0)
				{
					environment.logError("BF AppEngine : " + errorCount + " errors occured on parsing");
					parsingOk = false;
				}

				if (warningCount > 0)
					environment.logWarning("BF AppEngine : " + warningCount
						+ " warnings raised on parsing");
			}
		}

		if (!parsingOk)
			throw new RuntimeException("BF AppEngine : parsing of inputstream failed.");
	} // end method loadDocument
	
	/**
	 * Load an XML Document with a given property file to build appropriate nodes.
	 * @param xmlFile     Path of the file that contains the XML info.
	 * @param factory     used to build the default handler used by the parser to load the document
	 * @param classLoader used to build the entity resolver used by the parser to load the document
	 * @param environment the environment
	 */
	protected void loadDocument(String xmlFile, LyXMLElementFactory factory,
		ClassLoader classLoader, LyEnvironment environment)
	{
		XMLReader		  parser = null;
		_LyDefaultHandler handler = null;
		boolean			  parsingOk = false;

		if (xmlFile == null)
			throw new IllegalStateException("BF AppEngine : xmlFile can't be null");

		try
		{
			// Load the document.
			InputSource			   xmlSource = null;
			LySimpleEntityResolver entityResolver = null;

			// ** Create Sax Parser **
			String				   defaultParser = environment.getEnv("_LY_XML_PARSER");
			boolean				   forceParser = false;

			if (defaultParser == null)
			{
				defaultParser = System.getProperty("org.xml.sax.driver");

				if (defaultParser == null)
					defaultParser = "org.apache.xerces.parsers.SAXParser";
			}
			else
				forceParser = true;

			System.setProperty("org.xml.sax.driver", defaultParser);

			try
			{
				// We use a SAX2 parser API... (name space support for future use)
				// Attempt to use default parser
				// parser = XMLReaderFactory.createXMLReader();
				if (!forceParser)
				{
					SAXParserFactory factorySAX = SAXParserFactory.newInstance();

					if (factorySAX != null)
					{
						factorySAX.setNamespaceAware(true);
						SAXParser saxParser = factorySAX.newSAXParser();

						if (saxParser != null)
							parser = saxParser.getXMLReader();
					}
				}

				// May use apache's xerces parser as default
				if (parser == null)
					parser = XMLReaderFactory.createXMLReader(defaultParser);
			}
			catch (SAXException e)
			{
				// attempt with xerces (default : should work)
				parser = XMLReaderFactory.createXMLReader(defaultParser);
			}
			catch (ParserConfigurationException e)
			{
				// attempt with xerces (default : should work)
				parser = XMLReaderFactory.createXMLReader(defaultParser);
			}

			if (parser == null)
				environment.logInfo("Unable to load SAX parser : "
					+ defaultParser
					+ ".\n Try to set the org.xml.sax.driver system property to a valid sax 2 parser.");
			else
				environment.logInfo("Using SAX parser " + parser.getClass().getName());

			try
			{
				// Use a validating parser. Comment this line to turn off validation
				setParserFeatures(parser);
			}
			catch (SAXException e)
			{
				environment.logInfo("Invalid parser feature : " + e.getMessage());
			}

			if ((!xmlFile.toLowerCase().startsWith("http:"))
					&& (!xmlFile.toLowerCase().startsWith("file:")))
			{
				xmlFile = environment.getString().expand(xmlFile);
				xmlFile = xmlFile.replace('\\', '/');
				xmlFile = "file:///" + xmlFile;
			}

			InputStream autoDetectSchema = null;

			// Try to use a specific entity resolver (using CLASSPATH)
			if (RESOLVE_XML_ENTITIES)
			{
				environment.logTrace("Using entity Resolver ...");

				entityResolver = new LyEntityResolver(environment, classLoader);
				parser.setEntityResolver(entityResolver);
				xmlSource = entityResolver.getInputSource(xmlFile);

				if (_validating && (xmlSource != null))
					autoDetectSchema = entityResolver.getInputSource(xmlFile).getByteStream();
			}
			else
			{
				entityResolver = new LySimpleEntityResolver(environment);
				parser.setEntityResolver(entityResolver);

				if (_validating)
					autoDetectSchema = new InputSource(xmlFile).getByteStream();
			}

			// Try to autodetect XML Schema
			if (autoDetectSchema != null)
			{
				try
				{
					StringBuffer sb = new StringBuffer();
					int			 n = 0;
					char		 c;

					while ((n >= 0) && (autoDetectSchema.available() > 0))
					{
						n = autoDetectSchema.read();
						if (n > 0)
						{
							c = (char)n;
							sb.append(c);
							if ((c == '[') || (sb.length() > 1000))
								break;
						}
					}
					if (sb.toString().toLowerCase().indexOf(".dtd") < 0)
					{
						environment.logInfo("Using XML Schema grammar : http://www.w3.org/2001/XMLSchema");
						parser.setProperty("http://java.sun.com/xml/jaxp/properties/schemaLanguage",
							"http://www.w3.org/2001/XMLSchema");
						parser.setFeature("http://xml.org/sax/features/namespaces", true);
					}
					else
						environment.logInfo("Using DTD grammar.");
				}
				catch (Exception e1)
				{
					environment.writeException(e1);
				}
				finally
				{
					if (autoDetectSchema != null)
						autoDetectSchema.close();
				}
			}

			handler = new _LyDefaultHandler(factory, environment);
			parser.setContentHandler(handler);
			parser.setErrorHandler(handler);
			
			// Try to handle comments stored on infos (when needed)
			try
			{
				if (((COMMENTS_START_KEY != null)
							&& (COMMENTS_START_KEY.length() > 0))
						|| ((COMMENTS_END_KEY != null)
							&& (COMMENTS_END_KEY.length() > 0)))
					parser.setProperty("http://xml.org/sax/properties/lexical-handler", handler);
			}
			catch (Exception e1)
			{
				environment.writeException(e1);
			}

			if (xmlSource == null)
				parser.parse(xmlFile);
			else
				parser.parse(xmlSource);

			parsingOk = true;
		}
		catch (SAXException t)
		{
			// Display errors on the standard error
			if (t.getMessage() != null)
				environment.logFatal(t.getMessage());
			if ((t.getException() != null) && (environment.isDebugLevel(LyEnvironment.DEBUG_TRACE)))
				environment.writeException(t.getException());
		}
		catch (IOException t)
		{
			if (t.getMessage() != null)
				environment.logFatal(t.getMessage());
			// Display errors on the standard error
			environment.writeException(t);
		}

		// Clean up
		if (handler != null)
		{
			_rootElement = handler.getRootElement();
			handler.free();

			if (parsingOk)
			{
				short errorCount = handler.getErrorCount();
				short warningCount = handler.getWarningCount();

				if (errorCount > 0)
				{
					environment.logError("BF AppEngine : " + errorCount + " errors occured on parsing");
					parsingOk = false;
				}

				if (warningCount > 0)
					environment.logWarning("BF AppEngine : " + warningCount
						+ " warnings raised on parsing");
			}
		}

		if (!parsingOk)
			throw new RuntimeException("BF AppEngine : parsing of " + xmlFile + " failed.");
	} // end method loadDocument

	/**
	 * Gets the parser root element.
	 * @return Root element built by this parser.
	 */
	public Object getRootElement()
	{
		return _rootElement;
	}

	/**
	 * Checks whether this parser is validating or not.
	 * @return True if validating, false otherwise.
	 */
	public boolean isValidating()
	{
		return _validating;
	}

	/**
	 * Builds a default factory. <b>BE CAREFUL:</b> this method always throws an
	 * IllegalStateException!
	 * @return a factory
	 */
	protected LyXMLElementFactory buildDefaultFactory()
	{
		throw new IllegalStateException(
			"BF AppEngine : can't build a default factory for parser handler");
	}

	/**
	 * Internal class to display parsing errors and warnings.
	 * @see org.xml.sax.ErrorHandler
	 */
	class _LyDefaultHandler extends DefaultHandler
		implements org.xml.sax.ext.LexicalHandler
	{

		/** Instance associated to the XML tree root element. */
		Object			    _rootElement;

		/** Factory used to create the requested class instances associated to parsed nodes. */
		LyXMLElementFactory _factory;

		/** The parsing environment. */
		LyEnvironment	    _environment;

		/** Instance of current element being parsed. */
		_LyTreeStruct	    _currentLevel;

		/** Number of errors that occurred during parsing. */
		short			    _errorCount;

		/** Number of warnings raised during parsing. */
		short			    _warningCount;

		/** Last comment read from the XML file (stored on next info tag). */
		String			    _commentsStart;

		/** Last info to store 'endComments'. */
		LyCommentInterface  _lastCommentInfo;

		/**
		 * Default constructor.
		 *
		 * @param factory     the parser factory
		 * @param environment the environment
		 */
		public _LyDefaultHandler(LyXMLElementFactory factory, LyEnvironment environment)
		{
			_rootElement = null;
			_factory = factory;
			_environment = environment;
			_currentLevel = null;
			_errorCount = 0;
			_warningCount = 0;
			_commentsStart = null;
		}

		/**
		 * Utility method to display parsing errors to the standard output.
		 *
		 * @param level Indicates the severity of the error (warning, error, fatal error).
		 * @param e     The exception raised on the error.
		 */
		protected void displayMessage(String level, SAXParseException e)
		{
			_environment.logErr("\nBF AppEngine parsing error : " + level + ".");
			_environment.logErr("      URI = " + e.getSystemId());
			_environment.logErr("     Line = " + e.getLineNumber());
			_environment.logErr("  Message = " + e.getMessage());
		}

		/**
		 * Retrieve the root element of the parser.
		 *
		 * @return the root element
		 */
		public Object getRootElement()
		{
			return _rootElement;
		}

		/**
		 * Gets the number of errors that occured during the parisng.
		 *
		 * @return the error count
		 */
		public short getErrorCount()
		{
			return _errorCount;
		}

		/**
		 * Gets the number of warnings that occured during the parisng.
		 *
		 * @return the warning count
		 */
		public short getWarningCount()
		{
			return _warningCount;
		}

		/*
		 * (non-Javadoc)
		 *
		 * @see org.xml.sax.ContentHandler#startElement(java.lang.String,
		 *      java.lang.String, java.lang.String, org.xml.sax.Attributes)
		 */
		public void startElement(String uri, String localName, String qName, Attributes attributes)
			throws SAXException
		{
			if (_factory == null)
				_factory = LySAXParser.this.buildDefaultFactory();

			// One level deeper
            Object parentElement = null;

            if (_currentLevel != null)
                parentElement = _currentLevel._current;
            
			Object current = _factory.buildElement(uri, localName, qName, attributes, parentElement);

			if ((_commentsStart != null) && (current instanceof LyCommentInterface))
			{
				((LyCommentInterface)current).setCommentsStart(_commentsStart);
				_commentsStart = null;
			}

			// Keep current hierarchy (for this level only)
			// We just keep parent reference, not the whole XML tree
			if (_currentLevel == null)
			{
				_currentLevel = new _LyTreeStruct(null, current);
				// Set tree root
				if (_rootElement == null)
					_rootElement = current;
			}
			else
				_currentLevel = new _LyTreeStruct(_currentLevel, current);

			if (current instanceof LyCommentInterface)
				_lastCommentInfo = (LyCommentInterface)current;
		} // end method startElement

		/*
		 * (non-Javadoc)
		 *
		 * @see org.xml.sax.ContentHandler#endElement(java.lang.String,
		 *      java.lang.String, java.lang.String)
		 */
		public void endElement(String uri, String localName, String qName)
			throws SAXException
		{
			// End of current level
			if (_currentLevel != null)
			{
				Object currentElement = _currentLevel._current;
				Object parentElement = null;

				if (_currentLevel._parentTree != null)
					parentElement = _currentLevel._parentTree._current;

				_factory.endElement(uri, localName, qName, parentElement, currentElement);

				// One level up
				_currentLevel = _currentLevel._parentTree;

				// We keep last info interface to store endComments on next tag
				if (currentElement instanceof LyCommentInterface)
					_lastCommentInfo = (LyCommentInterface)currentElement;
			}
		} // end method endElement

		/**
		 * Stores comments read from the meta model file in order to restore them if the meta-model
		 * is written back (in Studio).
		 */
		public void comment(char[] ch, int start, int length)
			throws SAXException
		{
			if ((COMMENTS_START_KEY == null)
					&& (COMMENTS_END_KEY == null))
				return;

			String s = new String(ch, start, length);

			// Keep comment if START comment
			if ((COMMENTS_START_KEY != null)
					&& (s.indexOf(COMMENTS_START_KEY) >= 0))
				_commentsStart = s;

			// Keep comment if END comment
			if ((COMMENTS_END_KEY != null)
					&& (s.indexOf(COMMENTS_END_KEY) >= 0)
					&& (_lastCommentInfo != null))
				_lastCommentInfo.setCommentsEnd(s);
		}

		/**
		 * Does nothing.
		 */
		public void endCDATA()
			throws SAXException
		{
		}

		/**
		 * Does nothing.
		 */
		public void endDTD()
			throws SAXException
		{
		}

		/**
		 * Does nothing.
		 */
		public void endEntity(String name)
			throws SAXException
		{
		}

		/**
		 * Does nothing.
		 */
		public void startCDATA()
			throws SAXException
		{
		}

		/**
		 * Does nothing.
		 */
		public void startDTD(String name, String publicId, String systemId)
			throws SAXException
		{
		}

		/**
		 * Does nothing.
		 */
		public void startEntity(String name)
			throws SAXException
		{
		}

		/**
		 * Catches recoverable errors: if the {@link LyEnvironment#DEBUG_ERROR DEBUG_ERROR} level is
		 * set, displays the message thanks to the {@link #displayMessage(String,
		 * SAXParseException)} method then increments the number of occured errors.
		 *
		 * @see ErrorHandler#error(org.xml.sax.SAXParseException)
		 */
		public void error(SAXParseException e)
		{
			// normally a validity error
			if (_environment.isDebugLevel(LyEnvironment.DEBUG_ERROR))
				displayMessage("Error", e);
			_errorCount++;
		}

		/**
		 * Catches warnings: if the {@link LyEnvironment#DEBUG_WARNING DEBUG_WARNING} level is set,
		 * displays the message thanks to the {@link #displayMessage(String, SAXParseException)}
		 * method then increments the number of occured warnings.
		 *
		 * @see ErrorHandler#warning
		 */
		public void warning(SAXParseException e)
		{
			if (_environment.isDebugLevel(LyEnvironment.DEBUG_WARNING))
				displayMessage("Warning", e);
			_warningCount++;
		}

		/**
		 * Catches fatal errors: displays the message thanks to the {@link #displayMessage(String,
		 * SAXParseException)} method.
		 *
		 * @see ErrorHandler#fatalError
		 */
		public void fatalError(SAXParseException e)
		{
			displayMessage("FATAL ERROR", e);
		}

		/**
		 * Clear references. Sets the {@link #_rootElement root element}, the {@link #_factory
		 * factory} and the {@link #_currentLevel  current level} to null.
		 */
		public void free()
		{
			_rootElement = null;
			_factory = null;
			_currentLevel = null;
		}
	} // end class _LyDefaultHandler


	/**
	 * Simple struct class to keep hierarchical parent tree.
	 */
	class _LyTreeStruct
	{

		/** The parent tree. */
		public _LyTreeStruct _parentTree;

		/** The current element to link with the parent tree. */
		public Object		 _current;

		/**
		 * Constructor. Builds a new object that keep hierarchical parent tree.
		 *
		 * @param parentTree the parent tree
		 * @param current    the curent element
		 */
		public _LyTreeStruct(_LyTreeStruct parentTree, Object current)
		{
			_parentTree = parentTree;
			_current = current;
		}
	}
} // end class LySAXParser
