/*
 * LyXmlNode.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.parser;

import java.text.*;
import java.util.*;


/**
 * This class is used to offer temporary storage to our real node instances. We chose not to extend
 * ElementNode directly from our objects to save memory. In user execution mode, these instances can
 * be freed after loading because they are totally unuseful at this time.
 * @version $Id: LyXmlNode.java 34647 2012-03-14 11:10:10Z jlsaouli $
 * @author  Lyria
 */
public abstract class LyXmlNode
{
	public static final String		  __VERSION =
		"$Id: LyXmlNode.java 34647 2012-03-14 11:10:10Z jlsaouli $";


	/** Code to used to represent empty strings. */
	protected static final String     EMPTY_STRING = "_empty".intern();

	/** Empty values representation. */
	protected static final String     EMPTY_VALUE = "".intern();
    
    /** Empty values representation. */
    protected static final String     TRUE_VALUE = "true".intern();


	/** Date tags values format. */
	protected static SimpleDateFormat _format = new SimpleDateFormat();


	/**
	 * Constructor. Does nothing.
	 */
	public LyXmlNode()
	{
	}


	/**
	 * Gets the given attribute value. Should be implemented in subclasses.
	 * @param  attributeName the attribute name
	 * @return the attribute value as a string
	 */
	public abstract String getAttribute(String attributeName);

	/**
	 * Gets the given attribute value.
	 * @param  attributeName the attribute name.
	 * @param  defaultValue  the value to return for the attribute if no value is specified
	 * @return the value
	 */
	public String getString(String attributeName, String defaultValue)
	{
		String value = getAttribute(attributeName);

		if ((value == null) || (value.length() == 0))
			return defaultValue;

		value = value.intern();
		if (value == EMPTY_STRING)
			return EMPTY_VALUE;

		return value;
	}

	/**
	 * Gets the given attribute value. Only calls {@link #getString(String, String)
	 * getString(attributeName, null)}
	 * @param  attributeName the attribute name.
	 * @return the value
	 * @see    LyXmlNode#getString(String, String)
	 */
	public String getString(String attributeName)
	{
		return getString(attributeName, null);
	}

	/**
	 * Gets the given attribute value as an int.
	 * @param     attributeName the attribute name.
	 * @param     defaultValue  the value to return for the attribute if no value is specified
	 * @return    the value as an int
	 * @exception IllegalArgumentException If the given token cannot be parsed as an int.
	 */
	public int getInt(String attributeName, int defaultValue)
	{
		String value = getString(attributeName);

		if (value == null)
			return defaultValue;

		try
		{
			return Integer.parseInt(value);
		}
		catch (NumberFormatException e)
		{
			throw new IllegalArgumentException("BF AppEngine :  invalid int format " + value);
		}
	}

	/**
	 * Gets the given attribute value. Only calls {@link #getInt(String, int) getInt(attributeName,
	 * 1)}
	 * @param  attributeName the attribute name.
	 * @return the value as an int
	 * @see    LyXmlNode#getInt(String, int)
	 */
	public int getInt(String attributeName)
	{
		return getInt(attributeName, -1);
	}

	/**
	 * Gets the given attribute value as a double.
	 * @param     attributeName the attribute name.
	 * @param     defaultValue  the value to return for the attribute if no value is specified
	 * @return    the value as a double
	 * @exception IllegalArgumentException If the given token cannot be parsed as a double.
	 */
	public double getDouble(String attributeName, double defaultValue)
	{
		String value = getString(attributeName);

		if (value == null)
			return defaultValue;
		
		if ("Double.NEGATIVE_INFINITY".equals(value))
			return Double.NEGATIVE_INFINITY;
		
		if ("Double.POSITIVE_INFINITY".equals(value))
			return Double.POSITIVE_INFINITY;

		try
		{
			return Double.valueOf(value).doubleValue();
		}
		catch (NumberFormatException e)
		{
			throw new IllegalArgumentException("BF AppEngine :  invalid double format " + value);
		}
	}

	/**
	 * Gets the given attribute value. Only calls {@link #getDouble(String, int)
	 * getDouble(attributeName, -1)}
	 * @param  attributeName the attribute name.
	 * @return the value as a double
	 * @see    LyXmlNode#getDouble(String, int)
	 */
	public double getDouble(String attributeName)
	{
		return getDouble(attributeName, -1);
	}

	/**
	 * Gets the given attribute value as a float.
	 * @param     attributeName the attribute name
	 * @param     defaultValue  the value to return for the attribute if no value is specified
	 * @return    the value as a float
	 * @exception IllegalArgumentException If the given token cannot be parsed as a float.
	 */
	public float getFloat(String attributeName, float defaultValue)
	{
		String value = getString(attributeName);

		if (value == null)
			return defaultValue;

		try
		{
			return Float.valueOf(value).floatValue();
		}
		catch (NumberFormatException e)
		{
			throw new IllegalArgumentException("BF AppEngine :  invalid float format " + value);
		}
	}

	/**
	 * Gets the given attribute value. Only calls {@link #getFloat(String, int)
	 * getFloat(attributeName, -1)}
	 * @param  attributeName the attribute name.
	 * @return the value as a float
	 * @see    LyXmlNode#getFloat(String, int)
	 */
	public float getFloat(String attributeName)
	{
		return getFloat(attributeName, -1);
	}

	/**
	 * Gets the given attribute value as a Number.
	 * @param     attributeName the attribute name
	 * @param     numberClass   the Number class for result
	 * @return    the value as a Number
	 * @exception IllegalArgumentException If the given attribute value cannot be parsed as a number
	 *                                     of the given type or if the numberClass doesn't
	 *                                     correspond to any supported number class.
	 */
	public Number getNumber(String attributeName, Class<? extends Number> numberClass)
	{
		String value = getString(attributeName);

		if (value == null)
			return null;

		try
		{
			if (numberClass == Integer.class)
				return new Integer(value);
			if (numberClass == Short.class)
				return new Short(value);
			if (numberClass == Long.class)
				return new Long(value);
			if (numberClass == Double.class)
				return new Double(value);
			if (numberClass == Integer.class)
				return new Float(value);
		}
		catch (NumberFormatException e)
		{
			throw new IllegalArgumentException("BF AppEngine :  invalid number format " + value);
		}

		throw new IllegalArgumentException("BF AppEngine : invalid number class " + numberClass);
	}

	/**
	 * Gets the given attribute value as a date.
	 * @param     attributeName the attribute name
	 * @return    the value as a date
	 * @exception IllegalArgumentException If the given token cannot be parsed as a float.
	 */
	public Date getDate(String attributeName)
	{
		String value = getString(attributeName);

		if (value == null)
			return null;

		try
		{
			synchronized (_format)
			{
				return _format.parse(value);
			}
		}
		catch (ParseException e)
		{
			throw new IllegalArgumentException("BF AppEngine : invalid date format " + value);
		}
	}

	/**
	 * Gets the given attribute value as a float.
	 * @param  attributeName the attribute name
	 * @param  defaultValue  the value to return for the attribute if no value is specified
	 * @return return true if the value equals (ignore case) "true", false otherwise.
	 */
	public boolean getBoolean(String attributeName, boolean defaultValue)
	{
		String value = getString(attributeName);

		if (value == null)
			return defaultValue;

		return (value.intern() == TRUE_VALUE);
	}

	/**
	 * Gets the given attribute value. Only calls {@link #getBoolean(String, boolean)
	 * getBoolean(attributeName, false)}
	 * @param  attributeName the attribute name.
	 * @return the value as a boolean
	 * @see    LyXmlNode#getBoolean(String, boolean)
	 */
	public boolean getBoolean(String attributeName)
	{
		return getBoolean(attributeName, false);
	}
} // end class LyXmlNode
