/*
 * LyRequestEncoder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer;

import java.io.*;

import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.peer.util.*;


/**
 * This class is used to encode a request made on a data source and decode results to give a
 * response. This class is abstract and must be overriden to provide specific encoding and decoding.
 *
 * <p>The task of the request encoder is to process requests from a data source and to send
 * 'requests' using a connection instance. The request encoder class converts a data source demand
 * into a 'language' which is understood by the connection instance : the specific 'language' of the
 * underlying connection, like SQL, XML, method invocation on a Corba service.</p>
 *
 * <p>Additionally, the request encoder instance must 'decode' or interpret responses from the
 * specific connection to a 'language' understood by Leonardi. Basically, this class implements the
 * mapping between the chosen physical layer and the data representation in the specific connection.
 * </p>
 *
 * @see LyDataSource
 * @see LyResponse
 */
public abstract class LyBlobRequestEncoder extends LyRequestEncoder
{
	private static final String							CLOB_PEER_TYPE	= "CLOB".intern();
	private static final String							BLOB_PEER_TYPE	= "BLOB".intern();
	/** Amount of bytes to read in a single block for a blob (performance buffer). */
	public static int										MAX_BLOB_BUFFER = 250000;
	protected static String								BLOB_SEP = "";

	/**
	 * Constructs a request encoder. Initialization is done with the init method.
	 *
	 * @param provider   Provider that creates this request encoder.
	 * @param dataSource Data source client of the request.
	 */
	public LyBlobRequestEncoder(LyDataProvider provider, LyDataSource dataSource)
	{
		super(provider, dataSource);
	}

	/**
	 * Check if this field info is bound to a BLOB in DB.
	 *
	 * @param  field the field to check
	 * @return true if it is a BLOB
	 */
	protected boolean isBlob(LyColumnField field)
	{
		if (BLOB_PEER_TYPE == field.getPeerType())
			return true;
		return false;
	}

	/**
	 * Check if this field info is bound to a CLOB in DB.
	 *
	 * @param  field the field to check
	 * @return true if it is a CLOB
	 */
	protected boolean isClob(LyColumnField field)
	{
		if (CLOB_PEER_TYPE == field.getPeerType())
			return true;
		return false;
	}
	/**
	 * Returns the value to access the Blob/clob value and its length Must be a Reader (text) or an
	 * InputStream.
	 * @param field field associated to the blob value
	 * @param value the blob content file
	 * @return an array, at index 0 the InputStream or the Reader, at index 1 the size to read.
	 */
	protected Object[] getBlobInputStream(LyColumnField field, Object value)
	{
		if (isClob(field))
		{
			final String vStr = value.toString();

			return new Object[] { new StringReader(vStr), new Integer(vStr.length()), null };
		}
		else if (isBlob(field))
		{
			String vStr = value.toString();
			// Is it representing a File ?
			vStr = getEnvironment().getString().expand(vStr);

			File f = new File(vStr);

			if (!f.exists())
			{
				// try with tempDir first
				vStr = getEnvironment().getViewManager().getTemporaryDir() + File.separator + value.toString();
				vStr = getEnvironment().getString().expand(vStr);
				f = new File(vStr);
			}

			if (f.exists() && f.canRead() && f.isFile())
			{
				try
				{
					InputStream fin= new BufferedInputStream(new FileInputStream(f));

					// we have a file, likely a local file, and this is not like when getting the value from the db
					// we shall copy the local data to a temp file
					File	    ftemp;
					ftemp = createTempfile(LyFileTools.getFileName(vStr), null);

					writeDataToFile(fin, ftemp, 0, (int)f.length());
					fin.close();

					// reset
					fin = new BufferedInputStream(new FileInputStream(f));

					if ((!allowAdditionalInfoInBlob())
							|| ((isBlob(field) /* || isClob(field)*/)
							&& (field.getId().indexOf(BLOB_SEP) > 0)))
						return new Object[]
						{
							fin,
							new Integer((int)f.length()),
							ftemp.getAbsolutePath()
						};
					else
					{
						// save the value itself in the blob
						// BLOB_SEP+filename+BLOB_SEP+<binary content>
						byte[] fileNameBytes =
							(LyBlobRequestEncoder.BLOB_SEP
								+ LyFileTools.getFileName(vStr)
								+ LyBlobRequestEncoder.BLOB_SEP).getBytes("UTF-8");
						byte[] dataFull = new byte[(int)f.length() + fileNameBytes.length];

						for (int i = 0; i < fileNameBytes.length; i++)
							dataFull[i] = fileNameBytes[i];
						for (int i = 0, n = (int)f.length(); i < n; i++)
							dataFull[i + fileNameBytes.length] = (byte)fin.read();

						InputStream fin1 = new ByteArrayInputStream(dataFull);
						fin.close();
						return new Object[]
						{
							fin1,
							new Integer(dataFull.length),
							ftemp.getAbsolutePath()
						};
					}
				}
				catch (IOException e)
				{
					getEnvironment().writeException(e);
				}

			}

			// save the value itself in the blob
			byte[] blob = value.toString().getBytes();

			return new Object[] { new ByteArrayInputStream(blob), new Integer(blob.length), null };

		}
		return null;
		//return new StringReader(vStr);
	} // end method getBlobInputStream

	/**
	 * internal method to convert the return type of the getBlobInputStream methods to a byte array
	 * @param blobInfo
	 * @return
	 */
	protected byte[] getBytes(Object[] blobInfo)
	{
		if ((blobInfo == null))
			return null;

		try
		{
			final Integer size = (Integer)blobInfo[1];
			final byte[] result = new byte[size.intValue()];
			int   pos = 0;

			if (blobInfo[0] instanceof InputStream)
			{
				InputStream iStream = (InputStream)blobInfo[0];

				if (!(iStream instanceof BufferedInputStream))
					iStream = new BufferedInputStream(iStream);

				while (pos < size.intValue())
					result[pos++] = (byte)iStream.read();
			}
			else if (blobInfo[0] instanceof Reader)
			{
				throw new IllegalArgumentException("Can not safely convert Reader to bytes.");
			}
			else
				return null;
			return result;
		}
		catch (IOException  e)
		{
			getEnvironment().logError(e.getMessage());
			getEnvironment().writeException(e);
			return null;
		}
	}

	/**
	 * Internal method to get a temp file for a Blob.
	 *
	 * @param  fileNameStr current file name
	 * @return a new temlporary empty file
	 */
	protected File createTempfile(String fileNameStr, String suffix)
		throws IOException
	{
		String directory = null;

		if (getEnvironment().getViewManager() != null)
			directory = getEnvironment().getViewManager().getTemporaryDir();
		if (directory != null)
			directory = getEnvironment().getString().expand(directory);
		fileNameStr = LyFileTools.getFileName(fileNameStr);

		if (suffix == null)
			suffix = LyFileTools.getFileSuffix(fileNameStr);
		if ((suffix != null) && (!suffix.startsWith(".")))
			suffix = "."+suffix;
		if ((suffix != null) && (fileNameStr.endsWith(suffix)))
			fileNameStr = fileNameStr.substring(0, fileNameStr.length() - suffix.length());

		if (fileNameStr.length() < 3)
			fileNameStr = fileNameStr + "tmp";
		// we have got the file name, now we have to set the file content into a temporary file
		File dirFile = null;
		if (directory != null)
		{
			dirFile = new File(directory);
			if (!dirFile.exists())
				dirFile.mkdirs();
		}
		LyFileTools tools = new LyFileTools(getEnvironment());

		File f = tools.createTempFile(fileNameStr, suffix,
				((directory == null) ? null : new File(directory)), true);

		return f;
	}

	/**
	 * Internal method to get a temp file for a Blob.
	 *
	 * @param  iStream    stream to read
	 * @param  f          file to write
	 * @param  nBytesRead nb of bytes already read from the stream
	 * @param  length     length of the stream
	 * @return the new nBytesRead after the method
	 */
	protected int writeDataToFile(InputStream iStream, File f, int nBytesRead, int length)
		throws IOException
	{
		final OutputStream fOut = new BufferedOutputStream(new FileOutputStream(f));
		byte[]		 data;
		int			 newBytesRead;

		try
		{
			while (nBytesRead < length)
			{
				if ((length - nBytesRead) < MAX_BLOB_BUFFER)
					data = new byte[length - nBytesRead];
				else
					data = new byte[MAX_BLOB_BUFFER];
				newBytesRead = iStream.read(data);
				fOut.write(data, 0, newBytesRead);
				nBytesRead = nBytesRead + newBytesRead;
			}
		}
		catch (Exception e)
		{
			getEnvironment().logError("Error Reading BLOB to file "+f.getAbsolutePath()+" length ="+length);
			getEnvironment().logError(e.getMessage());
			getEnvironment().writeException(e);
		}
		fOut.close();
		return nBytesRead;
	}

	/**
	 * Indicates if the connector accepts the storage of addintional info in the blob (like file name)
	 */
	protected boolean allowAdditionalInfoInBlob()
	{
		return false;
	}

	/**
	 * Load the real blob content from the locator : return a physical locator like a file path
	 * @param object the object which helds the blob field and values
	 * @param fieldInfo blob fieldinfo to load
	 * @param locator value of the field, usually a locator for making object reference
	 * @return the physical locator
	 */
	public String getBlobContent(LyObject object, LyFieldInfo fieldInfo, String locator)
	{
		return locator;
	}
} // end class LyRequestEncoder
