/*
 * LyConnection.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer;

import java.rmi.server.*;
import java.util.*;

import leon.app.*;
import leon.misc.*;


/**
 * This class represents the physical link with a data supplier. It implements means to 'send
 * requests' and to 'receive responses' from the underlying physical connection.
 */
public abstract class LyConnection
{
	public static final String __VERSION =
		"$Id: LyConnection.java 35518 2012-05-16 15:39:46Z bangiole $";

	/** Application for which the connection is constructed. */
	protected LyApplication    _application;

	/** Last error message that occured on this connection. */
	protected String		   _error;

	/** Unique identifier of the transaction in which the connection is implied. */
	protected UID			   _uid;

	/** Data provider owner of this connection. */
	protected LyDataProvider   _provider;

	/** Boolean indicating if this connection is closed. */
	protected boolean		   _closed;

	/** Indicate which thread is currently using the connection. Null if no one use it. */
    protected Thread           _reserved;

	/** Date of reservation. */
	Date					   _date;

	/** Exception for keeping dump stack of reservation. */
	Exception				   _exception;

	/** Reservation thread. */
    protected String           _thread;

	/** Associations between keys and values used to evaluate contextual values. */
	protected Map<Object, Object> _values;

	/**
	 * Constructs a new connection for the given application.
	 *
	 * @param application Application for which the connection is constructed.
	 */
	public LyConnection(LyApplication application)
	{
		_application = application;

		init();
	}

	/**
	 * Initialization of the data of the instance.
	 */
	protected void init()
	{
		// Initialize class attributes.
		_error = null;
		_uid = null;
		_provider = null;
		_closed = false;
		_reserved = null;
		_thread = null;
	}

	/**
	 * Close connection : close associated physical connection.
	 */
	public boolean close()
	{
		_closed = true;
        _reserved = null;
        _exception = null;
		_thread = null;
		return true;
	}

	/**
	 * This method is called before a commit.
	 *
	 * @return true if the commi can be made, else returns false.
	 */
	public boolean prepareCommit()
	{
		return true;
	}

	/**
	 * To commit a pending transaction for this connection.
	 */
	public void commit()
	{
	}

	/**
	 * To rollback a pending transaction for this connection.
	 */
	public void rollback()
	{
	}

	/**
	 * Method called when the connection is reserved by the data provider. This method may be
	 * overidden to reserve some resources.
	 */
	public void reserved()
	{
		if ((_reserved != null) 
				&& (_reserved != Thread.currentThread()))
		{
			_application.getEnvironment().logWarning("Connection already reserved " + this
				+ " (thread : " + Thread.currentThread() + ")");
			Thread.dumpStack();
			dump();
		}

		_date = new Date();
		_exception = new Exception("Connection reserved");
		_reserved = Thread.currentThread();
		_thread = Thread.currentThread().toString();
	}

	/**
	 * Method called when the connection is returned to the data provider. This method may be
	 * overidden to free some resources.
	 */
	public void released()
	{
		if ((_thread != null)
			&& (_reserved == null))
		{
			_application.getEnvironment().logWarning("Connection already released " + this
				+ " (thread : " + Thread.currentThread() + ")");
			Thread.dumpStack();
			dump();
		}
		else if ((_reserved != null)
				&& (_reserved != Thread.currentThread()))
		{
			_application.getEnvironment().logError("Release of connection " + this
				+ " occurs in a different thread than reservation : "
				+ Thread.currentThread() + ")");
			Thread.dumpStack();
			dump();
		}

		_date = new Date();
		_exception = new Exception("Connection released");
		_reserved = null;
		_thread = Thread.currentThread().toString();
	}

	/**
	 * To set the commit mode for this connection. Default mode is auto commit.
	 */
	public void autoCommit(boolean status)
	{
	}

	/**
	 * Dump state of this connection.
	 */
	public void dump()
	{
        LyEnvironment environment = _application.getEnvironment();

		if (_reserved != null)
		    environment.logErr("Connection " + this + " reserved on " + _date
		    	+ ", thread : " + _thread);
		else
	        environment.logErr("Connection " + this + " released on " + _date
			    + ", thread : " + _thread);

	    if  (_exception != null)
	        environment.writeException(_exception);
	}

	/**
	 * Gets the last error message that occured on this connection.
	 *
	 * @return Last error message (may be null if no error occured).
	 */
	public String getError()
	{
		return _error;
	}

	/**
     * Checks close state.
	 */
	public boolean isClosed()
	{
		return _closed;
	}

    /**
     * Checks reserved state.
     */
	 public final boolean isReserved()
	{
	     return (_reserved == Thread.currentThread());
	}

	/**
	 * Gets the UID of this connection.
	 *
	 * @return The UID of this connection.
	 */
	public UID getUID()
	{
		if (_uid == null)
			_uid = new UID();
		return _uid;
	}

	/**
	 * Sets the last error message that occured on this connection.
	 *
	 * @param error Error message (null means no error).
	 */
	public void setError(String error)
	{
		_error = error;
		_application.setError(error);
	}

	/**
	 * Sets the UID of this connection.
	 *
	 * @param uid the UID to set
	 */
	public void setUID(UID uid)
	{
		_uid = uid;
	}

	/**
	 * Sets the data provider of this connection.
	 *
	 * @param provider the data provider of this connection.
	 */
	public void setDataProvider(LyDataProvider provider)
	{
		_provider = provider;
	}

	/**
	 * Adds a pair key-value to the session. May be used for example to store the login used to
	 * create the session.
	 *
	 * @param  key    The key that must be stored.
	 * @param  value  The value that must be stored with the key.
	 *
	 * @see    #getValue
	 */
	public void addValue(Object key, Object value)
	{
		if (_values == null)
			_values = new HashMap<Object, Object>();

		_values.put(key, value);
	}

	/**
	 * Retrieves a value from a registered key.
	 *
	 * @param   key  The key that must be used.
	 *
	 * @return  The value stored with the key or null if the key is not registered.
	 *
	 * @see     #addValue
	 */
	public Object getValue(Object key)
	{
		if ((_values == null) || (key == null))
			return null;

		Object result = _values.get(key);

		return result;
	}

} // end class LyConnection
