/*
 * LyDataProvider.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer;

import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.data.*;
import leon.info.*;
import leon.info.location.*;
import leon.info.notifier.*;
import leon.misc.*;
import leon.notifier.*;


/**
 * This class is used by DataSources to transmit their requests to an instance of RequestEncoder.
 *
 * <p>The LyDataProvider class implements the connector. Every instance of LyDataSource is
 * associated to an instance of LyDataProvider.</p>
 *
 * <p>The services of the LyDataProvider classes are :<br>
 * - Manage instances of the LyConnection (connection to the information system : RDBMS, files,
 * application server, corba connection ...).<br>
 * - Create the description of physical data in memory and make the link with the logical layer.<br>
 * - Create instances of LyRequestEncoder, associated to the LyDataSource instances : the encoder
 * implements the communication between the dataSource and the connection. (Request encoding,
 * connection service calls ...)<br>
 * - Accessors to valid LyConnection instances for the LyRequestEncoder instances.</p>
 *
 * <p>LyDataProvider implements a connection pool to manage multiple connections as a connection
 * stack. The number of connection in the pool is a parameter (default is 1 single connection).</p>
 *
 * <p>Remark : current implementation only uses a number of connection pool for SQL data providers,
 * or programmatically in the code of a specific data provider.</p>
 *
 * @see    LyRequestEncoder
 * @see    LyConnection
 * @author Lyria
 */
public abstract class LyDataProvider
{
	public static final String __VERSION =
		"$Id: LyDataProvider.java 34648 2012-03-14 11:12:14Z jlsaouli $";

	/** Reference to the application instance. */
	protected LyApplication           _application;

	/** Reference to the Location instance associated to this connector in the meta model. */
	protected LyLocation	          _location;

	/** Identifier of this provider (obtained from the meta model, in LyLocation). */
	protected String		          _identifier;

	/**
	 * Internal connection pool to enable multiple simultaneous accesses to the connection servcice.
	 */
	protected _LyConnectionPool       _connections;

	/** Number of connections available in the connection pool. */
	protected int			          _nbConnection;

	/**
	 * Notification service added to this data-provider. A notification service is used to post /
	 * receive data changes such as creation / deletion / modification. (For example JMS, Bus...)
	 * This may be null if no notification service is required.
	 */
	protected LyEventNotifier         _eventNotifier;

	/** Boolean that indicates if connection is ready or not. */
	protected boolean		          _connectionReady;

	/** Boolean that indicates if connection was lost or not. */
	protected boolean		          _lostConnection;
    
    /** Modified class infos and projects. */
    protected ArrayList<LyClass>      _modifiedClasses = null;

	/**
	 * Constructs a data provider. Initialization is done with the init method.
	 *
	 * @param application Current application instance
	 * @param location    This value is obtained from the location parameter in the meta model.
	 */
	public LyDataProvider(LyApplication application, LyLocation location)
	{
		_application = application;

		init();
		setLocation(location);
	}

	/**
	 * Constructs a data provider. Initialization is done with the init method.
	 *
	 * @param      application Current application instance
	 * @param      id          Identifier of this dataProvider. This value is obtained from the
	 *                         location parameter in the meta model.
	 * @deprecated should use LyDataProvider(LyApplication application, LyLocation location) instead
	 *             of this constructor.
	 */
	public LyDataProvider(LyApplication application, String id)
	{
		_application = application;
		_identifier = id;

		init();
	}

	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		_connections = null;
		_nbConnection = 1;
		_connectionReady = false;
		_lostConnection = false;
        _modifiedClasses = null;
	}

	/**
	 * Tests if the data connection is ready or not.
	 *
	 * @return true if the data connection is ready, false otherwise.
	 */
	public boolean isConnectionReady()
	{
		return _connectionReady && !_lostConnection;
	}

	/**
	 * Tests if the notifier is ready or not.
	 *
	 * @return true if notifier is ready, false otherwise.
	 */
	public boolean isNotifierReady()
	{
		return (_eventNotifier != null) && _eventNotifier.isReady();
	}

	/**
	 * Tests if the connection with the physical layer has been lost or not.
	 *
	 * @return true if the connection with the physical layer has been detected lost, false
	 *         otherwise.
	 */
	public boolean hasLostConnection()
	{
		return _lostConnection;
	}

	/**
	 * Sets location on provider.
	 *
	 * @param location New location for the provider.
	 */
	public void setLocation(LyLocation location)
	{
		boolean result = true;

		if (_location == null)
		{
			_location = location;
			_identifier = location.getId();
		}
		else if (location == _location)
		{
			sleep();
			result = wakeUp();
		}
		else
		{
			sleep();

			LyNotifier oldNotifier = _location.getNotifier();
			LyNotifier notifier = location.getNotifier();

			if (oldNotifier != notifier)
			{
				// Check if we have a notifier associated to this provider
				LyEnvironment   environment = _application.getEnvironment();
				LyEventNotifier eventNotifier = null;

				if ((!environment.testEnv("LY_USE_FILES"))
						&& (notifier != null))
				{
					// Create the notifier instance or get it if it already exists
					environment.logInfo("BF AppEngine : Found notifier for "
						+ location + " -> " + notifier);
					eventNotifier = LyNotifierFactory.getNotifier(_application, notifier);
				}

				setEventNotifier(eventNotifier);
			}

			_location = location;
			result = wakeUp();
		}

		if (!result)
			throw new IllegalStateException("BF AppEngine : failed to restart provider " + this
				+ " with given location " + location);
	} // end method setLocation

	/**
	 * Accessor to the application instance.
	 *
	 * @return the reference to the application
	 */
	public final LyApplication getApplication()
	{
		return _application;
	}

	/**
	 * Accessor to the location of the provider.
	 *
	 * @return Location of this provider.
	 */
	public final LyLocation getLocation()
	{
		return _location;
	}

	/**
	 * Accessor to the id of the provider.
	 *
	 * @return Id of this provider.
	 */
	public final String getId()
	{
		return _identifier;
	}

	/**
	 * Accessor to the environment instance.
	 *
	 * @return the reference to the environment
	 */
	public final LyEnvironment getEnvironment()
	{
		return _application.getEnvironment();
	}

	/**
	 * Accessor to the notification service added to this data-provider. A notification service is
	 * used to post / receive data changes such as creation / deletion / modification. (For example
	 * JMS, Bus...) This may be null if no notification service is required.
	 *
	 * @return The notification service associated to this data provider.
	 */
	public final LyEventNotifier getEventNotifier()
	{
		return _eventNotifier;
	}

	/**
	 * Accessor to the max number of connections available in the connection pool.
	 *
	 * @return number of connection available. If <= 1 , one connection should be used.
	 */
	public final int getNbConnection()
	{
        return _nbConnection;
	}

	/**
	 * Sets the number of connection available in the connection pool.
	 *
	 * @param nbConnection number of connection available. If <= 1 , one connection is used.
	 */
	public void setNbConnection(int nbConnection)
	{
		_nbConnection = nbConnection;
	}

	/**
     * Accessor to the number of connections available in the connection pool.
     *
     * @return number of connection available. If <= 1 , one connection should be used.
     */
    public final int getNbAvailableConnection()
    {
        return (_connections == null ? 0 : _connections.getSize());
    }
    
	/**
	 * Modifier of the notification service added to this data-provider. A notification service is
	 * used to post / receive data changes such as creation / deletion / modification. (For example
	 * JMS, Bus...) This may be null if no notification service is required.
	 *
	 * @param notifier the notification service to associate with this data provider.
	 */
	public void setEventNotifier(LyEventNotifier notifier)
	{
		_eventNotifier = notifier;
		if (_eventNotifier != null)
			_eventNotifier.setLocationId(_identifier);
	}

	/**
	 * This method is used to build all the class peers for all the classInfo.
	 *
	 * @return true if the valaitation was successful.
	 */
	public boolean validate()
	{
		LyInfoList<LyClassInfo> classInfos = _application.getClassInfos();

		return validate(classInfos);
	}

	/**
	 * This method is used to build all the class peers for all the classInfo. It must be
	 * specialized by subclasses. Basically, this method associates the dataProvider with local
	 * classes. A class is local if there is no binding.
	 *
	 * @param  classInfos Classes to validate.
	 * @return true if the validation was successful.
	 */
	public boolean validate(LyInfoList<LyClassInfo> classInfos)
	{
		// Nothing to do
		return true;
	}

	/**
	 * Validate the associated event notifier if exists.
	 *
	 * @param  classInfos Classes to validate.
	 * @return true if the validation was successful.
	 */
	public boolean validateNotifier(LyInfoList<LyClassInfo> classInfos)
	{
		// Validate notifier if any
		if (_eventNotifier != null)
			_eventNotifier.validate(classInfos);

		return true;
	}

	/**
	 * This method is called when the application is going on sleeping mode. Default implementation
	 * call close method and set the notifier into sleeping mode. (untill a further call to wakeUp()
	 *
	 * <P>Application will restart after calling wakeUp()</P>
	 *
	 * @see #close()
	 * @see #wakeUp()
	 */
	public void sleep()
	{
		close();

		// Notifier is also going on sleeping mode
		if (_eventNotifier != null)
			_eventNotifier.sleep();
	}

	/**
	 * This method is called when the application is going out of sleeping mode Default
	 * implementation call create the connection pool and wakes up the notifier.
	 *
	 * @see #sleep()
	 */
	public boolean wakeUp()
	{
		if (!createConnections(getNbConnection()))
			return false;

		// Notifier should also end sleeping mode
		if (_eventNotifier == null)
			return true;
		else
			return _eventNotifier.wakeUp();
	}

	/**
	 * Add the specified number of connection available. If this number is lower or equal to 1, one
	 * connection is used.
	 *
	 * @param  nbConnection Number of connections in the pool
	 * @return true if all connections have been created, false otherwise.
	 */
	private boolean createConnections(int nbConnection)
	{
		if (_connections == null)
			_connections = new _LyConnectionPool();

		LyConnection  connection;
		LyEnvironment environment = getEnvironment();

		if (nbConnection <= 0)
		{
			if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
				environment.logOut("Attempt to create no connection : setting to 1");

			nbConnection = 1;
		}

		if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
			environment.logOut("Creating a pool of " + nbConnection
				+ " connection(s) for provider " + this);

		updateConnectionParameters(_location);

		for (int i = 0; i < nbConnection; i++)
		{
			connection = createConnection();
			if (connection == null)
				return false;

			_connections.returnConnection(connection);
		}

		_connectionReady = true;
		_lostConnection = false;
		return true;
	} // end method createConnections

	/**
	 * Updates this connection parameters using location information and application environment.
	 */
	protected void updateConnectionParameters(LyLocation location)
	{
	}

	/**
	 * Closes all connections from the connection pool.
	 *
	 * @return true if the request completed successfully.
	 * @see    #createConnections(int)
	 */
	public boolean close()
	{
		LyEnvironment environment = getEnvironment();

		if (environment.testEnv("LY_SAVE_FILES")
				&& _application.isValid()
				&& _connectionReady)
			save(true);

		if (_connections != null)
		{
			// Close all the connections in the pool.
			LyConnection connection;

			synchronized (this)
			{
				while (!_connections.isEmpty())
				{
					connection = _connections.getConnection();
					connection.close();
				}
				
				_connectionReady = false;
				_connections = null;
			}
		}

		// Notifier should also be closed
		if (_eventNotifier != null)
			_eventNotifier.close();

		_lostConnection = true;
		return true;
	} // end method close

	/**
	 * To save data managed by this provider.
	 *
	 * @return true if all is all right, else false.
	 */
	public final boolean save()
	{
        if ((_modifiedClasses != null)
                && (_modifiedClasses.size() > 0))
            return save(false);
        else
            return true;
	}

	/**
	 * To save data managed by this provider.
	 *
	 * @param  loadEmptyCaches Indicated if empty caches must be loaded or not.
	 * @return true if all is all right, else false.
	 */
	protected boolean save(boolean loadEmptyCaches)
	{
	    setModified(false, null, null);
		return true;
	}

	/**
	 * This class is a factory to create a LyConnection instance. Current implementation only return
	 * null, so this method must be overriden. This instance depends on the nature of the connection
	 * and must provide services such as opening/closing the connection, and (specific) accessor to
	 * the connection service.
	 *
	 * <P>This method is ovberriden and used internally. Use getConnection() to access to a
	 * connection instance.</P>
	 *
	 * <P>Parent class LyDataProvider implements a connection pool so this Method may be called
	 * several times.</P>
	 *
	 * @return a new Connection instance. The connection instance is used in the specific LyRequest
	 *         encoder implementation.
	 * @see    #getConnection
	 * @see    #releaseConnection
	 */
	protected abstract LyConnection createConnection();

	/**
	 * Accessor to a connection instance from the connection pool.
	 *
	 * <P>If no connection is available currently, this method will block until a connection is
	 * released and available.</P>
	 *
	 * <P><b>Important !</b> The connection MUST be released after usage so it's available for the
	 * connection pool. Invoke releaseConnection when the connection is no longer needed.</P>
	 *
	 * @return a connection instance from the connection pool. The connection instance must be
	 *         released (releaseConnection) after usage.
	 * @see    #releaseConnection
	 */
	public final LyConnection getConnection()
	{
		LyTransactionManager tm = _application.getTransactionManager();

		return tm.getConnection(this);
	}

	/**
	 * Accessor to a connection instance from the connection pool. For internal use only.
	 *
	 * @param  b not used
	 * @return a connection from the pool.
	 */
	protected LyConnection getConnection(boolean b)
	{
		if (_lostConnection)
			return null;

		// Avoid conflict access to connections
		LyConnection connection = null;

		synchronized (this)
		{
			if (_connections == null)
				createConnections(_nbConnection);

			connection = _connections.getConnection();
		}

		if (connection == null)
			return null;

		connection.setError(null);
		connection.setDataProvider(this);
		connection.autoCommit(true);

		if (connection.isClosed())
			_connectionReady = false;
		if (!_lostConnection
				&& connection.isClosed()
			/*&& (_eventNotifier == null)*/)
		{
			// Can not use this connection : back inpool and close

			// Try to reconnect automatically
			sleep();
			wakeUp();

			if (_lostConnection)
			{
				LyLocation						location = getLocation();
				LyDataSession					session = _application.getCurrentSession();
				LyApplicationBehaviorInterface	behavior = _application.getApplicationBehavior();
				if (session != null)
				{
					LyDataSessionBehaviorInterface sessionBehavior = session.getSessionBehavior();
					sessionBehavior.connectionLost(session, location);
				}
			
				behavior.connectionLost(_application, location);
				return null;
			}
			else
				return getConnection(b);
		}

		return connection;
	} // end method getConnection

	/**
	 * Return a connection instance to the connection pool.
	 *
	 * <P>The connection reference was obtained by getConnection()</P>
	 *
	 * <P><b>Important !</b> The connection MUST be released after usage so it's available for the
	 * connection pool. Invoke releaseConnection when the connection is no longer needed.</P>
	 *
	 * @param connection a connection instance from the connection pool.
	 * @see   #releaseConnection
	 */
	public final void releaseConnection(LyConnection connection)
	{
		LyTransactionManager tm = _application.getTransactionManager();

		tm.releaseConnection(connection, this);
	}

	/**
	 * Return a connection instance to the connection pool. For internal use only.
	 *
	 * @param connection the conection to be returned to the pool
	 * @param t          not used
	 */
	protected void releaseConnection(LyConnection connection, boolean t)
	{
		if (_connections == null)
			return;

		if (connection != null)
		{
			if (!connection.isClosed())
				_connections.returnConnection(connection);
			else if (!_lostConnection)
			{
				_lostConnection = true;

				LyLocation						location = getLocation();
				LyDataSession					session = _application.getCurrentSession();
				LyApplicationBehaviorInterface	behavior = _application.getApplicationBehavior();
				
				if (session != null)
				{
					LyDataSessionBehaviorInterface sessionBehavior = session.getSessionBehavior();
					sessionBehavior.connectionLost(session, location);
				}
			
				behavior.connectionLost(_application, location);
			}
		}
	}

	/**
	 * Create a specific request encoder for the specified dataSource.
	 *
	 * <P>This method is used internally, but subclasses should override this method and create an
	 * instance of their own (specific) request encoder.</P>
	 *
	 * <P>Default returns a LyRequestEncoder instance.</P>
	 *
	 * @param  dataSource reference to the dataSource associated to this request encoder.
	 * @return the specific LyRequestEncoder for the specified dataSource.
	 */
	public abstract LyRequestEncoder getRequestEncoder(LyDataSource dataSource);

	/**
	 * Tests if this data provider supports the XA protocol or not. Subclasses may override this
	 * method. The default implementation is to return false.
	 *
	 * @return true if this provider supports the XA protocol, false otherwise.
	 */
	public boolean isXASupported()
	{
		return false;
	}

	/**
	 * Indicates if this data provider can be fully loaded.
	 * @return true if the provider can load all objects. False otherwise.
	 */
	public boolean canBeFullyLoaded(LyClass aClass)
	{
		return true;
	}
	
	/**
	 * A class which manages a connection pool.
	 */
	protected class _LyConnectionPool
	{
		Stack<LyConnection>  _pool;
		Vector<LyConnection> _current;
		LyEnvironment		 _environment;
		int					 _timeout;
		int					 _nbWaits;

		_LyConnectionPool()
		{
			_pool = new Stack<LyConnection>();
			_current = new Vector<LyConnection>(1);
			_environment = null;
			_nbWaits = 0;
			_timeout = -1;
		}

		synchronized LyConnection getConnection()
		{
			if (_pool.isEmpty() && _current.isEmpty())
				return null;

			if (_environment == null)
			{
				_environment = getEnvironment();

				String str = _environment.getResource("LY_CONNECTION_TIMEOUT");

				if (str != null)
				{
					_timeout = LyTools.intFromString(str);
					if (_timeout < 0)
						_timeout = 0;
				}
			}

			if (_environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
				_environment.logOut("Search for a connection...");

			// Wait for a valid connection
			// The thread enter the wait if some threads are still waiting
			// Avoids a  thread to steal the connection to a waiting one
			if (isEmpty() || (_nbWaits > 0))
			{
				_nbWaits++;
				if (_environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
					_environment.logOut("Waiting for a valid connection... ("
						+ _nbWaits + " waiting threads, timeout "
						+ _timeout + " ms)");
				try
				{
					// Mise en attente.
					if (_timeout > 0)
						this.wait(_timeout);
					else
						this.wait();
				}
				catch (Exception e)
				{
					// Error
					_environment.writeException(e);
				}
				finally
				{
					_nbWaits--;
				}

				if (isEmpty())
				{
					if (_timeout > 0)
						_environment.logErr("BF AppEngine : can't get connection, timeout expired");
					else
						_environment.logErr("BF AppEngine : can't get connection");
					dump();

					if (_timeout > 0)
						throw new IllegalStateException(
							"BF AppEngine : can't get connection, timeout expired");
					else
						throw new IllegalStateException("BF AppEngine : can't get connection");
				}
			}

			// Get the first connection.
			LyConnection connection = _pool.pop();

			if (_environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
				_environment.logOut("Connection found : " + connection);

			// Reserve connection
			connection.reserved();
			_current.add(connection);
			
	        if (_environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
            {
                _environment.logOut("[CONNECTION] Connection reserved : " + connection
                	+ " (thread " + Thread.currentThread() + "), "
                    + _current.size() + " in use, " + _pool.size() + " free.");
                Thread.dumpStack();
            }

			// End of waiting.
			return connection;
		} // end method getConnection

		synchronized void returnConnection(LyConnection connection)
		{
			if (_pool.contains(connection))
			{
				if (_environment.isDebugLevel(LyEnvironment.DEBUG_WARNING))
				{
					_environment.logWarning("Connection is already in the pool : " + connection);
					Thread.dumpStack();
					
					connection.dump();
				}

				return;
			}

			// Release connection
			_current.remove(connection);
			connection.released();
			_pool.push(connection);
			
            if ((_environment != null) // Init
                && _environment.isDebugLevel(LyEnvironment.DEBUG_TRANSACTION))
            {
                _environment.logOut("[CONNECTION] Connection released : " + connection
                	+ " (thread " + Thread.currentThread() + "), "
                    + _current.size() + " in use, " + _pool.size() + " free.");
                Thread.dumpStack();
            }

			// End of waiting.
			this.notify();
		}

		synchronized boolean isEmpty()
		{
			return _pool.isEmpty();
		}

        synchronized int getSize()
        {
            return _pool.size();
        }
	      
		protected void dump()
		{
			int			 i, nc = _current.size(), np = _pool.size();
			LyConnection connection;

			_environment.logErr("Dumping all connections for provider " + this
				+ ", " + nc + " connections in use, " + np + " free connections");

			for (i = 0; i < nc; i++)
			{
				connection = _current.elementAt(i);
				connection.dump();
			}
			for (i = 0; i < np; i++)
			{
				connection = _pool.elementAt(i);
				connection.dump();
			}
		}
	} // end class _LyConnectionPool

	public void dumpConnections()
	{
		_connections.dump();
	}
    
    /**
     * Sets the given class as modified on this provider.
     * @param modified      True if modified, false otherwise.
     * @param modifiedClass Class that was modified.
     * @param object        Modified instance(s) (may be null).
     */
    public void setModified(boolean modified, LyClass modifiedClass, Object object)
    {      
        if (modified)
        {
            if (modifiedClass == null)
                return;
            
            if (_modifiedClasses == null)
                _modifiedClasses = new ArrayList<LyClass>();
            else if (_modifiedClasses.contains(modifiedClass))
                return;
            
            _modifiedClasses.add(modifiedClass);
        }
        else
        {
            if (modifiedClass == null)
                _modifiedClasses = null;
            else if ((_modifiedClasses != null)
                     && _modifiedClasses.contains(modifiedClass))
                _modifiedClasses.remove(modifiedClass);
        }
    }
    
    /**
     * Checks if this provider was modified.
     * @return True if modified, false otherwise.
     */
    public boolean isModified()
    {
        return ((_modifiedClasses != null)
            && (_modifiedClasses.size() > 0));
    }
    
    /**
     * Checks if the given class was modified for this provider.
     * @param modifiedClass Class that is checked.
     * @return True if modified, false otherwise.
     */
    public boolean isModified(LyClass modifiedClass)
    {
        return ((_modifiedClasses != null)
            && (_modifiedClasses.contains(modifiedClass)));
    }
} // end class LyDataProvider
