/*
 * LyDataProviderFactory.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer;

import java.lang.reflect.*;

import leon.app.*;
import leon.info.location.*;
import leon.misc.*;
import leon.notifier.*;
import leon.peer.file.*;


/**
 * This class is used by LyApplication to create instances of LyDataProvider defined by LyLocation.
 *
 * <p>Existing providers are the following :</p>
 *
 * <table border='1'>
 * <tr>
 * <td>XML</td>
 * <td>leon.peer.xml.file.LyXmlFileDataProvider</td>
 * </tr>
 * <tr>
 * <td>CSV</td>
 * <td>leon.peer.file.LyFileDataProvider</td>
 * </tr>
 * <tr>
 * <td>INFO</td>
 * <td>leon.peer.file.LyFileInfoDataProvider</td>
 * </tr>
 * <tr>
 * <td>LDAP</td>
 * <td>eu.w4.ae.peer.ldap.LyLdapDataProvider</td>
 * </tr>
 * <tr>
 * <td>SQL</td>
 * <td>
 * <table border='1'>
 * <tr>
 * <td>Access</td>
 * <td>leon.peer.sql.access.LyAccessDataProvider</td>
 * </tr>
 * <tr>
 * <td>MySql</td>
 * <td>leon.peer.sql.mysql.LyMySqlDataProvider</td>
 * </tr>
 * <tr>
 * <td>Oracle</td>
 * <td>eu.w4.ae.peer.sql.oracle.LyOracleDataProvider</td>
 * </tr>
 * <tr>
 * <td>Sybase</td>
 * <td>eu.w4.ae.peer.sql.sybase.LySybaseDataProvider</td>
 * </tr>
 * <tr>
 * <td>generic</td>
 * <td>leon.peer.sql.LySqlDataProvider</td>
 * </tr>
 * </table>
 * </td>
 * </tr>
 * </table>
 *
 * @see  leon.app.LyApplication
 * @see  leon.info.location.LyLocation
 */
public final class LyDataProviderFactory
{
    public static final String __VERSION = "$Id: LyDataProviderFactory.java 35251 2012-04-17 15:32:36Z jlsaouli $";

    /**
     * Initialisation of the data of the instance.
     */
    public static LyDataProvider getProvider(LyApplication application, LyLocation location)
    {
        LyDataProvider provider = null;
        LyEnvironment  environment = application.getEnvironment();
        String         javaClass = null;

        if (environment.testEnv("LY_USE_FILES"))
            provider = new LyFileDataProvider(application,
                    new LyFileLocation(location.getId(), "LY_DATA_DIR", LyFileLocation.Type.INFO, null, null));
        else if (location instanceof LyFileLocation)
        {
        	LyFileLocation.Type type = ((LyFileLocation)location).getType();
            
            if (type == LyFileLocation.Type.XML)
                javaClass = "leon.peer.xml.file.LyXmlFileDataProvider";
            else if (type == LyFileLocation.Type.CSV)
                javaClass = "leon.peer.file.LyFileDataProvider";
            else if (type == LyFileLocation.Type.INFO)
                javaClass = "leon.peer.file.LyFileInfoDataProvider";
        }
        else if (location instanceof LySqlLocation)
        {
            // Check if the parameter is specified
            javaClass = ((LySqlLocation)location).getJavaName();

            // try to get from the driver
            if (javaClass == null)
            {
                String driver = ((LySqlLocation)location).getDriver();
                String url = ((LySqlLocation)location).getUrl();

                if (driver.charAt(0) == '$')
                {
                    String d = environment.getEnv(driver);

                    if (d == null)
                    {
                        String err = application.getEnvironment().getMessage("LY_ER_INIT_PARAM", driver);

                        application.getEnvironment().logErr(err);
                    }
                    else
                        driver = d;
                }

                if (url.charAt(0) == '$')
                {
                    String d = environment.getEnv(url);

                    if (d == null)
                    {
                        String err = application.getEnvironment().getMessage("LY_ER_INIT_PARAM", url);

                        application.getEnvironment().logErr(err);
                    }
                    else
                        url = d;
                }

                if (driver.indexOf("mysql") != -1)
                    javaClass = "leon.peer.sql.mysql.LyMySqlDataProvider";
                else if (driver.indexOf("oracle") != -1)
                    javaClass = "eu.w4.ae.peer.sql.oracle.LyOracleDataProvider";
                else if (driver.indexOf("odbc") != -1)
                    javaClass = "leon.peer.sql.access.LyAccessDataProvider";
                else if (driver.indexOf("sybase") != -1)
                    javaClass = "eu.w4.ae.peer.sql.sybase.LySybaseDataProvider";
                else if ((driver.indexOf("sqlserver") != -1)
                        || ((url.indexOf("sqlserver") != -1)
                            && (driver.indexOf("jtds") != -1)))
                    javaClass = "eu.w4.ae.peer.sql.sqlserver.LySqlServerDataProvider";
                else if (driver.indexOf("instantdb") != -1)
                    javaClass = "leon.peer.sql.idb.LyIdbDataProvider";
                else if (driver.indexOf("postgresql") != -1)
                    javaClass = "leon.peer.sql.postgresql.LyPostgresqlDataProvider";
                else if (driver.indexOf("progress") != -1)
                    javaClass = "eu.w4.ae.peer.sql.progress.LyProgressDataProvider";
                else if (driver.indexOf("hsqldb") != -1)
                    javaClass = "leon.peer.sql.hsqldb.LyHsqldbDataProvider";
                else
                    // Use generic SQL data provider
                    javaClass = "leon.peer.sql.LySqlDataProvider";
            }
        }
        else if (location instanceof LyGenLocation)
        {
        	javaClass = ((LyGenLocation)location).getJavaName();
        	
        	
        }
        else if (location instanceof LyLdapLocation)
            javaClass = "eu.w4.ae.peer.ldap.LyLdapDataProvider";

        if (javaClass != null)
        {
            if (javaClass.startsWith("$"))
                javaClass = environment.getEnv(javaClass);

            // For compatibility reason, translate old leonardi packages to w4 packages
            if ((javaClass.indexOf("leon.peer.w4") >=0)
            		|| (javaClass.indexOf("leon.peer.hibernate") >=0 ) 
            		|| (javaClass.indexOf("leon.peer.cms") >=0 )
            		|| (javaClass.indexOf("leon.peer.domino") >=0 )
            		|| (javaClass.indexOf("leon.peer.webservices") >=0 )
            		|| (javaClass.indexOf("leon.peer.sql.oracle") >=0 )
            		|| (javaClass.indexOf("leon.peer.sql.progress") >=0 )
            		|| (javaClass.indexOf("leon.peer.sql.sqlserver") >=0 )
            		|| (javaClass.indexOf("leon.peer.sql.sybase") >=0 )
            		|| (javaClass.indexOf("leon.peer.bfstore") >=0 ))
            {
            	javaClass = javaClass.replace("leon.peer", "eu.w4.ae.peer");	
            }
            
            Class<?>[] types = new Class[2];
            Object[]   parameters = new Object[2];

            types[0] = LyApplication.class;
            parameters[0] = application;

            types[1] = LyLocation.class;
            parameters[1] = location;

            // Build provider
            Class<?>       providerClass;
            Constructor<?> constructor;

            providerClass = application.loadClass(javaClass,
                    "BF AppEngine : unknown data provider class " + javaClass
                    + " in LyDataProviderFactory");
            if (providerClass == null)
                return null;

            try
            {
                constructor = providerClass.getConstructor(types);
                provider = (LyDataProvider)constructor.newInstance(parameters);
            }
            catch (InvocationTargetException e)
            {
                environment.writeException(e.getTargetException());
                return null;
            }
            catch (NoSuchMethodException e)
            {
                // Unknown controller
                environment.logError("BF AppEngine : constructor with parameters ("
                    + types[0] + ", " + types[1] + ") not found on class " + providerClass);
                environment.writeException(e);
                return null;
            }
            catch (Exception e)
            {
                environment.writeException(e);
                return null;
            }
        }

        // Check if we have a notifier associated to this provider
        if ((!environment.testEnv("LY_USE_FILES")) && (location.getNotifier() != null))
        {
            environment.logInfo("Found notifier for " + location + " -> " + location.getNotifier());

            // Create the notifier instance or get it if it already exists
            LyEventNotifier eventNotifier = LyNotifierFactory.getNotifier(application, location.getNotifier());

            provider.setEventNotifier(eventNotifier);
        }

        return provider;
    } // end method getProvider
} // end class LyDataProviderFactory
