/*
 * LyPeerClass.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer;

import java.lang.reflect.Array;
import java.util.Hashtable;
import java.util.Vector;

import leon.app.LyApplication;
import leon.data.LyClass;
import leon.data.LyDataSource;
import leon.data.LyField;
import leon.data.LyObjectId;
import leon.data.LyValue;
import leon.data.LyValueNotLoadedException;
import leon.data.LyValueSet;
import leon.misc.LyEnvironment;


/**
 * This class offers a default abstract implementation of LyClass interface for physical classes.
 *
 * <p>Using this class, 3 methods remain abstract :</p>
 *
 * <ul>
 *   <LI>Method newInstance : to create a new instance of this class,</LI>
 *   <LI>Method get : defines how to get a value for a field for a given object instance,</LI>
 *   <LI>Method set : defines how to set a value for a field on a given object instance.</LI>
 * </ul>
 *
 * @see LyPeerField
 */
public abstract class LyPeerClass
	implements LyClass
{
	public static final String			  __VERSION =
		"$Id: LyPeerClass.java 32456 2011-09-22 07:53:53Z jlsaouli $";

	/** Used to find unloaded values in the physical layer. */
	public static final Object			  VALUE_NOT_LOADED = "__Not_loaded__".intern();

	/** An unique id for the application (location + name). */
	protected String					  _id;

	/** Name of the class. */
	protected String					  _name;

	/** Java class used for instances. */
	protected Class<?>					  _objectClass;

	/** List of the peer fields. */
	protected LyPeerField[]			      _fields;

	/** Hastable of fields indexes to optimize accesses. */
	protected Hashtable<LyField, Integer> _indexes;

	/** List of id fields. */
	protected LyPeerField[]        		  _idFields;

	/** Data source used to managed objects of this class. */
	protected LyDataSource				  _dataSource;

	/**
	 * Constructs a peer class with no parameter. This class must be completed by using accessors.
	 */
	public LyPeerClass()
	{
		_id = null;
		_name = null;
		_objectClass = null;
		_fields = new LyPeerField[0];
		_indexes = new Hashtable<LyField, Integer>(1);
		_idFields = new LyPeerField[0];
		_dataSource = null;
	}

	/**
	 * Constructs a peer class with a given id and an objects class.
	 *
	 * @param id          Identifier of this class.
	 * @param objectClass Java class used for "peer objects".
	 */
	public LyPeerClass(String id, Class<?> objectClass)
	{
		this();

		_id = id;
		_objectClass = objectClass;
	}

	/**
	 * Gets the value corresponding to the given peer field for an "instance" of this class. This
	 * method is called by getValue method that performs controls on given parameters so these
	 * controls don't need to be done twice.
	 *
	 * @param  object Instance of this class.
	 * @param  field  Peer field for which a value is requested.
	 * @return Value of the given field on the specified instance.
	 * @see    LyPeerClass#getValue
	 */
	protected abstract Object get(Object object, LyPeerField field);

	/**
	 * Sets the value corresponding to the given peer field on an "instance" of this class. This
	 * method is called by setValue method that performs controls on given parameters so these
	 * controls don't need to be done twice.
	 *
	 * @param object Instance of this class on which a value is set.
	 * @param field  Field for which a value is set.
	 * @param value  Value set for the given field on the specified instance.
	 * @see   LyPeerClass#setValue
	 */
	protected abstract void set(Object object, LyPeerField field, Object value);

	/**
	 * Retrieves the object id for a given object (previously registered by setId method). This
	 * method is called by getObjectId method that has the capacity to calculate the object id if
	 * this method returns null, this treatment doesn't need to be reproduced. The only purpose of
	 * this call is to retrieve the object id if it was registered on the object or in a
	 * hashtable... The default implementation provided by this method returns null.
	 *
	 * @param  object Instance of this class.
	 * @return Object identifier of the given object.
	 * @see    LyPeerClass#setId
	 */
	protected LyObjectId getId(Object object)
	{
		return null;
	}

	/**
	 * Sets the object id on the given object. This method associate the object id to the object if
	 * possible to allow to getId method to retrieve it in a fast way. The idea if it's possible to
	 * to register the id on the given object. The default implementation does nothing.
	 *
	 * @param object   Instance of this class.
	 * @param objectId Object identifier of this instance.
	 * @see   LyPeerClass#getId
	 */
	protected void setId(Object object, LyObjectId objectId)
	{
	}

	/**
	 * To validate the new values (already known) and add new extra values. if needed. The default
	 * implementation does nothing.
	 *
	 * @param  object Instance of this class.
	 * @param  values Validated values.
	 * @return A values set containing values that
	 */
	public LyValueSet validate(Object object, LyValueSet values)
	{
		return values;
	}

	/**
	 * Cancels current modifications (i.e. delete new values)
	 *
	 * @param object Instance of this class on which modifications are cancelled.
	 */
	protected void cancel(Object object)
	{
	}

	/**
	 * Gets identifier of the class.
	 *
	 * @return Identifier of the class.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets name of the class. The name of the class is the name in the information system. By
	 * default, if no name is specified, this method returns the id.
	 *
	 * @return Name of the class.
	 */
	public String getName()
	{
		if (_name == null)
			return getId();

		return _name;
	}

	/**
	 * Gets the physical representation of the instances : java class that really holds instances of
	 * this class.
	 *
	 * @return Java class of the "objects" of this class.
	 */
	public Class<?> getObjectClass()
	{
		return _objectClass;
	}

	/**
	 * Gets the data source that manages all instances of this class.
	 *
	 * @return Data source that manages the instances of this class.
	 */
	public LyDataSource getDataSource()
	{
		return _dataSource;
	}

	/**
	 * Gets the list of fields of this class.
	 *
	 * @return List of all fields of this class.
	 */
	public Vector<LyField> getFields()
	{
        int              i, n = _fields.length;
        Vector<LyField>  fields = new Vector<LyField>(n);
        
        for (i = 0; i < n; i++)
            fields.add(_fields[i]);
		return fields;
	}

	/**
	 * Gets field at the given index.
	 *
	 * @param  index Index of searched field.
	 * @return Field at the given index.
	 */
	public LyPeerField getField(int index)
	{
		return _fields[index];
	}

	/**
	 * Gets the number of fields of this class.
	 *
	 * @return Number of fields of this class.
	 */
	public int getFieldCount()
	{
		return _fields.length;
	}

	/**
	 * Gets the list of id fields of this class.
	 *
	 * @return List of all id fields of this class.
	 */
	public Vector<LyField> getIdFields()
	{
        int              i, n = _idFields.length;
        Vector<LyField>  fields = new Vector<LyField>(n);
        
        for (i = 0; i < n; i++)
            fields.add(_idFields[i]);
        return fields;
	}

	/**
	 * Sets identifier of the class.
	 *
	 * @param     id Identifier of the class.
	 * @exception IllegalStateException If the class has already an id.
	 */
	public void setId(String id)
	{
		if (_id != null)
			throw new IllegalStateException(
				"BF AppEngine : Attempt to set id on a peer class that has already one");

		_id = id;
	}

	/**
	 * Sets name of the class.
	 *
	 * @param     name Name of the class.
	 * @exception IllegalStateException If the class has already a name.
	 */
	public void setName(String name)
	{
		if (_name != null)
			throw new IllegalStateException(
				"BF AppEngine : Attempt to set name on a peer class that has already one");

		_name = name;
	}

	/**
	 * Sets the data source of this class.
	 *
	 * @param     dataSource Data source of the class.
	 * @exception IllegalStateException If the class has already a data source.
	 */
	public void setDataSource(LyDataSource dataSource)
	{
		if (_dataSource != null)
			throw new IllegalStateException(
				"BF AppEngine : Attempt to set data source on a peer class that has already one");

		_dataSource = dataSource;
	}

	/**
	 * Gets index of the given peer field in this class.
	 *
	 * @param  field Peer field for which the index is searched.
	 * @return Index of the given field in this peer class.
	 */
	public final int getIndex(LyPeerField field)
	{
		Integer index = _indexes.get(field);

		return ((index == null) ? -1 : index.intValue());
	}

	/**
	 * Adds a peer field in this peer class.
	 *
	 * @param     field Peer field added to this class.
	 * @return    Index of the given field in this class.
	 * @exception IllegalArgumentException If the given field is null.
	 * @exception IllegalStateException    If the class has already a field with the same id.
	 */
	public int addField(LyPeerField field)
	{
		if (field == null)
			throw new IllegalArgumentException("BF AppEngine : field can't be null");

		String    id = field.getId();

		if (getField(id) != null)
			throw new IllegalStateException("BF AppEngine : peer class " + _id
				+ " has already a field with the  id " + id);

        LyPeerField[]   oldFields = _fields;
		int             index = _fields.length;
        
        _fields = new LyPeerField[index + 1];
        if (index > 0)
            System.arraycopy(oldFields, 0, _fields, 0, index);
        
		_fields[index] = field;
		_indexes.put(field, new Integer(index));

		return index;
	}
	
	/**
     * Removes a peer field in this peer class.
     *
     * @param     field Peer field removed to this class.
     * @exception IllegalArgumentException If the given field is null.
     */
    public void removeField(LyPeerField field)
    {
        if (field == null)
            throw new IllegalArgumentException("BF AppEngine : field can't be null");

        int       index = getIndex(field);

        if (index < 0)
            return;

        // Remove field from fields table
        LyPeerField[]   oldFields = _fields;
        int             nf = _fields.length;
        
        _fields = new LyPeerField[nf - 1];
        if (index > 0)
            System.arraycopy(oldFields, 0, _fields, 0, index);
        if (index < nf - 1)
            System.arraycopy(oldFields, index + 1, _fields, index, nf - index - 1);
        
        // Remove index and update following ones
        _indexes.remove(field);
        for (; index < nf - 1; index++)
            _indexes.put(_fields[index], new Integer(index));
        
        // Remove field from id fields table if needed
        int     i, ni = _idFields.length;
        
        for (i = 0; (i < ni) && (_idFields[i] != field); i++);
        if (i < ni)
        {
            LyPeerField[]   oldIdFields = _idFields;
            
            _idFields = new LyPeerField[ni - 1];
            if (i > 0)
                System.arraycopy(oldIdFields, 0, _idFields, 0, i);
            if (i < ni - 1)
                System.arraycopy(oldIdFields, i + 1, _idFields, i, ni - i - 1);
        }
    }

	/**
	 * Retrieves a field on this class by its identifier.
	 *
	 * @param     id Identifier of the searched field.
	 * @return    The field of this class which has the specified identifier, null if no field is
	 *            found.
	 * @exception IllegalArgumentException If the given id is null.
	 */
	public LyField getField(String id)
	{
		int     i, n = _fields.length;
		LyField field;

        id = id.intern();
		for (i = 0; i < n; i++)
		{
			field = _fields[i];
			if (field.getId().equals(id))
				return field;
		}

		return null;
	}

	/**
	 * Registers a peer field as an id field on this class.
	 *
	 * @param field Peer field to register as id field.
	 */
	public void addIdField(LyPeerField field)
	{
        int     i, n = _fields.length;
        
        for (i = 0; (i < n) && (_fields[i] != field); i++);
        if (i == n)
            addField(field);
               
        for (i = 0, n = _idFields.length; (i < n) && (_idFields[i] != field); i++);
        if (i == n)
        {
            LyPeerField[]   oldIdFields = _idFields;
            int             index = _idFields.length;
            
            _idFields = new LyPeerField[index + 1];
            if (index > 0)
                System.arraycopy(oldIdFields, 0, _idFields, 0, index);
            _idFields[index] = field;
        }
	}

	/**
	 * Gets identifier for an instance of this class. This method uses internal getId method to
	 * retrieve the object id for the given object, if this method returns null, the object id is
	 * calculated by constructing an object id from values of id fields and finally registered by
	 * using setId method.
	 *
	 * @param  object Instance of this class.
	 * @return Identifier of the given object.
	 * @see    LyPeerClass#getId
	 * @see    LyPeerClass#setId
	 */
	public LyObjectId getObjectId(Object object)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		LyObjectId id = getId(object);

		if (id != null)
			return id;

		id = getObjectIdNoCache(object);
		setId(object, id);

		return id;
	}

	/**
	 * Gets identifier for an instance of this class. This method uses internal getId methd to
	 * retrieve the object id for the given object, if this method returns null, the object id is
	 * calculated by contructing an object id from values of id fields and finally registered by
	 * using setId method.
	 *
	 * @param  object Instance of this class.
	 * @return Identifier of the given object.
	 * @see    LyPeerClass#getId
	 * @see    LyPeerClass#setId
	 */
	public LyObjectId getObjectIdNoCache(Object object)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		if (_idFields.length == 0)
			throw new IllegalStateException("BF AppEngine : no id fields in peer class " + _name);

		int		    i, n;
		LyPeerField peerField;
		Object	    result;

		n = _idFields.length;
		if (n == 1)
		{
			peerField = _idFields[0];
			result = getValue(object, peerField);
			if (result == null)
				result = object.hashCode(); // Fake id
		}
		else
		{
			result = new Object[n];
			for (i = 0; i < n; i++)
			{
				peerField = _idFields[i];
				((Object[])result)[i] = getValue(object, peerField);
			}
		}

		return new LyObjectId(this, result);
	} // end method getObjectIdNoCache

	/**
	 * Gets the value corresponding to the given field for an instance of this class. The
	 * implementation is to check parameters and call get method.
	 *
	 * @param     object Instance of this class.
	 * @param     field  Field for which a value is requested.
	 * @return    Value of the given field on the specified instance.
	 * @exception IllegalArgumentException If object or field is null.
	 */
	public Object getValue(Object object, LyField field)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");
		if (field == null)
			throw new IllegalArgumentException("BF AppEngine : field can't be null");

		return get(object, (LyPeerField)field);
	}

	/**
	 * Sets the value corresponding to the given field on an instance of this class. The
	 * implementation is to check parameters and call set method.
	 *
	 * @param     object Instance of this class on which a value is set.
	 * @param     field  Field for which a value is set.
	 * @param     value  Value set for the given field on the specified instance.
	 * @exception IllegalArgumentException If object or field is null.
	 * @exception IllegalArgumentException If the given value is not null but doesn't match the
	 *                                     value class of the field.
	 */
	public void setValue(Object object, LyField field, Object value)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");
		if (field == null)
			throw new IllegalArgumentException("BF AppEngine : field can't be null");

		set(object, (LyPeerField)field, value);
	}

	/**
	 * Gets all field values from an instance of this class.
	 *
	 * @param  object Instance of this class.
	 * @return All the field values of the given instance.
	 */
	public LyValueSet getValues(Object object)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");
		if (_fields == null)
			throw new IllegalStateException("BF AppEngine : no fields defined in the class " + this);

		int		    i, n = _fields.length;
		LyPeerField field;
		LyValueSet  result = null;
		Object	    value;

		for (i = 0; i < n; i++)
		{
			field = _fields[i];
			value = getValue(object, field);
			if (result == null)
				result = new LyValueSet(false);
			result.addValue(field, value);
		}
		result.enableCheckField();
		return result;
	}

	/**
	 * Copies all field values of an instance of this class on another instance.
	 *
	 * @param  object1 Object on which values are copied.
	 * @param  object2 Object from which values are get.
	 * @return Field values that are copied from object2 to object1.
	 */
	public LyValueSet copy(Object object1, Object object2)
	{
		if (object1 == null)
			throw new IllegalArgumentException("BF AppEngine : object1 can't be null");
		if (object2 == null)
			throw new IllegalArgumentException("BF AppEngine : object2 can't be null");

		LyValueSet values = getValues(object2);
		LyValue    value;
		Object     val1, val2;

		for (int i = 0; i < values.getSize(); i++)
		{
			value = values.getFieldValue(i);
			try
			{
				val1 = getValue(object1, value.getField());
			}
			catch (LyValueNotLoadedException e)
			{
				val1 = VALUE_NOT_LOADED;
			}
			val2 = value.getValue();

			if (val1 == VALUE_NOT_LOADED)
				setValue(object1, value.getField(), val2);
			else if ((val1 == null) && (val2 == null))
				continue;
			else if ((val1 != null) && (val1.equals(val2)))
				continue;
			else if ((val1 == null) || (val2 == null))
				setValue(object1, value.getField(), val2);
			else
			{
				if ((val1.getClass().isArray()) && (val2.getClass().isArray()))
				{
				  if (Array.getLength(val1) != Array.getLength(val2))
						setValue(object1, value.getField(), val2);
					else
					{
						int    j, n;
						Object subVal1, subVal2;

						n = Array.getLength(val1);
						for (j = 0; j < n; j++)
						{
							subVal1 = Array.get(val1, j); 
							subVal2 = Array.get(val2, j); 
							if ((subVal1 == null) && (subVal2 == null))
								continue;
							else if ((subVal1 != null) && (subVal1.equals(subVal2)))
								continue;
							else
								break;
						}

						if (j != n)
							setValue(object1, value.getField(), val2);
					}
				}
				else
					setValue(object1, value.getField(), val2);
			}
		}

		return validate(object1, null);
	} // end method copy

	/**
	 * Dump content of this peer class to output of the environment.
	 *
	 * @param environment The environment to dump the class.
	 */
	public void dump(LyEnvironment environment)
	{
		environment.logOut("Peer class : " + _id);
		environment.logOut("\tname = " + _name);
		environment.logOut("\tobjectClass = " + _objectClass);
		environment.logOut("\tList of peerFields :");

		LyPeerField field;
		int		    i, n = _fields.length;

		for (i = 0; i < n; i++)
		{
			field = _fields[i];
			environment.logOut("\t\t" + field);
		}
	}

	/**
	 * Returns a display string.
	 */
	public String toString()
	{
		return getName() + "(" + getId() + ")";
	}

	/**
	 * Gets current environment.
	 */
	protected LyEnvironment getEnvironment()
	{
		if (_dataSource == null)
			return null;
		LyApplication application = _dataSource.getApplication();

		return application.getEnvironment();
	}
} // end class LyPeerClass
