/* 
 * LyPeerDataProvider.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer;

import java.util.*;

import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.*;
import leon.info.location.*;
import leon.misc.*;
import leon.peer.file.*;


/**
 * This class is provided is ease implementations of data providers that use peer classes
 * (LyPeerClass) and peer fields (LyPeerField).
 *
 * <p>Two methods are abstract and must be implemented by subclasses :</p>
 *
 * <ul>
 *   <li>createPeerClass, creates the right instance of peer class,</li>
 *   <li>createPeerField, creates the right instance of peer field.</li>
 * </ul>
 *
 * @see LyPeerClass
 * @see LyPeerField
 */
public abstract class LyPeerDataProvider extends LyDataProvider
{
	public static final String				 __VERSION =
		"$Id: LyPeerDataProvider.java 40889 2012-09-21 09:20:18Z jlsaouli $";

	/** Indicates if this provider is valid or not. */
	protected boolean						 _valid;

	/**
	 * Hashtable of peer classes containing pairs of [id, peerClass] constructed by this data
	 * provider.
	 */
	protected Hashtable<String, LyPeerClass> _peerClasses;

	/**
	 * Constructs a peer data provider.
	 *
	 * @param application Current application instance
	 * @param location    This value is obtained from the location parameter in the meta model.
	 */
	public LyPeerDataProvider(LyApplication application, LyLocation location)
	{
		super(application, location);
	}

	/**
	 * Constructs a peer data provider.
	 *
	 * @param      application Current application instance
	 * @param      id          Identifier of this data provider.
	 * @deprecated should use LyDataProvider(LyApplication application, LyLocation location) instead
	 *             of this constructor.
	 */
	public LyPeerDataProvider(LyApplication application, String id)
	{
		super(application, id);
	}

	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		super.init();
		_valid = false;
	}

	/**
	 * Test if his provider is valid or not.
	 *
	 * @return true if the provider is valid, false otherwise.
	 */
	protected boolean isValid()
	{
		return _valid;
	}

	/**
	 * Retrieves a peer class from its id.
	 * @param id Id of the searched peer class.
	 * @return The peer class if found, null otherwise.
	 */
	public LyPeerClass getPeerClass(String id)
	{
	    return (_peerClasses == null ? null : _peerClasses.get(id));
	}
	
	/**
	 * To save data managed by this provider.
	 *
	 * @return true if all is all right, else false.
	 */
	protected boolean save(boolean loadEmptyCaches)
	{
		LyInfoList<LyClassInfo>	classInfos = _application.getClassInfos();
		int					    i, n = classInfos.getSize();
		LyDataProvider		    provider = getSaveFileDataProvider();
		LyClassInfo			    classInfo;
		LyClassBinding		    classBinding;
		ArrayList<LyDataSource> dataSources = new ArrayList<LyDataSource>();
		ArrayList<LyClassInfo>  classes = new ArrayList<LyClassInfo>();
		ArrayList<LyClassInfo>  otherClasses = new ArrayList<LyClassInfo>();
		ArrayList<LyResponse>   responses = new ArrayList<LyResponse>();
		LyObjectSource		    objectSource;
		LyCache				    cache;

		for (i = 0; i < n; i++)
		{
			classInfo = classInfos.getInfo(i);
			if (classInfo.isVirtual())
				continue;

			classBinding = classInfo.getBinding();

			if (!isValid(classBinding))
				continue;

			objectSource = _application.getObjectSource(classInfo);
			if (objectSource == null)
				continue;

			cache = objectSource.getCache();

			if (cache == null)
			{
				if (objectSource instanceof LyExtendObjectSource)
				{
					// Now an inherited class may not be abstract
					objectSource = ((LyExtendObjectSource)objectSource).getObjectSource();
					if (objectSource == null)
						// The class is really abstract !
						continue;
					cache = objectSource.getCache();
					if (cache == null)
						// May never occured
						continue;
				}
			}

			// If cache is empty and data are not loaded, we have nothing to do
			if (!loadEmptyCaches
					&& cache.isEmpty()
					&& !cache.isLoaded())
				otherClasses.add(classInfo);
			else
			{
				// This classInfo must be load, we store it.
				classes.add(classInfo);

				// Get all peer data sources for this classInfo.
				addDataSources(dataSources, classBinding);
			}
		}

		// Now we must test if classInfos not stored in the classes list have a
		// dataSource in the dataSources list : in this case, we must load this
		// classInfo too (case of many classInfos on the same peer class).
		ArrayList<LyDataSource> list;
		Iterator<LyClassInfo>   it = otherClasses.iterator();
		Iterator<LyDataSource>  it2;
		LyDataSource		    dataSource;

		while (it.hasNext())
		{
			classInfo = it.next();
			if (classes.contains(classInfo))
				continue;

			classBinding = classInfo.getBinding();
			if (!isValid(classBinding))
				continue;

			list = new ArrayList<LyDataSource>();
			addDataSources(list, classBinding);

			it2 = list.iterator();
			while (it2.hasNext())
			{
				dataSource = it2.next();
				if ((dataSources.contains(dataSource))
						&& (!classes.contains(classInfo)))
					classes.add(classInfo);
			}
		}

		// Now we can load the datas.
		it = classes.iterator();
		while (it.hasNext())
		{
			classInfo = it.next();
			loadData(responses, classInfo);
		}

		// Now we can save the datas.
		boolean   saved = true;
		    
		it2 = dataSources.iterator();
		while (it2.hasNext())
		{
			dataSource = it2.next();
			if (!loadEmptyCaches
			        && !isModified(dataSource.getDataClass()))
			    continue;
			
			if (save(dataSource, provider))
			    setModified(false, dataSource.getDataClass(), null);
			else
			    saved = false;
		}

		return saved;
	} // end method save

	/**
	 * Search if the specified classBinding is valid for this data provider (i.e. the location id is
	 * the same).
	 *
	 * @param  classBinding
	 * @return true if thid binding is valid, false otherwise.
	 */
	protected boolean isValid(LyClassBinding classBinding)
	{
		if (classBinding == null)
			return false;
		else if (classBinding instanceof LyUnionBinding)
		{
			LyUnionBinding unionBinding = (LyUnionBinding)classBinding;

			return (isValid(unionBinding.getClassBinding1())
					|| isValid(unionBinding.getClassBinding2()));
		}
		else
			return (_identifier.equals(classBinding.getLocation()));
	}

	/**
	 * Recursive method to get all data sources for the specified binding with classInfo to be load.
	 */
	protected void addDataSources(Collection<LyDataSource> dataSources, LyClassBinding classBinding)
	{
		if (classBinding == null)
			return;
		else if (classBinding instanceof LyUnionBinding)
		{
			LyUnionBinding unionBinding = (LyUnionBinding)classBinding;

			addDataSources(dataSources, unionBinding.getClassBinding1());
			addDataSources(dataSources, unionBinding.getClassBinding2());
		}
		else if (!_identifier.equals(classBinding.getLocation()))
			return;
		else
		{
			LyClass		 peerClass = classBinding.getPeerClass();
			LyDataSource dataSource = _application.getDataSource(peerClass);

			if (!dataSources.contains(dataSource))
				dataSources.add(dataSource);
		}
	}

	/**
	 * Recursive method to get all data sources for the specified binding.
	 */
	protected void loadData(Collection<LyResponse> responses, LyClassInfo classInfo)
	{
		LyObjectSource objectSource = _application.getObjectSource(classInfo);
		LyCache		   cache = objectSource.getCache();

		if ((cache == null) && (objectSource instanceof LyExtendObjectSource))
		{
			// Now an inherited class may not be abstract
			objectSource = ((LyExtendObjectSource)objectSource).getObjectSource();
			cache = objectSource.getCache();
		}

		if (cache == null)
			// May never occured.
			return;

		// Nothing to do if the cache is already loaded.
		if (cache.isLoaded())
			return;

		LyResponse response;

		response = objectSource.loadObjects(null, null, null, -1);
		responses.add(response);
	}

	/**
	 * Saves given peer data source.
	 */
	protected boolean save(LyDataSource dataSource, LyDataProvider provider)
	{
		LyClass				 peerClass = dataSource.getDataClass();
		LyFileRequestEncoder encoder;
		LyResponse			 response;
		LyCache				 cache;

		cache = dataSource.getCache();
		encoder = (LyFileRequestEncoder)provider.getRequestEncoder(dataSource);
		response = new LyResponse(LyResponse.Operation.SAVE, peerClass);
		encoder.setResponse(response);

		Vector<Object> objects = cache.getObjects();

		sort(objects, peerClass);
		return encoder.saveObjects(objects, null, -1);
	}

	/**
	 * Return the file data provider.
	 *
	 * @return the file data provider.
	 */
	protected LyDataProvider getSaveFileDataProvider()
	{
		LyLocation     location;
		LyDataProvider provider;

		location = new LyFileLocation(_identifier, "LY_DATA_DIR", LyFileLocation.Type.CSV, null,
				null);
		provider = LyDataProviderFactory.getProvider(_application, location);

		return provider;
	}

	/**
	 * Utility method that implements quick sorting. Sorts objets beetween the given indexes.
	 *
	 * @param lo0 Min index.
	 * @param hi0 Max index.
	 */
	void quickSort(Vector<Object> objects, LyClass peerClass, int lo0, int hi0)
	{
		if (hi0 < lo0)
			return;

		int    lo = lo0;
		int    hi = hi0;
		Object mid, low, high;

		// Arbitrarily establishing partition element as the midpoint of the array.
		mid = objects.elementAt((lo0 + hi0) / 2);

		// loop through the array until indices cross
		while (lo <= hi)
		{
			// Find the first element that is greater than or equal
			// to the partition element starting from the left Index.
			low = objects.elementAt(lo);

			while ((lo < hi0) && (compare(peerClass, low, mid) < 0))
			{
				++lo;
				low = objects.elementAt(lo);
			}

			// Find an element that is smaller than or equal to
			// the partition element starting from the right Index.
			high = objects.elementAt(hi);

			while ((hi > lo0) && (compare(peerClass, high, mid) > 0))
			{
				--hi;
				high = objects.elementAt(hi);
			}

			// if the indexes have not crossed, swap
			if (lo <= hi)
			{
				if (lo < hi)
				{
					objects.setElementAt(high, lo);
					objects.setElementAt(low, hi);
				}
				++lo;
				--hi;
			}
		}

		// If the right index has not reached the left side of array
		// must now sort the left partition.
		if (lo0 < hi)
			quickSort(objects, peerClass, lo0, hi);

		// If the left index has not reached the right side of array
		// must now sort the right partition.
		if (lo < hi0)
			quickSort(objects, peerClass, lo, hi0);
	} // end method quickSort

	protected int compare(LyClass peerClass, Object peerObject1, Object peerObject2)
	{
		if (peerObject1 == peerObject2)
			return 0;

		Vector<LyField> idFields = new Vector<LyField>(peerClass.getIdFields());
		int			    i, n = idFields.size();
		LyField		    field;
		Object		    value1, value2;
		int			    result = 0;

		for (i = 0; (i < n) && (result == 0); i++)
		{
			field = idFields.elementAt(i);
			value1 = peerClass.getValue(peerObject1, field);
			value2 = peerClass.getValue(peerObject2, field);
			if (value1 == value2)
				continue;

			if (value1 == null)
				result = -1;
			else if (value2 == null)
				result = 1;
			else if (Number.class.isAssignableFrom(field.getValueClass()))
				result = (new Double(((Number)value1).doubleValue())).compareTo((new Double(
								((Number)value2).doubleValue())));
			else if (field.getValueClass() == Date.class)
				result = ((Date)value1).compareTo((Date)value2);
			else
				result = value1.toString().compareTo(value2.toString());
		}

		return result;
	} // end method compare

	/**
	 * To sort a list of objects by their ids.
	 */
	protected void sort(Vector<Object> objects, LyClass peerClass)
	{
		if ((objects == null) || (objects.size() == 0))
			return;

		quickSort(objects, peerClass, 0, objects.size() - 1);
	}

	/**
	 * This method is used to build all the class peers for all the classInfos of the location
	 * managed by this provider. Validating a peer data provider consists to :
	 *
	 * <ul>
	 *   <li>start validation using startValidate method,</li>
	 *   <li>validate logical classes (LyClassInfo) binded on the location managed by this provider
	 *     by calling create,</li>
	 *   <li>end validation using startValidate method,</li>
	 * </ul>
	 *
	 * @param  classInfos Classes to validate.
	 * @return True if the validation was successful, false otherwise.
	 */
	public boolean validate(LyInfoList<LyClassInfo> classInfos)
	{
		// startValidate must be call only once
		if (!isValid())
		{
			if (!startValidate())
				return false;
		}

		// Creating the class and fields peers.
		int			   i, n = classInfos.getSize();
		LyClassInfo    classInfo;
		LyClassBinding classBinding;

		if (_peerClasses == null)
			_peerClasses = new Hashtable<String, LyPeerClass>();

		for (i = 0; i < n; i++)
		{
			classInfo = classInfos.getInfo(i);

			// Local classes must use a leon.peer.local.LyLocalDataProvider
			if (classInfo.hasMark(LyClassInfo.MARK_LOCAL)
					|| (classInfo.isAbstract()
						&& (classInfo.getBinding() == null)))
				continue;

			if (classInfo.isVirtual())
				continue;

			classBinding = classInfo.getBinding();
			if (classBinding != null)
			{
				if (_identifier.equals(classBinding.getLocation()))
					validate(classInfo); // The classInfo has the right provider id, validate it.
				else if (classBinding instanceof LyUnionBinding)
					validateBinding(classBinding, classInfo, false);
			}
		}

		// Now we must verify that all peer classes have some id peer fields.
		LyPeerClass		    peerClass;
		Enumeration<String> keys = _peerClasses.keys();
		String			    binding;
		Vector<LyField>	    idFields;

		while (keys.hasMoreElements())
		{
			binding = keys.nextElement();
			peerClass = _peerClasses.get(binding);
			idFields = peerClass.getIdFields();

			if ((idFields == null) || (idFields.size() == 0))
			{
				// Search in all classInfos binding to this peer class for id field infos.
				for (i = 0; i < n; i++)
				{
					classInfo = classInfos.getInfo(i);

					// Ignore local classInfos.
					if (classInfo.isLocal() 
							&& !(classInfo.isAbstract() && classInfo.getBinding() != null))
						continue;

					classBinding = classInfo.getBinding();
					getPeerIds(peerClass, classInfo, classBinding);
				}
			}

			// Now the peer class must have id fields.
			idFields = peerClass.getIdFields();
			if (((idFields == null) || (idFields.size() == 0))
			    && !binding.startsWith("ABSTRACT##"))
			{
				_application.setError("Peer class " + peerClass.getId()
					+ " has no id fields for provider : " + getId());
				return false;
			}
		}

		LyEnvironment environment = getEnvironment();

		if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
		{
			// Dumping to standard output all the class peers.
			Enumeration<LyPeerClass> peerClasses = _peerClasses.elements();

			environment.logOut("Class peers created for data provider " + _identifier + " : ");
			while (peerClasses.hasMoreElements())
			{
				peerClass = peerClasses.nextElement();
				peerClass.dump(getEnvironment());
			}
		}

		boolean result = endValidate();

		_valid = true;
		return result;
	} // end method validate

	/**
	 * Validate the associated event notifier if exists.
	 *
	 * @param  classInfos Classes to validate.
	 * @return true if the validation was successful.
	 */
	public boolean validateNotifier(LyInfoList<LyClassInfo> classInfos)
	{
		// Validate notifier if any
		if (_eventNotifier != null)
		{
			if (!_eventNotifier.validate(classInfos))
			{
				_eventNotifier.close();
				_eventNotifier = null;

				LyEnvironment env = getApplication().getEnvironment();

				try
				{
					env.raiseError(env.getMessage("LY_ER_NOTIFIER_NOT_READY"));
				}
				catch (Exception e)
				{
					env.logError(env.getMessage("LY_ER_NOTIFIER_NOT_READY"));
				}
				return true;
			}
		}
		return true;
	}

	/**
	 * Search for peer fields associated with fieldinfos with the mark ID.
	 */
	protected void getPeerIds(LyPeerClass peerClass, LyClassInfo classInfo,
		LyClassBinding classBinding)
	{
		if (classBinding == null)
			return;

		if (classBinding instanceof LyUnionBinding)
		{
			LyUnionBinding unionBinding = (LyUnionBinding)classBinding;

			getPeerIds(peerClass, classInfo, unionBinding.getClassBinding1());
			getPeerIds(peerClass, classInfo, unionBinding.getClassBinding2());
			return;
		}

		if (!_identifier.equals(classBinding.getLocation()))
			return;

		if (classBinding.getPeerClass() != peerClass)
			return;

		LyInfoList<LyFieldInfo> idFieldInfos = classInfo.getIdFieldInfos();
		int					    i, j, n = ((idFieldInfos == null) ? 0 : idFieldInfos.getSize());
		LyFieldInfo			    fieldInfo;
		LyField[]			    peerFields;
		LyPeerField			    peerField;

		for (i = 0; i < n; i++)
		{
			fieldInfo = idFieldInfos.getInfo(i);
			peerFields = classBinding.getPeerFields(fieldInfo);

			if (peerFields != null)
			{
				for (j = 0; j < peerFields.length; j++)
				{
					peerField = (LyPeerField)peerFields[j];
					if (peerClass.getField(peerField.getId()) != null)
						peerClass.addIdField(peerField);
				}
			}
		}
	} // end method getPeerIds

	/**
	 * Starts validation. Default implementation creates connections.
	 *
	 * @return True if the validation was successfully started, false otherwise.
	 * @see    LyDataProvider#createConnections
	 */
	protected boolean startValidate()
	{
		// Create and open the connection.
		return true;
	}

	/**
	 * Ends validation. Default implementation validate event notifier if there's one.
	 *
	 * @return True if the validation was successfully started, false otherwise.
	 * @see    leon.notifier.LyEventNotifier#validate
	 */
	protected boolean endValidate()
	{
		return true;
	}

	/**
	 * Validates the given class info : this mainly consists to create a peer class for each
	 * encountered class binding. Note : peer classes must be shared when several classes have the
	 * same binding. The process is the following :
	 *
	 * <ul>
	 *   <li>create object source for the class info,</li>
	 *   <li>create or retrieve the associated class peer and data source corresponding to the class
	 *     binding,</li>
	 *   <li>register the object source as listener of the data source,</li>
	 *   <li>register the peer class on the class binding,</li>
	 *   <li>validate field infos of this class info.</li>
	 * </ul>
	 *
	 * @param  classInfo Class info that is validated.
	 * @return True if the validation was successful, false otherwise.
	 * @see    LyObjectSource#LyObjectSource(LyDataProvider, LyClassInfo)
	 * @see    LyPeerDataProvider#createPeerClass
	 * @see    LyDataSource#LyDataSource(LyDataProvider, LyClass)
	 * @see    LyObjectSource#listenTo
	 * @see    LyClassBinding#setPeerClass
	 * @see    LyPeerDataProvider#validate(LyClassInfo, LyFieldInfo)
	 */
	protected void validate(LyClassInfo classInfo)
	{
		// XXX Workaround for a class created internally by a struct declared
		// in an include, the intern class is validated by both projects
		// This problem should be solved in LyProject.getLoadedClasses method
		try
		{
			if (_application.getDataSource(classInfo) != null)
			{
				// Already validated but we must see if
				// some fieldInfos have been added.
				validateFieldInfos(classInfo);
				return;
			}
		}
		catch (Exception e)
		{
		}

		LyClassBinding classBinding = classInfo.getBinding();
		String		   binding = classBinding.getBinding();
		LyPeerClass    peerClass;

		if (classInfo.isAbstract())
		{
			peerClass = createPeerClass(_identifier + '#' + binding, classInfo);
			if (!_peerClasses.containsKey("ABSTRACT##"+ binding))
			{
				_peerClasses.put("ABSTRACT##"+ binding, peerClass);
			}
		}
		else
		{
			LyObjectSource objectSource;
			LyDataSource   dataSource;

			// First we create the objectSource and we associate the infoProvider
			objectSource = createObjectSource(classInfo);

			// Now we create the peer class if it is not exist.
			if (!_peerClasses.containsKey(binding))
			{
				// Creating the peer class.
				peerClass = createPeerClass(_identifier + '#' + binding, classInfo);

				// Creating the dataSource.
				dataSource = new LyPeerDataSource(this, peerClass);
				peerClass.setDataSource(dataSource);

				// Adding the class in the list.
				_peerClasses.put(binding, peerClass);
			}
			else
			{
				// Get the class.
				peerClass = _peerClasses.get(binding);

				// Get the dataSource.
				dataSource = _application.getDataSource(peerClass);
			}

			// The objectSource must listen to dataSource events.
			objectSource.listenTo(dataSource);
		}

		// Associating the binding and the class peer.
		classBinding.setPeerClass(peerClass);

		// Searching for all fieldInfos in the classInfo and create
		// the field peers if they do not exist.
		validateFieldInfos(classInfo);
	} // end method validate

	/**
	 * Validates the field info list of the given class info : this mainly consists to create or
	 * retrieve the associated peer fields.
	 *
	 * @param classInfo Class info to which the field info list must be validated.
	 */
	protected void validateFieldInfos(LyClassInfo classInfo)
	{
		// Searching for all fieldInfos in the classInfo and create
		// the field peers if they do not exist.
		LyFieldInfo    fieldInfo;
		LyFieldBinding fieldBinding;
		int			   i, n = classInfo.getFieldInfoCount();

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			fieldBinding = fieldInfo.getBinding();
			if ((fieldBinding != null)
					&& ((fieldBinding.getClass() == LyFieldBinding.class)
						|| (fieldBinding.getClass() == LyFieldsBinding.class)))
				validate(classInfo, fieldInfo);
		}
	}

	/**
	 * Create an object source for the specified class info.
	 *
	 * @param  classInfo the class info for creating the object source
	 * @return the created object source for the specified classInfo.
	 */
	protected LyObjectSource createObjectSource(LyClassInfo classInfo)
	{
		return (new LyObjectSource(_application.getInfoProvider(), classInfo));
	}

	/**
	 * Validates the given class binding which is part of a given union class info : this mainly
	 * consists to create a peer class for each encountered class binding. Note : peer classes must
	 * be shared when several classes have the same binding. The process is the following :
	 *
	 * <ul>
	 *   <li>create or retrieve the associated class peer and data source corresponding to the class
	 *     binding,</li>
	 *   <li>register the peer class on the class binding,</li>
	 *   <li>create peer fields for field infos of the class info that corresponds to the given
	 *     binding.</li>
	 *   <li>associate peer fields to the peer class.</li>
	 * </ul>
	 *
	 * @param classBinding 		Class binding which is part of the binding of the given union class info.
	 * @param classInfo    		Class info that is validated.
	 * @param defaultBinding	Indicates default binding.
	 * @see   LyPeerDataProvider#createPeerClass
	 * @see   LyDataSource#LyDataSource(LyDataProvider, LyClass)
	 * @see   LyClassBinding#setPeerClass
	 * @see   LyPeerDataProvider#createPeerField(String id, LyFieldInfo fieldInfo, LyClassInfo
	 *        classInfo)
	 */
	protected void validateBinding(LyClassBinding classBinding, LyClassInfo classInfo, boolean defaultBinding)
	{
		if (classBinding == null)
			return;

		if (classBinding instanceof LyUnionBinding)
		{
			LyUnionBinding unionBinding = (LyUnionBinding)classBinding;

			validateBinding(unionBinding.getClassBinding1(), classInfo, true);
			validateBinding(unionBinding.getClassBinding2(), classInfo, false);
			return;
		}

		if (!_identifier.equals(classBinding.getLocation()))
			return;

		String		 binding = classBinding.getBinding();
		LyPeerClass  peerClass;
		LyDataSource dataSource;

		// Now we create the peer class if it is not exist.
		if (!_peerClasses.containsKey(binding))
		{
			// Creating the peer class.
			peerClass = createPeerClass(_identifier + '#' + binding, classBinding);

			// Creating the dataSource.
			dataSource = new LyPeerDataSource(this, peerClass);
			peerClass.setDataSource(dataSource);

			// Adding the class in the list.
			_peerClasses.put(binding, peerClass);
		}
		else // Get the class.
			peerClass = _peerClasses.get(binding);

		// Associating the binding and the class peer.
		classBinding.setPeerClass(peerClass);

		// Searching for all fieldInfos in the classInfo that corresponds to this sub-binding
		// and create the field peers if they do not exist.
		LyFieldInfo    fieldInfo;
		LyFieldBinding fieldBinding;
		int			   i, n = classInfo.getFieldInfoCount();

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			fieldBinding = fieldInfo.getBinding();
			if (fieldBinding == null)
				// Local field
				continue;

			String fbinding = fieldBinding.getBinding();
			String cbind = fieldBinding.getCBind();
			
			// validate the field if it belongs to the current class binding
			if (binding.equals(cbind)
				|| getId().equals(cbind)
				|| ((cbind == null)&& defaultBinding))
			{
				// try to get the peer field
				LyPeerField peerField = (LyPeerField)peerClass.getField(fbinding);

				if (peerField == null)
				{
					// create the peer field
					peerField = createPeerField(fbinding, fieldInfo, classBinding);

					// If the field is not in the class, add it.
					addPeerField(peerClass, peerField);
				}
				else
					checkPeerField(classInfo, fieldInfo, peerClass, peerField, fieldBinding);

				if (isIdField(peerClass, peerField, fieldBinding))
					addIdPeerField(peerClass, peerField);
				
				// Associate binding and field peer.
				registerPeerField(classBinding, fieldInfo, peerField);
			}
			else if (fieldBinding instanceof LyFieldsBinding)
			{
				LyFieldsBinding fieldsBinding = (LyFieldsBinding)fieldBinding;
				int			    j, m = fieldsBinding.getFieldBindingCount(), counter = 0;
				LyPeerField[]   peerFields = new LyPeerField[m];
				String		    subFieldBinding;
				String		    subCbind;

				for (j = 0; j < m; j++)
				{
					fieldBinding = fieldsBinding.getFieldBinding(j);
					subFieldBinding = fieldBinding.getBinding();
					subCbind = fieldBinding.getCBind();

					// validate the field if it belongs to the current class binding
					if (binding.equals(subCbind) || (getId().equals(subCbind)))
					{
						peerFields[counter] = (LyPeerField)peerClass.getField(subFieldBinding);

						if (peerFields[counter] == null)
						{
							// Create the field peer.
							peerFields[counter] = createPeerField(subFieldBinding, fieldInfo,
									classBinding);

							// If the field is not in the class, add it.
							addPeerField(peerClass, peerFields[counter]);
							if (isIdField(peerClass, peerFields[counter], fieldBinding))
								addIdPeerField(peerClass, peerFields[counter]);
						}
						else
							checkPeerField(classInfo, fieldInfo, peerClass, peerFields[counter], fieldBinding);

						counter++;
					}
				}

				// Associate binding and field peers.
				LyPeerField[] finalPeerFields = new LyPeerField[counter];

				for (j = 0; j < counter; j++)
					finalPeerFields[j] = peerFields[j];

				registerPeerFields(classBinding, fieldInfo, finalPeerFields);
			}
		}
	} // end method validateBinding

	/**
	 * Validates the field info of the given class info : this mainly consists to create or retrieve
	 * the associated peer field.
	 *
	 * @param  classInfo Class info to which the field info belongs to.
	 * @param  fieldInfo Field info that is validated.
	 */
	public void validate(LyClassInfo classInfo, LyFieldInfo fieldInfo)
	{
		LyFieldBinding fieldBinding = fieldInfo.getBinding();
		LyClassBinding classBinding = classInfo.getBinding();
		LyPeerClass    peerClass = (LyPeerClass)classBinding.getPeerClass();
		String		   binding;

		if (fieldBinding instanceof LyFieldsBinding)
		{
			LyFieldsBinding fieldsBinding = (LyFieldsBinding)fieldBinding;
			int			    i, n = fieldsBinding.getFieldBindingCount();
			LyPeerField[]   peerFields = new LyPeerField[n];

			for (i = 0; i < n; i++)
			{
				fieldBinding = fieldsBinding.getFieldBinding(i);
				binding = fieldBinding.getBinding();
				peerFields[i] = (LyPeerField)peerClass.getField(binding);

				if (peerFields[i] == null)
				{
					// Create the field peer.
					peerFields[i] = createPeerField(binding, fieldInfo, classInfo);

					// If the field is not in the class, add it.
					addPeerField(peerClass, peerFields[i]);
					if (isIdField(peerClass, peerFields[i], fieldBinding))
						addIdPeerField(peerClass, peerFields[i]);
				}
				else
					checkPeerField(classInfo, fieldInfo, peerClass, peerFields[i], fieldBinding);
			}

			// Associate binding and field peers.
			registerPeerFields(classBinding, fieldInfo, peerFields);
		}
		else
		{
			LyPeerField peerField;

			binding = fieldBinding.getBinding();
			peerField = (LyPeerField)peerClass.getField(binding);

			if (peerField == null)
			{
				// Create the field peer.
				peerField = createPeerField(binding, fieldInfo, classInfo);

				// If the field is not in the class, add it.
				addPeerField(peerClass, peerField);
				if (isIdField(peerClass, peerField, fieldBinding))
					addIdPeerField(peerClass, peerField);
			}
			else
				checkPeerField(classInfo, fieldInfo, peerClass, peerField, fieldBinding);

			// Associate binding and field peer.
			registerPeerField(classBinding, fieldInfo, peerField);
		}
	} // end method validate

	/**
     * Invalidates the field info of the given class info : this mainly consists to remove
     * the associated peer field.
     *
     * @param  classInfo Class info to which the field info belongs to.
     * @param  fieldInfo Field info that is invalidated.
     */
    public void invalidate(LyClassInfo classInfo, LyFieldInfo fieldInfo)
    {
        LyFieldBinding fieldBinding = fieldInfo.getBinding();
        LyClassBinding classBinding = classInfo.getBinding();
        LyPeerClass    peerClass = (LyPeerClass)classBinding.getPeerClass();
        String         binding;

        if (fieldBinding instanceof LyFieldsBinding)
        {
            LyFieldsBinding fieldsBinding = (LyFieldsBinding)fieldBinding;
            int             i, n = fieldsBinding.getFieldBindingCount();
            LyPeerField[]   peerFields = new LyPeerField[n];

            for (i = 0; i < n; i++)
            {
                fieldBinding = fieldsBinding.getFieldBinding(i);
                binding = fieldBinding.getBinding();
                peerFields[i] = (LyPeerField)peerClass.getField(binding);
                removePeerField(peerClass, peerFields[i]);
            }

            registerPeerFields(classBinding, fieldInfo, null);
        }
        else
        {
            LyPeerField peerField;

            binding = fieldBinding.getBinding();
            peerField = (LyPeerField)peerClass.getField(binding);
            removePeerField(peerClass, peerField);
            registerPeerField(classBinding, fieldInfo, peerField);
        }
    }
    
	/**
	 * To test if peer field is compatible with fieldInfo in given classInfo.
	 */
	protected void checkPeerField(LyClassInfo classInfo, LyFieldInfo fieldInfo,
	    LyPeerClass peerClass, LyPeerField peerField, LyFieldBinding fieldBinding)
	{	    
		LyEnvironment environment = getEnvironment();

		if (!environment.isDebugLevel(LyEnvironment.DEBUG_WARNING))
			return;

		boolean		    isId = isIdField(peerClass, peerField, fieldBinding);
		Vector<LyField> idFields = new Vector<LyField>(peerClass.getIdFields());
		boolean		    wasId = (idFields != null) && (idFields.contains(peerField));

		if (isId && !wasId)
			environment.logWarning("Peer field " + peerField + " is declared id in class info "
				+ classInfo + " but is not id in another class");
		else if (wasId && !isId)
			environment.logWarning("Peer field " + peerField + " is not declared id in class info "
				+ classInfo + " but is id in another class");
	}

	/**
	 * Create a peer class with the specified id for the given class info.
	 *
	 * @param  id        Identifier of the new peer class.
	 * @param  classInfo Class info for which a peer class is created.
	 * @return The new peer class.
	 */
	protected LyPeerClass createPeerClass(String id, LyClassInfo classInfo)
	{
		LyClassBinding classBinding = classInfo.getBinding();

		return createPeerClass(id, classBinding);
	}

	/**
	 * Create a peer class with the specified id for the class binding.
	 *
	 * @param  id           Identifier of the new peer class.
	 * @param  classBinding Class binding for which a peer class is created.
	 * @return The new peer class.
	 */
	protected abstract LyPeerClass createPeerClass(String id, LyClassBinding classBinding);

	/**
	 * Create a peer field with the specified id for the field info of the given class info.
	 *
	 * @param  id        Identifier of the new peer field.
	 * @param  fieldInfo Field info for which a peer field is created.
	 * @param  classInfo Class info to which the field info belongs to.
	 * @return The new peer field.
	 */
	protected LyPeerField createPeerField(String id, LyFieldInfo fieldInfo, LyClassInfo classInfo)
	{
		LyClassBinding classBinding = classInfo.getBinding();

		return createPeerField(id, fieldInfo, classBinding);
	}

	/**
	 * Create a peer field with the specified id for the field info of the given class binding.
	 *
	 * @param  id           Identifier of the new peer field.
	 * @param  fieldInfo    Field info for which a peer field is created.
	 * @param  classBinding Class binding to which the field info belongs to.
	 * @return The new peer field.
	 */
	protected abstract LyPeerField createPeerField(String id, LyFieldInfo fieldInfo,
		LyClassBinding classBinding);

	/**
	 * Add the peer field to the given peer class. The default implementation calls
	 * LyPeerClass.addField.
	 *
	 * @param peerClass Peer class to which the pier field is added.
	 * @param peerField Added peer field.
	 * @see   LyPeerClass#addField
	 */
	protected void addPeerField(LyPeerClass peerClass, LyPeerField peerField)
	{
		peerClass.addField(peerField);
	}
	
	/**
     * Add the peer field to the given peer class. The default implementation calls
     * LyPeerClass.addField.
     *
     * @param peerClass Peer class to which the pier field is added.
     * @param peerField Added peer field.
     * @see   LyPeerClass#addField
     */
    protected void removePeerField(LyPeerClass peerClass, LyPeerField peerField)
    {
        peerClass.removeField(peerField);
    }

	/**
	 * Register peer field on the given class binding corresponding to the given field info.
	 *
	 * @param classBinding Class binding on the which the peer field are registered.
	 * @param fieldInfo    Field info to which the peer field corresponds to.
	 * @param peerField    Peer field added to the class binding.
	 * @see   LyClassBinding#addPeerField
	 */
	protected void registerPeerField(LyClassBinding classBinding,
		LyFieldInfo fieldInfo, LyPeerField peerField)
	{
		// Associate binding and field peer.
		classBinding.addPeerField(fieldInfo, peerField);
	}

	/**
	 * Register peer fields on the given class binding corresponding to the given field info.
	 *
	 * @param classBinding Class binding on the which the peer fields are registered.
	 * @param fieldInfo    Field info to which the peer fields corresponds to.
	 * @param peerFields   Peer fields added to the class binding.
	 * @see   LyClassBinding#addPeerField
	 */
	protected void registerPeerFields(LyClassBinding classBinding,
		LyFieldInfo fieldInfo, LyPeerField[] peerFields)
	{
		// Associate binding and field peer.
		classBinding.addPeerFields(fieldInfo, peerFields);
	}

	/**
	 * Determines the field is an identifier on the data provider Internal use of the validate
	 * method.
	 *
	 * <P>Default return true if the mark ID is set.</P>
	 *
	 * @param  tableClass  the peer table class reference as specified in validate method.
	 * @param  columnField peer field associated with the value we store.
	 * @param  fieldInfo   data description of the meta model.
	 * @return the java class used to store data in memory.
	 * @see    #validate()
	 */
	protected boolean isIdField(LyPeerClass peerClass, LyPeerField peerField,
		LyFieldBinding fieldBinding)
	{
		return ((fieldBinding == null) ? false : fieldBinding.isId());
	}

	/**
	 * Adds the peer field as an id of the given peer class. The default implementation calls
	 * LyPeerClass.addIdField.
	 *
	 * @param peerClass Peer class to which the pier field is added as id field.
	 * @param peerField Peer field added as id field.
	 * @see   LyPeerClass#addField
	 */
	protected void addIdPeerField(LyPeerClass peerClass, LyPeerField peerField)
	{
		peerClass.addIdField(peerField);
	}
	
    /**
     * Load peer objects coming from given provider.
     * @param application     Application on which peer objects are imported.
     * @param importProvider  Provider that gives imported objects.
     * @param enableReloading Enable reloading of existing objects.
     */
    public void load(LyApplication application, LyPeerDataProvider importProvider, boolean enableReloading)
    {
        LyInfoList<LyClassInfo>   classes = application.getClassInfos();
        int                       i, n = classes.getSize();
        LyResponse                response;
        LyClassInfo               classInfo;
        LyClass                   peerClass;
        LyObjectSource            objectSource;
        LyDataSource              dataSource;
        LyRequestEncoder          requestEncoder;
        int                       j, nr;
        Object                    peer;
        
        for (i = 0; i < n; i++)
        {
            classInfo = classes.getInfo(i);
            objectSource = classInfo.getObjectSource();
            dataSource = objectSource.getPeerDataSource();
            
            if ((dataSource != null)
                && (dataSource.getDataProvider() == this))
            {
                peerClass = classInfo.getPeerClass();
                requestEncoder = importProvider.getRequestEncoder(dataSource);
                response = new LyResponse(LyResponse.Operation.LOAD, peerClass);
                response.enableReloading(false);
                requestEncoder.setResponse(response);
                requestEncoder.getObjects(null, null, null, -1);
                
                nr = response.getSize();
                if (nr > 0)
                {
                    setModified(true, peerClass, null);
                    
                    for (j = 0; j < nr; j++)
                    {
                        peer = response.getResponse(j);
                        peerImported(peerClass, peer);
                    }
                }
            }
        }
    }
    
    /**
     * Notify import.
     * @param peerClass Peer class corresponding imported object.
     * @param peer      Imported object.
     */
    protected void peerImported(LyClass peerClass, Object peer)
    {
    }

	public void resetPeerClasses()
	{
		_peerClasses = new Hashtable<String, LyPeerClass>();
	}
} // end class LyPeerDataProvider
