/*
 * LyPeerDataSource.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer;

import java.util.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.*;
import leon.misc.*;
import leon.notifier.*;


/**
 * This class implements a collection of data of the same class. It provides methods for accessing,
 * creating, deleting, modifying and counting datas.
 *
 * <p>Both physical class and logical class are managed by data sources. The only difference is that
 * physical classes use direct instances of this class, and logical classes are handled by instances
 * of an inherited class, LyObjectSource.</p>
 *
 * <p>Each data source is connected to a data provider which is responsible to manage real objects :
 * encoding requests and decoding results.</p>
 *
 * <p>Data sources are valid within an application.</p>
 *
 * <p>Objects returned via data sources are kept in caches.</p>
 *
 * @see LyClass
 * @see LyDataSource
 */
public class LyPeerDataSource extends LyDataSource
{
	public static final String			  __VERSION =
		"$Id: LyPeerDataSource.java 41712 2012-12-05 08:51:26Z bangiole $";

	/** Listeners registered on the data source. */
	protected Vector<_LyDataSourceClient> _listeners;

	/**
	 * Constructs a data source within the given application that manages the specified class.
	 *
	 * @param provider  Provider on which this data source is connected.
	 * @param dataClass Class of data handled by this new data source.
	 */
	public LyPeerDataSource(LyDataProvider provider, LyClass dataClass)
	{
		super(provider, dataClass);

		_listeners = null;
	}

	/**
	 * Sets the cache for this data source.
	 *
	 * @param     cachePolicy Informations for the cache to associate to this object source.
	 * @exception IllegalArgumentException If the given cache policy is null.
	 * @exception IllegalStateException    If the data source is already associated to a cache that
	 *                                     contains objects.
	 */
	public void setCache(LyInfoCache cachePolicy)
	{
		if (cachePolicy == null)
			throw new IllegalArgumentException("BF AppEngine : cachePolicy parameter can't be null");

		if ((_cache != null) && (!_cache.isEmpty()))
			throw new IllegalStateException(
				"BF AppEngine : this peer data source is already associated to a cache that contains objects");

		// Requests cache policy of the data source
		requestCachePolicy(cachePolicy);
	}

	/**
	 * Registers a listener to receive events of the given type on objects contained within this
	 * data source.
	 *
	 * @param     listener New listener. To receive these events, the listener have to implement the
	 *                     LyDataSourceListener interface.
	 * @param     type     Indicates which type of event the listener wants to receive. Types are
	 *                     defined as constant in LyDataSourceEvent class : TYPE_CREATE, TYPE_SET,
	 *                     TYPE_DELETE
	 * @exception IllegalArgumentException If the listener is null or if the given type is unknown.
	 */
	public void addListener(LyDataSourceListener listener, short type)
	{
		if (listener == null)
			throw new IllegalArgumentException("BF AppEngine : listener can't be null");

		if ((type != LyDataSourceEvent.TYPE_ALL)
				&& ((type < LyDataSourceEvent.TYPE_CREATE)
					|| (type > LyDataSourceEvent.TYPE_ERROR)))
			throw new IllegalArgumentException(
				"BF AppEngine : Attempt to add a listener with an unknown type of event");

		if (_listeners == null)
			_listeners = new Vector<_LyDataSourceClient>(1);

		synchronized (_listeners)
		{
			_LyDataSourceClient client = new _LyDataSourceClient(listener, type, null);

			_listeners.addElement(client);
		}
	}

	/**
	 * Registers a listener to receive modification events concerning given fields on objects
	 * contained within this data source. Warning ! This method only registers the listener for
	 * modification events.
	 *
	 * @param     listener New listener. To receive these events, the listener have to implement the
	 *                     LyDataSourceListener interface.
	 * @param     fields   List of fields on which the listener wants to be notified when they are
	 *                     modified.
	 * @exception IllegalArgumentException If the listener is null.
	 */
	public void addListener(LyDataSourceListener listener, Vector<LyField> fields)
	{
		if (listener == null)
			throw new IllegalArgumentException("BF AppEngine : listener can't be null");

		if (_listeners == null)
			_listeners = new Vector<_LyDataSourceClient>(1);

		synchronized (_listeners)
		{
			_LyDataSourceClient client = new _LyDataSourceClient(listener,
					LyDataSourceEvent.TYPE_SET, fields);

			_listeners.addElement(client);
		}
	}

	/**
	 * Unregisters a listener for notifications from this data source. The given listener is
	 * unregistered for all types of events he has been registered for.
	 *
	 * @param     listener The listener which shall not receive further events.
	 * @exception IllegalArgumentException If the listener is null or if it is not known as listener
	 *                                     on this data source.
	 */
	public void removeListener(LyDataSourceListener listener)
	{
		if (listener == null)
			throw new IllegalArgumentException("BF AppEngine : Listener can't be null");
		if (_listeners == null)
			throw new IllegalStateException(
				"BF AppEngine : Attempt to remove a non registered listener from a datasource");

		synchronized (_listeners)
		{
			_LyDataSourceClient client;
			int				    i, n;

			for (n = _listeners.size(), i = n - 1; i >= 0; i--)
			{
				client = _listeners.elementAt(i);
				if (client.getListener() == listener)
					_listeners.removeElementAt(i);
			}

			if (_listeners.size() == n)
				throw new IllegalArgumentException(
					"BF AppEngine : Attempt to remove a non registered listener from a datasource");
		}
	}

	/**
	 * Internal method to post an event to listeners.
	 *
	 * @param     event         Posted event.
	 * @param     skipListeners Listeners that were already notified and that must be skipped by
	 *                          this method to avoid a double notification of the same event.
	 * @return    True if the vent was sent and used by at least one listener, false otherwise.
	 * @exception IllegalArgumentException If the given event is null.
	 */
	protected boolean postEvent(LyDataSourceEvent event, Vector<LyDataSourceListener> skipListeners)
	{
		if (event == null)
			throw new IllegalArgumentException("BF AppEngine : event can't be null");

		// Update the last modified time.
		Date date = new Date();

		setLastModified(date.getTime());

		if ((_listeners == null) || (_listeners.size() == 0))
			return false;

		short				  type = event.getType(), clientType;
		int					  i, n;
		LyDataSourceListener  listener;
		_LyDataSourceClient[] clients;
		boolean				  used = false;

		synchronized (_listeners)
		{
			n = _listeners.size();
			clients = new _LyDataSourceClient[n];

			for (i = 0; i < n; i++)
				clients[i] = _listeners.elementAt(i);
		}

		for (i = 0; i < n; i++)
		{
			if (!_listeners.contains(clients[i]))
				// Be careful, some clients may disappear during this loop.
				continue;

			// Check if the client is interested by the event type
			clientType = clients[i].getType();
			if ((clientType != LyDataSourceEvent.TYPE_ALL)
					&& (clientType != type))
				continue;

			// Check if the listener must be skipped
			listener = clients[i].getListener();

			if ((skipListeners != null) && skipListeners.contains(listener))
				continue;

			// Call appropriate method
			try
			{
				switch (type)
				{
				case LyDataSourceEvent.TYPE_CREATE:
					used |= listener.objectCreated(event);
					break;

				case LyDataSourceEvent.TYPE_SET:
					if (clients[i].matches(event))
						used |= listener.objectModified(event);
					break;

				case LyDataSourceEvent.TYPE_DELETE:
					used |= listener.objectDeleted(event);
					break;
				}
			}
			catch (Exception e)
			{
				LyEnvironment environment = _application.getEnvironment();

				environment.writeException(e);
			}
		}

		return used;
	} // end method postEvent

	/**
	 * Gets an object from this data source by the specified identifier.
	 *
	 * @param  id Object identifier of the searched object.
	 * @return The requested object if it was found, or null if it could not be found.
	 */
	public Object getObject(LyObjectId id)
	{
		return getObject(id, null, null);
	}

	/**
	 * Return the number of objects within this data source matching the given filter.
	 *
	 * @param  filter Filter applied on objects for counting.
	 * @return Number of objects matching the given filter in this data source.
	 */
	public int getSize(LyFilter filter)
	{
		return getSize(filter, null);
	}

	/**
	 * Return the number of objects within this data source matching the given filter.
	 *
	 * @param  filter       Filter applied on objects for counting.
	 * @param  classBinding Binding between logical class and physical class. This binding may
	 *                      contain extra information required for encoding the request (such as
	 *                      extra filter).
	 * @return Number of objects matching the given filter in this data source.
	 */
	public int getSize(LyFilter filter, LyClassBinding classBinding)
	{
		if(filter != null)
			filter = filter.getPeerFilter();

		return super.getSize(filter, classBinding);
	}

	/**
	 * Internal method used to receive a new response from data provider to a creation request sent
	 * on createObject. Object is added in the cache of this data source and a creation event is
	 * sent to listeners. If the data provider has a notifier, the creation is notified.
	 *
	 * @param response Response to the create operation.
	 * @param status   Status of the new response. Valid status are defined in LyResponse.
	 * @param object   Newly created object.
	 * @see   #createObject
	 */
	protected void receiveCreate(LyResponse response, LyResponse.Status status, Object object)
	{
		// Recompute the id of the object.
		// XXX How to avoid this ? In most of cases we don't need to do this.
		_class.rehash(object);

		// Add the object in the cache.
		putObject(object);

		// Post the notification.
		LyEventNotifier notifier = null;

		if (_provider != null)
			notifier = _provider.getEventNotifier();

		if (notifier != null)
		{
			LyTransactionManager tm;
			LyTransaction		 transaction;
			LyEventEncoder		 encoder;

			tm = _application.getTransactionManager();
			transaction = tm.getTransaction(_provider);

			if (transaction != null)
				transaction.addNotification(notifier, LyEventNotifier.EventType.CREATE,
                    _class, object, null);
			else
			{
				encoder = notifier.getEventEncoder(this);
				encoder.postCreation(object);
			}
		}

		// Set the status of the response.
		response.setStatus(status);

		// Add the object in the response.
		response.addResponse(object, null);

		// Notify all the listeners.
		Vector<LyDataSourceListener>  skipListeners = response.getSkippedListeners();
		LyDataSourceEvent             event;

		event = eventFactory(LyDataSourceEvent.TYPE_CREATE, object, null, false, false, true);
		postEvent(event, skipListeners);

	} // end method receiveCreate

	/**
	 * Internal method used to receive a new response from data provider to a set request sent on
	 * setObject or setObjects. Object is modified and a set event is sent to listeners. If the data
	 * provider has a notifier, the set is notified.
	 *
	 * @param response  Response to the set operation.
	 * @param status    Status of the new response. Valid status are defined in LyResponse.
	 * @param object    Modified object.
	 * @param oldValues Old values of the object before the set operation.
	 * @see   #setObject
	 * @see   #setObjects
	 */
	protected void receiveSet(LyResponse response, LyResponse.Status status, Object object,
		LyValueSet oldValues)
	{
		// Recompute the id of the object.
		// XXX How to avoid this ? In most of cases we don't need to do this.
		_class.rehash(object);

		LyEventNotifier notifier = null;

		if (_provider != null)
			notifier = _provider.getEventNotifier();

		// Post the notification.
		if ((oldValues != null) && (notifier != null))
		{
			LyTransactionManager tm;
			LyTransaction		 transaction;
			LyEventEncoder		 encoder;

			tm = _application.getTransactionManager();
			transaction = tm.getTransaction(_provider);

			if (transaction != null)
			{
				LyValueSet notificationValues = new LyValueSet();
				LyValue    value;

				for (int i = 0; i < oldValues.getSize(); i++)
				{
					value = oldValues.getFieldValue(i);
					notificationValues.addValue(value);
				}
				transaction.addNotification(notifier, LyEventNotifier.EventType.UPDATE,
                    _class, object, notificationValues);
			}
			else
			{
				encoder = notifier.getEventEncoder(this);
				encoder.postModification(object, oldValues);
			}
		}

		// Set the status to the response.
		response.setStatus(status);

		// Add the object in the response.
		response.addResponse(object, oldValues);

		// Notify listeners.
		Vector<LyDataSourceListener>			  skipListeners = response.getSkippedListeners();
		LyDataSourceEvent event;

		event = eventFactory(LyDataSourceEvent.TYPE_SET, object, oldValues, false, false, true);
		postEvent(event, skipListeners);
	} // end method receiveSet

	/**
	 * Internal method used to receive a new response from data provider to a deletion request sent
	 * on deleteObject or deleteObjects. A deletion event is sent to listeners and the object is
	 * deleted from the cache If the data provider has a notifier, the deletion is notified.
	 *
	 * @param response Response to the deletion operation.
	 * @param status   Status of the new response. Valid status are defined in LyResponse.
	 * @param object   Deleted object.
	 * @see   #deleteObject
	 * @see   #deleteObjects
	 */
	protected void receiveDelete(LyResponse response, LyResponse.Status status, Object object)
	{
		// Mark object as removed in the cache.
		startRemoveObject(object);

		LyEventNotifier notifier = null;

		if (_provider != null)
			notifier = _provider.getEventNotifier();

		try
		{
			// Post the notification.
			if (notifier != null)
			{
				LyTransactionManager tm;
				LyTransaction		 transaction;
				LyEventEncoder		 encoder;

				tm = _application.getTransactionManager();
				transaction = tm.getTransaction(_provider);

				if (transaction != null)
					transaction.addNotification(notifier, LyEventNotifier.EventType.DELETE,
	                    _class, object, null);
				else
				{
					encoder = notifier.getEventEncoder(this);
					encoder.postDeletion(object);
				}
			}

			// Set the status to the response.
			response.setStatus(status);

			// Add the object in the response.
			response.addResponse(object, null);

			// Notify listeners.
			Vector<LyDataSourceListener>  skipListeners = response.getSkippedListeners();
			LyDataSourceEvent             event;

			event = eventFactory(LyDataSourceEvent.TYPE_DELETE, object, null, false, false, true);
			postEvent(event, skipListeners);
		}
		finally
		{
			// Delete object from the cache
			removeObject(object);
		}
	} // end method receiveDelete

	/**
	 * Internal method used to receive a new response from data provider to a load request sent on
	 * loadObject or loadObjects. If needed, the object is added in the cache or it is updated, and
	 * then, a load event is sent to listeners.
	 *
	 * @param response Response to the load operation.
	 * @param status   Status of the new response. Valid status are defined in LyResponse.
	 * @param object   Loaded object.
	 * @see   #loadObject
	 * @see   #loadObjects
	 */
	protected void receiveLoad(LyResponse response, LyResponse.Status status, Object object,
		LyValueSet oldValues)
	{
		Object			  cachedObject = null;
		LyValueSet		  values = oldValues;
		LyDataSourceEvent event = null;

		// If the object is already in the cache, we must only
		// change the values else we must put it in the cache.
		LyObjectId		  objectId = _class.getObjectId(object);

		try
		{
			cachedObject = _cache.getObject(objectId);
		}
		catch (LyNoSuchObjectException e)
		{
			cachedObject = null;
		}

		short eventType = -1;
		if (cachedObject == null)
		{
			// Add the object in the cache.
			putObject(object);
			cachedObject = object;
			values = null;
			// Prepare the event for the listeners.
			eventType = LyDataSourceEvent.TYPE_CREATE;
		}
		else if ((cachedObject != object)
		    && response.isReloadingEnabled())
		{
			// Set locally the values of cached object
			values = _class.copy(cachedObject, object);
			eventType = LyDataSourceEvent.TYPE_SET;
		}

		// Set the status to the response.
		response.setStatus(status);

		// Add the object in the response.
		response.addResponse(cachedObject, values);

		// Notify listeners.
		if ((eventType == LyDataSourceEvent.TYPE_CREATE)
				|| ((eventType == LyDataSourceEvent.TYPE_SET)
						&& (values != null)
						&& (values.getSize() != 0)))
		{	// Prepare the event for the listeners.
			event = eventFactory(eventType, cachedObject, values, true,
					response.isNotified(), response.isLocal());

			Vector<LyDataSourceListener> skipListeners = response.getSkippedListeners();

			postEvent(event, skipListeners);
		}
	} // end method receiveLoad

	/* Internal class */

	/**
	 * This class is used to store listener parameters given when the listener registered on the
	 * data source.
	 */
	final class _LyDataSourceClient
	{
		// Listener of the data source
		LyDataSourceListener _listener;

		// Type of events the listener is interested in
		short				 _type;

		// List of fields on which the listener wants to receive modification events
		Vector<LyField>		_fields;

		/**
		 * Constructs a client of the data source.
		 *
		 * @param listener Listener of the data source
		 * @param type     Type of events the listener is interested in.
		 * @param fields   List of fields on which the listener wants to receive modification events
		 *                 (may be null, null means all fields).
		 */
		_LyDataSourceClient(LyDataSourceListener listener, short type, Vector<LyField> fields)
		{
			_listener = listener;
			_type = type;
			_fields = fields;
		}

		/**
		 * Gets listener of this client.
		 *
		 * @return The listener that registers these parameters.
		 */
		LyDataSourceListener getListener()
		{
			return _listener;
		}

		/**
		 * Gets type of events the listener is interested in.
		 *
		 * @return The type of events this client wants to receive.
		 */
		short getType()
		{
			return _type;
		}

		/**
		 * Verifies if a field in the given event matches one of the fields on which the listener
		 * wants to receive modification events.
		 *
		 * @param  event Checked event.
		 * @return True if the listener needs to be notified, false otherwise.
		 */
		boolean matches(LyDataSourceEvent event)
		{
			if ((event == null)
					|| (event.getType() != LyDataSourceEvent.TYPE_SET)
					|| ((_type != LyDataSourceEvent.TYPE_ALL)
						&& (_type != LyDataSourceEvent.TYPE_SET)))
				return false;

			if (_fields == null)
				return true;

			LyValueSet values = event.getValues();

			if (values == null)
				return true;

			LyField field;
			LyValue value;
			int     i, n = values.getSize();

			for (i = 0; i < n; i++)
			{
				value = values.getFieldValue(i);
				field = value.getField();
				if (_fields.contains(field))
					return true;
			}

			return false;
		} // end method matches
	} // end class _LyDataSourceClient
} // end class LyPeerDataSource
