/*
 * LyRequestEncoder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer;

import java.util.*;
import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.*;
import leon.misc.*;


/**
 * This class is used to encode a request made on a data source and decode results to give a
 * response. This class is abstract and must be overriden to provide specific encoding and decoding.
 *
 * <p>The task of the request encoder is to process requests from a data source and to send
 * 'requests' using a connection instance. The request encoder class converts a data source demand
 * into a 'language' which is understood by the connection instance : the specific 'language' of the
 * underlying connection, like SQL, XML, method invocation on a Corba service.</p>
 *
 * <p>Additionally, the request encoder instance must 'decode' or interpret responses from the
 * specific connection to a 'language' understood by Leonardi. Basically, this class implements the
 * mapping between the chosen physical layer and the data representation in the specific connection.
 * </p>
 *
 * @see LyDataSource
 * @see LyResponse
 */
public abstract class LyRequestEncoder
{
	public static final String __VERSION =
		"$Id: LyRequestEncoder.java 29006 2010-12-23 17:44:46Z bmichel $";

	/**
	 * Converts an Object to a string.
	 *
	 * @param  object The object that must be converted.
	 * @return The string that represents the object or null if the object is null.
	 */
	public static String toString(Object object)
	{
		if (object == null)
			return null;
		if (object.getClass().isArray())
			return toString((Object[])object);
		if (object instanceof Vector<?>)
			return toString((Vector<?>)object);
		if (object instanceof Date)
			return LyTools.toString((Date)object);
		if (object instanceof BitSet)
			return LyTools.toString((BitSet)object);
		if (object instanceof LyObjectId)
			return ((LyObjectId)object).getFullString();
		else
			return object.toString();
	}

	/**
	 * Converts an Array to a string.
	 *
	 * @param  array The Array that must be converted.
	 * @return The string that represents the array.
	 */
	public static String toString(Object[] array)
	{
		StringBuffer buffer = new StringBuffer();
		int			 i, limit = array.length;

		buffer.append(LyTools.ARRAY_SEP);
		for (i = 0; i < limit; i++)
		{
			if (i > 0)
				buffer.append(LyTools.ARRAY_SEP);
			if (array[i] != null)
				buffer.append(toString(array[i]));
		}

		buffer.append(LyTools.ARRAY_SEP);
		return buffer.toString();
	}

	/**
	 * Converts a Vector to a string.
	 *
	 * @param  vector The vector that must be converted.
	 * @return The string that represents the vector.
	 */
	public static String toString(Vector<?> vector)
	{
		StringBuffer buffer = new StringBuffer();
		int			 i, limit = vector.size();

		buffer.append(LyTools.VECTOR_SEP);
		for (i = 0; i < limit; i++)
		{
			if (i > 0)
				buffer.append(LyTools.VECTOR_SEP);
			buffer.append(toString(vector.elementAt(i)));
		}

		buffer.append(LyTools.VECTOR_SEP);
		return buffer.toString();
	}

	/** Provider that created this request encoder. */
	protected LyDataProvider		  _provider;

	/** Data source client of the request. */
	protected LyDataSource			  _dataSource;

	/** Object class managed by the client data source. */
	protected LyClass				  _class;

	/**
	 * Binding between logical class and physical class. This binding may contain extra information
	 * required for encoding the request (such as extra filter).
	 */
	protected LyClassBinding		  _classBinding;

	/** Response used to return results to the data provider. */
	protected LyResponse			  _response;

	/** Boolean indicating if this request encoder is synchronous or not. */
	protected boolean				  _synchronous;

	/** Cache to keep the fields with the NOT_NULL mark. */
	protected LyInfoList<LyFieldInfo> _fieldsNotNull;

	/**
	 * Constructs a request encoder. Initialization is done with the init method.
	 *
	 * @param provider   Provider that creates this request encoder.
	 * @param dataSource Data source client of the request.
	 */
	public LyRequestEncoder(LyDataProvider provider, LyDataSource dataSource)
	{
		_provider = provider;
		_dataSource = dataSource;
		_class = dataSource.getDataClass();
		init();
	}

	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		// Initialize class attributes.
		_response = null;
		_classBinding = null;
		_fieldsNotNull = null;

		// Asynchronous by default
		_synchronous = false;
	}

	/**
	 * Gets the provider that created this request encoder.
	 *
	 * @return Provider of this request encoder.
	 */
	public final LyDataProvider getProvider()
	{
		return _provider;
	}

	/**
	 * Check if the request encoder is synchronous.
	 *
	 * @return True if the request is synchronous, false otherwise.
	 */
	public final boolean isSynchronous()
	{
		return _synchronous;
	}

	/**
	 * Get the response in which results are returned.
	 *
	 * @return Response of the request encode by this request encoder.
	 */
	public final LyResponse getResponse()
	{
		return _response;
	}

	/**
	 * Utility method to get current environment.
	 *
	 * @return Current environment.
	 */
	protected LyEnvironment getEnvironment()
	{
		LyApplication application = _provider.getApplication();

		return application.getEnvironment();
	}

	/**
	 * Gets the binding between logical class and physical class. Warning ! Trasmitting the binding
	 * is a workaround to retrieve information on class filter and may be suppressed in further
	 * version.
	 *
	 * @return Binding between logical class and physical class
	 */
	public final LyClassBinding getClassBinding()
	{
		return _classBinding;
	}

	/**
	 * Set the synchronous mode. Depending on the underlying technology, asynchronous or synchronous
	 * mode may be not supported.
	 *
	 * @param synchronous True if synchronous, false otherwise.
	 */
	protected void setSynchronous(boolean synchronous)
	{
		_synchronous = synchronous;
	}

	/**
	 * Set the request encoder response.
	 *
	 * @param     response Response in which results are returned.
	 * @exception IllegalArgumentException If the given response is null.
	 * @exception IllegalStateException    If this request encoder has already a response.
	 */
	public void setResponse(LyResponse response)
	{
		if (response == null)
			throw new IllegalArgumentException("BF AppEngine : response parameter can't be null.");
		if (_response != null)
			throw new IllegalStateException(
				"BF AppEngine : Attempt to set a response on a request encoder that already has one");

		_response = response;
	}

	/**
	 * Sets the binding between logical class and physical class. Warning ! Trasmitting this binding
	 * is a workaround to retrieve information on class filter and may be suppressed in further
	 * version.
	 *
	 * @param classBinding Binding between logical class and physical class
	 */
	public final void setClassBinding(LyClassBinding classBinding)
	{
		_classBinding = classBinding;
	}

	/* Methods for accessing objects */

	/**
	 * Load an object from an object id. This method forces the request to be sent to external data
	 * provider. Default implementation of this method is to call getObject(LyObjectId, Vector).
	 *
	 * @param  objectId Object identifier.
	 * @param  fields   Fields to load (if null, load all fields).
	 * @return True if the request was correctly sent to the external data provider, false
	 *         otherwise.
	 */
	public boolean loadObject(LyObjectId objectId, Vector<LyField> fields)
	{
		return getObject(objectId, fields);
	}

	/**
	 * Load an object from an object id. If the object is already in memory, returns this object and
	 * doesn't sent the request to the external data provider.
	 *
	 * @param  objectId Object identifier.
	 * @param  fields   Fields to load (if null, load all fields).
	 * @return True if the request was correctly sent to the external data provider, false
	 *         otherwise.
	 */
	public abstract boolean getObject(LyObjectId objectId, Vector<LyField> fields);

	/**
	 * Load a number of objects matching a given filter and sorted by a specified sort.
	 *
	 * @param  fields Fields to load (if null, load all fields).
	 * @param  sort   Gives sorting criterion, or null if default sort is used.
	 * @param  filter Specifies filtering criterion, or null if there is no filter.
	 * @param  n      Maximum number of objects to get (-1 means no limit).
	 * @return True if the request was correctly sent to the external data provider, false
	 *         otherwise.
	 */
	public abstract boolean getObjects(Vector<LyField> fields, LySort sort, LyFilter filter, int n);

	/**
	 * Get the number of objects that match a given filter.
	 *
	 * @param  filter Specifies filtering criterion, or null if there is no filter.
	 * @return True if the request was correctly sent to the external data provider, false
	 *         otherwise.
	 */
	public abstract boolean getSize(LyFilter filter);

	/* Methods for deleting objects */

	/**
	 * Delete an object.
	 *
	 * @param  object Object to delete.
	 * @return True if the request was correctly sent to the external data provider, false
	 *         otherwise.
	 */
	public abstract boolean deleteObject(Object object);

	/**
	 * To delete a list of objects.
	 *
	 * @param  objects Objects to delete.
	 * @return True if the request was correctly sent to the external data provider, false
	 *         otherwise.
	 */
	public boolean deleteObjects(Vector<Object> objects)
	{
		Object object;
		int    i, n = objects.size();

		for (i = 0; i < n; i++)
		{
			object = objects.elementAt(i);
			deleteObject(object);
		}

		return true;
	}

	/* Methods for creating objects */

	/**
	 * Create a new object.
	 *
	 * @param  object New object to create.
	 * @return True if the request was correctly sent to the external data provider, false
	 *         otherwise.
	 */
	public abstract boolean createObject(Object object);

	/* Methods for modifying objects */

	/**
	 * To set an object.
	 *
	 * @param  object Object to set.
	 * @return True if the request was correctly sent to the external data provider, false
	 *         otherwise.
	 */
	public abstract boolean setObject(Object object);

	/**
	 * To set a list of objects.
	 *
	 * @param  objects List of objects to set.
	 * @return True if the request was correctly sent to the external data provider, false
	 *         otherwise.
	 */
	public boolean setObjects(Vector<Object> objects)
	{
		Object object;
		int    i, n = objects.size();

		for (i = 0; i < n; i++)
		{
			object = objects.elementAt(i);
			setObject(object);
		}

		return true;
	}

	/**
	 * This method add to the specified object all the values for fieldInfos that have the mark
	 * NOT_NULL.
	 *
	 * @param object    the object to add the values
	 * @param onlyLocal spcifies if only local values are concerned.
	 */
	protected void addNotNullValues(LyObject object, boolean onlyLocal)
	{
		// We must verify that all fieldInfos that have the mark NOT_NULL
		// have a value.
		LyClassInfo classInfo = object.getClassInfo();
		LyFieldInfo field;
		int		    i;
		Object	    value, defValue;

		if (_fieldsNotNull == null)
			_fieldsNotNull = classInfo.getFieldInfos(LyFieldInfo.MARK_NOT_NULL);

		for (i = 0; i < _fieldsNotNull.getSize(); i++)
		{
			field = _fieldsNotNull.getInfo(i);

			if (onlyLocal && !field.isLocal())
				continue;

			value = classInfo.getValue(object, field);
			if (value == null)
			{
				defValue = field.getDefaultValue();

				//FFT00360
				// On the current session, we can try to translate the defValue.
				if ((defValue != null)
						&& (defValue instanceof String)
						&& (((String)defValue).charAt(0) == '$'))
				{
					LyApplication application = _provider.getApplication();
					LyDataSession session = application.getCurrentSession();

					if (session != null)
					{
						defValue = session.getValue(defValue);

						// We try to get a correct value for the current field
						if (defValue != null)
							defValue = field.buildFieldValue(defValue);
					}
				}

				if (defValue == null)
				{
					LyEnvironment environment = getEnvironment();

					environment.logWarning("BF AppEngine : Attempt to build not null value for class "
						+ classInfo.getId() + " on a field " + field.getId()
						+ " that doesn't have a default value.");
				}
				else
					classInfo.setValue(object, field, defValue);
			}
		}
	} // end method addNotNullValues

	/**
	 * This method removes to the specified object all the values for fieldInfos that have the mark
	 * PROVIDER_VALUE.
	 *
	 * @param object    the object to add the values
	 * @param onlyLocal specifies if only local values are concerned.
	 */
	protected void removeProviderValues(LyObject object, boolean onlyLocal)
	{
		// We must verify that all fieldInfos that have the mark PROVIDER_VALUE
		// haven't any value.
		LyClassInfo			    classInfo = object.getClassInfo();
		LyFieldInfo			    field;
		int					    i;
		Object				    value;
		LyInfoList<LyFieldInfo> fieldsValuedByProvider;

		fieldsValuedByProvider = classInfo.getFieldInfos(LyFieldInfo.MARK_PROVIDER_VALUE);

		for (i = 0; i < fieldsValuedByProvider.getSize(); i++)
		{
			field = fieldsValuedByProvider.getInfo(i);
			if (onlyLocal && !field.isLocal())
				continue;

			value = classInfo.getValue(object, field);
			if (value != null)
			{
				object.removeNewValue(field);
				object.removePendingValue(field);
			}
		}
	}

	/**
	 * Sets the physical data class of the request encoder. Warning ! Internal use only.
	 *
	 * @param classBinding Binding between logical class and physical class
	 */
	public final void setDataClass(LyClass c)
	{
		_class = c;
	}

	/**
	 * Accessor to the physical data class of the request encoder.
	 *
	 * @return c physical class
	 */
	public final LyClass getDataClass()
	{
		return _class;
	}
} // end class LyRequestEncoder
