/*
 * LyTransaction.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer;

import java.io.*;
import java.rmi.server.*;
import java.util.*;
import javax.naming.*;
import javax.naming.spi.*;
import javax.transaction.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.*;
import leon.info.location.*;
import leon.misc.*;
import leon.notifier.*;
import leon.peer.union.*;


/**
 * This class represents a transaction managed by the transaction manager. A transaction is
 * associated with a user session (LyDataSession) and a data provider (LyDataProvider). When a
 * transaction is open, all requests for the data provider in the user session are stored in the
 * transaction.
 *
 * @see    LyDataProvider
 * @see    LyRequestEncoder
 * @see    LyConnection
 * @see    LyTransactionManager
 * @author Lyria
 */
public class LyTransaction
{
	public static final String __VERSION =
		"$Id: LyTransaction.java 29535 2011-02-08 13:50:18Z bangiole $";

	/** Statuses of the transaction. */
	public enum Status // The transaction is started but there is no operation registered.
	{
		STARTED,
		// The transaction is started and there is one or more operations registered.
		RUNNING,
		// The transaction is rollbacked.
		ERROR,
		// The transaction is committed.
		COMPLETED
	}

	/** Operation types. */
	public enum Operation // Creation type of an operation or notification registered in the transaction.
	{
		CREATE,
		// Modification type of an operation or notification registered in the transaction.
		SET,
		// Deletion type of an operation or notification registered in the transaction.
		DELETE,
		// Loading type of an operation or notification registered in the transaction.
		LOAD,
		// Sizing type of an operation or notification registered in the transaction.
		SIZE
	}

	/** Constant for user transaction name parameter in the property file. */
	public static final String			  USER_TRANSACTION_NAME = "userTransactionName";

	/** Constant for user transaction timeout parameter in the property file. */
	public static final String			  TRANSACTION_TIMEOUT = "transactionTimeout";

	/** Reference to the current thread at the creation of this transaction. */
	protected Thread					  _thread;

	/** Reference to the application. */
	protected LyApplication				  _application;

	/** Reference to the user session. */
	protected LyDataSession				  _session;

	/** List of data providers used in this transaction. */
	protected Vector<LyDataProvider>	  _providers;

	/**
	 * List of connections used in this transaction. Connections have the same index than the
	 * provider they come from.
	 */
	protected Vector<LyConnection>		  _connections;

	/** Current status of the transaction. */
	protected LyTransaction.Status		  _status;

	/** Registered operations on this transaction. Used to rollback operation. */
	protected LinkedList<_LyOperation>    _operations;

	/** Registered notifications on this transaction. Used to commit operation. */
	protected LinkedList<_LyNotification> _eventTypes;

	/** Associated user transaction to manage a 2PC. */
	protected UserTransaction			  _ut;

	/** Unique identifier of the user transaction. */
	protected UID						  _uid;

	/** List of notifiers associated with this transaction : used for grouping notifications. */
	protected Vector<LyEventNotifier>     _notifiers;

	/** Boolean indicating if transaction has already been committed. */
	protected boolean					  _committed;

	/** Boolean indicating if transaction has already been rollbacked. */
	protected boolean					  _rollbacked;

	/**
	 * Constructs a transaction. Initialization is done with the init method.
	 *
	 * @param session User session associated with the transaction.
	 */
	protected LyTransaction(LyDataSession session)
	{
		init();

		if (session == null)
			throw new IllegalArgumentException("BF AppEngine : session parameter can't be null.");

		_application = session.getApplication().getRootApplication();
		_session = session;
		_thread = Thread.currentThread();
	}

	/**
	 * Constructs a transaction. Initialization is done with the init method.
	 *
	 * @param application Application for which the transaction is created. The created application
	 *                    is associated to the current session of this application.
	 */
	protected LyTransaction(LyApplication application)
	{
		init();

		if (application == null)
			throw new IllegalArgumentException("BF AppEngine : application parameter can't be null.");

		_application = application.getRootApplication();
		_session = application.getCurrentSession();
		_thread = Thread.currentThread();
	}

	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		_application = null;
		_session = null;
		_providers = new Vector<LyDataProvider>();
		_connections = new Vector<LyConnection>();
		_status = Status.STARTED;
		_operations = new LinkedList<_LyOperation>();
		_eventTypes = new LinkedList<_LyNotification>();
		_notifiers = null;
	}

	/**
	 * Accessor to the current status.
	 *
	 * @return the current status
	 */
	public final Status getStatus()
	{
		return _status;
	}

	/**
	 * Accessor to the user parent application.
	 *
	 * @return the parent application
	 */
	public final LyApplication getApplication()
	{
		return _application;
	}

	/**
	 * Accessor to the associated user session.
	 *
	 * @return the associated session
	 */
	public final LyDataSession getSession()
	{
		if (_session == null)
			return _application.getCurrentSession();
		return _session;
	}

	/**
	 * Accessor to the environment of this transaction.
	 *
	 * @return the corresponding environment (session or application environment)
	 */
	public final LyEnvironment getEnvironment()
	{
		if (_session != null)
			return _session.getEnvironment();
		else
			return _application.getEnvironment();
	}

	/**
	 * Accessor to the associated thread.
	 *
	 * @return the associated thread
	 */
	public final Thread getThread()
	{
		return _thread;
	}

	/**
	 * Add a data provider to this transaction.
	 *
	 * @param provider the added data provider
	 */
	public final void addDataProvider(LyDataProvider provider)
	{
		if (!(provider instanceof LyPeerDataProvider))
			return;

		if ((!_providers.contains(provider)) && (_providers.size() != 0))
		{
			LyEnvironment environment = provider.getEnvironment();

			if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
				environment.logWarning(
					"Transactions on more than one data provider are not supported."
					+ " The transactions will be managed in an independent way.");
		}

		if (!_providers.contains(provider))
		{
			_providers.addElement(provider);
			_connections.addElement(null);
		}
	}

	/**
	 * This method searches for a property file containing informations about a UserTRansaction.
	 *
	 * @return the properties defining a UserTransaction for this manager.
	 */
	protected Properties getProperties()
	{
		String extension, dir;
		String appId = _application.getId();
		int    p = appId.lastIndexOf('.');

		if (p < 0)
		{
			extension = appId;
			dir = appId;
		}
		else
		{
			extension = appId.substring(p + 1);
			dir = appId.replace('.', '/');
		}

		String		  filename = dir + "/properties/transaction_"
			+ extension + ".properties";
		LyEnvironment environment = _application.getEnvironment();
		InputStream   stream = environment.getResourceAsStream(filename,
				_application.getClassLoader());

		if (stream == null)
			return null;

		Properties result = new Properties();

		try
		{
			result.load(stream);
		}
		catch (Exception e)
		{
			if (_session != null)
				_session.handleException(e);
			else
				environment.writeException(e);
			result = null;
		}

		return result;
	} // end method getProperties

	/**
	 * Internal method to get the unique identifier of the user transaction.
	 *
	 * @return The unique identifier of the user transaction.
	 */
	public UID getUID()
	{
		return _uid;
	}

	/**
	 * Internal method to get a UserTransaction to manage a 2PC.
	 *
	 * @return a user transaction if the manager is configured to use a 2PC.
	 */
	protected UserTransaction getUserTransaction()
	{
		if (_ut != null)
			return _ut;

		LyTransactionManager tm = _application.getTransactionManager();

		if (!tm.hasUt())
			return null;

		Properties properties = getProperties();

		if (properties != null)
		{
			// Replace properties which may be set in the environment
			// These properties starts with $NAME and may be set in the ini file
			Enumeration<?>	names = properties.propertyNames();
			LyEnvironment	environment = _application.getEnvironment();
			String		 	name, val;

			while (names.hasMoreElements())
			{
				name = (String)names.nextElement();
				val = properties.getProperty(name);

				if ((val != null) && (val.startsWith("$")))
				{
					val = environment.getEnv(val);
					properties.setProperty(name, val);
				}
			}

			if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
			{
				environment.logOut("* LyTransaction : found properties");
				properties.list(environment.out);
			}

			UserTransaction userTransaction = null;
			String		    userTransactionName = properties.getProperty(USER_TRANSACTION_NAME);

			if (userTransactionName != null)
				properties.remove(USER_TRANSACTION_NAME);
			else
				userTransactionName = "UserTransaction";

			String transactionTimeout = properties.getProperty(TRANSACTION_TIMEOUT);
			int    timeout = -1;

			if (transactionTimeout != null)
			{
				properties.remove(TRANSACTION_TIMEOUT);
				try
				{
					timeout = Integer.parseInt(transactionTimeout);
				}
				catch (Exception e)
				{
					environment.logError("Invalid timeout for transaction " + transactionTimeout);
					timeout = -1;
				}
			}

			try
			{
				// Get the initial context of the naming service
				InitialContext namingService = new InitialContext(properties);

				// Build the name of the dataSource.
				NameParser     nameParser = namingService.getNameParser("");
				Name		   utName = nameParser.parse(userTransactionName);

				// Lookup the topic factory
				Object		   ut = namingService.lookup(utName);

				if (ut instanceof UserTransaction)
					userTransaction = (UserTransaction)ut;
				else if (ut instanceof Reference)
				{
					// This part of code is a copy from MySQL sample
					// For some reason, this comes back as a Reference
					// instance under CruiseControl !?
					Reference     objAsRef = (Reference)ut;
					Class<?>	  cls = Class.forName(objAsRef.getFactoryClassName());
					ObjectFactory factory = (ObjectFactory)cls.newInstance();

					userTransaction = (UserTransaction)factory.getObjectInstance(objAsRef, utName,
							namingService, new Hashtable<Object, Object>());
				}
			}
			catch (Exception e)
			{
				environment.logInfo("Can't get a valid user transaction for name "
					+ userTransactionName);
				if (environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
					environment.writeException(e);
				userTransaction = null;
			}

			if ((userTransaction != null) && (timeout > 0))
			{
				try
				{
					userTransaction.setTransactionTimeout(timeout);
				}
				catch (Exception e)
				{
					environment.logError("Can't set timeout to user transaction");
				}
			}

			return userTransaction;
		}
		tm.hasUt(false);
		return null;
	} // end method getUserTransaction

	/**
	 * To begin this transaction.
	 */
	public void begin()
	{
		// This is the first transaction of the list.
		// Get a user transaction if we use JTA.
		UserTransaction ut = getUserTransaction();

		if (ut != null)
		{
			// Make an association between ut and the transaction ?
			try
			{
				ut.begin();
			}
			catch (Exception e)
			{
				LyEnvironment environment = getEnvironment();

				environment.logError("Can't begin user transaction");
			}
			UID uid = new UID();

			setUserTransaction(ut, uid);
		}
	}

	/**
	 * This method is called before a commit operation. For each connection of this transaction, the
	 * method prepareCommit is called and if at least one of the connections can't commit, this
	 * method stops and returns false.
	 *
	 * @return true if the commit can be made, else returns false.
	 */
	public boolean prepareCommit()
	{
		int			 i, n = _providers.size();
		LyConnection connection;
		boolean		 result = true;

		for (i = 0; (i < n) && result; i++)
		{
			connection = _connections.elementAt(i);

			if (connection != null)
				result = connection.prepareCommit();
		}

		return result;
	}

	/**
	 * To commit the transaction. All registered notifications are sent.
	 */
	public void commit()
	{
		if (_committed)
			return; // Already committed
		if (_rollbacked)
			throw new IllegalStateException(
				"BF AppEngine : can't commit a transaction that has been rollbacked.");

		if (_ut != null)
		{
			twoPhasesCommit();
			return;
		}

		int							   i, n = _providers.size();
		LyDataProvider				   provider;
		LyConnection				   connection;
		Vector<String>				   connectionIds = new Vector<String>();
		Vector<String>				   locationIds = new Vector<String>();
		LyApplicationBehaviorInterface behavior = _application.getApplicationBehavior();

		for (i = 0; i < n; i++)
		{
			provider = _providers.elementAt(i);
			connection = _connections.elementAt(i);

			if (connection != null)
			{
				UID		   uid = connection.getUID();
				LyLocation location = provider.getLocation();

				connectionIds.add(uid.toString());
				locationIds.add(location.getId());

				try
				{
					// Notify beforeCommit to the main application
					behavior.beforeCommit(_application, uid.toString(), location.getId());
				}
				catch (Exception e)
				{
					LyEnvironment environment = getEnvironment();

					environment.logError(
						"BF AppEngine : exception caught on behavior when calling beforeCommit");
					environment.writeException(e);
				}

				connection.commit();
				if (connection.getError() != null)
				{
					if (!_rollbacked)
						// If an error occured during commit, rollback if not already done
						rollback();
					_status = Status.ERROR;
					return;
				}

				try
				{
					// Notify afterCommit to the main application
					behavior.afterCommit(_application, uid.toString(), location.getId());
				}
				catch (Exception e)
				{
					LyEnvironment environment = getEnvironment();

					environment.logError(
						"BF AppEngine : exception caught on behavior when calling afterCommit");
					environment.writeException(e);
				}

				provider.releaseConnection(connection, true);
			}
		}

		_status = Status.COMPLETED;
		_committed = true;

		// Now we can post notifications.
		_LyNotification notification;
		Iterator<_LyNotification>	it;

		it = _eventTypes.iterator();
		while (it.hasNext())
		{
			notification = it.next();
			it.remove();
			postNotification(notification);
		}

		if (_notifiers != null)
		{
			LyEventNotifier notifier;

			for (i = 0; i < _notifiers.size(); i++)
			{
				notifier = _notifiers.elementAt(i);
				notifier.sendGroupedEvents(_uid.toString());
			}
			_notifiers = null;
		}

		try
		{
			behavior.transactionCompleted(_application, connectionIds, locationIds);
		}
		catch (Exception e)
		{
			LyEnvironment environment = getEnvironment();

			environment.logError(
				"BF AppEngine : exception caught on behavior when calling transactionCompleted");
			environment.writeException(e);
		}
	} // end method commit

	/**
	 * Internal method to commit this transaction in case of a 2PC.
	 */
	protected void twoPhasesCommit()
	{
		// We can post notifications for XA notifiers.
		_LyNotification notification;
		LyEventNotifier notifier;
		Iterator<_LyNotification>	it;

		it = _eventTypes.iterator();
		while (it.hasNext())
		{
			notification = it.next();
			notifier = notification.getNotifier();
			if (!notifier.isXASupported())
				continue;
			it.remove();
			postNotification(notification);
		}

		// Notify beforeRollback to the main application
		UID							   uid;
		LyLocation					   location;
		LyApplicationBehaviorInterface behavior = _application.getApplicationBehavior();
		int							   i, n = _providers.size();
		LyDataProvider				   provider;
		LyConnection				   connection;

		for (i = 0; i < n; i++)
		{
			provider = _providers.elementAt(i);
			connection = _connections.elementAt(i);

			if (connection != null)
			{
				// Notify the afterRollback to the main application.
				uid = connection.getUID();
				location = provider.getLocation();

				try
				{
					behavior.beforeCommit(_application, uid.toString(), location.getId());
				}
				catch (Exception e)
				{
					LyEnvironment environment = getEnvironment();

					environment.logError(
						"BF AppEngine : exception caught on behavior when calling afterCommit");
					environment.writeException(e);
				}
			}
		}

		if (_notifiers != null)
		{
			for (i = 0; i < _notifiers.size(); i++)
			{
				notifier = _notifiers.elementAt(i);
				notifier.sendGroupedEvents(_uid.toString());
			}
			_notifiers = null;
		}

		if (_ut != null)
		{
			try
			{
				_ut.commit();
			}
			catch (Exception e)
			{
				LyEnvironment environment = getEnvironment();

				environment.logError("Can't commit user transaction : " + e.getMessage());
			}
		}

		for (i = 0; i < n; i++)
		{
			provider = _providers.elementAt(i);
			connection = _connections.elementAt(i);

			if (connection != null)
			{
				// Notify afterCommit to the main application
				uid = connection.getUID();
				location = provider.getLocation();
				try
				{
					behavior.afterCommit(_application, uid.toString(), location.getId());
				}
				catch (Exception e)
				{
					LyEnvironment environment = getEnvironment();

					environment.logError(
						"BF AppEngine : exception caught on behavior when calling afterCommit");
					environment.writeException(e);
				}

				provider.releaseConnection(connection, true);
			}
		}

		// Post notification for non XA notifiers (remainding ones)
		it = _eventTypes.iterator();
		while (it.hasNext())
		{
			notification = (_LyNotification)it.next();
			it.remove();
			postNotification(notification);
		}

		if (_notifiers != null)
		{
			for (i = 0; i < _notifiers.size(); i++)
			{
				notifier = _notifiers.elementAt(i);
				notifier.sendGroupedEvents(_uid.toString());
			}
			_notifiers = null;
		}
	} // end method twoPhasesCommit

	/**
	 * Internal method to post a notification.
	 *
	 * @param notification The notification to post.
	 */
	protected void postNotification(_LyNotification notification)
	{
		LyEventNotifier.EventType eventType;
		LyEventNotifier			  notifier;
		LyEventEncoder			  encoder;
		LyValueSet				  values;
		LyClass					  peerClass;
		LyDataSource			  source;
		Object					  object;

		notifier = notification.getNotifier();
		eventType = notification.getNotification();
		object = notification.getObject();
		values = notification.getValues();
		peerClass = notification.getPeerClass();
		source = peerClass.getDataSource();

		if (notifier.supportsGroupedEvents())
		{
			encoder = notifier.getEventEncoder(source);
			if (_uid == null)
				_uid = new UID();
			encoder.postGroupedEvent(_uid.toString(), eventType, object, values);
			if (_notifiers == null)
				_notifiers = new Vector<LyEventNotifier>();
			if (!_notifiers.contains(notifier))
				_notifiers.addElement(notifier);
		}
		else if (eventType == LyEventNotifier.EventType.CREATE)
		{
			encoder = notifier.getEventEncoder(source);
			encoder.postCreation(object);
		}
		else if (eventType == LyEventNotifier.EventType.UPDATE)
		{
			encoder = notifier.getEventEncoder(source);
			encoder.postModification(object, values);
		}
		else if (eventType == LyEventNotifier.EventType.DELETE)
		{
			encoder = notifier.getEventEncoder(source);
			encoder.postDeletion(object);
		}
	} // end method postNotification

	/**
	 * To rollback the transaction. Return to the initial state before starting the transaction.
	 */
	public void rollback()
	{
		if (_rollbacked)
			return; // Already committed
		if (_committed)
			throw new IllegalStateException(
				"BF AppEngine : can't rollback a transaction that has been committed.");

		if (_ut != null)
		{
			twoPhasesRollback();
			return;
		}

		int			   i, n = _providers.size();
		LyDataProvider provider;
		LyConnection   connection;

		for (i = 0; i < n; i++)
		{
			provider = _providers.elementAt(i);
			connection = _connections.elementAt(i);

			if (connection != null)
			{
				// Notify beforeRollback to the main application
				UID							   uid = connection.getUID();
				LyLocation					   location = provider.getLocation();
				LyApplicationBehaviorInterface behavior;

				behavior = _application.getApplicationBehavior();
				try
				{
					behavior.beforeRollback(_application, uid.toString(), location.getId());
				}
				catch (Exception e)
				{
					LyEnvironment environment = getEnvironment();

					environment.logError(
						"BF AppEngine : exception caught on behavior when calling afterCommit");
					environment.writeException(e);
				}

				connection.rollback();

				// Notify the afterRollback to the main application
				try
				{
					behavior.afterRollback(_application, uid.toString(), location.getId());
				}
				catch (Exception e)
				{
					LyEnvironment environment = getEnvironment();

					environment.logError(
						"BF AppEngine : exception caught on behavior when calling afterCommit");
					environment.writeException(e);
				}

				if (connection.isReserved())
					provider.releaseConnection(connection, true);
			}
		}

		_status = Status.ERROR;
		_rollbacked = true;

		// We must now return to the original state.
		_LyOperation operation;
		Iterator<_LyOperation>	it;

		it = _operations.iterator();
		while (it.hasNext())
		{
			operation = it.next();
			it.remove();
			operation.rollback();
		}

	} // end method rollback

	/**
	 * Internal method to rollback this transaction in case of a 2PC.
	 */
	protected void twoPhasesRollback()
	{
		int							   i, n = _providers.size();
		LyDataProvider				   provider;
		LyConnection				   connection;

		// Notify beforeRollback to the main application
		UID							   uid;
		LyLocation					   location;
		LyApplicationBehaviorInterface behavior;

		behavior = _application.getApplicationBehavior();

		for (i = 0; i < n; i++)
		{
			provider = _providers.elementAt(i);
			connection = _connections.elementAt(i);

			if (connection != null)
			{
				// Notify the afterRollback to the main application.
				uid = connection.getUID();
				location = provider.getLocation();

				try
				{
					behavior.beforeRollback(_application, uid.toString(), location.getId());
				}
				catch (Exception e)
				{
					LyEnvironment environment = getEnvironment();

					environment.logError(
						"BF AppEngine : exception caught on behavior when calling afterCommit");
					environment.writeException(e);
				}
			}
		}

		if (_ut != null)
		{
			try
			{
				_ut.rollback();
			}
			catch (Exception e)
			{
				LyEnvironment environment = getEnvironment();

				environment.logError("Can't rollback user transaction");
			}
		}

		for (i = 0; i < n; i++)
		{
			provider = _providers.elementAt(i);
			connection = _connections.elementAt(i);

			if (connection != null)
			{
				// Notify the afterRollback to the main application.
				uid = connection.getUID();
				location = provider.getLocation();
				try
				{
					behavior.afterRollback(_application, uid.toString(), location.getId());
				}
				catch (Exception e)
				{
					LyEnvironment environment = getEnvironment();

					environment.logError(
						"BF AppEngine : exception caught on behavior when calling afterCommit");
					environment.writeException(e);
				}

				provider.releaseConnection(connection, true);
			}
		}

		_status = Status.ERROR;
		_rollbacked = true;

		// We must now return to the original state.
		_LyOperation operation;
		Iterator<_LyOperation>	it;

		it = _operations.iterator();
		while (it.hasNext())
		{
			operation = it.next();
			it.remove();
			operation.rollback();
		}
	} // end method twoPhasesRollback

	/**
	 * To get the connection associated with this transaction and the specified provider.
	 *
	 * @param  provider the provider to get the connection
	 * @return a connection from the specified provider with auto commit mode disabled.
	 */
	public LyConnection getConnection(LyDataProvider provider)
	{
		addDataProvider(provider);

		int			 idx = _providers.indexOf(provider);
		LyConnection connection = _connections.elementAt(idx);

		if (connection == null)
		{
			connection = provider.getConnection(true);
			if (connection == null)
				return null;

			connection.autoCommit(false);

			// Gives an id of transaction to this connection to enable
			// it to notify the main application for updates.
			UID uid;

			if (!provider.isXASupported())
			{
				uid = new UID();
				connection.setUID(uid);
			}
			else
				uid = _uid;

			// Notify the begin of the transaction to the main application
			LyLocation					   location = provider.getLocation();
			LyApplicationBehaviorInterface behavior;

			behavior = _application.getApplicationBehavior();
			try
			{
				behavior.begin(_application, uid.toString(), location.getId());
			}
			catch (Exception e)
			{
				LyEnvironment environment = getEnvironment();

				environment.logError(
					"BF AppEngine : exception caught on behavior when calling afterCommit");
				environment.writeException(e);
			}

			_connections.setElementAt(connection, idx);
		}
		return connection;
	} // end method getConnection

	/**
	 * Add the provider of the specified data source to this transaction.
	 *
	 * @param dataSource the data source to add the provider.
	 */
	protected void addDataProvider(LyDataSource dataSource)
	{
		LyDataProvider provider = dataSource.getDataProvider();

		addDataProvider(provider);
	}

	/**
	 * Delete an object on the specified data source.
	 *
	 * @param  dataSource   The data source for deleting the object.
	 * @param  peer         Object to delete.
	 * @param  listener     Listener of the response.
	 * @param  classBinding Binding between logical class and physical class This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the deletion.
	 * @see    LyDataSource#deleteObject
	 */
	public LyResponse deleteObject(LyDataSource dataSource, Object peer,
		LyResponseListener listener, LyClassBinding classBinding)
	{
		addDataProvider(dataSource);

		LyResponse   response;
		_LyOperation operation;

		operation = new _LyOperation(Operation.DELETE, dataSource, peer, listener, classBinding);
		addOperation(operation);

		response = operation.send();

		return response;
	}

	/**
	 * Delete a list of objects on the specified data source.
	 *
	 * @param  dataSource   The data source for deleting the objects.
	 * @param  peers        List of objects to delete.
	 * @param  listener     Listener of the response.
	 * @param  classBinding Binding between logical class and physical class This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the deletion.
	 * @see    LyDataSource#deleteObjects
	 */
	public LyResponse deleteObjects(LyDataSource dataSource, Vector<Object> peers,
		LyResponseListener listener, LyClassBinding classBinding)
	{
		addDataProvider(dataSource);

		LyResponse   response;
		_LyOperation operation;

		operation = new _LyOperation(Operation.DELETE, dataSource, peers, listener, classBinding);
		addOperation(operation);

		response = operation.send();

		return response;
	}

	/**
	 * Create a new object on the specified data source.
	 *
	 * @param  dataSource   The data source for creating the object.
	 * @param  peer         Object to create.
	 * @param  listener     Listener of the response.
	 * @param  classBinding Binding between logical class and physical class. This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the creation.
	 * @see    LyDataSource#createObject
	 */
	public LyResponse createObject(LyDataSource dataSource, Object peer,
		LyResponseListener listener, LyClassBinding classBinding)
	{
		addDataProvider(dataSource);

		LyResponse   response;
		_LyOperation operation;

		operation = new _LyOperation(Operation.CREATE, dataSource, peer, listener, classBinding);
		addOperation(operation);

		response = operation.send();

		return response;
	}

	/**
	 * Set an object on the specified data source.
	 *
	 * @param  dataSource   The data source for setting the object.
	 * @param  peer         Object to modify.
	 * @param  listener     Listener of the response.
	 * @param  classBinding Binding between logical class and physical class This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the modification.
	 * @see    LyDataSource#setObject
	 */
	public LyResponse setObject(LyDataSource dataSource, Object peer,
		LyResponseListener listener, LyClassBinding classBinding)
	{
		addDataProvider(dataSource);

		LyResponse   response;
		_LyOperation operation;

		operation = new _LyOperation(Operation.SET, dataSource, peer, listener, classBinding);
		addOperation(operation);

		response = operation.send();

		return response;
	}

	/**
	 * Set a list of objects on the specified data source.
	 *
	 * @param  dataSource   The data source for setting the objects.
	 * @param  peers        List of objects to modify.
	 * @param  listener     Listener of the response.
	 * @param  classBinding Binding between logical class and physical class This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the modification.
	 * @see    LyDataSource#setObjects
	 */
	public LyResponse setObjects(LyDataSource dataSource, Vector<Object> peers,
		LyResponseListener listener, LyClassBinding classBinding)
	{
		addDataProvider(dataSource);

		LyResponse   response;
		_LyOperation operation;

		operation = new _LyOperation(Operation.SET, dataSource, peers, listener, classBinding);
		addOperation(operation);

		response = operation.send();

		return response;
	}

	/**
	 * Method to load an object from its id on the specified data source. Load fields of an object
	 * from a given object id.
	 *
	 * @param  dataSource   The data source for loading the object.
	 * @param  peerId       Object identifier corresponding to the object to load.
	 * @param  peerFields   Fields to load (null means all fields).
	 * @param  listener     Listener of the response.
	 * @param  classBinding binding between logical class and physical class This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the loading.
	 * @see    LyDataSource#loadObject
	 */
	public LyResponse loadObject(LyDataSource dataSource, LyObjectId peerId,
		Vector<LyField> peerFields, LyResponseListener listener, LyClassBinding classBinding)
	{
		addDataProvider(dataSource);

		LyResponse   response;
		_LyOperation operation;

		operation = new _LyOperation(Operation.LOAD, dataSource, peerId, peerFields, listener,
				classBinding);
		addOperation(operation);

		response = operation.send();

		return response;
	}

	/**
	 * Method for loading objects on the specified data source. Load fields of a specified number of
	 * objects sorted by a given sort and that match a given filter.
	 *
	 * @param  dataSource   The data source for loading the objects.
	 * @param  peerFields   Fields to load (null means all fields).
	 * @param  peerSort     Gives sort criteria, or null if there is no sort (the order is the one
	 *                      of the physical layer).
	 * @param  peerFilter   Gives filter criteria, or null if there is no filter (all objects).
	 * @param  n            Indicates the number of objects to load (-1 means no limit).
	 * @param  listener     Listener of the response.
	 * @param  classBinding binding between logical class and physical class This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the loading.
	 * @see    LyDataSource#loadObjects
	 */
	public LyResponse loadObjects(LyDataSource dataSource, Vector<LyField> peerFields,
		LySort peerSort, LyFilter peerFilter, int n, LyResponseListener listener,
		LyClassBinding classBinding)
	{
		addDataProvider(dataSource);

		LyResponse   response;
		_LyOperation operation;

		operation = new _LyOperation(Operation.LOAD, dataSource, peerFields, peerSort, peerFilter,
				n, listener, classBinding);
		addOperation(operation);

		response = operation.send();

		return response;
	}

	/**
	 * Return the number of objects within the specified data source matching the given filter.
	 *
	 * @param  dataSource   The data source for getting the number of objects.
	 * @param  peerFilter   Filter applied on objects for counting.
	 * @param  classBinding Binding between logical class and physical class. This binding may
	 *                      contain extra information required for encoding the request (such as
	 *                      extra filter).
	 * @return Number of objects matching the given filter in this data source.
	 * @see    LyDataSource#getSize
	 */
	public int getSize(LyDataSource dataSource, LyFilter peerFilter,
		LyClassBinding classBinding)
	{
		addDataProvider(dataSource);

		int			 size;
		_LyOperation operation;

		operation = new _LyOperation(Operation.SIZE, dataSource, peerFilter, classBinding);
		addOperation(operation);

		size = operation.getSize();

		return size;
	}

	/**
	 * Add an operation in the transaction to do the rollback.
	 *
	 * @param operation the operation to add.
	 */
	private void addOperation(_LyOperation operation)
	{
		LyDataSource   dataSource = operation._dataSource;
		LyDataProvider provider = dataSource.getDataProvider();

		// Do not register operation for non physical layer.
		if ((provider instanceof LyPeerDataProvider) || (provider instanceof LyUnionDataProvider))
			_operations.addFirst(operation);
	}

	/**
	 * Internal method to register notifications on the transaction. For internal use only.
	 *
	 * @param notifier     the notifier to send the notification.
	 * @param notification the type of the added notification.
	 * @param peerClass    the class of the object.
	 * @param object       the object on which is the notification.
	 * @param values       old values of the object when the notification is a modification (else
	 *                     null).
	 */
	public void addNotification(LyEventNotifier notifier, LyEventNotifier.EventType notification,
		LyClass peerClass, Object object, LyValueSet values)
	{
		_LyNotification n = new _LyNotification(notifier, notification, peerClass, object, values);

		addNotification(n);
	}

	/**
	 * To add a notification to the internal list of notifications of this transaction.
	 *
	 * @param notification the notification to add
	 */
	private void addNotification(_LyNotification notification)
	{
		_eventTypes.addLast(notification);
	}

	/**
	 * Add all providers, operations and notifications of the specified transaction to the current
	 * transaction.
	 *
	 * @param transaction the transaction to add to the current transaction.
	 */
	protected void addTransaction(LyTransaction transaction)
	{
		_LyOperation    operation;
		_LyNotification notification;

		Iterator<_LyOperation>it = transaction._operations.iterator();
		while (it.hasNext())
		{
			operation = it.next();
			addOperation(operation);
		}

		Iterator<_LyNotification> it2 = transaction._eventTypes.iterator();
		while (it2.hasNext())
		{
			notification = it2.next();
			addNotification(notification);
		}

		Status status = transaction.getStatus();

		if (status == Status.ERROR)
			_status = Status.ERROR;
		else
		{
			transaction._status = Status.COMPLETED;
			if ((_status == Status.STARTED) && (_status == status))
				// No operations on each transaction
				_status = Status.STARTED;
			else if (_status != Status.ERROR)
				// One or more operation on at least one transaction.
				_status = Status.RUNNING;
		}

		addProviders(transaction);

		UserTransaction ut = transaction.getUserTransaction();
		UID			    uid = transaction.getUID();

		if (ut != null)
			setUserTransaction(ut, uid);
	} // end method addTransaction

	/**
	 * Add all providers and all connections of the specified transaction to the current
	 * transaction.
	 *
	 * @param transaction the transaction to add to the current transaction.
	 */
	protected void addProviders(LyTransaction transaction)
	{
		int			   i, n = transaction._providers.size(), idx;
		LyDataProvider provider;
		LyConnection   connection;

		for (i = 0; i < n; i++)
		{
			provider = transaction._providers.elementAt(i);
			connection = transaction._connections.elementAt(i);
			addDataProvider(provider);
			idx = _providers.indexOf(provider);
			if (_connections.elementAt(idx) == null)
				_connections.setElementAt(connection, idx);
		}
	}

	/**
	 * Modifier of the associated user transaction and its unique identifier.
	 *
	 * @param ut  The user transaction.
	 * @param uid Unique identifier of the user transaction.
	 */
	protected void setUserTransaction(UserTransaction ut, UID uid)
	{
		_ut = ut;
		_uid = uid;
	}

	/**
	 * Get all data providers of this transaction.
	 *
	 * @return a Vector of the data providers.
	 */
	public Vector<LyDataProvider> getProviders()
	{
		Vector<LyDataProvider> providers = new Vector<LyDataProvider>();
		int					   i, n = _providers.size();

		for (i = 0; i < n; i++)
			providers.addElement(_providers.elementAt(i));

		return providers;
	}

	/* Internal classes */

	/**
	 * Internal class to manage operations in this transaction.
	 */
	class _LyOperation
		implements LyResponseListener,
			LyGetDataSourceListenersInterface
	{
		Operation		   _operation;

		LyDataSource	   _dataSource;

		Vector<Object>     _peers;

		Vector<LyObjectId> _peerIds;

		Vector<LyField>    _peerFields;

		LyObjectId		   _peerId;

		int				   _limit;

		LySort			   _peerSort;

		LyFilter		   _peerFilter;

		LyResponseListener _listener;

		LyClassBinding     _classBinding;

		/**
		 * Constructor.
		 *
		 * @param operation    Type of the operation
		 * @param dataSource   Data source concerned by the operation
		 * @param peer         Object to update
		 * @param listener     Listener of this operation
		 * @param classBinding Binding of the class.
		 */
		_LyOperation(Operation operation, LyDataSource dataSource, Object peer,
			LyResponseListener listener, LyClassBinding classBinding)
		{
			_operation = operation;
			_dataSource = dataSource;
			_listener = listener;
			_classBinding = classBinding;
			_peerId = null;
			_peerFields = null;

			_peers = new Vector<Object>();
			_peers.addElement(peer);

			initPeerIds();
		}

		/**
		 * Constructor.
		 *
		 * @param operation    Type of the operation
		 * @param dataSource   Data source concerned by the operation
		 * @param peers        List of objects to update.
		 * @param listener     Listener of this operation
		 * @param classBinding Binding of the class.
		 */
		_LyOperation(Operation operation, LyDataSource dataSource, Vector<Object> peers,
			LyResponseListener listener, LyClassBinding classBinding)
		{
			_operation = operation;
			_dataSource = dataSource;
			_listener = listener;
			_classBinding = classBinding;
			_peers = peers;

			initPeerIds();
		}

		/**
		 * Constructor used for a load operation.
		 *
		 * @param operation    Type of the operation
		 * @param dataSource   Data source concerned by the operation
		 * @param peerId       Id of the object to load.
		 * @param peerFields   List of fields to load.
		 * @param listener     Listener of this operation
		 * @param classBinding Binding of the class.
		 */
		_LyOperation(Operation operation, LyDataSource dataSource, LyObjectId peerId,
			Vector<LyField> peerFields, LyResponseListener listener, LyClassBinding classBinding)
		{
			_operation = operation;
			_dataSource = dataSource;
			_listener = listener;
			_classBinding = classBinding;
			_peerId = peerId;
			_peerFields = peerFields;
		}

		/**
		 * Constructor used for a load operation.
		 *
		 * @param operation    Type of the operation
		 * @param dataSource   Data source concerned by the operation
		 * @param peerFields   List of fields to load.
		 * @param peerSort     Sort criterion
		 * @param peerFilter   Filter criterion
		 * @param limit        Number of objects to load.
		 * @param listener     Listener of this operation
		 * @param classBinding Binding of the class.
		 */
		_LyOperation(Operation operation, LyDataSource dataSource, Vector<LyField> peerFields,
			LySort peerSort, LyFilter peerFilter, int limit,
			LyResponseListener listener, LyClassBinding classBinding)
		{
			_operation = operation;
			_dataSource = dataSource;
			_listener = listener;
			_classBinding = classBinding;
			_peerFields = peerFields;
			_peerSort = peerSort;
			_peerFilter = peerFilter;
			_limit = limit;
		}

		/**
		 * Constructor used for a load operation.
		 *
		 * @param operation    Type of the operation
		 * @param dataSource   Data source concerned by the operation
		 * @param peerFilter   Filter criterion
		 * @param classBinding Binding of the class.
		 */
		_LyOperation(Operation operation, LyDataSource dataSource, LyFilter peerFilter,
			LyClassBinding classBinding)
		{
			_operation = operation;
			_dataSource = dataSource;
			_classBinding = classBinding;
			_peerFilter = peerFilter;
		}

		/**
		 * Internal method to initialize a list of LyObjectId from the peer objects managed by this
		 * operation.
		 */
		void initPeerIds()
		{
			LyObjectId peerId;
			Object     peer;
			int		   i, n;
			LyClass    peerClass;

			// For sizing or loading operations we don't have any peer.
			if (_peers == null)
				return;

			n = _peers.size();
			_peerIds = new Vector<LyObjectId>(n);

			// For a creation we can't initialize ids here because the id may be
			// calculeted by the data provider.
			if (_operation == Operation.CREATE)
				return;

			peerClass = _dataSource.getDataClass();
			for (i = 0; i < n; i++)
			{
				peer = _peers.elementAt(i);
				peerId = peerClass.getObjectId(peer);
				_peerIds.addElement(peerId);
			}
		}

		/**
		 * Send this operation to the data source.
		 *
		 * @return The response of this request.
		 */
		LyResponse send()
		{
			LyResponse response = null;
			int		   n = ((_peers != null) ? _peers.size() : 0);
			Object     peer = ((n == 0) ? null : _peers.elementAt(0));

			_status = Status.RUNNING;
			try
			{
				if (_operation == Operation.CREATE)
					response = _dataSource.createObject(peer, this, _classBinding);
				else if (_operation == Operation.SET)
				{
					if (n == 1)
						response = _dataSource.setObject(peer, this, _classBinding);
					else
						response = _dataSource.setObjects(_peers, this, _classBinding);
				}
				else if (_operation == Operation.DELETE)
				{
					if (n == 1)
						response = _dataSource.deleteObject(peer, this, _classBinding);
					else
						response = _dataSource.deleteObjects(_peers, this, _classBinding);
				}
				else if (_operation == Operation.LOAD)
				{
					if (_peerId == null)
						response = _dataSource.loadObjects(_peerFields, _peerSort, _peerFilter,
								_limit, this, _classBinding);
					else
						response = _dataSource.loadObject(_peerId, _peerFields, this,
								_classBinding);
				}
			}
			catch (Exception e)
			{
				if (_session != null)
					_session.handleException(e);
				else
					getEnvironment().writeException(e);
				_status = Status.ERROR;
			}

			return response;
		} // end method send

		/**
		 * To get the number of objects corresponding to the filter criterion.
		 *
		 * @return The number of objects.
		 */
		int getSize()
		{
			_status = Status.RUNNING;
			int size = -1;

			try
			{
				size = _dataSource.getSize(_peerFilter, _classBinding);
			}
			catch (Exception e)
			{
				if (_session != null)
					_session.handleException(e);
				else
					getEnvironment().writeException(e);
				_status = Status.ERROR;
			}
			return size;
		}

		/**
		 * To rollback this operation.
		 */
		void rollback()
		{
			LyObjectId peerId;
			Object     peer;

			if (_peerIds == null)
				return;
			int i, n = _peerIds.size();

			for (i = 0; i < n; i++)
			{
				peerId = _peerIds.elementAt(i);

				try
				{
					if (_operation == Operation.CREATE)
					{
						// We must remove the object from memory.
						peer = peerId.getObject();
						if (peer != null)
						{
							_dataSource.setRollBack(true);
							_dataSource.deleteObject(peer, null, _classBinding);
							_dataSource.setRollBack(false);
						}
					}
					else if ((_operation == Operation.SET)
							|| (_operation == Operation.DELETE))
					{
						// We must reload the object.
						_dataSource.setRollBack(true);
						_dataSource.loadObject(peerId, null, null, _classBinding);
						_dataSource.setRollBack(false);
					}
				}
				catch (Exception e)
				{
					if (_session != null)
						_session.handleException(e);
					else
						getEnvironment().writeException(e);
				}
			}
		} // end method rollback

		/**
		 * Method for receiving a new object in a response. This method is called by LyResponse when
		 * a new objet is added in response and thus will be called several times for demands that
		 * return several objects.
		 *
		 * @param response  Response for which a new object is received.
		 * @param object    The new object returned in the response.
		 * @param oldValues When the demand was a modification of an object, this parameter may
		 *                  contains the list of values that were changed. In other cases, this
		 *                  parameter is null.
		 */
		public void newResponse(LyResponse response, Object object, LyValueSet oldValues)
		{
			if (response.getStatus() == LyResponse.Status.ERROR)
				_status = Status.ERROR;

			if ((_operation == Operation.CREATE) && (object != null))
			{
				// Update the peer id list.
				LyClass    peerClass = _dataSource.getDataClass();
				LyObjectId peerId;

				peerId = peerClass.getObjectId(object);
				_peerIds.addElement(peerId);
			}

			if (_listener != null)
				_listener.newResponse(response, object, oldValues);
		}

		/**
		 * Method for ending a response. This method is called by LyResponse when all results have
		 * been received to notify the response is terminated.
		 *
		 * @param response Response which end is notified.
		 */
		public void endResponse(LyResponse response)
		{
			if (response.getStatus() == LyResponse.Status.ERROR)
				_status = Status.ERROR;
			if (_listener != null)
				_listener.endResponse(response);
		}

		/**
		 * Gets list of listeners that have been notified of the response of the last request to
		 * request encoder.
		 *
		 * @return List of notified listeners (may be null).
		 */
		public Vector<LyDataSourceListener> getDataSourceListeners()
		{
			if (_listener instanceof LyGetDataSourceListenersInterface)
				return ((LyGetDataSourceListenersInterface)_listener).getDataSourceListeners();
			else
				return null;
		}
	} // end class _LyOperation


	/**
	 * Internal class to manage notifications of this transaction.
	 */
	class _LyNotification
	{
		LyEventNotifier.EventType _eventType;

		LyClass					  _peerClass;

		Object					  _object;

		LyValueSet				  _values;

		LyEventNotifier			  _notifier;

		/**
		 * Constructor.
		 *
		 * @param notifier     Notification service.
		 * @param notification Type of the notification event.
		 * @param peerClass    Class impacted by the notification.
		 * @param object       Object impacted by the notification.
		 * @param values       New values of the object for a modification event.
		 */
		_LyNotification(LyEventNotifier notifier, LyEventNotifier.EventType notification,
			LyClass peerClass, Object object, LyValueSet values)
		{
			_notifier = notifier;
			_eventType = notification;
			_peerClass = peerClass;
			_object = object;
			_values = values;
		}

		/**
		 * Accessor to the notification service.
		 *
		 * @return The notification service.
		 */
		LyEventNotifier getNotifier()
		{
			return _notifier;
		}

		/**
		 * Accessor to the type of the notification.
		 *
		 * @return The type of the notification.
		 */
		LyEventNotifier.EventType getNotification()
		{
			return _eventType;
		}

		/**
		 * Accessor to the class impacted by the notification.
		 *
		 * @return The class impacted by the notification.
		 */
		LyClass getPeerClass()
		{
			return _peerClass;
		}

		/**
		 * Accessor to the object impacted by the notification.
		 *
		 * @return The object impacted by the notification.
		 */
		Object getObject()
		{
			return _object;
		}

		/**
		 * Accessor to the values of this notification.
		 *
		 * @return The values of this notification.
		 */
		LyValueSet getValues()
		{
			return _values;
		}
	} // end class _LyNotification
} // end class LyTransaction
