/*
 * LyTransactionManager.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer;

import java.util.*;
import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.*;
import leon.misc.*;


/**
 * This class is used to manage transactions in an application.
 *
 * @see    LyDataProvider
 * @see    LyRequestEncoder
 * @see    LyConnection
 * @see    LyTransaction
 * @author Lyria
 */
public class LyTransactionManager
{
	public static final String		    __VERSION =
		"$Id: LyTransactionManager.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** Reference to the application instance. */
	protected LyApplication			    _application;

	/** List of transactions managed by the transaction manager. */
	protected LinkedList<LyTransaction> _transactions;

	/** Indicates if user transaction are managed or not.. */
	protected boolean				    _hasUt;

	/**
	 * Constructs a transaction manager. Initialization is done with the init method.
	 *
	 * @param application Current application instance
	 */
	public LyTransactionManager(LyApplication application)
	{
		init();

		_application = application;
	}

	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		_application = null;
		_transactions = new LinkedList<LyTransaction>();
		_hasUt = true;
	}

	/**
	 * To begin a new transaction. The new transaction is associated with the current user session.
	 *
	 * @return the new created transaction.
	 */
	public LyTransaction begin()
	{
		LyTransaction transaction, t;
		LyEnvironment environment = _application.getEnvironment();

		t = getTransaction();
		transaction = new LyTransaction(_application);
		if (t != null)
			transaction.addProviders(t);
		else
			transaction.begin();
		addTransaction(transaction);

		if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
			environment.logOut("Begin of transaction " + transaction);
			//Thread.dumpStack();

		return transaction;
	}

	/**
	 * To end a transaction. if an error has occured during the transaction, do a rollback else do a
	 * commit.
	 *
	 * @param  transaction The transaction to end
	 * @return true if all operations in the transactions are OK and false if an error occured.
	 *         WARNING : operations may be completed successfully but the current transaction may be
	 *         merged in another transactions. If one of the operations in this transaction is in
	 *         error, all operations of all transactions are rollbacked. When the ended transaction
	 *         is not the last begin transaction for the current user session (interlaced
	 *         transactions).
	 */
	public boolean end(LyTransaction transaction)
	{
        LyTransaction.Status  status = transaction.getStatus();
		boolean               commit = true;

		if (status == LyTransaction.Status.ERROR)
			commit = false;

		return end(transaction, commit);
	}

	/**
	 * To end a transaction.
	 *
	 * @param  transaction The transaction to end
	 * @param  commit      Specifiy if the transaction must be committed (true) or aborted (false).
	 * @return true if all operations in the transactions are OK and false if an error occured.
	 *         WARNING : operations may be completed successfully but the current transaction may be
	 *         merged in another transactions. If one of the operations in this transaction is in
	 *         error, all operations of all transactions are rollbacked. When the ended transaction
	 *         is not the last begin transaction for the current user session (interlaced
	 *         transactions).
	 */
	public boolean end(LyTransaction transaction, boolean commit)
	{
		LyEnvironment environment = _application.getEnvironment();
		LyTransaction t = getTransaction();

		if ((t != null) && (transaction != t))
		{
			if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
				environment.logOut("\tEnd of transaction " + transaction
					+ " Transaction interlaced");
		}

		removeTransaction(transaction);

		// Search for another transaction on the same criterias.
		Vector<LyDataProvider>	providers = transaction.getProviders();
		LyDataProvider 			provider;

		t = null;
		for (int i = 0; i < providers.size(); i++)
		{
			provider = providers.elementAt(i);
			t = getTransaction(provider);
			if (t != null)
				break;
		}
		if (t != null)
		{
			// Merge the transactions and do not perform any commit or rollback.
			t.addTransaction(transaction);
			if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
				environment.logOut("End of transaction " + transaction + " Transaction merged");
				//Thread.dumpStack();
		}
		else if ((commit) && (transaction.prepareCommit()))
		{
			transaction.commit();
			if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
				environment.logOut("End of transaction " + transaction + " Transaction committed");
				//Thread.dumpStack();
		}
		else
		{
			transaction.rollback();
			if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
				environment.logOut("End of transaction " + transaction
					+ " Transaction rollbacked");
				//Thread.dumpStack();
		}

		return (transaction.getStatus() != LyTransaction.Status.COMPLETED);
	} // end method end

	/**
	 * Abort all pending transactions (rollback) for a session.
	 *
	 * @param session the session
	 */
	public void abort(LyDataSession session)
	{
		LyTransaction transaction = getTransaction(session);

		while (transaction != null)
		{
			end(transaction, false);
			transaction = getTransaction(session);
		}
	}

	/**
	 * Abort all pending transactions (rollback) for a data provider.
	 *
	 * @param session the session
	 */
	public void abort(LyDataProvider provider)
	{
		LyTransaction transaction = getTransaction(provider);

		while (transaction != null)
		{
			end(transaction, false);
			transaction = getTransaction(provider);
		}
	}

	/**
	 * Accessor to a connection instance from the connection pool of the specified data provider.
	 *
	 * <P>If no connection is available currently, this method will block until a connection is
	 * released and available.</P>
	 *
	 * <P><b>Important !</b> The connection MUST be released after usage so it's available for the
	 * connection pool. Invoke releaseConnection when the connection is no longer needed.</P>
	 *
	 * @param  provider the data provider from which we want to get a connection.
	 * @return a connection instance from the connection pool of the specified data provider. The
	 *         connection instance must be released (releaseConnection) after usage.
	 * @see    #releaseConnection
	 */
	public LyConnection getConnection(LyDataProvider provider)
	{
		LyTransaction transaction = getTransaction();
		LyConnection  connection;

		if (transaction == null)
			connection = provider.getConnection(true);
		else
			connection = transaction.getConnection(provider);

		return connection;
	}

	/**
	 * Return a connection instance to the connection pool of the specified data provider.
	 *
	 * <P>The connection reference was obtained by getConnection()</P>
	 *
	 * <P><b>Important !</b> The connection MUST be released after usage so it's available for the
	 * connection pool. Invoke releaseConnection when the connection is no longer needed.</P>
	 *
	 * @param connection a connection instance from the connection pool of the specified data
	 *                   provider.
	 * @param provider   the provider to return the connection.
	 * @see   #releaseConnection
	 */
	public void releaseConnection(LyConnection connection, LyDataProvider provider)
	{
		LyTransaction transaction = getTransaction();

		if (transaction == null)
			provider.releaseConnection(connection, true);
	}

	/**
	 * To get the last open transaction on the current session associated with a data provider. If
	 * there is a transaction without any associated provider, it returns it. For internal use only.
	 *
	 * @param  provider the associated provider of the search transaction.
	 * @return the last open transaction associated with the specified provider if exists, else
	 *         null.
	 */
	public LyTransaction getTransaction(LyDataProvider provider)
	{
		LyDataSession session = _application.getCurrentSession();

		synchronized (_transactions)
		{
			Iterator<LyTransaction>	it = _transactions.iterator();
			LyTransaction 			transaction;
			Vector<LyDataProvider>	providers;
			Thread		  			thread;

			thread = Thread.currentThread();

			while (it.hasNext())
			{
				transaction = (LyTransaction)it.next();
				if ((transaction.getSession() == session)
						&& (transaction.getThread() == thread))
				{
					providers = transaction.getProviders();
					if ((providers.size() == 0) || (providers.contains(provider)))
						return transaction;
				}
			}
		}

		return null;
	} // end method getTransaction

	/**
	 * To get the last open transaction on the current session.
	 *
	 * @return the last open transaction on the current session if exists, else null.
	 */
	protected LyTransaction getTransaction()
	{
		LyDataSession session = _application.getCurrentSession();

		return getTransaction(session);
	}

	/**
	 * To get the last open transaction on a session.
	 *
	 * @param  session the session
	 * @return the last open transaction on the specified session if exists, else null.
	 */
	public LyTransaction getTransaction(LyDataSession session)
	{
		synchronized (_transactions)
		{
			Iterator<LyTransaction>	it = _transactions.iterator();
			LyTransaction			transaction;
			Thread					thread;

			thread = Thread.currentThread();

			while (it.hasNext())
			{
				transaction = it.next();
				if ((transaction.getSession() == session)
						&& (transaction.getThread() == thread))
					return transaction;
			}
		}

		return null;
	}

	/**
	 * Internal method to add a new transaction to the manager.
	 *
	 * @param transaction the transaction to add.
	 */
	protected void addTransaction(LyTransaction transaction)
	{
		synchronized (_transactions)
		{
			_transactions.addFirst(transaction);
		}
	}

	/**
	 * Internal method to remove a transaction from the manager.
	 *
	 * @param transaction the transaction to remove.
	 */
	protected void removeTransaction(LyTransaction transaction)
	{
		synchronized (_transactions)
		{
			_transactions.remove(transaction);
		}
	}

	/* Methods used to send requests */

	/**
	 * Method to load an object from its id on the specified data source. Load fields of an object
	 * from a given object id.
	 *
	 * @param  dataSource   The data source for loading the object.
	 * @param  peerId       Object identifier corresponding to the object to load.
	 * @param  peerFields   Fields to load (null means all fields).
	 * @param  listener     Listener of the response.
	 * @param  classBinding binding between logical class and physical class This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the loading.
	 * @see    LyDataSource#loadObject
	 */
	public LyResponse loadObject(LyDataSource dataSource, LyObjectId peerId,
		Vector<LyField> peerFields,
		LyResponseListener listener, LyClassBinding classBinding)
	{
		LyResponse    response;
		LyTransaction transaction = getTransaction();

		if (transaction != null)
			response = transaction.loadObject(dataSource, peerId, peerFields, listener,
					classBinding);
		else
			response = dataSource.loadObject(peerId, peerFields, listener, classBinding);

		return response;
	}

	/**
	 * Method for loading objects on the specified data source. Load fields of a specified number of
	 * objects sorted by a given sort and that match a given filter.
	 *
	 * @param  dataSource   The data source for loading the objects.
	 * @param  peerFields   Fields to load (null means all fields).
	 * @param  peerSort     Gives sort criteria, or null if there is no sort (the order is the one
	 *                      of the physical layer).
	 * @param  peerFilter   Gives filter criteria, or null if there is no filter (all objects).
	 * @param  n            Indicates the number of objects to load (-1 means no limit).
	 * @param  listener     Listener of the response.
	 * @param  classBinding binding between logical class and physical class This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the loading.
	 * @see    LyDataSource#loadObjects
	 */
	public LyResponse loadObjects(LyDataSource dataSource, Vector<LyField> peerFields,
		LySort peerSort,
		LyFilter peerFilter, int n, LyResponseListener listener, LyClassBinding classBinding)
	{
		LyResponse    response;
		LyTransaction transaction = getTransaction();

		if (transaction != null)
			response = transaction.loadObjects(dataSource, peerFields, peerSort, peerFilter, n,
					listener, classBinding);
		else
			response = dataSource.loadObjects(peerFields, peerSort, peerFilter, n, listener,
					classBinding);

		return response;
	}

	/**
	 * Return the number of objects within the specified data source matching the given filter.
	 *
	 * @param  dataSource   The data source for getting the number of objects.
	 * @param  peerFilter   Filter applied on objects for counting.
	 * @param  classBinding Binding between logical class and physical class. This binding may
	 *                      contain extra information required for encoding the request (such as
	 *                      extra filter).
	 * @return Number of objects matching the given filter in this data source.
	 * @see    LyDataSource#getSize
	 */
	public int getSize(LyDataSource dataSource, LyFilter peerFilter, LyClassBinding classBinding)
	{
		int			  size;

		LyTransaction transaction = getTransaction();

		if (transaction != null)
			size = transaction.getSize(dataSource, peerFilter, classBinding);
		else
			size = dataSource.getSize(peerFilter, classBinding);
		return size;
	}

	/**
	 * Delete an object on the specified data source.
	 *
	 * @param  dataSource   The data source for deleting the object.
	 * @param  peer         Object to delete.
	 * @param  listener     Listener of the response.
	 * @param  classBinding Binding between logical class and physical class This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the deletion.
	 * @see    LyDataSource#deleteObject
	 */
	public LyResponse deleteObject(LyDataSource dataSource, Object peer,
		LyResponseListener listener, LyClassBinding classBinding)
	{
		LyResponse    response;
		LyTransaction transaction = getTransaction();

		if (transaction != null)
			response = transaction.deleteObject(dataSource, peer, listener, classBinding);
		else
			response = dataSource.deleteObject(peer, listener, classBinding);

		return response;
	}

	/**
	 * Delete a list of objects on the specified data source.
	 *
	 * @param  dataSource   The data source for deleting the objects.
	 * @param  peers        List of objects to delete.
	 * @param  listener     Listener of the response.
	 * @param  classBinding Binding between logical class and physical class This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the deletion.
	 * @see    LyDataSource#deleteObjects
	 */
	public LyResponse deleteObjects(LyDataSource dataSource, Vector<Object> peers,
		LyResponseListener listener, LyClassBinding classBinding)
	{
		LyResponse    response;
		LyTransaction transaction = getTransaction();

		if (transaction != null)
			response = transaction.deleteObjects(dataSource, peers, listener, classBinding);
		else
			response = dataSource.deleteObjects(peers, listener, classBinding);

		return response;
	}

	/**
	 * Create a new object on the specified data source.
	 *
	 * @param  dataSource   The data source for creating the object.
	 * @param  peer         Object to create.
	 * @param  listener     Listener of the response.
	 * @param  classBinding Binding between logical class and physical class. This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the creation.
	 * @see    LyDataSource#createObject
	 */
	public LyResponse createObject(LyDataSource dataSource, Object peer,
		LyResponseListener listener, LyClassBinding classBinding)
	{
		LyResponse    response;
		LyTransaction transaction = getTransaction();

		if (transaction != null)
			response = transaction.createObject(dataSource, peer, listener, classBinding);
		else
			response = dataSource.createObject(peer, listener, classBinding);

		return response;
	}

	/**
	 * Set an object on the specified data source.
	 *
	 * @param  dataSource   The data source for setting the object.
	 * @param  peer         Object to modify.
	 * @param  listener     Listener of the response.
	 * @param  classBinding Binding between logical class and physical class This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the modification.
	 * @see    LyDataSource#setObject
	 */
	public LyResponse setObject(LyDataSource dataSource, Object peer,
		LyResponseListener listener, LyClassBinding classBinding)
	{
		LyResponse    response;
		LyTransaction transaction = getTransaction();

		if (transaction != null)
			response = transaction.setObject(dataSource, peer, listener, classBinding);
		else
			response = dataSource.setObject(peer, listener, classBinding);

		return response;
	}

	/**
	 * Set a list of objects on the specified data source.
	 *
	 * @param  dataSource   The data source for setting the objects.
	 * @param  peers        List of objects to modify.
	 * @param  listener     Listener of the response.
	 * @param  classBinding Binding between logical class and physical class This binding may
	 *                      contain extra information required for encoding the request.
	 * @return Response to the modification.
	 * @see    LyDataSource#setObjects
	 */
	public LyResponse setObjects(LyDataSource dataSource, Vector<Object> peers,
		LyResponseListener listener, LyClassBinding classBinding)
	{
		LyResponse    response;
		LyTransaction transaction = getTransaction();

		if (transaction != null)
			response = transaction.setObjects(dataSource, peers, listener, classBinding);
		else
			response = dataSource.setObjects(peers, listener, classBinding);

		return response;
	}

	/**
	 * Indicates if user transactions are managed or not by this context manager.
	 *
	 * @param hasUt true to indicate that user transactions are managed, false otherwise.
	 */
	protected void hasUt(boolean hasUt)
	{
		_hasUt = hasUt;
	}

	/**
	 * Return the _hasUt flag.
	 */
	protected boolean hasUt()
	{
		return _hasUt;
	}
} // end class LyTransactionManager
