/*
 * LyExtendDataProvider.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.extend;

import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.LyClassBinding;
import leon.info.location.*;
import leon.peer.*;


/**
 * This class represents the data provider used for local class infos.
 * @see LyClassInfo
 */
public final class LyExtendDataProvider extends LyDataProvider
{
	public static final String __VERSION =
		"$Id: LyExtendDataProvider.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Force an abstract class tu connect directly to the peer binding (not using concrete classbinding definition)
	 * but the abstract classbinding. This could be much more efficient on concrete classes with an union from an abstract class.
	 * INTERNAL USE : should remain false (this paramenetr is only a performance optimization)
	 */
	protected boolean _useClassBindingForAbstractClass = false;
	
	/**
	 * Constructs a local data provider.
	 * @param application current application instance
	 * @param location    this value is created in LyApplication.createProviders()
	 */
	public LyExtendDataProvider(LyApplication application, LyLocation location)
	{
		super(application, location);
	}

	/**
	 * Default constructor.
	 * @param      application Current application instance
	 * @param      id          Identifier of this dataProvider. This value is obtained from the
	 *                         location parameter in the meta model.
	 * @deprecated should use LyUnionDataProvider(LyApplication application, LyLocation location)
	 *             instead of this constructor.
	 */
	public LyExtendDataProvider(LyApplication application, String id)
	{
		super(application, id);
	}


	/**
	 * This method is used to build all the class peers for all the classInfo. It must be
	 * specialized by subclasses. Basically, this method associates the dataProvider with local
	 * classes. A class is local if there is no binding.
	 * @param  classInfos Classes to validate.
	 * @return True if the validation was successful.
	 */
	public boolean validate(LyInfoList<LyClassInfo> classInfos)
	{
		LyClassInfo classInfo, superClassInfo;
		int		    i, n = classInfos.getSize();

		for (i = 0; i < n; i++)
		{
			classInfo = classInfos.getInfo(i);
			if (classInfo.isVirtual())
				continue;

			if (classInfo.isAbstract())
				// The class is local.
				new LyExtendObjectSource(this, classInfo);
			else if (classInfo.getSubClasses() != null)
			{
				// Non abstract inherited class
				LyApplication  application = getApplication();
				LyObjectSource objectSource = application.getObjectSource(classInfo);

				application.removeDataSource(objectSource);
				new LyExtendObjectSource(this, classInfo, objectSource);
			}
			// Maybee this class has been added and it's parent classes has subclasses now and not before
			// Just check
			superClassInfo = classInfo.getSuperClass();
			if (superClassInfo != null)
			{
				// Non abstract inherited class
				LyApplication  application = getApplication();
				LyObjectSource objectSource = application.getObjectSource(superClassInfo);

				// this ObjectSoource must be an LyExtendObjectSource
				if ((objectSource != null)
						&& (!(objectSource instanceof LyExtendObjectSource)))
				{
					application.removeDataSource(objectSource);

					LyExtendObjectSource newSource = new LyExtendObjectSource(this, superClassInfo,
							objectSource);

					newSource.listenToSubClasses();

					// Clear the cache on the classInfo
					superClassInfo.clearObjectSourceCache();
					// Some listeners may already have registred for objectSource : they must listen to newSource now.
					LyEventManager em = _application.getEventManager();

					em.switchObjectSource(objectSource, newSource);
				}
			}
		}

		return super.validate(classInfos);
	} // end method validate

	/**
	 * Create a specific request encoder for the specified dataSource.
	 * @param  dataSource Data source for which is request encoder is requested.
	 * @return A LyExtendRequestEncoder for the specified dataSource.
	 */
	public LyRequestEncoder getRequestEncoder(LyDataSource dataSource)
	{
		synchronized (this)
		{
			if (_useClassBindingForAbstractClass)
			{
				LyClassInfo    classInfo = (LyClassInfo)dataSource.getDataClass();
				LyClassBinding classBinding = classInfo.getBinding();
				
				if (classBinding != null)
				{
					String 		locationId = classBinding.getLocation();
					int 		n = getApplication().getDataProviderCount();
					LyDataProvider provider;
					
					for (int i=0; i<n; i++)
					{
						provider = getApplication().getDataProvider(i);
						
						if (provider.getId().equals(locationId))
						{
							LyRequestEncoder encoder = provider.getRequestEncoder(dataSource);
							encoder.setDataClass(classBinding.getPeerClass());
							return encoder;
						}
					}
				}
				
			}
		}
		return new LyExtendRequestEncoder(this, dataSource);
	}

	/**
	 * This class is a factory to create a LyConnection instance. Current implementation only return
	 * null.
	 * @return a new Connection instance. The connection instance is used in the specific LyRequest
	 *         encoder implementation.
	 */
	protected LyConnection createConnection()
	{
		return null;
	}
	
	/**
	 * Force an abstract class tu connect directly to the peer binding (not using concrete classbinding definition)
	 * but the abstract classbinding. This could be much more efficient on concrete classes with an union from an abstract class.
	 * INTERNAL USE : should remain false (this paramenetr is only a performance optimization)
	 * @param new status (INTERNAL USE)
	 */
	public void setUseClassBindingForAbstractClass(boolean status)
	{
		_useClassBindingForAbstractClass = status;
	}
} // end class LyExtendDataProvider
