/*
 * LyExtendRequestEncoder.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.peer.extend;

import leon.app.*;
import leon.data.*;
import leon.data.cache.*;
import leon.info.*;
import leon.peer.*;

import java.util.*;


/**
 * Local request encoder associated to the local data provider.
 */
public class LyExtendRequestEncoder extends LyRequestEncoder
	implements LyResponseListener,
		LyGetDataSourceListenersInterface
{
	public static final String __VERSION = "$Id: LyExtendRequestEncoder.java 30500 2011-04-13 16:17:02Z jlsaouli $";

	/** Set of subclasses corresponding to sub requests. */
	protected Vector<LyClassInfo> _subClasses;

	/** Number of sub requests. */
	protected int _nbSubRequests;

	/** Set of objetcs corresponding to sub requests for a LOAD operation. */
	protected LySimpleObjectList _objects;
	protected LySimpleObjectList _subObjects;

	/** For a LOAD operation, the number of requested objects. */
	protected int _n;

	/**
	 * Default constructor Initialization is done with the init method.
	 */
	public LyExtendRequestEncoder(LyDataProvider provider, LyDataSource dataSource)
	{
		super(provider, dataSource);

		LyInfoList<LyClassInfo> subClasses = ((LyClassInfo)_class).getSubClasses();

		if (subClasses != null)
			_subClasses = subClasses.getInfos();
	}

	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		super.init();
		_subClasses = null;
		_nbSubRequests = 0;
		_objects = null;
		_subObjects = null;
		_n = -1;
	}

	/**
	 * To set the response properties according to the data response properties.
	 */
	protected void setResponseProperties(LyResponse response)
	{
		_response.setSorted(false);
		// We always can filter the responses.
		_response.setFiltered(true);
		_response.setCounted(false);
		_response.setDistinctValues(response.hasDistinctValues());
	}

	/**
	 * Get an object.
	 *
	 * @param   objectId  object identifier
	 * @param   fields    the fields to load (if null, load all)
	 *
	 * @return  true if the request was correctly sent to the external data provider.
	 */
	public boolean getObject(LyObjectId objectId, Vector<LyField> fields)
	{
		LyClassInfo			    classInfo = objectId.getClassInfo();
		LyInfoList<LyFieldInfo> idFields = classInfo.getIdFieldInfos();
		LyFilter			    idFilter = new LyFilter();
		LyFilterExpression	    exp = null;
		String[]			    ids = objectId.toStrings();

		if (idFields.getSize() != ids.length)
		{
			// Set the response in error mode
			_response.setStatus(LyResponse.Status.ERROR);
			_response.addResponse(null, null);
			return true;
		}

		for (int i = 0; i < idFields.getSize(); i++)
		{
			exp = new LyFilterExpression(idFields.getInfo(i), ids[i]);
			idFilter.addElement(exp);
		}

		return getObjects(fields, null, idFilter, 1);
	}

	/**
	 * Check if the given fields are all in the mother class.
	 *
	 * @param   fields  a vector of field
	 *
	 * @return  true if all the fields are declared in the mother class. False otherwise.
	 */
	protected boolean isFieldListInMotherClass(Vector<LyField> fields)
	{
		if (fields == null)
			return false;

		if (fields.size() == 0)
			return false;

		LyField field;
		Vector<? extends LyField>  classFields = _class.getFields();

		for (int i = 0; i < fields.size(); i++)
		{
			field = fields.elementAt(i);

			if (!classFields.contains(field))
				return false;
		}

		return true;
	}

	/**
	 * Get a set of objects.
	 *
	 * @param   fields  the fields to load (if null, load all)
	 * @param   sort    indicates sorting criterion, or null if default sort is used.
	 * @param   filter  indicates filtering criterion, or null if there is no filter.
	 * @param   n       maximum number of objects to get.
	 *
	 * @return  true if the request was correctly sent to the external data provider.
	 */
	public boolean getObjects(Vector<LyField> fields, LySort sort, LyFilter filter, int n)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// Create a subrequest for each sub classe.
		int		    i, nbClasses;
		LyClassInfo subClass;

		nbClasses = ((_subClasses == null) ? 0 : _subClasses.size());
		_nbSubRequests = 0;

		// We always stock the results for non abstract inherited classes.
		_objects = null;
		_subObjects = new LySimpleObjectList();
		_subObjects.setSort(sort);
		_subObjects.setSorted(false);
		_n = n;

		LyObjectSource objectSource = ((LyExtendObjectSource)_dataSource).getObjectSource();

		// Maybe the class is not abstract.
		if (objectSource != null)
			// Class is not abstract, we must send a request to its objectSource.
			_nbSubRequests++;

        // Test if all the fields are in the mother class
	    boolean   isFieldListInMotherClass = isFieldListInMotherClass(fields);
        boolean   sendSubRequests = (!isFieldListInMotherClass
            || (objectSource == null)
            || (objectSource.getCache() instanceof LyFullCache));

        if (sendSubRequests)
            _nbSubRequests += nbClasses;

        LyResponse response;
        boolean    local = _response.isLocal();
        int        no = 0;

		if (objectSource != null)
		{
			if (isFieldListInMotherClass
					&& (filter == null)
					&& (n < 0)) // Full load
				((LyExtendObjectSource)_dataSource).setLoaded(true, true);

			if(filter != null)
				filter = filter.getPeerFilter();
			
			// Class is not abstract, we must send a request to its objectSource.
			Object otemp = fields;
			@SuppressWarnings("unchecked")
			Vector<LyFieldInfo> temp = (Vector<LyFieldInfo>) otemp;
			
			response = objectSource.getObjects(temp, sort, filter, n, local, this);
            no = response.getSize();
            
            if ((n > 0)
                    && (no >= n)
                    && (fields == null))
                isFieldListInMotherClass = true;
		}

		// If all fields are not in the mother class, we have to load all sub classes...
		if (sendSubRequests)
		{
			for (i = nbClasses - 1; i >= 0; i--)
			{
				subClass = _subClasses.elementAt(i);
				objectSource = getObjectSource(subClass);
				
				Object otemp = fields;
				@SuppressWarnings("unchecked")
				Vector<LyFieldInfo> temp = (Vector<LyFieldInfo>)otemp;
				response = objectSource.getObjects(temp, sort, filter, n, local, this);
				no += response.getSize();
			}
		}

		return true;
	} // end method getObjects

	/**
	 * Return the number of objects from the data provider .
	 *
	 * @param   filter  indicates filtering criterion, or null if there is no filter.
	 *
	 * @return  true if the request was correctly sent to the external data provider.
	 */
	public boolean getSize(LyFilter filter)
	{
		if(filter != null)
			filter = filter.getPeerFilter();
		
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// Create a subrequest for each sub classe.
		int			   i, nbClasses, size;
		LyClassInfo    subClass;
		LyObjectSource objectSource;

		nbClasses = ((_subClasses == null) ? 0 : _subClasses.size());
		_nbSubRequests = 0;
		size = 0;

		// Maybe the class is not abstract.
		boolean local = _response.isLocal();

		objectSource = ((LyExtendObjectSource)_dataSource).getObjectSource();

		// Check if all filter field are in the mother class
		boolean filterInMotherClass = true;

		if (filter != null)
		{
			LyInfoList<?> fields = filter.getFieldInfos();
			@SuppressWarnings("unchecked")
			Vector<LyField> temp = (Vector<LyField>)fields.getInfos();
			
			filterInMotherClass = isFieldListInMotherClass(temp);
		}

		if ((objectSource != null) && filterInMotherClass)
		{
			// Class is not abstract, we must send a request to its objectSource.
			_nbSubRequests++;
			size += objectSource.getSize(filter, local);
		}
		else
		{
			_nbSubRequests += nbClasses;
			for (i = nbClasses - 1; i >= 0; i--)
			{
				subClass = _subClasses.elementAt(i);
				objectSource = getObjectSource(subClass);

				size += objectSource.getSize(filter, local);
			}
		}

		_response.setStatus(LyResponse.Status.COMPLETED);
		_response.addResponse(new Integer(size), null);

		return true;
	} // end method getSize

	/**
	 * To delete an object.
	 *
	 * @param   object  the object to delete
	 *
	 * @return  true if the request was correctly sent to the external data provider.
	 */
	public boolean deleteObject(Object object)
	{
		// Set the response in error mode
		_response.setStatus(LyResponse.Status.ERROR);
		_response.addResponse(null, null);
		return false;
	}

	/**
	 * To delete a list of objects.
	 *
	 * @param   objects  Objects to delete.
	 *
	 * @return  True if the request was correctly sent to the external data provider, false
	 *          otherwise.
	 */
	public boolean deleteObjects(Vector<Object> objects)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// Create a subrequest for each sub classe.
		int				 i, j, nbClasses;
		LyClassInfo		 subClass;
		LyObjectSource   objectSource;
		Vector<LyObject> subObjects;
		LyObject		 object;

		nbClasses = ((_subClasses == null) ? 0 : _subClasses.size());
		_nbSubRequests = 0;

		// Maybe the class is not abstract.
		boolean local = _response.isLocal();
		boolean refresh = _response.isRefresh();
		boolean deleteLinkedObjects = _response.deleteLinkedObjects();

		objectSource = ((LyExtendObjectSource)_dataSource).getObjectSource();
		if (objectSource != null)
		{
			// Class is not abstract, we must send a request to its objectSource.

			subClass = (LyClassInfo)objectSource.getDataClass();
			subObjects = new Vector<LyObject>();
			for (j = objects.size() - 1; j >= 0; j--)
			{
				object = (LyObject)objects.elementAt(j);
				if (object.getClassInfo() == subClass)
				{
					subObjects.addElement(object);
					objects.removeElementAt(j);
				}
			}

			if (subObjects.size() > 0)
			{
				_nbSubRequests++;
				objectSource.deleteObjects(subObjects, this, local, refresh, deleteLinkedObjects);
			}
		}

		for (i = nbClasses - 1; i >= 0; i--)
		{
			subClass = _subClasses.elementAt(i);
			objectSource = getObjectSource(subClass);

			subObjects = new Vector<LyObject>();
			for (j = 0; j < objects.size(); j++)
			{
				object = (LyObject)objects.elementAt(j);
				if (object.instanceOf(subClass))
					subObjects.addElement(object);
			}

			if (subObjects.size() > 0)
			{
				_nbSubRequests++;
				objectSource.deleteObjects(subObjects, this, local, refresh, deleteLinkedObjects);
			}
		}

		return true;
	} // end method deleteObjects

	/**
	 * To create a new object.
	 *
	 * @param   object  the new object to create
	 *
	 * @return  true if the request was correctly sent to the external data provider.
	 */
	public boolean createObject(Object object)
	{
		// Set the response in error mode
		_response.setStatus(LyResponse.Status.ERROR);
		_response.addResponse(null, null);
		return false;
	}

	/**
	 * To set an object.
	 *
	 * @param   object  the object to set
	 *
	 * @return  true if the request was correctly sent to the external data provider.
	 */
	public boolean setObject(Object object)
	{
		// Set the response in error mode
		_response.setStatus(LyResponse.Status.ERROR);
		_response.addResponse(null, null);

		return true;
	}

	/**
	 * To set a list of objects.
	 *
	 * @param   objects  List of objects to set.
	 *
	 * @return  True if the request was correctly sent to the external data provider, false
	 *          otherwise.
	 */
	public boolean setObjects(Vector<Object> objects)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// Create a subrequest for each sub classe.
		int				 i, j, nbClasses;
		LyClassInfo		 subClass;
		LyObjectSource   objectSource;
		Vector<LyObject> subObjects;
		LyObject		 object;

		nbClasses = ((_subClasses == null) ? 0 : _subClasses.size());
		_nbSubRequests = 0;

		// Maybe the class is not abstract.
		objectSource = ((LyExtendObjectSource)_dataSource).getObjectSource();
		if (objectSource != null)
		{
			// Class is not abstract, we must send a request to its objectSource.

			subClass = (LyClassInfo)objectSource.getDataClass();
			subObjects = new Vector<LyObject>();
			for (j = objects.size() - 1; j >= 0; j--)
			{
				object = (LyObject)objects.elementAt(j);
				if (object.getClassInfo() == subClass)
				{
					subObjects.addElement(object);
					objects.removeElementAt(j);
				}
			}

			if (subObjects.size() > 0)
			{
				_nbSubRequests++;
				objectSource.setObjects(subObjects, this);
			}
		}

		for (i = nbClasses - 1; i >= 0; i--)
		{
			subClass = _subClasses.elementAt(i);
			objectSource = getObjectSource(subClass);

			subObjects = new Vector<LyObject>();
			for (j = 0; j < objects.size(); j++)
			{
				object = (LyObject)objects.elementAt(j);
				if (object.instanceOf(subClass))
					subObjects.addElement(object);
			}

			if (subObjects.size() > 0)
			{
				_nbSubRequests++;
				objectSource.setObjects(subObjects, this);
			}
		}

		return true;
	} // end method setObjects

	/**
	 * Internal method to get the proper object source of a given class info : ask to the current
	 * session, then to the current application.
	 *
	 * @param   classInfo
	 *
	 * @return
	 */
	protected LyObjectSource getObjectSource(LyClassInfo classInfo)
	{
		LyApplication  application = _provider.getApplication();
		LyObjectSource objectSource = null;
		LyDataSession  session = application.getCurrentSession();

		if (session != null)
			objectSource = session.getObjectSource(classInfo);

		if (objectSource == null)
			objectSource = application.getObjectSource(classInfo);

		return objectSource;
	}

	/**
	 * Implementation of LyResponseListener.*
	 */

	/**
	 * Method for response changes.
	 *
	 * @param  response  the response
	 */
	public void newResponse(LyResponse response, Object object, LyValueSet values)
	{
		if (response == null)
			throw new IllegalArgumentException("BF AppEngine : response can't be null");

		setResponseProperties(response);

		LyResponse.Status status = response.getStatus();

		if (status == LyResponse.Status.ERROR)
		{
			_response.setError(response.getError());
			_response.addResponse(null, null);
		}
		else
		{
			LyResponse.Operation operation = response.getOperation();

			if (operation == LyResponse.Operation.LOAD)
			{
				if (_subObjects != null)
				{
					if (_objects != null)
					{
						LyObjectSource objectSource = ((LyExtendObjectSource)_dataSource).getObjectSource();

						if (objectSource != null)
						{
							// If the class in not abstract we can have many LyObjects with
							// the same peer parts (with an union)
							// Sample : classInfo1 bind on a table1
							// classInfo2 extends classInfo1 bind on union on table 1 and table 2
							// When we load the objects of classInfo1 we have the objects of classInfo2
							// twice : one time for classInfo1 and one time for classInfo2
							// We must manage this case.
							LyObjectId objectId = ((LyObject)object).getObjectId();
							LyObject   obj = _objects.getObject(objectId);

							if (obj != null)
								_objects.removeObject(obj);
						}
					}

					_subObjects.addObject((LyObject)object);
				}
				else
					_response.addResponse(object, null);
			}
			else if (operation == LyResponse.Operation.SET)
				_response.addResponse(object, values);
			else
				_response.addResponse(object, null);
		}
	} // end method newResponse

	/**
	 * Method for end of response.
	 *
	 * @param  response  the response
	 */
	public void endResponse(LyResponse response)
	{
		LyResponse.Status status = response.getStatus();

		if (status == LyResponse.Status.ERROR)
			_response.setError(response.getError());

		if ((_subObjects != null)
				&& !response.isSorted())
			_subObjects.setSorted(true);

		if (_objects == null)
			_objects = _subObjects;
		else
			_objects.addList(_subObjects);

		_subObjects = null;

		_nbSubRequests--;

		if (_nbSubRequests == 0)
		{
			if ((response.getOperation() == LyResponse.Operation.LOAD)
					&& (_objects != null))
			{
				_objects.setSorted(true);

				LyObject object;
				int		 i, n = _objects.getSize();

				if ((_n > 0) && (_n < n))
					n = _n;

				for (i = 0; i < n; i++)
				{
					object = _objects.getObject(i);
					_response.addResponse(object, null);
				}

				_response.setSorted(true);
			}

			_response.setStatus(LyResponse.Status.COMPLETED);
			_response.addResponse(null, null);
		}
		else
		{
			_subObjects = new LySimpleObjectList();
            if (_objects != null)
                _subObjects.setSort(_objects.getSort());
			_subObjects.setSorted(false);
		}
	} // end method endResponse

	/**
	 * Indicates which listeners have been notified of the Response.
	 *
	 * @see  LyGetDataSourceListenersInterface
	 */
	public Vector<LyDataSourceListener> getDataSourceListeners()
	{
		if (_dataSource == null)
			return null;

		Vector<LyDataSourceListener> v = new Vector<LyDataSourceListener>(1);
		LyDataSourceListener		 listener = ((LyObjectSource)_dataSource).getListener();

		v.addElement(listener);
		return v;
	}
} // end class LyExtendRequestEncoder
