/*
 * LyFileConnection.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.file;

import java.io.*;
import java.util.*;

import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.location.*;
import leon.misc.*;
import leon.peer.*;


/**
 * This class is used as a connection for local files.
 * @version $Id: LyFileConnection.java 35251 2012-04-17 15:32:36Z jlsaouli $
 * @author  Lyria
 * @see     LyConnection
 */
public class LyFileConnection extends LyConnection
{
	public static final String __VERSION =
		"$Id: LyFileConnection.java 35251 2012-04-17 15:32:36Z jlsaouli $";

	/** When reading file lines, check that default charset is correct. */
	public static boolean	 TEST_CHARSET = true;
	
	protected String		 _baseDirectory;

	/**
	 * Constructor.
	 * @param classInfo ClassInfo used by the request encoder.
	 */
	public LyFileConnection(LyApplication application, String baseDirectory)
	{
		super(application);

		if (baseDirectory == null)
			_baseDirectory = ".";
		else
			_baseDirectory = baseDirectory;
	}

	/**
	 * Return the lines describing the objects of the class. The first line contains information
	 * about the structure of the file.
	 * @return a list of Strings.
	 */
	public Vector<String> readFile(LyClass peerClass)
	{
		LyEnvironment environment = _application.getEnvironment();

		if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
			environment.logOut("Reading file for class " + peerClass.getId());

		// FFT00038 / V3.3
		// Get the path of the file and open an input stream
		String real_file_path = getFilePath(peerClass);
		String charset = getConnectionCharset();
		boolean fileEncoded = false;

		String file_path = real_file_path;
		File tmpFile = null;
		
		// DECODE FILE if necessary
		if (charset != null)
		{		
			try
			{
				LyFileTools tools = new LyFileTools(environment);

				tmpFile = tools.createTempFile("tmpFile", "", null, false);
				file_path = tmpFile.getAbsolutePath();
				String message = LyFileTools.decodeFile(environment, charset, real_file_path, file_path);
				fileEncoded = true;
				
				if (message != null)
				{
					file_path = real_file_path;
					environment.logError(message);
					fileEncoded = false;
				}
			}
			catch (IOException e1)
			{
				file_path = real_file_path;
				fileEncoded = false;
			}
		}
		boolean testCharSet = false; // FFT03820 : manage UTF-8 .res files
		
		if (charset == null)
		{
			charset = LyString.DEFAULT_CHARSET;
			testCharSet = TEST_CHARSET;
		}
		
		 Vector<String> lines = readLines(peerClass, environment, file_path, charset, fileEncoded, real_file_path);

		 if (testCharSet
			&& (lines != null)
			&& (lines.size() > 0))
		 {
		     boolean status = true;
		     int	 l, n = lines.size();
		     String  line;
		     
		     for (l = 0; (l < n) && status; l++)
		     {
	            try
	            {
			        line = lines.get(l);
	                status = LyString.testUTF8String(line);
	            }
	            catch (Exception e)
	            {
	                status = false;
	            }
		     }
		     
		     if (!status)
		     {
		         List<String> testedEncoding = new ArrayList<String>();
		                
		         testedEncoding.add("UTF-8");
		         testedEncoding.add("ISO-8859-1");
		         testedEncoding.remove(charset.toUpperCase());
		                
                // try other 'classic' encoding
		         Vector<String> tmpLines = null;
		         
		         while (!status && testedEncoding.size() > 0)
		         {
		             String tmpCharSet = testedEncoding.remove(0);
		             
		             try
		             {
		                 tmpLines = readLines(peerClass, environment, file_path, tmpCharSet, fileEncoded, real_file_path);
		                  
		                 status = true;
		                 for (l=0; (tmpLines != null) && (tmpLines.size() > 0) && (l<tmpLines.size()) && status; l++)
		                 {
		                    String tmpLine = tmpLines.get(l);
		                    
		                    try
		                    {
		                        status = LyString.testUTF8String(tmpLine);
		                    }
		                    catch (Exception e)
		                    {
		                        status = false;
		                    }
		                 }
                    }
                   catch (Exception e)
                   {
                   }
                   if (status && tmpLines != null && tmpLines.size() > 0)
                       lines = tmpLines;
		        }
		     }
		 }
		 
		if (tmpFile != null)
			tmpFile.delete();
		
		return lines;
	} // end method readFile

	
	protected Vector<String> readLines(LyClass peerClass, LyEnvironment environment, String file_path, 
	        String charset, boolean fileEncoded, String real_file_path)
	{
	       // Check if file does Exists
        File f = new File(file_path.toString());

        boolean     tryClasspath = environment.testEnv("LY_USE_FILES_FROM_CLASSPATH");
        if (!tryClasspath && !f.exists())
        {
            PrintWriter pw = openFile(peerClass);

            closeFile(peerClass, pw);
        }

        
        InputStream input = environment.getInputStream(file_path.toString(), !tryClasspath);

        // try (read only) from classpath
        if ((input == null) && (tryClasspath))
        {
            // FFT00038 / V3.3
            if (fileEncoded)
                file_path = real_file_path;
                
            String appId = _application.getId();

            appId = appId.replace('.', '/');

            String s = file_path.toString();

            s = LyString.replace(s, "\\", "/");
            s = LyString.replace(s, "#", "_"); // # is not supported in CLASSPATH file names.

            do
            {

                input = environment.getResourceAsStream(s, null);
                environment.logOut("SEARCH in classpath : " + s + " :: " + input);
                if (input == null)
                {
                    input = environment.getResourceAsStream(appId + "/" + s, null);
                    environment.logOut("SEARCH in classpath : " + appId + "/" + s + " :: " + input);
                }
                if ((input == null) && (s.length() > 0))
                {
                    int n = s.indexOf("/");

                    if ((n >= 0) && (n < s.length()))
                        s = s.substring(n + 1);
                    else
                        s = null; // Over
                }
            }
            while ((input == null) && (s != null));

            if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS) && (input != null))
                environment.logOut("* Found in classpath : " + s);
        }

        if (input == null) 
        {
            setError(environment.getMessage("LY_ER_GET_INPUT_STREAM", file_path));
            return null;
        }
            
        // FFT00038 / V3.3
        if (!real_file_path.equals(file_path))
            environment.logInfo("Reading file :" + real_file_path + " decoded into " + file_path);
        else
            environment.logInfo("Reading file :" + file_path);

        // Reading each line of the file
        Vector<String> lines = new Vector<String>();
        String line;
        
        while ((line = LyString.getLine(input, charset)) != null)
        {
            // Do not consider the empty lines nor the comment lines.
            if ((line.length() == 0)
                    || (line.charAt(0) == '#')
                    || (line.length() > 1 && line.charAt(0) == ((char) 65279) && (line
                            .charAt(1) == '#')))
                continue;
            if (line.startsWith(">>>") || line.startsWith("<<<"))
            {
                environment.logFatal("A conflict was discovered in data file "
                    + file_path + ", please solve it before restarting your application");
                throw new IllegalStateException("BF AppEngine : error while reading datas");
            }
            lines.addElement(line);
        }
        return lines;
	}
	
	/**
	 * Write lines in the file containing objects data.
	 * @return true if the operation was successful.
	 */
	public boolean writeFile(LyClass peerClass, Vector<String> lines)
	{
		// Get the path of file and directory
		StringBuffer file_path = new StringBuffer();
		String		 dir_path = getPath();

		file_path.append(getFilePath(peerClass));

		File		  file = new File(file_path.toString());
		File		  directory = new File(dir_path);
		LyEnvironment environment = _application.getEnvironment();

		// Try to create the directory if needed
		// TO DO : check if applet mode
		if (!directory.exists())
		{
			try
			{
				environment.logInfo("Creating directory :" + dir_path);
				directory.mkdirs();
			}
			catch (SecurityException e)
			{
				environment.logError("Unable to create directory :" + dir_path);
				setError(environment.getMessage("LY_ER_CREATE_DIR", dir_path.toString()));
				return false;
			}
		}
		else if (file.exists())
		{
			// Create bak file
			File bakFile = new File(file_path.toString() + ".bak");

			if (bakFile.exists())
				bakFile.delete();

			if (file.renameTo(bakFile))
				environment.logInfo("File :" + file_path + " renamed to .bak");
		}

		// Open the file
		FileOutputStream output;
		
		try
		{
			output = new FileOutputStream(file);
		}
		catch (Exception e)
		{
			environment.logError("Unable to open file :" + file_path);
			setError(environment.getMessage("LY_ER_OPEN_FILE", file_path.toString()));
			return false;
		}

		environment.logInfo("Saving file :" + file_path);

		//FFT00038 / V3.3
		// Write the file
		PrintWriter pw;// = new PrintWriter(output);

		try
		{
			// FFT00038 / V3.3
			String charset = getConnectionCharset();
			
			if (charset == null)
				charset = LyString.DEFAULT_CHARSET;
						
			pw = new PrintWriter(new BufferedWriter(
						new OutputStreamWriter(output, charset)), false);
		}
		catch (UnsupportedEncodingException e)
		{
			pw = new PrintWriter(output);
		}
		
		// Write header
		if (!writeHeader(peerClass, pw))
			return false;

		// Write each line
		int    i, nl = lines.size();
		String line;

		for (i = 0; i < nl; i++)
		{
			line = lines.elementAt(i);
			pw.println(line);
		}

		// Write Footer
		writeFooter(peerClass, pw);

		pw.close();
		return true;
	} // end method writeFile

	/**
	 * Open the file of peerClass and write the begining of file.
	 * @return the PrintWriter of file of peerClass.
	 */
	public PrintWriter openFile(LyClass peerClass)
	{
		// Get the path of file and directory
		StringBuffer file_path = new StringBuffer();
		String		 dir_path = getPath();

		file_path.append(getFilePath(peerClass));

		File		  file = new File(file_path.toString());
		File		  directory = new File(dir_path);
		LyEnvironment environment = _application.getEnvironment();

		// Try to create the directory if needed
		// TO DO : check if applet mode
		if (!directory.exists())
		{
			try
			{
				environment.logInfo("Creating directory :" + dir_path);
				directory.mkdirs();
			}
			catch (SecurityException e)
			{
				environment.logError("Unable to create directory :" + dir_path);
				setError(environment.getMessage("LY_ER_CREATE_DIR", dir_path.toString()));
				return null;
			}
		}
		else if (file.exists())
		{
			// Create bak file
			File bakFile = new File(file_path.toString() + ".bak");

			if (bakFile.exists())
				bakFile.delete();

			if (file.renameTo(bakFile))
				environment.logInfo("File :" + file_path + " renamed to .bak");
		}

		// Open the file
		OutputStream output;
		PrintWriter  pw;

		try
		{
			try
			{
				// FFT00038 / V3.3
				String charset = getConnectionCharset();
				
				if (charset == null)
					charset = LyString.DEFAULT_CHARSET;
				
				output = new FileOutputStream(file);
				Writer osw = new OutputStreamWriter(output, charset);

				pw = new PrintWriter(osw);
			}
			catch (UnsupportedEncodingException ex)
			{
				output = new FileOutputStream(file);
				pw = new PrintWriter(output);
			}
		}
		catch (FileNotFoundException e)
		{
			environment.logError("Unable to open file :" + file_path);
			setError(environment.getMessage("LY_ER_OPEN_FILE", file_path.toString()));
			return null;
		}

		environment.logInfo("Saving file :" + file_path);

		// Write header
		if (!writeHeader(peerClass, pw))
			return null;

		return pw;
	} // end method openFile

	/**
	 * Write lines in the file containing objects data.
	 * @return true if the operation was successful.
	 */
	public void writeLine(String line, PrintWriter pw)
	{
		pw.println(line);
		pw.flush();
	}

	/**
	 * Close the PrintWriter and write the end of file.
	 * @return true if the operation was successful.
	 */
	public void closeFile(LyClass peerClass, PrintWriter pw)
	{
		// Write Footer
		writeFooter(peerClass, pw);

		pw.close();		
		
		// Enode the file if necessary
		String file_path = getFilePath(peerClass);

		// FFT00038 / V3.3
		String charset = getConnectionCharset();
		
		if (file_path != null && charset != null)
		{
			LyEnvironment environment = _application.getEnvironment();
			
			// decode file
			String message = LyFileTools.encodeFile(environment, charset, file_path, file_path);
			
			if (message != null)
			{
				environment.logError(message);
			}
		}
	}

	// FFT00038 / V3.3
	/**
	 * Gets the connection charset. 
	 * 
	 * @return the charset if any, null otherwise
	 */
	public String getConnectionCharset() 
	{
		LyEnvironment environment = _application.getEnvironment();
		String lang = environment.getLanguage();
		String charset = null;
		
		if (lang != null)
			charset = environment.getCharset(lang, true);
		
		return charset;
	}

	/**
	 * Write header in the file containing objects data.
	 * @return true if the operation was successful.
	 */
	protected boolean writeHeader(LyClass peerClass, PrintWriter ps)
	{
		ps.println("#");
		ps.println("# Class \"" + peerClass.getId() + "\"");
		ps.println("#");

		return true;
	}

	/**
	 * Write header in the file containing objects data.
	 * @return true if the operation was successful.
	 */
	protected boolean writeFooter(LyClass peerClass, PrintWriter ps)
	{
		ps.println("# 	--------- END OF FILE -----------   #");

		return true;
	}

	/**
	 * Return the path of the file containing the data.
	 */
	protected String getFilePath(LyClass peerClass)
	{
		// Get the path of the directory containing the file
		StringBuffer file_path = new StringBuffer(getPath());

		file_path.append(_application.getEnvironment().getString().FILE_SEPARATOR);
		file_path.append((peerClass).getId());
		
		LyEnvironment  environment = _application.getEnvironment();

		if (environment.testEnv("LY_USE_FILES"))
		    file_path.append(".res");
		else if ((_provider != null)
			&& (_provider.getLocation() != null))
		{
			LyFileLocation.Type  type = ((LyFileLocation)_provider.getLocation()).getType();

		    if (type == LyFileLocation.Type.XML)
	            file_path.append(".xml");
	        else if (type == LyFileLocation.Type.CSV)
	            file_path.append(".csv");
	        else if (type == LyFileLocation.Type.INFO)
	            file_path.append(".res");           
		}
		else
            file_path.append(".res");           
			
		return file_path.toString();
	}

	/**
	 * Return the directory path of the file containing the data.
	 */
	protected String getPath()
	{
		return _baseDirectory;
	}

	/**
	 * Return the decoding type.
	 */
	protected int getDecodingType()
	{
        LyEnvironment  environment = _application.getEnvironment();
        
        if (environment.testEnv("LY_USE_FILES"))
            return LyFileRequestEncoder.RAW_ENCODING;
            
		if ((_provider != null)
		    && (_provider.getLocation() != null))
		{
	        LyFileLocation.Type type = ((LyFileLocation)_provider.getLocation()).getType();
	        
	        if (type == LyFileLocation.Type.XML)
	            return LyFileRequestEncoder.XML_EXPORT_ENCODING;
	        else if (type == LyFileLocation.Type.CSV)
	            return LyFileRequestEncoder.SPREADSHEET_ENCODING;
		}
		
        return LyFileRequestEncoder.RAW_ENCODING;
	}

	/**
	 * Return the list of field identifiers contained in the file.
	 */
	protected Vector<String> getFileFieldIds()
	{
		return null;
	}

	/**
	 * Return the logical class associated with this connection.
	 */
	protected LyClassInfo getLogicalClass()
	{
		return null;
	}

	/**
	 * Return the start line.
	 * @return the index of the first line of the imported file containing real data.
	 */
	public int getStartLine()
	{
		return 0;
	}
} // end class LyFileConnection
