/*
 * LyFileDataProvider.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.file;

import leon.app.*;
import leon.data.*;
import leon.info.location.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.util.*;


/**
 * This class is used by DataSources to transmit their requests to an instance of RequestEncoder.
 * @version $Id: LyFileDataProvider.java 35251 2012-04-17 15:32:36Z jlsaouli $
 * @author  Lyria
 * @see     LyRowObject
 * @see     LyTableClass
 * @see     LyColumnField
 * @see     LyFileRequestEncoder
 * @see     LyFileConnection
 */
public class LyFileDataProvider extends LyTableDataProvider
{
	public static final String __VERSION =
		"$Id: LyFileDataProvider.java 35251 2012-04-17 15:32:36Z jlsaouli $";

	public static boolean	SAVE_XML_FILES = false;
	
	/** The directory where the data files are saved. */
	protected String		   _baseDirectory;

	/** The separator for multiple value. */
	protected String		   _valueSep;

	/** The separator for fields. */
	protected String		   _fieldSep;


	/**
	 * Default constructor Initialization is done with the init method.
	 */
	public LyFileDataProvider(LyApplication application, LyLocation location)
	{
		super(application, location);
	}

	/**
	 * Constructor.
	 * @param      application
	 * @param      id
	 * @param      baseDirectory
	 * @param      valueSep
	 * @param      fieldSep
	 * @deprecated use LyFileDataProvider(LyApplication, LyLocation) instead.
	 */
	public LyFileDataProvider(LyApplication application, String id,
		String baseDirectory, String valueSep, String fieldSep)
	{
		this(application,
			new LyFileLocation(id, baseDirectory, LyFileLocation.Type.CSV, fieldSep, valueSep));
	}

	/**
	 * Sets location on provider.
	 * @param location New location for the provider.
	 */
	public void setLocation(LyLocation location)
	{
		if (!(location instanceof LyFileLocation))
			throw new IllegalArgumentException("BF AppEngine : invalid location type "
				+ location.getClass()
				+ " for file data provider (LyFileLocation expected)");

		if ((_location != null)
				&& (((LyFileLocation)_location).getType() != ((LyFileLocation)location).getType()))
			throw new IllegalArgumentException("BF AppEngine : invalid file location type "
				+ ((LyFileLocation)location).getType()
				+ ", it should be of the same type of the previous location "
				+ ((LyFileLocation)_location).getType());

		updateConnectionParameters(location);
		_valueSep = ((LyFileLocation)location).getValueSep();
		_fieldSep = ((LyFileLocation)location).getFieldSep();

		super.setLocation(location);
	}

	/**
	 * Updates this connection parameters using location information and application environment.
	 */
	protected void updateConnectionParameters(LyLocation location)
	{
		String		  url = location.getUrl();
		LyEnvironment env = _application.getEnvironment();

		_baseDirectory = env.getEnv(url);
		if (_baseDirectory == null)
			_baseDirectory = url;
		_baseDirectory = (env.getString()).expand(_baseDirectory, ".");
	}


	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		super.init();
		_baseDirectory = null;
		_valueSep = LyFileLocation.VALUE_SEP;
		_fieldSep = LyFileLocation.FIELD_SEP;
	}

	/**
	 * Create a specific request encoder for the specified dataSource.
	 * <P>This method is used internally, but subclasses should override this method and create an
	 * instance of their own (specific) request encoder.</P>
	 * <P>Default returns a LyRequestEncoder instance.</P>
	 * @param  dataSource reference to the dataSource associated to this request encoder.
	 * @return the specific LyRequestEncoder for the specified dataSource.
	 */
	public LyRequestEncoder getRequestEncoder(LyDataSource dataSource)
	{
		return new LyFileRequestEncoder(this, dataSource, _fieldSep, _valueSep);
	}

	/**
	 * This class is a factory to create a LyConnection instance. Current implementation only return
	 * null, so this method must be overriden. This instance depends on the nature of the connection
	 * and must provide services such as opening/closing the connection, and (specific) accessor to
	 * the connection service.
	 * <P>This method is ovberriden and used internally. Use getConnection() to access to a
	 * connection instance.</P>
	 * <P>Parent class LyDataProvider implements a connection pool so this Method may be called
	 * several times.</P>
	 * @return a new Connection instance. The connection instance is used in the specific LyRequest
	 *         encoder implementation.
	 * @see    #getConnection
	 * @see    #releaseConnection
	 */
	protected LyConnection createConnection()
	{
		return new LyFileConnection(_application, _baseDirectory);
	}

	/**
	 * Return the file data provider.
	 * @return the file data provider.
	 */
	protected LyDataProvider getSaveFileDataProvider()
	{
		if (SAVE_XML_FILES)
		{
			LyLocation       location = new LyFileLocation(_identifier, "LY_DATA_DIR", LyFileLocation.Type.XML, null, null);
			LyDataProvider   provider = LyDataProviderFactory.getProvider(_application, location);

			return provider;
		}
		
		return this;
	}

	/**
	 * Closes all connections from the connection pool.
	 * @return true if the request completed successfully.
	 * @see    #createConnections(int)
	 */
	public boolean close()
	{
		LyEnvironment environment = getEnvironment();

		if (environment.testEnv("LY_SAVE_FILES"))
		    // If LY_SAVE_FILES == true we let parent do the job.
			return super.close();

		if (environment.testEnv("_LY_SAVE_DATA")
				&& _application.isValid())
			save(false);

		return super.close();
	}
} // end class LyFileDataProvider
