/*
 * LyFileInfoDataProvider.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.file;

import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.*;
import leon.info.location.*;
import leon.peer.*;
import leon.peer.util.*;


/**
 * This class is used by DataSources to transmit their requests to an instance of RequestEncoder.
 * @version $Id: LyFileInfoDataProvider.java 35251 2012-04-17 15:32:36Z jlsaouli $
 * @author  Lyria
 * @see     LyRowObject
 * @see     LyTableClass
 * @see     LyColumnField
 * @see     LyFileRequestEncoder
 * @see     LyFileConnection
 */
public class LyFileInfoDataProvider extends LyFileDataProvider
{
	public static final String __VERSION =
		"$Id: LyFileInfoDataProvider.java 35251 2012-04-17 15:32:36Z jlsaouli $";


	/**
	 * Default constructor Initialization is done with the init method.
	 */
	public LyFileInfoDataProvider(LyApplication application, LyLocation location)
	{
		super(application, location);
	}

	/**
	 * Constructor.
	 * @param      application
	 * @param      id
	 * @param      baseDirectory
	 * @deprecated use LyFileInfoDataProvider(Lyapplication, LyLocation) instead.
	 */
	public LyFileInfoDataProvider(LyApplication application, String id,
		String baseDirectory)
	{
		this(application,
			new LyFileLocation(id, baseDirectory, LyFileLocation.Type.INFO, null, null));
	}


	/**
	 * Validates the field info of the given class info : this mainly consists to create or retrieve
	 * the associated peer field.
	 * @param  classInfo Class info to which the field info belongs to.
	 * @param  fieldInfo Field info that is validated.
	 * @return True if the validation was successful, false otherwise.
	 */
	public void validate(LyClassInfo classInfo, LyFieldInfo fieldInfo)
	{
		// Get the binding of the LyBinding.
		String		   binding = fieldInfo.getId();
		LyClassBinding classBinding = classInfo.getBinding();
		LyPeerClass    peerClass = (LyPeerClass)classBinding.getPeerClass();
		LyPeerField    peerField = (LyPeerField)peerClass.getField(binding);

		if (peerField == null)
		{
			// Create the field peer.
			peerField = createPeerField(binding, fieldInfo, classInfo);

			addPeerField(peerClass, peerField);
			if (isIdField(peerClass, peerField, fieldInfo.getBinding()))
				addIdPeerField(peerClass, peerField);
		}

		// Associate binding and field peer.
		registerPeerField(classBinding, fieldInfo, peerField);
	}

	/**
	 * Determine the class for memory storage associated to a columnField. This method is internally
	 * used by the validate method. Method should be overriden for optimization by subclasses.
	 * @param  tableClass  the peer table class reference as specified in validate method.
	 * @param  columnField peer field associated with the value we store.
	 * @param  fieldInfo   data description of the meta model.
	 * @return the java class used to store data in memory.
	 * @see    #validate()
	 */
	protected Class<?> getValueClass(LyTableClass tableClass, LyColumnField columnField,
		LyFieldInfo fieldInfo)
	{
		Class<?> valueClass = null;

		if (fieldInfo instanceof LyRelationFieldInfo)
		{
			if (fieldInfo.isMultiple())
				valueClass = LyObjectId[].class;
			else
				valueClass = LyObjectId.class;
		}
		else if (fieldInfo instanceof LyNumberFieldInfo)
			valueClass = fieldInfo.getValueClass();
		else if (fieldInfo instanceof LyDateFieldInfo)
		{
			valueClass = fieldInfo.getValueClass();
			columnField.setFormat(((LyDateFieldInfo)fieldInfo).getStringDateFormat());
		}
		else if (fieldInfo instanceof LyArrayFieldInfo)
			valueClass = String[].class;
		else
		{
			if (fieldInfo.isMultiple())
				valueClass = String[].class;
			else
				valueClass = String.class;
		}

		return valueClass;
	} // end method getValueClass

	/**
	 */
	public LyConnection createConnection()
	{
		return new LyFileInfoConnection(_application, _baseDirectory);
	}
} // end class LyFileInfoDataProvider
