/*
 * LyFileRequestEncoder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.file;

import java.io.*;
import java.text.*;
import java.util.*;

import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.data.*;
import leon.info.*;
import leon.info.location.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.util.*;


/**
 * This class is used for local files as data provider. Methods for creation, deletion and
 * modification are not overriden since they don't have to write anything in the corresponding
 * files. The specific method saveObjects must be called in order to make the objects persistent in
 * a file.
 * @version $Id: LyFileRequestEncoder.java 33278 2011-11-24 15:22:59Z nbodin $
 * @author  Lyria
 * @see     LyRequestEncoder
 */
public class LyFileRequestEncoder extends LyRequestEncoder
{
	public static final String __VERSION =
		"$Id: LyFileRequestEncoder.java 33278 2011-11-24 15:22:59Z nbodin $";


	/** Value for raw encoding (used in encodeValue). */
	public static final int    RAW_ENCODING = 1;

	/** Value for text encoding (used in encodeValue). */
	public static final int    TEXT_ENCODING = 2;

	/** Value for spreadsheet encoding (used in encodeValue). */
	public static final int    SPREADSHEET_ENCODING = 3;

	/** Value for export encoding (used in encodeValue). */
	public static final int    EXPORT_ENCODING = 4;

	/** Value for xml encoding (used in encodeValue). */
	public static final int    XML_EXPORT_ENCODING = 5;
      
    /** Empty char set. */
    public static final String EMPTY_CHARSET = "EmptyCharSet";
    

	/** Separator used between fields. */
	public String			   _fieldSep;

	/** Separator used between values. */
	public String			   _valueSep;

	/** Current environment if any, otherwise use dataprovider environment. */
	public LyEnvironment	   _environment;

	/**  */
	protected String		   _headerMark;
    
    /** */
    protected String           _charSet;


	/**
	 * Constructor.
	 * @param classInfo ClassInfo used by the request encoder.
	 * @param fieldSep  separator used between fields.
	 * @param valuedSep separator used between values.
	 */
	public LyFileRequestEncoder(LyDataProvider provider, LyDataSource dataSource,
		String fieldSep, String valueSep)
	{
		super(provider, dataSource);

		if (fieldSep != null)
			_fieldSep = fieldSep;
		if (valueSep != null)
			_valueSep = valueSep;

		_headerMark = getEnvironment().getString().EXPORT_HEADER_MARK;
        _charSet = null;
	}

	/**
	 * Sets the header mark.
	 * @param headerMark new header mark.
	 */
	public void setHeaderMark(String headerMark)
	{
		_headerMark = headerMark;
	}

	/**
	 * Sets the fields separator.
	 * @param fieldSep new fields separator.
	 */
	void setFieldSeparator(String fieldSep)
	{
		if (fieldSep == null)
			throw new IllegalArgumentException("BF AppEngine : "
				+ getEnvironment().getMessage("LY_ER_CANT_BE_NULL", "fieldSep"));
		_fieldSep = fieldSep;
	}

	/**
	 * Sets the values separator.
	 * @param valueSep new values separator.
	 */
	void setValueSeparator(String valueSep)
	{
		if (valueSep == null)
			throw new IllegalArgumentException("BF AppEngine : "
				+ getEnvironment().getMessage("LY_ER_CANT_BE_NULL", "valueSep"));
		_valueSep = valueSep;
	}


	/**
	 * Gets the fields separator.
	 * @return the fields separator.
	 */
	public String getFieldSeparator()
	{
		return _fieldSep;
	}

	/**
	 * Gets the values separator.
	 * @return the values separator.
	 */
	public String getValueSeparator()
	{
		return _valueSep;
	}


	/**
	 * Initialisation of the data of the instance.
	 */
	public void init()
	{
		super.init();

		_fieldSep = LyFileLocation.FIELD_SEP;
		_valueSep = LyFileLocation.VALUE_SEP;
		// Synchronous by default
		_synchronous = true;
		_environment = null;
	}

	/**
	 * Set the request encoder response. Inherited method from LyRequestEncoder to set properties to
	 * the response.
	 * @param the response
	 */
	public void setResponse(LyResponse response)
	{
		_response = response;

		// Set properties of the response.
		_response.setSorted(false);
		_response.setFiltered(false);
		_response.setCounted(true);
		_response.setDistinctValues(true);
	}

	/**
	 * Get an object.
	 * @param  objectId object identifier
	 * @param  fields   the fields to load (if null, load all)
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getObject(LyObjectId objectId, Vector<LyField> fields)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// Check the identifier
		if (objectId == null)
		{
			_response.setError(getEnvironment().getMessage("LY_ER_CANT_BE_NULL", "objectId"));
			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			return false;
		}

		// No need to encode the value
		// Get the data from the connection as a list of lines.
		Vector<String>     lines = readFromConnection();

		// Decode the values (the lines) returned by the file connection
		Vector<Object>     objects = decodeValues(lines, fields);

		// Parse the objects to find out the good one.
		int		   i, n = ((objects == null) ? 0 : objects.size());
		Object     object = null;
		LyObjectId id, peerId;

		peerId = getPeerId(objectId);
		for (i = 0; i < n; i++)
		{
			object = objects.elementAt(i);
			id = _class.getObjectId(object);

			if (id == null)
				continue;

			if (id.equals(peerId))
			{
				// If the object was found, add it to the response. Else, add nothing.
				_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object, null);
				return true;
			}
		}

		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);

		return true;
	} // end method getObject

	/**
	 */
	protected LyObjectId getPeerId(LyObjectId objectId)
	{
		if (objectId == null)
			return null;

		Vector<? extends LyField>   idFields;
		String[] values;
		int		 n, p;

		values = objectId.toStrings();
		idFields = _class.getIdFields();

		n = idFields.size();
		p = values.length;

		if (n != p)
			throw new IllegalStateException("BF AppEngine : "
				+ getEnvironment().getMessage("LY_ER_CONDITION_OBJECT_ID", objectId.toName(),
					_class.getId()));

		return new LyObjectId(_class, values);
	}

	/**
	 * Get a set of objects.
	 * @param  fields the fields to load (if null, load all)
	 * @param  sort   indicates sorting criterion, or null if default sort is used.
	 * @param  filter indicates filtering criterion, or null if there is no filter.
	 * @param  n      maximum number of objects to get.
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getObjects(Vector<LyField> fields, LySort sort, LyFilter filter, int n)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// No need to encode the value
		// Get the data from the connection as a list of lines.
		Vector<String> lines = readFromConnection();

		if (lines == null)
		{
			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			return false;
		}

		Vector<Object> objects = null;

		try
		{
			// Decode the values (the lines) returned by the file connection
			objects = decodeValues(lines, fields);
		}
		catch (Exception e)
		{
			LyEnvironment	env = getEnvironment();
			
			env.logError("[LyFileRequestEncoder] Error during parsing of file");
			env.writeException(e);
			_response.setError(e.getMessage());
			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
		}

		// Sort the objects
		// TODO

		// Keep the first n objects matching the filter (if given)
		int    i, nf = 0, no = ((objects == null) ? 0 : objects.size());
		Object object;

		for (i = 0; i < no; i++)
		{
			object = objects.elementAt(i);

			nf++;
			_dataSource.receiveResponse(_response, LyResponse.Status.RUNNING, object, null);

			// If the maximum size is reached
			if ((n > 0) && (nf >= n))
				break;
		}

		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);

		return true;
	} // end method getObjects

	/**
	 * Return the number of objects from the file.
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getSize()
	{
		return (getSize(null));
	}

	/**
	 * Return the number of objects from the file .
	 * @param  filter indicates filtering criterion, or null if there is no filter.
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getSize(LyFilter filter)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// No need to encode the value
		// Get the data from the connection as a list of lines.
		Vector<String> lines = readFromConnection();

		// Decode the values (the lines) returned by the file connection
		Vector<Object> objects = decodeValues(lines, null);
		int            nf = ((objects == null) ? 0 : objects.size());

		/* TODO : Keep the first n objects matching the filter (if given)
		int             i, nf = 0;
		Object  object;

		for (i = 0; i < no; i++)
		{
		        object = objects.elementAt(i);
		        nf++;
		}
*/

		// Add the Integer to the response
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, new Integer(nf), null);
		return true;
	} // end method getSize

	/**
	 * To create a new object.
	 * @param  object the new object to create
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean createObject(Object object)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// If the object has no Id, we try to get one.
		LyValueSet values = computeIds(object);
		LyValue    value;
		int		   i, n = ((values == null) ? 0 : values.getSize());

		for (i = 0; i < n; i++)
		{
			value = values.getFieldValue(i);
			_class.setValue(object, value.getField(), value.getValue());
		}

		// Validate the creation.
		validateChanges(object);

		// Send the response.
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object, null);

		return true;
	}

	/**
	 * Compute id for the given object.
	 */
	protected LyValueSet computeIds(Object object)
	{
		return computeIds(object, 0);
	}
	
	/**
	 * Compute id for the given object.
	 */
	protected LyValueSet computeIds(Object object, int offset)
	{	
		Vector<?> idFields = _class.getIdFields();

		if ((idFields == null) || (idFields.size() == 0))
			return null;

		int							   i, n = idFields.size();
		LyField						   field;
		Object						   value;
		Class<?>					   valueClass;
		LyApplication				   application = _provider.getApplication();
		LyApplicationBehaviorInterface behavior = application.getApplicationBehavior();

		// Maybe a composite id may have some null values. Verify that all values
		// are null before compute them.
		if (object != null)
		{
			for (i = 0; i < n; i++)
			{
				field = (LyField)idFields.elementAt(i);
				try
				{
					value = _class.getValue(object, field);
				}
                catch (LyValueNotLoadedException e)
				{
					// If the value is empty we receive this exception, normal case, handle it as null
					value = null;
				}
				if (value != null)
					return null;
			}
		}

		// All values for this id are null, we must compute a valid id.
		LyValueSet values = new LyValueSet();
		
		for (i = 0; i < n; i++)
		{
			field = (LyField)idFields.elementAt(i);
			valueClass = field.getValueClass();
			
			if (Number.class.isAssignableFrom(valueClass))
			{          
                if (valueClass == Integer.class)
                	value = behavior.getNextInt(application, _class);
                else if (valueClass == Short.class)
                	value = behavior.getNextShort(application, _class);
                else if (valueClass == Long.class)
                	value = behavior.getNextLong(application, _class);
                else if (valueClass == Float.class)
                	value = behavior.getNextFloat(application, _class);
                else
                	value = behavior.getNextDouble(application, _class);
			}
			else if (valueClass == Date.class)
			{
				value = new Date();
				// Be sure to have a different date next time
				try { Thread.sleep(1); } catch (Exception e) {}
			}
			else if (valueClass == String.class)
				value = Long.toString(behavior.getNextInt(application, _class));
			else
				throw new IllegalStateException("BF AppEngine : can't generate an id for field "
					+ field + " with type " + valueClass
					+ " belonging to class " + _class);

			values.addValue(field, value);
		}

		return values;
	} // end method computeIds

	/**
	 * To set an object.
	 * @param  object the object to set
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean setObject(Object object)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// Validate the modification.
		LyValueSet oldValues =  validateChanges(object);

		// Send the response.
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object, oldValues);

		return true;
	}

	/**
	 * To set a set of objects.
	 * @param  objects the objects to set
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean setObjects(Vector<Object> objects)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		Object     object;
		int		   i, n;
		LyValueSet oldValues;

		n = objects.size();

		for (i = 0; i < n; i++)
		{
			object = objects.elementAt(i);

			// Validate the modification.
			oldValues = validateChanges(object);

			// Send the response.
			_dataSource.receiveResponse(_response, LyResponse.Status.RUNNING, object, oldValues);
		}

		// Send the last response.
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);

		return true;
	}

	/**
	 * To delete an object.
	 * @param  object the object to delete
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean deleteObject(Object object)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object, null);
		return true;
	}

	/**
	 * Method used to saved all the objects contained in the corresponding data source cache.
	 * @param  fields fields to save (if null save all).
	 * @param  type   indicates the file format
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean saveObjects(Vector<Object> objects, Vector<LyField> fields, int type)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// Save the data in the file		
		writeToConnection(objects, fields, type);

		_response.setStatus(LyResponse.Status.COMPLETED);

		return true;
	}

	/**
	 * Sends a request to the file connection for reading data.
	 * @return lines read from the file.
	 */
	public Vector<String> readFromConnection()
	{
		// Get a valid connection
		LyFileConnection connection;

		connection = (LyFileConnection)_provider.getConnection();

		if (connection == null)
			return null;
        
        // Inhibate controls on objects to enhance performance
        LyObject.DEFAULT_CONTROL = LyObject.CONTROL_NONE;

		// Get the data from the connection as a list of lines.
		Vector<String> lines = connection.readFile(_class);
        
        // Restore controls on objects to enhance performance
        LyObject.DEFAULT_CONTROL = LyObject.CONTROL_FULL;

		_provider.releaseConnection(connection);

		return lines;
	}

	/**
	 * Sends a request to the file connection for writing data.
	 * @param  lines lines to write on the file.
	 * @return lines read from the file.
	 */
	public boolean writeToConnection(Vector<String> lines)
	{
		// Get a valid connection
		LyFileConnection connection;

		connection = (LyFileConnection)_provider.getConnection();

		if (connection == null)
			return false;

		// Write the data in the file.
		boolean result = connection.writeFile(_class, lines);

		_provider.releaseConnection(connection);

		return result;
	}

	/**
	 * Sends a request to the file connection for writing data.
	 * @param  objects the set of objects to encode.
	 * @param  fields  FieldInfos to save (if null save all).
	 * @param  type    indicates the file format.
	 * @return lines read from the file.
	 */
	public boolean writeToConnection(Vector<Object> objects, Vector<LyField> fields, int type)
	{
		// Get a valid connection
		LyFileConnection connection;

		connection = (LyFileConnection)_provider.getConnection();

		if (connection == null)
			return false;

		// Check type
	    if (type < 0)
            type = connection.getDecodingType();
	      
		// Write the data in the file.
		Vector<? extends LyField> collect;

		if ((fields == null) || (fields.size() == 0))
			collect = _class.getFields();
		else
			collect = fields;

		PrintWriter pw = connection.openFile(_class);
		boolean     result = false;
		String	    line;

		if (pw != null)
		{
			// Add header
			connection.writeLine(encodeHeader(collect, type), pw);

			// Add the lines
			int    i, no = ((objects != null) ? objects.size() : 0);
			Object object;

			for (i = 0; i < no; i++)
			{
				object = objects.elementAt(i);
				if (object == null)
					continue;

				line = encodeValue(object, collect, type);
				connection.writeLine(line, pw);
			}

			connection.closeFile(_class, pw);
			result = true;
		}

		//boolean result = connection.writeFile(_class, lines);

		_provider.releaseConnection(connection);

		return result;
	} // end method writeToConnection

	/**
	 *  Methods for decoding *
	 */

	/**
	 * Read a set of lines and return a set of objects.
	 * @param  lines  lines to read.
	 * @param  fields the fields to load (if null, load all)
	 * @return the objects read from the lines.
	 */
	public Vector<Object> decodeValues(Vector<String> lines, Vector<LyField> fields)
	{
		return decodeValues(lines, fields, true);
	}

	/**
	 * Read a set of lines and return a set of objects.
	 * @param  lines       lines to read.
	 * @param  fields      the fields to load (if null, load all)
	 * @param  checkObject indicates if we must check if object is already in data source
	 * @return the objects read from the lines.
	 */
	public Vector<Object> decodeValues(Vector<String> lines, Vector<LyField> fields, boolean checkObject)
	{
		if (lines == null)
			return null;

		// Reading each line of the file
		StringBuffer str_buffer = null;
		int			 idx, length, l, nl = lines.size(), i, n;
		Vector<Object>		 objects;
        Vector<LyField> fields_to_load, fields_in_file, missing_fields;
		String		 line;
		boolean		 first = true;
		LyField		 field;

		// Retrieve the fields that must be loaded
		fields_to_load = ((fields == null) ? new Vector<LyField>(_class.getFields()) : fields);
		fields_in_file = new Vector<LyField>(_class.getFields());
		missing_fields = new Vector<LyField>();

		// Get a valid connection to check if fields are given by the connection
		LyFileConnection connection = (LyFileConnection)_provider.getConnection();
		boolean			 fieldsGiven = false;
		boolean			 spreadSheetDecoding = false;
		int				 startLine = 0;

		if (connection != null)
		{
			Vector<String> fieldIds = connection.getFileFieldIds();

			startLine = connection.getStartLine();
			spreadSheetDecoding = (connection.getDecodingType() == SPREADSHEET_ENCODING);

			if ((fieldIds != null) && (fieldIds.size() > 0))
			{
				String fieldId;

				fieldsGiven = true;
				fields_in_file = new Vector<LyField>();

				for (i = 0; i < fieldIds.size(); i++)
				{
					fieldId = fieldIds.elementAt(i);
					fields_in_file.addElement(_class.getField(fieldId));
				}
			}

			_provider.releaseConnection(connection);
		}

		objects = new Vector<Object>(nl);

		for (l = Math.max(startLine, 0); l < nl; l++)
		{
			line = lines.elementAt(l);

			// Ignore tabulations at the beginning of the line
			length = line.length();
			for (idx = 0; (idx < length) && (line.charAt(idx) == '\t'); idx++)
				;
			if (idx > 0)
				line = line.substring(idx);

			if (str_buffer != null)
			{
				str_buffer.append(line);
				line = str_buffer.toString();
			}

			length = line.length();
			str_buffer = null;

			// Go through empty lines and comment lines
            if ((length == 0)
                    || (line.charAt(0) == '#')
                    || (line.length() > 1 && line.charAt(0) == ((char) 65279) && (line
                            .charAt(1) == '#')))
                continue;

			// Read the header and keep only fields that must be loaded
			if ((first && (_headerMark == null))
					|| line.startsWith(_headerMark))
			{
				if (fieldsGiven)
					continue;

				first = false;
				fields_in_file = decodeHeader(line);

				// For each field that must be loaded, check if found in the file
				for (i = 0, n = fields_to_load.size(); i < n; i++)
				{
					field = fields_to_load.elementAt(i);
					if (!fields_in_file.contains(field))
					{
						getEnvironment().logWarning("Field :" + field
							+ " not found for class :" + _class.getId() + " in data file.");
						missing_fields.addElement(field);
					}
				}

				// Read datas
				continue;
			}
			
			try
			{
				// Create a ValueSet with the content of the line
				LyValueSet values = decodeValue(line, fields_in_file, l + 1, spreadSheetDecoding);
	
				if (values == null) // An error occured on the object
					continue;
	
				// Add values for missing fields
				for (i = 0, n = missing_fields.size(); i < n; i++)
				{
					field = missing_fields.elementAt(i);
					values.addValue(missing_fields.elementAt(i), null);
				}
				// Remove values for non requested fields
				for (i = 0, n = fields_in_file.size(); i < n; i++)
				{
					field = fields_in_file.elementAt(i);
					if (!fields_to_load.contains(field))
						values.removeValue(field);
				}
	
				if (spreadSheetDecoding)
				{
					// Check if id fields are in the list of loaded fields
					Vector<? extends LyField>	idFields = _class.getIdFields();
					
					LyDataSession    session = _provider.getApplication().getCurrentSession();
					LyClassWorkSpace ws;
					LyObject		 object;
	
					for (i = 0, n = idFields.size(); i < n; i++)
					{
						field = (LyField)idFields.elementAt(i);
	
						if (!values.contains(field))
						{
							// Try to retrieve object by its name
							LyClassInfo		   		classInfo = connection.getLogicalClass();
							LyInfoList<LyFieldInfo>	nameFieldInfos = classInfo.getNameFieldInfos();
							LyFieldInfo		   		nameFieldInfo;
							LyValue			   		value;
							LyFilter		   		filter = new LyFilter(LyFilter.Type.AND);
							LyFilterExpression 		exp;
							int				   		j, m = nameFieldInfos.getSize();
	
							for (j = 0; j < m; j++)
							{
								nameFieldInfo = (LyFieldInfo)nameFieldInfos.getInfo(j);
								value = values.getFieldValue(nameFieldInfo.getId());
	
								if (value != null)
								{
									exp = new LyFilterExpression(nameFieldInfo, value.getString());
									filter.addElement(exp);
								}
							}
	
							ws = new LyClassWorkSpace(session, classInfo, filter, null, false, false);
	
							// If at least one object was found, use its identifier
							if (ws.getSize() > 0)
							{
								object = ws.getObject(0);
								values.addValue(field, object.getValue(field.getId()));
							}
							// Otherwise generate an identifier for the new object
							else
							{
								LyValueSet idValues = computeIds(null, objects.size());
	
								values.addValues(idValues);
							}
						}
					}
				}
	
				// Create the object with the set of values
				Object object = buildObject(values, checkObject);
	
				if ((objects != null) && (object != null))
					objects.addElement(object);
			}
			catch (RuntimeException e)
			{
				if (_provider instanceof LyImportFileDataProvider)
				{
					// When importing data from files, do not stop import process if an error
					// occurs : write the exception and keep on processing data.
					LyImportFileDataProvider	importProvider = (LyImportFileDataProvider)_provider;
					LyEnvironment				env = getEnvironment();
					
					env.logError("[LyFileRequestEncoder] Error during parsing of file");
					env.writeException(e);
					importProvider.addErrorLine(l + 1);
				}
				else
					throw e;
			}
		}

		return objects;
	} // end method decodeValues

	/**
	 */
	protected Object buildObject(LyValueSet values, boolean checkObject)
	{
		if (values == null)
			return null;

		// Create the object with the set of values
		LyRowObject object = new LyRowObject((LyTableClass)_class, values);

		object.validate(null);

		LyObjectId objectId = _class.getObjectId(object);

		if (checkObject
		    && _dataSource.checkObject(objectId))
		{
			// The object is already in the cache. We take its values and we set them
			// to the load object because in case of files we don't save the datas
			// during the life of the application. So the cached object is more
			// recent than the file object.		    
		    try
		    {
		        LyCache     cache = _dataSource.getCache();
    			LyRowObject cacheObject = (LyRowObject)cache.getObject(objectId);
    			LyValueSet  cacheValues = cacheObject.getValues();
    
    			object.validate(cacheValues);
		    }
		    catch (LyNoSuchObjectException e)
		    {
		        // Can't happen, we checked before that the object is in cache
		    }
		}

		return object;
	}

	/**
	 * Read a line and return a set of FieldInfos. The line must contain the attributes names
	 * separed by a specific separator.
	 * @param  line line to read.
	 * @return the set of FieldInfos read from the lines.
	 */
	public Vector<LyField> decodeHeader(String line)
	{
		if (line == null)
			return null;

		Vector<LyField>  fields = new Vector<LyField>();
		String  buffer;
		LyField field;
		int     idx, idx1 = ((_headerMark == null) ? 0 : _headerMark.length());
		int     length = line.length();
		int     idx2 = 0;
		int     l1 = _fieldSep.length();

		for (idx = idx1; idx < length; idx = idx2 + l1)
		{
			idx2 = line.indexOf(_fieldSep, idx);
			if (idx2 < 0)
				idx2 = length;

			buffer = line.substring(idx, idx2);
			buffer = asciiDecode(buffer);
			field = _class.getField(buffer);

			if (field == null)
				getEnvironment().logWarning("Field \"" + buffer
					+ "\" unknown on class " + _class.getId());

			// We add the field  even it is null (in this case we ignore it).
			fields.addElement(field);
		}

		return fields;
	} // end method decodeHeader

	/**
	 * Read a line and return a set of values.
	 * @param  line   line to read.
	 * @param  fields the list of FieldInfos.
	 * @param  l      position of line in the file.
	 * @param  spreadSheetDecoding indicates if spreadsheet decoding mode is set
	 * @return the set of values read from the lines.
	 */
	public LyValueSet decodeValue(String line, Vector<LyField> fields, int l, boolean spreadSheetDecoding)
	{
		LyField       field;
		String        buffer;
		int		      i, idx, n = fields.size();
		int		      idx2 = 0;
		int		      l1 = _fieldSep.length();
		int		      length = line.length();
		Object        value;
        LyValueSet    values = new LyValueSet(false);
        char          valueSep = LyTools.charFromString(_valueSep);

		for (i = 0, idx = 0; i < n; i++, idx = idx2 + l1)
		{
			field = fields.elementAt(i);
			idx2 = line.indexOf(_fieldSep, idx);

			if (idx2 < 0)
			{
				if (i == (n - 1)) // No separator at the end of line
					idx2 = length;
				else
				{
					getEnvironment().logError("[LyFileRequestEncoder] Error during parsing of file."
						+ " Missing separator at line " + l + " for attribute \"" + field.getId() + "\".");
					return null;
				}
			}

			if (idx2 == idx)
				buffer = null;
			else
			{
				buffer = line.substring(idx, idx2);

				// Replace the default separator used by the toString method with _valueSep.
				buffer = buffer.replace(valueSep, LyTools.ARRAY_SEP);
				
				if (!spreadSheetDecoding)
					buffer = asciiDecode(buffer);
			}

			if (field == null)
				continue;

			value = fromString(field, buffer, spreadSheetDecoding);
            values.addValue(field, value);
		}

		if (i < n) // An error occured on the object
			return null;
        
		return values;
	} // end method decodeValue

	/**
	 *
	 */
	public Object fromString(LyField field, String str)
	{
		return fromString(field, str, false);
	}

	/**
	 */
	public Object fromString(LyField field, String str, boolean spreadSheetDecoding)
	{
		if ((field == null) || (str == null))
			return null;

		Class<?>  valueClass;
		Object value;

		valueClass = field.getValueClass();

        if (spreadSheetDecoding)
        {
            // Retrieve the decoding type
            LyConnection        connection = _provider.getConnection();
            
    		if ((connection instanceof LyFileConnection)
    		    && (((LyFileConnection)connection).getLogicalClass() != null))
    		{
    		    // For spreadsheet decoding, use logical class to decode values
    		    LyFileConnection  fileConnection = (LyFileConnection)connection;
				LyClassInfo       classInfo = fileConnection.getLogicalClass();
				LyFieldInfo       fieldInfo = classInfo.getFieldInfo(field.getId());

				if (fieldInfo instanceof LyRelationFieldInfo)
				{
					// For relations, retrieve related objects by their name
					LyRelationFieldInfo		relationFieldInfo = (LyRelationFieldInfo)fieldInfo;
					String[]				objectNames = LyString.stringToList(str, LyTools.ARRAY_SEP);
					LyObject				object;
					LyDataSession			session = _provider.getApplication().getCurrentSession();
					LyObjectList			objects = session.getObjectsByName(relationFieldInfo.getClassInfos(), objectNames);
					
					if ((objects != null) && (objects.getSize() > 0))
					{
						if (relationFieldInfo.isMultiple())
						{
							LyObjectId[]	objectIds = new LyObjectId[objects.getSize()];
							int				i, n = objects.getSize();
							
							for (i = 0; i < n; i++)
							{
								object = objects.getObject(i);
								objectIds[i] = object.getObjectId();
							}
							
							return objectIds;
						}
						else
						{
							object = objects.getObject(0);
							
							return object.getObjectId();
						}
					}
				}
				else if (fieldInfo instanceof LyChoiceFieldInfo)
				{
					// For a choice field, retrieve corresponding options by their name
					LyChoiceFieldInfo choiceFieldInfo = (LyChoiceFieldInfo)fieldInfo;
					String[]		  choiceValues = LyString.stringToList(str, LyTools.ARRAY_SEP);
					String			  choiceValue, choiceItemValue;
					Vector<String>	  choiceIds = new Vector<String>();
					int				  i, n = ((choiceValues == null) ? 0 : choiceValues.length);
					short			  j, m = choiceFieldInfo.getChoiceItemCount();
					LyChoiceItem	  choiceItem;
					LyDataSession     session = _provider.getApplication().getCurrentSession();

					for (i = 0; i < n; i++)
					{
						choiceValue = choiceValues[i];

						for (j = 0; j < m; j++)
						{
							choiceItem = choiceFieldInfo.getChoiceItem(j);
							choiceItemValue = choiceItem.getName();
							choiceItemValue = session.getEnvironment().translate(choiceItemValue);

							if ((choiceItemValue != null) && (choiceItemValue.equals(choiceValue)))
							{
								choiceIds.addElement(choiceItem.getId());
								break;
							}
						}
					}

					if (choiceIds.size() > 0)
					{
						if (choiceFieldInfo.isMultiple())
						{
							n = choiceIds.size();

							String[] itemIds = new String[n];

							for (i = 0; i < n; i++)
								itemIds[i] = choiceIds.elementAt(i);

							return itemIds;
						}
						else
							return choiceIds.elementAt(0);
					}
				}
				else if (Number.class.isAssignableFrom(valueClass))
				{
					// Replace decimal separator by a dot
					LyDataSession		 session = _provider.getApplication().getCurrentSession();
					LyEnvironment		 environment = session.getEnvironment();
					DecimalFormatSymbols format = new DecimalFormatSymbols(environment.getLocale());
					char				 decimalSeparator = format.getDecimalSeparator();

					if (str != null)
					{
						str = str.replace(LyTools.ARRAY_SEP, '.');
						str = str.replace(decimalSeparator, '.');
					}
				}
			}
            
            if (connection != null)
                _provider.releaseConnection(connection);
		}
		
		if (valueClass.isArray())
		{
			Object[] result;
			int		 i, l;

			value = LyTools.ArrayFromString(str, String.class);
			l = ((Object[])value).length;
			if (valueClass == LyObjectId[].class)
				result = new LyObjectId[l];
			else
				result = new String[l];

			for (i = 0; i < l; i++)
			{
				if (valueClass == LyObjectId[].class)
					result[i] = new LyObjectId(_provider.getApplication(),
							(String)((Object[])value)[i]);
				else
					result[i] = ((Object[])value)[i];
			}

			value = result;
		}
		else
		{
			if (valueClass == LyObjectId.class)
			{
			    if (spreadSheetDecoding)
			        value = str;
			    else
			        value = new LyObjectId(_provider.getApplication(), str);
			}
			else if (valueClass == Integer.class)
				value = LyTools.IntegerFromString(str);
			else if (valueClass == Short.class)
				value = LyTools.ShortFromString(str);
			else if (valueClass == Long.class)
				value = LyTools.LongFromString(str);
			else if (valueClass == Double.class)
				value = LyTools.DoubleFromString(str);
			else if (valueClass == Float.class)
				value = LyTools.FloatFromString(str);
			else if (valueClass == Date.class)
			{
				if ((field instanceof LyColumnField)
						&& (((LyColumnField)field).getFormat() != null)
						&& spreadSheetDecoding)
					value = LyDateFormatter.parse(str, ((LyColumnField)field).getFormat());
				else
					value = LyTools.DateFromString(str);
			}
			else
				value = str;
		}

		return value;
	} // end method fromString

	/* Methods for encoding */

	/**
	 * Read a set of lines and return a set of objects.
	 * @param  objects the set of objects to encode.
	 * @param  fields  FieldInfos to save (if null save all).
	 * @param  type    indicates the file format.
	 * @return the set of lines describing the set of objects.
	 */
	public Vector<String> encodeValues(Vector<Object> objects, Vector<LyField> fields, int type)
	{
		// Get the fields : if fields is empty, get all the FieldInfos from
		// the ClassInfo.
		Vector<?> collect;

		if ((fields == null) || (fields.size() == 0))
			collect = _class.getFields();
		else
			collect = fields;

		Vector<String> result = new Vector<String>();
		String line;

		// Write the file header
		line = encodeHeader(collect, type);

		result.addElement(line);

		// For each object get the ValueSet and encode it.
		// FFT01068 test if objects is not null
		int    i, no = ((objects != null) ? objects.size() : 0);
		Object object;

		for (i = 0; i < no; i++)
		{
			object = objects.elementAt(i);
			if (object == null)
				continue;

			line = encodeValue(object, collect, type);
			result.addElement(line);
		}

		return result;
	} // end method encodeValues

	/**
	 * Return a line describing the FieldInfos.
	 * @param  fields the FieldInfos to encode.
	 * @param  type   indicates the file format.
	 * @return a String describing the FieldInfos.
	 */
	public String encodeHeader(Vector<?> fields, int type)
	{
		LyField		  field;
		int			  i, nf = fields.size();
		StringBuffer  result = new StringBuffer();
		String		  fieldName;
		LyEnvironment environment = getEnvironment();

		// Adding the export header mark if not null.
		if ((_headerMark != null)
	        && ((type == RAW_ENCODING)
                || (type == EXPORT_ENCODING)
                || (type == SPREADSHEET_ENCODING)))
			result.append(_headerMark);

		for (i = 0; i < nf; i++)
		{
			field = (LyField)fields.elementAt(i);

			if ((type == RAW_ENCODING)
			        || (type == EXPORT_ENCODING)
			        || !(field instanceof LyFieldInfo))
				fieldName = field.getId();
			else
			{
				fieldName = ((LyFieldInfo)field).getName();
				fieldName = environment.translate(fieldName);
			}

			result.append(asciiEncode(fieldName));

			if ((i != (nf - 1))
			        || (type == RAW_ENCODING)
			        || (type == EXPORT_ENCODING))
				result.append(_fieldSep);
		}

		return (result.toString());
	} // end method encodeHeader

	/**
	 * Return a line describing the ValueSet.
	 * @param  values the ValueSet to encode.
	 * @param  fields the FieldInfos to encode.
	 * @param  type   indicates the file format.
	 * @return the line describing the ValueSet.
	 */
	public String encodeValue(Object object, Vector<?> fields, int type)
	{
		LyField		 field;
		Object		 val;
		int			 i;
		int			 na = fields.size();
		StringBuffer result = new StringBuffer();
		String		 str = null;
		String		 toolsSep;

		// For each FieldInfo retrieve the value and write it
		for (i = 0; i < na; i++)
		{
			field = (LyField)fields.elementAt(i);
			if (i > 0)
				result.append(_fieldSep);

			try
			{
				val = _class.getValue(object, field);
			}
			catch (Exception e)
			{
				val = null;
			}

			if (val == null)
				continue;

			// For export : value of choice must be saved instead of index
			if ((field instanceof LyChoiceFieldInfo) && (type == EXPORT_ENCODING))
			{
				LyChoiceFieldInfo choice = (LyChoiceFieldInfo)field;

				if (!choice.isMultiple())
				{
					LyChoiceItem item = null;

					if (val instanceof Boolean)
						item = choice.getChoiceItem(choice.toString(val));
					else
						item = choice.getChoiceItem(((Short)val).shortValue());

					val = item.getValue();
				}
				else
				{
					BitSet		 bs = (BitSet)val;
					Vector<Object>		 vector = null;
					int			 j, size = bs.size();
					LyChoiceItem item;

					for (j = 0; j < size; j++)
					{
						if (bs.get(j))
						{
							if (vector == null)
								vector = new Vector<Object>();

							item = choice.getChoiceItem((short)j);
							vector.addElement(item.getValue());
						}
					}

					val = vector;
				}
			}

			str = toString(field, val, type);
			
			if ((field instanceof LyTextFieldInfo)
					&& (((LyTextFieldInfo)field).getType() == LyTextFieldInfo.Type.RICH)
					&& ((type == TEXT_ENCODING) || (type == SPREADSHEET_ENCODING)))
				str = LyTools.removeHtmlTags(str);
			
	         if ((type == SPREADSHEET_ENCODING)
                     && (str != null)
                     && str.contains(_fieldSep))
                 str = "\"" + str.replace("\"", "\"\"") + "\"";  

			if (str != null)
			{
				if ((type == RAW_ENCODING) || (type == EXPORT_ENCODING))
					str = asciiEncode(str);

				// Retrieve the separator for object type
				toolsSep = new Character(LyTools.getSeparator(val)).toString();

				// Replace the default separator used by the toString method
				// with _valueSep.
				str = LyString.replace(str, toolsSep, _valueSep);

				result.append(str);
			}
		}

		return (result.toString());
	} // end method encodeValue

	/**
	 * Gets the charset associated to the current data provider connection.
	 * 
	 * @return the charset if any, null otherwise
	 */
	protected String getCharset()
	{
        if (_charSet != null)
            return (_charSet == EMPTY_CHARSET ? null : _charSet);
        
		LyFileConnection connection = (LyFileConnection)_provider.getConnection();

		if (connection == null)
			return null;
		
        _charSet = connection.getConnectionCharset();
        if  (_charSet == null)
            _charSet = EMPTY_CHARSET;
        
        _provider.releaseConnection(connection);
		return _charSet;
	}

	/**
	 */
	protected String toString(LyField field, Object value, int type)
	{
		String result = "";

		if (value == null)
			return null;

		// FFT00016 SITA
		if ( ((type == SPREADSHEET_ENCODING) || (type == TEXT_ENCODING))
				&& ((field instanceof LyRelationFieldInfo)
						|| (field instanceof LyArrayFieldInfo)
						|| (field instanceof LyChoiceFieldInfo)))
		{
			if (field instanceof LyRelationFieldInfo)
				result = ((LyRelationFieldInfo)field).toName(value, _valueSep);

			if (field instanceof LyArrayFieldInfo)
				result = ((LyArrayFieldInfo)field).toName(value, _valueSep);
			
			if (field instanceof LyChoiceFieldInfo)
			{
				result = ((LyChoiceFieldInfo)field).toName(value, _valueSep);
				
				// Simple choice must be translated
				if (!((LyChoiceFieldInfo)field).isMultiple())
					result = getEnvironment().translate(result);
			}
		}
		// FFT00020 : adding spreadsheet export mode
		else if ( ((field instanceof LyFieldInfo) && (type == TEXT_ENCODING)) 
			|| ((field instanceof LyDateFieldInfo) && (type == SPREADSHEET_ENCODING)))
		{
			result = ((LyFieldInfo)field).toName(value);
			
			// Simple choice must be translated
			if ((field instanceof LyChoiceFieldInfo) && !((LyChoiceFieldInfo)field).isMultiple())
					result = getEnvironment().translate(result);
		}
		
		else if ((field instanceof LyColumnField)
				&& (((LyColumnField)field).getFormat() != null)
				&& (value instanceof Date))
			result = LyDateFormatter.format((Date)value, ((LyColumnField)field).getFormat());
		else
		{
			// Convert a boolean value. This implementation returns '0' for false and '1' for true.
			if ((type == RAW_ENCODING)
					&& (value instanceof Boolean))
			{
				Boolean	booleanValue = (Boolean)value;
				
				if (booleanValue.booleanValue())
					value = "1";
				else
					value = "0";
			}
			
			if (getEnvironment().testResource("LY_HIDE_UNKNOWN_OBJECTS") && (value instanceof LyObjectId[]))
			{
				LyObjectId[]		ids = (LyObjectId[])value;
				Vector<LyObjectId>	realIdsList = new Vector<LyObjectId>();
				int					i, n = ids.length;
				
				for (i = 0; i < n; i++)
				{
					if ((ids[i] != null) && (ids[i].getLogicalObject() != null))
						realIdsList.addElement(ids[i]);
				}

				value = (LyObjectId[])realIdsList.toArray(new LyObjectId[realIdsList.size()]);
			}
			
			result = LyRequestEncoder.toString(value);

			if ((field instanceof LyNumberFieldInfo)
					&& (type == SPREADSHEET_ENCODING))
			{
				DecimalFormatSymbols format = new DecimalFormatSymbols(getEnvironment()
							.getLocale());
				char				 decimalSeparator = format.getDecimalSeparator();

				result = result.replace('.', decimalSeparator);
			}
		}

		// Replace '\n' by spaces
		if (((type == TEXT_ENCODING) || (type == SPREADSHEET_ENCODING))
				&& (field instanceof LyTextFieldInfo))
		{
			result = result.replace('\n', ' ');
			result = result.replace('\r', ' ');
		}

		return result;
	} // end method toString

	/**
	 * Encode String in a quoted printable like manner Avoid conflicts in file.
	 */
	public String asciiEncode(String str)
	{
		int			 i, n = str.length();
		char		 c;
		StringBuffer sb = new StringBuffer((int)(n * 1.2));
		String		 hexStr;

		char		 fieldSep = _fieldSep.charAt(0);
		char		 valueSep = _valueSep.charAt(0);
		char		 headerMark = (_headerMark == null ? ' ' : _headerMark.charAt(0));
		boolean		 headerMarkNull = (_headerMark == null);

		for (i = 0; i < n; i++)
		{
			c = str.charAt(i);

			if ((c == fieldSep)
					|| (c == valueSep)
					|| (!headerMarkNull && (c == headerMark))
					|| (c == '=')
					|| (c == '#')
					|| (c == '\r')
					|| (c == '\n'))
			{
				hexStr = Integer.toHexString(c);

				if (hexStr.length() == 2)
					sb.append("=").append(hexStr);
				else if (hexStr.length() == 1)
					sb.append("=0").append(hexStr);
				else
					sb.append("=00");

			}
			else if ((c > 0xFFFD) || (c < 0x20)) // Invalid unicode or invalid XML char
			{
				// Save as unicode symbol
				hexStr = Integer.toHexString(c);
				if (hexStr.length() == 4)
					sb.append("=\\u").append(hexStr);
				else if (hexStr.length() == 3)
					sb.append("=\\u0").append(hexStr);
				else if (hexStr.length() == 2)
					sb.append("=\\u00").append(hexStr);
				else if (hexStr.length() == 1)
					sb.append("=\\u000").append(hexStr);
				else
					sb.append("=\\u0000");
			}
			else
				sb.append(c);
		}

		return sb.toString();
	} // end method asciiEncode

	/**
	 * Encode String in a qutoted printable like manner Avoid conflics in file.
	 */
	protected String asciiDecode(String str)
	{
        int          n = str.length();
        StringBuffer sb = null;
        int          idx, idx2;
        char         c;
        String       hexStr, val;
        Integer      asciiCode;
        int          skip;

        for (idx = 0, idx2 = 0; (idx < n) && ((idx2 = str.indexOf('=', idx)) >= 0);
                idx = idx2 + 1)
        {
            if (sb == null)
                sb = new StringBuffer(n);
            sb.append(str.substring(idx, idx2));

			// values \\uXXXX might have been encoded, try to recognise them
			c = str.charAt(idx2 + 1);
			if (c == '')
			{
				// native2ascii UTF-8 decode \u0004 into , 
				// for ascendent compatibility, re encode it.
				asciiCode = Integer.decode("0x0004");
				sb.append((char)asciiCode.intValue());
				idx2 += 1;
			}
			else
			{
				try
				{
					val = str.substring(idx2 + 1, idx2 + 3);
                    skip = 2; // Skip next 2 chars
                    
					if (val.equals("\\u"))
					{
						val = str.substring(idx2 + 3, idx2 + 7); // using unicode 4 letters in hexa \uA1FE
						skip += 4; // Skip 4 more chars
					}                     
                    
					hexStr = "0x" + val;
					asciiCode = Integer.decode(hexStr);
					sb.append((char)asciiCode.intValue());	
					idx2 += skip;
				}
				catch (NumberFormatException e)
				{
					sb.append('=');
					getEnvironment().writeException(e);
				}
			}
		}

        if (sb == null)
            return str;
        
        if (idx < n)
            sb.append(str.substring(idx, n));
        
		return sb.toString();
	} // end method asciiDecode

	/**
	 * Utility method to get current environment.
	 * @return Current environment.
	 */
	public LyEnvironment getEnvironment()
	{
		if (_environment != null)
			return _environment;

		return super.getEnvironment();
	}

	/**
	 * Utility method to set current environment.
	 * @param env Current environment.
	 */
	public void setEnvironment(LyEnvironment env)
	{
		_environment = env;
	}
	
	/**
	 * Validate changes on peer object.
	 */
	protected LyValueSet validateChanges(Object object)
	{
	    // Validate the modification.
	    return ((LyRowObject)object).validate(null);
	}
} // end class LyFileRequestEncoder
