/*
 * LyImportFileDataProvider.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.file;

import java.util.*;
import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.location.*;
import leon.peer.*;


/**
 * This class is used by DataSources to transmit their requests to an instance of RequestEncoder.
 * @version $Id: LyImportFileDataProvider.java 35251 2012-04-17 15:32:36Z jlsaouli $
 * @author  Lyria
 * @see     leon.peer.util.LyRowObject
 * @see     leon.peer.util.LyTableClass
 * @see     leon.peer.util.LyColumnField
 * @see     LyFileRequestEncoder
 * @see     LyFileConnection
 * @see     LyImportFileConnection
 */
public class LyImportFileDataProvider extends LyFileInfoDataProvider
{
	public static final String __VERSION =
		"$Id: LyImportFileDataProvider.java 35251 2012-04-17 15:32:36Z jlsaouli $";

	/** The path to the file containing data to import. */
	protected String		   _filePath;

	/**
	 * Decoding type. One of the following:
	 * <ul>
	 *   <li>{@link LyFileRequestEncoder#RAW_ENCODING}</li>
	 *   <li>{@link LyFileRequestEncoder#TEXT_ENCODING}</li>
	 *   <li>{@link LyFileRequestEncoder#SPREADSHEET_ENCODING}</li>
	 *   <li>{@link LyFileRequestEncoder#EXPORT_ENCODING}</li>
	 *   <li>{@link LyFileRequestEncoder#XML_EXPORT_ENCODING}</li>
	 * </ul>
	 */
	protected int			   _decodingType;

	/** The list of fields identifers contained in the file. */
	protected Vector<String>  _fileFieldIds;

	/**
	 * Logical class info associated with this file data provider. May be needed for spreadsheet
	 * import.
	 */
	protected LyClassInfo	   _classInfo;

	/** The index of the first line of the imported file containing real data. */
	protected int			   _startLine;
	
	/** List of lines containing errors. */
	protected BitSet		   _errorLines;


	/**
	 * Default constructor Initialization is done with the init method.
	 */
	public LyImportFileDataProvider(LyApplication application, LyLocation location)
	{
		super(application, location);
	}

	/**
	 * Constructor.
	 * @param      application
	 * @param      id
	 * @param      filePath
	 * @deprecated use LyImportFileDataProvider(LyApplication, LyLocation) instead.
	 */
	public LyImportFileDataProvider(LyApplication application, String id,
		String filePath)
	{
		this(application, new LyFileLocation(id, null, LyFileLocation.Type.INFO, null, null));
		_filePath = filePath;
		_decodingType = LyFileRequestEncoder.RAW_ENCODING;
		_fileFieldIds = null;
		_classInfo = null;
		_startLine = 0;
		_errorLines = null;
	}


	/**
	 */
	protected LyObjectSource createObjectSource(LyClassInfo classInfo)
	{
		return new LyTemporaryObjectSource(_application.getInfoProvider(), classInfo);
	}

	/**
	 */
	public void setFilePath(String filePath)
	{
		_filePath = filePath;
	}

	/**
	 * Sets the decoding type.
	 * @param decodingType one of the following:
	 *                     <ul>
	 *                       <li>{@link LyFileRequestEncoder#RAW_ENCODING}</li>
	 *                       <li>{@link LyFileRequestEncoder#TEXT_ENCODING}</li>
	 *                       <li>{@link LyFileRequestEncoder#SPREADSHEET_ENCODING}</li>
	 *                       <li>{@link LyFileRequestEncoder#EXPORT_ENCODING}</li>
	 *                       <li>{@link LyFileRequestEncoder#XML_EXPORT_ENCODING}</li>
	 *                     </ul>
	 */
	public void setDecodingType(int decodingType)
	{
		_decodingType = decodingType;
	}

	/**
	 * Sets the list of the fields contained in the file.
	 * @param fields the list of the fields contained in the file.
	 */
	public void setFileFieldIds(Vector<String> fieldIds)
	{
		_fileFieldIds = fieldIds;
	}

	/**
	 * Sets the logical class associated with this import data provider.
	 * @param fields the list of the fields contained in the file.
	 */
	public void setLogicalClass(LyClassInfo classInfo)
	{
		_classInfo = classInfo;
	}

	/**
	 * Sets the start line.
	 * @param startLine the index of the first line of the imported file containing real data.
	 */
	public void setStartLine(int startLine)
	{
		_startLine = startLine;
	}
	
	/**
	 * Adds a line containing an error.
	 * @param lineNumber the index of the line containing the error.
	 */
	public void addErrorLine(int lineNumber)
	{
		if (_errorLines == null)
			_errorLines = new BitSet();
		
		_errorLines.set(lineNumber);
	}
	
	/**
	 * Returns the list of lines containing errors.
	 * @return a bitset with the indexes of lines containing errors.
	 */
	public BitSet getErrorLines()
	{
		return _errorLines;
	}

	/**
	 */
	public LyConnection createConnection()
	{
		LyImportFileConnection connection = new LyImportFileConnection(_application, _filePath);

		connection.setDecodingType(_decodingType);
		connection.setFileFieldIds(_fileFieldIds);
		connection.setLogicalClass(_classInfo);
		connection.setStartLine(_startLine);

		return connection;
	}
} // end class LyImportFileDataProvider
