/*
 * LyInfoProvider.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.info;

import java.util.*;
import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.*;
import leon.info.location.*;
import leon.misc.*;
import leon.peer.*;


/**
 * This class is used by DataSources to transmit their requests to an instance of RequestEncoder.
 * The LyInfoProvider is used for info classes.
 * @version $Id: LyInfoProvider.java 28997 2010-12-23 10:17:15Z bmichel $
 * @author  Lyria
 * @see     LyInfoRequestEncoder
 */
public class LyInfoProvider extends LyDataProvider
{
	public static final String __VERSION =
		"$Id: LyInfoProvider.java 28997 2010-12-23 10:17:15Z bmichel $";


	/** List of LyClassInfos forwhich we can convert logical ids to physical ids. */
	protected Vector<LyClassInfo>		   _canConvertIdClassInfos;

	protected boolean					   _canBeFullyLoaded;
	/**
	 * Constructs a info data provider.
	 * @param application Current application instance
	 * @param location    This value is obtained from the location parameter in the meta model.
	 */
	public LyInfoProvider(LyApplication application, LyLocation location)
	{
		super(application, location);
		_canConvertIdClassInfos = new Vector<LyClassInfo>(1);
		_canBeFullyLoaded = true;
	}

	/**
	 * Constructs a info data provider.
	 * @param      application Current application instance
	 * @param      id          Identifier of this dataProvider. This value is obtained from the
	 *                         location parameter in the meta model.
	 * @deprecated should use LyInfoDataProvider(LyApplication application, LyLocation location)
	 *             instead of this constructor.
	 */
	public LyInfoProvider(LyApplication application, String id)
	{
		super(application, id);
		_canConvertIdClassInfos = new Vector<LyClassInfo>(1);
		_canBeFullyLoaded = true;
	}


	/**
	 */
	public LyRequestEncoder getRequestEncoder(LyDataSource dataSource)
	{
		return new LyInfoRequestEncoder(this, dataSource);
	}

	/**
	 * This class is a factory to create a LyConnection instance.
	 * @return a new Connection instance. The connection instance is used in the specific LyRequest
	 *         encoder implementation.
	 */
	protected LyConnection createConnection()
	{
		return null;
	}

	/**
	 * This method is used to build all the class peers for all the classInfos of the location
	 * managed by this provider. Validating a peer data provider consists to :
	 * <ul>
	 *   <li>start validation using startValidate method,</li>
	 *   <li>validate logical classes (LyClassInfo) binded on the location managed by this provider
	 *     by calling create,</li>
	 *   <li>end validation using startValidate method,</li>
	 * </ul>
	 * @param  classInfos Classes to validate.
	 * @return True if the validation was successful, false otherwise.
	 */
	public boolean validate(LyInfoList<LyClassInfo> classInfos)
	{
		// Creating the class and fields peers.
		int		    i, n = classInfos.getSize();
		LyClassInfo classInfo;

		for (i = 0; i < n; i++)
		{
			classInfo = classInfos.getInfo(i);
			if (classInfo.isVirtual())
				continue;

			if (classInfo.getBinding() instanceof LyClassInfoBinding)
				// The classInfo has the right provider id, validate it.
				validate(classInfo);

			// Check if logical object ids can be converted to peer objects ids.
			checkObjectIdConversion(classInfo);
		}

		return true;
	}

	/**
	 * Validates the given class info : this mainly consists to create a peer class for each
	 * encountered class binding. Note : peer classes must be shared when several classes have the
	 * same binding. The process is the following :
	 * <ul>
	 *   <li>create object source for the class info,</li>
	 *   <li>create or retrieve the associated class peer and data source corresponding to the class
	 *     binding,</li>
	 *   <li>register the object source as listener of the data source,</li>
	 *   <li>register the peer class on the class binding,</li>
	 *   <li>validate field infos of this class info.</li>
	 * </ul>
	 * @param  classInfo Class info that is validated.
	 * @return True if the validation was successful, false otherwise.
	 */
	protected void validate(LyClassInfo classInfo)
	{
		if (classInfo.isVirtual())
			return;

		LyClassBinding classBinding = classInfo.getBinding();
		String		   binding = classBinding.getBinding();
		LyClassInfo    peerClassInfo = (LyClassInfo)_application.getInfo(binding);

		if (peerClassInfo == null)
			throw new IllegalStateException("BF AppEngine : invalid class info binding " + binding);

		// First we create the objectSource and we associate the infoProvider
		LyObjectSource objectSource;
		LyDataSource   dataSource;

		objectSource = new LyObjectSource(_application.getInfoProvider(), classInfo);

		// Get the dataSource.
		dataSource = _application.getObjectSource(peerClassInfo);

		// The objectSource must listen to dataSource events.
		objectSource.listenTo(dataSource);

		// Associate the binding and the class peer.
		classBinding.setPeerClass(peerClassInfo);

		// Searching for all fieldInfos in the classInfo and create
		// the field peers if they do not exist.
		LyFieldInfo    fieldInfo;
		LyFieldBinding fieldBinding;
		int			   i, n = classInfo.getFieldInfoCount();

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			fieldBinding = fieldInfo.getBinding();
			if (fieldBinding == null)
				continue;

			if (fieldBinding instanceof LyFieldInfoBinding)
				validate(classInfo, fieldInfo);
			else
				throw new IllegalStateException("BF AppEngine : a F_BIND_INFO must be used for "
					+ fieldInfo + " (because it belongs to class info " + classInfo
					+ " that has a C_BIND_INFO)");
		}
	} // end method validate

	/**
	 * Validates the field info of the given class info : this mainly consists to create or retrieve
	 * the associated peer field.
	 * @param  classInfo Class info to which the field info belongs to.
	 * @param  fieldInfo Field info that is validated.
	 * @return True if the validation was successful, false otherwise.
	 */
	protected void validate(LyClassInfo classInfo, LyFieldInfo fieldInfo)
	{
		LyFieldBinding fieldBinding = fieldInfo.getBinding();
		String		   binding = fieldBinding.getBinding();
		LyClassBinding classBinding = classInfo.getBinding();
		LyClassInfo    peerClassInfo = (LyClassInfo)classBinding.getPeerClass();
		LyFieldInfo    peerFieldInfo;

		peerFieldInfo = peerClassInfo.getFieldInfo(binding);
		if (peerFieldInfo == null)
			throw new IllegalStateException("BF AppEngine : invalid field info binding "
				+ fieldInfo + " (no field info with id \"" + binding
				+ " found in class info \"" + peerClassInfo + ")");

		// Associate binding and field peer.
		classBinding.addPeerField(fieldInfo, peerFieldInfo);
	}

	/**
	 * Checks if conversion between logical object ids and physical object ids is enabled for the
	 * specified classInfo. This method stocks the classInfo in an internal list if the conversion
	 * is possible.
	 * @param classInfo The logical class to check.
	 */
	protected void checkObjectIdConversion(LyClassInfo classInfo)
	{
		if (classInfo.isLocal())
			return;

		LyClass peerClass = classInfo.getPeerClass();

		if (peerClass == null)
			throw new IllegalStateException("BF AppEngine : peer class is null for class " + classInfo);

		LyInfoList<LyFieldInfo>     idFieldInfos = classInfo.getIdFieldInfos();
		Vector<? extends LyField>	peerIdFields = peerClass.getIdFields();
		LyClassBinding classBinding = classInfo.getBinding();
		LyFieldInfo    fieldInfo;
		LyFieldBinding binding;
		int			   i, n;
		LyField		   field;
		boolean		   stop = false;

		@SuppressWarnings("unchecked")
		Vector<? extends LyField> temp = (Vector<? extends LyField>)peerIdFields.clone();
		peerIdFields = temp;
		
		@SuppressWarnings("unchecked")
		Vector<? extends LyField> originalPeerIdFields	= (Vector<? extends LyField>)peerIdFields.clone();
		n = idFieldInfos.getSize();
		for (i = 0; (i < n) && (!stop); i++)
		{
			fieldInfo = idFieldInfos.getInfo(i);
			if (fieldInfo.isLocal())
				continue;

			binding = fieldInfo.getBinding();
			if (binding instanceof LyFieldsBinding)
			{
				// FFT00025/AD6 : Warning sur binding de classe
				// Finds bindings on <F_BINDS> fields and treat them the same way as <F_BIND> ones
				LyField[] fields = classBinding.getPeerFields(fieldInfo);
				for (int j = 0; (j < fields.length) && (!stop); j++)
				{
					if (!originalPeerIdFields.contains(fields[j]))
						stop = true;
					else
						peerIdFields.remove(fields[j]);
				}
			}
			else
			{
				field = classBinding.getPeerField(fieldInfo);
				if (!originalPeerIdFields.contains(field))
					stop = true;
				else
					peerIdFields.remove(field);
			}
		}

		if ((peerIdFields.size() == 0) && (!stop))
			_canConvertIdClassInfos.addElement(classInfo);
		else
		{
			LyEnvironment environment = getEnvironment();

			environment.logWarning(
				"Can't convert logical objectIds to physical objectIds for classInfo " + classInfo);
		}
	} // end method checkObjectIdConversion

	/**
	 * To test if conversion between logical and physical object id is enable for the specified
	 * classInfo.
	 * @param  classInfo theclass info to check
	 * @return true if the conversion is possible, else false.
	 */
	protected boolean canConvertId(LyClassInfo classInfo)
	{
		return _canConvertIdClassInfos.contains(classInfo);
	}

	/**
	 * Closes all connections from the connection pool.
	 * @return true if the request completed successfully.
	 * @see    #createConnections(int)
	 */
	public boolean close()
	{
		_canConvertIdClassInfos = null;
		return super.close();
	}

	@Override
	public boolean canBeFullyLoaded(LyClass aClass)
	{
		LyClassInfo classInfo = (LyClassInfo)aClass;
		List<LyPeerDataProvider> dataProviders = classInfo.getPeerDataProviders();
		int i, n = (dataProviders == null ? 0 : dataProviders.size());
		LyPeerDataProvider peerDP;
		for (i = 0; i < n; i++)
		{
			peerDP = dataProviders.get(i);
			if (!peerDP.canBeFullyLoaded(classInfo.getPeerClass()))
				return false;
		}
		
		return true;
	}
	
	
} // end class LyInfoProvider
