/**
 * LyInfoRequestEncoder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.info;

import java.util.*;
import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.*;
import leon.info.location.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.util.*;


/**
 * This class is used to send requests to a data provider and to receive its answers. It is used by
 * LyObjectSource to send requests. All the objects in the parameter methods are LyObject. It gets
 * the peer of this objects and sends the request to the good LyDataSource. It waits for the
 * response and updates the response for the LyObjectSource.
 *
 * @version $Id: LyInfoRequestEncoder.java 34970 2012-03-22 15:14:47Z jlsaouli $
 * @author  Lyria
 * @see     LyObjectSource
 */
public class LyInfoRequestEncoder extends LyRequestEncoder
	implements LyResponseListener,
		LyGetDataSourceListenersInterface
{
	public static final String __VERSION =
		"$Id: LyInfoRequestEncoder.java 34970 2012-03-22 15:14:47Z jlsaouli $";

	/** Resource for activating controls on objects when loading. */
	public static boolean	   CONTROL_OBJECTS_ON_LOADING = false;
	
	/** Response used with the LyDataSource. */
	protected LyResponse	   _dataResponse;

	/** Physical class of the objects. */
	protected LyClass		   _peerClass;

	/** Data source for peer class. */
	protected LyDataSource     _peerDataSource;

	/**  */
	protected Vector<LyObject> _objects;

	/**  */
	protected LyFilter		   _filter;

	/**
	 * List of associated peers of the objects concerned by the request. The peers have the same
	 * index than the objects in _objects.
	 */
	protected Vector<Object>   _peerObjects;

	/** The transaction manager of the application. */
	LyTransactionManager	   _transactionManager;

	/** The transaction that owns this request. */
	LyTransaction			   _transaction;

	/** The fields to load in a LOAD request. */
	Vector<LyFieldInfo>		   _fieldsToLoad;

	/** Indicates that the current request is a full load (no filter and no limit). */
	boolean					   _fullLoad;

	/**
	 * Default constructor Initialization is done with the init method.
	 */
	public LyInfoRequestEncoder(LyDataProvider provider, LyDataSource dataSource)
	{
		super(provider, dataSource);

		LyApplication application = provider.getApplication();

		_peerClass = ((LyClassInfo)_class).getPeerClass();
		_peerDataSource = application.getDataSource(_peerClass);
		_transactionManager = application.getTransactionManager();
		_transaction = null;
	}

	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		super.init();
		_dataResponse = null;
		_peerClass = null;
		_peerDataSource = null;
		_peerObjects = null;
		_transactionManager = null;
		_fieldsToLoad = null;
	}

	/**
	 * To set the response properties according to the data response properties.
	 */
	protected void setResponseProperties(LyResponse response)
	{
		_response.setSorted(response.isSorted());
		// We can always filter the responses.
		_response.setFiltered(true);
		_response.setPaginated(response.isPaginated());
		_response.setPartialNotification(response.isPartialNotification());
		_response.setCounted(response.isCounted());
		_response.setPartial(response.isPartial());
		_response.setDistinctValues(response.hasDistinctValues());
	}

	/**
	 * To get the peer fields associated with the specified fields, filter and sort for a loading
	 * request.
	 *
	 * @param  fields the fields infos to get the peer fields. All extra fields will be added to
	 *                this list.
	 * @param  filter filter to get the peer fields.
	 * @param  sort   sort to get the peer fields.
	 * @return the peer fields associated with the specified fields, filter and sort and all peer
	 *         fields associated with fieldInfos which have mark id or name or load.
	 */
	protected Vector<LyField> getPeerFields(Vector<LyFieldInfo> fields, LyFilter filter, LySort sort)
	{
		if (fields == null)
			return null;
        
        LyApplication   application = _provider.getApplication();
        LyEnvironment   environment = application.getEnvironment();
        
        if (!environment.testResource("LY_PARTIAL_LOADING"))
            return null;
        
		LyDataProvider provider = _peerDataSource.getDataProvider();
		LyLocation     location = provider.getLocation();

		if ((location == null) || !location.isPersistent())
			return null;

		// We always load fieldinfos with mark id and load.
		LyInfoList<LyFieldInfo> fieldInfos;

		if (((LyClassInfo)_class).getParent() instanceof LyStructFieldInfo)
			// Get all fields for a struct class
			fieldInfos = ((LyClassInfo)_class).getFieldInfos();
		else
		{
			BitSet marks = new BitSet();

			marks.set(LyFieldInfo.MARK_ID);
			marks.set(LyFieldInfo.MARK_NAME);
			marks.set(LyFieldInfo.MARK_LOAD);
			fieldInfos = ((LyClassInfo)_class).getFieldInfos(marks);

			// Add requested fields
			LyFieldInfo fieldInfo;
			int		    i, n = (fields == null ? 0 : fields.size());

			for (i = 0; i < n; i++)
			{
				fieldInfo = fields.elementAt(i);
				fieldInfos.addInfo(fieldInfo);
			}

			// Add filter fields
			if (filter != null)
				fieldInfos.addList(filter.getFieldInfos(false));

			// Add all added fields...
            if (fields != null)
            {
    			for (i = 0, n = fieldInfos.getSize(); i < n; i++)
    			{
    				fieldInfo = fieldInfos.getInfo(i);
    				if (!fields.contains(fieldInfo))
    					fields.addElement(fieldInfo);
    			}
            }

			// Add fields of filter binding
			LyClassBinding classBinding = ((LyClassInfo)_class).getBinding();
			LyFilter	   bindingFilter = classBinding.getFilter();

			if (bindingFilter != null)
				fieldInfos.addList(bindingFilter.getFieldInfos());

			// Add sort fields
			if (sort != null)
				fieldInfos.addList(sort.getFieldInfos());
		}

		// Retrieve peer fields of for all these field infos
		LyClassBinding  classBinding = ((LyClassInfo)_class).getBinding();
		LyFieldInfo     fieldInfo;
		LyFieldBinding  binding;
		LyField[]	    peerFields;
		LyField		    peerField;
		int			    i, n = fieldInfos.getSize();
		int			    j, np;
		String		    id;
		Vector<LyField> result = new Vector<LyField>();

		for (i = 0; i < n; i++)
		{
			fieldInfo = fieldInfos.getInfo(i);
			binding = fieldInfo.getBinding();
			if (binding == null)
				continue;

			if (binding instanceof LyFieldsBinding)
			{
				peerFields = classBinding.getPeerFields(fieldInfo);
				if (peerFields == null)
					continue;

				for (j = 0, np = peerFields.length; j < np; j++)
				{
					id = peerFields[j].getId();
					if ((_peerClass.getField(id) != null) && !result.contains(peerFields[j]))
						result.addElement(peerFields[j]);
				}
			}
			else
			{
				peerField = classBinding.getPeerField(fieldInfo);
				if (peerField == null)
					continue;

				id = peerField.getId();
				if ((_peerClass.getField(id) != null) && !result.contains(peerField))
					result.addElement(peerField);
			}
		}

		// We always load peer id fields if they are missing from our list
		Vector<? extends LyField> idFields = _peerClass.getIdFields();

		n = idFields.size();
		for (i = 0; i < n; i++)
		{
			peerField = (LyField)idFields.elementAt(i);
			if (!result.contains(peerField))
				result.addElement(peerField);
		}

		return result;
	} // end method getPeerFields

	/**
	 * Get the peer object assosciated with the specified logical object ans set the new values to
	 * this peer if they exist. If the peer object does'nt exist, create it. Only this method is
	 * authorized to create peer instances.
	 *
	 * @param  object The logical object to get the peer instance.
	 * @param  create Indicates if we must create a new one or search an existent object.
	 * @return the associated peer of the specified logical object.
	 */
	protected Object getPeer(LyObject object, boolean create)
	{
		Object	    peer = null;
		LyClassInfo classInfo = (LyClassInfo)_class;
		LyClass     peerClass = classInfo.getPeerClass();

		if (create)
		{
			// Creating a new one.
			try
			{
				peer = peerClass.newInstance();
			}
			catch (InstantiationException e)
			{
				throw new RuntimeException(e);
			}
			catch (IllegalAccessException e)
			{
				throw new RuntimeException(e);
			}
		}
		else
			peer = object.getPeer();

		// Setting new values if exist
		LyValueSet[] values = new LyValueSet[2];

		values[0] = object.getPendingValues();
		values[1] = object.getNewValues();

		LyFieldInfo    fieldInfo;
		Object		   value, peerValue;
		LyValue		   newValue;
		LyFieldBinding binding;
		LyClassBinding classBinding = classInfo.getBinding();
		int			   i, j, n, m;
		LyField[]	   fields;
		Class<?>[]	   classes;
		Object[]	   objects;
		LyField		   field;

		for (int k = 0; k < values.length; k++)
		{
			n = ((values[k] == null) ? 0 : values[k].getSize());
			for (i = 0; i < n; i++)
			{
				newValue = values[k].getFieldValue(i);
				value = newValue.getValue();
				fieldInfo = (LyFieldInfo)newValue.getField();

				if (fieldInfo.isLocal() || classInfo.isLocal())
					continue;

				binding = fieldInfo.getBinding();
				if (binding instanceof LyFieldsBinding)
				{
					fields = classBinding.getPeerFields(fieldInfo);
					m = fields.length;
					classes = new Class[m];

					for (j = 0; j < m; j++)
						classes[j] = fields[j].getValueClass();

					objects = fieldInfo.getPeerValues(value, classes, object);

					for (j = 0; j < m; j++)
						peerClass.setValue(peer, fields[j],
							((objects == null) ? null : objects[j]));
				}
				else
				{
					field = classBinding.getPeerField(fieldInfo);
					peerValue = fieldInfo.getPeerValue(value, field.getValueClass(), object);

					peerClass.setValue(peer, field, peerValue);
				}
			}
		}

		// TODO Suppress dependencies with leon.peer.util package.
		if (create
				&& (peer instanceof LyRowObject))
		{
			// Unload all peer fields corresponding to a provider value fieldInfo.
			LyInfoList<LyFieldInfo> fieldsValuedByProvider = classInfo.getFieldInfos(
					LyFieldInfo.MARK_PROVIDER_VALUE);

			for (i = 0; i < fieldsValuedByProvider.getSize(); i++)
			{
				fieldInfo = fieldsValuedByProvider.getInfo(i);
				binding = fieldInfo.getBinding();

				if (binding instanceof LyFieldsBinding)
				{
					fields = classBinding.getPeerFields(fieldInfo);
					m = fields.length;
					for (j = 0; j < m; j++)
						((LyRowObject)peer).unload((LyColumnField)fields[j]);
				}
				else
				{
					field = classBinding.getPeerField(fieldInfo);
					((LyRowObject)peer).unload((LyColumnField)field);
				}
			}
		}

		return peer;
	} // end method getPeer

	/**
	 * To get the peer filter corresponding to the given filter.
	 *
	 * @param  filter Filter for which peer filter is requested.
	 * @return Peer filter corresponding to the given filter.
	 */
	protected LyFilter getPeerFilter(LyFilter filter)
	{
		if (!(_peerDataSource instanceof LyObjectSource))
			return filter;

		if (filter == null)
			return null;

		LyFilter		   peerFilter = null, subPeerFilter;
		LyFilterElement    element;
		LyFilterExpression filterExp;
		LyClassBinding     classBinding = ((LyClassInfo)_class).getBinding();
		LyFieldInfo		   fieldInfo;
		LyField			   peerField;
		int				   i, n = filter.getSize();
		String			   id;

		for (i = 0; i < n; i++)
		{
			element = filter.getElement(i);
			if (element instanceof LyFilter)
			{
				subPeerFilter = getPeerFilter((LyFilter)element);
				if (subPeerFilter == null)
					continue;

				if (peerFilter == null)
					peerFilter = new LyFilter(filter.getType());

				peerFilter.addElement(subPeerFilter);
			}
			else if (element instanceof LyFilterExpression)
			{
				filterExp = (LyFilterExpression)element;
				fieldInfo = filterExp.getFieldInfo();
				peerField = classBinding.getPeerField(fieldInfo);
				if (peerField == null)
					continue;

				id = peerField.getId();
				if (_peerClass.getField(id) == null)
					continue;

				if (peerFilter == null)
					peerFilter = new LyFilter(filter.getType());

				// FFT 596 : Be careful with object ID : when a logical class
				// is mapped on the logical class we must keep the
				// id (toString()) without class identifier (getFullString())
				Object val = filterExp.getValue();

				if (val instanceof LyObjectId)
					val = ((LyObjectId)val).toString();

				filterExp = new LyFilterExpression((LyFieldInfo)peerField, filterExp.isInverted(),
						filterExp.getCondition(), val);
				peerFilter.addElement(filterExp);
			}
			else if (element instanceof LyFilterRelation)
			{
				LyFilterRelation    filterRel = (LyFilterRelation)element;
				LyRelationFieldInfo relation = filterRel.getRelationFieldInfo();

				peerField = classBinding.getPeerField(relation);
				if (peerField == null)
					continue;

				LyFilterRelation peerFilterRel = new LyFilterRelation((LyRelationFieldInfo)
						peerField, filterRel.getFilter());

				for (int j = 0; j < filterRel.getAcceptedTargetCount(); j++)
					peerFilterRel.addAcceptedTargetClassInfo(filterRel.getAcceptedTargetClassInfos(
							j));

				if (peerFilter == null)
					peerFilter = new LyFilter(filter.getType());
				peerFilter.addElement(peerFilterRel);
			}
		}

		return peerFilter;
	} // end method getPeerFilter

	/**
	 * To get the peer sort corresponding to the given sort.
	 *
	 * @param  sort Sort for which peer sort is requested.
	 * @return Peer sort corresponding to the given sort.
	 */
	protected LySort getPeerSort(LySort sort)
	{
		if (!(_peerDataSource instanceof LyObjectSource))
			return sort;

		if (sort == null)
			return null;

		LySort		    peerSort = null;
		LySortCriterion criterion;
		LyClassBinding  classBinding = ((LyClassInfo)_class).getBinding();
		LyFieldInfo     fieldInfo;
		LyField		    peerField;
		int			    i, n = sort.getCriterionCount();
		String		    id;

		for (i = 0; i < n; i++)
		{
			criterion = sort.getCriterion(i);
			fieldInfo = criterion.getFieldInfo();
			peerField = classBinding.getPeerField(fieldInfo);
			if (peerField == null)
				continue;

			id = peerField.getId();
			if (_peerClass.getField(id) == null)
				continue;

			if (peerSort == null)
				peerSort = new LySort();

			peerSort.addCriterion((LyFieldInfo)peerField, criterion.isIncreasing());
		}

		return peerSort;
	} // end method getPeerSort

	/**
	 * To get the peer id corresponding to given logical id.
	 *
	 * @param  objectId Id for which peer id is requested.
	 * @return Id sort corresponding to the given logical id.
	 */
	protected LyFilter getFilter(LyObjectId objectId)
	{
		String[] 					ids = objectId.toStrings();
		int		 					nIds = ((ids == null) ? 0 : ids.length);
		Vector<? extends LyField>   idFields = _class.getIdFields();
		int		 					n = idFields.size();

		if (nIds != n)
			return null;

		LyFieldInfo		   fieldInfo;
		LyFilter		   filter = new LyFilter(LyFilter.Type.AND);
		LyFilterExpression expression;
		int				   i;

		for (i = 0; i < n; i++)
		{
			fieldInfo = (LyFieldInfo)idFields.elementAt(i);
			if (fieldInfo instanceof LyChoiceFieldInfo)
			{
				Short s;

				try
				{
					s = Short.decode(ids[i]);
				}
				catch (NumberFormatException e)
				{
					s = null;
				}
				expression = new LyFilterExpression(fieldInfo, s);
			}
			else
				expression = new LyFilterExpression(fieldInfo, ids[i]);
			expression.setModifier(LyFilterExpression.MOD_CASE_SENSITIVE);
			filter.addElement(expression);
		}

		return filter;
	} // end method getFilter

	/**
	 * To get the peer id corresponding to given logical id.
	 *
	 * @param  objectId Id for which peer id is requested.
	 * @return Id sort corresponding to the given logical id.
	 */
	protected LyObjectId getPeerObjectId(LyObjectId objectId)
	{
		// Make a translation only if peer class is not a LyClassInfo.
		if ((_peerDataSource instanceof LyObjectSource))
			return objectId;

		LyObjectId     				result;
		LyClassInfo    				classInfo = (LyClassInfo)_class;
		LyClassBinding 				classBinding = classInfo.getBinding();
		String[]	   				ids = objectId.toStrings();
		Vector<? extends LyField>	idFields = _class.getIdFields();
		LyFieldInfo    				fieldInfo;
		int			   				i, n = idFields.size();
		LyField[]	   				fields;
		Class<?>		   			idClass;
		Vector<Object> 				values = new Vector<Object>();
		Object		   				value, peerValue;
		LyFieldBinding 				binding;
		int			   				j, m, nId;
		Object[]	   				objects;
		LyField		   				field;
		Vector<String> 				peerIds = new Vector<String>();

		for (i = 0, nId = 0; i < n; i++)
		{
			fieldInfo = (LyFieldInfo)idFields.elementAt(i);
			if (fieldInfo.isLocal() || classInfo.isLocal())
				continue;

			binding = fieldInfo.getBinding();

			if (binding instanceof LyFieldsBinding)
			{
				fields = classBinding.getPeerFields(fieldInfo);
				m = fields.length;
				
				for (j = 0; j < m; j++)
				{
					if (!peerIds.contains(fields[j].getId()))
					{
						values.addElement(getValueFromIdString(null, ids[nId], fields[j].getValueClass()));
						peerIds.addElement(fields[j].getId());
					}

					nId++;
				}
			}
			else
			{
				field = classBinding.getPeerField(fieldInfo);
				
				if (!peerIds.contains(field.getId()))
				{
					idClass = fieldInfo.getValueClass();
					value = getValueFromIdString(fieldInfo, ids[nId], idClass);
					peerValue = fieldInfo.getPeerValue(value, field.getValueClass(), null);
					values.addElement(peerValue);
					peerIds.addElement(field.getId());
				}

				nId++;
			}
		}

		objects = values.toArray();
		result = new LyObjectId(_peerClass, objects);
		return result;
	} // end method getPeerObjectId

	protected Object getValueFromIdString(LyFieldInfo fieldInfo, String id, Class<?> idClass)
	{
		Object value = null;

		if (Number.class.isAssignableFrom(idClass))
		{
			// Check if id is a valid Number
			try
			{
				if (idClass == Integer.class)
					value = Integer.valueOf(id);
				else if (idClass == Short.class)
					value = Short.valueOf(id);
				else if (idClass == Long.class)
					value = Long.valueOf(id);
				else if (idClass == Double.class)
					value = Double.valueOf(id);
				else if (idClass == Float.class)
					value = Float.valueOf(id);
			}
			catch (NumberFormatException e)
			{
				value = null;
			}
		}
		else if (Date.class.isAssignableFrom(idClass))
			// Check if id is a valid Date
			value = LyDateFormatter.fromString(id);
		else if (fieldInfo != null)
			value = fieldInfo.parse(id);
		else
			value = id;

		return value;
	} // end method getValueFromIdString

	/**
	 * Load an object from the physical layer. This method is overriden to force the loading of the
	 * requested object.
	 *
	 * @param  objectId object identifier
	 * @param  fields   the fields to load (if null, load all)
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean loadObject(LyObjectId objectId, Vector<LyField> fields)
	{
		try
		{
			LyFilter filter = getFilter(objectId);

			if (filter != null)
				return getObjects(fields, null, filter, -1);
		}
		catch (Exception e)
		{
			// This could occur when the try to build a filter for F_BINDS failed
		}

		_dataResponse = _peerDataSource.loadObject(objectId, null, this, _classBinding);
		return true;
	}

	/**
	 * Get an object.
	 *
	 * @param  objectId object identifier
	 * @param  fields   the fields to load (if null, load all)
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getObject(LyObjectId objectId, Vector<LyField> fields)
	{  
		if (objectId == null)
			throw new IllegalArgumentException("BF AppEngine : objectId can't be null");

        LyApplication   application = _provider.getApplication();
        LyEnvironment   environment = application.getEnvironment();
        
		try
		{
			if (((LyInfoProvider)_provider).canConvertId((LyClassInfo)_class))
			{
				Object temp = (fields == null ? null : new Vector<LyField>(fields));
				@SuppressWarnings("unchecked")
                Vector<LyFieldInfo> fieldInfos = (Vector<LyFieldInfo>)temp;
                Vector<LyField>     peerFields = getPeerFields(fieldInfos, null, null);
                
				// We can convert ids so we try to retreive peer object
				// from its dataSource and so maybe in its cache.
				LyObjectId peerId = getPeerObjectId(objectId);
				Object     result = _peerDataSource.getObject(peerId, peerFields, _classBinding);

				// We must verify that result have all required fields.
				boolean reload = false;

				if (result != null)
				{
					if (peerFields == null)
						peerFields = new Vector<LyField>(_peerClass.getFields());

					LyField field;
					int     i, n = peerFields.size();

					for (i = 0; (i < n) && !reload; i++)
					{
						field = peerFields.elementAt(i);
						try
						{
							_peerClass.getValue(result, field);
						}
						catch (LyValueNotLoadedException e)
						{
							reload = true;
						}
					}
				}

				if (!reload)
				{
					_dataResponse = new LyResponse(LyResponse.Operation.LOAD, _peerClass);
					_dataResponse.setListener(this);
					_dataResponse.setFiltered(false);
					_dataResponse.setStatus(LyResponse.Status.RUNNING);
					_dataResponse.setStatus(LyResponse.Status.COMPLETED);
					_dataResponse.addResponse(result, null);
					_dataResponse.addResponse(null, null);
					return true;
				}
			}

			// We can't convert id or we must reload the object to have all required values.
			LyFilter filter = getFilter(objectId);

			if (filter != null)
				return getObjects(fields, null, filter, -1);
		}
		catch (Exception e)
		{
			if (environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
			{
				environment.logTrace("BF AppEngine : exception caught while getting the object "
					+ objectId + ". It will be reloaded");
				environment.writeException(e);
			}
			else
				environment.logWarning("BF AppEngine : " + e.getMessage());
		}

		// This could occurs when the try to build a filter for F_BINDS failed
		_transaction = _transactionManager.begin();
		_dataResponse = _transactionManager.loadObject(_peerDataSource, objectId, null, this,
				_classBinding);
		return true;
	} // end method getObject

	/**
	 * To test if the specified filter contains a local field or not.
	 *
	 * @param  filter the filter to test
	 * @return true if the filter does not contain any local field, false otherwise.
	 */
	protected boolean canFilter(LyFilter filter)
	{
		if (filter == null)
			return true;

		LyFilterElement element;
		int			    i, n = filter.getSize();
		boolean		    result = true;
		LyFieldInfo     fieldInfo;

		for (i = 0; result && (i < n); i++)
		{
			element = filter.getElement(i);
			if (element instanceof LyFilterExpression)
			{
				fieldInfo = ((LyFilterExpression)element).getFieldInfo();
				result = !fieldInfo.isLocal();
			}
			else if (element instanceof LyFilter)
				result = canFilter((LyFilter)element);
			else if (element instanceof LyFilterRelation)
			{
				fieldInfo = ((LyFilterRelation)element).getRelationFieldInfo();
				result = !fieldInfo.isLocal()
						&& (canFilter(((LyFilterRelation)element).getFilter()));
			}
			else if (element instanceof LyFilterCriterion)
				result = false;
		}
		return result;
	} // end method canFilter

	/**
	 * To get a new filter without any local field from the specified filter.
	 *
	 * @param  filter
	 * @return a filter without any local field
	 */
	protected LyFilter getNonLocalFilter(LyFilter filter)
	{
		if (filter == null)
			return null;

		LyFilterElement element;
		int			    i, n = filter.getSize();
		LyFilter	    result = null, subFilter;
		LyFieldInfo     fieldInfo;
		boolean		    or = (filter.getType() == LyFilter.Type.OR);

		for (i = 0; i < n; i++)
		{
			element = filter.getElement(i);
			if (element instanceof LyFilterExpression)
			{
				fieldInfo = ((LyFilterExpression)element).getFieldInfo();
				if (fieldInfo.isLocal() && or)
					return null;
				else if (fieldInfo.isLocal())
					continue;

				if (result == null)
					result = new LyFilter(filter.getType());
				result.addElement(element);
			}
			else if (element instanceof LyFilter)
			{
				subFilter = getNonLocalFilter((LyFilter)element);
				if ((subFilter == null) && or)
					return null;
				else if (subFilter == null)
					continue;

				if (result == null)
					result = new LyFilter(filter.getType());
				result.addElement(subFilter);
			}
			else if (element instanceof LyFilterRelation)
			{
				fieldInfo = ((LyFilterRelation)element).getRelationFieldInfo();
				if (fieldInfo.isLocal() && or)
					return null;
				else if (fieldInfo.isLocal())
					continue;

				subFilter = getNonLocalFilter(((LyFilterRelation)element).getFilter());
				if ((subFilter == null) && or)
					return null;

				if (result == null)
					result = new LyFilter(filter.getType());
				result.addElement(new LyFilterRelation((LyRelationFieldInfo)fieldInfo, subFilter));
			}
		}
		return result;
	} // end method getNonLocalFilter

	/**
	 * Get a set of objects.
	 *
	 * @param  fields the fields to load (if null, load all)
	 * @param  sort   indicates sorting criterion, or null if default sort is used.
	 * @param  filter indicates filtering criterion, or null if there is no filter.
	 * @param  n      maximum number of objects to get.
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getObjects(Vector<LyField> fields, LySort sort, LyFilter filter, int n)
	{
		if(filter != null)
			filter = filter.getPeerFilter();
		
        LyApplication       application = _provider.getApplication();
        LyEnvironment       environment = application.getEnvironment();
		LySort		        peerSort = getPeerSort(sort);
		LyFilter	        peerFilter = getPeerFilter(filter);
		
		Object temp = (fields == null ? null : new Vector<LyField>(fields));
		@SuppressWarnings("unchecked")
        Vector<LyFieldInfo> fieldInfos = (Vector<LyFieldInfo>)temp;
		Vector<LyField>     peerFields = getPeerFields(fieldInfos, filter, sort);

		_fullLoad = ((filter == null) // No filter
		        && (n < 0)            // No pagination
		        && (environment.getIntResource("LY_DATA_UPDATE_DELAY") < 0)); // No update
		_fieldsToLoad = fieldInfos;

		_filter = filter;
		_transaction = _transactionManager.begin();
		_dataResponse = _transactionManager.loadObjects(_peerDataSource, peerFields, peerSort, peerFilter, n, this, _classBinding);

		return true;
	}

	/**
	 * Return the number of objects from the data provider .
	 *
	 * @param  filter indicates filtering criterion, or null if there is no filter.
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getSize(LyFilter filter)
	{
		if(filter != null)
			filter = filter.getPeerFilter();
		
		LyFilter peerFilter = getPeerFilter(filter);

		_transaction = _transactionManager.begin();
		int n = _transactionManager.getSize(_peerDataSource, peerFilter, _classBinding);

		_transactionManager.end(_transaction);

		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, new Integer(n), null);

		return true;
	}

	/* Methods for deletion */

	/**
	 * Utility method that accesses to all realtion values to force their loading to be be sure that
	 * all relations controlled by daemons (or by specific code) are loaded to avoid missing sub
	 * objects or join objects when deleting.
	 */
	protected void completeObject(LyObject object)
	{
		LyClassInfo   classInfo = object.getClassInfo();
		LyApplication application = classInfo.getApplication();
		int			  i, n = classInfo.getFieldInfoCount();
		LyFieldInfo   fieldInfo;

		for (i = 0; i < n; i++)
		{
			fieldInfo = classInfo.getFieldInfo(i);
			if (((fieldInfo instanceof LyRelationFieldInfo)
						|| (fieldInfo instanceof LyArrayFieldInfo))
					&& (application.getDaemon(fieldInfo) != null))
				object.getValue(fieldInfo);
		}
	}

	/**
	 * To delete an object.
	 *
	 * @param  object the object to delete
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean deleteObject(Object object)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		// Retrieve assosciated peer.
		LyObject lyObject = (LyObject)object;
		Object   peer = getPeer(lyObject, false);

		_objects = new Vector<LyObject>();
		_objects.addElement(lyObject);
		_peerObjects = new Vector<Object>();
		_peerObjects.addElement(peer);

		// Complete object before accessing to sub objects or join objects
		completeObject(lyObject);

		// Get sub objects and join objects
		LyObjectList localSubObjects, subObjects, joinObjects, arrayObjects;

		arrayObjects = lyObject.getArrayObjects();
		localSubObjects = lyObject.getSubObjects(true);
		subObjects = lyObject.getSubObjects(false);
		joinObjects = lyObject.getJoinObjects();

		// Delete local join objects, sub objects, peer and finally remaining sub objects
		LyApplication application = _provider.getApplication();

		_transaction = _transactionManager.begin();

		// Delete join objects.
		if (joinObjects != null)
			joinObjects.delete();
		if (_transaction.getStatus() == LyTransaction.Status.ERROR)
		{
			// There is an error, we can stop now by rollbacking the transaction.
			_response.setStatus(LyResponse.Status.ERROR);
			_response.setError(application.getError());

			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			_transactionManager.end(_transaction, false);
			_transaction = null;
			return false;
		}

		// Delete array objects
		if (arrayObjects != null)
			arrayObjects.delete();
		if (_transaction.getStatus() == LyTransaction.Status.ERROR)
		{
			// There is an error, we can stop now by rollbacking the transaction.
			_response.setStatus(LyResponse.Status.ERROR);
			_response.setError(application.getError());

			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			_transactionManager.end(_transaction, false);
			_transaction = null;
			return false;
		}

		// Delete sub objects
		if (localSubObjects != null)
			localSubObjects.delete();
		if (_transaction.getStatus() == LyTransaction.Status.ERROR)
		{
			// There is an error, we can stop now by rollbacking the transaction.
			_response.setStatus(LyResponse.Status.ERROR);
			_response.setError(application.getError());

			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			_transactionManager.end(_transaction, false);
			_transaction = null;
			return false;
		}

		// Delete the object
		_dataResponse = _transactionManager.deleteObject(_peerDataSource, peer,
				new _LyDeleteResponseListener(), _classBinding);
		if (_transaction.getStatus() == LyTransaction.Status.ERROR)
		{
			// There is an error, we can stop now by rollbacking the transaction.
			_transactionManager.end(_transaction, false);
			_transaction = null;
			return false;
		}

		// Delete remaining sub objects
		if (subObjects != null)
			subObjects.delete();
		if (_transaction.getStatus() == LyTransaction.Status.ERROR)
		{
			// There is an error, we can stop now by rollbacking the transaction.
			_response.setStatus(LyResponse.Status.ERROR);
			_response.setError(application.getError());

			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			_transactionManager.end(_transaction, false);
			_transaction = null;
			return false;
		}

		// Now we can complete the response...
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);

		// ... and commit the transaction.
		_transactionManager.end(_transaction, true);

		return true;
	} // end method deleteObject

	/**
	 * To delete an object.
	 *
	 * @param  object the object to delete
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean deleteObjects(Vector<Object> objects)
	{
		if (objects == null)
			throw new IllegalArgumentException("BF AppEngine : objects can't be null");

		Object		   peer;
		LyObject	   object;
		Vector<Object> peers;
		int			   i, n;
		LyObjectList   subObjects = null, joinObjects = null, localSubObjects = null,
					   arrayObjects = null, objectList;

		_objects = new Vector<LyObject>();
		_peerObjects = new Vector<Object>();

		peers = new Vector<Object>();
		n = objects.size();

		for (i = 0; i < n; i++)
		{
			object = (LyObject)objects.elementAt(i);
			_objects.addElement(object);

			// Complete object before accessing to sub objects or join objects
			completeObject(object);

			// Get array objects
			objectList = object.getArrayObjects();
			if ((arrayObjects == null) && (objectList != null))
				arrayObjects = new LySimpleObjectList(objectList);
			else if ((arrayObjects != null) && (objectList != null))
				arrayObjects.addList(objectList);

			// Get local sub objects
			objectList = object.getSubObjects(true);
			if ((localSubObjects == null) && (objectList != null))
				localSubObjects = new LySimpleObjectList(objectList);
			else if ((localSubObjects != null) && (objectList != null))
				localSubObjects.addList(objectList);

			// Get non local sub objects
			objectList = object.getSubObjects(false);
			if ((subObjects == null) && (objectList != null))
				subObjects = new LySimpleObjectList(objectList);
			else if ((subObjects != null) && (objectList != null))
				subObjects.addList(objectList);

			// Get join objects
			objectList = object.getJoinObjects();
			if ((joinObjects == null) && (objectList != null))
				joinObjects = new LySimpleObjectList(objectList);
			else if ((joinObjects != null) && (objectList != null))
				joinObjects.addList(objectList);

			// Retrieving associated peer.
			peer = getPeer(object, false);
			_peerObjects.addElement(peer);

			peers.addElement(peer);
		}

		// Delete objects that depend on the object in a transaction.
		// We begin by join objects to avoid integrity constraints exceptions in the data provider.
		LyApplication application = _provider.getApplication();

		_transaction = _transactionManager.begin();

		// Delete join objects
		if (joinObjects != null)
			joinObjects.delete();
		if (_transaction.getStatus() == LyTransaction.Status.ERROR)
		{
			// There is an error, we can stop now by rollbacking the transaction.
			_response.setStatus(LyResponse.Status.ERROR);
			_response.setError(application.getError());

			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			_transactionManager.end(_transaction, false);
			_transaction = null;
			return false;
		}

		// Delete array objects
		if (arrayObjects != null)
			arrayObjects.delete();
		if (_transaction.getStatus() == LyTransaction.Status.ERROR)
		{
			// There is an error, we can stop now by rollbacking the transaction.
			_response.setStatus(LyResponse.Status.ERROR);
			_response.setError(application.getError());

			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			_transactionManager.end(_transaction, false);
			_transaction = null;
			return false;
		}

		// Delete sub objects
		if (localSubObjects != null)
			localSubObjects.delete();
		if (_transaction.getStatus() == LyTransaction.Status.ERROR)
		{
			// There is an error, we can stop now by rollbacking the transaction.
			_response.setStatus(LyResponse.Status.ERROR);
			_response.setError(application.getError());

			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			_transactionManager.end(_transaction, false);
			_transaction = null;
			return false;
		}

		// Delete the object
		if (peers.size() > 0)
			_dataResponse = _transactionManager.deleteObjects(_peerDataSource, peers,
					new _LyDeleteResponseListener(), _classBinding);
		if (_transaction.getStatus() == LyTransaction.Status.ERROR)
		{
			// There is an error, we can stop now by rollbacking the transaction.
			_transactionManager.end(_transaction, false);
			_transaction = null;
			return false;
		}

		// Delete remaining sub objects
		if (subObjects != null)
			subObjects.delete();
		if (_transaction.getStatus() == LyTransaction.Status.ERROR)
		{
			// There is an error, we can stop now by rollbacking the transaction.
			_response.setStatus(LyResponse.Status.ERROR);
			_response.setError(application.getError());

			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			_transactionManager.end(_transaction, false);
			_transaction = null;
			return false;
		}

		// Now we can complete the response...
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);

		// ... and commit the transaction
		_transactionManager.end(_transaction, true);

		return true;
	} // end method deleteObjects


	/* Methods for creation */

	/**
	 * To create a new object.
	 *
	 * @param  object the new object to create
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean createObject(Object object)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		// We must verify that all fieldInfos that have the mark NOT_NULL
		// have a value.
		addNotNullValues((LyObject)object, false);

		// We must verify that all fieldInfos that have the mark PROVIDER_VALUE
		// have'nt any value.
		removeProviderValues((LyObject)object, false);

		// Create a new instance of peer object and setting the new values.
		Object peer = getPeer((LyObject)object, true);

		if (peer == null)
			throw new IllegalArgumentException("BF AppEngine : object peer can't be null");

		_objects = new Vector<LyObject>();
		_objects.addElement((LyObject)object);

		_peerObjects = new Vector<Object>();
		_peerObjects.addElement(peer);

		_transaction = _transactionManager.begin();
		_dataResponse = _transactionManager.createObject(_peerDataSource, peer, this,
				_classBinding);

		return true;
	} // end method createObject

	/* Methods for modification */
	/**
	 * To set an object.
	 *
	 * @param  object the object to set
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean setObject(Object object)
	{
		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : object can't be null");

		LyValueSet[] values = new LyValueSet[2];

		values[0] = ((LyObject)object).getPendingValues();
		values[1] = ((LyObject)object).getNewValues();

		int[]	    i, n;
		int		    k = 0;
		LyValue     newValue;
		LyFieldInfo fieldInfo;

		i = new int[2];
		n = new int[2];
		for (k = 0; k < values.length; k++)
		{
			n[k] = ((values[k] == null) ? 0 : values[k].getSize());
			for (i[k] = 0; i[k] < n[k]; i[k]++)
			{
				newValue = values[k].getFieldValue(i[k]);
				fieldInfo = (LyFieldInfo)newValue.getField();
				if (!fieldInfo.isLocal())
					break;
			}
			if (i[k] != n[k])
				break;
		}

		if ((k == values.length) && ((i[0] > 0) || (i[1] > 0)))
		{
			// Object locally modified
			// We must even so begin a new transaction because some
			// local modifications could trigger some daemons.
			_transaction = _transactionManager.begin();

			// Get old local values
			LyValueSet oldValues = new LyValueSet();
			Object     oldValue;

			for (k = 0; k < values.length; k++)
			{
				n[k] = ((values[k] == null) ? 0 : values[k].getSize());
				for (i[k] = 0; i[k] < n[k]; i[k]++)
				{
					newValue = values[k].getFieldValue(i[k]);
					fieldInfo = (LyFieldInfo)newValue.getField();
					oldValue = ((LyObject)object).getLocalValue(fieldInfo, false);
					oldValues.addValue(fieldInfo, oldValue);
				}
			}
			_response.setStatus(LyResponse.Status.RUNNING);
			_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object, oldValues);

			// End the transaction.
			_transactionManager.end(_transaction);
			_transaction = null;

			return true;
		}

		// Retrieving associated peer and setting the new values
		Object peer = getPeer((LyObject)object, false);

		_objects = new Vector<LyObject>();
		_objects.addElement((LyObject)object);

		_peerObjects = new Vector<Object>();
		_peerObjects.addElement(peer);

		_transaction = _transactionManager.begin();
		_dataResponse = _transactionManager.setObject(_peerDataSource, peer, this, _classBinding);

		return true;
	} // end method setObject

	/**
	 * To set a set of objects.
	 *
	 * @param  objects the objects to set
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean setObjects(Vector<Object> objects)
	{
		if (objects == null)
			throw new IllegalArgumentException("BF AppEngine : objects can't be null");

		Object		   peer;
		LyObject	   object;
		Vector<Object> peers;
		int			   i, n = objects.size();

		_objects = new Vector<LyObject>();
		_peerObjects = new Vector<Object>();
		peers = new Vector<Object>();

		for (i = 0; i < n; i++)
		{
			object = (LyObject)objects.elementAt(i);
			_objects.addElement(object);

			// Retrieving associated peer and setting new values.
			peer = getPeer(object, false);
			_peerObjects.addElement(peer);

			peers.addElement(peer);
		}

		_transaction = _transactionManager.begin();
		_dataResponse = _transactionManager.setObjects(_peerDataSource, peers, this, _classBinding);
		if (objects.size() == 0)
		{
			_dataResponse.setStatus(LyResponse.Status.COMPLETED);
			_dataResponse.addResponse(null, null);
		}

		return true;
	} // end method setObjects

	/**
	 * To test if an object matches the filters associated with the request.
	 */
	protected boolean matchFilters(LyObject object)
	{
		if (object == null)
			return false;

		if (_classBinding != null)
		{
			LyFilter filter;

			filter = _classBinding.getFilter();

			if ((filter != null) && (!filter.match(object)))
				return false;
		}

		if ((_filter != null) && (!_filter.match(object)))
			return false;

		return true;
	}

	/* Implementation of LyResponseListener */

	/**
	 * Method for response changes.
	 *
	 * @param response the response
	 */
	public void newResponse(LyResponse response, Object object, LyValueSet values)
	{
		if (response == null)
			throw new IllegalArgumentException("BF AppEngine : response can't be null");

        LyResponse.Status status = response.getStatus();

		if (status == LyResponse.Status.ERROR)
		{
			_response.setError(response.getError());
			_dataSource.receiveResponse(_response, status, null, null);
		}
		else
		{
			LyObject lyObject = null;
            LyResponse.Operation operation = response.getOperation();

			if (operation == LyResponse.Operation.LOAD)
			{
				setResponseProperties(response);

				LyCache cache = _dataSource.getCache();

				if (cache != null)
					lyObject = cache.getObjectFromPeer(object);
				if (lyObject == null)
				{
					lyObject = ((LyClassInfo)_class).createInstance();
					lyObject.setPeer(object);

					LyClassInfo classInfo = ((LyClassInfo)_class).getObjectSource()
							.getRealClassInfo(lyObject);

					if ((classInfo != null) && (((LyClassInfo)_class) != classInfo))
					{

						if (classInfo.getBinding() instanceof LyUnionBinding)
						{
							LyObjectSource objectSource = classInfo.getApplication()
									.getObjectSource(classInfo);

							lyObject = (LyObject)objectSource.getObject(lyObject.getObjectId());
							if (lyObject == null)
								return;
						}
						else
						{
							lyObject = classInfo.createInstance();
							lyObject.setPeer(object);
						}
					}
				}

				// We must verify if the object match the filters.
				if (response.isFiltered() || matchFilters(lyObject))
				{
					LyDataSource ds = lyObject.getClassInfo().getDataSource();

					if (ds instanceof LyExtendObjectSource)
						ds = ((LyExtendObjectSource)ds).getObjectSource();
					if (ds == null)
						ds = _dataSource;
					ds.receiveResponse(_response, status, lyObject, values);
				}
			}
			else if (operation == LyResponse.Operation.COUNT)
				_dataSource.receiveResponse(_response, status, object, null);
			else
			{
				int index;

				// Retrieving the index of the peer.
				index = _peerObjects.indexOf(object);
				if (index < 0)
				    index = _peerObjects.size() - 1;
				_peerObjects.removeElementAt(index);

				// Retrieving the associated object.
				lyObject = _objects.elementAt(index);
				_objects.removeElementAt(index);

				if (operation == LyResponse.Operation.CREATE)
					lyObject.setPeer(object);
				else if (operation == LyResponse.Operation.SET)
				{
					// Get old local values if any.
					LyValueSet  newValues = lyObject.getNewValues();
					LyFieldInfo fieldInfo;
					Object	    oldValue;
					LyValue     newValue;

					for (int i = 0; i < newValues.getSize(); i++)
					{
						newValue = newValues.getFieldValue(i);
						fieldInfo = (LyFieldInfo)newValue.getField();
						if (fieldInfo.isLocal())
						{
							if (values == null)
								values = new LyValueSet();
							oldValue = lyObject.getLocalValue(fieldInfo);
							values.addValue(fieldInfo, oldValue);
						}
					}
				}

				// Send the response.
				_dataSource.receiveResponse(_response, status, lyObject, values);
			}
		}

		if ((status == LyResponse.Status.COMPLETED) && (_transaction != null))
		{
			_transactionManager.end(_transaction);
			_transaction = null;
		}
	} // end method newResponse

	/**
	 * Method for end of response.
	 *
	 * @param response the response
	 */
	public void endResponse(LyResponse response)
	{
        LyResponse.Status status = response.getStatus();

		if (status == LyResponse.Status.ERROR)
			_response.setError(response.getError());
		else if ((status == LyResponse.Status.COMPLETED) && _fullLoad)
		{
			_response.setFullLoad(true);
			if (response.isPartial())
				((LyObjectSource)_dataSource).addLoadedFieldInfos(_fieldsToLoad);
			else
				((LyObjectSource)_dataSource).addLoadedFieldInfos(null);
		}

		setResponseProperties(response);

		_dataSource.receiveResponse(_response, status, null, null);

		if (_transaction != null)
		{
			_transactionManager.end(_transaction);
			_transaction = null;
		}

		if (CONTROL_OBJECTS_ON_LOADING
			&& (response.getOperation() == LyResponse.Operation.LOAD)
			&& !((LyClassInfo)_class).isTemporary())
		{
			// We must control the loaded objects.
			int		   i, n = _response.getSize();
			LyObject   object;
			LyValueSet oldValues;

			for (i = 0; i < n; i++)
			{
				object = (LyObject)_response.getResponse(i);
				oldValues = _response.getValues(i);
				if (oldValues != null)
				{
					// We don't know what are the changes so we contro;all the fields.
					if (oldValues.getSize() == 0)
						oldValues = null;
					object.control(oldValues, true);
				}
			}
		}
	} // end method endResponse

	/**
	 * Indicates which listeners have been notified of the Response.
	 *
	 * @see LyGetDataSourceListenersInterface
	 */
	public Vector<LyDataSourceListener> getDataSourceListeners()
	{
		if (_dataSource == null)
			return null;

		Vector<LyDataSourceListener> v = new Vector<LyDataSourceListener>(1);

		v.addElement(((LyObjectSource)_dataSource).getListener());
		return v;
	}

	class _LyLoadResponseListener
		implements LyResponseListener,
			LyGetDataSourceListenersInterface
	{
		/**
		 * Method for response changes.
		 *
		 * @param response the response
		 */
		public void newResponse(LyResponse response, Object object, LyValueSet values)
		{
            LyResponse.Status status = response.getStatus();

			setResponseProperties(response);
			LyObject lyObject = (LyObject)object;

			// We must verify if the object match the filters.
			// TODO check if filtering can be avoided if response is completely filtered
			if (matchFilters(lyObject))
				_dataSource.receiveResponse(_response, status, lyObject, values);

			if ((status == LyResponse.Status.COMPLETED) && (_transaction != null))
			{
				_transactionManager.end(_transaction);
				_transaction = null;
			}
		}

		/**
		 * Method for end of response.
		 *
		 * @param response the response
		 */
		public void endResponse(LyResponse response)
		{
            LyResponse.Status status = response.getStatus();

			if (status == LyResponse.Status.ERROR)
				_response.setError(response.getError());
			else if ((status == LyResponse.Status.COMPLETED) && _fullLoad)
			{
				_response.setFullLoad(true);
				if (response.isPartial())
					((LyObjectSource)_dataSource).addLoadedFieldInfos(_fieldsToLoad);
				else
					((LyObjectSource)_dataSource).addLoadedFieldInfos(null);
			}

			setResponseProperties(response);

			_dataSource.receiveResponse(_response, status, null, null);

			if (_transaction != null)
			{
				_transactionManager.end(_transaction);
				_transaction = null;
			}
		}

		/**
		 * Indicates which listeners have been notified of the Response.
		 *
		 * @see LyGetDataSourceListenersInterface
		 */
		public Vector<LyDataSourceListener> getDataSourceListeners()
		{
			if (_dataSource == null)
				return null;

			Vector<LyDataSourceListener> v = new Vector<LyDataSourceListener>(1);

			v.addElement(((LyObjectSource)_dataSource).getListener());
			return v;
		}
	} // end class _LyLoadResponseListener


	class _LyDeleteResponseListener
		implements LyResponseListener,
			LyGetDataSourceListenersInterface
	{

		/**
		 * Method for response changes.
		 *
		 * @param response the response
		 */
		public void newResponse(LyResponse response, Object object, LyValueSet values)
		{
			if (response == null)
				throw new IllegalArgumentException("BF AppEngine : response can't be null");

            LyResponse.Status status = response.getStatus();

			if (status == LyResponse.Status.ERROR)
			{
				_response.setError(response.getError());
				_dataSource.receiveResponse(_response, status, null, null);
			}
			else
			{
				LyObject lyObject;
				int		 index;

				// Retrieving the index of the peer.
				index = _peerObjects.indexOf(object);
				_peerObjects.removeElementAt(index);

				// Retrieving the associated object.
				lyObject = _objects.elementAt(index);
				_objects.removeElementAt(index);

				// Send the response.
				_dataSource.receiveResponse(_response, status, lyObject, values);
			}
		} // end method newResponse

		/**
		 * Method for end of response.
		 *
		 * @param response the response
		 */
		public void endResponse(LyResponse response)
		{
            LyResponse.Status status = response.getStatus();

			if (status == LyResponse.Status.ERROR)
			{
				_response.setError(response.getError());
				_dataSource.receiveResponse(_response, status, null, null);
			}
		}

		/**
		 * Indicates which listeners have been notified of the Response.
		 *
		 * @see LyGetDataSourceListenersInterface
		 */
		public Vector<LyDataSourceListener> getDataSourceListeners()
		{
			if (_dataSource == null)
				return null;

			Vector<LyDataSourceListener> v = new Vector<LyDataSourceListener>(1);

			v.addElement(((LyObjectSource)_dataSource).getListener());
			return v;
		}
	} // end class _LyDeleteResponseListener
} // end class LyInfoRequestEncoder
