/*
 * LyJavaClass.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.java;

import java.lang.reflect.*;

import leon.data.*;
import leon.misc.*;
import leon.peer.*;


/**
 * This class is the implementation of the interface LyClass for Java data classes.
 * @version $Id: LyJavaClass.java 28997 2010-12-23 10:17:15Z bmichel $
 * @author  Lyria
 * @see     leon.data.LyClass
 * @see     LyAttributeField
 */
public class LyJavaClass extends LyPeerClass
{
	public static final String __VERSION =
		"$Id: LyJavaClass.java 28997 2010-12-23 10:17:15Z bmichel $";

    /** Constructor to be used for building new instances of this class. */
    
	protected Constructor<?>    _constructor;
    
	/**
	 * empty constructor.
	 */
	public LyJavaClass()
	{
	}

	/**
	 * Default constructor.
	 */
	public LyJavaClass(String id, String javaClassName)
	{
		setId(id);
		setName(javaClassName);
    }
    
    /**
     * Sets name of the class.
     *
     * @param     name Name of the class.
     * @exception IllegalStateException If the class has already a name.
     */
    public void setName(String name)
    {
        super.setName(name);
        
		try
		{
			_objectClass = Class.forName(name);
            
            try
            {
                // Check if we have a public constructor
                _objectClass.getConstructor((Class[])null);
            }
            catch (Exception e1)
            {
                initProtectedConstructor();
            }
		}
		catch (ClassNotFoundException e)
		{
			LyEnvironment env = getEnvironment();
			
			if (env != null)
				env.writeException(e);
			else
				e.printStackTrace();
		}
	}

    /**
     * Inits constructor when a public constuctor with no parameters was not found.
     */
    protected void initProtectedConstructor()
    {
        try
        {
            _constructor = _objectClass.getDeclaredConstructor((Class[])null);
        }
        catch (NoSuchMethodException e2)
        {
            _constructor = null;
        }     
    }
    
	/**
	 * Gets a value of a field for an instance of this class.
	 * @param  object an instance of this class
	 * @param  field  a field of this class
	 * @return the value of the specified field for the specified instance.
	 */
	protected Object get(Object object, LyPeerField field)
	{
		LyAttributeField attribute = (LyAttributeField)field;
		Method			 getMethod = attribute.getGetMethod(this);

		if (getMethod == null)
		{
			Class<?>  objectClass = getObjectClass();

			throw new IllegalStateException("BF AppEngine : In class " + getName()
                + " '" +objectClass.getName() + "' can't find get method '"
                + attribute.getGetMethodName(this)+"()' for field " + attribute.getId());
		}

		Object    result = null;
        boolean   accessible = getMethod.isAccessible();

		try
		{
            if (!accessible)
                getMethod.setAccessible(true);
            
			result = getMethod.invoke(object);
            
            if (!accessible)
                getMethod.setAccessible(false);
			return result;
		}
		catch (InvocationTargetException e)
		{
			LyEnvironment env = getEnvironment();

			env.logError("BF AppEngine : exception caught in getValue "
				+ e + ", " + object + ", " + attribute);
			env.writeException(e.getTargetException());
			return null;
		}
		catch (Exception e)
		{
			LyEnvironment env = getEnvironment();

			env.logError("BF AppEngine : exception caught in getValue "
				+ e + ", " + object + ", " + attribute);
			env.writeException(e);
			return null;
		}
	} // end method get

	/**
	 * Sets a value of a field for an instance of this class.
	 * @param object an instance of this class
	 * @param field  a field of this class
	 * @param value  the value to set.
	 */
	protected void set(Object object, LyPeerField field, Object value)
	{
		LyAttributeField attribute = (LyAttributeField)field;
		Method			 setMethod = attribute.getSetMethod(this);

		if (setMethod == null)
            throw new IllegalStateException("BF AppEngine : can't find set method for field " + attribute.getId()
                + " on object " + object);
                
		Object[]   args = new Object[] { value };
        boolean    accessible = setMethod.isAccessible();
        
		try
		{
            if (!accessible)
                setMethod.setAccessible(true);
            
			setMethod.invoke(object, args);
            
            if (!accessible)
                setMethod.setAccessible(false);
		}
		catch (InvocationTargetException e)
		{
			LyEnvironment env = getEnvironment();

			env.logError("BF AppEngine : exception caught in setValue "
				+ e + ", " + object + ", " + attribute + ", " + value);
			env.writeException(e.getTargetException());
		}
		catch (Exception e)
		{
			LyEnvironment env = getEnvironment();

			env.logError("BF AppEngine : exception caught in setValue "
				+ e + ", " + object + ", " + attribute + ", " + value);
			env.writeException(e);
		}
	} // end method set

	/**
	 * Creates a new instance of this class.
	 * @return New instance of this class.
	 */
	public Object newInstance()
		throws InstantiationException,
			IllegalAccessException
	{
        if (_constructor == null)
            return _objectClass.newInstance();
        
        try
        {
            Object  result;
            
            // Call protected constructor
            _constructor.setAccessible(true);
            result = _constructor.newInstance((Object[])null);
            _constructor.setAccessible(false);
            return result;
        }
        catch (Exception e)
        {
            throw new InstantiationException();
        }    
	}
    
    /**
     * Recompute the object id of the specified object.
     */
    public void rehash(Object object)
    {
        LyObjectId   oldObjectId = getObjectId(object);

        rehash(object, oldObjectId);
    }

	/**
	 * Recompute the object id of the specified object.
	 */
	protected void rehash(Object object, LyObjectId oldObjectId)
	{
		LyDataSource dataSource = getDataSource();
		LyCache		 cache = dataSource.getCache();

		cache.rehash(object, oldObjectId);
	}
} // end class LyJavaClass
