/*
 * LyLocalRequestEncoder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.local;

import java.util.*;
import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.util.*;


/**
 * Local request encoder associated to the local data provider.
 */
public class LyLocalRequestEncoder extends LyRequestEncoder
{
	public static final String __VERSION =
		"$Id: LyLocalRequestEncoder.java 28422 2010-11-03 12:03:38Z ehelmste $";

	protected static Random    _random = null;


	/**
	 * Default constructor Initialization is done with the init method.
	 */
	public LyLocalRequestEncoder(LyDataProvider provider, LyDataSource dataSource)
	{
		super(provider, dataSource);
		if (_random == null)
			_random = new Random(System.currentTimeMillis());
	}


	/**
	 * Get an object.
	 * @param  objectId object identifier
	 * @param  fields   the fields to load (if null, load all)
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getObject(LyObjectId objectId, Vector<LyField> fields)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);
		return true;
	}

	/**
	 * Get a set of objects.
	 * @param  fields the fields to load (if null, load all)
	 * @param  sort   indicates sorting criterion, or null if default sort is used.
	 * @param  filter indicates filtering criterion, or null if there is no filter.
	 * @param  n      maximum number of objects to get.
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getObjects(Vector<LyField> fields, LySort sort, LyFilter filter, int n)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);
		return true;
	}

	/**
	 * Return the number of objects from the data provider .
	 * @param  filter indicates filtering criterion, or null if there is no filter.
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getSize(LyFilter filter)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, new Integer(0), null);
		return true;
	}

	/**
	 * To delete an object.
	 * @param  object the object to delete
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean deleteObject(Object object)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// Get sub objects and join objects
		LyObject     lyObject = (LyObject)object;
		LyObjectList localSubObjects, subObjects, joinObjects, arrayObjects;

		if (_response.deleteLinkedObjects())
		{
			arrayObjects = lyObject.getArrayObjects();
			localSubObjects = lyObject.getSubObjects(true);
			subObjects = lyObject.getSubObjects(false);
			joinObjects = lyObject.getJoinObjects();
		}
		else
		{
			arrayObjects = null;
			localSubObjects = null;
			subObjects = null;
			joinObjects = null;
		}

		// Delete local join objects, sub objects, peer and finally remaining sub objects
		LyApplication		 application = _provider.getApplication();
		LyTransactionManager tm = application.getTransactionManager();
		LyTransaction		 transaction;
		boolean				 local = _response.isLocal();
		boolean				 refresh = _response.isRefresh();

		transaction = tm.begin();

		// Delete join objects.
		if (joinObjects != null)
			joinObjects.delete(local, refresh);
		if (transaction.getStatus() == LyTransaction.Status.ERROR)
		{
			// There is an error, we can stop now by rollbacking the transaction.
			_response.setStatus(LyResponse.Status.ERROR);
			_response.setError(application.getError());

			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			tm.end(transaction, false);
			return false;
		}

		// Delete array objects
		if (arrayObjects != null)
			arrayObjects.delete(local, refresh);
		if (transaction.getStatus() == LyTransaction.Status.ERROR)
		{
			// There is an error, we can stop now by rollbacking the transaction.
			_response.setStatus(LyResponse.Status.ERROR);
			_response.setError(application.getError());

			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			tm.end(transaction, false);
			return false;
		}

		// Delete sub objects
		if (localSubObjects != null)
			localSubObjects.delete(local, refresh);
		if (transaction.getStatus() == LyTransaction.Status.ERROR)
		{
			// There is an error, we can stop now by rollbacking the transaction.
			_response.setStatus(LyResponse.Status.ERROR);
			_response.setError(application.getError());

			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			tm.end(transaction, false);
			return false;
		}

		// Delete remaining sub objects
		if (subObjects != null)
			subObjects.delete(local, refresh);
		if (transaction.getStatus() == LyTransaction.Status.ERROR)
		{
			// There is an error, we can stop now by rollbacking the transaction.
			_response.setStatus(LyResponse.Status.ERROR);
			_response.setError(application.getError());

			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			tm.end(transaction, false);
			return false;
		}

		// Now we can complete the response...
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object, null);

		// ... and commit the transaction.
		tm.end(transaction, true);

		return true;
	} // end method deleteObject

	/**
	 * To create a new object.
	 * @param  object the new object to create
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean createObject(Object object)
	{
		if (object instanceof LyObject)
		{
			removeProviderValues((LyObject)object, false);
			addNotNullValues((LyObject)object, false);
		}

		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object, null);

		return true;
	}

	/**
	 * To set an object.
	 * @param  object the object to set
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean setObject(Object object)
	{
		// Set the response in running mode
		LyValueSet oldValues = null;

		// For import : we use a simple request encoder and need to get
		// the old values as in the other implementations of request
		// encoder (file, sql...).
		if (object instanceof LyObject)
		{
			addNotNullValues((LyObject)object, false);

			LyObject lyObject = (LyObject)object;
			Object   peer = lyObject.getPeer();

			if (peer instanceof LyRowObject)
				oldValues = ((LyRowObject)peer).validate(null);
		}

		_response.setStatus(LyResponse.Status.RUNNING);
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object, oldValues);

		return true;
	}

	/**
	 * This method add to the specified object all the values for fieldInfos that have the mark
	 * NOT_NULL.
	 * @param object    the object to add the values
	 * @param onlyLocal spcifies if only local values are concerned.
	 */
	protected void addNotNullValues(LyObject object, boolean onlyLocal)
	{
		super.addNotNullValues(object, onlyLocal);

		if (onlyLocal)
			return;

		// We must verify that the object has an id
		LyClassInfo classInfo = object.getClassInfo();

		// If classInfo is not local, this not to this encoder to
		// compute id.
		if (!classInfo.isLocal())
			return;

		LyInfoList<LyFieldInfo>  idFieldInfos = classInfo.getIdFieldInfos();
		LyFieldInfo fieldInfo;
		int		    i, n = idFieldInfos.getSize();
		LyValue     fieldValue;
		Object	    value;

		for (i = 0; i < n; i++)
		{
			fieldInfo = idFieldInfos.getInfo(i);
			value = classInfo.getValue(object, fieldInfo);
			if (value != null)
				continue;

			if (fieldInfo instanceof LyNumberFieldInfo)
			{
                LyNumberFieldInfo.Type type = ((LyNumberFieldInfo)fieldInfo).getType();

				if (type == LyNumberFieldInfo.Type.INT)
					fieldValue = fieldInfo.buildFieldValue(new Integer(_random.nextInt()));
				else if (type == LyNumberFieldInfo.Type.SHORT)
					fieldValue = fieldInfo.buildFieldValue(new Integer(
								_random.nextInt(Short.MAX_VALUE)));
				else if (type == LyNumberFieldInfo.Type.FLOAT)
					fieldValue = fieldInfo.buildFieldValue(new Float(_random.nextFloat()));
				else
				{
					// To avoid two dates in the same ms.
					long id = ((LyLocalDataProvider)_provider).getNextId();

					fieldValue = fieldInfo.buildFieldValue(new Long(id));
				}
			}
			else if (fieldInfo instanceof LyTextFieldInfo)
				fieldValue = fieldInfo.buildFieldValue(LyString.getRandomId());
			else if (fieldInfo instanceof LyDateFieldInfo)
				fieldValue = fieldInfo.buildFieldValue(new Date());
			else
				fieldValue = null;

			if (fieldValue != null)
				classInfo.setValue(object, fieldInfo, fieldValue.getValue());
		}
	} // end method addNotNullValues
} // end class LyLocalRequestEncoder
