/*
 * LySqlConnection.java
 * Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.peer.sql;

import java.sql.*;
import java.util.*;

import javax.sql.*;
import java.io.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.info.location.*;
import leon.misc.*;
import leon.peer.*;


/**
 * This class is used as a connection for relational database.
 * @version $Id: LySqlConnection.java 34477 2012-03-12 16:49:49Z jlsaouli $
 * @author  Lyria
 * @see     LyConnection
 */
public class LySqlConnection extends LyConnection
{
	public static final String __VERSION =
		"$Id: LySqlConnection.java 34477 2012-03-12 16:49:49Z jlsaouli $";

	/** Max length for username and password */
	static final int		   MAX_LENGTH = 30;
	
	/** Default statement timeout */
    public static int          STATEMENT_TIMEOUT = 30;

	/**  */
	protected String		   _name;

	/**  */
	protected String		   _username;

	/**  */
	protected String		   _password;

	/** Connection to the database. */
	protected Connection	   _connection;

	/**  */
	protected Statement		   _statement;

	/**  */
	protected DataSource	   _dataSource;

	/** Flag to indicate if we must call the hook for update requests. */
	protected boolean		   _hookForUpdate;

	/** Flag to indicate that this connection is in autocommit mode (true) or not. */
	protected boolean		   _isAutoCommit;
	
	/** Timeout used for statements of this connection. */
	protected int			   _statementTimeout;

	/**
	 * Constructor.
	 * @param application Application of the data provider that manages this connection.
	 * @param name        Name of the database.
	 */
	public LySqlConnection(LyApplication application, String name)
	{
		this(application, name, null);
	}

	/**
	 * Constructor.
	 * @param application Application of the data provider that manages this connection.
	 * @param name        Name of the database.
	 * @param dataSource  JDBC DataSource providing the JDBC connection.
	 */
	public LySqlConnection(LyApplication application, String name,
		DataSource dataSource)
	{
		super(application);
		_name = name;
		_dataSource = dataSource;
	}

	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		super.init();

		_connection = null;
		_name = null;
		_username = null;
		_password = null;
		_statement = null;
		_dataSource = null;
		_hookForUpdate = true;
		_isAutoCommit = true;
		_statementTimeout = STATEMENT_TIMEOUT;
	}

	/**
	 * To set the _hookForUpdate attribute.
	 * @param hookForUpdate true to call the hook for update request, false otherwise.
	 */
	protected void setHookForUpdate(boolean hookForUpdate)
	{
		_hookForUpdate = hookForUpdate;
	}
	
	/**
	 * Sets the <code>_statementTimeout</code> attribute for current connection.
	 * @param statementTimeout new value for the <code>_statementTimeout</code> attribute
	 */
	public void setStatementTimeout(int statementTimeout)
	{
		_statementTimeout = statementTimeout;
	}

	/**
	 * Open the physical connection to the database.
	 * @param  username user id.
	 * @param  password user password.
	 * @return true : connection established, false : connection refused
	 */
	public boolean open(String username, String password)
	{
		return open(username, password, null);
	}

	/**
	 * Open the physical connection to the database.
	 * @param  username   user id.
	 * @param  password   user password.
	 * @param  driverName The driver class name.
	 * @return true : connection established, false : connection refused
	 */
	public boolean open(String username, String password, String driverName)
	{
		// username and password must be compatible with the structures managed
		// by the DBMS
		if ((username != null) && (username.length() > MAX_LENGTH))
			return false;
		if ((password != null) && (password.length() > MAX_LENGTH))
			return false;

		_username = username;
		_password = password;

		// Open the connection with the DBMS
		LyEnvironment environment = _application.getEnvironment();

		if (_dataSource != null)
		{
			if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
				environment.logOut("Searching SQL connection for server " + _name);
			try
			{
				if (_dataSource instanceof XADataSource)
				{
					XAConnection xac;

					xac = ((XADataSource)_dataSource).getXAConnection(_username, _password);
					_connection = xac.getConnection();
				}
				else if (_dataSource instanceof ConnectionPoolDataSource)
				{
					PooledConnection pc;

					pc = ((ConnectionPoolDataSource)_dataSource).getPooledConnection(_username,
							_password);
					_connection = pc.getConnection();
				}
				else if ((_username != null) && (_username.trim().length() > 0))
				{
					_connection = _dataSource.getConnection(_username, _password);
				}
				else
					_connection = _dataSource.getConnection();
			}
			catch (SQLException e)
			{
				handleError(e);
				return false;
			}
		}
		else
		{
			environment.logInfo("Opening SQL connection with server " + _name);

			Properties properties = getProperties();
properties.put("includeSynonyms", "true");

			if (driverName != null )
			{
				try
				{
			    	
					Class<?>  driverClass = _application.loadClass(driverName,
							environment.getMessage("LY_ER_DRIVER_NOT_FOUND", driverName));

					if (driverClass == null)
					{
						environment.logError(environment.getMessage("LY_ER_REGISTER_DRIVER",
								driverName));
						return false;
					}
					
					Driver driver = (Driver)driverClass.newInstance();

					_connection = driver.connect(_name, properties);
				}
				catch (IllegalStateException e)
				{
					environment.writeException(e);
					_application.setError(e.getMessage());
					return false;
				}
				catch (Exception e)
				{
						environment.writeException(e);
						_application.setError(e.getMessage());
						return false;
				}
			}

			if (_connection == null)
			{
				try
				{
					_connection = DriverManager.getConnection(_name, properties);
				}
				catch (SQLException e)
				{
					handleError(e);
					return false;
				}
			}
		}

		if (environment.testResource("LY_DISABLE_AUTOCOMMIT_AT_OPEN"))
		{
			// Set the auto commit mode to false and the flag to prevent the use of the
			// autocommit method.
			autoCommit(false);
			_isAutoCommit = false;
		}

		_closed = false;
		return true;
	} // end method open

	/**
	 * Close the physical connection to the database.
	 * @return true : connection is closed, false : an error occured.
	 */
	public boolean close()
	{
		Connection connection = getConnection();
		
		if (_dataSource != null)
		{
			// Before closing the connection
			// we must close the current statement if any.
			closeStatement();
			
			try
			{
				if (connection != null)
					connection.close();
				LyEnvironment environment = _application.getEnvironment();

				environment.logInfo("Closing SQL connection with datasource"
					+ _name);
			}
			catch (SQLException e)
			{
				handleError(e);
				return false;
			}
			finally
			{
				_connection = null;
			}
		}
		else if (connection != null)
		{
			// Closing the connection
			try
			{
				
				LyEnvironment environment = _application.getEnvironment();

				environment.logInfo("Closing SQL connection with server "
					+ _name);
				connection.close();
			}
			catch (SQLException e)
			{
				_connection = null;
				_closed = true;
				handleError(e);
				return false;
			}

			_closed = true;
			_connection = null;
			return true;
		}

		_closed = true;
		return false;
	} // end method close

	/**
	 * Connection released.
	 */
	public void released()
	{
		closeStatement();
		super.released();
	}

	/**
	 * To get the properties of this SQL connection.
	 * @return The properties of this connection.
	 */
	public Properties getProperties()
	{
		Properties properties = new Properties();

		if (_username != null)
			properties.put("user", _username);
		if (_password != null)
			properties.put("password", _password);

		return properties;
	}

	/**
	 * To get meta-information from the database.
	 * @return the meta-information of the database.
	 */
	public DatabaseMetaData getMetaData()
	{
		DatabaseMetaData metaData = null;

		try
		{
			Connection connection = getConnection();
			
			metaData = connection.getMetaData();

			LyEnvironment environment = _application.getEnvironment();

			if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
				environment.logOut("Gets DatabaseMetaData of SQL connection with server "
					+ _name + "= " + metaData);
		}
		catch (SQLException e)
		{
			handleError(e);
			return null;
		}

		return metaData;
	}

	/**
	 */
	protected void closeStatement()
	{
		if (_statement == null)
			return;
		try
		{
			_statement.close();
		}
		catch (SQLException e)
		{
			handleError(e);
		}

		_statement = null;
	}

	/**
	 * Sends a request to the database. The request type must be : select, count.
	 * @param  request the sql request.
	 * @return the ResultSet returned by the driver.
	 */
	public ResultSet sendQueryRequest(String request)
	{
		return sendQueryRequest(request, -1);
	}

	/**
	 * Sends a request to the database. The request type must be : select, count.
	 * @param  request the sql request.
	 * @param  limit   maximum number of rows to fetch (-1 for no limit)
	 * @return the ResultSet returned by the driver.
	 */
	public ResultSet sendQueryRequest(String request, int limit)
	{
		LyEnvironment environment = _application.getEnvironment();

		if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
		{
			environment.logOut(">>> Request sent : " + request);
			if (environment.isDebugLevel(LyEnvironment.DEBUG_ALL))
				environment.writeException(new Throwable());
		}

		java.util.Date date = new java.util.Date();
		ResultSet	   rs = null;

		try
		{
			if (_statement != null)
				_statement.close();

			Connection connection = getConnection();

			if (connection == null)
			{
				setError("Can't get a valid connection. Connection is closed.");
				return null;
			}
			
			_statement = connection.createStatement();
			setQueryTimeout(_statementTimeout);
			setLimit(limit);
			
			rs = _statement.executeQuery(request);
		}
		catch (SQLException sqle)
		{
			handleError(sqle);
			return null;
		}

		if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
			environment.logOut("<<< End of request ("
				+ ((new java.util.Date()).getTime() - date.getTime()) + " ms)");

		return rs;
	} // end method sendQueryRequest

	/**
	 * Sends a request to the database. The request type must be : set, create, delete.
	 * @param  request the sql request.
	 * @return the number of lines returned by the driver.
	 */
	public int sendUpdateRequest(String request)
	{
		LyEnvironment environment = _application.getEnvironment();

		if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
		{
			environment.logOut(">>> Request sent : " + request);
			if (environment.isDebugLevel(LyEnvironment.DEBUG_ALL))
				environment.writeException(new Throwable());
		}

		java.util.Date date = new java.util.Date();
		int			   nbLignes;

		try
		{
			if (_statement != null)
				_statement.close();

			if (_hookForUpdate)
			{
				LyApplication				   application = _application.getRootApplication();
				LyApplicationBehaviorInterface behavior = application.getApplicationBehavior();
				LyLocation					   location = _provider.getLocation();

				try
				{
					behavior.update(application, getUID().toString(), location.getId(), request);
				}
				catch (Exception e)
				{
					// The hook has throwed an exception, we cancel the request.
					_statement = null;
					setError(e.getMessage());
					return -1;
					//environment.writeException(e);
				}
			}
			
			if (LySqlRequestEncoder.getCurrentBlob(0) == null)
			{	
				Connection connection = getConnection();

				if (connection == null)
				{
					setError("Can't get a valid connection. Connection is closed.");
					_statement = null;
					return -1;
				}

				_statement = connection.createStatement();
				setQueryTimeout(_statementTimeout);

				nbLignes = _statement.executeUpdate(request);
			}
			else
			{
				Connection connection = getConnection();

				if (connection == null)
				{
					setError("Can't get a valid connection. Connection is closed.");
					_statement = null;
					return -1;
				}

				_statement = connection.prepareStatement(request);
				setQueryTimeout(_statementTimeout);
	
				int 		index = 0;
				Object[] 	blobInfo = LySqlRequestEncoder.getCurrentBlob(index);

				while (blobInfo != null)
				{
					if (blobInfo[1] instanceof Reader)
						((PreparedStatement)_statement).setCharacterStream((index+1), (Reader)blobInfo[1], ((Integer)blobInfo[2]).intValue()); // be careful : SQL starts at index 1.
					else
						((PreparedStatement)_statement).setBinaryStream((index+1), (InputStream)blobInfo[1], ((Integer)blobInfo[2]).intValue()); // be careful : SQL starts at index 1.

					// check if we have more blobs ?
					index++;
					blobInfo = LySqlRequestEncoder.getCurrentBlob(index);
				}

				LySqlRequestEncoder.clearCurrentBlobs();

				nbLignes = ((PreparedStatement)_statement).executeUpdate();
			}
		}
		catch (SQLException e)
		{
			LySqlRequestEncoder.clearCurrentBlobs();
			handleError(e);
			return -1;
		}

		if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
			environment.logOut("<<< End of request ("
				+ ((new java.util.Date()).getTime() - date.getTime()) + " ms)");

		return nbLignes;
	} // end method sendUpdateRequest

	/**
	 * Method to handle SQLException.
	 * @param e SQLEXception to handle.
	 */
	public void handleError(SQLException e)
	{
		LyDataSession session = _application.getCurrentSession();

		if (session != null)
			session.handleException(e);

		LyEnvironment environment = _application.getEnvironment();
		StringBuffer  error = new StringBuffer();
		int			  i = 0;
		String		  message;
		String		  key;

		for (; e != null; e = e.getNextException())
		{
			if (i == 0)
				i++;
			else
				error.append("\n");

			key = e.getSQLState();
			message = environment.getMessage(key);
			if ((message == null) || message.equals(key))
			{
				key = "" + e.getErrorCode();
				message = environment.getMessage(key);
				if (message.equals(key))
					message = e.getMessage();
			}
			if (message.length() > 255)
			{
				int idx = message.indexOf('\n');

				if (idx > 0)
					message = message.substring(0, idx);
				else
				{
					message = message.substring(0, 255);
					message = message.concat("...");
				}
			}
			error.append(message);
		}

		setError(error.toString());

		if (!_closed)
		{
			try
			{
		       Connection connection = getConnection();
				
				if ((connection == null) || connection.isClosed())
					_closed = true;
			}
			catch (SQLException ex)
			{
				_closed = true;
			}
		}
		else
			environment.logError("Connection is already closed");
	} // end method handleError

	/**
	 */
	public void commit()
	{
		Connection connection = getConnection();

		if (connection == null)
			return;
		try
		{
			LyEnvironment environment = _application.getEnvironment();

			if ((environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
					&& (!connection.getAutoCommit()))
				environment.logOut("* commit");
			connection.commit();
		}
		catch (SQLException e)
		{
			handleError(e);
		}
	}

	/**
	 */
	public void rollback()
	{
		Connection connection = getConnection();

		if (connection == null)
			return;
		try
		{
			connection.rollback();
			LyEnvironment environment = _application.getEnvironment();

			if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
				environment.logOut("* rollback");
		}
		catch (SQLException e)
		{
			handleError(e);
		}
	}

	/**
	 * To change the autoCommit mode of the JDBC connection.
	 * @param status the new mode.
	 */
	public void autoCommit(boolean status)
	{
		Connection connection = getConnection();
		
		if ((connection == null)
				|| (!_isAutoCommit))
			return;

		try
		{
			LyEnvironment environment = _application.getEnvironment();

			connection.setAutoCommit(status);
			if (!status && (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)))
				environment.logOut("*  setAutoCommit "+status+" Transaction isolation = "
					+ connection.getTransactionIsolation());
		}
		catch (SQLException e)
		{
			handleError(e);
		}
	}

	/**
	 * Gets connection to the database.
	 * Warning ! Any modification on the returned object may cause major dysfunctions.
	 * @return The connection to the database.
	 */
	public Connection getConnection()
	{
		if (_closed)
		{
			LyEnvironment environment = _application.getEnvironment();
			
			environment.logWarning("[LySQLConnection] got a closed SQL Connection !");
			return null;
		}
		return _connection;
	}
	
    /**
     * Gets current statement to the database.
     * Warning ! Any modification on the returned object may cause major dysfunctions.
     * @return The current statement.
     */
    public Statement getStatement()
    {
        return _statement;
    }
    
    protected void setQueryTimeout(int timeout)
    {
        if (_statement == null)
            return;
        
        try
        {
            if (timeout > 0)
                _statement.setQueryTimeout(timeout);
        }
        catch (Exception e)
        {
            LyEnvironment environment = _application.getEnvironment();
            
            if (environment.isDebugLevel(LyEnvironment.DEBUG_ALL))
                environment.writeException(e);
        } 
    }
    
    protected void setLimit(int limit)
    {
        if (limit < 0)
            // 0 indicates no limit for JDBC.
            limit = 0;
        
        // Some drivers or RDBMS may not support this feature so we must
        // enclose it in a try-catch block.
        try
        {
            _statement.setMaxRows(limit);
        }
        catch (Exception e)
        {
            LyEnvironment environment = _application.getEnvironment();
            
            if (environment.isDebugLevel(LyEnvironment.DEBUG_ALL))
                environment.writeException(e);
        }
    }
} // end class LySqlConnection
