/*
 * LySqlDataProvider.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.peer.sql;

import java.io.*;
import java.sql.*;
import java.util.*;

import javax.naming.*;
import javax.naming.spi.*;
import javax.sql.*;

import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.*;
import leon.info.location.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.union.*;
import leon.peer.util.*;
import leon.peer.util.LyColumnField.*;


/**
 * This class is used by DataSources to transmit their requests to an instance of SqlRequestEncoder.
 *
 * @version $Id: LySqlDataProvider.java 41560 2012-11-19 12:04:44Z bangiole $
 * @author  Lyria
 * @see     LyDataProvider
 * @see     LySqlRequestEncoder
 * @see     LySqlConnection
 */
public class LySqlDataProvider extends LyTableDataProvider
	implements LyJoinDataProvider
{
	public static final String							    __VERSION =
		"$Id: LySqlDataProvider.java 41560 2012-11-19 12:04:44Z bangiole $";

	/** Constant for datasource name parameter in the property file. */
	public static final String							    DATASOURCE_NAME = "dataSourceName";

	/** Default delay. */
	static final int									    DEFAULT_DELAY = 10000;

	/**  */
	protected String									    _databaseName;

	/**  */
	protected String									    _driverName;

	/**  */
	protected String									    _connectionName;

	/**  */
	protected String									    _username;

	/**  */
	protected String									    _password;

	/**  */
	protected String									    _valueSep;

	/**  */
	protected boolean									    _trim;

	/**  */
	protected boolean									    _metadata;

	/**  */
	protected Hashtable<String, LyMetaDataTableInformation> _metaInformation;

	/**  */
	protected HashMap<LyField, LySqlSpecificEncoder>	    _specificEncoders;

	/**  */
	protected int										    _delay;

	/**  */
	protected DataSource								    _dataSource;

	/**
	 * Default constructor. Initialization is done with the init method.
	 *
	 * @param application the application of this provider.
	 * @param location    the info location defining this provider.
	 */
	public LySqlDataProvider(LyApplication application, LyLocation location)
	{
		super(application, location);
	}

	/**
	 * Updates this connection parameters using location information and application environment.
	 */
	protected void updateConnectionParameters(LyLocation location)
	{
		LySqlLocation sqlLocation = (LySqlLocation)location;

		_driverName = sqlLocation.getDriver();
		_connectionName = sqlLocation.getUrl();
		_username = sqlLocation.getUser();
		_password = sqlLocation.getPassword();
		_databaseName = sqlLocation.getDbName();

		// Check within environment if the values are in the ini file
		LyEnvironment env = sqlLocation.getEnvironment();

		if ((_driverName != null) && (_driverName.startsWith("$")))
		{
			String d = env.getEnv(_driverName);

			if (d == null)
			{
				String err = env.getMessage("LY_ER_INIT_PARAM", _driverName);

				env.logErr(err);
			}
			else
				_driverName = d;
		}

		if ((_connectionName != null) && (_connectionName.startsWith("$")))
		{
			String c = env.getEnv(_connectionName);

			if (c == null)
			{
				String err = env.getMessage("LY_ER_INIT_PARAM", _connectionName);

				env.logErr(err);
			}
			else
				_connectionName = c;
		}

		if ((_username != null) && (_username.startsWith("$")))
		{
			String c = env.getEnv(_username);

			if (c == null)
			{
				String err = env.getMessage("LY_ER_INIT_PARAM", _username);

				env.logErr(err);
			}
			else
				_username = c;
		}

		if ((_password != null) && (_password.startsWith("$")))
		{
			String c = env.getEnv(_password);

			if (c == null)
			{
				String err = env.getMessage("LY_ER_INIT_PARAM", _password);

				env.logErr(err);
			}
			else
				_password = c;
		}

		if ((_databaseName != null) && (_databaseName.startsWith("$")))
		{
			String c = env.getEnv(_databaseName);

			if (c == null)
			{
				String err = env.getMessage("LY_ER_INIT_PARAM", _databaseName);

				env.logErr(err);
			}
			else
				_databaseName = c;
		}

		String delayStr = env.getResource("LY_SQL_MAX_DELAY");

		_delay = ((delayStr == null) ? DEFAULT_DELAY : LyTools.intFromString(delayStr));
	} // end method updateConnectionParameters

	/**
	 * Sets location on this provider.
	 */
	public void setLocation(LyLocation location)
	{
		if (location instanceof LySqlLocation)
		{
			LySqlLocation sqlLocation = (LySqlLocation)location;

			// Try to extract the information from the location
			// Otherwise, the specific implemtation should set the parameters
			updateConnectionParameters(sqlLocation);
			_valueSep = sqlLocation.getValueSep();

			String trim = sqlLocation.getEnvironment().getEnv(sqlLocation.isTrim());
			if(trim != null)
				_trim = Boolean.parseBoolean(trim);
			else
				_trim = Boolean.parseBoolean(sqlLocation.isTrim());

			String metadata = sqlLocation.getEnvironment().getEnv(sqlLocation.useMetadata());
			if(metadata != null)
				_metadata = Boolean.parseBoolean(metadata);
			else
				_metadata = Boolean.parseBoolean(sqlLocation.useMetadata());

			String nbConStr = sqlLocation.getEnvironment().getEnv(sqlLocation.getNbConnection());
			int nbConnection;

			try
			{
				if(nbConStr != null)
					nbConnection = Integer.parseInt(nbConStr);
				else
					nbConnection = Integer.parseInt(sqlLocation.getNbConnection());
			}
			catch(NumberFormatException nfe)
			{
				nbConnection = 1;
			}

			setNbConnection(nbConnection);
		}

		super.setLocation(location);
	}

	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		super.init();

		_driverName = null;
		_connectionName = null;
		_username = null;
		_password = null;
		_trim = false;
		_metadata = false;
		_metaInformation = null;
		_specificEncoders = null;
		_delay = DEFAULT_DELAY;
		_dataSource = null;
	}

	/**
	 * Accessor to the delay member.
	 *
	 * @return the maximum delay for a request before displaying a trace.
	 */
	protected int getDelay()
	{
		return _delay;
	}

	/**
	 * Modifier of the trim member.
	 *
	 * @param trim true to trim all values provided by the database, false otherwise.
	 */
	public void setTrim(boolean trim)
	{
		_trim = trim;
	}

	/**
	 * This method searches for a property file containing informations about a DataSource.
	 *
	 * @return the properties defining a DataSource for this provider.
	 */
	protected Properties getProperties()
	{
		String extension, dir;
		String appId = _application.getId();
		int    p = appId.lastIndexOf('.');

		if (p < 0)
		{
			extension = appId;
			dir = appId;
		}
		else
		{
			extension = appId.substring(p + 1);
			dir = appId.replace('.', '/');
		}

		String		  filename = dir + "/properties/" + _location.getId() + "_"
			+ extension + ".properties";
		LyEnvironment environment = getEnvironment();
		InputStream   stream = environment.getResourceAsStream(filename,
				getApplication().getClassLoader());

		if (stream == null)
			return null;

		Properties result = new Properties();

		try
		{
			result.load(stream);
		}
		catch (Exception e)
		{
			if (environment.isDebugLevel(LyEnvironment.DEBUG_ERROR))
				environment.writeException(e);
			result = null;
		}

		return result;
	} // end method getProperties

	/**
	 * Accessor to a connection instance from the connection pool. For internal use only.
	 *
	 * @param  b for internal use only
	 * @return a connection from the connection pool
	 */
	protected LyConnection getConnection(boolean b)
	{
		if (_dataSource != null)
		{
			LyConnection connection = createConnection();

			connection.setDataProvider(this);
			connection.autoCommit(true);
			connection.setError(null);
			connection.reserved();
			return connection;
		}
		else
			return super.getConnection(b);
	}

	/**
	 * Return a connection instance to the connection pool. For internal use only.
	 *
	 * @param connection the connection to be returned to the pool.
	 * @param t          For internal use only
	 */
	protected void releaseConnection(LyConnection connection, boolean t)
	{
		if (_dataSource == null)
			super.releaseConnection(connection, t);
		else
			connection.close();
	}

	/**
	 * Initialize validation and create connections Overriden to load JDBC driver and load meta
	 * data.
	 *
	 * @return true is the operation completed successfully
	 */
	protected boolean startValidate()
	{
		// Search for a properties file containing JNDI informations for
		// a DataSource.
		Properties properties = getProperties();

		if (properties != null)
		{
			// Replace properties which may be set in the environment
			// These properties starts with $NAME and may be set in the ini file
			Enumeration<?>   names = properties.propertyNames();
			LyEnvironment environment = getEnvironment();
			String		  name, val;

			while (names.hasMoreElements())
			{
				name = (String)names.nextElement();
				val = properties.getProperty(name);

				if ((val != null) && (val.startsWith("$")))
				{
					val = environment.getEnv(val);
					properties.setProperty(name, val);
				}
			}

			if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
			{
				environment.logOut("* LySqlDataProvider : found properties");
				properties.list(environment.out);
			}

			String dataSourceName = properties.getProperty(DATASOURCE_NAME);

			if (dataSourceName != null)
				properties.remove(DATASOURCE_NAME);
			else
				dataSourceName = "jdbc/" + _location.getId();

			loadDataSource(dataSourceName, properties);
		}

		if ((_dataSource == null) && (_connectionName != null) && (!_connectionName.startsWith("jdbc:")))
		{
			if (getEnvironment().isDebugLevel(LyEnvironment.DEBUG_DBMS))
				getEnvironment().logOut("initialize datasource for name " + _connectionName);
			loadDataSource(_connectionName, properties);
		}

		if (_dataSource == null)
		{
			// Loading the driver.
			if (!loadDriver())
				return false;
		}

		_metaInformation = getMetaData();

		return true;
	} // end method startValidate

	/**
	 * This method is used to build all the class peers for all the classInfo. It must be
	 * specialized by subclasses. Basically, this method associates the dataProvider with local
	 * classes. A class is local if there is no binding.
	 *
	 * @return true if the end of validation is OK, false otherwise.
	 */
	public boolean endValidate()
	{
		// Free meta information
		_metaInformation = null;

		return super.endValidate();
	}

	/**
	 * To get meta-information from the database.
	 *
	 * @return Meta-information from the database.
	 */
	protected Hashtable<String, LyMetaDataTableInformation> getMetaData()
	{
		if (!_metadata)
			return null;

		LySqlConnection connection = (LySqlConnection)getConnection();

		if (connection == null)
			return null;

		DatabaseMetaData metaData = connection.getMetaData();

		releaseConnection(connection);

		if (metaData == null)
			return null;

		Hashtable<String, LyMetaDataTableInformation> result = getTables(metaData);

		if (result == null)
			return null;

		if (completeTables(metaData, result))
			return result;

		return null;
	}

	/**
	 * Factory to create a new LyColumnField instance. May be specialized in subclasses, default
	 * return a LyColumnField instance.
	 *
	 * @param  fieldBinding reference of the fieldBinding associated with the columnField to create
	 * @return a LyColumnField instance
	 */
	protected LyColumnField createColumnField(LyFieldBinding fieldBinding)
	{
		LySqlColumnField field = new LySqlColumnField();

		if (fieldBinding.getCBind() != null)
			field.setClassBinding(fieldBinding.getCBind());
		return field;
	}

	/**
	 * Determine the class for memory storage associated to a columnField. This method is internally
	 * used by the validate method. Method should be overriden for optimization by subclasses.
	 *
	 * @param  tableClass  the peer table class reference as specified in validate method.
	 * @param  columnField peer field associated with the value we store.
	 * @param  fieldInfo   data description of the meta model.
	 * @return the java class used to store data in memory.
	 * @see    #validate()
	 */
	protected Class<?> getValueClass(LyTableClass tableClass,
		LyColumnField columnField, LyFieldInfo fieldInfo)
	{
		Class<?> valueClass = null;
		LyMetaDataColumnInformation column = null;

		if (_metaInformation != null)
		{
			String					   name;
			LyMetaDataTableInformation table;

			name = tableClass.getName();
			table = _metaInformation.get(name);

			if (table != null)
			{
				name = columnField.getId();
				column = table.getColumn(name);
				if (column != null)
				{
					short type = column.getDataType();

					switch (type)
					{
					case Types.BOOLEAN:
						valueClass = Boolean.class;
						break;
					case Types.TINYINT:
						valueClass = Short.class;
						break;
					case Types.SMALLINT:
						valueClass = Short.class;
						break;
					case Types.INTEGER:
						valueClass = Integer.class;
						break;
					case Types.BIGINT:
						valueClass = Long.class;
						break;
					case Types.REAL:
						valueClass = Float.class;
						break;
					case Types.FLOAT:
					case Types.DOUBLE:
						valueClass = Double.class;
						break;
					case Types.CHAR:
					case Types.VARCHAR:
					case Types.LONGVARCHAR:
						if (fieldInfo.isMultiple())
							valueClass = String[].class;
						else
							valueClass = String.class;
						break;
					case Types.DATE:
					case Types.TIME:
					case Types.TIMESTAMP:
						valueClass = java.util.Date.class;
						((LySqlColumnField)columnField).setDateType(type);
						break;
					case Types.NULL:
					case Types.OTHER:
					case Types.DECIMAL:
					case Types.NUMERIC:
					case Types.BIT:
					case Types.BINARY:
					case Types.VARBINARY:
					case Types.LONGVARBINARY:
					default:
						break;
					}
				}
			}
		}
		// ** Detect CONSTRAINTS
		if (fieldInfo instanceof LyRelationFieldInfo)
		{
			if (!fieldInfo.isMultiple())
			{
				// detect foreign keys
				if ((column != null) && (column.isImportedKey()))
				{
					columnField.addContraint(CONSTRAINT.FOREIGN_KEY);
				}
			}
		}

		if (valueClass != null)
			return valueClass;

		if (fieldInfo instanceof LyChoiceFieldInfo)
		{
			LyChoiceFieldInfo choiceFieldInfo = (LyChoiceFieldInfo)fieldInfo;

			if (booleanTypeSupported()
					&& (choiceFieldInfo.getType() == LyChoiceFieldInfo.Type.BOOLEAN))
			{
				// Check if all values are boolean values i.e are equals ignoring cases to true or false
				LyChoiceFieldInfo choice = (LyChoiceFieldInfo)fieldInfo;
				LyChoiceItem	  item;
				short			  i, n = choice.getChoiceItemCount();
				boolean			  booleanValues = true;
				String			  value;

				for (i = 0; (i < n) && booleanValues; i++)
				{
					item = choice.getChoiceItem(i);
					value = item.getValue().toString();
					booleanValues &= (value.equalsIgnoreCase("true") || value.equalsIgnoreCase("false"));
				}
				if (!booleanValues)
				{
					// Check if all values can be casted to short
					boolean			  shortValues = true;

					for (i = 0; (i < n) && shortValues; i++)
					{
						item = choice.getChoiceItem(i);
						shortValues &= (item.getShortValue() >= 0);
					}

					if (shortValues)
						valueClass = Short.class;
					else
						valueClass = String.class;
				}
				else
					valueClass = Boolean.class;
			}
			else if (choiceFieldInfo.isMultiple())
			{
				LyChoiceItem choiceItem;
				short i, s;
				for (i = 0; i < (choiceFieldInfo.getChoiceItemCount()) && (valueClass == null); i++)
				{
					choiceItem = choiceFieldInfo.getChoiceItem(i);
					s = choiceItem.getShortValue();

					if (s < 0 || s > 63)
						valueClass = String.class; // not a string or above Long capacity
				}
				if (valueClass == null)
					valueClass = Long.class; // all numeric, may save a long value
			}
			else
			{
				// Check if all values can be casted to short
				LyChoiceFieldInfo choice = (LyChoiceFieldInfo)fieldInfo;
				LyChoiceItem	  item;
				short			  i, n = choice.getChoiceItemCount();
				boolean			  shortValues = true;

				for (i = 0; (i < n) && shortValues; i++)
				{
					item = choice.getChoiceItem(i);
					shortValues &= (item.getShortValue() >= 0);
				}

				if (shortValues)
					valueClass = Short.class;
				else
					valueClass = String.class;
			}
		}
		else if (fieldInfo instanceof LyRelationFieldInfo)
		{
			if (fieldInfo.isMultiple())
				valueClass = String[].class;
			else
			{
				// We search for the type of all fields id of the pointed
				// classes.
				// If they are all the same, the valueClass is the value class
				// of
				// the id fields else it's a string.
				int		    i, n;
				LyClassInfo classInfo;
				LyInfoList<LyFieldInfo>  idFields;
				LyFieldInfo idFieldInfo;

				n = ((LyRelationFieldInfo)fieldInfo).getClassInfoCount();
				for (i = 0; i < n; i++)
				{
					classInfo = ((LyRelationFieldInfo)fieldInfo).getClassInfo(i);
					idFields = classInfo.getIdFieldInfos();
					if ((idFields == null) || (idFields.getSize() != 1))
					{
						// If the field binding is a FBINDS, try to find the correct type
						LyFieldBinding binding = fieldInfo.getBinding();

						if (binding instanceof LyFieldsBinding)
						{
							LyFieldsBinding fbinds = (LyFieldsBinding)binding;
							int			    nbBinding = fbinds.getFieldBindingCount();
							int			    j;

							if (nbBinding != idFields.getSize())
							{
								valueClass = String.class;
								break;
							}

							for (j = 0; j < nbBinding; j++)
							{
								binding = fbinds.getFieldBinding(j);

								if (columnField.getId().equals(binding.getBinding()))
								{
									// We have found the correct F_BIND.
									idFieldInfo = idFields.getInfo(j);

									if (valueClass == null)
									{
										valueClass = idFieldInfo.getValueClass();
										// Workaround for relation fields whith tag id.
										if (valueClass == LyObjectId.class)
											valueClass = String.class;
									}
									else if (valueClass != idFieldInfo.getValueClass())
									{
										// The previous value class for this field is not the same has the current one,
										// use String.
										valueClass = String.class;
										break;
									}
									break;
								}
							}
						}
						else
						{
							valueClass = String.class;
							break;
						}
					}
					else
					{
						idFieldInfo = idFields.getInfo(0);
						if (valueClass == null)
						{
							valueClass = idFieldInfo.getValueClass();
							// Workaround for relation fields whith tag id.
							if (valueClass == LyObjectId.class)
								valueClass = String.class;
						}
						else if (valueClass != idFieldInfo.getValueClass())
						{
							valueClass = String.class;
							break;
						}
					}
				}

				if (valueClass == null)
					valueClass = String.class;
			}
		}
		else if (fieldInfo instanceof LyArrayFieldInfo)
		{
			LyFieldInfo fi = ((LyArrayFieldInfo)fieldInfo).getFieldInfo();

			valueClass = fi.getValueClass();
		}
		else if (fieldInfo instanceof LyDateFieldInfo)
		{
			valueClass = fieldInfo.getValueClass();

			if (((LyDateFieldInfo)fieldInfo).getType() != LyDateFieldInfo.Type.DATE_TIME)
				columnField.setFormat(((LyDateFieldInfo)fieldInfo).getStringDateFormat());
		}
		else
			valueClass = fieldInfo.getValueClass();

		if (valueClass == java.util.Date.class)
			((LySqlColumnField)columnField).setDateType(Types.TIMESTAMP);

		return valueClass;
	} // end method getValueClass

	/**
	 * Create a peer field with the specified id for the field info of the given class binding.
	 * Overridden to set the date type for a date field if the javaType is given in the model.
	 * @param  id           Identifier of the new peer field.
	 * @param  fieldInfo    Field info for which a peer field is created.
	 * @param  classBinding Class binding to which the field info belongs to.
	 * @return The new peer field.
	 */
	protected LyPeerField createPeerField(String id, LyFieldInfo fieldInfo,
			LyClassBinding classBinding)
	{
		LyPeerField		columnField = super.createPeerField(id, fieldInfo, classBinding);

		if (columnField.getValueClass() == java.util.Date.class)
		{
			LyFieldBinding fieldBinding = fieldInfo.getBinding();

			// Check that the binding is not a F_BINDS. If It's the case, we have to give the correct sub binding.
			if (fieldBinding instanceof LyFieldsBinding)
			{
				// we must find the correct binding
				LyFieldsBinding fbinds = (LyFieldsBinding)fieldBinding;

				for (int i = 0; i < fbinds.getFieldBindingCount(); i++)
				{
					fieldBinding = fbinds.getFieldBinding(i);
					if (id.equals(fieldBinding.getBinding()))
						break;
					else
						fieldBinding = null;
				}
			}

			if (("java.util.Date".equals(fieldBinding.getJavaType()))
							|| ("java.sql.Date".equals(fieldBinding.getJavaType())))
				((LySqlColumnField)columnField).setDateType(Types.TIMESTAMP);
		}

		return columnField;
	}

	/**
	 * Determine the field is an identifier on the data provider Internal use of the validate
	 * method.Overriden to check if the field is a primary key Default return true if the mark ID is
	 * set.
	 *
	 * @param  tableClass   the peer table class reference as specified in validate method.
	 * @param  columnField  peer field associated with the value we store.
	 * @param  fieldBinding binding description of the column field.
	 * @return true if the given field is a primary key for the given class, false otherwise.
	 * @see    #validate()
	 */
	protected boolean isIdField(LyTableClass tableClass,
		LySqlColumnField columnField, LyFieldBinding fieldBinding)
	{
		if (_metaInformation != null)
		{
			String					   name;
			LyMetaDataTableInformation table;

			name = tableClass.getName();
			table = _metaInformation.get(name);

			if (table != null)
			{
				LyMetaDataColumnInformation column;

				name = columnField.getId();
				column = table.getColumn(name);

				if (column != null)
				{
					int keyOrder = column.getKeyOrder();

					if (keyOrder != LyMetaDataColumnInformation.NO_KEY)
						return (keyOrder >= 0);
				}
			}
		}

		return super.isIdField(tableClass, columnField, fieldBinding);
	} // end method isIdField


	/**
	 * Connect using JNDI to a datasource name for RDBMS connection.
	 * @param dataSourceName name (as in JNDI) of the datasource
	 * @param properties : InitialContext properties to connect to the JNDI server, may be null if within a J2EE
	 * container with JNDI support (Tomcat/Weblogic/Websphere ...)
	 * @return true if the datasource was found, false otherwise.
	 */
	protected boolean loadDataSource(String dataSourceName, Properties properties)
	{
		try
		{
			if ((properties == null) && (_username != null) && (_username.trim().length() > 0))
			{
				properties = new Properties();
				properties.setProperty(InitialContext.SECURITY_PRINCIPAL, _username);
				properties.setProperty(InitialContext.SECURITY_CREDENTIALS, _password == null ? "" : _password);
			}

			// Get the initial context of the naming service
			Context			namingService;
			String			contextKey = getEnvironment().getEnv(LyEnvironment.NAMING_CONTEXT);

			if (!LyEnvironment.NAMING_CONTEXT.equals(contextKey)
					&& (contextKey != null)
					&& (dataSourceName != null)
					&& !dataSourceName.startsWith("java:"))
				namingService = (Context)new InitialContext(properties).lookup(contextKey);
			else
				namingService = new InitialContext(properties);

			// Build the name of the dataSource.
			if (getEnvironment().isDebugLevel(LyEnvironment.DEBUG_DBMS))
				getEnvironment().logOut("LookUp datasource for name " + dataSourceName+" ...");

			NameParser     nameParser = namingService.getNameParser("");
			Name		   dsName = nameParser.parse(dataSourceName);

			// Lookup the topic factory
			Object		   ds = namingService.lookup(dsName);

			if (ds instanceof DataSource)
				_dataSource = (DataSource)ds;
			else if (ds instanceof Reference)
			{
				// This part of code is a copy from MySQL sample
				// For some reason, this comes back as a Reference
				// instance under CruiseControl !?
				Reference     objAsRef = (Reference)ds;
				Class<?>	  cls = Class.forName(objAsRef.getFactoryClassName());
				ObjectFactory factory = (ObjectFactory)cls.newInstance();

				_dataSource = (DataSource)factory.getObjectInstance(objAsRef, dsName,
						namingService, new Hashtable<Object,Object>());
			}
			if (getEnvironment().isDebugLevel(LyEnvironment.DEBUG_DBMS))
			{
				if (_dataSource != null)
					getEnvironment().logOut("Datasource for name " + dataSourceName+" has been succesfully found");
				else
					getEnvironment().logOut("Datasource for name was not loaded " + dataSourceName+" , found "+ds);
			}
		}
		catch (Exception e)
		{
			LyEnvironment environment = getEnvironment();

			environment.logError("Can't get a valid datasource for name " + dataSourceName);
			environment.writeException(e);

			_dataSource = null;
		}
		return (_dataSource != null);
	}

	/**
	 * Load the JDBC driver for this provider.
	 *
	 * @return true if the driver is successfully loaded, false otherwise.
	 */
	protected boolean loadDriver()
	{
		Class<?>	  driverClass;
		LyEnvironment environment = getEnvironment();

		if (environment.testResource("LY_IGNORE_DRIVER")) // studio may create connection without driver
		{
			return true;
		}

		if (environment.testResource("LY_NO_DRIVER_MANAGER"))
		{
			try
			{
				if (_application.getClassLoader() == null)
				{
					_application.setError(environment.getMessage("LY_ER_DRIVER_NOT_FOUND_NO_DRIVERMANEGER_NO_CLASSLOADER", _driverName));
		    		return false;
				}
				_application.getClassLoader().loadClass(_driverName);
				return true;
			}
	    	catch (ClassNotFoundException e)
	    	{
	    		environment.writeException(e);
	    		_application.setError(environment.getMessage("LY_ER_DRIVER_NOT_FOUND", _driverName));
	    		return false;
	    	}
		}

		try
		{
			driverClass = _application.loadClass(_driverName,
					environment.getMessage("LY_ER_DRIVER_NOT_FOUND", _driverName));

			if (driverClass == null)
			{
				environment.logError(environment.getMessage("LY_ER_REGISTER_DRIVER", _driverName));
				return false;
			}

			DriverManager.registerDriver((Driver)(driverClass.newInstance()));
		}
		catch (IllegalStateException e)
		{
			environment.writeException(e);
			_application.setError(e.getMessage());
			return false;
		}
		catch (SQLException e)
		{
			environment.logError(environment.getMessage("LY_ER_REGISTER_DRIVER", _driverName));
			environment.writeException(e);
			_application.setError(e.getMessage());
			return false;
		}
		catch (Exception e)
		{
			environment.writeException(e);
			_application.setError(e.getMessage());
			return false;
		}
		return true;
	} // end method loadDriver

	/**
	 * This method is a factory to create a LyConnection instance.
	 *
	 * @return a new Connection instance. The connection instance is used in the specific LyRequest
	 *         encoder implementation.
	 * @see    #getConnection
	 * @see    #releaseConnection
	 */
	protected LyConnection createConnection()
	{
		LySqlConnection connection;

		connection = new LySqlConnection(_application, _connectionName, _dataSource);

		// Openning the connection.
		boolean connectionOpened;

		if (getEnvironment().testResource("LY_NO_DRIVER_MANAGER"))
			connectionOpened = connection.open(_username, _password, _driverName);
		else
			connectionOpened = connection.open(_username, _password);

		if (!connectionOpened)
			return null;

		return connection;
	}

	/**
	 * Create a specific request encoder for the specified dataSource.
	 *
	 * @param  dataSource reference to the dataSource associated to this request encoder.
	 * @return the specific LyRequestEncoder for the specified dataSource.
	 */
	public LyRequestEncoder getRequestEncoder(LyDataSource dataSource)
	{
		LySqlRequestEncoder encoder = new LySqlRequestEncoder(this, dataSource, _valueSep);

		encoder.setTrim(_trim);
		return encoder;
	}

	/**
	 * Create a specific join request encoder for the specified dataSource.
	 *
	 * @param  dataSource reference to the dataSource associated to this request encoder.
	 * @return the specific LyRequestEncoder for the specified dataSource.
	 */
	public LyJoinRequestEncoder getJoinRequestEncoder(LyDataSource dataSource)
	{
		return (LyJoinRequestEncoder)getRequestEncoder(dataSource);
	}

	/**
	 * To get the name of the catalog to use with metadata.
	 *
	 * @return the name of the catalog.
	 */
	protected String getCatalogPattern()
	{
		return null;
	}

	/**
	 * To get the name of the schema to use with metadata.
	 *
	 * @return the name of the schema.
	 */
	protected String getSchemaPattern()
	{
		return _databaseName;
	}

	/**
	 * Get metaTable informations.
	 *
	 * @param  metaData for the current database
	 * @return Hashtable corresponding informations for tables
	 */
	protected Hashtable<String, LyMetaDataTableInformation> getTables(DatabaseMetaData metaData)
	{
		return getTables(metaData, null);
	}

	/**
	 * Access data provider, only to get table names
	 *
	 * @param metaData metaData for the current database
	 * @return table names
	 */
	public ArrayList<String> getTableNames()
	{
		// Loading the driver.
		if ((_dataSource == null) && (!loadDriver()))
			return null;

		// Getting a valid connection.
		LySqlConnection  connection = null;
		DatabaseMetaData metaData = null;

		connection = (LySqlConnection)getConnection();
		if (connection == null)
			return null;

		metaData = connection.getMetaData();
		releaseConnection(connection);

		ArrayList<String>	result = new ArrayList<String>();

		if (metaData != null)
		{
			ResultSet					rs = null;
			String[]					types = new String[1];
			String						name;
			String						catalog, schema;
			LyEnvironment				environment = _application.getEnvironment();

			types[0] = "TABLE";

			try
			{
				catalog = getCatalogPattern();
				schema = getSchemaPattern();

				if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
					environment.logOut("DatabaseMetaData getTables(" + catalog
						+ "," + schema + ",null,types[0]=" + types[0] + ")");

				rs = metaData.getTables(catalog, schema, null, types);

				while (rs.next())
				{
					name = rs.getString("TABLE_NAME");
					result.add(name);
				}
			}
			catch (SQLException e)
			{
				environment.writeException(e);
				result = null;
			}
			finally
			{
				if (rs != null)
				{
					try
					{
						rs.close();
					}
					catch (SQLException e)
					{
						environment.writeException(e);
					}
				}
			}
		}
		else
			_application.setError("Can't get database metadatas");

		return result;
	}

	/**
	 * Get metaTable informations.
	 *
	 * @param  metaData for the current database
	 * @return Hashtable corresponding informations for tables
	 */
	protected Hashtable<String, LyMetaDataTableInformation> getTables(DatabaseMetaData metaData, List<String> tableNames)
	{
		ResultSet									  rs = null;
		String[]									  types = new String[1];
		String										  name, remarks;
		LyMetaDataTableInformation					  table;
		Hashtable<String, LyMetaDataTableInformation> result =
			new Hashtable<String, LyMetaDataTableInformation>();
		String										  catalog, schema;
		LyEnvironment								  environment = _application.getEnvironment();

		types[0] = "TABLE";

		try
		{
			catalog = getCatalogPattern();
			schema = getSchemaPattern();

			if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
				environment.logOut("DatabaseMetaData getTables(" + catalog
					+ "," + schema + ",null,types[0]=" + types[0] + ")");

			rs = metaData.getTables(catalog, schema, null, types);

			while (rs.next())
			{
				name = rs.getString("TABLE_NAME");

				if(!hasName(name, tableNames))
					continue;

				// Ignore tables names for old or delete tables
				if (name.indexOf('$') >= 0)
					continue;

				try
				{
					remarks = rs.getString("REMARKS");
				}
				catch (SQLException ex)
				{
					remarks = null;
				}

				if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
					environment.logOut("DatabaseMetaData getTables Result(TABLE_NAME="
						+ name + ",REMARKS=" + remarks + ")");

				table = new LyMetaDataTableInformation(name, remarks);
				result.put(name, table);
			}
		}
		catch (SQLException e)
		{
			environment.writeException(e);
			result = null;
		}
		finally
		{
			if (rs != null)
			{
				try
				{
					rs.close();
				}
				catch (SQLException e)
				{
					environment.writeException(e);
				}
			}
		}

		return result;
	} // end method getTables

	/**
	 * We test if the table list name contains the name or part of the name
	 * followed or previoused by *
	 * @param name			the name to test
	 * @param tableNames	the table names
	 * @return
	 */
	protected boolean hasName(String name, List<String> tableNames) {
		if(tableNames == null)
			return true;

		boolean matches = false;
		int i = 0;

		while(!matches && i <tableNames.size())
		{
			String tn = tableNames.get(i).replace("*", ".*");
			matches = name.toLowerCase().matches(tn.toLowerCase());
			i++;
		}

		return matches;
	}

	/**
	 * Complete tables meta-informations.
	 *
	 * @param  metaData for the current database.
	 * @param  tables   tables of the current database.
	 * @return boolean if the operation is succeded
	 */
	protected boolean completeTables(DatabaseMetaData metaData,
        Hashtable<String, LyMetaDataTableInformation> tables)
	{
		ResultSet				    rs = null;
		String					    str;
		Enumeration<LyMetaDataTableInformation>	elements = tables.elements();
		LyMetaDataTableInformation  table;
		LyMetaDataColumnInformation column;
		short					    keyseq;
		boolean					    hasPrimaryKey;
		String					    tableName;
		short					    dataType;
		String					    typeName;
		String					    name;
		int						    columnSize;
		int						    decimalDigits;
		int						    nullable;
		int						    octetLength;
		int						    position;
		String					    remarks;
		String					    isNullable;
		String					    def;
		String					    catalog, schema;
		boolean					    result = true;
		LyEnvironment			    environment = _application.getEnvironment();

		try
		{
			while (elements.hasMoreElements())
			{
				catalog = getCatalogPattern();
				schema = getSchemaPattern();
				table = (LyMetaDataTableInformation)elements.nextElement();
				str = table.getName();
				if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
					environment.logOut("Table " + str + " - columns : ");

				// Get the columns of the table.
				if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
					environment.logOut("DatabaseMetaData getColumns(" + catalog
						+ "," + schema + "," + str + ",null)");

				rs = metaData.getColumns(catalog, schema, str, null);

				while (rs.next())
				{
					tableName = rs.getString("TABLE_NAME");//.toLowerCase();
					dataType = rs.getShort("DATA_TYPE");
					typeName = rs.getString("TYPE_NAME");
					name = rs.getString("COLUMN_NAME");//.toLowerCase();
					if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
						environment.logOut("\t" + name + "\t" + dataType + "\t"
							+ typeName);

					columnSize = rs.getInt("COLUMN_SIZE");
					decimalDigits = rs.getInt("DECIMAL_DIGITS");
					nullable = rs.getInt("NULLABLE");
					remarks = rs.getString("REMARKS");
					isNullable = rs.getString("IS_NULLABLE");
					try
					{
						def = rs.getString("COLUMN_DEF");
					}
					catch (SQLException ex)
					{
						def = null;
					}

					octetLength = rs.getInt("CHAR_OCTET_LENGTH");
					position = rs.getInt("ORDINAL_POSITION");

					column = new LyMetaDataColumnInformation(name, remarks);
					column.setTableName(tableName);
					column.setDataType(dataType);
					column.setTypeName(typeName);
					column.setColumnSize(columnSize);
					column.setDecimalDigits(decimalDigits);
					column.setNullable(nullable);
					column.setIsNullable(isNullable);
					column.setDef(def);
					column.setOctetLength(octetLength);
					column.setPosition(position);
					table.addColumn(column);
				}

				rs.close();
				rs = null;

				// Get the primary keys of the table
				hasPrimaryKey = false;
				List<Object[]> primaryKeys = getPrimaryKeys(metaData, catalog, schema, str);
				if ((primaryKeys != null)
						&& !primaryKeys.isEmpty())
				{
					hasPrimaryKey = true;
					for (Object[] key : primaryKeys)
					{
						if (key[0] != null)
							name = (String)key[0];
						else
							name = null;
						if (key[1] != null)
							keyseq = (Short)key[1];
						else
							keyseq = 0;
						column = table.getColumn(name);
						if (column != null)
						{
							boolean isIdentity = false;
							typeName = column.getTypeName();
							remarks = column.getRemarks();

							if(typeName != null)
								isIdentity = typeName.contains("identity");
							if(remarks != null && !isIdentity)
								isIdentity = remarks.contains("auto_increment");

							column.setKeyOrder(keyseq);
							column.setIsPrimaryKey(isIdentity);
						}
					}
				}

				// Foreign Keys of the table
				try
				{
					rs = metaData.getImportedKeys(catalog, schema, str);
					if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
						environment.logOut("\t foreing Keys : ");

					while (rs.next())
					{
						String	fkName = rs.getString("FK_NAME");
						String	pkColumnName = rs.getString("PKCOLUMN_NAME");
						String	pkTableName = rs.getString("PKTABLE_NAME");

						if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
							environment.logOut("\t\t"
								+ rs.getString("FKCOLUMN_NAME"));
						column = table.getColumn(rs.getString("FKCOLUMN_NAME"));
						if (column != null)
							column.setIsImportedKey(fkName, pkTableName, pkColumnName);
					}
				}
				catch (Exception ex)
				{
					if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
						environment.writeException(ex);
				}
				finally
				{
					if (rs != null)
					{
						rs.close();
						rs = null;
					}
				}

				if (!hasPrimaryKey)
				{
					// The information is not available.
					Enumeration<LyMetaDataColumnInformation> columns = table.getColumns();

					while (columns.hasMoreElements())
					{
						column = columns.nextElement();
						column.setKeyOrder(LyMetaDataColumnInformation.NO_KEY);
					}
				}
			}
		}
		catch (SQLException e)
		{
			environment.writeException(e);
			result = false;
		}
		finally
		{
			if (rs != null)
			{
				try
				{
					rs.close();
				}
				catch (SQLException e)
				{
					environment.writeException(e);
				}
			}
		}

		return result;
	} // end method completeTables

	protected List<Object[]> getPrimaryKeys(DatabaseMetaData metaData, String catalog, String schema, String tableName)
	{
		LyEnvironment			    environment = _application.getEnvironment();
		String name;
		Short keyseq;
		ResultSet rs = null;
		Object[] result;
		List<Object[]> resultList = null;

		try
		{
			rs = metaData.getPrimaryKeys(catalog, schema, tableName);
			resultList = new ArrayList<Object[]>();
			if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
				environment.logOut("\t Primary Keys : ");

			while (rs.next())
			{
				name = rs.getString("COLUMN_NAME");
				if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
					environment.logOut("\t\t" + name);
				keyseq = rs.getShort("KEY_SEQ");
				result = new Object[2];
				result[0] = name;
				result[1] = keyseq;
				resultList.add(result);
			}
		}
		catch (Exception ex)
		{
			if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
				environment.writeException(ex);
		}
		finally
		{
			if (rs != null)
			{
				try
				{
					rs.close();
				}
				catch (SQLException e)
				{
				}
				rs = null;
			}
		}
		return resultList;
	}

	/**
	 * This method is use for get LyMetadataInformations.
	 *
	 * @return Object LyMetaDataInformation of this database
	 */
	public Object writeMetaData()
	{
		return writeMetaData(null);
	}

	/**
	 * This method is use for get LyMetadataInformations.
	 * @param tableNames	Names of the tables to discover
	 *
	 * @return Object LyMetaDataInformation of this database
	 */
	public Object writeMetaData(List<String> tableNames)
	{
		// Loading the driver.
		if ((_dataSource == null) && (!loadDriver()))
			return null;

		// Getting a valid connection.
		LySqlConnection  connection = null;
		DatabaseMetaData metaData = null;

		connection = (LySqlConnection)getConnection();
		if (connection == null)
			return null;

		metaData = connection.getMetaData();
		releaseConnection(connection);

		// Extracting metadatas.
		LyMetaDataInformation metaInformation = null;

		if (metaData != null)
		{
			// create the associate MetaDataInformation
			Hashtable<String, LyMetaDataTableInformation> result = getTables(metaData, tableNames);

			if (result == null)
			{
				_application.setError("Can't get tables informations");
				return null;
			}

			// complete tables
			if (completeTables(metaData, result))
			{
				LyEnvironment environment = _application.getEnvironment();

				metaInformation = new LyMetaDataInformation(metaData, _databaseName);
				if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
					environment.logOut("DatabaseMetaData create LyMetaDataInformation:"
						+ metaInformation.toString());

				metaInformation.setTables(result);
			}
			else
				_application.setError("Can't get columns informations");
		}
		else
			_application.setError("Can't get database metadatas");

		return metaInformation;
	} // end method writeMetaData

	/**
	 * Register a specific encoder for the given field.
	 *
	 * @param field   The field for which the encoder is registered.
	 * @param encoder The specific encoder to register.
	 */
	public void addSpecificEncoder(LyField field, LySqlSpecificEncoder encoder)
	{
		if (_specificEncoders == null)
			_specificEncoders = new HashMap<LyField, LySqlSpecificEncoder>(1);

		if (_specificEncoders.containsKey(field))
		{
			LyEnvironment environment = getEnvironment();

			environment.logWarning("Replacing SQL specific encoder for field "
				+ field.getId());
		}

		_specificEncoders.put(field, encoder);
	}

	/**
	 * To get the specific encoder associated to the specified field.
	 *
	 * @param  field the field
	 * @return the specific encoder associated to the field if exists, null otherwise.
	 */
	protected LySqlSpecificEncoder getSpecificEncoder(LyField field)
	{
		if (_specificEncoders == null)
			return null;
		else
			return _specificEncoders.get(field);
	}

	/**
	 * Tells if the current SQL data provider refers to a database that supports the boolean type.
	 *
	 * @return false
	 */
	protected boolean booleanTypeSupported()
	{
		return true;
	}

	/**
	 * Tests if this data provider supports the XA protocol or not. A SQL provider supports the XA
	 * protocol if a JDBC DataSource is defined for this provider ad if this DataSource is a
	 * XADataSource.
	 *
	 * @return true if this provider supports the XA protocol, false otherwise.
	 */
	public boolean isXASupported()
	{
		return ((_dataSource != null) && (_dataSource instanceof XADataSource));
	}
} // end class LySqlDataProvider
