/*
 * LySqlRequestEncoder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.sql;

import java.util.*;
import leon.data.*;
import leon.info.*;


/**
 * This class is used to keep the encoding context of a request when going into multiple encoders.
 *
 * @see LySqlRequestEncoder
 */
public class LySqlRequestInfo
{
	public static final String __VERSION =
		"$Id: LySqlRequestInfo.java 41657 2012-11-29 09:01:07Z bangiole $";

	/**
	 * Join type to merge at parent level the new aliases only. Table names used only once do not
	 * create an alias for parent level. This type is used for different clauses for instance in
	 * (xxx) OR (yyy) where the table names in xxx may be the same as in yyy
	 */
	public enum Join
    {
	    NEW_ALIAS_ONLY,
        SUM_ALIAS_COUNTS,
        NOTHING
    }

	protected static int										   _tableAliasCounter = 0;

	/** Current classes being used by the request. */
	protected HashMap<Thread, ArrayList<String>>				   _usedClassesList;

	/** Table of tables aliases used for request generation. */
	protected Map<_LySqlRequestInfoKey, Hashtable<String, String>> _tableAliases;

	/** The count of the table aliases (number of alias per table). */
	protected Map<String, Short>								   _tableAliasesCount;

	/** Unsused Aliases to skip */
	protected List<String>										   _skipUnusedAliases;

	/**
	 * Indicates if we need to write the table names before the fields of the request. Mandatory if
	 * two class have the same field name or if a table is using aliases
	 */
	protected boolean											   _writeTableNames;

	/** Parent request info. */
	protected LySqlRequestInfo									   _parent;

	/**
	 * children request info : we have a new request info when we switch from one
	 * LySqlRequestEncoder to another.
	 */
	protected ArrayList<LySqlRequestInfo>						   _children;

	/** Object class managed by the client data source. */
	protected LyClass											   _class;

	/** Type of child, one of the CHILD_xxx types. */
	protected Join												   _childType;

	/**
	 * This mode applies when an union select validation is requested Internal use only for SQL
	 * Optimization.
	 */
	protected boolean											   _detectingUnionSelect;

	/**
	 * This mode applies when an union select validation is requested Internal use only for SQL
	 * Optimization.
	 */
	protected ArrayList<Object[]>								   _unionSelectInfos;

	/**
	 * Internal search context for alias. Searching ab alias for a table is based on a couple
	 * (classInfo, fieldInfo) both optional. This parameter stores 'default' search parameters
	 */
	protected _LySqlRequestInfoKey								   _aliasContext;

	/**
	 * When an analysis has been done in first step, second step is not adding new aliases counts,
	 * keep current from parent.
	 */
	protected boolean											   _lockAliasCount;

	/** Debug information for request encoding. */
	protected LySqlRequestInfoDebugTask							   _currentTask;

	/** Couples of tablename / aliases to ignore in select from. */
	protected HashMap<String, Boolean>							   _ignoreInSelectFrom;

	 /** Force using alias an recursive contextual search*/
	protected boolean _forceUseAlias; // FFT00557 W4

	// Indicates if the request is using formatted date in the order by clause : in case of a distinct select we may need to add the formatted content also
	protected Map<String, String> _mapSQLFormatDateInOrderByClause = null;



	/**
	 * Constructor.
	 *
	 * @param provider
	 * @param dataSource
	 * @param valueSep
	 */
	public LySqlRequestInfo(LyClass c)
	{
		_class = c; // just for info
		_lockAliasCount = false;
		_forceUseAlias = false;
	}


	/**
	 * sets the Parent request info.
	 *
	 * @param parent    new parent, if null this object is removed from the parent children list
	 * @param childType the type of child, one of the TYPE_xxx constants. parameter is ignored when
	 *                  parent is null.
	 */
	public void setParent(LySqlRequestInfo parent, Join childType)
	{
		if (parent != null)
		{
			_parent = parent;
			_parent.addChild(this, childType);
			if ((_parent._currentTask != null) && (_currentTask == null))
			{
				_currentTask = new LySqlRequestInfoDebugTask();
				_parent._currentTask.addTask(_currentTask);
				_parent._currentTask.addInfo("--> add child task " + _currentTask.getCounter()
					+ " type = " + childType);
			}

		}
		else if (_parent != null)
		{
			if ((_children != null) && (_children.size() > 0))
				throw new IllegalStateException(
					"[LySqlRequestInfo]Should not remove a request info with sons :\n" + toString()
					+ "* parent :\n" + _parent.toString());

			if ((_parent._currentTask != null) && (_currentTask != null))
			{
				_parent._currentTask.addInfo("--> remove child task " + _currentTask.getCounter()
					+ " type = " + childType);
				_parent._currentTask.addInfo("[INFO child] " + this.toString());
			}
			_parent.removeChild(this);
			if (_parent._currentTask != null)
				_parent._currentTask.addInfo("[INFO parent] " + _parent.toString());
		}
	} // end method setParent

	/**
	 * Access to the root parent request info of this request info
	 * @return the root parent request info or null if this is a root context
	 */
	public LySqlRequestInfo getRootParent()
	{
		if (_parent == null)
			return this;

		return _parent.getRootParent();
	}

	/**
	 * Access to the parent info of this request info
	 * @return the parent request info or null if this is a root context
	 */
	public LySqlRequestInfo getParent()
	{
		return _parent;
	}

	/**
	 * Internal use (use setParent to define) add a child in children's list.
	 *
	 * @param child     new child
	 * @param childType the type of child, one of the TYPE_xxx constants
	 */
	protected void addChild(LySqlRequestInfo child, Join childType)
	{
		if (_children == null)
			_children = new ArrayList<LySqlRequestInfo>();
		_children.add(child);
		child._childType = childType;
		// if locked, transfer from PARENT to child
		if (_lockAliasCount && (childType != Join.NOTHING))
		{
			child.reportAliasCount(this, Join.NEW_ALIAS_ONLY);
			child.setLockAliasCount(_lockAliasCount);
			if (getForceUseAlias())
				child.setForceUseAlias(true); // FFT02391

		}
	}

	/**
	 * Internal use (use setParent to define) remove a child from children's list.
	 *
	 * @param child child to remove
	 */
	protected void removeChild(LySqlRequestInfo child)
	{
		if (isDetectingUnionSelect())
			return; // detection UNION SELECT, keep the children tree structure (until isDetectingUnionSelect turns back to false)

		if (_children != null)
		{
			if (_children.remove(child))
			{
				if (!_lockAliasCount)
					reportAliasCount(child, child.getChildType());
				reportUsedClasses(child);
			}

			if (_children.size() == 0)
				_children = null;
		}
		// *** When a child is removed we may need to report on parent level the aliases and usage count of table names
	}

	/**
	 * Store a table alias for a given logical class and a table name.
	 *
	 * @param classInfo target logical class (optional)
	 * @param classInfo target logical field for context (optional) // FFT FL 438 , when multiple
	 *                  relation to same table name within the classInfo. classInfoParameter must
	 *                  not be null
	 * @param classInfo target logical class
	 * @param tableName table name
	 * @param alias     table alias
	 * @param counted   indicates if this alias is COUNTED in the used alias or not. Excepted for
	 *                  internal use the counted status should be true.
	 * @param isTargetRelationClass used to distinguish class and target class on the case of a self (class
	 * and target class identical) pointing relation.
	 */
	protected void addTableAlias(String domain, LyClassInfo classInfo, LyFieldInfo fieldInfo,
		String tableName, String alias,
		boolean counted, boolean isTargetRelationClass)
	{
		if ((alias == null) || (tableName == null) || alias.equals(tableName))
			return;

		if (_lockAliasCount)
			return;
		// use context if any specified as default values
		if (_aliasContext != null)
		{
			if (domain == null)
				domain = _aliasContext._domain;
			if (classInfo == null)
				classInfo = _aliasContext._classInfo;
			if ((fieldInfo == null) && (classInfo == _aliasContext._classInfo))
				fieldInfo = _aliasContext._fieldInfo;
		}
		if (_tableAliases == null)
		{
			_tableAliases = Collections.synchronizedMap(
					new HashMap<_LySqlRequestInfoKey, Hashtable<String, String>>(3));
			_tableAliasesCount = Collections.synchronizedMap(new HashMap<String, Short>(5));
		}

		_LySqlRequestInfoKey key = new _LySqlRequestInfoKey(domain, classInfo, fieldInfo, isTargetRelationClass);

		getCurrentTask().addInfo("addTableAlias : tableName=" + tableName + " alias=" + alias
			+ " key =" + key);

		Hashtable<String, String> aliases = _tableAliases.get(key);

		if (aliases == null)
		{
			aliases = new Hashtable<String, String>(3);
			_tableAliases.put(key, aliases);
		}

		// store the alias
		aliases.put(tableName, alias);

		// update the count of aliases
		if (counted && !_lockAliasCount)
		{
			Short count = _tableAliasesCount.get(tableName);

			if (count == null)
				count = new Short((short)0);
			_tableAliasesCount.put(tableName, new Short((short)(count.shortValue() + 1)));
		}
	} // end method addTableAlias

	/**
	 * Search for the table alias of a given table for a given logical class.
	 *
	 * @param  classInfo target logical class
	 * @param  fieldInfo optioanl parameter to search the alias name in the context of this
	 *                   fieldInfo
	 * @param  tableName table name
	 * @return the table alias if found, the table name otherwise.
	 */
	protected String getTableAlias(LyClassInfo classInfo, LyFieldInfo fieldInfo, String tableName)
	{
		return getTableAlias(null, classInfo, fieldInfo, tableName, false, false);
	}

	/**
	 * Search for the table alias of a given table for a given logical class.
	 *
	 * @param  classInfo        target logical class
	 * @param  fieldInfo        optioanl parameter to search the alias name in the context of this
	 *                          fieldInfo
	 * @param  tableName        table name
	 * @param  returnIfUsedOnce return alias names which have been used only once (not yet sure to
	 *                          use it at the end)
   	 * @param selfTargetRelationClass used to distinguish class and target class on the case of a self (class
	 * and target class identical) pointing relation.
	 * @return the table alias if found, the table name otherwise.
	 */
	protected String getTableAlias(String domain, LyClassInfo classInfo, LyFieldInfo fieldInfo,
		String tableName, boolean returnIfUsedOnce, boolean selfTargetRelationClass)
	{
		//isTargetRelationClass = false;
		String resu = null;

		if ((resu == null) && (_tableAliases == null))
			resu = tableName;

		if (_tableAliases == null)
			return tableName;

		// use context if any specified as default values
		if (_aliasContext != null)
		{
			if (classInfo == null)
				classInfo = _aliasContext._classInfo;
			if ((fieldInfo == null) && (classInfo == _aliasContext._classInfo))
				fieldInfo = _aliasContext._fieldInfo;
			domain = _aliasContext._domain;
			if (_aliasContext._selfTargetRelationClass)
				selfTargetRelationClass = _aliasContext._selfTargetRelationClass;
		}

		// get the list of aliases for the given class info
		Hashtable<String, String> aliases = null;

		if (resu == null)
		{
			_LySqlRequestInfoKey key = new _LySqlRequestInfoKey(domain, classInfo, fieldInfo, selfTargetRelationClass);

			aliases = _tableAliases.get(key);

			if ((aliases == null) && (_lockAliasCount) && (classInfo != null))
			{
				// default alias ?
				key = new _LySqlRequestInfoKey(domain, null, null, selfTargetRelationClass);
				aliases = _tableAliases.get(key);
			}

			if (aliases == null)
				resu = tableName;
		}

		if ((resu == null) && (!getForceUseAlias()))
		{
			// if there is only one alias, return the name of the table
			Short count = _tableAliasesCount.get(tableName);

			if ((count == null) || (count.shortValue() <= 1))
				resu = tableName;
		}

		// get the alias
		if (resu == null)
		{
			String alias = aliases.get(tableName);

			if (alias == null)
			{
				if (getForceUseAlias())
				{
					_LySqlRequestInfoKey previousAliasContext = _aliasContext;// recurse without context to avoid stackoverflow
					_aliasContext = null; // FFT002455

					if (fieldInfo != null)
						resu =  getTableAlias(domain, classInfo, null, tableName, returnIfUsedOnce, false);
					else if (classInfo != null)
						resu =  getTableAlias(domain, null, null, tableName, returnIfUsedOnce, false);
					else if (domain != null)
						resu =  getTableAlias(null, null, null, tableName, returnIfUsedOnce, false);
					else
						resu= tableName;
					_aliasContext = previousAliasContext;// restore
				}
				else
					resu = tableName;
			}
			else
				resu = alias;
		}

		return resu;
	} // end method getTableAlias

	/**
	 * Search for the all the (aliases --> table) couples.
	 *
	 * @return the table alias map if found, an empty map otherwise.
	 */
	protected HashMap<String, String> getAllAliases()
	{
		HashMap<String, String> aliasMap = new HashMap<String, String>();

		if ((_tableAliasesCount == null) || (_tableAliases == null))
			return aliasMap;

		Iterator<_LySqlRequestInfoKey> keys = _tableAliases.keySet().iterator();
		Iterator<String>			   tableNameKeys;
		String						   tableName;
		Short						   count;
		Hashtable<String, String>	   aliases;
		_LySqlRequestInfoKey		   key;

		while (keys.hasNext())
		{
			key = keys.next();
			aliases = _tableAliases.get(key);
			tableNameKeys = aliases.keySet().iterator();

			while (tableNameKeys.hasNext())
			{
				tableName = tableNameKeys.next();
				count = _tableAliasesCount.get(tableName);

				if ((count != null) && (count.shortValue() > 1) && (aliases.get(tableName) != null))
				{
					String tableAlias = aliases.get(tableName);

					if (_skipUnusedAliases == null || !_skipUnusedAliases.contains(tableAlias))
						aliasMap.put(tableAlias, tableName);
				}
			}
		}
		return aliasMap;
	} // end method getAllAliases

	/**
	 * Search for the all the (aliases --> table) couples.
	 *
	 * @return the table alias map if found, an empty map otherwise.
	 */
	protected void skipUnusedAliases(String tableAlias)
	{
		if (tableAlias != null)
		{
			if (_skipUnusedAliases == null)
				_skipUnusedAliases = new ArrayList<String>();

			_skipUnusedAliases.add(tableAlias);
		}
	} // end method getAllAliases
	/**
	 * Retrieves the class of data handled by this data source.
	 *
	 * @return Class of the data managed in this data source.
	 */
	public final LyClass getDataClass()
	{
		return _class;
	}

	/**
	 * Overriden to display an understandable information for debug.
	 *
	 * @return a human readable details of current info and children
	 */
	public String toString()
	{
		StringBuffer     sb = new StringBuffer();
		StringBuffer     sbUnusedAliases = new StringBuffer();
		StringBuffer     tabs = new StringBuffer();
		LySqlRequestInfo parent = getParent();

		while (parent != null)
		{
			tabs.append("\t");
			parent = parent.getParent();
		}

		sb.append(tabs);
		sb.append("CLASS=").append(_class.getName());
		sb.append(" ALIASES={");

		sbUnusedAliases.append(" NOT_USED_ALIASES={");

		if (_tableAliases != null)
		{
			Iterator<_LySqlRequestInfoKey> keys = _tableAliases.keySet().iterator();
			Iterator<String>			   tableNameKeys;
			String						   tableName;
			Short						   count;
			boolean						   found = false, foundUsed = false, foundUnused = false;
			Hashtable<String, String>	   aliases;
			_LySqlRequestInfoKey		   key;

			while (keys.hasNext())
			{
				key = keys.next();
				aliases = _tableAliases.get(key);
				tableNameKeys = aliases.keySet().iterator();

				while (tableNameKeys.hasNext())
				{
					tableName = tableNameKeys.next();
					count = _tableAliasesCount.get(tableName);
					StringBuffer buf = null;

					if ((count != null) && (count.shortValue() > 1))
					{
						buf = sb;
						found = foundUsed;
						foundUsed = true;
					}
					else
					{
						buf = sbUnusedAliases;
						found = foundUnused;
						foundUnused = true;
					}

					if (found)
						buf.append(", ");
					buf.append(tableName);
					if (key._domain != null)
						buf.append('(').append(key._domain).append("#)");
					if (key._classInfo != null)
						buf.append('(').append(key._classInfo.getId()).append(")");
					if (key._fieldInfo != null)
						buf.append(".(").append(key._fieldInfo.getId()).append(")");
					buf.append("->").append(aliases.get(tableName)).append('(').append(count)
						.append(')');
				}
			}
		}
		sb.append("}");
		sbUnusedAliases.append("}");
		sb.append(" USED_CLASSES={");
		String    name;
		ArrayList<String> nameList = getCurrentUsedClasses();

		for (int i = 0; i < nameList.size(); i++)
		{
			name = nameList.get(i);
			if (i > 0)
				sb.append(", ");
			sb.append(name);
		}
		sb.append("}");
		if (_children != null)
		{
			for (int i = 0; i < _children.size(); i++)
				sb.append('\n').append(_children.get(i).toString());
		}
		sb.append(' ').append(sbUnusedAliases);

		return sb.toString();
	} // end method toString

	/**
	 * Report to current level the child aliases and counts from child level This transfer of
	 * information depends on the 'type' of child info we have.
	 *
	 * @param child    the targer request info to merge
	 * @param typejoin one of the TYPE_JOIN_XXXX (constants), usually child.getChildType()
	 */
	protected void reportAliasCount(LySqlRequestInfo child, Join join)
	{
		if (child == null)
			return;
		if (join == Join.NEW_ALIAS_ONLY)
		{
			if (child._tableAliases != null)
			{
				Iterator<_LySqlRequestInfoKey> keys = child._tableAliases.keySet().iterator();
				Iterator<String>			   tableNameKeys;
				String						   tableName, alias;
				Short						   count, childCount;
				Hashtable<String, String>	   aliases, childAliases = null;
				_LySqlRequestInfoKey		   key;

				while (keys.hasNext())
				{
					key = keys.next();
					aliases = ((_tableAliases == null) ? null : _tableAliases.get(key));

					if (child._tableAliases != null)
						childAliases = child._tableAliases.get(key);
					if (aliases == null)
						aliases = childAliases;
					else if (childAliases != null)
						aliases.putAll(childAliases);

					tableNameKeys = aliases.keySet().iterator();
					while (tableNameKeys.hasNext())
					{
						tableName = tableNameKeys.next();
						count = null;
						childCount = null;
						if (_tableAliasesCount != null)
							count = _tableAliasesCount.get(tableName);
						if (count == null)
							count = new Short((short)0);
						if (child._tableAliasesCount != null)
							childCount = child._tableAliasesCount.get(tableName);
						if (childCount == null)
							childCount = 0;
						if ((count > 1) || (childCount > 1)
								|| (count == 1 && getForceUseAlias())  // FFT02391
								|| (childCount == 1 && child.getForceUseAlias())) // FFT02391
						{
							// ** a 'real' alias is found in the child and used by the child : report it at parent level
							alias = aliases.get(tableName);
							addTableAlias(key._domain, key._classInfo, key._fieldInfo, tableName,
								alias, true, key._selfTargetRelationClass);
						}
					}
				}
			}
		}
		else if (join == Join.SUM_ALIAS_COUNTS)
		{
			if (child._tableAliases != null)
			{
				Iterator<_LySqlRequestInfoKey> keys = child._tableAliases.keySet().iterator();
				Iterator<String>			   tableNameKeys;
				_LySqlRequestInfoKey		   key;
				String						   tableName, alias;
				Map<String, Short>			   tableAliasesCount;
				Short						   count;
				Hashtable<String, String>	   aliases;

				while (keys.hasNext())
				{
					key = keys.next();
					// x aliases   = (Hashtable)(_tableAliases == null ? null : _tableAliases.get(key));

					// x if (aliases == null)
					{
						// unknown from parent : add it if child is using it
						// FL FFT438
						aliases = child._tableAliases.get(key);
						tableAliasesCount = child._tableAliasesCount;
					}
/*                                      else
                                        {
                                                tableAliasesCount = _tableAliasesCount;
                                        }*/
					tableNameKeys = aliases.keySet().iterator();

					while (tableNameKeys.hasNext())
					{
						tableName = tableNameKeys.next();
						count = ((tableAliasesCount == null)
								? new Short((short)0) : (Short)tableAliasesCount.get(tableName));

						if (count != null)
						{
							// ** an usage of the alias is done (real or juste used once) : report it at parent level
							alias = aliases.get(tableName);
							for (short p = 0; p < count.shortValue(); p++)
								addTableAlias(key._domain, key._classInfo, key._fieldInfo,
									tableName, alias, true, key._selfTargetRelationClass);
						}
					}
				}
			}
		}
		else if (join == Join.NOTHING)
		{
			// do nothing
		}
		else
			throw new IllegalStateException("Undefined child type : " + child.getChildType()
				+ " : must be one of the TYPE_xxx constants.");
	} // end method reportAliasCount

	/**
	 * Report to current level the child aliases and counts from child level This transfer of
	 * information depends on the 'type' of child info we have.
	 *
	 * @param child    the targer request info to merge
	 * @param typejoin one of the TYPE_JOIN_XXXX (constants), usually child.getChildType()
	 */
	protected void copyAliasFromDomain(LySqlRequestInfo requestInfo, String sourceDomain,
		String targetDomain)
	{
		if ((requestInfo == null) || (!_lockAliasCount))
			return;

		if (sourceDomain == targetDomain)
			return;
		if ((sourceDomain != null) && sourceDomain.equals(targetDomain))
			return;

		if (requestInfo._tableAliases != null)
		{
			// clone
			Set<_LySqlRequestInfoKey> keySet = new HashSet<_LySqlRequestInfoKey>();

			keySet.addAll(requestInfo._tableAliases.keySet());

			Iterator<_LySqlRequestInfoKey>	keys = keySet.iterator();

			Iterator<String>		  tableNameKeys;
			String					  tableName, alias;
			Short					  count, childCount;
			Hashtable<String, String> aliases, childAliases = null;
			_LySqlRequestInfoKey	  key;

			while (keys.hasNext())
			{
				key = keys.next();
				// From sourceDomain ONLY
				if (key._domain == null)
				{
					if (sourceDomain != null)
						continue;
				}
				else if (!key._domain.equals(sourceDomain))
					continue;

				aliases = ((_tableAliases == null) ? null : _tableAliases.get(key));

				if (requestInfo._tableAliases != null)
					childAliases = requestInfo._tableAliases.get(key);
				if (aliases == null)
					aliases = childAliases;
				else if (childAliases != null)
					aliases.putAll(childAliases);

				tableNameKeys = aliases.keySet().iterator();
				while (tableNameKeys.hasNext())
				{
					tableName = tableNameKeys.next();
					count = null;
					childCount = null;
					if (_tableAliasesCount != null)
						count = _tableAliasesCount.get(tableName);
					if (count == null)
						count = new Short((short)0);
					if (requestInfo._tableAliasesCount != null)
						childCount = requestInfo._tableAliasesCount.get(tableName);
					if (childCount == null)
						childCount = new Short((short)0);
					if ((count.shortValue() > 1) || (childCount.shortValue() > 1))
					{
						// ** a 'real' alias is found in the child and used by the child : report it at parent level
						alias = aliases.get(tableName);
						// do not override an existing alias in target domain
						if (tableName.equals(
									getTableAlias(targetDomain, key._classInfo, key._fieldInfo,
										tableName, false, false)))
						{
							boolean status = _lockAliasCount; // save

							_lockAliasCount = false; // temporary allow creations of new aliases for the copy
							addTableAlias(targetDomain, key._classInfo, key._fieldInfo, tableName,
								alias, true, false);
							_lockAliasCount = status; // restore status
						}
					}
				}
			}
		}
	} // end method copyAliasFromDomain

	/**
	 * Indicates if 'real' aliases are used in this request.
	 *
	 * @return true if at least a 'real' alias is used
	 */
	public boolean isUsingAlias()
	{
		if (_tableAliases == null)
			return false;

		Iterator<_LySqlRequestInfoKey>	keys = _tableAliases.keySet().iterator();
		Iterator<String>   tableNameKeys;
		_LySqlRequestInfoKey key;
		String	    tableName;
		Short	    count;
		Hashtable<String, String>   aliases;

		while (keys.hasNext())
		{
			key = keys.next();
			aliases = _tableAliases.get(key);

			tableNameKeys = aliases.keySet().iterator();
			while (tableNameKeys.hasNext())
			{
				tableName = tableNameKeys.next();
				count = _tableAliasesCount.get(tableName);

				if (count.shortValue() > 1)
					return true;
			}
		}
		return false;
	} // end method isUsingAlias

	/**
	 * Type of child, one of the CHILD_xxx types.
	 */
	public Join getChildType()
	{
		return _childType;
	}

	/**
	 * Report to parent the used table names.
	 */
	public void reportUsedClasses(LySqlRequestInfo child)
	{
		ArrayList<String> childNameList = child.getCurrentUsedClasses();
		String			  name;
		ArrayList<String> resu = getCurrentUsedClasses();

		for (int i = 0; i < childNameList.size(); i++)
		{
			name = childNameList.get(i);
			if (!resu.contains(name))
				resu.add(name);
		}
	}

	/**
	 * Add a name to the Classes names actually used in the where clause.
	 *
	 * @param name the name of the table which is used
	 */
	public void addUsedClass(String name)
	{
		if (name == null)
			return;

		ArrayList<String> resu = getCurrentUsedClasses();

		if (!resu.contains(name))
			resu.add(name);
	}

	/**
	 * Classes names actually used in the where clause.
	 *
	 * @return the list of table names used in the where clause
	 */
	public ArrayList<String> getCurrentUsedClasses()
	{
		if (_usedClassesList == null)
			_usedClassesList = new HashMap<Thread, ArrayList<String>>();

		ArrayList<String> resu = _usedClassesList.get(Thread.currentThread());

		if (resu == null)
		{
			resu = new ArrayList<String>();
			_usedClassesList.put(Thread.currentThread(), resu);
		}

		return resu;
	}

	/**
	 * Reset Classes names actually used in the where clause.
	 */
	public void clearCurrentUsedClasses()
	{
		if (_usedClassesList == null)
			return;

		_usedClassesList.remove(Thread.currentThread());

		for (int i = 0; (_children != null) && (i < _children.size()); i++)
			_children.get(i).clearCurrentUsedClasses();
	}

	public void removeUsedClass(String className)
	{
		if (_usedClassesList == null)
			return;

		ArrayList<String> strings = _usedClassesList.get(Thread.currentThread());

		if (strings != null)
			strings.remove(className);

		for (int i = 0; (_children != null) && (i < _children.size()); i++)
			_children.get(i).removeUsedClass(className);
	}
	/**
	 * This mode applies when an union select validation is requested Internal use only for SQL
	 * Optimization.
	 */
	public boolean isDetectingUnionSelect()
	{
		return _detectingUnionSelect;
	}

	/**
	 * This mode applies when an union select validation is requested Internal use only for SQL
	 * Optimization.
	 *
	 * @param true to activate detectiuon mode. Applicationshall never invoke this method directly
	 */
	public void setDetectingUnionSelect(boolean unionSelect)
	{
		if (_detectingUnionSelect && (!unionSelect))
		{
			_detectingUnionSelect = unionSelect;
			completeDetectingUnionSelect();
		}
		_detectingUnionSelect = unionSelect;
	}

	/**
	 * Internal use only : checks wether an union select is pertinent or not.
	 */
	public void completeDetectingUnionSelect()
	{
		// For each potential select union, find the used classes they need and that this parent info do not know yet
		boolean			  useUnionSelect = false;
		int				  currentChild = 0, lastChild, maxChildCount = 0;
		ArrayList<String> parentUsedClasses = getCurrentUsedClasses();
		ArrayList<String> childrenUsedClasses;
		ArrayList<String> childUsedClasses;
		ArrayList<String> currentUsedClasses;

		childrenUsedClasses = new ArrayList<String>();

		for (int i = 0; (_unionSelectInfos != null) && (i < _unionSelectInfos.size()); i++)
		{
			Object[] info = getUnionSelectInfos(i);

			lastChild = ((Integer)info[2]).intValue();

			// for each group of child request info : find the used class by this child only
			childUsedClasses = new ArrayList<String>();

			while (currentChild < lastChild)
			{
				LySqlRequestInfo child = _children.get(currentChild);

				currentUsedClasses = child.getCurrentUsedClasses();

				for (int j = 0; (currentUsedClasses != null) && (j < currentUsedClasses.size());
						j++)
				{
					String usedClass = currentUsedClasses.get(j);

					if (parentUsedClasses.contains(usedClass))
						; // ignore
					else if (childUsedClasses.contains(usedClass))
						; // already have it
					else
					{
						childUsedClasses.add(usedClass);
						if (!childrenUsedClasses.contains(usedClass))
							childrenUsedClasses.add(usedClass);
						if (childUsedClasses.size() > maxChildCount)
							maxChildCount = childUsedClasses.size();
					}
				}
				currentChild++;
			}
			info[0] = childUsedClasses; // store for future use
		}
		// when shall we try an union select ?
		if ((maxChildCount >= 2) && (childrenUsedClasses.size() >= 3))
			useUnionSelect = true;

		if (useUnionSelect)
			_children = null; // Do not commit children, use multiple select then union
		else
		{
			// The test conclusion is : DO NOT USE UNION SELECT. Commit all children
			while ((_children != null) && (_children.size() > 0))
			{
				LySqlRequestInfo child = _children.get(0);

				removeChild(child);
			}
			_unionSelectInfos = null; // Forget.
		}
	} // end method completeDetectingUnionSelect

	/**
	 * This mode applies when an union select validation is requested Internal use only for SQL
	 * Optimization.
	 *
	 * @return in [0] the 'additional' used classes fot the from request, in [1] the 'request
	 *         criteria' and in [2] the index of the child.
	 */
	public Object[] getUnionSelectInfos(int i)
	{
		if ((_unionSelectInfos == null) || (_unionSelectInfos.size() <= i))
			return null;

		return _unionSelectInfos.get(i);
	}

	/**
	 * This mode applies when an union select validation is requested Internal use only for SQL
	 * Optimization.
	 *
	 * @return in [0] the 'additional' used classes for the from request, in [1] the 'request
	 *         criteria' and in [2] the index of the child.
	 */
	public void addUnionSelectInfos(String filter)
	{
		if (!_lockAliasCount)
			return; // only in real mode, not analysis
		if ((_children == null) || (_children.size() == 0))
			return;

		if (_unionSelectInfos == null)
			_unionSelectInfos = new ArrayList<Object[]>();
		_unionSelectInfos.add(new Object[] { null, filter, new Integer(_children.size()) });
	}
	/**
	 * Internal search context for alias. Searching ab alias for a table is based on a couple
	 * (classInfo, fieldInfo) both optional. This parameter stores 'default' search parameters set
	 * to 'null' to clear the context
	 *
	 * @param classInfo target logical class
	 * @param fieldInfo optional parameter to search the alias name in the context of this fieldInfo
	 */
	protected void setAliasContext(String domain, LyClassInfo classInfo, LyFieldInfo fieldInfo)
	{
		setAliasContext(domain, classInfo, fieldInfo, false);
	}
	/**
	 * Internal search context for alias. Searching ab alias for a table is based on a couple
	 * (classInfo, fieldInfo) both optional. This parameter stores 'default' search parameters set
	 * to 'null' to clear the context
	 *
	 * @param classInfo target logical class
	 * @param fieldInfo optional parameter to search the alias name in the context of this fieldInfo
	 */
	protected void setAliasContext(String domain, LyClassInfo classInfo, LyFieldInfo fieldInfo, boolean isTargetRelationClass)
	{
		getCurrentTask().addInfo("setAliasContext : domain=" + domain + " classInfo="
			+ ((classInfo == null) ? "null" : classInfo.getId()) + " fieldInfo="
			+ ((fieldInfo == null) ? "null" : fieldInfo.getId()));
		if ((domain == null) && (classInfo == null))
			_aliasContext = null;
		else
			_aliasContext = new _LySqlRequestInfoKey(domain, classInfo, fieldInfo, isTargetRelationClass);
	}

	/**
	 * Internal search context for alias. Searching ab alias for a table is based on a couple
	 * (classInfo, fieldInfo) both optional. This parameter stores 'default' search parameters set
	 * to 'null' to clear the context
	 *
	 * @param classInfo target logical class
	 * @param fieldInfo optioanl parameter to search the alias name in the context of this fieldInfo
	 */
	protected String getDomain()
	{
		String domain = ((_aliasContext == null) ? null : _aliasContext._domain);

		return domain;
	}

	/**
	 * Internal search context for alias. Searching ab alias for a table is based on a couple
	 * (classInfo, fieldInfo) both optional. This parameter stores 'default' search parameters set
	 * to 'null' to clear the context
	 *
	 * @param classInfo target logical class
	 * @param fieldInfo optioanl parameter to search the alias name in the context of this fieldInfo
	 */
	protected void setAliasDomain(String domain)
	{
		if (_aliasContext == null)
			setAliasContext(domain, null, null);
		else
			setAliasContext(domain, _aliasContext._classInfo, _aliasContext._fieldInfo);
	}

	/**
	 * Set locked status for tranfering aliases When an analysis has been done in first step, second
	 * step is not adding new aliases counts, keep current counts and aliases from parent context
	 * and to not report when removing child context.
	 */
	protected void setLockAliasCount(boolean locked)
	{
		_lockAliasCount = locked;
	}

	/**
	 * Internal class stored as a key in the alias table names map : key is the couple classInfo,
	 * fieldInfo and both may be null.
	 *
	 * @author nbo
	 */
	protected class _LySqlRequestInfoKey
	{
		protected String	  _domain;
		protected LyClassInfo _classInfo;
		protected LyFieldInfo _fieldInfo;
		protected boolean     _selfTargetRelationClass;

		protected _LySqlRequestInfoKey(String domain, LyClassInfo classInfo, LyFieldInfo fieldInfo, boolean selfTargetRelationClass)
		{
			_domain = domain;
			_classInfo = classInfo;
			_fieldInfo = fieldInfo;
			_selfTargetRelationClass = selfTargetRelationClass;
		}

		/**
		 * In hashmap the hashCode method must be consistent with the equals method.
		 */
		public int hashCode()
		{
			if (_classInfo == null)
			{
				if (_domain == null)
					return 1;
				return _domain.hashCode();
			}

			int code = _classInfo.hashCode();

			if (_domain != null)
				code = code + _domain.hashCode();

			if (_fieldInfo != null)
				code = code + _fieldInfo.hashCode();
			code = code + (_selfTargetRelationClass ? 1 : 0);
			return code;
		}

		public boolean equals(Object other)
		{
			if (this == other)
				return true;

			// classinfo are null (both) or equals fieldInfo are null (both) or equals otherwise false.
			if (other instanceof _LySqlRequestInfoKey)
			{
				_LySqlRequestInfoKey otherKey = (_LySqlRequestInfoKey)other;

				if ((_domain == null) && (otherKey._domain != null))
					return false;
				if ((_domain != null) && (!_domain.equals(otherKey._domain)))
					return false;
				if (_selfTargetRelationClass != otherKey._selfTargetRelationClass)
					return false;
				if (_classInfo == null)
					return (otherKey._classInfo == null);

				if (_classInfo.equals(otherKey._classInfo))
				{
					if (_fieldInfo == null)
						return (otherKey._fieldInfo == null);
					else
						return (_fieldInfo.equals(otherKey._fieldInfo));
				}
			}
			return false;
		}

		public String toString()
		{
			StringBuffer sb = new StringBuffer();

			sb.append("_LySqlRequestInfoKey[");
			if (_domain != null)
				sb.append(_domain).append("#");

			if (_fieldInfo == null)
			{
				if (_classInfo != null)
					sb.append(_classInfo.getId());
			}
			else
				sb.append(_classInfo.getId()).append(',').append(_fieldInfo.getId());
			sb.append(',').append(_selfTargetRelationClass);
			sb.append(']');
			return sb.toString();
		}

	} // end class _LySqlRequestInfoKey

	/**
	 * Accessor on current debug task or create a new one if no task is found.
	 */
	protected LySqlRequestInfoDebugTask getCurrentTask()
	{
		if (_currentTask == null)
			_currentTask = new LySqlRequestInfoDebugTask();
		return _currentTask;
	}

	/**
	 * Accessor on current debug task or create a new one if no task is found.
	 */
	protected void setCurrentTask(LySqlRequestInfoDebugTask task)
	{
		if (task == null)
			_currentTask = null;
		else
		{
			if (_currentTask == null)
				_currentTask = task;
			else
				System.err.println("Task " + _currentTask.getCounter()
					+ " can not be overriden by task " + task.getCounter());
		}
	}

	/**
	 * May add ignore table names / aliases in select from (because this select is nested in another
	 * select and alias comes from parent).
	 *
	 * @param tableName name of the table
	 * @param alias     name of the alias
	 */
	protected void setIgnoreInSelectFrom(String tableName, String alias, boolean status)
	{
		if (_ignoreInSelectFrom == null)
			_ignoreInSelectFrom = new HashMap<String, Boolean>();

		String key = tableName + "##" + alias;

		if (!status)
			_ignoreInSelectFrom.remove(key);
		else
			_ignoreInSelectFrom.put(key, Boolean.TRUE);
	}

	/**
	 * May add ignore table names / aliases in select from (because this select is nested in another
	 * select and alias comes from parent).
	 *
	 * @param tableName name of the table
	 * @param alias     name of the alias
	 */
	protected boolean isIgnoreInSelectFrom(String tableName, String alias)
	{
		if (_ignoreInSelectFrom == null)
			return false;

		String  key = tableName + "##" + alias;
		Boolean b = _ignoreInSelectFrom.get(key);

		if (b != null)
			return b.booleanValue();
		else
			return false;
	}

	/**
	 * Force using alias an not just table names
	 * Example : see FFT00757 when using left join clauses
	 * @param b new alias using status
	 */
	public void setForceUseAlias(boolean b)
	{
		_forceUseAlias = b;
	}

	/**
	 * Accessor : Force using alias an not just table names
	 * Example : see FFT00757 when using left join clauses
	 * @return new alias using status
	 */
	public boolean getForceUseAlias()
	{
		return _forceUseAlias;
	}
	/**
	 * Force using alias an not just table names
	 * Example : see FFT00757 when using left join clauses
	 * @param b new alias using status
	 */
	public void setSQLFormatDateInOrderByClause(String id, String sqlFormattedDate)
	{
		if (_mapSQLFormatDateInOrderByClause == null)
		{
			_mapSQLFormatDateInOrderByClause = new HashMap<String, String>();
		}
		if (id != null && sqlFormattedDate != null)
			_mapSQLFormatDateInOrderByClause.put(id, sqlFormattedDate);
	}

	/**
	 * Accessor : Force using alias an not just table names
	 * Example : see FFT00757 when using left join clauses
	 * @return new alias using status
	 */
	public String getSQLFormatDateInOrderByClause(String id)
	{
		if (_mapSQLFormatDateInOrderByClause != null && id != null)
		{
			return _mapSQLFormatDateInOrderByClause.get(id);
		}
		return null;
	}

	public boolean isSQLFormatDateInOrderByClauseEmpty()
	{
		return _mapSQLFormatDateInOrderByClause == null;
	}

} // end class LySqlRequestInfo
