/*
 * LyAccessRequestEncoder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.sql.access;

import java.sql.*;
import java.util.*;
import leon.data.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.sql.*;
import leon.peer.util.*;


/**
 * This class is used to send requests to a database connection and to receive its answers.
 * <p>It uses an LySqlDataEncoder to encode and decode the values of requests and answers.</p>
 * @version $Id: LyAccessRequestEncoder.java 28997 2010-12-23 10:17:15Z bmichel $
 * @author  Lyria
 * @see     LyRequestEncoder
 */
public class LyAccessRequestEncoder extends LySqlRequestEncoder
{
	public static final String __VERSION =
		"$Id: LyAccessRequestEncoder.java 28997 2010-12-23 10:17:15Z bmichel $";

	/**
	 * Constructor.
	 * @param classInfo ClassInfo used by the sql request encoder.
	 */
	public LyAccessRequestEncoder(LyDataProvider provider, LyDataSource dataSource, String valueSep)
	{
		super(provider, dataSource, valueSep);
	}

	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		super.init();
		if (getEnvironment().testResource("LY_DBMS_QUOTE_TABLES_AND_COLUMNS"))
		{
			QUOTE_CHAR_LEFT = "\"";
			QUOTE_CHAR_RIGHT = "\"";
		}
	}


	/**
	 * Returns a boolean that indicates if case sensitive is supported. Default implementation
	 * returns false for Access
	 * @return True if supported, false otherwise.
	 */
	public boolean supportCaseSensitive()
	{
		return false;
	}

	/**
	 */
	protected String beginEncodeSelectRequest(int limit, boolean distinct)
	{
		String result = (distinct ? "select distinct " : "select ");

		if (limit > 0)
			result += "TOP " + limit + " ";

		return result;
	}

	/**
	 */
	protected String getSqlValue(LyColumnField field, java.util.Date value)
	{
		if (value == null)
			return null;

		Class<?> valueClass = field.getValueClass();

		if (valueClass == java.util.Date.class)
		{
			StringBuffer  result;
			long		  val;
			Time		  time;
			java.sql.Date sqlDate;

			val = value.getTime();

			time = new Time(val);
			sqlDate = new java.sql.Date(val);

			result = new StringBuffer();
			result.append("DateValue(");
			result.append(CHAR_QUOTE);
			result.append(sqlDate.toString());
			result.append(" ");
			result.append(time.toString());
			result.append(CHAR_QUOTE);
			result.append(")");

			return result.toString();
		}
		else
			return getSqlValue(field, value.toString());
	} // end method getSqlValue

	/**
	 * This method is used after an insert request to complete the object if some values or the
	 * identifier is set by the database.
	 */
	protected void completeObject(Object object, LySqlConnection connection)
	{
		LyEnvironment environment = getEnvironment();

		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : "
				+ environment.getMessage("LY_ER_CANT_BE_NULL", object));
		if (connection == null)
			throw new IllegalArgumentException("BF AppEngine : "
				+ environment.getMessage("LY_ER_CANT_BE_NULL", connection));

		// If the object has no identifier, it may be an identity.
		LyValueSet   newValues = ((LyRowObject)object).getNewValues();
		LyTableClass tableClass = ((LyRowObject)object).getTableClass();
		Vector<LyField>		 idFields = tableClass.getIdFields();

		if (idFields.size() != 1)
			// If there are several id fields, they can't be identity.
			return;

		LyColumnField columnField = (LyColumnField)idFields.firstElement();

		if (newValues.contains(columnField))
			// The identifier was not set by the database.
			return;

		// Create the request.
		StringBuffer request_text = new StringBuffer();

		request_text.append("select @@identity");

		// Send the request to the connection
		ResultSet rs = connection.sendQueryRequest(request_text.toString());

		if (rs == null)
			// An error occured while sending the request to the database
			environment.logError("An error occured while sending the request\n"
				+ request_text + "\nto the database.");
		else
		{
			// The column in the database is an integer.
			try
			{
				rs.next();
				Object value = getPeerValue(rs, columnField, 1, columnField.getValueClass());

				((LyRowObject)object).setValue(columnField, value);
			}
			catch (Exception e)
			{
				environment.writeException(e);
				return;
			}
		}
	} // end method completeObject
} // end class LyAccessRequestEncoder
