/*
 * LyIdbRequestEncoder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.sql.idb;

import java.sql.*;
import java.util.*;

import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.sql.*;
import leon.peer.util.*;


/**
 * This class is used to send requests to a SQL InstantDB database connection and to receive its
 * answers.
 * @version $Id: LyIdbRequestEncoder.java 29765 2011-02-22 13:37:57Z jlsaouli $
 * @author  Lyria
 * @see     LyRequestEncoder
 * @see     LySqlConnection
 * @see     LySqlDataProvider
 */
public class LyIdbRequestEncoder extends LySqlRequestEncoder
{
	public static final String    __VERSION =
		"$Id: LyIdbRequestEncoder.java 29765 2011-02-22 13:37:57Z jlsaouli $";

	protected static final String STRING_BACKSLASH = "\\";
	protected static final String STRING_TWO_BACKSLASHES = "\\\\";

	/**
	 * Constructor.
	 * @param provider
	 * @param dataSource
	 * @param valueSep
	 */
	public LyIdbRequestEncoder(LyDataProvider provider, LyDataSource dataSource, String valueSep)
	{
		super(provider, dataSource, valueSep);
	}
	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		super.init();
		if (getEnvironment().testResource("LY_DBMS_QUOTE_TABLES_AND_COLUMNS"))
		{
			QUOTE_CHAR_LEFT = "\"";
			QUOTE_CHAR_RIGHT = "\"";
		}
	}

	/**
	 */
	protected String getSqlValue(LyColumnField field, String value)
	{
		if (value == null)
			return null;

		String str = super.getSqlValue(field, value);

		// We must replace '\' by '\\'.
		str = LyString.replace(str, STRING_BACKSLASH, STRING_TWO_BACKSLASHES);

		// We must replace "''" by "\'"
		str = LyString.replace(str, STRING_TWO_QUOTES, "\\'");

		return str;
	}

	/**
	 */
	protected String encodeSort(LySortCriterion criterion, String tableName, String alias, Map<LyField, String> aliases)
	{
		if (criterion == null)
			return null;

		LyFieldInfo fieldInfo = criterion.getFieldInfo();

		if (fieldInfo.getBinding() == null)
		{
	        _response.setSorted(false);
			return null;
		}

		LyField[]    peerFields = _classBinding.getPeerFields(fieldInfo);
		String		 id;
		StringBuffer result = null;
		int			 i, n = peerFields.length;
		boolean		 caseSensitive = false;

		if (!supportCaseSensitive() || !(fieldInfo instanceof LyTextFieldInfo))
			caseSensitive = true;

		if (!caseSensitive)
			_response.setSorted(false);

		for (i = 0; i < n; i++)
		{
			id = peerFields[i].getId();
			if (_class.getField(id) == null)
				continue;

			if (tableName == null)
				tableName = getTableName(peerFields[i]);

			if (tableName != null)
				id = QUOTE_CHAR_LEFT+(alias != null? alias : getTableAlias(null, tableName))+ QUOTE_CHAR_RIGHT 
					+ "." + QUOTE_CHAR_LEFT + id + QUOTE_CHAR_RIGHT;

			if (result == null)
				result = new StringBuffer();
			else
				result.append(", ");

			result.append(id);
			if (criterion.isIncreasing())
				result.append(" ASC");
			else
				result.append(" DESC");
		}

		return ((result == null) ? null : result.toString());
	} // end method encodeSort

	/**
	 * This method is used after an insert request to complete the object if some values or the
	 * identifier is set by the database.
	 */
	protected void completeObject(Object object, LySqlConnection connection)
	{
		LyEnvironment environment = getEnvironment();

		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : "
				+ environment.getMessage("LY_ER_CANT_BE_NULL", object));
		if (connection == null)
			throw new IllegalArgumentException("BF AppEngine : "
				+ environment.getMessage("LY_ER_CANT_BE_NULL", connection));

		// If the object has no identifier, it may be an identity.
		LyValueSet   	newValues = ((LyRowObject)object).getNewValues();
		LyTableClass 	tableClass = ((LyRowObject)object).getTableClass();
		Vector<LyField>	idFields = tableClass.getIdFields();

		if (idFields.size() != 1)
			// If there are several id fields, they can't be identity.
			return;

		LyColumnField columnField = (LyColumnField)idFields.firstElement();

		if ((newValues != null) && (newValues.contains(columnField)))
			// The identifier was not set by the database.
			return;

		// Create the request.
		StringBuffer request_text = new StringBuffer();

		request_text.append("select __IDB_INTERNAL last_value from TEXT(\"");
		request_text.append(columnField.getId());
		request_text.append("\") in TABLE ").append(QUOTE_CHAR_LEFT).append(_class.getName()).append(QUOTE_CHAR_RIGHT);
		request_text.append(" from ").append(QUOTE_CHAR_LEFT).append(_class.getName()).append(QUOTE_CHAR_RIGHT);

		// Send the request to the connection
		ResultSet rs = connection.sendQueryRequest(request_text.toString());

		if (rs == null)
			// An error occured while sending the request to the database
			environment.logError("An error occured while sending the request\n"
				+ request_text + "\nto the database.");
		else
		{
			// The column in the database is an integer.
			try
			{
				rs.next();
				Object value = getPeerValue(rs, columnField, 1, columnField.getValueClass());

				((LyRowObject)object).setValue(columnField, value);
			}
			catch (Exception e)
			{
				environment.writeException(e);
			}
			finally
			{
				try
				{
					rs.close();
				}
				catch (SQLException sqle)
				{
					if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
							&& environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
						environment.writeException(sqle);
				}
			}
		}
	} // end method completeObject
} // end class LyIdbRequestEncoder
