/*
 * LyMySqlRequestEncoder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.sql.mysql;

import java.sql.*;
import java.util.*;

import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.sql.*;
import leon.peer.util.*;


/**
 * This class is used to send requests to a database connection and to receive its answers.
 * <p>It uses an LySqlDataEncoder to encode and decode the values of requests and answers.</p>
 * @version $Id: LyMySqlRequestEncoder.java 40675 2012-09-05 16:26:11Z jlsaouli $
 * @author  Lyria
 * @see     LyRequestEncoder
 */
public class LyMySqlRequestEncoder extends LySqlRequestEncoder
{
	public static final String    __VERSION =
		"$Id: LyMySqlRequestEncoder.java 40675 2012-09-05 16:26:11Z jlsaouli $";

	protected static final String STRING_BACKSLASH = "\\";
	protected static final String STRING_TWO_BACKSLASHES = "\\\\";

	/**
	 * Constructor.
	 * @param classInfo ClassInfo used by the sql request encoder.
	 */
	public LyMySqlRequestEncoder(LyDataProvider provider, LyDataSource dataSource, String valueSep)
	{
		super(provider, dataSource, valueSep);
	}
	
	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		super.init();
		if (getEnvironment().testResource("LY_DBMS_QUOTE_TABLES_AND_COLUMNS"))
		{
			QUOTE_CHAR_LEFT="`";
			QUOTE_CHAR_RIGHT="`";
		}
	}
	
	/**
	 * To create a new object.
	 * Overriden to handle transactional handling of primary keys
	 * @param  values the values of the new object
	 */
	public boolean createObject(Object object)
	{
		boolean 	permanentDisableForeignKeys = "ALWAYS".equalsIgnoreCase(getEnvironment().getResource("LY_DISABLE_FOREIGN_KEY_CONSTRAINT"));
		boolean 	neverDisableForeignKeys = "NEVER".equalsIgnoreCase(getEnvironment().getResource("LY_DISABLE_FOREIGN_KEY_CONSTRAINT"));
		boolean 	foreignKeyDetection = false;
		
		if ((!permanentDisableForeignKeys) && (!neverDisableForeignKeys))
		{
			LyValueSet 	valueSet = ((LyRowObject)object).getNewValues();
			int 		n = ((valueSet == null) ? 0 : valueSet.getSize());
			
			for(int i=0; i<n && !foreignKeyDetection; i++)
			{
				LyColumnField field = (LyColumnField)valueSet.get(i).getField();
				
				if (field.hasContraint(LyColumnField.CONSTRAINT.FOREIGN_KEY))
				{
					if (getEnvironment().isDebugLevel(LyEnvironment.DEBUG_DBMS))
						getEnvironment().logOut("MySQL Foreign key detection while inserting "+valueSet.get(i));
					foreignKeyDetection = true;
				}
			}
		}		
		if (permanentDisableForeignKeys || foreignKeyDetection)
		{
			// Get a valid connection
			LySqlConnection connection = (LySqlConnection)_provider.getConnection();

			// permanently disable : send once per session only
			if (connection != null)
			{
				if ((foreignKeyDetection) 
						|| (!"ALWAYS".equals(connection.getValue("LY_DISABLE_FOREIGN_KEY_CONSTRAINT"))))
				{
					connection.sendUpdateRequest("SET foreign_key_checks=0");
					if (permanentDisableForeignKeys) 
					{
						connection.addValue("LY_DISABLE_FOREIGN_KEY_CONSTRAINT", "ALWAYS");
						if (getEnvironment().isDebugLevel(LyEnvironment.DEBUG_DBMS))
							getEnvironment().logOut("MySQL Foreign key constraints permanently disabled");
					}
				}
			}
			_provider.releaseConnection(connection);
		}
		boolean result = super.createObject(object);
		if (foreignKeyDetection)
		{
			// Get a valid connection
			LySqlConnection connection = (LySqlConnection)_provider.getConnection();
	
			if (connection != null)
				connection.sendUpdateRequest("SET foreign_key_checks=1");
			_provider.releaseConnection(connection);
		}

		return result;
	} // end method createObject


    /**
     * Set the request encoder response. Inherited method of LyRequestEncoder to set properties to
     * the response.
     *
     * @param the response
     */
    public void setResponse(LyResponse response)
    {
        super.setResponse(response);
        
        response.setPaginated(true);
    }

	/**
	 */
	protected String endEncodeSelectRequest(LySort sort, int limit, String tableName,
        Map<LyField, String> aliases)
	{
		String result = encodeSort(sort, tableName, aliases);

		if (limit > 0)
		{
			if (result == null)
				result = new String();

			result += encodeLimitCondition(limit);
		}

		return result;
	}
	
	/**
	 * encode the limit condition 
	 * @param limit max number of rows fetched 
	 */
	public String encodeLimitCondition(int limit)
	{
		if (limit <=0)
			return null;
		
        int		n = limit, p = 0;
        int		maxLimit = LyPageWorkSpace.MAX_LIMIT;
        
        if (limit > maxLimit)
        {
            p = limit / maxLimit;
            n = limit % maxLimit;
            return " LIMIT " + (n * p) + "," + n;
        }
        else
            return " LIMIT " + limit;
	}
	
	/**
	 */
	protected String getSqlValue(LyColumnField field, String value)
	{
		if (value == null)
			return null;

		String str = super.getSqlValue(field, value);

		// We must replace '\' by '\\'.
		str = LyString.replace(str, STRING_BACKSLASH, STRING_TWO_BACKSLASHES);

		return str;
	}
	
	/**
	 * Utility internal method (may change based on database)
	 * @see #encodeSelectCountEnd()
	 */
	protected String encodeSelectCountBegin(LyFilter filter, boolean writeTableName)
	{
		// REMARK : SAME AS ORACLE, but use an alias AS xxx at the end
		//FFT Flandres 355
		StringBuffer sb = new StringBuffer();
		String  tableName = (writeTableName ? getTableName(_class) : null);

		// Test the filter needs a distinct on its key
		Vector<? extends LyField> idFields = _class.getIdFields();

		sb.append("select count(*) from (select distinct ");
		// ** A PARENTHESIS IS LEFT OPENED, closed in completeEncodeSelectCount()
		// add the name of the first key column
		LySqlColumnField field;

		for(int i=0; i<idFields.size(); i++)
		{
			if (i > 0)
				sb.append(", ");
			field = (LySqlColumnField)idFields.elementAt(i);

			if (writeTableName)
				sb.append(QUOTE_CHAR_LEFT).append(getTableAlias(null, tableName)).append(QUOTE_CHAR_RIGHT).append(".");
			else if (field.getClassBinding() != null)
				sb.append(QUOTE_CHAR_LEFT).append(getTableAlias(null, field.getClassBinding())).append(QUOTE_CHAR_RIGHT).append(".");
			sb.append(QUOTE_CHAR_LEFT).append(field.getId()).append(QUOTE_CHAR_RIGHT);
		}
		sb.append(" from ");

		return sb.toString();
	}

	/**
	 * Utility internal method (may change based on database)
	 * @see #encodeSelectCountBegin()
	 */
	protected String completeEncodeSelectCount()
	{
		// ** A PARENTHESIS IS LEFT OPENED in encodeSelectCountBegin()
		return ") AS DERIVED_TABLE_NAME";
	}	

	/**
	 * This method is used after an insert request to complete the object if some values or the
	 * identifier is set by the database.
	 */
	protected void completeObject(Object object, LySqlConnection connection)
	{
		LyEnvironment environment = getEnvironment();

		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : "
				+ environment.getMessage("LY_ER_CANT_BE_NULL", object));
		if (connection == null)
			throw new IllegalArgumentException("BF AppEngine : "
				+ environment.getMessage("LY_ER_CANT_BE_NULL", connection));

		// If the object has no identifier, it may be an identity.
		LyValueSet   newValues = ((LyRowObject)object).getNewValues();
		LyTableClass tableClass = ((LyRowObject)object).getTableClass();
		Vector<LyField>		 idFields = tableClass.getIdFields();
		LyColumnField columnField = null, columnFieldTmp;
		
		for (int i = 0; i < idFields.size(); i++)
		{
			columnFieldTmp = (LyColumnField)idFields.elementAt(i);

			if ((newValues == null) || !newValues.contains(columnFieldTmp))
			{
				// If there are several id fields, they can't be identity.
				if (columnField != null)
					return;

				// The identifier was not set by the database.
				columnField = columnFieldTmp;
			}
		}

		if (columnField == null)
			return;
		
		// Create the request.
		StringBuffer request_text = new StringBuffer();

		request_text.append("select LAST_INSERT_ID()");

		// Send the request to the connection
		ResultSet rs = connection.sendQueryRequest(request_text.toString());

		if (rs == null)
			// An error occured while sending the request to the database
			environment.logError("An error occured while sending the request\n"
				+ request_text + "\nto the database.");
		else
		{
			// The column in the database is an integer.
			try
			{
				rs.next();
				Object value = getPeerValue(rs, columnField, 1, columnField.getValueClass());

				((LyRowObject)object).setValue(columnField, value);
			}
			catch (Exception e)
			{
				environment.writeException(e);
			}
			finally
			{
				try
				{
					rs.close();
				}
				catch (SQLException sqle)
				{
					if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
							&& environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
						environment.writeException(sqle);
				}
			}
		}
	} // end method completeObject
	
	/**
	 * String force formating date for order by using FORMAT(quotedId, 'YYYY-MM-DD HH:MM:SS')
	 * but syntax is specific depending on database vendor.
	 * If not overridden this method does nothing and returns the quotedId to keep backward compatibility.
	 * @param quotedId id of the date field (using aliases, table names and quotes if needed)
	 * @param peerField peer field related to this column
	 * @return the sort criteria to use in the SQL request
	 */
	public String getSQLFormatDate(String quotedId, LyDateFieldInfo fieldInfo)
	{
		// syntax example : 
		// DATE_FORMAT(commande_date, '%Y-%m-%d %H:%i:%s')
		StringBuffer result = new StringBuffer();
		result.append("DATE_FORMAT(").append(quotedId).append(',');

		if (fieldInfo.getType() == LyDateFieldInfo.Type.DATE)
		{
			result.append("'%Y-%m-%d')");
		}
		else if (fieldInfo.getType() == LyDateFieldInfo.Type.DATE_HOUR_MIN)
		{
			result.append("'%Y-%m-%d %H:%i')");
		}
		else if (fieldInfo.getType() == LyDateFieldInfo.Type.HOUR_MIN)
		{
			result.append("'%H:%i')");
		}
		else if (fieldInfo.getType() == LyDateFieldInfo.Type.TIME)
		{
			result.append("'%H:%i:%s')");
		}
		else // DATE_TIME
			result.append("'%Y-%m-%d %H:%i:%s')"); // default format

		return result.toString();
	}
	
	/**
	 * Gets locate function.
	 */
	protected String getLocateFunction()
	{
		return "LOCATE";
	}
} // end class LyMySqlRequestEncoder
