/*
 * LyPostgresqlDataProvider.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.peer.sql.postgresql;

import java.sql.*;
import java.util.*;

import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.*;
import leon.info.location.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.sql.*;
import leon.peer.util.*;


/**
 * This class is used by DataSources to transmit their requests to an instance of
 * PostgresqlRequestEncoder.
 *
 * @version $Id: LyPostgresqlDataProvider.java 30865 2011-05-09 12:21:53Z bmichel $
 * @author  Lyria
 * @see     LySqlDataProvider
 * @see     LySqlRequestEncoder
 * @see     LySqlConnection
 */
public class LyPostgresqlDataProvider extends LySqlDataProvider
{
	public static final String __VERSION =
		"$Id: LyPostgresqlDataProvider.java 30865 2011-05-09 12:21:53Z bmichel $";

	/**
	 * Default constructor Initialization is done with the init method.
	 */
	public LyPostgresqlDataProvider(LyApplication application, LyLocation location)
	{
		super(application, location);
	}

	/**
	 */
	public LyRequestEncoder getRequestEncoder(LyDataSource dataSource)
	{
		LySqlRequestEncoder encoder = new LyPostgresqlRequestEncoder(this, dataSource, _valueSep);

		encoder.setTrim(_trim);
		return encoder;
	}

	/**
	 * Factory to create a new LyColumnField instance May be specialized in subclasses, default
	 * return a LyColumnField instance.
	 *
	 * @param  fieldBinding reference of the fieldBinding associated with the columnField to create
	 * @return a LyColumnField instance
	 */
	protected LyColumnField createColumnField(LyFieldBinding fieldBinding)
	{
		LyPostgresqlField result = new LyPostgresqlField();

		return result;
	}

	/**
	 * Determine the class for memory storage associated to a columnField. This method is internally
	 * used by the validate method. Method should be overriden for optimization by subclasses.
	 *
	 * @param  tableClass  the peer table class reference as specified in validate method.
	 * @param  columnField peer field associated with the value we store.
	 * @param  fieldInfo   data description of the meta model.
	 * @return the java class used to store data in memory.
	 * @see    #validate()
	 */
	protected Class<?> getValueClass(LyTableClass tableClass, LyColumnField columnField,
		LyFieldInfo fieldInfo)
	{
		if (fieldInfo instanceof LyChoiceFieldInfo)
		{
			LyChoiceFieldInfo choice = (LyChoiceFieldInfo)fieldInfo;
			int				  n = choice.getChoiceOptionCount();

			if (n == 2)
			{
				LyChoiceOption option1, option2;
				Object		   value1, value2;

				option1 = choice.getChoiceOption((short)0);
				option2 = choice.getChoiceOption((short)1);

				value1 = option1.getValue();
				value2 = option2.getValue();

				if ((value1 instanceof String) && (value2 instanceof String))
				{
					if (((value1.equals("true")) || (value1.equals("false")))
							&& (value2.equals("true") || (value2.equals("false"))))
						((LyPostgresqlField)columnField).setBoolean(true);
				}
			}
		}

		return super.getValueClass(tableClass, columnField, fieldInfo);
	} // end method getValueClass

	/**
	 * To get the name of the catalog to use with metadata.
	 *
	 * @return the name of the catalog.
	 */
	protected String getCatalogPattern()
	{
		return _databaseName;
	}

	/**
	 * To get the name of the schema to use with metadata.
	 *
	 * @return the name of the schema.
	 */
	protected String getSchemaPattern()
	{
		return null;
	}

	/**
	 * get metaTable informations Overriden to store the name only in uppercase in the hashtable.
	 *
	 * @param  DatabaseMetaData metadta for the current database
	 * @return Hashtable cooresponding informations for tables
	 */
	protected Hashtable<String, LyMetaDataTableInformation> getTables(DatabaseMetaData metaData)
	{
		ResultSet									  rs = null;
		String[]									  types = new String[1];
		String										  name, remarks;
		LyMetaDataTableInformation					  table;
		Hashtable<String, LyMetaDataTableInformation> result =
			new Hashtable<String, LyMetaDataTableInformation>();
		String										  catalog, schema;
		LyEnvironment								  environment = _application.getEnvironment();

		types[0] = "TABLE";

		try
		{
			catalog = getCatalogPattern();
			schema = getSchemaPattern();

			if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
				environment.logOut("DatabaseMetaData getTables(" + catalog + "," + schema
					+ ",null,types[0]=" + types[0] + ")");

			rs = metaData.getTables(catalog, schema, null, types);

			while (rs.next())
			{
				name = rs.getString("TABLE_NAME");
				try
				{
					remarks = rs.getString("REMARKS");
				}
				catch (SQLException ex)
				{
					remarks = null;
				}

				if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
					environment.logOut("DatabaseMetaData getTables Result(TABLE_NAME=" + name
						+ ",REMARKS=" + remarks + ")");

				table = new LyMetaDataTableInformation(name, remarks);
				result.put(name, table);
			}
		}
		catch (SQLException e)
		{
			environment.writeException(e);
			result = null;
		}
		finally
		{
			if (rs != null)
			{
				try
				{
					rs.close();
				}
				catch (SQLException e)
				{
				}
			}
		}

		return result;
	} // end method getTables

	/**
	 * complete tables Overriden to store the name only in uppercase in the hashtable.
	 *
	 * @param  DatabaseMetaData metadata for the current database
	 * @param  Hashtable        tables
	 * @return boolean if the operation is succeded
	 */
	protected boolean completeTables(DatabaseMetaData metaData,
		Hashtable<String, LyMetaDataTableInformation> tables)
	{
		ResultSet							    rs = null;
		String								    str;
		Enumeration<LyMetaDataTableInformation> elements = tables.elements();
		LyMetaDataTableInformation			    table;
		LyMetaDataColumnInformation			    column;
		short								    keyseq;
		boolean								    hasPrimaryKey;
		String								    tableName;
		short								    dataType;
		String								    typeName;
		String								    name;
		int									    columnSize;
		int									    decimalDigits;
		int									    nullable;
		int									    octetLength;
		int									    position;
		String								    remarks;
		String								    isNullable;
		String								    def;
		String								    catalog, schema;
		boolean								    result = true;
		LyEnvironment						    environment = _application.getEnvironment();

		try
		{
			while (elements.hasMoreElements())
			{
				catalog = getCatalogPattern();
				schema = getSchemaPattern();
				table = elements.nextElement();
				str = table.getName();
				if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
					environment.logOut("Table " + str + " - columns : ");

				// Get the columns of the table.
				if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
					environment.logOut("DatabaseMetaData getColumns("
						+ catalog + "," + schema + "," + str + ",null)");

				rs = metaData.getColumns(catalog, schema, str, null);

				while (rs.next())
				{
					tableName = rs.getString("TABLE_NAME");
					dataType = rs.getShort("DATA_TYPE");
					typeName = rs.getString("TYPE_NAME");
					name = rs.getString("COLUMN_NAME");

					if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
						environment.logOut("\t" + name + "\t" + dataType + "\t" + typeName);

					columnSize = rs.getInt("COLUMN_SIZE");
					
					decimalDigits = rs.getInt("DECIMAL_DIGITS");
					if(decimalDigits > 10)
						decimalDigits = 10;
					
					nullable = rs.getInt("NULLABLE");
					remarks = rs.getString("REMARKS");
					isNullable = rs.getString("IS_NULLABLE");
					try
					{
						def = rs.getString("COLUMN_DEF");
					}
					catch (SQLException ex)
					{
						def = null;
					}

					octetLength = rs.getInt("CHAR_OCTET_LENGTH");
					position = rs.getInt("ORDINAL_POSITION");

					column = new LyMetaDataColumnInformation(name, remarks);
					column.setTableName(tableName);
					column.setDataType(dataType);
					column.setTypeName(typeName);
					column.setColumnSize(columnSize);
					column.setDecimalDigits(decimalDigits);
					column.setNullable(nullable);
					column.setIsNullable(isNullable);
					column.setDef(def);
					column.setOctetLength(octetLength);
					column.setPosition(position);
					table.addColumn(column);

				}

				rs.close();
				rs = null;

				// Get the primary keys of the table
				try
				{
					rs = metaData.getPrimaryKeys(catalog, schema, str);
					hasPrimaryKey = false;
					if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
						environment.logOut("\t Primary Keys : ");

					while (rs.next())
					{
						hasPrimaryKey = true;
						name = rs.getString("COLUMN_NAME");
						if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
							environment.logOut("\t\t" + name);
						keyseq = rs.getShort("KEY_SEQ");

						column = table.getColumn(name);
						if (column != null)
						{
							boolean isIdentity = false;
							typeName = column.getTypeName();
							
							if(typeName != null)
								isIdentity = typeName.equals("serial");
							
							column.setKeyOrder(keyseq);
							column.setIsPrimaryKey(isIdentity);
						}
					}
				}
				catch (Exception ex)
				{
					if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
						environment.writeException(ex);
					hasPrimaryKey = false;
				}
				finally
				{
					if (rs != null)
					{
						rs.close();
						rs = null;
					}
				}

				// Foreign Keys of the table
				try
				{
					rs = metaData.getImportedKeys(catalog, schema, str);
					if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
						environment.logOut("\t foreing Keys : ");

					while (rs.next())
					{
						String	fkName = rs.getString("FK_NAME");
						String	pkColumnName = rs.getString("PKCOLUMN_NAME");
						String	pkTableName = rs.getString("PKTABLE_NAME");
						
						if (environment.isDebugLevel(LyEnvironment.DEBUG_PEER))
							environment.logOut("\t\t" + rs.getString("FKCOLUMN_NAME"));
						column = table.getColumn(rs.getString("FKCOLUMN_NAME"));
						if (column != null)
							column.setIsImportedKey(fkName, pkTableName, pkColumnName);
					}
				}
				catch (Exception ex)
				{
					if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS))
						environment.writeException(ex);
				}
				finally
				{
					if (rs != null)
					{
						rs.close();
						rs = null;
					}
				}

				if (!hasPrimaryKey)
				{
					// The information is not available.
					Enumeration<LyMetaDataColumnInformation> columns = table.getColumns();

					while (columns.hasMoreElements())
					{
						column = columns.nextElement();
						column.setKeyOrder(LyMetaDataColumnInformation.NO_KEY);
					}
				}
			}
		}
		catch (SQLException e)
		{
			environment.writeException(e);
			result = false;
		}
		finally
		{
			if (rs != null)
			{
				try
				{
					rs.close();
				}
				catch (SQLException e)
				{
				}
			}
		}

		return result;
	} // end method completeTables

} // end class LyPostgresqlDataProvider
