/*
 * LyPostgresqlRequestEncoder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.peer.sql.postgresql;

import java.sql.*;
import java.util.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.sql.*;
import leon.peer.util.*;


/**
 * This class is used to send requests to a SQL database connection and to receive its answers. This
 * is a specialization for postgresql database .
 * @version $Id: LyPostgresqlRequestEncoder.java 40675 2012-09-05 16:26:11Z jlsaouli $
 * @author  Lyria
 * @see     LyRequestEncoder
 * @see     LySqlConnection
 * @see     LyPostgresqlDataProvider
 */
public class LyPostgresqlRequestEncoder extends LySqlRequestEncoder
{
	public static final String    __VERSION =
		"$Id: LyPostgresqlRequestEncoder.java 40675 2012-09-05 16:26:11Z jlsaouli $";

	protected static final String STRING_BACKSLASH = "\\";
	protected static final String STRING_TWO_BACKSLASHES = "\\\\";

	/**
	 * @param provider
	 * @param dataSource
	 * @param valueSep
	 */
	public LyPostgresqlRequestEncoder(LyDataProvider provider, LyDataSource dataSource,
		String valueSep)
	{
		super(provider, dataSource, valueSep);
	}

	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		super.init();
		if (getEnvironment().testResource("LY_DBMS_QUOTE_TABLES_AND_COLUMNS"))
		{
			QUOTE_CHAR_LEFT = "\"";
			QUOTE_CHAR_RIGHT = "\"";
		}
	}

	/**
	 * Returns a boolean that indicates if case sensitive is supported.
	 * @return True if supported, false otherwise.
	 */
	public boolean supportCaseSensitive()
	{
		return false;
	}

	/**
	 * Gets the SQL encoded string corresponding to the given string value. This method is overriden
	 * to manage boolean values.
	 */
	protected String getSqlValue(LyColumnField field, String value)
	{
		if (value == null)
			return null;
		else if (value.equals("true") || value.equals("false"))
			return value;
		else
		{
			String str = super.getSqlValue(field, value);

			// We must replace '\' by '\\'.
			str = LyString.replace(str, STRING_BACKSLASH, STRING_TWO_BACKSLASHES);

			return str;
		}
	}

	/**
	 * Gets the SQL encoded String corresponding to the given date value This method is overriden to
	 * manage date and time value.
	 */
	protected String getSqlValue(LyColumnField field, java.util.Date value)
	{
		if (value == null)
			return null;

		Class<?> valueClass = field.getValueClass();

		if (valueClass == java.util.Date.class)
		{
			StringBuffer     result;
			long			 val;
			Time			 time;
			java.sql.Date    sqlDate;
			LySqlColumnField sqlField = (LySqlColumnField)field;
			int				 dateType = sqlField.getDateType();

			val = value.getTime();

			time = new Time(val);
			sqlDate = new java.sql.Date(val);

			result = new StringBuffer();
			result.append(CHAR_QUOTE);

			if (dateType != Types.TIME)
			{
				result.append(sqlDate.toString());
				result.append(" ");
			}

			result.append(time.toString());
			result.append(CHAR_QUOTE);

			return result.toString();
		}
		else
			return getSqlValue(field, value.toString());
	} // end method getSqlValue

	/**
	 * To decode a value given by the database. This method is overriden to manage boolean type.
	 */
	protected Object getPeerValue(ResultSet rs, LyColumnField field, int index, Class<?> valueClass)
		throws LySqlUnsupportedConversionException,
			SQLException
	{
		if (((LyPostgresqlField)field).isBoolean())
		{
			boolean val = rs.getBoolean(index);

			return (Boolean.valueOf(val).toString());
		}

		return super.getPeerValue(rs, field, index, valueClass);
	}

	/**
	 * This method is used after an insert request to complete the object if some values or the
	 * identifier is set by the database.
	 */
	protected void completeObject(Object object, LySqlConnection connection)
	{
		LyEnvironment environment = getEnvironment();

		if (object == null)
			throw new IllegalArgumentException("BF AppEngine : "
				+ environment.getMessage("LY_ER_CANT_BE_NULL", object));
		if (connection == null)
			throw new IllegalArgumentException("BF AppEngine : "
				+ environment.getMessage("LY_ER_CANT_BE_NULL", connection));

		// If the object has no identifier, it may be an identity.
		LyValueSet   	newValues = ((LyRowObject)object).getNewValues();
		LyTableClass	tableClass = ((LyRowObject)object).getTableClass();
		Vector<LyField>	idFields = tableClass.getIdFields();

		if (idFields.size() != 1)
			// If there are several id fields, they can't be identity.
			return;

		LyColumnField columnField = (LyColumnField)idFields.firstElement();

		if ((newValues != null) && newValues.contains(columnField))
			// The identifier was not set by the database.
			return;

		String		   tableName = _classBinding.getBinding();
		LyFieldBinding fieldBinding = _classBinding.getFieldBinding(columnField);
		String		   fieldName = fieldBinding.getBinding();

		// Create the request.
		StringBuffer   request_text = new StringBuffer();

		request_text.append("select currval('" + tableName + "_" + fieldName + "_seq')");

		// Send the request to the connection
		ResultSet rs = connection.sendQueryRequest(request_text.toString());

		if (rs == null)
			// An error occured while sending the request to the database
			environment.logError("An error occured while sending the request\n"
				+ request_text + "\nto the database.");
		else
		{
			// The column in the database is an integer.
			try
			{
				rs.next();
				Object value = getPeerValue(rs, columnField, 1, columnField.getValueClass());

				((LyRowObject)object).setValue(columnField, value);
			}
			catch (Exception e)
			{
				environment.writeException(e);
			}
			finally
			{
				try
				{
					rs.close();
				}
				catch (SQLException sqle)
				{
					if (environment.isDebugLevel(LyEnvironment.DEBUG_DBMS)
							&& environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
						environment.writeException(sqle);
				}
			}
		}
	} // end method completeObject

	@Override
	protected String endEncodeSelectRequest(LySort sort, int limit, String tableName,
			Map<LyField, String> sortAliases)
	{
		String result = encodeSort(sort, tableName, sortAliases);

		if (limit > 0)
		{
			if (result == null)
				result = new String();

			result += encodeLimitCondition(limit);
		}

		return result;
	}

	@Override
	public String encodeLimitCondition(int limit)
	{
		if (limit <=0)
			return null;

		int  	n = limit, p = 0;
        int		maxLimit = LyPageWorkSpace.MAX_LIMIT;

		if (limit > maxLimit)
		{
			p = limit / maxLimit;
			n = limit % maxLimit;
			return " limit " + n + " offset " + (n * p);
		}
		else
			return " limit " + limit;
	}

	 /**
	  * Set the request encoder response. Inherited method of LyRequestEncoder to set properties to
	  * the response.
	  *
	  * @param the response
	  */
	public void setResponse(LyResponse response)
	{
		super.setResponse(response);

		response.setPaginated(true);
	}

	/**
	 * String force formating date for order by using FORMAT(quotedId, 'YYYY-MM-DD HH:MM:SS')
	 * but syntax is specific depending on database vendor.
	 * If not overridden this method does nothing and returns the quotedId to keep backward compatibility.
	 * @param quotedId id of the date field (using aliases, table names and quotes if needed)
	 * @param peerField peer field related to this column
	 * @return the sort criteria to use in the SQL request
	 */
	public String getSQLFormatDate(String quotedId, LyDateFieldInfo fieldInfo)
	{
		// syntax example :
		// TO_CHAR(commande_date, 'YYYY-MM-DD HH24:MI:SS')
		StringBuffer result = new StringBuffer();
		result.append("TO_CHAR(").append(quotedId).append(',');

		if (fieldInfo.getType() == LyDateFieldInfo.Type.DATE)
		{
			result.append("'YYYY-MM-DD')");
		}
		else if (fieldInfo.getType() == LyDateFieldInfo.Type.DATE_HOUR_MIN)
		{
			result.append("'YYYY-MM-DD HH24:MI')");
		}
		else if (fieldInfo.getType() == LyDateFieldInfo.Type.HOUR_MIN)
		{
			result.append("'HH24:MI')");
		}
		else if (fieldInfo.getType() == LyDateFieldInfo.Type.TIME)
		{
			result.append("'HH24:MI:SS')");
		}
		else // DATE_TIME
			result.append("'YYYY-MM-DD HH24:MI:SS')"); // default format
		// FFT03609 and FFT04177
		getRequestInfo().setSQLFormatDateInOrderByClause(quotedId, result.toString());
		return result.toString();
	}

	/**
	 * Complete the sort criteria for ascending or descending SQL syntax.
	 *
	 * @param increasing true if we expect an increasing sort
	 * @param optional   is true if the field is optional.
	 */
	protected String getSqlSortCriterion(boolean increasing, boolean optional)
	{
		if (increasing)
			return " ASC NULLS FIRST";
		else
			return " DESC NULLS LAST";
	}

	/**
	 * Gets selected field for a multilingual field.
	 * @param lang		Selected lang.
	 * @param field		Selected field.
	 * @param tableName	Table name
	 * @return Selected lang field.
	 */
	protected String getSelectedLangField(String lang, LyField field, String tableName)
	{
		if (lang == null)
			return null;

		String		id = field.getId();
		String		quotedId;

		quotedId = QUOTE_CHAR_LEFT;
		if (tableName != null)
			quotedId += getTableAlias(null, tableName) + QUOTE_CHAR_RIGHT + "." + QUOTE_CHAR_LEFT;
		quotedId += id + QUOTE_CHAR_RIGHT;

		String		startLang = LyTextFieldInfo.OPEN_START_LANG + lang + LyTextFieldInfo.CLOSE_START_LANG;
		String		endLang = LyTextFieldInfo.OPEN_END_LANG + lang + LyTextFieldInfo.CLOSE_END_LANG;

		return "IF(POSITION('" + startLang + "' IN " +  quotedId + ")>0,SUBSTR("
			+ quotedId + ",POSITION('" + startLang + "' IN "
			+  quotedId + ")+" + startLang.length() + ",POSITION('" + endLang
			+ "' IN " + quotedId + ")-POSITION('" + startLang + "' IN " + quotedId + ")),"
			+ quotedId + ") as " + id + "_" + lang.toUpperCase();
	}

	/**
	 * Gets locate function.
	 */
	protected String getLocateFunction()
	{
		return "POSITION";
	}

	/**
	 * Gets substring function.
	 */
	protected String getSubstringFunction()
	{
		return "SUBSTR";
	}
} // end class LyPostgresqlRequestEncoder
