/*
 * LyUnionClass.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.union;

import java.util.*;
import leon.app.*;
import leon.app.behaviorinterface.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.*;
import leon.info.daemon.*;
import leon.misc.*;
import leon.notifier.*;
import leon.peer.*;


/**
 * This class is a simple peer class implementation that uses LyUnionObject as wrapper for
 * instances.
 * @see LyUnionObject
 */
public class LyUnionClass
	implements LyClass
{
	public static final String __VERSION =
		"$Id: LyUnionClass.java 35864 2012-06-14 08:20:45Z jlsaouli $";

	protected LyDataSource     _dataSource;

	protected LyUnionBinding   _binding;

	protected LyClassInfo	   _classInfo;

	/** First peer class of the union. */
	protected LyClass		   _class1;

	/** Second peer class of the union. */
	protected LyClass		   _class2;

	/** List of id fields. */
	protected Vector<LyField>	_idFields;

	/** Store objects of class1 during notification of creation. */
	protected Vector<Object>	_class1TemporaryObjects;

	/** Store objects of class2 during notification of creation. */
	protected Vector<Object>	_class2TemporaryObjects;
    
	/** Store the listeners of the 2 classes. */
	protected Vector<_LyUnionClassDataSourceListener>		   _listeners;

	/** List of objects to ignore during events. */
	protected Vector<Object>		   _ignoredObjects;

	/**
	 * Constructs a table class.
	 */
	public LyUnionClass(LyUnionBinding unionBinding, LyClassInfo classInfo)
	{
		_binding = unionBinding;
		_classInfo = classInfo;

		LyClassBinding classBinding1 = unionBinding.getClassBinding1();
		LyClassBinding classBinding2 = unionBinding.getClassBinding2();

		_class1 = classBinding1.getPeerClass();
		_class2 = classBinding2.getPeerClass();

		if ((_class1 == null) || (_class2 == null))
			throw new IllegalStateException("BF AppEngine : invalid union binding on " + classInfo);

		_idFields = null;

		setUnionOrder();

		LyDataSource ds1 = _class1.getDataSource();
		LyDataSource ds2 = _class2.getDataSource();

		_class1TemporaryObjects = new Vector<Object>(1);
		_class2TemporaryObjects = new Vector<Object>(1);

		_listeners = new Vector<_LyUnionClassDataSourceListener>(2);
		_LyUnionClassDataSourceListener listener = new _LyUnionClassDataSourceListener(_class1);

		ds1.addListener(listener);
		_listeners.addElement(listener);

		listener = new _LyUnionClassDataSourceListener(_class2);
		ds2.addListener(listener);
		_listeners.addElement(listener);

		_ignoredObjects = null;
	}

	/**
	 * Clears the content of the vector of temporary object used by this class.
	 */
	public void clearTemporaryObjects()
	{
		_class1TemporaryObjects = new Vector<Object>(1);
		_class2TemporaryObjects = new Vector<Object>(1);
	}
	
	/**
	 * Build the filter on the logical level to load the 2 union physical classes from the
	 * identifier of one physical classes This is used to retreive union object(s) from one of its
	 * parts.
	 * @return the filter
	 */
	@SuppressWarnings("deprecation")
	protected LyFilter buildInfoFilter(Object currentObject, LyClass currentClass)
	{
		LyFilter filter = null;

		if ((currentObject != null) && (currentClass != null))
		{
			// Fetch the values we have for the union
			LyUnionBinding binding = getUnionBinding();
			LyField		   currentField;
			int			   i, n = binding.getValueCount();
			Object		   currentValue;
			LyFieldInfo    fieldInfo;
			int			   j, p = _classInfo.getFieldInfoCount();
			LyFieldBinding fieldInfoBinding;

			// Get the Names of CBIND of the current class
			Vector<String>		   names = new Vector<String>(2);

			getNames(currentClass, names);

			for (i = 0; i < n; i++)
			{
				// Fetch on the physical level
				if (currentClass == _class1)
					currentField = currentClass.getField(binding.getValue1(i));
				else
					currentField = currentClass.getField(binding.getValue2(i));

				currentValue = currentClass.getValue(currentObject, currentField);

				// Link to logical level
				// At least one fieldInfo is related to currentClass / currentField in _classInfo
				// XXX TO DO F_BINDS ???
				for (j = 0; j < p; j++)
				{
					fieldInfo = _classInfo.getFieldInfo(j);
					// Let's compare binding
					fieldInfoBinding = fieldInfo.getBinding();

					if ((fieldInfoBinding != null)
							&& (fieldInfoBinding.getCBind() != null)
							&& (names.contains(fieldInfoBinding.getCBind()))
							&& (fieldInfoBinding.getBinding() != null)
							&& (fieldInfoBinding.getBinding().equals(currentField.getId())))
					{
						// Got it ! Found a criteria for our filter
						if (filter == null)
							filter = new LyFilter(LyFilter.Type.AND);

						LyFilterExpression exp = new LyFilterExpression(fieldInfo,
								LyFilterExpression.Condition.EQUALS, currentValue);

						filter.addElement(exp);
					}
				}
			}
		}
		return filter;
	} // end method buildInfoFilter

	protected void getNames(LyClass lyclass, Vector<String> names)
	{
		if (lyclass instanceof LyUnionClass)
		{
			getNames(((LyUnionClass)lyclass)._class1, names);
			getNames(((LyUnionClass)lyclass)._class2, names);
		}
		else
			names.add(lyclass.getName());
	}

	protected Vector<LyDataSourceListener> getListeners()
	{
        if (_listeners == null)
               return null;
        
		return new Vector<LyDataSourceListener>(_listeners);
	}

	/**
	 * Add an object to the list of ignored objects.
	 * @param object the new object to be ignored
	 */
	protected void addIgnoredObject(Object object)
	{
		if (object == null)
			return;

		if (_ignoredObjects == null)
			_ignoredObjects = new Vector<Object>();

		_ignoredObjects.addElement(object);
	}

	/**
	 * Remove an object from the list of ignored objects.
	 * @param object the object to be removed from the list
	 */
	protected void removeIgnoredObject(Object object)
	{
		if ((object == null) || (_ignoredObjects == null))
			return;

		if (_ignoredObjects.contains(object))
			_ignoredObjects.removeElement(object);
	}

	@SuppressWarnings("deprecation")
	protected void setUnionOrder()
	{
		Vector<String> values1 = new Vector<String>();
		Vector<String> values2 = new Vector<String>();
		int    i, n = _binding.getValueCount();
		String v1, v2;

		for (i = 0; i < n; i++)
		{
			v1 = _binding.getValue1(i);
			v2 = _binding.getValue2(i);

			if (_class1.getField(v1) != null)
			{
				// value1 exists in class1
				if (_class2.getField(v2) != null)
				{
					// value2 exists in class2
					values1.addElement(v1);
					values2.addElement(v2);
				}
				else
				{
					// value2 does not exist in class2
					if (_class2.getField(v1) != null)
					{
						// value1 exists in class2
						if (_class1.getField(v2) != null)
						{
							// value2 exists int class1
							values1.addElement(v2);
							values2.addElement(v1);
						}
						else
							// value2 does not exist int class1
							throw new IllegalStateException("BF AppEngine : invalid union binding on "
								+ _classInfo
								+ " : value2 does'nt exist in the two class binding.Expecting field "
								+ v2 + " in class " + _class1.getId());
					}
					else
						// value1 does not exist int class2
						throw new IllegalStateException("BF AppEngine : invalid union binding on "
							+ _classInfo
							+ " : value and value2 does'nt exist in the second class binding. Expecting field "
							+ v1 + " in class " + _class2.getId());
				}
			}
			else
			{
				// value1 does not exist in class1
				if (_class1.getField(v2) != null)
				{
					// value2 exists int class1
					if (_class2.getField(v1) != null)
					{
						// value1 exists in class2
						values1.addElement(v2);
						values2.addElement(v1);
					}
					else
						// value1 does not exist int class2
						throw new IllegalStateException("BF AppEngine : invalid union binding on "
							+ _classInfo
							+ " : value doesn't exist in the two class binding.Expecting field "
							+ v1 + " in class " + _class2.getId());
				}
				else
					// value2 does not exist in class1
					throw new IllegalStateException("BF AppEngine : invalid union binding on "
						+ _classInfo
						+ " : value and value2 doesn't exist in the first class binding.Expecting field "
						+ v2 + " in class " + _class1.getId());
			}
		}

		// Now we can set the values to the binding.
		// First reset old values.
		_binding.clearUnionList();

		for (i = 0; i < n; i++)
		{
			v1 = values1.elementAt(i);
			v2 = values2.elementAt(i);
			_binding.addUnion(new LyUnion(v1, v2));
		}
	} // end method setUnionOrder

	/**
	 * Gets identifier of the class.
	 * @return Identifier of the class.
	 */
	public String getId()
	{
		// Add hashcode because the same union may be used in different cases
		return hashCode() + "{" + _class1.getId() + "-" + _class2.getId() + "}";
	}

	/**
	 * Gets name of the class. The name of the class is the name in the information system. By
	 * default, if no name is specified, this method returns the id.
	 * @return Name of the class.
	 */
	public String getName()
	{
		return "{" + _class1.getName() + ", " + _class2.getName() + "}";
	}

	/**
	 * Gets the physical representation of the instances : java class that really holds instances of
	 * this class.
	 * @return Java class of the "objects" of this class.
	 */
	public Class<?> getObjectClass()
	{
		return LyUnionObject.class;
	}

	/**
	 *Gets the associated union binding
	 * @return the associated union binding
	 */
	public LyUnionBinding getUnionBinding()
	{
		return _binding;
	}

	/**
	 * Sets the data source of this class.
	 * @param     dataSource Data source of the class.
	 * @exception IllegalStateException If the class has already a data source.
	 */
	public void setDataSource(LyDataSource dataSource)
	{
		if (_dataSource != null)
			throw new IllegalStateException(
				"BF AppEngine : Attempt to set data source on a peer class that has already one");

		_dataSource = dataSource;
	}

	/**
	 * Gets the data source that manages all instances of this class.
	 * @return Data source that manages the instances of this class.
	 */
	public LyDataSource getDataSource()
	{
		return _dataSource;
	}

	/**
	 * Gets the list of fields that identifies instances of this class.
	 * @return List of fields from which an identifier can be calculated.
	 */
	@SuppressWarnings("deprecation")
	public Vector<LyField> getIdFields()
	{
		if (_idFields != null)
			return _idFields;

		Vector<LyField> ids1, ids2;

		_idFields = new Vector<LyField>();

		// Get the id fields of the two classes because sent loaded requests must
		// have the id of all the parts of the union to create valid objets.
		ids1 = new Vector<LyField>(_class1.getIdFields());
		ids2 = new Vector<LyField>(_class2.getIdFields());

		// If the union of the two class is on id peer fields, we can just take
		// one of the two parts.
		int     i, n = _binding.getValueCount();
		String  v1, v2;
		LyField field;

		for (i = 0; i < n; i++)
		{
			v1 = _binding.getValue1(i);
			field = _class1.getField(v1);
			if (ids1.contains(field))
			{
				v2 = _binding.getValue2(i);
				field = _class2.getField(v2);
				if (ids2.contains(field))
					ids2.removeElement(field);
			}
		}

		_idFields.addAll(ids1);
		_idFields.addAll(ids2);

		return _idFields;
	} // end method getIdFields

	/**
	 * Registers a peer field as an id field on this class.
	 * @param field Peer field to register as id field.
	 */
	public void addIdField(LyField field)
	{
		// Nothing to do in this implementation.
	}

	/**
	 * Gets the list of fields of this class.
	 * @return List of all fields of this class.
	 */
	public Vector<LyField> getFields()
	{
		Vector<LyField> fields = new Vector<LyField>();

		fields.addAll(_class1.getFields());
		fields.addAll(_class2.getFields());

		return fields;
	}

	/**
	 * Retrieves a field on this class by its identifier.
	 * @param  id Identifier of the searched field.
	 * @return The field of this class which has the specified identifier, null if no field is
	 *         found.
	 */
	public LyField getField(String id)
	{
		LyField field = _class1.getField(id);

		if (field != null)
			return field;
		else
			return _class2.getField(id);
	}

	/**
	 * Gets identifier for an instance of this class.
	 * @param  object Instance of this class.
	 * @return Identifier of the given object.
	 */
	public LyObjectId getObjectId(Object object)
	{
		return ((LyUnionObject)object).getObjectId();
	}

	/**
	 * Gets the value corresponding to the given field for an instance of this class.
	 * @param  object Instance of this class.
	 * @param  field  Field for which a value is requested.
	 * @return Value of the given field on the specified instance.
	 */
	public Object getValue(Object object, LyField field)
	{
		return ((LyUnionObject)object).getValue(field);
	}

	/**
	 * Sets the value corresponding to the given field on an instance of this class.
	 * @param object Instance of this class on which a value is set.
	 * @param field  Field for which a value is set.
	 * @param value  Value set for the given field on the specified instance.
	 */
	public void setValue(Object object, LyField field, Object value)
	{
		((LyUnionObject)object).setValue(field, value);
	}

	/**
	 * Copies all field values of an instance of this class on another instance.
	 * @param  target  Object on which values are copied.
	 * @param  pattern Object from which values are get.
	 * @return Field values that are copied from object2 to object1.
	 */
	public LyValueSet copy(Object target, Object pattern)
	{
		Object     targetObject = ((LyUnionObject)target).getObject1();
		Object     patternObject = ((LyUnionObject)pattern).getObject1();

		LyValueSet values1 = _class1.copy(((LyUnionObject)target).getObject1(),
				((LyUnionObject)pattern).getObject1());

		if (targetObject != patternObject)
		{
			LyDataSource ds = _class1.getDataSource();

			if (ds.checkObject(patternObject))
				((LyUnionObject)target)._object1 = patternObject;
		}

		targetObject = ((LyUnionObject)target).getObject2();
		patternObject = ((LyUnionObject)pattern).getObject2();

		LyValueSet values2 = _class2.copy(((LyUnionObject)target).getObject2(),
				((LyUnionObject)pattern).getObject2());

		if (targetObject != patternObject)
		{
			LyDataSource ds = _class2.getDataSource();

			if (ds.checkObject(patternObject))
				((LyUnionObject)target)._object2 = patternObject;
		}

		if ((values1 == null) && (values2 == null))
			return null;
		else
		{
			LyValueSet values = new LyValueSet();

			if (values1 != null)
			{
				// copy all value from values1 to values
				for (int i = 0; i < values1.getSize(); i++)
				{
					LyValue value = values1.getFieldValue(i);

					values.addValue(value);
				}
			}
			if (values2 != null)
			{
				// copy all value from values1 to values
				for (int i = 0; i < values2.getSize(); i++)
				{
					LyValue value = values2.getFieldValue(i);

					values.addValue(value);
				}
			}

			return values;
		}
	} // end method copy

	/**
	 * Gets all field values from an instance of this class.
	 * @param  object Instance of this class.
	 * @return All the field values of the given instance.
	 */
	public LyValueSet getValues(Object object)
	{
		return ((LyUnionObject)object).getValues();
	}

	/**
	 * Creates a new instance of this class.
	 * @return    New instance of this class.
	 * @exception IllegalAccessException If the class or initializer is not accessible.
	 * @exception InstantiationException If an application tries to instantiate an abstract class or
	 *                                   an interface, or if the instantiation fails for some other
	 *                                   reason.
	 */
	public Object newInstance()
		throws InstantiationException,
			IllegalAccessException
	{
		return (new LyUnionObject(this));
	}

	/**
	 * This utility method enables to group fields that belongs to a same sub-class of the current
	 * union class. The returned array actually contains two vectors, the first for _class1 ant
	 * the second for _class2.
	 * @param  fields a list of fields of the current union class.
	 * @return an array of vector, each containing a list of fields belonging to the same sub-class.
	 */
	protected Vector<LyField>[] splitFields(Vector<? extends LyField> fields)
	{
		@SuppressWarnings("unchecked")
		Vector<LyField>[] result = new Vector[2];

		if (fields == null)
			return result;

		result[0] = new Vector<LyField>();
		result[1] = new Vector<LyField>();
        
		// check for each field if it belongs to class 1 or class 2
		LyField field;
		Vector<LyField>  fields1, fields2;

		fields1 = new Vector<LyField>(_class1.getFields());
		fields2 = new Vector<LyField>(_class2.getFields());

		for (int i = 0; i < fields.size(); i++)
		{
			field = fields.elementAt(i);
			if (fields1.contains(field))
				result[0].addElement(field);
			else if (fields2.contains(field))
				result[1].addElement(field);
		}

		return result;
	}

	/**
	 * This utility method split a sort on the current union class to an array of sub-sorts, each
	 * corresponding to a sub-class of the union class.
	 * @param  sort a sort defined on the current union class.
	 * @return an array of sorts, each defined on a sub-class of the current union class.
	 */
	public LySort[] splitSort(LySort sort)
	{
		LySort[] result = new LySort[2];

		if (sort == null)
			return result;

		// split the sort into two sub-sorts
		LyFieldInfo    fieldInfo;
		LyFieldBinding fbinding = null;
		String		   cbind;

		result[0] = new LySort();
		result[1] = new LySort();
		// watch all criteria of the filter
		for (int i = 0; i < sort.getCriterionCount(); i++)
		{
			fieldInfo = sort.getCriterion(i).getFieldInfo();
			fbinding = fieldInfo.getBinding();
			if (fbinding != null)
			{
				cbind = fbinding.getCBind();
				if ((cbind == null)
						|| _class1.getName().equals(cbind))
					result[0].addCriterion(sort.getCriterion(i));
				else if (_class2.getName().equals(cbind))
					result[1].addCriterion(sort.getCriterion(i));

			}
		}
		// return sorts
		return result;
	} // end method splitSort

	/**
	 * This utility method split a filter on the current union class to an array of sub-filters,
	 * each corresponding to a sub-class of the union class.
	 * @param  filter a filter defined on the current union class.
	 * @return an array of filters, each defined on a sub-class of the current union class.
	 */
	public LyFilter[] splitFilter(LyFilter filter)
	{
		LyFilter[] result = new LyFilter[2];

		if (filter == null)
			return result;

		// create filter objects
		result[0] = new LyFilter(filter.getType());
		result[1] = new LyFilter(filter.getType());

		// split the filter into two sub filters
		LyFilterElement element;
		LyFieldInfo     fieldInfo;
		LyFieldBinding  fbinding = null;
		String		    cbind, fbind;
		Vector<String>  values;
		int			    index;
		LyInfoDaemon    daemon;

		// watch all elements of the filter
		for (int i = 0; i < filter.getSize(); i++)
		{
			element = filter.getElement(i);
			
			// case of a filter
			if (element instanceof LyFilter)
			{
				LyFilter[] subFilters = splitFilter((LyFilter)element);

				for (int k = 0; k < subFilters.length; k++)
					if ((subFilters[k] != null) && (subFilters[k].getSize() > 0))
						result[k].addElement(subFilters[k]);
			}
			// case of a filter expression
			else if (element instanceof LyFilterExpression)
			{
				fieldInfo = ((LyFilterExpression)element).getFieldInfo();
				fbinding = fieldInfo.getBinding();
				if (fbinding != null)
				{
					cbind = fbinding.getCBind();
					fbind = fbinding.getBinding();
					index = -1;
					values = null;
					if ((cbind == null)
						|| _class1.getName().equals(cbind))
					{
						result[0].addElement(element);
						values = _binding.getValues1();
						if (values.contains(fbind))
						{
							// Attempt to build another expression with the fieldinfo
							// corresponding to the other part of the union.
							index = values.indexOf(fbind);
							values = _binding.getValues2();
							cbind = _class2.getName();
							fbind = (String)values.elementAt(index);
							index = 1;
						}
					}
					else if (_class2.getName().equals(cbind))
					{
						result[1].addElement(element);
						values = _binding.getValues2();
						if (values.contains(fbind))
						{
							// Attempt to build another expression with the fieldinfo
							// corresponding to the other part of the union.
							index = values.indexOf(fbind);
							values = _binding.getValues1();
							cbind = _class1.getName();
							fbind = (String)values.elementAt(index);
							index = 0;
						}
					}

					if (index != -1)
					{
						String	fieldCBind;
						
						for (int j = 0; j < _classInfo.getFieldInfoCount(); j++)
						{
							fieldInfo = _classInfo.getFieldInfo(j);
							fbinding = fieldInfo.getBinding();
							if (fbinding == null)
								continue;
							
							fieldCBind = fbinding.getCBind();
							if (fieldCBind == null)
								fieldCBind = _class1.getName();
							
							if ((fbinding != null)
									&& fieldCBind.equals(cbind)
									&& fbinding.getBinding().equals(fbind))
							{
								element = ((LyFilterExpression)element).cloneFilterExpression();
								((LyFilterExpression)element).setFieldInfo(fieldInfo);
								result[index].addElement(element);
							}
						}
					}
				}
				else if (fieldInfo instanceof LyRelationFieldInfo)
				{
				    daemon = ((LyRelationFieldInfo)fieldInfo).getDaemon();
				    if (daemon instanceof LyJoinDaemon)
				    {
	                    LyClassInfo    joinClassInfo = ((LyJoinDaemon)daemon).getJoinClass();
	                    LyClassBinding joinClassBinding = joinClassInfo.getBinding();
	                    
				        if (_binding.getClassBinding1().getLocation() == joinClassBinding.getLocation())
				            result[0].addElement(element);
				        else if (_binding.getClassBinding2().getLocation() == joinClassBinding.getLocation())
				            result[1].addElement(element);
				    }
				    else if (daemon instanceof LyReverseDaemon)
				    {
			            LyClassInfo    targetClassInfo = ((LyRelationFieldInfo)fieldInfo).getClassInfo(0);
			            LyClassBinding classBinding = targetClassInfo.getBinding();
			            
                        if (_binding.getClassBinding1().getLocation() == classBinding.getLocation())
                            result[0].addElement(element);
                        else if (_binding.getClassBinding2().getLocation() == classBinding.getLocation())
                            result[1].addElement(element);
				    }
				}
			}
		}
		// return the sub filters
		return result;
	} // end method splitFilter

	/**
	 * Returns a display string.
	 */
	public String toString()
	{
		return getName() + "(" + getId() + ")";
	}

	/**
	 * Recompute the object id of the specified object.
	 */
	public void rehash(Object object)
	{
		LyDataSource dataSource = getDataSource();
		LyCache		 cache = dataSource.getCache();
		LyObjectId   oldObjectId = getObjectId(object);

		((LyUnionObject)object).setObjectId(null);
		cache.rehash(object, oldObjectId);
	}

	class _LyUnionClassDataSourceListener
		implements LyDataSourceListener
	{
		LyClass _class;
		Vector<LyObjectId>  _reloadingIds;

		_LyUnionClassDataSourceListener(LyClass subClass)
		{
			_class = subClass;
			_reloadingIds = null;
		}

		/**
		 * Retrieve union object from one of its sub-object.
		 */
		LyUnionObject getUnionObject(Object subObject)
		{
			Vector<Object>	objects = _dataSource.getCache().getObjects();
			int			  i, n = ((objects == null) ? 0 : objects.size());
			LyUnionObject object;

			for (i = 0; i < n; i++)
			{
				object = (LyUnionObject)objects.elementAt(i);
				if ((object.getObject1() == subObject)
						|| (object.getObject2() == subObject))
					return object;

			}

			return null;
		}

		@SuppressWarnings("deprecation")
		Object getAssociatedObject(Object currentObject, LyClass currentClass,
			Vector<Object> associatedObjects, LyClass associatedClass)
		{
			if (associatedObjects.size() == 0)
				return null;

			LyUnionBinding binding = getUnionBinding();
			LyField		   associatedField, currentField;
			int			   i, n = binding.getValueCount();
			Object		   associatedValue, currentValue;
			Vector<Object>		   values = new Vector<Object>();

			for (i = 0; i < n; i++)
			{
				if (currentClass == _class1)
					currentField = currentClass.getField(binding.getValue1(i));
				else
					currentField = currentClass.getField(binding.getValue2(i));

				currentValue = currentClass.getValue(currentObject, currentField);
				values.add(currentValue);
			}

			int    j, no = associatedObjects.size();
			Object associatedObject = null;

			for (j = 0; j < no; j++)
			{
				associatedObject = associatedObjects.elementAt(j);
				for (i = 0; i < n; i++)
				{
					currentValue = values.elementAt(i);
					if (associatedClass == _class1)
						associatedField = associatedClass.getField(binding.getValue1(i));
					else
						associatedField = associatedClass.getField(binding.getValue2(i));

					associatedValue = associatedClass.getValue(associatedObject, associatedField);
					if ((associatedValue != currentValue)
							&& ((associatedValue == null)
								|| (currentValue == null)
								|| !associatedValue.equals(currentValue)))
						break;
				}
				if (i == n)
					break;
			}

			if (j == no)
				return null;

			associatedObjects.removeElement(associatedObject);

			// Should be unlocked when union object is created
			return associatedObject;
		} // end method getAssociatedObject

		/**
		 * Method for receiving an object creation.
		 *
		 * @param  event Event notifying the creation of an object.
		 * @return True indicating if the event was used, false otherwise.
		 */
		public boolean objectCreated(LyDataSourceEvent event)
		{
			if (event.isLoading())
				return false;
			
			Object object = event.getObject();
			Object associatedObject = null;

			if (_class == _class1)
			{
				associatedObject = getAssociatedObject(object, _class1, _class2TemporaryObjects, _class2);
				if (associatedObject == null)
				{
					_class1TemporaryObjects.add(object);
					return false;
				}
			}
			else
			{
				associatedObject = getAssociatedObject(object, _class2, _class1TemporaryObjects, _class1);
				if (associatedObject == null)
				{
					_class2TemporaryObjects.add(object);

					// Force the complementary load if notified
					boolean notifying = isNotifying(event, true);

					if (notifying && (!isGroupedEvent(event)))
						reloadAssociatedObject(object, null);

					return false;
				}
			}

			// associated object is not null
			LyUnionObject finalObject;

			if (_class == _class1)
				finalObject = new LyUnionObject(LyUnionClass.this, object, associatedObject);
			else
				finalObject = new LyUnionObject(LyUnionClass.this, associatedObject, object);

			String groupId = event.getGroupId();

			_dataSource.receiveEvent(groupId, LyEventNotifier.EventType.CREATE, finalObject, null);

			return true;
		} // end method objectCreated

		/**
		 * Check if event is notified by JMS.
		 *
		 * @param  LyDataSourceEvent event event received
		 * @param  checkLoading      also check if the event is loading
		 * @return true if notified
		 */
		protected boolean isNotifying(LyDataSourceEvent event, boolean checkLoading)
		{
			return event.isNotified();
		}

		/**
		 * Check if event is notified by JMS within a grouped message (waiting for additional
		 * messages).
		 *
		 * @param  LyDataSourceEvent event event received
		 * @param  checkLoading      also check if the event is loading
		 * @return true if notified
		 */
		protected boolean isGroupedEvent(LyDataSourceEvent event)
		{
			// Check if the notifier is notifying an event with no values
			LyDataSource    source = event.getSource();
			LyDataProvider  provider = ((source != null) ? source.getDataProvider() : null);
			LyEventNotifier notifier = ((provider != null) ? provider.getEventNotifier() : null);

			if (notifier != null)
				return (notifier.isGroupedEvent());

			return false;
		}

		/**
		 * Method to force the reload of associated object One object of the bind method.
		 */
		public void reloadAssociatedObject(Object object, LyValueSet values)
		{
			LyObjectId objId = _class.getObjectId(object);

			if (_reloadingIds == null)
				_reloadingIds = new Vector<LyObjectId>();
			else
			{
				// Check if we are already reloading for this part (avoid loops)
				if (_reloadingIds.contains(objId))
				{
					// LOOP DETECTED ! Return
					_classInfo.getEnvironment().logError(
						"[LyUnionClass] loop detected in reloadAssociatedObject for "
						+ objId + " in class " + _class.getId()
						+ "(name : " + _class.getName() + ")");
					if (_classInfo.getEnvironment().isDebugLevel(LyEnvironment.DEBUG_TRACE))
						Thread.dumpStack();
					return;
				}
			}

			// first try to get the object with the behavior
			Object result = getObject(objId, values);

			// if object not found try old method
			if (result == null)
			{
				_reloadingIds.addElement(objId);

				LyFilter	   filter = buildInfoFilter(object, _class);

				LyObjectSource objectSource = _classInfo.getObjectSource();

				objectSource.loadObjects(null, null, filter, -1);

				_reloadingIds.removeElement(objId);

				if (_reloadingIds.size() == 0)
					_reloadingIds = null;
			}
		} // end method reloadAssociatedObject

		/**
		 * Gets an object from this object source by the specified identifier.
		 *
		 * @param  id     Object identifier of the searched object.
		 * @param  values a set of values when the object is modified.
		 * @return The requested object if it was found, or null if it could not be found. This
		 *         object is an instance of LyObject (or an inherited class).
		 */
		public Object getObject(LyObjectId id, LyValueSet values)
		{
			LyApplication				   application = _classInfo.getApplication();
			LyApplicationBehaviorInterface behavior = application.getApplicationBehavior();

			if (behavior == null)
				return null;

			// Try on behavior first
			LyInfoList<LyClassInfo> subClasses = behavior.getLogicalClasses(application, id, values);

			// If not found do not test generic classes
			if (subClasses == null)
				return null;

			LyClassInfo    classInfo;
			int			   i, n = subClasses.getSize();
			LyObjectId     objectId;
			LyObjectSource objectSource;
			String		   str = id.toString();
			Object		   result;

			for (i = 0; i < n; i++)
			{
				classInfo = subClasses.getInfo(i);
				objectId = new LyObjectId(classInfo, str);
				objectSource = application.getObjectSource(classInfo);
				result = objectSource.getObject(objectId);
				if (result != null)
					return result;
			}

			return null;
		} // end method getObject

		/**
		 * Method for receiving an object modification.
		 *
		 * @param  event Event notifying the modification of an object.
		 * @return True indicating if the event was used, false otherwise.
		 */
		public boolean objectModified(LyDataSourceEvent event)
		{
			if (event.isLoading())
				return false;
			
			Object object = event.getObject();

			if ((_ignoredObjects != null) && (_ignoredObjects.contains(object))
					&& (!event.isNotified()))
				return false;

			LyValueSet    values = event.getValues();
			LyUnionObject parent = getUnionObject(object);

			if (parent != null)
			{
				String groupId = event.getGroupId();

				_dataSource.receiveEvent(groupId, LyEventNotifier.EventType.UPDATE, parent, values);
				return true;
			}
			else
			{
				// Check if notifing : we may need to reload second part
				boolean notifying = isNotifying(event, false);

				if (notifying)
					reloadAssociatedObject(object, null);

				return false;
			}
		} // end method objectModified

		/**
		 * Method for receiving an object deletion.
		 *
		 * @param  event Event notifying the deletion of an object.
		 * @return True indicating if the event was used, false otherwise.
		 */
		public boolean objectDeleted(LyDataSourceEvent event)
		{
			Object object = event.getObject();

			if ((_ignoredObjects != null) && (_ignoredObjects.contains(object))
					&& (!event.isNotified()))
				return false;

			LyUnionObject parent = getUnionObject(object);

			if (parent != null)
			{
				String groupId = event.getGroupId();

				_dataSource.receiveEvent(groupId, LyEventNotifier.EventType.DELETE, parent, null);
				return true;
			}
			else
				return false;
		}
	} // end class _LyUnionClassDataSourceListener

	/**
	 * Accessor to the first class of the union
	 * @return First peer class of the union
	 */
	public LyClass getClass1()
	{
		return _class1;
	}

	/**
	 * Accessor to the second class of the union
	 * @return Second peer class of the union
	 */
	public LyClass getClass2()
	{
		return _class2;
	}
} // end class LyUnionClass
