/*
 * LyUnionDataProvider.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.union;

import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.*;
import leon.info.location.*;
import leon.misc.*;
import leon.peer.*;


/**
 * This class is used by DataSources to transmit their requests to an instance of RequestEncoder.
 * @version $Id: LyUnionDataProvider.java 35854 2012-06-13 16:43:14Z jlsaouli $
 * @author  Lyria
 * @see     leon.peer.info.LyInfoRequestEncoder
 */
public class LyUnionDataProvider extends LyDataProvider
{
	public static final String __VERSION =
		"$Id: LyUnionDataProvider.java 35854 2012-06-13 16:43:14Z jlsaouli $";

	/**
	 * Default constructor.
	 * @param application Current application instance
	 * @param location    This value is obtained from the location parameter in the meta model.
	 */
	public LyUnionDataProvider(LyApplication application, LyLocation location)
	{
		super(application, location);
	}

	/**
	 * Default constructor.
	 * @param      application Current application instance
	 * @param      id          Identifier of this dataProvider. This value is obtained from the
	 *                         location parameter in the meta model.
	 * @deprecated should use LyUnionDataProvider(LyApplication application, LyLocation location)
	 *             instead of this constructor.
	 */
	public LyUnionDataProvider(LyApplication application, String id)
	{
		super(application, id);
	}

	/**
	 * Gets request encoder for a given data source.
	 * @param dataSource Data source for which a request encoder is requiered.
	 */
	public LyRequestEncoder getRequestEncoder(LyDataSource dataSource)
	{
		return new LyUnionRequestEncoder(this, dataSource);
	}

	/**
	 * This class is a factory to create a LyConnection instance.
	 * @return a new Connection instance. The connection instance is used in the specific LyRequest
	 *         encoder implementation.
	 */
	protected LyConnection createConnection()
	{
		return null;
	}

	/**
	 * This method is used to build all the class peers for all the classInfos of the location
	 * managed by this provider.
	 * @param  classInfos Classes to validate.
	 * @return True if the validation was successful, false otherwise.
	 */
	public boolean validate(LyInfoList<LyClassInfo> classInfos)
	{
		// Creating the class and fields peers.
		int			   i, n = classInfos.getSize();
		LyClassInfo    classInfo;
		LyClassBinding classBinding;

		for (i = 0; i < n; i++)
		{
			classInfo = classInfos.getInfo(i);
			if (classInfo.isVirtual())
				continue;

			classBinding = classInfo.getBinding();
			if (!(classBinding instanceof LyUnionBinding))
				continue;

			// The classInfo has the right provider id, validate it.
			validate(classInfo);
		}

		return true;
	}

	/**
	 * Validates the given class info : this mainly consists to create a peer class for each
	 * encountered class binding.
	 * @param  classInfo Class info that is validated.
	 * @return True if the validation was successful, false otherwise.
	 */
	protected void validate(LyClassInfo classInfo)
	{
		// Don't revalidate classInfos which are already validate
		try
		{
			if (_application.getDataSource(classInfo) != null)
				// Already validated
				return;
		}
		catch (Exception e)
		{
		}

		LyUnionBinding unionBinding = (LyUnionBinding)classInfo.getBinding();

		if (classInfo.isAbstract())
		    validate(unionBinding, classInfo);
		else
		{
			// First we create the objectSource and we associate the infoProvider
			LyObjectSource		objectSource;
			LyDataSource		dataSource;

			objectSource = new LyObjectSource(_application.getInfoProvider(), classInfo);
			dataSource = validate(unionBinding, classInfo);

			// The objectSource must listen to dataSource events.
			objectSource.listenTo(dataSource);
		}
	}

	/**
	 * Validates given union binding.
	 */
	protected LyDataSource validate(LyUnionBinding unionBinding, LyClassInfo classInfo)
	{
		LyClassBinding binding1 = unionBinding.getClassBinding1();
		LyClassBinding binding2 = unionBinding.getClassBinding2();

		if (binding1 instanceof LyUnionBinding)
			validate((LyUnionBinding)binding1, classInfo);
		if (binding2 instanceof LyUnionBinding)
			validate((LyUnionBinding)binding2, classInfo);

		// Creating the peer class.
		LyUnionClass unionClass = new LyUnionClass(unionBinding, classInfo);
		LyDataSource dataSource = null;

		if (!classInfo.isAbstract())
		{
			// Creating the dataSource.
			dataSource = new LyPeerDataSource(this, unionClass);
			unionClass.setDataSource(dataSource);
		}

		// Associate the binding and the class peer.
		unionBinding.setPeerClass(unionClass);

		return dataSource;
	}

	/**
	 * Validates the field info of the given class info : for this data provider, this consit to
	 * retrieve and store id fields of the union class.
	 * @param  classInfo Class info to which the field info belongs to.
	 * @param  fieldInfo Field info that is validated.
	 * @return True if the validation was successful, false otherwise.
	 */
	protected void validate(LyClassInfo classInfo, LyFieldInfo fieldInfo)
	{
		// Nothing to do.
	}

	/**
	 * Determines the field is an identifier on the data provider Internal use of the validate
	 * method.
	 * <P>Default return true if the mark ID is set.</P>
	 * @param  unionClass the union class reference as specified in validate method.
	 * @param  field      peer field associated with the value we store.
	 * @param  fieldInfo  data description of the meta model.
	 * @return true or false.
	 * @see    #validate(LyClassInfo classInfo, LyFieldInfo fieldInfo)
	 */
	protected boolean isIdField(LyUnionClass unionClass, LyField field, LyFieldBinding fieldBinding)
	{
		return ((fieldBinding == null) ? false : fieldBinding.isId());
	}

	/**
	 * Adds the peer field as an id of the given peer class. The default implementation calls
	 * LyPeerClass.addIdField.
	 * @param unionClass Union class to which the pier field is added as id field.
	 * @param field      Peer field added as id field.
	 * @see   LyPeerClass#addField
	 */
	protected void addIdPeerField(LyUnionClass unionClass, LyField field)
	{
		unionClass.addIdField(field);
	}

	/**
	 * This method is used to build caches on datasources used in union if they have not one yet.
	 * @param classInfos List of classes for which caches are checked.
	 */
	public void completeCaches(LyInfoList<LyClassInfo> classInfos)
	{
		// Creating the class and fields peers.
		int			   i, n = classInfos.getSize();
		LyClassInfo    classInfo;
		LyClassBinding classBinding, classBinding1, classBinding2;
		LyUnionBinding unionBinding;
		LyObjectSource objectSource;
		LyProject	   project = _application.getProject();
		LyLocation     location, location2;
		String		   locationId;
		int			   nbloc = project.getLocationCount();
		LyInfoCache    cachePolicy;
		LyEnvironment  environment = _application.getEnvironment();
		boolean		   locationIsPersistent, location2IsPersistent;

		for (i = 0; i < n; i++)
		{
			classInfo = classInfos.getInfo(i);
			classBinding = classInfo.getBinding();
			if (!(classBinding instanceof LyUnionBinding))
				continue;

			if (classInfo.isAbstract())
			    continue;

			// The classInfo has the right provider id, validate it.
			unionBinding = (LyUnionBinding)classBinding;

			// Create cache for the object source.
			objectSource = classInfo.getObjectSource();
			cachePolicy = classInfo.getCachePolicy();
			
			classBinding1 = unionBinding.getClassBinding1();
			classBinding2 = unionBinding.getClassBinding2();

			if ((cachePolicy == null)
				|| cachePolicy.isDefault())
			{
				// Get first location.
				locationId = (classBinding1 == null) ? null : classBinding1.getLocation();
				if (locationId == null)
					locationId = _application.getDefaultLocationId();
				location = (LyLocation)_application.getInfo(locationId);
				if (classBinding1 instanceof LyUnionBinding)
					locationIsPersistent = true;
				else
					locationIsPersistent = location.isPersistent();
	
				// Get second location.
				locationId = (classBinding2 == null) ? null : classBinding2.getLocation();
				if (locationId == null)
					locationId = _application.getDefaultLocationId();
				location2 = (LyLocation)_application.getInfo(locationId);
				if (classBinding2 instanceof LyUnionBinding)
					location2IsPersistent = true;
				else
					location2IsPersistent = location2.isPersistent();
	
				// Set the object source cache.
				if ((nbloc > 0)
						&& ((!locationIsPersistent)
							|| (!location2IsPersistent)))
				{
					// If location can't save update on objects, we must use full caches.
					cachePolicy = new LyInfoCache(LyInfoCache.CachePolicy.FULL, null);
					environment.logWarning("Forcing cache type to FULL for classInfo " + classInfo
						+ " because one of the data provider of the union is not persistent");
				}
				else if (environment.testEnv("LY_SAVE_FILES") && (nbloc > 0))
				{
					if ((location instanceof LyFileLocation)
							|| (location2 instanceof LyFileLocation))
					{
						// If we want to save data at the end of the application,
						// we must use full caches.
						cachePolicy = new LyInfoCache(LyInfoCache.CachePolicy.FULL, null);
						environment.logWarning("Forcing cache type to FULL for classInfo "
							+ classInfo
							+ " because one of the data provider of the union is a file data provider");
					}
					else
						cachePolicy = new LyInfoCache(LyInfoCache.CachePolicy.AUTO, null);
				}
				else
					cachePolicy = new LyInfoCache(LyInfoCache.CachePolicy.AUTO, null);
			}
			
			objectSource.setCache(cachePolicy);

			// Now for sub-bindings
			completeCaches(classBinding1);
			completeCaches(classBinding2);
		}
	} // end method completeCaches

	/**
	 * This method is used to build caches on datasources used in union if they have not one yet.
	 */
	public void completeCaches(LyClassBinding classBinding)
	{
		if (classBinding == null)
			return;
		
		LyClass		 peerClass = classBinding.getPeerClass();
		LyDataSource dataSource = peerClass.getDataSource();

		if (dataSource.getCache() != null)
			return;

		LyLocation    location;
		String		  locationId;
		LyProject     project = _application.getProject();
		int			  nbloc = project.getLocationCount();
		LyInfoCache   cachePolicy;
		LyEnvironment environment = _application.getEnvironment();

		locationId = classBinding.getLocation();
		if (locationId == null)
			locationId = _application.getDefaultLocationId();
		location = (LyLocation)_application.getInfo(locationId);

		if (classBinding instanceof LyUnionBinding)
			cachePolicy = new LyInfoCache(LyInfoCache.CachePolicy.AUTO, null);
		else if ((nbloc > 0) && (!location.isPersistent()))
			// If location can't save update on objects, we must use full caches.
			cachePolicy = new LyInfoCache(LyInfoCache.CachePolicy.FULL, null);
		else if (environment.testEnv("LY_SAVE_FILES") && (nbloc > 0))
		{
			if (location instanceof LyFileLocation)
				// If we want to save data at the end of the application,
				// we must use full caches.
				cachePolicy = new LyInfoCache(LyInfoCache.CachePolicy.FULL, null);
			else
				cachePolicy = new LyInfoCache(LyInfoCache.CachePolicy.AUTO, null);
		}
		else
			cachePolicy = new LyInfoCache(LyInfoCache.CachePolicy.AUTO, null);

		dataSource.setCache(cachePolicy);

		if (classBinding instanceof LyUnionBinding)
		{
			completeCaches(((LyUnionBinding)classBinding).getClassBinding1());
			completeCaches(((LyUnionBinding)classBinding).getClassBinding2());
		}
	} // end method completeCaches
} // end class LyUnionDataProvider
