/*
 * LyUnionObject.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.union;

import java.util.*;
import leon.data.*;
import leon.info.binding.*;


/**
 * This class implements instances of LyUnionClass.
 */
public class LyUnionObject
{
	public static final String __VERSION =
		"$Id: LyUnionObject.java 35841 2012-06-12 16:11:13Z jlsaouli $";

	/** Union class to which the union object "belongs to". */
	protected LyUnionClass     _class;

	/** Sub-objects. */
	protected Object		   _object1;
	protected Object		   _object2;

	/** Object id of this instance. */
	protected LyObjectId	   _objectId;

	/**
	 * Constructs a new object for the given class.
	 * @param unionClass Union class to which this object belongs to.
	 */
	public LyUnionObject(LyUnionClass unionClass)
	{
		_class = unionClass;
		_object1 = null;
		_object2 = null;
	}

	/**
	 * Constructs a new object for the given class with two given sub-objects.
	 * @param unionClass Union class to which this object belongs to.
	 * @param object1    the first sub-object
	 * @param object2    the second sub-object
	 */
	public LyUnionObject(LyUnionClass unionClass, Object object1, Object object2)
	{
		_class = unionClass;

		_object1 = object1;
		_object2 = object2;
	}

	public LyUnionClass getUnionClass()
	{
		return _class;
	}

	/**
	 */
	public Object getObject1()
	{
		if (_object1 == null)
		{
			try
			{
				_object1 = _class._class1.newInstance();
			}
            catch (InstantiationException e)
            {
                throw new RuntimeException(e);
            }
            catch (IllegalAccessException e)
            {
                throw new RuntimeException(e);
            }
		}
		return _object1;
	}

	/**
	 */
	public Object getObject2()
	{
		if (_object2 == null)
		{
			try
			{
				_object2 = _class._class2.newInstance();
			}
            catch (InstantiationException e)
            {
                throw new RuntimeException(e);
            }
            catch (IllegalAccessException e)
            {
                throw new RuntimeException(e);
            }
		}
        
		return _object2;
	}

	/**
	 * Gets object id of this instance.
	 * @return Object identifier of this instance.
	 */
	protected LyObjectId getObjectId()
	{
		if (_objectId != null)
			return _objectId;

		Vector<LyField>  fields = _class.getIdFields();
		int              i, n = fields.size();
		Object           result;
		LyField          field;

		if (n == 1)
		{
			// The identifier of this object is the value of the field.
			field = fields.get(0);
			result = getValue(field);
			if (result instanceof LyObject)
				result = ((LyObject)result).getObjectId();
		}
		else
		{
			// The identifier of object is an array of values of all id fields.
			result = new Object[n];

			for (i = 0; i < n; i++)
			{
				field = fields.get(i);
				((Object[])result)[i] = getValue(field);
				if (((Object[])result)[i] instanceof LyObject)
					((Object[])result)[i] = ((LyObject)((Object[])result)[i]).getObjectId();
			}
		}

		if (result != null)
			_objectId = new LyObjectId(_class, result);
		return _objectId;
	} // end method getObjectId

	/**
	 * Sets object id for this instance.
	 * @param objectId Object identifier for this instance.
	 */
	protected void setObjectId(LyObjectId objectId)
	{
		_objectId = objectId;
	}

	/**
	 * Gets value for the specified column field.
	 * @param  columnField Column field for which a value is set on this object.
	 * @return Value corresponding to this field.
	 */
	public Object getValue(LyField field)
	{
		Object object;

		// Should never happen (probably an error in the metamodel...
		if (field == null)
			return null;

		Vector<? extends LyField>  class1Fields = _class._class1.getFields();

		if ((class1Fields != null) && (class1Fields.contains(field)))
		{
			object = getObject1();
			return _class._class1.getValue(object, field);
		}
		else
		{
			object = getObject2();
			return _class._class2.getValue(object, field);
		}
	}

	/**
	 * Sets value of the specified column field.
	 * @param columnField Column field for which a value is set on this object.
	 * @param value       Value corresponding to this field.
	 */
	public void setValue(LyField field, Object value)
	{
		Object object;
		Vector<? extends LyField>	class1Fields = _class._class1.getFields();

		if ((class1Fields != null) && (class1Fields.contains(field)))
		{
			// if a key field of class 1 is set, we have to propagate to class2 object
			object = getObject1();
			_class._class1.setValue(object, field, value);

			LyField		   currentField = _class._class1.getField(field.getId());
			LyUnionBinding unionBinding = _class.getUnionBinding();
			Vector<String> keyFields1 = unionBinding.getValues1();

			int			   pos = keyFields1.indexOf(currentField.getId());

			if (pos >= 0)
			{
				// we need to propagate
				object = getObject2();
				LyField keyField2 = _class._class2.getField(unionBinding.getValue2(pos));

				_class._class2.setValue(object, keyField2, value);
			}
		}
		else
		{
			object = getObject2();
			_class._class2.setValue(object, field, value);
		}
	} // end method setValue

	/**
	 * Gets all values of this object.
	 * @return Values set corresponding to all column fields for this object.
	 */
	public LyValueSet getValues()
	{
		Object     object1 = getObject1();
		Object     object2 = getObject2();
		LyValueSet values1 = _class._class1.getValues(object1);
		LyValueSet values2 = _class._class2.getValues(object2);

		if (values1 == null)
			return values2;
		else
		{
			values1.addValues(values2);
			return values1;
		}
	}

	/**
	 */
	protected void validate()
	{
	}
} // end class LyUnionObject
