/*
 * LyUnionRequestEncoder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.union;

import java.util.*;
import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.*;
import leon.peer.*;
import leon.peer.util.*;


/**
 * This class is used to send requests and to receive answers.
 *
 * @version $Id: LyUnionRequestEncoder.java 41563 2012-11-19 13:44:17Z jlsaouli $
 * @author  Lyria
 * @see     LyRequestEncoder
 */
public class LyUnionRequestEncoder extends LyRequestEncoder
{
	public static final String     __VERSION =
		"$Id: LyUnionRequestEncoder.java 41563 2012-11-19 13:44:17Z jlsaouli $";

	public char					   VALUE_SEP = (char)4;

	protected LyTransactionManager _tm;

	/**
	 * Constructor.
	 *
	 * @param provider
	 * @param dataSource
	 */
	public LyUnionRequestEncoder(LyDataProvider provider, LyDataSource dataSource)
	{
		super(provider, dataSource);
		LyApplication application = _provider.getApplication();

		_tm = application.getTransactionManager();
	}

	/**
	 * Set the request encoder response. Inherited method of LyRequestEncoder to set properties to
	 * the response.
	 *
	 * @param the response
	 */
	public void setResponse(LyResponse response)
	{
		super.setResponse(response);

		// Set properties of the response.
		_response.setSorted(false);
		_response.setFiltered(false);
		_response.setCounted(false);
		_response.setDistinctValues(false);
	}

	/**
	 * Get an object.
	 *
	 * @param  objectId object identifier
	 * @param  fields   the fields to load (if null, load all)
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getObject(LyObjectId objectId, Vector<LyField> fields)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// Don't check equality between the number of id fields and the values
		// of the id because in union we may have some differences when id and
		// union fields are the same.
		// We let the splitObjectId method do the job and complete the values
		// for missing id fields

		// split the objectId
		LyObjectId[] subIds = splitObjectId(objectId);

		// No Id found, no objects...
		if (subIds == null)
		{
			_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);
			return true;
		}

		// get the fields of union
		LyUnionClass	  classPeer = ((LyUnionClass)_class);
		Vector<LyField>[] subFields = classPeer.splitFields(fields);
		LyUnionBinding    unionBinding;

		if (_classBinding != null)
			unionBinding = (LyUnionBinding)_classBinding;
		else
			unionBinding = classPeer.getUnionBinding();

		// encode the request for the first object
		boolean		 status1 = true;
		LyDataSource source1 = classPeer._class1.getDataSource();
		LyResponse   response1 = null;
		int			 n1 = 0;

		if (subIds[0] != null)
		{
		    completeFields(classPeer._class1, subFields[0]);
			response1 = _tm.loadObject(source1, subIds[0], subFields[0], null, unionBinding.getClassBinding1());
			status1 = (response1.getStatus() != LyResponse.Status.ERROR);
			n1 = response1.getSize();
		}

		// encode the request for the second object
		boolean		 status2 = true;
		LyDataSource source2 = classPeer._class2.getDataSource();
		LyResponse   response2 = null;
		int			 n2 = 0;

		if (subIds[1] != null)
		{
		    completeFields(classPeer._class2, subFields[1]);
			response2 = _tm.loadObject(source2, subIds[1], subFields[1], null, unionBinding.getClassBinding2());
			status2 = (response2.getStatus() != LyResponse.Status.ERROR);
			n2 = response2.getSize();
		}

		if (!status1 && !status2)
		{
			if (_response != null)
				_response.setError(response1.getError() + "\n" + response2.getError());

			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			return false;
		}

		// get objects found
		if ((n1 < 1) || (n2 < 1))
		{
			_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);
			return true;
		}

		// get the first sub-object found
		Object object1 = response1.getResponse(0);

		// get the second sub-object found
		Object object2 = response2.getResponse(0);

		// send response
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED,
			new LyUnionObject(classPeer, object1, object2), null);
		return true;
	} // end method getObject

	/**
	 * This utility method split a given object id related to a union object to an array of
	 * sub-objects ids, each corresponding to a sub-class of the union class. There are a few rules
	 * to respect: All Id field in the union class must correspond to all Id field of class 1. The
	 * union field must be the Id field of class 2 (and all Id field of class 2 must be union field
	 * (it allows an object of 2 to be uniquely identify using the union fields) )
	 *
	 * @param  objectId an object id related to a union object
	 * @return an array of object id, each related to a sub-class of the current union class.
	 */
	public LyObjectId[] splitObjectId(LyObjectId objectId)
	{
		LyObjectId[]   result = new LyObjectId[2];

		// get the peer class
		LyUnionClass   classPeer = (LyUnionClass)_class;
		LyUnionBinding unionBinding;

		if (_classBinding != null)
			unionBinding = (LyUnionBinding)_classBinding;
		else
			unionBinding = classPeer.getUnionBinding();

		// look for the common value
		String[] values = objectId.toStrings();
		String   value = null;

		// Create LyObjectId for corresponding object of class 1
		if (values.length == 1)
			value = values[0];
		else
		{
			Vector<LyField> unionIdFields = classPeer.getIdFields();
			Vector<? extends LyField>	    class1IdFields;

			if (classPeer._class1 instanceof LyUnionClass)
				class1IdFields = getIdFields((LyUnionClass)classPeer._class1);
			else
				class1IdFields = classPeer._class1.getIdFields();

			if ((unionIdFields.size() == 1) && (class1IdFields.size() == 1))
				// All the Ids of the union in values are hold by union fields
				// so we can take one of the value (all are equals)
				value = values[0];
			else if (unionIdFields.size() == class1IdFields.size())
			{
				// All Id fields of class 1 must be id fields in union class
				value = null;
				for (int i = 0; i < values.length; i++)
				{
					if (value == null)
						value = values[i];
					else
						value = value + LyObjectId.FIELD_SEP + values[i];
				}
			}
			else if (unionIdFields.size() < class1IdFields.size())
				// We don't have enough Id to correctly find object of class 1
				throw new IllegalStateException("BF AppEngine : "
					+ getEnvironment().getMessage("LY_ER_CONDITION_OBJECT_ID", objectId.toName(),
						_class.getId()));
			else
			{
				//We have more id fields than required, we have to remove all value
				// that don't correspond to an Id field of class 1
				// We construct a new Id
				boolean findId = false;

				for (int i = 0; i < class1IdFields.size(); i++)
				{
					LyField class1Field = (LyField)class1IdFields.elementAt(i);
					String  Id = class1Field.getId();
					// for each field we try to find the corresponding value
					findId = false;
					for (int j = 0; j < unionIdFields.size(); j++)
					{
						LyField unionField = unionIdFields.elementAt(i);

						if (unionField.getId().equals(Id))
						{
							if (value == null)
								value = values[i];
							else
								value = value + LyObjectId.FIELD_SEP + values[i];
							findId = true;
							break;
						}
					}
					if (!findId)
						throw new IllegalStateException("BF AppEngine : "
							+ getEnvironment().getMessage("LY_ER_CONDITION_OBJECT_ID",
								objectId.toName(), _class.getId()));
				}
			}
		}

		LyObjectId   objectId1 = new LyObjectId(classPeer._class1, value);
		LyDataSource source1 = classPeer._class1.getDataSource();

		Object		 object1 = source1.getObject(objectId1, null, unionBinding.getClassBinding1());

		if (object1 == null)
			return null;
		// Create LyObjectId for object of class 2
		// Get value for union field from object1
		value = null;

		LyField keyField = null, keyField2;
		Vector<? extends LyField>  class2IdFields;
		String  relValue;

		if (classPeer._class2 instanceof LyUnionClass)
			class2IdFields = getIdFields((LyUnionClass)classPeer._class2);
		else
			class2IdFields = classPeer._class2.getIdFields();
		for (int i = 0; i < unionBinding.getUnionCount(); i++)
		{
			keyField = classPeer._class1.getField(unionBinding.getValue1(i));
			keyField2 = classPeer._class2.getField(unionBinding.getValue2(i));
			if (!class2IdFields.contains(keyField2))
				continue;

			// for each union field, get value
			relValue = (classPeer._class1.getValue(object1, keyField)).toString();
			if (value == null)
				value = relValue;
			else
				value = value + LyObjectId.FIELD_SEP + relValue;
		}

		LyObjectId objectId2 = new LyObjectId(classPeer._class2, value);

		result[0] = objectId1;
		result[1] = objectId2;

		return result;
	} // end method splitObjectId

	/**
	 * Complete fields for given class
	 */
	protected void completeFields(LyClass classe, Vector<LyField> fields)
	{
	    if (fields == null)
	        return;
	    
	    Vector<? extends LyField>   idFields = classe.getIdFields();
	    int       					i, n = idFields.size();
	    LyField   					field;
	    
	    for (i = 0; i < n; i++)
	    {
	        field = (LyField)idFields.get(i);
	        if (!fields.contains(field)
	        		&& (classe.getField(field.getId()) == field))
	            fields.add(field);
	    }
	}
	
	/**
	 * Get a set of objects.
	 *
	 * @param  fields the fields to load (if null, load all)
	 * @param  sort   indicates sorting criterion, or null if default sort is used.
	 * @param  filter indicates filtering criterion, or null if there is no filter.
	 * @param  n      maximum number of objects to get.
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getObjects(Vector<LyField> fields, LySort sort, LyFilter filter, int n)
	{
		// We must check if all id fields of the peer classes are in the list
		checkIdFields(fields);

		LyUnionClass	classPeer = ((LyUnionClass)_class);
		LyUnionBinding	unionBinding;

		if (_classBinding != null)
			unionBinding = (LyUnionBinding)_classBinding;
		else
			unionBinding = classPeer.getUnionBinding();

		boolean			strict = unionBinding.isStrict();
		
		if (strict)
		{
			// Special case for data providers supporting join requests
			LyJoinDataProvider provider = canSendJoinRequest();
	
			if (provider != null)
				return getJoinObjects(provider, fields, sort, filter, n);
		}

		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// split fields : for each field, find the class it belongs
		Vector<LyField>[] subFields = classPeer.splitFields(fields);

		// split sorts
		LySort[]		  subSorts = classPeer.splitSort(sort);

		// split filters
		LyFilter[]		  subFilters = classPeer.splitFilter(filter);

		// encode the request for the first class
		LyDataSource	  source1 = classPeer._class1.getDataSource();
		LyResponse		  response1;
		LyClassBinding    classBinding1 = classPeer.getUnionBinding().getClassBinding1();
		boolean			  status;

		completeFields(classPeer._class1, subFields[0]);
		response1 = _tm.loadObjects(source1, subFields[0], subSorts[0], subFilters[0], n, null, classBinding1);
		status = (response1.getStatus() != LyResponse.Status.ERROR);

		if (!status)
		{
			if (_response != null)
				_response.setError(response1.getError());

			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			return status;
		}

		// encode the request for the second class
		LyDataSource   source2 = classPeer._class2.getDataSource();
		LyResponse     response2;
		LyClassBinding classBinding2 = classPeer.getUnionBinding().getClassBinding2();

		completeFields(classPeer._class2, subFields[1]);
		response2 = _tm.loadObjects(source2, subFields[1], subSorts[1], subFilters[1], n, null, classBinding2);
		status = (response2.getStatus() != LyResponse.Status.ERROR);

		if (!status)
		{
			if (_response != null)
				_response.setError(response2.getError());

			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			return status;
		}

		// check if the sort must be taken from the 1rst or 2nd response
		boolean      sortOnFirstClass;
		
		if (((subSorts[0] == null) && (subSorts[1] == null))
		        || ((subSorts[0] != null) && (subSorts[0].getCriterionCount() > 0)))
		    sortOnFirstClass = true;
		else
		    sortOnFirstClass = false;

		// get key fields for union
		LyField		keyField1 = null;
		LyField		keyField2 = null;

		// create a list of union objects
		ArrayList<LyUnionObject> unionObjects = new ArrayList<LyUnionObject>();

		// get sub-objects from responses
		Object		object, object2;
		String		key;
		int			i, j, nv, nbobj1 = response1.getSize(), nbobj2 = response2.getSize();

		if ((n < 0)
			&& ((filter == null)
				|| (filter.getSize() == 1)
				|| (filter.getType() == LyFilter.Type.AND)))
		{
		    if (sortOnFirstClass)
		    {
    			// stored objects from response2 inside a map
    			HashMap<String, Object> objects = new HashMap<String, Object>(nbobj2);
    
    			for (i = 0; i < nbobj2; i++)
    			{
    				object = response2.getResponse(i);
    				
    				key = "";
    				for (j = 0, nv = unionBinding.getUnionCount(); j < nv; j++)
    				{
    					keyField2 = classPeer._class2.getField(unionBinding.getValue2(j));
    
    					key += (classPeer._class2.getValue(object, keyField2)).toString();
    					key += VALUE_SEP;
    				}
    				objects.put(key, object);
    			}
    			// create union objects in the order of response1
    			for (i = 0; i < nbobj1; i++)
    			{
    				object = response1.getResponse(i);
    
    				key = "";
    				for (j = 0, nv = unionBinding.getUnionCount(); j < nv; j++)
    				{
    					keyField1 = classPeer._class1.getField(unionBinding.getValue1(j));
    
    					key += (classPeer._class1.getValue(object, keyField1)).toString();
    					key += VALUE_SEP;
    				}
    				object2 = objects.get(key);
    				if ((object2 != null)
    						|| !strict)
    					unionObjects.add(new LyUnionObject(classPeer, object, object2));
    			}
		    }
		    else
		    {
	            // stored objects from response1 inside a map
                HashMap<String, Object> objects = new HashMap<String, Object>(nbobj1);
    
                for (i = 0; i < nbobj1; i++)
                {
                    object = response1.getResponse(i);
                    
                    key = "";
                    for (j = 0, nv = unionBinding.getUnionCount(); j < nv; j++)
                    {
                        keyField1 = classPeer._class1.getField(unionBinding.getValue1(j));
    
                        key += (classPeer._class1.getValue(object, keyField1)).toString();
                        key += VALUE_SEP;
                    }
                    objects.put(key, object);
                }
                // create union objects in the order of response2
                for (i = 0; i < nbobj2; i++)
                {
                    object = response2.getResponse(i);
    
                    key = "";
                    for (j = 0, nv = unionBinding.getUnionCount(); j < nv; j++)
                    {
                        keyField2 = classPeer._class2.getField(unionBinding.getValue2(j));
    
                        key += (classPeer._class2.getValue(object, keyField2)).toString();
                        key += VALUE_SEP;
                    }
                    object2 = objects.get(key);
    				if ((object2 != null)
    						|| !strict)
    					unionObjects.add(new LyUnionObject(classPeer, object2, object));
                }
		    }
		}
		else
		{
		    // create union objects in the order of response1
            Vector<LyField>     keyFields;
            String[]            keyIds;
            LyObjectId          objectId1, objectId2;
            LyResponse          response;
            Vector<Object>      objects = new Vector<Object>(nbobj1 + nbobj2);

            for (i = 0; i < nbobj1; i++)
            {
                object = response1.getResponse(i);
                if (objects.contains(object))
                    continue;
                
                objects.add(object);
                nv = unionBinding.getUnionCount();
                keyFields = new Vector<LyField>(nv);
                keyIds = new String[nv];
                
                for (j = 0; j < nv; j++)
                {
                    keyField1 = classPeer._class1.getField(unionBinding.getValue1(j));
                    keyIds[j] = classPeer._class1.getValue(object, keyField1).toString();
                    keyField2 = classPeer._class2.getField(unionBinding.getValue2(j));
                    keyFields.add(keyField2);
                }
                
                objectId2 = new LyObjectId(classPeer._class2, keyIds);
                response = _tm.loadObject(source2, objectId2, keyFields, null, unionBinding.getClassBinding1());
                if (response.getSize() == 1)
                    object2 = response.getResponse(0);
                else
                	object2 = null;
                
                if ((object2 != null)
                		|| !strict)
                	unionObjects.add(new LyUnionObject(classPeer, object, object2));
            }
            
            // create union objects in the order of response2
            for (i = 0; i < nbobj2; i++)
            {
                object = response2.getResponse(i);
                if (objects.contains(object))
                    continue;
                
                objects.add(object);
                nv = unionBinding.getUnionCount();
                keyFields = new Vector<LyField>(nv);
                keyIds = new String[nv];
                
                for (j = 0; j < nv; j++)
                {
                    keyField2 = classPeer._class2.getField(unionBinding.getValue2(j));
                    keyIds[j] = classPeer._class2.getValue(object, keyField2).toString();
                    keyField1 = classPeer._class1.getField(unionBinding.getValue1(j));
                    keyFields.add(keyField1);
                }
                
                objectId1 = new LyObjectId(classPeer._class1, keyIds);
                response = _tm.loadObject(source1, objectId1, keyFields, null, unionBinding.getClassBinding1());
                if (response.getSize() == 1)
                    object2 = response.getResponse(0);
                else
                	object2 = null;
                
                if ((object2 != null)
                		|| !strict)
                    unionObjects.add(new LyUnionObject(classPeer, object2, object));
            }
		}

		// send responses
		for (i = 0, n = unionObjects.size(); i < n; i++)
			_dataSource.receiveResponse(_response, LyResponse.Status.RUNNING, unionObjects.get(i), null);

		// terminate operation
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);
		return true;
	} // end method getObjects

	/**
	 * Return the number of objects from the data provider.
	 *
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getSize(LyFilter filter)
	{
		// Special case for data providers supporting join requests
		LyJoinDataProvider provider = canSendJoinRequest();

		if (provider != null)
			return getJoinSize(provider, filter);

		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// get the union class peer
		LyUnionClass   classPeer = ((LyUnionClass)_class);

		// split filters
		LyFilter[]     subFilters = classPeer.splitFilter(filter);

		// get key fields for union
		LyUnionBinding unionBinding;

		if (_classBinding != null)
			unionBinding = (LyUnionBinding)_classBinding;
		else
			unionBinding = classPeer.getUnionBinding();

		// encode the request for the first class
		LyDataSource source1 = classPeer._class1.getDataSource();
		int			 size1 = -1;

		size1 = _tm.getSize(source1, subFilters[0], unionBinding.getClassBinding1());

		// encode the request for the second class
		LyDataSource source2 = classPeer._class2.getDataSource();
		int			 size2 = -1;

		size2 = _tm.getSize(source2, subFilters[1], unionBinding.getClassBinding2());

		if (size1 >= size2)
			_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, new Integer(size1), null);
		else
			_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, new Integer(size2), null);

		return true;
	} // end method getSize

	/* Methods for deletion */

	/**
	 * To delete an object.
	 *
	 * @param  object the object to delete
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean deleteObject(Object object)
	{
		Vector<Object> objects = new Vector<Object>(1);

		objects.add(object);
		return deleteObjects(objects);
	}

	/**
	 * To delete a list of objects.
	 *
	 * @param  objects the objects to delete
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean deleteObjects(Vector<Object> objects)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// get the peer class
		LyUnionClass   classPeer = (LyUnionClass)_class;
		LyUnionBinding unionBinding;

		if (_classBinding != null)
			unionBinding = (LyUnionBinding)_classBinding;
		else
			unionBinding = classPeer.getUnionBinding();

		// create lists of sub-objects
		LyUnionObject  object;
		int			   nbobj = objects.size();
		Vector<Object> objects1 = new Vector<Object>(nbobj);
		Vector<Object> objects2 = new Vector<Object>(nbobj);
		LyDataSource   source2 = classPeer._class2.getDataSource();
		LyDataSource   source1 = classPeer._class1.getDataSource();

		for (int i = 0; i < objects.size(); i++)
		{
			object = (LyUnionObject)objects.get(i);
			if (source1.checkObject(object.getObject1()))
				objects1.add(object.getObject1());
			if (source2.checkObject(object.getObject2()))
				objects2.add(object.getObject2());
		}

		// FST0027 : delete the second object first!!
		// encode the request for the second objects
		LyResponse rsp;
		boolean    status;

		//FFT01508
		if (objects2.size() > 0)
		{
			rsp = _tm.deleteObjects(source2, objects2, new _LyUnionResponseListener(),
					unionBinding.getClassBinding2());

			status = (rsp.getStatus() != LyResponse.Status.ERROR);
			if (!status)
			{
				if (_response != null)
					_response.setError(rsp.getError());
				_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
				return status;
			}
		}

		// encode the request for the first objects
		// FFT01508
		if (objects1.size() > 0)
		{
			rsp = _tm.deleteObjects(source1, objects1, new _LyUnionResponseListener(),
					unionBinding.getClassBinding1());

			status = (rsp.getStatus() != LyResponse.Status.ERROR);
			if (!status)
			{
				if (_response != null)
					_response.setError(rsp.getError());
				_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
				return status;
			}
		}

		// send responses
		for (int i = 0; i < nbobj; i++)
		{
			object = (LyUnionObject)objects.get(i);
			_dataSource.receiveResponse(_response, LyResponse.Status.RUNNING, object, null);
		}
		// terminate operation
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);
		return true;
	} // end method deleteObjects

	/* Methods for creation */

	/**
	 * To create a new object.
	 *
	 * @param  values the values of the new object
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean createObject(Object object)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// get the peer class
		LyUnionClass   classPeer = (LyUnionClass)_class;
		LyUnionObject  unionObject = (LyUnionObject)object;
		LyUnionBinding unionBinding;

		if (_classBinding != null)
			unionBinding = (LyUnionBinding)_classBinding;
		else
			unionBinding = classPeer.getUnionBinding();

		// encode the request for the first object
		LyDataSource source1 = classPeer._class1.getDataSource();
		LyResponse   response;

		response = _tm.createObject(source1, unionObject.getObject1(),
				new _LyUnionResponseListener(), unionBinding.getClassBinding1());
		boolean status = (response.getStatus() != LyResponse.Status.ERROR);

		if (!status)
		{
			if (_response != null)
				_response.setError(response.getError());

			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			return status;
		}

		// The object 1 is created
		// We must check that all common values in object 2 are filled.
		// If they are not filled, use values from object 1.
		// This may be usefull if one of the field is created using sequence or autoincrement
		// in the database
		if (response.getSize() >= 1)
		{
			String		   bindingValue1;
			String		   bindingValue2;
			LyValue		   currentValue1;
			LyValue		   currentValue2;
			Object		   realValue;
			LyColumnField  field;

			LyUnionBinding binding = unionObject.getUnionClass().getUnionBinding();
			Object		   object1 = response.getResponse(0);
			Object		   object2 = unionObject.getObject2();

			LyValueSet     values1 = classPeer._class1.getValues(object1);
			LyValueSet     values2 = classPeer._class2.getValues(object2);

			for (int i = 0; i < binding.getUnionCount(); i++)
			{
				bindingValue1 = binding.getValue1(i);
				bindingValue2 = binding.getValue2(i);

				// Get the value of the field bindingValue1 for the object1
				currentValue1 = values1.getFieldValue(bindingValue1);

				// If there is no value in the object 1, nothing to do
				if ((currentValue1 == null) || (currentValue1.isNull()))
					// If currentValue1 is null, we can't do anything
					continue;

				// Get the value of the field bindingValue2 for the object2
				currentValue2 = values2.getFieldValue(bindingValue2);

				if ((currentValue2 != null) && (currentValue2.isNull()))
				{
					// Get the corresponding value
					realValue = currentValue1.getValue();
					// and get the corresponding field in object 2
					field = (LyColumnField)currentValue2.getField();
					classPeer._class2.setValue(object2, field, realValue);
				}
			}
		}

		// encode the request for the second object
		LyDataSource source2 = classPeer._class2.getDataSource();

		response = _tm.createObject(source2, unionObject.getObject2(),
				new _LyUnionResponseListener(), unionBinding.getClassBinding2());
		status = (response.getStatus() != LyResponse.Status.ERROR);

		if (!status)
		{
			if (_response != null)
				_response.setError(response.getError());
			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			return status;
		}

		// send response
		unionObject.validate();
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, unionObject, null);

		return true;
	} // end method createObject

	/* Methods for modification */

	/**
	 * To set an object.
	 *
	 * @param  object the object to set
	 * @param  values the new values of the object
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean setObject(Object object)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// get the peer class
		LyUnionClass   classPeer = (LyUnionClass)_class;
		LyUnionBinding unionBinding;

		if (_classBinding != null)
			unionBinding = (LyUnionBinding)_classBinding;
		else
			unionBinding = classPeer.getUnionBinding();

		// encode the request for the first object
		LyDataSource source1 = classPeer._class1.getDataSource();
		LyResponse   response1;
		boolean		 status;

		// Add the two objects to the list of objects to ignored on the
		// union class. If we don't do that, union class receive an event and
		// may clear values on the second object.
		LyUnionObject	unionObject = (LyUnionObject)object;
		Object		 	object1 = unionObject.getObject1();
		Object		 	object2 = unionObject.getObject2();

		classPeer.addIgnoredObject(object1);
		classPeer.addIgnoredObject(object2);

		if (source1.checkObject(object1))
			response1 = _tm.setObject(source1, object1, null, unionBinding.getClassBinding1());
		else
			response1 = _tm.createObject(source1, object1, null, unionBinding.getClassBinding1());
		
		status = (response1.getStatus() != LyResponse.Status.ERROR);

		if (!status)
		{
			_response.setError(response1.getError());
			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			return status;
		}

		// Encode the request for the second object
		LyDataSource source2 = classPeer._class2.getDataSource();
		LyResponse   response2;
		
		// Check if object2 needs to be set or created
		if (source2.checkObject(object2))
			response2 = _tm.setObject(source2, object2, null, unionBinding.getClassBinding2());
		else
		{
			// We must check that all common values in object 2 are filled.
			// If they are not filled, use values from object 1.
			// This may be useful if one of the field is created using sequence or auto-increment in the database
			String		   bindingValue1, bindingValue2;
			LyValue		   currentValue1, currentValue2;
			Object		   realValue;
			LyColumnField  field;
			LyUnionBinding binding = unionObject.getUnionClass().getUnionBinding();
			LyValueSet     values1 = classPeer._class1.getValues(object1);
			LyValueSet     values2 = classPeer._class2.getValues(object2);

			for (int i = 0; i < binding.getUnionCount(); i++)
			{
				bindingValue1 = binding.getValue1(i);
				bindingValue2 = binding.getValue2(i);

				// Get the value of the field bindingValue1 for the object1
				currentValue1 = values1.getFieldValue(bindingValue1);

				// If there is no value in the object 1, nothing to do
				if ((currentValue1 == null)
						|| (currentValue1.isNull()))
					// If currentValue1 is null, we can't do anything
					continue;

				// Get the value of the field bindingValue2 for the object2
				currentValue2 = values2.getFieldValue(bindingValue2);

				if ((currentValue2 != null)
					&& (currentValue2.isNull()))
				{
					// Get the corresponding value
					realValue = currentValue1.getValue();
					// and get the corresponding field in object 2
					field = (LyColumnField)currentValue2.getField();
					classPeer._class2.setValue(object2, field, realValue);
				}
			}
			
			response2 = _tm.createObject(source2, object2, null, unionBinding.getClassBinding2());
		}
		
		status = (response2.getStatus() != LyResponse.Status.ERROR);

		if (!status)
		{
			_response.setError(response2.getError());
			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			return status;
		}

		// get values from responses
		LyValueSet oldValues1 = response1.getValues(0);
		LyValueSet oldValues2 = response2.getValues(0);
		LyValueSet oldValues;

		if ((oldValues1 == null) && (oldValues2 == null))
			oldValues = null;
		else
		{
			oldValues = new LyValueSet();
			oldValues.addValues(oldValues1);
			oldValues.addValues(oldValues2);
		}
		// send response
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object, oldValues);

		classPeer.removeIgnoredObject(object1);
		classPeer.removeIgnoredObject(object2);

		return true;
	} // end method setObject

	/**
	 * To set a set of objects.
	 *
	 * @param  objects the objects to set
	 * @param  values  the new values of the objects
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean setObjects(Vector<Object> objects)
	{
		throw new LyUnsupportedOperationException("BF AppEngine : "
			+ getEnvironment().getMessage("LY_ER_UNSUPPORTED_OPERATION", "LyUnionRequestEncoder",
				"setObjects"));
	}

	/**
	 * To test if we can send a join request for loading objects rather than cutting it into smaller
	 * pieces. We can send a join request if all peer classes have the same data provider and if
	 * this data provider implements the LyJoinDataProvider interface.
	 */
	protected LyJoinDataProvider canSendJoinRequest()
	{
		Collection<LyClass>     classes = getPeerClasses();
		Iterator<LyClass>	   it = classes.iterator();
		LyApplication  application = _provider.getApplication();
		LyClass		   cls;
		LyDataProvider provider = null;
		LyDataSource   dataSource;

		while (it.hasNext())
		{
			cls = (LyClass)it.next();
			dataSource = application.getDataSource(cls);
			if (provider == null)
			{
				provider = dataSource.getDataProvider();
				if (!(provider instanceof LyJoinDataProvider))
					return null;
			}
			else if (provider != dataSource.getDataProvider())
				return null;
		}

		return (LyJoinDataProvider)provider;
	}

	/**
	 */
	protected boolean getJoinSize(LyJoinDataProvider provider, LyFilter filter)
	{
		_response.setStatus(LyResponse.Status.RUNNING);

		LyJoinRequestEncoder encoder = provider.getJoinRequestEncoder(_dataSource);

		// We must set the current response to the request encoder.
		((LyRequestEncoder)encoder).setResponse(_response);

		Integer size = encoder.getSize(this, filter);

		if (size == null)
		{
			// There are some errors.
			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			return false;
		}

		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, size, null);

		return true;
	}

	/**
	 */
	protected boolean getJoinObjects(LyJoinDataProvider provider, Vector<LyField> fields,
		LySort sort, LyFilter filter, int n)
	{
		_response.setStatus(LyResponse.Status.RUNNING);

		LyJoinRequestEncoder encoder = provider.getJoinRequestEncoder(_dataSource);

		// We must set the current response to the request encoder.
		((LyRequestEncoder)encoder).setResponse(_response);

		Collection<LyValueSet> values = encoder.getObjects(this, fields, sort, filter, n);

		if (values == null)
		{
			// There are some errors.
			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			return false;
		}

		// Now we have all values for all objects part of the union(s) object(s).
		// We must build all the parts and agregate it.
		Iterator<LyValueSet> it = values.iterator();
		LyValueSet			 objectValues;
		LyUnionClass		 cls = (LyUnionClass)_class;
		LyUnionObject		 object;

		while (it.hasNext())
		{
			objectValues = it.next();
			object = buildUnionObject(encoder, cls, objectValues);
			_dataSource.receiveResponse(_response, LyResponse.Status.RUNNING, object, null);
		}
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);

		return true;
	} // end method getJoinObjects

	/**
	 */
	protected LyUnionObject buildUnionObject(
		LyJoinRequestEncoder encoder,
		LyUnionClass cls,
		LyValueSet values)
	{
		LyClass[]    classes = new LyClass[2];
		Object[]     objects = new Object[2];
		Object		 val;
		LyResponse   response;
		LyDataSource dataSource;
		LyValue		 value;
		LyField		 field;
		Vector<? extends LyField>		 fields;

		classes[0] = cls._class1;
		classes[1] = cls._class2;

		for (int i = 0; i < classes.length; i++)
		{
			if (classes[i] instanceof LyUnionClass)
				objects[i] = buildUnionObject(encoder, (LyUnionClass)classes[i], values);
			else
			{
				// Create the new object and set its values.
				objects[i] = encoder.buildObject(this, classes[i]);
				fields = classes[i].getFields();
				for (int j = 0; j < values.getSize(); j++)
				{
					value = values.getFieldValue(j);
					field = value.getField();
					if (fields.contains(field))
					{
						val = value.getValue();
						classes[i].setValue(objects[i], field, val);
					}
				}
				((LyPeerClass)classes[i]).validate(objects[i], null);
			}

			// Simulate a loading response to the datasource.
			// We must do it in all cases else the object is not put
			// in the datasource cache.
			response = new LyResponse(LyResponse.Operation.LOAD, classes[i]);
			response.setStatus(LyResponse.Status.RUNNING);
			dataSource = classes[i].getDataSource();
			if (classes[i] instanceof LyUnionClass)
				response.setListener(new _LyUnionResponseListener2((LyUnionClass)classes[i]));
			else
				response.setListener(new _LyUnionResponseListener2(cls));

			dataSource.receiveResponse(response, LyResponse.Status.COMPLETED, objects[i], null);
			objects[i] = response.getResponse(0);
		}

		return new LyUnionObject(cls, objects[0], objects[1]);
	} // end method buildUnionObject

	/**
	 */
	public Collection<LyClass> getPeerClasses()
	{
		return getPeerClasses((LyUnionClass)_class);
	}

	/**
	 */
	protected Collection<LyClass> getPeerClasses(LyUnionClass cls)
	{
		ArrayList<LyClass>  result;
		Collection<LyClass> result1, result2;
		LyClass			    cls1 = cls._class1;
		LyClass			    cls2 = cls._class2;

		result = new ArrayList<LyClass>();
		if (cls1 instanceof LyUnionClass)
		{
			result1 = getPeerClasses((LyUnionClass)cls1);
			result.addAll(result1);
		}
		else
			result.add(cls1);

		if (cls2 instanceof LyUnionClass)
		{
			result2 = getPeerClasses((LyUnionClass)cls2);
			result.addAll(result2);
		}
		else
			result.add(cls2);

		return result;
	}

	protected void checkIdFields(Vector<LyField> fields)
	{
		// All fields are requested : nothing to do.
		if (fields == null)
			return;

		Collection<LyClass> peerClasses = getPeerClasses();
		Iterator<LyClass>   it = peerClasses.iterator();
		LyClass    peerClass;
		Vector<? extends LyField>     idFields;
		int		   i, n;
		LyField    peerField;

		while (it.hasNext())
		{
			peerClass = (LyClass)it.next();
			idFields = peerClass.getIdFields();
			n = idFields.size();
			for (i = 0; i < n; i++)
			{
				peerField = (LyField)idFields.elementAt(i);
				if (!fields.contains(peerField))
					fields.addElement(peerField);
			}
		}
	}

	protected Vector<LyField> getIdFields()
	{
		return getIdFields((LyUnionClass)_class);
	}

	/**
	 */
	protected Vector<LyField> getIdFields(LyUnionClass cls)
	{
		Collection<LyClass>				  peerClasses = getPeerClasses(cls);
		Iterator<LyClass>				  it = peerClasses.iterator();
		LyClass					  peerClass;
		Vector<? extends LyField> idFields;
		Vector<LyField>			  result = new Vector<LyField>();

		while (it.hasNext())
		{
			peerClass = (LyClass)it.next();
			idFields = peerClass.getIdFields();
			result.addAll(idFields);
		}

		return result;
	}

	/**
	 * This class does nothing as a responseListener. It is just implemented to add the union class
	 * in the list of skip listeners when we send a response
	 *
	 * @author ban
	 */
	class _LyUnionResponseListener
		implements LyResponseListener,
			LyGetDataSourceListenersInterface
	{

		/**
		 * @see leon.data.LyResponseListener#newResponse(leon.data.LyResponse, java.lang.Object,
		 *      leon.data.LyValueSet)
		 */
		public void newResponse(LyResponse response, Object object, LyValueSet oldValues)
		{

		}

		/**
		 * @see leon.data.LyResponseListener#endResponse(leon.data.LyResponse)
		 */
		public void endResponse(LyResponse response)
		{

		}

		/**
		 * @see leon.data.LyGetDataSourceListenersInterface#getDataSourceListeners()
		 */
		public Vector<LyDataSourceListener> getDataSourceListeners()
		{
			LyUnionClass unionClass = (LyUnionClass)_class;

			return unionClass.getListeners();
		}
	} // end class _LyUnionResponseListener


	/**
	 * This class does nothing as a responseListener. It is just implemented to add the union class
	 * in the list of skip listeners when we send a response
	 *
	 * @author ban
	 */
	class _LyUnionResponseListener2
		implements LyResponseListener,
			LyGetDataSourceListenersInterface
	{
		protected LyUnionClass _unionClass;

		/**
		 * Constructor.
		 *
		 * @param unionClass class to skip for listener events
		 */
		public _LyUnionResponseListener2(LyUnionClass unionClass)
		{
			_unionClass = unionClass;
		}

		/**
		 * @see leon.data.LyResponseListener#newResponse(leon.data.LyResponse, java.lang.Object,
		 *      leon.data.LyValueSet)
		 */
		public void newResponse(LyResponse response, Object object, LyValueSet oldValues)
		{

		}

		/**
		 * @see leon.data.LyResponseListener#endResponse(leon.data.LyResponse)
		 */
		public void endResponse(LyResponse response)
		{

		}

		/**
		 * @see leon.data.LyGetDataSourceListenersInterface#getDataSourceListeners()
		 */
		public Vector<LyDataSourceListener> getDataSourceListeners()
		{
			LyUnionClass				 unionClassParent = (LyUnionClass)_class;
			Vector<LyDataSourceListener> list = new Vector<LyDataSourceListener>();

			addUnionListeners(unionClassParent, list);
			if (unionClassParent != _unionClass)
				addUnionListeners(_unionClass, list);

			return list;
		}

		protected Vector<LyDataSourceListener> addUnionListeners(LyUnionClass unionClass,
			Vector<LyDataSourceListener> list)
		{
			Vector<LyDataSourceListener> current = unionClass.getListeners();

			for (int i = 0; i < current.size(); i++)
			{
				if (!list.contains(current.elementAt(i)))
					list.addElement(current.elementAt(i));
			}

			if (unionClass._class1 instanceof LyUnionClass)
				addUnionListeners((LyUnionClass)unionClass._class1, list);
			if (unionClass._class2 instanceof LyUnionClass)
				addUnionListeners((LyUnionClass)unionClass._class2, list);

			return list;
		}
	} // end class _LyUnionResponseListener2
} // end class LyUnionRequestEncoder
