/*
 * LyBlobFileHandler.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.util;

import java.io.*;
import leon.data.*;
import leon.peer.*;


/**
 * This class offers methods for fectching content from a 'blob' or embeded document in a data
 * provider and is used by leon.data.LyYFileValue to handle attachments.
 * @see leon.data.LyFileValue
 * @see LyObject
 */
public class LyBlobFileHandler
{
	public static final String __VERSION =
		"$Id: LyBlobFileHandler.java 36510 2012-08-17 15:43:06Z jlsaouli $";

	/**
	 * Default constructor.
	 */
	public LyBlobFileHandler()
	{
	}

	/**
	 * Gets the InputStream to the underlying value which is a 'File' The attachment handler MUST be
	 * "blob", additional FILE fields types may be supported later. The application may be able to
	 * read the content of the file. This inputStream is generated from the peer value (by the
	 * datasource)
	 * @param  internal locator (identifier) to get the attachment.
	 * @return the inputstream to read the file or null if it could not be found.
	 */
	public InputStream getInputStream(String locator)
	{
		if (locator != null)
		{
			try
			{
				BufferedInputStream bIn = new BufferedInputStream(new FileInputStream(locator));

				return bIn;
			}
			catch (IOException e)
			{
				e.printStackTrace();
			}
		}
		return null;
	}

	/**
	 * Sets the the content of the value which is a 'file'. The content should be read from the
	 * specified InputStream, and is stored in a file or in a blob or in a specific storage
	 * depending on the data provider. When the mthod exits, the inputStream is no longer used and
	 * may be closed. The attachment handler MUST be "blob", additional FILE fields types may be
	 * supported later.
	 * @param  iStream the inputstream used by the data provider to read the file.
	 * @param  dir     where to place the file
	 * @param  fieldId identifier of the field (to be part of the generated temp name)
	 * @return the locator path or null if not saved.
	 */
	public String setInputStream(String dir, String fieldId, InputStream iStream)
	{

		if (iStream == null)
			return null;

		String				 locator = null;
		BufferedOutputStream bOut = null;

		try
		{
			BufferedInputStream bIn = ((iStream instanceof BufferedInputStream)
					? (BufferedInputStream)iStream : new BufferedInputStream(iStream));
			File			    f = createTemporaryFile(dir, fieldId, null);

			bOut = new BufferedOutputStream(new FileOutputStream(f));
			// Copy
			int n = bIn.read();

			while (n >= 0)
			{
				bOut.write(n);
				n = bIn.read();
			}
			
			// Set locator to fetch it back
			locator = f.getAbsolutePath();
			
			// Close
			bOut.close();
			bOut = null;
			
			bIn.close();
			bIn = null;
		}
		catch (IOException e)
		{
			e.printStackTrace();
		}
		finally
		{
			try
			{
				if (bOut != null)
					bOut.close();
			}
			catch (IOException e)
			{
			}

		}
		return locator;
	} // end method setInputStream

	/**
	 * Create a temporary file to fecth attachments.
	 * @param  directory where to place the file
	 * @param  fieldId   identifier of the field (to be part of the generated temp name)
	 * @param  endname   optional end of the temporary fileName
	 * @return the File to create
	 */
	protected File createTemporaryFile(String dir, String fieldId, String endName)
	{
		File f, d;

		if (dir != null)
		{
			d = new File(dir);

			if (!d.exists())
				d.mkdir();

			// Delete all files existing for more than 10 minutes
			File[] files = d.listFiles();
			int    i, n = (files == null ? 0 : files.length);

			for (i = 0; i < n; i++)
			{
				if ((System.currentTimeMillis() - files[i].lastModified()) > (10 * 600000L))
				{
					try
					{
						files[i].delete();
					}
					catch (SecurityException e)
					{
					}
				}
			}
		}

		StringBuffer fileName = new StringBuffer();

		fileName.append("att_").append(fieldId).append("_").append(System.currentTimeMillis());
		if (endName == null)
			fileName.append(".tmp");
		else
		{
			int n1, n2, n3;

			n1 = endName.lastIndexOf("/");
			n2 = endName.lastIndexOf("\\");
			n3 = endName.lastIndexOf(" ");
			n1 = Math.max(n1, n2);
			n1 = Math.max(n1, n3);

			fileName.append("_").append(endName.substring(n1 + 1));
		}

		f = new File(dir, fileName.toString());
		f.deleteOnExit(); // enforce delete on exit in addition to timout deletion above
		return f;
	} // end method createTemporaryFile

	/**
	 * Fetch attachment from the peer data provider. This methods actually download or get from the
	 * blob the target attachment.
	 * @param  dir          where to place the file
	 * @param  fieldInfoId  identifier of the field (to be part of the generated temp name)
	 * @param  origFileName identifier of the field (to be part of the generated temp name)
	 * @return the locator path if attachment was found, null otherwise
	 */
	public String fetchAttachment(LyObject object, LyField field, String dir, String fieldInfoId,
		String origFileName)
	{
		// Searching for a non local value.
		Object peer = object.getPeer();

		if (peer == null)
			return null;

		String  locatorResult = null;

		LyClass peerClass = object.getClassInfo().getPeerClass();

		//LyField               field = object.getPeerField(fieldInfo);
		Object  peerVal = peerClass.getValue(peer, field);

		if (peerVal instanceof LyPeerFileValue)
		{
			String locator = ((LyPeerFileValue)peerVal).getLocator();
			//               We have to fetch the file from peer !
			// retreive content to a temporary location
			if (locator.startsWith("LyDataProvider://"))
			{
				String s = locator.substring("LyDataProvider://".length());
				String providerId = null, objectId = null, fieldId = null;
				int    n1 = s.indexOf("//"), n2;

				if (n1 > 0)
				{
					providerId = s.substring(0, n1);
					n2 = s.indexOf("//", n1 + 2);
					if (n2 > 0)
					{
						objectId = s.substring(n1 + 2, n2);
						fieldId = s.substring(n2 + 2);
					}
				}

				if ((providerId != null) && (fieldId != null) && (objectId != null))
				{
					LyDataProvider provider = peerClass.getDataSource().getDataProvider();

					if (providerId.equals(provider.getId()))
					{
						LyFetchAttachmentInterface encoder = (LyFetchAttachmentInterface)provider
								.getRequestEncoder(peerClass.getDataSource());
						BufferedOutputStream	   bOut = null;

						try
						{
							File f = createTemporaryFile(dir, fieldInfoId, origFileName);

							bOut = new BufferedOutputStream(new FileOutputStream(f));

							if (encoder.fetchAttachment(new LyObjectId(peerClass, objectId),
										fieldId, bOut))
								locatorResult = f.getAbsolutePath();
						}
						catch (IOException e)
						{
							e.printStackTrace();
						}
						finally
						{
							try
							{
								if (bOut != null)
									bOut.close();
							}
							catch (IOException e)
							{
							}
						}
					}
				}
			}
		}
		return locatorResult;
	} // end method fetchAttachment
} // end class LyBlobFileHandler
