/*
 * LyColumnField.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.util;

import java.util.ArrayList;
import java.util.List;

import leon.peer.*;


/**
 * This class provides a simple implementation of LyField.
 */
public class LyColumnField extends LyPeerField
{
	public static final String    __VERSION =
		"$Id: LyColumnField.java 34675 2012-03-15 08:02:42Z jlsaouli $";

	public enum CONSTRAINT {FOREIGN_KEY};
	
	protected static final String NO_DEFAULT = new String();

	/** Physical id of the field. */
	protected String			  _id;

	/** Class representing the type of the fieldPeer. */
	protected Class<?>			  _valueClass;

	/** Default value. */
	protected String			  _default;

	/** Format of this field. */
	protected String			  _format;

	/** Peer type (used by RDBMS which accepts BLOBS and/or CLOBS) otherwise value is 'DEFAULT'. */
	protected String		   	  _peerType;
	
	/** Store detected constrains on that field that may impact SQL generation **/
	protected List<CONSTRAINT>	  _constraints;

	/**
	 * Default Constructor.
	 */
	public LyColumnField()
	{
		_id = null;
		_valueClass = null;
		_default = NO_DEFAULT;
		_format = null;
	}

	/**
	 * Gets identifier of this field.
	 * @return Identifier of this field.
	 */
	public final String getId()
	{
		return _id;
	}

	/**
	 * Gets the java class representation of the value of this field.
	 * @return Java class used to represent values of this field.
	 */
	public final Class<?> getValueClass()
	{
		return _valueClass;
	}

	/**
	 * Gets the default value for this peer field as a string.
	 * @return Default value for this field given as a string.
	 */
	public final String getDefault()
	{
		return ((_default == NO_DEFAULT) ? null : _default);
	}

	/**
	 * Gets the format value for this peer field as a string.
	 * @return Format for this field given as a string.
	 */
	public final String getFormat()
	{
		return _format;
	}

	/**
	 * Peer type (used by RDBMS which accepts BLOBS and/or CLOBS) otherwise value is 'DEFAULT'
	 * @return Possible values are 'DEFAULT', 'BLOB' and 'CLOB'
	 */
	public final String getPeerType()
	{
		return _peerType;
	}

	/**
	 * Gets identifier of this field.
	 * @param id Identifier of this field.
	 */
	public void setId(String id)
	{
		if (id == null)
			throw new IllegalArgumentException("BF AppEngine : id can't be null");

		_id = id;
	}

	/**
	 * Sets the java class representation of the value of this field.
	 * @param valueClass Java class used to represent values of this field.
	 */
	public void setValueClass(Class<?> valueClass)
	{
		if (valueClass == null)
			throw new IllegalArgumentException("BF AppEngine : valueClass can't be null");

		_valueClass = valueClass;
	}

	/**
	 * Sets the default value for this peer field as a string.
	 * @param defaultValue Default value for this field given as a string.
	 */
	public void setDefault(String defaultValue)
	{
		if ((_default != NO_DEFAULT)
				&& (_default != defaultValue)
				&& (((_default == null) && (defaultValue != null))
					|| ((_default != null) && (defaultValue == null))
					|| ((_default != null) && !_default.equals(defaultValue))))
			System.out.println("[WARNING] - Leonardi : field " + this
				+ " has already a default value (" + _default + ")");

		_default = defaultValue;
	}

	/**
	 * Sets the format value for this peer field as a string.
	 * @param format Format for this field given as a string.
	 */
	public void setFormat(String format)
	{
		_format = format;
	}

	/**
	 * Sets the peer type (used by RDBMS which accepts BLOBS and/or CLOBS) otherwise value is 'DEFAULT'
	 * @param type Possible values are 'DEFAULT', 'BLOB' and 'CLOB'
	 */
	public void setPeerType(String peerType)
	{
		_peerType = peerType;
	}

	/**
	 * Returns this peer field as a String.
	 * @return This peer field returned as a string.
	 */
	public String toString()
	{
		return "id = " + _id + "\tvalueClass = " + _valueClass + "\tformat = " + _format;
	}
	
	/**
	 * Add a constraint to the list
	 * @param constraint new constraint to add
	 * @see removeConstraint
	 * @see hasConstraint
	 */
	public void addContraint(CONSTRAINT constraint)
	{
		if (constraint != null)
		{
			if (_constraints == null)
				_constraints = new ArrayList<CONSTRAINT>();
			if (!_constraints.contains(constraint))
				_constraints.add(constraint);
		}
	}

	/**
	 * remove a constraints from the list
	 * @param constraint new constraint to remove
	 * @see addConstraint
	 * @see hasConstraint
	 */
	public void removeContraint(CONSTRAINT constraint)
	{
		if (constraint != null)
		{
			if (_constraints != null)
				_constraints.remove(constraint);
		}
	}

	/**
	 * test if the field has a constraint
	 *  @return true if the constraint is found
	 */
	public boolean hasContraint(CONSTRAINT constraint)
	{
		if (constraint == null)
			return false;

		if ((_constraints != null)
				&& (_constraints.contains(constraint)))
			return true;
		else
			return false;
	}

} // end class LyColumnField
