/*
 * LyRowObject.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.util;

import java.util.*;

import leon.data.*;
import leon.peer.*;


/**
 * This class implements instances of LyTableClass. Its main work is to offer a storage for the
 * values and the object id of this instance.
 */
public class LyRowObject
{
	public static final String __VERSION =
		"$Id: LyRowObject.java 34676 2012-03-15 08:05:03Z jlsaouli $";

	/** Peer class to which this object "belongs to". */
	protected LyTableClass     _tableClass;

	/** Object id of this instance. */
	protected LyObjectId	   _objectId;

	/** List of the values. */
	protected Object[]		   _values;

	/** List of the new values : values that are changed but not yet comitted. */
	protected LyValueSet	   _newValues;

	/**
	 * Constructs a new object.
	 */
	public LyRowObject()
	{
		init();
	}
	
	/**
	 * Finalizes peer object.
	 */
	public void finalize()
	{
       LyDataSource dataSource = _tableClass.getDataSource();

        if (dataSource != null)
        {
            // Call contains (that class getFromCache) for this object to avoid keeping references
            // in cache for objects that are finalized (see LySimpleCache).
            LyCache cache = dataSource.getCache();

            if (cache != null)
                cache.contains(this);
        }
	}

	/**
	 * Init method.
	 */
	protected void init()
	{
		_values = null;
		_newValues = null;
		_tableClass = null;
		_objectId = null;
	}

	/**
	 * Constructs a new object for the given class.
	 * @param tableClass Table class to which this object belongs to.
	 */
	public LyRowObject(LyTableClass tableClass)
	{
		this();

		setTableClass(tableClass);
	}

	/**
	 * Constructs a new object for the given class with the specified list of values.
	 * @param tableClass Table class to which this object belongs to.
	 * @param values     List of values of the new instance.
	 */
	public LyRowObject(LyTableClass tableClass, LyValueSet values)
	{
		this(tableClass);

		int                   i, n = _tableClass.getFieldCount();
		LyField               field;
		LyValue               value;
        Map<LyField, LyValue> map = values.getMap();

		for (i = 0; i < n; i++)
		{
			field = _tableClass.getField(i);
			value = map.get(field);

			if (value == null)
				_values[i] = LyPeerClass.VALUE_NOT_LOADED;
			else
				_values[i] = value.getValue();
		}
	}

	/**
	 * Constructs a new object for the given class with the specified values.
	 * @param tableClass Table class to which this object belongs to.
	 * @param values     Values of the new instance.
	 */
	public LyRowObject(LyTableClass tableClass, Object[] values)
	{
		this();

		_tableClass = tableClass;
		_values = values;
	}

	/**
	 * Gets object id of this instance.
	 * @return Object identifier of thsi instance.
	 */
	public LyObjectId getObjectId()
	{
		return _objectId;
	}

	/**
	 * Gets table class to which this object belongs to.
	 * @return Table class of this object.
	 */
	public LyTableClass getTableClass()
	{
		return _tableClass;
	}

	/**
	 * Sets object id for this instance.
	 * @param objectId Object identifier for this instance.
	 */
	protected void setObjectId(LyObjectId objectId)
	{
		_objectId = objectId;
	}

	/**
	 * Sets table class to which this object belongs to.
	 * @param tableClass Table class of this object.
	 */
	public void setTableClass(LyTableClass tableClass)
	{
		_tableClass = tableClass;
		_values = new Object[_tableClass.getFieldCount()];
	}

	/**
	 * Gets value for the specified column field.
	 * @param  columnField Column field for which a value is set on this object.
	 * @return Value corresponding to this field.
	 */
	public Object getValue(LyColumnField columnField)
	{
		if (_newValues != null)
		{
			LyValue value = _newValues.getFieldValue(columnField);

			if (value != null)
				return value.getValue();
		}

		int index = _tableClass.getIndex(columnField);

		if ((index >= _values.length)
            || (_values[index] == LyPeerClass.VALUE_NOT_LOADED))
		{
			String message = "BF AppEngine : Value for columnField "
				+ columnField.getId() + " is not loaded.";

			throw new LyValueNotLoadedException(message);
		}

		return _values[index];
	}

	/**
	 * Sets value of the specified column field.
	 * @param columnField Column field for which a value is set on this object.
	 * @param value       Value corresponding to this field.
	 */
	public void setValue(LyColumnField columnField, Object value)
	{
		if (_newValues == null)
			_newValues = new LyValueSet();

		_newValues.addValue(columnField, value);
	}

	/**
	 * Gets all values of this object.
	 * @return Values set corresponding to all column fields for this object.
	 */
	public LyValueSet getValues()
	{
		LyValueSet        values = new LyValueSet(false);
		LyField           columnField;
		Object		      value;
		Vector<LyField>   fields = _tableClass.getFields();
        int               i, n = fields.size();

		synchronized(this)
		{
			for (i = 0; i < n; i++)
			{
				if ((i < _values.length)
                    && (_values[i] != LyPeerClass.VALUE_NOT_LOADED))
				{
					columnField = fields.elementAt(i);
					value = _values[i];
					values.addValue(columnField, value);
				}
			}
		}
		values.enableCheckField();
		return values;
	}

	/**
	 * Gets new values set on this object. Values set by the setValue method but not yet commited by
	 * validate.
	 * @see LyRowObject#setValue
	 * @see LyRowObject#validate
	 */
	public LyValueSet getNewValues()
	{
		return _newValues;
	}

	/**
	 * To validate the new values and add extra new values. This call commits all changes made on
	 * this object through setValue method and set the extra given values.
	 * @param  values Extra values set on this object.
	 * @return Previous values for changed fields.
	 */
	public LyValueSet validate(LyValueSet values)
	{
		LyValue		  value;
		LyColumnField columnField;
		LyValueSet    oldValues;
		int			  i, n;
		int			  index;

		oldValues = new LyValueSet();

		testCapacity();

		synchronized(this)
		{
			if (values != null)
			{
				n = values.getSize();
				for (i = 0; i < n; i++)
				{
					value = values.getFieldValue(i);
					columnField = (LyColumnField)value.getField();
					index = _tableClass.getIndex(columnField);
					 if ((_values[index] != LyPeerClass.VALUE_NOT_LOADED)
						&& (_values[index] != value.getValue()))
						oldValues.addValue(columnField, _values[index]);
					_values[index] = value.getValue();
				}
			}

			if (_newValues != null)
			{
				n = _newValues.getSize();
				for (i = 0; i < n; i++)
				{
					value = _newValues.getFieldValue(i);
					columnField = (LyColumnField)value.getField();
					index = _tableClass.getIndex(columnField);
					if (index < 0)
						continue;

					if (_values[index] != LyPeerClass.VALUE_NOT_LOADED)
						oldValues.addValue(columnField, _values[index]);
					_values[index] = value.getValue();
				}

				_newValues = null;
			}
		}

		// Consequence of LyObjectSource changes : distinguish no changes
		// null from unknown changes (empty)
		return ((oldValues.getSize() == 0) ? null : oldValues);
	} // end method validate

	/**
	 * Cancels current modifications on this object (i.e. delete new values).
	 */
	public void cancel()
	{
		_newValues = null;
	}

	/**
	 * Tests if the array of values has the same size than the number of fields in the peer class.
	 * If no, the array is increased.
	 * @return the array values
	 */
	protected Object[] testCapacity()
	{
		synchronized(this)
		{
			// A field may have been added : we must resize the array values.
			int n = _tableClass.getFieldCount();

			if (n > _values.length)
			{
				Object[] vals = new Object[_tableClass.getFieldCount()];
				int		 i;

				for (i = 0; i < _values.length; i++)
					vals[i] = _values[i];
				for (; i < vals.length; i++)
					vals[i] = LyPeerClass.VALUE_NOT_LOADED;
				_values = vals;
			}
		}
		return _values;
	}

	/**
	 * Force the value for the specified field to be reload the next time it will be accessed.
	 * @param columnField The field for the value to be unloaded.
	 */
	public void unload(LyColumnField columnField)
	{
		int index = _tableClass.getIndex(columnField);

		_values[index] = LyPeerClass.VALUE_NOT_LOADED;
	}
} // end class LyRowObject
