/*
 * LyTableClass.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.util;

import java.lang.reflect.*;

import leon.data.*;
import leon.misc.LyFileTools;
import leon.peer.*;


/**
 * This class is a simple peer class implementation that uses LyRowObject as wrapper for instances.
 * @see LyRowObject
 */
public class LyTableClass extends LyPeerClass
{
	public static final String __VERSION =
		"$Id: LyTableClass.java 29519 2011-02-07 16:00:36Z jlsaouli $";
	
	private static final String    BLOB_PEER_TYPE	= "BLOB".intern();

	/**
	 * Constructs a table class.
	 */
	public LyTableClass()
	{
		super(null, LyRowObject.class);
	}

	/**
	 * Retrieves the object id for a given object.
	 * @param  object Instance of this class. This object is an instance of LyRowObject.
	 * @return Object identifier of the given object.
	 * @see    LyRowObject#getObjectId
	 */
	protected LyObjectId getId(Object object)
	{
		return ((LyRowObject)object).getObjectId();
	}

	/**
	 * Sets the object id on the given object.
	 * @param object   Instance of this class. This object is an instance of LyRowObject.
	 * @param objectId Object identifier of this instance.
	 * @see   LyRowObject#setObjectId
	 */
	protected void setId(Object object, LyObjectId id)
	{
		((LyRowObject)object).setObjectId(id);
	}

	/**
	 * Gets the value corresponding to the given peer field for an "instance" of this class.
	 * @param  object Instance of this class. This object is an instance of LyRowObject.
	 * @param  field  Peer field for which a value is requested.
	 * @return Value of the given field on the specified instance.
	 * @see    LyRowObject#getValue
	 */
	protected Object get(Object object, LyPeerField field)
	{
		return ((LyRowObject)object).getValue((LyColumnField)field);
	}

	/**
	 * Sets the value corresponding to the given peer field on an "instance" of this class.
	 * @param object Instance of this class. This object is an instance of LyRowObject.
	 * @param field  Field for which a value is set.
	 * @param value  Value set for the given field on the specified instance.
	 * @see   LyRowObject#setValue
	 */
	protected void set(Object object, LyPeerField field, Object value)
	{
		((LyRowObject)object).setValue((LyColumnField)field, value);
	}

	/**
	 * To validate the new values (already known) and add new extra values.
	 * @param  object Instance of this class. This object is an instance of LyRowObject.
	 * @param  values Validated values.
	 * @return A values set containing values that
	 * @see    LyRowObject#validate
	 */
	public LyValueSet validate(Object object, LyValueSet values)
	{
		return ((LyRowObject)object).validate(values);
	}

	/**
	 * Cancels current modifications (i.e. delete new values)
	 * @param object Instance of this class on which modifications are cancelled.
	 */
	protected void cancel(Object object)
	{
		((LyRowObject)object).cancel();
	}

	/**
	 * Gets all field values from an "instance" of this class.
	 * @param  object Instance of this class. This object is an instance of LyRowObject.
	 * @return All the field values of the given instance.
	 * @see    LyRowObject#getValues
	 */
	public LyValueSet getValues(Object object)
	{
		return ((LyRowObject)object).getValues();
	}

	/**
	 * Creates a new instance of this class : creates a LyRowObject.
	 * @return New instance of this class.
	 * @see    LyRowObject#LyRowObject(LyTableClass)
	 */
	public Object newInstance()
		throws InstantiationException,
			IllegalAccessException
	{
		return new LyRowObject(this);
	}

	/**
	 * Recompute the object id of the specified object.
	 */
	public void rehash(Object object)
	{
		LyDataSource dataSource = getDataSource();
		LyCache		 cache = dataSource.getCache();
		LyObjectId   oldObjectId = getObjectId(object);

		((LyRowObject)object).setObjectId(null);
		cache.rehash(object, oldObjectId);
	}
	
	/**
	 * Copies all field values of an instance of this class on another instance.
	 *
	 * @param  object1 Object on which values are copied.
	 * @param  object2 Object from which values are get.
	 * @return Field values that are copied from object2 to object1.
	 */
	public LyValueSet copy(Object object1, Object object2)
	{
		if (object1 == null)
			throw new IllegalArgumentException("BF AppEngine : object1 can't be null");
		if (object2 == null)
			throw new IllegalArgumentException("BF AppEngine : object2 can't be null");

		LyRowObject     rowObject1 = (LyRowObject)object1;
		LyRowObject     rowObject2 = (LyRowObject)object2;
		int             i, n = _fields.length;
		LyValueSet      values = null;
		Object          val1, val2;

		for (i = 0; i < n; i++)
		{
			val2 = rowObject2._values[i];
			if (val2 == LyPeerClass.VALUE_NOT_LOADED)
				continue;
			
			val1 = rowObject1._values[i];
			if (val1 == val2)
				continue;
			if ((val1 != null) && val1.equals(val2))
				continue;
			
			if (val1 != LyPeerClass.VALUE_NOT_LOADED)
			{
				if ((val1 != null) 
						&& (val2 != null)
						&& val1.getClass().isArray() 
						&& val2.getClass().isArray()
						&& isSameArray(val1, val2))
					continue;

				// test BLOB
				LyColumnField field = (LyColumnField)rowObject1.getTableClass().getField(i);

				if (field.getPeerType() == BLOB_PEER_TYPE && (val1 != null) && (val2 != null))
				{
					// modification may be a fake change : just file name is changed.
					// like file_1.txt and file_2.txt or file.txt and file_1.txt
					// keep new file but the original name
					if (LyFileTools.compareTempFiles(val1.toString(), val2.toString()))
						continue;// keep val1
				}
				if (values == null)
					values = new LyValueSet(false);
				values.addValue(_fields[i], val1);
			}
			
			rowObject1._values[i] = val2;
		}

		if (values == null)
			return null; // No change
		
		values.enableCheckField();
		return values;
	} // end method copy

	/**
	 * Compare the content of two array values.
	 * 
	 * @param val1 the first array
	 * @param val2 the second array
	 * @return true if both array are the same, false otherwise.
	 */
	private boolean isSameArray(Object val1, Object val2)
	{
		if (Array.getLength(val1) != Array.getLength(val2))
			return false;
		
		int    i, n;
		Object subVal1, subVal2;

		n = Array.getLength(val1);
		for (i = 0; i < n; i++)
		{
			subVal1 = Array.get(val1, i); 
			subVal2 = Array.get(val1, i); 
			
			if (subVal1 == subVal2)
				continue;
			else if ((subVal1 != null) && (subVal1.equals(subVal2)))
				continue;
			else
				return false; // the content of the arrays are different...
		}
		
		return true;
	}
} // end class LyTableClass
