/**
 * LyTableDataProvider.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.util;

import leon.app.*;
import leon.info.*;
import leon.info.binding.*;
import leon.info.location.*;
import leon.peer.*;


/**
 * This is an abstract class that must be specialized.
 * <P>This data provider is based on leon.peer.util classes which are used by several data providers
 * such as File, XML and SQL data providers.</P>
 * <P>It may also be used as the basis of specific data providers which need a LyClass and LyField
 * implementation and use generic code to store data in memory.</P>
 * <P>The implementation (subclass) must implement the createConnection() and getRequestEncoder()
 * methods which are specific to the nature of their connector.</P>
 * @see LyTableClass
 * @see LyColumnField
 * @see LyRowObject
 */
public abstract class LyTableDataProvider extends LyPeerDataProvider
{
	public static final String __VERSION =
		"$Id: LyTableDataProvider.java 40908 2012-09-24 11:57:26Z jlsaouli $";

	/**
	 * Constructor This constructor is used by introspection mechanisms in the generic code.
	 * @param application reference to the application instance.
	 * @param location    reference to the location information retreived from the data description
	 *                    in the meta model.
	 */
	public LyTableDataProvider(LyApplication application, LyLocation location)
	{
		super(application, location);
	}

	/**
	 * Constructor This constructor is used by introspection mechanisms in the generic code.
	 * @param      application reference to the application instance.
	 * @param      id          identifier of this provider
	 * @deprecated should use LyTableDataProvider(LyApplication application, LyLocation location)
	 *             instead of this constructor.
	 */
	public LyTableDataProvider(LyApplication application, String id)
	{
		super(application, id);
	}

	/**
	 * Determine the class for memory storage associated to a columnField. This method is internally
	 * used by the validate method. Method should be overriden for optimization by subclasses.
	 * @param  tableClass  the peer table class reference as specified in validate method.
	 * @param  columnField peer field associated with the value we store.
	 * @param  fieldInfo   data description of the meta model.
	 * @return the java class used to store data in memory.
	 * @see    #validate()
	 */
	protected Class<?> getValueClass(LyTableClass tableClass, LyColumnField columnField,
		LyFieldInfo fieldInfo)
	{
		Class<?> valueClass = null;

		if (fieldInfo.isMultiple())
			valueClass = String[].class;
		else if (fieldInfo instanceof LyArrayFieldInfo)
		{
			LyFieldInfo fi = ((LyArrayFieldInfo) fieldInfo).getFieldInfo();
			if(fi.isMultiple())
				valueClass = String[][].class;
			else
				valueClass = String[].class;
		}
		else if (fieldInfo instanceof LyDateFieldInfo)
			valueClass = fieldInfo.getValueClass();
		else if (fieldInfo instanceof LyNumberFieldInfo)
			valueClass = fieldInfo.getValueClass();
		else
			valueClass = String.class;

		return valueClass;
	}

	/**
	 * Create a peer class with the specified id for the given class binding.
	 * @param  id        Identifier of the new peer class.
	 * @param  classInfo Class binding for which a peer class is created.
	 * @return The new peer class.
	 */
	protected LyPeerClass createPeerClass(String id, LyClassBinding classBinding)
	{
		// Creating the peer class.
		String		 binding = classBinding.getBinding();
		LyTableClass tableClass = createTableClass(classBinding);

		tableClass.setId(id);
		tableClass.setName(binding);
		return tableClass;
	}

	/**
	 * Factory to create a new LyTableClass instance May be specialized in subclasses, default
	 * return a LyTableClass instance.
	 * @param  classBinding reference of the classBinding associated with the tableClass to create
	 * @return a LyTableClass instance
	 */
	protected LyTableClass createTableClass(LyClassBinding classBinding)
	{
		return (new LyTableClass());
	}

	/**
	 * Create a peer field with the specified id for the field info of the given class binding.
	 * @param  id           Identifier of the new peer field.
	 * @param  fieldInfo    Field info for which a peer field is created.
	 * @param  classBinding Class binding to which the field info belongs to.
	 * @return The new peer field.
	 */
	protected LyPeerField createPeerField(String id, LyFieldInfo fieldInfo,
		LyClassBinding classBinding)
	{
		// Creating the field peer.
		LyFieldBinding fieldBinding = fieldInfo.getBinding();

		// Check that the binding is not a F_BINDS. If It's the case, we have to give the correct sub binding.
		if (fieldBinding instanceof LyFieldsBinding)
		{
			// we must find the correct binding
			LyFieldsBinding fbinds = (LyFieldsBinding)fieldBinding;

			for (int i = 0; i < fbinds.getFieldBindingCount(); i++)
			{
				fieldBinding = fbinds.getFieldBinding(i);
				if (id.equals(fieldBinding.getBinding()))
					break;
				else
					fieldBinding = null;
			}
		}

		LyColumnField	columnField = createColumnField(fieldBinding);
		Class<?>		valueClass = null;
		LyTableClass 	tableClass = (LyTableClass)classBinding.getPeerClass();

		columnField.setId(id);

		// If a javaType is provided, use it.
		if ((fieldBinding != null) && (fieldBinding.getJavaType() != null))
		{
			String	    javaType = fieldBinding.getJavaType();
			ClassLoader appClassLoader = _application.getClassLoader();
			// If we have a classLoader on the application, use it.
			try
			{
				if (appClassLoader != null)
					valueClass = Class.forName(javaType, false, appClassLoader);
				else
					valueClass = Class.forName(javaType);
			}
			catch (ClassNotFoundException e)
			{
				_application.getEnvironment().logError("The java type for the field "
					+ fieldInfo.getId() + " (with binding " + id + ") is invalid. Ignoring it.");
				valueClass = null;
			}
		}
		// BLOB/CLOB only apply on TEXT and FILE fields, and use null when DEFAULT ton speed up test in the request encoder
		if ((fieldBinding != null) 
				&& ((fieldInfo instanceof LyTextFieldInfo)
					||  (fieldInfo instanceof LyFileFieldInfo))
				&& (fieldBinding.getPeerType() != null)
				&& (!"DEFAULT".equals(fieldBinding.getPeerType())))
		{
			String	    peerType = fieldBinding.getPeerType();
			
			columnField.setPeerType(peerType);
		}
		if (valueClass == null)
			valueClass = getValueClass(tableClass, columnField, fieldInfo);
		columnField.setValueClass(valueClass);

		return columnField;
	} // end method createPeerField

	/**
	 * Factory to create a new LyColumnField instance May be specialized in subclasses, default
	 * return a LyColumnField instance.
	 * @param  fieldBinding reference of the fieldBinding associated with the columnField to create
	 * @return a LyColumnField instance
	 */
	protected LyColumnField createColumnField(LyFieldBinding fieldBinding)
	{
		return (new LyColumnField());
	}

	/**
	 * Register peer fields on the given class binding corresponding to the given field info.
	 * @param classBinding Class binding on the which the peer fields are registered.
	 * @param fieldInfo    Field info to which the peer fields corresponds to.
	 * @param peerField    Peer field added to the class binding.
	 */
	protected void registerPeerField(LyClassBinding classBinding,
		LyFieldInfo fieldInfo, LyPeerField peerField)
	{
		// Associate binding and peer fields.
		super.registerPeerField(classBinding, fieldInfo, peerField);

		if (peerField != null)
		{
    		LyFieldBinding fieldBinding = fieldInfo.getBinding();
    
    		((LyColumnField)peerField).setDefault(fieldBinding.getDefault());
		}
	}
	
    /**
     * Test if peer field is compatible with fieldInfo in given classInfo.
     */
    protected void checkPeerField(LyClassInfo classInfo, LyFieldInfo fieldInfo,
        LyPeerClass peerClass, LyPeerField peerField, LyFieldBinding fieldBinding)
    {
        super.checkPeerField(classInfo, fieldInfo, peerClass, peerField, fieldBinding);
        
        // Check class
        Class<?>   valueClass = getValueClass((LyTableClass)peerClass, (LyColumnField)peerField, fieldInfo);
        
        if (valueClass != peerField.getValueClass())
            _application.getEnvironment().logError("BF AppEngine : binding \"" + peerField.getId()
                + "\" is used several times for different attributes that need different types of values \""
                + valueClass + "\" and \"" + peerField.getValueClass()
                + "\", check consistency of all fields using this binding for class " + peerClass + ".");
    }
} // end class LyTableDataProvider
