/**
 * LyImportXmlDataProvider.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyImportXmlDataProvider.java 28997 2010-12-23 10:17:15Z bmichel $
 */
package leon.peer.xml;

import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.*;
import leon.info.location.*;
import leon.peer.*;
import leon.peer.util.*;


/**
 * This class is used by DataSources to transmit their requests to an instance of RequestEncoder.
 * @author Lyria
 * @see    LyXmlExportRequestEncoder
 */
public class LyImportXmlDataProvider extends LyXmlDataProvider
{
	public static final String __VERSION =
		"$Id: LyImportXmlDataProvider.java 28997 2010-12-23 10:17:15Z bmichel $";


	/** The xml file to import. */
	protected String		   _filePath;

	/** The xml buffer to import. */
	protected String		   _buffer;

	/* Constructor */

	/**
	 * Constructor.
	 */
	public LyImportXmlDataProvider(LyApplication application, LyLocation location)
	{
		super(application, location);

		_buffer = null;
	}


	/**
	 * Create a specific request encoder for the specified dataSource.
	 * <P>This method is used internally, but subclasses should override this method and create an
	 * instance of their own (specific) request encoder.</P>
	 * @param  dataSource reference to the dataSource associated to this request encoder.
	 * @return the specific LyRequestEncoder for the specified dataSource.
	 */
	public LyRequestEncoder getRequestEncoder(LyDataSource dataSource)
	{
		return new LyXmlExportRequestEncoder(this, dataSource, _fieldSep, _valueSep);
	}

	/**
	 * Set the path of the file to import.
	 * @param filePath The path of the file to import.
	 */
	public void setFilePath(String filePath)
	{
		_filePath = filePath;
	}

	/**
	 * Set the buffer to import.
	 * @param buffer The buffer to import.
	 */
	public void setBuffer(String buffer)
	{
		_buffer = buffer;
	}

	/**
	 * Create an object source for a given class info.
	 * @param  classInfo The concerned classInfo.
	 * @return An object source for the class info.
	 */
	protected LyObjectSource createObjectSource(LyClassInfo classInfo)
	{
		return new LyTemporaryObjectSource(_application.getInfoProvider(), classInfo);
	}

	/**
	 * Create a connection for the xml import.
	 * @return The connection.
	 */
	public LyConnection createConnection()
	{
		LyImportXmlConnection connection = new LyImportXmlConnection(_application, _filePath);

		if (_buffer != null)
			connection.setBuffer(_buffer);

		return connection;
	}

	/**
	 * Create a peer class with the specified id for the given class binding.
	 * @param  id        Identifier of the new peer class.
	 * @param  classInfo Class binding for which a peer class is created.
	 * @return The new peer class.
	 */
	protected LyPeerClass createPeerClass(String id, LyClassBinding classBinding)
	{
		// Creating the peer class.
		String		 binding = classBinding.getBinding();
		LyTableClass tableClass = createTableClass(classBinding);
		int			 cutIndex = binding.indexOf(":");

		tableClass.setId(id);
		tableClass.setName(binding.substring(0, cutIndex));
		return tableClass;
	}

	/**
	 * Create a peer field with the specified id for the field info of the given class binding.
	 * @param  id           Identifier of the new peer field.
	 * @param  fieldInfo    Field info for which a peer field is created.
	 * @param  classBinding Class binding to which the field info belongs to.
	 * @return The new peer field.
	 */
	protected LyPeerField createPeerField(String id, LyFieldInfo fieldInfo,
		LyClassBinding classBinding)
	{
		// Creating the field peer.
		LyFieldBinding fieldBinding = fieldInfo.getBinding();
		LyXmlField     xmlField = createXmlField(fieldBinding);
		Class<?>	   valueClass;
		LyXmlClass     xmlClass = (LyXmlClass)classBinding.getPeerClass();

		xmlField.setId(id);
		valueClass = getValueClass(xmlClass, xmlField, fieldInfo);
		xmlField.setValueClass(valueClass);
		xmlField.setOriginalId(fieldInfo.getId());

		return xmlField;
	}

	/**
	 * Factory to create a new LyXmlField instance.
	 * @param  fieldBinding reference of the fieldBinding associated with the xmlField to create
	 * @return a LyXmlField instance
	 */
	protected LyXmlField createXmlField(LyFieldBinding fieldBinding)
	{
		return (new LyXmlField());
	}

} // end class LyImportXmlDataProvider
