/*
 * LyXmlDataProvider.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.xml;

import leon.app.*;
import leon.data.*;
import leon.info.binding.*;
import leon.info.location.*;
import leon.peer.*;
import leon.peer.util.*;


/**
 * This class is used by DataSources to transmit their requests to an instance of RequestEncoder.
 * @version $Id: LyXmlDataProvider.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     LyXmlRequestEncoder
 * @see     LyXmlConnection
 */
public class LyXmlDataProvider extends LyTableDataProvider
{
	public static final String __VERSION =
		"$Id: LyXmlDataProvider.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The separator for multiple value. */
	protected String		   _valueSep;

	/** The separator for fields. */
	protected String		   _fieldSep;

	/**
	 * Default constructor Initialization is done with the init method.
	 */
	public LyXmlDataProvider(LyApplication application, LyLocation location)
	{
		super(application, location);
	}

	/**
	 * Sets location on provider.
	 * @param location New location for the provider.
	 */
	public void setLocation(LyLocation location)
	{
		_valueSep = ((LyFileLocation)location).getValueSep();
		_fieldSep = ((LyFileLocation)location).getFieldSep();

		super.setLocation(location);
	}

	/**
	 * This class is a factory to create a LyConnection instance. Current implementation only return
	 * null, so this method must be overriden. This instance depends on the nature of the connection
	 * and must provide services such as opening/closing the connection, and (specific) accessor to
	 * the connection service.
	 * <P>This method is ovberriden and used internally. Use getConnection() to access to a
	 * connection instance.</P>
	 * <P>Parent class LyDataProvider implements a connection pool so this Method may be called
	 * several times.</P>
	 * @return a new Connection instance. The connection instance is used in the specific LyRequest
	 *         encoder implementation.
	 * @see    #getConnection
	 * @see    #releaseConnection
	 */
	protected LyConnection createConnection()
	{
		return new LyXmlConnection(_application);
	}

	/**
	 * Create a specific request encoder for the specified dataSource.
	 * <P>This method is used internally, but subclasses should override this method and create an
	 * instance of their own (specific) request encoder.</P>
	 * <P>Default returns a LyRequestEncoder instance.</P>
	 * @param  dataSource reference to the dataSource associated to this request encoder.
	 * @return the specific LyRequestEncoder for the specified dataSource.
	 */
	public LyRequestEncoder getRequestEncoder(LyDataSource dataSource)
	{
		return new LyXmlRequestEncoder(this, dataSource, _valueSep, _fieldSep);
	}

	/**
	 * Factory to create a new LyTableClass instance May be specialized in subclasses, default
	 * return a LyTableClass instance.
	 * @param  classBinding reference of the classBinding associated with the tableClass to create
	 * @return a LyTableClass instance
	 */
	protected LyTableClass createTableClass(LyClassBinding classBinding)
	{
		return (new LyXmlClass());
	}

	/**
	 * Factory to create a new LyColumnField instance May be specialized in subclasses, default
	 * return a LyColumnField instance.
	 * @param  fieldBinding reference of the fieldBinding associated with the columnField to create
	 * @return a LyColumnField instance
	 */
	protected LyColumnField createColumnField(LyFieldBinding fieldBinding)
	{
		return (new LyXmlField());
	}
} // end class LyXmlDataProvider
