/**
 * LyXmlExportRequestEncoder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current version: $Id: LyXmlExportRequestEncoder.java 34773 2012-03-16 15:13:24Z jlsaouli $
 */
package leon.peer.xml;

import java.io.*;
import java.text.*;
import java.util.*;

import javax.xml.parsers.*;
import leon.app.*;
import leon.data.*;
import leon.info.*;
import leon.info.binding.*;
import leon.misc.*;
import leon.peer.*;

import org.xml.sax.*;
import org.xml.sax.helpers.*;


/**
 * This class is used to encode objects in the xml export format.
 *
 * @see    LyRequestEncoder
 * @author Lyria
 */
public class LyXmlExportRequestEncoder extends LyXmlRequestEncoder
{
	public static final String __VERSION =
		"$Id: LyXmlExportRequestEncoder.java 34773 2012-03-16 15:13:24Z jlsaouli $";

	public static final String LOCAL_FIELD_IMPORT_EXTENSION = "_import_local_";

	/**
	 * Return the list of class infos names from the given XML file.
	 *
	 * @param  filePath
	 * @return
	 */
	public static Vector<String> getClassInfosNames(String filePath)
	{
		SAXParserFactory    factory = SAXParserFactory.newInstance();
		_LyClassInfoHandler handler = null;

		factory.setNamespaceAware(true);

		try
		{
			SAXParser saxParser = factory.newSAXParser();

			// Create an instance of our SAX event handler and associate it
			// with the XMLReader
			handler = new _LyClassInfoHandler();

			// Parse the document
			saxParser.parse(new File(filePath), handler);
			
			return handler.getClassInfoNames();
		}
		catch (Exception e)
		{
			return null;
		}
	}

	/**
	 * Return the list of class infos names from the given XML file.
	 *
	 * @param  xml content
	 * @return
	 */
	public static Vector<String> getClassInfosNamesFromBuffer(String xml)
	{
		SAXParserFactory    factory = SAXParserFactory.newInstance();
		_LyClassInfoHandler handler = null;

		factory.setNamespaceAware(true);

		try
		{
			SAXParser saxParser = factory.newSAXParser();

			// Create an instance of our SAX event handler and associate it
			// with the XMLReader
			handler = new _LyClassInfoHandler();

			// Parse the document
			saxParser.parse(new InputSource(new StringReader(xml)), handler);
			
			return handler.getClassInfoNames();
		}
		catch (Exception e)
		{
			return null;
		}
	}

	/**
	 * Handler used to retrieve class infos contained in an XML file.
	 */
	public static class _LyClassInfoHandler extends DefaultHandler
	{
		int			   _xmlDepthLevel = 0;
		Vector<String> _classInfoNames = null;

		public void startDocument()
		{
			_xmlDepthLevel = 0;
			_classInfoNames = new Vector<String>();
		}

		public void startElement(String namespaceURI, String localName, String qName,
			Attributes atts)
			throws SAXException
		{
			if ((_xmlDepthLevel <= 0) && (localName.equals("DATA")))
				_xmlDepthLevel = 1; // Just below DATA
			else if (_xmlDepthLevel == 1)
			{
				if (!_classInfoNames.contains(localName) && (localName.length() > 0))
					_classInfoNames.addElement(localName);

				_xmlDepthLevel++;
			}
			else
				_xmlDepthLevel++;
		}

		public void endElement(String namespaceURI, String simpleName, String qualifiedName)
		{
			_xmlDepthLevel--;
		}

		public Vector<String> getClassInfoNames()
		{
			return _classInfoNames;
		}
	} // end class _LyClassInfoHandler

	/** Create a sub-directory for packages that contain classes. */
	public static boolean SUBDIR_FOR_SCHEMA = true;

	protected String		    _tab;

	protected LyApplication     _application;

	protected boolean		    _skipHeader;

	protected boolean		    _skipFooter;

	/**
	 * Constructor.
	 *
	 * @param provider   The data provider.
	 * @param dataSource The datasource.
	 * @param fieldSep   Separator between fields.
	 * @param valueSep   Separator between values.
	 */
	public LyXmlExportRequestEncoder(
		LyDataProvider provider,
		LyDataSource dataSource,
		String fieldSep,
		String valueSep)
	{
		super(provider, dataSource, fieldSep, valueSep);
		_tab = "";
		_skipHeader = false;
		_skipFooter = false;
		_application = provider.getApplication();
	}

	/**
	 * Write the header of the export xml file.
	 *
	 * @return the header of the export xml file.
	 */
	public String writeHeader()
	{
		StringBuffer result = new StringBuffer(2);

		if (!_skipHeader)
		{
			result.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
			result.append(
				"<DATA xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns='_ExportRootLevel_' xsi:schemaLocation='_ExportRootLevel_ leon/conf/schema/exportRootLevel.xsd'>");
		}
		addIncrement();

		return result.toString();
	}

	/**
	 * Write the footer of the export xml file.
	 *
	 * @return the footer of the export xml file.
	 */
	public String writeFooter()
	{
		removeIncrement();

		if (_skipFooter)
			return "";
		else
			return "</DATA>";
	}

	/**
	 * Read a set of objects and return a set of lines for xml export encoding.
	 *
	 * @param  classInfo the classInfo of the objects to encode.
	 * @param  fields    FieldInfos to save (if null save all).
	 * @return the set of lines describing the objects.
	 */
	public Vector<String> encodeClassInfo(LyClassInfo classInfo, Vector<LyFieldInfo> fields)
	{
		// Get the fields : if fields is empty, get all the FieldInfos from
		// the ClassInfo.
		Vector<LyField> collect;

		if ((fields == null) || (fields.size() == 0))
			collect = new Vector<LyField>(classInfo.getFields());
		else
		{
			@SuppressWarnings("unchecked")
			Vector<LyField> temp = (Vector<LyField>)orderFields(fields);
			collect = temp;
		}

		Vector<String> result = new Vector<String>();
		String		   line;

		// For each object get the ValueSet and encode it.
		// FFT01068 test if objects is not null
		int			   i, no;
		Object		   object;
		LyValueSet     values;
		// TODO : Get the class version when available.
		String		   version = "1.0";
		String		   classId = classInfo.getId();
		String		   schemaPath = getSchemaPath(classInfo);

		result.add(_tab + "<" + classId + " version=\"" + version + "\" xmlns=\"" + classId + "\"");
		result.add(_tab + "\txsi:schemaLocation=\"" + classId + " " + schemaPath + "\">\n");
		addIncrement();

		LyObjectList objects = getObjects(classInfo);

		no = objects.getSize();

		for (i = 0; i < no; i++)
		{
			object = objects.getObject(i);

			if (object == null)
				continue;

			values = classInfo.getValues(object);
			line = encodeValue(values, collect);

			result.addElement(line);
		}
		removeIncrement();
		result.add(_tab + "</" + classId + ">");
		return result;
	} // end method encodeClassInfo

	/**
	 * Accessor on objects to export.
	 *
	 * @param  classInfo current classInfo
	 * @return the list of objects
	 */
	protected LyObjectList getObjects(LyClassInfo classInfo)
	{
		LyClassWorkSpace objects = new LyClassWorkSpace(classInfo.getApplication()
					.getCurrentSession(), classInfo, false);

		return objects;
	}

	/**
	 * Read a set of objects and return a set of lines for xml export encoding.
	 *
	 * @param  objects the set of objects to encode.
	 * @param  fields  FieldInfos to save (if null save all).
	 * @return the set of lines describing the set of objects.
	 */
	public Vector<String> encodeValues(Vector<LyObject> objects, Vector<LyFieldInfo> fields)
	{
		Vector<String> result = new Vector<String>();

		if (objects == null)
			return result;

		// Get the fields : if fields is empty, get all the FieldInfos from
		// the ClassInfo.
		Vector<LyField> collect;

		if ((fields == null) || (fields.size() == 0))
			collect = new Vector<LyField>(_class.getFields());
		else
		{
			@SuppressWarnings("unchecked")
			Vector<LyField> temp = (Vector<LyField>)orderFields(fields);
			collect = temp;
		}

		String	    line;

		// For each object get the ValueSet and encode it.
		// FFT01068 test if objects is not null
		int		    i;
		LyObject    object;
		LyValueSet  values;
		LyClassInfo classInfo;
		String	    version;
		String	    classId;
		String	    schemaPath;

		while (objects.size() > 0)
		{
			object = objects.remove(0);
			if (object == null)
				continue;

			// TODO : Get the class version when available.
			version = "1.0";
			classInfo = object.getClassInfo();
			classId = classInfo.getId();
			schemaPath = getSchemaPath(classInfo);

			result.add(_tab + "<" + classId + " version=\""
				+ version + "\" xmlns=\"" + classId + "\"");
			result.add(_tab + "\txsi:schemaLocation=\""
				+ classId + " " + schemaPath + "\">\n");
			addIncrement();

			values = classInfo.getValues(object);
			line = encodeValue(values, collect);
			result.addElement(line);

			for (i = objects.size() - 1; i >= 0; i--)
			{
				object = objects.get(i);
				if (object == null)
					continue;

				if (object.getClassInfo() == classInfo)
				{
					values = classInfo.getValues(object);
					line = encodeValue(values, collect);
					result.addElement(line);
					objects.remove(i);
				}
			}

			removeIncrement();
			result.add(_tab + "</" + classId + ">");
		}

		return result;
	} // end method encodeValues

	/**
	 * Return a line describing the ValueSet in the xml export format.
	 *
	 * @param  values the ValueSet to encode.
	 * @param  fields the FieldInfos to encode.
	 * @return the line describing the ValueSet.
	 */
	protected String encodeValue(LyValueSet values, Vector<LyField> fields)
	{
		LyField		 field;
		LyValue		 value;
		int			 i;
		int			 na = fields.size();
		StringBuffer result = new StringBuffer();

		result.append(_tab);
		result.append("<OBJECT>\n");
		addIncrement();

		// For each FieldInfo retrieve the value and write it
		for (i = 0; i < na; i++)
		{
			field = fields.elementAt(i);
			value = values.getFieldValue(field);
			result.append(encodeField(field, value));
		}
		removeIncrement();
		result.append(_tab);
		result.append("</OBJECT>\n");

		return (result.toString());
	}

	/**
	 * Encode a field in the xml export format.
	 *
	 * @param  field The field to export.
	 * @return The field encoded.
	 */
	protected String encodeField(LyField field, LyValue value)
	{
		StringBuffer result = new StringBuffer();
		String		 str = null;
		String		 fieldId;
		Object		 val;

		if (value == null)
			return "";

		val = value.getValue();
		if (val == null)
			return "";

		if (field instanceof LyArrayFieldInfo)
			result.append(encodeArrayField((LyArrayFieldInfo)field, (LyArrayValue)value));
		else if ((field instanceof LyChoiceFieldInfo)
				&& (((LyChoiceFieldInfo)field).getType() == LyChoiceFieldInfo.Type.BOOLEAN))
			result.append(encodeChoiceField((LyChoiceFieldInfo)field, (LyBooleanValue)value));
		else if (field instanceof LyChoiceFieldInfo)
			result.append(encodeChoiceField((LyChoiceFieldInfo)field, (LyChoiceValue)value));
		else if (field instanceof LyStructFieldInfo)
			result.append(encodeStructField((LyStructFieldInfo)field, (LyRelationValue)value));
		else if (field instanceof LyRelationFieldInfo)
			result.append(encodeRelationField((LyRelationFieldInfo)field, (LyRelationValue)value));
		else if (field instanceof LyDateFieldInfo)
			result.append(encodeDateField((LyDateFieldInfo)field, (LyDateValue)value));
		else
		{
			fieldId = field.getId();
			result.append(_tab);
			result.append("<").append(fieldId).append(">");

			val = value.getValue();
			// No type, so put -1
			str = toString(field, val, -1);
			result.append(str);

			result.append("</").append(fieldId).append(">\n");
		}

		return result.toString();
	} // end method encodeField

	/**
	 * Encode an array field in the xml export format.
	 *
	 * @param  arrayField The array field to export.
	 * @param  value      The value of the array field.
	 * @return The formatted array field.
	 */
	protected String encodeArrayField(LyArrayFieldInfo arrayField, LyArrayValue value)
	{
		StringBuffer	   result = new StringBuffer();
		String			   arrayId = arrayField.getId();
		LyFieldInfoValue[] values = value.getArrayFieldValue();
		LyFieldInfo		   fieldInfo = arrayField.getFieldInfo();

		result.append(_tab);
		result.append("<").append(arrayId).append(">\n");
		addIncrement();

		for (int i = 0; i < values.length; i++)
			result.append(encodeField(fieldInfo, values[i]));

		removeIncrement();
		result.append(_tab);
		result.append("</").append(arrayId).append(">\n");

		return result.toString();
	}

	/**
	 * Encode a choice field in the xml export format.
	 *
	 * @param  choiceField The choice field to export.
	 * @param  value       The value of the choice field.
	 * @return The formatted choice field.
	 */
	protected String encodeChoiceField(LyChoiceFieldInfo choiceField, LyChoiceValue value)
	{
		StringBuffer result = new StringBuffer();
		String		 choiceId = choiceField.getId();
		Object		 val = value.getValue();
		String		 str = null;

		if (!choiceField.isMultiple())
		{
			result.append(_tab);
			result.append("<").append(choiceId).append(">");

			LyChoiceItem item = choiceField.getChoiceItem(((Short)val).shortValue());

			val = item.getValue();

			// No type here.
			str = toString(choiceField, val, -1);
			result.append(str);

			result.append("</").append(choiceId).append(">\n");
		}
		else
		{
			BitSet		 bs = (BitSet)val;
			int			 j, size = bs.size();
			LyChoiceItem item;

			for (j = 0; j < size; j++)
			{
				if (bs.get(j))
				{
					result.append(_tab);
					result.append("<").append(choiceId).append(">");
					item = choiceField.getChoiceItem((short)j);
					str = toString(choiceField, item.getValue(), -1);
					result.append(str);
					result.append("</").append(choiceId).append(">\n");
				}
			}
		}

		return result.toString();
	} // end method encodeChoiceField

	/**
	 * Encode a choice field in the xml export format.
	 *
	 * @param  choiceField The choice field to export.
	 * @param  value       The value of the choice field.
	 * @return The formatted choice field.
	 */
	protected String encodeChoiceField(LyChoiceFieldInfo choiceField,
		LyBooleanValue value)
	{
		StringBuffer result = new StringBuffer();
		String		 choiceId = choiceField.getId();
		boolean		 val = value.booleanValue();
		LyChoiceItem item;
		String		 str = null;

		result.append(_tab);
		result.append("<").append(choiceId).append(">");
		item = choiceField.getChoiceItem((short)(val ? 1 : 0));
		str = toString(choiceField, item.getValue(), -1);
		result.append(str);
		result.append("</").append(choiceId).append(">\n");

		return result.toString();
	}

	/**
	 * Encode a struct field in the xml export format.
	 *
	 * @param  structField The struct field to export.
	 * @param  value       The value of the struct field.
	 * @return The formatted struct field.
	 */
	protected String encodeStructField(LyStructFieldInfo structField, LyRelationValue value)
	{
		StringBuffer result = new StringBuffer();
		String		 structId = structField.getId();
		LyObjectId   objectId = value.getObjectId(0);
		StringBuffer str = new StringBuffer();
		LyObject     object = value.getObject(0);

		if (object == null)
		{
			// The value is not available
			_environment.logWarning("[LyXmlExportRequestEncoder] encodeStructField "
				+ structField.getId() + " struct object is not found : " + objectId);
			return "";
		}
		result.append(_tab);
		result.append("<").append(structId).append(">\n");
		addIncrement();

		LyClassInfo classInfo = objectId.getClassInfo();

		result.append(_tab);
		result.append("<classId>").append(classInfo.getId()).append("</classId>\n");

		Vector<LyField> fields = new Vector<LyField>(classInfo.getFields());
		LyValueSet	    values = classInfo.getValues(object);

		for (int i = 0; i < fields.size(); i++)
		{
			LyField field = fields.get(i);

			str.append(encodeField(field, values.getFieldValue(field)));
		}

		result.append(str.toString());
		removeIncrement();
		result.append(_tab);
		result.append("</").append(structId).append(">\n");

		return result.toString();
	} // end method encodeStructField

	/**
	 * Encode a relation field in the xml export format.
	 *
	 * @param  relationField The relation field to export.
	 * @param  value         The value of the relation field.
	 * @return The formatted relation field.
	 */
	protected String encodeRelationField(LyRelationFieldInfo relationField,
		LyRelationValue value)
	{
		int			 i, relationCount = value.getValueCount();
		StringBuffer result = new StringBuffer();
		String		 relationId = relationField.getId();
		LyObject     object;
		LyObjectId   id;

		for (i = 0; i < relationCount; i++)
		{
			id = value.getObjectId(i);
			object = value.getObject(i);

			if ((id == null) && (object == null))
				continue;
			
			if ((object == null) && getEnvironment().testResource("LY_HIDE_UNKNOWN_OBJECTS"))
				continue;
			
			result.append(_tab);
			result.append("<").append(relationId).append(">\n");
			addIncrement();
			result.append(_tab);
			result.append("<classId>");
			if (object != null)
				result.append(object.getClassInfo().getId());
			else
				result.append(id.getClassInfo().getId()); // FFT01597
			result.append("</classId>\n");
			result.append(_tab);
			result.append("<objectId>");
			result.append(toString(relationField, id.toString(), -1));
			result.append("</objectId>\n");
			removeIncrement();
			result.append(_tab);
			result.append("</").append(relationId).append(">\n");
		}

		return result.toString();
	} // end method encodeRelationField

	/**
	 * Encode a date field in the xml export format.
	 *
	 * @param  dateField The date field to export.
	 * @param  value     The value of the date field.
	 * @return The formatted date field.
	 */
	protected String encodeDateField(LyDateFieldInfo dateField, LyDateValue value)
	{
		StringBuffer		 result = new StringBuffer();
		String				 dateId = dateField.getId();
		Date				 date = value.getDateValue();
		LyDateFieldInfo.Type type = dateField.getType();
		SimpleDateFormat     dateFormat;

		result.append(_tab);
		result.append("<").append(dateId).append(">");

		if (type == LyDateFieldInfo.Type.DATE)
		{
			dateFormat = new SimpleDateFormat("yyyy-MM-dd");
			result.append(dateFormat.format(date));
		}
		else if ((type == LyDateFieldInfo.Type.TIME) || (type == LyDateFieldInfo.Type.HOUR_MIN))
		{
			dateFormat = new SimpleDateFormat("HH:mm:ss");
			result.append(dateFormat.format(date));
		}
		else
		{
			dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
			result.append(dateFormat.format(date));
		}

		result.append("</").append(dateId).append(">\n");

		return result.toString();
	} // end method encodeDateField

	protected void addIncrement()
	{
		_tab += "\t";
	}

	protected void removeIncrement()
	{
		int n = _tab.indexOf("\t");

		if (n == 0)
			_tab = _tab.substring(1);
		else
			_tab = _tab.substring(0, n) + _tab.substring(n + 1);
			// JDK 1.4 or above : _tab = _tab.replaceFirst("\t", "");
	}

	/**
	 * Get the path of the schema file for a given class.
	 *
	 * @param  classInfo The classInfo concerned.
	 * @return A String which contains the path of the schema file.
	 */
	protected String getSchemaPath(LyClassInfo classInfo)
	{
		StringBuffer path = new StringBuffer();

		if (SUBDIR_FOR_SCHEMA)
		{
			LyInfo parent = classInfo.getParent(); // LyProject or LyInclude

			while (parent != null)
			{
				if (parent instanceof LyProject)
					path.insert(0, parent.getId() + '/');
				parent =  parent.getParent();
			}
		}
		path.append(classInfo.getId());
		path.append(".xsd");

		return path.toString();
	}

	/**
	 * Order the fields.
	 *
	 * @param  classInfo A Vector which contains the concerned fields.
	 * @return A Vector which contains the fields ordered.
	 */
	
	protected Vector<? extends LyField> orderFields(Vector<?> fields)
	{
		Vector<? extends LyField> ordered = _class.getFields();
		Object field;

		for (int i = ordered.size() - 1; i >= 0; i--)
		{
			field = ordered.get(i);
			if (!fields.contains(field))
				ordered.remove(i);
		}

		return ordered;
	}

	/**
	 */
	public Object fromString(LyField field, String str)
	{
		if ((field == null) || (str == null))
			return null;

		Class<?>  valueClass;
		Object value;

		valueClass = field.getValueClass();

		// If we must change the date.
		if ((valueClass == Date.class) && (str.indexOf("-") > 0))
		{
			str = str.replace('T', ' ');
			str = str.replace('-', '/');

			StringTokenizer tok = new StringTokenizer(str, "/");
			String		    day = "";
			String		    month = "";
			String		    year = "";
			int			    count = 0;

			while ((tok.hasMoreTokens()) && (count < 3))
			{
				String token = tok.nextToken();

				switch (count)
				{
				case 0:
					year = token;
					break;
				case 1:
					month = token;
					break;
				case 2:
					day = token.substring(0, 2);
				}
				count++;
			}

			StringBuffer date = new StringBuffer(5);

			date.append(day);
			date.append("/");
			date.append(month);
			date.append("/");
			date.append(year);

			int spaceIndex = str.indexOf(" ");

			if (spaceIndex > 0)
				str = date.toString() + str.substring(spaceIndex);
			else
				str = date.toString();

		}

		value = super.fromString(field, str);

		return value;
	} // end method fromString

	/**
	 * Return the default handler for this provider.
	 *
	 * @return The default handler.
	 */
	public DefaultHandler buildHandler()
	{
		return new _LyXmlDefaultHandler();
	}

	/**
	 * DefaultHandler.*
	 */
	protected class _LyXmlDefaultHandler extends DefaultHandler
	{

		/** Instance associated with the root element of the XML tree. */
		protected Object	    _rootElement;

		/** Instance of current element being parsed. */
		protected _LyTreeStruct _currentLevel;

		/**
		 * Indicates if the value and the end element have to be checked. Equals 0 if they have to
		 * be checked. Otherwise, they won't.
		 */
		protected int		    _must;

		/** Says if the current class is the good one. */
		protected boolean	    _isGoodClass = false;

		/** The classInfo to import. */
		protected LyClassInfo   _classInfo;

		/** The buffer to stock attribute value. */
		protected StringBuffer  _attribute = new StringBuffer();

		public void startDocument()
			throws SAXException
		{
			// Initialize the vector to stock objects.
			_objects = new Vector<LyXmlObject>();
			_must = 0;
			_classInfo = _application.getClassInfo(((LyXmlClass)_class).getName());
			_environment = _application.getEnvironment();
		}

		public void characters(char[] buf, int offset, int len)
			throws SAXException
		{
			String s = new String(buf, offset, len);

			if ((!s.trim().equals("")) && (_must == 0) && (_isGoodClass))
				// we concat the value with the new value,
				// because sax may have splitted the value
				_attribute.append(s);
		}

		
		public void startElement(String uri, String localName, String qName,
			Attributes attributes)
			throws SAXException
		{
			_attribute = new StringBuffer();

			if (qName != null)
			{
				LyXmlClass xmlClass = (LyXmlClass)_class;

				if (qName.equals(xmlClass.getName()))
					startClassElement();
				else if (qName.equals("DATA"))
				{
				}
				else if ((qName.equals("OBJECT")) && (_isGoodClass))
					startObjectElement(xmlClass);
				// Nothing to do
				else if (qName.equals("classId"))
				{
				}
				// A relation
				else if ((qName.equals("objectId")) && (_isGoodClass))
					_currentLevel._realType = _LyTreeStruct.CURRENT_NODE_TYPE_RELATION;
				else if (_isGoodClass)
				{
					LyXmlField    current = null;
					LyXmlField    testField;
					_LyTreeStruct child;
					LyFieldInfo   fieldInfo = (LyFieldInfo)_application.getInfo(qName);

					if (fieldInfo instanceof LyArrayFieldInfo)
						startArrayElement(qName);
					else if (fieldInfo instanceof LyStructFieldInfo)
					{
						short	   relativeType;
						LyXmlField field;

						for (int i = 0; i < _fields.size(); i++)
						{
							field = (LyXmlField)_fields.elementAt(i);
							if (field.getOriginalId().equals("__" + qName))
							{
								current = field;
								break;
							}
						}

						if (current == null)
						{
							current = new LyXmlField();
							current.setOriginalId(qName);
							current.setId(qName);
						}

						if (_currentLevel._realType == _LyTreeStruct.CURRENT_NODE_TYPE_ARRAY)
							relativeType = _LyTreeStruct.CURRENT_NODE_TYPE_ARRAY_ELEMENT;
						else if (_currentLevel._realType == _LyTreeStruct.CURRENT_NODE_TYPE_STRUCT)
							relativeType = _LyTreeStruct.CURRENT_NODE_TYPE_STRUCT_ELEMENT;
						else
							relativeType = _LyTreeStruct.CURRENT_NODE_TYPE_OBJECT_ELEMENT;

						child = new _LyTreeStruct(_currentLevel, current, relativeType,
								_LyTreeStruct.CURRENT_NODE_TYPE_STRUCT,
								new Vector<_LyTreeStruct>());
						_currentLevel._childTrees.add(child);
						_currentLevel = child;
					}
					else
					{

						Vector<?> fields = ((_currentLevel._fields == null) ? _fields
																		 : _currentLevel._fields);

						for (int i = 0; i < fields.size(); i++)
						{
							testField = (LyXmlField)fields.elementAt(i);
							// A field of the _rootElement.
							if (testField.getOriginalId().equals(qName))
							{
								current = testField;
								if (_currentLevel._realType
										== _LyTreeStruct.CURRENT_NODE_TYPE_OBJECT)
								{
									child = new _LyTreeStruct(_currentLevel, current,
											_LyTreeStruct.CURRENT_NODE_TYPE_OBJECT_ELEMENT,
											_LyTreeStruct.CURRENT_NODE_TYPE_ELEMENT, null);
									_currentLevel._childTrees.add(child);
									_currentLevel = child;
								}
								break;
							}
						}
						// A field of a struct or array element
						if (current == null)
						{
							if ((_currentLevel._realType == _LyTreeStruct.CURRENT_NODE_TYPE_STRUCT)
									|| (_currentLevel._realType
										== _LyTreeStruct.CURRENT_NODE_TYPE_ARRAY))
							{
								short relativeType;

								if (_currentLevel._realType
										== _LyTreeStruct.CURRENT_NODE_TYPE_STRUCT)
									relativeType = _LyTreeStruct.CURRENT_NODE_TYPE_STRUCT_ELEMENT;
								else
									relativeType = _LyTreeStruct.CURRENT_NODE_TYPE_ARRAY_ELEMENT;

								LyXmlField field = new LyXmlField();

								field.setOriginalId(qName);
								field.setId(qName);
								child = new _LyTreeStruct(_currentLevel, field, relativeType,
										_LyTreeStruct.CURRENT_NODE_TYPE_ELEMENT, null);

								_currentLevel._childTrees.add(child);
								_currentLevel = child;
							}
							else
							{
								boolean found = false;
								// Maybe this is an imported local field
								// try again with extension...
								if (fieldInfo instanceof LyRelationFieldInfo)
								{
									for (int i = 0; i < _fields.size(); i++)
									{
										testField = (LyXmlField)_fields.elementAt(i);
										// A field of the _rootElement.
										if (testField.getOriginalId().equals(
													qName + LOCAL_FIELD_IMPORT_EXTENSION))
										{
											current = testField;
											if (_currentLevel._realType
													== _LyTreeStruct.CURRENT_NODE_TYPE_OBJECT)
											{
												child = new _LyTreeStruct(_currentLevel, current,
														_LyTreeStruct
															.CURRENT_NODE_TYPE_OBJECT_ELEMENT,
															_LyTreeStruct.CURRENT_NODE_TYPE_ELEMENT,
															null);
												_currentLevel._childTrees.add(child);
												_currentLevel = child;
											}
											found = true;
											break;
										}
									}
								}

								if (!found)
								{
									if (_currentLevel._realType
											== _LyTreeStruct.CURRENT_NODE_TYPE_OBJECT)
									{
										String message = _environment.getMessage(
												"LY_ER_IMPORT_UNKNOWN_FIELD", qName, _classInfo);

										_environment.logWarning(message);
									}
									_must++;
								}
							}
						}
					}
				}
			}
		} // end method startElement

		public void endElement(String uri, String localName, String qName)
			throws SAXException
		{
			String value = _attribute.toString();

			if ((_currentLevel != null) && (_currentLevel._values != null))
				_currentLevel._values.add(value);

			_attribute = new StringBuffer();

			if (!_isGoodClass)
				return;

			if (qName != null)
			{
				LyXmlClass xmlClass = (LyXmlClass)_class;

				if ((qName.equals(xmlClass.getName())) && (_isGoodClass))
					endClassElement();
				else if (qName.equals("DATA"))
				{
				}
				else if ((qName.equals("OBJECT")) && (_isGoodClass))
					endObjectElement();
				else
				{
					if (_currentLevel._current instanceof LyXmlField)
					{
						LyXmlField xmlField = (LyXmlField)_currentLevel._current;

						if (xmlField.getOriginalId().equals(qName)
								|| xmlField.getOriginalId().equals("__" + qName)
								|| xmlField.getOriginalId().equals(
									qName + LOCAL_FIELD_IMPORT_EXTENSION))
						{
							if (_must == 0)
								_currentLevel = _currentLevel._parentTree;
						}
					}
					if (_must > 0)
						_must--;
				}
			}
		} // end method endElement

		// treat validation errors as fatal
		public void error(SAXParseException e)
			throws SAXParseException
		{
			throw e;
		}

		// dump warnings too
		public void warning(SAXParseException err)
			throws SAXParseException
		{
			getEnvironment().logWarning("line "
				+ err.getLineNumber()
				+ ", uri "
				+ err.getSystemId()
				+ "\n\t"
				+ err.getMessage());
		}

		protected void startClassElement()
		{
			_isGoodClass = true;
		}

		protected void startObjectElement(LyXmlClass xmlClass)
		{
			LyXmlObject current = new LyXmlObject(xmlClass);

			_currentLevel = new _LyTreeStruct(null, current, _LyTreeStruct.CURRENT_NODE_TYPE_OBJECT,
					_LyTreeStruct.CURRENT_NODE_TYPE_OBJECT, new Vector<_LyTreeStruct>());
			// Set tree root
			_rootElement = current;
		}

		protected void startArrayElement(String id)
		{
			short	   relativeType;
			LyXmlField field, current = null;

			for (int i = 0; i < _fields.size(); i++)
			{
				field = (LyXmlField)_fields.elementAt(i);
				if (field.getOriginalId().equals(id))
				{
					current = field;
					break;
				}
			}

			if (current == null)
			{
				current = new LyXmlField();
				current.setOriginalId(id);
				current.setId(id);
			}

			if (_currentLevel._realType == _LyTreeStruct.CURRENT_NODE_TYPE_ARRAY)
				relativeType = _LyTreeStruct.CURRENT_NODE_TYPE_ARRAY_ELEMENT;
			else if (_currentLevel._realType == _LyTreeStruct.CURRENT_NODE_TYPE_STRUCT)
				relativeType = _LyTreeStruct.CURRENT_NODE_TYPE_STRUCT_ELEMENT;
			else
				relativeType = _LyTreeStruct.CURRENT_NODE_TYPE_OBJECT_ELEMENT;

			_LyTreeStruct child = new _LyTreeStruct(_currentLevel, current, relativeType,
					_LyTreeStruct.CURRENT_NODE_TYPE_ARRAY, new Vector<_LyTreeStruct>());

			_currentLevel._childTrees.add(child);
			_currentLevel = child;
		} // end method startArrayElement

		protected void endClassElement()
		{
			_isGoodClass = false;
		}

		protected void endObjectElement()
		{
			Object		  value = null;
			Vector<_LyTreeStruct>	childs = _currentLevel._childTrees;
			LyXmlField    xmlField;
			_LyTreeStruct child;
			LyXmlField    nextXmlField;
			String		  xmlFieldId;
			String		  nextXmlFieldId;
			_LyTreeStruct nextChild;

			if (childs == null)
				return;

			LyClassInfo classInfo;

			if (_currentLevel._classInfo != null)
				classInfo = _currentLevel._classInfo;
			else
				classInfo = _classInfo;

			// For each child element (field)
			for (int i = 0; i < childs.size();)
			{
				child = childs.remove(0);
				xmlField = (LyXmlField)child._current;
				value = null;

				// If it's a struct, we must construct the object
				if (child._realType == _LyTreeStruct.CURRENT_NODE_TYPE_STRUCT)
				{
					xmlField.setValueClass(LyObjectId.class);
					value = processStructElement(child);
				}
				// If it's an array, we must construct the object
				else if (child._realType == _LyTreeStruct.CURRENT_NODE_TYPE_ARRAY)
				{
					String			 arrayId = ((LyXmlField)child._current)._originalId;
					LyArrayFieldInfo arrayFieldInfo = (LyArrayFieldInfo)_application.getInfo(
							arrayId);

					xmlField.setValueClass(arrayFieldInfo.getValueClass());
					value = processArrayElement(child);
				}
				// If it's a relation
				else if (child._realType == _LyTreeStruct.CURRENT_NODE_TYPE_RELATION)
				{
					String			    relationId = ((LyXmlField)child._current)._originalId;
					LyRelationFieldInfo relation = (LyRelationFieldInfo)_application.getInfo(
							relationId);

					if (relation.isMultiple())
					{
						xmlField.setValueClass(LyObjectId[].class);

						Vector<LyObjectId> objectIdVector = new Vector<LyObjectId>();
						LyObjectId		   objectId;

						// first element
						String			   classId = child._values.remove(0);
						LyClassInfo		   classInfo2 = _application.getClassInfo(classId);

						if (child._values.size() > 0)
							value = child._values.remove(0);
						for (int j = 0; j < child._values.size();)
							value = (String)value + child._values.remove(0);

						value = decodeString((String)value);
						String[] strings = LyString.stringToList((String)value, LyTools.ARRAY_SEP);

						objectId = new LyObjectId(classInfo2, strings);
						objectIdVector.add(objectId);

						// seeking for more elements
						if (childs.size() > 0)
						{
							nextChild = childs.get(0);
							nextXmlField = (LyXmlField)nextChild._current;

							while (nextXmlField.getOriginalId().equals(xmlField.getOriginalId()))
							{
								if (nextChild._values.size() > 1)
								{
									classId = nextChild._values.remove(0);
									classInfo2 = _application.getClassInfo(classId);
									if (nextChild._values.size() > 0)
										value = nextChild._values.remove(0);
									for (int j = 0; j < nextChild._values.size();)
										value = (String)value + nextChild._values.remove(0);

									value = decodeString((String)value);
									strings = LyString.stringToList((String)value,
											LyTools.ARRAY_SEP);

									objectId = new LyObjectId(classInfo2, strings);
									objectIdVector.add(objectId);
								}
								childs.remove(0);
								if (childs.size() > 0)
								{
									nextChild = childs.get(0);
									nextXmlField = (LyXmlField)nextChild._current;
								}
								else
									break;
							}
						}
						LyObjectId[] objectIds = new LyObjectId[objectIdVector.size()];

						for (int j = 0; j < objectIds.length; j++)
							objectIds[j] = objectIdVector.get(j);
						value = objectIds;
					}
					else
					{
						xmlField.setValueClass(LyObjectId.class);
						String	    classId = child._values.remove(0);
						LyClassInfo classInfo2 = _application.getClassInfo(classId);

						if (child._values.size() > 0)
							value = child._values.remove(0);
						for (int j = 0; j < child._values.size();)
							value = (String)value + child._values.remove(0);
						value = decodeString((String)value);
						String[] strings = LyString.stringToList((String)value, LyTools.ARRAY_SEP);

						value = new LyObjectId(classInfo2, strings);
					}
				}
				// If it's the last element
				else if (childs.size() == 0)
				{
					if (child._values.size() > 0)
						value = child._values.remove(0);
					for (int j = 0; j < child._values.size();)
						value = (String)value + child._values.remove(0);
				}
				// Either a choice field or an other field
				else
				{
					nextChild = childs.get(0);
					nextXmlField = (LyXmlField)nextChild._current;
					xmlFieldId = xmlField.getOriginalId();
					nextXmlFieldId = nextXmlField.getOriginalId();
					LyFieldInfo fieldInfo = (LyFieldInfo)_application.getInfo(xmlFieldId);

					if (child._values.size() > 0)
						value = child._values.remove(0);
					for (int j = 0; j < child._values.size();)
						value = (String)value + child._values.remove(0);

					// If the next field have the same id, it's a choice field
					if (xmlFieldId.equals(nextXmlFieldId))
					{
						if ((fieldInfo instanceof LyChoiceFieldInfo)
								&& (fieldInfo.isMultiple()))
						{
							Object nextValue = null;

							if (nextChild._values.size() > 0)
								nextValue = nextChild._values.remove(0);
							for (int j = 0; j < nextChild._values.size();)
								nextValue = (String)nextValue
									+ nextChild._values.remove(0);

							if (nextValue != null)
								value = (String)value + _valueSep + (String)nextValue;

							// We skip this element
							childs.remove(0);

							if (childs.size() > 0)
							{
								nextChild = childs.get(0);
								nextXmlField = (LyXmlField)nextChild._current;

								while (nextXmlField.getOriginalId().equals(
											xmlField.getOriginalId()))
								{
									nextValue = null;

									if (nextChild._values.size() > 0)
										nextValue = nextChild._values.remove(0);
									for (int j = 0; j < nextChild._values.size();)
										nextValue = (String)nextValue
											+ nextChild._values.remove(0);

									if (nextValue != null)
										value = (String)value + _valueSep + (String)nextValue;

									childs.remove(0);

									if (childs.size() > 0)
									{
										nextChild = childs.get(0);
										nextXmlField = (LyXmlField)nextChild._current;
									}
									else
										break;
								}
							}
						}
						else
						{
							String message = _environment.getMessage("LY_ER_IMPORT_CARD",
									xmlFieldId);

							_environment.logWarning(message);
						}
					}
				}

				if (value != null)
				{
					if (value instanceof java.lang.String)
					{
						value = decodeString((String)value);
						value = fromString(xmlField, (String)value);
					}
				}
				// For nested objects within facets
				if ((_currentLevel != null)
						&& (_currentLevel._classInfo != null)
						&& (_currentLevel._current != null))
					((LyXmlObject)_currentLevel._current).setValue(xmlField, value);
				else
					((LyXmlObject)_rootElement).setValue(xmlField, value);
			}

			LyXmlObject objectToValidate;
			// For nested objects within facets
			if ((_currentLevel != null)
					&& (_currentLevel._classInfo != null)
					&& (_currentLevel._current != null))
				objectToValidate = (LyXmlObject)_currentLevel._current;
			else
			{
				_object = (LyXmlObject)_rootElement;
				_currentLevel = null;
				_rootElement = null;
				objectToValidate = _object;
			}
			objectToValidate.validate(null);

			// Check for the id fields
			LyInfoList<LyFieldInfo> idFields = classInfo.getIdFieldInfos();
			LyFieldBinding		    fieldBinding;
			LyFieldInfo			    fieldInfo;
			String				    id;
			String				    message;

			for (int i = 0; i < idFields.getSize(); i++)
			{
				fieldInfo = idFields.getInfo(i);
				fieldBinding = fieldInfo.getBinding();
				id = (fieldBinding == null) ? fieldInfo.getId() : fieldBinding.getBinding();
				if (objectToValidate.getValues().getFieldValue(id).getValue() == null)
				{
					message = _environment.getMessage("LY_ER_IMPORT_ID", id);
					_environment.logWarning(message);
					return;
				}
			}
			// For nested objects within facets
			if ((_currentLevel != null)
					&& (_currentLevel._classInfo != null)
					&& (_currentLevel._current != null))
			{
				// job is done in parent (see LyXmlObjectTool)
			}
			else
				_objects.addElement(_object);
		} // end method endObjectElement

		/**
		 * Construct the struct field.
		 *
		 * @param  tree the element tree.
		 * @return the objectId.
		 */
		protected Object processStructElement(_LyTreeStruct tree)
		{
			Vector<_LyTreeStruct>   childs = tree._childTrees;
			LyXmlField			    xmlField;
			_LyTreeStruct		    child;
			LyClassInfo			    classInfo = null, testClassInfo;
			LyInfoList<LyClassInfo> list = _application.getClassInfos();

			// No values
			if (tree._values.size() == 0)
				return null;

			for (int i = 0; i < list.getSize(); i++)
			{
				testClassInfo = list.getInfo(i);
				if (testClassInfo.getId().indexOf(tree._values.get(0) + ":") >= 0)
				{
					classInfo = testClassInfo;
					break;
				}
			}

			// ClassInfo not found
			if (classInfo == null)
				return null;

			LyXmlClass			    xmlClass = (LyXmlClass)classInfo.getPeerClass();
			LyXmlObject			    structObject = new LyXmlObject(xmlClass);
			LyXmlField			    testField;
			Object				    value = null;
			LyXmlField			    nextXmlField;
			String				    xmlFieldId;
			String				    nextXmlFieldId;
			_LyTreeStruct		    nextChild;
			LyInfoList<LyFieldInfo> idFields = classInfo.getIdFieldInfos();

			// For each child element (field)
			for (int i = 0; i < childs.size();)
			{
				child = childs.remove(0);
				xmlField = (LyXmlField)child._current;
				value = null;
				for (int j = 0; j < xmlClass.getFieldCount(); j++)
				{
					testField = (LyXmlField)xmlClass.getField(j);
					if (xmlField.getId().equals(testField.getOriginalId()))
					{
						xmlField = testField;
						break;
					}
				}
				// If it's a struct, we must construct the object
				if (child._realType == _LyTreeStruct.CURRENT_NODE_TYPE_STRUCT)
					value = processStructElement(child);
				// If it's an array, we must construct the object
				else if (child._realType == _LyTreeStruct.CURRENT_NODE_TYPE_ARRAY)
					value = processArrayElement(child);
				// If it's a relation
				else if (child._realType == _LyTreeStruct.CURRENT_NODE_TYPE_RELATION)
				{
					String			    relationId = ((LyXmlField)child._current)._originalId;
					LyRelationFieldInfo relation = (LyRelationFieldInfo)_application.getInfo(
							relationId);

					if (relation.isMultiple())
					{
						xmlField.setValueClass(LyObjectId[].class);

						Vector<LyObjectId> objectIdVector = new Vector<LyObjectId>();
						LyObjectId		   objectId;

						// first element
						String			   classId = child._values.remove(0);

						testClassInfo = _application.getClassInfo(classId);

						if (child._values.size() > 0)
							value = child._values.remove(0);
						for (int j = 0; j < child._values.size();)
							value = (String)value + child._values.remove(0);

						value = decodeString((String)value);
						String[] strings = LyString.stringToList((String)value, LyTools.ARRAY_SEP);

						objectId = new LyObjectId(testClassInfo, strings);
						objectIdVector.add(objectId);

						// seeking for more elements
						if (childs.size() > 0)
						{
							nextChild = childs.get(0);
							nextXmlField = (LyXmlField)nextChild._current;

							while (nextXmlField.getOriginalId().equals(xmlField.getOriginalId()))
							{

								if (nextChild._values.size() > 1)
								{
									classId = nextChild._values.remove(0);
									testClassInfo = _application.getClassInfo(classId);
									if (nextChild._values.size() > 0)
										value = nextChild._values.remove(0);
									for (int j = 0; j < nextChild._values.size();)
										value = (String)value
											+ nextChild._values.remove(0);

									value = decodeString((String)value);
									strings = LyString.stringToList((String)value,
											LyTools.ARRAY_SEP);

									objectId = new LyObjectId(testClassInfo, strings);
									objectIdVector.add(objectId);
								}
								childs.remove(0);

								if (childs.size() > 0)
								{
									nextChild = childs.get(0);
									nextXmlField = (LyXmlField)nextChild._current;
								}
								else
									break;
							}
						}
						LyObjectId[] objectIds = new LyObjectId[objectIdVector.size()];

						for (int j = 0; j < objectIds.length; j++)
							objectIds[j] = objectIdVector.get(j);
						value = objectIds;
					}
					else
					{
						xmlField.setValueClass(LyObjectId.class);
						String classId = child._values.remove(0);

						testClassInfo = _application.getClassInfo(classId);
						if (child._values.size() > 0)
							value = child._values.remove(0);
						for (int j = 0; j < child._values.size();)
							value = (String)value + child._values.remove(0);
						value = decodeString((String)value);
						String[] strings = LyString.stringToList((String)value, LyTools.ARRAY_SEP);

						value = new LyObjectId(testClassInfo, strings);
					}
				}
				// If it's the last element it can't be a multiple choice field
				else if (childs.size() == 0)
				{
					if (child._values.size() > 0)
						value = child._values.remove(0);
					for (int j = 0; j < child._values.size();)
						value = (String)value + child._values.remove(0);
				}
				// Either a choice field or an other field
				else
				{
					nextChild = childs.get(0);
					nextXmlField = (LyXmlField)nextChild._current;
					xmlFieldId = xmlField.getOriginalId();
					nextXmlFieldId = nextXmlField.getOriginalId();
					LyFieldInfo fieldInfo = (LyFieldInfo)_application.getInfo(xmlFieldId);

					if (child._values.size() > 0)
						value = child._values.remove(0);
					for (int j = 0; j < child._values.size();)
						value = (String)value + child._values.remove(0);

					// If the next field have the same id, it's a choice field
					if (xmlFieldId.equals(nextXmlFieldId))
					{
						if ((fieldInfo instanceof LyChoiceFieldInfo)
								&& (fieldInfo.isMultiple()))
						{
							Object nextValue = null;

							if (nextChild._values.size() > 0)
								nextValue = nextChild._values.remove(0);
							for (int j = 0; j < nextChild._values.size();)
								nextValue = (String)nextValue
									+ nextChild._values.remove(0);

							if (nextValue != null)
								value = (String)value + _valueSep + (String)nextValue;

							// We skip this element
							childs.remove(0);

							if (childs.size() > 0)
							{
								nextChild = childs.get(0);
								nextXmlField = (LyXmlField)nextChild._current;

								while (nextXmlField.getOriginalId().equals(
											xmlField.getOriginalId()))
								{
									nextValue = null;

									if (nextChild._values.size() > 0)
										nextValue = nextChild._values.remove(0);
									for (int j = 0; j < nextChild._values.size();)
										nextValue = (String)nextValue
											+ nextChild._values.remove(0);

									if (nextValue != null)
										value = (String)value + _valueSep + (String)nextValue;

									childs.remove(0);

									if (childs.size() > 0)
									{
										nextChild = childs.get(0);
										nextXmlField = (LyXmlField)nextChild._current;
									}
									else
										break;
								}
							}
						}
						else
						{
							String message = _environment.getMessage("LY_ER_IMPORT_CARD",
									xmlFieldId);

							_environment.logWarning(message);
						}
					}
				}

				if (value != null)
				{
					if (value instanceof java.lang.String)
					{
						value = decodeString((String)value);
						value = fromString(xmlField, (String)value);
					}
				}
				else
				{
					xmlFieldId = xmlField.getId();

					for (int j = 0; j < idFields.getSize(); j++)
					{
						if (xmlFieldId.equals(idFields.getInfo(j).getId()))
						{
							String message = _environment.getMessage("LY_ER_IMPORT_ID", xmlFieldId);

							_environment.logWarning(message);
							return null;
						}
					}
				}
				structObject.setValue(xmlField, value);
			}

			_currentLevel = null;

			structObject.validate(null);

			_application.getDataSource(xmlClass).receiveResponse(new LyResponse(
					LyResponse.Operation.LOAD, xmlClass), LyResponse.Status.RUNNING, structObject, null);

			return new LyObjectId(classInfo, xmlClass.getObjectId(structObject).toStrings());
		} // end method processStructElement

		/**
		 * Construct the array field.
		 *
		 * @param  tree the element tree.
		 * @return the object value array.
		 */
		protected Object processArrayElement(_LyTreeStruct tree)
		{
			String			 arrayId = ((LyXmlField)tree._current)._originalId;
			LyArrayFieldInfo arrayFieldInfo = (LyArrayFieldInfo)_application.getInfo(arrayId);
			LyFieldInfo		 fieldInfo = arrayFieldInfo.getFieldInfo();
			Vector<_LyTreeStruct>	childs = tree._childTrees;
			_LyTreeStruct    child;
			int				 nbChilds = childs.size(), j = 0;
			Object[]		 values = arrayFieldInfo.getArray(nbChilds);
			Object			 value = null;
			LyXmlField		 nextXmlField;
			String			 xmlFieldId;
			String			 nextXmlFieldId;
			_LyTreeStruct    nextChild;
			LyXmlField		 xmlField;

			// For each child element (field)
			for (int i = 0; i < childs.size(); j++)
			{
				child = childs.remove(0);
				xmlField = (LyXmlField)child._current;
				value = null;

				// If it's a struct, we must construct the object
				if (child._realType == _LyTreeStruct.CURRENT_NODE_TYPE_STRUCT)
					value = processStructElement(child);
				// If it's an array, we must construct the object
				else if (child._realType == _LyTreeStruct.CURRENT_NODE_TYPE_ARRAY)
					value = processArrayElement(child);
				// If it's a relation
				else if (child._realType == _LyTreeStruct.CURRENT_NODE_TYPE_RELATION)
				{
					String			    relationId = ((LyXmlField)child._current)._originalId;
					LyRelationFieldInfo relation = (LyRelationFieldInfo)_application.getInfo(
							relationId);
					LyClassInfo		    classInfo;

					if (relation.isMultiple())
					{
						xmlField.setValueClass(LyObjectId[].class);

						Vector<LyObjectId> objectIdVector = new Vector<LyObjectId>();
						LyObjectId		   objectId;

						// first element
						String			   classId = child._values.remove(0);

						classInfo = _application.getClassInfo(classId);

						if (child._values.size() > 0)
							value = child._values.remove(0);
						for (int k = 0; k < child._values.size();)
							value = (String)value + child._values.remove(0);

						value = decodeString((String)value);
						String[] strings = LyString.stringToList((String)value, LyTools.ARRAY_SEP);

						objectId = new LyObjectId(classInfo, strings);
						objectIdVector.add(objectId);

						// seeking for more elements
						if (childs.size() > 0)
						{
							nextChild = childs.get(0);
							nextXmlField = (LyXmlField)nextChild._current;

							while (nextXmlField.getOriginalId().equals(xmlField.getOriginalId()))
							{
								if (nextChild._values.size() > 1)
								{
									classId = nextChild._values.remove(0);
									classInfo = _application.getClassInfo(classId);
									if (nextChild._values.size() > 0)
										value = nextChild._values.remove(0);
									for (int k = 0; k < nextChild._values.size();)
										value = (String)value
											+ nextChild._values.remove(0);

									value = decodeString((String)value);
									strings = LyString.stringToList((String)value,
											LyTools.ARRAY_SEP);

									objectId = new LyObjectId(classInfo, strings);
									objectIdVector.add(objectId);
								}
								childs.remove(0);

								if (childs.size() > 0)
								{
									nextChild = childs.get(0);
									nextXmlField = (LyXmlField)nextChild._current;
								}
								else
									break;
							}
						}
						LyObjectId[] objectIds = new LyObjectId[objectIdVector.size()];

						for (int k = 0; k < objectIds.length; j++)
							objectIds[k] = objectIdVector.get(k);
						value = objectIds;
					}
					else
					{
						xmlField.setValueClass(LyObjectId.class);
						String classId = child._values.remove(0);

						classInfo = _application.getClassInfo(classId);
						if (child._values.size() > 0)
							value = child._values.remove(0);
						for (int k = 0; k < child._values.size();)
							value = (String)value + child._values.remove(0);
						value = decodeString((String)value);
						String[] strings = LyString.stringToList((String)value, LyTools.ARRAY_SEP);

						value = new LyObjectId(classInfo, strings);
					}
				}
				// If it's the last element it can't be a choice field
				else if (childs.size() == 0)
				{
					if (child._values.size() > 0)
						value = child._values.remove(0);
					for (int k = 0; k < child._values.size();)
						value = (String)value + child._values.remove(0);
				}
				// Either a choice field or an other field
				else
				{
					nextChild = childs.get(0);
					nextXmlField = (LyXmlField)nextChild._current;
					xmlFieldId = xmlField.getOriginalId();
					nextXmlFieldId = nextXmlField.getOriginalId();
					LyFieldInfo xmlFieldInfo = (LyFieldInfo)_application.getInfo(xmlFieldId);

					if (child._values.size() > 0)
						value = child._values.remove(0);
					for (int k = 0; k < child._values.size();)
						value = (String)value + child._values.remove(0);

					// If the next field have the same id, it's a choice field
					if (xmlFieldId.equals(nextXmlFieldId))
					{
						if ((xmlFieldInfo instanceof LyChoiceFieldInfo)
								&& (xmlFieldInfo.isMultiple()))
						{
							Object nextValue = null;

							if (nextChild._values.size() > 0)
								nextValue = nextChild._values.remove(0);
							for (int k = 0; k < nextChild._values.size();)
								nextValue = (String)nextValue
									+ nextChild._values.remove(0);

							if (nextValue != null)
								value = (String)value + _valueSep + (String)nextValue;

							// We skip this element
							childs.remove(0);

							if (childs.size() > 0)
							{
								nextChild = childs.get(0);
								nextXmlField = (LyXmlField)nextChild._current;

								while (nextXmlField.getOriginalId().equals(
											xmlField.getOriginalId()))
								{
									nextValue = null;

									if (nextChild._values.size() > 0)
										nextValue = nextChild._values.remove(0);
									for (int k = 0; k < nextChild._values.size();)
										nextValue = (String)nextValue
											+ nextChild._values.remove(0);

									if (nextValue != null)
										value = (String)value + _valueSep + (String)nextValue;

									childs.remove(0);

									if (childs.size() > 0)
									{
										nextChild = childs.get(0);
										nextXmlField = (LyXmlField)nextChild._current;
									}
									else
										break;
								}
							}
						}
						else
						{
							String message = _environment.getMessage("LY_ER_IMPORT_CARD",
									xmlFieldId);

							_environment.logWarning(message);
						}
					}
				}

				if (value != null)
				{
					if (value instanceof java.lang.String)
						value = decodeString((String)value);
				}
				values[j] = fieldInfo.buildFieldValue(value).getValue();
			}

			return values;
		} // end method processArrayElement

		/**
		 * Method to decode a string.
		 *
		 * @param  string The string to decode.
		 * @return The string decoded.
		 */
		protected String decodeString(String string)
		{
			string = LyString.replace(string, "&lt;", "<");
			string = LyString.replace(string, "&gt;", ">");
			string = LyString.replace(string, "&apos;", "'");
			string = LyString.replace(string, "&amp;", "&");
			string = LyString.replace(string, "&quot;", "\"");

			string = string.replace(LyTools.charFromString(_valueSep), LyTools.ARRAY_SEP);

			return asciiDecode(string);
		}
	} // end class _LyXmlDefaultHandler


	/**
	 * Simple struct class to keep hierarchical parent tree.
	 */
	protected class _LyTreeStruct
	{
		public static final short    CURRENT_NODE_TYPE_CLASS = 0;
		public static final short    CURRENT_NODE_TYPE_OBJECT = 1;
		public static final short    CURRENT_NODE_TYPE_RELATION = 2;
		public static final short    CURRENT_NODE_TYPE_ELEMENT = 3;
		public static final short    CURRENT_NODE_TYPE_STRUCT = 4;
		public static final short    CURRENT_NODE_TYPE_STRUCT_ELEMENT = 5;
		public static final short    CURRENT_NODE_TYPE_OBJECT_ELEMENT = 6;
		public static final short    CURRENT_NODE_TYPE_ARRAY = 7;
		public static final short    CURRENT_NODE_TYPE_ARRAY_ELEMENT = 8;

		public _LyTreeStruct		 _parentTree;
		public Vector<_LyTreeStruct> _childTrees;
		public Vector<String>		 _values;
		public Object				 _current;
		public short				 _relativeType;
		public short				 _realType;

		// Use to swap temporary the class and fields for facets
		public LyClassInfo			 _classInfo;
		public Vector<LyField>		 _fields;

		public _LyTreeStruct(_LyTreeStruct parentTree,
			Object current, short relativeType, short realType, Vector<_LyTreeStruct> childTrees)
		{
			_parentTree = parentTree;
			_current = current;
			_relativeType = relativeType;
			_realType = realType;
			_childTrees = childTrees;
			_fields = ((parentTree == null) ? null : _parentTree._fields);
			_classInfo = ((parentTree == null) ? null : _parentTree._classInfo);
			_values = new Vector<String>();
		}
	} // end class _LyTreeStruct

	/**
	 * Apply XSLT Stylesheets to convert the source version of the XML File to latest version.
	 *
	 * @deprecated OLD_IMPORT : reported in LyImportContentXmlController
	 * @param      input source for parsing (input version)
	 * @return     source the nw source after parsing (output version)
	 */
	public InputSource convertSourceVersion(InputSource source)
	{
		return source;
	} // end method convertSourceVersion

	/**
	 * Skip root tag generation (only write XML content in the root tag).
	 *
	 * @param status
	 */
	public void skipHeader(boolean status)
	{
		_skipHeader = status;
	}

	/**
	 * Skip root tag generation (only write XML content in the root tag).
	 *
	 * @param status
	 * @param status
	 */
	public void skipFooter(boolean status)
	{
		_skipFooter = status;
	}
} // end class LyXmlExportRequestEncoder
