/*
 * LyXmlRequestEncoder.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.peer.xml;

import org.xml.sax.*;
import org.xml.sax.helpers.*;
import java.io.*;
import java.util.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.peer.*;
import leon.peer.file.*;
import leon.peer.util.*;


/**
 * This class is used for local files as data provider. Methods for creation, deletion and
 * modification are not overriden since they don't have to write anything in the corresponding
 * files. The specific method saveObjects must be called in order to make the objects persistent in
 * a file.
 * @version $Id: LyXmlRequestEncoder.java 35185 2012-04-10 14:40:49Z jlsaouli $
 * @author  Lyria
 * @see     LyRequestEncoder
 */
public class LyXmlRequestEncoder extends LyFileRequestEncoder // TODO : remove extension for filerequestencoder here for computeId() only
{
	public static final String __VERSION =
		"$Id: LyXmlRequestEncoder.java 35185 2012-04-10 14:40:49Z jlsaouli $";

	/**
	 * Saving type : only default yet ,but here to keep file save cmpatibility and for further
	 * extension.
	 */
	public static final int    TYPE_DEFAULT = 1;

	/** To stock objects during loading. */
	protected Vector<LyXmlObject>		   _objects;

	/** To stock the current parsed object. */
	protected LyXmlObject	   _object;

	/** The requested fields. */
	protected Vector<?>		   _fields;

	/**
	 * Constructor.
	 * @param fieldSep  separator used between fields.
	 * @param valuedSep separator used between values.
	 */
	public LyXmlRequestEncoder(
		LyDataProvider provider,
		LyDataSource dataSource,
		String fieldSep,
		String valueSep)
	{
		//super(provider, dataSource);
		super(provider, dataSource, fieldSep, valueSep);
	}

	/**
	 * Initialisation of the data of the instance.
	 */
	public void init()
	{
		super.init();
		_objects = null;
	}

	/**
	 * Set the request encoder response. Inherited method from LyRequestEncoder to set properties to
	 * the response.
	 * @param the response
	 */
	public void setResponse(LyResponse response)
	{
		_response = response;

		// Set properties of the response.
		_response.setSorted(false);
		_response.setFiltered(false);
		_response.setCounted(true);
		_response.setDistinctValues(false);
	}

	/**
	 * Get an object.
	 * @param  objectId object identifier
	 * @param  fields   the fields to load (if null, load all)
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getObject(LyObjectId objectId, Vector<LyField> fields)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// Check the identifier
		if (objectId == null)
		{
			_response.setError(_provider.getEnvironment().getMessage("LY_ER_CANT_BE_NULL",
					"objectId"));
			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			return false;
		}

		// No need to encode the value
		// Get the data from the connection as a list of lines.
		Vector<LyXmlObject> objects = readFromConnection(fields);

		if (objects == null)
		{
			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			return false;
		}

		// Parse the objects to find out the good one.
		int		   i, n = objects.size();
		Object     object = null;
		LyObjectId id;

		for (i = 0; i < n; i++)
		{
			object = objects.elementAt(i);
			id = _class.getObjectId(object);

			if (id == null)
				continue;

			if (id.equals(objectId))
			{
				// If the object was found, add it to the response. Else, add nothing.
				_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object, null);
				return true;
			}
		}

		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);

		return true;
	} // end method getObject

	/**
	 * Get a set of objects.
	 * @param  fields the fields to load (if null, load all)
	 * @param  sort   indicates sorting criterion, or null if default sort is used.
	 * @param  filter indicates filtering criterion, or null if there is no filter.
	 * @param  n      maximum number of objects to get.
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getObjects(
		Vector<LyField> fields,
		LySort sort,
		LyFilter filter,
		int n)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// No need to encode the value
		// Get the data from the connection as a list of lines.
		Vector<LyXmlObject> objects = readFromConnection(fields);

		if (objects == null)
		{
			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			return false;
		}

		// Sort the objects
		// TODO

		// Keep the first n objects matching the filter (if given)
		int    i, nf = 0, no = ((objects == null) ? 0 : objects.size());
		Object object;

		for (i = 0; i < no; i++)
		{
			object = objects.elementAt(i);

			nf++;
			_dataSource.receiveResponse(_response, LyResponse.Status.RUNNING, object, null);

			// If the maximum size is reached
			if ((n > 0) && (nf >= n))
				break;
		}

		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);

		return true;
	} // end method getObjects

	/**
	 * Return the number of objects from the file.
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getSize()
	{
		return (getSize(null));
	}

	/**
	 * Return the number of objects from the file .
	 * @param  filter indicates filtering criterion, or null if there is no filter.
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean getSize(LyFilter filter)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// No need to encode the value
		// Get the data from the connection as a list of lines.
		Vector<LyXmlObject> objects = readFromConnection(null);

		if (objects == null)
		{
			_dataSource.receiveResponse(_response, LyResponse.Status.ERROR, null, null);
			return false;
		}

		int nf = objects.size();

		// TODO : Keep the first n objects matching the filter (if given)

		// Add the Integer to the response
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, new Integer(nf), null);
		return true;
	}

	/**
	 * To create a new object.
	 * @param  object the new object to create
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean createObject(Object object)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// If the object has no Id, we try to get one.
		computeIds(object);

		// Validate the creation.
		((LyXmlObject)object).validate(null);

		// Send the response.
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object, null);

		return true;
	}

	/**
	 * To set an object.
	 * @param  object the object to set
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean setObject(Object object)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// Validate the modification.
		LyValueSet oldValues = ((LyXmlObject)object).validate(null);

		// Send the response.
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object, oldValues);

		return true;
	}

	/**
	 * To set a set of objects.
	 * @param  objects the objects to set
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean setObjects(Vector<Object> objects)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		Object     object;
		int		   i, n;
		LyValueSet oldValues;

		n = objects.size();

		for (i = 0; i < n; i++)
		{
			object = objects.elementAt(i);

			// Validate the modification.
			oldValues = ((LyXmlObject)object).validate(null);

			// Send the response.
			_dataSource.receiveResponse(_response, LyResponse.Status.RUNNING, object, oldValues);
		}

		// Send the last response.
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, null, null);

		return true;
	}

	/**
	 * To delete an object.
	 * @param  object the object to delete
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean deleteObject(Object object)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);
		_dataSource.receiveResponse(_response, LyResponse.Status.COMPLETED, object, null);
		return true;
	}

	/**
	 * Find reader to dataSource.
	 */
	protected Reader getReader()
	{
		// Get a valid connection from the data provider.
		LyXmlConnection connection = (LyXmlConnection)_provider.getConnection();

		// Get a Reader from the connection.
		return connection.getReader(_class);
	}

	/**
	 */
	protected Vector<LyXmlObject> readFromConnection(Vector<?> fields)
	{
		// Set the requested fields;
		_fields = ((fields == null) ? _class.getFields() : fields);

		// Get a valid connection from the data provider.

		LyXmlConnection connection = (LyXmlConnection)_provider.getConnection();

		if (connection == null)
			return null; // No available connection.

		// Get a Reader from the connection.
		Reader reader = getReader();

		if (reader == null)
			return null; // File not found

		// Create the InputSource.
		InputSource source = new InputSource(reader);

		try
		{
			parseXMLSource(source);
		}
		catch (Exception e)
		{
			LyEnvironment environment = getEnvironment();

			environment.logError("Error during parsing of " + _class.getId());
			if (e instanceof SAXParseException)
				environment.logError("Line : " + ((SAXParseException)e).getLineNumber()
					+ ", column : " + ((SAXParseException)e).getColumnNumber());

			if (environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
				environment.writeException(e);

			// Reports the error.
			_response.setError(e.getMessage());
			_provider.getApplication().setError(e.getMessage());

			// Release the connection.
			_provider.releaseConnection(connection);

			return null;
		}

		// Release the connection.
		_provider.releaseConnection(connection);

		// Return the result.
		return _objects;
	} // end method readFromConnection

	/**
	 *Utility method for XML Source parsing
	 */
	protected void parseXMLSource(InputSource source)
		throws SAXException,
			IOException
	{
		LyEnvironment environment = _provider.getApplication().getEnvironment();
		String		  requestedParser = environment.getEnv("_LY_XML_PARSER");
		String		  defaultParser = "org.apache.xerces.parsers.SAXParser";

		if (requestedParser == null)
		{
			requestedParser = System.getProperty("org.xml.sax.driver");
			if (requestedParser == null)
				requestedParser = defaultParser;
		}

		// Create a new Parser.
		XMLReader	   parser = XMLReaderFactory.createXMLReader(requestedParser);

		// Create a default handler
		DefaultHandler handler = buildHandler();

		// Set the ContentHandler to the parser.
		parser.setContentHandler(handler);

		// Set the ErrorHandler to the parser.
		parser.setErrorHandler(handler);

		// Before we parse, we may apply stylesheets to convert the meta-model
		source = convertSourceVersion(source);

		// Parse.
		parser.parse(source);
	} // end method parseXMLSource

	/**
	 * Apply any transformation (like XSLT Stylesheets to convert the source version of the XML File
	 * to latest version current implementation does nothing : should be overriden in subclasses.
	 * @param  input source for parsing (input version)
	 * @return source the new source after parsing (output version)
	 */
	public InputSource convertSourceVersion(InputSource source)
	{
		return source;
	}

	/**
	 * Return the default handler for this provider.
	 * @return The default handler.
	 */
	public DefaultHandler buildHandler()
	{
		return new _LyXmlDefaultHandler();
	}

	/**
	 *  DefaultHandler *
	 */
	class _LyXmlDefaultHandler extends DefaultHandler
	{
		public void startDocument()
			throws SAXException
		{
			// Initialize the vector to stock objects.
			_objects = new Vector<LyXmlObject>();
		}

		public void startElement(String uri, String localName, String qName,
			Attributes attributes)
			throws SAXException
		{
			if (qName != null)
			{
				LyXmlClass xmlClass = (LyXmlClass)_class;

				if (qName.equals(xmlClass.getName()))
				{
					// The tag represents the class : we create a new LyXmlObject.
					_object = new LyXmlObject(xmlClass);

					// The attributes are the fields of the object.
					int		   i, n;
					LyXmlField xmlField;
					String     value;

					n = _fields.size();
					for (i = 0; i < n; i++)
					{
						xmlField = (LyXmlField)_fields.elementAt(i);
						value = attributes.getValue(xmlField.getId());

						// Specific XML encoding
						if (value != null)
						{
							value = LyString.replace(value, "&lt;", "<");
							value = LyString.replace(value, "&gt;", ">");
							value = LyString.replace(value, "&apos;", "'");
							value = LyString.replace(value, "&quot;", "\"");
							value = LyString.replace(value, "&amp;", "&");

							// Replace the default separator used by the toString method with _valueSep.
							value = value.replace(LyTools.charFromString(_valueSep),
									LyTools.ARRAY_SEP);
							value = asciiDecode(value);
						}
						Object valueObj = fromString(xmlField, value);

						_object.setValue(xmlField, valueObj);
					}

					_object.validate(null);
					_objects.addElement(_object);
				}
			}
		} // end method startElement

		// treat validation errors as fatal
		public void error(SAXParseException e)
			throws SAXParseException
		{
			throw e;
		}

		// dump warnings too
		public void warning(SAXParseException err)
			throws SAXParseException
		{
			getEnvironment().logWarning("line "
				+ err.getLineNumber()
				+ ", uri "
				+ err.getSystemId()
				+ "\n\t"
				+ err.getMessage());
		}
	} // end class _LyXmlDefaultHandler

	/**
	 * Method used to saved all the objects contained in the corresponding data source cache.
	 * @param  fields fields to save (if null save all).
	 * @param  type   indicates the file format
	 * @return true if the request was correctly sent to the external data provider.
	 */
	public boolean saveObjects(Vector<Object> objects, Vector<LyField> fields, int type)
	{
		// Set the response in running mode
		_response.setStatus(LyResponse.Status.RUNNING);

		// Save the data in the file
		writeToConnection(objects, fields, type);

		_response.setStatus(LyResponse.Status.COMPLETED);

		return true;
	}

	/**
	 * Read a set of lines and return a set of objects.
	 * @param  objects the set of objects to encode.
	 * @param  fields  FieldInfos to save (if null save all).
	 * @param  type    indicates the file format.
	 * @return the set of lines describing the set of objects.
	 */
	public Vector<String> encodeValues(Vector<Object> objects, Vector<LyField> fields, int type)
	{
		// Get the fields : if fields is empty, get all the FieldInfos from
		// the ClassInfo.
		Vector<?> collect;

		if ((fields == null) || (fields.size() == 0))
			collect = _class.getFields();
		else
			collect = fields;

		Vector<String>     result = new Vector<String>();
		String     line;

		// For each object get the ValueSet and encode it.
		// FFT01068 test if objects is not null
		int		   i, no = ((objects != null) ? objects.size() : 0);
		Object     object;
		LyValueSet values;

		for (i = 0; i < no; i++)
		{
			object = objects.elementAt(i);

			if (object == null)
				continue;

			values = _class.getValues(object);
			line = encodeValue(values, collect, type);

			result.addElement(line);
		}

		return result;
	} // end method encodeValues

	/**
	 * Return a line describing the ValueSet.
	 * @param  values the ValueSet to encode.
	 * @param  fields the FieldInfos to encode.
	 * @param  type   indicates the file format.
	 * @return the line describing the ValueSet.
	 */
	public String encodeValue(LyValueSet values, Vector<LyField> fields, int type)
	{
		LyField		 field;
		LyValue		 value;
		Object		 val;
		int			 i;
		int			 na = fields.size();
		StringBuffer result = new StringBuffer();
		String		 str = null;

		LyPeerClass  xmlClass = (LyPeerClass)_class;

		result.append("\t<" + xmlClass.getName());

		StringBuffer betweenAttributes = new StringBuffer();

		betweenAttributes.append("\n\t");
		for (i = 0; i < (xmlClass.getName().length() + 2); i++)
			betweenAttributes.append(" ");

		// For each FieldInfo retrieve the value and write it
		for (i = 0; i < na; i++)
		{
			field = fields.elementAt(i);

			value = values.getFieldValue(field);

			if (value == null)
				continue;

			val = value.getValue();
			if (val == null)
				continue;

			if (i == 0)
				result.append(" ");
			else
				result.append(betweenAttributes);

			result.append(field.getId()).append("='");

			// For export : value of choice must be saved instead of index
			if (field instanceof LyChoiceFieldInfo)
			{
				LyChoiceFieldInfo choice = (LyChoiceFieldInfo)field;

				if (!choice.isMultiple())
				{
					LyChoiceItem item = choice.getChoiceItem(((Short)val).shortValue());

					val = item.getValue();
				}
				else
				{
					BitSet		 bs = (BitSet)val;
					Vector<Object>		 vector = null;
					int			 j, size = bs.size();
					LyChoiceItem item;

					for (j = 0; j < size; j++)
					{
						if (bs.get(j))
						{
							if (vector == null)
								vector = new Vector<Object>();

							item = choice.getChoiceItem((short)j);
							vector.addElement(item.getValue());
						}
					}

					val = vector;
				}
			}

			str = toString(field, val, type);
			result.append(str);
			result.append("'");
		}
		result.append("/>\n");

		return (result.toString());
	} // end method encodeValue

	/**
	 * Method to get an objectId value as string.
	 * @param  objectId The objectId value.
	 * @return The value as a string (specific XML encoding must not be done).
	 */
	protected String getObjectIdValue(LyObjectId objectId)
	{
		return LyRequestEncoder.toString(objectId);
	}

	/**
	 * Method to get an array of objectId values as string.
	 * @param  objectIds The array of objectId values.
	 * @return The value as a string (specific XML encoding must not be done).
	 */
	protected String getObjectIdValue(LyObjectId[] objectIds)
	{
		StringBuffer buffer = new StringBuffer();
		int			 i, n = objectIds.length;

		buffer.append(LyTools.ARRAY_SEP);

		for (i = 0; i < n; i++)
		{
			if (i > 0)
				buffer.append(LyTools.ARRAY_SEP);
			if (objectIds[i] != null)
				buffer.append(getObjectIdValue(objectIds[i]));
		}

		buffer.append(LyTools.ARRAY_SEP);
		return buffer.toString();
	}

	/**
	 * encode value to XML attribute String.
	 */
	protected String toString(LyField field, Object value, int type)
	{
		String result;

		if (value == null)
			return null;

		if ((field instanceof LyRelationFieldInfo)
				&& (value.getClass() == field.getValueClass()))
		{
			if (value instanceof LyObjectId)
				result = getObjectIdValue((LyObjectId)value);
			else
				result = getObjectIdValue((LyObjectId[])value);
		}
		else if ((field instanceof LyColumnField)
				&& (((LyColumnField)field).getFormat() != null)
				&& (value instanceof Date))
			result = LyDateFormatter.format((Date)value, ((LyColumnField)field).getFormat());
		else
			result = LyRequestEncoder.toString(value);
		// Specific XML encoding
		if (result != null)
		{
			result = LyString.replace(result, "&", "&amp;");
			result = LyString.replace(result, "<", "&lt;");
			result = LyString.replace(result, ">", "&gt;");
			result = LyString.replace(result, "'", "&apos;");
			// Not needed here : result = LyString.replace(result, "\"", "&quot;");
			result = asciiEncode(result);

			// Retrieve the separator for object type
			String toolsSep = new Character(LyTools.getSeparator(value)).toString();

			// Replace the default separator used by the toString method
			// with _valueSep.
			result = LyString.replace(result, toolsSep, _valueSep);
		}
		return result;
	} // end method toString

	/**
	 * Sends a request to the file connection for writing data.
	 * @param  lines lines to write on the file.
	 * @return lines read from the file.
	 */
	public boolean writeToConnection(Vector<String> lines)
	{
		// Get a valid connection
		LyXmlConnection connection;

		connection = (LyXmlConnection)_provider.getConnection();
		if (connection == null)
			return false; // No available connection.

		// Write the data in the file.
		boolean result = connection.writeXml(_class, lines);

		_provider.releaseConnection(connection);

		return result;
	}

	/**
	 * Sends a request to the file connection for writing data.
	 * @param  objects the set of objects to encode.
	 * @param  fields  FieldInfos to save (if null save all).
	 * @param  type    indicates the file format.
	 * @return lines read from the file.
	 */
	public boolean writeToConnection(Vector<Object> objects, Vector<LyField> fields, int type)
	{
		// Get a valid connection
		LyXmlConnection connection;

		connection = (LyXmlConnection)_provider.getConnection();
		if (connection == null)
			return false; // No available connection.

		// Write the data in the file.
		Vector<?> collect;

		if ((fields == null) || (fields.size() == 0))
			collect = _class.getFields();
		else
			collect = fields;

		PrintWriter pw = connection.openFile(_class);
		boolean     result = false;
		String	    line;

		if (pw != null)
		{
			// Add header
			connection.writeLine(encodeHeader(collect, type), pw);

			// Add the lines
			int    i, no = ((objects != null) ? objects.size() : 0);
			Object object;

			for (i = 0; i < no; i++)
			{
				object = objects.elementAt(i);
				if (object == null)
					continue;

				line = encodeValue(((LyXmlObject)object).getValues(), collect, type);
				connection.writeLine(line, pw);
			}

			connection.closeFile(pw);
			result = true;
		}

		_provider.releaseConnection(connection);

		return result;
	} // end method writeToConnection

} // end class LyXmlRequestEncoder
