/*
 * LyArea.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.view;

import java.util.*;

import leon.misc.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;


/**
 * The LyArea class implements generic behavior for the AREA components as described in views.dtd
 * regardless of the requested target (win, swing, html...). It represents an area.
 *
 * @author   Lyria
 * @version  $Id: LyArea.java 29006 2010-12-23 17:44:46Z bmichel $
 * @see      LyAreaBase
 */
public class LyArea extends LyAreaBase
	implements LyGetRefreshInterface
{
	public static final String __VERSION = "$Id: LyArea.java 29006 2010-12-23 17:44:46Z bmichel $";

	public static final short NORTH = 0;

	public static final short SOUTH = 1;

	public static final short EAST = 2;

	public static final short WEST = 3;

	/** Specific environnement of this area. */
	LyEnvironment _environment;

	/** Border panels. */
	protected LyPanel[] _borderPanels;

	/** Size of the grid representing the layout for this area. */
	protected int[] _layoutSize;

	/** Position of each sub-panels of this area in the grid layout. */
	protected Hashtable<LyPanel, int[]> _panelPositions;
	
	/** Indicates whether all columns in the layout will be forced to have the same width. */
	protected boolean _isColumnsEqualWidth;
	
	/** Indicates that the field is hidden (not displayed). */
    protected boolean _hidden;

	/**
	 * Constructor with all required attributes.
	 */
	public LyArea()
	{
		_environment = null;
		_borderPanels = null;
		_hidden = false;
	}

	/**
	 * Construction from a LyViewNode.
	 */
	protected LyArea(LyViewNode node)
	{
		super(node);

		_environment = null;
		_borderPanels = null;
		_hidden = false;
	}

	/**
	 * End of construction : completion with child component.
	 *
	 * @param  child  Sub child component.
	 *
	 * @see    leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		if (child instanceof LySelectBar)
			setSelectBar((LySelectBar)child);
		else
			super.addChild(child);
	}

	/**
	 * Reset this component.<br />
	 * Overridden to reset border panels if needed.
	 */
	public void reset()
	{
		super.reset();

		// Free border panels
		if (_borderPanels != null)
		{
			int i;

			for (i = NORTH; i < WEST; i++)
			{
				if (_borderPanels[i] != null)
					_borderPanels[i].reset();
			}
		}
	}

	/**
	 * Free this component.<br />
	 * Overridden to free border panels if needed.
	 */
	public void free()
	{
		super.free();

		// Free border panels
		if (_borderPanels != null)
		{
			int i;

			for (i = NORTH; i < WEST; i++)
			{
				if (_borderPanels[i] != null)
					_borderPanels[i].free();
			}
		}

		// Free table containing panels positions
		if (_panelPositions != null)
		{
			_panelPositions.clear();
			_panelPositions = null;
		}

		_panels = null;
		_toolBars = null;
		_borderPanels = null;
		_titleBar = null;
		_tip = null;
	}

	/**
	 * Gets specific environment of this area.
	 *
	 * @return  specific environment of this area.
	 */
	public LyEnvironment getEnvironment()
	{
		if (_environment != null)
			return _environment;

		return super.getEnvironment();
	}

	/**
	 * Retrieve the scrollable attribute.
	 */
	public boolean isScrollable()
	{
		return ((getScrolling() == Scrolling.YES) || (getScrolling() == Scrolling.AUTO));
	}

	/**
	 * Sets specific environment of this area.
	 *
	 * @param  environment  Specific environment of this area.
	 */
	public void setEnvironment(LyEnvironment environment)
	{
		_environment = environment;
	}

	/**
	 * Return the top left toolbar.
	 *
	 * @return  the toolbar
	 */
	public LyToolBar getToolBar()
	{
		return getTopLeftToolBar();
	}

	/**
	 * Add a toolbar.
	 *
	 * @param  toolBar  a toolbar
	 */
	public void setToolBar(LyToolBar toolBar)
	{
		if ((toolBar == null) || (toolBar.getFill() != LyToolBar.Fill.VERTICAL))
			setTopLeftToolBar(toolBar);
		else if (toolBar.getFill() == LyToolBar.Fill.VERTICAL)
			setLeftToolBar(toolBar);
	}

	/**
	 * Return the top right toolbar.
	 *
	 * @return  the toolbar
	 */
	public LyToolBar getTopRightToolBar()
	{
		return getToolBar(LyToolBar.HAlign.RIGHT, LyToolBar.VAlign.TOP);
	}

	/**
	 * Add a toolbar at the top right position.
	 *
	 * @param  toolBar  a toolbar
	 */
	public void setTopRightToolBar(LyToolBar toolBar)
	{
		if ((toolBar != null)
				&& (toolBar.getHAlign() != LyToolBar.HAlign.RIGHT)
				&& (toolBar.getVAlign() != LyToolBar.VAlign.TOP))
			throw new IllegalStateException("BF AppEngine : attempting to set a toolbar at a wrong position");

		// get the current toolbar
		LyToolBar current = getTopRightToolBar();

		if (current != null)
			_toolBars.remove(current);

		if (toolBar != null)
			super.addToolBar(toolBar);
	}

	/**
	 * Return the top left toolbar.
	 *
	 * @return  the toolbar
	 */
	public LyToolBar getTopLeftToolBar()
	{
		return getToolBar(LyToolBar.HAlign.LEFT, LyToolBar.VAlign.TOP);
	}

	/**
	 * Add a toolbar at the top left position.
	 *
	 * @param  toolBar  a toolbar
	 */
	public void setTopLeftToolBar(LyToolBar toolBar)
	{
		if ((toolBar != null)
				&& (toolBar.getHAlign() != LyToolBar.HAlign.LEFT)
				&& (toolBar.getVAlign() != LyToolBar.VAlign.TOP))
			throw new IllegalStateException("BF AppEngine : attempting to set a toolbar at a wrong position");

		// get the current toolbar
		LyToolBar current = getTopLeftToolBar();

		if (current != null)
			_toolBars.remove(current);

		if (toolBar != null)
			super.addToolBar(toolBar);
	}

	/**
	 * Return the left toolbar.
	 *
	 * @return  the toolbar
	 */
	public LyToolBar getLeftToolBar()
	{
		return getToolBar(LyToolBar.HAlign.LEFT, LyToolBar.VAlign.MIDDLE);
	}

	/**
	 * Add a toolbar at the left position.
	 *
	 * @param  toolBar  a toolbar
	 */
	public void setLeftToolBar(LyToolBar toolBar)
	{
		if ((toolBar != null)
				&& (toolBar.getHAlign() != LyToolBar.HAlign.LEFT)
				&& (toolBar.getVAlign() != LyToolBar.VAlign.MIDDLE))
			throw new IllegalStateException("BF AppEngine : attempting to set a toolbar at a wrong position");

		// get the current toolbar
		LyToolBar current = getLeftToolBar();

		if (current != null)
			_toolBars.remove(current);

		if (toolBar != null)
			super.addToolBar(toolBar);
	}

	/**
	 * Return the right toolbar.
	 *
	 * @return  the toolbar
	 */
	public LyToolBar getRightToolBar()
	{
		return getToolBar(LyToolBar.HAlign.RIGHT, LyToolBar.VAlign.MIDDLE);
	}

	/**
	 * Add a toolbar at the right position.
	 *
	 * @param  toolBar  a toolbar
	 */
	public void setRightToolBar(LyToolBar toolBar)
	{
		if ((toolBar != null)
				&& (toolBar.getHAlign() != LyToolBar.HAlign.RIGHT)
				&& (toolBar.getVAlign() != LyToolBar.VAlign.MIDDLE))
			throw new IllegalStateException("BF AppEngine : attempting to set a toolbar at a wrong position");

		// get the current toolbar
		LyToolBar current = getRightToolBar();

		if (current != null)
			_toolBars.remove(current);

		if (toolBar != null)
			super.addToolBar(toolBar);
	}

	/**
	 * Return the bottom right toolbar.
	 *
	 * @return  the toolbar
	 */
	public LyToolBar getBottomRightToolBar()
	{
		return getToolBar(LyToolBar.HAlign.RIGHT, LyToolBar.VAlign.BOTTOM);
	}

	/**
	 * Add a toolbar at the bottom right position.
	 *
	 * @param  toolBar  a toolbar
	 */
	public void setBottomRightToolBar(LyToolBar toolBar)
	{
		if ((toolBar != null)
				&& (toolBar.getHAlign() != LyToolBar.HAlign.RIGHT)
				&& (toolBar.getVAlign() != LyToolBar.VAlign.BOTTOM))
			throw new IllegalStateException("BF AppEngine : attempting to set a toolbar at a wrong position");

		// get the current toolbar
		LyToolBar current = getBottomRightToolBar();

		if (current != null)
			_toolBars.remove(current);

		if (toolBar != null)
			super.addToolBar(toolBar);
	}

	/**
	 * Return the bottom left toolbar.
	 *
	 * @return  the toolbar
	 */
	public LyToolBar getBottomLeftToolBar()
	{
		return getToolBar(LyToolBar.HAlign.LEFT, LyToolBar.VAlign.BOTTOM);
	}

	/**
	 * Add a toolbar at the bottom left position.
	 *
	 * @param  toolBar  a toolbar
	 */
	public void setBottomLeftToolBar(LyToolBar toolBar)
	{
		if ((toolBar != null)
				&& (toolBar.getHAlign() != LyToolBar.HAlign.LEFT)
				&& (toolBar.getVAlign() != LyToolBar.VAlign.BOTTOM))
			throw new IllegalStateException("BF AppEngine : attempting to set a toolbar at a wrong position");

		// get the current toolbar
		LyToolBar current = getBottomLeftToolBar();

		if (current != null)
			_toolBars.remove(current);

		if (toolBar != null)
			super.addToolBar(toolBar);
	}

	/**
	 * Internal method to get the toolbar with the given alignments.
	 *
	 * @param   hAlign  horizontal alignment
	 * @param   vAlign  vertical alignment
	 *
	 * @return  a toolbar or null if not found
	 */
	private LyToolBar getToolBar(LyToolBar.HAlign hAlign, LyToolBar.VAlign vAlign)
	{
		if (_toolBars == null)
			return null;

		LyToolBar toolBar;

		for (int i = 0; i < _toolBars.size(); i++)
		{
			toolBar = _toolBars.get(i);
			if ((toolBar.getHAlign() == hAlign)
					&& (toolBar.getVAlign() == vAlign))
				return toolBar;
		}

		return null;
	}

	/**
	 * @see  leon.view.LyAreaBase#addToolBar(leon.view.LyToolBar)
	 */
	public void addToolBar(LyToolBar toolBar)
	{
		// If the toolBar is vertical, add it to the left or to the right
		// if it's V_ALIGN is TOP or BOTTOM, show a warning
		if ((toolBar.getFill() == LyToolBar.Fill.VERTICAL)
				|| (toolBar.getVAlign() == LyToolBar.VAlign.MIDDLE))
		{
			if (toolBar.getVAlign() != LyToolBar.VAlign.MIDDLE)
			{
				LyEnvironment env = getEnvironment();

				if (env != null)
					env.logWarning("Adding a vertical toolbar with a TOP or BOTTOM vertical alignment - changing it");
				else
					System.out.println(
						"Adding a vertical toolbar with a TOP or BOTTOM vertical alignment - changing it");

				toolBar.setVAlign(LyToolBar.VAlign.MIDDLE);
			}

			if (toolBar.getFill() != LyToolBar.Fill.VERTICAL)
			{
				LyEnvironment env = getEnvironment();

				if (env != null)
					env.logWarning("Adding a horizontal toolbar in a vertical emplacement - Changing it");
				else
					System.out.println("Adding a horizontal toolbar in a vertical emplacement - Changing it");

				toolBar.setFill(LyToolBar.Fill.VERTICAL);
			}

			if (toolBar.getHAlign() == LyToolBar.HAlign.LEFT)
				setLeftToolBar(toolBar);
			else
				setRightToolBar(toolBar);
		}
		else if (toolBar.getVAlign() == LyToolBar.VAlign.TOP)
		{
			if (toolBar.getHAlign() == LyToolBar.HAlign.LEFT)
				setTopLeftToolBar(toolBar);
			else
				setTopRightToolBar(toolBar);
		}
		else
		{
			if (toolBar.getHAlign() == LyToolBar.HAlign.LEFT)
				setBottomLeftToolBar(toolBar);
			else
				setBottomRightToolBar(toolBar);
		}

	} // end method addToolBar

	/**
	 * Get panel at the given pos.
	 *
	 * @param   position  Either NORTH, SOUTH, EAST or WEST position.
	 *
	 * @return  the searched panel.
	 */
	public LyPanel getBorderPanel(short position)
	{
		if (_borderPanels == null)
			return null;

		if ((position < NORTH) || (position > WEST))
			return null;

		return _borderPanels[position];
	}

	/**
	 * Set a border panel at the given position.
	 *
	 * @param  panel     the border panel.
	 * @param  position  Either NORTH, SOUTH, EAST or WEST position.
	 */
	public void setBorderPanel(LyPanel panel, short position)
	{
		if ((position < NORTH) || (position > WEST))
			throw new IllegalArgumentException("BF AppEngine : invalid position parameter " + position);

		if (_borderPanels == null)
			_borderPanels = new LyPanel[4];

		_borderPanels[position] = panel;

		if (panel != null)
			panel.setParent(this);

		if (_internalView instanceof LySetBorderPanelInterface)
			((LySetBorderPanelInterface)_internalView).setBorderPanel(panel, position);
	}

	/**
	 * Indicates whether given panel is one of the border panel for current area.
	 *
	 * @param   panel  the panel to be checked.
	 *
	 * @return  true if given panel is one of the border panel for current area, false otherwise.
	 */
	public boolean isBorderPanel(LyPanel panel)
	{
		if (_borderPanels == null)
			return false;

		for (int i = 0; i < _borderPanels.length; i++)
		{
			if (_borderPanels[i] == panel)
				return true;
		}

		return false;
	}

	/**
	 * Computes the position of each sub-panels for this area.<br />
	 * Those positions will be stored in the <code>_panelPositions</code> variable as an array of 4
	 * integers (column of the panel, row of the panel, width of the panel, height of the panel).
	 * <br />
	 * Global size of the grid will be stored in the <code>_layoutSize</code> variable as an array
	 * of two integers (width of the grid, height of the grid).
	 */
	public void computePanelsPositions()
	{
		int			 i, n = getPanelCount(), j, k;
		int			 rowIndex = 0, colIndex = 0, colspan, rowspan;
		boolean		 remainder;
		BitSet[]     grid = new BitSet[n];
		LyPanel		 subPanel;
		LyConstraint constraint;
		boolean		 onlyFill = true;
		
		_isColumnsEqualWidth = true;
		_layoutSize = new int[] { 0, 0 };
		_panelPositions = new Hashtable<LyPanel, int[]>(n);

		if (getLayout() == LyArea.Layout.BORDER)
		{
			_layoutSize = new int[] { 3, 3 };
			return;
		}
		
		for (i = 0; i < n; i++)
		{
			constraint = null;
			remainder = true;
			colspan = 1;
			rowspan = 1;
			subPanel = getPanel(i);

			// Retrieve constraint for sub-panel
			if (getLayout() == LyArea.Layout.FLOW)
				remainder = false;
			else if (subPanel instanceof LyGetConstraintInterface)
			{
				constraint = ((LyGetConstraintInterface)subPanel).getConstraint();

				if (constraint != null)
				{
					remainder = constraint.isRemainder();

					if (constraint.getColSpan() > 0)
						colspan = constraint.getColSpan();

					if (constraint.getRowSpan() > 0)
						rowspan = constraint.getRowSpan();
				}
				
				if ((constraint == null)
						|| (constraint.getFill() == LyConstraint.Fill.DEFAULT)
						|| (constraint.getFill() == LyConstraint.Fill.VERTICAL)
						|| (constraint.getFill() == LyConstraint.Fill.NONE))
				{
					onlyFill = false;
				}
			}

			// Fill the grid with cells where sub-panel is displayed
			for (j = rowIndex; (j < (rowIndex + rowspan)) && (j < grid.length); j++)
			{
				for (k = colIndex; k < (colIndex + colspan); k++)
				{
					if (grid[j] == null)
					{
						grid[j] = new BitSet();
						_layoutSize[1] = j + 1;
					}

					grid[j].set(k);
				}
			}

			_panelPositions.put(subPanel, new int[] { colIndex, rowIndex, colspan, rowspan });

			// Check if current column position is greater than before
			_layoutSize[0] = Math.max(colIndex + colspan, _layoutSize[0]);

			// Compute position of next sub-panel in the grid
			if (remainder)
			{
				rowIndex++;
				colIndex = 0;
				
				// If at least one line of sub-panels has only h-fill panels, set _isColumnsEqualWidth to true
				if (!onlyFill)
					_isColumnsEqualWidth = false;
				
				onlyFill = false;

				for (j = 0; (j < n) && (i < (n - 1)) && (grid[rowIndex] != null)
						&& (grid[rowIndex].get(j)); j++)
					colIndex++;
			}
			else
				colIndex = colIndex + colspan;
		}
	} // end method computePanelsPositions

	/**
	 * Returns the position of a sub-panel for this area.<br />
	 *
	 * @param   panel  the panel whose position must be returned.
	 *
	 * @return  the position of a sub-panel as an array of 4 integers (column of the panel, row of
	 *          the panel, width of the panel, height of the panel).
	 *
	 * @see     LyArea#computePanelsPositions()
	 */
	public int[] getPanelPosition(LyPanel panel)
	{
		return _panelPositions.get(panel);
	}

	/**
	 * Returns the size of the grid representing the layout of this area.
	 *
	 * @return  the size of the grid representing the layout of this area as an array of two
	 *          integers (width of the grid, height of the grid).
	 */
	public int[] getGridSize()
	{
		return _layoutSize;
	}
    
	/**
	 * Indicates whether all columns in the layout will be forced to have the same width.
	 * 
	 * @return true if all columns in the layout will be forced to have the same width, false otherwise.
	 */
	public boolean isColumnsEqualWidth()
	{
		return _isColumnsEqualWidth;
	}
	
    /**
     * Initializes attribute xmlFile to the given value.
     * @param xmlFile New value of xmlFile.
     */
    
	public void setXmlFile(String xmlFile)
    {
        super.setXmlFile(xmlFile);
        
        if (xmlFile == null)
            return;
        
        if  (_template != null)
            throw new IllegalStateException("BF AppEngine : can't set both template and xmlFile on area " + getId());
            
        // Parse XML file and add contents to this area
        LyEnvironment   environment = getEnvironment();
        LyViewParser    parser = new LyViewParser(environment, xmlFile);
        Object          rootElement = parser.getRootElement();

        if (rootElement instanceof List<?>)
        {
        	int length = ((List<?>)rootElement).size();
        	rootElement = ((List<?>)rootElement).get(length -1);
        }
        
        if (!(rootElement instanceof LyArea))
            throw new IllegalStateException("BF AppEngine : invalid xmlFile \"" + xmlFile
                    + "\" specified on area " + getId()
                    + ", root element is not an area (" + rootElement.getClass() + ")");

        setTemplate((LyArea)rootElement);
        _xmlFile = xmlFile;
    }
    
    /**
     * Checks if value of attribute hidden is set.
     * @return true if hidden is set, false otherwise.
     */
    public boolean isHidden()
    {
        return _hidden;
    }
    
    /**
     * Sets attribute hidden to the given value.
     * @param hidden Boolean indicating if attribute hidden is set or unset.
     */
    public void setHidden(boolean hidden)
    {
        if (hidden == _hidden)
            return;
        _hidden = hidden;
    }
    
    /*
     * (non-Javadoc)
     * @see leon.view.LyComponent#setVisible(boolean)
     */
    public void setVisible(boolean status)
    {   
        if (_hidden)
            status = false;
        super.setVisible(status);
    }
} // end class LyArea
