/*
 * LyAreaBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyAreaBase class is a generated Java implementation that corresponds to the element 
 * AREA of the descriptive grammar of the views (views.dtd).
 * <B>&lt;AREA&gt; : Area</B>
 * An area allows the positioning of various graphical components inside a main 
 * window or a dialog window.
 * It can contain any type of panel as well as a title bar, a tool bar and a 
 * command bar.
 * @see leon.view.LyContainer
 * @see leon.view.LyLyArea
 * See also tag <A href="../../conf/dtd/view/AREA.html">AREA</A> of grammar leon_views.dtd.
 */
public abstract class LyAreaBase extends LyContainer
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LyGetControllerInterface,
				LySetControllerInterface,
				LyGetRefreshInterface,
				LyGetConstraintInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'scrolling'. */
	public enum Scrolling
	{
		AUTO,
		YES,
		NO;
	}


	/** Possible values for attribute 'border'. */
	public enum Border
	{
		NONE,
		IN,
		OUT,
		SOLID,
		RAISED,
		LOWERED;
	}


	/** Possible values for attribute 'layout'. */
	public enum Layout
	{
		BORDER,
		CARD,
		FILL,
		FLOW,
		GRID;
	}


	/* Static methods */

	/**
	 * Utility method to get scrolling value from a string.
	 * @param token Token string
	 * @return Scrolling value corresponding to this token
	 */
	public static final Scrolling getScrolling(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("AUTO")
				|| token.equals("0"))
			return Scrolling.AUTO;
		else if (token.equalsIgnoreCase("YES")
				|| token.equals("1"))
			return Scrolling.YES;
		else if (token.equalsIgnoreCase("NO")
				|| token.equals("2"))
			return Scrolling.NO;
		else
			return Scrolling.NO;
	}

	/**
	 * Utility method to get border value from a string.
	 * @param token Token string
	 * @return Border value corresponding to this token
	 */
	public static final Border getBorder(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("NONE")
				|| token.equals("0"))
			return Border.NONE;
		else if (token.equalsIgnoreCase("IN")
				|| token.equals("1"))
			return Border.IN;
		else if (token.equalsIgnoreCase("OUT")
				|| token.equals("2"))
			return Border.OUT;
		else if (token.equalsIgnoreCase("SOLID")
				|| token.equals("3"))
			return Border.SOLID;
		else if (token.equalsIgnoreCase("RAISED")
				|| token.equals("4"))
			return Border.RAISED;
		else if (token.equalsIgnoreCase("LOWERED")
				|| token.equals("5"))
			return Border.LOWERED;
		else
			return Border.NONE;
	}

	/**
	 * Utility method to get layout value from a string.
	 * @param token Token string
	 * @return Layout value corresponding to this token
	 */
	public static final Layout getLayout(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("BORDER")
				|| token.equals("0"))
			return Layout.BORDER;
		else if (token.equalsIgnoreCase("CARD")
				|| token.equals("1"))
			return Layout.CARD;
		else if (token.equalsIgnoreCase("FILL")
				|| token.equals("2"))
			return Layout.FILL;
		else if (token.equalsIgnoreCase("FLOW")
				|| token.equals("3"))
			return Layout.FLOW;
		else if (token.equalsIgnoreCase("GRID")
				|| token.equals("4"))
			return Layout.GRID;
		else
			return Layout.GRID;
	}


	/* Attributes */

	/**
	 * Identifier of the area
	 */
	protected String			_id;

	/**
	 * Template of the area.
	 */
	protected LyArea			_template;

	/**
	 * Path of XML file containing area components.
	 */
	protected String			_xmlFile;

	/**
	 * Title of the area.
	 */
	protected String			_label;

	/**
	 * Tip associated to the area.
	 */
	protected String			_tip;

	/**
	 * Reference of the controller that manages this area (i.e. handles user demands 
	 * and refresh content).
	 */
	protected String			_controller;

	/**
	 * Indicates if the area should have scroll bars.The following scroll modes are 
	 * supported:
	 * AUTO The management of the scroll bars is automatic.
	 * If the graphical components contained in the area can managed their own scroll 
	 * bars, the area let them manage the scroll bars.
	 * YES The area must have scroll bars.
	 * NO The area does not have any scroll bar.
	 */
	protected Scrolling			_scrolling;

	/**
	 * Style of the border used for the area.
	 */
	protected Border			_border;

	/**
	 * Border width of the area.
	 */
	protected int				_borderWidth;

	/**
	 * Page to which the area is redirected.
	 * Ressource indicating to which URL the Web browser should be redirected when 
	 * displaying this area.
	 */
	protected String			_redirect;

	/**
	 * Refreshing interval of the area.
	 * Indicates (in seconds) the time after which the area should be updated.
	 */
	protected int				_refresh;

	/**
	 * Font used in this area.
	 */
	protected String			_font;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Font color of the component. 
	 */
	protected String			_fontColor;

	/**
	 * Vertical spacing of components displayed in the area.
	 */
	protected int				_verticalSpacing;

	/**
	 * Horizontal spacing of components displayed in the area.
	 */
	protected int				_horizontalSpacing;

	/**
	 * Layout used in the area.
	 * Supported values are : BORDER, CARD, FILL, FLOW, GRID (default layout).
	 */
	protected Layout			_layout;

	/**
	 * Attribute used to give additional information needed by the layout.
	 */
	protected String			_layoutInfo;

	/**
	 * Indicates whether the area is closable or not.
	 */
	protected boolean			_closable;

	/**
	 * Indicates whether the area is open or not (only if it is closable).
	 */
	protected boolean			_open;


	/* Sub components */

	/**
	 * Icon associated to the area.
	 */
	protected LyIcon			_icon;

	/**
	 * Title bar of the area.
	 */
	protected LyTitleBar		_titleBar;

	/**
	 * Toolbars of the area.
	 */
	protected ArrayList<LyToolBar>		_toolBars;

	/**
	 * Panels contained in the area.
	 */
	protected ArrayList<LyPanel>		_panels;

	/**
	 * Selection bar of the area.
	 */
	protected LySelectBar		_selectBar;

	/**
	 * Command bar of the area.
	 */
	protected LyCommands		_commands;

	/**
	 * Positioning constraints in the including area.
	 */
	protected LyConstraint		_constraint;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_xmlFile = null;
		_label = null;
		_tip = null;
		_controller = null;
		_scrolling = Scrolling.NO;
		_border = Border.NONE;
		_borderWidth = getIntValue("AREA.borderWidth", -1);
		_redirect = null;
		_refresh = getIntValue("AREA.refresh", 0);
		_font = getStringValue("AREA.font", "$ALL.font");
		_color = getStringValue("AREA.color", "$ALL.color");
		_fontColor = getStringValue("AREA.fontColor", "$ALL.fontColor");
		_verticalSpacing = getIntValue("AREA.verticalSpacing", -1);
		_horizontalSpacing = getIntValue("AREA.horizontalSpacing", -1);
		_layout = Layout.GRID;
		_layoutInfo = null;
		_closable = getBooleanValue("AREA.closable", false);
		_open = getBooleanValue("AREA.open", true);

		// Initialize sub components
		_icon = null;
		_titleBar = null;
		_toolBars = null;
		_panels = null;
		_selectBar = null;
		_commands = null;
		_constraint = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyArea	model = (LyArea)component;

		_model = component;
		_xmlFile = model._xmlFile;
		_label = model._label;
		_tip = model._tip;
		_controller = model._controller;
		_scrolling = model._scrolling;
		_border = model._border;
		_borderWidth = model._borderWidth;
		_redirect = model._redirect;
		_refresh = model._refresh;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;
		_verticalSpacing = model._verticalSpacing;
		_horizontalSpacing = model._horizontalSpacing;
		_layout = model._layout;
		_layoutInfo = model._layoutInfo;
		_closable = model._closable;
		_open = model._open;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		if (model._icon != null)
			addChild(model._icon.cloneComponent());
		if (model._titleBar != null)
			addChild(model._titleBar.cloneComponent());
		LyToolBar	toolBars;
		int			i, n;

		n = (model._toolBars == null ? 0 : model._toolBars.size());
		for (i = 0; i < n; i++)
		{
			toolBars = model._toolBars.get(i);
			if (toolBars != null)
				toolBars = (LyToolBar)toolBars.cloneComponent();
			addChild(toolBars);
		}

		LyPanel	panels;

		n = (model._panels == null ? 0 : model._panels.size());
		for (i = 0; i < n; i++)
		{
			panels = model._panels.get(i);
			if (panels != null)
				panels = (LyPanel)panels.cloneComponent();
			addChild(panels);
		}
		if (model._selectBar != null)
			addChild(model._selectBar.cloneComponent());
		if (model._commands != null)
			addChild(model._commands.cloneComponent());
		if (model._constraint != null)
			addChild(model._constraint.cloneComponent());
		if (model._datas != null)
			_datas = new Hashtable<String, String>(model._datas);
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyAreaBase()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyAreaBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyArea)node.getComponent("template");
		_xmlFile = node.getString("xmlFile");

		if (_template != null)
			initFromModel(_template);
		if (_xmlFile != null)
			setXmlFile(_xmlFile);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("tip"))
			_tip = node.getString("tip");
		if (node.isSpecified("controller"))
			_controller = node.getString("controller");
		if (node.isSpecified("scrolling"))
			_scrolling = getScrolling(node.getString("scrolling"));
		if (node.isSpecified("border"))
			_border = getBorder(node.getString("border"));
		if (node.isSpecified("borderWidth"))
			_borderWidth = node.getInt("borderWidth");
		if (node.isSpecified("redirect"))
			_redirect = node.getString("redirect");
		if (node.isSpecified("refresh"))
			_refresh = node.getInt("refresh");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("verticalSpacing"))
			_verticalSpacing = node.getInt("verticalSpacing");
		if (node.isSpecified("horizontalSpacing"))
			_horizontalSpacing = node.getInt("horizontalSpacing");
		if (node.isSpecified("layout"))
			_layout = getLayout(node.getString("layout"));
		if (node.isSpecified("layoutInfo"))
			_layoutInfo = node.getString("layoutInfo");
		if (node.isSpecified("closable"))
			_closable = node.getBoolean("closable");
		if (node.isSpecified("open"))
			_open = node.getBoolean("open");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyIcon)
			setIcon((LyIcon)child);
		else if (child instanceof LyTitleBar)
			setTitleBar((LyTitleBar)child);
		else if (child instanceof LyToolBar)
			addToolBar((LyToolBar)child);
		else if (child instanceof LyPanel)
			addPanel((LyPanel)child);
		else if (child instanceof LySelectBar)
			setSelectBar((LySelectBar)child);
		else if (child instanceof LyCommands)
			setCommands((LyCommands)child);
		else if (child instanceof LyConstraint)
			setConstraint((LyConstraint)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyArea clone = new LyArea();

		clone.initFromModel(this);
		clone._template = (LyArea)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyArea getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of xmlFile.
	 * @return value of xmlFile.
	 */
	public String getXmlFile()
	{
		return _xmlFile;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of tip.
	 * @return value of tip.
	 */
	public String getTip()
	{
		return _tip;
	}

	/**
	 * Gets value of controller.
	 * @return value of controller.
	 */
	public String getController()
	{
		return _controller;
	}

	/**
	 * Gets value of scrolling.
	 * @return value of scrolling.
	 */
	public Scrolling getScrolling()
	{
		return _scrolling;
	}

	/**
	 * Gets value of border.
	 * @return value of border.
	 */
	public Border getBorder()
	{
		return _border;
	}

	/**
	 * Gets value of borderWidth.
	 * @return value of borderWidth.
	 */
	public int getBorderWidth()
	{
		return _borderWidth;
	}

	/**
	 * Gets value of redirect.
	 * @return value of redirect.
	 */
	public String getRedirect()
	{
		return _redirect;
	}

	/**
	 * Gets value of refresh.
	 * @return value of refresh.
	 */
	public int getRefresh()
	{
		return _refresh;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("AREA.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("AREA.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("AREA.fontColor"));
	}

	/**
	 * Gets value of verticalSpacing.
	 * @return value of verticalSpacing.
	 */
	public int getVerticalSpacing()
	{
		return _verticalSpacing;
	}

	/**
	 * Gets value of horizontalSpacing.
	 * @return value of horizontalSpacing.
	 */
	public int getHorizontalSpacing()
	{
		return _horizontalSpacing;
	}

	/**
	 * Gets value of layout.
	 * @return value of layout.
	 */
	public Layout getLayout()
	{
		return _layout;
	}

	/**
	 * Gets value of layoutInfo.
	 * @return value of layoutInfo.
	 */
	public String getLayoutInfo()
	{
		return _layoutInfo;
	}

	/**
	 * Checks if value of attribute closable is set.
	 * @return true if closable is set, false otherwise.
	 */
	public boolean isClosable()
	{
		return _closable;
	}

	/**
	 * Checks if value of attribute open is set.
	 * @return true if open is set, false otherwise.
	 */
	public boolean isOpen()
	{
		return _open;
	}

	/**
	 * Get sub-component Icon.
	 * @return sub-component Icon.
	 */
	public LyIcon getIcon()
	{
		return _icon;
	}

	/**
	 * Get sub-component TitleBar.
	 * @return sub-component TitleBar.
	 */
	public LyTitleBar getTitleBar()
	{
		return _titleBar;
	}

	/**
	 * Get sub-component ToolBar of index i.
	 * @param i the index of the ToolBar to get.
	 * @return sub-component ToolBar of index i.
	 */
	public LyToolBar getToolBar(int i)
	{
		if (_toolBars == null)
			return null;
		if ((i < 0) || (i >= _toolBars.size()))
			return null;
		return _toolBars.get(i);
	}

	/**
	 * Get sub-component ToolBar count.
	 * @return sub-component ToolBar count.
	 */
	public int getToolBarCount()
	{
		if (_toolBars == null)
			return 0;
		return _toolBars.size();
	}

	/**
	 * Get sub-component Panel of index i.
	 * @param i the index of the Panel to get.
	 * @return sub-component Panel of index i.
	 */
	public LyPanel getPanel(int i)
	{
		if (_panels == null)
			return null;
		if ((i < 0) || (i >= _panels.size()))
			return null;
		return _panels.get(i);
	}

	/**
	 * Get sub-component Panel count.
	 * @return sub-component Panel count.
	 */
	public int getPanelCount()
	{
		if (_panels == null)
			return 0;
		return _panels.size();
	}

	/**
	 * Get sub-component SelectBar.
	 * @return sub-component SelectBar.
	 */
	public LySelectBar getSelectBar()
	{
		return _selectBar;
	}

	/**
	 * Get sub-component Commands.
	 * @return sub-component Commands.
	 */
	public LyCommands getCommands()
	{
		return _commands;
	}

	/**
	 * Get sub-component Constraint.
	 * @return sub-component Constraint.
	 */
	public LyConstraint getConstraint()
	{
		return _constraint;
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyArea template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Initializes attribute xmlFile to the given value.
	 * @param xmlFile New value of xmlFile.
	 */
	public void setXmlFile(String xmlFile)
	{
		if ((xmlFile == _xmlFile)
			|| ((xmlFile != null)
				&& (_xmlFile != null)
				&& xmlFile.equals(_xmlFile)))
			return;

		_xmlFile = xmlFile;
	}

	/**
	 * Initializes attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
	}

	/**
	 * Initializes attribute tip to the given value.
	 * @param tip New value of tip.
	 */
	public void setTip(String tip)
	{
		if ((tip == _tip)
			|| ((tip != null)
				&& (_tip != null)
				&& tip.equals(_tip)))
			return;

		_tip = tip;
	}

	/**
	 * Initializes attribute controller to the given value.
	 * @param controller New value of controller.
	 */
	public void setController(String controller)
	{
		if ((controller == _controller)
			|| ((controller != null)
				&& (_controller != null)
				&& controller.equals(_controller)))
			return;

		_controller = controller;
	}

	/**
	 * Initializes attribute scrolling to the given value.
	 * @param scrolling New value of scrolling.
	 * @exception IllegalArgumentException if scrolling parameter is null.
	 */
	public void setScrolling(Scrolling scrolling)
	{
		if (scrolling == _scrolling)
			return;
		if (scrolling == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute scrolling");

		_scrolling = scrolling;
	}

	/**
	 * Initializes attribute border to the given value.
	 * @param border New value of border.
	 * @exception IllegalArgumentException if border parameter is null.
	 */
	public void setBorder(Border border)
	{
		if (border == _border)
			return;
		if (border == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute border");

		_border = border;
	}

	/**
	 * Initializes attribute borderWidth to the given value.
	 * @param borderWidth New value of borderWidth.
	 */
	public void setBorderWidth(int borderWidth)
	{
		if (borderWidth == _borderWidth)
			return;
		_borderWidth = borderWidth;
	}

	/**
	 * Sets attribute redirect to the given value.
	 * @param redirect New value of redirect.
	 */
	public void setRedirect(String redirect)
	{
		if ((redirect == _redirect)
			|| ((redirect != null)
				&& (_redirect != null)
				&& redirect.equals(_redirect)))
			return;

		_redirect = redirect;
		if (_internalView != null)
			((LyGraphicalArea)_internalView).setRedirect(redirect);
	}

	/**
	 * Sets attribute refresh to the given value.
	 * @param refresh New value of refresh.
	 */
	public void setRefresh(int refresh)
	{
		if (refresh == _refresh)
			return;
		_refresh = refresh;
		if (_internalView != null)
			((LyGraphicalArea)_internalView).setRefresh(refresh);
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Initializes attribute verticalSpacing to the given value.
	 * @param verticalSpacing New value of verticalSpacing.
	 */
	public void setVerticalSpacing(int verticalSpacing)
	{
		if (verticalSpacing == _verticalSpacing)
			return;
		_verticalSpacing = verticalSpacing;
	}

	/**
	 * Initializes attribute horizontalSpacing to the given value.
	 * @param horizontalSpacing New value of horizontalSpacing.
	 */
	public void setHorizontalSpacing(int horizontalSpacing)
	{
		if (horizontalSpacing == _horizontalSpacing)
			return;
		_horizontalSpacing = horizontalSpacing;
	}

	/**
	 * Initializes attribute layout to the given value.
	 * @param layout New value of layout.
	 * @exception IllegalArgumentException if layout parameter is null.
	 */
	public void setLayout(Layout layout)
	{
		if (layout == _layout)
			return;
		if (layout == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute layout");

		_layout = layout;
	}

	/**
	 * Sets attribute layoutInfo to the given value.
	 * @param layoutInfo New value of layoutInfo.
	 */
	public void setLayoutInfo(String layoutInfo)
	{
		if ((layoutInfo == _layoutInfo)
			|| ((layoutInfo != null)
				&& (_layoutInfo != null)
				&& layoutInfo.equals(_layoutInfo)))
			return;

		_layoutInfo = layoutInfo;
		if (_internalView != null)
			((LyGraphicalArea)_internalView).setLayoutInfo(layoutInfo);
	}

	/**
	 * Initializes attribute closable to the given value.
	 * @param closable Boolean indicating if attribute closable is set or unset.
	 */
	public void setClosable(boolean closable)
	{
		if (closable == _closable)
			return;
		_closable = closable;
	}

	/**
	 * Sets attribute open to the given value.
	 * @param open Boolean indicating if attribute open is set or unset.
	 */
	public void setOpen(boolean open)
	{
		if (open == _open)
			return;
		_open = open;
		if (_internalView != null)
			((LyGraphicalArea)_internalView).setOpen(open);
	}

	/**
	 * Sets sub-component icon.
	 * @param icon New value of icon
	 */
	public void setIcon(LyIcon icon)
	{
		if (_icon != null)
			_icon.setParent(null);
		_icon = icon;
		if (icon != null)
			icon.setParent(this);
	}

	/**
	 * Sets sub-component titleBar.
	 * @param titleBar New value of titleBar
	 */
	public void setTitleBar(LyTitleBar titleBar)
	{
		if (_titleBar != null)
			_titleBar.setParent(null);
		_titleBar = titleBar;
		if (titleBar != null)
			titleBar.setParent(this);
	}

	/**
	 * Adds a sub-component toolBar.
	 * @param toolBar The toolBar to add to toolBars list.
	 */
	public void addToolBar(LyToolBar toolBar)
	{
		insertToolBar(toolBar, -1);
	}

	/**
	 * Inserts a sub-component toolBar at given position.
	 * @param toolBar The toolBar to add to toolBars list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertToolBar(LyToolBar toolBar, int pos)
	{
		if (_toolBars == null)
			_toolBars = new ArrayList<LyToolBar>(1);
		else if (_toolBars.contains(toolBar))
			return;
		if (pos >= 0)
			_toolBars.add(pos,toolBar);
		else
			_toolBars.add(toolBar);
		if (toolBar != null)
			toolBar.setParent(this);
	}

	/**
	 * Adds a sub-component panel.
	 * @param panel The panel to add to panels list.
	 */
	public void addPanel(LyPanel panel)
	{
		insertPanel(panel, -1);
	}

	/**
	 * Inserts a sub-component panel at given position.
	 * @param panel The panel to add to panels list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertPanel(LyPanel panel, int pos)
	{
		if (_panels == null)
			_panels = new ArrayList<LyPanel>(1);
		else if (_panels.contains(panel))
			return;
		if (pos >= 0)
			_panels.add(pos,panel);
		else
			_panels.add(panel);
		if (panel != null)
			panel.setParent(this);
		if (_internalView != null)
			((LyGraphicalArea)_internalView).addPanel(panel, pos);
	}

	/**
	 * Removes a sub-component panel.
	 * @param panel The panel to remove to panels list.
	 * @exception IllegalStateException if panels list is empty.
	 * @exception IllegalArgumentException if panels list doesn't contain panel.
	 */
	public void removePanel(LyPanel panel)
	{
		if (_panels == null)
			throw new IllegalStateException("BF AppEngine : panels list is empty");
		if (!_panels.contains(panel))
			throw new IllegalArgumentException("BF AppEngine : panels list doesn't contain panel");

		if (_internalView != null)
			((LyGraphicalArea)_internalView).removePanel(panel);
		_panels.remove(panel);
		panel.setParent(null);
	}

	/**
	 * Sets sub-component selectBar.
	 * @param selectBar New value of selectBar
	 */
	public void setSelectBar(LySelectBar selectBar)
	{
		if (_selectBar != null)
			_selectBar.setParent(null);
		_selectBar = selectBar;
		if (selectBar != null)
			selectBar.setParent(this);
	}

	/**
	 * Sets sub-component commands.
	 * @param commands New value of commands
	 */
	public void setCommands(LyCommands commands)
	{
		if (_commands != null)
			_commands.setParent(null);
		_commands = commands;
		if (commands != null)
			commands.setParent(this);
	}

	/**
	 * Sets sub-component constraint.
	 * @param constraint New value of constraint
	 */
	public void setConstraint(LyConstraint constraint)
	{
		if (_constraint != null)
			_constraint.setParent(null);
		_constraint = constraint;
		if (constraint != null)
			constraint.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<AREA");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if ((_template != null)
				&& (_xmlFile == null))
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_xmlFile != null))
				|| ((_template != null) && (_xmlFile != _template._xmlFile)))
			pw.print(" xmlFile='" + LyString.getXmlString(_xmlFile) + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (((_template == null) && (_tip != null))
				|| ((_template != null) && (_tip != _template._tip)))
			pw.print(" tip='" + LyString.getXmlString(_tip) + "'");
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlString(_controller) + "'");
		if ((_template == null)
			|| (_scrolling != _template._scrolling))
		{
			if (_scrolling == Scrolling.AUTO)
				pw.print(" scrolling='AUTO'");
			else if (_scrolling == Scrolling.YES)
				pw.print(" scrolling='YES'");
			else if ((_scrolling == Scrolling.NO) && (_template != null))
				pw.print(" scrolling='NO'");
		}
		if ((_template == null)
			|| (_border != _template._border))
		{
			if ((_border == Border.NONE) && (_template != null))
				pw.print(" border='NONE'");
			else if (_border == Border.IN)
				pw.print(" border='IN'");
			else if (_border == Border.OUT)
				pw.print(" border='OUT'");
			else if (_border == Border.SOLID)
				pw.print(" border='SOLID'");
			else if (_border == Border.RAISED)
				pw.print(" border='RAISED'");
			else if (_border == Border.LOWERED)
				pw.print(" border='LOWERED'");
		}
		if (((_template == null) && (_borderWidth != getIntValue("AREA.borderWidth", -1)))
				|| ((_template != null) && (_borderWidth != _template._borderWidth)))
			pw.print(" borderWidth='" + _borderWidth + "'");
		if (((_template == null) && (_redirect != null))
				|| ((_template != null) && (_redirect != _template._redirect)))
			pw.print(" redirect='" + LyString.getXmlString(_redirect) + "'");
		if (((_template == null) && (_refresh != getIntValue("AREA.refresh", 0)))
				|| ((_template != null) && (_refresh != _template._refresh)))
			pw.print(" refresh='" + _refresh + "'");
		if (((_template == null) && (_font != getStringValue("AREA.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("AREA.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("AREA.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && (_verticalSpacing != getIntValue("AREA.verticalSpacing", -1)))
				|| ((_template != null) && (_verticalSpacing != _template._verticalSpacing)))
			pw.print(" verticalSpacing='" + _verticalSpacing + "'");
		if (((_template == null) && (_horizontalSpacing != getIntValue("AREA.horizontalSpacing", -1)))
				|| ((_template != null) && (_horizontalSpacing != _template._horizontalSpacing)))
			pw.print(" horizontalSpacing='" + _horizontalSpacing + "'");
		if ((_template == null)
			|| (_layout != _template._layout))
		{
			if (_layout == Layout.BORDER)
				pw.print(" layout='BORDER'");
			else if (_layout == Layout.CARD)
				pw.print(" layout='CARD'");
			else if (_layout == Layout.FILL)
				pw.print(" layout='FILL'");
			else if (_layout == Layout.FLOW)
				pw.print(" layout='FLOW'");
			else if ((_layout == Layout.GRID) && (_template != null))
				pw.print(" layout='GRID'");
		}
		if (((_template == null) && (_layoutInfo != null))
				|| ((_template != null) && (_layoutInfo != _template._layoutInfo)))
			pw.print(" layoutInfo='" + LyString.getXmlString(_layoutInfo) + "'");
		if (((_template == null) && _closable)
				|| ((_template != null) && (_closable != _template._closable)))
			pw.print(" closable='" + _closable + "'");
		if (((_template == null) && !_open)
				|| ((_template != null) && (_open != _template._open)))
			pw.print(" open='" + _open + "'");

		if ((_xmlFile != null)
			|| ((getChildCount() == 0) && (_datas == null)))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		if (_icon != null)
			_icon.writeXml(pw, debug);
		if (_titleBar != null)
			_titleBar.writeXml(pw, debug);
		LyComponent	toolBars;
		int			i, n;

		n = (_toolBars == null ? 0 : _toolBars.size());
		for (i = 0; i < n; i++)
		{
			toolBars = _toolBars.get(i);
			toolBars.writeXml(pw, debug);
		}

		LyComponent	panels;

		n = (_panels == null ? 0 : _panels.size());
		for (i = 0; i < n; i++)
		{
			panels = _panels.get(i);
			panels.writeXml(pw, debug);
		}

		if (_selectBar != null)
			_selectBar.writeXml(pw, debug);
		if (_commands != null)
			_commands.writeXml(pw, debug);
		if (_constraint != null)
			_constraint.writeXml(pw, debug);

		writeDatas(pw);

		pw.println("</AREA>");
	}
}
