/*
 * LyCanvasBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyCanvasBase class is a generated Java implementation that corresponds to the element 
 * CANVAS of the descriptive grammar of the views (views.dtd).
 * <B>&lt;CANVAS&gt; : Canvas</B>
 * A canvas is a drawing area fully controlled by an external component. It allows 
 * a simple integration of external graphical libraries in a Application Engine 
 * application.
 * @see leon.view.LyPanel
 * @see leon.view.LyLyCanvas
 * See also tag <A href="../../conf/dtd/view/CANVAS.html">CANVAS</A> of grammar leon_views.dtd.
 */
public abstract class LyCanvasBase extends LyPanel
	implements  LyGetIdInterface,
				LyGetControllerInterface,
				LySetControllerInterface,
				LyGetConstraintInterface
{
	public static final String			__VERSION = "$Id$";


	/* Attributes */

	/**
	 * Identifier of the component.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyCanvas			_template;

	/**
	 * Java full name of the class that manages the drawing on the canvas and 
	 * interactions of the end-user on this drawing area.
	 */
	protected String			_javaName;

	/**
	 * Reference of the controller that manages this canvas (i.e. handles user demands 
	 * and refresh content).
	 */
	protected String			_controller;

	/**
	 * User data that can be used to give parameters for the management of the canvas 
	 * encoded in a string.
	 */
	protected String			_userData;


	/* Sub components */

	/**
	 * Positioning constraint of the canvas.
	 */
	protected LyConstraint		_constraint;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_javaName = null;
		_controller = null;
		_userData = null;

		// Initialize sub components
		_constraint = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyCanvas	model = (LyCanvas)component;

		_model = component;
		_javaName = model._javaName;
		_controller = model._controller;
		_userData = model._userData;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		if (model._constraint != null)
			addChild(model._constraint.cloneComponent());
		if (model._datas != null)
			_datas = new Hashtable<String, String>(model._datas);
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	protected LyCanvasBase()
	{
	}

	/**
	 * Constructor with all required attributes.
	 * @param javaName Java full name of the class that manages the drawing on the canvas and interactions of the end-user on this drawing area.
	 */
	public LyCanvasBase(String javaName)
	{
		// Fill up required attributes with parameters
		_javaName = javaName;
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyCanvasBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyCanvas)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("javaName"))
			_javaName = node.getString("javaName");
		if (node.isSpecified("controller"))
			_controller = node.getString("controller");
		if (node.isSpecified("userData"))
			_userData = node.getString("userData");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyConstraint)
			setConstraint((LyConstraint)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyCanvas clone = new LyCanvas();

		clone.initFromModel(this);
		clone._template = (LyCanvas)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyCanvas getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of javaName.
	 * @return value of javaName.
	 */
	public String getJavaName()
	{
		return _javaName;
	}

	/**
	 * Gets value of controller.
	 * @return value of controller.
	 */
	public String getController()
	{
		return _controller;
	}

	/**
	 * Gets value of userData.
	 * @return value of userData.
	 */
	public String getUserData()
	{
		return _userData;
	}

	/**
	 * Get sub-component Constraint.
	 * @return sub-component Constraint.
	 */
	public LyConstraint getConstraint()
	{
		return _constraint;
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyCanvas template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Initializes attribute javaName to the given value.
	 * @param javaName New value of javaName.
	 * @exception IllegalArgumentException if javaName is null.
	 */
	public void setJavaName(String javaName)
	{
		if (javaName == null)
			throw new IllegalArgumentException("BF AppEngine : attribute javaName is required (can't be null)");

		if ((javaName == _javaName)
			|| ((javaName != null)
				&& (_javaName != null)
				&& javaName.equals(_javaName)))
			return;

		_javaName = javaName;
	}

	/**
	 * Initializes attribute controller to the given value.
	 * @param controller New value of controller.
	 */
	public void setController(String controller)
	{
		if ((controller == _controller)
			|| ((controller != null)
				&& (_controller != null)
				&& controller.equals(_controller)))
			return;

		_controller = controller;
	}

	/**
	 * Initializes attribute userData to the given value.
	 * @param userData New value of userData.
	 */
	public void setUserData(String userData)
	{
		if ((userData == _userData)
			|| ((userData != null)
				&& (_userData != null)
				&& userData.equals(_userData)))
			return;

		_userData = userData;
	}

	/**
	 * Sets sub-component constraint.
	 * @param constraint New value of constraint
	 */
	public void setConstraint(LyConstraint constraint)
	{
		if (_constraint != null)
			_constraint.setParent(null);
		_constraint = constraint;
		if (constraint != null)
			constraint.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<CANVAS");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		pw.print(" javaName='" + LyString.getXmlString(_javaName) + "'");
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlString(_controller) + "'");
		if (((_template == null) && (_userData != null))
				|| ((_template != null) && (_userData != _template._userData)))
			pw.print(" userData='" + LyString.getXmlString(_userData) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		if (_constraint != null)
			_constraint.writeXml(pw, debug);

		writeDatas(pw);

		pw.println("</CANVAS>");
	}
}
