/*
 * LyChartBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyChartBase class is a generated Java implementation that corresponds to the element 
 * CHART of the descriptive grammar of the views (views.dtd).
 * <B>&lt;CHART&gt; : Statistic Chart</B>
 * A chart allows displaying a graph representing statistics on data.
 * @see leon.view.LyPanel
 * @see leon.view.LyLyChart
 * See also tag <A href="../../conf/dtd/view/CHART.html">CHART</A> of grammar leon_views.dtd.
 */
public abstract class LyChartBase extends LyPanel
	implements  LyGetIdInterface,
				LyGetControllerInterface,
				LySetControllerInterface,
				LyGetConstraintInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'graphType'. */
	public enum GraphType
	{
		STACK_BAR,
		BAR,
		PIE,
		SCATTER_PLOT,
		PLOT,
		STACK_AREA;
	}


	/** Possible values for attribute 'xVal'. */
	public enum XVal
	{
		X_REAL,
		X_INTERVAL;
	}


	/** Possible values for attribute 'yVal'. */
	public enum YVal
	{
		Y_REAL,
		Y_SUM_REAL,
		Y_COUNT_INTERVAL,
		Y_COUNT_REAL;
	}


	/** Possible values for attribute 'xType'. */
	public enum Type
	{
		CHARTNUMBER,
		CHARTTEXT,
		CHARTDATE;
	}


	/** Possible values for attribute 'yType'. */


	/* Static methods */

	/**
	 * Utility method to get graphType value from a string.
	 * @param token Token string
	 * @return GraphType value corresponding to this token
	 */
	public static final GraphType getGraphType(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("STACK_BAR")
				|| token.equals("0"))
			return GraphType.STACK_BAR;
		else if (token.equalsIgnoreCase("BAR")
				|| token.equals("1"))
			return GraphType.BAR;
		else if (token.equalsIgnoreCase("PIE")
				|| token.equals("2"))
			return GraphType.PIE;
		else if (token.equalsIgnoreCase("SCATTER_PLOT")
				|| token.equals("3"))
			return GraphType.SCATTER_PLOT;
		else if (token.equalsIgnoreCase("PLOT")
				|| token.equals("4"))
			return GraphType.PLOT;
		else if (token.equalsIgnoreCase("STACK_AREA")
				|| token.equals("5"))
			return GraphType.STACK_AREA;
		else
			return GraphType.STACK_BAR;
	}

	/**
	 * Utility method to get xVal value from a string.
	 * @param token Token string
	 * @return XVal value corresponding to this token
	 */
	public static final XVal getXVal(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("X_REAL")
				|| token.equals("0"))
			return XVal.X_REAL;
		else if (token.equalsIgnoreCase("X_INTERVAL")
				|| token.equals("1"))
			return XVal.X_INTERVAL;
		else
			return XVal.X_REAL;
	}

	/**
	 * Utility method to get yVal value from a string.
	 * @param token Token string
	 * @return YVal value corresponding to this token
	 */
	public static final YVal getYVal(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("Y_REAL")
				|| token.equals("0"))
			return YVal.Y_REAL;
		else if (token.equalsIgnoreCase("Y_SUM_REAL")
				|| token.equals("1"))
			return YVal.Y_SUM_REAL;
		else if (token.equalsIgnoreCase("Y_COUNT_INTERVAL")
				|| token.equals("2"))
			return YVal.Y_COUNT_INTERVAL;
		else if (token.equalsIgnoreCase("Y_COUNT_REAL")
				|| token.equals("3"))
			return YVal.Y_COUNT_REAL;
		else
			return YVal.Y_REAL;
	}

	/**
	 * Utility method to get xType value from a string.
	 * @param token Token string
	 * @return Type value corresponding to this token
	 */
	public static final Type getType(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("CHARTNUMBER")
				|| token.equals("0"))
			return Type.CHARTNUMBER;
		else if (token.equalsIgnoreCase("CHARTTEXT")
				|| token.equals("1"))
			return Type.CHARTTEXT;
		else if (token.equalsIgnoreCase("CHARTDATE")
				|| token.equals("2"))
			return Type.CHARTDATE;
		else
			return Type.CHARTTEXT;
	}


	/* Attributes */

	/**
	 * Identifier of the statistic.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyChart			_template;

	/**
	 * Reference of the controller that manages this chart (i.e. handles user demands 
	 * and refresh content).
	 */
	protected String			_controller;

	/**
	 * Types of available graphs.The following types of graphs are managed:
	 * STACK_BAR: Stacked bar chart
	 * BAR: Bar chart
	 * PIE: Pie chart
	 * SCATTER_PLOT: Point chart
	 * PLOT: Plot chart
	 * STACK_AREA: Stacked area chart
	 * By default, a STACK_BAR is selected.
	 */
	protected GraphType			_graphType;

	/**
	 * Indicates the way the values of X should be considered:
	 * X_REAL: Real values.
	 * X_INTERVAL: Grouping the values of X by intervals By default, an X_REAL is 
	 * selected.
	 */
	protected XVal				_xVal;

	/**
	 * Indicates the way the values of Y should be considered:
	 * Y_REAL: Real values.
	 * Y_SUM_REAL: Sum of the values of Y (only available for the numeric variables).
	 * Y_COUNT_INTERVAL: Grouping the values of Y by intervals, and counting each 
	 * group.
	 * Y_COUNT_REAL: Counting by the value of Y.
	 * By default, a Y_REAL is selected.
	 */
	protected YVal				_yVal;

	/**
	 * Indicates if the pie should correspond to the variables to analyze or to the 
	 * values of the 'according to' variable. By default, there is a pie per variable 
	 * to analyze.
	 */
	protected boolean			_pieBy;

	/**
	 * Allows defining a title that will be displayed on the graphical window.
	 */
	protected String			_title;

	/**
	 * Allows defining a title that will be displayed on the axis of the abscissa of 
	 * the graphical window.
	 * By default, this is the name of the variable in X.
	 */
	protected String			_xTitle;

	/**
	 * Allows to define a title that will be displayed on the axis of the ordinate of 
	 * the graphical window.  By default, this is the name of the variable in y, if 
	 * there is only one.
	 */
	protected String			_yTitle;

	/**
	 * Allows to define the type of the X variable in a broad sense: text, number, 
	 * date.
	 * CHARTNUMBER: Number values.
	 * CHARTTEXT: Strings.
	 * CHARTDATE: Dates.
	 * By default the type is CHARTTEXT (for example a choice is considered as a text).
	 */
	protected Type				_xType;

	/**
	 * Allows to define the type of the Y variables in a broad sense: text, number, 
	 * date. If you want to select several variables to study them, they must be of the
	 * same type.
	 * CHARTNUMBER: Number values.
	 * CHARTTEXT: Strings.
	 * CHARTDATE: Dates.
	 * By default the type is CHARTTEXT (for example a choice is considered as a text).
	 */
	protected Type				_yType;

	/**
	 * Allows defining the format of the dates to display in X.
	 * It corresponds to the formats used for the input fields of date type.
	 */
	protected String			_xDateType;

	/**
	 * Allows defining the format of the dates to display en Y.
	 * It corresponds to the formats used for the input fields of date type.
	 */
	protected String			_yDateType;

	/**
	 * Format of the LyFieldInfo ion X.
	 */
	protected String			_xFormat;

	/**
	 * Format of the LyFieldInfo in Y.
	 */
	protected String			_yFormat;

	/**
	 * Allows displaying vertically the labels of abscissa axis.
	 * This is useful when the labels are too long and they could be not all displayed.
	 */
	protected boolean			_xVerticalLabels;

	/**
	 * Allows drawing a vertical grid corresponding to the graduations of the 
	 * horizontal axis (only when the type of graph is not PIE).
	 */
	protected boolean			_xGrid;

	/**
	 * Allows drawing a horizontal grid corresponding to the graduations of the 
	 * vertical axis (only when the type of graph is not PIE).
	 */
	protected boolean			_yGrid;

	/**
	 * Allows drawing the graph with a 3D effect (only when the type of graph is PIE, 
	 * BAR or STACK_BAR).
	 */
	protected boolean			_d3Effect;

	/**
	 * Allows to display the legend
	 */
	protected boolean			_showLegend;

	/**
	 * If this attribute is to true, all the tips indicating the values of all the 
	 * points are displayed.
	 * Else, they are displayed only when the mouse cursor is moved over the points, if
	 * the viewer supports them (this is not the case for the HTML viewer)
	 */
	protected boolean			_showTips;

	/**
	 * Allows calculating percentages.
	 * By default, they are not calculated.
	 */
	protected boolean			_ratios;

	/**
	 * This attribute is set to true if there is no variable according to which the 
	 * variable(s) to analyze are studied.
	 */
	protected boolean			_noXField;

	/**
	 * Indicates if empty values must be displayed in the legend.
	 */
	protected boolean			_displayLegendEmptyValues;


	/* Sub components */

	/**
	 * Vectors of pairs of values (x, y). There is one COLUMN element per variable to 
	 * study.
	 */
	protected ArrayList<LyColumn>		_columns;

	/**
	 * Defines the intervals used to group together the values in X and in Y, as long 
	 * as one of the xVal or yVal attributes of the CHART element corresponds to 
	 * intervals.
	 */
	protected LyIntervals		_intervals;

	/**
	 * Defines the intervals on which the axis will be displayed, if the type of graph 
	 * that has been chosen is not pie.
	 */
	protected LyAxis			_axis;

	/**
	 * Positioning constraint of the chart.
	 */
	protected LyConstraint		_constraint;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_controller = null;
		_graphType = GraphType.STACK_BAR;
		_xVal = XVal.X_REAL;
		_yVal = YVal.Y_REAL;
		_pieBy = getBooleanValue("CHART.pieBy", false);
		_title = null;
		_xTitle = null;
		_yTitle = null;
		_xType = Type.CHARTTEXT;
		_yType = Type.CHARTTEXT;
		_xDateType = null;
		_yDateType = null;
		_xFormat = null;
		_yFormat = null;
		_xVerticalLabels = getBooleanValue("CHART.xVerticalLabels", false);
		_xGrid = getBooleanValue("CHART.xGrid", true);
		_yGrid = getBooleanValue("CHART.yGrid", true);
		_d3Effect = getBooleanValue("CHART.d3Effect", true);
		_showLegend = getBooleanValue("CHART.showLegend", true);
		_showTips = getBooleanValue("CHART.showTips", false);
		_ratios = getBooleanValue("CHART.ratios", false);
		_noXField = getBooleanValue("CHART.noXField", true);
		_displayLegendEmptyValues = getBooleanValue("CHART.displayLegendEmptyValues", true);

		// Initialize sub components
		_columns = null;
		_intervals = null;
		_axis = null;
		_constraint = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyChart	model = (LyChart)component;

		_model = component;
		_controller = model._controller;
		_graphType = model._graphType;
		_xVal = model._xVal;
		_yVal = model._yVal;
		_pieBy = model._pieBy;
		_title = model._title;
		_xTitle = model._xTitle;
		_yTitle = model._yTitle;
		_xType = model._xType;
		_yType = model._yType;
		_xDateType = model._xDateType;
		_yDateType = model._yDateType;
		_xFormat = model._xFormat;
		_yFormat = model._yFormat;
		_xVerticalLabels = model._xVerticalLabels;
		_xGrid = model._xGrid;
		_yGrid = model._yGrid;
		_d3Effect = model._d3Effect;
		_showLegend = model._showLegend;
		_showTips = model._showTips;
		_ratios = model._ratios;
		_noXField = model._noXField;
		_displayLegendEmptyValues = model._displayLegendEmptyValues;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyColumn	columns;
		int			i, n;

		n = (model._columns == null ? 0 : model._columns.size());
		for (i = 0; i < n; i++)
		{
			columns = model._columns.get(i);
			if (columns != null)
				columns = (LyColumn)columns.cloneComponent();
			addChild(columns);
		}
		if (model._intervals != null)
			addChild(model._intervals.cloneComponent());
		if (model._axis != null)
			addChild(model._axis.cloneComponent());
		if (model._constraint != null)
			addChild(model._constraint.cloneComponent());
		if (model._datas != null)
			_datas = new Hashtable<String, String>(model._datas);
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyChartBase()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyChartBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyChart)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("controller"))
			_controller = node.getString("controller");
		if (node.isSpecified("graphType"))
			_graphType = getGraphType(node.getString("graphType"));
		if (node.isSpecified("xVal"))
			_xVal = getXVal(node.getString("xVal"));
		if (node.isSpecified("yVal"))
			_yVal = getYVal(node.getString("yVal"));
		if (node.isSpecified("pieBy"))
			_pieBy = node.getBoolean("pieBy");
		if (node.isSpecified("title"))
			_title = node.getString("title");
		if (node.isSpecified("xTitle"))
			_xTitle = node.getString("xTitle");
		if (node.isSpecified("yTitle"))
			_yTitle = node.getString("yTitle");
		if (node.isSpecified("xType"))
			_xType = getType(node.getString("xType"));
		if (node.isSpecified("yType"))
			_yType = getType(node.getString("yType"));
		if (node.isSpecified("xDateType"))
			_xDateType = node.getString("xDateType");
		if (node.isSpecified("yDateType"))
			_yDateType = node.getString("yDateType");
		if (node.isSpecified("xFormat"))
			_xFormat = node.getString("xFormat");
		if (node.isSpecified("yFormat"))
			_yFormat = node.getString("yFormat");
		if (node.isSpecified("xVerticalLabels"))
			_xVerticalLabels = node.getBoolean("xVerticalLabels");
		if (node.isSpecified("xGrid"))
			_xGrid = node.getBoolean("xGrid");
		if (node.isSpecified("yGrid"))
			_yGrid = node.getBoolean("yGrid");
		if (node.isSpecified("d3Effect"))
			_d3Effect = node.getBoolean("d3Effect");
		if (node.isSpecified("showLegend"))
			_showLegend = node.getBoolean("showLegend");
		if (node.isSpecified("showTips"))
			_showTips = node.getBoolean("showTips");
		if (node.isSpecified("ratios"))
			_ratios = node.getBoolean("ratios");
		if (node.isSpecified("noXField"))
			_noXField = node.getBoolean("noXField");
		if (node.isSpecified("displayLegendEmptyValues"))
			_displayLegendEmptyValues = node.getBoolean("displayLegendEmptyValues");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyColumn)
			addColumn((LyColumn)child);
		else if (child instanceof LyIntervals)
			setIntervals((LyIntervals)child);
		else if (child instanceof LyAxis)
			setAxis((LyAxis)child);
		else if (child instanceof LyConstraint)
			setConstraint((LyConstraint)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyChart clone = new LyChart();

		clone.initFromModel(this);
		clone._template = (LyChart)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyChart getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of controller.
	 * @return value of controller.
	 */
	public String getController()
	{
		return _controller;
	}

	/**
	 * Gets value of graphType.
	 * @return value of graphType.
	 */
	public GraphType getGraphType()
	{
		return _graphType;
	}

	/**
	 * Gets value of xVal.
	 * @return value of xVal.
	 */
	public XVal getXVal()
	{
		return _xVal;
	}

	/**
	 * Gets value of yVal.
	 * @return value of yVal.
	 */
	public YVal getYVal()
	{
		return _yVal;
	}

	/**
	 * Checks if value of attribute pieBy is set.
	 * @return true if pieBy is set, false otherwise.
	 */
	public boolean isPieBy()
	{
		return _pieBy;
	}

	/**
	 * Gets value of title.
	 * @return value of title.
	 */
	public String getTitle()
	{
		return _title;
	}

	/**
	 * Gets value of xTitle.
	 * @return value of xTitle.
	 */
	public String getXTitle()
	{
		return _xTitle;
	}

	/**
	 * Gets value of yTitle.
	 * @return value of yTitle.
	 */
	public String getYTitle()
	{
		return _yTitle;
	}

	/**
	 * Gets value of xType.
	 * @return value of xType.
	 */
	public Type getXType()
	{
		return _xType;
	}

	/**
	 * Gets value of yType.
	 * @return value of yType.
	 */
	public Type getYType()
	{
		return _yType;
	}

	/**
	 * Gets value of xDateType.
	 * @return value of xDateType.
	 */
	public String getXDateType()
	{
		return _xDateType;
	}

	/**
	 * Gets value of yDateType.
	 * @return value of yDateType.
	 */
	public String getYDateType()
	{
		return _yDateType;
	}

	/**
	 * Gets value of xFormat.
	 * @return value of xFormat.
	 */
	public String getXFormat()
	{
		return _xFormat;
	}

	/**
	 * Gets value of yFormat.
	 * @return value of yFormat.
	 */
	public String getYFormat()
	{
		return _yFormat;
	}

	/**
	 * Checks if value of attribute xVerticalLabels is set.
	 * @return true if xVerticalLabels is set, false otherwise.
	 */
	public boolean isXVerticalLabels()
	{
		return _xVerticalLabels;
	}

	/**
	 * Checks if value of attribute xGrid is set.
	 * @return true if xGrid is set, false otherwise.
	 */
	public boolean isXGrid()
	{
		return _xGrid;
	}

	/**
	 * Checks if value of attribute yGrid is set.
	 * @return true if yGrid is set, false otherwise.
	 */
	public boolean isYGrid()
	{
		return _yGrid;
	}

	/**
	 * Checks if value of attribute d3Effect is set.
	 * @return true if d3Effect is set, false otherwise.
	 */
	public boolean isD3Effect()
	{
		return _d3Effect;
	}

	/**
	 * Checks if value of attribute showLegend is set.
	 * @return true if showLegend is set, false otherwise.
	 */
	public boolean isShowLegend()
	{
		return _showLegend;
	}

	/**
	 * Checks if value of attribute showTips is set.
	 * @return true if showTips is set, false otherwise.
	 */
	public boolean isShowTips()
	{
		return _showTips;
	}

	/**
	 * Checks if value of attribute ratios is set.
	 * @return true if ratios is set, false otherwise.
	 */
	public boolean isRatios()
	{
		return _ratios;
	}

	/**
	 * Checks if value of attribute noXField is set.
	 * @return true if noXField is set, false otherwise.
	 */
	public boolean isNoXField()
	{
		return _noXField;
	}

	/**
	 * Checks if value of attribute displayLegendEmptyValues is set.
	 * @return true if displayLegendEmptyValues is set, false otherwise.
	 */
	public boolean isDisplayLegendEmptyValues()
	{
		return _displayLegendEmptyValues;
	}

	/**
	 * Get sub-component Column of index i.
	 * @param i the index of the Column to get.
	 * @return sub-component Column of index i.
	 */
	public LyColumn getColumn(int i)
	{
		if (_columns == null)
			return null;
		if ((i < 0) || (i >= _columns.size()))
			return null;
		return _columns.get(i);
	}

	/**
	 * Get sub-component Column count.
	 * @return sub-component Column count.
	 */
	public int getColumnCount()
	{
		if (_columns == null)
			return 0;
		return _columns.size();
	}

	/**
	 * Get sub-component Intervals.
	 * @return sub-component Intervals.
	 */
	public LyIntervals getIntervals()
	{
		return _intervals;
	}

	/**
	 * Get sub-component Axis.
	 * @return sub-component Axis.
	 */
	public LyAxis getAxis()
	{
		return _axis;
	}

	/**
	 * Get sub-component Constraint.
	 * @return sub-component Constraint.
	 */
	public LyConstraint getConstraint()
	{
		return _constraint;
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyChart template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Initializes attribute controller to the given value.
	 * @param controller New value of controller.
	 */
	public void setController(String controller)
	{
		if ((controller == _controller)
			|| ((controller != null)
				&& (_controller != null)
				&& controller.equals(_controller)))
			return;

		_controller = controller;
	}

	/**
	 * Initializes attribute graphType to the given value.
	 * @param graphType New value of graphType.
	 * @exception IllegalArgumentException if graphType parameter is null.
	 */
	public void setGraphType(GraphType graphType)
	{
		if (graphType == _graphType)
			return;
		if (graphType == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute graphType");

		_graphType = graphType;
	}

	/**
	 * Initializes attribute xVal to the given value.
	 * @param xVal New value of xVal.
	 * @exception IllegalArgumentException if xVal parameter is null.
	 */
	public void setXVal(XVal xVal)
	{
		if (xVal == _xVal)
			return;
		if (xVal == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute xVal");

		_xVal = xVal;
	}

	/**
	 * Initializes attribute yVal to the given value.
	 * @param yVal New value of yVal.
	 * @exception IllegalArgumentException if yVal parameter is null.
	 */
	public void setYVal(YVal yVal)
	{
		if (yVal == _yVal)
			return;
		if (yVal == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute yVal");

		_yVal = yVal;
	}

	/**
	 * Initializes attribute pieBy to the given value.
	 * @param pieBy Boolean indicating if attribute pieBy is set or unset.
	 */
	public void setPieBy(boolean pieBy)
	{
		if (pieBy == _pieBy)
			return;
		_pieBy = pieBy;
	}

	/**
	 * Initializes attribute title to the given value.
	 * @param title New value of title.
	 */
	public void setTitle(String title)
	{
		if ((title == _title)
			|| ((title != null)
				&& (_title != null)
				&& title.equals(_title)))
			return;

		_title = title;
	}

	/**
	 * Initializes attribute xTitle to the given value.
	 * @param xTitle New value of xTitle.
	 */
	public void setXTitle(String xTitle)
	{
		if ((xTitle == _xTitle)
			|| ((xTitle != null)
				&& (_xTitle != null)
				&& xTitle.equals(_xTitle)))
			return;

		_xTitle = xTitle;
	}

	/**
	 * Initializes attribute yTitle to the given value.
	 * @param yTitle New value of yTitle.
	 */
	public void setYTitle(String yTitle)
	{
		if ((yTitle == _yTitle)
			|| ((yTitle != null)
				&& (_yTitle != null)
				&& yTitle.equals(_yTitle)))
			return;

		_yTitle = yTitle;
	}

	/**
	 * Initializes attribute xType to the given value.
	 * @param xType New value of xType.
	 * @exception IllegalArgumentException if xType parameter is null.
	 */
	public void setXType(Type xType)
	{
		if (xType == _xType)
			return;
		if (xType == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute xType");

		_xType = xType;
	}

	/**
	 * Initializes attribute yType to the given value.
	 * @param yType New value of yType.
	 * @exception IllegalArgumentException if yType parameter is null.
	 */
	public void setYType(Type yType)
	{
		if (yType == _yType)
			return;
		if (yType == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute yType");

		_yType = yType;
	}

	/**
	 * Initializes attribute xDateType to the given value.
	 * @param xDateType New value of xDateType.
	 */
	public void setXDateType(String xDateType)
	{
		if ((xDateType == _xDateType)
			|| ((xDateType != null)
				&& (_xDateType != null)
				&& xDateType.equals(_xDateType)))
			return;

		_xDateType = xDateType;
	}

	/**
	 * Initializes attribute yDateType to the given value.
	 * @param yDateType New value of yDateType.
	 */
	public void setYDateType(String yDateType)
	{
		if ((yDateType == _yDateType)
			|| ((yDateType != null)
				&& (_yDateType != null)
				&& yDateType.equals(_yDateType)))
			return;

		_yDateType = yDateType;
	}

	/**
	 * Initializes attribute xFormat to the given value.
	 * @param xFormat New value of xFormat.
	 */
	public void setXFormat(String xFormat)
	{
		if ((xFormat == _xFormat)
			|| ((xFormat != null)
				&& (_xFormat != null)
				&& xFormat.equals(_xFormat)))
			return;

		_xFormat = xFormat;
	}

	/**
	 * Initializes attribute yFormat to the given value.
	 * @param yFormat New value of yFormat.
	 */
	public void setYFormat(String yFormat)
	{
		if ((yFormat == _yFormat)
			|| ((yFormat != null)
				&& (_yFormat != null)
				&& yFormat.equals(_yFormat)))
			return;

		_yFormat = yFormat;
	}

	/**
	 * Initializes attribute xVerticalLabels to the given value.
	 * @param xVerticalLabels Boolean indicating if attribute xVerticalLabels is set or unset.
	 */
	public void setXVerticalLabels(boolean xVerticalLabels)
	{
		if (xVerticalLabels == _xVerticalLabels)
			return;
		_xVerticalLabels = xVerticalLabels;
	}

	/**
	 * Initializes attribute xGrid to the given value.
	 * @param xGrid Boolean indicating if attribute xGrid is set or unset.
	 */
	public void setXGrid(boolean xGrid)
	{
		if (xGrid == _xGrid)
			return;
		_xGrid = xGrid;
	}

	/**
	 * Initializes attribute yGrid to the given value.
	 * @param yGrid Boolean indicating if attribute yGrid is set or unset.
	 */
	public void setYGrid(boolean yGrid)
	{
		if (yGrid == _yGrid)
			return;
		_yGrid = yGrid;
	}

	/**
	 * Initializes attribute d3Effect to the given value.
	 * @param d3Effect Boolean indicating if attribute d3Effect is set or unset.
	 */
	public void setD3Effect(boolean d3Effect)
	{
		if (d3Effect == _d3Effect)
			return;
		_d3Effect = d3Effect;
	}

	/**
	 * Initializes attribute showLegend to the given value.
	 * @param showLegend Boolean indicating if attribute showLegend is set or unset.
	 */
	public void setShowLegend(boolean showLegend)
	{
		if (showLegend == _showLegend)
			return;
		_showLegend = showLegend;
	}

	/**
	 * Initializes attribute showTips to the given value.
	 * @param showTips Boolean indicating if attribute showTips is set or unset.
	 */
	public void setShowTips(boolean showTips)
	{
		if (showTips == _showTips)
			return;
		_showTips = showTips;
	}

	/**
	 * Initializes attribute ratios to the given value.
	 * @param ratios Boolean indicating if attribute ratios is set or unset.
	 */
	public void setRatios(boolean ratios)
	{
		if (ratios == _ratios)
			return;
		_ratios = ratios;
	}

	/**
	 * Initializes attribute noXField to the given value.
	 * @param noXField Boolean indicating if attribute noXField is set or unset.
	 */
	public void setNoXField(boolean noXField)
	{
		if (noXField == _noXField)
			return;
		_noXField = noXField;
	}

	/**
	 * Initializes attribute displayLegendEmptyValues to the given value.
	 * @param displayLegendEmptyValues Boolean indicating if attribute displayLegendEmptyValues is set or unset.
	 */
	public void setDisplayLegendEmptyValues(boolean displayLegendEmptyValues)
	{
		if (displayLegendEmptyValues == _displayLegendEmptyValues)
			return;
		_displayLegendEmptyValues = displayLegendEmptyValues;
	}

	/**
	 * Adds a sub-component column.
	 * @param column The column to add to columns list.
	 */
	public void addColumn(LyColumn column)
	{
		insertColumn(column, -1);
	}

	/**
	 * Inserts a sub-component column at given position.
	 * @param column The column to add to columns list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertColumn(LyColumn column, int pos)
	{
		if (_columns == null)
			_columns = new ArrayList<LyColumn>(1);
		else if (_columns.contains(column))
			return;
		if (pos >= 0)
			_columns.add(pos,column);
		else
			_columns.add(column);
		if (column != null)
			column.setParent(this);
		if (_internalView != null)
			((LyGraphicalChart)_internalView).addColumn(column, pos);
	}

	/**
	 * Removes a sub-component column.
	 * @param column The column to remove to columns list.
	 * @exception IllegalStateException if columns list is empty.
	 * @exception IllegalArgumentException if columns list doesn't contain column.
	 */
	public void removeColumn(LyColumn column)
	{
		if (_columns == null)
			throw new IllegalStateException("BF AppEngine : columns list is empty");
		if (!_columns.contains(column))
			throw new IllegalArgumentException("BF AppEngine : columns list doesn't contain column");

		if (_internalView != null)
			((LyGraphicalChart)_internalView).removeColumn(column);
		_columns.remove(column);
		column.setParent(null);
	}

	/**
	 * Sets sub-component intervals.
	 * @param intervals New value of intervals
	 */
	public void setIntervals(LyIntervals intervals)
	{
		if (_intervals != null)
			_intervals.setParent(null);
		_intervals = intervals;
		if (intervals != null)
			intervals.setParent(this);
	}

	/**
	 * Sets sub-component axis.
	 * @param axis New value of axis
	 */
	public void setAxis(LyAxis axis)
	{
		if (_axis != null)
			_axis.setParent(null);
		_axis = axis;
		if (axis != null)
			axis.setParent(this);
	}

	/**
	 * Sets sub-component constraint.
	 * @param constraint New value of constraint
	 */
	public void setConstraint(LyConstraint constraint)
	{
		if (_constraint != null)
			_constraint.setParent(null);
		_constraint = constraint;
		if (constraint != null)
			constraint.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<CHART");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlString(_controller) + "'");
		if ((_template == null)
			|| (_graphType != _template._graphType))
		{
			if ((_graphType == GraphType.STACK_BAR) && (_template != null))
				pw.print(" graphType='STACK_BAR'");
			else if (_graphType == GraphType.BAR)
				pw.print(" graphType='BAR'");
			else if (_graphType == GraphType.PIE)
				pw.print(" graphType='PIE'");
			else if (_graphType == GraphType.SCATTER_PLOT)
				pw.print(" graphType='SCATTER_PLOT'");
			else if (_graphType == GraphType.PLOT)
				pw.print(" graphType='PLOT'");
			else if (_graphType == GraphType.STACK_AREA)
				pw.print(" graphType='STACK_AREA'");
		}
		if ((_template == null)
			|| (_xVal != _template._xVal))
		{
			if ((_xVal == XVal.X_REAL) && (_template != null))
				pw.print(" xVal='X_REAL'");
			else if (_xVal == XVal.X_INTERVAL)
				pw.print(" xVal='X_INTERVAL'");
		}
		if ((_template == null)
			|| (_yVal != _template._yVal))
		{
			if ((_yVal == YVal.Y_REAL) && (_template != null))
				pw.print(" yVal='Y_REAL'");
			else if (_yVal == YVal.Y_SUM_REAL)
				pw.print(" yVal='Y_SUM_REAL'");
			else if (_yVal == YVal.Y_COUNT_INTERVAL)
				pw.print(" yVal='Y_COUNT_INTERVAL'");
			else if (_yVal == YVal.Y_COUNT_REAL)
				pw.print(" yVal='Y_COUNT_REAL'");
		}
		if (((_template == null) && _pieBy)
				|| ((_template != null) && (_pieBy != _template._pieBy)))
			pw.print(" pieBy='" + _pieBy + "'");
		if (((_template == null) && (_title != null))
				|| ((_template != null) && (_title != _template._title)))
			pw.print(" title='" + LyString.getXmlString(_title) + "'");
		if (((_template == null) && (_xTitle != null))
				|| ((_template != null) && (_xTitle != _template._xTitle)))
			pw.print(" xTitle='" + LyString.getXmlString(_xTitle) + "'");
		if (((_template == null) && (_yTitle != null))
				|| ((_template != null) && (_yTitle != _template._yTitle)))
			pw.print(" yTitle='" + LyString.getXmlString(_yTitle) + "'");
		if ((_template == null)
			|| (_xType != _template._xType))
		{
			if (_xType == Type.CHARTNUMBER)
				pw.print(" xType='CHARTNUMBER'");
			else if ((_xType == Type.CHARTTEXT) && (_template != null))
				pw.print(" xType='CHARTTEXT'");
			else if (_xType == Type.CHARTDATE)
				pw.print(" xType='CHARTDATE'");
		}
		if ((_template == null)
			|| (_yType != _template._yType))
		{
			if (_yType == Type.CHARTNUMBER)
				pw.print(" yType='CHARTNUMBER'");
			else if ((_yType == Type.CHARTTEXT) && (_template != null))
				pw.print(" yType='CHARTTEXT'");
			else if (_yType == Type.CHARTDATE)
				pw.print(" yType='CHARTDATE'");
		}
		if (((_template == null) && (_xDateType != null))
				|| ((_template != null) && (_xDateType != _template._xDateType)))
			pw.print(" xDateType='" + LyString.getXmlString(_xDateType) + "'");
		if (((_template == null) && (_yDateType != null))
				|| ((_template != null) && (_yDateType != _template._yDateType)))
			pw.print(" yDateType='" + LyString.getXmlString(_yDateType) + "'");
		if (((_template == null) && (_xFormat != null))
				|| ((_template != null) && (_xFormat != _template._xFormat)))
			pw.print(" xFormat='" + LyString.getXmlString(_xFormat) + "'");
		if (((_template == null) && (_yFormat != null))
				|| ((_template != null) && (_yFormat != _template._yFormat)))
			pw.print(" yFormat='" + LyString.getXmlString(_yFormat) + "'");
		if (((_template == null) && _xVerticalLabels)
				|| ((_template != null) && (_xVerticalLabels != _template._xVerticalLabels)))
			pw.print(" xVerticalLabels='" + _xVerticalLabels + "'");
		if (((_template == null) && !_xGrid)
				|| ((_template != null) && (_xGrid != _template._xGrid)))
			pw.print(" xGrid='" + _xGrid + "'");
		if (((_template == null) && !_yGrid)
				|| ((_template != null) && (_yGrid != _template._yGrid)))
			pw.print(" yGrid='" + _yGrid + "'");
		if (((_template == null) && !_d3Effect)
				|| ((_template != null) && (_d3Effect != _template._d3Effect)))
			pw.print(" d3Effect='" + _d3Effect + "'");
		if (((_template == null) && !_showLegend)
				|| ((_template != null) && (_showLegend != _template._showLegend)))
			pw.print(" showLegend='" + _showLegend + "'");
		if (((_template == null) && _showTips)
				|| ((_template != null) && (_showTips != _template._showTips)))
			pw.print(" showTips='" + _showTips + "'");
		if (((_template == null) && _ratios)
				|| ((_template != null) && (_ratios != _template._ratios)))
			pw.print(" ratios='" + _ratios + "'");
		if (((_template == null) && !_noXField)
				|| ((_template != null) && (_noXField != _template._noXField)))
			pw.print(" noXField='" + _noXField + "'");
		if (((_template == null) && !_displayLegendEmptyValues)
				|| ((_template != null) && (_displayLegendEmptyValues != _template._displayLegendEmptyValues)))
			pw.print(" displayLegendEmptyValues='" + _displayLegendEmptyValues + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	columns;
		int			i, n;

		n = (_columns == null ? 0 : _columns.size());
		for (i = 0; i < n; i++)
		{
			columns = _columns.get(i);
			columns.writeXml(pw, debug);
		}

		if (_intervals != null)
			_intervals.writeXml(pw, debug);
		if (_axis != null)
			_axis.writeXml(pw, debug);
		if (_constraint != null)
			_constraint.writeXml(pw, debug);

		writeDatas(pw);

		pw.println("</CHART>");
	}
}
