/*
 * LyChoiceBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.getinterface.*;
import leon.misc.LyString;


/**
 * The LyChoiceBase class is a generated Java implementation that corresponds to the element 
 * CHOICE of the descriptive grammar of the views (views.dtd).
 * <B>&lt;CHOICE&gt; : Input Field of Choice Type</B>
 * An input field of Choice type allows displaying a graphical component containing
 * a set of check boxes.
 * It allows choosing one or more values among a discrete list of values.
 * It is possible to specify whether the choice is multiple or not and whether the 
 * boxes should be arranged horizontally or vertically.
 * @see leon.view.LyFieldItem
 * @see leon.view.LyLyChoice
 * See also tag <A href="../../conf/dtd/view/CHOICE.html">CHOICE</A> of grammar leon_views.dtd.
 */
public abstract class LyChoiceBase extends LyFieldItem
	implements  LyGetIdInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'fill'. */
	public enum Fill
	{
		HORIZONTAL,
		VERTICAL;
	}


	/* Static methods */

	/**
	 * Utility method to get fill value from a string.
	 * @param token Token string
	 * @return Fill value corresponding to this token
	 */
	public static final Fill getFill(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("HORIZONTAL")
				|| token.equals("0"))
			return Fill.HORIZONTAL;
		else if (token.equalsIgnoreCase("VERTICAL")
				|| token.equals("1"))
			return Fill.VERTICAL;
		else
			return Fill.HORIZONTAL;
	}


	/* Attributes */

	/**
	 * Identifier of the component.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyChoice			_template;

	/**
	 * Indicates whether the choice is multiple or not.
	 */
	protected boolean			_multiple;

	/**
	 * Indicates whether the boxes should be arranged horizontally or vertically.
	 */
	protected Fill				_fill;

	/**
	 * Indicates the maximum number of boxes on the same line (case of an horizontal 
	 * alignment) or on the same column (case of a vertical alignment).
	 */
	protected int				_maxSpan;

	/**
	 * Allows specifying the name of a Java class that replaces the default graphical 
	 * component.
	 */
	protected String			_bean;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * Gives the number of columns of the choice (given in number of characters).
	 */
	protected int				_nCols;

	/**
	 * Gives the number of lines on which the options are shown.
	 */
	protected int				_nRows;


	/* Sub components */

	/**
	 * List of the options offered for this choice
	 */
	protected ArrayList<LyOption>		_options;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_multiple = getBooleanValue("CHOICE.multiple", false);
		_fill = Fill.HORIZONTAL;
		_maxSpan = getIntValue("CHOICE.maxSpan", -1);
		_bean = null;
		_font = getStringValue("CHOICE.font", "$ALL.font");
		_color = getStringValue("CHOICE.color", "$ALL.color");
		_fontColor = getStringValue("CHOICE.fontColor", "$ALL.fontColor");
		_nCols = getIntValue("CHOICE.nCols", -1);
		_nRows = getIntValue("CHOICE.nRows", -1);

		// Initialize sub components
		_options = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyChoice	model = (LyChoice)component;

		_model = component;
		_multiple = model._multiple;
		_fill = model._fill;
		_maxSpan = model._maxSpan;
		_bean = model._bean;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;
		_nCols = model._nCols;
		_nRows = model._nRows;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyOption	options;
		int			i, n;

		n = (model._options == null ? 0 : model._options.size());
		for (i = 0; i < n; i++)
		{
			options = model._options.get(i);
			if (options != null)
				options = (LyOption)options.cloneComponent();
			addChild(options);
		}
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyChoiceBase()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyChoiceBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyChoice)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("multiple"))
			_multiple = node.getBoolean("multiple");
		if (node.isSpecified("fill"))
			_fill = getFill(node.getString("fill"));
		if (node.isSpecified("maxSpan"))
			_maxSpan = node.getInt("maxSpan");
		if (node.isSpecified("bean"))
			_bean = node.getString("bean");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("nCols"))
			_nCols = node.getInt("nCols");
		if (node.isSpecified("nRows"))
			_nRows = node.getInt("nRows");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyOption)
			addOption((LyOption)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyChoice clone = new LyChoice();

		clone.initFromModel(this);
		clone._template = (LyChoice)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyChoice getTemplate()
	{
		return _template;
	}

	/**
	 * Checks if value of attribute multiple is set.
	 * @return true if multiple is set, false otherwise.
	 */
	public boolean isMultiple()
	{
		return _multiple;
	}

	/**
	 * Gets value of fill.
	 * @return value of fill.
	 */
	public Fill getFill()
	{
		return _fill;
	}

	/**
	 * Gets value of maxSpan.
	 * @return value of maxSpan.
	 */
	public int getMaxSpan()
	{
		return _maxSpan;
	}

	/**
	 * Gets value of bean.
	 * @return value of bean.
	 */
	public String getBean()
	{
		return _bean;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("CHOICE.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("CHOICE.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("CHOICE.fontColor"));
	}

	/**
	 * Gets value of nCols.
	 * @return value of nCols.
	 */
	public int getNCols()
	{
		return _nCols;
	}

	/**
	 * Gets value of nRows.
	 * @return value of nRows.
	 */
	public int getNRows()
	{
		return _nRows;
	}

	/**
	 * Get sub-component Option of index i.
	 * @param i the index of the Option to get.
	 * @return sub-component Option of index i.
	 */
	public LyOption getOption(int i)
	{
		if (_options == null)
			return null;
		if ((i < 0) || (i >= _options.size()))
			return null;
		return _options.get(i);
	}

	/**
	 * Get sub-component Option count.
	 * @return sub-component Option count.
	 */
	public int getOptionCount()
	{
		if (_options == null)
			return 0;
		return _options.size();
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyChoice template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Initializes attribute multiple to the given value.
	 * @param multiple Boolean indicating if attribute multiple is set or unset.
	 */
	public void setMultiple(boolean multiple)
	{
		if (multiple == _multiple)
			return;
		_multiple = multiple;
	}

	/**
	 * Initializes attribute fill to the given value.
	 * @param fill New value of fill.
	 * @exception IllegalArgumentException if fill parameter is null.
	 */
	public void setFill(Fill fill)
	{
		if (fill == _fill)
			return;
		if (fill == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute fill");

		_fill = fill;
	}

	/**
	 * Initializes attribute maxSpan to the given value.
	 * @param maxSpan New value of maxSpan.
	 */
	public void setMaxSpan(int maxSpan)
	{
		if (maxSpan == _maxSpan)
			return;
		_maxSpan = maxSpan;
	}

	/**
	 * Initializes attribute bean to the given value.
	 * @param bean New value of bean.
	 */
	public void setBean(String bean)
	{
		if ((bean == _bean)
			|| ((bean != null)
				&& (_bean != null)
				&& bean.equals(_bean)))
			return;

		_bean = bean;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Initializes attribute nCols to the given value.
	 * @param nCols New value of nCols.
	 */
	public void setNCols(int nCols)
	{
		if (nCols == _nCols)
			return;
		_nCols = nCols;
	}

	/**
	 * Initializes attribute nRows to the given value.
	 * @param nRows New value of nRows.
	 */
	public void setNRows(int nRows)
	{
		if (nRows == _nRows)
			return;
		_nRows = nRows;
	}

	/**
	 * Adds a sub-component option.
	 * @param option The option to add to options list.
	 */
	public void addOption(LyOption option)
	{
		insertOption(option, -1);
	}

	/**
	 * Inserts a sub-component option at given position.
	 * @param option The option to add to options list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertOption(LyOption option, int pos)
	{
		if (_options == null)
		{
			_options = new ArrayList<LyOption>(1);
			_children = _options;
		}
		else if (_options.contains(option))
			return;
		if (pos >= 0)
			_options.add(pos,option);
		if (option != null)
			option.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<CHOICE");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && _multiple)
				|| ((_template != null) && (_multiple != _template._multiple)))
			pw.print(" multiple='" + _multiple + "'");
		if ((_template == null)
			|| (_fill != _template._fill))
		{
			if ((_fill == Fill.HORIZONTAL) && (_template != null))
				pw.print(" fill='HORIZONTAL'");
			else if (_fill == Fill.VERTICAL)
				pw.print(" fill='VERTICAL'");
		}
		if (((_template == null) && (_maxSpan != getIntValue("CHOICE.maxSpan", -1)))
				|| ((_template != null) && (_maxSpan != _template._maxSpan)))
			pw.print(" maxSpan='" + _maxSpan + "'");
		if (((_template == null) && (_bean != null))
				|| ((_template != null) && (_bean != _template._bean)))
			pw.print(" bean='" + LyString.getXmlString(_bean) + "'");
		if (((_template == null) && (_font != getStringValue("CHOICE.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("CHOICE.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("CHOICE.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && (_nCols != getIntValue("CHOICE.nCols", -1)))
				|| ((_template != null) && (_nCols != _template._nCols)))
			pw.print(" nCols='" + _nCols + "'");
		if (((_template == null) && (_nRows != getIntValue("CHOICE.nRows", -1)))
				|| ((_template != null) && (_nRows != _template._nRows)))
			pw.print(" nRows='" + _nRows + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	options;
		int			i, n;

		n = (_options == null ? 0 : _options.size());
		for (i = 0; i < n; i++)
		{
			options = _options.get(i);
			options.writeXml(pw, debug);
		}


		pw.println("</CHOICE>");
	}
}
