/*
 * LyColumn.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.misc.LyString;


/**
 * The LyColumn class is a generated Java implementation that corresponds to the element 
 * COLUMN of the descriptive grammar of the views (views.dtd).
 * <B>&lt;COLUMN&gt; : Set of Statistic Values</B>
 * A COLUMN element is a set of pairs (x, y) where y is a value (a real or a sum or
 * a count) of a variable to analyze and x is either a value of the variable 
 * according to which y is studied (if there is one) or the name of the variable to
 * study. There is one COLUMN element per variable to analyze.
 * @see leon.view.LyComponent
 * See also tag <A href="../../conf/dtd/view/COLUMN.html">COLUMN</A> of grammar leon_views.dtd.
 */
public class LyColumn extends LyComponent
	implements  LyGetLabelInterface
{
	public static final String			__VERSION = "$Id$";


	/* Attributes */

	/**
	 * Name of the variable to study.
	 */
	protected String			_label;


	/* Sub components */

	/**
	 * Pairs of values (x, y).
	 */
	protected ArrayList<LyChartValue>		_chartValues;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_label = null;

		// Initialize sub components
		_chartValues = new ArrayList<LyChartValue>(1);
		_children = _chartValues;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyColumn	model = (LyColumn)component;

		_model = component;
		_label = model._label;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyChartValue	chartValues;
		int			i, n;

		n = model._chartValues.size();
		for (i = 0; i < n; i++)
		{
			chartValues = model._chartValues.get(i);
			if (chartValues != null)
				chartValues = (LyChartValue)chartValues.cloneComponent();
			addChild(chartValues);
		}
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyColumn()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyColumn(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_label = node.getString("label");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyChartValue)
			addChartValue((LyChartValue)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyColumn clone = new LyColumn();

		clone.initFromModel(this);
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Get sub-component ChartValue of index i.
	 * @param i the index of the ChartValue to get.
	 * @return sub-component ChartValue of index i.
	 */
	public LyChartValue getChartValue(int i)
	{
		if ((i < 0) || (i >= _chartValues.size()))
			return null;
		return _chartValues.get(i);
	}

	/**
	 * Get sub-component ChartValue count.
	 * @return sub-component ChartValue count.
	 */
	public int getChartValueCount()
	{
		return _chartValues.size();
	}


	/* Modifiers */

	/**
	 * Initializes attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
	}

	/**
	 * Adds a sub-component chartValue.
	 * @param chartValue The chartValue to add to chartValues list.
	 */
	public void addChartValue(LyChartValue chartValue)
	{
		insertChartValue(chartValue, -1);
	}

	/**
	 * Inserts a sub-component chartValue at given position.
	 * @param chartValue The chartValue to add to chartValues list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertChartValue(LyChartValue chartValue, int pos)
	{
		if (pos >= 0)
			_chartValues.add(pos,chartValue);
		if (chartValue != null)
			chartValue.setParent(this);
		if (_internalView != null)
			((LyGraphicalColumn)_internalView).addChartValue(chartValue, pos);
	}

	/**
	 * Removes a sub-component chartValue.
	 * @param chartValue The chartValue to remove to chartValues list.
	 * @exception IllegalStateException if chartValues list is empty.
	 * @exception IllegalArgumentException if chartValues list doesn't contain chartValue.
	 */
	public void removeChartValue(LyChartValue chartValue)
	{
		if (!_chartValues.contains(chartValue))
			throw new IllegalArgumentException("BF AppEngine : chartValues list doesn't contain chartValue");

		if (_internalView != null)
			((LyGraphicalColumn)_internalView).removeChartValue(chartValue);
		chartValue.setParent(null);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<COLUMN");
		if (_label != null)
			pw.print(" label='" + LyString.getXmlString(_label) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	chartValues;
		int			i, n;

		n = _chartValues.size();
		for (i = 0; i < n; i++)
		{
			chartValues = _chartValues.get(i);
			chartValues.writeXml(pw, debug);
		}


		pw.println("</COLUMN>");
	}
}
