/*
 * LyCommand.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyCommand class is a generated Java implementation that corresponds to the element 
 * COMMAND of the descriptive grammar of the views (views.dtd).
 * <B>&lt;COMMAND&gt; : Command Button</B>
 * A command button allows executing an action.
 * There are predefined types of command giving information about the nature of the
 * action to execute (closing, cancellation, help, etc.).
 * @see leon.view.LyPanel
 * See also tag <A href="../../conf/dtd/view/COMMAND.html">COMMAND</A> of grammar leon_views.dtd.
 */
public class LyCommand extends LyPanel
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface,
				LySetEnabledInterface,
				LyGetConstraintInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'action'. */
	public enum Action
	{
		CANCEL,
		SUBMIT,
		RESET,
		HELP,
		OTHER,
		CLOSE;
	}


	/* Static methods */

	/**
	 * Utility method to get action value from a string.
	 * @param token Token string
	 * @return Action value corresponding to this token
	 */
	public static final Action getAction(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("CANCEL")
				|| token.equals("0"))
			return Action.CANCEL;
		else if (token.equalsIgnoreCase("SUBMIT")
				|| token.equals("1"))
			return Action.SUBMIT;
		else if (token.equalsIgnoreCase("RESET")
				|| token.equals("2"))
			return Action.RESET;
		else if (token.equalsIgnoreCase("HELP")
				|| token.equals("3"))
			return Action.HELP;
		else if (token.equalsIgnoreCase("OTHER")
				|| token.equals("4"))
			return Action.OTHER;
		else if (token.equalsIgnoreCase("CLOSE")
				|| token.equals("5"))
			return Action.CLOSE;
		else
			return Action.OTHER;
	}


	/* Attributes */

	/**
	 * Identifier of the command button.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyCommand			_template;

	/**
	 * Text of the command button.
	 */
	protected String			_label;

	/**
	 * Image of the command button.
	 */
	protected String			_image;

	/**
	 * Indicates whether the command button is enabled or not.
	 */
	protected boolean			_enabled;

	/**
	 * Type of action generated by the command button.
	 * The following types of action are supported:
	 * CANCEL Action of cancellation type (example: on a modification form).
	 * SUBMIT Action of validation type (example: on a creation form).
	 * RESET Action of reset type (example: on a modification form).
	 * HELP Action of help type.
	 * OTHER Action which type does not belong to any predefined type.
	 * CLOSE Action of closing the window type.
	 */
	protected Action			_action;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;


	/* Sub components */

	/**
	 * Positioning constraints in the including area.
	 */
	protected LyConstraint		_constraint;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_image = null;
		_enabled = getBooleanValue("COMMAND.enabled", true);
		_action = Action.OTHER;
		_font = getStringValue("COMMAND.font", "$ALL.font");
		_color = getStringValue("COMMAND.color", "$ALL.color");
		_fontColor = getStringValue("COMMAND.fontColor", "$ALL.fontColor");

		// Initialize sub components
		_constraint = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyCommand	model = (LyCommand)component;

		_model = component;
		_label = model._label;
		_image = model._image;
		_enabled = model._enabled;
		_action = model._action;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		if (model._constraint != null)
			addChild(model._constraint.cloneComponent());
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	protected LyCommand()
	{
		_id = getInternalId();
	}

	/**
	 * Constructor with all required attributes.
	 * @param id Identifier of the command button.
	 * @param label Text of the command button.
	 */
	public LyCommand(String id, String label)
	{
		// Fill up required attributes with parameters
		_id = (id != null ? id.intern() : null);
		_label = label;
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyCommand(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyCommand)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("image"))
			_image = node.getString("image");
		if (node.isSpecified("enabled"))
			_enabled = node.getBoolean("enabled");
		if (node.isSpecified("action"))
			_action = getAction(node.getString("action"));
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyConstraint)
			setConstraint((LyConstraint)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyCommand clone = new LyCommand();

		clone.initFromModel(this);
		clone._template = this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyCommand getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of image.
	 * @return value of image.
	 */
	public String getImage()
	{
		return _image;
	}

	/**
	 * Checks if value of attribute enabled is set.
	 * @return true if enabled is set, false otherwise.
	 */
	public boolean isEnabled()
	{
		return _enabled;
	}

	/**
	 * Gets value of action.
	 * @return value of action.
	 */
	public Action getAction()
	{
		return _action;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("COMMAND.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("COMMAND.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("COMMAND.fontColor"));
	}

	/**
	 * Get sub-component Constraint.
	 * @return sub-component Constraint.
	 */
	public LyConstraint getConstraint()
	{
		return _constraint;
	}


	/* Modifiers */

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyCommand template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 * @exception IllegalArgumentException if label is null.
	 */
	public void setLabel(String label)
	{
		if (label == null)
			throw new IllegalArgumentException("BF AppEngine : attribute label is required (can't be null)");

		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView != null)
			((LyGraphicalCommand)_internalView).setLabel(label);
	}

	/**
	 * Initializes attribute image to the given value.
	 * @param image New value of image.
	 */
	public void setImage(String image)
	{
		if ((image == _image)
			|| ((image != null)
				&& (_image != null)
				&& image.equals(_image)))
			return;

		_image = image;
	}

	/**
	 * Sets attribute enabled to the given value.
	 * @param enabled Boolean indicating if attribute enabled is set or unset.
	 */
	public void setEnabled(boolean enabled)
	{
		if (enabled == _enabled)
			return;
		_enabled = enabled;
		if (_internalView != null)
			((LyGraphicalCommand)_internalView).setEnabled(enabled);
	}

	/**
	 * Initializes attribute action to the given value.
	 * @param action New value of action.
	 * @exception IllegalArgumentException if action parameter is null.
	 */
	public void setAction(Action action)
	{
		if (action == _action)
			return;
		if (action == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute action");

		_action = action;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Sets sub-component constraint.
	 * @param constraint New value of constraint
	 */
	public void setConstraint(LyConstraint constraint)
	{
		if (_constraint != null)
			_constraint.setParent(null);
		_constraint = constraint;
		if (constraint != null)
			constraint.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<COMMAND");
		pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (((_template == null) && (_image != null))
				|| ((_template != null) && (_image != _template._image)))
			pw.print(" image='" + LyString.getXmlString(_image) + "'");
		if (((_template == null) && !_enabled)
				|| ((_template != null) && (_enabled != _template._enabled)))
			pw.print(" enabled='" + _enabled + "'");
		if ((_template == null)
			|| (_action != _template._action))
		{
			if (_action == Action.CANCEL)
				pw.print(" action='CANCEL'");
			else if (_action == Action.SUBMIT)
				pw.print(" action='SUBMIT'");
			else if (_action == Action.RESET)
				pw.print(" action='RESET'");
			else if (_action == Action.HELP)
				pw.print(" action='HELP'");
			else if ((_action == Action.OTHER) && (_template != null))
				pw.print(" action='OTHER'");
			else if (_action == Action.CLOSE)
				pw.print(" action='CLOSE'");
		}
		if (((_template == null) && (_font != getStringValue("COMMAND.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("COMMAND.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("COMMAND.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		if (_constraint != null)
			_constraint.writeXml(pw, debug);

		pw.println("</COMMAND>");
	}
}
