/*
 * LyComponent.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import leon.misc.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;

import java.io.*;
import java.util.*;


/**
 * This class implements the generic behavior of the classes of the viewer package. LyComponents are
 * created when parsing the XML tree describing the graphical objects that must be shown to the
 * user. LyComponent implements LyActionTransmitter methods, but the interface is NOT declared.
 * Inherited classes just have to indicate that they implement this interface to benefit from
 * transmission features.
 *
 * @version  $Id: LyComponent.java 41911 2013-01-10 13:38:36Z smaupin $
 * @author   Lyria
 */
public abstract class LyComponent
{
	public static final String __VERSION = "$Id: LyComponent.java 41911 2013-01-10 13:38:36Z smaupin $";

	/** Id for sep components. */
    public static final String    SEP_ID = "_sep".intern();
    
    /**
     * Boolean used for activating raw XML writing of components
     * This boolean inhibits attributes initialization from properties and initFromModel.
     * Reserved for internal use.
     */
    public static boolean   REWRITING_MODE = false;
    
    /**
     * Boolean used for activating the search for id based 
     * If set to false, the initForId method for a component won't be called.
     * Reserved for internal use.
     */
    public static boolean   INIT_RESOURCE_ID = true;
    
	/** Separator used in internal ids. */
	public static String    INTERNAL_ID_SEP = "_";

	public static int _nbComponents;
    
	/** Used in record or replay modes. */
	protected static HashMap<String, LyComponent> _components = new HashMap<String, LyComponent>();
	protected static int						  _counter  = 0;

	/** Identifier for the default function of an event type. */
	private static final String EVENT_DEFAULT_FUNCTION = "EVENT_DEFAULT_FUNCTION";
	
	/** Type of events available on abstract components. */
	public enum EventType
	{
		MOUSE_ENTERED,
		MOUSE_EXITED,
		MOUSE_CLICKED,
		MOUSE_PRESSED,
		MOUSE_RELEASED,
		KEY_TYPED,
		KEY_RELEASED,
		KEY_PRESSED,
		FOCUS_GAINED,
		FOCUS_LOST,
		PASTE;
	}

	/**
	 * Internal view implementation. The nature of this object depends on the requested target (win,
	 * swing, html...) and is accessed through the setinterface package.
	 */
	protected LyViewComponent _internalView;

	/** Parent component. */
	protected LyComponent _parent;

	/** Son components list. */
	
	protected ArrayList<?> _children;

	/** Optimization to enhance children seek. */
	protected Hashtable<String, LyComponent> _childCache;

	/** Optimization to enhance children seek. */
	protected Hashtable<String, LyComponent> _childCacheForView;

	/** Optimization to enhance children seek. */
	protected Hashtable<LyComponent, int[]> _childCacheBounds;

	/** Indicates if component is freeing. */
	protected boolean _freeing;

	/** Indicates if component is freed. */
	protected boolean _freed;

	/** Internal id used in record or replay mode. */
	protected String _internalId;
    
    /** User datas. */
    protected Hashtable<String, String> _datas;
    
    /** Model used for cloning component. */
    protected LyComponent _model;
    
    /** List of functions that must be called for a given type of event. */
    protected HashMap<EventType, Vector<String>> _eventFunctions;
    

	/**
	 * Method to register a component in the association table. For internal use only (in record or
	 * replay mode).
	 */
	public static void addComponent(LyComponent component)
	{
		String internalId = component.getInternalId();

		_components.put(internalId, component);
	}

	/**
	 * Method used to get a component by its internal id. For internal use only (in record or replay
	 * mode).
	 */
	public static LyComponent getComponentByInternalId(String internalId)
	{
		return _components.get(internalId);
	}

	/**
	 * Method to remove a component from the association table. For internal use only (in record or
	 * replay mode).
	 */
	public static void removeComponent(LyComponent component)
	{
		_components.remove(component._internalId);
	}

	/**
	 * Default constructor. Initialization is done with the init method.
	 */
	public LyComponent()
	{
		init();
		_nbComponents++;
	}

	protected void finalize()
	{
		_nbComponents--;
	}

	/**
	 * Initialisation of the data of the instance.
	 */
	protected void init()
	{
		// Initialize class attributes.
		_internalView = null;
		_children = null;
		_parent = null;
		_childCache = null;
		_childCacheForView = null;
		_freed = false;
        _datas = null;
        _model = null;
        _eventFunctions = null;
	}
    
    /**
     * Inits component from a view node.
     */
    protected abstract void initFromViewNode(LyViewNode viewNode);
    
    /**
     * Utility method to retrieves the value of an attribute from a key corresponding to an entry in a values file.
     */
    protected String getStringValue(String str)
    {
        return getStringValue(str, null);
    }
    
    /**
     * Utility method to retrieves the value of an attribute from a key corresponding to an entry in a values file.
     */ 
    protected String getStringValue(String str, String defaultValue)
    {
        if (REWRITING_MODE)
        {
            if ((defaultValue != null)
                    && (defaultValue.length() > 0)
                    && (defaultValue.charAt(0) == '$'))
                return null;
            
            return defaultValue;
        }
        
    	 LyEnvironment   env = getEnvironment();
         String          value = (env == null ? str : env.getViewAttValue(str));
         
         if (value == str)
             value = defaultValue;
         if ((value != null)
         		&& (value.length() > 0)
                && (value.charAt(0) == '$'))
             value = getStringValue(value.substring(1));
         if (value == null)
         	value = env.getString(str);
         if (value == str) // if the value is still equals to str, put value to null...
        	 value = null;

         if (value != null)
        	 value = value.intern();
         
         return value;
    }
   
    /**
     * Utility method to retrieves the value of an attribute from a key corresponding to an entry in a values file.
     */
    protected String getValue(String str)
    {        
        String  value = getStringValue(str);
        
        if (value == null)
            return null;
        
        LyEnvironment   env = getEnvironment();
        
        if (env == null)
            return str;
        
        value = env.getResource(value);
        return value;
    }
    
    /**
     * Utility method to retrieves the value of an attribute from a key corresponding to an entry in a values file.
     */
    protected boolean getBooleanValue(String str, boolean defaultValue)
    {
        String  value = getValue(str);
        
        return (value == null  ? defaultValue : LyTools.booleanFromString(value));
    }
    
    /**
     * Utility method to retrieves the value of an attribute from a key corresponding to an entry in a values file.
     */
    protected int getIntValue(String str, int defaultValue)
    {
        String  value = getValue(str);
        
        return (value == null  ? defaultValue : LyTools.intFromString(value));
    }
    
    /**
     * Utility method to retrieves the value of an attribute from a key corresponding to an entry in a values file.
     */
    protected short getShortValue(String str, short defaultValue)
    {
        String  value = getValue(str);
        
        return (value == null  ? defaultValue : LyTools.shortFromString(value));
    }
    
    /**
     * Utility method to retrieves the value of an attribute from a key corresponding to an entry in a values file.
     */
    protected double getDoubleValue(String str, double defaultValue)
    {
        String  value = getValue(str);
        
        return (value == null  ? defaultValue : LyTools.doubleFromString(value));
    }
    
    /**
     * Utility method to retrieves the value of an attribute from a key corresponding to an entry in a values file.
     */
    protected float getFloatValue(String str, float defaultValue)
    {
        String  value = getValue(str);
        
        return (value == null  ? defaultValue : LyTools.floatFromString(value));
    }
    
    /**
     * Method for cloning a component.
     */
    public abstract LyComponent cloneComponent();
    
	/**
	 * Returns the internal id of this component (may be null).
	 */
	public String getInternalId()
	{
		if (_internalId != null)
			return _internalId;

		if (this instanceof LyGetIdInterface)
			_internalId = ((LyGetIdInterface)this).getId();

		if ((_internalId == null)
		    && (_parent != null))
		{
			int i, n = _parent.getChildCount();

			for (i = 0; i < n; i++)
				if (_parent.getChild(i) == this)
				{
					_internalId = _parent.getInternalId() + INTERNAL_ID_SEP + i;
					break;
				}
		}

		if (_internalId == null)
		{
			_internalId = getClass().getName() + _counter;
			_counter++;
		}

		_internalId = _internalId.intern();
		return _internalId;
	}

    /**
     * Gets tag name corresponding to this component.
     *
     * @return  The corresponding tag name.
     */
    public String getTag()
    {
        return LyViewNodeFactory.getTag(getClass().getName());
    }

    /**
     * Gets component id (may be null).
     */
    public String getId()
    {
        // By default, returns null
        return null;
    }
    
    /**
     * Internal code used as identifier for web components.
     * @return Internal code as a string.
     */
    public String getIdCode()
    {
       if (LyEnvironment.TEST_MODE)
	    {
	    	String internalId = getInternalId();
	    	
	    	LyViewManager viewManager = getViewManager();
	    	if (viewManager != null && viewManager.isMultiSession() && internalId != null)
	    	{
	    		internalId = LyString.replace(internalId, new char[] {(char)3, '\\'}, new char[] {'_', '_'});
	    	}
	    	return internalId;
	    }
	    else
	    {
	        return Integer.toString(hashCode());
	    }
    }
    
	/**
	 * Returns the component environment.
	 */
	public LyEnvironment getEnvironment()
	{
		if (_parent != null)
			return _parent.getEnvironment();

		return LyEnvironment.getCurrentEnvironment();
	}

	/**
	 * Gets current view manager.
	 *
	 * @return  The current view manager.
	 */
	public final LyViewManager getViewManager()
	{
        LyEnvironment   environment = getEnvironment();
        
		return (environment == null ? null : (LyViewManager)environment.getViewManager());
	}

	/**
	 * Retrieve a node from a view by its id. Access is made from the root view only.
	 *
	 * @param   id  unique id of the node in the view.
	 *
	 * @return  the node if found, null otherwise.
	 */
	public final LyComponent getComponent(String id)
	{
		return getComponent(id, false);
	}

	/**
	 * Retrieve a node from a view by its id. Access is made from the root view only.
	 *
	 * @param   id              unique id of the node in the view.
	 * @param   componentClass  expected class
	 *
	 * @return  the node if found and if class of node matches expected class, null otherwise.
	 */
	public LyComponent getComponent(String id, Class<?> componentClass)
	{
		LyComponent component = getComponent(id);

		if (componentClass == null)
			return component;

		if (component == null)
			return null;

		if (!componentClass.isAssignableFrom(component.getClass()))
		{
			// Warns user and return null
			Exception e = new IllegalStateException("BF AppEngine : component with id "
					+ id + " (class " + component.getClass()
					+ ") doesn't match the expected class " + componentClass);

			e.printStackTrace();
			return null;
		}

		return component;
	}

	/**
	 * Retrieve a node from a view by its id. Access is made from the root view only.
	 *
	 * @param   id  unique id of the node in the view.
	 *
	 * @return  the node if found, null otherwise.
	 */
	public final LyComponent getComponent(String id, boolean skipCurrent)
	{
		return getComponent(id, skipCurrent, false);
	}

	/**
	 * Retrieve a node from a view by its id. Access is made from the root view only.
	 *
	 * @param   id             unique id of the node in the view.
	 * @param   skipCurrent    indicates whether current component must be tested or not.
	 * @param   useInternalId  indicates whether internal identifiers may be used to retrieve node
	 *                         with given id.
	 *
	 * @return  the node if found, null otherwise.
	 */
	public LyComponent getComponent(String id, boolean skipCurrent, boolean useInternalId)
	{
		if (id == null)
			return null;

		id = id.intern();
		
		if (!skipCurrent
				&& (this instanceof LyGetIdInterface)
				&& (id == ((LyGetIdInterface)this).getId()))
			return this;

		if (useInternalId
				&& !skipCurrent
				&& (id == getInternalId()))
			return this;

		if (_children == null)
			return null;

		LyComponent result = getChildFromCache(id);

		if (result != null)
			return result;

	    synchronized (_children)
	    {
	        LyComponent child;
	        int		    i, n = _children.size();

    		for (i = 0; i < n; i++)
    		{
    			child = (LyComponent)_children.get(i);
    			if (child == null)
    				continue;
    
    			result = child.getComponent(id, false, useInternalId);
    			if (result != null)
    			{
    				addChildInCache(result, false);
    				return result;
    			}
    		}
		}

		return null;
	} // end method getComponent

	/**
	 * Retrieve nodes from a view by their ids. Note that this method does not use the component
	 * cache.
	 *
	 * @param  id      id of the nodes in the view.
	 * @param  result  The vector containing all the components which match the id.
	 */
	public void getComponent(String id, Vector<LyComponent> result)
	{
		if (id == null)
			return;

		if ((this instanceof LyGetIdInterface)
				&& (id == ((LyGetIdInterface)this).getId()))
			result.addElement(this);
		
		if (_children == null)
		    return;
 
        synchronized (_children)
        {
            LyComponent   child;
            int           i, n = _children.size();

    		for (i = 0; i < n; i++)
    		{
    			child = (LyComponent)_children.get(i);
    
    			if (child == null)
    				continue;
    
    			child.getComponent(id, result);
    		}
	    }
	}

	/**
	 * Retrieve nodes from a view by their class. Note that this method does not use the component
	 * cache.
	 *
	 * @param  compClass  class of the nodes in the view.
	 * @param  result     The vector containing all the components which belongs to the given class.
	 */
	public <T> void getComponents(Class<T> compClass, Vector<T> result)
	{
		if (compClass == null)
			return;
		
		if (_children == null)
		    return;

        synchronized (_children)
        {
    		LyComponent child;
    		int		    i, n = _children.size();
    
    		for (i = 0; i < n; i++)
    		{
    			child = (LyComponent)_children.get(i);
    			if (child == null)
    				continue;
    
    			if (compClass.isAssignableFrom(child.getClass()))
    			{
    				@SuppressWarnings("unchecked")
    				T temp = (T)child;
    				result.addElement(temp);
    			}
    
    			child.getComponents(compClass, result);
    		}
        }
	}

	/**
	 * Retrieve a node from a view. Access is made from the root view only.
	 *
	 * @param   label  label of the node in the view.
	 *
	 * @return  the node if found, null otherwise.
	 */
	public LyComponent getComponentByLabel(String label)
	{
		if (label == null)
			return null;

		if ((this instanceof LyGetLabelInterface)
				&& label.equals(((LyGetLabelInterface)this).getLabel()))
			return this;

		if (_children == null)
		    return null;
		
        synchronized (_children)
        {
    		LyComponent child, result;
    		int		    i, n = _children.size();
    
    		for (i = 0; i < n; i++)
    		{
    		    child = (LyComponent)_children.get(i);;
    			if (child == null)
    				continue;
    
    			result = child.getComponentByLabel(label);
    			if (result != null)
    				return result;
    		}
        }

		return null;
	}

    /**
     * Retrieve a child component from a model.
     *
     * @param   model  Model of the searched component.
     * @return  The child found or null if not found.
     */
    public LyComponent getComponentWithModel(LyComponent model)
    {
        if (model == null)
            return null;
        if (_children == null)
            return null;
        
        synchronized (_children)
        {
            LyComponent child;
            int         i, n = _children.size();
    
            for (i = 0; i < n; i++)
            {
                child = (LyComponent)_children.get(i);
                if (child._model == model)
                    return child;
            }
        }

        return null;
    }
    
	/**
	 * Internal view implementation. The nature of this object depends on the requested target (win,
	 * swing, html...) and is accessed through the setinterface package.
	 */
	public final LyViewComponent getInternalView()
	{
		return getInternalView(true);
	}

	/**
	 * Get internal view implementation.
	 *
	 * @param  build  Indicates if this representation should be built if it doesn't exist.
	 */
	public LyViewComponent getInternalView(boolean build)
	{
		if (_internalView != null)
			return _internalView;
		
		if (!build)
			return null;
		
		synchronized (getLock())
		{
			if ((_internalView == null) && build) // test again with the lock
			{
				if (_freeing || _freed)
					return null;

				LyViewManager viewManager = getViewManager();

                if (viewManager != null)
                    _internalView = viewManager.buildInternalView(this);
			}
			
			return _internalView;
		}
	}

	/**
	 * Reserved for internal use, do not use.
	 */
	public void setInternalView(LyViewComponent viewComponent)
	{
		if ((_internalView != null)
				&& (_internalView != viewComponent))
			throw new IllegalStateException("BF AppEngine : Internal view can't be set twice");

		_internalView = viewComponent;
	}

	/**
	 * Returns child component of index i.
	 *
	 * @param   i  index of component to get.
	 *
	 * @return  child component of index i.
	 */
	public LyComponent getChild(int i)
	{
		if (i < 0)
			return null;
		if (_children == null)
		    return null;
		
		synchronized (_children)
		{
		    if (i >= _children.size())
		        return null;
		    
		    return (LyComponent)_children.get(i);
		}
	}

	/**
	 * Returns child component count.
	 */
	public int getChildCount()
	{
		return ((_children == null) ? 0 : _children.size());
	}

	/**
	 * Get parent component.
	 *
	 * @return  parent component.
	 */
	public final LyComponent getParent()
	{
		return _parent;
	}

    /**
     * Gets model of this component.
     * @return Model of this component if component was built from template or cloneComponent.
     */
    public final LyComponent getModel()
    {
        return _model;
    }
    
	/**
	 * Returns the position and size of current component relatively to given component.
	 *
	 * @param   parentComponent  the component used to compute relative position of current
	 *                           component. If null, position of current component is computed with
	 *                           the top-left corner of the screen as the origin.
	 *
	 * @return  an array of 4 integers (x position, y position, width, height) representing the
	 *          position and size of current component.
	 */
	public int[] getBounds(LyComponent parentComponent)
	{
		if (parentComponent != null)
		{
			int[] result = parentComponent.getChildrenBoundsFromCache(this);

			if (result != null)
				return result;
		}

		LyViewComponent internal = getInternalView(false);

		if (internal instanceof LyGraphicalObject)
		{
			LyGraphicalObject object = (LyGraphicalObject)internal;

			return object.getBounds(parentComponent);
		}

		return null;
	}

	/**
	 * Returns the position and size of given sub component relatively to current component.
	 *
	 * @param   childrenComponent  the sub component whose relative position must be returned.
	 *
	 * @return  an array of 4 integers (x position, y position, width, height) representing the
	 *          position and size of given sub component.
	 */
	public int[] getChildrenBounds(LyComponent childrenComponent)
	{
		LyViewComponent internal = getInternalView(false);

		if (internal instanceof LyGraphicalObject)
		{
			LyGraphicalObject object = (LyGraphicalObject)internal;

			return object.getChildrenBounds(childrenComponent);
		}

		return null;
	}

	/**
	 * Returns the position and size of given sub component relatively to current component.
	 * Position is retrieved from internal cache.
	 *
	 * @param   childrenComponent  the sub component whose relative position must be returned.
	 *
	 * @return  an array of 4 integers (x position, y position, width, height) representing the
	 *          position and size of given sub component.
	 */
	private int[] getChildrenBoundsFromCache(LyComponent component)
	{
		if (_childCacheBounds != null)
			return _childCacheBounds.get(component);

		return null;
	}

	/**
	 * Registers in the internal cache position ans size of all sub components of current component.
	 */
	public void registerChildrenBounds()
	{
		// Retrieve all the components contained in this component.
		LyComponent		    component;
		Vector<LyComponent> components = new Vector<LyComponent>();

		getComponents(LyComponent.class, components);

		int   i, n = components.size();
		int[] bounds;

		if (_childCacheBounds == null)
			_childCacheBounds = new Hashtable<LyComponent, int[]>();
		else
			_childCacheBounds.clear();

		for (i = n - 1; i >= 0; i--)
		{
			component = components.elementAt(i);
			bounds = getChildrenBounds(component);

			if (bounds != null)
				_childCacheBounds.put(component, bounds);
		}
	}

	/**
	 * Clears the content of the internal cache containing sub components sizes and positions.
	 */
	public void clearChildrenBounds()
	{
		if (_childCacheBounds != null)
			_childCacheBounds.clear();
	}

	/**
	 * Returns the component at given position with an error margin.
	 *
	 * @param   position  the position of the component we want to retrieve.
	 * @param   hMargin   the horizontal margin used to retrieve component at given position.
	 * @param   vMargin   the vertical margin used to retrieve component at given position.
	 *
	 * @return  the component at given x and y position, or null if no component was found.
	 */
	public final LyComponent getComponentAt(int[] position, int hMargin, int vMargin)
	{
	    return getComponentAt(position, hMargin, vMargin, LyComponent.class);
	}
	
	/**
     * Returns the component at given position with an error margin.
     *
     * @param   position        the position of the component we want to retrieve.
     * @param   hMargin         the horizontal margin used to retrieve component at given position.
     * @param   vMargin         the vertical margin used to retrieve component at given position.
     * @param   componentClass  expected class
     * @return  the component at given x and y position, or null if no component was found.
     */
    public <T extends LyComponent> T getComponentAt(int[] position, int hMargin, int vMargin, Class<T> componentClass)
    {
		if ((position == null) || (position.length != 2))
			return null;

		// Retrieve all the components contained in this component.
		T             component;
		Vector<T>     components = new Vector<T>();

		getComponents(componentClass, components);

		int   i, n = components.size();
		int[] bounds;

		// Retrieve the first component at given position
		for (i = n - 1; i >= 0; i--)
		{
			component = components.elementAt(i);

			if ((_childCacheBounds != null)
			       && _childCacheBounds.containsKey(component))
				bounds = _childCacheBounds.get(component);
			else
				bounds = getChildrenBounds(component);

			if ((bounds != null)
					&& (position[0] >= (bounds[0] - hMargin))
					&& (position[0] <= (bounds[0] + bounds[2] + hMargin))
					&& (position[1] >= (bounds[1] - vMargin))
					&& (position[1] <= (bounds[1] + bounds[3] + vMargin)))
				return component;
		}

		return null;
	} // end method getComponentAt

	/**
	 * Add child component.
	 *
	 * @param  child  component to add.
	 */
	
	void addChildComponent(LyComponent child)
	{
		if (_children == null)
			_children = new ArrayList<LyComponent>(1);
        
        synchronized (_children)
        {
    		if (!(this instanceof LyColumn) // Optimization for charts
    				&& _children.contains(child))
    			return;
    
    		@SuppressWarnings("unchecked")
    		ArrayList<LyComponent> temp = (ArrayList<LyComponent>)_children;
    		
    		temp.add(child);
        }
	}

	/**
	 * Remove child component.
	 *
	 * @param  child  component to remove.
	 */
	void removeChildComponent(LyComponent child)
	{
		if (_children == null)
			return;
        
        synchronized (_children)
        {            
            if (_children.remove(child))
            	removeChildFromCache(child);
        }
	}

	/**
	 * Empty children cache.<br />
	 * This method clears the content of the children caches. It must be called only when caches
	 * must be cleared for a good reason because it cancels cache optimization for this component.
	 */
	public void resetChildrenCache()
	{
		if (_childCache != null)
		{
			_childCache.clear();
			_childCache = null;
		}

		if (_childCacheForView != null)
		{
			_childCacheForView.clear();
			_childCacheForView = null;
		}
	}

	/**
	 * Removes the given id from child cache.
	 *
	 * @param  id  id to remove from local cache.
	 */
	void removeChildFromCache(LyComponent child)
	{
		LyComponent parent = this;

		while (parent != null)
		{
			parent.resetChildrenCache();
			parent = parent.getParent();
		}
	}

	/**
	 * Add the given id / component from child cache.
	 *
	 * @param  child          id to add in local cache.
	 * @param  recurseToView  indicates if we store it temporary for top view
	 */
	protected void addChildInCache(LyComponent child, boolean recurseToView)
	{
		if (!(child instanceof LyGetIdInterface))
			return;

		String id = ((LyGetIdInterface)child).getId();

		if (id == null)
			return;

		if (recurseToView)
		{
			if (_childCacheForView == null)
				_childCacheForView = new Hashtable<String, LyComponent>(5);

			_childCacheForView.put(id, child);
		}
		else
		{
			if (_childCache == null)
				_childCache = new Hashtable<String, LyComponent>(5);

			_childCache.put(id, child);
		}
	}

	/**
	 * Add the given id / component from child cache.
	 *
	 * @param   id  to search in local cache.
	 *
	 * @return  child if found or null
	 */
	public LyComponent getChildFromCache(String id)
	{
		if ((_childCache == null) || (id == null))
			return null;

		return _childCache.get(id);
	}

	/**
	 * Set parent component.
	 *
	 * @param  parent  the new parent for this component.
	 */
	void setParent(LyComponent parent)
	{
		if (_parent == parent)
			return;

		if (_parent != null)
			_parent.removeChildComponent(this);

		_parent = parent;

		if (_parent != null)
		{
			_parent.addChildComponent(this);

			if (_childCacheForView != null)
			{
				// Transfert it
				LyComponent parentView = getParentView();
				boolean     recurseToView = false;

				if (parentView == null)
				{
					// Do our best
					parentView = _parent;

					while (parentView.getParent() != null)
						parentView = parentView.getParent();

					recurseToView = true;
				}

				Iterator<LyComponent> iter = _childCacheForView.values().iterator();

				while (iter.hasNext())
				{
					LyComponent comp = iter.next();

					parentView.addChildInCache(comp, recurseToView);
				}

				_childCacheForView = null; // Over
			}
		}
	} // end method setParent

	/**
	 * Add child component.
	 *
	 * @param  child  component to add.
	 */
	protected void addChild(LyComponent child)
	{
	}

	/**
	 * Retrieves controller id for the component. If the component has no controller, search
	 * recursively for the parent's controller.
	 *
	 * @return  the id of the root controller for this component.
	 */
	public String getRootController()
	{
		// If the component has a (not null) controller
		if (this instanceof LyGetControllerInterface)
		{
			String controller = ((LyGetControllerInterface)this).getController();

			if (controller != null)
				return controller;
		}

		// If controller not found, return parent controller
		if (_parent != null)
			return _parent.getRootController();

		// If parent is null, return null
		return null;
	}

	/**
	 * Write XML description corresponding to this component.
	 *
	 * @param  pw  Print writer where the description is written.
	 */
	public final void writeXml(PrintWriter pw)
	{
		writeXml(pw, true);
	}

	/**
	 * Write XML description corresponding to this component.
	 *
	 * @param  pw     Print writer where the description is written.
	 * @param  debug  True if debug mode (full dump includinf optional ids and references of
	 *                controllers).
	 */
	public abstract void writeXml(PrintWriter pw, boolean debug);

	/**
	 * Retrieve closest parent that implements the LySetVisibleInterface interface.
	 *
	 * @return  closest parent that implements the LySetVisibleInterface interface.
	 */
	public LySetVisibleInterface getSetVisibleParent()
	{
		LyComponent parent = getParent();

		if (parent == null)
			return null;

		Object internalParent = parent.getInternalView();

		if (internalParent instanceof LySetVisibleInterface)
			return (LySetVisibleInterface)internalParent;
		else
			return parent.getSetVisibleParent();
	}

	/**
	 * Retrieve closest frame or dialog that contains this component.
	 *
	 * @return  closest frame or dialog that contains this component.
	 */
	public LyView getParentView()
	{
		LyComponent parent = getParent();

		if (parent == null)
			return null;

		if (parent instanceof LyView)
			return (LyView)parent;
		else
			return parent.getParentView();
	}

	/**
	 * Retrieve closest area that contains this component.
	 *
	 * @return  closest area that contains this component.
	 */
	public LyArea getParentArea()
	{
		LyComponent parent = getParent();

		if (parent == null)
			return null;

		if (parent instanceof LyArea)
			return (LyArea)parent;
		else
			return parent.getParentArea();
	}

	/**
	 * Retrieve closest form that contains this component.
	 *
	 * @return  closest form that contains this component.
	 */
	public LyForm getParentForm()
	{
		LyComponent parent = getParent();

		if (parent == null)
			return null;

		if (parent instanceof LyForm)
			return (LyForm)parent;
		else
			return parent.getParentForm();
	}

	/**
	 * Retrieve closest tabs that contains this component.
	 *
	 * @return  closest tabs that contains this component.
	 */
	public final LyTabs getParentTabs()
	{
		return getParentTabs(null);
	}

	/**
	 * Retrieve closest tabs that contains this component. Don't search after a fixed limit.
	 *
	 * @param   topComponent  limit of the search
	 *
	 * @return  closest tabs that contains this component.
	 */
	public LyTabs getParentTabs(LyComponent topComponent)
	{
		LyComponent parent = getParent();

		if (parent == null)
			return null;

		if (parent == topComponent)
			return null;

		if (parent instanceof LyTabs)
			return (LyTabs)parent;
		else
			return parent.getParentTabs(topComponent);
	}

	/**
	 * Retrieve closest splitter that contains this component.
	 *
	 * @return  closest splitter that contains this component.
	 */
	public final LySplitter getParentSplitter()
	{
		return getParentSplitter(null);
	}

	/**
	 * Retrieve closest splitter that contains this component. Don't search after a fixed limit.
	 *
	 * @param   topComponent  limit of the search
	 *
	 * @return  closest splitter that contains this component.
	 */
	public LySplitter getParentSplitter(LyComponent topComponent)
	{
		LyComponent parent = getParent();

		if (parent == null)
			return null;

		if (parent == topComponent)
			return null;

		if (parent instanceof LySplitter)
			return (LySplitter)parent;
		else
			return parent.getParentSplitter(topComponent);
	}

	/**
	 * Retrieve closest tab that contains this component.
	 *
	 * @return  closest tab that contains this component.
	 */
	public final LyTab getParentTab()
	{
		return getParentTab(null);
	}

	/**
	 * Retrieve closest tab that contains this component. Don't search after a fixed limit.
	 *
	 * @param   topComponent  limit of the search
	 *
	 * @return  closest tab that contains this component.
	 */
	public LyTab getParentTab(LyComponent topComponent)
	{
		LyComponent parent = getParent();

		if (parent == null)
			return null;

		if (parent == topComponent)
			return null;

		if (parent instanceof LyTab)
			return (LyTab)parent;
		else
			return parent.getParentTab(topComponent);
	}

	/**
	 * Retrieve closest formField id that contains this component.
	 *
	 * @return  closest formField id that contains this component, null otherwise.
	 */
	public String getParentFormFieldId()
	{
		LyComponent parent = getParent();

		if (parent == null)
			return null;

		if (parent instanceof LyFormField)
			return ((LyFormField)parent).getId();
		else
			return parent.getParentFormFieldId();
	}

	/**
	 * Reset this component.
	 */
	public void reset()
	{
		synchronized (getLock())
		{
		    if (_children != null)
		    {
    		    synchronized (_children)
    		    {
        			// Recurse into children
        			LyComponent child;
        			int		    i, n = _children.size();
        
        			for (i = n - 1; i >= 0; i--)
        			{
        				child = (LyComponent)_children.get(i);
        				if (child != null)
        					child.reset();
        			}
    		    }
		    }

			if (_internalView instanceof LyFreeInterface)
				((LyFreeInterface)_internalView).free();

			// Clean up all
			_internalView = null;
		}
	}

	/**
	 * Free this component.
	 */
	public void free()
	{
		synchronized (getLock())
		{
			if (_freed || _freeing)
				return;

			_freeing = true;

			if (_children != null)
			{
			    synchronized (_children)
			    {
		             // Recurse into children
	                LyComponent child;
	                int         i, n = _children.size();
	    
	                for (i = n - 1; i >= 0; i--)
	                {
	                    child = (LyComponent)_children.get(i);
	                    if (child != null)
	                        child.free();
	                }
			    }
			}

			if (_internalView instanceof LyFreeInterface)
				((LyFreeInterface)_internalView).free();

            // Unregister this component as definition if needed
            LyViewManager   viewManager = getViewManager();
            
            if (viewManager != null)
                viewManager.unregisterDefinition(this);
            
			// Clean up all
			_internalView = null;
			_parent = null;
			_childCache = null;
			_childCacheForView = null;
			_freeing = false;
			_freed = true;
            _datas = null;
            _model = null;

			if (LyEnvironment.TEST_MODE)
				LyComponent.removeComponent(this);

			_internalId = null;
			_children = null;
		}
	} // end method free

	/**
	 * Checks if this component is freeing.
	 *
	 * @return  True id this component is freeing, false otherwise.
	 */
	public final boolean isFreeing()
	{
		return _freeing
			|| ((_parent != null) && _parent.isFreeing());
	}

	/**
	 * Checks if this component was freed.
	 *
	 * @return  True id this component was freed, false otherwise.
	 */
	public final boolean isFreed()
	{
		return _freed;
	}

	/**
	 * Shows or hides the component/node.
	 *
	 * @param  status  new visibility status
	 *
	 * @see    LySetVisibleInterface#setVisible
	 */
	public void setVisible(boolean status)
	{
		if (LyEnvironment.TEST_MODE && status)
			LyComponent.addComponent(this);

		if (_internalView instanceof LySetVisibleInterface)
			((LySetVisibleInterface)_internalView).setVisible(status);

		if (_children == null)
		    return;
		
		synchronized (_children)
		{
    		// Recurse into children
    		LyComponent child;
    		int		    i, n = _children.size();
    
    		for (i = 0; i < n; i++)
    		{
    			child = (LyComponent)_children.get(i);
    			if (child != null)
    				child.setVisible(status);
    		}
		}
	}

	/**
	 * Freeze or unfreeze current view.
	 *
	 * @param  status  indicates whether current view must be frozen or not.
	 */
	public void freeze(boolean status)
	{
		// Retrieve the list of components that must be freezed and keep their status
		// in the _componentsFreezeState list.
		Vector<LyFormField> formFields = new Vector<LyFormField>();
		Vector<LyCommand>   commands = new Vector<LyCommand>();
		Vector<LyHRef>	    hrefs = new Vector<LyHRef>();
		Vector<LyTool>	    tools = new Vector<LyTool>();
		Vector<LyCaption>   captions = new Vector<LyCaption>();
		Vector<LyTable>     tables = new Vector<LyTable>();

		getComponents(LyFormField.class, formFields);
		getComponents(LyCommand.class, commands);
		getComponents(LyHRef.class, hrefs);
		getComponents(LyTool.class, tools);
		getComponents(LyCaption.class, captions);
		getComponents(LyTable.class, tables);

		int			    i, n;
		LyFormField     formField;
		LyCommand	    command;
		LyHRef		    href;
		LyTool		    tool;
		LyCaption	    caption;
		LyTable		    table;
		LyViewComponent internal;

		// Freeze or unfreeze forms
		n = formFields.size();
		for (i = 0; i < n; i++)
		{
			formField = formFields.elementAt(i);
			internal = formField.getInternalView(false);
			if (internal instanceof LyGraphicalFormField)
				if (status)
					((LyGraphicalFormField)internal).setEditable(false);
				else
					((LyGraphicalFormField)internal).setEditable(formField.isEditable());
		}

		// Freeze or unfreeze commands
		n = commands.size();
		for (i = 0; i < n; i++)
		{
			command = commands.elementAt(i);
			internal = command.getInternalView(false);
			if (internal instanceof LyGraphicalCommand)
				if (status)
					((LyGraphicalCommand)internal).setEnabled(false);
				else
					((LyGraphicalCommand)internal).setEnabled(command.isEnabled());
		}

		// Freeze or unfreeze hrefs
		n = hrefs.size();
		for (i = 0; i < n; i++)
		{
			href = hrefs.elementAt(i);
			internal = href.getInternalView(false);
			if (internal instanceof LyGraphicalCommand)
				if (status)
					((LySetSensitiveInterface)internal).setSensitive(false);
				else
					((LySetSensitiveInterface)internal).setSensitive(href.isSensitive());
		}

		// Freeze or unfreeze tools
		n = tools.size();
		for (i = 0; i < n; i++)
		{
			tool = tools.elementAt(i);
			internal = tool.getInternalView(false);
			if (internal instanceof LyGraphicalTool)
				if (status)
					((LyGraphicalTool)internal).setEnabled(false);
				else
					((LyGraphicalTool)internal).setEnabled(tool.isEnabled());
		}

		// Freeze or unfreeze captions
		n = captions.size();
		for (i = 0; i < n; i++)
		{
			caption = captions.elementAt(i);
			internal = caption.getInternalView(false);
			if (internal instanceof LyGraphicalCaption)
				if (status)
					((LyGraphicalCaption)internal).setEnabled(false);
				else
					((LyGraphicalCaption)internal).setEnabled(caption.isEnabled());
		}

		// Freeze or unfreeze tables
		n = tables.size();
		for (i = 0; i < n; i++)
		{
			table = tables.elementAt(i);
			internal = table.getInternalView(false);
			if (internal instanceof LyGraphicalTable)
				if (status)
					((LyGraphicalTable)internal).setEnabled(false);
				else
					((LyGraphicalTable)internal).setEnabled(table.isEnabled());
		}
	} // end method freeze

	/**
	 * Get a lock to prevent concurrent access Returns the one provided by the internal view
	 * implementation, otherwise returns this.
	 *
	 * @return  a valid lock for synchronized(), may not return null
	 */
	public Object getLock()
	{
		if (_internalView instanceof LyGetLockInterface)
			return ((LyGetLockInterface)_internalView).getLock();
		else
			return this;
	}
    
    /**
     * Sets datas.
     *
     * @param  datas  Datas of the component.
     */
    public void setDatas(Hashtable<String, String> datas)
    {
        if (_datas == null)
            _datas = datas;
        else
            _datas.putAll(datas);
    }

    /**
     * Get data corresponding to the given data name.
     *
     * @param   name  Name of the data.
     *
     * @return  Value of the data, null if not found.
     */
    public final String getData(String name)
    {
        return (_datas == null? null : _datas.get(name));
    }

    /**
     * Gets datas.
     *
     * @return  Datas of the component.
     */
    public final Hashtable<String, String> getDatas()
    {
        return _datas;
    }
    
    /**
     * Add data to this component.
     *
     * @param  name   Name of the data.
     * @param  value  Value of the data.
     */
    public void addData(String name, String value)
    {
        if (_datas == null)
            _datas = new Hashtable<String, String>();

        _datas.put(name, value);
    }
    
    /**
     * Write datas.
     */
    protected void writeDatas(PrintWriter pw)
    {
        if (_datas == null)
            return;
        
        Enumeration<String>     keys = _datas.keys();
        ArrayList<String>       names = new ArrayList<String>(_datas.size());
        String                  key, value;
        int                     i, n = 0;
        
        // Sort keys to have a constant order
        while (keys.hasMoreElements())
        {
            key = keys.nextElement();
            for (i = 0; (i < n) && (key.compareTo(names.get(i)) > 0); i++);
            names.add(i, key);
            n++;
        }
            
        // Write datas
        for (String name : names)
        {
            value = getData(name);
            pw.println("<DATA name=\"" + LyString.getXmlString(name) + "\" value=\"" + LyString.getXmlString(value) + "\"/>");
        }
    }
    
    /**
     * Add a function to the list of functions that must be called for a given event type.
     * For example calling this function with the <code>EventType.FOCUS_LOST</code> and the
     * <code>"CheckValue()"</code> parameters with the web viewer will call the CheckValue()
     * Javascript function when focus will be lost on current component.
     * 
     * @param eventType	the type of event that will call the function
     * @param function  the function to be called by the event
     */
    public void addEventFunction(EventType eventType, String function)
    {
    	if (_eventFunctions == null)
    		_eventFunctions = new HashMap<EventType, Vector<String>>();
    	
    	Vector<String>	functions = _eventFunctions.get(eventType);
    	
    	if (functions == null)
    	{
    		functions = new Vector<String>();
    		functions.add(EVENT_DEFAULT_FUNCTION);
    	}
    	
    	functions.add(function);
    	_eventFunctions.put(eventType, functions);
    }
    
    /**
     * Remove all the functionc to the list of functions that must be called for a given event type,
     * even the generic ones.
     * 
     * @param eventType	the type of event that will call the function
     */
    public void resetEventFunctions(EventType eventType)
    {
    	if (_eventFunctions == null)
    		_eventFunctions = new HashMap<EventType, Vector<String>>();
    	
    	_eventFunctions.put(eventType, new Vector<String>());
    }
    
    /**
     * Indicates whether the default Leonardi function must be called for the given
     * type of event.
     * 
     * @param eventType the type of event (see EventType enum)
     * @return true if the default function must be called, false otherwise
     */
    public boolean useEventDefaultFunction(EventType eventType)
    {
    	if (_eventFunctions == null)
    		return true;
    	
    	Vector<String>	functions = _eventFunctions.get(eventType);
    	
    	return ((functions == null) || functions.contains(EVENT_DEFAULT_FUNCTION));
    }
    
    /**
     * Return the list of functions that must be called for a given event type.
     * 
     * @param eventType the type of event that will call the function
     * @param sep       the separator used to concatenate the function names
     * @return the list of functions to be called separated by the <code>sep</code> character or an
     * empty string if no additional function has been set for this type of event.
     */
    public String getEventFunctions(EventType eventType, char sep)
    {
    	if (_eventFunctions == null)
    		return "";
    	
    	Vector<String>	functions = _eventFunctions.get(eventType);
    	
    	if ((functions == null) || functions.isEmpty())
    		return "";
    	
    	StringBuffer	result = new StringBuffer();
    	
    	for (int i = 0; i < functions.size(); i++)
    	{
    		if (functions.elementAt(i) != EVENT_DEFAULT_FUNCTION)
    		{
    			result.append(functions.elementAt(i));
				result.append(sep);
    		}
    	}
    	
    	return result.toString();
    }
} // end class LyComponent
