package leon.view;

import java.util.*;

import leon.app.*;
import leon.app.behavior.*;
import leon.control.*;
import leon.data.*;
import leon.misc.*;

/**
 * 
 * @author smaupin
 *
 */
public class LyConversation
{
	
	/** id of the conversation tab */
	private String _tabId;
	
	/** session list. */
	private Map<String, LySession>	_sessionList;

	
	private Map<String, LyAreaController> _controllerList;
	
	/** default constructor. */
	public LyConversation()
	{
		_sessionList = new HashMap<String, LySession>();
		_controllerList = new HashMap<String, LyAreaController>();
	}

	public String getTabId()
	{
		return _tabId;
	}

	public void setTabId(String tabId)
	{
		this._tabId = tabId;
	}
	
	/**
	 * Add session to conversation.
	 * @param session The session to add
	 */
	public void addSession(LySession session, LyAreaController controller)
	{
		if (session != null && session.getUser() != null)
		{
			_sessionList.put(session.getId(), session);
			_controllerList.put(session.getId(), controller);
		}
	}
	
	/**
	 * Remove session.
	 * @param session session to remove.
	 */
	public void removeSession(LySession session)
	{
		if (session != null)
		{
			_sessionList.remove(session.getId());
			_controllerList.remove(session.getId());
		}
	}
	
	/**
	 * Get controller with session id.
	 * @param sessionId sessionId.
	 * @return the controller.
	 */
	public LyAreaController getController(String sessionId)
	{
		return _controllerList.get(sessionId);
	}
	
	
	/**
	 * Get list of controllers.
	 * @return map map of controller/session
	 */
	public  Map<String, LyAreaController> getControllers()
	{
		return _controllerList;
	}
	
	/**
	 * Set session list.
	 * @param sessionList session List
	 */
	public void setSessions(Map<String, LySession> sessionList)
	{
		this._sessionList = sessionList;
	}

	/**
	 * Get Sessions for conversation.
	 * @return sessions.
	 */
	public Map<String, LySession> getSessions()
	{
		return _sessionList;
	}
	
	/**
	 * Get Session from userId.
	 * @param userId The user id.
	 * @return The user session
	 */
	public LySession getTargetUserSession(String sessionId)
	{
		return _sessionList.get(sessionId);
	}

	/**
	 * Send message to all users in conversation.
	 * @param inputValue Message
	 * @param sender Sender of message
	 */
	public void sendMessage(String inputValue, LySession senderSession, boolean serverMessage, boolean key)
	{
		if (senderSession != null && _sessionList.get(senderSession.getId()) != null)
		{
			for (LySession session : _sessionList.values())
			{
				LyController chatController = _controllerList.get(session.getId());
				if (key)
				{
					LyEnvironment targetEnv = session.getEnvironment();
					inputValue = targetEnv.translate(inputValue);
				}
				if (chatController != null)
				{
					LyChatBehavior behavior = (LyChatBehavior)chatController.getBehavior();
					if (behavior != null)
					{
						behavior.writeMessage(inputValue, senderSession.getUser(), serverMessage);
					}
				}
			}
		}
	}

	/**
	 * Sends an object to all users in conversation.
	 * @param sentObject object sent to the other users
	 * @param senderSession Sender of message
	 */
	public void sendObject(LyObject sentObject, LySession senderSession)
	{
		if (senderSession != null && _sessionList.get(senderSession.getId()) != null)
		{
			for (LySession session : _sessionList.values())
			{
				LyController chatController = _controllerList.get(session.getId());

				if (chatController != null)
				{
					LyChatBehavior behavior = (LyChatBehavior)chatController.getBehavior();
					
					if (behavior != null)
					{
						String	message;
						String	objectId = sentObject.getObjectId().getFullString();
						String	link = "<a class=hrefLabel style=color:blue" +
							" href=javascript:invokeHrefLink(\"" + chatController.getId() + "\",\"\",\"" +
							LyString.replace(objectId, "\\", "\\\\") + "\");>" + sentObject.getName() + "</a>";

						if (chatController.getSession() == senderSession)
							message = chatController.getEnvironment().getMessage("LY_SENT_OBJECT_LINK", link);
						else
							message = chatController.getEnvironment().getMessage("LY_SEND_OBJECT_LINK", senderSession.getUser().getName(), link);

						behavior.writeMessage(message, null, true);
					}
				}
			}
		}
	}
	
	/**
	 * Checks whether an object is visible by all the users of the conversation.
	 * 
	 * @param sentObject the object whose visibility is checked
	 * @return the list of users that may not access the object
	 */
	public Vector<LyObject> isObjectVisible(LyObject sentObject, LySession senderSession)
	{
		Vector<LyObject>	result = null;
		
		if (senderSession != null && _sessionList.get(senderSession.getId()) != null)
		{
			for (LySession session : _sessionList.values())
			{
				if (!session.isObjectVisible(sentObject))
				{
					if (result == null)
						result = new Vector<LyObject>();
					
					result.add(session.getUser());
				}
			}
		}
		
		return result;
	}

	/**
	 * Update actors labels for each users in conversation.
	 */
	public void updateActorLabels()
	{
		for (LySession session : _sessionList.values())
		{
			LyController chatController = _controllerList.get(session.getId());
			if (chatController != null)
			{
				LyChatBehavior behavior = (LyChatBehavior)chatController.getBehavior();
				if (behavior != null)
				{
					behavior.updateActorsLabel();
				}
			}
		}
		
	}
}

