/*
 * LyDialogBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyDialogBase class is a generated Java implementation that corresponds to the element 
 * DIALOG of the descriptive grammar of the views (views.dtd).
 * <B>&lt;DIALOG&gt; : Dialog Window</B>
 * A dialog window is, generally, linked to a main window of an application.
 * Unlike the main window, it contains neither a menu, nor a message bar.
 * Its display can be modal (blocking) or not. This type of window is generally 
 * used to allow a user to interact with an application: input request, 
 * presentation of the results, warning messages, questions, etc.
 * @see leon.view.LyView
 * @see leon.view.LyLyDialog
 * See also tag <A href="../../conf/dtd/view/DIALOG.html">DIALOG</A> of grammar leon_views.dtd.
 */
public abstract class LyDialogBase extends LyView
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface,
				LyGetControllerInterface,
				LySetControllerInterface
{
	public static final String			__VERSION = "$Id$";


	/* Attributes */

	/**
	 * Identifier of the dialog window
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyDialog			_template;

	/**
	 * Title of the dialog window
	 */
	protected String			_label;

	/**
	 * Reference of the controller that manages this dialog window (i.e. handles user 
	 * demands and refresh content).
	 */
	protected String			_controller;

	/**
	 * Internal identifier of the application of the parent window of the current 
	 * dialog window.
	 * The parent window can be a main window or another dialog window.
	 */
	protected String			_parentId;

	/**
	 * Initial width of the dialog window.
	 */
	protected int				_width;

	/**
	 * Initial height of the dialog window.
	 */
	protected int				_height;

	/**
	 * Indicates whether the dialog window can be resized or not.
	 */
	protected boolean			_resizable;

	/**
	 * Minimum width of the dialog window.
	 */
	protected int				_minWidth;

	/**
	 * Minimum height of the dialog window.
	 */
	protected int				_minHeight;

	/**
	 * Initial position in abscissa of the dialog window.
	 */
	protected int				_x;

	/**
	 * Initial position in ordinate of the dialog window.
	 */
	protected int				_y;

	/**
	 * Indicates whether the dialog window is modal or not.
	 */
	protected boolean			_modal;

	/**
	 * Indicates whether the dialog window is decorated or not.
	 */
	protected boolean			_decorated;

	/**
	 * Identifier of the display zone for the dialog window.
	 */
	protected String			_target;


	/* Sub components */

	/**
	 * Main area contained in the dialog window.
	 */
	protected LyArea			_area;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_controller = null;
		_parentId = null;
		_width = getIntValue("DIALOG.width", -1);
		_height = getIntValue("DIALOG.height", -1);
		_resizable = getBooleanValue("DIALOG.resizable", false);
		_minWidth = getIntValue("DIALOG.minWidth", -1);
		_minHeight = getIntValue("DIALOG.minHeight", -1);
		_x = getIntValue("DIALOG.x", 0);
		_y = getIntValue("DIALOG.y", 0);
		_modal = getBooleanValue("DIALOG.modal", false);
		_decorated = getBooleanValue("DIALOG.decorated", true);
		_target = null;

		// Initialize sub components
		_area = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyDialog	model = (LyDialog)component;

		_model = component;
		_label = model._label;
		_controller = model._controller;
		_parentId = model._parentId;
		_width = model._width;
		_height = model._height;
		_resizable = model._resizable;
		_minWidth = model._minWidth;
		_minHeight = model._minHeight;
		_x = model._x;
		_y = model._y;
		_modal = model._modal;
		_decorated = model._decorated;
		_target = model._target;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		if (model._area != null)
			addChild(model._area.cloneComponent());
		if (model._datas != null)
			_datas = new Hashtable<String, String>(model._datas);
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	protected LyDialogBase()
	{
		_id = getInternalId();
	}

	/**
	 * Constructor with all required attributes.
	 * @param id Identifier of the dialog window
	 */
	public LyDialogBase(String id)
	{
		// Fill up required attributes with parameters
		_id = (id != null ? id.intern() : null);
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyDialogBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyDialog)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("controller"))
			_controller = node.getString("controller");
		if (node.isSpecified("parentId"))
			_parentId = node.getString("parentId");
		if (node.isSpecified("width"))
			_width = node.getInt("width");
		if (node.isSpecified("height"))
			_height = node.getInt("height");
		if (node.isSpecified("resizable"))
			_resizable = node.getBoolean("resizable");
		if (node.isSpecified("minWidth"))
			_minWidth = node.getInt("minWidth");
		if (node.isSpecified("minHeight"))
			_minHeight = node.getInt("minHeight");
		if (node.isSpecified("x"))
			_x = node.getInt("x");
		if (node.isSpecified("y"))
			_y = node.getInt("y");
		if (node.isSpecified("modal"))
			_modal = node.getBoolean("modal");
		if (node.isSpecified("decorated"))
			_decorated = node.getBoolean("decorated");
		if (node.isSpecified("target"))
			_target = node.getString("target");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyArea)
			setArea((LyArea)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyDialog clone = new LyDialog();

		clone.initFromModel(this);
		clone._template = (LyDialog)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyDialog getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of controller.
	 * @return value of controller.
	 */
	public String getController()
	{
		return _controller;
	}

	/**
	 * Gets value of parentId.
	 * @return value of parentId.
	 */
	public String getParentId()
	{
		return _parentId;
	}

	/**
	 * Gets value of width.
	 * @return value of width.
	 */
	public int getWidth()
	{
		return _width;
	}

	/**
	 * Gets value of height.
	 * @return value of height.
	 */
	public int getHeight()
	{
		return _height;
	}

	/**
	 * Checks if value of attribute resizable is set.
	 * @return true if resizable is set, false otherwise.
	 */
	public boolean isResizable()
	{
		return _resizable;
	}

	/**
	 * Gets value of minWidth.
	 * @return value of minWidth.
	 */
	public int getMinWidth()
	{
		return _minWidth;
	}

	/**
	 * Gets value of minHeight.
	 * @return value of minHeight.
	 */
	public int getMinHeight()
	{
		return _minHeight;
	}

	/**
	 * Gets value of x.
	 * @return value of x.
	 */
	public int getX()
	{
		return _x;
	}

	/**
	 * Gets value of y.
	 * @return value of y.
	 */
	public int getY()
	{
		return _y;
	}

	/**
	 * Checks if value of attribute modal is set.
	 * @return true if modal is set, false otherwise.
	 */
	public boolean isModal()
	{
		return _modal;
	}

	/**
	 * Checks if value of attribute decorated is set.
	 * @return true if decorated is set, false otherwise.
	 */
	public boolean isDecorated()
	{
		return _decorated;
	}

	/**
	 * Gets value of target.
	 * @return value of target.
	 */
	public String getTarget()
	{
		return _target;
	}

	/**
	 * Get sub-component Area.
	 * @return sub-component Area.
	 */
	public LyArea getArea()
	{
		return _area;
	}


	/* Modifiers */

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyDialog template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView != null)
			((LyGraphicalDialog)_internalView).setLabel(label);
	}

	/**
	 * Initializes attribute controller to the given value.
	 * @param controller New value of controller.
	 */
	public void setController(String controller)
	{
		if ((controller == _controller)
			|| ((controller != null)
				&& (_controller != null)
				&& controller.equals(_controller)))
			return;

		_controller = controller;
	}

	/**
	 * Initializes attribute parentId to the given value.
	 * @param parentId New value of parentId.
	 */
	public void setParentId(String parentId)
	{
		if ((parentId == _parentId)
			|| ((parentId != null)
				&& (_parentId != null)
				&& parentId.equals(_parentId)))
			return;

		_parentId = parentId;
	}

	/**
	 * Initializes attribute width to the given value.
	 * @param width New value of width.
	 */
	public void setWidth(int width)
	{
		if (width == _width)
			return;
		_width = width;
	}

	/**
	 * Initializes attribute height to the given value.
	 * @param height New value of height.
	 */
	public void setHeight(int height)
	{
		if (height == _height)
			return;
		_height = height;
	}

	/**
	 * Initializes attribute resizable to the given value.
	 * @param resizable Boolean indicating if attribute resizable is set or unset.
	 */
	public void setResizable(boolean resizable)
	{
		if (resizable == _resizable)
			return;
		_resizable = resizable;
	}

	/**
	 * Initializes attribute minWidth to the given value.
	 * @param minWidth New value of minWidth.
	 */
	public void setMinWidth(int minWidth)
	{
		if (minWidth == _minWidth)
			return;
		_minWidth = minWidth;
	}

	/**
	 * Initializes attribute minHeight to the given value.
	 * @param minHeight New value of minHeight.
	 */
	public void setMinHeight(int minHeight)
	{
		if (minHeight == _minHeight)
			return;
		_minHeight = minHeight;
	}

	/**
	 * Initializes attribute x to the given value.
	 * @param x New value of x.
	 */
	public void setX(int x)
	{
		if (x == _x)
			return;
		_x = x;
	}

	/**
	 * Initializes attribute y to the given value.
	 * @param y New value of y.
	 */
	public void setY(int y)
	{
		if (y == _y)
			return;
		_y = y;
	}

	/**
	 * Initializes attribute modal to the given value.
	 * @param modal Boolean indicating if attribute modal is set or unset.
	 */
	public void setModal(boolean modal)
	{
		if (modal == _modal)
			return;
		_modal = modal;
	}

	/**
	 * Initializes attribute decorated to the given value.
	 * @param decorated Boolean indicating if attribute decorated is set or unset.
	 */
	public void setDecorated(boolean decorated)
	{
		if (decorated == _decorated)
			return;
		_decorated = decorated;
	}

	/**
	 * Initializes attribute target to the given value.
	 * @param target New value of target.
	 */
	public void setTarget(String target)
	{
		if ((target == _target)
			|| ((target != null)
				&& (_target != null)
				&& target.equals(_target)))
			return;

		_target = target;
	}

	/**
	 * Sets sub-component area.
	 * @param area New value of area
	 */
	public void setArea(LyArea area)
	{
		if (_area != null)
			_area.setParent(null);
		_area = area;
		if (area != null)
			area.setParent(this);
		if (_internalView != null)
			((LyGraphicalDialog)_internalView).setArea(area);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<DIALOG");
		pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlString(_controller) + "'");
		if (((_template == null) && (_parentId != null))
				|| ((_template != null) && (_parentId != _template._parentId)))
			pw.print(" parentId='" + LyString.getXmlString(_parentId) + "'");
		if (((_template == null) && (_width != getIntValue("DIALOG.width", -1)))
				|| ((_template != null) && (_width != _template._width)))
			pw.print(" width='" + _width + "'");
		if (((_template == null) && (_height != getIntValue("DIALOG.height", -1)))
				|| ((_template != null) && (_height != _template._height)))
			pw.print(" height='" + _height + "'");
		if (((_template == null) && _resizable)
				|| ((_template != null) && (_resizable != _template._resizable)))
			pw.print(" resizable='" + _resizable + "'");
		if (((_template == null) && (_minWidth != getIntValue("DIALOG.minWidth", -1)))
				|| ((_template != null) && (_minWidth != _template._minWidth)))
			pw.print(" minWidth='" + _minWidth + "'");
		if (((_template == null) && (_minHeight != getIntValue("DIALOG.minHeight", -1)))
				|| ((_template != null) && (_minHeight != _template._minHeight)))
			pw.print(" minHeight='" + _minHeight + "'");
		if (((_template == null) && (_x != getIntValue("DIALOG.x", 0)))
				|| ((_template != null) && (_x != _template._x)))
			pw.print(" x='" + _x + "'");
		if (((_template == null) && (_y != getIntValue("DIALOG.y", 0)))
				|| ((_template != null) && (_y != _template._y)))
			pw.print(" y='" + _y + "'");
		if (((_template == null) && _modal)
				|| ((_template != null) && (_modal != _template._modal)))
			pw.print(" modal='" + _modal + "'");
		if (((_template == null) && !_decorated)
				|| ((_template != null) && (_decorated != _template._decorated)))
			pw.print(" decorated='" + _decorated + "'");
		if (((_template == null) && (_target != null))
				|| ((_template != null) && (_target != _template._target)))
			pw.print(" target='" + LyString.getXmlString(_target) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		if (_area != null)
			_area.writeXml(pw, debug);

		writeDatas(pw);

		pw.println("</DIALOG>");
	}
}
