/*
 * LyFieldItem.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.view;

import leon.misc.*;
import leon.view.beaninterface.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;

import java.util.*;


/**
 * This class implements the generic behavior of the following classes : LyText, LyNumber, LyDate,
 * LyChoice, LyOption, LyRelation, LyFile, LyFont, LyColor, LyForm.
 *
 * @see      LyText
 * @see      LyNumber
 * @see      LyDate
 * @see      LyChoice
 * @see      LyOption
 * @see      LyRelation
 * @see      LyFile
 * @see      LyForm
 * @version  $Id: LyFieldItem.java 28997 2010-12-23 10:17:15Z bmichel $
 * @author   Lyria
 */
public abstract class LyFieldItem extends LyComponent
	implements LyUpdateViewContentInterface
{
	public static final String __VERSION = "$Id: LyFieldItem.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** Bean component. */
	protected LyFieldBeanInterface _beanComponent;

	/** Field value represented in the field item. */
	protected Object _fieldValue;

	/**
	 * Default constructor. Initialization is done with the init method.
	 */
	public LyFieldItem()
	{
		super();

		_beanComponent = null;
		_fieldValue = null;
	}

	/**
	 * Sets attribute label to the given value.
	 *
	 * @param  label  new value of label.
	 */
	public void setLabel(String label)
	{
	}

	/**
	 * Sets attribute bean to the given value.
	 *
	 * @param  bean  new value of bean.
	 */
	public void setBean(String bean)
	{
		// No bean by default
	}

	/**
	 * Sets bean parameters.
	 *
	 * @param  datas  Datas of the bean.
     * @deprecated Use LyComponent.setDatas.
	 */
	public void setBeanDatas(Hashtable<String, String> datas)
	{
		setDatas(datas);
	}

	/**
	 * Get data corresponding to the given data name.
	 *
	 * @param   name  Name of the data.
	 *
	 * @return  Value of the data, null if not found.
     * @deprecated Use LyComponent.getData.
	 */
	public String getBeanData(String name)
	{
		return getData(name);
	}

	/**
	 * Add data for the associated bean.
	 *
	 * @param  name   Name of the data.
	 * @param  value  Value of the data.
     * @deprecated Use LyComponent.addData.
	 */
	public void addBeanData(String name, String value)
	{
		addData(name, value);
	}

	/**
	 * Sets the field value presented in this field item.
	 *
	 * @param  value  field value.
	 */
	public void setFieldValue(Object value)
	{
		_fieldValue = value;
	}

	/**
	 * Gets the field value presented in this field item.
	 *
	 * @param  value  field value.
	 */
	public Object getFieldValue()
	{
		return _fieldValue;
	}

	/**
	 * Returns the name of the bean associated to this field item.
	 */
	public String getBean()
	{
		// No bean by default
		return null;
	}

	/**
	 * Gets bean parameters.
	 *
	 * @return  Datas of the bean.
     * @deprecated Use LyComponent.getDatas.
	 */
	public Hashtable<String, String> getBeanDatas()
	{
		return getDatas();
	}

	/**
	 * Gets bean component.
	 *
	 * @return  The bean component if this field item has one, null otherwise.
	 */
	public LyFieldBeanInterface getBeanComponent()
	{
		return _beanComponent;
	}

	/**
	 * Sets bean component.
	 *
	 * @param  beanComponent  The bean component if this field item has one, null otherwise.
	 */
	public void setBeanComponent(LyFieldBeanInterface beanComponent)
	{
		if (_beanComponent == null)
			_beanComponent = beanComponent;
	}

	/**
	 * Builds bean instance for this field item corresponding to bean name.
	 *
	 * @return  The bean component if this field item has one, null otherwise.
	 */
	
	public LyViewComponent buildBean()
	{
		String javaName = getBean();

		if (javaName == null)
			return null;

		LyViewComponent beanInstance = null;
		LyEnvironment   environment = getEnvironment();
		Class<?>	    beanClass;
		ClassLoader     classLoader = null;

		if (environment != null)
			classLoader = environment.getClassLoader();

		try
		{
			if (classLoader != null)
				beanClass = classLoader.loadClass(javaName);
			else
				beanClass = Class.forName(javaName);

			beanInstance = (LyViewComponent)beanClass.newInstance();

			if ((_datas != null)
					&& (beanInstance instanceof LyFieldBeanDatasInterface))
				((LyFieldBeanDatasInterface)beanInstance).setDatas(_datas);

			if (beanInstance instanceof LyFieldBeanInterface)
			{
				_beanComponent = (LyFieldBeanInterface)beanInstance;
				_beanComponent.setViewComponent(this);
			}
		}
		catch (Exception e)
		{
			// Dump stack
			getEnvironment().writeException(e);
		}

		return beanInstance;
	} // end method buildBean

	/**
	 * Returns child component count.
	 */
	public int getChildCount()
	{
		if ((_beanComponent != null)
				&& (_beanComponent instanceof LyGetLyComponentInterface))
			return (super.getChildCount() + 1);
		else
			return super.getChildCount();
	}

	/**
	 * Returns child component of index i.
	 *
	 * @param   i  index of component to get.
	 *
	 * @return  child component of index i.
	 */
	public LyComponent getChild(int i)
	{
		if ((i == super.getChildCount())
				&& (_beanComponent != null)
				&& (_beanComponent instanceof LyGetLyComponentInterface))
			return ((LyGetLyComponentInterface)_beanComponent).getLyComponent();

		return super.getChild(i);
	}

	/**
	 * Default accessor to the value.
	 */
	public String getValue()
	{
		if (_internalView instanceof LyGetValueInterface)
			return ((LyGetValueInterface)_internalView).getValue();

		if (this instanceof LyGetLabelInterface)
			return ((LyGetLabelInterface)this).getLabel();

		// Should be overriden
		return null;
	}

	/**
	 * Validate input done on this field item.
	 */
	public short validateInput()
	{
		if (_internalView instanceof LyValidateInterface)
			return ((LyValidateInterface)_internalView).validateInput();

		return LyValidateInterface.STATUS_OK;
	}

	/**
	 * Returns the error message associated to this field item.
	 *
	 * @return  an error message if the value of the field item is not correct.
	 */
	public String getErrorMessage()
	{
		if (_internalView instanceof LyGetErrorMessageInterface)
			return ((LyGetErrorMessageInterface)_internalView).getErrorMessage();

		return null;
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.setinterface.LyUpdateViewContentInterface#updateViewContent()
	 */
	public void updateViewContent()
	{
		if (_internalView instanceof LyUpdateViewContentInterface)
			((LyUpdateViewContentInterface)_internalView).updateViewContent();
	}

	/**
	 * Free this field item.
	 */
	public void free()
	{
		if ((_beanComponent != null)
				&& (_beanComponent instanceof LyFreeInterface))
			((LyFreeInterface)_beanComponent).free();

		super.free();
	}
} // end class LyFieldItem
