/*
 * LyFile.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;
import java.io.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyFile class is a generated Java implementation that corresponds to the element 
 * FILE of the descriptive grammar of the views (views.dtd).
 * <B>&lt;FILE&gt; : Input Field of File Type</B>
 * An input field of File type allows displaying a graphical component that allows 
 * selecting a file. The type of file can be specified (text, image, audio, video).
 * According to the form in which this field is present, the graphical component 
 * allows to choose the location of the file and/or to view the access path to the 
 * file.
 * @see leon.view.LyFieldItem
 * See also tag <A href="../../conf/dtd/view/FILE.html">FILE</A> of grammar leon_views.dtd.
 */
public class LyFile extends LyFieldItem
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'type'. */
	public enum Type
	{
		TEXT,
		IMAGE,
		AUDIO,
		VIDEO,
		DIRECTORY;
	}


	/** Possible values for attribute 'border'. */
	public enum Border
	{
		NONE,
		SINGLE,
		DOUBLE,
		RAISED,
		LOWERED;
	}


	/* Static methods */

	/**
	 * Utility method to get type value from a string.
	 * @param token Token string
	 * @return Type value corresponding to this token
	 */
	public static final Type getType(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("TEXT")
				|| token.equals("0"))
			return Type.TEXT;
		else if (token.equalsIgnoreCase("IMAGE")
				|| token.equals("1"))
			return Type.IMAGE;
		else if (token.equalsIgnoreCase("AUDIO")
				|| token.equals("2"))
			return Type.AUDIO;
		else if (token.equalsIgnoreCase("VIDEO")
				|| token.equals("3"))
			return Type.VIDEO;
		else if (token.equalsIgnoreCase("DIRECTORY")
				|| token.equals("4"))
			return Type.DIRECTORY;
		else
			return Type.TEXT;
	}

	/**
	 * Utility method to get border value from a string.
	 * @param token Token string
	 * @return Border value corresponding to this token
	 */
	public static final Border getBorder(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("NONE")
				|| token.equals("0"))
			return Border.NONE;
		else if (token.equalsIgnoreCase("SINGLE")
				|| token.equals("1"))
			return Border.SINGLE;
		else if (token.equalsIgnoreCase("DOUBLE")
				|| token.equals("2"))
			return Border.DOUBLE;
		else if (token.equalsIgnoreCase("RAISED")
				|| token.equals("3"))
			return Border.RAISED;
		else if (token.equalsIgnoreCase("LOWERED")
				|| token.equals("4"))
			return Border.LOWERED;
		else
			return Border.NONE;
	}


	/* Attributes */

	/**
	 * Identifier of the component.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyFile			_template;

	/**
	 * Location of the file on the disk.
	 */
	protected String			_label;

	/**
	 * The type of file that can be selected. The following types are managed:
	 * TEXT: Text file
	 * IMAGE: Image file
	 * AUDIO: Audio file
	 * VIDEO: Video file
	 */
	protected Type				_type;

	/**
	 * Allows specifying the name of a Java class that replaces the default graphical 
	 * component.
	 */
	protected String			_bean;

	/**
	 * This attribute specifies the directory where files are stored on server. The 
	 * default value is a directory named "files".
	 */
	protected String			_attachmentDirectory;

	/**
	 * Attachment manager.
	 * If a value is set on this attribute, then the form sends not only the name but 
	 * also the content of the attached file.
	 * This file can then be kept on the server. By default, the sending of attachments
	 * is disabled.
	 * It is possible to specify the default manager, with ?default? as value (the 
	 * manager that is used is then leon.view.web.LyFileAttachmentHandler), or else a 
	 * specific manager inheriting the class that is used by default.
	 */
	protected String			_attachmentHandler;

	/**
	 * Directory from which search of files is done. If a file is chosen in a 
	 * sub-directory of directtory, the value is relative to this directory (its value 
	 * doesn't contain the directory part).
	 */
	protected String			_directory;

	/**
	 * "Filter for searched
	 * files (example: ""*.doc"")"
	 */
	protected String			_filter;

	/**
	 * Width of the file component.
	 */
	protected int				_nCols;

	/**
	 * Height of the file component.
	 */
	protected int				_nRows;

	/**
	 * Allows defining whether a border will be displayed around the image or not 
	 * (image type only):
	 * NONE: No border (default value)
	 * SINGLE: Simple border
	 * DOUBLE: Double border
	 * RAISED: Border simulating a raised 3D frame
	 * LOWERED: Border simulating a lowered 3D frame
	 */
	protected Border			_border;

	/**
	 * Indicates the image used for file selection button.
	 */
	protected String			_image;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Background of the file selector when enabled.
	 */
	protected String			_enabledColor;

	/**
	 * Background of the file selector when disabled.
	 */
	protected String			_disabledColor;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_type = Type.TEXT;
		_bean = null;
		_attachmentDirectory = "files";
		_attachmentHandler = null;
		_directory = null;
		_filter = null;
		_nCols = getIntValue("FILE.nCols", -1);
		_nRows = getIntValue("FILE.nRows", -1);
		_border = Border.NONE;
		_image = getStringValue("FILE.image", "$ALL.image");
		_font = getStringValue("FILE.font", "$ALL.font");
		_enabledColor = getStringValue("FILE.enabledColor", "$ALL.enabledColor");
		_disabledColor = getStringValue("FILE.disabledColor", "$ALL.disabledColor");
		_fontColor = getStringValue("FILE.fontColor", "$ALL.fontColor");
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyFile	model = (LyFile)component;

		_model = component;
		_label = model._label;
		_type = model._type;
		_bean = model._bean;
		_attachmentDirectory = model._attachmentDirectory;
		_attachmentHandler = model._attachmentHandler;
		_directory = model._directory;
		_filter = model._filter;
		_nCols = model._nCols;
		_nRows = model._nRows;
		_border = model._border;
		_image = model._image;
		_font = model._font;
		_enabledColor = model._enabledColor;
		_disabledColor = model._disabledColor;
		_fontColor = model._fontColor;
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyFile()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyFile(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyFile)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("type"))
			_type = getType(node.getString("type"));
		if (node.isSpecified("bean"))
			_bean = node.getString("bean");
		if (node.isSpecified("attachmentDirectory"))
			_attachmentDirectory = node.getString("attachmentDirectory");
		if (node.isSpecified("attachmentHandler"))
			_attachmentHandler = node.getString("attachmentHandler");
		if (node.isSpecified("directory"))
			_directory = node.getString("directory");
		if (node.isSpecified("filter"))
			_filter = node.getString("filter");
		if (node.isSpecified("nCols"))
			_nCols = node.getInt("nCols");
		if (node.isSpecified("nRows"))
			_nRows = node.getInt("nRows");
		if (node.isSpecified("border"))
			_border = getBorder(node.getString("border"));
		if (node.isSpecified("image"))
			_image = node.getString("image");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("enabledColor"))
			_enabledColor = node.getString("enabledColor");
		if (node.isSpecified("disabledColor"))
			_disabledColor = node.getString("disabledColor");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyFile clone = new LyFile();

		clone.initFromModel(this);
		clone._template = this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyFile getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of type.
	 * @return value of type.
	 */
	public Type getType()
	{
		return _type;
	}

	/**
	 * Gets value of bean.
	 * @return value of bean.
	 */
	public String getBean()
	{
		return _bean;
	}

	/**
	 * Gets value of attachmentDirectory.
	 * @return value of attachmentDirectory.
	 */
	public String getAttachmentDirectory()
	{
		return _attachmentDirectory;
	}

	/**
	 * Gets value of attachmentHandler.
	 * @return value of attachmentHandler.
	 */
	public String getAttachmentHandler()
	{
		return _attachmentHandler;
	}

	/**
	 * Gets value of directory.
	 * @return value of directory.
	 */
	public String getDirectory()
	{
		return _directory;
	}

	/**
	 * Gets value of filter.
	 * @return value of filter.
	 */
	public String getFilter()
	{
		return _filter;
	}

	/**
	 * Gets value of nCols.
	 * @return value of nCols.
	 */
	public int getNCols()
	{
		return _nCols;
	}

	/**
	 * Gets value of nRows.
	 * @return value of nRows.
	 */
	public int getNRows()
	{
		return _nRows;
	}

	/**
	 * Gets value of border.
	 * @return value of border.
	 */
	public Border getBorder()
	{
		return _border;
	}

	/**
	 * Gets value of image.
	 * @return value of image.
	 */
	public String getImage()
	{
		return _image;
	}

	/**
	 * Checks whether the image attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultImage()
	{
		return (_image == getStringValue("FILE.image"));
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("FILE.font"));
	}

	/**
	 * Gets value of enabledColor.
	 * @return value of enabledColor.
	 */
	public String getEnabledColor()
	{
		return _enabledColor;
	}

	/**
	 * Checks whether the enabledColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultEnabledColor()
	{
		return (_enabledColor == getStringValue("FILE.enabledColor"));
	}

	/**
	 * Gets value of disabledColor.
	 * @return value of disabledColor.
	 */
	public String getDisabledColor()
	{
		return _disabledColor;
	}

	/**
	 * Checks whether the disabledColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultDisabledColor()
	{
		return (_disabledColor == getStringValue("FILE.disabledColor"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("FILE.fontColor"));
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyFile template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView instanceof LySetLabelInterface)
			((LySetLabelInterface)_internalView).setLabel(label);
	}

	/**
	 * Initializes attribute type to the given value.
	 * @param type New value of type.
	 * @exception IllegalArgumentException if type parameter is null.
	 */
	public void setType(Type type)
	{
		if (type == _type)
			return;
		if (type == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute type");

		_type = type;
	}

	/**
	 * Initializes attribute bean to the given value.
	 * @param bean New value of bean.
	 */
	public void setBean(String bean)
	{
		if ((bean == _bean)
			|| ((bean != null)
				&& (_bean != null)
				&& bean.equals(_bean)))
			return;

		_bean = bean;
	}

	/**
	 * Initializes attribute attachmentDirectory to the given value.
	 * @param attachmentDirectory New value of attachmentDirectory.
	 */
	public void setAttachmentDirectory(String attachmentDirectory)
	{
		if ((attachmentDirectory == _attachmentDirectory)
			|| ((attachmentDirectory != null)
				&& (_attachmentDirectory != null)
				&& attachmentDirectory.equals(_attachmentDirectory)))
			return;

		_attachmentDirectory = attachmentDirectory;
	}

	/**
	 * Initializes attribute attachmentHandler to the given value.
	 * @param attachmentHandler New value of attachmentHandler.
	 */
	public void setAttachmentHandler(String attachmentHandler)
	{
		if ((attachmentHandler == _attachmentHandler)
			|| ((attachmentHandler != null)
				&& (_attachmentHandler != null)
				&& attachmentHandler.equals(_attachmentHandler)))
			return;

		_attachmentHandler = attachmentHandler;
	}

	/**
	 * Initializes attribute directory to the given value.
	 * @param directory New value of directory.
	 */
	public void setDirectory(String directory)
	{
		if ((directory == _directory)
			|| ((directory != null)
				&& (_directory != null)
				&& directory.equals(_directory)))
			return;

		_directory = directory;
	}

	/**
	 * Initializes attribute filter to the given value.
	 * @param filter New value of filter.
	 */
	public void setFilter(String filter)
	{
		if ((filter == _filter)
			|| ((filter != null)
				&& (_filter != null)
				&& filter.equals(_filter)))
			return;

		_filter = filter;
	}

	/**
	 * Initializes attribute nCols to the given value.
	 * @param nCols New value of nCols.
	 */
	public void setNCols(int nCols)
	{
		if (nCols == _nCols)
			return;
		_nCols = nCols;
	}

	/**
	 * Initializes attribute nRows to the given value.
	 * @param nRows New value of nRows.
	 */
	public void setNRows(int nRows)
	{
		if (nRows == _nRows)
			return;
		_nRows = nRows;
	}

	/**
	 * Initializes attribute border to the given value.
	 * @param border New value of border.
	 * @exception IllegalArgumentException if border parameter is null.
	 */
	public void setBorder(Border border)
	{
		if (border == _border)
			return;
		if (border == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute border");

		_border = border;
	}

	/**
	 * Initializes attribute image to the given value.
	 * @param image New value of image.
	 */
	public void setImage(String image)
	{
		if ((image == _image)
			|| ((image != null)
				&& (_image != null)
				&& image.equals(_image)))
			return;

		_image = image;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute enabledColor to the given value.
	 * @param enabledColor New value of enabledColor.
	 */
	public void setEnabledColor(String enabledColor)
	{
		if ((enabledColor == _enabledColor)
			|| ((enabledColor != null)
				&& (_enabledColor != null)
				&& enabledColor.equals(_enabledColor)))
			return;

		_enabledColor = enabledColor;
	}

	/**
	 * Initializes attribute disabledColor to the given value.
	 * @param disabledColor New value of disabledColor.
	 */
	public void setDisabledColor(String disabledColor)
	{
		if ((disabledColor == _disabledColor)
			|| ((disabledColor != null)
				&& (_disabledColor != null)
				&& disabledColor.equals(_disabledColor)))
			return;

		_disabledColor = disabledColor;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<FILE");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if ((_template == null)
			|| (_type != _template._type))
		{
			if ((_type == Type.TEXT) && (_template != null))
				pw.print(" type='TEXT'");
			else if (_type == Type.IMAGE)
				pw.print(" type='IMAGE'");
			else if (_type == Type.AUDIO)
				pw.print(" type='AUDIO'");
			else if (_type == Type.VIDEO)
				pw.print(" type='VIDEO'");
			else if (_type == Type.DIRECTORY)
				pw.print(" type='DIRECTORY'");
		}
		if (((_template == null) && (_bean != null))
				|| ((_template != null) && (_bean != _template._bean)))
			pw.print(" bean='" + LyString.getXmlString(_bean) + "'");
		if (((_template == null) && (_attachmentDirectory != "files"))
				|| ((_template != null) && (_attachmentDirectory != _template._attachmentDirectory)))
			pw.print(" attachmentDirectory='" + LyString.getXmlString(_attachmentDirectory) + "'");
		if (((_template == null) && (_attachmentHandler != null))
				|| ((_template != null) && (_attachmentHandler != _template._attachmentHandler)))
			pw.print(" attachmentHandler='" + LyString.getXmlString(_attachmentHandler) + "'");
		if (((_template == null) && (_directory != null))
				|| ((_template != null) && (_directory != _template._directory)))
			pw.print(" directory='" + LyString.getXmlString(_directory) + "'");
		if (((_template == null) && (_filter != null))
				|| ((_template != null) && (_filter != _template._filter)))
			pw.print(" filter='" + LyString.getXmlString(_filter) + "'");
		if (((_template == null) && (_nCols != getIntValue("FILE.nCols", -1)))
				|| ((_template != null) && (_nCols != _template._nCols)))
			pw.print(" nCols='" + _nCols + "'");
		if (((_template == null) && (_nRows != getIntValue("FILE.nRows", -1)))
				|| ((_template != null) && (_nRows != _template._nRows)))
			pw.print(" nRows='" + _nRows + "'");
		if ((_template == null)
			|| (_border != _template._border))
		{
			if ((_border == Border.NONE) && (_template != null))
				pw.print(" border='NONE'");
			else if (_border == Border.SINGLE)
				pw.print(" border='SINGLE'");
			else if (_border == Border.DOUBLE)
				pw.print(" border='DOUBLE'");
			else if (_border == Border.RAISED)
				pw.print(" border='RAISED'");
			else if (_border == Border.LOWERED)
				pw.print(" border='LOWERED'");
		}
		if (((_template == null) && (_image != getStringValue("FILE.image", "$ALL.image")))
				|| ((_template != null) && (_image != _template._image)))
			pw.print(" image='" + LyString.getXmlString(_image) + "'");
		if (((_template == null) && (_font != getStringValue("FILE.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_enabledColor != getStringValue("FILE.enabledColor", "$ALL.enabledColor")))
				|| ((_template != null) && (_enabledColor != _template._enabledColor)))
			pw.print(" enabledColor='" + LyString.getXmlString(_enabledColor) + "'");
		if (((_template == null) && (_disabledColor != getStringValue("FILE.disabledColor", "$ALL.disabledColor")))
				|| ((_template != null) && (_disabledColor != _template._disabledColor)))
			pw.print(" disabledColor='" + LyString.getXmlString(_disabledColor) + "'");
		if (((_template == null) && (_fontColor != getStringValue("FILE.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		pw.println("/>");
	}
}
