/*
 * LyFormBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyFormBase class is a generated Java implementation that corresponds to the element 
 * FORM of the descriptive grammar of the views (views.dtd).
 * <B>&lt;FORM&gt; : Form</B>
 * A form allows showing a set of fields.
 * This type of component is used to allow an application user to consult 
 * application data (non editable form) or else to input data (editable form). The 
 * default arrangement of the fields inside a form is based on a grid of two 
 * columns comprising as many lines as there are fields.
 * The first column contains the labels of the fields and the second one contains 
 * the input fields.
 * @see leon.view.LyPanel
 * @see leon.view.LyLyForm
 * See also tag <A href="../../conf/dtd/view/FORM.html">FORM</A> of grammar leon_views.dtd.
 */
public abstract class LyFormBase extends LyPanel
	implements  LyGetIdInterface,
				LySetEditableInterface,
				LyGetControllerInterface,
				LySetControllerInterface,
				LyGetConstraintInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'align'. */
	public enum Align
	{
		LEFT,
		RIGHT,
		CENTER;
	}


	/* Static methods */

	/**
	 * Utility method to get align value from a string.
	 * @param token Token string
	 * @return Align value corresponding to this token
	 */
	public static final Align getAlign(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("LEFT")
				|| token.equals("0"))
			return Align.LEFT;
		else if (token.equalsIgnoreCase("RIGHT")
				|| token.equals("1"))
			return Align.RIGHT;
		else if (token.equalsIgnoreCase("CENTER")
				|| token.equals("2"))
			return Align.CENTER;
		else
			return Align.LEFT;
	}


	/* Attributes */

	/**
	 * Identifier of the form.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyForm			_template;

	/**
	 * Indicates whether the form is editable or not.
	 * However, the fields contained in an editable form can be non editable (editable 
	 * attribute of a field).
	 */
	protected boolean			_editable;

	/**
	 * Indicates the alignment of the fields contained in the form (left, right or 
	 * center).
	 * The default value is left.
	 */
	protected Align				_align;

	/**
	 * Reference of the controller that manages the form (i.e. handles user demands and
	 * refresh content).
	 */
	protected String			_controller;

	/**
	 * Indicates if the form is outlined with a border.
	 */
	protected boolean			_border;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * Color used to display a field on warning.
	 */
	protected String			_warningColor;

	/**
	 * Color used to display a field on error.
	 */
	protected String			_errorColor;


	/* Sub components */

	/**
	 * List of the fields contained in the form.
	 */
	protected ArrayList<LyFormField>		_formFields;

	/**
	 * Positioning constraint of the form.
	 */
	protected LyConstraint		_constraint;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_editable = getBooleanValue("FORM.editable", true);
		_align = Align.LEFT;
		_controller = null;
		_border = getBooleanValue("FORM.border", true);
		_font = getStringValue("FORM.font", "$ALL.font");
		_color = getStringValue("FORM.color", "$ALL.color");
		_fontColor = getStringValue("FORM.fontColor", "$ALL.fontColor");
		_warningColor = getStringValue("FORM.warningColor", "$ALL.warningColor");
		_errorColor = getStringValue("FORM.errorColor", "$ALL.errorColor");

		// Initialize sub components
		_formFields = null;
		_constraint = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyForm	model = (LyForm)component;

		_model = component;
		_editable = model._editable;
		_align = model._align;
		_controller = model._controller;
		_border = model._border;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;
		_warningColor = model._warningColor;
		_errorColor = model._errorColor;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyFormField	formFields;
		int			i, n;

		n = (model._formFields == null ? 0 : model._formFields.size());
		for (i = 0; i < n; i++)
		{
			formFields = model._formFields.get(i);
			if (formFields != null)
				formFields = (LyFormField)formFields.cloneComponent();
			addChild(formFields);
		}
		if (model._constraint != null)
			addChild(model._constraint.cloneComponent());
		if (model._datas != null)
			_datas = new Hashtable<String, String>(model._datas);
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	protected LyFormBase()
	{
		_id = getInternalId();
	}

	/**
	 * Constructor with all required attributes.
	 * @param id Identifier of the form.
	 */
	public LyFormBase(String id)
	{
		// Fill up required attributes with parameters
		_id = (id != null ? id.intern() : null);
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyFormBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyForm)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("editable"))
			_editable = node.getBoolean("editable");
		if (node.isSpecified("align"))
			_align = getAlign(node.getString("align"));
		if (node.isSpecified("controller"))
			_controller = node.getString("controller");
		if (node.isSpecified("border"))
			_border = node.getBoolean("border");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("warningColor"))
			_warningColor = node.getString("warningColor");
		if (node.isSpecified("errorColor"))
			_errorColor = node.getString("errorColor");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyFormField)
			addFormField((LyFormField)child);
		else if (child instanceof LyConstraint)
			setConstraint((LyConstraint)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyForm clone = new LyForm();

		clone.initFromModel(this);
		clone._template = (LyForm)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyForm getTemplate()
	{
		return _template;
	}

	/**
	 * Checks if value of attribute editable is set.
	 * @return true if editable is set, false otherwise.
	 */
	public boolean isEditable()
	{
		return _editable;
	}

	/**
	 * Gets value of align.
	 * @return value of align.
	 */
	public Align getAlign()
	{
		return _align;
	}

	/**
	 * Gets value of controller.
	 * @return value of controller.
	 */
	public String getController()
	{
		return _controller;
	}

	/**
	 * Checks if value of attribute border is set.
	 * @return true if border is set, false otherwise.
	 */
	public boolean isBorder()
	{
		return _border;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("FORM.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("FORM.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("FORM.fontColor"));
	}

	/**
	 * Gets value of warningColor.
	 * @return value of warningColor.
	 */
	public String getWarningColor()
	{
		return _warningColor;
	}

	/**
	 * Checks whether the warningColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultWarningColor()
	{
		return (_warningColor == getStringValue("FORM.warningColor"));
	}

	/**
	 * Gets value of errorColor.
	 * @return value of errorColor.
	 */
	public String getErrorColor()
	{
		return _errorColor;
	}

	/**
	 * Checks whether the errorColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultErrorColor()
	{
		return (_errorColor == getStringValue("FORM.errorColor"));
	}

	/**
	 * Get sub-component FormField of index i.
	 * @param i the index of the FormField to get.
	 * @return sub-component FormField of index i.
	 */
	public LyFormField getFormField(int i)
	{
		if (_formFields == null)
			return null;
		if ((i < 0) || (i >= _formFields.size()))
			return null;
		return _formFields.get(i);
	}

	/**
	 * Get sub-component FormField count.
	 * @return sub-component FormField count.
	 */
	public int getFormFieldCount()
	{
		if (_formFields == null)
			return 0;
		return _formFields.size();
	}

	/**
	 * Get sub-component Constraint.
	 * @return sub-component Constraint.
	 */
	public LyConstraint getConstraint()
	{
		return _constraint;
	}


	/* Modifiers */

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyForm template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute editable to the given value.
	 * @param editable Boolean indicating if attribute editable is set or unset.
	 */
	public void setEditable(boolean editable)
	{
		if (editable == _editable)
			return;
		_editable = editable;
		if (_internalView != null)
			((LyGraphicalForm)_internalView).setEditable(editable);
	}

	/**
	 * Initializes attribute align to the given value.
	 * @param align New value of align.
	 * @exception IllegalArgumentException if align parameter is null.
	 */
	public void setAlign(Align align)
	{
		if (align == _align)
			return;
		if (align == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute align");

		_align = align;
	}

	/**
	 * Initializes attribute controller to the given value.
	 * @param controller New value of controller.
	 */
	public void setController(String controller)
	{
		if ((controller == _controller)
			|| ((controller != null)
				&& (_controller != null)
				&& controller.equals(_controller)))
			return;

		_controller = controller;
	}

	/**
	 * Initializes attribute border to the given value.
	 * @param border Boolean indicating if attribute border is set or unset.
	 */
	public void setBorder(boolean border)
	{
		if (border == _border)
			return;
		_border = border;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Initializes attribute warningColor to the given value.
	 * @param warningColor New value of warningColor.
	 */
	public void setWarningColor(String warningColor)
	{
		if ((warningColor == _warningColor)
			|| ((warningColor != null)
				&& (_warningColor != null)
				&& warningColor.equals(_warningColor)))
			return;

		_warningColor = warningColor;
	}

	/**
	 * Initializes attribute errorColor to the given value.
	 * @param errorColor New value of errorColor.
	 */
	public void setErrorColor(String errorColor)
	{
		if ((errorColor == _errorColor)
			|| ((errorColor != null)
				&& (_errorColor != null)
				&& errorColor.equals(_errorColor)))
			return;

		_errorColor = errorColor;
	}

	/**
	 * Adds a sub-component formField.
	 * @param formField The formField to add to formFields list.
	 */
	public void addFormField(LyFormField formField)
	{
		insertFormField(formField, -1);
	}

	/**
	 * Inserts a sub-component formField at given position.
	 * @param formField The formField to add to formFields list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertFormField(LyFormField formField, int pos)
	{
		if (_formFields == null)
			_formFields = new ArrayList<LyFormField>(1);
		else if (_formFields.contains(formField))
			return;
		if (pos >= 0)
			_formFields.add(pos,formField);
		else
			_formFields.add(formField);
		if (formField != null)
			formField.setParent(this);
	}

	/**
	 * Sets sub-component constraint.
	 * @param constraint New value of constraint
	 */
	public void setConstraint(LyConstraint constraint)
	{
		if (_constraint != null)
			_constraint.setParent(null);
		_constraint = constraint;
		if (constraint != null)
			constraint.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<FORM");
		pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && !_editable)
				|| ((_template != null) && (_editable != _template._editable)))
			pw.print(" editable='" + _editable + "'");
		if ((_template == null)
			|| (_align != _template._align))
		{
			if ((_align == Align.LEFT) && (_template != null))
				pw.print(" align='LEFT'");
			else if (_align == Align.RIGHT)
				pw.print(" align='RIGHT'");
			else if (_align == Align.CENTER)
				pw.print(" align='CENTER'");
		}
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlString(_controller) + "'");
		if (((_template == null) && !_border)
				|| ((_template != null) && (_border != _template._border)))
			pw.print(" border='" + _border + "'");
		if (((_template == null) && (_font != getStringValue("FORM.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("FORM.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("FORM.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && (_warningColor != getStringValue("FORM.warningColor", "$ALL.warningColor")))
				|| ((_template != null) && (_warningColor != _template._warningColor)))
			pw.print(" warningColor='" + LyString.getXmlString(_warningColor) + "'");
		if (((_template == null) && (_errorColor != getStringValue("FORM.errorColor", "$ALL.errorColor")))
				|| ((_template != null) && (_errorColor != _template._errorColor)))
			pw.print(" errorColor='" + LyString.getXmlString(_errorColor) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	formFields;
		int			i, n;

		n = (_formFields == null ? 0 : _formFields.size());
		for (i = 0; i < n; i++)
		{
			formFields = _formFields.get(i);
			formFields.writeXml(pw, debug);
		}

		if (_constraint != null)
			_constraint.writeXml(pw, debug);

		writeDatas(pw);

		pw.println("</FORM>");
	}
}
