/*
 * LyFormField.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved. Current Version: $Id: LyFormField.java,v 6.9
 * 2008/03/27 14:53:49 jls Exp $.
 */

package leon.view;

import leon.view.graphic.*;
import leon.view.setinterface.*;


/**
 * The LyFormField class implements generic behavior for the FIELD components as described in
 * views.dtd regardless of the requested target (win, swing, html...). It represents a form field.
 *
 * @see  leon.view.LyComponent
 */
public class LyFormField extends LyFormFieldBase
{
	public static final String __VERSION = "$Id: LyFormField.java 31240 2011-05-30 09:48:37Z jlsaouli $";

	/**
     * Default constructor.
     */
    protected LyFormField()
    {
    }
    
	/**
	 * Constructor with all required attributes.
	 */
	public LyFormField(String id)
	{
		super(id);
	}

	/* Protected methods for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 *
	 * @param  node  Node from which the component is constructed.
	 *
	 * @see    leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyFormField(LyViewNode node)
	{
		super(node);
	}

	/**
	 * Checks if value of attribute valid is set.
	 *
	 * @return      true if valid is set, false otherwise.
	 *
	 * @deprecated  Use getState method
	 */
	public boolean isValid()
	{
		return (_state != STATE_ERROR);
	}

	/**
	 * Sets attribute valid to the given value.
	 *
	 * @param       valid  Boolean indicating if attribute valid is set or unset.
	 *
	 * @deprecated  Use setState method
	 */
	public void setValid(boolean valid)
	{
		setState(valid ? STATE_NORMAL : STATE_ERROR);
	}

	/**
	 * Sets attribute state to the given value. This method is overriden to allow states greater
	 * than STATE_WARNING.
	 *
	 * @param      state  New value of STATE_
	 *
	 * @exception  IllegalArgumentException  if state is not a known value for this attribute.
	 */
	public void setState(short state)
	{
		if (state == _state)
			return;

		if (state < STATE_NORMAL)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute state");

		if (state == STATE_NORMAL)
		    setMessage(null);
		
		_state = state;
		if (_internalView != null)
			((LyGraphicalFormField)_internalView).setState(state);
	}

	/**
	 * Reset content of this form field.
	 */
	public void resetContent()
	{
		// Force a refresh of the 'internal' structure
		Object internal = getInternalView();

		if (internal instanceof LyResetContentInterface)
			((LyResetContentInterface)internal).resetContent();
	}

	/**
	 * Sets id of this formField.
	 *
	 * @param  id  New id of the formField.
	 */
	public void setId(String id)
	{
		_id = (id != null ? id.intern() : null);
	}

	/**
	 * To give the focus to this formField.
	 */
	public void setFocus()
	{
		// Force a refresh of the 'internal' structure
		Object internal = getInternalView(false);

		if (internal instanceof LySetFocusInterface)
			((LySetFocusInterface)internal).setFocus();
	}

	/**
	 * End of construction : completion with child component.
	 *
	 * @param  child  Sub child component.
	 *
	 * @see    leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyFieldUnit)
			setFieldUnit((LyFieldUnit)child);
		else
			super.addChild(child);
	}

	/**
	 * Set parent component.
	 *
	 * @param  parent  the new parent for this component.
	 */
	void setParent(LyComponent parent)
	{
		super.setParent(parent);

		// To optimize the search,add myself on the top dialog/frame
		LyView view = getParentView();

		if (view != null)
			view.addChildInCache(this, false);
		else
		{
			while ((parent != null) && (parent.getParent() != null))
				parent = parent.getParent();

			if (parent != null)
				parent.addChildInCache(this, true);
		}
	}
	
	/*
	 * (non-Javadoc)
	 * @see leon.view.LyComponent#setVisible(boolean)
	 */
	public void setVisible(boolean status)
	{
		if (_hidden && status)
			return;
		
		super.setVisible(status);
	}

	/**
	 * Returns the position and size of current component relatively to given component. Overridden
	 * to compute bounds for both caption and field item.
	 *
	 * @param   parentComponent  the component used to compute relative position of current
	 *                           component. If null, position of current component is computed with
	 *                           the top-left corner of the screen as the origin.
	 *
	 * @return  an array of 4 integers (x position, y position, width, height) representing the
	 *          position and size of current component.
	 */
	public int[] getBounds(LyComponent parentComponent)
	{
		LyCaption   caption = this.getCaption();
		LyFieldItem fieldItem = this.getFieldItem();
		int[]	    captionBounds = (caption == null) ? null : caption.getBounds(parentComponent);
		int[]	    itemBounds = (fieldItem == null) ? null : fieldItem.getBounds(parentComponent);
		int[]	    result = new int[4];

		if (captionBounds == null)
			return itemBounds;
		else if (itemBounds == null)
			return captionBounds;
		else
		{
			int x2 = Math.max(captionBounds[0] + captionBounds[2], itemBounds[0] + itemBounds[2]);
			int y2 = Math.max(captionBounds[1] + captionBounds[3], itemBounds[1] + itemBounds[3]);

			result[0] = Math.min(captionBounds[0], itemBounds[0]);
			result[1] = Math.min(captionBounds[1], itemBounds[1]);
			result[2] = x2 - result[0];
			result[3] = y2 - result[1];

			return result;
		}
	}

	/**
	 * Free this component.
	 */
	public void free()
	{
		super.free();
	}
} // end class LyFormField
