/*
 * LyFormFieldBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyFormFieldBase class is a generated Java implementation that corresponds to the element 
 * FIELD of the descriptive grammar of the views (views.dtd).
 * <B>&lt;FIELD&gt; : Field</B>
 * A field is contained in a form and allows to show or to input application data.
 * It is principally made up of a label (optional) and an input field depending of 
 * the type of the data (text, number, date, color...).
 * It is possible to associate a tool bar and a help entry to this field
 * @see leon.view.LyPanel
 * @see leon.view.LyLyFormField
 * See also tag <A href="../../conf/dtd/view/FIELD.html">FIELD</A> of grammar leon_views.dtd.
 */
public abstract class LyFormFieldBase extends LyPanel
	implements  LyGetIdInterface,
				LySetEditableInterface,
				LySetOptionalInterface,
				LySetEnabledInterface,
				LySetStateInterface,
				LySetHiddenInterface,
				LyGetControllerInterface,
				LySetControllerInterface,
				LyGetConstraintInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'state'. */
	public static final short	STATE_NORMAL = 0;
	public static final short	STATE_ERROR = 1;
	public static final short	STATE_WARNING = 2;


	/* Static methods */

	/**
	 * Utility method to get state value from a string.
	 * @param token Token string
	 * @return short value corresponding to this token
	 */
	public static final short getshort(String token)
	{
		if (token == null)
			return -1;
		else if (token.equalsIgnoreCase("NORMAL")
				|| token.equals("0"))
			return STATE_NORMAL;
		else if (token.equalsIgnoreCase("ERROR")
				|| token.equals("1"))
			return STATE_ERROR;
		else if (token.equalsIgnoreCase("WARNING")
				|| token.equals("2"))
			return STATE_WARNING;
		else
			return -1;
	}


	/* Attributes */

	/**
	 * Identifier of the form.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyFormField			_template;

	/**
	 * If the form is editable, the value of this attribute indicates if the field is 
	 * editable.
	 * If the form in which this field is placed is not editable, the value of this 
	 * attribute is ignored (the field is then not editable).
	 */
	protected boolean			_editable;

	/**
	 * Indicates whether the field is optional or not.
	 * A non optional field (i.e. mandatory) must be input.
	 * The only effect of this attribute at the level of the view is that the label 
	 * will be displayed in bold.
	 */
	protected boolean			_optional;

	/**
	 * Indicates whether the field is enabled or not (true if it is enabled).
	 * If a field is editable but is not enabled, it is not possible to input any 
	 * value.
	 */
	protected boolean			_enabled;

	/**
	 * Indicates whether the field is valid or not.
	 * A field declared as non-valid is a field for which the input value is wrong 
	 * (examples: null value on a mandatory field, format that is not respected,...).
	 * At the level of the view, the effect of this attribute is that the label will be
	 * displayed in red.
	 */
	protected boolean			_valid;

	/**
	 * Indicates that the field has an associated check box that allows 
	 * enabling/disabling the corresponding input field.
	 */
	protected boolean			_checkBox;

	/**
	 * Indicates whether the associated check box is checked or not.
	 */
	protected boolean			_checked;

	/**
	 * Indicates if the modification of the value of this field implies the update of 
	 * another field (true if an update should be performed).
	 */
	protected boolean			_needCheck;

	/**
	 * State of the field. The state of the field has one of the following values:
	 * NORMAL: Normal state,
	 * WARNING: warning state (non-blocking state)
	 * ERROR: error state (blocking state).
	 */
	protected short				_state;

	/**
	 * Error or warning message displayed on input field.
	 */
	protected String			_message;

	/**
	 * Indicates that the field is hidden (not displayed).
	 */
	protected boolean			_hidden;

	/**
	 * Reference of the controller that manages the field (i.e. handles user demands 
	 * and refresh content).
	 */
	protected String			_controller;


	/* Sub components */

	/**
	 * Legend of the field.
	 */
	protected LyCaption			_caption;

	/**
	 * Input field of the field.
	 */
	protected LyFieldItem		_fieldItem;

	/**
	 * Tool bar associated to the field.
	 */
	protected LyToolBar			_toolBar;

	/**
	 * Positioning constraint of the field within its parent area.
	 */
	protected LyConstraint		_constraint;

	/**
	 * Tooltip associated to the field.
	 */
	protected LyTip			_tip;

	/**
	 * Input field of Relation type.
	 */
	protected LyRelation		_relation;

	/**
	 * Unit of the field.
	 */
	protected LyFieldUnit		_fieldUnit;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_editable = getBooleanValue("FIELD.editable", true);
		_optional = getBooleanValue("FIELD.optional", true);
		_enabled = getBooleanValue("FIELD.enabled", true);
		_valid = getBooleanValue("FIELD.valid", true);
		_checkBox = getBooleanValue("FIELD.checkBox", false);
		_checked = getBooleanValue("FIELD.checked", false);
		_needCheck = getBooleanValue("FIELD.needCheck", false);
		_state = getShortValue("FIELD.state", STATE_NORMAL);
		_message = null;
		_hidden = getBooleanValue("FIELD.hidden", false);
		_controller = null;

		// Initialize sub components
		_caption = null;
		_fieldItem = null;
		_toolBar = null;
		_constraint = null;
		_tip = null;
		_relation = null;
		_fieldUnit = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyFormField	model = (LyFormField)component;

		_model = component;
		_editable = model._editable;
		_optional = model._optional;
		_enabled = model._enabled;
		_valid = model._valid;
		_checkBox = model._checkBox;
		_checked = model._checked;
		_needCheck = model._needCheck;
		_state = model._state;
		_message = model._message;
		_hidden = model._hidden;
		_controller = model._controller;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		if (model._caption != null)
			addChild(model._caption.cloneComponent());
		if (model._fieldItem != null)
			addChild(model._fieldItem.cloneComponent());
		if (model._toolBar != null)
			addChild(model._toolBar.cloneComponent());
		if (model._constraint != null)
			addChild(model._constraint.cloneComponent());
		if (model._tip != null)
			addChild(model._tip.cloneComponent());
		if (model._datas != null)
			_datas = new Hashtable<String, String>(model._datas);
		if (model._relation != null)
			addChild(model._relation.cloneComponent());
		if (model._fieldUnit != null)
			addChild(model._fieldUnit.cloneComponent());
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	protected LyFormFieldBase()
	{
		_id = getInternalId();
	}

	/**
	 * Constructor with all required attributes.
	 * @param id Identifier of the form.
	 */
	public LyFormFieldBase(String id)
	{
		// Fill up required attributes with parameters
		_id = (id != null ? id.intern() : null);
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyFormFieldBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyFormField)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("editable"))
			_editable = node.getBoolean("editable");
		if (node.isSpecified("optional"))
			_optional = node.getBoolean("optional");
		if (node.isSpecified("enabled"))
			_enabled = node.getBoolean("enabled");
		if (node.isSpecified("valid"))
			_valid = node.getBoolean("valid");
		if (node.isSpecified("checkBox"))
			_checkBox = node.getBoolean("checkBox");
		if (node.isSpecified("checked"))
			_checked = node.getBoolean("checked");
		if (node.isSpecified("needCheck"))
			_needCheck = node.getBoolean("needCheck");
		if (node.isSpecified("state"))
			_state = getshort(node.getString("state"));
		if (node.isSpecified("message"))
			_message = node.getString("message");
		if (node.isSpecified("hidden"))
			_hidden = node.getBoolean("hidden");
		if (node.isSpecified("controller"))
			_controller = node.getString("controller");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyCaption)
			setCaption((LyCaption)child);
		else if (child instanceof LyFieldItem)
			setFieldItem((LyFieldItem)child);
		else if (child instanceof LyToolBar)
			setToolBar((LyToolBar)child);
		else if (child instanceof LyConstraint)
			setConstraint((LyConstraint)child);
		else if (child instanceof LyTip)
			setTip((LyTip)child);
		else if (child instanceof LyRelation)
			setRelation((LyRelation)child);
		else if (child instanceof LyFieldUnit)
			setFieldUnit((LyFieldUnit)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyFormField clone = new LyFormField();

		clone.initFromModel(this);
		clone._template = (LyFormField)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyFormField getTemplate()
	{
		return _template;
	}

	/**
	 * Checks if value of attribute editable is set.
	 * @return true if editable is set, false otherwise.
	 */
	public boolean isEditable()
	{
		return _editable;
	}

	/**
	 * Checks if value of attribute optional is set.
	 * @return true if optional is set, false otherwise.
	 */
	public boolean isOptional()
	{
		return _optional;
	}

	/**
	 * Checks if value of attribute enabled is set.
	 * @return true if enabled is set, false otherwise.
	 */
	public boolean isEnabled()
	{
		return _enabled;
	}

	/**
	 * Checks if value of attribute valid is set.
	 * @return true if valid is set, false otherwise.
	 */
	public boolean isValid()
	{
		return _valid;
	}

	/**
	 * Checks if value of attribute checkBox is set.
	 * @return true if checkBox is set, false otherwise.
	 */
	public boolean isCheckBox()
	{
		return _checkBox;
	}

	/**
	 * Checks if value of attribute checked is set.
	 * @return true if checked is set, false otherwise.
	 */
	public boolean isChecked()
	{
		return _checked;
	}

	/**
	 * Checks if value of attribute needCheck is set.
	 * @return true if needCheck is set, false otherwise.
	 */
	public boolean isNeedCheck()
	{
		return _needCheck;
	}

	/**
	 * Gets value of state.
	 * @return value of state.
	 */
	public short getState()
	{
		return _state;
	}

	/**
	 * Gets value of message.
	 * @return value of message.
	 */
	public String getMessage()
	{
		return _message;
	}

	/**
	 * Checks if value of attribute hidden is set.
	 * @return true if hidden is set, false otherwise.
	 */
	public boolean isHidden()
	{
		return _hidden;
	}

	/**
	 * Gets value of controller.
	 * @return value of controller.
	 */
	public String getController()
	{
		return _controller;
	}

	/**
	 * Get sub-component Caption.
	 * @return sub-component Caption.
	 */
	public LyCaption getCaption()
	{
		return _caption;
	}

	/**
	 * Get sub-component FieldItem.
	 * @return sub-component FieldItem.
	 */
	public LyFieldItem getFieldItem()
	{
		return _fieldItem;
	}

	/**
	 * Get sub-component ToolBar.
	 * @return sub-component ToolBar.
	 */
	public LyToolBar getToolBar()
	{
		return _toolBar;
	}

	/**
	 * Get sub-component Constraint.
	 * @return sub-component Constraint.
	 */
	public LyConstraint getConstraint()
	{
		return _constraint;
	}

	/**
	 * Get sub-component Tip.
	 * @return sub-component Tip.
	 */
	public LyTip getTip()
	{
		return _tip;
	}

	/**
	 * Get sub-component Relation.
	 * @return sub-component Relation.
	 */
	public LyRelation getRelation()
	{
		return _relation;
	}

	/**
	 * Get sub-component FieldUnit.
	 * @return sub-component FieldUnit.
	 */
	public LyFieldUnit getFieldUnit()
	{
		return _fieldUnit;
	}


	/* Modifiers */

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyFormField template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute editable to the given value.
	 * @param editable Boolean indicating if attribute editable is set or unset.
	 */
	public void setEditable(boolean editable)
	{
		if (editable == _editable)
			return;
		_editable = editable;
		if (_internalView != null)
			((LyGraphicalFormField)_internalView).setEditable(editable);
	}

	/**
	 * Sets attribute optional to the given value.
	 * @param optional Boolean indicating if attribute optional is set or unset.
	 */
	public void setOptional(boolean optional)
	{
		if (optional == _optional)
			return;
		_optional = optional;
		if (_internalView != null)
			((LyGraphicalFormField)_internalView).setOptional(optional);
	}

	/**
	 * Sets attribute enabled to the given value.
	 * @param enabled Boolean indicating if attribute enabled is set or unset.
	 */
	public void setEnabled(boolean enabled)
	{
		if (enabled == _enabled)
			return;
		_enabled = enabled;
		if (_internalView != null)
			((LyGraphicalFormField)_internalView).setEnabled(enabled);
	}

	/**
	 * Initializes attribute valid to the given value.
	 * @param valid Boolean indicating if attribute valid is set or unset.
	 */
	public void setValid(boolean valid)
	{
		if (valid == _valid)
			return;
		_valid = valid;
	}

	/**
	 * Initializes attribute checkBox to the given value.
	 * @param checkBox Boolean indicating if attribute checkBox is set or unset.
	 */
	public void setCheckBox(boolean checkBox)
	{
		if (checkBox == _checkBox)
			return;
		_checkBox = checkBox;
	}

	/**
	 * Sets attribute checked to the given value.
	 * @param checked Boolean indicating if attribute checked is set or unset.
	 */
	public void setChecked(boolean checked)
	{
		if (checked == _checked)
			return;
		_checked = checked;
		if (_internalView != null)
			((LyGraphicalFormField)_internalView).setChecked(checked);
	}

	/**
	 * Initializes attribute needCheck to the given value.
	 * @param needCheck Boolean indicating if attribute needCheck is set or unset.
	 */
	public void setNeedCheck(boolean needCheck)
	{
		if (needCheck == _needCheck)
			return;
		_needCheck = needCheck;
	}

	/**
	 * Sets attribute state to the given value.
	 * @param state New value of state.
	 * @exception IllegalArgumentException if state parameter is null.
	 */
	public void setState(short state)
	{
		if (state == _state)
			return;
		if ((state < STATE_NORMAL) || (state > STATE_WARNING))
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute state");

		_state = state;
		if (_internalView != null)
			((LyGraphicalFormField)_internalView).setState(state);
	}

	/**
	 * Sets attribute message to the given value.
	 * @param message New value of message.
	 */
	public void setMessage(String message)
	{
		if ((message == _message)
			|| ((message != null)
				&& (_message != null)
				&& message.equals(_message)))
			return;

		_message = message;
		if (_internalView != null)
			((LyGraphicalFormField)_internalView).setMessage(message);
	}

	/**
	 * Sets attribute hidden to the given value.
	 * @param hidden Boolean indicating if attribute hidden is set or unset.
	 */
	public void setHidden(boolean hidden)
	{
		if (hidden == _hidden)
			return;
		_hidden = hidden;
		if (_internalView != null)
			((LyGraphicalFormField)_internalView).setHidden(hidden);
	}

	/**
	 * Initializes attribute controller to the given value.
	 * @param controller New value of controller.
	 */
	public void setController(String controller)
	{
		if ((controller == _controller)
			|| ((controller != null)
				&& (_controller != null)
				&& controller.equals(_controller)))
			return;

		_controller = controller;
	}

	/**
	 * Sets sub-component caption.
	 * @param caption New value of caption
	 */
	public void setCaption(LyCaption caption)
	{
		if (_caption != null)
			_caption.setParent(null);
		_caption = caption;
		if (caption != null)
			caption.setParent(this);
	}

	/**
	 * Sets sub-component fieldItem.
	 * @param fieldItem New value of fieldItem
	 */
	public void setFieldItem(LyFieldItem fieldItem)
	{
		if (_fieldItem != null)
			_fieldItem.setParent(null);
		_fieldItem = fieldItem;
		if (fieldItem != null)
			fieldItem.setParent(this);
	}

	/**
	 * Sets sub-component toolBar.
	 * @param toolBar New value of toolBar
	 */
	public void setToolBar(LyToolBar toolBar)
	{
		if (_toolBar != null)
			_toolBar.setParent(null);
		_toolBar = toolBar;
		if (toolBar != null)
			toolBar.setParent(this);
		if (_internalView != null)
			((LyGraphicalFormField)_internalView).setToolBar(toolBar);
	}

	/**
	 * Sets sub-component constraint.
	 * @param constraint New value of constraint
	 */
	public void setConstraint(LyConstraint constraint)
	{
		if (_constraint != null)
			_constraint.setParent(null);
		_constraint = constraint;
		if (constraint != null)
			constraint.setParent(this);
	}

	/**
	 * Sets sub-component tip.
	 * @param tip New value of tip
	 */
	public void setTip(LyTip tip)
	{
		if (_tip != null)
			_tip.setParent(null);
		_tip = tip;
		if (tip != null)
			tip.setParent(this);
	}

	/**
	 * Sets sub-component relation.
	 * @param relation New value of relation
	 */
	public void setRelation(LyRelation relation)
	{
		if (_relation != null)
			_relation.setParent(null);
		_relation = relation;
		if (relation != null)
			relation.setParent(this);
	}

	/**
	 * Sets sub-component fieldUnit.
	 * @param fieldUnit New value of fieldUnit
	 */
	public void setFieldUnit(LyFieldUnit fieldUnit)
	{
		if (_fieldUnit != null)
			_fieldUnit.setParent(null);
		_fieldUnit = fieldUnit;
		if (fieldUnit != null)
			fieldUnit.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<FIELD");
		pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && !_editable)
				|| ((_template != null) && (_editable != _template._editable)))
			pw.print(" editable='" + _editable + "'");
		if (((_template == null) && !_optional)
				|| ((_template != null) && (_optional != _template._optional)))
			pw.print(" optional='" + _optional + "'");
		if (((_template == null) && !_enabled)
				|| ((_template != null) && (_enabled != _template._enabled)))
			pw.print(" enabled='" + _enabled + "'");
		if (((_template == null) && !_valid)
				|| ((_template != null) && (_valid != _template._valid)))
			pw.print(" valid='" + _valid + "'");
		if (((_template == null) && _checkBox)
				|| ((_template != null) && (_checkBox != _template._checkBox)))
			pw.print(" checkBox='" + _checkBox + "'");
		if (((_template == null) && _checked)
				|| ((_template != null) && (_checked != _template._checked)))
			pw.print(" checked='" + _checked + "'");
		if (((_template == null) && _needCheck)
				|| ((_template != null) && (_needCheck != _template._needCheck)))
			pw.print(" needCheck='" + _needCheck + "'");
		if ((_template == null)
			|| (_state != _template._state))
		{
			if ((_state == STATE_NORMAL) && (_template != null))
				pw.print(" state='NORMAL'");
			else if (_state == STATE_ERROR)
				pw.print(" state='ERROR'");
			else if (_state == STATE_WARNING)
				pw.print(" state='WARNING'");
		}
		if (((_template == null) && (_message != null))
				|| ((_template != null) && (_message != _template._message)))
			pw.print(" message='" + LyString.getXmlString(_message) + "'");
		if (((_template == null) && _hidden)
				|| ((_template != null) && (_hidden != _template._hidden)))
			pw.print(" hidden='" + _hidden + "'");
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlString(_controller) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		if (_caption != null)
			_caption.writeXml(pw, debug);
		if (_fieldItem != null)
			_fieldItem.writeXml(pw, debug);
		if (_toolBar != null)
			_toolBar.writeXml(pw, debug);
		if (_constraint != null)
			_constraint.writeXml(pw, debug);
		if (_tip != null)
			_tip.writeXml(pw, debug);

		writeDatas(pw);
		if (_relation != null)
			_relation.writeXml(pw, debug);
		if (_fieldUnit != null)
			_fieldUnit.writeXml(pw, debug);

		pw.println("</FIELD>");
	}
}
