/*
 * LyFrameBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyFrameBase class is a generated Java implementation that corresponds to the element 
 * FRAME of the descriptive grammar of the views (views.dtd).
 * <B>&lt;FRAME&gt; : Main Window</B>
 * The root of an XML file describing a complete Application Engine view is an 
 * element of main window or dialog window type (see the next chapter).
 * @see leon.view.LyView
 * @see leon.view.LyLyFrame
 * See also tag <A href="../../conf/dtd/view/FRAME.html">FRAME</A> of grammar leon_views.dtd.
 */
public abstract class LyFrameBase extends LyView
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface,
				LyGetControllerInterface,
				LySetControllerInterface
{
	public static final String			__VERSION = "$Id$";


	/* Attributes */

	/**
	 * Identifier of the main window
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyFrame			_template;

	/**
	 * Title of the main window
	 */
	protected String			_label;

	/**
	 * Reference of the controller that manages the main window (i.e. handles user 
	 * demands and refresh content).
	 */
	protected String			_controller;

	/**
	 * Identifier internal to the application of the main window that is parent of the 
	 * current main window.
	 */
	protected String			_parentId;

	/**
	 * Initial width of the main window.
	 */
	protected int				_width;

	/**
	 * Initial height of the main window.
	 */
	protected int				_height;

	/**
	 * Indicates whether the main window can be resized or not.
	 */
	protected boolean			_resizable;

	/**
	 * Indicates whether the main window is decorated or not.
	 */
	protected boolean			_decorated;

	/**
	 * Minimum width of the main window.
	 */
	protected int				_minWidth;

	/**
	 * Minimum height of the main window.
	 */
	protected int				_minHeight;

	/**
	 * Initial position in abscissa of the main window.
	 */
	protected int				_x;

	/**
	 * Initial position in ordinate of the main window.
	 */
	protected int				_y;

	/**
	 * Name of the frame where the view is displayed (used with multi-frames viewers 
	 * like web viewers or eclipse viewer).
	 */
	protected String			_target;


	/* Sub components */

	/**
	 * Icon of the main window.
	 */
	protected LyIcon			_icon;

	/**
	 * Menu bar of the main window.
	 */
	protected LyMenuBar			_menuBar;

	/**
	 * Main area contained in the main window.
	 */
	protected LyArea			_area;

	/**
	 * Message bar of the main window.
	 */
	protected LyMessages		_messages;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_controller = null;
		_parentId = null;
		_width = getIntValue("FRAME.width", -1);
		_height = getIntValue("FRAME.height", -1);
		_resizable = getBooleanValue("FRAME.resizable", true);
		_decorated = getBooleanValue("FRAME.decorated", true);
		_minWidth = getIntValue("FRAME.minWidth", -1);
		_minHeight = getIntValue("FRAME.minHeight", -1);
		_x = getIntValue("FRAME.x", 0);
		_y = getIntValue("FRAME.y", 0);
		_target = null;

		// Initialize sub components
		_icon = null;
		_menuBar = null;
		_area = null;
		_messages = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyFrame	model = (LyFrame)component;

		_model = component;
		_label = model._label;
		_controller = model._controller;
		_parentId = model._parentId;
		_width = model._width;
		_height = model._height;
		_resizable = model._resizable;
		_decorated = model._decorated;
		_minWidth = model._minWidth;
		_minHeight = model._minHeight;
		_x = model._x;
		_y = model._y;
		_target = model._target;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		if (model._icon != null)
			addChild(model._icon.cloneComponent());
		if (model._menuBar != null)
			addChild(model._menuBar.cloneComponent());
		if (model._area != null)
			addChild(model._area.cloneComponent());
		if (model._messages != null)
			addChild(model._messages.cloneComponent());
		if (model._datas != null)
			_datas = new Hashtable<String, String>(model._datas);
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	protected LyFrameBase()
	{
		_id = getInternalId();
	}

	/**
	 * Constructor with all required attributes.
	 * @param id Identifier of the main window
	 */
	public LyFrameBase(String id)
	{
		// Fill up required attributes with parameters
		_id = (id != null ? id.intern() : null);
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyFrameBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyFrame)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("controller"))
			_controller = node.getString("controller");
		if (node.isSpecified("parentId"))
			_parentId = node.getString("parentId");
		if (node.isSpecified("width"))
			_width = node.getInt("width");
		if (node.isSpecified("height"))
			_height = node.getInt("height");
		if (node.isSpecified("resizable"))
			_resizable = node.getBoolean("resizable");
		if (node.isSpecified("decorated"))
			_decorated = node.getBoolean("decorated");
		if (node.isSpecified("minWidth"))
			_minWidth = node.getInt("minWidth");
		if (node.isSpecified("minHeight"))
			_minHeight = node.getInt("minHeight");
		if (node.isSpecified("x"))
			_x = node.getInt("x");
		if (node.isSpecified("y"))
			_y = node.getInt("y");
		if (node.isSpecified("target"))
			_target = node.getString("target");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyIcon)
			setIcon((LyIcon)child);
		else if (child instanceof LyMenuBar)
			setMenuBar((LyMenuBar)child);
		else if (child instanceof LyArea)
			setArea((LyArea)child);
		else if (child instanceof LyMessages)
			setMessages((LyMessages)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyFrame clone = new LyFrame();

		clone.initFromModel(this);
		clone._template = (LyFrame)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyFrame getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of controller.
	 * @return value of controller.
	 */
	public String getController()
	{
		return _controller;
	}

	/**
	 * Gets value of parentId.
	 * @return value of parentId.
	 */
	public String getParentId()
	{
		return _parentId;
	}

	/**
	 * Gets value of width.
	 * @return value of width.
	 */
	public int getWidth()
	{
		return _width;
	}

	/**
	 * Gets value of height.
	 * @return value of height.
	 */
	public int getHeight()
	{
		return _height;
	}

	/**
	 * Checks if value of attribute resizable is set.
	 * @return true if resizable is set, false otherwise.
	 */
	public boolean isResizable()
	{
		return _resizable;
	}

	/**
	 * Checks if value of attribute decorated is set.
	 * @return true if decorated is set, false otherwise.
	 */
	public boolean isDecorated()
	{
		return _decorated;
	}

	/**
	 * Gets value of minWidth.
	 * @return value of minWidth.
	 */
	public int getMinWidth()
	{
		return _minWidth;
	}

	/**
	 * Gets value of minHeight.
	 * @return value of minHeight.
	 */
	public int getMinHeight()
	{
		return _minHeight;
	}

	/**
	 * Gets value of x.
	 * @return value of x.
	 */
	public int getX()
	{
		return _x;
	}

	/**
	 * Gets value of y.
	 * @return value of y.
	 */
	public int getY()
	{
		return _y;
	}

	/**
	 * Gets value of target.
	 * @return value of target.
	 */
	public String getTarget()
	{
		return _target;
	}

	/**
	 * Get sub-component Icon.
	 * @return sub-component Icon.
	 */
	public LyIcon getIcon()
	{
		return _icon;
	}

	/**
	 * Get sub-component MenuBar.
	 * @return sub-component MenuBar.
	 */
	public LyMenuBar getMenuBar()
	{
		return _menuBar;
	}

	/**
	 * Get sub-component Area.
	 * @return sub-component Area.
	 */
	public LyArea getArea()
	{
		return _area;
	}

	/**
	 * Get sub-component Messages.
	 * @return sub-component Messages.
	 */
	public LyMessages getMessages()
	{
		return _messages;
	}


	/* Modifiers */

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyFrame template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView != null)
			((LyGraphicalFrame)_internalView).setLabel(label);
	}

	/**
	 * Initializes attribute controller to the given value.
	 * @param controller New value of controller.
	 */
	public void setController(String controller)
	{
		if ((controller == _controller)
			|| ((controller != null)
				&& (_controller != null)
				&& controller.equals(_controller)))
			return;

		_controller = controller;
	}

	/**
	 * Initializes attribute parentId to the given value.
	 * @param parentId New value of parentId.
	 */
	public void setParentId(String parentId)
	{
		if ((parentId == _parentId)
			|| ((parentId != null)
				&& (_parentId != null)
				&& parentId.equals(_parentId)))
			return;

		_parentId = parentId;
	}

	/**
	 * Initializes attribute width to the given value.
	 * @param width New value of width.
	 */
	public void setWidth(int width)
	{
		if (width == _width)
			return;
		_width = width;
	}

	/**
	 * Initializes attribute height to the given value.
	 * @param height New value of height.
	 */
	public void setHeight(int height)
	{
		if (height == _height)
			return;
		_height = height;
	}

	/**
	 * Initializes attribute resizable to the given value.
	 * @param resizable Boolean indicating if attribute resizable is set or unset.
	 */
	public void setResizable(boolean resizable)
	{
		if (resizable == _resizable)
			return;
		_resizable = resizable;
	}

	/**
	 * Initializes attribute decorated to the given value.
	 * @param decorated Boolean indicating if attribute decorated is set or unset.
	 */
	public void setDecorated(boolean decorated)
	{
		if (decorated == _decorated)
			return;
		_decorated = decorated;
	}

	/**
	 * Initializes attribute minWidth to the given value.
	 * @param minWidth New value of minWidth.
	 */
	public void setMinWidth(int minWidth)
	{
		if (minWidth == _minWidth)
			return;
		_minWidth = minWidth;
	}

	/**
	 * Initializes attribute minHeight to the given value.
	 * @param minHeight New value of minHeight.
	 */
	public void setMinHeight(int minHeight)
	{
		if (minHeight == _minHeight)
			return;
		_minHeight = minHeight;
	}

	/**
	 * Initializes attribute x to the given value.
	 * @param x New value of x.
	 */
	public void setX(int x)
	{
		if (x == _x)
			return;
		_x = x;
	}

	/**
	 * Initializes attribute y to the given value.
	 * @param y New value of y.
	 */
	public void setY(int y)
	{
		if (y == _y)
			return;
		_y = y;
	}

	/**
	 * Initializes attribute target to the given value.
	 * @param target New value of target.
	 */
	public void setTarget(String target)
	{
		if ((target == _target)
			|| ((target != null)
				&& (_target != null)
				&& target.equals(_target)))
			return;

		_target = target;
	}

	/**
	 * Sets sub-component icon.
	 * @param icon New value of icon
	 */
	public void setIcon(LyIcon icon)
	{
		if (_icon != null)
			_icon.setParent(null);
		_icon = icon;
		if (icon != null)
			icon.setParent(this);
	}

	/**
	 * Sets sub-component menuBar.
	 * @param menuBar New value of menuBar
	 */
	public void setMenuBar(LyMenuBar menuBar)
	{
		if (_menuBar != null)
			_menuBar.setParent(null);
		_menuBar = menuBar;
		if (menuBar != null)
			menuBar.setParent(this);
	}

	/**
	 * Sets sub-component area.
	 * @param area New value of area
	 */
	public void setArea(LyArea area)
	{
		if (_area != null)
			_area.setParent(null);
		_area = area;
		if (area != null)
			area.setParent(this);
	}

	/**
	 * Sets sub-component messages.
	 * @param messages New value of messages
	 */
	public void setMessages(LyMessages messages)
	{
		if (_messages != null)
			_messages.setParent(null);
		_messages = messages;
		if (messages != null)
			messages.setParent(this);
		if (_internalView != null)
			((LyGraphicalFrame)_internalView).setMessages(messages);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<FRAME");
		pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlString(_controller) + "'");
		if (((_template == null) && (_parentId != null))
				|| ((_template != null) && (_parentId != _template._parentId)))
			pw.print(" parentId='" + LyString.getXmlString(_parentId) + "'");
		if (((_template == null) && (_width != getIntValue("FRAME.width", -1)))
				|| ((_template != null) && (_width != _template._width)))
			pw.print(" width='" + _width + "'");
		if (((_template == null) && (_height != getIntValue("FRAME.height", -1)))
				|| ((_template != null) && (_height != _template._height)))
			pw.print(" height='" + _height + "'");
		if (((_template == null) && !_resizable)
				|| ((_template != null) && (_resizable != _template._resizable)))
			pw.print(" resizable='" + _resizable + "'");
		if (((_template == null) && !_decorated)
				|| ((_template != null) && (_decorated != _template._decorated)))
			pw.print(" decorated='" + _decorated + "'");
		if (((_template == null) && (_minWidth != getIntValue("FRAME.minWidth", -1)))
				|| ((_template != null) && (_minWidth != _template._minWidth)))
			pw.print(" minWidth='" + _minWidth + "'");
		if (((_template == null) && (_minHeight != getIntValue("FRAME.minHeight", -1)))
				|| ((_template != null) && (_minHeight != _template._minHeight)))
			pw.print(" minHeight='" + _minHeight + "'");
		if (((_template == null) && (_x != getIntValue("FRAME.x", 0)))
				|| ((_template != null) && (_x != _template._x)))
			pw.print(" x='" + _x + "'");
		if (((_template == null) && (_y != getIntValue("FRAME.y", 0)))
				|| ((_template != null) && (_y != _template._y)))
			pw.print(" y='" + _y + "'");
		if (((_template == null) && (_target != null))
				|| ((_template != null) && (_target != _template._target)))
			pw.print(" target='" + LyString.getXmlString(_target) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		if (_icon != null)
			_icon.writeXml(pw, debug);
		if (_menuBar != null)
			_menuBar.writeXml(pw, debug);
		if (_area != null)
			_area.writeXml(pw, debug);
		if (_messages != null)
			_messages.writeXml(pw, debug);

		writeDatas(pw);

		pw.println("</FRAME>");
	}
}
