/*
 * LyHeader.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.misc.LyString;


/**
 * The LyHeader class is a generated Java implementation that corresponds to the element 
 * HEADER of the descriptive grammar of the views (views.dtd).
 * <B>&lt;HEADER&gt; : Header cell</B>
 * A header is the top label of column in a table.
 * @see leon.view.LyComponent
 * See also tag <A href="../../conf/dtd/view/HEADER.html">HEADER</A> of grammar leon_views.dtd.
 */
public class LyHeader extends LyComponent
	implements  LyGetIdInterface,
				LyGetLabelInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'align'. */
	public enum Align
	{
		LEFT,
		RIGHT,
		CENTER;
	}


	/* Static methods */

	/**
	 * Utility method to get align value from a string.
	 * @param token Token string
	 * @return Align value corresponding to this token
	 */
	public static final Align getAlign(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("LEFT")
				|| token.equals("0"))
			return Align.LEFT;
		else if (token.equalsIgnoreCase("RIGHT")
				|| token.equals("1"))
			return Align.RIGHT;
		else if (token.equalsIgnoreCase("CENTER")
				|| token.equals("2"))
			return Align.CENTER;
		else
			return Align.LEFT;
	}


	/* Attributes */

	/**
	 * Identifier of the header.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyHeader			_template;

	/**
	 * Label of the header.
	 */
	protected String			_label;

	/**
	 * Image of the header.
	 */
	protected String			_image;

	/**
	 * Sorting on the column corresponding to the cell.
	 * The expected value is a positive integer value or a null.
	 * If a sorting on the table is performed over several columns, the provided value 
	 * indicates the order in which the sorting is made (0 for the first column that is
	 * used for the sorting, 1 for the second one, etc.).
	 * This value is used only if the value of the ?sortable? attribute is true.
	 */
	protected int				_sort;

	/**
	 * Indicates whether the sorting is in ascending or descending order (true for the 
	 * ascending order).
	 */
	protected boolean			_increasing;

	/**
	 * Indicates if the column corresponding to the header can be sorted.
	 */
	protected boolean			_sortable;

	/**
	 * Indicates if the column corresponding to the header can be resized.
	 */
	protected boolean			_resizable;

	/**
	 * Indicates if the column corresponding to the header is fixed.
	 * A fixed column is not resized when using scroll bars.
	 */
	protected boolean			_fixed;

	/**
	 * Indicates if the label of the header can be wrapped over several lines.
	 * The true value indicates that the text can be wrapped.
	 */
	protected boolean			_wrap;

	/**
	 * HEADER.hidden
	 */
	protected boolean			_hidden;

	/**
	 * Alignment of the labels of the cells of the column corresponding to the header 
	 * (left, right or centered).
	 */
	protected Align				_align;

	/**
	 * Text of the tooltip for the header.
	 */
	protected String			_tooltip;

	/**
	 * Overview label for the header.
	 */
	protected String			_overview;

	/**
	 * Width of the column associated to the header in number of characters.
	 */
	protected double			_columnWidth;

	/**
	 * Width of the border associated to the header in number of pixels.
	 */
	protected int				_borderWidth;

	/**
	 * Color of vertical border corresponding to the header.
	 */
	protected String			_borderColor;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_image = getStringValue("HEADER.image", "$ALL.image");
		_sort = getIntValue("HEADER.sort", -1);
		_increasing = getBooleanValue("HEADER.increasing", false);
		_sortable = getBooleanValue("HEADER.sortable", true);
		_resizable = getBooleanValue("HEADER.resizable", true);
		_fixed = getBooleanValue("HEADER.fixed", false);
		_wrap = getBooleanValue("HEADER.wrap", false);
		_hidden = getBooleanValue("HEADER.hidden", false);
		_align = Align.LEFT;
		_tooltip = null;
		_overview = null;
		_columnWidth = getDoubleValue("HEADER.columnWidth", 0F);
		_borderWidth = getIntValue("HEADER.borderWidth", -1);
		_borderColor = getStringValue("HEADER.borderColor", "$ALL.borderColor");
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyHeader	model = (LyHeader)component;

		_model = component;
		_label = model._label;
		_image = model._image;
		_sort = model._sort;
		_increasing = model._increasing;
		_sortable = model._sortable;
		_resizable = model._resizable;
		_fixed = model._fixed;
		_wrap = model._wrap;
		_hidden = model._hidden;
		_align = model._align;
		_tooltip = model._tooltip;
		_overview = model._overview;
		_columnWidth = model._columnWidth;
		_borderWidth = model._borderWidth;
		_borderColor = model._borderColor;
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyHeader()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyHeader(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyHeader)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("image"))
			_image = node.getString("image");
		if (node.isSpecified("sort"))
			_sort = node.getInt("sort");
		if (node.isSpecified("increasing"))
			_increasing = node.getBoolean("increasing");
		if (node.isSpecified("sortable"))
			_sortable = node.getBoolean("sortable");
		if (node.isSpecified("resizable"))
			_resizable = node.getBoolean("resizable");
		if (node.isSpecified("fixed"))
			_fixed = node.getBoolean("fixed");
		if (node.isSpecified("wrap"))
			_wrap = node.getBoolean("wrap");
		if (node.isSpecified("hidden"))
			_hidden = node.getBoolean("hidden");
		if (node.isSpecified("align"))
			_align = getAlign(node.getString("align"));
		if (node.isSpecified("tooltip"))
			_tooltip = node.getString("tooltip");
		if (node.isSpecified("overview"))
			_overview = node.getString("overview");
		if (node.isSpecified("columnWidth"))
			_columnWidth = node.getDouble("columnWidth");
		if (node.isSpecified("borderWidth"))
			_borderWidth = node.getInt("borderWidth");
		if (node.isSpecified("borderColor"))
			_borderColor = node.getString("borderColor");
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyHeader clone = new LyHeader();

		clone.initFromModel(this);
		clone._template = this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyHeader getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of image.
	 * @return value of image.
	 */
	public String getImage()
	{
		return _image;
	}

	/**
	 * Checks whether the image attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultImage()
	{
		return (_image == getStringValue("HEADER.image"));
	}

	/**
	 * Gets value of sort.
	 * @return value of sort.
	 */
	public int getSort()
	{
		return _sort;
	}

	/**
	 * Checks if value of attribute increasing is set.
	 * @return true if increasing is set, false otherwise.
	 */
	public boolean isIncreasing()
	{
		return _increasing;
	}

	/**
	 * Checks if value of attribute sortable is set.
	 * @return true if sortable is set, false otherwise.
	 */
	public boolean isSortable()
	{
		return _sortable;
	}

	/**
	 * Checks if value of attribute resizable is set.
	 * @return true if resizable is set, false otherwise.
	 */
	public boolean isResizable()
	{
		return _resizable;
	}

	/**
	 * Checks if value of attribute fixed is set.
	 * @return true if fixed is set, false otherwise.
	 */
	public boolean isFixed()
	{
		return _fixed;
	}

	/**
	 * Checks if value of attribute wrap is set.
	 * @return true if wrap is set, false otherwise.
	 */
	public boolean isWrap()
	{
		return _wrap;
	}

	/**
	 * Checks if value of attribute hidden is set.
	 * @return true if hidden is set, false otherwise.
	 */
	public boolean isHidden()
	{
		return _hidden;
	}

	/**
	 * Gets value of align.
	 * @return value of align.
	 */
	public Align getAlign()
	{
		return _align;
	}

	/**
	 * Gets value of tooltip.
	 * @return value of tooltip.
	 */
	public String getTooltip()
	{
		return _tooltip;
	}

	/**
	 * Gets value of overview.
	 * @return value of overview.
	 */
	public String getOverview()
	{
		return _overview;
	}

	/**
	 * Gets value of columnWidth.
	 * @return value of columnWidth.
	 */
	public double getColumnWidth()
	{
		return _columnWidth;
	}

	/**
	 * Gets value of borderWidth.
	 * @return value of borderWidth.
	 */
	public int getBorderWidth()
	{
		return _borderWidth;
	}

	/**
	 * Gets value of borderColor.
	 * @return value of borderColor.
	 */
	public String getBorderColor()
	{
		return _borderColor;
	}

	/**
	 * Checks whether the borderColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultBorderColor()
	{
		return (_borderColor == getStringValue("HEADER.borderColor"));
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyHeader template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Initializes attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
	}

	/**
	 * Initializes attribute image to the given value.
	 * @param image New value of image.
	 */
	public void setImage(String image)
	{
		if ((image == _image)
			|| ((image != null)
				&& (_image != null)
				&& image.equals(_image)))
			return;

		_image = image;
	}

	/**
	 * Sets attribute sort to the given value.
	 * @param sort New value of sort.
	 */
	public void setSort(int sort)
	{
		if (sort == _sort)
			return;
		_sort = sort;
		if (_internalView != null)
			((LyGraphicalHeader)_internalView).setSort(sort);
	}

	/**
	 * Sets attribute increasing to the given value.
	 * @param increasing Boolean indicating if attribute increasing is set or unset.
	 */
	public void setIncreasing(boolean increasing)
	{
		if (increasing == _increasing)
			return;
		_increasing = increasing;
		if (_internalView != null)
			((LyGraphicalHeader)_internalView).setIncreasing(increasing);
	}

	/**
	 * Initializes attribute sortable to the given value.
	 * @param sortable Boolean indicating if attribute sortable is set or unset.
	 */
	public void setSortable(boolean sortable)
	{
		if (sortable == _sortable)
			return;
		_sortable = sortable;
	}

	/**
	 * Initializes attribute resizable to the given value.
	 * @param resizable Boolean indicating if attribute resizable is set or unset.
	 */
	public void setResizable(boolean resizable)
	{
		if (resizable == _resizable)
			return;
		_resizable = resizable;
	}

	/**
	 * Sets attribute fixed to the given value.
	 * @param fixed Boolean indicating if attribute fixed is set or unset.
	 */
	public void setFixed(boolean fixed)
	{
		if (fixed == _fixed)
			return;
		_fixed = fixed;
		if (_internalView != null)
			((LyGraphicalHeader)_internalView).setFixed(fixed);
	}

	/**
	 * Initializes attribute wrap to the given value.
	 * @param wrap Boolean indicating if attribute wrap is set or unset.
	 */
	public void setWrap(boolean wrap)
	{
		if (wrap == _wrap)
			return;
		_wrap = wrap;
	}

	/**
	 * Initializes attribute hidden to the given value.
	 * @param hidden Boolean indicating if attribute hidden is set or unset.
	 */
	public void setHidden(boolean hidden)
	{
		if (hidden == _hidden)
			return;
		_hidden = hidden;
	}

	/**
	 * Initializes attribute align to the given value.
	 * @param align New value of align.
	 * @exception IllegalArgumentException if align parameter is null.
	 */
	public void setAlign(Align align)
	{
		if (align == _align)
			return;
		if (align == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute align");

		_align = align;
	}

	/**
	 * Initializes attribute tooltip to the given value.
	 * @param tooltip New value of tooltip.
	 */
	public void setTooltip(String tooltip)
	{
		if ((tooltip == _tooltip)
			|| ((tooltip != null)
				&& (_tooltip != null)
				&& tooltip.equals(_tooltip)))
			return;

		_tooltip = tooltip;
	}

	/**
	 * Initializes attribute overview to the given value.
	 * @param overview New value of overview.
	 */
	public void setOverview(String overview)
	{
		if ((overview == _overview)
			|| ((overview != null)
				&& (_overview != null)
				&& overview.equals(_overview)))
			return;

		_overview = overview;
	}

	/**
	 * Sets attribute columnWidth to the given value.
	 * @param columnWidth New value of columnWidth.
	 */
	public void setColumnWidth(double columnWidth)
	{
		if (columnWidth == _columnWidth)
			return;
		_columnWidth = columnWidth;
		if (_internalView != null)
			((LyGraphicalHeader)_internalView).setColumnWidth(columnWidth);
	}

	/**
	 * Sets attribute borderWidth to the given value.
	 * @param borderWidth New value of borderWidth.
	 */
	public void setBorderWidth(int borderWidth)
	{
		if (borderWidth == _borderWidth)
			return;
		_borderWidth = borderWidth;
		if (_internalView != null)
			((LyGraphicalHeader)_internalView).setBorderWidth(borderWidth);
	}

	/**
	 * Sets attribute borderColor to the given value.
	 * @param borderColor New value of borderColor.
	 */
	public void setBorderColor(String borderColor)
	{
		if ((borderColor == _borderColor)
			|| ((borderColor != null)
				&& (_borderColor != null)
				&& borderColor.equals(_borderColor)))
			return;

		_borderColor = borderColor;
		if (_internalView != null)
			((LyGraphicalHeader)_internalView).setBorderColor(borderColor);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<HEADER");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (((_template == null) && (_image != getStringValue("HEADER.image", "$ALL.image")))
				|| ((_template != null) && (_image != _template._image)))
			pw.print(" image='" + LyString.getXmlString(_image) + "'");
		if (((_template == null) && (_sort != getIntValue("HEADER.sort", -1)))
				|| ((_template != null) && (_sort != _template._sort)))
			pw.print(" sort='" + _sort + "'");
		if (((_template == null) && _increasing)
				|| ((_template != null) && (_increasing != _template._increasing)))
			pw.print(" increasing='" + _increasing + "'");
		if (((_template == null) && !_sortable)
				|| ((_template != null) && (_sortable != _template._sortable)))
			pw.print(" sortable='" + _sortable + "'");
		if (((_template == null) && !_resizable)
				|| ((_template != null) && (_resizable != _template._resizable)))
			pw.print(" resizable='" + _resizable + "'");
		if (((_template == null) && _fixed)
				|| ((_template != null) && (_fixed != _template._fixed)))
			pw.print(" fixed='" + _fixed + "'");
		if (((_template == null) && _wrap)
				|| ((_template != null) && (_wrap != _template._wrap)))
			pw.print(" wrap='" + _wrap + "'");
		if (((_template == null) && _hidden)
				|| ((_template != null) && (_hidden != _template._hidden)))
			pw.print(" hidden='" + _hidden + "'");
		if ((_template == null)
			|| (_align != _template._align))
		{
			if ((_align == Align.LEFT) && (_template != null))
				pw.print(" align='LEFT'");
			else if (_align == Align.RIGHT)
				pw.print(" align='RIGHT'");
			else if (_align == Align.CENTER)
				pw.print(" align='CENTER'");
		}
		if (((_template == null) && (_tooltip != null))
				|| ((_template != null) && (_tooltip != _template._tooltip)))
			pw.print(" tooltip='" + LyString.getXmlString(_tooltip) + "'");
		if (((_template == null) && (_overview != null))
				|| ((_template != null) && (_overview != _template._overview)))
			pw.print(" overview='" + LyString.getXmlString(_overview) + "'");
		if (((_template == null) && (_columnWidth != getDoubleValue("HEADER.columnWidth", 0F)))
				|| ((_template != null) && (_columnWidth != _template._columnWidth)))
			pw.print(" columnWidth='" + _columnWidth + "'");
		if (((_template == null) && (_borderWidth != getIntValue("HEADER.borderWidth", -1)))
				|| ((_template != null) && (_borderWidth != _template._borderWidth)))
			pw.print(" borderWidth='" + _borderWidth + "'");
		if (((_template == null) && (_borderColor != getStringValue("HEADER.borderColor", "$ALL.borderColor")))
				|| ((_template != null) && (_borderColor != _template._borderColor)))
			pw.print(" borderColor='" + LyString.getXmlString(_borderColor) + "'");
		pw.println("/>");
	}
}
