/*
 * LyHeaders.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved. Current Version: $Id: LyHeaders.java,v 6.8 2008/04/08
 * 15:52:51 nha Exp $.
 */

package leon.view;

import java.util.*;


/**
 * The LyHeaders class implements generic behavior for the HEADERS components as described in
 * views.dtd regardless of the requested target (win, swing, html...). It represents headers.
 *
 * @see  leon.view.LyComponent
 */
public class LyHeaders extends LyHeadersBase
{
	public static final String __VERSION = "$Id: LyHeaders.java 32494 2011-09-23 11:47:57Z jlsaouli $";

	/** List of sub-header. */
	protected Vector<LyHeader> _subHeader;

	/** List of sub-headers. */
	protected Vector<LyHeaders> _subHeaders;

	/** Indicates if headers is hierarchical. */
	protected boolean _hierarchical;

	/** Indicates if headers may be reordered. */
	protected boolean _reorderingAllowed;

	/**
	 * Initialisation of the data structure.
	 *
	 * @see  leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();
		_subHeader = new Vector<LyHeader>();
		_subHeaders = new Vector<LyHeaders>();
		_hierarchical = false;
		_reorderingAllowed = true;
	}

	/**
	 * Constructor with all required attributes.
	 */
	public LyHeaders()
	{
		super();
	}

	/* Protected methods for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 *
	 * @param  node  Node from which the component is constructed.
	 *
	 * @see    leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyHeaders(LyViewNode node)
	{
		super(node);
	}

	/**
	 * Inserts a sub-component header at given position.
	 *
	 * @param  header  The header to add to components list.
	 * @param  pos     Position where the sub-component is inserted.
	 */
	public void insertHeader(LyHeader header, int pos)
	{
		super.insertHeader(header, pos);

		updateHeaderList();

		Vector<LyHeaders> topHeadersList = getParentHeadersList();

		if ((topHeadersList != null) && (topHeadersList.size() > 0))
		{
			int		  i, n = topHeadersList.size();
			LyHeaders topHeaders;

			for (i = 0; i < n; i++)
			{
				topHeaders = topHeadersList.elementAt(i);
				topHeaders.updateHeaderList();
			}
		}
	}

	/**
	 * Update the list of sub-header.
	 */
	protected void updateHeaderList()
	{
		_subHeader = new Vector<LyHeader>();
		getComponents(LyHeader.class, _subHeader);
	}

	/**
	 * Inserts a sub-component headers at given position.
	 *
	 * @param  headers  The headers to add to components list.
	 * @param  pos      Position where the sub-component is inserted.
	 */
	public void insertHeaders(LyHeaders headers, int pos)
	{
		super.insertHeaders(headers, pos);

		_hierarchical = true;

		updateHeaderList();
		updateHeadersList();

		Vector<LyHeaders> topHeadersList = getParentHeadersList();

		if ((topHeadersList != null) && (topHeadersList.size() > 0))
		{
			int		  i, n = topHeadersList.size();
			LyHeaders topHeaders;

			for (i = 0; i < n; i++)
			{
				topHeaders = topHeadersList.elementAt(i);
				topHeaders.updateHeadersList();
			}
		}
	}

	/**
	 * Update the list of sub-headers.
	 */
	protected void updateHeadersList()
	{
		_subHeaders = new Vector<LyHeaders>();
		getComponents(LyHeaders.class, _subHeaders);
	}

	/**
	 * Return the parent headers list.
	 *
	 * @return  the parent headers list
	 */
	public Vector<LyHeaders> getParentHeadersList()
	{
		LyComponent		  parent = getParent();
		Vector<LyHeaders> list = new Vector<LyHeaders>();

		while ((parent != null)
				&& (parent instanceof LyHeaders))
		{
			list.addElement((LyHeaders)parent);
			parent = parent.getParent();
		}

		return list;
	}

	/**
	 * Return the parent table.
	 *
	 * @return  the parent table
	 */
	public LyTable getParentTable()
	{
		LyComponent parent = getParent();

		while (!(parent instanceof LyTable) && (parent != null))
			parent = parent.getParent();

		return (LyTable)parent;
	}

	/**
	 * Return sub-header at given position.
	 *
	 * @param   i  the position of the sub-header
	 *
	 * @return  the sub-header at the given position
	 */
	public LyHeader getHeader(int i)
	{
		if ((i < 0) || (i >= _subHeader.size()))
			return null;
		
		return _subHeader.elementAt(i);
	}

	/**
	 * Return sub-headers at given position.
	 *
	 * @param   i  the position of the sub-header
	 *
	 * @return  the sub-header at the given position
	 */
	public LyHeaders getHeaders(int i)
	{
		if ((i < 0) || (i >= _subHeaders.size()))
			return null;
		
		return _subHeaders.elementAt(i);
	}

	/**
	 * The total number of sub-header.
	 *
	 * @return  the total number of sub-header
	 */
	public int getHeaderCount()
	{
		return _subHeader.size();
	}

	/**
	 * The total number of sub-headers.
	 *
	 * @return  the total number of sub-headers
	 */
	public int getHeadersCount()
	{
		return _subHeaders.size();
	}

	/**
	 * Indicates if the headers component has sub-headers components.
	 *
	 * @return  true if headers is hierarchical, false otherwise
	 */
	public boolean isHierarchical()
	{
		return _hierarchical;
	}

	/**
	 * Sets whether headers may be reordered.
	 *
	 * @param  reorder  true if headers may be reordered, false otherwise.
	 */
	public void setReorderingAllowed(boolean reorder)
	{
		_reorderingAllowed = reorder;
	}

	/**
	 * Indicates whether headers may be reordered.
	 *
	 * @return  true if headers may be reordered, false otherwise.
	 */
	public boolean isReorderingAllowed()
	{
		return _reorderingAllowed;
	}
} // end class LyHeaders
