/*
 * LyHeadersBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.getinterface.*;
import leon.misc.LyString;


/**
 * The LyHeadersBase class is a generated Java implementation that corresponds to the element 
 * HEADERS of the descriptive grammar of the views (views.dtd).
 * <B>&lt;HEADERS&gt; : Headers group</B>
 * A headers allows displaying header at the top of a table. Headers element does 
 * not contain any element: it is only a non-empty list of headers or sub-headers.
 * @see leon.view.LyComponent
 * @see leon.view.LyLyHeaders
 * See also tag <A href="../../conf/dtd/view/HEADERS.html">HEADERS</A> of grammar leon_views.dtd.
 */
public abstract class LyHeadersBase extends LyComponent
	implements  LyGetIdInterface,
				LyGetLabelInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'align'. */


	/* Attributes */

	/**
	 * Identifier of the headers.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyHeaders			_template;

	/**
	 * Label of the headers.
	 */
	protected String			_label;

	/**
	 * Align of the label of the headers (left, right or center)
	 */
	protected LyHeader.Align			_align;

	/**
	 * Tooltip of the headers.
	 */
	protected String			_tooltip;

	/**
	 * Overview label of the headers.
	 */
	protected String			_overview;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Font used to display sorted headers.
	 */
	protected String			_sortableFont;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;


	/* Sub components */

	/**
	 * Components of headers group : either header or sub-groups of headers.
	 */
	protected ArrayList<LyComponent>		_components;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_align = LyHeader.Align.CENTER;
		_tooltip = null;
		_overview = null;
		_font = getStringValue("HEADERS.font", "$ALL.font");
		_sortableFont = getStringValue("HEADERS.sortableFont", "$ALL.sortableFont");
		_color = getStringValue("HEADERS.color", "$ALL.color");
		_fontColor = getStringValue("HEADERS.fontColor", "$ALL.fontColor");

		// Initialize sub components
		_components = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyHeaders	model = (LyHeaders)component;

		_model = component;
		_label = model._label;
		_align = model._align;
		_tooltip = model._tooltip;
		_overview = model._overview;
		_font = model._font;
		_sortableFont = model._sortableFont;
		_color = model._color;
		_fontColor = model._fontColor;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyComponent	components;
		int			i, n;

		n = (model._components == null ? 0 : model._components.size());
		for (i = 0; i < n; i++)
		{
			components = model._components.get(i);
			if (components != null)
				components = components.cloneComponent();
			addChild(components);
		}
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyHeadersBase()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyHeadersBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyHeaders)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("align"))
			_align = LyHeader.getAlign(node.getString("align"));
		if (node.isSpecified("tooltip"))
			_tooltip = node.getString("tooltip");
		if (node.isSpecified("overview"))
			_overview = node.getString("overview");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("sortableFont"))
			_sortableFont = node.getString("sortableFont");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyHeader)
			addHeader((LyHeader)child);
		else if (child instanceof LyHeaders)
			addHeaders((LyHeaders)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyHeaders clone = new LyHeaders();

		clone.initFromModel(this);
		clone._template = (LyHeaders)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyHeaders getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of align.
	 * @return value of align.
	 */
	public LyHeader.Align getAlign()
	{
		return _align;
	}

	/**
	 * Gets value of tooltip.
	 * @return value of tooltip.
	 */
	public String getTooltip()
	{
		return _tooltip;
	}

	/**
	 * Gets value of overview.
	 * @return value of overview.
	 */
	public String getOverview()
	{
		return _overview;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("HEADERS.font"));
	}

	/**
	 * Gets value of sortableFont.
	 * @return value of sortableFont.
	 */
	public String getSortableFont()
	{
		return _sortableFont;
	}

	/**
	 * Checks whether the sortableFont attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultSortableFont()
	{
		return (_sortableFont == getStringValue("HEADERS.sortableFont"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("HEADERS.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("HEADERS.fontColor"));
	}

	/**
	 * Get sub-component Component of index i.
	 * @param i the index of the Component to get.
	 * @return sub-component Component of index i.
	 */
	public LyComponent getComponent(int i)
	{
		if (_components == null)
			return null;
		if ((i < 0) || (i >= _components.size()))
			return null;
		return _components.get(i);
	}

	/**
	 * Get sub-component Component count.
	 * @return sub-component Component count.
	 */
	public int getComponentCount()
	{
		if (_components == null)
			return 0;
		return _components.size();
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyHeaders template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Initializes attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
	}

	/**
	 * Initializes attribute align to the given value.
	 * @param align New value of align.
	 * @exception IllegalArgumentException if align parameter is null.
	 */
	public void setAlign(LyHeader.Align align)
	{
		if (align == _align)
			return;
		if (align == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute align");

		_align = align;
	}

	/**
	 * Initializes attribute tooltip to the given value.
	 * @param tooltip New value of tooltip.
	 */
	public void setTooltip(String tooltip)
	{
		if ((tooltip == _tooltip)
			|| ((tooltip != null)
				&& (_tooltip != null)
				&& tooltip.equals(_tooltip)))
			return;

		_tooltip = tooltip;
	}

	/**
	 * Initializes attribute overview to the given value.
	 * @param overview New value of overview.
	 */
	public void setOverview(String overview)
	{
		if ((overview == _overview)
			|| ((overview != null)
				&& (_overview != null)
				&& overview.equals(_overview)))
			return;

		_overview = overview;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute sortableFont to the given value.
	 * @param sortableFont New value of sortableFont.
	 */
	public void setSortableFont(String sortableFont)
	{
		if ((sortableFont == _sortableFont)
			|| ((sortableFont != null)
				&& (_sortableFont != null)
				&& sortableFont.equals(_sortableFont)))
			return;

		_sortableFont = sortableFont;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Adds a sub-component header.
	 * @param header The header to add to components list.
	 */
	public void addHeader(LyHeader header)
	{
		insertHeader(header, -1);
	}

	/**
	 * Inserts a sub-component header at given position.
	 * @param header The header to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertHeader(LyHeader header, int pos)
	{
		if (_components == null)
		{
			_components = new ArrayList<LyComponent>(1);
			_children = _components;
		}
		else if (_components.contains(header))
			return;
		if (pos >= 0)
			_components.add(pos,header);
		if (header != null)
			header.setParent(this);
	}

	/**
	 * Adds a sub-component headers.
	 * @param headers The headers to add to components list.
	 */
	public void addHeaders(LyHeaders headers)
	{
		insertHeaders(headers, -1);
	}

	/**
	 * Inserts a sub-component headers at given position.
	 * @param headers The headers to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertHeaders(LyHeaders headers, int pos)
	{
		if (_components == null)
		{
			_components = new ArrayList<LyComponent>(1);
			_children = _components;
		}
		else if (_components.contains(headers))
			return;
		if (pos >= 0)
			_components.add(pos,headers);
		if (headers != null)
			headers.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<HEADERS");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if ((_template == null)
			|| (_align != _template._align))
		{
			if (_align == LyHeader.Align.LEFT)
				pw.print(" align='LEFT'");
			else if (_align == LyHeader.Align.RIGHT)
				pw.print(" align='RIGHT'");
			else if ((_align == LyHeader.Align.CENTER) && (_template != null))
				pw.print(" align='CENTER'");
		}
		if (((_template == null) && (_tooltip != null))
				|| ((_template != null) && (_tooltip != _template._tooltip)))
			pw.print(" tooltip='" + LyString.getXmlString(_tooltip) + "'");
		if (((_template == null) && (_overview != null))
				|| ((_template != null) && (_overview != _template._overview)))
			pw.print(" overview='" + LyString.getXmlString(_overview) + "'");
		if (((_template == null) && (_font != getStringValue("HEADERS.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_sortableFont != getStringValue("HEADERS.sortableFont", "$ALL.sortableFont")))
				|| ((_template != null) && (_sortableFont != _template._sortableFont)))
			pw.print(" sortableFont='" + LyString.getXmlString(_sortableFont) + "'");
		if (((_template == null) && (_color != getStringValue("HEADERS.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("HEADERS.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	components;
		int			i, n;

		n = (_components == null ? 0 : _components.size());
		for (i = 0; i < n; i++)
		{
			components = _components.get(i);
			components.writeXml(pw, debug);
		}


		pw.println("</HEADERS>");
	}
}
