/*
 * LyInterval.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyInterval class is a generated Java implementation that corresponds to the element 
 * INTERVAL of the descriptive grammar of the views (views.dtd).
 * <B>&lt;INTERVAL&gt; : Interval</B>
 * An interval is defined by a begin date and an end date. If the end date is not 
 * specified, the interval is replaced by a point (an image representing a date). 
 * An interval has a label and a color. It may be resizable (it is possible to 
 * change its begin and end dates), movable on its parent line and may have 
 * vertical lines on its extremities. It may also have dependencies to other 
 * intervals represented by arrows linking this interval with the pointed 
 * intervals. An interval may also contain periods allowing splitting it into 
 * several sub-intervals. These periods are described by a begin date and an end 
 * date.
 * @see leon.view.LyComponent
 * See also tag <A href="../../conf/dtd/view/INTERVAL.html">INTERVAL</A> of grammar leon_views.dtd.
 */
public class LyInterval extends LyComponent
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface,
				LySetFontInterface,
				LySetColorInterface
{
	public static final String			__VERSION = "$Id$";


	/* Attributes */

	/**
	 * Identifier of the interval.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyInterval			_template;

	/**
	 * Label of the interval.
	 */
	protected String			_label;

	/**
	 * Center label of the interval.
	 */
	protected String			_centerLabel;

	/**
	 * Tooltip
	 * of the interval.
	 */
	protected String			_tooltip;

	/**
	 * Font used for the label of the interval.
	 */
	protected String			_font;

	/**
	 * Background of the interval.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * Begin date of the interval (must match the format of the planning).
	 */
	protected String			_beginDate;

	/**
	 * End date of the interval (must match the format of the planning). If this date 
	 * is not specified, the interval is represented by a point.
	 */
	protected String			_endDate;

	/**
	 * Indicates if this interval is resizable from the left.
	 */
	protected boolean			_leftResizable;

	/**
	 * Indicates if this interval is resizable from the right.
	 */
	protected boolean			_rightResizable;

	/**
	 * Indicates if this interval is movable on its parent line.
	 */
	protected boolean			_movable;

	/**
	 * Indicates if vertical lines must be drawn at the extremities of the interval on 
	 * the planning (only one line if the interval is represented by a point).
	 */
	protected boolean			_drawVerticalLines;

	/**
	 * Color of vertical lines.
	 */
	protected String			_verticalLinesColor;

	/**
	 * List of identifiers of the intervals related to this interval.
	 */
	protected String			_dependencies;


	/* Sub components */

	/**
	 * List of periods.
	 */
	protected ArrayList<LyPeriod>		_periods;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_centerLabel = null;
		_tooltip = null;
		_font = getStringValue("INTERVAL.font", "$ALL.font");
		_color = getStringValue("INTERVAL.color", "$ALL.color");
		_fontColor = getStringValue("INTERVAL.fontColor", "$ALL.fontColor");
		_beginDate = null;
		_endDate = null;
		_leftResizable = getBooleanValue("INTERVAL.leftResizable", true);
		_rightResizable = getBooleanValue("INTERVAL.rightResizable", true);
		_movable = getBooleanValue("INTERVAL.movable", true);
		_drawVerticalLines = getBooleanValue("INTERVAL.drawVerticalLines", false);
		_verticalLinesColor = getStringValue("INTERVAL.verticalLinesColor", "$ALL.verticalLinesColor");
		_dependencies = null;

		// Initialize sub components
		_periods = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyInterval	model = (LyInterval)component;

		_model = component;
		_label = model._label;
		_centerLabel = model._centerLabel;
		_tooltip = model._tooltip;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;
		_beginDate = model._beginDate;
		_endDate = model._endDate;
		_leftResizable = model._leftResizable;
		_rightResizable = model._rightResizable;
		_movable = model._movable;
		_drawVerticalLines = model._drawVerticalLines;
		_verticalLinesColor = model._verticalLinesColor;
		_dependencies = model._dependencies;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyPeriod	periods;
		int			i, n;

		n = (model._periods == null ? 0 : model._periods.size());
		for (i = 0; i < n; i++)
		{
			periods = model._periods.get(i);
			if (periods != null)
				periods = (LyPeriod)periods.cloneComponent();
			addChild(periods);
		}
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyInterval()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyInterval(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyInterval)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("centerLabel"))
			_centerLabel = node.getString("centerLabel");
		if (node.isSpecified("tooltip"))
			_tooltip = node.getString("tooltip");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("beginDate"))
			_beginDate = node.getString("beginDate");
		if (node.isSpecified("endDate"))
			_endDate = node.getString("endDate");
		if (node.isSpecified("leftResizable"))
			_leftResizable = node.getBoolean("leftResizable");
		if (node.isSpecified("rightResizable"))
			_rightResizable = node.getBoolean("rightResizable");
		if (node.isSpecified("movable"))
			_movable = node.getBoolean("movable");
		if (node.isSpecified("drawVerticalLines"))
			_drawVerticalLines = node.getBoolean("drawVerticalLines");
		if (node.isSpecified("verticalLinesColor"))
			_verticalLinesColor = node.getString("verticalLinesColor");
		if (node.isSpecified("dependencies"))
			_dependencies = node.getString("dependencies");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyPeriod)
			addPeriod((LyPeriod)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyInterval clone = new LyInterval();

		clone.initFromModel(this);
		clone._template = this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyInterval getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of centerLabel.
	 * @return value of centerLabel.
	 */
	public String getCenterLabel()
	{
		return _centerLabel;
	}

	/**
	 * Gets value of tooltip.
	 * @return value of tooltip.
	 */
	public String getTooltip()
	{
		return _tooltip;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("INTERVAL.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("INTERVAL.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("INTERVAL.fontColor"));
	}

	/**
	 * Gets value of beginDate.
	 * @return value of beginDate.
	 */
	public String getBeginDate()
	{
		return _beginDate;
	}

	/**
	 * Gets value of endDate.
	 * @return value of endDate.
	 */
	public String getEndDate()
	{
		return _endDate;
	}

	/**
	 * Checks if value of attribute leftResizable is set.
	 * @return true if leftResizable is set, false otherwise.
	 */
	public boolean isLeftResizable()
	{
		return _leftResizable;
	}

	/**
	 * Checks if value of attribute rightResizable is set.
	 * @return true if rightResizable is set, false otherwise.
	 */
	public boolean isRightResizable()
	{
		return _rightResizable;
	}

	/**
	 * Checks if value of attribute movable is set.
	 * @return true if movable is set, false otherwise.
	 */
	public boolean isMovable()
	{
		return _movable;
	}

	/**
	 * Checks if value of attribute drawVerticalLines is set.
	 * @return true if drawVerticalLines is set, false otherwise.
	 */
	public boolean isDrawVerticalLines()
	{
		return _drawVerticalLines;
	}

	/**
	 * Gets value of verticalLinesColor.
	 * @return value of verticalLinesColor.
	 */
	public String getVerticalLinesColor()
	{
		return _verticalLinesColor;
	}

	/**
	 * Checks whether the verticalLinesColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultVerticalLinesColor()
	{
		return (_verticalLinesColor == getStringValue("INTERVAL.verticalLinesColor"));
	}

	/**
	 * Gets value of dependencies.
	 * @return value of dependencies.
	 */
	public String getDependencies()
	{
		return _dependencies;
	}

	/**
	 * Get sub-component Period of index i.
	 * @param i the index of the Period to get.
	 * @return sub-component Period of index i.
	 */
	public LyPeriod getPeriod(int i)
	{
		if (_periods == null)
			return null;
		if ((i < 0) || (i >= _periods.size()))
			return null;
		return _periods.get(i);
	}

	/**
	 * Get sub-component Period count.
	 * @return sub-component Period count.
	 */
	public int getPeriodCount()
	{
		if (_periods == null)
			return 0;
		return _periods.size();
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyInterval template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView != null)
			((LyGraphicalInterval)_internalView).setLabel(label);
	}

	/**
	 * Sets attribute centerLabel to the given value.
	 * @param centerLabel New value of centerLabel.
	 */
	public void setCenterLabel(String centerLabel)
	{
		if ((centerLabel == _centerLabel)
			|| ((centerLabel != null)
				&& (_centerLabel != null)
				&& centerLabel.equals(_centerLabel)))
			return;

		_centerLabel = centerLabel;
		if (_internalView != null)
			((LyGraphicalInterval)_internalView).setCenterLabel(centerLabel);
	}

	/**
	 * Sets attribute tooltip to the given value.
	 * @param tooltip New value of tooltip.
	 */
	public void setTooltip(String tooltip)
	{
		if ((tooltip == _tooltip)
			|| ((tooltip != null)
				&& (_tooltip != null)
				&& tooltip.equals(_tooltip)))
			return;

		_tooltip = tooltip;
		if (_internalView != null)
			((LyGraphicalInterval)_internalView).setTooltip(tooltip);
	}

	/**
	 * Sets attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
		if (_internalView != null)
			((LyGraphicalInterval)_internalView).setFont(font);
	}

	/**
	 * Sets attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
		if (_internalView != null)
			((LyGraphicalInterval)_internalView).setColor(color);
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Sets attribute beginDate to the given value.
	 * @param beginDate New value of beginDate.
	 */
	public void setBeginDate(String beginDate)
	{
		if ((beginDate == _beginDate)
			|| ((beginDate != null)
				&& (_beginDate != null)
				&& beginDate.equals(_beginDate)))
			return;

		_beginDate = beginDate;
		if (_internalView != null)
			((LyGraphicalInterval)_internalView).setBeginDate(beginDate);
	}

	/**
	 * Sets attribute endDate to the given value.
	 * @param endDate New value of endDate.
	 */
	public void setEndDate(String endDate)
	{
		if ((endDate == _endDate)
			|| ((endDate != null)
				&& (_endDate != null)
				&& endDate.equals(_endDate)))
			return;

		_endDate = endDate;
		if (_internalView != null)
			((LyGraphicalInterval)_internalView).setEndDate(endDate);
	}

	/**
	 * Sets attribute leftResizable to the given value.
	 * @param leftResizable Boolean indicating if attribute leftResizable is set or unset.
	 */
	public void setLeftResizable(boolean leftResizable)
	{
		if (leftResizable == _leftResizable)
			return;
		_leftResizable = leftResizable;
		if (_internalView != null)
			((LyGraphicalInterval)_internalView).setLeftResizable(leftResizable);
	}

	/**
	 * Sets attribute rightResizable to the given value.
	 * @param rightResizable Boolean indicating if attribute rightResizable is set or unset.
	 */
	public void setRightResizable(boolean rightResizable)
	{
		if (rightResizable == _rightResizable)
			return;
		_rightResizable = rightResizable;
		if (_internalView != null)
			((LyGraphicalInterval)_internalView).setRightResizable(rightResizable);
	}

	/**
	 * Sets attribute movable to the given value.
	 * @param movable Boolean indicating if attribute movable is set or unset.
	 */
	public void setMovable(boolean movable)
	{
		if (movable == _movable)
			return;
		_movable = movable;
		if (_internalView != null)
			((LyGraphicalInterval)_internalView).setMovable(movable);
	}

	/**
	 * Sets attribute drawVerticalLines to the given value.
	 * @param drawVerticalLines Boolean indicating if attribute drawVerticalLines is set or unset.
	 */
	public void setDrawVerticalLines(boolean drawVerticalLines)
	{
		if (drawVerticalLines == _drawVerticalLines)
			return;
		_drawVerticalLines = drawVerticalLines;
		if (_internalView != null)
			((LyGraphicalInterval)_internalView).setDrawVerticalLines(drawVerticalLines);
	}

	/**
	 * Sets attribute verticalLinesColor to the given value.
	 * @param verticalLinesColor New value of verticalLinesColor.
	 */
	public void setVerticalLinesColor(String verticalLinesColor)
	{
		if ((verticalLinesColor == _verticalLinesColor)
			|| ((verticalLinesColor != null)
				&& (_verticalLinesColor != null)
				&& verticalLinesColor.equals(_verticalLinesColor)))
			return;

		_verticalLinesColor = verticalLinesColor;
		if (_internalView != null)
			((LyGraphicalInterval)_internalView).setVerticalLinesColor(verticalLinesColor);
	}

	/**
	 * Sets attribute dependencies to the given value.
	 * @param dependencies New value of dependencies.
	 */
	public void setDependencies(String dependencies)
	{
		if ((dependencies == _dependencies)
			|| ((dependencies != null)
				&& (_dependencies != null)
				&& dependencies.equals(_dependencies)))
			return;

		_dependencies = dependencies;
		if (_internalView != null)
			((LyGraphicalInterval)_internalView).setDependencies(dependencies);
	}

	/**
	 * Adds a sub-component period.
	 * @param period The period to add to periods list.
	 */
	public void addPeriod(LyPeriod period)
	{
		insertPeriod(period, -1);
	}

	/**
	 * Inserts a sub-component period at given position.
	 * @param period The period to add to periods list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertPeriod(LyPeriod period, int pos)
	{
		if (_periods == null)
		{
			_periods = new ArrayList<LyPeriod>(1);
			_children = _periods;
		}
		else if (_periods.contains(period))
			return;
		if (pos >= 0)
			_periods.add(pos,period);
		if (period != null)
			period.setParent(this);
		if (_internalView != null)
			((LyGraphicalInterval)_internalView).addPeriod(period, pos);
	}

	/**
	 * Removes a sub-component period.
	 * @param period The period to remove to periods list.
	 * @exception IllegalStateException if periods list is empty.
	 * @exception IllegalArgumentException if periods list doesn't contain period.
	 */
	public void removePeriod(LyPeriod period)
	{
		if (_periods == null)
			throw new IllegalStateException("BF AppEngine : periods list is empty");
		if (!_periods.contains(period))
			throw new IllegalArgumentException("BF AppEngine : periods list doesn't contain period");

		if (_internalView != null)
			((LyGraphicalInterval)_internalView).removePeriod(period);
		period.setParent(null);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<INTERVAL");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (((_template == null) && (_centerLabel != null))
				|| ((_template != null) && (_centerLabel != _template._centerLabel)))
			pw.print(" centerLabel='" + LyString.getXmlString(_centerLabel) + "'");
		if (((_template == null) && (_tooltip != null))
				|| ((_template != null) && (_tooltip != _template._tooltip)))
			pw.print(" tooltip='" + LyString.getXmlString(_tooltip) + "'");
		if (((_template == null) && (_font != getStringValue("INTERVAL.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("INTERVAL.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("INTERVAL.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && (_beginDate != null))
				|| ((_template != null) && (_beginDate != _template._beginDate)))
			pw.print(" beginDate='" + LyString.getXmlString(_beginDate) + "'");
		if (((_template == null) && (_endDate != null))
				|| ((_template != null) && (_endDate != _template._endDate)))
			pw.print(" endDate='" + LyString.getXmlString(_endDate) + "'");
		if (((_template == null) && !_leftResizable)
				|| ((_template != null) && (_leftResizable != _template._leftResizable)))
			pw.print(" leftResizable='" + _leftResizable + "'");
		if (((_template == null) && !_rightResizable)
				|| ((_template != null) && (_rightResizable != _template._rightResizable)))
			pw.print(" rightResizable='" + _rightResizable + "'");
		if (((_template == null) && !_movable)
				|| ((_template != null) && (_movable != _template._movable)))
			pw.print(" movable='" + _movable + "'");
		if (((_template == null) && _drawVerticalLines)
				|| ((_template != null) && (_drawVerticalLines != _template._drawVerticalLines)))
			pw.print(" drawVerticalLines='" + _drawVerticalLines + "'");
		if (((_template == null) && (_verticalLinesColor != getStringValue("INTERVAL.verticalLinesColor", "$ALL.verticalLinesColor")))
				|| ((_template != null) && (_verticalLinesColor != _template._verticalLinesColor)))
			pw.print(" verticalLinesColor='" + LyString.getXmlString(_verticalLinesColor) + "'");
		if (((_template == null) && (_dependencies != null))
				|| ((_template != null) && (_dependencies != _template._dependencies)))
			pw.print(" dependencies='" + LyString.getXmlString(_dependencies) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	periods;
		int			i, n;

		n = (_periods == null ? 0 : _periods.size());
		for (i = 0; i < n; i++)
		{
			periods = _periods.get(i);
			periods.writeXml(pw, debug);
		}


		pw.println("</INTERVAL>");
	}
}
