/*
 * LyItem.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyItem class is a generated Java implementation that corresponds to the element 
 * ITEM of the descriptive grammar of the views (views.dtd).
 * <B>&lt;ITEM&gt; : Menu entry</B>
 * A menu entry allows executing an action, as when using an action button. The 
 * entry can be of check box type (checkbox attribute) and can open a dialog window
 * (dialog attribute).
 * @see leon.view.LyComponent
 * See also tag <A href="../../conf/dtd/view/ITEM.html">ITEM</A> of grammar leon_views.dtd.
 */
public class LyItem extends LyComponent
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface,
				LyGetControllerInterface,
				LySetControllerInterface,
				LySetEnabledInterface,
				LySetImageInterface
{
	public static final String			__VERSION = "$Id$";


	/* Attributes */

	/**
	 * Identifier of the menu entry.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyItem			_template;

	/**
	 * Text of the menu entry.
	 */
	protected String			_label;

	/**
	 * Keyboard shortcut of menu entry.
	 * The expected value is an alphanumeric character that is not case sensitive.
	 */
	protected String			_shortcut;

	/**
	 * Reference of the controller that manages the menu entry (i.e. handles user 
	 * demands and refresh content).
	 */
	protected String			_controller;

	/**
	 * Indicates whether the menu entry is enabled or not (true if it is enabled).
	 */
	protected boolean			_enabled;

	/**
	 * Indicates whether the menu entry is of check box type or not.
	 */
	protected boolean			_checkBox;

	/**
	 * Indicates whether the menu entry is selected or not.
	 * This value is used only if the entry is of check box type.
	 */
	protected boolean			_checked;

	/**
	 * Indicates whether the action generated by the menu entry opens a dialog window 
	 * or not.
	 * At the display time, suspension points are added if the attribute is set to 
	 * ?true?
	 */
	protected boolean			_dialog;

	/**
	 * Image associated to the menu item.
	 */
	protected String			_image;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_shortcut = null;
		_controller = null;
		_enabled = getBooleanValue("ITEM.enabled", true);
		_checkBox = getBooleanValue("ITEM.checkBox", false);
		_checked = getBooleanValue("ITEM.checked", false);
		_dialog = getBooleanValue("ITEM.dialog", false);
		_image = getStringValue("ITEM.image", "$ALL.image");
		_font = getStringValue("ITEM.font", "$ALL.font");
		_color = getStringValue("ITEM.color", "$ALL.color");
		_fontColor = getStringValue("ITEM.fontColor", "$ALL.fontColor");
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyItem	model = (LyItem)component;

		_model = component;
		_label = model._label;
		_shortcut = model._shortcut;
		_controller = model._controller;
		_enabled = model._enabled;
		_checkBox = model._checkBox;
		_checked = model._checked;
		_dialog = model._dialog;
		_image = model._image;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	protected LyItem()
	{
		_id = getInternalId();
	}

	/**
	 * Constructor with all required attributes.
	 * @param id Identifier of the menu entry.
	 * @param label Text of the menu entry.
	 */
	public LyItem(String id, String label)
	{
		// Fill up required attributes with parameters
		_id = (id != null ? id.intern() : null);
		_label = label;
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyItem(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyItem)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("shortcut"))
			_shortcut = node.getString("shortcut");
		if (node.isSpecified("controller"))
			_controller = node.getString("controller");
		if (node.isSpecified("enabled"))
			_enabled = node.getBoolean("enabled");
		if (node.isSpecified("checkBox"))
			_checkBox = node.getBoolean("checkBox");
		if (node.isSpecified("checked"))
			_checked = node.getBoolean("checked");
		if (node.isSpecified("dialog"))
			_dialog = node.getBoolean("dialog");
		if (node.isSpecified("image"))
			_image = node.getString("image");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyItem clone = new LyItem();

		clone.initFromModel(this);
		clone._template = this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyItem getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of shortcut.
	 * @return value of shortcut.
	 */
	public String getShortcut()
	{
		return _shortcut;
	}

	/**
	 * Gets value of controller.
	 * @return value of controller.
	 */
	public String getController()
	{
		return _controller;
	}

	/**
	 * Checks if value of attribute enabled is set.
	 * @return true if enabled is set, false otherwise.
	 */
	public boolean isEnabled()
	{
		return _enabled;
	}

	/**
	 * Checks if value of attribute checkBox is set.
	 * @return true if checkBox is set, false otherwise.
	 */
	public boolean isCheckBox()
	{
		return _checkBox;
	}

	/**
	 * Checks if value of attribute checked is set.
	 * @return true if checked is set, false otherwise.
	 */
	public boolean isChecked()
	{
		return _checked;
	}

	/**
	 * Checks if value of attribute dialog is set.
	 * @return true if dialog is set, false otherwise.
	 */
	public boolean isDialog()
	{
		return _dialog;
	}

	/**
	 * Gets value of image.
	 * @return value of image.
	 */
	public String getImage()
	{
		return _image;
	}

	/**
	 * Checks whether the image attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultImage()
	{
		return (_image == getStringValue("ITEM.image"));
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("ITEM.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("ITEM.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("ITEM.fontColor"));
	}


	/* Modifiers */

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyItem template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 * @exception IllegalArgumentException if label is null.
	 */
	public void setLabel(String label)
	{
		if (label == null)
			throw new IllegalArgumentException("BF AppEngine : attribute label is required (can't be null)");

		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView != null)
			((LyGraphicalItem)_internalView).setLabel(label);
	}

	/**
	 * Initializes attribute shortcut to the given value.
	 * @param shortcut New value of shortcut.
	 */
	public void setShortcut(String shortcut)
	{
		if ((shortcut == _shortcut)
			|| ((shortcut != null)
				&& (_shortcut != null)
				&& shortcut.equals(_shortcut)))
			return;

		_shortcut = shortcut;
	}

	/**
	 * Initializes attribute controller to the given value.
	 * @param controller New value of controller.
	 */
	public void setController(String controller)
	{
		if ((controller == _controller)
			|| ((controller != null)
				&& (_controller != null)
				&& controller.equals(_controller)))
			return;

		_controller = controller;
	}

	/**
	 * Sets attribute enabled to the given value.
	 * @param enabled Boolean indicating if attribute enabled is set or unset.
	 */
	public void setEnabled(boolean enabled)
	{
		if (enabled == _enabled)
			return;
		_enabled = enabled;
		if (_internalView != null)
			((LyGraphicalItem)_internalView).setEnabled(enabled);
	}

	/**
	 * Initializes attribute checkBox to the given value.
	 * @param checkBox Boolean indicating if attribute checkBox is set or unset.
	 */
	public void setCheckBox(boolean checkBox)
	{
		if (checkBox == _checkBox)
			return;
		_checkBox = checkBox;
	}

	/**
	 * Sets attribute checked to the given value.
	 * @param checked Boolean indicating if attribute checked is set or unset.
	 */
	public void setChecked(boolean checked)
	{
		if (checked == _checked)
			return;
		_checked = checked;
		if (_internalView != null)
			((LyGraphicalItem)_internalView).setChecked(checked);
	}

	/**
	 * Initializes attribute dialog to the given value.
	 * @param dialog Boolean indicating if attribute dialog is set or unset.
	 */
	public void setDialog(boolean dialog)
	{
		if (dialog == _dialog)
			return;
		_dialog = dialog;
	}

	/**
	 * Sets attribute image to the given value.
	 * @param image New value of image.
	 */
	public void setImage(String image)
	{
		if ((image == _image)
			|| ((image != null)
				&& (_image != null)
				&& image.equals(_image)))
			return;

		_image = image;
		if (_internalView != null)
			((LyGraphicalItem)_internalView).setImage(image);
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<ITEM");
		pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (((_template == null) && (_shortcut != null))
				|| ((_template != null) && (_shortcut != _template._shortcut)))
			pw.print(" shortcut='" + LyString.getXmlString(_shortcut) + "'");
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlString(_controller) + "'");
		if (((_template == null) && !_enabled)
				|| ((_template != null) && (_enabled != _template._enabled)))
			pw.print(" enabled='" + _enabled + "'");
		if (((_template == null) && _checkBox)
				|| ((_template != null) && (_checkBox != _template._checkBox)))
			pw.print(" checkBox='" + _checkBox + "'");
		if (((_template == null) && _checked)
				|| ((_template != null) && (_checked != _template._checked)))
			pw.print(" checked='" + _checked + "'");
		if (((_template == null) && _dialog)
				|| ((_template != null) && (_dialog != _template._dialog)))
			pw.print(" dialog='" + _dialog + "'");
		if (((_template == null) && (_image != getStringValue("ITEM.image", "$ALL.image")))
				|| ((_template != null) && (_image != _template._image)))
			pw.print(" image='" + LyString.getXmlString(_image) + "'");
		if (((_template == null) && (_font != getStringValue("ITEM.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("ITEM.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("ITEM.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		pw.println("/>");
	}
}
