/*
 * LyLabel.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;
import java.io.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyLabel class is a generated Java implementation that corresponds to the element 
 * LABEL of the descriptive grammar of the views (views.dtd).
 * <B>&lt;LABEL&gt; : Input Field of Label Type</B>
 * An input field of label type allows displaying a simple, non-editable label 
 * inside a form.
 * It is used in the consultation forms to show different types of fields (date, 
 * number, choice, etc.). It can be made up of one or several lines. It does not 
 * comprise any element.
 * @see leon.view.LyFieldItem
 * See also tag <A href="../../conf/dtd/view/LABEL.html">LABEL</A> of grammar leon_views.dtd.
 */
public class LyLabel extends LyFieldItem
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface,
				LySetImageInterface,
				LySetFontInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'hAlign'. */
	public enum HAlign
	{
		LEFT,
		CENTER,
		RIGHT;
	}


	/* Static methods */

	/**
	 * Utility method to get hAlign value from a string.
	 * @param token Token string
	 * @return HAlign value corresponding to this token
	 */
	public static final HAlign getHAlign(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("LEFT")
				|| token.equals("0"))
			return HAlign.LEFT;
		else if (token.equalsIgnoreCase("CENTER")
				|| token.equals("1"))
			return HAlign.CENTER;
		else if (token.equalsIgnoreCase("RIGHT")
				|| token.equals("2"))
			return HAlign.RIGHT;
		else
			return HAlign.LEFT;
	}


	/* Attributes */

	/**
	 * Identifier of the component.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyLabel			_template;

	/**
	 * Text of the label.
	 */
	protected String			_label;

	/**
	 * Image of the label.
	 */
	protected String			_image;

	/**
	 * Horizontal alignment of the label.
	 */
	protected HAlign			_hAlign;

	/**
	 * Allows specifying the name of a Java class that replaces the default graphical 
	 * component.
	 */
	protected String			_bean;

	/**
	 * Font of the label.
	 */
	protected String			_font;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * LABEL.highlightColor
	 */
	protected String			_highlightColor;

	/**
	 * Gives the height (number of columns) on which the label is shown (given in 
	 * number of characters).
	 */
	protected int				_nCols;

	/**
	 * Gives the width (number of rows) on which the label is shown (given in number of
	 * characters).
	 */
	protected int				_nRows;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_image = getStringValue("LABEL.image", "$ALL.image");
		_hAlign = HAlign.LEFT;
		_bean = null;
		_font = getStringValue("LABEL.font", "$ALL.font");
		_color = getStringValue("LABEL.color", "$ALL.color");
		_fontColor = getStringValue("LABEL.fontColor", "$ALL.fontColor");
		_highlightColor = getStringValue("LABEL.highlightColor", "$ALL.highlightColor");
		_nCols = getIntValue("LABEL.nCols", -1);
		_nRows = getIntValue("LABEL.nRows", -1);
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyLabel	model = (LyLabel)component;

		_model = component;
		_label = model._label;
		_image = model._image;
		_hAlign = model._hAlign;
		_bean = model._bean;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;
		_highlightColor = model._highlightColor;
		_nCols = model._nCols;
		_nRows = model._nRows;
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyLabel()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyLabel(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyLabel)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("image"))
			_image = node.getString("image");
		if (node.isSpecified("hAlign"))
			_hAlign = getHAlign(node.getString("hAlign"));
		if (node.isSpecified("bean"))
			_bean = node.getString("bean");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("highlightColor"))
			_highlightColor = node.getString("highlightColor");
		if (node.isSpecified("nCols"))
			_nCols = node.getInt("nCols");
		if (node.isSpecified("nRows"))
			_nRows = node.getInt("nRows");
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyLabel clone = new LyLabel();

		clone.initFromModel(this);
		clone._template = this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyLabel getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of image.
	 * @return value of image.
	 */
	public String getImage()
	{
		return _image;
	}

	/**
	 * Checks whether the image attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultImage()
	{
		return (_image == getStringValue("LABEL.image"));
	}

	/**
	 * Gets value of hAlign.
	 * @return value of hAlign.
	 */
	public HAlign getHAlign()
	{
		return _hAlign;
	}

	/**
	 * Gets value of bean.
	 * @return value of bean.
	 */
	public String getBean()
	{
		return _bean;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("LABEL.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("LABEL.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("LABEL.fontColor"));
	}

	/**
	 * Gets value of highlightColor.
	 * @return value of highlightColor.
	 */
	public String getHighlightColor()
	{
		return _highlightColor;
	}

	/**
	 * Checks whether the highlightColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultHighlightColor()
	{
		return (_highlightColor == getStringValue("LABEL.highlightColor"));
	}

	/**
	 * Gets value of nCols.
	 * @return value of nCols.
	 */
	public int getNCols()
	{
		return _nCols;
	}

	/**
	 * Gets value of nRows.
	 * @return value of nRows.
	 */
	public int getNRows()
	{
		return _nRows;
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyLabel template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView instanceof LySetLabelInterface)
			((LySetLabelInterface)_internalView).setLabel(label);
	}

	/**
	 * Sets attribute image to the given value.
	 * @param image New value of image.
	 */
	public void setImage(String image)
	{
		if ((image == _image)
			|| ((image != null)
				&& (_image != null)
				&& image.equals(_image)))
			return;

		_image = image;
		if (_internalView instanceof LySetImageInterface)
			((LySetImageInterface)_internalView).setImage(image);
	}

	/**
	 * Initializes attribute hAlign to the given value.
	 * @param hAlign New value of hAlign.
	 * @exception IllegalArgumentException if hAlign parameter is null.
	 */
	public void setHAlign(HAlign hAlign)
	{
		if (hAlign == _hAlign)
			return;
		if (hAlign == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute hAlign");

		_hAlign = hAlign;
	}

	/**
	 * Initializes attribute bean to the given value.
	 * @param bean New value of bean.
	 */
	public void setBean(String bean)
	{
		if ((bean == _bean)
			|| ((bean != null)
				&& (_bean != null)
				&& bean.equals(_bean)))
			return;

		_bean = bean;
	}

	/**
	 * Sets attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
		if (_internalView instanceof LySetFontInterface)
			((LySetFontInterface)_internalView).setFont(font);
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Initializes attribute highlightColor to the given value.
	 * @param highlightColor New value of highlightColor.
	 */
	public void setHighlightColor(String highlightColor)
	{
		if ((highlightColor == _highlightColor)
			|| ((highlightColor != null)
				&& (_highlightColor != null)
				&& highlightColor.equals(_highlightColor)))
			return;

		_highlightColor = highlightColor;
	}

	/**
	 * Initializes attribute nCols to the given value.
	 * @param nCols New value of nCols.
	 */
	public void setNCols(int nCols)
	{
		if (nCols == _nCols)
			return;
		_nCols = nCols;
	}

	/**
	 * Initializes attribute nRows to the given value.
	 * @param nRows New value of nRows.
	 */
	public void setNRows(int nRows)
	{
		if (nRows == _nRows)
			return;
		_nRows = nRows;
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<LABEL");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (((_template == null) && (_image != getStringValue("LABEL.image", "$ALL.image")))
				|| ((_template != null) && (_image != _template._image)))
			pw.print(" image='" + LyString.getXmlString(_image) + "'");
		if ((_template == null)
			|| (_hAlign != _template._hAlign))
		{
			if ((_hAlign == HAlign.LEFT) && (_template != null))
				pw.print(" hAlign='LEFT'");
			else if (_hAlign == HAlign.CENTER)
				pw.print(" hAlign='CENTER'");
			else if (_hAlign == HAlign.RIGHT)
				pw.print(" hAlign='RIGHT'");
		}
		if (((_template == null) && (_bean != null))
				|| ((_template != null) && (_bean != _template._bean)))
			pw.print(" bean='" + LyString.getXmlString(_bean) + "'");
		if (((_template == null) && (_font != getStringValue("LABEL.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("LABEL.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("LABEL.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && (_highlightColor != getStringValue("LABEL.highlightColor", "$ALL.highlightColor")))
				|| ((_template != null) && (_highlightColor != _template._highlightColor)))
			pw.print(" highlightColor='" + LyString.getXmlString(_highlightColor) + "'");
		if (((_template == null) && (_nCols != getIntValue("LABEL.nCols", -1)))
				|| ((_template != null) && (_nCols != _template._nCols)))
			pw.print(" nCols='" + _nCols + "'");
		if (((_template == null) && (_nRows != getIntValue("LABEL.nRows", -1)))
				|| ((_template != null) && (_nRows != _template._nRows)))
			pw.print(" nRows='" + _nRows + "'");
		pw.println("/>");
	}
}
