/*
 * LyLayout.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;
import java.io.*;
import leon.view.graphic.*;
import leon.misc.LyString;


/**
 * The LyLayout class is a generated Java implementation that corresponds to the element 
 * LAYOUT of the descriptive grammar of the views (views.dtd).
 * <B>&lt;LAYOUT&gt; : GEOGRAPHIC Places the nodes, the regions and the images according to their coordinates (x, y, z attributes).</B>
 * A layout manager allows placing the nodes of a map according to a predefined 
 * layout.
 * In most cases, the coordinates of the nodes are not used.
 * @see leon.view.LyComponent
 * See also tag <A href="../../conf/dtd/view/LAYOUT.html">LAYOUT</A> of grammar leon_views.dtd.
 */
public class LyLayout extends LyComponent
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'style'. */
	public enum Style
	{
		CIRCLE,
		ELLIPSE,
		MATRIX,
		LINEAR,
		GEOGRAPHIC,
		COORDINATES,
		TREEMAP;
	}


	/** Possible values for attribute 'hAlign'. */
	public enum HAlign
	{
		LEFT,
		CENTER,
		RIGHT;
	}


	/** Possible values for attribute 'vAlign'. */
	public enum VAlign
	{
		TOP,
		MIDDLE,
		BOTTOM;
	}


	/** Possible values for attribute 'fill'. */
	public enum Fill
	{
		HORIZONTAL,
		VERTICAL;
	}


	/* Static methods */

	/**
	 * Utility method to get style value from a string.
	 * @param token Token string
	 * @return Style value corresponding to this token
	 */
	public static final Style getStyle(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("CIRCLE")
				|| token.equals("0"))
			return Style.CIRCLE;
		else if (token.equalsIgnoreCase("ELLIPSE")
				|| token.equals("1"))
			return Style.ELLIPSE;
		else if (token.equalsIgnoreCase("MATRIX")
				|| token.equals("2"))
			return Style.MATRIX;
		else if (token.equalsIgnoreCase("LINEAR")
				|| token.equals("3"))
			return Style.LINEAR;
		else if (token.equalsIgnoreCase("GEOGRAPHIC")
				|| token.equals("4"))
			return Style.GEOGRAPHIC;
		else if (token.equalsIgnoreCase("COORDINATES")
				|| token.equals("5"))
			return Style.COORDINATES;
		else if (token.equalsIgnoreCase("TREEMAP")
				|| token.equals("6"))
			return Style.TREEMAP;
		else
			return Style.GEOGRAPHIC;
	}

	/**
	 * Utility method to get hAlign value from a string.
	 * @param token Token string
	 * @return HAlign value corresponding to this token
	 */
	public static final HAlign getHAlign(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("LEFT")
				|| token.equals("0"))
			return HAlign.LEFT;
		else if (token.equalsIgnoreCase("CENTER")
				|| token.equals("1"))
			return HAlign.CENTER;
		else if (token.equalsIgnoreCase("RIGHT")
				|| token.equals("2"))
			return HAlign.RIGHT;
		else
			return HAlign.LEFT;
	}

	/**
	 * Utility method to get vAlign value from a string.
	 * @param token Token string
	 * @return VAlign value corresponding to this token
	 */
	public static final VAlign getVAlign(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("TOP")
				|| token.equals("0"))
			return VAlign.TOP;
		else if (token.equalsIgnoreCase("MIDDLE")
				|| token.equals("1"))
			return VAlign.MIDDLE;
		else if (token.equalsIgnoreCase("BOTTOM")
				|| token.equals("2"))
			return VAlign.BOTTOM;
		else
			return VAlign.TOP;
	}

	/**
	 * Utility method to get fill value from a string.
	 * @param token Token string
	 * @return Fill value corresponding to this token
	 */
	public static final Fill getFill(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("HORIZONTAL")
				|| token.equals("0"))
			return Fill.HORIZONTAL;
		else if (token.equalsIgnoreCase("VERTICAL")
				|| token.equals("1"))
			return Fill.VERTICAL;
		else
			return Fill.HORIZONTAL;
	}


	/* Attributes */

	/**
	 * Type of manager. The different types of manager are:
	 * CIRCL Performs a circular distribution of the nodes.
	 * The circle on which the nodes are placed occupies the maximum place available.
	 * ELLIPSE Performs an elliptic distribution of the nodes.
	 * The ellipse on which the nodes are placed occupies the maximum place available.
	 * MATRIX Performs a matrix distribution of the nodes.
	 * The dimension of the matrix (number of columns / number of lines) is 
	 * proportional to the dimensions of the map (width / height).
	 * LINEAR Performs a linear distribution of the nodes.
	 * They are displayed the one after the others in an horizontal or vertical 
	 * direction (according to the value of the fill attribute) and with a spacing that
	 * depends on the value of the deltaNode attribute .
	 */
	protected Style				_style;

	/**
	 * Reference to the identifier of the node that is considered as center for the 
	 * current manager.
	 * This attribute is used only for the managers of CIRCLE and ELLIPSE type.
	 * The referred node is placed in the middle of the ellipse or of the circle on 
	 * which the other nodes are placed.
	 */
	protected String			_center;

	/**
	 * Spacing between the nodes.
	 * The expected value is a positive integer value representing the space to insert 
	 * between each node of the map.
	 * This value is used only for the managers of MATRIX and LINEAR type.
	 */
	protected int				_deltaNode;

	/**
	 * Spacing between the nodes and the edge of the map.
	 * The expected value is a positive integer value representing the minimum space to
	 * insert between each node of the map and the edges of this one (top, bottom, left
	 * and right edge).
	 * This value is used only for the managers of CIRCLE, ELLIPSE, MATRIX and LINEAR 
	 * type.
	 */
	protected int				_deltaBorder;

	/**
	 * Size of the selection border. This value is used only for the managers of 
	 * TREEMAP type.
	 */
	protected String			_selectBorderSize;

	/**
	 * Sizes of the separation borders for the different levels of hierarchy. This 
	 * value is used only for the managers of TREEMAP type. The different sizes must be
	 * separated by spaces.
	 */
	protected String			_borderSizes;

	/**
	 * Styles of the labels for the different levels of hierarchy. This value is used 
	 * only for the managers of TREEMAP type. The different types of styles are:
	 * OUTSIDE Label outside the region, centered above the region.
	 * TOPLEFT Label at the upper-left corner of the region.
	 * CENTER Label at the center of the region.
	 * NONE No label.
	 * The different styles must be separated by spaces.
	 */
	protected String			_labelStyles;

	/**
	 * Horizontal alignment of the elements that are managed by the manager. The 
	 * possible values are: left aligned (LEFT), right aligned (RIGHT), centered 
	 * (CENTER).
	 */
	protected HAlign			_hAlign;

	/**
	 * Vertical alignment of the elements managed by the manager. The possible values 
	 * are: bottom aligned (BOTTOM), top aligned (TOP), centered (MIDDLE).
	 */
	protected VAlign			_vAlign;

	/**
	 * Filling direction of the map (either horizontal or vertical). The value of this 
	 * attribute is used for the managers of LINEAR type.
	 */
	protected Fill				_fill;

	/**
	 * Name of the java class that manages the layout of the map elements.
	 */
	protected String			_javaName;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_style = Style.GEOGRAPHIC;
		_center = null;
		_deltaNode = getIntValue("LAYOUT.deltaNode", -1);
		_deltaBorder = getIntValue("LAYOUT.deltaBorder", -1);
		_selectBorderSize = null;
		_borderSizes = null;
		_labelStyles = null;
		_hAlign = HAlign.LEFT;
		_vAlign = VAlign.TOP;
		_fill = Fill.HORIZONTAL;
		_javaName = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyLayout	model = (LyLayout)component;

		_model = component;
		_style = model._style;
		_center = model._center;
		_deltaNode = model._deltaNode;
		_deltaBorder = model._deltaBorder;
		_selectBorderSize = model._selectBorderSize;
		_borderSizes = model._borderSizes;
		_labelStyles = model._labelStyles;
		_hAlign = model._hAlign;
		_vAlign = model._vAlign;
		_fill = model._fill;
		_javaName = model._javaName;
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyLayout()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyLayout(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_style = getStyle(node.getString("style"));
		_center = node.getString("center");
		_deltaNode = node.getInt("deltaNode");
		_deltaBorder = node.getInt("deltaBorder");
		_selectBorderSize = node.getString("selectBorderSize");
		_borderSizes = node.getString("borderSizes");
		_labelStyles = node.getString("labelStyles");
		_hAlign = getHAlign(node.getString("hAlign"));
		_vAlign = getVAlign(node.getString("vAlign"));
		_fill = getFill(node.getString("fill"));
		_javaName = node.getString("javaName");
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyLayout clone = new LyLayout();

		clone.initFromModel(this);
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of style.
	 * @return value of style.
	 */
	public Style getStyle()
	{
		return _style;
	}

	/**
	 * Gets value of center.
	 * @return value of center.
	 */
	public String getCenter()
	{
		return _center;
	}

	/**
	 * Gets value of deltaNode.
	 * @return value of deltaNode.
	 */
	public int getDeltaNode()
	{
		return _deltaNode;
	}

	/**
	 * Gets value of deltaBorder.
	 * @return value of deltaBorder.
	 */
	public int getDeltaBorder()
	{
		return _deltaBorder;
	}

	/**
	 * Gets value of selectBorderSize.
	 * @return value of selectBorderSize.
	 */
	public String getSelectBorderSize()
	{
		return _selectBorderSize;
	}

	/**
	 * Gets value of borderSizes.
	 * @return value of borderSizes.
	 */
	public String getBorderSizes()
	{
		return _borderSizes;
	}

	/**
	 * Gets value of labelStyles.
	 * @return value of labelStyles.
	 */
	public String getLabelStyles()
	{
		return _labelStyles;
	}

	/**
	 * Gets value of hAlign.
	 * @return value of hAlign.
	 */
	public HAlign getHAlign()
	{
		return _hAlign;
	}

	/**
	 * Gets value of vAlign.
	 * @return value of vAlign.
	 */
	public VAlign getVAlign()
	{
		return _vAlign;
	}

	/**
	 * Gets value of fill.
	 * @return value of fill.
	 */
	public Fill getFill()
	{
		return _fill;
	}

	/**
	 * Gets value of javaName.
	 * @return value of javaName.
	 */
	public String getJavaName()
	{
		return _javaName;
	}


	/* Modifiers */

	/**
	 * Initializes attribute style to the given value.
	 * @param style New value of style.
	 * @exception IllegalArgumentException if style parameter is null.
	 */
	public void setStyle(Style style)
	{
		if (style == _style)
			return;
		if (style == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute style");

		_style = style;
	}

	/**
	 * Initializes attribute center to the given value.
	 * @param center New value of center.
	 */
	public void setCenter(String center)
	{
		if ((center == _center)
			|| ((center != null)
				&& (_center != null)
				&& center.equals(_center)))
			return;

		_center = center;
	}

	/**
	 * Initializes attribute deltaNode to the given value.
	 * @param deltaNode New value of deltaNode.
	 */
	public void setDeltaNode(int deltaNode)
	{
		if (deltaNode == _deltaNode)
			return;
		_deltaNode = deltaNode;
	}

	/**
	 * Initializes attribute deltaBorder to the given value.
	 * @param deltaBorder New value of deltaBorder.
	 */
	public void setDeltaBorder(int deltaBorder)
	{
		if (deltaBorder == _deltaBorder)
			return;
		_deltaBorder = deltaBorder;
	}

	/**
	 * Initializes attribute selectBorderSize to the given value.
	 * @param selectBorderSize New value of selectBorderSize.
	 */
	public void setSelectBorderSize(String selectBorderSize)
	{
		if ((selectBorderSize == _selectBorderSize)
			|| ((selectBorderSize != null)
				&& (_selectBorderSize != null)
				&& selectBorderSize.equals(_selectBorderSize)))
			return;

		_selectBorderSize = selectBorderSize;
	}

	/**
	 * Sets attribute borderSizes to the given value.
	 * @param borderSizes New value of borderSizes.
	 */
	public void setBorderSizes(String borderSizes)
	{
		if ((borderSizes == _borderSizes)
			|| ((borderSizes != null)
				&& (_borderSizes != null)
				&& borderSizes.equals(_borderSizes)))
			return;

		_borderSizes = borderSizes;
		if (_internalView != null)
			((LyGraphicalLayout)_internalView).setBorderSizes(borderSizes);
	}

	/**
	 * Sets attribute labelStyles to the given value.
	 * @param labelStyles New value of labelStyles.
	 */
	public void setLabelStyles(String labelStyles)
	{
		if ((labelStyles == _labelStyles)
			|| ((labelStyles != null)
				&& (_labelStyles != null)
				&& labelStyles.equals(_labelStyles)))
			return;

		_labelStyles = labelStyles;
		if (_internalView != null)
			((LyGraphicalLayout)_internalView).setLabelStyles(labelStyles);
	}

	/**
	 * Initializes attribute hAlign to the given value.
	 * @param hAlign New value of hAlign.
	 * @exception IllegalArgumentException if hAlign parameter is null.
	 */
	public void setHAlign(HAlign hAlign)
	{
		if (hAlign == _hAlign)
			return;
		if (hAlign == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute hAlign");

		_hAlign = hAlign;
	}

	/**
	 * Initializes attribute vAlign to the given value.
	 * @param vAlign New value of vAlign.
	 * @exception IllegalArgumentException if vAlign parameter is null.
	 */
	public void setVAlign(VAlign vAlign)
	{
		if (vAlign == _vAlign)
			return;
		if (vAlign == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute vAlign");

		_vAlign = vAlign;
	}

	/**
	 * Initializes attribute fill to the given value.
	 * @param fill New value of fill.
	 * @exception IllegalArgumentException if fill parameter is null.
	 */
	public void setFill(Fill fill)
	{
		if (fill == _fill)
			return;
		if (fill == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute fill");

		_fill = fill;
	}

	/**
	 * Initializes attribute javaName to the given value.
	 * @param javaName New value of javaName.
	 */
	public void setJavaName(String javaName)
	{
		if ((javaName == _javaName)
			|| ((javaName != null)
				&& (_javaName != null)
				&& javaName.equals(_javaName)))
			return;

		_javaName = javaName;
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<LAYOUT");
		if (_style == Style.CIRCLE)
			pw.print(" style='CIRCLE'");
		else if (_style == Style.ELLIPSE)
			pw.print(" style='ELLIPSE'");
		else if (_style == Style.MATRIX)
			pw.print(" style='MATRIX'");
		else if (_style == Style.LINEAR)
			pw.print(" style='LINEAR'");
		else if (_style == Style.COORDINATES)
			pw.print(" style='COORDINATES'");
		else if (_style == Style.TREEMAP)
			pw.print(" style='TREEMAP'");
		if (_center != null)
			pw.print(" center='" + LyString.getXmlString(_center) + "'");
		if (_deltaNode != getIntValue("LAYOUT.deltaNode", -1))
			pw.print(" deltaNode='" + _deltaNode + "'");
		if (_deltaBorder != getIntValue("LAYOUT.deltaBorder", -1))
			pw.print(" deltaBorder='" + _deltaBorder + "'");
		if (_selectBorderSize != null)
			pw.print(" selectBorderSize='" + LyString.getXmlString(_selectBorderSize) + "'");
		if (_borderSizes != null)
			pw.print(" borderSizes='" + LyString.getXmlString(_borderSizes) + "'");
		if (_labelStyles != null)
			pw.print(" labelStyles='" + LyString.getXmlString(_labelStyles) + "'");
		if (_hAlign == HAlign.CENTER)
			pw.print(" hAlign='CENTER'");
		else if (_hAlign == HAlign.RIGHT)
			pw.print(" hAlign='RIGHT'");
		if (_vAlign == VAlign.MIDDLE)
			pw.print(" vAlign='MIDDLE'");
		else if (_vAlign == VAlign.BOTTOM)
			pw.print(" vAlign='BOTTOM'");
		if (_fill == Fill.VERTICAL)
			pw.print(" fill='VERTICAL'");
		if (_javaName != null)
			pw.print(" javaName='" + LyString.getXmlString(_javaName) + "'");
		pw.println("/>");
	}
}
