/*
 * LyList.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyList class is a generated Java implementation that corresponds to the element 
 * LIST of the descriptive grammar of the views (views.dtd).
 * <B>&lt;LIST&gt; : Input Field of List Type</B>
 * An input field of List type allows choosing one or more values among a discrete 
 * list of values.
 * It is possible to specify if the selection is multiple or not as well as the 
 * height of the list.
 * @see leon.view.LyFieldItem
 * See also tag <A href="../../conf/dtd/view/LIST.html">LIST</A> of grammar leon_views.dtd.
 */
public class LyList extends LyFieldItem
	implements  LyGetIdInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'selectionMode'. */
	public enum SelectionMode
	{
		SINGLE,
		MULTIPLE;
	}


	/* Static methods */

	/**
	 * Utility method to get selectionMode value from a string.
	 * @param token Token string
	 * @return SelectionMode value corresponding to this token
	 */
	public static final SelectionMode getSelectionMode(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("SINGLE")
				|| token.equals("0"))
			return SelectionMode.SINGLE;
		else if (token.equalsIgnoreCase("MULTIPLE")
				|| token.equals("1"))
			return SelectionMode.MULTIPLE;
		else
			return SelectionMode.SINGLE;
	}


	/* Attributes */

	/**
	 * Identifier of the component.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyList			_template;

	/**
	 * Indicates the type of selection (simple or multiple).
	 */
	protected SelectionMode			_selectionMode;

	/**
	 * Gives the height (number of columns) on which the list is shown (given in number
	 * of characters).
	 */
	protected int				_nCols;

	/**
	 * Gives the width (number of rows) on which the list is shown (given in number of 
	 * characters).
	 */
	protected int				_nRows;

	/**
	 * Allows specifying the name of a Java class that replaces the default graphical 
	 * component.
	 */
	protected String			_bean;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * LIST.tipFont
	 */
	protected String			_tipFont;

	/**
	 * LIST.tipColor
	 */
	protected String			_tipColor;

	/**
	 * LIST.tipFontColor
	 */
	protected String			_tipFontColor;

	/**
	 * LIST.selectedColor
	 */
	protected String			_selectedColor;

	/**
	 * LIST.inactiveSelectColor
	 */
	protected String			_inactiveSelectColor;

	/**
	 * LIST.selectedFontColor
	 */
	protected String			_selectedFontColor;

	/**
	 * LIST.inactiveSelectFontColor
	 */
	protected String			_inactiveSelectFontColor;


	/* Sub components */

	/**
	 * List of the options offered for this list.
	 */
	protected ArrayList<LyListOption>		_listOptions;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_selectionMode = SelectionMode.SINGLE;
		_nCols = getIntValue("LIST.nCols", -1);
		_nRows = getIntValue("LIST.nRows", -1);
		_bean = null;
		_font = getStringValue("LIST.font", "$ALL.font");
		_color = getStringValue("LIST.color", "$ALL.color");
		_fontColor = getStringValue("LIST.fontColor", "$ALL.fontColor");
		_tipFont = getStringValue("LIST.tipFont", "$ALL.tipFont");
		_tipColor = getStringValue("LIST.tipColor", "$ALL.tipColor");
		_tipFontColor = getStringValue("LIST.tipFontColor", "$ALL.tipFontColor");
		_selectedColor = getStringValue("LIST.selectedColor", "$ALL.selectedColor");
		_inactiveSelectColor = getStringValue("LIST.inactiveSelectColor", "$ALL.inactiveSelectColor");
		_selectedFontColor = getStringValue("LIST.selectedFontColor", "$ALL.selectedFontColor");
		_inactiveSelectFontColor = getStringValue("LIST.inactiveSelectFontColor", "$ALL.inactiveSelectFontColor");

		// Initialize sub components
		_listOptions = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyList	model = (LyList)component;

		_model = component;
		_selectionMode = model._selectionMode;
		_nCols = model._nCols;
		_nRows = model._nRows;
		_bean = model._bean;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;
		_tipFont = model._tipFont;
		_tipColor = model._tipColor;
		_tipFontColor = model._tipFontColor;
		_selectedColor = model._selectedColor;
		_inactiveSelectColor = model._inactiveSelectColor;
		_selectedFontColor = model._selectedFontColor;
		_inactiveSelectFontColor = model._inactiveSelectFontColor;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyListOption	listOptions;
		int			i, n;

		n = (model._listOptions == null ? 0 : model._listOptions.size());
		for (i = 0; i < n; i++)
		{
			listOptions = model._listOptions.get(i);
			if (listOptions != null)
				listOptions = (LyListOption)listOptions.cloneComponent();
			addChild(listOptions);
		}
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyList()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyList(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyList)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("selectionMode"))
			_selectionMode = getSelectionMode(node.getString("selectionMode"));
		if (node.isSpecified("nCols"))
			_nCols = node.getInt("nCols");
		if (node.isSpecified("nRows"))
			_nRows = node.getInt("nRows");
		if (node.isSpecified("bean"))
			_bean = node.getString("bean");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("tipFont"))
			_tipFont = node.getString("tipFont");
		if (node.isSpecified("tipColor"))
			_tipColor = node.getString("tipColor");
		if (node.isSpecified("tipFontColor"))
			_tipFontColor = node.getString("tipFontColor");
		if (node.isSpecified("selectedColor"))
			_selectedColor = node.getString("selectedColor");
		if (node.isSpecified("inactiveSelectColor"))
			_inactiveSelectColor = node.getString("inactiveSelectColor");
		if (node.isSpecified("selectedFontColor"))
			_selectedFontColor = node.getString("selectedFontColor");
		if (node.isSpecified("inactiveSelectFontColor"))
			_inactiveSelectFontColor = node.getString("inactiveSelectFontColor");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyListOption)
			addListOption((LyListOption)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyList clone = new LyList();

		clone.initFromModel(this);
		clone._template = this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyList getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of selectionMode.
	 * @return value of selectionMode.
	 */
	public SelectionMode getSelectionMode()
	{
		return _selectionMode;
	}

	/**
	 * Gets value of nCols.
	 * @return value of nCols.
	 */
	public int getNCols()
	{
		return _nCols;
	}

	/**
	 * Gets value of nRows.
	 * @return value of nRows.
	 */
	public int getNRows()
	{
		return _nRows;
	}

	/**
	 * Gets value of bean.
	 * @return value of bean.
	 */
	public String getBean()
	{
		return _bean;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("LIST.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("LIST.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("LIST.fontColor"));
	}

	/**
	 * Gets value of tipFont.
	 * @return value of tipFont.
	 */
	public String getTipFont()
	{
		return _tipFont;
	}

	/**
	 * Checks whether the tipFont attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultTipFont()
	{
		return (_tipFont == getStringValue("LIST.tipFont"));
	}

	/**
	 * Gets value of tipColor.
	 * @return value of tipColor.
	 */
	public String getTipColor()
	{
		return _tipColor;
	}

	/**
	 * Checks whether the tipColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultTipColor()
	{
		return (_tipColor == getStringValue("LIST.tipColor"));
	}

	/**
	 * Gets value of tipFontColor.
	 * @return value of tipFontColor.
	 */
	public String getTipFontColor()
	{
		return _tipFontColor;
	}

	/**
	 * Checks whether the tipFontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultTipFontColor()
	{
		return (_tipFontColor == getStringValue("LIST.tipFontColor"));
	}

	/**
	 * Gets value of selectedColor.
	 * @return value of selectedColor.
	 */
	public String getSelectedColor()
	{
		return _selectedColor;
	}

	/**
	 * Checks whether the selectedColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultSelectedColor()
	{
		return (_selectedColor == getStringValue("LIST.selectedColor"));
	}

	/**
	 * Gets value of inactiveSelectColor.
	 * @return value of inactiveSelectColor.
	 */
	public String getInactiveSelectColor()
	{
		return _inactiveSelectColor;
	}

	/**
	 * Checks whether the inactiveSelectColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultInactiveSelectColor()
	{
		return (_inactiveSelectColor == getStringValue("LIST.inactiveSelectColor"));
	}

	/**
	 * Gets value of selectedFontColor.
	 * @return value of selectedFontColor.
	 */
	public String getSelectedFontColor()
	{
		return _selectedFontColor;
	}

	/**
	 * Checks whether the selectedFontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultSelectedFontColor()
	{
		return (_selectedFontColor == getStringValue("LIST.selectedFontColor"));
	}

	/**
	 * Gets value of inactiveSelectFontColor.
	 * @return value of inactiveSelectFontColor.
	 */
	public String getInactiveSelectFontColor()
	{
		return _inactiveSelectFontColor;
	}

	/**
	 * Checks whether the inactiveSelectFontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultInactiveSelectFontColor()
	{
		return (_inactiveSelectFontColor == getStringValue("LIST.inactiveSelectFontColor"));
	}

	/**
	 * Get sub-component ListOption of index i.
	 * @param i the index of the ListOption to get.
	 * @return sub-component ListOption of index i.
	 */
	public LyListOption getListOption(int i)
	{
		if (_listOptions == null)
			return null;
		if ((i < 0) || (i >= _listOptions.size()))
			return null;
		return _listOptions.get(i);
	}

	/**
	 * Get sub-component ListOption count.
	 * @return sub-component ListOption count.
	 */
	public int getListOptionCount()
	{
		if (_listOptions == null)
			return 0;
		return _listOptions.size();
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyList template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Initializes attribute selectionMode to the given value.
	 * @param selectionMode New value of selectionMode.
	 * @exception IllegalArgumentException if selectionMode parameter is null.
	 */
	public void setSelectionMode(SelectionMode selectionMode)
	{
		if (selectionMode == _selectionMode)
			return;
		if (selectionMode == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute selectionMode");

		_selectionMode = selectionMode;
	}

	/**
	 * Initializes attribute nCols to the given value.
	 * @param nCols New value of nCols.
	 */
	public void setNCols(int nCols)
	{
		if (nCols == _nCols)
			return;
		_nCols = nCols;
	}

	/**
	 * Initializes attribute nRows to the given value.
	 * @param nRows New value of nRows.
	 */
	public void setNRows(int nRows)
	{
		if (nRows == _nRows)
			return;
		_nRows = nRows;
	}

	/**
	 * Initializes attribute bean to the given value.
	 * @param bean New value of bean.
	 */
	public void setBean(String bean)
	{
		if ((bean == _bean)
			|| ((bean != null)
				&& (_bean != null)
				&& bean.equals(_bean)))
			return;

		_bean = bean;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Initializes attribute tipFont to the given value.
	 * @param tipFont New value of tipFont.
	 */
	public void setTipFont(String tipFont)
	{
		if ((tipFont == _tipFont)
			|| ((tipFont != null)
				&& (_tipFont != null)
				&& tipFont.equals(_tipFont)))
			return;

		_tipFont = tipFont;
	}

	/**
	 * Initializes attribute tipColor to the given value.
	 * @param tipColor New value of tipColor.
	 */
	public void setTipColor(String tipColor)
	{
		if ((tipColor == _tipColor)
			|| ((tipColor != null)
				&& (_tipColor != null)
				&& tipColor.equals(_tipColor)))
			return;

		_tipColor = tipColor;
	}

	/**
	 * Initializes attribute tipFontColor to the given value.
	 * @param tipFontColor New value of tipFontColor.
	 */
	public void setTipFontColor(String tipFontColor)
	{
		if ((tipFontColor == _tipFontColor)
			|| ((tipFontColor != null)
				&& (_tipFontColor != null)
				&& tipFontColor.equals(_tipFontColor)))
			return;

		_tipFontColor = tipFontColor;
	}

	/**
	 * Initializes attribute selectedColor to the given value.
	 * @param selectedColor New value of selectedColor.
	 */
	public void setSelectedColor(String selectedColor)
	{
		if ((selectedColor == _selectedColor)
			|| ((selectedColor != null)
				&& (_selectedColor != null)
				&& selectedColor.equals(_selectedColor)))
			return;

		_selectedColor = selectedColor;
	}

	/**
	 * Initializes attribute inactiveSelectColor to the given value.
	 * @param inactiveSelectColor New value of inactiveSelectColor.
	 */
	public void setInactiveSelectColor(String inactiveSelectColor)
	{
		if ((inactiveSelectColor == _inactiveSelectColor)
			|| ((inactiveSelectColor != null)
				&& (_inactiveSelectColor != null)
				&& inactiveSelectColor.equals(_inactiveSelectColor)))
			return;

		_inactiveSelectColor = inactiveSelectColor;
	}

	/**
	 * Initializes attribute selectedFontColor to the given value.
	 * @param selectedFontColor New value of selectedFontColor.
	 */
	public void setSelectedFontColor(String selectedFontColor)
	{
		if ((selectedFontColor == _selectedFontColor)
			|| ((selectedFontColor != null)
				&& (_selectedFontColor != null)
				&& selectedFontColor.equals(_selectedFontColor)))
			return;

		_selectedFontColor = selectedFontColor;
	}

	/**
	 * Initializes attribute inactiveSelectFontColor to the given value.
	 * @param inactiveSelectFontColor New value of inactiveSelectFontColor.
	 */
	public void setInactiveSelectFontColor(String inactiveSelectFontColor)
	{
		if ((inactiveSelectFontColor == _inactiveSelectFontColor)
			|| ((inactiveSelectFontColor != null)
				&& (_inactiveSelectFontColor != null)
				&& inactiveSelectFontColor.equals(_inactiveSelectFontColor)))
			return;

		_inactiveSelectFontColor = inactiveSelectFontColor;
	}

	/**
	 * Adds a sub-component listOption.
	 * @param listOption The listOption to add to listOptions list.
	 */
	public void addListOption(LyListOption listOption)
	{
		insertListOption(listOption, -1);
	}

	/**
	 * Inserts a sub-component listOption at given position.
	 * @param listOption The listOption to add to listOptions list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertListOption(LyListOption listOption, int pos)
	{
		if (_listOptions == null)
		{
			_listOptions = new ArrayList<LyListOption>(1);
			_children = _listOptions;
		}
		else if (_listOptions.contains(listOption))
			return;
		if (pos >= 0)
			_listOptions.add(pos,listOption);
		if (listOption != null)
			listOption.setParent(this);
		if (_internalView instanceof LySetListOptionsInterface)
			((LySetListOptionsInterface)_internalView).addListOption(listOption, pos);
	}

	/**
	 * Removes a sub-component listOption.
	 * @param listOption The listOption to remove to listOptions list.
	 * @exception IllegalStateException if listOptions list is empty.
	 * @exception IllegalArgumentException if listOptions list doesn't contain listOption.
	 */
	public void removeListOption(LyListOption listOption)
	{
		if (_listOptions == null)
			throw new IllegalStateException("BF AppEngine : listOptions list is empty");
		if (!_listOptions.contains(listOption))
			throw new IllegalArgumentException("BF AppEngine : listOptions list doesn't contain listOption");

		if (_internalView instanceof LySetListOptionsInterface)
			((LySetListOptionsInterface)_internalView).removeListOption(listOption);
		listOption.setParent(null);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<LIST");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if ((_template == null)
			|| (_selectionMode != _template._selectionMode))
		{
			if ((_selectionMode == SelectionMode.SINGLE) && (_template != null))
				pw.print(" selectionMode='SINGLE'");
			else if (_selectionMode == SelectionMode.MULTIPLE)
				pw.print(" selectionMode='MULTIPLE'");
		}
		if (((_template == null) && (_nCols != getIntValue("LIST.nCols", -1)))
				|| ((_template != null) && (_nCols != _template._nCols)))
			pw.print(" nCols='" + _nCols + "'");
		if (((_template == null) && (_nRows != getIntValue("LIST.nRows", -1)))
				|| ((_template != null) && (_nRows != _template._nRows)))
			pw.print(" nRows='" + _nRows + "'");
		if (((_template == null) && (_bean != null))
				|| ((_template != null) && (_bean != _template._bean)))
			pw.print(" bean='" + LyString.getXmlString(_bean) + "'");
		if (((_template == null) && (_font != getStringValue("LIST.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("LIST.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("LIST.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && (_tipFont != getStringValue("LIST.tipFont", "$ALL.tipFont")))
				|| ((_template != null) && (_tipFont != _template._tipFont)))
			pw.print(" tipFont='" + LyString.getXmlString(_tipFont) + "'");
		if (((_template == null) && (_tipColor != getStringValue("LIST.tipColor", "$ALL.tipColor")))
				|| ((_template != null) && (_tipColor != _template._tipColor)))
			pw.print(" tipColor='" + LyString.getXmlString(_tipColor) + "'");
		if (((_template == null) && (_tipFontColor != getStringValue("LIST.tipFontColor", "$ALL.tipFontColor")))
				|| ((_template != null) && (_tipFontColor != _template._tipFontColor)))
			pw.print(" tipFontColor='" + LyString.getXmlString(_tipFontColor) + "'");
		if (((_template == null) && (_selectedColor != getStringValue("LIST.selectedColor", "$ALL.selectedColor")))
				|| ((_template != null) && (_selectedColor != _template._selectedColor)))
			pw.print(" selectedColor='" + LyString.getXmlString(_selectedColor) + "'");
		if (((_template == null) && (_inactiveSelectColor != getStringValue("LIST.inactiveSelectColor", "$ALL.inactiveSelectColor")))
				|| ((_template != null) && (_inactiveSelectColor != _template._inactiveSelectColor)))
			pw.print(" inactiveSelectColor='" + LyString.getXmlString(_inactiveSelectColor) + "'");
		if (((_template == null) && (_selectedFontColor != getStringValue("LIST.selectedFontColor", "$ALL.selectedFontColor")))
				|| ((_template != null) && (_selectedFontColor != _template._selectedFontColor)))
			pw.print(" selectedFontColor='" + LyString.getXmlString(_selectedFontColor) + "'");
		if (((_template == null) && (_inactiveSelectFontColor != getStringValue("LIST.inactiveSelectFontColor", "$ALL.inactiveSelectFontColor")))
				|| ((_template != null) && (_inactiveSelectFontColor != _template._inactiveSelectFontColor)))
			pw.print(" inactiveSelectFontColor='" + LyString.getXmlString(_inactiveSelectFontColor) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	listOptions;
		int			i, n;

		n = (_listOptions == null ? 0 : _listOptions.size());
		for (i = 0; i < n; i++)
		{
			listOptions = _listOptions.get(i);
			listOptions.writeXml(pw, debug);
		}


		pw.println("</LIST>");
	}
}
