/*
 * LyMapBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyMapBase class is a generated Java implementation that corresponds to the element 
 * MAP of the descriptive grammar of the views (views.dtd).
 * <B>&lt;MAP&gt; : Map</B>
 * A map allows displaying any type of map view: maps representing geographic or 
 * synoptic areas, displaying of elements with different types of positioning, 
 * displaying of nodes, links, images, etc. There are four types of elements that 
 * can be placed on a map: the nodes, the links, the regions and the images.
 * The elements placed on a map can be sensitive or not (except the images that can
 * never be sensitive).
 * It is possible to specify for the nodes and the links the layer on which they 
 * are.
 * This layer allows to place the elements of a higher layer over the elements of a
 * lower layer (for example, the elements of the layer 2 are placed over the 
 * elements of the layer 1). The nodes allow defining elements that can be placed 
 * at any location on the map and that can be moved.
 * By using a special positioning manager, it is possible to place these nodes 
 * according to a predefined geometry (matrix, circle, ellipse, line,...).
 * When zooming on the map, a node keeps always the same aspect unlike an element 
 * of image type. The links allow defining segments between two nodes of the map.
 * If one of the two nodes defining the link is moved, the link follows the moved 
 * node so that it always starts from its source node and that it always arrives to
 * its destination node. The regions allow representing geometric areas defined by 
 * a discrete set of points.
 * These areas can be sensitive.
 * It is notably possible to lead to other map views when launching an action over 
 * one of these regions. It is possible to define the background that will be used 
 * for the map (defined by an image), the selection mode (none, simple or 
 * multiple), the visualization mode (through a viewer or with scroll bars) and to 
 * specify if the map is in two or three dimensions. The possible movements on a 
 * map are translations, zooms (forward/backward) and rotations.
 * A map in three dimensions allows spatial rotations enabling to display the 
 * regions in three dimensions.
 * A map in two dimensions allows the axial rotations around the center of the map.
 * @see leon.view.LyPanel
 * @see leon.view.LyLyMap
 * See also tag <A href="../../conf/dtd/view/MAP.html">MAP</A> of grammar leon_views.dtd.
 */
public abstract class LyMapBase extends LyPanel
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface,
				LyGetControllerInterface,
				LySetControllerInterface,
				LySetViewerInterface,
				LyGetConstraintInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'selectionMode'. */
	public enum SelectionMode
	{
		NONE,
		SINGLE,
		MULTIPLE;
	}


	/** Possible values for attribute 'selectionDraw'. */
	public enum SelectionDraw
	{
		FILL,
		BORDER;
	}


	/** Possible values for attribute 'viewerStatus'. */
	public enum ViewerStatus
	{
		SHOW,
		HIDE;
	}


	/** Possible values for attribute 'zoomStatus'. */
	public enum ZoomStatus
	{
		SHOW,
		HIDE;
	}


	/** Possible values for attribute 'translationStatus'. */
	public enum TranslationStatus
	{
		SHOW,
		HIDE;
	}


	/** Possible values for attribute 'style'. */
	public enum Style
	{
		STANDARD,
		HYBRID,
		ROADMAP,
		SATELLITE,
		TERRAIN;
	}


	/* Static methods */

	/**
	 * Utility method to get selectionMode value from a string.
	 * @param token Token string
	 * @return SelectionMode value corresponding to this token
	 */
	public static final SelectionMode getSelectionMode(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("NONE")
				|| token.equals("0"))
			return SelectionMode.NONE;
		else if (token.equalsIgnoreCase("SINGLE")
				|| token.equals("1"))
			return SelectionMode.SINGLE;
		else if (token.equalsIgnoreCase("MULTIPLE")
				|| token.equals("2"))
			return SelectionMode.MULTIPLE;
		else
			return SelectionMode.MULTIPLE;
	}

	/**
	 * Utility method to get selectionDraw value from a string.
	 * @param token Token string
	 * @return SelectionDraw value corresponding to this token
	 */
	public static final SelectionDraw getSelectionDraw(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("FILL")
				|| token.equals("0"))
			return SelectionDraw.FILL;
		else if (token.equalsIgnoreCase("BORDER")
				|| token.equals("1"))
			return SelectionDraw.BORDER;
		else
			return SelectionDraw.FILL;
	}

	/**
	 * Utility method to get viewerStatus value from a string.
	 * @param token Token string
	 * @return ViewerStatus value corresponding to this token
	 */
	public static final ViewerStatus getViewerStatus(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("SHOW")
				|| token.equals("0"))
			return ViewerStatus.SHOW;
		else if (token.equalsIgnoreCase("HIDE")
				|| token.equals("1"))
			return ViewerStatus.HIDE;
		else
			return ViewerStatus.SHOW;
	}

	/**
	 * Utility method to get zoomStatus value from a string.
	 * @param token Token string
	 * @return ZoomStatus value corresponding to this token
	 */
	public static final ZoomStatus getZoomStatus(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("SHOW")
				|| token.equals("0"))
			return ZoomStatus.SHOW;
		else if (token.equalsIgnoreCase("HIDE")
				|| token.equals("1"))
			return ZoomStatus.HIDE;
		else
			return ZoomStatus.SHOW;
	}

	/**
	 * Utility method to get translationStatus value from a string.
	 * @param token Token string
	 * @return TranslationStatus value corresponding to this token
	 */
	public static final TranslationStatus getTranslationStatus(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("SHOW")
				|| token.equals("0"))
			return TranslationStatus.SHOW;
		else if (token.equalsIgnoreCase("HIDE")
				|| token.equals("1"))
			return TranslationStatus.HIDE;
		else
			return TranslationStatus.SHOW;
	}

	/**
	 * Utility method to get style value from a string.
	 * @param token Token string
	 * @return Style value corresponding to this token
	 */
	public static final Style getStyle(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("STANDARD")
				|| token.equals("0"))
			return Style.STANDARD;
		else if (token.equalsIgnoreCase("HYBRID")
				|| token.equals("1"))
			return Style.HYBRID;
		else if (token.equalsIgnoreCase("ROADMAP")
				|| token.equals("2"))
			return Style.ROADMAP;
		else if (token.equalsIgnoreCase("SATELLITE")
				|| token.equals("3"))
			return Style.SATELLITE;
		else if (token.equalsIgnoreCase("TERRAIN")
				|| token.equals("4"))
			return Style.TERRAIN;
		else
			return Style.STANDARD;
	}


	/* Attributes */

	/**
	 * Identifier of the map.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyMap				_template;

	/**
	 * Label of the map.
	 */
	protected String			_label;

	/**
	 * Reference of the controller that manages the map (i.e. handles user demands and 
	 * refresh content).
	 */
	protected String			_controller;

	/**
	 * Selection mode of the elements of the map. The following modes are managed:
	 * NONE: No selection is possible
	 * SINGLE: Selection of only one node at a time
	 * MULTIPLE: Selection of several elements at a time
	 */
	protected SelectionMode			_selectionMode;

	/**
	 * Draw style for the elements that are selected on the map.
	 * This type allows defining the way the selection of an element of the map is 
	 * represented. The following types are managed:
	 * FILL: Filled frame style
	 * BORDER: Outlined frame style
	 */
	protected SelectionDraw			_selectionDraw;

	/**
	 * Visualization mode of the map.
	 * The true value indicates a visualization by using a viewer representing a 
	 * small-scale.
	 * The viewer allows locating its position and to move on the map.
	 * In the absence of a viewer, the map manages scroll bars.
	 */
	protected boolean			_viewer;

	/**
	 * Status of the viewer in the viewer mode.
	 * The viewer can have the two following statuses:
	 * SHOW: the viewer is shown.
	 * HIDE: the viewer is hidden.
	 */
	protected ViewerStatus			_viewerStatus;

	/**
	 * X position of the viewer in the map.
	 */
	protected int				_viewerX;

	/**
	 * Y position of the viewer in the map.
	 */
	protected int				_viewerY;

	/**
	 * Indicates if the zoom widget must be displayed on the map.
	 */
	protected boolean			_zoom;

	/**
	 * Indicates if the zoom widget is visible or hidden.
	 */
	protected ZoomStatus			_zoomStatus;

	/**
	 * Step of the zoom widget (must be greater than 1).
	 */
	protected float				_zoomStep;

	/**
	 * Current zoom factor of the map.
	 */
	protected float				_zoomFactor;

	/**
	 * Indicates if the translation widget must be displayed on the map.
	 */
	protected boolean			_translation;

	/**
	 * Step of the translation widget. Used by the translation widget to move the map 
	 * and also by the Struts map component to compute the number of sub-images that 
	 * must be generated for the map background.
	 */
	protected int				_translationStep;

	/**
	 * Indicates if the translation widget is visible or hidden.
	 */
	protected TranslationStatus			_translationStatus;

	/**
	 * Indicates if the map is in three dimensions.
	 * A map in two dimensions manages only the coordinates in abscissa and in ordinate
	 * of the elements that it contains.
	 */
	protected boolean			_spatial;

	/**
	 * Background of the map.
	 * The expected value is the key of an image or else the name of an image.
	 * The pattern representing this image is repeated in order to recover the whole 
	 * visible part of the map.
	 * All the elements that are available on the map are displayed over this 
	 * background.
	 */
	protected String			_background;

	/**
	 * Font of the map title.
	 */
	protected String			_titleFont;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * Selection color used when nodes or links are selected within the map.
	 */
	protected String			_selectedColor;

	/**
	 * Color used for outline.
	 */
	protected String			_outlineColor;

	/**
	 * Width of the map component.
	 */
	protected int				_width;

	/**
	 * Height of the map component.
	 */
	protected int				_height;

	/**
	 * Style of the map component.
	 */
	protected Style				_style;

	/**
	 * Latitude of the map component.
	 */
	protected String			_latitude;

	/**
	 * Longitude of the map component.
	 */
	protected String			_longitude;


	/* Sub components */

	/**
	 * Components of the map. These components can be map nodes, map links or map 
	 * regions.
	 */
	protected ArrayList<LyComponent>		_components;

	/**
	 * Additional elements of the map.
	 * These elements represent the static part of the elements of the map.
	 * They are not intended to evolve dynamically when the map is used.
	 * Generally, these elements are defined in a file that is associated to the action
	 * used to display the map.
	 */
	protected LyMapContent		_mapContent;

	/**
	 * Layout manager of the map.
	 */
	protected LyLayout			_layout;

	/**
	 * Positioning constraint of the map.
	 */
	protected LyConstraint		_constraint;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_controller = null;
		_selectionMode = SelectionMode.MULTIPLE;
		_selectionDraw = SelectionDraw.FILL;
		_viewer = getBooleanValue("MAP.viewer", true);
		_viewerStatus = ViewerStatus.SHOW;
		_viewerX = getIntValue("MAP.viewerX", -1);
		_viewerY = getIntValue("MAP.viewerY", -1);
		_zoom = getBooleanValue("MAP.zoom", true);
		_zoomStatus = ZoomStatus.SHOW;
		_zoomStep = getFloatValue("MAP.zoomStep", 0F);
		_zoomFactor = getFloatValue("MAP.zoomFactor", 0F);
		_translation = getBooleanValue("MAP.translation", true);
		_translationStep = getIntValue("MAP.translationStep", -1);
		_translationStatus = TranslationStatus.SHOW;
		_spatial = getBooleanValue("MAP.spatial", false);
		_background = null;
		_titleFont = getStringValue("MAP.titleFont", "$ALL.titleFont");
		_color = getStringValue("MAP.color", "$ALL.color");
		_fontColor = getStringValue("MAP.fontColor", "$ALL.fontColor");
		_selectedColor = getStringValue("MAP.selectedColor", "$ALL.selectedColor");
		_outlineColor = getStringValue("MAP.outlineColor", "$ALL.outlineColor");
		_width = getIntValue("MAP.width", -1);
		_height = getIntValue("MAP.height", -1);
		_style = Style.STANDARD;
		_latitude = null;
		_longitude = null;

		// Initialize sub components
		_components = null;
		_mapContent = null;
		_layout = null;
		_constraint = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyMap	model = (LyMap)component;

		_model = component;
		_label = model._label;
		_controller = model._controller;
		_selectionMode = model._selectionMode;
		_selectionDraw = model._selectionDraw;
		_viewer = model._viewer;
		_viewerStatus = model._viewerStatus;
		_viewerX = model._viewerX;
		_viewerY = model._viewerY;
		_zoom = model._zoom;
		_zoomStatus = model._zoomStatus;
		_zoomStep = model._zoomStep;
		_zoomFactor = model._zoomFactor;
		_translation = model._translation;
		_translationStep = model._translationStep;
		_translationStatus = model._translationStatus;
		_spatial = model._spatial;
		_background = model._background;
		_titleFont = model._titleFont;
		_color = model._color;
		_fontColor = model._fontColor;
		_selectedColor = model._selectedColor;
		_outlineColor = model._outlineColor;
		_width = model._width;
		_height = model._height;
		_style = model._style;
		_latitude = model._latitude;
		_longitude = model._longitude;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyComponent	components;
		int			i, n;

		n = (model._components == null ? 0 : model._components.size());
		for (i = 0; i < n; i++)
		{
			components = model._components.get(i);
			if (components != null)
				components = components.cloneComponent();
			addChild(components);
		}
		if (model._mapContent != null)
			addChild(model._mapContent.cloneComponent());
		if (model._layout != null)
			addChild(model._layout.cloneComponent());
		if (model._constraint != null)
			addChild(model._constraint.cloneComponent());
		if (model._datas != null)
			_datas = new Hashtable<String, String>(model._datas);
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyMapBase()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyMapBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyMap)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("controller"))
			_controller = node.getString("controller");
		if (node.isSpecified("selectionMode"))
			_selectionMode = getSelectionMode(node.getString("selectionMode"));
		if (node.isSpecified("selectionDraw"))
			_selectionDraw = getSelectionDraw(node.getString("selectionDraw"));
		if (node.isSpecified("viewer"))
			_viewer = node.getBoolean("viewer");
		if (node.isSpecified("viewerStatus"))
			_viewerStatus = getViewerStatus(node.getString("viewerStatus"));
		if (node.isSpecified("viewerX"))
			_viewerX = node.getInt("viewerX");
		if (node.isSpecified("viewerY"))
			_viewerY = node.getInt("viewerY");
		if (node.isSpecified("zoom"))
			_zoom = node.getBoolean("zoom");
		if (node.isSpecified("zoomStatus"))
			_zoomStatus = getZoomStatus(node.getString("zoomStatus"));
		if (node.isSpecified("zoomStep"))
			_zoomStep = node.getFloat("zoomStep");
		if (node.isSpecified("zoomFactor"))
			_zoomFactor = node.getFloat("zoomFactor");
		if (node.isSpecified("translation"))
			_translation = node.getBoolean("translation");
		if (node.isSpecified("translationStep"))
			_translationStep = node.getInt("translationStep");
		if (node.isSpecified("translationStatus"))
			_translationStatus = getTranslationStatus(node.getString("translationStatus"));
		if (node.isSpecified("spatial"))
			_spatial = node.getBoolean("spatial");
		if (node.isSpecified("background"))
			_background = node.getString("background");
		if (node.isSpecified("titleFont"))
			_titleFont = node.getString("titleFont");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("selectedColor"))
			_selectedColor = node.getString("selectedColor");
		if (node.isSpecified("outlineColor"))
			_outlineColor = node.getString("outlineColor");
		if (node.isSpecified("width"))
			_width = node.getInt("width");
		if (node.isSpecified("height"))
			_height = node.getInt("height");
		if (node.isSpecified("style"))
			_style = getStyle(node.getString("style"));
		if (node.isSpecified("latitude"))
			_latitude = node.getString("latitude");
		if (node.isSpecified("longitude"))
			_longitude = node.getString("longitude");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyMapNode)
			addMapNode((LyMapNode)child);
		else if (child instanceof LyMapLink)
			addMapLink((LyMapLink)child);
		else if (child instanceof LyMapRegion)
			addMapRegion((LyMapRegion)child);
		else if (child instanceof LyMapContent)
			setMapContent((LyMapContent)child);
		else if (child instanceof LyLayout)
			setLayout((LyLayout)child);
		else if (child instanceof LyConstraint)
			setConstraint((LyConstraint)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyMap clone = new LyMap();

		clone.initFromModel(this);
		clone._template = (LyMap)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyMap getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of controller.
	 * @return value of controller.
	 */
	public String getController()
	{
		return _controller;
	}

	/**
	 * Gets value of selectionMode.
	 * @return value of selectionMode.
	 */
	public SelectionMode getSelectionMode()
	{
		return _selectionMode;
	}

	/**
	 * Gets value of selectionDraw.
	 * @return value of selectionDraw.
	 */
	public SelectionDraw getSelectionDraw()
	{
		return _selectionDraw;
	}

	/**
	 * Checks if value of attribute viewer is set.
	 * @return true if viewer is set, false otherwise.
	 */
	public boolean isViewer()
	{
		return _viewer;
	}

	/**
	 * Gets value of viewerStatus.
	 * @return value of viewerStatus.
	 */
	public ViewerStatus getViewerStatus()
	{
		return _viewerStatus;
	}

	/**
	 * Gets value of viewerX.
	 * @return value of viewerX.
	 */
	public int getViewerX()
	{
		return _viewerX;
	}

	/**
	 * Gets value of viewerY.
	 * @return value of viewerY.
	 */
	public int getViewerY()
	{
		return _viewerY;
	}

	/**
	 * Checks if value of attribute zoom is set.
	 * @return true if zoom is set, false otherwise.
	 */
	public boolean isZoom()
	{
		return _zoom;
	}

	/**
	 * Gets value of zoomStatus.
	 * @return value of zoomStatus.
	 */
	public ZoomStatus getZoomStatus()
	{
		return _zoomStatus;
	}

	/**
	 * Gets value of zoomStep.
	 * @return value of zoomStep.
	 */
	public float getZoomStep()
	{
		return _zoomStep;
	}

	/**
	 * Gets value of zoomFactor.
	 * @return value of zoomFactor.
	 */
	public float getZoomFactor()
	{
		return _zoomFactor;
	}

	/**
	 * Checks if value of attribute translation is set.
	 * @return true if translation is set, false otherwise.
	 */
	public boolean isTranslation()
	{
		return _translation;
	}

	/**
	 * Gets value of translationStep.
	 * @return value of translationStep.
	 */
	public int getTranslationStep()
	{
		return _translationStep;
	}

	/**
	 * Gets value of translationStatus.
	 * @return value of translationStatus.
	 */
	public TranslationStatus getTranslationStatus()
	{
		return _translationStatus;
	}

	/**
	 * Checks if value of attribute spatial is set.
	 * @return true if spatial is set, false otherwise.
	 */
	public boolean isSpatial()
	{
		return _spatial;
	}

	/**
	 * Gets value of background.
	 * @return value of background.
	 */
	public String getBackground()
	{
		return _background;
	}

	/**
	 * Gets value of titleFont.
	 * @return value of titleFont.
	 */
	public String getTitleFont()
	{
		return _titleFont;
	}

	/**
	 * Checks whether the titleFont attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultTitleFont()
	{
		return (_titleFont == getStringValue("MAP.titleFont"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("MAP.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("MAP.fontColor"));
	}

	/**
	 * Gets value of selectedColor.
	 * @return value of selectedColor.
	 */
	public String getSelectedColor()
	{
		return _selectedColor;
	}

	/**
	 * Checks whether the selectedColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultSelectedColor()
	{
		return (_selectedColor == getStringValue("MAP.selectedColor"));
	}

	/**
	 * Gets value of outlineColor.
	 * @return value of outlineColor.
	 */
	public String getOutlineColor()
	{
		return _outlineColor;
	}

	/**
	 * Checks whether the outlineColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultOutlineColor()
	{
		return (_outlineColor == getStringValue("MAP.outlineColor"));
	}

	/**
	 * Gets value of width.
	 * @return value of width.
	 */
	public int getWidth()
	{
		return _width;
	}

	/**
	 * Gets value of height.
	 * @return value of height.
	 */
	public int getHeight()
	{
		return _height;
	}

	/**
	 * Gets value of style.
	 * @return value of style.
	 */
	public Style getStyle()
	{
		return _style;
	}

	/**
	 * Gets value of latitude.
	 * @return value of latitude.
	 */
	public String getLatitude()
	{
		return _latitude;
	}

	/**
	 * Gets value of longitude.
	 * @return value of longitude.
	 */
	public String getLongitude()
	{
		return _longitude;
	}

	/**
	 * Get sub-component Component of index i.
	 * @param i the index of the Component to get.
	 * @return sub-component Component of index i.
	 */
	public LyComponent getComponent(int i)
	{
		if (_components == null)
			return null;
		if ((i < 0) || (i >= _components.size()))
			return null;
		return _components.get(i);
	}

	/**
	 * Get sub-component Component count.
	 * @return sub-component Component count.
	 */
	public int getComponentCount()
	{
		if (_components == null)
			return 0;
		return _components.size();
	}

	/**
	 * Get sub-component MapContent.
	 * @return sub-component MapContent.
	 */
	public LyMapContent getMapContent()
	{
		return _mapContent;
	}

	/**
	 * Get sub-component Layout.
	 * @return sub-component Layout.
	 */
	public LyLayout getLayout()
	{
		return _layout;
	}

	/**
	 * Get sub-component Constraint.
	 * @return sub-component Constraint.
	 */
	public LyConstraint getConstraint()
	{
		return _constraint;
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyMap template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView != null)
			((LyGraphicalMap)_internalView).setLabel(label);
	}

	/**
	 * Initializes attribute controller to the given value.
	 * @param controller New value of controller.
	 */
	public void setController(String controller)
	{
		if ((controller == _controller)
			|| ((controller != null)
				&& (_controller != null)
				&& controller.equals(_controller)))
			return;

		_controller = controller;
	}

	/**
	 * Initializes attribute selectionMode to the given value.
	 * @param selectionMode New value of selectionMode.
	 * @exception IllegalArgumentException if selectionMode parameter is null.
	 */
	public void setSelectionMode(SelectionMode selectionMode)
	{
		if (selectionMode == _selectionMode)
			return;
		if (selectionMode == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute selectionMode");

		_selectionMode = selectionMode;
	}

	/**
	 * Initializes attribute selectionDraw to the given value.
	 * @param selectionDraw New value of selectionDraw.
	 * @exception IllegalArgumentException if selectionDraw parameter is null.
	 */
	public void setSelectionDraw(SelectionDraw selectionDraw)
	{
		if (selectionDraw == _selectionDraw)
			return;
		if (selectionDraw == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute selectionDraw");

		_selectionDraw = selectionDraw;
	}

	/**
	 * Sets attribute viewer to the given value.
	 * @param viewer Boolean indicating if attribute viewer is set or unset.
	 */
	public void setViewer(boolean viewer)
	{
		if (viewer == _viewer)
			return;
		_viewer = viewer;
		if (_internalView != null)
			((LyGraphicalMap)_internalView).setViewer(viewer);
	}

	/**
	 * Sets attribute viewerStatus to the given value.
	 * @param viewerStatus New value of viewerStatus.
	 * @exception IllegalArgumentException if viewerStatus parameter is null.
	 */
	public void setViewerStatus(ViewerStatus viewerStatus)
	{
		if (viewerStatus == _viewerStatus)
			return;
		if (viewerStatus == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute viewerStatus");

		_viewerStatus = viewerStatus;
		if (_internalView != null)
			((LyGraphicalMap)_internalView).setViewerStatus(viewerStatus);
	}

	/**
	 * Initializes attribute viewerX to the given value.
	 * @param viewerX New value of viewerX.
	 */
	public void setViewerX(int viewerX)
	{
		if (viewerX == _viewerX)
			return;
		_viewerX = viewerX;
	}

	/**
	 * Initializes attribute viewerY to the given value.
	 * @param viewerY New value of viewerY.
	 */
	public void setViewerY(int viewerY)
	{
		if (viewerY == _viewerY)
			return;
		_viewerY = viewerY;
	}

	/**
	 * Sets attribute zoom to the given value.
	 * @param zoom Boolean indicating if attribute zoom is set or unset.
	 */
	public void setZoom(boolean zoom)
	{
		if (zoom == _zoom)
			return;
		_zoom = zoom;
		if (_internalView != null)
			((LyGraphicalMap)_internalView).setZoom(zoom);
	}

	/**
	 * Sets attribute zoomStatus to the given value.
	 * @param zoomStatus New value of zoomStatus.
	 * @exception IllegalArgumentException if zoomStatus parameter is null.
	 */
	public void setZoomStatus(ZoomStatus zoomStatus)
	{
		if (zoomStatus == _zoomStatus)
			return;
		if (zoomStatus == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute zoomStatus");

		_zoomStatus = zoomStatus;
		if (_internalView != null)
			((LyGraphicalMap)_internalView).setZoomStatus(zoomStatus);
	}

	/**
	 * Initializes attribute zoomStep to the given value.
	 * @param zoomStep New value of zoomStep.
	 */
	public void setZoomStep(float zoomStep)
	{
		if (zoomStep == _zoomStep)
			return;
		_zoomStep = zoomStep;
	}

	/**
	 * Sets attribute zoomFactor to the given value.
	 * @param zoomFactor New value of zoomFactor.
	 */
	public void setZoomFactor(float zoomFactor)
	{
		if (zoomFactor == _zoomFactor)
			return;
		_zoomFactor = zoomFactor;
		if (_internalView != null)
			((LyGraphicalMap)_internalView).setZoomFactor(zoomFactor);
	}

	/**
	 * Sets attribute translation to the given value.
	 * @param translation Boolean indicating if attribute translation is set or unset.
	 */
	public void setTranslation(boolean translation)
	{
		if (translation == _translation)
			return;
		_translation = translation;
		if (_internalView != null)
			((LyGraphicalMap)_internalView).setTranslation(translation);
	}

	/**
	 * Initializes attribute translationStep to the given value.
	 * @param translationStep New value of translationStep.
	 */
	public void setTranslationStep(int translationStep)
	{
		if (translationStep == _translationStep)
			return;
		_translationStep = translationStep;
	}

	/**
	 * Sets attribute translationStatus to the given value.
	 * @param translationStatus New value of translationStatus.
	 * @exception IllegalArgumentException if translationStatus parameter is null.
	 */
	public void setTranslationStatus(TranslationStatus translationStatus)
	{
		if (translationStatus == _translationStatus)
			return;
		if (translationStatus == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute translationStatus");

		_translationStatus = translationStatus;
		if (_internalView != null)
			((LyGraphicalMap)_internalView).setTranslationStatus(translationStatus);
	}

	/**
	 * Sets attribute spatial to the given value.
	 * @param spatial Boolean indicating if attribute spatial is set or unset.
	 */
	public void setSpatial(boolean spatial)
	{
		if (spatial == _spatial)
			return;
		_spatial = spatial;
		if (_internalView != null)
			((LyGraphicalMap)_internalView).setSpatial(spatial);
	}

	/**
	 * Initializes attribute background to the given value.
	 * @param background New value of background.
	 */
	public void setBackground(String background)
	{
		if ((background == _background)
			|| ((background != null)
				&& (_background != null)
				&& background.equals(_background)))
			return;

		_background = background;
	}

	/**
	 * Initializes attribute titleFont to the given value.
	 * @param titleFont New value of titleFont.
	 */
	public void setTitleFont(String titleFont)
	{
		if ((titleFont == _titleFont)
			|| ((titleFont != null)
				&& (_titleFont != null)
				&& titleFont.equals(_titleFont)))
			return;

		_titleFont = titleFont;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Initializes attribute selectedColor to the given value.
	 * @param selectedColor New value of selectedColor.
	 */
	public void setSelectedColor(String selectedColor)
	{
		if ((selectedColor == _selectedColor)
			|| ((selectedColor != null)
				&& (_selectedColor != null)
				&& selectedColor.equals(_selectedColor)))
			return;

		_selectedColor = selectedColor;
	}

	/**
	 * Initializes attribute outlineColor to the given value.
	 * @param outlineColor New value of outlineColor.
	 */
	public void setOutlineColor(String outlineColor)
	{
		if ((outlineColor == _outlineColor)
			|| ((outlineColor != null)
				&& (_outlineColor != null)
				&& outlineColor.equals(_outlineColor)))
			return;

		_outlineColor = outlineColor;
	}

	/**
	 * Initializes attribute width to the given value.
	 * @param width New value of width.
	 */
	public void setWidth(int width)
	{
		if (width == _width)
			return;
		_width = width;
	}

	/**
	 * Initializes attribute height to the given value.
	 * @param height New value of height.
	 */
	public void setHeight(int height)
	{
		if (height == _height)
			return;
		_height = height;
	}

	/**
	 * Initializes attribute style to the given value.
	 * @param style New value of style.
	 * @exception IllegalArgumentException if style parameter is null.
	 */
	public void setStyle(Style style)
	{
		if (style == _style)
			return;
		if (style == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute style");

		_style = style;
	}

	/**
	 * Initializes attribute latitude to the given value.
	 * @param latitude New value of latitude.
	 */
	public void setLatitude(String latitude)
	{
		if ((latitude == _latitude)
			|| ((latitude != null)
				&& (_latitude != null)
				&& latitude.equals(_latitude)))
			return;

		_latitude = latitude;
	}

	/**
	 * Initializes attribute longitude to the given value.
	 * @param longitude New value of longitude.
	 */
	public void setLongitude(String longitude)
	{
		if ((longitude == _longitude)
			|| ((longitude != null)
				&& (_longitude != null)
				&& longitude.equals(_longitude)))
			return;

		_longitude = longitude;
	}

	/**
	 * Adds a sub-component mapNode.
	 * @param mapNode The mapNode to add to components list.
	 */
	public void addMapNode(LyMapNode mapNode)
	{
		insertMapNode(mapNode, -1);
	}

	/**
	 * Inserts a sub-component mapNode at given position.
	 * @param mapNode The mapNode to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertMapNode(LyMapNode mapNode, int pos)
	{
		if (_components == null)
			_components = new ArrayList<LyComponent>(1);
		else if (_components.contains(mapNode))
			return;
		if (pos >= 0)
			_components.add(pos,mapNode);
		else
			_components.add(mapNode);
		if (mapNode != null)
			mapNode.setParent(this);
		if (_internalView != null)
			((LyGraphicalMap)_internalView).addComponent(mapNode, pos);
	}

	/**
	 * Removes a sub-component mapNode.
	 * @param mapNode The mapNode to remove to components list.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain mapNode.
	 */
	public void removeMapNode(LyMapNode mapNode)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if (!_components.contains(mapNode))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain mapNode");

		if (_internalView != null)
			((LyGraphicalMap)_internalView).removeComponent(mapNode);
		_components.remove(mapNode);
		mapNode.setParent(null);
	}

	/**
	 * Adds a sub-component mapLink.
	 * @param mapLink The mapLink to add to components list.
	 */
	public void addMapLink(LyMapLink mapLink)
	{
		insertMapLink(mapLink, -1);
	}

	/**
	 * Inserts a sub-component mapLink at given position.
	 * @param mapLink The mapLink to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertMapLink(LyMapLink mapLink, int pos)
	{
		if (_components == null)
			_components = new ArrayList<LyComponent>(1);
		else if (_components.contains(mapLink))
			return;
		if (pos >= 0)
			_components.add(pos,mapLink);
		else
			_components.add(mapLink);
		if (mapLink != null)
			mapLink.setParent(this);
		if (_internalView != null)
			((LyGraphicalMap)_internalView).addComponent(mapLink, pos);
	}

	/**
	 * Removes a sub-component mapLink.
	 * @param mapLink The mapLink to remove to components list.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain mapLink.
	 */
	public void removeMapLink(LyMapLink mapLink)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if (!_components.contains(mapLink))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain mapLink");

		if (_internalView != null)
			((LyGraphicalMap)_internalView).removeComponent(mapLink);
		_components.remove(mapLink);
		mapLink.setParent(null);
	}

	/**
	 * Adds a sub-component mapRegion.
	 * @param mapRegion The mapRegion to add to components list.
	 */
	public void addMapRegion(LyMapRegion mapRegion)
	{
		insertMapRegion(mapRegion, -1);
	}

	/**
	 * Inserts a sub-component mapRegion at given position.
	 * @param mapRegion The mapRegion to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertMapRegion(LyMapRegion mapRegion, int pos)
	{
		if (_components == null)
			_components = new ArrayList<LyComponent>(1);
		else if (_components.contains(mapRegion))
			return;
		if (pos >= 0)
			_components.add(pos,mapRegion);
		else
			_components.add(mapRegion);
		if (mapRegion != null)
			mapRegion.setParent(this);
		if (_internalView != null)
			((LyGraphicalMap)_internalView).addComponent(mapRegion, pos);
	}

	/**
	 * Removes a sub-component mapRegion.
	 * @param mapRegion The mapRegion to remove to components list.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain mapRegion.
	 */
	public void removeMapRegion(LyMapRegion mapRegion)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if (!_components.contains(mapRegion))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain mapRegion");

		if (_internalView != null)
			((LyGraphicalMap)_internalView).removeComponent(mapRegion);
		_components.remove(mapRegion);
		mapRegion.setParent(null);
	}

	/**
	 * Sets sub-component mapContent.
	 * @param mapContent New value of mapContent
	 */
	public void setMapContent(LyMapContent mapContent)
	{
		if (_mapContent != null)
			_mapContent.setParent(null);
		_mapContent = mapContent;
		if (mapContent != null)
			mapContent.setParent(this);
		if (_internalView != null)
			((LyGraphicalMap)_internalView).setMapContent(mapContent);
	}

	/**
	 * Sets sub-component layout.
	 * @param layout New value of layout
	 */
	public void setLayout(LyLayout layout)
	{
		if (_layout != null)
			_layout.setParent(null);
		_layout = layout;
		if (layout != null)
			layout.setParent(this);
		if (_internalView != null)
			((LyGraphicalMap)_internalView).setLayout(layout);
	}

	/**
	 * Sets sub-component constraint.
	 * @param constraint New value of constraint
	 */
	public void setConstraint(LyConstraint constraint)
	{
		if (_constraint != null)
			_constraint.setParent(null);
		_constraint = constraint;
		if (constraint != null)
			constraint.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<MAP");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlString(_controller) + "'");
		if ((_template == null)
			|| (_selectionMode != _template._selectionMode))
		{
			if (_selectionMode == SelectionMode.NONE)
				pw.print(" selectionMode='NONE'");
			else if (_selectionMode == SelectionMode.SINGLE)
				pw.print(" selectionMode='SINGLE'");
			else if ((_selectionMode == SelectionMode.MULTIPLE) && (_template != null))
				pw.print(" selectionMode='MULTIPLE'");
		}
		if ((_template == null)
			|| (_selectionDraw != _template._selectionDraw))
		{
			if ((_selectionDraw == SelectionDraw.FILL) && (_template != null))
				pw.print(" selectionDraw='FILL'");
			else if (_selectionDraw == SelectionDraw.BORDER)
				pw.print(" selectionDraw='BORDER'");
		}
		if (((_template == null) && !_viewer)
				|| ((_template != null) && (_viewer != _template._viewer)))
			pw.print(" viewer='" + _viewer + "'");
		if ((_template == null)
			|| (_viewerStatus != _template._viewerStatus))
		{
			if ((_viewerStatus == ViewerStatus.SHOW) && (_template != null))
				pw.print(" viewerStatus='SHOW'");
			else if (_viewerStatus == ViewerStatus.HIDE)
				pw.print(" viewerStatus='HIDE'");
		}
		if (((_template == null) && (_viewerX != getIntValue("MAP.viewerX", -1)))
				|| ((_template != null) && (_viewerX != _template._viewerX)))
			pw.print(" viewerX='" + _viewerX + "'");
		if (((_template == null) && (_viewerY != getIntValue("MAP.viewerY", -1)))
				|| ((_template != null) && (_viewerY != _template._viewerY)))
			pw.print(" viewerY='" + _viewerY + "'");
		if (((_template == null) && !_zoom)
				|| ((_template != null) && (_zoom != _template._zoom)))
			pw.print(" zoom='" + _zoom + "'");
		if ((_template == null)
			|| (_zoomStatus != _template._zoomStatus))
		{
			if ((_zoomStatus == ZoomStatus.SHOW) && (_template != null))
				pw.print(" zoomStatus='SHOW'");
			else if (_zoomStatus == ZoomStatus.HIDE)
				pw.print(" zoomStatus='HIDE'");
		}
		if (((_template == null) && (_zoomStep != getFloatValue("MAP.zoomStep", 0F)))
				|| ((_template != null) && (_zoomStep != _template._zoomStep)))
			pw.print(" zoomStep='" + _zoomStep + "'");
		if (((_template == null) && (_zoomFactor != getFloatValue("MAP.zoomFactor", 0F)))
				|| ((_template != null) && (_zoomFactor != _template._zoomFactor)))
			pw.print(" zoomFactor='" + _zoomFactor + "'");
		if (((_template == null) && !_translation)
				|| ((_template != null) && (_translation != _template._translation)))
			pw.print(" translation='" + _translation + "'");
		if (((_template == null) && (_translationStep != getIntValue("MAP.translationStep", -1)))
				|| ((_template != null) && (_translationStep != _template._translationStep)))
			pw.print(" translationStep='" + _translationStep + "'");
		if ((_template == null)
			|| (_translationStatus != _template._translationStatus))
		{
			if ((_translationStatus == TranslationStatus.SHOW) && (_template != null))
				pw.print(" translationStatus='SHOW'");
			else if (_translationStatus == TranslationStatus.HIDE)
				pw.print(" translationStatus='HIDE'");
		}
		if (((_template == null) && _spatial)
				|| ((_template != null) && (_spatial != _template._spatial)))
			pw.print(" spatial='" + _spatial + "'");
		if (((_template == null) && (_background != null))
				|| ((_template != null) && (_background != _template._background)))
			pw.print(" background='" + LyString.getXmlString(_background) + "'");
		if (((_template == null) && (_titleFont != getStringValue("MAP.titleFont", "$ALL.titleFont")))
				|| ((_template != null) && (_titleFont != _template._titleFont)))
			pw.print(" titleFont='" + LyString.getXmlString(_titleFont) + "'");
		if (((_template == null) && (_color != getStringValue("MAP.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("MAP.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && (_selectedColor != getStringValue("MAP.selectedColor", "$ALL.selectedColor")))
				|| ((_template != null) && (_selectedColor != _template._selectedColor)))
			pw.print(" selectedColor='" + LyString.getXmlString(_selectedColor) + "'");
		if (((_template == null) && (_outlineColor != getStringValue("MAP.outlineColor", "$ALL.outlineColor")))
				|| ((_template != null) && (_outlineColor != _template._outlineColor)))
			pw.print(" outlineColor='" + LyString.getXmlString(_outlineColor) + "'");
		if (((_template == null) && (_width != getIntValue("MAP.width", -1)))
				|| ((_template != null) && (_width != _template._width)))
			pw.print(" width='" + _width + "'");
		if (((_template == null) && (_height != getIntValue("MAP.height", -1)))
				|| ((_template != null) && (_height != _template._height)))
			pw.print(" height='" + _height + "'");
		if ((_template == null)
			|| (_style != _template._style))
		{
			if ((_style == Style.STANDARD) && (_template != null))
				pw.print(" style='STANDARD'");
			else if (_style == Style.HYBRID)
				pw.print(" style='HYBRID'");
			else if (_style == Style.ROADMAP)
				pw.print(" style='ROADMAP'");
			else if (_style == Style.SATELLITE)
				pw.print(" style='SATELLITE'");
			else if (_style == Style.TERRAIN)
				pw.print(" style='TERRAIN'");
		}
		if (((_template == null) && (_latitude != null))
				|| ((_template != null) && (_latitude != _template._latitude)))
			pw.print(" latitude='" + LyString.getXmlString(_latitude) + "'");
		if (((_template == null) && (_longitude != null))
				|| ((_template != null) && (_longitude != _template._longitude)))
			pw.print(" longitude='" + LyString.getXmlString(_longitude) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	components;
		int			i, n;

		n = (_components == null ? 0 : _components.size());
		for (i = 0; i < n; i++)
		{
			components = _components.get(i);
			components.writeXml(pw, debug);
		}

		if (_mapContent != null)
			_mapContent.writeXml(pw, debug);
		if (_layout != null)
			_layout.writeXml(pw, debug);
		if (_constraint != null)
			_constraint.writeXml(pw, debug);

		writeDatas(pw);

		pw.println("</MAP>");
	}
}
