/*
 * LyMapLink.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyMapLink class is a generated Java implementation that corresponds to the element 
 * MAP_LINK of the descriptive grammar of the views (views.dtd).
 * <B>&lt;MAP_LINK&gt; : Map link</B>
 * This element represents a link between two nodes in a map.
 * @see leon.view.LyComponent
 * See also tag <A href="../../conf/dtd/view/MAP_LINK.html">MAP_LINK</A> of grammar leon_views.dtd.
 */
public class LyMapLink extends LyComponent
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface,
				LySetSelectedInterface,
				LySetHiddenInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'style'. */
	public enum Style
	{
		LINE,
		DOT,
		DASHED,
		TUBE;
	}


	/** Possible values for attribute 'arrow'. */
	public enum Arrow
	{
		NONE,
		START,
		END,
		BOTH;
	}


	/** Possible values for attribute 'startArrowType'. */
	public enum ArrowType
	{
		TRIANGLE,
		CIRCLE,
		SQUARE;
	}


	/** Possible values for attribute 'startArrowFillMode'. */
	public enum ArrowFill
	{
		EMPTY,
		FULL;
	}


	/** Possible values for attribute 'endArrowType'. */


	/** Possible values for attribute 'endArrowFillMode'. */


	/* Static methods */

	/**
	 * Utility method to get style value from a string.
	 * @param token Token string
	 * @return Style value corresponding to this token
	 */
	public static final Style getStyle(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("LINE")
				|| token.equals("0"))
			return Style.LINE;
		else if (token.equalsIgnoreCase("DOT")
				|| token.equals("1"))
			return Style.DOT;
		else if (token.equalsIgnoreCase("DASHED")
				|| token.equals("2"))
			return Style.DASHED;
		else if (token.equalsIgnoreCase("TUBE")
				|| token.equals("3"))
			return Style.TUBE;
		else
			return Style.LINE;
	}

	/**
	 * Utility method to get arrow value from a string.
	 * @param token Token string
	 * @return Arrow value corresponding to this token
	 */
	public static final Arrow getArrow(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("NONE")
				|| token.equals("0"))
			return Arrow.NONE;
		else if (token.equalsIgnoreCase("START")
				|| token.equals("1"))
			return Arrow.START;
		else if (token.equalsIgnoreCase("END")
				|| token.equals("2"))
			return Arrow.END;
		else if (token.equalsIgnoreCase("BOTH")
				|| token.equals("3"))
			return Arrow.BOTH;
		else
			return Arrow.NONE;
	}

	/**
	 * Utility method to get startArrowType value from a string.
	 * @param token Token string
	 * @return ArrowType value corresponding to this token
	 */
	public static final ArrowType getArrowType(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("TRIANGLE")
				|| token.equals("0"))
			return ArrowType.TRIANGLE;
		else if (token.equalsIgnoreCase("CIRCLE")
				|| token.equals("1"))
			return ArrowType.CIRCLE;
		else if (token.equalsIgnoreCase("SQUARE")
				|| token.equals("2"))
			return ArrowType.SQUARE;
		else
			return ArrowType.TRIANGLE;
	}

	/**
	 * Utility method to get startArrowFillMode value from a string.
	 * @param token Token string
	 * @return ArrowFill value corresponding to this token
	 */
	public static final ArrowFill getArrowFill(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("EMPTY")
				|| token.equals("0"))
			return ArrowFill.EMPTY;
		else if (token.equalsIgnoreCase("FULL")
				|| token.equals("1"))
			return ArrowFill.FULL;
		else
			return ArrowFill.FULL;
	}


	/* Attributes */

	/**
	 * Identifier of the link.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyMapLink			_template;

	/**
	 * Label of the link.
	 */
	protected String			_label;

	/**
	 * Start label of the link.
	 */
	protected String			_startLabel;

	/**
	 * End label of the link.
	 */
	protected String			_endLabel;

	/**
	 * Tooltip of the link.
	 */
	protected String			_tooltip;

	/**
	 * Gives the overview label of the map link.
	 */
	protected String			_overview;

	/**
	 * Start node of the link.
	 * A link has always a start and a destination node.
	 * The expected value is the identifier of a node of the map.
	 */
	protected String			_startNode;

	/**
	 * Destination node of the link.
	 * The expected value is the identifier of a node.
	 */
	protected String			_endNode;

	/**
	 * Indicates whether the link is sensitive or not.
	 * A sensitive link can be selected and allows launching actions (by 
	 * double-clicking for example).
	 */
	protected boolean			_sensitive;

	/**
	 * Indicates whether the link is selected or not.
	 * The value of this attribute is used only if the link is sensitive.
	 */
	protected boolean			_selected;

	/**
	 * Indicates if the line representing the link should be drawn with a double or a 
	 * simple line (the true value indicates a double line).
	 * A link of TUBE type cannot be drawn with a double line.
	 */
	protected boolean			_doubleLink;

	/**
	 * Width of the link.
	 * If the link is represented by a double line, the value of this attribute defines
	 * the space between the two lines.
	 */
	protected int				_width;

	/**
	 * Style of the link. The following styles are managed:
	 * LINE Link represented by a continuous line (or by a double line if the 
	 * doubleLink attribute is true).
	 * DOT Link represented by a dotted line (or by two lines if the value of the width
	 * attribute is positive).
	 * DASHED Link represented by a dashed line (or by two lines if the value of the 
	 * width attribute is positive).
	 * TUBE Link represented with a tube appearance.
	 */
	protected Style				_style;

	/**
	 * Indicates if the link should starts or ends with arrows. The NONE value 
	 * corresponds to no arrow, the START value corresponds to an arrow on the side of 
	 * the source node, the END value corresponds to an arrow on the side of the 
	 * destination node and the BOTH value corresponds to arrows on both sides.
	 */
	protected Arrow				_arrow;

	/**
	 * Style of the start arrow. The following style are managed:
	 * TRIANGLE the start arrow of the link is a triangle (Default).
	 * CIRCLE the start arrow of the link is a circle.
	 * SQUARE the start arrow of the link is a square. 
	 */
	protected ArrowType			_startArrowType;

	/**
	 * The fill mode of the start arrow. The following fill mode are managed:
	 * FULL the arrow is filled with the link color (Default).
	 * EMPTY the arrow is not filled.
	 */
	protected ArrowFill			_startArrowFillMode;

	/**
	 * Style of the end arrow. The following style are managed:
	 * TRIANGLE the end arrow of the link is a triangle (Default).
	 * CIRCLE the end arrow of the link is a circle.
	 * SQUARE the end arrow of the link is a square. 
	 */
	protected ArrowType			_endArrowType;

	/**
	 * The fill mode of the end arrow. The following fill mode are managed:
	 * FULL the arrow is filled with the link color (Default).
	 * EMPTY the arrow is not filled.
	 */
	protected ArrowFill			_endArrowFillMode;

	/**
	 * Repetitive pattern used to draw the link
	 */
	protected String			_motif;

	/**
	 * Source image of the link (displayed in the middle of the first third of the 
	 * link).
	 */
	protected String			_imageStart;

	/**
	 * Source image of the link (displayed in the middle of the link).
	 */
	protected String			_imageMiddle;

	/**
	 * Source image of the link (displayed in the middle of the last third of the 
	 * link).
	 */
	protected String			_imageEnd;

	/**
	 * Color of the first third of the link.
	 */
	protected String			_colorStart;

	/**
	 * Color of the last third of the link.
	 */
	protected String			_colorEnd;

	/**
	 * Layer of the link.
	 */
	protected int				_layer;

	/**
	 * Indicates if the link is hidden (not displayed) or not.
	 */
	protected boolean			_hidden;

	/**
	 * Specifies if the image on the beginning of the link should blink.
	 */
	protected boolean			_imageStartBlink;

	/**
	 * Specifies if the image in the middle of the link should blink.
	 */
	protected boolean			_imageMiddleBlink;

	/**
	 * Specifies if the image at the end of the link should blink.
	 */
	protected boolean			_imageEndBlink;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Color of the link.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * MAP_LINK.selectedColor
	 */
	protected String			_selectedColor;

	/**
	 * MAP_LINK.inactiveSelectedColor
	 */
	protected String			_inactiveSelectedColor;

	/**
	 * MAP_LINK.outlineColor
	 */
	protected String			_outlineColor;


	/* Sub components */

	/**
	 * Points used to draw the link when a specific drawing is needed between start and
	 * end nodes.
	 */
	protected ArrayList<LyPoint>		_points;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_startLabel = null;
		_endLabel = null;
		_tooltip = null;
		_overview = null;
		_startNode = null;
		_endNode = null;
		_sensitive = getBooleanValue("MAP_LINK.sensitive", true);
		_selected = getBooleanValue("MAP_LINK.selected", false);
		_doubleLink = getBooleanValue("MAP_LINK.doubleLink", false);
		_width = getIntValue("MAP_LINK.width", -1);
		_style = Style.LINE;
		_arrow = Arrow.NONE;
		_startArrowType = ArrowType.TRIANGLE;
		_startArrowFillMode = ArrowFill.FULL;
		_endArrowType = ArrowType.TRIANGLE;
		_endArrowFillMode = ArrowFill.FULL;
		_motif = null;
		_imageStart = null;
		_imageMiddle = null;
		_imageEnd = null;
		_colorStart = null;
		_colorEnd = null;
		_layer = getIntValue("MAP_LINK.layer", -1);
		_hidden = getBooleanValue("MAP_LINK.hidden", false);
		_imageStartBlink = getBooleanValue("MAP_LINK.imageStartBlink", false);
		_imageMiddleBlink = getBooleanValue("MAP_LINK.imageMiddleBlink", false);
		_imageEndBlink = getBooleanValue("MAP_LINK.imageEndBlink", false);
		_font = getStringValue("MAP_LINK.font", "$ALL.font");
		_color = getStringValue("MAP_LINK.color", "$ALL.color");
		_fontColor = getStringValue("MAP_LINK.fontColor", "$ALL.fontColor");
		_selectedColor = getStringValue("MAP_LINK.selectedColor", "$ALL.selectedColor");
		_inactiveSelectedColor = getStringValue("MAP_LINK.inactiveSelectedColor", "$ALL.inactiveSelectedColor");
		_outlineColor = getStringValue("MAP_LINK.outlineColor", "$ALL.outlineColor");

		// Initialize sub components
		_points = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyMapLink	model = (LyMapLink)component;

		_model = component;
		_label = model._label;
		_startLabel = model._startLabel;
		_endLabel = model._endLabel;
		_tooltip = model._tooltip;
		_overview = model._overview;
		_startNode = model._startNode;
		_endNode = model._endNode;
		_sensitive = model._sensitive;
		_selected = model._selected;
		_doubleLink = model._doubleLink;
		_width = model._width;
		_style = model._style;
		_arrow = model._arrow;
		_startArrowType = model._startArrowType;
		_startArrowFillMode = model._startArrowFillMode;
		_endArrowType = model._endArrowType;
		_endArrowFillMode = model._endArrowFillMode;
		_motif = model._motif;
		_imageStart = model._imageStart;
		_imageMiddle = model._imageMiddle;
		_imageEnd = model._imageEnd;
		_colorStart = model._colorStart;
		_colorEnd = model._colorEnd;
		_layer = model._layer;
		_hidden = model._hidden;
		_imageStartBlink = model._imageStartBlink;
		_imageMiddleBlink = model._imageMiddleBlink;
		_imageEndBlink = model._imageEndBlink;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;
		_selectedColor = model._selectedColor;
		_inactiveSelectedColor = model._inactiveSelectedColor;
		_outlineColor = model._outlineColor;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyPoint	points;
		int			i, n;

		n = (model._points == null ? 0 : model._points.size());
		for (i = 0; i < n; i++)
		{
			points = model._points.get(i);
			if (points != null)
				points = (LyPoint)points.cloneComponent();
			addChild(points);
		}
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	protected LyMapLink()
	{
	}

	/**
	 * Constructor with all required attributes.
	 * @param startNode Start node of the link.<BR>A link has always a start and a destination node.<BR>The expected value is the identifier of a node of the map.
	 * @param endNode Destination node of the link.<BR>The expected value is the identifier of a node.
	 */
	public LyMapLink(String startNode, String endNode)
	{
		// Fill up required attributes with parameters
		_startNode = startNode;
		_endNode = endNode;
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyMapLink(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyMapLink)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("startLabel"))
			_startLabel = node.getString("startLabel");
		if (node.isSpecified("endLabel"))
			_endLabel = node.getString("endLabel");
		if (node.isSpecified("tooltip"))
			_tooltip = node.getString("tooltip");
		if (node.isSpecified("overview"))
			_overview = node.getString("overview");
		if (node.isSpecified("startNode"))
			_startNode = node.getString("startNode");
		if (node.isSpecified("endNode"))
			_endNode = node.getString("endNode");
		if (node.isSpecified("sensitive"))
			_sensitive = node.getBoolean("sensitive");
		if (node.isSpecified("selected"))
			_selected = node.getBoolean("selected");
		if (node.isSpecified("doubleLink"))
			_doubleLink = node.getBoolean("doubleLink");
		if (node.isSpecified("width"))
			_width = node.getInt("width");
		if (node.isSpecified("style"))
			_style = getStyle(node.getString("style"));
		if (node.isSpecified("arrow"))
			_arrow = getArrow(node.getString("arrow"));
		if (node.isSpecified("startArrowType"))
			_startArrowType = getArrowType(node.getString("startArrowType"));
		if (node.isSpecified("startArrowFillMode"))
			_startArrowFillMode = getArrowFill(node.getString("startArrowFillMode"));
		if (node.isSpecified("endArrowType"))
			_endArrowType = getArrowType(node.getString("endArrowType"));
		if (node.isSpecified("endArrowFillMode"))
			_endArrowFillMode = getArrowFill(node.getString("endArrowFillMode"));
		if (node.isSpecified("motif"))
			_motif = node.getString("motif");
		if (node.isSpecified("imageStart"))
			_imageStart = node.getString("imageStart");
		if (node.isSpecified("imageMiddle"))
			_imageMiddle = node.getString("imageMiddle");
		if (node.isSpecified("imageEnd"))
			_imageEnd = node.getString("imageEnd");
		if (node.isSpecified("colorStart"))
			_colorStart = node.getString("colorStart");
		if (node.isSpecified("colorEnd"))
			_colorEnd = node.getString("colorEnd");
		if (node.isSpecified("layer"))
			_layer = node.getInt("layer");
		if (node.isSpecified("hidden"))
			_hidden = node.getBoolean("hidden");
		if (node.isSpecified("imageStartBlink"))
			_imageStartBlink = node.getBoolean("imageStartBlink");
		if (node.isSpecified("imageMiddleBlink"))
			_imageMiddleBlink = node.getBoolean("imageMiddleBlink");
		if (node.isSpecified("imageEndBlink"))
			_imageEndBlink = node.getBoolean("imageEndBlink");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("selectedColor"))
			_selectedColor = node.getString("selectedColor");
		if (node.isSpecified("inactiveSelectedColor"))
			_inactiveSelectedColor = node.getString("inactiveSelectedColor");
		if (node.isSpecified("outlineColor"))
			_outlineColor = node.getString("outlineColor");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyPoint)
			addPoint((LyPoint)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyMapLink clone = new LyMapLink();

		clone.initFromModel(this);
		clone._template = this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyMapLink getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of startLabel.
	 * @return value of startLabel.
	 */
	public String getStartLabel()
	{
		return _startLabel;
	}

	/**
	 * Gets value of endLabel.
	 * @return value of endLabel.
	 */
	public String getEndLabel()
	{
		return _endLabel;
	}

	/**
	 * Gets value of tooltip.
	 * @return value of tooltip.
	 */
	public String getTooltip()
	{
		return _tooltip;
	}

	/**
	 * Gets value of overview.
	 * @return value of overview.
	 */
	public String getOverview()
	{
		return _overview;
	}

	/**
	 * Gets value of startNode.
	 * @return value of startNode.
	 */
	public String getStartNode()
	{
		return _startNode;
	}

	/**
	 * Gets value of endNode.
	 * @return value of endNode.
	 */
	public String getEndNode()
	{
		return _endNode;
	}

	/**
	 * Checks if value of attribute sensitive is set.
	 * @return true if sensitive is set, false otherwise.
	 */
	public boolean isSensitive()
	{
		return _sensitive;
	}

	/**
	 * Checks if value of attribute selected is set.
	 * @return true if selected is set, false otherwise.
	 */
	public boolean isSelected()
	{
		return _selected;
	}

	/**
	 * Checks if value of attribute doubleLink is set.
	 * @return true if doubleLink is set, false otherwise.
	 */
	public boolean isDoubleLink()
	{
		return _doubleLink;
	}

	/**
	 * Gets value of width.
	 * @return value of width.
	 */
	public int getWidth()
	{
		return _width;
	}

	/**
	 * Gets value of style.
	 * @return value of style.
	 */
	public Style getStyle()
	{
		return _style;
	}

	/**
	 * Gets value of arrow.
	 * @return value of arrow.
	 */
	public Arrow getArrow()
	{
		return _arrow;
	}

	/**
	 * Gets value of startArrowType.
	 * @return value of startArrowType.
	 */
	public ArrowType getStartArrowType()
	{
		return _startArrowType;
	}

	/**
	 * Gets value of startArrowFillMode.
	 * @return value of startArrowFillMode.
	 */
	public ArrowFill getStartArrowFillMode()
	{
		return _startArrowFillMode;
	}

	/**
	 * Gets value of endArrowType.
	 * @return value of endArrowType.
	 */
	public ArrowType getEndArrowType()
	{
		return _endArrowType;
	}

	/**
	 * Gets value of endArrowFillMode.
	 * @return value of endArrowFillMode.
	 */
	public ArrowFill getEndArrowFillMode()
	{
		return _endArrowFillMode;
	}

	/**
	 * Gets value of motif.
	 * @return value of motif.
	 */
	public String getMotif()
	{
		return _motif;
	}

	/**
	 * Gets value of imageStart.
	 * @return value of imageStart.
	 */
	public String getImageStart()
	{
		return _imageStart;
	}

	/**
	 * Gets value of imageMiddle.
	 * @return value of imageMiddle.
	 */
	public String getImageMiddle()
	{
		return _imageMiddle;
	}

	/**
	 * Gets value of imageEnd.
	 * @return value of imageEnd.
	 */
	public String getImageEnd()
	{
		return _imageEnd;
	}

	/**
	 * Gets value of colorStart.
	 * @return value of colorStart.
	 */
	public String getColorStart()
	{
		return _colorStart;
	}

	/**
	 * Gets value of colorEnd.
	 * @return value of colorEnd.
	 */
	public String getColorEnd()
	{
		return _colorEnd;
	}

	/**
	 * Gets value of layer.
	 * @return value of layer.
	 */
	public int getLayer()
	{
		return _layer;
	}

	/**
	 * Checks if value of attribute hidden is set.
	 * @return true if hidden is set, false otherwise.
	 */
	public boolean isHidden()
	{
		return _hidden;
	}

	/**
	 * Checks if value of attribute imageStartBlink is set.
	 * @return true if imageStartBlink is set, false otherwise.
	 */
	public boolean isImageStartBlink()
	{
		return _imageStartBlink;
	}

	/**
	 * Checks if value of attribute imageMiddleBlink is set.
	 * @return true if imageMiddleBlink is set, false otherwise.
	 */
	public boolean isImageMiddleBlink()
	{
		return _imageMiddleBlink;
	}

	/**
	 * Checks if value of attribute imageEndBlink is set.
	 * @return true if imageEndBlink is set, false otherwise.
	 */
	public boolean isImageEndBlink()
	{
		return _imageEndBlink;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("MAP_LINK.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("MAP_LINK.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("MAP_LINK.fontColor"));
	}

	/**
	 * Gets value of selectedColor.
	 * @return value of selectedColor.
	 */
	public String getSelectedColor()
	{
		return _selectedColor;
	}

	/**
	 * Checks whether the selectedColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultSelectedColor()
	{
		return (_selectedColor == getStringValue("MAP_LINK.selectedColor"));
	}

	/**
	 * Gets value of inactiveSelectedColor.
	 * @return value of inactiveSelectedColor.
	 */
	public String getInactiveSelectedColor()
	{
		return _inactiveSelectedColor;
	}

	/**
	 * Checks whether the inactiveSelectedColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultInactiveSelectedColor()
	{
		return (_inactiveSelectedColor == getStringValue("MAP_LINK.inactiveSelectedColor"));
	}

	/**
	 * Gets value of outlineColor.
	 * @return value of outlineColor.
	 */
	public String getOutlineColor()
	{
		return _outlineColor;
	}

	/**
	 * Checks whether the outlineColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultOutlineColor()
	{
		return (_outlineColor == getStringValue("MAP_LINK.outlineColor"));
	}

	/**
	 * Get sub-component Point of index i.
	 * @param i the index of the Point to get.
	 * @return sub-component Point of index i.
	 */
	public LyPoint getPoint(int i)
	{
		if (_points == null)
			return null;
		if ((i < 0) || (i >= _points.size()))
			return null;
		return _points.get(i);
	}

	/**
	 * Get sub-component Point count.
	 * @return sub-component Point count.
	 */
	public int getPointCount()
	{
		if (_points == null)
			return 0;
		return _points.size();
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyMapLink template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView != null)
			((LyGraphicalMapLink)_internalView).setLabel(label);
	}

	/**
	 * Sets attribute startLabel to the given value.
	 * @param startLabel New value of startLabel.
	 */
	public void setStartLabel(String startLabel)
	{
		if ((startLabel == _startLabel)
			|| ((startLabel != null)
				&& (_startLabel != null)
				&& startLabel.equals(_startLabel)))
			return;

		_startLabel = startLabel;
		if (_internalView != null)
			((LyGraphicalMapLink)_internalView).setStartLabel(startLabel);
	}

	/**
	 * Sets attribute endLabel to the given value.
	 * @param endLabel New value of endLabel.
	 */
	public void setEndLabel(String endLabel)
	{
		if ((endLabel == _endLabel)
			|| ((endLabel != null)
				&& (_endLabel != null)
				&& endLabel.equals(_endLabel)))
			return;

		_endLabel = endLabel;
		if (_internalView != null)
			((LyGraphicalMapLink)_internalView).setEndLabel(endLabel);
	}

	/**
	 * Initializes attribute tooltip to the given value.
	 * @param tooltip New value of tooltip.
	 */
	public void setTooltip(String tooltip)
	{
		if ((tooltip == _tooltip)
			|| ((tooltip != null)
				&& (_tooltip != null)
				&& tooltip.equals(_tooltip)))
			return;

		_tooltip = tooltip;
	}

	/**
	 * Initializes attribute overview to the given value.
	 * @param overview New value of overview.
	 */
	public void setOverview(String overview)
	{
		if ((overview == _overview)
			|| ((overview != null)
				&& (_overview != null)
				&& overview.equals(_overview)))
			return;

		_overview = overview;
	}

	/**
	 * Initializes attribute startNode to the given value.
	 * @param startNode New value of startNode.
	 * @exception IllegalArgumentException if startNode is null.
	 */
	public void setStartNode(String startNode)
	{
		if (startNode == null)
			throw new IllegalArgumentException("BF AppEngine : attribute startNode is required (can't be null)");

		if ((startNode == _startNode)
			|| ((startNode != null)
				&& (_startNode != null)
				&& startNode.equals(_startNode)))
			return;

		_startNode = startNode;
	}

	/**
	 * Initializes attribute endNode to the given value.
	 * @param endNode New value of endNode.
	 * @exception IllegalArgumentException if endNode is null.
	 */
	public void setEndNode(String endNode)
	{
		if (endNode == null)
			throw new IllegalArgumentException("BF AppEngine : attribute endNode is required (can't be null)");

		if ((endNode == _endNode)
			|| ((endNode != null)
				&& (_endNode != null)
				&& endNode.equals(_endNode)))
			return;

		_endNode = endNode;
	}

	/**
	 * Initializes attribute sensitive to the given value.
	 * @param sensitive Boolean indicating if attribute sensitive is set or unset.
	 */
	public void setSensitive(boolean sensitive)
	{
		if (sensitive == _sensitive)
			return;
		_sensitive = sensitive;
	}

	/**
	 * Sets attribute selected to the given value.
	 * @param selected Boolean indicating if attribute selected is set or unset.
	 */
	public void setSelected(boolean selected)
	{
		if (selected == _selected)
			return;
		_selected = selected;
		if (_internalView != null)
			((LyGraphicalMapLink)_internalView).setSelected(selected);
	}

	/**
	 * Initializes attribute doubleLink to the given value.
	 * @param doubleLink Boolean indicating if attribute doubleLink is set or unset.
	 */
	public void setDoubleLink(boolean doubleLink)
	{
		if (doubleLink == _doubleLink)
			return;
		_doubleLink = doubleLink;
	}

	/**
	 * Initializes attribute width to the given value.
	 * @param width New value of width.
	 */
	public void setWidth(int width)
	{
		if (width == _width)
			return;
		_width = width;
	}

	/**
	 * Initializes attribute style to the given value.
	 * @param style New value of style.
	 * @exception IllegalArgumentException if style parameter is null.
	 */
	public void setStyle(Style style)
	{
		if (style == _style)
			return;
		if (style == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute style");

		_style = style;
	}

	/**
	 * Initializes attribute arrow to the given value.
	 * @param arrow New value of arrow.
	 * @exception IllegalArgumentException if arrow parameter is null.
	 */
	public void setArrow(Arrow arrow)
	{
		if (arrow == _arrow)
			return;
		if (arrow == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute arrow");

		_arrow = arrow;
	}

	/**
	 * Initializes attribute startArrowType to the given value.
	 * @param startArrowType New value of startArrowType.
	 * @exception IllegalArgumentException if startArrowType parameter is null.
	 */
	public void setStartArrowType(ArrowType startArrowType)
	{
		if (startArrowType == _startArrowType)
			return;
		if (startArrowType == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute startArrowType");

		_startArrowType = startArrowType;
	}

	/**
	 * Initializes attribute startArrowFillMode to the given value.
	 * @param startArrowFillMode New value of startArrowFillMode.
	 * @exception IllegalArgumentException if startArrowFillMode parameter is null.
	 */
	public void setStartArrowFillMode(ArrowFill startArrowFillMode)
	{
		if (startArrowFillMode == _startArrowFillMode)
			return;
		if (startArrowFillMode == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute startArrowFillMode");

		_startArrowFillMode = startArrowFillMode;
	}

	/**
	 * Initializes attribute endArrowType to the given value.
	 * @param endArrowType New value of endArrowType.
	 * @exception IllegalArgumentException if endArrowType parameter is null.
	 */
	public void setEndArrowType(ArrowType endArrowType)
	{
		if (endArrowType == _endArrowType)
			return;
		if (endArrowType == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute endArrowType");

		_endArrowType = endArrowType;
	}

	/**
	 * Initializes attribute endArrowFillMode to the given value.
	 * @param endArrowFillMode New value of endArrowFillMode.
	 * @exception IllegalArgumentException if endArrowFillMode parameter is null.
	 */
	public void setEndArrowFillMode(ArrowFill endArrowFillMode)
	{
		if (endArrowFillMode == _endArrowFillMode)
			return;
		if (endArrowFillMode == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute endArrowFillMode");

		_endArrowFillMode = endArrowFillMode;
	}

	/**
	 * Initializes attribute motif to the given value.
	 * @param motif New value of motif.
	 */
	public void setMotif(String motif)
	{
		if ((motif == _motif)
			|| ((motif != null)
				&& (_motif != null)
				&& motif.equals(_motif)))
			return;

		_motif = motif;
	}

	/**
	 * Initializes attribute imageStart to the given value.
	 * @param imageStart New value of imageStart.
	 */
	public void setImageStart(String imageStart)
	{
		if ((imageStart == _imageStart)
			|| ((imageStart != null)
				&& (_imageStart != null)
				&& imageStart.equals(_imageStart)))
			return;

		_imageStart = imageStart;
	}

	/**
	 * Initializes attribute imageMiddle to the given value.
	 * @param imageMiddle New value of imageMiddle.
	 */
	public void setImageMiddle(String imageMiddle)
	{
		if ((imageMiddle == _imageMiddle)
			|| ((imageMiddle != null)
				&& (_imageMiddle != null)
				&& imageMiddle.equals(_imageMiddle)))
			return;

		_imageMiddle = imageMiddle;
	}

	/**
	 * Initializes attribute imageEnd to the given value.
	 * @param imageEnd New value of imageEnd.
	 */
	public void setImageEnd(String imageEnd)
	{
		if ((imageEnd == _imageEnd)
			|| ((imageEnd != null)
				&& (_imageEnd != null)
				&& imageEnd.equals(_imageEnd)))
			return;

		_imageEnd = imageEnd;
	}

	/**
	 * Sets attribute colorStart to the given value.
	 * @param colorStart New value of colorStart.
	 */
	public void setColorStart(String colorStart)
	{
		if ((colorStart == _colorStart)
			|| ((colorStart != null)
				&& (_colorStart != null)
				&& colorStart.equals(_colorStart)))
			return;

		_colorStart = colorStart;
		if (_internalView != null)
			((LyGraphicalMapLink)_internalView).setColorStart(colorStart);
	}

	/**
	 * Sets attribute colorEnd to the given value.
	 * @param colorEnd New value of colorEnd.
	 */
	public void setColorEnd(String colorEnd)
	{
		if ((colorEnd == _colorEnd)
			|| ((colorEnd != null)
				&& (_colorEnd != null)
				&& colorEnd.equals(_colorEnd)))
			return;

		_colorEnd = colorEnd;
		if (_internalView != null)
			((LyGraphicalMapLink)_internalView).setColorEnd(colorEnd);
	}

	/**
	 * Initializes attribute layer to the given value.
	 * @param layer New value of layer.
	 */
	public void setLayer(int layer)
	{
		if (layer == _layer)
			return;
		_layer = layer;
	}

	/**
	 * Sets attribute hidden to the given value.
	 * @param hidden Boolean indicating if attribute hidden is set or unset.
	 */
	public void setHidden(boolean hidden)
	{
		if (hidden == _hidden)
			return;
		_hidden = hidden;
		if (_internalView != null)
			((LyGraphicalMapLink)_internalView).setHidden(hidden);
	}

	/**
	 * Initializes attribute imageStartBlink to the given value.
	 * @param imageStartBlink Boolean indicating if attribute imageStartBlink is set or unset.
	 */
	public void setImageStartBlink(boolean imageStartBlink)
	{
		if (imageStartBlink == _imageStartBlink)
			return;
		_imageStartBlink = imageStartBlink;
	}

	/**
	 * Initializes attribute imageMiddleBlink to the given value.
	 * @param imageMiddleBlink Boolean indicating if attribute imageMiddleBlink is set or unset.
	 */
	public void setImageMiddleBlink(boolean imageMiddleBlink)
	{
		if (imageMiddleBlink == _imageMiddleBlink)
			return;
		_imageMiddleBlink = imageMiddleBlink;
	}

	/**
	 * Initializes attribute imageEndBlink to the given value.
	 * @param imageEndBlink Boolean indicating if attribute imageEndBlink is set or unset.
	 */
	public void setImageEndBlink(boolean imageEndBlink)
	{
		if (imageEndBlink == _imageEndBlink)
			return;
		_imageEndBlink = imageEndBlink;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Initializes attribute selectedColor to the given value.
	 * @param selectedColor New value of selectedColor.
	 */
	public void setSelectedColor(String selectedColor)
	{
		if ((selectedColor == _selectedColor)
			|| ((selectedColor != null)
				&& (_selectedColor != null)
				&& selectedColor.equals(_selectedColor)))
			return;

		_selectedColor = selectedColor;
	}

	/**
	 * Initializes attribute inactiveSelectedColor to the given value.
	 * @param inactiveSelectedColor New value of inactiveSelectedColor.
	 */
	public void setInactiveSelectedColor(String inactiveSelectedColor)
	{
		if ((inactiveSelectedColor == _inactiveSelectedColor)
			|| ((inactiveSelectedColor != null)
				&& (_inactiveSelectedColor != null)
				&& inactiveSelectedColor.equals(_inactiveSelectedColor)))
			return;

		_inactiveSelectedColor = inactiveSelectedColor;
	}

	/**
	 * Initializes attribute outlineColor to the given value.
	 * @param outlineColor New value of outlineColor.
	 */
	public void setOutlineColor(String outlineColor)
	{
		if ((outlineColor == _outlineColor)
			|| ((outlineColor != null)
				&& (_outlineColor != null)
				&& outlineColor.equals(_outlineColor)))
			return;

		_outlineColor = outlineColor;
	}

	/**
	 * Adds a sub-component point.
	 * @param point The point to add to points list.
	 */
	public void addPoint(LyPoint point)
	{
		insertPoint(point, -1);
	}

	/**
	 * Inserts a sub-component point at given position.
	 * @param point The point to add to points list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertPoint(LyPoint point, int pos)
	{
		if (_points == null)
		{
			_points = new ArrayList<LyPoint>(1);
			_children = _points;
		}
		else if (_points.contains(point))
			return;
		if (pos >= 0)
			_points.add(pos,point);
		if (point != null)
			point.setParent(this);
		if (_internalView != null)
			((LyGraphicalMapLink)_internalView).addPoint(point, pos);
	}

	/**
	 * Removes a sub-component point.
	 * @param point The point to remove to points list.
	 * @exception IllegalStateException if points list is empty.
	 * @exception IllegalArgumentException if points list doesn't contain point.
	 */
	public void removePoint(LyPoint point)
	{
		if (_points == null)
			throw new IllegalStateException("BF AppEngine : points list is empty");
		if (!_points.contains(point))
			throw new IllegalArgumentException("BF AppEngine : points list doesn't contain point");

		if (_internalView != null)
			((LyGraphicalMapLink)_internalView).removePoint(point);
		point.setParent(null);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<MAP_LINK");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (((_template == null) && (_startLabel != null))
				|| ((_template != null) && (_startLabel != _template._startLabel)))
			pw.print(" startLabel='" + LyString.getXmlString(_startLabel) + "'");
		if (((_template == null) && (_endLabel != null))
				|| ((_template != null) && (_endLabel != _template._endLabel)))
			pw.print(" endLabel='" + LyString.getXmlString(_endLabel) + "'");
		if (((_template == null) && (_tooltip != null))
				|| ((_template != null) && (_tooltip != _template._tooltip)))
			pw.print(" tooltip='" + LyString.getXmlString(_tooltip) + "'");
		if (((_template == null) && (_overview != null))
				|| ((_template != null) && (_overview != _template._overview)))
			pw.print(" overview='" + LyString.getXmlString(_overview) + "'");
		pw.print(" startNode='" + LyString.getXmlString(_startNode) + "'");
		pw.print(" endNode='" + LyString.getXmlString(_endNode) + "'");
		if (((_template == null) && !_sensitive)
				|| ((_template != null) && (_sensitive != _template._sensitive)))
			pw.print(" sensitive='" + _sensitive + "'");
		if (((_template == null) && _selected)
				|| ((_template != null) && (_selected != _template._selected)))
			pw.print(" selected='" + _selected + "'");
		if (((_template == null) && _doubleLink)
				|| ((_template != null) && (_doubleLink != _template._doubleLink)))
			pw.print(" doubleLink='" + _doubleLink + "'");
		if (((_template == null) && (_width != getIntValue("MAP_LINK.width", -1)))
				|| ((_template != null) && (_width != _template._width)))
			pw.print(" width='" + _width + "'");
		if ((_template == null)
			|| (_style != _template._style))
		{
			if ((_style == Style.LINE) && (_template != null))
				pw.print(" style='LINE'");
			else if (_style == Style.DOT)
				pw.print(" style='DOT'");
			else if (_style == Style.DASHED)
				pw.print(" style='DASHED'");
			else if (_style == Style.TUBE)
				pw.print(" style='TUBE'");
		}
		if ((_template == null)
			|| (_arrow != _template._arrow))
		{
			if ((_arrow == Arrow.NONE) && (_template != null))
				pw.print(" arrow='NONE'");
			else if (_arrow == Arrow.START)
				pw.print(" arrow='START'");
			else if (_arrow == Arrow.END)
				pw.print(" arrow='END'");
			else if (_arrow == Arrow.BOTH)
				pw.print(" arrow='BOTH'");
		}
		if ((_template == null)
			|| (_startArrowType != _template._startArrowType))
		{
			if ((_startArrowType == ArrowType.TRIANGLE) && (_template != null))
				pw.print(" startArrowType='TRIANGLE'");
			else if (_startArrowType == ArrowType.CIRCLE)
				pw.print(" startArrowType='CIRCLE'");
			else if (_startArrowType == ArrowType.SQUARE)
				pw.print(" startArrowType='SQUARE'");
		}
		if ((_template == null)
			|| (_startArrowFillMode != _template._startArrowFillMode))
		{
			if (_startArrowFillMode == ArrowFill.EMPTY)
				pw.print(" startArrowFillMode='EMPTY'");
			else if ((_startArrowFillMode == ArrowFill.FULL) && (_template != null))
				pw.print(" startArrowFillMode='FULL'");
		}
		if ((_template == null)
			|| (_endArrowType != _template._endArrowType))
		{
			if ((_endArrowType == ArrowType.TRIANGLE) && (_template != null))
				pw.print(" endArrowType='TRIANGLE'");
			else if (_endArrowType == ArrowType.CIRCLE)
				pw.print(" endArrowType='CIRCLE'");
			else if (_endArrowType == ArrowType.SQUARE)
				pw.print(" endArrowType='SQUARE'");
		}
		if ((_template == null)
			|| (_endArrowFillMode != _template._endArrowFillMode))
		{
			if (_endArrowFillMode == ArrowFill.EMPTY)
				pw.print(" endArrowFillMode='EMPTY'");
			else if ((_endArrowFillMode == ArrowFill.FULL) && (_template != null))
				pw.print(" endArrowFillMode='FULL'");
		}
		if (((_template == null) && (_motif != null))
				|| ((_template != null) && (_motif != _template._motif)))
			pw.print(" motif='" + LyString.getXmlString(_motif) + "'");
		if (((_template == null) && (_imageStart != null))
				|| ((_template != null) && (_imageStart != _template._imageStart)))
			pw.print(" imageStart='" + LyString.getXmlString(_imageStart) + "'");
		if (((_template == null) && (_imageMiddle != null))
				|| ((_template != null) && (_imageMiddle != _template._imageMiddle)))
			pw.print(" imageMiddle='" + LyString.getXmlString(_imageMiddle) + "'");
		if (((_template == null) && (_imageEnd != null))
				|| ((_template != null) && (_imageEnd != _template._imageEnd)))
			pw.print(" imageEnd='" + LyString.getXmlString(_imageEnd) + "'");
		if (((_template == null) && (_colorStart != null))
				|| ((_template != null) && (_colorStart != _template._colorStart)))
			pw.print(" colorStart='" + LyString.getXmlString(_colorStart) + "'");
		if (((_template == null) && (_colorEnd != null))
				|| ((_template != null) && (_colorEnd != _template._colorEnd)))
			pw.print(" colorEnd='" + LyString.getXmlString(_colorEnd) + "'");
		if (((_template == null) && (_layer != getIntValue("MAP_LINK.layer", -1)))
				|| ((_template != null) && (_layer != _template._layer)))
			pw.print(" layer='" + _layer + "'");
		if (((_template == null) && _hidden)
				|| ((_template != null) && (_hidden != _template._hidden)))
			pw.print(" hidden='" + _hidden + "'");
		if (((_template == null) && _imageStartBlink)
				|| ((_template != null) && (_imageStartBlink != _template._imageStartBlink)))
			pw.print(" imageStartBlink='" + _imageStartBlink + "'");
		if (((_template == null) && _imageMiddleBlink)
				|| ((_template != null) && (_imageMiddleBlink != _template._imageMiddleBlink)))
			pw.print(" imageMiddleBlink='" + _imageMiddleBlink + "'");
		if (((_template == null) && _imageEndBlink)
				|| ((_template != null) && (_imageEndBlink != _template._imageEndBlink)))
			pw.print(" imageEndBlink='" + _imageEndBlink + "'");
		if (((_template == null) && (_font != getStringValue("MAP_LINK.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("MAP_LINK.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("MAP_LINK.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && (_selectedColor != getStringValue("MAP_LINK.selectedColor", "$ALL.selectedColor")))
				|| ((_template != null) && (_selectedColor != _template._selectedColor)))
			pw.print(" selectedColor='" + LyString.getXmlString(_selectedColor) + "'");
		if (((_template == null) && (_inactiveSelectedColor != getStringValue("MAP_LINK.inactiveSelectedColor", "$ALL.inactiveSelectedColor")))
				|| ((_template != null) && (_inactiveSelectedColor != _template._inactiveSelectedColor)))
			pw.print(" inactiveSelectedColor='" + LyString.getXmlString(_inactiveSelectedColor) + "'");
		if (((_template == null) && (_outlineColor != getStringValue("MAP_LINK.outlineColor", "$ALL.outlineColor")))
				|| ((_template != null) && (_outlineColor != _template._outlineColor)))
			pw.print(" outlineColor='" + LyString.getXmlString(_outlineColor) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	points;
		int			i, n;

		n = (_points == null ? 0 : _points.size());
		for (i = 0; i < n; i++)
		{
			points = _points.get(i);
			points.writeXml(pw, debug);
		}


		pw.println("</MAP_LINK>");
	}
}
