/*
 * LyMapNode.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyMapNode class is a generated Java implementation that corresponds to the element 
 * MAP_NODE of the descriptive grammar of the views (views.dtd).
 * <B>&lt;MAP_NODE&gt; : Map node</B>
 * This element represents a node displayed in a map.
 * @see leon.view.LyComponent
 * See also tag <A href="../../conf/dtd/view/MAP_NODE.html">MAP_NODE</A> of grammar leon_views.dtd.
 */
public class LyMapNode extends LyComponent
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface,
				LySetSelectedInterface,
				LySetHiddenInterface,
				LySetImageInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'style'. */
	public enum Style
	{
		IMAGE,
		CIRCLE,
		SQUARE,
		TRIANGLE,
		ELLIPSE;
	}


	/** Possible values for attribute 'viewerStyle'. */


	/** Possible values for attribute 'labelAlign'. */
	public enum LabelAlign
	{
		LEFT,
		RIGHT,
		TOP,
		BOTTOM,
		CENTER,
		EXTERN;
	}


	/* Static methods */

	/**
	 * Utility method to get style value from a string.
	 * @param token Token string
	 * @return Style value corresponding to this token
	 */
	public static final Style getStyle(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("IMAGE")
				|| token.equals("0"))
			return Style.IMAGE;
		else if (token.equalsIgnoreCase("CIRCLE")
				|| token.equals("1"))
			return Style.CIRCLE;
		else if (token.equalsIgnoreCase("SQUARE")
				|| token.equals("2"))
			return Style.SQUARE;
		else if (token.equalsIgnoreCase("TRIANGLE")
				|| token.equals("3"))
			return Style.TRIANGLE;
		else if (token.equalsIgnoreCase("ELLIPSE")
				|| token.equals("4"))
			return Style.ELLIPSE;
		else
			return Style.CIRCLE;
	}

	/**
	 * Utility method to get labelAlign value from a string.
	 * @param token Token string
	 * @return LabelAlign value corresponding to this token
	 */
	public static final LabelAlign getLabelAlign(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("LEFT")
				|| token.equals("0"))
			return LabelAlign.LEFT;
		else if (token.equalsIgnoreCase("RIGHT")
				|| token.equals("1"))
			return LabelAlign.RIGHT;
		else if (token.equalsIgnoreCase("TOP")
				|| token.equals("2"))
			return LabelAlign.TOP;
		else if (token.equalsIgnoreCase("BOTTOM")
				|| token.equals("3"))
			return LabelAlign.BOTTOM;
		else if (token.equalsIgnoreCase("CENTER")
				|| token.equals("4"))
			return LabelAlign.CENTER;
		else if (token.equalsIgnoreCase("EXTERN")
				|| token.equals("5"))
			return LabelAlign.EXTERN;
		else
			return LabelAlign.BOTTOM;
	}


	/* Attributes */

	/**
	 * Identifier of the node.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyMapNode			_template;

	/**
	 * Label of the node.
	 */
	protected String			_label;

	/**
	 * Left label of the node.
	 */
	protected String			_leftLabel;

	/**
	 * Right label of the node.
	 */
	protected String			_rightLabel;

	/**
	 * Top label of the node.
	 */
	protected String			_topLabel;

	/**
	 * Bottom label of the node.
	 */
	protected String			_bottomLabel;

	/**
	 * Tooltip of the node.
	 */
	protected String			_tooltip;

	/**
	 * Overview label of the node.
	 */
	protected String			_overview;

	/**
	 * Parent node of the current node.
	 * The parent node is used by certain positioning managers to indicate a link from 
	 * a node to another.
	 */
	protected String			_parentNode;

	/**
	 * Style of the node. The following styles are managed:
	 * IMAGE: Node represented by an image
	 * CIRCLE: Node represented by a circle
	 * SQUARE: Node represented by a square
	 * TRIANGLE: Node represented by a triangle
	 */
	protected Style				_style;

	/**
	 * Style of the node in the viewer. The following styles are managed:
	 * IMAGE: Node represented by an image
	 * CIRCLE: Node represented by a circle
	 * SQUARE: Node represented by a square
	 * TRIANGLE: Node represented by a triangle
	 */
	protected Style				_viewerStyle;

	/**
	 * Alignment of the label of the node.
	 * This attribute determines the positioning of the label of the node in comparison
	 * with the image representing the node.
	 * It is used only if the label of the node is specified. The following alignments 
	 * are managed:
	 * LEFT: Label on the left of the image
	 * RIGHT: Label on the right of the image
	 * TOP: Label on the top of the image
	 * BOTTOM: Label on the bottom of the image
	 * CENTER: Label in the middle of the image
	 * EXTERN: Label outside the image in comparison with the center of the map
	 */
	protected LabelAlign			_labelAlign;

	/**
	 * Abscissa of the node.
	 */
	protected float				_x;

	/**
	 * Ordinate of the node.
	 */
	protected float				_y;

	/**
	 * Altitude of the node (used only for a map in three dimensions)
	 */
	protected float				_z;

	/**
	 * Indicates whether the node is sensitive or not.
	 * A sensitive node can be selected and allows launching actions (by 
	 * double-clicking for example).
	 */
	protected boolean			_sensitive;

	/**
	 * Indicates whether the node can be moved or not.
	 */
	protected boolean			_movable;

	/**
	 * Indicates whether the node is selected or not.
	 * The value of this attribute is used only if the node is sensitive.
	 */
	protected boolean			_selected;

	/**
	 * Indicates whether the node can be resized or not when zooming the map.
	 */
	protected boolean			_resizable;

	/**
	 * Indicates if the node is hidden (not displayed) or not.
	 */
	protected boolean			_hidden;

	/**
	 * Image used for the node.
	 * The image is used only if the style of the node is IMAGE.
	 */
	protected String			_image;

	/**
	 * Image used for the node in the viewer if the viewer style is IMAGE.
	 * The image is used only if the style of the node is IMAGE.
	 */
	protected String			_viewerImage;

	/**
	 * Image used as superimposed on the image of the node.
	 * It is displayed in the bottom right-hand corner of the image representing the 
	 * node.
	 */
	protected String			_smallImage;

	/**
	 * Size of the node.
	 * The expected value is a positive integer value defining the size of the node in 
	 * pixels (width and height).
	 * This value is used only if the style of the node is CIRCLE, SQUARE or TRIANGLE.
	 */
	protected int				_size;

	/**
	 * MAP_NODE.height
	 */
	protected int				_height;

	/**
	 * Layer of the node.
	 */
	protected int				_layer;

	/**
	 * Color of the internal frame of the node.
	 * If this attribute is specified, a frame of the indicated color is drawn around 
	 * the node.
	 */
	protected String			_colorFrameInt;

	/**
	 * Color of external frame of the node.
	 * If this attribute is specified, a frame of the indicated color is drawn around 
	 * the node, and around the internal frame if there is one.
	 */
	protected String			_colorFrameExt;

	/**
	 * Color of the cross drawn on the node.
	 * This cross is drawn only if this attribute has been specified.
	 */
	protected String			_colorCross;

	/**
	 * Positioning data.
	 * This data is used by certain positioning managers to set the node in a special 
	 * way. The possible values are the following: north, south, east, west, northeast,
	 * northwest, southeast and southwest: Values interpreted by the manager of LINEAR 
	 * type.
	 * It indicates the location of the node in question in comparison with its parent 
	 * node (in the case where a parent node has actually been specified).
	 */
	protected String			_layoutInfo;

	/**
	 * Indicates if the image is blinking or not.
	 */
	protected boolean			_blink;

	/**
	 * Indicates if the small image drawn over the image of the node is blinking or 
	 * not.
	 */
	protected boolean			_smallImageBlink;

	/**
	 * Java full name of the class that manages the map node.
	 */
	protected String			_javaName;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Color of the node.
	 * If the style of the node is CIRCLE, SQUARE or TRIANGLE the color is used to draw
	 * the node.
	 * If the style of the node is IMAGE the color is used to replace the black color 
	 * contained in the image of the node.
	 * This last possibility allows among others to use a same image for a node that 
	 * switches from a state to another (for a major warning state, the black color of 
	 * the image can be replace by the red color, for a minor state by the blue color, 
	 * etc.).
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * MAP_NODE.selectedColor
	 */
	protected String			_selectedColor;

	/**
	 * MAP_NODE.inactiveSelectedColor
	 */
	protected String			_inactiveSelectedColor;

	/**
	 * MAP_NODE.outlineColor
	 */
	protected String			_outlineColor;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_leftLabel = null;
		_rightLabel = null;
		_topLabel = null;
		_bottomLabel = null;
		_tooltip = null;
		_overview = null;
		_parentNode = null;
		_style = Style.CIRCLE;
		_viewerStyle = Style.CIRCLE;
		_labelAlign = LabelAlign.BOTTOM;
		_x = getFloatValue("MAP_NODE.x", 0F);
		_y = getFloatValue("MAP_NODE.y", 0F);
		_z = getFloatValue("MAP_NODE.z", 0F);
		_sensitive = getBooleanValue("MAP_NODE.sensitive", true);
		_movable = getBooleanValue("MAP_NODE.movable", true);
		_selected = getBooleanValue("MAP_NODE.selected", false);
		_resizable = getBooleanValue("MAP_NODE.resizable", false);
		_hidden = getBooleanValue("MAP_NODE.hidden", false);
		_image = null;
		_viewerImage = null;
		_smallImage = null;
		_size = getIntValue("MAP_NODE.size", -1);
		_height = getIntValue("MAP_NODE.height", -1);
		_layer = getIntValue("MAP_NODE.layer", -1);
		_colorFrameInt = null;
		_colorFrameExt = null;
		_colorCross = null;
		_layoutInfo = null;
		_blink = getBooleanValue("MAP_NODE.blink", false);
		_smallImageBlink = getBooleanValue("MAP_NODE.smallImageBlink", false);
		_javaName = null;
		_font = getStringValue("MAP_NODE.font", "$ALL.font");
		_color = getStringValue("MAP_NODE.color", "$ALL.color");
		_fontColor = getStringValue("MAP_NODE.fontColor", "$ALL.fontColor");
		_selectedColor = getStringValue("MAP_NODE.selectedColor", "$ALL.selectedColor");
		_inactiveSelectedColor = getStringValue("MAP_NODE.inactiveSelectedColor", "$ALL.inactiveSelectedColor");
		_outlineColor = getStringValue("MAP_NODE.outlineColor", "$ALL.outlineColor");
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyMapNode	model = (LyMapNode)component;

		_model = component;
		_label = model._label;
		_leftLabel = model._leftLabel;
		_rightLabel = model._rightLabel;
		_topLabel = model._topLabel;
		_bottomLabel = model._bottomLabel;
		_tooltip = model._tooltip;
		_overview = model._overview;
		_parentNode = model._parentNode;
		_style = model._style;
		_viewerStyle = model._viewerStyle;
		_labelAlign = model._labelAlign;
		_x = model._x;
		_y = model._y;
		_z = model._z;
		_sensitive = model._sensitive;
		_movable = model._movable;
		_selected = model._selected;
		_resizable = model._resizable;
		_hidden = model._hidden;
		_image = model._image;
		_viewerImage = model._viewerImage;
		_smallImage = model._smallImage;
		_size = model._size;
		_height = model._height;
		_layer = model._layer;
		_colorFrameInt = model._colorFrameInt;
		_colorFrameExt = model._colorFrameExt;
		_colorCross = model._colorCross;
		_layoutInfo = model._layoutInfo;
		_blink = model._blink;
		_smallImageBlink = model._smallImageBlink;
		_javaName = model._javaName;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;
		_selectedColor = model._selectedColor;
		_inactiveSelectedColor = model._inactiveSelectedColor;
		_outlineColor = model._outlineColor;
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyMapNode()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyMapNode(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyMapNode)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("leftLabel"))
			_leftLabel = node.getString("leftLabel");
		if (node.isSpecified("rightLabel"))
			_rightLabel = node.getString("rightLabel");
		if (node.isSpecified("topLabel"))
			_topLabel = node.getString("topLabel");
		if (node.isSpecified("bottomLabel"))
			_bottomLabel = node.getString("bottomLabel");
		if (node.isSpecified("tooltip"))
			_tooltip = node.getString("tooltip");
		if (node.isSpecified("overview"))
			_overview = node.getString("overview");
		if (node.isSpecified("parentNode"))
			_parentNode = node.getString("parentNode");
		if (node.isSpecified("style"))
			_style = getStyle(node.getString("style"));
		if (node.isSpecified("viewerStyle"))
			_viewerStyle = getStyle(node.getString("viewerStyle"));
		if (node.isSpecified("labelAlign"))
			_labelAlign = getLabelAlign(node.getString("labelAlign"));
		if (node.isSpecified("x"))
			_x = node.getFloat("x");
		if (node.isSpecified("y"))
			_y = node.getFloat("y");
		if (node.isSpecified("z"))
			_z = node.getFloat("z");
		if (node.isSpecified("sensitive"))
			_sensitive = node.getBoolean("sensitive");
		if (node.isSpecified("movable"))
			_movable = node.getBoolean("movable");
		if (node.isSpecified("selected"))
			_selected = node.getBoolean("selected");
		if (node.isSpecified("resizable"))
			_resizable = node.getBoolean("resizable");
		if (node.isSpecified("hidden"))
			_hidden = node.getBoolean("hidden");
		if (node.isSpecified("image"))
			_image = node.getString("image");
		if (node.isSpecified("viewerImage"))
			_viewerImage = node.getString("viewerImage");
		if (node.isSpecified("smallImage"))
			_smallImage = node.getString("smallImage");
		if (node.isSpecified("size"))
			_size = node.getInt("size");
		if (node.isSpecified("height"))
			_height = node.getInt("height");
		if (node.isSpecified("layer"))
			_layer = node.getInt("layer");
		if (node.isSpecified("colorFrameInt"))
			_colorFrameInt = node.getString("colorFrameInt");
		if (node.isSpecified("colorFrameExt"))
			_colorFrameExt = node.getString("colorFrameExt");
		if (node.isSpecified("colorCross"))
			_colorCross = node.getString("colorCross");
		if (node.isSpecified("layoutInfo"))
			_layoutInfo = node.getString("layoutInfo");
		if (node.isSpecified("blink"))
			_blink = node.getBoolean("blink");
		if (node.isSpecified("smallImageBlink"))
			_smallImageBlink = node.getBoolean("smallImageBlink");
		if (node.isSpecified("javaName"))
			_javaName = node.getString("javaName");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("selectedColor"))
			_selectedColor = node.getString("selectedColor");
		if (node.isSpecified("inactiveSelectedColor"))
			_inactiveSelectedColor = node.getString("inactiveSelectedColor");
		if (node.isSpecified("outlineColor"))
			_outlineColor = node.getString("outlineColor");
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyMapNode clone = new LyMapNode();

		clone.initFromModel(this);
		clone._template = this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyMapNode getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of leftLabel.
	 * @return value of leftLabel.
	 */
	public String getLeftLabel()
	{
		return _leftLabel;
	}

	/**
	 * Gets value of rightLabel.
	 * @return value of rightLabel.
	 */
	public String getRightLabel()
	{
		return _rightLabel;
	}

	/**
	 * Gets value of topLabel.
	 * @return value of topLabel.
	 */
	public String getTopLabel()
	{
		return _topLabel;
	}

	/**
	 * Gets value of bottomLabel.
	 * @return value of bottomLabel.
	 */
	public String getBottomLabel()
	{
		return _bottomLabel;
	}

	/**
	 * Gets value of tooltip.
	 * @return value of tooltip.
	 */
	public String getTooltip()
	{
		return _tooltip;
	}

	/**
	 * Gets value of overview.
	 * @return value of overview.
	 */
	public String getOverview()
	{
		return _overview;
	}

	/**
	 * Gets value of parentNode.
	 * @return value of parentNode.
	 */
	public String getParentNode()
	{
		return _parentNode;
	}

	/**
	 * Gets value of style.
	 * @return value of style.
	 */
	public Style getStyle()
	{
		return _style;
	}

	/**
	 * Gets value of viewerStyle.
	 * @return value of viewerStyle.
	 */
	public Style getViewerStyle()
	{
		return _viewerStyle;
	}

	/**
	 * Gets value of labelAlign.
	 * @return value of labelAlign.
	 */
	public LabelAlign getLabelAlign()
	{
		return _labelAlign;
	}

	/**
	 * Gets value of x.
	 * @return value of x.
	 */
	public float getX()
	{
		return _x;
	}

	/**
	 * Gets value of y.
	 * @return value of y.
	 */
	public float getY()
	{
		return _y;
	}

	/**
	 * Gets value of z.
	 * @return value of z.
	 */
	public float getZ()
	{
		return _z;
	}

	/**
	 * Checks if value of attribute sensitive is set.
	 * @return true if sensitive is set, false otherwise.
	 */
	public boolean isSensitive()
	{
		return _sensitive;
	}

	/**
	 * Checks if value of attribute movable is set.
	 * @return true if movable is set, false otherwise.
	 */
	public boolean isMovable()
	{
		return _movable;
	}

	/**
	 * Checks if value of attribute selected is set.
	 * @return true if selected is set, false otherwise.
	 */
	public boolean isSelected()
	{
		return _selected;
	}

	/**
	 * Checks if value of attribute resizable is set.
	 * @return true if resizable is set, false otherwise.
	 */
	public boolean isResizable()
	{
		return _resizable;
	}

	/**
	 * Checks if value of attribute hidden is set.
	 * @return true if hidden is set, false otherwise.
	 */
	public boolean isHidden()
	{
		return _hidden;
	}

	/**
	 * Gets value of image.
	 * @return value of image.
	 */
	public String getImage()
	{
		return _image;
	}

	/**
	 * Gets value of viewerImage.
	 * @return value of viewerImage.
	 */
	public String getViewerImage()
	{
		return _viewerImage;
	}

	/**
	 * Gets value of smallImage.
	 * @return value of smallImage.
	 */
	public String getSmallImage()
	{
		return _smallImage;
	}

	/**
	 * Gets value of size.
	 * @return value of size.
	 */
	public int getSize()
	{
		return _size;
	}

	/**
	 * Gets value of height.
	 * @return value of height.
	 */
	public int getHeight()
	{
		return _height;
	}

	/**
	 * Gets value of layer.
	 * @return value of layer.
	 */
	public int getLayer()
	{
		return _layer;
	}

	/**
	 * Gets value of colorFrameInt.
	 * @return value of colorFrameInt.
	 */
	public String getColorFrameInt()
	{
		return _colorFrameInt;
	}

	/**
	 * Gets value of colorFrameExt.
	 * @return value of colorFrameExt.
	 */
	public String getColorFrameExt()
	{
		return _colorFrameExt;
	}

	/**
	 * Gets value of colorCross.
	 * @return value of colorCross.
	 */
	public String getColorCross()
	{
		return _colorCross;
	}

	/**
	 * Gets value of layoutInfo.
	 * @return value of layoutInfo.
	 */
	public String getLayoutInfo()
	{
		return _layoutInfo;
	}

	/**
	 * Checks if value of attribute blink is set.
	 * @return true if blink is set, false otherwise.
	 */
	public boolean isBlink()
	{
		return _blink;
	}

	/**
	 * Checks if value of attribute smallImageBlink is set.
	 * @return true if smallImageBlink is set, false otherwise.
	 */
	public boolean isSmallImageBlink()
	{
		return _smallImageBlink;
	}

	/**
	 * Gets value of javaName.
	 * @return value of javaName.
	 */
	public String getJavaName()
	{
		return _javaName;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("MAP_NODE.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("MAP_NODE.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("MAP_NODE.fontColor"));
	}

	/**
	 * Gets value of selectedColor.
	 * @return value of selectedColor.
	 */
	public String getSelectedColor()
	{
		return _selectedColor;
	}

	/**
	 * Checks whether the selectedColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultSelectedColor()
	{
		return (_selectedColor == getStringValue("MAP_NODE.selectedColor"));
	}

	/**
	 * Gets value of inactiveSelectedColor.
	 * @return value of inactiveSelectedColor.
	 */
	public String getInactiveSelectedColor()
	{
		return _inactiveSelectedColor;
	}

	/**
	 * Checks whether the inactiveSelectedColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultInactiveSelectedColor()
	{
		return (_inactiveSelectedColor == getStringValue("MAP_NODE.inactiveSelectedColor"));
	}

	/**
	 * Gets value of outlineColor.
	 * @return value of outlineColor.
	 */
	public String getOutlineColor()
	{
		return _outlineColor;
	}

	/**
	 * Checks whether the outlineColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultOutlineColor()
	{
		return (_outlineColor == getStringValue("MAP_NODE.outlineColor"));
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyMapNode template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView != null)
			((LyGraphicalMapNode)_internalView).setLabel(label);
	}

	/**
	 * Sets attribute leftLabel to the given value.
	 * @param leftLabel New value of leftLabel.
	 */
	public void setLeftLabel(String leftLabel)
	{
		if ((leftLabel == _leftLabel)
			|| ((leftLabel != null)
				&& (_leftLabel != null)
				&& leftLabel.equals(_leftLabel)))
			return;

		_leftLabel = leftLabel;
		if (_internalView != null)
			((LyGraphicalMapNode)_internalView).setLeftLabel(leftLabel);
	}

	/**
	 * Sets attribute rightLabel to the given value.
	 * @param rightLabel New value of rightLabel.
	 */
	public void setRightLabel(String rightLabel)
	{
		if ((rightLabel == _rightLabel)
			|| ((rightLabel != null)
				&& (_rightLabel != null)
				&& rightLabel.equals(_rightLabel)))
			return;

		_rightLabel = rightLabel;
		if (_internalView != null)
			((LyGraphicalMapNode)_internalView).setRightLabel(rightLabel);
	}

	/**
	 * Sets attribute topLabel to the given value.
	 * @param topLabel New value of topLabel.
	 */
	public void setTopLabel(String topLabel)
	{
		if ((topLabel == _topLabel)
			|| ((topLabel != null)
				&& (_topLabel != null)
				&& topLabel.equals(_topLabel)))
			return;

		_topLabel = topLabel;
		if (_internalView != null)
			((LyGraphicalMapNode)_internalView).setTopLabel(topLabel);
	}

	/**
	 * Sets attribute bottomLabel to the given value.
	 * @param bottomLabel New value of bottomLabel.
	 */
	public void setBottomLabel(String bottomLabel)
	{
		if ((bottomLabel == _bottomLabel)
			|| ((bottomLabel != null)
				&& (_bottomLabel != null)
				&& bottomLabel.equals(_bottomLabel)))
			return;

		_bottomLabel = bottomLabel;
		if (_internalView != null)
			((LyGraphicalMapNode)_internalView).setBottomLabel(bottomLabel);
	}

	/**
	 * Initializes attribute tooltip to the given value.
	 * @param tooltip New value of tooltip.
	 */
	public void setTooltip(String tooltip)
	{
		if ((tooltip == _tooltip)
			|| ((tooltip != null)
				&& (_tooltip != null)
				&& tooltip.equals(_tooltip)))
			return;

		_tooltip = tooltip;
	}

	/**
	 * Initializes attribute overview to the given value.
	 * @param overview New value of overview.
	 */
	public void setOverview(String overview)
	{
		if ((overview == _overview)
			|| ((overview != null)
				&& (_overview != null)
				&& overview.equals(_overview)))
			return;

		_overview = overview;
	}

	/**
	 * Initializes attribute parentNode to the given value.
	 * @param parentNode New value of parentNode.
	 */
	public void setParentNode(String parentNode)
	{
		if ((parentNode == _parentNode)
			|| ((parentNode != null)
				&& (_parentNode != null)
				&& parentNode.equals(_parentNode)))
			return;

		_parentNode = parentNode;
	}

	/**
	 * Initializes attribute style to the given value.
	 * @param style New value of style.
	 * @exception IllegalArgumentException if style parameter is null.
	 */
	public void setStyle(Style style)
	{
		if (style == _style)
			return;
		if (style == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute style");

		_style = style;
	}

	/**
	 * Initializes attribute viewerStyle to the given value.
	 * @param viewerStyle New value of viewerStyle.
	 * @exception IllegalArgumentException if viewerStyle parameter is null.
	 */
	public void setViewerStyle(Style viewerStyle)
	{
		if (viewerStyle == _viewerStyle)
			return;
		if (viewerStyle == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute viewerStyle");

		_viewerStyle = viewerStyle;
	}

	/**
	 * Initializes attribute labelAlign to the given value.
	 * @param labelAlign New value of labelAlign.
	 * @exception IllegalArgumentException if labelAlign parameter is null.
	 */
	public void setLabelAlign(LabelAlign labelAlign)
	{
		if (labelAlign == _labelAlign)
			return;
		if (labelAlign == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute labelAlign");

		_labelAlign = labelAlign;
	}

	/**
	 * Sets attribute x to the given value.
	 * @param x New value of x.
	 */
	public void setX(float x)
	{
		if (x == _x)
			return;
		_x = x;
		if (_internalView != null)
			((LyGraphicalMapNode)_internalView).setX(x);
	}

	/**
	 * Sets attribute y to the given value.
	 * @param y New value of y.
	 */
	public void setY(float y)
	{
		if (y == _y)
			return;
		_y = y;
		if (_internalView != null)
			((LyGraphicalMapNode)_internalView).setY(y);
	}

	/**
	 * Sets attribute z to the given value.
	 * @param z New value of z.
	 */
	public void setZ(float z)
	{
		if (z == _z)
			return;
		_z = z;
		if (_internalView != null)
			((LyGraphicalMapNode)_internalView).setZ(z);
	}

	/**
	 * Initializes attribute sensitive to the given value.
	 * @param sensitive Boolean indicating if attribute sensitive is set or unset.
	 */
	public void setSensitive(boolean sensitive)
	{
		if (sensitive == _sensitive)
			return;
		_sensitive = sensitive;
	}

	/**
	 * Initializes attribute movable to the given value.
	 * @param movable Boolean indicating if attribute movable is set or unset.
	 */
	public void setMovable(boolean movable)
	{
		if (movable == _movable)
			return;
		_movable = movable;
	}

	/**
	 * Sets attribute selected to the given value.
	 * @param selected Boolean indicating if attribute selected is set or unset.
	 */
	public void setSelected(boolean selected)
	{
		if (selected == _selected)
			return;
		_selected = selected;
		if (_internalView != null)
			((LyGraphicalMapNode)_internalView).setSelected(selected);
	}

	/**
	 * Initializes attribute resizable to the given value.
	 * @param resizable Boolean indicating if attribute resizable is set or unset.
	 */
	public void setResizable(boolean resizable)
	{
		if (resizable == _resizable)
			return;
		_resizable = resizable;
	}

	/**
	 * Sets attribute hidden to the given value.
	 * @param hidden Boolean indicating if attribute hidden is set or unset.
	 */
	public void setHidden(boolean hidden)
	{
		if (hidden == _hidden)
			return;
		_hidden = hidden;
		if (_internalView != null)
			((LyGraphicalMapNode)_internalView).setHidden(hidden);
	}

	/**
	 * Sets attribute image to the given value.
	 * @param image New value of image.
	 */
	public void setImage(String image)
	{
		if ((image == _image)
			|| ((image != null)
				&& (_image != null)
				&& image.equals(_image)))
			return;

		_image = image;
		if (_internalView != null)
			((LyGraphicalMapNode)_internalView).setImage(image);
	}

	/**
	 * Sets attribute viewerImage to the given value.
	 * @param viewerImage New value of viewerImage.
	 */
	public void setViewerImage(String viewerImage)
	{
		if ((viewerImage == _viewerImage)
			|| ((viewerImage != null)
				&& (_viewerImage != null)
				&& viewerImage.equals(_viewerImage)))
			return;

		_viewerImage = viewerImage;
		if (_internalView != null)
			((LyGraphicalMapNode)_internalView).setViewerImage(viewerImage);
	}

	/**
	 * Initializes attribute smallImage to the given value.
	 * @param smallImage New value of smallImage.
	 */
	public void setSmallImage(String smallImage)
	{
		if ((smallImage == _smallImage)
			|| ((smallImage != null)
				&& (_smallImage != null)
				&& smallImage.equals(_smallImage)))
			return;

		_smallImage = smallImage;
	}

	/**
	 * Initializes attribute size to the given value.
	 * @param size New value of size.
	 */
	public void setSize(int size)
	{
		if (size == _size)
			return;
		_size = size;
	}

	/**
	 * Initializes attribute height to the given value.
	 * @param height New value of height.
	 */
	public void setHeight(int height)
	{
		if (height == _height)
			return;
		_height = height;
	}

	/**
	 * Sets attribute layer to the given value.
	 * @param layer New value of layer.
	 */
	public void setLayer(int layer)
	{
		if (layer == _layer)
			return;
		_layer = layer;
		if (_internalView != null)
			((LyGraphicalMapNode)_internalView).setLayer(layer);
	}

	/**
	 * Sets attribute colorFrameInt to the given value.
	 * @param colorFrameInt New value of colorFrameInt.
	 */
	public void setColorFrameInt(String colorFrameInt)
	{
		if ((colorFrameInt == _colorFrameInt)
			|| ((colorFrameInt != null)
				&& (_colorFrameInt != null)
				&& colorFrameInt.equals(_colorFrameInt)))
			return;

		_colorFrameInt = colorFrameInt;
		if (_internalView != null)
			((LyGraphicalMapNode)_internalView).setColorFrameInt(colorFrameInt);
	}

	/**
	 * Sets attribute colorFrameExt to the given value.
	 * @param colorFrameExt New value of colorFrameExt.
	 */
	public void setColorFrameExt(String colorFrameExt)
	{
		if ((colorFrameExt == _colorFrameExt)
			|| ((colorFrameExt != null)
				&& (_colorFrameExt != null)
				&& colorFrameExt.equals(_colorFrameExt)))
			return;

		_colorFrameExt = colorFrameExt;
		if (_internalView != null)
			((LyGraphicalMapNode)_internalView).setColorFrameExt(colorFrameExt);
	}

	/**
	 * Initializes attribute colorCross to the given value.
	 * @param colorCross New value of colorCross.
	 */
	public void setColorCross(String colorCross)
	{
		if ((colorCross == _colorCross)
			|| ((colorCross != null)
				&& (_colorCross != null)
				&& colorCross.equals(_colorCross)))
			return;

		_colorCross = colorCross;
	}

	/**
	 * Initializes attribute layoutInfo to the given value.
	 * @param layoutInfo New value of layoutInfo.
	 */
	public void setLayoutInfo(String layoutInfo)
	{
		if ((layoutInfo == _layoutInfo)
			|| ((layoutInfo != null)
				&& (_layoutInfo != null)
				&& layoutInfo.equals(_layoutInfo)))
			return;

		_layoutInfo = layoutInfo;
	}

	/**
	 * Initializes attribute blink to the given value.
	 * @param blink Boolean indicating if attribute blink is set or unset.
	 */
	public void setBlink(boolean blink)
	{
		if (blink == _blink)
			return;
		_blink = blink;
	}

	/**
	 * Initializes attribute smallImageBlink to the given value.
	 * @param smallImageBlink Boolean indicating if attribute smallImageBlink is set or unset.
	 */
	public void setSmallImageBlink(boolean smallImageBlink)
	{
		if (smallImageBlink == _smallImageBlink)
			return;
		_smallImageBlink = smallImageBlink;
	}

	/**
	 * Initializes attribute javaName to the given value.
	 * @param javaName New value of javaName.
	 */
	public void setJavaName(String javaName)
	{
		if ((javaName == _javaName)
			|| ((javaName != null)
				&& (_javaName != null)
				&& javaName.equals(_javaName)))
			return;

		_javaName = javaName;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Sets attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
		if (_internalView != null)
			((LyGraphicalMapNode)_internalView).setFontColor(fontColor);
	}

	/**
	 * Initializes attribute selectedColor to the given value.
	 * @param selectedColor New value of selectedColor.
	 */
	public void setSelectedColor(String selectedColor)
	{
		if ((selectedColor == _selectedColor)
			|| ((selectedColor != null)
				&& (_selectedColor != null)
				&& selectedColor.equals(_selectedColor)))
			return;

		_selectedColor = selectedColor;
	}

	/**
	 * Initializes attribute inactiveSelectedColor to the given value.
	 * @param inactiveSelectedColor New value of inactiveSelectedColor.
	 */
	public void setInactiveSelectedColor(String inactiveSelectedColor)
	{
		if ((inactiveSelectedColor == _inactiveSelectedColor)
			|| ((inactiveSelectedColor != null)
				&& (_inactiveSelectedColor != null)
				&& inactiveSelectedColor.equals(_inactiveSelectedColor)))
			return;

		_inactiveSelectedColor = inactiveSelectedColor;
	}

	/**
	 * Initializes attribute outlineColor to the given value.
	 * @param outlineColor New value of outlineColor.
	 */
	public void setOutlineColor(String outlineColor)
	{
		if ((outlineColor == _outlineColor)
			|| ((outlineColor != null)
				&& (_outlineColor != null)
				&& outlineColor.equals(_outlineColor)))
			return;

		_outlineColor = outlineColor;
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<MAP_NODE");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (((_template == null) && (_leftLabel != null))
				|| ((_template != null) && (_leftLabel != _template._leftLabel)))
			pw.print(" leftLabel='" + LyString.getXmlString(_leftLabel) + "'");
		if (((_template == null) && (_rightLabel != null))
				|| ((_template != null) && (_rightLabel != _template._rightLabel)))
			pw.print(" rightLabel='" + LyString.getXmlString(_rightLabel) + "'");
		if (((_template == null) && (_topLabel != null))
				|| ((_template != null) && (_topLabel != _template._topLabel)))
			pw.print(" topLabel='" + LyString.getXmlString(_topLabel) + "'");
		if (((_template == null) && (_bottomLabel != null))
				|| ((_template != null) && (_bottomLabel != _template._bottomLabel)))
			pw.print(" bottomLabel='" + LyString.getXmlString(_bottomLabel) + "'");
		if (((_template == null) && (_tooltip != null))
				|| ((_template != null) && (_tooltip != _template._tooltip)))
			pw.print(" tooltip='" + LyString.getXmlString(_tooltip) + "'");
		if (((_template == null) && (_overview != null))
				|| ((_template != null) && (_overview != _template._overview)))
			pw.print(" overview='" + LyString.getXmlString(_overview) + "'");
		if (((_template == null) && (_parentNode != null))
				|| ((_template != null) && (_parentNode != _template._parentNode)))
			pw.print(" parentNode='" + LyString.getXmlString(_parentNode) + "'");
		if ((_template == null)
			|| (_style != _template._style))
		{
			if (_style == Style.IMAGE)
				pw.print(" style='IMAGE'");
			else if ((_style == Style.CIRCLE) && (_template != null))
				pw.print(" style='CIRCLE'");
			else if (_style == Style.SQUARE)
				pw.print(" style='SQUARE'");
			else if (_style == Style.TRIANGLE)
				pw.print(" style='TRIANGLE'");
			else if (_style == Style.ELLIPSE)
				pw.print(" style='ELLIPSE'");
		}
		if ((_template == null)
			|| (_viewerStyle != _template._viewerStyle))
		{
			if (_viewerStyle == Style.IMAGE)
				pw.print(" viewerStyle='IMAGE'");
			else if ((_viewerStyle == Style.CIRCLE) && (_template != null))
				pw.print(" viewerStyle='CIRCLE'");
			else if (_viewerStyle == Style.SQUARE)
				pw.print(" viewerStyle='SQUARE'");
			else if (_viewerStyle == Style.TRIANGLE)
				pw.print(" viewerStyle='TRIANGLE'");
			else if (_viewerStyle == Style.ELLIPSE)
				pw.print(" viewerStyle='ELLIPSE'");
		}
		if ((_template == null)
			|| (_labelAlign != _template._labelAlign))
		{
			if (_labelAlign == LabelAlign.LEFT)
				pw.print(" labelAlign='LEFT'");
			else if (_labelAlign == LabelAlign.RIGHT)
				pw.print(" labelAlign='RIGHT'");
			else if (_labelAlign == LabelAlign.TOP)
				pw.print(" labelAlign='TOP'");
			else if ((_labelAlign == LabelAlign.BOTTOM) && (_template != null))
				pw.print(" labelAlign='BOTTOM'");
			else if (_labelAlign == LabelAlign.CENTER)
				pw.print(" labelAlign='CENTER'");
			else if (_labelAlign == LabelAlign.EXTERN)
				pw.print(" labelAlign='EXTERN'");
		}
		if (((_template == null) && (_x != getFloatValue("MAP_NODE.x", 0F)))
				|| ((_template != null) && (_x != _template._x)))
			pw.print(" x='" + _x + "'");
		if (((_template == null) && (_y != getFloatValue("MAP_NODE.y", 0F)))
				|| ((_template != null) && (_y != _template._y)))
			pw.print(" y='" + _y + "'");
		if (((_template == null) && (_z != getFloatValue("MAP_NODE.z", 0F)))
				|| ((_template != null) && (_z != _template._z)))
			pw.print(" z='" + _z + "'");
		if (((_template == null) && !_sensitive)
				|| ((_template != null) && (_sensitive != _template._sensitive)))
			pw.print(" sensitive='" + _sensitive + "'");
		if (((_template == null) && !_movable)
				|| ((_template != null) && (_movable != _template._movable)))
			pw.print(" movable='" + _movable + "'");
		if (((_template == null) && _selected)
				|| ((_template != null) && (_selected != _template._selected)))
			pw.print(" selected='" + _selected + "'");
		if (((_template == null) && _resizable)
				|| ((_template != null) && (_resizable != _template._resizable)))
			pw.print(" resizable='" + _resizable + "'");
		if (((_template == null) && _hidden)
				|| ((_template != null) && (_hidden != _template._hidden)))
			pw.print(" hidden='" + _hidden + "'");
		if (((_template == null) && (_image != null))
				|| ((_template != null) && (_image != _template._image)))
			pw.print(" image='" + LyString.getXmlString(_image) + "'");
		if (((_template == null) && (_viewerImage != null))
				|| ((_template != null) && (_viewerImage != _template._viewerImage)))
			pw.print(" viewerImage='" + LyString.getXmlString(_viewerImage) + "'");
		if (((_template == null) && (_smallImage != null))
				|| ((_template != null) && (_smallImage != _template._smallImage)))
			pw.print(" smallImage='" + LyString.getXmlString(_smallImage) + "'");
		if (((_template == null) && (_size != getIntValue("MAP_NODE.size", -1)))
				|| ((_template != null) && (_size != _template._size)))
			pw.print(" size='" + _size + "'");
		if (((_template == null) && (_height != getIntValue("MAP_NODE.height", -1)))
				|| ((_template != null) && (_height != _template._height)))
			pw.print(" height='" + _height + "'");
		if (((_template == null) && (_layer != getIntValue("MAP_NODE.layer", -1)))
				|| ((_template != null) && (_layer != _template._layer)))
			pw.print(" layer='" + _layer + "'");
		if (((_template == null) && (_colorFrameInt != null))
				|| ((_template != null) && (_colorFrameInt != _template._colorFrameInt)))
			pw.print(" colorFrameInt='" + LyString.getXmlString(_colorFrameInt) + "'");
		if (((_template == null) && (_colorFrameExt != null))
				|| ((_template != null) && (_colorFrameExt != _template._colorFrameExt)))
			pw.print(" colorFrameExt='" + LyString.getXmlString(_colorFrameExt) + "'");
		if (((_template == null) && (_colorCross != null))
				|| ((_template != null) && (_colorCross != _template._colorCross)))
			pw.print(" colorCross='" + LyString.getXmlString(_colorCross) + "'");
		if (((_template == null) && (_layoutInfo != null))
				|| ((_template != null) && (_layoutInfo != _template._layoutInfo)))
			pw.print(" layoutInfo='" + LyString.getXmlString(_layoutInfo) + "'");
		if (((_template == null) && _blink)
				|| ((_template != null) && (_blink != _template._blink)))
			pw.print(" blink='" + _blink + "'");
		if (((_template == null) && _smallImageBlink)
				|| ((_template != null) && (_smallImageBlink != _template._smallImageBlink)))
			pw.print(" smallImageBlink='" + _smallImageBlink + "'");
		if (((_template == null) && (_javaName != null))
				|| ((_template != null) && (_javaName != _template._javaName)))
			pw.print(" javaName='" + LyString.getXmlString(_javaName) + "'");
		if (((_template == null) && (_font != getStringValue("MAP_NODE.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("MAP_NODE.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("MAP_NODE.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && (_selectedColor != getStringValue("MAP_NODE.selectedColor", "$ALL.selectedColor")))
				|| ((_template != null) && (_selectedColor != _template._selectedColor)))
			pw.print(" selectedColor='" + LyString.getXmlString(_selectedColor) + "'");
		if (((_template == null) && (_inactiveSelectedColor != getStringValue("MAP_NODE.inactiveSelectedColor", "$ALL.inactiveSelectedColor")))
				|| ((_template != null) && (_inactiveSelectedColor != _template._inactiveSelectedColor)))
			pw.print(" inactiveSelectedColor='" + LyString.getXmlString(_inactiveSelectedColor) + "'");
		if (((_template == null) && (_outlineColor != getStringValue("MAP_NODE.outlineColor", "$ALL.outlineColor")))
				|| ((_template != null) && (_outlineColor != _template._outlineColor)))
			pw.print(" outlineColor='" + LyString.getXmlString(_outlineColor) + "'");
		pw.println("/>");
	}
}
