/*
 * LyMapRegionBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyMapRegionBase class is a generated Java implementation that corresponds to the element 
 * MAP_REGION of the descriptive grammar of the views (views.dtd).
 * <B>&lt;MAP_REGION&gt; : Map region</B>
 * A region allows displaying on a map a graphical element made up of a finite set 
 * of points.
 * These points define polygons that are displayed on the map with a given color.
 * These elements follow the zooms made on the map (their size increases or 
 * decreases when zooming in or out).
 * @see leon.view.LyComponent
 * @see leon.view.LyLyMapRegion
 * See also tag <A href="../../conf/dtd/view/MAP_REGION.html">MAP_REGION</A> of grammar leon_views.dtd.
 */
public abstract class LyMapRegionBase extends LyComponent
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetSelectedInterface,
				LySetHiddenInterface
{
	public static final String			__VERSION = "$Id$";


	/* Attributes */

	/**
	 * Identifier of the region.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyMapRegion			_template;

	/**
	 * Additional data file for the region.
	 * The indicated file must contain the description of one or more polygons that 
	 * will be added to the region at the moment of its first display. The content of 
	 * the file can be of two types: An XML file containing the description of a 
	 * polygon A DXF file (Autocad data format)
	 */
	protected String			_file;

	/**
	 * Label of the region.
	 */
	protected String			_label;

	/**
	 * Tooltip of the region.
	 */
	protected String			_tooltip;

	/**
	 * Image used as superimposed on the representation of the region.
	 */
	protected String			_smallImage;

	/**
	 * Tooltip of the image used as superimposed on the representation of the region.
	 */
	protected String			_smallImageTooltip;

	/**
	 * Weight of the region. This value is used only for the managers of TREEMAP type.
	 */
	protected float				_weight;

	/**
	 * Overview label of the region.
	 */
	protected String			_overview;

	/**
	 * Indicates whether the region is sensitive or not, that is it can be selected.
	 */
	protected boolean			_sensitive;

	/**
	 * Indicates whether the region is selected or not.
	 */
	protected boolean			_selected;

	/**
	 * Indicates if the region is hidden (not displayed) or not.
	 */
	protected boolean			_hidden;

	/**
	 * Indicates whether the region is a map background region or not.
	 * A map background region is not sensitive, is displayed under the other regions 
	 * and is drawn with a shadow effect.
	 */
	protected boolean			_background;

	/**
	 * Layer of the region.
	 */
	protected int				_layer;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Border color used for drawing the region.
	 */
	protected String			_borderColor;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * MAP_REGION.tooltipFont
	 */
	protected String			_tooltipFont;

	/**
	 * MAP_REGION.tooltipColor
	 */
	protected String			_tooltipColor;

	/**
	 * MAP_REGION.selectedColor
	 */
	protected String			_selectedColor;

	/**
	 * MAP_REGION.inactiveSelectedColor
	 */
	protected String			_inactiveSelectedColor;

	/**
	 * MAP_REGION.outlineColor
	 */
	protected String			_outlineColor;


	/* Sub components */

	/**
	 * Polygons describing the region.
	 */
	protected ArrayList<LyPolygon>		_polygons;

	/**
	 * Content of the region.
	 * This content is not displayed on the map where the region is.
	 * However, when launching an action on the concerned region, the content of the 
	 * map is reloaded by taking as content the one of the activated region.
	 */
	protected LyMapContent		_mapContent;

	/**
	 * Sub-regions of the region.
	 */
	protected ArrayList<LyMapRegion>		_mapRegions;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_file = null;
		_label = null;
		_tooltip = null;
		_smallImage = getStringValue("MAP_REGION.smallImage", "$ALL.smallImage");
		_smallImageTooltip = null;
		_weight = getFloatValue("MAP_REGION.weight", 0F);
		_overview = null;
		_sensitive = getBooleanValue("MAP_REGION.sensitive", true);
		_selected = getBooleanValue("MAP_REGION.selected", false);
		_hidden = getBooleanValue("MAP_REGION.hidden", false);
		_background = getBooleanValue("MAP_REGION.background", false);
		_layer = getIntValue("MAP_REGION.layer", -1);
		_font = getStringValue("MAP_REGION.font", "$ALL.font");
		_color = getStringValue("MAP_REGION.color", "$ALL.color");
		_borderColor = getStringValue("MAP_REGION.borderColor", "$ALL.borderColor");
		_fontColor = getStringValue("MAP_REGION.fontColor", "$ALL.fontColor");
		_tooltipFont = getStringValue("MAP_REGION.tooltipFont", "$ALL.tooltipFont");
		_tooltipColor = getStringValue("MAP_REGION.tooltipColor", "$ALL.tooltipColor");
		_selectedColor = getStringValue("MAP_REGION.selectedColor", "$ALL.selectedColor");
		_inactiveSelectedColor = getStringValue("MAP_REGION.inactiveSelectedColor", "$ALL.inactiveSelectedColor");
		_outlineColor = getStringValue("MAP_REGION.outlineColor", "$ALL.outlineColor");

		// Initialize sub components
		_polygons = null;
		_mapContent = null;
		_mapRegions = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyMapRegion	model = (LyMapRegion)component;

		_model = component;
		_file = model._file;
		_label = model._label;
		_tooltip = model._tooltip;
		_smallImage = model._smallImage;
		_smallImageTooltip = model._smallImageTooltip;
		_weight = model._weight;
		_overview = model._overview;
		_sensitive = model._sensitive;
		_selected = model._selected;
		_hidden = model._hidden;
		_background = model._background;
		_layer = model._layer;
		_font = model._font;
		_color = model._color;
		_borderColor = model._borderColor;
		_fontColor = model._fontColor;
		_tooltipFont = model._tooltipFont;
		_tooltipColor = model._tooltipColor;
		_selectedColor = model._selectedColor;
		_inactiveSelectedColor = model._inactiveSelectedColor;
		_outlineColor = model._outlineColor;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyPolygon	polygons;
		int			i, n;

		n = (model._polygons == null ? 0 : model._polygons.size());
		for (i = 0; i < n; i++)
		{
			polygons = model._polygons.get(i);
			if (polygons != null)
				polygons = (LyPolygon)polygons.cloneComponent();
			addChild(polygons);
		}
		if (model._mapContent != null)
			addChild(model._mapContent.cloneComponent());

		LyMapRegion	mapRegions;

		n = (model._mapRegions == null ? 0 : model._mapRegions.size());
		for (i = 0; i < n; i++)
		{
			mapRegions = model._mapRegions.get(i);
			if (mapRegions != null)
				mapRegions = (LyMapRegion)mapRegions.cloneComponent();
			addChild(mapRegions);
		}
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyMapRegionBase()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyMapRegionBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyMapRegion)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("file"))
			_file = node.getString("file");
		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("tooltip"))
			_tooltip = node.getString("tooltip");
		if (node.isSpecified("smallImage"))
			_smallImage = node.getString("smallImage");
		if (node.isSpecified("smallImageTooltip"))
			_smallImageTooltip = node.getString("smallImageTooltip");
		if (node.isSpecified("weight"))
			_weight = node.getFloat("weight");
		if (node.isSpecified("overview"))
			_overview = node.getString("overview");
		if (node.isSpecified("sensitive"))
			_sensitive = node.getBoolean("sensitive");
		if (node.isSpecified("selected"))
			_selected = node.getBoolean("selected");
		if (node.isSpecified("hidden"))
			_hidden = node.getBoolean("hidden");
		if (node.isSpecified("background"))
			_background = node.getBoolean("background");
		if (node.isSpecified("layer"))
			_layer = node.getInt("layer");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("borderColor"))
			_borderColor = node.getString("borderColor");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("tooltipFont"))
			_tooltipFont = node.getString("tooltipFont");
		if (node.isSpecified("tooltipColor"))
			_tooltipColor = node.getString("tooltipColor");
		if (node.isSpecified("selectedColor"))
			_selectedColor = node.getString("selectedColor");
		if (node.isSpecified("inactiveSelectedColor"))
			_inactiveSelectedColor = node.getString("inactiveSelectedColor");
		if (node.isSpecified("outlineColor"))
			_outlineColor = node.getString("outlineColor");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyPolygon)
			addPolygon((LyPolygon)child);
		else if (child instanceof LyMapContent)
			setMapContent((LyMapContent)child);
		else if (child instanceof LyMapRegion)
			addMapRegion((LyMapRegion)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyMapRegion clone = new LyMapRegion();

		clone.initFromModel(this);
		clone._template = (LyMapRegion)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyMapRegion getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of file.
	 * @return value of file.
	 */
	public String getFile()
	{
		return _file;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of tooltip.
	 * @return value of tooltip.
	 */
	public String getTooltip()
	{
		return _tooltip;
	}

	/**
	 * Gets value of smallImage.
	 * @return value of smallImage.
	 */
	public String getSmallImage()
	{
		return _smallImage;
	}

	/**
	 * Checks whether the smallImage attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultSmallImage()
	{
		return (_smallImage == getStringValue("MAP_REGION.smallImage"));
	}

	/**
	 * Gets value of smallImageTooltip.
	 * @return value of smallImageTooltip.
	 */
	public String getSmallImageTooltip()
	{
		return _smallImageTooltip;
	}

	/**
	 * Gets value of weight.
	 * @return value of weight.
	 */
	public float getWeight()
	{
		return _weight;
	}

	/**
	 * Gets value of overview.
	 * @return value of overview.
	 */
	public String getOverview()
	{
		return _overview;
	}

	/**
	 * Checks if value of attribute sensitive is set.
	 * @return true if sensitive is set, false otherwise.
	 */
	public boolean isSensitive()
	{
		return _sensitive;
	}

	/**
	 * Checks if value of attribute selected is set.
	 * @return true if selected is set, false otherwise.
	 */
	public boolean isSelected()
	{
		return _selected;
	}

	/**
	 * Checks if value of attribute hidden is set.
	 * @return true if hidden is set, false otherwise.
	 */
	public boolean isHidden()
	{
		return _hidden;
	}

	/**
	 * Checks if value of attribute background is set.
	 * @return true if background is set, false otherwise.
	 */
	public boolean isBackground()
	{
		return _background;
	}

	/**
	 * Gets value of layer.
	 * @return value of layer.
	 */
	public int getLayer()
	{
		return _layer;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("MAP_REGION.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("MAP_REGION.color"));
	}

	/**
	 * Gets value of borderColor.
	 * @return value of borderColor.
	 */
	public String getBorderColor()
	{
		return _borderColor;
	}

	/**
	 * Checks whether the borderColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultBorderColor()
	{
		return (_borderColor == getStringValue("MAP_REGION.borderColor"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("MAP_REGION.fontColor"));
	}

	/**
	 * Gets value of tooltipFont.
	 * @return value of tooltipFont.
	 */
	public String getTooltipFont()
	{
		return _tooltipFont;
	}

	/**
	 * Checks whether the tooltipFont attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultTooltipFont()
	{
		return (_tooltipFont == getStringValue("MAP_REGION.tooltipFont"));
	}

	/**
	 * Gets value of tooltipColor.
	 * @return value of tooltipColor.
	 */
	public String getTooltipColor()
	{
		return _tooltipColor;
	}

	/**
	 * Checks whether the tooltipColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultTooltipColor()
	{
		return (_tooltipColor == getStringValue("MAP_REGION.tooltipColor"));
	}

	/**
	 * Gets value of selectedColor.
	 * @return value of selectedColor.
	 */
	public String getSelectedColor()
	{
		return _selectedColor;
	}

	/**
	 * Checks whether the selectedColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultSelectedColor()
	{
		return (_selectedColor == getStringValue("MAP_REGION.selectedColor"));
	}

	/**
	 * Gets value of inactiveSelectedColor.
	 * @return value of inactiveSelectedColor.
	 */
	public String getInactiveSelectedColor()
	{
		return _inactiveSelectedColor;
	}

	/**
	 * Checks whether the inactiveSelectedColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultInactiveSelectedColor()
	{
		return (_inactiveSelectedColor == getStringValue("MAP_REGION.inactiveSelectedColor"));
	}

	/**
	 * Gets value of outlineColor.
	 * @return value of outlineColor.
	 */
	public String getOutlineColor()
	{
		return _outlineColor;
	}

	/**
	 * Checks whether the outlineColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultOutlineColor()
	{
		return (_outlineColor == getStringValue("MAP_REGION.outlineColor"));
	}

	/**
	 * Get sub-component Polygon of index i.
	 * @param i the index of the Polygon to get.
	 * @return sub-component Polygon of index i.
	 */
	public LyPolygon getPolygon(int i)
	{
		if (_polygons == null)
			return null;
		if ((i < 0) || (i >= _polygons.size()))
			return null;
		return _polygons.get(i);
	}

	/**
	 * Get sub-component Polygon count.
	 * @return sub-component Polygon count.
	 */
	public int getPolygonCount()
	{
		if (_polygons == null)
			return 0;
		return _polygons.size();
	}

	/**
	 * Get sub-component MapContent.
	 * @return sub-component MapContent.
	 */
	public LyMapContent getMapContent()
	{
		return _mapContent;
	}

	/**
	 * Get sub-component MapRegion of index i.
	 * @param i the index of the MapRegion to get.
	 * @return sub-component MapRegion of index i.
	 */
	public LyMapRegion getMapRegion(int i)
	{
		if (_mapRegions == null)
			return null;
		if ((i < 0) || (i >= _mapRegions.size()))
			return null;
		return _mapRegions.get(i);
	}

	/**
	 * Get sub-component MapRegion count.
	 * @return sub-component MapRegion count.
	 */
	public int getMapRegionCount()
	{
		if (_mapRegions == null)
			return 0;
		return _mapRegions.size();
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyMapRegion template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Initializes attribute file to the given value.
	 * @param file New value of file.
	 */
	public void setFile(String file)
	{
		if ((file == _file)
			|| ((file != null)
				&& (_file != null)
				&& file.equals(_file)))
			return;

		_file = file;
	}

	/**
	 * Initializes attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
	}

	/**
	 * Initializes attribute tooltip to the given value.
	 * @param tooltip New value of tooltip.
	 */
	public void setTooltip(String tooltip)
	{
		if ((tooltip == _tooltip)
			|| ((tooltip != null)
				&& (_tooltip != null)
				&& tooltip.equals(_tooltip)))
			return;

		_tooltip = tooltip;
	}

	/**
	 * Sets attribute smallImage to the given value.
	 * @param smallImage New value of smallImage.
	 */
	public void setSmallImage(String smallImage)
	{
		if ((smallImage == _smallImage)
			|| ((smallImage != null)
				&& (_smallImage != null)
				&& smallImage.equals(_smallImage)))
			return;

		_smallImage = smallImage;
		if (_internalView != null)
			((LyGraphicalMapRegion)_internalView).setSmallImage(smallImage);
	}

	/**
	 * Sets attribute smallImageTooltip to the given value.
	 * @param smallImageTooltip New value of smallImageTooltip.
	 */
	public void setSmallImageTooltip(String smallImageTooltip)
	{
		if ((smallImageTooltip == _smallImageTooltip)
			|| ((smallImageTooltip != null)
				&& (_smallImageTooltip != null)
				&& smallImageTooltip.equals(_smallImageTooltip)))
			return;

		_smallImageTooltip = smallImageTooltip;
		if (_internalView != null)
			((LyGraphicalMapRegion)_internalView).setSmallImageTooltip(smallImageTooltip);
	}

	/**
	 * Sets attribute weight to the given value.
	 * @param weight New value of weight.
	 */
	public void setWeight(float weight)
	{
		if (weight == _weight)
			return;
		_weight = weight;
		if (_internalView != null)
			((LyGraphicalMapRegion)_internalView).setWeight(weight);
	}

	/**
	 * Initializes attribute overview to the given value.
	 * @param overview New value of overview.
	 */
	public void setOverview(String overview)
	{
		if ((overview == _overview)
			|| ((overview != null)
				&& (_overview != null)
				&& overview.equals(_overview)))
			return;

		_overview = overview;
	}

	/**
	 * Initializes attribute sensitive to the given value.
	 * @param sensitive Boolean indicating if attribute sensitive is set or unset.
	 */
	public void setSensitive(boolean sensitive)
	{
		if (sensitive == _sensitive)
			return;
		_sensitive = sensitive;
	}

	/**
	 * Sets attribute selected to the given value.
	 * @param selected Boolean indicating if attribute selected is set or unset.
	 */
	public void setSelected(boolean selected)
	{
		if (selected == _selected)
			return;
		_selected = selected;
		if (_internalView != null)
			((LyGraphicalMapRegion)_internalView).setSelected(selected);
	}

	/**
	 * Sets attribute hidden to the given value.
	 * @param hidden Boolean indicating if attribute hidden is set or unset.
	 */
	public void setHidden(boolean hidden)
	{
		if (hidden == _hidden)
			return;
		_hidden = hidden;
		if (_internalView != null)
			((LyGraphicalMapRegion)_internalView).setHidden(hidden);
	}

	/**
	 * Initializes attribute background to the given value.
	 * @param background Boolean indicating if attribute background is set or unset.
	 */
	public void setBackground(boolean background)
	{
		if (background == _background)
			return;
		_background = background;
	}

	/**
	 * Initializes attribute layer to the given value.
	 * @param layer New value of layer.
	 */
	public void setLayer(int layer)
	{
		if (layer == _layer)
			return;
		_layer = layer;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute borderColor to the given value.
	 * @param borderColor New value of borderColor.
	 */
	public void setBorderColor(String borderColor)
	{
		if ((borderColor == _borderColor)
			|| ((borderColor != null)
				&& (_borderColor != null)
				&& borderColor.equals(_borderColor)))
			return;

		_borderColor = borderColor;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Initializes attribute tooltipFont to the given value.
	 * @param tooltipFont New value of tooltipFont.
	 */
	public void setTooltipFont(String tooltipFont)
	{
		if ((tooltipFont == _tooltipFont)
			|| ((tooltipFont != null)
				&& (_tooltipFont != null)
				&& tooltipFont.equals(_tooltipFont)))
			return;

		_tooltipFont = tooltipFont;
	}

	/**
	 * Initializes attribute tooltipColor to the given value.
	 * @param tooltipColor New value of tooltipColor.
	 */
	public void setTooltipColor(String tooltipColor)
	{
		if ((tooltipColor == _tooltipColor)
			|| ((tooltipColor != null)
				&& (_tooltipColor != null)
				&& tooltipColor.equals(_tooltipColor)))
			return;

		_tooltipColor = tooltipColor;
	}

	/**
	 * Initializes attribute selectedColor to the given value.
	 * @param selectedColor New value of selectedColor.
	 */
	public void setSelectedColor(String selectedColor)
	{
		if ((selectedColor == _selectedColor)
			|| ((selectedColor != null)
				&& (_selectedColor != null)
				&& selectedColor.equals(_selectedColor)))
			return;

		_selectedColor = selectedColor;
	}

	/**
	 * Initializes attribute inactiveSelectedColor to the given value.
	 * @param inactiveSelectedColor New value of inactiveSelectedColor.
	 */
	public void setInactiveSelectedColor(String inactiveSelectedColor)
	{
		if ((inactiveSelectedColor == _inactiveSelectedColor)
			|| ((inactiveSelectedColor != null)
				&& (_inactiveSelectedColor != null)
				&& inactiveSelectedColor.equals(_inactiveSelectedColor)))
			return;

		_inactiveSelectedColor = inactiveSelectedColor;
	}

	/**
	 * Initializes attribute outlineColor to the given value.
	 * @param outlineColor New value of outlineColor.
	 */
	public void setOutlineColor(String outlineColor)
	{
		if ((outlineColor == _outlineColor)
			|| ((outlineColor != null)
				&& (_outlineColor != null)
				&& outlineColor.equals(_outlineColor)))
			return;

		_outlineColor = outlineColor;
	}

	/**
	 * Adds a sub-component polygon.
	 * @param polygon The polygon to add to polygons list.
	 */
	public void addPolygon(LyPolygon polygon)
	{
		insertPolygon(polygon, -1);
	}

	/**
	 * Inserts a sub-component polygon at given position.
	 * @param polygon The polygon to add to polygons list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertPolygon(LyPolygon polygon, int pos)
	{
		if (_polygons == null)
			_polygons = new ArrayList<LyPolygon>(1);
		else if (_polygons.contains(polygon))
			return;
		if (pos >= 0)
			_polygons.add(pos,polygon);
		else
			_polygons.add(polygon);
		if (polygon != null)
			polygon.setParent(this);
	}

	/**
	 * Sets sub-component mapContent.
	 * @param mapContent New value of mapContent
	 */
	public void setMapContent(LyMapContent mapContent)
	{
		if (_mapContent != null)
			_mapContent.setParent(null);
		_mapContent = mapContent;
		if (mapContent != null)
			mapContent.setParent(this);
	}

	/**
	 * Adds a sub-component mapRegion.
	 * @param mapRegion The mapRegion to add to mapRegions list.
	 */
	public void addMapRegion(LyMapRegion mapRegion)
	{
		insertMapRegion(mapRegion, -1);
	}

	/**
	 * Inserts a sub-component mapRegion at given position.
	 * @param mapRegion The mapRegion to add to mapRegions list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertMapRegion(LyMapRegion mapRegion, int pos)
	{
		if (_mapRegions == null)
		{
			_mapRegions = new ArrayList<LyMapRegion>(1);
			_children = _mapRegions;
		}
		else if (_mapRegions.contains(mapRegion))
			return;
		if (pos >= 0)
			_mapRegions.add(pos,mapRegion);
		if (mapRegion != null)
			mapRegion.setParent(this);
		if (_internalView != null)
			((LyGraphicalMapRegion)_internalView).addMapRegion(mapRegion, pos);
	}

	/**
	 * Removes a sub-component mapRegion.
	 * @param mapRegion The mapRegion to remove to mapRegions list.
	 * @exception IllegalStateException if mapRegions list is empty.
	 * @exception IllegalArgumentException if mapRegions list doesn't contain mapRegion.
	 */
	public void removeMapRegion(LyMapRegion mapRegion)
	{
		if (_mapRegions == null)
			throw new IllegalStateException("BF AppEngine : mapRegions list is empty");
		if (!_mapRegions.contains(mapRegion))
			throw new IllegalArgumentException("BF AppEngine : mapRegions list doesn't contain mapRegion");

		if (_internalView != null)
			((LyGraphicalMapRegion)_internalView).removeMapRegion(mapRegion);
		mapRegion.setParent(null);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<MAP_REGION");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_file != null))
				|| ((_template != null) && (_file != _template._file)))
			pw.print(" file='" + LyString.getXmlString(_file) + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (((_template == null) && (_tooltip != null))
				|| ((_template != null) && (_tooltip != _template._tooltip)))
			pw.print(" tooltip='" + LyString.getXmlString(_tooltip) + "'");
		if (((_template == null) && (_smallImage != getStringValue("MAP_REGION.smallImage", "$ALL.smallImage")))
				|| ((_template != null) && (_smallImage != _template._smallImage)))
			pw.print(" smallImage='" + LyString.getXmlString(_smallImage) + "'");
		if (((_template == null) && (_smallImageTooltip != null))
				|| ((_template != null) && (_smallImageTooltip != _template._smallImageTooltip)))
			pw.print(" smallImageTooltip='" + LyString.getXmlString(_smallImageTooltip) + "'");
		if (((_template == null) && (_weight != getFloatValue("MAP_REGION.weight", 0F)))
				|| ((_template != null) && (_weight != _template._weight)))
			pw.print(" weight='" + _weight + "'");
		if (((_template == null) && (_overview != null))
				|| ((_template != null) && (_overview != _template._overview)))
			pw.print(" overview='" + LyString.getXmlString(_overview) + "'");
		if (((_template == null) && !_sensitive)
				|| ((_template != null) && (_sensitive != _template._sensitive)))
			pw.print(" sensitive='" + _sensitive + "'");
		if (((_template == null) && _selected)
				|| ((_template != null) && (_selected != _template._selected)))
			pw.print(" selected='" + _selected + "'");
		if (((_template == null) && _hidden)
				|| ((_template != null) && (_hidden != _template._hidden)))
			pw.print(" hidden='" + _hidden + "'");
		if (((_template == null) && _background)
				|| ((_template != null) && (_background != _template._background)))
			pw.print(" background='" + _background + "'");
		if (((_template == null) && (_layer != getIntValue("MAP_REGION.layer", -1)))
				|| ((_template != null) && (_layer != _template._layer)))
			pw.print(" layer='" + _layer + "'");
		if (((_template == null) && (_font != getStringValue("MAP_REGION.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("MAP_REGION.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_borderColor != getStringValue("MAP_REGION.borderColor", "$ALL.borderColor")))
				|| ((_template != null) && (_borderColor != _template._borderColor)))
			pw.print(" borderColor='" + LyString.getXmlString(_borderColor) + "'");
		if (((_template == null) && (_fontColor != getStringValue("MAP_REGION.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && (_tooltipFont != getStringValue("MAP_REGION.tooltipFont", "$ALL.tooltipFont")))
				|| ((_template != null) && (_tooltipFont != _template._tooltipFont)))
			pw.print(" tooltipFont='" + LyString.getXmlString(_tooltipFont) + "'");
		if (((_template == null) && (_tooltipColor != getStringValue("MAP_REGION.tooltipColor", "$ALL.tooltipColor")))
				|| ((_template != null) && (_tooltipColor != _template._tooltipColor)))
			pw.print(" tooltipColor='" + LyString.getXmlString(_tooltipColor) + "'");
		if (((_template == null) && (_selectedColor != getStringValue("MAP_REGION.selectedColor", "$ALL.selectedColor")))
				|| ((_template != null) && (_selectedColor != _template._selectedColor)))
			pw.print(" selectedColor='" + LyString.getXmlString(_selectedColor) + "'");
		if (((_template == null) && (_inactiveSelectedColor != getStringValue("MAP_REGION.inactiveSelectedColor", "$ALL.inactiveSelectedColor")))
				|| ((_template != null) && (_inactiveSelectedColor != _template._inactiveSelectedColor)))
			pw.print(" inactiveSelectedColor='" + LyString.getXmlString(_inactiveSelectedColor) + "'");
		if (((_template == null) && (_outlineColor != getStringValue("MAP_REGION.outlineColor", "$ALL.outlineColor")))
				|| ((_template != null) && (_outlineColor != _template._outlineColor)))
			pw.print(" outlineColor='" + LyString.getXmlString(_outlineColor) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	polygons;
		int			i, n;

		n = (_polygons == null ? 0 : _polygons.size());
		for (i = 0; i < n; i++)
		{
			polygons = _polygons.get(i);
			polygons.writeXml(pw, debug);
		}

		if (_mapContent != null)
			_mapContent.writeXml(pw, debug);
		LyComponent	mapRegions;

		n = (_mapRegions == null ? 0 : _mapRegions.size());
		for (i = 0; i < n; i++)
		{
			mapRegions = _mapRegions.get(i);
			mapRegions.writeXml(pw, debug);
		}


		pw.println("</MAP_REGION>");
	}
}
