/*
 * LyMenuBar.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.view;

import leon.view.getinterface.*;
import leon.view.setinterface.*;

import java.util.*;


/**
 * The LyMenuBar class implements generic behavior for the MENUBAR components as described in
 * views.dtd regardless of the requested target (win, swing, html...). It represents a menu bar.
 *
 * @author   Lyria
 * @version  $Id: LyMenuBar.java 35018 2012-03-23 15:17:40Z jlsaouli $
 * @see      LyComponent
 */
public class LyMenuBar extends LyMenuBarBase
	implements LyGetIdInterface
{
	public static final String __VERSION = "$Id: LyMenuBar.java 35018 2012-03-23 15:17:40Z jlsaouli $";

	/** Not found entry. */
	protected static final Object NOT_FOUND = new Object();

	/** Automatically sort menu content or not. */
	protected boolean _sorted = true;

	/** Items cache. */
	protected Hashtable<String, Object> _menusCache;

	/**
	 * Constructor with all required attributes.
	 */
	public LyMenuBar()
	{
	}

	/* Protected methods for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 *
	 * @param  node  Node from which the component is constructed.
	 *
	 * @see    LyViewNodeFactory#buildElement
	 */
	protected LyMenuBar(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id", null);
	}

	/* Utility methods */

	/**
	 * Looks for the LyMenu object that has a given id and a given controller. The research is made
	 * through the hierarchical tree, all the children are taken into account. This is very usefull
	 * to retrieve a menu associated to a sub-controller inside a compound view.
	 *
	 * @param   id            the identifier of the menu
	 * @param   controllerId  the identifier of the controller to which the menu should belong.
	 *
	 * @return  the the corresponding LyMenu object or null if not was found.
	 */
	public LyMenu findMenu(String id, String controllerId)
	{
		if ((_menusCache != null)
				&& _menusCache.containsKey(id + controllerId))
		{
			Object menu = _menusCache.get(id + controllerId);

			return ((menu instanceof LyMenu) ? (LyMenu)menu : null);
		}

		LyMenu			   result = null, menu;
		LyComponent		   component;
		int				   i, n;
		LinkedList<LyMenu> watchedMenus = new LinkedList<LyMenu>();

		watchedMenus.addAll(_menus);

		// search while there are menus to watch
		while ((result == null) && !watchedMenus.isEmpty())
		{
			menu = watchedMenus.removeFirst();

			if ((menu.getId() != null)
					&& menu.getId().equals(id)
					&& menu.getController().equals(controllerId))
				result = menu;
			else
				for (i = 0, n = menu.getComponentCount(); i < n; i++)
				{
					component = menu.getComponent(i);
					if (component instanceof LyMenu)
						watchedMenus.addLast((LyMenu)component);
				}
		}

		if (_menusCache == null)
			_menusCache = new Hashtable<String, Object>();

		if (result == null)
			_menusCache.put(id + controllerId, NOT_FOUND);
		else
			_menusCache.put(id + controllerId, result);

		return result;
	} // end method findMenu

	/**
	 * Looks for the LyItem object that has a given id and a given controller. The research is made
	 * through the hierarchical tree, all the children are taken into account. This is very usefull
	 * to retrieve a menu item associated to a sub-controller inside a compound view.
	 *
	 * @param   id            the identifier of the menu item
	 * @param   controllerId  the identifier of the controller to which the menu item should belong.
	 *
	 * @return  the the corresponding LyItem object or null if not found.
	 */
	public LyItem findItem(String id, String controllerId)
	{
		if ((_menusCache != null)
				&& _menusCache.containsKey(id + controllerId))
		{
			Object item = _menusCache.get(id + controllerId);

			return ((item instanceof LyItem) ? (LyItem)item : null);
		}

		LyItem			   result = null, item;
		LyMenu			   menu;
		LyComponent		   component;
		int				   i, n;
		LinkedList<LyMenu> watchedMenus = new LinkedList<LyMenu>();

		watchedMenus.addAll(_menus);

		id = id.intern();
		controllerId = controllerId.intern();

		// search while there are menus to watch
		while ((result == null) && !watchedMenus.isEmpty())
		{
			menu = watchedMenus.removeFirst();
			for (i = 0, n = menu.getComponentCount(); i < n; i++)
			{
				component = menu.getComponent(i);
				if (component instanceof LyMenu) // add the menu found into the list of menu to watch
					watchedMenus.addLast((LyMenu)component);
				else if (component instanceof LyItem)
				{
					item = (LyItem)component;
					if (item.getId().equals(id)
							&& item.getController().equals(controllerId))
					{ // if the item is found : break the loop and return the result.
						result = item;
						break;
					}
				}
			}
		}

		if (_menusCache == null)
			_menusCache = new Hashtable<String, Object>();

		if (result == null)
			_menusCache.put(id + controllerId, NOT_FOUND);
		else
			_menusCache.put(id + controllerId, result);

		return result;
	} // end method findItem

	/**
	 * Indicates wether this menubar is idle or not (idle menu bar do nor refresh the menu contents
	 * in views dynamicallly.
	 *
	 * @return  idle status
	 */
	public boolean isIdle()
	{
		if (_internalView instanceof LyIsIdleInterface)
			return ((LyIsIdleInterface)_internalView).isIdle();

		return false;
	}

	/**
	 * Indicates wether this menubar is idle or not (idle menu bar do nor refresh the menu contents
	 * in views dynamicallly.
	 *
	 * @param  idle  status
	 */
	public void setIdle(boolean idle)
	{
		if (_internalView instanceof LySetIdleInterface)
			((LySetIdleInterface)_internalView).setIdle(idle);
	}

	/**
	 * Adds a sub-component menu.
	 *
	 * @param  menu  The menu to add to components list.
	 */
	public void addMenu(LyMenu menu)
	{
		if ((_sorted) && (getMenuCount() > 0))
		{
			// DEFAULT ORDER IS LY_MENU_FILE, LY_MENU_EDITION, LY_MENU_DISPLAY, LY_MENU_ACTIONS then ...
			Hashtable<String, Integer> order = new Hashtable<String, Integer>();

			order.put("LY_MENU_FILE", new Integer(0));
			order.put("LY_MENU_WINDOW", new Integer(1));
			order.put("LY_MENU_EDITION", new Integer(2));
			order.put("LY_MENU_DISPLAY", new Integer(3));
			order.put("LY_MENU_ACTIONS", new Integer(4));

			Integer index = order.get(menu.getLabel());
			Integer index2;
			LyMenu  menu2;

			if (index != null)
			{
				int insertIndex = 0;

				for (int i = 0; i < getMenuCount(); i++)
				{
					menu2 = getMenu(i);
					index2 = order.get(menu2.getLabel());

					if (index2 != null)
					{
						if (index.intValue() > index2.intValue())
							insertIndex = i + 1; // After this one
						else
						{
							insertIndex = i;
							break; // Before this one
						}
					}
				}

				if (insertIndex < getMenuCount())
				{
					insertMenu(menu, insertIndex);
					return;
				}
			}
		}

		insertMenu(menu, -1);
	} // end method addMenu

	/**
	 * Adds an item to the cache of items.
	 *
	 * @param  item  the item to be added to the cache.
	 */
	public void addItemToCache(LyItem item)
	{
		if (item == null)
			return;

		if (_menusCache == null)
			_menusCache = new Hashtable<String, Object>();

		_menusCache.put(item.getId() + item.getController(), item);
	}

	/**
	 * Indictates wether this menubar is Automatically sorted for its menu content or not.
	 *
	 * @param  sorted  status
	 */
	public void setSorted(boolean sorted)
	{
		_sorted = sorted;
	}

	/**
	 * Removes a sub-component menu.
	 *
	 * @param      menu  The menu to remove to menus list.
	 *
	 * @exception  IllegalStateException     if menus list is empty.
	 * @exception  IllegalArgumentException  if menus list doesn't contain menu.
	 */
	public void removeMenu(LyMenu menu)
	{
		_menusCache = null;
		super.removeMenu(menu);
	}
} // end class LyMenuBar
