/*
 * LyMenuBarBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.misc.LyString;


/**
 * The LyMenuBarBase class is a generated Java implementation that corresponds to the element 
 * MENUBAR of the descriptive grammar of the views (views.dtd).
 * <B>&lt;MENUBAR&gt; : Menu bar</B>
 * A menu bar allows displaying a set of menus.
 * It can be displayed in a main window of a Application Engine application
 * @see leon.view.LyComponent
 * @see leon.view.LyLyMenuBar
 * See also tag <A href="../../conf/dtd/view/MENUBAR.html">MENUBAR</A> of grammar leon_views.dtd.
 */
public abstract class LyMenuBarBase extends LyComponent
	implements  LyGetIdInterface
{
	public static final String			__VERSION = "$Id$";


	/* Attributes */

	/**
	 * Identifier of the menu bar.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyMenuBar			_template;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;


	/* Sub components */

	/**
	 * Icon of the menu bar.
	 */
	protected LyIcon			_icon;

	/**
	 * Menus contained in the menu bar.
	 */
	protected ArrayList<LyMenu>		_menus;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_font = getStringValue("MENUBAR.font", "$ALL.font");
		_color = getStringValue("MENUBAR.color", "$ALL.color");
		_fontColor = getStringValue("MENUBAR.fontColor", "$ALL.fontColor");

		// Initialize sub components
		_icon = null;
		_menus = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyMenuBar	model = (LyMenuBar)component;

		_model = component;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		if (model._icon != null)
			addChild(model._icon.cloneComponent());
		LyMenu	menus;
		int			i, n;

		n = (model._menus == null ? 0 : model._menus.size());
		for (i = 0; i < n; i++)
		{
			menus = model._menus.get(i);
			if (menus != null)
				menus = (LyMenu)menus.cloneComponent();
			addChild(menus);
		}
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyMenuBarBase()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyMenuBarBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyMenuBar)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyIcon)
			setIcon((LyIcon)child);
		else if (child instanceof LyMenu)
			addMenu((LyMenu)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyMenuBar clone = new LyMenuBar();

		clone.initFromModel(this);
		clone._template = (LyMenuBar)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyMenuBar getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("MENUBAR.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("MENUBAR.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("MENUBAR.fontColor"));
	}

	/**
	 * Get sub-component Icon.
	 * @return sub-component Icon.
	 */
	public LyIcon getIcon()
	{
		return _icon;
	}

	/**
	 * Get sub-component Menu of index i.
	 * @param i the index of the Menu to get.
	 * @return sub-component Menu of index i.
	 */
	public LyMenu getMenu(int i)
	{
		if (_menus == null)
			return null;
		if ((i < 0) || (i >= _menus.size()))
			return null;
		return _menus.get(i);
	}

	/**
	 * Get sub-component Menu count.
	 * @return sub-component Menu count.
	 */
	public int getMenuCount()
	{
		if (_menus == null)
			return 0;
		return _menus.size();
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyMenuBar template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Sets sub-component icon.
	 * @param icon New value of icon
	 */
	public void setIcon(LyIcon icon)
	{
		if (_icon != null)
			_icon.setParent(null);
		_icon = icon;
		if (icon != null)
			icon.setParent(this);
	}

	/**
	 * Adds a sub-component menu.
	 * @param menu The menu to add to menus list.
	 */
	public void addMenu(LyMenu menu)
	{
		insertMenu(menu, -1);
	}

	/**
	 * Inserts a sub-component menu at given position.
	 * @param menu The menu to add to menus list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertMenu(LyMenu menu, int pos)
	{
		if (_menus == null)
		{
			_menus = new ArrayList<LyMenu>(1);
			_children = _menus;
		}
		else if (_menus.contains(menu))
			return;
		if (pos >= 0)
			_menus.add(pos,menu);
		if (menu != null)
			menu.setParent(this);
		if (_internalView != null)
			((LyGraphicalMenuBar)_internalView).addMenu(menu, pos);
	}

	/**
	 * Removes a sub-component menu.
	 * @param menu The menu to remove to menus list.
	 * @exception IllegalStateException if menus list is empty.
	 * @exception IllegalArgumentException if menus list doesn't contain menu.
	 */
	public void removeMenu(LyMenu menu)
	{
		if (_menus == null)
			throw new IllegalStateException("BF AppEngine : menus list is empty");
		if (!_menus.contains(menu))
			throw new IllegalArgumentException("BF AppEngine : menus list doesn't contain menu");

		if (_internalView != null)
			((LyGraphicalMenuBar)_internalView).removeMenu(menu);
		menu.setParent(null);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<MENUBAR");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_font != getStringValue("MENUBAR.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("MENUBAR.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("MENUBAR.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		if (_icon != null)
			_icon.writeXml(pw, debug);
		LyComponent	menus;
		int			i, n;

		n = (_menus == null ? 0 : _menus.size());
		for (i = 0; i < n; i++)
		{
			menus = _menus.get(i);
			menus.writeXml(pw, debug);
		}


		pw.println("</MENUBAR>");
	}
}
