/*
 * LyMenuBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyMenuBase class is a generated Java implementation that corresponds to the element 
 * MENU of the descriptive grammar of the views (views.dtd).
 * <B>&lt;MENU&gt; : Menu</B>
 * A menu contains a set of menu, sub-menus entries and separators.
 * @see leon.view.LyComponent
 * @see leon.view.LyLyMenu
 * See also tag <A href="../../conf/dtd/view/MENU.html">MENU</A> of grammar leon_views.dtd.
 */
public abstract class LyMenuBase extends LyComponent
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LyGetControllerInterface,
				LySetControllerInterface,
				LySetEnabledInterface
{
	public static final String			__VERSION = "$Id$";


	/* Attributes */

	/**
	 * Identifier of the menu.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyMenu			_template;

	/**
	 * Text of the menu.
	 */
	protected String			_label;

	/**
	 * Reference of the controller that manages the menu (i.e. handles user demands and
	 * refresh content).
	 */
	protected String			_controller;

	/**
	 * Indicates whether the menu is enabled or not (true if it is enabled).
	 */
	protected boolean			_enabled;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;


	/* Sub components */

	/**
	 * Components of a menu : either menu items, separators or sub-menus. 
	 */
	protected ArrayList<LyComponent>		_components;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_controller = null;
		_enabled = getBooleanValue("MENU.enabled", true);
		_font = getStringValue("MENU.font", "$ALL.font");
		_color = getStringValue("MENU.color", "$ALL.color");
		_fontColor = getStringValue("MENU.fontColor", "$ALL.fontColor");

		// Initialize sub components
		_components = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyMenu	model = (LyMenu)component;

		_model = component;
		_label = model._label;
		_controller = model._controller;
		_enabled = model._enabled;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyComponent	components;
		int			i, n;

		n = (model._components == null ? 0 : model._components.size());
		for (i = 0; i < n; i++)
		{
			components = model._components.get(i);
			if (components != null)
				components = components.cloneComponent();
			addChild(components);
		}
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	protected LyMenuBase()
	{
	}

	/**
	 * Constructor with all required attributes.
	 * @param label Text of the menu.
	 */
	public LyMenuBase(String label)
	{
		// Fill up required attributes with parameters
		_label = label;
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyMenuBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyMenu)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("controller"))
			_controller = node.getString("controller");
		if (node.isSpecified("enabled"))
			_enabled = node.getBoolean("enabled");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyItem)
			addItem((LyItem)child);
		else if (child == null)
			addSep();
		else if (child instanceof LyMenu)
			addMenu((LyMenu)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyMenu clone = new LyMenu();

		clone.initFromModel(this);
		clone._template = (LyMenu)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyMenu getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of controller.
	 * @return value of controller.
	 */
	public String getController()
	{
		return _controller;
	}

	/**
	 * Checks if value of attribute enabled is set.
	 * @return true if enabled is set, false otherwise.
	 */
	public boolean isEnabled()
	{
		return _enabled;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("MENU.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("MENU.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("MENU.fontColor"));
	}

	/**
	 * Get sub-component Component of index i.
	 * @param i the index of the Component to get.
	 * @return sub-component Component of index i.
	 */
	public LyComponent getComponent(int i)
	{
		if (_components == null)
			return null;
		if ((i < 0) || (i >= _components.size()))
			return null;
		return _components.get(i);
	}

	/**
	 * Get sub-component Component count.
	 * @return sub-component Component count.
	 */
	public int getComponentCount()
	{
		if (_components == null)
			return 0;
		return _components.size();
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyMenu template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Initializes attribute label to the given value.
	 * @param label New value of label.
	 * @exception IllegalArgumentException if label is null.
	 */
	public void setLabel(String label)
	{
		if (label == null)
			throw new IllegalArgumentException("BF AppEngine : attribute label is required (can't be null)");

		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
	}

	/**
	 * Initializes attribute controller to the given value.
	 * @param controller New value of controller.
	 */
	public void setController(String controller)
	{
		if ((controller == _controller)
			|| ((controller != null)
				&& (_controller != null)
				&& controller.equals(_controller)))
			return;

		_controller = controller;
	}

	/**
	 * Sets attribute enabled to the given value.
	 * @param enabled Boolean indicating if attribute enabled is set or unset.
	 */
	public void setEnabled(boolean enabled)
	{
		if (enabled == _enabled)
			return;
		_enabled = enabled;
		if (_internalView != null)
			((LyGraphicalMenu)_internalView).setEnabled(enabled);
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Adds a sub-component item.
	 * @param item The item to add to components list.
	 */
	public void addItem(LyItem item)
	{
		insertItem(item, -1);
	}

	/**
	 * Inserts a sub-component item at given position.
	 * @param item The item to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertItem(LyItem item, int pos)
	{
		if (_components == null)
		{
			_components = new ArrayList<LyComponent>(1);
			_children = _components;
		}
		else if (_components.contains(item))
			return;
		if (pos >= 0)
			_components.add(pos,item);
		if (item != null)
			item.setParent(this);
		if (_internalView != null)
			((LyGraphicalMenu)_internalView).addComponent(item, pos);
	}

	/**
	 * Removes a sub-component item.
	 * @param item The item to remove to components list.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain item.
	 */
	public void removeItem(LyItem item)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if (!_components.contains(item))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain item");

		if (_internalView != null)
			((LyGraphicalMenu)_internalView).removeComponent(item);
		item.setParent(null);
	}

	/**
	 * Adds a separator.
	 */
	public void addSep()
	{
		insertSep(-1);
	}

	/**
	 * Inserts a separator at the given position.
	 * @param pos	Position where the separator is inserted.
	 */
	public void insertSep(int pos)
	{
		LyItem sep = new LyItem(SEP_ID, null);
		insertItem(sep, pos);
	}

	/**
	 * Removes a separator at the given position.
	 * @param pos	Position where the separator is removed.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain a separator at the given position.
	 */
	public void removeSep(int pos)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if ((pos < 0)
				|| (pos >= _components.size())
				|| ((_components.get(pos) != null)
					&& (_components.get(pos).getId() != SEP_ID)))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain a separator at pos " + pos);

		_components.remove(pos);
	}

	/**
	 * Adds a sub-component menu.
	 * @param menu The menu to add to components list.
	 */
	public void addMenu(LyMenu menu)
	{
		insertMenu(menu, -1);
	}

	/**
	 * Inserts a sub-component menu at given position.
	 * @param menu The menu to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertMenu(LyMenu menu, int pos)
	{
		if (_components == null)
		{
			_components = new ArrayList<LyComponent>(1);
			_children = _components;
		}
		else if (_components.contains(menu))
			return;
		if (pos >= 0)
			_components.add(pos,menu);
		if (menu != null)
			menu.setParent(this);
		if (_internalView != null)
			((LyGraphicalMenu)_internalView).addComponent(menu, pos);
	}

	/**
	 * Removes a sub-component menu.
	 * @param menu The menu to remove to components list.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain menu.
	 */
	public void removeMenu(LyMenu menu)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if (!_components.contains(menu))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain menu");

		if (_internalView != null)
			((LyGraphicalMenu)_internalView).removeComponent(menu);
		menu.setParent(null);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<MENU");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlString(_controller) + "'");
		if (((_template == null) && !_enabled)
				|| ((_template != null) && (_enabled != _template._enabled)))
			pw.print(" enabled='" + _enabled + "'");
		if (((_template == null) && (_font != getStringValue("MENU.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("MENU.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("MENU.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	components;
		int			i, n;

		n = (_components == null ? 0 : _components.size());
		for (i = 0; i < n; i++)
		{
			components = _components.get(i);
			if ((components == null)
					|| (components.getId() == SEP_ID))
				pw.println("<SEP/>");
			else
				components.writeXml(pw, debug);
		}


		pw.println("</MENU>");
	}
}
