/*
 * LyMessage.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyMessage class is a generated Java implementation that corresponds to the element 
 * MESSAGE of the descriptive grammar of the views (views.dtd).
 * <B>&lt;MESSAGE&gt; : Message</B>
 * A message is placed inside a message bar.
 * It allows displaying a text and an image.
 * @see leon.view.LyComponent
 * See also tag <A href="../../conf/dtd/view/MESSAGE.html">MESSAGE</A> of grammar leon_views.dtd.
 */
public class LyMessage extends LyComponent
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface,
				LySetImageInterface,
				LyGetControllerInterface,
				LySetControllerInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'align'. */
	public enum Align
	{
		LEFT,
		RIGHT,
		CENTER;
	}


	/* Static methods */

	/**
	 * Utility method to get align value from a string.
	 * @param token Token string
	 * @return Align value corresponding to this token
	 */
	public static final Align getAlign(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("LEFT")
				|| token.equals("0"))
			return Align.LEFT;
		else if (token.equalsIgnoreCase("RIGHT")
				|| token.equals("1"))
			return Align.RIGHT;
		else if (token.equalsIgnoreCase("CENTER")
				|| token.equals("2"))
			return Align.CENTER;
		else
			return Align.LEFT;
	}


	/* Attributes */

	/**
	 * Identifier of the message bar.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyMessage			_template;

	/**
	 * Text of the message.
	 */
	protected String			_label;

	/**
	 * Image associated to the message.
	 */
	protected String			_image;

	/**
	 * Reference of the controller that manages the message (i.e. handles user demands 
	 * and refresh content).
	 */
	protected String			_controller;

	/**
	 * Size of the message.
	 * The expected value is a positive integer value corresponding to a weight (the 
	 * weights of the different messages contained in the message bar allow performing 
	 * the barycentric distribution of the space allocated to each of them).
	 * A negative or null value is equivalent to the value 1.
	 */
	protected int				_size;

	/**
	 * Alignment of the message.
	 */
	protected Align				_align;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_image = getStringValue("MESSAGE.image", "$ALL.image");
		_controller = null;
		_size = getIntValue("MESSAGE.size", 0);
		_align = Align.LEFT;
		_font = getStringValue("MESSAGE.font", "$ALL.font");
		_color = getStringValue("MESSAGE.color", "$ALL.color");
		_fontColor = getStringValue("MESSAGE.fontColor", "$ALL.fontColor");
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyMessage	model = (LyMessage)component;

		_model = component;
		_label = model._label;
		_image = model._image;
		_controller = model._controller;
		_size = model._size;
		_align = model._align;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	protected LyMessage()
	{
		_id = getInternalId();
	}

	/**
	 * Constructor with all required attributes.
	 * @param id Identifier of the message bar.
	 * @param label Text of the message.
	 */
	public LyMessage(String id, String label)
	{
		// Fill up required attributes with parameters
		_id = (id != null ? id.intern() : null);
		_label = label;
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyMessage(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyMessage)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("image"))
			_image = node.getString("image");
		if (node.isSpecified("controller"))
			_controller = node.getString("controller");
		if (node.isSpecified("size"))
			_size = node.getInt("size");
		if (node.isSpecified("align"))
			_align = getAlign(node.getString("align"));
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyMessage clone = new LyMessage();

		clone.initFromModel(this);
		clone._template = this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyMessage getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of image.
	 * @return value of image.
	 */
	public String getImage()
	{
		return _image;
	}

	/**
	 * Checks whether the image attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultImage()
	{
		return (_image == getStringValue("MESSAGE.image"));
	}

	/**
	 * Gets value of controller.
	 * @return value of controller.
	 */
	public String getController()
	{
		return _controller;
	}

	/**
	 * Gets value of size.
	 * @return value of size.
	 */
	public int getSize()
	{
		return _size;
	}

	/**
	 * Gets value of align.
	 * @return value of align.
	 */
	public Align getAlign()
	{
		return _align;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("MESSAGE.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("MESSAGE.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("MESSAGE.fontColor"));
	}


	/* Modifiers */

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyMessage template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 * @exception IllegalArgumentException if label is null.
	 */
	public void setLabel(String label)
	{
		if (label == null)
			throw new IllegalArgumentException("BF AppEngine : attribute label is required (can't be null)");

		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView != null)
			((LyGraphicalMessage)_internalView).setLabel(label);
	}

	/**
	 * Sets attribute image to the given value.
	 * @param image New value of image.
	 */
	public void setImage(String image)
	{
		if ((image == _image)
			|| ((image != null)
				&& (_image != null)
				&& image.equals(_image)))
			return;

		_image = image;
		if (_internalView != null)
			((LyGraphicalMessage)_internalView).setImage(image);
	}

	/**
	 * Initializes attribute controller to the given value.
	 * @param controller New value of controller.
	 */
	public void setController(String controller)
	{
		if ((controller == _controller)
			|| ((controller != null)
				&& (_controller != null)
				&& controller.equals(_controller)))
			return;

		_controller = controller;
	}

	/**
	 * Initializes attribute size to the given value.
	 * @param size New value of size.
	 */
	public void setSize(int size)
	{
		if (size == _size)
			return;
		_size = size;
	}

	/**
	 * Initializes attribute align to the given value.
	 * @param align New value of align.
	 * @exception IllegalArgumentException if align parameter is null.
	 */
	public void setAlign(Align align)
	{
		if (align == _align)
			return;
		if (align == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute align");

		_align = align;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<MESSAGE");
		pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (((_template == null) && (_image != getStringValue("MESSAGE.image", "$ALL.image")))
				|| ((_template != null) && (_image != _template._image)))
			pw.print(" image='" + LyString.getXmlString(_image) + "'");
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlString(_controller) + "'");
		if (((_template == null) && (_size != getIntValue("MESSAGE.size", 0)))
				|| ((_template != null) && (_size != _template._size)))
			pw.print(" size='" + _size + "'");
		if ((_template == null)
			|| (_align != _template._align))
		{
			if ((_align == Align.LEFT) && (_template != null))
				pw.print(" align='LEFT'");
			else if (_align == Align.RIGHT)
				pw.print(" align='RIGHT'");
			else if (_align == Align.CENTER)
				pw.print(" align='CENTER'");
		}
		if (((_template == null) && (_font != getStringValue("MESSAGE.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("MESSAGE.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("MESSAGE.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		pw.println("/>");
	}
}
