/*
 * LyNumberBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;
import java.io.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyNumberBase class is a generated Java implementation that corresponds to the element 
 * NUMBER of the descriptive grammar of the views (views.dtd).
 * <B>&lt;NUMBER&gt; : Input Field of Number Type</B>
 * An input field of Number type allows displaying a graphical component allowing 
 * to input or to consult a number value.
 * It does not comprise any element.
 * It is possible to specify the minimum and the maximum value, the incrementation 
 * step, the number of decimals and the unit.
 * @see leon.view.LyFieldItem
 * @see leon.view.LyLyNumber
 * See also tag <A href="../../conf/dtd/view/NUMBER.html">NUMBER</A> of grammar leon_views.dtd.
 */
public abstract class LyNumberBase extends LyFieldItem
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface
{
	public static final String			__VERSION = "$Id$";


	/* Attributes */

	/**
	 * Identifier of the component.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyNumber			_template;

	/**
	 * Value displayed for the number.
	 */
	protected String			_label;

	/**
	 * Display format of the number.
	 */
	protected String			_format;

	/**
	 * Minimum value.
	 * The provided value can be a positive or a negative integer, or else a non 
	 * integer value (the decimal delimiter being a dot).
	 */
	protected double			_vMin;

	/**
	 * Maximum value.
	 * The provided value can be a positive or a negative integer, or else a non 
	 * integer value (the decimal delimiter being a dot).
	 */
	protected double			_vMax;

	/**
	 * Incrementation step.
	 * The provided value can be a positive or a negative integer, or else a non 
	 * integer value (the decimal delimiter being a dot).
	 */
	protected double			_pitch;

	/**
	 * Number of decimals.
	 * The expected value is a positive integer value.
	 * Any other value is equivalent to the value 0 (indicating a field of integer 
	 * type).
	 */
	protected int				_decimals;

	/**
	 * Indicates if the number value is a modulo of its maximum value. By default the 
	 * number value is not a modulo of its maximum value.
	 */
	protected boolean			_modulo;

	/**
	 * Unit.
	 * The expected value is a string which format is free (this label should 
	 * nevertheless remain short because it it intended to be displayed after the 
	 * graphical component).
	 * Its content is purely informative.
	 */
	protected String			_unit;

	/**
	 * Allows specifying the name of a Java class that replaces the default graphical 
	 * component.
	 */
	protected String			_bean;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * NUMBER.enabledColor
	 */
	protected String			_enabledColor;

	/**
	 * NUMBER.disabledColor
	 */
	protected String			_disabledColor;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * Indicates if arrows are displayed or not on the numeric field.
	 */
	protected boolean			_arrows;

	/**
	 * Image of the button used for increasing the current value.
	 */
	protected String			_arrowUp;

	/**
	 * Image of the button used for decreasing the current value.
	 */
	protected String			_arrowDown;

	/**
	 * Gives the number of digits used to present the numeric field.
	 */
	protected int				_nCols;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_format = null;
		_vMin = getDoubleValue("NUMBER.vMin", Double.NEGATIVE_INFINITY);
		_vMax = getDoubleValue("NUMBER.vMax", Double.POSITIVE_INFINITY);
		_pitch = getDoubleValue("NUMBER.pitch", 0);
		_decimals = getIntValue("NUMBER.decimals", -1);
		_modulo = getBooleanValue("NUMBER.modulo", false);
		_unit = null;
		_bean = null;
		_font = getStringValue("NUMBER.font", "$ALL.font");
		_enabledColor = getStringValue("NUMBER.enabledColor", "$ALL.enabledColor");
		_disabledColor = getStringValue("NUMBER.disabledColor", "$ALL.disabledColor");
		_fontColor = getStringValue("NUMBER.fontColor", "$ALL.fontColor");
		_arrows = getBooleanValue("NUMBER.arrows", true);
		_arrowUp = getStringValue("NUMBER.arrowUp", "$ALL.arrowUp");
		_arrowDown = getStringValue("NUMBER.arrowDown", "$ALL.arrowDown");
		_nCols = getIntValue("NUMBER.nCols", -1);
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyNumber	model = (LyNumber)component;

		_model = component;
		_label = model._label;
		_format = model._format;
		_vMin = model._vMin;
		_vMax = model._vMax;
		_pitch = model._pitch;
		_decimals = model._decimals;
		_modulo = model._modulo;
		_unit = model._unit;
		_bean = model._bean;
		_font = model._font;
		_enabledColor = model._enabledColor;
		_disabledColor = model._disabledColor;
		_fontColor = model._fontColor;
		_arrows = model._arrows;
		_arrowUp = model._arrowUp;
		_arrowDown = model._arrowDown;
		_nCols = model._nCols;
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyNumberBase()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyNumberBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyNumber)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("format"))
			_format = node.getString("format");
		if (node.isSpecified("vMin"))
			_vMin = node.getDouble("vMin");
		if (node.isSpecified("vMax"))
			_vMax = node.getDouble("vMax");
		if (node.isSpecified("pitch"))
			_pitch = node.getDouble("pitch");
		if (node.isSpecified("decimals"))
			_decimals = node.getInt("decimals");
		if (node.isSpecified("modulo"))
			_modulo = node.getBoolean("modulo");
		if (node.isSpecified("unit"))
			_unit = node.getString("unit");
		if (node.isSpecified("bean"))
			_bean = node.getString("bean");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("enabledColor"))
			_enabledColor = node.getString("enabledColor");
		if (node.isSpecified("disabledColor"))
			_disabledColor = node.getString("disabledColor");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("arrows"))
			_arrows = node.getBoolean("arrows");
		if (node.isSpecified("arrowUp"))
			_arrowUp = node.getString("arrowUp");
		if (node.isSpecified("arrowDown"))
			_arrowDown = node.getString("arrowDown");
		if (node.isSpecified("nCols"))
			_nCols = node.getInt("nCols");
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyNumber clone = new LyNumber();

		clone.initFromModel(this);
		clone._template = (LyNumber)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyNumber getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of format.
	 * @return value of format.
	 */
	public String getFormat()
	{
		return _format;
	}

	/**
	 * Gets value of vMin.
	 * @return value of vMin.
	 */
	public double getVMin()
	{
		return _vMin;
	}

	/**
	 * Gets value of vMax.
	 * @return value of vMax.
	 */
	public double getVMax()
	{
		return _vMax;
	}

	/**
	 * Gets value of pitch.
	 * @return value of pitch.
	 */
	public double getPitch()
	{
		return _pitch;
	}

	/**
	 * Gets value of decimals.
	 * @return value of decimals.
	 */
	public int getDecimals()
	{
		return _decimals;
	}

	/**
	 * Checks if value of attribute modulo is set.
	 * @return true if modulo is set, false otherwise.
	 */
	public boolean isModulo()
	{
		return _modulo;
	}

	/**
	 * Gets value of unit.
	 * @return value of unit.
	 */
	public String getUnit()
	{
		return _unit;
	}

	/**
	 * Gets value of bean.
	 * @return value of bean.
	 */
	public String getBean()
	{
		return _bean;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("NUMBER.font"));
	}

	/**
	 * Gets value of enabledColor.
	 * @return value of enabledColor.
	 */
	public String getEnabledColor()
	{
		return _enabledColor;
	}

	/**
	 * Checks whether the enabledColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultEnabledColor()
	{
		return (_enabledColor == getStringValue("NUMBER.enabledColor"));
	}

	/**
	 * Gets value of disabledColor.
	 * @return value of disabledColor.
	 */
	public String getDisabledColor()
	{
		return _disabledColor;
	}

	/**
	 * Checks whether the disabledColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultDisabledColor()
	{
		return (_disabledColor == getStringValue("NUMBER.disabledColor"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("NUMBER.fontColor"));
	}

	/**
	 * Checks if value of attribute arrows is set.
	 * @return true if arrows is set, false otherwise.
	 */
	public boolean isArrows()
	{
		return _arrows;
	}

	/**
	 * Gets value of arrowUp.
	 * @return value of arrowUp.
	 */
	public String getArrowUp()
	{
		return _arrowUp;
	}

	/**
	 * Checks whether the arrowUp attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultArrowUp()
	{
		return (_arrowUp == getStringValue("NUMBER.arrowUp"));
	}

	/**
	 * Gets value of arrowDown.
	 * @return value of arrowDown.
	 */
	public String getArrowDown()
	{
		return _arrowDown;
	}

	/**
	 * Checks whether the arrowDown attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultArrowDown()
	{
		return (_arrowDown == getStringValue("NUMBER.arrowDown"));
	}

	/**
	 * Gets value of nCols.
	 * @return value of nCols.
	 */
	public int getNCols()
	{
		return _nCols;
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyNumber template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView instanceof LySetLabelInterface)
			((LySetLabelInterface)_internalView).setLabel(label);
	}

	/**
	 * Sets attribute format to the given value.
	 * @param format New value of format.
	 */
	public void setFormat(String format)
	{
		if ((format == _format)
			|| ((format != null)
				&& (_format != null)
				&& format.equals(_format)))
			return;

		_format = format;
		if (_internalView instanceof LySetNumberInterface)
			((LySetNumberInterface)_internalView).setFormat(format);
	}

	/**
	 * Sets attribute vMin to the given value.
	 * @param vMin New value of vMin.
	 */
	public void setVMin(double vMin)
	{
		if (vMin == _vMin)
			return;
		_vMin = vMin;
		if (_internalView instanceof LySetNumberInterface)
			((LySetNumberInterface)_internalView).setVMin(vMin);
	}

	/**
	 * Sets attribute vMax to the given value.
	 * @param vMax New value of vMax.
	 */
	public void setVMax(double vMax)
	{
		if (vMax == _vMax)
			return;
		_vMax = vMax;
		if (_internalView instanceof LySetNumberInterface)
			((LySetNumberInterface)_internalView).setVMax(vMax);
	}

	/**
	 * Sets attribute pitch to the given value.
	 * @param pitch New value of pitch.
	 */
	public void setPitch(double pitch)
	{
		if (pitch == _pitch)
			return;
		_pitch = pitch;
		if (_internalView instanceof LySetNumberInterface)
			((LySetNumberInterface)_internalView).setPitch(pitch);
	}

	/**
	 * Sets attribute decimals to the given value.
	 * @param decimals New value of decimals.
	 */
	public void setDecimals(int decimals)
	{
		if (decimals == _decimals)
			return;
		_decimals = decimals;
		if (_internalView instanceof LySetNumberInterface)
			((LySetNumberInterface)_internalView).setDecimals(decimals);
	}

	/**
	 * Initializes attribute modulo to the given value.
	 * @param modulo Boolean indicating if attribute modulo is set or unset.
	 */
	public void setModulo(boolean modulo)
	{
		if (modulo == _modulo)
			return;
		_modulo = modulo;
	}

	/**
	 * Initializes attribute unit to the given value.
	 * @param unit New value of unit.
	 */
	public void setUnit(String unit)
	{
		if ((unit == _unit)
			|| ((unit != null)
				&& (_unit != null)
				&& unit.equals(_unit)))
			return;

		_unit = unit;
	}

	/**
	 * Initializes attribute bean to the given value.
	 * @param bean New value of bean.
	 */
	public void setBean(String bean)
	{
		if ((bean == _bean)
			|| ((bean != null)
				&& (_bean != null)
				&& bean.equals(_bean)))
			return;

		_bean = bean;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute enabledColor to the given value.
	 * @param enabledColor New value of enabledColor.
	 */
	public void setEnabledColor(String enabledColor)
	{
		if ((enabledColor == _enabledColor)
			|| ((enabledColor != null)
				&& (_enabledColor != null)
				&& enabledColor.equals(_enabledColor)))
			return;

		_enabledColor = enabledColor;
	}

	/**
	 * Initializes attribute disabledColor to the given value.
	 * @param disabledColor New value of disabledColor.
	 */
	public void setDisabledColor(String disabledColor)
	{
		if ((disabledColor == _disabledColor)
			|| ((disabledColor != null)
				&& (_disabledColor != null)
				&& disabledColor.equals(_disabledColor)))
			return;

		_disabledColor = disabledColor;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Initializes attribute arrows to the given value.
	 * @param arrows Boolean indicating if attribute arrows is set or unset.
	 */
	public void setArrows(boolean arrows)
	{
		if (arrows == _arrows)
			return;
		_arrows = arrows;
	}

	/**
	 * Initializes attribute arrowUp to the given value.
	 * @param arrowUp New value of arrowUp.
	 */
	public void setArrowUp(String arrowUp)
	{
		if ((arrowUp == _arrowUp)
			|| ((arrowUp != null)
				&& (_arrowUp != null)
				&& arrowUp.equals(_arrowUp)))
			return;

		_arrowUp = arrowUp;
	}

	/**
	 * Initializes attribute arrowDown to the given value.
	 * @param arrowDown New value of arrowDown.
	 */
	public void setArrowDown(String arrowDown)
	{
		if ((arrowDown == _arrowDown)
			|| ((arrowDown != null)
				&& (_arrowDown != null)
				&& arrowDown.equals(_arrowDown)))
			return;

		_arrowDown = arrowDown;
	}

	/**
	 * Initializes attribute nCols to the given value.
	 * @param nCols New value of nCols.
	 */
	public void setNCols(int nCols)
	{
		if (nCols == _nCols)
			return;
		_nCols = nCols;
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<NUMBER");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (((_template == null) && (_format != null))
				|| ((_template != null) && (_format != _template._format)))
			pw.print(" format='" + LyString.getXmlString(_format) + "'");
		if (((_template == null) && (_vMin != getDoubleValue("NUMBER.vMin", Double.NEGATIVE_INFINITY)))
				|| ((_template != null) && (_vMin != _template._vMin)))
			pw.print(" vMin='" + _vMin + "'");
		if (((_template == null) && (_vMax != getDoubleValue("NUMBER.vMax", Double.POSITIVE_INFINITY)))
				|| ((_template != null) && (_vMax != _template._vMax)))
			pw.print(" vMax='" + _vMax + "'");
		if (((_template == null) && (_pitch != getDoubleValue("NUMBER.pitch", 0)))
				|| ((_template != null) && (_pitch != _template._pitch)))
			pw.print(" pitch='" + _pitch + "'");
		if (((_template == null) && (_decimals != getIntValue("NUMBER.decimals", -1)))
				|| ((_template != null) && (_decimals != _template._decimals)))
			pw.print(" decimals='" + _decimals + "'");
		if (((_template == null) && _modulo)
				|| ((_template != null) && (_modulo != _template._modulo)))
			pw.print(" modulo='" + _modulo + "'");
		if (((_template == null) && (_unit != null))
				|| ((_template != null) && (_unit != _template._unit)))
			pw.print(" unit='" + LyString.getXmlString(_unit) + "'");
		if (((_template == null) && (_bean != null))
				|| ((_template != null) && (_bean != _template._bean)))
			pw.print(" bean='" + LyString.getXmlString(_bean) + "'");
		if (((_template == null) && (_font != getStringValue("NUMBER.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_enabledColor != getStringValue("NUMBER.enabledColor", "$ALL.enabledColor")))
				|| ((_template != null) && (_enabledColor != _template._enabledColor)))
			pw.print(" enabledColor='" + LyString.getXmlString(_enabledColor) + "'");
		if (((_template == null) && (_disabledColor != getStringValue("NUMBER.disabledColor", "$ALL.disabledColor")))
				|| ((_template != null) && (_disabledColor != _template._disabledColor)))
			pw.print(" disabledColor='" + LyString.getXmlString(_disabledColor) + "'");
		if (((_template == null) && (_fontColor != getStringValue("NUMBER.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && !_arrows)
				|| ((_template != null) && (_arrows != _template._arrows)))
			pw.print(" arrows='" + _arrows + "'");
		if (((_template == null) && (_arrowUp != getStringValue("NUMBER.arrowUp", "$ALL.arrowUp")))
				|| ((_template != null) && (_arrowUp != _template._arrowUp)))
			pw.print(" arrowUp='" + LyString.getXmlString(_arrowUp) + "'");
		if (((_template == null) && (_arrowDown != getStringValue("NUMBER.arrowDown", "$ALL.arrowDown")))
				|| ((_template != null) && (_arrowDown != _template._arrowDown)))
			pw.print(" arrowDown='" + LyString.getXmlString(_arrowDown) + "'");
		if (((_template == null) && (_nCols != getIntValue("NUMBER.nCols", -1)))
				|| ((_template != null) && (_nCols != _template._nCols)))
			pw.print(" nCols='" + _nCols + "'");
		pw.println("/>");
	}
}
