/*
 * LyPolygon.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.misc.LyString;


/**
 * The LyPolygon class is a generated Java implementation that corresponds to the element 
 * POLYGON of the descriptive grammar of the views (views.dtd).
 * <B>&lt;POLYGON&gt; : Polygon</B>
 * A polygon is a set of points that defines a map region.
 * @see leon.view.LyComponent
 * See also tag <A href="../../conf/dtd/view/POLYGON.html">POLYGON</A> of grammar leon_views.dtd.
 */
public class LyPolygon extends LyComponent
{
	public static final String			__VERSION = "$Id$";


	/* Attributes */

	/**
	 * Filling color of the polygon.
	 */
	protected String			_color;

	/**
	 * Border color of the polygon.
	 */
	protected String			_borderColor;


	/* Sub components */

	/**
	 * Set of points defining the polygon.
	 */
	protected ArrayList<LyPoint>		_points;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_color = getStringValue("POLYGON.color", "$ALL.color");
		_borderColor = getStringValue("POLYGON.borderColor", "$ALL.borderColor");

		// Initialize sub components
		_points = new ArrayList<LyPoint>(1);
		_children = _points;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyPolygon	model = (LyPolygon)component;

		_model = component;
		_color = model._color;
		_borderColor = model._borderColor;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyPoint	points;
		int			i, n;

		n = model._points.size();
		for (i = 0; i < n; i++)
		{
			points = model._points.get(i);
			if (points != null)
				points = (LyPoint)points.cloneComponent();
			addChild(points);
		}
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyPolygon()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyPolygon(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_color = node.getString("color");
		_borderColor = node.getString("borderColor");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyPoint)
			addPoint((LyPoint)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyPolygon clone = new LyPolygon();

		clone.initFromModel(this);
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("POLYGON.color"));
	}

	/**
	 * Gets value of borderColor.
	 * @return value of borderColor.
	 */
	public String getBorderColor()
	{
		return _borderColor;
	}

	/**
	 * Checks whether the borderColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultBorderColor()
	{
		return (_borderColor == getStringValue("POLYGON.borderColor"));
	}

	/**
	 * Get sub-component Point of index i.
	 * @param i the index of the Point to get.
	 * @return sub-component Point of index i.
	 */
	public LyPoint getPoint(int i)
	{
		if ((i < 0) || (i >= _points.size()))
			return null;
		return _points.get(i);
	}

	/**
	 * Get sub-component Point count.
	 * @return sub-component Point count.
	 */
	public int getPointCount()
	{
		return _points.size();
	}


	/* Modifiers */

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute borderColor to the given value.
	 * @param borderColor New value of borderColor.
	 */
	public void setBorderColor(String borderColor)
	{
		if ((borderColor == _borderColor)
			|| ((borderColor != null)
				&& (_borderColor != null)
				&& borderColor.equals(_borderColor)))
			return;

		_borderColor = borderColor;
	}

	/**
	 * Adds a sub-component point.
	 * @param point The point to add to points list.
	 */
	public void addPoint(LyPoint point)
	{
		insertPoint(point, -1);
	}

	/**
	 * Inserts a sub-component point at given position.
	 * @param point The point to add to points list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertPoint(LyPoint point, int pos)
	{
		if (pos >= 0)
			_points.add(pos,point);
		if (point != null)
			point.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<POLYGON");
		if (_color != getStringValue("POLYGON.color", "$ALL.color"))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (_borderColor != getStringValue("POLYGON.borderColor", "$ALL.borderColor"))
			pw.print(" borderColor='" + LyString.getXmlString(_borderColor) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	points;
		int			i, n;

		n = _points.size();
		for (i = 0; i < n; i++)
		{
			points = _points.get(i);
			points.writeXml(pw, debug);
		}


		pw.println("</POLYGON>");
	}
}
