/*
 * LyPopupMenuBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyPopupMenuBase class is a generated Java implementation that corresponds to the element 
 * POPUP_MENU of the descriptive grammar of the views (views.dtd).
 * <B>&lt;POPUP_MENU&gt; : Popup Menu</B>
 * A popup menu is a menu which positioning and content depends on the context that
 * made possible to display it.
 * This context depends, most of the time, of an element belonging to a high-level 
 * graphical object (tree node, element of a map, line of a list,...).
 * If need be, this context is used as context for the launched action.
 * The context-sensitive menus are available only with the Swing and SWT viewers.
 * @see leon.view.LyComponent
 * @see leon.view.LyLyPopupMenu
 * See also tag <A href="../../conf/dtd/view/POPUP_MENU.html">POPUP_MENU</A> of grammar leon_views.dtd.
 */
public abstract class LyPopupMenuBase extends LyComponent
	implements  LyGetIdInterface,
				LyGetControllerInterface,
				LySetControllerInterface,
				LySetEnabledInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'style'. */
	public enum Style
	{
		STANDARD,
		LIGHT;
	}


	/* Static methods */

	/**
	 * Utility method to get style value from a string.
	 * @param token Token string
	 * @return Style value corresponding to this token
	 */
	public static final Style getStyle(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("STANDARD")
				|| token.equals("0"))
			return Style.STANDARD;
		else if (token.equalsIgnoreCase("LIGHT")
				|| token.equals("1"))
			return Style.LIGHT;
		else
			return Style.STANDARD;
	}


	/* Attributes */

	/**
	 * Identifier of the popup menu.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyPopupMenu			_template;

	/**
	 * Reference of the controller that manages the popup menu (i.e. handles user 
	 * demands and refresh content).
	 */
	protected String			_controller;

	/**
	 * Indicates whether the popup menu is enabled or not (true if it is enabled).
	 */
	protected boolean			_enabled;

	/**
	 * Style of the popup menu.
	 * The following styles are managed:
	 * STANDARD: Standard mode
	 * LIGHT: Light mode (items with no label and horizontally displayed)
	 */
	protected Style				_style;

	/**
	 * Position in abscissa of the popup menu.
	 */
	protected int				_x;

	/**
	 * Position en ordinate of the popup menu.
	 */
	protected int				_y;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;


	/* Sub components */

	/**
	 * Components of a popup menu : either menu items, separators or sub-menus. 
	 */
	protected ArrayList<LyComponent>		_components;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_controller = null;
		_enabled = getBooleanValue("POPUP_MENU.enabled", true);
		_style = Style.STANDARD;
		_x = getIntValue("POPUP_MENU.x", 0);
		_y = getIntValue("POPUP_MENU.y", 0);
		_font = getStringValue("POPUP_MENU.font", "$ALL.font");
		_color = getStringValue("POPUP_MENU.color", "$ALL.color");
		_fontColor = getStringValue("POPUP_MENU.fontColor", "$ALL.fontColor");

		// Initialize sub components
		_components = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyPopupMenu	model = (LyPopupMenu)component;

		_model = component;
		_controller = model._controller;
		_enabled = model._enabled;
		_style = model._style;
		_x = model._x;
		_y = model._y;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyComponent	components;
		int			i, n;

		n = (model._components == null ? 0 : model._components.size());
		for (i = 0; i < n; i++)
		{
			components = model._components.get(i);
			if (components != null)
				components = components.cloneComponent();
			addChild(components);
		}
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyPopupMenuBase()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyPopupMenuBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyPopupMenu)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("controller"))
			_controller = node.getString("controller");
		if (node.isSpecified("enabled"))
			_enabled = node.getBoolean("enabled");
		if (node.isSpecified("style"))
			_style = getStyle(node.getString("style"));
		if (node.isSpecified("x"))
			_x = node.getInt("x");
		if (node.isSpecified("y"))
			_y = node.getInt("y");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyItem)
			addItem((LyItem)child);
		else if (child == null)
			addSep();
		else if (child instanceof LyMenu)
			addMenu((LyMenu)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyPopupMenu clone = new LyPopupMenu();

		clone.initFromModel(this);
		clone._template = (LyPopupMenu)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyPopupMenu getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of controller.
	 * @return value of controller.
	 */
	public String getController()
	{
		return _controller;
	}

	/**
	 * Checks if value of attribute enabled is set.
	 * @return true if enabled is set, false otherwise.
	 */
	public boolean isEnabled()
	{
		return _enabled;
	}

	/**
	 * Gets value of style.
	 * @return value of style.
	 */
	public Style getStyle()
	{
		return _style;
	}

	/**
	 * Gets value of x.
	 * @return value of x.
	 */
	public int getX()
	{
		return _x;
	}

	/**
	 * Gets value of y.
	 * @return value of y.
	 */
	public int getY()
	{
		return _y;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("POPUP_MENU.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("POPUP_MENU.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("POPUP_MENU.fontColor"));
	}

	/**
	 * Get sub-component Component of index i.
	 * @param i the index of the Component to get.
	 * @return sub-component Component of index i.
	 */
	public LyComponent getComponent(int i)
	{
		if (_components == null)
			return null;
		if ((i < 0) || (i >= _components.size()))
			return null;
		return _components.get(i);
	}

	/**
	 * Get sub-component Component count.
	 * @return sub-component Component count.
	 */
	public int getComponentCount()
	{
		if (_components == null)
			return 0;
		return _components.size();
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyPopupMenu template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Initializes attribute controller to the given value.
	 * @param controller New value of controller.
	 */
	public void setController(String controller)
	{
		if ((controller == _controller)
			|| ((controller != null)
				&& (_controller != null)
				&& controller.equals(_controller)))
			return;

		_controller = controller;
	}

	/**
	 * Sets attribute enabled to the given value.
	 * @param enabled Boolean indicating if attribute enabled is set or unset.
	 */
	public void setEnabled(boolean enabled)
	{
		if (enabled == _enabled)
			return;
		_enabled = enabled;
		if (_internalView != null)
			((LyGraphicalPopupMenu)_internalView).setEnabled(enabled);
	}

	/**
	 * Initializes attribute style to the given value.
	 * @param style New value of style.
	 * @exception IllegalArgumentException if style parameter is null.
	 */
	public void setStyle(Style style)
	{
		if (style == _style)
			return;
		if (style == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute style");

		_style = style;
	}

	/**
	 * Initializes attribute x to the given value.
	 * @param x New value of x.
	 */
	public void setX(int x)
	{
		if (x == _x)
			return;
		_x = x;
	}

	/**
	 * Initializes attribute y to the given value.
	 * @param y New value of y.
	 */
	public void setY(int y)
	{
		if (y == _y)
			return;
		_y = y;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Adds a sub-component item.
	 * @param item The item to add to components list.
	 */
	public void addItem(LyItem item)
	{
		insertItem(item, -1);
	}

	/**
	 * Inserts a sub-component item at given position.
	 * @param item The item to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertItem(LyItem item, int pos)
	{
		if (_components == null)
		{
			_components = new ArrayList<LyComponent>(1);
			_children = _components;
		}
		else if (_components.contains(item))
			return;
		if (pos >= 0)
			_components.add(pos,item);
		if (item != null)
			item.setParent(this);
		if (_internalView != null)
			((LyGraphicalPopupMenu)_internalView).addComponent(item, pos);
	}

	/**
	 * Removes a sub-component item.
	 * @param item The item to remove to components list.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain item.
	 */
	public void removeItem(LyItem item)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if (!_components.contains(item))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain item");

		if (_internalView != null)
			((LyGraphicalPopupMenu)_internalView).removeComponent(item);
		item.setParent(null);
	}

	/**
	 * Adds a separator.
	 */
	public void addSep()
	{
		insertSep(-1);
	}

	/**
	 * Inserts a separator at the given position.
	 * @param pos	Position where the separator is inserted.
	 */
	public void insertSep(int pos)
	{
		LyItem sep = new LyItem(SEP_ID, null);
		insertItem(sep, pos);
	}

	/**
	 * Removes a separator at the given position.
	 * @param pos	Position where the separator is removed.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain a separator at the given position.
	 */
	public void removeSep(int pos)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if ((pos < 0)
				|| (pos >= _components.size())
				|| ((_components.get(pos) != null)
					&& (_components.get(pos).getId() != SEP_ID)))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain a separator at pos " + pos);

		_components.remove(pos);
	}

	/**
	 * Adds a sub-component menu.
	 * @param menu The menu to add to components list.
	 */
	public void addMenu(LyMenu menu)
	{
		insertMenu(menu, -1);
	}

	/**
	 * Inserts a sub-component menu at given position.
	 * @param menu The menu to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertMenu(LyMenu menu, int pos)
	{
		if (_components == null)
		{
			_components = new ArrayList<LyComponent>(1);
			_children = _components;
		}
		else if (_components.contains(menu))
			return;
		if (pos >= 0)
			_components.add(pos,menu);
		if (menu != null)
			menu.setParent(this);
		if (_internalView != null)
			((LyGraphicalPopupMenu)_internalView).addComponent(menu, pos);
	}

	/**
	 * Removes a sub-component menu.
	 * @param menu The menu to remove to components list.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain menu.
	 */
	public void removeMenu(LyMenu menu)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if (!_components.contains(menu))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain menu");

		if (_internalView != null)
			((LyGraphicalPopupMenu)_internalView).removeComponent(menu);
		menu.setParent(null);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<POPUP_MENU");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlString(_controller) + "'");
		if (((_template == null) && !_enabled)
				|| ((_template != null) && (_enabled != _template._enabled)))
			pw.print(" enabled='" + _enabled + "'");
		if ((_template == null)
			|| (_style != _template._style))
		{
			if ((_style == Style.STANDARD) && (_template != null))
				pw.print(" style='STANDARD'");
			else if (_style == Style.LIGHT)
				pw.print(" style='LIGHT'");
		}
		if (((_template == null) && (_x != getIntValue("POPUP_MENU.x", 0)))
				|| ((_template != null) && (_x != _template._x)))
			pw.print(" x='" + _x + "'");
		if (((_template == null) && (_y != getIntValue("POPUP_MENU.y", 0)))
				|| ((_template != null) && (_y != _template._y)))
			pw.print(" y='" + _y + "'");
		if (((_template == null) && (_font != getStringValue("POPUP_MENU.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("POPUP_MENU.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("POPUP_MENU.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	components;
		int			i, n;

		n = (_components == null ? 0 : _components.size());
		for (i = 0; i < n; i++)
		{
			components = _components.get(i);
			if ((components == null)
					|| (components.getId() == SEP_ID))
				pw.println("<SEP/>");
			else
				components.writeXml(pw, debug);
		}


		pw.println("</POPUP_MENU>");
	}
}
