/*
 * LyRowBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyRowBase class is a generated Java implementation that corresponds to the element 
 * ROW of the descriptive grammar of the views (views.dtd).
 * <B>&lt;ROW&gt; : Table row</B>
 * A row of a list is a set of non-empty cells (or input fields).
 * It is possible to indicate if the row is selected, if an image is assigned to it
 * and if a tooltip is associated to it.
 * @see leon.view.LyComponent
 * @see leon.view.LyLyRow
 * See also tag <A href="../../conf/dtd/view/ROW.html">ROW</A> of grammar leon_views.dtd.
 */
public abstract class LyRowBase extends LyComponent
	implements  LyGetIdInterface,
				LySetImageInterface,
				LyGetLabelInterface,
				LySetLabelInterface,
				LySetHiddenInterface,
				LySetColorInterface
{
	public static final String			__VERSION = "$Id$";


	/* Attributes */

	/**
	 * Identifier of the line.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyRow				_template;

	/**
	 * Indicates if the row is selected (only for the ROW selection type on the list).
	 */
	protected boolean			_selected;

	/**
	 * Indicates if the row is sensitive (a sensitive row may be selected, may have a 
	 * popup menu...).
	 */
	protected boolean			_sensitive;

	/**
	 * Indicates if the row is open (only if the row contains sub-rows).
	 */
	protected boolean			_open;

	/**
	 * Image assigned to the row.
	 * This image is displayed at the beginning of the row in a column that is 
	 * specially dedicated to the images.
	 */
	protected String			_image;

	/**
	 * Label of the row (displayed if row does not contain any cell).
	 */
	protected String			_label;

	/**
	 * ROW.index
	 */
	protected String			_index;

	/**
	 * Overview label of the row.
	 */
	protected String			_tooltip;

	/**
	 * Overview label of the row.
	 */
	protected String			_overview;

	/**
	 * Indicates if the row is hidden.
	 */
	protected boolean			_hidden;

	/**
	 * Height of the border associated to the row in number of pixels.
	 */
	protected int				_borderHeight;

	/**
	 * Color of horizontal border corresponding to the row.
	 */
	protected String			_borderColor;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Color of the row.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;


	/* Sub components */

	/**
	 * Components of a row. They can be table cells, input fields or lists of table 
	 * rows.
	 */
	protected ArrayList<LyComponent>		_components;

	/**
	 * Sub-rows of the row when the table is hierarchical.
	 */
	protected LyRows			_rows;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_selected = getBooleanValue("ROW.selected", false);
		_sensitive = getBooleanValue("ROW.sensitive", true);
		_open = getBooleanValue("ROW.open", false);
		_image = getStringValue("ROW.image", "$ALL.image");
		_label = null;
		_index = null;
		_tooltip = null;
		_overview = null;
		_hidden = getBooleanValue("ROW.hidden", false);
		_borderHeight = getIntValue("ROW.borderHeight", -1);
		_borderColor = getStringValue("ROW.borderColor", "$ALL.borderColor");
		_font = getStringValue("ROW.font", "$ALL.font");
		_color = getStringValue("ROW.color", "$ALL.color");
		_fontColor = getStringValue("ROW.fontColor", "$ALL.fontColor");

		// Initialize sub components
		_components = null;
		_rows = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyRow	model = (LyRow)component;

		_model = component;
		_selected = model._selected;
		_sensitive = model._sensitive;
		_open = model._open;
		_image = model._image;
		_label = model._label;
		_index = model._index;
		_tooltip = model._tooltip;
		_overview = model._overview;
		_hidden = model._hidden;
		_borderHeight = model._borderHeight;
		_borderColor = model._borderColor;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyComponent	components;
		int			i, n;

		n = (model._components == null ? 0 : model._components.size());
		for (i = 0; i < n; i++)
		{
			components = model._components.get(i);
			if (components != null)
				components = components.cloneComponent();
			addChild(components);
		}
		if (model._rows != null)
			addChild(model._rows.cloneComponent());
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyRowBase()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyRowBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyRow)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("selected"))
			_selected = node.getBoolean("selected");
		if (node.isSpecified("sensitive"))
			_sensitive = node.getBoolean("sensitive");
		if (node.isSpecified("open"))
			_open = node.getBoolean("open");
		if (node.isSpecified("image"))
			_image = node.getString("image");
		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("index"))
			_index = node.getString("index");
		if (node.isSpecified("tooltip"))
			_tooltip = node.getString("tooltip");
		if (node.isSpecified("overview"))
			_overview = node.getString("overview");
		if (node.isSpecified("hidden"))
			_hidden = node.getBoolean("hidden");
		if (node.isSpecified("borderHeight"))
			_borderHeight = node.getInt("borderHeight");
		if (node.isSpecified("borderColor"))
			_borderColor = node.getString("borderColor");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyCell)
			addCell((LyCell)child);
		else if (child instanceof LyFormField)
			addFormField((LyFormField)child);
		else if (child instanceof LyToolBar)
			addToolBar((LyToolBar)child);
		else if (child instanceof LyRows)
			setRows((LyRows)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyRow clone = new LyRow();

		clone.initFromModel(this);
		clone._template = (LyRow)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyRow getTemplate()
	{
		return _template;
	}

	/**
	 * Checks if value of attribute selected is set.
	 * @return true if selected is set, false otherwise.
	 */
	public boolean isSelected()
	{
		return _selected;
	}

	/**
	 * Checks if value of attribute sensitive is set.
	 * @return true if sensitive is set, false otherwise.
	 */
	public boolean isSensitive()
	{
		return _sensitive;
	}

	/**
	 * Checks if value of attribute open is set.
	 * @return true if open is set, false otherwise.
	 */
	public boolean isOpen()
	{
		return _open;
	}

	/**
	 * Gets value of image.
	 * @return value of image.
	 */
	public String getImage()
	{
		return _image;
	}

	/**
	 * Checks whether the image attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultImage()
	{
		return (_image == getStringValue("ROW.image"));
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of index.
	 * @return value of index.
	 */
	public String getIndex()
	{
		return _index;
	}

	/**
	 * Gets value of tooltip.
	 * @return value of tooltip.
	 */
	public String getTooltip()
	{
		return _tooltip;
	}

	/**
	 * Gets value of overview.
	 * @return value of overview.
	 */
	public String getOverview()
	{
		return _overview;
	}

	/**
	 * Checks if value of attribute hidden is set.
	 * @return true if hidden is set, false otherwise.
	 */
	public boolean isHidden()
	{
		return _hidden;
	}

	/**
	 * Gets value of borderHeight.
	 * @return value of borderHeight.
	 */
	public int getBorderHeight()
	{
		return _borderHeight;
	}

	/**
	 * Gets value of borderColor.
	 * @return value of borderColor.
	 */
	public String getBorderColor()
	{
		return _borderColor;
	}

	/**
	 * Checks whether the borderColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultBorderColor()
	{
		return (_borderColor == getStringValue("ROW.borderColor"));
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("ROW.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("ROW.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("ROW.fontColor"));
	}

	/**
	 * Get sub-component Component of index i.
	 * @param i the index of the Component to get.
	 * @return sub-component Component of index i.
	 */
	public LyComponent getComponent(int i)
	{
		if (_components == null)
			return null;
		if ((i < 0) || (i >= _components.size()))
			return null;
		return _components.get(i);
	}

	/**
	 * Get sub-component Component count.
	 * @return sub-component Component count.
	 */
	public int getComponentCount()
	{
		if (_components == null)
			return 0;
		return _components.size();
	}

	/**
	 * Get sub-component Rows.
	 * @return sub-component Rows.
	 */
	public LyRows getRows()
	{
		return _rows;
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyRow template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Initializes attribute selected to the given value.
	 * @param selected Boolean indicating if attribute selected is set or unset.
	 */
	public void setSelected(boolean selected)
	{
		if (selected == _selected)
			return;
		_selected = selected;
	}

	/**
	 * Initializes attribute sensitive to the given value.
	 * @param sensitive Boolean indicating if attribute sensitive is set or unset.
	 */
	public void setSensitive(boolean sensitive)
	{
		if (sensitive == _sensitive)
			return;
		_sensitive = sensitive;
	}

	/**
	 * Sets attribute open to the given value.
	 * @param open Boolean indicating if attribute open is set or unset.
	 */
	public void setOpen(boolean open)
	{
		if (open == _open)
			return;
		_open = open;
		if (_internalView != null)
			((LyGraphicalRow)_internalView).setOpen(open);
	}

	/**
	 * Sets attribute image to the given value.
	 * @param image New value of image.
	 */
	public void setImage(String image)
	{
		if ((image == _image)
			|| ((image != null)
				&& (_image != null)
				&& image.equals(_image)))
			return;

		_image = image;
		if (_internalView != null)
			((LyGraphicalRow)_internalView).setImage(image);
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView != null)
			((LyGraphicalRow)_internalView).setLabel(label);
	}

	/**
	 * Sets attribute index to the given value.
	 * @param index New value of index.
	 */
	public void setIndex(String index)
	{
		if ((index == _index)
			|| ((index != null)
				&& (_index != null)
				&& index.equals(_index)))
			return;

		_index = index;
		if (_internalView != null)
			((LyGraphicalRow)_internalView).setIndex(index);
	}

	/**
	 * Sets attribute tooltip to the given value.
	 * @param tooltip New value of tooltip.
	 */
	public void setTooltip(String tooltip)
	{
		if ((tooltip == _tooltip)
			|| ((tooltip != null)
				&& (_tooltip != null)
				&& tooltip.equals(_tooltip)))
			return;

		_tooltip = tooltip;
		if (_internalView != null)
			((LyGraphicalRow)_internalView).setTooltip(tooltip);
	}

	/**
	 * Initializes attribute overview to the given value.
	 * @param overview New value of overview.
	 */
	public void setOverview(String overview)
	{
		if ((overview == _overview)
			|| ((overview != null)
				&& (_overview != null)
				&& overview.equals(_overview)))
			return;

		_overview = overview;
	}

	/**
	 * Sets attribute hidden to the given value.
	 * @param hidden Boolean indicating if attribute hidden is set or unset.
	 */
	public void setHidden(boolean hidden)
	{
		if (hidden == _hidden)
			return;
		_hidden = hidden;
		if (_internalView != null)
			((LyGraphicalRow)_internalView).setHidden(hidden);
	}

	/**
	 * Sets attribute borderHeight to the given value.
	 * @param borderHeight New value of borderHeight.
	 */
	public void setBorderHeight(int borderHeight)
	{
		if (borderHeight == _borderHeight)
			return;
		_borderHeight = borderHeight;
		if (_internalView != null)
			((LyGraphicalRow)_internalView).setBorderHeight(borderHeight);
	}

	/**
	 * Sets attribute borderColor to the given value.
	 * @param borderColor New value of borderColor.
	 */
	public void setBorderColor(String borderColor)
	{
		if ((borderColor == _borderColor)
			|| ((borderColor != null)
				&& (_borderColor != null)
				&& borderColor.equals(_borderColor)))
			return;

		_borderColor = borderColor;
		if (_internalView != null)
			((LyGraphicalRow)_internalView).setBorderColor(borderColor);
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Sets attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
		if (_internalView != null)
			((LyGraphicalRow)_internalView).setColor(color);
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Adds a sub-component cell.
	 * @param cell The cell to add to components list.
	 */
	public void addCell(LyCell cell)
	{
		insertCell(cell, -1);
	}

	/**
	 * Inserts a sub-component cell at given position.
	 * @param cell The cell to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertCell(LyCell cell, int pos)
	{
		if (_components == null)
			_components = new ArrayList<LyComponent>(1);
		else if (_components.contains(cell))
			return;
		if (pos >= 0)
			_components.add(pos,cell);
		else
			_components.add(cell);
		if (cell != null)
			cell.setParent(this);
		if (_internalView != null)
			((LyGraphicalRow)_internalView).addComponent(cell, pos);
	}

	/**
	 * Removes a sub-component cell.
	 * @param cell The cell to remove to components list.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain cell.
	 */
	public void removeCell(LyCell cell)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if (!_components.contains(cell))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain cell");

		if (_internalView != null)
			((LyGraphicalRow)_internalView).removeComponent(cell);
		_components.remove(cell);
		cell.setParent(null);
	}

	/**
	 * Adds a sub-component formField.
	 * @param formField The formField to add to components list.
	 */
	public void addFormField(LyFormField formField)
	{
		insertFormField(formField, -1);
	}

	/**
	 * Inserts a sub-component formField at given position.
	 * @param formField The formField to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertFormField(LyFormField formField, int pos)
	{
		if (_components == null)
			_components = new ArrayList<LyComponent>(1);
		else if (_components.contains(formField))
			return;
		if (pos >= 0)
			_components.add(pos,formField);
		else
			_components.add(formField);
		if (formField != null)
			formField.setParent(this);
		if (_internalView != null)
			((LyGraphicalRow)_internalView).addComponent(formField, pos);
	}

	/**
	 * Removes a sub-component formField.
	 * @param formField The formField to remove to components list.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain formField.
	 */
	public void removeFormField(LyFormField formField)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if (!_components.contains(formField))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain formField");

		if (_internalView != null)
			((LyGraphicalRow)_internalView).removeComponent(formField);
		_components.remove(formField);
		formField.setParent(null);
	}

	/**
	 * Adds a sub-component toolBar.
	 * @param toolBar The toolBar to add to components list.
	 */
	public void addToolBar(LyToolBar toolBar)
	{
		insertToolBar(toolBar, -1);
	}

	/**
	 * Inserts a sub-component toolBar at given position.
	 * @param toolBar The toolBar to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertToolBar(LyToolBar toolBar, int pos)
	{
		if (_components == null)
			_components = new ArrayList<LyComponent>(1);
		else if (_components.contains(toolBar))
			return;
		if (pos >= 0)
			_components.add(pos,toolBar);
		else
			_components.add(toolBar);
		if (toolBar != null)
			toolBar.setParent(this);
		if (_internalView != null)
			((LyGraphicalRow)_internalView).addComponent(toolBar, pos);
	}

	/**
	 * Removes a sub-component toolBar.
	 * @param toolBar The toolBar to remove to components list.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain toolBar.
	 */
	public void removeToolBar(LyToolBar toolBar)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if (!_components.contains(toolBar))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain toolBar");

		if (_internalView != null)
			((LyGraphicalRow)_internalView).removeComponent(toolBar);
		_components.remove(toolBar);
		toolBar.setParent(null);
	}

	/**
	 * Sets sub-component rows.
	 * @param rows New value of rows
	 */
	public void setRows(LyRows rows)
	{
		if (_rows != null)
			_rows.setParent(null);
		_rows = rows;
		if (rows != null)
			rows.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<ROW");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && _selected)
				|| ((_template != null) && (_selected != _template._selected)))
			pw.print(" selected='" + _selected + "'");
		if (((_template == null) && !_sensitive)
				|| ((_template != null) && (_sensitive != _template._sensitive)))
			pw.print(" sensitive='" + _sensitive + "'");
		if (((_template == null) && _open)
				|| ((_template != null) && (_open != _template._open)))
			pw.print(" open='" + _open + "'");
		if (((_template == null) && (_image != getStringValue("ROW.image", "$ALL.image")))
				|| ((_template != null) && (_image != _template._image)))
			pw.print(" image='" + LyString.getXmlString(_image) + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (((_template == null) && (_index != null))
				|| ((_template != null) && (_index != _template._index)))
			pw.print(" index='" + LyString.getXmlString(_index) + "'");
		if (((_template == null) && (_tooltip != null))
				|| ((_template != null) && (_tooltip != _template._tooltip)))
			pw.print(" tooltip='" + LyString.getXmlString(_tooltip) + "'");
		if (((_template == null) && (_overview != null))
				|| ((_template != null) && (_overview != _template._overview)))
			pw.print(" overview='" + LyString.getXmlString(_overview) + "'");
		if (((_template == null) && _hidden)
				|| ((_template != null) && (_hidden != _template._hidden)))
			pw.print(" hidden='" + _hidden + "'");
		if (((_template == null) && (_borderHeight != getIntValue("ROW.borderHeight", -1)))
				|| ((_template != null) && (_borderHeight != _template._borderHeight)))
			pw.print(" borderHeight='" + _borderHeight + "'");
		if (((_template == null) && (_borderColor != getStringValue("ROW.borderColor", "$ALL.borderColor")))
				|| ((_template != null) && (_borderColor != _template._borderColor)))
			pw.print(" borderColor='" + LyString.getXmlString(_borderColor) + "'");
		if (((_template == null) && (_font != getStringValue("ROW.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("ROW.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("ROW.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	components;
		int			i, n;

		n = (_components == null ? 0 : _components.size());
		for (i = 0; i < n; i++)
		{
			components = _components.get(i);
			components.writeXml(pw, debug);
		}

		if (_rows != null)
			_rows.writeXml(pw, debug);

		pw.println("</ROW>");
	}
}
