/*
 * LyRows.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.view;

import leon.misc.*;
import leon.view.setinterface.*;

import java.util.*;


/**
 * This class represents a Rows component.
 *
 * @author   Lyria
 * @version  $Id: LyRows.java 32527 2011-09-28 06:45:53Z jlsaouli $
 * @see      LyComponent
 */
public class LyRows extends LyRowsBase
{
	public static final String __VERSION = "$Id: LyRows.java 32527 2011-09-28 06:45:53Z jlsaouli $";

	// Be sure to avoid having the same row several times
	protected Hashtable<String, LyRow> _rowsById;

	protected boolean _hierarchical = false;

	protected int _hiddenRowCount;

	/**
	 * Constructor with all required attributes.
	 */
	public LyRows()
	{
		super();
	}

	/* Protected methods for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 *
	 * @param  node  node from which the component is constructed.
	 *
	 * @see    LyViewNodeFactory#buildElement
	 */
	protected LyRows(LyViewNode node)
	{
		super(node);
	}

	/**
	 * Initialisation of the data structure.
	 *
	 * @see  leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();
		_rowsById = new Hashtable<String, LyRow>();
		_hierarchical = false;
		_hiddenRowCount = 0;
	}

	/**
	 * Gets parent table.
	 *
	 * @return  parent table (internal view) if instance of LySetTableInterface, or null.
	 */
	public LySetTableInterface getTableInterface()
	{
		LyTable parentTable = getParentTable();

		if (parentTable == null)
			return null;

		if (parentTable._internalView == null)
			return null;

		Object internal = parentTable.getInternalView();

		if (internal instanceof LySetTableInterface)
			return (LySetTableInterface)internal;
		else
			return null;
	}

	/**
	 * Returns the position of the rows component in the parent table component.
	 *
	 * @return  0 if parent table is not hierarchical (only one rows component) or the rows
	 *          component position in the table otherwise.
	 */
	public int getPositionInTable()
	{
		if (_parent == null)
			return 0;
		else if (_parent instanceof LyTable)
			return 0;
		else
		{
			if (_parent instanceof LyRow)
			{
				LyRow	    parent = (LyRow)_parent;
				LyComponent component = _parent.getParent();

				if (component instanceof LyRows)
				{
					// Retrieve position of the parent row in its own parent
					LyRows grandParent = (LyRows)component;
					int    parentIndex = grandParent.getRowIndex(parent);

					parentIndex = grandParent.getSubRowCount(0, parentIndex, false);

					return ((LyRows)component).getPositionInTable() + parentIndex + 1;
				}
			}

			return 0;
		}
	} // end method getPositionInTable

	/**
	 * Return the number of sub-rows, browsing all the tree of sub-rows.
	 *
	 * @param   displayed  indicates if only displayed rows must be considered
	 *
	 * @return  the number of sub-rows
	 */
	public int getSubRowCount(boolean displayed)
	{
		int n = (displayed ? getDisplayedRowCount() : getRowCount());

		return getSubRowCount(0, n, displayed);
	}

	/**
	 * Return the number of sub-rows, browsing all the tree of sub-rows from row at start index
	 * until row at end index (excluded).
	 *
	 * @param   start      index of the first row
	 * @param   end        index of the last row
	 * @param   displayed  indicates if only displayed rows must be considered
	 *
	 * @return  the number of sub-rows
	 */
	protected int getSubRowCount(int start, int end, boolean displayed)
	{
		if (!_hierarchical)
			return end;

		int    i, total = 0;
		LyRow  row;
		LyRows rows;

		for (i = start; i < end; i++)
		{
			row = getRow(i);
			rows = row.getRows();
			total++;

			if ((rows != null) && (row.isOpen() || !displayed))
				total += rows.getSubRowCount(displayed);
		}

		return total;
	}

	/**
	 * Get sub-component displayed Row count.
	 *
	 * @return  sub-component displayed Row count.
	 */
	public int getDisplayedRowCount()
	{
		int n = ((_rows == null) ? 0 : _rows.size());

		return (n - _hiddenRowCount);
	}

	/**
	 * Retrieve all the displayed sub-rows of the rows component and add them to the given list,
	 * browsing all the tree of sub-rows.
	 *
	 * @param  list  the list containing the sub-rows
	 */
	public void getDisplayedSubRows(Vector<LyRow> list)
	{
		int    i, n = getRowCount();
		LyRow  row;
		LyRows rows;

		for (i = 0; i < n; i++)
		{
			row = getRow(i);
			rows = row.getRows();

			if (row.isHidden())
				continue;

			list.addElement(row);

			if (row.isOpen() && (rows != null))
				rows.getDisplayedSubRows(list);
		}
	}

	/**
	 * Indicates if the rows component is hierarchical (if at least one of its children has
	 * children).
	 *
	 * @return  true if the rows component is hierarchical, false otherwise.
	 */
	public boolean isHierarchical()
	{
		return _hierarchical;
	}

	/**
	 * Indicates if the rows component has at least one hidden row.
	 *
	 * @return  true if the rows component has at least one hidden row, false otherwise.
	 */
	public boolean hasHiddenRow()
	{
		return (_hiddenRowCount > 0);
	}

	/**
	 * Gets parent table.
	 *
	 * @return  parent table.
	 */
	public LyTable getParentTable()
	{
		if (_parent == null)
			return null;

		if (_parent instanceof LyTable)
			return (LyTable)_parent;
		// If table is hierarchical, retrieve the table from the parent row
		else if (_parent instanceof LyRow)
		{
			LyComponent component = _parent.getParent();

			if (component instanceof LyRows)
				return ((LyRows)component).getParentTable();
		}

		return null;
	}

	/**
	 * Inserts a sub-component row at given position.
	 *
	 * @param  row  row that must be added.
	 * @param  pos  index where row must be added.
	 */
	public void insertRow(LyRow row, int pos)
	{
		String rowId = row.getId();
		LyRow  oldRow = null;

		synchronized (getLock())
		{
			// Check if the ROW is not here
			if (rowId != null)
			{
				oldRow = _rowsById.get(rowId);
				_rowsById.put(rowId, row);
			}

			super.insertRow(row, pos);

			// AWT MUST be consistent before removing old one
			LySetTableInterface table = getTableInterface();

			if (table != null)
			{
				int rowsPosition = getPositionInTable();

				// If table is hierarchical, compute the new row position
				if (rowsPosition > 0)
				{
					if (pos < 0)
						pos = rowsPosition + getSubRowCount(false) - 1;
					else
						pos = rowsPosition + getSubRowCount(0, pos, false);
				}

				table.addRow(row, pos);
			}

			if (oldRow != null)
			{
				LyEnvironment env = getEnvironment();

				// Should never happen
				if (env != null)
				{
					env.logWarning("Attemp to add a row in a table : "
						+ row.getId()
						+ ". The row already exists : old one is removed.");
					if (env.isDebugLevel(LyEnvironment.DEBUG_TRACE))
						Thread.dumpStack();
				}

				removeRow(oldRow);
				// Warning ! removeRow removes rowId from ids table
				_rowsById.put(rowId, row);

			}

			if (row._hierarchical)
				setHierarchical(true);

			if (row.isHidden())
				addHiddenRow();
		} // end method insertRow
	} // end method insertRow

	/**
	 * Removes a sub-component row.
	 *
	 * @param  row  row that must be removed.
	 */
	public void removeRow(LyRow row)
	{
		synchronized (getLock())
		{
			int index = _rows.indexOf(row);

			removeRow(index);
		}
	}

	/**
	 * Remove row at the given index.
	 *
	 * @param  index  index of the row that must be removed.
	 */
	public void removeRow(int index)
	{
		synchronized (getLock())
		{
			// Remove our row from local id cache
			LyRow oldRow = _rows.get(index);

			// Check if the ROW is not here
			if (oldRow.getId() != null)
				_rowsById.remove(oldRow.getId());

			LySetTableInterface table = getTableInterface();

			if (table != null)
			{
				int rowsPosition = getPositionInTable();
				int tableIndex = getSubRowCount(0, index, false);

				// If table is hierarchical, compute the new row position
				if (rowsPosition > 0)
					tableIndex += rowsPosition;

				// Also remove sub-rows of current rows instance
				if (oldRow.getRows() != null)
				{
					LyRows rows = oldRow.getRows();
					int    i, n = rows.getSubRowCount(false);

					for (i = n; i > 0; i--)
						table.removeRow(tableIndex + i);
				}

				table.removeRow(tableIndex);
			}

			_rows.remove(index);
			// Clean parent cache.
			if (_parent != null)
				_parent.removeChildFromCache(oldRow);

			if (oldRow.isHidden())
				removeHiddenRow();
		}
	} // end method removeRow

	/**
	 * Swap rows at the given indexes.
	 *
	 * @param  index1  index of first row to swap.
	 * @param  index2  index of second row to swap.
	 */
	public void swapRows(int index1, int index2)
	{
		synchronized (getLock())
		{
			LySetTableInterface table = getTableInterface();
			LyRow			    row1 = getRow(index1);
			LyRow			    row2 = getRow(index2);

			_rows.set(index2, row1);
			_rows.set(index1, row2);

			if (table != null)
			{
				int rowsPosition = getPositionInTable();

				// If table is hierarchical, compute the new rows positions
				if (rowsPosition > 0)
				{
					index1 += rowsPosition;
					index2 += rowsPosition;
				}

				table.swapRows(index1, index2);
			}
		}
	}

	/**
	 * Accessor to the index of a row.
	 *
	 * @param   row  row we are looking for the index
	 *
	 * @return  the index of the row or -1 if it could not be found
	 */
	public int getRowIndex(LyRow row)
	{
		if ((_rows == null) || (row == null))
			return -1;

		int		index = _rows.indexOf(row);
		
		if (index >= 0)
			return index;
		
        int     i, n = _rows.size();
        LyRow   subRow = null;
        LyRows  subRows;
        int		total = 0;
        
        for (i = 0; i < n; i++)
        {
            subRow = _rows.get(i);
            subRows = subRow.getRows();
			total++;

            if ((subRows != null)
                && subRow.isOpen())
            {
                index = subRows.getRowIndex(row);
                if (index >= 0)
                    return total + index;
                else
    				total += subRows.getSubRowCount(true);
            }
        }
        
		return -1;
	}

	/**
	 * Retrieves a row by its id.
	 *
	 * @param   id  Id of the row.
	 *
	 * @return  The corresponding row if found, null otherwise.
	 */
	public LyRow getRow(String id)
	{
		return _rowsById.get(id);
	}

	/**
	 * Accessor indicates a hierarchical table.
	 *
	 * @param  hierarchical
	 */
	protected void setHierarchical(boolean hierarchical)
	{
		if (_hierarchical == hierarchical)
			return;

		_hierarchical = hierarchical;
		if (_parent instanceof LyRow)
			((LyRow)_parent).setHierarchical(true);
	}

	/**
	 * Cached value : indicates a new hidden row in the table.
	 */
	protected void addHiddenRow()
	{
		_hiddenRowCount++;
	}

	/**
	 * Cached value : indicates a new hidden row in the table.
	 */
	protected void removeHiddenRow()
	{
		_hiddenRowCount--;
	}

	/**
	 * Get a lock to prevent concurrent access Returns the one provided by the internal view
	 * implementation, otherwise returns this.
	 *
	 * @return  a valid lock for synchronized(), may not return null
	 */
	public Object getLock()
	{
		LyTable parent = getParentTable();

		if (parent != null)
			return parent.getLock();
		else
			return super.getLock();
	}

} // end class LyRows
