/*
 * LySelectBar.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LySelectBar class is a generated Java implementation that corresponds to the element 
 * SELECTBAR of the descriptive grammar of the views (views.dtd).
 * <B>&lt;SELECTBAR&gt; : Navigation bar</B>
 * A navigation bar allows displaying a set of page numbers (or page names) in 
 * order to go directly to the selected page.
 * This bar must be associated to an area presenting paginated data (a long list 
 * for example).
 * @see leon.view.LyPanel
 * See also tag <A href="../../conf/dtd/view/SELECTBAR.html">SELECTBAR</A> of grammar leon_views.dtd.
 */
public class LySelectBar extends LyPanel
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface,
				LyGetControllerInterface,
				LySetControllerInterface,
				LyGetConstraintInterface
{
	public static final String			__VERSION = "$Id$";


	/* Attributes */

	/**
	 * Identifier of the navigation bar.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LySelectBar			_template;

	/**
	 * Displayable label of the navigation bar.
	 */
	protected String			_label;

	/**
	 * Start page
	 */
	protected int				_startItem;

	/**
	 * Number of items viewed in the navigation bar.
	 */
	protected int				_itemCount;

	/**
	 * Reference of the controller that manages this navigation bar (i.e. handles user 
	 * demands and refresh content).
	 */
	protected String			_controller;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * SELECTBAR.hrefColor
	 */
	protected String			_hrefColor;

	/**
	 * SELECTBAR.disabledColor
	 */
	protected String			_disabledColor;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * SELECTBAR.highlightColor
	 */
	protected String			_highlightColor;

	/**
	 * SELECTBAR.visitedColor
	 */
	protected String			_visitedColor;

	/**
	 * SELECTBAR.fontTitle
	 */
	protected String			_fontTitle;

	/**
	 * SELECTBAR.pageFirstImage
	 */
	protected String			_pageFirstImage;

	/**
	 * SELECTBAR.pagePreviousImage
	 */
	protected String			_pagePreviousImage;

	/**
	 * SELECTBAR.pageShowPreviousImage
	 */
	protected String			_pageShowPreviousImage;

	/**
	 * SELECTBAR.pageLastImage
	 */
	protected String			_pageLastImage;

	/**
	 * SELECTBAR.pageNextImage
	 */
	protected String			_pageNextImage;

	/**
	 * SELECTBAR.pageShowNextImage
	 */
	protected String			_pageShowNextImage;


	/* Sub components */

	/**
	 * Components of a navigation bar : either selection items or separators.
	 */
	protected ArrayList<LyComponent>		_components;

	/**
	 * Input field displayed on the right side of a navigation bar.
	 */
	protected LyFormField		_formField;

	/**
	 * Positioning constraints
	 */
	protected LyConstraint		_constraint;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_startItem = getIntValue("SELECTBAR.startItem", 0);
		_itemCount = getIntValue("SELECTBAR.itemCount", 20);
		_controller = null;
		_color = getStringValue("SELECTBAR.color", "$ALL.color");
		_hrefColor = getStringValue("SELECTBAR.hrefColor", "$ALL.hrefColor");
		_disabledColor = getStringValue("SELECTBAR.disabledColor", "$ALL.disabledColor");
		_font = getStringValue("SELECTBAR.font", "$ALL.font");
		_fontColor = getStringValue("SELECTBAR.fontColor", "$ALL.fontColor");
		_highlightColor = getStringValue("SELECTBAR.highlightColor", "$ALL.highlightColor");
		_visitedColor = getStringValue("SELECTBAR.visitedColor", "$ALL.visitedColor");
		_fontTitle = getStringValue("SELECTBAR.fontTitle", "$ALL.fontTitle");
		_pageFirstImage = getStringValue("SELECTBAR.pageFirstImage", "$ALL.pageFirstImage");
		_pagePreviousImage = getStringValue("SELECTBAR.pagePreviousImage", "$ALL.pagePreviousImage");
		_pageShowPreviousImage = getStringValue("SELECTBAR.pageShowPreviousImage", "$ALL.pageShowPreviousImage");
		_pageLastImage = getStringValue("SELECTBAR.pageLastImage", "$ALL.pageLastImage");
		_pageNextImage = getStringValue("SELECTBAR.pageNextImage", "$ALL.pageNextImage");
		_pageShowNextImage = getStringValue("SELECTBAR.pageShowNextImage", "$ALL.pageShowNextImage");

		// Initialize sub components
		_components = null;
		_formField = null;
		_constraint = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LySelectBar	model = (LySelectBar)component;

		_model = component;
		_label = model._label;
		_startItem = model._startItem;
		_itemCount = model._itemCount;
		_controller = model._controller;
		_color = model._color;
		_hrefColor = model._hrefColor;
		_disabledColor = model._disabledColor;
		_font = model._font;
		_fontColor = model._fontColor;
		_highlightColor = model._highlightColor;
		_visitedColor = model._visitedColor;
		_fontTitle = model._fontTitle;
		_pageFirstImage = model._pageFirstImage;
		_pagePreviousImage = model._pagePreviousImage;
		_pageShowPreviousImage = model._pageShowPreviousImage;
		_pageLastImage = model._pageLastImage;
		_pageNextImage = model._pageNextImage;
		_pageShowNextImage = model._pageShowNextImage;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyComponent	components;
		int			i, n;

		n = (model._components == null ? 0 : model._components.size());
		for (i = 0; i < n; i++)
		{
			components = model._components.get(i);
			if (components != null)
				components = components.cloneComponent();
			addChild(components);
		}
		if (model._formField != null)
			addChild(model._formField.cloneComponent());
		if (model._constraint != null)
			addChild(model._constraint.cloneComponent());
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	protected LySelectBar()
	{
		_id = getInternalId();
	}

	/**
	 * Constructor with all required attributes.
	 * @param id Identifier of the navigation bar.
	 */
	public LySelectBar(String id)
	{
		// Fill up required attributes with parameters
		_id = (id != null ? id.intern() : null);
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LySelectBar(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LySelectBar)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("startItem"))
			_startItem = node.getInt("startItem");
		if (node.isSpecified("itemCount"))
			_itemCount = node.getInt("itemCount");
		if (node.isSpecified("controller"))
			_controller = node.getString("controller");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("hrefColor"))
			_hrefColor = node.getString("hrefColor");
		if (node.isSpecified("disabledColor"))
			_disabledColor = node.getString("disabledColor");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("highlightColor"))
			_highlightColor = node.getString("highlightColor");
		if (node.isSpecified("visitedColor"))
			_visitedColor = node.getString("visitedColor");
		if (node.isSpecified("fontTitle"))
			_fontTitle = node.getString("fontTitle");
		if (node.isSpecified("pageFirstImage"))
			_pageFirstImage = node.getString("pageFirstImage");
		if (node.isSpecified("pagePreviousImage"))
			_pagePreviousImage = node.getString("pagePreviousImage");
		if (node.isSpecified("pageShowPreviousImage"))
			_pageShowPreviousImage = node.getString("pageShowPreviousImage");
		if (node.isSpecified("pageLastImage"))
			_pageLastImage = node.getString("pageLastImage");
		if (node.isSpecified("pageNextImage"))
			_pageNextImage = node.getString("pageNextImage");
		if (node.isSpecified("pageShowNextImage"))
			_pageShowNextImage = node.getString("pageShowNextImage");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LySelectItem)
			addSelectItem((LySelectItem)child);
		else if (child == null)
			addSep();
		else if (child instanceof LyFormField)
			setFormField((LyFormField)child);
		else if (child instanceof LyConstraint)
			setConstraint((LyConstraint)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LySelectBar clone = new LySelectBar();

		clone.initFromModel(this);
		clone._template = this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LySelectBar getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of startItem.
	 * @return value of startItem.
	 */
	public int getStartItem()
	{
		return _startItem;
	}

	/**
	 * Gets value of itemCount.
	 * @return value of itemCount.
	 */
	public int getItemCount()
	{
		return _itemCount;
	}

	/**
	 * Gets value of controller.
	 * @return value of controller.
	 */
	public String getController()
	{
		return _controller;
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("SELECTBAR.color"));
	}

	/**
	 * Gets value of hrefColor.
	 * @return value of hrefColor.
	 */
	public String getHrefColor()
	{
		return _hrefColor;
	}

	/**
	 * Checks whether the hrefColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultHrefColor()
	{
		return (_hrefColor == getStringValue("SELECTBAR.hrefColor"));
	}

	/**
	 * Gets value of disabledColor.
	 * @return value of disabledColor.
	 */
	public String getDisabledColor()
	{
		return _disabledColor;
	}

	/**
	 * Checks whether the disabledColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultDisabledColor()
	{
		return (_disabledColor == getStringValue("SELECTBAR.disabledColor"));
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("SELECTBAR.font"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("SELECTBAR.fontColor"));
	}

	/**
	 * Gets value of highlightColor.
	 * @return value of highlightColor.
	 */
	public String getHighlightColor()
	{
		return _highlightColor;
	}

	/**
	 * Checks whether the highlightColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultHighlightColor()
	{
		return (_highlightColor == getStringValue("SELECTBAR.highlightColor"));
	}

	/**
	 * Gets value of visitedColor.
	 * @return value of visitedColor.
	 */
	public String getVisitedColor()
	{
		return _visitedColor;
	}

	/**
	 * Checks whether the visitedColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultVisitedColor()
	{
		return (_visitedColor == getStringValue("SELECTBAR.visitedColor"));
	}

	/**
	 * Gets value of fontTitle.
	 * @return value of fontTitle.
	 */
	public String getFontTitle()
	{
		return _fontTitle;
	}

	/**
	 * Checks whether the fontTitle attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontTitle()
	{
		return (_fontTitle == getStringValue("SELECTBAR.fontTitle"));
	}

	/**
	 * Gets value of pageFirstImage.
	 * @return value of pageFirstImage.
	 */
	public String getPageFirstImage()
	{
		return _pageFirstImage;
	}

	/**
	 * Checks whether the pageFirstImage attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultPageFirstImage()
	{
		return (_pageFirstImage == getStringValue("SELECTBAR.pageFirstImage"));
	}

	/**
	 * Gets value of pagePreviousImage.
	 * @return value of pagePreviousImage.
	 */
	public String getPagePreviousImage()
	{
		return _pagePreviousImage;
	}

	/**
	 * Checks whether the pagePreviousImage attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultPagePreviousImage()
	{
		return (_pagePreviousImage == getStringValue("SELECTBAR.pagePreviousImage"));
	}

	/**
	 * Gets value of pageShowPreviousImage.
	 * @return value of pageShowPreviousImage.
	 */
	public String getPageShowPreviousImage()
	{
		return _pageShowPreviousImage;
	}

	/**
	 * Checks whether the pageShowPreviousImage attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultPageShowPreviousImage()
	{
		return (_pageShowPreviousImage == getStringValue("SELECTBAR.pageShowPreviousImage"));
	}

	/**
	 * Gets value of pageLastImage.
	 * @return value of pageLastImage.
	 */
	public String getPageLastImage()
	{
		return _pageLastImage;
	}

	/**
	 * Checks whether the pageLastImage attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultPageLastImage()
	{
		return (_pageLastImage == getStringValue("SELECTBAR.pageLastImage"));
	}

	/**
	 * Gets value of pageNextImage.
	 * @return value of pageNextImage.
	 */
	public String getPageNextImage()
	{
		return _pageNextImage;
	}

	/**
	 * Checks whether the pageNextImage attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultPageNextImage()
	{
		return (_pageNextImage == getStringValue("SELECTBAR.pageNextImage"));
	}

	/**
	 * Gets value of pageShowNextImage.
	 * @return value of pageShowNextImage.
	 */
	public String getPageShowNextImage()
	{
		return _pageShowNextImage;
	}

	/**
	 * Checks whether the pageShowNextImage attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultPageShowNextImage()
	{
		return (_pageShowNextImage == getStringValue("SELECTBAR.pageShowNextImage"));
	}

	/**
	 * Get sub-component Component of index i.
	 * @param i the index of the Component to get.
	 * @return sub-component Component of index i.
	 */
	public LyComponent getComponent(int i)
	{
		if (_components == null)
			return null;
		if ((i < 0) || (i >= _components.size()))
			return null;
		return _components.get(i);
	}

	/**
	 * Get sub-component Component count.
	 * @return sub-component Component count.
	 */
	public int getComponentCount()
	{
		if (_components == null)
			return 0;
		return _components.size();
	}

	/**
	 * Get sub-component FormField.
	 * @return sub-component FormField.
	 */
	public LyFormField getFormField()
	{
		return _formField;
	}

	/**
	 * Get sub-component Constraint.
	 * @return sub-component Constraint.
	 */
	public LyConstraint getConstraint()
	{
		return _constraint;
	}


	/* Modifiers */

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LySelectBar template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView != null)
			((LyGraphicalSelectBar)_internalView).setLabel(label);
	}

	/**
	 * Sets attribute startItem to the given value.
	 * @param startItem New value of startItem.
	 */
	public void setStartItem(int startItem)
	{
		if (startItem == _startItem)
			return;
		_startItem = startItem;
		if (_internalView != null)
			((LyGraphicalSelectBar)_internalView).setStartItem(startItem);
	}

	/**
	 * Initializes attribute itemCount to the given value.
	 * @param itemCount New value of itemCount.
	 */
	public void setItemCount(int itemCount)
	{
		if (itemCount == _itemCount)
			return;
		_itemCount = itemCount;
	}

	/**
	 * Initializes attribute controller to the given value.
	 * @param controller New value of controller.
	 */
	public void setController(String controller)
	{
		if ((controller == _controller)
			|| ((controller != null)
				&& (_controller != null)
				&& controller.equals(_controller)))
			return;

		_controller = controller;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute hrefColor to the given value.
	 * @param hrefColor New value of hrefColor.
	 */
	public void setHrefColor(String hrefColor)
	{
		if ((hrefColor == _hrefColor)
			|| ((hrefColor != null)
				&& (_hrefColor != null)
				&& hrefColor.equals(_hrefColor)))
			return;

		_hrefColor = hrefColor;
	}

	/**
	 * Initializes attribute disabledColor to the given value.
	 * @param disabledColor New value of disabledColor.
	 */
	public void setDisabledColor(String disabledColor)
	{
		if ((disabledColor == _disabledColor)
			|| ((disabledColor != null)
				&& (_disabledColor != null)
				&& disabledColor.equals(_disabledColor)))
			return;

		_disabledColor = disabledColor;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Initializes attribute highlightColor to the given value.
	 * @param highlightColor New value of highlightColor.
	 */
	public void setHighlightColor(String highlightColor)
	{
		if ((highlightColor == _highlightColor)
			|| ((highlightColor != null)
				&& (_highlightColor != null)
				&& highlightColor.equals(_highlightColor)))
			return;

		_highlightColor = highlightColor;
	}

	/**
	 * Initializes attribute visitedColor to the given value.
	 * @param visitedColor New value of visitedColor.
	 */
	public void setVisitedColor(String visitedColor)
	{
		if ((visitedColor == _visitedColor)
			|| ((visitedColor != null)
				&& (_visitedColor != null)
				&& visitedColor.equals(_visitedColor)))
			return;

		_visitedColor = visitedColor;
	}

	/**
	 * Initializes attribute fontTitle to the given value.
	 * @param fontTitle New value of fontTitle.
	 */
	public void setFontTitle(String fontTitle)
	{
		if ((fontTitle == _fontTitle)
			|| ((fontTitle != null)
				&& (_fontTitle != null)
				&& fontTitle.equals(_fontTitle)))
			return;

		_fontTitle = fontTitle;
	}

	/**
	 * Initializes attribute pageFirstImage to the given value.
	 * @param pageFirstImage New value of pageFirstImage.
	 */
	public void setPageFirstImage(String pageFirstImage)
	{
		if ((pageFirstImage == _pageFirstImage)
			|| ((pageFirstImage != null)
				&& (_pageFirstImage != null)
				&& pageFirstImage.equals(_pageFirstImage)))
			return;

		_pageFirstImage = pageFirstImage;
	}

	/**
	 * Initializes attribute pagePreviousImage to the given value.
	 * @param pagePreviousImage New value of pagePreviousImage.
	 */
	public void setPagePreviousImage(String pagePreviousImage)
	{
		if ((pagePreviousImage == _pagePreviousImage)
			|| ((pagePreviousImage != null)
				&& (_pagePreviousImage != null)
				&& pagePreviousImage.equals(_pagePreviousImage)))
			return;

		_pagePreviousImage = pagePreviousImage;
	}

	/**
	 * Initializes attribute pageShowPreviousImage to the given value.
	 * @param pageShowPreviousImage New value of pageShowPreviousImage.
	 */
	public void setPageShowPreviousImage(String pageShowPreviousImage)
	{
		if ((pageShowPreviousImage == _pageShowPreviousImage)
			|| ((pageShowPreviousImage != null)
				&& (_pageShowPreviousImage != null)
				&& pageShowPreviousImage.equals(_pageShowPreviousImage)))
			return;

		_pageShowPreviousImage = pageShowPreviousImage;
	}

	/**
	 * Initializes attribute pageLastImage to the given value.
	 * @param pageLastImage New value of pageLastImage.
	 */
	public void setPageLastImage(String pageLastImage)
	{
		if ((pageLastImage == _pageLastImage)
			|| ((pageLastImage != null)
				&& (_pageLastImage != null)
				&& pageLastImage.equals(_pageLastImage)))
			return;

		_pageLastImage = pageLastImage;
	}

	/**
	 * Initializes attribute pageNextImage to the given value.
	 * @param pageNextImage New value of pageNextImage.
	 */
	public void setPageNextImage(String pageNextImage)
	{
		if ((pageNextImage == _pageNextImage)
			|| ((pageNextImage != null)
				&& (_pageNextImage != null)
				&& pageNextImage.equals(_pageNextImage)))
			return;

		_pageNextImage = pageNextImage;
	}

	/**
	 * Initializes attribute pageShowNextImage to the given value.
	 * @param pageShowNextImage New value of pageShowNextImage.
	 */
	public void setPageShowNextImage(String pageShowNextImage)
	{
		if ((pageShowNextImage == _pageShowNextImage)
			|| ((pageShowNextImage != null)
				&& (_pageShowNextImage != null)
				&& pageShowNextImage.equals(_pageShowNextImage)))
			return;

		_pageShowNextImage = pageShowNextImage;
	}

	/**
	 * Adds a sub-component selectItem.
	 * @param selectItem The selectItem to add to components list.
	 */
	public void addSelectItem(LySelectItem selectItem)
	{
		insertSelectItem(selectItem, -1);
	}

	/**
	 * Inserts a sub-component selectItem at given position.
	 * @param selectItem The selectItem to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertSelectItem(LySelectItem selectItem, int pos)
	{
		if (_components == null)
			_components = new ArrayList<LyComponent>(1);
		else if (_components.contains(selectItem))
			return;
		if (pos >= 0)
			_components.add(pos,selectItem);
		else
			_components.add(selectItem);
		if (selectItem != null)
			selectItem.setParent(this);
		if (_internalView != null)
			((LyGraphicalSelectBar)_internalView).addComponent(selectItem, pos);
	}

	/**
	 * Removes a sub-component selectItem.
	 * @param selectItem The selectItem to remove to components list.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain selectItem.
	 */
	public void removeSelectItem(LySelectItem selectItem)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if (!_components.contains(selectItem))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain selectItem");

		if (_internalView != null)
			((LyGraphicalSelectBar)_internalView).removeComponent(selectItem);
		_components.remove(selectItem);
		selectItem.setParent(null);
	}

	/**
	 * Adds a separator.
	 */
	public void addSep()
	{
		insertSep(-1);
	}

	/**
	 * Inserts a separator at the given position.
	 * @param pos	Position where the separator is inserted.
	 */
	public void insertSep(int pos)
	{
		LySelectItem sep = new LySelectItem(SEP_ID, null);
		insertSelectItem(sep, pos);
	}

	/**
	 * Removes a separator at the given position.
	 * @param pos	Position where the separator is removed.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain a separator at the given position.
	 */
	public void removeSep(int pos)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if ((pos < 0)
				|| (pos >= _components.size())
				|| ((_components.get(pos) != null)
					&& (_components.get(pos).getId() != SEP_ID)))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain a separator at pos " + pos);

		_components.remove(pos);
	}

	/**
	 * Sets sub-component formField.
	 * @param formField New value of formField
	 */
	public void setFormField(LyFormField formField)
	{
		if (_formField != null)
			_formField.setParent(null);
		_formField = formField;
		if (formField != null)
			formField.setParent(this);
	}

	/**
	 * Sets sub-component constraint.
	 * @param constraint New value of constraint
	 */
	public void setConstraint(LyConstraint constraint)
	{
		if (_constraint != null)
			_constraint.setParent(null);
		_constraint = constraint;
		if (constraint != null)
			constraint.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<SELECTBAR");
		pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (((_template == null) && (_startItem != getIntValue("SELECTBAR.startItem", 0)))
				|| ((_template != null) && (_startItem != _template._startItem)))
			pw.print(" startItem='" + _startItem + "'");
		if (((_template == null) && (_itemCount != getIntValue("SELECTBAR.itemCount", 20)))
				|| ((_template != null) && (_itemCount != _template._itemCount)))
			pw.print(" itemCount='" + _itemCount + "'");
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlString(_controller) + "'");
		if (((_template == null) && (_color != getStringValue("SELECTBAR.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_hrefColor != getStringValue("SELECTBAR.hrefColor", "$ALL.hrefColor")))
				|| ((_template != null) && (_hrefColor != _template._hrefColor)))
			pw.print(" hrefColor='" + LyString.getXmlString(_hrefColor) + "'");
		if (((_template == null) && (_disabledColor != getStringValue("SELECTBAR.disabledColor", "$ALL.disabledColor")))
				|| ((_template != null) && (_disabledColor != _template._disabledColor)))
			pw.print(" disabledColor='" + LyString.getXmlString(_disabledColor) + "'");
		if (((_template == null) && (_font != getStringValue("SELECTBAR.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_fontColor != getStringValue("SELECTBAR.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && (_highlightColor != getStringValue("SELECTBAR.highlightColor", "$ALL.highlightColor")))
				|| ((_template != null) && (_highlightColor != _template._highlightColor)))
			pw.print(" highlightColor='" + LyString.getXmlString(_highlightColor) + "'");
		if (((_template == null) && (_visitedColor != getStringValue("SELECTBAR.visitedColor", "$ALL.visitedColor")))
				|| ((_template != null) && (_visitedColor != _template._visitedColor)))
			pw.print(" visitedColor='" + LyString.getXmlString(_visitedColor) + "'");
		if (((_template == null) && (_fontTitle != getStringValue("SELECTBAR.fontTitle", "$ALL.fontTitle")))
				|| ((_template != null) && (_fontTitle != _template._fontTitle)))
			pw.print(" fontTitle='" + LyString.getXmlString(_fontTitle) + "'");
		if (((_template == null) && (_pageFirstImage != getStringValue("SELECTBAR.pageFirstImage", "$ALL.pageFirstImage")))
				|| ((_template != null) && (_pageFirstImage != _template._pageFirstImage)))
			pw.print(" pageFirstImage='" + LyString.getXmlString(_pageFirstImage) + "'");
		if (((_template == null) && (_pagePreviousImage != getStringValue("SELECTBAR.pagePreviousImage", "$ALL.pagePreviousImage")))
				|| ((_template != null) && (_pagePreviousImage != _template._pagePreviousImage)))
			pw.print(" pagePreviousImage='" + LyString.getXmlString(_pagePreviousImage) + "'");
		if (((_template == null) && (_pageShowPreviousImage != getStringValue("SELECTBAR.pageShowPreviousImage", "$ALL.pageShowPreviousImage")))
				|| ((_template != null) && (_pageShowPreviousImage != _template._pageShowPreviousImage)))
			pw.print(" pageShowPreviousImage='" + LyString.getXmlString(_pageShowPreviousImage) + "'");
		if (((_template == null) && (_pageLastImage != getStringValue("SELECTBAR.pageLastImage", "$ALL.pageLastImage")))
				|| ((_template != null) && (_pageLastImage != _template._pageLastImage)))
			pw.print(" pageLastImage='" + LyString.getXmlString(_pageLastImage) + "'");
		if (((_template == null) && (_pageNextImage != getStringValue("SELECTBAR.pageNextImage", "$ALL.pageNextImage")))
				|| ((_template != null) && (_pageNextImage != _template._pageNextImage)))
			pw.print(" pageNextImage='" + LyString.getXmlString(_pageNextImage) + "'");
		if (((_template == null) && (_pageShowNextImage != getStringValue("SELECTBAR.pageShowNextImage", "$ALL.pageShowNextImage")))
				|| ((_template != null) && (_pageShowNextImage != _template._pageShowNextImage)))
			pw.print(" pageShowNextImage='" + LyString.getXmlString(_pageShowNextImage) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	components;
		int			i, n;

		n = (_components == null ? 0 : _components.size());
		for (i = 0; i < n; i++)
		{
			components = _components.get(i);
			if ((components == null)
					|| (components.getId() == SEP_ID))
				pw.println("<SEP/>");
			else
				components.writeXml(pw, debug);
		}

		if (_formField != null)
			_formField.writeXml(pw, debug);
		if (_constraint != null)
			_constraint.writeXml(pw, debug);

		pw.println("</SELECTBAR>");
	}
}
