/*
 * LySplitter.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.misc.LyString;


/**
 * The LySplitter class is a generated Java implementation that corresponds to the element 
 * SPLITTER of the descriptive grammar of the views (views.dtd).
 * <B>&lt;SPLITTER&gt; : Panels splitter</B>
 * A container of panel type with splitter allows displaying different panels that 
 * are splitted by vertical or horizontal separation bars.
 * These bars can be moved so as to allocating more or less place to each of the 
 * panels.
 * The minimum number of panels that is required is two panels.
 * There is no constraint on the maximum number of panels allowed in such a 
 * container. Including such containers in containers of same type allows getting 
 * windows combining vertical and horizontal separators.
 * @see leon.view.LyContainer
 * See also tag <A href="../../conf/dtd/view/SPLITTER.html">SPLITTER</A> of grammar leon_views.dtd.
 */
public class LySplitter extends LyContainer
	implements  LyGetIdInterface,
				LyGetConstraintInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'style'. */
	public enum Style
	{
		HORIZONTAL,
		VERTICAL;
	}


	/** Possible values for attribute 'activePart'. */
	public enum ActivePart
	{
		AUTO,
		FIRST,
		SECOND;
	}


	/* Static methods */

	/**
	 * Utility method to get style value from a string.
	 * @param token Token string
	 * @return Style value corresponding to this token
	 */
	public static final Style getStyle(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("HORIZONTAL")
				|| token.equals("0"))
			return Style.HORIZONTAL;
		else if (token.equalsIgnoreCase("VERTICAL")
				|| token.equals("1"))
			return Style.VERTICAL;
		else
			return Style.HORIZONTAL;
	}

	/**
	 * Utility method to get activePart value from a string.
	 * @param token Token string
	 * @return ActivePart value corresponding to this token
	 */
	public static final ActivePart getActivePart(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("AUTO")
				|| token.equals("0"))
			return ActivePart.AUTO;
		else if (token.equalsIgnoreCase("FIRST")
				|| token.equals("1"))
			return ActivePart.FIRST;
		else if (token.equalsIgnoreCase("SECOND")
				|| token.equals("2"))
			return ActivePart.SECOND;
		else
			return ActivePart.FIRST;
	}


	/* Attributes */

	/**
	 * Identifier of the component.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LySplitter			_template;

	/**
	 * Indicates whether the panels should be placed horizontally or vertically.
	 * HORIZONTAL: Indicates that the panels are placed horizontally and that the 
	 * separation bar is vertical.
	 * VERTICAL: Indicates that the panels are placed vertically and that the 
	 * separation bar is horizontal.
	 */
	protected Style				_style;

	/**
	 * SPLITTER.activePart
	 */
	protected ActivePart			_activePart;

	/**
	 * Indicates the proportional size of each panel inside the container.
	 * The expected values are positive integer values separated by blanks.
	 * The umpteenth provided value corresponds to the umpteenth panel inserted in the 
	 * container.
	 * The value 0 corresponds to the null size for a given panel.
	 * A negative value indicates that the place occupied by the concerned panel should
	 * be a fixed size and should correspond to the default size of the panel.
	 * If a non integer value (examples: 'tkj', '1.34', '3%',...) is specified, this 
	 * value is equivalent to the value 0. If the number of sizes is greater than the 
	 * total number of panels, the additional sizes are ignored.
	 * If the number
	 * of sizes is lower than the total number of panels, the missing size are replaced
	 * by the value 10.
	 */
	protected String			_sizes;

	/**
	 * Indicates if the separators can be moved.
	 */
	protected boolean			_resizable;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Tooltip of the button that hides or shows the first part of the splitter.
	 */
	protected String			_firstTooltip;

	/**
	 * Tooltip of the button that hides or shows the second part of the splitter.
	 */
	protected String			_secondTooltip;


	/* Sub components */

	/**
	 * List of additional panels of the container.
	 * At least one is needed in order to have at least two panels.
	 */
	protected ArrayList<LyPanel>		_panels;

	/**
	 * Positioning constraint of the panels splitter.
	 */
	protected LyConstraint		_constraint;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_style = Style.HORIZONTAL;
		_activePart = ActivePart.FIRST;
		_sizes = null;
		_resizable = getBooleanValue("SPLITTER.resizable", true);
		_color = getStringValue("SPLITTER.color", "$ALL.color");
		_firstTooltip = null;
		_secondTooltip = null;

		// Initialize sub components
		_panels = null;
		_constraint = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LySplitter	model = (LySplitter)component;

		_model = component;
		_style = model._style;
		_activePart = model._activePart;
		_sizes = model._sizes;
		_resizable = model._resizable;
		_color = model._color;
		_firstTooltip = model._firstTooltip;
		_secondTooltip = model._secondTooltip;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyPanel	panels;
		int			i, n;

		n = (model._panels == null ? 0 : model._panels.size());
		for (i = 0; i < n; i++)
		{
			panels = model._panels.get(i);
			if (panels != null)
				panels = (LyPanel)panels.cloneComponent();
			addChild(panels);
		}
		if (model._constraint != null)
			addChild(model._constraint.cloneComponent());
		if (model._datas != null)
			_datas = new Hashtable<String, String>(model._datas);
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LySplitter()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LySplitter(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LySplitter)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("style"))
			_style = getStyle(node.getString("style"));
		if (node.isSpecified("activePart"))
			_activePart = getActivePart(node.getString("activePart"));
		if (node.isSpecified("sizes"))
			_sizes = node.getString("sizes");
		if (node.isSpecified("resizable"))
			_resizable = node.getBoolean("resizable");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("firstTooltip"))
			_firstTooltip = node.getString("firstTooltip");
		if (node.isSpecified("secondTooltip"))
			_secondTooltip = node.getString("secondTooltip");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyPanel)
			addPanel((LyPanel)child);
		else if (child instanceof LyConstraint)
			setConstraint((LyConstraint)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LySplitter clone = new LySplitter();

		clone.initFromModel(this);
		clone._template = this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LySplitter getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of style.
	 * @return value of style.
	 */
	public Style getStyle()
	{
		return _style;
	}

	/**
	 * Gets value of activePart.
	 * @return value of activePart.
	 */
	public ActivePart getActivePart()
	{
		return _activePart;
	}

	/**
	 * Gets value of sizes.
	 * @return value of sizes.
	 */
	public String getSizes()
	{
		return _sizes;
	}

	/**
	 * Checks if value of attribute resizable is set.
	 * @return true if resizable is set, false otherwise.
	 */
	public boolean isResizable()
	{
		return _resizable;
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("SPLITTER.color"));
	}

	/**
	 * Gets value of firstTooltip.
	 * @return value of firstTooltip.
	 */
	public String getFirstTooltip()
	{
		return _firstTooltip;
	}

	/**
	 * Gets value of secondTooltip.
	 * @return value of secondTooltip.
	 */
	public String getSecondTooltip()
	{
		return _secondTooltip;
	}

	/**
	 * Get sub-component Panel of index i.
	 * @param i the index of the Panel to get.
	 * @return sub-component Panel of index i.
	 */
	public LyPanel getPanel(int i)
	{
		if (_panels == null)
			return null;
		if ((i < 0) || (i >= _panels.size()))
			return null;
		return _panels.get(i);
	}

	/**
	 * Get sub-component Panel count.
	 * @return sub-component Panel count.
	 */
	public int getPanelCount()
	{
		if (_panels == null)
			return 0;
		return _panels.size();
	}

	/**
	 * Get sub-component Constraint.
	 * @return sub-component Constraint.
	 */
	public LyConstraint getConstraint()
	{
		return _constraint;
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LySplitter template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Initializes attribute style to the given value.
	 * @param style New value of style.
	 * @exception IllegalArgumentException if style parameter is null.
	 */
	public void setStyle(Style style)
	{
		if (style == _style)
			return;
		if (style == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute style");

		_style = style;
	}

	/**
	 * Sets attribute activePart to the given value.
	 * @param activePart New value of activePart.
	 * @exception IllegalArgumentException if activePart parameter is null.
	 */
	public void setActivePart(ActivePart activePart)
	{
		if (activePart == _activePart)
			return;
		if (activePart == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute activePart");

		_activePart = activePart;
		if (_internalView != null)
			((LyGraphicalSplitter)_internalView).setActivePart(activePart);
	}

	/**
	 * Sets attribute sizes to the given value.
	 * @param sizes New value of sizes.
	 */
	public void setSizes(String sizes)
	{
		if ((sizes == _sizes)
			|| ((sizes != null)
				&& (_sizes != null)
				&& sizes.equals(_sizes)))
			return;

		_sizes = sizes;
		if (_internalView != null)
			((LyGraphicalSplitter)_internalView).setSizes(sizes);
	}

	/**
	 * Initializes attribute resizable to the given value.
	 * @param resizable Boolean indicating if attribute resizable is set or unset.
	 */
	public void setResizable(boolean resizable)
	{
		if (resizable == _resizable)
			return;
		_resizable = resizable;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute firstTooltip to the given value.
	 * @param firstTooltip New value of firstTooltip.
	 */
	public void setFirstTooltip(String firstTooltip)
	{
		if ((firstTooltip == _firstTooltip)
			|| ((firstTooltip != null)
				&& (_firstTooltip != null)
				&& firstTooltip.equals(_firstTooltip)))
			return;

		_firstTooltip = firstTooltip;
	}

	/**
	 * Initializes attribute secondTooltip to the given value.
	 * @param secondTooltip New value of secondTooltip.
	 */
	public void setSecondTooltip(String secondTooltip)
	{
		if ((secondTooltip == _secondTooltip)
			|| ((secondTooltip != null)
				&& (_secondTooltip != null)
				&& secondTooltip.equals(_secondTooltip)))
			return;

		_secondTooltip = secondTooltip;
	}

	/**
	 * Adds a sub-component panel.
	 * @param panel The panel to add to panels list.
	 */
	public void addPanel(LyPanel panel)
	{
		insertPanel(panel, -1);
	}

	/**
	 * Inserts a sub-component panel at given position.
	 * @param panel The panel to add to panels list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertPanel(LyPanel panel, int pos)
	{
		if (_panels == null)
			_panels = new ArrayList<LyPanel>(1);
		else if (_panels.contains(panel))
			return;
		if (pos >= 0)
			_panels.add(pos,panel);
		else
			_panels.add(panel);
		if (panel != null)
			panel.setParent(this);
	}

	/**
	 * Sets sub-component constraint.
	 * @param constraint New value of constraint
	 */
	public void setConstraint(LyConstraint constraint)
	{
		if (_constraint != null)
			_constraint.setParent(null);
		_constraint = constraint;
		if (constraint != null)
			constraint.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<SPLITTER");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if ((_template == null)
			|| (_style != _template._style))
		{
			if ((_style == Style.HORIZONTAL) && (_template != null))
				pw.print(" style='HORIZONTAL'");
			else if (_style == Style.VERTICAL)
				pw.print(" style='VERTICAL'");
		}
		if ((_template == null)
			|| (_activePart != _template._activePart))
		{
			if (_activePart == ActivePart.AUTO)
				pw.print(" activePart='AUTO'");
			else if ((_activePart == ActivePart.FIRST) && (_template != null))
				pw.print(" activePart='FIRST'");
			else if (_activePart == ActivePart.SECOND)
				pw.print(" activePart='SECOND'");
		}
		if (((_template == null) && (_sizes != null))
				|| ((_template != null) && (_sizes != _template._sizes)))
			pw.print(" sizes='" + LyString.getXmlString(_sizes) + "'");
		if (((_template == null) && !_resizable)
				|| ((_template != null) && (_resizable != _template._resizable)))
			pw.print(" resizable='" + _resizable + "'");
		if (((_template == null) && (_color != getStringValue("SPLITTER.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_firstTooltip != null))
				|| ((_template != null) && (_firstTooltip != _template._firstTooltip)))
			pw.print(" firstTooltip='" + LyString.getXmlString(_firstTooltip) + "'");
		if (((_template == null) && (_secondTooltip != null))
				|| ((_template != null) && (_secondTooltip != _template._secondTooltip)))
			pw.print(" secondTooltip='" + LyString.getXmlString(_secondTooltip) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	panels;
		int			i, n;

		n = (_panels == null ? 0 : _panels.size());
		for (i = 0; i < n; i++)
		{
			panels = _panels.get(i);
			panels.writeXml(pw, debug);
		}

		if (_constraint != null)
			_constraint.writeXml(pw, debug);

		writeDatas(pw);

		pw.println("</SPLITTER>");
	}
}
