/*
 * LyTable.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyTable class is a generated Java implementation that corresponds to the element 
 * TABLE of the descriptive grammar of the views (views.dtd).
 * <B>&lt;TABLE&gt; : Table</B>
 * A table allows displaying a set of rows and columns containing a set of cells.
 * It can contain a header providing information about the content of the cells of 
 * the table.
 * It is possible to define its selection modes (simple, extended, multiple or 
 * none) and finally to indicate if horizontal and/or vertical separation lines 
 * should be drawn.
 * @see leon.view.LyPanel
 * See also tag <A href="../../conf/dtd/view/TABLE.html">TABLE</A> of grammar leon_views.dtd.
 */
public class LyTable extends LyPanel
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LyGetControllerInterface,
				LySetControllerInterface,
				LySetEnabledInterface,
				LyGetConstraintInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'selectionMode'. */
	public enum SelectionMode
	{
		NONE,
		SINGLE,
		MULTIPLE,
		SINGLE_CHECKBOX,
		MULTIPLE_CHECKBOX;
	}


	/* Static methods */

	/**
	 * Utility method to get selectionMode value from a string.
	 * @param token Token string
	 * @return SelectionMode value corresponding to this token
	 */
	public static final SelectionMode getSelectionMode(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("NONE")
				|| token.equals("0"))
			return SelectionMode.NONE;
		else if (token.equalsIgnoreCase("SINGLE")
				|| token.equals("1"))
			return SelectionMode.SINGLE;
		else if (token.equalsIgnoreCase("MULTIPLE")
				|| token.equals("2"))
			return SelectionMode.MULTIPLE;
		else if (token.equalsIgnoreCase("SINGLE_CHECKBOX")
				|| token.equals("3"))
			return SelectionMode.SINGLE_CHECKBOX;
		else if (token.equalsIgnoreCase("MULTIPLE_CHECKBOX")
				|| token.equals("4"))
			return SelectionMode.MULTIPLE_CHECKBOX;
		else
			return SelectionMode.MULTIPLE;
	}


	/* Attributes */

	/**
	 * Identifier of the table.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyTable			_template;

	/**
	 * Label of the table
	 */
	protected String			_label;

	/**
	 * Reference of the controller that manages the table (i.e. handles user demands 
	 * and refresh content).
	 */
	protected String			_controller;

	/**
	 * Selection mode of the elements of the table.
	 * The elements that can be selected are defined by the type of selection (ROW, 
	 * COLUMN, CELL). The following modes are managed:
	 * NONE: No selection is possible
	 * SINGLE: Selection of only one element at a time
	 * MULTIPLE: Selection of several elements at a time
	 * SINGLE_CHECKBOX: Selection of only one element at a time with a checkbox
	 * MULTIPLE_CHECKBOX: Selection of several elements at a time with a checkbox
	 */
	protected SelectionMode			_selectionMode;

	/**
	 * Indicates if horizontal separation lines should be drawn.
	 */
	protected boolean			_hLines;

	/**
	 * Indicates if vertical separation lines should be drawn
	 */
	protected boolean			_vLines;

	/**
	 * Indicates if the table is the context of another element.
	 */
	protected boolean			_isContext;

	/**
	 * Indicates whether the table is enabled or not.
	 */
	protected boolean			_enabled;

	/**
	 * Gives the number of lines on which the table is presented.
	 */
	protected int				_nRows;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * Color used to draw the table grid.
	 */
	protected String			_gridColor;

	/**
	 * Color used to display selected rows.
	 */
	protected String			_selectedColor;

	/**
	 * Font color of the selected rows.
	 */
	protected String			_selectedFontColor;

	/**
	 * Text of the tooltip for the table.
	 */
	protected String			_tooltip;


	/* Sub components */

	/**
	 * Header of the table.
	 */
	protected LyHeaders			_headers;

	/**
	 * Lines of the table.
	 */
	protected LyRows			_rows;

	/**
	 * Footer of the table.
	 */
	protected LyFooters			_footers;

	/**
	 * Positioning constraint of the table.
	 */
	protected LyConstraint		_constraint;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_controller = null;
		_selectionMode = SelectionMode.MULTIPLE;
		_hLines = getBooleanValue("TABLE.hLines", false);
		_vLines = getBooleanValue("TABLE.vLines", false);
		_isContext = getBooleanValue("TABLE.isContext", false);
		_enabled = getBooleanValue("TABLE.enabled", true);
		_nRows = getIntValue("TABLE.nRows", -1);
		_font = getStringValue("TABLE.font", "$ALL.font");
		_color = getStringValue("TABLE.color", "$ALL.color");
		_fontColor = getStringValue("TABLE.fontColor", "$ALL.fontColor");
		_gridColor = getStringValue("TABLE.gridColor", "$ALL.gridColor");
		_selectedColor = getStringValue("TABLE.selectedColor", "$ALL.selectedColor");
		_selectedFontColor = getStringValue("TABLE.selectedFontColor", "$ALL.selectedFontColor");
		_tooltip = null;

		// Initialize sub components
		_headers = null;
		_rows = null;
		_footers = null;
		_constraint = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyTable	model = (LyTable)component;

		_model = component;
		_label = model._label;
		_controller = model._controller;
		_selectionMode = model._selectionMode;
		_hLines = model._hLines;
		_vLines = model._vLines;
		_isContext = model._isContext;
		_enabled = model._enabled;
		_nRows = model._nRows;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;
		_gridColor = model._gridColor;
		_selectedColor = model._selectedColor;
		_selectedFontColor = model._selectedFontColor;
		_tooltip = model._tooltip;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		if (model._headers != null)
			addChild(model._headers.cloneComponent());
		if (model._rows != null)
			addChild(model._rows.cloneComponent());
		if (model._footers != null)
			addChild(model._footers.cloneComponent());
		if (model._constraint != null)
			addChild(model._constraint.cloneComponent());
		if (model._datas != null)
			_datas = new Hashtable<String, String>(model._datas);
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	protected LyTable()
	{
		_id = getInternalId();
	}

	/**
	 * Constructor with all required attributes.
	 * @param id Identifier of the table.
	 */
	public LyTable(String id)
	{
		// Fill up required attributes with parameters
		_id = (id != null ? id.intern() : null);
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyTable(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyTable)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("controller"))
			_controller = node.getString("controller");
		if (node.isSpecified("selectionMode"))
			_selectionMode = getSelectionMode(node.getString("selectionMode"));
		if (node.isSpecified("hLines"))
			_hLines = node.getBoolean("hLines");
		if (node.isSpecified("vLines"))
			_vLines = node.getBoolean("vLines");
		if (node.isSpecified("isContext"))
			_isContext = node.getBoolean("isContext");
		if (node.isSpecified("enabled"))
			_enabled = node.getBoolean("enabled");
		if (node.isSpecified("nRows"))
			_nRows = node.getInt("nRows");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("gridColor"))
			_gridColor = node.getString("gridColor");
		if (node.isSpecified("selectedColor"))
			_selectedColor = node.getString("selectedColor");
		if (node.isSpecified("selectedFontColor"))
			_selectedFontColor = node.getString("selectedFontColor");
		if (node.isSpecified("tooltip"))
			_tooltip = node.getString("tooltip");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyHeaders)
			setHeaders((LyHeaders)child);
		else if (child instanceof LyRows)
			setRows((LyRows)child);
		else if (child instanceof LyFooters)
			setFooters((LyFooters)child);
		else if (child instanceof LyConstraint)
			setConstraint((LyConstraint)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyTable clone = new LyTable();

		clone.initFromModel(this);
		clone._template = this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyTable getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of controller.
	 * @return value of controller.
	 */
	public String getController()
	{
		return _controller;
	}

	/**
	 * Gets value of selectionMode.
	 * @return value of selectionMode.
	 */
	public SelectionMode getSelectionMode()
	{
		return _selectionMode;
	}

	/**
	 * Checks if value of attribute hLines is set.
	 * @return true if hLines is set, false otherwise.
	 */
	public boolean isHLines()
	{
		return _hLines;
	}

	/**
	 * Checks if value of attribute vLines is set.
	 * @return true if vLines is set, false otherwise.
	 */
	public boolean isVLines()
	{
		return _vLines;
	}

	/**
	 * Checks if value of attribute isContext is set.
	 * @return true if isContext is set, false otherwise.
	 */
	public boolean isIsContext()
	{
		return _isContext;
	}

	/**
	 * Checks if value of attribute enabled is set.
	 * @return true if enabled is set, false otherwise.
	 */
	public boolean isEnabled()
	{
		return _enabled;
	}

	/**
	 * Gets value of nRows.
	 * @return value of nRows.
	 */
	public int getNRows()
	{
		return _nRows;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("TABLE.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("TABLE.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("TABLE.fontColor"));
	}

	/**
	 * Gets value of gridColor.
	 * @return value of gridColor.
	 */
	public String getGridColor()
	{
		return _gridColor;
	}

	/**
	 * Checks whether the gridColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultGridColor()
	{
		return (_gridColor == getStringValue("TABLE.gridColor"));
	}

	/**
	 * Gets value of selectedColor.
	 * @return value of selectedColor.
	 */
	public String getSelectedColor()
	{
		return _selectedColor;
	}

	/**
	 * Checks whether the selectedColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultSelectedColor()
	{
		return (_selectedColor == getStringValue("TABLE.selectedColor"));
	}

	/**
	 * Gets value of selectedFontColor.
	 * @return value of selectedFontColor.
	 */
	public String getSelectedFontColor()
	{
		return _selectedFontColor;
	}

	/**
	 * Checks whether the selectedFontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultSelectedFontColor()
	{
		return (_selectedFontColor == getStringValue("TABLE.selectedFontColor"));
	}

	/**
	 * Gets value of tooltip.
	 * @return value of tooltip.
	 */
	public String getTooltip()
	{
		return _tooltip;
	}

	/**
	 * Get sub-component Headers.
	 * @return sub-component Headers.
	 */
	public LyHeaders getHeaders()
	{
		return _headers;
	}

	/**
	 * Get sub-component Rows.
	 * @return sub-component Rows.
	 */
	public LyRows getRows()
	{
		return _rows;
	}

	/**
	 * Get sub-component Footers.
	 * @return sub-component Footers.
	 */
	public LyFooters getFooters()
	{
		return _footers;
	}

	/**
	 * Get sub-component Constraint.
	 * @return sub-component Constraint.
	 */
	public LyConstraint getConstraint()
	{
		return _constraint;
	}


	/* Modifiers */

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyTable template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Initializes attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
	}

	/**
	 * Initializes attribute controller to the given value.
	 * @param controller New value of controller.
	 */
	public void setController(String controller)
	{
		if ((controller == _controller)
			|| ((controller != null)
				&& (_controller != null)
				&& controller.equals(_controller)))
			return;

		_controller = controller;
	}

	/**
	 * Initializes attribute selectionMode to the given value.
	 * @param selectionMode New value of selectionMode.
	 * @exception IllegalArgumentException if selectionMode parameter is null.
	 */
	public void setSelectionMode(SelectionMode selectionMode)
	{
		if (selectionMode == _selectionMode)
			return;
		if (selectionMode == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute selectionMode");

		_selectionMode = selectionMode;
	}

	/**
	 * Initializes attribute hLines to the given value.
	 * @param hLines Boolean indicating if attribute hLines is set or unset.
	 */
	public void setHLines(boolean hLines)
	{
		if (hLines == _hLines)
			return;
		_hLines = hLines;
	}

	/**
	 * Initializes attribute vLines to the given value.
	 * @param vLines Boolean indicating if attribute vLines is set or unset.
	 */
	public void setVLines(boolean vLines)
	{
		if (vLines == _vLines)
			return;
		_vLines = vLines;
	}

	/**
	 * Initializes attribute isContext to the given value.
	 * @param isContext Boolean indicating if attribute isContext is set or unset.
	 */
	public void setIsContext(boolean isContext)
	{
		if (isContext == _isContext)
			return;
		_isContext = isContext;
	}

	/**
	 * Sets attribute enabled to the given value.
	 * @param enabled Boolean indicating if attribute enabled is set or unset.
	 */
	public void setEnabled(boolean enabled)
	{
		if (enabled == _enabled)
			return;
		_enabled = enabled;
		if (_internalView != null)
			((LyGraphicalTable)_internalView).setEnabled(enabled);
	}

	/**
	 * Sets attribute nRows to the given value.
	 * @param nRows New value of nRows.
	 */
	public void setNRows(int nRows)
	{
		if (nRows == _nRows)
			return;
		_nRows = nRows;
		if (_internalView != null)
			((LyGraphicalTable)_internalView).setNRows(nRows);
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Initializes attribute gridColor to the given value.
	 * @param gridColor New value of gridColor.
	 */
	public void setGridColor(String gridColor)
	{
		if ((gridColor == _gridColor)
			|| ((gridColor != null)
				&& (_gridColor != null)
				&& gridColor.equals(_gridColor)))
			return;

		_gridColor = gridColor;
	}

	/**
	 * Initializes attribute selectedColor to the given value.
	 * @param selectedColor New value of selectedColor.
	 */
	public void setSelectedColor(String selectedColor)
	{
		if ((selectedColor == _selectedColor)
			|| ((selectedColor != null)
				&& (_selectedColor != null)
				&& selectedColor.equals(_selectedColor)))
			return;

		_selectedColor = selectedColor;
	}

	/**
	 * Initializes attribute selectedFontColor to the given value.
	 * @param selectedFontColor New value of selectedFontColor.
	 */
	public void setSelectedFontColor(String selectedFontColor)
	{
		if ((selectedFontColor == _selectedFontColor)
			|| ((selectedFontColor != null)
				&& (_selectedFontColor != null)
				&& selectedFontColor.equals(_selectedFontColor)))
			return;

		_selectedFontColor = selectedFontColor;
	}

	/**
	 * Initializes attribute tooltip to the given value.
	 * @param tooltip New value of tooltip.
	 */
	public void setTooltip(String tooltip)
	{
		if ((tooltip == _tooltip)
			|| ((tooltip != null)
				&& (_tooltip != null)
				&& tooltip.equals(_tooltip)))
			return;

		_tooltip = tooltip;
	}

	/**
	 * Sets sub-component headers.
	 * @param headers New value of headers
	 */
	public void setHeaders(LyHeaders headers)
	{
		if (_headers != null)
			_headers.setParent(null);
		_headers = headers;
		if (headers != null)
			headers.setParent(this);
		if (_internalView != null)
			((LyGraphicalTable)_internalView).setHeaders(headers);
	}

	/**
	 * Sets sub-component rows.
	 * @param rows New value of rows
	 */
	public void setRows(LyRows rows)
	{
		if (_rows != null)
			_rows.setParent(null);
		_rows = rows;
		if (rows != null)
			rows.setParent(this);
	}

	/**
	 * Sets sub-component footers.
	 * @param footers New value of footers
	 */
	public void setFooters(LyFooters footers)
	{
		if (_footers != null)
			_footers.setParent(null);
		_footers = footers;
		if (footers != null)
			footers.setParent(this);
		if (_internalView != null)
			((LyGraphicalTable)_internalView).setFooters(footers);
	}

	/**
	 * Sets sub-component constraint.
	 * @param constraint New value of constraint
	 */
	public void setConstraint(LyConstraint constraint)
	{
		if (_constraint != null)
			_constraint.setParent(null);
		_constraint = constraint;
		if (constraint != null)
			constraint.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<TABLE");
		pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlString(_controller) + "'");
		if ((_template == null)
			|| (_selectionMode != _template._selectionMode))
		{
			if (_selectionMode == SelectionMode.NONE)
				pw.print(" selectionMode='NONE'");
			else if (_selectionMode == SelectionMode.SINGLE)
				pw.print(" selectionMode='SINGLE'");
			else if ((_selectionMode == SelectionMode.MULTIPLE) && (_template != null))
				pw.print(" selectionMode='MULTIPLE'");
			else if (_selectionMode == SelectionMode.SINGLE_CHECKBOX)
				pw.print(" selectionMode='SINGLE_CHECKBOX'");
			else if (_selectionMode == SelectionMode.MULTIPLE_CHECKBOX)
				pw.print(" selectionMode='MULTIPLE_CHECKBOX'");
		}
		if (((_template == null) && _hLines)
				|| ((_template != null) && (_hLines != _template._hLines)))
			pw.print(" hLines='" + _hLines + "'");
		if (((_template == null) && _vLines)
				|| ((_template != null) && (_vLines != _template._vLines)))
			pw.print(" vLines='" + _vLines + "'");
		if (((_template == null) && _isContext)
				|| ((_template != null) && (_isContext != _template._isContext)))
			pw.print(" isContext='" + _isContext + "'");
		if (((_template == null) && !_enabled)
				|| ((_template != null) && (_enabled != _template._enabled)))
			pw.print(" enabled='" + _enabled + "'");
		if (((_template == null) && (_nRows != getIntValue("TABLE.nRows", -1)))
				|| ((_template != null) && (_nRows != _template._nRows)))
			pw.print(" nRows='" + _nRows + "'");
		if (((_template == null) && (_font != getStringValue("TABLE.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("TABLE.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("TABLE.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && (_gridColor != getStringValue("TABLE.gridColor", "$ALL.gridColor")))
				|| ((_template != null) && (_gridColor != _template._gridColor)))
			pw.print(" gridColor='" + LyString.getXmlString(_gridColor) + "'");
		if (((_template == null) && (_selectedColor != getStringValue("TABLE.selectedColor", "$ALL.selectedColor")))
				|| ((_template != null) && (_selectedColor != _template._selectedColor)))
			pw.print(" selectedColor='" + LyString.getXmlString(_selectedColor) + "'");
		if (((_template == null) && (_selectedFontColor != getStringValue("TABLE.selectedFontColor", "$ALL.selectedFontColor")))
				|| ((_template != null) && (_selectedFontColor != _template._selectedFontColor)))
			pw.print(" selectedFontColor='" + LyString.getXmlString(_selectedFontColor) + "'");
		if (((_template == null) && (_tooltip != null))
				|| ((_template != null) && (_tooltip != _template._tooltip)))
			pw.print(" tooltip='" + LyString.getXmlString(_tooltip) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		if (_headers != null)
			_headers.writeXml(pw, debug);
		if (_rows != null)
			_rows.writeXml(pw, debug);
		if (_footers != null)
			_footers.writeXml(pw, debug);
		if (_constraint != null)
			_constraint.writeXml(pw, debug);

		writeDatas(pw);

		pw.println("</TABLE>");
	}
}
