/*
 * LyTabs.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.misc.LyString;


/**
 * The LyTabs class is a generated Java implementation that corresponds to the element 
 * TABS of the descriptive grammar of the views (views.dtd).
 * <B>&lt;TABS&gt; : Tabs list</B>
 * A container of tab type allows displaying different panels in the same graphical
 * area, only one panel being visible at once.
 * Each panel is actually contained in a tab, the container being made up of a set 
 * of tabs.
 * The displaying that corresponds to a container of tab type must present two 
 * areas: a first one that shows names of the different tabs and a second one that 
 * allows displaying the panels.
 * @see leon.view.LyContainer
 * See also tag <A href="../../conf/dtd/view/TABS.html">TABS</A> of grammar leon_views.dtd.
 */
public class LyTabs extends LyContainer
	implements  LyGetIdInterface,
				LyGetConstraintInterface
{
	public static final String			__VERSION = "$Id$";


	/* Attributes */

	/**
	 * Identifier of the component.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyTabs			_template;

	/**
	 * Identifier of the selected tab.
	 */
	protected String			_selectedTab;

	/**
	 * Indicates whether tabs are closable or not.
	 */
	protected boolean			_closable;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * Color of the selected tab.
	 */
	protected String			_selectedColor;


	/* Sub components */

	/**
	 * List of tabs.
	 */
	protected ArrayList<LyTab>		_tabs;

	/**
	 * Positioning constraint of the tabs list.
	 */
	protected LyConstraint		_constraint;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_selectedTab = null;
		_closable = getBooleanValue("TABS.closable", false);
		_font = getStringValue("TABS.font", "$ALL.font");
		_color = getStringValue("TABS.color", "$ALL.color");
		_fontColor = getStringValue("TABS.fontColor", "$ALL.fontColor");
		_selectedColor = getStringValue("TABS.selectedColor", "$ALL.selectedColor");

		// Initialize sub components
		_tabs = null;
		_constraint = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyTabs	model = (LyTabs)component;

		_model = component;
		_selectedTab = model._selectedTab;
		_closable = model._closable;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;
		_selectedColor = model._selectedColor;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyTab	tabs;
		int			i, n;

		n = (model._tabs == null ? 0 : model._tabs.size());
		for (i = 0; i < n; i++)
		{
			tabs = model._tabs.get(i);
			if (tabs != null)
				tabs = (LyTab)tabs.cloneComponent();
			addChild(tabs);
		}
		if (model._constraint != null)
			addChild(model._constraint.cloneComponent());
		if (model._datas != null)
			_datas = new Hashtable<String, String>(model._datas);
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyTabs()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyTabs(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyTabs)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("selectedTab"))
			_selectedTab = node.getString("selectedTab");
		if (node.isSpecified("closable"))
			_closable = node.getBoolean("closable");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("selectedColor"))
			_selectedColor = node.getString("selectedColor");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyTab)
			addTab((LyTab)child);
		else if (child instanceof LyConstraint)
			setConstraint((LyConstraint)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyTabs clone = new LyTabs();

		clone.initFromModel(this);
		clone._template = this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyTabs getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of selectedTab.
	 * @return value of selectedTab.
	 */
	public String getSelectedTab()
	{
		return _selectedTab;
	}

	/**
	 * Checks if value of attribute closable is set.
	 * @return true if closable is set, false otherwise.
	 */
	public boolean isClosable()
	{
		return _closable;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("TABS.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("TABS.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("TABS.fontColor"));
	}

	/**
	 * Gets value of selectedColor.
	 * @return value of selectedColor.
	 */
	public String getSelectedColor()
	{
		return _selectedColor;
	}

	/**
	 * Checks whether the selectedColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultSelectedColor()
	{
		return (_selectedColor == getStringValue("TABS.selectedColor"));
	}

	/**
	 * Get sub-component Tab of index i.
	 * @param i the index of the Tab to get.
	 * @return sub-component Tab of index i.
	 */
	public LyTab getTab(int i)
	{
		if (_tabs == null)
			return null;
		if ((i < 0) || (i >= _tabs.size()))
			return null;
		return _tabs.get(i);
	}

	/**
	 * Get sub-component Tab count.
	 * @return sub-component Tab count.
	 */
	public int getTabCount()
	{
		if (_tabs == null)
			return 0;
		return _tabs.size();
	}

	/**
	 * Get sub-component Constraint.
	 * @return sub-component Constraint.
	 */
	public LyConstraint getConstraint()
	{
		return _constraint;
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyTabs template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute selectedTab to the given value.
	 * @param selectedTab New value of selectedTab.
	 */
	public void setSelectedTab(String selectedTab)
	{
		if ((selectedTab == _selectedTab)
			|| ((selectedTab != null)
				&& (_selectedTab != null)
				&& selectedTab.equals(_selectedTab)))
			return;

		_selectedTab = selectedTab;
		if (_internalView != null)
			((LyGraphicalTabs)_internalView).setSelectedTab(selectedTab);
	}

	/**
	 * Initializes attribute closable to the given value.
	 * @param closable Boolean indicating if attribute closable is set or unset.
	 */
	public void setClosable(boolean closable)
	{
		if (closable == _closable)
			return;
		_closable = closable;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Initializes attribute selectedColor to the given value.
	 * @param selectedColor New value of selectedColor.
	 */
	public void setSelectedColor(String selectedColor)
	{
		if ((selectedColor == _selectedColor)
			|| ((selectedColor != null)
				&& (_selectedColor != null)
				&& selectedColor.equals(_selectedColor)))
			return;

		_selectedColor = selectedColor;
	}

	/**
	 * Adds a sub-component tab.
	 * @param tab The tab to add to tabs list.
	 */
	public void addTab(LyTab tab)
	{
		insertTab(tab, -1);
	}

	/**
	 * Inserts a sub-component tab at given position.
	 * @param tab The tab to add to tabs list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertTab(LyTab tab, int pos)
	{
		if (_tabs == null)
			_tabs = new ArrayList<LyTab>(1);
		else if (_tabs.contains(tab))
			return;
		if (pos >= 0)
			_tabs.add(pos,tab);
		else
			_tabs.add(tab);
		if (tab != null)
			tab.setParent(this);
		if (_internalView != null)
			((LyGraphicalTabs)_internalView).addTab(tab, pos);
	}

	/**
	 * Removes a sub-component tab.
	 * @param tab The tab to remove to tabs list.
	 * @exception IllegalStateException if tabs list is empty.
	 * @exception IllegalArgumentException if tabs list doesn't contain tab.
	 */
	public void removeTab(LyTab tab)
	{
		if (_tabs == null)
			throw new IllegalStateException("BF AppEngine : tabs list is empty");
		if (!_tabs.contains(tab))
			throw new IllegalArgumentException("BF AppEngine : tabs list doesn't contain tab");

		if (_internalView != null)
			((LyGraphicalTabs)_internalView).removeTab(tab);
		_tabs.remove(tab);
		tab.setParent(null);
	}

	/**
	 * Sets sub-component constraint.
	 * @param constraint New value of constraint
	 */
	public void setConstraint(LyConstraint constraint)
	{
		if (_constraint != null)
			_constraint.setParent(null);
		_constraint = constraint;
		if (constraint != null)
			constraint.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<TABS");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_selectedTab != null))
				|| ((_template != null) && (_selectedTab != _template._selectedTab)))
			pw.print(" selectedTab='" + LyString.getXmlString(_selectedTab) + "'");
		if (((_template == null) && _closable)
				|| ((_template != null) && (_closable != _template._closable)))
			pw.print(" closable='" + _closable + "'");
		if (((_template == null) && (_font != getStringValue("TABS.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("TABS.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("TABS.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && (_selectedColor != getStringValue("TABS.selectedColor", "$ALL.selectedColor")))
				|| ((_template != null) && (_selectedColor != _template._selectedColor)))
			pw.print(" selectedColor='" + LyString.getXmlString(_selectedColor) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	tabs;
		int			i, n;

		n = (_tabs == null ? 0 : _tabs.size());
		for (i = 0; i < n; i++)
		{
			tabs = _tabs.get(i);
			tabs.writeXml(pw, debug);
		}

		if (_constraint != null)
			_constraint.writeXml(pw, debug);

		writeDatas(pw);

		pw.println("</TABS>");
	}
}
