/*
 * LyTextBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyTextBase class is a generated Java implementation that corresponds to the element 
 * TEXT of the descriptive grammar of the views (views.dtd).
 * <B>&lt;TEXT&gt; : Input Field of Text Type</B>
 * An input field of Text type allows displaying a graphical input or text 
 * consultation component.
 * It does not comprise any element.
 * It is possible to specify if the text should be hidden, if it should respect a 
 * given format, if it can be divided into several lines and the maximum number of 
 * characters that it can contain.
 * @see leon.view.LyFieldItem
 * @see leon.view.LyLyText
 * See also tag <A href="../../conf/dtd/view/TEXT.html">TEXT</A> of grammar leon_views.dtd.
 */
public abstract class LyTextBase extends LyFieldItem
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'textAlign'. */
	public enum TextAlign
	{
		LEFT,
		CENTER,
		RIGHT;
	}


	/** Possible values for attribute 'type'. */
	public enum Type
	{
		STANDARD,
		RICH;
	}


	/** Possible values for attribute 'scroll'. */
	public enum Scroll
	{
		TOP,
		BOTTOM;
	}


	/* Static methods */

	/**
	 * Utility method to get textAlign value from a string.
	 * @param token Token string
	 * @return TextAlign value corresponding to this token
	 */
	public static final TextAlign getTextAlign(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("LEFT")
				|| token.equals("0"))
			return TextAlign.LEFT;
		else if (token.equalsIgnoreCase("CENTER")
				|| token.equals("1"))
			return TextAlign.CENTER;
		else if (token.equalsIgnoreCase("RIGHT")
				|| token.equals("2"))
			return TextAlign.RIGHT;
		else
			return TextAlign.LEFT;
	}

	/**
	 * Utility method to get type value from a string.
	 * @param token Token string
	 * @return Type value corresponding to this token
	 */
	public static final Type getType(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("STANDARD")
				|| token.equals("0"))
			return Type.STANDARD;
		else if (token.equalsIgnoreCase("RICH")
				|| token.equals("1"))
			return Type.RICH;
		else
			return Type.STANDARD;
	}

	/**
	 * Utility method to get scroll value from a string.
	 * @param token Token string
	 * @return Scroll value corresponding to this token
	 */
	public static final Scroll getScroll(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("TOP")
				|| token.equals("0"))
			return Scroll.TOP;
		else if (token.equalsIgnoreCase("BOTTOM")
				|| token.equals("1"))
			return Scroll.BOTTOM;
		else
			return Scroll.TOP;
	}


	/* Attributes */

	/**
	 * Identifier of the component.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyText			_template;

	/**
	 * Value displayed for the text.
	 */
	protected String			_label;

	/**
	 * Indicates whether the input field should be hidden or not.
	 * If the text should be hidden, each character of the text is replaced by an 
	 * asterisk (' * ').
	 */
	protected boolean			_secret;

	/**
	 * Number of columns of the graphical component representing the input field. The 
	 * expected value is a positive integer value.
	 * A negative or null value lets the viewer free to choose the number of columns.
	 * Any other value is equivalent to the null value. This field does not limit the 
	 * input of the text.
	 * Moreover, it can be overstepped when using constraints.
	 */
	protected int				_nCols;

	/**
	 * Number of lines of the graphical component representing the input field. The 
	 * expected value is a positive integer value.
	 * A negative or null value lets the viewer free to choose the number of de lines.
	 * Any other value is equivalent to the null value. This attribute does not limit 
	 * the input of the text.
	 * Moreover, it can be overstepped when using constraints.
	 */
	protected int				_nRows;

	/**
	 * Maximum number of characters allowed for the text. The expected value is a 
	 * positive integer value.
	 * A negative or null value implies that there is no limitation.
	 * Any other value is equivalent to the null value.
	 */
	protected int				_nChars;

	/**
	 * Input format of the text. The expected format is a string containing fixed and 
	 * numeric parts.
	 * The numeric parts are defined by an interval of integer values.
	 * They allow enabling only the input of numbers which values are comprised in the 
	 * interval corresponding to their position. Example of format for an IP address, 
	 * for which it is allowed to input 4 series of numbers comprised between 0 and 
	 * 255, and delimited by dots: "[0-255].[0-255].[0-255].[0-255]".
	 */
	protected String			_format;

	/**
	 * Indicates whether the text is editable or not.
	 */
	protected boolean			_editable;

	/**
	 * Indicates whether the text can be wrapped on several lines or if it should 
	 * remain on a single line. The true value indicates that the text can be wrapped 
	 * on several lines.
	 */
	protected boolean			_wrap;

	/**
	 * Indicates whether the modification of the text value must be notified or not 
	 * (even before focus lost). The true value indicates that the modification of the 
	 * text value must be notified.
	 */
	protected boolean			_notifyValueChanged;

	/**
	 * List of values used for auto-completion. Values must be separated by 
	 * LyTools.VALUE_SEP.
	 */
	protected String			_autoCompletionValues;

	/**
	 * Allows specifying the name of a Java class that replaces the default graphical 
	 * component.
	 */
	protected String			_bean;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * TEXT.enabledColor
	 */
	protected String			_enabledColor;

	/**
	 * TEXT.disabledColor
	 */
	protected String			_disabledColor;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * Text alignment of the label of the component (left, right or centered).
	 */
	protected TextAlign			_textAlign;

	/**
	 * Indicates if arrow buttons are displayed when text has a format.
	 */
	protected boolean			_arrows;

	/**
	 * TEXT.arrowUp
	 */
	protected String			_arrowUp;

	/**
	 * TEXT.arrowDown
	 */
	protected String			_arrowDown;

	/**
	 * Indicates if a rich text editor must be used to edit this text (web viewer only)
	 */
	protected Type				_type;

	/**
	 * Indicates for a multiple text the initial position of vertical scrollbar (web 
	 * viewer only).
	 */
	protected Scroll			_scroll;


	/* Sub components */

	/**
	 * Set of characters allowed in the text field.
	 */
	protected ArrayList<LyCharType>		_charTypes;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_secret = getBooleanValue("TEXT.secret", false);
		_nCols = getIntValue("TEXT.nCols", -1);
		_nRows = getIntValue("TEXT.nRows", -1);
		_nChars = getIntValue("TEXT.nChars", -1);
		_format = null;
		_editable = getBooleanValue("TEXT.editable", true);
		_wrap = getBooleanValue("TEXT.wrap", true);
		_notifyValueChanged = getBooleanValue("TEXT.notifyValueChanged", false);
		_autoCompletionValues = null;
		_bean = null;
		_font = getStringValue("TEXT.font", "$ALL.font");
		_enabledColor = getStringValue("TEXT.enabledColor", "$ALL.enabledColor");
		_disabledColor = getStringValue("TEXT.disabledColor", "$ALL.disabledColor");
		_fontColor = getStringValue("TEXT.fontColor", "$ALL.fontColor");
		_textAlign = TextAlign.LEFT;
		_arrows = getBooleanValue("TEXT.arrows", true);
		_arrowUp = getStringValue("TEXT.arrowUp", "$ALL.arrowUp");
		_arrowDown = getStringValue("TEXT.arrowDown", "$ALL.arrowDown");
		_type = Type.STANDARD;
		_scroll = Scroll.TOP;

		// Initialize sub components
		_charTypes = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyText	model = (LyText)component;

		_model = component;
		_label = model._label;
		_secret = model._secret;
		_nCols = model._nCols;
		_nRows = model._nRows;
		_nChars = model._nChars;
		_format = model._format;
		_editable = model._editable;
		_wrap = model._wrap;
		_notifyValueChanged = model._notifyValueChanged;
		_autoCompletionValues = model._autoCompletionValues;
		_bean = model._bean;
		_font = model._font;
		_enabledColor = model._enabledColor;
		_disabledColor = model._disabledColor;
		_fontColor = model._fontColor;
		_textAlign = model._textAlign;
		_arrows = model._arrows;
		_arrowUp = model._arrowUp;
		_arrowDown = model._arrowDown;
		_type = model._type;
		_scroll = model._scroll;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyCharType	charTypes;
		int			i, n;

		n = (model._charTypes == null ? 0 : model._charTypes.size());
		for (i = 0; i < n; i++)
		{
			charTypes = model._charTypes.get(i);
			if (charTypes != null)
				charTypes = (LyCharType)charTypes.cloneComponent();
			addChild(charTypes);
		}
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyTextBase()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyTextBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyText)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("secret"))
			_secret = node.getBoolean("secret");
		if (node.isSpecified("nCols"))
			_nCols = node.getInt("nCols");
		if (node.isSpecified("nRows"))
			_nRows = node.getInt("nRows");
		if (node.isSpecified("nChars"))
			_nChars = node.getInt("nChars");
		if (node.isSpecified("format"))
			_format = node.getString("format");
		if (node.isSpecified("editable"))
			_editable = node.getBoolean("editable");
		if (node.isSpecified("wrap"))
			_wrap = node.getBoolean("wrap");
		if (node.isSpecified("notifyValueChanged"))
			_notifyValueChanged = node.getBoolean("notifyValueChanged");
		if (node.isSpecified("autoCompletionValues"))
			_autoCompletionValues = node.getString("autoCompletionValues");
		if (node.isSpecified("bean"))
			_bean = node.getString("bean");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("enabledColor"))
			_enabledColor = node.getString("enabledColor");
		if (node.isSpecified("disabledColor"))
			_disabledColor = node.getString("disabledColor");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("textAlign"))
			_textAlign = getTextAlign(node.getString("textAlign"));
		if (node.isSpecified("arrows"))
			_arrows = node.getBoolean("arrows");
		if (node.isSpecified("arrowUp"))
			_arrowUp = node.getString("arrowUp");
		if (node.isSpecified("arrowDown"))
			_arrowDown = node.getString("arrowDown");
		if (node.isSpecified("type"))
			_type = getType(node.getString("type"));
		if (node.isSpecified("scroll"))
			_scroll = getScroll(node.getString("scroll"));
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyCharType)
			addCharType((LyCharType)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyText clone = new LyText();

		clone.initFromModel(this);
		clone._template = (LyText)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyText getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Checks if value of attribute secret is set.
	 * @return true if secret is set, false otherwise.
	 */
	public boolean isSecret()
	{
		return _secret;
	}

	/**
	 * Gets value of nCols.
	 * @return value of nCols.
	 */
	public int getNCols()
	{
		return _nCols;
	}

	/**
	 * Gets value of nRows.
	 * @return value of nRows.
	 */
	public int getNRows()
	{
		return _nRows;
	}

	/**
	 * Gets value of nChars.
	 * @return value of nChars.
	 */
	public int getNChars()
	{
		return _nChars;
	}

	/**
	 * Gets value of format.
	 * @return value of format.
	 */
	public String getFormat()
	{
		return _format;
	}

	/**
	 * Checks if value of attribute editable is set.
	 * @return true if editable is set, false otherwise.
	 */
	public boolean isEditable()
	{
		return _editable;
	}

	/**
	 * Checks if value of attribute wrap is set.
	 * @return true if wrap is set, false otherwise.
	 */
	public boolean isWrap()
	{
		return _wrap;
	}

	/**
	 * Checks if value of attribute notifyValueChanged is set.
	 * @return true if notifyValueChanged is set, false otherwise.
	 */
	public boolean isNotifyValueChanged()
	{
		return _notifyValueChanged;
	}

	/**
	 * Gets value of autoCompletionValues.
	 * @return value of autoCompletionValues.
	 */
	public String getAutoCompletionValues()
	{
		return _autoCompletionValues;
	}

	/**
	 * Gets value of bean.
	 * @return value of bean.
	 */
	public String getBean()
	{
		return _bean;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("TEXT.font"));
	}

	/**
	 * Gets value of enabledColor.
	 * @return value of enabledColor.
	 */
	public String getEnabledColor()
	{
		return _enabledColor;
	}

	/**
	 * Checks whether the enabledColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultEnabledColor()
	{
		return (_enabledColor == getStringValue("TEXT.enabledColor"));
	}

	/**
	 * Gets value of disabledColor.
	 * @return value of disabledColor.
	 */
	public String getDisabledColor()
	{
		return _disabledColor;
	}

	/**
	 * Checks whether the disabledColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultDisabledColor()
	{
		return (_disabledColor == getStringValue("TEXT.disabledColor"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("TEXT.fontColor"));
	}

	/**
	 * Gets value of textAlign.
	 * @return value of textAlign.
	 */
	public TextAlign getTextAlign()
	{
		return _textAlign;
	}

	/**
	 * Checks if value of attribute arrows is set.
	 * @return true if arrows is set, false otherwise.
	 */
	public boolean isArrows()
	{
		return _arrows;
	}

	/**
	 * Gets value of arrowUp.
	 * @return value of arrowUp.
	 */
	public String getArrowUp()
	{
		return _arrowUp;
	}

	/**
	 * Checks whether the arrowUp attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultArrowUp()
	{
		return (_arrowUp == getStringValue("TEXT.arrowUp"));
	}

	/**
	 * Gets value of arrowDown.
	 * @return value of arrowDown.
	 */
	public String getArrowDown()
	{
		return _arrowDown;
	}

	/**
	 * Checks whether the arrowDown attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultArrowDown()
	{
		return (_arrowDown == getStringValue("TEXT.arrowDown"));
	}

	/**
	 * Gets value of type.
	 * @return value of type.
	 */
	public Type getType()
	{
		return _type;
	}

	/**
	 * Gets value of scroll.
	 * @return value of scroll.
	 */
	public Scroll getScroll()
	{
		return _scroll;
	}

	/**
	 * Get sub-component CharType of index i.
	 * @param i the index of the CharType to get.
	 * @return sub-component CharType of index i.
	 */
	public LyCharType getCharType(int i)
	{
		if (_charTypes == null)
			return null;
		if ((i < 0) || (i >= _charTypes.size()))
			return null;
		return _charTypes.get(i);
	}

	/**
	 * Get sub-component CharType count.
	 * @return sub-component CharType count.
	 */
	public int getCharTypeCount()
	{
		if (_charTypes == null)
			return 0;
		return _charTypes.size();
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyText template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView instanceof LySetLabelInterface)
			((LySetLabelInterface)_internalView).setLabel(label);
	}

	/**
	 * Initializes attribute secret to the given value.
	 * @param secret Boolean indicating if attribute secret is set or unset.
	 */
	public void setSecret(boolean secret)
	{
		if (secret == _secret)
			return;
		_secret = secret;
	}

	/**
	 * Initializes attribute nCols to the given value.
	 * @param nCols New value of nCols.
	 */
	public void setNCols(int nCols)
	{
		if (nCols == _nCols)
			return;
		_nCols = nCols;
	}

	/**
	 * Initializes attribute nRows to the given value.
	 * @param nRows New value of nRows.
	 */
	public void setNRows(int nRows)
	{
		if (nRows == _nRows)
			return;
		_nRows = nRows;
	}

	/**
	 * Initializes attribute nChars to the given value.
	 * @param nChars New value of nChars.
	 */
	public void setNChars(int nChars)
	{
		if (nChars == _nChars)
			return;
		_nChars = nChars;
	}

	/**
	 * Initializes attribute format to the given value.
	 * @param format New value of format.
	 */
	public void setFormat(String format)
	{
		if ((format == _format)
			|| ((format != null)
				&& (_format != null)
				&& format.equals(_format)))
			return;

		_format = format;
	}

	/**
	 * Initializes attribute editable to the given value.
	 * @param editable Boolean indicating if attribute editable is set or unset.
	 */
	public void setEditable(boolean editable)
	{
		if (editable == _editable)
			return;
		_editable = editable;
	}

	/**
	 * Initializes attribute wrap to the given value.
	 * @param wrap Boolean indicating if attribute wrap is set or unset.
	 */
	public void setWrap(boolean wrap)
	{
		if (wrap == _wrap)
			return;
		_wrap = wrap;
	}

	/**
	 * Initializes attribute notifyValueChanged to the given value.
	 * @param notifyValueChanged Boolean indicating if attribute notifyValueChanged is set or unset.
	 */
	public void setNotifyValueChanged(boolean notifyValueChanged)
	{
		if (notifyValueChanged == _notifyValueChanged)
			return;
		_notifyValueChanged = notifyValueChanged;
	}

	/**
	 * Initializes attribute autoCompletionValues to the given value.
	 * @param autoCompletionValues New value of autoCompletionValues.
	 */
	public void setAutoCompletionValues(String autoCompletionValues)
	{
		if ((autoCompletionValues == _autoCompletionValues)
			|| ((autoCompletionValues != null)
				&& (_autoCompletionValues != null)
				&& autoCompletionValues.equals(_autoCompletionValues)))
			return;

		_autoCompletionValues = autoCompletionValues;
	}

	/**
	 * Initializes attribute bean to the given value.
	 * @param bean New value of bean.
	 */
	public void setBean(String bean)
	{
		if ((bean == _bean)
			|| ((bean != null)
				&& (_bean != null)
				&& bean.equals(_bean)))
			return;

		_bean = bean;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute enabledColor to the given value.
	 * @param enabledColor New value of enabledColor.
	 */
	public void setEnabledColor(String enabledColor)
	{
		if ((enabledColor == _enabledColor)
			|| ((enabledColor != null)
				&& (_enabledColor != null)
				&& enabledColor.equals(_enabledColor)))
			return;

		_enabledColor = enabledColor;
	}

	/**
	 * Initializes attribute disabledColor to the given value.
	 * @param disabledColor New value of disabledColor.
	 */
	public void setDisabledColor(String disabledColor)
	{
		if ((disabledColor == _disabledColor)
			|| ((disabledColor != null)
				&& (_disabledColor != null)
				&& disabledColor.equals(_disabledColor)))
			return;

		_disabledColor = disabledColor;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Initializes attribute textAlign to the given value.
	 * @param textAlign New value of textAlign.
	 * @exception IllegalArgumentException if textAlign parameter is null.
	 */
	public void setTextAlign(TextAlign textAlign)
	{
		if (textAlign == _textAlign)
			return;
		if (textAlign == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute textAlign");

		_textAlign = textAlign;
	}

	/**
	 * Initializes attribute arrows to the given value.
	 * @param arrows Boolean indicating if attribute arrows is set or unset.
	 */
	public void setArrows(boolean arrows)
	{
		if (arrows == _arrows)
			return;
		_arrows = arrows;
	}

	/**
	 * Initializes attribute arrowUp to the given value.
	 * @param arrowUp New value of arrowUp.
	 */
	public void setArrowUp(String arrowUp)
	{
		if ((arrowUp == _arrowUp)
			|| ((arrowUp != null)
				&& (_arrowUp != null)
				&& arrowUp.equals(_arrowUp)))
			return;

		_arrowUp = arrowUp;
	}

	/**
	 * Initializes attribute arrowDown to the given value.
	 * @param arrowDown New value of arrowDown.
	 */
	public void setArrowDown(String arrowDown)
	{
		if ((arrowDown == _arrowDown)
			|| ((arrowDown != null)
				&& (_arrowDown != null)
				&& arrowDown.equals(_arrowDown)))
			return;

		_arrowDown = arrowDown;
	}

	/**
	 * Initializes attribute type to the given value.
	 * @param type New value of type.
	 * @exception IllegalArgumentException if type parameter is null.
	 */
	public void setType(Type type)
	{
		if (type == _type)
			return;
		if (type == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute type");

		_type = type;
	}

	/**
	 * Initializes attribute scroll to the given value.
	 * @param scroll New value of scroll.
	 * @exception IllegalArgumentException if scroll parameter is null.
	 */
	public void setScroll(Scroll scroll)
	{
		if (scroll == _scroll)
			return;
		if (scroll == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute scroll");

		_scroll = scroll;
	}

	/**
	 * Adds a sub-component charType.
	 * @param charType The charType to add to charTypes list.
	 */
	public void addCharType(LyCharType charType)
	{
		insertCharType(charType, -1);
	}

	/**
	 * Inserts a sub-component charType at given position.
	 * @param charType The charType to add to charTypes list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertCharType(LyCharType charType, int pos)
	{
		if (_charTypes == null)
		{
			_charTypes = new ArrayList<LyCharType>(1);
			_children = _charTypes;
		}
		else if (_charTypes.contains(charType))
			return;
		if (pos >= 0)
			_charTypes.add(pos,charType);
		if (charType != null)
			charType.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<TEXT");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (((_template == null) && _secret)
				|| ((_template != null) && (_secret != _template._secret)))
			pw.print(" secret='" + _secret + "'");
		if (((_template == null) && (_nCols != getIntValue("TEXT.nCols", -1)))
				|| ((_template != null) && (_nCols != _template._nCols)))
			pw.print(" nCols='" + _nCols + "'");
		if (((_template == null) && (_nRows != getIntValue("TEXT.nRows", -1)))
				|| ((_template != null) && (_nRows != _template._nRows)))
			pw.print(" nRows='" + _nRows + "'");
		if (((_template == null) && (_nChars != getIntValue("TEXT.nChars", -1)))
				|| ((_template != null) && (_nChars != _template._nChars)))
			pw.print(" nChars='" + _nChars + "'");
		if (((_template == null) && (_format != null))
				|| ((_template != null) && (_format != _template._format)))
			pw.print(" format='" + LyString.getXmlString(_format) + "'");
		if (((_template == null) && !_editable)
				|| ((_template != null) && (_editable != _template._editable)))
			pw.print(" editable='" + _editable + "'");
		if (((_template == null) && !_wrap)
				|| ((_template != null) && (_wrap != _template._wrap)))
			pw.print(" wrap='" + _wrap + "'");
		if (((_template == null) && _notifyValueChanged)
				|| ((_template != null) && (_notifyValueChanged != _template._notifyValueChanged)))
			pw.print(" notifyValueChanged='" + _notifyValueChanged + "'");
		if (((_template == null) && (_autoCompletionValues != null))
				|| ((_template != null) && (_autoCompletionValues != _template._autoCompletionValues)))
			pw.print(" autoCompletionValues='" + LyString.getXmlString(_autoCompletionValues) + "'");
		if (((_template == null) && (_bean != null))
				|| ((_template != null) && (_bean != _template._bean)))
			pw.print(" bean='" + LyString.getXmlString(_bean) + "'");
		if (((_template == null) && (_font != getStringValue("TEXT.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_enabledColor != getStringValue("TEXT.enabledColor", "$ALL.enabledColor")))
				|| ((_template != null) && (_enabledColor != _template._enabledColor)))
			pw.print(" enabledColor='" + LyString.getXmlString(_enabledColor) + "'");
		if (((_template == null) && (_disabledColor != getStringValue("TEXT.disabledColor", "$ALL.disabledColor")))
				|| ((_template != null) && (_disabledColor != _template._disabledColor)))
			pw.print(" disabledColor='" + LyString.getXmlString(_disabledColor) + "'");
		if (((_template == null) && (_fontColor != getStringValue("TEXT.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if ((_template == null)
			|| (_textAlign != _template._textAlign))
		{
			if ((_textAlign == TextAlign.LEFT) && (_template != null))
				pw.print(" textAlign='LEFT'");
			else if (_textAlign == TextAlign.CENTER)
				pw.print(" textAlign='CENTER'");
			else if (_textAlign == TextAlign.RIGHT)
				pw.print(" textAlign='RIGHT'");
		}
		if (((_template == null) && !_arrows)
				|| ((_template != null) && (_arrows != _template._arrows)))
			pw.print(" arrows='" + _arrows + "'");
		if (((_template == null) && (_arrowUp != getStringValue("TEXT.arrowUp", "$ALL.arrowUp")))
				|| ((_template != null) && (_arrowUp != _template._arrowUp)))
			pw.print(" arrowUp='" + LyString.getXmlString(_arrowUp) + "'");
		if (((_template == null) && (_arrowDown != getStringValue("TEXT.arrowDown", "$ALL.arrowDown")))
				|| ((_template != null) && (_arrowDown != _template._arrowDown)))
			pw.print(" arrowDown='" + LyString.getXmlString(_arrowDown) + "'");
		if ((_template == null)
			|| (_type != _template._type))
		{
			if ((_type == Type.STANDARD) && (_template != null))
				pw.print(" type='STANDARD'");
			else if (_type == Type.RICH)
				pw.print(" type='RICH'");
		}
		if ((_template == null)
			|| (_scroll != _template._scroll))
		{
			if ((_scroll == Scroll.TOP) && (_template != null))
				pw.print(" scroll='TOP'");
			else if (_scroll == Scroll.BOTTOM)
				pw.print(" scroll='BOTTOM'");
		}

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	charTypes;
		int			i, n;

		n = (_charTypes == null ? 0 : _charTypes.size());
		for (i = 0; i < n; i++)
		{
			charTypes = _charTypes.get(i);
			charTypes.writeXml(pw, debug);
		}


		pw.println("</TEXT>");
	}
}
