/*
 * LyToolBarBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyToolBarBase class is a generated Java implementation that corresponds to the element 
 * TOOLBAR of the descriptive grammar of the views (views.dtd).
 * <B>&lt;TOOLBAR&gt; : Tool bar</B>
 * A tool bar allows displaying a set of action buttons.
 * A tool bar can itself contain tool bars, which allows grouping together the 
 * action buttons into logical groups and sub-groups.
 * @see leon.view.LyComponent
 * @see leon.view.LyLyToolBar
 * See also tag <A href="../../conf/dtd/view/TOOLBAR.html">TOOLBAR</A> of grammar leon_views.dtd.
 */
public abstract class LyToolBarBase extends LyComponent
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LyGetControllerInterface,
				LySetControllerInterface,
				LySetImageInterface,
				LySetColorInterface,
				LyGetConstraintInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'fill'. */
	public enum Fill
	{
		DEFAULT,
		HORIZONTAL,
		VERTICAL;
	}


	/** Possible values for attribute 'style'. */
	public enum Style
	{
		DEFAULT,
		BORDER,
		MENU,
		SEPARATOR,
		EXPAND;
	}


	/** Possible values for attribute 'hAlign'. */
	public enum HAlign
	{
		LEFT,
		RIGHT;
	}


	/** Possible values for attribute 'vAlign'. */
	public enum VAlign
	{
		TOP,
		BOTTOM,
		MIDDLE;
	}


	/* Static methods */

	/**
	 * Utility method to get fill value from a string.
	 * @param token Token string
	 * @return Fill value corresponding to this token
	 */
	public static final Fill getFill(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("DEFAULT")
				|| token.equals("0"))
			return Fill.DEFAULT;
		else if (token.equalsIgnoreCase("HORIZONTAL")
				|| token.equals("1"))
			return Fill.HORIZONTAL;
		else if (token.equalsIgnoreCase("VERTICAL")
				|| token.equals("2"))
			return Fill.VERTICAL;
		else
			return Fill.DEFAULT;
	}

	/**
	 * Utility method to get style value from a string.
	 * @param token Token string
	 * @return Style value corresponding to this token
	 */
	public static final Style getStyle(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("DEFAULT")
				|| token.equals("0"))
			return Style.DEFAULT;
		else if (token.equalsIgnoreCase("BORDER")
				|| token.equals("1"))
			return Style.BORDER;
		else if (token.equalsIgnoreCase("MENU")
				|| token.equals("2"))
			return Style.MENU;
		else if (token.equalsIgnoreCase("SEPARATOR")
				|| token.equals("3"))
			return Style.SEPARATOR;
		else if (token.equalsIgnoreCase("EXPAND")
				|| token.equals("4"))
			return Style.EXPAND;
		else
			return Style.DEFAULT;
	}

	/**
	 * Utility method to get hAlign value from a string.
	 * @param token Token string
	 * @return HAlign value corresponding to this token
	 */
	public static final HAlign getHAlign(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("LEFT")
				|| token.equals("0"))
			return HAlign.LEFT;
		else if (token.equalsIgnoreCase("RIGHT")
				|| token.equals("1"))
			return HAlign.RIGHT;
		else
			return HAlign.LEFT;
	}

	/**
	 * Utility method to get vAlign value from a string.
	 * @param token Token string
	 * @return VAlign value corresponding to this token
	 */
	public static final VAlign getVAlign(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("TOP")
				|| token.equals("0"))
			return VAlign.TOP;
		else if (token.equalsIgnoreCase("BOTTOM")
				|| token.equals("1"))
			return VAlign.BOTTOM;
		else if (token.equalsIgnoreCase("MIDDLE")
				|| token.equals("2"))
			return VAlign.MIDDLE;
		else
			return VAlign.TOP;
	}


	/* Attributes */

	/**
	 * Identifier of the tool bar.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyToolBar			_template;

	/**
	 * Text of the tool bar.
	 */
	protected String			_label;

	/**
	 * Reference of the controller that manages the tool bar (i.e. handles user demands
	 * and refresh content).
	 */
	protected String			_controller;

	/**
	 * Image of the tool bar.
	 */
	protected String			_image;

	/**
	 * Allows specifying whether the bar is horizontal or vertical.
	 * The DEFAULT value lets the viewer decide the way the component will be placed. A
	 * vertical fill automatically set the vAlign field to the MIDDLE value.
	 */
	protected Fill				_fill;

	/**
	 * Style of the tool bar.
	 * The following styles are supported:
	 * DEFAULT The viewer decides the style that will be attributed to the tool bar.
	 * BORDER The tool bar is outlined by a border.
	 * MENU The tool bar is shown under the form of a menu.
	 * EXPAND The tool bar is shown under the form of an expand bar.
	 * A tool bar contained in such a bar becomes then a sub-menu.
	 * SEPARATOR The tool bar is flanked on both sides by separating lines which 
	 * direction is the same as the one of the bar.
	 */
	protected Style				_style;

	/**
	 * Horizontal position of the tool bar.
	 * The following positions are supported:
	 * LEFT The toolbar is at the left of the view.
	 * RIGHT The toolbar is at the right of the view.
	 */
	protected HAlign			_hAlign;

	/**
	 * Vertical position
	 * of the tool bar.
	 * The following position are supported:
	 * TOP The toolbar is at the top of the view. This position implies an horizontal 
	 * fill for the toolbar.
	 * BOTTOM The tool bar is at the bottom of the view. This position implies an 
	 * horizontal fill for the toolbar.
	 * MIDDLE The tool bar is on the left or the right side of the view. This position 
	 * implies a vertical fill for the toolbar.
	 */
	protected VAlign			_vAlign;

	/**
	 * Indicates if the tool bar is opened or not when first displayed.
	 */
	protected boolean			_open;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Background color of the tool bar.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;


	/* Sub components */

	/**
	 * Components of a toolbar are tools, separators or sub toolbars.
	 */
	protected ArrayList<LyComponent>		_components;

	/**
	 * Positioning constraint of the tool bar.
	 */
	protected LyConstraint		_constraint;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_controller = null;
		_image = getStringValue("TOOLBAR.image", "$ALL.image");
		_fill = Fill.DEFAULT;
		_style = Style.DEFAULT;
		_hAlign = HAlign.LEFT;
		_vAlign = VAlign.TOP;
		_open = getBooleanValue("TOOLBAR.open", true);
		_font = getStringValue("TOOLBAR.font", "$ALL.font");
		_color = getStringValue("TOOLBAR.color", "$ALL.color");
		_fontColor = getStringValue("TOOLBAR.fontColor", "$ALL.fontColor");

		// Initialize sub components
		_components = null;
		_constraint = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyToolBar	model = (LyToolBar)component;

		_model = component;
		_label = model._label;
		_controller = model._controller;
		_image = model._image;
		_fill = model._fill;
		_style = model._style;
		_hAlign = model._hAlign;
		_vAlign = model._vAlign;
		_open = model._open;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyComponent	components;
		int			i, n;

		n = (model._components == null ? 0 : model._components.size());
		for (i = 0; i < n; i++)
		{
			components = model._components.get(i);
			if (components != null)
				components = components.cloneComponent();
			addChild(components);
		}
		if (model._constraint != null)
			addChild(model._constraint.cloneComponent());
		if (model._datas != null)
			_datas = new Hashtable<String, String>(model._datas);
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyToolBarBase()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyToolBarBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyToolBar)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("controller"))
			_controller = node.getString("controller");
		if (node.isSpecified("image"))
			_image = node.getString("image");
		if (node.isSpecified("fill"))
			_fill = getFill(node.getString("fill"));
		if (node.isSpecified("style"))
			_style = getStyle(node.getString("style"));
		if (node.isSpecified("hAlign"))
			_hAlign = getHAlign(node.getString("hAlign"));
		if (node.isSpecified("vAlign"))
			_vAlign = getVAlign(node.getString("vAlign"));
		if (node.isSpecified("open"))
			_open = node.getBoolean("open");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyTool)
			addTool((LyTool)child);
		else if (child instanceof LyToolBar)
			addToolBar((LyToolBar)child);
		else if (child == null)
			addSep();
		else if (child instanceof LyLabel)
			addLabel((LyLabel)child);
		else if (child instanceof LyFormField)
			addFormField((LyFormField)child);
		else if (child instanceof LyConstraint)
			setConstraint((LyConstraint)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyToolBar clone = new LyToolBar();

		clone.initFromModel(this);
		clone._template = (LyToolBar)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyToolBar getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of controller.
	 * @return value of controller.
	 */
	public String getController()
	{
		return _controller;
	}

	/**
	 * Gets value of image.
	 * @return value of image.
	 */
	public String getImage()
	{
		return _image;
	}

	/**
	 * Checks whether the image attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultImage()
	{
		return (_image == getStringValue("TOOLBAR.image"));
	}

	/**
	 * Gets value of fill.
	 * @return value of fill.
	 */
	public Fill getFill()
	{
		return _fill;
	}

	/**
	 * Gets value of style.
	 * @return value of style.
	 */
	public Style getStyle()
	{
		return _style;
	}

	/**
	 * Gets value of hAlign.
	 * @return value of hAlign.
	 */
	public HAlign getHAlign()
	{
		return _hAlign;
	}

	/**
	 * Gets value of vAlign.
	 * @return value of vAlign.
	 */
	public VAlign getVAlign()
	{
		return _vAlign;
	}

	/**
	 * Checks if value of attribute open is set.
	 * @return true if open is set, false otherwise.
	 */
	public boolean isOpen()
	{
		return _open;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("TOOLBAR.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("TOOLBAR.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("TOOLBAR.fontColor"));
	}

	/**
	 * Get sub-component Component of index i.
	 * @param i the index of the Component to get.
	 * @return sub-component Component of index i.
	 */
	public LyComponent getComponent(int i)
	{
		if (_components == null)
			return null;
		if ((i < 0) || (i >= _components.size()))
			return null;
		return _components.get(i);
	}

	/**
	 * Get sub-component Component count.
	 * @return sub-component Component count.
	 */
	public int getComponentCount()
	{
		if (_components == null)
			return 0;
		return _components.size();
	}

	/**
	 * Get sub-component Constraint.
	 * @return sub-component Constraint.
	 */
	public LyConstraint getConstraint()
	{
		return _constraint;
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyToolBar template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Initializes attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
	}

	/**
	 * Initializes attribute controller to the given value.
	 * @param controller New value of controller.
	 */
	public void setController(String controller)
	{
		if ((controller == _controller)
			|| ((controller != null)
				&& (_controller != null)
				&& controller.equals(_controller)))
			return;

		_controller = controller;
	}

	/**
	 * Sets attribute image to the given value.
	 * @param image New value of image.
	 */
	public void setImage(String image)
	{
		if ((image == _image)
			|| ((image != null)
				&& (_image != null)
				&& image.equals(_image)))
			return;

		_image = image;
		if (_internalView != null)
			((LyGraphicalToolBar)_internalView).setImage(image);
	}

	/**
	 * Initializes attribute fill to the given value.
	 * @param fill New value of fill.
	 * @exception IllegalArgumentException if fill parameter is null.
	 */
	public void setFill(Fill fill)
	{
		if (fill == _fill)
			return;
		if (fill == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute fill");

		_fill = fill;
	}

	/**
	 * Initializes attribute style to the given value.
	 * @param style New value of style.
	 * @exception IllegalArgumentException if style parameter is null.
	 */
	public void setStyle(Style style)
	{
		if (style == _style)
			return;
		if (style == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute style");

		_style = style;
	}

	/**
	 * Initializes attribute hAlign to the given value.
	 * @param hAlign New value of hAlign.
	 * @exception IllegalArgumentException if hAlign parameter is null.
	 */
	public void setHAlign(HAlign hAlign)
	{
		if (hAlign == _hAlign)
			return;
		if (hAlign == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute hAlign");

		_hAlign = hAlign;
	}

	/**
	 * Initializes attribute vAlign to the given value.
	 * @param vAlign New value of vAlign.
	 * @exception IllegalArgumentException if vAlign parameter is null.
	 */
	public void setVAlign(VAlign vAlign)
	{
		if (vAlign == _vAlign)
			return;
		if (vAlign == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute vAlign");

		_vAlign = vAlign;
	}

	/**
	 * Sets attribute open to the given value.
	 * @param open Boolean indicating if attribute open is set or unset.
	 */
	public void setOpen(boolean open)
	{
		if (open == _open)
			return;
		_open = open;
		if (_internalView != null)
			((LyGraphicalToolBar)_internalView).setOpen(open);
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Sets attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
		if (_internalView != null)
			((LyGraphicalToolBar)_internalView).setColor(color);
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Adds a sub-component tool.
	 * @param tool The tool to add to components list.
	 */
	public void addTool(LyTool tool)
	{
		insertTool(tool, -1);
	}

	/**
	 * Inserts a sub-component tool at given position.
	 * @param tool The tool to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertTool(LyTool tool, int pos)
	{
		if (_components == null)
			_components = new ArrayList<LyComponent>(1);
		else if (_components.contains(tool))
			return;
		if (pos >= 0)
			_components.add(pos,tool);
		else
			_components.add(tool);
		if (tool != null)
			tool.setParent(this);
		if (_internalView != null)
			((LyGraphicalToolBar)_internalView).addComponent(tool, pos);
	}

	/**
	 * Removes a sub-component tool.
	 * @param tool The tool to remove to components list.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain tool.
	 */
	public void removeTool(LyTool tool)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if (!_components.contains(tool))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain tool");

		if (_internalView != null)
			((LyGraphicalToolBar)_internalView).removeComponent(tool);
		_components.remove(tool);
		tool.setParent(null);
	}

	/**
	 * Adds a sub-component toolBar.
	 * @param toolBar The toolBar to add to components list.
	 */
	public void addToolBar(LyToolBar toolBar)
	{
		insertToolBar(toolBar, -1);
	}

	/**
	 * Inserts a sub-component toolBar at given position.
	 * @param toolBar The toolBar to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertToolBar(LyToolBar toolBar, int pos)
	{
		if (_components == null)
			_components = new ArrayList<LyComponent>(1);
		else if (_components.contains(toolBar))
			return;
		if (pos >= 0)
			_components.add(pos,toolBar);
		else
			_components.add(toolBar);
		if (toolBar != null)
			toolBar.setParent(this);
		if (_internalView != null)
			((LyGraphicalToolBar)_internalView).addComponent(toolBar, pos);
	}

	/**
	 * Removes a sub-component toolBar.
	 * @param toolBar The toolBar to remove to components list.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain toolBar.
	 */
	public void removeToolBar(LyToolBar toolBar)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if (!_components.contains(toolBar))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain toolBar");

		if (_internalView != null)
			((LyGraphicalToolBar)_internalView).removeComponent(toolBar);
		_components.remove(toolBar);
		toolBar.setParent(null);
	}

	/**
	 * Adds a separator.
	 */
	public void addSep()
	{
		insertSep(-1);
	}

	/**
	 * Inserts a separator at the given position.
	 * @param pos	Position where the separator is inserted.
	 */
	public void insertSep(int pos)
	{
		LyTool sep = new LyTool(SEP_ID, null);
		insertTool(sep, pos);
	}

	/**
	 * Removes a separator at the given position.
	 * @param pos	Position where the separator is removed.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain a separator at the given position.
	 */
	public void removeSep(int pos)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if ((pos < 0)
				|| (pos >= _components.size())
				|| ((_components.get(pos) != null)
					&& (_components.get(pos).getId() != SEP_ID)))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain a separator at pos " + pos);

		_components.remove(pos);
	}

	/**
	 * Adds a sub-component label.
	 * @param label The label to add to components list.
	 */
	public void addLabel(LyLabel label)
	{
		insertLabel(label, -1);
	}

	/**
	 * Inserts a sub-component label at given position.
	 * @param label The label to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertLabel(LyLabel label, int pos)
	{
		if (_components == null)
			_components = new ArrayList<LyComponent>(1);
		else if (_components.contains(label))
			return;
		if (pos >= 0)
			_components.add(pos,label);
		else
			_components.add(label);
		if (label != null)
			label.setParent(this);
		if (_internalView != null)
			((LyGraphicalToolBar)_internalView).addComponent(label, pos);
	}

	/**
	 * Removes a sub-component label.
	 * @param label The label to remove to components list.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain label.
	 */
	public void removeLabel(LyLabel label)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if (!_components.contains(label))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain label");

		if (_internalView != null)
			((LyGraphicalToolBar)_internalView).removeComponent(label);
		_components.remove(label);
		label.setParent(null);
	}

	/**
	 * Adds a sub-component formField.
	 * @param formField The formField to add to components list.
	 */
	public void addFormField(LyFormField formField)
	{
		insertFormField(formField, -1);
	}

	/**
	 * Inserts a sub-component formField at given position.
	 * @param formField The formField to add to components list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertFormField(LyFormField formField, int pos)
	{
		if (_components == null)
			_components = new ArrayList<LyComponent>(1);
		else if (_components.contains(formField))
			return;
		if (pos >= 0)
			_components.add(pos,formField);
		else
			_components.add(formField);
		if (formField != null)
			formField.setParent(this);
		if (_internalView != null)
			((LyGraphicalToolBar)_internalView).addComponent(formField, pos);
	}

	/**
	 * Removes a sub-component formField.
	 * @param formField The formField to remove to components list.
	 * @exception IllegalStateException if components list is empty.
	 * @exception IllegalArgumentException if components list doesn't contain formField.
	 */
	public void removeFormField(LyFormField formField)
	{
		if (_components == null)
			throw new IllegalStateException("BF AppEngine : components list is empty");
		if (!_components.contains(formField))
			throw new IllegalArgumentException("BF AppEngine : components list doesn't contain formField");

		if (_internalView != null)
			((LyGraphicalToolBar)_internalView).removeComponent(formField);
		_components.remove(formField);
		formField.setParent(null);
	}

	/**
	 * Sets sub-component constraint.
	 * @param constraint New value of constraint
	 */
	public void setConstraint(LyConstraint constraint)
	{
		if (_constraint != null)
			_constraint.setParent(null);
		_constraint = constraint;
		if (constraint != null)
			constraint.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<TOOLBAR");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlString(_controller) + "'");
		if (((_template == null) && (_image != getStringValue("TOOLBAR.image", "$ALL.image")))
				|| ((_template != null) && (_image != _template._image)))
			pw.print(" image='" + LyString.getXmlString(_image) + "'");
		if ((_template == null)
			|| (_fill != _template._fill))
		{
			if ((_fill == Fill.DEFAULT) && (_template != null))
				pw.print(" fill='DEFAULT'");
			else if (_fill == Fill.HORIZONTAL)
				pw.print(" fill='HORIZONTAL'");
			else if (_fill == Fill.VERTICAL)
				pw.print(" fill='VERTICAL'");
		}
		if ((_template == null)
			|| (_style != _template._style))
		{
			if ((_style == Style.DEFAULT) && (_template != null))
				pw.print(" style='DEFAULT'");
			else if (_style == Style.BORDER)
				pw.print(" style='BORDER'");
			else if (_style == Style.MENU)
				pw.print(" style='MENU'");
			else if (_style == Style.SEPARATOR)
				pw.print(" style='SEPARATOR'");
			else if (_style == Style.EXPAND)
				pw.print(" style='EXPAND'");
		}
		if ((_template == null)
			|| (_hAlign != _template._hAlign))
		{
			if ((_hAlign == HAlign.LEFT) && (_template != null))
				pw.print(" hAlign='LEFT'");
			else if (_hAlign == HAlign.RIGHT)
				pw.print(" hAlign='RIGHT'");
		}
		if ((_template == null)
			|| (_vAlign != _template._vAlign))
		{
			if ((_vAlign == VAlign.TOP) && (_template != null))
				pw.print(" vAlign='TOP'");
			else if (_vAlign == VAlign.BOTTOM)
				pw.print(" vAlign='BOTTOM'");
			else if (_vAlign == VAlign.MIDDLE)
				pw.print(" vAlign='MIDDLE'");
		}
		if (((_template == null) && !_open)
				|| ((_template != null) && (_open != _template._open)))
			pw.print(" open='" + _open + "'");
		if (((_template == null) && (_font != getStringValue("TOOLBAR.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("TOOLBAR.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("TOOLBAR.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	components;
		int			i, n;

		n = (_components == null ? 0 : _components.size());
		for (i = 0; i < n; i++)
		{
			components = _components.get(i);
			if ((components == null)
					|| (components.getId() == SEP_ID))
				pw.println("<SEP/>");
			else
				components.writeXml(pw, debug);
		}

		if (_constraint != null)
			_constraint.writeXml(pw, debug);

		writeDatas(pw);

		pw.println("</TOOLBAR>");
	}
}
