/*
 * LyToolBase.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyToolBase class is a generated Java implementation that corresponds to the element 
 * TOOL of the descriptive grammar of the views (views.dtd).
 * <B>&lt;TOOL&gt; : Action button</B>
 * An action button allows executing an action.
 * In the case where the button contains an area (AREA), this area is displayed in 
 * a new window when launching the action.
 * @see leon.view.LyPanel
 * @see leon.view.LyLyTool
 * See also tag <A href="../../conf/dtd/view/TOOL.html">TOOL</A> of grammar leon_views.dtd.
 */
public abstract class LyToolBase extends LyPanel
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface,
				LySetImageInterface,
				LySetEnabledInterface,
				LyGetConstraintInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'style'. */
	public enum Style
	{
		DEFAULT,
		TOGGLE;
	}


	/* Static methods */

	/**
	 * Utility method to get style value from a string.
	 * @param token Token string
	 * @return Style value corresponding to this token
	 */
	public static final Style getStyle(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("DEFAULT")
				|| token.equals("0"))
			return Style.DEFAULT;
		else if (token.equalsIgnoreCase("TOGGLE")
				|| token.equals("1"))
			return Style.TOGGLE;
		else
			return Style.DEFAULT;
	}


	/* Attributes */

	/**
	 * Identifier of the action button.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyTool			_template;

	/**
	 * Text of the action button.
	 */
	protected String			_label;

	/**
	 * Tooltip associated to the action button.
	 */
	protected String			_tooltip;

	/**
	 * Overview label.
	 */
	protected String			_overview;

	/**
	 * Image of the action button.
	 */
	protected String			_image;

	/**
	 * Style of the tool
	 * The following styles are supported:
	 * DEFAULT The viewer decides the style that will be attributed to the tool.
	 * TOGGLE The tool is a two-state button.
	 */
	protected Style				_style;

	/**
	 * Indicates whether the tool is pressed or not (TOGGLE style only).
	 */
	protected boolean			_pressed;

	/**
	 * Indicates whether the action button is enabled or disabled (true if is enabled).
	 * Disabled means that no interaction is possible on the corresponding graphical 
	 * component.
	 */
	protected boolean			_enabled;

	/**
	 * Indicates whether the action launched by the button opens a dialog window or 
	 * not.
	 * In dialog mode, the text of the action is followed by suspension points.
	 */
	protected boolean			_dialog;

	/**
	 * Width of the tool in pixels.
	 */
	protected int				_width;

	/**
	 * Height of the tool in pixels.
	 */
	protected int				_height;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * Color associated to the click action (mobile only)
	 */
	protected String			_clickColor;


	/* Sub components */

	/**
	 * Component displayed when selecting the tool (may be an area or a popup menu).
	 */
	protected LyComponent		_component;

	/**
	 * Positioning constraints in the including area.
	 */
	protected LyConstraint		_constraint;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_tooltip = null;
		_overview = null;
		_image = getStringValue("TOOL.image", "$ALL.image");
		_style = Style.DEFAULT;
		_pressed = getBooleanValue("TOOL.pressed", false);
		_enabled = getBooleanValue("TOOL.enabled", true);
		_dialog = getBooleanValue("TOOL.dialog", false);
		_width = getIntValue("TOOL.width", -1);
		_height = getIntValue("TOOL.height", -1);
		_font = getStringValue("TOOL.font", "$ALL.font");
		_color = getStringValue("TOOL.color", "$ALL.color");
		_fontColor = getStringValue("TOOL.fontColor", "$ALL.fontColor");
		_clickColor = getStringValue("TOOL.clickColor", "$ALL.clickColor");

		// Initialize sub components
		_component = null;
		_constraint = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyTool	model = (LyTool)component;

		_model = component;
		_label = model._label;
		_tooltip = model._tooltip;
		_overview = model._overview;
		_image = model._image;
		_style = model._style;
		_pressed = model._pressed;
		_enabled = model._enabled;
		_dialog = model._dialog;
		_width = model._width;
		_height = model._height;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;
		_clickColor = model._clickColor;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		if (model._component != null)
			addChild(model._component.cloneComponent());
		if (model._constraint != null)
			addChild(model._constraint.cloneComponent());
		if (model._datas != null)
			_datas = new Hashtable<String, String>(model._datas);
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	protected LyToolBase()
	{
		_id = getInternalId();
	}

	/**
	 * Constructor with all required attributes.
	 * @param id Identifier of the action button.
	 * @param label Text of the action button.
	 */
	public LyToolBase(String id, String label)
	{
		// Fill up required attributes with parameters
		_id = (id != null ? id.intern() : null);
		_label = label;
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyToolBase(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyTool)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("tooltip"))
			_tooltip = node.getString("tooltip");
		if (node.isSpecified("overview"))
			_overview = node.getString("overview");
		if (node.isSpecified("image"))
			_image = node.getString("image");
		if (node.isSpecified("style"))
			_style = getStyle(node.getString("style"));
		if (node.isSpecified("pressed"))
			_pressed = node.getBoolean("pressed");
		if (node.isSpecified("enabled"))
			_enabled = node.getBoolean("enabled");
		if (node.isSpecified("dialog"))
			_dialog = node.getBoolean("dialog");
		if (node.isSpecified("width"))
			_width = node.getInt("width");
		if (node.isSpecified("height"))
			_height = node.getInt("height");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("clickColor"))
			_clickColor = node.getString("clickColor");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyArea)
			setComponent(child);
		else if (child instanceof LyPopupMenu)
			setComponent(child);
		else if (child instanceof LyConstraint)
			setConstraint((LyConstraint)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyTool clone = new LyTool();

		clone.initFromModel(this);
		clone._template = (LyTool)this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyTool getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of tooltip.
	 * @return value of tooltip.
	 */
	public String getTooltip()
	{
		return _tooltip;
	}

	/**
	 * Gets value of overview.
	 * @return value of overview.
	 */
	public String getOverview()
	{
		return _overview;
	}

	/**
	 * Gets value of image.
	 * @return value of image.
	 */
	public String getImage()
	{
		return _image;
	}

	/**
	 * Checks whether the image attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultImage()
	{
		return (_image == getStringValue("TOOL.image"));
	}

	/**
	 * Gets value of style.
	 * @return value of style.
	 */
	public Style getStyle()
	{
		return _style;
	}

	/**
	 * Checks if value of attribute pressed is set.
	 * @return true if pressed is set, false otherwise.
	 */
	public boolean isPressed()
	{
		return _pressed;
	}

	/**
	 * Checks if value of attribute enabled is set.
	 * @return true if enabled is set, false otherwise.
	 */
	public boolean isEnabled()
	{
		return _enabled;
	}

	/**
	 * Checks if value of attribute dialog is set.
	 * @return true if dialog is set, false otherwise.
	 */
	public boolean isDialog()
	{
		return _dialog;
	}

	/**
	 * Gets value of width.
	 * @return value of width.
	 */
	public int getWidth()
	{
		return _width;
	}

	/**
	 * Gets value of height.
	 * @return value of height.
	 */
	public int getHeight()
	{
		return _height;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("TOOL.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("TOOL.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("TOOL.fontColor"));
	}

	/**
	 * Gets value of clickColor.
	 * @return value of clickColor.
	 */
	public String getClickColor()
	{
		return _clickColor;
	}

	/**
	 * Checks whether the clickColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultClickColor()
	{
		return (_clickColor == getStringValue("TOOL.clickColor"));
	}

	/**
	 * Get sub-component Component.
	 * @return sub-component Component.
	 */
	public LyComponent getComponent()
	{
		return _component;
	}

	/**
	 * Get sub-component Constraint.
	 * @return sub-component Constraint.
	 */
	public LyConstraint getConstraint()
	{
		return _constraint;
	}


	/* Modifiers */

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyTool template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 * @exception IllegalArgumentException if label is null.
	 */
	public void setLabel(String label)
	{
		if (label == null)
			throw new IllegalArgumentException("BF AppEngine : attribute label is required (can't be null)");

		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView != null)
			((LyGraphicalTool)_internalView).setLabel(label);
	}

	/**
	 * Sets attribute tooltip to the given value.
	 * @param tooltip New value of tooltip.
	 */
	public void setTooltip(String tooltip)
	{
		if ((tooltip == _tooltip)
			|| ((tooltip != null)
				&& (_tooltip != null)
				&& tooltip.equals(_tooltip)))
			return;

		_tooltip = tooltip;
		if (_internalView != null)
			((LyGraphicalTool)_internalView).setTooltip(tooltip);
	}

	/**
	 * Initializes attribute overview to the given value.
	 * @param overview New value of overview.
	 */
	public void setOverview(String overview)
	{
		if ((overview == _overview)
			|| ((overview != null)
				&& (_overview != null)
				&& overview.equals(_overview)))
			return;

		_overview = overview;
	}

	/**
	 * Sets attribute image to the given value.
	 * @param image New value of image.
	 */
	public void setImage(String image)
	{
		if ((image == _image)
			|| ((image != null)
				&& (_image != null)
				&& image.equals(_image)))
			return;

		_image = image;
		if (_internalView != null)
			((LyGraphicalTool)_internalView).setImage(image);
	}

	/**
	 * Initializes attribute style to the given value.
	 * @param style New value of style.
	 * @exception IllegalArgumentException if style parameter is null.
	 */
	public void setStyle(Style style)
	{
		if (style == _style)
			return;
		if (style == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute style");

		_style = style;
	}

	/**
	 * Sets attribute pressed to the given value.
	 * @param pressed Boolean indicating if attribute pressed is set or unset.
	 */
	public void setPressed(boolean pressed)
	{
		if (pressed == _pressed)
			return;
		_pressed = pressed;
		if (_internalView != null)
			((LyGraphicalTool)_internalView).setPressed(pressed);
	}

	/**
	 * Sets attribute enabled to the given value.
	 * @param enabled Boolean indicating if attribute enabled is set or unset.
	 */
	public void setEnabled(boolean enabled)
	{
		if (enabled == _enabled)
			return;
		_enabled = enabled;
		if (_internalView != null)
			((LyGraphicalTool)_internalView).setEnabled(enabled);
	}

	/**
	 * Initializes attribute dialog to the given value.
	 * @param dialog Boolean indicating if attribute dialog is set or unset.
	 */
	public void setDialog(boolean dialog)
	{
		if (dialog == _dialog)
			return;
		_dialog = dialog;
	}

	/**
	 * Initializes attribute width to the given value.
	 * @param width New value of width.
	 */
	public void setWidth(int width)
	{
		if (width == _width)
			return;
		_width = width;
	}

	/**
	 * Initializes attribute height to the given value.
	 * @param height New value of height.
	 */
	public void setHeight(int height)
	{
		if (height == _height)
			return;
		_height = height;
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Initializes attribute clickColor to the given value.
	 * @param clickColor New value of clickColor.
	 */
	public void setClickColor(String clickColor)
	{
		if ((clickColor == _clickColor)
			|| ((clickColor != null)
				&& (_clickColor != null)
				&& clickColor.equals(_clickColor)))
			return;

		_clickColor = clickColor;
	}

	/**
	 * Sets sub-component component.
	 * @param component New value of component
	 */
	public void setComponent(LyComponent component)
	{
		if (_component != null)
			_component.setParent(null);
		_component = component;
		if (component != null)
			component.setParent(this);
	}

	/**
	 * Sets sub-component constraint.
	 * @param constraint New value of constraint
	 */
	public void setConstraint(LyConstraint constraint)
	{
		if (_constraint != null)
			_constraint.setParent(null);
		_constraint = constraint;
		if (constraint != null)
			constraint.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<TOOL");
		pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (((_template == null) && (_tooltip != null))
				|| ((_template != null) && (_tooltip != _template._tooltip)))
			pw.print(" tooltip='" + LyString.getXmlString(_tooltip) + "'");
		if (((_template == null) && (_overview != null))
				|| ((_template != null) && (_overview != _template._overview)))
			pw.print(" overview='" + LyString.getXmlString(_overview) + "'");
		if (((_template == null) && (_image != getStringValue("TOOL.image", "$ALL.image")))
				|| ((_template != null) && (_image != _template._image)))
			pw.print(" image='" + LyString.getXmlString(_image) + "'");
		if ((_template == null)
			|| (_style != _template._style))
		{
			if ((_style == Style.DEFAULT) && (_template != null))
				pw.print(" style='DEFAULT'");
			else if (_style == Style.TOGGLE)
				pw.print(" style='TOGGLE'");
		}
		if (((_template == null) && _pressed)
				|| ((_template != null) && (_pressed != _template._pressed)))
			pw.print(" pressed='" + _pressed + "'");
		if (((_template == null) && !_enabled)
				|| ((_template != null) && (_enabled != _template._enabled)))
			pw.print(" enabled='" + _enabled + "'");
		if (((_template == null) && _dialog)
				|| ((_template != null) && (_dialog != _template._dialog)))
			pw.print(" dialog='" + _dialog + "'");
		if (((_template == null) && (_width != getIntValue("TOOL.width", -1)))
				|| ((_template != null) && (_width != _template._width)))
			pw.print(" width='" + _width + "'");
		if (((_template == null) && (_height != getIntValue("TOOL.height", -1)))
				|| ((_template != null) && (_height != _template._height)))
			pw.print(" height='" + _height + "'");
		if (((_template == null) && (_font != getStringValue("TOOL.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("TOOL.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("TOOL.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && (_clickColor != getStringValue("TOOL.clickColor", "$ALL.clickColor")))
				|| ((_template != null) && (_clickColor != _template._clickColor)))
			pw.print(" clickColor='" + LyString.getXmlString(_clickColor) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		if (_component != null)
			_component.writeXml(pw, debug);
		if (_constraint != null)
			_constraint.writeXml(pw, debug);

		writeDatas(pw);

		pw.println("</TOOL>");
	}
}
