/*
 * LyTree.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyTree class is a generated Java implementation that corresponds to the element 
 * TREE of the descriptive grammar of the views (views.dtd).
 * <B>&lt;TREE&gt; : Tree</B>
 * A tree allows displaying a set of elements under the form of a tree.
 * It is possible to specify its selection mode (none, simple, multiple or 
 * sensitive). Such a component is interesting to display a hierarchy of objects of
 * an application.
 * It allows not displaying some sub-sets of these objects.
 * @see leon.view.LyPanel
 * See also tag <A href="../../conf/dtd/view/TREE.html">TREE</A> of grammar leon_views.dtd.
 */
public class LyTree extends LyPanel
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface,
				LyGetControllerInterface,
				LySetControllerInterface,
				LyGetConstraintInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'selectionMode'. */
	public enum SelectionMode
	{
		NONE,
		SINGLE,
		MULTIPLE,
		SENSITIVE,
		CHECKBOX;
	}


	/** Possible values for attribute 'selectionType'. */
	public enum SelectionType
	{
		ALL,
		LEAVES_ONLY;
	}


	/* Static methods */

	/**
	 * Utility method to get selectionMode value from a string.
	 * @param token Token string
	 * @return SelectionMode value corresponding to this token
	 */
	public static final SelectionMode getSelectionMode(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("NONE")
				|| token.equals("0"))
			return SelectionMode.NONE;
		else if (token.equalsIgnoreCase("SINGLE")
				|| token.equals("1"))
			return SelectionMode.SINGLE;
		else if (token.equalsIgnoreCase("MULTIPLE")
				|| token.equals("2"))
			return SelectionMode.MULTIPLE;
		else if (token.equalsIgnoreCase("SENSITIVE")
				|| token.equals("3"))
			return SelectionMode.SENSITIVE;
		else if (token.equalsIgnoreCase("CHECKBOX")
				|| token.equals("4"))
			return SelectionMode.CHECKBOX;
		else
			return SelectionMode.SINGLE;
	}

	/**
	 * Utility method to get selectionType value from a string.
	 * @param token Token string
	 * @return SelectionType value corresponding to this token
	 */
	public static final SelectionType getSelectionType(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("ALL")
				|| token.equals("0"))
			return SelectionType.ALL;
		else if (token.equalsIgnoreCase("LEAVES_ONLY")
				|| token.equals("1"))
			return SelectionType.LEAVES_ONLY;
		else
			return SelectionType.ALL;
	}


	/* Attributes */

	/**
	 * Identifier of the tree.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyTree			_template;

	/**
	 * Label of the tree.
	 */
	protected String			_label;

	/**
	 * Reference of the controller that manages the tree (i.e. handles user demands and
	 * refresh content).
	 */
	protected String			_controller;

	/**
	 * Selection mode of the nodes of the tree. The following modes are managed:
	 * NONE No selection is possible,
	 * SINGLE Selection of only one node at a time,
	 * MULTIPLESelection of several nodes at a time,
	 * SENSITIVE Selection of only a node at a time with selection of hyperlink type
	 */
	protected SelectionMode			_selectionMode;

	/**
	 * Selection typr of the nodes of the tree. The following types are managed:
	 * ALL All the nodes are selectable,
	 * LEAVES_ONLY Only leaf nodes are selectable
	 */
	protected SelectionType			_selectionType;

	/**
	 * Indicates if the tree is the context of an other element.
	 */
	protected boolean			_isContext;

	/**
	 * Gives the last node used in the tree so as to keep it visible during the next 
	 * displaying.
	 * Used internally for scrolling problems.
	 */
	protected String			_activeNode;

	/**
	 * Font of the tree.
	 */
	protected String			_font;

	/**
	 * Background color of the tree.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * Color used for selected nodes.
	 */
	protected String			_selectedColor;

	/**
	 * Color used for labels of selected nodes.
	 */
	protected String			_selectedFontColor;


	/* Sub components */

	/**
	 * Root node of the tree.
	 */
	protected LyTreeNode		_treeNode;

	/**
	 * Positioning constraint of the tree.
	 */
	protected LyConstraint		_constraint;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_controller = null;
		_selectionMode = SelectionMode.SINGLE;
		_selectionType = SelectionType.ALL;
		_isContext = getBooleanValue("TREE.isContext", false);
		_activeNode = null;
		_font = getStringValue("TREE.font", "$ALL.font");
		_color = getStringValue("TREE.color", "$ALL.color");
		_fontColor = getStringValue("TREE.fontColor", "$ALL.fontColor");
		_selectedColor = getStringValue("TREE.selectedColor", "$ALL.selectedColor");
		_selectedFontColor = getStringValue("TREE.selectedFontColor", "$ALL.selectedFontColor");

		// Initialize sub components
		_treeNode = null;
		_constraint = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyTree	model = (LyTree)component;

		_model = component;
		_label = model._label;
		_controller = model._controller;
		_selectionMode = model._selectionMode;
		_selectionType = model._selectionType;
		_isContext = model._isContext;
		_activeNode = model._activeNode;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;
		_selectedColor = model._selectedColor;
		_selectedFontColor = model._selectedFontColor;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		if (model._treeNode != null)
			addChild(model._treeNode.cloneComponent());
		if (model._constraint != null)
			addChild(model._constraint.cloneComponent());
		if (model._datas != null)
			_datas = new Hashtable<String, String>(model._datas);
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	protected LyTree()
	{
		_id = getInternalId();
	}

	/**
	 * Constructor with all required attributes.
	 * @param id Identifier of the tree.
	 */
	public LyTree(String id)
	{
		// Fill up required attributes with parameters
		_id = (id != null ? id.intern() : null);
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyTree(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyTree)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("controller"))
			_controller = node.getString("controller");
		if (node.isSpecified("selectionMode"))
			_selectionMode = getSelectionMode(node.getString("selectionMode"));
		if (node.isSpecified("selectionType"))
			_selectionType = getSelectionType(node.getString("selectionType"));
		if (node.isSpecified("isContext"))
			_isContext = node.getBoolean("isContext");
		if (node.isSpecified("activeNode"))
			_activeNode = node.getString("activeNode");
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("selectedColor"))
			_selectedColor = node.getString("selectedColor");
		if (node.isSpecified("selectedFontColor"))
			_selectedFontColor = node.getString("selectedFontColor");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyTreeNode)
			setTreeNode((LyTreeNode)child);
		else if (child instanceof LyConstraint)
			setConstraint((LyConstraint)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyTree clone = new LyTree();

		clone.initFromModel(this);
		clone._template = this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyTree getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of controller.
	 * @return value of controller.
	 */
	public String getController()
	{
		return _controller;
	}

	/**
	 * Gets value of selectionMode.
	 * @return value of selectionMode.
	 */
	public SelectionMode getSelectionMode()
	{
		return _selectionMode;
	}

	/**
	 * Gets value of selectionType.
	 * @return value of selectionType.
	 */
	public SelectionType getSelectionType()
	{
		return _selectionType;
	}

	/**
	 * Checks if value of attribute isContext is set.
	 * @return true if isContext is set, false otherwise.
	 */
	public boolean isIsContext()
	{
		return _isContext;
	}

	/**
	 * Gets value of activeNode.
	 * @return value of activeNode.
	 */
	public String getActiveNode()
	{
		return _activeNode;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("TREE.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("TREE.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("TREE.fontColor"));
	}

	/**
	 * Gets value of selectedColor.
	 * @return value of selectedColor.
	 */
	public String getSelectedColor()
	{
		return _selectedColor;
	}

	/**
	 * Checks whether the selectedColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultSelectedColor()
	{
		return (_selectedColor == getStringValue("TREE.selectedColor"));
	}

	/**
	 * Gets value of selectedFontColor.
	 * @return value of selectedFontColor.
	 */
	public String getSelectedFontColor()
	{
		return _selectedFontColor;
	}

	/**
	 * Checks whether the selectedFontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultSelectedFontColor()
	{
		return (_selectedFontColor == getStringValue("TREE.selectedFontColor"));
	}

	/**
	 * Get sub-component TreeNode.
	 * @return sub-component TreeNode.
	 */
	public LyTreeNode getTreeNode()
	{
		return _treeNode;
	}

	/**
	 * Get sub-component Constraint.
	 * @return sub-component Constraint.
	 */
	public LyConstraint getConstraint()
	{
		return _constraint;
	}


	/* Modifiers */

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyTree template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView != null)
			((LyGraphicalTree)_internalView).setLabel(label);
	}

	/**
	 * Initializes attribute controller to the given value.
	 * @param controller New value of controller.
	 */
	public void setController(String controller)
	{
		if ((controller == _controller)
			|| ((controller != null)
				&& (_controller != null)
				&& controller.equals(_controller)))
			return;

		_controller = controller;
	}

	/**
	 * Initializes attribute selectionMode to the given value.
	 * @param selectionMode New value of selectionMode.
	 * @exception IllegalArgumentException if selectionMode parameter is null.
	 */
	public void setSelectionMode(SelectionMode selectionMode)
	{
		if (selectionMode == _selectionMode)
			return;
		if (selectionMode == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute selectionMode");

		_selectionMode = selectionMode;
	}

	/**
	 * Initializes attribute selectionType to the given value.
	 * @param selectionType New value of selectionType.
	 * @exception IllegalArgumentException if selectionType parameter is null.
	 */
	public void setSelectionType(SelectionType selectionType)
	{
		if (selectionType == _selectionType)
			return;
		if (selectionType == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute selectionType");

		_selectionType = selectionType;
	}

	/**
	 * Initializes attribute isContext to the given value.
	 * @param isContext Boolean indicating if attribute isContext is set or unset.
	 */
	public void setIsContext(boolean isContext)
	{
		if (isContext == _isContext)
			return;
		_isContext = isContext;
	}

	/**
	 * Sets attribute activeNode to the given value.
	 * @param activeNode New value of activeNode.
	 */
	public void setActiveNode(String activeNode)
	{
		if ((activeNode == _activeNode)
			|| ((activeNode != null)
				&& (_activeNode != null)
				&& activeNode.equals(_activeNode)))
			return;

		_activeNode = activeNode;
		if (_internalView != null)
			((LyGraphicalTree)_internalView).setActiveNode(activeNode);
	}

	/**
	 * Initializes attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
	}

	/**
	 * Initializes attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Initializes attribute selectedColor to the given value.
	 * @param selectedColor New value of selectedColor.
	 */
	public void setSelectedColor(String selectedColor)
	{
		if ((selectedColor == _selectedColor)
			|| ((selectedColor != null)
				&& (_selectedColor != null)
				&& selectedColor.equals(_selectedColor)))
			return;

		_selectedColor = selectedColor;
	}

	/**
	 * Initializes attribute selectedFontColor to the given value.
	 * @param selectedFontColor New value of selectedFontColor.
	 */
	public void setSelectedFontColor(String selectedFontColor)
	{
		if ((selectedFontColor == _selectedFontColor)
			|| ((selectedFontColor != null)
				&& (_selectedFontColor != null)
				&& selectedFontColor.equals(_selectedFontColor)))
			return;

		_selectedFontColor = selectedFontColor;
	}

	/**
	 * Sets sub-component treeNode.
	 * @param treeNode New value of treeNode
	 */
	public void setTreeNode(LyTreeNode treeNode)
	{
		if (_treeNode != null)
			_treeNode.setParent(null);
		_treeNode = treeNode;
		if (treeNode != null)
			treeNode.setParent(this);
		if (_internalView != null)
			((LyGraphicalTree)_internalView).setTreeNode(treeNode);
	}

	/**
	 * Sets sub-component constraint.
	 * @param constraint New value of constraint
	 */
	public void setConstraint(LyConstraint constraint)
	{
		if (_constraint != null)
			_constraint.setParent(null);
		_constraint = constraint;
		if (constraint != null)
			constraint.setParent(this);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<TREE");
		pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlString(_controller) + "'");
		if ((_template == null)
			|| (_selectionMode != _template._selectionMode))
		{
			if (_selectionMode == SelectionMode.NONE)
				pw.print(" selectionMode='NONE'");
			else if ((_selectionMode == SelectionMode.SINGLE) && (_template != null))
				pw.print(" selectionMode='SINGLE'");
			else if (_selectionMode == SelectionMode.MULTIPLE)
				pw.print(" selectionMode='MULTIPLE'");
			else if (_selectionMode == SelectionMode.SENSITIVE)
				pw.print(" selectionMode='SENSITIVE'");
			else if (_selectionMode == SelectionMode.CHECKBOX)
				pw.print(" selectionMode='CHECKBOX'");
		}
		if ((_template == null)
			|| (_selectionType != _template._selectionType))
		{
			if ((_selectionType == SelectionType.ALL) && (_template != null))
				pw.print(" selectionType='ALL'");
			else if (_selectionType == SelectionType.LEAVES_ONLY)
				pw.print(" selectionType='LEAVES_ONLY'");
		}
		if (((_template == null) && _isContext)
				|| ((_template != null) && (_isContext != _template._isContext)))
			pw.print(" isContext='" + _isContext + "'");
		if (((_template == null) && (_activeNode != null))
				|| ((_template != null) && (_activeNode != _template._activeNode)))
			pw.print(" activeNode='" + LyString.getXmlString(_activeNode) + "'");
		if (((_template == null) && (_font != getStringValue("TREE.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("TREE.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("TREE.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && (_selectedColor != getStringValue("TREE.selectedColor", "$ALL.selectedColor")))
				|| ((_template != null) && (_selectedColor != _template._selectedColor)))
			pw.print(" selectedColor='" + LyString.getXmlString(_selectedColor) + "'");
		if (((_template == null) && (_selectedFontColor != getStringValue("TREE.selectedFontColor", "$ALL.selectedFontColor")))
				|| ((_template != null) && (_selectedFontColor != _template._selectedFontColor)))
			pw.print(" selectedFontColor='" + LyString.getXmlString(_selectedFontColor) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		if (_treeNode != null)
			_treeNode.writeXml(pw, debug);
		if (_constraint != null)
			_constraint.writeXml(pw, debug);

		writeDatas(pw);

		pw.println("</TREE>");
	}
}
