/*
 * LyTreeNode.java
 * Copyright (c) 2000-2010, Lyria-W4.
 * All rights reserved.
 */

package leon.view;

import java.util.*;
import java.io.*;
import leon.view.graphic.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.misc.LyString;


/**
 * The LyTreeNode class is a generated Java implementation that corresponds to the element 
 * TREE_NODE of the descriptive grammar of the views (views.dtd).
 * <B>&lt;TREE_NODE&gt; : Tree node</B>
 * A tree node can contain a set of nodes (this node is then called parent node of 
 * these last ones).
 * It is possible to assign a label and several images to a tree node, and to 
 * indicate if it is selected, valid or open. It is also possible to indicate the 
 * font and the different colors of the node.
 * @see leon.view.LyComponent
 * See also tag <A href="../../conf/dtd/view/TREE_NODE.html">TREE_NODE</A> of grammar leon_views.dtd.
 */
public class LyTreeNode extends LyComponent
	implements  LyGetIdInterface,
				LyGetLabelInterface,
				LySetLabelInterface,
				LySetSelectedInterface,
				LySetHiddenInterface,
				LySetImageInterface,
				LySetValidityInterface,
				LySetFontInterface,
				LySetColorInterface
{
	public static final String			__VERSION = "$Id$";


	/** Possible values for attribute 'validity'. */
	public enum Validity
	{
		VALID,
		INVALID,
		PARTIALLY;
	}


	/* Static methods */

	/**
	 * Utility method to get validity value from a string.
	 * @param token Token string
	 * @return Validity value corresponding to this token
	 */
	public static final Validity getValidity(String token)
	{
		if (token == null)
			return null;
		else if (token.equalsIgnoreCase("VALID")
				|| token.equals("0"))
			return Validity.VALID;
		else if (token.equalsIgnoreCase("INVALID")
				|| token.equals("1"))
			return Validity.INVALID;
		else if (token.equalsIgnoreCase("PARTIALLY")
				|| token.equals("2"))
			return Validity.PARTIALLY;
		else
			return Validity.VALID;
	}


	/* Attributes */

	/**
	 * Identifier of the node of the tree.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyTreeNode			_template;

	/**
	 * Label of the node of the tree.
	 */
	protected String			_label;

	/**
	 * Tooltip of a given node.
	 */
	protected String			_tooltip;

	/**
	 * Indicates if the node is selected.
	 * This attribute is not used for the NONE selection mode.
	 */
	protected boolean			_selected;

	/**
	 * Indicates if the node is open.
	 */
	protected boolean			_open;

	/**
	 * Indicates if the node is hidden.
	 */
	protected boolean			_hidden;

	/**
	 * Specifies if the tree node is expandable or not (can be opened).
	 */
	protected boolean			_expandable;

	/**
	 * Image associated to the node.
	 */
	protected String			_image;

	/**
	 * Indicates if the node is valid, invalid or partially valid.
	 */
	protected Validity			_validity;

	/**
	 * Font of the node.
	 */
	protected String			_font;

	/**
	 * Background color of the node.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * TREE_NODE.selectedBackgroundColor
	 */
	protected String			_selectedBackgroundColor;

	/**
	 * TREE_NODE.inactiveSelectedBackgroundColor
	 */
	protected String			_inactiveSelectedBackgroundColor;

	/**
	 * TREE_NODE.selectedFontColor
	 */
	protected String			_selectedFontColor;

	/**
	 * TREE_NODE.emptyCheckboxImage
	 */
	protected String			_emptyCheckboxImage;

	/**
	 * TREE_NODE.selectedCheckboxImage
	 */
	protected String			_selectedCheckboxImage;

	/**
	 * TREE_NODE.openDirImage
	 */
	protected String			_openDirImage;

	/**
	 * TREE_NODE.closedDirImage
	 */
	protected String			_closedDirImage;


	/* Sub components */

	/**
	 * Images of a tree node.
	 */
	protected ArrayList<LyImage>		_images;

	/**
	 * Sub-nodes of the tree node.
	 */
	protected ArrayList<LyTreeNode>		_treeNodes;


	/* Initialization */

	/**
	 * Initialization of the data structure.
	 * @see leon.view.LyComponent#init
	 */
	protected void init()
	{
		super.init();

		// Initialize attributes
		_id = null;
		_template = null;
		_label = null;
		_tooltip = null;
		_selected = getBooleanValue("TREE_NODE.selected", false);
		_open = getBooleanValue("TREE_NODE.open", false);
		_hidden = getBooleanValue("TREE_NODE.hidden", false);
		_expandable = getBooleanValue("TREE_NODE.expandable", false);
		_image = null;
		_validity = Validity.VALID;
		_font = getStringValue("TREE_NODE.font", "$ALL.font");
		_color = getStringValue("TREE_NODE.color", "$ALL.color");
		_fontColor = getStringValue("TREE_NODE.fontColor", "$ALL.fontColor");
		_selectedBackgroundColor = getStringValue("TREE_NODE.selectedBackgroundColor", "$ALL.selectedBackgroundColor");
		_inactiveSelectedBackgroundColor = getStringValue("TREE_NODE.inactiveSelectedBackgroundColor", "$ALL.inactiveSelectedBackgroundColor");
		_selectedFontColor = getStringValue("TREE_NODE.selectedFontColor", "$ALL.selectedFontColor");
		_emptyCheckboxImage = getStringValue("TREE_NODE.emptyCheckboxImage", "$ALL.emptyCheckboxImage");
		_selectedCheckboxImage = getStringValue("TREE_NODE.selectedCheckboxImage", "$ALL.selectedCheckboxImage");
		_openDirImage = getStringValue("TREE_NODE.openDirImage", "$ALL.openDirImage");
		_closedDirImage = getStringValue("TREE_NODE.closedDirImage", "$ALL.closedDirImage");

		// Initialize sub components
		_images = null;
		_treeNodes = null;
	}

	/**
	 * Method for initializing a component from a model.
	 */
	protected void initFromModel(LyComponent component)
	{
		if ((component == null)
				|| (_model != null)) // Already initialized from its model
			return;

		LyTreeNode	model = (LyTreeNode)component;

		_model = component;
		_label = model._label;
		_tooltip = model._tooltip;
		_selected = model._selected;
		_open = model._open;
		_hidden = model._hidden;
		_expandable = model._expandable;
		_image = model._image;
		_validity = model._validity;
		_font = model._font;
		_color = model._color;
		_fontColor = model._fontColor;
		_selectedBackgroundColor = model._selectedBackgroundColor;
		_inactiveSelectedBackgroundColor = model._inactiveSelectedBackgroundColor;
		_selectedFontColor = model._selectedFontColor;
		_emptyCheckboxImage = model._emptyCheckboxImage;
		_selectedCheckboxImage = model._selectedCheckboxImage;
		_openDirImage = model._openDirImage;
		_closedDirImage = model._closedDirImage;

		if (REWRITING_MODE)
			return;

		// Copy sub-components
		LyImage	images;
		int			i, n;

		n = (model._images == null ? 0 : model._images.size());
		for (i = 0; i < n; i++)
		{
			images = model._images.get(i);
			if (images != null)
				images = (LyImage)images.cloneComponent();
			addChild(images);
		}

		LyTreeNode	treeNodes;

		n = (model._treeNodes == null ? 0 : model._treeNodes.size());
		for (i = 0; i < n; i++)
		{
			treeNodes = model._treeNodes.get(i);
			if (treeNodes != null)
				treeNodes = (LyTreeNode)treeNodes.cloneComponent();
			addChild(treeNodes);
		}
	}


	/* Constructors */

	/**
	 * Default constructor.
	 */
	public LyTreeNode()
	{
	}


	/* Protected method for construction from parser  */

	/**
	 * Construction from a LyViewNode.
	 * @param node	Node from which the component is constructed.
	 * @see leon.view.LyViewNodeFactory#buildElement
	 */
	protected LyTreeNode(LyViewNode node)
	{
		initFromViewNode(node);
	}

	/**
	 * Inits from a LyViewNode.
	 * @param node    Node from which the component is initialized.
	 */
	protected void initFromViewNode(LyViewNode node)
	{
		// Retrieve attributes from node attributes
		_id = node.getString("id");
		_template = (LyTreeNode)node.getComponent("template");

		if (_template != null)
			initFromModel(_template);

		if (node.isSpecified("label"))
			_label = node.getString("label");
		if (node.isSpecified("tooltip"))
			_tooltip = node.getString("tooltip");
		if (node.isSpecified("selected"))
			_selected = node.getBoolean("selected");
		if (node.isSpecified("open"))
			_open = node.getBoolean("open");
		if (node.isSpecified("hidden"))
			_hidden = node.getBoolean("hidden");
		if (node.isSpecified("expandable"))
			_expandable = node.getBoolean("expandable");
		if (node.isSpecified("image"))
			_image = node.getString("image");
		if (node.isSpecified("validity"))
			_validity = getValidity(node.getString("validity"));
		if (node.isSpecified("font"))
			_font = node.getString("font");
		if (node.isSpecified("color"))
			_color = node.getString("color");
		if (node.isSpecified("fontColor"))
			_fontColor = node.getString("fontColor");
		if (node.isSpecified("selectedBackgroundColor"))
			_selectedBackgroundColor = node.getString("selectedBackgroundColor");
		if (node.isSpecified("inactiveSelectedBackgroundColor"))
			_inactiveSelectedBackgroundColor = node.getString("inactiveSelectedBackgroundColor");
		if (node.isSpecified("selectedFontColor"))
			_selectedFontColor = node.getString("selectedFontColor");
		if (node.isSpecified("emptyCheckboxImage"))
			_emptyCheckboxImage = node.getString("emptyCheckboxImage");
		if (node.isSpecified("selectedCheckboxImage"))
			_selectedCheckboxImage = node.getString("selectedCheckboxImage");
		if (node.isSpecified("openDirImage"))
			_openDirImage = node.getString("openDirImage");
		if (node.isSpecified("closedDirImage"))
			_closedDirImage = node.getString("closedDirImage");
	}

	/**
	 * End of construction : completion with child component.
	 * @param child Sub child component.
	 * @see leon.view.LyComponent#addChild
	 */
	protected void addChild(LyComponent child)
	{
		// Add sub-components
		if (child instanceof LyImage)
			addImage((LyImage)child);
		else if (child instanceof LyTreeNode)
			addTreeNode((LyTreeNode)child);
	}

	/**
	 * Method for cloning a component.
	 */
	public LyComponent cloneComponent()
	{
		LyTreeNode clone = new LyTreeNode();

		clone.initFromModel(this);
		clone._template = this;
		return clone;
	}


	/* Accessors */

	/**
	 * Gets value of id.
	 * @return value of id.
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * Gets value of template.
	 * @return value of template.
	 */
	public LyTreeNode getTemplate()
	{
		return _template;
	}

	/**
	 * Gets value of label.
	 * @return value of label.
	 */
	public String getLabel()
	{
		return _label;
	}

	/**
	 * Gets value of tooltip.
	 * @return value of tooltip.
	 */
	public String getTooltip()
	{
		return _tooltip;
	}

	/**
	 * Checks if value of attribute selected is set.
	 * @return true if selected is set, false otherwise.
	 */
	public boolean isSelected()
	{
		return _selected;
	}

	/**
	 * Checks if value of attribute open is set.
	 * @return true if open is set, false otherwise.
	 */
	public boolean isOpen()
	{
		return _open;
	}

	/**
	 * Checks if value of attribute hidden is set.
	 * @return true if hidden is set, false otherwise.
	 */
	public boolean isHidden()
	{
		return _hidden;
	}

	/**
	 * Checks if value of attribute expandable is set.
	 * @return true if expandable is set, false otherwise.
	 */
	public boolean isExpandable()
	{
		return _expandable;
	}

	/**
	 * Gets value of image.
	 * @return value of image.
	 */
	public String getImage()
	{
		return _image;
	}

	/**
	 * Gets value of validity.
	 * @return value of validity.
	 */
	public Validity getValidity()
	{
		return _validity;
	}

	/**
	 * Gets value of font.
	 * @return value of font.
	 */
	public String getFont()
	{
		return _font;
	}

	/**
	 * Checks whether the font attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFont()
	{
		return (_font == getStringValue("TREE_NODE.font"));
	}

	/**
	 * Gets value of color.
	 * @return value of color.
	 */
	public String getColor()
	{
		return _color;
	}

	/**
	 * Checks whether the color attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultColor()
	{
		return (_color == getStringValue("TREE_NODE.color"));
	}

	/**
	 * Gets value of fontColor.
	 * @return value of fontColor.
	 */
	public String getFontColor()
	{
		return _fontColor;
	}

	/**
	 * Checks whether the fontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultFontColor()
	{
		return (_fontColor == getStringValue("TREE_NODE.fontColor"));
	}

	/**
	 * Gets value of selectedBackgroundColor.
	 * @return value of selectedBackgroundColor.
	 */
	public String getSelectedBackgroundColor()
	{
		return _selectedBackgroundColor;
	}

	/**
	 * Checks whether the selectedBackgroundColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultSelectedBackgroundColor()
	{
		return (_selectedBackgroundColor == getStringValue("TREE_NODE.selectedBackgroundColor"));
	}

	/**
	 * Gets value of inactiveSelectedBackgroundColor.
	 * @return value of inactiveSelectedBackgroundColor.
	 */
	public String getInactiveSelectedBackgroundColor()
	{
		return _inactiveSelectedBackgroundColor;
	}

	/**
	 * Checks whether the inactiveSelectedBackgroundColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultInactiveSelectedBackgroundColor()
	{
		return (_inactiveSelectedBackgroundColor == getStringValue("TREE_NODE.inactiveSelectedBackgroundColor"));
	}

	/**
	 * Gets value of selectedFontColor.
	 * @return value of selectedFontColor.
	 */
	public String getSelectedFontColor()
	{
		return _selectedFontColor;
	}

	/**
	 * Checks whether the selectedFontColor attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultSelectedFontColor()
	{
		return (_selectedFontColor == getStringValue("TREE_NODE.selectedFontColor"));
	}

	/**
	 * Gets value of emptyCheckboxImage.
	 * @return value of emptyCheckboxImage.
	 */
	public String getEmptyCheckboxImage()
	{
		return _emptyCheckboxImage;
	}

	/**
	 * Checks whether the emptyCheckboxImage attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultEmptyCheckboxImage()
	{
		return (_emptyCheckboxImage == getStringValue("TREE_NODE.emptyCheckboxImage"));
	}

	/**
	 * Gets value of selectedCheckboxImage.
	 * @return value of selectedCheckboxImage.
	 */
	public String getSelectedCheckboxImage()
	{
		return _selectedCheckboxImage;
	}

	/**
	 * Checks whether the selectedCheckboxImage attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultSelectedCheckboxImage()
	{
		return (_selectedCheckboxImage == getStringValue("TREE_NODE.selectedCheckboxImage"));
	}

	/**
	 * Gets value of openDirImage.
	 * @return value of openDirImage.
	 */
	public String getOpenDirImage()
	{
		return _openDirImage;
	}

	/**
	 * Checks whether the openDirImage attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultOpenDirImage()
	{
		return (_openDirImage == getStringValue("TREE_NODE.openDirImage"));
	}

	/**
	 * Gets value of closedDirImage.
	 * @return value of closedDirImage.
	 */
	public String getClosedDirImage()
	{
		return _closedDirImage;
	}

	/**
	 * Checks whether the closedDirImage attribute is valuated to the default value.
	 * @return True if default, false otherwise.
	 */
	public boolean isDefaultClosedDirImage()
	{
		return (_closedDirImage == getStringValue("TREE_NODE.closedDirImage"));
	}

	/**
	 * Get sub-component Image of index i.
	 * @param i the index of the Image to get.
	 * @return sub-component Image of index i.
	 */
	public LyImage getImage(int i)
	{
		if (_images == null)
			return null;
		if ((i < 0) || (i >= _images.size()))
			return null;
		return _images.get(i);
	}

	/**
	 * Get sub-component Image count.
	 * @return sub-component Image count.
	 */
	public int getImageCount()
	{
		if (_images == null)
			return 0;
		return _images.size();
	}

	/**
	 * Get sub-component TreeNode of index i.
	 * @param i the index of the TreeNode to get.
	 * @return sub-component TreeNode of index i.
	 */
	public LyTreeNode getTreeNode(int i)
	{
		if (_treeNodes == null)
			return null;
		if ((i < 0) || (i >= _treeNodes.size()))
			return null;
		return _treeNodes.get(i);
	}

	/**
	 * Get sub-component TreeNode count.
	 * @return sub-component TreeNode count.
	 */
	public int getTreeNodeCount()
	{
		if (_treeNodes == null)
			return 0;
		return _treeNodes.size();
	}


	/* Modifiers */

	/**
	 * Initializes attribute id to the given value.
	 * @param id New value of id.
	 */
	public void setId(String id)
	{
		if ((id == _id)
			|| ((id != null)
				&& (_id != null)
				&& id.equals(_id)))
			return;

		_id = (id != null ? id.intern() : null);
	}

	/**
	 * Initializes attribute template to the given value.
	 * @param template New value of template.
	 */
	public void setTemplate(LyTreeNode template)
	{
		if (template == null)
			return;
		if (_template != null)
			throw new IllegalStateException("BF AppEngine : attribute template was already set.");

		initFromModel(template);
		_template = template;
	}

	/**
	 * Sets attribute label to the given value.
	 * @param label New value of label.
	 */
	public void setLabel(String label)
	{
		if ((label == _label)
			|| ((label != null)
				&& (_label != null)
				&& label.equals(_label)))
			return;

		_label = label;
		if (_internalView != null)
			((LyGraphicalTreeNode)_internalView).setLabel(label);
	}

	/**
	 * Sets attribute tooltip to the given value.
	 * @param tooltip New value of tooltip.
	 */
	public void setTooltip(String tooltip)
	{
		if ((tooltip == _tooltip)
			|| ((tooltip != null)
				&& (_tooltip != null)
				&& tooltip.equals(_tooltip)))
			return;

		_tooltip = tooltip;
		if (_internalView != null)
			((LyGraphicalTreeNode)_internalView).setTooltip(tooltip);
	}

	/**
	 * Sets attribute selected to the given value.
	 * @param selected Boolean indicating if attribute selected is set or unset.
	 */
	public void setSelected(boolean selected)
	{
		if (selected == _selected)
			return;
		_selected = selected;
		if (_internalView != null)
			((LyGraphicalTreeNode)_internalView).setSelected(selected);
	}

	/**
	 * Sets attribute open to the given value.
	 * @param open Boolean indicating if attribute open is set or unset.
	 */
	public void setOpen(boolean open)
	{
		if (open == _open)
			return;
		_open = open;
		if (_internalView != null)
			((LyGraphicalTreeNode)_internalView).setOpen(open);
	}

	/**
	 * Sets attribute hidden to the given value.
	 * @param hidden Boolean indicating if attribute hidden is set or unset.
	 */
	public void setHidden(boolean hidden)
	{
		if (hidden == _hidden)
			return;
		_hidden = hidden;
		if (_internalView != null)
			((LyGraphicalTreeNode)_internalView).setHidden(hidden);
	}

	/**
	 * Sets attribute expandable to the given value.
	 * @param expandable Boolean indicating if attribute expandable is set or unset.
	 */
	public void setExpandable(boolean expandable)
	{
		if (expandable == _expandable)
			return;
		_expandable = expandable;
		if (_internalView != null)
			((LyGraphicalTreeNode)_internalView).setExpandable(expandable);
	}

	/**
	 * Sets attribute image to the given value.
	 * @param image New value of image.
	 */
	public void setImage(String image)
	{
		if ((image == _image)
			|| ((image != null)
				&& (_image != null)
				&& image.equals(_image)))
			return;

		_image = image;
		if (_internalView != null)
			((LyGraphicalTreeNode)_internalView).setImage(image);
	}

	/**
	 * Sets attribute validity to the given value.
	 * @param validity New value of validity.
	 * @exception IllegalArgumentException if validity parameter is null.
	 */
	public void setValidity(Validity validity)
	{
		if (validity == _validity)
			return;
		if (validity == null)
			throw new IllegalArgumentException("BF AppEngine : invalid value for attribute validity");

		_validity = validity;
		if (_internalView != null)
			((LyGraphicalTreeNode)_internalView).setValidity(validity);
	}

	/**
	 * Sets attribute font to the given value.
	 * @param font New value of font.
	 */
	public void setFont(String font)
	{
		if ((font == _font)
			|| ((font != null)
				&& (_font != null)
				&& font.equals(_font)))
			return;

		_font = font;
		if (_internalView != null)
			((LyGraphicalTreeNode)_internalView).setFont(font);
	}

	/**
	 * Sets attribute color to the given value.
	 * @param color New value of color.
	 */
	public void setColor(String color)
	{
		if ((color == _color)
			|| ((color != null)
				&& (_color != null)
				&& color.equals(_color)))
			return;

		_color = color;
		if (_internalView != null)
			((LyGraphicalTreeNode)_internalView).setColor(color);
	}

	/**
	 * Initializes attribute fontColor to the given value.
	 * @param fontColor New value of fontColor.
	 */
	public void setFontColor(String fontColor)
	{
		if ((fontColor == _fontColor)
			|| ((fontColor != null)
				&& (_fontColor != null)
				&& fontColor.equals(_fontColor)))
			return;

		_fontColor = fontColor;
	}

	/**
	 * Initializes attribute selectedBackgroundColor to the given value.
	 * @param selectedBackgroundColor New value of selectedBackgroundColor.
	 */
	public void setSelectedBackgroundColor(String selectedBackgroundColor)
	{
		if ((selectedBackgroundColor == _selectedBackgroundColor)
			|| ((selectedBackgroundColor != null)
				&& (_selectedBackgroundColor != null)
				&& selectedBackgroundColor.equals(_selectedBackgroundColor)))
			return;

		_selectedBackgroundColor = selectedBackgroundColor;
	}

	/**
	 * Initializes attribute inactiveSelectedBackgroundColor to the given value.
	 * @param inactiveSelectedBackgroundColor New value of inactiveSelectedBackgroundColor.
	 */
	public void setInactiveSelectedBackgroundColor(String inactiveSelectedBackgroundColor)
	{
		if ((inactiveSelectedBackgroundColor == _inactiveSelectedBackgroundColor)
			|| ((inactiveSelectedBackgroundColor != null)
				&& (_inactiveSelectedBackgroundColor != null)
				&& inactiveSelectedBackgroundColor.equals(_inactiveSelectedBackgroundColor)))
			return;

		_inactiveSelectedBackgroundColor = inactiveSelectedBackgroundColor;
	}

	/**
	 * Initializes attribute selectedFontColor to the given value.
	 * @param selectedFontColor New value of selectedFontColor.
	 */
	public void setSelectedFontColor(String selectedFontColor)
	{
		if ((selectedFontColor == _selectedFontColor)
			|| ((selectedFontColor != null)
				&& (_selectedFontColor != null)
				&& selectedFontColor.equals(_selectedFontColor)))
			return;

		_selectedFontColor = selectedFontColor;
	}

	/**
	 * Initializes attribute emptyCheckboxImage to the given value.
	 * @param emptyCheckboxImage New value of emptyCheckboxImage.
	 */
	public void setEmptyCheckboxImage(String emptyCheckboxImage)
	{
		if ((emptyCheckboxImage == _emptyCheckboxImage)
			|| ((emptyCheckboxImage != null)
				&& (_emptyCheckboxImage != null)
				&& emptyCheckboxImage.equals(_emptyCheckboxImage)))
			return;

		_emptyCheckboxImage = emptyCheckboxImage;
	}

	/**
	 * Initializes attribute selectedCheckboxImage to the given value.
	 * @param selectedCheckboxImage New value of selectedCheckboxImage.
	 */
	public void setSelectedCheckboxImage(String selectedCheckboxImage)
	{
		if ((selectedCheckboxImage == _selectedCheckboxImage)
			|| ((selectedCheckboxImage != null)
				&& (_selectedCheckboxImage != null)
				&& selectedCheckboxImage.equals(_selectedCheckboxImage)))
			return;

		_selectedCheckboxImage = selectedCheckboxImage;
	}

	/**
	 * Initializes attribute openDirImage to the given value.
	 * @param openDirImage New value of openDirImage.
	 */
	public void setOpenDirImage(String openDirImage)
	{
		if ((openDirImage == _openDirImage)
			|| ((openDirImage != null)
				&& (_openDirImage != null)
				&& openDirImage.equals(_openDirImage)))
			return;

		_openDirImage = openDirImage;
	}

	/**
	 * Initializes attribute closedDirImage to the given value.
	 * @param closedDirImage New value of closedDirImage.
	 */
	public void setClosedDirImage(String closedDirImage)
	{
		if ((closedDirImage == _closedDirImage)
			|| ((closedDirImage != null)
				&& (_closedDirImage != null)
				&& closedDirImage.equals(_closedDirImage)))
			return;

		_closedDirImage = closedDirImage;
	}

	/**
	 * Adds a sub-component image.
	 * @param image The image to add to images list.
	 */
	public void addImage(LyImage image)
	{
		insertImage(image, -1);
	}

	/**
	 * Inserts a sub-component image at given position.
	 * @param image The image to add to images list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertImage(LyImage image, int pos)
	{
		if (_images == null)
			_images = new ArrayList<LyImage>(1);
		else if (_images.contains(image))
			return;
		if (pos >= 0)
			_images.add(pos,image);
		else
			_images.add(image);
		if (image != null)
			image.setParent(this);
		if (_internalView != null)
			((LyGraphicalTreeNode)_internalView).addImage(image, pos);
	}

	/**
	 * Removes a sub-component image.
	 * @param image The image to remove to images list.
	 * @exception IllegalStateException if images list is empty.
	 * @exception IllegalArgumentException if images list doesn't contain image.
	 */
	public void removeImage(LyImage image)
	{
		if (_images == null)
			throw new IllegalStateException("BF AppEngine : images list is empty");
		if (!_images.contains(image))
			throw new IllegalArgumentException("BF AppEngine : images list doesn't contain image");

		if (_internalView != null)
			((LyGraphicalTreeNode)_internalView).removeImage(image);
		_images.remove(image);
		image.setParent(null);
	}

	/**
	 * Adds a sub-component treeNode.
	 * @param treeNode The treeNode to add to treeNodes list.
	 */
	public void addTreeNode(LyTreeNode treeNode)
	{
		insertTreeNode(treeNode, -1);
	}

	/**
	 * Inserts a sub-component treeNode at given position.
	 * @param treeNode The treeNode to add to treeNodes list.
	 * @param pos	Position where the sub-component is inserted.
	 */
	public void insertTreeNode(LyTreeNode treeNode, int pos)
	{
		if (_treeNodes == null)
		{
			_treeNodes = new ArrayList<LyTreeNode>(1);
			_children = _treeNodes;
		}
		else if (_treeNodes.contains(treeNode))
			return;
		if (pos >= 0)
			_treeNodes.add(pos,treeNode);
		if (treeNode != null)
			treeNode.setParent(this);
		if (_internalView != null)
			((LyGraphicalTreeNode)_internalView).addTreeNode(treeNode, pos);
	}

	/**
	 * Removes a sub-component treeNode.
	 * @param treeNode The treeNode to remove to treeNodes list.
	 * @exception IllegalStateException if treeNodes list is empty.
	 * @exception IllegalArgumentException if treeNodes list doesn't contain treeNode.
	 */
	public void removeTreeNode(LyTreeNode treeNode)
	{
		if (_treeNodes == null)
			throw new IllegalStateException("BF AppEngine : treeNodes list is empty");
		if (!_treeNodes.contains(treeNode))
			throw new IllegalArgumentException("BF AppEngine : treeNodes list doesn't contain treeNode");

		if (_internalView != null)
			((LyGraphicalTreeNode)_internalView).removeTreeNode(treeNode);
		treeNode.setParent(null);
	}


	/* XML Writing */

	/**
	 * Method to rewrite XML corresponding to this component.
	 * @param pw	the printWriter on which the XML is written.
	 * @param debug	true if debug mode (full dump).
	 */
	public void writeXml(PrintWriter pw, boolean debug)
	{
		// Write tag with its attributes
		pw.print("<TREE_NODE");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template._label)))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		if (((_template == null) && (_tooltip != null))
				|| ((_template != null) && (_tooltip != _template._tooltip)))
			pw.print(" tooltip='" + LyString.getXmlString(_tooltip) + "'");
		if (((_template == null) && _selected)
				|| ((_template != null) && (_selected != _template._selected)))
			pw.print(" selected='" + _selected + "'");
		if (((_template == null) && _open)
				|| ((_template != null) && (_open != _template._open)))
			pw.print(" open='" + _open + "'");
		if (((_template == null) && _hidden)
				|| ((_template != null) && (_hidden != _template._hidden)))
			pw.print(" hidden='" + _hidden + "'");
		if (((_template == null) && _expandable)
				|| ((_template != null) && (_expandable != _template._expandable)))
			pw.print(" expandable='" + _expandable + "'");
		if (((_template == null) && (_image != null))
				|| ((_template != null) && (_image != _template._image)))
			pw.print(" image='" + LyString.getXmlString(_image) + "'");
		if ((_template == null)
			|| (_validity != _template._validity))
		{
			if ((_validity == Validity.VALID) && (_template != null))
				pw.print(" validity='VALID'");
			else if (_validity == Validity.INVALID)
				pw.print(" validity='INVALID'");
			else if (_validity == Validity.PARTIALLY)
				pw.print(" validity='PARTIALLY'");
		}
		if (((_template == null) && (_font != getStringValue("TREE_NODE.font", "$ALL.font")))
				|| ((_template != null) && (_font != _template._font)))
			pw.print(" font='" + LyString.getXmlString(_font) + "'");
		if (((_template == null) && (_color != getStringValue("TREE_NODE.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlString(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("TREE_NODE.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlString(_fontColor) + "'");
		if (((_template == null) && (_selectedBackgroundColor != getStringValue("TREE_NODE.selectedBackgroundColor", "$ALL.selectedBackgroundColor")))
				|| ((_template != null) && (_selectedBackgroundColor != _template._selectedBackgroundColor)))
			pw.print(" selectedBackgroundColor='" + LyString.getXmlString(_selectedBackgroundColor) + "'");
		if (((_template == null) && (_inactiveSelectedBackgroundColor != getStringValue("TREE_NODE.inactiveSelectedBackgroundColor", "$ALL.inactiveSelectedBackgroundColor")))
				|| ((_template != null) && (_inactiveSelectedBackgroundColor != _template._inactiveSelectedBackgroundColor)))
			pw.print(" inactiveSelectedBackgroundColor='" + LyString.getXmlString(_inactiveSelectedBackgroundColor) + "'");
		if (((_template == null) && (_selectedFontColor != getStringValue("TREE_NODE.selectedFontColor", "$ALL.selectedFontColor")))
				|| ((_template != null) && (_selectedFontColor != _template._selectedFontColor)))
			pw.print(" selectedFontColor='" + LyString.getXmlString(_selectedFontColor) + "'");
		if (((_template == null) && (_emptyCheckboxImage != getStringValue("TREE_NODE.emptyCheckboxImage", "$ALL.emptyCheckboxImage")))
				|| ((_template != null) && (_emptyCheckboxImage != _template._emptyCheckboxImage)))
			pw.print(" emptyCheckboxImage='" + LyString.getXmlString(_emptyCheckboxImage) + "'");
		if (((_template == null) && (_selectedCheckboxImage != getStringValue("TREE_NODE.selectedCheckboxImage", "$ALL.selectedCheckboxImage")))
				|| ((_template != null) && (_selectedCheckboxImage != _template._selectedCheckboxImage)))
			pw.print(" selectedCheckboxImage='" + LyString.getXmlString(_selectedCheckboxImage) + "'");
		if (((_template == null) && (_openDirImage != getStringValue("TREE_NODE.openDirImage", "$ALL.openDirImage")))
				|| ((_template != null) && (_openDirImage != _template._openDirImage)))
			pw.print(" openDirImage='" + LyString.getXmlString(_openDirImage) + "'");
		if (((_template == null) && (_closedDirImage != getStringValue("TREE_NODE.closedDirImage", "$ALL.closedDirImage")))
				|| ((_template != null) && (_closedDirImage != _template._closedDirImage)))
			pw.print(" closedDirImage='" + LyString.getXmlString(_closedDirImage) + "'");

		if ((getChildCount() == 0) && (_datas == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	images;
		int			i, n;

		n = (_images == null ? 0 : _images.size());
		for (i = 0; i < n; i++)
		{
			images = _images.get(i);
			images.writeXml(pw, debug);
		}

		LyComponent	treeNodes;

		n = (_treeNodes == null ? 0 : _treeNodes.size());
		for (i = 0; i < n; i++)
		{
			treeNodes = _treeNodes.get(i);
			treeNodes.writeXml(pw, debug);
		}


		pw.println("</TREE_NODE>");
	}
}
