/*
 * LyView.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.view;

import leon.misc.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;

import java.io.*;
import java.util.*;


/**
 * This class implements the common description of a Frame object and a Dialog object as described
 * in the dtd.
 *
 * @version  $Id: LyView.java 35182 2012-04-10 13:54:06Z jlsaouli $
 * @author   Lyria
 * @see      LyComponent
 */
public abstract class LyView extends LyComponent
	implements LyGetIdInterface,
		LyGetControllerInterface,
		LySetVisibleInterface,
		LyGetBoundsInterface,
		LySetBoundsInterface,
		LySetWaitingInterface
{
	public static final String __VERSION = "$Id: LyView.java 35182 2012-04-10 13:54:06Z jlsaouli $";

	/** Number of views for debug. */
	public static int _nbViews;

	/** Current environment. */
	protected LyEnvironment _environment;

	/** Keep setVisible time to allow waiting status Workaround for display bug in setWaiting. */
	protected long _visibleTime;

	/** Minimum delay before allowing waiting status Workaround for display bug. */
	protected long MIN_VISIBLE_DELAY = 500; // ms

	/**
	 * Default constructor.
	 */
	public LyView()
	{
		_visibleTime = -1;

		_nbViews++;
	}

	/**
	 * Constructor.
	 */
	public LyView(LyEnvironment environment)
	{
		this();
		setEnvironment(environment);
	}

	protected void finalize()
	{
		super.finalize();
		_nbViews--;
	}

	/**
	 */
	public void setEnvironment(LyEnvironment environment)
	{
		if (_environment != null)
			throw new IllegalStateException(
				"BF AppEngine : attempt to set an environment on a view which has already an environment");

		_environment = environment;

		// Register on viewManager
		LyViewManager manager = getViewManager();
		
		if (manager != null)
			manager.addView(this);
	}

	/**
	 */
	public LyEnvironment getEnvironment()
	{
		return (_environment != null ? _environment : LyEnvironment.getCurrentEnvironment());
	}

	/**
	 * Shows or hides the component/node.
	 *
	 * @param  status  new visibility status
	 *
	 * @see    LySetVisibleInterface#setVisible
	 */
	public void setVisible(boolean status)
	{
		if (status)
			getInternalView(true);

		// Workaround for setvisible lock : even if the view
		// is visible, some component may need repaint yet.
		if (status)
			_visibleTime = new Date().getTime();
		else
			_visibleTime = -1;

		super.setVisible(status);
	}

	/**
	 * Checks if the component is visible or not.
	 *
	 * @return  true if the component is showing (a visible component is never freed()).
	 *
	 * @see     LySetVisibleInterface#setVisible
	 */
	public boolean isVisible()
	{
		Object internal = getInternalView(false);

		if (internal instanceof LyIsVisibleInterface)
			return ((LyIsVisibleInterface)internal).isVisible();
		else
			return false;
	}

	/**
	 * Accessor to the location.
	 *
	 * @see  LySetBoundsInterface#setLocation
	 */
	public void setLocation(int x, int y)
	{
		Object internal = getInternalView();

		if (internal instanceof LySetBoundsInterface)
			((LySetBoundsInterface)internal).setLocation(x, y);
	}

	/**
	 * Accessor to the Size.
	 *
	 * @see  LySetBoundsInterface#setSize
	 */
	public void setSize(int w, int h)
	{
		Object internal = getInternalView();

		if (internal instanceof LySetBoundsInterface)
			((LySetBoundsInterface)internal).setSize(w, h);
	}

	/**
	 * Accessor to the location.
	 *
	 * @see  LyGetBoundsInterface#getLocation
	 */
	public int[] getLocation()
	{
		Object internal = getInternalView(false);

		if (internal instanceof LyGetBoundsInterface)
			return ((LyGetBoundsInterface)internal).getLocation();
		else
			return null;
	}

	/**
	 * Accessor to the Size.
	 *
	 * @see  LyGetBoundsInterface#getSize
	 */
	public int[] getSize()
	{
		Object internal = getInternalView();

		if (internal instanceof LyGetBoundsInterface)
			return ((LyGetBoundsInterface)internal).getSize();
		else
			return null;
	}

	/**
	 * Update the waiting status of current view.
	 *
	 * @see  LySetWaitingInterface#setWaiting
	 */
	public void setWaiting(boolean status)
	{
		if (_visibleTime <= 0)
			return; // SKIP (workaround for display bug if not visible or partially visible)

		if (status)
		{
			// Compare to current (how long this component has been showing ?)
			long current = new Date().getTime();

			if ((current - _visibleTime) < MIN_VISIBLE_DELAY)
				return; // Just shown ! Skip
		}

		Object internal = getInternalView();

		if (internal instanceof LySetWaitingInterface)
			((LySetWaitingInterface)internal).setWaiting(status);
	}

	/**
	 * Free this view.
	 */
	public void free()
	{
		if (_freed)
			return;

		if (_environment != null)
		{
			// Unregister manager
			LyViewManager manager = getViewManager();

			if (manager != null)
				manager.removeView(this);
		}

		// Free this
		super.free();

		_environment = null;
	}

	/**
	 * Gets value of id.
	 *
	 * @return  value of id
	 */
	public abstract String getId();

	/**
	 * Gets value of controller.
	 *
	 * @return  value of controller
	 */
	public abstract String getController();

	/**
	 * Get sub-component Area.
	 *
	 * @return  sub-component Area
	 */
	public abstract LyArea getArea();

	/**
	 * Sets sub-component area.
	 *
	 * @param      area  New value of area
	 *
	 * @exception  IllegalArgumentException  if area is null
	 */
	public abstract void setArea(LyArea area);

	/**
	 * Change resizable state on this view.
	 */
	public abstract void setResizable(boolean resizable);

	/**
	 * Checks if value of attribute resizable is set.
	 *
	 * @return  true if resizable is set, false otherwise.
	 */
	public abstract boolean isResizable();

	/**
	 * Gets value of minWidth.
	 *
	 * @return  value of minWidth.
	 */
	public abstract int getMinWidth();

	/**
	 * Gets value of minHeight.
	 *
	 * @return  value of minHeight.
	 */
	public abstract int getMinHeight();

	/**
	 * Sets title of this view.
	 */
	public abstract void setLabel(String label);

	/**
	 * Write XML description of this view.
	 *
	 * @param  file      the path of the file
	 * @param  debug     True if view is dumped for debug.
	 * @param  fullView  True if the full view is dumped or only its content.
	 */
	public void writeXml(String file, boolean debug, boolean fullView)
	{
		try
		{
			if (file == null)
			{
				// If a file is not given, try to write to a file located in LY_APP_DIR
				String path = _environment.getEnv("LY_APP_DIR");

				if (_environment.testEnv("_LY_DATA_IN_TEMP_FILE"))
					path = _environment.getEnv("LY_TEMP_DIR") + "/data";

				if ((path != null) && (!path.equals("LY_APP_DIR")))
					file = path + "/" + "dump.vml";
				else
					file = "dump.vml";
			}

			FileOutputStream output = new FileOutputStream(file);
			PrintWriter		 pw = new PrintWriter(output);

			pw.println("<?xml version='1.0' encoding='" + LyString.DEFAULT_CHARSET + "'?>");

			if (!fullView)
			{
				LyArea area = getArea();

				pw.println("<!DOCTYPE AREA SYSTEM 'leon/conf/dtd/views.dtd'>");
				area.writeXml(pw, debug);
			}
			else if (this instanceof LyFrame)
			{
				pw.println("<!DOCTYPE FRAME SYSTEM 'leon/conf/dtd/views.dtd'>");
				writeXml(pw, debug);
			}
			else if (this instanceof LyDialog)
			{
				pw.println("<!DOCTYPE DIALOG SYSTEM 'leon/conf/dtd/views.dtd'>");
				writeXml(pw, debug);
			}

			pw.close();
		}
		catch (Exception e)
		{
			_environment.writeException(e);
		}
	} // end method writeXml
} // end class LyView
