/*
 * LyViewManager.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.view;

import leon.control.*;
import leon.misc.*;
import leon.view.event.*;
import leon.view.print.*;
import leon.view.setinterface.*;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;


/**
 * Communication control layer --> View Will implement remote update notifications.
 *
 * @version  $Id: LyViewManager.java 41416 2012-10-31 09:34:36Z jlsaouli $
 * @author   Lyria
 */
public abstract class LyViewManager
	implements LyViewManagerInterface
{
	public static final String __VERSION = "$Id: LyViewManager.java 41416 2012-10-31 09:34:36Z jlsaouli $";

	/** For message window : window closed (no button pressed). */
	public static final short COMMAND_CLOSED = 0;

	/** For message window : existence of the associated button. */
	public static final short COMMAND_OK = 1;

	/** For message window : existence of the associated button. */
	public static final short COMMAND_YES = 2;

	/** For message window : existence of the associated button. */
	public static final short COMMAND_ALL = 3;

	/** For message window : existence of the associated button. */
	public static final short COMMAND_NONE = 4;

	/** For message window : existence of the associated button. */
	public static final short COMMAND_NO = 5;

	/** For message window : existence of the associated button. */
	public static final short COMMAND_CANCEL = 6;

	/** For message window : existence of the associated button. */
	public static final short COMMAND_HELP = 7;
    
	/** Width of an empty area when displayed in the graphical builder. */
	public static int	EMPTY_AREA_WIDTH = 120;
	
	/** Height of an empty area when displayed in the graphical builder. */
	public static int	EMPTY_AREA_HEIGHT = 80;
    
	/** Width of an empty area with no parent area when displayed in the graphical builder. */
	public static int	EMPTY_TOP_AREA_WIDTH = 300;
	
	/** Height of an empty area with no parent area when displayed in the graphical builder. */
	public static int	EMPTY_TOP_AREA_HEIGHT = 200;
	
	/** Width of an empty form when displayed in the graphical builder. */
	public static int	EMPTY_FORM_WIDTH = 200;
	
	/** Height of an empty form when displayed in the graphical builder. */
	public static int	EMPTY_FORM_HEIGHT = 200;
	
	/** Width of an empty form field when displayed in the graphical builder. */
	public static int	EMPTY_FORM_FIELD_WIDTH = 50;
	
	/** Height of an empty form field when displayed in the graphical builder. */
	public static int	EMPTY_FORM_FIELD_HEIGHT = 10;
	
	/** Width of an empty label when displayed in the graphical builder. */
	public static int	EMPTY_LABEL_WIDTH = 50;
	
	/** Height of an empty label when displayed in the graphical builder. */
	public static int	EMPTY_LABEL_HEIGHT = 15;
	
	/** Width of an empty messages when displayed in the graphical builder. */
	public static int	EMPTY_MESSAGES_WIDTH = 50;
	
	/** Height of an empty messages when displayed in the graphical builder. */
	public static int	EMPTY_MESSAGES_HEIGHT = 10;
	
	/** Width of an empty table when displayed in the graphical builder. */
	public static int	EMPTY_TABLE_WIDTH = 480;
	
	/** Height of an empty table when displayed in the graphical builder. */
	public static int	EMPTY_TABLE_HEIGHT = 160;
	
	/** Width of an empty toolbar when displayed in the graphical builder. */
	public static int	EMPTY_TOOLBAR_WIDTH = 10;
	
	/** Height of an empty toolbar when displayed in the graphical builder. */
	public static int	EMPTY_TOOLBAR_HEIGHT = 10;
	
	/** Height of an empty menubar when displayed in the graphical builder. */
	public static int	EMPTY_MENUBAR_HEIGHT = 19;
	
	/** For show location : x position of the newly opened window. */
	public static final short SHOW_LOCATION_PARAMETER_X = 0;

	/** For show location : y position of the newly opened window. */
	public static final short SHOW_LOCATION_PARAMETER_Y = 1;

	/** For show location : width of the newly opened window. */
	public static final short SHOW_LOCATION_PARAMETER_WIDTH = 2;

	/** For show location : height of the newly opened window. */
	public static final short SHOW_LOCATION_PARAMETER_HEIGHT = 3;

	/** For show location : indicates whether newly opened window must be full screen. */
	public static final short SHOW_LOCATION_PARAMETER_FULL_SCREEN = 4;

	/** View list. */
	protected Vector<LyView> _viewList;
	
	/** View list Lock. */
	protected Object _viewListLock = new Object();

	/** Environment that created this view manager. */
	protected LyEnvironment _environment;

	/** List of properties of the view manager. */
	protected Hashtable<String, String> _properties;

	/** Beans. */
	protected LyResources _beans;
    /** Processing state. */
    protected boolean _processing;
    
    // List of read components
    protected Hashtable<String, LyComponent>   _definitions;


	/**
	 * The DemandTransmitter is a gateway between control and presentation layers. Used for
	 * notification from presentation layer to control layer.
	 */
	protected LySimpleDemandTransmitter _demandTransmitter;

	/** The plugin id. */
	protected String _pluginId;

	/**
	 * Stacks of shown controllers count. The stack is on per-view basis (default,
	 * target_popup_window ...) This stack is used to navigate through the views shown to the user
	 * (used in servlet mode).
	 */
	protected Hashtable<String, Integer> _controllerStackList;
	
	/** Counter for setWaiting calls. */
	protected int                       _waitingCounter;
	
	/** cache to store temporary directory, computed on first call to getTemporaryDir() **/
	protected String 				    _temporaryDirectory = null;
	
	/**
	 * Constructor.
	 */
	public LyViewManager(LyEnvironment environment)
	{
		_environment = environment;
		_viewList = new Vector<LyView>(50);
		_beans = null;
		_demandTransmitter = null;
		_pluginId = null;
		_controllerStackList = null;
		_waitingCounter = 0;

		setProperties();
	}

	/**
	 * Free the view manager. Clears all
	 */
	public void free()
	{
		if (_viewList != null)
		{
			LyView view;

			while (getViewCount() > 0)
			{
				view = getView(0);
				removeView(view);
			}

			_viewList = null;
		}

		if (_beans != null)
		{
			_beans.free();
			_beans = null;
		}

		if (_demandTransmitter != null)
		{
			_demandTransmitter.free();
			_demandTransmitter = null;
		}

        _definitions = null;
		_properties = null;
		_environment = null;
	} // end method free

	/**
	 * Prepares GUI to start.
	 */
	public void prepareGui()
	{
	}

	/**
	 * Starts GUI.
	 */
	public void startGui(LyGuiClientInterface client)
	{
        // Put environment as current one
        LyEnvironment oldEnvironment = LyEnvironment.setCurrentEnvironment(_environment);
        
        try
        {
            client.guiStart();
        }
        finally
        {
            // Remove current environment
            LyEnvironment.setCurrentEnvironment(oldEnvironment);
        }
	}

	/**
	 * Stops GUI.
	 */
	public void stopGui(LyGuiClientInterface client)
	{
		client.guiStop();
	}

	/**
	 * Causes the run() method of the given runnable to be invoked by the user-interface thread
	 * either synchronously or asynchronously. If asynchronous, the caller of this method continues
	 * to run in parallel, and is not notified when the runnable has completed. Otherwise, the
	 * thread which calls this method is suspended until the runnable completes.
	 *
	 * @param  runnable    Code to run on the user-interface thread.
	 * @param  sychronous  True is synchronous, false if asynchronous.
	 */
	public void execGui(Runnable runnable, boolean synchronous)
	{
		if (synchronous)
			runnable.run();
		else
			new _LyViewEvent(runnable);
	}

	/**
	 * Causes the run() method of the given runnable to be invoked by the user-interface thread
	 * asynchronously. The caller of this method continues to run in parallel, and is not notified
	 * when the runnable has completed.
	 *
	 * @param  runnable  Code to run on the user-interface thread.
	 */
	public final void execGui(Runnable runnable)
	{
		execGui(runnable, false);
	}

	/**
	 * Build internal view.
	 */
	public abstract LyViewComponent buildInternalView(LyComponent component);

	/**
	 * Build internal view.
	 */
	public LyViewComponent buildInternalView(String packageName, String prefixName, LyComponent component)
	{
		Class<?>  componentClass = component.getClass();
		String    componentClassName = componentClass.getName();

		while ((componentClassName.indexOf("leon.view.Ly") < 0)
				&& (componentClass.getSuperclass() != null))
		{
			componentClass = componentClass.getSuperclass();
			componentClassName = componentClass.getName();
		}

		String viewClassName = packageName + "." + prefixName
			+ componentClassName.substring("leon.view.Ly".length());

		return buildInternalView(viewClassName, component);
	}

	/**
	 * Build internal view.
	 */
	public LyViewComponent buildInternalView(String viewClassName, LyComponent component)
	{
		Class<?>	    componentClass = component.getClass();
		String		    componentClassName = componentClass.getName();
		Class<?>[]		types = new Class[1];
		Object[]	    parameters = new Object[1];
		LyViewComponent result = null;

		while ((componentClassName.indexOf("leon.view.Ly") < 0)
				&& (componentClass.getSuperclass() != null))
		{
			componentClass = componentClass.getSuperclass();
			componentClassName = componentClass.getName();
		}

		types[0] = componentClass;
		parameters[0] = component;

		try
		{
			ClassLoader cl = (_environment == null ? null : _environment.getClassLoader());
			Class<?>    viewClass;
			
			if (cl != null)
				viewClass = cl.loadClass(viewClassName);
			else
				viewClass = Class.forName(viewClassName);
			
			Constructor<?> constructor = viewClass.getConstructor(types);

			result = (LyViewComponent)constructor.newInstance(parameters);
		}
		catch (InvocationTargetException e1)
		{
			Throwable    targetE = e1.getTargetException();
			StringBuffer msg = new StringBuffer();

			msg.append("BF AppEngine : can't build internal view for ").append(componentClass);
			msg.append(", specific view class : ").append(viewClassName);
			msg.append(".\ntarget exception : ").append(targetE.getClass());

			if (targetE.getMessage() != null)
				msg.append(" : ").append(targetE.getMessage());

			IllegalStateException exception = new IllegalStateException(msg.toString());

			exception.setStackTrace(targetE.getStackTrace());

			throw exception;
		}
		catch (Exception e)
		{
			IllegalStateException exception = new IllegalStateException("BF AppEngine : can't build internal view for "
					+ componentClass + ", specific view class : " + viewClassName);

			exception.setStackTrace(e.getStackTrace());
			throw exception;
		}

		return result;
	} // end method buildInternalView

	/**
	 * Add a view.
	 */
	public void addView(LyView view)
	{
		synchronized (_viewListLock)
		{
			if (view == null)
				return;

			_viewList.addElement(view);
		}
	}

	/**
	 * Removes a view.
	 */
	public void removeView(LyView view)
	{
		synchronized (_viewListLock)
		{
			if (view == null)
				return;

			_viewList.removeElement(view);
		}
	}

	/**
	 * Checks if a view is still known or was freed.
	 */
	public boolean containsView(LyView view)
	{
		synchronized (_viewListLock)
		{
			return _viewList.contains(view);
		}
	}

	/**
	 * Get a view from a specified index.
	 *
	 * @param   idx  index of the view
	 *
	 * @return  the view or null if out of bounds (getViewCount)
	 */
	public LyView getView(int idx)
	{
		synchronized (_viewListLock)
		{
			if ((idx < 0) || (idx >= getViewCount()))
				return null;

			return _viewList.elementAt(idx);
		}
	}

	/**
	 * Get a view from a specified index.
	 *
	 * @param   idx  index of the view
	 *
	 * @return  the view
	 */
	public LyView getView(String id)
	{
		// Check validity
		if (id == null)
			return null;

		LyView view;
		int    i, n = getViewCount();

		synchronized (_viewListLock)
		{
			for (i = 0; i < n; i++)
			{
				// Fin the view for specified id
				view = _viewList.elementAt(i);

				if (id.equals(view.getId()))
					return view; // Found
			}
		}

		return null; // Not found
	}

	/**
	 * Retrieve a view.
	 */
	public int getViewCount()
	{
		synchronized (_viewListLock)
		{
			return (_viewList.size());
		}
	}

	/**
	 * Updates the _controllerStackList parameter with the count of controllers from the
	 * controllersStackList parameter. controllersStackList must contain names of controller stacks
	 * as keys and stack of controllers as values. _controllerStackList must contain names of
	 * controller stacks as keys and stack of controllers as values.
	 *
	 * @param  controllersStackList
	 */
	public void updateControllersStackList(Hashtable<String, Stack<LyViewController>> controllersStackList)
	{
		if (_controllerStackList == null)
			_controllerStackList = new Hashtable<String, Integer>();

		_controllerStackList.clear();

		if ((controllersStackList == null) || (controllersStackList.isEmpty()))
			return;

		Iterator<String> keys = controllersStackList.keySet().iterator();
		String   key;
		Stack<LyViewController>    stack;
		int		 count;

		while (keys.hasNext())
		{
			key = (String)keys.next();
			stack = controllersStackList.get(key);
			count = ((stack == null) ? 0 : stack.size());
			_controllerStackList.put(key, new Integer(count));
		}
	}

	/**
	 * Returns the stack size (number of controller) for the given stack (or null for default).
	 *
	 * @param   stackName  the name of the controller stack or null if default stack is considered
	 *
	 * @return  the number of controllers or null if stack is empty.
	 */
	public int getControllerStackSize(String stackName)
	{
		if ((stackName == null) || (_controllerStackList == null))
			return -1;

		Integer count = _controllerStackList.get(stackName);

		if (count == null)
			return 0;
		else
			return count.intValue();
	}

	/**
	 */
	protected void setProperties()
	{
	}

	/**
	 */
	protected void setProperty(String property)
	{
		if (property == null)
			return;

		if (_properties == null)
			_properties = new Hashtable<String, String>(20);

		_properties.put(property, property);
	}

	/**
	 */
	protected void removeProperty(String property)
	{
		if (property == null)
			return;

		if (_properties == null)
			return;

		if (_properties.containsKey(property))
			_properties.remove(property);
	}

	/**
	 */
	public boolean hasProperty(String property)
	{
		if ((property == null) || (_properties == null))
			return false;

		return (_properties.get(property) != null);
	}

	/**
	 */
	public LyEnvironment getEnvironment()
	{
		return _environment;
	}

	/**
	 * Build an instance of LyChartConverter corresponding to the internal view of the specified
	 * chart.
	 *
	 * @param   chart  the chart to convert
	 *
	 * @return  an instance of chart converter corresponding to the internal view of the specified
	 *          chart.
	 */
	public LyChartConverter buildChartConverter(LyChart chart)
	{
		String className = _beans.getString("CHART_CONVERTER");

		if ((className == null) || (className.length() == 0))
			className = "leon.view.common.LyCommonFreeChartConverter";

		Class<?>[]  types = new Class[1];
		Object[]    parameters = new Object[1];

		types[0] = LyChart.class;
		parameters[0] = chart;

		LyChartConverter converter = null;

		try
		{
			Class<?>	   viewClass = Class.forName(className);
			Constructor<?> constructor = viewClass.getConstructor(types);

			converter = (LyChartConverter)constructor.newInstance(parameters);
		}
		catch (InvocationTargetException e1)
		{
			_environment.writeException(e1.getTargetException());
			_environment.writeException(e1);
			throw new IllegalArgumentException("BF AppEngine : can't build chart converter class : "
				+ className);
		}
		catch (Exception e)
		{
			_environment.writeException(e);
			throw new IllegalArgumentException("BF AppEngine : can't build chart converter class : "
				+ className);
		}

		return converter;
	} // end method buildChartConverter

	/**
	 * Clear all the temporary resources used by the viewManager Implemented in subclasses.
	 */
	public abstract void clearTemporaryResources();

	/**
	 * Return the temporary directory used to store temporary resources but dependant from the
	 * viewer type (for ex: url / partial url for web).
	 */
	public abstract String getTemporaryLocation();

	/**
	 * Return the action that allow to download temporary resources but dependant from the
	 * viewer type (for ex: url / partial url for web).
	 */
	public abstract String getDownloadAction();
	
	/**
	 * Set Waiting status for all view of the application Implemented in subclasses.
	 */
	public abstract void setWaiting(boolean status);

	/**
	 * Shows a warning message to the user.
	 *
	 * @param  source   view element which raises the warning
	 * @param  message  warning message
	 */
	public abstract void showWarning(LyComponent source, String message);

	/**
	 * Shows en error message to the end-user.
	 *
	 * @param  message  error message
	 */
	public final void raiseError(String message)
	{
		showError(null, message);
	}

	/**
	 * Shows a warning message to the user.
	 *
	 * @param  source   view element which raises the error
	 * @param  message  error message
	 */
	public abstract void showError(LyComponent source, String message);

	/**
	 * Ask a question to the user The method is asynchronous and the result is returned to the
	 * client in messageAnswer. This method will block untill validation is done A timeout may also
	 * be implemented by subclasses.
	 *
	 * @param  source      view element which asks the question
	 * @param  message     text of the question
	 * @param  client      registred to receive the response.
	 * @param  clientData  freely set data which will be returned with the response.
	 */
	public abstract void askQuestion(LyComponent source, String message, LyQuestionAnswerInterface client,
		Object clientData);

	/**
	 * Show a message to the user and return the selection code.
	 *
	 * @param  source      view element which asks the question
	 * @param  title       title of the message window
	 * @param  icon        icon shown in the message window
	 * @param  message     text of the message
	 * @param  commands    bitset which contains COMMAND_xxx values
	 * @param  client      registred to receive the response.
	 * @param  clientData  freely set data which will be returned with the response.
	 */
	public abstract void showMessage(LyComponent source, String title, String icon, String message, BitSet commands,
		LyMessageAnswerInterface client, Object clientData);

	/**
	 * Shows a location to the user.
	 *
	 * @param  source    view element which raises the location
	 * @param  location  location
	 * @param  newWindow indicates whether a new window must be opened
	 */
	public abstract void showLocation(LyComponent source, String location, boolean newWindow);

	/**
	 * Shows a location to the user.
	 *
	 * @param  source     view element which raises the location
	 * @param  location   location
	 * @param  newWindow  indicates whether a new window must be opened
	 * @param  parameters list of parameters
	 */
	public abstract void showLocation(LyComponent source, String location, boolean newWindow, Hashtable<Short, String> parameters);

	/**
	 * Shows a file to the user.
	 *
	 * @param  source    view element which raises the location
	 * @param  location  location
	 */
	public abstract void showFile(LyComponent source, String location);

	/**
	 * Shows a PDF file to the user.<br>
	 * This makes a preview for printings.
	 *
	 * @param  source    view element which raises the location
	 * @param  location  location
	 */
	public abstract void showPDFPreview(LyComponent source, String location);

	/**
	 * To exit an application.
	 *
	 * @param  code  exit code of the application.
	 */
	public abstract void exit(int code);

	/**
	 * Copy a given text to the system clipboard.
	 *
	 * @param  text  the text to be copied to the clipboard.
	 */
	public abstract void copyToClipboard(String text);

	/**
	 * Retrieves the screen size.
	 *
	 * @param   component  The component used to retrieve the screen size.
	 *
	 * @return  The dimension of the screen used for displaying views.
	 */
	public int[] getScreenSize(Object component)
	{
		return new int[] { 0, 0, 0, 0 };
	}

	/**
	 * Retrieves the viewport size available on the computer.
	 *
	 * @return  The addition of all screen sizes.
	 */
	public int[] getMaxScreenSize()
	{
		return new int[] { 0, 0, 0, 0 };
	}
	
	/**
	 * Retrieves the maximum size of a window on the computer.
	 */
	public int[] getMaxWindowSize()
	{
		return new int[] {0, 0, 0, 0};
	}

	/**
	 * Assoc for answer notification.
	 */
	protected class _LyMessageAnswerInfo
	{
		public _LyMessageAnswerInfo(LyQuestionAnswerInterface client, Object clientData)
		{
			_clientQuestion = client;
			_clientData = clientData;
		}

		public _LyMessageAnswerInfo(LyMessageAnswerInterface client, Object clientData)
		{
			_clientMessage = client;
			_clientData = clientData;
		}

		public LyQuestionAnswerInterface _clientQuestion;
		public LyMessageAnswerInterface  _clientMessage;
		public Object					 _clientData;
	}

	/**
	 * Indicates wether this view manager handles multiple user sessions or not.
	 *
	 * @return  true if the view manager uses multiple user sessions (like web server)
	 */
	public abstract boolean isMultiSession();

	/**
	 * Emits a brief sound signal.
	 */
	public void beep()
	{
	}

	/**
	 * Retrieves a Color object from a LyColor.
	 *
	 * @param   key  The leonardi color.
	 *
	 * @return  A reference to the requested color, or null if it could not be found.
	 *
	 * @see     leon.misc.LyColors
	 */
	public abstract Object getColor(LyColor color);

	/**
	 * Retrieves a Font object from a LyFont.
	 *
	 * @param   key  The leonardi font.
	 *
	 * @return  A reference to the requested font, or null if it could not be found.
	 *
	 * @see     leon.misc.LyColors
	 */
	public abstract Object getFont(LyFont font);

	/**
	 * Retrieves the width of a string for a given LyFont.
	 *
	 * @param   font  The leonardi font.
	 * @param   text  The string value.
	 *
	 * @return  The width of given string for given font.
	 */
	public abstract int getStringWidth(LyFont font, String text);

	/**
	 * Constructs a default image.
	 */
	public abstract Object getDefaultImage();

	/**
	 * Gets images sizes.
	 *
	 * @return  Image size (table with 2 entries).
	 */
	public abstract int[] getSizes(Object image);

	/**
	 * Creates a scaled version of the given image.
	 *
	 * @param   image   The image to scale.
	 * @param   width   The width to which to scale the image.
	 * @param   height  The height to which to scale the image.
	 *
	 * @return  A scaled version of the image.
	 */
	public abstract Object getScaledInstance(Object image, int width, int height);

	/**
	 * Waits for the image loaging.
	 *
	 * @param  image  Image to load.
	 */
	public abstract void waitForLoading(Object component, Object image) throws InterruptedException;

	/**
	 * Retrieves an image from a Stream.
	 *
	 * @param   component  The client of the image creation.
	 * @param   resource   stream inputStream to the image data
	 *
	 * @return  the image or null if an error occurs
	 */
	public abstract Object loadImageFromStream(Object component, InputStream resource) throws IOException;

	/**
	 * Compare two images pixel by pixel.
	 *
	 * @param   image1  first image to compare
	 * @param   image2  second image to compare
	 *
	 * @return  true if images are identical.
	 *
	 * @throws  InterruptedException  exception if problems on grabber or loading.
	 */
	public abstract boolean isSameImage(Object image1, Object image2) throws InterruptedException;

	/**
	 * Free the resources of the image if not in the cache. If the image is in the cache, this
	 * method does nothing.
	 *
	 * @param  image  The image that must be freed.
	 */
	public abstract void freeImage(Object image);
	
	/**
	 * Indicates whether given image may be cached.
	 * Default implementation returns true.
	 * 
	 * @param path  The path of image that may be cached.
	 * @param image The image that may be cached.
	 * @return true if image may be cached, false otherwise
	 */
	public abstract boolean useCacheForImage(String path, Object image);

	/**
	 * Accessor to the target popup window name if using popup.This is stored on a per-thread basis
	 * Default mplementation always return null (used for web modes).
	 *
	 * @return  the window name or null for default main browser window
	 */
	public String getCurrentWindowName()
	{
		return null;
	}

	/**
	 * Accessor to the target popup window name if using popup.This is stored on a per-thread basis
	 * Default mplementation always return null (used for web modes).
	 *
	 * @return  the window name or null for default main browser window
	 */
	public void setCurrentWindowName(String name)
	{
	}

	/**
	 * Gets print manager.
	 */
	public LyPrintManager getPrintManager(LyPrintDataInterface printDataInterface, Properties properties)
	{
		return null; // XXX
	}

	/**
	 * Gets PDF print manager.
	 */
	public LyPrintManager getPdfPrintManager(LyPrintDataInterface printDataInterface, OutputStream output)
	{
		return null; // XXX
	}

	/**
	 * Generate a valid error image for the object The image is added into the environment.
	 *
	 * @param   imageKey    the key to the image without error
	 * @param   errorLevel  current error level.
	 *
	 * @return  the new key fot the image or imageKey if unchanged.
	 */
	public String buildErrorImage(String imageKey, short errorLevel)
	{
		return imageKey;
	}

	/**
	 * Show the perspective corresponding to the perspectiveId. Default implementation always does
	 * nothing (used for eclipse mode).
	 *
	 * @param  moduleId       The id of the module which demands this display.
	 * @param  perspectiveId  The id of the perspective to display.
	 */
	public void showPerspective(String moduleId, String perspectiveId)
	{
	}

	/**
	 * Indicates if the view manager handles perspectives.
	 *
	 * @return  true if the view manager handles perspectives, false otherwise.
	 */
	public boolean hasPerspective()
	{
		return false;
	}

	/**
	 * Encodes an image to JPEG format.
	 *
	 * @param   image      the image to encode
	 * @param   imageDir   temporary directory (optional)
	 * @param   imageFile  the JPEG File
	 *
	 * @return  true if the file was generated, false otherwise.
	 */
	public boolean encodeImage(Object image, String imageDir, String imageFile)
	{
		return false;
	}

	// Extern events queue
	
	_LyViewEventQueue    _eventQueue;
	int					 _nEvent;

	class _LyViewEventQueue implements Runnable
	{
		boolean _stopped = true;
		Vector<_LyViewEvent> _eventQueueList;
		
		public synchronized void add(_LyViewEvent event)
		{
			if (_eventQueueList == null)
				_eventQueueList = new Vector<_LyViewEvent>();
				
			_eventQueueList.add(event);
			
			if (_stopped)
			{
				_stopped = false;
				// Rearm events post
				Thread	t = new Thread(this);
				
				t.start();
			}
		}
		
		public int getSize()
		{
			if (_eventQueueList == null)
				return 0;
			
			return _eventQueueList.size();
		}
		
		public void run()
		{
			_LyViewEvent event;
			while (_eventQueueList.size() > 0)
			{
				event = _eventQueueList.remove(0);
				event.execute();
			}
			_stopped = true;
		}
	} // end class _LyViewEventQueue
	
	class _LyViewEvent
	{
		Runnable _runnable;

		_LyViewEvent(Runnable runnable)
		{
			_runnable = runnable;
			_nEvent++;

			if (_eventQueue == null)
				_eventQueue = new _LyViewEventQueue();
			
			_eventQueue.add(this);
			
			LyEnvironment environment = getEnvironment();
			environment.logTrace("Event " + _nEvent
				+ " reposted in View-Event-Queue for " + _runnable
				+ " from thread " + Thread.currentThread()
				+ ", " + (_eventQueue.getSize() - 1) + " remaining events");
		}

		public void execute()
		{
			_runnable.run();
		}
	} // end class _LyViewEvent

	/**
	 * Method for checking if events on data need to be reposted in GUI event queue.
	 *
	 * @return  True if the event need to be reposted, false otherwise.
	 */
	public boolean repostGuiEvents()
	{
		return false;
	}

	/**
	 * Method used to call back the listener of the specified message window for the specified
	 * result.
	 *
	 * @param  messageWindow
	 * @param  cmdResult
	 */
	public LyAnswerInterface messageActionPerformed(Object message, int cmdResult)
	{
		if (!(message instanceof LyMessageWindowInterface))
			throw new IllegalStateException("BF AppEngine : invalid message window " + message);

		LyMessageWindowInterface messageWindow = (LyMessageWindowInterface)message;
		LyDemandTransmitter		 transmitter = getDemandTransmitter();

		if (transmitter instanceof LyMessageWindowRegister)
			((LyMessageWindowRegister)transmitter).setMessageWindowResult(messageWindow, cmdResult);

		_LyMessageAnswerInfo info = (_LyMessageAnswerInfo)messageWindow.getClientData();
		LyAnswerInterface    answerInterface = null;

		if (info != null)
		{
			if (info._clientQuestion != null)
			{
				boolean result = (cmdResult == COMMAND_OK)
						|| (cmdResult == COMMAND_YES);

				info._clientQuestion.questionAnswer(result, info._clientData);
				answerInterface = info._clientQuestion;
			}
			else if (info._clientMessage != null)
			{
				info._clientMessage.messageAnswer(cmdResult, info._clientData);
				answerInterface = info._clientMessage;
			}
		}

		messageWindow.free();
		return answerInterface;
	} // end method messageActionPerformed

	/**
	 * Retrieves the DemandTransmitter which is a gateway between control and presentation layers.
	 *
	 * @return  The LyDemandTransmitter instance associated with the environment or the one
	 *          associated with the parent if the environment does not have a view manager.
	 */
	public LyDemandTransmitter getDemandTransmitter()
	{
		if (_demandTransmitter != null)
			return _demandTransmitter;

		if (_environment == null)
			return null;

		_demandTransmitter = new LySimpleDemandTransmitter(_environment);
		if (_environment.testEnv("LY_TEST_MONITOR"))
			setProperty("NO_MODAL");

		return _demandTransmitter;
	}

	/**
	 * Plays next item in test file.
	 */
	public boolean replay()
	{
		if (!_environment.testEnv("LY_REPLAY"))
			return false;

		LyDemandTransmitter transmitter = getDemandTransmitter();

		return transmitter.playNext();
	}

	/**
	 * Gets delay between plays.
	 */
	public int getDelay()
	{
		if (!_environment.testEnv("LY_REPLAY"))
			return -1;

		LyDemandTransmitter transmitter = getDemandTransmitter();

		return transmitter.getDelay();
	}

	public void takeSnapshot(LyView view, String file)
	{

	}

	public void takeSnapshot(LyArea area, String file)
	{

	}

	/**
	 * Gets the plugin id.
	 *
	 * @return  The plugin id.
	 */
	public String getPluginId()
	{
		return _pluginId;
	}

	/**
	 * Sets the plugin id.
	 *
	 * @param  The  plugin id.
	 */
	public void setPluginId(String pluginId)
	{
		_pluginId = pluginId;
	}

	/**
	 * Adds the specified demand listener to receive demands (events) from this component.
	 * (LyDemandTransmitter interface)
	 *
	 * @param  listener  client for the demand
	 */
	public void addDemandListener(LyDemandListener listener)
	{
		getDemandTransmitter().addDemandListener(listener);
	}

	/**
	 * Removes the specified demand listener to receive demands from this component.
	 * (LyDemandTransmitter interface)
	 *
	 * @param  listener  client for the demand
	 */
	public void removeDemandListener(LyDemandListener listener)
	{
		getDemandTransmitter().removeDemandListener(listener);
	}

	/**
	 * Processes demands on this component.
	 *
	 * @param  controllerId  the demand is sent to the specified listener (LyDemandTransmitter
	 *                       interface)
	 */
	public void processDemand(String listenerId, LyDemand demand)
	{
        // Put environment as current one
        LyEnvironment oldEnvironment = LyEnvironment.setCurrentEnvironment(_environment);

        try
        {
            _processing = true;
            getDemandTransmitter().processDemand(listenerId, demand);
        }
        finally
        {
            // Restore current environment
            LyEnvironment.setCurrentEnvironment(oldEnvironment);
            
            _processing = false;
        }
	}

	/**
	 * Moves mouse on component.
	 */
	public void moveCursor(LyComponent component)
	{
	}

	/**
	 * Show given comments to end-user.
	 */
	public void showComments(String comments)
	{
	}

	/**
	 * Swap two sessions.
	 *
	 * @param  newSession     the new session
	 * @param  oldSession     the old session
	 * @param  applicationId  the application id
	 */
	public void swapSession(Object newSession, Object oldSession, String applicationId)
	{
	}

	/**
	 * Share a session with other sessions.
	 *
	 * @param  sharedSession   the shared session
	 * @param  sharingSessions the list of sessions sharing the shared session
	 * @param  applicationId   the application id
	 */
	public void shareSession(Object sharedSession, Vector<Object> oldSession, String applicationId)
	{
	}
	
    /**
     * Indicates whether current view manager is being shared.
     * 
     * @return true if current view manager is being shared, false otherwise.
     */
	public boolean isShared()
	{
		return false;
	}

	/**
	 * Check if this view manager is compatible with the given view manager.
	 *
	 * @param   otherViewManager  the other view manager
	 *
	 * @return  true if both view manager are compatible
	 */
	public boolean areViewManagerCompatible(LyViewManager otherViewManager)
	{
		if (this.getClass() != otherViewManager.getClass())
			return false;

		return true;
	}
	
	/**
     * Gets web root URL.
     */
    public String getRootURL()
    {
        return (_environment == null ? null : _environment.getEnv("LY_ROOT_URL"));
    }
    
	 /**
     * Load external definitions that can be used as templates for new components.
     * @param xmlStream An xml stream that contains the templates definitions.
     * @return List of components loaded on parsing given XML file.
     */
	public Vector<LyComponent> loadDefinitions(InputStream xmlStream)
    {  
    	LyViewParser    parser = new LyViewParser(_environment, xmlStream);
        Object          rootElement = parser.getRootElement();
        
        if (rootElement instanceof Vector<?>)
        {
        	@SuppressWarnings("unchecked")
        	Vector<LyComponent> temp = (Vector<LyComponent>)rootElement;
            return temp;
        }
        
        else if (rootElement instanceof LyComponent)
        {
            Vector<LyComponent>     result = new Vector<LyComponent>(1);
            
            result.add((LyComponent)rootElement);
            return result;
        }
        else
            return null;
    }
	
    /**
     * Load external definitions that can be used as templates for new components.
     * @param xmlFile File path of the XML file that contains the templates definitions.
     * @return List of components loaded on parsing given XML file.
     */
    public Vector<LyComponent> loadDefinitions(String xmlFile)
    {        
    	if (xmlFile == null)
    		return null;
    	
    	InputStream stream;
    	Vector<LyComponent> result = null;
    	try
		{
			stream = new FileInputStream(xmlFile);
			result = loadDefinitions(stream);
		}
		catch (FileNotFoundException e)
		{
			getEnvironment().writeException(e);
			result = null;
		}
		return result;
    }
    
    /**
     * Register component as definition.
     * @param component Component to register in definitions list.
     * @exception IllegalArgumentException If a definition with given id was already registered.
     */
    public void registerDefinition(LyComponent component)
    {
        String  id = (component == null ? null : component.getId());
        
        if (id != null)
        {
            if (_definitions == null)
                _definitions = new Hashtable<String, LyComponent>();
            else
            {
                LyComponent     definition = _definitions.get(id);
                
                if ((definition != null)
                        && (definition != component))
                    throw new IllegalArgumentException("BF AppEngine : a component with id \"" + id
                        + "\" was already registered on view manager (component " + definition + ").");
            }
                
            _definitions.put(id, component);
        }
    }
    
    /**
     * Unregister component definition.
     * @param component Component to unregister from definitions list.
     * @return The unregistered component if found, false otherwise.
     */
    public LyComponent unregisterDefinition(LyComponent component)
    {
        String  id = (component == null ? null : component.getId());
        
        if ((id != null)
                && (_definitions != null))
            return _definitions.remove(id);
        else
            return null;
    }
    
    /**
     * Retrieve a registered component by its id.
     * @param Id of the searched component.
     * @return The definition corresponding to the given id if found, false otherwise.
     * @exception IllegalArgumentException If no definition found with given id.
     */
    public LyComponent findDefinition(String componentId)
    {
        if (componentId == null)
            return null;
        if (_definitions == null)
            return null;
        
        LyComponent definition = _definitions.get(componentId);
        
        if (definition == null)
            throw new IllegalArgumentException("BF AppEngine : No definition with id \"" + componentId
                + "\" was found on view manager.");
        
        return definition;
    }
    
    /**
     * Gets waiting counter. Each call to setWaiting(true) increments counter,
     * each call to setWaiting(false) decrements it (until 0 is reached).
     * @return The waiting counter.
     */
    public int getWaitingCounter()
    {
        return _waitingCounter;
    }
    
    /**
     * Accessor to the properties file or parameter if specified in constructor.
     *
     * @param prefix    Prefix used for properties files name.
     * @return the properties.
     */
    public Properties getPropertiesFromContext(String prefix)
    {
        return null;
    }
	
    /**
     * Return the temporary directory used to store temporary resources.
     */
    public String getTemporaryDir()
    {
    	if (_temporaryDirectory == null)
    	{
    		try
    		{
    			String tempDir = getEnvironment().getFile("LY_TEMP_DOC_DIR");

    			if (tempDir == null || tempDir.equals("LY_TEMP_DOC_DIR") || tempDir.length() <= 0)
    				tempDir = System.getProperty("java.io.tmpdir");
    			if(tempDir != null)
    			{
    				_temporaryDirectory  = getEnvironment().getString().expand(tempDir);
    				
    				// if needed try to create the tmp dir
    				File tmpDirFile = new File(_temporaryDirectory);
    				if (!tmpDirFile.exists())
    				{
    					tmpDirFile.mkdirs();
    				}
    				getEnvironment().logTrace("Leonardi : temporary directory is set to "+_temporaryDirectory);
    			}
    		}
    		catch (Throwable e)
    		{	
    		}
    	}
    	return _temporaryDirectory;
    }
    
    /**
     * Indicates whether RTL (right to left) mode is activated
     * @return true is RTL mode is activated, false otherwise.
     */
    public boolean isRtlMode()
    {
    	return false;
    }
    
    /**
     * Indicates whether this view manager is in portlet mode.
     * @return true if portlet mode is activated, false otherwise.
     */
    public boolean isPortletMode()
    {
    	return false;
    }
    
    /**
     * Returns the name of the Java class that must be used to instanciate a new slave session.
     * 
     * @return null if default session class must be used, a full java path otherwise.
     */
    public String getSlaveSessionClassName()
    {
    	return null;
    }
} // end class LyViewManager
