/**
 * LyNodeFactory.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.view;

import java.lang.reflect.*;
import java.util.*;

import leon.misc.*;
import leon.parser.*;

import org.xml.sax.*;


/**
 * This class is used to build instances from the SAX Parser.
 *
 * @version  $Id: LyViewNodeFactory.java 29006 2010-12-23 17:44:46Z bmichel $
 * @author   Lyria
 */
public class LyViewNodeFactory extends LyAbstractViewNodeFactory
	implements LyXMLElementFactory
{
	public static final String __VERSION = "$Id: LyViewNodeFactory.java 29006 2010-12-23 17:44:46Z bmichel $";


	/**
	 * Utility method to tag name from class name.
	 *
	 * @param   className  className
	 *
	 * @return  The corresponding class name.
	 */
	public static String getTag(String className)
	{
		String tag = _tags.get(className);

		if (tag != null)
			return tag;

		String name;

		if (className.startsWith("leon.view.Ly"))
			name = className.substring(12);
		else
			name = className;

		// First, some exceptions in naming
		if (name.equals("FormField"))
			tag = TAG_FIELD;
		else if (name.equals("MenuBar"))
			tag = TAG_MENUBAR;
		else if (name.equals("ToolBar"))
			tag = TAG_TOOLBAR;
		else if (name.equals("TitleBar"))
			tag = TAG_TITLEBAR;
		else if (name.equals("SelectBar"))
			tag = TAG_SELECTBAR;
		else if (name.equals("SelectItem"))
			tag = TAG_SELECTITEM;
		else if (name.equals("HRef"))
			tag = TAG_HREF;
		else if (name.equals("HRefs"))
			tag = TAG_HREFS;
		else if (name.equals("FieldUnit"))
			tag = TAG_UNIT;
		else
		{
			// General case
			StringBuffer buffer = new StringBuffer();
			int			 i, nc = name.length();
			char		 c;

			for (i = 0; i < nc; i++)
			{
				c = name.charAt(i);
				if ((i > 0) && Character.isUpperCase(c))
					buffer.append('_');

				buffer.append(Character.toUpperCase(c));
			}

			tag = buffer.toString();
		}

		_tags.put(className, tag);
		return tag;
	} // end method getTag

    // Current environment of view parser
	protected LyEnvironment                    _environment;
    
    // List of read components
    protected Hashtable<String, LyComponent>   _components;
    
    // Boolean indicating if components have to be registered on viewManager
    protected boolean                          _registerComponents;
    

	/**
	 * Constructor.
	 */
	public LyViewNodeFactory(LyEnvironment environment)
	{
		_environment = environment;
        _components = null;
        _registerComponents = false;
	}
    
    /**
     * Access to the environment.
     */
    public LyEnvironment getEnvironment()
    {
        return _environment;
    }
    
    /**
     * Register component.
     */
    protected void registerComponent(LyComponent component)
    {
        String  id = (component == null ? null : component.getId());
        
        if (id != null)
        {
            if (_components == null)
                _components = new Hashtable<String, LyComponent>();
            _components.put(id, component);
            
            if (_registerComponents)
            {
                LyViewManager   viewManager = component.getViewManager();
                
                viewManager.registerDefinition(component);
            }
        }
    }
    
    /**
     * Find a component by its id.
     */
    protected LyComponent findComponent(String componentId)
    {
        if (componentId == null)
            return null;
        if (_components == null)
            return null;
        
        return _components.get(componentId);
    }
    
	/**
	 * Builds the result corresponding to each node. The result depends on the node tag. This method
	 * is invoked by SAXParser at the begening of each new tag.
	 *
	 * @param      uri         The Namespace URI.
	 * @param      localName   The Namespace local name.
	 * @param      qName       The qualified (prefixed) name.
	 * @param      attributes  The SAX2 attributes.
     * @param      parent      The parent node (created by {@link #buildElement(String, String, String,
     *                         Attributes)})
	 *
	 * @exception  SAXException  if the result could not have been built.
	 */
	public Object buildElement(String uri, String localName, String qName, Attributes attributes,
        Object parent) throws SAXException
	{
		// Get some common information
		LyComponent result = null;

		// Build the appropriate result corresponding to each node.
		try
		{
			// Find class corresponding to tag then get a constructor
			// that takes a parameter of type LyViewNode and invoke it
			String tag = qName.intern();

            if (tag == TAG_DEFINITION)
            {
                _registerComponents = true;
                return new Vector<LyComponent>();
            }
            else if (tag == TAG_DESCRIPTION)
            {
                return new Vector<LyComponent>(1);
            }
            else if (tag == TAG_EXTERN)
            {
                String          id = attributes.getValue(ATT_ID);
                LyViewManager   viewManager = (LyViewManager)_environment.getViewManager();
                
                return viewManager.findDefinition(id);
            }
            else if (tag == TAG_DATA)
			{
				String name = attributes.getValue(ATT_NAME);
				String value = attributes.getValue(ATT_VALUE);

				if ((name != null) && (value != null))
					return new _LyData(name, value);
				else
					return null;
			}
            else if (tag == TAG_VIEWS)
            {
                return new Vector<LyView>(5);
            }

			String className = getClassName(tag);

			if (className == null)
				return null;

            // Get some common information

            String      templateId = attributes.getValue(ATT_TEMPLATE);
            LyComponent template = findComponent(templateId);

            // Forbid forward reference on template
            if (templateId != null)
            {
                if ((templateId.length() > 0)
                        && (template == null))
                    throw new IllegalStateException(
                        "BF AppEngine : Forward references to template objects are not supported. Element '"
                        + tag + "' with template : '" + templateId + "'.");
                
                if  ((parent instanceof LyComponent)
                     && (((LyComponent)parent).getModel() != null)
                     && (((LyComponent)parent).getModel().getComponent(templateId, true) != null))
                {             
                    result = ((LyComponent)parent).getComponentWithModel(template);
                    if (result != null)
                    {
                        LyViewNode node = new LyViewNode(attributes, this);
                        
                        result.initFromViewNode(node);
                        return result;
                    }
                }
            }
            
			LyViewNode node = new LyViewNode(attributes, this);
			Class<?>[] types = new Class[1];
			Object[]   parameters = new Object[1];

			types[0] = LyViewNode.class;
			parameters[0] = node;

			try
			{         
				Class<?>	   componentClass = Class.forName(className);
				Constructor<?> constructor = componentClass.getDeclaredConstructor(types);

				result = (LyComponent)constructor.newInstance(parameters);
			}
			catch (InvocationTargetException e)
			{
				_environment.logError("Tag : " + tag + " - className : " + className);
				_environment.writeException(e.getTargetException());
				throw new SAXException(e.getMessage());
			}
			catch (Exception e)
			{
				_environment.logError("Tag : " + tag + " - className : " + className);
				_environment.writeException(e);
				throw new SAXException(e.getMessage());
			}
		}
		catch (Exception e)
		{
			_environment.writeException(e);
		}

		return result;
	} // end method buildElement

	/**
	 * End of Build of a node. The result depends on the node tag. This method is invoked by
	 * SAXParser at the begening of each new tag.
	 *
	 * @param      uri        The Namespace URI.
	 * @param      localName  The Namespace local name.
	 * @param      qName      The qualified (prefixed) name.
	 * @param      parent     - The parent node (created by buildElement)
	 * @param      current    - The current node (created by buildElement)
	 *
	 * @exception  SAXException  if the result could not have been built.
	 */
	public void endElement(String uri, String localName, String qName, Object parent, Object current)
		throws SAXException
	{
		if ((current instanceof _LyData)
			&& (parent instanceof LyComponent))
		{
            if ((parent instanceof LyFormField)
                    && (((LyFormField)parent).getFieldItem() != null))
                // For upward compatibility, datas on form field are also added on field item
                ((LyFormField)parent).getFieldItem().addData(((_LyData)current)._name, ((_LyData)current)._value);
            
            ((LyComponent)parent).addData(((_LyData)current)._name, ((_LyData)current)._value);
			return;
		}
        
        if (qName == TAG_DEFINITION)
        {
            _registerComponents = false;
            return;
        }
        
        if ((current instanceof LyComponent) || (current == null))
        {
            LyComponent     component = (LyComponent)current;
            
            if (parent instanceof Vector<?>)
            {
            	@SuppressWarnings("unchecked")
            	Vector<Object> temp = (Vector<Object>)parent;
                temp.add(component);
            }
            
            else if (parent instanceof LyComponent)
            {
        		// Add child on parent
        		try
        		{
        			((LyComponent)parent).addChild(component);
        		}
        		catch (Exception e)
        		{
        			_environment.writeException(e);
        		}
            }
            
            registerComponent(component);
        }
	}

	class _LyData
	{
		String _name;
		String _value;

		_LyData(String name, String value)
		{
			_name = name;
			_value = value;
		}
	}
} // end class LyViewNodeFactory
