/*
 * Ly3DBox.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.common;

import java.util.*;
import leon.view.*;


/**
 * This class defines a box in a 3D space.
 *
 * @version $Id: Ly3DBox.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class Ly3DBox
{
	public static final String __VERSION = "$Id: Ly3DBox.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Start point. */
	protected LyPoint		   _startPoint;

	/** End point. */
	protected LyPoint		   _endPoint;

	/**
	 * Constructor.
	 */
	public Ly3DBox()
	{
		this(new LyPoint(), new LyPoint());
	}

	/**
	 * Constructor.
	 *
	 * @param x1 Start point x value
	 * @param y1 Start point y value
	 * @param z1 Start point z value
	 * @param x2 End point x value
	 * @param y2 End point y value
	 * @param z2 End point z value
	 */
	public Ly3DBox(float x1, float y1, float z1,
		float x2, float y2, float z2)
	{
		this(new LyPoint(x1, y1, z1), new LyPoint(x2, y2, z2));
	}

	/**
	 * Constructor.
	 *
	 * @param startPoint Start point
	 * @param endPoint   End point
	 */
	public Ly3DBox(LyPoint startPoint, LyPoint endPoint)
	{
		init();

		setStartPoint(startPoint);
		setEndPoint(endPoint);
	}

	/**
	 * Init method.
	 */
	public void init()
	{
		_startPoint = new LyPoint();
		_endPoint = new LyPoint();
	}

	/**
	 * Calculates the union of the current box and the given box.
	 *
	 * @param box the box to use for union
	 */
	public Ly3DBox union(Ly3DBox box)
	{
		if (box == null)
			return this;

		float   minx, miny, minz, maxx, maxy, maxz, val1, val2;

		LyPoint startPoint = box.getStartPoint();
		LyPoint endPoint = box.getEndPoint();

		// Get the min x value
		val1 = Math.min(_startPoint.getX(), _endPoint.getX());
		val2 = Math.min(startPoint.getX(), endPoint.getX());
		minx = Math.min(val1, val2);

		// Get the min y value
		val1 = Math.min(_startPoint.getY(), _endPoint.getY());
		val2 = Math.min(startPoint.getY(), endPoint.getY());
		miny = Math.min(val1, val2);

		// Get the min z value
		val1 = Math.min(_startPoint.getZ(), _endPoint.getZ());
		val2 = Math.min(startPoint.getZ(), endPoint.getZ());
		minz = Math.min(val1, val2);

		// Get the max x value
		val1 = Math.max(_startPoint.getX(), _endPoint.getX());
		val2 = Math.max(startPoint.getX(), endPoint.getX());
		maxx = Math.max(val1, val2);

		// Get the max y value
		val1 = Math.max(_startPoint.getY(), _endPoint.getY());
		val2 = Math.max(startPoint.getY(), endPoint.getY());
		maxy = Math.max(val1, val2);

		// Get the max z value
		val1 = Math.max(_startPoint.getZ(), _endPoint.getZ());
		val2 = Math.max(startPoint.getZ(), endPoint.getZ());
		maxz = Math.max(val1, val2);

		return new Ly3DBox(minx, miny, minz, maxx, maxy, maxz);
	} // end method union

	/**
	 * Returns a string representation of the box.
	 */
	public String toString()
	{
		StringBuffer buffer = new StringBuffer();

		buffer.append("[[");
		buffer.append(_startPoint.getX()).append(",").append(_startPoint.getY()).append(",").append(
			_startPoint.getZ());
		buffer.append("] , [");
		buffer.append(_endPoint.getX()).append(",").append(_endPoint.getY()).append(",").append(
			_endPoint.getZ());
		buffer.append("]]");

		return (buffer.toString());
	}

	/**
	 * Sets the start point of the box.
	 *
	 * @param startPoint start point
	 */
	public void setStartPoint(LyPoint startPoint)
	{
		_startPoint = startPoint;
	}

	/**
	 * Sets the end point of the box.
	 *
	 * @param startPoint end point
	 */
	public void setEndPoint(LyPoint endPoint)
	{
		_endPoint = endPoint;
	}

	/**
	 * Returns the start point of the box.
	 */
	public LyPoint getStartPoint()
	{
		return _startPoint;
	}

	/**
	 * Returns the end point of the box.
	 */
	public LyPoint getEndPoint()
	{
		return _endPoint;
	}

	/**
	 * Returns all the points of the box.
	 */
	public Vector<LyPoint> getPoints()
	{
		Vector<LyPoint> points = new Vector<LyPoint>(6);

		points.addElement(_startPoint);
		points.addElement(_endPoint);

		LyPoint point;

		point = new LyPoint(_startPoint.getX(), _startPoint.getY(), _endPoint.getZ());
		points.addElement(point);
		point = new LyPoint(_startPoint.getX(), _endPoint.getY(), _startPoint.getZ());
		points.addElement(point);
		point = new LyPoint(_endPoint.getX(), _startPoint.getY(), _startPoint.getZ());
		points.addElement(point);
		point = new LyPoint(_endPoint.getX(), _endPoint.getY(), _startPoint.getZ());
		points.addElement(point);
		point = new LyPoint(_endPoint.getX(), _startPoint.getY(), _endPoint.getZ());
		points.addElement(point);
		point = new LyPoint(_startPoint.getX(), _endPoint.getY(), _endPoint.getZ());
		points.addElement(point);

		return points;
	}

	public float getDiameter()
	{
		float x1 = _startPoint.getX();
		float y1 = _startPoint.getY();
		float z1 = _startPoint.getZ();
		float x2 = _endPoint.getX();
		float y2 = _endPoint.getY();
		float z2 = _endPoint.getZ();

		float sum = ((x2 - x1) * (x2 - x1)) + ((y2 - y1) * (y2 - y1))
			+ ((z2 - z1) * (z2 - z1));

		return (float)Math.sqrt(sum);
	}
} // end class Ly3DBox
