/*
 * LyCommonFreeChartConverter.java
 * Copyright (c) 2007, Lyria.
 * All rights reserved.
 */

package leon.view.common;

import org.jfree.chart.*;
import org.jfree.chart.axis.*;
import org.jfree.chart.entity.*;
import org.jfree.chart.imagemap.*;
import org.jfree.chart.labels.*;
import org.jfree.chart.plot.*;
import org.jfree.chart.renderer.category.*;
import org.jfree.chart.renderer.xy.*;
import org.jfree.chart.title.*;
import org.jfree.chart.urls.*;
import org.jfree.data.category.*;
import org.jfree.data.general.*;
import org.jfree.data.xy.*;
import org.jfree.ui.*;
import org.jfree.util.*;

import java.awt.*;
import java.awt.geom.*;
import java.io.*;
import java.text.*;
import java.util.*;
import java.util.Vector;

import leon.misc.*;
import leon.view.*;
import leon.view.getinterface.*;


/**
 * This class implements the Chart object (as described in the dtd) for the viewer.
 *
 * @version $Id: LyCommonFreeChartConverter.java 36616 2012-08-28 12:35:15Z jlsaouli $
 * @author  Lyria
 * @see     LyChart
 */
public class LyCommonFreeChartConverter extends LyChartConverter
{
    public static final String __VERSION =
    	"$Id: LyCommonFreeChartConverter.java 36616 2012-08-28 12:35:15Z jlsaouli $";
    
    /** Default chart width. */
    public static int		   DEFAULT_CHART_WIDTH = 650;
    
    /** Default chart height. */
    public static int		   DEFAULT_CHART_HEIGHT = 520;
    
    /** Default alpha ratio for transparency. */
    public static float		   ALPHA_RATIO = 0.8F;
    
    /** Height of the link used to display values in plot and 3D bar charts. */
    public static int		   LINK_HEIGHT = 5;
    
    /** Max number of lines for splitting the labels in the chart action. */
    public static int		   MAX_LABEL_LINES = 2;
    
    /**
     * Visible separator used for labels of multiple points in scatter plot mode
     */
    public static String	   SCATTER_PLOT_VISIBLE_LABEL_SEPARATOR = "-";
    
    /**
     * Internal separator used for labels of multiple points in scatter plot mode
     */
    public static char		   SCATTER_PLOT_INTERNAL_LABEL_SEPARATOR = LyTools.VALUE_SEP;
    
    /** Separator used for labels of multiple points in scatter plot mode. */
    public static String	   SCATTER_PLOT_LABEL_SEPARATOR = "-";
    
    /** Separator used for tooltips of points in scatter plot mode. */
    public static String	   SCATTER_PLOT_TOOLTIP_SEPARATOR = " / ";
    
    /** Number of decimals for position tooltips in scatter plot mode. */
    public static int		   SCATTER_PLOT_DECIMALS_NUMBER = 2;
    
    protected JFreeChart	   _chartMgr;
    protected Paint[]		   _paints;
    protected Color			   _tooltipColor;
     
    protected int              _chartWidth;
    protected int              _chartHeight;

	/**
	 * Constructor.
	 */
	public LyCommonFreeChartConverter()
	{
        _chartWidth = DEFAULT_CHART_WIDTH;
        _chartHeight = DEFAULT_CHART_HEIGHT;
	}

	/**
	 * Constructor.
	 *
	 * @param chart the chart created by LyChartController
	 * @see   leon.control.LyChartController
	 */
	public LyCommonFreeChartConverter(LyChart chart)
	{
		super(chart);
        
        LyEnvironment   environment = chart.getEnvironment();
        
        _chartWidth = environment.getIntResource("LY_CHART_WIDTH");
        if (_chartWidth < 0)
            _chartWidth = DEFAULT_CHART_WIDTH;
        _chartHeight = environment.getIntResource("LY_CHART_HEIGHT"); 
        if (_chartHeight < 0)
            _chartHeight = DEFAULT_CHART_HEIGHT;
	}

    /**
     * Sets the width and height of the generated image.
     *
     * @param width  the new width of the generated image
     * @param height the new height of the generated image
     */
    public void setDimension(int width, int height)
    {
        if (width > 0)
            _chartWidth = width;
        if (height > 0)
            _chartHeight = height;
    }
    
	/**
	 * Initialization method.
	 */
	public void initialize()
	{
		_chartMgr = null;
		_paints = ChartColor.createDefaultPaintArray();

		LyColor color = _chart.getEnvironment().getLyColor("LY_TOOLTIP_COLOR");

		_tooltipColor = new Color(color.getRed(), color.getGreen(), color.getBlue());
	}

	/**
	 * Reset the display from the underlying XML data Main method to construct the chart.
	 */
	@SuppressWarnings("deprecation")
	public void reset()
	{
		java.util.Vector<String>  xSerie = new java.util.Vector<String>();
		java.util.Vector<String>  ySerie = new java.util.Vector<String>();
		boolean unknownValues = true;
		String  title = "", xTitle = "", yTitle = "";

		_data = new Vector<Vector<Object>>();
		_seriesLabel = new Vector<String>();
		createChartableData(xSerie, ySerie, unknownValues);

		// Create the set of data to be displayed
		CategoryDataset    categoryData = null;
		XYSeriesCollection xyDataset = null;

		if (_chart.getGraphType() != LyChartBase.GraphType.SCATTER_PLOT)
			categoryData = createCategoryDataset();

		// Sets chart title
		if ((_chart.getTitle() != null) && (!_chart.getTitle().equals("")))
		{
			LyViewManager viewManager = _chart.getViewManager();

			if ((viewManager != null) && !viewManager.hasProperty(_chart.getTitle()))
				title = _chart.getEnvironment().translate(_chart.getTitle());
			else
				title = _chart.getTitle();
		}

		if (_chart.getXTitle() != null)
			xTitle = _chart.getXTitle();

		if (_chart.getYTitle() != null)
			yTitle = _chart.getYTitle();

		// Create the requested chart type
		if (_chart.getGraphType() == LyChartBase.GraphType.PIE)
		{
			if (categoryData.getColumnCount() > 1)
			{
				// Create a multiple pie chart
				if (_chart.isD3Effect())
					_chartMgr = ChartFactory.createMultiplePieChart3D(title, categoryData,
							TableOrder.BY_COLUMN, _chart.isShowLegend(), false, false);
				else
					_chartMgr = ChartFactory.createMultiplePieChart(title, categoryData,
							TableOrder.BY_COLUMN, _chart.isShowLegend(), false, false);
			}
			else
			{
				// Create a pie chart with its own dataset
				DefaultPieDataset dataset = new DefaultPieDataset();
				int				  i, n = categoryData.getRowCount();

				for (i = 0; i < n; i++)
					dataset.setValue(categoryData.getRowKey(i), categoryData.getValue(i, 0));

				if (_chart.isD3Effect())
					_chartMgr = ChartFactory.createPieChart3D(title, dataset, _chart.isShowLegend(),
							false, false);
				else
					_chartMgr = ChartFactory.createPieChart(title, dataset, _chart.isShowLegend(),
							false, false);
			}
		}
		else if (_chart.getGraphType() == LyChartBase.GraphType.SCATTER_PLOT)
		{
			xyDataset = createXYDataset(true);
			_chartMgr = ChartFactory.createScatterPlot(title, xTitle, yTitle, xyDataset,
					PlotOrientation.VERTICAL, _chart.isShowLegend(), false, false);
		}
		else if (_chart.getGraphType() == LyChartBase.GraphType.PLOT)
		{
			xyDataset = createXYDataset(false);
			_chartMgr = ChartFactory.createXYLineChart(title, xTitle, yTitle, xyDataset,
					PlotOrientation.VERTICAL, _chart.isShowLegend(), false, false);

			if (_chart.getXType() == LyChart.Type.CHARTDATE)
			{
				XYPlot   plot = (XYPlot)_chartMgr.getPlot();
				DateAxis dateAxis = new DateAxis();

				plot.setDomainAxis(0, dateAxis);
			}

		}
		else if (_chart.getGraphType() == LyChartBase.GraphType.BAR)
		{
			if (_chart.isD3Effect())
				_chartMgr = ChartFactory.createBarChart3D(title, xTitle, yTitle, categoryData,
						PlotOrientation.VERTICAL, _chart.isShowLegend(), false, false);
			else
				_chartMgr = ChartFactory.createBarChart(title, xTitle, yTitle, categoryData,
						PlotOrientation.VERTICAL, _chart.isShowLegend(), false, false);
		}
		else if (_chart.getGraphType() == LyChartBase.GraphType.STACK_BAR)
		{
			if (_chart.isD3Effect())
				_chartMgr = ChartFactory.createStackedBarChart3D(title, xTitle, yTitle,
						categoryData, PlotOrientation.VERTICAL, _chart.isShowLegend(), false,
						false);
			else
				_chartMgr = ChartFactory.createStackedBarChart(title, xTitle, yTitle, categoryData,
						PlotOrientation.VERTICAL, _chart.isShowLegend(), false, false);
		}
		else if (_chart.getGraphType() == LyChartBase.GraphType.STACK_AREA)
		{
			DefaultTableXYDataset tablexyDataset = createTableXYDataset();

			_chartMgr = ChartFactory.createStackedXYAreaChart(title, xTitle, yTitle, tablexyDataset,
					PlotOrientation.VERTICAL, _chart.isShowLegend(), false, false);

			XYPlot				   plot = (XYPlot)_chartMgr.getPlot();
			StackedXYAreaRenderer2 renderer = new StackedXYAreaRenderer2();

			plot.setRenderer(0, renderer);

			if (_chart.getXType() == LyChart.Type.CHARTDATE)
			{
				DateAxis dateAxis = new DateAxis();

				plot.setDomainAxis(0, dateAxis);
			}
		}

		Plot bPlot = _chartMgr.getPlot();

		// Specific behaviors for XY plots
		if (bPlot instanceof XYPlot)
		{
			XYItemRenderer renderer = null;

			if (_chart.getGraphType() == LyChartBase.GraphType.SCATTER_PLOT)
			{
				renderer = new LyXYLineAndShapeRenderer(_paints);
				((XYPlot)bPlot).setRenderer(renderer);

				int		   i, n = xyDataset.getSeriesCount(), j, m;
				XYSeries   series;
				XYDataItem item;
				float	   xMin = Float.MAX_VALUE, xMax = Float.MIN_VALUE, x;
				float	   yMin = Float.MAX_VALUE, yMax = Float.MIN_VALUE, y;

				for (i = 0; i < n; i++)
				{
					series = xyDataset.getSeries(i);
					m = series.getItemCount();

					for (j = 0; j < m; j++)
					{
						item = series.getDataItem(j);
						x = item.getX().floatValue();
						y = item.getY().floatValue();

						xMin = Math.min(xMin, x);
						yMin = Math.min(yMin, y);
						xMax = Math.max(xMax, x);
						yMax = Math.max(yMax, y);
					}
				}

				float ratio = (_chart.isShowTips() ? 0.08F : 0.02F);
				float smallRatio = 0.02F;

				xMin = Math.min(xMin - ((xMax - xMin) * ratio), 0);
				xMax = xMax + ((xMax - xMin) * ratio);
				yMin = Math.min(yMin - ((yMax - yMin) * smallRatio), 0);
				yMax = yMax + ((yMax - yMin) * (ratio + smallRatio));

				NumberAxis rAxis = (NumberAxis)((XYPlot)bPlot).getRangeAxis();

				rAxis.setAutoRange(true);
				rAxis.setLowerBound(yMin);
				rAxis.setUpperBound(yMax);

				NumberAxis dAxis = (NumberAxis)((XYPlot)bPlot).getDomainAxis();

				dAxis.setAutoRange(true);
				dAxis.setLowerBound(xMin);
				dAxis.setUpperBound(xMax);

		        if (_chart.isShowLegend())
	            {
		        	// Change the items of the legend
		            LegendTitle				legend = _chartMgr.getLegend();
					ValueAxis				cAxis = ((XYPlot)bPlot).getDomainAxis();
	                LegendItemCollection	items = new LegendItemCollection();
	    			LegendItemCollection	legendItems = renderer.getLegendItems();
	                String					label;

	                if ((legendItems != null)
	                		&& (legendItems.getItemCount() > 0))
	                {
		                LegendItem		legendItem = legendItems.get(0), newItem;

		                if (legendItem != null)
		                {
		                    for (i = 0; i < xyDataset.getSeriesCount(); i++)
		                    {
		    	                legendItem = renderer.getLegendItem(0, i);
		                    	label = xyDataset.getSeries(i).getKey().toString();
		                    	
		                    	String[]	labels = LyString.stringToList(label, SCATTER_PLOT_INTERNAL_LABEL_SEPARATOR);
		                    	
		                    	for (int k = 0; k < labels.length; k++)
		                    	{
			                    	label = labels[k];
			                    	
			                    	if (_specificLegendValues.get(label) != null)
			                    		label = _specificLegendValues.get(label);
			                    	
			                       	newItem = new LegendItem(label, null, null, null, legendItem.getShape(),
			                        	_paints[i % _paints.length], legendItem.getOutlineStroke(), legendItem.getOutlinePaint());
		    	                    
		        	                items.add(newItem);
		                    	}
		                    }
		    	            
		                    ((XYPlot)bPlot).setFixedLegendItems(items);
		                }
	                }
	                
	                legend.setItemFont(cAxis.getTickLabelFont());
	            }
			}
			else
				renderer = ((XYPlot)bPlot).getRenderer();
			LyAxis	axis = _chart.getAxis();
			
			if (axis != null)
			{
				NumberAxis rAxis = (NumberAxis)((XYPlot)bPlot).getRangeAxis();
				
				if ((axis.getYBegin() != null) && (axis.getYBegin().length() > 0))
					rAxis.setLowerBound(Double.parseDouble(axis.getYBegin()));
	
				if ((axis.getYEnd() != null) && (axis.getYEnd().length() > 0))
					rAxis.setUpperBound(Double.parseDouble(axis.getYEnd()));
				
				if ((axis.getYWidth() != null) && (axis.getYWidth().length() > 0))
					rAxis.setTickUnit(new NumberTickUnit(Double.parseDouble(axis.getYWidth())));
				
				ValueAxis dAxis = ((XYPlot)bPlot).getDomainAxis();
				
				if (dAxis instanceof NumberAxis)
				{
					if ((axis.getXBegin() != null) && (axis.getXBegin().length() > 0))
						dAxis.setLowerBound(Double.parseDouble(axis.getXBegin()));
	
					if ((axis.getXEnd() != null) && (axis.getXEnd().length() > 0))
						dAxis.setUpperBound(Double.parseDouble(axis.getXEnd()));
					
					if ((axis.getXWidth() != null) && (axis.getXWidth().length() > 0))
						((NumberAxis)dAxis).setTickUnit(new NumberTickUnit(Double.parseDouble(axis.getXWidth())));
				}
				else if (dAxis instanceof DateAxis)
				{
					if ((axis.getXBegin() != null) && (axis.getXBegin().length() > 0))
						((DateAxis)dAxis).setMinimumDate(new Date(Long.parseLong(axis.getXBegin())));
					
					if ((axis.getXEnd() != null) && (axis.getXEnd().length() > 0))
						((DateAxis)dAxis).setMaximumDate(new Date(Long.parseLong(axis.getXEnd())));
					
					if ((axis.getXWidth() != null) && (axis.getXWidth().length() > 0))
					{
						long	width = Long.parseLong(axis.getXWidth());
						
						((DateAxis)dAxis).setTickUnit(new DateTickUnit(DateTickUnitType.MINUTE, (int)(width / 60000)), true, true);
						
						// If width lower than 3 hours, use time formatter
						if (width < 10800000L)
							((DateAxis)dAxis).setDateFormatOverride(DateFormat.getTimeInstance(DateFormat.SHORT, _chart.getEnvironment().getLocale()));
						// If width lower than 24 hours, use date and time formatter
						else if (width < 86400000L)
							((DateAxis)dAxis).setDateFormatOverride(DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.SHORT, _chart.getEnvironment().getLocale()));
						// If width greater than 24 hours, use date formatter
						else
							((DateAxis)dAxis).setDateFormatOverride(DateFormat.getDateInstance(DateFormat.SHORT, _chart.getEnvironment().getLocale()));
					}
				}
			}
			
			((XYPlot)bPlot).setDomainGridlinesVisible(_chart.isXGrid());
			((XYPlot)bPlot).setRangeGridlinesVisible(_chart.isYGrid());

	        if (_chart.isD3Effect() && (renderer instanceof XYLineAndShapeRenderer))
	        	((XYLineAndShapeRenderer)renderer).setShapesVisible(true);
			
			if (_chart.isShowTips())
			{
				renderer.setItemLabelsVisible(true);
				renderer.setItemLabelGenerator(new StandardXYItemLabelGenerator("{0}",
						NumberFormat.getNumberInstance(), NumberFormat.getNumberInstance()));
			}

			// Generation of imagemap
			renderer.setURLGenerator(new LyStandardXYURLGenerator());

			// Generation of tooltips
			renderer.setToolTipGenerator(new LyStandardXYTooltipGenerator());
		}

		// Specific behaviors for category plots
		if (bPlot instanceof CategoryPlot)
		{
			boolean columnChangeColor = (_chart.getColumnCount() <= 1);

			((CategoryPlot)bPlot).setDomainGridlinesVisible(_chart.isXGrid());
			((CategoryPlot)bPlot).setRangeGridlinesVisible(_chart.isYGrid());

			LyAxis axis = _chart.getAxis();

			if (axis != null)
			{
				// If Y axis is count type, generate 10 thicks with only integer values
				if (((_chart.getYVal() == LyChart.YVal.Y_COUNT_INTERVAL)
							|| (_chart.getYVal() == LyChart.YVal.Y_COUNT_REAL))
						&& ((CategoryPlot)bPlot).getRangeAxis() instanceof NumberAxis)
				{
					NumberAxis		numberAxis = (NumberAxis)((CategoryPlot)bPlot).getRangeAxis();
					int				nbThicks = 10;
					double			diff = numberAxis.getUpperBound() - numberAxis.getLowerBound();

					diff = diff / nbThicks;

					double	diffTmp = diff;
					int count = 0;
					
					while (diffTmp > 10)
					{
						diffTmp = diffTmp / 10;
						count++;
					}
					
					if (diffTmp <= 1)
						diffTmp = 1;
					else if (diffTmp <= 2)
						diffTmp = 2;
					else if (diffTmp <= 5)
						diffTmp = 5;
					else
						diffTmp = 10;
					
					for (int i = 0; i < count; i++)
						diffTmp = diffTmp * 10;
					
					NumberTickUnit	ntu = new NumberTickUnit(diffTmp);
					
					numberAxis.setTickUnit(ntu);
				}
				
				if ((((CategoryPlot)bPlot).getDomainAxis() != null) 
						&& (MAX_LABEL_LINES > 0))
					((CategoryPlot)bPlot).getDomainAxis().setMaximumCategoryLabelLines(MAX_LABEL_LINES);
				
				if ((axis.getYBegin() != null) && (axis.getYBegin().length() > 0))
					((CategoryPlot)bPlot).getRangeAxis().setLowerBound(Double.parseDouble(axis.getYBegin()));

				if ((axis.getYEnd() != null) && (axis.getYEnd().length() > 0))
					((CategoryPlot)bPlot).getRangeAxis().setUpperBound(Double.parseDouble(axis.getYEnd()));
			}

			BarRenderer renderer = null;

			if (_chart.getGraphType() == LyChartBase.GraphType.BAR)
			{
				if (_chart.isD3Effect())
					renderer = new LyBarRenderer3D(_paints, columnChangeColor);
				else
					renderer = new LyBarRenderer(_paints, columnChangeColor);
			}
			else if (_chart.getGraphType() == LyChartBase.GraphType.STACK_BAR)
			{
				if (_chart.isD3Effect())
					renderer = new LyStackedBarRenderer3D(_paints, columnChangeColor);
				else
					renderer = new LyStackedBarRenderer(_paints, columnChangeColor);
			}

			if (renderer != null)
			{
				renderer.setGradientPaintTransformer(new StandardGradientPaintTransformer(
						GradientPaintTransformType.VERTICAL));
				((CategoryPlot)bPlot).setRenderer(renderer);
				renderer.setOutlinePaint(Color.gray);

				if (columnChangeColor
						&& _chart.isShowLegend()
						&& (renderer.getColumnCount() > 0)
						&& (renderer.getRowCount() > 0))
				{
					// Change the items of the legend
					LegendTitle			 legend = _chartMgr.getLegend();
					CategoryAxis		 cAxis = ((CategoryPlot)bPlot).getDomainAxis();
					LegendItemCollection items = new LegendItemCollection();
	    			LegendItemCollection legendItems = renderer.getLegendItems();
					String				 label;

					if ((legendItems != null)
							&& (legendItems.getItemCount() > 0))
					{
						LegendItem	item = legendItems.get(0), newItem;
	
						if (item != null)
						{
							for (int i = 0; i < _points.length; i++)
							{
								label = _points[i];
	
								if (_specificLegendValues.get(label) != null)
									label = _specificLegendValues.get(label);
	
								newItem = new LegendItem(label, null, null, null, item.getShape(),
										_paints[i % _paints.length], item.getOutlineStroke(),
										item.getOutlinePaint());
	
								items.add(newItem);
							}
	
							((CategoryPlot)bPlot).setFixedLegendItems(items);
						}
					}
					
					legend.setItemFont(cAxis.getTickLabelFont());
				}

				// Generation of imagemap
				renderer.setItemURLGenerator(new LyStandardCategoryURLGenerator());

				NumberAxis	nAxis = (NumberAxis)((CategoryPlot)bPlot).getRangeAxis();
	        	
	        	nAxis.setAutoRange(true);
	        	
				if (_chart.isShowTips())
				{
					nAxis.setUpperMargin(0.08);
		        	nAxis.setLowerMargin(0.08);
			        renderer.setItemLabelsVisible(true);
			        renderer.setItemLabelGenerator(new StandardCategoryItemLabelGenerator());
			        
					if (_chart.isRatios())
						renderer.setItemLabelGenerator(new LyCategoryItemLabelGenerator());
				}
			}
		}

		if (bPlot instanceof PiePlot)
		{
			if (_chart.isShowTips())
			{
				if (_chart.isRatios())
					((PiePlot)bPlot).setLabelGenerator(new StandardPieSectionLabelGenerator("{0} : {2}", 
			                NumberFormat.getNumberInstance(), 
			                NumberFormat.getPercentInstance()));
				else
					((PiePlot)bPlot).setLabelGenerator(new StandardPieSectionLabelGenerator("{0} : {1}", 
			                NumberFormat.getNumberInstance(), 
			                NumberFormat.getNumberInstance()));
			}
			else
				((PiePlot)bPlot).setLabelGenerator(null);

			// Generation of imagemap
			((PiePlot)bPlot).setURLGenerator(new LyStandardPieURLGenerator());
		}

		if (bPlot instanceof MultiplePiePlot)
		{
			JFreeChart	chart = ((MultiplePiePlot)bPlot).getPieChart();
			PiePlot		piePlot = (PiePlot)chart.getPlot();
			
			if (_chart.isShowTips())
			{
				if (_chart.isRatios())
					piePlot.setLabelGenerator(new StandardPieSectionLabelGenerator("{0} : {2}", 
			                NumberFormat.getNumberInstance(), 
			                NumberFormat.getPercentInstance()));
				else
					piePlot.setLabelGenerator(new StandardPieSectionLabelGenerator("{0} : {1}", 
			                NumberFormat.getNumberInstance(), 
			                NumberFormat.getNumberInstance()));
			}
			else
				piePlot.setLabelGenerator(null);

			// Generation of imagemap
			piePlot.setURLGenerator(new LyStandardPieURLGenerator());
		}

		if (_chart.getGraphType() == LyChartBase.GraphType.PLOT)
		{
			XYLineAndShapeRenderer renderer = (XYLineAndShapeRenderer)((XYPlot)bPlot).getRenderer();

			if (_chart.isShowTips())
			{
				renderer.setItemLabelsVisible(true);
				renderer.setItemLabelGenerator(new StandardXYItemLabelGenerator());
			}
		}

		if (_chart.getGraphType() == LyChartBase.GraphType.STACK_AREA)
		{
			StackedXYAreaRenderer2 renderer = (StackedXYAreaRenderer2)((XYPlot)bPlot).getRenderer();

			renderer.setURLGenerator(new LyStandardXYURLGenerator());
		}

		// Checks if Labels should display Horizontally or vertically
		if (_chart.isXVerticalLabels())
		{
			CategoryAxis cAxis = null;

			if (bPlot instanceof CategoryPlot)
				cAxis = ((CategoryPlot)bPlot).getDomainAxis();
			else if (bPlot instanceof XYPlot)
			{
				ValueAxis vAxis = ((XYPlot)bPlot).getDomainAxis();

				vAxis.setVerticalTickLabels(true);
			}

			if (cAxis != null)
				cAxis.setCategoryLabelPositions(CategoryLabelPositions
						.createUpRotationLabelPositions(Math.PI / 2.0));
		}

		// Convert data to Percentiles when needed
		if (_chart.isRatios())
		{
			ValueAxis axis = null;

			if (bPlot instanceof XYPlot)
				axis = ((XYPlot)bPlot).getRangeAxis();
			else if (bPlot instanceof CategoryPlot)
				axis = ((CategoryPlot)bPlot).getRangeAxis();

			if (axis instanceof NumberAxis)
			{
				DecimalFormat formatter = new DecimalFormat("0.##'%'");

				((NumberAxis)axis).setTickUnit(new NumberTickUnit(5.0, formatter));
			}
		}

		// Set the transparency
		bPlot.setForegroundAlpha(ALPHA_RATIO);

		LyColor background = _chart.getEnvironment().getLyColor("LY_DEFAULT_BACKGROUND");

		if (!background.useSystemColor())
			_chartMgr.setBackgroundPaint(new Color(background.getRed(), background.getGreen(),
					background.getBlue()));
	} // end method reset

	/**
	 * Creates and returns a category dataset for the charts.
	 */
	public CategoryDataset createCategoryDataset()
	{
		try
		{
			Number[][] dataValues = null;
			String[]   seriesNames = null;
			int		   i, j, nValues;

			// In the data we have couples x1, y1, x2, y2 and so on : read every 2 to get y
			for (i = 1; i < _data.size(); i = i + 2)
			{
				nValues = _data.elementAt(i).size();

				if (dataValues == null)
				{
					dataValues = new Number[_data.size() / 2][nValues];
					seriesNames = new String[_data.size() / 2];
				}

				seriesNames[i / 2] = _seriesLabel.elementAt(i / 2);

				for (j = 0; j < nValues; j++)
				{
					Object chartValue = _data.elementAt(i).elementAt(j);

					Number v_n = null;

					if (chartValue instanceof Number)
						v_n = (Number)chartValue; // We already have a numeric value
					else if (chartValue != null)
					{
						if (v_n == null)
							v_n = LyTools.DoubleFromString(chartValue.toString());
					}

					if ((v_n != null)
							&& (dataValues[i / 2].length > j))
						dataValues[i / 2][j] = v_n;
				}
			}

			DefaultCategoryDataset dataset = new DefaultCategoryDataset();

			for (i = 0; i < ((seriesNames == null) ? 0 : seriesNames.length); i++)
			{
				for (j = 0; j < ((_points == null) ? 0 : _points.length); j++)
				{
					if (_chart.isDisplayLegendEmptyValues() || dataValues[i][j].floatValue() != 0)
						dataset.setValue(dataValues[i][j], seriesNames[i], _points[j]);
					else if (_chart.getGraphType() == LyChartBase.GraphType.PIE)
						dataset.setValue(Integer.MIN_VALUE, seriesNames[i], _points[j]);
				}
			}

			return dataset;
		}
		catch (Exception e)
		{
			_chart.getEnvironment().writeException(e);
		}

		return null;
	} // end method createCategoryDataset

	/**
	 * Creates and returns a XYDataset for the demo charts.
	 */
	public XYSeriesCollection createXYDataset(boolean groupValues)
	{
		try
		{
			Object[][][] dataValues = null;
			String[]     seriesNames = null;

			int			 i, j, k, nValues;

			// In the data we have couples x1, y1, x2, y2 and so on : read every 2 to get [x, y]
			for (i = 1; i < _data.size(); i = i + 2)
			{
				nValues = _data.elementAt(i).size();

				if (dataValues == null)
				{
					dataValues = new Object[_data.size() / 2][nValues][2];
					seriesNames = new String[_data.size() / 2];
				}

				seriesNames[i / 2] = _seriesLabel.elementAt(i / 2);

				for (j = 0; j < nValues; j++)
				{
					// Get X value
					Object chartValue = _data.elementAt(i - 1).elementAt(j);

					Number v_n = null;

					if (chartValue instanceof Number)
						v_n = (Number)chartValue; // We already have a numeric value
					else if (chartValue != null)
					{
						// Try to parse a numeric valuie from data
						LyTools.IntegerFromString(chartValue.toString());

						if (v_n == null)
							v_n = LyTools.DoubleFromString(chartValue.toString());
					}

					if (v_n != null)
						dataValues[i / 2][j][0] = v_n;

					// Get Y value
					chartValue = _data.elementAt(i).elementAt(j);
					v_n = null;

					if (chartValue instanceof Number)
						v_n = (Number)chartValue; // We already have a numeric value
					else if (chartValue != null)
					{
						// Try to parse a numeric valuie from data
						LyTools.IntegerFromString(chartValue.toString());

						if (v_n == null)
							v_n = LyTools.DoubleFromString(chartValue.toString());
					}

					if (v_n != null)
						dataValues[i / 2][j][1] = v_n;
				}
			}

			XYSeriesCollection dataset = new XYSeriesCollection();
			java.util.Vector<LyScatterPlotData>  scatterPlotData = new java.util.Vector<LyScatterPlotData>();
			int				   seriesNameLength = ((seriesNames == null) ? 0 : seriesNames.length);

			// Check if some points have exactly the same coordonates
			for (i = 0; i < seriesNameLength; i++)
			{
				LyScatterPlotData data = new LyScatterPlotData(seriesNames[i]), dataTmp;
				Double			  x, y;
				boolean			  valueAlreadyExists, hasValue = false;

				for (j = 0; j < dataValues[i].length; j++)
				{
					valueAlreadyExists = false;
					x = (Double)dataValues[i][j][0];
					y = (Double)dataValues[i][j][1];

					// Check if this point already exists
					for (k = 0; k < scatterPlotData.size(); k++)
					{
						dataTmp = scatterPlotData.elementAt(k);
						valueAlreadyExists = (groupValues ? dataTmp.hasValue(x, y) : false);

						// If point already exists, concatenate the labels
						if (valueAlreadyExists)
						{
							dataTmp.setLabel(dataTmp.getLabel() + SCATTER_PLOT_INTERNAL_LABEL_SEPARATOR
								+ data.getLabel());
							break;
						}
					}

					if (!valueAlreadyExists)
					{
						data.addValue((Double)dataValues[i][j][0], (Double)dataValues[i][j][1]);
						hasValue = true;
					}
				}

				if (hasValue)
					scatterPlotData.addElement(data);
			}

			// Create JFreeChart series
			for (i = 0; i < scatterPlotData.size(); i++)
			{
				LyScatterPlotData dataTmp = scatterPlotData.elementAt(i);
				XYSeries		  series = new XYSeries(dataTmp.getLabel());
				Double[][]		  values = dataTmp.getValues();

				for (j = 0; j < values.length; j++)
					series.add(values[j][0], values[j][1]);

				dataset.addSeries(series);
			}

			return dataset;
		}
		catch (Exception e)
		{
			_chart.getEnvironment().writeException(e);
		}

		return null;
	} // end method createXYDataset

	/**
	 * Creates and returns a XYDataset for the demo charts.
	 */
	public DefaultTableXYDataset createTableXYDataset()
	{
		try
		{
			Object[][][] dataValues = null;
			String[]     seriesNames = null;

			int			 i, j, nValues;

			// In the data we have couples x1, y1, x2, y2 and so on : read every 2 to get [x, y]
			for (i = 1; i < _data.size(); i = i + 2)
			{
				nValues = _data.elementAt(i).size();

				if (dataValues == null)
				{
					dataValues = new Object[_data.size() / 2][nValues][2];
					seriesNames = new String[_data.size() / 2];
				}

				seriesNames[i / 2] = _seriesLabel.elementAt(i / 2);

				for (j = 0; j < nValues; j++)
				{
					for (j = 0; j < nValues; j++)
					{
						// Get X value
						Object chartValue = _data.elementAt(i - 1).elementAt(j);

						Number v_n = null;

						if (chartValue instanceof Number)
							v_n = (Number)chartValue; // We already have a numeric value
						else if (chartValue != null)
						{
							// Try to parse a numeric valuie from data
							LyTools.IntegerFromString(chartValue.toString());

							if (v_n == null)
								v_n = LyTools.DoubleFromString(chartValue.toString());
						}

						if (v_n != null)
							dataValues[i / 2][j][0] = v_n;

						// Get Y value
						chartValue = _data.elementAt(i).elementAt(j);
						v_n = null;

						if (chartValue instanceof Number)
							v_n = (Number)chartValue; // We already have a numeric value
						else if (chartValue != null)
						{
							// Try to parse a numeric valuie from data
							LyTools.IntegerFromString(chartValue.toString());

							if (v_n == null)
								v_n = LyTools.DoubleFromString(chartValue.toString());
						}

						if (v_n != null)
							dataValues[i / 2][j][1] = v_n;
					}
				}
			}

			DefaultTableXYDataset dataset = new DefaultTableXYDataset();
			java.util.Vector<LyScatterPlotData>	scatterPlotData = new java.util.Vector<LyScatterPlotData>();
			int					  seriesNameLength = ((seriesNames == null) ? 0
																			: seriesNames.length);

			for (i = 0; i < seriesNameLength; i++)
			{
				LyScatterPlotData data = new LyScatterPlotData(seriesNames[i]);
				boolean			  hasValue = false;

				for (j = 0; j < dataValues[i].length; j++)
				{
					data.addValue((Double)dataValues[i][j][0], (Double)dataValues[i][j][1]);
					hasValue = true;
				}

				if (hasValue)
					scatterPlotData.addElement(data);
			}

			// Create JFreeChart series
			for (i = 0; i < scatterPlotData.size(); i++)
			{
				LyScatterPlotData dataTmp = scatterPlotData.elementAt(i);
				XYSeries		  series = new XYSeries(dataTmp.getLabel(), true, false);
				Double[][]		  values = dataTmp.getValues();

				for (j = 0; j < values.length; j++)
					series.add(values[j][0], values[j][1]);

				dataset.addSeries(series);
			}

			return dataset;
		}
		catch (Exception e)
		{
			_chart.getEnvironment().writeException(e);
		}

		return null;
	} // end method createTableXYDataset

	/**
	 * Set a specific palette for colors used in chart component.
	 *
	 * @param paints the array of new colors that must be used in the chart component
	 */
	@SuppressWarnings("deprecation")
	public void setPaints(Paint[] paints)
	{
		_paints = paints;

		Plot bPlot = _chartMgr.getPlot();

		if (bPlot instanceof CategoryPlot)
		{
			CategoryItemRenderer renderer = ((CategoryPlot)bPlot).getRenderer();

			if (renderer instanceof LyBarRenderer3D)
				((LyBarRenderer3D)renderer).setPaints(paints);
			else if (renderer instanceof LyBarRenderer)
				((LyBarRenderer)renderer).setPaints(paints);
			else if (renderer instanceof LyStackedBarRenderer3D)
				((LyStackedBarRenderer3D)renderer).setPaints(paints);
			else if (renderer instanceof LyStackedBarRenderer)
				((LyStackedBarRenderer)renderer).setPaints(paints);
		}
		else if (bPlot instanceof PiePlot)
		{
			for (int i = 0; i < paints.length; i++)
			{
				((PiePlot)bPlot).setSectionPaint(i, paints[i]);
			}
		}
		else if (bPlot instanceof XYPlot)
		{
			XYItemRenderer	renderer = ((XYPlot)bPlot).getRenderer(0);
			
			for (int i = 0; i < paints.length; i++)
			{
				renderer.setSeriesPaint(i, paints[i]);
			}
		}
	}
	
	/**
	 * Returns the specific palette for colors used in chart component.
	 */
	public Paint[] getPaints()
	{
		return _paints;
	}

	/**
	 * Change the background color of the chart component.
	 *
	 * @param color the new background color
	 */
	public void setBackgroundColor(Color color)
	{
		_chartMgr.setBackgroundPaint(color);
	}

	/**
	 * Set the specific legend values.
	 *
	 * @param values
	 */
	public void setSpecificLegendValue(Hashtable<String, String> values)
	{
		super.setSpecificLegendValue(values);

		Plot bPlot = _chartMgr.getPlot();

		if (bPlot instanceof CategoryPlot)
		{
			CategoryItemRenderer itemRenderer = ((CategoryPlot)bPlot).getRenderer();

			if (!(itemRenderer instanceof BarRenderer))
				return;

			boolean              columnChangeColor = (_chart.getColumnCount() <= 1);
			BarRenderer          renderer = (BarRenderer)itemRenderer;
			Paint[]              paints = ChartColor.createDefaultPaintArray();
			LegendItemCollection legendItems = renderer.getLegendItems();

			if (columnChangeColor
				&& _chart.isShowLegend()
				&& (legendItems != null)
				&& (legendItems.getItemCount() > 0))
			{
				// Change the items of the legend
				LegendTitle			 legend = _chartMgr.getLegend();
				CategoryAxis		 cAxis = ((CategoryPlot)bPlot).getDomainAxis();
				LegendItemCollection items = new LegendItemCollection();
				LegendItem			 item = legendItems.get(0), newItem;
				String				 label;

				if (item != null)
				{
					for (int i = 0; i < _points.length; i++)
					{
						label = _points[i];

						if (_specificLegendValues.get(label) != null)
							label = _specificLegendValues.get(label);

						newItem = new LegendItem(label, null, null, null, item.getShape(),
								paints[i % paints.length], item.getOutlineStroke(),
								item.getOutlinePaint());

						items.add(newItem);
					}

					((CategoryPlot)bPlot).setFixedLegendItems(items);
				}

				legend.setItemFont(cAxis.getTickLabelFont());
			}
		}
		else if (bPlot instanceof XYPlot)
		{
	        if (_chart.isShowLegend())
            {
	        	// Change the items of the legend
	        	XYPlot					xyPlot = (XYPlot)bPlot;
				XYItemRenderer			itemRenderer = xyPlot.getRenderer();
	            LegendTitle				legend = _chartMgr.getLegend();
				ValueAxis				cAxis = xyPlot.getDomainAxis();
				LegendItemCollection	legendItems = itemRenderer.getLegendItems();
                String					label;

                if ((legendItems != null)
    				&& (legendItems.getItemCount() > 0))
                {
                    LegendItemCollection items = new LegendItemCollection();
    				LegendItem			 legendItem, newItem;
                	XYDataset			 xyDataset = xyPlot.getDataset();
                	
                    for (int i = 0; i < xyDataset.getSeriesCount(); i++)
                    {
    	                legendItem = itemRenderer.getLegendItem(0, i);
                    	label = xyDataset.getSeriesKey(i).toString();
                    	
                    	String[]	labels = LyString.stringToList(label, SCATTER_PLOT_INTERNAL_LABEL_SEPARATOR);
                    	
                    	for (int k = 0; k < labels.length; k++)
                    	{
	                    	label = labels[k];
	                    	
	                    	if (_specificLegendValues.get(label) != null)
	                    		label = _specificLegendValues.get(label);
	                    	
	                       	newItem = new LegendItem(label, null, null, null, legendItem.getShape(),
	                        	_paints[i % _paints.length], legendItem.getOutlineStroke(), legendItem.getOutlinePaint());
    	                    
        	                items.add(newItem);
                    	}
                    }
    	            
                    xyPlot.setFixedLegendItems(items);
                }

                legend.setItemFont(cAxis.getTickLabelFont());
            }
		}
	} // end method setSpecificLegendValue

	/**
	 * Sets the legendRightPosition parameter.
	 *
	 * @param rightPosition the new legendRightPosition value
	 */
	public void setLegendRightPosition(boolean rightPosition)
	{
		super.setLegendRightPosition(rightPosition);

		// Set the legend position
		if (_chart.isShowLegend() && _legendRightPosition)
		{
			LegendTitle legend = _chartMgr.getLegend();

			legend.setHorizontalAlignment(HorizontalAlignment.RIGHT);
			legend.setVerticalAlignment(VerticalAlignment.CENTER);
			legend.setPosition(RectangleEdge.RIGHT);
		}
	}

	/**
	 * Retrieve the beginning of the label when it has more than one line.
	 *
	 * @return for example : null for String, and <html>... for HTML
	 */
	public String getMultiLineBegin()
	{
		return "";
	}

	/**
	 * Retrieve the end of the label when it has more than one line.
	 *
	 * @return for example : null for String, and ... for HTML
	 */
	public String getMultiLineEnd()
	{
		return "";
	}

	/**
	 * Retrieve the line separator of the label when it has more than one line.
	 *
	 * @return for example : \n for String, and /p> <p for HTML
	 */
	public String getMultiLineSep()
	{
		return (" ");
	}

	/**
	 * Save the image of the chart.
	 *
	 * @param tmpDir    temporary directory
	 * @param imageFile the PNG File
	 * @param buffer    the buffer where imagemap must be generated
	 */
	public void saveImg(String tmpDir, String imageFile, StringBuffer buffer, boolean deleteOnExit)
	{
		File			   f = null;
		try
		{
			f = new File(imageFile);
			
			if (deleteOnExit)
				f.deleteOnExit(); // temp file
			
			ChartRenderingInfo info = new ChartRenderingInfo(new StandardEntityCollection());
			
			// Retrieve chat dimension to generate an image with the same size as the one
			// displayed on the screen.
			LyGetChartConverterInterface	internal = (LyGetChartConverterInterface)_chart.getInternalView();
			int[]							dimension = internal.getComponentDimension();
			
			if (dimension == null)
			    dimension = new int[] { _chartWidth, _chartHeight };
			else
			{
			    if (dimension[0] <= 0)
			        dimension[0] = _chartWidth;
                if (dimension[1] <= 0)
                    dimension[1] = _chartHeight;
			}
			
			ChartUtilities.saveChartAsPNG(f, getFreeChart(), dimension[0], dimension[1], info);

			if (buffer != null)
			{
				String imageMap = ImageMapUtilities.getImageMap("chart_imagemap", info,
						new StandardToolTipTagFragmentGenerator(),
						new StandardURLTagFragmentGenerator());

				buffer.append(imageMap);
			}
		}
		catch (IOException e)
		{
			if (_chart.getEnvironment().isDebugLevel(LyEnvironment.DEBUG_ERROR))
			{
				_chart.getEnvironment().logErr("[LyCommonfreeChartConverter] error writing file : "+(f == null ? "null" : f.getAbsolutePath()));
				_chart.getEnvironment().writeException(e);
			}
		}
	}

	/**
	 * Generate an URL for a given source id (used in web mode only).
	 *
	 * @param  sourceId the id of the source
	 * @return the URL used to call the web server
	 */
	public String generateUrl(String sourceId)
	{
		if (!generateSensitiveAreas())
			return "";

		String controllerId = _chart.getRootController();
		String url = "javascript:InvokeAction('";

		sourceId = sourceId.replace('\'', '_');

		url += LyString.encodeURL(_chart.getId()) + "','";
		url += LyString.encodeURL("/event.do','reload_frames','");
		url += controllerId + "','" + sourceId + "');";

		return url;
	}

	/**
	 * Retreives the Internal Free chart component.
	 *
	 * @return the chart component
	 */
	public JFreeChart getFreeChart()
	{
		if (_chartMgr == null)
			setContent();

		return _chartMgr;
	}

	/* Internal Class*/

	/**
	 * A custom renderer that returns a different color for each item in a single series.
	 */
	static class LyBarRenderer extends BarRenderer
	{

		/**
		 * 
		 */
		private static final long serialVersionUID = -6433207801896348419L;

		/** The colors to be used. */
		private Paint[] paints;

		private Paint[] colors;
		/** Indicates if the alternance of colors is made with columns. */
		private boolean columnChangeColor;

		/**
		 * Creates a new renderer.
		 *
		 * @param colors            the colors
		 * @param columnChangeColor indicates if the alternance of colors is made with columns
		 */
		public LyBarRenderer(Paint[] colors, boolean columnChangeColor)
		{
			setPaints(colors);
			this.columnChangeColor = columnChangeColor;
		}

		/**
		 * Set the colors used to fill columns.
		 *
		 * @param colors the array of colors
		 */
		public void setPaints(Paint[] colors)
		{
			paints = new Paint[colors.length];

			for (int i = 0; i < colors.length; i++)
				paints[i] = new GradientPaint(0f, 0f,
						((Color)colors[i]).brighter().brighter().brighter(), 0f, 0f,
						(Color)colors[i]);
			this.colors = colors;
		}

		/**
		 * Returns the paint for an item. Overrides the default behaviour inherited from
		 * AbstractSeriesRenderer.
		 *
		 * @param  row    the series.
		 * @param  column the category.
		 * @return The item color.
		 */
		public Paint getItemPaint(int row, int column)
		{
			if (columnChangeColor)
				return paints[column % this.paints.length];
			else
				return paints[row % this.paints.length];
		}

		/**
		 * Returns a legend item for a series.
		 * 
		 * @param datasetIndex - the dataset index (zero-based).
		 * @param series - the series index (zero-based). 
		 */
		public LegendItem getLegendItem(int datasetIndex, int series)
		{	
			LegendItem	item = super.getLegendItem(datasetIndex, series); 
			Paint		color = colors[series % this.paints.length];
			
			LyLegendItem	li = new LyLegendItem(item.getLabel(), item.getDescription(), item.getToolTipText(),
					item.getURLText(), item.isShapeVisible(), item.getShape(), item.isShapeFilled(), color,
					item.isShapeOutlineVisible(), item.getOutlinePaint(), item.getOutlineStroke(), item.isLineVisible(),
					item.getLine(), item.getLineStroke(), item.getLinePaint());
			
			return li;
		}
	} // end class LyBarRenderer


	/**
	 * A custom renderer that returns a different color for each item in a single series.
	 */
	class LyBarRenderer3D extends BarRenderer3D
	{

		/**
		 * 
		 */
		private static final long serialVersionUID = -5124308726027226293L;

		/** The paints to be used. */
		private Paint[] paints;

		/** The colors to be used. */
		private Paint[] colors;

		/** Indicates if the alternance of colors is made with columns. */
		private boolean columnChangeColor;

		/**
		 * Creates a new renderer.
		 *
		 * @param colors            the colors
		 * @param columnChangeColor indicates if the alternance of colors is made with columns
		 */
		public LyBarRenderer3D(Paint[] colors, boolean columnChangeColor)
		{
			setPaints(colors);
			this.columnChangeColor = columnChangeColor;
		}

		/**
		 * Set the colors used to fill columns.
		 *
		 * @param colors the array of colors
		 */
		public void setPaints(Paint[] colors)
		{
			paints = new Paint[colors.length];

			for (int i = 0; i < colors.length; i++)
				paints[i] = new GradientPaint(0f, 0f,
						((Color)colors[i]).brighter().brighter().brighter(), 0f, 0f,
						(Color)colors[i]);

			this.colors = colors;
		}

		/**
		 * Returns the paint for an item. Overrides the default behaviour inherited from
		 * AbstractSeriesRenderer.
		 *
		 * @param  row    the series.
		 * @param  column the category.
		 * @return The item color.
		 */
		public Paint getItemPaint(int row, int column)
		{
			if (columnChangeColor)
				return paints[column % this.paints.length];
			else
				return paints[row % this.paints.length];
		}

		/**
		 * Returns the paint for an item. Overrides the default behaviour inherited from
		 * AbstractSeriesRenderer.
		 *
		 * @param  row    the series.
		 * @param  column the category.
		 * @return The item color.
		 */
		public Paint getItemColor(int row, int column)
		{
			if (columnChangeColor)
				return colors[column % this.paints.length];
			else
				return colors[row % this.paints.length];
		}

		/**
		 * Returns a legend item for a series.
		 * 
		 * @param datasetIndex - the dataset index (zero-based).
		 * @param series - the series index (zero-based). 
		 */
		public LegendItem getLegendItem(int datasetIndex, int series) {
			
			LegendItem	item = super.getLegendItem(datasetIndex, series); 
			Paint		color = colors[series % this.paints.length];
			
			LyLegendItem	li = new LyLegendItem(item.getLabel(), item.getDescription(), item.getToolTipText(),
					item.getURLText(), item.isShapeVisible(), item.getShape(), item.isShapeFilled(), color,
					item.isShapeOutlineVisible(), item.getOutlinePaint(), item.getOutlineStroke(), item.isLineVisible(),
					item.getLine(), item.getLineStroke(), item.getLinePaint());
			
			return li;
		}

		/**
		 * Draws a 3D bar to represent one data item.
		 *
		 * @param g2         the graphics device.
		 * @param state      the renderer state.
		 * @param dataArea   the area for plotting the data.
		 * @param plot       the plot.
		 * @param domainAxis the domain axis.
		 * @param rangeAxis  the range axis.
		 * @param dataset    the dataset.
		 * @param row        the row index (zero-based).
		 * @param column     the column index (zero-based).
		 * @param pass       the pass index.
		 */
		public void drawItem(Graphics2D g2,
			CategoryItemRendererState state,
			Rectangle2D dataArea,
			CategoryPlot plot,
			CategoryAxis domainAxis,
			ValueAxis rangeAxis,
			CategoryDataset dataset,
			int row,
			int column,
			int pass)
		{

			// check the value we are plotting...
			Number dataValue = dataset.getValue(row, column);

			if (dataValue == null)
				return;

			double		    value = dataValue.doubleValue();

			Rectangle2D     adjusted = new Rectangle2D.Double(dataArea.getX(),
					dataArea.getY() + getYOffset(), dataArea.getWidth() - getXOffset(),
					dataArea.getHeight() - getYOffset());

			PlotOrientation orientation = plot.getOrientation();

			double		    barW0 = calculateBarW0(plot, orientation, adjusted, domainAxis, state,
					row, column);
			double[]	    barL0L1 = calculateBarL0L1(value);

			if (barL0L1 == null)
				return; // the bar is not visible

			RectangleEdge edge = plot.getRangeAxisEdge();
			double		  transL0 = rangeAxis.valueToJava2D(barL0L1[0], adjusted, edge);
			double		  transL1 = rangeAxis.valueToJava2D(barL0L1[1], adjusted, edge);
			double		  barL0 = Math.min(transL0, transL1);
			double		  barLength = Math.abs(transL1 - transL0);

			// draw the bar...
			Rectangle2D   bar = null;

			if (orientation == PlotOrientation.HORIZONTAL)
				bar = new Rectangle2D.Double(barL0, barW0, barLength, state.getBarWidth());
			else
				bar = new Rectangle2D.Double(barW0, barL0, state.getBarWidth(), barLength);
			Paint					 itemPaint = getItemPaint(row, column);
			GradientPaintTransformer t = getGradientPaintTransformer();

			if ((t != null) && (itemPaint instanceof GradientPaint))
				itemPaint = t.transform((GradientPaint)itemPaint, bar);
			g2.setPaint(itemPaint);
			g2.fill(bar);

			itemPaint = getItemColor(row, column);
			g2.setPaint(itemPaint);

			double	    x0 = bar.getMinX();
			double	    x1 = x0 + getXOffset();
			double	    x2 = bar.getMaxX();
			double	    x3 = x2 + getXOffset();

			double	    y0 = bar.getMinY() - getYOffset();
			double	    y1 = bar.getMinY();
			double	    y2 = bar.getMaxY() - getYOffset();
			double	    y3 = bar.getMaxY();

			GeneralPath bar3dRight = null;
			GeneralPath bar3dTop = null;

			if (barLength > 0.0)
			{
				bar3dRight = new GeneralPath();
				bar3dRight.moveTo((float)x2, (float)y3);
				bar3dRight.lineTo((float)x2, (float)y1);
				bar3dRight.lineTo((float)x3, (float)y0);
				bar3dRight.lineTo((float)x3, (float)y2);
				bar3dRight.closePath();

				if (itemPaint instanceof Color)
					g2.setPaint(((Color)itemPaint).darker());
				g2.fill(bar3dRight);
			}

			bar3dTop = new GeneralPath();
			bar3dTop.moveTo((float)x0, (float)y1);
			bar3dTop.lineTo((float)x1, (float)y0);
			bar3dTop.lineTo((float)x3, (float)y0);
			bar3dTop.lineTo((float)x2, (float)y1);
			bar3dTop.closePath();
			g2.fill(bar3dTop);

			if (isDrawBarOutline()
					&& (state.getBarWidth() > BAR_OUTLINE_WIDTH_THRESHOLD))
			{
				g2.setStroke(getItemOutlineStroke(row, column));
				g2.setPaint(getItemOutlinePaint(row, column));
				g2.draw(bar);
				if (bar3dRight != null)
					g2.draw(bar3dRight);
				if (bar3dTop != null)
					g2.draw(bar3dTop);
			}

			CategoryItemLabelGenerator generator = getItemLabelGenerator(row, column);

			if ((generator != null) && isItemLabelVisible(row, column))
				drawItemLabel(g2, dataset, row, column, plot, generator, bar, (value < 0.0));

			// add an item entity, if this information is being collected
			EntityCollection entities = state.getEntityCollection();

			if (entities != null)
			{
				GeneralPath barOutline = new GeneralPath();

				barOutline.moveTo((float)x0, (float)y3);
				barOutline.lineTo((float)x0, (float)y1);
				barOutline.lineTo((float)x1, (float)y0);
				barOutline.lineTo((float)x3, (float)y0);
				barOutline.lineTo((float)x3, (float)y2);
				barOutline.lineTo((float)x2, (float)y3);
				barOutline.closePath();
				addItemEntity(entities, dataset, row, column, barOutline);
			}
		} // end method drawItem

		/**
		 * Draws an item label. This method is overridden so that the bar can be used to calculate
		 * the label anchor point.
		 *
		 * @param g2        the graphics device.
		 * @param data      the dataset.
		 * @param row       the row.
		 * @param column    the column.
		 * @param plot      the plot.
		 * @param generator the label generator.
		 * @param bar       the bar.
		 * @param negative  a flag indicating a negative value.
		 */
		protected void drawItemLabel(Graphics2D g2,
			CategoryDataset data,
			int row,
			int column,
			CategoryPlot plot,
			CategoryItemLabelGenerator generator,
			Rectangle2D bar,
			boolean negative)
		{

			String label = generator.generateLabel(data, row, column);

			if (label == null)
				return; // nothing to do

			Font	    labelFont = getItemLabelFont(row, column);
			FontMetrics fm = g2.getFontMetrics(labelFont);

			g2.setFont(labelFont);
			g2.setPaint(_tooltipColor);

			int minY = (int)bar.getMinY();
			int minX = (int)bar.getMinX();
			int maxX = (int)bar.getMaxX();
			int x = ((minX + maxX) / 2) + 6;
			int labelLength = fm.stringWidth(label);
			int labelHeight = fm.getHeight();
			int labelMargin = 5;

            if (negative)
            {
            	minY = (int)bar.getMaxY();
                g2.fillRect(x - labelLength / 2 - labelMargin, minY + LINK_HEIGHT / 2,
                    	labelLength + 2 * labelMargin, labelHeight);
                g2.setPaint(Color.black);
                g2.drawRect(x - labelLength / 2 - labelMargin, minY + LINK_HEIGHT / 2,
                	labelLength + 2 * labelMargin, labelHeight);
                g2.drawLine(x, minY, x, minY + LINK_HEIGHT / 2);
                g2.drawString(label, x - labelLength / 2,
                	minY + fm.getAscent() + LINK_HEIGHT / 2);
            }            
            else
            {
				g2.fillRect(x - (labelLength / 2) - labelMargin, minY - labelHeight - LINK_HEIGHT,
					labelLength + (2 * labelMargin), labelHeight);
				g2.setPaint(Color.black);
				g2.drawRect(x - (labelLength / 2) - labelMargin, minY - labelHeight - LINK_HEIGHT,
					labelLength + (2 * labelMargin), labelHeight);
				g2.drawLine(x, minY - 5, x, minY - LINK_HEIGHT);
				g2.drawString(label, x - (labelLength / 2),
					minY + fm.getAscent() - labelHeight - LINK_HEIGHT);
            }
		} // end method drawItemLabel
	} // end class LyBarRenderer3D


	/**
	 * A custom renderer that returns a different color for each item in a single series.
	 */
	static class LyStackedBarRenderer extends StackedBarRenderer
	{

		/**
		 * 
		 */
		private static final long serialVersionUID = -8923168248634436559L;
		/** The colors to be used. */
		private Paint[] paints;
		private Paint[] colors;

		/** Indicates if the alternance of colors is made with columns. */
		private boolean columnChangeColor;

		/**
		 * Creates a new renderer.
		 *
		 * @param colors            the colors
		 * @param columnChangeColor indicates if the alternance of colors is made with columns
		 */
		public LyStackedBarRenderer(Paint[] colors, boolean columnChangeColor)
		{
			setPaints(colors);
			this.columnChangeColor = columnChangeColor;
		}

		/**
		 * Set the colors used to fill columns.
		 *
		 * @param colors the array of colors
		 */
		public void setPaints(Paint[] colors)
		{
			paints = new Paint[colors.length];

			for (int i = 0; i < colors.length; i++)
				paints[i] = new GradientPaint(0f, 0f,
						((Color)colors[i]).brighter().brighter().brighter(), 0f, 0f,
						(Color)colors[i]);
			this.colors = colors;
		}

		/**
		 * Returns the paint for an item. Overrides the default behaviour inherited from
		 * AbstractSeriesRenderer.
		 *
		 * @param  row    the series.
		 * @param  column the category.
		 * @return The item color.
		 */
		public Paint getItemPaint(int row, int column)
		{
			if (columnChangeColor)
				return paints[column % this.paints.length];
			else
				return paints[row % this.paints.length];
		}

		/**
		 * Returns a legend item for a series.
		 * 
		 * @param datasetIndex - the dataset index (zero-based).
		 * @param series - the series index (zero-based). 
		 */
		public LegendItem getLegendItem(int datasetIndex, int series) {
			
			LegendItem	item = super.getLegendItem(datasetIndex, series); 
			Paint		color = colors[series % this.paints.length];
			
			LyLegendItem	li = new LyLegendItem(item.getLabel(), item.getDescription(), item.getToolTipText(),
					item.getURLText(), item.isShapeVisible(), item.getShape(), item.isShapeFilled(), color,
					item.isShapeOutlineVisible(), item.getOutlinePaint(), item.getOutlineStroke(), item.isLineVisible(),
					item.getLine(), item.getLineStroke(), item.getLinePaint());
			
			return li;
		}
	} // end class LyStackedBarRenderer


	/**
	 * Specific legend item implementation used to change legend items colors.
	 */
	public static class LyLegendItem extends LegendItem
	{
		/**
		 * 
		 */
		private static final long serialVersionUID = -8293805138551464900L;

		public LyLegendItem(String label, String description, String toolTipText,
				String urlText, boolean shapeVisible, Shape shape, boolean shapeFilled, Paint fillPaint, boolean
				shapeOutlineVisible, Paint outlinePaint, Stroke outlineStroke, boolean lineVisible, Shape line,
				Stroke lineStroke, Paint linePaint)
		{
			super(label, description, toolTipText,
				urlText, shapeVisible, shape, shapeFilled, fillPaint, shapeOutlineVisible, outlinePaint, outlineStroke, lineVisible, line,
				lineStroke, linePaint);
		}
	}
	
	/**
	 * A custom renderer that returns a different color for each item in a single series.
	 */
	static class LyStackedBarRenderer3D extends StackedBarRenderer3D
	{

		/**
		 * 
		 */
		private static final long serialVersionUID = -7059931371058928761L;

		/** The paints to be used. */
		private Paint[] paints;

		/** The colors to be used. */
		private Paint[] colors;

		/** Indicates if the alternance of colors is made with columns. */
		private boolean columnChangeColor;

		/**
		 * Creates a new renderer.
		 *
		 * @param colors            the colors
		 * @param columnChangeColor indicates if the alternance of colors is made with columns
		 */
		public LyStackedBarRenderer3D(Paint[] colors, boolean columnChangeColor)
		{
			setPaints(colors);
			this.columnChangeColor = columnChangeColor;
		}

		/**
		 * Set the colors used to fill columns.
		 *
		 * @param colors the array of colors
		 */
		public void setPaints(Paint[] colors)
		{
			paints = new Paint[colors.length];

			for (int i = 0; i < colors.length; i++)
				paints[i] = new GradientPaint(0f, 0f,
						((Color)colors[i]).brighter().brighter().brighter(), 0f, 0f,
						(Color)colors[i]);

			this.colors = colors;
		}

		/**
		 * Returns the paint for an item. Overrides the default behaviour inherited from
		 * AbstractSeriesRenderer.
		 *
		 * @param  row    the series.
		 * @param  column the category.
		 * @return The item color.
		 */
		public Paint getItemPaint(int row, int column)
		{
			if (columnChangeColor)
				return paints[column % this.paints.length];
			else
				return paints[row % this.paints.length];
		}

		/**
		 * Returns the paint for an item. Overrides the default behaviour inherited from
		 * AbstractSeriesRenderer.
		 *
		 * @param  row    the series.
		 * @param  column the category.
		 * @return The item color.
		 */
		public Paint getItemColor(int row, int column)
		{
			if (columnChangeColor)
				return colors[column % this.paints.length];
			else
				return colors[row % this.paints.length];
		}

		/**
		 * Returns a legend item for a series.
		 * 
		 * @param datasetIndex - the dataset index (zero-based).
		 * @param series - the series index (zero-based). 
		 */
		public LegendItem getLegendItem(int datasetIndex, int series) {
			
			LegendItem	item = super.getLegendItem(datasetIndex, series); 
			Paint		color = colors[series % this.paints.length];
			
			LyLegendItem	li = new LyLegendItem(item.getLabel(), item.getDescription(), item.getToolTipText(),
					item.getURLText(), item.isShapeVisible(), item.getShape(), item.isShapeFilled(), color,
					item.isShapeOutlineVisible(), item.getOutlinePaint(), item.getOutlineStroke(), item.isLineVisible(),
					item.getLine(), item.getLineStroke(), item.getLinePaint());
			
			return li;
		}

		/**
		 * Draws a 3D bar to represent one data item.
		 *
		 * @param g2         the graphics device.
		 * @param state      the renderer state.
		 * @param dataArea   the area for plotting the data.
		 * @param plot       the plot.
		 * @param domainAxis the domain axis.
		 * @param rangeAxis  the range axis.
		 * @param dataset    the dataset.
		 * @param row        the row index (zero-based).
		 * @param column     the column index (zero-based).
		 * @param pass       the pass index.
		 */
		public void drawItem(Graphics2D g2,
			CategoryItemRendererState state,
			Rectangle2D dataArea,
			CategoryPlot plot,
			CategoryAxis domainAxis,
			ValueAxis rangeAxis,
			CategoryDataset dataset,
			int row,
			int column,
			int pass)
		{

			// check the value we are plotting...
			Number dataValue = dataset.getValue(row, column);

			if (dataValue == null)
				return;

			double		    value = dataValue.doubleValue();

			Rectangle2D     adjusted = new Rectangle2D.Double(dataArea.getX(),
					dataArea.getY() + getYOffset(), dataArea.getWidth() - getXOffset(),
					dataArea.getHeight() - getYOffset());

			PlotOrientation orientation = plot.getOrientation();

			double		    barW0 = domainAxis.getCategoryMiddle(column, getColumnCount(), adjusted,
					plot.getDomainAxisEdge()) - (state.getBarWidth() / 2.0);

			double		    positiveBase = getBase();
			double		    negativeBase = positiveBase;

			for (int i = 0; i < row; i++)
			{
				Number v = dataset.getValue(i, column);

				if (v != null)
				{
					double d = v.doubleValue();

					if (d > 0)
						positiveBase = positiveBase + d;
					else
						negativeBase = negativeBase + d;
				}
			}

			double		  translatedBase;
			double		  translatedValue;
			RectangleEdge location = plot.getRangeAxisEdge();

			if (value > 0.0)
			{
				translatedBase = rangeAxis.valueToJava2D(positiveBase, adjusted, location);
				translatedValue = rangeAxis.valueToJava2D(positiveBase + value, adjusted, location);
			}
			else
			{
				translatedBase = rangeAxis.valueToJava2D(negativeBase, adjusted, location);
				translatedValue = rangeAxis.valueToJava2D(negativeBase + value, adjusted, location);
			}
			double	    barL0 = Math.min(translatedBase, translatedValue);
			double	    barLength = Math.max(Math.abs(translatedValue - translatedBase),
					getMinimumBarLength());

			Rectangle2D bar = null;

			if (orientation == PlotOrientation.HORIZONTAL)
				bar = new Rectangle2D.Double(barL0, barW0, barLength, state.getBarWidth());
			else
				bar = new Rectangle2D.Double(barW0, barL0, state.getBarWidth(), barLength);
			Paint					 itemPaint = getItemPaint(row, column);
			GradientPaintTransformer t = getGradientPaintTransformer();

			if ((t != null) && (itemPaint instanceof GradientPaint))
				itemPaint = t.transform((GradientPaint)itemPaint, bar);
			g2.setPaint(itemPaint);
			g2.fill(bar);

			itemPaint = getItemColor(row, column);
			g2.setPaint(itemPaint);

			if (pass == 0)
			{
				double	    x0 = bar.getMinX();
				double	    x1 = x0 + getXOffset();
				double	    x2 = bar.getMaxX();
				double	    x3 = x2 + getXOffset();

				double	    y0 = bar.getMinY() - getYOffset();
				double	    y1 = bar.getMinY();
				double	    y2 = bar.getMaxY() - getYOffset();
				double	    y3 = bar.getMaxY();

				GeneralPath bar3dRight = null;
				GeneralPath bar3dTop = null;

				if ((value > 0.0) || (orientation == PlotOrientation.VERTICAL))
				{
					bar3dRight = new GeneralPath();
					bar3dRight.moveTo((float)x2, (float)y3);
					bar3dRight.lineTo((float)x2, (float)y1);
					bar3dRight.lineTo((float)x3, (float)y0);
					bar3dRight.lineTo((float)x3, (float)y2);
					bar3dRight.closePath();

					if (itemPaint instanceof Color)
						g2.setPaint(((Color)itemPaint).darker());
					g2.fill(bar3dRight);
				}

				if ((value > 0.0) || (orientation == PlotOrientation.HORIZONTAL))
				{
					bar3dTop = new GeneralPath();
					bar3dTop.moveTo((float)x0, (float)y1);
					bar3dTop.lineTo((float)x1, (float)y0);
					bar3dTop.lineTo((float)x3, (float)y0);
					bar3dTop.lineTo((float)x2, (float)y1);
					bar3dTop.closePath();
					g2.fill(bar3dTop);
				}

				if (isDrawBarOutline() && (state.getBarWidth() > 3))
				{
					g2.setStroke(getItemOutlineStroke(row, column));
					g2.setPaint(getItemOutlinePaint(row, column));
					g2.draw(bar);
					if (bar3dRight != null)
						g2.draw(bar3dRight);
					if (bar3dTop != null)
						g2.draw(bar3dTop);
				}

				// add an item entity, if this information is being collected
				EntityCollection entities = state.getEntityCollection();

				if (entities != null)
					addItemEntity(entities, dataset, row, column, bar);

				CategoryItemLabelGenerator generator = getItemLabelGenerator(row, column);

				if ((generator != null) && isItemLabelVisible(row, column))
					drawItemLabel(g2, dataset, row, column, plot, generator, bar, (value < 0.0));
			}
		} // end method drawItem
	} // end class LyStackedBarRenderer3D


	class LyXYLineAndShapeRenderer extends XYLineAndShapeRenderer
	{
		/**
		 * 
		 */
		private static final long serialVersionUID = 3111295999182137253L;
		private Paint[] colors;
		
		/**
		 * Creates a new renderer.
		 *
		 * @param colors            the colors
		 * @param columnChangeColor indicates if the alternance of colors is made with columns
		 */
		public LyXYLineAndShapeRenderer(Paint[] colors)
		{
			setPaints(colors);
		}

		/**
		 * Set the colors used to fill columns.
		 *
		 * @param colors the array of colors
		 */
		public void setPaints(Paint[] colors)
		{
			this.colors = colors;
		}

		/**
		 * Returns a legend item for a series.
		 * 
		 * @param datasetIndex - the dataset index (zero-based).
		 * @param series - the series index (zero-based). 
		 */
		public LegendItem getLegendItem(int datasetIndex, int series) {
			
			LegendItem	item = super.getLegendItem(datasetIndex, series); 
			Paint		color = colors[series % this.colors.length];
			
			LyLegendItem	li = new LyLegendItem(item.getLabel(), item.getDescription(), item.getToolTipText(),
					item.getURLText(), item.isShapeVisible(), item.getShape(), item.isShapeFilled(), color,
					item.isShapeOutlineVisible(), item.getOutlinePaint(), item.getOutlineStroke(), item.isLineVisible(),
					item.getLine(), item.getLineStroke(), item.getLinePaint());
			
			return li;
		}

		/**
		 * Draws an item label. This method is overridden so that the bar can be used to calculate
		 * the label anchor point.
		 *
		 * @param g2        the graphics device.
		 * @param data      the dataset.
		 * @param row       the row.
		 * @param column    the column.
		 * @param plot      the plot.
		 * @param generator the label generator.
		 * @param bar       the bar.
		 * @param negative  a flag indicating a negative value.
		 */
		protected void drawItemLabel(Graphics2D g2, PlotOrientation orientation,
			XYDataset dataset, int series, int item, double x, double y,
			boolean negative)
		{
			XYItemLabelGenerator generator = getItemLabelGenerator(series, item);

			if (generator != null)
			{
				String label = generator.generateLabel(dataset, series, item);

				if (label == null)
					return; // nothing to do
	            
	            label = LyString.replace(label, SCATTER_PLOT_INTERNAL_LABEL_SEPARATOR + "", SCATTER_PLOT_VISIBLE_LABEL_SEPARATOR);
	            
				Font	    labelFont = getItemLabelFont(series, item);
				FontMetrics fm = g2.getFontMetrics(labelFont);

				g2.setFont(labelFont);
				g2.setPaint(_tooltipColor);

				int yPos = (int)y;
				int xPos = (int)x;
				int labelLength = fm.stringWidth(label);
				int labelHeight = fm.getHeight();
				int labelMargin = 5;

				g2.fillRect(xPos - (labelLength / 2) - labelMargin,
					yPos - labelHeight - LINK_HEIGHT, labelLength + (2 * labelMargin), labelHeight);
				g2.setPaint(Color.black);
				g2.drawRect(xPos - (labelLength / 2) - labelMargin,
					yPos - labelHeight - LINK_HEIGHT, labelLength + (2 * labelMargin), labelHeight);
				g2.drawLine(xPos, yPos - 5, xPos, yPos - LINK_HEIGHT);
				g2.drawString(label, xPos - (labelLength / 2),
					yPos + fm.getAscent() - labelHeight - LINK_HEIGHT);
			}
		} // end method drawItemLabel

	} // end class LyXYLineAndShapeRenderer


	class LyStandardCategoryURLGenerator
		implements CategoryURLGenerator
	{
		
		public String generateURL(CategoryDataset dataset, int series, int category)
		{
			Comparable<?> categoryKey = dataset.getColumnKey(category);
			Comparable<?> rowKey = dataset.getRowKey(series);

			if (_chart.getColumnCount() <= 1)
				return generateUrl(categoryKey.toString());
			else
				return generateUrl(rowKey.toString());
		}
	}


	class LyStandardPieURLGenerator
		implements PieURLGenerator
	{
		public String generateURL(PieDataset dataset, @SuppressWarnings("rawtypes") Comparable key, int pieIndex)
		{
			return generateUrl(key.toString());
		}
	}


	class LyStandardXYURLGenerator
		implements XYURLGenerator
	{
		
		public String generateURL(XYDataset dataset, int series, int item)
		{
			Comparable<?> categoryKey = dataset.getSeriesKey(series);

			return generateUrl(categoryKey.toString());
		}
	}


	class LyStandardXYTooltipGenerator
		implements XYToolTipGenerator
	{
		public String generateToolTip(XYDataset dataset, int series, int item)
		{
			double		 x = dataset.getXValue(series, item);
			double		 y = dataset.getYValue(series, item);
			StringBuffer buildFormat = new StringBuffer();

			buildFormat.append("###,###,###,##0");

			if (SCATTER_PLOT_DECIMALS_NUMBER > 0)
				buildFormat.append('.');

			for (int i = 0; i < SCATTER_PLOT_DECIMALS_NUMBER; i++)
				buildFormat.append('0');

	        if (_chart.getXType() == LyChart.Type.CHARTDATE)
	        {
	        	DateFormat		format;
	        	DateFormat		dateFormat = _chart.getEnvironment().dateFormatFactory(LyEnvironment.TYPE_DATE);
	        	DateFormat		dateTimeFormat = _chart.getEnvironment().dateFormatFactory(LyEnvironment.TYPE_DATE_TIME);
	        	NumberFormat	numberFormat = _chart.getEnvironment().getNumberFormat(buildFormat.toString());
	        	Calendar		calendar = Calendar.getInstance();
	        	
	        	calendar.setTimeInMillis((long)x);
	        	
	        	if ((calendar.get(Calendar.HOUR) == 0)
	        			&& (calendar.get(Calendar.MINUTE) == 0)
	        			&& (calendar.get(Calendar.SECOND) == 0))
	        		format = dateFormat;
	        	else
	        		format = dateTimeFormat;
	        		
	        	return format.format(new Date((long)x)) + SCATTER_PLOT_TOOLTIP_SEPARATOR + numberFormat.format(y);
	        }
	        else
	        {
				NumberFormat numberFormat = _chart.getEnvironment().getNumberFormat(buildFormat
							.toString());
	
				return numberFormat.format(x) + SCATTER_PLOT_TOOLTIP_SEPARATOR + numberFormat.format(y);
	        }
		}
	}

	class LyPieSectionLabelGenerator extends StandardPieSectionLabelGenerator
	{
		/**
		 * 
		 */
		private static final long serialVersionUID = -1268949087419384839L;

		public String generateSectionLabel(PieDataset dataset, @SuppressWarnings("rawtypes") Comparable key)
		{
			String result = super.generateSectionLabel(dataset, key) + "%";

			return result;
		}
	}

	class LyCategoryItemLabelGenerator extends StandardCategoryItemLabelGenerator
	{
		/**
		 * 
		 */
		private static final long serialVersionUID = -1343293479955645785L;

		protected String generateLabelString(CategoryDataset dataset, int row, int column)
		{
			String result = super.generateLabelString(dataset, row, column) + "%";

			return result;
		}
	}


	class LyScatterPlotData
	{
		java.util.Vector<Double[]> _values;
		String _label;

		LyScatterPlotData(String label)
		{
			_values = new java.util.Vector<Double[]>();
			_label = label;
		}

		public void addValue(Double x, Double y)
		{
			_values.addElement(new Double[] { x, y });
		}

		public void setLabel(String label)
		{
			_label = label;
		}

		public String getLabel()
		{
			return _label;
		}

		public Double[][] getValues()
		{
			int		   i, n = _values.size();
			Double[][] values = new Double[n][2];
			Double[]   value;

			for (i = 0; i < n; i++)
			{
				value = _values.elementAt(i);
				values[i][0] = value[0];
				values[i][1] = value[1];
			}

			return values;
		}

		public boolean hasValue(Double x, Double y)
		{
			int		 i, n = _values.size();
			Double[] value;

			for (i = 0; i < n; i++)
			{
				value = _values.elementAt(i);
				if (value[0].equals(x) && value[1].equals(y))
					return true;
			}

			return false;
		}
	} // end class LyScatterPlotData
} // end class LyCommonFreeChartConverter
