/*
 * LyImageUtils.java Copyright (c) 2000-2009, Lyria-W4. All rights reserved.
 */

package leon.view.common;

import com.sun.image.codec.jpeg.*;

import leon.misc.*;
import leon.view.win.swing.*;

import java.awt.*;
import java.awt.image.*;
import java.io.*;
import java.util.*;

import javax.imageio.*;


/**
 * This class gives utility methods to handle images.
 *
 * @version  $Id: LyImageUtils.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author   Lyria
 */
public class LyImageUtils
{
	public static final String __VERSION = "$Id: LyImageUtils.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Number of loaded images. */
	protected static int nLoaded;

	/**
	 * Default Frame used in some parts of the code when a default AWT component is needed (for
	 * example to create an image).
	 */
	public static Frame _frame;

	/**
	 * Returns a default frame.
	 *
	 * @return  The default frame.
	 */
	static Frame getDefaultFrame()
	{
		if (_frame == null)
		{
			_frame = new Frame();
			// ensure that the peer is ready.
			// a peer is required for some methods such as createImage()
			// see also : LyEnvironment.getImage()
			if (LyEnvironment.isOperatingSystem(LyEnvironment.OS_WINDOWS))
				_frame.addNotify();
		}

		return _frame;
	}

	/**
	 * Clean all statics used by this class.
	 */
	public static void clean()
	{
		if (_frame != null)
			_frame.dispose();

		_frame = null;
	}

	/**
	 * Encodes an image to JPEG format.
	 *
	 * @param   image      the image to encode
	 * @param   imageDir   temporary directory
	 * @param   imageFile  the JPEG File
	 *
	 * @return  true if the file was generated, false otherwise.
	 */
	public static boolean encodeImage(Image image, String imageDir, String imageFile)
	{
		OutputStream oStream = encodeImageToStream(image, imageDir, imageFile);

		if (oStream != null)
		{
			try
			{
				oStream.flush();
				oStream.close();
			}
			catch (IOException e)
			{
			}

			return true;
		}
		else
			return false;
	} // end method encodeImage

	/**
	 * Encodes an image to JPEG format.
	 *
	 * @param   image      the image to encode
	 * @param   imageDir   temporary directory
	 * @param   imageFile  the JPEG File
	 *
	 * @return  true if the file was generated, false otherwise.
	 */
	public static OutputStream encodeImageToStream(Image image, String imageDir, String imageFile)
	{
		if (image == null)
			return null;

		try
		{
			if (imageDir != null)
			{
				File imageRep = new File(imageDir);

				imageRep.mkdir();
			}

			OutputStream oStream = null;
			File		 dataFile = new File(imageFile);

			oStream = new FileOutputStream(dataFile);

			BufferedImage buffImage = new BufferedImage(image.getWidth(null), image.getHeight(null),
					BufferedImage.TYPE_INT_BGR);
			Graphics	  g = buffImage.getGraphics();

			g.drawImage(image, 0, 0, null);

			JPEGEncodeParam encodeParam = JPEGCodec.getDefaultJPEGEncodeParam(buffImage);

			// set the quality param (1.0 is highest, 0.0 is lowest)
			// I think that 0.9 is a good choice between quality and size
			encodeParam.setQuality((float)0.90, true);
			JPEGImageEncoder encoder = JPEGCodec.createJPEGEncoder(oStream, encodeParam);

			encoder.encode(buffImage);

			oStream.flush();
			oStream.close();

			return oStream;
		}
		catch (Exception e)
		{
			e.printStackTrace();
			return null;
		}
	} // end method encodeImage

	/**
	 * Constructs a default image.
	 *
	 * @return  the default image
	 */
	public static Image getDefaultImage()
	{
		Image    image = LySwingViewManager.getDefaultFrame().createImage(20, 20);
		Graphics g = image.getGraphics();

		g.setColor(Color.white);
		g.fillRect(0, 0, 20, 20);
		g.setColor(Color.black);
		g.drawRect(0, 0, 19, 19);
		g.drawLine(0, 0, 19, 19);
		g.drawLine(0, 19, 19, 0);

		return image;
	}

	/**
	 * Gets images sizes.
	 *
	 * @param   image
	 *
	 * @return  Image size (table with 2 entries).
	 */
	public static int[] getSizes(Image image)
	{
		if (image == null)
			return null;

		return new int[] { image.getWidth(null), image.getHeight(null) };
	}

	/**
	 * Creates a scaled version of the given image.
	 *
	 * @param   image   The image to scale.
	 * @param   width   The width to which to scale the image.
	 * @param   height  The height to which to scale the image.
	 *
	 * @return  A scaled version of the image.
	 */
	public static Object getScaledInstance(Image image, int width, int height)
	{
		// return image.getScaledInstance(width, height, Image.SCALE_SMOOTH);
		// Due to Java Bug#4937376, we can not use the above line of code.
		// The following code do the same thing...
		// We could also have replace SMOOTH by REPLICATE, but the resulting image is not 'SMOOTH'

		BufferedImage bi = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB); // USE ARGB to keep alpha chanel
		Graphics2D    g = (Graphics2D)bi.getGraphics();

		g.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
		g.drawImage(image, 0, 0, width, height, null);
		return bi;

	}

	/**
	 * Waits for the image loaging.
	 *
	 * @param   component
	 * @param   image      Image to load.
	 *
	 * @throws  InterruptedException
	 */
	public static void waitForLoading(Component component, Image image) throws InterruptedException
	{
		if (component == null)
			component = getDefaultFrame();

		MediaTracker mt = new MediaTracker(component);

		mt.addImage(image, nLoaded);
		mt.waitForID(nLoaded);

		nLoaded++;
	} // end method waitForLoading

	/**
	 * retreives an image from a Stream.
	 *
	 * @param   component  The client of the image creation.
	 * @param   resource   stream inputStream to the image data
	 *
	 * @return  the image or null if an error occurs
	 *
	 * @throws  IOException  if something is wrong during loading
	 */
	public static Object loadImageFromStream(Component component, InputStream resource) throws IOException
	{
		if (resource == null)
			return null;

		InputStream resourceToRead = resource;
		byte[]	    content = null;

		if (!resource.markSupported())
		{
			// Create an other inputStream...
			ByteArrayOutputStream bStream = new ByteArrayOutputStream(resource.available());
			int					  n = 0;

			while (true)
			{
				n = resource.read();
				if (n >= 0)
					bStream.write((byte)n);
				else
					break;
			}

			content = bStream.toByteArray();
			resourceToRead = new ByteArrayInputStream(content);

			bStream.flush();
			bStream.close();

		}

		// Add a mark to the resource to allow a rewind if the first method for loading the image failed.
		resourceToRead.mark(-1);

		Image result = null;

		try
		{
			result = ImageIO.read(resourceToRead);
			resourceToRead.close();
		}
		catch (Exception e)
		{
			// If any exception occured, try to load image using toolkit...

			// Returns to the beginning of the inputstream.
			resourceToRead.reset();

			Toolkit tk = Toolkit.getDefaultToolkit();

			if (content == null)
			{
				ByteArrayOutputStream bStream = new ByteArrayOutputStream(resourceToRead.available());
				int					  n = 0;

				while (true)
				{
					n = resourceToRead.read();
					if (n >= 0)
						bStream.write((byte)n);
					else
						break;
				}

				content = bStream.toByteArray();
				bStream.flush();
				bStream.close();
			}

			resourceToRead.close();

			result = tk.createImage(content);

		}

		return result;
	} // end method loadImageFromStream

	/**
	 * Free the resources of the image if not in the cache. If the image is in the cache, this
	 * method does nothing.
	 *
	 * @param  image  The image that must be freed.
	 */
	public static void freeImage(Image image)
	{
		image.flush();
	}

	/**
	 * Compare two images pixel by pixel.
	 *
	 * @param   image1  first image to compare
	 * @param   image2  second image to compare
	 *
	 * @return  true if images are the same.
	 *
	 * @throws  InterruptedException  exception if problems on grabber or loading.
	 */
	public static boolean isSameImage(Image image1, Image image2) throws InterruptedException
	{
		if (image1 == image2)
			return true;

		if ((image1 == null) || (image2 == null))
			return false;

		// Size must be the same.
		int w = image1.getWidth(null);
		int h = image1.getHeight(null);

		if ((w == -1) || (h == -1))
		{
			waitForLoading(null, image1);
			w = image1.getWidth(null);
			h = image1.getHeight(null);
		}

		int w1 = image2.getWidth(null);
		int h1 = image2.getHeight(null);

		if ((w1 == -1) || (h1 == -1))
		{
			waitForLoading(null, image2);
			w1 = image2.getWidth(null);
			h1 = image2.getHeight(null);
		}

		if ((w == -1) || (h == -1) || (w1 == -1) || (h1 == -1))
			return false;

		if ((w != w1) || (h != h1))
			return false;

		// compare image pixel by pixel with array of lines (because of memory)

		int[]		 pixels1 = new int[w];
		int[]		 pixels2 = new int[w];
		PixelGrabber pg;

		for (int i = 0; i < h; i++)
		{
			pg = new PixelGrabber(image1, 0, i, w, 1, pixels1, 0, w);
			pg.grabPixels();
			pg = new PixelGrabber(image2, 0, i, w, 1, pixels2, 0, w);
			pg.grabPixels();

			if (!Arrays.equals(pixels1, pixels2))
				return false;
		}

		return true;
	} // end method isSameImage

} // end class LyImageUtils
