/*
 * LyActionDemand.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.event;

import java.util.*;


/**
 * This class implements an action event created from the presentation layer, which is sent to the
 * associated controller and sent to its controller using RMI.
 *
 * @version $Id: LyActionDemand.java 28997 2010-12-23 10:17:15Z bmichel $
 * @author  Lyria
 * @see     LyDemand
 * @see     LySelectionDemand
 * @see     LyHelpDemand
 * @see     LyMenuDemand
 */
public class LyActionDemand extends LyDemand
{
	/**
	 * 
	 */
	private static final long serialVersionUID = -4703698467912925400L;

	public static final String __VERSION =
		"$Id: LyActionDemand.java 28997 2010-12-23 10:17:15Z bmichel $";

	/** This event indicates that the default action occured for instance button pressed. */
	public static final short  TYPE_ACTION = 1;

	/** The value in a component has changed. */
	public static final short  TYPE_VALUE_CHANGED = 2;

	/** This event indicates that a window is closing and indicates a return code. */
	public static final short  TYPE_WINDOW_CLOSED = 3;

	/** This event indicates that user has requested a new sort in the view. */
	public static final short  TYPE_SORT = 4;

	/** This event indicates that user has requested a new sort in the view. */
	public static final short  TYPE_EXTEND_SORT = 5;

	/** This event indicates that user has requested an hypertext in the view. */
	public static final short  TYPE_HREF = 6;

	/** This event indicates that a node has been expanded. */
	public static final short  TYPE_EXPAND = 7;

	/** This event indicates that some nodes have been moved (map...). */
	public static final short  TYPE_MOVE = 8;

	/** This event indicates that user has requested a deletion in the view. */
	public static final short  TYPE_DELETE = 9;

	/** This event indicates that user has selected status image. */
	public static final short  TYPE_STATUS = 10;

	/** This event indicates that user has requested a insert in the view. */
	public static final short  TYPE_INSERT = 11;

	/** The unit in a component has changed. */
	public static final short  TYPE_UNIT_CHANGED = 12;

	/** The type in a component has changed. */
	public static final short  TYPE_TYPE_CHANGED = 13;

	/** This event indicates that a formatted string has to be completed. */
	public static final short  TYPE_COMPLETE_VALUE = 14;

	/** This event indicates a divide interval request. */
	public static final short  TYPE_DIVIDE = 15;

	/** This event indicates a tab next request. */
	public static final short  TYPE_TAB_NEXT = 16;

	/** This event indicates a tab previous request. */
	public static final short  TYPE_TAB_PREVIOUS = 17;

	/** This event indicates a close tab request. */
	public static final short  TYPE_CLOSE_TAB = 18;
	
	/* Internal Types */

	/**
	 * This event indicates that a timeout has expired (Ex : waiting for component initialization
	 * with periodic notifications). Reserved for internal used in view package.
	 */
	public static final short  TYPE_TIMEOUT = 50;

	/** The mouse is over the item. Reserved for internal used in view package. */
	public static final short  TYPE_MOUSE_OVER = 51;

	/** This event indicates a open line request. Reserved for internal used in view package. */
	public static final short  TYPE_OPEN = 52;

	/** Optional value associated to this action event. */
	protected String		   _value;

	/** Optional additional value associated to this action event. */
	protected String		   _additionalValue;

	/**
	 * Optional selection context associated to an action event NOTE : this is an indication from
	 * the view, wich may not be set neither complete depending on view capacity.
	 */
	protected Vector<String>   _viewSelection;

	/**
	 * Constructor.
	 */
	public LyActionDemand(int demandType, String sourceId)
	{
		super(demandType, sourceId);

		if (demandType > TYPE_OPEN)
			throw new IllegalStateException("BF AppEngine : invalid demand type " + demandType);

		_viewSelection = null;
		_value = null;
	}

	/**
	 * Accessor to the value associated to this action. The meaning of the value depends on the type
	 * of the event.
	 *
	 * @return the value
	 */
	public String getValue()
	{
		return (_value);
	}

	/**
	 * Accessor to the value associated to this action. The meaning of the value depends on the type
	 * of the event.
	 *
	 * @param value the value
	 */
	public void setValue(String value)
	{
		_value = value;
	}

	/**
	 * Accessor to the additional value associated to this action. The meaning of the value depends
	 * on the type of the event.
	 *
	 * @return the value
	 */
	public String getAdditionalValue()
	{
		return (_additionalValue);
	}

	/**
	 * Accessor to the additional value associated to this action. The meaning of the value depends
	 * on the type of the event.
	 *
	 * @param value the value
	 */
	public void setAdditionalValue(String value)
	{
		_additionalValue = value;
	}

	/**
	 * Accessor to a possible selection context of invocation in an action. This context may be
	 * null, unknown from the view or even partial. Decision of context is provided by the
	 * controller, this accessor is only indicative.
	 *
	 * @return the selection ids or null.
	 */
	public Vector<String> getViewSelection()
	{
		return _viewSelection;
	}

	/**
	 * Accessor to a possible selection context of invocation in an action. This context may be
	 * null, unknown from the view or even partial. Decision of context is provided by the
	 * controller, this accessor is only indicative.
	 *
	 * @param value the value
	 */
	public void setViewSelection(Vector<String> selection)
	{
		_viewSelection = selection;
	}

	/**
	 * Export.
	 */
	public String toString()
	{
		StringBuffer sb = new StringBuffer();

		sb.append("** Action Demand **\n");
		sb.append(super.toString());
		sb.append("- value   : ").append(_value).append('\n');

		return (sb.toString());
	}

	/**
	 * Convert the demand type to a string (print and dump).
	 */
	protected String getDemandType(int typ)
	{
		switch (typ)
		{
		case TYPE_ACTION:
			return "TYPE_ACTION";

		case TYPE_VALUE_CHANGED:
			return "TYPE_VALUE_CHANGED";

		case TYPE_WINDOW_CLOSED:
			return "TYPE_WINDOW_CLOSED";

		case TYPE_SORT:
			return "TYPE_SORT";

		case TYPE_EXTEND_SORT:
			return "TYPE_EXTEND_SORT";

		case TYPE_HREF:
			return "TYPE_HREF";

		case TYPE_EXPAND:
			return "TYPE_EXPAND";

		case TYPE_MOVE:
			return "TYPE_MOVE";

		case TYPE_DELETE:
			return "TYPE_DELETE";

		case TYPE_STATUS:
			return "TYPE_STATUS";

		case TYPE_INSERT:
			return "TYPE_INSERT";

		case TYPE_UNIT_CHANGED:
			return "TYPE_UNIT_CHANGED";
			
		case TYPE_TYPE_CHANGED:
			return "TYPE_TYPE_CHANGED";

		case TYPE_COMPLETE_VALUE:
			return "TYPE_COMPLETE_VALUE";

		case TYPE_DIVIDE:
			return "TYPE_DIVIDE";

		case TYPE_TAB_NEXT:
			return "TYPE_TAB_NEXT";

		case TYPE_TAB_PREVIOUS:
			return "TYPE_TAB_PREVIOUS";

		case TYPE_CLOSE_TAB:
			return "TYPE_CLOSE_TAB";

		case TYPE_TIMEOUT:
			return "TYPE_TIMEOUT";

		default:
			return super.getDemandType(typ);
		}
	} // end method getDemandType
	
} // end class LyActionDemand
