/*
 * LyDemand.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.event;

import java.io.*;
import java.util.HashMap;

import leon.misc.*;


/**
 * This class implements an event from the presentation layer and sent to its controller using RMI
 * This class is specialised into several event types.
 *
 * @version $Id: LyDemand.java 28997 2010-12-23 10:17:15Z bmichel $
 * @author  Lyria
 * @see     LySelectionDemand
 * @see     LyActionDemand
 * @see     LyHelpDemand
 * @see     LyMenuDemand
 */
public class LyDemand
	implements Serializable
{

	/**
	 * 
	 */
	private static final long serialVersionUID = -7247728470611170603L;

	public static final String  __VERSION =
		"$Id: LyDemand.java 28997 2010-12-23 10:17:15Z bmichel $";

	public static char		    SEPARATOR = '|';

	/** Identifier of the demand. */
	protected int			    _id;

	/** Demand type. */
	protected int			    _demandType;

	/** Source : identifier of the object which originally caused the demand. */
	protected String		    _sourceId;

	// Reference to the selection index
	protected int[]			    _index_list;

	/** Sub-index of the demand. */
	protected int			    _sub_index;

	/** Parameter values. */
	protected LyParameterValues _parameterValues;

	/** Associations between keys and values used to valuate contextual values. */
	protected HashMap<String, Object>		    _attributes;

	/**
	 * Constructor.
	 *
	 * @param id or -1 for automatic generation.
	 */
	public LyDemand(int demandType, String sourceId)
	{
		_id = LyTools.getNextIdentifier();
		_demandType = demandType;
		_sourceId = sourceId;
		_index_list = null;
		_parameterValues = null;
		_sub_index = -1;
	}

	/**
	 * Accessor to the identifier of the demand.
	 *
	 * @return the demand id
	 */
	public int getId()
	{
		return (_id);
	}

	/**
	 * Accessor to the type of the demand Types are described in specific implementations :
	 * selectDemand, actionDemand ...
	 *
	 * @return the demand type
	 */
	public int getType()
	{
		return (_demandType);
	}

	/**
	 * Accessor to the identifier of the source of this demand.
	 *
	 * @return the source id
	 */
	public String getSource()
	{
		return (_sourceId);
	}

	/**
	 * Accessor to the index the (un-)selected object.
	 *
	 * @param  pos index of the selection
	 * @return the index, or -1 if not found
	 */
	public int getIndex(int pos)
	{
		if (_index_list != null)
		{
			if ((pos >= 0) && (pos < _index_list.length))
				return (_index_list[pos]);
		}
		return -1;
	}

	/**
	 * Add a new index for a (un-)selected object.
	 *
	 * @param pos index of the selection
	 */
	public void addIndex(int idx)
	{
		int pos = 0;

		// check size
		if (_index_list != null)
		{
			boolean spaceLeft = false;

			for (pos = 0; (pos < _index_list.length) && (!spaceLeft); pos++)
			{
				if (_index_list[pos] == -1)
				{
					spaceLeft = true;
					break;
				}
			}

			if (!spaceLeft)
			{
				// increase size
				int   newSize = (_index_list.length * 2) + 5;
				int[] tmpArray = new int[newSize];

				// copy current
				for (int i = 0; i < _index_list.length; i++)
					tmpArray[i] = _index_list[i];
				//prepare spaceLeft
				for (int i = _index_list.length; i < tmpArray.length; i++)
					tmpArray[i] = -1;
				// done
				_index_list = tmpArray;
			}
		}
		else
			_index_list = new int[1];

		_index_list[pos] = idx;
	} // end method addIndex

	/**
	 * Accessor to info description associated to this demand The ParameterValues contains pairs id
	 * / values for the current view.
	 *
	 * @return the demand info if any, null otherwise
	 */
	public LyParameterValues getParameterValues()
	{
		return _parameterValues;
	}

	/**
	 * Accessor to info description associated to this demand The ParameterValues contains pairs id
	 * / values for the current view.
	 *
	 * @return the demand info if any, null otherwise
	 */
	public void setParameterValues(LyParameterValues parameterValues)
	{
		_parameterValues = parameterValues;
	}

	/**
	 * Sub-index associated to this event (optional).
	 */
	public int getSubIndex()
	{
		return _sub_index;
	}

	/**
	 * Value associated to this event (optional).
	 */
	public void setSubIndex(int sub_index)
	{
		_sub_index = sub_index;
	}

	/**
	 * Export.
	 */
	public String toString()
	{
		StringBuffer sb = new StringBuffer();

		sb.append("- class   : ").append(getClass()).append('\n');
		sb.append("- id (demand) : ").append(_id).append('\n');
		sb.append("- type    : ").append(getDemandType(_demandType)).append('\n');
		sb.append("- id (source) : ").append(_sourceId).append('\n');
		sb.append("- index  : ");

		int n = (_index_list == null) ? 0 : _index_list.length;

		for (int i = 0; i < n; i++)
		{
			if (getIndex(i) < 0)
				break;
			if (i > 0)
				sb.append(" , ");
			sb.append(getIndex(i));
		}
		sb.append('\n');

		return (sb.toString());
	}

	/**
	 * Export.
	 */
	public String toShortString()
	{
		StringBuffer sb = new StringBuffer();

		sb.append("demand #").append(_id);
		sb.append(", ").append(getDemandType(_demandType));
		sb.append(", ").append(_sourceId);

		return sb.toString();
	}

	/**
	 * Convert the demand type to a string.
	 */
	public String getDemandType()
	{
		return getDemandType(_demandType);
	}

	/**
	 * Convert the demand type to a string (print and dump).
	 */
	protected String getDemandType(int typ)
	{
		return Integer.toString(typ);
	}
	
	/**
	 * Adds a pair key-value to the session. May be used for example to store the login used to
	 * create the session.
	 *
	 * @param key   The key that must be stored.
	 * @param value The value that must be stored with the key.
	 * @see   #getValue
	 */
	public void setAttribute(String key, Object value)
	{
		if (_attributes == null)
			_attributes = new HashMap<String, Object>();

		_attributes.put(key, value);
	}

	/**
	 * Retrieves a value from a registered key.
	 *
	 * @param  key The key that must be used.
	 * @return The value stored with the key or null if the key is not registered.
	 * @see    #addValue
	 */
	public Object getAttribute(String key)
	{
		if ((_attributes == null) || (key == null))
			return null;

		return _attributes.get(key);
	}

} // end class LyDemand
