/*
 * LySelectDemand.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.event;

import java.util.*;
import leon.misc.*;


/**
 * This class implements a selection event created from the presentation layer, which is sent to the
 * associated controller and sent to its controller using RMI.
 *
 * @version $Id: LySelectionDemand.java 34656 2012-03-14 16:20:01Z nhamon $
 * @author  Lyria
 * @see     LyDemand
 * @see     LyActionDemand
 * @see     LyHelpDemand
 * @see     LyMenuDemand
 */
public class LySelectionDemand extends LyDemand
{
	/**
	 * 
	 */
	private static final long serialVersionUID = 8692775297829676492L;

	public static final String __VERSION =
		"$Id: LySelectionDemand.java 34656 2012-03-14 16:20:01Z nhamon $";

	/**
	 * This event indicates the selection of one one several items Items id are indicated in the
	 * getSelectionId() method.
	 */
	public static final short  TYPE_SELECT = 101;

	/**
	 * This event indicates the unselection of one one several items Items id are indicated in the
	 * getSelectionId() method.
	 */
	public static final short  TYPE_UNSELECT = 102;

	/** This event indicates that all itemls are selected. */
	public static final short  TYPE_SELECT_ALL = 103;

	/** This event indicates that all items are unselected. */
	public static final short  TYPE_UNSELECT_ALL = 104;

	// Reference to the selection id : may be null, String or Vector of String
	protected Object		   _selectionId;

	// Reference to the selection index
	protected int			   _selectionSize;

	// Indicates wether this notification is and extension of (de)selection, or
	// if it contains all selected objects.
	protected boolean		   _extend;

	/**
	 * Constructor.
	 *
	 * @param id or -1 fgor automatic generation.
	 */
	public LySelectionDemand(int demandType, String sourceId)
	{
		super(demandType, sourceId);

		_selectionSize = -1;
		_extend = false;
	}

	/**
	 * Accessor to the identifier of the (un-)selected object. if multiple objects are selecte or
	 * unselected, all ids are returned with a separator.
	 */
	public String getSelectionId()
	{
		if (_selectionId instanceof String)
			return ((String)_selectionId);
		else if (_selectionId instanceof Vector<?>)
		{
			// Aggregate id into a String and return it
			String[] result = new String[getSelectionSize()];

			((Vector<?>)_selectionId).copyInto(result);

			return (LyString.listToString(result, SEPARATOR));
		}
		else
			return (null);
	}

	/**
	 * Accessor to the identifier of the (un-)selected object.
	 *
	 * @param  idx index of the selection
	 * @return the id, or null if no id specified for the idx
	 */
	public String getSelectionId(int idx)
	{
		if ((idx < 0) || (idx >= getSelectionSize()))
			return null; // no such idx

		if (_selectionId instanceof String)
		{
			if (idx == 0)
				return ((String)_selectionId);
		}
		else if (_selectionId instanceof Vector<?>)
		{
			Vector<?> v = (Vector<?>)_selectionId;
			//	FFT00463 with edit table, the last line may have no id (yet) in creation
			//  may have no id for last selection
			if (v.size()> idx)  
				return ((String)v.elementAt(idx));
		}
		return (null);
	}
	
	/**
	 * Indicates whether demand contains given value in its list of selection ids.
	 */
	public boolean hasSelectionId(String value)
	{
		if (value == null)
			return false;
		
		if (_selectionId instanceof String)
			return value.equals(_selectionId);
		else if (_selectionId instanceof Vector<?>)
			return ((Vector<?>)_selectionId).contains(value);
		
		return false;
	}

	/**
	 * Accessor to the number of selections.
	 */
	public void setSelectionSize(int size)
	{
		_selectionSize = size;
		if (size == 0)
			_selectionId = null;
	}

	/**
	 * Returns the number of ids in the list.<br />
	 * This method checks the real size of selection if _selectionSize attribute has
	 * not been set.
	 *
	 * @return the number of ids in the list
	 */
	public int getSelectionSize()
	{
		if (_selectionSize >= 0)
			return _selectionSize;
		else if (_selectionId instanceof String)
			return (1);
		else if (_selectionId instanceof Vector<?>)
			return ((Vector<?>)_selectionId).size();
		else if (_index_list != null)
		{
			for (int pos = 0; pos < _index_list.length; pos++)
			{
				if (_index_list[pos] == -1)
					return pos;
			}
			return _index_list.length;
		}
		else
			return (0);
	}
	
	/**
	 * Returns the value of the _selectionSize attribute.<br />
	 * This method does not check the real size of selection.
	 * 
	 * @return the value of the _selectionSize attribute if positive or null, otherwise return -1.
	 */
	public int getSelectionSizeParam()
	{
		if (_selectionSize >= 0)
			return _selectionSize;
		else
			return -1;
	}

	/**
	 * Adds a new id to the selection.
	 *
	 * @param newId new object identifier
	 */
	public void addSelectionId(String newId)
	{
		if (newId != null)
		{
			if (_selectionId == null)
				_selectionId = newId;
			else if (_selectionId instanceof Vector<?>)
			{
				@SuppressWarnings("unchecked")
				Vector<Object> temp = (Vector<Object>)_selectionId;
				temp.addElement(newId);
			}
			else if (_selectionId instanceof String)
			{
				Vector<String> v = new Vector<String>();

				v.addElement((String)_selectionId);
				v.addElement(newId);

				_selectionId = v;
			}
		}
	}

	/**
	 * Export.
	 */
	public String toString()
	{
		StringBuffer sb = new StringBuffer();

		sb.append(super.toString());
		sb.append("- selectionSize  : ").append(getSelectionSize()).append('\n');
		sb.append("- selection  : ");

		int n = getSelectionSize();

		for (int i = 0; i < n; i++)
		{
			if (getSelectionId(i) == null)
				break;
			if (i > 0)
				sb.append(" , ");
			sb.append(getSelectionId(i));
		}

		sb.append('\n');

		return (sb.toString());
	}

	/**
	 * Convert the demand type to a string (print and dump).
	 */
	protected String getDemandType(int typ)
	{
		switch (typ)
		{
		case TYPE_SELECT:
			return "TYPE_SELECT";

		case TYPE_UNSELECT:
			return "TYPE_UNSELECT";

		case TYPE_SELECT_ALL:
			return "TYPE_SELECT_ALL";

		case TYPE_UNSELECT_ALL:
			return "TYPE_UNSELECT_ALL";

		default:
			return super.getDemandType(typ);
		}
	}

	/**
	 * Accessor to the new extension status.
	 *
	 * @return true if the selection extends previous notification
	 */
	public boolean isExtend()
	{
		return _extend;
	}

	/**
	 * Accessor to the new extension status. Default is false.
	 *
	 * @param extend true if the selection extends previous notification.
	 */
	public void setExtend(boolean extend)
	{
		_extend = extend;
	}
} // end class LySelectionDemand
