/*
 * LySimpleDemandTransmitter.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.event;

import java.util.*;
import leon.misc.*;


/**
 * Client may register on components which implement this interface to receive notifications from
 * the component.
 *
 * @version $Id: LySimpleDemandTransmitter.java 29006 2010-12-23 17:44:46Z bmichel $
 * @author  Lyria
 */
public class LySimpleDemandTransmitter
	implements LyDemandTransmitter,
		LyMessageWindowRegister
{
	public static final String __VERSION =
		"$Id: LySimpleDemandTransmitter.java 29006 2010-12-23 17:44:46Z bmichel $";

	// List of listeners for demands without Identification
	protected Vector<LyDemandListener>		       _demandListeners;

	// Cache to optimize search for identified objects
	protected Hashtable<String, LyDemandListener>  _controllers;

	// Environment of this demand transmitter
	protected LyEnvironment    _environment;

	// Demand player recorded on this transmitter
	private LyDemandPlayer     _player;

	/**
	 * Constructor.
	 */
	public LySimpleDemandTransmitter(LyEnvironment environment)
	{
		_environment = environment;
	}

	/**
	 * Free the demand listener.
	 */
	public void free()
	{
		if (_player != null)
		{
			_player.free();
			_player = null;
		}
	}

	/**
	 * Adds the specified selection listener to receive demand events from this component.
	 * (LyDemandTransmitter interface)
	 *
	 * @param listener client for the action event
	 */
	public synchronized void addDemandListener(LyDemandListener listener)
	{
		// add it to the cache item if possible
		if ((listener.getId() != null)
				&& (listener.getId().length() > 0))
		{
			if (_controllers == null)
				_controllers = new Hashtable<String, LyDemandListener>();
			_controllers.put(listener.getId(), listener);
		}

		// use full list
		if (_demandListeners == null)
			_demandListeners = new Vector<LyDemandListener>();
	}

	/**
	 * Removes the specified demand listener from this component. (LyDemandTransmitter interface)
	 *
	 * @param listener client for the action event
	 */
	public synchronized void removeDemandListener(LyDemandListener listener)
	{
		// remove from cache
		if ((_controllers != null)
				&& (listener.getId() != null)
				&& (listener.getId().length() > 0))
		{
			LyDemandListener registeredListener = _controllers.get(listener.getId());
			if (listener == registeredListener )
				_controllers.remove(listener.getId());
		}

		// remove from list
		if ((_demandListeners != null)
				&& (listener != null))
			_demandListeners.removeElement(listener);
	}

	/**
	 * Processes demands on this component.
	 */
	public void processDemand(LyDemand demand)
	{
		processDemand(null, demand);
	}

	/**
	 * Processes demands on this component.
	 *
	 * @param listenerId the demand is sent to the specified listener (LyDemandTransmitter
	 *                   interface)
	 */
	public void processDemand(String listenerId, LyDemand demand)
	{
		LyDemandPlayer player = getPlayer();

		if (player != null)
		{
			// Send demand to player and return if it was processed
			if (player.processDemand(this, listenerId, demand))
				return;
		}

		process(listenerId, demand);
	}

	/**
	 * Processes demands on this component.
	 *
	 * @param  listenerId the demand is sent to the specified listener (LyDemandTransmitter
	 *                    interface)
	 * @return The listener that processes the demand.
	 */
	public LyDemandListener process(String listenerId, LyDemand demand)
	{
		LyDemandListener listener;

		// ** First pass : attempt to use CACHE (if id specified) **
		if ((listenerId != null)
			&& (_controllers != null))
		{
			listener = _controllers.get(listenerId);

			if (listener != null)
			{
				try
				{
					listener.demandPerformed(demand);
				}
				catch (Exception e)
				{
					e.printStackTrace();
				}
				return listener;
			}
		}

		// ** Otherwise manage full list **
		if (_demandListeners != null)
		{
			int i, n = _demandListeners.size();

			if (n == 0)
				return null;

			// Duplicate list if several listeners (some may remove themselves)
			Vector<LyDemandListener> listeners = null;

			synchronized (_demandListeners)
			{
				listeners = new Vector<LyDemandListener>(_demandListeners);
			}

			n = listeners.size();

			for (i = 0; i < n; i++)
			{
				listener = listeners.elementAt(i);
				if (listenerId == null)
					listener.demandPerformed(demand);
				else if (listenerId.equals(listener.getId()))
				{
					// found id
					listener.demandPerformed(demand);
					return listener;
				}
			}
		}

		return null;
	} // end method process

	/**
	 * Retrieve a LyDemandListener by its id.
	 *
	 * @param listenerId which has to be retrieve
	 */
	public synchronized LyDemandListener getListenerById(String listenerId)
	{
		if ((_controllers == null) || (listenerId == null))
			return null;

		return _controllers.get(listenerId);
	}

	/**
	 * Register a message.
	 *
	 * @param messageId The message id to register.
	 * @param type      Type of the message.
	 * @param message   The message.
	 */
	public void registerMessageWindow(Object messageId, String type, String message)
	{
		LyDemandPlayer player = getPlayer();

		if (player instanceof LyMessageWindowRegister)
			((LyMessageWindowRegister)player).registerMessageWindow(messageId, type, message);
	}

	/**
	 * Sets the result for a message.
	 *
	 * @param message The message id.
	 * @param result  The result.
	 */
	public void setMessageWindowResult(Object messageId, int result)
	{
		LyDemandPlayer player = getPlayer();

		if (player != null)
			((LyMessageWindowRegister)player).setMessageWindowResult(messageId, result);
	}

	/**
	 * Gets demand player.
	 */
	public LyDemandPlayer getPlayer()
	{
		if (_player != null)
			return _player;
		else if (_environment != null)
			return (LyDemandPlayer)_environment.getPlayer();
		else
			return null;
	}

	/**
	 * Sets demand player.
	 */
	public void setPlayer(LyDemandPlayer player)
	{
		_player = player;
	}

	/**
	 * Used in test mode to play next demand in test file.
	 */
	public boolean playNext()
	{
		return ((_player == null) ? false : _player.playNext());
	}

	/**
	 * Gets delay between plays.
	 */
	public int getDelay()
	{
		int delay = ((_player == null) ? 0 : _player.getDelay());

		if (delay < 1000)
			return 1;

		return (delay / 1000);
	}
} // end class LySimpleDemandTransmitter
