/*
 * LySimpleParameterValues.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current Version: $Id: LySimpleParameterValues.java 33969 2012-01-30 09:12:52Z nhamon $.
 */

package leon.view.event;

import java.io.*;
import java.util.*;
import leon.misc.*;
import leon.view.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;


/**
 * Retrieve pairs id / values when submiting data from a view.
 *
 * @author Lyria
 */
public class LySimpleParameterValues
	implements LyParameterValues,
		Serializable
{
	/**
	 * 
	 */
	private static final long serialVersionUID = 1204009833727551160L;

	public static final String				    __VERSION = "$Id: LySimpleParameterValues.java 33969 2012-01-30 09:12:52Z nhamon $";

	/** Separator used in javascript method for relations. */
	public static char						    SEP = LyTools.OBJECT_SEP;

	// Reference to the data Structure
	protected transient LyComponent			    _root;

	// Values stored by association id, value. (used only in replay mode).
	protected transient HashMap<String, String> _values;

	/**
	 * Constructor.
	 */
	public LySimpleParameterValues(LyComponent root)
	{
		_root = root;
		_values = null;
	}

	/**
	 * Returns a string containing the lone value of the specified parameter, or null if the
	 * parameter does not exist.
	 */
	public String getParameter(String param)
	{
		if (_values != null)
		{
			String val = _values.get(param);

			if (val != null)
				return val;
		}

		String[] values = getParameters(param, _root);

		if (values == null)
			return null;
		else if (values.length == 1)
			return values[0];
		else
			return LyString.listToString(values, SEP);
	}

	/**
	 * recursive method to seek the formFields.
	 */
	protected String[] getParameters(String itemId, LyComponent currentComp)
	{
		if (currentComp == null)
			return null;
		if (currentComp instanceof LyFormField)
		{
			String id = ((LyFormField)currentComp).getId();

			if ((id != null) && id.equals(itemId))
			{
				// Found
				LyFieldItem item = ((LyFormField)currentComp).getFieldItem();
				String	    value = item.getValue();

				return LyString.stringToList(value, SEP);
			}
		}

		LyComponent child;
		int		    i, n = currentComp.getChildCount();
		String[]    parameters = null;

		// Go through component list at this level
		for (i = 0; (i < n) && (parameters == null); i++)
		{
			child = currentComp.getChild(i);
			parameters = getParameters(itemId, child);
		}

		return parameters;
	} // end method getParameters

	/**
	 * Returns the parameter names for this request as an enumeration of strings, or an empty
	 * enumeration if there are no parameters or the input stream is empty.
	 */
	public String[] getParameterNames()
	{
		Vector<String> result = new Vector<String>(50);

		getParameterNames(result, _root);

		if (result.size() == 0)
			return null; // Empty

		// Build list
		String[] names = new String[result.size()];

		result.copyInto(names);
		return names;
	}

	/**
	 * recursive method to seek the formFields.
	 */
	protected void getParameterNames(Vector<String> result, LyComponent currentComp)
	{
		if (currentComp == null)
			return;

		if (currentComp instanceof LyFormField)
		{
			String id = ((LyFormField)currentComp).getId();

			if (id != null)
				result.addElement(id);
		}

		LyComponent child;
		int		    i, n = currentComp.getChildCount();
		LyFormField	formField;
		LyForm		form;

		// Go through component list at this level
		for (i = 0; i < n; i++)
		{
			child = currentComp.getChild(i);

			if (child instanceof LyFormField)
			{
				formField = (LyFormField)child;
				form = formField.getParentForm();  
				
				if ((form != null)
					&& form.isEditable())
				{
					String	id = formField.getId();
	
					if (id != null)
						result.addElement(id);
				}
			}
			else
				getParameterNames(result, child);
		}
	} // end method getParameterNames

	/**
	 * Returns the values of the specified parameter for the request as an array of strings, or null
	 * if the named parameter does not exist.
	 */
	public String[] getParameterValues(String param)
	{
		return (getParameters(param, _root));
	}

	/**
	 * Update a view with the specified values.
	 */
	public void updateViewFromValues(LyComponent component)
	{
		updateViewFromValues(component, false);
	}

	/**
	 * Update a view with the specified values.
	 */
	public void updateViewFromValues(LyComponent component, boolean needValueChange)
	{
		updateViewFromValues(component, null, needValueChange);
	}
	
	/**
	 * Update a view with the specified values.
	 */
	public void updateViewFromValues(LyComponent component, LyComponent priorityComponent, boolean needValueChange)
	{
		// Do nothing : view is updated by win components
		if (_values == null)
			return;

		// In replay mode we must update the view layer with values.
		int					   l, j;
		LyComponent			   subComponent = null;
		LyFieldItem			   fieldItem;
		LySetLabelInterface    setLabel;
		String				   parameterName;
		String[]			   parameterValues;
		String				   parameterValue = null;
		boolean				   status;
		int					   nParameters;
		Vector<String>		   selList;
		LyActionDemand		   demand;
		int					   demandType = LyActionDemand.TYPE_VALUE_CHANGED;
		String				   source;
		Set<String>			   keys = _values.keySet();
		Iterator<String>	   it = keys.iterator();
		String				   label;
		LyBoolean.Status	   booleanStatus;
		boolean				   changed;
		Vector<LyActionDemand> demands = null;
		Vector<LyComponent>    changedComponents = null;

		while (it.hasNext())
		{
			demand = null;
			setLabel = null;
			subComponent = null;
			parameterName = it.next();
			parameterValue = _values.get(parameterName);
			parameterValues = LyString.stringToList(parameterValue, SEP);
			changed = false;

			source = parameterName;

			// Get rid of dummy parameters (takes a long tilme to search for nothing)
			if (parameterName.endsWith("_date"))
			{
				String realParam = parameterName.substring(0, parameterName.length() - 5);

				if (getParameter(realParam) != null)
					continue; // Dummy duplicate of realParam
			}
			if (parameterName.endsWith("File"))
			{
				String realParam = parameterName.substring(0, parameterName.length() - 4);

				if (getParameter(realParam) != null)
					continue; // Dummy duplicate of realParam
			}

			// Find the component with the id of the parameter
			// First of all, look on top level (optimization) : fieldsinfos are stored here
			// which avoid a deep recursive search in the tree
			LyView view = component.getParentView();

			if (view != null)
				subComponent = view.getChildFromCache(parameterName);

			if (subComponent == null)
				subComponent = component.getComponent(parameterName);

			if (subComponent instanceof LySetLabelInterface)
				setLabel = (LySetLabelInterface)subComponent;
			else if (subComponent instanceof LyFormField)
			{
				demand = new LyActionDemand(demandType, source);
				demand.setParameterValues(this);

				// If form field retrieve the field item
				fieldItem = ((LyFormField)subComponent).getFieldItem();

				if (fieldItem instanceof LySetLabelInterface)
					setLabel = (LySetLabelInterface)fieldItem;
				else
				{
					nParameters = ((parameterValues == null) ? 0 : parameterValues.length);
					if (nParameters != 0)
					{
						// create a vector of selected items
						selList = new Vector<String>(nParameters);
						for (j = 0; j < nParameters; j++)
							selList.addElement(parameterValues[j]);
					}
					else
						selList = new Vector<String>();

					if (fieldItem instanceof LyRelation)
					{
						int    nLink = ((LyRelation)fieldItem).getLinkCount();
						LyLink link;

						for (j = 0; j < nLink; j++)
						{
							link = ((LyRelation)fieldItem).getLink(j);
							status = selList.contains(link.getId());
							if (link.isSelected() != status)
							{
								link.setSelected(status);
								changed = true;
							}
						}
					}
					else if (fieldItem instanceof LyChoice)
					{
						int		 nChoice = ((LyChoice)fieldItem).getOptionCount();
						LyOption option;

						for (j = 0; j < nChoice; j++)
						{
							option = ((LyChoice)fieldItem).getOption(j);
							status = selList.contains(option.getId());
							if (option.isChecked() != status)
							{
								option.setChecked(status);
								changed = true;
							}
						}
					}
					else if (fieldItem instanceof LyList)
					{
						int			 nList = ((LyList)fieldItem).getListOptionCount();
						LyListOption option;

						for (j = 0; j < nList; j++)
						{
							option = ((LyList)fieldItem).getListOption(j);
							status = selList.contains(option.getId());
							if (option.isSelected() != status)
							{
								option.setSelected(status);
								changed = true;
							}
						}
					}
					else if (fieldItem instanceof LyBoolean)
					{
                        if (selList.size() == 0)
                        {
                            if (!((LyFormField)subComponent).isOptional()
                                    && (fieldItem.getViewManager().getEnvironment().getPlayer() != null))
                                // Auto-correction for replay mode when choice type was changed to boolean (AC)
                                booleanStatus = LyBoolean.Status.UNCHECKED;
                            else
                                booleanStatus = LyBoolean.Status.EMPTY;
                        }
                        else if (selList.elementAt(0).equals(LyBoolean.Status.EMPTY.toString()))
                            booleanStatus = LyBoolean.Status.EMPTY;
                        else if (selList.elementAt(0).equals(LyBoolean.Status.UNCHECKED.toString()))
                            booleanStatus = LyBoolean.Status.UNCHECKED;
                        else
                            booleanStatus = LyBoolean.Status.CHECKED;

						if (booleanStatus != ((LyBoolean)fieldItem).getStatus())
						{
							((LyBoolean)fieldItem).setStatus(booleanStatus);
							changed = true;
						}
					}
					else if (fieldItem instanceof LyDictionary)
					{
						Vector<LyEntry> result = new Vector<LyEntry>();
						LyEntry		    entry;

						((LyDictionary)fieldItem).getComponents(LyEntry.class, result);
						l = result.size();

						for (j = 0; j < l; j++)
						{
							entry = result.elementAt(j);
							changed |= updateEntryFromValues(selList, entry);
						}
					}
				}
			}

			if ((setLabel instanceof LyText) && ((LyText)setLabel).isSecret()
					&& (parameterValues != null))
			{
				parameterValue = LyString.listToString(parameterValues, SEP); // XXX : SEP ?

				if ((parameterValue != null) && (parameterValue.trim().length() > 0)
						&& (parameterValue.replace('*', ' ').trim().length() == 0))
					setLabel = null; // secret field has not been modified ('****')
			}

			if (setLabel != null)
			{
				demand = new LyActionDemand(demandType, source);
				demand.setParameterValues(this);

				parameterValue = LyString.listToString(parameterValues, SEP); // XXX : SEP ?

				if (parameterValue != null)
				{
					if ((setLabel instanceof LyText) && (((LyText)setLabel).getNRows() > 1))
						// Remove '\r' from multi-line text fields
						parameterValue = LyString.replace(parameterValue, "\r", "");

					if ((setLabel instanceof LyDate) && (parameterValue.length() == 0))
						// Date case : if value is "", use null instead
						parameterValue = null;
				}

				if (setLabel instanceof LyGetLabelInterface)
				{
					label = ((LyGetLabelInterface)setLabel).getLabel();
					if (((label == null)
								&& (parameterValue != null))
							|| ((label != null)
								&& (parameterValue == null))
							|| ((label != null)
								&& (parameterValue != null)
								&& !label.equals(parameterValue)))
						changed = true;
				}

				setLabel.setLabel(parameterValue);
			}

			if (needValueChange
					&& changed
					&& (demand != null)
					&& (subComponent != null))
			{
				if (demands == null)
				{
					demands = new Vector<LyActionDemand>(1);
					changedComponents = new Vector<LyComponent>(1);
				}
				demands.add(demand);
				changedComponents.add(subComponent);
			}
		}

		if (demands == null)
			return;

		// Post demands when all changes are identified before other indirect changes
		LyViewManager vm = component.getViewManager();
		String		  rootControllerId;
		int			  i, n;

		for (i = 0, n = demands.size(); i < n; i++)
		{
			demand = demands.elementAt(i);
			subComponent = changedComponents.elementAt(i);

			rootControllerId = subComponent.getRootController();
			vm.processDemand(rootControllerId, demand);
		}
	} // end method updateViewFromValues

	private boolean updateEntryFromValues(Vector<String> list, LyEntry entry)
	{
		int     i, n = entry.getEntryCount();
		LyEntry innerEntry;
		boolean checked = list.contains(entry.getId());
		boolean changed = false;

		if (entry.isChecked() != checked)
		{
			entry.setChecked(checked);
			changed = true;
		}

		for (i = 0; i < n; i++)
		{
			innerEntry = entry.getEntry(i);
			changed |= updateEntryFromValues(list, innerEntry);
		}

		return changed;
	}

	public LyComponent getComponent()
	{
		return _root;
	}

	public HashMap<String, String> getValues()
	{
		return _values;
	}

	public void setComponent(LyComponent root)
	{
		_root = root;
	}

	public void setValues(HashMap<String, String> values)
	{
		_values = values;
	}
} // end class LySimpleParameterValues
