/*
 * LyEmbossImageFilter.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.filters;

import java.awt.*;


/**
 * This class is used to create embossed images.
 * @version $Id: LyEmbossImageFilter.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyEmbossImageFilter extends LyImageFilter
{
	public static final String __VERSION =
		"$Id: LyEmbossImageFilter.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Name of the filter used for keys in table of images. */
	public static String	   EMBOSS_FILTER_NAME = "_embossFilter";

	/**
	 * Constructor with no parameter. Calls the super implementation.
	 */
	public LyEmbossImageFilter()
	{
		super();
	}

	/**
	 * Gets the filtered image associated to the given image and background color. It it is not
	 * found, creates it.
	 * @param  image      The image that must be filtered.
	 * @param  background The color that must be used every two pixels.
	 * @return The filtered image.
	 */
	public static Image getFilteredImage(Image image, Color background)
	{
		if ((image == null) || (background == null))
			return null;

		// Creation of a unique key : imSource, color
		StringBuffer key = new StringBuffer(10);

		key.append(image.hashCode());
		key.append('-');
		key.append(background.hashCode());
		key.append('-');
		key.append(EMBOSS_FILTER_NAME);

		Image result = null;

		if (_images != null)
			result = (Image)_images.get(key.toString());

		// If image not found, create a new one
		if (result == null)
		{
			int[]     pixels = getPixels(image);
			Dimension imageDim = getImageSize(image);

			pixels = emboss(pixels, imageDim.width, imageDim.height, 45, 2, background.getRed(),
					background.getGreen(), background.getBlue());
			result = createImage(pixels, imageDim.width, imageDim.height);

			putCache(key.toString(), result);
		}

		return result;
	} // end method getFilteredImage

	/**
	 * Creates an array of pixels representing the filtered data for the given array of pixels.
	 * @param  sourcePixels The source pixels that must be filtered.
	 * @param  width        The width of the image represented by the given array of pixels.
	 * @param  height       The height of the image represented by the given array of pixels.
	 * @param  angle        The angle that must be used by the filter.
	 * @param  power        The power that must be used by the filter.
	 * @param  red          The red component of the color that must be used by the filter.
	 * @param  green        The green component of the color that must be used by the filter.
	 * @param  blue         The blue component of the color that must be used by the filter.
	 * @return The array of pixels that represents the filtered given array of pixels.
	 */
	protected static int[] emboss(int[] sourcePixels, int width, int height, double angle,
		double power, int red, int green, int blue)
	{
		int[]  pixels = new int[width * height];
		double angleRadians = angle / (180.0 / Math.PI);
		int    i, c, r1, g1, b1, r2, g2, b2, r, g, b;
		int    light = (int)(Math.round(Math.sin(angleRadians)) * width)
			+ (int)(Math.round(Math.cos(angleRadians)));

		for (i = width + 1; i < ((width * height) - width - 1); i++)
		{
			c = sourcePixels[i];
			r1 = (c & 0xff0000) >> 16;
			g1 = (c & 0xff00) >> 8;
			b1 = c & 0xff;

			c = sourcePixels[i - light];
			r2 = (c & 0xff0000) >> 16;
			g2 = (c & 0xff00) >> 8;
			b2 = c & 0xff;

			r = Math.min(Math.max(red + (int)((r2 - r1) * power), 0), 255);
			g = Math.min(Math.max(green + (int)((g2 - g1) * power), 0), 255);
			b = Math.min(Math.max(blue + (int)((b2 - b1) * power), 0), 255);

			pixels[i] = (c & 0xff000000) + (r << 16) + (g << 8) + b;
		}

		int color = (red << 16) + (green << 8) + blue;

		for (i = 0; i < width; i++)
		{
			pixels[i] = (sourcePixels[i] & 0xff000000) + color;
			pixels[(width * height) - i - 1] = (sourcePixels[(width * height) - i - 1] & 0xff000000)
				+ color;
		}

		for (i = 0; i < (width * height); i += width)
		{
			pixels[i] = (sourcePixels[i] & 0xff000000) + color;
			pixels[i + width - 1] = (sourcePixels[i + width - 1] & 0xff000000) + color;
		}

		return pixels;
	} // end method emboss

} // end class LyEmbossImageFilter
