/*
 * LyGrayImageFilter.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.filters;

import java.awt.*;


/**
 * This class is used to create grayed images. The proportion of gray may be customized.
 * @version $Id: LyGrayImageFilter.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyGrayImageFilter extends LyImageFilter
{
	public static final String __VERSION =
		"$Id: LyGrayImageFilter.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Name of the filter used for keys in table of images. */
	public static String	   GRAY_FILTER_NAME = "_grayFilter";

	/** Default gray proportion. */
	public static final int    GRAY_PROP = 75;

	/**
	 * Constructor with no parameter. Calls the super implementation.
	 */
	public LyGrayImageFilter()
	{
		super();
	}

	/**
	 * Gets a filtered image with the default gray proportion. Calls {@link #getFilteredImage(Image,
	 * int) getFilteredImage(image, GRAY_PROP)}.
	 * @param  image The image that must be filtered.
	 * @return The filtered image.
	 */
	public static Image getFilteredImage(Image image)
	{
		return getFilteredImage(image, GRAY_PROP);
	}

	/**
	 * Gets the filtered image associated to the given image and background color. It it is not
	 * found, creates it.
	 * @param  image    The image that must be filtered.
	 * @param  grayProp The proportion of gray that must be used by the filter.
	 * @return The filtered image.
	 */
	public static Image getFilteredImage(Image image, int grayProp)
	{
		if (image == null)
			return null;

		// Creation of a unique key : imSource, color
		StringBuffer key = new StringBuffer(10);

		key.append(image.hashCode());
		key.append('-');
		key.append(GRAY_FILTER_NAME);

		Image result = null;

		if (_images != null)
			result = (Image)_images.get(key.toString());

		// If image not found, create a new one
		if (result == null)
		{
			int[]     pixels = getPixels(image);
			Dimension imageDim = getImageSize(image);

			pixels = gray(pixels, imageDim.width, imageDim.height, grayProp);
			result = createImage(pixels, imageDim.width, imageDim.height);

			putCache(key.toString(), result);
		}

		return result;
	} // end method getFilteredImage

	/**
	 * Creates an array of pixels representing the filtered data for the given array of pixels.
	 * @param  sourcePixels The source pixels that must be filtered.
	 * @param  width        The width of the image represented by the given array of pixels.
	 * @param  height       The height of the image represented by the given array of pixels.
	 * @param  grayProp     The proportion of gray that must be used by the filter.
	 * @return The array of pixels that represents the filtered given array of pixels.
	 */
	protected static int[] gray(int[] sourcePixels, int width, int height, int grayProp)
	{
		int[] pixels = new int[width * height];
		int   t, c, r, g, b, gray, prop, col, index;

		for (index = 0; index < (width * height); index++)
		{
			prop = 0;
			col = 1;

			c = sourcePixels[index];
			t = ((c & 0xff000000));
			r = (c & 0xff0000) >> 16;
			g = (c & 0xff00) >> 8;
			b = c & 0xff;
			gray = ((r * 3) + (g * 4) + (b * 2)) / 9;

			// Calculating the gray proportion
			if (grayProp >= 100)
			{
				prop = 1;
				col = 0;
			}
			else if (grayProp <= 0)
			{
				prop = 0;
				col = 1;
			}
			else
			{
				prop = grayProp;
				col = 100 - grayProp;
			}

			pixels[index] = (t | ((((prop * gray) + (col * r)) / (prop + col)) << 16)
						| ((((prop * gray) + (col * g)) / (prop + col)) << 8)
						| (((prop * gray) + (col * b)) / (prop + col)));
		}

		return pixels;
	} // end method gray
} // end class LyGrayImageFilter
