/*
 * LyImageFilter.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.filters;

import java.awt.*;
import java.awt.image.*;
import java.util.*;


/**
 * This class is used to create filtered images. This class is abstract since it does not represent
 * a real filter. It gives only useful static methods to create and manipulate images. Each type of
 * filter must inherits this class and implements its own filtering algorithm in a method.
 * @version $Id: LyImageFilter.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public abstract class LyImageFilter
{
	public static final String __VERSION =
		"$Id: LyImageFilter.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Table of images (for optimization). Keys are strings that identify images and filters used.
	 * Objects stored are the associated filtered images. This is an image cache.
	 */
	protected static Hashtable<String, Object> _images;

	/**
	 * Constructor with no parameter. Simply calls the {@link #init()} method.
	 */
	public LyImageFilter()
	{
		init();
	}

	/**
	 * Clears the images cache.
	 * @see LyImageFilter#putCache(String,Image)
	 */
	public static void clearCache()
	{
		if(_images == null)
			return;
		
		Enumeration<Object> elements = _images.elements();
		Object elem;
		while(elements.hasMoreElements())
		{
			elem = elements.nextElement();
			if (elem instanceof Image)
			{
				((Image)elem).flush();
			}
		}
		_images = null;
	}

	/**
	 * Adds the specified pair (key-image) to the images cache.
	 * @param key   The key used to register the filtered image. This key must be unique for a given
	 *              image and a given filter.
	 * @param image The filtered image that must be registered.
	 * @see   LyImageFilter#clearCache()
	 */
	public static void putCache(String key, Image image)
	{
		if (_images == null)
			_images = new Hashtable<String, Object>(10);

		_images.put(key, image);
	}

	/**
	 * Creates an array of pixels from the given image.
	 * @param  image The image that must be analyzed.
	 * @return An array of integers that describes the pixels of the image.
	 */
	protected static int[] getPixels(Image image)
	{
		int			 width = image.getWidth(null);
		int			 height = image.getHeight(null);
		int[]		 pixels = new int[width * height];
		PixelGrabber pg = new PixelGrabber(image, 0, 0, width, height, pixels, 0, width);

		try
		{
			pg.grabPixels();
		}
		catch (InterruptedException e)
		{
		}

		return pixels;
	}

	/**
	 * Gets the given image dimension.
	 * @param  image The image that must be analyzed.
	 * @return The dimension of the given image.
	 */
	protected static Dimension getImageSize(Image image)
	{
		int width = image.getWidth(null);
		int height = image.getHeight(null);

		return new Dimension(width, height);
	}

	/**
	 * Creates an image from an array of pixels.
	 * @param  pixels The array of pixels used to create the image.
	 * @param  width  The with of the created image.
	 * @param  height The height of the created image.
	 * @return The image created from the given array of pixels.
	 */
	protected static Image createImage(int[] pixels, int width, int height)
	{
		Toolkit tk = Toolkit.getDefaultToolkit();

		return tk.createImage(new MemoryImageSource(width, height, pixels, 0, width));
	}

	/**
	 * Init method used to initialize attributes of a new image filter. This method is called by
	 * constructors and may be overriden in subclasses to complete initialization. Does nothing
	 * here.
	 */
	protected void init()
	{
	}
} // end class LyImageFilter
