/*
 * LyReplaceColorImageFilter.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.filters;

import java.awt.*;


/**
 * This class is used to replace a color by another one in an image.
 * @version $Id: LyReplaceColorImageFilter.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyReplaceColorImageFilter extends LyImageFilter
{
	public static final String __VERSION =
		"$Id: LyReplaceColorImageFilter.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Name of the filter used for keys in table of images. */
	public static String	   REPLACE_COLOR_FILTER_NAME = "_replaceColorFilter";

	/**
	 * Constructor with no parameter. Calls the super implementation.
	 */
	public LyReplaceColorImageFilter()
	{
		super();
	}

	/**
	 * Gets the filtered image associated to the given image and background color. It it is not
	 * found, creates it.
	 * @param  image       The image that must be filtered.
	 * @param  colorSource The color that must be replaced. A null parameter means that transparency
	 *                     must be replaced.
	 * @param  colorDest   The color that must be used to replace the source color. A null parameter
	 *                     means that the source color must be replaced by transparency.
	 * @return The filtered image.
	 */
	public static Image getFilteredImage(Image image, Color colorSource, Color colorDest)
	{
		// If colorSource is null, replace transparency by the new color
		if (image == null)
			return null;

		// Creation of a unique key : imSource, color
		StringBuffer key = new StringBuffer(10);

		key.append(image.hashCode()).append('-');

		if (colorSource == null)
			key.append("null");
		else
			key.append(colorSource.hashCode());

		key.append('-');

		if (colorDest == null)
			key.append("null");
		else
			key.append(colorDest.hashCode());

		key.append('-');
		key.append(REPLACE_COLOR_FILTER_NAME);

		Image result = null;

		if (_images != null)
			result = (Image)_images.get(key.toString());

		// If image not found, create a new one
		if (result == null)
		{
			int[]     pixels = getPixels(image);
			Dimension imageDim = getImageSize(image);

			pixels = replaceColor(pixels, imageDim.width, imageDim.height, colorSource, colorDest);
			result = createImage(pixels, imageDim.width, imageDim.height);

			putCache(key.toString(), result);
		}

		return result;
	} // end method getFilteredImage

	/**
	 * Creates an array of pixels representing the filtered data for the given array of pixels.
	 * @param  sourcePixels The source pixels that must be filtered.
	 * @param  width        The width of the image represented by the given array of pixels.
	 * @param  height       The height of the image represented by the given array of pixels.
	 * @param  colorSource  The color that must be replaced. A null parameter means that
	 *                      transparency must be replaced.
	 * @param  colorDest    The color that must be used to replace the source color. A null
	 *                      parameter means that the source color must be replaced by transparency.
	 * @return The array of pixels that represents the filtered given array of pixels.
	 */
	protected static int[] replaceColor(int[] sourcePixels, int width, int height,
		Color colorSource, Color colorDest)
	{
		int[]   pixels = new int[width * height];
		int     i, j;
		boolean replaceTransparency = (colorSource == null);
		boolean setTransparency = (colorDest == null);
		int     rgbSource = 0;
		int     rgbDest = 0;

		if (colorSource != null)
			rgbSource = colorSource.getRGB();

		if (colorDest != null)
			rgbDest = colorDest.getRGB();

		for (i = 0; i < height; i++)
		{
			for (j = 0; j < width; j++)
			{
				if (replaceTransparency)
				{
					if (((sourcePixels[(width * i) + j] & 0xff000000) >> 24) == 0)
						pixels[(width * i) + j] = 0xff000000 | rgbDest;

				}
				else if (setTransparency)
				{
					if ((sourcePixels[(width * i) + j] & 0x00ffffff) == (rgbSource & 0x00ffffff))
						pixels[(width * i) + j] = 0;
					else
						pixels[(width * i) + j] = sourcePixels[(width * i) + j];
				}
				else
				{
					if ((sourcePixels[(width * i) + j] & 0x00ffffff) == (rgbSource & 0x00ffffff))
						pixels[(width * i) + j] = (sourcePixels[(width * i) + j] & 0xff000000)
								| (rgbDest & 0x00ffffff);
					else
						pixels[(width * i) + j] = sourcePixels[(width * i) + j];
				}
			}
		}

		return pixels;
	} // end method replaceColor

} // end class LyReplaceColorImageFilter
