/*
 * LyGraphicalObject.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current Version: $Id: LyGraphicalObject.java 30998 2011-05-13 13:28:12Z jlsaouli $.
 */

package leon.view.graphic;

import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;


/**
 * The LyGraphicalObject class implements root object for all win components.
 */
public abstract class LyGraphicalObject
	implements LyGraphicalComponent
{
	public static final String __VERSION =
		"$Id: LyGraphicalObject.java 30998 2011-05-13 13:28:12Z jlsaouli $";

	/** View component. */
	protected LyComponent	   _viewComponent;
	
	/** Environment. */
	protected LyEnvironment	   _environment;

	/**
	 * Constructs abstract graphical component.
	 */
	public LyGraphicalObject(LyComponent viewComponent)
	{
		initViewComponent(viewComponent);
	}

	/**
	 * Initialize view component for graphical object.
	 *
	 * @param viewComponent current view component
	 */
	protected void initViewComponent(LyComponent viewComponent)
	{
		_viewComponent = viewComponent;
		_viewComponent.setInternalView(this);
		_environment = viewComponent.getEnvironment();
	}

	/**
	 * Free resources allocated to this component.
	 *
	 * @see leon.view.setinterface.LyFreeInterface#free
	 */
	public void free()
	{
		_viewComponent = null;
		_environment = null;
	}

	/**
     * Internal code used as identifier for web components.
     * @return Internal code as a string.
     */
    public String getIdCode()
    {
        if (LyEnvironment.TEST_MODE
                && (_viewComponent != null))
            return _viewComponent.getIdCode();
        else
            return Integer.toString(hashCode());
    }
    
	/**
	 * Checks if graphical object is freeing or not.
	 *
	 * @return True if freed, false otherwise.
	 */
	public boolean isFreeing()
	{
		LyComponent component = _viewComponent;

		while (component != null)
		{
			if (component.isFreeing())
				return true;
			else
				component = component.getParent();
		}

		return false;
	}

	/**
	 * Checks if graphical object was freed or not.
	 *
	 * @return True if freed, false otherwise.
	 */
	public boolean isFreed()
	{
		return (_viewComponent == null);
	}

	/**
	 * Gets view component corresponding to this graphical object.
	 *
	 * @return The view component for this graphical component.
	 */
	public final LyComponent getViewComponent()
	{
		return _viewComponent;
	}

	/**
	 * Returns the position and size of current component relatively to given component.
	 * 
	 * @param parentComponent the component used to compute relative position of
	 * current component. If null, position of current component is computed with the top-left
	 * corner of the screen as the origin.
	 * @return an array of 4 integers (x position, y position, width, height) representing
	 * the position and size of current component.
	 */
	public int[] getBounds(LyComponent parent)
	{
		return null;
	}
	
	/**
	 * Returns the position and size of given sub component relatively to current component.
	 * 
	 * @param childrenComponent the sub component whose relative position must be returned.
	 * @return an array of 4 integers (x position, y position, width, height) representing
	 * the position and size of given sub component.
	 */
	public int[] getChildrenBounds(LyComponent children)
	{
		return null;
	}
	
	/**
	 * Retrieves controller id of this graphical component. Warning ! This method may be overridden
	 * in sub-classes to redirect on parent controller for example.
	 *
	 * @return The controller id of this component, may be null.
	 */
	public String getController()
	{
		if (_viewComponent instanceof LyGetControllerInterface)
			return ((LyGetControllerInterface)_viewComponent).getController();
		else
			return null;
	}

	/**
	 * Gets environment of this graphical object.
	 *
	 * @return The environment of this graphical object.
	 */
	public final LyEnvironment getEnvironment()
	{
		if ((_environment == null)
				&& (_viewComponent != null))
			_environment = _viewComponent.getEnvironment();
		
		return _environment;
	}

	/**
	 * Gets current view manager.
	 *
	 * @return The current view manager.
	 */
	public final LyViewManager getViewManager()
	{
		return (LyViewManager)getEnvironment().getViewManager();
	}

	/**
	 * Retrieves a value from the contextual environment.
	 *
	 * @param  The key associated with the environment parameter.
	 * @return The environment variable or null if the value could not be found.
	 * @see    LyEnvironment#getEnv
	 */
	public final String getEnv(String key)
	{
		return getEnvironment().getEnv(key);
	}

	/**
	 * Convenient routine to get a formated message without argument.
	 *
	 * @see LyEnvironment#getString(String)
	 */
	public final String getString(String key)
	{
		return getEnvironment().getString(key);
	}

	/**
	 * Convenient routine to get a formated message without a single argument.
	 *
	 * @see LyEnvironment#getString(String,Object)
	 */
	public final String getString(String key, Object argument)
	{
		return getEnvironment().getString(key, argument);
	}

	/**
	 * Constructs a string from a key and an array of arguments.
	 *
	 * @param  key      The key to be translated.
	 * @param  argument The array of arguments used to construct the string.
	 * @return The string formated with the arguments.
	 * @see    LyEnvironment#getString(String,Object[])
	 */
	public final String getString(String key, Object[] arguments)
	{
		return getEnvironment().getString(key, arguments);
	}

	/**
	 * Convenient routine to get a formated message without argument.
	 *
	 * @see LyEnvironment#getMessage
	 */
	public final String getMessage(String key)
	{
		return getMessage(key, null);
	}

	/**
	 * Convenient routine to get a formated message with given arguments.
	 *
	 * @see LyEnvironment#getMessage
	 */
	public String getMessage(String key, Object[] arguments)
	{
		return getEnvironment().getMessage(key, arguments);
	}

	/**
	 * Retrieves a resource from the application environment.
	 *
	 * @param  key The resource name (key).
	 * @return The value of the requested resource, or null if it could not be found.
	 * @see    LyEnvironment#getResource
	 */
	public final String getResource(String key)
	{
		return getEnvironment().getResource(key);
	}

	/**
	 * Checks if a resource is set to true or 1.
	 *
	 * @param  key The resource name (key).
	 * @return True if the resource exits and is set to true or 1, false otherwise.
	 * @see    LyEnvironment#testResource(String)
	 */
	public final boolean testResource(String key)
	{
		return getEnvironment().testResource(key);
	}

	/**
	 * Retrieves an integer resource from the application environment.
	 *
	 * @param  key The resource name (key).
	 * @return The value of the requested resource, or -1 if it could not be found.
	 * @see    LyEnvironment#getResource
	 */
	public final int getIntResource(String key)
	{
		return getEnvironment().getIntResource(key);
	}

	/**
	 * Translates the given key to a String that may be shown to the user.
	 *
	 * @param  key The key to be translated.
	 * @return The translated key if found in local dictionary. If not found, the key is returned
	 *         'as is'.
	 * @see    LyEnvironment#translate
	 */
	public final String translate(String key)
	{
		return getEnvironment().translate(key);
	}

	/**
	 * Replaces within a given String some $properties$ found in the environment.
	 *
	 * @param  text The original string, before any substitution.
	 * @return The string after all substitutions.
	 * @see    LyString#expand
	 */
	public String expand(String text)
	{
		return getEnvironment().getString().expand(text);
	}

	/**
	 * Checks if an image is animated.
	 *
	 * @param  key The resource name (key) of the image.
	 * @return True if the image is animated, false otherwise.
	 * @see    LyEnvironment#isAnimated
	 */
	public final boolean isImageAnimated(String key)
	{
		return getEnvironment().isImageAnimated(key);
	}

	/**
	 * Processes demands on this component.
	 *
	 * @param demand the demand to process.
	 */
	public final void processDemand(LyDemand demand)
	{
		processDemand(getController(), demand);
	}

	/**
	 * Processes demands on this component.
	 *
	 * @param controllerId the demand is sent to the specified controller.
	 * @param demand       the demand to process.
	 */
	public void processDemand(String controller, LyDemand demand)
	{
		if (controller == null)
			return;

		getViewManager().processDemand(controller, demand);
	}
} // end class LyGraphicalObject
