package leon.view.mobile;

import java.io.*;
import java.util.*;

import leon.misc.*;
import leon.view.*;
import leon.view.LyMapBase.SelectionDraw;
import leon.view.LyMapBase.SelectionMode;
import leon.view.LyMapBase.Style;
import leon.view.LyMapBase.TranslationStatus;
import leon.view.LyMapBase.ViewerStatus;
import leon.view.LyMapBase.ZoomStatus;
import leon.view.graphic.*;

public class LyMobileMap extends LyMobileObject
implements LyGraphicalMap
{
	/** The map generic component. */
	protected LyMap						_map;


	/**
	 * Identifier of the map.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyMap				_template;

	/**
	 * Label of the map.
	 */
	protected String			_label;

	/**
	 * Reference of the controller that manages the map (i.e. handles user demands and 
	 * refresh content).
	 */
	protected String			_controller;

	/**
	 * Selection mode of the elements of the map. The following modes are managed:
	 * NONE: No selection is possible
	 * SINGLE: Selection of only one node at a time
	 * MULTIPLE: Selection of several elements at a time
	 */
	protected SelectionMode			_selectionMode;

	/**
	 * Draw style for the elements that are selected on the map.
	 * This type allows defining the way the selection of an element of the map is 
	 * represented. The following types are managed:
	 * FILL: Filled frame style
	 * BORDER: Outlined frame style
	 */
	protected SelectionDraw			_selectionDraw;

	/**
	 * Visualization mode of the map.
	 * The true value indicates a visualization by using a viewer representing a 
	 * small-scale.
	 * The viewer allows locating its position and to move on the map.
	 * In the absence of a viewer, the map manages scroll bars.
	 */
	protected boolean			_viewer;

	/**
	 * Status of the viewer in the viewer mode.
	 * The viewer can have the two following statuses:
	 * SHOW: the viewer is shown.
	 * HIDE: the viewer is hidden.
	 */
	protected ViewerStatus			_viewerStatus;

	/**
	 * X position of the viewer in the map.
	 */
	protected int				_viewerX;

	/**
	 * Y position of the viewer in the map.
	 */
	protected int				_viewerY;

	/**
	 * Indicates if the zoom widget must be displayed on the map.
	 */
	protected boolean			_zoom;

	/**
	 * Indicates if the zoom widget is visible or hidden.
	 */
	protected ZoomStatus			_zoomStatus;

	/**
	 * Step of the zoom widget (must be greater than 1).
	 */
	protected float				_zoomStep;

	/**
	 * Current zoom factor of the map.
	 */
	protected float				_zoomFactor;

	/**
	 * Indicates if the translation widget must be displayed on the map.
	 */
	protected boolean			_translation;

	/**
	 * Step of the translation widget. Used by the translation widget to move the map 
	 * and also by the Struts map component to compute the number of sub-images that 
	 * must be generated for the map background.
	 */
	protected int				_translationStep;

	/**
	 * Indicates if the translation widget is visible or hidden.
	 */
	protected TranslationStatus			_translationStatus;

	/**
	 * Indicates if the map is in three dimensions.
	 * A map in two dimensions manages only the coordinates in abscissa and in ordinate
	 * of the elements that it contains.
	 */
	protected boolean			_spatial;

	/**
	 * Background of the map.
	 * The expected value is the key of an image or else the name of an image.
	 * The pattern representing this image is repeated in order to recover the whole 
	 * visible part of the map.
	 * All the elements that are available on the map are displayed over this 
	 * background.
	 */
	protected String			_background;

	/**
	 * Font of the map title.
	 */
	protected String			_titleFont;

	/**
	 * Background color of the component.
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * Selection color used when nodes or links are selected within the map.
	 */
	protected String			_selectedColor;

	/**
	 * Color used for outline.
	 */
	protected String			_outlineColor;

	/**
	 * Width of the map component.
	 */
	protected int				_width;

	/**
	 * Height of the map component.
	 */
	protected int				_height;

	/**
	 * Style of the map component.
	 */
	protected Style				_style;

	/**
	 * Latitude of the map component.
	 */
	protected String			_latitude;

	/**
	 * Longitude of the map component.
	 */
	protected String			_longitude;


	/* Sub components */

	/**
	 * Components of the map. These components can be map nodes, map links or map 
	 * regions.
	 */
	protected ArrayList<LyComponent>		_components;

	/**
	 * Additional elements of the map.
	 * These elements represent the static part of the elements of the map.
	 * They are not intended to evolve dynamically when the map is used.
	 * Generally, these elements are defined in a file that is associated to the action
	 * used to display the map.
	 */
	protected LyMapContent		_mapContent;

	/**
	 * Layout manager of the map.
	 */
	protected LyLayout			_layout;

	/**
	 * Positioning constraint of the map.
	 */
	protected LyConstraint		_constraint;


	public LyMobileMap(LyMap map)
	{
		super(map);

		_map = map;
		_components = new ArrayList<LyComponent>();
	}


	public void writeXml(PrintWriter pw)
	{
		init();
		
		pw.print("<MAP");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlStringStrict(_id) + "'");
		/*if (_template != null)
			pw.print(" template='" + _template.getId() + "'");*/
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template.getLabel())))
			pw.print(" label='" + LyString.getXmlStringStrict(_label) + "'");
		if (_controller != null)
			pw.print(" controller='" + LyString.getXmlStringStrict(_controller) + "'");
		/*if ((_template == null)
				|| (_selectionMode != _template._selectionMode))
		{
			if (_selectionMode == SelectionMode.NONE)
				pw.print(" selectionMode='NONE'");
			else if (_selectionMode == SelectionMode.SINGLE)
				pw.print(" selectionMode='SINGLE'");
			else if ((_selectionMode == SelectionMode.MULTIPLE) && (_template != null))
				pw.print(" selectionMode='MULTIPLE'");
		}
		if ((_template == null)
				|| (_selectionDraw != _template._selectionDraw))
		{
			if ((_selectionDraw == SelectionDraw.FILL) && (_template != null))
				pw.print(" selectionDraw='FILL'");
			else if (_selectionDraw == SelectionDraw.BORDER)
				pw.print(" selectionDraw='BORDER'");
		}
		if (((_template == null) && !_viewer)
				|| ((_template != null) && (_viewer != _template._viewer)))
			pw.print(" viewer='" + _viewer + "'");
		if ((_template == null)
				|| (_viewerStatus != _template._viewerStatus))
		{
			if ((_viewerStatus == ViewerStatus.SHOW) && (_template != null))
				pw.print(" viewerStatus='SHOW'");
			else if (_viewerStatus == ViewerStatus.HIDE)
				pw.print(" viewerStatus='HIDE'");
		}
		if (((_template == null) && (_viewerX != getIntValue("MAP.viewerX", -1)))
				|| ((_template != null) && (_viewerX != _template._viewerX)))
			pw.print(" viewerX='" + _viewerX + "'");
		if (((_template == null) && (_viewerY != getIntValue("MAP.viewerY", -1)))
				|| ((_template != null) && (_viewerY != _template._viewerY)))
			pw.print(" viewerY='" + _viewerY + "'");*/
		if (((_template == null) && !_zoom)
				|| ((_template != null) && (_zoom != _template.isZoom())))
			pw.print(" zoom='" + _zoom + "'");
		if ((_template == null)
				|| (_zoomStatus != _template.getZoomStatus()))
		{
			if ((_zoomStatus == ZoomStatus.SHOW) && (_template != null))
				pw.print(" zoomStatus='SHOW'");
			else if (_zoomStatus == ZoomStatus.HIDE)
				pw.print(" zoomStatus='HIDE'");
		}
		if (((_template == null) && (_zoomStep != getFloatValue("MAP.zoomStep", 0F)))
				|| ((_template != null) && (_zoomStep != _template.getZoomStep())))
			pw.print(" zoomStep='" + _zoomStep + "'");
		if (((_template == null) && (_zoomFactor != getFloatValue("MAP.zoomFactor", 0F)))
				|| ((_template != null) && (_zoomFactor != _template.getZoomFactor())))
			pw.print(" zoomFactor='" + _zoomFactor + "'");
	/*	if (((_template == null) && !_translation)
				|| ((_template != null) && (_translation != _template._translation)))
			pw.print(" translation='" + _translation + "'");
		if (((_template == null) && (_translationStep != getIntValue("MAP.translationStep", -1)))
				|| ((_template != null) && (_translationStep != _template._translationStep)))
			pw.print(" translationStep='" + _translationStep + "'");
		if ((_template == null)
				|| (_translationStatus != _template._translationStatus))
		{
			if ((_translationStatus == TranslationStatus.SHOW) && (_template != null))
				pw.print(" translationStatus='SHOW'");
			else if (_translationStatus == TranslationStatus.HIDE)
				pw.print(" translationStatus='HIDE'");
		}
		if (((_template == null) && _spatial)
				|| ((_template != null) && (_spatial != _template._spatial)))
			pw.print(" spatial='" + _spatial + "'");
		if (((_template == null) && (_background != null))
				|| ((_template != null) && (_background != _template._background)))
			pw.print(" background='" + LyString.getXmlStringStrict(_background) + "'");
		if (((_template == null) && (_titleFont != getStringValue("MAP.titleFont", "$ALL.titleFont")))
				|| ((_template != null) && (_titleFont != _template._titleFont)))
			pw.print(" titleFont='" + LyString.getXmlStringStrict(_titleFont) + "'");
		if (((_template == null) && (_color != getStringValue("MAP.color", "$ALL.color")))
				|| ((_template != null) && (_color != _template._color)))
			pw.print(" color='" + LyString.getXmlStringStrict(_color) + "'");
		if (((_template == null) && (_fontColor != getStringValue("MAP.fontColor", "$ALL.fontColor")))
				|| ((_template != null) && (_fontColor != _template._fontColor)))
			pw.print(" fontColor='" + LyString.getXmlStringStrict(_fontColor) + "'");
		if (((_template == null) && (_selectedColor != getStringValue("MAP.selectedColor", "$ALL.selectedColor")))
				|| ((_template != null) && (_selectedColor != _template._selectedColor)))
			pw.print(" selectedColor='" + LyString.getXmlStringStrict(_selectedColor) + "'");
		if (((_template == null) && (_outlineColor != getStringValue("MAP.outlineColor", "$ALL.outlineColor")))
				|| ((_template != null) && (_outlineColor != _template._outlineColor)))
			pw.print(" outlineColor='" + LyString.getXmlStringStrict(_outlineColor) + "'");
		if (((_template == null) && (_width != getIntValue("MAP.width", -1)))
				|| ((_template != null) && (_width != _template._width)))
			pw.print(" width='" + _width + "'");
		if (((_template == null) && (_height != getIntValue("MAP.height", -1)))
				|| ((_template != null) && (_height != _template._height)))
			pw.print(" height='" + _height + "'");*/
		if ((_template == null)
				|| (_style != _template.getStyle()))
		{
			if ((_style == Style.STANDARD) && (_template != null))
				pw.print(" style='STANDARD'");
			else if (_style == Style.HYBRID)
				pw.print(" style='HYBRID'");
			else if (_style == Style.ROADMAP)
				pw.print(" style='ROADMAP'");
			else if (_style == Style.SATELLITE)
				pw.print(" style='SATELLITE'");
			else if (_style == Style.TERRAIN)
				pw.print(" style='TERRAIN'");
		}
		if (((_template == null) && (_latitude != null))
				|| ((_template != null) && (_latitude != _template.getLatitude())))
			pw.print(" latitude='" + LyString.getXmlStringStrict(_latitude) + "'");
		if (((_template == null) && (_longitude != null))
				|| ((_template != null) && (_longitude != _template.getLongitude())))
			pw.print(" longitude='" + LyString.getXmlStringStrict(_longitude) + "'");

		if ((_map.getChildCount() == 0) && (_map.getDatas() == null))
		{
			pw.println("/>");
			return;
		}
		else
			pw.println(">");

		// Write sub-components
		LyComponent	components;
		int			i, n;

		n = (_components == null ? 0 : _components.size());
		for (i = 0; i < n; i++)
		{
			components = _components.get(i);
			if(components instanceof LyMapNode)
			{
				((LyMobileMapNode)components.getInternalView()).writeXml(pw);
			}
			
		}

		/*if (_mapContent != null)
			_mapContent.writeXml(pw, debug);
		if (_layout != null)
			_layout.writeXml(pw, debug);
		if (_constraint != null)
			_constraint.writeXml(pw, debug);*/

		//writeDatas(pw);

		pw.println("</MAP>");

	}
	
	public void writeJson(PrintWriter pw, String sessionId)
	{
		init();
		if (_map.getStyle() != LyMap.Style.STANDARD)
		{
			boolean isFirstElement = true;
			
			pw.println("{\"MAP\": [");
			
			if (_id != null)
			{
				isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
				pw.print("{\"id\": \"" + LyString.getJsonString(_id) + "\"}");
			}
				
			if (((_template == null) && (_label != null))
					|| ((_template != null) && (_label != _template.getLabel())))
			{
				isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
				pw.print("{\"label\": \"" + LyString.getJsonString(_label) + "\"}");
			}

			if (_controller != null)
			{
				isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
				pw.print("{\"controller\": \"" + LyString.getJsonString(_controller) + "\"}");
			}
			if (((_template == null) && !_zoom)
					|| ((_template != null) && (_zoom != _template.isZoom())))
			{
				isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
				pw.print("{\"zoom\": \"" + _zoom + "\"}");
			}
			if ((_template == null)
					|| (_zoomStatus != _template.getZoomStatus()))
			{
				
				if ((_zoomStatus == ZoomStatus.SHOW) && (_template != null))
				{
					isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
					pw.print("{\"zoomStatus\": \"SHOW\"}");
				}
				else if (_zoomStatus == ZoomStatus.HIDE)
				{
					isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
					pw.print("{\"zoomStatus\": \"HIDE\"}");
				}
			}
			if (((_template == null) && (_zoomStep != getFloatValue("MAP.zoomStep", 0F)))
					|| ((_template != null) && (_zoomStep != _template.getZoomStep())))
			{
				isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
				pw.print("{\"zoomStep\": \"" + _zoomStep + "\"}");
			}
			if (((_template == null) && (_zoomFactor != getFloatValue("MAP.zoomFactor", 0F)))
					|| ((_template != null) && (_zoomFactor != _template.getZoomFactor())))
			{
				isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
				pw.print("{\"zoomFactor\": \"" + _zoomFactor + "\"}");
			}
			if ((_template == null)
					|| (_style != _template.getStyle()))
			{
				isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
				if ((_style == Style.STANDARD) && (_template != null))
					pw.print("{\"style\": \"STANDARD\"}");
				else if (_style == Style.HYBRID)
					pw.print("{\"style\": \"HYBRID\"}");
				else if (_style == Style.ROADMAP)
					pw.print("{\"style\": \"ROADMAP\"}");
				else if (_style == Style.SATELLITE)
					pw.print("{\"style\": \"SATELLITE\"}");
				else if (_style == Style.TERRAIN)
					pw.print("{\"style\": \"TERRAIN\"}");
			}
			if (((_template == null) && (_latitude != null))
					|| ((_template != null) && (_latitude != _template.getLatitude())))
			{
				isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
				pw.print("{\"latitude\": \"" + LyString.getJsonString(_latitude)  + "\"}");
			}
			if (((_template == null) && (_longitude != null))
					|| ((_template != null) && (_longitude != _template.getLongitude())))
			{
				isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
				pw.print("{\"longitude\": \"" + LyString.getJsonString(_longitude) + "\"}");
			}

			if ((_map.getChildCount() == 0) && (_map.getDatas() == null))
			{
				pw.println("]}");
				return;
			}

			// Write sub-components
			LyComponent	components;
			int			i, n;

			n = (_components == null ? 0 : _components.size());
			for (i = 0; i < n; i++)
			{
				isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
				components = _components.get(i);
				if(components instanceof LyMapNode)
				{
					((LyMobileMapNode)components.getInternalView()).writeJson(pw, sessionId);
				}
				
			}

			pw.println("]}");
		}
		else
		{
			createWebview(pw, sessionId);
		}
	}
	
	public void init()
	{
		_id = _map.getId();

		_template = _map.getTemplate();

		_label = _map.getLabel();

		_controller = _map.getController();

		_selectionMode = _map.getSelectionMode();

		_selectionDraw = _map.getSelectionDraw();

		_viewer = _map.isViewer();

		_viewerStatus = _map.getViewerStatus();

		_viewerX = _map.getViewerX();

		_viewerY = _map.getViewerY();

		_zoom = _map.isZoom();

		_zoomStatus = _map.getZoomStatus();

		_zoomStep = _map.getZoomStep();

		_zoomFactor = _map.getZoomFactor();

		_translation = _map.isTranslation();

		_translationStep = _map.getTranslationStep();

		_translationStatus = _map.getTranslationStatus();

		_spatial = _map.isSpatial();

		_background = _map.getBackground();

		_titleFont = _map.getTitleFont();

		_color = _map.getColor();

		_fontColor = _map.getFontColor();

		_selectedColor = _map.getSelectedColor();

		_outlineColor = _map.getOutlineColor();

		_width = _map.getWidth();

		_height = _map.getHeight();

		_style = _map.getStyle();

		_latitude = _map.getLatitude();

		_longitude = _map.getLongitude();

		_components.clear();

		for(int i = 0; i < _map.getComponentCount(); i++)
		{
			_components.add(_map.getComponent(i));
		}

		_mapContent = _map.getMapContent();

		_layout = _map.getLayout();

		_constraint = _map.getConstraint();

	}


	public void setLabel(String label)
	{
		// TODO Auto-generated method stub

	}


	public void setViewer(boolean viewer)
	{
		// TODO Auto-generated method stub

	}


	public void setViewerStatus(ViewerStatus viewerStatus)
	{
		// TODO Auto-generated method stub

	}


	public void setZoom(boolean zoom)
	{
		// TODO Auto-generated method stub

	}


	public void setZoomStatus(ZoomStatus zoomStatus)
	{
		// TODO Auto-generated method stub

	}


	public void setZoomFactor(float zoomFactor)
	{
		// TODO Auto-generated method stub

	}


	public void setTranslation(boolean translation)
	{
		// TODO Auto-generated method stub

	}


	public void setTranslationStatus(TranslationStatus translationStatus)
	{
		// TODO Auto-generated method stub

	}


	public void setSpatial(boolean spatial)
	{
		// TODO Auto-generated method stub

	}


	public void addComponent(LyComponent components, int pos)
	{
		// TODO Auto-generated method stub

	}


	public void removeComponent(LyComponent components)
	{
		// TODO Auto-generated method stub

	}


	public void setMapContent(LyMapContent mapContent)
	{
		// TODO Auto-generated method stub

	}


	public void setLayout(LyLayout layout)
	{
		// TODO Auto-generated method stub

	}

}
