package leon.view.mobile;

import java.io.*;

import leon.misc.*;
import leon.view.*;
import leon.view.LyMapNode.*;
import leon.view.graphic.*;

public class LyMobileMapNode extends LyMobileObject implements LyGraphicalMapNode
{
	/** The mapNode generic component. */
	protected LyMapNode			_mapNode;

	/**
	 * Identifier of the node.
	 */
	protected String			_id;

	/**
	 * Template of the component.
	 */
	protected LyMapNode			_template;

	/**
	 * Label of the node.
	 */
	protected String			_label;

	/**
	 * Abscissa of the node.
	 */
	protected float				_x;

	/**
	 * Ordinate of the node.
	 */
	protected float				_y;

	/**
	 * Altitude of the node (used only for a map in three dimensions)
	 */
	protected float				_z;


	/**
	 * Indicates if the node is hidden (not displayed) or not.
	 */
	protected boolean			_hidden;

	/**
	 * Image used for the node.
	 * The image is used only if the style of the node is IMAGE.
	 */
	protected String			_image;

	protected String			_rootUrl;

	/**
	 * Left label of the node.
	 */
	protected String			_leftLabel;

	/**
	 * Right label of the node.
	 */
	protected String			_rightLabel;

	/**
	 * Top label of the node.
	 */
	protected String			_topLabel;

	/**
	 * Bottom label of the node.
	 */
	protected String			_bottomLabel;

	/**
	 * Tooltip of the node.
	 */
	protected String			_tooltip;

	/**
	 * Overview label of the node.
	 */
	protected String			_overview;

	/**
	 * Parent node of the current node.
	 * The parent node is used by certain positioning managers to indicate a link from 
	 * a node to another.
	 */
	protected String			_parentNode;

	/**
	 * Style of the node. The following styles are managed:
	 * IMAGE: Node represented by an image
	 * CIRCLE: Node represented by a circle
	 * SQUARE: Node represented by a square
	 * TRIANGLE: Node represented by a triangle
	 */
	protected Style				_style;

	/**
	 * Style of the node in the viewer. The following styles are managed:
	 * IMAGE: Node represented by an image
	 * CIRCLE: Node represented by a circle
	 * SQUARE: Node represented by a square
	 * TRIANGLE: Node represented by a triangle
	 */
	protected Style				_viewerStyle;

	/**
	 * Alignment of the label of the node.
	 * This attribute determines the positioning of the label of the node in comparison
	 * with the image representing the node.
	 * It is used only if the label of the node is specified. The following alignments 
	 * are managed:
	 * LEFT: Label on the left of the image
	 * RIGHT: Label on the right of the image
	 * TOP: Label on the top of the image
	 * BOTTOM: Label on the bottom of the image
	 * CENTER: Label in the middle of the image
	 * EXTERN: Label outside the image in comparison with the center of the map
	 */
	protected LabelAlign			_labelAlign;

	/**
	 * Indicates whether the node is sensitive or not.
	 * A sensitive node can be selected and allows launching actions (by 
	 * double-clicking for example).
	 */
	protected boolean			_sensitive;

	/**
	 * Indicates whether the node can be moved or not.
	 */
	protected boolean			_movable;

	/**
	 * Indicates whether the node is selected or not.
	 * The value of this attribute is used only if the node is sensitive.
	 */
	protected boolean			_selected;

	/**
	 * Indicates whether the node can be resized or not when zooming the map.
	 */
	protected boolean			_resizable;
	/**
	 * Image used for the node in the viewer if the viewer style is IMAGE.
	 * The image is used only if the style of the node is IMAGE.
	 */
	protected String			_viewerImage;

	/**
	 * Image used as superimposed on the image of the node.
	 * It is displayed in the bottom right-hand corner of the image representing the 
	 * node.
	 */
	protected String			_smallImage;

	/**
	 * Size of the node.
	 * The expected value is a positive integer value defining the size of the node in 
	 * pixels (width and height).
	 * This value is used only if the style of the node is CIRCLE, SQUARE or TRIANGLE.
	 */
	protected int				_size;

	/**
	 * MAP_NODE.height
	 */
	protected int				_height;

	/**
	 * Layer of the node.
	 */
	protected int				_layer;

	/**
	 * Color of the internal frame of the node.
	 * If this attribute is specified, a frame of the indicated color is drawn around 
	 * the node.
	 */
	protected String			_colorFrameInt;

	/**
	 * Color of external frame of the node.
	 * If this attribute is specified, a frame of the indicated color is drawn around 
	 * the node, and around the internal frame if there is one.
	 */
	protected String			_colorFrameExt;

	/**
	 * Color of the cross drawn on the node.
	 * This cross is drawn only if this attribute has been specified.
	 */
	protected String			_colorCross;

	/**
	 * Positioning data.
	 * This data is used by certain positioning managers to set the node in a special 
	 * way. The possible values are the following: north, south, east, west, northeast,
	 * northwest, southeast and southwest: Values interpreted by the manager of LINEAR 
	 * type.
	 * It indicates the location of the node in question in comparison with its parent 
	 * node (in the case where a parent node has actually been specified).
	 */
	protected String			_layoutInfo;

	/**
	 * Indicates if the image is blinking or not.
	 */
	protected boolean			_blink;

	/**
	 * Indicates if the small image drawn over the image of the node is blinking or 
	 * not.
	 */
	protected boolean			_smallImageBlink;

	/**
	 * Java full name of the class that manages the map node.
	 */
	protected String			_javaName;

	/**
	 * Text font used on the component.
	 */
	protected String			_font;

	/**
	 * Color of the node.
	 * If the style of the node is CIRCLE, SQUARE or TRIANGLE the color is used to draw
	 * the node.
	 * If the style of the node is IMAGE the color is used to replace the black color 
	 * contained in the image of the node.
	 * This last possibility allows among others to use a same image for a node that 
	 * switches from a state to another (for a major warning state, the black color of 
	 * the image can be replace by the red color, for a minor state by the blue color, 
	 * etc.).
	 */
	protected String			_color;

	/**
	 * Text color used on the component.
	 */
	protected String			_fontColor;

	/**
	 * MAP_NODE.selectedColor
	 */
	protected String			_selectedColor;

	/**
	 * MAP_NODE.inactiveSelectedColor
	 */
	protected String			_inactiveSelectedColor;

	/**
	 * MAP_NODE.outlineColor
	 */
	protected String			_outlineColor;


	public LyMobileMapNode(LyMapNode mapNode)
	{
		super(mapNode);

		_mapNode = mapNode;
		_rootUrl = ((LyMobileViewManager) getViewManager()).getRealRootURL();
	}

	public void writeXml(PrintWriter pw)
	{
		// Write tag with its attributes
		init();
		pw.print("<MAP_NODE");
		if (_id != null)
			pw.print(" id='" + LyString.getXmlString(_id) + "'");
		if (_template != null)
			pw.print(" template='" + _template.getId() + "'");
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template.getLabel())))
			pw.print(" label='" + LyString.getXmlString(_label) + "'");
		
		if (((_template == null) && (_x != getFloatValue("MAP_NODE.x", 0F)))
				|| ((_template != null) && (_x != _template.getX())))
			pw.print(" x='" + _x + "'");
		if (((_template == null) && (_y != getFloatValue("MAP_NODE.y", 0F)))
				|| ((_template != null) && (_y != _template.getY())))
			pw.print(" y='" + _y + "'");
		if (((_template == null) && (_z != getFloatValue("MAP_NODE.z", 0F)))
				|| ((_template != null) && (_z != _template.getZ())))
			pw.print(" z='" + _z + "'");
		if (((_template == null) && _hidden)
				|| ((_template != null) && (_hidden != _template.isHidden())))
			pw.print(" hidden='" + _hidden + "'");
		if (((_template == null) && (_image != null))
				|| ((_template != null) && (_image != _template.getImage())))
			pw.print(" image='" + _rootUrl+ getImageUrl(_image) + "'");
		
		pw.println("/>");
	}
	
	public void writeJson(PrintWriter pw, String sessionId)
	{
		boolean isFirstElement = true;
		init();
		pw.println("{\"MAP_NODE\": [");
		
		if (_id != null)
		{
			isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
			pw.print("{\"id\": \"" + LyString.getJsonString(_id) + "\"}");
		}
			
		if (((_template == null) && (_label != null))
				|| ((_template != null) && (_label != _template.getLabel())))
		{
			isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
			pw.print("{\"label\": \"" + LyString.getJsonString(_label) + "\"}");
		}

		if (((_template == null) && (_x != getFloatValue("MAP_NODE.x", 0F)))
				|| ((_template != null) && (_x != _template.getX())))
		{
			isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
			pw.print("{\"x\": \"" + _x + "\"}");
		}
		if (((_template == null) && (_y != getFloatValue("MAP_NODE.y", 0F)))
				|| ((_template != null) && (_y != _template.getY())))
		{
			isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
			pw.print("{\"y\": \"" + _y + "\"}");
		}
		if (((_template == null) && (_z != getFloatValue("MAP_NODE.z", 0F)))
				|| ((_template != null) && (_z != _template.getZ())))
		{
			isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
			pw.print("{\"z\": \"" + _z + "\"}");
		}
		if (((_template == null) && _hidden)
				|| ((_template != null) && (_hidden != _template.isHidden())))
		{
			isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);
			pw.print("{\"hidden\": \"" + _hidden + "\"}");
		}
		if (((_template == null) && (_image != null))
				|| ((_template != null) && (_image != _template.getImage())))
		{
			String imageUrl = getImageUrl(_image);
			isFirstElement = printJsonAttributeSeparator(isFirstElement, pw);	
			pw.print("{\"image\": \"" + 
					(imageUrl.startsWith(_rootUrl) ? imageUrl : _rootUrl+imageUrl)  + "\"}");
		}

		pw.println("]}");
	}

	public void init()
	{
		_id = _mapNode.getId();

		_template = _mapNode.getTemplate();

		_label = _mapNode.getLabel();

		_x = _mapNode.getX();

		_y = _mapNode.getY();

		_z = _mapNode.getZ();

		_hidden = _mapNode.isHidden();

		_image = _mapNode.getImage();
		

		_leftLabel = _mapNode.getLeftLabel();

		_rightLabel = _mapNode.getRightLabel();

		_topLabel = _mapNode.getTopLabel();

		_bottomLabel = _mapNode.getBottomLabel();

		_tooltip = _mapNode.getTooltip();

		_overview = _mapNode.getOverview();

		_parentNode = _mapNode.getParentNode();

		_style = _mapNode.getStyle();

		_viewerStyle = _mapNode.getViewerStyle();

		_labelAlign = _mapNode.getLabelAlign();

		_sensitive = _mapNode.isSensitive();

		_movable = _mapNode.isMovable();

		_selected = _mapNode.isSelected();

		_resizable = _mapNode.isResizable();
		
		_viewerImage = _mapNode.getViewerImage();

		_smallImage = _mapNode.getSmallImage();

		_size = _mapNode.getSize();

		_height = _mapNode.getHeight();

		_layer = _mapNode.getLayer();

		_colorFrameInt = _mapNode.getColorFrameInt();

		_colorFrameExt = _mapNode.getColorFrameExt();

		_colorCross = _mapNode.getColorCross();

		_layoutInfo = _mapNode.getLayoutInfo();

		_blink = _mapNode.isBlink();

		_smallImageBlink = _mapNode.isSmallImageBlink();

		_javaName = _mapNode.getJavaName();

		_font = _mapNode.getFont();

	}

	public void setLabel(String label)
	{
		// TODO Auto-generated method stub
		
	}

	public void setLeftLabel(String leftLabel)
	{
		// TODO Auto-generated method stub
		
	}

	public void setRightLabel(String rightLabel)
	{
		// TODO Auto-generated method stub
		
	}

	public void setTopLabel(String topLabel)
	{
		// TODO Auto-generated method stub
		
	}

	public void setBottomLabel(String bottomLabel)
	{
		// TODO Auto-generated method stub
		
	}

	public void setX(float x)
	{
		// TODO Auto-generated method stub
		
	}

	public void setY(float y)
	{
		// TODO Auto-generated method stub
		
	}

	public void setZ(float z)
	{
		// TODO Auto-generated method stub
		
	}
	
	public void setLayer(int layer)
	{
		// TODO Auto-generated method stub
		
	}

	public void setSelected(boolean selected)
	{
		if (_mapNode.isSensitive() && (_mapNode.isSelected() != selected))
            _mapNode.setSelected(selected);
		
	}

	public void setHidden(boolean hidden)
	{
		// TODO Auto-generated method stub
		
	}

	public void setImage(String image)
	{
		// TODO Auto-generated method stub
		
	}

	public void setViewerImage(String viewerImage)
	{
		// TODO Auto-generated method stub
		
	}

	public void setColorFrameInt(String colorFrameInt)
	{
		// TODO Auto-generated method stub
		
	}

	public void setColorFrameExt(String colorFrameExt)
	{
		// TODO Auto-generated method stub
		
	}

	public void setFontColor(String fontColor)
	{
		// TODO Auto-generated method stub
		
	}

}
