/*
 * LyFileAttachmentHandler.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web;

import java.io.*;
import java.security.*;
import java.util.*;
import javax.activation.*;
import javax.mail.*;
import javax.mail.internet.*;
import javax.servlet.*;
import javax.servlet.http.*;

import org.apache.commons.fileupload.util.Streams;

import org.apache.commons.fileupload.*;
import org.apache.commons.fileupload.servlet.*;

import leon.misc.*;
import leon.view.setinterface.*;


/**
 * This class handles file attachments in doPost method. Main services are :
 *
 * <UL>
 *   <LI>Extract the raw data from the file submit</LI>
 *   <LI>Retreive the param part session as a valid HttpServletRequest for generic processing</LI>
 * </UL>
 *
 * @version $Id: LyFileAttachmentHandler.java 41816 2012-12-19 15:18:30Z smaupin $
 * @author  Lyria
 */
public class LyFileAttachmentHandler
	implements LyFreeInterface
{
	public static final String			 __VERSION =
		"$Id: LyFileAttachmentHandler.java 41816 2012-12-19 15:18:30Z smaupin $";

	/** Current environment. */
	protected LyEnvironment				 _environment;

	/** Request from client (POST) which contains attached document. */
	protected HttpServletRequest		 _request;

	/** Internal request wrapper (cached for internal use only). */
	protected _LyMultipartRequestWrapper _internalRequest;

	/** Server directory to store files. */
	protected String					 _attachmentDirectory;

	/**
	 * Constructor.
	 */
	public LyFileAttachmentHandler()
	{
		_environment = null;
		_request = null;
		_attachmentDirectory = null;
	}

	/**
	 * Gets the environment.
	 *
	 * @param environment Current environment
	 */
	public void setEnvironment(LyEnvironment environment)
	{
		_environment = environment;
	}

	/**
	 * Gets the request.
	 *
	 * @param request current user Http request
	 */
	public void setRequest(HttpServletRequest request)
	{
		_request = request;
	}

	/**
	 * Parses the content of the request.<br>
	 * Result of this method invocation is :
	 *
	 * <UL>
	 *   <LI>Retreive all parameters / values which are now accessible</LI>
	 *   <LI>Use a LyFilePartWriter instance for each file to store attachements</LI>
	 * </ul>
	 *
	 * @throws IOException
	 * @throws FileUploadException
	 */
	public void parseContent()
		throws IOException, FileUploadException
	{
		getParameterPartRequest(); // Call factory (get ready)
		if (_internalRequest._request == null)
			_internalRequest._request  = _request;
		_internalRequest.parseContent();
	}

	/**
	 * Gets the parameter part of the request.<br>
	 * Result is wrapped into a HttpServletRequest as if there was only one part.
	 *
	 * @return the http servlet request corresponding to the parameter part of the request
	 * @throws IOException
	 */
	public HttpServletRequest getParameterPartRequest()
		throws IOException
	{
		if (_internalRequest == null)
		{
			_internalRequest = new _LyMultipartRequestWrapper(_environment, _request); // Use
			// wrapper.
			_internalRequest.setFilePartWriter(filePartWriterFactory());
		}

		return _internalRequest;
	}

	/**
	 * Factory to create File attachment handler instance. This may be overriden to specify server
	 * location, and ways to save/retreive the attachment.
	 *
	 * @return a LyFilePartWriter instance.
	 * @throws IOException
	 */
	public LyFilePartWriter filePartWriterFactory()
		throws IOException
	{
		LyFilePartWriter fp = new LyFilePartWriter();

		fp.setAttachmentDirectory(getAttachmentDirectory());
		return fp;
	}

	/**
	 * Gets the LyFilePartWriter instance to manage file attachments.
	 *
	 * @return current file management class instance.
	 */
	public LyFilePartWriter getFilePartWriter()
	{
		return _internalRequest.getFilePartWriter();
	}

	/**
	 * Gets the supported encoding type.
	 *
	 * @return the supported encoding type
	 */
	public String getEncodingType()
	{
		return LyWebConstants.CONTENT_MULTIPART;
	}

	/**
	 * Gets the attachment directory from meta-model.
	 *
	 * @return the attachment directory from meta-model
	 * @see    #setAttachmentDirectory(String)
	 */
	public String getAttachmentDirectory()
	{
		return _attachmentDirectory;
	}

	/**
	 * Sets the attachment directory.
	 *
	 * @param attachmentDirectory the new attachment directory
	 * @see   #getAttachmentDirectory()
	 */
	public void setAttachmentDirectory(String attachmentDirectory)
	{
		_attachmentDirectory = attachmentDirectory;
	}

	/**
	 * Clean up everything.
	 */
	public void free()
	{
		_environment = null;
		_request = null;
		_attachmentDirectory = null;

		if (_internalRequest != null)
		{
			_internalRequest.free();
			_internalRequest = null;
		}
	}
} // end class LyFileAttachmentHandler



/**
 * Extracts the parameter part from the stream and return it as a 'non Multipart'
 * HttpServletRequest.
 *
 * @version $Id: LyFileAttachmentHandler.java 41816 2012-12-19 15:18:30Z smaupin $
 * @author  Lyria
 */
class _LyMultipartRequestWrapper
	implements HttpServletRequest,
		LyFreeInterface
{
	public static final String		    __VERSION =
		"$Id: LyFileAttachmentHandler.java 41816 2012-12-19 15:18:30Z smaupin $";

	/** Current environment. */
	protected LyEnvironment			    _environment;

	/** Request from client (POST) which contains attached document. */
	protected HttpServletRequest	    _request;

	/** Parameters tables (stored when parsing multipart). */
	protected Hashtable<String, Object> _parameters;

	/** LyFilePartWriter is cached for internal use only. */
	protected LyFilePartWriter		    _filePartWriter;

	/**
	 * Constructor.
	 *
	 * @param  environment current environment receinving the file
	 * @param  request     current user Http request
	 * @throws IOException
	 */
	public _LyMultipartRequestWrapper(LyEnvironment environment, HttpServletRequest request)
		throws IOException
	{
		_request = request;
		_parameters = new Hashtable<String, Object>();
		_environment = environment;
	}

    public void parseContentStreaming() throws IOException, FileUploadException
    {
        // Create a new file upload handler
        ServletFileUpload upload = new ServletFileUpload();
        // Parse the request
        FileItemIterator iter = upload.getItemIterator(_request);

        String EXTENSION = "File";

        while (iter.hasNext())
        {
            FileItemStream item = iter.next();
            InputStream stream = item.openStream();
            String fileName = item.getName();
            String paramName = item.getFieldName();
            if (!item.isFormField())
            {
                LyFilePartWriter writer = getFilePartWriter();

                if (fileName != null)
                    fileName = new String(fileName.getBytes(), LyString.DEFAULT_CHARSET);

                if ((fileName != null) && (fileName.equals("") || fileName.equals("#")))
                {
                    // Leave BLANK : do not save
                    _parameters.put(item.getFieldName(), fileName);
                }
                else
                {
                    // Set the information we have
                    writer.setEnvironment(_environment);
                    writer.setClientFileName(fileName);
                    writer.setContentType(item.getContentType());

                    // Save file on server side (server location is
                    // computed)
                    writer.saveFile(stream);

                    // Save the name of the file on server side.
                    // XXX May have ressources here to tell what to do !
                    if (paramName.endsWith(EXTENSION))
                    {
                        String fieldParam = paramName.substring(0, paramName.length() - EXTENSION.length());
                        String tmpDir = _environment.getViewManager().getTemporaryDir();

                        tmpDir = _environment.getString().expand(tmpDir);

                        if (!writer.getServerFileName().startsWith(tmpDir))

                            // Override fieldParam (we loose what the user typed
                            // !!)
                            // XXX could do better but it would require 2 fields
                            // in database
                            // or a special separator-based encoding.

                            _parameters.put(fieldParam, writer.getFileLocation());
                        else
                            // blob or clob : we can keep original name (no risk
                            // of conflict in db)
                            _parameters.put(fieldParam, writer.getServerFileName());
                    }

                    _parameters.put(paramName, writer.getServerFileName());
                }

            }
            else
            {
                _parameters.put(item.getFieldName(), Streams.asString(stream));
            }
        }
    }

	/**
	 * Parses the content of the request.<br>
	 * Result of this method invocation is :
	 *
	 * <UL>
	 *   <LI>Retreive all parameters / values which are now accessible</LI>
	 *   <LI>Use a LyFilePartWriter instance for each file to store attachements</LI>
	 * </UL>
	 *
	 * @throws IOException
	 * @throws FileUploadException
	 */
	public void parseContent()
		throws IOException, FileUploadException
	{
		DataSource datasource = dataSourceFactory(_request.getRemoteAddr(),
				_request.getContentType(), _request.getInputStream(), _request.getContentLength());

		boolean isMultipart = ServletFileUpload.isMultipartContent(_request);
		if (isMultipart)
		{
		    parseContentStreaming();
		    return;
		}


		try
		{
			// First of all, retreive original request parameters
			Enumeration<?> e = _request.getParameterNames();
			String	    param;
			String[]    value;

			if (e != null)
			{
				while (e.hasMoreElements())
				{
					param = (String)e.nextElement();
					value = _request.getParameterValues(param);
					if (value != null)
						_parameters.put(param, value);
				}
			}

			MimeMultipart    mimeReader = new MimeMultipart(datasource);
			BodyPart		 bodyPart = null;
			// Seek for parameter part ...
			// Order is not forced.
			int				 i, partCount = mimeReader.getCount();
			String[]		 vals;
			String			 paramName, content, contentType, fileName, fieldParam;
			LyFilePartWriter writer;
			String			 EXTENSION = "File";
			String			 contentDisposition = null;

			for (i = 0; i < partCount; ++i)
			{
				bodyPart = mimeReader.getBodyPart(i);
				vals = bodyPart.getHeader("Content-Disposition");
				paramName = null;

				if ((vals != null) && (vals.length > 0))
				{
					contentDisposition = vals[0];
					paramName = extractHeaderParameter(contentDisposition, "name");
				}

				// Is it a file part or a parameter part ?
				if (bodyPart.getFileName() == null)
				{
					if (paramName != null)
					{
						// Parameters
						content = bodyPart.getContent().toString();
						_parameters.put(paramName, content);
					}
				}
				else
				{
					// File content...
					contentType = bodyPart.getContentType();
					fileName = null;
					if (contentDisposition != null)
						fileName = extractHeaderParameter(contentDisposition, "filename");
					if (fileName == null)
						fileName = bodyPart.getFileName();

					// File name may not be correctly encoded
					if (fileName != null)
						fileName = new String(fileName.getBytes(), LyString.DEFAULT_CHARSET);

					if ((fileName != null) && (fileName.equals("") || fileName.equals("#")))
						// Leave BLANK : do not save
						_parameters.put(paramName, fileName);
					else
					{
						writer = getFilePartWriter();

						// Set the information we have
						writer.setEnvironment(_environment);
						writer.setClientFileName(fileName);
						writer.setContentType(contentType);

						// Save file on server side (server location is
						// computed)
						writer.saveFile(bodyPart.getInputStream());

						// Save the name of the file on server side.
						// XXX May have ressources here to tell what to do !
						if (paramName.endsWith(EXTENSION))
						{
							fieldParam = paramName.substring(0,
									paramName.length()
									- EXTENSION.length());
							String tmpDir = _environment.getViewManager().getTemporaryDir();

							tmpDir = _environment.getString().expand(tmpDir);

							if (!writer.getServerFileName().startsWith(tmpDir))

								// Override fieldParam (we loose what the user typed
								// !!)
								// XXX could do better but it would require 2 fields
								// in database
								// or a special separator-based encoding.

								_parameters.put(fieldParam, writer.getFileLocation());
							else
								//   blob or clob : we can keep original name (no risk of conflict in db)
								_parameters.put(fieldParam, writer.getServerFileName());
						}

						_parameters.put(paramName, writer.getServerFileName());
					}
				}
			}

			if (datasource instanceof LyFreeInterface)
				((LyFreeInterface)datasource).free();
		}
		catch (MessagingException e)
		{
			_environment.writeException(e);
		}
	} // end method parseContent

	/**
	 * Extract a header parameter name from the specified String like :<br>
	 * <code>'name'</code> from the string <code>Content-Disposition: form-data;
	 * name="actscript_fileFile"</code>.
	 *
	 * @param  headerLine the header line to parse
	 * @param  param      the param to find
	 * @return the value of the found parameter, empty String if not found
	 */
	public String extractHeaderParameter(String headerLine, String param)
	{
		// Get the name of the parameter
		String value = null;
		int    k, j = headerLine.indexOf(";");

		if (j >= 0)
		{
			j = headerLine.indexOf(param, j);
			if (j >= 0)
			{
				j = headerLine.indexOf("=", j);
				if (j >= 0)
				{
					k = headerLine.indexOf(";", j);
					if (k < 0)
						value = headerLine.substring(j + 1);
					else
						value = headerLine.substring(j + 1, k);

					value = value.replace('\"', ' ').trim();
				}
			}
		}

		return value;
	}

	/**
	 * Factory to wrap the input stream of the request to a parsable structure for the MimeDecoder.
	 *
	 * @param  name        data source name
	 * @param  contentType data MIME type
	 * @param  iStream     the used input stream
	 * @param  length      length of the data source
	 * @return a data source
	 * @see    _LyMultipartDataSource#_LyMultipartDataSource(String, String, InputStream, int)
	 */
	public DataSource dataSourceFactory(String name, String contentType, InputStream iStream,
		int length)
	{
		return new _LyMultipartDataSource(name, contentType, iStream, length);
	}

	/**
	 * Gets the created File attachment handler instance.
	 *
	 * @return a LyFilePartWriter instance.
	 * @see    #setFilePartWriter(LyFilePartWriter)
	 */
	public LyFilePartWriter getFilePartWriter()
	{
		return _filePartWriter;
	}

	/**
	 * Sets the File attachment handler instance.
	 *
	 * @param filePartWriter the new LyFilePartWriter instance.
	 * @see   #getFilePartWriter()
	 */
	public void setFilePartWriter(LyFilePartWriter filePartWriter)
	{
		_filePartWriter = filePartWriter;
	}

	/**
	 * Clean up everything. Keep parameters and internal request for later use.
	 */
	public void free()
	{
		_environment = null;

		if (_filePartWriter != null)
		{
			_filePartWriter.free();
			_filePartWriter = null;
		}

	}

	/**
	 * Returns the name of the authentication scheme used to protect the servlet, for example,
	 * "BASIC" or "SSL," or null if the servlet was not protected. Same as the value of the CGI
	 * variable AUTH_TYPE.
	 *
	 * @return a String specifying the name of the authentication scheme, or null if the request was
	 *         not authenticated
	 * @see    javax.servlet.http.HttpServletRequest#getAuthType()
	 */
	public String getAuthType()
	{
		return _request.getAuthType();
	}

	/**
	 * Returns the value of the specified request header as a long value that represents a Date
	 * object. Use this method with headers that contain dates, such as If-Modified-Since. The date
	 * is returned as the number of milliseconds since January 1, 1970 GMT. The header name is case
	 * insensitive. If the request did not have a header of the specified name, this method returns
	 * -1. If the header can't be converted to a date, the method throws an
	 * IllegalArgumentException.
	 *
	 * @param  name a String specifying the name of the header
	 * @return a long value representing the date specified in the header expressed as the number of
	 *         milliseconds since January 1, 1970 GMT, or -1 if the named header was not included
	 *         with the reqest
	 * @throws java.lang.IllegalArgumentException - If the header value can't be converted to a date
	 * @see    javax.servlet.http.HttpServletRequest#getDateHeader(java.lang.String)
	 */
	public long getDateHeader(String name)
	{
		return _request.getDateHeader(name);
	}

	/**
	 * Returns the value of the specified request header as a String. If the request did not include
	 * a header of the specified name, this method returns null. The header name is case
	 * insensitive. You can use this method with any request header.
	 *
	 * @param  name - a String specifying the header name
	 * @return a String containing the value of the requested header, or null if the request does
	 *         not have a header of that name
	 * @see    javax.servlet.http.HttpServletRequest#getHeader(java.lang.String)
	 */
	public String getHeader(String name)
	{
		return _request.getHeader(name);
	}

	/**
	 * Gets an array containing all of the Cookie objects the client sent with this request. This
	 * method returns null if no cookies were sent.
	 *
	 * @return an array of all the Cookies included with this request, or null if the request has no
	 *         cookies
	 * @see    javax.servlet.http.HttpServletRequest#getCookies()
	 */
	public Cookie[] getCookies()
	{
		return _request.getCookies();
	}

	/**
	 * Returns the portion of the request URI that indicates the context of the request. The context
	 * path always comes first in a request URI. The path starts with a "/" character but does not
	 * end with a "/" character. For servlets in the default (root) context, this method returns "".
	 *
	 * @return a String specifying the portion of the request URI that indicates the context of the
	 *         request
	 * @see    javax.servlet.http.HttpServletRequest#getContextPath()
	 */
	public String getContextPath()
	{
		return _request.getContextPath();
	}

	/**
	 * Gets an enumeration of all the header names this request contains. If the request has no
	 * headers, this method returns an empty enumeration. Some servlet containers do not allow do
	 * not allow servlets to access headers using this method, in which case this method returns
	 * null
	 *
	 * @return an enumeration of all the header names sent with this request; if the request has no
	 *         headers, an empty enumeration; if the servlet container does not allow servlets to
	 *         use this method, null
	 * @see    javax.servlet.http.HttpServletRequest#getHeaderNames()
	 */
	public Enumeration<?> getHeaderNames()
	{
		return _request.getHeaderNames();
	}

	/**
	 * Gets all the values of the specified request header as an Enumeration of String objects. Some
	 * headers, such as Accept-Language can be sent by clients as several headers each with a
	 * different value rather than sending the header as a comma separated list. If the request did
	 * not include any headers of the specified name, this method returns an empty Enumeration. The
	 * header name is case insensitive. You can use this method with any request header.
	 *
	 * @param  name a String specifying the header name
	 * @return a Enumeration containing the values of the requested header, or null if the request
	 *         does not have any headers of that name
	 * @see    javax.servlet.http.HttpServletRequest#getHeaders(java.lang.String)
	 */
	public Enumeration<?> getHeaders(String name)
	{
		return _request.getHeaders(name);
	}

	/**
	 * Returns the value of the specified request header as an int. If the request does not have a
	 * header of the specified name, this method returns 1. If the header cannot be converted to an
	 * integer, this method throws a NumberFormatException. The header name is case insensitive.
	 *
	 * @param  name a String specifying the name of a request header
	 * @return an integer expressing the value of the request header or -1 if the request doesn't
	 *         have a header of this name
	 * @throws java.lang.NumberFormatException If the header value can't be converted to an int
	 * @see    javax.servlet.http.HttpServletRequest#getIntHeader(java.lang.String)
	 */
	public int getIntHeader(String name)
	{
		return _request.getIntHeader(name);
	}

	/**
	 * Returns any extra path information associated with the URL the client sent when it made this
	 * request. The extra path information follows the servlet path but precedes the query string.
	 * This method returns null if there was no extra path information. Same as the value of the CGI
	 * variable PATH_INFO.
	 *
	 * @return a String specifying extra path information that comes after the servlet path but
	 *         before the query string in the request URL; or null if the URL does not have any
	 *         extra path information
	 * @see    javax.servlet.http.HttpServletRequest#getPathInfo()
	 */
	public String getPathInfo()
	{
		return _request.getPathInfo();
	}

	/**
	 * Returns the name of the HTTP method with which this request was made, for example, GET, POST,
	 * or PUT. Same as the value of the CGI variable REQUEST_METHOD.
	 *
	 * @return a String specifying the name of the method with which this request was made
	 * @see    javax.servlet.http.HttpServletRequest#getMethod()
	 */
	public String getMethod()
	{
		return _request.getMethod();
	}

	/**
	 * Returns any extra path information after the servlet name but before the query string, and
	 * translates it to a real path. Same as the value of the CGI variable PATH_TRANSLATED. If the
	 * URL does not have any extra path information, this method returns null.
	 *
	 * @return a String specifying the real path, or null if the URL does not have any extra path
	 *         information
	 * @see    javax.servlet.http.HttpServletRequest#getPathTranslated()
	 */
	public String getPathTranslated()
	{
		return _request.getPathTranslated();
	}

	/**
	 * Returns the login of the user making this request, if the user has been authenticated, or
	 * null if the user has not been authenticated. Whether the user name is sent with each
	 * subsequent request depends on the browser and type of authentication. Same as the value of
	 * the CGI variable REMOTE_USER.
	 *
	 * @return a String specifying the login of the user making this request, or null
	 * @see    javax.servlet.http.HttpServletRequest#getRemoteUser()
	 */
	public String getRemoteUser()
	{
		return _request.getRemoteUser();
	}

	/**
	 * Returns the query string that is contained in the request URL after the path. This method
	 * returns null if the URL does not have a query string. Same as the value of the CGI variable
	 * QUERY_STRING.
	 *
	 * @return a String containing the query string or null if the URL contains no query string
	 * @see    javax.servlet.http.HttpServletRequest#getQueryString()
	 */
	public String getQueryString()
	{
		return _request.getQueryString();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.servlet.ServletRequest#getRequestDispatcher(java.lang.String)
	 */
	public RequestDispatcher getRequestDispatcher(String path)
	{
		return _request.getRequestDispatcher(path);
	}

	/**
	 * Returns the part of this request's URL from the protocol name up to the query string in the
	 * first line of the HTTP request. For example:
	 *
	 * <table>
	 *   <tr>
	 *     <td><b>First line of HTTP request</b></td>
	 *     <td><b>Returned Value</b></td>
	 *   </tr>
	 *   <tr>
	 *     <td><code>POST /some/path.html HTTP/1.1</code></td>
	 *     <td><code>/some/path.html</code></td>
	 *   </tr>
	 *   <tr>
	 *     <td><code>GET http://foo.bar/a.html HTTP/1.0</code></td>
	 *     <td><code>http://foo.bar/a.html</code></td>
	 *   </tr>
	 *   <tr>
	 *     <td><code>HEAD /xyz?a=b HTTP/1.1</code></td>
	 *     <td><code>/xyz</code></td>
	 *   </tr>
	 * </table>
	 *
	 * <table>
	 *   <tbody>
	 *     <tr>
	 *       <td>To reconstruct an URL with a scheme and host, use
	 *         HttpUtils.getRequestURL(javax.servlet.http.HttpServletRequest).</td>
	 *     </tr>
	 *   </tbody>
	 * </table>
	 *
	 * @return a String containing the part of the URL from the protocol name up to the query string
	 * @see    javax.servlet.http.HttpUtils#getRequestURL(javax.servlet.http.HttpServletRequest)
	 * @see    javax.servlet.http.HttpServletRequest#getRequestURI()
	 */
	public String getRequestURI()
	{
		return _request.getRequestURI();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.servlet.http.HttpServletRequest#getRequestURL()
	 */
	public StringBuffer getRequestURL()
	{
		throw new IllegalArgumentException("BF AppEngine : Extension in J2EE not*yet* activated");
		// XXX uncomment below
		// return _request.getRequestURL();
	}

	/**
	 * Returns the session ID specified by the client. This may not be the same as the ID of the
	 * actual session in use. For example, if the request specified an old (expired) session ID and
	 * the server has started a new session, this method gets a new session with a new ID. If the
	 * request did not specify a session ID, this method returns null.
	 *
	 * @return a String specifying the session ID, or null if the request did not specify a session
	 *         ID
	 * @see    #isRequestedSessionIdValid()
	 * @see    javax.servlet.http.HttpServletRequest#getRequestedSessionId()
	 */
	public String getRequestedSessionId()
	{
		return _request.getRequestedSessionId();
	}

	/**
	 * Returns the part of this request's URL that calls the servlet. This includes either the
	 * servlet name or a path to the servlet, but does not include any extra path information or a
	 * query string. Same as the value of the CGI variable SCRIPT_NAME.
	 *
	 * @return a String containing the name or path of the servlet being called, as specified in the
	 *         request URL
	 * @see    javax.servlet.http.HttpServletRequest#getServletPath()
	 */
	public String getServletPath()
	{
		return _request.getServletPath();
	}

	/**
	 * Returns the current session associated with this request, or if the request does not have a
	 * session, creates one.
	 *
	 * @return the HttpSession associated with this request
	 * @see    #getSession(boolean)
	 * @see    javax.servlet.http.HttpServletRequest#getSession()
	 */
	public HttpSession getSession()
	{
		return _request.getSession();
	}

	/**
	 * Returns the current HttpSession associated with this request or, if if there is no current
	 * session and create is true, returns a new session.<br>
	 * If create is false and the request has no valid HttpSession, this method returns null.<br>
	 * To make sure the session is properly maintained, you must call this method before the
	 * response is committed.
	 *
	 * @param  create true to create a new session for this request if necessary; false to return
	 *                null if there's no current session
	 * @return the HttpSession associated with this request or null if create is false and the
	 *         request has no valid session
	 * @see    #getSession()
	 * @see    javax.servlet.http.HttpServletRequest#getSession(boolean)
	 */
	public HttpSession getSession(boolean create)
	{
		return _request.getSession(create);
	}

	/**
	 * Checks whether the requested session ID came in as a cookie.
	 *
	 * @return true if the session ID came in as a cookie; otherwise, false
	 * @see    #getSession(boolean)
	 * @see    javax.servlet.http.HttpServletRequest#isRequestedSessionIdFromCookie()
	 */
	public boolean isRequestedSessionIdFromCookie()
	{
		return _request.isRequestedSessionIdFromCookie();
	}

	/**
	 * Checks whether the requested session ID came in as part of the request URL.
	 *
	 * @return true if the session ID came in as part of a URL; otherwise, false
	 * @see    #getSession(boolean)
	 * @see    javax.servlet.http.HttpServletRequest#isRequestedSessionIdFromURL()
	 */
	public boolean isRequestedSessionIdFromURL()
	{
		return _request.isRequestedSessionIdFromURL();
	}

	/**
	 * Returns a java.security.Principal object containing the name of the current authenticated
	 * user. If the user has not been authenticated, the method returns null.
	 *
	 * @return a java.security.Principal containing the name of the user making this request; null
	 *         if the user has not been authenticated
	 * @see    javax.servlet.http.HttpServletRequest#getUserPrincipal()
	 */
	public Principal getUserPrincipal()
	{
		return _request.getUserPrincipal();
	}

	/**
	 * Returns a boolean indicating whether the authenticated user is included in the specified
	 * logical "role". Roles and role membership can be defined using deployment descriptors. If the
	 * user has not been authenticated, the method returns false.
	 *
	 * @param  role a String specifying the name of the role
	 * @return a boolean indicating whether the user making this request belongs to a given role;
	 *         false if the user has not been authenticated
	 * @see    javax.servlet.http.HttpServletRequest#isUserInRole(java.lang.String)
	 */
	public boolean isUserInRole(String role)
	{
		return _request.isUserInRole(role);
	}

	/**
	 * @deprecated As of Version 2.1 of the Java Servlet API, use isRequestedSessionIdFromURL()
	 *             instead.
	 * @see        javax.servlet.http.HttpServletRequest#isRequestedSessionIdFromUrl()
	 */
	public boolean isRequestedSessionIdFromUrl()
	{
		return _request.isRequestedSessionIdFromURL();
	}

	/**
	 * Checks whether the requested session ID is still valid.
	 *
	 * @return true if this request has an id for a valid session in the current session context;
	 *         false otherwise
	 * @see    #getRequestedSessionId()
	 * @see    #getSession(boolean)
	 * @see    javax.servlet.http.HttpSessionContext
	 * @see    javax.servlet.http.HttpServletRequest#isRequestedSessionIdValid()
	 */
	public boolean isRequestedSessionIdValid()
	{
		return _request.isRequestedSessionIdValid();
	}

	/* From ServletRequest */

	/**
	 * Gets the value of the named attribute as an Object, or null if no attribute of the given name
	 * exists. Attributes can be set two ways. The servlet container may set attributes to make
	 * available custom information about a request. For example, for requests made using HTTPS, the
	 * attribute javax.servlet.request.X509Certificate can be used to retrieve information on the
	 * certificate of the client. Attributes can also be set programatically using
	 * setAttribute(java.lang.String, java.lang.Object). This allows information to be embedded into
	 * a request before a RequestDispatcher call. Attribute names should follow the same conventions
	 * as package names. This specification reserves names matching java.*, javax.*, and sun.*.
	 *
	 * @param  name a String specifying the name of the attribute
	 * @return an Object containing the value of the attribute, or null if the attribute does not
	 *         exist
	 * @see    javax.servlet.ServletRequest#getAttribute(java.lang.String)
	 * @see    #setAttribute(String, Object)
	 */
	public Object getAttribute(String name)
	{
		return _request.getAttribute(name);
	}

	/**
	 * Gets an Enumeration containing the names of the attributes available to this request. This
	 * method returns an empty Enumeration if the request has no attributes available to it.
	 *
	 * @return an Enumeration of strings containing the names of the request's attributes
	 * @see    javax.servlet.ServletRequest#getAttributeNames()
	 */
	public Enumeration<?> getAttributeNames()
	{
		return _request.getAttributeNames();
	}

	/**
	 * Gets the name of the character encoding used in the body of this request. This method returns
	 * null if the request does not specify a character encoding
	 *
	 * @return a String containing the name of the chararacter encoding, or null if the request does
	 *         not specify a character encoding
	 * @see    javax.servlet.ServletRequest#getCharacterEncoding()
	 * @see    #setCharacterEncoding(String)
	 */
	public String getCharacterEncoding()
	{
		return _request.getCharacterEncoding();
	}

	/**
	 * Overrides the name of the character encoding used in the body of this request. This method
	 * must be called prior to reading request parameters or reading input using getReader().
	 *
	 * @param  enc a String containing the name of the chararacter encoding.
	 * @throws java.io.UnsupportedEncodingException if this is not a valid encoding
	 * @see    javax.servlet.ServletRequest#setCharacterEncoding(java.lang.String)
	 * @see    #getCharacterEncoding()
	 */
	public void setCharacterEncoding(String enc)
		throws UnsupportedEncodingException
	{
		throw new IllegalArgumentException("BF AppEngine : Extension in J2EE not*yet* activated");
		// XXX uncomment below
		// _request.setCharacterEncoding(enc);
	}

	/**
	 * Gets the length, in bytes, of the request body and made available by the input stream, or -1
	 * if the length is not known. For HTTP servlets, same as the value of the CGI variable
	 * CONTENT_LENGTH.
	 *
	 * @return an integer containing the length of the request body or -1 if the length is not known
	 * @see    javax.servlet.ServletRequest#getContentLength()
	 */
	public int getContentLength()
	{
		return _request.getContentLength();
	}

	/**
	 * Gets the MIME type of the body of the request, or null if the type is not known. For HTTP
	 * servlets, same as the value of the CGI variable CONTENT_TYPE.
	 *
	 * @return a String containing the name of the MIME type of the request, or null if the type is
	 *         not known
	 * @see    javax.servlet.ServletRequest#getContentType()
	 */
	public String getContentType()
	{
		return _request.getContentType();
	}

	/**
	 * Retrieves the body of the request as binary data using a ServletInputStream. Either this
	 * method or getReader() may be called to read the body, not both.
	 *
	 * @return a ServletInputStream object containing the body of the request
	 * @throws java.lang.IllegalStateException if the getReader() method has already been called for
	 *                                         this request
	 * @throws java.io.IOException             if an input or output exception occurred
	 * @see    javax.servlet.ServletRequest#getInputStream()
	 */
	public ServletInputStream getInputStream()
		throws IOException
	{
		return _request.getInputStream();
	}

	/**
	 * Returns the preferred Locale that the client will accept content in, based on the
	 * Accept-Language header. If the client request doesn't provide an Accept-Language header, this
	 * method returns the default locale for the server.
	 *
	 * @return the preferred Locale for the client
	 * @see    javax.servlet.ServletRequest#getLocale()
	 */
	public Locale getLocale()
	{
		return _request.getLocale();
	}

	/**
	 * Returns an Enumeration of Locale objects indicating, in decreasing order starting with the
	 * preferred locale, the locales that are acceptable to the client based on the Accept-Language
	 * header. If the client request doesn't provide an Accept-Language header, this method returns
	 * an Enumeration containing one Locale, the default locale for the server.
	 *
	 * @return an Enumeration of preferred Locale objects for the client
	 * @see    javax.servlet.ServletRequest#getLocales()
	 */
	public Enumeration<?> getLocales()
	{
		return _request.getLocales();
	}

	/**
	 * Returns the value of a request parameter as a String, or null if the parameter does not
	 * exist. Request parameters are extra information sent with the request. For HTTP servlets,
	 * parameters are contained in the query string or posted form data. You should only use this
	 * method when you are sure the parameter has only one value. If the parameter might have more
	 * than one value, use getParameterValues(java.lang.String). If you use this method with a
	 * multivalued parameter, the value returned is equal to the first value in the array returned
	 * by getParameterValues. If the parameter data was sent in the request body, such as occurs
	 * with an HTTP POST request, then reading the body directly via getInputStream() or getReader()
	 * can interfere with the execution of this method.
	 *
	 * @param  name a String specifying the name of the parameter
	 * @return a String representing the single value of the parameter
	 * @see    #getParameterValues(java.lang.String)
	 * @see    javax.servlet.ServletRequest#getParameter(java.lang.String)
	 */
	public String getParameter(String name)
	{
		Object param = _parameters.get(name);

		if (param instanceof String[])
		{
			String[] vals = (String[])param;

			if (vals.length > 0)
				return (vals[0]);
			else
				return null;
		}
		return ((String)param);
	}

	/**
	 * Returns an Enumeration of String objects containing the names of the parameters contained in
	 * this request. If the request has no parameters, the method returns an empty Enumeration.
	 *
	 * @return an Enumeration of String objects, each String containing the name of a request
	 *         parameter; or an empty Enumeration if the request has no parameters
	 * @see    javax.servlet.ServletRequest#getParameterNames()
	 */
	public Enumeration<?> getParameterNames()
	{
		return (_parameters.keys());
	}

	/**
	 * Returns a java.util.Map of the parameters of this request. Request parameters are extra
	 * information sent with the request. For HTTP servlets, parameters are contained in the query
	 * string or posted form data.
	 *
	 * @return an immutable java.util.Map containing parameter names as keys and parameter values as
	 *         map values. The keys in the parameter map are of type String. The values in the
	 *         parameter map are of type String array.
	 * @see    javax.servlet.ServletRequest#getParameterMap()
	 */
	public Map<?,?> getParameterMap()
	{
		throw new IllegalArgumentException("BF AppEngine : Extension in J2EE not*yet* activated");
		// XXX uncomment below
		// return (_request.getParameterMap());
	}

	/**
	 * Returns an array of String objects containing all of the values the given request parameter
	 * has, or null if the parameter does not exist. If the parameter has a single value, the array
	 * has a length of 1.
	 *
	 * @param  name a String containing the name of the parameter whose value is requested
	 * @return an array of String objects containing the parameter's values
	 * @see    #getParameter(java.lang.String)
	 * @see    javax.servlet.ServletRequest#getParameterValues(java.lang.String)
	 */
	public String[] getParameterValues(String name)
	{
		Object param = _parameters.get(name);

		if (param == null)
			return null;
		if (param instanceof String[])
			return ((String[])param);

		String[] values = new String[1];

		values[0] = (String)param;
		return (values);
	}

	/**
	 * Returns the name and version of the protocol the request uses in the form
	 * protocol/majorVersion.minorVersion, for example, HTTP/1.1. For HTTP servlets, the value
	 * returned is the same as the value of the CGI variable SERVER_PROTOCOL.
	 *
	 * @return a String containing the protocol name and version number
	 * @see    javax.servlet.ServletRequest#getProtocol()
	 */
	public String getProtocol()
	{
		return _request.getProtocol();
	}

	/**
	 * Returns the name of the scheme used to make this request, for example, http, https, or ftp.
	 * Different schemes have different rules for constructing URLs, as noted in RFC 1738.
	 *
	 * @return a String containing the name of the scheme used to make this request
	 * @see    javax.servlet.ServletRequest#getScheme()
	 */
	public String getScheme()
	{
		return _request.getScheme();
	}

	/**
	 * Returns the host name of the server that received the request. For HTTP servlets, same as the
	 * value of the CGI variable SERVER_NAME.
	 *
	 * @return a String containing the name of the server to which the request was sent
	 * @see    javax.servlet.ServletRequest#getServerName()
	 */
	public String getServerName()
	{
		return _request.getServerName();
	}

	/**
	 * Returns the port number on which this request was received. For HTTP servlets, same as the
	 * value of the CGI variable SERVER_PORT.
	 *
	 * @return an integer specifying the port number
	 * @see    javax.servlet.ServletRequest#getServerPort()
	 */
	public int getServerPort()
	{
		return _request.getServerPort();
	}

	/**
	 * Retrieves the body of the request as character data using a BufferedReader. The reader
	 * translates the character data according to the character encoding used on the body. Either
	 * this method or getInputStream() may be called to read the body, not both.
	 *
	 * @return a BufferedReader containing the body of the request
	 * @throws java.io.UnsupportedEncodingException - if the character set encoding used is not
	 *                                              supported and the text cannot be decoded
	 * @throws java.lang.IllegalStateException      - if getInputStream() method has been called on
	 *                                              this request
	 * @throws java.io.IOException                  - if an input or output exception occurred
	 * @see    #getInputStream()
	 * @see    javax.servlet.ServletRequest#getReader()
	 */
	public java.io.BufferedReader getReader()
		throws IOException
	{
		return _request.getReader();
	}

	/**
	 * Returns the Internet Protocol (IP) address of the client that sent the request. For HTTP
	 * servlets, same as the value of the CGI variable REMOTE_ADDR.
	 *
	 * @return a String containing the IP address of the client that sent the request
	 * @see    javax.servlet.ServletRequest#getRemoteAddr()
	 */
	public String getRemoteAddr()
	{
		return _request.getRemoteAddr();
	}

	/**
	 * Returns the fully qualified name of the client that sent the request. If the engine cannot or
	 * chooses not to resolve the hostname (to improve performance), this method returns the
	 * dotted-string form of the IP address. For HTTP servlets, same as the value of the CGI
	 * variable REMOTE_HOST.
	 *
	 * @return a String containing the fully qualified name of the client
	 * @see    javax.servlet.ServletRequest#getRemoteHost()
	 */
	public String getRemoteHost()
	{
		return _request.getRemoteHost();
	}

	/**
	 * Removes an attribute from this request. This method is not generally needed as attributes
	 * only persist as long as the request is being handled. Attribute names should follow the same
	 * conventions as package names. Names beginning with java.*, javax.*, and com.sun.*, are
	 * reserved for use by Sun Microsystems.
	 *
	 * @param key a String specifying the name of the attribute to remove
	 * @see   javax.servlet.ServletRequest#removeAttribute(java.lang.String)
	 */
	public void removeAttribute(String key)
		throws IllegalStateException
	{
		_request.removeAttribute(key);
	}

	/**
	 * Stores an attribute in this request. Attributes are reset between requests. This method is
	 * most often used in conjunction with RequestDispatcher. Attribute names should follow the same
	 * conventions as package names. Names beginning with java.*, javax.*, and com.sun.*, are
	 * reserved for use by Sun Microsystems. If the value passed in is null, the effect is the same
	 * as calling removeAttribute(java.lang.String).
	 *
	 * @param key a String specifying the name of the attribute
	 * @param o   the Object to be stored
	 * @see   javax.servlet.ServletRequest#setAttribute(java.lang.String, java.lang.Object)
	 * @see   #getAttribute(String)
	 */
	public void setAttribute(String key, Object o)
		throws IllegalStateException
	{
		_request.setAttribute(key, o);
	}

	/**
	 * Returns a boolean indicating whether this request was made using a secure channel, such as
	 * HTTPS.
	 *
	 * @return a boolean indicating if the request was made using a secure channel
	 * @see    javax.servlet.ServletRequest#isSecure()
	 */
	public boolean isSecure()
	{
		return _request.isSecure();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see javax.servlet.ServletRequest#getRealPath(java.lang.String)
	 */
	@SuppressWarnings("deprecation")
	public String getRealPath(String path)
	{
		return _request.getRealPath(path);
	}
} // end class _LyMultipartRequestWrapper



/**
 * Data source that store different types of attachment files.
 *
 * @version $Id: LyFileAttachmentHandler.java 41816 2012-12-19 15:18:30Z smaupin $
 * @author  Lyria
 */
class _LyMultipartDataSource
	implements DataSource,
		LyFreeInterface
{
	public static final String __VERSION =
		"$Id: LyFileAttachmentHandler.java 41816 2012-12-19 15:18:30Z smaupin $";

	/** The input stream used. */
	protected InputStream	   _iStream;

	/** The MIME type of the data. */
	protected String		   _contentType;

	/** The name of the object. */
	protected String		   _name;

	/** Maximum lenght of the data source. */
	protected int			   _length;

	/** Buffer for the file save. */
	protected byte[]		   _buffer;

	/**
	 * Constructor. Initializes class attributes with the given values.
	 *
	 * @param name        the data source name
	 * @param contentType the MIME type of data
	 * @param iStream     the used input stream
	 * @param length      the lenght of the data source
	 */
	public _LyMultipartDataSource(String name, String contentType, InputStream iStream, int length)
	{
		_name = name;
		_contentType = contentType;
		_iStream = iStream;
		_length = Math.max(4096, length);
	}

	/**
	 * This method returns the MIME type of the data in the form of a string. It should always
	 * return a valid type. It is suggested that getContentType return "application/octet-stream" if
	 * the DataSource implementation can not determine the data type.
	 *
	 * @return the MIME Type
	 */
	public String getContentType()
	{
		return _contentType;
	}

	/**
	 * Return the name of this object where the name of the object is dependant on the nature of the
	 * underlying objects. DataSources encapsulating files may choose to return the filename of the
	 * object. (Typically this would be the last component of the filename, not an entire pathname.)
	 *
	 * @return the name of the object
	 */
	public String getName()
	{
		return _name;
	}

	/**
	 * This method returns an InputStream representing the data and throws the appropriate exception
	 * if it can not do so. Note that a new InputStream object must be returned each time this
	 * method is called, and the stream must be positioned at the beginning of the data.
	 *
	 * @return an InputStream
	 * @throws java.io.IOException
	 */
	public InputStream getInputStream()
		throws IOException
	{
		if (_buffer == null)
			return _iStream;
		else
			return (new ByteArrayInputStream(_buffer));
	}

	/**
	 * This method returns an OutputStream where the data can be written and throws the appropriate
	 * exception if it can not do so. Note that a new OutputStream object must be returned each time
	 * this method is called, and the stream must be positioned at the location the data is to be
	 * written.
	 *
	 * @return an OutputStream
	 * @throws java.io.IOException
	 */
	public OutputStream getOutputStream()
		throws IOException
	{
		throw new UnsupportedOperationException(
			"BF AppEngine : getOutputStream() not implemented in Multipart servlet reader.");
	}

	/**
	 * Closes stream access on this dataSource Memory is freed.
	 */
	public void free()
	{
		_buffer = null;
		_iStream = null;
	}
} // end class _LyMultipartDataSource
