/*
 * LyFilePartWriter.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web;

import java.io.*;
import java.util.*;
import leon.misc.*;
import leon.view.setinterface.*;


/**
 * This class saves attachment on server side and retrieve them on request. This class should be
 * overridden to specify location.
 *
 * @version $Id: LyFilePartWriter.java 33204 2011-11-18 14:49:18Z nhamon $
 * @author  Lyria
 */
public class LyFilePartWriter
	implements LyFreeInterface
{
	public static final String __VERSION =
		"$Id: LyFilePartWriter.java 33204 2011-11-18 14:49:18Z nhamon $";

	/** Current environment. */
	protected LyEnvironment    _environment;

	/** Original filename on client side. */
	protected String		   _clientFileName;

	/** Current server fileName. */
	protected String		   _serverFileName;

	/** File content type (sent by the client, can't verify). */
	protected String		   _contentType;

	/** Server directory to store files. */
	protected String		   _attachmentDirectory;

	/**
	 * Constructor.
	 */
	public LyFilePartWriter()
	{
		_environment = null;
	}

	/**
	 * Gets the environment.
	 *
	 * @param environment Current environment
	 */
	public void setEnvironment(LyEnvironment environment)
	{
		_environment = environment;
	}

	/**
	 * Saves the stream to a file on server side Takes current parameters, and if needed compute a
	 * serverFileName. File is saved to a temporary file until commit. (handle close / valid)
	 *
	 * @param  iStream input stream to read the data from.
	 * @throws IOException
	 * @see    #computeServerFileName()
	 */
	public void saveFile(InputStream iStream)
		throws IOException
	{
		if (getServerFileName() == null)
			computeServerFileName();

		String fileName = getServerFileName();

		if (fileName == null)
		{
			_environment.logError("LyFilePartWriter : can not save client file : "
				+ getClientFileName() + " server filename is null (undefined).");
			return;
		}

		File   f;
		
		if (_environment.getViewManager().getTemporaryDir().equals(_attachmentDirectory))
		{
			fileName = fileName.replace('\\', '/');
			
			String suffix = null;

			fileName = fileName.replace('\\', '/');
			
			int n1 = fileName.lastIndexOf('/');
			if (n1 >=0)
				fileName = fileName.substring(n1+1);
			
			n1 = fileName.lastIndexOf('.');
			if ((n1 >=0) && ((fileName.length() - n1)>3))
			{
				suffix = fileName.substring(n1);
				fileName = fileName.substring(0, n1);
			}
			if (fileName.length()<3)
				fileName = fileName+"tmp";
			
			// Check if _attachmentDirectory exist
			File attachDir =  new File(_attachmentDirectory);
			if (!attachDir.exists())
				attachDir.mkdirs();
			LyFileTools tools = new LyFileTools(_environment);

			f = tools.createTempFile(fileName, suffix, attachDir, true);
			setServerFileName(f.getAbsolutePath());
		}
		else
			f = new File(fileName+ ".tmp"); // Use temporary file
		String d = f.getParent();

		if (d != null)
		{
			File dir = new File(d);

			if (!dir.exists())
				dir.mkdirs();
		}
		// Dump stream
		int i = iStream.read();

		// FFT 397
		if (i < 0)
		{
			// Empty file (EOF).
			// W4/FFT01063
			f.createNewFile();
			
			try
			{
				iStream.close();
			}
			catch (IOException e)
			{
			}
			
			return;
		}
		
		OutputStream oStream = null;
		
		try
		{
			oStream = new BufferedOutputStream(new FileOutputStream(f));
			
			while (i >= 0)
			{
				oStream.write(i);
				i = iStream.read();
			}
		}
		finally
		{
			if (oStream != null)
			{
				try
				{
					oStream.close();
					iStream.close();
				}
				catch (IOException e)
				{
				}
			}
		}

		if (_environment != null)
			_environment.logTrace("LyFilePartWriter : File : " + getClientFileName()
				+ " was successfully saved as : " + f.getAbsolutePath());
	} // end method saveFile

	/**
	 * Validates file upload. File is stored permanently on server side, or removed if the parameter
	 * is false.
	 *
	 * @param  keep true if the file is committed, false otherwise.
	 * @throws IOException
	 */
	public void commit(boolean keep)
		throws IOException
	{
		String fileName = getServerFileName();
		
		if (fileName == null)
			return;

		if (_environment != null && (_environment.getViewManager() != null))
		{
			// Used by BLOBs
			String tmpDir = _environment.getViewManager().getTemporaryDir();
			
			tmpDir = _environment.getString().expand(tmpDir);
			if (fileName.startsWith(tmpDir))
			{
				// temporary file for upload, do nothing
				return;
			}
		}
		
		File   f2 = new File(fileName + ".tmp");

		// Get rid of temporary file if commit aborts.
		if (!keep)
		{
			if (f2.exists())
				f2.delete();
			if (_environment != null)
				_environment.logTrace(" LyFilePartWriter COMMIT delete temporary file : " + f2);
			return;
		}

		// Replace temporary and actual file
		if (!f2.exists())
		{
			if (_environment != null)
				_environment.logTrace("LyFilePartWriter : can't commit. file not found : " + f2);
			return;
		}

		File f1 = new File(fileName);

		if (f1.exists())
		{
			if (_environment != null)
				_environment.logTrace(
					"LyFilePartWriter : Server file already exists. Overwriting in commit : "
					+ f1);
			f1.delete();
		}

		f2.renameTo(f1);

		if (_environment != null)
			_environment.logTrace("LyFilePartWriter : Committed : " + f2 + " to " + f1);
	} // end method commit

	/**
	 * Cleans up everything Keep parameters and internal request for later use.
	 */
	public void free()
	{
		_environment = null;
		_clientFileName = null;
		_serverFileName = null;
		_contentType = null;
	}

	/**
	 * Gets the original filename on client side.
	 *
	 * @return original fileName or null if unknown
	 * @see    #setClientFileName(String)
	 */
	public String getClientFileName()
	{
		return _clientFileName;
	}

	/**
	 * Sets the filename on client side.
	 *
	 * @param fileName the new fileName
	 * @see   #getClientFileName()
	 */
	public void setClientFileName(String fileName)
	{
		_clientFileName = fileName;
	}

	/**
	 * Gets server filename (local name). This is the actual local location of the file on the
	 * server. The value stored in database (client side location) is getFileLocation which return
	 * the same value.
	 *
	 * @return local file path on server or null if unknown
	 * @see    #getFileLocation()
	 * @see    #setServerFileName(String)
	 */
	public String getServerFileName()
	{
		return _serverFileName;
	}

	/**
	 * Location (address) to access to this file (stored in database) default returns
	 * getServerFileName().
	 *
	 * @return server location or null if unknown
	 * @see    #getServerFileName()
	 */
	public String getFileLocation()
	{
		String fileLocation = getServerFileName().replace('\\', '/');
		int    i = fileLocation.lastIndexOf('/');
		String shortName;

		if (i >= 0)
			shortName = fileLocation.substring(i + 1);
		else
			shortName = fileLocation;

		return (_environment.getString().expand(shortName));
	}

	/**
	 * Automatically compute a valid server file name from current properties.
	 *
	 * <P>May be overriden by subclasses. This method is called when saving data and when no server
	 * filename is specified.</P>
	 *
	 * @see #setServerFileName(String)
	 */
	public void computeServerFileName()
	{
		if (_clientFileName == null)
			throw new IllegalArgumentException("BF AppEngine : Can not compute server fileName.");
		
		_clientFileName = _clientFileName.replace('\\', '/');
		int    i = _clientFileName.lastIndexOf('/');
		String shortName;

		if (i >= 0)
			shortName = _clientFileName.substring(i + 1);
		else
			shortName = _clientFileName;

		if (_environment.getViewManager().getTemporaryDir().equals(_attachmentDirectory))
		{
			// likely a BLOB, use a temporary File (will be stored in DB)
			setServerFileName(_environment.getString().expand(_attachmentDirectory+"/"+shortName));
		}
		else
		{	
	
			String docDir = _environment.getEnv("LY_APP_DOC_DIR");
			String fileDir = docDir + "/" + _attachmentDirectory;
	
	
			setServerFileName(_environment.getString().expand(
					fileDir + "/" + "__" + (new Date()).getTime() + "__" + shortName));
		}
	}

	/**
	 * Server filename (local name).
	 *
	 * @param fileName the new server file name
	 * @see   #getServerFileName()
	 */
	public void setServerFileName(String fileName)
	{
		_serverFileName = fileName;
	}

	/**
	 * Gets file content type (sent by the client, can't verify).
	 *
	 * @return content type (mime type) or null if unknown
	 * @see    #setContentType(String)
	 */
	public String getContentType()
	{
		return _contentType;
	}

	/**
	 * Sets file content type.
	 *
	 * @param contentType the new content type (mime type). May be null.
	 * @see   #getContentType()
	 */
	public void setContentType(String contentType)
	{
		_contentType = contentType;
	}

	/**
	 * Gets the attachment directory from meta-model.
	 *
	 * @return the attachment directory
	 * @see    #setAttachmentDirectory(String)
	 */
	public String getAttachmentDirectory()
	{
		return _attachmentDirectory;
	}

	/**
	 * Sets the attachment directory.
	 *
	 * @param attachmentDirectory the new attachment directory
	 * @see   #getAttachmentDirectory()
	 */
	public void setAttachmentDirectory(String attachmentDirectory)
	{
		_attachmentDirectory = attachmentDirectory;
	}
} // end class LyFilePartWriter
