/*
 * LyWebObject.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 * Current Version: $Id: LyWebObject.java 35801 2012-06-11 08:33:47Z nhamon $.
 */

package leon.view.web;

import leon.misc.*;
import leon.view.*;
import leon.view.graphic.*;

import java.awt.*;

import java.io.*;


/**
 * The LyWebObject class implements root object for all Web components.
 */
public abstract class LyWebObject extends LyGraphicalObject implements LyWebComponent
{
    public static final String __VERSION =
        "$Id: LyWebObject.java 35801 2012-06-11 08:33:47Z nhamon $";

    public static final String WEB_COMPONENT_ID = "/component.htm?_componentId=".intern();
    public static final String WEB_CONTROLLER_ID = "&_controllerId=".intern();
    public static final String WEB_SESSION_ID = "&_sessionId=".intern();

    /** Name of the frame where this object was last generated. */
    protected String _lastFrameName;

    /** Name of the window where this object was last generated. */
    protected String _lastWindowName;

    /** Id of the writer of the frame where this object was last generated. */
    protected String _lastFrameWriterId;

    /** Indicates if component is being updated by parameter values. */
    protected boolean _updating;

    /**
     * Constructs abstract graphical component.
     *
     * @param viewComponent the view component
     */
    public LyWebObject(LyComponent viewComponent)
    {
        super(viewComponent);

        _lastFrameName = null;
        _lastWindowName = null;
        _lastFrameWriterId = null;
        _updating = false;
    }

    /**
     * Encode an URL by replacing unauthorized characters.
     *
     * @param  url the URL to encode
     * @return the encoded URL
     */
    public static String encodeUrl(String url)
    {
        String newUrl;

        newUrl = LyString.replace(url, " ", "%20");
        // Needed for Javascript
        newUrl = LyString.replace(newUrl, "\\", "\\\\");
        newUrl = LyString.replace(newUrl, "\"", "&#34;");
        newUrl = LyString.replace(newUrl, "'", "\\%27");

        return newUrl;
    }

    /**
     * Encode an URL by replacing unauthorized characters.
     *
     * @param  url the URL to encode
     * @return the encoded URL
     */
    public static String encodeUrlXhtml(String url)
    {
        String newUrl = encodeUrl(url);

        newUrl = LyString.replace(url, "&", "&#38;");

        return newUrl;
    }

    /**
     * Encode an URL by replacing unauthorized characters.
     *
     * @param  url the URL to encode
     * @return the encoded URL
     */
    public static String decodeUrl(String url)
    {
        String newUrl;

        newUrl = LyString.replace(url, "%20", " ");
        newUrl = LyString.replace(newUrl, "\\\\", "\\");

        return newUrl;
    }

    /**
     * Get servlet root URL : this is "http://"+local host + LY_APP_DOC rebuild the root Url.
     *
     * @param  environment the application environment
     * @return the root URL
     */
    public static String getRootURL(LyEnvironment environment)
    {
        if (environment == null)
            return null;

        LyWebViewManager viewManager = (LyWebViewManager) environment.getViewManager();

        if (viewManager == null)
            return null;

        return viewManager.getRootURL();
    }

    /**
     * Indicates if an up page is displayed in the web browser.
     *
     * @param  env the application environment
     * @return true if an up page is displayed in the web browser, false otherwise
     */
    public static boolean displayUpPage(LyEnvironment env)
    {
        String file = env.getFile("LY_UP_PAGE");

        return ((file != null) && !"LY_UP_PAGE".equals(file) && !"_void".equals(file));
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LyFreeInterface#free()
     */
    public void free()
    {
        _lastFrameName = null;
        _lastWindowName = null;
        _lastFrameWriterId = null;
    }

    /**
     * Gets web view manager.
     *
     * @return the web view manager
     */
    public final LyWebViewManager getWebViewManager()
    {
        return (LyWebViewManager) getViewManager();
    }

    /**
     * Gets encoded HTML string from given string.
     *
     * @param  label the string to encode
     * @return the HTML encoded string
     * @see    LyString#getHtmlString(String)
     */
    public final String getHtmlString(String label)
    {
        return LyString.getHtmlString(translate(label));
    }

    /**
     * Gets encoded HTML string from given string.
     *
     * @param  label the string to encode
     * @return the HTML encoded string
     * @see    LyString#getHtmlEntitiesString(String)
     */
    public final String getHtmlEntitiesString(String label)
    {
        return LyString.getHtmlEntitiesString(label);
    }

    /**
     * Gets encoded Javascript string from given string.
     *
     * @param  label the given string
     * @return the encoded Javascript string
     * @see    LyString#getJavascriptString(String)
     */
    public final String getJavascriptString(String label)
    {
        return LyString.getJavascriptString(translate(label));
    }

    /**
     * Gets font corresponding to the given key.
     *
     * @param  key the font key
     * @return the font
     */
    public final LyFont getLyFont(String key)
    {
        return getEnvironment().getLyFont(key);
    }

    /**
     * Gets font corresponding to the given key.
     *
     * @param  key the font key
     * @return the font size
     */
    public final int getFontSize(String key)
    {
        LyFont font = getLyFont(key);

        return ((font == null) ? 0 : font.getSize());
    }

    /**
     * Gets root URL.
     *
     * @return the root URL
     * @see    #getRootURL(LyEnvironment)
     */
    public String getRootUrl()
    {
        return getRootURL(getEnvironment());
    }

    /**
     * Replaces within a given String some $properties$ found in the system environment. This method
     * is looking for macros between '$' and replaces these macros by a property if any (such as
     * $LY_APP_DIR$/data/test.html).
     *
     * @param  text The original string, before any substitution.
     * @return The string after all substitutions.
     */
    public final String expandUrl(String text)
    {
        return getEnvironment().getString().expandUrl(text);
    }

    /**
     * Retrieves an url from the application environment.
     *
     * @param  key The ressource name (key) of the url
     * @return A reference to the requested url, or null if it could not be found.
     * @see    LyEnvironment#getUrl(String)
     */
    public String getUrl(String key)
    {
        return getEnvironment().getUrl(key);
    }

    /**
     * Retrieves an image url from the application environment.
     *
     * @param  key The ressource name (key) of the image.
     * @return The image url, or the key if it could not be found.
     * @see    LyEnvironment#getImageUrl(String)
     */
    public final String getImageUrl(String key)
    {
        return getEnvironment().getImageUrl(key);
    }

    /**
     * Retrieves a disabled image url from the application environment.
     *
     * @param  key The ressource name (key) of the image.
     * @return The image url, or the key if it could not be found.
     * @see    LyEnvironment#getDisabledImageUrl(String)
     */
    public final String getDisabledImageUrl(String key)
    {
        return getImageUrl(key);
    }

    /**
     * Retrieves the dimensions of an image.
     *
     * @param  key the key of the image or the path of the image
     * @return An array of two integers (width, height) describing the size of the image or null if
     *         no size is given.
     */
    public final int[] getImageSize(String key)
    {
        return getImageSize(key, false);
    }

    /**
     * Retrieves the dimensions of an image.
     *
     * @param  key             the key of the image or the path of the image
     * @param  enableLoadImage indicates wether the application is allowed to load the image
     * @return An array of two integers (width, height) describing the size of the image or null if
     *         no size is given.
     */
    public final int[] getImageSize(String key, boolean enableLoadImage)
    {
        LyEnvironment environment = getEnvironment();
        String        image = environment.getImagePath(key);

        return ((image == null) ? null : environment.getImageSize(image, enableLoadImage));
    }

    /**
     * Convenient routine to get a formated message with given arguments.
     *
     * @see LyEnvironment#getMessage(String, Object[])
     */
    public final String getMessage(String key, Object[] arguments)
    {
        String message = super.getMessage(key, arguments);

        message = LyWebViewManager.encodeString(message);
        message = message.replace('\"', '\'');
        message = LyString.replace(message, "\'", "\\\'");

        return message;
    }

    /**
     * Get hexadecimal color from given key.
     *
     * @param  key The ressource name (key) of the Color.
     * @return the hexadecimal color for the given key (eg: #FF0000)
     */
    public String getHexColor(String key)
    {
        LyColor color = getEnvironment().getLyColor(key);

        if (color == null)
            return null;

        return "#" + LyColors.getHexString(color);
    }

    /**
     * Get hexadecimal color from given key.
     *
     * @param  key        The ressource name (key) of the Color.
     * @param  defaultKey The default ressource name (key) to be used if no color was found for
     *                    given key
     * @return the hexadecimal color for the given key (eg: #FF0000)
     */
    public String getHexColor(String key, String defaultKey)
    {
        String hexColor = getHexColor(key);

        if (hexColor != null)
            return hexColor;
        else if (defaultKey != null)
            return getHexColor(defaultKey);
        else
            return null;
    }

    /**
     * Gets font description.
     *
     * @param  key The ressource name (key) of the Font.
     * @return a string describing the font. This result may be used in a css stylesheet.
     */
    public final String getFontString(String key)
    {
        return getFontString(key, null);
    }

    /**
     * Gets font description.
     *
     * @param  key   The ressource name (key) of the Font.
     * @param  color Indicates whether the color for given key must be retrieved.
     * @return a string describing the font. This result may be used in a css stylesheet.
     */
    public String getFontString(String key, boolean color)
    {
        String hexColor = (color ? getHexColor(key) : null);

        return getFontString(key, hexColor);
    }

    /**
     * Gets font and color description.
     *
     * @param  key   The ressource name (key) of the Font.
     * @param  color The ressource name (key) of the Color.
     * @return a string describing the font. This result may be used in a css stylesheet.
     */
    public String getFontString(String key, String color)
    {
        if ((color != null) && (color.charAt(0) != '#'))
            color = getHexColor(color);

        LyEnvironment env = getEnvironment();
        LyFont        font = ((env == null) ? null : env.getLyFont(key));

        if (font == null)
            return "";

        StringBuffer buffer = new StringBuffer();

        buffer.append("font-family:").append(font.getFontName()).append(";");
        buffer.append("font-size:").append(font.getSize()).append("px;");

        if (font.isBold())
            buffer.append("font-weight:bold;");

        if (font.isItalic())
            buffer.append("font-style:italic;");

        if (color != null)
            buffer.append("color:").append(color).append(";");

        return buffer.toString();
    }

    /**
     * Update the refresh info of this object. This method updates information such as
     * _lastFrameName and _lastWindowName with values given by the page writer parameter.
     *
     * @param pw the current page writer
     */
    public void updateRefreshInfo(LyWebPageWriter pw)
    {
        updateRefreshInfo(pw, false);
    }

    /**
     * Update the refresh info of this object. This method updates information such as
     * _lastFrameName, _lastWindowName and _lastFrameWriterId with values given by the page writer
     * parameter.
     *
     * @param pw     the current page writer
     * @param keepId indicates whether the _lastFrameWriterId parameter must be updated
     */
    protected void updateRefreshInfo(LyWebPageWriter pw, boolean keepId)
    {
        LyComponent component = pw.getViewComponent();
        boolean     writeDiv = !pw.writeDivOnly() || (component != _viewComponent);

        // In print mode do not register refresh information
        if (writeDiv && !pw.isPrintMode())
        {
            _lastFrameName = pw.getActiveName();

            if (pw.getActiveName() == LyWebPageWriter.MAIN_BUFFER)
            {
                _lastFrameName = LyWebConstants.TARGET_FRAME;

                if (LyWebConstants.POPUP_VIEW.equals(pw.getWindowName()))
                    _lastFrameName = LyWebConstants.POPUP_VIEW;
            }

            _lastWindowName = pw.getWindowName();

            if (keepId)
                _lastFrameWriterId = pw.getId();
        }
    }

    /**
     * Return the name of the frame where this object was last generated.
     *
     * @return the value of the _lastFrameName attribute.
     */
    public String getLastFrameName()
    {
        return _lastFrameName;
    }

    /**
     * Refresh frame containing this web object.
     */
    public void refresh()
    {
        LyWebViewManager viewManager = (LyWebViewManager) getViewManager();

        // Get current frame name and set refresh
        if (_lastFrameName != null)
        {
            String frameName;

            if (_lastWindowName == null)
                frameName = _lastFrameName;
            else
                frameName = _lastWindowName + "__" + _lastFrameName;

            viewManager.addReloadFrame(frameName, _lastWindowName);
            viewManager.addPopupViewMapping(frameName, _lastWindowName);
        }
    }

    /**
     * Add an external Javascript file to the view manager scripts.
     *
     * @param fileName the name of the external Javascript file
     */
    public void addFileScript(String fileName)
    {
        if (_viewComponent != null)
        {
            String           scriptRoot = ((LyWebViewManager) _viewComponent.getViewManager())
                .getScriptsDir();
            LyWebViewManager viewManager = getWebViewManager();

            viewManager.addAdditionalRefreshScripts("<script src='" + scriptRoot + "/" + fileName
                + "'></script>\n");
        }
    }

    /**
     * Replace an URL pattern http:// inside the document by href tags. This method is a recursive
     * building of links like :
     *
     * @param  label     the label containing URLs
     * @param  target    the target frame of the URL
     * @param  printMode indicates if print mode is active
     * @return the new URL
     */
    public String replaceUrl(String label, String target, boolean printMode)
    {
        return replaceUrl(label, target, printMode, -1, "LY_DEFAULT_FONT");
    }

    /**
     * Replace an URL pattern http:// inside the document by href tags. This method is a recursive
     * building of links like :
     *
     * @param  label     the label containing URLs
     * @param  target    the target frame of the URL
     * @param  printMode indicates if print mode is active
     * @param  maxLength indicates the maxLength of the generated link
     * @param  fontString the font property used to display label
     * @return the new URL
     */
    public String replaceUrl(String label, String target, boolean printMode, int maxLength,
        String fontString)
    {
        String HTTP_PATTERN = "http://";
        String MAILTO_PATTERN = "mailto:";
        String A_HREF_PATTERN = "href=";
        String WWW_PATTERN = "www.";
        int    n1 = label.indexOf(HTTP_PATTERN);
        int    n1_mail = label.indexOf(MAILTO_PATTERN);
        int    n1_href = label.indexOf(A_HREF_PATTERN);
        int    n1_www = label.indexOf(WWW_PATTERN);

        if ((n1 < 0) && (n1_mail < 0) && (n1_www < 0))
        {
            if (maxLength > 0)
                label = getReducedString(label, fontString, maxLength);

            return "<font class='label'>" + label + "</font>"; // exit
            // condition,
            // nothing more
        }

        if (n1_href > 0)
            return label; // Label contains the hyperlink : no need to generate it

        if ((n1_mail >= 0) && ((n1_mail < n1) || (n1 < 0)))
            n1 = n1_mail; // mode mailto

        if ((n1_www >= 0) && (n1 < 0))
            n1 = n1_www; // mode www.

        StringBuffer sb = new StringBuffer(label.length() + 20);
        int          n2 = label.indexOf("<br />", n1);
        int          n3 = label.indexOf(',', n1);
        int          n4 = label.indexOf(' ', n1);

        if (n2 < 0)
        {
            if (n3 < 0)
            {
                if (n4 < 0)
                    n2 = label.length(); // end of label
                else
                    n2 = n4;
            }
            else
                n2 = n3;
        }

        if (n1 > 0)
        {
            sb.append("<font class='label'>");
            sb.append(getReducedString(label.substring(0, n1), fontString, maxLength)); // content before
            sb.append("</font>");
        }

        if (((maxLength > 0) && (maxLength >= n1)) || (maxLength < 0))
        {
            if (!printMode)
            {
                sb.append("<a href='");

                if (n1_www == n1)
                    sb.append(HTTP_PATTERN);

                sb.append(label.substring(n1, n2));
                sb.append("' target=\"").append(target).append("\">");
            }

            sb.append("<font class='label'>");

            String labelTmp;

            if (n1_mail == n1)
                labelTmp = label.substring(n1 + MAILTO_PATTERN.length(), n2); // hide
            // mailto
            else
                labelTmp = label.substring(n1, n2);

            if (maxLength > 0)
                labelTmp = getReducedString(labelTmp, fontString, maxLength - n1);

            sb.append(labelTmp);
            sb.append("</font>");

            if (!printMode)
                sb.append("</a>");

            if (n2 < label.length())
                sb.append(replaceUrl(label.substring(n2), target, printMode)); // recurse
        }

        return sb.toString();
    } // end method replaceUrl

    /**
     * Computes the reduced string for a given label and a given length.
     */
    @SuppressWarnings("deprecation")
    protected String getReducedString(String label, String fontString, int width)
    {
		LyEnvironment environment = getEnvironment();
		Font		  defaultFont = (Font)environment.getFont("LY_TABLE");
		Font		  font = (Font)environment.getFont(fontString);
		String		  initialLabel = label + "";
		int			  pos = ((label == null) ? 0 : label.length());
		Toolkit		  toolkit = Toolkit.getDefaultToolkit();
		FontMetrics   defaultFm = toolkit.getFontMetrics(defaultFont);
		FontMetrics   fm = toolkit.getFontMetrics(font);
		float		  charWidth = ((float)defaultFm.stringWidth("0123456789")) / 10;

        if (width > 0)
        {
            while ((pos > 0) && (fm.stringWidth(label) > (charWidth * width)))
			{
				pos--;
				label = initialLabel.substring(0, pos) + "...";
            }
        }

		return label;
    }

    /**
     * Indicates whether this component needs scrollbars when inserted in a splitter.
     *
     * @param pw current page writer
     * @return a positive value if this component needs scrollbars, a negative or null value otherwise.
     */
    public int getScrollHeightForSplitter(LyWebPageWriter pw)
    {
        return -1;
    }

    /**
     * Indicates if a modification of the component is in process.
     *
     * @param updating true if component is being updated, false otherwise.
     */
    public void setUpdating(boolean updating)
    {
        _updating = updating;
    }

    /**
     * Generate a webview on a mobile
     * @param pw PrintWriter
     * @param sessionId the session id
     */
    public void writeJson(PrintWriter pw, String sessionId)
    {
        LyComponent component = getViewComponent();
        String      path = getViewManager().getRootURL() + WEB_COMPONENT_ID + component.getId()
            + WEB_CONTROLLER_ID + component.getRootController() + WEB_SESSION_ID + sessionId;

        pw.println("{\"WEBVIEW\": [");
        pw.print("{\"link\": \"" + path + "\"},");
        pw.print("{\"controller\": \"" + component.getRootController() + "\"}");
        pw.println("]}");
    }
} // end class LyWebObject
