/*
 * LyWebPageWriter.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web;

import java.io.*;
import java.util.*;

import javax.servlet.http.*;

import leon.misc.*;
import leon.view.*;
import leon.view.getinterface.*;
import leon.view.web.struts.*;


/**
 * It must be inherited in order to generate HTML pages.
 *
 * @version $Id: LyWebPageWriter.java 35027 2012-03-23 17:03:08Z jlsaouli $
 * @author  Lyria
 */
public abstract class LyWebPageWriter extends LyWebObject
	implements LyWebComponent
{
	public static final String __VERSION =
		"$Id: LyWebPageWriter.java 35027 2012-03-23 17:03:08Z jlsaouli $";

	/** The main buffer name. */
	public static final String MAIN_BUFFER = "main".intern();

	/** Indicates whether Ajax mode must be forced in every page. */
	public static boolean	   FORCE_AJAX = false;

	/** Default Background color. */
	protected String		   _bgColor;

	/** Default Background motif. */
	protected String		   _bgMotif;

	/** Background color for page frames. */
	protected Hashtable<String, String>		   _bgColorTable;

	/** Root path for javascripts. */
	protected String		   _scriptRoot;

	/** This page writer id. */
	protected String		   _id;

	/** This page writer name. */
	protected String		   _name;

	/** The title of this component. */
	protected String		   _title;

	/** Mapping between a frame (key) and the buffer containing its HTML representation. */
	protected Hashtable<String, StringBuffer>		   _buffers;

	/** Mapping between a frame (key) and the list of its begin scripts HTML representation. */
	protected Hashtable<String, Vector<_LyScript>>		   _beginScripts;

	/** Mapping between a frame (key) and the list of its end scripts HTML representation. */
	protected Hashtable<String, Vector<_LyScript>>		   _endScripts;

	/** Mapping between a frame (key) and the list of its CSS scripts HTML representation. */
	protected Hashtable<String, Vector<_LyCssScript>>	   _cssScripts;

	/** Mapping between a frame (key) and the identifier of the form it contains. */
	protected Hashtable<String, String>		   _formIds;

	/**
	 * Mapping between the active buffer (key) and the attachment handler class used to manage file
	 * submission in forms.
	 */
	protected Hashtable<String, String>		   _formAttachmentHandler;

	/**
	 * Active buffer name.
	 *
	 * @see #_activeBuffer
	 */
	protected String		   _activeName;

	/** Active buffer: the buffer in which HTML information of the page will be wrote. */
	protected StringBuffer     _activeBuffer;

	/**
	 * List of begin scripts associated to the active buffer.
	 *
	 * @see #_activeBuffer
	 */
	protected Vector<_LyScript>		   _activeBeginScripts;

	/**
	 * List of end scripts associated to the active buffer.
	 *
	 * @see #_activeBuffer
	 */
	protected Vector<_LyScript>		   _activeEndScripts;

	/**
	 * List of CSS scripts associated to the active buffer.
	 *
	 * @see #_activeBuffer
	 */
	protected Vector<_LyCssScript>		   _activeCssScripts;

	/**
	 * Form identifier associated to the active buffer.
	 *
	 * @see #_activeBuffer
	 */
	protected String		   _activeFormId;

	/** Indicates if framesets is used or not. */
	protected boolean		   _noFrames;

	/** Indicates the print mode of the page. */
	protected boolean		   _printMode;

	//CLEANCODE : xslClassInfo, _xslActionId and _xslRootActionId not used
	//protected String                      _xslClassInfo;
	//protected String                      _xslActionId;
	//protected String                      _xslRootActionId;

	/** Indicates if this page has to force its reload or not. */
	protected boolean		   _forceReloadState;

	/** The time since the last reload perform. */
	protected long			   _lastReloadTime;

	/** The time since the last modification. */
	protected long			   _lastModifiedTime;

	/** Ids of modified elements. */
	protected Vector<String>   _oldIds;

	/** Indicates if this component is visible or not. */
	protected boolean		   _visible;

	/** The current javascript id. */
	protected long			   _javascriptId;

	/** Name of the target Window when displayed or null for main Window. */
	protected String		   _windowName;

	/**
	 * List of properties (value of type java.util.Hashtable) associated to a buffer (key equals to
	 * a buffer name). Properties are themselves a list of properties (values of type String)
	 * associated to their name (keys value of type String).
	 */
	protected Hashtable<String, Hashtable<String, String>>		   _properties_by_buffer;

	/** Indicates that we use only a partial rendering of a DIV tag for DHTML. */
	protected boolean		   _writeDivOnly;

	/** Indicates if we are in a field of a form. */
	protected boolean		   _innerView;

	/** The name of the CSS class to be used for body tag. */
	protected String			_bodyClassName;

	/** Indentation level. */
    protected int              _indent = 0;

    /** */
    protected boolean          _endsWithCR;


	/**
	 * Constructor.
	 *
	 * @param component the component of this page
	 */
	public LyWebPageWriter(LyComponent component)
	{
		super(component);

		// Give an id to the page writer
		_id = LyString.getRandomId();
	}

	/**
	 * Initialization method.<br>
	 * Sets the main buffer as the active buffer, register this page writer on its view manager and
	 * set a default value to the class attributes.
	 */
	public void init()
	{
		_buffers = new Hashtable<String, StringBuffer>();
		_beginScripts = new Hashtable<String, Vector<_LyScript>>();
		_endScripts = new Hashtable<String, Vector<_LyScript>>();
		_cssScripts = new Hashtable<String, Vector<_LyCssScript>>();
		_formIds = new Hashtable<String, String>();
		_formAttachmentHandler = new Hashtable<String, String>();
		_activeBuffer = null;
		_activeBeginScripts = null;
		_activeEndScripts = null;
		_activeCssScripts = null;
		_activeFormId = null;
		_name = null;
		_title = null;
		_bgColor = null;
		_bgMotif = null;
		_scriptRoot = null;
		_noFrames = false;
		_printMode = false;
		_bodyClassName = null;

		// CLEANCODE : _xslClassInfo, _xslActionId and _xslRootActionId not used
		//_xslClassInfo = null;
		//_xslActionId = null;
		//_xslRootActionId = null;

		_forceReloadState = false;
		_lastReloadTime = 0;
		_lastModifiedTime = -1;
		_visible = false;
		setActiveBuffer(MAIN_BUFFER);
		_javascriptId = 0;
		_properties_by_buffer = new Hashtable<String, Hashtable<String, String>>();
		_oldIds = null;
		_writeDivOnly = false;
		_innerView = false;

		// Register this page writer on its view manager
		LyWebViewManager viewManager = getWebViewManager();

		viewManager.addPageWriter(this);
	} // end method init

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyFreeInterface#free()
	 */
	public synchronized void free()
	{
		// Unregister this page writer on its view manager
		LyEnvironment environment = getEnvironment();

		if (environment != null)
		{
			// Unregister manager
			LyWebViewManager viewManager = (LyWebViewManager)environment.getViewManager();

			if (viewManager != null)
				viewManager.removePageWriter(this);
		}

		// Clean up references
		_bgColorTable = null;
		_buffers = null;
		_beginScripts = null;
		_endScripts = null;
		_cssScripts = null;
		_formIds = null;
		_formAttachmentHandler = null;
		_activeBeginScripts = null;
		_activeEndScripts = null;
		_oldIds = null;
		_properties_by_buffer = null;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#getId()
	 */
	public String getId()
	{
		return _id;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#writeHtml(java.io.Writer, java.lang.String)
	 */
	public synchronized void writeHtml(Writer pw, String frame)
		throws IOException
	{
		writeBeginPage(pw, frame);
		writeBeginScripts(pw, frame);
		writePageContent(pw, frame);
		writeEndScripts(pw, frame);
		writeEndPage(pw, frame);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#preparePage(java.lang.String, java.lang.String)
	 */
	public void preparePage(String frame, String targetWindow)
	{
		_javascriptId = 0;
		if (targetWindow != null)
			setWindowName(targetWindow); // keep the name of the window where shown (may be null or popup)
		clearBeginScripts(frame);
		clearEndScripts(frame);
		clearCssScripts(frame);
	}

	/**
	 * Writes the beginning of a HTML page (head, scripts and open the body).
	 *
	 * @param  pw    the page writer
	 * @param  frame the frame where the page is written
	 * @throws IOException if something is wrong during page writing
	 */
	public void writeBeginPage(Writer pw, String frame)
		throws IOException
	{
		LyWebDecoration deco = getWebViewManager().getDecoration();

		pw.write("<!-- Generated by Leonardi (Lyria) -->\n");
		pw.write("<html><head>\n");
		pw.write("<meta name=\"AUTHOR\" content=\"Lyria\"/>\n");
		pw.write("<meta name=\"GENERATOR\" content=\"Leonardi\"/>\n");
		pw.write("<meta name=\"Content-Type\" content=\"text/html;charset=" + LyString.DEFAULT_CHARSET + "\"/>\n");

		if ((frame != null) && frame.equals(LyWebConstants.PILOT_FRAME))
			pw.write("<meta http-equiv='Page-Enter' content='blendTrans(Duration=0.2)'>");

		pw.write("<title>");
		pw.write((_title == null) ? "" : translate(_title));
		pw.write("</title>\n");
		writeCssScripts(pw, frame);
		// Store pageWriterId
		pw.write("</head>\n");

		if ((frame == null) || frame.equals(LyWebConstants.TARGET_FRAME))
			frame = MAIN_BUFFER;

		pw.write("<body ");

		String	bodyClass = getBodyClassString();

		if (bodyClass != null)
			pw.write(" " + bodyClass);

		LyEnvironment		   env = getEnvironment();
		LyWebViewManager	   manager = getWebViewManager();

		if (manager.hasProperty("DYNAMIC")
		        && !(this instanceof LyStrutsDictionary))
			pw.write(" scroll='nO'");

		if (deco.getSkinFile() == null)
		{
			if (getBackgroundColor(frame) != null)
			{
				pw.write(" bgcolor='");
				pw.write(getBackgroundColor(frame));
				pw.write("'");
			}

			if (getBackground() != null)
			{
				pw.write(" background='");
				pw.write(getBackground());
				pw.write("'");
			}
		}
		pw.write(" topmargin='0' leftmargin='0' style='height:100%;width:100%;' " + getBodyScript()
			+ ">\n");

		String	scripts = manager.generateScriptsInclusion();

		pw.write(scripts);
		pw.write("<script type='text/javascript'>var pageWriterId='" + getId() + "';</script>\n");

		if (!env.testEnv("LY_REPLAY")
				&& ((getAttachmentHandler(frame) == null) || (manager instanceof LyStrutsViewManager))
				&& (manager.hasProperty("DYNAMIC")))
			pw.write("<script type='text/javascript'>var AJAX_MODE='" + (FORCE_AJAX ? "force" : true) + "';</script>\n");
		else
			pw.write("<script type='text/javascript'>var AJAX_MODE='false';</script>\n");

		// Manage Struts mode
		if (manager instanceof LyStrutsViewManager)
		{
			String	actionUrl = ((LyStrutsViewManager)manager).getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);

			pw.write("<script type='text/javascript'>var STRUTS_EVENT_URL='" + actionUrl + "';</script>\n");
		}

		pw.write("<script type='text/javascript'>var isInPopUp='" + isInPopUp() + "';</script>\n");
	} // end method writeBeginPage

	/**
	 * Returns the CSS class string to be used for body tag.
	 *
	 * @return the CSS class string or null if no special class string has to be used.
	 */
	protected String getBodyClassString()
	{
		if (_bodyClassName != null)
			return "class='" + _bodyClassName + "'";

		LyWebDecoration deco = getWebViewManager().getDecoration();

		return deco.getBodyClassString(this);
	}

	/**
	 * Sets the name of the CSS class to be used for body tag.
	 *
	 * @param bodyClassName the CSS class name to be used.
	 */
	public void setBodyClassName(String bodyClassName)
	{
		_bodyClassName = bodyClassName;
	}

	/**
	 * Indicates if the component is in a pop up frame.
	 *
	 * @return true if the component is in a pop up frame, false otherwise
	 */
	protected boolean isInPopUp()
	{
		LyEnvironment		   env = _viewComponent.getEnvironment();
		LyViewManagerInterface manager = env.getViewManager();

		if (manager.hasProperty("DYNAMIC"))
		{
			if (LyWebConstants.POPUP_VIEW.equals(getWindowName()))
				return true;
		}

		return false;
	}

	/**
	 * Fetches additional information for body HTML tag.
	 *
	 * @return a String with information to add to the body HTML tag if the view manager is
	 *         dynamic, empty String otherwise
	 */
	protected String getBodyScript()
	{
		if (_viewComponent == null)
			return null;
		
		LyEnvironment		   env = _viewComponent.getEnvironment();
		LyViewManagerInterface manager = env.getViewManager();

		if (manager.hasProperty("DYNAMIC"))
		{
			if (LyWebConstants.POPUP_VIEW.equals(getWindowName()))
				return "onload='checkSizePopupView(null, true);ResetFormFocus(false);'";
			else
				return "onload='checkSizeMainView();ResetFormFocus(false);' onResize='checkSizeMainView()'";
		}
		else
			return "";
	}

	/**
	 * Writes the begin scripts of the given frame.
	 *
	 * @param  pw    the page writer
	 * @param  frame the wrame
	 * @throws IOException
	 * @see    #getBeginScriptsAsHtml(String)
	 */
	public void writeBeginScripts(Writer pw, String frame)
		throws IOException
	{
		// Write the begin scripts
		pw.write(getBeginScriptsAsHtml(frame).toString());
		pw.write("\n");
	}

	/**
	 * Writes the page content of the frame and the given writer.
	 *
	 * @param  pw    the page writer
	 * @param  frame the frame to write
	 * @throws IOException
	 */
	public void writePageContent(Writer pw, String frame)
		throws IOException
	{
		if ((frame == null) || frame.equals(LyWebConstants.TARGET_FRAME))
			frame = MAIN_BUFFER;

		// Use one single form for a same frame
		pw.write(getOpenFormContent(frame));

		StringBuffer buffer = getBuffer(frame);

		if (buffer == null)
			throw new RuntimeException("BF AppEngine : Buffer not ready for this page : " + this
				+ " frame : " + LyString.getHtmlEntitiesString(frame));

		// Write the corresponding buffer
		pw.write(buffer.toString());
		pw.write("\n");
	} // end method writePageContent

	/**
	 * Writes the open form tag of the frame and the given writer.
	 *
	 * @param  frame the frame to write
	 */
	public String getOpenFormContent(String frame)
	{
		return getOpenFormContent(frame, false);
	}

	/**
	 * Writes the open form tag of the frame and the given writer.
	 *
	 * @param  frame the frame to write
	 * @param  force indicates whether form must be generated event if no form
	 * name has been specified on this page writer
	 */
	public String getOpenFormContent(String frame, boolean force)
	{
		StringBuffer	buffer = new StringBuffer();
		String			attachmentHandlerClass = getAttachmentHandler(frame);
		String			action = getActionUrl();
		String			formName = getFormId(frame);

		if (force || ((formName != null) && (formName.length() != 0)))
		{
			buffer.append("<form name='");

			if (force && ((formName == null) || (formName.length() == 0)))
				buffer.append("_default");
			else
				buffer.append(formName);

			if (attachmentHandlerClass != null)
			{
				buffer.append("' enctype='");
				buffer.append(LyWebConstants.CONTENT_MULTIPART); // XX should get
				// from the
				// attachmentHandlerClass,
				// only this one
				// supported.
			}

			if (action != null)
			{
				buffer.append("' action='");
				buffer.append(action);
			}

			buffer.append("' method='post' style='margin:0px;'>\n");

			buffer.append("<input type='hidden' name='" + LyWebConstants.CONTROLLER + "' value='"
				+ formName + "'/>\n");
			buffer.append("<input type='hidden' name='" + LyWebConstants.SOURCE + "' value=''/>\n");
			buffer.append("<input type='hidden' name='" + LyWebConstants.HIDDEN + "' value=''/>\n");
			buffer.append("<input type='hidden' name='" + LyWebConstants.COMMAND + "' value=''/>\n");
			buffer.append("<input type='hidden' name='" + LyWebConstants.VALUE + "' value=''/>\n");
			buffer.append("<input type='hidden' name='" + LyWebConstants.TYPE + "' value=''/>\n");
			buffer.append("<input type='hidden' name='" + LyWebConstants.FRAME + "' value=''/>\n");
			buffer.append("<input type='hidden' name='" + LyWebConstants.INDEX + "' value=''/>\n");
			buffer.append("<input type='hidden' name='" + LyWebConstants.EXTEND + "' value=''/>\n");
			buffer.append("<input type='hidden' name='" + LyWebConstants.POSITION_X + "' value=''/>\n");
			buffer.append("<input type='hidden' name='" + LyWebConstants.POSITION_Y + "' value=''/>\n");
			buffer.append("<input type='hidden' name='" + LyWebConstants.SUBMITTED + "' value=''/>\n");
			buffer.append("<input type='hidden' name='" + LyWebConstants.TARGET_CONTROLLER
				+ "' value=''/>\n");

			LyWebRequestInfo	info = LyStrutsServlet.getWebRequestInfo();
			HttpServletRequest	request = (info == null ? null : info.getRequest());
			String				target = (request == null ? null : (String)request.getParameter(LyWebConstants.TARGET_PARAM));

			if (target != null)
				buffer.append("<input type='hidden' name='" + LyWebConstants.TARGET_PARAM + "' value='" + target + "'/>\n");

			// Write ajax mode ressource
			LyEnvironment		   env = getEnvironment();
			LyViewManagerInterface manager = env.getViewManager();

			if (!env.testEnv("LY_REPLAY")
					&& ((getAttachmentHandler(frame) == null) || (manager instanceof LyStrutsViewManager))
					&& (manager.hasProperty("DYNAMIC")))
				buffer.append("<input type='hidden' name='AJAX_MODE' value='" + (FORCE_AJAX ? "force" : true) + "'/>\n");
			else
				buffer.append("<input type='hidden' name='AJAX_MODE' value='false'/>\n");
		}

		return buffer.toString();
	}

	/**
	 * Writes CSS scritps associated to the given frame into the given writer.
	 *
	 * @param  pw    the page writer
	 * @param  frame the frame to write CSS scripts
	 * @throws IOException
	 * @see    #getCssScriptsAsHtml(String, boolean)
	 */
	public void writeCssScripts(Writer pw, String frame)
		throws IOException
	{
		String css = getCssScriptsAsHtml(frame, false).toString();

		if ((css != null) && (css.length() > 0))
		{
			pw.write("<style type=\"text/css\">\n");
			pw.write("<!--\n");
			pw.write(css);
			pw.write("-->\n");
			pw.write("</style>\n");
		}

		LyWebDecoration deco = getWebViewManager().getDecoration();

		if (deco.getSkinFile() != null)
		{
			pw.write("<style type='text/css' media='screen'>\n @import url('" + deco.getSkinFile()
				+ "');\n</style>");

            if (isPrintMode() && (deco.getSkinPath() != null))
            {
                // Generate the content of CSS file in order to get a correct print result
                LyEnvironment   env = getEnvironment();
                InputStream     input = env.getInputStream(deco.getSkinPath());
                String          line;

                pw.append("<style type='text/css'>\n");
                pw.append("<!--\n");

                while ((line = LyString.getLine(input)) != null)
                {
                    if (line.trim().startsWith("overflow:"))
                        // Workaround : printing cut on first page with Firefox
                        continue;

                    pw.append(line);
                    pw.append('\n');
                }

                input.close();

                pw.append("-->\n");
                pw.append("</style>\n");
            }
		}

		if (deco.getSpecificCSSFile() != null)
			pw.write("<style type='text/css' media='screen'>\n @import url('" + deco.getSpecificCSSFile()
				+ "');\n</style>");

		pw.write(getCssScriptsAsHtml(frame, true).toString());
	}

	/**
	 * Writes the end scripts of the given frame.
	 *
	 * @param  pw    the page writer
	 * @param  frame the wrame
	 * @throws IOException
	 * @see    #getEndScriptsAsHtml(String)
	 */
	public void writeEndScripts(Writer pw, String frame)
		throws IOException
	{
		// Write the end scripts
		pw.write(getEndScriptsAsHtml(frame).toString());
		pw.write("\n");
	}

	/**
	 * Writes the end of HTML page. Closes the body and html tags and the form tag if it exists.
	 *
	 * @param  pw    the page writer
	 * @param  frame the frame where the page is written
	 * @throws IOException if something is wrong during end page writes
	 */
	public void writeEndPage(Writer pw, String frame)
		throws IOException
	{
		String formName = getFormId(frame);

		if ((formName != null) && (formName.length() != 0))
			pw.write("</form>\n");

		pw.write("</body>\n");
		pw.write("</html>\n");
		pw.write("<!-- End of HTML generation by Lyria -->\n");
	}

	/**
	 * Gets the buffer associated to the given frame.
	 *
	 * @param  frame the frame
	 * @return the buffer associated to the given frame if it exists, null otherwise
	 */
	public StringBuffer getBuffer(String frame)
	{
		if ((frame == null) || frame.equals(LyWebConstants.TARGET_FRAME))
			frame = MAIN_BUFFER;

		return _buffers.get(frame);
	}

	/**
	 * Gets the vector of all begin scripts whose associated to the given frame.
	 *
	 * @param  key the frame whose begin scripts are wanted
	 * @return the list of begin scripts
	 */
	protected Vector<_LyScript> getBeginScripts(String key)
	{
		if ((key == null) || key.equals(LyWebConstants.TARGET_FRAME))
			key = MAIN_BUFFER;

		return _beginScripts.get(key);
	}

	/**
	 * Gets the vector of all end scripts whose associated to the given frame.
	 *
	 * @param  key the frame whose end scripts are wanted
	 * @return the list of begin scripts
	 */
	protected Vector<_LyScript> getEndScripts(String key)
	{
		if ((key == null) || key.equals(LyWebConstants.TARGET_FRAME))
			key = MAIN_BUFFER;

		return _endScripts.get(key);
	}

	/**
	 * Gets the vector of css scripts associated to the given frame.
	 *
	 * @param  frame the frame of the wanted scripts
	 * @return the list of CSS scritps associated to the frame
	 */
	protected Vector<_LyCssScript> getCssScripts(String frame)
	{
		if ((frame == null) || frame.equals(LyWebConstants.TARGET_FRAME))
			frame = MAIN_BUFFER;

		return _cssScripts.get(frame);
	}

	/**
	 * Gets the formId associated to the given frame.
	 *
	 * @param  frame the frame
	 * @return the form id associated to the given frame if it exists, null otherwise
	 * @see    #setFormId(String)
	 */
	public String getFormId(String frame)
	{
		if ((frame == null) || frame.equals(LyWebConstants.TARGET_FRAME))
			frame = MAIN_BUFFER;

		String formId = _formIds.get(frame);

		// '.' in form names may not work in Javascript commands.
		// Replace it by '_'.
		if (formId != null)
			formId = formId.replace('.', '_');

		return formId;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#getAttachmentHandler(java.lang.String)
	 */
	public String getAttachmentHandler(String key)
	{
		if ((key == null) || key.equals(LyWebConstants.TARGET_FRAME))
			key = MAIN_BUFFER;

		return _formAttachmentHandler.get(key);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#getAttachmentDirectory(java.lang.String)
	 */
	public String getAttachmentDirectory(String key)
	{
		if ((key == null) || key.equals(LyWebConstants.TARGET_FRAME))
			key = MAIN_BUFFER;

		return _formAttachmentHandler.get(key + ".att_dir");
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#clearBeginScripts()
	 */
	public void clearBeginScripts()
	{
		_beginScripts.put(MAIN_BUFFER, new Vector<_LyScript>());
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#clearEndScripts()
	 */
	public void clearEndScripts()
	{
		_endScripts.put(MAIN_BUFFER, new Vector<_LyScript>());
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#clearCssScripts()
	 */
	public void clearCssScripts()
	{
		_cssScripts.put(MAIN_BUFFER, new Vector<_LyCssScript>());
	}

	/**
	 * Clears the begin scripts of the given frame. This method is not called during the generic
	 * action, but it may be used in specific code.
	 *
	 * @param name the name of the frame begin scripts to clear.
	 */
	public void clearBeginScripts(String name)
	{
		if ((name == null) || name.equals(LyWebConstants.TARGET_FRAME))
			name = MAIN_BUFFER;

		if (_beginScripts == null)
			_beginScripts = new Hashtable<String, Vector<_LyScript>>();

		_beginScripts.put(name, new Vector<_LyScript>());
	}

	/**
	 * Clears the end scripts of the given frame. This method is not called during the generic
	 * action, but it may be used in specific code.
	 *
	 * @param name the name of the frame end scripts to clear.
	 */
	public void clearEndScripts(String name)
	{
		if ((name == null) || name.equals(LyWebConstants.TARGET_FRAME))
			name = MAIN_BUFFER;

		_endScripts.put(name, new Vector<_LyScript>());
	}

	/**
	 * Clears all CSS scripts of the given frame. This method is not called during the generic
	 * action, but it may be used in specific code.
	 *
	 * @param name the name of the frame of all CSS scripts to clear.
	 */
	public void clearCssScripts(String name)
	{
		if ((name == null) || name.equals(LyWebConstants.TARGET_FRAME))
			name = MAIN_BUFFER;

		_cssScripts.put(name, new Vector<_LyCssScript>());
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#clearBuffer(java.lang.String)
	 */
	public void clearBuffer(String name)
	{
		if ((name == null) || name.equals(LyWebConstants.TARGET_FRAME))
			name = MAIN_BUFFER;

		_buffers.put(name, new StringBuffer());
		_formIds.put(name, "");
		_formAttachmentHandler.remove(name);
		_formAttachmentHandler.remove(name + ".att_dir");
		if (_properties_by_buffer != null)
			_properties_by_buffer.remove(name);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#setActiveBuffer(java.lang.String)
	 */
	public void setActiveBuffer(String name)
	{
		if ((name == null) || name.equals(LyWebConstants.TARGET_FRAME))
			name = MAIN_BUFFER;

		StringBuffer buffer = getBuffer(name);
		Vector<_LyScript>		 beginScripts = getBeginScripts(name);
		Vector<_LyScript>		 endScripts = getEndScripts(name);
		Vector<_LyCssScript>     cssScripts = getCssScripts(name);
		String		 formId = getFormId(name);

		if (buffer == null)
		{
			buffer = new StringBuffer();
			_buffers.put(name, buffer);
		}
		if (beginScripts == null)
		{
			beginScripts = new Vector<_LyScript>();
			_beginScripts.put(name, beginScripts);
		}
		if (endScripts == null)
		{
			endScripts = new Vector<_LyScript>();
			_endScripts.put(name, endScripts);
		}
		if (cssScripts == null)
		{
			cssScripts = new Vector<_LyCssScript>();
			_cssScripts.put(name, cssScripts);
		}

		if (formId == null)
		{
			formId = ""; // Does not like empty strings in hashtable !
			_formIds.put(name, "");
			_formAttachmentHandler.remove(name);
			_formAttachmentHandler.remove(name + ".att_dir");
		}

		_activeName = name;
		_activeBuffer = buffer;
		_activeBeginScripts = beginScripts;
		_activeEndScripts = endScripts;
		_activeCssScripts = cssScripts;
		_activeFormId = formId;
	} // end method setActiveBuffer

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#append(java.lang.String)
	 */
	public final void append(String string)
	{
	    if ((string == null)
	            || (string.length() == 0))
	        return;

		// final to enhance speed (frequently used method).
		if (_activeBuffer == null)
			_activeBuffer = getBuffer(MAIN_BUFFER);

		if (_indent > 0)
		{
            int   i, p;

            if ((string.charAt(0) == '\n')
                    && _endsWithCR)
                string = string.substring(1);

            if (_endsWithCR)
            {
                for (i = 0; i < _indent; i++)
                    _activeBuffer.append('\t');
                _endsWithCR = false;
            }

		    while ((p = string.indexOf('\n')) >= 0)
		    {
		    	boolean indent = false;

		    	if (p + 1 < string.length())
		    	{
		    		if (string.charAt(p + 1) == '<')
		    			indent = true;
		    	}

	            _activeBuffer.append(string.substring(0, p + 1));

	            if (p + 1 < string.length())
	            {
	            	if (indent)
	            	{
		                for (i = 0; i < _indent; i++)
		                    _activeBuffer.append('\t');
	            	}

	                string = string.substring(p + 1);
	            }
                else
                {
                    _endsWithCR = true;
                    return;
                }
		    }
		}

		_activeBuffer.append(string);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#getActiveName()
	 */
	public String getActiveName()
	{
		return _activeName;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#addBeginScript(java.lang.String, boolean)
	 */
	public void addBeginScript(String text, boolean isFile)
	{
		if (text == null)
			return;

		if (_activeBeginScripts == null)
			_activeBeginScripts = getBeginScripts(MAIN_BUFFER);

		_LyScript script = new _LyScript(text, isFile);

		// Add script if not already added
		if (!_activeBeginScripts.contains(script))
			_activeBeginScripts.addElement(script);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#addEndScript(java.lang.String, boolean)
	 */
	public void addEndScript(String text, boolean isFile)
	{
		addEndScript(text, isFile, true);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#addEndScript(java.lang.String, boolean)
	 */
	public void addEndScript(String text, boolean isFile, boolean insertEnd)
	{
		if (text == null)
			return;

		if (_activeEndScripts == null)
			_activeEndScripts = getEndScripts(MAIN_BUFFER);

		_LyScript script = new _LyScript(text, isFile);

		if (!_activeEndScripts.contains(script))
		{
			// Add script if not already added
			if (insertEnd)
				_activeEndScripts.addElement(script);
			else
				_activeEndScripts.insertElementAt(script, 0);
		}
		else
		{
			// Add script at the right position in the list
			_activeEndScripts.removeElement(script);

			if (insertEnd)
				_activeEndScripts.addElement(script);
			else
				_activeEndScripts.insertElementAt(script, 0);
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#addCssScript(java.lang.String, boolean)
	 */
	public void addCssScript(String text, boolean isFile)
	{
		if (text == null)
			return;

		if (_activeCssScripts == null)
			_activeCssScripts = getCssScripts(MAIN_BUFFER);

		_LyCssScript script = new _LyCssScript(text, isFile);

		// Add script if not already added
		if (!_activeCssScripts.contains(script))
			_activeCssScripts.addElement(script);
	}

	/**
	 * Copy Begin, End and CSS scripts to the specified page writer. Utility methods for tabs within
	 * tabs (COMPOSITION RELATION).
	 *
	 * @param pw the page writer
	 */
	public void copyScriptsTo(LyWebPageWriter pw)
	{
		_LyScript script;
		int		  i, n;

		if (_activeBeginScripts != null)
		{
			for (i = 0, n = _activeBeginScripts.size(); i < n; i++)
			{
				script = _activeBeginScripts.elementAt(i);
				pw.addBeginScript(script.getText(), script.isFile());
			}
		}

		if (_activeEndScripts != null)
		{
			for (i = 0, n = _activeEndScripts.size(); i < n; i++)
			{
				script = _activeEndScripts.elementAt(i);
				pw.addEndScript(script.getText(), script.isFile());
			}
		}

		if (_activeCssScripts != null)
		{
			for (i = 0, n = _activeCssScripts.size(); i < n; i++)
			{
				script = _activeCssScripts.elementAt(i);
				pw.addCssScript(script.getText(), script.isFile());
			}
		}
	} // end method copyScriptsTo

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#setFormId(java.lang.String)
	 */
	public void setFormId(String formId)
	{
		if (_activeName == null)
			_activeName = MAIN_BUFFER;

		_formIds.put(_activeName, formId);
	}

	/**
	 * Accessor to the attachment handler. Default value is null, This value is set by a File
	 * fieldItem in a form to submit file content on server side for storage.
	 *
	 * @param handler is set by a File fieldItem. default handler is
	 *                leon.view.web.LyFileAttachmentHandler
	 * @see   LyWebForm#getAttachmentHandler
	 */
	public void setAttachmentHandler(String attachmentHandler)
	{
		String activeName = ((_activeName == null) ? MAIN_BUFFER : _activeName);
		
		setAttachmentHandler(activeName, attachmentHandler);
	}
		
	/**
	 * Accessor to the attachment handler. Default value is null, This value is set by a File
	 * fieldItem in a form to submit file content on server side for storage.
	 *
	 * @param key key used to save attachment handler
	 * @param handler is set by a File fieldItem. default handler is
	 *                leon.view.web.LyFileAttachmentHandler
	 * @see   LyWebForm#getAttachmentHandler
	 */
	public void setAttachmentHandler(String key, String attachmentHandler)
	{
		if (attachmentHandler != null)
			_formAttachmentHandler.put(key, attachmentHandler);
		else
			_formAttachmentHandler.remove(key);
	}

	/**
	 * Accessor to the attachment handler directory. Default value is null, This value is set by a
	 * File fieldItem in a form to submit file content directory on server side for storage.
	 *
	 * @param key key used to save attachment directory
	 * @param handler is set by a File fieldItem. default directory is 'files'
	 * @see   LyWebForm#getAttachmentHandler
	 */
	public void setAttachmentDirectory(String attachmentDirectory)
	{
		String activeName = ((_activeName == null) ? MAIN_BUFFER : _activeName);
		
		setAttachmentDirectory(activeName, attachmentDirectory);
	}

	/**
	 * Accessor to the attachment handler directory. Default value is null, This value is set by a
	 * File fieldItem in a form to submit file content directory on server side for storage.
	 *
	 * @param handler is set by a File fieldItem. default directory is 'files'
	 * @see   LyWebForm#getAttachmentHandler
	 */
	public void setAttachmentDirectory(String key, String attachmentDirectory)
	{
		if (attachmentDirectory != null)
			_formAttachmentHandler.put(key + ".att_dir", attachmentDirectory);
		else
			_formAttachmentHandler.remove(key + ".att_dir");
	}

	/**
	 * Implementation of LyWebPageWriter.
	 */
	public String getActiveFormId()
	{
		return getFormId(getActiveName());
	}

	/**
	 * Gets a StringBuffer that contains the HTML for the begin scripts associated to the given
	 * frame.
	 *
	 * @param  frame the given frame
	 * @return the StringBuffer containing the HTML for the begin scripts
	 * @see    #getBeginScripts(String)
	 */
	public StringBuffer getBeginScriptsAsHtml(String frame)
	{
		if ((frame == null) || frame.equals(LyWebConstants.TARGET_FRAME))
			frame = MAIN_BUFFER;

		Vector<_LyScript>     scripts = getBeginScripts(frame);
		StringBuffer          buffer = new StringBuffer();

		if (scripts != null)
		{
			_LyScript script;
			int		  i, n = scripts.size();

			for (i = 0; i < n; i++)
			{
				script = scripts.elementAt(i);
				script.writeHtml(buffer);
			}
		}

		return buffer;
	}

	/**
	 * Gets a StringBuffer that contains the HTML for the end scripts associated to the given frame.
	 *
	 * @param  frame the given frame
	 * @return the StringBuffer containing the HTML for the end scripts
	 * @see    #getEndScripts(String)
	 */
	public StringBuffer getEndScriptsAsHtml(String frame)
	{
		if ((frame == null) || frame.equals(LyWebConstants.TARGET_FRAME))
			frame = MAIN_BUFFER;

		Vector<_LyScript>	  scripts = getEndScripts(frame);
		StringBuffer          buffer = new StringBuffer();

		if (scripts != null)
		{
			_LyScript script;
			int		  i, n = scripts.size();

			for (i = 0; i < n; i++)
			{
				script = scripts.elementAt(i);
				script.writeHtml(buffer);
			}
		}

		return buffer;
	}

	/**
	 * Gets a StringBuffer that contains the HTML code of the given frame CSS scripts.
	 *
	 * @param  frame the frame
	 * @param  file  indicates if the CSS script is in a file or not
	 * @return a StringBuffer containing the HTML code of the CSS scripts of the given frame
	 * @see    #getCssScripts(String)
	 */
	public StringBuffer getCssScriptsAsHtml(String frame, boolean file)
	{
		if ((frame == null) || frame.equals(LyWebConstants.TARGET_FRAME))
			frame = MAIN_BUFFER;

		Vector<_LyCssScript>  scripts = getCssScripts(frame);
		StringBuffer          buffer = new StringBuffer();

		if (scripts != null)
		{
			_LyCssScript script;
			int			 i, n = scripts.size();

			for (i = 0; i < n; i++)
			{
				script = scripts.elementAt(i);
				if (script.isFile() == file)
					script.writeHtml(buffer);
			}
		}

		return buffer;
	}

	/**
	 * Gets the HTML code corresponding to a CSS file style sheet declaration of the given frame.
	 *
	 * @param  frame the frame
	 * @return StringBuffer containg the HTML code of a CSS file declaration
	 */
	public StringBuffer getCssFileAsHtml(String frame)
	{
		if ((frame == null) || frame.equals(LyWebConstants.TARGET_FRAME))
			frame = MAIN_BUFFER;

		Vector<_LyCssScript>  scripts = getCssScripts(frame);
		StringBuffer          buffer = new StringBuffer();

		if (scripts != null)
		{
			_LyCssScript script;
			int			 i, n = scripts.size();

			for (i = 0; i < n; i++)
			{
				script = scripts.elementAt(i);
				if (script.isFile())
					script.writeHtml(buffer);
			}
		}

		return buffer;
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.setinterface.LySetVisibleInterface#setVisible(boolean)
	 */
	public void setVisible(boolean status)
	{
		LyWebViewManager viewManager = getWebViewManager();

		if (status)
			viewManager.showView(this);
		else
			viewManager.hideView(this);

		_visible = status;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#setInnerView(boolean)
	 */
	public void setInnerView(boolean innerView)
	{
		_innerView = innerView;
	}

	/**
	 * Indicates if the component is visible or not.
	 *
	 * @return true if it is visible, false otherwise
	 */
	public boolean isVisible()
	{
		return _visible;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#isInnerView()
	 */
	public boolean isInnerView()
	{
		return _innerView;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#isFrameSet()
	 */
	public boolean isFrameSet()
	{
		return false;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#getPageWriters()
	 */
	public abstract Vector<LyWebPageWriter> getPageWriters();

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#setFrameName(java.lang.String)
	 */
	public void setFrameName(String name)
	{
		_name = name;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#setWindowName(java.lang.String)
	 */
	public void setWindowName(String name)
	{
		_windowName = name;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#getWindowName()
	 */
	public String getWindowName()
	{
		if (_windowName == null && _lastWindowName != null)
			_windowName = _lastWindowName;

		return _windowName;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#getTarget()
	 */
	public String getTarget()
	{
		return null;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#setTitle(java.lang.String)
	 */
	public void setTitle(String title)
	{
		_title = title;
	}

	/**
	 * Set the background color of the page and store the mapping between the color and frame.
	 *
	 * @param hexColor the string that represents the hexadecimal code for the color
	 * @param frame    the frame to associate to this background color
	 * @see   #getBackgroundColor(String)
	 */
	public void setBackgroundColor(String hexColor, String frame)
	{
		if (frame == null)
			_bgColor = hexColor;
		else
		{
			if (_bgColorTable == null)
				_bgColorTable = new Hashtable<String, String>();

			_bgColorTable.put(frame, hexColor);
		}
	}

	/**
	 * Gets the background color of the given frame.
	 *
	 * @param  frame the frame
	 * @return the background color of the given frame
	 * @see    #setBackgroundColor(String, String)
	 */
	public String getBackgroundColor(String frame)
	{
		// Check if a color is defined for current frame
		if ((_bgColorTable != null) && (frame != null))
		{
			String col = _bgColorTable.get(frame);

			if (col != null)
				return col;
		}

		// Check if a default color is defined, or use default bg
		if (_bgColor == null)
		{
			_bgColor = getHexColor("LY_DEFAULT_BACKGROUND");
			if (_bgColor == null)
				_bgColor = "#FFFFFF"; // Default is white
		}

		return _bgColor;
	}

	/**
	 * Sets the background motif of the page.
	 *
	 * @param image the image that must be used as a motif
	 * @see   #getBackgroundColor(String)
	 */
	public void setBackground(String image)
	{
		_bgMotif = image;
	}

	/**
	 * Gets the background motif of the page.
	 *
	 * @return the image that is used as a motif
	 * @see    #setBackground(String)
	 */
	public String getBackground()
	{
		return _bgMotif;
	}

	/**
	 * Gets the script root path of the page.
	 *
	 * @return the script root path of the page
	 */
	public String getScriptRoot()
	{
		if (_scriptRoot == null)
			_scriptRoot = getWebViewManager().getScriptsDir();

		return _scriptRoot;
	}

	/**
	 * Prevent from using frameset.
	 *
	 * @param noFrames if true, means that no frame set is used, false otherwise
	 */
	public void setNoFrames(boolean noFrames)
	{
		_noFrames = noFrames;
	}

	/**
	 * Indicates if frameset are used or not.
	 *
	 * @return true if no frame set is used, false otherwise
	 */
	public boolean isNoFrames()
	{
		return _noFrames;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#getFrameName()
	 */
	public String getFrameName()
	{
		return _name;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#setLastModified(leon.view.LyViewManager, long)
	 */
	public void setLastModified(LyViewManager viewManager, long lastModified)
	{
		if ((_lastModifiedTime == lastModified) || (lastModified < 0))
			return; // Nothing to do

		if (_oldIds == null)
			_oldIds = new Vector<String>();
		_oldIds.addElement(_id);

		// ((LyWebViewManager)viewManager).removePageWriter(this);

		int p = _id.indexOf("*");

		if (p > 0)
			_id = _id.substring(0, p) + "*" + lastModified;
		else
			_id = _id + "*" + lastModified;

		((LyWebViewManager)viewManager).addPageWriter(this);
		((LyWebViewManager)viewManager).addReloadFrame(getFrameName());
		((LyWebViewManager)viewManager).addPopupViewMapping(getFrameName(), getWindowName());

		_lastModifiedTime = lastModified;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#setForceReloadState(boolean)
	 */
	public void setForceReloadState(boolean force)
	{
		_forceReloadState = force;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#getForceReloadState()
	 */
	public boolean getForceReloadState()
	{
		return _forceReloadState;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#getLastReloadTime()
	 */
	public long getLastReloadTime()
	{
		return _lastReloadTime;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#setPrintMode(boolean)
	 */
	public void setPrintMode(boolean printMode)
	{
		_printMode = printMode;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#isPrintMode()
	 */
	public boolean isPrintMode()
	{
		return _printMode;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#getOldIds()
	 */
	public Vector<String> getOldIds()
	{
		return _oldIds;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#getNextJavaScriptId()
	 */
	public long getNextJavaScriptId()
	{
		long current = _javascriptId;

		_javascriptId++;
		return current;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#setProperty(java.lang.String, java.lang.String)
	 */
	public void setProperty(String name, String value)
	{
		// Get by active buffer
		String key = getActiveName();

		if ((key == null) || key.equals(LyWebConstants.TARGET_FRAME))
			key = MAIN_BUFFER;

		if (_properties_by_buffer == null)
			_properties_by_buffer = new Hashtable<String, Hashtable<String, String>>();

		Hashtable<String, String> properties = _properties_by_buffer.get(key);

		if (properties == null)
		{
			properties = new Hashtable<String, String>();
			_properties_by_buffer.put(key, properties);
		}

		if (value == null)
			properties.remove(name);
		else
			properties.put(name, value);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#getProperty(java.lang.String)
	 */
	public String getProperty(String name)
	{
		// Get by active buffer
		String key = getActiveName();

		if ((key == null) || key.equals(LyWebConstants.TARGET_FRAME))
			key = MAIN_BUFFER;

		if (_properties_by_buffer == null)
			_properties_by_buffer = new Hashtable<String, Hashtable<String, String>>();

		Hashtable<String, String> properties = _properties_by_buffer.get(key);

		if (properties == null)
			return null;
		return properties.get(name);
	}

	/**
	 * Change the _writeDivOnly parameter.
	 */
	public void setWriteDivOnly(boolean writeDivOnly)
	{
		_writeDivOnly = writeDivOnly;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#writeDivOnly()
	 */
	public boolean writeDivOnly()
	{
		return _writeDivOnly;
	}

	/**
	 * Builds the URL of the <b><code>action</code></b> attribute of the <b>
	 * <code>&lt;FORM&gt;</code></b> HTML tag.
	 *
	 * @return the URL of the action or null.
	 */
	public String getActionUrl()
	{
		return null;
	}

	/**
	 * Internal classes.
	 */

	/**
	 * Class used to describe a script.
	 */
	protected class _LyScript
	{

		/** Text of the script or name of the javascript file. */
		String  _text;

		/** Indicates whether this script is a file or not. */
		boolean _isFile;

		/**
		 * Constructor.
		 *
		 * @param text   text of the script or name of the javascript file
		 * @param isFile indicates whether this script is a file or not
		 */
		_LyScript(String text, boolean isFile)
		{
			_text = text.intern();
			_isFile = isFile;
		}

		/**
		 * Write HTML description corresponding to this component.
		 *
		 * @param buffer the page writer
		 */
		void writeHtml(StringBuffer buffer)
		{
			if (_isFile)
				buffer.append("<script charset='").append(LyString.DEFAULT_CHARSET).append("' src='").append(getScriptRoot()).append("/").append(_text)
					.append("'></script>\n");
			else
				buffer.append(_text).append('\n');
		}

		/*
		 * (non-Javadoc)
		 *
		 * @see java.lang.Object#equals(java.lang.Object)
		 */
		public boolean equals(Object object)
		{
			if (!(object instanceof _LyScript))
				return false;

			_LyScript script = (_LyScript)object;

			if ((!script.isFile() && isFile())
			        || (script.isFile() && !isFile()))
				return false;

			return (_text == script.getText());
		}

		/**
		 * Indicates whether this script is a file or not.
		 *
		 * @return true if the script is a file, false otherwise
		 */
		boolean isFile()
		{
			return _isFile;
		}

		/**
		 * Gets the text of the script or the name of the javascript file.
		 *
		 * @return the text of the script or the name of its file
		 */
		String getText()
		{
			return _text;
		}
	} // end class _LyScript


	/**
	 * Class used to describe a CSS script.
	 */
	protected class _LyCssScript extends _LyScript
	{
		/**
		 * Constructor.
		 *
		 * @param text   text of the script or name of the CSS file
		 * @param isFile indicates whether this script is a file or not
		 */
		_LyCssScript(String text, boolean isFile)
		{
			super(text, isFile);
		}

		/*
		 * (non-Javadoc)
		 *
		 * @see leon.view.web.LyWebPage._LyScript#writeHtml(java.lang.StringBuffer)
		 */
		void writeHtml(StringBuffer buffer)
		{
			if (_isFile)
				buffer.append("<link rel='stylesheet' type='text/css' href='").append(_text).append(
					"'/>\n");
			else
				buffer.append(_text).append('\n');
		}
	} // end class _LyCssScript

	/**
	 * Gets the url for an alternate Web page.
	 *
	 * @param  frame the frame to redirect
	 * @return the URL of the alternate Web page
	 */
	public abstract String getRedirect(String frame);

	/**
     * Adds an indentation level.
     */
    protected void addIndent()
    {
        _indent++;
    }

    /**
     * Removes an indentation level.
     */
    protected void removeIndent()
    {
        _indent--;
    }

    /**
     * Start writing of given component.
     */
    public void startComponent(LyWebObject webObject)
    {
        addIndent();

        LyEnvironment environment = getEnvironment();

        if (environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
        {
            LyComponent   component = webObject.getViewComponent();

            if (component != null)
            {
                String      tag = component.getTag();

                if (tag != null)
                {
                    append("\n<!-- <" + tag);

                    if (component instanceof LyGetIdInterface)
                    {
                        String  id = ((LyGetIdInterface)component).getId();

                        if (id != null)
                            append(" " + id);
                    }

                    append("> -->\n");
                }
            }
        }
    }

    /**
     * End writing of given component.
     */
    public void endComponent(LyWebObject webObject)
    {
        LyEnvironment environment = getEnvironment();

        if (environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
        {
            LyComponent component = webObject.getViewComponent();

            if (component != null)
            {
                String      tag = component.getTag();

                if (tag != null)
                    append("\n<!-- </" + tag + "> -->\n");
            }
        }

        removeIndent();
    }
} // end class LyWebPageWriter
