/*
 * LyWebParameterValues.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web;

import java.util.*;

import javax.servlet.http.*;

import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;


/**
 * Submit data from one or several forms.
 *
 * @version $Id: LyWebParameterValues.java 33969 2012-01-30 09:12:52Z nhamon $
 * @author  Lyria
 */
public class LyWebParameterValues
	implements LyAdvancedParameterValues
{
	public static final String   __VERSION =
		"$Id: LyWebParameterValues.java 33969 2012-01-30 09:12:52Z nhamon $";

	/**
	 * This separator is used with {@link #HTML_ENCODED_SEP}. They are separators used in forms.<br>
	 * For internal use only.<br>
	 * 261/octal == 177/decimal
	 */
	public static String		 HTML_SEP = "MULTI_VAL_SEP";

	/**
	 * This separator is used with {@link #HTML_SEP}. They are separators used in forms.<br>
	 * For internal use only.<br>
	 * 261/octal == 177/decimal
	 */
	public static String		 HTML_ENCODED_SEP = "MULTI_VAL_SEP";

	/**
	 * This separator is used to replace '\n' character in links identifiers.
	 */
	public static String		 HTML_LINK_CR_SEP = "__HTML_LINK_CR_SEP__";

	/** Reference to the http request. */
	protected HttpServletRequest _request;

	/** Reference to the LyString. */
	protected LyString			 _string;

	/**
	 * Copy of the parameter names. If keepValues() is invoked, all values from request are kept in
	 * an internal copy. This may be used when the values are intended to be valid between a
	 * request.
	 */
	protected String[]			 _copy_parameter_names;

	/**
	 * Copy of the parameter value (as String). If keepValues() is invoked, all values from request
	 * are kept in an internal copy. This may be used when the Values are intended to be valid
	 * between a request.
	 */
	protected Hashtable<String, String>	_copy_parameter_value;

	/**
	 * Copy of the parameter values (as String[]). If keepValues() is invoked, all values from
	 * request are kept in an internal copy. This may be used when the values are intended to be
	 * valid between a request.
	 */
	
	protected Hashtable<String, String[]>	_copy_parameter_values;

	/**
	 * Constructor.
	 *
	 * @param request the HTTP request
	 * @param string  the reference to the LyString
	 */
	public LyWebParameterValues(HttpServletRequest request, LyString string)
	{
		_request = request;
		_string = string;
		_copy_parameter_names = null;
		_copy_parameter_value = null;
		_copy_parameter_values = null;
	}

	/**
	 * Returns a string containing the lone value of the specified parameter, or null if the
	 * parameter does not exist.
	 */
	public String getParameter(String param)
	{
		String parameters;

		if (_copy_parameter_value != null)
			parameters = (String)_copy_parameter_value.get(param);
		else
			parameters = _request.getParameter(param);

		parameters = LyString.replace(parameters, String.valueOf(HTML_SEP),
				String.valueOf(LyTools.OBJECT_SEP));
		return parameters;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyParameterValues#getParameterNames()
	 */
	public String[] getParameterNames()
	{
		if (_copy_parameter_names != null)
			return _copy_parameter_names;

		Enumeration<?>    parameterNames = _request.getParameterNames();
		Vector<String> result = new Vector<String>();

		while (parameterNames.hasMoreElements())
			result.addElement((String)parameterNames.nextElement());

		// Build list
		String[] names = new String[result.size()];

		result.copyInto(names);
		return names;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyParameterValues#getParameterValues(java.lang.String)
	 */
	public String[] getParameterValues(String param)
	{
		if (_copy_parameter_values != null)
			return (String[])_copy_parameter_values.get(param);

		String[] values = _request.getParameterValues(param);

		if (values == null)
			return null;

		if (values.length == 1)
		{
			if ((values[0] == null) || (values[0].length() == 0))
				return null;

			return LyString.stringToList(values[0], HTML_SEP);
		}

		return values;
	}

	/**
	 * Forces to keep a copy of internal values.
	 */
	public void keepValues()
	{
		_copy_parameter_names = getParameterNames();
		if ((_copy_parameter_names != null) && (_copy_parameter_names.length > 0))
		{
			Hashtable<String, String>   value = new Hashtable<String, String>();
			Hashtable<String, String[]> values = new Hashtable<String, String[]>();
			String					    name, val;
			String[]				    vals;

			for (int i = 0; i < _copy_parameter_names.length; i++)
			{
				name = _copy_parameter_names[i];
				val = getParameter(name);
				vals = getParameterValues(name);
				if (val != null)
					value.put(name, val);
				if (vals != null)
					values.put(name, vals);
			}

			_copy_parameter_value = value;
			_copy_parameter_values = values;
		}
		else
		{
			_copy_parameter_value = null;
			_copy_parameter_values = null;
		}
	} // end method keepValues

    /**
     * (non-Javadoc).
     *
     * @see leon.view.event.LyParameterValues#updateViewFromValues(leon.view.LyComponent)
     */
    public void updateViewFromValues(LyComponent component)
    {
        updateViewFromValues(component, false);
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.event.LyParameterValues#updateViewFromValues(leon.view.LyComponent, boolean)
     */
    public void updateViewFromValues(LyComponent component, boolean needValueChange)
	{
    	updateViewFromValues(component, null, needValueChange);
	}
    
    /**
     * (non-Javadoc).
     *
     * @see leon.view.event.LyParameterValues#updateViewFromValues(leon.view.LyComponent, boolean)
     */
    public void updateViewFromValues(LyComponent component, LyComponent priorityComponent, boolean needValueChange)
	{
		String[]			   parameterNames = getParameterNames();
		int					   i, n = parameterNames.length;
		String				   parameterName;
		String[]			   parameterValues;
		LyComponent			   subComponent = null;
		LyFieldItem			   fieldItem;
		LySetLabelInterface    setLabel;
		String				   realParam;
		String				   parameterValue = null;
		boolean				   status;
		int					   nParameters;
		Vector<String>		   selList;
		LyActionDemand		   demand;
		int					   demandType = LyActionDemand.TYPE_VALUE_CHANGED;
		List<String>		   skippedParameters = new ArrayList<String>();
		LyView				   view = component.getParentView();
		int					   j, l;
		boolean				   changed;
		String				   label;
		LyBoolean.Status	   state;
		double				   d;
		Vector<LyActionDemand> demands = null;
		Vector<LyComponent>    changedComponents = null;

		skippedParameters.add(LyWebConstants.HIDDEN);
		skippedParameters.add(LyWebConstants.SOURCE);
		skippedParameters.add(LyWebConstants.CONTROLLER);
		skippedParameters.add(LyWebConstants.VALUE);
		skippedParameters.add(LyWebConstants.RELOAD);
		skippedParameters.add("reload");
		skippedParameters.add(LyWebConstants.FRAME);
		skippedParameters.add(LyWebConstants.EXTEND);
		skippedParameters.add(LyWebConstants.POSITION_X);
		skippedParameters.add(LyWebConstants.POSITION_Y);
		skippedParameters.add(LyWebConstants.COMMAND);
		skippedParameters.add(LyWebConstants.PAGE);
		skippedParameters.add(LyWebConstants.TYPE);
		skippedParameters.add(LyWebConstants.TARGET_CONTROLLER);
		skippedParameters.add(LyWebConstants.SUBMITTED);
		skippedParameters.add(LyWebConstants.INDEX);
		skippedParameters.add(LyWebConstants.TARGET_WINDOW);
		skippedParameters.add(LyWebConstants.AJAX);

		for (i = 0; i < n; i++)
		{
			demand = null;
			setLabel = null;
			subComponent = null;
			changed = false;

			parameterName = parameterNames[i];
			parameterValues = getParameterValues(parameterName);

			// Manage toggle tools
			if ((parameterName.equals(LyWebConstants.SOURCE))
					&& (parameterValues != null)
					&& (parameterValues.length == 1))
			{
				LyComponent sourceComponent = component.getComponent(parameterValues[0]);

				if ((sourceComponent instanceof LyTool)
						&& (((LyTool)sourceComponent).getStyle() == LyTool.Style.TOGGLE))
					((LyTool)sourceComponent).setPressed(!((LyTool)sourceComponent).isPressed());
			}

			// Get rid of dummy parameters (takes a long time to search for
			// nothing)
			if (skippedParameters.contains(parameterName))
				continue;
			if (parameterName.startsWith("LySelectionItem."))
				continue;
			if (parameterName.endsWith("_date"))
			{
				realParam = parameterName.substring(0, parameterName.length() - 5);
				if (getParameter(realParam) != null)
					continue; // Dummy duplicate of realParam
			}
			if (parameterName.endsWith("File"))
			{
				realParam = parameterName.substring(0, parameterName.length() - 4);
				if (getParameter(realParam) != null)
					continue; // Dummy duplicate of realParam
			}
			if (parameterName.endsWith("_empty_value") && (parameterValues != null))
			{
				parameterName = parameterName.substring(0, parameterName.length() - 12);
				parameterValue = null;
				parameterValues = null;
			}

			if (priorityComponent != null)
				subComponent = priorityComponent.getComponent(parameterName);
			
			// Find the component with the id of the parameter
			// First of all, look on top level (optimization) : fieldsinfos are
			// stored here
			// which avoid a deep recursive search in the tree
			if ((subComponent == null) && (view != null))
				subComponent = view.getChildFromCache(parameterName);

			if (subComponent == null)
				subComponent = component.getComponent(parameterName);

			if (subComponent instanceof LySetLabelInterface)
				setLabel = (LySetLabelInterface)subComponent;
			else if (subComponent instanceof LyFormField)
			{
				demand = new LyActionDemand(demandType, parameterName);
				demand.setParameterValues(this);

				// If form field retrieve the field item
				fieldItem = ((LyFormField)subComponent).getFieldItem();

				if (fieldItem instanceof LySetLabelInterface)
					setLabel = (LySetLabelInterface)fieldItem;
				else
				{
					nParameters = ((parameterValues == null) ? 0 : parameterValues.length);
					if (nParameters != 0)
					{
						// create a vector of selected items
						selList = new Vector<String>(nParameters);
						for (j = 0; j < nParameters; j++)
							selList.addElement(parameterValues[j]);
					}
					else
						selList = new Vector<String>();

					if (fieldItem instanceof LyRelation)
					{
						if (fieldItem.getInternalView() instanceof LyWebObject)
							((LyWebObject)fieldItem.getInternalView()).setUpdating(true);

						// Put selected links at the end of the list to keep selection order
						LyLink		   link = null;
						int			   ns = selList.size();
						int			   k, nl = ((LyRelation)fieldItem).getLinkCount();
						Vector<LyLink> selectedOrderedLinks = new Vector<LyLink>(ns);
						String		   selElement;

						// First pass : retrieve selected links in order
						for (j = 0; j < ns; j++)
						{
							selElement = selList.elementAt(j);
							
							// Value sent by navigator may contain the '\n' character. Replace it by the
							// '\' and 'n' characters used in link identifiers.
							if ((selElement != null) && (selElement.indexOf('\n') >= 0))
								selElement = LyString.replace(selElement, "\n", "\\n");

							if ((selElement != null) && (selElement.indexOf(HTML_LINK_CR_SEP) >= 0))
								selElement = LyString.replace(selElement, HTML_LINK_CR_SEP, "\\n");
							
							for (k = 0; k < nl; k++)
							{
								link = ((LyRelation)fieldItem).getLink(k);
								if (link.getId().equals(selElement))
								{
									selectedOrderedLinks.addElement(link);
									break;
								}
							}
						}

						// Retrieve old selection list in order
						Vector<LyLink> oldSelectedOrderedLinks = new Vector<LyLink>(nl);

						for (k = 0; k < nl; k++)
						{
							link = ((LyRelation)fieldItem).getLink(k);
							if (link.isSelected())
								oldSelectedOrderedLinks.addElement(link);
						}

						// Reset link status only if something has changed in the selection
						if (!selectedOrderedLinks.equals(oldSelectedOrderedLinks))
						{
							// Second pass : unselect links that are not selected
							for (k = 0; k < nl; k++)
							{
								link = ((LyRelation)fieldItem).getLink(k);

								if (!selectedOrderedLinks.contains(link))
									link.setSelected(false);
							}

							// Third pass : select links in order
							nl = selectedOrderedLinks.size();

							for (k = 0; k < nl; k++)
							{
								link = selectedOrderedLinks.elementAt(k);
								link.setSelected(true);

								// If relation is multiple, put selected links at the end of the relation
								if (((LyRelation)fieldItem).isMultiple())
								{
									((LyRelation)fieldItem).removeLink(link);
									((LyRelation)fieldItem).addLink(link);
								}
							}

							changed = true;
						}
						
						if (fieldItem.getInternalView() instanceof LyWebObject)
							((LyWebObject)fieldItem.getInternalView()).setUpdating(false);
					}
					else if (fieldItem instanceof LyChoice)
					{
						int		 nChoice = ((LyChoice)fieldItem).getOptionCount();
						LyOption option;

						for (j = 0; j < nChoice; j++)
						{
							option = ((LyChoice)fieldItem).getOption(j);
							status = selList.contains(option.getId());
							if (option.isChecked() != status)
							{
								option.setChecked(status);
								changed = true;
							}
						}
					}
					else if (fieldItem instanceof LyList)
					{
						int			 nList = ((LyList)fieldItem).getListOptionCount();
						LyListOption option;

						for (j = 0; j < nList; j++)
						{
							option = ((LyList)fieldItem).getListOption(j);
							status = selList.contains(option.getId());
							if (option.isSelected() != status)
							{
								option.setSelected(status);
								changed = true;
							}
						}
					}
					else if (fieldItem instanceof LyBoolean)
					{
						if (selList.size() == 0)
							state = LyBoolean.Status.EMPTY;
						else if (selList.elementAt(0).equals(LyBoolean.Status.CHECKED.toString()))
							state = LyBoolean.Status.CHECKED;
						else if (selList.elementAt(0).equals(LyBoolean.Status.EMPTY.toString()))
							state = LyBoolean.Status.EMPTY;
						else
							state = LyBoolean.Status.UNCHECKED;

						if (state != ((LyBoolean)fieldItem).getStatus())
						{
							((LyBoolean)fieldItem).setStatus(state);
							changed = true;
						}
					}
					else if (fieldItem instanceof LyDictionary)
					{
						Vector<LyEntry> result = new Vector<LyEntry>();
						LyEntry		    entry;

						((LyDictionary)fieldItem).getComponents(LyEntry.class, result);
						l = result.size();

						for (j = 0; j < l; j++)
						{
							entry = result.elementAt(j);
							changed |= updateEntryFromValues(selList, entry);
						}
					}
				}
			}

			if ((setLabel instanceof LyText)
					&& ((LyText)setLabel).isSecret()
					&& (parameterValues != null))
			{
				parameterValue = LyString.listToString(parameterValues, _string.SEP);
				// XXX : SEP ?
				if ((parameterValue != null)
						&& (parameterValue.trim().length() > 0)
						&& (parameterValue.replace('*', ' ').trim().length() == 0))
					setLabel = null; // secret field has not been modified
			}
			
			// Manage the checkbox generated with a form field with the isCheckbox=true
			if ((subComponent == null) && parameterName.startsWith("_checkbox_state_"))
			{
				if ((parameterValues != null) && (parameterValues.length > 0))
				{
					String		fieldId = parameterName.substring("_checkbox_state_".length());
					LyComponent formField = component.getComponent(fieldId);
					
					if (formField instanceof LyFormField)
						((LyFormField)formField).setChecked(Boolean.parseBoolean(parameterValues[0]));
				}
			}

			// File case : set value to null if checkbox used to empty file is checked.
			if ((setLabel instanceof LyFile)
					&& (parameterValues != null))
			{
				String[] result = getParameterValues(parameterName + "_empty_value");

				if (result != null)
				{
					parameterValue = null;
					parameterValues = null;
				}
				else
				{
					// Check if file needs to be updated
					setLabel = needFileUpdate((LyFile)setLabel);
				}
			}

			if (setLabel != null)
			{
				demand = new LyActionDemand(demandType, parameterName);
				demand.setParameterValues(this);

				if (parameterValues != null)
				{
					if ((parameterValues.length >= 2) && (setLabel instanceof LyDate))
					{
						parameterValue = parameterValues[0] + " " + parameterValues[1];
						
						if (parameterValues.length > 2)
							parameterValue += " " + parameterValues[2];
					}
					else if ((parameterValues.length > 0)
							&& ((setLabel instanceof LyText)
									|| (setLabel instanceof LyDate)
									|| (setLabel instanceof LyNumber)))
						parameterValue = parameterValues[0];
					else
						parameterValue = LyString.listToString(parameterValues, _string.SEP);
				}
				// XXX : SEP ?
				else
					parameterValue = null;

				if (parameterValue != null)
				{
					if ((setLabel instanceof LyText)
							&& (((LyText)setLabel).getNRows() > 1))
						// Remove '\r' from multi-line text fields
						parameterValue = LyString.replace(parameterValue, "\r", "");
					else if ((setLabel instanceof LyDate)
							&& (parameterValue.length() == 0))
						// Date case : if value is "", use null instead
						parameterValue = null;
					else if ((setLabel instanceof LyNumber)
							&& (parameterValue.indexOf(',') > 0))
					{
						parameterValue = parameterValue.replace(',', '.');
						if(LyTools.isDouble(parameterValue))
						{
							d = Double.parseDouble(parameterValue);
							parameterValue = Double.toString(d);
						}
					}
				}

				if (setLabel instanceof LyGetLabelInterface)
				{
					label = ((LyGetLabelInterface)setLabel).getLabel();
					if (((label == null)
								&& (parameterValue != null))
							|| ((label != null)
								&& (parameterValue == null))
							|| ((label != null)
								&& (parameterValue != null)
								&& !label.equals(parameterValue)))
						changed = true;
				}
				
				if ((setLabel instanceof LyComponent)
						&& (((LyComponent)setLabel).getInternalView() instanceof LyWebObject))
					((LyWebObject)((LyComponent)setLabel).getInternalView()).setUpdating(true);

				setLabel.setLabel(parameterValue);
				
				if ((setLabel instanceof LyComponent)
						&& (((LyComponent)setLabel).getInternalView() instanceof LyWebObject))
					((LyWebObject)((LyComponent)setLabel).getInternalView()).setUpdating(false);
			}

			if (needValueChange
                && changed
                && (demand != null)
                && (subComponent != null))
			{
				if (demands == null)
				{
					demands = new Vector<LyActionDemand>(1);
					changedComponents = new Vector<LyComponent>(1);
				}
				demands.add(demand);
				changedComponents.add(subComponent);
			}
		}

		if (demands == null)
			return;

		// Post demands when all changes are identified before other indirect changes
		LyViewManager vm = component.getViewManager();
		String		  rootControllerId;

		for (i = 0, n = demands.size(); i < n; i++)
		{
			demand = demands.elementAt(i);
			subComponent = changedComponents.elementAt(i);

			rootControllerId = subComponent.getRootController();
			vm.processDemand(rootControllerId, demand);
        }
	} // end method updateViewFromValues

    /**
     * Checks if a file must be updated by the updateViewFromValues process.<br />
     * If file must be updated, this method must return the given file. Otherwise, it must
     * return null.
     * 
     * @param file the file component that is being checked.
     * @return the file component itself if it has to be updated, null otherwise.
     */
    protected LyFile needFileUpdate(LyFile file)
    {
    	return file;
    }
    
	/**
	 * Updates the given entry: check the entry and all its inner entries who are listed in <code>
	 * list</code>.
	 *
	 * @param  list  list of checked entries
	 * @param  entry entry to update
	 * @return True if changed, false otherwise.
	 */
	private boolean updateEntryFromValues(Vector<String> list, LyEntry entry)
	{
		int     i, n = entry.getEntryCount();
		LyEntry innerEntry;
		boolean checked = list.contains(entry.getId());
		boolean changed = false;

		if (entry.isChecked() != checked)
		{
			entry.setChecked(checked);
			changed = true;
		}

		for (i = 0; i < n; i++)
		{
			innerEntry = entry.getEntry(i);
			changed |= updateEntryFromValues(list, innerEntry);
		}

		return changed;
	}
} // end class LyWebParameterValues
