/*
 * LyWebViewManager.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;

import javax.servlet.*;
import javax.servlet.http.*;

import leon.app.*;
import leon.control.*;
import leon.misc.*;
import leon.view.*;
import leon.view.common.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.setinterface.*;
import leon.view.web.struts.*;


/**
 * Communication control layer --> View Will implement remote update notifications.
 *
 * @version $Id: LyWebViewManager.java 41344 2012-10-25 15:45:02Z nhamon $
 * @author  Lyria
 */
public abstract class LyWebViewManager extends LyViewManager
{
	public static final String						  __VERSION =
		"$Id: LyWebViewManager.java 41344 2012-10-25 15:45:02Z nhamon $";

	/**
	 * The type of a message that have no type. Other types of message can be :
	 *
	 * <ul>
	 *   <li>{@link #TYPE_ALERT}</li>
	 *   <li>{@link #TYPE_QUESTION}</li>
	 *   <li>{@link #TYPE_LOCATION}</li>
	 *   <li>{@link #TYPE_LOCATION_NEW_WINDOW}</li>
	 *   <li>{@link #TYPE_SHOW_FILE}</li>
	 * </ul>
	 */
	public final static int							  TYPE_NO = 0;

	/**
	 * The alert message type. Indicates that the message is an alert.
	 *
	 * @see #TYPE_NO
	 */
	public final static int							  TYPE_ALERT = 1;

	/**
	 * The question message type. Indicates that the message shown to the user is a question.
	 *
	 * @see #TYPE_NO
	 */
	public final static int							  TYPE_QUESTION = 2;

	/**
	 * The location in current window message type. Those messages are used to show a location in
	 * the window to the user.
	 *
	 * @see #TYPE_NO
	 */
	public final static int							  TYPE_LOCATION = 3;

	/**
	 * The location in new window message type. Those messages are used to show a location in a new
	 * window to the user.
	 *
	 * @see #TYPE_NO
	 */
	public final static int							  TYPE_LOCATION_NEW_WINDOW = 4;

	/**
	 * The show file message type. Indicates that the message have to show a file.
	 *
	 * @see #TYPE_NO
	 */
	public final static int							  TYPE_SHOW_FILE = 5;

	/**
	 * The MicroSoft Internet Explorer browser.
	 *
	 * @see #BROWSER_OTHER
	 */
	public final static int							  BROWSER_MSIE = 0;

	/**
	 * The Netscape browser.
	 *
	 * @see #BROWSER_OTHER
	 */
	public final static int							  BROWSER_NETSCAPE = 1;

	/**
	 * The Opera browser.
	 *
	 * @see #BROWSER_OTHER
	 */
	public final static int							  BROWSER_OPERA = 2;

	/**
	 * Unspecified browser.<br>
	 * Means other browsers than:
	 *
	 * <ul>
	 *   <li>{@link #BROWSER_FIREFOX}</li>
	 *   <li>{@link #BROWSER_MSIE}</li>
	 *   <li>{@link #BROWSER_NETSCAPE}</li>
	 *   <li>{@link #BROWSER_OPERA}</li>
	 * </ul>
	 */
	public final static int							  BROWSER_OTHER = 3;

	/**
	 * The Firefox browser.
	 *
	 * @see #BROWSER_OTHER
	 */
	public final static int							  BROWSER_FIREFOX = 4;

	/**
	 * The Chrome browser.
	 *
	 * @see #BROWSER_OTHER
	 */
	public final static int							  BROWSER_CHROME = 5;

	/** The MicroSoft Internet Explorer browser identifier. */
	public final static String						  BROWSER_MSIE_ID = "MSIE";

	/** The Firefox browser identifier. */
	public final static String						  BROWSER_FIREFOX_ID = "FIREFOX";

	/** The Chrome browser identifier. */
	public final static String						  BROWSER_CHROME_ID = "CHROME";

	/** The Netscape browser identifier. */
	public final static String						  BROWSER_NETSCAPE_ID = "MOZILLA";

	/** The Opera browser identifier. */
	public final static String						  BROWSER_OPERA_ID = "OPERA";

	/** For show location : indicates the name of the newly opened window. */
	public static final short						  SHOW_LOCATION_PARAMETER_NAME = 5;

	/** For show location : indicates whether scrollbars must be set in newly opened window. */
	public static final short						  SHOW_LOCATION_PARAMETER_SCROLLBARS = 6;

	/** For show location : indicates whether newly opened window must be resizable. */
	public static final short						  SHOW_LOCATION_PARAMETER_RESIZABLE = 7;

	/** For show location : indicates whether toolbar must be set in newly opened window. */
	public static final short						  SHOW_LOCATION_PARAMETER_TOOLBAR = 8;

	/** For show location : indicates whether statusbar must be set in newly opened window. */
	public static final short						  SHOW_LOCATION_PARAMETER_STATUSBAR = 9;

	/** For show location : indicates whether menubar must be set in newly opened window. */
	public static final short						  SHOW_LOCATION_PARAMETER_MENUBAR = 10;

	/** For show location : indicates whether directories must be set in newly opened window. */
	public static final short						  SHOW_LOCATION_PARAMETER_DIRECTORIES = 11;

	/** For show location : indicates whether location must be set in newly opened window. */
	public static final short						  SHOW_LOCATION_PARAMETER_LOCATION = 12;

	/**
	 * The page writer id separator : used within a frame name to introduce the page writer
	 * identifier.
	 */
	public final static String						  SEP_PW_ID = "_#PW_ID#_";

	/** Separator used within a message text to introduce parameters. */
	public final static String						  PARAMETERS = "_#PARAMETERS#_";

	/** Indicates whether page writers must be registered on view managers. */
	public static boolean							  REGISTER_PAGE_WRITERS = true;

	/** Indicates whether the TABLE HTML tags must be optimized. */
	public static boolean							  OPTIMIZE_TABLE_GENERATION = false;
	
	/**
	 * Indicates the size (in pixel) used for a new window
	 * if nothing is specified for width or height.
	 */
	public static int								  DEFAULT_POPUP_WIDTH = 800;
	public static int								  DEFAULT_POPUP_HEIGHT = 500;
	
    /**
     * Indicates that we open a new window when there is a file to show. 
     * If set to false, show the window in the target frame.
     */
    public static boolean	  						  WEB_FILE_NEW_WINDOW = true;
    
    /**
     * Maximum size of the images that may be cached (in pixels).
     */
    public static int								  CACHED_IMAGE_MAX_SIZE = 200000;

	/** List of page writers identified by their ids. */
	protected Hashtable<String, LyWebPageWriter>	  _pageWritersId;

	/** List of page writers identified by their thread. */
	protected Hashtable<Thread, LyWebPageWriter>	  _pageWriters;

	/** List of messages identified by their thread. */
	protected Hashtable<Object, Vector<_LyWebMessage>>				  _messages;

	/** List of associations between messages and answers. */
	protected Hashtable<String, _LyMessageAnswerInfo> _messageAnswers;

	/** List of associations between thread and close requests. */
	protected Hashtable<Thread, Boolean>			  _closeRequests;

	/** List of associations between thread and window names. */
	protected Hashtable<Thread, String>				  _windowNames;

	/** Indicates if the view manager reload mode is true or false. */
	protected boolean								  _reload;

	/** localhost root URL. */
	public String									  _root_url;
	
	/** Mobile component web view URL. */
	public String									  _mobileStartUrl;

	/** List of frames (html) which need refresh. */
	protected Vector<String>						  _reloadFrames;

	/**
	 * Mapping between a frame and its target view (default is supposed in main view if not
	 * present).
	 */
	protected HashMap<String, String>				  _framesPopupViewMapping;

	/** List of views which require the auto refresh mode. */
	protected Vector<LyGetRefreshInterface>			  _refreshingViews;

	/**
	 * List of additional scripts used during a view refresh. For example the script used (during
	 * refresh) to close all pop up windows.
	 */
	private Vector<String>						      _additionalRefreshScripts;

	/** Temporary buffer containing additional refresh scripts. */
	private Vector<String>							  _additionalRefreshScriptsBuffer;

	/** Scripts lock. */
    protected Object                                  _refreshScriptsLock = new Object();

	/** refreshingViews lock. */
	protected Object                                  _refreshingViewsLock = new Object();

	/**
	 * Current browser name. Possible values are :
	 *
	 * <ul>
	 *   <li>{@link #BROWSER_FIREFOX}</li>
	 *   <li>{@link #BROWSER_MSIE}</li>
	 *   <li>{@link #BROWSER_NETSCAPE}</li>
	 *   <li>{@link #BROWSER_OPERA}</li>
	 *   <li>{@link #BROWSER_OTHER}</li>
	 * </ul>
	 */
	protected int									  _browser;

	/** Current browser version. */
	protected double								  _browserVersion;

	/** Indicates if any specific frame has been found. */
	protected boolean								  _foundSpecificFrame;

	/** View manager decoration. */
	protected LyWebDecoration						  _deco;

	/** Indicates if DIV tags may be written. */
	protected boolean								  _writeDivs;

	/** Indicates if scrolling attribute must be generated in iframe tags. */
	protected boolean								  _forceScrolling;

	/** Indicates if popup view is allowed. */
	protected boolean								  _usePopupView;

	/** View controller that was last displayed in the popup view. */
	protected LyViewController						  _lastPopupViewController;

	/** The target code (use in multi-windows mode). */
	private String									  _targetCode;

	/**
	 * Constructor.
	 *
	 * @param environment the application environment
	 */
	public LyWebViewManager(LyEnvironment environment)
	{
		super(environment);

		_pageWritersId = null;
		_pageWriters = null;
		_messages = null;
		_messageAnswers = null;
		_closeRequests = null;
		_windowNames = null;
		_reload = false;
		_reloadFrames = null;
		_framesPopupViewMapping = null;
		_refreshingViews = null;
		_browser = BROWSER_OTHER;
		_browserVersion = -1.0;
		_foundSpecificFrame = false;
		_deco = null;
		_writeDivs = true;
		_forceScrolling = false;
		_usePopupView = true;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.misc.LyViewManagerInterface#free()
	 */
	public void free()
	{
        synchronized (_refreshScriptsLock)
        {
            _additionalRefreshScripts = null;
            _additionalRefreshScriptsBuffer = null;
        }

		super.free();

		if (_pageWritersId != null)
			_pageWritersId.clear();

		_pageWritersId = null;
		_pageWriters = null;
		_messages = null;
		_messageAnswers = null;
		_closeRequests = null;
		_windowNames = null;
		_root_url = null;
		_mobileStartUrl = null;
		_reloadFrames = null;
		_framesPopupViewMapping = null;
		_refreshingViews = null;
		_deco = null;
	}

	/**
	 * Sets view manager properties such as MANAGE_PRINT, MANAGE_TIMEOUT...
	 */
	protected void setProperties()
	{
		super.setProperties();

		setProperty("MANAGE_PRINT");
		setProperty("HANDLE_TIMEOUT");
		setProperty("NO_ASYNC_PRINT"); // Not use asynchrone for web
		setProperty("RE_ORDER_EXPORT");
		setProperty("ONLY_TARGET_POPUP");
		setProperty("MANAGE_URL"); // The web can display url correctly (via
		// the browser)
		setProperty("ASK_MULTIPLE_OBJECTS_DELETION"); // The web does not ask
		// for each object when multiple deletion
		setProperty("MANAGE_ATTACHMENT"); // We can upload files from a web browser.
		setProperty("WEB_MANAGER");
		setProperty("DELAYED_CLOSING");
		setProperty("LAYOUT_MANAGER"); // manage its layout itself
		setProperty("NOTFORMOBILE");
	}

	/**
	 * Builds an instance of LyWebDecoration corresponding to current view manager.
	 *
	 * @param  environment the environment of the web decoration.
	 * @return an instance of web decoration corresponding to current view manager.
	 */
	public LyWebDecoration buildWebDecoration(LyEnvironment environment)
	{
		// Try to load view files
		if (_beans == null)
			loadBeans();

		String className = _beans.getString("WEB_DECORATION");

		if ((className == null) || (className.length() == 0))
			className = "leon.view.web.LyWebDecoration";

		Class<?>[]  types = new Class[1];
		Object[]    parameters = new Object[1];

		types[0] = LyEnvironment.class;
		parameters[0] = environment;

		LyWebDecoration decoration = null;

		try
		{
			Class<?>	   viewClass = Class.forName(className);
			Constructor<?> constructor = viewClass.getConstructor(types);

			decoration = (LyWebDecoration)constructor.newInstance(parameters);
		}
		catch (InvocationTargetException e1)
		{
			_environment.writeException(e1.getTargetException());
			_environment.writeException(e1);
			throw new IllegalArgumentException("BF AppEngine : can't build web decoration class : "
				+ className);
		}
		catch (Exception e)
		{
			_environment.writeException(e);
			throw new IllegalArgumentException("BF AppEngine : can't build web decoration class : "
				+ className);
		}

		return decoration;
	} // end method buildWebDecoration

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.view.LyViewManager#buildInternalView(leon.view.LyComponent)
	 */
	public LyViewComponent buildInternalView(LyComponent component)
	{
		// Check if field item and if a bean must be used
		if (component instanceof LyFieldItem)
		{
			LyFieldItem item = (LyFieldItem)component;

			if (item.getBean() != null)
			{
				LyViewComponent bean = item.buildBean();

				if (bean instanceof LyWebComponent)
					return bean;
				else if (bean != null)
					_environment.logWarning("Can't use bean " + bean.getClass().getName()
						+ " because it's not a LyWebComponent");
			}
		}

		if (_beans == null)
			// try to load view files
			loadBeans();

		if (_beans != null)
		{
			String tag = LyViewNodeFactory.getTag(component.getClass().getName());

			// Special case for treemaps in HTML mode (using DHTML) : retrieve
			// TREEMAP and TREEMAP_REGION tags
			if ((component instanceof LyMap) || (component instanceof LyMapRegion))
			{
				LyComponent parent = component;

				while (!(parent instanceof LyMap) && (parent != null))
					parent = parent.getParent();

				if (parent instanceof LyMap)
				{
					LyMap    map = (LyMap)parent;
					LyLayout layout = map.getLayout();

					if ((layout != null) && (layout.getStyle() == LyLayout.Style.TREEMAP))
						tag = "TREE" + tag;
				}
			}

			String beanClassName = _beans.getString(tag);

			if (beanClassName != null)
			{
				if (beanClassName.length() == 0)
					return null;
				else
					return buildInternalView(beanClassName, component);
			}

			// Try to find bean class name on the super class.
			Class<?>  superClass = component.getClass().getSuperclass();

			while (superClass != null)
			{
				String parentTag = LyViewNodeFactory.getTag(superClass.getName());
				String parentBeanClassName = _beans.getString(parentTag);

				if (parentBeanClassName != null)
				{
					if (parentBeanClassName.length() == 0)
						return null;
					else
						return buildInternalView(parentBeanClassName, component);
				}

				superClass = superClass.getSuperclass();
			}
		}

		if ((component instanceof LyItem) || (component instanceof LyMenu)
				|| (component instanceof LyMenuBar) || (component instanceof LyPopupMenu))
			return null;

		return buildInternalView("leon.view.web", "LyWeb", component);
	} // end method buildInternalView

	/**
	 * Load views properties files.
	 */
	protected abstract void loadBeans();

	/**
	 * Load beans depending on navigator.
	 *
	 * @param beansFile file describing beans
	 * @see   #addBeans(String)
	 */
	protected void loadBeans(String beansFile)
	{
		addBeans(beansFile);

		if (_browser == BROWSER_MSIE)
			addBeans(beansFile + "_ie");
		else if (_browser == BROWSER_NETSCAPE)
			addBeans(beansFile + "_ns");
	}

	/**
	 * Utility method to add extra beans.
	 *
	 * @param beansFile file describing beans
	 */
	protected void addBeans(String beansFile)
	{
		// Add all values in _beans
		LyResources beans = new LyResources(_environment, beansFile);

		if (_beans == null)
		{
			_beans = beans;
			return;
		}

		String                key, value;
		Enumeration<String>   e = beans.getKeys(true);

		while (e.hasMoreElements())
		{
			key = e.nextElement();
			value = beans.get(key);
			_beans.put(key, value);
		}

		beans.free();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.misc.LyViewManagerInterface#clearTemporaryResources()
	 */
	public void clearTemporaryResources()
	{
		// Clear all previous temp directory
		String tmpDir = getTemporaryDir(); // FFT01251
		File   dir = new File(tmpDir);

		if ((dir == null) || !dir.exists())
			return;

		String[] files = dir.list();
		int		 i, n = files.length;
		File     tmpFile;

		for (i = n - 1; i >= 0; i--)
		{
			tmpFile = new File(tmpDir + "/" + files[i]);
			if (!tmpFile.delete())
				_environment.logWarning("File :" + tmpDir + "/" + files[i] + " cannot be deleted");
		}
	}

	/**
	 * Returns the directory that contains the Javascript files.
	 *
	 * @return the directory that contains the Javascript files.
	 */
	public String getScriptsDir()
	{
		String	scriptRoot = _environment.getEnv("LY_LEON_DOC") + "/scripts";

		return scriptRoot;
	}

	/**
	 * Returns the path of the directory that contains the Javascript files.
	 *
	 * @return the path of the directory that contains the Javascript files.
	 */
	public String getScriptsDocDir()
	{
		String	scriptRoot = _environment.getEnv("LY_LEON_DOC_DIR") + "/scripts";

		return scriptRoot;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.misc.LyViewManagerInterface#getTemporaryLocation()
	 */
	public String getTemporaryLocation()
	{
		return _environment.getUrl("LY_TEMP_DOC");
	}

	/**
	 * Returns the name of the resource to be used for default page.
	 */
	public String getDefaultPageResource()
	{
		return "LY_DEFAULT_PAGE";
	}

	/**
	 * Generates a string containing all the Javascript files inclusion.
	 */
	public String generateScriptsInclusion()
	{
		String			scriptRootDir = getScriptsDocDir();
		String			scriptRoot = getScriptsDir();
        File            scriptDir = new File(scriptRootDir);
        File            scriptFull = new File(scriptRootDir + File.separator + "LyFullScript.js");
		StringBuffer	buffer = new StringBuffer();
		String			versionString = "?_version=" + getEnvironment().translate("LY_PRODUCT_VERSION");

		if (scriptFull.exists())
		{
            buffer.append("<script charset='");
            buffer.append(LyString.DEFAULT_CHARSET);
            buffer.append("' src='");
            buffer.append(scriptRoot);
            buffer.append("/LyFullScript.js");
            buffer.append(versionString);
            buffer.append("'></script>\n");
            return buffer.toString();
		}

		if (scriptDir.exists()
		    && scriptDir.isDirectory())
		{
			String[]	files = scriptDir.list(new FilenameFilter() {

				public boolean accept(File dir, String name)
				{
					return ((name != null) && name.endsWith(".js"));
				}
			});

			int			i, n = (files == null ? 0 : files.length);

			for (i = 0; i < n; i++)
			{
				buffer.append("<script charset='");
				buffer.append(LyString.DEFAULT_CHARSET);
				buffer.append("' src='");
				buffer.append(scriptRoot);
				buffer.append('/');
				buffer.append(files[i]);
	            buffer.append(versionString);
				buffer.append("'></script>\n");
			}
		}

		return buffer.toString();
	}

	/**
	 * Shows the view wrote in the given page writer.
	 *
	 * @param pageWriter the page writer
	 * @see   #setPageWriter(LyWebPageWriter)
	 */
	public void showView(LyWebPageWriter pageWriter)
	{
		setPageWriter(pageWriter);
	}

	/**
	 * Hides the view wrote in the given page writer.
	 *
	 * @param pageWriter the page writer containing view to hide
	 */
	public void hideView(LyWebPageWriter pageWriter)
	{
		setClose(pageWriter, true);
	}

	/**
	 * To exit an application. This method does nothing.
	 */
	public void exit(int code)
	{
	}

	/**
	 * Sets the Waiting status for all view of the application.<br>
	 * This method does nothing.
	 */
	public void setWaiting(boolean status)
	{
	    if (status)
	        _waitingCounter++;
	    else if (_waitingCounter > 0)
	        _waitingCounter--;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.view.LyViewManager#showWarning(leon.view.LyComponent,
	 *      java.lang.String)
	 */
	public void showWarning(LyComponent source, String message)
	{
		showMessage(source, "LY_WARNING_TITLE", "LY_WARNING_ICON", message, null /* OK */, null);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.view.LyViewManager#showError(leon.view.LyComponent,
	 *      java.lang.String)
	 */
	public void showError(LyComponent source, String message)
	{
		// Workaround for message whish is only information for a web server
		// Not always a bug
		if (message.equals(getEnvironment().getMessage("LY_ER_REMAINING_PENDING_TRANSACTION")))
			getEnvironment().getLogger().warning(message);
		else
			showMessage(source, "LY_ERROR_TITLE", "LY_ERROR_ICON", message, null /* OK */, null);
	}

	/**
	 * Ask a question to the user.
	 *
	 * @see LyViewManager#showWarning(LyComponent, String)
	 */
	public void askQuestion(LyComponent source, String message, LyQuestionAnswerInterface client,
		Object clientData)
	{
		BitSet bs = new BitSet();

		bs.set(LyViewManager.COMMAND_YES);
		bs.set(LyViewManager.COMMAND_NO);

		_LyMessageAnswerInfo info = new _LyMessageAnswerInfo(client, clientData);

		showMessage(source, "LY_QUESTION_TITLE", "LY_QUESTION_ICON", message, bs, info);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.view.LyViewManager#showFile(leon.view.LyComponent,
	 *      java.lang.String)
	 */
	public void showFile(LyComponent source, String location)
	{
		showMessage(source, "LY_WARNING_TITLE", "LY_WARNING_ICON", location, null, null,
			TYPE_SHOW_FILE);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.view.LyViewManager#showLocation(leon.view.LyComponent,
	 *      java.lang.String, boolean)
	 */
	public void showLocation(LyComponent source, String location, boolean newWindow)
	{
		showLocation(source, location, newWindow, null);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.view.LyViewManager#showLocation(leon.view.LyComponent,
	 *      java.lang.String, boolean)
	 */
	public void showLocation(LyComponent source, String location, boolean newWindow, Hashtable<Short, String> parameters)
	{
		String docDir = _environment.getEnv("LY_APP_DOC_DIR");
		// case a file being consulted with an hyperlink.
		String fileRef = "file://";

		if ((location.startsWith(docDir)) || (location.startsWith(fileRef+docDir)))
		{
			String appDoc = _environment.getEnv("LY_APP_DOC");

			// use a relative URL to the document, not file://
			if ((location.startsWith(docDir)))
			{
				location = location.replace('\\', '/');

				location = appDoc+"/"+location.substring(docDir.length()+1);
			}
			else if ((location.startsWith(fileRef+docDir)))
			{
				location = location.replace('\\', '/');

				location = appDoc+"/"+location.substring(fileRef.length()+docDir.length()+1);
			}
		}
		else
		{
			// use a  file:// reference to the document
			if (getBrowser() == BROWSER_FIREFOX)
				location = location.replace('\\', '/');
			if (location.startsWith(fileRef))
				location = location.substring(fileRef.length());
		}

		if (newWindow)
			showMessage(source, "LY_WARNING_TITLE", "LY_WARNING_ICON", location, null, null,
				TYPE_LOCATION_NEW_WINDOW, parameters);
		else
			showMessage(source, "LY_WARNING_TITLE", "LY_WARNING_ICON", location, null, null,
				TYPE_LOCATION);
	}

	protected Vector<String>		_newWindowScripts;

	public Vector<String> getNewWindowScripts()
	{
		return _newWindowScripts;
	}

	public void addNewWindowScript(String script)
	{
		if (script == null)
			return;

		if (_newWindowScripts == null)
			_newWindowScripts = new Vector<String>();

		if (!_newWindowScripts.contains(script))
			_newWindowScripts.addElement(script);
		else
		{
			// Add script at the right position in the list
			_newWindowScripts.removeElement(script);
			_newWindowScripts.addElement(script);
		}
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.view.LyViewManager#showPDFPreview(leon.view.LyComponent,
	 *      java.lang.String)
	 */
	public void showPDFPreview(LyComponent source, String location)
	{
		// location is an absolute path on the server
		// try to get a relative path on the web application on the server
		StringBuffer url = new StringBuffer();
		String		 appDocDir = null;

		if (source != null)
			appDocDir = source.getEnvironment().getEnv("LY_APP_DOC_DIR");

		location = location.replace('\\', '/');

		if (appDocDir != null)
		{
			appDocDir = appDocDir.replace('\\', '/');

			if (location.startsWith(appDocDir))
			{
				url.append(source.getEnvironment().getEnv("LY_APP_DOC"));
				url.append(location.substring(appDocDir.length()));
			}
		}
		if (url.length() == 0)
		{
			String root = getRootURL();

			root = root.replace('\\', '/');

			String[] locationPath = LyString.stringToList(location, '/');
			String[] serverPath = LyString.stringToList(root, '/');

			int		 i, j, n = serverPath.length, m = locationPath.length;
			boolean  makeUrl = false;

			for (i = 0; i < n; i++)
			{
				if (serverPath[i].trim().length() == 0)
					continue;
				for (j = 0; j < m; j++)
				{
					if (makeUrl)
					{
						url.append(locationPath[j]);
						if (j < (m - 1))
							url.append('/');
					}

					if (!makeUrl && locationPath[j].equals(serverPath[i]))
						makeUrl = true;
				}

				if (makeUrl)
					break;
			}
		}
		// show the file to the user
		showMessage(source, "LY_ACTION_PRINT", "LY_ACTION_PRINT", url.toString(), null, null,
			getShowPDFPreviewType());
	} // end method showPDFPreview

	/**
	 * Returns the type of message that must be used to show a PDF preview.
	 */
	protected int getShowPDFPreviewType()
	{
		return TYPE_LOCATION_NEW_WINDOW;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.view.LyViewManager#showMessage(leon.view.LyComponent,
	 *      java.lang.String, java.lang.String, java.lang.String,
	 *      java.util.BitSet, leon.view.setinterface.LyMessageAnswerInterface,
	 *      java.lang.Object)
	 */
	public void showMessage(LyComponent source, String title, String icon, String message,
		BitSet commands, LyMessageAnswerInterface client, Object clientData)
	{
		_LyMessageAnswerInfo info = new _LyMessageAnswerInfo(client, clientData);

		showMessage(source, title, icon, message, commands, info);
	}

	/**
	 * Shows a message to the user.
	 *
	 * @param source   view element which asks the question
	 * @param title    title of the message window
	 * @param icon     icon shown in the message window
	 * @param message  text of the message
	 * @param commands bitset which contains COMMAND_xxx values
	 * @param info     the message answer information : the client registred to receive the response
	 *                 and the client data that freely set data which will be returned with the
	 *                 response.
	 */
	protected void showMessage(LyComponent source, String title, String icon, String message,
		BitSet commands, _LyMessageAnswerInfo info)
	{
		showMessage(source, title, icon, message, commands, info, TYPE_NO);
	}

	/**
	 * Shows a message to the user.
	 *
	 * @param      source   not used any more. Kept for ascendent compatibility
	 * @param      title    title of the message window
	 * @param      icon     icon shown in the message window
	 * @param      message  text of the message
	 * @param      commands bitset which contains COMMAND_xxx values
	 * @param      info     the message answer information : the client registred to receive the
	 *                      response and the client data that freely set data which will be returned
	 *                      with the response.
	 * @param      type     the message type, can be
	 *
	 *                      <ul>
	 *                        <li>{@link #TYPE_ALERT}</li>
	 *                        <li>{@link #TYPE_QUESTION}</li>
	 *                        <li>{@link #TYPE_LOCATION}</li>
	 *                        <li>{@link #TYPE_LOCATION_NEW_WINDOW}</li>
	 *                        <li>{@link #TYPE_SHOW_FILE}</li>
	 *                      </ul>
	 */
	protected void showMessage(LyComponent source, String title, String icon, String message,
		BitSet commands, _LyMessageAnswerInfo info, int type)
	{
		showMessage(source, title, icon, message, commands, info, type, null);
	}

	/**
	 * Shows a message to the user.
	 *
	 * @param      source   not used any more. Kept for ascendent compatibility
	 * @param      title    title of the message window
	 * @param      icon     icon shown in the message window
	 * @param      message  text of the message
	 * @param      commands bitset which contains COMMAND_xxx values
	 * @param      info     the message answer information : the client registred to receive the
	 *                      response and the client data that freely set data which will be returned
	 *                      with the response.
	 * @param      type     the message type, can be
	 *
	 *                      <ul>
	 *                        <li>{@link #TYPE_ALERT}</li>
	 *                        <li>{@link #TYPE_QUESTION}</li>
	 *                        <li>{@link #TYPE_LOCATION}</li>
	 *                        <li>{@link #TYPE_LOCATION_NEW_WINDOW}</li>
	 *                        <li>{@link #TYPE_SHOW_FILE}</li>
	 *                      </ul>
	 */
	protected void showMessage(LyComponent source, String title, String icon, String message,
		BitSet commands, _LyMessageAnswerInfo info, int type, Hashtable<Short, String> parameters)
	{
		String bgColor = null;
		String messageType;

		if ((icon == null) || (icon.equals("LY_WARNING_ICON")))
		{
			bgColor = "LY_WARNING_BACKGROUND";
			messageType = "INFO";
		}
		else if (icon.equals("LY_ERROR_ICON"))
		{
			bgColor = "LY_ERROR_BACKGROUND";
			messageType = "ERROR";
		}
		else if (icon.equals("LY_QUESTION_ICON"))
		{
			bgColor = "LY_QUESTION_BACKGROUND";
			messageType = "QUESTION";
		}
		else
		{
			if (type == TYPE_QUESTION)
				messageType = "QUESTION";
			else if (type == TYPE_ALERT)
				messageType = "ERROR";
			else
				messageType = "INFO";
		}

		message = _environment.translate(message);
		title = _environment.translate(title);

		if (icon != null)
			icon = _environment.getImageUrl(icon);
		else
			icon = "";

		if (commands != null)
			type = TYPE_QUESTION;

		// Give an id to the message answer info and register it
		String infoId = LyString.getRandomId();

		if (info != null)
			addMessageAnswer(info, infoId);

		// Set the message
		setMessage(message, title, icon, type, bgColor, commands, infoId, parameters);

		LyDemandTransmitter transmitter = getDemandTransmitter();

		if (transmitter instanceof LyMessageWindowRegister)
			((LyMessageWindowRegister)transmitter).registerMessageWindow(infoId, messageType,
				message);
	} // end method showMessage

	/**
	 * Adds the mapping between the given message answer <code>info</code> and the id <code>
	 * infoId</code> into the {@link #_messageAnswers} table.
	 *
	 * @param info   the message info
	 * @param infoId the info ID
	 */
	protected void addMessageAnswer(_LyMessageAnswerInfo info, String infoId)
	{
		if (_messageAnswers == null)
			_messageAnswers = new Hashtable<String, _LyMessageAnswerInfo>();

		if (infoId != null)
			_messageAnswers.put(infoId, info);
	}

	/**
	 * Gets the message answer associated to the given id.
	 *
	 * @param  infoId the id associated to the wanted message answer
	 * @return the message answer if it is found, null otherwise
	 */
	protected _LyMessageAnswerInfo getMessageAnswer(String infoId)
	{
		if (_messageAnswers == null)
			return null;

		if (infoId == null)
			return null;

		return _messageAnswers.get(infoId);
	}

	/**
	 * Removes the message answer associated to the given ID.
	 *
	 * @param infoId the id of the message answer to remove
	 */
	public void removeMessageAnswer(String infoId)
	{
		if ((_messageAnswers != null) && (infoId != null))
		{
			_messageAnswers.remove(infoId);
		}

		removeMessages(infoId);
	}

	/**
	 * Removes the message associated to the given ID from the list of messages.
	 * If no ID is given, message of type TYPE_LOCATION, TYPE_LOCATION_NEW_WINDOW, TYPE_NO and TYPE_SHOW_FILE.
	 *
	 * @param infoId the id of the message to remove
	 */
	protected void removeMessages(String infoId)
	{
		// Remove message from list of messages
		if (_messages != null)
		{
			// Remove message from list of messages
			Set<Object>	keys = _messages.keySet();

			for (Object key : keys)
			{
				Vector<_LyWebMessage>	messages = _messages.get(key);

				if (messages == null)
			        return;
				else
				{
					int	messageSize = messages.size();

					for (int messageIndex = messageSize - 1; messageIndex >= 0; messageIndex--)
					{
						_LyWebMessage	message = messages.elementAt(messageIndex);
						boolean			removeMessage = false;

						if ((infoId != null) && infoId.equals(message._infoId))
							removeMessage = true;
						else if ((message._type == TYPE_LOCATION)
								|| (message._type == TYPE_LOCATION_NEW_WINDOW)
								|| (message._type == TYPE_SHOW_FILE))
							removeMessage = true;

						if (removeMessage)
						{
							messages.remove(message);
						}
					}

					if (messages.size() == 0)
					{
						_messages.remove(key);
					}
				}
			}
		}
	}

	/**
	 * Returns the URL that must be used to answer messages.
	 *
	 * @return the URL that must answer to the user messages.
	 */
	public String getMessageAnswerURL()
	{
		return LyWebObject.getRootURL(getEnvironment());
	}

	/**
	 * Sets the pageWriter for current context (Thread).
	 *
	 * @param pageWriter the pagewriter we are currently working on when generating this page
	 * @see   #getPageWriter()
	 */
	public void setPageWriter(LyWebPageWriter pageWriter)
	{
		if (_pageWriters == null)
			_pageWriters = new Hashtable<Thread, LyWebPageWriter>();

		if (pageWriter == null)
		{
			_pageWriters.remove(Thread.currentThread());

			// clear possibly dead old threads
			if (_pageWriters.size() > 0)
			{
				Enumeration<Thread> e = _pageWriters.keys();

				while (e.hasMoreElements())
				{
					Thread t = e.nextElement();

					if (!t.isAlive())
						_pageWriters.remove(t);
				}
			}

			return;
		}

		// _foundSpecificFrame is added just to ensure compatibility with
		// existing applications : should be removed later
		if (_foundSpecificFrame
				&& (((pageWriter.getWindowName() == null) && (getCurrentWindowName() != null))
					|| ((pageWriter.getWindowName() != null)
						&& (!pageWriter.getWindowName().equals(getCurrentWindowName())))))
		{
			// Do it only for specific frame
			if (("" + pageWriter.getWindowName()).endsWith("_specific_frame"))
				return; // keep current
			if (("" + getCurrentWindowName()).endsWith("_specific_frame"))
				return; // keep current
		}

		_pageWriters.put(Thread.currentThread(), pageWriter);
	} // end method setPageWriter

	/**
	 * Gets the pageWriter for current context (Thread).
	 *
	 * @return the pagewriter we are currently working on when generating this page
	 * @see    #setPageWriter(LyWebPageWriter)
	 */
	public LyWebPageWriter getPageWriter()
	{
		if (_pageWriters == null)
			return null;

		return _pageWriters.get(Thread.currentThread());
	}

	/**
	 * Adds the given PageWriter in the known page writer list. This could be retreived by its id
	 * afterwards.
	 *
	 * @param page the new PageWriter
	 */
	public void addPageWriter(LyWebPageWriter page)
	{
		if (!REGISTER_PAGE_WRITERS)
			return;

		if (_pageWritersId == null)
			_pageWritersId = new Hashtable<String, LyWebPageWriter>();

		String pageId = page.getId();

		if (pageId != null)
			_pageWritersId.put(pageId, page);
	}

	/**
	 * Removes the given PageWriter from the known page writer list.
	 *
	 * @param page the PageWriterwhich should be removed
	 */
	public void removePageWriter(LyWebPageWriter page)
	{
		if (_pageWritersId == null)
			return;

		String pageId = page.getId();

		if (pageId != null)
			_pageWritersId.remove(pageId);

		Vector<String> oldIds = page.getOldIds();

		if (oldIds != null)
		{
			int i, n = oldIds.size();

			for (i = 0; i < n; i++)
			{
				pageId = oldIds.elementAt(i);
				_pageWritersId.remove(pageId);
			}
		}
	}

	/**
	 * Gets the PageWriter associated to the given id.
	 *
	 * @param  id identifier of the PageWriter
	 * @return the page writer if it is found, null otherwise
	 * @see    #setPageWriter(LyWebPageWriter)
	 */
	public LyWebPageWriter getPageWriter(String id)
	{
		if (_pageWritersId == null)
			return null;

		if (id == null)
			return null;

		return _pageWritersId.get(id);
	}

	/**
	 * <b>If</b> the current page writer is not the given page writer,<br>
	 * <b>if</b> <code>close=true</code><br>
	 * and <b>if</b> the page writer window is a pop up window<br>
	 * <b>then</b> adds a script to hide the pop up view<br>
	 * <b>Else if</b> <code>close=true</code><br>
	 * and <b>if</b> the given page writer's window is null<br>
	 * <b>then</b> adds the current target frame to the list of frame to reload and adds the current
	 * thread to associations between thread and close requests.
	 *
	 * @param pageWriter the page writer to close.
	 * @param close      if true, indicates that the page of the page writer has to be closed
	 * @see   #getClose()
	 */
	public void setClose(LyWebPageWriter pageWriter, boolean close)
	{
		// Check if close is called on current page writer
		LyWebPageWriter currentWriter = getPageWriter();

		if (close
				&& LyWebConstants.POPUP_VIEW.equals(pageWriter.getWindowName())
				&& (getControllerStackSize(LyWebConstants.POPUP_VIEW) <= 0))
		{
			addAdditionalRefreshScripts("<script type='text/javascript'>hidePopupView('"
				+ LyWebConstants.POPUP_VIEW + "');</script>");
			return;
		}

		if ((currentWriter != null) && (currentWriter != pageWriter))
			return;

		if (close)
		{
			if (pageWriter.getWindowName() == null)
				addReloadFrame(LyWebConstants.TARGET_FRAME);
				// else
				// System.out.println("Skipping reload :
				// "+pageWriter.getWindowName());
				// ELSE : Closing a popup Window : nothing to do
		}

		if (_closeRequests == null)
			_closeRequests = new Hashtable<Thread, Boolean>();

		_closeRequests.put(Thread.currentThread(), Boolean.valueOf(close));
	} // end method setClose

	/**
	 * Gets the current thread close request status from the {@link #_closeRequests} list. If the
	 * current thread is found in the list, the remove it.
	 *
	 * @return true if the current thread is found in the {@link #_closeRequests} list associated to
	 *         the value <code>true</code>, false otherwise
	 * @see    #setClose(LyWebPageWriter, boolean)
	 */
	public boolean getClose()
	{
		if (_closeRequests == null)
			return false;

		Boolean close = _closeRequests.get(Thread.currentThread());
		boolean result = ((close == null) ? false : close.booleanValue());

		_closeRequests.remove(Thread.currentThread());
		return result;
	}

	/**
	 * Sets the reload status with the given value.
	 *
	 * @param reload the status of the reload mode
	 */
	public void setReload(boolean reload)
	{
		_reload = reload;
	}

	/**
	 * Gets the reload mode status of the view.
	 *
	 * @return true if the view has to be reloaded, false otherwise
	 */
	public boolean isReload()
	{
		return _reload;
	}

	/**
	 * Adds the given message to the list of messages identified by their thread.
	 *
	 * @param message the message to add
	 * @see   #setMessage(String, int, String) with values (message, NO_TYPE, null)
	 * @see   #getMessage()
	 */
	public void setMessage(String message)
	{
		setMessage(message, null, null, TYPE_NO, null, null, null, null);
	}

	/**
	 * Adds the given message to the list of messages identified by their thread or remove it if
	 * message=null.
	 *
	 * @param message the content of the message to add. If null the remove the message from the
	 *                list.
	 * @param title   title of the message window
	 * @param icon    icon shown in the message window
	 * @param type    the type of the message from the values
	 *
	 *                <ul>
	 *                  <li>{@link #TYPE_ALERT}</li>
	 *                  <li>{@link #TYPE_QUESTION}</li>
	 *                  <li>{@link #TYPE_LOCATION}</li>
	 *                  <li>{@link #TYPE_LOCATION_NEW_WINDOW}</li>
	 *                  <li>{@link #TYPE_SHOW_FILE}</li>
	 *                </ul>
	 * @param commands bitset which contains COMMAND_xxx values
	 * @param bgColor  background color of the message window
	 * @param infoId   id of the info answer that must be called when answer is given
	 * @see   #getMessage()
	 */
	public void setMessage(String message, String title, String icon, int type, String bgColor,
			BitSet commands, String infoId, Hashtable<Short, String> parameters)
	{
		if (message != null)
		{
			if (_messages == null)
				_messages = new Hashtable<Object, Vector<_LyWebMessage>>();

			// If no request info was found, register given message current the view manager.
			// Otherwise register it with current thread.
            LyWebRequestInfo        info = null; // LyStrutsServlet.getWebRequestInfo();
            _LyWebMessage           webMessage = new _LyWebMessage(message, title, icon, bgColor, type, commands, infoId, parameters);
            Object                  key = (info == null ? this : Thread.currentThread());
            Vector<_LyWebMessage>   messages = _messages.get(key);

			if (messages == null)
				messages = new Vector<_LyWebMessage>(1);

			if (messages.contains(webMessage))
				removeMessageAnswer(infoId);
			else
			{
				messages.add(webMessage);
				_messages.put(key, messages);
			}
		}
		else if (_messages != null)
		{
			removeMessages(null);
		}
	}

	/**
	 * This method looks for the current message and sets the URL for the answer to the question.
	 * <br>
	 * <u>Note:</u> This, of course, is only usefull for messages of type = {@link #TYPE_QUESTION}.
	 *
	 * @param answerUrl The absolute URL to the server page (servlet or JSP) that manages the answer
	 *                  of the question.
	 */
	public void setMessageAnswerUrl(String answerUrl)
	{
		if (_messages == null)
			return;

		Vector<_LyWebMessage> messages = _messages.get(Thread.currentThread());

		if (messages == null)
			messages = _messages.get(this);

		if (messages != null)
		{
			_LyWebMessage	message;
			int				i, n = messages.size();

			for (i = 0; i < n; i++)
			{
				message = messages.elementAt(i);
				message.setAnswerUrl(answerUrl);
			}
		}
	}

	/**
	 * Gets the message associated to the current thread.
	 *
	 * @return the message if it is found, null otherwise
	 * @see    #setMessage(String)
	 * @see    #setMessage(String, int, String)
	 */
	public String getMessage()
	{
		return getMessage(false);
	}

	/**
	 * Gets the message associated to the current thread.
	 *
	 * @param  anyThread take the first available message (from any thread) (used by hidden frame)
	 * @return the message if it is found, null otherwise
	 * @see    #setMessage(String)
	 * @see    #setMessage(String, int, String)
	 */
	public String getMessage(boolean anyThread)
	{
		return getMessage(anyThread, false);
	}

	/**
	 * Gets the message associated to the current thread.
	 *
	 * @param  anyThread take the first available message (from any thread) (used by hidden frame)
	 * @param  ignoreSentMessages indicates whether messages that were already sent must be retrieved
	 * @return the message if it is found, null otherwise
	 * @see    #setMessage(String)
	 * @see    #setMessage(String, int, String)
	 */
	public String getMessage(boolean anyThread, boolean ignoreSentMessages)
	{
		if (_messages == null)
			return null;

		Vector<_LyWebMessage>	managerMessages = _messages.get(this);
		Vector<_LyWebMessage>	messages = _messages.get(Thread.currentThread());
		Vector<_LyWebMessage>	resultMessages = messages;

		if (resultMessages == null)
			resultMessages = managerMessages;

		if ((resultMessages == null)
				&& (anyThread)
				&& (_messages.size() > 0))
		{
			Iterator<Object> keyIterator = _messages.keySet().iterator();
			Object           key;

			while (keyIterator.hasNext() && (resultMessages == null))
			{
				key = keyIterator.next();
				resultMessages = _messages.get(key);

				if (resultMessages != null)
				{
					// Affect this message to current Thread
					_messages.remove(key);
					_messages.put(Thread.currentThread(), resultMessages);
				}
			}
		}

		if ((resultMessages != null) && (resultMessages.size() > 0))
		{
			_LyWebMessage	message = resultMessages.elementAt(0);

			if (!ignoreSentMessages || !message._sent)
			{
				message._sent = true;

				return generateMessage(message);
			}
		}

		return null;
	} // end method getMessage

	/**
	 * Generates the HTML code for a <code>_LyWebMessage</code> instance.
	 */
	protected String generateMessage(_LyWebMessage message)
	{
		return message.toString();
	}

	/**
	 * Answer the interface by calling the answer associated to message associated to the given ID.
	 *
	 * @param  infoId the id of the answer message
	 * @param  result true if the user chossed "yes" or "ok", false otherwise
	 * @return the answer interface
	 */
	public LyAnswerInterface answerMessage(String infoId, boolean result)
	{
		// Retrieve the message answer info with its id
		_LyMessageAnswerInfo info = getMessageAnswer(infoId);
		LyAnswerInterface answerInterface = null;

		removeMessageAnswer(infoId);

		if (info != null)
		{
			if (info._clientQuestion != null)
			{
				info._clientQuestion.questionAnswer(result, info._clientData);
				answerInterface = info._clientQuestion;
			}
			else if (info._clientMessage != null)
			{
				// Answer COMMAND_OK or COMMAND_NO.
				// XXX TO DO : answer one of the given commands
				info._clientMessage.messageAnswer((result ? COMMAND_YES : COMMAND_NO),
					info._clientData);
				answerInterface = info._clientMessage;
			}
		}

		return answerInterface;
	}

	/**
	 * Answer the interface by calling the answer associated to message associated to the given ID.
	 *
	 * @param  infoId the id of the answer message
	 * @param  result true if the user chossed "yes" or "ok", false otherwise
	 * @return the answer interface
	 */
	public LyAnswerInterface answerMessage(String infoId, short result)
	{
		LyDemandTransmitter		 transmitter = getDemandTransmitter();

		if (transmitter instanceof LyMessageWindowRegister)
			((LyMessageWindowRegister)transmitter).setMessageWindowResult(infoId, result);

		// Retrieve the message answer info with its id
		_LyMessageAnswerInfo info = getMessageAnswer(infoId);
		LyAnswerInterface answerInterface = null;

		removeMessageAnswer(infoId);

		if (info != null)
		{
			if (info._clientQuestion != null)
			{
				boolean booleanResult = (result == COMMAND_OK)
					|| (result == COMMAND_YES);

				info._clientQuestion.questionAnswer(booleanResult, info._clientData);
				answerInterface = info._clientQuestion;
			}
			else if (info._clientMessage != null)
			{
				// Answer COMMAND_OK or COMMAND_NO.
				// XXX TO DO : answer one of the given commands
				info._clientMessage.messageAnswer(result, info._clientData);
				answerInterface = info._clientMessage;
			}
		}

		return answerInterface;
	}

	/**
	 * Encodes a String to enable displaying it in an alert message box.
	 *
	 * @param  message the message to encode
	 * @return the encoded string
	 */
	public static String encodeString(String message)
	{
		if (message == null)
			return message;

		int			 i, n = message.length();
		StringBuffer sb = new StringBuffer((int)(n * 1.2)); // Size : just
		// guess
		char		 c;

		for (i = 0; i < n; i++)
		{
			c = message.charAt(i);

			// Rever quotes
			if (c == '"')
				sb.append('\'');
			else if (c == '\n')
				sb.append('\\').append('n');
			// FFT00095 : Javascript error on multilines text fields
			else if (c == '\r')
				sb.append('\\').append('r');
			else if (c == '\\')
				sb.append('\\').append('\\');
			else
				sb.append(c);
		}

		return sb.toString();
	} // end method encodeString

	/**
	 * Gets the main target frame suffix.
	 *
	 * @return the main target frame suffix
	 */
	public String getMainTargetFrame()
	{
		return LyWebConstants.MAIN_FRAME + "." + LyWebConstants.TARGET_FRAME;
	}

	/**
	 * Gets servlet root URL : this is "http://"+local host + LY_APP_DOC.<br>
	 * Rebuild the root URL.
	 *
	 * @return the root URL
	 * @see    #setRootURL(String)
	 */
	public final String getRootURL()
	{
		return _root_url;
	}

	/**
	 * Sets servlet root URL. This URL is built by server : application should never sets URL
	 * directly.
	 *
	 * @param rootUrl the new root URL
	 * @see   #getRootURL()
	 */
	public void setRootURL(String rootUrl)
	{
		_root_url = rootUrl;
	}

	/**
	 * Gets the component web view URL.
	 *
	 * @return the component web view URL
	 * @see    #setMobileStartURL(String)
	 */
	public final String getMobileStartURL()
	{
		return _mobileStartUrl;
	}

	/**
	 * Sets component web view URL.
	 *
	 * @param mobileStartUrl the new component web view URL
	 * @see   #getMobileStartURL()
	 */
	public void setMobileStartURL(String mobileStartUrl)
	{
		_mobileStartUrl = mobileStartUrl;
	}

	/**
	 * Sets the target code used in multi-windows mode.
	 *
	 * @param code the new code
	 */
	public void setTargetCode(String targetCode)
	{
		_targetCode = targetCode;
	}

	/**
	 * Return the target code used in multi-windows mode.
	 *
	 * @return the target code is it has been set, null otherwise
	 */
	public String getTargetCode()
	{
		return _targetCode;
	}

	/**
	 * Adds the given frame to the list of {@link #_reloadFrames reload frames}.
	 *
	 * @param frame a frame which will need to be refreshed next time
	 * @see   #addReloadFrame(String, String) where page writer id = null
	 */
	public void addReloadFrame(String frame)
	{
		addReloadFrame(frame, null);
	}

	/**
	 * Adds the given frame to the list of {@link #_reloadFrames reload frames}.
	 *
	 * @param frame        a frame which will need to be refreshed next time
	 * @param pageWriterId optional pagewriter identifier we check before refreshing.
	 */
	public void addReloadFrame(String frame, String pageWriterId)
	{
		if ((frame == null) || (frame.length() == 0))
			return;

		synchronized (this)
		{
			if (_reloadFrames == null)
				_reloadFrames = new Vector<String>();

			if (frame.startsWith(LyWebConstants.POPUP_VIEW))
				frame = LyWebConstants.POPUP_VIEW;
			if (frame.indexOf("_specific_frame") >= 0)
			{
				int n1 = "_specific_frame".length();
				int n2 = frame.indexOf("_specific_frame");

				frame = frame.substring(0, n1 + n2);
				_foundSpecificFrame = true; // Found specific frame :
				// TARGET_FRAME is no longer alone
			}
			else if (LyWebConstants.PILOT_FRAME.equals(frame))
			{
				_foundSpecificFrame = true; // Found PILOT_FRAME : TARGET_FRAME
				// is no longer alone
			}

			frame = frame.intern();
			if (frame == LyWebPageWriter.MAIN_BUFFER)
				frame = LyWebConstants.TARGET_FRAME; // Anacronisme : why
			// MAIN ?

			// Everything is going to be updated -> reload...
			if (_reloadFrames.contains(LyWebConstants.TOP_FRAME))
				return;
			if ((!_foundSpecificFrame) && _reloadFrames.contains(LyWebConstants.TARGET_FRAME)
					&& (frame != LyWebConstants.PILOT_FRAME))
				return;

			if (((frame == LyWebConstants.TOP_FRAME)
						|| ((frame == LyWebConstants.TARGET_FRAME) && (!_foundSpecificFrame)))
					&& (pageWriterId == null) && (_reloadFrames.size() > 0))
				// Clear current reload frame list
				_reloadFrames = new Vector<String>();
			else if (frame == LyWebConstants.POPUP_VIEW)
			{
				// Clean up all subframes of the popup view since we reload all the popup anyway
				// Remove all existing frame with a controller id
			    Enumeration<String>  e = _reloadFrames.elements();

				while (e.hasMoreElements())
				{
					String name = e.nextElement();

					if (getPopupViewMapping(name) == LyWebConstants.POPUP_VIEW)
						_reloadFrames.remove(name);
				}
			}
			// Before adding:
			// - if the frame exists with a controller id and if the given
			// controller id is null
			// remove the old one
			// - if there is already a frame store, ignore the new one (even if
			// the controller id is specified)
			if (!_reloadFrames.contains(frame))
			{
				if (pageWriterId != null)
				{
					frame = frame + SEP_PW_ID + pageWriterId;
					if (!_reloadFrames.contains(frame))
						_reloadFrames.addElement(frame);
				}
				else
				{
					// Remove all existing frame with a controller id
				    Enumeration<String> e = _reloadFrames.elements();

					while (e.hasMoreElements())
					{
						String name = e.nextElement();

						if (name.startsWith(frame + SEP_PW_ID))
							_reloadFrames.remove(name);
					}

					_reloadFrames.addElement(frame);
				}
			}
		}
	} // end method addReloadFrame

	/**
	 * Adds the mapping between the given frame and its given popup view.
	 *
	 * @param frame the frame to map to the view
	 * @param view  the view to map to the frame
	 */
	public void addPopupViewMapping(String frame, String view)
	{
		if ((frame != null) && (view != null))
		{
			if (_framesPopupViewMapping == null)
				_framesPopupViewMapping = new HashMap<String, String>();

			_framesPopupViewMapping.put(frame, view.intern());
		}
	}

	/**
	 * Gets the popup view mapped to the given frame.
	 *
	 * @param  frame the frame
	 * @return the pop up view mapped to the frame
	 */
	public String getPopupViewMapping(String frame)
	{
		if ((_framesPopupViewMapping == null) || (frame == null))
			return null;

		return _framesPopupViewMapping.get(frame);
	}

	/**
	 * Removes a frames from the {@link #_reloadFrames} list.
	 *
	 * @param frame a frame which will not need to be refreshed next time
	 */
	public void clearReloadFrame(String frame)
	{
		if ((frame == null) || (frame.length() == 0) || (_reloadFrames == null))
			return;

		synchronized (this)
		{
			frame = frame.intern();
			if (frame == LyWebPageWriter.MAIN_BUFFER)
				frame = LyWebConstants.TARGET_FRAME;

			_reloadFrames.remove(frame);
		}
	}

	/**
	 * Gets the list of frames which need to be refreshed.
	 *
	 * @return the list of frames name, or null if no frame needs reload.
	 */
	public Vector<String> getReloadFrames()
	{
		return _reloadFrames;
	}

	/**
	 * Clears the list of frames which need to be reloaded.
	 */
	public void clearReloadFrames()
	{
		synchronized (this)
		{
			_reloadFrames = null;
		}
	}

	/**
	 * Indicates whether the client needs an update.
	 */
	public boolean needsUpdate()
	{
		Vector<String>	reloadFrames = getReloadFrames();

		if (reloadFrames != null)
		{
			for (int i = 0; i < reloadFrames.size(); i++)
			{
				if (reloadFrames.elementAt(i) != null)
					return true;
			}
		}

		Vector<String>	addScripts = null;

		synchronized (_refreshScriptsLock)
	    {
			addScripts = _additionalRefreshScripts;

			if (addScripts != null)
			{
				String	script;

				for (int i = 0; (addScripts != null) && (i < addScripts.size()); i++)
				{
					script = addScripts.get(i);

					if (script != null)
						return true;
				}
			}
	    }

		if (getMessage(true, true) != null)
			return true;

		return false;
	}

	/**
	 * Adds a refreshing view in the list : this enables auto-refresh mode.
	 *
	 * @param viewComp the refreshed view
	 */
	public void addRefreshingView(LyGetRefreshInterface viewComp)
	{
		if (viewComp == null)
			return;

		synchronized (_refreshingViewsLock)
		{
			if (_refreshingViews == null)
				_refreshingViews = new Vector<LyGetRefreshInterface>();
			_refreshingViews.addElement(viewComp);
		}
	}

	/**
	 * Removes the given view from the {@link #_refreshingViews} list.
	 *
	 * @param viewComp the refreshed view to remove from the list
	 */
	public void clearRefreshingView(LyGetRefreshInterface viewComp)
	{
		if ((viewComp == null) || (_refreshingViews == null))
			return;

		synchronized (_refreshingViewsLock)
		{
			_refreshingViews.removeElement(viewComp);

			if (_refreshingViews.size() <= 0)
				_refreshingViews = null;
		}
	}

	/**
	 * Gets the refresh rate for the auto refresh views.
	 *
	 * @return refresh value (duration is seconds) or 0 if no refresh is set
	 */
	public int getRefreshRate()
	{
		if (getEnvironment().testEnv("LY_REPLAY"))
		{
			String d = getEnvironment().getEnv("LY_TEST_DELAY");

			try
			{
				if (d != null)
                {
					return Integer.parseInt(d) / 1000;
                }
			}
			catch (NumberFormatException e)
			{
			}

			return 1;
		}

		int delay = getDelay();

		if (delay > 0)
			return delay;

		// Initial value is a resource
		int refresh = LyTools.intFromString(_environment.getResource("LY_AUTO_REFRESH_VIEWS"));

		if (_refreshingViews == null)
			return refresh;

		synchronized (_refreshingViewsLock)
		{
			LyComponent     viewComp;
			int			    i, n = _refreshingViews.size();
			boolean		    visible;
			int			    r;
			LyComponent     parentView;
			LyViewComponent parentViewComp;

			// Go down since we may remove obsolete views
			for (i = n - 1; i >= 0; i--)
			{
				visible = true;
				viewComp = (LyComponent)_refreshingViews.elementAt(i);
				if (viewComp == null)
					continue;

				parentView = viewComp.getParentView();
				if (parentView == null)
					continue;

				parentViewComp = parentView.getInternalView(false);

				if (parentViewComp instanceof LyStrutsFrame)
					// xx workaround for isVisible in LyStrutsFrame
					visible = ((LyStrutsFrame)parentViewComp).isSuperVisible();
				else if (parentViewComp instanceof LyWebPageWriter)
					visible = ((LyWebPageWriter)parentViewComp).isVisible();

				if (visible && (viewComp instanceof LyGetRefreshInterface))
				{
					r = ((LyGetRefreshInterface)viewComp).getRefresh();
					// Take the minimum refresh rate
					if ((r > 0) && ((refresh <= 0) || (refresh > r)))
						refresh = r;
				}
			}
		}

		return refresh;
	} // end method getRefreshRate

	/**
	 * Indicates if DIV tags may be written.
	 *
	 * @return true if DIV tags may be written, false otherwise
	 */
	public boolean writeDivs()
	{
		return _writeDivs;
	}

	/**
	 * Indicates if popup view is allowed.
	 *
	 * @return true if popup view is allowed, false otherwise
	 */
	public boolean usePopupView()
	{
		return _usePopupView;
	}

	/**
	 * Indicates if IFRAME tags must contain the scrolling attribute set.
	 *
	 * @return true if IFRAME tags must contain the scrolling attribute set, false otherwise
	 */
	public boolean forceScrolling()
	{
		return _forceScrolling;
	}

	/**
	 * Sets the browser type and eventually properties that are not supported by this browser.
	 *
	 * @param userAgent the userAgent string send by the browser.
	 * @see   #getBrowser()
	 */
	public void setBrowser(String userAgent)
	{
		// If already set, return
		if (_browserVersion > 0)
			return;

		userAgent = userAgent.toUpperCase();

		// Check for Chrome browser or compatible
		int index = userAgent.indexOf(BROWSER_CHROME_ID);

		if (index >= 0)
		{
			_browser = BROWSER_CHROME;

			// find version
			int    pos = userAgent.indexOf("/", index);
			String verAsString = userAgent.substring(pos + 1);

			_browserVersion = getBrowserVersion(verAsString);
			return;
		}

		// Check for Firefox browser or compatible
		index = userAgent.indexOf(BROWSER_FIREFOX_ID);

		if (index >= 0)
		{
			_browser = BROWSER_FIREFOX;

			// find version
			int    pos = userAgent.indexOf("/", index);
			String verAsString = userAgent.substring(pos + 1);

			_browserVersion = getBrowserVersion(verAsString);
			_writeDivs = false;
			_forceScrolling = true;
			return;
		}

		// Check for MSIE browser or compatible
		index = userAgent.indexOf(BROWSER_MSIE_ID);

		if (index >= 0)
		{
			_browser = BROWSER_MSIE;

			// find version
			int    pos = userAgent.indexOf(";", index);
			String verAsString = userAgent.substring(index + BROWSER_MSIE_ID.length(), pos);

			_browserVersion = getBrowserVersion(verAsString);
			return;
		}

		// Check for netscape browser (look for MOZILLA String) or compatible
		index = userAgent.indexOf(BROWSER_NETSCAPE_ID);
		if (index >= 0)
		{
			_browser = BROWSER_NETSCAPE;

			int    pos = userAgent.indexOf(' ');
			String verAsString = userAgent.substring(index + BROWSER_NETSCAPE_ID.length() + 1, pos);

			_browserVersion = getBrowserVersion(verAsString);

			return;
		}

		// Check for Opera browser or compatible
		index = userAgent.indexOf(BROWSER_OPERA_ID);
		if (index >= 0)
		{
			_browser = BROWSER_OPERA;

			int    pos = userAgent.indexOf(' ');
			String verAsString = userAgent.substring(index + BROWSER_OPERA_ID.length() + 1, pos);

			_browserVersion = getBrowserVersion(verAsString);

			return;
		}

		_browser = BROWSER_OTHER;
		_browserVersion = -1.0;
	} // end method setBrowser

	/**
	 * Extracts the browser version from header info.
	 *
	 * @param  verAsString version as a string
	 * @return version as a float
	 */
	protected double getBrowserVersion(String verAsString)
	{
		double ver = 0.0;

		if (verAsString == null)
			return ver;

		verAsString = verAsString.trim();

		StringBuffer buf = new StringBuffer();

		for (int i=0; i<verAsString.length(); i++)
		{
			char c =verAsString.charAt(i);

			if (c == '.' || Character.isDigit(c))
				buf.append(c);
			else
				break;
		}
		verAsString = buf.toString();

		int n1 = verAsString.indexOf(".");

		if (n1 > 0)
		{
			int n2 = verAsString.indexOf(".", n1 + 1);

			// When version is like 1.1.2 keep only major and minor (1.1)
			if (n2 > 0)
				verAsString = verAsString.substring(0, n2);
		}

		n1 = verAsString.indexOf(" ");

		// If there is a space in the version string, only keep first part
		// which should contains the real version number.
		if (n1 > 0)
			verAsString = verAsString.substring(0, n1);

		// For a BETA build, try to find a B in the string
		n1 = verAsString.indexOf("B");
		if (n1 > 0)
			verAsString = verAsString.substring(0, n1);

		// Skip all other non numeric values...
		String version = new String();
		int i = 0;
		boolean cont = true;
		while (cont && (i < verAsString.length()))
		{
			if ( Character.isDigit(verAsString.charAt(i))
					|| verAsString.charAt(i) == '.')
			{
				version += verAsString.charAt(i);
				i++;
			}
			else
				cont = false;
		}

		try
		{
			ver = Double.parseDouble(version);
		}
		catch (Exception e)
		{
			getEnvironment().writeException(e);
		}
		return ver;
	} // end method getBrowserVersion

	/**
	 * Gets the browser type code. Possible values are :
	 *
	 * <ul>
	 *   <li>{@link #BROWSER_FIREFOX}</li>
	 *   <li>{@link #BROWSER_MSIE}</li>
	 *   <li>{@link #BROWSER_NETSCAPE}</li>
	 *   <li>{@link #BROWSER_OPERA}</li>
	 *   <li>{@link #BROWSER_OTHER}</li>
	 * </ul>
	 *
	 * @return browser the browser type
	 * @see    #setBrowser(String)
	 */
	public int getBrowser()
	{
		return _browser;
	}

	/**
	 * Gets the browser version.
	 *
	 * @return browser the browser version
	 */
	public double getBrowserVersion()
	{
		return _browserVersion;
	}

	/**
	 * Indicates wether this view manager handles multiple user sessions or not. Current
	 * implementation always returns true.
	 *
	 * @return true if the view manager uses multiple user sessions (like web server)
	 */
	public boolean isMultipleSession()
	{
		return true;
	}

	/**
	 * Updates the _controllerStackList parameter with the count of controllers from the
	 * controllersStackList parameter. controllersStackList must contain names of controller stacks
	 * as keys and stack of controllers as values. _controllerStackList must contain names of
	 * controller stacks as keys and stack of controllers as values.
	 *
	 * @param controllersStackList
	 */

	public void updateControllersStackList(Hashtable<String, Stack<LyViewController>> controllersStackList)
	{
		super.updateControllersStackList(controllersStackList);

		_lastPopupViewController = null;

		if ((controllersStackList == null) || (controllersStackList.isEmpty()))
			return;

		Iterator<?>	keys = controllersStackList.keySet().iterator();
		String		key;
		Stack<LyViewController>	stack, defaultStack;

		while (keys.hasNext())
		{
			key = (String)keys.next();

			if (LyWebConstants.POPUP_VIEW.equals(key))
			{
				stack = controllersStackList.get(key);

				if ((stack != null) && !stack.isEmpty())
				{
					LyViewController	popupController = stack.peek();
					int					i, n = (popupController == null ? 0 : popupController.getSonsCount());
					boolean				hasNonPopupSon = false;

					for (i = 0; i < n; i++)
					{
						if (!stack.contains(popupController.getSon(i)))
						{
						    defaultStack = controllersStackList.get(LySession.DEFAULT);

							if ((defaultStack != null) && !defaultStack.isEmpty() && defaultStack.contains(popupController.getSon(i)))
							{
								hasNonPopupSon = true;
							}
						}
					}

					if (!hasNonPopupSon && popupController.getParent() != null)
						_lastPopupViewController = popupController;
				}
			}
		}
	}

	/**
	 * Returns the view controller that was last displayed in the popup view.
	 */
	public LyViewController getLastPopupViewController()
	{
		return _lastPopupViewController;
	}

	/**
	 * Indicates if current browser is supported with the version of Leonardi.<br />
	 * Supported browsers are :
	 *
	 * <ul>
	 *   <li>MSIE 5.0 or greater</li>
	 *   <li>Firefox 1.0 or greater</li>
	 * </ul>
	 *
	 * @return true if the browser is supported, false otherwise.
	 */
	public boolean isBrowserSupported()
	{
		boolean result = (((_browser == BROWSER_FIREFOX) && (_browserVersion >= 1.0))
				|| ((_browser == BROWSER_MSIE) && (_browserVersion >= 5.0)));

		if ((!result) && getEnvironment().isDebugLevel(LyEnvironment.DEBUG_ALL))
		{
			getEnvironment().logWarning("Browser "+_browser+" version "+_browserVersion+" is not supported, accepted only in DEBUG_ALL mode.");
			_browser = BROWSER_FIREFOX;
			_browserVersion = 2.0;
			return true;
		}
		return result;
	}

	/**
	 * Generate the Javascript parameter used to open a new window for a given location and given parameters.
	 *
	 * @param location        the location that must be displayed in the new window
	 * @param parametersTable the list of parameters used to open the new window
	 * @return the Javascript code
	 */
	protected String getNewWindowScriptParameters(String location, Hashtable<Short, String> parametersTable)
	{
		// From the resource file, try to get the size of the new window
		String parameters;
		String x = "";
		String y = "";
		String width = null;
		String height = null;
		String fullscreen = "";
		String name = (parametersTable != null ? parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_NAME)) : null);

		if ((parametersTable != null) && (parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_X)) != null))
			x = parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_X));

		if ((parametersTable != null) && (parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_Y)) != null))
			y = parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_Y));

		if ((parametersTable != null) && (parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_WIDTH)) != null))
			width = parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_WIDTH));

		if ((parametersTable != null) && (parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_HEIGHT)) != null))
			height = parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_HEIGHT));

		if ((parametersTable != null) && (parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_FULL_SCREEN)) != null))
			fullscreen = parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_FULL_SCREEN));

		if ((parametersTable != null) && (parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_SCROLLBARS)) != null))
			parameters = "scrollbars=" + parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_SCROLLBARS)) + ",";
		else
			parameters = "scrollbars=yes,";

		if ((parametersTable != null) && (parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_RESIZABLE)) != null))
			parameters += "resizable=" + parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_RESIZABLE)) + ",";
		else
			parameters += "resizable=yes,";

		if ((parametersTable != null) && (parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_TOOLBAR)) != null))
			parameters += "toolbar=" + parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_TOOLBAR)) + ",";
		else
			parameters += "toolbar=yes,";

		if ((parametersTable != null) && (parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_STATUSBAR)) != null))
			parameters += "status=" + parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_STATUSBAR)) + ",";
		else
			parameters += "status=yes,";

		if ((parametersTable != null) && (parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_MENUBAR)) != null))
			parameters += "menubar=" + parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_MENUBAR)) + ",";
		else
			parameters += "menubar=no,";

		if ((parametersTable != null) && (parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_DIRECTORIES)) != null))
			parameters += "directories=" + parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_DIRECTORIES)) + ",";
		else
			parameters += "directories=no,";

		if ((parametersTable != null) && (parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_LOCATION)) != null))
			parameters += "location=" + parametersTable.get(new Short(SHOW_LOCATION_PARAMETER_LOCATION));
		else
			parameters += "location=no";

		// If parameters are given in the string, use them instead of
		// default parameters
		if (location.indexOf(PARAMETERS) >= 0)
		{
			int pos = location.indexOf(PARAMETERS);

			parameters = location.substring(pos);
			location = location.substring(0, pos);
		}

		if ((width == null)
				|| (width.length() == 0)
				|| width.equals("-1"))
			width = Integer.toString(DEFAULT_POPUP_WIDTH);

		if ((height == null)
				|| (height.length() == 0)
				|| height.equals("-1"))
			height = Integer.toString(DEFAULT_POPUP_WIDTH);

		if ((name == null) || (name.length() == 0))
			name = "Help" + System.currentTimeMillis();

		return "new Array('" + name + "','" + location + "','" + parameters + "','"
			+ x + "','" + y + "','" + width + "','" + height + "','" + fullscreen + "')";
	}

	/**
	 * Class used to describe a message.
	 */
	protected class _LyWebMessage
	{
		/** Text of the message. */
		public String _text;

		/** Title of the message. */
		public String _title;

		/** Icon of the message. */
		public String _icon;

		/** Background color of the message. */
		public String _bgColor;

		/** Indicates whether this message is a question or not. */
		public int    _type;

		/** Id of the info answer that must be called when answer is given. */
		public String _infoId;

		/** Commands of the message. */
		public BitSet _commands;

		/** The url of the question answer. */
		public String _answerUrl;

		/** List of parameters of the message. */
		public Hashtable<Short, String>	_parameters;

		/** Indicates whether this message has already been sent. */
		public boolean _sent;

		/**
		 * Constructor.
		 *
		 * @param text     the message text
		 * @param title    the message title
		 * @param icon     the message icon
		 * @param bgColor  the message background color
		 * @param type     the message type
		 * @param commands the message commands
		 * @param infoId   the id of the information answer
		 */
		_LyWebMessage(String text, String title, String icon, String bgColor, int type, BitSet commands, String infoId,
				Hashtable<Short, String> parameters)
		{
			_text = text;
			_title = title;
			_icon = icon;
			_bgColor = bgColor;
			_type = type;
			_commands = commands;
			_infoId = infoId;
			_answerUrl = null;
			_parameters = parameters;
			_sent = false;
		}

		/*
		 * (non-Javadoc)
		 *
		 * @see java.lang.Object#toString()
		 */
		public String toString()
		{
			if ((_type != TYPE_LOCATION)
					&& (_type != TYPE_LOCATION_NEW_WINDOW)
					&& (_type != TYPE_SHOW_FILE))
			{
				String messageWindowClassName = (_beans == null ? null : _beans.getString("MESSAGE_WINDOW"));

				if ((messageWindowClassName != null) && (messageWindowClassName.length() > 0))
				{
					if (_answerUrl == null)
						_answerUrl = getMessageAnswerURL();

					Class<?>[]  types = new Class[2];
					Object[]    parameters = new Object[2];

					types[0] = LyViewManager.class;
					parameters[0] = LyWebViewManager.this;

					types[1] = String.class;
					parameters[1] = _answerUrl;

					LyMessageWindowInterface messageWindow = null;

					try
					{
						Class<?>	   viewClass = Class.forName(messageWindowClassName);
						Constructor<?> constructor = viewClass.getConstructor(types);

						messageWindow = (LyMessageWindowInterface)constructor.newInstance(parameters);
					}
					catch (InvocationTargetException e1)
					{
						_environment.writeException(e1.getTargetException());
						_environment.writeException(e1);
						throw new IllegalArgumentException("BF AppEngine : can't build message window class : "
							+ messageWindowClassName);
					}
					catch (Exception e)
					{
						_environment.writeException(e);
						throw new IllegalArgumentException("BF AppEngine : can't build message window class : "
							+ messageWindowClassName);
					}

					messageWindow.setBackgroundColor(_bgColor);
					messageWindow.setIcon(_icon);
					messageWindow.setTitle(_title);
					messageWindow.setMessage(_text);
					messageWindow.setCommands(_commands);
					messageWindow.setClientData(_infoId);

					return messageWindow.toString();
				}
			}

			String text = LyWebViewManager.encodeString(_text);

			if (_type == TYPE_QUESTION)
			{
				if (_answerUrl == null)
					_answerUrl = getMessageAnswerURL();

				return ("<script type='text/javascript'>\n//<![CDATA[\nAskQuestion(\"" + text
					+ "\",\"" + _infoId + "\",\"" + _answerUrl + "\")\n//]]>\n</script>");
			}

			if (_type == TYPE_LOCATION)
				return ("<script type='text/javascript'>\n//<![CDATA[\nShowLocation(\"" + text
					+ "\")\n//]]>\n</script>");

			if (_type == TYPE_LOCATION_NEW_WINDOW)
			{
				String	parameters = getNewWindowScriptParameters(text, _parameters);

				return ("<script type='text/javascript'>NewWindow(" + parameters + ");</script>");
			}
			else if (_type == TYPE_SHOW_FILE)
			{
				LyEnvironment lyEnvironment = getEnvironment();
				String		  message = lyEnvironment.translate("LY_SHOW_FILE_MESSAGE");
				String		  alternate = lyEnvironment.translate("LY_SHOW_FILE_ALTERNATE");
				String		  title = lyEnvironment.translate("LY_SHOW_FILE_TITLE");
				String		  click = lyEnvironment.translate("LY_SHOW_FILE_CLICK");

				return ("<script type='text/javascript'>\n//<![CDATA[\nShowFile(\"" + text
					+ "\",\"" + title + "\",\"" + message + "\",\"" + click + "\",\""
					+ alternate + "\")\n//]]>\n</script>");
			}
			else
			{
				getPageWriter();

				// CODE REMOVE BY NBO : the script below is doing nothing.
				// if (pageWriter != null)
				// return ("<script type='text/javascript'>\n//<![CDATA[\n(\"" +
				// text + "\");\n//]]>\n</script>");

				return ("<script type='text/javascript'>\n//<![CDATA[\nShowMessage(\"" + text
					+ "\",\"" + _infoId + "\")\n//]]>\n</script>");
			}
		} // end method toString

		/**
		 * Sets the URL for the answer to a question.<br>
		 * Default is rootUrl.
		 *
		 * @param url the url for answer.
		 */
		public void setAnswerUrl(String url)
		{
			_answerUrl = url;
		}

		@Override
		public boolean equals(Object obj)
		{
			if (!(obj instanceof _LyWebMessage))
				return false;

			_LyWebMessage	message = (_LyWebMessage)obj;

			if (((_text == null) && (message._text != null))
					|| ((message._text == null) && (_text != null))
					|| ((_text != null) && !_text.equals(message._text)))
				return false;

			if (((_title == null) && (message._title != null))
					|| ((message._title == null) && (_title != null))
					|| ((_title != null) && !_title.equals(message._title)))
				return false;

			if (((_icon == null) && (message._icon != null))
					|| ((message._icon == null) && (_icon != null))
					|| ((_icon != null) && !_icon.equals(message._icon)))
				return false;

			if (_type != message._type)
				return false;

			BitSet	bs1 = _commands;
			BitSet	bs2 = message._commands;

			if (bs1 == null)
			{
				bs1 = new BitSet();
				bs1.set(LyViewManager.COMMAND_OK);
			}

			if (bs2 == null)
			{
				bs2 = new BitSet();
				bs2.set(LyViewManager.COMMAND_OK);
			}

			if (!bs1.equals(bs2))
				return false;

			if (((_parameters == null) && (message._parameters != null))
					|| ((message._parameters == null) && (_parameters != null))
					|| ((_parameters != null) && !_parameters.equals(message._parameters)))
				return false;

			return true;
		}
	} // end class _LyWebMessage

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.view.LyViewManager#setCurrentWindowName(java.lang.String)
	 */
	public void setCurrentWindowName(String name)
	{
		if (name != null)
		{
			if (_windowNames == null)
				_windowNames = new Hashtable<Thread, String>();

			_windowNames.put(Thread.currentThread(), name);
		}
		else if (_windowNames != null)
		{
			_windowNames.remove(Thread.currentThread()); // clear

			// clear possibly dead old threads
			if (_windowNames.size() > 0)
			{
				Enumeration<Thread> e = _windowNames.keys();

				while (e.hasMoreElements())
				{
					Thread t = e.nextElement();

					if (!t.isAlive())
						_windowNames.remove(t);
				}
			}

		}

	} // end method setCurrentWindowName

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.view.LyViewManager#getCurrentWindowName()
	 */
	public String getCurrentWindowName()
	{
		if (_windowNames == null)
			return null;

		return _windowNames.get(Thread.currentThread());
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.view.LyViewManager#encodeImage(java.lang.Object,
	 *      java.lang.String, java.lang.String)
	 */
	public boolean encodeImage(Object image, String imageDir, String imageFile)
	{
		return LyImageUtils.encodeImage((java.awt.Image)image, imageDir, imageFile);
	}
	
	/**
	 * Indicates whether given image may be cached.
	 * Default implementation returns true.
	 * 
	 * @param path  The path of image that may be cached.
	 * @param image The image that may be cached.
	 * @return true if image may be cached, false otherwise
	 */
	public boolean useCacheForImage(String path, Object image)
	{
		if ((image == null) || !(image instanceof java.awt.Image))
			return false;
		
		java.awt.Image	awtImage = (java.awt.Image)image;
		int				imageSize = awtImage.getWidth(null) * awtImage.getHeight(null);
		
		if (imageSize > CACHED_IMAGE_MAX_SIZE)
			return false;

		return true;
	}

	/**
	 * Gets the list of additional refresh scripts {@link #_additionalRefreshScripts}.
	 *
	 * @return the list of additional refresh scripts
	 */
	public Vector<String> getAdditionalRefreshScripts()
	{
	    return getAdditionalRefreshScripts(false);
	}

	/**
	 * Gets the list of additional refresh scripts {@link #_additionalRefreshScripts}.
	 *
	 * @param clearScripts indicates whether additional refresh scripts must be cleared.
	 * @return the list of additional refresh scripts
	 */
	public Vector<String> getAdditionalRefreshScripts(boolean clearScripts)
	{
	    synchronized (_refreshScriptsLock)
	    {
	    	if (clearScripts)
	    	{
				if (_additionalRefreshScripts != null)
				{
			    	Vector<String>	result = new Vector<String>();
		    		String			script;
		    		
					for (int i = 0; (_additionalRefreshScripts != null) && (i < _additionalRefreshScripts.size()); i++)
					{
						script = _additionalRefreshScripts.get(i);
						
						if (script != null)
							result.add(script);
					}
					
					_additionalRefreshScripts = null;
					
					return result;
				}
	    	}

			return _additionalRefreshScripts;
	    }
	}

	/**
	 * Clear the list of additional refresh scripts {@link #_additionalRefreshScripts}.
	 */
	public void clearAdditionalRefreshScripts()
	{
        synchronized (_refreshScriptsLock)
        {
            _additionalRefreshScripts = null;
            _additionalRefreshScriptsBuffer = null;
        }
	}

	/**
	 * Adds the given additional refresh script to the {@link #_additionalRefreshScripts} list.
	 *
	 * @param script the HTML representation of the script to add
	 */
	public final void addAdditionalRefreshScripts(String script)
	{
		addAdditionalRefreshScripts(script, true);
	}

	/**
	 * Adds the given additional refresh script to the {@link #_additionalRefreshScripts} list.
	 *
	 * @param script the HTML representation of the script to add
	 * @param insertEnd indicates whether the script must be inserted at the end of the list
	 */
	public void addAdditionalRefreshScripts(String script, boolean insertEnd)
	{
		if (script == null)
			return;

	    synchronized (_refreshScriptsLock)
	    {
	    	Vector<String>	scriptsVector;

	    	if (_additionalRefreshScriptsBuffer == null)
	    	{
	    		if (_additionalRefreshScripts == null)
	    			_additionalRefreshScripts = new Vector<String>();
	    		else if (_additionalRefreshScripts.contains(script))
	                _additionalRefreshScripts.removeElement(script);

	    		scriptsVector = _additionalRefreshScripts;
	    	}
	    	else
	    	{
	    		if (_additionalRefreshScriptsBuffer.contains(script))
	    			_additionalRefreshScriptsBuffer.removeElement(script);

	    		scriptsVector = _additionalRefreshScriptsBuffer;
	    	}

            if (insertEnd)
            	scriptsVector.addElement(script);
            else
            	scriptsVector.insertElementAt(script, 0);
	    }
	}

	/**
	 * Enables or disables the use of a temporary buffer for additional refresh scripts.
	 *
	 * @param activate true if buffer must be activated, false otherwise.
	 */
	public void activateAdditionalRefreshScriptsBuffer(boolean activate)
	{
	    synchronized (_refreshScriptsLock)
	    {
			if (activate)
				_additionalRefreshScriptsBuffer = new Vector<String>();
			else
			{
				if (_additionalRefreshScriptsBuffer != null)
				{
					String	script;

					for (int i = 0; (_additionalRefreshScriptsBuffer != null) && (i < _additionalRefreshScriptsBuffer.size()); i++)
					{
						script = _additionalRefreshScriptsBuffer.get(i);

						if (_additionalRefreshScripts == null)
			    			_additionalRefreshScripts = new Vector<String>();
			    		else if (_additionalRefreshScripts.contains(script))
			                _additionalRefreshScripts.removeElement(script);

						_additionalRefreshScripts.addElement(script);
					}

					_additionalRefreshScriptsBuffer = null;
				}
			}
	    }
	}

	/**
	 * Gets the current decoration page.
	 *
	 * @return the decoration.
	 * @see    #setDecoration(LyWebDecoration)
	 */
	public LyWebDecoration getDecoration()
	{
		if (_deco == null)
			_deco = buildWebDecoration(_environment);

		return _deco;
	}

	/**
	 * Sets the decoration page.
	 *
	 * @param deco The new decoration.
	 * @see   #getDecoration()
	 */
	public void setDecoration(LyWebDecoration deco)
	{
		_deco = deco;
	}

	/**
	 * Method used to call back the listener of the specified message window for the specified
	 * result.
	 *
	 * @param  cmdResult the comand result : 1 if the user chose 'YES' or 'OK', 0 otherwise
	 * @param  infoId    the message info Id (of type {@link String}
	 * @return the answer to the message
	 * @see    #answerMessage(String, boolean)
	 */
	public LyAnswerInterface messageActionPerformed(Object infoId, int cmdResult)
	{
		return answerMessage((String)infoId, ((cmdResult == COMMAND_YES) ? true : false));
	}

	/**
	 * Checks whether this manager handles multiple session. Return always <code>true</code>.
	 */
	public boolean isMultiSession()
	{
		return true;
	}

	/**
     * Accessor to the properties file or parameter if specified in constructor.
     *
     * @param prefix    Prefix used for properties files name.
     * @return the properties.
     */
	public Properties getPropertiesFromContext(String prefix)
    {
        if (prefix == null)
            return null;

        if (getWebRequestInfo() == null)
            return null;

        HttpServlet servlet = getWebRequestInfo().getServlet();
        Properties  properties = null;

        if (servlet != null)
        {
            ServletContext context = servlet.getServletContext();

            if (context != null)
            {
            	@SuppressWarnings("unchecked")
            	Enumeration<String> paramNames = context.getInitParameterNames();

            	while (paramNames.hasMoreElements())
                {
                    String param = paramNames.nextElement();

                    if (param.startsWith(prefix))
                    {
                        if (properties == null)
                            properties = new Properties();
                        properties.put(param.substring(prefix.length()), context.getInitParameter(param));
                    }
                }
            }

            ServletConfig config = servlet.getServletConfig();

            if (config != null)
            {
            	@SuppressWarnings("unchecked")
            	Enumeration<String> paramNames = config.getInitParameterNames();
                while (paramNames.hasMoreElements())
                {
                    String param = paramNames.nextElement();

                    if (param.startsWith(prefix))
                    {
                        if (properties == null)
                            properties = new Properties();
                        properties.put(param.substring(prefix.length()), config.getInitParameter(param));
                    }
                }
            }
        }

        return properties;
    }

	/**
	 * Return the current web request info object.
	 */
	protected abstract LyWebRequestInfo getWebRequestInfo();
} // end class LyWebViewManager
