/*
 * LyXmlValidator.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.view.web;

import org.xml.sax.*;
import org.xml.sax.helpers.*;
import java.io.*;
import java.util.*;
import javax.xml.parsers.*;


/**
 * Debug utility class to check XHTML Validity.
 *
 * @version $Id: LyXmlValidator.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 */
public class LyXmlValidator
{
	public static final String __VERSION =
		"$Id: LyXmlValidator.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/**
	 * Debug utility class to check XHTML Validity of Html pages. Activated by DEBUG_VALIDATE_XHTML
	 * in LY_DEBUG_LEVEL.
	 *
	 * @param f file to parse and validate
	 */
	public void validate(File f)
	{
		try
		{
			// Create a "namespace aware" SAX2 XML Reader using standards JAXP
			// methods
			SAXParserFactory factory = SAXParserFactory.newInstance();

			factory.setValidating(true);
			SAXParser     saxParser = factory.newSAXParser();

			// Create an instance of our SAX event handler and associate it
			// with the XMLReader
			_LySAXHandler handler = new _LySAXHandler(f);

			// Parse the document
			saxParser.parse(f, handler);

		}
		catch (Exception e)
		{
			System.out.println(e.toString());
		}
	}

	/**
	 * Internal class to handle errors and warning during sax parsing.
	 *
	 * @version $Id: LyXmlValidator.java 28422 2010-11-03 12:03:38Z ehelmste $
	 * @author  Lyria
	 */
	class _LySAXHandler extends DefaultHandler
	{
		/** The parsed file. */
		File			  _file;

		/** Indicates if an error or a warning has already been shown to the user. */
		boolean			  _first = true;

		/** List of skipped errors (errors ignored in the output). */
		ArrayList<String> _skippedErrors;

		/**
		 * Constructor. Initialize all skipped errors.
		 *
		 * @param f the parsed file
		 */
		public _LySAXHandler(File f)
		{
			_file = f;
			_skippedErrors = new ArrayList<String>();

			// This errors are ignored in the output
			_skippedErrors.add("Attribute \"scroll\" must be declared for element type \"body\".");
			_skippedErrors.add(
				"Attribute \"topmargin\" must be declared for element type \"body\".");
			_skippedErrors.add(
				"Attribute \"leftmargin\" must be declared for element type \"body\".");
			_skippedErrors.add(
				"Attribute \"onResize\" must be declared for element type \"body\".");
			_skippedErrors.add(
				"Attribute \"type\" is required and must be specified for element type \"script\".");
			_skippedErrors.add("Attribute \"height\" must be declared for element type \"table\".");
			_skippedErrors.add("Attribute \"height\" must be declared for element type \"tr\".");
			_skippedErrors.add(
				"Attribute \"alt\" is required and must be specified for element type \"img\".");
			_skippedErrors.add("Attribute \"onMouseOut\" must be declared for element type \"a\".");
			_skippedErrors.add(
				"Attribute \"onMouseOut\" must be declared for element type \"tr\".");
			_skippedErrors.add(
				"Attribute \"onMouseOut\" must be declared for element type \"td\".");
			_skippedErrors.add(
				"Attribute \"onMouseOver\" must be declared for element type \"tr\".");
			_skippedErrors.add(
				"Attribute \"onMouseOver\" must be declared for element type \"td\".");
			_skippedErrors.add(
				"Attribute \"onMouseOver\" must be declared for element type \"a\".");
			_skippedErrors.add(
				"Attribute \"onResize\" must be declared for element type \"body\".");
			_skippedErrors.add(
				"The content of element type \"head\" is incomplete, it must match \"((script|style|meta|link|object|isindex)*,((title,(script|style|meta|link|object|isindex)*,(base,(script|style|meta|link|object|isindex)*)?)|(base,(script|style|meta|link|object|isindex)*,(title,(script|style|meta|link|object|isindex)*))))\".");
			_skippedErrors.add(
				"Attribute \"nowrap\" with value \"true\" must have a value from the list \"nowrap \".");
			_skippedErrors.add(
				"Attribute \"ly_debug\" must be declared for element type \"iframe\".");
			_skippedErrors.add(
				"Attribute \"ly_width_scroll\" must be declared for element type \"iframe\".");
			_skippedErrors.add(
				"Attribute \"ly_height_scroll\" must be declared for element type \"iframe\".");
			_skippedErrors.add(
				"Attribute \"action\" is required and must be specified for element type \"form\".");
			_skippedErrors.add(
				"Attribute \"onChange\" must be declared for element type \"select\".");
			_skippedErrors.add(
				"Attribute \"selected\" with value \"true\" must have a value from the list \"selected \".");
			_skippedErrors.add(
				"Attribute \"background\" must be declared for element type \"td\".");
			_skippedErrors.add(
				"Attribute \"marginwidth\" must be declared for element type \"body\".");
			_skippedErrors.add(
				"Attribute \"marginheight\" must be declared for element type \"body\".");
			_skippedErrors.add("Attribute \"objId\" must be declared for element type \"tr\".");
			_skippedErrors.add("Attribute \"bgColor\" must be declared for element type \"tr\".");
			_skippedErrors.add("Attribute \"bgColor\" must be declared for element type \"tr\".");
			_skippedErrors.add("Attribute \"onClick\" must be declared for element type \"tr\".");
			_skippedErrors.add(
				"Attribute \"onDblClick\" must be declared for element type \"tr\".");

		}

		/* (non-Javadoc)
		 * @see org.xml.sax.ErrorHandler#error(org.xml.sax.SAXParseException)
		 */
		public void error(SAXParseException e)
			throws SAXException
		{
			if (_first && (!_skippedErrors.contains(e.getMessage())))
			{
				_first = false;
				System.err.println("IN FILE : " + _file.getName());
			}
			if (!_skippedErrors.contains(e.getMessage()))
				System.err.println("[ERROR XHTML] " + e.getMessage() + " (" + e.getLineNumber()
					+ "," + e.getColumnNumber() + ")");
		}

		/* (non-Javadoc)
		 * @see org.xml.sax.ErrorHandler#fatalError(org.xml.sax.SAXParseException)
		 */
		public void fatalError(SAXParseException e)
			throws SAXException
		{
			if (_first && (!_skippedErrors.contains(e.getMessage())))
			{
				_first = false;
				System.err.println("IN FILE : " + _file.getName());
			}
			if (!_skippedErrors.contains(e.getMessage()))
				System.err.println("[FATAL XHTML] " + e.getMessage() + " (" + e.getLineNumber()
					+ "," + e.getColumnNumber() + ")");
		}

		/* (non-Javadoc)
		 * @see org.xml.sax.ErrorHandler#warning(org.xml.sax.SAXParseException)
		 */
		public void warning(SAXParseException e)
			throws SAXException
		{
			if (_first && (!_skippedErrors.contains(e.getMessage())))
			{
				_first = false;
				System.err.println("IN FILE : " + _file.getName());
			}
			if (!_skippedErrors.contains(e.getMessage()))
				System.err.println("[WARNING XHTML] " + e.getMessage() + " (" + e.getLineNumber()
					+ "," + e.getColumnNumber() + ")");
		}
	} // end class _LySAXHandler

} // end class LyXmlValidator
