/*
 * LyStrutsCell.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web.struts;

import java.awt.*;

import leon.misc.*;
import leon.view.*;
import leon.view.graphic.*;
import leon.view.web.*;


/**
 * This class implements the Cell object (as described in the dtd) for Struts viewer.
 * <p>
 * <b>CELL: Cell</b><br>
 * A cell may contain sub-cells to allow the construction of complex tables.
 *
 * @version $Id: LyStrutsCell.java 41220 2012-10-16 07:57:21Z nhamon $
 * @author Lyria
 * @see leon.view.LyCell
 */
public class LyStrutsCell extends LyWebObject
    implements LyGraphicalCell
{
	public static final String __VERSION =
		"$Id: LyStrutsCell.java 41220 2012-10-16 07:57:21Z nhamon $";

	/** Cell component. */
    protected LyCell           _cell;

    /** Force refresh mode. */
    protected boolean          _forceRefresh;

	/**
	 * Class constructor of a Cell object for the struts target.
	 *
	 * @param cell the generic Cell defined for every type of view targets.
	 */
	public LyStrutsCell(LyCell cell)
	{
        super(cell);
        _cell = cell;
        _forceRefresh = false;
	}

	/**
	 * Gets the Javascript code used for the HRef of the cell.
	 *
	 * @param pw       the page writer
	 * @return the String describing the Javascript code used for the HRef of the cell
	 */
	protected String getHRefScript(LyWebPageWriter pw)
	{
		return getHRefScript(pw, _cell.getObjectId());
	}

	/**
	 * Gets the Javascript code used for the HRef of the cell.
	 *
	 * @param pw       the page writer
	 * @param objectId identifier of the hyperlink
	 * @return the String describing the Javascript code used for the HRef of the cell
	 */
	protected String getHRefScript(LyWebPageWriter pw, String objectId)
	{
		StringBuffer	script = new StringBuffer();
		LyRow		 	row = getParentRow();
		LyRows		 	rows = (LyRows)row.getParent();
		LyTable		 	table = rows.getParentTable();

		objectId = LyString.getJavascriptString(objectId);

		// \n have to be splitted into two parts to be correctly sent to the Javascript function
		if (pw.writeDivOnly())
			objectId = LyString.replace(objectId, "\\n", "\\' + 'n");

		// Retrieve action URL
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);

		script.append("javascript:InvokeHRefClick('");
		script.append(objectId);
		script.append("','");
		script.append(table.getRootController());
		script.append("','");
		script.append(actionUrl);
		script.append("');");

		return script.toString();
	}

	/**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
     */
    public void setLabel(String label)
    {
		boolean	isUrl = (label != null)
			&& ((label.indexOf("http://") >= 0)
			    || (label.indexOf("https://") >= 0)
			    || (label.indexOf("mailto:") >= 0));
		
		boolean containsScript = (label != null) && (label.indexOf("</script>") >= 0);

		if (!_forceRefresh && !isUrl && !containsScript)
        {
            LyWebViewManager viewManager = getWebViewManager();
            StringBuffer     buffer = new StringBuffer();
            int[]            position = getCellPosition();
            LyRow            row = getParentRow();
            String           id = "cell_" + row.getIdCode() + "_" + position[0] + "_" + position[1];

            if (_cell.isEditable())
            {
                // If cell is editable, change the content of edition div instead
                id = "cell_edition_" + row.getIdCode() + "_" + position[1];
            }

            if (_cell.getType() != LyCell.Type.RICH)
            {
				label = LyString.replace(label, "<", "&lt;");
				label = LyString.replace(label, ">", "&gt;");
            }
            
            label = LyString.replace(label, '\n' + "", "<br/>");
            label = LyString.replace(label, "\\", "\\\\");
            label = LyString.replace(label, "\\\\n", "\\n");
            label = LyString.replace(label, "\"", "\\\"");
            label = LyString.replace(label, "\\'", "\\\\'");

            if ((label == null) || (label.length() == 0))
                label = "&nbsp;";
			else
			{
				LyRows			rows = (LyRows)row.getParent();
				LyTable			table = rows.getParentTable();
				LyRows			rootRows = table.getRows();
				LyStrutsTable	dhtmlTable = (LyStrutsTable)table.getInternalView();
				LyStrutsRow		htmlRow = (LyStrutsRow)row.getInternalView();

				if (rootRows.isHierarchical())
				{
					int	firstDataColumn = (dhtmlTable.sortOnStatus() ? 1 : 0);

					if (position[1] == firstDataColumn)
					{
						int    i, rowDepth = htmlRow.getDepthInTable();
						String blankImage = getImageUrl("LY_BLANK_ROW");

						// If row is a parent row, add an arrow to open/close the row
						if (row.getRows() != null)
							rowDepth--;

						String	images = "";

						for (i = 0; i < (rowDepth + 1); i++)
							images += "<img border='0' align='bottom' src='" + blankImage + "'/>";

						label = images + "&nbsp;" + label;
					}
				}
			}

            if (LyEnvironment.TEST_MODE)
            {
            	id = LyString.getJavascriptString(id);
            }

            buffer.append("<script type='text/javascript'>\n//<![CDATA[\nreplaceDivInnerHtml('")
                .append(_lastFrameName).append("','").append(id).append("',\"")
                .append(label).append("\");\n//]]>\n</script>");

            viewManager.addAdditionalRefreshScripts(buffer.toString());
        }
        else
            refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalCell#setObjectId(java.lang.String)
     */
    public void setObjectId(String objectId)
    {
		if ((objectId != null) && (objectId.length() > 0))
			_forceRefresh = true;

		refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetImageInterface#setImage(java.lang.String)
     */
    public void setImage(String key)
    {
        refresh();
    }

    /**
     * (non-Javadoc)
     *
     * @see leon.view.graphic.LyGraphicalCell#setEditable(boolean)
     */
    public void setEditable(boolean editable)
    {
        refresh();
    }

    /**
     * (non-Javadoc)
     *
     * @see leon.view.graphic.LyGraphicalCell#setSelected(boolean)
     */
    public void setSelected(boolean selected)
    {
        refresh();
    }

    /**
     * Gets the parent row.
     *
     * @return the parent row
     */
    protected LyRow getParentRow()
    {
        LyComponent parent = _cell;

        while (!(parent instanceof LyRow))
            parent = parent.getParent();

        return (LyRow)parent;
    }

    /**
     * Gets the parent table.
     *
     * @return the parent table
     */
    protected LyTable getParentTable()
    {
        LyComponent parent = _cell;

        while (!(parent instanceof LyTable))
            parent = parent.getParent();

        return (LyTable)parent;
    }

    /**
     * Returns the position of current cell in its parent table.
     *
     * @return an array containing row index and column index of current cell.
     */
    public int[] getCellPosition()
    {
        LyRow       row = getParentRow();
        LyStrutsRow strutsRow = (LyStrutsRow)row.getInternalView();
        LyComponent component;
        int         i, columnIndex = 0, n = row.getComponentCount();

        for (i = 0; i < n; i++)
        {
            component = row.getComponent(i);

            if (component == _cell)
            {
                columnIndex = i;
                break;
            }
        }

        return new int[] {strutsRow._rowIndex, columnIndex};
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
     */
    public void writeHtml(LyWebPageWriter pw)
    {
    }

    /**
     * Write HTML description corresponding to this cell.
     *
     * @param pw           the page writer
     * @param rowIndex     the cell rox index
     * @param columnIndex  the cell column index
     * @param rowspanValue the cell row span value
     * @param extraContent extra content to write in the cell
     */
    public void writeHtml(LyWebPageWriter pw, int rowIndex, int columnIndex, int rowspanValue,
        String extraContent)
    {
        pw.startComponent(this);

        updateRefreshInfo(pw);

        String       label, noWrap = "";

        LyRow        row = getParentRow();
        LyRows       rows = (LyRows)row.getParent();
        LyTable      table = rows.getParentTable();

        LyStrutsRow   htmlRow = (LyStrutsRow)row.getInternalView();

        String       classType = htmlRow.getRowStyle(rowIndex);
        String       additionalStyle = htmlRow.getAdditionalStyle();
        String       align = htmlRow.getAlignment();
		String		 cellAlign = LyStrutsTable.getHtmlHAligment(_cell.getAlign());
		LyHeader.Align		 headerAlign = htmlRow.getHeaderAlignment();

		if ((cellAlign != null) && (cellAlign.length() > 0))
			align = cellAlign;

		String		 cellColor = _cell.getColor();
        String       rowspan = "";

        if (rowspanValue > 1)
            rowspan = " rowspan='" + rowspanValue + "'";

		if (_cell.isDefaultColor() && _cell.isDefaultFont() && _cell.isDefaultFontColor())
            additionalStyle = htmlRow.getAdditionalStyle();
        else
		{
			additionalStyle = "";

			if (cellColor != null)
				additionalStyle += "background-color:" + getHexColor(cellColor) + ";";

			if (!_cell.isDefaultFont() || !_cell.isDefaultFontColor())
				additionalStyle += getFontString(_cell.getFont(), _cell.getFontColor());
        }

        // Check if a specific LY_TABLE resource has been set on parent table
        LyFont  tableFont = _cell.getEnvironment().getLyFont("LY_TABLE");
        LyFont  managerFont = getViewManager().getEnvironment().getLyFont("LY_TABLE");

        if ((tableFont != null)
                && !tableFont.equals(managerFont))
            additionalStyle += getFontString("LY_TABLE");

        if (_cell.isEditable())
        {
        	if (headerAlign == LyHeader.Align.RIGHT)
        		additionalStyle += "background:url(" + getImageUrl("LY_EDITABLE_CELL_MARK") + ");background-repeat:no-repeat;background-position:left top;";
        	else
        		additionalStyle += "background:url(" + getImageUrl("LY_EDITABLE_CELL_MARK") + ");background-repeat:no-repeat;background-position:right top;";
        }

        // Add cell borders
        if (_cell.getLeftBorderColor() != null)
        	additionalStyle += "border-left:1px solid " + getHexColor(_cell.getLeftBorderColor()) + ";";

        if (_cell.getRightBorderColor() != null)
        	additionalStyle += "border-right:1px solid " + getHexColor(_cell.getRightBorderColor()) + ";";

        if (_cell.getTopBorderColor() != null)
        	additionalStyle += "border-top:1px solid " + getHexColor(_cell.getTopBorderColor()) + ";";

        if (_cell.getBottomBorderColor() != null)
        	additionalStyle += "border-bottom:1px solid " + getHexColor(_cell.getBottomBorderColor()) + ";";

        if ((_cell.getTopLeftImage() != null)
        		|| (_cell.getTopRightImage() != null)
        		|| (_cell.getBottomLeftImage() != null)
        		|| (_cell.getBottomRightImage() != null))
        	additionalStyle += "position:relative;";

        if ((additionalStyle != null) && (additionalStyle.length() > 0))
            additionalStyle = " style=\"" + additionalStyle + "\"";

        // get cell label
        label = _cell.getLabel();
        label = LyString.getHtmlString(label);
        if (label.length() == 0)
            label = "&nbsp;";

        // Check label with header width
        String    displayedLabel = _cell.getLabel();
        LyHeaders headers = table.getHeaders();
        boolean   labelReduced = false;
        String    htmlLabel = label;
        String    tooltip = _cell.getTooltip();
		boolean   isWrap = false;

        if (tooltip != null)
            tooltip = translate(tooltip);

		if ((headers != null) && !headers.isFreed() && !headers.isFreeing())
		{
            LyHeader      header = headers.getHeader(columnIndex);

            if (header != null)
            {
				if ((displayedLabel != null) && (displayedLabel.length() > 0))
				{
		            LyStrutsHeader strutsHeader = (LyStrutsHeader)header.getInternalView();
		            String        fullLabel = displayedLabel + "";
		            String		  labelReducedValue = strutsHeader.checkHeaderWidth(displayedLabel, "LY_TABLE", false);

		            if (header.getColumnWidth() >= 1)
		                _forceRefresh = true;

		            if (_cell.getType() != LyCell.Type.RICH)
		            	displayedLabel = labelReducedValue;

		            if (!labelReducedValue.equals(fullLabel))
		                labelReduced = true;
				}

				isWrap = header.isWrap();
            }
        }

		if (!isWrap)
			noWrap = " nowrap='true'";

        pw.append("<td valign='middle'");
        pw.append(" class='");
        pw.append(classType);
        pw.append("'");

        pw.append(additionalStyle);
        pw.append(rowspan);

        pw.append(align);
        pw.append(noWrap);

        if (labelReduced && (tooltip == null))
        {
        	if (_cell.getType() != LyCell.Type.RICH)
        		tooltip = htmlLabel;
        	else
        		tooltip = displayedLabel;
        }

        if (labelReduced)
            htmlLabel = LyString.getHtmlString(displayedLabel);

        if (htmlLabel.length() == 0)
            htmlLabel = "&nbsp;";

        if (tooltip != null)
        {
            tooltip = LyString.replace(tooltip, "&quot;", "\"");
            tooltip = LyString.replace(tooltip, "<br />", "\n");
            tooltip = LyTools.removeHtmlTags(tooltip);
            pw.append(" title='");
            pw.append(LyString.replace(tooltip, "'", "&#39;"));
            pw.append("'");
        }

        if (!_forceRefresh)
        {
            pw.append(" id='cell_");
            pw.append(getParentRow().getIdCode() + "_" + rowIndex + "_" + columnIndex);
            pw.append("'");
        }

        // Check whether cell is editable or not.
        boolean     hasComponents = htmlRow.hasComponents(true);

        if (hasComponents && (table.getSelectionMode() != LyTable.SelectionMode.NONE))
        {
            pw.append(" onClick=\"");
            pw.append(htmlRow.getClickScript(pw, rowIndex));

            if (_cell.isEditable())
                pw.append(htmlRow.getCellEditionScript(pw, rowIndex, columnIndex));

            pw.append("\"");
        }

        if (_cell.isEditable())
        {
            pw.append(" onDblClick=\"");
            pw.append(htmlRow.getCellEditionScript(pw, rowIndex, columnIndex));
            pw.append("\"");
        }
        else if (hasComponents || LyStrutsRow.USE_COLUMN_INDEX_FOR_DOUBLE_CLICK)
        {
            pw.append(" onDblClick=\"");
            pw.append(htmlRow.getDoubleClickScript(pw, rowIndex, columnIndex));
            pw.append("\"");
        }

        pw.append(" onContextMenu=\"lastCellSelected=" + columnIndex + ";\"");
        pw.append(">");

        // Add cell images
        if (_cell.getTopLeftImage() != null)
      	  pw.append("<img src='" + getImageUrl(_cell.getTopLeftImage()) + "' style='position:absolute;top:0px;left:0px;'/>");

        if (_cell.getTopRightImage() != null)
      	  pw.append("<img src='" + getImageUrl(_cell.getTopRightImage()) + "' style='position:absolute;top:0px;right:0px;'/>");

        if (_cell.getBottomLeftImage() != null)
      	  pw.append("<img src='" + getImageUrl(_cell.getBottomLeftImage()) + "' style='position:absolute;bottom:0px;left:0px;'/>");

        if (_cell.getBottomRightImage() != null)
      	  pw.append("<img src='" + getImageUrl(_cell.getBottomRightImage()) + "' style='position:absolute;bottom:0px;right:0px;'/>");

        if (_cell.isEditable())
        {
        	boolean		isIe = (getWebViewManager().getBrowser() == LyWebViewManager.BROWSER_MSIE);

            pw.append("<div style='position:relative;");

            if (!isIe)
        		pw.append("width:100%;");

            pw.append("'>");
            pw.append("<div id='cell_edition_");
            pw.append(row.getIdCode() + "_" + columnIndex);
            pw.append("'");

            if (!isIe)
                pw.append(" style='float:" + headerAlign + ";'");

            pw.append(">");
        }

        writeCellContent(pw, columnIndex);

        if (extraContent != null)
            pw.append(extraContent);

        if (_cell.isEditable())
        {
            pw.append("</div>");
            pw.append("</div>");
        }

        pw.append("</td>\n");

        pw.endComponent(this);
    } // end method writeHtml

    /**
     * Write HTML content of this cell.
     *
     * @param pw           the page writer
     * @param columnIndex  the cell column index
     */
    @SuppressWarnings("deprecation")
	public void writeCellContent(LyWebPageWriter pw, int columnIndex)
    {
        updateRefreshInfo(pw);

        String       label, objectId;
        boolean      isFile, relation, multiRelation, isMail, isUrl, isUrlRecurse;
        String       javascriptToolTip;
        String       target = ""; // Use default value for target

        LyRow        row = getParentRow();
        LyRows       rows = (LyRows)row.getParent();
        LyTable      table = rows.getParentTable();

        LyStrutsTable strutsTable = (LyStrutsTable)table.getInternalView();
        LyStrutsRow   htmlRow = (LyStrutsRow)row.getInternalView();

        String       additionalStyle = htmlRow.getAdditionalStyle();

        String       cellColor = _cell.getColor();
        String       font = _cell.getFont();

        String       image = _cell.getImage();
        String       imageUrl = null;
        String       docDir;
        String       docRoot = expandUrl(getEnv("LY_APP_DOC"));

        LyWebDecoration deco = getWebViewManager().getDecoration();
        String          hyperlinkClass = deco.getClassName(_cell, pw);

        if ((image != null) && (image.length() > 0))
        {
            image = expandUrl(image);
            docDir = expandUrl(getEnv("LY_APP_DOC_DIR"));

            if (image.startsWith(docDir))
            {
                image = image.substring(docDir.length());
                image = getEnv("LY_APP_DOC") + image;
            }
            else
            {
                docDir = expandUrl(getEnv("LY_LEON_DOC_DIR"));
                if (image.startsWith(docDir))
                {
                    image = image.substring(docDir.length());
                    image = getEnv("LY_LEON_DOC") + image;
                }
            }

            if (image.startsWith(docRoot))
                imageUrl = expandUrl(image);
            else if (image.startsWith("file"))
                imageUrl = LyString.encodeFileURL(image);
            else if (image.startsWith("http") || image.startsWith("ftp"))
                imageUrl = image;
            else
                imageUrl = getImageUrl(image);

            LyEnvironment environment = getEnvironment();

            imageUrl = "<img border='0'  align='middle' src='"
                + environment.getPathWithoutSize(imageUrl) + "'";

            int[] size = getImageSize(_cell.getImage());

            if (size != null)
            {
                if (size.length == 2)
                    imageUrl += " width='" + size[0] + "' height='" + size[1] + "'";
                else
                    imageUrl += " height='" + size[0] + "'";
            }

            imageUrl += "/>";
        }

        if ((cellColor == null) && (font == null))
            additionalStyle = htmlRow.getAdditionalStyle();
        else
        {
            additionalStyle = "";
            if (cellColor != null)
                additionalStyle += "background-color:" + getHexColor(cellColor) + ";";

            if (font != null)
                additionalStyle += getFontString(font);
        }

		boolean		 generateTable = (imageUrl != null);

		if (generateTable && (font == null))
			additionalStyle += getFontString("LY_TABLE");

        if ((additionalStyle != null) && (additionalStyle.length() > 0))
            additionalStyle = " style=\"" + additionalStyle + "\"";

        // get cell label
        label = _cell.getLabel();

        String htmlLabel = (_cell.getType() == LyCell.Type.RICH ? label : LyString.getHtmlString(label));

        label = htmlLabel;

        if ((label == null) || (label.length() == 0))
            label = "&nbsp;";

        // Check label with header width
        String    displayedLabel = _cell.getLabel();
        LyHeaders headers = table.getHeaders();
		int		  columnWidth = -1;

		if ((headers != null) && !headers.isFreed() && !headers.isFreeing() && (displayedLabel != null) && (displayedLabel.length() > 0))
		{
			LyHeader	   header = headers.getHeader(columnIndex);

			if (header != null)
			{
				LyStrutsHeader dhtmlHeader = (LyStrutsHeader)header.getInternalView();
				double		   headerWidth = header.getColumnWidth();

				if ((dhtmlHeader != null) && (_cell.getType() != LyCell.Type.RICH))
					displayedLabel = dhtmlHeader.checkHeaderWidth(displayedLabel, "LY_TABLE", false);

				if (headerWidth >= 1)
				{
					columnWidth = (int)header.getColumnWidth();

					// If header is sorted, add 2 characters for cell content
					if (header.getSort() >= 0)
						columnWidth += 2;
				}
			}

			if (!displayedLabel.equals( _cell.getLabel() ))
				htmlLabel =  LyString.getHtmlString(displayedLabel);
		}

        if ((htmlLabel == null) || (htmlLabel.length() == 0))
            htmlLabel = "&nbsp;";

        objectId = _cell.getObjectId();
        relation = (objectId != null) && (objectId.length() != 0);
		multiRelation = relation && (objectId.indexOf(LyTools.VALUE_SEP) >= 0) && (label.indexOf(LyTools.VALUE_SEP) >= 0);
        isFile = (_cell.getHLink() != null);
        isMail = false;
        // FFT00180/IpBoss : strings containing @
        if ((label.indexOf(' ') < 0) && (label.indexOf('@') > 0)
                && (label.indexOf("mailto:") < 0))
        {
            // Checks that the domain part of the label is an IP address
            String domain = label.substring(label.lastIndexOf('@') + 1);
            String[] domainParts = domain.split("\\.");
            boolean isIp = true;
            if (domainParts.length == 4)
            {
                for (int i = 0; i < domainParts.length && isIp; i++)
                {
                    String part = domainParts[i];
                    isIp &= part.matches("(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)");
                }
            }
            // If the domain is not an IP address, then the label is an email
            if (!isIp)
                isMail = true;
        }

        // contains one or more http:// or starts with www.
        isUrlRecurse = (label.indexOf("http://") >= 0)
                || (label.indexOf("https://") >= 0)
        		|| (label.indexOf("www.") > 0)
                || ((label.indexOf("mailto:") >= 0)
                    && !isMail);
        isUrl =  isUrlRecurse || (label.indexOf("www.") == 0);

        if (_cell.getType() == LyCell.Type.RICH)
        {
        	isFile = false;
        	isUrl = false;
        	isMail = false;
        	isUrlRecurse = false;
        }

        if (isFile || isMail || isUrl || relation || (imageUrl != null) || (_cell.getType() == LyCell.Type.RICH))
            _forceRefresh = true;

        LyRows rootRows = table.getRows();

        // If table is hierarchical, add space and arrow if necessary
        if (rootRows.isHierarchical())
        {
            int firstDataColumn = (strutsTable.sortOnStatus() ? 1 : 0);

            if (columnIndex == firstDataColumn)
            {
                int    i, rowDepth = htmlRow.getDepthInTable();
                String blankImage = getImageUrl("LY_BLANK_ROW");
                String closedRowImage = getImageUrl("LY_CLOSED_ROW");
                String openRowImage = getImageUrl("LY_OPEN_ROW");

                // If row is a parent row, add an arrow to open/close the row
                if (row.getRows() != null)
                    rowDepth--;

                for (i = 0; i < rowDepth; i++)
                    pw.append("<img border='0' align='bottom' src='" + blankImage + "'/>");

                if (row.getRows() != null)
                {
                    boolean      isOpen = row.isOpen();
                    String       expandRowScript = htmlRow.getExpandRowScript(pw);

                    if (isOpen)
                        javascriptToolTip = getJavascriptString("LY_CLOSE_ROW");
                    else
                        javascriptToolTip = getJavascriptString("LY_OPEN_ROW");

                    if (!pw.isPrintMode() && (expandRowScript != null))
                    {
                        pw.append("<a href='" + expandRowScript.toString());
                        pw.append("' onmouseover=\'window.status=\"");
                        pw.append(javascriptToolTip);
                        pw.append("\";return true\'>");
                    }

                    pw.append("<img border='0' align='bottom' src='"
                        + (isOpen ? openRowImage : closedRowImage) + "'/>");

                    if (!pw.isPrintMode() && (expandRowScript != null))
                        pw.append("</a>");
                }
                if (!generateTable)
                  pw.append("&nbsp;");
            }
        }

		StringBuffer linkBuffer = null;
		if (isFile && !pw.isPrintMode())
        {
            javascriptToolTip = getJavascriptString("LY_LINK_TO" + " " + _cell.getLabel());
            if (LyWebViewManager.WEB_FILE_NEW_WINDOW)
                target = "_blank";
            else
                target = LyWebConstants.TARGET_FRAME;

			linkBuffer = new StringBuffer();
			linkBuffer.append("<a class='" + hyperlinkClass + "' href='");
			linkBuffer.append(_cell.getHLink());
			linkBuffer.append("'");
			linkBuffer.append(" target='");
			linkBuffer.append(target);
			linkBuffer.append("' onmouseover=\'window.status=\"");
			linkBuffer.append(javascriptToolTip);
			linkBuffer.append("\";return true\'>");
            isUrlRecurse = false;
        }
		else if (multiRelation && !pw.isPrintMode()
				&& testResource("LY_WEB_TABLE_TRANSLATE_RELATION_TO_LINK"))
		{
			String[]	ids = LyString.stringToList(objectId, LyTools.VALUE_SEP);
			String[]	htmlLabels = LyString.stringToList(htmlLabel, LyTools.VALUE_SEP);
			int			i, n = (ids == null ? 0 : ids.length);

			for (i = 0; i < n; i++)
			{
				pw.append("<a class='" + hyperlinkClass + "' href=\"");
				pw.append(getHRefScript(pw, ids[i]));
				pw.append("\">");
				pw.append(htmlLabels[i]);
				pw.append("</a>");
			}
		}
        else if (relation && !pw.isPrintMode()
                && testResource("LY_WEB_TABLE_TRANSLATE_RELATION_TO_LINK"))
        {
            javascriptToolTip = getJavascriptString("LY_LINK_TO" + " " + _cell.getLabel());
			linkBuffer = new StringBuffer();
			linkBuffer.append("<a class='" + hyperlinkClass + "' href=\"");
			linkBuffer.append(getHRefScript(pw));
			linkBuffer.append("\" onmouseover=\'window.status=\"");
			linkBuffer.append(javascriptToolTip);
			linkBuffer.append("\";return true\'>");
        }
        else if (isMail && !pw.isPrintMode())
        {
			linkBuffer = new StringBuffer();
			linkBuffer.append("<a class='" + hyperlinkClass + "' href='mailto:" + label + "'>");
        }
        else if (isUrl)
        {
            if (LyWebViewManager.WEB_FILE_NEW_WINDOW)
                target = "_blank";
            else
                target = LyWebConstants.TARGET_FRAME;

			linkBuffer = new StringBuffer();

			if (isUrlRecurse)
				pw.append(replaceUrl(label, target, pw.isPrintMode(), columnWidth, "LY_TABLE"));
            else if (!pw.isPrintMode())
            	linkBuffer.append("<a class='" + hyperlinkClass + "' href='http://" + label + "' target=\"" + target + "\">");
        }

        if (imageUrl != null)
		{
			if (generateTable)
				pw.append("<table cellpadding='0' cellspacing='0'><tr><td width='1' valign='middle'>");

			if (linkBuffer != null)
				pw.append(linkBuffer.toString());

			pw.append(imageUrl);

			if (linkBuffer != null)
				pw.append("</a>");

			if (generateTable)
				pw.append("</td>");
		}

		if (!isUrlRecurse
				&& (!multiRelation || pw.isPrintMode() || !testResource("LY_WEB_TABLE_TRANSLATE_RELATION_TO_LINK")))
        {
			if (generateTable)
				pw.append("<td valign='middle' " + additionalStyle + ">");

			if ((_cell.getType() == LyCell.Type.RICH) && (columnWidth >= 1))
			{
	            LyEnvironment environment = getEnvironment();
	            Font          defaultFont = (Font)environment.getFont("LY_TABLE");
	            Toolkit       toolkit = Toolkit.getDefaultToolkit();
	            FontMetrics   defaultFm = toolkit.getFontMetrics(defaultFont);
	            int           charWidth = defaultFm.stringWidth("0123456789") / 10;

	            pw.append("<div style='width:" + (charWidth * columnWidth) + "px;overflow:hidden;'>");
			}

			if (linkBuffer != null)
				pw.append(linkBuffer.toString());

            pw.append(htmlLabel);

			if (linkBuffer != null)
				pw.append("</a>");

			if ((_cell.getType() == LyCell.Type.RICH) && (columnWidth >= 1))
				pw.append("</div>");

			if (generateTable)
				pw.append("</td></tr></table>");
        }
		else if (generateTable)
			pw.append("</tr></table>");
    } // end method writeCellContent

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetColorInterface#setColor(java.lang.String)
     */
    public void setColor(String color)
    {
        if (!_forceRefresh)
        {
            String           cellColor = ((color == null) ? "" : getHexColor(color));
            LyWebViewManager viewManager = getWebViewManager();
            StringBuffer     buffer = new StringBuffer();
            int[]            position = getCellPosition();
            String           id = "cell_" + getParentRow().getIdCode() + "_" + position[0] + "_" + position[1];
            LyTable			 table = getParentTable();
            String           selectColor = getHexColor(table.getSelectedColor());
            String    		 selectFontColor = getHexColor(table.getSelectedFontColor());

            // Update the color of the cell
            buffer.append("<script type='text/javascript'>\n//<![CDATA[\nSetObjectColor('").append(
                _lastFrameName).append("','").append(id).append("','").append(cellColor).append(
                "');\n//]]>\n</script>");

            viewManager.addAdditionalRefreshScripts(buffer.toString());

            // Add script used to update rows color
            StringBuffer    updateBuffer = new StringBuffer();

            updateBuffer.append("<script type='text/javascript'>\n//<![CDATA[\nUpdateRowsColors('");
            updateBuffer.append(selectColor);
            updateBuffer.append("','").append(_lastFrameName).append("','" + selectFontColor + "');\n//]]>\n</script>");

            viewManager.addAdditionalRefreshScripts(updateBuffer.toString());
        }
        else
            refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetFontInterface#setFont(java.lang.String)
     */
    public void setFont(String font)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalCell#addCell(leon.view.LyCell, int)
     */
    public void addCell(LyCell cell, int pos)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalCell#removeCell(leon.view.LyCell)
     */
    public void removeCell(LyCell cell)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalCell#setTooltip(java.lang.String)
     */
    public void setTooltip(String tooltip)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalCell#setBottomBorderColor(java.lang.String)
     */
    public void setBottomBorderColor(String bottomBorderColor)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalCell#setLeftBorderColor(java.lang.String)
     */
    public void setLeftBorderColor(String leftBorderColor)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalCell#setRightBorderColor(java.lang.String)
     */
    public void setRightBorderColor(String rightBorderColor)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalCell#setTopBorderColor(java.lang.String)
     */
    public void setTopBorderColor(String topBorderColor)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalCell#setBottomLeftImage(java.lang.String)
     */
    public void setBottomLeftImage(String bottomLeftImage)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalCell#setBottomRightImage(java.lang.String)
     */
    public void setBottomRightImage(String bottomRightImage)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalCell#setTopLeftImage(java.lang.String)
     */
    public void setTopLeftImage(String topLeftImage)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalCell#setTopRightImage(java.lang.String)
     */
    public void setTopRightImage(String topRightImage)
    {
        refresh();
    }

    /**
     * Overriden to call the refresh method on the parent row.
     */
    public void refresh()
    {
        LyTable table = getParentTable();

        if (table != null)
        {
            LyWebObject strutsTable = (LyWebObject)table.getInternalView();

            strutsTable.refresh();
        }
        else
            super.refresh();
    }
} // end class LyStrutsCell
