/*
 * LyStrutsFieldItem.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web.struts;

import leon.view.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.web.*;


/**
 * This class implements the common behavior for field items objects for the HTML viewer.
 *
 * <p>Fields items are implemented in Struts by those classes : LyStrutsText, LyStrutsNumber,
 * LyStrutsDate, LyStrutsChoice, LyStrutsOption, LyStrutsRelation, LyStrutsFile, LyStrutsForm.</p>
 *
 * @see     leon.view.web.struts.LyStrutsText
 * @see     leon.view.web.struts.LyStrutsNumber
 * @see     leon.view.web.struts.LyStrutsDate
 * @see     leon.view.web.struts.LyStrutsChoice
 * @see     leon.view.web.struts.LyStrutsOption
 * @see     leon.view.web.struts.LyStrutsRelation
 * @see     leon.view.web.struts.LyStrutsFile
 * @see     leon.view.web.struts.LyStrutsForm
 * @version $Id: LyStrutsFieldItem.java 16172 2009-01-06 11:04:44Z jls $
 * @author  Lyria
 * @see     leon.view.LyFieldItem
 */
public abstract class LyStrutsFieldItem extends LyWebObject
	implements LyGraphicalFieldItem,
		LySetLabelInterface
{
	public static final String __VERSION =
		"$Id: LyStrutsFieldItem.java 16172 2009-01-06 11:04:44Z jls $";

	/** The form field component that contains this item. */
	protected LyFormField	   _formField;

	/** The name of the html form where this item is generated. */
	protected String		   _formName;

	/**
	 * Class constructor of a Field Item object for the struts target.
	 *
	 * @param fieldItem the generic Field Item defined for every type of view targets
	 */
	protected LyStrutsFieldItem(LyFieldItem fieldItem)
	{
		super(fieldItem);

		LyComponent parent = fieldItem.getParent();

		if (parent instanceof LyFormField)
			_formField = (LyFormField)parent;
		else if (parent.getParent() instanceof LyFormField)
			_formField = (LyFormField)parent.getParent();
		else
			_formField = null;

		_formName = null;
	}

	/**
	 * Gets the field item.
	 *
	 * @return the field item
	 */
	public LyFieldItem getFieldItem()
	{
		return (LyFieldItem)_viewComponent;
	}

	/**
	 * Gets the form field that contains this item.
	 *
	 * @return the form field
	 */
	public LyFormField getFormField()
	{
		return _formField;
	}

	/**
	 * Gets the id of the form field that contains this item.
	 *
	 * @return the id
	 */
	public String getId()
	{
		return _formField.getId();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalObject#getController()
	 */
	public String getController()
	{
		if (_formField.getParent() instanceof LyGetControllerInterface)
			return ((LyGetControllerInterface)_formField.getParent()).getController();

		return null;
	}

	/**
	 * Gets the caption of the parent form field.
	 *
	 * @return the caption
	 */
	public String getCaption()
	{
		LyCaption caption = _formField.getCaption();

		return ((caption != null) ? caption.getLabel() : _formField.getId());
	}

	/**
	 * Gets the form ancestor.
	 *
	 * @return the form ancestor
	 */
	public LyForm getForm()
	{
		// Look for an ancestor of type LyForm
		LyComponent ancestor = _formField;

		do
			ancestor = ancestor.getParent();
		while ((ancestor != null) && !(ancestor instanceof LyForm));

		return (LyForm)ancestor;
	}

	/**
	 * Gets the Struts form.
	 *
	 * @return the web form that contains this item
	 */
	public LyStrutsForm getHtmlForm()
	{
		LyForm form = getForm();

		return ((form == null) ? null : (LyStrutsForm)form.getInternalView());
	}

	/**
	 * Gets the ancestor form name.
	 *
	 * @param  pw the page writer
	 * @return the form name if found, the String "default" otherwise
	 */
	public String getFormName(LyWebPageWriter pw)
	{
		// If formName was already registered, return it directly.
		// May be needed for Ajax requests.
		if (_formName != null)
			return _formName;

		String formName = null;

		if (!(_formField.getParent() instanceof LyForm)) // Editable table
			formName = pw.getActiveFormId();
		else
		{
			LyStrutsForm form = getHtmlForm();

			formName = ((form == null) ? "default" : form.getName());
		}

		// '.' in form names may not work in Javascript commands.
		// Replace it by '_'.
		if (formName != null)
			formName = formName.replace('.', '_');

		_formName = formName;

		return formName;
	}

	/**
	 * Indicates if the Web browser used is Netscape.
	 *
	 * @return true if it is Netscape, false otherwise
	 */
	public boolean isNetscape()
	{
		LyWebViewManager viewManager = getWebViewManager();

		return (viewManager.getBrowser() == LyWebViewManager.BROWSER_NETSCAPE);
	}

	/**
	 * Indicates if the Web browser can disabled its elements.
	 *
	 * @return true if it can, false otherwise
	 */
	public boolean canDisabled()
	{
		LyWebViewManager viewManager = getWebViewManager();

		return (!isNetscape() || (viewManager.getBrowserVersion() >= 5.0));
	}

	/**
	 * Indicates if the field item is editable.
	 *
	 * @return true if it is editable, false otherwise
	 */
	public boolean isEditable()
	{
		if (!_formField.isEditable() || !_formField.isEnabled())
			return false;

		LyForm form = ((_formField.getParent() instanceof LyForm) ? (LyForm)_formField.getParent()
																  : null);

		return ((form == null) || form.isEditable());
	}

	/**
	 * Indicates if the field item is optional.
	 *
	 * @return true if it is optional, false otherwise
	 */
	public boolean isOptional()
	{
		return _formField.isOptional();
	}

	/**
	 * Indicates if the field item needs check.
	 *
	 * @return true if it needs check, false otherwise
	 */
	public boolean needCheck()
	{
		return _formField.isNeedCheck();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
	 */
	public void setLabel(String label)
	{
		refresh();
	}
} // end class LyStrutsFieldItem
