/*
 * LyStrutsFile.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.view.web.struts;

import java.awt.*;
import java.io.*;

import leon.misc.*;
import leon.view.*;
import leon.view.graphic.*;
import leon.view.web.*;


/**
 * This class implements the File object (as described in the dtd) for web viewer.
 *
 * <p><b>FILE: Input Field of File Type</b><br>
 * An input field of File type allows to display a graphical component allowing to select a file.
 * The type of file can be specified (text, image, audio, video). According to the form in which
 * this field is present, the graphical component allows to choose the location of the file and/or
 * to view the access path to the file.</p>
 *
 * @version $Id: LyStrutsFile.java 35023 2012-03-23 15:50:24Z nhamon $
 * @author  Lyria S.A.
 */
public class LyStrutsFile extends LyStrutsFieldItem
    implements LyGraphicalFile
{
	public static final String __VERSION =
		"$Id: LyStrutsFile.java 35023 2012-03-23 15:50:24Z nhamon $";

	/** Indicates whether the name or full path of previously selected file must be displayed. */
	public static boolean DISPLAY_PREVIOUS_FILE_NAME = false;

	/** The file generic component. */
    protected LyFile           _file;

	/**
	 * Constructor.
	 *
	 * @param file the file component
	 */
	public LyStrutsFile(LyFile file)
	{
        super(file);

        _file = file;
	}

	/**
	 * Gets the additional HTML code placed at the right of the file input.
	 *
	 * @param pw the page writer
	 */
	protected String getAdditionalHtml(LyWebPageWriter pw)
	{
		return null;
	}

	/**
	 * Gets the Javascript code used when changing the value of the file.
	 *
	 * @param pw the page writer
	 * @return the String describing the Javascript code when changing the value of the file
	 */
	protected String getOnChangeScript(LyWebPageWriter pw)
	{
		if (_file.getAttachmentHandler() == null)
			return null;

		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String				uploadUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_HOME);
		String				fileErrorMessage = getMessage("LY_FORM_FILE_ERROR");
		String				formName = "forms[0]";//getFormName(pw);

		if (pw.writeDivOnly())
			fileErrorMessage = LyString.replace(fileErrorMessage, "\\n", "\\\\n");

		String	sendFileScript = "SendFile(window.document." + formName + ",'"
			+ _file.getRootController() + "','" + getId() + "','file_frame_" + _file.getParentForm().getIdCode() + "','" + fileErrorMessage + "','"
			+ uploadUrl + "','" + LyWebConstants.CONTENT_MULTIPART + "','" + viewManager.getTargetCode() + "'";
		
		if (viewManager.getTargetCode() != null)
			sendFileScript += ",'" + viewManager.getTargetCode() + "'";
		
		sendFileScript += ");";

		return sendFileScript;
	}

	/**
	 * Returns the name to display for previously selected file.
	 * @return the full path of the file if <code>DISPLAY_PREVIOUS_FILE_NAME</code> equals <code>false</code>,
	 * file name otherwise.
	 */
	protected String getPreviousFileName()
	{
		String result = _file.getLabel();

		if ((result != null) && DISPLAY_PREVIOUS_FILE_NAME)
		{
			String	sep = File.separator;
			int		pos = result.lastIndexOf(sep);

			if (pos >= 0)
				result = result.substring(pos + sep.length());
		}

		return result;
	}

	/**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
     */
    public void setLabel(String label)
    {
        LyComponent parent = _file.getParent();

        if (parent instanceof LyFormField)
        {
            LyFormField       formField = (LyFormField)parent;
            LyStrutsFormField webField = (LyStrutsFormField)formField.getInternalView();

            webField.refreshDivContent();
        }
        else
            refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
     */
    public void writeHtml(LyWebPageWriter pw)
    {
        pw.startComponent(this);

        updateRefreshInfo(pw);

        LyStrutsForm    form = getHtmlForm();

        // Check encoding type : if a handler is specify, force file upload.
        if ((form != null) && (_file.getAttachmentHandler() != null))
        {
        	String	controller = _file.getRootController();
        	
            form.setAttachmentHandler(_file.getAttachmentHandler());

            if (controller != null)
        		pw.setAttachmentHandler(controller, _file.getAttachmentHandler());
        	else
        		pw.setAttachmentHandler(_file.getAttachmentHandler());
        		
            if (_file.getAttachmentDirectory() != null)
            {
                form.setAttachmentDirectory(_file.getAttachmentDirectory());

                if (controller != null)
            		pw.setAttachmentDirectory(controller, _file.getAttachmentDirectory());
            	else
            		pw.setAttachmentDirectory(_file.getAttachmentDirectory());
            }

            // Try to set attachment handler and directory on the main page writer
            LyView  view = _file.getParentView();

            if (view != null)
            {
                LyWebPageWriter pageWriter = (LyWebPageWriter)view.getInternalView();

                if (controller != null)
                	pageWriter.setAttachmentHandler(controller, _file.getAttachmentHandler());
            	else
            		pageWriter.setAttachmentHandler(_file.getAttachmentHandler());

                if (_file.getAttachmentDirectory() != null)
                {
                    if (controller != null)
                    	pageWriter.setAttachmentDirectory(controller, _file.getAttachmentDirectory());
                	else
                		pageWriter.setAttachmentDirectory(_file.getAttachmentDirectory());
                }
            }
        }

		String  style = "";

		if (!_file.isDefaultFont() || !_file.isDefaultFontColor())
			style += getFontString(_file.getFont(),_file.getFontColor());

		if (isEditable())
		{
			if (!_file.isDefaultEnabledColor())
			{
				style += " background-color:" + getHexColor(_file.getEnabledColor()) + ";";
				style += " background-image:none;";
			}
		}
		else
		{
			if (!_file.isDefaultDisabledColor())
			{
				style += " background-color:" + getHexColor(_file.getDisabledColor()) + ";";
				style += " background-image:none;";
			}
		}

		String label = _file.getLabel();
		String previousFileLabel = getPreviousFileName();

        // If not editable, just write text
        if (!isEditable())
        {
            if (previousFileLabel != null)
                pw.append("<font class='label'>" + previousFileLabel + "</font>\n");

            pw.endComponent(this);
            return;
        }

        String formName = "forms[0]";//getFormName(pw);
        String name = getId();

        // Previous file value is not shown and may be lost when modifying an
        // object.
        String  previous = null;
        boolean isDirectory = (_file.getType() == LyFile.Type.DIRECTORY);

        // Do not generate checkbox used to empty value if field is not optional.
        boolean generateCheckbox = ((_formField == null) || _formField.isOptional());
        int     colspan = 1;

        if (label == null)
            label = "";

	    if (previousFileLabel == null)
	    	previousFileLabel = "";

        if (previousFileLabel.length() != 0)
        {
            String  resource = (generateCheckbox ? "LY_PREVIOUS_FILE_VALUE_CHECKBOX" : "LY_PREVIOUS_FILE_VALUE");

            if (previousFileLabel.startsWith("__"))
            {
            	int n1 = -1, n2 = -1;
            	n1 = previousFileLabel.indexOf("__");
            	if (n1 >= 0)
            	{
            		n2 = previousFileLabel.indexOf("__", n1+1);
            	}
            	
            	if ((n1 >= 0) && (n2 >= 0))
            	{
            		String newLabel = previousFileLabel.substring(0, n1) + previousFileLabel.substring(n2 + 2);
            		previous =  getString(resource, newLabel);
            	}
            	else
            		previous = getString(resource, previousFileLabel);
            }
            else
                previous = getString(resource, previousFileLabel);
        }
        else
            previous = getString("LY_PREVIOUS_FILE_EMPTY", previousFileLabel);

        pw.append("<table border='0' cellpadding='0' cellspacing='0'><tr>");

		if (generateCheckbox
				&& (previousFileLabel.length() > 0)
				&& (previous != null)
				&& (previous.trim().length() > 0))
        {
            // Generate a check box used to empty file value
            pw.append("<td><input type='checkbox' name='" + name + "_empty_value' value='1'");

            if (canDisabled())
            {
                pw.append(" onClick=\"DisableFileInput(window.document." + formName + ",'"
                        + name + "',this.checked);\"");
            }

            pw.append("/></td>");
            colspan++;
        }

        int     width = _file.getNCols();
		int		height = -1;

		// If a constraint is set on the file, use the nCols and nRows values to compute
		// file size in pixels.
		if ((_file.getParent() instanceof LyFormField)
				&& (((LyFormField)(_file.getParent())).getConstraint() != null))
		{
			LyConstraint	constraint = ((LyFormField)(_file.getParent())).getConstraint();

			if (constraint != null)
			{
				if (constraint.getNCols() > 0)
					width = constraint.getNCols();

				if (constraint.getNRows() > 0)
					height = constraint.getNRows();
			}
		}

		if (width > 0)
			style += "width:" + width + "px;";
		else
		{
			int		nCols = getIntResource("LY_FILE_SELECTOR_WIDTH");

			if (nCols < 0)
				nCols = 40;

			style += "width:" + getFileWidth(nCols) + "px;";
		}

		if (height > 0)
			style += "height:" + height + "px;";

		if (style.trim().length() > 0)
			style =" style=\""+style+"\"";

		pw.append("<td><input name=\"" + name + "\" type=\"hidden\""
            + " value=\"" + label + "\" />\n");
		pw.append("<input "+ style +" name=\"" + name + "File\" type=\"file\""
            + " value=\"" + label + "\"");
        pw.append(" onBlur=\"UpdateFile(window.document." + formName + ",'"
            + name + "'," + isDirectory + ")\"");

        String  onChangeScript = getOnChangeScript(pw);
        String  additionalHtml = getAdditionalHtml(pw);

		pw.append(" onChange=\"");

		if (onChangeScript != null)
			pw.append(onChangeScript);

		pw.append("UpdateFile(window.document." + formName + ",'" + name + "'," + isDirectory + ");");
		
		pw.append("\"");
        pw.append("/>\n");
        pw.append("</td>");

        if (additionalHtml != null)
        {
            colspan++;
            pw.append("<td>");
            pw.append(additionalHtml);
            pw.append("</td>");
        }

        pw.append("</tr>");

		if ((previous != null) && (previous.trim().length() > 0))
        {
            pw.append("<tr>");

            if (label.length() > 0)
                pw.append("<td wrap='yes' colspan='" + colspan + "'>");
            else
                pw.append("<td nowrap='yes' colspan='" + colspan + "'>");

            if (previous != null)
                pw.append("<font class='label'>" + previous + "</font>\n");

            pw.append("</td></tr>");
        }

        pw.append("</table>");

        pw.endComponent(this);
    } // end method writeHtml

	/**
	 * Returns the CSS style to be used for date input.
	 */
	@SuppressWarnings("deprecation")
	protected int getFileWidth(int nCols)
	{
		Font			font = (Font)getEnvironment().getFont(_file.getFont());
		Toolkit			toolkit = Toolkit.getDefaultToolkit();
		FontMetrics		fm = toolkit.getFontMetrics(font);
		int				wCharWidth = fm.charWidth('W');
		int				averageCharWidth = fm.stringWidth("abcdefghijklmnopqrstuvwxyz") / 26;
		int sizeX = 0;

		if (nCols < 10)
            // With small size, the text field is sometimes not wide enough...
			sizeX = wCharWidth * (nCols + 1);
        else
            // Note : the 1.5 factor allows to have almost the same width as AWT...
        	sizeX = (averageCharWidth * nCols * 3) / 2;

		return sizeX + 4;
	}
}
