/*
 * LyStrutsForm.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web.struts;

import java.util.*;

import leon.view.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.web.*;


/**
 * This class implements the Form object (as described in the dtd) for web viewer.
 *
 * <p><b>FORM: Form</b><br>
 * A form allows to show a set of fields. This type of component is used to allow a user of an
 * application to consult application data (non editable form) or else to input data (editable
 * form). The default arrangement of the fields inside a form is based on a grid of two columns
 * comprising as many lines as there are fields. The first column contains the labels of the fields
 * and the second one contains the input fields.</p>
 *
 * @version $Id: LyStrutsForm.java 24563 2010-03-04 08:15:19Z jlsaouli $
 * @author  Lyria
 * @see     leon.view.LyForm
 */
public class LyStrutsForm extends LyStrutsPanel
	implements LyGraphicalForm,
		LySetEnabledInterface,
		LyResetContentInterface
{
	public static final String __VERSION =
		"$Id: LyStrutsForm.java 24563 2010-03-04 08:15:19Z jlsaouli $";

	/** Workaround for IE 5.0 and tables width within forms. */
	public static boolean	   COMPOSITION_TABLE_WORKAROUND = true;

	/** Constant used to generate CSS class name for caption. */
	public static int		   DECO_CAPTION_PARAMETER = 0;

	/** Constant used to generate CSS class name for field item. */
	public static int		   DECO_ITEM_PARAMETER = 1;

	/** Constant used to generate CSS class name for form. */
	public static int		   DECO_FORM_PARAMETER = 3;

    /** Maximum length for not wrapped labels. */
    public static int          LABEL_LENGTH_WRAP = 50;

    /** Width of the border used for HTML tables (for debug). */
    public static int		   DEBUG_BORDER = 0;


	/**
	 * Gets the html horizontal alignment string corresponding to the LyConstraint alignment.
	 *
	 * @param  alignCode the LyConstraint halign code.
	 * @return the html horizontal alignment string.
	 */
	protected static String getHtmlHAlignment(LyConstraint.HAlign alignCode)
	{
		if (alignCode == LyConstraint.HAlign.LEFT)
			return "left";
		else if (alignCode == LyConstraint.HAlign.RIGHT)
			return "right";
		else if (alignCode == LyConstraint.HAlign.CENTER)
			return "center";
		else
			return null;
	}

	/**
	 * Gets the html vertical alignment string corresponding to the LyConstraint alignment.
	 *
	 * @param  alignCode the LyConstraint valign code.
	 * @return the html vertical alignment string.
	 */
	protected static String getHtmlVAlignment(LyConstraint.VAlign alignCode)
	{
		if (alignCode == LyConstraint.VAlign.TOP)
			return "top";
		else if (alignCode == LyConstraint.VAlign.MIDDLE)
			return "middle";
		else if (alignCode == LyConstraint.VAlign.BOTTOM)
			return "bottom";
		else
			return null;
	}

	/** The form generic component. */
	protected LyForm _form;

	/** Form name. */
	protected String _name;

	/**
	 * Current Attachment handler. Default is null, and possible value is {@link
	 * leon.view.web.LyFileAttachmentHandler} which uses enctype="multipart/form-data" for file
	 * submit in forms.
	 */
	protected String _attachmentHandler;

	/** Current Attachment handler dirrectory. */
	protected String _attachmentDirectory;

	/**
	 * Class constructor of a Form object for the struts target.
	 *
	 * @param form the generic Form defined for every type of view targets
	 */
	public LyStrutsForm(LyForm form)
	{
		super(form);
		_form = form;

		initialize();
	}

	/**
	 * Initialization method.<br>
	 * Sets the form name, attachment handler and directory to <code>null</code>.<br>
	 * Sets the label font and colors with values defined in resources files.
	 */
	public void initialize()
	{
		_name = null;
		_attachmentHandler = null;
		_attachmentDirectory = null;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
	 */
	public void writeHtml(LyWebPageWriter pw)
	{
	    pw.startComponent(this);

		updateRefreshInfo(pw);

		String activeFormId = pw.getActiveFormId();

		if ((activeFormId != null) && (activeFormId.length() != 0))
			_name = activeFormId;

		String name = getName();
		String bgColor = getHexColor(_form.getColor());

		if (_form.getParent().getParent() instanceof LyTab)
			translate(((LyTab)_form.getParent().getParent()).getLabel());

		// If form only contains hidden fields, do not generate form but call
		// writeHtml() method
		// in order to update refresh info of the html fields
		LyConstraint formConstraint = _form.getConstraint();
		int			 i, nFields = _form.getFormFieldCount();
		LyFormField item;
		boolean     visible = false;

		for (i = 0; i < nFields; i++)
		{
			item = _form.getFormField(i);

			if (!item.isHidden())
			{
				visible = true;
				break;
			}
		}

		if ((formConstraint != null)
				&& ((formConstraint.getNCols() > 0)
						|| (formConstraint.getNRows() > 0)))
			visible = true;

		if (!visible)
		{
			LyStrutsFormField internal;

			for (i = 0; i < nFields; i++)
			{
				item = _form.getFormField(i);
				internal = (LyStrutsFormField)item.getInternalView();
				internal.writeHtml(pw);
			}

			pw.endComponent(this);
			return;
		}

		LyWebDecoration deco = getWebViewManager().getDecoration();

		if (LyWebConstants.TOOL_POPUP.equals(pw.getActiveName()))
			pw.append("<div>\n");
		else
			deco.writePageDecoration(pw, LyWebDecoration.OPEN_ZONE_1);

		LyConstraint	constraint = _form.getConstraint();
		boolean			generateExternalTable = generateExternalTable(pw);
		String			width = "width='100%' ";
		String			height = "";

		if (constraint != null)
		{
			LyConstraint.Fill	fill = constraint.getFill();

			if (constraint.getNCols() > 0)
				width = "width='" + constraint.getNCols() + "px' ";
			else if ((fill == LyConstraint.Fill.NONE) || (fill == LyConstraint.Fill.VERTICAL))
				width = "";

			if (constraint.getNRows() > 0)
				height = "height='" + constraint.getNRows() + "px' ";
		}

		pw.append("<!-- skin form fields begin -->\n");

		if (generateExternalTable)
		{
			String	className = deco.getClassName(_form, DECO_FORM_PARAMETER, pw);

			pw.append("<table class='" + className + "' border='0' ");
			pw.append(width);
			pw.append(height);

			String	align = "left";

				if ((_form.getAlign() == LyForm.Align.CENTER)
						|| ((formConstraint != null) && (formConstraint.getHAlign() == LyConstraint.HAlign.CENTER)))
					align = "center";
				else if ((_form.getAlign() == LyForm.Align.RIGHT)
						|| ((formConstraint != null) && (formConstraint.getHAlign() == LyConstraint.HAlign.RIGHT)))
					align = "right";

				if (bgColor != null)
					pw.append("style='background-color:" + bgColor + ";' ");

			pw.append("cellpadding='0' cellspacing='0' valign='top' float='" + align + "' align='" + align + "'>\n<tr>\n<td valign='top'>\n");
		}

		int	cellspacing = getCellSpacing();
		int	cellpadding = getCellPadding();

		if (LyWebConstants.TOOL_POPUP.equals(pw.getActiveName()))
		{
			cellspacing = 0;
			cellpadding = 0;
		}

		pw.append("<table id='form_table_" + _form.getIdCode() + "' ");

		if (!generateExternalTable)
		{
			pw.append(width);
			pw.append(height);
		}
		else
		{
			pw.append("width='100%' ");
			pw.append("height='100%' ");
		}

		if ((bgColor != null) && !generateExternalTable)
			pw.append(" style='background-color:" + bgColor + ";'");

		pw.append("border='" + DEBUG_BORDER + "' cellspacing='" + cellspacing + "' cellpadding='" + cellpadding + "' valign='top'>\n");

		if ((nFields == 0) && visible)
			pw.append("<tr><td>");

		// FFT00146 : use one single form for a same frame
		pw.setFormId(name);

		writeFields(pw);

		if ((nFields == 0) && visible)
			pw.append("</td></tr>");

		// Manage file uploads
		pw.setAttachmentHandler(_attachmentHandler);

		if (getAttachmentDirectory() != null)
			pw.setAttachmentDirectory(getAttachmentDirectory());

		StringBuffer	code = new StringBuffer();

		// Generate an hidden frame used to redirect the upload responses of file uploads.
		code.append("<tr><td><iframe frameborder='0' scrolling='no'");
		code.append(" name='file_frame_" + _form.getIdCode() + "'");
		code.append(" src='");
		code.append(getWebViewManager().getScriptsDir() + "/blank.html");
		code.append("' style='width:2px;height:2px;display:none;'></iframe></td></tr>");

		pw.append(code.toString());

		pw.append("\n");
		pw.append("</table>");

		if (generateExternalTable)
			pw.append("</td>\n</tr></table>\n");

		pw.append("<!-- skin form field end -->\n");

		deco.writePageDecoration(pw, LyWebDecoration.CLOSE_ZONE);

		String	popupMenuView = getPopupMenuView();

		pw.append(popupMenuView);

		pw.endComponent(this);
	} // end method writeHtml

	/**
	 * Indicates whether an HTML table must be generated around form table.
	 *
	 * @param pw the page writer
	 * @return true if the table must be generated, false otherwise.
	 */
	protected boolean generateExternalTable(LyWebPageWriter pw)
	{
		return !LyWebConstants.TOOL_POPUP.equals(pw.getActiveName());
	}

	/**
	 * Gets the HTML name of the form.
	 *
	 * @return the form name
	 */
	public String getName()
	{
		// FFT00146 : use one single form for a same frame
		if (_name == null)
		{
			// Check if the form is included within a parent form (COMPOSITION
			// RELATION)
			// TABS >> TAB >> AREA >> FORM
			if ((_form.getParent() != null) && (_form.getParent().getParent() != null)
					&& (_form.getParent().getParent().getParent() instanceof LyTabs))
			{
				LyTabs myTabs = (LyTabs)_form.getParent().getParent().getParent();

				// Tabs within tabs in a FORM
				if (myTabs.getParentFormFieldId() != null)
				{
					LyComponent parent = myTabs.getParent().getParent();

					_name = parent.getRootController();
				}
			}

			if (_name == null)
				_name = _form.getRootController();
		}

		return _name;
	} // end method getName

	/**
	 * Gets the form generic component controller id.
	 *
	 * @return the controller id
	 */
	public String getControllerId()
	{
		return _form.getRootController();
	}

	/**
	 * Writes HTML description corresponding to the form fields.
	 *
	 * @param pw the page writer
	 */
	protected void writeFields(LyWebPageWriter pw)
	{
		int				 i, nFields = _form.getFormFieldCount();
		LyFormField		 item;
		LyCaption		 caption;
		LyStrutsFormField internal;
		LyConstraint     constraint;
		int				 rowspan, colspan, currentCol = 0;
		boolean			 jump = false, captionJump, remainder;
		boolean			 openedTR = false;
        LyConstraint.HAlign hAlignCode;
        LyConstraint.VAlign vAlignCode;
        int              colWeight;
		String			 hAlign = null;
		String			 vAlign = null;
		int				 maxNbCols = 2;
		int				 maxUnitNbCols = 1;
		int				 hightNbCols = 2;
		int				 tempNbCols = 0;
		int				 tempHightNbCols = 0;
		LyWebDecoration	 deco = getWebViewManager().getDecoration();
		boolean			 hasColsOffset = false;
		String			 oldColsString = "", colsString = "";
		Vector<String>	 endScripts = new Vector<String>();

		String	width = "";
		String	height = "";

		boolean hasFullWidthArea = false;
		BitSet	areaColspans = new BitSet();

		// compute the nomber ncols of the hight object (hightNbCols)
		// and compute the number ncols of hight line (maxNbCols)
		for (i = 0; i < nFields; i++)
		{
			item = _form.getFormField(i);

			caption = item.getCaption();
			if (caption != null)
			{
				constraint = caption.getConstraint();
				if (constraint != null)
				{
					colspan = constraint.getColSpan();
					remainder = constraint.isRemainder();

					if (colspan < 0)
						colspan = 1;

					if (remainder)
						jump = true;
				}
				else
					colspan = 1;

				colsString += colspan;

				if (jump)
				{
					tempNbCols = 0;
					tempHightNbCols = 0;
				}
				else
				{
					tempNbCols += colspan;
					tempHightNbCols += colspan;
				}
			}

			if (jump)
			{
				if (oldColsString.length() > 0 && !oldColsString.equals(colsString))
					hasColsOffset = true;

				if (tempNbCols > maxNbCols)
					maxNbCols = tempNbCols;

				tempNbCols = 0;
				jump = false;
				oldColsString = colsString;
				colsString = "";
			}

			constraint = item.getConstraint();

			if (constraint != null)
			{
				colspan = constraint.getColSpan();
				remainder = constraint.isRemainder();

				if (remainder || (colspan == LyConstraint.DEFAULT_VALUE))
					jump = true;

				if (colspan < 0)
					colspan = 1;
			}
			else
			{
				colspan = 1;
				jump = true;
			}
			tempNbCols += colspan;
			tempHightNbCols += colspan;
			colsString += colspan;

			if (tempHightNbCols > hightNbCols)
				hightNbCols = tempHightNbCols;

			if (colspan > maxUnitNbCols)
				maxUnitNbCols = colspan;

			tempHightNbCols = 0;

			if (jump)
			{
				if (oldColsString.length() > 0 && !oldColsString.equals(colsString))
					hasColsOffset = true;

				if (tempNbCols > maxNbCols)
					maxNbCols = tempNbCols;

				tempNbCols = 0;
				jump = false;
				oldColsString = colsString;
				colsString = "";
			}

			if (item.getFieldItem() instanceof LyArea)
				areaColspans.set(colspan);
		}

		if (areaColspans.get(hightNbCols))
			hasFullWidthArea = true;

		jump = true;

		boolean generateCellForHiddenFields = generateCellForHiddenFields();

		for (i = 0; i < nFields; i++)
		{
			hAlign = "right";
			vAlign = "top";
			item = _form.getFormField(i);
			internal = (LyStrutsFormField)item.getInternalView();

			if (internal == null)
				continue;

			int		cellWidth = getCellWidth(item);
			int		cellHeight = getCellHeight(item);
			int		cellLeftPadding = getCellLeftPadding(item);
			int		cellRightPadding = getCellRightPadding(item);
			int		cellTopPadding = getCellTopPadding(item);
			int		cellBottomPadding = getCellBottomPadding(item);

			width  = "";
			height = "";

			if (cellWidth > 0)
				width = "width='" + cellWidth + "px' ";

			if (cellHeight > 0)
				height = "height='" + cellHeight + "px' ";

			boolean	jumpAfterItem = true;

			if (jump)
			{
				constraint = item.getConstraint();

				if (constraint != null)
				{
					colspan = constraint.getColSpan();
					remainder = constraint.isRemainder();

					if (!remainder && (colspan != LyConstraint.DEFAULT_VALUE))
						jumpAfterItem = false;
				}
			}

			if (item.isHidden())
			{
				// Call writeHtml() method in order to update refresh info of
				// the html field
				internal.writeHtml(pw);

				boolean	onlyFieldOnTR = (jump && jumpAfterItem);

				if (!generateCellForHiddenFields && !onlyFieldOnTR)
					continue;
			}

			if (jump)
			{
				// If caption and item are hidden and displayed on a single line without any other field,
				// do not display line (line will be displayed when setHidden(false) will be called on field.
				String	parentTRId = internal.getIdCode() + "_parentTR";

				pw.append("<tr id='" + parentTRId + "'");

				if (jumpAfterItem)
				{
					internal.setParentTRId(parentTRId);

					if (item.isHidden())
						pw.append(" style='display:none;'");
				}

				pw.append(">");
				openedTR = true;
				currentCol = 0;
			}

			caption = item.getCaption();
			captionJump = false;

			if (caption != null)
			{
				pw.append("\n");
				jump = false;
				constraint = caption.getConstraint();

				if (constraint != null)
				{
					rowspan = constraint.getRowSpan();
					colspan = constraint.getColSpan();
					remainder = constraint.isRemainder();
					colWeight = (int)constraint.getColWeight();

					if (remainder)
						jump = true;

					if (colspan < 0)
						colspan = 1;

					if (rowspan == 0)
						rowspan = 1;
					else if (rowspan < 0)
						rowspan = 1;

					hAlignCode = constraint.getHAlign();

					if (getHtmlHAlignment(hAlignCode) != null)
						hAlign = getHtmlHAlignment(hAlignCode);

					if (pw.isPrintMode())
						hAlign = getHtmlHAlignment(LyConstraint.HAlign.LEFT);

					vAlignCode = constraint.getVAlign();
					if (getHtmlVAlignment(vAlignCode) != null)
						vAlign = getHtmlVAlignment(vAlignCode);
				}
				else
				{
					rowspan = 1;
					colspan = 1;
					colWeight = 0;
				}

				int diffColspan = colspan;

				currentCol += colspan;
				if (jump)
				{
					// workaround for tables in forms : the table does not take
					// the entire line if alone
					if (COMPOSITION_TABLE_WORKAROUND)
						// colspan++;
						captionJump = true; // workaround
					if ((maxNbCols - currentCol) > 0)
						diffColspan += (maxNbCols - currentCol);
				}
				if (vAlign.equals("center"))
					vAlign = "middle"; // Consistency with HTML DTD

				pw.append("<td colspan='" + diffColspan + "' rowspan='" + rowspan + "'");
				pw.append(" align='" + hAlign + "'");
				pw.append(" valign='" + vAlign + "'");

				if (colWeight <= 0)
				{
					if (_form.getAlign() != LyForm.Align.LEFT)
						colWeight = 30; // %
					else
						colWeight = 0; // %
				}

				pw.append(" nowrap='true' width='" + colWeight + "%' class='" +
					deco.getClassName(item, DECO_CAPTION_PARAMETER, pw) + "'>");

				internal.writeCaption(pw);
				pw.append("</td>");

				if (jump)
				{
					String	parentCaptionTRId = internal.getIdCode() + "_parentCaptionTR";

					pw.append("</tr><tr id='" + parentCaptionTRId + "'");

					if (jumpAfterItem && item.isHidden())
					{
						internal.setParentCaptionTRId(parentCaptionTRId);
						pw.append(" style='display:none;'");
					}

					pw.append(">");
				}
			}

			if (jump)
			{
				jump = false;
				currentCol = 0;
			}

			hAlign = null;
			vAlign = null;
			colWeight = 0;

			constraint = item.getConstraint();

			if (constraint != null)
			{
				rowspan = constraint.getRowSpan();
				colspan = constraint.getColSpan();
				remainder = constraint.isRemainder();
				colWeight = (int)constraint.getColWeight();

				if (remainder || (colspan == LyConstraint.DEFAULT_VALUE))
					jump = true;

				if (colspan == LyConstraint.DEFAULT_VALUE)
					colspan = maxNbCols - currentCol;

				if (colspan <= 0)
					colspan = 1;

				if (rowspan == 0)
					rowspan = 1;
				else if (rowspan < 0)
					rowspan = 1;

				if (colWeight == -2)
					colWeight = 100;

				if (colWeight < 0)
					colWeight = 0;

				hAlignCode = constraint.getHAlign();
				hAlign = getHtmlHAlignment(hAlignCode);
				vAlignCode = constraint.getVAlign();
				vAlign = getHtmlVAlignment(vAlignCode);
			}
			else
			{
				rowspan = 1;
				colWeight = 0;
				colspan = 1;
				jump = true;
			}

			pw.append("\n");

			currentCol += colspan;

			int diffColspan = colspan;

			if (((currentCol + 1) < hightNbCols) && (i < (nFields - 1)))
				jump = false;
			else if (i == (nFields - 1))
				jump = true;

			String	style = "";

			// Complete colspan if field is the last one on its line and only
			// if it is left-aligned. Otherwise field may not be placed correctly.
			if (jump
					&& ((hAlign == null) || "left".equals(hAlign)))
			{
				if ((maxNbCols - currentCol) > 0)
					diffColspan += (maxNbCols - currentCol);
			}

			if (COMPOSITION_TABLE_WORKAROUND && captionJump)
			{
				// workaround for tables in forms : the table does not take the
				// entire line if alone
				// diffColspan += 1;
				pw.append("<td " + width + height + "colspan='" + diffColspan + "' rowspan='" + rowspan + "'");
				captionJump = false;
			}
			else
			{
				if (hasColsOffset && cellWidth > 0)
				{
					style += "width:2px;";
					pw.append("<td " + height + "colspan='" + diffColspan + "' rowspan='" + rowspan + "'");
				}
				else
					pw.append("<td " + width + height + "colspan='" + diffColspan + "' rowspan='" + rowspan + "'");
			}

			if (hAlign != null)
				pw.append(" float='" + hAlign + "' align='" + hAlign + "'");
			if (vAlign == null)
				vAlign = "middle";
			else if (vAlign.equals("center"))
				vAlign = "middle"; // Consistency with HTML DTD

			pw.append(" valign='" + vAlign + "'");

			if (jump)
			{
				if (((maxUnitNbCols != hightNbCols) || (colWeight != 100))
						&& (cellWidth < 0))
				{
					// Giving a 100% width to a cell does not work with IE nor Chrome.
					// Compute well width by Javascript.
					if (hasColsOffset
							&& ((getWebViewManager().getBrowser() == LyWebViewManager.BROWSER_MSIE)
									|| (getWebViewManager().getBrowser() == LyWebViewManager.BROWSER_CHROME)))
					{
						pw.append(" id='form_cell_" + i + "_" + _form.getIdCode() + "'");
						endScripts.add("<script type='text/javascript'>setTimeout(\"ExtendCellToFullWidth('form_cell_" + i + "_" + _form.getIdCode() +
								"', 'form_table_" + _form.getIdCode() + "')\", 200);</script>");
					}
					else
						pw.append(" width='" + colWeight + "%'");
				}
				else
				{
					// If last field is contained in a form and area with an horizontal fill,
					// give it a 100% width.
					LyConstraint	formConstraint = _form.getConstraint();
					LyConstraint	areaConstraint = _form.getParentArea().getConstraint();

					if ((constraint != null)
							&& ((constraint.getFill() == LyConstraint.Fill.HORIZONTAL)
									|| (constraint.getFill() == LyConstraint.Fill.BOTH))
							&& ((formConstraint == null)
									|| (formConstraint.getFill() == LyConstraint.Fill.HORIZONTAL)
									|| (formConstraint.getFill() == LyConstraint.Fill.BOTH))
							&& (areaConstraint != null)
							&& ((areaConstraint.getFill() == LyConstraint.Fill.HORIZONTAL)
									|| (areaConstraint.getFill() == LyConstraint.Fill.BOTH)))
						pw.append(" width='100%'");
				}
			}
			else
			{
				if (cellWidth < 0)
					pw.append(" width='0%'");
			}

			boolean nowrap = false;

			if (item.getFieldItem() instanceof LyLabel)
            {
                String label = ((LyLabel)item.getFieldItem()).getLabel();

				nowrap = (label == null) || (label.length() < LABEL_LENGTH_WRAP);
            }
			else if ((item.getFieldItem() instanceof LyText) && !item.isEditable()
					&& (((LyText)item.getFieldItem()).getNRows() <= 1))
				nowrap = true;

			if (nowrap)
				pw.append(" nowrap='true'");

			if (cellLeftPadding >  0)
				style += "padding-left:" + cellLeftPadding + "px;";

			if (cellRightPadding >  0)
				style += "padding-right:" + cellRightPadding + "px;";

			if (cellTopPadding >  0)
				style += "padding-top:" + cellTopPadding + "px;";

			if (cellBottomPadding >  0)
				style += "padding-bottom:" + cellBottomPadding + "px;";

			if (style.length() > 0)
				pw.append(" style='" + style + "'");

			pw.append(" class='" + deco.getClassName(item, DECO_ITEM_PARAMETER, pw) + "'>");

			if (hasColsOffset && cellWidth > 0)
				pw.append("<table cellpadding='0' cellspacing='0' border='0' width='" + cellWidth + "'><tr><td>");

			boolean currentInnerView = pw.isInnerView();

			pw.setInnerView(true);
			internal.writeItem(pw);
			pw.setInnerView(currentInnerView);

			if (hasColsOffset && cellWidth > 0)
				pw.append("</td></tr></table>");

			pw.append("</td>");

			if (jump)
			{
				pw.append("</tr>");
				openedTR = false;

				if (hasColsOffset && hasFullWidthArea && (i < nFields - 1))
				{
					// If one of the field is a full width area, generate another HTML table around it
					// to avoid graphical issues with the other fields.
					LyFormField	nextItem = _form.getFormField(i + 1);

					if ((item.getFieldItem() instanceof LyArea
								&& item.getConstraint() != null
								&& item.getConstraint().getColSpan() == hightNbCols)
							|| (nextItem.getFieldItem() instanceof LyArea
									&& nextItem.getConstraint() != null
									&& nextItem.getConstraint().getColSpan() == hightNbCols))
					{
						pw.append("</table>");
						pw.append("<table ");
						pw.append("width='100%' ");
						pw.append("height='100%' ");
						pw.append("border='" + DEBUG_BORDER + "' cellspacing='" + getCellSpacing() + "' cellpadding='" + getCellPadding() + "' valign='top'>\n");
					}
				}
			}
		}

		if (openedTR)
			pw.append("</tr>");

		for (String s : endScripts)
		{
			getWebViewManager().addAdditionalRefreshScripts(s);
			pw.addEndScript(s, false);
		}
	} // end method writeFields

	/**
	 * Sets the attachment handler. Default value is null, this value is set by a File fieldItem in
	 * the form.
	 *
	 * @param handler the new handler, or null if none is specified
	 * @see   #getAttachmentHandler()
	 */
	public void setAttachmentHandler(String handler)
	{
		_attachmentHandler = handler;
	}

	/**
	 * Gets the attachment handler.
	 *
	 * @return the attachment handler
	 * @see    #setAttachmentHandler(String)
	 */
	public String getAttachmentHandler()
	{
		return _attachmentHandler;
	}

	/**
	 * Sets the attachment handler directory.
	 *
	 * @param attachmentDirectory the new attachment directory
	 * @see   #getAttachmentDirectory()
	 */
	public void setAttachmentDirectory(String attachmentDirectory)
	{
		_attachmentDirectory = attachmentDirectory;
	}

	/**
	 * Gets the attachment handler directory.
	 *
	 * @return the attachment handler directory
	 * @see    #setAttachmentDirectory(String)
	 */
	public String getAttachmentDirectory()
	{
		return _attachmentDirectory;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetEnabledInterface#setEnabled(boolean)
	 */
	public void setEnabled(boolean enabled)
	{
		int		    i, n = _form.getFormFieldCount();
		LyFormField field;

		for (i = 0; i < n; i++)
		{
			field = _form.getFormField(i);
			field.setEnabled(enabled);
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetEditableInterface#setEditable(boolean)
	 */
	public void setEditable(boolean editable)
	{
		refresh();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyResetContentInterface#resetContent()
	 */
	public void resetContent()
	{
		resetContent(false);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyResetContentInterface#resetContent(boolean)
	 */
	public void resetContent(boolean full)
	{
		if (full)
		{
			LyArea	area = _form.getParentArea();

			if ((area != null)
					&& (area.getParentArea() != null)
					&& (area.getParentArea().isBorderPanel(area)))
			{
				StringBuffer	buffer = new StringBuffer();

				buffer.append("var parentIframes = parent.document.getElementsByTagName('iframe');");
				buffer.append("var i;");
				buffer.append("for (i = 0; i < parentIframes.length; i++) {");
				buffer.append("if (parentIframes[i].name == '" + _lastFrameName + "') {");
				buffer.append("parentIframes[i].setAttribute('ly_height_scroll', '0');");
				buffer.append("parentIframes[i].scrolling = 'no';");
				buffer.append("}");
				buffer.append("}");
				buffer.append("checkSizeMainView();");

				getWebViewManager().addAdditionalRefreshScripts("<script type='text/javascript'>\n//<![CDATA[\n" + buffer + "\n//]]>\n</script>");
			}

			refresh();
		}
	}

	/**
	 * Returns the width of the HTML cell containing given field if fixed.
	 * @return the width of the cell or a negative value if there is no fixed width.
	 */
	public int getCellWidth(LyFormField formField)
	{
		return -1;
	}

	/**
	 * Returns the height of the HTML cell containing given field if fixed.
	 * @return the height of the cell or a negative value if there is no fixed height.
	 */
	public int getCellHeight(LyFormField formField)
	{
		return -1;
	}

	/**
	 * Returns the left padding of the HTML cell containing given field.
	 * @return the left padding of the cell or a negative value if there is no left padding.
	 */
	public int getCellLeftPadding(LyFormField formField)
	{
		LyConstraint	constraint = formField.getConstraint();

		if (constraint != null)
		{
			if (constraint.getLeftPadding() > 0)
				return constraint.getLeftPadding();

			return constraint.getCellPadding();
		}

		return 0;
	}

	/**
	 * Returns the right padding of the HTML cell containing given field.
	 * @return the right padding of the cell or a negative value if there is no right padding.
	 */
	public int getCellRightPadding(LyFormField formField)
	{
		LyConstraint	constraint = formField.getConstraint();

		if (constraint != null)
		{
			if (constraint.getRightPadding() > 0)
				return constraint.getRightPadding();

			return constraint.getCellPadding();
		}

		return 0;
	}

	/**
	 * Returns the top padding of the HTML cell containing given field.
	 * @return the top padding of the cell or a negative value if there is no top padding.
	 */
	public int getCellTopPadding(LyFormField formField)
	{
		LyConstraint	constraint = formField.getConstraint();

		if (constraint != null)
		{
			if (constraint.getTopPadding() > 0)
				return constraint.getTopPadding();

			return constraint.getCellPadding();
		}

		return 0;
	}

	/**
	 * Returns the bottom padding of the HTML cell containing given field.
	 * @return the bottom padding of the cell or a negative value if there is no bottom padding.
	 */
	public int getCellBottomPadding(LyFormField formField)
	{
		LyConstraint	constraint = formField.getConstraint();

		if (constraint != null)
		{
			if (constraint.getBottomPadding() > 0)
				return constraint.getBottomPadding();

			return constraint.getCellPadding();
		}

		return 0;
	}

	/**
	 * Returns the cell spacing that must be used for the internal HTML table generated for form.
	 *
	 * @return the cell spacing in pixels.
	 */
	public int getCellSpacing()
	{
		return 1;
	}

	/**
	 * Returns the cell padding that must be used for the internal HTML table generated for form.
	 *
	 * @return the cell padding in pixels.
	 */
	public int getCellPadding()
	{
		return 1;
	}

	/**
	 * Indicates whether an HTML cell must be used for hidden fields.
	 * @return true if an HTML cell must be used, false otherwise.
	 */
	public boolean generateCellForHiddenFields()
	{
		return false;
	}
} // end class LyStrutsForm
