/*
 * LyStrutsFormField.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web.struts;

import leon.misc.*;
import leon.view.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.web.*;


/**
 * This class implements the Field object (as described in the dtd) for Struts viewer.
 *
 * @version $Id: LyStrutsFormField.java 36163 2012-07-03 16:41:33Z nhamon $
 * @author  Lyria
 * @see     leon.view.LyFormField
 */
public class LyStrutsFormField extends LyWebObject
    implements LyGraphicalFormField,
               LyResetContentInterface,
               LySetFocusInterface
{
	public static final String __VERSION =
		"$Id: LyStrutsFormField.java 36163 2012-07-03 16:41:33Z nhamon $";

    /** Maximum length for not wrapped captions. */
    public static int          CAPTION_LENGTH_WRAP = 50;
    
	/** The toolbar generic component of the form field generic component. */
    protected LyToolBar        _toolbar;

    /** The parent form id. */
    protected String           _formId;

    /** The form field generic component. */
    protected LyFormField      _field;
    
    /** Identifier of the parent TR containing the caption. */
    protected String           _parentCaptionTRId;
    
    /** Identifier of the parent TR containing the field. */
    protected String           _parentTRId;
    
    
	/**
	 * Class constructor of a Form Field object for the Struts target.
	 *
	 * @param field the generic Form Field defined for every type of view targets
	 */
	public LyStrutsFormField(LyFormField field)
	{
        super(field);
        
        _field = field;
        _parentCaptionTRId = null;
        _parentTRId = null;
        initialize();
	}
	
	/**
	 * Returns the url sent to Javascript function for popup menus.
	 * 
	 * @return the graphical action URL.
	 */
	protected String getPopupMenuUrl()
	{
		// get the url to the event handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		
		return actionUrl;
	}
	
	   /**
     * Initialization method.<br>
     * Sets the toolbar generic component and the form Id to <code>null</code>.<br>
     * If the generic form field component is not enable, sets the enable attribute of this
     * component to false.
     */
    public void initialize()
    {
        if (!_field.isEnabled())
            setEnabled(false);

        _toolbar = null;
        _formId = null;
    }

    /**
     * Gets the name of the CSS style to be used for this field.
     *
     * @return the name of the CSS style to use
     */
    protected String getCssClass()
    {
        LyWebDecoration deco = getWebViewManager().getDecoration();

        return deco.getClassName(_field.getCaption(), null);
    }

    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebObject#refresh()
     */
    public void refresh()
    {
        boolean refresh = true;
        LyTabs  tabs = _field.getParentTabs();

        if (tabs != null)
        {
            LyStrutsTabs webTabs = (LyStrutsTabs)tabs.getInternalView();
            LyTab        selectedTab = webTabs.getSelectedTab();
            LyTab        fieldTab = _field.getParentTab();

            // Refresh only if field is displayed
            if ((selectedTab != null) && (fieldTab != null) && (fieldTab != selectedTab))
                refresh = false;
        }

        if (refresh)
            super.refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
     */
    public void writeHtml(LyWebPageWriter pw)
    {       
        updateRefreshInfo(pw);

        // if the field is hidden : do not write
        if (_field.isHidden())
            return;

        pw.startComponent(this);
          
        // When a form field is not displayed in a form, an HTML table is generated
        LyCaption	caption = _field.getCaption();
        boolean     generateTable = !(_field.getParent() instanceof LyForm) && (caption != null);
        
        if (generateTable)
            pw.append("<table border='0' cellpadding='0' cellspacing='0'><tr><td nowrap='yes'>");
        
        writeCaption(pw);
        
        if (generateTable)
            pw.append("</td><td nowrap='yes'>");
        
        writeItem(pw);
        
        if (generateTable)
            pw.append("</td></tr></table>");
        
        pw.endComponent(this);
    }

    /*
     * (non-Javadoc)
     * 
     * @see leon.view.graphic.LyGraphicalFormField#setChecked(boolean)
     */
    public void setChecked(boolean checked)
    {
        if (_field.isCheckBox())
        {
            LyWebViewManager viewManager = getWebViewManager();
            StringBuffer     buffer = new StringBuffer();

            buffer.append("<script type='text/javascript'>\n//<![CDATA[\nSelectCheckbox('");
            buffer.append(_formId);
            buffer.append("','_checkbox_");
            buffer.append(_field.getId());
            buffer.append("','',");
            buffer.append(checked);
            buffer.append(");\n//]]>\n</script>");
    
            viewManager.addAdditionalRefreshScripts(buffer.toString());
        }
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalFormField#setHidden(boolean)
     */
    public void setHidden(boolean hidden)
    {
        if (_parentTRId != null)
        {
            LyWebViewManager viewManager = getWebViewManager();
            StringBuffer     buffer = new StringBuffer();

            buffer.append("<script type='text/javascript'>\n//<![CDATA[\ndisplayElement('");
            buffer.append(_lastFrameName);
            buffer.append("','");
            buffer.append(_parentTRId);
            buffer.append("',");
            buffer.append(!hidden);
            buffer.append(");\n//]]>\n</script>");
            
            if (_parentCaptionTRId != null)
            {
                buffer.append("<script type='text/javascript'>\n//<![CDATA[\ndisplayElement('");
                buffer.append(_lastFrameName);
                buffer.append("','");
                buffer.append(_parentCaptionTRId);
                buffer.append("',");
                buffer.append(!hidden);
                buffer.append(");\n//]]>\n</script>");
            }
    
            viewManager.addAdditionalRefreshScripts(buffer.toString());
        }
        else
            refresh();
    }

    /**
     * Writes HTML description corresponding to the Caption label of the field.
     *
     * @param pw the page writer
     * @see   #writeCaption(LyWebPageWriter, boolean) with parameters (pw, true)
     */
    protected void writeCaption(LyWebPageWriter pw)
    {
        writeCaption(pw, !pw.isPrintMode());
    }

    /**
     * Writes HTML description corresponding to the Caption label of the field.
     *
     * @param pw       the page writer
     * @param writeDiv indicates if an HTML division must include the HTML description of the label
     */
    protected void writeCaption(LyWebPageWriter pw, boolean writeDiv)
    {
        if (writeDiv)
            pw.append("<div id=\"" + getIdCode() + "_caption\">");

        updateRefreshInfo(pw);

        // if the fiel is hidden : do not write
        if (_field.isHidden() && (_parentTRId == null))
            return;

        pw.append("<table><tr>");

        if (!pw.isPrintMode() && _field.isCheckBox())
        {
            LyForm      form = (LyForm)_field.getParent();
            LyStrutsForm webForm = (LyStrutsForm)form.getInternalView();

            pw.append("<td><input type='hidden' value='" + _field.isChecked() + "' name=\"_checkbox_state_");
            pw.append(_field.getId());
            pw.append("\" /><input type='checkbox' value='' name=\"_checkbox_");
            pw.append(_field.getId());
            pw.append("\" onClick=\"javascript:window.document.forms[0].elements['_checkbox_state_" + _field.getId()
                + "'].value=this.checked;FieldChange(window.document." + webForm.getName()
                + ",this,'" + _field.getRootController() + "')\"");

            if (_field.isChecked())
                pw.append(" checked='true'");
            
            pw.append("/></td>");
        }

        LyTip         tip = _field.getTip();
        LyEnvironment env = _field.getEnvironment();
        boolean       displayTipIcon = env.testResource("LY_DISPLAY_TIP_ICON");

        if (!pw.isPrintMode() && (tip != null) && displayTipIcon)
        {
            LyStrutsTip webTip = (LyStrutsTip)tip.getInternalView();

            pw.append("<td>");
            webTip.writeHtml(pw);
            pw.append("</td>");
        }

        LyCaption caption = _field.getCaption();

        if (caption != null)
        {
            LyStrutsCaption internal = (LyStrutsCaption)caption.getInternalView();
            String         cssClass = getCssClass();
            boolean        nowrap = false;

            pw.append("<td id='" + _field.getId() + "_td' class='" + cssClass + "'");
			
			String  style = "";
      if(!caption.isDefaultFont() || !caption.isDefaultFontColor()) {
        style += getFontString(caption.getFont(),caption.getFontColor());
      }
      if(!caption.isDefaultColor()) {
        style += " background-color:"+getHexColor(caption.getColor())+";";
        style += " background-image:none;";
      }
      
      if (style.trim().length() == 0) {
        if (_field.getParent() instanceof LyForm) {
          LyForm form = (LyForm) _field.getParent();
          if (!form.isDefaultFont() || !form.isDefaultFontColor()) {
            style += getFontString(form.getFont(), form.getFontColor());
          }
          if (!form.isDefaultColor()) {
            style += " background-color:" + getHexColor(form.getColor()) + ";";
            style += " background-image:none;";
          }
        }
      }
           
      if(style.trim().length() > 0) {
	        pw.append(" style=\'"+style+"\'");
	    } 
	      
			if (!pw.isPrintMode() && (tip != null) && !displayTipIcon)
            {
                String label = tip.getLabel();

                // translate
                label = translate(label);
                label = getHtmlEntitiesString(label);

                pw.append(" onMouseOver=\"this.style.cursor='help'\" title='" + label + "'");
            }
            
            if (caption.isEnabled()
                && !pw.isPrintMode())
            {
                pw.append(" onContextMenu=\"javascript:InvokePopup('_caption_" + _field.getId() + "','");
                pw.append("-1','" + getPopupMenuUrl() + "','" + _field.getParent().getRootController() + "',event.ctrlKey,event.clientX,event.clientY);");
                pw.append("return false;\"");
            }
            
            String captionLabel = caption.getLabel();
            
            captionLabel = translate(captionLabel);
            nowrap = (captionLabel == null) || (captionLabel.length() < CAPTION_LENGTH_WRAP);
            
            if (nowrap)
                pw.append(" noWrap='true'");

            pw.append(">");

            if (_field.getState() > LyFormField.STATE_WARNING)
            {
                String hexColor = getHexColor("LY_LABEL_STATE_" + _field.getState());

                pw.append("<font color=\"" + hexColor + "\">");
                internal.writeHtml(pw);
                pw.append("</font>");
            }
            else
                internal.writeHtml(pw);
            
            pw.append("</td>");
        }

        pw.append("</tr></table>");

        if (writeDiv)
            pw.append("</div>");
    } // end method writeCaption

    /**
     * Writes HTML description corresponding to the form field item.
     *
     * @param pw the page writer
     * @see   #writeItem(LyWebPageWriter, boolean) with parameters (pw, true)
     */
    protected void writeItem(LyWebPageWriter pw)
    {
        writeItem(pw, !pw.isPrintMode());
    }

    /**
     * Writes HTML description corresponding to the form field item.
     *
     * @param pw       the page writer
     * @param writeDiv indicates if a division must include the item HTML description
     */
    protected void writeItem(LyWebPageWriter pw, boolean writeDiv)
    {
        updateRefreshInfo(pw);

        // if the fiel is hidden : do not write
        if (_field.isHidden() && (_parentTRId == null))
            return;

        _toolbar = _field.getToolBar();

        LyComponent    item = _field.getFieldItem();
        LyFieldUnit    unit = _field.getFieldUnit();
        LyWebComponent internal;
        boolean        writeTable = ((_toolbar != null) || (unit != null));

        if (writeDiv)
            pw.append("<div id=\"" + getIdCode() + "_item\">");

        if (writeTable)
            pw.append("<table border='0' cellpadding='0' cellspacing='0'>\n<tr height='100%'>\n");

        if (item != null)
        {
            if (writeTable)
                pw.append("<td>");

            internal = (LyWebComponent)item.getInternalView();

            // There must be no frame within an HTML form
            if (internal instanceof LyWebPageWriter)
            {
                // Structis / FFT00048
                if ((internal instanceof LyStrutsArea)
                        && !(internal instanceof LyStrutsArea)
                        && !pw.isPrintMode())
                    // DHTML : use an iframe with everything in it
                    ((LyStrutsArea)internal).writeGlobalFrameSet(pw, false, false);
                else
                {
                    ((LyWebPageWriter)internal).setNoFrames(true);
                    
                    if (!(internal instanceof LyStrutsArea))
                        ((LyWebPageWriter)internal).preparePage(null, null);
                    
                    internal.writeHtml(pw);
                }
            }
            else if ((internal instanceof LyStrutsTable)
                    && (((LyTable)item).getSelectionMode() != LyTable.SelectionMode.NONE)
                    && !(getWebViewManager() instanceof LyStrutsViewManager))
            {
                // create an iframe as a container for this table
                String elementId = LyStrutsArea.ELEMENT + ((LyGetIdInterface)item).getId();

                // Be sure the page is prepared for a callback
                String activeName = pw.getActiveName();

                pw.clearBuffer(elementId); // Clear any previous info
                pw.setActiveBuffer(elementId); // Activate buffer for callback
                internal.writeHtml(pw);
                pw.setActiveBuffer(activeName); // Restore current writing
                // context

                // Write an iframe for callback
                String maxHeight = " ly_max_height='" + LyStrutsArea.FRAME_MAX_HEIGHT + "'";
                String maxWidth = " ly_max_width='" + LyStrutsArea.FRAME_MAX_WIDTH + "'";

                LyStrutsArea.writeFrame(pw.getId(), pw.getWindowName(), pw, getRootUrl(), elementId,
                    "", false, false, false, maxHeight, maxWidth);
            }
            else
                internal.writeHtml(pw);

            if (writeTable)
                pw.append("</td>\n");
        }

        if (unit != null)
        {
            pw.append("<td nowrap='yes' valign='middle'>&nbsp;");

            LyFieldItem unitItem = unit.getFieldItem();

            internal = (LyWebComponent)unitItem.getInternalView();
            internal.writeHtml(pw);

            pw.append("</td>\n");
        }

        if (_toolbar != null)
        {
            pw.append("<td nowrap='yes'>");
            internal = (LyStrutsToolBar)_toolbar.getInternalView();
            internal.writeHtml(pw);
            pw.append("</td>\n");
        }

        if (writeTable)
            pw.append("</tr>\n</table>\n");

        if (writeDiv)
            pw.append("</div>");
    } // end method writeItem

    /**
     * Sets the identifier of the parent TR containing the field.
     */
    public void setParentTRId(String parentTRId)
    {
        _parentTRId = parentTRId;
    }

    /**
     * Sets the identifier of the parent TR containing the caption.
     */
    public void setParentCaptionTRId(String parentCaptionTRId)
    {
        _parentCaptionTRId = parentCaptionTRId;
    }
    
    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalFormField#setToolBar(leon.view.LyToolBar)
     */
    public void setToolBar(LyToolBar toolBar)
    {
        // Compare the old toolbar and the new one and refresh only if needed
        boolean refresh = compareToolbarComponents(_toolbar, toolBar);

        if (refresh)
            refreshDivContent();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetOptionalInterface#setOptional(boolean)
     */
    public void setOptional(boolean optional)
    {
        if ((_field.getEnvironment() != null) 
                && _field.getEnvironment().testEnv("LY_REPLAY"))
        {
            // Need to refresh the whole view (a call to refresh() only does not
            // work)
            _lastFrameName = LyWebConstants.TARGET_FRAME;
            refresh();
            return;
        }

        // If item is a date, perform a full refresh to update the generated UpdateDateOnBlur() call
        if (_field.getFieldItem() instanceof LyDate)
        	refreshDivContent();
        
        String	asterisk = getString("LY_CAPTION_ASTERISK", "");
        
        if ((asterisk != null) && (asterisk.trim().length() > 0))
        {
        	refreshCaption();
        	return;
        }
        
        LyWebViewManager viewManager = getWebViewManager();
        StringBuffer     buffer = new StringBuffer();
        String           cssClass = getCssClass();

        buffer.append("<script type='text/javascript'>\n//<![CDATA[\nChangeObjectClass('").append(
            _field.getId()).append("_td','").append(cssClass).append("');\n//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(buffer.toString());
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LyResetContentInterface#resetContent()
     */
    public void resetContent()
    {
        resetContent(false);
    }
    
    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LyResetContentInterface#resetContent(boolean)
     */
    public void resetContent(boolean full)
    {
        if (full)
            refresh();
        else
            refreshDivContent();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetEditableInterface#setEditable(boolean)
     */
    public void setEditable(boolean editable)
    {
        refreshDivContent();
        
        if (_field.getFieldItem() instanceof LyText)
        {
        	LyText	text = (LyText)_field.getFieldItem();

        	if (!text.isNotifyValueChanged() && (text.getType() != LyText.Type.RICH))
			{
				// Text label may not be computed correctly when updated by Ajax : calling the setLabel() method
        		// to make sure that label will be correctly displayed.
	        	LyStrutsText	strutsText = (LyStrutsText)text.getInternalView();
	        	
	        	if (strutsText != null)
	        		strutsText.setLabel(text.getLabel());
			}
        }
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetEnabledInterface#setEnabled(boolean)
     */
    public void setEnabled(boolean enabled)
    {
        refreshDivContent();
    }
    
    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetFocusInterface#setFocus()
     */
    public void setFocus()
    {
        LyWebViewManager    viewManager = getWebViewManager();
        StringBuffer        buffer = new StringBuffer();

        // Focus may be given only to texts, numbers, dates and simple relations
        LyFieldItem     item = _field.getFieldItem();
        String          fieldId = _field.getId();
        
        if (item instanceof LyDate)
        {
            LyDate.Type   type = ((LyDate)item).getType();
            boolean       addDate = (type == LyDate.Type.DATE_TIME) || (type == LyDate.Type.DATE_HOUR_MIN)
                || (type == LyDate.Type.DATE);
            
            if (addDate)
                fieldId += "_date";
            else
                fieldId += "_time";
        }
        
        buffer.append("<script type='text/javascript'>\n//<![CDATA[\nsetTimeout(\"ForceFocus('" + _formId + "', '" +
                fieldId + "', false)\", 180);\n//]]>\n</script>");
        
        viewManager.addAdditionalRefreshScripts(buffer.toString());
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetStateInterface#setState(short)
     */
    public void setState(short state)
    {
        refreshCaption();
    } // end method setState

    /**
     * Updates attribute message to the given value.
     * @param message New value of message.
     */
    public void setMessage(String message)
    {
    }
    
    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebObject#updateRefreshInfo(leon.view.web.LyWebPageWriter,
     *      boolean)
     */
    protected void updateRefreshInfo(LyWebPageWriter pw, boolean keepId)
    {
        super.updateRefreshInfo(pw, keepId);

        if (!pw.writeDivOnly() || (_formId == null))
            _formId = pw.getActiveFormId();
    }

    /**
     * Refreshes the content of the div containing the item description.
     */
    public void refreshDivContent()
    {
        if ((_field.getEnvironment() != null)
                && _field.getEnvironment().testEnv("LY_REPLAY"))
        {
            // Need to refresh the whole view (a call to refresh() only does not
            // work)
            _lastFrameName = LyWebConstants.TARGET_FRAME;
            refresh();
            return;
        }

        boolean refresh = true;
        LyTabs  tabs = _field.getParentTabs();

        if (tabs != null)
        {
            LyStrutsTabs webTabs = (LyStrutsTabs)tabs.getInternalView();
            LyTab        selectedTab = webTabs.getSelectedTab();
            LyTab        fieldTab = _field.getParentTab();

            // Refresh only if field is displayed. Test if field tab is selected
            // tab and its label
            // is not null (tab with a null label contains fields that are
            // allways displayed)
            if ((selectedTab != null) && (fieldTab != null) && (fieldTab != selectedTab)
                    && (fieldTab.getLabel() != null) && (fieldTab.getLabel().length() > 0))
                refresh = false;
        }

        if (!refresh)
            return;

        // Not needed any more
        // If field contains escape character, perform a global refresh
//      LyFieldItem item = _field.getFieldItem();
//
//      if (item instanceof LyGetLabelInterface)
//      {
//          String  label = ((LyGetLabelInterface)item).getLabel();
//          boolean hasEscapeChar = ((label != null) && (label.indexOf('\\') >= 0));
//
//          if (hasEscapeChar)
//          {
//              refresh();
//              return;
//          }
//      }

        LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
        String             id = getIdCode() + "_item";

        viewManager.addReloadDiv(_lastFrameName, id, this);

        if (LyWebConstants.POPUP_VIEW.equals(_lastWindowName))
            viewManager.addAdditionalRefreshScripts("<script type='text/javascript'>\n//<![CDATA[\n" +
                "checkSizePopupView(null, false);\n//]]>\n</script>");
    } // end method refreshDivContent

    /**
     * Refreshes the content of the div containing the caption description.
     */
    public void refreshCaption()
    {
        if ((_field.getEnvironment() != null)
                &&_field.getEnvironment().testEnv("LY_REPLAY"))
        {
            // Need to refresh the whole view (a call to refresh() only does not
            // work)
            _lastFrameName = LyWebConstants.TARGET_FRAME;
            refresh();
            return;
        }

        boolean refresh = true;
        LyTabs  tabs = _field.getParentTabs();

        if (tabs != null)
        {
            LyStrutsTabs webTabs = (LyStrutsTabs)tabs.getInternalView();
            LyTab        selectedTab = webTabs.getSelectedTab();
            LyTab        fieldTab = _field.getParentTab();

            // Refresh only if field is displayed
            if ((selectedTab != null) && (fieldTab != null) && (fieldTab != selectedTab))
                refresh = false;
        }

        if (!refresh)
            return;

        LyWebViewManager viewManager = getWebViewManager();
        StringBuffer     buffer = new StringBuffer();
        LyWebPageWriter  pw = new LyStrutsPage(_field, true);

        pw.setFormId(_formId);
        writeCaption(pw, false);

        StringBuffer innerHtml = pw.getBuffer(pw.getActiveName());
        String       innerHtmlStr = LyString.replace(innerHtml.toString(), "\n", "\\n");

        innerHtmlStr = LyString.replace(innerHtmlStr, "\\", "\\\\");
        innerHtmlStr = LyString.replace(innerHtmlStr, "\\\\n", "\\n");
        innerHtmlStr = LyString.replace(innerHtmlStr, "\"", "\\\"");
        innerHtmlStr = LyString.replace(innerHtmlStr, "\\'", "\\\\'");

        buffer.append("<script type='text/javascript'>\n//<![CDATA[\nreplaceDivInnerHtml('").append(
            _lastFrameName).append("','").append(getIdCode()).append("_caption',\"").append(
            innerHtmlStr).append("\");\n//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(buffer.toString());
        pw.free();
    }
    
    /**
     * Checks differences between components found in a toolbar (toolbars or tools).
     *
     * @param  comp1 the first component of comparison
     * @param  comp2 the second component of comparison
     * @return true if the components differ, false otherwise
     */
    public boolean compareToolbarComponents(LyComponent comp1, LyComponent comp2)
    {
        boolean differ = false;

        if (((comp1 != null) && (comp2 == null)) || ((comp1 == null) && (comp2 != null)))
            differ = true;
        else if ((comp1 != null) && (comp2 != null))
        {
            if ((comp1 instanceof LyTool) && (comp2 instanceof LyTool))
            {
                // Compare tools
                LyTool tool1 = (LyTool)comp1, tool2 = (LyTool)comp2;

                if (((tool1.getImage() == null) && (tool2.getImage() != null))
                        || ((tool1.getImage() != null) && (tool2.getImage() == null))
                        || ((tool1.getImage() != null)
                            && !(tool1.getImage().equals(tool2.getImage()))))
                    differ = true;

                if (testResource("LY_WRITE_TOOL_LABEL"))
                {
                    if (((tool1.getLabel() == null) && (tool2.getLabel() != null))
                            || ((tool1.getLabel() != null) && (tool2.getLabel() == null))
                            || ((tool1.getLabel() != null)
                                && !(tool1.getLabel().equals(tool2.getLabel()))))
                        differ = true;
                }

                if (((tool1.getTooltip() == null) && (tool2.getTooltip() != null))
                        || ((tool1.getTooltip() != null) && (tool2.getTooltip() == null))
                        || ((tool1.getTooltip() != null)
                            && !(tool1.getTooltip().equals(tool2.getTooltip()))))
                    differ = true;

                if (((tool1.getId() == null) && (tool2.getId() != null))
                        || ((tool1.getId() != null) && (tool2.getId() == null))
                        || ((tool1.getId() != null) && !(tool1.getId().equals(tool2.getId()))))
                    differ = true;
            }
            else if ((comp1 instanceof LyToolBar) && (comp2 instanceof LyToolBar))
            {
                // Compare toolbars
                LyToolBar   toolbar1 = (LyToolBar)comp1, toolbar2 = (LyToolBar)comp2;
                int         i, n = toolbar1.getComponentCount(), m = toolbar2.getComponentCount();
                LyComponent subComp1, subComp2;

                if (n != m)
                    differ = true;
                else
                {
                    for (i = 0; i < n; i++)
                    {
                        subComp1 = toolbar1.getComponent(i);
                        subComp2 = toolbar2.getComponent(i);
                        differ |= compareToolbarComponents(subComp1, subComp2);
                    }
                }
            }
            else
                differ = true;
        }

        return differ;
    } // end method compareToolbarComponents
} // end class LyStrutsFormField
