/*
 * LyStrutsHRefs.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.view.web.struts;

import leon.misc.*;
import leon.view.*;
import leon.view.graphic.*;
import leon.view.web.*;


/**
 * This class implements the HRefs object (as described in the dtd) for web viewer.
 *
 * <p><b>HREFS: List of Hyperlinks</b><br>
 * An input field of Hyperlinks type allows to display a graphical component containing a set of
 * hyperlinks. Such a component is a simple list of hyperlinks such as described below. It does not
 * contain any attribute.</p>
 *
 * @version $Id: LyStrutsHRefs.java 41081 2012-10-09 07:04:51Z nhamon $
 * @author  Lyria S.A.
 */
public class LyStrutsHRefs extends LyStrutsFieldItem
    implements LyGraphicalHRefs
{
	public static final String __VERSION =
		"$Id: LyStrutsHRefs.java 41081 2012-10-09 07:04:51Z nhamon $";

	/** The HRefs generic component. */
    protected LyHRefs          _hRefs;
    
	/**
	 * Constructor.
	 *
	 * @param hRefs the hrefs component
	 */
	public LyStrutsHRefs(LyHRefs hRefs)
	{
        super(hRefs);
        
        _hRefs = hRefs;
	}

	/**
	 * Overriden to get the appplication context URL.
	 *
	 * @see LyWebViewManager#getRootURL()
	 */
	protected String getBaseUrl(LyWebPageWriter pw)
	{
		// get the application context url.
		return ((LyWebViewManager)getViewManager()).getRootURL();
	}
	
	   /**
     * Refresh frame containing this web object.<br>
     * Overriden in order to refresh directly the parent of this href if it is a form field, else
     * calls the super impplementation.
     *
     * @see leon.view.web.LyWebObject#refresh()
     */
    public void refresh()
    {
        LyComponent parent = _hRefs.getParent();

        if (parent instanceof LyFormField)
        {
            LyFormField      formField = (LyFormField)parent;
            LyStrutsFormField webField = (LyStrutsFormField)formField.getInternalView();

            webField.refreshDivContent();
        }
        else
            super.refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
     */
    public void writeHtml(LyWebPageWriter pw)
    {
        pw.startComponent(this);
        
        updateRefreshInfo(pw);

        LyHRef      href;
        LyStrutsHRef internal;
        boolean     drawImage = containsImage();

        // Compute base URL
        LyEnvironment   environment = getEnvironment();
        String          baseUrl = getBaseUrl(pw);
        int             height = _hRefs.getNRows();

        if (height <= 0)
            height = environment.getIntResource("LY_MULTIPLE_RELATION_CONSULT_HEIGHT");
        
        height = 22 * height + 2;

        pw.addBeginScript("<script type='text/javascript'>\n//<![CDATA[\nvar _rootUrl='" + baseUrl
            + "';\n//]]>\n</script>", false);

        pw.append("<div class='readOnlyField'");
        
        if (!pw.isPrintMode())
        {
            pw.append(" style='");
            
            if (pw.getActiveName() != LyWebConstants.TOOL_POPUP)
            {
                pw.append("overflow:auto;height:" + height + "px;");

                if (environment.getIntResource("LY_MULTIPLE_RELATION_CONSULT_WIDTH") > 0)
                    pw.append("width:" + environment.getIntResource("LY_MULTIPLE_RELATION_CONSULT_WIDTH") + ";");
            }
            
            if (!_hRefs.isBorder())
                pw.append("border:0px;padding:2px;");
            
            pw.append("'");
        }
        
        pw.append(">");

        if (drawImage)
            pw.append("<table border='0' cellspacing='0' cellpadding='1'>");

        int i, n = _hRefs.getHRefCount();

        for (i = 0; i < n; i++)
        {
            href = _hRefs.getHRef(i);
            internal = (LyStrutsHRef)href.getInternalView();

            pw.append("\n");

            if (drawImage)
            {
                pw.append("<tr valign='top'><td width='0' valign='middle'>");
                internal.writeImage(pw);
                pw.append("</td><td valign='middle' nowrap='yes'>");
            }

            internal.writeLabel(pw);

            if (drawImage)
                pw.append("</td></tr>");
            else
                pw.append("<br />");
        }

        if (drawImage)
            pw.append("</table>");
        
        pw.append("</div>");
        
        pw.endComponent(this);
    } // end method writeHtml

    /**
     * Indicates if at least one of the href has an image.
     *
     * @return true if at least one of the href has an image, false otherwise
     */
    protected boolean containsImage()
    {
        int    i, n = _hRefs.getHRefCount();
        LyHRef href;
        String image;

        for (i = 0; i < n; i++)
        {
            href = _hRefs.getHRef(i);
            image = href.getImage();

            // If an image was found return true
            if ((image != null) && (image.length() != 0))
                return true;
        }

        return false;
    }

    /* (non-Javadoc)
     * @see leon.view.setinterface.LySetHRefsInterface#removeHRef(leon.view.LyHRef)
     */
    public void removeHRef(LyHRef hRef)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetHRefsInterface#addHRef(leon.view.LyHRef,
     *      int)
     */
    public void addHRef(LyHRef hRef, int pos)
    {
        refresh();
    }
}