/*
 * LyStrutsHeader.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web.struts;

import java.awt.*;

import leon.misc.*;
import leon.view.*;
import leon.view.graphic.*;
import leon.view.web.*;


/**
 * This class implements the Header object (as described in the dtd) for Struts viewer.
 *
 * <p><b>HEADER: Header Cell</b><br>
 * This represents the header of a cell.</p>
 *
 * @version $Id: LyStrutsHeader.java 35801 2012-06-11 08:33:47Z nhamon $
 * @author  Lyria
 * @see     leon.view.LyHeader
 */
public class LyStrutsHeader extends LyWebObject
    implements LyGraphicalHeader
{
	public static final String __VERSION =
		"$Id: LyStrutsHeader.java 35801 2012-06-11 08:33:47Z nhamon $";

	/** The Header generic component. */
    protected LyHeader         _header;

    /** Image for increase button. */
    protected String           _increaseButton;

    /** Image for decrease button. */
    protected String           _decreaseButton;

    /** Sort tooltip shown to the user. */
    protected String           _sortBy;
    
    
	/**
	 * Class constructor of a Header object for the Struts target.
	 *
	 * @param header the generic Header defined for every type of view targets
	 */
	public LyStrutsHeader(LyHeader header)
	{
        super(header);
        
        _header = header;
        initialize();
	}
	
	/**
	 * Performs a refresh of the table.
	 */
	public void refresh()
	{
		LyHeaders	   headers = (LyHeaders)_header.getParent();
		LyTable		   table = headers.getParentTable();
		LyStrutsTable   htmlTable = (LyStrutsTable)table.getInternalView();
		
		htmlTable.refresh();
	}

	/**
	 * Returns the script that must be used to sort header.
	 * 
	 * @param pw current page writer.
	 * @param headerIndex index of the header.
	 */
	protected String getSortScript(LyWebPageWriter pw, int headerIndex)
	{
		LyHeaders	   headers = (LyHeaders)_header.getParent();
		LyTable		   table = headers.getParentTable();
		StringBuffer   script = new StringBuffer();

		// Retrieve action URL
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);

		script.append("javascript:InvokeHeaderClick('");
		script.append(table.getRootController());
		script.append("','");
		script.append(headerIndex);
		script.append("','");
		script.append(actionUrl);
		script.append("',event);");
		
		return script.toString();
	}
	
	/**
	 * Returns the url sent to Javascript function for popup menus.
	 * 
	 * @return the graphical action URL.
	 */
	protected String getPopupMenuUrl()
	{
		// get the url to the event handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		
		return actionUrl;
	}
	
	   /**
     * Initialization method.<br>
     * Sets the buttons images with values from resources files. Sets the sort tooltip with an HTML
     * compatible representation of the label defined in the resource LY_MENU_SORT_BY.
     */
    public void initialize()
    {
        _increaseButton = getImageUrl("LY_TABLE_SORT_ASC");
        _decreaseButton = getImageUrl("LY_TABLE_SORT_DES");

        _sortBy = translate("LY_MENU_SORT_BY");
        _sortBy = getJavascriptString(_sortBy);

    }

    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
     */
    @SuppressWarnings("deprecation")
	public void writeHtml(LyWebPageWriter pw)
    {
    	if (_header.isFreed() || _header.isFreeing())
    		return;
        pw.startComponent(this);
           
        updateRefreshInfo(pw);

        LyHeaders      headers = (LyHeaders)_header.getParent();
        LyTable        table = headers.getParentTable();
        LyStrutsHeaders htmlHeaders = (LyStrutsHeaders)headers.getInternalView();
        int            headerIndex = getHeaderIndex();
        String         rowspan = ((htmlHeaders.getRowSpan() > 1)
                ? (" rowspan='" + htmlHeaders.getRowSpan()
                    + "'") : "");

        String  script = getSortScript(pw, headerIndex);
        boolean sortable = _header.isSortable();
        int     sort = _header.getSort();

        boolean increasing = _header.isIncreasing();
        String  align = LyStrutsTable.getHtmlHAligment(_header.getAlign());
        boolean isWrap = _header.isWrap();
        String  tooltip = translate(_header.getTooltip());

        tooltip = LyString.replace(tooltip, "\"", "\\\"");

        // Check label with header width
        String displayedLabel = translate(_header.getLabel());

        displayedLabel = checkHeaderWidth(displayedLabel,
                (sortable ? "LY_TABLE_LABEL_SORTABLE" : "LY_TABLE_LABEL"), true);

        String htmlLabel = getHtmlString(displayedLabel);

        if ((htmlLabel == null) || (htmlLabel.trim().length() == 0))
            htmlLabel = "&nbsp;";
        
        pw.append("<td");
        pw.append(align);
        pw.append(rowspan);
        
        double headerWidth = _header.getColumnWidth();
        
        if (headerWidth >= 1)
        {
			LyEnvironment environment = getEnvironment();
			Font		  defaultFont = (Font)environment.getFont("LY_TABLE");
			Toolkit		  toolkit = Toolkit.getDefaultToolkit();
			FontMetrics   defaultFm = toolkit.getFontMetrics(defaultFont);
			float		  charWidth = ((float)defaultFm.stringWidth("0123456789")) / 10;
			
			pw.append(" width='" + (int)(charWidth * headerWidth) + "px' ");
        }
        else if (headerWidth > 0)
            pw.append(" width='" + headerWidth * 100 + "%' ");
        
        if (!isWrap)
            pw.append(" nowrap='true'");

        LyWebDecoration deco = getWebViewManager().getDecoration();
        String          style = deco.getClassName(_header, pw);

        pw.append(" class='" + style + "'");
		
		String localStyle = "";

		if (!headers.isDefaultColor())
		{
			localStyle += " background-image:none;";
			localStyle += " background-color:" + getHexColor(headers.getColor()) + ";";
		}

		if (!headers.isDefaultFont())
			localStyle += getFontString(headers.getFont(), null);
	
		if (sortable && !pw.isPrintMode())
        {
			pw.append(" style='cursor:hand;cursor:pointer;" + localStyle);
            
            if (_header.getBorderWidth() > 0)
                pw.append("border-right:0px;");

            pw.append("' onClick=\"" + script.toString() + "\"");

            if (tooltip != null)
            {
                pw.append(" title='");
                pw.append(LyString.replace(tooltip, "'", "&#39;"));
                pw.append("'");
            }

            if (style.equals("headerSorted") || style.equals("sortableheader"))
            {
                pw.append(
                    " onMouseOut=\"javascript:changeStyle(this,'" + style + "','" + style + "AndHighlighted')\" ");
                pw.append(
                    " onMouseOver=\"javascript:changeStyle(this,'" + style + "','" + style + "AndHighlighted');return true;\" ");
            }
        }
        else
        {
            if (_header.getBorderWidth() > 0)
				pw.append(" style='border-right:0px;'"+localStyle);
            
            if (!pw.isPrintMode() && (tooltip != null))
            {
                pw.append(" title='");
                pw.append(LyString.replace(tooltip, "'", "&#39;"));
                pw.append("'");
            }
        }

        pw.append(" onContextMenu=\"javascript:InvokePopup('" + table.getId() + "_headers','");
        pw.append(getHeaderIndex() + "','" + getPopupMenuUrl() + "','" + table.getController() + "',event.ctrlKey,event.clientX,event.clientY);");
        pw.append("return false;\"");
        pw.append(">");

		if (_header.getImage() != null)
			pw.append("<img border='0' align='middle' src='" + getImageUrl(_header.getImage()) + "'/>&nbsp;");
		
		if (!headers.isDefaultFontColor())
			pw.append("<font style='color:" + getHexColor(headers.getFontColor()) + ";'>");

		pw.append(htmlLabel);
		
		if (!headers.isDefaultFontColor())
			pw.append("</font>");

		if (sort >= 0)
		{
            pw.append("&nbsp;<img src='");
            if (increasing)
                pw.append(_increaseButton);
            else
                pw.append(_decreaseButton);

            pw.append("' hspace='2' vspace='0' /><font face='Arial' size='-3'>");
            pw.append(Integer.toString(sort + 1));
            pw.append("</font>");
        }

        pw.append("</td>\n");
        
        // Write the border if necessary
        if (_header.getBorderWidth() > 0)
        {
            pw.append("<td bgColor='" + getHexColor(_header.getBorderColor())
                    + "'><div style='width:" + _header.getBorderWidth() + "px;'></div></td>\n");
        }
        
        pw.endComponent(this);
    } // end method writeHtml

    /**
     * Checks a label is displayable in the header according to header width.
     *
     * @param  label      the label
     * @param  fontString the font property used to display label
     * @param  isHeader   indicates if label is displayed in an header
     * @return the label adjusted to header width if it is necessary
     */
    public String checkHeaderWidth(String label, String fontString, boolean isHeader)
    {
        // Check header width
        double headerWidth = _header.getColumnWidth();

        if (headerWidth < 1)
            return label;

        // If header is sorted, add 2 characters for cell content
        if (!isHeader && (_header.getSort() >= 0))
            headerWidth += 2;

        // If label is a multi-line label, compute reduced string for each line.
        if ((label != null) && (label.indexOf('\n') >= 0))
        {
            StringBuffer    result = new StringBuffer();
            String          subResult;
            String[]        list = LyString.stringToList(label, '\n');
            int             i, n = (list != null ? list.length : 0);
            
            for (i = 0; i < n; i++)
            {
                if (i > 0)
                    result.append('\n');
                
                subResult = checkHeaderWidth(list[i], fontString, isHeader);
                
                if (subResult != null)
                    result.append(subResult);
            }
            
            return result.toString();
        }

		label = getReducedString(label, fontString, (int)headerWidth);

        return label;
    } // end method checkHeaderWidth

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalHeader#setIncreasing(boolean)
     */
    public void setIncreasing(boolean increasing)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalHeader#setFixed(boolean)
     */
    public void setFixed(boolean fixed)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalHeader#setSort(int)
     */
    public void setSort(int sort)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalHeader#setBorderColor(String)
     */
    public void setBorderColor(String borderColor)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalHeader#setBorderWidth(int)
     */
    public void setBorderWidth(int borderWidth)
    {
        refresh();
    }

    /**
     * Gets the header index from the headers list.
     *
     * @return the header index
     */
    public int getHeaderIndex()
    {
    	if (_header.isFreed() || (_header.isFreeing()))
    		return -1;
        LyHeaders headers = (LyHeaders)_header.getParent();
        LyTable   table = headers.getParentTable();
        LyHeader  header;

        // Retrieve top level headers
        headers = table.getHeaders();

        int i, n = headers.getHeaderCount();

        for (i = 0; i < n; i++)
        {
            header = headers.getHeader(i);
            if (header == _header)
                return i;
        }

        return -1;
    }

	/**
	 * Sets attribute columnWidth to the given value.
	 * @param columnWidth New value of columnWidth.
	 */
	public void setColumnWidth(double columnWidth)
	{
	}
} // end class LyStrutsHeader
