/*
 * LyStrutsInterval.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web.struts;

import java.text.*;
import java.util.*;

import leon.view.*;
import leon.view.graphic.*;
import leon.view.web.*;


/**
 * This class implements the Interval object (as described in the dtd) for web viewer.
 *
 * <p><b>INTERVAL: Interval</b><br>
 * An interval is defined by a begin date and an end date. If the end date is not specified, the
 * interval is replaced by a point (an image representing a date). An interval has a label and a
 * color. It may be resizable (it is possible to change its begin and end dates), movable on its
 * parent line and may have vertical lines on its extremities. It may also have dependencies to
 * other intervals represented by arrows linking this interval with the pointed intervals. An
 * interval may also contain periods allowing to split it into several sub-intervals. These periods
 * are described by a begin date and an end date.</p>
 *
 * @version $Id: LyStrutsInterval.java 16172 2009-01-06 11:04:44Z jls $
 * @author  Lyria
 * @see     leon.view.LyInterval
 */
public class LyStrutsInterval extends LyWebObject
	implements LyGraphicalInterval
{
	public static final String __VERSION =
		"$Id: LyStrutsInterval.java 16172 2009-01-06 11:04:44Z jls $";

	/** The interval generic component. */
	protected LyInterval	   _interval;
	
	/** Begin date cache **/
	protected Date _beginDate;
	
	/** End date cache **/
	protected Date _endDate;

	/**
	 * Class constructor of an Interval object for the struts target.
	 *
	 * @param interval the generic Interval defined for every type of view targets
	 */
	public LyStrutsInterval(LyInterval interval)
	{
		super(interval);
		_interval = interval;
	}

	/**
	 * This method does nothing.
	 */
	public void writeHtml(LyWebPageWriter pw)
	{
	}

	/**
	 * Writes HTML description corresponding to this interval.
	 *
	 * @param pw the page writer
	 */
	public void writeInterval(LyWebPageWriter pw)
	{
		updateRefreshInfo(pw);
	}

	/**
	 * Return the begin date of the interval using the date format.
	 *
	 * @return the begin date
	 */
	public Date getBeginDate()
	{
		if (_beginDate == null)
			_beginDate = parseDate(_interval.getBeginDate());
		return _beginDate;
	}

	/**
	 * Return the end date of the interval using the date format.
	 *
	 * @return the end date
	 */
	public Date getEndDate()
	{
		if (_endDate == null)
			_endDate = parseDate(_interval.getEndDate());
		return _endDate;
	}
	/**
	 * Parses a date value using date time format and date format.
	 *
	 * @param  value the date value
	 * @return the parsed date
	 */
	public Date parseDate(String value)
	{
		DateFormat dateFormat = getDateTimeFormat();
		Date	   date = null;

		try
		{
			date = dateFormat.parse(value);
		}
		catch (Exception e)
		{
			try
			{
				// If date could not be parsed with date and time format, try
				// with date only
				dateFormat = getDateFormat();
				date = dateFormat.parse(value);
			}
			catch (Exception e2)
			{
			}
		}

		return date;
	}

	/**
	 * Gets the date and time format associated with this interval.
	 *
	 * @return the date and time format
	 */
	protected DateFormat getDateTimeFormat()
	{
		// Get the date format from the Gantt component
		LyGantt		 gantt = getParentGantt();
		LyStrutsGantt htmlGantt = (LyStrutsGantt)gantt.getInternalView();
		DateFormat   dateFormat = htmlGantt.getDateTimeFormat();

		return dateFormat;
	}

	/**
	 * Gets the date format associated with this interval.
	 *
	 * @return the date format
	 */
	protected DateFormat getDateFormat()
	{
		// Get the date format from the Gantt component
		LyGantt		 gantt = getParentGantt();
		LyStrutsGantt htmlGantt = (LyStrutsGantt)gantt.getInternalView();
		DateFormat   dateFormat = htmlGantt.getDateFormat();

		return dateFormat;
	}

	/**
	 * Gets the parent Gantt component of the interval.
	 *
	 * @return the parent Gantt component
	 */
	protected LyGantt getParentGantt()
	{
		LyComponent parent = _interval.getParent();

		if (parent instanceof LyLine)
		{
			LyLine	    line = (LyLine)parent;
			LyStrutsLine htmlLine = (LyStrutsLine)line.getInternalView();

			return htmlLine.getParentGantt();
		}

		return null;
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalInterval#addPeriod(leon.view.LyPeriod,
	 *      int)
	 */
	public void addPeriod(LyPeriod period, int pos)
	{
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalInterval#removePeriod(leon.view.LyPeriod)
	 */
	public void removePeriod(LyPeriod period)
	{
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalInterval#setBeginDate(java.lang.String)
	 */
	public void setBeginDate(String beginDate)
	{
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.setinterface.LySetColorInterface#setColor(java.lang.String)
	 */
	public void setColor(String color)
	{
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalInterval#setDependencies(java.lang.String)
	 */
	public void setDependencies(String dependencies)
	{
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalInterval#setDrawVerticalLines(boolean)
	 */
	public void setDrawVerticalLines(boolean drawVerticalLines)
	{
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalInterval#setEndDate(java.lang.String)
	 */
	public void setEndDate(String endDate)
	{
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.setinterface.LySetFontInterface#setFont(java.lang.String)
	 */
	public void setFont(String font)
	{
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalInterval#setForegroundColor(java.lang.String)
	 */
	public void setForegroundColor(String foregroundColor)
	{
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
	 */
	public void setLabel(String label)
	{
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.setinterface.LySetLabelInterface#setTooltip(java.lang.String)
	 */
	public void setTooltip(String tooltip)
	{
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.setinterface.LySetLabelInterface#setCenterLabel(java.lang.String)
	 */
	public void setCenterLabel(String label)
	{
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalInterval#setLeftResizable(boolean)
	 */
	public void setLeftResizable(boolean leftResizable)
	{
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalInterval#setMovable(boolean)
	 */
	public void setMovable(boolean movable)
	{
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalInterval#setRightResizable(boolean)
	 */
	public void setRightResizable(boolean rightResizable)
	{
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalInterval#setVerticalLinesColor(java.lang.String)
	 */
	public void setVerticalLinesColor(String verticalLinesColor)
	{
	}
	
	/**
	 * Refresh frame containing this web object.
	 */
	public void refresh()
	{
		// clear cache
		_beginDate = null;
		_endDate = null;

		super.refresh();
	}
} // end class LyStrutsInterval
