/*
 * LyStrutsLine.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web.struts;

import java.awt.*;
import java.util.*;

import leon.misc.*;
import leon.view.*;
import leon.view.graphic.*;
import leon.view.web.*;
import leon.view.web.struts.gantt.*;
import leon.view.web.struts.gantt.LyStrutsGanttElement.*;


/**
 * This class implements the Line object (as described in the dtd) for web viewer.
 *
 * <p><b>LINE: Line of a planning</b><br>
 * </p>
 *
 * @version $Id: LyStrutsLine.java 36570 2012-08-24 09:35:45Z nhamon $
 * @author  Lyria
 * @see     leon.view.LyTreeNode
 */
public class LyStrutsLine extends LyWebObject
    implements LyGraphicalLine
{
	public static final String __VERSION =
		"$Id: LyStrutsLine.java 36570 2012-08-24 09:35:45Z nhamon $";

	/** The line generic component. */
    protected LyLine           _line;

    /** The parent line HTML tag representation. */
    protected String           _parentString;

    /** Is last is true if this line hasn't any child, false otherwise. */
    protected boolean          _isLast;
    
	/**
	 * Class constructor of a Line object for the Struts target.
	 *
	 * @param treeNode the generic Tree Node defined for every type of view targets
	 */
	public LyStrutsLine(LyLine line)
	{
        super(line);
        
        _line = line;
        _parentString = null;
	}

	/**
	 * Gets the javascript code used to consult a line in the current Gantt.
	 *
	 * @param  pw the page writer.
	 * @return the javascript code used to consult a line in the current Gantt
	 */
	protected String getConsultNodeScript(LyWebPageWriter pw)
	{
		// get the url to the selection handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		String				id = _line.getId();

		id = LyString.getJavascriptString(id);

		if (pw.writeDivOnly())
			id = LyString.replace(id, "\\", "\\\\");
		
		return "javascript:ConsultTreeNode(\"" + id + "\", \"" + _line.getRootController() + "\", \"" + getParentGantt().getId() + "\", \"" + actionUrl + "\");";
	}

	/**
	 * Gets the javascript code used to select a line in the current Gantt.
	 *
	 * @param  pw the page writer.
	 * @return the javascript code used to select a line in the current Gantt
	 */
	protected String getSelectNodeScript(LyWebPageWriter pw)
	{
		// get the url to the selection handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    selectionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_SELECTEVENT);
		String				id = _line.getId();

		id = LyString.getJavascriptString(id);

		if (pw.writeDivOnly())
			id = LyString.replace(id, "\\", "\\\\");
		
		return "javascript:SelectTreeNode(\"" + id + "\", \"" + _line.getRootController() + "\", \"" + selectionUrl + "\");";
	}

	/**
	 * Gets the javascript code used to expand a line in the current Gantt.
	 *
	 * @param  pw the page writer.
	 * @return the javascript code used to expand a line in the current Gantt
	 */
	protected String getExpandNodeScript(LyWebPageWriter pw)
	{
		// get the url to the selection handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		String				id = _line.getId();

		id = LyString.getJavascriptString(id);

		if (pw.writeDivOnly())
			id = LyString.replace(id, "\\", "\\\\");
		
		return "javascript:ExpandTreeNode(\"" + id + "\", \"" + _line.getRootController() + "\", \"" + actionUrl + "\");";
	}

	/**
	 * Gets the javascript code used for popup menu in the current Gantt.
	 *
	 * @param  pw the page writer.
	 * @return the javascript code used for popup menu in the current Gantt
	 */
	protected String getPopupMenuNodeScript(LyWebPageWriter pw)
	{
		// get the url to the selection handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		String			    selectionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_SELECTEVENT);
		String				id = _line.getId();

		id = LyString.getJavascriptString(id);

		if (pw.writeDivOnly())
			id = LyString.replace(id, "\\", "\\\\");
		
		return "javascript:InvokeTreeNodePopup('" + id + "', '', '" + getParentGantt().getId() + "', '" +
			_line.getRootController() + "', '" + actionUrl + "', '" + selectionUrl + "', event);return false;";
	}
	
	   /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
     */
    public void writeHtml(LyWebPageWriter pw)
    {
    }

    /**
     * Gets the begining of the HTML representation : the parent line HTML representation.
     *
     * @return the parent line String representation
     */
    public String getBeginning()
    {
        return _parentString;
    }

    /**
     * Prepare the beginnining : builds the given parent line string representation.
     *
     * @param parent the parent line
     * @param isLast true if the line is the last
     */
    public void prepareBeginning(LyStrutsLine parent, boolean isLast)
    {
        if (parent == null)
            return;

        StringBuffer buffer = new StringBuffer();
        String       nodeIcon, parentString;

        if (isLast)
            nodeIcon = getImageUrl("LY_EMPTY_NODE");
        else
            nodeIcon = getImageUrl("LY_VERTICAL_LINE");

        parentString = parent.getBeginning();
        if (parentString != null)
            buffer.append(parentString);
        buffer.append("<td><img src=\"" + nodeIcon + "\" /></td>");

        _parentString = buffer.toString();
    }

    /**
     * Gets the number of visible children lines of the given line.
     *
     * @param  line the given line
     * @return the number of children which are visible
     */
    protected int getVisibleLineCount(LyLine line)
    {
        if (line == null)
            return 0;

        int         i, n = line.getComponentCount();
        int         count = 0;
        LyComponent son;

        for (i = 0; i < n; i++)
        {
            son = line.getComponent(i);

            if ((son instanceof LyLine) && !((LyLine)son).isHidden())
                count++;
        }

        return count;
    }

    /**
     * Computes the height of the line.
     *
     * @return the line height
     */
    public int computeHeight()
    {
        int   lineHeight = 22;
        int[] imageSize = getImageSize("LY_NODE");

        if ((imageSize != null) && (imageSize.length > 0))
            lineHeight = imageSize[1];

        return lineHeight;
    }

    /**
     * Compute line bounds and send it to the Gantt component.
     */
    public void computeLineBounds()
    {
        int              lineHeight = computeHeight();

        // Register line position on the gantt component
        LyGantt      gantt = getParentGantt();
        LyStrutsGantt htmlGantt = (LyStrutsGantt)gantt.getInternalView();

        htmlGantt.setLineBounds(this, lineHeight);
        
        boolean      isOpen = _line.isOpen();

        // Write recursively all child nodes
        if (isOpen)
        {
            LyComponent comp;
            LyStrutsLine webLine;
            int         i;
            int         n = _line.getComponentCount();

            for (i = 0; i < n; i++)
            {
                comp = _line.getComponent(i);

                if (!(comp instanceof LyLine) || ((LyLine)comp).isHidden())
                    continue;

                webLine = (LyStrutsLine)comp.getInternalView();
                webLine.computeLineBounds();
            }
        }
    }
    
    /**
     * Write HTML description corresponding to this line.
     *
     * @param pw     the page writer
     * @param parent the parent line
     * @param isLast true if this line is the last one
     */
    public void writeLine(LyWebPageWriter pw, LyStrutsLine parent, boolean isLast)
    {
        updateRefreshInfo(pw);

        _isLast = isLast;

        // First Write the node
        int              nVisible = getVisibleLineCount(_line);
        String           label = _line.getLabel();
        String           translatedLabel = translate(label);
        String           image = _line.getImage();
        boolean          isOpen = _line.isOpen();
        boolean          isSelected = _line.isSelected();
        String           icon;
        String           nodeIcon;
        String           beginning = null;
        String           openNode = null;
        String           consultLink = null;
        String           selectLink = null;
        String           selectedColor = getHexColor("LY_TREENODE_SELECTED_BG_COLOR");
        String           selectedTextColor;
        String           toolTip = getHtmlString(translatedLabel);
        String           javascriptToolTip;

        selectedTextColor = getHexColor("LY_TREENODE_SELECTED_FONT_COLOR");
        label = getHtmlString(translatedLabel);

        if (parent != null)
            beginning = parent.getBeginning();
        prepareBeginning(parent, isLast);

        if (isLast)
        {
            if (nVisible > 0)
            {
                if (isOpen)
                    nodeIcon = getImageUrl("LY_LAST_NODE_OPEN");
                else
                    nodeIcon = getImageUrl("LY_LAST_NODE_CLOSED");
            }
            else
                nodeIcon = getImageUrl("LY_LAST_NODE");
        }
        else if (nVisible > 0)
        {
            if (isOpen)
                nodeIcon = getImageUrl("LY_NODE_OPEN");
            else
                nodeIcon = getImageUrl("LY_NODE_CLOSED");
        }
        else
            nodeIcon = getImageUrl("LY_NODE");

        if (image != null)
            icon = getImageUrl(image);
        else
        {
            if (isOpen)
                icon = getImageUrl("LY_DIR_OPEN");
            else
                icon = getImageUrl("LY_DIR_CLOSED");
        }

        String lineId = getLineId();
        String id = _line.getId();

        id = LyString.getJavascriptString(id);

        if (pw.writeDivOnly())
            id = LyString.replace(id, "\\", "\\\\");

        if (nVisible > 0)
            openNode = getExpandNodeScript(pw);

        consultLink = getConsultNodeScript(pw);

        selectLink = getSelectNodeScript(pw);

        pw.append("<div id=\"" + lineId + "_div\">");
        pw.append("<table border='0' cellspacing='0' cellpadding='0'>\n");
        pw.append("<tr>");
        if (beginning != null)
            pw.append(beginning);

        // Don't draw nodeIcon, if we are the root node
        if (parent != null)
        {
            pw.append("<td>");
            if (openNode != null)
            {
                if (isOpen)
                    javascriptToolTip = LyString.getJavascriptString(translate("LY_CLOSE_NODE")
                            + " " + translatedLabel);
                else
                    javascriptToolTip = LyString.getJavascriptString(translate("LY_OPEN_NODE")
                            + " " + translatedLabel);

                pw.append("<a href='" + openNode.toString());
                pw.append("' onmouseover=\'window.status=\"");
                pw.append(javascriptToolTip);
                pw.append("\";return true\'>");
            }
            pw.append("<img src=\"" + nodeIcon + "\" border='no' title='");
            pw.append(toolTip);
            pw.append("'/>");
            if (openNode != null)
                pw.append("</a>");
            pw.append("</td>");
        }

        // the node icon is used to consult a node
        pw.append("<td>");
        pw.append("<a ");

        javascriptToolTip = LyString.getJavascriptString(translate("LY_LINK_TO") + " "
                + translatedLabel);

        pw.append("href='");
        pw.append(consultLink.toString());
        pw.append("' onmouseover=\'window.status=\"");
        pw.append(javascriptToolTip);
        pw.append("\";return true\'>");
        pw.append("<img src=\"" + icon + "\" width='20' height='20' border='no' title='");
        pw.append(toolTip);
        pw.append("'/></a></td>");

        pw.append("<td>&nbsp;</td>");

        if (isSelected)
            pw.append("<td id='" + lineId + "_td' bgcolor='" + selectedColor + "' nowrap='true'>");
        else if (_line.getColor() != null)
        {
            String  nodeColor = getHexColor(_line.getColor());
            
            pw.append("<td id='" + lineId + "_td' bgcolor='" + nodeColor + "' nowrap='true'>");
        }
        else
            pw.append("<td id='" + lineId + "_td' nowrap='true'>");

        javascriptToolTip = LyString.getJavascriptString(translate("LY_SELECT_NODE") + " "
                + translatedLabel);

        pw.append("<a href='");
        pw.append(selectLink.toString());
        pw.append("' class='treeNodeLabel' onmouseover=\'window.status=\"");
        pw.append(javascriptToolTip);
        pw.append("\";return true\'>");
        pw.append("<font id='" + lineId + "_font'");

        if (isSelected)
            pw.append(" color='" + selectedTextColor + "'");
        else
            pw.append(" color='" + getHexColor(_line.getFontColor()) + "'");

        pw.append(">");
        pw.append("&nbsp;" + label + "&nbsp;");
        pw.append("</font>");
        pw.append("</a>");

        pw.append("</td>");
        pw.append("</tr>");
        pw.append("</table>\n");

        writeChildren(pw);

        pw.append("</div>\n");
    } // end method writeLine

    /**
     * Write the HTML code for children of current tree node.
     *
     * @param pw current page writer
     */
    protected void writeChildren(LyWebPageWriter pw)
    {
        int          nVisible = getVisibleLineCount(_line);
        boolean      isOpen = _line.isOpen();
        StringBuffer selectLink = null;

        selectLink = new StringBuffer();
        selectLink.append(getSelectNodeScript(pw));

        // Write recursively all child nodes
        if (isOpen)
        {
            LyComponent comp;
            LyStrutsLine webLine;
            int         i, j = 0;
            int         n = _line.getComponentCount();

            for (i = 0; i < n; i++)
            {
                comp = _line.getComponent(i);

                if (comp instanceof LyInterval)
                {
                    LyInterval      interval = (LyInterval)comp;
                    LyStrutsInterval htmlInterval = (LyStrutsInterval)interval.getInternalView();

                    htmlInterval.writeInterval(pw);
                }

                if (!(comp instanceof LyLine) || ((LyLine)comp).isHidden())
                    continue;

                webLine = (LyStrutsLine)comp.getInternalView();
                webLine.writeLine(pw, this, (j == (nVisible - 1)));
                j++;
            }
        }
        // look for selected children then add script to select current node.
        else if (hasChildSelected())
        {
            StringBuffer buffer = new StringBuffer();

            buffer.append("<script type='text/javascript'>\n//<![CDATA[\n");
            buffer.append(selectLink.toString());
            buffer.append("\n//]]>\n</script>");

            if (!pw.writeDivOnly())
                pw.append(buffer.toString());
            else
            {
                LyWebViewManager viewManager = getWebViewManager();

                viewManager.addAdditionalRefreshScripts(buffer.toString());
            }
        }
    } // end method writeChildren

    /**
     * Gets the identifier of this line in HTML code.
     *
     * @return the line id
     */
    protected String getLineId()
    {
        String lineId = _line.getId();

        lineId = lineId.replace('@', '_');
        lineId = lineId.replace('\\', '_');

        return lineId;
    }

    /**
     * Draws the line on the planning.
     *
     * @param g           the graphic context
     * @param bounds      the bounds of place reserved for the line
     * @param scaleHeight scale height
     */
    public void drawLine(Graphics g, Rectangle bounds, int scaleHeight)
    {
        LyGantt            gantt = getParentGantt();
        
        if (gantt == null)
            return;
        
        LyStrutsGantt       htmlGantt = (LyStrutsGantt)gantt.getInternalView();
        LyStrutsGanttCanvas ganttCanvas = htmlGantt.getGanttCanvas();

        if (isSuperLine())
        {
            if (!_line.isSuperLineHidden())
            {
                Date beginDate = getLimitDate(false);
                Date endDate = getLimitDate(true);

                if ((beginDate != null) && (endDate != null))
                {
                    LyStrutsGanttElement ganttElement = new LyStrutsGanttElement(
                            LyStrutsGanttElement.TYPE_SUPER_LINE);
                    int                 beginX = ganttCanvas.getDateAbscisse(beginDate);
                    int                 endX = ganttCanvas.getDateAbscisse(endDate);
                    String              superLineColor = _line.getSuperLineColor();

                    if ((superLineColor != null) && (superLineColor.length() > 0))
                    {
                        ganttElement.setColor(htmlGantt.getColor(superLineColor));
                        ganttElement.setLabelColor(htmlGantt.getColor(superLineColor));
                    }
                    else
                    {
                        ganttElement.setColor(htmlGantt.getColor("LY_GANTT_SUPERINTERVAL_COLOR"));
                        ganttElement.setLabelColor(htmlGantt.getColor(
                                "LY_GANTT_SUPERINTERVAL_COLOR"));
                    }

                    ganttElement.setLabel(_line.getSuperLineLabel());
                    ganttElement.setTooltip(_line.getTooltip());
                    ganttElement.setId(_line.getId());
                    ganttElement.setFont(htmlGantt.getFont("LY_GANTT_FONT"));
                    ganttElement.draw(g, ganttCanvas, beginX, endX, bounds.y + scaleHeight,
                        bounds.y + scaleHeight + bounds.height);
                }
            }
        }
        else
        {
            int i, n = _line.getComponentCount();

            for (i = 0; i < n; i++)
            {
                LyComponent component = _line.getComponent(i);

                if (component instanceof LyInterval)
                {
                    LyInterval          interval = (LyInterval)component;
                    LyStrutsInterval     htmlInterval = (LyStrutsInterval)interval.getInternalView();
                    Date                beginDate = htmlInterval.getBeginDate();
                    Date                endDate = htmlInterval.getEndDate();
                    int                 beginX = ganttCanvas.getDateAbscisse(beginDate);
                    int                 endX = ((endDate == null)
                            ? beginX : ganttCanvas.getDateAbscisse(endDate));
                    LyStrutsGanttElement ganttElement = new LyStrutsGanttElement((endDate == null)
                            ? LyStrutsGanttElement.TYPE_POINT_INTERVAL
                            : LyStrutsGanttElement.TYPE_REAL_INTERVAL);
                    String              label = interval.getLabel();
                    String              centerLabel = interval.getCenterLabel();
                    String              tooltip = interval.getTooltip();
                    String              color = interval.getColor();
                    String              font = interval.getFont();

                    if ((color != null) && (color.length() > 0))
                        ganttElement.setColor(htmlGantt.getColor(color));
                    else
                        ganttElement.setColor(htmlGantt.getColor("LY_GANTT_INTERVAL_COLOR"));

                    if ((font != null) && (font.length() > 0))
                        ganttElement.setFont(htmlGantt.getFont(font));
                    else
                        ganttElement.setFont(htmlGantt.getFont("LY_GANTT_FONT"));

                    color = interval.getFontColor();

                    if ((color != null) && (color.length() > 0))
                        ganttElement.setLabelColor(htmlGantt.getColor(color));
                    else
                        ganttElement.setLabelColor(htmlGantt.getColor("LY_GANTT_FONT_COLOR"));

                    color = interval.getVerticalLinesColor();

                    if ((color != null) && (color.length() > 0))
                        ganttElement.setVerticalLinesColor(htmlGantt.getColor(color));

                    ganttElement.setId(interval.getId());
                    ganttElement.setLabel(translate(label));
                    ganttElement.setCenterLabel(translate(centerLabel));
                    ganttElement.setTooltip(translate(tooltip));
                    ganttElement.setDrawVerticalLines(interval.isDrawVerticalLines());
                    ganttElement.setVerticalLinesColor(htmlGantt.getColor(
                            interval.getVerticalLinesColor()));

                    // Compute dependencies
                    String dependencies = interval.getDependencies();
                    Vector<LyInterval> result = new Vector<LyInterval>();

                    if ((dependencies != null) && (dependencies.length() > 0))
                    {
                        String[] dependenciesArray = LyString.stringToList(dependencies, ' ');

                        if (dependenciesArray != null)
                        {
                            // For each dependency, retrieve the pointed
                            // interval and its Gantt element
                            // and add it to the list of dependencies
                            for (int k = 0; k < dependenciesArray.length; k++)
                            {
                                String     intervalId = dependenciesArray[k];
                                LyInterval intervalTmp = (LyInterval)gantt.getComponent(intervalId);

                                if (intervalTmp != null)
                                    result.addElement(intervalTmp);
                            }

                            ganttElement.setDependencies(result);
                        }
                    }

                    // Compute periods
                    int      j, m = interval.getPeriodCount();
                    LyPeriod period;
                    Vector<_LyPeriod>   periods = ((n == 0) ? null : new Vector<_LyPeriod>(n));

                    for (j = 0; j < m; j++)
                    {
                        period = interval.getPeriod(j);
                        periods.addElement(new _LyPeriod(
                                htmlInterval.parseDate(period.getBeginDate()),
                                htmlInterval.parseDate(period.getEndDate())));
                    }

                    ganttElement.setPeriods(periods);

					if (beginDate != null)
						ganttElement.draw(g, ganttCanvas, beginX, endX, bounds.y + scaleHeight,
								bounds.y + scaleHeight + bounds.height);
                }
            }
        }
    } // end method drawLine

    /**
     * Indicates if current line is a super line. (A super line is a parent line).
     *
     * @return true if current line is a super line, false otherwise
     */
    protected boolean isSuperLine()
    {
        int i, n = _line.getComponentCount();

        for (i = 0; i < n; i++)
        {
            LyComponent component = _line.getComponent(i);

            if (component instanceof LyLine)
                return true;
        }

        return false;
    }

    /**
     * Looks for selected children inside this line.
     *
     * @return true if at least one children is selected, false otherwise
     */
    public boolean hasChildSelected()
    {
        int n = _line.getComponentCount();

        if (n == 0)
            return false;

        LyComponent comp;
        LyStrutsLine webLine;
        boolean     result = false;
        int         i = 0;

        for (i = 0; (i < n) && !result; i++)
        {
            comp = _line.getComponent(i);

            if ((comp instanceof LyLine) && ((LyLine)comp).isSelected()
                    && !((LyLine)comp).isHidden())
                result = true;
            else if (comp instanceof LyLine)
            {
                webLine = (LyStrutsLine)comp.getInternalView();
                result = webLine.hasChildSelected();
            }
        }

        return result;
    } // end method hasChildSelected

    /**
     * Gets the limit date for this line (minimum or maximum). Browse all line sub-components and
     * retrieve their minimum (or maximum) dates.
     *
     * @param  max indicate if maximum or minimum date must be returned and if end date is null, we
     *             consider begin date instead.
     * @return the maximum date if max is true and if the end date is not null, the minimum date
     *         otherwise
     */
    public Date getLimitDate(boolean max)
    {
        Date date = null;
        Date dateTmp = null;
        int  i, n = _line.getComponentCount();

        // Browse all the sub-component
        for (i = 0; i < n; i++)
        {
            LyComponent component = _line.getComponent(i);

            dateTmp = null;

            if (component instanceof LyLine)
            {
                LyLine      line = (LyLine)component;
                LyStrutsLine internal = (LyStrutsLine)line.getInternalView();

                dateTmp = internal.getLimitDate(max);
            }
            else if (component instanceof LyInterval)
            {
                LyInterval      interval = (LyInterval)component;
                LyStrutsInterval internal = (LyStrutsInterval)interval.getInternalView();

                if (max)
                {
                    dateTmp = internal.getEndDate();

                    if (dateTmp == null)
                        // If end date is null, we consider the begin date
                        dateTmp = internal.getBeginDate();
                }
                else
                    dateTmp = internal.getBeginDate();
            }

            if (dateTmp == null)
                continue;

            if (date == null)
                date = dateTmp;
            else
            {
                if (max && dateTmp.after(date))
                    date = dateTmp;

                if (!max && dateTmp.before(date))
                    date = dateTmp;
            }

        }

        return date;
    } // end method getLimitDate

    /**
     * Gets the parent Gantt component of the line.
     *
     * @return the parent Gantt component
     */
    protected LyGantt getParentGantt()
    {
        LyComponent parent = _line.getParent();

        if (parent instanceof LyLine)
        {
            LyLine       line = (LyLine)parent;
            LyStrutsLine webLine = (LyStrutsLine)line.getInternalView();

            return webLine.getParentGantt();
        }
        else if (parent instanceof LyGantt)
            return (LyGantt)parent;

        return null;
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
     */
    public void setLabel(String label)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalLine#setTooltip(java.lang.String)
     */
    public void setTooltip(String tooltip)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetSelectedInterface#setSelected(boolean)
     */
    public void setSelected(boolean status)
    {
        // Manage Ajax mode
        LyWebViewManager viewManager = getWebViewManager();
        StringBuffer     buffer = new StringBuffer();
        String           color = null;
        String           textColor = null;

        if (status)
        {
            color = getHexColor("LY_TREENODE_SELECTED_BG_COLOR");
            textColor = getHexColor("LY_TREENODE_SELECTED_FONT_COLOR");
        }
        else
        {
            if (_line.getFontColor() != null)
                color = getHexColor(_line.getColor());

            if (color == null)
                color = getHexColor("LY_TREE_BACKGROUND");

            if (_line.getFontColor() != null)
                textColor = getHexColor(_line.getFontColor());

            if (textColor == null)
                textColor = getHexColor("LY_TREENODE_FONT_COLOR");
        }

        buffer.append("<script type='text/javascript'>\n//<![CDATA[\nSetTreeNodeColor('").append(
            getLineId()).append("_td','").append(getLineId()).append("_font','").append(color)
            .append("','").append(textColor).append("');\n//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(buffer.toString());
    } // end method setSelected

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetHiddenInterface#setHidden(boolean)
     */
    public void setHidden(boolean hidden)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetImageInterface#setImage(java.lang.String)
     */
    public void setImage(String key)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalLine#setOpen(boolean)
     */
    public void setOpen(boolean open)
    {
        // Manage Ajax mode
        LyWebViewManager viewManager = getWebViewManager();
        StringBuffer     buffer = new StringBuffer();
        String           treeNodeId = getLineId();

        // Update the div containing the children
        LyWebPageWriter  pw = new LyStrutsPage(_line, true);
        LyComponent      parent = _line.getParent();
        LyStrutsLine      htmlParentLine = null;

        if (parent instanceof LyLine)
        {
            LyLine parentLine = (LyLine)parent;

            htmlParentLine = (LyStrutsLine)parentLine.getInternalView();
        }
        else
            htmlParentLine = new LyStrutsLine(new LyLine());

        writeLine(pw, htmlParentLine, _isLast);

        StringBuffer innerHtml = pw.getBuffer(pw.getActiveName());
        String       innerHtmlStr = LyString.replace(innerHtml.toString(), "\n", "\\n");

        innerHtmlStr = LyString.replace(innerHtmlStr, "\"", "&quot;");

        buffer.append("<script type='text/javascript'>\n//<![CDATA[\nreplaceDivInnerHtml(\"")
            .append(_lastFrameName).append("\",\"").append(treeNodeId).append("_div\",\"").append(
            innerHtmlStr).append("\"").append(");\n//]]>\n</script>");

        // Compute new image for planning
        LyGantt      gantt = getParentGantt();
        LyStrutsGantt htmlGantt = (LyStrutsGantt)gantt.getInternalView();

        pw.free();

        pw = new LyStrutsPage(gantt, true);

        htmlGantt.writePlanning(pw);

        innerHtml = pw.getBuffer(pw.getActiveName());
        innerHtmlStr = LyString.replace(innerHtml.toString(), "\n", "\\n");
        innerHtmlStr = LyString.replace(innerHtmlStr, "\"", "&quot;");

        buffer.append("<script type='text/javascript'>\n//<![CDATA[\nreplaceDivInnerHtml(\"")
            .append(_lastFrameName).append("\",\"planning_").append(gantt.getIdCode()).append(
            "\",\"").append(innerHtmlStr).append("\"").append(");\n//]]>\n</script>");

        // Update the scroll if necessary
        if (open)
            buffer.append(
                "<script type='text/javascript'>\n//<![CDATA[\ncheckSizeMainView(this.parent);\n//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(buffer.toString());
        pw.free();
    } // end method setOpen

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetColorInterface#setColor(java.lang.String)
     */
    public void setColor(String color)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetFontInterface#setFont(java.lang.String)
     */
    public void setFont(String font)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalLine#setForegroundColor(java.lang.String)
     */
    public void setForegroundColor(String color)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalLine#addComponent(leon.view.LyComponent,
     *      int)
     */
    public void addComponent(LyComponent component, int pos)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalLine#removeComponent(leon.view.LyComponent)
     */
    public void removeComponent(LyComponent component)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalLine#setSuperLineColor(java.lang.String)
     */
    public void setSuperLineColor(String superLineColor)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalLine#setSuperLineLabel(java.lang.String)
     */
    public void setSuperLineLabel(String label)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalLine#setSuperLineHidden(boolean)
     */
    public void setSuperLineHidden(boolean superLineHidden)
    {
        refresh();
    }
} // end class LyStrutsLine
