/*
 * LyStrutsMap.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web.struts;

import java.awt.*;
import java.awt.image.*;
import java.io.*;
import java.util.*;

import leon.misc.*;
import leon.view.*;
import leon.view.LyMapBase.*;
import leon.view.common.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.web.*;
import leon.view.web.struts.map.*;


/**
 * This class implements the Map object (as described in the dtd) for Struts viewer.
 *
 * @version $Id: LyStrutsMap.java 36510 2012-08-17 15:43:06Z jlsaouli $
 * @author  Lyria
 * @see     leon.view.LyMap
 */
public class LyStrutsMap extends LyWebPageWriter
    implements LyGraphicalMap,
               LySetVisibleInterface,
               LyResetContentInterface,
               LySetIdleInterface,
               LySetPopupMenuInterface
{
	public static final String    __VERSION =
		"$Id: LyStrutsMap.java 36510 2012-08-17 15:43:06Z jlsaouli $";
	
	   /** The <b>hidden</b> String value used within frames names. */
    protected static final String HIDDEN = "hidden".intern();

    /** The <b>map_component</b> String value used within frames names. */
    protected static final String MAP_COMPONENT = "map_component".intern();

    /** Color used for the border of the viewer. */
    public static Color           VIEWER_BORDER_COLOR = Color.RED;

    /** Color used for the border of the displayed part in the viewer. */
    public static Color           VIEWER_DISPLAYED_PART_BORDER_COLOR = Color.GREEN;
    
    /** Number of steps available in the zoom widget (positive as well as negative). */
    public static int             ZOOM_STEPS_NB = 6;
    
    /** The map width. */
    public static int             MAP_WIDTH = 600;

    /** the map height. */
    public static int             MAP_HEIGHT = 550;

    /**
     * The maximum number of temporary files.
     * @deprecated
     */
    public static int             MAX_TEMP_FILES = 20;
        
    public static boolean		  NO_BACKGROUND = false;
    
    /** Force the map to keep the ten last map files and to destroy the oldest. */
    public static boolean		  CLEAR_MAP_FILES = false;

    /** Default frame. */
    protected static Frame        _frame;

    /* Static methods
    */

    /**
     * Gets the default frame.
     *
     * @return the default frame.
     */
    public static Frame getDefaultFrame()
    {
        return _frame;
    }

    /**
     * Gets the hexadecimal string representing the given color.
     *
     * @param  color the color to convert.
     * @return the hexadecimal code for the color.
     */
    public static String getHexString(Color color)
    {
        if (color == null)
            return "000000";

        return LyColors.getHexString(color.getRed(), color.getGreen(), color.getBlue());
    }

    /** The struts map box. */
    protected LyStrutsMapBox _mapbox;

    /** The map generic component. */
    protected LyMap         _map;

    /** Store the previous filename in order to remove it. */
    protected Vector<String>    _lastFileNames;

    /** Forced width of the map. */
    protected int           _mapWidth;

    /** Forced height of the map. */
    protected int           _mapHeight;
    
    /** Indicates if the background image must be refreshed. */
    protected boolean       _refreshBackgroundImage;
    
	/** Directions that may be used to translate the map component. */
	public enum TranslationDirection
	{
		CENTER,
		NORTH,
		EAST,
		SOUTH,
		WEST;
	}
	
	/** Possible positions of the viewer widget. */
	public enum ViewerPosition
	{
		NORTH_WEST,
		NORTH_EAST,
		SOUTH_EAST,
		SOUTH_WEST;
	}
	
	/** Idle mode of the map component. */
	protected boolean		_idle;
	
	/** Indicates if the viewer image must be refreshed. */
	protected boolean		_refreshViewerImage;
	
	/** Counter used to generate map images DIV. */
	protected int			_mapImagesCounter;
	
	/** Store the previous viewer filename in order to remove it. */
	protected Vector<String>	_lastViewerFileNames;

	/**
	 * Class constructor of a Map object for the struts target.
	 *
	 * @param map the generic Map defined for every type of view targets.
	 */
	public LyStrutsMap(LyMap map)
	{
        super(map);
        _map = map;
        _mapWidth = -1;
        _mapHeight = -1;
        _refreshBackgroundImage = false;

        init();
        setContent();
		
		_idle = false;
		_refreshViewerImage = false;
		_mapImagesCounter = 0;
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
	 */
	public void writeHtml(LyWebPageWriter pw)
	{
	    pw.startComponent(this);
	    
		updateRefreshInfo(pw);
		
		String controllerId = _map.getRootController();

		if (isFrameSet())
		{
			LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
			String			    pageUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_PAGE);
			StringBuffer		script = new StringBuffer();
			
			// Adjust the frame height to the message bar at the bottom of the view if there is one.
			LyView		view = _map.getParentView();
			LyMessages	messages = (view instanceof LyFrame ? ((LyFrame)view).getMessages() : null);

			// Call the RefreshDiv function to refresh the content of tabs.
			// Call it on the top frame to prevent from blocking the wait mechanism.
			script.append("getLeonardiTargetFrame().InitResizeMap('" + MAP_COMPONENT + controllerId + "', '" + getId() + "','"
					+ (messages == null ? "" : "messages_" + messages.getIdCode()) + "','" + pageUrl + "','false');");

			pw.append("<iframe frameborder='0' scrolling='no'");
			pw.append(" name='" + MAP_COMPONENT + controllerId + "'");
			pw.append(" onLoad=\"" + script.toString() + "\"");
			pw.append(" src='");
			pw.append(getWebViewManager().getScriptsDir() + "/blank.html");
			pw.append("' style='width:100%;height:100%;'></iframe>");
			
			getWebViewManager().addAdditionalRefreshScripts("<script type='text/javascript'>\n//<![CDATA[\n" +
					"getLeonardiTargetFrame().ReinitMap();\n//]]>\n</script>");
		}
		else
			writeMapFrame(pw);
		
		pw.endComponent(this);
	} // end method writeHtml
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebPageWriter#writeHtml(java.io.Writer, java.lang.String)
	 */
	public synchronized void writeHtml(Writer pw, String frame)
		throws IOException
	{
		String			controllerId = _map.getRootController();
		String			prefix = MAP_COMPONENT + controllerId;
		
		if ((frame != null) && (frame.indexOf(prefix) >= 0))
		{
			int    pos = frame.indexOf(prefix);
			int    pos2 = frame.indexOf("*");
			String width = frame.substring(pos + prefix.length(), pos2);
			String height = frame.substring(pos2 + 1, frame.length());

			_mapWidth = Math.max(20, Integer.parseInt(width));
			_mapHeight = Math.max(20, Integer.parseInt(height));

			frame = frame.substring(0, pos + prefix.length());

			LyStrutsPage		page = new LyStrutsPage(_map, true);

			page.setActiveBuffer(frame);
			page.setWriteDivOnly(false);
			page.setFrameName(frame);
			page.setWindowName(getWindowName());
			page.setFormId(_map.getRootController());
			
			writeMapFrame(page);

			StringBuffer		buffer = page.getBuffer(frame);
			LyWebViewManager	manager = getWebViewManager();
			String				scripts = manager.generateScriptsInclusion();

			pw.write("<html><head>\n");
			page.writeCssScripts(pw, frame);
			pw.write("</head>\n");
			
			LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
			String			    pageUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_PAGE);
			StringBuffer		script = new StringBuffer();
			
			// Adjust the frame height to the message bar at the bottom of the view if there is one.
			LyView		view = _map.getParentView();
			LyMessages	messages = (view instanceof LyFrame ? ((LyFrame)view).getMessages() : null);

			// Call the RefreshDiv function to refresh the content of tabs.
			// Call it on the top frame to prevent from blocking the wait mechanism.
			script.append("getLeonardiTargetFrame().ResizeMap('" + MAP_COMPONENT + controllerId + "', '" + getId() + "','"
					+ (messages == null ? "" : "messages_" + messages.getIdCode()) + "','" + pageUrl + "','false');");
			
			pw.write("<body scroll='no' topmargin='0' leftmargin='0' onresize=\"" + script + "\">\n");
			pw.write(scripts);
			page.writeBeginScripts(pw, frame);
			pw.write(page.getOpenFormContent(frame));
			pw.write(buffer.toString());
			pw.write("</form>\n");
			pw.write("</body>\n");
			page.writeEndScripts(pw, frame);
			pw.write("</html>\n");
			
			page.free();
		}
	}

	/**
	 * Write HTML description of the map frame.
	 *
	 * @param pw the page writer
	 */
	public void writeMapFrame(LyWebPageWriter pw)
	{
		updateRefreshInfo(pw);

		synchronized (this)
		{
			LyStrutsMapCanvas	mapCanvas = _mapbox.getMapCanvas();
			LyLayout			mapLayout = _map.getLayout();
			LyStrutsLayout		webLayout = (mapLayout == null ? null : (LyStrutsLayout)mapLayout.getInternalView());
			boolean				useTreemapLayout = (mapLayout != null) && (mapLayout.getStyle() == LyLayout.Style.TREEMAP);
			LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
			boolean				useGeographicalLayout = (mapLayout == null)
				|| (mapLayout.getStyle() == LyLayout.Style.GEOGRAPHIC) || (mapLayout.getStyle() == LyLayout.Style.COORDINATES);

			if (webLayout != null)
				webLayout.updateRefreshInfo(pw);
			
			// Use a special color to set the transparency of the background.
			// Replace filter doesn't seem to work with all the colors,
			// so we use a standard color (magenta).
			String hexColor = "#" + getHexString(mapCanvas.getBackground());

			pw.setBackgroundColor(hexColor, pw.getActiveName());

			Dimension	imageSize = mapCanvas.getSize();
			int			mapWidth = imageSize.width, mapHeight = imageSize.height;
			
			// Add internal CSS
			pw.addCssScript(".tooltip{" + getFontString("LY_TOOL_TIP_FONT") + "}", false);

			// Add script for rootMapUrl
			String		 controllerId = _map.getRootController();
			StringBuffer url = null;

			pw.addBeginScript("<script type='text/javascript'>\n//<![CDATA[\nvar _rootUrl='"
				+ getRootUrl() + "';\n//]]>\n</script>", false);

			pw.addBeginScript("<script type='text/javascript'>var _controllerId='" + controllerId
				+ "';</script>", false);

			pw.addBeginScript("<script type='text/javascript'>var mapWidth = " + mapWidth
				+ ";</script>", false);
			viewManager.addAdditionalRefreshScripts("<script type='text/javascript'>var mapWidth = " + mapWidth
				+ ";</script>");
			pw.addBeginScript("<script type='text/javascript'>var mapHeight = " + mapHeight
				+ ";</script>", false);
			viewManager.addAdditionalRefreshScripts("<script type='text/javascript'>var mapHeight = " + mapHeight
				+ ";</script>");

			if (_map.getSelectionMode() != LyMap.SelectionMode.NONE)
			{
				String	selectionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_SELECTEVENT);

				if (_map.getSelectionMode() == LyMap.SelectionMode.SINGLE)
					pw.addBeginScript(
						"<script type='text/javascript'>var selectionSingle=true;</script>", false);

				url = new StringBuffer();
				url.append("<script type='text/javascript'>\n//<![CDATA[\n");
				url.append("var selectedMapRegion;\n");
				url.append("function InvokeMapSelection(source, elementId, controllerId)\n");
				url.append("{\n");
				url.append("InvokeSelection(source, elementId, controllerId, 'false', '" + selectionUrl + "');");
				url.append("}\n");
				url.append("\n//]]>\n</script>\n");
				pw.addBeginScript(url.toString(), false);
			}

			if (useTreemapLayout)
			{
				LyStrutsTreemapLayout treemapLayout = (LyStrutsTreemapLayout)mapCanvas
						.getLayoutManager();

				url = new StringBuffer();
				url.append("<script>\n");

				// Limited to Single selection for treemaps
				url.append("var selectedRegions = new Array();\n");
				url.append("function TreemapSelection(regionId, select)\n");
				url.append("{\n");
				url.append("var doc = window.document;\n");
				url.append("var region = doc.getElementById(regionId);\n");
				url.append("var wasSelected = (region.style.borderWidth.substring(0, 1) != '0');\n");
				url.append("var newSelect = !wasSelected;\n");
				url.append("if (select == 'true')\nnewSelect = true;\n");
				url.append("if (select == 'false')\nnewSelect = false;\n");
				url.append("var selectString = (newSelect ? '" + LyWebConstants.SELECT_ITEM + "' : '"
						+ LyWebConstants.UNSELECT_ITEM + "');\n");

				url.append("if (selectedRegions.length > 0)\n");
				url.append("{\n");
				url.append("for (var i = 0; i < selectedRegions.length; i++)\n");
				url.append("{if (regionId == selectedRegions[i].id && select == 'true')\ncontinue;");
				url.append("if (regionId != selectedRegions[i].id && select == 'false')\ncontinue;");
				url.append("selectedRegions[i].style.borderWidth = 0;\n");
				
				if (viewManager.getBrowser() == LyWebViewManager.BROWSER_FIREFOX)
				{
					url.append("var oldWidth = selectedRegions[i].style.width;\n");
					url.append("var oldHeight = selectedRegions[i].style.height;\n");
					url.append("oldWidth = oldWidth.substring(i, oldWidth.length - 2);\n");
					url.append("oldHeight = oldHeight.substring(0, oldHeight.length - 2);\n");
					url.append("selectedRegions[i].style.width = oldWidth - (-"
						+ (2 * treemapLayout.getSelectBorderSize()) + ");\n");
					url.append("selectedRegions[i].style.height = oldHeight - (-"
						+ (2 * treemapLayout.getSelectBorderSize()) + ");\n");
				}
				
				url.append("}\n");
				url.append("}\n");

				url.append("var elt = doc.getElementById('_parentRegionTopSelection');\n");
				url.append("if (elt != null)\n");
				url.append("elt.style.backgroundColor = '';\n");

				url.append("elt = doc.getElementById('_parentRegionRihtSelection');\n");
				url.append("if (elt != null)\n");
				url.append("elt.style.backgroundColor = '';\n");

				url.append("elt = doc.getElementById('_parentRegionBottomSelection');\n");
				url.append("if (elt != null)\n");
				url.append("elt.style.backgroundColor = '';\n");

				url.append("elt = doc.getElementById('_parentRegionLeftSelection');\n");
				url.append("if (elt != null)\n");
				url.append("elt.style.backgroundColor = '';\n");

				url.append("if (newSelect)\n");
				url.append("{\n");
				
				if (viewManager.getBrowser() == LyWebViewManager.BROWSER_FIREFOX)
				{
					url.append("if (!wasSelected)\n");
					url.append("{\n");
					url.append("var oldWidth = region.style.width;\n");
					url.append("var oldHeight = region.style.height;\n");
					url.append("oldWidth = oldWidth.substring(0, oldWidth.length - 2);\n");
					url.append("oldHeight = oldHeight.substring(0, oldHeight.length - 2);\n");
					url.append("region.style.width = oldWidth - "
						+ (2 * treemapLayout.getSelectBorderSize()) + ";\n");
					url.append("region.style.height = oldHeight - "
						+ (2 * treemapLayout.getSelectBorderSize()) + ";\n");
					url.append("}\n");
				}
				
				url.append("region.style.borderWidth = " + treemapLayout.getSelectBorderSize()
					+ ";\n");
				url.append("selectedRegions = new Array();\n");
				url.append("selectedRegions.push(region);\n");
				url.append("}\n");

				url.append("else\n");
				url.append("{\n");
				
				if (viewManager.getBrowser() == LyWebViewManager.BROWSER_FIREFOX)
				{
					url.append("if (region.style.borderWidth.substring(0, 1) != '0')\n");
					url.append("{\n");
					url.append("var oldWidth = region.style.width;\n");
					url.append("var oldHeight = region.style.height;\n");
					url.append("if (endsWith(oldWidth, 'px')) oldWidth = oldWidth.substring(0, oldWidth.length - 2);\n");
					url.append("if (endsWith(oldHeight, 'px')) oldHeight = oldHeight.substring(0, oldHeight.length - 2);\n");
					url.append("region.style.width = oldWidth - (-"
						+ (2 * treemapLayout.getSelectBorderSize()) + ");\n");
					url.append("region.style.height = oldHeight - (-"
						+ (2 * treemapLayout.getSelectBorderSize()) + ");\n");
					url.append("}\n");
				}

				url.append("region.style.borderWidth = 0;\n");
				url.append("selectedRegions = new Array();\n");
				url.append("}\n");

				url.append("if ((select != 'true') && (select != 'false')) InvokeMapSelection(selectString, regionId, '" + _map.getRootController() + "');\n");
				url.append("}\n");
				url.append("</script>\n");
				pw.addBeginScript(url.toString(), false);
			}

			// get the url to the handlers
			String		selectionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_SELECTEVENT);
			String		actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
			String		mapUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_MAP);
			Dimension	dimension = getMapDimension();
			boolean		addBorder = ((getColor("LY_MAP_OUTLINE_COLOR") != null) && !NO_BACKGROUND && !useTreemapLayout);

			if (!pw.writeDivOnly() || (pw.getViewComponent() != _map))
			{
			String color = getHexColor("LY_MAP_BACKGROUND");

				pw.append("<center>\n");
			pw.append("<table ");
			
			if ((color != null) && !NO_BACKGROUND)
				pw.append("bgColor='" + color + "' ");
			
			pw.append("border='0' id='map_" + _map.getIdCode() + "' cellpadding='0' cellspacing='0'>\n");
			pw.append("<tr valign='middle'>\n");
			pw.append("<td valign='top' align='center'><table cellpadding='0' cellspacing='0'><tr><td>\n\n"); 

			pw.append("<div id='mapDiv_" + _map.getIdCode() + "' style='overflow:hidden;position:relative;z-index:0;");
			pw.append("width:" + (dimension.width + (addBorder ? 2 : 0)) + "px;height:" + (dimension.height + (addBorder ? 2 : 0)) + "px;");

			if (addBorder)
				pw.append("border:1px solid " + getHexColor("LY_MAP_OUTLINE_COLOR") + ";");
			
			pw.append("'>");

			if ((_map.getSelectionMode() == LyMap.SelectionMode.MULTIPLE) && !useTreemapLayout)
			{
				// Add divs for rubber banding
				pw.append("<div id='rubberband" + _map.getIdCode() + "' class='mapRubber'></div>\n");
				pw.append("<div id='rubberband" + _map.getIdCode() + "_moving' style='position:absolute;visibility:hidden;z-index:500;");
				pw.append("background-color:#FFFFFF;filter:alpha(opacity=0);opacity:0;'></div>\n\n");
			}
			
			if (useGeographicalLayout && _map.isZoom())
			{
				StringBuffer	endScripts = new StringBuffer();
				float			zoomStep = _map.getZoomStep();
				
				if (zoomStep <= 1)
					zoomStep = 2;
				
				float	zoomFactor = _map.getZoomFactor();
				boolean	rebuild = false;
				
				if (zoomFactor <= 0)
				{
					zoomFactor = getMapCanvas().getZoom();
					rebuild = true;
				}
				
				int		scale = getMapZoomScale();
				
				zoomFactor = (float)Math.pow(zoomStep, scale);
				
				if (rebuild)
					_map.setZoomFactor(zoomFactor);

				endScripts.append("<script type='text/javascript'>\n//<![CDATA[\n");
				endScripts.append("setCurrentScaleValue('" + _lastFrameName + "','" + getId() + "'," + scale + ");\n");
				endScripts.append("setTimeout(\"setMapZoomScale(getCurrentScaleValue('" + _lastFrameName + "','"
						+ getId() + "'), '', '', 'map_zoom" + _map.getIdCode() + "')\", 500);\n");
				
				if (viewManager.getBrowser() == LyWebViewManager.BROWSER_FIREFOX)
				{
					endScripts.append("setTimeout(\"addFirefoxWheelEvent('" + mapUrl + "', 'mapDiv_" + _map.getIdCode() +
						"', 'mapContent_" + _map.getIdCode() + "', 'map_zoom" + _map.getIdCode() + "', '" + getId() + "', '" + controllerId + "')\", 100);");
				}
				
				endScripts.append("\n//]]>\n</script>\n");
				
				pw.addEndScript(endScripts.toString(), false);
				viewManager.addAdditionalRefreshScripts(endScripts.toString());
			}
			
			mapCanvas.setSize(dimension.width, dimension.height);
			mapCanvas.setZoom(_map.getZoomFactor());
			mapCanvas.resetTranslation();

			LyStrutsLayoutManager layoutManager = mapCanvas.getLayoutManager();
			Rectangle			  boundingRect = mapCanvas.getBoundingRectangle();
			int					  left = (boundingRect.x + mapCanvas._scrollx);
			int					  top = (boundingRect.y + mapCanvas._scrolly);
			int					  width = boundingRect.width;
			int					  height = boundingRect.height;

			if ((layoutManager != null) && layoutManager.centerMapContent())
			{
				if ((width <= 0) || (height <= 0))
				{
					width = dimension.width;
					height = dimension.height;
				}
				
				left = (dimension.width - width) /2;
				top = (dimension.height - height) /2;
			}
			
			pw.append("<div id='mapContent_" + _map.getIdCode() + "' style='position:absolute;left:" +
					left + "px;top:" + top + "px;width:" +
					width + "px;height:" + height + "px;'");
			
			if ((_map.getSelectionMode() == LyMap.SelectionMode.MULTIPLE) && !useTreemapLayout)
			{
				pw.append(" onMouseOver=\"javascript:setRubberBandListener('rubberband" + _map.getIdCode() + "', true, 'map_" + _map.getIdCode() +
						"', 'mapDiv_" + _map.getIdCode() + "', 'mapContent_" + _map.getIdCode() + "', '" + _map.getRootController() + "', '" +
						getId() + "', '" + mapUrl + "', '" + selectionUrl + "');SetSelectionStartMode(false);\"");
				pw.append(" onMouseOut=\"javascript:setRubberBandListener('rubberband" + _map.getIdCode() + "', false);SetSelectionStartMode(true);\"");
			}

			if (_map.isTranslation())
			{
				String	moveCursorString = "<script type='text/javascript'>setUseRubber('" + _lastFrameName + "', '" + getId()
						+ "', " + isRubberDefaultMouseAction() + ");setMoveMapCursors('" + _lastFrameName
					+ "','url(" + getImageUrl("LY_MAP_MOVE_CURSOR") + "), default','url(" + getImageUrl("LY_MAP_MOVING_CURSOR") + "), default');</script>";
				
				pw.addEndScript(moveCursorString, false);
				getWebViewManager().addAdditionalRefreshScripts(moveCursorString);
			}
			else
				pw.addEndScript("<script type='text/javascript'>setUseRubber('" + _lastFrameName + "', '" + getId() + "', true);</script>", false);
			
			if (useGeographicalLayout && _map.isZoom())
			{
				pw.append(" onmousewheel=\"javascript:wheelMoved(event, '" + mapUrl + "', 'mapDiv_" + _map.getIdCode() +
						"', 'mapContent_" + _map.getIdCode() + "', 'map_zoom" + _map.getIdCode() + "', '" + getId() + "', '" + controllerId + "');\"");
			}				

			if ((_map.getSelectionMode() != LyMap.SelectionMode.NONE) && !useTreemapLayout)
			{
				pw.append(
					" onClick='javascript:InvokeBackgroundClick(\"map_" + _map.getIdCode() + "\", \"" +
					_map.getRootController() + "\", \"" + selectionUrl + "\");'");
			}
			
			pw.append(" ondblClick='javascript:InvokeBackgroundDblCLick(\"" + _map.getId() + "\", \"" +
					_map.getRootController() + "\", \"" + actionUrl + "\");' >\n\n");
			}

			if (!NO_BACKGROUND)
			{
				String	mieCSS = ".MIE {position:absolute;}";
				String	miCSS = ".MI {position:absolute;}";

				pw.addCssScript(mieCSS, false);
				pw.addCssScript(miCSS, false);
				pw.append("<style type=\"text/css\">\n<!--\n" + mieCSS + "\n" + miCSS + "\n-->\n</style>\n");
				pw.append("\n\n<div style='position:absolute;left:0px;top:0px;z-index:0;' id='mapImages_" + _map.getIdCode() + "_" +
						_mapImagesCounter + "'>\n" + generateMapImages() + "\n</div>\n\n");
				
				// Needed for usemap
				pw.append("<img src='about:blank' border='0' usemap=\"#" + _map.getId()
						+ "\" id=\"" + _map.getId() + "_background_overlay\"" +
						" style='position:absolute;left:0px;top:0px;width:100%;height:100%;" +
						"filter:alpha(opacity=0);opacity:0;z-index:1;'/>");
			}

			// Write Html for all element...
			Vector<LyComponent>	elements = getMapElements();
			LyComponent			mapElement;
			LyStrutsMapElement	internal;
			int					i, count = elements.size();

			for (i = 0; i < count; i++)
			{
				mapElement = elements.elementAt(i);

				internal = (LyStrutsMapElement)mapElement.getInternalView();

				if (mapCanvas.needDrawing(internal) || internal.isHidden())
					continue; // This one is drawn on background or hidden

				internal.drawHtml(mapCanvas, pw);
			}

			if (useTreemapLayout)
			{
				// Draw border for selection
				LyStrutsTreemapLayout treemapLayout = (LyStrutsTreemapLayout)mapCanvas
						.getLayoutManager();
				int					 borderWidth = treemapLayout.getSelectBorderSize();
				String				 selectedColor = getHexColor("LY_MAP_SELECTED_COLOR");

				pw.append(
					"<div id='_select_border_up' style='visibility:hidden;position:absolute;z-index:2;"
					+ ";background-color:" + selectedColor + "'/>\n");
				pw.append(
					"<div id='_select_border_right' style='visibility:hidden;position:absolute;z-index:2;"
					+ ";background-color:" + selectedColor + "'/>\n");
				pw.append(
					"<div id='_select_border_down' style='visibility:hidden;position:absolute;z-index:2;"
					+ ";background-color:" + selectedColor + "'/>\n");
				pw.append(
					"<div id='_select_border_left' style='visibility:hidden;position:absolute;z-index:2;"
					+ ";background-color:" + selectedColor + "'/>\n");
				// Next div is needed when height is too small
				pw.append("<div/>\n");

				pw.addBeginScript("<script type='text/javascript'>var selectBorderWidth = "
					+ borderWidth + ";</script>", false);
			}

			if (!pw.writeDivOnly() || (pw.getViewComponent() != _map))
			{
			pw.append("</div>\n");

			if (useGeographicalLayout)
				generateMapWidgets(pw);

			pw.append("</div>\n\n");

			// Add map
			pw.append("<map name='" + _map.getId() + "' id='" + _map.getId() + "'>\n");
			pw.append(getUseMapContent());
			pw.append("</map>\n");
				pw.append("</td></tr></table></td></tr>");
				pw.append("</table>\n");
			pw.append("</center>\n");
		}
		}
	} // end method writeMapFrame

	/**
	 * Indicates whether rubber mode is the default mode when map is opened and when translation and
	 * multi-selection are available.
	 * @return true if rubber mode is the default mode, false otherwise
	 */
	protected boolean isRubberDefaultMouseAction()
	{
		return false;
	}
	
	/**
	 * Return the list of elements contained by the map.
	 */
	protected Vector<LyComponent> getMapElements()
	{
		int						i, count = _map.getComponentCount();
		LyComponent				mapElement;
		LyStrutsMapElement		internal;
		Vector<LyComponent>		elements = new Vector<LyComponent>(count);
		LyMapContent			mapContent = _map.getMapContent(); ;

		for (i = 0; i < count; i++)
		{
			mapElement = _map.getComponent(i);

			// XXX : dead code, LyStrutsMapRegion.USE_DIV_FOR_TREEMAP_LAYOUT is always true, not sure what to do
			/*if (!LyStrutsMapRegion.USE_DIV_FOR_TREEMAP_LAYOUT
					&& (mapElement instanceof LyMapRegion)
					&& (((LyMapRegion)mapElement).getMapRegionCount() > 0))
				computeDisplayedMapRegions((LyMapRegion)mapElement, elements);
			else*/
				elements.addElement(mapElement);
		}

		if (mapContent != null)
		{
			int nContent = mapContent.getComponentCount();

			for (i = 0; i < nContent; i++)
			{
				mapElement = mapContent.getComponent(i);
				elements.addElement(mapElement);
			}
		}

		// Sort elements by their layer
		// First pass to set existing layers
		int    				layer = 0;
		Vector<LyComponent>	sortedElements = new Vector<LyComponent>(count);
		BitSet				existingLayers = new BitSet();

		count = elements.size();

		for (i = 0; i < count; i++)
		{
			mapElement = elements.get(i);
			internal = (LyStrutsMapElement)mapElement.getInternalView();

			if (internal.getLayer() > 0)
				existingLayers.set(internal.getLayer());
		}

		if (elements.size() != 1)
		{
			// Now we sort the content
			for (layer = 0; elements.size() > 0; layer++)
			{
				if (!existingLayers.get(layer))
					continue; // Layer is empty
                
				Vector<LyComponent> toRemove = new Vector<LyComponent>();

				count = elements.size();

				for (i = 0; i < count; i++)
				{
					mapElement = elements.get(i);
					internal = (LyStrutsMapElement)mapElement.getInternalView();

					if (internal.getLayer() <= layer)
					{
						sortedElements.add(mapElement);
						toRemove.add(mapElement);
					}
				}
				for (i = 0; i < toRemove.size(); i++)
					elements.remove(toRemove.get(i));
			}
			elements = sortedElements;
			count = elements.size();
		}
		
		return elements;
	}
	
	/**
	 * Returns the HTML code that must be inserted in the <code>map</code> tag.
	 */
	protected String getUseMapContent()
	{
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String				selectionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_SELECTEVENT);
		String				actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		LyStrutsMapCanvas	mapCanvas = _mapbox.getMapCanvas();

		mapCanvas.setSize(getMapDimension());
		mapCanvas.setZoom(_map.getZoomFactor());
		mapCanvas.resetTranslation();
		
		Vector<LyComponent>	elements = getMapElements();
		int					i, count = elements.size();
		Rectangle			rect = null;
		StringBuffer		pw = new StringBuffer(), url;
		LyComponent			mapElement;
		LyStrutsMapElement	internal;
		
		for (i = 0; i < count; i++)
		{
			mapElement = elements.elementAt(i);
			internal = (LyStrutsMapElement)mapElement.getInternalView();

			if ((!internal.isSensitive()) || (internal.getId() == null))
				continue;
			if (internal instanceof LyStrutsMapNode)
				continue;

			url = new StringBuffer();

			url.append("&#38;");
			url.append(LyWebConstants.VALUE);
			url.append("=");
			url.append(internal.getId());

			pw.append("<area ondblclick=\"javascript:AreaDblClick('" + LyString.getJavascriptString(internal.getId()) + "','" + _map.getId() + "','" +
					_map.getRootController() + "','" + actionUrl + "');\"");

			if (internal.getTooltip() != null)
				pw.append(" title='" + internal.getTooltip() + "'");

			if (internal instanceof LyStrutsMapLink)
			{
				Polygon poly = ((LyStrutsMapLink)internal).getBoundingShape(mapCanvas);

				if ((_map.getSelectionMode() != LyMap.SelectionMode.NONE))
				{
					url = new StringBuffer();
					url.append("\"javascript:InvokeMapClick('");
					url.append(LyString.getJavascriptString(internal.getId()));
					url.append("','");
					url.append(_map.getRootController());
					url.append("','map_");
					url.append(_map.getIdCode());
					url.append("','");
					url.append(selectionUrl);
					url.append("',event);\"");

					pw.append(" onclick=" + url.toString() + " onmousedown=\"setRubberBandListener('rubberband" + _map.getIdCode() + "',false);\""
							+ " onmouseup=\"setRubberBandListener('rubberband" + _map.getIdCode() + "',true);\"'");
				}

				pw.append(" shape='poly' coords='");

				int[] xpoints = poly.xpoints;
				int[] ypoints = poly.ypoints;

				for (int j = 0; j < poly.npoints; j++)
				{
					if (j != 0)
						pw.append(",");

					pw.append(xpoints[j] + "," + ypoints[j]);
				}
			}
			else if (internal instanceof LyStrutsMapRegion)
			{
				Polygon poly = ((LyStrutsMapRegion)internal).getBoundingShape(mapCanvas);

				if (internal.isSensitive())
				{
					boolean extend = true;

					if (_map.getSelectionMode() != LyMap.SelectionMode.NONE)
					{
						if (_map.getSelectionMode() == LyMap.SelectionMode.SINGLE)
							extend = false;

						pw.append(" onClick=\"javascript:InvokeMapClick(this,'");
						pw.append("&#38;" + LyWebConstants.SOURCE + "="
							+ LyWebConstants.SELECT_ITEM);
						pw.append("','");
						pw.append("&#38;" + LyWebConstants.SOURCE + "="
							+ LyWebConstants.UNSELECT_ITEM);
						pw.append("','");
						pw.append("&#38;" + LyWebConstants.EXTEND + "=" + extend);
						pw.append("','");
						pw.append("&#38;" + LyWebConstants.VALUE + "="
							+ LyWebObject.encodeUrl(internal.getId()));
						pw.append("');\"");
					}
				}

				pw.append(" shape='poly' coords='");

				int[] xpoints = poly.xpoints;
				int[] ypoints = poly.ypoints;

				for (int j = 0; j < poly.npoints; j++)
				{
					if (j != 0)
						pw.append(",");

					pw.append(xpoints[j] + "," + ypoints[j]);
				}
			}
			else
			{
				rect = internal.getBoundingRectangle(mapCanvas);
				pw.append(" shape='rect' coords='" + rect.x + "," + rect.y + ","
					+ (rect.x + rect.width) + "," + (rect.y + rect.height));
			}

			// Close coordinates attribute.
			pw.append("'/>\n");
		}
		
		return pw.toString();
	}
	
	public String generateMapImage(int xoffset, int yoffset, int width, int height, boolean viewer)
	{
		synchronized (this)
		{
			LyStrutsMapCanvas	mapCanvas = _mapbox.getMapCanvas();
			Ly3DBox				boundingBox = mapCanvas.getBoundingBox();
			int					bbWidth = (int)(boundingBox.getEndPoint().getX() - boundingBox.getStartPoint().getX());
			int					bbHeight = (int)(boundingBox.getEndPoint().getY() - boundingBox.getStartPoint().getY());
			LyLayout			mapLayout = _map.getLayout();
			boolean				useTreemapLayout = (mapLayout != null) && (mapLayout.getStyle() == LyLayout.Style.TREEMAP);
			
			if (useTreemapLayout)
			{
				Rectangle	r = mapCanvas.getBoundingRectangle();
				
				mapCanvas.setSize(r.width, r.height);
			}
			else
				mapCanvas.setSize(bbWidth, bbHeight);
			
			mapCanvas.setZoom(1);
			mapCanvas.resetTranslation();
			mapCanvas.translate(-bbWidth / 2, -bbHeight / 2);
			
			float			zoom = _map.getZoomFactor();
			int				subImageWidth = (viewer ? LyStrutsMapCanvas.VIEWER_SIZE : width);
			int				subImageHeight = (viewer ? LyStrutsMapCanvas.VIEWER_SIZE : height);
			String			mapId = _map.getId();
			String			fileName = "map_" + mapId + (viewer ? "_viewer_" : "_" + xoffset + "_" + yoffset + "_") + System.currentTimeMillis() + ".jpg";
			BufferedImage	image = new BufferedImage(subImageWidth, subImageHeight, BufferedImage.TYPE_INT_BGR);
			Graphics		g = image.getGraphics();
	
			if (!viewer)
			{
				mapCanvas.setZoom(zoom);
				
				float	newZoom = (float)(zoom - 0.5);
				float	tx = bbWidth * newZoom;
				float	ty = bbHeight * newZoom;
				
				Rectangle	boundingRectangle = mapCanvas.getBoundingRectangle();
				
				tx = -boundingRectangle.x;
				ty = -boundingRectangle.y;
				
				mapCanvas.translate((int)tx, (int)ty);
				mapCanvas.setSize(subImageWidth, subImageHeight);
				mapCanvas.translate(-xoffset, -yoffset);
				mapCanvas.paint(g);
			}
			else
			{
				mapCanvas.setSize(getMapDimension());
				mapCanvas.setZoom(_map.getZoomFactor());
				mapCanvas.resetTranslation();
				mapCanvas.paintViewer(image.getGraphics());
			}
			
	        LyViewManager viewManager = getViewManager();
	        String        tmpDir = viewManager.getTemporaryDir();
	        String        tmpRoot = viewManager.getTemporaryLocation();
			String	      subImageFile = tmpDir + "/" + fileName;
			String	      subImagePath = tmpRoot + "/" + fileName;
	
//			if (!viewer)
//			{
//				g.setColor(Color.BLACK);
//				g.setFont(new Font("Arial", 0, 20));
//				g.drawRect(0, 0, subImageWidth - 1, subImageHeight - 1);
//				g.drawString("Image " + xoffset + " / " + yoffset, subImageWidth / 2 - 60, subImageHeight / 2);
//			}

			if (viewer)
				_lastViewerFileNames.add(subImageFile);
			else
			_lastFileNames.add(subImageFile);
			
			LyImageUtils.encodeImage(image, tmpDir, subImageFile);
			image.flush();
			image = null;
			
			File   temp = new File(subImageFile);

			if (temp.exists())
				temp.deleteOnExit();

			return subImagePath;
		}
	}

	protected String generateMapImages()
	{
		synchronized (this)
		{
			// First clean tmp directory
			clearMapFiles(false);
			
			StringBuffer		result = new StringBuffer();
			boolean				generateImage;
			int					x = 0, y = 0;
			LyStrutsMapCanvas	mapCanvas = getMapCanvas();
			Rectangle			boundingRect = mapCanvas.getBoundingRectangle();
			Dimension			dimension = getMapDimension();
			int					boundingRectWidth = boundingRect.width;
			int					boundingRectHeight = boundingRect.height;
			int					step = _map.getTranslationStep();
			int					defaultImageWidth = step, defaultImageHeight = step;
			
			if (step <= 0)
			{
				defaultImageWidth = dimension.width;
				defaultImageHeight = dimension.height;
			}
	
			for (y = 0; y < boundingRectHeight; y += defaultImageHeight)
			{
				for (x = 0; x < boundingRectWidth; x += defaultImageWidth)
				{
					int		imageWidth = defaultImageWidth;
					int		imageHeight = defaultImageHeight;
					
					generateImage = true;
					
					if (x + imageWidth > boundingRectWidth)
						imageWidth = boundingRectWidth - x;
					
					if (y + imageHeight > boundingRectHeight)
						imageHeight = boundingRectHeight - y;
					
					if ((x + imageWidth < -boundingRect.x - mapCanvas._scrollx) || (x > (-boundingRect.x - mapCanvas._scrollx + dimension.width)))
						generateImage = false;
					
					if ((y + imageHeight < -boundingRect.y - mapCanvas._scrolly) || (y > (-boundingRect.y - mapCanvas._scrolly + dimension.height)))
						generateImage = false;
					
					result.append("<div id='MI" + _mapImagesCounter + "_" + x + "_" + y + "_" + imageWidth + "_" + imageHeight + "' class='" + (generateImage ? "MI" : "MIE") +
							"' style='left:" + x + "px;top:" + y + "px;");
					
	//				if (imageWidth != defaultImageWidth)
						result.append("width:" + imageWidth + "px;");
					
	//				if (imageHeight != defaultImageHeight)
						result.append("height:" + imageHeight + "px;");
	
					result.append("'>");
					
					if (generateImage)
					{
						String	mapImage = generateMapImage(x, y, imageWidth, imageHeight, false);
						
						result.append("<img src='" + mapImage + "' />");
					}
					
					result.append("</div>");
				}
			}
			
			return result.toString();
		}
	}

	protected void clearMapFiles()
	{
		clearMapFiles(true);
		clearMapFiles(false);
	}
	
	protected void clearMapFiles(boolean viewer)
	{
		// If the current mode is clearing previous map file
		if (!CLEAR_MAP_FILES)
			return;
		
		Vector<String>	fileNames = (viewer ? _lastViewerFileNames : _lastFileNames);
		int				i, n = fileNames.size();
		
		for (i = n - 1; i >= 0; i--)
		{
			// Remove previous map files...
			String	path = fileNames.elementAt(i);
			File	temp = new File(path);

			if (temp.exists())
				temp.delete();
		}

		fileNames.clear();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyFreeInterface#free()
	 */
	public synchronized void free()
	{
		super.free();
		clearMapFiles();
	}

	/**
	 * Generate the HTML code for map widgets (viewer, translation, zoom).
	 * 
	 * @param pw the page writer
	 */
	protected void generateMapWidgets(LyWebPageWriter pw)
	{
		String	collapseImage = getImageUrl("LY_MAP_COLLAPSE_WIDGET");
		String	expandImage = getImageUrl("LY_MAP_EXPAND_WIDGET");
		String	controllerId = _map.getRootController();
		String	actionUrl = ((LyStrutsViewManager)getWebViewManager()).getActionURL(LyStrutsConstants.PATH_ACTION_MAP);
		boolean	displaySwitchImages = false;
		int		widgetLeftOffset = (displaySwitchImages ? 5 : 0);
		int		widgetTopOffset = (displaySwitchImages ? 5 : 0);
		ViewerPosition	viewerDefaultPosition = getViewerDefaultPosition();

		if (_map.isViewer())
		{
			// Compute the viewer position
			int				viewerX = _map.getViewerX();
			int				viewerY = _map.getViewerY();
			
			if (viewerX < 0)
			{
				switch (viewerDefaultPosition)
				{
				case NORTH_EAST :
				case SOUTH_EAST :
				{
					Dimension	dimension = getMapDimension();

					viewerX = dimension.width - LyStrutsMapCanvas.VIEWER_SIZE - 2;
					break;
				}
				case NORTH_WEST :
				case SOUTH_WEST :
				default:
					viewerX = 2;
				}
			}
			
			if (viewerY < 0)
			{
				switch (viewerDefaultPosition)
				{
				case SOUTH_WEST :
				case SOUTH_EAST :
				{
					Dimension	dimension = getMapDimension();

					viewerY = dimension.height - LyStrutsMapCanvas.VIEWER_SIZE - 2;
					break;
				}
				case NORTH_EAST :
				case NORTH_WEST :
				default:
					viewerY = 2;
				}
			}

			String	expandViewer = translate("LY_MAP_EXPAND_VIEWER");
			String	collapseViewer = translate("LY_MAP_COLLAPSE_VIEWER");
			boolean	viewerVisible = (_map.getViewerStatus() == LyMap.ViewerStatus.SHOW);
			String	viewerSwitchImage = (viewerVisible ? collapseImage : expandImage);
			String	viewerSwitchTooltip = (viewerVisible ? collapseViewer : expandViewer);
			
			clearMapFiles(true);
			
			pw.append("<div id='map_viewer" + _map.getIdCode() + "' style='position:absolute;left:" + viewerX +"px;top:" + viewerY +"px;z-index:50;'\n");
			pw.append(" onMouseOver=\"javascript:prepareMovingViewer('mapDiv_" + _map.getIdCode() + "', 'map_viewer" + _map.getIdCode() +
					"', 'map_viewer_translate" + _map.getIdCode() + "', 'map_viewer_moving" + _map.getIdCode() + "');\">\n");
			
			if (displaySwitchImages)
			{
				pw.append("<div id='map_viewer_switch" + _map.getIdCode() + "' style='position:absolute;left:" + (viewerX + 2) + "px;top:" + (viewerY + 2) + "px;z-index:100;'>");
				pw.append("<a href=\"javascript:switchMapWidgetState('displayViewer', '" + actionUrl + "', '" + getId() + "', 'map_viewer_img" + _map.getIdCode() + "', '" + controllerId + "');\">");
				pw.append("<img id='map_viewer_switch_img' src='" + viewerSwitchImage + "' border='0' title=\"" + viewerSwitchTooltip + "\"/></a></div>\n");
			}
			
			pw.append("<div id='map_viewer_img" + _map.getIdCode() + "' style='position:relative;left:0px;top:0px;z-index:1;");
			
			if (!viewerVisible)
				pw.append("display:none;");
			
			pw.append("'>");
			generateViewerWidget(pw);
			pw.append("</div>\n");
			
			pw.append("<div id='map_viewer_moving" + _map.getIdCode() + "'\n");
			pw.append("	style='cursor:move;z-index:50;background-color:#000000;filter:alpha(opacity=0);opacity:0;-moz-opacity:0;position:absolute;");
			
			if (!viewerVisible)
				pw.append("display:none;");
			
			pw.append("'\n onMouseDown=\"startMovingViewer('mapDiv_" + _map.getIdCode() + "', 'map_viewer_translate" + _map.getIdCode() +
					"', 'map_viewer_moving" + _map.getIdCode() + "');\"\n");
			pw.append("	onMouseUp=\"endMovingViewer('mapDiv_" + _map.getIdCode() + "', 'map_viewer" + _map.getIdCode() + "', 'map_viewer_translate" + _map.getIdCode() +
					"', 'map_viewer_moving" + _map.getIdCode() + "', '" + actionUrl + "', '" + getId() + "', '" + controllerId + "')\"\n");
			pw.append("	onMouseMove=\"moveViewer(event, 'mapDiv_" + _map.getIdCode() + "', 'map_viewer" + _map.getIdCode() +
					"', 'map_viewer_img" + _map.getIdCode() + "', 'map_viewer_switch" + _map.getIdCode() + "')\"\n");
			pw.append("	onMouseOut=\"endMovingViewer('mapDiv_" + _map.getIdCode() + "', 'map_viewer" + _map.getIdCode() + "', 'map_viewer_translate" + _map.getIdCode() +
					"', 'map_viewer_moving" + _map.getIdCode() + "', '" + actionUrl + "', '" + getId() + "', '" + controllerId + "');\"></div>\n");

			pw.append("<div id='map_viewer_translate" + _map.getIdCode() + "'\n");
			pw.append("	style='cursor:hand;cursor:pointer;z-index:60;background-color:#000000;filter:alpha(opacity=0);opacity:0;-moz-opacity:0;position:absolute;");
			
			if (!viewerVisible)
				pw.append("display:none;");
			
			pw.append("'\n onMouseDown=\"translateViewer(event, '" + actionUrl + "', '" + getId() + "', 'map_viewer" + _map.getIdCode() + "', '" + controllerId + "');\"></div>\n");
		}
		
		if (_map.isTranslation())
		{
			String	expandTranslation = translate("LY_MAP_EXPAND_TRANSLATION");
			String	collapseTranslation = translate("LY_MAP_COLLAPSE_TRANSLATION");
			boolean	translationVisible = (_map.getTranslationStatus() == LyMap.TranslationStatus.SHOW);
			String	translationSwitchImage = (translationVisible ? collapseImage : expandImage);
			String	translationSwitchTooltip = (translationVisible ? collapseTranslation : expandTranslation);
			int		top = (_map.isViewer() && (viewerDefaultPosition == ViewerPosition.NORTH_WEST) ?
					LyStrutsMapCanvas.VIEWER_SIZE + 3 : 0) + widgetTopOffset;

			if (displaySwitchImages)
			{
				pw.append("<div id='map_translation_switch" + _map.getIdCode() + "' style='position:absolute;left:5px;top:" + top + "px;z-index:100;'>\n");
				pw.append("<a href=\"javascript:switchMapWidgetState('displayTranslation', '" + actionUrl + "', '" + getId() + "', 'map_translation" + _map.getIdCode() + "', '" + controllerId + "');\">");
				pw.append("<img id='map_translation_switch_img" + _map.getIdCode() + "' src='" + translationSwitchImage + "' border='0' title=\"" + translationSwitchTooltip + "\"/></a>\n");
				pw.append("</div>\n");
			}
			
			pw.append("<div id='map_translation" + _map.getIdCode() + "' style='position:absolute;left:" + widgetLeftOffset + "px;top:" + (top + 2) + "px;z-index:50;");
			
			if (!translationVisible)
				pw.append("display:none;");
			
			pw.append("'>\n");
			generateTranslationWidget(pw);
			pw.append("</div>\n");
			
			if (_map.getSelectionMode() == LyMap.SelectionMode.MULTIPLE)
			{
				pw.append("<div id='map_selection_mode" + _map.getIdCode() + "' style='position:absolute;left:" + widgetLeftOffset + "px;top:" + (top + 50) +
						"px;z-index:50;");
				
				if (!translationVisible)
					pw.append("display:none;");
				
				pw.append("'>\n");
				generateMouseModeWidget(pw);
				pw.append("</div>\n");
			}
		}
		
		if (_map.isZoom())
		{
			String	expandZoom = translate("LY_MAP_EXPAND_ZOOM");
			String	collapseZoom = translate("LY_MAP_COLLAPSE_ZOOM");
			boolean	zoomVisible = (_map.getZoomStatus() == LyMap.ZoomStatus.SHOW);
			String	zoomSwitchImage = (zoomVisible ? collapseImage : expandImage);
			String	zoomSwitchTooltip = (zoomVisible ? collapseZoom : expandZoom);
			int		top = (_map.isViewer() && (viewerDefaultPosition == ViewerPosition.NORTH_WEST) ?
					LyStrutsMapCanvas.VIEWER_SIZE + 3 : 0) + (_map.isTranslation() ? 45 : 0)
					+ (_map.isTranslation() && _map.getSelectionMode() == LyMap.SelectionMode.MULTIPLE ? 28 : 0) + (widgetTopOffset * 2);

			if (displaySwitchImages)
			{
				pw.append("<div id='map_zoom" + _map.getIdCode() + "_switch' style='position:absolute;left:5px;top:" + top + "px;z-index:100;'>\n");
				pw.append("<a href=\"javascript:switchMapWidgetState('displayZoom', '" + actionUrl + "', '" + getId() + "', 'map_zoom" + _map.getIdCode() + "', '" + controllerId + "');");
				pw.append("setMapZoomScale(currentScale['" + getId() + "'], '', '', 'map_zoom" + _map.getIdCode() + "');\">\n");
				pw.append("<img id='map_zoom_switch_img" + _map.getIdCode() + "' src='" + zoomSwitchImage + "' border='0' title=\"" + zoomSwitchTooltip + "\"/></a>\n");
				pw.append("</div>\n");
			}
			
			pw.append("<div id='map_zoom" + _map.getIdCode() + "' style='position:absolute;left:" + widgetLeftOffset + "px;top:" + (top + 7) + "px;z-index:50;");
			
			if (!zoomVisible)
				pw.append("display:none;");
			
			pw.append("'>\n");
			generateZoomWidget(pw);
			pw.append("</div>\n");

			pw.append("<div id='map_zoom" + _map.getIdCode() + "_cursor_moving'\n");
			pw.append("style='cursor:hand;cursor:pointer;z-index:100;background-color:#000000;filter:alpha(opacity=0);opacity:0;-moz-opacity:0;visibility:hidden;position:absolute;'\n");
			pw.append("onMouseDown=\"startMovingZoom('mapDiv_" + _map.getIdCode() + "', 'map_zoom" + _map.getIdCode() + "');\"\n");
			pw.append("onMouseUp=\"endMovingZoom('" + actionUrl + "', '" + getId() + "', 'map_zoom" + _map.getIdCode() + "', " + -ZOOM_STEPS_NB + ", '" + controllerId + "')\"\n");
			pw.append("onMouseMove=\"moveZoom(event, 'mapDiv_" + _map.getIdCode() + "', 'map_zoom" + _map.getIdCode() + "')\">\n");
			pw.append("<table><tr><td></td></tr></table></div>\n");
		}
	}
	
	/**
	 * Generate the HTML code for viewer widget.
	 * 
	 * @param pw the page writer
	 */
	protected void generateViewerWidget(LyWebPageWriter pw)
	{
		String	viewerImagePath = generateMapImage(-1, -1, -1, -1, true);

		pw.append("<img id='" + _map.getId() + "_viewer_image' src='" + viewerImagePath + "' style='z-index:1;border:1px solid #" +
				getHexString(VIEWER_BORDER_COLOR) + ";'/></div>\n");
	}

	/**
	 * Return the default position of the viewer widget.
	 * 
	 * @return default implementation returns ViewerPosition.NORTH_WEST.
	 */
	protected ViewerPosition getViewerDefaultPosition()
	{
		return ViewerPosition.NORTH_WEST;
	}

	/**
	 * Generate the HTML code for translation widget.
	 * 
	 * @param pw the page writer
	 */
	protected void generateTranslationWidget(LyWebPageWriter pw)
	{
		String	controllerId = _map.getRootController();
		String	actionUrl = ((LyStrutsViewManager)getWebViewManager()).getActionURL(LyStrutsConstants.PATH_ACTION_MAP);
		String	translationUp = translate("LY_MAP_TRANSLATION_UP");
		String	translationDown = translate("LY_MAP_TRANSLATION_DOWN");
		String	translationLeft = translate("LY_MAP_TRANSLATION_LEFT");
		String	translationRight = translate("LY_MAP_TRANSLATION_RIGHT");
		String	translationCenter = translate("LY_MAP_TRANSLATION_CENTER");
		String	translateUpImage = getImageUrl("LY_MAP_TRANSLATE_UP");
		String	translateDownImage = getImageUrl("LY_MAP_TRANSLATE_DOWN");
		String	translateLeftImage = getImageUrl("LY_MAP_TRANSLATE_LEFT");
		String	translateRightImage = getImageUrl("LY_MAP_TRANSLATE_RIGHT");
			String	emptyImage = getImageUrl("LY_EMPTY_IMAGE");
		String	resetImage = getImageUrl("LY_MAP_RESET_WIDGET");
			
		pw.append("<table border='0' cellspacing='0' cellpadding='0' align='center'>\n");
			pw.append("<tr>\n");
			pw.append("<td></td>\n");
			pw.append("<td align='center'><a href=\"javascript:translateMap('" + actionUrl + "', '" + TranslationDirection.NORTH + "', '" +
					getId() + "', 'mapDiv_" + _map.getIdCode() + "', 'mapContent_" + _map.getIdCode() + "', '" + _map.getId() + "', '" +
					emptyImage + "', '" + controllerId + "');\"><img src='" + translateUpImage + "' border='0' title=\"" + translationUp + "\"/></a></td>\n");
			pw.append("<td></td>\n");
			pw.append("</tr>\n");
			pw.append("<tr>\n");
			pw.append("<td align='center'><a href=\"javascript:translateMap('" + actionUrl + "', '" + TranslationDirection.WEST + "', '" +
					getId() + "', 'mapDiv_" + _map.getIdCode() + "', 'mapContent_" + _map.getIdCode() + "', '" + _map.getId() + "', '" +
					emptyImage + "', '" + controllerId + "');\"><img src='" + translateLeftImage + "' border='0'/ title=\"" + translationLeft + "\"/></a></td>\n");
			pw.append("<td><a href=\"javascript:translateMap('" + actionUrl + "', '" + TranslationDirection.CENTER + "', '" +
					getId() + "', 'mapDiv_" + _map.getIdCode() + "', 'mapContent_" + _map.getIdCode() + "', '" + _map.getId() + "', '" +
					emptyImage + "', '" + controllerId + "');\"><img src='" + resetImage + "' border='0' title=\"" + translationCenter + "\"/></a></td>\n");
			pw.append("<td align='center'><a href=\"javascript:translateMap('" + actionUrl + "', '" + TranslationDirection.EAST + "', '" +
					getId() + "', 'mapDiv_" + _map.getIdCode() + "', 'mapContent_" + _map.getIdCode() + "', '" + _map.getId() + "', '" +
					emptyImage + "', '" + controllerId + "');\"><img src='" + translateRightImage + "' border='0' title=\"" + translationRight + "\"/></a></td>\n");
			pw.append("</tr>\n");
			pw.append("<tr>\n");
			pw.append("<td></td>\n");
			pw.append("<td align='center'><a href=\"javascript:translateMap('" + actionUrl + "', '" + TranslationDirection.SOUTH + "', '" +
					getId() + "', 'mapDiv_" + _map.getIdCode() + "', 'mapContent_" + _map.getIdCode() + "', '" + _map.getId() + "', '" +
					emptyImage + "', '" + controllerId + "');\"><img src='" + translateDownImage + "' border='0' title=\"" + translationDown + "\"/></a></td>\n");
			pw.append("<td></td>\n");
			pw.append("</tr>\n");
			pw.append("</table>\n");
	}
			
	/**
	 * Generate the HTML code for mouse mode widget.
	 * 
	 * @param pw the page writer
	 */
	protected void generateMouseModeWidget(LyWebPageWriter pw)
			{
				boolean	isRubber = isRubberDefaultMouseAction();
				
		pw.append("<table border='0' cellspacing='3' cellpadding='1' align='center'>\n");
				pw.append("<tr>\n");
				pw.append("<td id='map_selection_mode_select" + _map.getIdCode() + "'");
				
				if (isRubber)
					pw.append(" class='pressedTool'");
				
				pw.append(">\n");
				pw.append("<img src='" + getImageUrl("LY_MAP_SELECT_ICON") + "' onclick=\"javascript:setMapRubberMode('" + _lastFrameName + "', '" +
						getId() + "', true, 'map_selection_mode_select" +
						_map.getIdCode() + "', 'map_selection_mode_move" + _map.getIdCode() + "');\"/>");
				pw.append("</td>\n");
				pw.append("<td id='map_selection_mode_move" + _map.getIdCode() + "'");
				
				if (!isRubber)
					pw.append(" class='pressedTool'");
				
				pw.append(">\n");
				pw.append("<img src='" + getImageUrl("LY_MAP_MOVE_ICON") + "' onclick=\"javascript:setMapRubberMode('" + _lastFrameName + "', '" +
						getId() + "', false, 'map_selection_mode_select" +
						_map.getIdCode() + "', 'map_selection_mode_move" + _map.getIdCode() + "');\"/>");
				pw.append("</td>\n");
				pw.append("</tr>\n");
				pw.append("</table>\n");
			}
		
	/**
	 * Generate the HTML code for zoom widget.
	 * 
	 * @param pw the page writer
	 */
	protected void generateZoomWidget(LyWebPageWriter pw)
		{
		String	controllerId = _map.getRootController();
		String	actionUrl = ((LyStrutsViewManager)getWebViewManager()).getActionURL(LyStrutsConstants.PATH_ACTION_MAP);
			String	zoomIn = translate("LY_MAP_ZOOM_IN");
			String	zoomOut = translate("LY_MAP_ZOOM_OUT");
			String	zoomReset = translate("LY_MAP_ZOOM_RESET");
			String	zoomInImage = getImageUrl("LY_MAP_ZOOM_IN");
			String	zoomOutImage = getImageUrl("LY_MAP_ZOOM_OUT");
			String	zoomCursorImage = getImageUrl("LY_MAP_ZOOM_CURSOR");
			String	zoomScaleImage = getImageUrl("LY_MAP_ZOOM_SCALE");
		String	resetImage = getImageUrl("LY_MAP_RESET_WIDGET");

		pw.append("<table border='0' cellspacing='0' cellpadding='0' align='center'>\n");
			pw.append("<tr>\n");
			pw.append("<td align='center' width='16'><font class='label'>&nbsp;</font></td>\n");
			pw.append("<td align='center' valign='bottom'><a href=\"javascript:setMapZoomScale(currentScale['" + getId() + "'] + 1, '" + actionUrl + "', '" + getId() +
					"', 'map_zoom" + _map.getIdCode() + "', null, null, '" + controllerId + "');\"><img src='" + zoomInImage + "' border='0' title=\"" + zoomIn + "\"/></a></td>\n");
			pw.append("<td align='center' width='16'><font class='label'>&nbsp;</font></td>\n");
			pw.append("</tr>\n");
			pw.append("<tr>\n");
			pw.append("<td align='center' width='16'><font class='label'>&nbsp;</font></td>\n");
			pw.append("<td align='center'>\n");
			pw.append("<div id='map_zoom" + _map.getIdCode() + "_scale'>\n");
			
			for (int scale = ZOOM_STEPS_NB; scale >= -ZOOM_STEPS_NB; scale--)
			{
				pw.append("<div id='map_zoom" + _map.getIdCode() + "_scale" + scale + "'><a href=\"javascript:setMapZoomScale(" + scale + ", '" + actionUrl +
						"', '" + getId() + "', 'map_zoom" + _map.getIdCode() + "', null, null, '" + controllerId + "');\">");
				pw.append("<img src='" + zoomScaleImage + "' border='0'/></a></div>\n");
			}

			pw.append("</div>\n");
			pw.append("</td>\n");
			pw.append("<td align='center' width='16' valign='middle'><a href=\"javascript:setMapZoomScale(0, '" + actionUrl + "', '" + getId() +
					"', 'map_zoom" + _map.getIdCode() + "', null, null, '" + controllerId + "')\">\n");
			pw.append("<img src='" + resetImage + "' border='0' title=\"" + zoomReset + "\"/></a></td>\n");
			pw.append("</tr>\n");
			pw.append("<tr>\n");
			pw.append("<td align='center' width='16'><font class='label'>&nbsp;</font></td>\n");
			pw.append("<td align='center' valign='top'><a href=\"javascript:setMapZoomScale(currentScale['" + getId() + "'] - 1, '" + actionUrl +
					"', '" + getId() + "', 'map_zoom" + _map.getIdCode() + "', null, null, '" + controllerId + "');\">\n");
			pw.append("<img src='" + zoomOutImage + "' border='0' title=\"" + zoomOut + "\"/></a></td>\n");
			pw.append("<td align='center' width='16'><font class='label'>&nbsp;</font></td>\n");
			pw.append("</tr>\n");
			pw.append("</table>\n");
		
			pw.append("<div id='map_zoom" + _map.getIdCode() + "_cursor' style='position:absolute;left:0px;top:0px;z-index:10;visibility:hidden;'\n");
			pw.append("onMouseOver=\"javascript:prepareMovingZoom(event, 'mapDiv_" + _map.getIdCode() + "', 'map_zoom" + _map.getIdCode() + "');\">\n");
			pw.append("<img src='" + zoomCursorImage + "'/></div>\n");
		}

	protected int getMapZoomScale()
	{
		float	zoomFactor = _map.getZoomFactor();
		float	zoomStep = _map.getZoomStep();
		
		if (zoomStep <= 1)
			zoomStep = 2;
		
		int		scale = (int)Math.floor((Math.log(zoomFactor) / Math.log(zoomStep)) + 0.001);
		
		if (scale > ZOOM_STEPS_NB)
			scale = ZOOM_STEPS_NB;
		else if (scale < -ZOOM_STEPS_NB)
			scale = -ZOOM_STEPS_NB;
		
		return scale;
	}
	
	/**
	 * Set the <code>_refreshViewerImage</code> attribute in order to refresh the viewer image used by the map.
	 */
	public void refreshViewerImage()
	{
		_refreshViewerImage = true;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetIdleInterface#setIdle(boolean)
	 */
	public void setIdle(boolean status)
	{
		_idle = status;

        if (!status && _refreshBackgroundImage)
        {
            updateBackgroundImage();
            _refreshBackgroundImage = false;
        }

		if (!status && _refreshViewerImage)
		{
			updateViewerImage();
			_refreshViewerImage = false;
		}
	}
	
	/**
	 * Refresh the background image used by the map.
	 */
	public void updateBackgroundImage()
	{
		LyStrutsMapCanvas	mapCanvas = getMapCanvas();
		Dimension			dimension = getMapDimension();
		
		mapCanvas.setSize(dimension.width, dimension.height);
		mapCanvas.setZoom(_map.getZoomFactor());
		mapCanvas.resetTranslation();
		
		Rectangle			boundingRect = mapCanvas.getBoundingRectangle();
		String				mapContentId = "mapContent_" + _map.getIdCode();
		StringBuffer		script = new StringBuffer();
		LyWebViewManager	viewManager = getWebViewManager();
		String				images = generateMapImages();
		String				mapImagesPrefix = "mapImages_" + _map.getIdCode() + "_";
		String				mapImagesIdOlder = mapImagesPrefix + (_mapImagesCounter - 1);
		String				mapImagesIdOld = mapImagesPrefix + _mapImagesCounter;
		
		_mapImagesCounter++;
		
		String				mapImagesId = mapImagesPrefix + _mapImagesCounter;

		script.append("<script type='text/javascript'>\n//<![CDATA[");
		script.append("var frame = findTarget(null, '" + _lastFrameName + "', null);\n");
		script.append("if (frame != null)\n");
		script.append("{\n");
		script.append("var parentDiv = frame.document.getElementById('" + mapContentId + "');\n");
		script.append("var imagesDiv = frame.document.getElementById('" + mapImagesIdOld + "');\n");
		
		if ((boundingRect.x > 0) || (boundingRect.y > 0))
			script.append("var timer = 0;\n");
		else
			script.append("var timer = 500;\n");
		
		int					 left = (boundingRect.x + mapCanvas._scrollx);
		int					 top = (boundingRect.y + mapCanvas._scrolly);
		LyStrutsLayoutManager layoutManager = mapCanvas.getLayoutManager();
		
		if ((layoutManager != null) && layoutManager.centerMapContent())
		{
			left = (dimension.width - boundingRect.width) /2;
			top = (dimension.height - boundingRect.height) /2;
		}

		script.append("if (imagesDiv != null)\n");
		script.append("{\n");
		script.append("var newLeft = parseInt(imagesDiv.style.left) + parseInt(parentDiv.style.left) - " + (boundingRect.x + mapCanvas._scrollx) + ";\n");
		script.append("var newTop = parseInt(imagesDiv.style.top) + parseInt(parentDiv.style.top) - " + (boundingRect.y + mapCanvas._scrolly) + ";\n");
		script.append("imagesDiv.style.left = newLeft;\n");
		script.append("imagesDiv.style.top = newTop;\n");
		script.append("}\n");
		script.append("}\n");
		script.append("SetDivSize('" + mapContentId + "', " + boundingRect.width + ", " + boundingRect.height + ");");
		script.append("SetDivPosition('" + mapContentId + "', " + left + ", " + top + ");");
		script.append("removeDiv('" + _lastFrameName + "', '" + mapImagesIdOlder + "');");
		script.append("setTimeout(\"removeDiv('" + _lastFrameName + "', '" + mapImagesIdOld + "')\", timer);");
		script.append("var newDiv = addDiv('" + _lastFrameName + "', '" + mapContentId + "', \"" + images + "\", '" + mapImagesId + "');");
		script.append("if ((newDiv != null) && (typeof newDiv != 'undefined')) {");
		script.append("newDiv.style.position = 'absolute';");
		script.append("newDiv.style.left = '0px';");
		script.append("newDiv.style.top = '0px';");
		script.append("}");
		
		String	useMapContent = getUseMapContent();

		useMapContent = LyString.replace(useMapContent, "\\", "\\\\");
		useMapContent = LyString.replace(useMapContent, "\"", "\\\"");
		useMapContent = LyString.replace(useMapContent, "\n", "");
		
		script.append("replaceDivInnerHtml('" + _lastFrameName + "', '" + _map.getId() + "', \"" + useMapContent + "\");");

		script.append("//]]>\n</script>");

		viewManager.addAdditionalRefreshScripts(script.toString());
	}

	/**
	 * Refresh the viewer image used by the map.
	 */
	public void updateViewerImage()
	{
		if (_map.isViewer())
		{
			clearMapFiles(true);

			String				viewerImagePath = generateMapImage(-1, -1, -1, -1, true);
			LyWebViewManager	viewManager = getWebViewManager();
			StringBuffer		script = new StringBuffer();

			script.append("<script type='text/javascript'>\n//<![CDATA[\nSetImageSrc('").append(_map.getId()).
				append("_viewer_image', '").append(viewerImagePath).append("');\n//]]>\n</script>");

			viewManager.addAdditionalRefreshScripts(script.toString());
		}
	}
	
	/**
	 * Adds the given component.
	 *
	 * @param component the component to add
	 * @param center    in case of ellipse layout, center is the id of the node that must be placed
	 *                  at the center, null otherwise
	 * @since leonardi V4.0, replace {@link #addComponent(LyComponent, int, String)} where the
	 *        second parameter was not used.
	 */
	protected void addComponent(LyComponent component, String center)
	{
		addComponent(component, center, true);
	}
	
	/**
	 * Adds the given component.
	 *
	 * @param component       the component to add
	 * @param center          in case of ellipse layout, center is the id of the node that must be placed
	 *                        at the center, null otherwise
	 * @param generateScripts indicates whether a Javascript script must be generated to add component
	 */
	protected void addComponent(LyComponent component, String center, boolean generateScripts)
	{
		String			  name, id;
		LyStrutsMapElement element;

		if (component == null)
			return;

		LyLayout mapLayout = _map.getLayout();
		boolean  useCoordinates = (mapLayout == null)
			|| (mapLayout.getStyle() == LyLayout.Style.GEOGRAPHIC)
			|| (mapLayout.getStyle() == LyLayout.Style.COORDINATES);
		
		if ((component instanceof LyMapNode) && useCoordinates)
		{
			synchronized (this)
			{
				LyMapNode mapNode = (LyMapNode)component;

				element = (LyStrutsMapElement)component.getInternalView();
				name = null;

				if ((center != null) && (component instanceof LyGetIdInterface))
				{
					id = ((LyGetIdInterface)component).getId();

					if ((id != null) && (id.equals(center)))
						name = "Center";
				}

				// Retrieve the parent node (if any)
				String				nodeId;
				LyStrutsMapNode		node;
				LyStrutsMapCanvas	mapCanvas = _mapbox.getMapCanvas();

				nodeId = mapNode.getParentNode();
				node = getNodeById(nodeId);
				((LyStrutsMapNode)element).setParentNode(node);

				_mapbox.addElement(name, element);

				if (generateScripts && !mapCanvas.needDrawing(element) && !element.isHidden())
				{
					// Add refresh script
					LyWebViewManager viewManager = getWebViewManager();
					StringBuffer     buffer = new StringBuffer();
					LyWebPageWriter  pw = new LyStrutsPage(_map, true);

					pw.setActiveBuffer(_activeName);
					pw.setWindowName(_lastFrameName);
					((LyStrutsMapNode)element).drawHtml(mapCanvas, pw);

					StringBuffer innerHtml = pw.getBuffer(pw.getActiveName());
					String		 innerHtmlStr = LyString.replace(innerHtml.toString(), "\n", "\\n");

					innerHtmlStr = LyString.replace(innerHtmlStr, "\\", "\\\\");
					innerHtmlStr = LyString.replace(innerHtmlStr, "\\\\n", "\\n");
					innerHtmlStr = LyString.replace(innerHtmlStr, "\"", "\\\"");
					innerHtmlStr = LyString.replace(innerHtmlStr, "\\'", "\\\\'");

					buffer.append("<script type='text/javascript'>\n//<![CDATA[\naddDiv('").append(
						_lastFrameName).append("','mapContent_" + _map.getIdCode() + "',\"").
						append(innerHtmlStr).append("\");\n//]]>\n</script>");

					viewManager.addAdditionalRefreshScripts(buffer.toString());
					pw.free();
				}

				refreshViewerImage();
			}
		}
		else
		{
			synchronized (this)
			{

				element = (LyStrutsMapElement)component.getInternalView();
				name = null;
				if ((center != null) && (component instanceof LyGetIdInterface))
				{
					id = ((LyGetIdInterface)component).getId();
					if ((id != null) && (id.equals(center)))
						name = "Center";
				}

				if (element instanceof LyStrutsMapLink)
				{
					// We must search for extremities.
					String		   nodeId;
					LyStrutsMapNode node;

					nodeId = ((LyMapLink)component).getStartNode();
					node = getNodeById(nodeId);
					((LyStrutsMapLink)element).setStartNode(node);

					nodeId = ((LyMapLink)component).getEndNode();
					node = getNodeById(nodeId);
					((LyStrutsMapLink)element).setEndNode(node);
				}
				else if (element instanceof LyStrutsMapNode)
				{
					// Retrieve the parent node (if any)
					String		   nodeId;
					LyStrutsMapNode node;

					nodeId = ((LyMapNode)component).getParentNode();
					node = getNodeById(nodeId);
					((LyStrutsMapNode)element).setParentNode(node);
				}

				_mapbox.addElement(name, element);
				refresh();
				
				if (mapLayout.getStyle() == LyLayout.Style.TREEMAP)
					getWebViewManager().addAdditionalRefreshScripts("<script type='text/javascript'>\n//<![CDATA[\n" +
							"getLeonardiTargetFrame().ReinitMap();\n//]]>\n</script>");
			}
		}
	} // end method addComponent

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalMap#removeComponent(leon.view.LyComponent)
	 */
	public void removeComponent(LyComponent component)
	{
		if (component == null)
			return;

		LyLayout mapLayout = _map.getLayout();
		boolean  useCoordinates = (mapLayout == null)
			|| (mapLayout.getStyle() == LyLayout.Style.GEOGRAPHIC)
			|| (mapLayout.getStyle() == LyLayout.Style.COORDINATES);

		if ((component instanceof LyMapNode) && useCoordinates)
		{
			// Remove the DIV for node image and node label
			LyMapNode		  mapNode = (LyMapNode)component;
			LyStrutsMapElement element = (LyStrutsMapElement)component.getInternalView();

			_mapbox.removeElement(element);

			LyWebViewManager viewManager = getWebViewManager();
			StringBuffer     script = new StringBuffer();

			script.append("<script type='text/javascript'>\n//<![CDATA[\nremoveDiv(\"");
			script.append(_lastFrameName).append("\",\"mapnode_");
			script.append(mapNode.getIdCode()).append("_node\");\n");
			script.append("removeDiv(\"");
			script.append(_lastFrameName).append("\",\"mapnode_");
			script.append(mapNode.getIdCode()).append("_Label\");\n//]]>\n</script>");

			viewManager.addAdditionalRefreshScripts(script.toString());

			refreshViewerImage();
		}
		else
		{
			LyStrutsMapElement element = (LyStrutsMapElement)component.getInternalView();

			_mapbox.removeElement(element);

			refresh();
		}
	} // end method removeComponent
	
	/*
	 * (non-Javadoc)
	 * @see leon.view.web.struts.LyStrutsPanel#getScrollHeightForSplitter(leon.view.web.LyWebPageWriter)
	 */
	public int getScrollHeightForSplitter(LyWebPageWriter pw)
	{
		LyStrutsMapCanvas	mapCanvas = _mapbox.getMapCanvas();
		LyLayout			mapLayout = _map.getLayout();
		boolean				useTreemapLayout = (mapLayout != null) && (mapLayout.getStyle() == LyLayout.Style.TREEMAP);
		int					mapHeight = 0;

		if (!useTreemapLayout)
		{
			mapHeight = MAP_HEIGHT;

			if ((mapLayout != null)
					&& ((mapLayout.getStyle() == LyLayout.Style.MATRIX)
						|| (mapLayout.getStyle() == LyLayout.Style.LINEAR)))
			{
				mapCanvas.resetView();

				LyStrutsLayoutManager htmlLayout = mapCanvas.getLayoutManager();

				if (htmlLayout != null)
				{
					htmlLayout.layoutMap(mapCanvas);

					Rectangle bb = htmlLayout.getBoundingRectangle(mapCanvas);
	
					mapHeight = bb.height;
				}
			}
		}

		return (mapHeight >= LyStrutsSplitter.MAXIMUM_COMPONENT_HEIGHT ? LyStrutsSplitter.MAXIMUM_COMPONENT_HEIGHT : -1);
	}

	/**
	 * Returns the map canvas.
	 */
	public LyStrutsMapCanvas getMapCanvas()
	{
		return _mapbox.getMapCanvas();
	}
	
	/**
	 * Resets the position of all the elements contained in the map.
	 */
	public void resetElementsPositions()
	{
		int		i, n = _mapbox.getMapCanvas().getElementCount();
		
		for (i = 0; i < n; i++)
		{
			LyStrutsMapElement	element = _mapbox.getMapCanvas().getElementAt(i);

			element.resetPosition();
		}
	}

	/**
	 * Checks if the component is idle or not.
	 *
	 * @return true if the component is idle (it is not refreshed when modifications are applied).
	 */
	public boolean isIdle()
	{
		return _idle;
	}

	public void setZoomFactor(float zoomFactor)
	{
        _mapbox.getMapCanvas().setZoom(zoomFactor);
		
		StringBuffer	script = new StringBuffer();
		int				scale = getMapZoomScale();

		script.append("<script type='text/javascript'>\n//<![CDATA[\n");
		script.append("setMapZoomScale(" + scale + ", '', '', 'map_zoom" + _map.getIdCode() + "');\n");
		script.append("\n//]]>\n</script>\n");
		
		getWebViewManager().addAdditionalRefreshScripts(script.toString());

		resetElementsPositions();

		if (!isIdle())
		{
			updateBackgroundImage();
			updateViewerImage();
		}
		else
		{
			_refreshBackgroundImage = true;
			_refreshViewerImage = true;
		}
	}

	/**
	 * (non-Javadoc).
	 * 
	 * @see leon.view.graphic.LyGraphicalMap#setTranslationStatus
	 */
	public void setTranslationStatus(TranslationStatus translationStatus)
	{
		boolean			open = (translationStatus == LyMap.TranslationStatus.SHOW);
		StringBuffer	script = new StringBuffer();
		String			image = getImageUrl(open ? "LY_MAP_COLLAPSE_WIDGET" : "LY_MAP_EXPAND_WIDGET");
		String			translationJS = translate(open ? "LY_MAP_COLLAPSE_TRANSLATION" : "LY_MAP_EXPAND_TRANSLATION");
		
		translationJS = LyString.replace(translationJS, "'", "\\'");
		script.append("<script type='text/javascript'>\n//<![CDATA[\n");
		script.append("setMapWidgetState(" + open + ", 'map_translation" + _map.getIdCode() + "', 'map_translation_switch_img" + _map.getIdCode() + "', '" + image + "', '" + translationJS + "');");
		script.append("setMapWidgetState(" + open + ", 'map_selection_mode" + _map.getIdCode() + "');");
		script.append("\n//]]>\n</script>\n");
		
		getWebViewManager().addAdditionalRefreshScripts(script.toString());
	}

	/**
	 * (non-Javadoc).
	 * 
	 * @see leon.view.graphic.LyGraphicalMap#setZoomStatus
	 */
	public void setZoomStatus(ZoomStatus zoomStatus)
	{
		boolean			open = (zoomStatus == LyMap.ZoomStatus.SHOW);
		StringBuffer	script = new StringBuffer();
		String			image = getImageUrl(open ? "LY_MAP_COLLAPSE_WIDGET" : "LY_MAP_EXPAND_WIDGET");
		String			translationJS = translate(open ? "LY_MAP_COLLAPSE_ZOOM" : "LY_MAP_EXPAND_ZOOM");
		
		translationJS = LyString.replace(translationJS, "'", "\\'");
		script.append("<script type='text/javascript'>\n//<![CDATA[\n");
		script.append("setMapWidgetState(" + open + ", 'map_zoom" + _map.getIdCode() + "', 'map_zoom_switch_img" + _map.getIdCode() + "', '" + image + "', '" + translationJS + "');");
		script.append("setMapZoomScale(currentScale['" + getId() + "'], '', '', 'map_zoom" + _map.getIdCode() + "');");
		script.append("\n//]]>\n</script>\n");
		
		getWebViewManager().addAdditionalRefreshScripts(script.toString());
	}
	
	/**
	 * (non-Javadoc).
	 * 
	 * @see leon.view.graphic.LyGraphicalMap#setViewerStatus
	 */
	public void setViewerStatus(ViewerStatus viewerStatus)
	{
		boolean			open = (viewerStatus == LyMap.ViewerStatus.SHOW);
		StringBuffer	script = new StringBuffer();
		String			image = getImageUrl(open ? "LY_MAP_COLLAPSE_WIDGET" : "LY_MAP_EXPAND_WIDGET");
		String			translationJS = translate(open ? "LY_MAP_COLLAPSE_VIEWER" : "LY_MAP_EXPAND_VIEWER");
		
		translationJS = LyString.replace(translationJS, "'", "\\'");
		script.append("<script type='text/javascript'>\n//<![CDATA[\n");
		script.append("setMapWidgetState(" + open + ", 'map_viewer_img" + _map.getIdCode() + "', 'map_viewer_switch_img', '" + image + "', '" + translationJS + "');");
		script.append("setMapWidgetState(" + open + ", 'map_viewer_moving" + _map.getIdCode() + "', '');");
		script.append("setMapWidgetState(" + open + ", 'map_viewer_translate" + _map.getIdCode() + "', '');");
		script.append("\n//]]>\n</script>\n");
		
		getWebViewManager().addAdditionalRefreshScripts(script.toString());
	}

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebPage#writeBeginPage(java.io.Writer, java.lang.String)
     */
    public void writeBeginPage(Writer pw, String frame)
        throws IOException
    {
        String controllerId = _map.getRootController();

        if ((controllerId + MAP_COMPONENT).equals(frame))
        {
            pw.write("<!-- Generated by Leonardi (Lyria) -->\n");
            pw.write("<html><head>\n");
            pw.write("<meta name=\"AUTHOR\" content=\"Lyria\"/>\n");
            pw.write("<meta name=\"GENERATOR\" content=\"Leonardi\"/>\n");
    		pw.write("<meta name=\"Content-Type\" content=\"text/html;charset=" + LyString.DEFAULT_CHARSET + "\"/>\n");
            pw.write("<title>");
            pw.write((_title == null) ? "" : translate(_title));
            pw.write("</title>\n");
            writeCssScripts(pw, frame);
            pw.write("</head>\n");
            pw.write(
                "<body scroll='no' topmargin='0' leftmargin='0' style='height:100%;width:100%;'>\n");
            pw.write("<script>var pageWriterId='" + getId() + "';</script>\n");
        }
        else
            super.writeBeginPage(pw, frame);
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebPageWriter#preparePage(java.lang.String, java.lang.String)
     */
    public void preparePage(String frame, String targetWindow)
    {
        super.preparePage(frame, targetWindow);

        String controllerId = _map.getRootController();

        clearBuffer(controllerId + HIDDEN);
        setActiveBuffer(controllerId + HIDDEN);

        writeHiddenFrame(this);
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebPage#init()
     */
    public void init()
    {
        super.init();
        BitSet bs = new BitSet();

        _lastFileNames = new Vector<String>();
		_lastViewerFileNames = new Vector<String>();

        LyMap.SelectionMode selectionMode = _map.getSelectionMode();

        _mapbox = createMapBox(bs, selectionMode);

        LyStrutsMapCanvas canvas = _mapbox.getMapCanvas();

        if (_map.getBackground() != null)
            canvas.setBackground(getImage(_map.getBackground()));

        // Always put a background color it is used even if we have a background
        // image
        Color col = getColor("LY_MAP_BACKGROUND");

        if (col == null)
            col = getColor("LY_DEFAULT_BACKGROUND");

        if (col != null)
            canvas.setBackground(col);

        if (_map.getLabel() != null)
            canvas.setTitle(translate(_map.getLabel()));

        canvas.setSelectionColor(getColor("LY_MAP_SELECTED_COLOR"));
    } // end method init

    /**
     * Creates the map box of this map.
     *
     * @param  bs            the map style
     * @param  selectionMode the selection mode
     * @return the created map box
     */
    protected LyStrutsMapBox createMapBox(BitSet bs, LyMap.SelectionMode selectionMode)
    {
        return new LyStrutsMapBox(bs, selectionMode, this);
    }

    /**
     * Sets the popup menu with the given value.
     *
     * @param popup the new popup menu
     */
    public void setPopupMenu(LyPopupMenu popup)
    {
        LyStrutsViewManager viewManager = (LyStrutsViewManager)getWebViewManager();
        LyView             view = _viewComponent.getParentView();
        
        viewManager.setPopupMenu(popup, LyWebConstants.POPUP_MENU + view.getIdCode());
    }

    /**
     * Sets the content of the map.
     */
    public void setContent()
    {
        // Building the layout manager.
        synchronized (this)
        {
            String center = null;

            if (_map.getLayout() != null)
            {
                boolean              isModelisation = (_map.getLayout().getStyle()
                            == LyLayout.Style.LINEAR);
                LyStrutsLayout        internal;
                LyStrutsLayoutManager layout;

                internal = (LyStrutsLayout)_map.getLayout().getInternalView();

                layout = internal.layoutFactory();

                center = _map.getLayout().getCenter();
                _mapbox.getMapCanvas().setLayoutManager(layout);

                if (isModelisation)
                    _map.setSelectionDraw(LyMap.SelectionDraw.BORDER);
            }

            // Reset the elements of the mapbox
            _mapbox.resetElements();

            // Add the elements to the map
            LyComponent  component;
            LyMapContent content = _map.getMapContent();
            // Retrieve the mapContent view (in order to instanciate it)
            Vector<LyComponent>      components = new Vector<LyComponent>();
            int          i, n;

            if (content != null)
            {
                n = content.getComponentCount();

                for (i = 0; i < n; i++)
                    components.addElement(content.getComponent(i));
            }

            n = _map.getComponentCount();

            for (i = 0; i < n; i++)
                components.addElement(_map.getComponent(i));

            n = components.size();

            for (i = 0; i < n; i++)
            {
                component = components.elementAt(i);
                addComponent(component, center, false);
            }

            if (_map.getBackground() != null)
                _mapbox.getMapCanvas().setBackground(getImage(_map.getBackground()));
        }
    } // end method setContent

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalMap#addComponent(leon.view.LyComponent, int)
     */
    public void addComponent(LyComponent component, int pos)
    {
        //CLEANCODE : the method addComponent(component, pos, null) is deprecated
        //addComponent(component, pos, null);
        addComponent(component, null);
    }

    /**
     * Adds the given component.
     *
     * @param      component the component to add
     * @param      pos       unused paramter
     * @param      center    in case of ellipse layout, center is the id of the node that must be
     *                       placed at the center, null otherwise
     * @deprecated replaced by {@link #addComponent(LyComponent, String)} because the parameter
     *             <code>pos</code> is not used.
     */
    protected void addComponent(LyComponent component, int pos, String center)
    {
        addComponent(component, center);
    }

    /**
     * Writes HTML description corresponding to this component.
     *
     * @param pw the page writer
     */
    public void writeHiddenFrame(LyWebPageWriter pw)
    {
        String forceReload = "";
        String targetWindow = pw.getWindowName(), targetWindowStr;

        if (getForceReloadState())
        {
            _lastReloadTime = System.currentTimeMillis();
            setForceReloadState(false);
        }

        if (targetWindow == null)
            targetWindowStr = "";
        else
            targetWindowStr = "&#38;" + LyWebConstants.TARGET_WINDOW + "=" + targetWindow;

        if (_lastReloadTime > 0)
            forceReload = "&#38;reload=" + _lastReloadTime;

        String controllerId = _map.getRootController();
        String url = getRootUrl() + "?" + LyWebConstants.PAGE + "=" + getId() + forceReload
            + targetWindowStr;

        pw.addEndScript("<script type='text/javascript'>ResizeMap('" + controllerId + MAP_COMPONENT
            + "', '" + url + "');</script>", false);
    } // end method writeHiddenFrame

    /**
     * Computes the list of map regions that area really displayed on the map.
     *
     * @param region   the root region of the computation
     * @param elements the list of elements
     */
    protected void computeDisplayedMapRegions(LyMapRegion region, Vector<LyComponent> elements)
    {
        // If element is a map region and has sub-map regions, do not consider
        // it as sensitive (but its children)
        int i, n = region.getMapRegionCount();

        if (n == 0)
            elements.addElement(region);
        else
        {
            for (i = 0; i < n; i++)
                computeDisplayedMapRegions(region.getMapRegion(i), elements);

            // Check if region has a sensitive label
            LyStrutsMapRegion htmlMapRegion = (LyStrutsMapRegion)region.getInternalView();

            if (htmlMapRegion.getLabelFrame() != null)
                elements.addElement(region);
        }
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebPage#setVisible(boolean)
     */
    public void setVisible(boolean status)
    {
        if (status)
        {
            LySetVisibleInterface visible = _map.getSetVisibleParent();

            if (visible != null)
                visible.setVisible(true);
        }
    }

    /**
     * Sets sub-component mapContent.
     * @param mapContent New value of mapContent
     */
    public void setMapContent(LyMapContent mapContent)
    {
        setContent();
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LyResetContentInterface#resetContent()
     */
    public void resetContent()
    {
        resetContent(false);
    }
    
    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LyResetContentInterface#resetContent(boolean)
     */
    public void resetContent(boolean full)
    {
        synchronized (this)
        {
            LyStrutsMapCanvas mapCanvas = _mapbox.getMapCanvas();

            mapCanvas.resetView();
        }
    }
    
    /**
     * Returns the dimension of the generated image.
     * 
     * @return the dimension of the generated image.
     */
    public Dimension getMapDimension()
    {
        LyLayout            mapLayout = _map.getLayout();
        LyStrutsMapCanvas    mapCanvas = _mapbox.getMapCanvas();
        int                 mapWidth = MAP_WIDTH;
        int                 mapHeight = MAP_HEIGHT;

        if (_map.getWidth() > 0)
            mapWidth = _map.getWidth();

        if (_map.getHeight() > 0)
            mapHeight = _map.getHeight();

        if ((mapLayout != null)
                && ((mapLayout.getStyle() == LyLayout.Style.MATRIX)
                    || (mapLayout.getStyle() == LyLayout.Style.LINEAR)))
        {
            LyStrutsLayoutManager htmlLayout = mapCanvas.getLayoutManager();

            if (htmlLayout != null)
            {
                htmlLayout.layoutMap(mapCanvas);

	            Rectangle bb = htmlLayout.getBoundingRectangle(mapCanvas);
	
	            mapWidth = bb.width;
	            mapHeight = bb.height;
            }
        }
        else if ((mapLayout != null) && (mapLayout.getStyle() == LyLayout.Style.TREEMAP))
        {
            if (_mapWidth > 0)
                mapWidth = _mapWidth;

            if (_mapHeight > 0)
                mapHeight = _mapHeight;
        }

        // Event if the map has no content, we need to draw it (may be
        // usefull
        // to go up in the map hierarchy by double-clicking on the
        // background)
        mapWidth = Math.max(mapWidth, 20);
        mapHeight = Math.max(mapHeight, 20);
        
        return new Dimension(mapWidth, mapHeight);
    }

    /**
     * Generate the image of the map and returns the path of the generated file.
     * 
     * @return the path of generated image.
     */
    protected String generateMapImage()
    {
        BufferedImage image = null;
        String        mapId = _map.getId();
        String        dateStamp = "";

        dateStamp = "_" + System.currentTimeMillis();

        LyStrutsMapCanvas mapCanvas = _mapbox.getMapCanvas();

        if (mapCanvas.isInitialSettings())
            mapCanvas.resetView();
        
        Dimension   dimension = getMapDimension();
        int         mapWidth = dimension.width;
        int         mapHeight = dimension.height;

        mapCanvas.setSize(mapWidth, mapHeight);

        String imagePath = null;

        if (!NO_BACKGROUND)
        {
            image = new BufferedImage(mapWidth, mapHeight, BufferedImage.TYPE_INT_BGR);

            mapCanvas.paint(image.getGraphics());

            // Create the image
            LyViewManager viewManager = getViewManager();
            String        tmpDir = viewManager.getTemporaryDir();
            String        tmpRoot = viewManager.getTemporaryLocation();
            String        imageFile;

            imageFile = tmpDir + "/map_" + mapId + dateStamp + ".jpg";
            imagePath = tmpRoot + "/map_" + mapId + dateStamp + ".jpg";
            LyImageUtils.encodeImage(image, tmpDir, imageFile);
        }

        if (image != null)
        {
            image.flush();
            image = null;
        }
        
        return imagePath;
    }
    
    /**
     * Gets the map box component.
     *
     * @return the map box component
     */
    public Component getComponent()
    {
        return _mapbox;
    }

    /**
     * Gets the node cprresponding to the given Id.
     *
     * @param  nodeId the id of the wanted node
     * @return the node if it is found, null otherwise
     */
    public LyStrutsMapNode getNodeById(String nodeId)
    {
        int         i, n = _map.getComponentCount();
        LyComponent element;

        for (i = 0; i < n; i++)
        {
            element = _map.getComponent(i);

            if ((element instanceof LyMapNode) && (((LyMapNode)element).getId().equals(nodeId)))
                return (LyStrutsMapNode)element.getInternalView();
        }

        // If node not found in the map, check if in the map content
        LyMapContent content = _map.getMapContent();

        if (content != null)
        {
            n = content.getComponentCount();

            for (i = 0; i < n; i++)
            {
                element = content.getComponent(i);

                if ((element instanceof LyMapNode)
                        && (((LyMapNode)element).getId().equals(nodeId)))
                    return (LyStrutsMapNode)element.getInternalView();
            }
        }

        return null;
    } // end method getNodeById

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
     */
    public void setLabel(String label)
    {
        LyEnvironment env = getEnvironment();

        label = env.translate(label);
        _mapbox.setTitle(label);

        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalMap#setSpatial(boolean)
     */
    public void setSpatial(boolean spatial)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetViewerInterface#setViewer(boolean)
     */
    public void setViewer(boolean viewer)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalMap#setLayout(leon.view.LyLayout)
     */
    public void setLayout(LyLayout layout)
    {
        refresh();
    }

    /**
     * Set the <code>_refreshBackgroundImage</code> attribute in order to refresh the background image used by the map.
     */
    public void refreshMapImage()
    {
        _refreshBackgroundImage = true;
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebPageWriter#getPageWriters()
     */
    public Vector<LyWebPageWriter> getPageWriters()
    {
        Vector<LyWebPageWriter> v = new Vector<LyWebPageWriter>(1);

        v.addElement(this);
        return v;
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebPageWriter#getRedirect(java.lang.String)
     */
    public String getRedirect(String frame)
    {
        return null;
    }

    /**
     * Retrieves a Color from the current environment.
     *
     * @param  key The ressource name (key) of the Color.
     * @return A reference to the requested Color, or null if it could not be found.
     * @see    LyEnvironment#getColor(String)
     */
    public final Color getColor(String key)
    {
        return (Color)getEnvironment().getColor(key);
    }

    /**
     * Retrieves a Font from the current environment.
     *
     * @param  key The ressource name (key) of the font.
     * @return A reference to the requested Font, or a default font if it could not be found.
     * @see    LyEnvironment#getFont(String)
     */
    public final Font getFont(String key)
    {
        return (Font)getEnvironment().getFont(key);
    }

    /**
     * Retrieves an image from the current environment.
     *
     * @param  key The ressource name (key) of the image.
     * @return A reference to the image object, or null if it could not be found.
     * @see    LyEnvironment#getImage(String)
     */
    public final Image getImage(String key)
    {
        return (Image)getEnvironment().getImage(key);
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebPageWriter#isFrameSet()
     */
    public boolean isFrameSet()
    {
        LyLayout mapLayout = _map.getLayout();

        if ((mapLayout != null) && (mapLayout.getStyle() == LyLayout.Style.TREEMAP))
            return (!_noFrames);

        return false;
    }

    public void setTranslation(boolean translation)
    {
        refresh();
    }

    public void setZoom(boolean zoom)
    {
        refresh();
    }

} // end class LyStrutsMap
