/*
 * LyStrutsMapLink.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web.struts;

import java.awt.*;

import leon.misc.*;
import leon.view.*;
import leon.view.common.*;
import leon.view.graphic.*;
import leon.view.web.*;
import leon.view.web.struts.map.*;
import leon.view.win.swing.*;


/**
 * This class implements the Map Link object (as described in the dtd) for Struts viewer.
 *
 * <p><b>MAP_LINK: Map Link</b><br>
 * Map links allow to define segments between two {@link leon.view.web.struts.LyStrutsMapNode nodes}
 * of the map. If one of the two nodes defining the link is moved, the link follows the moved node
 * so that it always starts from its source node and that it always arrives to its destination node.
 * </p>
 *
 * @version $Id: LyStrutsMapLink.java 24563 2010-03-04 08:15:19Z jlsaouli $
 * @author  Lyria
 * @see     leon.view.LyMapLink
 */
public class LyStrutsMapLink extends LyStrutsMapObject
	implements LyGraphicalMapLink,
		LyStrutsMapElement
{
	public static final String __VERSION =
		"$Id: LyStrutsMapLink.java 24563 2010-03-04 08:15:19Z jlsaouli $";

	/** Offset between the link and the selection area. */
	public static final int    SELECTION_OFFSET = 6;

	/** Enum with label alignement **/
	private enum LabelAlignment { START, CENTER, END }
	
	/** Start node. */
	protected LyStrutsMapNode   _startNode;

	/** End node. */
	protected LyStrutsMapNode   _endNode;

	/** The map link generic component. */
	protected LyMapLink		   _mapLink;

	/** Local cache for the image shown on the start part of the link. */
	protected Image			   _imageStart;

	/** Resource key of the image shown on the start part of the link. */
	protected String		   _imageStartString;

	/** Local cache for the image shown on the middle part of the link. */
	protected Image			   _imageMiddle;

	/** Resource key of the image shown on the middle part of the link. */
	protected String		   _imageMiddleString;

	/** Image shown on the end part of the link. */
	protected Image			   _imageEnd;

	/** Resource key of the image shown on the end part of the link. */
	protected String		   _imageEndString;

	/** Local cache for the color used for the start part of the link. */
	protected Color			   _colorStart;

	/** Resource key of the color used for the start part of the link. */
	protected String		   _colorStartString;

	/** Local cache for the color used for the middle part of the link. */
	protected Color			   _color;

	/** Resource key of the color used for the middle part of the link. */
	protected String		   _colorString;

	/** Local cache for the color used for the end part of the link. */
	protected Color			   _colorEnd;

	/** Resource key of the color used for the end part of the link. */
	protected String		   _colorEndString;

	/** The color of the link when it is selected. */
	protected Color			   _selectedColor;

	/** Font color of the map link. */
	protected Color			   _fontColor;

	/** Outlines colors of the map link. */
	protected Color			   _outlineColor;

	/** Font of the node. */
	protected Font			   _graphicalFont;

	/** Cache to translate label only once (may access very often). */
	protected String		   _labelCache;

	/** Width of the label (for optimization). */
	protected int			   _labelWidth;

	/** Height of the label (for optimization). */
	protected int			   _labelHeight;

	/** Cached translated label that can be accessed many times. */
	protected String		   _endLabelCache;

	/** Width of the label (for optimization). */
	protected int			   _endLabelWidth;

	/** Height of the label (for optimization). */
	protected int			   _endLabelHeight;
	
	/** Cached translated label that can be accessed many times. */
	protected String		   _startLabelCache;

	/** Width of the label (for optimization). */
	protected int			   _startLabelWidth;

	/** Height of the label (for optimization). */
	protected int			   _startLabelHeight;
	
	/**
	 * Array with link's images positions.<br>
	 * It contains:
	 *
	 * <ul>
	 *   <li>[0] = start image position</li>
	 *   <li>[1] = middle image position</li>
	 *   <li>[2] = end image position</li>
	 * </ul>
	 */
	protected Point[]		   _imagePos;

	/**
	 * Class constructor of a Map Link object for the struts target.
	 *
	 * @param mapLink the generic Map Link defined for every type of view targets
	 */
	public LyStrutsMapLink(LyMapLink mapLink)
	{
		super(mapLink);
		_mapLink = mapLink;

		initialize();
	}

	/**
	 * Initialization method.<br>
	 * Sets the font of the node with the value from resource.<br>
	 * Sets check box images with the appropriate resource switch the parent map selection mode is
	 * multiple or not.<br>
	 * Sets other attributes to null or -1.
	 */
	protected void initialize()
	{
		// Get images String
		_imageStartString = null;
		_imageMiddleString = null;
		_imageEndString = null;
		_imageStart = null;
		_imageMiddle = null;
		_imageEnd = null;

		_colorStart = null;
		_colorStartString = null;
		_color = null;
		_colorString = null;
		_colorEnd = null;
		_colorEndString = null;

		_selectedColor = null;
		_fontColor = null;

		_labelCache = null;
		_labelWidth = -1;
		_labelHeight = -1;

		LyComponent parent = _mapLink.getParent();

		if (parent instanceof LyMapContent)
			parent = parent.getParent();

		getImages();

		_graphicalFont = getFont("LY_MAPNODE_FONT");
	} // end method initialize

	@Override
	public void free()
	{
		LyEnvironment	env = getEnvironment();
		LyViewManager	viewManager = getViewManager();
		
		if ((env != null) && (viewManager != null))
		{
			if ((_imageStart != null) && !env.isImageCached(_imageStart))
				viewManager.freeImage(_imageStart);

			if ((_imageMiddle != null) && !env.isImageCached(_imageMiddle))
				viewManager.freeImage(_imageMiddle);

			if ((_imageEnd != null) && !env.isImageCached(_imageEnd))
				viewManager.freeImage(_imageEnd);
		}
		
		super.free();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#drawHtml(leon.view.web.struts.map.LyStrutsMapCanvas,
	 *      leon.view.web.LyWebPageWriter)
	 */
	public void drawHtml(LyStrutsMapCanvas map, LyWebPageWriter pw)
	{
		// Nothing to do for a link
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
	 */
	public void writeHtml(LyWebPageWriter pw)
	{
		// Nothing to do
	}

	/**
	 * Gets images used by this element and sets images class attributes with those new values.
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#getImages()
	 */
	public Image[] getImages()
	{
		Image imageStart = null;
		Image imageMiddle = null;
		Image imageEnd = null;

		if (_mapLink.getImageStart() != null)
		{
			if ((_imageStartString == null) || !_imageStartString.equals(_mapLink.getImageStart()))
			{
				_imageStartString = _mapLink.getImageStart();
				_imageStart = getImage(_mapLink.getImageStart());
			}

			imageStart = _imageStart;
		}

		if (_mapLink.getImageMiddle() != null)
		{
			if ((_imageMiddleString == null)
					|| !_imageMiddleString.equals(_mapLink.getImageMiddle()))
			{
				_imageMiddleString = _mapLink.getImageMiddle();
				_imageMiddle = getImage(_mapLink.getImageMiddle());
			}

			imageMiddle = _imageMiddle;
		}

		if (_mapLink.getImageEnd() != null)
		{
			if ((_imageEndString == null) || !_imageEndString.equals(_mapLink.getImageEnd()))
			{
				_imageEndString = _mapLink.getImageEnd();
				_imageEnd = getImage(_mapLink.getImageEnd());
			}

			imageEnd = _imageEnd;
		}

		return (new Image[] { imageStart, imageMiddle, imageEnd });
	} // end method getImages

	/**
	 * Gets the color of the start part of the link.
	 *
	 * @return the color of the start part of the link
	 */
	public Color getColorStart()
	{
		if (_mapLink.getColorStart() != null)
		{
			if ((_colorStartString != null) && _colorStartString.equals(_mapLink.getColorStart()))
				return _colorStart;

			_colorStartString = _mapLink.getColorStart();
			_colorStart = getColor(_mapLink.getColorStart());
		}
		else
		{
			if (_colorStartString != null)
				return _colorStart;

			_colorStart = getColor();
			_colorStartString = _colorString;
		}

		return _colorStart;
	}

	/**
	 * Gets the color of the middle part of the link.
	 *
	 * @return the color of the middle part of the link
	 */
	public Color getColor()
	{
		if (_mapLink.getColor() != null)
		{
			if ((_colorString != null) && _colorString.equals(_mapLink.getColor()))
				return _color;

			_colorString = _mapLink.getColor();
			_color = getColor(_mapLink.getColor());
		}
		else
		{
			if (_colorString != null)
				return _color;

			_colorString = new String("LY_MAP_DEFAULT_LINK_COLOR");
			_color = getColor("LY_MAP_DEFAULT_LINK_COLOR");
		}

		// Use default.
		if (_color == null)
			_color = getColor("LY_MAP_DEFAULT_LINK_COLOR");

		return _color;
	}

	/**
	 * Gets the color of the end part of the link.
	 *
	 * @return the color of the end part of the link
	 */
	public Color getColorEnd()
	{
		if (_mapLink.getColorEnd() != null)
		{
			if ((_colorEndString != null) && _colorEndString.equals(_mapLink.getColorEnd()))
				return _colorEnd;

			_colorEndString = _mapLink.getColorEnd();
			_colorEnd = getColor(_mapLink.getColorEnd());
		}
		else
		{
			if (_colorEndString != null)
				return _colorEnd;

			_colorEnd = getColor();
			_colorEndString = _colorString;
		}

		return _colorEnd;
	}

	/**
	 * Gets the color of a selected link.
	 *
	 * @return the colorof a selected link
	 */
	public Color getSelectedColor()
	{
		if (_selectedColor != null)
			return _selectedColor;

		_selectedColor = getColor("LY_MAP_SELECTED_COLOR");
		return _selectedColor;
	}

	/**
	 * Gets the font color of the link.
	 *
	 * @return the font color of the link
	 */
	public Color getFontColor()
	{
		if (_fontColor != null)
			return _fontColor;

		_fontColor = getColor("LY_MAP_FONT_COLOR");
		return _fontColor;
	}

	/**
	 * Gets the outline color of this map link.
	 *
	 * @return the outline color of the link
	 */
	public Color getOutlineColor()
	{
		if (_outlineColor != null)
			return _outlineColor;

		_outlineColor = getColor("LY_MAP_OUTLINE_COLOR");
		return _outlineColor;
	}

	/**
	 * Gets the dimension of the label part of the link.
	 *
	 * @param  g the graphics context
	 * @return the dimension of the label part of the link
	 */
	@SuppressWarnings("deprecation")
	public Dimension getLabelSize(Graphics g)
	{
		if ((_labelCache == null) || (_labelWidth < 0) || (_labelHeight < 0))
		{
			String label = getLabel();

			if (label == null)
				return new Dimension(0, 0);

			FontMetrics fm;

			if (g == null)
			{
				Toolkit toolkit = Toolkit.getDefaultToolkit();

				fm = toolkit.getFontMetrics(_graphicalFont);
			}
			else
				fm = g.getFontMetrics(_graphicalFont);

			_labelHeight = fm.getHeight();
			_labelWidth = fm.stringWidth(label);
		}

		return (new Dimension(_labelWidth, _labelHeight));
	}

	/**
	 * Returns the dimension of the label part of the link. If the {@link #_labelWidth} and {@link
	 * #_labelHeight} values of this link are not set (equal -1), the width and height are computed
	 * to the correct values for the label to be displayed.
	 *
	 * @param  g the graphics context
	 * @return the label size of the current node
	 */
	@SuppressWarnings("deprecation")
	public Dimension getStartLabelSize(Graphics g)
	{
		if ((_startLabelCache == null) || (_startLabelWidth < 0) || (_startLabelHeight < 0))
		{
			String label = getStartLabel();

			if (label == null)
				return new Dimension(0, 0);

			FontMetrics fm;

			if (g == null)
			{
				Toolkit toolkit = Toolkit.getDefaultToolkit();

				fm = toolkit.getFontMetrics(_graphicalFont);
			}
			else
				fm = g.getFontMetrics(_graphicalFont);

			_startLabelHeight = fm.getHeight();
			_startLabelWidth = fm.stringWidth(label);
		}

		return (new Dimension(_startLabelWidth, _startLabelHeight));
	}

	/**
	 * Returns the dimension of the label part of the link. If the {@link #_labelWidth} and {@link
	 * #_labelHeight} values of this link are not set (equal -1), the width and height are computed
	 * to the correct values for the label to be displayed.
	 *
	 * @param  g the graphics context
	 * @return the label size of the current node
	 */
	@SuppressWarnings("deprecation")
	public Dimension getEndLabelSize(Graphics g)
	{
		if ((_endLabelCache == null) || (_endLabelWidth < 0) || (_endLabelHeight < 0))
		{
			String label = getEndLabel();

			if (label == null)
				return new Dimension(0, 0);

			FontMetrics fm;

			if (g == null)
			{
				Toolkit toolkit = Toolkit.getDefaultToolkit();

				fm = toolkit.getFontMetrics(_graphicalFont);
			}
			else
				fm = g.getFontMetrics(_graphicalFont);

			_endLabelHeight = fm.getHeight();
			_endLabelWidth = fm.stringWidth(label);
		}

		return (new Dimension(_endLabelWidth, _endLabelHeight));
	}

	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#draw(leon.view.web.struts.map.LyStrutsMapCanvas,
	 *      java.awt.Graphics, int)
	 */
	public void draw(LyStrutsMapCanvas map, Graphics g, int mode)
	{
		if ((_startNode == null) || (_endNode == null))
			return;

		drawLink(map, g, mode);
		drawLabels(map, g, mode);
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#drawInViewer(leon.view.web.struts.map.LyStrutsMapCanvas,
	 *      java.awt.Graphics, int)
	 */
	public void drawInViewer(LyStrutsMapCanvas map, Graphics g)
	{
		if ((_startNode == null) || (_endNode == null))
			return;

		LyPoint start = _startNode.getCenter();
		LyPoint end = _endNode.getCenter();
		Point	startPoint = map.convert(start);
		Point	endPoint = map.convert(end);
		Color	color = getColor();
		
		startPoint = map.convertToViewer(startPoint);
		endPoint = map.convertToViewer(endPoint);
		g.setColor(color);
		g.drawLine(startPoint.x, startPoint.y, endPoint.x, endPoint.y);
	}

	/**
	 * Draws the label of this link on the specified map.
	 *
	 * @param map  the map containing the label to draw
	 * @param g    the graphics context
	 * @param mode the mode of the drawing, not used in this implementation
	 */
	public void drawLabels(LyStrutsMapCanvas map, Graphics g, int mode)
	{
		Dimension   labelDim = getLabelSize(g);
		Dimension   startLabelDim = getStartLabelSize(g);
		Dimension   endLabelDim = getEndLabelSize(g);
		String	    label = getLabel();
		String		startLabel = getStartLabel();
		String		endLabel = getEndLabel();
		
		LyComponent parent = _mapLink.getParent();
		boolean     fill = true;

		if (parent instanceof LyMapContent)
			parent = parent.getParent();

		if (((LyMap)parent).getSelectionDraw() == LyMapBase.SelectionDraw.BORDER)
			fill = false;

		drawLabel(label, labelDim, map, g, mode, fill, LabelAlignment.CENTER);
		drawLabel(startLabel, startLabelDim, map, g, mode, fill, LabelAlignment.START);
		drawLabel(endLabel, endLabelDim, map, g, mode, fill, LabelAlignment.END);
	} // end method drawLabels

	/**
	 * Draw a label
	 * @param label
	 * @param labelDim
	 * @param map
	 * @param g
	 * @param mode
	 * @param fill
	 * @param align
	 */
	private void drawLabel(String label, Dimension labelDim, LyStrutsMapCanvas map, Graphics g, 
			int mode, boolean fill, LabelAlignment align)
	{
		if ((label == null) || (label.trim().length() == 0))
			return;
		
		FontMetrics   fm = g.getFontMetrics(_graphicalFont);
		Point		  startPoint = map.convert(_startNode.getCenter());
		Point		  endPoint = map.convert(_endNode.getCenter());

		int			  labelX = 0;//(startPoint.x + endPoint.x) / 2;
		int			  labelY = 0;//(startPoint.y + endPoint.y) / 2;
				
		if ((align == null) || align.equals(LabelAlignment.CENTER))
		{
			if (_mapLink.getPointCount() >= 1)
		    {
		        int     i, np = _mapLink.getPointCount();
		        int     middleX = (startPoint.x + endPoint.x) / 2;
		        Point   start = startPoint;
		        Point   end;
		        
	            for (i = 0; i < np; i++)
                {
	                end = map.convert(_mapLink.getPoint(i));
	                
	                if ((middleX >= start.x) && (middleX <= end.x))
	                {
	                    startPoint = start;
	                    endPoint = end;
	                }
	                else
	                    start = end;
                }
		    }

			labelX = (startPoint.x + endPoint.x) / 2 - (labelDim.width / 2);
			labelY = (startPoint.y + endPoint.y) / 2;
		}
		else if (align.equals(LabelAlignment.START))
		{
			Dimension dimStart = _startNode.getImageSize(map);
			
			double    radius = Math.sqrt((dimStart.width * dimStart.width)
					+ (dimStart.height
							* dimStart.height)) / 2;
			if (startPoint.x < endPoint.x)
				labelX = startPoint.x + (int)radius;
			else
				labelX = startPoint.x - labelDim.width - (int)radius;
			labelY = startPoint.y;
		}
		else if (align.equals(LabelAlignment.END))
		{
			Dimension dimEnd = _endNode.getImageSize(map);
			
			double    radius = Math.sqrt((dimEnd.width * dimEnd.width)
					+ (dimEnd.height
							* dimEnd.height)) / 2;
			if (startPoint.x < endPoint.x)
				labelX = endPoint.x - labelDim.width - (int)radius;
			else
				labelX = endPoint.x + (int)radius;
			labelY = endPoint.y;
		}

		// If selected, draw a rectangle around the label
		if (_mapLink.isSelected())
			g.setColor(getSelectedColor());
		else
			g.setColor(map.getBackground());

		int offset = (((_mapLink.getWidth() % 2) == 0) ? 0 : 1);

		if (fill)
			g.fillRect(labelX,
					labelY + ((_mapLink.getWidth() + SELECTION_OFFSET) / 2)
				+ offset, labelDim.width, labelDim.height);
		else if (_mapLink.isSelected())
			g.drawRect(labelX - 2,
					labelY + ((_mapLink.getWidth() + SELECTION_OFFSET) / 2)
				+ offset, labelDim.width + 4, labelDim.height);

		// Draw the label
		String labelColor = _mapLink.getFontColor();
		if (labelColor == null)
			g.setColor(getFontColor());
		else
			g.setColor(getColor(labelColor));
		g.setFont(_graphicalFont);
		g.drawString(label, labelX,
			labelY + fm.getAscent()
			+ ((_mapLink.getWidth() + SELECTION_OFFSET) / 2) + offset);
	}
	
	/**
	 * Draws the link.
	 *
	 * @param map  the map of the link
	 * @param g    the graphics context
	 * @param mode the drawing mode (can be {@link LyStrutsMapElement#DRAW_MODE_NORMAL} or {@link
	 *             LyStrutsMapElement#DRAW_MODE_REDUCED})
	 */
	public void drawLink(LyStrutsMapCanvas map, Graphics g, int mode)
	{
		if ((_startNode == null) || (_endNode == null))
			return;

		boolean isNormalMode = (mode == LyStrutsMapElement.DRAW_MODE_NORMAL);

		if (!isNormalMode)
			return;

		Point startPoint, endPoint;
		Color colorStart, colorMiddle, colorEnd;

		colorStart = getColorStart();
		colorMiddle = getColor();
		colorEnd = getColorEnd();
		startPoint = map.convert(_startNode.getCenter());
		endPoint = map.convert(_endNode.getCenter());

		int		    x1 = startPoint.x;
		int		    y1 = startPoint.y;
		int		    x2 = endPoint.x;
		int		    y2 = endPoint.y;

		// Divide the link in three parts
		int		    firstSectionX = ((2 * x1) + x2) / 3;
		int		    firstSectionY = ((2 * y1) + y2) / 3;
		int		    secondSectionX = (x1 + (2 * x2)) / 3;
		int		    secondSectionY = (y1 + (2 * y2)) / 3;

		// Modify sections if there are intermediate points (sub-component)
		if (_mapLink.getPointCount() >= 1)
		{
			firstSectionX = map.convert(_mapLink.getPoint(0)).x;
			firstSectionY = map.convert(_mapLink.getPoint(0)).y;

			if (_mapLink.getPointCount() >= 2)
			{
				int lastPoint = _mapLink.getPointCount() - 1;

				secondSectionX = map.convert(_mapLink.getPoint(lastPoint)).x;
				secondSectionY = map.convert(_mapLink.getPoint(lastPoint)).y;
			}
			else
			{
				secondSectionX = ((2 * firstSectionX) + x2) / 3;
				secondSectionY = ((2 * firstSectionY) + y2) / 3;
			}
		}

		// Check whether three different parts must be generated or only the last one
		boolean	onlyDrawLastPart = true;
		
		onlyDrawLastPart &= (_mapLink.getPointCount() == 0);
		onlyDrawLastPart &= (_mapLink.getColorStart() == null) && (_mapLink.getColorEnd() == null);
		onlyDrawLastPart &= (_mapLink.getArrow() != LyMapLink.Arrow.START) && (_mapLink.getArrow() != LyMapLink.Arrow.BOTH);
		
		if (onlyDrawLastPart)
		{
			secondSectionX = x1;
			secondSectionY = y1;
		}
		
		LyComponent parent = _mapLink.getParent();
		boolean     fill = true;

		if (parent instanceof LyMapContent)
			parent = parent.getParent();

		if (((LyMap)parent).getSelectionDraw() == LyMap.SelectionDraw.BORDER)
			fill = false;

		// If selected, draw a polygon under the tube
		if (_mapLink.isSelected())
		{
			g.setColor(getSelectedColor());
			Point tmp = new Point();

			tmp.x = x1;
			tmp.y = y1;
			// If there are intermediate points
			for (int i = 0; i < _mapLink.getPointCount(); i++)
			{
				Point aPoint = map.convert(_mapLink.getPoint(i));
			
				drawLine(g, tmp.x, tmp.y, aPoint.x, aPoint.y, _mapLink.getWidth()
						+ SELECTION_OFFSET, LyMapLink.Style.LINE, !fill, false);
				tmp.x = aPoint.x;
				tmp.y = aPoint.y;
			}
			drawLine(g, tmp.x, tmp.y, x2, y2, _mapLink.getWidth()
					+ SELECTION_OFFSET, LyMapLink.Style.LINE, !fill, false);
		}

		if (_mapLink.getMotif() != null)
		{
			Image image = getImage(_mapLink.getMotif());
			int   length = Math.max(image.getWidth(null), image.getHeight(null));
			Point tmp = new Point();

			tmp.x = x1;
			tmp.y = y1;
			// If there are intermediate points
			for (int i = 0; i < _mapLink.getPointCount(); i++)
			{
				Point aPoint = map.convert(_mapLink.getPoint(i));
			
				drawDashedLine(g, tmp.x, tmp.y, aPoint.x, aPoint.y, length / 2, image);
				tmp.x = aPoint.x;
				tmp.y = aPoint.y;
			}
			drawDashedLine(g, tmp.x, tmp.y, x2, y2, length / 2, image);
		}
		else
		{
			if (!onlyDrawLastPart)
			{
				// Draw the first part
				if ((_mapLink.getArrow() == LyMapLink.Arrow.START)
						|| (_mapLink.getArrow() == LyMapLink.Arrow.BOTH))
				{
					boolean fillArrow = (_mapLink.getStartArrowFillMode() == LyMapLink.ArrowFill.FULL);
					Dimension dimStart = _startNode.getImageSize(map);
					double    startRadius = Math.sqrt((dimStart.width * dimStart.width)
							+ (dimStart.height
									* dimStart.height)) / 2;
					double    distance = Math.sqrt(((x2 - x1) * (x2 - x1)) + ((y2 - y1) * (y2 - y1)));
					int		  xArrow = (int)((((distance - startRadius) * x1) + (startRadius * x2))
								/ distance);
					int		  yArrow = (int)((((distance - startRadius) * y1) + (startRadius * y2))
								/ distance);
	
					firstSectionX = ((2 * xArrow) + x2) / 3;
					firstSectionY = ((2 * yArrow) + y2) / 3;
	
					// If the points are superposed, do nothing
					if (distance != 0)
					{
						g.setColor(colorStart);
						if (_mapLink.getStartArrowType() == LyMapLink.ArrowType.CIRCLE)
							drawCircle(g, xArrow, yArrow, firstSectionX, firstSectionY, _mapLink.getWidth(),
									_mapLink.getStyle(), _mapLink.isDoubleLink(), true, fillArrow);
						else if (_mapLink.getStartArrowType() == LyMapLink.ArrowType.SQUARE)
							drawSquare(g, xArrow, yArrow, firstSectionX, firstSectionY, _mapLink.getWidth(),
									_mapLink.getStyle(), _mapLink.isDoubleLink(), true, fillArrow);
						else
							drawArrow(g, xArrow, yArrow, firstSectionX, firstSectionY, _mapLink.getWidth(),
									_mapLink.getStyle(), _mapLink.isDoubleLink(), true, fillArrow);
					}
				}
				else
				{
					g.setColor(colorStart);
					
					if (_mapLink.getStyle() == LyMapLink.Style.TUBE)
						drawTube(g, x1, y1, firstSectionX, firstSectionY, _mapLink.getWidth(), colorStart);
					else
						drawLine(g, x1, y1, firstSectionX, firstSectionY, _mapLink.getWidth(),
								_mapLink.getStyle(), _mapLink.isDoubleLink(), true);
				}
			}
			
			// Draw the third part
			g.setColor(colorEnd);

			if ((_mapLink.getArrow() == LyMapLink.Arrow.END)
					|| (_mapLink.getArrow() == LyMapLink.Arrow.BOTH))
			{
				boolean fillArrow = (_mapLink.getEndArrowFillMode() == LyMapLink.ArrowFill.FULL);
				Dimension dimEnd = _endNode.getImageSize(map);
				double    endRadius = Math.sqrt((dimEnd.width * dimEnd.width)
						+ (dimEnd.height
								* dimEnd.height)) / 2;
				double    distance = Math.sqrt(((x2 - x1) * (x2 - x1)) + ((y2 - y1) * (y2 - y1)));
				int		  xArrow = (int)((((distance - endRadius) * x2) + (endRadius * x1))
							/ distance);
				int		  yArrow = (int)((((distance - endRadius) * y2) + (endRadius * y1))
							/ distance);

				//secondSectionX = (x1 + (2 * xArrow)) / 3;
				//secondSectionY = (y1 + (2 * yArrow)) / 3;

				// If the points are superposed, do nothing
				if (distance != 0)
				{
					g.setColor(colorEnd);
					
					if (_mapLink.getEndArrowType() == LyMapLink.ArrowType.CIRCLE)
						drawCircle(g, xArrow, yArrow, secondSectionX, secondSectionY,
								_mapLink.getWidth(), _mapLink.getStyle(), _mapLink.isDoubleLink(), true, fillArrow);	
					else if (_mapLink.getEndArrowType() == LyMapLink.ArrowType.SQUARE)
						drawSquare(g, xArrow, yArrow, secondSectionX, secondSectionY,
							_mapLink.getWidth(), _mapLink.getStyle(), _mapLink.isDoubleLink(), true, fillArrow);		
					else
						drawArrow(g, xArrow, yArrow, secondSectionX, secondSectionY,
								_mapLink.getWidth(), _mapLink.getStyle(), _mapLink.isDoubleLink(), true, fillArrow);
				}
			}
			else
			{
				g.setColor(colorEnd);
				
				if (_mapLink.getStyle() == LyMapLink.Style.TUBE)
					drawTube(g, secondSectionX, secondSectionY, x2, y2, _mapLink.getWidth(), colorEnd);
				else
					drawLine(g, secondSectionX, secondSectionY, x2, y2, _mapLink.getWidth(),
							_mapLink.getStyle(), _mapLink.isDoubleLink(), true);
			}

			if (!onlyDrawLastPart)
			{
				// Draw the middle part
				g.setColor(colorMiddle);
				
				Point	tmp = new Point();
				
	            tmp.x = firstSectionX;
	            tmp.y = firstSectionY;
	            // If there are intermediate points
	            for (int i = 1; i < (_mapLink.getPointCount() - 1); i++) {
	              Point aPoint = map.convert(_mapLink.getPoint(i));

	              drawLine(g, tmp.x, tmp.y, aPoint.x, aPoint.y, _mapLink.getWidth(),
	                  _mapLink.getStyle(), _mapLink.isDoubleLink(), true);
	              tmp.x = aPoint.x;
	              tmp.y = aPoint.y;
	            }
				
				if (_mapLink.getStyle() == LyMapLink.Style.TUBE)
					drawTube(g, firstSectionX, firstSectionY, secondSectionX, secondSectionY, _mapLink.getWidth(), colorMiddle);
				else
					drawLine(g, firstSectionX, firstSectionY, secondSectionX, secondSectionY,
							_mapLink.getWidth(), _mapLink.getStyle(), _mapLink.isDoubleLink(), true);
			}
		}

		// Draw the images
		int     width, height;
		int     xImage, yImage;
		Image[] images = getImages();
		Point   imageStartPos = null;
		Point   imageMiddlePos = null;
		Point   imageEndPos = null;

		if (images != null)
		{
			Image imageStart = images[0];
			Image imageMiddle = images[1];
			Image imageEnd = images[2];

			if (imageStart != null)
			{
				width = imageStart.getWidth(map);
				height = imageStart.getHeight(map);
				xImage = (x1 + firstSectionX) / 2;
				yImage = (y1 + firstSectionY) / 2;

				imageStartPos = new Point(xImage, yImage);

				g.drawImage(imageStart, xImage - (width / 2), yImage - (height / 2), width, height,
					map);
			}

			if (imageMiddle != null)
			{
				width = imageMiddle.getWidth(map);
				height = imageMiddle.getHeight(map);
				xImage = (secondSectionX + firstSectionX) / 2;
				yImage = (secondSectionY + firstSectionY) / 2;

				imageMiddlePos = new Point(xImage, yImage);

				g.drawImage(imageMiddle, xImage - (width / 2), yImage - (height / 2), width, height,
					map);
			}

			if (imageEnd != null)
			{
				width = imageEnd.getWidth(map);
				height = imageEnd.getHeight(map);
				xImage = (secondSectionX + x2) / 2;
				yImage = (secondSectionY + y2) / 2;

				imageEndPos = new Point(xImage, yImage);

				g.drawImage(imageEnd, xImage - (width / 2), yImage - (height / 2), width, height,
					map);
			}

			_imagePos = new Point[] { imageStartPos, imageMiddlePos, imageEndPos };
		}
	} // end method drawLink

	/**
	 * Gets the checkbox position from the start node, end node and the middle image.
	 *
	 * @param  map    the map of the link
	 * @param  width  the width of the checkbox
	 * @param  height the height of the checkbox
	 * @return the checkbox position
	 */
	protected Point getCheckBoxPosition(LyStrutsMapCanvas map, int width, int height)
	{
		if ((_startNode == null) || (_endNode == null))
			return null;

		Point   startPoint = map.convert(_startNode.getCenter());
		Point   endPoint = map.convert(_endNode.getCenter());
		int     x1 = startPoint.x;
		int     y1 = startPoint.y;
		int     x2 = endPoint.x;
		int     y2 = endPoint.y;
		int     middleX = (x1 + x2) / 2;
		int     middleY = (y1 + y2) / 2;
		Image[] images = getImages();

		if ((images != null) && (images[1] != null))
		{
			Image imageMiddle = images[1];
			int   imageWidth = imageMiddle.getWidth(map);
			int   x = middleX - (imageWidth / 2) - width - 3;

			return new Point(x, middleY - (height / 2));
		}
		else
		{
			// Compute vector perpendicular to link
			int    linkVectorY = -x2 + x1;
			int    linkVectorX = y2 - y1;

			// Normalize vector
			double length = Math.sqrt((linkVectorX * linkVectorX) + (linkVectorY * linkVectorY));
			double normVectorX = (_mapLink.getWidth()) * linkVectorX / length;
			double normVectorY = (_mapLink.getWidth()) * linkVectorY / length;
			int    x = (int)(middleX + normVectorX - (width / 2));
			int    y = (int)(middleY + normVectorY - (height / 2));

			return new Point(x, y);
		}
	} // end method getCheckBoxPosition

	/**
	 * Draws a line for this link.
	 *
	 * @param g       the graphics context
	 * @param x1      starting x coordonate of the line
	 * @param y1      starting y coordonate of the line
	 * @param x2      ending x coordonate of the line
	 * @param y2      ending y coordonate of the line
	 * @param width   width of the line
	 * @param style   style of the line
	 * @param dble    indicates a double line
	 * @param outline outlines the line
	 */
	protected void drawLine(Graphics g, int x1, int y1, int x2, int y2, int width,
        LyMapLink.Style style, boolean dble, boolean outline)
	{
		// Keep the color in memory
		Color color = g.getColor();

		if (style == LyMapLink.Style.LINE)
		{
			// For a linear wide line
			if (width > 1)
			{
				// Calculate the polygon containing the link
				Point[] polygon = getPolygon(x1, y1, x2, y2, width);

				if (dble)
				{
					g.drawLine(polygon[0].x, polygon[0].y, polygon[1].x, polygon[1].y);
					g.drawLine(polygon[2].x, polygon[2].y, polygon[3].x, polygon[3].y);
				}
				else
				{
					int[] xPoints = new int[]
						{
							polygon[0].x,
							polygon[1].x,
							polygon[2].x,
							polygon[3].x
						};
					int[] yPoints = new int[]
						{
							polygon[0].y,
							polygon[1].y,
							polygon[2].y,
							polygon[3].y
						};

					g.fillPolygon(xPoints, yPoints, 4);

					if (outline)
					{
						g.setColor(getOutlineColor());
						g.drawLine(polygon[0].x, polygon[0].y, polygon[1].x, polygon[1].y);
						g.drawLine(polygon[2].x, polygon[2].y, polygon[3].x, polygon[3].y);
						g.setColor(color);
					}
				}
			}
			else
				g.drawLine(x1, y1, x2, y2);
		}
		// If dot or dashed line, draw a dashed line with the corresponding
		// length for segments.
		else
		{
			int length = 1;

			if (style == LyMapLink.Style.DASHED)
				length = 5;

			if (dble)
			{
				Point[] polygon = getPolygon(x1, y1, x2, y2, width);

				drawDashedLine(g, polygon[0].x, polygon[0].y, polygon[1].x, polygon[1].y, length);
				drawDashedLine(g, polygon[2].x, polygon[2].y, polygon[3].x, polygon[3].y, length);
			}
			else
				drawDashedLine(g, x1, y1, x2, y2, length);
		}
	} // end method drawLine

	/**
	 * Draws a tube-style line.
	 *
	 * @param g     the graphics context
	 * @param x1    starting x coordonate of the tube-style line
	 * @param y1    starting y coordonate of the tube-style line
	 * @param x2    ending x coordonate of the tube-style line
	 * @param y2    ending y coordonate of the tube-style line
	 * @param width width of the tube-style line
	 * @param color the color of the tube-style line
	 */
	protected void drawTube(Graphics g, int x1, int y1, int x2, int y2, int width, Color color)
	{
		if (width <= 0)
			return;

		int[] xPoints;
		int[] yPoints;
		int   red = color.getRed(), green = color.getGreen(), blue = color.getBlue();
		int   w, offset = 400 / width;

		for (w = width; w > 0; w = w - 2)
		{
			// Calculate the polygon containing the link
			Point[] polygon = getPolygon(x1, y1, x2, y2, w);

			xPoints = new int[] { polygon[0].x, polygon[1].x, polygon[2].x, polygon[3].x };
			yPoints = new int[] { polygon[0].y, polygon[1].y, polygon[2].y, polygon[3].y };

			g.setColor(color);
			g.fillPolygon(xPoints, yPoints, 4);

			red = ((red > (255 - offset)) ? 255 : (red + offset));
			green = ((green > (255 - offset)) ? 255 : (green + offset));
			blue = ((blue > (255 - offset)) ? 255 : (blue + offset));

			color = new Color(red, green, blue);
		}
	} // end method drawTube

	/**
	 * Gets the polygon where (x1,y1) and (x2,y2) are center of two opposite sides of the given
	 * width.
	 *
	 * @param  x1    x coordonate of the first side center
	 * @param  y1    y coordonate of the first side center
	 * @param  x2    x coordonate of the second side center
	 * @param  y2    y coordonate of the second side center
	 * @param  width width of the polygon
	 * @return the polygon
	 */
	protected Point[] getPolygon(int x1, int y1, int x2, int y2, int width)
	{
		double alpha;
		
		if (x2 < x1)
		{
			int	xTmp = x2;
			int	yTmp = y2;
			
			x2 = x1;
			x1 = xTmp;
			y2 = y1;
			y1 = yTmp;
		}

		if (y2 == y1)
			alpha = (Math.PI) / 2;
		else
			alpha = Math.atan(((float)(x2 - x1)) / (y2 - y1));

		width = width + 1;

		int   dx = (int)((float)(width * Math.cos(alpha))) / 2;
		int   dy = -(int)((float)(width * Math.sin(alpha))) / 2;
		
		if (dy > 0)
		{
			dx = -dx;
			dy = -dy;
		}

		Point point1 = new Point(x1 - dx, y1 - dy);
		Point point2 = new Point(x2 - dx, y2 - dy);
		Point point3 = new Point(x2 + dx, y2 + dy);
		Point point4 = new Point(x1 + dx, y1 + dy);

		return new Point[] { point1, point2, point3, point4 };
	}

	/**
	 * Draws an arrow.
	 *
	 * @param g       the graphics context
	 * @param x1      starting x coordonate of the arrow
	 * @param y1      starting y coordonate of the arrow
	 * @param x2      ending x coordonate of the arrow
	 * @param y2      ending y coordonate of the arrow
	 * @param width   width of the arrow
	 * @param style   style of the arrow
	 * @param dble    indicates a double arrow
	 * @param outline outlines the arrow
	 * @param fill	  fills the arrow
	 */
	protected void drawArrow(Graphics g, int x1, int y1, int x2, int y2, int width,
        LyMapLink.Style style, boolean dble, boolean outline, boolean fill)
	{
		// Keep the color in memory
		Color	color = g.getColor();
		Point[]	polygon = getPolygon(x1, y1, x2, y2, width + 5);

		double	lineLength = Math.sqrt((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1));
		double	ratio = lineLength / 10;
		
		// Calculate the base of the pointer
		int xc = (int)((ratio * x1 + x2) / (ratio + 1));
		int yc = (int)((ratio * y1 + y2) / (ratio + 1));

		// Calculate the 2 other extremities of the pointer
		int xa = xc + (int)((polygon[0].x - x1) * 1.5);
		int ya = yc + (int)((polygon[0].y - y1) * 1.5);
		int xb = xc + (int)((polygon[3].x - x1) * 1.5);
		int yb = yc + (int)((polygon[3].y - y1) * 1.5);

		// Get the polygon that represents the body of the arrow
		Point[] polygon2 = getPolygon(xc, yc, x2, y2, width);
		int[]   xPoints;
		int[]   yPoints;

		if (style == LyMapLink.Style.LINE)
		{
			if (dble)
			{
				// Draw the pointer
				g.drawLine(x1, y1, xa, ya);
				g.drawLine(x1, y1, xb, yb);

				g.drawLine(polygon2[0].x, polygon2[0].y, xa, ya);
				g.drawLine(polygon2[3].x, polygon2[3].y, xb, yb);

				// Draw the body
				g.drawLine(polygon2[0].x, polygon2[0].y, polygon2[1].x, polygon2[1].y);
				g.drawLine(polygon2[2].x, polygon2[2].y, polygon2[3].x, polygon2[3].y);
			}
			else
			{
				// Draw the pointer
				xPoints = new int[] { x1, xa, xb };
				yPoints = new int[] { y1, ya, yb };

				if (fill)
					g.fillPolygon(xPoints, yPoints, 3);

				if (outline)
				{
					g.setColor(getOutlineColor());

					// Draw the pointer
					g.drawLine(x1, y1, xa, ya);
					g.drawLine(x1, y1, xb, yb);

					g.drawLine(polygon2[0].x, polygon2[0].y, xa, ya);
					g.drawLine(polygon2[3].x, polygon2[3].y, xb, yb);

					// Draw the body
					g.drawLine(polygon2[0].x, polygon2[0].y, polygon2[1].x, polygon2[1].y);
					g.drawLine(polygon2[2].x, polygon2[2].y, polygon2[3].x, polygon2[3].y);

					g.setColor(color);
				}

				// Draw the body
				if (width > 1)
				{
					xPoints = new int[]
						{
							polygon2[0].x,
							polygon2[1].x,
							polygon2[2].x,
							polygon2[3].x
						};
					yPoints = new int[]
						{
							polygon2[0].y,
							polygon2[1].y,
							polygon2[2].y,
							polygon2[3].y
						};
					
					if (fill)
						g.fillPolygon(xPoints, yPoints, 4);

					if (outline)
					{
						g.setColor(getOutlineColor());

						// Draw the pointer
						g.drawLine(x1, y1, xa, ya);
						g.drawLine(x1, y1, xb, yb);

						g.drawLine(polygon2[0].x, polygon2[0].y, xa, ya);
						g.drawLine(polygon2[3].x, polygon2[3].y, xb, yb);

						// Draw the body
						g.drawLine(polygon2[0].x, polygon2[0].y, polygon2[1].x, polygon2[1].y);
						g.drawLine(polygon2[2].x, polygon2[2].y, polygon2[3].x, polygon2[3].y);

						g.setColor(color);
					}
				}
				else
					g.drawLine(xc, yc, x2, y2);
			}
		}
		else
		{
			int length = 1;

			if (style == LyMapLink.Style.DASHED)
				length = 4;

			if (dble)
			{
				// Draw the pointer
				drawDashedLine(g, x1, y1, xa, ya, length);
				drawDashedLine(g, x1, y1, xb, yb, length);

				drawDashedLine(g, polygon2[0].x, polygon2[0].y, xa, ya, length);
				drawDashedLine(g, polygon2[3].x, polygon2[3].y, xb, yb, length);

				// Draw the body
				drawDashedLine(g, polygon2[0].x, polygon2[0].y, polygon2[1].x, polygon2[1].y,
					length);
				drawDashedLine(g, polygon2[2].x, polygon2[2].y, polygon2[3].x, polygon2[3].y,
					length);
			}
			else
			{
				xPoints = new int[] { x1, xa, xb };
				yPoints = new int[] { y1, ya, yb };

				// Draw the pointer
				if (fill)
					g.fillPolygon(xPoints, yPoints, 3);

				// Draw the body
				drawDashedLine(g, xc, yc, x2, y2, length);
			}
		}
	} // end method drawArrow

	/**
	 * Draws a circle from the (x1,y1) coordinate to the (x2,y2) coordinate with the specified
	 * properties.
	 *
	 * @param g       the graphics context
	 * @param x1      starting x coordonate of the circle
	 * @param y1      starting y coordonate of the circle
	 * @param x2      ending x coordonate of the circle
	 * @param y2      ending y coordonate of the circle
	 * @param width   width of the circle
	 * @param style   style of the circle
	 * @param dble    indicates a double circle
	 * @param outline outlines the circle
	 * @param fill	  fills the circle
	 */
	protected void drawCircle(Graphics g, int x1, int y1, int x2, int y2, int width,
	        LyMapLink.Style style, boolean dble, boolean outline, boolean fill)
	{
		// Keep the color in memory
		Color		  color = g.getColor();
		double		  lineLength = Math.sqrt((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1));
		double		  ratio = lineLength / 10;
		
		// Calculate the base of the pointer
		int			  xc = (int)((ratio * x1 + x2) / (ratio + 1));
		int 		  yc = (int)((ratio * y1 + y2) / (ratio + 1));
		// Get the polygon that represents the body of the arrow
		Point[]		  polygon2 = LySwingUtils.getPolygon(xc, yc, x2, y2, width);
		
		int x = (x1 + polygon2[0].x) / 2;
		int y = (y1 + polygon2[0].y) / 2;
		
		// compute center
		int xWidth = Math.abs((x1 - polygon2[0].x) / 2);
		int yWidth = Math.abs((y1 - polygon2[0].y) / 2);
		int rayon = (int)Math.sqrt(xWidth * xWidth + yWidth * yWidth) + 1;

		x = x - rayon;
		y = y - rayon;
		if (fill)
			g.fillOval( x , y, 2 * rayon, 2 * rayon);
		if (outline)
		{
			g.setColor(getOutlineColor());
			g.drawOval( x , y, 2 * rayon, 2 * rayon);
			g.setColor(color);
		}
		g.drawLine(xc, yc, x2, y2);
		
	}
	
	/**
	 * Draws a square from the (x1,y1) coordinate to the (x2,y2) coordinate with the specified
	 * properties.
	 *
	 * @param g       the graphics context
	 * @param x1      starting x coordonate of the square
	 * @param y1      starting y coordonate of the square
	 * @param x2      ending x coordonate of the square
	 * @param y2      ending y coordonate of the square
	 * @param width   width of the square
	 * @param style   style of the square
	 * @param dble    indicates a double square
	 * @param outline outlines the square
	 * @param fill	  fills the square
	 */
	protected void drawSquare(Graphics g, int x1, int y1, int x2, int y2, int width,
	        LyMapLink.Style style, boolean dble, boolean outline, boolean fill)
	{
		Color		  color = g.getColor();
		Point[]		  polygon = LySwingUtils.getPolygon(x1, y1, x2, y2, width + 5);
		double		  lineLength = Math.sqrt((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1));
		double		  ratio = lineLength / 10;
		
		// Calculate the base of the pointer
		int			  xc = (int)((ratio * x1 + x2) / (ratio + 1));
		int 		  yc = (int)((ratio * y1 + y2) / (ratio + 1));

		// Calculate the 2 other extremities of the pointer
		int 		  xa = xc + (int)((polygon[0].x - x1) * 1.5);
		int 		  ya = yc + (int)((polygon[0].y - y1) * 1.5);
		int 		  xb = xc + (int)((polygon[3].x - x1) * 1.5);
		int 		  yb = yc + (int)((polygon[3].y - y1) * 1.5);

		// Get the polygon that represents the body of the arrow
		Point[]		  polygon2 = LySwingUtils.getPolygon(xc, yc, x2, y2, width);
		
		// compute center
		int xWidth = (x1 - polygon2[0].x) / 2;
		int yWidth = (y1 - polygon2[0].y) / 2;
		
		Polygon p = new Polygon();
		p.addPoint(x1, y1);
		p.addPoint(xa + xWidth, ya + yWidth);
		p.addPoint(xc, yc);
		p.addPoint(xb + xWidth, yb + yWidth);
		
		if (fill)
			g.fillPolygon(p);
		if (outline)
		{
			g.setColor(getOutlineColor());
			g.drawPolygon(p);
			g.setColor(color);
		}
		
		g.drawLine(xc, yc, x2, y2);
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.getinterface.LyGetIdInterface#getId()
	 */
	public String getId()
	{
		return _mapLink.getId();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#getBoundingBox(leon.view.web.struts.map.LyStrutsMapCanvas)
	 */
	public Ly3DBox getBoundingBox(LyStrutsMapCanvas map)
	{
		LyPoint start3D = new LyPoint();
		LyPoint end3D = new LyPoint();

		if (_startNode != null)
			start3D = _startNode.getCenter();
		if (_endNode != null)
			end3D = _endNode.getCenter();

		return new Ly3DBox(start3D, end3D);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#getBoundingRectangle(leon.view.web.struts.map.LyStrutsMapCanvas)
	 */
	public Rectangle getBoundingRectangle(LyStrutsMapCanvas map)
	{
		LyPoint start3D = new LyPoint();
		LyPoint end3D = new LyPoint();

		if (_startNode != null)
			start3D = _startNode.getCenter();
		if (_endNode != null)
			end3D = _endNode.getCenter();

		Point start = map.convert(start3D);
		Point end = map.convert(end3D);
		int   x1, y1;

		x1 = ((end.x < start.x) ? end.x : start.x);
		y1 = ((end.y < start.y) ? end.y : start.y);

		return new Rectangle(x1, y1, Math.abs(end.x - start.x), Math.abs(end.y - start.y));
	}

	/**
	 * Gets the bounding shape of the link.
	 *
	 * @param  map the map of the link
	 * @return the bounding shape of the link
	 */
	public Polygon getBoundingShape(LyStrutsMapCanvas map)
	{
		LyPoint start3D = new LyPoint();
		LyPoint end3D = new LyPoint();

		if (_startNode != null)
			start3D = _startNode.getCenter();
		if (_endNode != null)
			end3D = _endNode.getCenter();

		Point   start = map.convert(start3D);
		Point   end = map.convert(end3D);
		Polygon polygon = new Polygon();
		
		Rectangle	boundingRect = map.getBoundingRectangle();

		if (_mapLink.getPointCount() > 0)
		{
			Point[] poly = getPolygon(start.x, start.y, map.convert(_mapLink.getPoint(0)).x, map.convert(_mapLink.getPoint(0)).y, _mapLink.getWidth() + 4);

			polygon.addPoint(poly[3].x - boundingRect.x, poly[3].y - boundingRect.y);
			polygon.addPoint(poly[0].x - boundingRect.x, poly[0].y - boundingRect.y);

			for (int i = 0; i < _mapLink.getPointCount() - 1; i++)
			{
				poly = getPolygon(map.convert(_mapLink.getPoint(i)).x, map.convert(_mapLink.getPoint(i)).y,
						map.convert(_mapLink.getPoint(i + 1)).x, map.convert(_mapLink.getPoint(i + 1)).y, _mapLink.getWidth() + 4);
				polygon.addPoint(poly[0].x - boundingRect.x, poly[0].y - boundingRect.y);
				polygon.addPoint(poly[1].x - boundingRect.x, poly[1].y - boundingRect.y);
			}
			
			poly = getPolygon(map.convert(_mapLink.getPoint(_mapLink.getPointCount() - 1)).x, map.convert(_mapLink.getPoint(_mapLink.getPointCount() - 1)).y,
					end.x, end.y, _mapLink.getWidth() + 4);
			polygon.addPoint(poly[0].x - boundingRect.x, poly[0].y - boundingRect.y);
			polygon.addPoint(poly[1].x - boundingRect.x, poly[1].y - boundingRect.y);
			polygon.addPoint(poly[2].x - boundingRect.x, poly[2].y - boundingRect.y);
			polygon.addPoint(poly[3].x - boundingRect.x, poly[3].y - boundingRect.y);

			for (int i = _mapLink.getPointCount() - 2; i >= 0; i--)
			{
				poly = getPolygon(map.convert(_mapLink.getPoint(i)).x, map.convert(_mapLink.getPoint(i)).y,
						map.convert(_mapLink.getPoint(i + 1)).x, map.convert(_mapLink.getPoint(i + 1)).y, _mapLink.getWidth() + 4);
				polygon.addPoint(poly[2].x - boundingRect.x, poly[2].y - boundingRect.y);
				polygon.addPoint(poly[3].x - boundingRect.x, poly[3].y - boundingRect.y);
			}
		}
		else
		{
			Point[] poly = getPolygon(start.x, start.y, end.x, end.y, _mapLink.getWidth() + 4);
			int     i, n = poly.length;
		
			for (i = 0; i < n; i++)
				polygon.addPoint(poly[i].x - boundingRect.x, poly[i].y - boundingRect.y);
		}
		
		return polygon;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#getClippingArea(leon.view.web.struts.map.LyStrutsMapCanvas,
	 *      java.awt.Image)
	 */
	public Shape getClippingArea(LyStrutsMapCanvas map, Image image)
	{
		Image[] images = getImages();

		if ((image != null) && (images != null))
		{
			int   xImage;
			int   yImage;
			int   width;
			int   height;
			Image imageStart = images[0];
			Image imageMiddle = images[1];
			Image imageEnd = images[2];
			Point imageStartPos = _imagePos[0];
			Point imageMiddlePos = _imagePos[1];
			Point imageEndPos = _imagePos[2];

			if ((imageStart != null) && (image == imageStart) && (imageStartPos != null))
			{
				width = imageStart.getWidth(map);
				height = imageStart.getHeight(map);
				xImage = _imagePos[0].x;
				yImage = _imagePos[0].y;

				return new Rectangle(xImage - 1 - (width / 2), yImage - 1 - (height / 2), width + 2,
					height + 2);
			}

			if ((imageMiddle != null) && (image == imageMiddle) && (imageMiddlePos != null))
			{
				width = imageMiddle.getWidth(map);
				height = imageMiddle.getHeight(map);
				xImage = _imagePos[1].x;
				yImage = _imagePos[1].y;

				return new Rectangle(xImage - 1 - (width / 2), yImage - 1 - (height / 2), width + 2,
					height + 2);
			}

			if ((imageEnd != null) && (image == imageEnd) && (imageEndPos != null))
			{
				width = imageEnd.getWidth(map);
				height = imageEnd.getHeight(map);
				xImage = _imagePos[2].x;
				yImage = _imagePos[2].y;

				return new Rectangle(xImage - 1 - (width / 2), yImage - 1 - (height / 2), width + 2,
					height + 2);
			}
		}
		else
			return getBoundingRectangle(map);

		return null;
	} // end method getClippingArea

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#getLayer()
	 */
	public int getLayer()
	{
		int layer = _mapLink.getLayer();

		return ((layer < 0) ? LyStrutsMapCanvas.LINK_LAYER : layer);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#getLabel()
	 */
	public String getLabel()
	{
		if (_labelCache == null)
			_labelCache = translate(_mapLink.getLabel());

		return _labelCache;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#getLabel()
	 */
	public String getStartLabel()
	{
		if (_startLabelCache == null)
			_startLabelCache = translate(_mapLink.getStartLabel());

		return _startLabelCache;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.map.LyMapElement#getLabel()
	 */
	public String getEndLabel()
	{
		if (_endLabelCache == null)
			_endLabelCache = translate(_mapLink.getEndLabel());

		return _endLabelCache;
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
	 */
	public void setLabel(String label)
	{
		// Get rid of cache
		_labelCache = null;
		refresh();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#getTooltip()
	 */
	public String getTooltip()
	{
		return translate(_mapLink.getTooltip());
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#isAtPosition(leon.view.web.struts.map.LyStrutsMapCanvas,
	 *      java.awt.Graphics, int, int)
	 */
	public boolean isAtPosition(LyStrutsMapCanvas map, Graphics g, int x, int y)
	{
		if ((_startNode == null) || (_endNode == null))
			return false;

		Point   startPoint = map.convert(_startNode.getCenter());
		Point   endPoint = map.convert(_endNode.getCenter());
		int     x1 = startPoint.x;
		int     y1 = startPoint.y;
		int     x2 = endPoint.x;
		int     y2 = endPoint.y;
		Point[] polygon = getPolygon(x1, y1, x2, y2, _mapLink.getWidth() + 4);
		int[]   xPoints = new int[] { polygon[0].x, polygon[1].x, polygon[2].x, polygon[3].x };
		int[]   yPoints = new int[] { polygon[0].y, polygon[1].y, polygon[2].y, polygon[3].y };
		Polygon poly = new Polygon(xPoints, yPoints, 4);

		if (poly.contains(x, y))
			return true;

		String label = getLabel();

		if (label == null)
			return false;

		Dimension labelDim = getLabelSize(g);
		Rectangle rectLabel = new Rectangle((x1 + x2 - labelDim.width) / 2,
				(y1 + y2
					+ _mapLink.getWidth() + SELECTION_OFFSET) / 2, labelDim.width, labelDim.height);

		return rectLabel.contains(x, y);
	} // end method isAtPosition

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetSelectedInterface#setSelected(boolean)
	 */
	public void setSelected(boolean selected)
	{
		if (_mapLink.isSensitive() && (_mapLink.isSelected() != selected))
			_mapLink.setSelected(selected);
		
		refresh();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#isSelected()
	 */
	public boolean isSelected()
	{
		return _mapLink.isSelected();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetHiddenInterface#setHidden(boolean)
	 */
	public void setHidden(boolean hidden)
	{
		refresh();
	}

	/**
	 * This method does nothing.
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#translate(float, float, float)
	 */
	public void translate(float dx, float dy, float dz)
	{
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#isOverlay()
	 */
	public boolean isOverlay()
	{
		if ((_startNode == null) || (_endNode == null))
			return false;

		if (_mapLink.isSelected() || _startNode.isOverlay() || _endNode.isOverlay())
			return true;
		else
			return false;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#isHidden()
	 */
	public boolean isHidden()
	{
		return _mapLink.isHidden();
		
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#isSensitive()
	 */
	public boolean isSensitive()
	{
		return _mapLink.isSensitive();
	}

	/**
	 * Draws a dashedline in the given graphics using the given coordinates.
	 *
	 * @param g      the graphics context
	 * @param x1     starting x coordonate of the line
	 * @param y1     starting y coordonate of the line
	 * @param x2     ending x coordonate of the line
	 * @param y2     ending y coordonate of the line
	 * @param length length of the dashed line
	 * @see   LyStrutsMapLink#drawDashedLine(Graphics, int, int, int, int, int, Image)
	 */
	protected void drawDashedLine(Graphics g, int x1, int y1, int x2, int y2, int length)
	{
		drawDashedLine(g, x1, y1, x2, y2, length, null);
	}

	/**
	 * Draws a dashedline in the given graphics using the given coordinates and the given image.
	 *
	 * @param g      the graphics context
	 * @param x1     starting x coordonate of the dashed line
	 * @param y1     starting y coordonate of the dashed line
	 * @param x2     ending x coordonate of the dashed line
	 * @param y2     ending y coordonate of the dashed line
	 * @param length length of the dashed line
	 * @param image  the image of the dashed line
	 */
	protected void drawDashedLine(Graphics g, int x1, int y1, int x2, int y2, int length,
		Image image)
	{
		int     width = (x2 - x1);
		int     height = (y2 - y1);
		int     invert = 1;
		int     maxDim = Math.max(Math.abs(width), Math.abs(height));
		boolean horizontalParsing = (maxDim == Math.abs(width));
		int     i, coord, coord2;
		int     draw = 0;
		int     imageWidth, imageHeight;

		if (horizontalParsing)
		{
			if (x2 < x1)
				invert = -1;

			for (i = 0; i < maxDim; i += length)
			{
				coord = (int)((invert * i * ((float)height / width)) + y1);
				coord2 = (int)((invert * (i + (length / 2)) * ((float)height / width)) + y1);

				if ((draw % 3) == 0)
				{
					if (image != null)
					{
						imageWidth = image.getWidth(null);
						imageHeight = image.getHeight(null);
						g.drawImage(image, x1 + (invert * i) - (imageWidth / 2),
							coord
							- (imageHeight
									/ 2), imageWidth, imageHeight, null);
					}
					else
						g.drawLine(x1 + (invert * i), coord, x1 + (invert * (i + (length / 2))),
							coord2);
				}

				draw++;
			}
		}
		else
		{
			if (y2 < y1)
				invert = -1;

			for (i = 0; i < maxDim; i += length)
			{
				coord = (int)((invert * i * ((float)width / height)) + x1);
				coord2 = (int)((invert * (i + (length / 2)) * ((float)width / height)) + x1);

				if ((draw % 3) == 0)
				{
					if (image != null)
					{
						imageWidth = image.getWidth(null);
						imageHeight = image.getHeight(null);
						g.drawImage(image, coord - (imageWidth / 2),
							y1 + (invert * i)
							- (imageHeight
									/ 2), imageWidth, imageHeight, null);
					}
					else
						g.drawLine(coord, y1 + (invert * i), coord2,
							y1 + (invert * (i + (length / 2))));
				}

				draw++;
			}
		}
	} // end method drawDashedLine

	/**
	 * Sets the start node.
	 *
	 * @param node the new start node
	 */
	public void setStartNode(LyStrutsMapNode node)
	{
		_startNode = node;
	}

	/**
	 * Sets the end node.
	 *
	 * @param node the new end node
	 */
	public void setEndNode(LyStrutsMapNode node)
	{
		_endNode = node;
	}

	/**
	 * This method does nothing.
	 *
	 * @see leon.view.graphic.LyGraphicalMapLink#addPoint(leon.view.LyPoint, int)
	 */
	public void addPoint(LyPoint point, int pos)
	{

	}

	/**
	 * This method does nothing.
	 *
	 * @see leon.view.graphic.LyGraphicalMapLink#removePoint(leon.view.LyPoint)
	 */
	public void removePoint(LyPoint point)
	{

	}
	
	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalMapLink#setColorStart(java.lang.String)
	 */
	public void setColorStart(String colorStart)
	{
		_mapLink.setColorStart(colorStart);
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalMapLink#setColor(java.lang.String)
	 */
	public void setColor(String color)
	{
		_mapLink.setColor(color);
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalMapLink#setColorEnd(java.lang.String)
	 */
	public void setColorEnd(String colorEnd)
	{
		_mapLink.setColorEnd(colorEnd);
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalMapLink#setLabelColor(java.lang.String)
	 */
	public void setFontColor(String labelColor)
	{
		_mapLink.setFontColor(labelColor);
		refresh();
	}

	/**
	 * Refresh map containing this map link.<br />
	 * Call the <code>refreshMapImage()</code> on the <code>LyStrutsMapLink</code> instance.
	 */
	public void refresh()
	{
		LyComponent		parent = _mapLink.getParent();
		
		while ((parent != null) && !(parent instanceof LyMap))
		{
			parent = parent.getParent();
		}
		
		if (parent instanceof LyMap)
		{
			LyMap		map = (LyMap)parent;
			LyStrutsMap	webMap = (LyStrutsMap)map.getInternalView();
			
			webMap.refreshMapImage();
			webMap.refreshViewerImage();
		}
		else
			super.refresh();
	}

	/**
	 * Indicates is the component is drawable or not in web.
	 *
	 * @return always false
	 * @see    leon.view.web.struts.map.LyStrutsMapElement#isDrawableInterface(leon.view.web.struts.map.LyStrutsMapCanvas)
	 */
	public boolean isDrawableInterface(LyStrutsMapCanvas map)
	{
		return false;
	}
	
	/**
	 * Set the left label and clear the label cache.
	 * @see leon.view.graphic.LyGraphicalMapLink#setLeftLabel(java.lang.String)
	 */
	public void setEndLabel(String leftLabel)
	{
		_endLabelCache = null;
		
	}

	/**
	 * Set the right label and clear the label cache.
	 * @see leon.view.graphic.LyGraphicalMapLink#setRightLabel(java.lang.String)
	 */
	public void setStartLabel(String rightLabel)
	{
		_startLabelCache = null;
		
	}
	
	/**
	 * Generates the Javascript code that resets the position of the element.
	 */
	public void resetPosition()
	{
	}
} // end class LyStrutsMapLink
