/*
 * LyStrutsMapNode.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web.struts;

import java.awt.*;
import java.awt.image.*;
import java.io.*;

import com.sun.image.codec.jpeg.*;

import leon.misc.*;
import leon.view.*;
import leon.view.common.*;
import leon.view.filters.*;
import leon.view.graphic.*;
import leon.view.web.*;
import leon.view.web.struts.map.*;


/**
 * This class implements the Map Node object (as described in the dtd) for Struts viewer.
 *
 * <p><b>MAP_NODE: Map Node</b><br>
 * Nodes allow to define elements that can be placed at any location on the map and that can be
 * moved. By using a special positioning manager, it is possible to place these nodes according to a
 * predefined geometry (matrix, circle, ellipse, line,...). When zooming on the map, a node keeps
 * always the same aspect unlike an element of image type.</p>
 *
 * @version $Id: LyStrutsMapNode.java 41709 2012-12-05 08:01:31Z nhamon $
 * @author  Lyria
 * @see     leon.view.LyMap
 */
public class LyStrutsMapNode extends LyStrutsMapObject
    implements LyGraphicalMapNode,
               LyStrutsLayElement,
               LyStrutsMapElement
{
	public static final String    __VERSION =
		"$Id: LyStrutsMapNode.java 41709 2012-12-05 08:01:31Z nhamon $";

	   /** Default size. */
    public static int          DEFAULT_SIZE = 15;

    /** Center of the node. */
    protected LyPoint          _center;

    /** Font of the node. */
    protected Font             _graphicalFont;

    /** Width of the image (for optimization). */
    protected int              _imageWidth;

    /** Height of the image (for optimization). */
    protected int              _imageHeight;

    /** Width of the label (for optimization). */
    protected int              _labelWidth;

    /** Height of the label (for optimization). */
    protected int              _labelHeight;

    /** Width of the left label (for optimization). */
    protected int               _leftLabelWidth;

    /** Height of the left label (for optimization). */
    protected int               _leftLabelHeight;

    /** Width of the right label (for optimization). */
    protected int               _rightLabelWidth;

    /** Height of the right label (for optimization). */
    protected int               _rightLabelHeight;

    /** Width of the top label (for optimization). */
    protected int               _topLabelWidth;

    /** Height of the top label (for optimization). */
    protected int               _topLabelHeight;

    /** Width of the bottom label (for optimization). */
    protected int               _bottomLabelWidth;

    /** Height of the label (for optimization). */
    protected int               _bottomLabelHeight;

    /** Label offset (for optimization). */
    protected Point            _labelOffset;

    /** Left label offset (for optimization). */
    protected Point            _leftLabelOffset;

    /** Right label offset (for optimization). */
    protected Point            _rightLabelOffset;

    /** Top label offset (for optimization). */
    protected Point            _topLabelOffset;

    /** Bottom label offset (for optimization). */
    protected Point            _bottomLabelOffset;

    /** The map node generic component. */
    protected LyMapNode        _mapNode;

    /** Parent node (optional). */
    protected LyStrutsMapNode   _parentNode;

    /** Cache to translate label only once (may access very often). */
    protected String           _labelCache;
    
    /** Cache to translate left label only once (may access very often). */
    protected String           _leftLabelCache;
    
    /** Cache to translate right label only once (may access very often). */
    protected String           _rightLabelCache;
    
    /** Cache to translate top label only once (may access very often). */
    protected String           _topLabelCache;
    
    /** Cache to translate bottom label only once (may access very often). */
    protected String           _bottomLabelCache;

    /** The image of the node. */
    protected Image            _image;

    /** Resource key of the image node. */
    protected String           _imageString;

    /** The small image of the node. */
    protected Image            _smallImage;

    /** Resource key of the small image node. */
    protected String           _smallImageString;

    /** The image of a checkbox. */
    protected Image            _checkboxImage;

    /** The image of a selected checkbox. */
    protected Image            _selectedCheckboxImage;

    /** Color of the node. */
    protected Color            _nodeColor;

    /** Resource key of the node color. */
    protected String           _nodeColorString;

    /** Internal frame color. */
    protected Color            _internalFrameColor;

    /** Resource key of the internal frame color. */
    protected String           _internalFrameColorString;

    /** External frame color. */
    protected Color            _externalFrameColor;

    /** Resource key of the external frame color. */
    protected String           _externalFrameColorString;

    /** Cross color. */
    protected Color            _crossColor;

    /** Resource key of the cross color. */
    protected String           _crossColorString;

    /** Outline color. */
    protected Color            _outlineColor;

    /** Map canvas containing the node. */
    protected LyStrutsMapCanvas _map;
    
	/**
	 * Class constructor of a Map object for the struts target.
	 *
	 * @param map the generic Map defined for every type of view targets.
	 */
	public LyStrutsMapNode(LyMapNode mapNode)
	{
        super(mapNode);
        _mapNode = mapNode;

        initialize();
	}

	@Override
	public void free()
	{
		LyEnvironment	env = getEnvironment();
		LyViewManager	viewManager = getViewManager();
		
		if ((env != null) && (viewManager != null))
		{
			if ((_image != null) && !env.isImageCached(_image))
				viewManager.freeImage(_image);

			if ((_smallImage != null) && !env.isImageCached(_smallImage))
				viewManager.freeImage(_smallImage);
		}
		
		super.free();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.map.LyStrutsMapElement#drawHtml(leon.view.web.struts.map.LyStrutsMapCanvas,
	 *      leon.view.web.LyWebPageWriter)
	 */
	public void drawHtml(LyStrutsMapCanvas map, LyWebPageWriter pw)
	{
        updateRefreshInfo(pw);
        _map = map;

        LyComponent parent = _mapNode.getParent();

        if (parent instanceof LyMapContent)
            parent = parent.getParent();

        LyMap     parentMap = (LyMap)parent;
        Dimension dim = getImageSize(map);
        float     zoom = map.getZoom();
        Color     internalFrameColor = getInternalFrameColor();
        Color     externalFrameColor = getExternalFrameColor();

        if (_mapNode.isResizable())
            dim = new Dimension((int)(dim.width * zoom), (int)(dim.height * zoom));

        // Add a few extra pixel for selection
        if (_mapNode.isSensitive())
            dim = new Dimension(dim.width + 8, dim.height + 8);
        
        // Add a few extra pixels for external frame
        if (externalFrameColor != null)
            dim = new Dimension(dim.width + 16, dim.height + 16);
        else if (internalFrameColor != null)
            dim = new Dimension(dim.width + 8, dim.height + 8);

        int          maxWidth = dim.width;
        int          maxHeight = dim.height;
        StringBuffer onDblClick = new StringBuffer();
        StringBuffer onClick = new StringBuffer();
        StringBuffer onContextMenu = new StringBuffer();

        if (_mapNode.isSensitive())
        {
            boolean extend = true;

            onDblClick.append(" onDblClick=\"");
            onDblClick.append(getDoubleClickScript(pw));
            onDblClick.append("\"");

            if (parentMap.getSelectionMode() != LyMap.SelectionMode.NONE)
            {
                if (parentMap.getSelectionMode() == LyMap.SelectionMode.SINGLE)
                    extend = false;

                onClick.append(" onClick=\"");
                onClick.append(getClickScript(pw, extend));
                onClick.append("\"");
            }

            // Node is sensitive ignore resizable node
            onDblClick.append(" onContextMenu=\"");
            onDblClick.append(getContextMenuScript(pw, extend));
            onDblClick.append("\"");
        }

        // Ignore all node type except image
        // Draw the image representing the node
        map.resetTranslation();

        Rectangle   boundingRect = map.getBoundingRectangle();
        Point       point = map.convert(_center);
        int         x = point.x - boundingRect.x, y = point.y - boundingRect.y;
        int         layer = _mapNode.getLayer() + 2;
        String      display = "block";

        pw.append("<div id='mapnode_" + _mapNode.getIdCode() + "_global' style='position:absolute;left:" + x + "px;top:"
                + y + "px;display:" + display + ";z-index:" + layer + ";'>");
        pw.append("<div style='position:relative;'>");
        pw.append("<div id='mapnode_" + _mapNode.getIdCode() + "_node' nodeId='" + getId() + "'");
        
        if (isSelected())
            pw.append(" class='nodeSelectedImage'");
        else
            pw.append(" class='nodeImage'");
        
        pw.append(" style='position:absolute;left:" + (-maxWidth / 2) + "px;top:" + (-maxHeight / 2) + ";width:" + maxWidth + "px;height:" + maxHeight
            + "px;'");

        // add Event when sensitive and when may move
        LyLayout mapLayout = parentMap.getLayout();
        boolean  movable = _mapNode.isSensitive()
                && _mapNode.isMovable()
                && (mapLayout != null)
                && ((mapLayout.getStyle() == LyLayout.Style.COORDINATES)
                    || (mapLayout.getStyle() == LyLayout.Style.GEOGRAPHIC));
        
        if (movable)
        {
            pw.append(" onmousedown=\"");
            pw.append(getDragStartScript(pw));
            pw.append("\"");
        }
        
        pw.append(onDblClick.toString());
        pw.append(onClick.toString());
        pw.append(onContextMenu.toString());
        
        pw.append(">\n");

        writeHtml(pw);

        pw.append("</div>\n");

        // Write label if not null and not 'blank'.
        String  labelDiv = getLabelDivContent(map, "", layer, x, y);
        
        if (labelDiv != null)
            pw.append(labelDiv);

        // Write left label if not null and not 'blank'.
        String  leftLabelDiv = getLabelDivContent(map, "left", layer, x, y);
        
        if (leftLabelDiv != null)
            pw.append(leftLabelDiv);

        // Write right label if not null and not 'blank'.
        String  rightLabelDiv = getLabelDivContent(map, "right", layer, x, y);
        
        if (rightLabelDiv != null)
            pw.append(rightLabelDiv);

        // Write top label if not null and not 'blank'.
        String topLabelDiv = getLabelDivContent(map, "top", layer, x, y);
        
        if (topLabelDiv != null)
            pw.append(topLabelDiv);

        // Write bottom label if not null and not 'blank'.
        String bottomLabelDiv = getLabelDivContent(map, "bottom", layer, x, y);
        
        if (bottomLabelDiv != null)
            pw.append(bottomLabelDiv);

        pw.append("</div>\n");
        pw.append("</div>\n");
		
		if (isSelected())
		{
			if (parent instanceof LyMapContent)
				parent = parent.getParent();

			StringBuffer	script = new StringBuffer();
			String			rowId = _mapNode.getId();
			String			javascriptRowId = LyString.getJavascriptString(rowId);
			
			script.append("<script type='text/javascript'>\n//<![CDATA[\n");
			script.append("addSelectionId('map_" + parentMap.getIdCode() + "', '" + javascriptRowId + "');");
			script.append("\n//]]>\n</script>\n");
	
			if (pw.writeDivOnly())
				getWebViewManager().addAdditionalRefreshScripts(script.toString());
			else
				pw.addBeginScript(script.toString(), false);
		}
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalMapNode#setX(float)
	 */
	public void setX(float x)
	{
		if (_center.getX() != x)
		{
			_center.setX(x);

			resetPosition();
			refreshViewerImage();
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalMapNode#setY(float)
	 */
	public void setY(float y)
	{
		if (_center.getY() != y)
		{
			_center.setY(y);

			resetPosition();
			refreshViewerImage();
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalMapNode#setZ(float)
	 */
	public void setZ(float z)
	{
		if (_center.getZ() != z)
		{
			_center.setZ(z);

			resetPosition();
			refreshViewerImage();
		}
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalMapNode#setLayer(int)
	 */
	public void setLayer(int layer)
	{
	}
	
	/**
	 * Generates the Javascript code that resets the position of the element.
	 */
	public void resetPosition()
	{	
		if (_map != null)
		{
			_map.resetTranslation();

			Rectangle	boundingRect = _map.getBoundingRectangle();
			Point		point = _map.convert(_center);
			int			x = point.x - boundingRect.x, y = point.y - boundingRect.y;
			
			LyWebViewManager viewManager = getWebViewManager();
			StringBuffer     script = new StringBuffer();

			script.append("<script type='text/javascript'>\n//<![CDATA[\n");
			script.append("moveDiv(\"");
			script.append(_lastFrameName).append("\",\"mapnode_");
			script.append(_mapNode.getIdCode()).append("_global\",").append(x);
			script.append(",").append(y).append(");");
			script.append("\n//]]>\n</script>");

			viewManager.addAdditionalRefreshScripts(script.toString());
		}
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.map.LyStrutsLayElement#setPosition(leon.view.LyPoint)
	 */
	public void setPosition(LyPoint point3D)
	{
		_center = point3D;
		_mapNode.setX(_center.getX());
		_mapNode.setY(_center.getY());
		_mapNode.setZ(_center.getZ());

		// Reset image and label positions
		setImage(_mapNode.getImage());
		setLabel(_mapNode.getLabel());
		refreshViewerImage();
	}

	/**
	 * Gets the Javascript code used when clicking on the node.
	 *
	 * @param pw     the page writer
	 * @param extend indicates whether extended selection is allowed
	 * @return the String describing the Javascript code when clicking on the node
	 */
	protected String getClickScript(LyWebPageWriter pw, boolean extend)
	{
		LyComponent parent = _mapNode.getParent();

		if (parent instanceof LyMapContent)
			parent = parent.getParent();

		LyMap  parentMap = (LyMap)parent;

		// get the url to the selection handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    selectionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_SELECTEVENT);
		String				nodeId = _mapNode.getId();
		String				javascriptNodeId = LyString.getJavascriptString(nodeId);
		StringBuffer		buffer = new StringBuffer();
		
		buffer.append("javascript:InvokeMapClick('");
		buffer.append(javascriptNodeId);
		buffer.append("','");
		buffer.append(_mapNode.getRootController());
		buffer.append("','map_");
		buffer.append(parentMap.getIdCode());
		buffer.append("','");
		buffer.append(selectionUrl);
		buffer.append("',event);");
		
		return buffer.toString();
	}

	/**
	 * Gets the Javascript code used when double-clicking on the node.
	 *
	 * @param pw     the page writer
	 * @return the String describing the Javascript code when double-clicking on the node
	 */
	protected String getDoubleClickScript(LyWebPageWriter pw)
	{
		LyComponent parent = _mapNode.getParent();

		if (parent instanceof LyMapContent)
			parent = parent.getParent();

		LyMap  parentMap = (LyMap)parent;
		
		// get the url to the selection handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		String				nodeId = _mapNode.getId();
		String				javascriptNodeId = LyString.getJavascriptString(nodeId);
		StringBuffer		buffer = new StringBuffer();
		
		buffer.append("javascript:InvokeMapDblClick('");
		buffer.append(javascriptNodeId);
		buffer.append("','");
		buffer.append(_mapNode.getRootController());
		buffer.append("','");
		buffer.append(parentMap.getId());
		buffer.append("','");
		buffer.append(actionUrl);
		buffer.append("');");
		
		return buffer.toString();
	}

	/**
	 * Gets the Javascript code used when a context menu is performed on the node.
	 *
	 * @param pw     the page writer
	 * @param extend indicates whether extended selection is allowed
	 * @return the String describing the Javascript code when a context menu is performed on the node.
	 */
	protected String getContextMenuScript(LyWebPageWriter pw, boolean extend)
	{
		LyComponent parent = _mapNode.getParent();

		if (parent instanceof LyMapContent)
			parent = parent.getParent();

		LyMap  parentMap = (LyMap)parent;
		
		// get the url to the selection handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		String			    selectionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_SELECTEVENT);
		String				nodeId = _mapNode.getId();
		String				javascriptNodeId = LyString.getJavascriptString(nodeId);
		StringBuffer		buffer = new StringBuffer();
		
		buffer.append("javascript:InvokeMapPopup('");
		buffer.append(javascriptNodeId);
		buffer.append("','','");
		buffer.append(_mapNode.getRootController());
		buffer.append("','map_");
		buffer.append(parentMap.getIdCode());
		buffer.append("','");
		buffer.append(parentMap.getId());
		buffer.append("','");
		buffer.append(actionUrl);
		buffer.append("','");
		buffer.append(selectionUrl);
		buffer.append("',event);return false;");
		
		return buffer.toString();
	}
	
	/**
	 * Gets the Javascript code used when a drag event is performed on the node.
	 *
	 * @param pw the page writer
	 * @return the String describing the Javascript code when a drag event is performed on the node.
	 */
	protected String getDragStartScript(LyWebPageWriter pw)
	{
		LyComponent parent = _mapNode.getParent();

		if (parent instanceof LyMapContent)
			parent = parent.getParent();

		LyMap  		parentMap = (LyMap)parent;
		LyStrutsMap	strutsMap = (LyStrutsMap)parentMap.getInternalView();
		
		// get the url to the selection handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_MAP);
		StringBuffer		buffer = new StringBuffer();
		
		buffer.append("javascript:dragStart(event,'mapnode_");
		buffer.append(LyWebObject.encodeUrl(_mapNode.getIdCode() + ""));
		buffer.append("_global','");
		buffer.append(LyWebObject.encodeUrl(getId()));
		buffer.append("','");
		buffer.append(strutsMap.getId());
		buffer.append("','");
		buffer.append("mapDiv_" + parentMap.getIdCode());
		buffer.append("','");
		buffer.append("mapContent_" + parentMap.getIdCode());
		buffer.append("','");
		buffer.append(_mapNode.getRootController());
		buffer.append("','");
		buffer.append(actionUrl);
		buffer.append("');");

		return buffer.toString();
	}
	
	   /**
     * Initialization method.
     */
    protected void initialize()
    {
        _parentNode = null;

        _image = null;
        _imageString = null;
        _smallImage = null;
        _smallImageString = null;

        _nodeColor = null;
        _nodeColorString = null;
        _internalFrameColor = null;
        _internalFrameColorString = null;
        _externalFrameColor = null;
        _externalFrameColorString = null;
        _crossColor = null;
        _crossColorString = null;

        _outlineColor = null;

        _checkboxImage = null;
        _selectedCheckboxImage = null;

        _center = new LyPoint(_mapNode.getX(), _mapNode.getY(), _mapNode.getZ());

        _graphicalFont = getFont(_mapNode.getFont());

        LyComponent parent = _mapNode.getParent();

        if (parent instanceof LyMapContent)
            parent = parent.getParent();

        if (((LyMap)parent).getSelectionMode() == LyMap.SelectionMode.MULTIPLE)
        {
            _selectedCheckboxImage = getImage("LY_NODE_CHECKBOX_SELECTED");
            _checkboxImage = getImage("LY_NODE_CHECKBOX");
        }
        else if (((LyMap)parent).getSelectionMode() == LyMap.SelectionMode.SINGLE)
        {
            _selectedCheckboxImage = getImage("LY_NODE_RADIO_SELECTED");
            _checkboxImage = getImage("LY_NODE_RADIO");
        }

        _imageWidth = -1;
        _imageHeight = -1;
        _labelWidth = -1;
        _labelHeight = -1;
        _leftLabelWidth = -1;
        _leftLabelHeight = -1;
        _rightLabelWidth = -1;
        _rightLabelHeight = -1;
        _topLabelWidth = -1;
        _topLabelHeight = -1;
        _bottomLabelWidth = -1;
        _bottomLabelHeight = -1;
        _labelOffset = null;
        _leftLabelOffset = null;
        _rightLabelOffset = null;
        _topLabelOffset = null;
        _bottomLabelOffset = null;
        checkLabels();

        _checkboxImage = null;
        _selectedCheckboxImage = null;
    } // end method initialize

    /**
     * Returns the content of one of the label DIV.
     */
    protected String getLabelDivContent(LyStrutsMapCanvas map, String labelPosition, int layer, int x, int y)
    {
        String label;
        
        if ("top".equals(labelPosition))
            label = getTopLabel();
        else if ("left".equals(labelPosition))
            label = getLeftLabel();
        else if ("right".equals(labelPosition))
            label = getRightLabel();
        else if ("bottom".equals(labelPosition))
            label = getBottomLabel();
        else
            label = getLabel();

        if ( (label != null) && (label.trim().length() > 0) )
        {
            StringBuffer    buffer = new StringBuffer();
            Point           pos = getLeftLabelOffset(map, map.getGraphics(), true);
            Dimension       labelDim = getLeftLabelSize(map.getGraphics());
            
            if ("top".equals(labelPosition))
            {
                pos = getTopLabelOffset(map, map.getGraphics(), true);
                labelDim = getTopLabelSize(map.getGraphics());
            }
            else if ("left".equals(labelPosition))
            {
                pos = getLeftLabelOffset(map, map.getGraphics(), true);
                labelDim = getLeftLabelSize(map.getGraphics());
            }
            else if ("right".equals(labelPosition))
            {
                pos = getRightLabelOffset(map, map.getGraphics(), true);
                labelDim = getRightLabelSize(map.getGraphics());
            }
            else if ("bottom".equals(labelPosition))
            {
                pos = getBottomLabelOffset(map, map.getGraphics(), true);
                labelDim = getBottomLabelSize(map.getGraphics());
            }
            else
            {
                pos = getLabelOffset(map, map.getGraphics(), x, y, true);
                labelDim = getLabelSize(map.getGraphics());
            }

            buffer.append("<div");
            
            if (isSelected() )
                buffer.append(" class='nodeSelectedLabel'");
            else
                buffer.append(" class='nodelabel'");
            
            buffer.append(" id='mapnode_" + _mapNode.getIdCode() + "_" + labelPosition + "Label' style='position:absolute;left:" +
                (pos.x) + "px;top:" + (pos.y) + "px;width:" +
                labelDim.width + "px;");
            
            if (!testResource("LY_FILL_MAP_NODE_LABEL"))
                buffer.append("background-color:transparent;");
            
            if (!_mapNode.isDefaultFont())
            {
            	String	font = getFontString(_mapNode.getFont());
            	
            	buffer.append(font);
            }
                        
            buffer.append("'>");
            buffer.append(label);
            buffer.append("</div>");
            
            return buffer.toString();
        }
        
        return null;
    }
    
    /**
     * Returns the content of the right label DIV.
     */
    protected String getRightLabelDivContent(LyStrutsMapCanvas map, int layer, int x, int y)
    {
        String rightLabel = getRightLabel();

        if ( (rightLabel != null) && (rightLabel.trim().length() > 0) )
        {
            StringBuffer    buffer = new StringBuffer();
            Point           pos = getRightLabelOffset(map, map.getGraphics(), true);
            Dimension       labelDim = getRightLabelSize(map.getGraphics());

            buffer.append("<div");
            
            if (isSelected() )
                buffer.append(" class='nodeSelectedLabel'");
            else
                buffer.append(" class='nodelabel'");
            
            buffer.append(" id='mapnode_" + _mapNode.getIdCode() + "_rightLabel' style='position:absolute;left:" +
                (x + pos.x) + "px;top:" + (y + pos.y) + "px;width:" +
                labelDim.width + "px;z-index:" + layer + ";");
            
            if (!testResource("LY_FILL_MAP_NODE_LABEL"))
                buffer.append("background-color:transparent;");
            
            buffer.append("'>");
            buffer.append(rightLabel);
            buffer.append("</div>");
            
            return buffer.toString();
        }
        
        return null;
    }
    
    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#draw(leon.view.web.struts.map.LyStrutsMapCanvas,
     *      java.awt.Graphics, int)
     */
    public void draw(LyStrutsMapCanvas map, Graphics g, int mode)
    {
        if (_mapNode.isSensitive())
            return;

        Point point = map.convert(_center);

        int   x = point.x, y = point.y;

        drawImage(map, g, x, y, mode);
        drawLabel(map, g, x, y);
        drawLeftLabel(map, g, x, y);
        drawRightLabel(map, g, x, y);
        drawTopLabel(map, g, x, y);
        drawBottomLabel(map, g, x, y);
    }
    
    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#drawInViewer(leon.view.web.struts.map.LyStrutsMapCanvas,
     *      java.awt.Graphics, int)
     */
    public void drawInViewer(LyStrutsMapCanvas map, Graphics g)
    {
        Point   point = map.convert(_center);
        Color   color = getNodeColor();

        point = map.convertToViewer(point);
        g.setColor(color);

        switch (_mapNode.getViewerStyle())
        {
        case IMAGE:
            // Draw the image
            Image graphicalImage = getImage(_mapNode.getViewerImage());

            if (graphicalImage == null)
                graphicalImage = getImage();
            
            if (graphicalImage != null)
            {
                if (_mapNode.isResizable())
                {
                    Dimension dim = getImageSize(map);
                    int width = dim.width;
                    int height = dim.height;

                    LyPoint point3D1 = new LyPoint(_center.getX() - (width / 2), _center.getY()
                            - (height / 2), _center.getZ());
                    LyPoint point3D2 = new LyPoint(_center.getX() + (width / 2), _center.getY()
                            + (height / 2), _center.getZ());

                    Point point1 = map.convert(point3D1);
                    Point point2 = map.convert(point3D2);

                    point1 = map.convertToViewer(point1);
                    point2 = map.convertToViewer(point2);

                    g.drawImage(graphicalImage, point1.x, point1.y, point2.x - point1.x, point2.y
                            - point1.y, map);
                }
                else
                    g.drawImage(graphicalImage, point.x - 3, point.y - 3, 6, 6, map);
            }
            break;

        case CIRCLE:
            // Draw a circle
            g.fillOval(point.x - 2, point.y - 2, 4, 4);
            break;

        case SQUARE:
            g.fillRect(point.x - 2, point.y - 2, 4, 4);
            break;

        case TRIANGLE:
            g.fillPolygon(new int[] { point.x - 3, point.x, point.x + 3 }, new int[] { point.y + 3,
                    point.y - 3, point.y + 3 }, 3);
            break;

        case ELLIPSE:
            // Draw an ellipse
            g.fillOval(point.x - 2, point.y - 2, 4, 2);
            break;
        }
    }

    /**
     * Gets the small image HTML tag.
     *
     * @param  map       the map of the node
     * @param  maxWidth  the maximum width of the image
     * @param  maxHeight the maximum height of the image
     * @return the HTML tag used to write the small image of the node
     */
    protected String getSmallImageTag(LyStrutsMapCanvas map, int maxWidth, int maxHeight)
    {
        Image smallImage = getSmallImage();

        if (smallImage != null)
        {
            int    widthSmall = smallImage.getWidth(map);
            int    heightSmall = smallImage.getHeight(map);
            String imageUrl = getImageUrl(_smallImageString);

            String tag = "<img src='" + imageUrl + "' width='" + widthSmall + "' height='"
                + heightSmall + "'" + " style='position:absolute;left:" + (maxWidth / 2)
                + "px;top:" + (maxHeight / 2) + "px;z-index:3'/>\n";

            return tag;
        }
        
        return null;
    }

    /**
     * Gets the node HTML tag.
     *
     * @param  map       the map of the node
     * @param  maxWidth  the maximum width of the node
     * @param  maxHeight the maximum height of the node
     * @return the HTML tag used to write the node
     */
    protected String getNodeTag(LyStrutsMapCanvas map, int maxWidth, int maxHeight)
    {
        Dimension symbolSize = getSymbolSize(map);
        float     zoom = map.getZoom();

        if (_mapNode.isResizable())
            symbolSize = new Dimension((int)(symbolSize.width * zoom),
                    (int)(symbolSize.height * zoom));

        String imagePath = "";
        int    width = 0;
        int    height = 0;
        LyMapNode.Style     style = _mapNode.getStyle();
        
        if (style == LyMapNode.Style.IMAGE)
        {
            int[]   dim = getImageSize();
            
            width = dim[0];
            height = dim[1];
            
            String image = _imageString;
    		String docDir;
    		String docRoot = expandUrl(getEnv("LY_APP_DOC"));

    		image = expandUrl(image);
    		docDir = expandUrl(getEnv("LY_APP_DOC_DIR"));

    		if (image.startsWith(docDir))
    		{
    			image = image.substring(docDir.length());
    			image = getEnv("LY_APP_DOC") + image;
    		}
    		else
    		{
    			docDir = expandUrl(getEnv("LY_LEON_DOC_DIR"));
    			if (image.startsWith(docDir))
    			{
    				image = image.substring(docDir.length());
    				image = getEnv("LY_LEON_DOC") + image;
    			}
    		}

    		if (image.startsWith(docRoot))
    		{
    			
    			image = getEnvironment().getPathWithoutSize(image);
    			
    			String	test = expandUrl(getEnv("LY_APP_DOC_DIR")) + image.substring(docRoot.length());
    			Image	im = getImage(test);
    			
    			if (im != null)
    			{
                    int[]	dimTest = getEnvironment().getImageSize(_imageString, false);
        			
    				width = im.getWidth(null);
    				height = im.getHeight(null);
    				
    				if (dimTest != null)
    				{
    					if (dimTest.length == 1)
    					{
    						width = width * dimTest[0] / height;
    						height = dimTest[0];
    					}
    					else if (dimTest.length > 1)
    					{
    						width = dimTest[0];
    						height = dimTest[1];
    					}
    				}
    			}

    			imagePath = expandUrl(image);
    		}
    		else if (image.startsWith("file"))
    			imagePath = LyString.encodeFileURL(image);
    		else if (image.startsWith("http") || image.startsWith("ftp"))
    			imagePath = image;
    		else
    			imagePath = getImageUrl(image);
    		
    		imagePath = getEnvironment().getPathWithoutSize(imagePath);
        }
        else
        {
            imagePath = generateNodeImage(map, maxWidth, maxHeight);
            width = symbolSize.width;
            height = symbolSize.height;
        }

        // At this point, the image file exist, we can return the IMG tag for
        // this image

        StringBuffer tag = new StringBuffer();

        if (imagePath != null)
        {
            tag.append("<img src='" + imagePath + "' width='" + width + "' height='" + height + "'");
            tag.append(" style='position:absolute;left:" + ((maxWidth - width) / 2) + "px;top:"
                + ((maxHeight - height) / 2) + "px;");
            
            if (_mapNode.isSensitive())
                tag.append("cursor:hand;cursor:pointer;");
            
            tag.append("' />\n");
        }

        return tag.toString();

    } // end method getNodeTag

    /**
     * Creates an image for the node.
     *
     * @param  map       the map of the node
     * @param  maxWidth  the maximum width of the image
     * @param  maxHeight the maximum height of the image
     * @return the image path
     * @see    java.awt.Component#createImage(int, int)
     */
    protected String generateNodeImage(LyStrutsMapCanvas map, int maxWidth, int maxHeight)
    {
        // Dimension dim = getImageSize(map);
        if (_mapNode.getStyle() == LyMapNode.Style.IMAGE)
            return null;

        Dimension symbolSize = getSymbolSize(map);
        float     zoom = map.getZoom();
        String    nodeId = "";

        if (_mapNode.isResizable())
            // dim = new Dimension((int)(dim.width * zoom), (int)(dim.height *
            // zoom));
            symbolSize = new Dimension((int)(symbolSize.width * zoom),
                    (int)(symbolSize.height * zoom));

        Color color = getNodeColor();
        LyMapNode.Style style = _mapNode.getStyle();
        
        if (style == LyMapNode.Style.CIRCLE)
            nodeId = "C";
        else if (style == LyMapNode.Style.SQUARE)
            nodeId = "S";
        else if (style == LyMapNode.Style.TRIANGLE)
            nodeId = "T";
        else if (style == LyMapNode.Style.ELLIPSE)
            nodeId = "E";

        nodeId += "" + symbolSize.width + "_" + getHexString(color);

        String imagePath = saveImage(null, nodeId);

        if (imagePath != null)
            return imagePath;

        Image    image = map.createImage(symbolSize.width, symbolSize.height);

        if (image == null)
            return null;
        
        Graphics g = image.getGraphics();

        g.setColor(Color.magenta);
        g.fillRect(0, 0, symbolSize.width, symbolSize.height);

        // Draw the image representing the node        
        if (style == LyMapNode.Style.IMAGE)
            return null;
        
        else if (style == LyMapNode.Style.CIRCLE)
        {
            // Draw a circle
            g.setColor(color);
            g.fillOval(0, 0, symbolSize.width, symbolSize.height);
            g.setColor(getOutlineColor());
            g.drawOval(0, 0, symbolSize.width, symbolSize.height);
            nodeId = "C";
        }
        else if (style == LyMapNode.Style.SQUARE)
        {
            g.setColor(color);
            g.fillRect(0, 0, symbolSize.width, symbolSize.height);
            g.setColor(getOutlineColor());
            g.drawRect(0, 0, symbolSize.width, symbolSize.height);
            nodeId = "S";
        }
        else if (style == LyMapNode.Style.TRIANGLE)
        {
            g.setColor(color);
            g.fillPolygon(new int[] { 0, symbolSize.width / 2, symbolSize.width },

            new int[] { symbolSize.height, 0, symbolSize.height }, 3);
            g.setColor(getOutlineColor());
            g.drawPolygon(new int[] { 0, symbolSize.width / 2, symbolSize.width },
                new int[]
                {
                    symbolSize.height, 0, symbolSize.height
                }, 3);
            nodeId = "T";
        }
        else if (style == LyMapNode.Style.ELLIPSE)
        {
            // Draw an ellipse
            g.setColor(color);
            g.fillOval(0, 0, symbolSize.width, symbolSize.height);
            g.setColor(getOutlineColor());
            g.drawOval(0, 0, symbolSize.width, symbolSize.height);
            nodeId = "E";
        }

        // write file to disk
        image = LyReplaceColorImageFilter.getFilteredImage(image, Color.magenta, null);

        imagePath = saveImage(image, nodeId);

        if (image != null)
        {
            image.flush();
            image = null;
        }

        return imagePath;
    } // end method generateNodeImage

    /**
     * Creates the image of a cross and build the HTML tag used to write it.
     *
     * @param  map       the map of the node
     * @param  maxWidth  the maximum width of the cross image
     * @param  maxHeight the maximum height of the cross image
     * @return the HTML tag used to write the cross image
     */
    protected String generateCrossImage(LyStrutsMapCanvas map, int maxWidth, int maxHeight)
    {
        Dimension symbolSize = getSymbolSize(map);
        float     zoom = map.getZoom();

        if (_mapNode.isResizable())
            symbolSize = new Dimension((int)(symbolSize.width * zoom),
                    (int)(symbolSize.height * zoom));

        Color  crossColor = getCrossColor();
        String crossId = "CR" + symbolSize.width + "_" + symbolSize.height + "_"
            + getHexString(crossColor);

        if (crossColor == null)
            return null;

        // Before drawing, check if the image exists...
        String    imagePath = saveImage(null, crossId);
        Dimension crossDim = new Dimension(symbolSize.width - 2, symbolSize.height - 2);

        if ((imagePath == null) || (imagePath.length() == 0))
        {
            Image    image = map.createImage(crossDim.width + 1, crossDim.height + 1);
            Graphics g = image.getGraphics();

            g.setColor(Color.magenta);
            g.fillRect(0, 0, crossDim.width, crossDim.height);

            g.setColor(crossColor);
            g.drawLine(0, 0, crossDim.width, crossDim.height);
            g.drawLine(1, 0, crossDim.width, crossDim.height - 1);
            g.drawLine(2, 0, crossDim.width, crossDim.height - 2);
            g.drawLine(0, 1, crossDim.width - 1, crossDim.height);
            g.drawLine(0, 2, crossDim.width - 2, crossDim.height);

            g.drawLine(0, crossDim.height, crossDim.width, 0);
            g.drawLine(1, crossDim.height, crossDim.width, 1);
            g.drawLine(2, crossDim.height, crossDim.width, 2);
            g.drawLine(0, crossDim.height - 1, crossDim.width - 1, 0);
            g.drawLine(0, crossDim.height - 2, crossDim.width - 2, 0);

            image = LyReplaceColorImageFilter.getFilteredImage(image, Color.magenta, null);

            imagePath = saveImage(image, crossId);
            if (image != null)
            {
                image.flush();
                image = null;
            }
        }

        int          width = crossDim.width;
        int          height = crossDim.height;

        StringBuffer tag = new StringBuffer();

        tag.append("<img src='" + imagePath + "' width='" + width + "' height='" + height + "'");
        tag.append(" style='position:absolute;left:" + ((maxWidth - width) / 2) + "px;top:"
            + ((maxHeight - height) / 2) + "px;z-index:4'");
        tag.append(" />\n");

        return tag.toString();

    } // end method generateCrossImage

    /**
     * Creates the image of the node's frame and build the HTML tag used to write it. If internal
     * and external frame color are not defined return null, else creates image and tag.
     *
     * @param  map       the map of the node
     * @param  maxWidth  the maximum width of the frame image
     * @param  maxHeight the maximum height of the frame image
     * @return the HTML tag used to write the frame
     */
    protected String generateFrameImage(LyStrutsMapCanvas map, int maxWidth, int maxHeight)
    {
        Color internalFrameColor = getInternalFrameColor();
        Color externalFrameColor = getExternalFrameColor();
        int   externalGap = 0, internalGap = 0;

        if ((externalFrameColor == null) && (internalFrameColor == null))
            return null;
        else if (externalFrameColor != null)
        {
            externalGap = 4;
            internalGap = 8;
        }
        else if (internalFrameColor != null)
        {
            internalGap = 4;
        }
        
        StringBuffer tag = new StringBuffer();
        Color        outlineColor = getColor("LY_MAP_OUTLINE_COLOR");

        if (externalFrameColor != null)
        {
            tag.append("<div id='" + getId() + "_external_frame' style='position:absolute;left:" + externalGap + "px;top:"
                    + externalGap + "px;z-index:1;width:"
                    + (maxWidth - 2*externalGap) + "px;height:"
                    + (maxHeight - 2*externalGap) + "px;"
                    + "background-color:#" + getHexString(externalFrameColor) + ";"
                    + "border:1px solid;border-color:#" + getHexString(outlineColor) + ";'>&nbsp;</div>");
        }
        
        if (internalFrameColor != null)
        {
            tag.append("<div id='" + getId() + "_internal_frame' style='position:absolute;left:" + internalGap + "px;top:"
                    + internalGap + "px;z-index:1;width:"
                    + (maxWidth - 2*internalGap) + "px;height:"
                    + (maxHeight - 2*internalGap) + "px;"
                    + "background-color:#" + getHexString(internalFrameColor) + ";"
                    + "border:1px solid;border-color:#" + getHexString(outlineColor) + ";'>&nbsp;</div>");
        }
        
        return tag.toString();

    } // end method generateFrameImage

    /**
     * Saves the given image with the given file name and return the image path if save is
     * successful.<br>
     * The file path is : <i>"LY_APP_DOC"/tmp/<code>filename</code>.gif</i>
     *
     * @param  image    the image to save
     * @param  filename the file name of the image (without extention)
     * @return return null if the image is null or if an exception occur during process, return the
     *         image path otherwise
     */
    public String saveImage(Image image, String filename)
    {
        LyViewManager viewManager = getViewManager();
        String        tmpDir = viewManager.getTemporaryDir();
        String        tmpRoot = viewManager.getTemporaryLocation();
        String        imageFile = "";
        String        imagePath = "";

        imageFile = tmpDir + "/" + filename + ".gif";
        imagePath = tmpRoot + "/" + filename + ".gif";

        File file = new File(imageFile);

        if (file.exists())
            return imagePath;

        if (image == null)
            return null;

        try
        {
            OutputStream fileOut = new FileOutputStream(imageFile);
            BufferedImage buffImage = new BufferedImage(image.getWidth(null), image.getHeight(null),
                    BufferedImage.TYPE_INT_BGR);
            Graphics      g = buffImage.getGraphics();

            g.drawImage(image, 0, 0, null);

            JPEGEncodeParam encodeParam = JPEGCodec.getDefaultJPEGEncodeParam(buffImage);

            // set the quality param (1.0 is highest, 0.0 is lowest)
            // I think that 0.9 is a good choice between quality and size
            encodeParam.setQuality((float)0.90, true);
            JPEGImageEncoder encoder = JPEGCodec.createJPEGEncoder(fileOut, encodeParam);

            encoder.encode(buffImage);
            fileOut.flush();
            fileOut.close();
        }
        catch (Exception e)
        {
            e.printStackTrace();
            return null;
        }

        return imagePath;
    } // end method saveImage

    /**
     * Draws the image of the node depending on its style. If the mode is not normal, doesn't draw
     * anything.
     *
     * @param map  the map of the node
     * @param g    the graphics context
     * @param x    the x position of the node
     * @param y    the y position of the node
     * @param mode the style of the node : {@link LyStrutsMapElement#DRAW_MODE_NORMAL} or {@link
     *             LyStrutsMapElement#DRAW_MODE_REDUCED}.
     */
    protected void drawImage(LyStrutsMapCanvas map, Graphics g, int x, int y, int mode)
    {
        boolean isNormalMode = (mode == LyStrutsMapElement.DRAW_MODE_NORMAL);

        if (!isNormalMode)
            return;

        Dimension dim = getImageSize(map);
        Dimension symbolSize = getSymbolSize(map);
        float     zoom = map.getZoom();

        if (_mapNode.isResizable())
        {
            dim = new Dimension((int)(dim.width * zoom), (int)(dim.height * zoom));
            symbolSize = new Dimension((int)(symbolSize.width * zoom),
                    (int)(symbolSize.height * zoom));
        }

        Color internalFrameColor;
        Color externalFrameColor;
        Color color;

        internalFrameColor = getInternalFrameColor();
        externalFrameColor = getExternalFrameColor();
        color = getNodeColor();

        // Draw a single frame
        if ((externalFrameColor != null) || (internalFrameColor != null))
        {
            Color outLineColor = getOutlineColor();

            // Draw the external frame
            if (externalFrameColor != null)
            {
                g.setColor(externalFrameColor);
                g.fillRect(x - (dim.width / 2), y - (dim.height / 2), dim.width, dim.height);

                // Draw the external border
                g.setColor(outLineColor);
                g.drawRect(x - (dim.width / 2), y - (dim.height / 2), dim.width, dim.height);

                // Draw the internal frame
                if (internalFrameColor != null)
                {
                    g.setColor(internalFrameColor);
                    g.fillRect(x - (dim.width / 2) + 4, y - (dim.height / 2) + 4, dim.width - 8,
                        dim.height - 8);

                    // Draw the internal border
                    g.setColor(outLineColor);
                    g.drawRect(x - (dim.width / 2) + 4, y - (dim.height / 2) + 4, dim.width - 8,
                        dim.height - 8);
                }
            }
            else if (internalFrameColor != null)
            {
                // Draw the single frame
                g.setColor(internalFrameColor);
                g.fillRect(x - (dim.width / 2), y - (dim.height / 2), dim.width, dim.height);

                // Draw the external border
                g.setColor(outLineColor);
                g.drawRect(x - (dim.width / 2), y - (dim.height / 2), dim.width, dim.height);
            }
        }

        // Draw the image representing the node
        LyMapNode.Style style = _mapNode.getStyle();
        
        if (style == LyMapNode.Style.IMAGE)
        {
            // Draw the image
            Image graphicalImage = getImage();

            // Replace the black color with the node color
            if (_mapNode.getColor() != null)
                graphicalImage = LyReplaceColorImageFilter.getFilteredImage(graphicalImage,
                        Color.black, color);

            int[]   size = getImageSize();          
            int     width = size[0];
            int     height = size[1];

            if (_mapNode.isResizable())
            {
                width = (int)(width * zoom);
                height = (int)(height * zoom);
            }

            g.drawImage(graphicalImage, x - (width / 2), y - (height / 2), width, height, map);
        }
        else if (style == LyMapNode.Style.CIRCLE)
        {
            // Draw a circle
            g.setColor(color);
            g.fillOval(x - (symbolSize.width / 2), y - (symbolSize.height / 2), symbolSize.width,
                symbolSize.height);
            g.setColor(getOutlineColor());
            g.drawOval(x - (symbolSize.width / 2), y - (symbolSize.height / 2),
                symbolSize.width - 1, symbolSize.height - 1);
        }
        else if (style == LyMapNode.Style.SQUARE)
        {
            g.setColor(color);
            g.fillRect(x - (symbolSize.width / 2), y - (symbolSize.height / 2), symbolSize.width,
                symbolSize.height);
            g.setColor(getOutlineColor());
            g.drawRect(x - (symbolSize.width / 2), y - (symbolSize.height / 2),
                symbolSize.width - 1, symbolSize.height - 1);
        }
        else if (style == LyMapNode.Style.TRIANGLE)
        {
            g.setColor(color);
            g.fillPolygon(new int[] { x - (symbolSize.width / 2), x, x + (symbolSize.width / 2) },
                new int[]
                {
                    y + (symbolSize.height / 2),
                    y - (symbolSize.height / 2),
                    y + (symbolSize.height / 2)
                }, 3);
            g.setColor(getOutlineColor());
            g.drawPolygon(
                new int[] { x - (symbolSize.width / 2) + 1, x, x + (symbolSize.width / 2) - 1 },
                new int[]
                {
                    y + (symbolSize.height / 2) - 1,
                    y - (symbolSize.height / 2) + 1,
                    y + (symbolSize.height / 2) - 1
                }, 3);
        }
        else if (style == LyMapNode.Style.ELLIPSE)
        {
            // Draw an ellipse
            g.setColor(color);
            g.fillOval(x - (symbolSize.width / 2), y - (symbolSize.height / 2), symbolSize.width,
                symbolSize.height);
            g.setColor(getOutlineColor());
            g.drawOval(x - (symbolSize.width / 2), y - (symbolSize.height / 2),
                symbolSize.width - 1, symbolSize.height - 1);
        }

        LyComponent parent = _mapNode.getParent();

        if (parent instanceof LyMapContent)
            parent = parent.getParent();

        if ((((LyMap)parent).getSelectionMode() != LyMap.SelectionMode.NONE)
                && _mapNode.isSensitive())
            drawCheckbox(map, g);

        Color crossColor = getCrossColor();

        if (crossColor != null)
        {
            // Draw a cross on the node
           g.setColor(crossColor);

            Dimension crossDim = new Dimension(symbolSize.width - 2, symbolSize.height - 2);

            g.drawLine(x - (crossDim.width / 2), y - (crossDim.height / 2),
                x + (crossDim.width / 2), y
                + (crossDim.height / 2));
            g.drawLine(x - (crossDim.width / 2) + 1, y - (crossDim.height / 2),
                x + (crossDim.width / 2), y + (crossDim.height / 2) - 1);
            g.drawLine(x - (crossDim.width / 2) + 2, y - (crossDim.height / 2),
                x + (crossDim.width / 2), y + (crossDim.height / 2) - 2);
            g.drawLine(x - (crossDim.width / 2), y - (crossDim.height / 2) + 1,
                x + (crossDim.width / 2)
                - 1, y + (crossDim.height / 2));
            g.drawLine(x - (crossDim.width / 2), y - (crossDim.height / 2) + 2,
                x + (crossDim.width / 2)
                - 2, y + (crossDim.height / 2));

            g.drawLine(x - (crossDim.width / 2), y + (crossDim.height / 2),
                x + (crossDim.width / 2), y
                - (crossDim.height / 2));
            g.drawLine(x - (crossDim.width / 2) + 1, y + (crossDim.height / 2),
                x + (crossDim.width / 2), y - (crossDim.height / 2) + 1);
            g.drawLine(x - (crossDim.width / 2) + 2, y + (crossDim.height / 2),
                x + (crossDim.width / 2), y - (crossDim.height / 2) + 2);
            g.drawLine(x - (crossDim.width / 2), y + (crossDim.height / 2) - 1,
                x + (crossDim.width / 2)
                - 1, y - (crossDim.height / 2));
            g.drawLine(x - (crossDim.width / 2), y + (crossDim.height / 2) - 2,
                x + (crossDim.width / 2)
                - 2, y - (crossDim.height / 2));
        }

        // Draw the small image
        Image smallImage = getSmallImage();

        if (smallImage != null)
        {
            int width = smallImage.getWidth(map);
            int height = smallImage.getHeight(map);

            g.drawImage(smallImage, x + (dim.width / 2) - width, y + (dim.height / 2) - height,
                map);
        }
    } // end method drawImage

    /**
     * Draws a checkbox with the node.
     *
     * @param map the map of the node
     * @param g   the graphics context
     */
    protected void drawCheckbox(LyStrutsMapCanvas map, Graphics g)
    {
        Image checkBox;

        if (_mapNode.isSelected())
            checkBox = _selectedCheckboxImage;
        else
            checkBox = _checkboxImage;

        if (checkBox == null)
            return;

        int   width = checkBox.getWidth(null);
        int   height = checkBox.getHeight(null);
        Point position = getCheckBoxPosition(map, width, height);

        g.drawImage(checkBox, position.x, position.y, width, height, map);
    }

    /**
     * Gets the checkbox position of the node switch its label alignment.
     *
     * @param  map         the map of the link
     * @param  imageWidth  the width of the checkbox image
     * @param  imageHeight the height of the checkbox image
     * @return the checkbox position
     */
    protected Point getCheckBoxPosition(LyStrutsMapCanvas map, int imageWidth, int imageHeight)
    {
        Rectangle imageRect = getNodeBoundingRectangle(map);
        String    oldCache = _labelCache;
        String    label = getLabel();
        int       x = 0, y = 0;

        if ((label == null) || (label.length() == 0))
        {
            x = imageRect.x - imageWidth - 2;
            y = imageRect.y + imageRect.height - imageHeight;
        }
        else
        {
            LyMapNode.LabelAlign    lAlign = _mapNode.getLabelAlign();
            
            if (lAlign == LyMapNode.LabelAlign.LEFT)
            {
                x = imageRect.x + imageRect.width + 2;
                y = imageRect.y + imageRect.height - imageHeight;
            }
            else if (lAlign == LyMapNode.LabelAlign.CENTER)
            {
                x = imageRect.x + (imageRect.width / 2) - (imageWidth / 2);
                y = imageRect.y + imageRect.height + 2;
            }
            else if (lAlign == LyMapNode.LabelAlign.EXTERN)
            {
                Point center = map.convert(_center);

                if (_labelOffset == null)
                {
                    getLabelSize(null, true);
                    getLabelOffset(map, null, center.x, center.y, true);
                }

                Rectangle labelRect = new Rectangle(center.x + _labelOffset.x,
                        center.y
                        + _labelOffset.y, _labelWidth, _labelHeight);
                Rectangle excludeZone = new Rectangle(0, imageRect.y, imageRect.x,
                        imageRect.height);

                if (labelRect.intersects(excludeZone))
                {
                    x = imageRect.x + imageRect.width + 2;
                    y = imageRect.y + imageRect.height - imageHeight;
                }
                else
                {
                    x = imageRect.x - imageWidth - 2;
                    y = imageRect.y + imageRect.height - imageHeight;
                }
            }
            else
            {
                x = imageRect.x - imageWidth - 2;
                y = imageRect.y + imageRect.height - imageHeight;
            }
        }

        _labelCache = oldCache;
        return new Point(x, y);
    } // end method getCheckBoxPosition

    /**
     * Gets the offset between the center of the node and the left upper corner of the label.
     *
     * @param  map the node map
     * @param  g   the graphics context
     * @param  x   the x position of the node center
     * @param  y   the y position of the node center
     * @return the offset point
     */
    protected Point getLabelOffset(LyStrutsMapCanvas map, Graphics g, int x, int y)
    {
        return getLabelOffset(map, g, x, y, false);
    }

    /**
     * Returns the offset between the center of the node and the left upper corner of the label.
     *
     * @param  map         the node map
     * @param  g           the graphics context
     * @param  x           the x position of the node center
     * @param  y           the y position of the node center
     * @param  recalculate indicates if the offset has to be recalculated or not
     * @return the label offset
     */
    @SuppressWarnings("deprecation")
	protected Point getLabelOffset(LyStrutsMapCanvas map, Graphics g, int x, int y,
        boolean recalculate)
    {
        if ((_labelOffset == null) || recalculate || _mapNode.isResizable())
        {
            FontMetrics fm;

            if (g == null)
            {
                Toolkit toolkit = Toolkit.getDefaultToolkit();

                fm = toolkit.getFontMetrics(_graphicalFont);
            }
            else
                fm = g.getFontMetrics(_graphicalFont);

            Dimension labelDim = getLabelSize(g);
            Dimension imageDim = getImageSize(map);
            int       offx = 0, offy = 0;
            Color     internalFrameColor = getInternalFrameColor();
            Color     externalFrameColor = getExternalFrameColor();

            if (_mapNode.isResizable())
            {
                float zoom = map.getZoom();

                imageDim = new Dimension((int)(imageDim.width * zoom),
                        (int)(imageDim.height * zoom));
            }
            
            // Add a few extra pixels for external frame
            if (externalFrameColor != null)
                imageDim = new Dimension(imageDim.width + 16, imageDim.height + 16);
            else if (internalFrameColor != null)
                imageDim = new Dimension(imageDim.width + 8, imageDim.height + 8);

            LyMapNode.LabelAlign    lAlign = _mapNode.getLabelAlign();
            
            if (lAlign == LyMapNode.LabelAlign.BOTTOM)
            {
                offx = -labelDim.width / 2;
                offy = (imageDim.height / 2) + 3;
            }
            else if (lAlign == LyMapNode.LabelAlign.TOP)
            {
                offx = -labelDim.width / 2;
                offy = (-imageDim.height / 2) - fm.getHeight();
            }
            else if (lAlign == LyMapNode.LabelAlign.LEFT)
            {
                offx = (-imageDim.width / 2) - labelDim.width - 1;
                offy = -fm.getHeight() / 2;
            }
            else if (lAlign == LyMapNode.LabelAlign.RIGHT)
            {
                offx = (imageDim.width / 2) + 4;
                offy = -fm.getHeight() / 2;
            }
            else if (lAlign == LyMapNode.LabelAlign.CENTER)
            {
                offx = -labelDim.width / 2;
                offy = -fm.getHeight() / 2;
            }
            else if (lAlign == LyMapNode.LabelAlign.EXTERN)
            {
                Dimension dim = map.getSize();
                Point     mapCenter = new Point(dim.width / 2, dim.height / 2);
                double    alpha;

                // If the center of the node and the center of the map have the
                // same x coordonate, alpha equals PI/2 or -PI/2
                if (mapCenter.x == x)
                {
                    if (y > mapCenter.y)
                        alpha = -Math.PI / 2;
                    else
                        alpha = Math.PI / 2;
                }
                // Else calculate alpha in common case
                else
                    alpha = -Math.atan(((float)(y - mapCenter.y)) / (x - mapCenter.x));

                // Give alpha a value between -PI and PI
                if (x < mapCenter.x)
                {
                    if (y > mapCenter.y)
                        alpha -= Math.PI;
                    else
                        alpha += Math.PI;
                }

                // Treat each of the four sectors
                if ((alpha >= (-Math.PI / 4)) && (alpha < (Math.PI / 4)))
                {
                    offx = (imageDim.width / 2) + 4;
                    offy = (int)((-fm.getHeight() / 2)
                            - ((fm.getHeight() + imageDim.height)
                                    * Math.tan(alpha) / 2));
                }
                else if ((alpha >= (Math.PI / 4)) && (alpha < (3 * Math.PI / 4)))
                {
                    offx = (int)((-labelDim.width / 2)
                            - ((labelDim.width + imageDim.width)
                                    * Math.tan(alpha + (Math.PI / 2)) / 2));
                    offy = (-imageDim.height / 2) - fm.getHeight();
                }
                else if (((alpha >= (3 * Math.PI / 4)) && (alpha <= Math.PI))
                        || ((alpha >= -Math.PI) && (alpha < (-3 * Math.PI / 4))))
                {
                    offx = (-imageDim.width / 2) - labelDim.width - 3;
                    offy = (int)((-fm.getHeight() / 2)
                            + ((fm.getHeight() + imageDim.height)
                                    * Math.tan(alpha) / 2));
                }
                else
                {
                    offx = (int)((-labelDim.width / 2)
                            + ((labelDim.width + imageDim.width)
                                    * Math.tan(alpha + (Math.PI / 2)) / 2));
                    offy = (imageDim.height / 2) + 1 + 3;
                }
            }

            _labelOffset = new Point(offx, offy);
        }

        return _labelOffset;
    } // end method getLabelOffset

    /**
     * Gets the offset between the center of the node and the left upper corner of the left label.
     *
     * @param  map the node map
     * @param  g   the graphics context
     * @return the offset point
     */
    protected Point getLeftLabelOffset(LyStrutsMapCanvas map, Graphics g)
    {
        return getLeftLabelOffset(map, g, false);
    }

    /**
     * Returns the offset between the center of the node and the left upper corner of the left label.
     *
     * @param  map         the node map
     * @param  g           the graphics context
     * @param  recalculate indicates if the offset has to be recalculated or not
     * @return the label offset
     */
    @SuppressWarnings("deprecation")
	protected Point getLeftLabelOffset(LyStrutsMapCanvas map, Graphics g, boolean recalculate)
    {
        if (_leftLabelOffset == null || recalculate || _mapNode.isResizable())
        {
            FontMetrics     fm;

            if (g == null)
            {
                Toolkit toolkit = Toolkit.getDefaultToolkit();
                fm = toolkit.getFontMetrics(_graphicalFont);
            }
            else
                fm = g.getFontMetrics(_graphicalFont);

            //Recovery of the rectangle of the largest child node.(in order to modify the '_leftLabelOffset')
            Dimension   labelDim = getLeftLabelSize(g);
            Dimension   imageDim = getImageSize(map);
            int         offx = 0, offy = 0;

            if (_mapNode.isResizable())
            {
                float       zoom = map.getZoom();

                imageDim = new Dimension((int)(imageDim.height * zoom),
                    (int)(imageDim.height * zoom));
            }

            offx = -imageDim.width / 2 - labelDim.width - 2;
            offy = -fm.getHeight() / 2;

            // Enlarge the size if there is a border
            if (_mapNode.getColorFrameExt() != null)
                offx -= 8;
            else if (_mapNode.getColorFrameInt() != null)
                offx -= 4;

            _leftLabelOffset = new Point(offx, offy);
        }

        return _leftLabelOffset;
    }

    /**
     * Gets the offset between the center of the node and the left upper corner of the right label.
     *
     * @param  map the node map
     * @param  g   the graphics context
     * @return the offset point
     */
    protected Point getRightLabelOffset(LyStrutsMapCanvas map, Graphics g)
    {
        return getRightLabelOffset(map, g, false);
    }

    /**
     * Returns the offset between the center of the node and the left upper corner of the right label.
     *
     * @param  map         the node map
     * @param  g           the graphics context
     * @param  recalculate indicates if the offset has to be recalculated or not
     * @return the label offset
     */
    @SuppressWarnings("deprecation")
	protected Point getRightLabelOffset(LyStrutsMapCanvas map, Graphics g, boolean recalculate)
    {
        if (_rightLabelOffset == null || recalculate || _mapNode.isResizable())
        {
            FontMetrics     fm;

            if (g == null)
            {
                Toolkit toolkit = Toolkit.getDefaultToolkit();
                fm = toolkit.getFontMetrics(_graphicalFont);
            }
            else
                fm = g.getFontMetrics(_graphicalFont);

            //Recovery of the rectangle of the largest child node.(in order to modify the '_rightLabelOffset')
            Dimension   imageDim = getImageSize(map);
            int         offx = 0, offy = 0;

            if (_mapNode.isResizable())
            {
                float       zoom = map.getZoom();

                imageDim = new Dimension((int)(imageDim.width * zoom),
                    (int)(imageDim.height * zoom));
            }

            offx = imageDim.width / 2 + 2;
            offy = -fm.getHeight() / 2;

            // Enlarge the size if there is a border
            if (_mapNode.getColorFrameExt() != null)
                offx += 8;
            else if (_mapNode.getColorFrameInt() != null)
                offx += 4;

            _rightLabelOffset = new Point(offx, offy);
        }

        return _rightLabelOffset;
    }

    /**
     * Gets the offset between the center of the node and the left upper corner of the top label.
     *
     * @param  map the node map
     * @param  g   the graphics context
     * @return the offset point
     */
    protected Point getTopLabelOffset(LyStrutsMapCanvas map, Graphics g)
    {
        return getTopLabelOffset(map, g, false);
    }

    /**
     * Returns the offset between the center of the node and the left upper corner of the top label.
     *
     * @param  map         the node map
     * @param  g           the graphics context
     * @param  recalculate indicates if the offset has to be recalculated or not
     * @return the label offset
     */
    @SuppressWarnings("deprecation")
	protected Point getTopLabelOffset(LyStrutsMapCanvas map, Graphics g, boolean recalculate)
    {
        if (_topLabelOffset == null || recalculate || _mapNode.isResizable())
        {
            FontMetrics     fm;

            if (g == null)
            {
                Toolkit toolkit = Toolkit.getDefaultToolkit();
                fm = toolkit.getFontMetrics(_graphicalFont);
            }
            else
                fm = g.getFontMetrics(_graphicalFont);

            //Recovery of the rectangle of the largest child node.(in order to modify the '_topLabelOffset')
            Dimension       labelDim = getTopLabelSize(g);
            Dimension       imageDim = getImageSize(map);
            int             offx = 0, offy = 0;

            if (_mapNode.isResizable())
            {
                float       zoom = map.getZoom();

                imageDim = new Dimension((int)(imageDim.width * zoom),
                    (int)(imageDim.height * zoom));
            }

            offx = -labelDim.width / 2;
            offy = -imageDim.height / 2 - fm.getHeight() - 2;

            // Enlarge the size if there is a border
            if (_mapNode.getColorFrameExt() != null)
                offy -= 8;
            else if (_mapNode.getColorFrameInt() != null)
                offy -= 4;

            _topLabelOffset = new Point(offx, offy);
        }

        return _topLabelOffset;
    }

    /**
     * Gets the offset between the center of the node and the left upper corner of the bottom label.
     *
     * @param  map the node map
     * @param  g   the graphics context
     * @return the offset point
     */
    protected Point getBottomLabelOffset(LyStrutsMapCanvas map, Graphics g)
    {
        return getBottomLabelOffset(map, g, false);
    }

    /**
     * Returns the offset between the center of the node and the left upper corner of the bottom label.
     *
     * @param  map         the node map
     * @param  g           the graphics context
     * @param  recalculate indicates if the offset has to be recalculated or not
     * @return the label offset
     */
    protected Point getBottomLabelOffset(LyStrutsMapCanvas map, Graphics g, boolean recalculate)
    {
        if (_bottomLabelOffset == null || recalculate || _mapNode.isResizable())
        {
            //Recovery of the rectangle of the largest child node.(in order to modify the '_bottomLabelOffset')
            Dimension       labelDim = getBottomLabelSize(g);
            Dimension       imageDim = getImageSize(map);
            int             offx = 0, offy = 0;
            Color           internalFrameColor = getInternalFrameColor();
            Color           externalFrameColor = getExternalFrameColor();

            if (_mapNode.isResizable())
            {
                float       zoom = map.getZoom();

                imageDim = new Dimension((int)(imageDim.width * zoom),
                    (int)(imageDim.height * zoom));
            }

            // Add a few extra pixels for external frame
            if (externalFrameColor != null)
                imageDim = new Dimension(imageDim.width + 16, imageDim.height + 16);
            else if (internalFrameColor != null)
                imageDim = new Dimension(imageDim.width + 8, imageDim.height + 8);

            offx = -labelDim.width / 2;
            offy = imageDim.height / 2 + 3;

            // Enlarge the size if there is a border
            if (_mapNode.getColorFrameExt() != null)
                offy += 8;
            else if (_mapNode.getColorFrameInt() != null)
                offy += 4;

            _bottomLabelOffset = new Point(offx, offy);
        }

        return _bottomLabelOffset;
    }

    /**
     * Draws the label of the node.
     *
     * @param map the node map
     * @param g   the graphics context
     * @param x   the x position of the node center
     * @param y   the y position of the node center
     */
    protected void drawLabel(LyStrutsMapCanvas map, Graphics g, int x, int y)
    {
        // Recalculate the offset and size if label is extern or if labelCache
        // is null
        boolean recalculate = (_mapNode.getLabelAlign() == LyMapNode.LabelAlign.EXTERN);

        if (_labelCache == null)
            recalculate = true;

        String label = getLabel();

        if ((label == null) || (label.length() == 0))
            return;

        Dimension   labelDim = getLabelSize(g, recalculate);
        Point       labelOffset = getLabelOffset(map, g, x, y, recalculate);
        int         xstart, ystart;

        // Calculate the label position
        xstart = x + labelOffset.x;
        ystart = y + labelOffset.y;
        
        drawLabelAtPosition(map, g, label, xstart, ystart, labelDim);
    } // end method drawLabel

    /**
     * Draws the left label of the node.
     *
     * @param map the node map
     * @param g   the graphics context
     * @param x   the x position of the node center
     * @param y   the y position of the node center
     */
    protected void drawLeftLabel(LyStrutsMapCanvas map, Graphics g, int x, int y)
    {
        // Recalculate the offset and size if label cache is null
        boolean recalculate = (_leftLabelCache == null);
        String  label = getLeftLabel();

        if ((label == null)
            || (label.length() == 0))
            return;

        Dimension       labelDim = getLeftLabelSize(g, recalculate);
        Point           labelOffset = getLeftLabelOffset(map, g, recalculate);
        int             xstart, ystart;

        // Calculate the label position
        xstart = x + labelOffset.x;
        ystart = y + labelOffset.y;
        
        drawLabelAtPosition(map, g, label, xstart, ystart, labelDim);
    }

    /**
     * Draws the right label of the node.
     *
     * @param map the node map
     * @param g   the graphics context
     * @param x   the x position of the node center
     * @param y   the y position of the node center
     */
    protected void drawRightLabel(LyStrutsMapCanvas map, Graphics g, int x, int y)
    {
        // Recalculate the offset and size if label cache is null
        boolean recalculate = (_rightLabelCache == null);
        String  label = getRightLabel();

        if ((label == null)
            || (label.length() == 0))
            return;

        Dimension       labelDim = getRightLabelSize(g, recalculate);
        Point           labelOffset = getRightLabelOffset(map, g, recalculate);
        int             xstart, ystart;

        // Calculate the label position
        xstart = x + labelOffset.x;
        ystart = y + labelOffset.y;
        
        drawLabelAtPosition(map, g, label, xstart, ystart, labelDim);
    }

    /**
     * Draws the top label of the node.
     *
     * @param map the node map
     * @param g   the graphics context
     * @param x   the x position of the node center
     * @param y   the y position of the node center
     */
    protected void drawTopLabel(LyStrutsMapCanvas map, Graphics g, int x, int y)
    {
        // Recalculate the offset and size if label cache is null
        boolean recalculate = (_topLabelCache == null);
        String  label = getTopLabel();

        if ((label == null)
            || (label.length() == 0))
            return;

        Dimension       labelDim = getTopLabelSize(g, recalculate);
        Point           labelOffset = getTopLabelOffset(map, g, recalculate);
        int             xstart, ystart;

        // Calculate the label position
        xstart = x + labelOffset.x;
        ystart = y + labelOffset.y;
        
        drawLabelAtPosition(map, g, label, xstart, ystart, labelDim);
    }

    /**
     * Draws the bottom label of the node.
     *
     * @param map the node map
     * @param g   the graphics context
     * @param x   the x position of the node center
     * @param y   the y position of the node center
     */
    protected void drawBottomLabel(LyStrutsMapCanvas map, Graphics g, int x, int y)
    {
        // Recalculate the offset and size if label cache is null
        boolean recalculate = (_bottomLabelCache == null);
        String  label = getBottomLabel();

        if ((label == null)
            || (label.length() == 0))
            return;

        Dimension       labelDim = getBottomLabelSize(g, recalculate);
        Point           labelOffset = getBottomLabelOffset(map, g, recalculate);
        int             xstart, ystart;

        // Calculate the label position
        xstart = x + labelOffset.x;
        ystart = y + labelOffset.y;
        
        drawLabelAtPosition(map, g, label, xstart, ystart, labelDim);
    }

    /**
     * Draws a label of the node.
     *
     * @param map      the node map
     * @param g        the graphics context
     * @param label    the label to be drawn
     * @param xstart   the x position of the node center
     * @param ystart   the y position of the node center
     * @param labelDim the dimension of the label to be drawn
     */
    protected void drawLabelAtPosition(LyStrutsMapCanvas map, Graphics g, String label,
        int xstart, int ystart, Dimension labelDim)
    {
        boolean fillLabel = testResource("LY_FILL_MAP_NODE_LABEL");

        if (fillLabel)
        {
            g.setColor(map.getBackground());
            g.fillRect(xstart, ystart, labelDim.width, labelDim.height);
        }

        // Draw the label
        FontMetrics fm = g.getFontMetrics(_graphicalFont);

        String labelColor = _mapNode.getFontColor();
        
        if (labelColor == null)
            g.setColor(getColor("LY_MAP_FONT_COLOR"));
        else
            g.setColor(getColor(labelColor));
        g.setFont(_graphicalFont);
        g.drawString(label, xstart, ystart + fm.getAscent());
    } // end method drawLabel

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#getBoundingBox(leon.view.web.struts.map.LyStrutsMapCanvas)
     */
    public Ly3DBox getBoundingBox(LyStrutsMapCanvas map)
    {
        Dimension dim = getImageSize(map);
        int       width = dim.width;
        int       height = dim.height;

        return new Ly3DBox(_center.getX() - (width / 2), _center.getY() - (height / 2),
            _center.getZ(), _center.getX() + (width / 2), _center.getY() + (height / 2),
            _center.getZ());
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#getBoundingRectangle(leon.view.web.struts.map.LyStrutsMapCanvas)
     */
    public Rectangle getBoundingRectangle(LyStrutsMapCanvas map)
    {
        Point       center = map.convert(_center);
        Dimension   dimImage = getImageSize(map);
        Point       labelOffset = getLabelOffset(map, null, center.x, center.y);
        Dimension   labelSize = getLabelSize(map.getGraphics());
        Point       leftLabelOffset = getLeftLabelOffset(map, null);
        Dimension   leftLabelSize = getLeftLabelSize(map.getGraphics());
        Point       rightLabelOffset = getRightLabelOffset(map, null);
        Dimension   rightLabelSize = getRightLabelSize(map.getGraphics());
        Point       topLabelOffset = getTopLabelOffset(map, null);
        Dimension   topLabelSize = getTopLabelSize(map.getGraphics());
        Point       bottomLabelOffset = getBottomLabelOffset(map, null);
        Dimension   bottomLabelSize = getBottomLabelSize(map.getGraphics());
        float       zoom = map.getZoom();

        if (_mapNode.isResizable())
        {
            dimImage = new Dimension((int)(dimImage.width * zoom),
                (int)(dimImage.height * zoom));
        }
        
        int     xLabelOffset = (labelOffset == null ? 0 : labelOffset.x);
        int     yLabelOffset = (labelOffset == null ? 0 : labelOffset.y);
        int     labelWidth = (labelSize == null ? 0 : labelSize.width);
        int     labelHeight = (labelSize == null ? 0 : labelSize.height);
        int     xLeftLabelOffset = (leftLabelOffset == null ? 0 : leftLabelOffset.x);
        int     yLeftLabelOffset = (leftLabelOffset == null ? 0 : leftLabelOffset.y);
        int     leftLabelWidth = (leftLabelSize == null ? 0 : leftLabelSize.width);
        int     leftLabelHeight = (leftLabelSize == null ? 0 : leftLabelSize.height);
        int     xRightLabelOffset = (rightLabelOffset == null ? 0 : rightLabelOffset.x);
        int     yRightLabelOffset = (rightLabelOffset == null ? 0 : rightLabelOffset.y);
        int     rightLabelWidth = (rightLabelSize == null ? 0 : rightLabelSize.width);
        int     rightLabelHeight = (rightLabelSize == null ? 0 : rightLabelSize.height);
        int     xTopLabelOffset = (topLabelOffset == null ? 0 : topLabelOffset.x);
        int     yTopLabelOffset = (topLabelOffset == null ? 0 : topLabelOffset.y);
        int     topLabelWidth = (topLabelSize == null ? 0 : topLabelSize.width);
        int     topLabelHeight = (topLabelSize == null ? 0 : topLabelSize.height);
        int     xBottomLabelOffset = (bottomLabelOffset == null ? 0 : bottomLabelOffset.x);
        int     yBottomLabelOffset = (bottomLabelOffset == null ? 0 : bottomLabelOffset.y);
        int     bottomLabelWidth = (bottomLabelSize == null ? 0 : bottomLabelSize.width);
        int     bottomLabelHeight = (bottomLabelSize == null ? 0 : bottomLabelSize.height);

        Rectangle   imageRect = new Rectangle(center.x - dimImage.width / 2,
            center.y - dimImage.height / 2, dimImage.width, dimImage.height);
        Rectangle   labelRect = new Rectangle(center.x + xLabelOffset,
            center.y + yLabelOffset, labelWidth, labelHeight);
        Rectangle   leftLabelRect = new Rectangle(center.x + xLeftLabelOffset,
            center.y + yLeftLabelOffset, leftLabelWidth, leftLabelHeight);
        Rectangle   rightLabelRect = new Rectangle(center.x + xRightLabelOffset,
            center.y + yRightLabelOffset, rightLabelWidth, rightLabelHeight);
        Rectangle   topLabelRect = new Rectangle(center.x + xTopLabelOffset,
            center.y + yTopLabelOffset, topLabelWidth, topLabelHeight);
        Rectangle   bottomLabelRect = new Rectangle(center.x + xBottomLabelOffset,
            center.y + yBottomLabelOffset, bottomLabelWidth, bottomLabelHeight);
        Rectangle   checkboxRect = getSelectBoundingRectangle(map);

        Rectangle   result = imageRect;
        
        if ((_labelWidth >= 0) && (_labelHeight >= 0))
            result = result.union(labelRect);

        if ((_leftLabelWidth >= 0) && (_leftLabelHeight >= 0))
            result = result.union(leftLabelRect);

        if ((_rightLabelWidth >= 0) && (_rightLabelHeight >= 0))
            result = result.union(rightLabelRect);

        if ((_topLabelWidth >= 0) && (_topLabelHeight >= 0))
            result = result.union(topLabelRect);

        if ((_bottomLabelWidth >= 0) && (_bottomLabelHeight >= 0))
            result = result.union(bottomLabelRect);

        if (checkboxRect != null)
            result = result.union(checkboxRect);
        
        return result;
    } // end method getBoundingRectangle

    /**
     * Gets the node bounding rectangle.
     *
     * @param  map the node map
     * @return the bounding rectangle
     */
    public Rectangle getNodeBoundingRectangle(LyStrutsMapCanvas map)
    {
        Point     center = map.convert(_center);
        Dimension dimImage = getImageSize(map);
        Rectangle imageRect = new Rectangle(center.x - (dimImage.width / 2),
                center.y
                - (dimImage.height / 2), dimImage.width, dimImage.height);

        return imageRect;
    }

    /**
     * Gets the bounding rectangle where the checkbox can be selected.
     *
     * @param  map the node map
     * @return the bounding rectangle where the checkbox can be selected
     */
    public Rectangle getSelectBoundingRectangle(LyStrutsMapCanvas map)
    {
        Image checkBox;

        if (_mapNode.isSelected())
            checkBox = _selectedCheckboxImage;
        else
            checkBox = _checkboxImage;

        if (checkBox == null)
            return null;

        int   width = checkBox.getWidth(null);
        int   height = checkBox.getHeight(null);
        Point position = getCheckBoxPosition(map, width, height);

        return new Rectangle(position.x, position.y, width, height);
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#getClippingArea(leon.view.web.struts.map.LyStrutsMapCanvas,
     *      java.awt.Image)
     */
    public Shape getClippingArea(LyStrutsMapCanvas map, Image image)
    {
        return getBoundingRectangle(map);
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#getLayer()
     */
    public int getLayer()
    {
        int layer = _mapNode.getLayer();

        return ((layer < 0) ? LyStrutsMapCanvas.NODE_LAYER : layer);
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#getLabel()
     */
    public String getLabel()
    {
        if (_labelCache == null)
            _labelCache = translate(_mapNode.getLabel());

        return _labelCache;
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalMapNode#getLeftLabel()
     */
    public String getLeftLabel()
    {
        if (_leftLabelCache == null)
            _leftLabelCache = translate(_mapNode.getLeftLabel());

        return _leftLabelCache;
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalMapNode#getRightLabel()
     */
    public String getRightLabel()
    {
        if (_rightLabelCache == null)
            _rightLabelCache = translate(_mapNode.getRightLabel());

        return _rightLabelCache;
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalMapNode#getTopLabel()
     */
    public String getTopLabel()
    {
        if (_topLabelCache == null)
            _topLabelCache = translate(_mapNode.getTopLabel());

        return _topLabelCache;
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalMapNode#getBottomLabel()
     */
    public String getBottomLabel()
    {
        if (_bottomLabelCache == null)
            _bottomLabelCache = translate(_mapNode.getBottomLabel());

        return _bottomLabelCache;
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#getTooltip()
     */
    public String getTooltip()
    {
        return translate(_mapNode.getTooltip());
    }

    /**
     * Sets attribute tooltip to the given value. This method does nothing.
     *
     * @param tooltip New value of the tooltip
     * @see   #getTooltip()
     */
    public void setTooltip(String tooltip)
    {
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#isAtPosition(leon.view.web.struts.map.LyStrutsMapCanvas,
     *      java.awt.Graphics, int, int)
     */
    public boolean isAtPosition(LyStrutsMapCanvas map, Graphics g, int x, int y)
    {
        Dimension dimImage = getImageSize(map);

        if (_mapNode.isResizable())
        {
            float zoom = map.getZoom();

            dimImage = new Dimension((int)(dimImage.width * zoom), (int)(dimImage.height * zoom));
        }

        Dimension dimLabel = getLabelSize(g);
        Point     point = map.convert(_center);
        Point     labelOffset = getLabelOffset(map, g, x, y);
        int       xnode = point.x, ynode = point.y;
        int       xlabel = (point.x + labelOffset.x);
        int       ylabel = (point.y + labelOffset.y);

        // Check if (x, y) in the image
        if ((x >= (xnode - (dimImage.width / 2))) && (x <= (xnode + (dimImage.width / 2)))
                && (y >= (ynode - (dimImage.height / 2))) && (y <= (ynode + (dimImage.height / 2))))
            return true;

        // Check if (x,y) in the label
        if ((x >= xlabel) && (x <= (xlabel + dimLabel.width)) && (y >= ylabel)
                && (y <= (ylabel + dimLabel.height)))
            return true;

        return false;
    } // end method isAtPosition

    /**
     * (non-Javadoc).
     *
     * @see leon.view.getinterface.LyGetIdInterface#getId()
     */
    public String getId()
    {
        return _mapNode.getId();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetHiddenInterface#setHidden(boolean)
     */
    public void setHidden(boolean hidden)
    {
        LyComponent parent = _mapNode.getParent();

        if (parent instanceof LyMapContent)
            parent = parent.getParent();

        if (parent instanceof LyMap)
        {
            LyMap       map = (LyMap)parent;
            LyStrutsMap  webMap = (LyStrutsMap)map.getInternalView();
            
            if (!hidden)
                webMap.addComponent(_mapNode, null);
            else
                webMap.removeComponent(_mapNode);
        }
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#isSelected()
     */
    public boolean isSelected()
    {
        return _mapNode.isSelected();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#translate(float, float, float)
     */
    public void translate(float dx, float dy, float dz)
    {
        _center.translate(dx, dy, dz);
        _mapNode.setX(_center.getX());
        _mapNode.setY(_center.getY());
        _mapNode.setZ(_center.getZ());
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#isOverlay()
     */
    public boolean isOverlay()
    {
        return true;
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#isHidden()
     */
    public boolean isHidden()
    {
        return _mapNode.isHidden();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#isSensitive()
     */
    public boolean isSensitive()
    {
        return (_mapNode.isSensitive() && !_mapNode.isHidden());
    }

    /**
     * Sets the parent node.
     *
     * @param node the new parent node
     * @see   #getParentNode()
     */
    public void setParentNode(LyStrutsMapNode node)
    {
        _parentNode = node;
    }

    /**
     * Gets the center of the node.
     *
     * @return the node center
     */
    public LyPoint getCenter()
    {
        return _center;
    }

    /**
     * Gets the parent node of the node.
     *
     * @return the parent node
     * @see    #setParentNode(LyStrutsMapNode)
     */
    public LyStrutsMapNode getParentNode()
    {
        return _parentNode;
    }

    /**
     * Gets the layout info of the node.
     *
     * @return the layout info of the node
     */
    public String getLayoutInfo()
    {
        return _mapNode.getLayoutInfo();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
     */
    public void writeHtml(LyWebPageWriter pw)
    {
        Dimension dim = getImageSize(_map);
        float     zoom = _map.getZoom();
        Color     internalFrameColor = getInternalFrameColor();
        Color     externalFrameColor = getExternalFrameColor();

        if (_mapNode.isResizable())
            dim = new Dimension((int)(dim.width * zoom), (int)(dim.height * zoom));

        // Add a few extra pixels for selection
        if (_mapNode.isSensitive())
            dim = new Dimension(dim.width + 8, dim.height + 8);

        // Add a few extra pixels for external frame
        if (externalFrameColor != null)
            dim = new Dimension(dim.width + 16, dim.height + 16);
        else if (internalFrameColor != null)
            dim = new Dimension(dim.width + 8, dim.height + 8);

        int    maxWidth = dim.width;
        int    maxHeight = dim.height;

        // Generate frame if needed
        String frameTag = generateFrameImage(_map, maxWidth, maxHeight);
        // generate cross if needed
        String crossTag = generateCrossImage(_map, maxWidth, maxHeight);
        // Generate node if needed
        String nodeTag = getNodeTag(_map, maxWidth, maxHeight);

        // Generate Small image
        String smallImageTag = getSmallImageTag(_map, maxWidth, maxHeight);

        if (isSelected())
            pw.append("<input type='hidden' name='" + LyWebConstants.SELECTION_ITEM + getId()
                + "' value='" + LyWebConstants.SELECTION_ITEM + getId() + "' />");

        if (frameTag != null)
            pw.append(frameTag);

        if (nodeTag != null)
            pw.append(nodeTag);

        if (smallImageTag != null)
            pw.append(smallImageTag);

        if (crossTag != null)
            pw.append(crossTag);
    } // end method writeHtml

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#getImages()
     */
    public Image[] getImages()
    {
        Image image = null;
        Image smallImage = null;

        if (_mapNode.getStyle() == LyMapNode.Style.IMAGE)
        {
            if (_mapNode.getImage() != null)
                image = getImage(_mapNode.getImage());
            else
                image = getImage("LY_DEFAULT_MAP_NODE");
        }

        if (_mapNode.getSmallImage() != null)
            smallImage = getImage(_mapNode.getSmallImage());
        else
            smallImage = null;

        return (new Image[] { image, smallImage });
    }

    /**
     * Gets the image of the node.
     *
     * @return the image of the node
     * @see    #setImage(String)
     */
    public Image getImage()
    {
        if (_mapNode.getStyle() == LyMapNode.Style.IMAGE)
        {
            if (_mapNode.getImage() != null)
            {
                if ((_imageString == null) || !_imageString.equals(_mapNode.getImage()) || (_image == null))
                {
                    _imageString = _mapNode.getImage();
                    _image = getImage(_mapNode.getImage());
                }
            }
            else
            {
                _imageString = new String("LY_DEFAULT_MAP_NODE");
                _image = getImage("LY_DEFAULT_MAP_NODE");
            }
        }

        return _image;
    }

    /**
     * Gets the small image of the node.
     *
     * @return the small image of the node
     */
    public Image getSmallImage()
    {
        if (_mapNode.getSmallImage() != null)
        {
            if ((_smallImageString == null) || !_smallImageString.equals(_mapNode.getSmallImage()))
            {
                _smallImageString = _mapNode.getSmallImage();
                _smallImage = getImage(_smallImageString);
            }
        }
        else
        {
            _smallImageString = null;
            _smallImage = null;
        }

        return _smallImage;
    }

    /**
     * Gets the node color.
     *
     * @return the node color
     */
    public Color getNodeColor()
    {
        if (_mapNode.getColor() != null)
        {
            if ((_nodeColorString != null) && _nodeColorString.equals(_mapNode.getColor()))
                return _nodeColor;

            _nodeColorString = _mapNode.getColor();
            _nodeColor = getColor(_nodeColorString);
        }
        else if (_nodeColor == null)
            _nodeColor = getColor("LY_MAP_NODE_COLOR");

        return _nodeColor;
    }

    /**
     * Gets the cross color.
     *
     * @return the cross color
     */
    public Color getCrossColor()
    {
        if (_mapNode.getColorCross() != null)
        {
            if ((_crossColorString == null) || !_crossColorString.equals(_mapNode.getColorCross()))
            {
                _crossColorString = _mapNode.getColorCross();
                _crossColor = getColor(_crossColorString);
            }

            return _crossColor;
        }
        else
            return null;
    }

    /**
     * Gets the internal frame color.
     *
     * @return the internal frame color
     */
    public Color getInternalFrameColor()
    {
        if (_mapNode.getColorFrameInt() != null)
        {
            if ((_internalFrameColorString == null)
                    || !_internalFrameColorString.equals(_mapNode.getColorFrameInt()))
            {
                _internalFrameColorString = _mapNode.getColorFrameInt();
                _internalFrameColor = getColor(_internalFrameColorString);
            }

            return _internalFrameColor;
        }
        else
            return null;
    }

    /**
     * Gets the external frame color.
     *
     * @return the external frame color
     */
    public Color getExternalFrameColor()
    {
        if (_mapNode.getColorFrameExt() != null)
        {
            if ((_externalFrameColorString == null)
                    || !_externalFrameColorString.equals(_mapNode.getColorFrameExt()))
            {
                _externalFrameColorString = _mapNode.getColorFrameExt();
                _externalFrameColor = getColor(_externalFrameColorString);
            }

            return _externalFrameColor;
        }
        else
            return null;
    }

    /**
     * Gets the outline color.
     *
     * @return the internal frame color
     */
    public Color getOutlineColor()
    {
        if (_outlineColor == null)
            _outlineColor = getColor("LY_MAP_OUTLINE_COLOR");

        return _outlineColor;
    }

    /**
     * Gets the dimension of the image part of the node.
     *
     * @return the dimension of the image part of the node
     * @see    LyStrutsMapNode#getImageSize(Component, boolean)
     */
    public int[] getImageSize()
    {
        if (_imageString == null)
            _imageString = _mapNode.getImage();
        
        if ((_imageWidth < 0) || (_imageHeight < 0))
        {
            LyEnvironment   environment = getEnvironment();
            int[]           dim = environment.getImageSize(_imageString, false);
           
            if (dim == null)
            {
                Image graphicalImage = getImage();
                
                if (graphicalImage != null)
                {  
                    _imageWidth = graphicalImage.getWidth(null);
                    _imageHeight = graphicalImage.getHeight(null);
                    
                }
            }
            else
            {
                _imageWidth = dim[0];
                _imageHeight = (dim.length > 1 ? dim[1] : dim[0]);
            }
        }
            
        return new int[] { _imageWidth, _imageHeight };
    }
    
    /**
     * Gets the dimension of the image part of the node.
     *
     * @param  comp the map of the node
     * @return the dimension of the image part of the node
     * @see    LyStrutsMapNode#getImageSize(Component, boolean)
     */
    public Dimension getImageSize(Component comp)
    {
        return getImageSize(comp, false);
    }

    /**
     * Gets the dimension of the image part of the node (without the frame) if the style of the node
     * is {@link leon.view.LyMapNode#Style.IMAGE}, the default size otherwise.
     *
     * @param  comp the map
     * @return the dimension of the image part of the node (without the frame)
     */
    public Dimension getSymbolSize(Component comp)
    {
        int sizeParam = _mapNode.getSize();
        int heightParam = _mapNode.getHeight();
        int width = (sizeParam > 0 ? sizeParam : DEFAULT_SIZE);
        int height = (heightParam > 0 ? heightParam : width);
        LyMapNode.Style style = _mapNode.getStyle();
        
        if (style == LyMapNode.Style.IMAGE)
        {
            int[]   dim = getImageSize();
            
            width = dim[0];
            height = dim[1];
        }

        Dimension frameSize = new Dimension(width, height);

        return frameSize;
    }

    /**
     * Gets the dimension of the image part of the node.
     *
     * @param  comp        the map
     * @param  recalculate indicates if the size must be recalculated
     * @return the dimension of the image part of the node
     */
    public Dimension getImageSize(Component comp, boolean recalculate)
    {
        if (recalculate || (_imageWidth < 0) || (_imageHeight < 0))
        {
            int dx = 0;
            int dy = 0;

            // Enlarge the size if there is a border
            if (_mapNode.getColorFrameExt() != null)
            {
                dx = 16;
                dy = 16;
            }
            else if (_mapNode.getColorFrameInt() != null)
            {
                dx = 10;
                dy = 10;
            }

            LyMapNode.Style  style = _mapNode.getStyle();
            
            if (style == LyMapNode.Style.IMAGE)
            {
                int[]   dim = getImageSize();
                
                _imageWidth = dim[0];
                _imageHeight = dim[1];
            }
            else
            {
                _imageWidth = _mapNode.getSize();
                if (_imageWidth < 0)
                    _imageWidth = DEFAULT_SIZE;
                _imageHeight = _imageWidth;
            }

            _imageWidth += dx;
            _imageHeight += dy;
        }

        Dimension result = new Dimension(_imageWidth, _imageHeight);

        return result;
    } // end method getImageSize

    /**
     * Gets the dimension of the label part of the node.
     *
     * @param  g the graphics context
     * @return the dimension of the label part of the node
     */
    public Dimension getLabelSize(Graphics g)
    {
        return getLabelSize(g, false);
    }

    /**
     * Gets the dimension of the label part of the node.
     *
     * @param  g           the graphics context
     * @param  recalculate indicates if the size must be recalculated
     * @return the dimension of the label part of the node
     */
    @SuppressWarnings("deprecation")
	public Dimension getLabelSize(Graphics g, boolean recalculate)
    {
        if ((recalculate) || (_labelWidth < 0) || (_labelHeight < 0) || (_labelCache == null))
        {
            String label = getLabel();

            if (label == null)
                return new Dimension(0, 0);

            FontMetrics fm;

            if (g == null)
            {
                Toolkit toolkit = Toolkit.getDefaultToolkit();

                fm = toolkit.getFontMetrics(_graphicalFont);
            }
            else
                fm = g.getFontMetrics(_graphicalFont);

            _labelHeight = fm.getHeight();

            // Handle multiple lines in label separated by <br/>
            _labelWidth = 0;
            String s = label;
            int    idx = s.indexOf("<br/>"), w;

            while (idx >= 0)
            {
                w = fm.stringWidth(s.substring(0, idx));
                if (_labelWidth < w)
                    _labelWidth = w;
                s = s.substring(idx + 5);
                idx = s.indexOf("<br/>");
            }
            w = fm.stringWidth(s);
            if (_labelWidth < w)
                _labelWidth = w;
        }

        return (new Dimension(_labelWidth, _labelHeight));
    } // end method getLabelSize

    /**
     * Returns the dimension of the left label part of the node.
     * 
     * @param  g the graphics context
     * @return the dimension of the left label part of the node
     */
    public Dimension getLeftLabelSize(Graphics g)
    {
        return getLeftLabelSize(g, false);
    }

    /**
     * Returns the dimension of the left label part of the node.
     * 
     * @param  g the graphics context
     * @param  recalculate indicates if the size must be recalculated
     * @return the dimension of the left label part of the node
     */
    @SuppressWarnings("deprecation")
	public Dimension getLeftLabelSize(Graphics g, boolean recalculate)
    {
        if ((recalculate) || (_leftLabelWidth < 0) ||
            (_leftLabelHeight < 0) || (_leftLabelCache == null))
        {
            String  label = getLeftLabel();
            
            if (label == null)
                return new Dimension(0, 0);

            FontMetrics     fm;

            if (g == null)
            {
                Toolkit toolkit = Toolkit.getDefaultToolkit();
                fm = toolkit.getFontMetrics(_graphicalFont);
            }
            else
                fm = g.getFontMetrics(_graphicalFont);

            _leftLabelHeight = fm.getHeight();
            _leftLabelWidth = fm.stringWidth(label);
        }

        return (new Dimension(_leftLabelWidth, _leftLabelHeight));
    }

    /**
     * Returns the dimension of the right label part of the node.
     * 
     * @param  g the graphics context
     * @return the dimension of the right label part of the node
     */
    public Dimension getRightLabelSize(Graphics g)
    {
        return getRightLabelSize(g, false);
    }

    /**
     * Returns the dimension of the right label part of the node.
     * 
     * @param  g the graphics context
     * @param  recalculate indicates if the size must be recalculated
     * @return the dimension of the right label part of the node
     */
    @SuppressWarnings("deprecation")
	public Dimension getRightLabelSize(Graphics g, boolean recalculate)
    {
        if ((recalculate) || (_rightLabelWidth < 0) ||
            (_rightLabelHeight < 0) || (_rightLabelCache == null))
        {
            String  label = getRightLabel();
            
            if (label == null)
                return new Dimension(0, 0);

            FontMetrics     fm;

            if (g == null)
            {
                Toolkit toolkit = Toolkit.getDefaultToolkit();
                fm = toolkit.getFontMetrics(_graphicalFont);
            }
            else
                fm = g.getFontMetrics(_graphicalFont);

            _rightLabelHeight = fm.getHeight();
            _rightLabelWidth = fm.stringWidth(label);
        }

        return (new Dimension(_rightLabelWidth, _rightLabelHeight));
    }

    /**
     * Returns the dimension of the top label part of the node.
     * 
     * @param  g the graphics context
     * @return the dimension of the top label part of the node
     */
    public Dimension getTopLabelSize(Graphics g)
    {
        return getTopLabelSize(g, false);
    }

    /**
     * Returns the dimension of the top label part of the node.
     * 
     * @param  g the graphics context
     * @param  recalculate indicates if the size must be recalculated
     * @return the dimension of the top label part of the node
     */
    @SuppressWarnings("deprecation")
	public Dimension getTopLabelSize(Graphics g, boolean recalculate)
    {
        if ((recalculate) || (_topLabelWidth < 0) ||
            (_topLabelHeight < 0) || (_topLabelCache == null))
        {
            String  label = getTopLabel();
            
            if (label == null)
                return new Dimension(0, 0);

            FontMetrics     fm;

            if (g == null)
            {
                Toolkit toolkit = Toolkit.getDefaultToolkit();
                fm = toolkit.getFontMetrics(_graphicalFont);
            }
            else
                fm = g.getFontMetrics(_graphicalFont);

            _topLabelHeight = fm.getHeight();
            _topLabelWidth = fm.stringWidth(label);
        }

        return (new Dimension(_topLabelWidth, _topLabelHeight));
    }

    /**
     * Returns the dimension of the bottom label part of the node.
     * 
     * @param  g the graphics context
     * @return the dimension of the bottom label part of the node
     */
    public Dimension getBottomLabelSize(Graphics g)
    {
        return getBottomLabelSize(g, false);
    }

    /**
     * Returns the dimension of the bottom label part of the node.
     * 
     * @param  g the graphics context
     * @param  recalculate indicates if the size must be recalculated
     * @return the dimension of the bottom label part of the node
     */
    @SuppressWarnings("deprecation")
	public Dimension getBottomLabelSize(Graphics g, boolean recalculate)
    {
        if ((recalculate) || (_bottomLabelWidth < 0) ||
            (_bottomLabelHeight < 0) || (_bottomLabelCache == null))
        {
            String  label = getBottomLabel();
            
            if (label == null)
                return new Dimension(0, 0);

            FontMetrics     fm;

            if (g == null)
            {
                Toolkit toolkit = Toolkit.getDefaultToolkit();
                fm = toolkit.getFontMetrics(_graphicalFont);
            }
            else
                fm = g.getFontMetrics(_graphicalFont);

            _bottomLabelHeight = fm.getHeight();
            _bottomLabelWidth = fm.stringWidth(label);
        }

        return (new Dimension(_bottomLabelWidth, _bottomLabelHeight));
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#isDrawableInterface(leon.view.web.struts.map.LyStrutsMapCanvas)
     */
    public boolean isDrawableInterface(LyStrutsMapCanvas map)
    {
        return true;
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetSelectedInterface#setSelected(boolean)
     */
    public void setSelected(boolean selected)
    {
        if (_mapNode.isSensitive() && (_mapNode.isSelected() != selected))
            _mapNode.setSelected(selected);

        LyWebViewManager viewManager = getWebViewManager();
        StringBuffer     script = new StringBuffer();
        String           cssClass = "nodeSelectedImage";

        if (!selected)
            cssClass = "nodeImage";

        script.append("<script type='text/javascript'>\n//<![CDATA[\nChangeObjectClass(\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_node\",\"");
        script.append(cssClass);
        script.append("\");\n");

        cssClass = "nodeSelectedLabel";

        if (!selected)
            cssClass = "nodeLabel";
        
        script.append("ChangeObjectClass(\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_Label\",\"");
        script.append(cssClass);
        script.append("\");\n");
        
        script.append("ChangeObjectClass(\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_leftLabel\",\"");
        script.append(cssClass);
        script.append("\");\n");
        
        script.append("ChangeObjectClass(\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_rightLabel\",\"");
        script.append(cssClass);
        script.append("\");\n");
        
        script.append("ChangeObjectClass(\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_topLabel\",\"");
        script.append(cssClass);
        script.append("\");\n");
        
        script.append("ChangeObjectClass(\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_bottomLabel\",\"");
        script.append(cssClass);
        script.append("\");\n");

        script.append("//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(script.toString());
    } // end method setSelected

    /**
     * Refresh map containing this map link.<br />
     * Call the <code>refreshMapImage()</code> on the <code>LyStrutsMapLink</code> instance.
     */
    public void refresh()
    {
        if (_lastFrameName == null)
        {
            LyComponent     parent = _mapNode.getParent();
            
            while ((parent != null) && !(parent instanceof LyMap))
            {
                parent = parent.getParent();
            }
            
            if (parent instanceof LyMap)
            {
                LyMap       map = (LyMap)parent;
                LyStrutsMap  webMap = (LyStrutsMap)map.getInternalView();
                
                webMap.refreshMapImage();
                webMap.refreshViewerImage();
            }
        }
        else
            super.refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
     */
    public void setLabel(String label)
    {
        // Get rid of cache and reload label
        _labelCache = null;
        _labelWidth = -1;
        _labelHeight = -1;
        _labelOffset = null;
        label = getLabel();

        if (_map == null)
        {
            refresh();
            return;
        }
        
        Point point = _map.convert(_center);
        int   x = point.x, y = point.y;

        Point pos = getLabelOffset(_map, _map.getGraphics(), x, y, true);

        x = pos.x;
        y = pos.y;

        LyWebViewManager viewManager = getWebViewManager();
        StringBuffer     script = new StringBuffer();
        int              layer = _mapNode.getLayer() + 2;
        String           labelDiv = getLabelDivContent(_map, "", layer, x, y);

        script.append("<script type='text/javascript'>\n//<![CDATA[\n");
        
        if (labelDiv != null)
        {
            LyComponent         parent = _mapNode.getParent();

            if (parent instanceof LyMapContent)
                parent = parent.getParent();

            labelDiv = LyString.replace(labelDiv, "\\", "\\\\");
            labelDiv = LyString.replace(labelDiv, "\\\\n", "\\n");
            labelDiv = LyString.replace(labelDiv, "\"", "\\\"");
            labelDiv = LyString.replace(labelDiv, "\\'", "\\\\'");

            script.append("addDiv(\"");
            script.append(_lastFrameName).append("\",\"mapnode_").append(_mapNode.getIdCode()).append("_global\",\"");
            script.append(labelDiv).append("\",\"mapnode_").append(_mapNode.getIdCode()).append("_Label\");\n");
        }
        
        script.append("replaceDivInnerHtml(\"");
        script.append(_lastFrameName).append("\",\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_Label\",\"");
        script.append(label);
        script.append("\");\n");
        script.append("SetDivSize(\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_Label\",").append(_labelWidth);
        script.append(",").append(_labelHeight).append(");\n");
        script.append("moveDiv(\"");
        script.append(_lastFrameName).append("\",\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_Label\",").append(x);
        script.append(",").append(y).append(");\n");

        String	cssClass = "nodeSelectedLabel";

        if (!_mapNode.isSelected())
            cssClass = "nodeLabel";
        
        script.append("ChangeObjectClass(\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_Label\",\"");
        script.append(cssClass);
        script.append("\");\n");
        script.append("ChangeObjectPositionStyle(\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_Label\",\"absolute\");\n");

        script.append("//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(script.toString());
        checkLabels();
    } // end method setLabel

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalMapNode#setLeftLabel(String)
     */
    public void setLeftLabel(String label)
    {
        // Get rid of cache and reload label
        _leftLabelCache = null;
        _leftLabelWidth = -1;
        _leftLabelHeight = -1;
        _leftLabelOffset = null;
        label = getLeftLabel();

        if (_map == null)
        {
            refresh();
            return;
        }
        
        Point   point = _map.convert(_center);
        int     x = point.x, y = point.y;
        Point   pos = getLeftLabelOffset(_map, _map.getGraphics(), true);

        x = pos.x;
        y = pos.y;

        LyWebViewManager    viewManager = getWebViewManager();
        StringBuffer        script = new StringBuffer();
        int                 layer = _mapNode.getLayer() + 2;
        String              leftLabelDiv = getLabelDivContent(_map, "left", layer, x, y);

        script.append("<script type='text/javascript'>\n//<![CDATA[\n");
        
        if (leftLabelDiv != null)
        {
            LyComponent         parent = _mapNode.getParent();

            if (parent instanceof LyMapContent)
                parent = parent.getParent();

            leftLabelDiv = LyString.replace(leftLabelDiv, "\\", "\\\\");
            leftLabelDiv = LyString.replace(leftLabelDiv, "\\\\n", "\\n");
            leftLabelDiv = LyString.replace(leftLabelDiv, "\"", "\\\"");
            leftLabelDiv = LyString.replace(leftLabelDiv, "\\'", "\\\\'");

            script.append("addDiv(\"");
            script.append(_lastFrameName).append("\",\"mapnode_").append(_mapNode.getIdCode()).append("_global\",\"");
            script.append(leftLabelDiv).append("\",\"mapnode_").append(_mapNode.getIdCode()).append("_leftLabel\");\n");
        }
        
        script.append("replaceDivInnerHtml(\"");
        script.append(_lastFrameName).append("\",\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_leftLabel\",\"");
        script.append(label);
        script.append("\");\n");
        script.append("SetDivSize(\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_leftLabel\",").append(_leftLabelWidth);
        script.append(",").append(_leftLabelHeight).append(");\n");
        script.append("moveDiv(\"");
        script.append(_lastFrameName).append("\",\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_leftLabel\",").append(x);
        script.append(",").append(y).append(");\n//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(script.toString());
        checkLabels();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalMapNode#setRightLabel(String)
     */
    public void setRightLabel(String label)
    {
        // Get rid of cache and reload label
        _rightLabelCache = null;
        _rightLabelWidth = -1;
        _rightLabelHeight = -1;
        _rightLabelOffset = null;
        label = getRightLabel();

        if (_map == null)
        {
            refresh();
            return;
        }
        
        Point   point = _map.convert(_center);
        int     x = point.x, y = point.y;
        Point   pos = getRightLabelOffset(_map, _map.getGraphics(), true);

        x = pos.x;
        y = pos.y;

        LyWebViewManager    viewManager = getWebViewManager();
        StringBuffer        script = new StringBuffer();
        int                 layer = _mapNode.getLayer() + 2;
        String              rightLabelDiv = getLabelDivContent(_map, "right", layer, x, y);

        script.append("<script type='text/javascript'>\n//<![CDATA[\n");
        
        if (rightLabelDiv != null)
        {
            LyComponent         parent = _mapNode.getParent();

            if (parent instanceof LyMapContent)
                parent = parent.getParent();

            rightLabelDiv = LyString.replace(rightLabelDiv, "\\", "\\\\");
            rightLabelDiv = LyString.replace(rightLabelDiv, "\\\\n", "\\n");
            rightLabelDiv = LyString.replace(rightLabelDiv, "\"", "\\\"");
            rightLabelDiv = LyString.replace(rightLabelDiv, "\\'", "\\\\'");

            script.append("addDiv(\"");
            script.append(_lastFrameName).append("\",\"mapnode_").append(_mapNode.getIdCode()).append("_global\",\"");
            script.append(rightLabelDiv).append("\",\"mapnode_").append(_mapNode.getIdCode()).append("_rightLabel\");\n");
        }

        script.append("replaceDivInnerHtml(\"");
        script.append(_lastFrameName).append("\",\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_rightLabel\",\"");
        script.append(label);
        script.append("\");\n");
        script.append("SetDivSize(\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_rightLabel\",").append(_rightLabelWidth);
        script.append(",").append(_rightLabelHeight).append(");\n");
        script.append("moveDiv(\"");
        script.append(_lastFrameName).append("\",\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_rightLabel\",").append(x);
        script.append(",").append(y).append(");\n//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(script.toString());
        checkLabels();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalMapNode#setTopLabel(String)
     */
    public void setTopLabel(String label)
    {
        // Get rid of cache and reload label
        _topLabelCache = null;
        _topLabelWidth = -1;
        _topLabelHeight = -1;
        _topLabelOffset = null;
        label = getTopLabel();

        if (_map == null)
        {
            refresh();
            return;
        }
        
        Point   point = _map.convert(_center);
        int     x = point.x, y = point.y;

        Point pos = getTopLabelOffset(_map, _map.getGraphics(), true);

        x = pos.x;
        y = pos.y;

        LyWebViewManager    viewManager = getWebViewManager();
        StringBuffer        script = new StringBuffer();
        int                 layer = _mapNode.getLayer() + 2;
        String              topLabelDiv = getLabelDivContent(_map, "top", layer, x, y);

        script.append("<script type='text/javascript'>\n//<![CDATA[\n");
        
        if (topLabelDiv != null)
        {
            LyComponent         parent = _mapNode.getParent();

            if (parent instanceof LyMapContent)
                parent = parent.getParent();

            topLabelDiv = LyString.replace(topLabelDiv, "\\", "\\\\");
            topLabelDiv = LyString.replace(topLabelDiv, "\\\\n", "\\n");
            topLabelDiv = LyString.replace(topLabelDiv, "\"", "\\\"");
            topLabelDiv = LyString.replace(topLabelDiv, "\\'", "\\\\'");

            script.append("addDiv(\"");
            script.append(_lastFrameName).append("\",\"mapnode_").append(_mapNode.getIdCode()).append("_global\",\"");
            script.append(topLabelDiv).append("\",\"mapnode_").append(_mapNode.getIdCode()).append("_topLabel\");\n");
        }
        
        script.append("replaceDivInnerHtml(\"");
        script.append(_lastFrameName).append("\",\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_topLabel\",\"");
        script.append(label);
        script.append("\");\n");
        script.append("SetDivSize(\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_topLabel\",").append(_topLabelWidth);
        script.append(",").append(_topLabelHeight).append(");\n");
        script.append("moveDiv(\"");
        script.append(_lastFrameName).append("\",\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_topLabel\",").append(x);
        script.append(",").append(y).append(");\n//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(script.toString());
        checkLabels();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalMapNode#setBottomLabel(String)
     */
    public void setBottomLabel(String label)
    {
        // Get rid of cache and reload label
        _bottomLabelCache = null;
        _bottomLabelWidth = -1;
        _bottomLabelHeight = -1;
        _bottomLabelOffset = null;
        label = getBottomLabel();

        if (_map == null)
        {
            refresh();
            return;
        }
        
        Point   point = _map.convert(_center);
        int     x = point.x, y = point.y;

        Point pos = getBottomLabelOffset(_map, _map.getGraphics(), true);

        x = pos.x;
        y = pos.y;

        LyWebViewManager    viewManager = getWebViewManager();
        StringBuffer        script = new StringBuffer();
        int                 layer = _mapNode.getLayer() + 2;
        String              bottomLabelDiv = getLabelDivContent(_map, "bottom", layer, x, y);

        script.append("<script type='text/javascript'>\n//<![CDATA[\n");
        
        if (bottomLabelDiv != null)
        {
            LyComponent         parent = _mapNode.getParent();

            if (parent instanceof LyMapContent)
                parent = parent.getParent();

            bottomLabelDiv = LyString.replace(bottomLabelDiv, "\\", "\\\\");
            bottomLabelDiv = LyString.replace(bottomLabelDiv, "\\\\n", "\\n");
            bottomLabelDiv = LyString.replace(bottomLabelDiv, "\"", "\\\"");
            bottomLabelDiv = LyString.replace(bottomLabelDiv, "\\'", "\\\\'");

            script.append("addDiv(\"");
            script.append(_lastFrameName).append("\",\"mapnode_").append(_mapNode.getIdCode()).append("_global\",\"");
            script.append(bottomLabelDiv).append("\",\"mapnode_").append(_mapNode.getIdCode()).append("_bottomLabel\");\n");
        }
        
        script.append("replaceDivInnerHtml(\"");
        script.append(_lastFrameName).append("\",\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_bottomLabel\",\"");
        script.append(label);
        script.append("\");\n");
        script.append("SetDivSize(\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_bottomLabel\",").append(_bottomLabelWidth);
        script.append(",").append(_bottomLabelHeight).append(");\n");
        script.append("moveDiv(\"");
        script.append(_lastFrameName).append("\",\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_bottomLabel\",").append(x);
        script.append(",").append(y).append(");\n//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(script.toString());
        checkLabels();
    }
    
    /**
     * Checks the consistency of the label values (label, leftLabel, rightLabel, topLabel
     * and bottomLabel). For example, if both label with a left alignement and leftLabel are set,
     * a warning will be displayed to indicate that labels may overlap.
     */
    protected void checkLabels()
    {
        String  warning = null;
        
        if ((_mapNode.getLabel() != null)
                && (_mapNode.getLabelAlign() == LyMapNode.LabelAlign.LEFT)
                && (_mapNode.getLeftLabel() != null))
            warning = "Both left label and label with left align are set on map node " + _mapNode.getId();
        else if ((_mapNode.getLabel() != null)
                && (_mapNode.getLabelAlign() == LyMapNode.LabelAlign.RIGHT)
                && (_mapNode.getRightLabel() != null))
            warning = "Both right label and label with right align are set on map node " + _mapNode.getId();
        else if ((_mapNode.getLabel() != null)
                && (_mapNode.getLabelAlign() == LyMapNode.LabelAlign.TOP)
                && (_mapNode.getTopLabel() != null))
            warning = "Both top label and label with top align are set on map node " + _mapNode.getId();
        else if ((_mapNode.getLabel() != null)
                && (_mapNode.getLabelAlign() == LyMapNode.LabelAlign.BOTTOM)
                && (_mapNode.getBottomLabel() != null))
            warning = "Both bottom label and label with bottom align are set on map node " + _mapNode.getId();
        
        if (warning != null)
        {
            LyEnvironment   environment = getEnvironment();
            
            environment.logWarning("[LyStrutsMapNode] " + warning + ". Label will not be displayed correctly.");
        }
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetImageInterface#setImage(java.lang.String)
     */
    public void setImage(String image)
    {
        // Get rid of cache and reload image
        _image = null;
        _imageString = image;
        _imageHeight = -1;
        _imageWidth = -1;
        getImageSize();
        
        if (_map == null)
        {
            refresh();
            return;
        }

        // Change the image
        LyWebViewManager viewManager = getWebViewManager();
        StringBuffer     script = new StringBuffer();
        LyWebPageWriter  pw = new LyStrutsPage(_mapNode, true);

        writeHtml(pw);

        StringBuffer innerHtml = pw.getBuffer(pw.getActiveName());
        String       innerHtmlStr = LyString.replace(innerHtml.toString(), "\n", "\\n");

        innerHtmlStr = LyString.replace(innerHtmlStr, "\\", "\\\\");
        innerHtmlStr = LyString.replace(innerHtmlStr, "\\\\n", "\\n");
        innerHtmlStr = LyString.replace(innerHtmlStr, "\"", "\\\"");
        innerHtmlStr = LyString.replace(innerHtmlStr, "\\'", "\\\\'");

        script.append("<script type='text/javascript'>\n//<![CDATA[\nreplaceDivInnerHtml(\"");
        script.append(_lastFrameName).append("\",\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_node\",\"");
        script.append(innerHtmlStr);
        script.append("\");\n//]]>\n</script>");

        // Resize and move div tag
        Dimension dim = getImageSize(_map, true);
        float     zoom = _map.getZoom();

        if (_mapNode.isResizable())
            dim = new Dimension((int)(dim.width * zoom), (int)(dim.height * zoom));

        if (_mapNode.isSensitive())
            dim = new Dimension(dim.width + 8, dim.height + 8);

        int   maxWidth = dim.width;
        int   maxHeight = dim.height;
        
        script.append("<script type='text/javascript'>\n//<![CDATA[\nmoveDiv(\"");
        script.append(_lastFrameName).append("\",\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_node\",").append((-maxWidth / 2));
        script.append(",").append((-maxHeight / 2)).append(");\n//]]>\n</script>");
        script.append("<script type='text/javascript'>\n//<![CDATA[\nresizeDiv(\"");
        script.append(_lastFrameName).append("\",\"mapnode_");
        script.append(_mapNode.getIdCode()).append("_node\",").append(maxWidth);
        script.append(",").append(maxHeight).append(");\n//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(script.toString());
        pw.free();
        refreshViewerImage();
    } // end method setImage

    /**
     * Refresh the viewer image of the parent map.
     */
    protected void refreshViewerImage()
    {
        LyComponent     parent = _mapNode.getParent();
        
        while ((parent != null) && !(parent instanceof LyMap))
        {
            parent = parent.getParent();
        }
        
        if (parent instanceof LyMap)
        {
            LyMap       map = (LyMap)parent;
            LyStrutsMap  webMap = (LyStrutsMap)map.getInternalView();
            
            webMap.refreshViewerImage();
        }
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalMapNode#setColorFrameInt(java.lang.String)
     */
    public void setColorFrameInt(String colorFrameInt)
    {
        if (_map == null)
        {
            refresh();
            return;
        }
        
        // Change div color and visibility
        LyWebViewManager    viewManager = getWebViewManager();
        StringBuffer        script = new StringBuffer();

        if (colorFrameInt == null)
        {
            script.append("<script type='text/javascript'>\n//<![CDATA[\nSetDivVisibility('").
                append(getId()).append("_internal_frame', false);\n//]]>\n</script>");
        }
        else
        {
            script.append("<script type='text/javascript'>\n//<![CDATA[\nSetObjectColor('").
                append(_lastFrameName).append("', '").append(getId()).append("_internal_frame', '").
                append(getHexColor(colorFrameInt)).append("');\n//]]>\n</script>");
            script.append("<script type='text/javascript'>\n//<![CDATA[\nSetDivVisibility('").
            append(getId()).append("_internal_frame', true);\n//]]>\n</script>");
        }
        
        String              innerHtmlStr = LyString.replace(script.toString(), "\\", "\\\\");

        viewManager.addAdditionalRefreshScripts(innerHtmlStr.toString());
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalMapNode#setColorFrameExt(java.lang.String)
     */
    public void setColorFrameExt(String colorFrameExt)
    {
        if (_map == null)
        {
            refresh();
            return;
        }
        
        // Change div color and visibility
        LyWebViewManager    viewManager = getWebViewManager();
        StringBuffer        script = new StringBuffer();

        if (colorFrameExt == null)
        {
            script.append("<script type='text/javascript'>\n//<![CDATA[\nSetDivVisibility('").
                append(getId()).append("_external_frame', false);\n//]]>\n</script>");
        }
        else
        {
            script.append("<script type='text/javascript'>\n//<![CDATA[\nSetObjectColor('").
                append(_lastFrameName).append("', '").append(getId()).append("_external_frame', '").
                append(getHexColor(colorFrameExt)).append("');\n//]]>\n</script>");
            script.append("<script type='text/javascript'>\n//<![CDATA[\nSetDivVisibility('").
            append(getId()).append("_external_frame', true);\n//]]>\n</script>");
        }
        
        String              innerHtmlStr = LyString.replace(script.toString(), "\\", "\\\\");

        viewManager.addAdditionalRefreshScripts(innerHtmlStr.toString());
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalMapNode#setFontColor(java.lang.String)
     */
    public void setFontColor(String labelColor)
    {
        _mapNode.setFontColor(labelColor);
    }

    public void setViewerImage(String viewerImage)
    {
    }
} // end class LyStrutsMapNode
