/*
 * LyStrutsMapRegion.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web.struts;

import java.awt.*;
import java.util.*;

import leon.misc.*;
import leon.view.*;
import leon.view.common.*;
import leon.view.graphic.*;
import leon.view.web.*;
import leon.view.web.struts.map.*;


/**
 * This class implements the Map Region object (as described in the dtd) for Struts viewer.
 *
 * <p><b>MAP_REGION: Map Region</b><br>
 * Regions allow representing geometric areas, who can be sensitive, defined by a discrete set of
 * points. These points define polygons that are displayed on the map with a given color. It is
 * notably possible to lead to other map views when launching an action over one of these regions.
 * <br>
 * Regions follow zooms made on the map (their size increases or decreases when zooming in or out).
 * </p>
 *
 * @version $Id: LyStrutsMapRegion.java 36510 2012-08-17 15:43:06Z jlsaouli $
 * @author  Lyria
 * @see     leon.view.LyMap
 */
public class LyStrutsMapRegion extends LyStrutsMapObject
    implements LyGraphicalMapRegion,
               LyStrutsMapElement
{
	public static final String    __VERSION =
		"$Id: LyStrutsMapRegion.java 36510 2012-08-17 15:43:06Z jlsaouli $";

	/** Indicates id the HTML tag DIV has to be used for treemap layout. */
    public static final boolean USE_DIV_FOR_TREEMAP_LAYOUT = true;

    /** Offset for region shadow. */
    public static final int     SHADOW_OFFSET = 3;

    /**
     * List of 2D polygons representing the region (Polygon) in the map. Used to determine if a
     * click happened in the region.
     */
    protected Vector<Polygon>   _drawnPolygons;

    /** The region generic component. */
    protected LyMapRegion       _region;

    /** Place dedicated to the frame used for the outside label. */
    protected Rectangle         _labelFrame;

    /** Bounds of the small image. */
    protected Rectangle         _smallImageBounds;

    /** Font of the region. */
    protected Font              _graphicalFont;

    /** Color of the font of the region. */
    protected Color             _fontColor;
    
    /** Indicates whether this map region is displayed in a treemap. */
    protected boolean           _isTreemapLayout;
    
	/**
	 * Class constructor of a Map region object for the Struts target.
	 *
	 * @param region the generic Map region defined for every type of view targets.
	 */
	public LyStrutsMapRegion(LyMapRegion region)
	{
        super(region);
        _labelFrame = null;
        _smallImageBounds = null;
        _region = region;
        _isTreemapLayout = false;

        setContent();
	}

	/**
	 * Gets the Javascript code used when a context menu is performed on the region.
	 *
	 * @param pw     the page writer
	 * @param map    parent map of the region
	 * @param extend indicates whether extended selection is allowed
	 * @return the String describing the Javascript code when a context menu is performed on the region.
	 */
	protected String getContextMenuScript(LyWebPageWriter pw, LyMap map, boolean extend)
	{
		// get the url to the selection handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		String			    selectionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_SELECTEVENT);
		String				regionId = _region.getId();
		String				javascriptRegionId = LyString.getJavascriptString(regionId);
		String				javascriptDivId = LyString.getJavascriptString(getId());
		StringBuffer		buffer = new StringBuffer();
		
		buffer.append("javascript:InvokeMapPopup('");
		buffer.append(javascriptRegionId);
		buffer.append("','");
		buffer.append(javascriptDivId);
		buffer.append("','");
		buffer.append(_region.getRootController());
		buffer.append("','map_");
		buffer.append(map.getIdCode());
		buffer.append("','");
		buffer.append(map.getId());
		buffer.append("','");
		buffer.append(actionUrl);
		buffer.append("','");
		buffer.append(selectionUrl);
		buffer.append("',event);return false;");
		
		return buffer.toString();
	}
	
	   /**
     * Sets the content of the map region.
     */
    public void setContent()
    {
        // Build the polygons of the region (if given in an extra file)
        _region.loadPolygonsFile(getEnvironment());
        _graphicalFont = getFont("LY_MAPNODE_FONT");
        _fontColor = getColor("LY_MAP_FONT_COLOR");
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
     */
    public void writeHtml(LyWebPageWriter pw)
    {
        // Nothing to do
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#getImages()
     */
    public Image[] getImages()
    {
        return null;
    }
    
    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#drawInViewer(leon.view.web.struts.map.LyStrutsMapCanvas,
     *      java.awt.Graphics, int)
     */
    public void drawInViewer(LyStrutsMapCanvas map, Graphics g)
    {
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#draw(leon.view.web.struts.map.LyStrutsMapCanvas,
     *      java.awt.Graphics, int)
     */
    public void draw(LyStrutsMapCanvas map, Graphics g, int mode)
    {
        LyStrutsLayoutManager layout = map.getLayoutManager();
        boolean              isTreemapLayout = (layout instanceof LyStrutsTreemapLayout);

        if (USE_DIV_FOR_TREEMAP_LAYOUT && isTreemapLayout)
            return;

        draw(map, g, mode, false, 0);
    }

    /**
     * Draws the map region.
     *
     * @param map       the map of this region
     * @param g         ths graphics context
     * @param mode      the map style
     * @param forceDraw indicates if the region must be drawn
     * @param level     te level of the region (its children will be at the level <code>
     *                  level+1</code> for example)
     */
    public void draw(LyStrutsMapCanvas map, Graphics g, int mode, boolean forceDraw, int level)
    {
        // If map region has a parent map region, parent will draw its sub
        // elements itself
        if (!forceDraw && (_region.getParent() instanceof LyMapRegion))
            return;

        LyStrutsLayoutManager layout = map.getLayoutManager();
        boolean              isTreemapLayout = (layout instanceof LyStrutsTreemapLayout);

        _smallImageBounds = null;

        // If region has no weight, do not display it
        if (isTreemapLayout && (_region.getWeight() <= 0))
            return;

        _drawnPolygons = new Vector<Polygon>();

        int       i, n = _region.getPolygonCount();
        LyPolygon polygon;

        synchronized (_region)
        {
            // If a map region has children, draw its children, otherwise draw
            // its polygon
            if (_region.getMapRegionCount() == 0)
            {
                for (i = 0; i < n; i++)
                {
                    polygon = _region.getPolygon(i);
                    drawPolygon(map, g, polygon, false, mode);
                }
            }
            else
            {
                LyMapRegion      region;
                LyStrutsMapRegion winRegion;

                n = _region.getMapRegionCount();

                for (i = 0; i < n; i++)
                {
                    region = _region.getMapRegion(i);
                    winRegion = (LyStrutsMapRegion)region.getInternalView();
                    winRegion.draw(map, g, mode, true, level + 1);
                }

                // Draw selection even if region has children
                n = _region.getPolygonCount();

                if (_region.isSelected())
                {
                    for (i = 0; i < n; i++)
                    {
                        polygon = _region.getPolygon(i);

                        drawPolygon(map, g, polygon, false, LyStrutsMapElement.DRAW_MODE_REDUCED);
                    }
                }
            }

            if (isTreemapLayout)
            {
                // For treemap layout draw the label of the region
                LyStrutsTreemapLayout treemapLayout = (LyStrutsTreemapLayout)layout;
                short                labelStyle = treemapLayout.getLabelStyle(level);
                Rectangle            bounds = getBoundingRectangle(map);
                String               label = getLabel();
                int                  smallImageWidth = 0;
                int                  smallImageHeight = 0;
                boolean              displaySmallImage = false;

                // Draw the small image
                Image                smallImage = getImage(_region.getSmallImage());
                int                  xImage = -1, yImage = -1;

                if (smallImage != null)
                {
                    smallImageWidth = smallImage.getWidth(map) + 1;
                    smallImageHeight = smallImage.getHeight(map);
                    displaySmallImage = ((bounds.height > (smallImageHeight + 1))
                                && (bounds.width > (smallImageWidth + 1)));
                }

                if ((label != null) && (label.length() > 0))
                {
                    FontMetrics fm = g.getFontMetrics(_graphicalFont);
                    int         xLabel = 0, yLabel = 0;
                    int         labelWidth = fm.stringWidth(label), l = label.length();
                    int         width = bounds.width;

                    if (labelStyle == LyStrutsTreemapLayout.LABEL_OUTSIDE)
                        bounds = _labelFrame;

                    if (_labelFrame != null)
                    {
                        Rectangle border = new Rectangle(_labelFrame.x, _labelFrame.y,
                                _labelFrame.width, _labelFrame.height).intersection(bounds);

                        g.setColor(getColor("LY_TREEMAP_OUTSIDE_BACKGROUND_COLOR"));
                        g.fillRect(border.x, border.y, border.width, border.height);
                        g.setColor(getColor("LY_TREEMAP_OUTSIDE_BORDER_COLOR"));
                        g.drawRect(border.x, border.y, border.width, border.height);
                    }

                    // Do not display label if not enough height
                    if ((bounds != null) && (bounds.height >= fm.getHeight()))
                    {
                        // Do not display entire label if not enough space
                        if ((labelWidth + smallImageWidth) > width)
                        {
                            while ((labelWidth + smallImageWidth) > width)
                            {
                                if (label.length() <= 1)
                                    break;

                                l--;

                                if (l == 0)
                                {
                                    label = "";
                                    break;
                                }

                                label = label.substring(0, l) + "...";
                                labelWidth = fm.stringWidth(label);
                            }
                        }

                        if (labelStyle == LyStrutsTreemapLayout.LABEL_CENTER)
                        {
                            xImage = bounds.x
                                + ((bounds.width - fm.stringWidth(label) - smallImageWidth) / 2);
                            yImage = bounds.y + ((bounds.height - smallImageHeight) / 2);
                            xLabel = xImage + smallImageWidth + 1;
                            yLabel = bounds.y + ((bounds.height - fm.getHeight()) / 2)
                                + fm.getAscent();
                        }
                        else if (labelStyle == LyStrutsTreemapLayout.LABEL_TOPLEFT)
                        {
                            xImage = bounds.x + 1;
                            xLabel = xImage + smallImageWidth + 1;

                            if (smallImageHeight > fm.getHeight())
                            {
                                yImage = bounds.y + 1;
                                yLabel = bounds.y + ((smallImageHeight + 1 - fm.getHeight()) / 2)
                                    + fm.getAscent();
                            }
                            else
                            {
                                yImage = bounds.y + ((fm.getHeight() - smallImageHeight) / 2);
                                yLabel = bounds.y + fm.getAscent();
                            }
                        }
                        else if (labelStyle == LyStrutsTreemapLayout.LABEL_OUTSIDE)
                        {
                            xImage = bounds.x + 1;
                            yImage = bounds.y + 1;
                            xLabel = xImage + ((bounds.width - fm.stringWidth(label)) / 2);
                            yLabel = bounds.y + bounds.height - fm.getHeight() + fm.getAscent();
                        }

                        if (labelStyle != LyStrutsTreemapLayout.LABEL_NONE)
                        {
                            g.setFont(_graphicalFont);
                            g.setColor(_fontColor);

                            if (_labelFrame != null)
                                g.setColor(getColor("LY_TREEMAP_OUTSIDE_LABEL_COLOR"));

                            g.drawString(label, xLabel, yLabel);
                        }
                    }
                }

                if (displaySmallImage)
                {
                    // If image position has not been computed (no label for
                    // example), put the image
                    // in the top left corner
                    if ((xImage < 0) && (yImage < 0))
                    {
                        xImage = bounds.x + 1;
                        yImage = bounds.y + 1;
                    }

                    _smallImageBounds = new Rectangle(xImage, yImage, smallImageWidth,
                            smallImageHeight);
                    g.drawImage(smallImage, xImage, yImage, map);
                }
            }
        }
    } // end method draw

    /**
     * Draws the given polygon of the region's given map.
     *
     * @param map      the region map
     * @param g        the cgraphics context
     * @param polygon  the polygon to draw
     * @param inViewer indicates if the region is in the viewer or not
     * @param mode     the drawing mode. Possible values are {@link
     *                 LyStrutsMapElement#DRAW_MODE_NORMAL} and {@link
     *                 LyStrutsMapElement#DRAW_MODE_REDUCED}.
     */
    protected void drawPolygon(LyStrutsMapCanvas map, Graphics g, LyPolygon polygon,
        boolean inViewer, int mode)
    {
        LyPoint point3D;
        Point   point, prevPoint = null;
        Polygon poly = new Polygon();
        int     j, nPoints = polygon.getPointCount();
        boolean isNormalMode = (mode == LyStrutsMapElement.DRAW_MODE_NORMAL);
        int     pitch = 1;

        if (!isNormalMode)
            pitch = Math.max((nPoints / 100), 1);

        for (j = 0; j < nPoints; j += pitch)
        {
            point3D = polygon.getPoint(j);
            point = map.convert(point3D);

            // Optimization : don't draw the point if same as previous
            if ((prevPoint == null) || (prevPoint.x != point.x) || (prevPoint.y != point.y))
                poly.addPoint(point.x, point.y);

            prevPoint = point;
        }

        // Draw a shadow of the region if background style and not in viewer
        if (isNormalMode)
        {
            if (!inViewer && _region.isBackground())
            {
                poly.translate(SHADOW_OFFSET, SHADOW_OFFSET);
                g.setColor(getColor("LY_MAP_OUTLINE_COLOR"));
                g.fillPolygon(poly);
                poly.translate(-SHADOW_OFFSET, -SHADOW_OFFSET);
            }

            // Retrieve the polygon color
            String color = polygon.getColor();
            Color  graphicalColor = null;

            if ((color != null) && (color.length() != 0))
                graphicalColor = getColor(color);
            else
                graphicalColor = getColor("LY_DEFAULT_MAP_REGION_COLOR");

            g.setColor(graphicalColor);
            g.fillPolygon(poly);
        }

        if (!inViewer)
            _drawnPolygons.addElement(poly);

        if ((!_region.isBackground() && !inViewer) || (!isNormalMode))
        {
            LyStrutsLayoutManager layout = map.getLayoutManager();
            boolean              isTreemapLayout = (layout instanceof LyStrutsTreemapLayout);

            if (_region.isSelected())
                g.setColor(getColor("LY_MAP_SELECTED_COLOR"));
            else if (!isTreemapLayout)
                g.setColor(getColor("LY_MAP_OUTLINE_COLOR"));

            g.drawPolygon(poly);

            if (isTreemapLayout && _region.isSelected())
            {
                // If select border size is greater than one, draw others
                // polygons
                if (((LyStrutsTreemapLayout)layout).getSelectBorderSize() > 1)
                {
                    int       borderWidth = ((LyStrutsTreemapLayout)layout).getSelectBorderSize();
                    Rectangle bounds = getBoundingRectangle(map);

                    borderWidth = Math.min(borderWidth, bounds.width);
                    borderWidth = Math.min(borderWidth, bounds.height);

                    for (j = 0; j < (borderWidth - 1); j++)
                    {
                        poly = new Polygon(
                                new int[]
                                {
                                    poly.xpoints[0] + 1,
                                    poly.xpoints[1] - 1,
                                    poly.xpoints[2] - 1,
                                    poly.xpoints[3] + 1
                                },
                                new int[]
                                {
                                    poly.ypoints[0] + 1,
                                    poly.ypoints[1] + 1,
                                    poly.ypoints[2] - 1,
                                    poly.ypoints[3] - 1
                                }, poly.npoints);
                        g.drawPolygon(poly);
                    }
                }
            }
        }
    } // end method drawPolygon

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#getBoundingBox(leon.view.web.struts.map.LyStrutsMapCanvas)
     */
    public Ly3DBox getBoundingBox(LyStrutsMapCanvas map)
    {
        Vector<LyPoint> points = getPoints();
        LyPoint         point3D;
        int             i, n = points.size();

        // If there is no point, return an empty bouding box
        if (n == 0)
            return new Ly3DBox(0, 0, 0, 0, 0, 0);

        float xmin = Float.MAX_VALUE;
        float xmax = -Float.MAX_VALUE;
        float ymin = Float.MAX_VALUE;
        float ymax = -Float.MAX_VALUE;
        float zmin = Float.MAX_VALUE;
        float zmax = -Float.MAX_VALUE;

        for (i = 0; i < n; i++)
        {
            point3D = points.elementAt(i);

            if (point3D.getX() > xmax)
                xmax = point3D.getX();
            if (point3D.getX() < xmin)
                xmin = point3D.getX();
            if (point3D.getY() > ymax)
                ymax = point3D.getY();
            if (point3D.getY() < ymin)
                ymin = point3D.getY();
            if (point3D.getZ() > zmax)
                zmax = point3D.getZ();
            if (point3D.getZ() < zmin)
                zmin = point3D.getZ();
        }

        return new Ly3DBox(xmin, ymin, zmin, xmax, ymax, zmax);
    } // end method getBoundingBox

    /**
     * Gets the polygon that contains this region.
     *
     * @param  map the region map
     * @return the polygon that contains this region
     */
    public Polygon getBoundingShape(LyStrutsMapCanvas map)
    {
        Vector<LyPoint> points = getPoints();
        LyPoint         point3D;
        Point           point;
        int             i, n = points.size();
        int             delta = ((n <= 40) ? 1 : (n / 40));
        Polygon         polygon = new Polygon();

        for (i = 0; i < n; i = i + delta)
        {
            point3D = points.elementAt(i);
            point = map.convert(point3D);
            polygon.addPoint(point.x, point.y);
        }

        return polygon;
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#getBoundingRectangle(leon.view.web.struts.map.LyStrutsMapCanvas)
     */
    public Rectangle getBoundingRectangle(LyStrutsMapCanvas map)
    {
        LyStrutsLayoutManager layout = map.getLayoutManager();
        boolean              isTreemapLayout = (layout instanceof LyStrutsTreemapLayout);

        Vector<LyPoint>      points = getPoints();
        LyPoint              point3D;
        Point                point;
        int                  i, n = points.size();
        int                  xmin = Integer.MAX_VALUE;
        int                  xmax = Integer.MIN_VALUE;
        int                  ymin = Integer.MAX_VALUE;
        int                  ymax = Integer.MIN_VALUE;
        int                  x, y;

        for (i = 0; i < n; i++)
        {
            point3D = points.elementAt(i);

            if (!isTreemapLayout)
            {
                point = map.convert(point3D);
                x = point.x;
                y = point.y;
            }
            else
            {
                x = (int)point3D.getX();
                y = (int)point3D.getY();
            }

            if (x > xmax)
                xmax = x;
            if (x < xmin)
                xmin = x;
            if (y > ymax)
                ymax = y;
            if (y < ymin)
                ymin = y;
        }

        return new Rectangle(xmin, ymin, (xmax - xmin), (ymax - ymin));
    } // end method getBoundingRectangle

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#getClippingArea(leon.view.web.struts.map.LyStrutsMapCanvas,
     *      java.awt.Image)
     */
    public Shape getClippingArea(LyStrutsMapCanvas map, Image image)
    {
        return getBoundingRectangle(map);
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#getLayer()
     */
    public int getLayer()
    {
        if (_region.getLayer() > 0)
            return _region.getLayer();
        
        if (_region.isBackground())
            return LyStrutsMapCanvas.REGION_BACKGROUND_LAYER;
        else
            return LyStrutsMapCanvas.REGION_LAYER;
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#getLabel()
     */
    public String getLabel()
    {
        LyEnvironment env = getEnvironment();
        String        label = _region.getLabel();

        return env.translate(label);
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#drawHtml(leon.view.web.struts.map.LyStrutsMapCanvas,
     *      leon.view.web.LyWebPageWriter)
     */
    public void drawHtml(LyStrutsMapCanvas map, LyWebPageWriter pw)
    {
        drawHtml(map, pw, 0);
    }

    /**
     * Draws the HTML tag for this region.
     *
     * @param map   the ragion map
     * @param pw    the page writer
     * @param level the level of the region (for example, its children are at the level <code>
     *              level</code>+1
     */
    public void drawHtml(LyStrutsMapCanvas map, LyWebPageWriter pw, int level)
    {
        updateRefreshInfo(pw);

        LyStrutsLayoutManager layout = map.getLayoutManager();
        boolean              isTreemapLayout = (layout instanceof LyStrutsTreemapLayout);

        if (!isTreemapLayout)
            return;

        // If region has no weight, do not display it
        if (isTreemapLayout && (_region.getWeight() <= 0))
            return;

        if (USE_DIV_FOR_TREEMAP_LAYOUT)
        {
            _isTreemapLayout = true;
            
            int                  i, n = _region.getPolygonCount();
            LyPolygon            polygon;
            Rectangle            bounds = getBoundingRectangle(map);
            LyStrutsTreemapLayout treemapLayout = (LyStrutsTreemapLayout)layout;

            synchronized (_region)
            {
                // If a map region has children, draw its children, otherwise
                // draw its polygon
                if (_region.getMapRegionCount() == 0)
                {
                    for (i = 0; i < n; i++)
                    {
                        polygon = _region.getPolygon(i);
                        drawHtmlPolygon(map, pw, polygon, _region.isSelected());
                    }
                }
                else
                {
                    LyMapRegion      region;
                    LyStrutsMapRegion winRegion;

                    n = _region.getMapRegionCount();

                    for (i = 0; i < n; i++)
                    {
                        region = _region.getMapRegion(i);
                        winRegion = (LyStrutsMapRegion)region.getInternalView();
                        winRegion.drawHtml(map, pw, level + 1);
                    }

                    if (_region.isSelected())
                    {
                        int    borderWidth = treemapLayout.getSelectBorderSize();
                        String selectedColor = getHexColor("LY_MAP_SELECTED_COLOR");

                        // Draw selection even if region has children
                        pw.append(
                            "<div id='_parentRegionTopSelection' style='position:absolute;z-index:2;left:"
                            + bounds.x
                            + ";top:"
                            + bounds.y
                            + ";width:"
                            + (bounds.width + 1)
                            + ";height:"
                            + borderWidth
                            + ";background-color:"
                            + selectedColor
                            + "'/><div></div></div>\n");
                        pw.append(
                            "<div id='_parentRegionRightSelection' style='position:absolute;z-index:2;left:"
                            + (bounds.x + bounds.width - borderWidth + 1)
                            + ";top:"
                            + bounds.y
                            + ";width:"
                            + borderWidth
                            + ";height:"
                            + (bounds.height + 1)
                            + ";background-color:"
                            + selectedColor + "'/><div></div></div>\n");
                        pw.append(
                            "<div id='_parentRegionBottomSelection' style='position:absolute;z-index:2;left:"
                            + bounds.x
                            + ";top:"
                            + (bounds.y + bounds.height - 1)
                            + ";width:"
                            + (bounds.width + 1)
                            + ";height:"
                            + borderWidth
                            + ";background-color:"
                            + selectedColor
                            + "'/><div></div></div>\n");
                        pw.append(
                            "<div id='_parentRegionLeftSelection' style='position:absolute;z-index:2;left:"
                            + bounds.x
                            + ";top:"
                            + bounds.y
                            + ";width:"
                            + borderWidth
                            + ";height:"
                            + (bounds.height + 1)
                            + ";background-color:"
                            + selectedColor + "'/><div></div></div>\n");
                    }
                }

                boolean displaySmallImage = false;
                int     smallImageWidth = 0;
                int     smallImageHeight = 0;
                Image   smallImage = getImage(_region.getSmallImage());
                int     xImage = -1, yImage = -1;

                // For treemap layout draw the label of the region
                short   labelStyle = treemapLayout.getLabelStyle(level);
                String  label = getLabel();

                if (smallImage != null)
                {
                    smallImageWidth = smallImage.getWidth(map) + 1;
                    smallImageHeight = smallImage.getHeight(map);
                    displaySmallImage = ((bounds.height > (smallImageHeight + 1))
                                && (bounds.width > (smallImageWidth + 1)));
                }

                if ((label != null) && (label.length() > 0))
                {
                    FontMetrics fm = map.getFontMetrics(_graphicalFont);
                    int         xLabel = 0, yLabel = 0;
                    int         labelWidth = fm.stringWidth(label), l = label.length();
                    int         width = bounds.width;

                    if (labelStyle == LyStrutsTreemapLayout.LABEL_OUTSIDE)
                        bounds = _labelFrame;

                    if (_labelFrame != null)
                    {
                        Rectangle border = new Rectangle(_labelFrame.x, _labelFrame.y,
                                _labelFrame.width, _labelFrame.height).intersection(bounds);

                        pw.append("<div style='position:absolute;z-index:0;left:" + border.x
                            + ";top:" + border.y + ";width:" + (border.width + 1) + ";height:"
                            + (border.height + 1) + ";background-color:"
                            + getHexColor("LY_TREEMAP_OUTSIDE_BORDER_COLOR")
                            + "'/><div style='visibility:hidden'></div></div>");
                        pw.append("<div style='position:absolute;z-index:0;left:" + (border.x + 1)
                            + ";top:" + (border.y + 1) + ";width:" + (border.width - 1)
                            + ";height:" + (border.height - 1) + ";background-color:"
                            + getHexColor("LY_TREEMAP_OUTSIDE_BACKGROUND_COLOR")
                            + "'/><div style='visibility:hidden'></div></div>");
                    }

                    // Do not display label if not enough height
                    if ((bounds != null) && (bounds.height >= fm.getHeight()))
                    {
                        // Do not display entire label if not enough space
                        if ((labelWidth + smallImageWidth) > width)
                        {
                            while ((labelWidth + smallImageWidth) > width)
                            {
                                if (label.length() <= 1)
                                    break;

                                l--;

                                if (l == 0)
                                {
                                    label = "";
                                    break;
                                }

                                label = label.substring(0, l) + "...";
                                labelWidth = fm.stringWidth(label);
                            }
                        }

                        if (labelStyle == LyStrutsTreemapLayout.LABEL_CENTER)
                        {
                            xImage = bounds.x
                                + ((bounds.width - fm.stringWidth(label) - smallImageWidth) / 2);
                            yImage = bounds.y + ((bounds.height - smallImageHeight) / 2);
                            xLabel = xImage + smallImageWidth + 1;
                            yLabel = bounds.y + ((bounds.height - fm.getHeight()) / 2);
                        }
                        else if (labelStyle == LyStrutsTreemapLayout.LABEL_TOPLEFT)
                        {
                            xImage = bounds.x + 1;
                            xLabel = xImage + smallImageWidth + 1;

                            if (smallImageHeight > fm.getHeight())
                            {
                                yImage = bounds.y + 1;
                                yLabel = bounds.y + ((smallImageHeight + 1 - fm.getHeight()) / 2);
                            }
                            else
                            {
                                yImage = bounds.y + ((fm.getHeight() - smallImageHeight) / 2);
                                yLabel = bounds.y;
                            }
                        }
                        else if (labelStyle == LyStrutsTreemapLayout.LABEL_OUTSIDE)
                        {
                            xImage = bounds.x + 1;
                            yImage = bounds.y + bounds.height + 2;
                            xLabel = xImage + ((bounds.width - fm.stringWidth(label)) / 2);
                            yLabel = bounds.y;
                        }

                        if (labelStyle != LyStrutsTreemapLayout.LABEL_NONE)
                        {
                            String color = "LY_MAP_FONT_COLOR";
                            String classe = "regionLabel";

                            if (_labelFrame != null)
                            {
                                color = "LY_TREEMAP_OUTSIDE_LABEL_COLOR";
                                classe = "regionOutsideLabel";
                            }

                            String fontColor = "color:" + getHexColor(color) + ";";

                            pw.addCssScript("td." + classe + "{" + fontColor
                                + getFontString("LY_MAPNODE_FONT") + "}", false);

                            pw.append("<div style='position:absolute;z-index:6;left:" + xLabel
                                + ";top:" + yLabel + "'>");
                            pw.append(
                                "<table cellspacing='0' cellpadding='0'><tr><td nowrap='yes' class='"
                                + classe + "'>" + label + "</td></tr></table>");
                            pw.append("</div>");
                        }
                    }
                }

                if (displaySmallImage)
                {
                    // If image position has not been computed (no label for
                    // example), put the image
                    // in the top left corner
                    if ((xImage < 0) && (yImage < 0))
                    {
                        xImage = bounds.x + 1;
                        yImage = bounds.y + 1;
                    }

                    _smallImageBounds = new Rectangle(xImage, yImage, smallImageWidth,
                            smallImageHeight);
                    bounds = getBoundingRectangle(map);

                    // Draw the small image
                    String urlSmallImage = getImageUrl(_region.getSmallImage());

                    if (smallImage != null)
                    {
                        pw.append("<div style='position:absolute;z-index:3;left:" + xImage
                            + ";top:" + yImage + "'" + " onmouseover=\"mouseOverRegion('"
                            + LyString.getJavascriptString(getId()) + "',true,true," + bounds.x
                            + "," + bounds.y + "," + (bounds.width + 1) + ","
                            + (bounds.height + 1) + ")\" onmouseout=\"mouseOverRegion('"
                            + LyString.getJavascriptString(getId()) + "',false,true)\">");
                        pw.append("<img src='" + urlSmallImage + "'></img>");
                        pw.append("</div>");
                    }
                }
                else
                    _smallImageBounds = null;
            }
        }

        drawHtmlTooltip(map, pw, false);

        if (_smallImageBounds != null)
            drawHtmlTooltip(map, pw, true);
    } // end method drawHtml

    /**
     * Draws the HTML tag for a polygon on the map canvas.
     *
     * @param mapCanvas the map canvas
     * @param pw        the page writer
     * @param polygon   the polygon to draw
     * @param selected  indicates whether polygon is selected or not
     */
    public void drawHtmlPolygon(LyStrutsMapCanvas mapCanvas, LyWebPageWriter pw, LyPolygon polygon,
        boolean selected)
    {
        LyStrutsLayoutManager layout = mapCanvas.getLayoutManager();
        String               graphicalColor = "";
        String               color = polygon.getColor();
        String               selectedColor = getHexColor("LY_MAP_SELECTED_COLOR");

        // Call getInternalView() method to be sure that web polygon will be created
        // and will refresh itself.
        polygon.getInternalView();

        if ((color != null) && (color.length() != 0))
            graphicalColor = getHexColor(color);
        else
            graphicalColor = getHexColor("LY_DEFAULT_MAP_REGION_COLOR");

        StringBuffer url = new StringBuffer();

        if (mapCanvas.getSelectionMode() == LyMap.SelectionMode.MULTIPLE)
        {
            if (selected)
            {
                pw.append("<input type='hidden' name='" + LyWebConstants.SELECTION_ITEM + getId()
                    + "' value='" + LyWebConstants.SELECTION_ITEM + getId() + "' />");

                url = new StringBuffer();
                url.append("\"javascript:InvokeSelection('" + LyWebConstants.SELECT_ITEM + "'");
                url.append(",'"
                    + LyString.getJavascriptString(LyWebConstants.SELECTION_ITEM + getId())
                    + "'");
                url.append(",'" + LyWebConstants.RELOAD_FRAMES + "'");
                url.append(",false");
                url.append(");\"");
            }
            else
            {
                url = new StringBuffer();
                url.append("\"javascript:InvokeSelection('" + LyWebConstants.SELECT_ITEM + "'");
                url.append(",'"
                    + LyString.getJavascriptString(LyWebConstants.SELECTION_ITEM + getId())
                    + "'");
                url.append(",'" + LyWebConstants.RELOAD_FRAMES + "'");
                url.append(",false");
                url.append(");\"");
            }
        }
        else
        {
            url = new StringBuffer();
            url.append("\"javascript:TreemapSelection('" + LyString.getJavascriptString(getId())
                + "');\"");
        }

        Rectangle bounds = getBoundingRectangle(mapCanvas);
        Rectangle mouseOverBounds = new Rectangle(bounds);

        if ((bounds.width <= 0) || (bounds.height <= 0))
            return;

        if (selected && (getWebViewManager().getBrowser() == LyWebViewManager.BROWSER_FIREFOX))
        {
            bounds.width -= ((LyStrutsTreemapLayout)layout).getSelectBorderSize() * 2;
            bounds.height -= ((LyStrutsTreemapLayout)layout).getSelectBorderSize() * 2;
        }

        pw.append("<div id='" + getId() + "' style='background-color:" + graphicalColor
            + ";position:absolute;z-index:1;left:");
        pw.append(bounds.x + "px;top:" + bounds.y + "px;width:" + (bounds.width + 1) + ";height:"
            + (bounds.height + 1));

        pw.append(";border-width:"
            + (selected ? ((LyStrutsTreemapLayout)layout).getSelectBorderSize() : 0)
            + "px;border-style:solid;border-color:" + selectedColor);

        pw.append("' onMouseOver=\"mouseOverRegion('" + LyString.getJavascriptString(getId())
            + "',true,false," + mouseOverBounds.x + "," + mouseOverBounds.y + "," + (mouseOverBounds.width + 1) + ","
            + (mouseOverBounds.height + 1) + ")\" ");
        pw.append(" onMouseOut=\"mouseOverRegion('" + LyString.getJavascriptString(getId())
            + "',false,false)\" onClick=" + url.toString() + "");
        
        boolean extend = true;
        LyComponent parent = _region.getParent();

        while (parent != null && !(parent instanceof LyMap))
            parent = parent.getParent();

        if (parent instanceof LyMap)
        {
            LyMap  parentMap = (LyMap)parent;

            if (parentMap.getSelectionMode() != LyMap.SelectionMode.NONE)
            {
                if (parentMap.getSelectionMode() == LyMap.SelectionMode.SINGLE)
                    extend = false;
            }
            
            StringBuffer onContextMenu = new StringBuffer();

            onContextMenu.append(" onContextMenu=\"");
            onContextMenu.append(getContextMenuScript(pw, parentMap, extend));
            onContextMenu.append("\"");

            pw.append(onContextMenu.toString());
        }
        
        // These strange divs are needed to correctly display div that have a
        // small height
        pw.append("/><div></div></div>\n");
    } // end method drawHtmlPolygon
    
    /**
     * Draws the HTML tag for the tooltip of the current map region or its small image.
     *
     * @param mapCanvas  the map canvas
     * @param pw         the web page writer
     * @param smallImage indicates whether tooltip is needed for the small image or not
     */
    public void drawHtmlTooltip(LyStrutsMapCanvas mapCanvas, LyWebPageWriter pw, boolean smallImage)
    {
        LyEnvironment env = getEnvironment();
        String        tooltipColor = getHexColor("LY_TOOLTIP_COLOR");
        Font          font = getFont("LY_TOOL_TIP_FONT");
        FontMetrics   fm = mapCanvas.getFontMetrics(font);
        Rectangle     bounds = getBoundingRectangle(mapCanvas);
        Rectangle     mapBounds = mapCanvas.getBounds();
        int           gap = 5;
        int           x = 0, y = 0, compWidth, width = 0, height = gap;
        String        tooltip = null;

        if (smallImage)
        {
            x = _smallImageBounds.x + (_smallImageBounds.width / 2);
            y = _smallImageBounds.y + (_smallImageBounds.height / 2);
            compWidth = _smallImageBounds.width;
            tooltip = _region.getSmallImageTooltip();
        }
        else
        {
            x = bounds.x + (bounds.width / 2);
            y = bounds.y + (bounds.height / 2);
            compWidth = bounds.width;
            tooltip = _region.getTooltip();
        }

        if ((tooltip != null) && (tooltip.length() > 0))
        {
            tooltip = env.translate(tooltip);

            StringTokenizer t = new StringTokenizer(tooltip, "\n");
            String          line;
            Vector<String>  lines = new Vector<String>();
            int             i, n = 0;

            while (t.hasMoreElements())
            {
                line = t.nextToken();
                lines.addElement(line);
                width = Math.max(width, fm.stringWidth(line));
                height += font.getSize() + 2;
                n++;
            }

            width += (2 * gap) + 2;
            height += 2 * gap;

            if (x > (mapBounds.width / 2))
                // Draw on the left
                x = Math.max(gap, x - (compWidth / 2) - width - gap);
            else
                // Draw on the right
                x = Math.min(mapBounds.width - width - gap, x + (compWidth / 2) + gap);

            y = y - (height / 2);
            y = Math.max(gap, y);
            y = Math.min(mapBounds.height - height - gap, y);

            // Check if tooltip is over region
            if (bounds.intersects(x, y, width, height))
            {
                int upMargin = bounds.y;
                int bottomMargin = mapBounds.height - (bounds.y + bounds.height);

                if (upMargin >= bottomMargin)
                {
                    y = bounds.y - gap - height;
                    y = Math.max(gap, y);
                }
                else
                {
                    y = bounds.y + bounds.height + gap;
                    y = Math.min(mapBounds.height - height - gap, y);
                }
            }

            if (smallImage)
                pw.append("<div id='" + _region.getId()
                    + "_smallimage_tooltip' style='visibility:hidden;position:absolute;left:"
                    + x + "px;top:" + y + "px;z-index:10'>\n");
            else
                pw.append("<div id='" + _region.getId()
                    + "_tooltip' style='visibility:hidden;position:absolute;left:" + x
                    + "px;top:" + y + "px;z-index:10'>\n");

            pw.append("<table><tr><td bgColor='#000000'><table border='0' bgColor='" + tooltipColor
                + "'><tr><td class='tooltip' nowrap='yes'>");

            for (i = 0; i < n; i++)
            {
                line = lines.elementAt(i);
                if (i > 0)
                    pw.append("<br/>");

                pw.append(line);
            }

            pw.append("</td></tr></table></td></tr></table></div>\n");
        }
    } // end method drawHtmlTooltip

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#getTooltip()
     */
    public String getTooltip()
    {
        return null;
    }

    /**
     * Gets the tooltip for small image.
     *
     * @return always null
     * @see    #setSmallImageTooltip(String)
     */
    public String getSmallImageTooltip()
    {
        return null;
    }

    /**
     * Indicates if the component is drawable or not in web.
     *
     * @return if the map is a Treemap the component is Drawable so return true in other case return
     *         false
     */
    public boolean isDrawableInterface(LyStrutsMapCanvas map)
    {
        LyStrutsLayoutManager layout = map.getLayoutManager();
        boolean              isTreemapLayout = (layout instanceof LyStrutsTreemapLayout);

        return isTreemapLayout;
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#isAtPosition(leon.view.web.struts.map.LyStrutsMapCanvas,
     *      java.awt.Graphics, int, int)
     */
    public boolean isAtPosition(LyStrutsMapCanvas map, Graphics g, int x, int y)
    {
        if (_drawnPolygons == null)
            return false;

        int     i, n = _drawnPolygons.size();
        Polygon polygon;

        for (i = 0; i < n; i++)
        {
            polygon = _drawnPolygons.elementAt(i);
            if (polygon.contains(x, y))
                return true;
        }

        return false;
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.getinterface.LyGetIdInterface#getId()
     */
    public String getId()
    {
        return _region.getId();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetSelectedInterface#setSelected(boolean)
     */
    public void setSelected(boolean selected)
    {
        if (!_region.isBackground())
        {
            _region.setSelected(selected);
            
            if (_isTreemapLayout)
            {
                LyWebViewManager    viewManager = getWebViewManager();
                StringBuffer        buffer = new StringBuffer();

                buffer.append("<script type='text/javascript'>\n//<![CDATA[\n");
                buffer.append("var mapFrame = findTarget(null, '" + _lastFrameName + "');\n");
                buffer.append("if (mapFrame != null) {\n");
                buffer.append("mapFrame.TreemapSelection('").
                    append(getId()).append("', '" + selected + "');\n}\n//]]>\n</script>");

                String              innerHtmlStr = LyString.replace(buffer.toString(), "\\", "\\\\");

                viewManager.addAdditionalRefreshScripts(innerHtmlStr.toString());
            }
        }
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalMapRegion#setSmallImageTooltip(java.lang.String)
     */
    public void setSmallImageTooltip(String tooltip)
    {
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalMapRegion#setWeight(float)
     */
    public void setWeight(float weight)
    {
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalMapRegion#setSmallImage(java.lang.String)
     */
    public void setSmallImage(String smallImage)
    {
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#isSelected()
     */
    public boolean isSelected()
    {
        return _region.isSelected();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#translate(float, float, float)
     */
    public void translate(float dx, float dy, float dz)
    {
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#isOverlay()
     */
    public boolean isOverlay()
    {
        return isSelected();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#isHidden()
     */
    public boolean isHidden()
    {
        return _region.isHidden();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetHiddenInterface#setHidden(boolean)
     */
    public void setHidden(boolean hidden)
    {
        if (_region.isHidden() != hidden)
            _region.setHidden(hidden);
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.map.LyStrutsMapElement#isSensitive()
     */
    public boolean isSensitive()
    {
        return (_region.isSensitive() && !_region.isBackground());
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalMapRegion#addMapRegion(leon.view.LyMapRegion, int)
     */
    public void addMapRegion(LyMapRegion mapRegion, int pos)
    {
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalMapRegion#removeMapRegion(leon.view.LyMapRegion)
     */
    public void removeMapRegion(LyMapRegion mapRegion)
    {
    }

    /**
     * Sets the frame reserved for the label.
     *
     * @param rect the new label frame
     * @see   #getLabelFrame()
     */
    public void setLabelFrame(Rectangle rect)
    {
        _labelFrame = rect;
    }

    /**
     * Gets all points of this region polygon and of all its subcomponents.
     *
     * @return the list of points
     */
    public Vector<LyPoint> getPoints()
    {
        Vector<LyPoint> points = new Vector<LyPoint>();
        int             i, n = _region.getPolygonCount();
        LyPolygon       polygon;
        int             j, nPoints;

        for (i = 0; i < n; i++)
        {
            polygon = _region.getPolygon(i);
            nPoints = polygon.getPointCount();

            for (j = 0; j < nPoints; j++)
                points.addElement(polygon.getPoint(j));
        }

        return points;
    }

    /**
     * Gets the rectangle frame where the label can be drawn.
     *
     * @return the label frame
     * @see    #setLabelFrame(Rectangle)
     */
    public Rectangle getLabelFrame()
    {
        return _labelFrame;
    }

    /**
     * Gets the small image bounds.
     *
     * @return the small image bounds
     */
    public Rectangle getSmallImageBounds()
    {
        return _smallImageBounds;
    }
    
    /**
     * Generates the Javascript code that resets the position of the element.
     */
    public void resetPosition()
    {
    }
} // end class LyStrutsMapNode
