/*
 * LyStrutsRelation.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web.struts;

import java.awt.*;

import leon.misc.*;
import leon.view.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.web.*;


/**
 * This class implements the Relation object (as described in the dtd) for struts viewer.
 *
 * <p><b>RELATION: Input Field of Relation Type</b><br>
 * An input field of Relation type allows to choose one or more values among a list of values. This
 * list offers generally application objects belonging to a same class and is used to establish a
 * relation between an object and one or more objects of this class. It is possible to specify if
 * the relation is multiple or not.</p>
 *
 * @version  $Id: LyStrutsRelation.java 24563 2010-03-04 08:15:19Z jlsaouli $
 * @author   Lyria
 * @see      leon.view.LyRelation
 */
public class LyStrutsRelation extends LyStrutsFieldItem
    implements LyGraphicalRelation,
        LyGetValueInterface
{
    public static final String __VERSION = "$Id: LyStrutsRelation.java 24563 2010-03-04 08:15:19Z jlsaouli $";

    /** Default number of rows for list selectors. */
    static int DEFAULT_HEIGHT = 6;

    /** Default width for list selectors. */
    static int DEFAULT_WIDTH = 20;

    /** Label for right button. */
    protected String _rightStatus;

    /** Label for right all button. */
    protected String _rightAllStatus;

    /** Label for left button. */
    protected String _leftStatus;

    /** Label for left all button. */
    protected String _leftAllStatus;

    /** Label for up button. */
    protected String _upStatus;

    /** Label for down button. */
    protected String _downStatus;

    /** Image for right button. */
    protected String _rightButton;

    /** Image for right all button. */
    protected String _rightAllButton;

    /** Image for left button. */
    protected String _leftButton;

    /** Image for left all button. */
    protected String _leftAllButton;

    /** Image for up button. */
    protected String _upButton;

    /** Image for down button. */
    protected String _downButton;

    /** Disabled image for right button. */
    protected String _disabledRightButton;

    /** Disabled image for right all button. */
    protected String _disabledRightAllButton;

    /** Disabled image for left button. */
    protected String _disabledLeftButton;

    /** Disabled image for left all button. */
    protected String _disabledLeftAllButton;

    /** Disabled image for up button. */
    protected String _disabledUpButton;

    /** Disabled image for down button. */
    protected String _disabledDownButton;

    /** Image for auto-completion selection. */
    protected String _autoCompletionSelectButton;

    /** Image for auto-completion unselection. */
    protected String _autoCompletionUnselectButton;

    /** The relation generic component. */
    protected LyRelation _relation;

    /** Number of rows for list selectors. */
    protected int _height;

    /** Width for list selectors. */
    protected int _width;

	/** Minimum width for list selectors. */
	protected int _minWidth;

	/** Maximum width for list selectors. */
	protected int _maxWidth;

    /** Id of the component. */
    protected String _id;

    /** Indicates if a modification of the relation must send an Ajax update. */
    protected boolean _notify;

    /**
     * Class constructor of a Relation object for the struts target.
     *
     * @param  relation  the generic Relation defined for every type of view targets.
     */
    public LyStrutsRelation(LyRelation relation)
    {
        super(relation);
        _relation = relation;
        _notify = true;

        initialize();
    }

    /**
     * Initialization method.<br>
     * Sets the relation height and buttons images and labels.<br>
     */
    @SuppressWarnings("deprecation")
    public void initialize()
    {
		// Computing relation's width
		// Simple relation rules :
		//  - if positive, we use nCols converted in pixels
		//  - otherwise we check component width with the getMinWidth() and getMaxWidth() values that are already in pixels
		// Multiple relation rules :
		//  - if positive, we use nCols converted in pixels
		//  - otherwise if positive, we use LY_MULTIPLE_RELATION_WIDTH converted in pixels
		//  - otherwise we check component width with the getMinWidth() and getMaxWidth() values converted in pixels
		_width = -1;
		_minWidth = -1;
		_maxWidth = -1;
		
        int     nCols = _relation.getNCols();
        boolean	convertMinMaxWidth = false;

        if ((nCols <= 0) && _relation.isMultiple())
        {
            String      s = getResource("LY_MULTIPLE_RELATION_WIDTH");
            
            nCols = LyTools.intFromString(s);
        }
        
        if (nCols > 0)
        	_width = nCols;
        else
        {
            _minWidth = _relation.getMinWidth();
            _maxWidth = _relation.getMaxWidth();
            convertMinMaxWidth = _relation.isMultiple();
        }
        
        // Calculate estimation of size in pixels
        LyEnvironment   environment = getEnvironment();
        Font            font = (Font)environment.getFont("LY_DEFAULT_FONT");
        Toolkit         toolkit = Toolkit.getDefaultToolkit();
        FontMetrics     fm = toolkit.getFontMetrics(font);
        
        if (_width > 0)
        	_width = fm.charWidth('W') * _width + 20;
        
        if (convertMinMaxWidth)
        {
        	if (_minWidth > 0)
        		_minWidth = fm.charWidth('W') * _minWidth + 20;
        	
        	if (_maxWidth > 0)
        		_maxWidth = fm.charWidth('W') * _maxWidth + 20;
        }

        _height = _relation.getNRows();

        if (_height <= 0)
            _height = getIntResource("LY_MULTIPLE_RELATION_HEIGHT");

        if (_height <= 0)
            _height = DEFAULT_HEIGHT;

        _rightButton = getImageUrl(_relation.getSelectImage());
        _rightAllButton = getImageUrl(_relation.getSelectAllImage());
        _leftButton = getImageUrl(_relation.getDeselectImage());
        _leftAllButton = getImageUrl(_relation.getDeselectAllImage());
        _upButton = getImageUrl(_relation.getMoveUpImage());
        _downButton = getImageUrl(_relation.getMoveDownImage());
        _disabledRightButton = getDisabledImageUrl(_relation.getSelectImage());
        _disabledRightAllButton = getDisabledImageUrl(_relation.getSelectAllImage());
        _disabledLeftButton = getDisabledImageUrl(_relation.getDeselectImage());
        _disabledLeftAllButton = getDisabledImageUrl(_relation.getDeselectAllImage());
        _disabledUpButton = getDisabledImageUrl(_relation.getMoveUpImage());
        _disabledDownButton = getDisabledImageUrl(_relation.getMoveDownImage());
        _autoCompletionSelectButton = getImageUrl("LY_MULTI_SELECTOR_SELECT");
        _autoCompletionUnselectButton = getImageUrl("LY_MULTI_SELECTOR_UNSELECT");
        _rightStatus = getHtmlString("LY_SELECT");
        _rightAllStatus = getHtmlString("LY_SELECT_ALL");
        _leftStatus = getHtmlString("LY_UNSELECT");
        _leftAllStatus = getHtmlString("LY_UNSELECT_ALL");
        _upStatus = getHtmlString("LY_TOOL_MOVE_UP");
        _downStatus = getHtmlString("LY_TOOL_MOVE_DOWN");
    }

    /**
     * Indicates if a modification of the relation must send an Ajax update.
     *
     * @param  notify  true if an Ajax update must be sent, false otherwise.
     */
    public void setNotify(boolean notify)
    {
        _notify = notify;
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetLinksInterface#addLink(leon.view.LyLink,
     *      int)
     */
    public void addLink(LyLink link, int pos)
    {
        LyComponent parent = _relation.getParent();

        if (_notify)
        {
			if (!_relation.isMultiple() && (_relation.getAutoCompletionMax() > 0))
			{
				LyStrutsViewManager	viewManager = (LyStrutsViewManager)getViewManager();
				String				id = getIdCode() + "_noDiv";

				viewManager.addReloadDiv(_lastFrameName, id, this);
			}
			else if (parent instanceof LyFormField)
            {
                LyFormField       formField = (LyFormField)parent;
                LyStrutsFormField webField = (LyStrutsFormField)formField.getInternalView();

                webField.refreshDivContent();
            }
            else
                refresh();
        }
    }

    /*
    * (non-Javadoc)
    *
    * @see leon.view.setinterface.LySetLinksInterface#removeLink(leon.view.LyLink)
    */
    public void removeLink(LyLink link)
    {
        LyComponent parent = _relation.getParent();

        if (_notify)
        {
			if (!_relation.isMultiple() && (_relation.getAutoCompletionMax() > 0))
			{
				LyStrutsViewManager	viewManager = (LyStrutsViewManager)getViewManager();
				String				id = getIdCode() + "_noDiv";

				viewManager.addReloadDiv(_lastFrameName, id, this);
			}
			else if (parent instanceof LyFormField)
            {
                LyFormField       formField = (LyFormField)parent;
                LyStrutsFormField webField = (LyStrutsFormField)formField.getInternalView();

                webField.refreshDivContent();
            }
            else
                refresh();
        }
    }

    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
     */
    public void writeHtml(LyWebPageWriter pw)
    {
        pw.startComponent(this);

        updateRefreshInfo(pw);

        String       name;
        String       formName = "forms[0]"; // getFormName(pw);
        boolean      isMultiple = _relation.isMultiple();
        boolean      isEditable = isEditable();
        boolean      needCheck = needCheck();
        int          i, nLinks = _relation.getLinkCount();
        LyLink       link;
        LyStrutsLink internal;
        String       sep = LyWebParameterValues.HTML_ENCODED_SEP;
        String       linkCrSep = LyWebParameterValues.HTML_LINK_CR_SEP;
        String       disabledColor = getHexColor("LY_DISABLED_BACKGROUND");
        boolean		 isRelationUnit = _relation.getParent() instanceof LyFieldUnit;
        LyForm		 parentForm = _relation.getParentForm();

        if (isRelationUnit)
        {
        	name = "LyUnitRelationSelector." + getId();
        	needCheck = true;
        }
        else if (_id == null)
            name = getId();
        else
            name = _id;

        String onMouseOverFunctions = _relation.getEventFunctions(LyComponent.EventType.MOUSE_ENTERED, ';');
        String onMouseOutFunctions = _relation.getEventFunctions(LyComponent.EventType.MOUSE_EXITED, ';');

        if (isEditable)
        {
            // If not multiple, create a combobox
            if (!isMultiple)
            {
                if (_relation.getAutoCompletionMax() == 0)
                {
                    String selectName = name;
                    String onChange = "";
                    String styleParam = getSimpleRelationStyle();

                    // Select input may not be updated on Firefox. Use a hidden input to
                    // change the field value and a combo with another name to display it.
                    if (!LyEnvironment.TEST_MODE
                            && (getWebViewManager().getBrowser() == LyWebViewManager.BROWSER_FIREFOX)
                            && getWebViewManager().getBrowserVersion() < 4.0)
                    {
                        String initialValue = ""; // Don't use empty value, there is nothing selected

                        for (i = nLinks - 1; i >= 0; i--)
                        {
                            link = _relation.getLink(i);

                            if (((link != null) && link.isSelected())
                            		|| (i == 0))
                            {
                                initialValue = link.getId();
                                initialValue = LyString.getHtmlEntitiesString(initialValue);
                                break;
                            }
                        }

                        pw.append("<input type='hidden' name='");
                        pw.append(name);
                        pw.append("' value='");
                        pw.append(initialValue);
                        pw.append("' />\n");

                        selectName += System.currentTimeMillis();
                        onChange = " onChange=\"javascript:window.document." + formName;
                        onChange += ".elements['" + name + "'].value=this.value;";

                        if (needCheck)
                        {
                            onChange += "javascript:FieldChange(";
                            onChange += "window.document." + formName;
                            onChange += ",window.document." + formName + ".elements['" + name + "'],'";
                            onChange += _relation.getRootController();
                            onChange += "');";
                        }

                        onChange += "\"";
                    }

                    pw.append("<select name='");
					pw.append(selectName);
	                pw.append("' id='");
					pw.append(selectName);
                    pw.append("'");
					pw.append(onChange);

                    if (needCheck && (onChange.length() == 0))
                    {
                        pw.append(" onChange=\"");
                        pw.append("javascript:FieldChange(");
                        pw.append("window.document." + formName);
                        pw.append(",this,'");
                        pw.append(_relation.getRootController());
                        pw.append("');\"");
                    }

                    if (styleParam != null)
                        pw.append(" style='" + styleParam + "'");

                    if (onMouseOverFunctions.length() > 0)
                        pw.append(" onmouseover=\"" + onMouseOverFunctions + "\"");

                    if (onMouseOutFunctions.length() > 0)
                        pw.append(" onmouseout=\"" + onMouseOutFunctions + "\"");

                    // Only fields that are in a form may get automatic focus
                    if (parentForm == null)
                    	pw.append(" noAutomaticFocus='true'");
                    
                    pw.append(">\n");

                    for (i = 0; i < nLinks; i++)
                    {
                        link = _relation.getLink(i);
                        
                        if (link != null)
                        {
	                        internal = (LyStrutsLink)link.getInternalView();
	                        internal.writeHtml(pw);
                        }
                    }

                    pw.append("</select>\n");

					if ((_minWidth > 0) || (_maxWidth > 0))
					{
						String	script = "<script type='text/javascript'>CheckComponentMinMaxWidth('" + selectName + "'," + _minWidth + "," + _maxWidth + ");</script>";
						
						getWebViewManager().addAdditionalRefreshScripts(script);
						pw.addEndScript(script, false);
					}
                }
                else
                    writeAutoCompletionSelector(pw, false, true);
            }
            // If multiple, create a multi-selector
            else
            {
                StringBuffer value = new StringBuffer();

                for (i = 0; i < nLinks; i++)
                {
                    link = _relation.getLink(i);

                    if ((link != null) && link.isSelected())
                    {
                        if (value.length() > 0)
                            value.append(sep);

                        value.append(link.getId());
                    }
                }

                String nameFrom = name + "_from";
                String nameTo = name + "_to";
                String controllerId = _relation.getRootController();

                if (!needCheck)
                    controllerId = "false";

                if (_relation.getAutoCompletionMax() != 0)
                    writeAutoCompletionSelector(pw, true, true);
                else
                {
                    pw.append("<input type='hidden' name='");
                    pw.append(name);
                    pw.append("' value='");
                    pw.append(LyString.replace(value.toString(), "\\n", LyWebParameterValues.HTML_LINK_CR_SEP));
                    pw.append("' />\n");
                }

                pw.append("<table border='0'>\n");

                String  leftHeader = translate("LY_MULTISELECTOR_LEFT_HEADER");
                String  rightHeader = translate("LY_MULTISELECTOR_RIGHT_HEADER");
                String  disabledToolStyle = " class='disabledTool' style='cursor:default;'";
                boolean hasUnselectedLink = false;
                boolean hasSelectedLink = false;

                pw.append("<tr>\n");

                if (_relation.getAutoCompletionMax() == 0)
                {
                    // Adding the first list
                    pw.append("<td>\n");

                    // Adding the left label if needed
                    if ((leftHeader != null) && (leftHeader.length() > 0))
                    {
                        pw.append("<table border='0' cellpadding='3' cellspacing='0' width='100%'>\n");
                        pw.append("<tr>\n");
                        pw.append("<td class='header' nowrap='yes'><font class='label'>");
                        pw.append(getHtmlString(leftHeader));
                        pw.append("</font></td>\n");
                        pw.append("<tr>\n");
                        pw.append("</table>\n");
                    }

                    String style = getSimpleRelationStyle();

                    // The name of the first list is name + "_from"
					pw.append("<select name='");
					pw.append(nameFrom);
	                pw.append("' id='");
					pw.append(nameFrom);
					pw.append("' size='" + _height + "'");
					
                    if (style != null)
                        pw.append(" style='" + style + "'");
					
                    pw.append(" ondblclick=\"javascript:InvokeAction(null, null, null, '" + controllerId
                        + "', this.value, null);;\"");
                    pw.append(" onChange=\"javascript:UpdateTools(window.document." + formName + ",'" + name + "');\"");
                    pw.append(" multiple='true'>\n");

                    for (i = 0; i < nLinks; i++)
                    {
                        link = _relation.getLink(i);
                        
                        if (link != null)
                        {
	                        if (!link.isSelected())
	                        {
	                            internal = (LyStrutsLink)link.getInternalView();
	                            internal.writeHtml(pw);
	                            hasUnselectedLink = true;
	                        }
	                        else
	                            hasSelectedLink = true;
                        }
                    }

                    pw.append("</select>\n");
                    pw.append("</td>\n");
                    pw.append("<td valign='middle'>\n");

                    // Adding the four center buttons
                    pw.append("<table>\n");
                    pw.append("<tr>\n");
                    pw.append("<td>\n");
                    pw.append("<a href=\"javascript:SelectAllOptions(window.document." + formName + ",'"
                        + nameFrom + "','" + nameTo + "','" + nameTo + "','" + name + "','" + sep + "','" + linkCrSep
                        + "','" + controllerId + "')\">\n");
                    pw.append("<img id='" + name + "_selectAll' src=\"" + _rightAllButton + "\" border='0' title='"
                        + _rightAllStatus + "'"
                        + (hasUnselectedLink ? "" : disabledToolStyle) + " /></a></td>\n");
                    pw.append("</tr>\n");

                    pw.append("<tr>\n");
                    pw.append("<td>\n");
                    pw.append("<a href=\"javascript:SelectOptions(window.document." + formName + ",'"
                        + nameFrom + "','" + nameTo + "','" + nameTo + "','" + name + "','" + sep + "','" + linkCrSep
                        + "','" + controllerId + "')\">\n");
                    pw.append("<img id='" + name + "_select' src=\"" + _rightButton + "\" border='0' title='"
                        + _rightStatus + "'"
                        + disabledToolStyle + " /></a></td>\n");
                    pw.append("</tr>\n");

                    pw.append("<tr>\n");
                    pw.append("<td>\n");
                    pw.append("<a href=\"javascript:SelectOptions(window.document." + formName + ",'"
                        + nameTo + "','" + nameFrom + "','" + nameTo + "','" + name + "','" + sep + "','" + linkCrSep
                        + "','" + controllerId + "',true)\">\n");
                    pw.append("<img id='" + name + "_unselect' src=\"" + _leftButton + "\" border='0' title='"
                        + _leftStatus + "'"
                        + disabledToolStyle + " /></a></td>\n");
                    pw.append("</tr>\n");

                    pw.append("<tr>\n");
                    pw.append("<td>\n");
                    pw.append("<a href=\"javascript:SelectAllOptions(window.document." + formName
                        + ",'" + nameTo + "','" + nameFrom + "','" + nameTo + "','" + name + "','"
                        + sep + "','" + linkCrSep + "','" + controllerId + "')\">\n");
                    pw.append("<img id='" + name + "_unselectAll' src=\"" + _leftAllButton + "\" border='0' title='"
                        + _leftAllStatus + "'"
                        + (hasSelectedLink ? "" : disabledToolStyle) + " /></a></td>\n");
                    pw.append("</tr>\n");
                    pw.append("</table>\n");
                    pw.append("</td>\n");
                }

                // Adding the second list
                pw.append("<td>\n");

                // Adding the right label if needed
                if ((_relation.getAutoCompletionMax() == 0) && (rightHeader != null) && (rightHeader.length() > 0))
                {
                    pw.append("<table border='0' cellpadding='3' cellspacing='0' width='100%'>\n");
                    pw.append("<tr>\n");
                    pw.append("<td class='header' nowrap='yes'><font class='label'>");
                    pw.append(getHtmlString(rightHeader));
                    pw.append("</font></td>\n");
                    pw.append("<tr>\n");
                    pw.append("</table>\n");
                }

                String style = getSimpleRelationStyle();

				pw.append("<select name='");
				pw.append(nameTo);
                pw.append("' id='");
				pw.append(nameTo);
				pw.append("' size='" + _height + "'");
				
                if (style != null)
                    pw.append(" style='" + style + "'");
				
                pw.append(" ondblclick=\"javascript:InvokeAction(null, null, null, '" + controllerId
                    + "', this.value, null);;\"");
                pw.append(" onChange=\"javascript:UpdateTools(window.document." + formName + ",'" + name + "');\"");
                pw.append(" multiple='true'>\n");

                for (i = 0; i < nLinks; i++)
                {
                    link = _relation.getLink(i);
                    
                    if ((link != null) && link.isSelected())
                    {
                        internal = (LyStrutsLink)link.getInternalView();
                        internal.writeHtml(pw);
                    }
                }

                pw.append("</select>\n");
                pw.append("</td>\n");

                if (_relation.isOrdered())
                {
                    pw.append("<td valign='top'>\n");
                    pw.append("<table>\n");
                    pw.append("<tr>\n");

                    // Adding the up button
                    pw.append("<td>\n");
                    pw.append("<a href=\"javascript:MoveUpOptions(window.document.");
                    pw.append(formName);
                    pw.append(",'");
                    pw.append(nameTo);
                    pw.append("','");
                    pw.append(name);
                    pw.append("','" + sep);
                    pw.append("','" + linkCrSep);
                    pw.append("','");
                    pw.append(controllerId);
                    pw.append("')\">\n");
                    pw.append("<img src=\"");
                    pw.append(_upButton);
                    pw.append("\" id='" + name + "_up' border='0' title='" + _upStatus + "'"
                        + disabledToolStyle + " /></a>\n");
                    pw.append("</td>\n");
                    pw.append("</tr>\n");

                    // Adding the down button
                    pw.append("<tr>\n");
                    pw.append("<td>\n");
                    pw.append("<a href=\"javascript:MoveDownOptions(window.document.");
                    pw.append(formName);
                    pw.append(",'");
                    pw.append(nameTo);
                    pw.append("','");
                    pw.append(name);
                    pw.append("','" + sep);
                    pw.append("','" + linkCrSep);
                    pw.append("','");
                    pw.append(controllerId);
                    pw.append("')\">\n");
                    pw.append("<img src=\"");
                    pw.append(_downButton);
                    pw.append("\" id='" + name + "_down' border='0' title='" + _downStatus + "'"
                        + disabledToolStyle + " /></a>\n");
                    pw.append("</td>");
                    pw.append("</tr>");
                    pw.append("</table>");
                    pw.append("</td>");
                }

                pw.append("</tr>\n");
                pw.append("</table>\n");

                StringBuffer script = new StringBuffer();

                script.append("<script type='text/javascript'>");
                script.append("UpdateTools(window.document." + formName + ",'" + name + "');");

				if ((_minWidth > 0) || (_maxWidth > 0))
					script.append("CheckComponentMinMaxWidth('" + nameFrom + "'," + _minWidth + "," + _maxWidth + ",'" + nameTo + "');");

                script.append("</script>");
                pw.addEndScript(script.toString(), false);
                getWebViewManager().addAdditionalRefreshScripts(script.toString());
            }
        }
        else
        {
            // If not multiple, just write the selected link
            if (!isMultiple)
            {
                if (canDisabled())
                // when disabled property may be use on <select>
                {
                    String styleParam = getSimpleRelationStyle();

	                pw.append("<select name='");
					pw.append(name);
	                pw.append("' id='");
					pw.append(name);
					pw.append("' disabled='true'");

                    if (styleParam != null)
                        pw.append(" style='" + styleParam + "'");

                    pw.append(">\n");

                    for (i = 0; i < nLinks; i++)
                    {
                        link = _relation.getLink(i);
                        
                        if (link != null)
                        {
	                        internal = (LyStrutsLink)link.getInternalView();
	                        internal.writeHtml(pw);
                        }
                    }

                    pw.append("</select>\n");

                    if (_relation.getAutoCompletionMax() != 0)
                        writeAutoCompletionSelector(pw, false, false);

					if ((_minWidth > 0) || (_maxWidth > 0))
					{
						String	script = "<script type='text/javascript'>CheckComponentMinMaxWidth('" + name + "'," + _minWidth + "," + _maxWidth + ");</script>";
						
						getWebViewManager().addAdditionalRefreshScripts(script);
						pw.addEndScript(script, false);
					}
                }
                else
                {
                    String label;

                    for (i = 0; i < nLinks; i++)
                    {
                        link = _relation.getLink(i);
                        
                        if ((link != null) && link.isSelected())
                        {
                            label = link.getLabel();
                            label = translate(label);
                            pw.append("<font class='label'>");
                            pw.append(getHtmlString(label));
                            pw.append("</font>");
                            break;
                        }
                    }

                    if (_relation.getAutoCompletionMax() != 0)
                        writeAutoCompletionSelector(pw, false, false);
                }
            }
            // If multiple, create a disabled multi-selector
            else
            {
                String leftHeader = translate("LY_MULTISELECTOR_LEFT_HEADER");
                String rightHeader = translate("LY_MULTISELECTOR_RIGHT_HEADER");
                String nameFrom = name + "_from";
                String nameTo = name + "_to";

                // Adding the first list
                pw.append("\n");

                pw.append("<table border='0'>\n");
                pw.append("<tr>\n");
                pw.append("<td rowspan='4'>\n");

                String style = getSimpleRelationStyle();

                // Adding the left label if needed
                if ((leftHeader != null) && (leftHeader.length() > 0))
                {
                    pw.append("<table border='0' cellpadding='3' cellspacing='0' width='100%'>\n");
                    pw.append("<tr>\n");
                    pw.append("<td class='header' nowrap='yes'><font class='label'>");
                    pw.append(getHtmlString(leftHeader));
                    pw.append("</font></td>\n");
                    pw.append("<tr>\n");
                    pw.append("</table>\n");
                }

                // The name of the first list is name + "_from"
				pw.append("<select name='");
				pw.append(nameFrom);
                pw.append("' id='");
				pw.append(nameFrom);
				pw.append("' size='" + _height + "'");
				
                if (style != null)
                    pw.append(" style='" + style + "'");
				
                pw.append("' multiple='true'");
                pw.append("  disabled='true'");
				
				if ((_width > 0) || (disabledColor != null))
				{
					pw.append(" style='");

					if (_width > 0)
						pw.append("width:" + _width + "px;");
				
					if (disabledColor != null)
						pw.append("background-color:#" + disabledColor + ";");
					
					pw.append("'");
				}
				
                pw.append(">\n");

                for (i = 0; i < nLinks; i++)
                {
                    link = _relation.getLink(i);
                    
                    if ((link != null) && !link.isSelected())
                    {
                        internal = (LyStrutsLink)link.getInternalView();
                        internal.writeHtml(pw);
                    }
                }

                pw.append("</select>\n");
                pw.append("</td>\n");
                pw.append("<td valign='middle'>\n");

                // Adding the four center buttons
                pw.append("<table>\n");
                pw.append("<tr>\n");
                pw.append("<td>\n");
                pw.append("<img id='" + name + "_selectAll' src=\"" + _rightAllButton + "\" border='0' title='"
                    + _rightAllStatus + "'"
                    + " class='disabledTool' style='cursor:default;'/></td>\n");
                pw.append("</tr>\n");

                pw.append("<tr>\n");
                pw.append("<td>\n");
                pw.append("<img id='" + name + "_select' src=\"" + _rightButton + "\" border='0' title='" + _rightStatus
                    + "'"
                    + " class='disabledTool' style='cursor:default;'/></td>\n");
                pw.append("</tr>\n");

                pw.append("<tr>\n");
                pw.append("<td>\n");
                pw.append("<img id='" + name + "_unselect' src=\"" + _leftButton + "\" border='0' title='" + _leftStatus
                    + "'"
                    + " class='disabledTool' style='cursor:default;'/></td>\n");
                pw.append("</tr>\n");

                pw.append("<tr>\n");
                pw.append("<td>\n");
                pw.append("<img id='" + name + "_unselectAll' src=\"" + _leftAllButton + "\" border='0' title='"
                    + _leftAllStatus + "'"
                    + " class='disabledTool' style='cursor:default;'/></td>\n");
                pw.append("</tr>\n");
                pw.append("</table>\n");
                pw.append("</td>\n");

                // Adding the second list
                pw.append("<td>\n");

                // Adding the right label if needed
                if ((_relation.getAutoCompletionMax() == 0) && (rightHeader != null) && (rightHeader.length() > 0))
                {
                    pw.append("<table border='0' cellpadding='3' cellspacing='0' width='100%'>\n");
                    pw.append("<tr>\n");
                    pw.append("<td class='header' nowrap='yes'><font class='label'>");
                    pw.append(getHtmlString(rightHeader));
                    pw.append("</font></td>\n");
                    pw.append("<tr>\n");
                    pw.append("</table>\n");
                }

				pw.append("<select name='");
				pw.append(nameTo);
                pw.append("' id='");
				pw.append(nameTo);
				pw.append("' size='" + _height);
				pw.append("' multiple='true'");
				pw.append(" disabled='true'");
				
				if ((_width > 0) || (disabledColor != null))
				{
					pw.append(" style='");

					if (_width > 0)
						pw.append("width:" + _width + "px;");
				
					if (disabledColor != null)
						pw.append("background-color:#" + disabledColor + ";");
					
					pw.append("'");
				}

                pw.append(">\n");

                for (i = 0; i < nLinks; i++)
                {
                    link = _relation.getLink(i);
                    
                    if ((link != null) && link.isSelected())
                    {
                        internal = (LyStrutsLink)link.getInternalView();
                        internal.writeHtml(pw);
                    }
                }

                pw.append("</select>\n");
                pw.append("</td>\n");
                pw.append("</tr>\n");
                pw.append("</table>\n");

                if (_relation.getAutoCompletionMax() != 0)
                    writeAutoCompletionSelector(pw, true, false);

				if ((_minWidth > 0) || (_maxWidth > 0))
				{
					StringBuffer	script = new StringBuffer();
					
					script.append("<script type='text/javascript'>");
					script.append("CheckComponentMinMaxWidth('" + nameFrom + "'," + _minWidth + "," + _maxWidth + ",'" + nameTo + "');");
					script.append("</script>");

					pw.addEndScript(script.toString(), false);
					getWebViewManager().addAdditionalRefreshScripts(script.toString());
				}
            }
        }

        pw.endComponent(this);
    } // end method writeHtml

	public String getRefreshContent(LyWebPageWriter pw)
	{
		String	    hiddenTextName;
		String	    formName = "forms[0]";//getFormName(pw);
		boolean     needCheck = needCheck();
		int			maxChars = (_relation.getNCols() > 0 ? _relation.getNCols() : 20);
		boolean		multiple = _relation.isMultiple();

		if (_id == null)
			hiddenTextName = getId();
		else
			hiddenTextName = _id;
		
		if (multiple)
			hiddenTextName += "_internal";
		
		String		 jsName = "AutoCompletionListe_" + _relation.getIdCode();
		int			 autoCompletionMax = _relation.getAutoCompletionMax();
		StringBuffer listeVar = new StringBuffer();
		String		 colorCode = "LY_HIGHLIGHTED_BACKGROUND";
		String		 bgColor = "";
		String		 controllerParam = (needCheck ? ", '" + _relation.getRootController()
				+ "'" : ", ''");

		int			 autoCompletionwidth = (maxChars * 5) + 33;

		if (colorCode != null)
			bgColor = getHexColor(colorCode);
		
		listeVar.append(writeAutoCompletionArray(pw));
		listeVar.append("refreshAutoCompletion((typeof window.document." + formName
				+ " == 'undefined' ? null : window.document." + formName
				+ ".name), '" + _relation.getIdCode() + "', '" + getJavascriptString(hiddenTextName) + "', '"
				+ jsName + "', '" + bgColor + "'," + autoCompletionMax
				+ ",'" + (autoCompletionwidth - 3) + "', null" + controllerParam + ", false, false);");
		
		return listeVar.toString();
	}
	
	protected String writeAutoCompletionArray(LyWebPageWriter pw)
	{
		String		linkLabel;
		int		    i, nLinks = _relation.getLinkCount();
		int			maxChars = 20;
		LyLink	    link;
		boolean		multiple = _relation.isMultiple();
		
		String		 jsName = "AutoCompletionListe_" + _relation.getIdCode();
		String[][]   autoCompletionList = new String[nLinks][3];
		String		 autoCompletionId, autoCompletionName;
		StringBuffer listeVar = new StringBuffer();
		int			 n = nLinks;

		for (i = 0; i < nLinks; i++)
		{
			link = _relation.getLink(i);
			// Force construction of internal view to receive Ajax demands on change
			link.getInternalView();
			linkLabel = translate(link.getLabel());
			
			if ((linkLabel != null) && (linkLabel.length() > maxChars))
				maxChars = linkLabel.length();
			
			autoCompletionList[i][0] = link.getId();
			autoCompletionList[i][1] = linkLabel;
			autoCompletionList[i][2] = Boolean.toString(!multiple || !link.isSelected());
		}

        // If write div only, var was already generated
        if (!pw.writeDivOnly())
            listeVar.append("var ");

        listeVar.append("" + jsName + " = new Array();\n");

        for (i = 0; i < n; i++)
        {
            autoCompletionId = autoCompletionList[i][0];
            autoCompletionId = LyString.replace(autoCompletionId, "\\", "\\\\");
            autoCompletionId = LyString.replace(autoCompletionId, "\"", "\\\"");
            autoCompletionName = autoCompletionList[i][1];
            autoCompletionName = LyString.replace(autoCompletionName, "\\", "\\\\");
            autoCompletionName = LyString.replace(autoCompletionName, "\"", "\\\"");
            listeVar.append(jsName + "[" + i + "] = [\"" + autoCompletionId + "\", \"" + autoCompletionName + "\", "
                + autoCompletionList[i][2] + "];\n");
        }

        listeVar.append("var relationFrame = findTarget(null, '" + _lastFrameName + "', null);\n");
        listeVar.append("if (relationFrame != null) relationFrame.addAutoCompletionValues('" + jsName + "', " + jsName
            + ");\n");
        listeVar.append("else addAutoCompletionValues('" + jsName + "', " + jsName + ")\n");
		
		return listeVar.toString();
	}
	
    /**
     * Write HTML description corresponding to an auto-completion selector.
     *
     * @param  pw          the page writer.
     * @param  multiple    indicates whether selector is a multi-selector or a single one.
     * @param  isEditable  indicates that the relation is editable
     */
    protected void writeAutoCompletionSelector(LyWebPageWriter pw, boolean multiple, boolean isEditable)
    {
        String  hiddenTextName, textName, hiddenName;
        String  initialHiddenTextValue = "", initialHiddenValue = "", initialName = "";
        String  formName = "forms[0]"; // getFormName(pw);
        String  linkLabel;
        boolean needCheck = needCheck();
        int     i, nLinks = _relation.getLinkCount();
        int     maxChars = (_relation.getNCols() > 0 ? _relation.getNCols() : 20);
        LyLink  link;

        if (_id == null)
            hiddenTextName = getId();
        else
            hiddenTextName = _id;

        hiddenName = hiddenTextName;

        if (multiple)
            hiddenTextName += "_internal";

        textName = hiddenTextName + "_text";

        if (isEditable)
        {
            pw.append("<table border='0' cellpadding='0' cellspacing='0'><tr><td>");
            pw.append("<input");
            pw.append(" name='" + textName + "' size='" + maxChars + "'");
        }

        String       jsName = "AutoCompletionListe_" + _relation.getIdCode();
        int          autoCompletionMax = _relation.getAutoCompletionMax();
        int          autoCompletionHeight = 0;
        String[][]   autoCompletionList = new String[nLinks][3];
        StringBuffer listeVar = new StringBuffer();
        int          n = nLinks;
        String       colorCode = "LY_HIGHLIGHTED_BACKGROUND";
        String       bgColor = "";
        String       sep = LyWebParameterValues.HTML_ENCODED_SEP;
        String       controllerParam = (needCheck ? (", '" + _relation.getRootController() + "'") : ", ''");
        
        for (i = 0; i < nLinks; i++)
        {
            link = _relation.getLink(i);
            
            if (link == null)
            	continue;
            
            // Force construction of internal view to receive Ajax demands on change
            link.getInternalView();
            linkLabel = translate(link.getLabel());

            if (linkLabel != null)
                linkLabel = linkLabel.replace('\n', ':');

            if ((linkLabel != null)
                    && (linkLabel.length() > maxChars))
                maxChars = linkLabel.length();

            autoCompletionList[i][0] = link.getId();
            autoCompletionList[i][1] = linkLabel;
            autoCompletionList[i][2] = Boolean.toString(!multiple || !link.isSelected());

            if ((!multiple
                    && link.isSelected())
                || (multiple
                    && !link.isSelected()
                    && (initialHiddenTextValue.length() == 0)))
            {
                initialHiddenTextValue = autoCompletionList[i][0];
                initialName = autoCompletionList[i][1];
            }

            if (multiple
                && link.isSelected())
            {
                if (initialHiddenValue.length() > 0)
                    initialHiddenValue += sep;

                initialHiddenValue += autoCompletionList[i][0];
            }
        }

        int autoCompletionwidth = (maxChars * 5) + 33;

        bgColor = getHexColor(colorCode);

        listeVar.append("<script type='text/javascript'>\n");
		
		listeVar.append(writeAutoCompletionArray(pw));

        listeVar.append("</script>\n");

        getWebViewManager().addAdditionalRefreshScripts(listeVar.toString());
        pw.addBeginScript(listeVar.toString(), false);

        if (isEditable)
        {
            if (autoCompletionMax > 0)
                n = autoCompletionMax;

            autoCompletionHeight = Math.min(150, (n * 22) + 2);

            pw.append(" value='" + getHtmlEntitiesString(initialName) + "'");
            
			String	refreshScript = "refreshAutoCompletion((typeof window.document." + formName
						+ " == 'undefined' ? null : window.document." + formName
						+ ".name), '" + _relation.getIdCode() + "', '" + getJavascriptString(hiddenTextName) + "', '"
						+ jsName + "', '" + bgColor + "'," + autoCompletionMax
						+ ",'" + (autoCompletionwidth - 3) + "', event" + controllerParam + ", false, false);";

			if (autoCompletionMax > 0)
			{
				LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
				String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);

				pw.append(" onKeyUp=\"var key = (!document.all) ? event.which : event.keyCode;");
				pw.append("if ((key == 38) || (key == 40) || (key == 13)) " + refreshScript + " else ");
				pw.append("sendAutoCompletionValue('" + actionUrl + "','" + _relation.getRootController()
						+ "', event, this.value);\"");
			}
			else
				pw.append(" onKeyUp=\"" + refreshScript + "\"");
			
            pw.append(" onFocus=\"javascript:resetAutoCompletionFocus(window.document." + formName
                + ".name, '" + _relation.getIdCode() + "', '" + getJavascriptString(hiddenTextName) + "');\"");
            pw.append(" onBlur=\"javascript:hideAutoCompletionTimeout(window.document." + formName
                + ".name, '" + _relation.getIdCode() + "', '" + getJavascriptString(hiddenTextName) + "');\"");
            pw.append(" autocomplete='off'>\n");
            pw.append("<input name='" + hiddenTextName + "' type='hidden' value='" + initialHiddenTextValue + "'/>\n");

            if (multiple)
                pw.append("<input name='" + hiddenName + "' type='hidden' value='" + initialHiddenValue + "'/>\n");

            pw.append("</td><td>");
            pw.append("<img src='" + getImageUrl("LY_ARROW_DOWN_BIG_BUTTON") + "'");
            pw.append(" onClick=\"switchAutoCompletion(window.document." + formName
                + ".name, '" + _relation.getIdCode() + "', '" + getJavascriptString(hiddenTextName) + "', '"
                + jsName + "', '" + bgColor + "'," + autoCompletionMax
                + ",'" + (autoCompletionwidth - 3) + "', event" + controllerParam + ");\"");
            pw.append("/>");
            pw.append("</td>");

            if (multiple)
            {
                pw.append("<td><img src='" + _autoCompletionSelectButton
                    + "' onClick=\"AutoCompletionSelectOption(window.document." + formName
                    + ".name, '" + getJavascriptString(hiddenTextName) + "', '" + getJavascriptString(textName)
                    + "', '" + getJavascriptString(hiddenName + "_to") + "', '" + getJavascriptString(hiddenName)
                    + "', '" + _relation.getIdCode() + "', " + jsName + ", '" + sep + "'" + controllerParam
                    + ");\"/></td>");
                pw.append("<td><img src='" + _autoCompletionUnselectButton
                    + "' onClick=\"AutoCompletionUnselectOption(window.document." + formName
                    + ".name, '" + getJavascriptString(hiddenTextName) + "', '" + getJavascriptString(textName)
                    + "', '" + getJavascriptString(hiddenName + "_to") + "', '" + getJavascriptString(hiddenName)
                    + "', '" + _relation.getIdCode() + "', " + jsName + ", '" + sep + "'" + controllerParam
                    + ");\"/></td>");
            }

            pw.append("</tr></table>");
            pw.append("<input name='AutoCompletionFocus_" + _relation.getIdCode()
                + "' type='hidden' value='false'/>\n");
            pw.append("<input name='AutoCompletionIndex_" + _relation.getIdCode()
                + "' type='hidden' value='-1'/>\n");
            pw.append("<div class='autoCompletion' id='AutoCompletion_" + _relation.getIdCode()
                + "'");
            pw.append(" style='position:absolute;width:" + autoCompletionwidth + ";height:"
                + autoCompletionHeight);
            pw.append(
                ";visibility:hidden;overflow:auto;z-index:100;table-layout:fixed;border:solid 1px #000000;background-color:#FFFFFF;'");
            pw.append(" scrolling='auto' frameborder='0'");
            pw.append(" onFocus=\"resetAutoCompletionFocus(window.document." + formName
                + ".name, '" + _relation.getIdCode() + "', '" + getJavascriptString(hiddenTextName) + "');\"");
            pw.append(" onBlur=\"hideAutoCompletionTimeout(window.document." + formName
                + ".name, '" + _relation.getIdCode() + "', '" + getJavascriptString(hiddenTextName) + "'"
                + controllerParam
                + ");\">");
            pw.append("</div>\n");

            // Add an iframe to hide other components that may appear over the div
            // (bug on IE with div and select components).
            if (getWebViewManager().getBrowser() == LyWebViewManager.BROWSER_MSIE)
            {
                pw.append("<iframe name='AutoCompletionFrame_" + _relation.getIdCode() + "'");
                pw.append(" style='position:absolute;width=" + autoCompletionwidth
                    + ";height=" + autoCompletionHeight);
                pw.append(";visibility:hidden;z-index:90;'");
                pw.append(" scrolling='no' frameborder='0'>");
                pw.append("</iframe>\n");
            }
        }

        pw.endComponent(this);
    }

    /**
     * Returns the CSS style that must be used for a simple relation.
     *
     * @return  a string representing the CSS style to be used for a simple relation.
     */
    protected String getSimpleRelationStyle()
    {
        boolean isEditable = isEditable();
        boolean widthSet = false;
        String  style = "";

        if (!isEditable && !_relation.isDefaultDisabledColor())
            style += " background-color:" + getHexColor(_relation.getDisabledColor()) + ";";
        else if (!_relation.isDefaultEnabledColor())
            style += " background-color:" + getHexColor(_relation.getEnabledColor()) + ";";

        if (!_relation.isDefaultFont() || !_relation.isDefaultFontColor())
            style += getFontString(_relation.getFont(), _relation.getFontColor());

        // If a constraint is set on the relation, use the nCols and nRows values to compute
        // relation size in pixels.
        if ((_relation.getParent() instanceof LyFormField)
                && (((LyFormField)(_relation.getParent())).getConstraint() != null))
        {
            LyConstraint constraint = ((LyFormField)(_relation.getParent())).getConstraint();

            if (constraint != null)
            {
                if (constraint.getNCols() > 0)
                {
                    int width = constraint.getNCols();

                    if (_relation.isMultiple())
                        width = (width - 50) / 2;

                    style += "width:" + width + "px;";
                    widthSet = true;
                }

                if (constraint.getNRows() > 0)
                    style += "height:" + constraint.getNRows() + "px;";
            }
        }

		if (!widthSet && (_width > 0))
			style += "width:" + _width + "px;";

        // compatibility with old code.
        if (style.trim().length() == 0)
            style = null;

        return style;
    }

    /**
     * Sets id of this formField.
     *
     * @param  id  New id of the formField.
     */
    public void setId(String id)
    {
        _id = id;
    }

    /*
     * (non-Javadoc)
     * @see leon.view.getinterface.LyGetValueInterface#getValue()
     */
    public String getValue()
    {
        int          i, n = _relation.getLinkCount();
        LyLink       link;
        StringBuffer sb = new StringBuffer();
        char         sep = getEnvironment().getString().SEP;

        for (i = 0; i < n; i++)
        {
            link = _relation.getLink(i);

            if ((link == null) || !link.isSelected())
                continue;

            if (sb.length() > 0)
                sb.append(sep);

            sb.append(link.getId());
        }

        if (sb.length() == 0)
            return null;
        else
            return (sb.toString());
    }
} // end class LyStrutsRelation
