/*
 * LyStrutsRow.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web.struts;

import java.util.*;

import leon.misc.*;
import leon.view.*;
import leon.view.LyTable.*;
import leon.view.graphic.*;
import leon.view.web.*;


/**
 * This class implements the Row object (as described in the dtd) for struts viewer.
 *
 * <p><b>ROW: Row</b><br>
 * A row of a list is a set of non-empty cells (or input fields). It is possible to indicate if the
 * row is selected, if an image is assigned to it and if a tooltip is associated to it.</p>
 *
 * @version $Id: LyStrutsRow.java 41069 2012-10-08 14:30:49Z nhamon $
 * @author  Lyria
 * @see     leon.view.LyRow
 */
public class LyStrutsRow extends LyWebObject
    implements LyGraphicalRow
{
	public static final String __VERSION =
		"$Id: LyStrutsRow.java 41069 2012-10-08 14:30:49Z nhamon $";

	   /** Indicates if column index must be sent to server for double-clicks */
    public static boolean      USE_COLUMN_INDEX_FOR_DOUBLE_CLICK = false;
    
    /** Indicates if context menu must be used on row. */
    public static boolean      USE_CONTEXT_MENU = true;

    /** The row generic component. */
    protected LyRow            _row;

    /** Even row Color. */
    protected String           _rowEvenColor;

    /** Odd row Color. */
    protected String           _rowOddColor;

    /** Image for main action button. */
    protected String           _mainActionButton;

    /** Html code describing the alignment of the row. */
    protected String           _align;

    /** Alignment of the row. */
    protected LyHeader.Align   _headerAlign;

    /** Indicates if a sub-cell is being writed. */
    protected boolean          _writingSubCell;

    /** Index of row in the table. */
    protected int              _rowIndex;

    /** The parent form id. */
    protected String           _formId;
    
    /** Javascript row id (kept to avoid multiple encodings).*/
    protected String           _javascriptRowId;
    
	/**
	 * Class constructor of a Row object for the Struts target. Calls the super implementation
	 *
	 * @param row the generic Row defined for every type of view targets.
	 */
	public LyStrutsRow(LyRow row)
	{
        super(row);
        
        _row = row;
        _formId = null;
        _headerAlign = LyHeader.Align.LEFT;

        initialize();
	}
	
	/**
	 * Writes HTML description corresponding to this row.
	 *
	 * @param pw       the page writer
	 * @param rowIndex the index of this row
	 * @see   leon.view.web.LyWebComponent#writeHtml(LyWebPageWriter)
	 */
	public void writeHtml(LyWebPageWriter pw, int rowIndex)
	{
		writeHtml(pw, rowIndex, -1);
	}
	
	/**
	 * Writes HTML description corresponding to this row.
	 *
	 * @param pw               the page writer
	 * @param rowIndex         the index of this row
	 * @param rowIndexInParent the index of this row in this parent rows
	 * @see   leon.view.web.LyWebComponent#writeHtml(LyWebPageWriter)
	 */
	public void writeHtml(LyWebPageWriter pw, int rowIndex, int rowIndexInParent)
	{
        pw.startComponent(this);
        
        updateRefreshInfo(pw);
        _formId = pw.getActiveFormId();
        _rowIndex = rowIndex;

        LyRows       rows = (LyRows)_row.getParent();
        final LyTable      table = rows.getParentTable();
        LyHeaders    headers = table.getHeaders();
        LyHeader     header;
        LyStrutsTable htmlTable = (LyStrutsTable)table.getInternalView();
        boolean      addLastCell = htmlTable.addLastCell(pw);
        int          nCells = _row.getComponentCount();
        String       additionalStyle = getAdditionalStyle();
        String		 additionalStyleWithLeftEdition = additionalStyle;
        String		 additionalStyleWithRightEdition = additionalStyle;
        String       classType = getRowStyle(rowIndexInParent >= 0 ? rowIndexInParent : rowIndex);
        String       rowId = _row.getId();
        String       htmlRowId = this.getHtmlEntitiesString(rowId);
        String       tooltip = translate(_row.getTooltip());
        String       javascriptToolTip = LyString.getJavascriptString(tooltip);
        StringBuffer onClick = new StringBuffer();
        StringBuffer onDblClick = new StringBuffer();
        
        String       image = _row.getImage();
        boolean      drawImages = htmlTable.drawImages();
        boolean      sortOnStatus = htmlTable.sortOnStatus();
        String       rowColor = _row.getColor();
		
		if ((rowColor != null) && (rowColor.trim().length() == 0))
			rowColor = null;
		
        if ((additionalStyle != null) && (additionalStyle.length() > 0))
        {
			additionalStyleWithLeftEdition = " style=\"" + additionalStyle + ";background:url(" +
				getImageUrl("LY_EDITABLE_CELL_MARK") + ");background-repeat:no-repeat;background-position:left top;\"";
			additionalStyleWithRightEdition = " style=\"" + additionalStyle + ";background:url(" +
				getImageUrl("LY_EDITABLE_CELL_MARK") + ");background-repeat:no-repeat;background-position:right top;\"";
            additionalStyle = " style=\"" + additionalStyle + "\"";
        }
        else
        {
        	additionalStyleWithLeftEdition = " style=\"background:url(" +
				getImageUrl("LY_EDITABLE_CELL_MARK") + ");background-repeat:no-repeat;background-position:left top;\"";
        	additionalStyleWithRightEdition = " style=\"background:url(" +
				getImageUrl("LY_EDITABLE_CELL_MARK") + ");background-repeat:no-repeat;background-position:right top;\"";
        }

        String	rowIds = getRowIdsString(rowIndex, -1, false);
        
        if (pw.writeDivOnly())
        	rowIds = LyString.replace(rowIds, "\\n", "&#92;n");
        
        pw.append("\n<tr ");
        pw.append(rowIds);

        String colorCode = "LY_HIGHLIGHTED_BACKGROUND";
        String bgColor = "";

        if (colorCode != null)
            bgColor = getHexColor(colorCode);

        if (classType.equals("RE") || classType.equals("RO"))
        {
            String color = (classType.equals("RE") ? "LY_ROW_COLOR_EVEN" : "LY_ROW_COLOR_ODD");
            
            pw.append(" bgColor='");
            
            if (!_row.isDefaultColor() && (rowColor != null))
                pw.append(getHexColor(rowColor));
            else
                pw.append(getHexColor(color));
            
            pw.append("'");

            if (!pw.isPrintMode() && (rowColor == null))
            {
                pw.append(" onMouseOver=\"");
                pw.append(getMouseOverScript(pw, rowIndex, bgColor));
                pw.append("\" onMouseOut=\"");
                pw.append(getMouseOverScript(pw, rowIndex, getHexColor(color)));
                pw.append("\"");
            }
        }

        String  colorString = getColorString();
        
        if (colorString != null)
            pw.append(" color='" + colorString + "'");
        
        StringBuffer url = new StringBuffer();

        url.append("&#38;");
        url.append(LyWebConstants.VALUE);
        url.append("=");
        url.append(rowId);
        url.append("&#38;");
        url.append(LyWebConstants.INDEX);
        url.append("=" + rowIndex);

        if (table.getSelectionMode() != LyTable.SelectionMode.NONE)
        {
            onClick.append(" onClick=\"");
            onClick.append(getClickScript(pw, rowIndex));
            onClick.append("\"");
        }

        // If row has components, do not manage double-click on the row but on each cell.
        boolean     hasComponents = hasComponents(true);
        String      onDblClickScript = getDoubleClickScript(pw, rowIndex, -1);
		boolean   checkboxSelection = false ;//testResource("LY_DHTML_TABLE_ROW_SELECTION_BY_CHECKBOX");
    
		// Set the checkbox selection value
		if (table.getSelectionMode() == SelectionMode.SINGLE_CHECKBOX 
				|| table.getSelectionMode() == SelectionMode.MULTIPLE_CHECKBOX)
			checkboxSelection = true;
		
		if (!hasComponents)
		{
			if (!checkboxSelection)
				pw.append(onClick.toString());
        }
        else
        {
            StringBuffer    manageTabKeyBuffer = new StringBuffer();
            
            manageTabKeyBuffer.append("<script type='text/javascript'>\n//<![CDATA[\n");
            manageTabKeyBuffer.append("var frame = findTarget(null, '" + _lastFrameName + "', null);if (frame != null) frame.manageTabKey = true;");
            manageTabKeyBuffer.append("\n//]]>\n</script>");
            
            getWebViewManager().addAdditionalRefreshScripts(manageTabKeyBuffer.toString());
            pw.addBeginScript(manageTabKeyBuffer.toString(), false);
        }
        
        onDblClick.append(" onDblClick=\"");
        onDblClick.append(onDblClickScript);
        onDblClick.append("\"");
        
        if (!hasComponents && !USE_COLUMN_INDEX_FOR_DOUBLE_CLICK)
            pw.append(onDblClick.toString());
        
        if (USE_CONTEXT_MENU && table.getSelectionMode() != LyTable.SelectionMode.NONE)
        {
            pw.append(" onContextMenu=\"");
			pw.append(getContextMenuScript(pw, rowIndex, getHexColor(table.getSelectedColor())));
            pw.append("\"");
        }
        
        pw.append(">\n");

        // Compute height data for sub-cells (if needed)
        int              i, j;
        boolean          hasSubCells = hasSubCells();
        Vector<String>   levels = new Vector<String>();
        LyComponent      comp;
        _LyCellList[]    cellLists = null;
        int[]            totalHeight = null;
        int              maxCellCount = 1;
        int              rowSpanValue, maxRowSpanValue = 1;
        String           maxRowSpan;
        int              subTrCount = 0;
        int              colspan = nCells;
        String           subRowInput = null;

        if (hasSubCells)
        {
            // Compute the levels found in the row
            for (j = 0; j < nCells; j++)
            {
                comp = _row.getComponent(j);

                if (comp instanceof LyCell)
                    computeCellHeights((LyCell)comp, levels, null);
            }

            cellLists = new _LyCellList[nCells];
            totalHeight = new int[nCells];

            // For each cell, compute the list of sub-cells with their rowspan
            for (j = 0; j < nCells; j++)
            {
                comp = _row.getComponent(j);
                totalHeight[j] = 0;

                if (comp instanceof LyCell)
                {
                    cellLists[j] = new _LyCellList();
                    buildCellList(cellLists[j], (LyCell)comp, null, levels);
                    maxCellCount = Math.max(maxCellCount, cellLists[j].getCellCount());
                    maxRowSpanValue = Math.max(maxRowSpanValue, cellLists[j].getRowspan());
                }
            }
        }

        maxRowSpan = ((maxRowSpanValue > 1) ? (" rowspan='" + maxRowSpanValue + "'") : "");
        
        if (htmlTable.generateExternalBorder(pw))
        {
            pw.append("<td class='tableLeft' width='1'");
            pw.append(maxRowSpan);
            pw.append("></td>\n");
        }
        
        boolean     selectionInputAdded = false;

		// Add the checkbox to select a row
		if (checkboxSelection)
		{
			pw.append("<td class='status'");
			
			if (!headers.isDefaultColor())
			{
				pw.append(" style='background-image:none;");
				pw.append("background-color:" + getHexColor(headers.getColor()) + ";'");
			}
			
			pw.append(" width='1'><input type='checkbox' "
				+ getRowIdsString(rowIndex, -1, true)
				+ " "
                + onClick.toString());
            
            if (_row.isSelected())
                pw.append(" checked='yes'");
            
            pw.append("/>");
            
            if (!addLastCell)
            {
                // Add an hidden field if the row is selected
                if (addHiddenInputForSelection()
                        && (table.getSelectionMode() != LyTable.SelectionMode.NONE)
                        && (_row.isSelected()))
                    pw.append("<input type='hidden' name='" + LyWebConstants.SELECTION_ITEM
                        + htmlRowId + "' value='" + LyWebConstants.SELECTION_ITEM + htmlRowId
                        + "' />");
                selectionInputAdded = true;
            }
            
            pw.append("</td>");
        }
        
        // Adding delete button
        if (drawImages && !sortOnStatus)
        {
            if ((image != null) && (image.length() != 0))
            {
                int[] size;

                size = getImageSize(image, true);
                image = getImageUrl(image);

                pw.append("<td class='");
                pw.append(classType);
                // Witdh = 1% -> style is (RO|RE)
                pw.append("W'");
                pw.append(additionalStyle);
                pw.append(maxRowSpan);

                if (!pw.isPrintMode())
                {
                    pw.append(" style='cursor:hand;cursor:pointer;'");

                    pw.append(" onClick=\"");
                    pw.append(getStatusScript(pw));
                    pw.append("\"");
                }

                pw.append(" align='center'>");
                url.append("&#38;");
                url.append(LyWebConstants.SOURCE);
                url.append("=");
                url.append(LyWebConstants.STATUS);

                pw.append("<img id='" + _row.getIdCode() + "_image' src='");
                pw.append(image);
                pw.append("' border='0'");

                if (tooltip != null)
                {
                    pw.append(" title=\"");
                    pw.append(tooltip);
                    pw.append("\"");
                }

                if (size != null)
                {
                    if (size.length == 2)
                        pw.append(" width='" + size[0] + "' height='" + size[1] + "'");
                    else
                        pw.append(" height='" + size[0] + "'");
                }

                pw.append("/>");
                
                if (!addLastCell)
                {
                    // Add an hidden field if the row is selected
                    if (addHiddenInputForSelection()
                            && (table.getSelectionMode() != LyTable.SelectionMode.NONE)
                            && (_row.isSelected()))
                        pw.append("<input type='hidden' name='" + LyWebConstants.SELECTION_ITEM
                            + htmlRowId + "' value='" + LyWebConstants.SELECTION_ITEM + htmlRowId
                            + "' />");
                    
                    selectionInputAdded = true;
                }

                pw.append("</td>\n");
            }
            else
            {
                pw.append("<td class='");
                pw.append(classType);
                pw.append("'");
                pw.append(additionalStyle);
                pw.append(maxRowSpan);
                pw.append(">&nbsp;</td>\n");
            }
        }
        
        for (i = 0; i < maxRowSpanValue; i++)
        {
            for (j = 0; j < nCells; j++)
            {
                header = null;
                
                // Manage sort headers.
                if (headers != null)
                {
                    header = headers.getHeader(j);

                    if (header != null)
                    {
                        if ((header.getLabel() != null) && header.getLabel().equals("LY_STATUS_SORT"))
                        {
                            if (drawImages && (i == 0))
                            {
                                // String image = _row.getImage();

                                if ((image != null) && (image.length() != 0))
                                {
                                    int[] sizeImg = getImageSize(image, true);

                                    image = getImageUrl(image);

                                    pw.append("<td class='status'");
									
									if (!headers.isDefaultColor())
									{
										pw.append(" style='background-image:none;");
										pw.append("background-color:" + getHexColor(headers.getColor()) + ";'");
									}

									url.append("&#38;");
                                    url.append(LyWebConstants.SOURCE);
                                    url.append("=");
                                    url.append(LyWebConstants.STATUS);

                                    if (!pw.isPrintMode())
                                    {
                                        pw.append(" style='cursor:hand;cursor:pointer;'");
                                        pw.append(
                                            " onMouseDown='javascript:changeStyle(this,\"statusHighlighted\",\"statusPressed\");'");
                                        pw.append(
                                            " onMouseUp='javascript:changeStyle(this,\"statusHighlighted\",\"statusPressed\");'");

                                        pw.append(
                                            " onmouseout='javascript:changeStyle(this,\"status\",\"statusHighlighted\");\'");
                                        pw.append(
                                            " onmouseover='javascript:changeStyle(this,\"status\",\"statusHighlighted\");return true\'");

                                        pw.append(" onClick=\"");
                                        pw.append(getStatusScript(pw));
                                        pw.append("\"");
                                    }

                                    pw.append(maxRowSpan);
                                    pw.append(" align='center'>");
                                    pw.append("<img id='" + _row.getIdCode() + "_image' src='");
                                    pw.append(image);
                                    pw.append("' border='0'");

                                    if (tooltip != null)
                                    {
                                        pw.append(" title=\"");
                                        pw.append(tooltip);
                                        pw.append("\"");
                                    }

                                    if (sizeImg != null)
                                    {
                                        if (sizeImg.length == 2)
                                        {
                                            pw.append(" width='");
                                            pw.append(Integer.toString(sizeImg[0]));
                                            pw.append("' height='");
                                            pw.append(Integer.toString(sizeImg[1]));
                                            pw.append("'/>");
                                        }
                                        else
                                        {
                                            pw.append(" height='");
                                            pw.append(Integer.toString(sizeImg[0]));
                                            pw.append("'/>");
                                        }
                                    }
                                    else
                                        pw.append("/>");

                                    if (!pw.isPrintMode())
                                        pw.append("</a>");

                                    pw.append("</td>\n");
                                }
                                else
                                {
                                    pw.append("<td class='status'");

									if (!headers.isDefaultColor())
									{
										pw.append(" style='background-image:none;");
										pw.append("background-color:" + getHexColor(headers.getColor()) + ";'");
									}
									
									pw.append(maxRowSpan);
                                    pw.append(">&nbsp;</td>\n");
                                }

                                continue;
                            }
                        }

                        _align = LyStrutsTable.getHtmlHAligment(header.getAlign());
                        _headerAlign = header.getAlign();
                    }
                    else
                    {
                        _align = "";
                        _headerAlign = LyHeader.Align.LEFT;
                    }
                }
                else
                {
                    _align = "";
                    _headerAlign = LyHeader.Align.LEFT;
                }

                comp = _row.getComponent(j);
                LyWebComponent webComp = (LyWebComponent)comp.getInternalView();

                if ((comp instanceof LyFormField) && (i == 0))
                {
                    LyFormField  formField = (LyFormField)comp;
                    String       noWrap = " nowrap='true'";
                    boolean      editable = formField.isEditable();
                    
                    pw.append("<td ");
                    pw.append(_align);
                    pw.append(noWrap);
                    pw.append(maxRowSpan);
                    
                    if (editable)
                    {
                    	if (_headerAlign == LyHeader.Align.RIGHT)
                    		pw.append(additionalStyleWithLeftEdition);
                    	else
                    		pw.append(additionalStyleWithRightEdition);
                    }
                    else
                    	pw.append(additionalStyle);
                    
                    pw.append(" class='");
                    pw.append(classType);
                    pw.append("'");
                    pw.append(" id='cell_");
                    pw.append(_row.getIdCode() + "_" + rowIndex + "_" + j);
                    pw.append("'");
                    pw.append(" onClick=\"");
                    
                    if (hasComponents)
                        pw.append(getClickScript(pw, rowIndex));
                    
                    pw.append(getCellEditionScript(pw, rowIndex, j));
                    pw.append("\"");
                    pw.append(">");
                    
                	boolean		isIe = (getWebViewManager().getBrowser() == LyWebViewManager.BROWSER_MSIE);
                	
                    pw.append("<div style='position:relative;");

                    if (!isIe)
                		pw.append("width:100%;");

                    pw.append("'>");
                    pw.append("<div id='cell_edition_");
                    pw.append(_row.getIdCode() + "_" + j);
                    pw.append("' manageTabKey='true'");

                    if (!isIe)
                        pw.append(" style='float:" + _headerAlign + ";'");
                    
                    pw.append(">");

                    webComp.writeHtml(pw);

                    pw.append("</div>");
                    pw.append("</div>");
                    pw.append("</td>\n");
                }
                else if (comp instanceof LyCell)
                {
                    if (!hasSubCells)
                    {
                        String  extra = null;
                        
                        if (addHiddenInputForSelection()
                                && !addLastCell
                                && !selectionInputAdded
                                && (i == 0)
                                && (j == 0)
                                && (table.getSelectionMode() != LyTable.SelectionMode.NONE)
                                && (_row.isSelected()))
                        {
                            extra = "<input type='hidden' name='" + LyWebConstants.SELECTION_ITEM
                                    + htmlRowId + "' value='" + LyWebConstants.SELECTION_ITEM + htmlRowId
                                    + "' />";
                        }
                        
                        ((LyStrutsCell)webComp).writeHtml(pw, rowIndex, j, 1, extra);
                    }
                    else
                    {
                        if (cellLists[j] != null)
                        {
                            rowSpanValue = cellLists[j].getCellHeight(i);
                            comp = cellLists[j].getCell(i);

                            if (rowSpanValue > 0)
                            {
                                totalHeight[j] += rowSpanValue;
                                webComp = (LyStrutsCell)comp.getInternalView();

                                if (rowSpanValue < maxRowSpanValue)
                                    _writingSubCell = true;

                                ((LyStrutsCell)webComp).writeHtml(pw, rowIndex, j, rowSpanValue,
                                    subRowInput);
                                subRowInput = null;
                                _writingSubCell = false;
                            }
                            else if ((comp == null) && (totalHeight[j] < maxRowSpanValue))
                            {
                                pw.append("<td class='");
                                pw.append(classType);
                                pw.append("'");
                                pw.append(additionalStyle);
                                pw.append(" rowspan='" + (maxRowSpanValue - i) + "'");
                                
                                if (USE_COLUMN_INDEX_FOR_DOUBLE_CLICK)
                                {
                                    onDblClickScript = getDoubleClickScript(pw, rowIndex, j);
                                    onDblClick.append(" onDblClick=\"");
                                    onDblClick.append(onDblClickScript);
                                    onDblClick.append("\"");
                                }
                                
                                pw.append(">");
                                pw.append("</td>\n");
                                totalHeight[j] = maxRowSpanValue;
                            }
                        }
                    }
                }
                else if (i == 0)
                {
                    pw.append("<td ");
                    pw.append(_align);
                    pw.append(maxRowSpan);
                    pw.append(additionalStyle);
                    pw.append(" class='");
                    pw.append(classType);
                    pw.append("'>");
                    webComp.writeHtml(pw);
                    pw.append("</td>");
                }

                // Write the border if necessary
                if ((i == 0) && (header != null) && (header.getBorderWidth() > 0))
                {
                    pw.append("<td bgColor='" + getHexColor(header.getBorderColor())
                            + "'><div style='width:" + header.getBorderWidth() + "px;'></div></td>\n");
                    colspan++;
                }
            }

            if (i == 0)
            {
                // Section row
                if ((nCells == 0) && (_row.getLabel() != null))
                {
                    int tableColspan = ((headers != null) ? headers.getHeaderCount() : 0);
                    
                    if (sortOnStatus)
                        tableColspan--;
                    
                    if (addLastCell)
                        tableColspan++;
                    
                    if (drawImages)
                        pw.append("<td class='status'>&nbsp;</td>");
                    
                    pw.append("<td colspan='" + tableColspan);
                    pw.append("' class='");
                    pw.append(classType);
                    pw.append("'");
                    
                    if ((_row.getFont() != null) || (_row.getFontColor() != null))
                    {
                        pw.append(" style='");
                        pw.append(getFontString(_row.getFont(), _row.getFontColor()));
                        pw.append("'");
                    }
                    
                    pw.append(">");
                    
                    int    k, rowDepth = getDepthInTable();
                    String blankImage = getImageUrl("LY_BLANK_ROW");
                    String closedRowImage = getImageUrl("LY_CLOSED_ROW");
                    String openRowImage = getImageUrl("LY_OPEN_ROW");

                    // If row is a parent row, add an arrow to open/close the row
                    if (_row.getRows() != null)
                        rowDepth--;

                    for (k = 0; k< (rowDepth + 1); k++)
                        pw.append("<img border='0' align='bottom' src='" + blankImage + "'/>");

                    if (_row.getRows() != null)
                    {
                        boolean      isOpen = _row.isOpen();
                        String       expandRowScript = getExpandRowScript(pw);

                        if (isOpen)
                            javascriptToolTip = getJavascriptString("LY_CLOSE_ROW");
                        else
                            javascriptToolTip = getJavascriptString("LY_OPEN_ROW");

                        if (!pw.isPrintMode() && (expandRowScript != null))
                        {
                            pw.append("<a href=\"" + expandRowScript.toString());
                            pw.append("\" onmouseover=\'window.status=\"");
                            pw.append(javascriptToolTip);
                            pw.append("\";return true\'>");
                        }

                        pw.append("<img border='0' align='bottom' src='"
                            + (isOpen ? openRowImage : closedRowImage) + "'/>");

                        if (!pw.isPrintMode() && (expandRowScript != null))
                            pw.append("</a>");
                    }

                    pw.append("&nbsp;");                    
                    pw.append(getHtmlString(_row.getLabel()));
                    pw.append("</td>");
                }
                else if (addLastCell)
                {
                    // Add last cell
                    pw.append("<td class='");
                    pw.append(classType);
                    // Width = 100% -> (RE|RO)F
                    pw.append("F'");
                    pw.append(additionalStyle);
                    pw.append(maxRowSpan);
                    
                    if (hasComponents)
                        pw.append(onClick.toString());
                    
                    // If row has editable cells, each cell has to manage its own double-click
                    if (hasComponents)
                        pw.append(onDblClick.toString());

                    pw.append(">&nbsp;");
                    colspan++;
                }

                // Add an hidden field if the row is selected
                if (addHiddenInputForSelection()
                        && !selectionInputAdded
                        && (table.getSelectionMode() != LyTable.SelectionMode.NONE)
                        && (_row.isSelected()))
                    pw.append("<input type='hidden' name='" + LyWebConstants.SELECTION_ITEM
                        + htmlRowId + "' value='" + LyWebConstants.SELECTION_ITEM + htmlRowId
                        + "' />");
                
                if (addLastCell)
                    pw.append("</td>\n");
                
                if (htmlTable.generateExternalBorder(pw))
                {
                    pw.append("<td class='tableRight' width='1'");
                    pw.append(maxRowSpan);
                    pw.append("></td>\n");
                }
            }

            if (i < (maxRowSpanValue - 1))
            {
                pw.append("</tr><tr ");
                pw.append(getRowIdsString(rowIndex, subTrCount, false));
                
                if (classType.equals("RE") || classType.equals("RO"))
                {
                    String color = (classType.equals("RE") ? "LY_ROW_COLOR_EVEN" : "LY_ROW_COLOR_ODD");
                    
                    pw.append(" bgColor='");
                    
                    if (rowColor != null)
                        pw.append(getHexColor(rowColor));
                    else
                        pw.append(getHexColor(color));

                    pw.append("'");

                    if (!pw.isPrintMode())
                    {
                        pw.append(" onMouseOver=\"");
                        pw.append(getMouseOverScript(pw, rowIndex, bgColor));
                        pw.append("\" onMouseOut=\"");
                        pw.append(getMouseOverScript(pw, rowIndex, getHexColor(color)));
                        pw.append("\"");
                    }
                }
                
                if (!hasComponents)
                    pw.append(onClick.toString());
                
                if (!hasComponents)
                    pw.append(onDblClick.toString());
                
                pw.append(">\n");
                    
                StringBuffer subRowInputBuffer = new StringBuffer();

                if (addHiddenInputForSelection())
                {
                    subRowInputBuffer.append("<input type='hidden' name='");
                    subRowInputBuffer.append("SUB_" + subTrCount + LyWebConstants.SELECTION_ITEM);
                    subRowInputBuffer.append(htmlRowId);
                    subRowInputBuffer.append("' value='");
                    subRowInputBuffer.append("SUB_" + subTrCount + LyWebConstants.SELECTION_ITEM);
                    subRowInputBuffer.append(htmlRowId);
                    subRowInputBuffer.append("'/>");
                }

                subRowInput = subRowInputBuffer.toString();
                subTrCount++;
            }
        }

        pw.append("</tr>\n");
        
        if (_row.getBorderHeight() > 0)
        {
            pw.append("<tr><td height='" + _row.getBorderHeight() + "' colspan='" + colspan + "' bgColor='" + getHexColor(_row.getBorderColor())
                    + "'></td></tr>\n");
        }
		
		if (_row.isSelected())
		{
			StringBuffer	script = new StringBuffer();
			String			selectionId = _javascriptRowId;
			
	        if (!pw.writeDivOnly())
	        	selectionId = LyString.replace(selectionId, "&#92;", "\\");
			
			script.append("<script type='text/javascript'>\n//<![CDATA[\n");
			script.append("addSelectionId('table_" + table.getIdCode() + "', '" + selectionId + "');");
			script.append("\n//]]>\n</script>\n");

			if (pw.writeDivOnly())
				getWebViewManager().addAdditionalRefreshScripts(script.toString());
			else
				pw.addBeginScript(script.toString(), false);
		}
        
        pw.endComponent(this);
	}

	/**
	 * Returns the string used to identify the TR object.
	 * 
	 * @param rowIndex   the index of this row
	 * @param subrow     the index of the sub row (positive if id must be generated for a sub-row)
	 * @param isCheckbox indicates if identifier will be used by selection checkbox
	 * @return the string used to identify the TR object.
	 */
	protected String getRowIdsString(int rowIndex, int subrow, boolean isCheckbox)
	{
		String		 rowId = getHtmlEntitiesString(_row.getId());

		if (isCheckbox)
			return "id='checkboxSelection_row" + rowId + "' name='checkboxSelection_row" + rowId + "'";
		else if (subrow >= 0)
			return "id='" + rowId + "SUB_" + subrow + "'";
		else
			return "id='" + rowId + "'";
	}

	/**
	 * Gets the Javascript code used when mouse is over the row.
	 *
	 * @param pw       the page writer
	 * @param rowIndex the index of this row
	 * @param color    the color to be passed to the Javascript function
	 * @return the String describing the Javascript code when mouse is over the row
	 */
	public String getMouseOverScript(LyWebPageWriter pw, int rowIndex, String color)
	{
		StringBuffer	script = new StringBuffer();
		
		script.append("javascript:callOnMouseOver('");
		script.append(_javascriptRowId);
		script.append("','");
		script.append(color);
		script.append("');");
		
		return script.toString();
	}

	/**
	 * Gets the Javascript code used when clicking on the row.
	 *
	 * @param pw       the page writer
	 * @param rowIndex the index of this row
	 * @return the String describing the Javascript code when clicking on the row
	 */
	public String getClickScript(LyWebPageWriter pw, int rowIndex)
	{
		if (!_row.isSensitive())
			return "";
		
		// get the url to the selection handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    selectionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_SELECTEVENT);
		LyRows		 		rows = (LyRows)_row.getParent();
		LyTable		 		table = rows.getParentTable();
		String  			selectFontColor = table.getSelectedFontColor();

        selectFontColor = getHexColor("LY_TABLE_SELECTED_FONT_COLOR");
		// Leave this comment until LyComponent.getStringValue(String, String) method is commented
        // selectFontColor = getHexColor(selectFontColor);
        		
		String				selectColor = getHexColor("LY_TABLE_SELECT_COLOR");
		StringBuffer	    script = new StringBuffer();

		script.append("InvokeRowClick('");
		script.append(_javascriptRowId);
		script.append("','");
		script.append(table.getRootController());
		script.append("','table_");
		script.append(table.getIdCode());
		script.append("','");
		script.append(selectColor);
		script.append("','");
		script.append(selectionUrl);
		script.append("',");
		script.append(table.getSelectionMode() == LyTable.SelectionMode.SINGLE);
		script.append(",event,'");
		script.append(selectFontColor);
		script.append("');");
		
		return script.toString();
	}

	/**
	 * Gets the Javascript code used when double-clicking on the row.
	 *
	 * @param pw       the page writer
	 * @param rowIndex the index of this row
	 * @param columnIndex the index of the column
	 * @return the String describing the Javascript code when double-clicking on the row
	 */
	public String getDoubleClickScript(LyWebPageWriter pw, int rowIndex, int columnIndex)
	{
		if (!_row.isSensitive())
			return "";
		
		StringBuffer	script = new StringBuffer();
		LyRows		 	rows = (LyRows)_row.getParent();
		LyTable		 	table = rows.getParentTable();
		String			selectColor = getHexColor("LY_TABLE_SELECT_COLOR");

		// Retrieve action URL
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		
		script.append("javascript:InvokeRowDoubleClick('");
		script.append(table.getId());
		script.append("','");
		script.append(table.getRootController());
		script.append("','");
		script.append(_javascriptRowId);
		script.append("','");
		script.append(selectColor);
		script.append("',");
		script.append(columnIndex);
		script.append(",'");
		script.append(LyWebConstants.SUB_INDEX_SEPARATOR);
		script.append("','");
		script.append(actionUrl);
		script.append("')");
		
		return script.toString();
	}

	/**
	 * Gets the Javascript code used when clicking on the status of the row.
	 *
	 * @param pw       the page writer
	 * @return the String describing the Javascript code when clicking on the status of the row
	 */
	public String getStatusScript(LyWebPageWriter pw)
	{
		StringBuffer	script = new StringBuffer();
		LyRows		 	rows = (LyRows)_row.getParent();
		LyTable		 	table = rows.getParentTable();

		// Retrieve action URL
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);

		script.append("javascript:InvokeStatusClick('");
		script.append(_javascriptRowId);
		script.append("','");
		script.append(table.getRootController());
		script.append("','");
		script.append(actionUrl);
		script.append("',event);");
		
		return script.toString();
	}

	/**
	 * Gets the expand row action and writes associated javascript function for this table.
	 *
	 * @param  pw the page writer
	 * @return the expand row action javascript tag
	 */
	public String getExpandRowScript(LyWebPageWriter pw)
	{
		StringBuffer	script = new StringBuffer();
		LyRows		 	rows = (LyRows)_row.getParent();
		LyTable		 	table = rows.getParentTable();
		String			rowId = _row.getId();
        String			javascriptRowId = LyString.getJavascriptString(rowId);

        javascriptRowId = LyString.replace(javascriptRowId, "&#39;", "\\&#39;");

		// Retrieve action URL
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);

		script.append("javascript:InvokeExpandRowClick('");
		script.append(javascriptRowId);
		script.append("','");
		script.append(table.getRootController());
		script.append("','");
		script.append(actionUrl);
		script.append("');");
		
		return script.toString();
	} // end method getExpandRowAction

	/**
	 * Gets the Javascript code used when a context menu is performed.
	 *
	 * @param pw          the page writer
	 * @param rowIndex    the index of this row
	 * @param selectColor the color to be passed to the Javascript function
	 * @return the String describing the Javascript code when a context menu is performed.
	 */
	public String getContextMenuScript(LyWebPageWriter pw, int rowIndex, String selectColor)
	{
		if (!_row.isSensitive())
			return "";
		
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		String			    selectionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_SELECTEVENT);
		LyRows				rows = (LyRows)_row.getParent();
		LyTable				table = rows.getParentTable();
		String    		    selectFontColor = table.getSelectedFontColor();
		StringBuffer		script = new StringBuffer();

        selectFontColor = getHexColor("LY_TABLE_SELECTED_FONT_COLOR");
		// Leave this comment until LyComponent.getStringValue(String, String) method is commented
        // selectFontColor = getHexColor(selectFontColor);
        		
		script.append("callOnContextMenu('");
		script.append(_javascriptRowId);
		script.append("','");
		script.append(table.getRootController());
		script.append("','table_");
		script.append(table.getIdCode());
		script.append("','");
		script.append(table.getId());
		script.append("','");
		script.append(selectColor);
		script.append("','");
		script.append(actionUrl);
		script.append("','");
		script.append(selectionUrl);
		script.append("',event,'" + selectFontColor + "');return false;");
		
		return script.toString();
	}

	/**
	 * Gets the Javascript code used when double-clicking on an editable cell.
	 *
	 * @param pw the page writer
	 * @param rowIndex the index of this row
	 * @param columnIndex the index of the cell column
	 * @return the String describing the Javascript code when double-clicking on the row
	 */
	public String getCellEditionScript(LyWebPageWriter pw, int rowIndex, int columnIndex)
	{
		StringBuffer	script = new StringBuffer();
		LyRows		 	rows = (LyRows)_row.getParent();
		LyTable		 	table = rows.getParentTable();

		// Retrieve action URL
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		
		script.append("startCellEdition('");
		script.append(table.getId());
		script.append("','");
		script.append(_row.getRootController());
		script.append("',");
		script.append(rowIndex);
		script.append(",");
		script.append(columnIndex);
		script.append(",'");
		script.append(LyTools.PIPE_SEP);
		script.append("','");
		script.append(actionUrl);
		script.append("',event);");
		
		return script.toString();
	}

	/**
	 * Indicates whether a hidden input must be generated in the HTML page for selection.
	 * 
	 * @return default implementation returns false.
	 */
	protected boolean addHiddenInputForSelection()
	{
		return false;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.view.setinterface.LySetColorInterface#setColor(java.lang.String)
	 */
	public void setColor(String key)
	{
		String			 rowColor = ((key == null) ? "" : getHexColor(key));
		LyWebViewManager viewManager = getWebViewManager();
		StringBuffer     buffer = new StringBuffer();

		// Update the color of the row
		buffer.append("<script type='text/javascript'>\n//<![CDATA[\nSetRowColor('").append(
			_javascriptRowId).append("','").append(rowColor).append("','").append(_lastFrameName).append(
			"', '").append(_rowEvenColor).append("', '").append(_rowOddColor).append("');\n//]]>\n</script>");

		// Update selection status if row is selected
		if (_row.isSelected())
		{
			LyRows		rows = (LyRows)_row.getParent();
			LyTable		table = rows.getParentTable();
			String		selectColor = getHexColor("LY_TABLE_SELECT_COLOR");

			buffer.append("<script type='text/javascript'>\n//<![CDATA[\nselectRow('table_").append(table.getIdCode())
				.append("','").append(_javascriptRowId).append("','").append(selectColor).append(
				"','").append(_lastFrameName).append("','false');\n//]]>\n</script>");
		}

		viewManager.addAdditionalRefreshScripts(buffer.toString());
	}
	
	   /**
     * Initialization method.<br>
     * Sets row style, additional style, row index and the sub cell writing status to default
     * values.<br>
     * Sets colors and button image with default values defined in resources.
     */
    public void initialize()
    {
        _writingSubCell = false;
        _rowIndex = -1;

        _rowEvenColor = getHexColor("LY_ROW_COLOR_EVEN");
        _rowOddColor = getHexColor("LY_ROW_COLOR_ODD");
        
        _javascriptRowId = LyString.getJavascriptString(_row.getId());
        _javascriptRowId = LyString.replace(_javascriptRowId, "&#39;", "\\'");
        _javascriptRowId = LyString.replace(_javascriptRowId, "\\\\n", "&#92;&#92;n");
    }
    
    /*
     * (non-Javadoc)
     *
     * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
     */
    public void writeHtml(LyWebPageWriter pw)
    {
    }

    /**
     * Returns the color string to be used for the TR element.
     * 
     * @return the color code or null if no specific color must be used.
     */
    protected String getColorString()
    {
        return null;
    }
    
    /**
     * Indicates if current row contains at least one cell with sub-cells.
     *
     * @return true if contains sub-cells, false otherwise
     */
    protected boolean hasSubCells()
    {
        int         i, n = _row.getComponentCount();
        LyComponent comp;

        for (i = 0; i < n; i++)
        {
            comp = _row.getComponent(i);

            if ((comp instanceof LyCell) && (((LyCell)comp).getCellCount() > 0))
                return true;
        }

        return false;
    }

    /**
     * Indicates if current row contains at least one component (field,
     * editable cell or toolbar).
     *
     * @param testToolbar indicates if we must test the presence of a toolbar
     * @return true if contains at least one component, false otherwise
     */
    public boolean hasComponents(boolean testToolbar)
    {
        int         i, n = _row.getComponentCount();
        LyComponent comp;

        for (i = 0; i < n; i++)
        {
            comp = _row.getComponent(i);

            if (((comp instanceof LyCell) && ((LyCell)comp).isEditable())
                    || (testToolbar && (comp instanceof LyToolBar))
                    || (comp instanceof LyFormField))
                return true;
        }

        return false;
    }

    /**
     * Computes the height of given cell and its sub-cells and add it to the list of heights with
     * their level.
     *
     * @param cell   the cell to compute
     * @param levels the list of levels of the given cell and its sub-cells
     * @param level  the level of the cell
     */
    protected void computeCellHeights(LyCell cell, Vector<String> levels, String level)
    {
        if (level == null)
            level = "";

        if (!levels.contains(level))
            levels.addElement(level);

        int    i, n = cell.getCellCount();
        LyCell subCell;

        for (i = 0; i < n; i++)
        {
            subCell = cell.getCell(i);
            computeCellHeights(subCell, levels, level + LyTools.PIPE_SEP + i);
        }
    }

    /**
     * Computes the height of the given level.
     *
     * @param  level  the level
     * @param  levels the list of levels and sub-levels (build in {@link
     *                LyStrutsRow#computeCellHeights(LyCell, Vector, String)})
     * @return the height of the level
     */
    protected int computeLevelHeight(String level, Vector<String> levels)
    {
        if (level == null)
            level = "";

        int    i, n = levels.size();
        String subLevel, subLevelTmp;
        int    pos;
        int    height = 0;

        for (i = 0; i < n; i++)
        {
            subLevel = levels.elementAt(i);
            pos = subLevel.lastIndexOf(LyTools.PIPE_SEP);

            if (pos >= 0)
            {
                subLevelTmp = subLevel.substring(0, pos);

                if (level.equals(subLevelTmp))
                    height += computeLevelHeight(subLevel, levels);
            }
        }

        return ((height == 0) ? 1 : height);
    }

    /**
     * Builds the sub-cells list of the given cell and associate them to their height.
     *
     * @param cellLists the sub-cells list
     * @param cell      the given cell
     * @param level     the level of the cell
     * @param levels    the list of levels of the given cell and its sub-cells
     */
    protected void buildCellList(_LyCellList cellLists, LyCell cell, String level,
        Vector<String> levels)
    {
        if (level == null)
            level = "";

        int    i, n = cell.getCellCount();
        LyCell subCell;

        if (n == 0)
            cellLists.addCell(cell, computeLevelHeight(level, levels));

        for (i = 0; i < n; i++)
        {
            subCell = cell.getCell(i);
            buildCellList(cellLists, subCell, level + LyTools.PIPE_SEP + i, levels);
        }
    }

    /**
     * Computes an additional style for a row.
     *
     * @return a string containing the CSS additional style
     */
    public String getAdditionalStyle()
	{
		String style = "";
		String linesStyle = getLinesAdditionalStyle();
		
		if (linesStyle != null)
			style += linesStyle;
		
		if (!_row.isDefaultColor())
			style += "background-color:" + getHexColor(_row.getColor()) + ";";
		
		if (!_row.isDefaultFontColor() || !_row.isDefaultFont())
			style += getFontString(_row.getFont(), _row.getFontColor());
	
		return style;
	}

	/**
	 * Computes an additional style for row lines.
	 *
	 * @return a string containing the CSS additional style for row lines
	 */
	public String getLinesAdditionalStyle()
	{
		String		style = "";
		LyRows		rows = (LyRows)_row.getParent();
		LyTable		table = rows.getParentTable();
		
		if (!table.isHLines())
			style += "border-top-width:0px;border-bottom-width:0px;";
		
		if (!table.isVLines())
			style += "border-left-width:0px;border-right-width:0px;";
		
		if (style.length() > 0)
			return style;
		
		return null;
	}

	/**
	 * Gets the row style.
     *
     * @param  rowIndex the index of the row
     * @return if the row is an even row, return "RE"<br />
     *         else if the row is an odd row, return "RO"<br/>
     *         if a sub-cell is written, add the prefix "SUB".
     */
    public String getRowStyle(int rowIndex)
    {
        String result = "";

        if ((rowIndex % 2) == 0)
            result = "RE";
        else
            result = "RO";

        if (_writingSubCell)
            result = "SUB" + result;

        return result;
    }

    /**
     * Gets the HTML CSS style piece of tag for the background color of the row.
     *
     * @return the String describing the background color of the row
     */
    public String getBackgroundColorStyle()
    {
        String result = null;

        if ((_row.getColor() != null) && (_row.getColor().trim().length() > 0))
        {
            String rowColor = getHexColor(_row.getColor());

            result = "background-color:" + rowColor + ";";
        }
        return result;
    }
    
    /**
     * Gets the alignment of the row.
     *
     * @return the row alignment
     */
    public String getAlignment()
    {
        return _align;
    }
    
    /**
     * Gets the header alignment of the row.
     *
     * @return the header alignment
     */
    public LyHeader.Align getHeaderAlignment()
    {
        return _headerAlign;
    }

    /**
     * Gets the depth of current row in the table.
     *
     * @return the depth of current row in the table if hierarchical, 0 otherwise.
     */
    protected int getDepthInTable()
    {
        LyRows  rows = (LyRows)_row.getParent();
        LyTable table = rows.getParentTable();
        LyRows  rootRows = table.getRows();
        LyRow   row;
        int     depth = 0;

        while (rows != rootRows)
        {
            depth++;
            row = (LyRow)rows.getParent();
            rows = (LyRows)row.getParent();
        }

        return depth;
    }

    /*
     * (non-Javadoc)
     *
     * @see leon.view.setinterface.LySetImageInterface#setImage(java.lang.String)
     */
    public void setImage(String key)
    {
        LyWebViewManager viewManager = getWebViewManager();

        // Manage Ajax mode
        StringBuffer     buffer = new StringBuffer();

        if ((key != null) && (key.length() != 0))
        {
            key = getImageUrl(key);

            // Update the image
            buffer.append("<script type='text/javascript'>SetImageSrc('").append(_row.getIdCode())
                .append("_image','").append(key).append("');</script>");

            viewManager.addAdditionalRefreshScripts(buffer.toString());
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see leon.view.graphic.LyGraphicalRow#setTooltip(java.lang.String)
     */
    public void setTooltip(String tooltip)
    {
        LyWebViewManager viewManager = getWebViewManager();

        // Manage Ajax mode
        StringBuffer     buffer = new StringBuffer();

        tooltip = translate(tooltip);
        tooltip = LyString.getJavascriptString(tooltip);

        // Update the tooltip of the image
        buffer.append("<script type='text/javascript'>\n//<![CDATA[\nSetImageTooltip('").append(_row
                .getIdCode()).append("_image','").append(tooltip).append("');\n//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(buffer.toString());
    }

    /*
     * (non-Javadoc)
     *
     * @see leon.view.graphic.LyGraphicalRow#setOpen(boolean)
     */
    public void setOpen(boolean open)
    {
    	LyComponent parent = _row;

        if(parent == null)
        {
            super.refresh();
            return;
        }
    	
        while (!(parent instanceof LyTable))
            parent = parent.getParent();

        LyTable table = (LyTable)parent;
        LyWebObject webTable = (LyWebObject)table.getInternalView();
        webTable.refresh();
    }

    /*
     * (non-Javadoc)
     *
     * @see leon.view.setinterface.LySetHiddenInterface#setHidden(boolean)
     */
    public void setHidden(boolean hidden)
    {
        refresh();
    }
    
    /*
     * (non-Javadoc)
     *
     * @see leon.view.graphic.LyGraphicalRow#setBorderHeight(int)
     */
    public void setBorderHeight(int height)
    {
        refresh();
    }
    
    /*
     * (non-Javadoc)
     *
     * @see leon.view.graphic.LyGraphicalRow#setBorderColor(String)
     */
    public void setBorderColor(String color)
    {
        refresh();
    }
    
    /*
     * (non-Javadoc)
     *
     * @see leon.view.setinterface.LySetColorInterface#setColor(java.lang.String)
     */
    public void setLabel(String label)
    {
        refresh();
    }
    
	
	/**
	 * @see leon.view.setinterface.LyGraphicalRow#setIndex(java.lang.String)
	 */
	public void setIndex(String index)
	{
		refresh();
	}

    /*
     * (non-Javadoc)
     * 
     * @see leon.view.graphic.LyGraphicalRow#addComponent(LyComponent, int)
     */
    public void addComponent(LyComponent component, int pos)
    {
        // Refresh cell content
        LyWebViewManager    viewManager = getWebViewManager();
        StringBuffer        buffer = new StringBuffer();
        LyWebObject         webObject = (LyWebObject)component.getInternalView();
        LyWebPageWriter     pw = new LyStrutsPage(_row, true);
        boolean             resize = false;

        pw.setActiveBuffer(_lastFrameName);
        pw.setFormId(_formId);
        
        if (component instanceof LyCell)
        {
            LyCell          cell = (LyCell)component;
            LyStrutsCell    webCell = (LyStrutsCell)webObject;
            LyTable         table = webCell.getParentTable();
            LyStrutsTable   webTable = (LyStrutsTable)table.getInternalView();
            
            webCell.writeCellContent(pw, pos);
            
            // Update the HTML cell color
            String          cellColor = (cell.getColor() == null ? "" : getHexColor(cell.getColor()));
            StringBuffer    colorBuffer = new StringBuffer();
            String          cellId = "cell_" + _row.getIdCode() + "_" + _rowIndex + "_" + pos;
            
            if (cellColor == null)
            	cellColor = "";
            
            colorBuffer.append("<script type='text/javascript'>\n//<![CDATA[\nSetObjectColor('").append(
                _lastFrameName).append("','").append(cellId).append("','").append(cellColor).append(
                "');\n//]]>\n</script>");

            // Force refresh of line color
            String			javascriptRowId = LyString.getJavascriptString(_row.getId());

            colorBuffer.append("<script type='text/javascript'>\n//<![CDATA[\ngetLeonardiTopFrame().rowIdToRefresh = '" + javascriptRowId + "';\n//]]>\n</script>");
                
            viewManager.addAdditionalRefreshScripts(colorBuffer.toString());

            // Add script used to update rows color
            StringBuffer    updateBuffer = new StringBuffer();
            
            updateBuffer.append("<script type='text/javascript'>\n//<![CDATA[\n");
            updateBuffer.append(webTable.getUpdateRowsColorsScript());
            updateBuffer.append("\n//]]>\n</script>");
            
            viewManager.addAdditionalRefreshScripts(updateBuffer.toString());
        }
        else
        {
            webObject.writeHtml(pw);
            resize = true;
        }
        
        StringBuffer innerHtml = pw.getBuffer(pw.getActiveName());
        String       innerHtmlStr = LyString.replace(innerHtml.toString(), "\n", "\\n");

        innerHtmlStr = LyString.replace(innerHtmlStr, "\\", "\\\\");
        innerHtmlStr = LyString.replace(innerHtmlStr, "\\\\n", "\\n");
        innerHtmlStr = LyString.replace(innerHtmlStr, "\"", "\\\"");
        innerHtmlStr = LyString.replace(innerHtmlStr, "\\'", "\\\\'");

        buffer.append("<script type='text/javascript'>\n//<![CDATA[\nreplaceDivInnerHtml('")
        .append(_lastFrameName).append("','cell_edition_").append(_row.getIdCode()).append("_").append(pos).append("',\"")
        .append(innerHtmlStr).append("\");");
        buffer.append("SetDivAttribute('")
        .append(_lastFrameName).append("','cell_edition_").append(_row.getIdCode()).append("_").append(pos).append("','manageTabKey','true');");
        
        if (resize)
        {
            if (LyWebConstants.POPUP_VIEW.equals(_lastWindowName))
                buffer.append("checkSizePopupView(null, false);");
            else
                buffer.append("checkSizeMainView(null);");
        }
        
        buffer.append("\n//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(buffer.toString());
        pw.free();
        
        // Try to give the focus to the new component
        if (webObject instanceof LyStrutsFormField)
            ((LyStrutsFormField)webObject).setFocus();
    }

    /*
     * (non-Javadoc)
     * 
     * @see leon.view.graphic.LyGraphicalRow#removeComponent(LyComponent)
     */
    public void removeComponent(LyComponent components)
    {
    }

    /*
     * Internal class
     * -----------------------------------------------------------------------
     */

    /**
     * Internal class describing sub-cells of a cell with their height.
     */
    protected class _LyCellList
    {
        /** The list of sub-cells. */
        Vector<LyCell> _cells;

        /** The list of height of sub-cells. */
        Vector<Integer> _heights;

        /** The total height of the cell. */
        int    _totalHeight;

        /**
         * Class constructor of a Cell list.
         */
        public _LyCellList()
        {
            _cells = null;
            _heights = null;
            _totalHeight = 0;
        }

        /**
         * Adds a cell to the list of sub-cells.
         *
         * @param cell   the cell to add
         * @param height the heigt of the cell to add
         */
        public void addCell(LyCell cell, int height)
        {
            if (_cells == null)
            {
                _cells = new Vector<LyCell>();
                _heights = new Vector<Integer>();
            }

            _cells.addElement(cell);
            _heights.addElement(new Integer(height));
            _totalHeight += height;
        }

        /**
         * Gets the number of sub-cells.
         *
         * @return the number of sub-cells
         */
        public int getCellCount()
        {
            return _cells.size();
        }

        /**
         * Gets the sub-cell at the given index.
         *
         * @param  index the index of the wanted cell
         * @return the cell at given index
         */
        public LyCell getCell(int index)
        {
            if (index >= _totalHeight)
                return null;

            int pos = 0;
            int height = 0;

            while (height < index)
            {
                height += _heights.elementAt(pos).intValue();
                pos++;
            }

            if (height > index)
                pos--;

            if (pos >= _cells.size())
                return null;

            return _cells.elementAt(pos);
        }

        /**
         * Gets the height of the cell at the given index.
         *
         * @param  index the index of the wanted height
         * @return the height at the given index
         */
        public int getCellHeight(int index)
        {
            if (index >= _totalHeight)
                return -1;

            int pos = 0;
            int height = 0;

            while (height < index)
            {
                height += _heights.elementAt(pos).intValue();
                pos++;
            }

            if (height > index)
                return -1;

            return _heights.elementAt(pos).intValue();
        }

        /**
         * Gets the row span of the cell. It is the sum of all heights in the height list.
         *
         * @return the row span
         */
        public int getRowspan()
        {
            int     i, n = _heights.size();
            Integer integer;
            int     result = 0;

            for (i = 0; i < n; i++)
            {
                integer = _heights.elementAt(i);
                result += integer.intValue();
            }

            return result;
        }
    } // end class _LyCellList
} // end class LyStrutsRow
