/*
 * LyStrutsSelectBar.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.view.web.struts;

import leon.view.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.web.*;


/**
 * This class implements the Select Bar object (as described in the dtd) for struts viewer.
 *
 * <p><b>SELECTBAR: Navigation Bar</b><br>
 * A navigation bar allows displaying a set of page numbers (or page names) in order to go directly
 * to the selected page. This bar must be associated to an area presenting paginated data (a long
 * list for example).</p>
 *
 * @version $Id: LyStrutsSelectBar.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria S.A.
 */
public class LyStrutsSelectBar extends LyStrutsPanel
    implements LyGraphicalSelectBar,
               LyGetBoundsInterface
{
	public static final String __VERSION =
		"$Id: LyStrutsSelectBar.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** Constant used to generate CSS class name for border. */
    public static int          DECO_BORDER_PARAMETER = 0;

    /** Constant used to generate CSS class name for select bar. */
    public static int          DECO_SELECTBAR_PARAMETER = 1;

    /** Constant used to generate CSS class name for select item. */
    public static int          DECO_SELECTITEM_PARAMETER = 2;

    /** The select bar generic component. */
    protected LySelectBar      _selectBar;

    /** The select bar background color. */
    protected String           _barColor;
    
    
	/**
	 * Constructor.
	 *
	 * @param selectBar the select bar component
	 */
	public LyStrutsSelectBar(LySelectBar selectBar)
	{
        super(selectBar);
        _selectBar = selectBar;

        initialize();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.LyStrutsSelectBar#getBaseUrl(leon.view.web.LyWebPageWriter)
	 */
	protected String getBaseUrl(LyWebPageWriter pw)
	{
		// get the application context url.
		return ((LyWebViewManager)getViewManager()).getRootURL();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.LyStrutsSelectBar#buildToolPath(java.lang.String,
	 *      leon.view.web.LyWebPageWriter)
	 */
	public String buildToolPath(String toolPath, LyWebPageWriter pw)
	{
		String			    targetWindow = pw.getWindowName();

		// get the url to the event handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);

		// build the URL
		StringBuffer	    toolUrl = new StringBuffer();

		toolUrl.append(actionUrl);
		toolUrl.append("?");
		toolUrl.append(LyWebConstants.CONTROLLER);
		toolUrl.append("=");
		toolUrl.append(_selectBar.getController());

		if (targetWindow != null)
		{
			toolUrl.append("&#38;");
			toolUrl.append(LyWebConstants.TARGET_WINDOW);
			toolUrl.append("=");
			toolUrl.append(targetWindow);
		}

		if (pw.getLastReloadTime() > 0)
		{
			// Fix navigator cache bug : no access to server when URL is already
			// in cache
			toolUrl.append("&#38;");
			toolUrl.append(LyWebConstants.RELOAD);
			toolUrl.append("=");
			toolUrl.append(Long.toString(pw.getLastReloadTime()));
		}

		toolUrl.append("&#38;");
		toolUrl.append(LyWebConstants.SOURCE);
		toolUrl.append("=");
		toolUrl.append(toolPath);

		return "javascript:InvokeAction('" + toolPath + "','" + toolUrl.toString() + "','"
			+ LyWebConstants.RELOAD_FRAMES + "','" + _selectBar.getRootController() + "');";
	} // end method buildToolPath
	
	   /**
     * Initialization method.<br>
     * Sets the select bar backgroung color.
     */
    public void initialize()
    {
        _barColor = getHexColor("LY_SELECTBAR_BACKGROUND");
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
     */
    public void writeHtml(LyWebPageWriter pw)
    {
        pw.startComponent(this);
        
        if (!pw.writeDivOnly())
            pw.append("<div id=\"" + "_" + _selectBar.getIdCode() + "\">\n");

        updateRefreshInfo(pw);

        // Don't print selectbars
        if (pw.isPrintMode())
        {
            pw.endComponent(this);
            return;
        }

        // First add the needed script
        if (!pw.writeDivOnly())
        {
            String rootUrl = getBaseUrl(pw);

            // Add needed scripts
            pw.addBeginScript("<script type='text/javascript'>window.defaultStatus='';</script>",
                false);

            // Add var for rootTableUrl
            pw.addBeginScript("<script type='text/javascript'>var _rootUrl='"
                + LyWebObject.encodeUrl(rootUrl) + "';</script>", false);
            pw.addBeginScript("<script type='text/javascript'>var selectionContext='true'</script>",
                false);
            pw.addBeginScript("<script type='text/javascript'>var _controllerId='"
                + _selectBar.getController() + "';</script>", false);
        }

        LyWebDecoration deco = getWebViewManager().getDecoration();

        deco.writePageDecoration(pw, LyWebDecoration.WINDOW_BEGIN);
        deco.writePageDecoration(pw, LyWebDecoration.WINDOW_TOP);
        deco.writePageDecoration(pw, LyWebDecoration.WINDOW_LEFT);

        pw.append("<table class='" + deco.getClassName(_selectBar, DECO_BORDER_PARAMETER, pw) +
            "' width='100%' height='100%' border='0' cellpadding='0' cellspacing='0'><tr><td>\n");
        pw.append("<table class='" + deco.getClassName(_selectBar, DECO_SELECTBAR_PARAMETER, pw) +
            "' border='0' cellpadding='0' cellspacing='0' height='100%' width='100%'>\n<tr><td valign='bottom'  align='center'>");
        pw.append("<table border='0' cellpadding='0' cellspacing='0'>\n");

        LySelectItem   item;
        LyWebComponent itemComp;
        int            selectedIndex = -1;
        int            i, n = _selectBar.getComponentCount();

        for (i = 0; (i < n) && (selectedIndex < 0); i++)
        {
            item = (LySelectItem)_selectBar.getComponent(i);
            if (item.isSelected())
                selectedIndex = i; // Exits loop we have the answer
        }

        // Now add the list of items
        int start = _selectBar.getStartItem();

        pw.append("<tr>\n<td align='right' class='" + deco.getClassName(_selectBar, DECO_SELECTITEM_PARAMETER, pw) +
            "' nowrap='true'>");
        pw.append(getMessage("LY_PAGE_INDEX",
                new String[] { Integer.toString(selectedIndex + 1) }));
        pw.append("</td>");

        // If we do not start at the begining : add navigation tool (previous)
        if (start > 0)
        {
            // First
            pw.append("  <td class='" + deco.getClassName(_selectBar, DECO_SELECTITEM_PARAMETER, pw) +
                "' valign='middle'>\n");
            pw.append("\n<a href=\"");
            pw.append(buildToolPath("_select_page_first", pw));
            pw.append("\" title='");
            pw.append(translate("LY_FIRST"));
            pw.append("'><img src='");
            pw.append(getImageUrl("LY_PAGE_FIRST"));
            pw.append("' border='0'></a></td>\n");

            // Previous
            pw.append("  <td class='" + deco.getClassName(_selectBar, DECO_SELECTITEM_PARAMETER, pw) +
                "' valign='middle'>\n");
            pw.append("\n<a href=\"");
            pw.append(buildToolPath("_select_page_previous", pw));
            pw.append("\" title='");
            pw.append(translate("LY_PREVIOUS"));
            pw.append("'><img src='");
            pw.append(getImageUrl("LY_PAGE_PREVIOUS"));
            pw.append("' border='0'></a></td>\n");
        }

        if (_selectBar.getComponentCount() > 1)
        {
            // Show previous
            pw.append("  <td class='" + deco.getClassName(_selectBar, DECO_SELECTITEM_PARAMETER, pw) +
                "' valign='middle'>\n");
            pw.append("\n<a href=\"");
            pw.append(buildToolPath("_select_page_showprevious", pw));
            pw.append("\" title='");
            pw.append(translate("LY_SHOW_PREVIOUS"));
            pw.append("'><img src='");
            pw.append(getImageUrl("LY_PAGE_SHOW_PREVIOUS"));
            pw.append("' border='0'/></a></td>\n");
        }

        // Add page numbers and hrefs
        n = Math.min(_selectBar.getComponentCount(), start + _selectBar.getItemCount());

        pw.append("<td>&nbsp;</td>");

        for (i = start; i < n; i++)
        {
            item = (LySelectItem)_selectBar.getComponent(i);
            pw.append("  <td class='" + deco.getClassName(_selectBar, DECO_SELECTITEM_PARAMETER, pw) +
            "' valign='middle' align='center'>\n");
            
            if ((item == null)
                || (item.getId() == LyComponent.SEP_ID))
            {
                pw.append("<img src='" + getImageUrl("LY_VERTICAL_SEPARATOR") + "' border='0'/>");
            }
            else
            {
                itemComp = (LyWebComponent)item.getInternalView();
                itemComp.writeHtml(pw);
            }
            pw.append("\n  </td>\n");
        }

        pw.append("<td>&nbsp;</td>");

        if (_selectBar.getComponentCount() > 1)
        {
            // Show next
            pw.append("  <td class='" + deco.getClassName(_selectBar, DECO_SELECTITEM_PARAMETER, pw) +
                "' valign='middle'>\n");
            pw.append("\n<a href=\"");
            pw.append(buildToolPath("_select_page_shownext", pw));
            pw.append("\" title='");
            pw.append(translate("LY_SHOW_NEXT"));
            pw.append("'><img src='");
            pw.append(getImageUrl("LY_PAGE_SHOW_NEXT"));
            pw.append("' border='0'/></a></td>\n");
        }

        // If we do not start at the begining : add navigation tool (previous)
        if (n < _selectBar.getComponentCount())
        {
            // Next
            pw.append("  <td class='" + deco.getClassName(_selectBar, DECO_SELECTITEM_PARAMETER, pw) +
                "' valign='middle'>\n");
            pw.append("\n<a href=\"");
            pw.append(buildToolPath("_select_page_next", pw));
            pw.append("\" title='");
            pw.append(translate("LY_NEXT"));
            pw.append("'><img src='");
            pw.append(getImageUrl("LY_PAGE_NEXT"));
            pw.append("' border='0'></a></td>\n");

            // Last
            pw.append("  <td class='" + deco.getClassName(_selectBar, DECO_SELECTITEM_PARAMETER, pw) +
                "' valign='middle'>\n");
            pw.append("\n<a href=\"");
            pw.append(buildToolPath("_select_page_last", pw));
            pw.append("\" title='");
            pw.append(translate("LY_LAST"));
            pw.append("'><img src='");
            pw.append(getImageUrl("LY_PAGE_LAST"));
            pw.append("' border='0'></a></td>\n");
        }
        
        // Write form field if set
        LyFormField     formField = _selectBar.getFormField();
        
        if (formField != null)
        {
            LyWebComponent  internal = (LyWebComponent)formField.getInternalView();

            pw.append("  <td class='" + deco.getClassName(_selectBar, DECO_SELECTITEM_PARAMETER, pw) +
                "' valign='middle'>\n");
            pw.setFormId(formField.getRootController());
            internal.writeHtml(pw);
            pw.append("</td>\n");
        }

        pw.append("</tr></table>\n");
        pw.append("</td></tr></table>\n");
        pw.append("</td></tr></table>\n");

        deco.writePageDecoration(pw, LyWebDecoration.WINDOW_RIGHT);
        deco.writePageDecoration(pw, LyWebDecoration.WINDOW_BOTTOM);
        deco.writePageDecoration(pw, LyWebDecoration.WINDOW_END);

        if (!pw.writeDivOnly())
            pw.append("</div>\n");
        
        pw.endComponent(this);
    } // end method writeHtml

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebObject#refresh()
     */
    public void refresh()
    {
        // In DHTML the refresh is more clever : we refresh the DIV content
        String             id = "_" + _selectBar.getIdCode();

        LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();

        viewManager.addReloadDiv(_lastFrameName, id, this);
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.getinterface.LyGetBoundsInterface#getLocation()
     */
    public int[] getLocation()
    {
        // Not used
        return null;
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.getinterface.LyGetBoundsInterface#getSize()
     */
    public int[] getSize()
    {
        return new int[] { -1, 24 };
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalSelectBar#setStartItem(int)
     */
    public void setStartItem(int startItem)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.LyStrutsFieldItem#setLabel(java.lang.String)
     */
    public void setLabel(String label)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalSelectBar#addComponent(leon.view.LyComponent, int)
     */
    public void addComponent(LyComponent component, int pos)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalSelectBar#removeComponent(leon.view.LyComponent)
     */
    public void removeComponent(LyComponent component)
    {
        refresh();
    }

} // end class LyStrutsSelectBar
