/*
 * LyStrutsTable.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.view.web.struts;

import java.util.*;

import leon.misc.*;
import leon.view.*;
import leon.view.LyTable.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.web.*;


/**
 * This class implements the Table object (as described in the dtd) for web viewer.
 *
 * <p><b>TABLE: Table</b><br>
 * A table allows displaying a set of rows and columns containing a set of cells. It can contain a
 * header providing information about the content of the cells of the list. It is possible to define
 * its selection modes (simple, extended, multiple or none) and finally to indicate if horizontal
 * and/or vertical separation lines should be drawn.</p>
 *
 * @version $Id: LyStrutsTable.java 40777 2012-09-13 12:26:21Z nhamon $
 * @author  Lyria S.A.
 */
public class LyStrutsTable extends LyStrutsPanel
    implements LyGraphicalTable,
               LySetVisibleInterface,
               LySetTableInterface,
               LySetIdleInterface
{
	public static final String __VERSION =
		"$Id: LyStrutsTable.java 40777 2012-09-13 12:26:21Z nhamon $";

	   /** The <b>extend_sort</b> String value. */
    public final static String    EXTEND_SORT = "extend_sort".intern();

    /** The <b>extend_sort_image</b> String value. */
    public final static String    EXTEND_SORT_IMAGE = "extend_sort_image".intern();

    /** The <b>select_all</b> String value. */
    public final static String    SELECT_ALL = "select_all".intern();
    
    /** Indicates whether text selection must be disabled. */
    public static boolean         DISABLE_TEXT_SELECTION = true;

    /**
     * The <b>InvkDef</b> String value.<br />
     * Used in a javascript to invoke the table {@link LyStrutsTable#_defaultAction default action}.
     */
    protected final static String ACTION_DEFAULT = "InvkDef".intern();

    /**
     * The <b>InvkStatus</b> String value.<br />
     * Used in a javascript to invoke the table {@link LyStrutsTable#_statusAction status action}.
     */
    protected final static String ACTION_STATUS = "InvkStatus".intern();

    /**
     * The <b>InvkHref</b> String value.<br />
     * Used in a javascript to invoke the table {@link LyStrutsTable#_hrefAction href action}.
     */
    protected final static String ACTION_HREF = "InvkHref".intern();

    /**
     * The <b>InvkSel</b> String value.<br />
     * Used in a javascript to invoke the table {@link LyStrutsTable#_selectionAction selection
     * action}.
     */
    protected final static String ACTION_SELECTION = "InvkSel".intern();

    /**
     * The <b>SortCol</b> String value.<br />
     * Used in a javascript to invoke the table {@link LyStrutsTable#_sortAction sort action}.
     */
    protected final static String ACTION_SORT = "SortCol".intern();

    /**
     * The <b>ExpandRow</b> String value.<br />
     * Used in a javascript to invoke the table {@link LyStrutsTable#_expandRowAction expand row
     * action}.
     */
    protected final static String ACTION_EXPAND_ROW = "ExpandRow".intern();

    /**
     * Gets the HTML horizontal alignment string corresponding to the {@link LyHeader} alignment.
     *
     * @param  alignCode the LyHeader halign code. Possible values are {@link
     *                   LyHeader#Align.CENTER}, {@link LyHeader#Align.LEFT} and {@link
     *                   LyHeader#Align.RIGHT}.
     * @return the html horizontal alignment string.
     */
    public static String getHtmlHAligment(LyHeader.Align alignCode)
    {
        if (alignCode == LyHeader.Align.LEFT)
            return "";
        else if (alignCode == LyHeader.Align.RIGHT)
            return " align='right'";
        else if (alignCode == LyHeader.Align.CENTER)
            return "  align='center'";
        else
            return null;
    }

    /** Activated row (row that must be shown to the user). */
    protected String  _activeRow;

    /** The table generic component. */
    protected LyTable _table;

    /** The default action of the table. */
    protected String  _defaultAction;

    /** The status action. */
    protected String  _statusAction;

    /** The href action. */
    protected String  _hrefAction;

    /** The selection action. */
    protected String  _selectionAction;

    /** The sort action. */
    protected String  _sortAction;

    /** The expand row action. */
    protected String  _expandRowAction;

    /** The root URL. */
    protected String  _rootUrl;

    /** The javascript ID. */
    protected long    _javascriptId;

    /** Cell font. */
    protected String  _bodyFont;

    /** First and even row Color. */
    protected String  _rowEvenColor;

    /** Second and odd row Color. */
    protected String  _rowOddColor;

    /** CellSpacing for inner table. */
    protected int     _cellSpacing;
    
	/** Absolute URL to the Leonardi struts 'graphical event handler'. */
	private String	  _eventUrl;
	
	/** Add external border. */
	protected boolean _externalBorder;

	
	/**
	 * Constructor.
	 *
	 * @param table the table component.
	 */
	public LyStrutsTable(LyTable table)
	{
        super(table);
        
        _table = table;
        initialize();
        _cellSpacing = 0;
        _activeRow = null;
		_eventUrl = null;
	}

	/**
	 * Writes the needed javascript for this table.
	 *
	 * @param pw the page writer
	 */
	protected void writeScripts(LyWebPageWriter pw)
	{
		if (!pw.isPrintMode())
		{
			String	updateRowsColorsScript = "<script type='text/javascript'>" + getUpdateRowsColorsScript() + "</script>";

			getWebViewManager().addAdditionalRefreshScripts(updateRowsColorsScript);
			pw.addEndScript(updateRowsColorsScript, false);
		}

		String rootUrl = getBaseUrl(pw);

		pw.addBeginScript("<script type='text/javascript'>\n//<![CDATA[\nvar _rootUrl='"
			+ LyWebObject.encodeUrl(rootUrl) + "';\n//]]>\n</script>", false);

		// If a row has been activated, show it
		if (_activeRow != null)
			pw.addEndScript("<script type='text/javascript'>ShowActiveRow('" + _activeRow + "','"
				+ _lastFrameName + "');</script>", false);
		
		getDefaultAction(pw);
		getHRefAction(pw);
		getStatusAction(pw);
	} // end method writeScripts

	/**
	 * Return the scripts that must be added to the DIV component.
	 * 
	 * @param pw current page writer
	 */
	protected String getTableDivScripts(LyWebPageWriter pw)
	{
		LyRows	rows = _table.getRows();
		
		if ((rows != null) && (rows.getRow(0) != null))
		{
			// Check if table contains components (for performance, just check it on the first row).
			// If so, do not disable selection mode.
			LyRow		row = rows.getRow(0);
			LyStrutsRow	webRow = (LyStrutsRow)row.getInternalView();
			
			if (webRow.hasComponents(false))
				return "";
		}
		
		return " onMouseOver='SetSelectionStartMode(false);' onMouseOut='SetSelectionStartMode(true);'";
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.setinterface.LySetTableInterface#selectAll(boolean,
	 *      boolean)
	 */
	public void selectAll(boolean select, boolean updateView)
	{
		if (!updateView)
			return;

		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();

		StringBuffer	   script = new StringBuffer();
		String			   selectColor = getHexColor("LY_TABLE_SELECT_COLOR");
		String    		   selectFontColor = getHexColor(_table.getSelectedFontColor());

		script.append("<script type='text/javascript'>\n//<![CDATA[\n\tselectAllInTable('table_");
		script.append(_table.getIdCode()).append("','");
		script.append(_lastFrameName).append("', ").append(select);
		script.append(", '").append(selectColor).append("', '").append(selectFontColor).append("'");
		script.append(");\n//]]>\n</script>");

		viewManager.addAdditionalRefreshScripts(script.toString());
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.setinterface.LySetTableInterface#selectRow(int, boolean)
	 */
	public void selectRow(int idx, boolean select)
	{
		// Nothing to do (Javascript) in normal mode
		if (_table.getEnvironment().testEnv("LY_REPLAY"))
		{
			// Need to refresh the whole view (a call to refresh() only does not
			// work)
			_lastFrameName = LyWebConstants.TARGET_FRAME;
			refresh();
		}

		LyWebViewManager viewManager = getWebViewManager();
		StringBuffer     buffer = new StringBuffer();
		String		     selectColor = getHexColor("LY_TABLE_SELECT_COLOR");
		LyRow			 row = getRow(idx);
		String			 rowId = row.getId();
		String			 javascriptRowId = LyString.getJavascriptString(rowId);

		javascriptRowId = LyString.replace(javascriptRowId, "&#39;", "\\'");

		if (select)
			buffer.append("<script type='text/javascript'>\n//<![CDATA[\nselectRow('table_").append(_table.getIdCode())
				.append("','").append(javascriptRowId).append("','").append(selectColor).append(
				"','").append(_lastFrameName).append("','false');\n//]]>\n</script>");
		else
			buffer.append("<script type='text/javascript'>\n//<![CDATA[\nunselectRow('table_").append(_table.getIdCode())
				.append("','").append(javascriptRowId).append("','").append(_lastFrameName).append("');\n//]]>\n</script>");

		viewManager.addAdditionalRefreshScripts(buffer.toString());
		
		// Add script used to update rows color
		StringBuffer	updateBuffer = new StringBuffer();
		
		updateBuffer.append("<script type='text/javascript'>\n//<![CDATA[\n");
		updateBuffer.append(getUpdateRowsColorsScript());
		updateBuffer.append("\n//]]>\n</script>");
		
		viewManager.addAdditionalRefreshScripts(updateBuffer.toString());
	} // end method selectRow

	/**
	 * Method overriden to get the application context URL.
	 *
	 * @see LyWebViewManager#getRootURL()
	 */
	protected String getBaseUrl(LyWebPageWriter pw)
	{
		// get the application context url.
		return ((LyWebViewManager)getViewManager()).getRootURL();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.LyStrutsTable#getDefaultAction(leon.view.web.LyWebPageWriter)
	 */
	public String getDefaultAction(LyWebPageWriter pw)
	{
		return null;
	} // end method getDefaultAction

	/**
	 * Gets the href action and writes associated javascript function for this table.
	 *
	 * @param  pw the page writer
	 * @return the href action
	 */
	public String getHRefAction(LyWebPageWriter pw)
	{
		return null;
	} // end method getHRefAction

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.LyStrutsTable#getStatusAction(leon.view.web.LyWebPageWriter)
	 */
	public String getStatusAction(LyWebPageWriter pw)
	{
		return null;
	} // end method getStatusAction

	/**
	 * Gets the sort action and writes associated javascript function for this table.
	 *
	 * @param  pw the page writer
	 * @return the sort action javascript tag
	 */
	public String getSortAction(LyWebPageWriter pw)
	{
		return null;
	} // end method getSortAction

	/**
	 * This method is overriden to redirect the SORT, HREF and EXPAND actions to the Leonardi struts
	 * 'graphical event handler'.
	 *
	 * @return The absolute URL of the target action.
	 */
	public String getTableRootUrl()
	{
		if (_eventUrl != null)
			return _eventUrl;

		// get the request parameters
		Map<String, String> parameters = new HashMap<String, String>(2);

		parameters.put(LyWebConstants.CONTROLLER, _table.getRootController());
		parameters.put(LyWebConstants.SOURCE, _table.getId());

		// get the url to the selection handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();

		_eventUrl = viewManager.getActionAbsoluteUrl(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT,
				parameters, true);

		return _eventUrl;
	}

	/**
	 * Returns the minimum number of rows displayed.
	 * 
	 * @return default implementation returns 15.
	 */
	protected int getDefaultNRows()
	{
		return 8;
	}
	
	/**
	 * @see leon.view.setinterface.LySetTableInterface#setHeaders(LyHeaders)
	 */
	public void setHeaders(LyHeaders headers)
	{
		refresh();
	}
	
	/**
	 * @see leon.view.setinterface.LySetTableInterface#setFooters(LyFooters)
	 */
	public void setFooters(LyFooters footers)
	{
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.setinterface.LySetTableInterface#addRow(leon.view.LyRow)
	 */
	public void addRow(LyRow row)
	{
		if ((row != null) && row.isSelected())
			addSelectionScript(row, true);
		
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.setinterface.LySetTableInterface#addRow(leon.view.LyRow,
	 *      int)
	 */
	public void addRow(LyRow row, int idx)
	{
		if ((row != null) && row.isSelected())
			addSelectionScript(row, true);
		
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.setinterface.LySetTableInterface#removeRow(leon.view.LyRow)
	 */
	public void removeRow(LyRow row)
	{
		if ((row != null) && row.isSelected())
			addSelectionScript(row, false);
		
		refresh();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.setinterface.LySetTableInterface#removeRow(int)
	 */
	public void removeRow(int idx)
	{
		LyRow     row = getRow(idx);

		if ((row != null) && row.isSelected())
			addSelectionScript(row, false);
		
		refresh();
	}

	/**
	 * Generates a selection script and adds it to the view manager.
	 * 
	 * @param row the row that must be selected or unselected.
	 * @param addSelection indicates whether this row must be added to Javascript
	 * selection or not.
	 */
	protected void addSelectionScript(LyRow row, boolean addSelection)
	{
		LyStrutsViewManager	viewManager = (LyStrutsViewManager)getViewManager();
		StringBuffer		script = new StringBuffer();
		LyRows		 		rows = (LyRows)row.getParent();
		String				rowId = row.getId();
		String				javascriptRowId = LyString.getJavascriptString(rowId);
		LyTable		 		table = rows.getParentTable();

		javascriptRowId = LyString.replace(javascriptRowId, "&#39;", "\\'");

		script.append("<script type='text/javascript'>");
		
		if (addSelection)
			script.append("addSelectionId('table_" + table.getIdCode() + "', '" + javascriptRowId + "');");
		else
			script.append("removeSelectionId('table_" + table.getIdCode() + "', '" + javascriptRowId + "');");
			
		script.append("</script>\n");

		viewManager.addAdditionalRefreshScripts(script.toString());
	}
	
	/**
	 * Performs a refresh of the table.
	 */
	public void refresh()
	{
		boolean refresh = true;
		LyTabs  tabs = _table.getParentTabs();

		if (tabs != null)
		{
			LyStrutsTabs webTabs = (LyStrutsTabs)tabs.getInternalView();
			LyTab	     selectedTab = webTabs.getSelectedTab();
			LyTab	     fieldTab = _table.getParentTab();

			// Refresh only if table is displayed. Test if table tab is selected
			// tab and its label
			// is not null (tab with a null label contains fields that are
			// allways displayed)
			if ((selectedTab != null) && (fieldTab != null) && (fieldTab != selectedTab)
					&& (fieldTab.getLabel() != null) && (fieldTab.getLabel().length() > 0))
				refresh = false;
		}

		if (!refresh)
			return;
		
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			   id = "tableDiv_" + _table.getIdCode();

		viewManager.addReloadDiv(_lastFrameName, id, this);
		
		// If table is displayed in a popup, check popup size
		if (LyWebConstants.POPUP_VIEW.equals(_lastFrameName))
		{
			StringBuffer		script = new StringBuffer();
	
			script.append("<script type='text/javascript'>\n//<![CDATA[\n");
			script.append("checkSizePopupView(null, false)\n");
			script.append("//]]>\n</script>");
			viewManager.addAdditionalRefreshScripts(script.toString());
		}
	}

	/**
	 * Returns the Javascript code to be used to update table colors.
	 * 
	 * @return the Javascript code to be used to update table colors.
	 */
	public String getUpdateRowsColorsScript()
	{
		String		    selectColor = getHexColor(_table.getSelectedColor());
		String    		selectFontColor = getHexColor(_table.getSelectedFontColor());
		StringBuffer	updateBuffer = new StringBuffer();

		updateBuffer.append("UpdateRowsColors('table_");
		updateBuffer.append(_table.getIdCode());
		updateBuffer.append("','");
		updateBuffer.append(selectColor);
		updateBuffer.append("','").append(_lastFrameName).append("','" + selectFontColor + "');");
		
		return updateBuffer.toString();
	}

	/**
	 * Gets the Javascript code used when a context menu is performed.
	 *
	 * @return the String describing the Javascript code when a context menu is performed.
	 */
	protected String getPopupMenuScript()
	{
		// get the url to the event handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		StringBuffer		script = new StringBuffer();
		String		    	selectColor = getHexColor(_table.getSelectedColor());
		String    			selectFontColor = getHexColor(_table.getSelectedFontColor());
        		
		script.append("callOnContextMenu('','");
		script.append(_table.getRootController());
		script.append("','table_");
		script.append(_table.getIdCode());
		script.append("','");
		script.append(_table.getId());
		script.append("','");
		script.append(selectColor);
		script.append("','");
		script.append(actionUrl);
		script.append("','',event,'" + selectFontColor + "');return false;");
		
		return script.toString();
	}
	
	/*
	 * (non-Javadoc)
	 * @see leon.view.web.struts.LyStrutsPanel#getScrollHeightForSplitter(leon.view.web.LyWebPageWriter)
	 */
	public int getScrollHeightForSplitter(LyWebPageWriter pw)
	{
		LyRows    rows = _table.getRows();
		boolean   hierarchical = ((rows != null) ? rows.isHierarchical() : false);
		int		  nRows = ((rows != null)
				? (hierarchical ? rows.getSubRowCount(true) : rows.getRowCount()) : 0);
		
		nRows = Math.max(_table.getNRows(), nRows);
		
		return (nRows >= LyStrutsSplitter.MAXIMUM_COMPONENT_HEIGHT / 20 ? LyStrutsSplitter.MAXIMUM_COMPONENT_HEIGHT : -1);
	}
	
	   /**
     * Initialization method.<br>
     * Sets actions and the root URL to <code>null</code>.<br>
     * Sets the body font and row even and odd colors with properties defined in resources files.
     * <br>
     * Sets the JavaSript Id with the default value.
     */
    public void initialize()
    {
        _defaultAction = null;
        _statusAction = null;
        _hrefAction = null;
        _selectionAction = null;
        _sortAction = null;
        _expandRowAction = null;
        _rootUrl = null;

        _bodyFont = getFontString("LY_TABLE");
        _rowEvenColor = getHexColor("LY_ROW_COLOR_EVEN");
        _rowOddColor = getHexColor("LY_ROW_COLOR_ODD");
        
        String  topLeft = getEnvironment().getImageUrl("LY_TABLE_TOP_LEFT", false, false);
        
        _externalBorder = (_table.getHeaders() != null) && (topLeft != null);

        _javascriptId = -1;
    }

    /**
     * Indicates if al rows of the table are selected or not.
     *
     * @return true if all rows are selectes, false otherwise
     */
    protected boolean areAllRowsSelected()
    {
        LyRows rows = _table.getRows();
        Object lock = ((rows == null) ? new Object() : rows.getLock());

        synchronized (lock)
        {
            LyRow   row;
            int     i, nRows = ((rows == null) ? 0 : rows.getRowCount());
            boolean result = (nRows != 0);

            for (i = 0; i < nRows; i++)
            {
                row = rows.getRow(i);
                if (!row.isSelected() && !row.isHidden())
                    return false;
            }

            return result;
        }
    }

    /**
     * Indicates if at least one item (one row) is selected.
     *
     * @return true if at least one row is selected, false otherwise
     */
    protected boolean isOneItemSelected()
    {
        LyRows rows = _table.getRows();
        Object lock = ((rows == null) ? new Object() : rows.getLock());

        synchronized (lock)
        {
            LyRow row;
            int   i, nRows = ((rows == null) ? 0 : rows.getRowCount());

            for (i = 0; i < nRows; i++)
            {
                row = rows.getRow(i);
                if (row.isSelected() && !row.isHidden())
                    return true;
            }

            return false;
        }
    }

    /**
     * Gets the depth of sub-headers for current table component.
     *
     * @return the depth of sub-headers.
     */
    protected int getHeadersDepth()
    {
        LyHeaders headers = _table.getHeaders();

        if (headers == null)
            return 0;
        else
        {
            LyStrutsHeaders internal = (LyStrutsHeaders)headers.getInternalView();

            return internal.getDepth();
        }
    }

    /**
     * Writes the needed styleSheet for this component.
     *
     * @param pw the page writer
     */
    protected void writeStyleSheet(LyWebPageWriter pw)
    {
        // Add css for forms (needed for edit tables)
		String labelFont = getFontString(_table.getFont());

        pw.addCssScript("input{" + labelFont + "}", false);
        pw.addCssScript("textarea{" + labelFont + "}", false);
        pw.addCssScript("select{" + labelFont + "}", false);
        pw.addCssScript("font.label{" + labelFont + "}", false);
    } // end method writeStyleSheet

    /**
     * Completes the table with empty rows if needed.
     *
     * @param pw the page writer
     * @param fill indicates if table must be filled with empty rows
     */
    protected void writeEmptyRows(LyWebPageWriter pw, boolean fill)
    {
        LyRows    rows = _table.getRows();
        LyHeaders headers = _table.getHeaders();
        int       i, j;
        boolean   hierarchical = ((rows != null) ? rows.isHierarchical() : false);
        int       nRows = ((rows != null)
                ? (hierarchical ? rows.getSubRowCount(true) : rows.getRowCount()) : 0);
        int       nHeaders = ((headers != null) ? headers.getHeaderCount() : 0);
        LyHeader  header;
        int       defaultRows = getDefaultNRows();
        int       totalRows = (fill ? defaultRows : 1);
		String    classType;
		int		  preferedNRows = _table.getNRows();
		boolean	  checkboxSelection = false ;
		
		// Set the checkbox selection value
		if ((_table.getSelectionMode() == SelectionMode.SINGLE_CHECKBOX )
				|| (_table.getSelectionMode() == SelectionMode.MULTIPLE_CHECKBOX))
			checkboxSelection = true;
		
		if ((preferedNRows < 1) && fill)
			preferedNRows = defaultRows;
		
        // If there is no headers, retrieve the number of columns with the first
        // row
        if ((headers == null) && (nRows > 0))
        {
            LyRow row = rows.getRow(0);

            nHeaders = row.getComponentCount();
            
            // Add the checkbox to select a row
            if (checkboxSelection)
                nHeaders = nHeaders + 1;
            
            if (row.getImage() != null)
                nHeaders = nHeaders + 1;
        }

        LyComponent     parent = _table.getParent();
        LyWebPageWriter parentPage;

        if (parent.getInternalView() instanceof LyWebPageWriter)
        {
            parentPage = (LyWebPageWriter)parent.getInternalView();

            if (parentPage.isNoFrames())
                totalRows = 4;
            
            if (pw.isPrintMode())
                totalRows = 0;
        }

        if (preferedNRows > 0)
            totalRows = preferedNRows;
        
        if (nRows < totalRows)
        {
            String      popupMenuScript = getPopupMenuScript();
            
            for (i = nRows; i < totalRows; i++)
            {
                pw.append("<tr bgColor='");
                if ((i % 2) == 0)
                {
                    classType = "RE";
                    pw.append(_rowEvenColor);
                }
                else
                {
                    classType = "RO";
                    pw.append(_rowOddColor);
                }
                pw.append("\' ");
                
                if (LyStrutsRow.USE_CONTEXT_MENU
                    && (_table.getSelectionMode() != LyTable.SelectionMode.NONE))
                {
                    pw.append(" onContextMenu=\"");
                    pw.append(popupMenuScript);
                    pw.append("\"");
                }
                
                pw.append(">");
                
                if (generateExternalBorder(pw))
                    pw.append("<td class='tableLeft' width='1'></td>\n");

				if (checkboxSelection)
				{
					pw.append("<td class='status'");
					
					if ((headers != null) && !headers.isDefaultColor())
					{
						pw.append(" style='background-image:none;");
						pw.append("background-color:" + getHexColor(headers.getColor()) + ";'");
					}

					pw.append(" width='1'>&nbsp;</td>\n");
				}
				
				for (j = 0; j < nHeaders; j++)
                {
                    header = (headers == null ? null : headers.getHeader(j));
                    pw.append("<td nowrap='true' class='");

					if ((sortOnStatus()) && (j == 0))
					{
						pw.append("status'");

						if (!headers.isDefaultColor())
						{
							pw.append(" style='background-image:none;");
							pw.append("background-color:" + getHexColor(headers.getColor()) + ";'");
						}
					}
					else
						pw.append(classType);

					pw.append("'>&nbsp;</td>\n");
                    
                    if ((header != null) && (header.getBorderWidth() > 0))
                    {
                        pw.append("<td bgColor='" + getHexColor(header.getBorderColor())
                            + "'><div style='width:" + header.getBorderWidth() + "px;'></div></td>\n");
                    }
                }

                // Adding a cell for last empty coumn
                if (addLastCell(pw))
                {
                    pw.append("<td nowrap='true' class='");
                    pw.append(classType);
                    // Witdh 100% -> style (RE|RO)F
                    pw.append("F'>&nbsp;</td>\n");
                }
                
                if (generateExternalBorder(pw))
                    pw.append("<td class='tableRight' width='1'></td>\n");

                pw.append("</tr>\n");
            }
        }
    } // end method writeEmptyRows
    
    /**
     * Indicates if an empty cell must be added at the end of each row.
     *
     * @return true if an empty cell must be added, false otherwise
     */
    public boolean addLastCell(LyWebPageWriter pw)
    {
        if (pw.isPrintMode())
        {
          return false;
        }

        if  (this.getEnvironment().testResource("LY_TABLE_FILL_NONE"))
        {
          return false;
        }

        LyConstraint constraint = _table.getConstraint();
        LyComponent  parent = _table.getParent();
        
        if (constraint == null)
        {
            if (_table.getParentFormFieldId() != null)
            {
                // Retrieve the parent form field
                while ((parent != null) && !(parent instanceof LyFormField))
                    parent = parent.getParent();
                
                if (parent instanceof LyFormField)
                {
                    LyFormField field = (LyFormField)parent;
                    
                    constraint = field.getConstraint();
                }
            }
        }

        if (constraint != null)
        {
            if ((constraint.getFill() != LyConstraint.Fill.BOTH)
                    && (constraint.getFill() != LyConstraint.Fill.HORIZONTAL))
                return false;
        }

        return true;
    }

    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
     */
    public void writeHtml(LyWebPageWriter pw)
    {
        pw.startComponent(this);
        
        LyWebDecoration deco = getWebViewManager().getDecoration();
        LyComponent     parent = _table.getParent();

        // Use constraint set on table
        int    cellspacing = 0;
        int    cellpadding = 0;

        LyConstraint    constraint = _table.getConstraint();
        String          align = "center";
        boolean         vFill = true;
		boolean			hasVerticalConstraint = false;
        
        // By default, tables are centered except in form fields
        if (_table.getParent() instanceof LyFormField)
            align = "left";
        
        if (constraint == null)
        {
            if (_table.getParentFormFieldId() != null)
            {
                // Retrieve the parent form field
                while ((parent != null) && !(parent instanceof LyFormField))
                    parent = parent.getParent();
                
                if (parent instanceof LyFormField)
                {
                    LyFormField field = (LyFormField)parent;
                    
                    constraint = field.getConstraint();
                }
            }
        }

        if (constraint != null)
        {
            cellspacing = constraint.getCellSpacing();
            cellspacing = Math.max(cellspacing, 0);
            cellpadding = constraint.getCellPadding();
            cellpadding = Math.max(cellpadding, 0);

            if ((constraint.getFill() != LyConstraint.Fill.BOTH)
					&& (constraint.getFill() != LyConstraint.Fill.VERTICAL))
				vFill = false;
			else
				hasVerticalConstraint = true;
			
            if (constraint.getHAlign() == LyConstraint.HAlign.LEFT)
                align = "left";
            else if (constraint.getHAlign() == LyConstraint.HAlign.CENTER)
                align = "center";
            else if (constraint.getHAlign() == LyConstraint.HAlign.RIGHT)
                align = "right";
        }
		
		if ((_table.getParent() instanceof LyArea)
				&& (((LyArea)_table.getParent()).getLayout() == LyArea.Layout.FILL))
			hasVerticalConstraint = true;

		if (!testResource("LY_WEB_TABLE_FILL_VERTICAL"))
            vFill = false;

        if (!pw.isPrintMode())
            updateRefreshInfo(pw);

        if (pw.writeDivOnly() && (pw.getViewComponent() == _table))
        {
            if (pw.isPrintMode())
                pw.append("<table class='" + deco.getClassName(_table, pw) + "' border='1' cellspacing='0' cellpadding='2' width='100%'>\n");
            else
                pw.append("<table class='" + deco.getClassName(_table, pw) + "' id='table_" + _table.getIdCode() +
                    "' border='0' cellspacing='0' cellpadding='0' width='100%'>\n");
            
            writeScripts(pw);
            writeElements(pw, vFill);
            
            pw.append("</table>\n");
            
            pw.endComponent(this);
            return;
        }
        
        if (getWebViewManager().writeDivs())
            pw.append("<div id=\"zone1\">");

        String      controllerId = _table.getRootController();
        LyWebObject parentPage = (LyWebObject)parent.getInternalView();

        // Reset value, to readd script when refreshing...
        _defaultAction = null;
        _statusAction = null;
        _hrefAction = null;
        _selectionAction = null;
        _sortAction = null;
        _expandRowAction = null;
        _rootUrl = null;

        // FFT00146 : use one single form for a same frame
        String activeFormId = pw.getActiveFormId();

        if ((activeFormId == null) || (activeFormId.length() == 0))
            pw.setFormId(controllerId);

		String bgColor = getHexColor(_table.getColor());

        if ((bgColor != null) && !pw.isPrintMode() && (parentPage instanceof LyWebPageWriter)
                && (!((LyWebPageWriter)parentPage).isNoFrames()))
            // Set the background color
            pw.setBackgroundColor(bgColor, pw.getActiveName());

        // Add style sheet for table font
        writeStyleSheet(pw);

        // Add needed scripts
        writeScripts(pw);

        deco.writePageDecoration(pw, LyWebDecoration.WINDOW_BEGIN);
        deco.writePageDecoration(pw, LyWebDecoration.WINDOW_LEFT);

        String          width = getTableWidthParameter(constraint);
        
        pw.append("<table border='0' cellspacing='" + cellspacing + "' ");
        pw.append("cellpadding='" + cellpadding + "'");
        pw.append(width);
		pw.append(" style='" + getAdditionalStyle() + "'");
        pw.append(" align='" + align + "'");
		
		if (hasVerticalConstraint)
			pw.append(" height='100%' bgColor='" + getHexColor(_table.getColor()) + "'");

		pw.append(">");
		pw.append("\n<tr><td align='center' vAlign='top'>");
        
        pw.append("<div id='tableDiv_" + _table.getIdCode() + "'");
		pw.append(getTableDivScripts(pw));
		
		if ((constraint != null)
				&& ((constraint.getNCols() > 0) || (constraint.getNRows() > 0)))
		{
			 pw.append(" style='overflow:auto;background:" + getHexColor(_table.getColor()) + ";");
			 
			 if (constraint.getNCols() > 0)
				 pw.append("width:" + constraint.getNCols() + "px;");
			 
			 if (constraint.getNRows() > 0)
				 pw.append("height:" + constraint.getNRows() + "px;");

			 pw.append("'");
		}
		
		pw.append(">\n");
        
        // Create a first table under the real table. This is done to create a
        // single line border.
        if (pw.isPrintMode())
            pw.append("<table class='" + deco.getClassName(_table, pw) + "' border='1' cellspacing='0' cellpadding='2' width='100%'>\n");
        else
            pw.append("<table class='" + deco.getClassName(_table, pw) + "' id='table_" + _table.getIdCode() +
                    "' border='0' cellspacing='0' cellpadding='0' width='100%'>\n");

        writeElements(pw, vFill);

        pw.append("</table>\n");

        pw.append("</div>\n");
        
        pw.append("</td></tr></table>");
        
        deco.writePageDecoration(pw, LyWebDecoration.WINDOW_RIGHT);
        deco.writePageDecoration(pw, LyWebDecoration.WINDOW_END);
        
        if (getWebViewManager().writeDivs())
            pw.append("</div>");
        
        String  popupMenuView = getPopupMenuView();
        
        pw.append(popupMenuView);
        
        pw.endComponent(this);
    } // end method writeHtml

    /**
     * Indicates whether an external border must be generated for this table.
     */
    public final boolean generateExternalBorder(LyWebPageWriter pw)
    {
        return (pw.isPrintMode() ? false : _externalBorder);
    }
    
    /**
     * Returns the width parameter to be used in the HTML table element.
     */
    protected String getTableWidthParameter(LyConstraint constraint)
    {
        String width = " width='100%'";

        if (constraint != null)
        {
            if ((constraint.getFill() != LyConstraint.Fill.BOTH)
                    && (constraint.getFill() != LyConstraint.Fill.HORIZONTAL))
                width = "";
        }

        return width;
    }
    
    /**
     * Check if there no headers visible.
     * @return true if all headers are hidden, false otherwise
     */
    protected boolean areAllHeadersHidden() {
    	LyHeaders headers = _table.getHeaders();
    	if (headers != null)
    	{
    		for (int i = 0; i < headers.getHeaderCount(); i++)
    		{
				LyHeader header = headers.getHeader(i);
				if (!header.isHidden())
				{
					return false;
				}
			}
    	}
    	return true;
    }
    
    /**
     * Writes the content of the table : headers and rows.
     *
     * @param pw the page writer
     * @param addEmptyRows indicates if empty rows must be generated
     */
    protected void writeElements(LyWebPageWriter pw, boolean addEmptyRows)
    {
        LyHeaders headers = _table.getHeaders();

        if (headers != null && !areAllHeadersHidden())
        {
            LyStrutsHeaders comp = (LyStrutsHeaders)headers.getInternalView();

			pw.append("<thead>");
			
			if (comp != null)
				comp.writeHtml(pw, drawImages(), sortOnStatus());
			
			pw.append("</thead>");
        }

        LyRows rows = _table.getRows();

		pw.append("<tbody>");

		if (rows != null)
        {
            LyWebComponent comp = (LyWebComponent)rows.getInternalView();

            comp.writeHtml(pw);
        }

        // Complete table with empty rows if needed
        writeEmptyRows(pw, addEmptyRows);

        pw.append("</tbody>");

        LyFooters	footers = _table.getFooters();
        boolean		generateFooter = (footers != null) || generateExternalBorder(pw);

        if (generateFooter)
			pw.append("<tfoot>");

        if (footers != null)
        {
            LyWebComponent comp = (LyWebComponent)footers.getInternalView();

            comp.writeHtml(pw);
        }
        
        if (generateExternalBorder(pw))
        {
            int         colspan = 0;
            int         i, nHeaders  = ((headers != null) ? headers.getHeaderCount() : 0);
            LyHeader    header;

            colspan += nHeaders;
            
            for (i = 0; i < nHeaders; i++)
            {
                header = headers.getHeader(i);
                
                if ((header != null) && (header.getBorderWidth() > 0))
                    colspan++;
            }

            if (testResource("LY_DHTML_TABLE_ROW_SELECTION_BY_CHECKBOX"))
                colspan++;
            
            if (addLastCell(pw))
                colspan++;
            
            LyEnvironment   environment = getEnvironment();
            String          bottomLeft = environment.getImageUrl("LY_TABLE_BOTTOM_LEFT", false, false);
            String          bottomRight = environment.getImageUrl("LY_TABLE_BOTTOM_RIGHT", false, false);

            pw.append("<tr>");
            
            if (bottomLeft != null)
                pw.append("<td><img src='" + bottomLeft + "' /></td>\n");
            
            pw.append("<td colspan='" + colspan + "' class='tableBottom'></td>\n");
            
            if (bottomRight != null)
                pw.append("<td><img src='" + bottomRight + "' /></td>\n");
            
            pw.append("</tr>\n");
        }

        if (generateFooter)
			pw.append("</tfoot>");
    } // end method writeElements

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetVisibleInterface#setVisible(boolean)
     */
    public void setVisible(boolean status)
    {
        if (status)
        {
            LySetVisibleInterface visible = _table.getSetVisibleParent();

            if (visible != null)
                visible.setVisible(true);
        }
    }

    /**
     * Indicates if images must be drawn.
     *
     * @return true if at least one row of the table has an image to draw, false othewise
     */
    public boolean drawImages()
    {
        LyRows rows = _table.getRows();
        Object lock = ((rows == null) ? new Object() : rows.getLock());

        synchronized (lock)
        {
            int    i, nRows = ((rows == null) ? 0 : rows.getRowCount());
            LyRow  row;
            String image;

            if ((rows == null) || !rows.isHierarchical())
            {
                // Parse the rows to check if a column must be added for images
                for (i = 0; i < nRows; i++)
                {
                    row = rows.getRow(i);
                    if (row == null)
                        continue;
                    
                    image = row.getImage();
                    if ((image != null) && (image.length() != 0))
                        return true;
                }
            }
            else
            {
                Vector<LyRow>   subrows = new Vector<LyRow>();
                
                rows.getComponents(LyRow.class, subrows);
                nRows = subrows.size();

                // Parse the rows to check if a column must be added for images
                for (i = 0; i < nRows; i++)
                {
                    row = subrows.elementAt(i);
                    if (row == null)
                        continue;
                    
                    image = row.getImage();
                    if ((image != null) && (image.length() != 0))
                        return true;
                }
            }
        }

        return false;
    }

    /**
     * Indicates there is header for sorting status.
     *
     * @return true if there is at least one header for the sorting status, false otherwise
     */
    // FFT00402
    public boolean sortOnStatus()
    {
        LyHeaders headers = _table.getHeaders();

        if (headers == null)
            return false;

        LyHeader header;
        int      i, nHeaders = headers.getHeaderCount();

        // Parse the rows to check if a column must be added for images
        for (i = 0; i < nHeaders; i++)
        {
            header = headers.getHeader(i);
            
            if ((header.getLabel() != null) && header.getLabel().equals("LY_STATUS_SORT"))
                return true;
        }

        return false;
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetTableInterface#swapRows(int, int)
     */
    public void swapRows(int idx1, int idx2)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetTableInterface#selectRow(leon.view.LyRow,
     *      boolean)
     */
    public void selectRow(LyRow row, boolean select)
    {
        // Nothing to do (Javascript) in normal mode
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetTableInterface#hideRow(int, boolean)
     */
    public void hideRow(int idx, boolean hide)
    {
        refresh();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetTableInterface#updateRow(leon.view.LyRow)
     */
    public void updateRow(LyRow row)
    {
        // Nothing to do (not supported)
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetTableInterface#freezeColumn(int)
     */
    public void freezeColumn(int idx)
    {
        // Nothing to do (not supported)
    }

    /**
     * Computes if we can have an extended sort (more than one header with sort enable).
     *
     * @return true if at least one header of the table is sortable, false otherwise
     */
    public boolean allowExtendedSort()
    {
        LyHeaders headers = _table.getHeaders();

        if (headers == null)
            return false;

        LyHeader header;
        int      i, nHeaders = headers.getHeaderCount();

        // Parse the rows to check if a column must be added for images
        for (i = 0; i < nHeaders; i++)
        {
            header = headers.getHeader(i);
            if (header.isSortable())
                return true;
        }

        return false;
    }

    /**
     * Gets the sort action and writes associated javascript function for this table.
     *
     * @param  pw the page writer
     * @return the sort action javascript tag
     */
    public String getSortActionFunction()
    {
        return _sortAction;
    }

    /**
     * Gets the expand row action and writes associated javascript function for this table.
     *
     * @param  pw the page writer
     * @return the expand row action javascript tag
     */
    public String getExpandRowAction(LyWebPageWriter pw)
    {
        if (_expandRowAction != null)
            return _expandRowAction;

        if (_javascriptId == -1)
            _javascriptId = LyTools.getNextIdentifier();

        _expandRowAction = ACTION_EXPAND_ROW + _javascriptId;

        // Add script for expand Action
        StringBuffer expand = new StringBuffer();

        expand.append("<script type='text/javascript'>\n//<![CDATA[\nfunction ").append(
            _expandRowAction);
        expand.append("(rowId)\n");
        expand.append("{\n");
        expand.append("if (typeof _rootUrl != \"undefined\"){\n");
        expand.append("var temp = _rootUrl;");
        expand.append("_rootUrl = \"\";}");
        expand.append("InvokeLink('");
        expand.append(getTableRootUrl());
        expand.append("&#38;");
        expand.append(LyWebConstants.VALUE);
        expand.append("='+rowId+'");
        expand.append("&#38;");
        expand.append(LyWebConstants.TYPE);
        expand.append("=expand','reload_frames','true'");
        expand.append(")\n");
        expand.append("if (typeof _rootUrl != \"undefined\")\n");
        expand.append("\t_rootUrl = temp;\n");
        expand.append("}\n//]]>\n</script>");

        pw.addBeginScript(expand.toString(), false);

        return _expandRowAction;
    } // end method getExpandRowAction

    /**
     * Gets the selection action javascript function name for this table.
     *
     * @param  pw the page writer
     * @return the selection action
     */
    public String getSelectionAction(LyWebPageWriter pw)
    {
        if (_selectionAction != null)
            return _selectionAction;

        if (_javascriptId == -1)
            _javascriptId = pw.getNextJavaScriptId();

        _selectionAction = ACTION_SELECTION + _javascriptId;
        return _selectionAction;
    }

    /**
     * Writes the selection action javascript function for this table.
     *
     * @param pw the page writer
     */
    protected void writeSelectionAction(LyWebPageWriter pw)
    {
        if (_table.getSelectionMode() != LyTable.SelectionMode.NONE)
        {
            // Multi selection
            StringBuffer select = new StringBuffer();

            select.append(
                "<script type='text/javascript'>\n//<![CDATA[\nfunction tableSelection");
            select.append("()\n");
            select.append("{\n");
            select.append("InvokeSelection('");
            select.append(LyWebConstants.SELECT_ITEM);
            select.append("','','',false,'");
            select.append(_table.getRootController());
            select.append("');");
            select.append("}\n//]]>\n</script>");

            pw.addBeginScript(select.toString(), false);
            select = null;
        }
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetTableInterface#showRow(int)
     */
    public void showRow(int idx)
    {
		LyRow	row = getRow(idx);
        String	parentFieldId = _table.getParentFormFieldId();
		
		_activeRow = (row != null ? LyString.getJavascriptString(row.getId()) : null);

        // If a row has been activated and table is not in a form, show it
        if ((_activeRow != null)
                && ((parentFieldId == null) || (parentFieldId.length() == 0)))
        {
            LyWebViewManager viewManager = getWebViewManager();
            StringBuffer     buffer = new StringBuffer();

            buffer.append("<script type='text/javascript'>\n//<![CDATA[\nShowActiveRow('").append(
            	_activeRow).append("','").append(_lastFrameName).append("');\n//]]>\n</script>");

            viewManager.addAdditionalRefreshScripts(buffer.toString());
        }
    }
    
    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetIdleInterface#setIdle(boolean)
     */
    public void setIdle(boolean status)
    {
        if (!status)
            _activeRow = null;
    }

    /**
     * @see leon.view.setinterface.LySetEnabledInterface#setEnabled(boolean)
     */
    public void setEnabled(boolean status)
    {
    }
	
	/**
	 * @see leon.view.graphic.LyGraphicalTable#setNRows(int)
	 */
	public void setNRows(int nRows)
	{
		refresh();
	}

    /**
     * Gets the row at the specified position.
     *
     * @param  i the row position
     * @return the LySwingRow at position i
     */
    public LyRow getRow(int i)
    {
        LyRows rows = _table.getRows();

        if (rows == null)
            return null;

        if (!rows.isHierarchical()
            && !rows.hasHiddenRow())
        {
            LyRow subRow = rows.getRow(i);
            
            return subRow;
        }
        else
        {
            Vector<LyRow>       subRows = new Vector<LyRow>();

            rows.getDisplayedSubRows(subRows);
            
            if (i >= subRows.size())
                return null;
            else
                return subRows.elementAt(i);
        }
    }
	
	/**
	 * Computes an additional style for the table.
	 *
	 * @return a string containing the CSS additional style
	 */
	public String getAdditionalStyle()
	{
		String style ="";
		
		if (!_table.isDefaultColor())
			style += "background-color:" + getHexColor(_table.getColor()) + ";";
		
		if (!_table.isDefaultFontColor() || !_table.isDefaultFont())
			style += getFontString(_table.getFont(), _table.getFontColor());
		
		return style;
	}
} // end class LyStrutsTable
