/*
 * LyStrutsTabs.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.view.web.struts;

import java.io.*;
import java.util.*;

import leon.misc.*;
import leon.view.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.web.*;
import leon.view.web.struts.action.*;


/**
 * This class implements the Tabs object (as described in the dtd) for Struts viewer.
 *
 * <p><b>TABS: Tabs</b><br>
 * A container of tab type allows displaying different panels in the same graphical area, only one
 * panel being visible at once. Each panel is contained in a tab, the container being made up of a
 * set of tabs. The displaying that corresponds to a container of tab type must present two areas: a
 * first one showing the names of the different tabs and a second one allowing displaying the
 * panels.</p>
 *
 * @version $Id: LyStrutsTabs.java 41135 2012-10-10 16:17:13Z nhamon $
 * @author  Lyria S.A.
 */
public class LyStrutsTabs extends LyStrutsContainer
    implements LyGraphicalTabs,
               LySetTabInterface
{
	public static final String __VERSION =
		"$Id: LyStrutsTabs.java 41135 2012-10-10 16:17:13Z nhamon $";


	/**
	 * Indicates if tabs must wrap automatically in web mode.  Can be one of the three following values:
	 * true = tabs wrap automatically
	 * false = the window  width is increased in order to have tabs on the same line
	 * arrows = arrows appear to scroll the different tabs without increasing the window width
	 */
	public static String		  TABS_WRAP = "true";
	
	/** The <b>tabs</b> String value used within frames names. */
    protected static final String TABS = "tabs".intern();

    /** The <b>arrows</b> String value used within frames names. */
    protected static final String ARROWS = "arrows".intern();

    /** The <b>tab</b> String value used within frames names. */
    protected static final String TAB = "tab".intern();

    /** Constant used to generate CSS class name for label. */
    public static int             DECO_LABEL_PARAMETER = 0;

    /** Constant used to generate CSS class name for selected label. */
    public static int             DECO_DISABLED_LABEL_PARAMETER = 1;

    /** Constant used to generate CSS class name for selected label. */
    public static int             DECO_SELECTED_LABEL_PARAMETER = 2;

    /** Constant used to generate CSS class name for selected label. */
    public static int             DECO_DISABLED_SELECTED_LABEL_PARAMETER = 3;

    /** Constant used to generate CSS class name for tab content. */
    public static int             DECO_TAB_CONTENT_PARAMETER = 4;

    /** Constant used to generate CSS class name for headers. */
    public static int             DECO_HEADERS_PARAMETER = 5;

    /** The selected generic tab component of this tabs component. */
    protected LyTab               _selectedTab;

    /** The tabs generic component. */
    protected LyTabs              _tabs;

    /** The header frame name. */
    protected String              _lastHeaderFrameName;

    /** The last tab frame name. */
    protected String              _lastTabFrameName;

    /** The tabs ID getted at random. */
    protected String              _randomTabsId;

    /** List of enabled tabs : for example if the tab at index i is enable, _enabledTab[i]=true. */
    protected boolean[]           _enabledTab;
    
    /** Tab instance that is being removed from tabs component. */
    protected LyTab               _removedTab;
    
	/**
	 * Constructor.
	 *
	 * @param tabs the tabs component
	 */
	public LyStrutsTabs(LyTabs tabs)
	{
        super(tabs);
        
        _tabs = tabs;

        init();

        _selectedTab = getSelectedTab();
        _removedTab = null;
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
	 */
	public synchronized void writeHtml(LyWebPageWriter pw)
	{
	    pw.startComponent(this);
		updateRefreshInfo(pw);

		String			width = " width='100%'";
		String			height = "";
		LyConstraint	constraint = _tabs.getConstraint();
		LyView			view = _tabs.getParentView();
		boolean			isFullDialog = (view instanceof LyDialog) && !LyWebConstants.POPUP_VIEW.equals(pw.getWindowName());
		boolean			generateTable = !LyWebViewManager.OPTIMIZE_TABLE_GENERATION && !isFullDialog;
		LyWebDecoration	deco = getWebViewManager().getDecoration();
		boolean			useArrows = "arrows".equalsIgnoreCase(TABS_WRAP);
		boolean			generateResizeScript = false;
		
		if (constraint != null)
		{
			if ((constraint.getFill() != LyConstraint.Fill.BOTH)
					&& (constraint.getFill() != LyConstraint.Fill.HORIZONTAL))
				width = "";
			
			if ((constraint.getFill() != LyConstraint.Fill.BOTH)
					&& (constraint.getFill() != LyConstraint.Fill.HORIZONTAL)
					&& (constraint.getNCols() > 0))
			{
				width = " width='" + constraint.getNCols() + "px'";
				generateTable = true;
			}

			if ((constraint.getFill() == LyConstraint.Fill.BOTH)
					|| (constraint.getFill() == LyConstraint.Fill.VERTICAL))
			{
				if ((getWebViewManager().getBrowser() == LyWebViewManager.BROWSER_MSIE)
						&& !LyWebConstants.POPUP_VIEW.equals(_lastWindowName)
						&& (_tabs.getParentView() instanceof LyFrame)
						&& (_tabs.getParent() instanceof LyArea)
						&& (((LyArea)_tabs.getParent()).getPanelCount() == 1))
					generateResizeScript = true;
				else
					height = " height='100%'";
			}
			
			if (constraint.getNRows() > 0)
			{
				height = " height='" + constraint.getNRows() + "px'";
				generateTable = true;
			}
		}
		
		if (useArrows)
			generateTable = true;

		String			style = (useArrows ? "overflow:hidden;width:200px;" : "");

		if (generateTable)
		{
			pw.append("<table id='tabs_full_content_" + _tabs.getIdCode() + "' border='0' cellpadding='0' cellspacing='0'");
			pw.append(width);
			pw.append(height);
			pw.append(">\n");
			pw.append("<tr height='1'>\n");
			pw.append("<td valign='top' id='tabs_content_" + _tabs.getIdCode() + "'>\n");
		}

		pw.append("<div id='tabs_div_" + _tabs.getIdCode() + "' style='position:relative;" + style + "'>\n");
		pw.append("<div id='" + TABS + _tabs.getIdCode() + "' style='position:relative;height:100%;top:0px;left:0px;'>\n");
		writeTabsHeader(pw);
		pw.append("</div>\n");
		
		if (useArrows)
		{
			pw.append("<div style='position:absolute;height:28px;top:0px;right:-1px;background-color:" + getHexColor("LY_DEFAULT_BACKGROUND") + ";'>\n");
			writeArrows(pw);
			pw.append("</div>\n");
			
			StringBuffer	buffer = new StringBuffer();
			
			buffer.append("<script type='text/javascript'>");
			buffer.append("checkScrollTabHeaders('" + _lastFrameName + "','" + TABS + _tabs.getIdCode() +
					"','tabs_headers_table_" + _tabs.getIdCode() + "','tabs_content_" + _tabs.getIdCode() +
					"','selectedLeftImage_" + _tabs.getIdCode() + "','selectedRightImage_" + _tabs.getIdCode() + "');");
			buffer.append("</script>\n");
			
			pw.addEndScript(buffer.toString(), false);
		}

		pw.append("</div>\n");

		if (generateTable)
		{
			pw.append("</td>\n");
			pw.append("</tr>\n");
			pw.append("<tr height='100%'>\n");
			pw.append("<td class='" + deco.getClassName(_tabs, DECO_TAB_CONTENT_PARAMETER, pw)
					+ "' valign='top' height='100%' id='tab_content_" + _tabs.getIdCode() + "'>\n");
		}

		pw.append("<div id='" + TAB + _tabs.getIdCode() + "'");
		
		if (!LyWebConstants.POPUP_VIEW.equals(_lastWindowName)
				&& (getWebViewManager().getBrowser() != LyWebViewManager.BROWSER_MSIE))
			pw.append(" style='height:100%;'");
		
		pw.append(">");
		
		if (pw.isPrintMode())
		{
			int		i, n = _tabs.getTabCount();
			LyTab	tab;
			String	tabName;
			
			pw.append("<table width='100%' cellpadding='0' cellspacing='0'>");
			
			for (i = 0; i < n; i++)
			{
				tab = _tabs.getTab(i);
				tabName = translate(tab.getLabel());

				if ((tabName != null) && (tabName.endsWith("\\n") || tabName.endsWith("\n")))
				{
					// Tab name ends with '\n' : jump to next line
					if (tabName.endsWith("\\n"))
						tabName = tabName.substring(0, tabName.length() - 2);
					else if (tabName.endsWith("\n"))
						tabName = tabName.substring(0, tabName.length() - 1);
				}

				tabName = getHtmlString(tabName);

				if ((tabName == null) || (tabName.length() == 0))
					continue;
				
				int		classType = DECO_LABEL_PARAMETER;
				
				if (!tab.isEnabled())
					classType = DECO_DISABLED_LABEL_PARAMETER;
				
				pw.append("<tr>");
				pw.append("<td>");
				pw.append("<font class='" + deco.getClassName(_tabs, classType, pw) + "'>");
				pw.append("&nbsp;" + tabName + "&nbsp;</font><br/><hr/>");

				LyWebObject		internal = (LyWebObject)tab.getInternalView();
				
				internal.writeHtml(pw);
				
				if (i != n - 1)
				{
					pw.append("<tr height='20'>");
					pw.append("<td>");
					pw.append("</td>");
					pw.append("</tr>");
				}

				pw.append("</td>");
				pw.append("</tr>");
			}
			
			pw.append("</table>");
		}
		else
		{
	        // Build all tabs
            int     i, n = _tabs.getTabCount();
                
            for (i = 0; i < n; i++)
                _tabs.getTab(i).getInternalView();
            
            // Write selected one
			LyTab    tab = getSelectedTab();
		
			if (tab != null)
			{
				LyWebObject		internal = (LyWebObject)tab.getInternalView();
			
				internal.writeHtml(pw);

				// If tab contains a form, give focus to first input in the form
				Vector<LyForm>	forms = new Vector<LyForm>();
				
				tab.getComponents(LyForm.class, forms);
				
				if (forms.size() > 0)
				{
					pw.append("<script type='text/javascript'>");
					pw.append("getLeonardiTargetFrame().ResetFormFocusImmediatly(false);");
					pw.append("</script>\n");
				}
			}
		}
		
		pw.append("</div>");

		if (generateTable)
		{
			pw.append("</td>\n");
			pw.append("</tr>\n");
			pw.append("</table>\n");
		}
		
		if (generateResizeScript)
		{
			StringBuffer	script = new StringBuffer();
			LyArea			area = _tabs.getParentArea();
			
			script.append("<script type='text/javascript'>");
			script.append("getLeonardiTopFrame().addSplitterIds(new Array(true,'mainScroll_" + area.getIdCode() +
					"','tabs_full_content_" + _tabs.getIdCode() + "','','fullHeight',false));");
			script.append("</script>");
	
			getWebViewManager().addAdditionalRefreshScripts(script.toString());
			pw.addEndScript(script.toString(), false, false);
		}

		// Tabs headers width is computed with the Javascript function
		if (useArrows)
		{
			StringBuffer	script = new StringBuffer();
			
			script.append("<script type='text/javascript'>");
			script.append("getLeonardiTopFrame().addSplitterIds(new Array(false,'tabs_content_" + _tabs.getIdCode() +
					"','tabs_div_" + _tabs.getIdCode() + "','','fullWidth',false));");
			script.append("</script>");
	
			getWebViewManager().addAdditionalRefreshScripts(script.toString());
			pw.addEndScript(script.toString(), false);
		}
		
		pw.endComponent(this);
	} // end method writeHtml
	
	/*
	 * (non-Javadoc).
	 * @see leon.view.web.LyWebPageWriter#writeHtml(java.io.Writer, java.lang.String)
	 */
	public synchronized void writeHtml(Writer pw, String frame)
		throws IOException
	{
	    
		LyStrutsPage		page = new LyStrutsPage(_tabs, true);

		page.setWriteDivOnly(false);
		page.setFrameName(getLastFrameName());
		page.setWindowName(getWindowName());
		
		if ((TABS + _tabs.getIdCode()).equals(frame))
		{
			writeTabsHeader(page);
			
			if ("arrows".equalsIgnoreCase(TABS_WRAP))
			{
				StringBuffer	buffer = new StringBuffer();
				
				buffer.append("<script type='text/javascript'>");
				buffer.append("checkScrollTabHeaders('" + _lastFrameName + "','" + TABS + _tabs.getIdCode() +
						"','tabs_headers_table_" + _tabs.getIdCode() + "','tabs_content_" + _tabs.getIdCode() +
						"','selectedLeftImage_" + _tabs.getIdCode() + "','selectedRightImage_" + _tabs.getIdCode() + "');");
				buffer.append("</script>\n");
				
				getWebViewManager().addAdditionalRefreshScripts(buffer.toString());
			}
		}
		else if ((TAB + _tabs.getIdCode()).equals(frame))
		{
			LyTab			tab = getSelectedTab();
			LyWebObject		internal = (LyWebObject)tab.getInternalView();
			
			internal.writeHtml(page);

			// If tab contains a form, give focus to first input in the form
			Vector<LyForm>	forms = new Vector<LyForm>();
			
			tab.getComponents(LyForm.class, forms);
			
			if (forms.size() > 0)
			{
				StringBuffer	buffer = new StringBuffer();
				
				buffer.append("<script type='text/javascript'>");
				buffer.append("getLeonardiTargetFrame().ResetFormFocusImmediatly(false);");
				buffer.append("</script>\n");
				
				getWebViewManager().addAdditionalRefreshScripts(buffer.toString());
			}
		}

		StringBuffer buffer = page.getBuffer(null);

		pw.write(buffer.toString());
		
		page.free();
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.web.struts.LyStrutsTabs#refreshFrames(boolean)
	 */
	protected void refreshFrames(boolean refreshTabFrame)
	{
		if (_lastFrameName == null)
			return;
		
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_PAGE);
		StringBuffer		script = new StringBuffer();
		String				target = viewManager.getTargetCode();

		script.append("<script type='text/javascript'>\n//<![CDATA[\n");
		
		// Call the RefreshDiv function to refresh the content of tabs.
		// Call it on the top frame to prevent from blocking the wait mechanism.
		if (refreshTabFrame)
			script.append("getLeonardiTopFrame().RefreshDiv('" + _lastFrameName + "', '" + getId() + "', '" +
					TAB + _tabs.getIdCode() + "', '" + TAB + _tabs.getIdCode() + "', '" +
					actionUrl + "', '" + LyStrutsPageAction.CODE_SEPARATOR + "','" + target + "');\n");
		
		script.append("getLeonardiTopFrame().RefreshDiv('" + _lastFrameName + "', '" + getId() + "', '" +
				TABS + _tabs.getIdCode() + "', '" + TABS + _tabs.getIdCode() + "', '" +
				actionUrl + "', '" + LyStrutsPageAction.CODE_SEPARATOR + "','" + target + "');\n");
		script.append("//]]>\n</script>");
		
		viewManager.addAdditionalRefreshScripts(script.toString());
	}

	/**
	 * Indicates whether arrows must be generated in the HTML page.
	 * 
	 * @param pw current page writer
	 * @return true if arrows must be generated in the HTML page, false otherwise.
	 */
	protected boolean displayArrows(LyWebPageWriter pw)
	{
		return false;
	}

	/**
	 * Writes arrows used to move tabs headers.
	 *
	 * @param pw the page writer
	 */
	protected void writeArrows(LyWebPageWriter pw)
	{
		if (!pw.writeDivOnly())
		{
			// Add two arrows for scrolling
			String	titleLeft = "title='" + translate("LY_TABS_MOVE_LEFT") + "'";
			String	titleRight = "title='" + translate("LY_TABS_MOVE_RIGHT") + "'";
			
			pw.append("<table border='0' cellpadding='0' cellspacing='2' height='100%'><tr>");
			pw.append("<td><a href=\"javascript:scrollTabHeaders(false,'" + _lastFrameName + "','" + TABS + _tabs.getIdCode() +
					"','tabs_headers_table_" + _tabs.getIdCode() + "','tabs_content_" + _tabs.getIdCode() + "');\"><img src='");
			pw.append(getImageUrl("LY_TAB_MOVE_LEFT"));
			pw.append("' border='0' " + titleLeft + "/></a></td>");
			pw.append("");
			pw.append("<td><a href=\"javascript:scrollTabHeaders(true,'" + _lastFrameName + "','" + TABS + _tabs.getIdCode() +
					"','tabs_headers_table_" + _tabs.getIdCode() + "','tabs_content_" + _tabs.getIdCode() + "');\"><img src='");
			pw.append(getImageUrl("LY_TAB_MOVE_RIGHT"));
			pw.append("' border='0' " + titleRight + "/></a></td>");
			pw.append("</tr></table>");
		}
	}

	/**
	 * Method overriden to return the URL to the "Command" struts action path.
	 */
	public String getActionUrl()
	{
		// get the view manager
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();

		// build the URL for the Leonardi struts command action
		String			    url = viewManager.getActionAbsoluteUrl(
				LyStrutsConstants.PATH_ACTION_COMMAND, null, true);

		return url;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.LyStrutsTabs#getTabListUrl(leon.view.web.LyWebPageWriter)
	 */
	protected String getTabListUrl(LyWebPageWriter pw)
	{
		// prepare the parameters
		String			    targetWindow = pw.getWindowName();

		// set the parameters
		Map<String, String> parameters = new HashMap<String, String>(2);

		parameters.put(LyWebConstants.PAGE, getId());
		parameters.put(LyWebConstants.FRAME, _randomTabsId + TABS);
		if (_lastReloadTime > 0)
			parameters.put("reload", String.valueOf(_lastReloadTime));
		if (targetWindow != null)
			parameters.put(LyWebConstants.TARGET_WINDOW, targetWindow);

		// get the url to the page action
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    url = viewManager.getActionAbsoluteUrl(
				LyStrutsConstants.PATH_ACTION_PAGE, parameters, true);

		return url;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.LyStrutsTabs#getTabArrowsUrl(leon.view.web.LyWebPageWriter)
	 */
	protected String getTabArrowsUrl(LyWebPageWriter pw)
	{
		// prepare the parameters
		String			    targetWindow = pw.getWindowName();

		// set the parameters
		Map<String, String> parameters = new HashMap<String, String>(2);

		parameters.put(LyWebConstants.PAGE, getId());
		parameters.put(LyWebConstants.FRAME, _randomTabsId + ARROWS);
		if (_lastReloadTime > 0)
			parameters.put("reload", String.valueOf(_lastReloadTime));
		if (targetWindow != null)
			parameters.put(LyWebConstants.TARGET_WINDOW, targetWindow);

		// get the url to the page action
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    url = viewManager.getActionAbsoluteUrl(
				LyStrutsConstants.PATH_ACTION_PAGE, parameters, true);

		return url;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.LyStrutsTabs#getTabContentUrl(leon.view.web.LyWebPageWriter)
	 */
	protected String getTabContentUrl(LyWebPageWriter pw)
	{
		// prepare the parameters
		String			    targetWindow = pw.getWindowName();
		String			    controllerId = _tabs.getRootController();

		// set the parameters
		Map<String, String> parameters = new HashMap<String, String>(2);

		parameters.put(LyWebConstants.PAGE, getId());
		parameters.put(LyWebConstants.FRAME, controllerId + TAB);
		if (_lastReloadTime > 0)
			parameters.put("reload", String.valueOf(_lastReloadTime));
		if (targetWindow != null)
			parameters.put(LyWebConstants.TARGET_WINDOW, targetWindow);

		// get the url to the page action
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    url = viewManager.getActionAbsoluteUrl(
				LyStrutsConstants.PATH_ACTION_PAGE, parameters, true);

		return url.toString();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.LyStrutsTabs#getTabURL(leon.view.web.LyWebPageWriter, int)
	 */
	protected String getTabURL(LyWebPageWriter pw, int tabIndex)
	{
		// prepare the parameters
		String			    controllerId = _tabs.getRootController();

		// get the action URL
        LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		StringBuffer	    url = new StringBuffer();

		url.append(actionUrl).append('?');
		url.append(LyWebConstants.CONTROLLER).append('=').append(controllerId);
		url.append("&#38;");
		url.append(LyWebConstants.SOURCE).append('=').append("_tabs");
		url.append("&#38;");
		url.append(LyWebConstants.INDEX).append('=').append("_tabs");
		url.append("&#38;");
		url.append(LyWebConstants.TYPE).append('=').append("tab");
		url.append("&#38;");
		url.append(LyWebConstants.PAGE).append('=').append(pw.getId());
		url.append("&#38;");
		url.append(LyWebConstants.FRAME).append('=').append(controllerId + TAB);

		return url.toString();
	}
	
	   /**
     * Initialization method.<br>
     * Overriden to add initialization of this class attributes.
     *
     * @see leon.view.web.LyWebPage#init()
     */
    public void init()
    {
        super.init();

        _randomTabsId = LyString.getRandomId() + "_";

        int nbTab = _tabs.getTabCount();

        _enabledTab = new boolean[nbTab];
        for (int i = 0; i < nbTab; i++)
            _enabledTab[i] = true;

    }

    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebPageWriter#preparePage(java.lang.String,
     *      java.lang.String)
     */
    public synchronized void preparePage(String frame, String targetWindow)
    {
        super.preparePage(frame, targetWindow);

        // Write the buffers for areas
        int            i, nTabs = _tabs.getTabCount();
        LyTab          tab;
        LyPanel        panel;
        LyWebComponent webComponent;
        String         tabName;
        String         controllerId = _tabs.getRootController();

        clearBuffer(controllerId + TAB);
        setActiveBuffer(controllerId + TAB);

        String  selectedTab = _tabs.getSelectedTab();
        boolean firstWritten = false;

        if ((selectedTab != null) && (selectedTab.length() == 0))
            selectedTab = null;

        LyWebDecoration deco = getWebViewManager().getDecoration();

        deco.writePageDecoration(this, LyWebDecoration.WINDOW_BEGIN);
        deco.writePageDecoration(this, LyWebDecoration.WINDOW_TOP);
        deco.writePageDecoration(this, LyWebDecoration.WINDOW_LEFT);

        deco.writePageDecoration(this, LyWebDecoration.CONTENT_BEGIN_FULL_HEIGHT);
        deco.writePageDecoration(this, LyWebDecoration.CONTENT_TOP);
        deco.writePageDecoration(this, LyWebDecoration.CONTENT_LEFT);

        for (i = 0; i < nTabs; i++)
        {
            tab = _tabs.getTab(i);

            if ((tab.getLabel() == null) || (tab.getLabel().length() == 0)
                    || (tab.getLabel().equals(selectedTab))
                    || ((tab.getId() != null) && tab.getId().equals(selectedTab))
                    || ((selectedTab == null) && (!firstWritten)) || (isPrintMode()))
            {
                tabName = translate(tab.getLabel());
                panel = tab.getPanel();
                webComponent = (LyWebComponent)panel.getInternalView();

                if ((webComponent instanceof LyWebPageWriter)
                        && (((LyWebPageWriter)webComponent).isFrameSet() || _noFrames))
                    ((LyWebPageWriter)webComponent).preparePage(frame, null);

                if ((tab.getLabel() != null) && (tab.getLabel().length() != 0) && isPrintMode())
                {
                    append("<br/>\n");
                    append("<table  align='center' width='100%'>\n");
                    append("<tr><td width='25'><hr/></td>\n");
                    append("<td width='1'><font class='"
                        + deco.getClassName(_tabs, DECO_SELECTED_LABEL_PARAMETER, this) + "'>" + tabName
                        + "</font></td>\n");
                    append("<td><hr/></td>\n");
                    append("</tr></table>\n");
                }

                webComponent.writeHtml(this);

                if ((selectedTab == null) && (tab.getLabel() != null)
                        && (tab.getLabel().length() != 0))
                    firstWritten = true;

                if (((tab.getLabel() == null) || (tab.getLabel().length() == 0))
                        && (!isPrintMode()))
                {
                    String top = getImageUrl("LY_TAB_TOP");

                    append("<table  align='center' width='100%'>\n");
                    append("<tr><td background='" + top + "'>&nbsp;</td></tr>\n");
                    append("</table>\n");
                }
            }
        }

        if (!isPrintMode())
        {
            deco.writePageDecoration(this, LyWebDecoration.CONTENT_RIGHT);
            deco.writePageDecoration(this, LyWebDecoration.CONTENT_FOOTER);
            deco.writePageDecoration(this, LyWebDecoration.CONTENT_END);
        }
        deco.writePageDecoration(this, LyWebDecoration.WINDOW_RIGHT);
        deco.writePageDecoration(this, LyWebDecoration.WINDOW_BOTTOM);
        deco.writePageDecoration(this, LyWebDecoration.WINDOW_END);

        clearBuffer(_randomTabsId + TABS);
        setActiveBuffer(_randomTabsId + TABS);

        deco.writePageDecoration(this, LyWebDecoration.WINDOW_BEGIN);
        deco.writePageDecoration(this, LyWebDecoration.WINDOW_HEADER_LEFT);

        writeTabsHeader(this);

        deco.writePageDecoration(this, LyWebDecoration.WINDOW_HEADER_RIGHT);
        deco.writePageDecoration(this, LyWebDecoration.WINDOW_END);

        // Write the buffers for areas
        nTabs = _tabs.getTabCount();
        controllerId = _tabs.getRootController();

        clearBuffer(controllerId + TAB);
        setActiveBuffer(controllerId + TAB);
        _selectedTab = getSelectedTab();

        selectedTab = _tabs.getSelectedTab();
        firstWritten = false;

        if ((selectedTab != null) && (selectedTab.length() == 0))
            selectedTab = null;

        if (getWebViewManager().writeDivs())
            append("<div id=\"zone1\">");
        else
            append(
                "<table border='0' cellpadding='0' cellspacing='0' width='100%' height='100%'><tr><td id=\"background\">");
        
        // To avoid a double scrolling with firefox when tabs are displayed in the popup view,
        // generate tabs headers in the same iframe as the tab content.
        if ((getWebViewManager().getBrowser() == LyWebViewManager.BROWSER_FIREFOX)
                && LyWebConstants.POPUP_VIEW.equals(targetWindow))
            writeTabsHeader(this);

        append("<table border='0' cellpadding='0' cellspacing='0' width='100%' height='100%'>");

        for (i = 0; i < nTabs; i++)
        {
            tab = _tabs.getTab(i);

            if ((tab.getLabel() == null) || (tab.getLabel().length() == 0)
                    || (tab.getLabel().equals(selectedTab))
                    || ((tab.getId() != null) && tab.getId().equals(selectedTab))
                    || ((selectedTab == null) && (!firstWritten)) || (isPrintMode()))
            {

                append("<tr><td width='1%'>");
                tabName = translate(tab.getLabel());
                panel = tab.getPanel();
                webComponent = (LyWebComponent)panel.getInternalView();

                // Now done in the computeNoFrame() method of the LyStrutsArea
                // class
                if ((webComponent instanceof LyWebPageWriter)
                        && (((LyWebPageWriter)webComponent).isFrameSet() || _noFrames))
                    ((LyWebPageWriter)webComponent).preparePage(frame, null);

                if ((tab.getLabel() != null) && (tab.getLabel().length() != 0) && isPrintMode())
                {
                    append("<br />\n");
                    append("<table  align='center' width='100%'>\n");
                    append("<tr><td width='25'><hr/></td>\n");
                    append("<td width='1'><font class='" + deco.getClassName(_tabs, DECO_SELECTED_LABEL_PARAMETER, this)
                        + "'>" + tabName + "</font></td>\n");
                    append("<td><hr/></td>\n");
                    append("</tr></table>\n");
                }

                webComponent.writeHtml(this);

                if ((selectedTab == null) && (tab.getLabel() != null)
                        && (tab.getLabel().length() != 0))
                    firstWritten = true;

                if (((tab.getLabel() == null) || (tab.getLabel().length() == 0))
                        && (!isPrintMode()))
                {
                    append("</td></tr><tr><td width='1%'><table  align='center' width='100%'>\n");
                    append("<tr><td></td></tr>\n");
                    append("</table>\n");
                }

                append("</td></tr>\n");
            }
        }

        // If the tabs isn't in a dialog (it's a tab action) it must take
        // maximum size
        if (_tabs.getParentView() instanceof LyDialog)
        {
            append("<tr height='100%'>\n <td>\n");
            append(" </td>\n</tr>\n");
        }

        append("</table>\n");

        if (getWebViewManager().writeDivs())
            append("</div>\n");
        else
            append("</td></tr></table>");

        clearBuffer(_randomTabsId + TABS);
        setActiveBuffer(_randomTabsId + TABS);

        writeTabsHeader(this);

        clearBuffer(_randomTabsId + ARROWS);
        setActiveBuffer(_randomTabsId + ARROWS);

        writeArrows(this);

    } // end method preparePage

    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebPageWriter#getRedirect(java.lang.String)
     */
    public String getRedirect(String frame)
    {
        String controllerId = _tabs.getRootController();

        if ((controllerId + TAB).equals(frame))
        {

            LyPanel        panel;
            LyWebComponent webComponent;
            
            if (getSelectedTab() == null)
                return null;
            
            panel = getSelectedTab().getPanel();

            webComponent = (LyWebComponent)panel.getInternalView();

            if (webComponent instanceof LyWebPageWriter)
                return ((LyWebPageWriter)webComponent).getRedirect(frame);
        }
        return null;
    }

    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebPage#getBodyScript()
     */
    protected String getBodyScript()
    {
        if (LyWebConstants.POPUP_VIEW.equals(getWindowName()))
            return "onload='checkSizePopupView(null, \"resetScroll\");ResetFormFocus(false);'";
        else
            return "onload='checkSizeMainView();ResetFormFocus(false);' onResize='checkSizeMainView();try {placeScrollDiv();}catch (ex) {};'";

    }

    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebPage#isInPopUp()
     */
    protected boolean isInPopUp()
    {
        if (LyWebConstants.POPUP_VIEW.equals(getWindowName()))
            return true;
        else
            return false;
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalTabs#setSelectedTab(java.lang.String)
     */
    public void setSelectedTab(String selectedTab)
    {
        if (_tabs.getEnvironment().testEnv("LY_REPLAY"))
        {
            // Need to refresh the whole view (a call to refresh() only does not
            // work)
            _lastFrameName = LyWebConstants.TARGET_FRAME;
            refresh();
            return;
        }

        // If selected tab is still the same, no need to refresh
        LyTab newSelectedTab = getSelectedTab();
        
        if (newSelectedTab == _selectedTab)
            return;
        
        _selectedTab = newSelectedTab;
        refreshFrames(true);
    }

    /**
     * Writes the beginning of the page. Overriden to add the selection of a given tab on load.
     *
     * @see LyWebPage#writeBeginPage(Writer, String)
     */
    public void writeBeginPage(Writer pw, String frame)
        throws IOException
    {
        String controllerId = _tabs.getRootController();

        if ((controllerId + TAB).equals(frame))
        {
            LyTab tab = getSelectedTab();

            if (tab != null)
            {
                LyPanel        panel = tab.getPanel();
                LyWebComponent webComponent = (LyWebComponent)panel.getInternalView();

                if ((webComponent instanceof LyWebPageWriter)
                        && (((LyWebPageWriter)webComponent).isFrameSet()))
                    return;
            }
        }
        else if ((_randomTabsId + TABS).equals(frame))
        {
            pw.write("<!-- Generated by Leonardi (Lyria) -->\n");
            pw.write("<html><head>\n");
            pw.write("<meta name=\"AUTHOR\" content=\"Lyria\"/>\n");
            pw.write("<meta name=\"GENERATOR\" content=\"Leonardi\"/>\n");
    		pw.write("<meta name=\"Content-Type\" content=\"text/html;charset=" + LyString.DEFAULT_CHARSET + "\"/>\n");
            writeCssScripts(pw, frame);
            pw.write("</head>\n");
            pw.write("<body ");

            LyWebDecoration        deco = getWebViewManager().getDecoration();
            LyEnvironment          env = getEnvironment();
            LyViewManagerInterface manager = env.getViewManager();

            if (deco.isUsingDefaultBackground())
                pw.write("bgcolor='" + getBackgroundColor(frame) + "' ");
            pw.write("topmargin='3' leftmargin='0' " + getBodyScript() + ">\n");
            pw.write("<script type='text/javascript'>var pageWriterId='" + getId()
                + "';</script>\n");

            if (!env.testEnv("LY_REPLAY") && (manager.hasProperty("DYNAMIC")))
                pw.write("<script type='text/javascript'>var AJAX_MODE='" + (FORCE_AJAX ? "force" : true) + "';</script>\n");
            else
                pw.write("<script type='text/javascript'>var AJAX_MODE='false';</script>\n");

            // Manage Struts mode
            if (manager instanceof LyStrutsViewManager)
            {
                String  actionUrl = ((LyStrutsViewManager)manager).getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);

                pw.write("<script type='text/javascript'>var STRUTS_EVENT_URL='" + actionUrl + "';</script>\n");
            }

            pw.write("<script type='text/javascript'>var isInPopUp='" + isInPopUp()
                + "';</script>\n");
            return;
        }

        super.writeBeginPage(pw, frame);
    } // end method writeBeginPage

    /**
     * Writes the end of the page. Overriden to add the selection of a given tab on load.
     *
     * @see LyWebPage#writeEndPage(Writer, String)
     */
    public void writeEndPage(Writer pw, String frame)
        throws IOException
    {
        String controllerId = _tabs.getRootController();

        if ((controllerId + TAB).equals(frame))
        {
            LyTab tab = getSelectedTab();

            if (tab != null)
            {
                LyPanel        panel = tab.getPanel();
                LyWebComponent webComponent = (LyWebComponent)panel.getInternalView();

                if ((webComponent instanceof LyWebPageWriter)
                        && (((LyWebPageWriter)webComponent).isFrameSet()))
                    return;
            }
        }

        super.writeEndPage(pw, frame);
    }

    /**
     * Gets the selected tab in the current tabs view.
     *
     * @return the selected tab in the current tabs view
     * @see    #setSelectedTab(String)
     */
    public LyTab getSelectedTab()
    {
        // Write the buffers for areas
        int    i, nTabs = _tabs.getTabCount();
        LyTab  tab;
        String selectedTab = _tabs.getSelectedTab();

        if ((selectedTab != null) && (selectedTab.length() == 0))
            selectedTab = null;

        selectedTab = translate(selectedTab);

        boolean selectFirst = (selectedTab == null);
        String  label, id;

        for (i = 0; i < nTabs; i++)
        {
            tab = _tabs.getTab(i);
            label = translate(tab.getLabel());
            id = tab.getId();
            
            if ((tab.getLabel() == null) || (tab.getLabel().length() == 0))
                continue;
            
            if (((id != null) && id.equals(selectedTab))
                    || ((label != null) && label.equals(selectedTab))
                    || (selectFirst))
                return tab;
        }

        return ((nTabs > 0) ? _tabs.getTab(0) : null);
    }

    /**
     * Writes all tabs header.
     *
     * @param pw the page writer
     */
    protected void writeTabsHeader(LyWebPageWriter pw)
    {
        if (pw.isPrintMode())
            return;

        _lastHeaderFrameName = pw.getActiveName();
        
        if (pw.getActiveName() == LyWebPageWriter.MAIN_BUFFER)
        {
            _lastHeaderFrameName = LyWebConstants.TARGET_FRAME;
            
            if (LyWebConstants.POPUP_VIEW.equals(pw.getWindowName()))
                _lastHeaderFrameName = LyWebConstants.POPUP_VIEW;
        }

        LyTab   tab = null;
        LyTab   selectedTab = getSelectedTab();
        int     i, nTabs = _tabs.getTabCount();
        String  tabName;
        boolean previousShown = false;
        String  bgColor = getHexColor("LY_FORM_BACKGROUND");

        LyWebDecoration deco = getWebViewManager().getDecoration();

        if (deco.isUsingDefaultBackground())
        {
            if (bgColor != null)
                // Set the background color
                setBackgroundColor(bgColor, _randomTabsId + TABS);
        }

        pw.addBeginScript("<script type='text/javascript'>window.defaultStatus='';</script>",
            false);

        String  top = getImageUrl("LY_TAB_TOP");
        String  topOver = getImageUrl("LY_TAB_TOP_OVER");
        String  bottom = getImageUrl("LY_TAB_BOTTOM");
        String  both = getImageUrl("LY_TAB_BOTH");
        String  bothOver = getImageUrl("LY_TAB_BOTH_OVER");
        String  first = getImageUrl("LY_TAB_FIRST");
        String  firstSelected = getImageUrl("LY_TAB_FIRST_SELECTED");
        String  last = getImageUrl("LY_TAB_LAST");
        String  lastSelected = getImageUrl("LY_TAB_LAST_SELECTED");
        String  close = getImageUrl("LY_CLOSE_TAB");
        String  closeOver = getImageUrl("LY_CLOSE_TAB_OVER");
        String  leftImage, rightImage, middleImage, middleImageOver;
        String  url;
        String  controllerId, javascriptToolTip;
        boolean jump = false;
        boolean wrap = !"true".equalsIgnoreCase(TABS_WRAP)
        	&& !"arrows".equalsIgnoreCase(TABS_WRAP);

        if (!pw.writeDivOnly())
            pw.append("<div id='" + getIdCode() + "_tabs_header'>");

        pw.append(
            "<table id='tabs_headers_table_" + _tabs.getIdCode() + "' border='0' cellspacing='0' cellpadding='0' width='100%'>\n");
        pw.append("<tr height='1'>\n");
        pw.append("<td" + (wrap ? "" : " width='1'") + " class='");
        pw.append(deco.getClassName(_tabs, DECO_HEADERS_PARAMETER, pw));
        pw.append("' background='" + bottom + "'>\n");
        
        // Remove _removedTab from tab list
        Vector<LyTab>       tabs = new Vector<LyTab>(nTabs);
        
        for (i = 0; i < nTabs; i++)
        {
            tab = _tabs.getTab(i);

            if (tab != _removedTab)
                tabs.addElement(tab);
        }
        
        nTabs = tabs.size();
        
        // First pass in order to create the images for separation
        for (i = 0; i < tabs.size(); i++)
        {
            jump = false;
            tab = tabs.elementAt(i);
            tabName = translate(tab.getLabel());

            if ((tabName == null) || (tabName.length() == 0))
            {
                previousShown = false;
                continue;
            }

            if (!previousShown)
            {
                if (tab == selectedTab)
                    leftImage = firstSelected;
                else
                    leftImage = first;
            }
            else
            {
                if (tab == selectedTab)
                    leftImage = firstSelected;
                else
                    leftImage = first;
            }

            previousShown = true;

            if (tab == selectedTab)
            {
                middleImage = top;
                middleImageOver = topOver;
            }
            else
            {
                middleImage = both;
                middleImageOver = bothOver;
            }

            if (wrap)
                pw.append("<div style='float:left;'>");
            
            pw.append("<table border='0' cellspacing='0' cellpadding='0'><tr>");
            pw.append("<td>");
            pw.append("<img src='" + leftImage + "'");
            
            if (tab == selectedTab)
                pw.append(" id='selectedLeftImage_" + _tabs.getIdCode() + "'");

            pw.append("/></td>\n");
            pw.append("<td id='tabBg" + _tabs.getIdCode() + i + "' background='" + middleImage + "'");
            pw.append(" valign='middle' nowrap='true'");
            
            if (tab.isEnabled() && (tab != selectedTab))
            {
                pw.append(" onMouseOver=\"changeLineBg(document.getElementById('tabBg" + _tabs.getIdCode() + i + "'), \'");
                pw.append("\', \'" + middleImageOver + "\');;\"");
                pw.append(" onMouseOut=\"changeLineBg(document.getElementById('tabBg" + _tabs.getIdCode() + i + "'), \'");
                pw.append("\',\'" + middleImage + "\');\"");
            }

            pw.append(">\n");
            
            tabName = translate(tab.getLabel());

            if ((tabName != null) && (tabName.endsWith("\\n") || tabName.endsWith("\n")))
            {
                // Tab name ends with '\n' : jump to next line
                if (tabName.endsWith("\\n"))
                    tabName = tabName.substring(0, tabName.length() - 2);
                else if (tabName.endsWith("\n"))
                    tabName = tabName.substring(0, tabName.length() - 1);

                jump = true;
                previousShown = false;
            }

            tabName = getHtmlString(tabName);

            if ((tabName == null) || (tabName.length() == 0))
                continue;

            url = getTabURL(pw, i);
            controllerId = _tabs.getRootController();

            String      ctrlBis = controllerId;
            boolean     useTable = (tab.getImage() != null) || _tabs.isClosable();

            javascriptToolTip = LyString.getJavascriptString(translate("LY_SELECT_TAB") + " "
                    + tabName);
            
            if (useTable)
                pw.append("<table border='0' cellpadding='0' cellspacing='0'><tr><td nowrap='yes'>");

            if (tab.getImage() != null)
            {
                if ((tab != selectedTab) && tab.isEnabled())
                {
                    pw.append("<a class='" + deco.getClassName(_tabs, DECO_LABEL_PARAMETER, pw) +
                        "' href=\"javascript:SelectTab('" + url + "','" + i + "','"
                        + ctrlBis + "','" + pw.getId() + "','" + controllerId + TAB + "'");
                    if (tab.getTargetFormId() != null)
                        pw.append(",'" + tab.getTargetFormId() + "'");
                    else
                        pw.append(",null");
                    pw.append(",null");
                    pw.append(",'')\" onmouseover=\'window.status=\"");
                    pw.append(javascriptToolTip);
                    pw.append("\";return true\'>");
                }

                String  image = tab.getImage();
                int[]   size = getImageSize(image);
                
                image = getImageUrl(image);
                
                pw.append("<img src='" + image + "' border='0' ");

                if (size != null)
                {
                    if (size.length == 2)
                        pw.append("width='" + size[0] + "' height='" + size[1] + "'");
                    else
                        pw.append("height='" + size[0] + "'");

                }
                
                pw.append("/>");

                if ((tab != selectedTab) && tab.isEnabled())
                    pw.append("</a>");

                pw.append("</td><td nowrap='yes'>");
            }

            if ((tab != selectedTab) && tab.isEnabled())
            {
                pw.append("<a class='" + deco.getClassName(_tabs, DECO_LABEL_PARAMETER, pw) +
                    "' href=\"javascript:SelectTab('" + url + "','" + i + "','"
                    + ctrlBis + "','" + pw.getId() + "','" + controllerId + TAB + "'");
                if (tab.getTargetFormId() != null)
                    pw.append(",'" + tab.getTargetFormId() + "'");
                else
                    pw.append(",null");
                pw.append(",null");
                pw.append(",'')\" onmouseover=\'window.status=\"");
                pw.append(javascriptToolTip);
                pw.append("\";return true\'>");
            }
            
            int     classType = DECO_LABEL_PARAMETER;
            
            if (tab == selectedTab)
            {
                if (tab.isEnabled())
                    classType = DECO_SELECTED_LABEL_PARAMETER;
                else
                    classType = DECO_DISABLED_SELECTED_LABEL_PARAMETER;
            }
            else if (!tab.isEnabled())
                classType = DECO_DISABLED_LABEL_PARAMETER;
            
            pw.append("<font class='" + deco.getClassName(_tabs, classType, pw) + "'>");
            pw.append("<div id='tab_" + tab.getIdCode() + "'>&nbsp;" + tabName + "&nbsp;</div></font>");

            if ((tab != selectedTab) && tab.isEnabled())
                pw.append("</a>");

            if (_tabs.isClosable())
            {
                pw.append("</td><td>");
                pw.append("<a class='" + deco.getClassName(_tabs, DECO_LABEL_PARAMETER, pw));
                
                if (tabs.size() == 1)
                {
                    pw.append("' href=\"javascript:InvokeAction('_tabs', '" + url + "','target_frame', '" + controllerId +
                            "', '" + i + "', 'closeTab')\"");
                }
                else
                {
                    pw.append("' href=\"javascript:SelectTab('" + url + "','" + i + "','"
                            + ctrlBis + "','" + pw.getId() + "','" + controllerId + TAB + "'");
                    
                    if (tab.getTargetFormId() != null)
                        pw.append(",'" + tab.getTargetFormId() + "'");
                    else
                        pw.append(",null");
                    
                    pw.append(",null");
                    pw.append(",'closeTab')\"");
                }

                pw.append(" onmouseover=\'window.status=\"");
                pw.append(javascriptToolTip);
                pw.append("\";return true\'>");
                pw.append("<img src='" + close + "' onMouseOver=\"this.src='" + closeOver +
                        "'\" onMouseOut=\"this.src='" + close + "'\" border='0'/></a>");
            }
            
            if (useTable)
                pw.append("</td></tr></table>");

            pw.append("</td>\n");

            if (tab == selectedTab)
                rightImage = lastSelected;
            else
                rightImage = last;

            pw.append("<td>");
            pw.append("<img src='" + rightImage + "'");
            
            if (tab == selectedTab)
                pw.append(" id='selectedRightImage_" + _tabs.getIdCode() + "'");

            pw.append("/></td>\n");
            pw.append("</tr></table>");

            if (wrap)
                pw.append("</div>");

            if (jump)
            {
                pw.append("</td>");
                
                if (!wrap)
                {
                    pw.append("<td class='");
                    pw.append(deco.getClassName(_tabs, DECO_HEADERS_PARAMETER, pw));
                    pw.append("' background='" + bottom + "'>&nbsp;</td>");
                }

                pw.append("</tr>");
                pw.append("<tr height='1'>\n");
                pw.append("<td class='");
                pw.append(deco.getClassName(_tabs, DECO_HEADERS_PARAMETER, pw));
                pw.append("' background='" + bottom + "'>\n");
            }
            else if (!wrap)
            {
                pw.append("</td><td width='1' class='");
                pw.append(deco.getClassName(_tabs, DECO_HEADERS_PARAMETER, pw));
                pw.append("' background='" + bottom + "'>");
            }
        }

        pw.append("</td>\n");

        if (!wrap)
        {
            pw.append("<td class='");
            pw.append(deco.getClassName(_tabs, DECO_HEADERS_PARAMETER, pw));
            pw.append("' background='" + bottom + "'>&nbsp;</td>");
        }
        
        pw.append("</tr>\n");
        pw.append("</table>\n");

        if (!pw.writeDivOnly())
            pw.append("</div>");
        
        if (displayArrows(pw) && (tab == selectedTab))
        {
            StringBuffer    buffer = new StringBuffer();

            buffer.append("<script type='text/javascript'>");
            buffer.append("var frame = findTarget(null, '" + _randomTabsId + ARROWS + "', null);\n");
            buffer.append("if (frame != null)\n\tframe.setTimeout('scrollTabHeaders(true)', 300);\n");
            buffer.append("</script>");

            getWebViewManager().addAdditionalRefreshScripts(buffer.toString());
            pw.addEndScript(buffer.toString(), false);
        }
    } // end method writeTabsHeader
    
    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebPageWriter#clearBeginScripts()
     */
    public void clearBeginScripts()
    {
        super.clearBeginScripts();
        
        String controllerId = _tabs.getRootController();

        _beginScripts.put(_randomTabsId + TABS, new Vector<_LyScript>());
        _beginScripts.put(controllerId + TAB, new Vector<_LyScript>());
    }

    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebPageWriter#clearEndScripts()
     */
    public void clearEndScripts()
    {
        super.clearEndScripts();

        String controllerId = _tabs.getRootController();

        _endScripts.put(_randomTabsId + TABS, new Vector<_LyScript>());
        _endScripts.put(controllerId + TAB, new Vector<_LyScript>());
    }

    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebPageWriter#clearCssScripts()
     */
    public void clearCssScripts()
    {
        super.clearCssScripts();

        String controllerId = _tabs.getRootController();

        _cssScripts.put(_randomTabsId + TABS, new Vector<_LyCssScript>());
        _cssScripts.put(controllerId + TAB, new Vector<_LyCssScript>());
    }

    /**
     * Gets the key used for Hrefs of the tab at the given index.
     *
     * @param  i the index of the tab
     * @return the key used for hrefs
     */
    protected String getTabKey(int i)
    {
        int nTabs = _tabs.getTabCount();

        if (i >= nTabs)
            return null;

        LyTab  tab = _tabs.getTab(i);
        String tabKey = tab.getLabel();

        // Replace ' ' by '_' since blanks are not allowed in hrefs.
        tabKey = tabKey.replace(' ', '_');
        return tabKey;
    }

    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebPageWriter#isFrameSet()
     */
    public boolean isFrameSet()
    {
        return (!_noFrames);
    }

    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebPageWriter#getPageWriters()
     */
    public Vector<LyWebPageWriter> getPageWriters()
    {
        // Don't know yet how to use a frameset in tabs...
        Vector<LyWebPageWriter> v = new Vector<LyWebPageWriter>(1);

        v.addElement(this);
        return v;
    }

    /*
     * (non-Javadoc)
     * @see leon.view.web.LyWebPageWriter#setForceReloadState(boolean)
     */
    public void setForceReloadState(boolean force)
    {
        int            i, nElements = _tabs.getTabCount();
        LyTab          tab;
        LyPanel        panel;
        LyWebComponent internal;

        _forceReloadState = force;

        // if force is true, we set the state of all the sub pageWriter
        if (force)
        {
            for (i = 0; i < nElements; i++)
            {
                tab = _tabs.getTab(i);
                panel = tab.getPanel();
                internal = (LyWebComponent)panel.getInternalView();

                if (internal instanceof LyWebPageWriter)
                    ((LyWebPageWriter)internal).setForceReloadState(true);
            }
        }
    }

    /**
     * Sets force reload to <code>true</code>.
     *
     * @see LyGraphicalTabs#addTab(LyTab, int)
     * @see LyStrutsTabs#setForceReloadState(boolean)
     */
    public void addTab(LyTab tab, int pos)
    {
        refreshFrames(tab == getSelectedTab());
    }

    /**
     * Sets force reload to <code>true</code>.
     *
     * @see LyGraphicalTabs#removeTab(LyTab)
     * @see LyStrutsTabs#setForceReloadState(boolean)
     */
    public void removeTab(LyTab tab)
    {
        _removedTab = tab;
        refreshFrames(tab == getSelectedTab());
        _removedTab = null;
    }

    /*
     * (non-Javadoc)
     * @see leon.view.setinterface.LySetTabInterface#setEnabled(java.lang.String,
     *      boolean)
     */
    public void setEnabled(String tabId, boolean enabled)
    {
        if (tabId == null)
            return;
            // Find corresponding tab

        tabId = translate(tabId);

        int     nbTab = _tabs.getTabCount();
        LyTab   tab;
        int     tabIndex = -1;
        String  label, id;

        for (int i = 0; i < nbTab; i++)
        {
            tab = _tabs.getTab(i);
            label = translate(tab.getLabel());
            id = tab.getId();
            
            if (tabId.equals(id)
                    || tabId.equals(label))
            {
                tabIndex = i;
                break;
            }
        }

        if (tabIndex < 0)
            return;

        if (_enabledTab[tabIndex] == enabled)
            return;

        _enabledTab[tabIndex] = enabled;

        refreshFrames(false);
    } // end method setEnabled
    
    /**
     * This method allows to change the label of a specific tab.
     *
     * @param tabId the label or identifier of the tab to be changed
     * @param newLabel the new label of the tab to be changed
     */
    public void setLabel(String tabId, String newLabel)
    {
        LyTab   tab = getTab(tabId);
        
        if (tab == null)
            return;
        
        LyWebViewManager viewManager = getWebViewManager();
        StringBuffer     script = new StringBuffer();
        String           label = getHtmlString(newLabel);

        label = "&nbsp;" + label + "&nbsp;";
        
        script.append("<script type='text/javascript'>\n//<![CDATA[\nreplaceDivInnerHtml(\"");
        script.append(_lastFrameName).append("\",\"");
        script.append("tab_").append(tab.getIdCode()).append("\",\"");
        script.append(label);
        script.append("\");\n//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(script.toString());
    }
    
    /**
     * Retrieve a tab from its id or label.
     *
     * @param tabId The id or label of the tab that must be retrieved.
     */
    protected LyTab getTab(String tabId)
    {
        if (tabId == null)
            return null;

        tabId = translate(tabId);

        int     nbTab = _tabs.getTabCount();
        LyTab   tab;
        String  label, id;

        for (int i = 0; i < nbTab; i++)
        {
            tab = _tabs.getTab(i);
            label = translate(tab.getLabel());
            id = tab.getId();
            
            if (tabId.equals(id)
                    || tabId.equals(label))
                return tab;
        }

        return null;
    }
} // end class LyStrutsTabs
