/*
 * LyStrutsTool.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.view.web.struts;

import leon.misc.*;
import leon.view.*;
import leon.view.graphic.*;
import leon.view.web.*;


/**
 * This class implements the Tool object (as described in the dtd) for struts viewer.
 *
 * <p><b>TOOL: Action Button</b><br>
 * An action button allows executing an action. In the case where the button contains an {@link
 * leon.view.LyArea area (AREA)}, this area is displayed in a new window when launching the action.
 * </p>
 *
 * @version $Id: LyStrutsTool.java 35032 2012-03-23 17:17:16Z jlsaouli $
 * @author  Lyria S.A.
 */
public class LyStrutsTool extends LyWebObject
    implements LyGraphicalTool
{
	public static final String __VERSION =
		"$Id: LyStrutsTool.java 35032 2012-03-23 17:17:16Z jlsaouli $";

	/**
     * Indicates whether an hyperlink element must be generated around tool.
     * This hyperlink is used to launch tool action with the keyboard by using tab and Enter keys.
     */
    public static boolean      GENERATE_HREF = false;
    
    /** The tool generic component. */
    protected LyTool           _tool;

    /** Kept for optimization. */
    protected LyWebPageWriter  _webDialog;
    
	/**
	 * Constructor.
	 *
	 * @param tool the tool component
	 */
	public LyStrutsTool(LyTool tool)
	{
        super(tool);
        
        _tool = tool;
        initialize();
	}

	/**
	 * Method overriden to get the application context URL.
	 *
	 * @see LyWebViewManager#getRootURL()
	 */
	protected String getBaseUrl(LyWebPageWriter pw)
	{
		// get the application context url.
		return ((LyWebViewManager)getViewManager()).getRootURL();
	}

	/**
	 * Indicates whether a specific case is made for print items.
	 */
	protected boolean managePrintItem()
	{
		return false;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.LyStrutsTool#getHRefUrl(leon.view.web.LyWebPageWriter)
	 */
	protected String getHRefUrl(LyWebPageWriter pw)
	{
		// get the url to the event handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);

		return actionUrl.toString();
	} // end method getHRefUrl
	
	   /**
     * Initialization method.<br>
     * Sets the font name and size with default values or with values defined in resources
     * (LY_BUTTON_FONT) Sets the tool font background color with resource value
     * (LY_TOOL_LABEL_COLOR).<br>
     * Sets the on mouse over label color with the value defined in resources files
     * (LY_TOOL_LABEL_HIGHLIGHT_COLOR).
     */
    public void initialize()
    {
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
     */
    public void writeHtml(LyWebPageWriter pw)
    {
        writeHtml(pw, true);
    }

    /**
     * Write HTML description corresponding to this tool.
     *
     * @param pw       the page writer.
     * @param writeDiv indicates whether div tags must be generated or not.
     * @see   writeHtml(leon.view.web.LyWebPageWriter)
     */
    public void writeHtml(LyWebPageWriter pw, boolean writeDiv)
    {
        pw.startComponent(this);
		
		LyToolBar		toolbar = (_tool.getParent() instanceof LyToolBar ? (LyToolBar)_tool.getParent() : null);
		LyStrutsToolBar	webToolbar = (toolbar != null ? (LyStrutsToolBar)toolbar.getInternalView() : null);
		boolean			isWrap = (webToolbar != null ? webToolbar.isWrap() : true);
		LyWebDecoration	deco = getWebViewManager().getDecoration();
        String          toolClass = deco.getClassName(_tool, pw);
        String          controllerId = _tool.getRootController();
        String          label = translate(_tool.getLabel());
        String          tooltip = translate(_tool.getTooltip());
        String          javascriptTooltip = null;
        String          url = getHRefUrl(pw);
        boolean         writeArea = false;
        boolean         isVertical = (webToolbar != null ? !webToolbar.isHorizontal() : false);

        label = getHtmlString(label);

        if ((tooltip == null) || (tooltip.length() == 0))
            tooltip = label;
        
        javascriptTooltip = LyString.getJavascriptString(tooltip);
		
		String	onMouseOverFunctions = _tool.getEventFunctions(LyComponent.EventType.MOUSE_ENTERED, ';');
		String	onMouseOutFunctions = _tool.getEventFunctions(LyComponent.EventType.MOUSE_EXITED, ';');

        // If parent area has no frames, tool must be activated anyway (since
        // frame may not be reload when selecting an object).
        LyArea          area = (_tool.getComponent() instanceof LyArea ? (LyArea)_tool.getComponent() : null);
        LyWebPageWriter webArea = null;

        if (area != null)
        {
            webArea = (LyWebPageWriter)area.getInternalView();
            // Write the area if no multi-frame : we can't write a frameset
            // since there is no call to the server side when opening popups.
            writeArea = !webArea.isFrameSet();
        }

        // The name of the var used is popupText_ + a unique number.
        // This is done to make the difference between all the vars
        // that may be created.
        String      popupText = null;
        LyDialog    dialog = null;

        if (writeArea && (!containsToolBar() || !webToolbar.isExpandable()))
        {
            if (_webDialog == null)
                dialog = new LyDialog(LyWebConstants.TOOL_POPUP, getEnvironment());
            else
                dialog = (LyDialog)_webDialog.getViewComponent();
            
            popupText = "popupText_" + dialog.getIdCode();
        }

        if (!isWrap)
        {
            pw.append("<td");
            
            if ((_tool.getWidth() > 0) || (_tool.getHeight() > 0))
            {
                if (_tool.getWidth() > 0)
                    pw.append(" align='center'");

                pw.append(" style='");

                if (_tool.getWidth() > 0)
                    pw.append("width:" + _tool.getWidth() + "px;");

                if (_tool.getHeight() > 0)
                    pw.append("height:" + _tool.getHeight() + "px;");

                pw.append("'");
            }

            pw.append(" nowrap='true' class='" + toolClass + "'");
            
            if (!isVertical
                    && !testResource("LY_TOOL_LABEL_RIGHT_POSITION"))
                pw.append(" valign='top'");
            
            pw.append(" onMouseOver=\'window.status=\"");
            pw.append(label);
            pw.append("\";");
    
            if (_tool.isEnabled())
                pw.append("this.className=\"toolOver\";");
            
            pw.append("return true;\' ");
            
            if (_tool.isEnabled())
                pw.append("onMouseOut=\"this.className='" + toolClass + "'\" ");
            
            if (isEnabled())
                writeHRef(pw, LyWebObject.encodeUrl(url), controllerId, javascriptTooltip, false, false, popupText);
            
            pw.append(">");
        }
        
        if (writeDiv && !isWrap)
        {
            pw.append("<div id=\"" + "_" + _tool.getIdCode() + "\"");
			
			if (!_tool.isDefaultColor())
				pw.append(" style='background-color:" + getHexColor(_tool.getColor()) + ";'");
            
            if (_tool.isPressed())
                pw.append(" class=\"pressedTool\"");
            
            pw.append(">\n");
        }

        updateRefreshInfo(pw);

        String           baseUrl = getBaseUrl(pw);
        String           path;
        int[]            size;
        String           imageSize = null;
        boolean          isEnabled = _tool.isEnabled();

        if (isEnabled)
            path = getImageUrl(_tool.getImage());
        else
            path = getDisabledImage();

        if (containsToolBar() && webToolbar.isExpandable())
        {
            String image = null;

            if (area.getToolBar(0).isOpen())
                image = "LY_ARROW_DOWN_BIG";
            else
                image = "LY_ARROW_RIGHT_BIG";

            size = getImageSize(image, true);
            path = getImageUrl(image);
        }
        else
            size = getImageSize(_tool.getImage(), true);

        if (_tool.getImage() == null)
            path = null;
        
        if (size != null)
        {
            if (size.length == 2)
                imageSize = "width='" + size[0] + "' height='" + size[1] + "'";
            else
                imageSize = "height='" + size[0] + "'";
        }

        // If small dashboard, use small icons
        if (isVertical
                && !webToolbar.isExpandable())
            imageSize = "width='20' height='20'";

        // if the action opens a dialog, append "..."
        if (_tool.isDialog())
        {
            if ((tooltip != null) && (tooltip.length() > 0))
                tooltip += "...";
        }

        int             border = 0;
        boolean         writeImage = true;
        boolean         writeLabel = true;
        boolean         generateTable = false;
        boolean         labelRightPosition = testResource("LY_TOOL_LABEL_RIGHT_POSITION");

        if (_tool.getImage() == null)
            writeImage = false;
        
        if (!testResource("LY_WRITE_TOOL_LABEL") && writeImage)
            writeLabel = false;

        if (!isWrap
            && ((writeArea
                    && (!containsToolBar()
                        || !webToolbar.isExpandable()))
                || (writeImage
                    && writeLabel
                    && ((_tool.getWidth() > 0)
                        || !labelRightPosition))))
            generateTable = true;

        if (!isVertical && !isWrap)
            pw.append("<center>");

        pw.addBeginScript("<script type='text/javascript'>\n//<![CDATA[\nvar _rootUrl='" + baseUrl
            + "';\n//]]>\n</script>", false);
        pw.addBeginScript("<script type='text/javascript'>var _controllerId='" + controllerId
            + "';</script>", false);
        pw.addBeginScript("<script type='text/javascript'>window.defaultStatus='';</script>",
            false);
        
        // Generates an hyperlink used to launch tool action with the keyboard by using tab
        // and Enter keys.
        if (isEnabled() && GENERATE_HREF)
        {
			pw.append("<a class='toolLabel' href=\"javascript:EmptyFunction();\"");

			if (!_tool.isDefaultFont() || !_tool.isDefaultFontColor())
				pw.append(" style='" + getFontString(_tool.getFont(), _tool.getFontColor()) + ";'");

			pw.append(">");
            pw.addBeginScript("<script type='text/javascript'>\nfunction EmptyFunction() {\n}\n</script>", false);
        }

        if (generateTable)
        {
            pw.append("<table cellspacing='0' cellpadding='0' border='0' width='100%'");

            if (writeArea)
                pw.append(" id='" + popupText + "_anchor'");

			if (onMouseOverFunctions.length() > 0)
				pw.append(" onmouseover=\"" + onMouseOverFunctions + "\"");
			
			if (onMouseOutFunctions.length() > 0)
				pw.append(" onmouseout=\"" + onMouseOutFunctions + "\"");

            pw.append("><tr>");
        }
        
        if (writeImage)
        {
            if (generateTable)
            {
                pw.append("<td width='1%'");
                if (!labelRightPosition)
                    pw.append(" align='center'");
                pw.append(">");
            }
            
            if (isEnabled)
            {
                border = 0;

                if (isWrap)
                    writeHRef(pw, LyWebObject.encodeUrl(url), controllerId, javascriptTooltip, true, false, popupText);
            }

            if (path != null)
            {
                pw.append("<img src='");
                pw.append(path);
                pw.append("' ");

                if (writeArea && !generateTable)
                    pw.append("id='" + popupText + "_anchor' ");
                
                if (!isEnabled)
                    pw.append("class='disabledTool' ");

                pw.append("title=\"");
                pw.append(tooltip);
                pw.append("\" ");

                // If there is an expandable toolbar we center the image.
                if (containsToolBar() && webToolbar.isExpandable())
                    pw.append("valign='middle'");
                else
                    pw.append("style='vertical-align:middle'");

                if (imageSize != null)
                {
                    pw.append(" ");
                    pw.append(imageSize);
                }

                pw.append(" border='");
                pw.append(Integer.toString(border));
                pw.append("'/>");
            }

            if (isWrap && isEnabled && (isVertical))
                pw.append("</a>");

            if (generateTable)
            {
                pw.append("</td>");
                
                if (!labelRightPosition
                        && (!writeArea || (containsToolBar() && webToolbar.isExpandable())))
                    pw.append("</tr><tr>");
            }
        }

        StringBuffer scriptBuffer = new StringBuffer();
        
        if (writeArea && ( !containsToolBar() || !webToolbar.isExpandable()))
        {
            String       openButton = getImageUrl("LY_ARROW_RIGHT_BIG");

            if (containsToolBar() && webToolbar.isExpandable() && isEnabled)
            {
                LyToolBar       subToolbar = area.getToolBar(0);
                
                // If tool with a subtoolbar, write label
                writeLabel = true;

                if (subToolbar.isOpen())
                    openButton = getImageUrl("LY_ARROW_DOWN_BIG");
            }
            else
            {
                openButton = getImageUrl("LY_ARROW_DOWN_BIG");
            
                // Create a web dialog to write the area. Then insert the
                // output produced in a javascript var used when popup
                // button is pressed.
                if (_webDialog == null)
                {
                    dialog.setArea(area);
                    _webDialog = (LyWebPageWriter)dialog.getInternalView();

                    // Do not register on view manager (dummy view)
                    getWebViewManager().removePageWriter(_webDialog);
                }

                // Avoid repetition of area in the popup
                _webDialog.clearBuffer(LyWebConstants.TOOL_POPUP);
                _webDialog.setActiveBuffer(LyWebConstants.TOOL_POPUP);
                webArea.writeHtml(_webDialog);

                scriptBuffer = _webDialog.getBuffer(LyWebConstants.TOOL_POPUP);

                // Include css generated in the dialog buffer
                pw.addCssScript(_webDialog.getCssScriptsAsHtml(LyWebConstants.TOOL_POPUP, false)
                        .toString(), false);

                // Update view (area must be replugged with its tool)
                _tool.setComponent(area);
            }

            int rowspan = 1;
            
            if (writeImage && writeLabel && !labelRightPosition)
                rowspan = 2;
            
            if (generateTable)
                pw.append("<td valign='middle' rowspan='" + rowspan + "'>");

            pw.append("<img src='");
            pw.append(openButton);
            pw.append("' align='middle' border='0'/>");

            if (generateTable)
                pw.append("</td>");
        }
        
        if (writeImage && generateTable)
        {
            if (!labelRightPosition
                    && (writeArea && (!containsToolBar() || !webToolbar.isExpandable())))
                pw.append("</tr><tr>");
        }

        if (isEnabled && writeLabel)
        {
            if (isWrap && writeImage && !isVertical)
                pw.append("</a>");

            if (generateTable)
                pw.append("<td align='center'>");
            
            if ((label != null) && (label.length() > 0))
                pw.append(" ");

            // FFT01488
            if (isWrap)
                writeHRef(pw, LyWebObject.encodeUrl(url), controllerId, javascriptTooltip, true, true, popupText);
            else
            {
    			pw.append("<font class='toolLabel'");
    			
    			if (!_tool.isDefaultFont() || !_tool.isDefaultFontColor())
    				pw.append(" style='" + getFontString(_tool.getFont(), _tool.getFontColor()) + ";'");

    			pw.append(">");
            }
            
            pw.append(label);

            if ((label != null) && (label.length() > 0))
                pw.append(" ");

            if (isWrap)
                pw.append("</a>");
            else
                pw.append("</font>");

            if (generateTable)
                pw.append("</td>");
        }
        else if (isWrap && isEnabled && !isVertical)
            pw.append("</a>");
        else if (!isEnabled && writeLabel)
        {
            // FFT01486
            // if (writeImage && !isVertical && !rightLabel)
            // pw.append("<br />");
            
            if (generateTable)
                pw.append("<td align='center'>");

            //if (!containsToolBar())
            if ((label != null) && (label.length() > 0))
                pw.append(" ");

			pw.append("<font class='toolLabelDisabled'");
			
			if (!_tool.isDefaultFont() || !_tool.isDefaultFontColor())
				pw.append(" style='" + getFontString(_tool.getFont(), _tool.getFontColor()) + ";'");

			pw.append(">");
            pw.append(label);
            
            if ((label != null) && (label.length() > 0))
                pw.append(" ");

            pw.append("</font>");

            if (generateTable)
                pw.append("</td>");
        }

        if (generateTable)
            pw.append("</tr></table>");
        
        if (isEnabled() && GENERATE_HREF)
            pw.append("</a>");
        
        if (!isVertical && !isWrap)
            pw.append("</center>");

        if (containsToolBar() && (webToolbar.isExpandable()) && isEnabled)
        {
            LyToolBar      subToolbar = area.getToolBar(0);
            LyStrutsToolBar subWebToolbar = (LyStrutsToolBar)subToolbar.getInternalView();

            if (subToolbar.isOpen())
            {
                if ((subToolbar.getFill() == LyToolBar.Fill.HORIZONTAL) && !isVertical)
                    pw.append("</td></tr><tr><td></td><td>\n");
                else
                    pw.append("</td></tr><tr><td>\n");

                subWebToolbar.writeHtml(pw);
            }
        }

        if (writeArea && (!containsToolBar() || !webToolbar.isExpandable()))
        {
            pw.append("<div class='toolPopup' id='" + popupText + "' style='position:absolute;top:0px;left:0px;z-index:100;'>");
            pw.append(scriptBuffer.toString());
            pw.append("</div>\n");
        }

        if (writeDiv && !isWrap)
            pw.append("</div>\n");
        
        if (!isWrap)
            pw.append("</td>\n");
        
        pw.endComponent(this);
    } // end method writeHtml

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebObject#refresh()
     */
    public void refresh()
    {
        LyToolBar        toolbar = (LyToolBar)_tool.getParent();
        LyStrutsToolBar   webToolbar = (LyStrutsToolBar)toolbar.getInternalView();

        if (webToolbar != null)
            webToolbar.refresh();
    } // end method refresh

    /**
     * Gets the disable image of the tool image.
     *
     * @return the disabled tool image
     * @see    leon.view.web.LyWebObject#getDisabledImageUrl(String key)
     */
    protected String getDisabledImage()
    {
        return getDisabledImageUrl(_tool.getImage());
    }

    /**
     * Indicates if the tool is enabele or not.
     *
     * @return true if the tool is enable, false otherwise.
     */
    public boolean isEnabled()
    {
        return _tool.isEnabled();
    }

    /**
     * Writes the HTML tag of the tool link.
     *
     * @param pw           the page writer
     * @param url          the link URL
     * @param controllerId the controller ID
     * @param label        the window status when mouse is over the tool
     * @param isHref       indicates whether a hyperlink or a onClick script should be written
     */
    protected void writeHRef(LyWebPageWriter pw, String url, String controllerId, String label,
            boolean isHref, boolean isLabel, String popupText)
    {
        String toolId = _tool.getId();
        String link = (isHref ? (isLabel ? "<a class='toolLabel' hRef=" : "<a hRef=") : " onClick=");

        if (popupText != null)
        {
            pw.append(link);
            pw.append("\"javascript:SetDivAnchor('" + popupText + "', '" + popupText + "_anchor');SwitchDivVisibility('" + popupText + "')\"");
        }
        else if (managePrintItem()
                && toolId.startsWith("_print") && !toolId.startsWith("_printXSLFO")
                && !toolId.startsWith("_print" + LyWebConstants.CONTROLLER_ID_SEP)
                && !toolId.equals("_print"))
        {
            pw.append(link);
            pw.append("\"javascript:InvokePrint('");
            pw.append(LyWebObject.encodeUrl(url.toString()));
            pw.append("',null,'");
            pw.append(_tool.getParentArea().getId());
            pw.append("','");
            pw.append(toolId);
            pw.append("')\"");
        }
        else
        {
            String          target = "'" + LyWebConstants.RELOAD_FRAMES + "'";
            LyToolBar       toolbar = (LyToolBar)_tool.getParent();
            LyStrutsToolBar webToolbar = (LyStrutsToolBar)toolbar.getInternalView();
            boolean         expand = false;

            if (containsToolBar() && webToolbar.isExpandable())
                expand = true;

            if (!expand)
            {
                // Tool may be in another form with a controller id different
                // from the root controller id.
                // In this case, we change the controller id used when calling
                // invokeAction in order to
                // retrieve root form. Controller id will then be changed in
                // InvokeAction Javascript function.
                boolean changeControllerId = false;
                LyArea  parentArea = _tool.getParentArea();

                if (parentArea != null)
                {
                    if (parentArea.getParentFormFieldId() != null)
                        changeControllerId = true;
                }

                pw.append(link);
                pw.append("\"");

                if (_tool.getStyle() == LyTool.Style.TOGGLE)
                {
                    if (_tool.isPressed())
                        pw.append("javascript:ChangeObjectClass('_" + _tool.getIdCode() + "', '');");
                    else
                        pw.append("javascript:ChangeObjectClass('_" + _tool.getIdCode() + "', 'pressedTool');");
                }
                
                pw.append("javascript:InvokeAction('");
                pw.append(LyString.replace(toolId, "\\", "\\\\"));
                pw.append("','");
                pw.append(LyWebObject.encodeUrl(url.toString()));
                pw.append("',");
                pw.append(target);

                if (changeControllerId)
                    pw.append(",'" + _tool.getRootController() + "_CONTROLLER_CHANGED_ID')\"");
                else
                    pw.append(",'" + _tool.getRootController() + "')\"");
            }
            else
            {
                LyArea          area = (LyArea)_tool.getComponent();
                LyToolBar       subToolbar = area.getToolBar(0);
                String          script;
                
                webToolbar = (LyStrutsToolBar)subToolbar.getInternalView();
                script = webToolbar.getOpenToolbarScript(pw);

                pw.append(link);
                pw.append("\"javascript:");
                pw.append(script);
                pw.append("\"");
            }
        }
        
        if (isHref)
        {
            pw.append(" onmouseover='window.status=\"");
            pw.append(label);
            pw.append("\";return true;'>");
        }
    } // end method writeHRef

    /**
     * Writes the HTML tag of the tool pop up link.
     *
     * @param pw                the page writer
     * @param controllerId      the controller ID
     * @param popupText         text written in the popup
     * @param popupSize         the size of the popup window [width, height]
     * @param javascriptTooltip the window status when mouse is over the tool
     * @deprecated
     */
    protected void writePopupRef(LyWebPageWriter pw, String controllerId, String popupText,
        int[] popupSize, String javascriptTooltip)
    {
        pw.append("<a class='toolLabel' href=\"javascript:SwitchDivVisibility('" + popupText
            + "')\"");
        pw.append(" onmouseover=\'window.status=\"");
        pw.append(javascriptTooltip);
        pw.append("\";return true\'>");
    }

    /**
     * Indicates if the tool contains an area which contains only a toolbar or not.
     *
     * @return true if the tool contains an area which contains only a toolbar, false otherwise
     */
    protected boolean containsToolBar()
    {
        LyComponent component = _tool.getComponent();

        if ((component instanceof LyArea)
                && (((LyArea)component).getPanelCount() == 0)
                && (((LyArea)component).getToolBarCount() > 0))
            return true;
        else
            return false;
    }

    /**
     * Gets the popup size from the toolbar of the given area.
     *
     * @param  area the area to show in popup
     * @return the popup size
     * @see    LyStrutsToolBar#computeSize()
     */
    protected int[] computePopupSize(LyArea area)
    {
        // Check if only the area contains only a toolbar
        int       nPanel = area.getPanelCount();
        LyToolBar toolbar = area.getToolBar();

        if ((nPanel != 0) || (toolbar == null))
            return null;

        LyStrutsToolBar webToolbar = (LyStrutsToolBar)toolbar.getInternalView();

        return webToolbar.computeSize();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
     */
    public void setLabel(String label)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetEnabledInterface#setEnabled(boolean)
     */
    public void setEnabled(boolean status)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetImageInterface#setImage(java.lang.String)
     */
    public void setImage(String key)
    {
        refresh();
    }

    /**
     * Updates attribute pressed to the given value.
     * 
     * @param pressed Boolean indicating if attribute pressed is set or unset.
     */
    public void setPressed(boolean pressed)
    {
        String           id = "_" + _tool.getIdCode();
        LyWebViewManager viewManager = getWebViewManager();
        StringBuffer     buffer = new StringBuffer();

        buffer.append("<script type='text/javascript'>\n//<![CDATA[\nChangeObjectClass('").append(id).
            append("',\"").append(pressed ? "pressedTool" : "").append("\");\n//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(buffer.toString());
    }

    /**
     * Sets the tool tooltip.
     *
     * @param tooltip the new tooltip
     */
    public void setTooltip(String tooltip)
    {
        refresh();
    }
} // end class LyStrutsTool
