/*
 * LyStrutsToolBar.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.view.web.struts;

import leon.view.*;
import leon.view.graphic.*;
import leon.view.web.*;


/**
 * This class implements the ToolBar object (as described in the dtd) for web viewer.
 *
 * <p><b>TOOLBAR: Tool Bar</b><br>
 * A tool bar allows displaying a set of action buttons. A tool bar can itself contain tool bars,
 * which allows to group together the action buttons into logical groups and sub-groups.</p>
 *
 * @version $Id: LyStrutsToolBar.java 35068 2012-03-27 17:18:53Z jlsaouli $
 * @author  Lyria S.A.
 */
public class LyStrutsToolBar extends LyWebObject
    implements LyGraphicalToolBar
{
	public static final String __VERSION =
		"$Id: LyStrutsToolBar.java 35068 2012-03-27 17:18:53Z jlsaouli $";
	
    /** Indicates if dashboards must be horizontal. */
    public static boolean	   HORIZONTAL_DASHBOARD = false;
    
    /** Indicates if dashboards may be expanded. */
    public static boolean	   EXPAND_DASHBOARD = true;
    
    /** Indicates if labels in the dashboards must be nowrap. */
    public static boolean	   LABEL_WRAP = false;
    
    /** The tool bar generic component. */
    protected LyToolBar        _toolBar;

    /** Specific style of a toolbar used in dashboards. */
    protected boolean          _dashboardSpecificStyle;
    
	/**
	 * Constructor.
	 *
	 * @param toolBar the tool bar component
	 */
	public LyStrutsToolBar(LyToolBar toolBar)
	{
        super(toolBar);
        
        _toolBar = toolBar;
        _dashboardSpecificStyle = false;
	}

	/**
	 * Overriden to redirect open tolbar script to the graphical event handler action.
	 *
	 * @param  pw the current page writer.
	 * @return The javascript for opening a toolbar.
	 */
	protected String getOpenToolbarScript(LyWebPageWriter pw)
	{
		// get toolbar infos
		String			    controllerId = _toolBar.getRootController();

		// get the url to the event handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);

		// build the javascript
		StringBuffer	    script = new StringBuffer();

		script.append("InvokeAction('");
		script.append(_toolBar.getId());
		script.append("', '");
		script.append(actionUrl);
		script.append("', null, '");
		script.append(controllerId);
		script.append("', '");
		script.append(!_toolBar.isOpen());
		script.append("', null)");

		return script.toString();
	} // end method getOpenToolbarScript

	/**
	 * Method overriden to get the application context URL.
	 *
	 * @see LyWebViewManager#getRootURL()
	 */
	protected String getBaseUrl(LyWebPageWriter pw)
	{
		// get the application context url.
		return ((LyWebViewManager)getViewManager()).getRootURL();
	}
	
	   /**
     * Sets the specific style of a toolbar (used in dashboards).
     * 
     * @param specific true if the specific style must be set, false otherwise.
     */
    protected void setSpecificStyle(boolean specific)
    {
        _dashboardSpecificStyle = specific;
    }
    
    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
     */
    public void writeHtml(LyWebPageWriter pw)
    {
        LyComponent component = pw.getViewComponent();
        boolean     writeDiv = (component != _toolBar) || (!pw.writeDivOnly() && isWrap());

        updateRefreshInfo(pw);

        // Don't print toolbars
        if (pw.isPrintMode())
            return;

        pw.startComponent(this);
        
        LyToolBar.Style    oldStyle = _toolBar.getStyle();
        LyToolBar.Fill     oldFill = _toolBar.getFill();
        LyToolBar.HAlign   oldAlign = _toolBar.getHAlign();

        if (_dashboardSpecificStyle)
        {
            _toolBar.setStyle(LyToolBar.Style.DEFAULT);
            _toolBar.setFill(LyToolBar.Fill.DEFAULT);
            _toolBar.setHAlign(LyToolBar.HAlign.LEFT);
        }
        
        LyToolBar.Style style = _toolBar.getStyle();
        int             border = 0;
        String          frame = "box";
        String          baseUrl = getBaseUrl(pw);
        String          controllerId = _toolBar.getRootController();

        pw.addBeginScript("<script type='text/javascript'>\n//<![CDATA[\nvar _rootUrl='" + baseUrl
            + "';\n//]]>\n</script>", false);
        pw.addBeginScript("<script type='text/javascript'>var _controllerId='" + controllerId
            + "';</script>", false);
        pw.addBeginScript("<script type='text/javascript'>window.defaultStatus='';</script>", false);

        if (style == LyToolBar.Style.BORDER)
            border = 0;
        else if (style == LyToolBar.Style.SEPARATOR)
        {
            LyComponent parent = _toolBar.getParent();

            if (parent instanceof LyArea)
            {
                LyArea area = (LyArea)parent;

                if (area.getTitleBar() != null)
                    border = 0;
                    // frame = "hsides";
            }
        }

        // If vertical set background color with dashboard color
        String          bgColor = null;
        String          colorCode = _toolBar.getColor();
        LyWebDecoration deco = getWebViewManager().getDecoration();
		boolean			isSubExpandBar = false;
        boolean         horizontal = isHorizontal();
		
		if ((_toolBar.getStyle() == LyToolBar.Style.EXPAND)
		        && (_toolBar.getParent() instanceof LyToolBar))
			isSubExpandBar = true;
		
		// Workaround for backward compatibility
		if ((colorCode != null) && !_toolBar.isDefaultColor())
			bgColor = getHexColor(colorCode);

        // If small dashboard style and vertical, no border
        if (!horizontal)
            border = 0;
        
        if (writeDiv)
        {
            String  divStyle = "height:100%;";
            
            pw.append("<div id='_" + _toolBar.getIdCode() + "'");

            if ((_toolBar.getStyle() == LyToolBar.Style.EXPAND)
            	&& !(_toolBar.getParent() instanceof LyToolBar))
            {
                int dashboardSize = getIntResource("LY_DASHBOARD_SIZE");
                
                if (dashboardSize > 0)
                    divStyle += "width:" + dashboardSize + ";";
                
                pw.append(" class='expandBar'");
            }

            if ((bgColor != null) && !isSubExpandBar)
            	divStyle += "background-color:" + bgColor + ";";
            
            pw.append(" style='" + divStyle + "'");
            
            pw.append(">\n");
        }
        
        // cellpadding is 2 for horizontal and vertical
        pw.append("<table id='toolbar_" + _toolBar.getIdCode() + "'");

        String      width = " width='20'";
        int         borderBeginStyle = LyWebDecoration.AREA_BORDER_BEGIN_NOT_FULL_WIDTH_AND_HEIGHT;
        int         borderLeftStyle = LyWebDecoration.AREA_BORDER_LEFT_NOT_FULL_WIDTH_AND_HEIGHT;
        int         borderRightStyle = LyWebDecoration.AREA_BORDER_RIGHT;
        int         borderEndStyle = LyWebDecoration.AREA_BORDER_END;
        String      label = _toolBar.getLabel();

        if ((style == LyToolBar.Style.SEPARATOR)
            || isWrap())
        {
            width = " width='100%'";
            borderBeginStyle = LyWebDecoration.AREA_BORDER_BEGIN_NOT_FULL_HEIGHT;
            borderLeftStyle = LyWebDecoration.AREA_BORDER_LEFT_NOT_FULL_HEIGHT;
        }
        
        if (!horizontal)
        {
            width = " width='100%'";
            borderBeginStyle = LyWebDecoration.VERTICAL_TOOLBAR_BEGIN;
            borderLeftStyle = LyWebDecoration.VERTICAL_TOOLBAR_LEFT;
            borderRightStyle = LyWebDecoration.VERTICAL_TOOLBAR_RIGHT;
            borderEndStyle = LyWebDecoration.VERTICAL_TOOLBAR_END;
        }

        pw.append(" border='" + Integer.toString(border) + "'");
        
        if (style == LyToolBar.Style.SEPARATOR)
            pw.append(" cellpadding='0'");
        else
            pw.append(" cellpadding='1'");

        pw.append(" cellspacing='0' frame='" + frame + "'");
        pw.append(width);
        pw.append(">\n");

        // Use a second table for simulating separators instead of <HR> because
        // <HR> add too much space before and after the toolBar
        if (style == LyToolBar.Style.SEPARATOR)
        {
            pw.append("<tr");
            if (_toolBar.getHAlign() == LyToolBar.HAlign.LEFT)
                pw.append(" align='left'>\n");
            else
                pw.append(" align='right'>\n");
            pw.append("<td>\n<table cellpadding='1' cellspacing='0'>\n");
        }

        if ((label != null)
            && (label.length() > 0))
        {
			String	font = (_toolBar.isDefaultFont() ? null : _toolBar.getFont());
			String	fontColor = (_toolBar.isDefaultFontColor() ? null : _toolBar.getFontColor());
			String	titleStyle = null;
			String	labelString = "";
			
			if ((font != null) || (fontColor != null))
				titleStyle = getFontString(font, fontColor);
			
            String      fontClass = deco.getClassName(_toolBar, pw);
			
			labelString = "<font";
			
			if ((font == null) || (fontColor == null))
				labelString += " class='" + fontClass + "'";

			if (titleStyle != null)
				labelString += " style='" + titleStyle + "'";
			
			labelString += ">&nbsp;" + translate(label) + "&nbsp;</font>";
            
            pw.append("<tr><td>");
            
            if (isExpandable())
                deco.writePageDecoration(pw, borderBeginStyle,
                    new String[] {labelString, getOpenToolbarScript(pw), _toolBar.isOpen() + ""});
            else
                deco.writePageDecoration(pw, borderBeginStyle, labelString);
            
            deco.writePageDecoration(pw, borderLeftStyle);
            pw.append("<table" + width + " border='0'>");
        }
        
        if ((label == null)
            || (label.length() == 0)
            || !isExpandable()
            || _toolBar.isOpen())
            writeTools(pw);
        
        if ((label != null)
            && (label.length() > 0))
        {
            pw.append("</table>");

            deco.writePageDecoration(pw, borderRightStyle);
            deco.writePageDecoration(pw, borderEndStyle);
            pw.append("</td></tr>");
        }
        
        if (style == LyToolBar.Style.SEPARATOR)
            pw.append("</table>\n</td>\n</tr>\n\n");

        boolean          writeLifts = writeLifts();

        if (writeLifts)
        {
            String up = getImageUrl("LY_ARROW_UP_BIG");
            String upPopup = translate("LY_GO_TOP_PAGE");

            pw.append("<tr><td align='left'>&nbsp;");

            pw.addEndScript("<a name='_bottom'></a>", false);

            pw.append("<a href='#_top'");
            pw.append(" onmouseover='window.status=\"");
            pw.append(upPopup);
            pw.append("\";return true;'>");
            pw.append("<img src='");
            pw.append(up);
            pw.append("' border='0' title='");
            pw.append(upPopup);
            pw.append("'/>");
            pw.append("</a></td></tr>");
        }

        pw.append("</table>\n");

        if (writeDiv)
            pw.append("</div>\n");

        if (_dashboardSpecificStyle)
        {
            _toolBar.setStyle(oldStyle);
            _toolBar.setFill(oldFill);
            _toolBar.setHAlign(oldAlign);
        }
        
        pw.endComponent(this);
    } // end method writeHtml

    /**
     * Writes the HTML description corresponding to the tools.
     *
     * @param pw the page writer
     */
    protected void writeTools(LyWebPageWriter pw)
    {
        if (!isWrap())
        {
            String         imagePath = getImagePath();
            String         imageSize = getImageSize();
            int            i, nTools = _toolBar.getComponentCount();
            LyComponent    component;
            LyWebComponent internal;
            boolean        horizontal = isHorizontal();
            boolean        writeImage = true;

            if (horizontal)
            {
                pw.append("<tr>");

                // Add the toolBar image (if needed)
                if (imagePath != null)
                {
                    pw.append("<td align='center'><img src='");
                    pw.append(imagePath);
                    pw.append("'");
                    if (imageSize != null)
                    {
                        pw.append(" ");
                        pw.append(imageSize);
                    }
                    pw.append("/></td>\n");
                }

                // Then write the tools (toolBar or tool)
                String  separator = getImageUrl("LY_VERTICAL_SEPARATOR");
                
                for (i = 0; i < nTools; i++)
                {
                    component = _toolBar.getComponent(i);

                    if ((component == null))
                    {
                        // Separator
                        if (writeImage)
                        {
                            pw.append("<td nowrap='true'>&nbsp;<img src='");
                            pw.append(separator);
                            pw.append("' valign='middle'/>&nbsp;</td>\n");
                        }
                        else
                            pw.append("<td width='10'>&nbsp;/&nbsp;</td>\n");
                    }
                    else if (component != null)
                    {
                        internal = (LyWebComponent)component.getInternalView();

                        pw.append("<td nowrap='true'>");

                        if ((component instanceof LyFormField) && !((LyFormField)component).isHidden())
                            pw.setFormId(component.getRootController());

                        internal.writeHtml(pw);
                        
                        pw.append("</td>\n");
                    }
                }

                pw.append("</tr>\n");
            }
            else
            // Vertical toolBar
            {
                boolean         wrap = testResource("LY_DASHBOARD_LABEL_WRAP");
                boolean         indent = false;
                LyComponent     parent = _toolBar.getParent();
                
                while (!indent && (parent != null))
                {
                    if (parent instanceof LyTool)
                        indent = true;
                    else
                        parent = parent.getParent();
                }

                // If toolbar has an image and a top-right toolbar as its first component
                // use an internal table to display image and toolbar on a single line.
                boolean     hasImageandTopRightToolbar = false;
                
                if ((imagePath != null)
                        && (nTools > 0)
                        && (_toolBar.getComponent(0) instanceof LyToolBar)
                        && (((LyToolBar)_toolBar.getComponent(0)).getHAlign() == LyToolBar.HAlign.RIGHT)
                        && (((LyToolBar)_toolBar.getComponent(0)).getVAlign() == LyToolBar.VAlign.TOP))
                    hasImageandTopRightToolbar = true;
                
                // Add the toolBar image (if needed)
                if (imagePath != null)
                {
                    pw.append("<tr>");
                    
                    if (hasImageandTopRightToolbar)
                        pw.append("<td><table border='0' width='100%' cellpadding='0' cellspacing='0'>");
                    
                    pw.append("<td align='left'><img src=\"");
                    pw.append(imagePath);
                    pw.append("\"/></td>");
                    
                    if (hasImageandTopRightToolbar)
                    {
                        pw.append("<td align='right' valign='middle'><table border='0' cellpadding='0' cellspacing='0'><tr><td>");
                        
                        internal = (LyWebComponent)_toolBar.getComponent(0).getInternalView();
                        internal.writeHtml(pw);
                        
                        pw.append("</td></tr></table></td>");
                        pw.append("</table></td>");
                    }

                    pw.append("</tr>\n");
                }
                
                for (i = 0; i < nTools; i++)
                {
                    if (hasImageandTopRightToolbar && (i == 0))
                        continue;
                    
                    component = _toolBar.getComponent(i);

                    if (component == null)
                    {
                        // Separator
                        pw.append("<tr height='10'><td nowrap='true'><hr /></td></tr>\n");
                        continue;
                    }

                    internal = (LyWebComponent)component.getInternalView();

                    if ((component == null)
                        || (component.getId() == LyComponent.SEP_ID))
                    {
                        // Separator
                        pw.append("<tr height='10'><td nowrap='true'><hr /></td></tr>\n");
                        continue;
                    }
                    
                    pw.append("<tr>");

                    if (internal instanceof LyStrutsTool)
                    {
                        if (indent)
                            // Add indent for sub-toolbars
                            pw.append("<td>&nbsp;&nbsp;&nbsp;</td>");
                    }
                    else
                    {
                        pw.append("<td width='100%'");
                        
                        if (!wrap)
                            pw.append(" nowrap='true'>");
                    }

                    if (component instanceof LyFormField)
                        pw.setFormId(component.getRootController());

                    internal.writeHtml(pw);

                    if (!(internal instanceof LyStrutsTool))
                        pw.append("</td>\n");
                    
                    pw.append("</tr>\n");
                }
            }
            
            return;
        }

        int             i, nTools = _toolBar.getComponentCount();
        LyComponent     component;
        LyWebComponent  internal;
        pw.append("<tr>");

        // Then write the tools (toolBar or tool)
        String  separator = getImageUrl("LY_VERTICAL_SEPARATOR");

        pw.append("<td>\n");

        for (i = 0; i < nTools; i++)
        {
            component = _toolBar.getComponent(i);

            if ((component == null)
                || (component.getId() == LyComponent.SEP_ID))
            {
                // Separator
                pw.append("</td><td>&nbsp;<img src='");
                pw.append(separator);
                pw.append("'  align='middle'/>&nbsp;\n</td><td>");
            }
            else
            {
                internal = (LyWebComponent)component.getInternalView();

                if (component instanceof LyFormField)
                    pw.setFormId(component.getRootController());
                
                internal.writeHtml(pw);
            }
        }

        pw.append("</td>\n");
        pw.append("</tr>\n");
    } // end method writeTools

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetColorInterface#setColor(java.lang.String)
     */
    public void setColor(String color)
    {
        String           toolbarColor = ((color == null) ? "" : getHexColor(color));
        LyWebViewManager viewManager = getWebViewManager();
        StringBuffer     buffer = new StringBuffer();
        String           id = "toolbar_" + _toolBar.getIdCode();

        // Update the color of the toolbar
        buffer.append("<script type='text/javascript'>\n//<![CDATA[\nSetObjectColor('").append(
            _lastFrameName).append("','").append(id).append("','").append(toolbarColor).append(
            "');\n//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(buffer.toString());
    }

    /**
     * Indicates the toolbar has writes lifts or not.
     *
     * @return false if the toolbar is expandable or horizontal, if its parent is a toolbar, or if
     *         it doesn't contains a sub toolbar, true otherwise
     */
    protected boolean writeLifts()
    {
        if (isExpandable() 
                || (_toolBar.getParent() instanceof LyToolBar) 
                || isHorizontal())
            return false;

        int         i, nTools = _toolBar.getComponentCount();
        int         count = 0;
        LyComponent component;

        for (i = 0; i < nTools; i++)
        {
            component = _toolBar.getComponent(i);
            if (component instanceof LyToolBar)
                count++;
        }

        // If at least one sub toolbar write lifts
        return (count > 1);
    }

    /**
     * Indicates if the toolbar is expandable or not.
     *
     * @return false if dashboards can't be expanded id EXPAND_DASHBOARD is set to false
     * or if the toolbar is horizontal or if its parent is a form field, true otherwise.
     */
    public boolean isExpandable()
    {
        if (_toolBar.getParent() instanceof LyFormField)
            return false;
        if (isHorizontal()
                || !EXPAND_DASHBOARD)
            return false;

        return true;
    }

    /**
     * Indicates if the toolbar is horizontal or not.
     *
     * @return true if the toolbar is horizontal, false otherwise
     */
    public boolean isHorizontal()
    {
        return ((_toolBar.getFill() == LyToolBar.Fill.HORIZONTAL)
                || HORIZONTAL_DASHBOARD);
    }

    /**
     * Indicates if the toolbar is wrapped or not.
     *
     * @return true if the toolbar is wrapped, false otherwise.
     */
    public boolean isWrap()
    {
        int             i, nComponents = _toolBar.getComponentCount();
        LyComponent     component;
        
        for (i = 0; i < nComponents; i++)
        {
            component = _toolBar.getComponent(i);
            
            // If toolbar contains at least one field, toolbar may not be wrapped.
            if (component instanceof LyFormField)
                return false;
            // If toolbar contains at least one toggle button, toolbar may not be wrapped.
            else if ((component instanceof LyTool)
                    && (((LyTool)component).getStyle() == LyTool.Style.TOGGLE))
                return false;
        }
        
        return (isHorizontal()
                && !containsLabel()
                && (getImagePath() == null)
                && testResource("LY_WRITE_TOOL_LABEL")
                && testResource("LY_TOOL_LABEL_RIGHT_POSITION"));
    }

    /**
     * Gets the tooltip image URL.
     *
     * @return the tooltip image URL if there is an image, null otherwise
     */
    protected String getImagePath()
    {
        String image = _toolBar.getImage();

        return (((image == null) || (image.length() == 0)) ? null : getImageUrl(image));
    }

    /**
     * Gets the tooltip image size.
     *
     * @return the String representation of width and height HTML attributes<br>
     *         (format : "width='<code>image_width</code>' height='<code>image_height</code>'"<br>
     *         or "height='<code>image_height</code>'" if the width is not defined)
     */
    protected String getImageSize()
    {
        String image = _toolBar.getImage();

        if (image != null)
            return null;

        int[] size = getImageSize(_toolBar.getImage());

        if (size == null)
            return null;

        String imageSize = null;

        if (size.length == 2)
            imageSize = "width='" + size[0] + "' height='" + size[1] + "'";
        else
            imageSize = "height='" + size[0] + "'";

        return imageSize;
    }

    /**
     * Indicates if at least one of the tools of the toolbar contains a label.
     *
     * @return true if at least one of the tools of the toolbar contains a label, false otherwise
     */
    public boolean containsLabel()
    {
        int         i, nTools = _toolBar.getComponentCount();
        LyComponent component;
        String      label;

        // Parse the tools contained in this toolBar and search if at least one
        // is a toolBar and has a label
        for (i = 0; i < nTools; i++)
        {
            component = _toolBar.getComponent(i);
            if (component instanceof LyToolBar)
            {
                label = ((LyToolBar)component).getLabel();
                if ((label != null) && (label.length() != 0))
                    return true;
            }
        }

        return false;
    }

    /**
     * Computes the toolbar size.
     *
     * @return an array where width is at position 0 and height at position 1
     */
    public int[] computeSize()
    {
        LyWebViewManager viewManager = getWebViewManager();
        int              i, nTools = _toolBar.getComponentCount();
        LyComponent      component;
        Object           webToolBar;
        int              width = 0, height = 0;
        int              subToolbarHeight = 0;
        int              defaultToolWidth;
        int              defaultToolHeight;
        boolean          isHorizontal = isHorizontal();
        boolean          writeImage = true;
        boolean          writeLabel = true;
        int              imageHeight = 20;
        int              additionalHeight = 0;

        if ((viewManager.getBrowser() != LyWebViewManager.BROWSER_MSIE)
                && (viewManager.getBrowserVersion() < 5.0))
            additionalHeight = 5;

        // If no tool, size is null
        if (nTools == 0)
            return new int[] { 0, 0 };

        String toolbarImageHeightResource = getResource("LY_TOOLBAR_IMAGE_HEIGHT");
        int    toolbarImageHeight = -1;

        if ((toolbarImageHeightResource != null) && (toolbarImageHeightResource.length() > 0))
            toolbarImageHeight = Integer.parseInt(toolbarImageHeightResource);

        if (toolbarImageHeight > 0)
            imageHeight = toolbarImageHeight;

        // If image is drawn, check if label should be too.
        if (!testResource("LY_WRITE_TOOL_LABEL") && writeImage)
            writeLabel = false;

        if (isHorizontal)
        {
            if (writeImage)
            {
                defaultToolWidth = (writeLabel ? 50 : 0) + imageHeight;
                defaultToolHeight = (writeLabel ? getFontSize("LY_BUTTON_FONT") : 0) + 15
                    + imageHeight;
            }
            else
            {
                defaultToolWidth = 70;
                defaultToolHeight = getFontSize("LY_BUTTON_FONT");
            }
        }
        else
        {
            defaultToolWidth = (writeLabel ? 65 : 10) + imageHeight;
            if (writeImage)
                defaultToolHeight = imageHeight;
            else
                defaultToolHeight = 10;
        }

        for (i = 0; i < nTools; i++)
        {
            component = _toolBar.getComponent(i);
            if (component instanceof LyToolBar)
            {
                webToolBar = ((LyToolBar)component).getInternalView();

                // Retrieve toolbar width or height
                if (isHorizontal)
                {
                    width += ((LyStrutsToolBar)webToolBar).computeSize()[0];
                    subToolbarHeight = getFontSize("LY_TOOLBAR_TAB_FONT") - 2; // use
                    // font
                    // size
                    // minus
                    // 2
                }
                else
                    height += ((LyStrutsToolBar)webToolBar).computeSize()[1] + 15;
            }
            else
            {
                if (isHorizontal)
                    width += defaultToolWidth;
                else
                {
                    int currentWidth = imageHeight;

                    if (component == null)
                        height += 7;
                    else
                    {
                        String label = ((LyTool)component).getLabel();

                        if ((label != null) && (label.length() > 0))
                        {
                            label = translate(label);
                            currentWidth += label.length() * getFontSize("LY_BUTTON_FONT");
                        }
                        else
                            currentWidth += (writeLabel ? 65 : 10);
                        width = Math.max(width, currentWidth);

                        height += defaultToolHeight;
                    }
                }
            }

            // Separation between tools
            if (isHorizontal)
                width += 7;
            else
                height += 7;
        }

        if (isHorizontal)
            return new int[] { width, defaultToolHeight + subToolbarHeight + additionalHeight };
        else
            return new int[] { defaultToolWidth, height };
    } // end method computeSize

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalToolBar#setOpen(boolean)
     */
    public void setOpen(boolean open)
    {
        if (_toolBar.getParent() != null)
        {
            LyComponent parent = _toolBar.getParent().getParent();
        
            // If toolbar is contained in a tool, refresh this tool
            if (parent instanceof LyTool)
            {
                ((LyStrutsTool)parent.getInternalView()).refresh();
                return;
            }
        }

        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetImageInterface#setImage(java.lang.String)
     */
    public void setImage(String image)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalToolBar#addComponent(leon.view.LyComponent, int)
     */
    public void addComponent(LyComponent component, int pos)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalToolBar#removeComponent(leon.view.LyComponent)
     */
    public void removeComponent(LyComponent component)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebObject#refresh()
     */
    public void refresh()
    {
        String               id = "_" + _toolBar.getIdCode();
        LyStrutsViewManager   webviewManager = (LyStrutsViewManager)getViewManager();

        webviewManager.addReloadDiv(_lastFrameName, id, this);
    } // end method refresh
} // end class LyStrutsToolBar
