/*
 * LyStrutsTree.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.view.web.struts;

import java.util.*;

import leon.misc.*;
import leon.view.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.web.*;


/**
 * This class implements the Tree object (as described in the dtd) for web viewer.
 *
 * <p><b></b><br>
 * </p>
 *
 * <p><b>TREE: Tree</b><br>
 * A tree allows displaying a set of elements under the form of a tree. It is possible to specify
 * its selection mode. Such a component is interesting to display a hierarchy of objects of an
 * application. It allows not displaying some sub-sets of these objects.</p>
 *
 * @version $Id: LyStrutsTree.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria S.A.
 */
public class LyStrutsTree extends LyStrutsPanel
    implements LyGraphicalTree,
               LySetVisibleInterface
{
	public static final String __VERSION =
		"$Id: LyStrutsTree.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The <b>activeNode</b> String value.<br>
     * Used in a JavaScript to open an active node. */
    protected static final String ACTIVE_NODE = "activeNode".intern();

    /** The tree generic component. */
    protected LyTree              _tree;

    /** Background color. */
    protected String              _bgColor;
    
    /** Indicates whether the root node must be shown or not. */
    protected boolean             _hideRoot;
    
	/**
	 * Constructor.
	 *
	 * @param tree the tree component
	 */
	public LyStrutsTree(LyTree tree)
	{
        super(tree);
        
        _tree = tree;
        _hideRoot = false;

        initialize();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.LyStrutsTree#getBaseUrl(leon.view.web.LyWebPageWriter)
	 */
	protected String getBaseUrl(LyWebPageWriter pw)
	{
		// prepare the parameters
		Map<String, String> parameters = new HashMap<String, String>(2);

		parameters.put(LyWebConstants.CONTROLLER, _tree.getRootController());
		parameters.put(LyWebConstants.SOURCE, _tree.getId());

		// get the absolute url to the event handler action
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    baseUrl = viewManager.getActionAbsoluteUrl(
				LyStrutsConstants.PATH_ACTION_GRAPHICEVENT, parameters, true);

		return baseUrl;
	}

	/**
	 * Gets the javascript code used to open a node in the current tree.
	 *
	 * @param  pw the page writer.
	 * @return the javascript code used to open a node in the current tree
	 */
	protected String getOpenNodeScript(LyWebPageWriter pw)
	{
		return null;
	}
	
	/**
	 * Gets the javascript code used to consult a node in the current tree.
	 *
	 * @param  pw the page writer
	 * @return the javascript code used to consult a node in the current tree
	 */
	protected String getConsultNodeScript(LyWebPageWriter pw)
	{
		return null;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.struts.LyStrutsTree#getSelectNodeScript(leon.view.web.LyWebPageWriter)
	 */
	protected String getSelectNodeScript(LyWebPageWriter pw)
	{
		return null;
	}
	
	/*
	 * (non-Javadoc)
	 * @see leon.view.web.struts.LyStrutsPanel#getScrollHeightForSplitter(leon.view.web.LyWebPageWriter)
	 */
	public int getScrollHeightForSplitter(LyWebPageWriter pw)
	{
		LyTreeNode treeNode = _tree.getTreeNode();
		int		   visibleNodes = getVisibleTreeNodeCount(treeNode);
		
		// If more than 25 nodes, need a scroll 
		return (visibleNodes >= LyStrutsSplitter.MAXIMUM_COMPONENT_HEIGHT / 24 ? LyStrutsSplitter.MAXIMUM_COMPONENT_HEIGHT : -1);
	}
	
	/**
	 * Returns the number of visible sons for a given tree node.
	 * 
	 * @param treeNode the tree node.
	 * @return the number of visible sons.
	 */
	protected int getVisibleTreeNodeCount(LyTreeNode treeNode)
	{
		if (treeNode == null)
			return 0;
		
		int		   i, n = treeNode.getTreeNodeCount();
		int		   count = 0;
		LyTreeNode son;

		if (!treeNode.isHidden())
		{
			count++;
			
			if (treeNode.isOpen())
			{
				for (i = 0; i < n; i++)
				{
					son = treeNode.getTreeNode(i);
					count += getVisibleTreeNodeCount(son);
				}
			}
		}
		
		return count;
	}
	
	   /**
     * Initialization method.<br>
     * Sets the background color and the node font.
     */
    public void initialize()
    {
        _bgColor = null;

        if (_tree.getColor() != null)
            _bgColor = getHexColor(_tree.getColor());

        if (_bgColor == null)
            _bgColor = getHexColor("LY_TREE_BACKGROUND");

        if (_bgColor == null)
            _bgColor = getHexColor("LY_DEFAULT_BACKGROUND");
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalTree#setActiveNode(java.lang.String)
     */
    public void setActiveNode(String activeNode)
    {
    }

    /**
     * Gets the background color.
     *
     * @return the node background color
     */
    public String getBackgroundHexColor()
    {
        return _bgColor;
    }
    
    /**
     * Indicates whether the root node must be shown or not.
     * 
     * @param hideRoot true if root node must be hidden, false otherwise.
     */
    public void setHideRoot(boolean hideRoot)
    {
        _hideRoot = hideRoot;
    }
    
    /**
     * Indicates whether the root node must be shown or not.
     * 
     * @return true if root node must be shown, false otherwise.
     */
    public boolean isRootVisible()
    {
        return !_hideRoot;
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
     */
    public void writeHtml(LyWebPageWriter pw)
    {
        pw.startComponent(this);
        
        updateRefreshInfo(pw);

        String          label = _tree.getLabel();
        LyTree.SelectionMode  selectionMode = _tree.getSelectionMode();
        String          controllerId = _tree.getRootController();
        LyWebDecoration deco = getWebViewManager().getDecoration();

        // Set the background color
        pw.setBackgroundColor(_bgColor, pw.getActiveName());

        // First, add the needed script to the page
        String selectedColor = getHexColor("LY_TREENODE_SELECTED_BG_COLOR");
        String selectedTextColor = getHexColor("LY_TREENODE_SELECTED_FONT_COLOR");

        pw.addBeginScript("<script type='text/javascript'>var _defaultColor='" + _bgColor
            + "';</script>", false);
        pw.addBeginScript(
            "<script type='text/javascript'>var _defaultTextColor='#000000';</script>", false);
        pw.addBeginScript("<script type='text/javascript'>var _selectColor='" + selectedColor
            + "';</script>", false);
        pw.addBeginScript("<script type='text/javascript'>var _selectTextColor='"
            + selectedTextColor + "';</script>", false);
        pw.addBeginScript("<script type='text/javascript'>var _monoSelection='"
            + (selectionMode == LyTree.SelectionMode.SINGLE) + "';</script>", false);

        pw.addBeginScript("<script type='text/javascript'>window.defaultStatus='';</script>",
            false);

        // add the script to select a node
        String selectScript = getSelectNodeScript(pw);

        if (selectScript != null)
            pw.addBeginScript(selectScript, false);

        // add the script to open a node
        String openScript = getOpenNodeScript(pw);

        if (openScript != null)
            pw.addBeginScript(openScript, false);

        // add the script to consult a node
        String consultScript = getConsultNodeScript(pw);

        if (consultScript != null)
            pw.addBeginScript(consultScript, false);

        // Use one single form for a same frame
        pw.setFormId(controllerId);

        // Use InvokeSelection in tree selections
        pw.append("<input type='hidden' name='" + LyWebConstants.SELECTION_ITEM + "' value=''/>\n");
        pw.append("<input type='hidden' name='" + LyWebConstants.REFRESH + "' value='false'/>\n");

        label = translate(label);
        label = getHtmlString(label);

        LyTreeNode      node = _tree.getTreeNode();
        LyStrutsTreeNode internal = ((node == null) ? null
                                                   : (LyStrutsTreeNode)node.getInternalView());

        // Compute tree validity status
        if ((internal != null) && getEnvironment().testResource("LY_TREE_VALIDITY"))
            internal.computeTreeNodeValidity(node);

        pw.append("<table border='0' width='100%' height='100%' cellpadding='0' cellspacing='0' class='"
            + deco.getClassName(_tree, pw) + "'><tr><td id='tree_" + _tree.getIdCode() +"' valign='top'>\n");

        writeRootNode(pw);

        pw.append("\n</td></tr></table>");
        
        String  popupMenuView = getPopupMenuView();
        
        pw.append(popupMenuView);
        
        pw.endComponent(this);
    } // end method writeHtml

    /**
     * Writes the HTML code for root node.
     * 
     * @param pw current page writer.
     */
    protected void writeRootNode(LyWebPageWriter pw)
    {
        LyTree.SelectionMode  selectionMode = _tree.getSelectionMode();
        LyTree.SelectionType  selectionType = _tree.getSelectionType();
        LyTreeNode            node = _tree.getTreeNode();

        if (node != null)
        {
            LyStrutsTreeNode parentWebNode = (LyStrutsTreeNode)node.getInternalView();

            if (!isRootVisible())
            {
                int     i, n = node.getTreeNodeCount();
                
                for (i = 0; i < n; i++)
                {
                    LyTreeNode      subNode = node.getTreeNode(i);
                    LyStrutsTreeNode webNode = (LyStrutsTreeNode)subNode.getInternalView();

                    if (webNode != null)
                        webNode.writeNode(pw, parentWebNode, (i == n - 1), selectionMode, selectionType);
                }
            }
            else if (parentWebNode != null)
                parentWebNode.writeNode(pw, null, false, selectionMode, selectionType);
        }
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetVisibleInterface#setVisible(boolean)
     */
    public void setVisible(boolean status)
    {
        if (status)
        {
            LySetVisibleInterface visible = _tree.getSetVisibleParent();

            if (visible != null)
                visible.setVisible(true);
        }
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalTree#setTreeNode(leon.view.LyTreeNode)
     */
    public void setTreeNode(LyTreeNode treeNode)
    {
        boolean refresh = true;
        LyTabs  tabs = _tree.getParentTabs();

        if (tabs != null)
        {
            LyStrutsTabs webTabs = (LyStrutsTabs)tabs.getInternalView();
            LyTab        selectedTab = webTabs.getSelectedTab();
            LyTab        fieldTab = _tree.getParentTab();

            // Refresh only if table is displayed. Test if table tab is selected
            // tab and its label
            // is not null (tab with a null label contains fields that are
            // allways displayed)
            if ((selectedTab != null) && (fieldTab != null) && (fieldTab != selectedTab)
                    && (fieldTab.getLabel() != null) && (fieldTab.getLabel().length() > 0))
                refresh = false;
        }

        if (!refresh)
            return;

        // Update the div containing the children
        // Manage Ajax mode
        LyWebViewManager viewManager = getWebViewManager();
        StringBuffer     buffer = new StringBuffer();
        LyWebPageWriter  pw = new LyStrutsPage(_tree, true);

        pw.setActiveBuffer(_lastFrameName);

        writeRootNode(pw);

        StringBuffer innerHtml = pw.getBuffer(pw.getActiveName());
        String       innerHtmlStr = LyString.replace(innerHtml.toString(), "\n", "\\n");

        innerHtmlStr = LyString.replace(innerHtmlStr, "\"", "&quot;");

        buffer.append("<script type='text/javascript'>\n//<![CDATA[\nreplaceDivInnerHtml(\"")
            .append(_lastFrameName).append("\",\"tree_").append(_tree.getIdCode()).append("\",\"").append(
            innerHtmlStr).append("\"").append(");");
        buffer.append("//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(buffer.toString());
        pw.free();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.struts.LyStrutsFieldItem#setLabel(java.lang.String)
     */
    public void setLabel(String label)
    {
        // XX Nothing to do
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetColorInterface#setColor(java.lang.String)
     */
    public void setColor(String color)
    {
        // TODO
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetFontInterface#setFont(java.lang.String)
     */
    public void setFont(String font)
    {
        // TODO
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalTree#setForegroundColor(java.lang.String)
     */
    public void setForegroundColor(String color)
    {
        // TODO
    }
} // end class LyStrutsTree
