/*
 * LyStrutsTreeNode.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.web.struts;

import leon.misc.*;
import leon.view.*;
import leon.view.graphic.*;
import leon.view.web.*;


/**
 * This class implements the TreeNode object (as described in the dtd) for web viewer.
 *
 * <p><b>TREE_NODE: Tree node</b><br>
 * A tree node can contain a set of nodes (this node is then called parent node of those last
 * nodes). It is possible to assign a label and several images to a tree node, and to indicate if it
 * is selected, valid or open. It is also possible to indicate the font and the different colors of
 * the node.</p>
 *
 * @version $Id: LyStrutsTreeNode.java 35067 2012-03-27 17:03:22Z jlsaouli $
 * @author  Lyria
 * @see     leon.view.LyTreeNode
 */
public class LyStrutsTreeNode extends LyWebObject
    implements LyGraphicalTreeNode
{
	public static final String __VERSION =
		"$Id: LyStrutsTreeNode.java 35067 2012-03-27 17:03:22Z jlsaouli $";

	   /** The tree node generic component. */
    protected LyTreeNode       _treeNode;

    /** The parent node HTML tag representation. */
    protected String           _parentString;

    /** The URL of the icon representing an open node. */
    protected String           _openNodeIcon;

    /** The URL of the icon representing a closed node. */
    protected String           _closedNodeIcon;

    /**
     * The tree selection mode.<br>
     * Possible values are the LyTree selection mode values :
     *
     * <ul>
     *   <li>{@link LyTree#SelectionMode.NONE}</li>
     *   <li>{@link LyTree#SelectionMode.CHECKBOX}</li>
     *   <li>{@link LyTree#SelectionMode.MULTIPLE}</li>
     *   <li>{@link LyTree#SelectionMode.SENSITIVE}</li>
     *   <li>{@link LyTree#SelectionMode.SINGLE}</li>
     * </ul>
     */
    protected LyTree.SelectionMode   _selectionMode;

    /**
     * The tree selection type.<br>
     * Possible values are the LyTree selection type values :
     *
     * <ul>
     *   <li>{@link LyTree#SelectionType.LEAVES_ONLY}</li>
     *   <li>{@link LyTree#SelectionType.ALL}</li>
     * </ul>
     */
    protected LyTree.SelectionType   _selectionType;

    /** Indicates if the node is the last of its bothers or not. */
    protected boolean          _isLast;
    
    /** Tree node instance that is being removed from current tree node component. */
    protected LyTreeNode       _removedTreeNode;
    
	/**
	 * Class constructor of a Tree Node object for the struts target.
	 *
	 * @param treeNode the generic Tree Node defined for every type of view targets
	 */
	public LyStrutsTreeNode(LyTreeNode treeNode)
	{
        super(treeNode);
        
        _treeNode = treeNode;
        _parentString = null;
        _openNodeIcon = null;
        _closedNodeIcon = null;
        _removedTreeNode = null;
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.web.dhtml.LyDhtmlTreeNode#addTreeNode(leon.view.LyTreeNode, int)
	 */
	public void addTreeNode(LyTreeNode treeNode, int pos)
	{
        refreshDivContent();

		if ((treeNode != null) && treeNode.isSelected())
			addSelectionScript(treeNode, true);
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.web.dhtml.LyDhtmlTreeNode#removeTreeNode(leon.view.LyTreeNode)
	 */
	public void removeTreeNode(LyTreeNode treeNode)
	{
        _removedTreeNode = treeNode;
        refreshDivContent();
        _removedTreeNode = null;

		if ((treeNode != null) && treeNode.isSelected())
			addSelectionScript(treeNode, false);
	}

	/**
	 * Generates a selection script and adds it to the view manager.
	 *
	 * @param row the row that must be selected or unselected.
	 * @param addSelection indicates whether this tree node must be added to Javascript
	 * selection or not.
	 */
	protected void addSelectionScript(LyTreeNode treeNode, boolean addSelection)
	{
		LyStrutsViewManager	viewManager = (LyStrutsViewManager)getViewManager();
		StringBuffer		script = new StringBuffer();
		LyTree				tree = getParentTree();
		String				rowId = treeNode.getId();
		String				javascriptTreeNodeId = LyString.getJavascriptString(rowId);

		script.append("<script type='text/javascript'>");

		if (addSelection)
			script.append("addSelectionId('tree_" + tree.getIdCode() + "', '" + javascriptTreeNodeId + "');");
		else
			script.append("removeSelectionId('tree_" + tree.getIdCode() + "', '" + javascriptTreeNodeId + "');");

		script.append("</script>\n");

		viewManager.addAdditionalRefreshScripts(script.toString());
	}

	/**
	 * Gets the javascript code used to consult a node in the current tree.
	 *
	 * @param  pw the page writer.
	 * @return the javascript code used to consult a node in the current tree
	 */
	protected String getConsultNodeScript(LyWebPageWriter pw)
	{
		// get the url to the selection handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		String				id = _treeNode.getId();

		id = LyString.getJavascriptString(id);

		if (pw.writeDivOnly())
			id = LyString.replace(id, "\\", "\\\\");

		return "javascript:ConsultTreeNode(\"" + id + "\", \"" + _treeNode.getRootController() + "\", \"" + getParentTree().getId() + "\", \"" + actionUrl + "\");";
	}

	/**
	 * Gets the javascript code used to select a node in the current tree.
	 *
	 * @param  pw the page writer.
	 * @return the javascript code used to select a node in the current tree
	 */
	protected String getSelectNodeScript(LyWebPageWriter pw)
	{
		// get the url to the selection handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    selectionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_SELECTEVENT);
		String				id = _treeNode.getId();
		LyTree				tree = getParentTree();

		id = LyString.getJavascriptString(id);

		if (pw.writeDivOnly())
			id = LyString.replace(id, "\\", "\\\\");

		return "javascript:InvokeTreeNodeSelection(\"tree_" + tree.getIdCode() + "\",\"" + id + "\",\"" + _treeNode.getRootController() +
			"\",\"" + selectionUrl + "\"," + (tree.getSelectionMode() == LyTree.SelectionMode.SINGLE) + ",event);";
	}

	/**
	 * Gets the javascript code used to expand a node in the current tree.
	 *
	 * @param  pw the page writer.
	 * @return the javascript code used to expand a node in the current tree
	 */
	protected String getExpandNodeScript(LyWebPageWriter pw)
	{
		// get the url to the selection handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		String				id = _treeNode.getId();

		id = LyString.getJavascriptString(id);

		if (pw.writeDivOnly())
			id = LyString.replace(id, "\\", "\\\\");

		return "javascript:ExpandTreeNode(\"" + id + "\", \"" + _treeNode.getRootController() + "\", \"" + actionUrl + "\");";
	}

	/**
	 * Gets the javascript code used for popup menu in the current tree.
	 *
	 * @param  pw the page writer.
	 * @return the javascript code used for popup menu in the current tree
	 */
	protected String getPopupMenuNodeScript(LyWebPageWriter pw)
	{
		// get the url to the selection handler
		LyStrutsViewManager viewManager = (LyStrutsViewManager)getViewManager();
		String			    actionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_GRAPHICEVENT);
		String			    selectionUrl = viewManager.getActionURL(LyStrutsConstants.PATH_ACTION_SELECTEVENT);
		String				id = _treeNode.getId();
		LyTree				tree = getParentTree();

		id = LyString.getJavascriptString(id);

		if (pw.writeDivOnly())
			id = LyString.replace(id, "\\", "\\\\");

		return "javascript:InvokeTreeNodePopup(\"" + id + "\",\"tree_" + tree.getIdCode() + "\",\"" + getParentTree().getId() + "\",\"" +
			_treeNode.getRootController() + "\",\"" + actionUrl + "\",\"" + selectionUrl + "\",event);return false;";
	}
	
	   /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetSelectedInterface#setSelected(boolean)
     */
    public void setSelected(boolean status)
    {
        // Manage Ajax mode
        LyTree tree = getParentTree();

        if (!getEnvironment().testResource("LY_TREE_VALIDITY")
        	&& (tree != null)
            && (tree.getSelectionMode() != LyTree.SelectionMode.CHECKBOX))
        {
            LyWebViewManager viewManager = getWebViewManager();
            StringBuffer     buffer = new StringBuffer();
            String           color = null;
            String           textColor = null;

            if (status)
            {
                color = getHexColor("LY_TREENODE_SELECTED_BG_COLOR");
                textColor = getHexColor("LY_TREENODE_SELECTED_FONT_COLOR");
            }
            else
            {
                LyStrutsTree htmlTree = (LyStrutsTree)tree.getInternalView();

                color = htmlTree.getBackgroundHexColor();

				if (!_treeNode.isDefaultFontColor())
                    textColor = getHexColor(_treeNode.getFontColor());

				if ((textColor == null) && (tree != null) && !tree.isDefaultFontColor())
                    textColor = getHexColor(tree.getFontColor());

                if (textColor == null)
                    textColor = getHexColor("LY_TREENODE_FONT_COLOR");
            }

            buffer.append("<script type='text/javascript'>\n//<![CDATA[\nSetTreeNodeColor('")
                .append(getTreeNodeId()).append("_td','").append(getTreeNodeId()).append("_font','")
                .append(color).append("','").append(textColor).append("');\n//]]>\n</script>");

            viewManager.addAdditionalRefreshScripts(buffer.toString());
        }
        else
            refreshDivContent();
        
        addSelectionScript(_treeNode, status);
    } // end method setSelected

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalTreeNode#setOpen(boolean)
     */
    public void setOpen(boolean open)
    {
        refreshDivContent();
    } // end method setOpen

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebComponent#writeHtml(leon.view.web.LyWebPageWriter)
     */
    public void writeHtml(LyWebPageWriter pw)
    {
    }

    /**
     * Gets the begining of the HTML representation of the node (its parent representation).
     *
     * @return the HTML representation of its parent node
     */
    public String getBeginning()
    {
        return _parentString;
    }

    /**
     * Prepare the begining of the HTML representation of the node (its parent representation).
     *
     * @param parent the parent node of this node
     * @param isLast true is this node is the last of the britherhood, false otherwise
     */
    public void prepareBeginning(LyStrutsTreeNode parent, boolean isLast)
    {
        if (parent == null)
            return;
        
        // If root must be hidden, do not generate beginning for nodes at level 1
        // (just under root node)
        LyTree      tree = getParentTree();
        LyStrutsTree htmlTree = (LyStrutsTree)tree.getInternalView();
        boolean     rootVisible = htmlTree.isRootVisible();
        
        if (!rootVisible
                && !_treeNode.isExpandable()
                && (getVisibleTreeNodeCount(_treeNode) <= 0)
                && (_treeNode.getParent() != null)
                && (_treeNode.getParent().getParent() instanceof LyTree))
            return;

        StringBuffer buffer = new StringBuffer();
        String       nodeIcon, parentString;

        if (isLast)
            nodeIcon = getImageUrl("LY_EMPTY_NODE");
        else
            nodeIcon = getImageUrl("LY_VERTICAL_LINE");

        parentString = parent.getBeginning();
        if (parentString != null)
            buffer.append(parentString);
        buffer.append("<td><img src=\"" + nodeIcon + "\" /></td>");

        _parentString = buffer.toString();
    }

    /**
     * Gets the number of sub-nodes shown to the user.
     *
     * @param  treeNode the parent node
     * @return the number of sub-nodes shown to the user
     */
    protected int getVisibleTreeNodeCount(LyTreeNode treeNode)
    {
        if (treeNode == null)
            return 0;

        int        i, n = treeNode.getTreeNodeCount();
        int        count = 0;
        LyTreeNode son;

        for (i = 0; i < n; i++)
        {
            son = treeNode.getTreeNode(i);

            if (!son.isHidden() && (son != _removedTreeNode))
                count++;
        }

        return count;
    }

    /**
     * Gets parent tree instance.
     *
     * @return parent tree instance
     */
    public LyTree getParentTree()
    {
        LyComponent parent = _treeNode.getParent();

        while ((parent != null) && !(parent instanceof LyTree))
            parent = parent.getParent();

        return (LyTree)parent;
    }

    /**
     * Gets the last frame name where this tree node where generated.
     *
     * @return the frame name
     */
    public String getLastFrameName()
    {
        return _lastFrameName;
    }

    /**
     * Writes the HTML representation of this node.
     *
     * @param pw            the page writer
     * @param parent        the parent of this node, null it this is the root
     * @param isLast        indicates if this node is the last of the brotherhood or not
     * @param selectionMode the selection mode
     * @param selectionType the selection type
     */
    public void writeNode(LyWebPageWriter pw, LyStrutsTreeNode parent, boolean isLast,
        LyTree.SelectionMode selectionMode, LyTree.SelectionType selectionType)
    {
        pw.startComponent(this);
        
        updateRefreshInfo(pw);

        _selectionMode = selectionMode;
        _selectionType = selectionType;
        _isLast = isLast;

        // First Write the node
        LyTree           tree = getParentTree();
        int              nVisible = getVisibleTreeNodeCount(_treeNode),
                         n = _treeNode.getTreeNodeCount();
        String           label = _treeNode.getLabel();
        String           translatedLabel = translate(label);
        String           image = _treeNode.getImage();
        boolean          isOpen = _treeNode.isOpen();
        boolean          isSelected = _treeNode.isSelected();
        String           icon;
        String           nodeIcon;
        String           beginning = null;
        String           openNode = null;
        String           selectLink = getSelectNodeScript(pw);
        String           selectedColor = getHexColor("LY_TREENODE_SELECTED_BG_COLOR");
        String           textColor = null, selectedTextColor;
        String           translatedTootip = translate(_treeNode.getTooltip());
        
        if ((translatedTootip == null) || translatedTootip.equals(""))
            translatedTootip = translatedLabel;
        
        String  toolTip = getHtmlString(translatedTootip);
        String  javascriptToolTip;
        String  nodeFont = null;
        String  treeNodeId = getTreeNodeId();

		if (!_treeNode.isDefaultFont())
        {
            nodeFont = getFontString(_treeNode.getFont());

            pw.addCssScript("." + treeNodeId + " { text-decoration:NONE;" + nodeFont
                + ";color:black}", false);
        }

		if (!_treeNode.isDefaultFontColor())
            textColor = getHexColor(_treeNode.getFontColor());

        if ((textColor == null) && (tree != null) && (tree.getFontColor() != null))
            textColor = getHexColor(tree.getFontColor());

        selectedTextColor = getHexColor("LY_TREENODE_SELECTED_FONT_COLOR");
        label = getHtmlString(translatedLabel);

        if (parent != null)
            beginning = parent.getBeginning();
        prepareBeginning(parent, isLast);

        if (isLast)
        {
            if ((nVisible > 0) || _treeNode.isExpandable())
            {
                _openNodeIcon = getImageUrl("LY_LAST_NODE_OPEN");
                _closedNodeIcon = getImageUrl("LY_LAST_NODE_CLOSED");

                if (isOpen)
                    nodeIcon = _openNodeIcon;
                else
                    nodeIcon = _closedNodeIcon;
            }
            else
            {
                _openNodeIcon = getImageUrl("LY_LAST_NODE");
                _closedNodeIcon = getImageUrl("LY_LAST_NODE");

                nodeIcon = _openNodeIcon;
            }
        }
        else if ((nVisible > 0) || _treeNode.isExpandable())
        {
            _openNodeIcon = getImageUrl("LY_NODE_OPEN");
            _closedNodeIcon = getImageUrl("LY_NODE_CLOSED");

            if (isOpen)
                nodeIcon = _openNodeIcon;
            else
                nodeIcon = _closedNodeIcon;
        }
        else
        {
            _openNodeIcon = getImageUrl("LY_NODE");
            _closedNodeIcon = getImageUrl("LY_NODE");

            nodeIcon = _openNodeIcon;
        }

        if (image != null)
            icon = getImageUrl(image);
        else
        {
            if (isOpen)
                icon = getImageUrl("LY_DIR_OPEN");
            else
                icon = getImageUrl("LY_DIR_CLOSED");
        }

        String id = _treeNode.getId();

        if (pw.writeDivOnly())
            id = LyString.replace(id, "\\", "\\\\");

        if ((nVisible > 0) || _treeNode.isExpandable())
            openNode = getExpandNodeScript(pw);

        pw.append("<div id=\"" + treeNodeId + "_div\">");
        pw.append("<table border='0' cellspacing='0' cellpadding='0'>\n");
        pw.append("<tr id=\"" + id + "\"");

        // Check selection mode
        boolean isSelectable = true;

        if (selectionMode == LyTree.SelectionMode.NONE)
            isSelectable = false;
        else if ((selectionType == LyTree.SelectionType.LEAVES_ONLY) && (n > 0))
            isSelectable = false;
        
        if (isSelectable)
        {
            String  popupMenuScript = getPopupMenuNodeScript(pw);
            
            pw.append(" onContextMenu='");
            pw.append(popupMenuScript);
            pw.append("'");
        }
        
        pw.append(">");

        if (beginning != null)
            pw.append(beginning);

        // Don't draw nodeIcon, if we are the root node
        if (parent != null)
        {
            pw.append("<td >");

            if (openNode != null)
            {
                if (isOpen)
                    javascriptToolTip = LyString.getJavascriptString(translate("LY_CLOSE_NODE")
                            + " " + translatedLabel);
                else
                    javascriptToolTip = LyString.getJavascriptString(translate("LY_OPEN_NODE")
                            + " " + translatedLabel);

                pw.append("<a style='cursor:hand;cursor:pointer;' onclick='" + openNode.toString());
                pw.append("' onmouseover=\'return true\'>");
            }

            pw.append("<img src=\"" + nodeIcon + "\" border='no'/>");

            if (openNode != null)
                pw.append("</a>");
            pw.append("</td>");
        }

        // draw the checkbox for checkbox selection mode only
        if ((selectionMode == LyTree.SelectionMode.CHECKBOX) && isSelectable)
        {
            String checkbox = null;

            if (isSelected)
                checkbox = getImageUrl("LY_TREENODE_CHECKED_CHECKBOX");
            else
                checkbox = getImageUrl("LY_TREENODE_EMPTY_CHECKBOX");

            pw.append("<td>");

            javascriptToolTip = LyString.getJavascriptString(translate("LY_SELECT_NODE") + " "
                    + translatedLabel);

            pw.append("<a style='cursor:hand;cursor:pointer;' onclick='");
            pw.append(selectLink.toString());
            pw.append("' class='treeNodeLabel' onmouseover=\'window.status=\"");
            pw.append(javascriptToolTip);
            pw.append("\";return true\'>");
            pw.append("<img src=\"" + checkbox + "\" border='no'/>");
            pw.append("</a>");
            pw.append("</td>");
            pw.append("<td>&nbsp;</td>");
        }

        if ((label == null) || (label.length() == 0))
            pw.append("<td><table border='0' cellpadding='2' cellspacing='0'><tr>");

        // the node icon is used to consult a node
        if ((parent != null) || !getEnvironment().testResource("LY_TREE_VALIDITY"))
        {
            if (isSelected && (selectionMode != LyTree.SelectionMode.CHECKBOX)
                    && ((label == null) || (label.length() == 0)))
                pw.append("<td bgcolor='" + selectedColor + "'>");
            else
                pw.append("<td>");

            pw.append("<a ");

            String activeNode = tree.getActiveNode();

            javascriptToolTip = LyString.getJavascriptString(translate("LY_SELECT_NODE") + " "
                    + translatedLabel);

            if ((activeNode != null) && activeNode.equals(_treeNode.getId()))
                pw.append("name='" + LyStrutsTree.ACTIVE_NODE + "' ");

            pw.append("style='cursor:hand;cursor:pointer;' onclick='");
            pw.append(selectLink.toString());

            pw.append("' onmouseover=\'SetSelectionStartMode(false);window.status=\"");
            pw.append(javascriptToolTip);
            pw.append("\";return true\' onmouseout=\'SetSelectionStartMode(true);\'>");

            // Retrieving LY_DEFAULT_IMAGE_HEIGHT resource
            // and using it for width and height of node image
            String imageHeightValue = getEnvironment().getResource("LY_DEFAULT_IMAGE_HEIGHT");
            int    imageHeight = LyTools.intFromString(imageHeightValue);

            pw.append("<img id=\"" + treeNodeId + "_icon\" src=\"" + icon + "\"");

            if (imageHeight > 0)
            {
                pw.append(" width='" + imageHeight + "'");
                pw.append(" height='" + imageHeight + "'");
            }

            pw.append(" border='no' title='");
            pw.append(toolTip);
            pw.append("'/></a></td>");
        }

        // Draw other images
        int     iImages, nImages = _treeNode.getImageCount();
        LyImage subImage = null;

        for (iImages = 0; iImages < nImages; iImages++)
        {
            subImage = _treeNode.getImage(iImages);

            if (isSelected && (selectionMode != LyTree.SelectionMode.CHECKBOX)
                    && ((label == null) || (label.length() == 0)))
                pw.append("<td bgcolor='" + selectedColor + "'>");
            else
                pw.append("<td>");

            pw.append("<img src=\"" + getImageUrl(subImage.getValue()) + "\" border='no'/>");
            pw.append("</td>");
        }

        // Draw validity
        if (getEnvironment().testResource("LY_TREE_VALIDITY"))
        {
            pw.append("<td>");
            pw.append("<img src=\""
                + getImageUrl("LY_TREE_VALIDITY_" + _treeNode.getValidity().toString())
                + "\" border='no'/>");
            pw.append("</td>");
        }

        pw.append("<td>&nbsp;</td>");

        if (isSelected && (selectionMode != LyTree.SelectionMode.CHECKBOX)
                && ((label != null) && (label.length() != 0)))
            pw.append("<td id=\"" + treeNodeId + "_td\" bgcolor='" + selectedColor
                + "' nowrap='true'>");
        else if ((_treeNode.getColor() != null) && !_treeNode.isDefaultColor())
        {
            String  nodeColor = getHexColor(_treeNode.getColor());
            
            pw.append("<td id=\"" + treeNodeId + "_td\" bgcolor='" + nodeColor
                + "' nowrap='true'>");
        }
        else
            pw.append("<td id=\"" + treeNodeId + "_td\" nowrap='true'>");

        if (isSelectable)
        {
            javascriptToolTip = LyString.getJavascriptString(translate("LY_SELECT_NODE") + " "
                    + translatedLabel);

            pw.append("<a style='cursor:hand;cursor:pointer;' onclick='");
            pw.append(selectLink.toString());

            if (nodeFont != null)
                pw.append("' class=\"" + treeNodeId + "\" onmouseover=\'window.status=\"");
            else
                pw.append("' class='treeNodeLabel' onmouseover=\'window.status=\"");

            pw.append(javascriptToolTip);
            pw.append("\";SetSelectionStartMode(false);return true\' onmouseout=\'SetSelectionStartMode(true);\'>");

            pw.append("<font id=\"" + treeNodeId + "_font\"");

            if (isSelected && (selectionMode != LyTree.SelectionMode.CHECKBOX))
                pw.append(" color='" + selectedTextColor + "'");
            else if (textColor != null)
                pw.append(" color='" + textColor + "'");

            pw.append(">");
        }
        else
        {
            pw.append("<font class='label'");

			if (textColor != null)
				pw.append(" color='" + textColor + "'");

			pw.append(">");
        }
        
        if (pw.writeDivOnly())
            pw.append("&nbsp;" + LyString.replace(label, "\\", "\\\\") + "&nbsp;");
        else
            pw.append("&nbsp;" + label + "&nbsp;");
        
        pw.append("</font>");

        if (isSelectable)
            pw.append("</a>");

        if ((label == null) || (label.length() == 0))
            pw.append("</tr></table></td>");

        pw.append("</td>");
        pw.append("</tr>");
        pw.append("</table>\n");

        writeChildren(pw);

        pw.append("</div>\n");
        
        if (_treeNode.isSelected())
        {
            StringBuffer    script = new StringBuffer();
            String          javascriptId;
            
            id = _treeNode.getId();
            javascriptId = LyString.getJavascriptString(id);
            tree = getParentTree();

            script.append("<script type='text/javascript'>\n//<![CDATA[\n");
            script.append("addSelectionId('tree_" + tree.getIdCode() + "', '" + javascriptId + "');");
            script.append("\n//]]>\n</script>\n");

            if (pw.writeDivOnly())
                getWebViewManager().addAdditionalRefreshScripts(script.toString());
            else
                pw.addBeginScript(script.toString(), false);
        }
        
        pw.endComponent(this);
    } // end method writeNode

    /**
     * Writes the HTML code for children of current tree node.
     *
     * @param pw the page writer
     */
    protected void writeChildren(LyWebPageWriter pw)
    {
        int          nVisible = getVisibleTreeNodeCount(_treeNode),
                     n = _treeNode.getTreeNodeCount();
        boolean      isOpen = _treeNode.isOpen();
        String       selectLink = getSelectNodeScript(pw);

        // Write recursively all child nodes
        if (isOpen)
        {
            LyTreeNode      node;
            LyStrutsTreeNode webNode;
            int             i, j = 0;

            for (i = 0; i < n; i++)
            {
                node = _treeNode.getTreeNode(i);

                if (node.isHidden() || (node == _removedTreeNode))
                    continue;

                webNode = (LyStrutsTreeNode)node.getInternalView();
                webNode.writeNode(pw, this, (j == (nVisible - 1)), _selectionMode, _selectionType);
                j++;
            }
        }
        // look for selected children then add script to select current node.
        else if (hasChildSelected())
        {
            StringBuffer buffer = new StringBuffer();

            buffer.append("<script type='text/javascript'>\n//<![CDATA[\n");
            buffer.append(selectLink.toString());
            buffer.append("\n//]]>\n</script>");

            if (!pw.writeDivOnly())
                pw.append(buffer.toString());
            else
            {
                LyWebViewManager viewManager = getWebViewManager();

                viewManager.addAdditionalRefreshScripts(buffer.toString());
            }
        }
    } // end method writeChildren

    /**
     * Gets the identifier of this tree node in HTML code.
     *
     * @return the id of this node in HTML code
     */
    protected String getTreeNodeId()
    {
        String treeNodeId = _treeNode.getId();
        if (treeNodeId == null)
            return null;

        treeNodeId = treeNodeId.replace('\'', '_');
        treeNodeId = treeNodeId.replace('@', '_');
        treeNodeId = treeNodeId.replace('\\', '_');

        return treeNodeId;
    }

    /**
     * Indicates if this node has at least one descendant selected.
     *
     * @return true if at least one child, child of child, etc. is selected, false otherwise
     */
    public boolean hasChildSelected()
    {
        int n = _treeNode.getTreeNodeCount();

        if (n == 0)
            return false;

        LyTreeNode      node;
        LyStrutsTreeNode webNode;
        boolean         result = false;
        int             i = 0;

        for (i = 0; (i < n) && !result; i++)
        {
            node = _treeNode.getTreeNode(i);
            if (node.isSelected() && !node.isHidden() && (node != _removedTreeNode))
                result = true;
            else
            {
                webNode = (LyStrutsTreeNode)node.getInternalView();
                result = webNode.hasChildSelected();
            }
        }

        return result;
    }

    /**
     * Updates the tree node validity from the view component.
     */
    public void updateValidityFromView()
    {
        int                     i, n = _treeNode.getTreeNodeCount();
        LyTreeNode              currentNode;
        LyTreeNode.Validity     validity = _treeNode.getValidity();

        // Update children validity status
        for (i = 0; i < n; i++)
        {
            currentNode = _treeNode.getTreeNode(i);

            if (validity.equals(LyTreeNode.Validity.VALID)
                    || validity.equals(LyTreeNode.Validity.INVALID))
                currentNode.setValidity(validity);
        }
    }

    /**
     * Computes the sub-tree validity.
     *
     * @param node current node
     */
    public void computeTreeNodeValidity(LyComponent node)
    {
        LyTreeNode treeNode = (LyTreeNode)node;

        if (treeNode == null)
            return;

        int n = treeNode.getTreeNodeCount();

        // Sets default value to valid.
        if (treeNode.getValidity() == null)
            treeNode.setValidity(LyTreeNode.Validity.VALID);

        if (n == 0)
            return; // Nothing to do on leaves

        boolean    validFound = false;
        boolean    invalidFound = false;
        LyTreeNode currentNode;

        for (int i = 0; i < n; i++)
        {
            currentNode = treeNode.getTreeNode(i);
            computeTreeNodeValidity(currentNode); // Recurse deeper

            // We try to compute state at this level as long as partially is not
            // already detected
            if ((!validFound) || (!invalidFound))
            {
                if ((!validFound)
                        && (currentNode.getValidity() == LyTreeNode.Validity.VALID))
                    validFound = true;
                else if ((!invalidFound)
                        && (currentNode.getValidity() == LyTreeNode.Validity.INVALID))
                    invalidFound = true;
                else if (currentNode.getValidity() == LyTreeNode.Validity.PARTIALLY)
                {
                    validFound = true;
                    invalidFound = true;
                }
            }
        }
        // Partially valid is detected
        if (validFound && invalidFound)
        {
            if (treeNode.getValidity() != LyTreeNode.Validity.PARTIALLY)
                treeNode.setValidity(LyTreeNode.Validity.PARTIALLY);
        }
        else if (validFound)
        {
            if (treeNode.getValidity() != LyTreeNode.Validity.VALID)
                treeNode.setValidity(LyTreeNode.Validity.VALID);
        }
        else if (invalidFound)
        {
            if (treeNode.getValidity() == LyTreeNode.Validity.INVALID)
                treeNode.setValidity(LyTreeNode.Validity.INVALID);
        }
    } // end method computeTreeNodeValidity
    
    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
     */
    public void setLabel(String label)
    {
        refreshDivContent();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetHiddenInterface#setHidden(boolean)
     */
    public void setHidden(boolean hidden)
    {
        refresh();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetImageInterface#setImage(java.lang.String)
     */
    public void setImage(String key)
    {
        String           treeNodeId = getTreeNodeId();
        LyWebViewManager viewManager = getWebViewManager();

        // Manage Ajax mode
        StringBuffer    buffer = new StringBuffer();
        String          path = getImageUrl(key);
        
        // Update the image
        buffer.append("<script type='text/javascript'>SetImageSrc('").append(treeNodeId)
            .append("_icon','").append(path).append("');</script>");
        
        viewManager.addAdditionalRefreshScripts(buffer.toString());
    }
    
    /**
     * Refreshes the content of the div containing the tree node description.
     */
    protected void refreshDivContent()
    {
        // Manage Ajax mode
        LyWebViewManager viewManager = getWebViewManager();
        StringBuffer     buffer = new StringBuffer();
        String           treeNodeId = getTreeNodeId();
        LyTree           tree = getParentTree();
        LyTreeNode       root = tree.getTreeNode();
        LyStrutsTree      htmlTree = (LyStrutsTree)tree.getInternalView();
        boolean          rootVisible = htmlTree.isRootVisible();

        if (_lastFrameName == null)
        {
            LyStrutsTreeNode htmlRoot = (LyStrutsTreeNode)root.getInternalView();

            _lastFrameName = htmlRoot.getLastFrameName();
        }

        // Update the div containing the children
        LyWebPageWriter pw = new LyStrutsPage(_treeNode, true);
        LyComponent     parent = _treeNode.getParent();
        LyStrutsTreeNode parentNode = null;

        pw.setActiveBuffer(_lastFrameName);

        if (parent instanceof LyTreeNode)
        {
            LyTreeNode parentTreeNode = (LyTreeNode)parent;

            parentNode = (LyStrutsTreeNode)parentTreeNode.getInternalView();
        }
        
        if (!rootVisible
                && !_treeNode.isExpandable()
                && (getVisibleTreeNodeCount(_treeNode) <= 0)
                && (parent == root))
            parentNode = null;

        writeNode(pw, parentNode, _isLast, _selectionMode, _selectionType);

        StringBuffer innerHtml = pw.getBuffer(pw.getActiveName());
        String       innerHtmlStr = LyString.replace(innerHtml.toString(), "\n", "\\n");

        innerHtmlStr = LyString.replace(innerHtmlStr, "\"", "&quot;");

        buffer.append("<script type='text/javascript'>\n//<![CDATA[\nreplaceDivInnerHtml(\"")
            .append(_lastFrameName).append("\",\"").append(treeNodeId).append("_div\",\"").append(
            innerHtmlStr).append("\"").append(");");

        // Update the scroll if necessary
        if (LyWebConstants.POPUP_VIEW.equals(_lastWindowName))
            buffer.append("checkSizePopupView(null, false);");
        else
            buffer.append("checkSizeMainView();");

        buffer.append("//]]>\n</script>");

        viewManager.addAdditionalRefreshScripts(buffer.toString());
        pw.free();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetColorInterface#setColor(java.lang.String)
     */
    public void setColor(String color)
    {
        // TO DO
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetFontInterface#setFont(java.lang.String)
     */
    public void setFont(String font)
    {
        // TO DO
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalTreeNode#setForegroundColor(java.lang.String)
     */
    public void setForegroundColor(String color)
    {
        // TO DO
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalTreeNode#addImage(leon.view.LyImage, int)
     */
    public void addImage(LyImage image, int pos)
    {
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalTreeNode#removeImage(leon.view.LyImage)
     */
    public void removeImage(LyImage image)
    {
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.setinterface.LySetValidityInterface#setValidity(java.lang.String)
     */
    public void setValidity(LyTreeNode.Validity validity)
    {
        refresh();
        updateValidityFromView();
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.view.graphic.LyGraphicalTreeNode#setExpandable(boolean)
     */
    public void setExpandable(boolean expandable)
    {
        refreshDivContent();
    }

    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalTreeNode#setTooltip(java.lang.String)
     */
    public void setTooltip(String tooltip) 
    {
        refresh();
    }
} // end class LyStrutsTreeNode
