/*
 * LyStrutsViewManager.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.view.web.struts;

import org.apache.struts.*;
import org.apache.struts.config.*;

import java.awt.*;
import java.io.*;
import java.net.*;
import java.util.*;

import javax.servlet.*;
import javax.servlet.http.*;

import leon.app.*;
import leon.control.*;
import leon.misc.*;
import leon.view.*;
import leon.view.common.*;
import leon.view.web.*;
import leon.view.web.struts.action.*;


/**
 * Communication control layer --> View Will implement remote update notifications.
 *
 * @version $Id: LyStrutsViewManager.java 35556 2012-05-23 07:06:09Z nhamon $
 * @author  Lyria S.A.
 */
public class LyStrutsViewManager extends LyWebViewManager
{
	public static final String __VERSION =
		"$Id: LyStrutsViewManager.java 35556 2012-05-23 07:06:09Z nhamon $";

	   /** The wait counter. */
    public static int          WAIT_COUNTER = 50;

    /**
     * Default Frame used in some parts of the code when a default AWT component is needed (for
     * example to create an image).
     */
    public static Frame        _frame;

    /** Indicates whether an absolute URL must be used as a root url. */
    public static boolean	   USE_ABSOLUTE_ROOT_URL = true;
    
    /** Enable/Disable popup views. */
    public static boolean		DHTML_POPUP_VIEW = true;
    

    /**
     * Creates and gets the default frame.
     *
     * @return The default frame.
     */
    public static Frame getDefaultFrame()
    {
        if (_frame == null)
        {
            _frame = new Frame();
            // ensure that the peer is ready.
            // a peer is required for some methods such as createImage()
            // see also : LyEnvironment.getImage()
            if (LyEnvironment.isOperatingSystem(LyEnvironment.OS_WINDOWS))
                _frame.addNotify();
        }

        return _frame;
    }

    /**
     * Removes the default frame.
     */
    public static void removeDefaultFrame()
    {
        if (_frame != null)
            _frame.dispose();

        _frame = null;
    }

    /**
     * List of HTML division that needs to be reloaded.<br>
     * It is used in order to don't reload all div any times.
     */
    protected HashMap<String, LyWebObject> _divToReplace;

    /** Id of the last shown popup menu. */
    protected String                       _lastPopupMenuId;

	/** The current struts module configuration : usefull to build URLs to Leonardi actions. */
	private ModuleConfig	   _moduleConfig;

	/** The current servlet context : the configuration of the WEB application. */
	private ServletContext     _context;

	/** The context url **/
	private String			   _contextUrl;

	/**
	 * Constructor.
	 *
	 * @param environment the application environment
	 */
	public LyStrutsViewManager(LyEnvironment environment)
	{
        super(environment);

        _beans = null;
        _lastPopupMenuId = null;
	}

	public String getDisplayedViewPage(LySession session, String targetWindow)
	{
		// get the last view controller
		LyController controller = session.getLastViewController();

		if (controller == null)
			controller = session.getController(0);

		if (controller != null)
			controller.showView();

		// get the page writer
		LyWebPageWriter pageWriter = getPageWriter();

		// the frame is the target frame
		String		    frame = LyWebConstants.TARGET_FRAME;

		// From this point (till we write it), the pagewriter must not change
		// !!!
		if (pageWriter != null)
		{
			synchronized (pageWriter)
			{
				// We prepare the current page
				pageWriter.preparePage(frame, targetWindow);

				// write the page
				try
				{
					StringWriter	sw = new StringWriter();

					if ("begin".equals(targetWindow))
					{
						pageWriter.writeCssScripts(sw, frame);
						pageWriter.writeBeginScripts(sw, frame);
					}
					else if ("end".equals(targetWindow))
					{
						pageWriter.writeEndScripts(sw, frame);
					}
					else
					{
						pageWriter.writePageContent(sw, frame);
					}

					return sw.toString();
				}
				catch (Exception e)
				{
					if ((pageWriter != null) && (pageWriter.isFreed() || pageWriter.isFreeing()))
						session.getEnvironment().logError("pageWriter freed while writing page; this causes following exception " + e.getMessage());
				}
			}
		}

		return null;
	}

	/**
	 * <b>If</b> the current page writer is not the given page writer,<br>
	 * <b>if</b> <code>close=true</code><br>
	 * and <b>if</b> the page writer window is a pop up window<br>
	 * <b>then</b> adds a script to hide the pop up view<br>
	 * <b>Else if</b> <code>close=true</code><br>
	 * and <b>if</b> the given page writer's window is null<br>
	 * <b>then</b> adds the current target frame to the list of frame to reload and adds the current
	 * thread to associations between thread and close requests.
	 *
	 * @param pageWriter the page writer to close.
	 * @param close      if true, indicates that the page of the page writer has to be closed
	 * @see   #getClose()
	 */
	public void setClose(LyWebPageWriter pageWriter, boolean close)
	{
		// Check if close is called on current page writer
		LyWebPageWriter currentWriter = getPageWriter();

		if (close
				&& LyWebConstants.POPUP_VIEW.equals(pageWriter.getWindowName())
				&& (getControllerStackSize(LyWebConstants.POPUP_VIEW) <= 0))
		{
//			addAdditionalRefreshScripts("<script type='text/javascript'>hidePopupView('"
//				+ LyWebConstants.POPUP_VIEW + "');</script>");

			return;
		}

		if ((currentWriter != null) && (currentWriter != pageWriter))
			return;

		if (close)
		{
			if ((pageWriter.getWindowName() == null) && !pageWriter.isPrintMode())
				addReloadFrame(LyWebConstants.TARGET_FRAME);
				// else
				// System.out.println("Skipping reload :
				// "+pageWriter.getWindowName());
				// ELSE : Closing a popup Window : nothing to do
		}

		if (_closeRequests == null)
			_closeRequests = new Hashtable<Thread, Boolean>();

		_closeRequests.put(Thread.currentThread(), Boolean.valueOf(close));
	} // end method setClose

	/**
	 * Updates the _controllerStackList parameter with the count of controllers from the
	 * controllersStackList parameter. controllersStackList must contain names of controller stacks
	 * as keys and stack of controllers as values. _controllerStackList must contain names of
	 * controller stacks as keys and stack of controllers as values.
	 *
	 * @param controllersStackList
	 */
	public void updateControllersStackList(Hashtable<String, Stack<LyViewController>> controllersStackList)
	{
		boolean	hadPopup = (_lastPopupViewController != null);

		super.updateControllersStackList(controllersStackList);

		if (!hadPopup && (_lastPopupViewController != null))
		{
			String	popupUrl = getPopupFrameAbsoluteUrl();
			String	title = _environment.translate(_lastPopupViewController.getName());

			addAdditionalRefreshScripts("<script type='text/javascript'>\n//<![CDATA[\nShowPopupView(true, '" +
					LyString.getJavascriptString(title) + "', '" + popupUrl + "');\n//]]>\n</script>");
		}
	}

	/**
	 * Encode the given URL by replacing unauthorized characters.
	 *
	 * @param  url the URL to encode
	 * @return the URL with only authorized characters
	 */
	public String encodeUrl(String url)
	{
		String newUrl;

		newUrl = LyString.replace(url, " ", "%20");
		// Needed for Javascript
		newUrl = LyString.replace(newUrl, "\\", "\\\\");
		newUrl = LyString.replace(newUrl, "\"", "&#34;");
		newUrl = LyString.replace(newUrl, "'", "\\%27");

		return newUrl;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.misc.LyViewManagerInterface#free()
	 */
	public void free()
	{
		// Generic free
		super.free();

		// free module config
		_moduleConfig = null;

		// free the servlet context
		_context = null;
	}

	/**
	 * Returns the url of the directory that contains the Javascript files.
	 *
	 * @return the url of the directory that contains the Javascript files.
	 */
	public String getScriptsDir()
	{
		String	scriptRoot = _environment.getEnv("LY_LEON_DOC") + "/scripts/struts";

		return scriptRoot;
	}

	/**
	 * Returns the path of the directory that contains the Javascript files.
	 *
	 * @return the path of the directory that contains the Javascript files.
	 */
	public String getScriptsDocDir()
	{
		String	scriptRoot = _environment.getEnv("LY_LEON_DOC_DIR") + "/scripts/struts";

		return scriptRoot;
	}

	/**
	 * Get the absolute URL for a context relative url.
	 *
	 * @param  url             A context relative URL.
	 * @param  parameters      The parameters used to build the query String.
	 * @param  encodeAmpersand Tells if the "&" character should be encoded.
	 * @return an absolute URL (scheme + server + port + context path + url + "?" + param1=value1 +
	 *         "&" + param2=value2, etc...).
	 * @throws MalformedURLException if the URL is not well formed
	 */
	public String getAbsoluteURL(String url, Map<String, String> parameters, boolean encodeAmpersand)
		throws MalformedURLException
	{
		String		 absoluteUrl = getRootURL() + url;
		StringBuffer result = new StringBuffer(absoluteUrl);

		// add the query string if needed
		if ((parameters != null) && (parameters.size() > 0))
		{
			result.append("?"); // add the query separator

			Map.Entry<String, String>    entry;
			Iterator<Map.Entry<String, String>>   it = parameters.entrySet().iterator();

			while (it.hasNext())
			{
				entry = it.next();
				result.append(entry.getKey()).append('=').append(entry.getValue());

				// add the parameter separator
				if (it.hasNext())
				{
					if (encodeAmpersand)
						result.append("&#38;");
					else
						result.append('&');
				}
			}
		}

		return result.toString();
	} // end method getAbsoluteURL

	/**
	 * Get the partial URL for a context relative url.
	 *
	 * @param  url             A context relative URL.
	 * @param  parameters      The parameters used to build the query String.
	 * @param  encodeAmpersand Tells if the "&" character should be encoded.
	 * @return an absolute URL (scheme + server + port + context path + url + "?" + param1=value1 +
	 *         "&" + param2=value2, etc...).
	 * @throws MalformedURLException if the URL is not well formed
	 */
	public String getPartialURL(String url, Map<String, String> parameters, boolean encodeAmpersand)
		throws MalformedURLException
	{
		String		 partialUrl = _contextUrl + url;
		StringBuffer result = new StringBuffer(partialUrl);

		// add the query string if needed
		if ((parameters != null) && (parameters.size() > 0))
		{
			result.append("?"); // add the query separator

			Map.Entry<String, String> entry;

			for (Iterator<Map.Entry<String, String>> it = parameters.entrySet().iterator(); it.hasNext();)
			{
				entry = it.next();
				result.append(entry.getKey()).append('=').append(entry.getValue());

				// add the parameter separator
				if (it.hasNext())
				{
					if (encodeAmpersand)
						result.append("&#38;");
					else
						result.append('&');
				}
			}
		}

		return result.toString();
	} // end method getAbsoluteURL

	/**
	 * Method that builds an absolute URL to a leonardi action configured in the current struts
	 * module.<br>
	 * This is a shortcut to the use of :<br>
	 * <b><code>String path = getActionPath(MyLeonardiAction.class.getName());<br>
	 * String relativeUrl = getActionURL(path);<br>
	 * String absoluteUrl = getAbsoluteURL(relativeUrl, parameters, encodeAmpersand);</code></b>
	 *
	 * @param  action          An action path relative to the context.
	 * @param  parameters      The parameters used to build the query String. (may be null).
	 * @param  encodeAmpersand Tells if the "&" character should be encoded.
	 * @return an absolute URL on the current server with the query string
	 */
	public String getActionAbsoluteUrl(String action, Map<String, String> parameters, boolean encodeAmpersand)
	{
		// get relative URL to the current servlet context
		String relativeUrl = getActionURL(action);

		// get absolute URL
		try
		{
			String absoluteUrl = getAbsoluteURL(relativeUrl, parameters, encodeAmpersand);

			return absoluteUrl;
		}
		catch (MalformedURLException e)
		{
			_environment.writeException(e);
			throw new IllegalStateException(
				"BF AppEngine : Could not build the absolute URL for the action " + action);
		}
	}

	/**
	 * Method that builds an partial URL to a leonardi action configured in the current struts
	 * module.<br>
	 * This is a shortcut to the use of :<br>
	 * <b><code>String path = getActionPath(MyLeonardiAction.class.getName());<br>
	 * String relativeUrl = getActionURL(path);<br>
	 * String absoluteUrl = getAbsoluteURL(relativeUrl, parameters, encodeAmpersand);</code></b>
	 *
	 * @param  action          An action path relative to the context.
	 * @param  parameters      The parameters used to build the query String. (may be null).
	 * @param  encodeAmpersand Tells if the "&" character should be encoded.
	 * @return an absolute URL on the current server with the query string
	 */
	public String getActionPartialUrl(String action, Map<String, String> parameters, boolean encodeAmpersand)
	{
		// get relative URL to the current servlet context
		String relativeUrl = getActionURL(action);

		// get absolute URL
		try
		{
			String absoluteUrl = getPartialURL(relativeUrl, parameters, encodeAmpersand);

			return absoluteUrl;
		}
		catch (MalformedURLException e)
		{
			_environment.writeException(e);
			throw new IllegalStateException(
				"BF AppEngine : Could not build the absolute URL for the action " + action);
		}
	}

	/**
	 * Method that builds an absolute URL to the leonardi popup action configured in the current struts
	 * module.<br>
	 */
	public String getPopupFrameAbsoluteUrl()
	{
		LyWebRequestInfo		info = LyStrutsServlet.getWebRequestInfo();
		HttpServletRequest		request = (info == null ? null : info.getRequest());
		String					target = (request == null ? null : (String)request.getParameter(LyWebConstants.TARGET_PARAM));
		HashMap<String, String>	parameters = new HashMap<String, String>();

		parameters.put(LyWebConstants.RELOAD, new Date().getTime() + "");

		if (target != null)
			parameters.put(LyWebConstants.TARGET_PARAM, target);
		else if (getTargetCode() != null)
			parameters.put(LyWebConstants.TARGET_PARAM, getTargetCode());

		return getActionAbsoluteUrl(LyStrutsConstants.PATH_ACTION_POPUPFRAME, parameters, false);
	}

	/**
	 * Method that builds a server relative URL for an action path.<br>
	 * It takes a context relative URL and add the URL pattern in order to generate an URL that
	 * matches the current ActionServlet mapping.<br>
	 * <u>Example:</u><br>
	 * if action is <code>/myaction</code> and the current ActionServlet URL pattern is <code>
	 * *.do</code>, then the result is <code>/myaction.do</code>.
	 *
	 * @param  action An action path relative to the context.
	 * @return an URL that matches the current ActionServlet mapping.
	 */
	public String getActionURL(String action)
	{
		// get the ActionServlet url-pattern : it is stored at startup in the
		// servlet context

		// 2009-02-11 MBI Modification for using in builder graphics a local viewManager
		// with null context.
		String urlPattern = null;

		if (_context != null)
			urlPattern = (String)_context.getAttribute(Globals.SERVLET_KEY);

		// add the pattern of the ActionServlet to the current action path
		// usually, this pattern is "*.do"
		StringBuffer sb = new StringBuffer();

		if (urlPattern == null)
			return "";

		if (urlPattern.endsWith("/*"))
		{
			sb.append(urlPattern.substring(0, urlPattern.length() - 2));
			sb.append(action);
		}
		else if (urlPattern.startsWith("*."))
		{
			if (_moduleConfig != null)
				sb.append(_moduleConfig.getPrefix());

			sb.append(action);
			sb.append(urlPattern.substring(1));
		}
		else
			throw new IllegalArgumentException(urlPattern);

		// return the action URL
		return sb.toString();
	}

	/**
	 * Returns the URL that must be used to answer messages.
	 *
	 * @return the URL that must answer to the user messages.
	 */
	public String getMessageAnswerURL()
	{
		return getActionAbsoluteUrl(LyStrutsConstants.PATH_ACTION_ANSWER, null, true);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see leon.view.LyViewManager#showLocation(leon.view.LyComponent,
	 *      java.lang.String, boolean, Hashtable<Short, String>)
	 */
	public void showLocation(LyComponent source, String location, boolean newWindow, Hashtable<Short, String> parameters)
	{
		if ((location.startsWith(LyWebConstants.TARGET_PARAM)))
		{
			// Add root url with home path before target param
			String	name = LyString.replace(location, "=", "");

			name = LyString.replace(name, ".", "");
			location = getActionAbsoluteUrl(LyStrutsConstants.PATH_ACTION_HOME, null, true) + '?' + location;

			if (parameters == null)
				parameters = new Hashtable<Short, String>(1);

			parameters.put(SHOW_LOCATION_PARAMETER_NAME, name);

			// Register newly opened window in the list of slave windows
			String	jsParams = getNewWindowScriptParameters(LyWebViewManager.encodeString(location), parameters);
			String	newWindowScript = "RegisterNewWindow(" + jsParams + ");";

			addNewWindowScript(newWindowScript);

			StringBuffer	script = new StringBuffer();

			script.append("<script type='text/javascript'>");
			script.append("getLeonardiTopFrame().RegisterNewWindow(" + jsParams + ");");
			script.append("</script>");

			addAdditionalRefreshScripts(script.toString());
		}
		else
			super.showLocation(source, location, newWindow, parameters);
	}

    /**
     * Returns the name of the Java class that must be used to instanciate a new slave session.
     *
     * @return null if default session class must be used, a full java path otherwise.
     */
    public String getSlaveSessionClassName()
    {
    	return "leon.view.web.struts.LyStrutsSession";
    }

	/**
	 * Swap two sessions.
	 * @param newSession the new session
	 * @param oldSession the old session
	 * @param applicationId the application id
	 */
	public void swapSession(Object newSession, Object oldSession, String applicationId)
	{
		// we need to swap some specific configuration: -> _root_url
		LySession oldLySession = (LySession)oldSession;
		LyViewManager viewManager = oldLySession.getViewManager();
		if (viewManager instanceof LyStrutsViewManager)
		{
			String newRoot = getRootURL();
			((LyStrutsViewManager)viewManager).setRootURL(newRoot);
		}

		LyStrutsSessionBean.swapSession(newSession, oldSession, applicationId);
	}

	/**
	 * Gets the current struts module configuration.
	 *
	 * @return the current struts module configuration
	 * @see    #setModuleConfig(ModuleConfig)
	 */
	public ModuleConfig getModuleConfig()
	{
		return _moduleConfig;
	}

	/**
	 * This methods is used by the LyStrutsSessionBean to intialize the struts module configuration
	 * at startup.
	 *
	 * @param config The struts module configuration to set.
	 * @see   #getModuleConfig()
	 */
	public void setModuleConfig(ModuleConfig config)
	{
		_moduleConfig = config;
	}

	/**
	 * Set the context Url
	 * @param contextUrl the context url
	 */
	public void setContextUrl(String contextUrl)
	{
		_contextUrl = contextUrl;
	}

	/**
	 * Gets the current servlet context.
	 *
	 * @return the current servlet context
	 * @see    #setServletContext(ServletContext)
	 */
	public ServletContext getServletContext()
	{
		return _context;
	}

	/**
	 * This methods is used by the LyStrutsSessionBean to intialize the WEB application
	 * configuration at startup.
	 *
	 * @param context The Servlet context to set.
	 * @see   #getServletContext()
	 */
	public void setServletContext(ServletContext context)
	{
		_context = context;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.LyViewManager#setProperties()
	 */
	protected void setProperties()
	{
	    super.setProperties();

        setProperty("MENU");
        setProperty("PARTIAL_MENU");
        setProperty("DOUBLE_CLICK");
        setProperty("POPUP_VIEW");
        setProperty("DHTML_NO_SCROLLING");
        setProperty("LOCAL_MAP_SELECTION");
        setProperty("DYNAMIC");
		setProperty("MULTI_FRAME_FORM");
		setProperty("ADVANCED_GANTT_FEATURES");
		setProperty("ALLOW_TABLE_HEADERS_UPDATE");
		setProperty("DISPLAY_RICH_TEXT");

		removeProperty("ASK_MULTIPLE_OBJECTS_DELETION");
		removeProperty("LAYOUT_MANAGER");
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.web.LyWebViewManager#loadBeans()
	 */
	protected void loadBeans()
	{
		// Add STRUTS beans
		loadBeans("views_struts");
	}

	public File getTemporaryFile(String fileDesc)
	{
		String path = getTemporaryDir() + "/" + fileDesc;
		File result = new File(path);
		return result;
	}

	/**
	 * Return the action that allow to download temporary resources but dependant from the
	 * viewer type (for ex: url / partial url for web).
	 */
	public String getDownloadAction()
	{
		String	actionUrl = getActionPartialUrl(LyStrutsConstants.PATH_ACTION_DOWNLOAD, null, false);
		String	params = "?";

		if (getTargetCode() != null)
			params += LyWebConstants.TARGET_PARAM + "=" + getTargetCode() + "&";

		params += "_file=";

		return actionUrl + params;
	}

    /**
     * (non-Javadoc).
     *
     * @see leon.view.LyViewManager#buildInternalView(leon.view.LyComponent)
     */
    public LyViewComponent buildInternalView(LyComponent component)
    {
        LyViewComponent internalView = super.buildInternalView(component);

        if (internalView != null)
            return internalView;

        if (_beans != null)
        {
            String tag = component.getTag();
            String beanClassName = _beans.getString(tag);

            if (beanClassName != null)
                return internalView;
        }

        if ((component instanceof LyItem) || (component instanceof LyMenu)
                || (component instanceof LyMenuBar) || (component instanceof LyPopupMenu))
            return null;

        if (_beans != null)
        {
            String tag = component.getTag();
            String beanClassName = _beans.getString(tag);

            if (beanClassName != null)
                return internalView;
        }

        if (component instanceof LyPopupMenu)
            return null;

        return buildInternalView("leon.view.web.struts", "LyStruts", component);
    } // end method buildInternalView

    /**
     * (non-Javadoc).
     *
     * @see leon.view.web.LyWebViewManager#getAdditionalRefreshScripts(boolean)
     */
    public Vector<String> getAdditionalRefreshScripts(boolean clearScripts)
    {
        // If we have any reloadDiv to to, time has come now.
        if ((_divToReplace != null)
            && !_divToReplace.isEmpty())
        {
            HashMap<String, LyWebObject> divToReplaceCopy;

            // Copy keys to avoid deadlock
            synchronized (_divToReplace)
            {
                divToReplaceCopy = new HashMap<String, LyWebObject>(_divToReplace);
                // Remove all
                _divToReplace.clear();
            }

            // propagate changes
            Iterator<String> keys = divToReplaceCopy.keySet().iterator();

            while (keys.hasNext())
            {
                String	    key = keys.next();
                int         n1 = key.indexOf("###");

                // Retrieve all information about the frame, the div and the
                // content
                String      frame = key.substring(0, n1);
                String      divId = key.substring(n1 + 3);
                LyWebObject obj = divToReplaceCopy.get(key);

                if (obj.isFreed() || obj.isFreeing())
                    continue;

                // Now try to build the HTML content
                LyComponent viewComp = obj.getViewComponent();

                if ((viewComp != null)
                    && !viewComp.isFreeing()
                    && !viewComp.isFreed()
                    && (viewComp.getEnvironment() != null)
                    && (viewComp.getViewManager() != null))
                {
                    LyWebPageWriter     pw = new LyStrutsPage(viewComp, true);
                    StringBuffer        script = new StringBuffer(200);

                    if (obj.getLastFrameName() != null)
                        pw.setActiveBuffer(obj.getLastFrameName());

                    // If div to replace comes from a form field, generate item only
                    if (obj instanceof LyStrutsMenuBar)
                    {
                        LyStrutsMenuBar     webMenuBar = (LyStrutsMenuBar)obj;
                        String              content = webMenuBar.getRefreshContent(pw);

                        content = LyString.replace(content, "\\", "\\\\");
                        content = LyString.replace(content, "\\\\n", "\\n");
                        content = LyString.replace(content, "\"", "\\\"");
                        content = LyString.replace(content, "\\'", "\\\\'");

                        script.append("<script type='text/javascript'>\n//<![CDATA[\n");
                        script.append(content);
                        script.append("\n//]]>\n</script>\n");
                    }
					else if (obj instanceof LyStrutsRelation && (((LyRelation)viewComp).getAutoCompletionMax() > 0))
					{
						LyStrutsRelation	webRelation = (LyStrutsRelation)obj;
						String				content = webRelation.getRefreshContent(pw);

						script.append("<script type='text/javascript'>\n//<![CDATA[\n");
						script.append(content);
						script.append("\n//]]>\n</script>\n");
					}
                    else
                    {
                        if (obj instanceof LyStrutsFormField)
                            ((LyStrutsFormField)obj).writeItem(pw, false);
                        else
                            obj.writeHtml(pw);

                        StringBuffer innerHtml = pw.getBuffer(pw.getActiveName());
                        String       innerHtmlStr = innerHtml.toString();

                        innerHtmlStr = LyString.replace(innerHtmlStr, "\\", "\\\\");
                        innerHtmlStr = LyString.replace(innerHtmlStr, "\"", "\\\"");
                        innerHtmlStr = LyString.replace(innerHtmlStr, "\\'", "\\\\'");
                        innerHtmlStr = LyString.replace(innerHtmlStr, "\\\\n", "\\n");
                        innerHtmlStr = LyString.replace(innerHtmlStr, "\n", "\\n");

                        script.append("<script type='text/javascript'>\n//<![CDATA[\nreplaceDivInnerHtml(\"");
                        script.append(frame).append("\",\"");
                        script.append(divId).append("\",\"");
                        script.append(innerHtmlStr).append("\");\n//]]>\n</script>\n");
                    }

                    addAdditionalRefreshScripts(script.toString(), false);
                    pw.free();
                }
            }
        }

        if (_lastPopupMenuId != null)
        {
            if (_lastPopupMenuId.startsWith("RESET"))
                _lastPopupMenuId = _lastPopupMenuId.substring("RESET".length());
            else
            {
                StringBuffer    hideMenu = new StringBuffer();

                hideMenu.append("<script type='text/javascript'>\n//<![CDATA[\n");
                hideMenu.append("var subMenu = document.getElementById('cmPopupSubMenuID' + cmLeonardiLastPopupMenuId);");
                hideMenu.append("if (subMenu != null) {");
                hideMenu.append("var info = new Object ();");
                hideMenu.append("info.div = '" + _lastPopupMenuId + "';");
                hideMenu.append("info.menu = '';");
                hideMenu.append("info.orient = 'hbr';");
                hideMenu.append("info.nodeProperties = cmLeonardiPopup;");
                hideMenu.append("info.prefix = 'Leonardi';");
                hideMenu.append("cmHideSubMenu(subMenu, info);}");
                hideMenu.append("\n//]]>\n</script>\n");

                addAdditionalRefreshScripts(hideMenu.toString());
                _lastPopupMenuId = null;
            }
        }

        return super.getAdditionalRefreshScripts(clearScripts);
    } // end method getAdditionalRefreshScripts

    /**
     * Adds reload of a DIV in a frame.
     *
     * @param frame   frame which will need to be refreshed next time
     * @param divId   the identifier of the div tag wich needs refresh
     * @param content the HTML content for replacement
     */
    public void addReloadDiv(String frame, String divId, LyWebObject content)
    {
        if ((frame == null) || (frame.length() == 0))
            return;
        if ((divId == null) || (content == null))
            return;

        if (_divToReplace == null)
            _divToReplace = new HashMap<String, LyWebObject>();

        synchronized (_divToReplace)
        {
            if (frame.startsWith(LyWebConstants.POPUP_VIEW))
                frame = LyWebConstants.POPUP_VIEW;
            frame = frame.intern();
            if (frame == LyWebPageWriter.MAIN_BUFFER)
                frame = LyWebConstants.TARGET_FRAME; // Anacronisme : why
                                                     // MAIN ?

            _divToReplace.put(frame + "###" + divId, content);
        }
    }

    /**
     * Returns the list of HTML division that needs to be reloaded.
     */
    public HashMap<String, LyWebObject> getReloadDivs()
    {
        return _divToReplace;
    }

	/**
	 * Indicates whether the client needs an update.
	 */
	public boolean needsUpdate()
	{
		HashMap<String, LyWebObject>	reloadDivs = getReloadDivs();

		if ((reloadDivs != null) && !reloadDivs.isEmpty())
			return true;

		return super.needsUpdate();
	}

    /**
     * Sets the popup menu with the given value.
     *
     * @param popup the new popup menu
     */
    public void setPopupMenu(LyPopupMenu popup, String popupMenuDivId)
    {
        int             i, menuCount = (popup == null ? 0 : popup.getComponentCount());

        if (menuCount == 0)
        	return;

        LyComponent     subComponent;
        LyStrutsMenu    webMenu;
        LyStrutsItem    webItem;
        StringBuffer    divContent = new StringBuffer();

        // Build a dummy main item, hidden for the end-user.
        divContent.append("var popupMenu = [");
        divContent.append("['&nbsp;','', null, null, '',");

        for (i = 0; i < menuCount; i++)
        {
            if (i > 0)
                divContent.append(",");

            subComponent = popup.getComponent(i);

            if (subComponent instanceof LyMenu)
            {
                webMenu = (LyStrutsMenu)subComponent.getInternalView();
                divContent.append(webMenu.writeContent(null, null));
            }
            else if (subComponent instanceof LyItem)
            {
                webItem = (LyStrutsItem)subComponent.getInternalView();
                divContent.append(webItem.writeContent(null, null));
            }
        }

        divContent.append(" ]");
        divContent.append("];");

        divContent.append("if (cmLeonardiPopup != null)\n\tcmLeonardiPopup.folderRight = '<img src=\""
            + getEnvironment().getImageUrl("LY_MENU_ARROW") + "\"/>';");

        // Change the name of sub-menus to avoid interferences with frame menus
        divContent.append("_cmIDName = 'cmPopupSubMenuID';");
        divContent.append("cmLeonardiLastPopupMenuId = _cmIDCount + 1;");

        // Create the sub-menus then show the first one which is the "real" popup menu
        divContent.append("var hiddenMenu = document.getElementById('LeonardiMainItem0');");

        // If an hiden menu already exists, it may come from another component. Just change its
        // identifier to ignore it.
        divContent.append("if ((hiddenMenu != null) && (typeof hiddenMenu != 'undefined')) hiddenMenu.id = 'oldPopup_"
                + System.currentTimeMillis() + "';");
        // Remove effect since popup menu may not be displayed at the right place with effect
        divContent.append("cmLeonardiPopup.effect = null;");
        divContent.append("cmDraw ('" + popupMenuDivId + "', popupMenu, 'hbr', cmLeonardiPopup, 'Leonardi');");
        divContent.append("hiddenMenu = document.getElementById('LeonardiMainItem0');");
        divContent.append("var subMenu = document.getElementById('cmPopupSubMenuID' + cmLeonardiLastPopupMenuId);");
        divContent.append("SetDivPosition('");
        divContent.append(popupMenuDivId);

        // Use an offset to display popup menu at the right place.
        // Vertical offset is due to hidden dummy div.
		int					xOffset = -4, yOffset = -26;
		LyWebRequestInfo	info = getWebRequestInfo();
		HttpServletRequest	request = (info == null ? null : info.getRequest());
		String				userAgent = (request == null ? null : request.getHeader("User-Agent"));

		// On Linux popup menu will appear when mouseDown event is thrown : reduce the offset to avoid
		// an automatic clic on first popup menu entry.
		if ((userAgent != null) && (userAgent.toUpperCase().indexOf("LINUX") >= 0))
		{
			xOffset = 2;
			yOffset = -20;
		}

		divContent.append("', " + (popup.getX() + xOffset) + ", " + (popup.getY() + yOffset) + ");");
        divContent.append("var info = new Object ();");
        divContent.append("info.div = '" + popupMenuDivId + "';");
        divContent.append("info.menu = popupMenu;");
        divContent.append("info.orient = 'hbr';");
        divContent.append("info.nodeProperties = cmLeonardiPopup;");
        divContent.append("info.prefix = 'Leonardi';");
        divContent.append("cmShowSubMenu(hiddenMenu, 'Leonardi', subMenu, info);");

        if (getBrowser() == BROWSER_CHROME)
        {
            // On Chrome, popup menu will not be displayed if popup div is hidden.
            // Hide hidden menu to avoid displaying an empty parent menu.
            divContent.append("var topPopupMenu = document.getElementById('" + popupMenuDivId + "');");
            divContent.append("topPopupMenu.style.visibility = 'visible';");
            divContent.append("hiddenMenu.style.display = 'none';");
        }

        StringBuffer        sb = new StringBuffer();

        sb.append("<script type='text/javascript'>\n//<![CDATA[\n");
        sb.append(divContent).append("\n");

        // Reset the popup position to avoid popup from being displayed where the last popup menu has been displayed
        sb.append("popupLeftOffset = popupDefaultLeftOffset;popupTopOffset = popupDefaultTopOffset;");
        sb.append("\n//]]>\n</script>");

        addAdditionalRefreshScripts(sb.toString());
        _lastPopupMenuId = "RESET" + popupMenuDivId;
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.misc.LyViewManagerInterface#getColor(leon.misc.LyColor)
     */
    public Object getColor(LyColor color)
    {
        if (color == null)
            return null;

        return new Color(color.getRGB());
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.misc.LyViewManagerInterface#getFont(leon.misc.LyFont)
     */
    public Object getFont(LyFont font)
    {
        if (font == null)
            return LyStrutsDrawingArea.getDefaultFrame().getFont();

        int style = 0;

        if (font.isBold())
            style |= Font.BOLD;
        if (font.isItalic())
            style |= Font.ITALIC;

        return new Font(font.getFontName(), style, font.getSize());
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.misc.LyViewManagerInterface#getStringWidth(leon.misc.LyFont, java.lang.String)
     */
    public int getStringWidth(LyFont font, String text)
    {
        Font        f = (Font)getFont(font);
        FontMetrics fm = getDefaultFrame().getFontMetrics(f);

        return fm.stringWidth(text);
    }

    /**
     * Construct a default image.
     *
     * @see LyImageUtils#getDefaultImage()
     * @see leon.misc.LyViewManagerInterface#getDefaultImage()
     */
    public Object getDefaultImage()
    {
        return LyImageUtils.getDefaultImage();
    }

    /**
     * Gets images sizes.
     *
     * @see LyImageUtils#getSizes(Image)
     * @see leon.misc.LyViewManagerInterface#getSizes(Object)
     */
    public int[] getSizes(Object image)
    {
        return LyImageUtils.getSizes((Image)image);
    }

    /**
     * Creates a scaled version of the given image.
     *
     * @see LyImageUtils#getScaledInstance(Image, int, int)
     * @see leon.misc.LyViewManagerInterface#getScaledInstance(Object, int, int)
     */
    public Object getScaledInstance(Object image, int width, int height)
    {
        return LyImageUtils.getScaledInstance((Image)image, width, height);
    }

    /**
     * Waits for the image loading if the headless mode is null or set to true.
     *
     * @see LyImageUtils#waitForLoading(Component, Image)
     * @see leon.misc.LyViewManagerInterface#waitForLoading(Object, Object)
     */
    public void waitForLoading(Object component, Object image)
        throws InterruptedException
    {
        // In headless mode, waitForLoading can't be used because the method
        // creates a new Frame
        Image  pic = (Image)image;
        int    counter = 0;

        if (GraphicsEnvironment.isHeadless())
        {
             while ((pic.getWidth(null) < 0) && (counter < WAIT_COUNTER))
             {
                 Thread.sleep(20);
                 counter++;
             }
        }
        else
            LyImageUtils.waitForLoading((Component)component, (Image)image);
    }

    /**
     * Retrieves an image from a Stream.
     *
     * @see LyImageUtils#loadImageFromStream(Component, InputStream)
     * @see leon.misc.LyViewManagerInterface#loadImageFromStream(Object, InputStream)
     */
    public Object loadImageFromStream(Object component, InputStream resource)
        throws IOException
    {
        return LyImageUtils.loadImageFromStream((Component)component, resource);
    }

    /**
     * Compare two images pixel by pixel.
     *
     * @see LyImageUtils#isSameImage(Image, Image)
     * @see leon.view.LyViewManager#isSameImage(java.lang.Object, java.lang.Object)
     */
    public boolean isSameImage(Object image1, Object image2)
        throws InterruptedException
    {
        return LyImageUtils.isSameImage((Image)image1, (Image)image2);
    }

    /**
     * Free the resources of the image if not in the cache.
     *
     * @see LyImageUtils#freeImage(Image)
     * @see leon.misc.LyViewManagerInterface#freeImage(java.lang.Object)
     */
    public void freeImage(Object image)
    {
        LyImageUtils.freeImage((Image)image);
    }

    /**
     * (non-Javadoc).
     *
     * @see leon.misc.LyViewManagerInterface#exit(int)
     */
    public void exit(int code)
    {
        // Do some cleaning
        LyImageUtils.clean();
        LyStrutsViewManager.removeDefaultFrame();
        super.exit(code);
    }

    /**
     * Copy a given text to the system clipboard.
     *
     * @param text the text to be copied to the clipboard.
     */
    public void copyToClipboard(String text)
    {
        // Nothing to do in web mode
    }

    /**
     * Check if this view manager is compatible with the given view manager
     * @param otherViewManager the other view manager
     * @return true if both view manager are compatible
     */
    public boolean areViewManagerCompatible(LyViewManager otherViewManager)
    {
        boolean result = super.areViewManagerCompatible(otherViewManager);

        if (!result)
            return result;
        if (this.getBrowser() == ((LyStrutsViewManager)otherViewManager).getBrowser())
            return true;

        return false;
    }

	/**
	 * Return the current web request info object.
	 */
	protected LyWebRequestInfo getWebRequestInfo()
	{
		return LyStrutsServlet.getWebRequestInfo();
	}

	/**
	 * Indicates whether push mode is activated or not.
	 * Result depends on the LY_WEB_PUSH_MODE resource.
	 *
	 * @return true is push mode is activated, false otherwise.
	 */
	public boolean isPushModeActivated()
	{
		LyEnvironment	env = getEnvironment();
		boolean			pushMode = (env == null ? false : env.testResource("LY_WEB_PUSH_MODE"));

		return pushMode;
	}
} // end class LyStrutsViewManager
