/*
 * LyStrutsLeonardiAction.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.view.web.struts.action;

import org.apache.struts.action.*;
import org.apache.struts.config.*;
import java.io.*;
import java.text.*;
import java.util.*;
import java.util.zip.*;
import javax.servlet.http.*;

import leon.app.*;
import leon.control.*;
import leon.data.*;
import leon.info.*;
import leon.misc.*;
import leon.view.web.*;
import leon.view.web.struts.*;


/**
 * Type : leon.view.web.struts.LyStrutsLeonardiAction<br>
 * This Leonardi class implements the Action layer of the Struts framework.<br>
 * This action is abstract and contains only usefull methods to be used for all Leonardi Struts
 * actions.
 *
 * @version $Id: LyStrutsLeonardiAction.java 40694 2012-09-06 15:32:54Z nhamon $
 * @author  Lyria S.A.
 */
public abstract class LyStrutsLeonardiAction extends Action
{
	public static final String __VERSION =
		"$Id: LyStrutsLeonardiAction.java 40694 2012-09-06 15:32:54Z nhamon $";

	/**
	 * Maximum length of Ajax refresh scripts before an entire refresh will be performed.
	 * A negative value indicates that refresh scripts are always sent to the client.
	 * It is not recommended to set a value lower than 100000.
	 */
	public static int	ADDITIONAL_SCRIPTS_MAX_SIZE = -1;

	/**
	 * Maximum length of generated pages.
	 * A negative value indicates that pages are always sent to the client.
	 */
	public static int	GENERATED_PAGE_MAX_SIZE = -1;
	
	/**
	 * This basic implementation of the Struts Action Model returns null (no response).
	 *
	 * @param  mappings - The ActionMapping used to select this instance
	 * @param  bean     - The optional ActionForm bean for this request (if any)
	 * @param  request  - The HTTP request we are processing
	 * @param  response - The HTTP response we are creating
	 * @return null : no response.
	 * @throws Exception
	 */
	public ActionForward execute(ActionMapping mappings, ActionForm bean,
		HttpServletRequest request, HttpServletResponse response)
		throws Exception
	{
		return null;
	}
	
	/**
	 * Check if current session is valid and redirect to home page if not.
	 * 
	 * @param  bean     - The optional ActionForm bean for this request (if any)
	 * @param  request  - The HTTP request we are processing
	 * @param  response - The HTTP response we are creating
	 * @return true if session is valid, false otherwise.
	 */
	public boolean checkSessionRegistered(LyStrutsSessionBean bean, HttpServletRequest request,
	        HttpServletResponse response)
		throws Exception
	{
		LySession session = bean.getSession();
		boolean   isSessionValid = ((session != null)
                        && session.isRegistered()
                        && (session.getState() != LySession.State.CLOSING)
                        && (session.getState() != LySession.State.CLOSED));
		
		if (isSessionValid)
		    return true;
		
		LyStrutsViewManager   viewManager = bean.getViewManager();
		
		if (viewManager == null)
		    return false;
		
		String                ajax = getParameter(request, LyWebConstants.AJAX);

		if ("true".equals(ajax))
		{
			// If Ajax mode, send a Javascript to reload Leonardi top frame
			Writer	pw = getPrintWriter(request, response, viewManager.getEnvironment(), false);

			writeJavaScriptCode(pw, "getLeonardiTopFrame().location.reload();");
		}
		else
		{
			// make the redirection
			String url = viewManager.getActionAbsoluteUrl(LyStrutsConstants.PATH_ACTION_HOME, null, false);

			response.sendRedirect(url);
		}
		
		return false;
	}

	/**
	 * Performs some specific operations at the begining of a request.<br>
	 * Do the following operations:
	 *
	 * <ul>
	 *   <li>Log the current request</li>
	 *   <li>Set the response content type</li>
	 *   <li>Set the response status</li>
	 *   <li>Set the user-agent on the view manager</li>
	 *   <li>Set the no-cache property in the HTTP response header if _LY_NO_CACHE is on</li>
	 *   <li>Set the ajax mode if the AJAX parameter is found in the request</li>
	 * </ul>
	 *
	 * @param request     The current HTTP request.
	 * @param response    The current HTTP response.
	 * @param viewManager The current view manager.
	 * @return the output writer.
	 */
	public Writer beginRequest(HttpServletRequest request, HttpServletResponse response,
		LyStrutsViewManager viewManager)
	{
	    LyEnvironment   environment = viewManager.getEnvironment();
	    
	    LyEnvironment.setCurrentEnvironment(environment);
	       
		// Set the ajax mode if needed
		String	ajaxParam = getParameter(request, LyWebConstants.AJAX);
		boolean	isAjax = "true".equals(ajaxParam);

		if (isAjax)
			setNoCacheHeaders(response, 0);
		
		response.setContentType("text/html;charset=" + LyString.DEFAULT_CHARSET);
		
		// log the request
		Writer			pw = getPrintWriter(request, response, viewManager.getEnvironment(), false);
		
		logRequest(request, viewManager.getEnvironment());

		response.setStatus(HttpServletResponse.SC_OK);

		// set the navigator type
		String userAgent = request.getHeader("User-Agent");

		if (userAgent != null)
			viewManager.setBrowser(userAgent);

		// set the no cache property
		if (environment.testEnv("_LY_NO_CACHE"))
			setNoCacheHeaders(response, 0);
		
		// Register exit info if answer leads to the exit
		addRequestAttributes(request, response, viewManager);
		
		return pw;
	} // end method beginRequest

	/**
	 * Sets some attributes on the request.
	 *
	 * @param request     The current HTTP request.
	 * @param response    The current HTTP response.
	 * @param viewManager The current view manager.
	 * @return the output writer.
	 */
	protected void addRequestAttributes(HttpServletRequest request, HttpServletResponse response,
			LyStrutsViewManager viewManager)
	{
		LyEnvironment	environment = viewManager.getEnvironment();

		String	cssPath = environment.getUrl("LY_CSS_FILE");
		String	logo = environment.getImageUrl("LY_COMPANY_LOGO");

		request.setAttribute("language", environment.getLanguage());
		request.setAttribute("cssPath", cssPath);
		request.setAttribute("logo", logo);
		request.setAttribute(LyWebConstants.TARGET_PARAM, request.getParameter(LyWebConstants.TARGET_PARAM));
	}
	
	/**
	 * Decodes an URL by replacing unauthorized characters.
	 *
	 * @param  url The URL to decode.
	 * @return The URL without unauthorized characters.
	 */
	public static String decodeUrl(String url)
	{
		String newUrl;

		newUrl = LyString.replace(url, "%20", " ");
		newUrl = LyString.replace(newUrl, "\\\\", "\\");

		return newUrl;
	}

	/**
	 * Performs some specific operations at the end of a request.<br>
	 * Clear the current page writer and the current window name.
	 *
	 * @param request     The current HTTP request.
	 * @param response    The current HTTP response.
	 * @param viewManager The current view manager.
	 */
	public void endRequest(HttpServletRequest request, HttpServletResponse response, LyStrutsViewManager viewManager,
			Writer out)
		throws IOException
	{
		// Clear page writer
		viewManager.setPageWriter(null);

		// Clear current window context
		viewManager.setCurrentWindowName(null);
		
		out = dumpToFile(viewManager, request, response, out);
		out.close();
		logEndRequest(request, viewManager.getEnvironment());
		
		LyEnvironment.setCurrentEnvironment(null);
	}

	/**
	 * Dump HTML content to files for debug.
	 *
	 * @param  viewManager the view manager
	 * @param  request     the request
	 * @param  response    the response
	 * @param  out         the out writer
	 * @return the writer  with the writed content
	 * @throws IOException
	 */
	public Writer dumpToFile(LyStrutsViewManager viewManager, HttpServletRequest request,
		HttpServletResponse response, Writer out)
		throws IOException
	{
		if ((LyWebConstants.WRITTEN_PAGES_COUNTER >= 0) && (out instanceof StringWriter))
		{
			LyEnvironment environment = viewManager.getEnvironment();
			String		  content = out.toString();

			out = getPrintWriter(request, response, environment, true);
			out.write(content); // Write result to real stream

			// Write debug
			if ((content != null) && (content.length() > 0))
			{
				LyWebRequestInfo	info = LyStrutsServlet.getWebRequestInfo();
				int					requ_count = (info == null ? LyWebConstants.REQU_COUNT : info.getRequestCount());
				DecimalFormat		ft = new DecimalFormat("00000");
				String		 		file = "$LY_APP_DIR$/log/" + ft.format(new Integer(requ_count))
					+ "_Page.html";

				file = environment.getString().expand(file);
				File f = new File(file);

				f.mkdirs();

				if (f.exists())
					f.delete();

				PrintWriter output = new PrintWriter(new FileWriter(f));

				// DEBUG Check if XHTML
				boolean     parse_xhtml = false;

				if ((environment.getEnv("LY_DEBUG_LEVEL") != null)
						&& (environment.getEnv("LY_DEBUG_LEVEL").indexOf("DEBUG_VALIDATE_XHTML") >= 0))
					parse_xhtml = true;

				if (parse_xhtml && (content.length() > 5))
					output.write("<!DOCTYPE html SYSTEM 'http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd'>\n");
				
				output.write(content);
				output.close();

				// REPARSE RESULT
				if (parse_xhtml && (content.length() > 5))
				{
					LyXmlValidator validator = new LyXmlValidator();

					validator.validate(f);
				}
			}
		}

		return out;
	} // end method dumpToFile

	/**
	 * Generates the scripts created by web components in the standard output.
	 * 
	 * @param session  The current session.
	 * @param request  The current HTTP request.
	 * @param response The current HTTP response.
	 * @param out      The response output.
	 * @throws IOException
	 */
	public static void generateAdditionalScripts(LySession session, HttpServletRequest request, HttpServletResponse response,
			Writer out)
		throws IOException
	{
		// get a writer
		LyEnvironment		environment = session.getEnvironment();
		LyStrutsViewManager viewManager = (LyStrutsViewManager)session.getViewManager();

		if (viewManager == null)
			return;
		
		synchronized (viewManager)
		{
			Writer			 	pw = out;
			
			if (pw == null)
				pw = getPrintWriter(request, response, environment, false);

			boolean showPopup = (viewManager.getLastPopupViewController() != null);

			if (!showPopup)
			{
				String	popupUrl = viewManager.getPopupFrameAbsoluteUrl();
				
				pw.write("<script type='text/javascript'>\n//<![CDATA[\n");
				pw.write("ShowPopupView(false, '', '" + popupUrl + "');");
				pw.write("\n//]]>\n</script>\n");
			}

			Vector<String>	reloadFrames = viewManager.getReloadFrames();
			
			if (reloadFrames != null)
			{
				String			reloadFrame;
				
				for (int i = 0; i < reloadFrames.size(); i++)
				{
					if (reloadFrames.elementAt(i) == null)
						continue;

					reloadFrame = reloadFrames.elementAt(i);
					pw.append(getReloadFrameScript(session, pw, reloadFrame));
				}
					
				viewManager.clearReloadFrames();
			}
			
			Vector<String>	addScripts = viewManager.getAdditionalRefreshScripts(true);
			Object		  	script;
			String			scriptString;
			StringBuffer	scriptBuffer = new StringBuffer();
			int				scriptsSize = 0;

			if (addScripts != null)
			{
				for (int i = 0; (addScripts != null) && (i < addScripts.size()); i++)
				{
					script = addScripts.get(i);
					
					if (script != null)
					{
						scriptString = script.toString();
						scriptsSize += scriptString.length();
						scriptBuffer.append(scriptString);
						scriptBuffer.append("\n");
					}
				}
			}

			if ((ADDITIONAL_SCRIPTS_MAX_SIZE > 0) && (scriptsSize > ADDITIONAL_SCRIPTS_MAX_SIZE))
				// Scripts size is greater than ADDITIONAL_SCRIPTS_MAX_SIZE : perform a full refresh
				pw.append(getReloadFrameScript(session, pw, LyWebConstants.TARGET_FRAME));
			else
				pw.append(scriptBuffer.toString());

			String message = viewManager.getMessage();
			
			if (message != null)
			{
				pw.write(message);
				viewManager.setMessage(null);
			}
			
			// Add refresh script only if a refresh is needed (refresh > 0) and refresh value is lower than
			// default parameter. Otherwise refresh will be performed automaticaly in the LyStrutsRefreshAction.
			// Do not perform refresh in push mode since updates are managed by the push action.
			boolean	pushMode = environment.testResource("LY_WEB_PUSH_MODE");
			boolean	replayMode = environment.testEnv("LY_REPLAY");
			int		refreshViews = LyTools.intFromString(environment.getResource("LY_AUTO_REFRESH_VIEWS"));
			int		refresh = viewManager.getRefreshRate();
			
			if (!pushMode
					&& !replayMode
					&& (refresh > 0)
					&& ((refreshViews <= 0) || (refresh < refreshViews)))
			{
				pw.write("<script type='text/javascript'>\n//<![CDATA[\n");
				pw.write("setTimeout(\"getLeonardiTopFrame().AutoRefresh(false)\", " +
						(refresh * 1000) + ");");
				pw.write("\n//]]>\n</script>\n");
			}
		}
	}
	
	/**
	 * Generates the script that must be used to perform a frame reload.
	 */
	protected static String getReloadFrameScript(LySession session, Writer pw, String reloadFrame)
	{
		LyStrutsViewManager viewManager = (LyStrutsViewManager)session.getViewManager();
		LyEnvironment	environment = session.getEnvironment();
		StringBuffer	buffer;
		String			checkCtrlId;
		int				sepPos;

		sepPos = reloadFrame.indexOf(LyWebViewManager.SEP_PW_ID);
		
		if (sepPos > 0)
		{
			checkCtrlId = reloadFrame.substring(sepPos + LyWebViewManager.SEP_PW_ID.length());
			reloadFrame = reloadFrame.substring(0, sepPos);
		}
		else
			checkCtrlId = null;
		
		buffer = new StringBuffer();
		buffer.append("<script type='text/javascript'>\n//<![CDATA[\n");
		buffer.append("var reloadFrame = findTarget(null,'" + reloadFrame + "'");

		if (checkCtrlId != null)
			buffer.append(",'").append(checkCtrlId).append("'");
		
		buffer.append(");");
					buffer.append("if ((reloadFrame != null) && ((typeof reloadFrame) != 'undefined')) {");

		HashMap<String, String>	parameters = new HashMap<String, String>();
		String					target = viewManager.getTargetCode();
		
		if (target != null)
			parameters.put(LyWebConstants.TARGET_PARAM, target);

		if (LyWebConstants.TARGET_FRAME.equals(reloadFrame))
		{
			boolean	urlNavigation = environment.testResource("LY_WEB_DIRECT_URL_NAVIGATION");
			String	actionAbsoluteUrl = viewManager.getActionAbsoluteUrl(LyStrutsConstants.PATH_ACTION_TARGETFRAME, parameters, false);
						
			if (urlNavigation)
			{
				String	redirect = (String)session.getValue("TARGET_ACTION_URL");
							
				if (redirect != null)
					actionAbsoluteUrl = redirect;
			}
						
			buffer.append("reloadFrame.location.replace('" + actionAbsoluteUrl + "');");
		}
		else if (LyWebConstants.POPUP_VIEW.equals(reloadFrame))
		{
			String	actionAbsoluteUrl = viewManager.getActionAbsoluteUrl(LyStrutsConstants.PATH_ACTION_POPUPFRAME, parameters, false);
			buffer.append("reloadFrame.location.replace('" + actionAbsoluteUrl + "');");
		}
		else
			buffer.append("reloadFrame.location.reload();");

		buffer.append("}");
		buffer.append(";\n//]]>\n</script>\n");
		
		return buffer.toString();
	}
	
	/**
	 * Removes the context path from URLS to enable forwarding into a JSP.<br>
	 * Make a path relative to the Servlet context for the given page.<br>
	 * The path beeing build is relative to the root directory of the current WEB application.
	 *
	 * @param  request Current HTTP request.
	 * @param  page    A JSP or HTML page with possibly a context path.
	 * @return the page ready for a forward.
	 */
	public String getContextReativePath(HttpServletRequest request, String page)
	{
		if ((request == null) || (page == null))
			return page;

		String contextPath = request.getContextPath();

		// We do a forward, so this url may be relative to current servlet
		if (page.startsWith(contextPath))
			return page.substring(contextPath.length());
		else
			return page;
	}

	/**
	 * Retrieves a parameter in an http request and decode its value.<br>
	 * If the value is multiple, this method just returns the last value.
	 *
	 * @param  request The HTTP request
	 * @param  name    The name of the parameter.
	 * @return The value found in the given HTTP request.
	 */
	public static String getParameter(HttpServletRequest request, String name)
	{
		String[] values = request.getParameterValues(name);

		if (values == null)
			return null;

		// Just return the last value
		String value = values[values.length - 1];

		value = decodeUrl(value);
		return value;
	}

	/**
	 * Sets an expiration delay used by the cache : modify the HTTP response header.
	 *
	 * @param response Current HTTP response.
	 * @param delay    The expiration delay.
	 */
	public void setNoCacheHeaders(HttpServletResponse response, int delay)
	{
		try
		{
			// Always control the pages.
			response.setHeader("Cache-control", "no-cache");
			response.setHeader("Pragma", "no-cache");
			response.setDateHeader("Expires", (new Date()).getTime() + delay);
		}
		catch (Exception e)
		{
			// An exception has been reported in some cases when using WebSphere
		}
	}

	/**
	 * Stores some usefull information as attribute inside an HTTP request. Then forward to a JSP:
	 *
	 * <ul>
	 *   <li>leonardiSession : The current LySession</li>
	 *   <li>servletRootURL : The current absolute URL of the servlet</li>
	 *   <li>leonardiRootURL : The URL to the leonardi documents: $LY_LEON_DOC$</li>
	 *   <li>appRootURL : The URL to the application documens: $LY_APP_DOC$</li>
	 * </ul>
	 *
	 * @param request Current HTTP request.
	 * @param session Current user session
	 */
	public void setRequestAttributes(HttpServletRequest request, LySession session)
	{
		if (session == null)
			return;

		request.setAttribute(LyWebConstants.RA_LY_SESSION, session);
		request.getSession().setAttribute(LyWebConstants.RA_LY_SESSION, session);

		LyEnvironment environment = session.getEnvironment();

		if (environment != null)
		{
			LyWebViewManager viewManager = (LyWebViewManager)environment.getViewManager();

			request.setAttribute(LyWebConstants.RA_LY_ROOT_URL, viewManager.getRootURL());

			// set the leonardi root url
			String leonardiRootUrl;
			String leonardiExternalUrl;
			String LEON_DOC = "$LY_LEON_DOC$";
			String LEON_DOC_EXTERNAL = "$LY_LEON_DOC$";
			String cssPath = environment.getUrl("LY_CSS_FILE");

			if (environment.getEnv("_LY_LEON_DOC_EXTERNAL") != null)
				LEON_DOC_EXTERNAL = "$_LY_LEON_DOC_EXTERNAL$";

			leonardiRootUrl = environment.getString().expandUrl(LEON_DOC);
			leonardiExternalUrl = environment.getString().expandUrl(LEON_DOC_EXTERNAL);

			request.setAttribute(LyWebConstants.RA_LY_LEON_DOC_URL, leonardiRootUrl);
			request.setAttribute(LyWebConstants.RA_LY_LEON_DOC_EXTERNAL_URL, leonardiExternalUrl);
			request.setAttribute(LyWebConstants.RA_LY_SKIN_URL, cssPath);
			request.setAttribute(LyWebConstants.RA_LY_APP_DOC_URL,
				environment.getString().expandUrl("$LY_APP_DOC$"));
			request.getSession().setAttribute(LyWebConstants.RA_LY_LEON_DOC_URL, leonardiRootUrl);
			request.getSession().setAttribute(LyWebConstants.RA_LY_LEON_DOC_EXTERNAL_URL, leonardiExternalUrl);
			request.getSession().setAttribute(LyWebConstants.RA_LY_SKIN_URL, cssPath);
			request.getSession().setAttribute(LyWebConstants.RA_LY_APP_DOC_URL,
				environment.getString().expandUrl("$LY_APP_DOC$"));
		}
	} // end method setRequestAttributes

	/**
	 * Computes the new reload state of the current page writer.
	 *
	 * @param pageWriter Current page writer.
	 * @param controller Controller of the view beeing written.
	 * @param response   Current HTTP response.
	 */
	protected void checkReloadState(LyWebPageWriter pageWriter, LyController controller,
		HttpServletResponse response)
	{
		long currentTime;

		if (pageWriter != null)
			currentTime = pageWriter.getLastReloadTime() / 1000;
		else
			currentTime = new Date().getTime() / 1000;
		currentTime = (currentTime * 1000) - 1000;

		long time = controller.getLastModified() / 1000;

		time = time * 1000;

		if (time >= currentTime)
		{
			if (pageWriter != null)
				pageWriter.setForceReloadState(true);

			if (response != null)
				response.setHeader("Last-Modified", "" + time);
		}
	}

	/**
	 * This method builds a dynamic action forward to a Leonardi Struts action.
	 *
	 * @param  path        The path of the action class.
	 * @param  viewManager The current view manager.
	 * @param  config      The current struts module configuration.
	 * @return an ActionForward instance.
	 */
	protected ActionForward getForwardToAction(String path, LyStrutsViewManager viewManager,
		ModuleConfig config)
	{
		// the view manager may have been freed meanwhile : set module config
		// and servlet context
		// this happens while closing the user session
		boolean freed = (viewManager.getModuleConfig() == null);

		if (freed)
		{
			viewManager.setModuleConfig(config);
			ActionServlet servlet = getServlet();

			viewManager.setServletContext(servlet.getServletContext());
		}

		// Get the URL to close action
		String url = viewManager.getActionURL(path);

		// the view manager was freed : unset module config and servlet context
		if (freed)
		{
			viewManager.setModuleConfig(null);
			viewManager.setServletContext(null);
		}

		// forward to close action
		return new ActionForward(url);
	} // end method getForwardToAction

	/**
	 * Gets a printWriter corresponding to the Accept-Encoding header passed by the browser.
	 *
	 * <ul>
	 *   <li>gzip compression if the browser support gzip</li>
	 *   <li>zip compression if the browsert support zip</li>
	 *   <li>standard if the browser doesn't support compression</li>
	 * </ul>
	 *
	 * <p>Set the corresponding header if necessary.</p>
	 *
	 * @param  request      the browser request
	 * @param  response     the response
	 * @param  environment  the application environment
	 * @param  directOutput indicates that we skip any debug stream and we want the real output.
	 * @return a print writer.
	 */
	protected static Writer getPrintWriter(HttpServletRequest request, HttpServletResponse response,
		LyEnvironment environment, boolean directOutput)
	{
		PrintWriter pw;
		String	    encodings = request.getHeader("Accept-Encoding");
		String		ajax = getParameter(request, LyWebConstants.AJAX);
		boolean     compress = !"true".equals(ajax) && environment.testResource("LY_WEB_USE_COMPRESSION");

		// Activate DEBUG MODE and dump HTML pages to DISK
		if ((LyWebConstants.WRITTEN_PAGES_COUNTER >= 0) && !directOutput)
			return (new StringWriter());

		try
		{
			if (compress && (encodings != null) && (encodings.indexOf("gzip") != -1))
			{
				// use gzip compression
				response.setHeader("Content-Encoding", "gzip");
				response.setHeader("Vary", "Accept-Encoding");

				GZIPOutputStream temp = new GZIPOutputStream(response.getOutputStream());

				pw = new PrintWriter(temp);
			}
			else if (compress && (encodings != null) && (encodings.indexOf("compress") != -1))
			{
				// use zip compression
				response.setHeader("Content-Encoding", "x-compress");
				response.setHeader("Vary", "Accept-Encoding");

				ZipOutputStream temp = new ZipOutputStream(response.getOutputStream());

				temp.putNextEntry(new ZipEntry("dummy name"));
				pw = new PrintWriter(temp);
			}
			else
				pw = response.getWriter();
		}
		catch (IOException e)
		{
			// if something go wrong, we return null
			// the caller has to take care of this case !!!
			environment.writeException(e);

			try
			{
				pw = response.getWriter();
			}
			catch (IOException e1)
			{
				environment.writeException(e1);
				pw = null;
			}
		}

		return pw;
	} // end method getPrintWriter

	/**
	 * Asks the given session to invoke the given leonardi action.<br>
	 * It uses the last view controller on the session to manage the context on the newly invoked
	 * action.
	 *
	 * @param  bean    Current session bean.
	 * @param  action  a Leonardi action.
	 * @return a controller.
	 */
	protected LyController invokeLeonardiAction(LyStrutsSessionBean bean, LyAction action)
	{
		return invokeLeonardiAction(bean, action, null, null);
	}
	
	/**
	 * Asks the given session to invoke the given leonardi action.<br>
	 * It uses the last view controller on the session to manage the context on the newly invoked
	 * action.
	 *
	 * @param  bean    Current session bean.
	 * @param  action  a Leonardi action.
	 * @param  classInfo a classinfo parameter for the action (default is action.getClassInfo() when set to null)
	 * @return a controller.
	 */
	protected LyController invokeLeonardiAction(LyStrutsSessionBean bean, LyAction action, LyClassInfo classInfo, LyObjectList context)
	{
		LySession		 session = bean.getSession();

		if (classInfo == null)
			classInfo = action.getClassInfo();
		
		int				 steps = 0;
		LyViewController viewController = session.getLastViewController(steps);

		// Check if controller was created from a static action (in this
		// case, controller
		// should not be displayed here)
		while ((viewController != null) && bean.isStaticActionController(viewController))
		{
			steps++;
			viewController = session.getLastViewController(steps);
		}

		LyAreaController parent = null;

		if (viewController != null)
		{
			parent = viewController.getAreaController();
			
			if ((parent instanceof LyDashboardController)
					&& (((LyDashboardController)parent).getDisplayedAreaController() != null))
				parent = ((LyDashboardController)parent).getDisplayedAreaController();
		}

		if (action.getMin() < 0)
		{
			// MIN < 0 : means any selection allowed, use content if nothing
			// selected
			if (context == null)
			{
				if (action.getMax() == 0)
					context = parent.getContent();
				else if (parent != null)
				{
					// take selection, even if any...
					context = parent.getSelection();
	
					// ... otherwise take content
					if ((context == null) || (context.getSize() == 0))
						context = parent.getContent();
				}
			}
		}
		else if (action.getMin() == 0)
		{
			// MIN = 0 : means any selection allowed, use empty if nothing
			// selected
			if (action.getMax() == 0)
			{
				// MIN = 0 and MAX = 0 action is independant from selection
				if (classInfo == null)
					classInfo = action.getClassInfo();

				if (classInfo == null)
					classInfo = parent.getClassInfo();
			}
			else if (context == null)
				// take selection, even if empty
				context = parent.getSelection();
		}
		else if ((parent != null) && (context == null))
			// take selection, even if empty
			context = parent.getSelection();

		if ((classInfo == null) && (parent != null))
			classInfo = parent.getClassInfo();

		// ask the session to process the action
		return session.invokeAction(parent, action, classInfo, context, false);
	} // end method invokeLeonardiAction

	/**
	 * Utility method that logs a request into the LY_OUTPUT_FILE when the DEBUG_TRACE debug level
	 * is on.
	 *
	 * @param request     An HTTP request.
	 * @param environment Current environment.
	 */
	protected void logRequest(HttpServletRequest request, LyEnvironment environment)
	{
		if (environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
		{
			// count the number of request
			LyWebRequestInfo	info = LyStrutsServlet.getWebRequestInfo();
			int					requ_count = (info == null ? LyWebConstants.REQU_COUNT : info.getRequestCount());
			StringBuffer		sb = new StringBuffer();
			Date				debugBeginDate = new Date();

			request.setAttribute("debugBeginDate", debugBeginDate);

			sb.append(
				"----------------------------------------------------------------------------------------------------\n");
			sb.append(">> Leonardi action ").append(this.getClass().getName()).append("\n");
			sb.append(">> Request n° ").append(requ_count).append("\n");
			sb.append(">> Request : ").append(request.getMethod());
			sb.append(" ").append(request.getContextPath());
			sb.append(request.getServletPath()).append(", ").append(debugBeginDate);

			Map<?, ?> parameters = request.getParameterMap();

			if (parameters != null)
			{
				Object       value;
				String[]     values;
				int		     i, n;
				Iterator<?> it = parameters.entrySet().iterator();

				while (it.hasNext())
				{
					Map.Entry<?,?>   entry = (Map.Entry<?,?>)it.next();
					
					sb.append("\n    - ").append(entry.getKey()).append(" = ");
					value = entry.getValue();
					if (value instanceof String[])
					{
						values = (String[])value;
						n = values.length;
						for (i = 0; i < n; i++)
						{
							sb.append(values[i]);
							if (i < (n - 1))
								sb.append(", ");
						}
					}
					else
						sb.append(value);
				}
			}

			environment.logTrace(sb.toString());
		}
	} // end method logRequest

	/**
	 * Utility method that logs a request into the LY_OUTPUT_FILE when the DEBUG_TRACE debug level
	 * is on.
	 *
	 * @param request     An HTTP request.
	 * @param environment Current environment.
	 */
	protected void logEndRequest(HttpServletRequest request, LyEnvironment environment)
	{
		if ((environment != null)
		    && environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
		{
			StringBuffer	sb = new StringBuffer();
			Date			debugBeginDate = (Date)request.getAttribute("debugBeginDate");

			if (debugBeginDate == null)
				debugBeginDate = new Date();

			sb.append(
				"----------------------------------------------------------------------------------------------------\n");
			sb.append(">> Ending Request for \n");
			sb.append(">> Leonardi action ").append(this.getClass().getName()).append("\n");
			sb.append(">> Request : ").append(request.getMethod());
			sb.append(" ").append(request.getContextPath());
			sb.append(request.getServletPath()).append(", duration : ");
			sb.append((System.currentTimeMillis() - debugBeginDate.getTime()));
			
			environment.logTrace(sb.toString());
		}
	} // end method logRequest

	
	/**
	 * Prepare the current page for viewing if needed.<br>
	 * Do nothing if the page writer is null or if the view is hidden.
	 *
	 * @param  pageWriter   Current page writer current environment
	 * @param  frame        current frame
	 * @param  targetWindow target window indicates that the result is to be shown into a popup
	 *                      frame. This parameter is the name of the popup frame, or null if in
	 *                      Leonardi main window
	 * @throws IOException
	 */
	protected void preparePage(LyWebPageWriter pageWriter, String frame, String targetWindow)
		throws IOException
	{
		if (pageWriter == null)
			return;

		try
		{
			pageWriter.preparePage(frame, targetWindow);
		}
		catch (RuntimeException e)
		{
			if (pageWriter.isFreed() || pageWriter.isFreeing())
				throw new WriteAbortedException(
					"pageWriter freed while preparing page; this causes following exception ", e);
			else
				throw e;
		}
	}

	/**
	 * This method stores the current window target in the current thread while handling an HTTP
	 * request.
	 *
	 * @param viewManager  The current view manager.
	 * @param targetWindow The name of the target window.
	 * @param frame        The name of the target frame
	 * @param source       The source component that sent the current HTTP request.
	 */
	protected void storeCurrentWindow(LyStrutsViewManager viewManager, String targetWindow,
		String frame, String source)
	{
		// Keep current window target name for session (controller stack is
		// managed there)
		if ((targetWindow == null)
				&& (frame != null)
				&& ((frame.equals(LyWebConstants.POPUP_VIEW))
					|| (frame.endsWith("_specific_frame"))))
			viewManager.setCurrentWindowName(frame); // Using an internal
		// popup view (iframe)
		else if (LyWebConstants.POPUP_VIEW.equals(source))
			// Popup view is being initialized or Reinitialized
			viewManager.setCurrentWindowName(LyWebConstants.POPUP_VIEW);
		else
			viewManager.setCurrentWindowName(targetWindow);
	}

	/**
	 * Writes a &lt;SCRIPT&gt; tag inside the given writer to add a piece of javascript code into an
	 * HTML page.
	 *
	 * @param  pw         Writer to the response.
	 * @param  javascript A piece of javascript code.
	 * @throws IOException
	 */
	protected void writeJavaScriptCode(Writer pw, String javascript)
		throws IOException
	{
		pw.write("<script type='text/javascript'>\n//<![CDATA[\n");
		pw.write(javascript);
		pw.write("\n//]]>\n</script>\n");
	}

	/**
	 * Writes a &lt;SCRIPT&gt; tag inside the given writer to access the javascript file.
	 *
	 * @param  environment    Current environment
	 * @param  pw             Current writer to write into
	 * @param  javascriptFile The name of a javascript file.
	 * @throws IOException
	 */
	protected void writeJavaScriptFile(LyEnvironment environment, Writer pw, String javascriptFile)
		throws IOException
	{
		if ((environment == null) || (environment.getViewManager() == null))
			return;
		
		String scriptRoot = ((LyWebViewManager)environment.getViewManager()).getScriptsDir();

		pw.write("<script src='");
		pw.write(scriptRoot);
		pw.write('/');
		pw.write(javascriptFile);
		pw.write("'></script>\n");
	}

	/**
	 * Writes message to specified output.
	 *
	 * @param  message     Message to write
	 * @param  environment current application environment
	 * @param  pw          current print writer for response.
	 * @throws IOException
	 */
	protected void writeMessage(String message, LyEnvironment environment, Writer pw)
		throws IOException
	{
		if (message == null)
			return;

		// add javascript links
		writeJavaScriptFile(environment, pw, LyWebConstants.JS_STRING);
		writeJavaScriptFile(environment, pw, LyWebConstants.JS_UTILS);
		writeJavaScriptFile(environment, pw, LyWebConstants.JS_ACTIONS);
		writeJavaScriptFile(environment, pw, LyWebConstants.JS_MESSAGE);

		// write the message
		pw.write(message);
	}

	/**
	 * Writes the current page writer in the given output.
	 *
	 * @param  request      Current HTTP request.
	 * @param  session      Current Leonardi user session.
	 * @param  environment  Current environment.
	 * @param  viewManager  Current view manager.
	 * @param  pageWriter   A page to be written.
	 * @param  pw           Current writer (writes into the output stream).
	 * @param  frame        current frame to write.
	 * @param  targetWindow The target window in the frameset.
	 * @throws IOException
	 */
	protected void writePage(HttpServletRequest request, LySession session,
		LyEnvironment environment, LyStrutsViewManager viewManager, LyWebPageWriter pageWriter,
		Writer pw, String frame, String targetWindow)
		throws IOException
	{
		try
		{
			if (environment.isDebugLevel(LyEnvironment.DEBUG_TRACE))
				environment.logTrace(">> writePage" + "\n\tpageWriter : " + pageWriter
					+ "\n\tframe : " + frame + "\n\ttargetWindow : "
					+ targetWindow);

			if (pageWriter != null)
				pageWriter.writeHtml(pw, frame);
		}
		catch (RuntimeException e)
		{
			if ((pageWriter != null) && (pageWriter.isFreed() || pageWriter.isFreeing()))
				throw new WriteAbortedException(
					"pageWriter freed while writing page; this causes following exception ", e);
			else
				throw e;
		}
	} // end method writePage

	/**
	 * Writes a page (depending on a url) at the top of the application (!= from the top window of
	 * the browser).
	 *
	 * @param  url         The page URL.
	 * @param  environment The application environment.
	 * @param  pw          The page writer.
	 * @throws IOException
	 */
	protected void writeUrlPage(String url, LyEnvironment environment, Writer pw)
		throws IOException
	{
		pw.write("<html>\n<head></head>\n<body>\n");

		// add javascripts
		writeJavaScriptFile(environment, pw, LyWebConstants.JS_STRING);
		writeJavaScriptFile(environment, pw, LyWebConstants.JS_UTILS);

		// write message if needed
		if ((environment != null) && (environment.getViewManager() != null))
		{
			LyWebViewManager viewManager = (LyWebViewManager)environment.getViewManager();
			String			 message = viewManager.getMessage();

			if (message != null)
			{
				writeMessage(message, environment, pw);
				viewManager.setMessage(null);
			}
		}

		// add javascript to load the page
		writeJavaScriptCode(pw, "loadUrlPage('" + url + "');");

		// close page
		pw.write("\n</body>\n</html>\n");
	}
	
	/**
	 * Check Session attributes in URL parameters in Direct URL navigation mode
	 */
	protected void checkSessionAttributes(LySession session, HttpServletRequest request)
	{
		String	navigationMode = session.getEnvironment().getResource("LY_WEB_DIRECT_URL_NAVIGATION");
		
		if ("session".equals(navigationMode)
				|| session.getEnvironment().testResource("LY_WEB_DIRECT_URL_NAVIGATION"))
		{
			// check if we have some parameters for the session
			Enumeration<?>   params = request.getParameterNames();
			String           param;
			
			while (params.hasMoreElements())
			{
				param = (String)params.nextElement();
				if (param.startsWith("$"))
					session.addValue(param, request.getParameter(param));
			}
		}
	}
} // end class LyStrutsLeonardiAction
