/*
 * LyStrutsLeonardiPlugin.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.view.web.struts.action;

import javax.servlet.*;
import javax.servlet.http.*;

import leon.app.*;
import leon.misc.*;
import leon.view.web.*;
import leon.view.web.struts.*;

import org.apache.struts.action.*;
import org.apache.struts.config.*;


/**
 * Type : leon.view.web.struts.LyStrutsLeonardiPlugin<br>
 * This class implements the <b><code>org.apache.struts.action.PlugIn</code></b> interface of the
 * Struts framework.<br>
 * This class manages the link between the Leonardi applications, Leonardi sessions and the WEB
 * Container.<br>
 * Once created, this plugin is stored inside the current ServletContext with the key <code>
 * <b>LEONARDI_PLUGIN</b></code> so that it may be retreived by any Leonardi Struts Action.<br>
 * To configure this plugin, add a plugin declaration inside the <code>struts-config.xml</code>
 * file:<br>
 * <code>&lt;plug-in className="leon.view.web.struts.LyStrutsLeonardiPlugin"&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&lt;set-property property="initName" value="$identfier of your Leonardi
 * application$"/&gt;<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&lt;set-property property="initFile" value="$intitialisation file name of
 * your Leonardi application$"/&gt;<br>
 * &lt;/plug-in&gt;</code><br>
 * This class is also a LyApplication and LySession factory : see the <code>
 * <b>createLySession</b></code> and <code><b>createLyApplication</b></code> methods.
 *
 * @version $Id: LyStrutsLeonardiPlugin.java 36511 2012-08-17 15:54:04Z jlsaouli $
 * @author  Lyria S.A.
 */
public class LyStrutsLeonardiPlugin
	implements PlugIn
{

	public static final String __VERSION =
		"$Id: LyStrutsLeonardiPlugin.java 36511 2012-08-17 15:54:04Z jlsaouli $";

	/** The key to store the Leonardi plugin inside the context. */
	public static final String LEONARDI_PLUGIN = "leon.view.web.struts.LyStrutsLeonardiPlugin";

	/** A lock for the HTTP SESSION. */
	public final String		   LOCK_HTTP_SESSION = "LOCK_HTTP_SESSION";
	
	/** Thread used to preload application. */
	protected Thread		   _preloadApplicationThread = null;


	/**
	 * Method that search for the Leonardi plugin inside the given context.
	 *
	 * @param  context A servlet context, i.e. the object that wraps the WEB application
	 *                 configuration.
	 * @return The current Leonardi Struts plugin.
	 * @throws IllegalStateException An illegal state exception if the plugin is not found.
	 */
	public static LyStrutsLeonardiPlugin getPlugin(ServletContext context)
		throws IllegalStateException
	{
		// search for the Leonardi plugin inside the servlet context
		LyStrutsLeonardiPlugin plugin = (LyStrutsLeonardiPlugin)context.getAttribute(
				LEONARDI_PLUGIN);

		if (plugin == null)
			throw new IllegalStateException(
				"BF AppEngine : Could not find the struts Leonardi plugin : "
				+ "check the struts-config.xml file contains a <plug-in "
				+ "className=\"leon.view.web.struts.LyStrutsLeonardiPlugin\"> declaration.");

		return plugin;
	}

	/**
	 * Factory method for creating a new LyApplication.<br>
	 * Retrieves the current plugin instance in the servlet context and create an application with
	 * the <code>initName</code> and <code>initFile</code> parameters.
	 *
	 * @param  context A servlet context, i.e. the object that wraps the WEB application
	 *                 configuration.
	 * @return a new LyApplication.
	 */
	public static LyApplication createApplication(ServletContext context)
	{
		// get the plugin
		LyStrutsLeonardiPlugin plugin = getPlugin(context);
		String				   applicationId = plugin.getInitName();
		String				   initFile = plugin.getInitFile();

		// Compute Base dir from request
		String				   path = context.getRealPath("");

		path = LyString.replace(path, "\\", "/");

		LyApplication application = new LyWebApplication(applicationId, initFile, path);

		if (application.getEnvironment() != null)
			application.getEnvironment().logTrace("Create a new application "
				+ application.getId());

		return application;
	}

	/**
	 * Factory method for creating a new LySession.<br>
	 * Retrieves the current plugin instance in the servlet context and create a session with the
	 * current application.
	 *
	 * @param  context A servlet context, i.e. the object that wraps the WEB application
	 *                 configuration.
	 * @return a new LySession.
	 */
	public static LySession createSession(ServletContext context)
	{
		return createSession(context, null);
	}

	/**
	 * Factory method for creating a new LySession.<br>
	 * Retrieves the current plugin instance in the servlet context and create a session with the
	 * current application.
	 *
	 * @param  context A servlet context, i.e. the object that wraps the WEB application
	 *                 configuration.
	 * @param  sessionCode hashcode of the session that must be retrieved
	 * @return a new LySession.
	 */
	public static LySession createSession(ServletContext context, String sessionCode)
	{
		// get the plugin
		LyStrutsLeonardiPlugin plugin = getPlugin(context);
		String				   applicationId = plugin.getInitName();
		LyApplication          application;

		synchronized (context)
		{
    		// retrieve the current LyApplication
    		application = LyApplication.getApplication(applicationId);
    
    		if (application == null)
    			// Start the Leonardi application now.
    			application = createApplication(context);
		}

		// Open a new user session
		// The session is Left UNREGISTRED --> Until registration in doGet
		// When the page structure is ready.
		LyEnvironment	environment = application.getEnvironment();
		
		if (environment == null)
		{
			// If environment is null, application has not been created correctly.
			// Retrieve application error message and register it for the exception handler.
			context.setAttribute("leonardi.exception", application.getError());
			return null;
		}

		// create the session
		LySession session;

		if (sessionCode != null)
		{
            session = (LySession)application.getSession(sessionCode);
            if (session != null)
                return session;
		}
		
		session = new LyStrutsSession(application, false, sessionCode);
		
		return session;
	}

	/** The name of the Leonardi application. */
	private String		   _initName;

	/** The name of the ".ini" file. */
	private String		   _initFile;
	
	/** Indicates whether application must be preloaded. */
	private String		   _isPreloadApplication;

	/** The Servlet context. */
	private ServletContext _context;

	/**
	 * Default constructor without parameters.<br>
	 * This constructor should be kept for JAVA reflection.
	 */
	public LyStrutsLeonardiPlugin()
	{
	}

	/**
	 * Implementation of the PlugIn interface, intialise the default Leonardi servlet for request
	 * processing.
	 *
	 * @param  servlet The current Struts Servlet.
	 * @param  config  The Configuration of the Struts module realted to this plugin.
	 * @throws ServletException
	 */
	public void init(ActionServlet servlet, ModuleConfig config)
		throws ServletException
	{
		// store the struts Servlet and the context
		_context = servlet.getServletContext();

		// store the plugin
		_context.setAttribute(LEONARDI_PLUGIN, this);
		
		if ("true".equalsIgnoreCase(_isPreloadApplication))
		{
			Runnable r = new Runnable()
			{
				public void run()
				{
					createApplication(getServletContext());
					_preloadApplicationThread = null;
				};
			};
			
			_preloadApplicationThread = new Thread(r);
			_preloadApplicationThread.start();
		}
	}

	/**
	 * Implementation of the PlugIn interface, calls for the destruction of all Leonardi application
	 * with the current id.
	 */
	public void destroy()
	{
		// close all applications with the current id.
		LyApplication application;
		String		  id = getInitName();
		int			  n = 1;

		while ((application = LyApplication.getApplication(id)) != null)
		{
			application.close();
			n++;
			id = application.getId() + n;
		}

		// remove the plugin
		_context.removeAttribute(LEONARDI_PLUGIN);
	}
	
	/**
	 * Indicates whether application is being loaded.
	 */
	public boolean isLoadingData()
	{
		return ((_preloadApplicationThread != null) && _preloadApplicationThread.isAlive());
	}

	/**
	 * Removes the context path from URLS to enable forwarding into a JSP.
	 *
	 * @param  request request which should contains the attrributes
	 * @param  page    with possibly a context path
	 * @return the page for the forward
	 */
	public String clearContextPath(HttpServletRequest request, String page)
	{
		if ((request == null) || (page == null))
			return page;

		String contextPath = request.getContextPath();

		// We do a forward, so this url may be relative to current servlet
		if (page.startsWith(contextPath))
			return page.substring(contextPath.length());
		else
			return page;
	}

	/**
	 * Gets the ".ini" file name.
	 *
	 * @return the name of the ".ini" file
	 * @see    #setInitFile(String)
	 */
	public String getInitFile()
	{
		return _initFile;
	}

	/**
	 * Sets the ".ini" file name.
	 *
	 * @param file the new name of the ".ini" file.
	 * @see   #getInitFile()
	 */
	public void setInitFile(String file)
	{
		_initFile = file;
	}

	/**
	 * Gets the name of the Leonardi application.
	 *
	 * @return Leonardi application's name.
	 * @see    #setInitName(String)
	 */
	public String getInitName()
	{
		return _initName;
	}

	/**
	 * Sets the name of the Leonardi application.
	 *
	 * @param name the new Leonardi application's name.
	 * @see   #getInitName()
	 */
	public void setInitName(String name)
	{
		_initName = name;
	}

	/**
	 * Gets the value of the isPreloadApplication parameter.
	 *
	 * @return the value of the isPreloadApplication parameter.
	 * @see    #setIsPreloadApplication(String)
	 */
	public String isPreloadApplication()
	{
		return _isPreloadApplication;
	}

	/**
	 * Sets the name value of the isPreloadApplication parameter.
	 *
	 * @param isPreloadApplication the new isPreloadApplication parameter's value.
	 * @see   #isPreloadApplication()
	 */
	public void setIsPreloadApplication(String isPreloadApplication)
	{
		_isPreloadApplication = isPreloadApplication;
	}

	/**
	 * Method overriden to return the context of the Struts servlet.
	 *
	 * @return the servlet context.
	 */
	public ServletContext getServletContext()
	{
		return _context;
	}
} // end class LyStrutsLeonardiPlugin
