/*
 * LyStrutsRefreshAction.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.view.web.struts.action;

import java.io.*;
import java.util.*;

import javax.servlet.http.*;

import leon.app.*;
import leon.misc.*;
import leon.view.web.*;
import leon.view.web.struts.*;

import org.apache.struts.action.*;


/**
 * Type : leon.view.web.struts.action.LyStrutsPageAction<br>
 * This action generates push mechanism for graphical updates.<br>
 *
 * @version $Id: LyStrutsPushAction.java 41635 2012-11-28 13:24:43Z bangiole $
 * @author  Lyria S.A.
 */
public class LyStrutsPushAction extends LyStrutsLeonardiAction
{
	public static final String __VERSION =
		"$Id: LyStrutsPushAction.java 41635 2012-11-28 13:24:43Z bangiole $";

	/** Time to sleep for push thread (in milliseconds). */
	public static int		PUSH_CHECK_DELAY = 500;

	/** Max delay for push response (in milliseconds). */
	public static int		PUSH_MAX_DELAY = 60000;

	/** Number of iterations for push response. */
	public static int		PUSH_ITERATIONS = PUSH_MAX_DELAY / PUSH_CHECK_DELAY;

	/** Number of iterations in error before a notification is shown to the end-user. */
	public static int		PUSH_ERROR_ITERATIONS = 3;

	/** Last time when a push refresh came from the browser. */
	protected Hashtable<HttpSession, Long>		_lastPushTimes;


	/**
	 * Default constructor without parameters.<br>
	 * This constructor should be kept for Java reflection.
	 */
	public LyStrutsPushAction()
	{
		_lastPushTimes = new Hashtable<HttpSession, Long>();
	}

	/**
	 * Generic treatment for the push mechanism of the Leoanrdi User Interface.
	 *
	 * @param  mappings - The ActionMapping used to select this instance
	 * @param  bean     - The optional ActionForm bean for this request (if any)
	 * @param  request  - The HTTP request we are processing
	 * @param  response - The HTTP response we are creating
	 * @return null : an empty action forward while the action has written the response.
	 */
	public ActionForward execute(ActionMapping mappings, ActionForm bean,
		HttpServletRequest request, HttpServletResponse response)
		throws Exception
	{
		// Get the session from the session bean
		LyStrutsSessionBean sessionBean = (LyStrutsSessionBean)bean;
		LySession		    mainSession = sessionBean.getSession();
		LyStrutsViewManager mainViewManager = sessionBean.getViewManager();

		// Check if session is registered, otherwise redirect to home page
		if (!checkSessionRegistered(sessionBean, request, response))
			return null;

		String	      timeParam = getParameter(request, "_time");
		long          time = LyTools.longFromString(timeParam);
		Long          lastPushTime;
		Writer	      pw = beginRequest(request, response, mainViewManager);
		int           loop = 0;
		boolean	      stop = false;
		boolean	      recurse = true;
		HttpSession   session = request.getSession();

		lastPushTime = _lastPushTimes.get(session);

		if ((lastPushTime == null)
		        || (time > lastPushTime.longValue()))
			_lastPushTimes.put(session, time);

		while ((loop < PUSH_ITERATIONS)
				&& !stop
				&& (time >= (_lastPushTimes.get(session)).longValue()))
		{
			Thread.sleep(PUSH_CHECK_DELAY);

			if ((mainSession == null)
					|| (mainSession.getState() == LySession.State.CLOSING)
					|| (mainSession.getState() == LySession.State.CLOSED))
			{
				stop = true;
				recurse = false;
				break;
			}

			try
			{
				mainSession.getSoftLock();
    			// Generate update scripts for main sessions if needed
    			if (needsUpdate(mainViewManager) && ! sessionBean.isSessionInProgress())
    			{
    				generateAdditionalScripts(mainSession, request, response, pw);
    				stop = true;
    			}

    			// Retrieve slave-sessions for current session (multi-sessions mode)
    			Vector<LySession>	slaveSessions = mainSession.getSlaveSessions();
    			int					i, n = (slaveSessions == null ? 0 : slaveSessions.size());

    			// Generate update scripts for slave sessions if needed
    			for (i = 0; (i < n) && !stop; i++)
    			{
    				LySession	slaveSession = slaveSessions.elementAt(i);

    				if (slaveSession == null)
    					continue;

    				LyStrutsViewManager	slaveViewManager = (LyStrutsViewManager)slaveSession.getViewManager();

    				if (needsUpdate(slaveViewManager) && !sessionBean.isSessionInProgress())
    				{
    					try
    					{
    						slaveSession.getSoftLock();
    						String	frameName = slaveSession.getId();

    						frameName = LyString.replace(frameName, ".", "");
    						pw.append("<LEONARDI_WINDOW " + LyWebConstants.TARGET_PARAM + frameName + ">");
    						generateAdditionalScripts(slaveSession, request, response, pw);
    						pw.append("</LEONARDI_WINDOW>");
    						stop = true;
    					}
    					finally
    					{
    						slaveSession.releaseSoftLock();
    					}
    				}
    			}
	         }
            catch (Throwable e)
            {
                // An exception occurred but it will logged on next refresh, trace only.
                LyEnvironment    environment = mainSession.getEnvironment();

                environment.logTrace("Leonardi : refresh error detected on push.");
                if (environment.isDebugLevel(LyEnvironment.DEBUG_ALL))
			    	environment.writeException(e);
            }
			finally
			{
				mainSession.releaseSoftLock();
			}

			loop++;
		}

		if (recurse && (time >= (_lastPushTimes.get(session)).longValue()))
			pw.write("<script type='text/javascript'>PushRefresh();</script>\n");

		addAdditionalScripts(pw);

		// end request
		endRequest(request, response, mainViewManager, pw);

		return null;
	} // end method execute

	/**
	 * Checks if an update must be sent to the browser.
	 *
	 * @param viewManager current view manager
	 * @return true if an update must be sent, false otherwise
	 */
	protected boolean needsUpdate(LyStrutsViewManager viewManager)
	{
		return viewManager.needsUpdate();
	}

	/**
	 * Add Javascript code that must be sent to the push answer.
	 *
	 * @param pw
	 * @throws Exception
	 */
	protected void addAdditionalScripts(Writer pw) throws Exception
	{
		// Default implementation does nothing
	}
} // end class LyStrutsPageAction
