/*
 * LyStrutsSessionBean.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */
package leon.view.web.struts.action;

import java.net.*;
import java.util.*;

import javax.servlet.*;
import javax.servlet.http.*;

import leon.app.*;
import leon.control.*;
import leon.misc.*;
import leon.view.web.*;
import leon.view.web.struts.*;

import org.apache.struts.action.*;
import org.apache.struts.config.*;
import org.apache.struts.util.*;


/**
 * Type : leon.view.web.struts.LyStrutsSessionBean<br>
 * This bean should be declared as a global bean in the <code>struts-config.xml</code> file with the
 * <code>&lt;form-beans&gt;</code>tag.<br>
 * It should also be used in the <b>session</b> scope of all Leonardi action mappings declarations.
 * <br>
 * <br>
 * This bean manages the <code>LySession</code> birth and death and makes the link with the <code>
 * HttpSession</code> provided it has been properly declared.
 *
 * @version $Id: LyStrutsSessionBean.java 34889 2012-03-21 10:33:38Z jlsaouli $
 * @author  Lyria S.A.
 */
public class LyStrutsSessionBean extends ActionForm
{
	/**
	 * 
	 */
	private static final long serialVersionUID = -7162011611755625150L;

	public static final String  __VERSION =
		"$Id: LyStrutsSessionBean.java 34889 2012-03-21 10:33:38Z jlsaouli $";

	/** Identifier of the default session. */
	public final static String	DEFAULT_SESSION = "DEFAULT_SESSION";
		
	/** Number of current user sessions. */
	protected static Integer    NB_SESSION = new Integer(0);
	
	/** Lock */
	protected static Object		LOCK = new Object();

	/** Keep the information between sessions and beans. */
	public static HashMap<LySession, LyStrutsSessionBean>	   _sessionBeans =
		new HashMap<LySession, LyStrutsSessionBean>();

	/** The associated LySession instance of the HttpSession. */
	private transient LySession	_currentSession;
	
	/** The associated LySession instance of the HttpSession. */
	private transient boolean	_sessionInProgress;

	/** The list of associated LySession instances of the HttpSession. */
	private transient Map<String, LySession>	_sessions;

	/**
	 * The association between the static actions and their view controllers for the current user
	 * session.
	 */
	private transient Map<String, String>	    _staticActions;
	
	/** Last shown controller for this session (used only in LY_REPLAY mode). */
	private LyController	_lastShownController;
	
	/** Indicates if current session is in replay mode. */
	private boolean			_replayMode;
	
	/**
     * Swap two sessions.
     * 
     * @param session the new session
     * @param oldSession the old session
     * @param applicationId the application id
     */
	public static void swapSession(Object session, Object oldSession, String applicationId)
	{
		LyStrutsSessionBean binding = _sessionBeans.get(session);
		LyStrutsSessionBean oldBinding = _sessionBeans.get(oldSession);

		binding.setSession((LySession)oldSession);
		if (oldBinding != null)
			oldBinding.setSession(null);
		binding._sessions.put(DEFAULT_SESSION, (LySession)oldSession);
		if (oldBinding != null)
			oldBinding._sessions.remove(DEFAULT_SESSION);

		((LyStrutsSession)oldSession).setSessionBean(binding, ((LyStrutsSession)session).getHttpSession());
		((LyStrutsSession)session).setSessionBean(null, null);
		((LySession)session).close();
		_sessionBeans.remove(session);
	}
	
	/**
	 * Empty constructor : used for Java reflection.
	 */
	public LyStrutsSessionBean()
	{
		init();
	}

	/**
	 * Initializes session bean.
	 */
	protected void init()
	{
	    _staticActions = Collections.synchronizedMap(new HashMap<String, String>(8));
	    _sessions = Collections.synchronizedMap(new HashMap<String, LySession>(1));
	    _replayMode = false;
	}
	
	/**
	 * Stores the current session in the Leonardi application.<br>
	 * Adds a lock to the Leonardi application.
	 */
	public synchronized void beginRequest()
	{
		LySession	currentSession = getSession();
		
		if (currentSession == null)
			return;
		_sessionInProgress = true;
		LyWebApplication appli = getApplication();

		appli.setCurrentSession(currentSession);
		appli.lock();
	}

	/**
	 * Removes the current session from the Leonardi application.<br>
	 * Removes a lock from the Leonardi application.
	 */
	public synchronized void endRequest()
	{
		LySession	currentSession = getSession();
		
		if (currentSession == null)
			return;

		LyWebApplication appli = getApplication();

		appli.setCurrentSession(null);
		appli.unlock();
		_sessionInProgress = false;
	}

	public synchronized boolean isSessionInProgress()
	{
		LySession	currentSession = getSession();
		
		if (currentSession == null)
			return false;
		
		return _sessionInProgress;
	}
	/**
	 * Gets the application of the current session.<br>
	 * It wakes it up if necessary.
	 *
	 * @return the application of the current session.
	 */
	public LyWebApplication getApplication()
	{
		LySession	currentSession = getSession();
		
		if (currentSession == null)
			return null;
		
		LyApplication application = currentSession.getApplication();

		if (application.isSleeping())
		{
			LyEnvironment environment = getEnvironment();

			environment.logTrace("Waking up application " + application.getId());
			application.wakeUp();
		}

		return (LyWebApplication)application;
	}

	/**
	 * Gets the environment of the current session.
	 *
	 * @return the environment of the current session.
	 */
	public LyEnvironment getEnvironment()
	{
		LySession	currentSession = getSession();
		
		if (currentSession == null)
			return null;
		
		return currentSession.getEnvironment();
	}

	/**
	 * Gets the associated LySession.
	 *
	 * @return the current Leonardi Session
	 * @see    #setSession(LySession)
	 */
	public LySession getSession()
	{
		LyWebRequestInfo	info = LyStrutsServlet.getWebRequestInfo();
		HttpServletRequest	request = (info != null ? info.getRequest() : null);
		LySession			currentRequestSession = (request != null ? (LySession)request.getAttribute("_currentSession") : null);
		
		if (currentRequestSession != null)
			return currentRequestSession;
		
		return _currentSession;
	}

	/**
	 * Gets the id of static action previously stored in this bean.
	 *
	 * @param  actionId Identifier of a Leonardi action.
	 * @return The identifier of the related view controller or null if not found
	 * @see    #setStaticAction(String, String)
	 */
	public String getStaticAction(String actionId)
	{
		return _staticActions.get(actionId);
	}

	/**
	 * Gets the view manager related to the current session.
	 *
	 * @return the view manager related to the current session.
	 */
	public LyStrutsViewManager getViewManager()
	{
		LyEnvironment env = getEnvironment();

		if (env == null)
			return null;
		
		return (LyStrutsViewManager)env.getViewManager();
	}

	/**
	 * Set the last shown controller (used only in LY_REPLAY mode).
	 */
	public void setLastShownController(LyController controller)
	{
		_lastShownController = controller;
	}
	
	/**
	 * Returns the last shown controller (used only in LY_REPLAY mode).
	 */
	public LyController getLastShownController()
	{
		return _lastShownController;
	}
	
	/**
	 * Indicates if current session is in replay mode.
	 */
	public boolean isReplayMode()
	{
		return _replayMode;
	}

	/**
	 * This method is called by the Struts request processor each time this bean is involed in a
	 * request to reset its properties.<br>
	 * Specific Leonardi treatments are done :<br>
	 *
	 * <ul>
	 *   <li>Initiliaze the root URL on the view manager if needed.</li>
	 *   <li>Initiliaze the <code>LyEnvironment</code> Locale with the struts Locale if needed.</li>
	 *   <li>Store The LY_EXIT_PAGE resource in the HTTP session</li>
	 * </ul>
	 *
	 * @param mapping The mapping used to select this instance
	 * @param request The servlet request we are processing
	 */
	public void reset(ActionMapping mapping, HttpServletRequest request)
	{
	    if (_sessions == null)
	        init();
	    
		// Call the checkSession() method to ensure a valid Leonardi session exists.
		// In a normal case, this method should have been called from the valueBound() method but in
		// some case (restart of web server for example), the http session may be still alive and
		// valueBound() method has not be called. So we create the Leonardi session now. 
		checkSession(request);
		
		// The session should already have been set
		LySession     session = getSession();

	    if (session == null)
	        return;
	      
		// initialize the root URL if needed
		LyEnvironment environment = session.getEnvironment();

		if (environment == null)
		    return;
		
		LyStrutsViewManager viewManager = (LyStrutsViewManager)environment.getViewManager();

		// the view manager may have been freed
		if ((viewManager != null)
				&& (viewManager.getEnvironment() != null)
				&& (viewManager.getRootURL() == null))
		{
			// set the root url
			String rootUrl = computeRootUrl(request, environment);

			viewManager.setRootURL(rootUrl);
			viewManager.setContextUrl(request.getContextPath());
			
			// set the module config instance
			ModuleConfig config = mapping.getModuleConfig();

			viewManager.setModuleConfig(config);

			// set the servlet context
			HttpSession    httpSession = request.getSession();
			ServletContext context = httpSession.getServletContext();

			viewManager.setServletContext(context);
			
			String	target = request.getParameter(LyWebConstants.TARGET_PARAM);
			
			if (session.getMasterSession() != null
					|| ((target != null) && !DEFAULT_SESSION.equals(target)))
				viewManager.setTargetCode(session.getId() + "");

			// store the url of exit page in the session while
			// the freeing of the environment may cause the lost of this
			// resource
			httpSession.setAttribute("LY_EXIT_PAGE", environment.getUrl("LY_EXIT_PAGE"));
			httpSession.setAttribute(LyWebConstants.RA_LY_ROOT_URL, rootUrl);
			
		    _replayMode = environment.testEnv("LY_REPLAY");
		    
		    if (_replayMode)
		    {
				String	cssPath = environment.getUrl("LY_CSS_FILE");
				String	logo = environment.getImageUrl("LY_COMPANY_LOGO");

				httpSession.setAttribute("language", environment.getLanguage());
				httpSession.setAttribute("cssPath", cssPath);
				httpSession.setAttribute("logo", logo);
		    }
		}
	} // end method reset

	/**
	 * Sets the associated LySession.
	 *
	 * @param session The leonardi session.
	 * @see   #getSession()
	 */
	public void setSession(LySession session)
	{
		_currentSession = session;
		
		if (session != null)
			_sessionBeans.put(session, this);

		setRequestSessionAttribute(session);
	}
	
	/**
	 * Set the _currentSession attribute for current http request.
	 */
	protected void setRequestSessionAttribute(LySession session)
	{
		LyWebRequestInfo	info = LyStrutsServlet.getWebRequestInfo();
		HttpServletRequest	request = (info != null ? info.getRequest() : null);
		
		if (request != null)
			request.setAttribute("_currentSession", session);
	}

	/**
	 * Close the current session and close http session if needed.
	 * 
	 * @param request current http request
	 */
	public void closeCurrentSession(HttpServletRequest request)
	{
		closeCurrentSession(request, null);
	}

	/**
	 * Close the current session and close http session if needed.
	 * 
	 * @param request current http request
	 * @param session current http session
	 */
	public void closeCurrentSession(HttpServletRequest request, HttpSession session)
	{
		closeCurrentSession(request, session, true);
	}
	
	/**
	 * Close the current session and close http session if needed.
	 * 
	 * @param request current http request
	 * @param session current http session
	 * @param closeHttpSession indicates whether http session must be closed
	 */
	public void closeCurrentSession(HttpServletRequest request, HttpSession session, boolean closeHttpSession)
	{
		LySession	currentSession = getSession();
		
		if (currentSession != null)
		{
			boolean			lastSession = (_sessions.size() == 1);
			String			target = (request == null ? null : (String)request.getAttribute(LyWebConstants.TARGET_PARAM));
			
			if (request == null)
			{
				// Try to retrieve target parameter from _sessions table
				Set<Map.Entry<String, LySession>>		entries = _sessions.entrySet();
				
				synchronized(_sessions)
				{
				Iterator<Map.Entry<String, LySession>>	iterator = entries.iterator();
				
				while (iterator.hasNext())
				{
					Map.Entry<String, LySession>	entry = iterator.next();
					LySession						lysession = entry.getValue();
	
					if (lysession == _currentSession)
						target = entry.getKey();
				}
			}
			}
			
			if ((target == null) || "null".equals(target))
				target = DEFAULT_SESSION;

			if ((lastSession || (target == DEFAULT_SESSION)) && closeHttpSession)
			{
				// After the exit, we must invalidate the HttpSession.
				HttpSession		httpSession;
				
				if (session != null)
					httpSession = session;
				else
					httpSession = request.getSession(false);

				if (httpSession != null)
				{
					try
					{
						httpSession.invalidate();
					}
					catch (IllegalStateException e)
					{
					}
				}
			}
			else
			{
				// Remove the binding info associated to the session
				_sessionBeans.remove(currentSession);
				_sessions.remove(target);
				
				if (target == DEFAULT_SESSION)
				{
					Set<Map.Entry<String, LySession>>		entries = _sessions.entrySet();

					synchronized(_sessions)
					{
					Iterator<Map.Entry<String, LySession>>	iterator = entries.iterator();
					
					while (iterator.hasNext())
					{
						Map.Entry<String, LySession>	entry = iterator.next();
						LySession						lysession = entry.getValue();
		
						// Remove the binding info associated to the session
						_sessionBeans.remove(lysession);
					}
					}
					
					_sessions.clear();
				}
			}
			
			_currentSession = null;
			setRequestSessionAttribute(null);
		}
	}

	/**
	 * Stores a static action.
	 *
	 * @param actionId     The identifier of a Leonardi action.
	 * @param controllerId The identifier of the related view controller.
	 * @see   #getStaticAction(String)
	 */
	public void setStaticAction(String actionId, String controllerId)
	{
		_staticActions.put(actionId, controllerId);
	}

	/**
	 * Indicates if the given controller manages a static action.
	 *
	 * @param  controller instance of LyController to be checked
	 * @return true if the controller manages a static action, false otherwise
	 */
	protected boolean isStaticActionController(LyController controller)
	{
		if ((controller == null) || (_staticActions == null))
			return false;

		return (_staticActions.containsValue(controller.getId()));
	}

	/**
	 * Checks the existence of a Leonardi session.<br />
	 * If no session exists, a new Leonardi session is created.
	 * 
	 * @param request current http request
	 */
	public void checkSession(HttpServletRequest request)
	{
		String		target = request.getParameter(LyWebConstants.TARGET_PARAM);
		
		if ((target == null) || "null".equals(target))
			target = DEFAULT_SESSION;
		
		LySession	currentSession = _sessions.get(target);
		
		if (currentSession != null)
		{
			_currentSession = currentSession;
			setRequestSessionAttribute(currentSession);
			return;
		}

		try
		{
			HttpSession	   httpSession = request.getSession();
			ServletContext context = httpSession.getServletContext();

			// Create a new LySession
			currentSession = LyStrutsLeonardiPlugin.createSession(context, target);
			
			if (currentSession == null)
			{
				// If an error has been set on the context, set it on the request.
				String	error = (String)context.getAttribute("leonardi.exception");
				
				request.setAttribute("leonardi.exception", error);
				return;
			}
			
			((LyStrutsSession)currentSession).setSessionBean(this, httpSession);
			setSession(currentSession);
			_sessions.put(target, currentSession);

			// initialize session timeout
			int timer = currentSession.getSessionTimeout();

			if (timer > 0)
				httpSession.setMaxInactiveInterval(timer);
			
			// Create a session listener for valueUnbound()
			LyApplication  application = getApplication();
			
			if (httpSession.getAttribute(application.getId()) == null)
				httpSession.setAttribute(application.getId(), new _LyStrutsSessionListener());
			
			// get the environment
			LyEnvironment environment = currentSession.getEnvironment();

			// log session count
			synchronized (LOCK)
			{
				NB_SESSION = new Integer(NB_SESSION.intValue() + 1);

				if (environment != null)
				{
					environment.logInfo("LyStrutsSessionBean.valueBound " + currentSession + " at "
						+ new Date() + " : " + httpSession + " (" + NB_SESSION + " sessions).");

					if (environment.isDebugLevel(LyEnvironment.DEBUG_PERF))
					{
						System.gc();
						environment.logOut("\n\tCurrent number of LySession = "
							+ LySession._nbSessions + "\n\tCurrent number of LyObject = "
							+ leon.data.LyObject._nbObjects
							+ "\n\tCurrent number of LyController = "
							+ leon.control.LyController._nbControllers
							+ "\n\tCurrent number of LyView = " + leon.view.LyView._nbViews
							+ "\n\tCurrent number of LyComponent = "
							+ leon.view.LyComponent._nbComponents);
					}
				}
			}

			if (LyWebConstants.WRITTEN_PAGES_COUNTER == -2)
			{
				if (environment.isDebugLevel(LyEnvironment.DEBUG_DUMP_HTML))
					LyWebConstants.WRITTEN_PAGES_COUNTER = 0;
				else
					LyWebConstants.WRITTEN_PAGES_COUNTER = -1;
			}
		}
		catch (Exception e)
		{
			// Prevent any crash
			if ((currentSession != null) && (currentSession.getEnvironment() != null))
				currentSession.getEnvironment().writeException(e);
			else
				e.printStackTrace();
		}
	}

	/**
	 * This method computes the root URL of the current leonardi application.<br>
	 * This is the path to the current server + the path of the current WEB application.<br>
	 * Change also the scheme from http to https if LY_USE_HTTPS is on.
	 *
	 * @param  request     The current HTTP request.
	 * @param  environment The current environment
	 * @return the root URL of the current leonardi application.
	 */
	protected String computeRootUrl(HttpServletRequest request, LyEnvironment environment)
	{
		String rootUrl = environment.getEnv("LY_ROOT_URL");

		// get the root url : the root servlet context
		try
		{
			if (LyStrutsViewManager.USE_ABSOLUTE_ROOT_URL)
			{
				URL serverUrl = RequestUtils.serverURL(request);
				URL url = new URL(serverUrl, request.getContextPath());
	
				rootUrl = url.toString();
			}
			else
				rootUrl = request.getContextPath();
		}
		catch (MalformedURLException e)
		{
			// note: if request.getRequestURL() (Servlet 2.3) does not
			// work, try HttpUtils.getRequestURL(request) (Servlet 2.2) instead
			rootUrl = request.getRequestURL().toString();
			int index = rootUrl.lastIndexOf('/');

			if (index > 0)
				rootUrl = rootUrl.substring(0, index);
		}

		// change the scheme according to SSL layer
		String httpScheme;

		if (environment.testEnv("LY_USE_HTTPS"))
			httpScheme = "https:";
		else
			httpScheme = request.getScheme() + ":";

		rootUrl = LyString.replace(rootUrl, "http:", httpScheme);
		rootUrl = LyString.replace(rootUrl, "https:", httpScheme);

		return rootUrl;
	} // end method computeRootUrl

	/**
	 * Tests if a session is closed.
	 *
	 * @param  session the session to test.
	 * @return true if the specified session is closed.
	 */
	protected boolean sessionIsClosed(LySession session)
	{
		return ((session == null)
				|| (session.getState() == LySession.State.CLOSED)
				|| (session.getState() == LySession.State.CLOSING));
	}
	
	/**
	 * Internal class used to listen to session bounds / unbounds.
	 */
	class _LyStrutsSessionListener implements HttpSessionBindingListener
	{	
		/**
		 * Implementation of HttpSessionBindingListener interface.<br>
		 * Creates the LySession and initialize the timeout.
		 *
		 * @see HttpSessionBindingListener#valueBound(javax.servlet.http.HttpSessionBindingEvent)
		 */
		public void valueBound(HttpSessionBindingEvent event)
		{
		}

		/**
		 * Implementation of HttpSessionBindingListener interface.<br>
		 * Destroys the related LySession.
		 *
		 * @see HttpSessionBindingListener#valueUnbound(javax.servlet.http.HttpSessionBindingEvent)
		 */
		public void valueUnbound(HttpSessionBindingEvent event)
		{
			// get the HttpSession
			HttpSession httpSession = event.getSession();

			// get the current application
			LyApplication application = getApplication();

			try
			{
				Set<Map.Entry<String, LySession>>		entries = _sessions.entrySet();

				synchronized(_sessions)
				{
				Iterator<Map.Entry<String, LySession>>	iterator = entries.iterator();
				
				while (iterator.hasNext())
				{
					Map.Entry<String, LySession>	entry = iterator.next();
					LySession						session = entry.getValue();
					LyEnvironment					environment = (session == null ? null : session.getEnvironment());
	
					if ((environment == null) && (application != null))
						environment = application.getEnvironment();
	
					// Remove the binding info associated to the session
					_sessionBeans.remove(session);
					
					if (session == _currentSession)
					{
					    _currentSession = null;
						setRequestSessionAttribute(null);
					}
	
					// log session count
					synchronized (LOCK)
					{
						NB_SESSION = new Integer(NB_SESSION.intValue() - 1);
	
						if (environment != null)
						{
							environment.logInfo("LyStrutsSessionBean.valueUnbound " + session + " at " + new Date()
								+ " : " + httpSession + " (" + NB_SESSION + " sessions).");
	
							if (environment.isDebugLevel(LyEnvironment.DEBUG_PERF))
							{
								System.gc();
								environment.logOut("\n\tCurrent number of LySession = "
									+ LySession._nbSessions + "\n\tCurrent number of LyObject = "
									+ leon.data.LyObject._nbObjects
									+ "\n\tCurrent number of LyController = "
									+ leon.control.LyController._nbControllers
									+ "\n\tCurrent number of LyView = " + leon.view.LyView._nbViews
									+ "\n\tCurrent number of LyComponent = "
									+ leon.view.LyComponent._nbComponents);
							}
						}
					}
					
					// Closing the user session.
					if ((session != null) && (!sessionIsClosed(session)))
						session.sessionTimeout(false);
				}
				}
				
				_sessions.clear();

				try
				{
					String	appId = (application == null ? null : application.getId());

					if (appId != null)
						httpSession.setAttribute(appId, null);
				}
				catch (IllegalStateException e)
				{
					// With some servers, this produces an exception "Session
					// already invalidated"
				}
			}
			catch (Exception e)
			{
				// Prevent any crash
				if ((application != null) && (application.getEnvironment() != null))
					application.getEnvironment().writeException(e);
				else
					e.printStackTrace();
			}
		}
	}
} // end class LyStrutsSessionBean
