/*
 * LySwingArea.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import net.miginfocom.swing.*;
import java.awt.*;
import java.util.*;
import javax.swing.*;
import javax.swing.border.*;

import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Area object (as described in the dtd) for swing viewer.
 *
 * <p>An area allows the positioning of various graphical components inside a main window or a
 * dialog window. It can contains any type of panel as well as a title bar, a tool bar and a command
 * bar.</p>
 *
 * @version $Id: LySwingArea.java 34996 2012-03-23 09:40:11Z jlsaouli $
 * @author  Lyria
 * @see     LyArea
 */
public class LySwingArea extends LySwingContainer
	implements LySwingComponent,
		LyGraphicalArea,
		LySetWaitingInterface,
		LySetBorderPanelInterface,
        LySetVisibleInterface
{
	public static final String  __VERSION =
		"$Id: LySwingArea.java 34996 2012-03-23 09:40:11Z jlsaouli $";

	/**
	 * A toolbar with an internal panel. This internal panel may use the <code>_scrollingArea</code>
	 * if it is scrollable.
	 */
	protected Container		    _subRootPanel;

	/** The titlebar. */
	protected Container		    _toolTitlePanel;

	/** The south panel. */
	protected Container		    _southPanel;

	/** The view independent area description. */
	protected LyArea		    _area;

	/** The root panel. */
	protected Container		    _rootPanel;

	/** A scrolling area. */
	protected Container		    _scrollingArea;

	/** An internal panel. */
	protected Container		    _internalPanel;

	/** Cached toolbar, used in the {@link #setWaiting(boolean)} method. */
	protected Component		    _titleBar;

	/** Cached toolbar, used in the {@link #setWaiting(boolean)} method. */
	protected Component		    _selectBar;

	/** Cached toolbar, used in the {@link #setWaiting(boolean)} method. */
	protected Component		    _commandBar;

	/** Cached toolbar, used in the {@link #setWaiting(boolean)} method. */
	protected Vector<Component> _toolBars;

	/** The border panels. */
	protected LySwingPanel[]    _borderPanels;

	/**
	 * Creates the action demand object for the specified action event. This action demand will be
	 * the link between the View and Controller layers.
	 *
	 * @param  actionEvent the action event
	 * @return the action demand associated with the actionEvent
	 */
	public static LyDemand buildActionDemand(LyActionEvent actionEvent)
	{
		LyActionDemand demand;

		demand = new LyActionDemand(actionEvent.getType(), actionEvent.getSourceId());

		// attempt to get the value
		demand.setValue(actionEvent.getValue());
		demand.setAdditionalValue(actionEvent.getAdditionalValue());
		demand.setSubIndex(actionEvent.getSubIndex());
		// check index
		int i, idx = 0;

		for (i = 0; idx >= 0; i++)
		{
			idx = actionEvent.getIndex(i);
			if (idx >= 0)
				demand.addIndex(idx);
		}

		return (demand);
	}

	/**
	 * Class constructor of an Area object for the Swing target.
	 *
	 * @param area an AREA defined for every type of view targets. This area is initialized
	 *             specificaly for the Swing view.
	 */
	public LySwingArea(LyArea area)
	{
		super(area);
		_area = area;

		initialize();
		setContent();

		if (_area.isClosable() && (_internalPanel instanceof LyJBorderPanel))
		{
			Image buttonClose = getImage("LY_AREA_BUTTON_CLOSE");
			Image buttonOpen = getImage("LY_AREA_BUTTON_OPEN");

			((LyJBorderPanel)_internalPanel).setClosable(true);
			((LyJBorderPanel)_internalPanel).setOpen(_area.isOpen());
			((LyJBorderPanel)_internalPanel).setOpenCloseImages(buttonOpen, buttonClose);
			((LyJBorderPanel)_internalPanel).setParentView(_area.getParentView());
		}
	}

	/**
	 * Initializes the Area object and prepares all its panels.
	 */
	public void initialize()
	{
		// Prepare the internal panel
		_internalPanel = createInternalPanel();

		// Prepare a scrolling area as needed
		_scrollingArea = createScrollingArea();

		// Prepare the root panel
		_rootPanel = createRootPanel();

		Color color = getColor("LY_AREA_BACKGROUND");

		((JComponent)_internalPanel).setOpaque(color != null);
		((JComponent)_internalPanel.getParent()).setOpaque(color != null);
		((JComponent)_scrollingArea).setOpaque(color != null);
		((JComponent)_rootPanel).setOpaque(color != null);

		if (_toolTitlePanel != null)
			((JComponent)_toolTitlePanel).setOpaque(false);

		// If no image is given as a background, set the root panel opaque
		Image image = (Image)getEnvironment().getImage("LY_AREA_BACKGROUND_IMAGE", null, false);

		((JComponent)_subRootPanel).setOpaque(color != null || image == null);
	}

	/**
	 * Fills all the panels with the right values, if needed.
	 */
	public void setContent()
	{
		LySwingComponent internal;
		Component		 comp;

		// add title bar

		LyTitleBar		 titleBar = _area.getTitleBar();

		if (titleBar != null)
			addTitleBar(titleBar);

		// add action bar
		int				  i, n = _area.getToolBarCount();
		LyToolBar		  actions;
		Vector<LyToolBar> toolBars = new Vector<LyToolBar>();

		for (i = 0; i < n; i++)
		{
			actions = _area.getToolBar(i);
			toolBars.addElement(actions);
			if (actions != null)
			{
				internal = (LySwingComponent)actions.getInternalView();

				// register events
				if (internal instanceof LyEventTransmitter)
				{
					LyActionTransmitter transmitter = ((LyEventTransmitter)internal)
							.getActionTransmitter();

					transmitter.addActionListener(this);
				}
			}
		}
		addToolBars(toolBars);

		String tip = _area.getTip();

		if (tip != null)
			addTip();

		// Add sub-panels
		LyPanel item;

		n = _area.getPanelCount();
		_area.computePanelsPositions();

		for (i = 0; i < n; i++)
		{
			item = _area.getPanel(i);
			internal = (LySwingComponent)item.getInternalView();
			if (internal == null)
                continue;

			comp = internal.getComponent();
			_internalPanel.add(comp, getLayoutInfo(item));

			// register events
			if (internal instanceof LyEventTransmitter)
			{
				LyActionTransmitter transmitter = ((LyEventTransmitter)internal)
						.getActionTransmitter();

				transmitter.addActionListener(this);
			}
		}

		checkScrollPane();

		// add select bar
		LySelectBar selectBar = _area.getSelectBar();

		if (selectBar != null)
		{
			addSelectBar(selectBar);
			internal = (LySwingComponent)selectBar.getInternalView();
			// register events
			if (internal instanceof LyEventTransmitter)
			{
				LyActionTransmitter transmitter = ((LyEventTransmitter)internal)
						.getActionTransmitter();

				transmitter.addActionListener(this);
			}
		}

		// add commands
		LyCommands commands = _area.getCommands();

		if (commands != null)
		{
			addCommands(commands);
			internal = (LySwingComponent)commands.getInternalView();
			// register events
			if (internal instanceof LyEventTransmitter)
			{
				LyActionTransmitter transmitter = ((LyEventTransmitter)internal)
						.getActionTransmitter();

				transmitter.addActionListener(this);
			}
		}

		LyPanel panel;

		for (i = LyArea.NORTH; i <= LyArea.WEST; i++)
		{
			panel = _area.getBorderPanel((short)i);
			if (panel != null)
				setBorderPanel(panel, (short)i);
		}
	} // end method setContent

	/**
	 * Return the layout info for given panel.
	 *
	 * @param  panel the panel whose layout info must be returned.
	 * @return a string representing the layout information in a MigLayout for given panel.
	 */
	protected String getLayoutInfo(LyPanel panel)
	{
		int[]  position = _area.getPanelPosition(panel);
		double growx = 100, growy = 100, alignx = 0, aligny = 0;

		if ((panel instanceof LyGetConstraintInterface))
		{
			LyConstraint constraint = ((LyGetConstraintInterface)panel).getConstraint();

			if (constraint != null)
			{
				if (constraint.getColWeight() >= 0)
					growx = constraint.getColWeight();

				if (constraint.getRowWeight() >= 0)
					growy = constraint.getRowWeight();

				if (constraint.getHAlign() == LyConstraint.HAlign.CENTER)
					alignx = 50;
				else if (constraint.getHAlign() == LyConstraint.HAlign.RIGHT)
					alignx = 100;

				if (constraint.getVAlign() == LyConstraint.VAlign.MIDDLE)
					aligny = 50;
				else if (constraint.getVAlign() == LyConstraint.VAlign.BOTTOM)
					aligny = 100;
			}
		}

		String layoutInfo = "cell " + position[0] + " " + position[1] + ", span " + position[2]
			+ " " + position[3]
			+ ", growx " + growx + ", growy " + growy + ", align " + alignx + "% " + aligny + "%";

		return layoutInfo;
	} // end method getLayoutInfo

	/**
	 * Creates and configures the internal panel. Initializes its display properties as well as the
	 * layout to use.
	 *
	 * @return the configured internal panel
	 */
	public Container createInternalPanel()
	{
		Container internalPanel;
		Color     areaBackground = getColor("LY_AREA_BACKGROUND");

		if (areaBackground == null)
			areaBackground = getColor("LY_DEFAULT_BACKGROUND");

		boolean	fillx = true, filly = true;
		
		if (_area.getConstraint() != null)
		{
			LyConstraint.Fill	fill = _area.getConstraint().getFill();
			
			if (fill == LyConstraint.Fill.HORIZONTAL)
				filly = false;
			else if (fill == LyConstraint.Fill.VERTICAL)
				fillx = false;
			if (fill == LyConstraint.Fill.NONE)
			{
				fillx = false;
				filly = false;
			}
		}
		
		String	  fillParam = (fillx ? (filly ? "fill" : "fillx") : (filly ? "filly" : ""));
		String	  growx = (fillx ? "grow" : "");
		String	  growy = (filly ? "grow" : "");
		MigLayout layout = new MigLayout(fillParam + ", insets 0 0 0 0, gapx 0, gapy 0", growx, growy);

		internalPanel = new LyJBorderPanel(_area.getBorder());
		((LyJBorderPanel)internalPanel).setThickness(_area.getLabel() == null ? 1 : 0);
		((LyJBorderPanel)internalPanel).setGap(0);
		((LyJBorderPanel)internalPanel).setText(translate(_area.getLabel()));
		((LyJBorderPanel)internalPanel).setTextFont(getFont("LY_GROUP_FONT"));

		if (_area.getLabel() != null)
			((LyJBorderPanel)internalPanel).setColor(getColor("LY_GROUP_FONT_COLOR"));

		if (areaBackground != null)
			internalPanel.setBackground(areaBackground);

		internalPanel.setLayout(layout);
		return internalPanel;
	}

	/**
	 * Creates and configures the scrolling area. Depending on the scrollable property of this area,
	 * the scrolling will be automaticaly, always or never activated. This method takes care of
	 * disabling the scrolling borders if scrolling area imbrications can be avoided.
	 *
	 * @return a scrolling area
	 */
	public Container createScrollingArea()
	{
		JScrollPane scrollingArea = null;
		boolean     cancelScrollingAuto = false;

		if (_area.getScrolling() == LyArea.Scrolling.YES)
			scrollingArea = new JScrollPane(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS,
					ScrollPaneConstants.HORIZONTAL_SCROLLBAR_ALWAYS);
		else if (_area.getScrolling() == LyArea.Scrolling.AUTO)
		{
			if (!isClientPanelScrollable())
				scrollingArea = new JScrollPane(ScrollPaneConstants.VERTICAL_SCROLLBAR_NEVER,
						ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
			else
			{
				// If parent is a scrolling auto area with only this area as a child,
				// remove scrolling for this area to avoid a scroll on both components.
				LyComponent parent = _area.getParent();

				if ((parent instanceof LyArea) && (((LyArea)parent).getPanelCount() == 1)
						&& (((LyArea)parent).getScrolling() == LyArea.Scrolling.AUTO))
				{
					scrollingArea = new JScrollPane(ScrollPaneConstants.VERTICAL_SCROLLBAR_NEVER,
							ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
					scrollingArea.setBorder(null);
					cancelScrollingAuto = true;
				}
				else
					scrollingArea = new JScrollPane(
							ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
							ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
			}
		}
		else
		// _area.getScrolling() == LyArea.Scrolling.NO
		{
			scrollingArea = new JScrollPane(ScrollPaneConstants.VERTICAL_SCROLLBAR_NEVER,
					ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
		}

		if (scrollingArea.getVerticalScrollBar() != null)
			scrollingArea.getVerticalScrollBar().setUnitIncrement(10);

		if (scrollingArea.getHorizontalScrollBar() != null)
			scrollingArea.getHorizontalScrollBar().setUnitIncrement(10);

		scrollingArea.getViewport().setMinimumSize(new Dimension(0, 0));

		if (!cancelScrollingAuto)
		{
			if (_area.getBorder() == LyArea.Border.NONE)
			{
				// Check if a border must be added on the scrolling panel.
				LyComponent parent = _area.getParent();
				boolean     addBorder = false;

				if ((parent instanceof LyArea)
						&& (((LyArea)parent).getPanelCount() == 1)
						&& (((LyArea)parent).getScrolling() == LyArea.Scrolling.NO))
					parent = parent.getParent();

				if ((parent instanceof LyView) || (parent instanceof LySplitter))
					addBorder = true;
				
				if ((_area.getPanelCount() == 1)
						&& (_area.getPanel(0) instanceof LyTabs))
					addBorder = false;

				if (addBorder)
					scrollingArea.setBorder(BorderFactory.createCompoundBorder(
							BorderFactory.createEmptyBorder(0, 1, 0,
									1), BorderFactory.createEtchedBorder(EtchedBorder.LOWERED)));
				else
					scrollingArea.setBorder(null);
			}
			else
				scrollingArea.setBorder(null);
		}

		return scrollingArea;
	} // end method createScrollingArea

	/**
	 * Creates and configures the root panel. This panel is composed by the {@link #_internalPanel}
	 * and can be scollable. It is also composed by a titlebar, toolbars and an help tip represented
	 * as a String.
	 *
	 * @return the root panel with its sub-areas.
	 * @see    LyAreaBase#getTip()
	 */
	public Container createRootPanel()
	{
		Container rootPanel;
		Color     areaBackground = getColor("LY_AREA_BACKGROUND");

		if (areaBackground == null)
			areaBackground = getColor("LY_DEFAULT_BACKGROUND");

		_toolTitlePanel = null;

		// Prepare the sub-root panel
		String  imagePath = getEnvironment().getImagePath("LY_AREA_BACKGROUND_IMAGE");
		LyArea  parentArea = _area.getParentArea();
		String  parentBgImagePath = ((parentArea == null)? null
				: parentArea.getEnvironment().getImagePath("LY_AREA_BACKGROUND_IMAGE"));

		// If there is no sub-panel, give to the area the toolbars background
		if (_area.getPanelCount() == 0)
			imagePath = getEnvironment().getImagePath("LY_TOOLBAR_BACKGROUND_IMAGE");

		boolean isBorderPanel = (parentArea != null) && parentArea.isBorderPanel(_area);

		// Use background only for top areas
		if ((_area != null)
				&& !(_area.getParent() instanceof LyTab)
				&& !(_area.getParent() instanceof LySplitter)
				&& !isBorderPanel)
		{
			// Check if image is the same as the parent
			boolean isSameBgAsParent = imagePath.equals(parentBgImagePath);

			if (isSameBgAsParent)
				imagePath = null;
		}

        Image  image = (Image)getEnvironment().getImage("LY_AREA_BACKGROUND_IMAGE", null, false);
        
		_subRootPanel = new LyJBackgroundPanel(image);
		((JComponent)_subRootPanel).setBorder(null);
		_subRootPanel.setLayout(new BorderLayout());

		// Add the scrollPane if client panel is scrollable
		if (isClientPanelScrollable())
		{
			((JScrollPane)_scrollingArea).setViewportView(_internalPanel);
			_subRootPanel.add(_scrollingArea, BorderLayout.CENTER);
		}
		else
			_subRootPanel.add(_internalPanel, BorderLayout.CENTER);

		// Prepare the root panel
		rootPanel = new JPanel();
		rootPanel.setLayout(new BorderLayout());
		((JPanel)rootPanel).setBorder(null);
		// Prepare the panel containing titleBar, Top toolbars and tip
		if ((_area.getTitleBar() != null) || (_area.getTip() != null)
				|| (_area.getToolBarCount() > 0))
		{
			_toolTitlePanel = new JPanel();
			((JComponent)_toolTitlePanel).setBorder(null);
			_toolTitlePanel.setLayout(new GridBagLayout());
		}

		if (areaBackground != null)
		{
			rootPanel.setBackground(areaBackground);
			_subRootPanel.setBackground(areaBackground);

			if (_toolTitlePanel != null)
				_toolTitlePanel.setBackground(areaBackground);
		}

		if (_toolTitlePanel != null)
			rootPanel.add(_toolTitlePanel, BorderLayout.NORTH);

		if (_area.getPanelCount() > 0)
			rootPanel.add(_subRootPanel, BorderLayout.CENTER);

		return rootPanel;
	} // end method createRootPanel

	/**
	 * Determines if this area needs scrollbars. It is not necessary to have scrollbars if there is
	 * only one panel in the area, or if it is a notice element.
	 *
	 * @return true if the area needs scrollbars
	 */
	public boolean isClientPanelScrollable()
	{
		int     i, n = _area.getPanelCount();
		boolean result = true;
		LyTab   parentTab = _area.getParentTab();

		if ((parentTab == null) || (n == 1))
		{
			for (i = 0; i < n; i++)
			{
				LyPanel panel = _area.getPanel(i);

				// Only form is not scrollable !
				// XXX Improve with interfaces later
				if (panel instanceof LyNotice)
					result &= false;

				result &= panel.isScrollable();
			}
		}

		return result;
	}

	/**
	 * Checks that there aren't a scrolling area alone in another scrolling area. Called when a
	 * panel is added or removed.
	 */
	public void checkScrollPane()
	{
		// If root component is the internal panel, no need to check scroll pane.
		if (getComponent() == _internalPanel)
			return;

		if (_area.getPanelCount() > 1)
			return;

		if (_internalPanel.getComponentCount() == 0)
		{
			if (_subRootPanel.isAncestorOf(_scrollingArea))
				_subRootPanel.remove(_scrollingArea);
			if (_subRootPanel.isAncestorOf(_internalPanel))
				_subRootPanel.remove(_internalPanel);
		}

		// TO AVOID A SCROLLPANE CONTAINING ONLY ANOTHER SCROLLPANE
		if (_internalPanel.getComponentCount() == 1)
		{
			if ((_internalPanel.getComponent(0) instanceof JScrollPane)
					|| (_internalPanel.getComponent(0) instanceof LyJSplitterComponent)
					|| (_internalPanel.getComponent(0) instanceof LyJNoScrollPanel)
					|| (((JScrollPane)_scrollingArea).getHorizontalScrollBarPolicy()
						== ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER))
			{
				if (_scrollingArea.isAncestorOf(_internalPanel))
					_scrollingArea.remove(_internalPanel);

				if (_subRootPanel.isAncestorOf(_scrollingArea))
					_subRootPanel.remove(_scrollingArea);

				if (!_subRootPanel.isAncestorOf(_internalPanel))
					_subRootPanel.add(_internalPanel, BorderLayout.CENTER);
			}
			else
			{
				if (_subRootPanel.isAncestorOf(_internalPanel))
					_subRootPanel.remove(_internalPanel);

				if (!_scrollingArea.isAncestorOf(_internalPanel))
					_scrollingArea.add(_internalPanel);

				if (!_subRootPanel.isAncestorOf(_scrollingArea))
					_subRootPanel.add(_scrollingArea, BorderLayout.CENTER);
			}
		}
		if (_internalPanel.getComponentCount() > 1)
		{
			if (_subRootPanel.isAncestorOf(_internalPanel))
				_subRootPanel.remove(_internalPanel);

			if (!_scrollingArea.isAncestorOf(_internalPanel))
				_scrollingArea.add(_internalPanel);

			if (!_subRootPanel.isAncestorOf(_scrollingArea))
				_subRootPanel.add(_scrollingArea, BorderLayout.CENTER);
		}
	} // end method checkScrollPane

	/**
	 * Creates and adds the titlebar in this area.
	 *
	 * @param titleBar the LyComponent to add
	 */
	public void addTitleBar(LyTitleBar titleBar)
	{
		if (_toolTitlePanel == null)
			return;

		_titleBar = ((LySwingComponent)titleBar.getInternalView()).getComponent();
		GridBagConstraints c = this.buildTitleBarConstraint(titleBar);

		_toolTitlePanel.add(_titleBar, c);
	}

	/**
	 * Creates and adds the toolbars in this area. The <code>_toolBars</code> field is <code>
	 * null</code> if the <code>toolbars</code> parameter is <code>null</code>, else it is a vector
	 * of all the toolbars around this area.
	 *
	 * @param toolbars the LyComponent to add
	 */
	public void addToolBars(Vector<LyToolBar> toolbars)
	{
		if (_toolTitlePanel == null)
			return;

		if ((toolbars == null) || (toolbars.size() == 0))
		{
			_toolBars = null;
			return;
		}

		_toolBars = new Vector<Component>();

		Container		   subPanel = null;
		GridBagConstraints c;
		LyToolBar		   toolBar;
		LySwingComponent   internal;

		if (_area.getPanelCount() > 0)
			subPanel = _subRootPanel;
		else
			subPanel = _rootPanel;

		LyToolBar topRightToolBar = _area.getTopRightToolBar();
		boolean   hasTopRight = (topRightToolBar != null);

		toolBar = _area.getTopLeftToolBar();

		Image  image = (Image)getEnvironment().getImage("LY_TOOLBAR_BACKGROUND_IMAGE", null, false);
		String fieldId = _area.getParentFormFieldId();
		JPanel toolbarsPanel;

		if ((toolBar != null) || hasTopRight)
		{
			if ((fieldId == null) || (fieldId.length() == 0))
				toolbarsPanel = new LyJBackgroundPanel(image);
			else
			{
				toolbarsPanel = new JPanel();
				toolbarsPanel.setOpaque(false);
			}

			toolbarsPanel.setLayout(new GridBagLayout());
			c = buildActionConstraint(null);
			_toolTitlePanel.add(toolbarsPanel, c);
		}
		else
			toolbarsPanel = (JPanel)_toolTitlePanel;

		if (toolBar != null)
		{
			internal = (LySwingComponent)toolBar.getInternalView();
			_toolBars.addElement(internal.getComponent());
			c = this.buildActionConstraint(toolBar, !hasTopRight);
			toolbarsPanel.add(internal.getComponent(), c);
		}

		toolBar = topRightToolBar;
		if (toolBar != null)
		{
			internal = (LySwingComponent)toolBar.getInternalView();
			_toolBars.addElement(internal.getComponent());
			c = this.buildActionConstraint(toolBar);
			
			if (toolBar.getComponentCount() == 0)
				c.weightx = 0.0;
			
			toolbarsPanel.add(internal.getComponent(), c);
		}

		toolBar = _area.getLeftToolBar();
		if (toolBar != null)
		{
			internal = (LySwingComponent)toolBar.getInternalView();
			_toolBars.addElement(internal.getComponent());

			if ((fieldId == null) || (fieldId.length() == 0))
				toolbarsPanel = new LyJBackgroundPanel(image);
			else
			{
				toolbarsPanel = new JPanel();
				toolbarsPanel.setOpaque(false);
			}

			toolbarsPanel.setLayout(new GridBagLayout());
			c = buildActionConstraint(null);
			
			if (toolBar.getStyle() != LyToolBar.Style.EXPAND)
			{
				c.anchor = GridBagConstraints.WEST;
				c.fill = GridBagConstraints.HORIZONTAL;
			}

			if (_rootPanel == _internalPanel)
			{
				toolbarsPanel.add(internal.getComponent(), c);
				c = this.buildActionConstraint(toolBar);
				subPanel.add(toolbarsPanel, c);
			}
			else
			{
				toolbarsPanel.add(internal.getComponent(), c);
				subPanel.add(toolbarsPanel, BorderLayout.WEST);
			}
		}

		toolBar = _area.getRightToolBar();
		if (toolBar != null)
		{
			internal = (LySwingComponent)toolBar.getInternalView();
			_toolBars.addElement(internal.getComponent());

			if ((fieldId == null) || (fieldId.length() == 0))
				toolbarsPanel = new LyJBackgroundPanel(image);
			else
			{
				toolbarsPanel = new JPanel();
				toolbarsPanel.setOpaque(false);
			}

			toolbarsPanel.setLayout(new GridBagLayout());
			c = buildActionConstraint(null);

			if (toolBar.getStyle() != LyToolBar.Style.EXPAND)
			{
				c.anchor = GridBagConstraints.WEST;
				c.fill = GridBagConstraints.HORIZONTAL;
			}

			if (_rootPanel == _internalPanel)
			{
				toolbarsPanel.add(internal.getComponent(), c);
				c = this.buildActionConstraint(toolBar);
				subPanel.add(toolbarsPanel, c);
			}
			else
			{
				toolbarsPanel.add(internal.getComponent(), c);
				subPanel.add(toolbarsPanel, BorderLayout.EAST);
			}

		}

		LyToolBar bottomLeftToolBar = _area.getBottomLeftToolBar();
		LyToolBar bottomRightToolBar = _area.getBottomRightToolBar();
		boolean   hasBottomLeft = (bottomLeftToolBar != null);
		boolean   hasBottomRight = (bottomRightToolBar != null);

		if (hasBottomLeft || hasBottomRight)
		{
			// Create a panel with a gridbag layout
			JPanel bottomPanel = new JPanel();

			bottomPanel.setBackground(_rootPanel.getBackground());
			bottomPanel.setLayout(new GridBagLayout());

			if (hasBottomLeft)
			{
				toolBar = bottomLeftToolBar;
				c = this.buildActionConstraint(toolBar, !hasBottomRight);
				c.weightx = 1.0;
				internal = (LySwingComponent)toolBar.getInternalView();
				_toolBars.addElement(internal.getComponent());
				bottomPanel.add(internal.getComponent(), c);
			}

			if (hasBottomRight)
			{
				toolBar = bottomRightToolBar;
				c = this.buildActionConstraint(toolBar);
				internal = (LySwingComponent)toolBar.getInternalView();
				_toolBars.addElement(internal.getComponent());
				bottomPanel.add(internal.getComponent(), c);
			}

			// If needed add commands in this panel
			if (_area.getCommands() != null)
			{
				LyCommands commands = _area.getCommands();

				_commandBar = ((LySwingComponent)commands.getInternalView()).getComponent();

				c = buildCommandConstraint(commands);

				bottomPanel.add(_commandBar, c);
			}

			subPanel.add(bottomPanel, BorderLayout.SOUTH);
		}

	} // end method addToolBars

	/**
	 * Creates and adds the tip label in this area if it has a title or a tool panel and if this
	 * area has a tip.
	 */
	public void addTip()
	{
		if (_toolTitlePanel == null)
			return;

		if (_area.getTip() != null)
		{
			Image			   image = (Image)getEnvironment().getImage("LY_AREA_BACKGROUND_IMAGE",
					null, false);
			LyJBackgroundPanel panel = new LyJBackgroundPanel(image);

			panel.setLayout(new BorderLayout());
			int n = _toolTitlePanel.getComponentCount();

			panel.setBorder(null);

			// Add a separator if at least one component is displayed in the tooltitle panel
			if (n > 0)
			{
				Component separator = new LyJSeparator(SwingConstants.HORIZONTAL,
						LyJSeparator.TOOLBAR_STYLE, Color.darkGray);

				panel.add(separator, BorderLayout.NORTH);
			}

			Component tip = getTipComponent();

			panel.add(tip, BorderLayout.CENTER);

			if (_area.getPanelCount() > 0)
			{
				Component separator = new LyJSeparator(SwingConstants.HORIZONTAL,
						LyJSeparator.TOOLBAR_STYLE, Color.darkGray);

				panel.add(separator, "South");
			}

			GridBagConstraints c = new GridBagConstraints();

			c.gridwidth = GridBagConstraints.REMAINDER;
			c.weightx = 1.0;
			c.weighty = 1.0;
			c.insets = new Insets(0, 0, 0, 0);
			c.anchor = GridBagConstraints.NORTHWEST;
			c.fill = GridBagConstraints.BOTH;
			_toolTitlePanel.add(panel, c);
		}
	} // end method addTip

	/**
	 * Gets the graphical component that displays the tip of this area.
	 *
	 * <p>This tip is a {@link LyJMultiLineLabel} that can have any number of characters, a margin
	 * width of 3 pixels and no height margin.
	 *
	 * @return the tip of this area
	 */
	protected Component getTipComponent()
	{
		Font   font = getFont("LY_TIP_FONT");
		Color  color = getColor("LY_TIP_COLOR");
		String tip = _area.getTip();

		tip = translate(tip);

		JTextArea text = new JTextArea(tip);
		JPanel    panel = new JPanel(new GridBagLayout());

		// Tips may be reduced to a minimum size
		text.setMinimumSize(new Dimension(10, 10));
		panel.setBorder(BorderFactory.createEmptyBorder(4, 4, 4, 4));
		panel.setBackground(null);
		panel.setOpaque(false);

		GridBagConstraints c = new GridBagConstraints();

		c.gridwidth = GridBagConstraints.REMAINDER;
		c.weightx = 1.0;
		c.weighty = 1.0;
		c.insets = new Insets(0, 0, 0, 0);
		c.anchor = GridBagConstraints.NORTHWEST;
		c.fill = GridBagConstraints.BOTH;
		panel.add(text, c);

		text.setLineWrap(true);
		text.setWrapStyleWord(true);
		text.setBackground(null);
		text.setOpaque(false);
		text.setFont(font);
		text.setForeground(color);
		text.setEditable(false);

		return panel;
	} // end method getTipComponent

	/**
	 * Creates and adds the select bar in this area. If this area also have a command bar, the both
	 * of them are added in a new panel.
	 *
	 * @param selectBar the select bar to add
	 */
	public void addSelectBar(LySelectBar selectBar)
	{
		_selectBar = ((LySwingComponent)selectBar.getInternalView()).getComponent();

		if (_southPanel == null)
			buildSouthPanel();

		if (_area.getCommands() == null)
			_southPanel.add(_selectBar);
		else
		{
			// ADD BOTH : selectBar + commands
			LySwingComponent internal = (LySwingComponent)_area.getCommands().getInternalView();

			_southPanel.setLayout(new GridLayout(2, 1));
			_southPanel.add(_selectBar);
			_southPanel.add(internal.getComponent());
		}
	}

	/**
	 * Initializes the commands in this area. If this area has no toolbars in its bottom, the
	 * command bar, and the select bar if any, are added.
	 *
	 * @param commands the LyComponent to add
	 */
	public void addCommands(LyCommands commands)
	{
		_commandBar = ((LySwingComponent)commands.getInternalView()).getComponent();

		if (_southPanel == null)
			buildSouthPanel();

		if ((_area.getBottomLeftToolBar() == null) && (_area.getBottomRightToolBar() == null))
		{
			if (_area.getSelectBar() == null)
				_southPanel.add(_commandBar);
			else
			{
				// ADD BOTH : selectBar + commands
				_southPanel.setLayout(new GridLayout(2, 1));
				_southPanel.add(_selectBar);
				_southPanel.add(_commandBar);
			}
		}
	}

	/**
	 * Build the south panel which may contain the select bar, the command bar and south border
	 * panel.
	 */
	protected void buildSouthPanel()
	{
		_southPanel = new JPanel();
		_southPanel.setBackground(_rootPanel.getBackground());
		_southPanel.setLayout(new GridLayout(1, 1));
		_rootPanel.add(_southPanel, BorderLayout.SOUTH);
	}

	/**
	 * Sets the waiting status of the specified component. This waiting status changes the cursor
	 * image (<code>LySwingViewManager.WAIT_CURSOR</code> or <code>
	 * LySwingViewManager.DEFAULT_CURSOR</code>) and the enabled state of the component.
	 *
	 * @param c      the component to set
	 * @param status sets the waiting property of the given component, false unsets it.
	 */
	protected void setWaiting(Component c, boolean status)
	{
		if (c != null)
		{
			if (status)
				c.setCursor(LySwingViewManager.WAIT_CURSOR);
			else
				c.setCursor(LySwingViewManager.DEFAULT_CURSOR);

			c.setEnabled(!status);
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalArea#addPanel(leon.view.LyPanel, int)
	 */
	public void addPanel(LyPanel panel, int pos)
	{
		insertPanel(panel, pos);

		LyView	    view = _area.getParentView();
		LySwingView swingView = ((view == null) ? null : (LySwingView)view.getInternalView());

		if (swingView != null)
		{
			swingView.configureDefaultButton();

			if (swingView instanceof LySwingDialog)
				((LySwingDialog)swingView).handleFocus();
		}

		// The size of the panel may have changed !
		// Check in the hierachy if another component may accept
		// the resetContent action.
		LyComponent parent = _area.getParent();
		Object	    internal;
		boolean     resetContent;

		while (parent != null)
		{
			internal = parent.getInternalView();

			if (internal instanceof LyResetContentInterface)
			{
				resetContent = true;

				// Reset content only if parent is not a resizable frame (otherwise the frame
				// may flash too much)
				if ((internal instanceof LySwingFrame) && (((LySwingFrame)internal).isResizable()))
					resetContent = false;

				// If parent is a non resizable dialog which has a default width or height,
				// do not change size
				if ((internal instanceof LySwingDialog)
						&& !(((LyDialog)parent).isResizable()))
				{
					LyDialog dialog = (LyDialog)parent;

					resetContent = ((dialog.getHeight() <= 0) || (dialog.getWidth() <= 0));
				}

				if (resetContent)
					((LyResetContentInterface)internal).resetContent();

				// reach the form
				if (!(internal instanceof LySwingFormField))
					return;
			}

			parent = parent.getParent();
		}
	} // end method addPanel

	/**
	 * Appends the specified panel at the position given inside this area.
	 *
	 * @param panel the panel to insert
	 * @param pos   position of the panel inside this container
	 * @see   #addPanel
	 */
	public void insertPanel(LyPanel panel, int pos)
	{
		LySwingComponent comPanel = (LySwingComponent)panel.getInternalView();
		Component		 comp = comPanel.getComponent();

		if (isClientPanelScrollable())
			_subRootPanel.add(_scrollingArea, BorderLayout.CENTER);

		// Do not remove following line : used to force the panel to
		// repaint all its sub-components (doLayout() does not work here).
		// comp.resize(0, 0);

		_area.computePanelsPositions();
		_internalPanel.add(comp, getLayoutInfo(panel));

		// Add or remove the scrolling area
		checkScrollPane();

		// register events
		if (comPanel instanceof LyEventTransmitter)
		{
			LyActionTransmitter transmitter = ((LyEventTransmitter)comPanel).getActionTransmitter();

			transmitter.addActionListener(this);
		}

		((JComponent)_internalPanel).repaint();
		((JComponent)_internalPanel).revalidate();
	} // end method insertPanel

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.graphic.LyGraphicalArea#removePanel(leon.view.LyPanel)
	 */
	public void removePanel(LyPanel panel)
	{
		LySwingComponent comPanel = (LySwingComponent)panel.getInternalView();

		if (comPanel == null)
			return;

		Component comp = comPanel.getComponent();

		_internalPanel.remove(comp);
        panel.reset();

		if (_internalPanel.getComponentCount() == 0)
			checkScrollPane();

		// unregister events
		if (comPanel instanceof LyEventTransmitter)
		{
			LyActionTransmitter transmitter = ((LyEventTransmitter)comPanel).getActionTransmitter();

			transmitter.removeActionListener(this);
		}
	}

	// FFT00557
	/**
	 * Sets the scrolling property of the different panels composing this area. The scrolling is
	 * enabled when the scrolling area is not already used. The same applies when it is disabled. In
	 * both cases, the <code>_subRootPanel</code> is modified to use the <code>_scrollingArea</code>
	 * or not.
	 *
	 * @param enabled true to enable the scrolling property, false to disable
	 */
	public void setScrollingEnabled(boolean enabled)
	{
		if (enabled)
		{
			if (!_scrollingArea.isAncestorOf(_internalPanel))
			{
				_subRootPanel.remove(_internalPanel);
				_scrollingArea.add(_internalPanel);
				_subRootPanel.add(_scrollingArea);
			}
		}
		else
		{
			if (_scrollingArea.isAncestorOf(_internalPanel))
			{
				_scrollingArea.remove(_internalPanel);
				_subRootPanel.remove(_scrollingArea);
				_subRootPanel.add(_internalPanel);
			}
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetWaitingInterface#setWaiting(boolean)
	 */
	public void setWaiting(boolean status)
	{
		Object lock = ((_area == null) ? this : _area.getLock());

		synchronized (lock)
		{
			setWaiting(_rootPanel, status);
			setWaiting(_titleBar, status);
			if ((_toolBars != null) && (_toolBars.size() > 0))
			{
				for (int i = 0; i < _toolBars.size(); i++)
					setWaiting(_toolBars.elementAt(i), status);
			}
			setWaiting(_selectBar, status);
			setWaiting(_commandBar, status);
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		// If area is not scrollable and only contains sub-areas, just return internal panel.
		if ((_area.getScrolling() == LyArea.Scrolling.NO)
				&& (_area.getPanelCount() > 1)
				&& (_area.getTitleBar() == null)
				&& (_area.getTip() == null)
				&& (_area.getToolBarCount() == 0)
				&& (_area.getSelectBar() == null)
				&& (_area.getCommands() == null))
			return _internalPanel;

		return _rootPanel;
	}

	/**
	 * Builds a contraint object for the specified toolbar. This constraint force the toolbar to be
	 * added at the end of this area.
	 *
	 * @param  actions Specifies the toolbar that its constrains are built.
	 * @return the constraints of the specified toolbar
	 * @see    #buildActionConstraint(LyToolBar, boolean)
	 */
	public GridBagConstraints buildActionConstraint(LyToolBar actions)
	{
		return buildActionConstraint(actions, true);
	}

	/**
	 * Builds a contraint object for the specified toolbar. If the toolbar <code>actions</code> is
	 * <code>null</code>, the constraint object is also <code>null</code>. Else, the constraint
	 * object that is returned is a {@link GridBagConstraints} that follows the toolbar properties
	 * for its filling and alignment.
	 *
	 * @param  actions   Specifies the toolbar that its constrains are built.
	 * @param  remainder Specifies if the toolbar is the last component of its row or column.
	 * @return the constraints of the specified toolbar
	 */
	public GridBagConstraints buildActionConstraint(LyToolBar actions, boolean remainder)
	{
		GridBagConstraints c = new GridBagConstraints();

		if (actions == null)
		{
			c.gridwidth = GridBagConstraints.REMAINDER;
			c.weightx = 1.0;
			c.weighty = 1.0;
			c.insets = new Insets(0, 0, 0, 0);
			c.anchor = GridBagConstraints.NORTHWEST;
			c.fill = GridBagConstraints.BOTH;

			return c;
		}

		// default
		if (actions.getFill() == LyToolBar.Fill.VERTICAL)
		{
			c.gridwidth = 1;
			c.gridheight = 1;
		}
		else
		{
			c.gridwidth = 1;
			c.gridheight = 1;
		}

		c.weightx = 1.0;
		c.weighty = 0.0;

		if (actions.getHAlign() == LyToolBar.HAlign.LEFT)
			c.anchor = GridBagConstraints.WEST;
		else
			c.anchor = GridBagConstraints.EAST;

		c.fill = GridBagConstraints.HORIZONTAL;
		// user
		LyConstraint requestedConstraint = null;

		// do we have any constrain specified for the node ?
		if (actions != null)
			requestedConstraint = ((LyGetConstraintInterface)actions).getConstraint();

		c = LySwingConstraint.buildConstraint(actions, requestedConstraint, c);
		return (c);
	} // end method buildActionConstraint

	/**
	 * Builds a contraint object for a title bar. The titleBar object is not used, as the
	 * constraints are the same from one titlebar to another.
	 *
	 * @param  titleBar a title bar, may be <code>null</code>
	 * @return the constraints of a generic title bar
	 * @see    #buildDefaultConstraint(LyComponent)
	 */
	public GridBagConstraints buildTitleBarConstraint(LyTitleBar titleBar)
	{
		GridBagConstraints c = buildDefaultConstraint(titleBar);

		c.insets = new Insets(0, 0, 0, 0);

		return c;
	}

	/**
	 * Builds a contraint object for the select bar. If the select bar is <code>null</code>, the
	 * returned <code>GridBagConstraints</code> is <code>null</code>. The final constraint for the
	 * specified select bar is based on its current constraints that are extended with a generic
	 * <code>GridBagConstraints</code>.
	 *
	 * @param  selectBar the select bar to build its constraints
	 * @return the constraints of the specified select bar
	 * @see    #buildCommandConstraint(LyCommands)
	 */
	public GridBagConstraints buildSelectBarConstraint(LySelectBar selectBar)
	{
		if (selectBar == null)
			return null;

		GridBagConstraints c = new GridBagConstraints();

		c.gridwidth = GridBagConstraints.REMAINDER;
		c.gridheight = 1;
		c.weightx = 0.0;
		c.weighty = 0.0;
		c.anchor = GridBagConstraints.NORTHWEST;
		c.fill = GridBagConstraints.BOTH;

		// user
		LyConstraint requestedConstraint = null;

		// do we have any constraint specified for the node ?
		if (selectBar != null)
			requestedConstraint = ((LyGetConstraintInterface)selectBar).getConstraint();

		c = LySwingConstraint.buildConstraint(selectBar, requestedConstraint, c);
		return (c);
	}

	/**
	 * Builds a contraint object for the command bar. If the command bar is <code>null</code>, the
	 * returned <code>GridBagConstraints</code> is <code>null</code>. The final constraint for the
	 * specified command bar is based on its current constraints that are extended with a generic
	 * <code>GridBagConstraints</code>.
	 *
	 * @param  commands the command bar to build its constraints
	 * @return the constraints of the specified command bar
	 * @see    #buildSelectBarConstraint(LySelectBar)
	 */
	public GridBagConstraints buildCommandConstraint(LyCommands commands)
	{
		if (commands == null)
			return null;

		GridBagConstraints c = new GridBagConstraints();

		c.gridwidth = GridBagConstraints.REMAINDER;
		c.gridheight = 1;
		c.weightx = 0.0;
		c.weighty = 0.0;
		c.anchor = GridBagConstraints.NORTHWEST;
		c.fill = GridBagConstraints.BOTH;

		// user
		LyConstraint requestedConstraint = null;

		// do we have any constraint specified for the node ?
		if (commands != null)
			requestedConstraint = ((LyGetConstraintInterface)commands).getConstraint();

		c = LySwingConstraint.buildConstraint(commands, requestedConstraint, c);
		return c;
	}

	/**
	 * Builds a contraint object for the layout of the specified label or component. The component
	 * object is not used, thus the constraints are generic from one component to another.
	 *
	 * @param  comp a component, not used in this implementation
	 * @return the constraints of a generic component
	 */
	public GridBagConstraints buildDefaultConstraint(LyComponent comp)
	{
		GridBagConstraints c = new GridBagConstraints();

		c.gridwidth = GridBagConstraints.REMAINDER;
		c.weightx = 1.0;
		c.weighty = 0.0;
		c.anchor = GridBagConstraints.NORTHWEST;
		c.fill = GridBagConstraints.BOTH;

		return c;
	}

	/**
	 * Builds a contraint object for the layout of the specified label or component. The requested
	 * constraint is applied to the default constraint generated for every component (see {@link
	 * #buildDefaultConstraint(LyComponent)}).
	 *
	 * @param  comp                the component to build its constraints
	 * @param  requestedConstraint the constraints to apply on the given component
	 * @return the constraints of the specified component
	 */
	public GridBagConstraints buildConstraint(LyComponent comp, LyConstraint requestedConstraint)
	{
		GridBagConstraints c = buildDefaultConstraint(comp);

		c = LySwingConstraint.buildConstraint(comp, requestedConstraint, c);
		return c;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		// FFT00468
		if (_area == null)
			return;

		// Notify controller if any
		if (_area.getController() != null)
		{
			// Notify controller
			LyDemand demand = LySwingArea.buildActionDemand(actionEvent);

			// Set info :
			demand.setParameterValues(new LySwingParameterValues(_area));

			processDemand(demand);
			actionEvent.consume();
		}

		// notify other listeners
		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(actionEvent);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyFreeInterface#free()
	 */
	public void free()
	{
		Object lock = ((_area == null) ? this : _area.getLock());

		synchronized (lock)
		{
			_rootPanel = null;
			_scrollingArea = null;
			_internalPanel = null;
			_area = null;
			_toolBars = null;
			_commandBar = null;

			super.free();
		}
	}

	/**
	 * Do not update the refresh value.
	 *
	 * @param duration this parameter is not read
	 * @see   LyGraphicalArea#setRefresh(int)
	 */
	public void setRefresh(int duration)
	{
	}
	
	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalArea#setOpen(boolean)
	 */
	public void setOpen(boolean open)
	{
		((LyJBorderPanel)_internalPanel).setOpen(_area.isOpen());
	}
	
	/**
	 * Set a border panel at the given position.
	 *
	 * @param panel    the border panel.
	 * @param position Either NORTH, SOUTH, EAST or WEST position.
	 */
	public void setBorderPanel(LyPanel panel, short position)
	{
		if (panel == null)
		{
			if (_borderPanels == null)
				return;

			LySwingPanel borderPanel = _borderPanels[position];

			if (borderPanel == null)
				return;

			Component compPanel = borderPanel.getComponent();

			if (position == LyArea.NORTH)
			{
				if (_toolTitlePanel != null)
					_toolTitlePanel.remove(compPanel);
				else
					_rootPanel.remove(compPanel);
			}
			else if (position == LyArea.SOUTH)
			{
				GridLayout layout = (GridLayout)_southPanel.getLayout();

				layout.setRows(layout.getRows() - 1);
				_southPanel.remove(compPanel);
			}
			else if (position == LyArea.WEST)
				_rootPanel.remove(compPanel);
			else if (position == LyArea.EAST)
				_rootPanel.remove(compPanel);

			// unregister events
			LyActionTransmitter transmitter = borderPanel.getActionTransmitter();

			transmitter.removeActionListener(this);
			_borderPanels[position] = null;
		}
		else
		{
			LySwingPanel borderPanel = null;

			if (_borderPanels == null)
				_borderPanels = new LySwingPanel[4];
			else
			{
				borderPanel = _borderPanels[position];

				if (borderPanel != null)
					setBorderPanel(null, position);
			}

			borderPanel = (LySwingPanel)panel.getInternalView();
			_borderPanels[position] = borderPanel;

			Container compPanel = (Container)borderPanel.getComponent();

			if (position == LyArea.NORTH)
			{
				GridBagConstraints c = this.buildBorderPanelConstraint(panel);

				if (_toolTitlePanel != null)
					_toolTitlePanel.add(compPanel, c);
				else
					_rootPanel.add(compPanel, BorderLayout.NORTH);
			}
			else if (position == LyArea.SOUTH)
			{
				if (_southPanel == null)
					buildSouthPanel();
				else
				{
					GridLayout layout = (GridLayout)_southPanel.getLayout();

					layout.setRows(layout.getRows() + 1);
				}

				_southPanel.add(compPanel);
			}
			else if (position == LyArea.WEST)
				_rootPanel.add(compPanel, BorderLayout.WEST);
			else if (position == LyArea.EAST)
				_rootPanel.add(compPanel, BorderLayout.EAST);

			// register events
			LyActionTransmitter transmitter = borderPanel.getActionTransmitter();

			transmitter.addActionListener(this);
		}

		((JComponent)_internalPanel).repaint();
		((JComponent)_internalPanel).revalidate();
	} // end method setBorderPanel

	/**
	 * Builds a constraint object for the given panel.
	 */
	protected GridBagConstraints buildBorderPanelConstraint(LyPanel panel)
	{
		if (panel == null)
			return null;

		GridBagConstraints c = new GridBagConstraints();

		c.gridwidth = GridBagConstraints.REMAINDER;
		c.gridheight = 1;
		c.weightx = 0.0;
		c.weighty = 0.0;
		c.fill = GridBagConstraints.BOTH;

		// Do we have any constraint specified for the panel ?
		LyConstraint requestedConstraint = ((LyGetConstraintInterface)panel).getConstraint();

		c = LySwingConstraint.buildConstraint(panel, requestedConstraint, c);
		return c;
	}
    
    /**
     * Sets the visibility status of this component.
     * @param status New visibility status.
     */
    public void setVisible(boolean visible)
    {
        if (!(_area.getParent() instanceof LyTab))
        {
            Component root = getComponent();
            
            root.setVisible(visible);
        }
    }
    
    /**
     * (non-Javadoc)
     * 
     * @see leon.view.graphic.LyGraphicalArea#setRedirect(java.lang.String)
     */
    public void setRedirect(String redirect)
    {
    }
    
    /*
     * (non-Javadoc)
     * @see leon.view.graphic.LyGraphicalArea#setLayoutInfo(java.lang.String)
     */
    public void setLayoutInfo(String layoutInfo)
    {
    }
} // end class LySwingArea
