/*
 * LySwingBoolean.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Boolean object (as described in the dtd) for swing viewer.
 *
 * <p>An input field of boolean type allows to display a graphical component with a three status
 * checkbox: not specified (empty), true (checked with a green sign), false (with a red cross). It
 * is possible to specify a label for each of these statuss and to make a negation of the boolean
 * value.</p>
 *
 * @version $Id: LySwingBoolean.java 29256 2011-01-19 08:45:17Z bangiole $
 * @author  Lyria
 * @see     LyBoolean
 */
public class LySwingBoolean extends LySwingFieldItem
	implements LyGraphicalBoolean,
		LySetEditableInterface,
		LySetEnabledInterface,
		LySetOptionalInterface,
		LyGetValueInterface,
		ActionListener
{
	public static final String __VERSION =
		"$Id: LySwingBoolean.java 29256 2011-01-19 08:45:17Z bangiole $";

	/** The view independent boolean description. */
	protected LyBoolean		   _booleanView;

	/** The swing boolean component. */
	protected Component		   _comp;
	
	/** The editable status of this component */
	protected boolean		   _isEditable;
	
	/** The enable status of this component */
	protected boolean		   _isEnable;

	/**
	 * Class constructor of a Boolean object for the Swing target.
	 *
	 * @param booleanViewComp a BOOLEAN component defined for every type of view targets. This
	 *                        boolean is initialized specificaly for the Swing view.
	 */
	public LySwingBoolean(LyBoolean booleanViewComp)
	{
		super(booleanViewComp);
		_booleanView = booleanViewComp;

		initialize();
		setContent();
	}

	/**
	 * Initializes the Boolean object by creating the associated Swing boolean component.
	 */
	public void initialize()
	{
		_comp = createBooleanComponent();
		_isEditable = true;
		_isEnable = true;
	}

	/**
	 * Returns the boolean manager.
	 *
	 * @return the boolean manager
	 */
	public LyAbstractBooleanPanelManager getBooleanManager()
	{
		return (LyAbstractBooleanPanelManager)((LyJBooleanPanel)_comp).getManager();
	}

	/**
	 * Returns a new Swing boolean component.
	 *
	 * @return a new boolean component
	 */
	public Component createBooleanComponent()
	{
		LyJBooleanPanel panel = new LyJBooleanPanel();

		panel.setOpaque(false);

		return panel;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetOptionalInterface#setOptional(boolean)
	 */
	public void setOptional(boolean optional)
	{
		getBooleanManager().setOptional(optional);
	}

	/**
	 * Returns the boolean value. The value may be Boolean.TRUE, Boolean.FALSE or <code>null</code>
	 * if this component is empty.
	 *
	 * @return the boolean value
	 * @see    #setBooleanValue(Boolean)
	 */
	public Boolean getBooleanValue()
	{
		return getBooleanManager().getValue();
	}

	/**
	 * Sets the new boolean value of this component. The value may be Boolean.TURE, Boolean.FALSE or
	 * <code>null</code> if the value is empty.
	 *
	 * @param value the new boolean value
	 * @see   #getBooleanValue()
	 */
	public void setBooleanValue(Boolean value)
	{
		getBooleanManager().setValue(value);
	}

	/**
	 * Sets the images of this component.
	 *
	 * @param back   the background image
	 * @param hoover the hoover image
	 * @param empty  the image used when BOOLEAN_EMPTY status is selected
	 * @param falseV the image used when BOOLEAN_FALSE status is selected
	 * @param trueV  the image used when BOOLEAN_TRUE status is selected
	 * @see   LyAbstractBooleanPanelManager#setImages(Image, Image, Image, Image, Image)
	 */
	public void setImages(Image back, Image hoover, Image empty, Image falseV, Image trueV)
	{
		getBooleanManager().setImages(back, hoover, empty, falseV, trueV);
	}

	/**
	 * Sets the labels of this component.
	 *
	 * @param labelEmpty     label displayed when the selection is empty
	 * @param labelChecked   label displayed when the selection is checked
	 * @param labelUnchecked label displayed when the selection is unchecked
	 * @see   LyAbstractBooleanPanelManager#setLabels(String, String, String)
	 */
	public void setLabels(String labelEmpty, String labelChecked, String labelUnchecked)
	{
		getBooleanManager().setLabels(labelEmpty, labelChecked, labelUnchecked);
	}

	/**
	 * Sets the content of this component. The images are retrieved as well as the labels and the
	 * font this component uses then these resources are applied on this boolean component.
	 */
	public void setContent()
	{
		Image  back = this.getImage("LY_BOOLEAN_BACKGROUND");
		Image  hoover = this.getImage("LY_BOOLEAN_BACKGROUND_HOOVER");
		Image  empty = this.getImage("LY_BOOLEAN_EMPTY");
		Image  trueV = this.getImage("LY_BOOLEAN_CHECKED");
		Image  falseV = this.getImage("LY_BOOLEAN_UNCHECKED");
		String labelEmpty = translate(_booleanView.getLabelEmpty());
		String labelChecked = translate(_booleanView.getLabelChecked());
		String labelUnchecked = translate(_booleanView.getLabelUnchecked());
		Font   font = this.getFont("LY_DEFAULT_FONT");

		// Set the images used by the canvas
		setImages(back, hoover, empty, falseV, trueV);

		// do not invert label
		// For examples: compta/fournisseur if we have a not on the VAT field
		// (so we fill 'no VAT: Yes/No') we expect the check mark to stay yes
		setLabels(labelEmpty, labelChecked, labelUnchecked);

		// Set the font
		_comp.setFont(font);

		LyFormField parentFormField = (LyFormField)_booleanView.getParent();

		setOptional(parentFormField.isOptional());

		// Register this class as a selection listener to handle value changed
		getBooleanManager().addActionListener(this);

		// Initialize the value of the component
		// do not handle 'not' here, it is handle in the setValue/getValue method
		switch (_booleanView.getStatus())
		{
		case CHECKED:
			setValue(Boolean.TRUE);
			break;
		case EMPTY:
			setValue(null);
			break;
		case UNCHECKED:
			setValue(Boolean.FALSE);
			break;
		}

		LySwingFormKeyHandler keyHandler = new LySwingFormKeyHandler(getActionTransmitter());

		keyHandler.setNotifyValidate(false); // already default in swing
		_comp.addKeyListener(keyHandler);
	} // end method setContent

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetEditableInterface#setEditable(boolean)
	 */
	public void setEditable(boolean status)
	{
		_isEditable = status;
		_comp.setEnabled(_isEditable && _isEnable);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetEnabledInterface#setEnabled(boolean)
	 */
	public void setEnabled(boolean status)
	{
		_isEnable = status;
		_comp.setEnabled(_isEditable && _isEnable);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetstatusInterface#setstatus(short)
	 */
	public void setStatus(LyBoolean.Status status)
	{
		if (status == LyBoolean.Status.EMPTY)
			setValue(null);
		else if (status == LyBoolean.Status.CHECKED)
			setValue(Boolean.TRUE);
		else if (status == LyBoolean.Status.UNCHECKED)
			setValue(Boolean.FALSE);
	}

	/**
	 * Computes the value of this boolean component. This is done by retreiving the boolean
	 * component value and reversing the value if the property {@link LyBoolean#isNot()} is set. If
	 * there is no value, <code>null</code> is returned.
	 *
	 * @return the current value. It may be Boolean.TRUE, Boolean.FALSE or <code>null</code> if
	 *         there is no value for this component.
	 */
	public Boolean computeBooleanValue()
	{
		// We get the value from the graphical component
		Boolean value = getBooleanValue();

		// check if we must invert the value
		boolean not = _booleanView.isNot();

		if (value == null) // no value, no invert
			return null;
		else if (value.booleanValue()) // its true, so if we must invert, we return false
			return (not ? Boolean.FALSE : value);
		else
			// it's false, so if we invert we return true
			return (not ? Boolean.TRUE : value);
	}

	/**
	 * Returns the boolean internal status as a string.
	 *
	 * @see #setValue(Boolean)
	 * @see LyGetValueInterface#getValue()
	 */
	public String getValue()
	{
		return _booleanView.getStatus().toString();
	}

	/**
	 * Sets the value of this swing boolean component. The values are:
	 *
	 * <ul>
	 *   <li>{@link Boolean#TRUE}</li>
	 *   <li>{@link Boolean#FALSE}</li>
	 *   <li><code>null</code> if the component is empty</li>
	 * </ul>
	 *
	 * @param value the new boolean value
	 * @see   #getValue()
	 */
	public void setValue(Boolean value)
	{
		// We get a value, we may need to invert it
		boolean not = _booleanView.isNot();

		Boolean           result = null;
        LyBoolean.Status   status = LyBoolean.Status.EMPTY;

		if (value == null)
			setBooleanValue(null);
		else if (value.booleanValue())
		{
			// The value is inverted (if needed) for the graphical component
			result = (not ? Boolean.FALSE : value);
			// but not for the view component (it's the real value)
			status = LyBoolean.Status.CHECKED;
		}
		else
		{
			// the value is inverted (if needed) for the graphical component
			result = (not ? Boolean.TRUE : value);
			// but not for the view component (it's the real value)
			status = LyBoolean.Status.UNCHECKED;
		}

		// Update both graphical and view component
		_booleanView.setStatus(status);
		setBooleanValue(result);
	} // end method setValue

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent e)
	{
		Boolean		  val = computeBooleanValue();

		String		  value = ((val == null) ? null : val.toString());

		LyActionEvent event = new LyActionEvent(this, null, value,
				LyActionDemand.TYPE_VALUE_CHANGED);

		setValue(val);
		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(event);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _comp;
	}

	/**
	 * Sets the background color of this component.
	 *
	 * @param color the background color
	 */
	protected void setBackground(Color color)
	{
		_comp.setBackground(color);
	}

} // end class LySwingBoolean
