/*
 * LySwingCaption.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;

import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.graphic.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Label object (as described in the dtd) for swing viewer.
 *
 * <p>An input field of label type allows to display a simple, non-editable label inside a form. It
 * is used in the consultation forms to show different types of fields (date, number, choice,...).
 * It can be made up of one or several lines. It does not comprise any element.</p>
 *
 * @version $Id: LySwingCaption.java 34954 2012-03-22 13:54:06Z jlsaouli $
 * @author  Lyria
 * @see     LyLabel
 */
public class LySwingCaption extends LySwingObject
	implements LySwingComponent,
		LyGraphicalCaption,
		ItemListener,
		MouseListener
{
	public static final String  __VERSION =
		"$Id: LySwingCaption.java 34954 2012-03-22 13:54:06Z jlsaouli $";

	private static final String CHECKBOX_SOURCE_ID = "_checkbox_";

	private static final String CAPTION_SOURCE_ID = "_caption_";

	/**
	 * Number of pixels of the border.
	 */
	public int				    LABEL_BORDER = 3;

	/** The view independent caption description. */
	protected LyCaption		    _caption;

	/** The swing component used to display this caption. */
	protected Component		    _labelComp;
	
	/** The swing panel used to display this caption and its tip. */
	protected Container			_tippedCaptionPanel;

	/**
	 * Class constructor of a caption object for the Swing target.
	 *
	 * @param caption a CAPTION component defined for every type of view targets. The caption border
	 *                is initialized if its value is negative, then the content of the caption is
	 *                filled.
	 */
	public LySwingCaption(LyCaption caption)
	{
		super(caption);

		_caption = caption;
		_tippedCaptionPanel = null;

		setContent();
	}

	/**
	 * Creates and initializes a caption. The constraints are computed and applied on the new Swing
	 * component for this caption. If an help tip is defined, it is used as a tooltip with the
	 * associated mouse cursor.
	 *
	 * @param  label the text displayed by the caption
	 * @return a new caption with the specified label
	 */
	public Component createCaption(String label)
	{
		// By default, captions are right aligned
		LyConstraint constraint = _caption.getConstraint();
		short		 align = LyJMultiLineLabel.RIGHT;

		if (constraint == null)
			align = LyJMultiLineLabel.RIGHT;
		else if (constraint.getHAlign() == LyConstraintBase.HAlign.CENTER)
			align = LyJMultiLineLabel.CENTER;
		else if (constraint.getHAlign() == LyConstraintBase.HAlign.LEFT)
			align = LyJMultiLineLabel.LEFT;

		LyJMultiLineLabel caption = new LyJMultiLineLabel(label, -1, LABEL_BORDER, LABEL_BORDER,
				align);
		LyEnvironment     env = _caption.getEnvironment();
		boolean			  displayTipIcon = env.testResource("LY_DISPLAY_TIP_ICON");
		LyFormField		  formField = (LyFormField)_caption.getParent();
		LyTip			  tip = formField.getTip();

		if (!displayTipIcon && (tip != null))
		{
			caption.setToolTipText(translate(tip.getLabel()));
			caption.setCursor(new Cursor(Cursor.HAND_CURSOR));
		}

		caption.setOpaque(false);

		return caption;
	} // end method createCaption

	/**
	 * Creates and configures a captionned check box.
	 *
	 * @param  label   the text displayed around the check box
	 * @param  checked indicates if check box is checked or not.
	 * @return a new labeled checkbox
	 */
	public Component createCheckCaption(String label, boolean checked)
	{
		JCheckBox	caption = new JCheckBox(label, checked);
		LyFormField formField = (LyFormField)_caption.getParent();

		caption.setOpaque(false);
		caption.setSelected(formField.isChecked());
		
		return caption;
	}

	/**
	 * Creates and configures a new panel used to display a tooltip.
	 *
	 * @return a panel with the layout tuned to display a tip
	 */
	public Container createPanelForTippedCaption()
	{
		// By default, captions are right aligned
		LyConstraint constraint = _caption.getConstraint();
		short		 align = FlowLayout.RIGHT;

		if (constraint == null)
			align = FlowLayout.RIGHT;
		else if (constraint.getHAlign() == LyConstraintBase.HAlign.CENTER)
			align = FlowLayout.CENTER;
		else if (constraint.getHAlign() == LyConstraintBase.HAlign.LEFT)
			align = FlowLayout.LEFT;

		// No gap : FFT 786
		JPanel result = new JPanel(new FlowLayout(align, 0, 0));

		result.setOpaque(false);

		return result;
	}

	/**
	 * Returns a new blank <code>JPanel</code> for spaces on tooltips.
	 *
	 * @return a blank panel
	 * @see    LySwingCaption#getComponent()
	 */
	public Component getTipCaptionBlank()
	{
		JPanel result = new JPanel(new FlowLayout(FlowLayout.LEFT, 2, 2));

		result.setOpaque(false);

		return result;
	}

	/* (non-Javadoc)
	 * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
	 */
	public void setLabel(String label)
	{
		if (_labelComp != null)
		{
			((LyJMultiLineLabel)_labelComp).setLabel(label);

			// We resize the component all the time
			if (_labelComp.getParent() != null)
			{
				_labelComp.invalidate();
				((JComponent)_labelComp.getParent()).revalidate();
			}
		}
	}
	
	/* (non-Javadoc)
	 * @see leon.view.setinterface.LySetEnabledInterface#setEnabled(boolean)
	 */
	public void setEnabled(boolean status)
	{
		LyFormField formField = (LyFormField)_caption.getParent();

		if (formField.isCheckBox())
			_labelComp.setEnabled(status);
	}

	/**
	 * Sets the content of this caption. If this caption has a label, it is used. Its specialized
	 * Swing component is created (as a default caption or as a checkbox if the property is set)
	 */
	public void setContent()
	{
	    LyFormField formField = (LyFormField)_caption.getParent();
		String      label = _caption.getLabel();

		if ((label != null) && (label.length() > 0))
		{
			label = translate(label);
			if (label != null)
			{
				label = label.trim();
				if (label.length() > 0)
				{			
				    if (!formField.isOptional())
				        label += getString("LY_CAPTION_ASTERISK", "");
                    if (_caption.isColon())
                        label += getString("LY_CAPTION_COLON", "");
				}
			}
			else
				label = "";
		}
		else
			label = "";

		boolean     checkbox = formField.isCheckBox();

		if (!checkbox)
			_labelComp = createCaption(label);
		else
		{
			_labelComp = createCheckCaption(label, formField.isEnabled());
			if (_labelComp instanceof ItemSelectable)
				((ItemSelectable)_labelComp).addItemListener(this);
		}

		_labelComp.addMouseListener(this);
		
		if (_labelComp instanceof LyJMultiLineLabel)
		{
			Font f = getFont("LY_DEFAULT_BOLD_FONT");

			((LyJMultiLineLabel)_labelComp).setBigFont(f);
		}
	} // end method setContent

	/**
	 * Returns the label of this caption.
	 *
	 * @return the label of this caption
	 */
	public Component getCaptionLabel()
	{
		return _labelComp;
	}

	/* (non-Javadoc)
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		// Set the help button
		LyFormField   formField = (LyFormField)_caption.getParent();
		LyTip		  tip = formField.getTip();
		LyEnvironment env = formField.getEnvironment();
		boolean		  displayTipIcon = env.testResource("LY_DISPLAY_TIP_ICON");

		if ((tip != null) && displayTipIcon)
		{
			if (_tippedCaptionPanel != null)
				return _tippedCaptionPanel;
			
			// We have an help on this item
			_tippedCaptionPanel = createPanelForTippedCaption();
			_tippedCaptionPanel.setBackground(getColor("LY_DEFAULT_COLOR"));

			if (tip != null)
			{
				LySwingComponent internal = (LySwingComponent)tip.getInternalView();

				_tippedCaptionPanel.add(internal.getComponent());
				Component blank = getTipCaptionBlank(); // XXX leave a blank : FFT 786

				blank.setBackground(getColor("LY_DEFAULT_COLOR"));
				_tippedCaptionPanel.add(blank);
			}

			_tippedCaptionPanel.add(_labelComp);
			
			return _tippedCaptionPanel;
		}

		return _labelComp;
	} // end method getComponent

	/* (non-Javadoc)
	 * @see java.awt.event.ItemListener#itemStateChanged(java.awt.event.ItemEvent)
	 */
	public void itemStateChanged(ItemEvent e)
	{
		if (!_caption.isEnabled())
			return;
		
		LyFormField		 formField = (LyFormField)_caption.getParent();
		LySwingComponent internal = (LySwingComponent)formField.getInternalView();
		JCheckBox		 checkbox = (JCheckBox)e.getSource();
		
		formField.setChecked(checkbox.isSelected());
		
		if (internal instanceof LyEventTransmitter)
		{
			LyActionTransmitter actionTransmitter = ((LyEventTransmitter)internal)
					.getActionTransmitter();
			LyActionEvent	    actionEvent = new LyActionEvent(this, null, null,
					LyActionDemand.TYPE_VALUE_CHANGED);

			actionEvent.setSourceId(CHECKBOX_SOURCE_ID + formField.getId());
			actionTransmitter.processActionEvent(actionEvent);
		}
	}
	
	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	public void mouseClicked(MouseEvent e)
	{
		if (!_caption.isEnabled())
			return;
		
		LyFormField		 formField = (LyFormField)_caption.getParent();
		LySwingComponent internal = (LySwingComponent)formField.getInternalView();

		// Right click on a caption send a menu demand
		if ((internal instanceof LyEventTransmitter)
				&& ((e.getModifiers() & InputEvent.BUTTON3_MASK) > 0))
		{
			LyActionTransmitter	actionTransmitter = ((LyEventTransmitter)internal)
					.getActionTransmitter();
			LyActionEvent		actionEvent = new LyActionEvent(this, null, null,
					LyMenuDemand.TYPE_POPUP);
			Point				componentPosition = _labelComp.getLocationOnScreen();

			actionEvent.setParameter(new int[] {e.getPoint().x + componentPosition.x,
					e.getPoint().y + componentPosition.y});
			actionEvent.setSourceId(CAPTION_SOURCE_ID + formField.getId());
			actionTransmitter.processActionEvent(actionEvent);
		}
	}
	
	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
	 */
	public void mousePressed(MouseEvent e)
	{
	}
	
	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	public void mouseReleased(MouseEvent e)
	{
	}
	
	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
	 */
	public void mouseEntered(MouseEvent e)
	{
	}
	
	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
	 */
	public void mouseExited(MouseEvent e)
	{
	}
	
} // end class LySwingCaption
