/*
 * LySwingCell.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;

import leon.misc.*;
import leon.view.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;
import leon.view.win.swing.table.*;


/**
 * This class implements the Cell object (as described in the dtd) for swing viewer.
 *
 * @version $Id: LySwingCell.java 36459 2012-08-09 13:52:03Z jlsaouli $
 * @author  Lyria
 * @see     LyCell
 */
public class LySwingCell extends LySwingObject
	implements LySetLabelInterface,
		LySetImageInterface,
		LyGetLabelInterface,
		LyGetIdInterface,
		LyGraphicalCell
{
	public static final String __VERSION =
		"$Id: LySwingCell.java 36459 2012-08-09 13:52:03Z jlsaouli $";

	/** The label displayed by this cell. */
	protected String		   _label;

	/** The image displayed by this cell. */
	protected Image			   _image;
	
	/** Image for editable cell. */
	protected Image			   _editionImage;

	/** The background color of this cell. */
	protected Color			   _color;
	
	/** The foreground color of this cell. */
	protected Color			   _fontColor;

	/** The font of this cell. */
	protected Font			   _font;

	/** The tooltip of this cell. */
	protected String		   _tooltip;

	/** The view independent cell description. */
	protected LyCell		   _cell;

	/** internal use to store cell component cache **/
	protected Component		   _component;
	
	/**
	 * Class constructor of a cell object for the Swing target.
	 *
	 * @param cell a CELL component defined for every type of view targets. This cell is initialized
	 *             specificaly for the Swing view.
	 */
	public LySwingCell(LyCell cell)
	{
		super(cell);
		_cell = cell;
		initialize();
		setContent();
	}

	/**
	 * Initializes this object. Sets its images, fonts, tooltip and color.
	 */
	public void initialize()
	{
		if (_cell.getImage() != null)
		{
			Image img = getImage(_cell.getImage());

			if (img != null)
				_image = img;
		}
		else
			_image = null;

		if (_cell.getFont() != null && !_cell.isDefaultFont())
		{
			Font font = getFont(_cell.getFont());

			if (font != null)
				_font = font;
		}
		else
			_font = null;

		setColor(_cell.getColor());
		setTooltip(_cell.getTooltip());
		_label = _cell.getLabel();
		_component = null;
	}

	/**
	 * Sets the content of this object. The cell is empty by default, so nothing is done here.
	 */
	public void setContent()
	{
	}

	/* (non-Javadoc)
	 * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
	 */
	public void setLabel(String label)
	{
		LyTable table = getParentTable();
		//	clear cache
		_component = null;
		
		if (table != null)
		{
			LySwingTable swingTable = (LySwingTable)table.getInternalView();
			LyRow		 row = getParentRow();
			int			 oldHeight = swingTable.computeRowHeight(row);

			_label = label;

			LyRows		 rows = table.getRows();
			int			 idx = rows.getRowIndex(row);
			int			 i , n = row.getComponentCount();
			boolean		 columnUpdated = false;
			
			for (i = 0; i < n; i++)
			{
				if (row.getComponent(i) == _cell)
				{
					swingTable.ensureColumnsWidth(idx, i);
					columnUpdated = true;
				}
			}
			
			if (!columnUpdated)
				swingTable.ensureColumnsWidth(idx);
			
			int		newHeight = swingTable.computeRowHeight(row);
			
			if (oldHeight != newHeight)
			{
				swingTable.checkComputeHeight();
				swingTable.computeRowHeights();
				swingTable.cacheHeight();
				
				LyJTableStatusColumn	statusColumn = swingTable.getStatusColumn();
				
				if (statusColumn != null)
					statusColumn.doLayout();
			}
		}
		else
			_label = label;

		updateTable(false);
	}

	/**
	 * The objectId of this object cannot change. The specified objectId parameter is not used.
	 */
	public void setObjectId(String objectId)
	{
	}

	/* (non-Javadoc)
	 * @see leon.view.getinterface.LyGetLabelInterface#getLabel()
	 */
	public String getLabel()
	{
		return _label;
	}
	
	/**
	 * Returns the width of this cell.
	 * 
	 * @param fm the font metrics used to compute text width.
	 * @return the preferred width of the cell.
	 */
	public int getWidth(FontMetrics fm)
	{
		int		width = 0;

		if (_cell.getCellCount() <= 0)
		{
			if (getFont() != null)
			{
				LyTable table = getParentTable();

				if (table != null)
				{
					LySwingTable	  swingTable = (LySwingTable)table.getInternalView();
					LyJTableComponent comp = swingTable.getTableComponent();

					fm = comp.getFontMetrics(getFont());
				}
			}
			
			String	label = getLabel();
			Image	image = getImage();

			if (label != null)
			{
				if (label.indexOf('\n') >= 0)
				{
					String[]	list = LyString.stringToList(label, '\n');
					
					if ((list != null) && (list.length > 1))
					{
						for (int i = 0; i < list.length; i++)
						{
							width = Math.max(fm.stringWidth(list[i]) + 10, width);
						}
					}
				}

				if (width <= 0)
					width = fm.stringWidth(label) + 10;
			}

			if (image != null)
				width += image.getWidth(null) + 2;
			
			LyRow		 parentRow = getParentRow();

			if (parentRow != null)
			{
				// Compute the offset for first column only
				LyTable table = getParentTable();

				if (table != null)
				{
					LySwingTable	  swingTable = (LySwingTable)table.getInternalView();
					int				  offset = 0;

					if (swingTable.isHierarchical() && (parentRow.getComponent(0) == _cell))
					{
						LySwingRow	swingRow = (LySwingRow)parentRow.getInternalView();
						
						offset = swingRow.getDepth();

						if (offset > 0)
							width += LyJCellRenderer.HIERARCHY_OFFSET_WIDTH * offset;
					}
				}
			}
		}
		else
		{
			int			i, n = _cell.getCellCount();
			LyCell		subCell;
			LySwingCell	swingSubCell;
			
			for (i = 0; i < n; i++)
			{
				subCell = _cell.getCell(i);
				swingSubCell = (LySwingCell)subCell.getInternalView();
				width = Math.max(width, swingSubCell.getWidth(fm));
			}
		}
		
		return width;
	}

	/* (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalCell#setTooltip(java.lang.String)
	 */
	public void setTooltip(String tooltip)
	{
		_tooltip = _cell.getEnvironment().translate(tooltip);
	}

	/**
	 * Returns the tooltip of this cell.
	 *
	 * @return the tooltip associated with this cell
	 * @see    #setTooltip(String)
	 */
	public String getTooltip()
	{
		return _tooltip;
	}

	/* (non-Javadoc)
	 * @see leon.view.setinterface.LySetImageInterface#setImage(java.lang.String)
	 */
	public void setImage(String image)
	{
		if (image != null)
		{
			Image img = getImage(image);

			if (img != null)
				_image = img;
		}
		else
			_image = null;

		updateTable(true);
	}

	/**
	 * Returns the image of this cell.
	 *
	 * @return the image displayed by this cell
	 * @see    #setImage(String)
	 */
	public Image getImage()
	{
		return _image;
	}

	/**
	 * Returns the image for an editable cell.
	 *
	 * @return the edition image
	 */
	public Image getEditionImage()
	{
		if (_editionImage == null)
			_editionImage = getImage("LY_EDITABLE_CELL_MARK");

		return _editionImage;
	}

	/* (non-Javadoc)
	 * @see leon.view.getinterface.LyGetIdInterface#getId()
	 */
	public String getId()
	{
		return _cell.getObjectId();
	}

	/**
	 * Gets the background color of this cell.
	 *
	 * @return the color used for the background of the cell
	 * @see    #setColor(String)
	 */
	public Color getColor()
	{
		if (_color != null)
			return _color;

		LyRow	row = getParentRow();
		
		if (row == null)
			return null;
		
		return ((LySwingRow)row.getInternalView()).getColor();
	}

	/**
	 * Gets the background color of this cell.
	 *
	 * @return the color used for the background of the cell
	 * @see    #setColor(String)
	 */
	public Color getFontColor()
	{
		if ((_fontColor == null)
				&& (_cell.getFontColor() != null))
			_fontColor = getColor(_cell.getFontColor());

		return _fontColor;
	}
	
	/* (non-Javadoc)
	 * @see leon.view.setinterface.LySetColorInterface#setColor(java.lang.String)
	 */
	public void setColor(String color)
	{
		if (color != null)
		{
			Color col = getColor(color);

			if (col != null)
				_color = col;
		}
		else
			_color = null;

		updateTable(false);
	}

	/*
	 * (non-Javadoc).
	 * @see leon.view.graphic.LyGraphicalCell#setBottomBorderColor(java.lang.String)
	 */
	public void setBottomBorderColor(String bottomBorderColor)
	{
		updateTable(false);
	}

	/*
	 * (non-Javadoc).
	 * @see leon.view.graphic.LyGraphicalCell#setLeftBorderColor(java.lang.String)
	 */
	public void setLeftBorderColor(String leftBorderColor)
	{
		updateTable(false);
	}

	/*
	 * (non-Javadoc).
	 * @see leon.view.graphic.LyGraphicalCell#setRightBorderColor(java.lang.String)
	 */
	public void setRightBorderColor(String rightBorderColor)
	{
		updateTable(false);
	}

	/*
	 * (non-Javadoc).
	 * @see leon.view.graphic.LyGraphicalCell#setTopBorderColor(java.lang.String)
	 */
	public void setTopBorderColor(String topBorderColor)
	{
		updateTable(false);
	}

	/*
	 * (non-Javadoc).
	 * @see leon.view.graphic.LyGraphicalCell#setBottomLeftImage(java.lang.String)
	 */
	public void setBottomLeftImage(String bottomLeftImage)
	{
		updateTable(false);
	}

	/*
	 * (non-Javadoc).
	 * @see leon.view.graphic.LyGraphicalCell#setBottomRightImage(java.lang.String)
	 */
	public void setBottomRightImage(String bottomRightImage)
	{
		updateTable(false);
	}

	/*
	 * (non-Javadoc).
	 * @see leon.view.graphic.LyGraphicalCell#setTopLeftImage(java.lang.String)
	 */
	public void setTopLeftImage(String topLeftImage)
	{
		updateTable(false);
	}

	/*
	 * (non-Javadoc).
	 * @see leon.view.graphic.LyGraphicalCell#setTopRightImage(java.lang.String)
	 */
	public void setTopRightImage(String topRightImage)
	{
		updateTable(false);
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalCell#setEditable(boolean)
	 */
	public void setEditable(boolean editable)
	{
		updateTable(false);
	}

	/*
	 * (non-Javadoc)
	 * @see leon.view.graphic.LyGraphicalCell#setSelected(boolean)
	 */
	public void setSelected(boolean selected)
	{
		updateTable(false);
	}
	
	/* (non-Javadoc)
	 * @see leon.view.setinterface.LySetFontInterface#setFont(java.lang.String)
	 */
	public void setFont(String font)
	{
		if (font != null)
		{
			Font police = getFont(font);

			if (police != null)
				_font = police;
		}
		else
			_font = null;

		updateTable(true);
	}

	/**
	 * Gets the font used for this cell.
	 *
	 * @return the cell font
	 * @see    #setFont(String)
	 */
	public Font getFont()
	{
		return _font;
	}

	/**
	 * This cell cannot have sub-cells, thus this method does not do anything.
	 *
	 * @see #removeCell(LyCell)
	 * @see LyGraphicalCell#addCell(LyCell, int)
	 */
	public void addCell(LyCell cell, int pos)
	{
		updateTable(true);
	}

	/**
	 * This cell cannot have sub-cells, thus this method does not do anything.
	 *
	 * @see #addCell(LyCell, int)
	 * @see LyGraphicalCell#removeCell(LyCell)
	 */
	public void removeCell(LyCell cell)
	{
	}

	/**
	 * Returns the parent row component.
	 *
	 * @return the parent row component
	 */
	public LyRow getParentRow()
	{
		LyComponent		parent = _cell.getParent();

		if (parent == null)
			return null;
		
		if (parent instanceof LyRow)
			return (LyRow)parent;
		LyCell	    cell = (LyCell)parent;
		LySwingCell internal = (LySwingCell)cell.getInternalView();

		return internal.getParentRow();
	}

	/**
	 * Returns the parent table component.
	 *
	 * @return the parent table component
	 */
	public LyTable getParentTable()
	{
		LyRow row = getParentRow();

		if (row == null)
			return null;

		LyRows rows = (LyRows)row.getParent();

		return ((rows == null) ? null : rows.getParentTable());
	}

	/**
	 * Updates the table component.
	 *
	 * @param refresh compute the table heights again
	 */
	protected void updateTable(boolean refresh)
	{
		LyTable table = getParentTable();

		if (table != null)
		{
			//			 clear cache
			_component = null;
			
			LySwingTable	  swingTable = (LySwingTable)table.getInternalView();
			LyJTableComponent comp = swingTable.getTableComponent();

			if (comp != null)
			{
				if (refresh
				    && !swingTable.isIdle())
				{
					swingTable.checkComputeHeight();
					swingTable.computeRowHeights();
					
					LyJTableStatusColumn	statusColumn = swingTable.getStatusColumn();
					
					if (statusColumn != null)
						statusColumn.doLayout();
				}
				else
					comp.repaint();
				
				swingTable.checkEditionStatus(null);
			}
		}
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _component;
	}
	
	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public void setComponent(Component c)
	{
		_component = c;
	}
} // end class LySwingCell
