/*
 * LySwingChoice.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import leon.misc.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.getinterface.*;
import leon.view.graphic.*;
import leon.view.setinterface.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Choice object (as described in the dtd) for Swing viewer.
 *
 * <p>An input field of Choice type allows to display a graphical component containing a set of
 * check boxes. It allows to choose one or more values among a discrete list of values. It is
 * possible to specify whether the choice is multiple or not and whether the boxes should be
 * arranged horizontally or vertically.</p>
 *
 * @version $Id: LySwingChoice.java 28745 2010-12-01 08:13:48Z jlsaouli $
 * @author  Lyria
 * @see     LyChoice
 */
public class LySwingChoice extends LySwingFieldItem
	implements LySwingComponent,
		LyGraphicalChoice,
		ItemListener,
		LySetEnabledInterface,
		LySetEditableInterface,
		LyGetValueInterface,
		LySetFocusInterface
{
	public static final String				 __VERSION =
		"$Id: LySwingChoice.java 28745 2010-12-01 08:13:48Z jlsaouli $";

	/** The view independent choices description. */
	protected LyChoice						 _choice;

	/** A group of checkboxes. */
	protected Object						 _checkboxGroup;

	/** The list of checkboxes in this choice component. */
	protected Vector<Component>				 _checkboxes;

	/** Association of checkboxex and their options. */
	protected Hashtable<Component, LyOption> _checkboxAssoc;

	/** The swing container for all the checkboxes of this choice object. */
	protected Container						 _panel;

	/** Last editable status of this choice sub-components. */
	protected boolean						 _last_editable;

	/** Last enabled status of this choice sub-components. */
	protected boolean						 _last_enabled;

	/**
	 * Class constructor of a choice object for the Swing target.
	 *
	 * @param choice a CHOICE component defined for every type of view targets. This choice
	 *               component is initialized specificaly for the Swing view.
	 */
	public LySwingChoice(LyChoice choice)
	{
		super(choice);
		_choice = choice;

		initialize();
		setContent();
	}

	/**
	 * Creates and configures the object that manage a group of checkboxes (logically).
	 *
	 * @return a new check box group
	 */
	public Object createCheckBoxGroup()
	{
		return new ButtonGroup();
	}

	/**
	 * Adds a checkbox to the checkbox group.
	 *
	 * @param box the checkbox to add. It may be a {@link JCheckBox} if the choices are multiple,
	 *            else a {@link JRadioButton}.
	 */
	public void addCheckBoxToGroup(Component box)
	{
		// FFT00999 If a choice is not multiple, it must be a JRadioButton
		if (_choice.isMultiple())
			((ButtonGroup)getCheckboxGroup()).add((JCheckBox)box);
		else
			((ButtonGroup)getCheckboxGroup()).add((JRadioButton)box);
	}

	/**
	 * Creates and configures the panel that contains all the choices.
	 *
	 * @param  fill a short that specifies the way to place the choices
	 * @return a panel initialized to contain all the choices of this component
	 */
	public Container createChoicePanel(LyChoiceBase.Fill fill)
	{
		_LyJChoicePanel panel = new _LyJChoicePanel();

		/* Set the background color same as its parent */
		panel.setBackground(null);
		panel.setOpaque(false);

		panel.setMaxSpan(_choice.getMaxSpan());
		panel.setFill(fill);

		return panel;
	}

	/**
	 * Checks or unchecks a checkbox or a group of checkboxes, depending on this choice object.
	 *
	 * @param box       the checkbox to modify
	 * @param isChecked the state to set for the specified checkbox
	 */
	public void setChecked(Component box, boolean isChecked)
	{
		if (!_choice.isMultiple())
		{
			// FFT00999
			((ButtonGroup)getCheckboxGroup()).setSelected(((JRadioButton)box).getModel(),
				isChecked);

			box.repaint();
		}
		else
			((JCheckBox)box).setSelected(isChecked);
	}

	/**
	 * Returns the state of a given box.
	 *
	 * @param  box the box to test
	 * @return true if the box is checked, false otherwise
	 */
	public boolean isChecked(Component box)
	{
		// FFT00999
		return ((JToggleButton)box).isSelected();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see java.awt.event.ItemListener#itemStateChanged(java.awt.event.ItemEvent)
	 */
	public void itemStateChanged(ItemEvent e)
	{
		String val = getValue();

		if ((e.getItem() instanceof JRadioButton) && (val == null))
			// Ignore unselect on radioButton
			return;

		LyActionEvent event = new LyActionEvent(this, null, val, LyActionDemand.TYPE_VALUE_CHANGED);

		setValue(val);
		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(event);
	}

	/**
	 * Adds a choice element to the panel. A handler is associated to the specified choice for it to
	 * validate or close the form respectively with the ENTER and ESCAPE keys.
	 *
	 * @param choice the choice to add
	 */
	public void addChoiceToPanel(Component choice)
	{
		((_LyJChoicePanel)_panel).addChoice(choice);

		// This handler is used to validate the form using ENTER key or close using ESCAPE key
		LySwingFormKeyHandler keyHandler = new LySwingFormKeyHandler(getActionTransmitter());

		keyHandler.setNotifyValidate(false); // already default in swing
		choice.addKeyListener(keyHandler);
	}

	/**
	 * Initializes this component and creates a new checkbox group if it is a single choice object.
	 */
	public void initialize()
	{
		_last_editable = true;
		_last_enabled = true;

		if (!_choice.isMultiple())
			_checkboxGroup = createCheckBoxGroup();
	}

	/**
	 * Sets the content of this choice component. Every checkboxes it contains are added to the
	 * checkbox list and initialized giving their properties, then they are added to this
	 * component's panel.
	 */
	public void setContent()
	{
		LyOption	  option;
		LySwingOption internal;
		int			  i, n = _choice.getOptionCount();
		Component     box;

		for (i = 0; i < n; i++)
		{
			option = _choice.getOption(i);
			internal = (LySwingOption)option.getInternalView();
			internal.setParent(this);

			box = internal.getComponent();

			// set checkboxgroup to 'link' radio buttons
			if (!_choice.isMultiple())
			{
				addCheckBoxToGroup(box);

				if (option.isChecked())
					setChecked(box, true);
			}

			((ItemSelectable)box).addItemListener(this);

			addCheckBox(box, option);
		}

		// ## CREATE PANEL ##
		_panel = createChoicePanel(_choice.getFill());

		Font	   font = getFont("LY_DEFAULT_FONT");
		JComponent component;

		n = ((_checkboxes == null) ? 0 : _checkboxes.size());
		for (i = 0; i < n; i++)
		{
			component = ((JComponent)_checkboxes.elementAt(i));

			// Set the background color same as its parent
			component.setBackground(null);
			component.setOpaque(false);
			// Set the default font.
			component.setFont(font);
			addChoiceToPanel(component);
		}
	} // end method setContent

	/**
	 * Sets the editable status of all the choices elements of this component.
	 *
	 * @param editable new status
	 * @see   LySetEditableInterface#setEditable(boolean)
	 */
	public void setEditable(boolean editable)
	{
		LyOption  option;
		Component box;
		int		  i, n = ((_checkboxes == null) ? 0 : _checkboxes.size());
		boolean   enabled;

		for (i = 0; i < n; i++)
		{
			option = _choice.getOption(i);
			box = _checkboxes.elementAt(i);
			enabled = editable && _last_enabled && option.isEnabled();
			box.setEnabled(enabled);
		}

		_last_editable = editable;
	}

	/**
	 * Sets the enabled status of all the choices elements of this component.
	 *
	 * @param status new status
	 * @see   LySetEnabledInterface#setEnabled(boolean)
	 */
	public void setEnabled(boolean status)
	{
		LyOption  option;
		Component box;
		int		  i, n = ((_checkboxes == null) ? 0 : _checkboxes.size());
		boolean   enabled;

		for (i = 0; i < n; i++)
		{
			option = _choice.getOption(i);
			box = _checkboxes.elementAt(i);
			enabled = status && _last_editable && option.isEnabled();
			box.setEnabled(enabled);
		}

		_last_enabled = status;
	}

	/**
	 * Sets the focus to this component.
	 *
	 * @see LySetFocusInterface#setFocus()
	 */
	public void setFocus()
	{
		if ((_checkboxes == null) || _checkboxes.isEmpty())
			return;

		Component box = _checkboxes.elementAt(0);

		box.requestFocus();
	}

	/**
	 * Represents a Swing panel containing the representation of the choice objects.
	 */
	public class _LyJChoicePanel extends JPanel
	{
		private static final long   serialVersionUID = 3236444197380349971L;

		/** The number of elements per line in this panel. */
		protected int			    _maxSpan;

		/** The fill policy. */
		protected LyChoiceBase.Fill _fill;

		/** The sub-panels of the current choice panel. */
		protected Vector<JPanel>    _subPanels;

		/** The number of choice in this panel. */
		protected int			    _choiceCount;

		/** The number of line displayed in this panel. */
		protected int			    _lineCount;

		/**
		 * Class constructor of a choice object for the Swing target. By default, this object can
		 * have only one element.
		 */
		public _LyJChoicePanel()
		{
			super();
			_lineCount = 1;
			_choiceCount = 0;
			_subPanels = new Vector<JPanel>();
			setMaxSpan(1);
			setFill(LyChoiceBase.Fill.HORIZONTAL);
			addSubPanel();
		}

		/**
		 * Sets the maximum span.
		 *
		 * @param maxSpan the new max span
		 */
		public void setMaxSpan(int maxSpan)
		{
			_maxSpan = maxSpan;
		}

		/**
		 * Sets the fill policy of the root panel and all its sub-panels.
		 *
		 * @param fill the fill policy
		 */
		public void setFill(LyChoiceBase.Fill fill)
		{
			_fill = fill;

			setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
			setSubPanelsFill(fill);
		}

		/**
		 * Sets the fill policy of the sub-panels.
		 *
		 * @param fill the fill policy. This parameter is not used and the sub-panels layout is
		 *             automaticaly set to a vertical FlowLayout.
		 * @see   VFlowLayout
		 */
		public void setSubPanelsFill(LyChoiceBase.Fill fill)
		{
			int i, n = _subPanels.size();

			for (i = 0; i < n; i++)
			{
				VFlowLayout layout = new VFlowLayout(VFlowLayout.TOP, 0, 0, false);

				layout.setAlignment(FlowLayout.LEFT);
				_subPanels.elementAt(i).setLayout(layout);
			}
		}

		/**
		 * Adds a new choice component to the current sub-panels. The new component is added
		 * verticaly or horizontaly depending on the fill policy used and on the number of objects
		 * already on this panel.
		 *
		 * @param choice the new choice component
		 */
		public void addChoice(Component choice)
		{
			_choiceCount++;

			if (_fill == LyChoiceBase.Fill.VERTICAL)
			{
				int newLineCount = (int)(Math.ceil(((double)_choiceCount) / ((double)_maxSpan)));

				if (newLineCount > _lineCount)
				{
					_lineCount = newLineCount;
					addSubPanel();
				}

				_subPanels.lastElement().add(choice);
			}
			else
			{
				if ((_choiceCount <= _maxSpan) || (_maxSpan <= 0))
				{
					if (_choiceCount != 1)
						addSubPanel();

					_subPanels.lastElement().add(choice);
				}
				else
				{
					int hPos = (_choiceCount % _maxSpan);

					if (hPos == 0)
						hPos = _maxSpan;
					if (hPos < 0)
						hPos = 1;

					_subPanels.elementAt(hPos - 1).add(choice);
				}
			}
		} // end method addChoice

		/**
		 * Adds a new empty sub-panel.
		 */
		public void addSubPanel()
		{
			JPanel p;

			if ((_lineCount != 1)
					|| ((_fill == LyChoiceBase.Fill.HORIZONTAL) && (_choiceCount > 1)))
				// to force p having the same height that the first panel
				p = new _SpecialPanel(_subPanels.elementAt(0));
			else
				p = new JPanel();

			VFlowLayout layout = new VFlowLayout(VFlowLayout.TOP, 0, 0, false);

			layout.setAlignment(FlowLayout.LEFT);
			p.setLayout(layout);

			p.setBackground(null);
			p.setOpaque(false);
			_subPanels.add(p);
			add(p);
		}

		/*
		 * (non-Javadoc)
		 * @see javax.swing.JComponent#requestFocus()
		 */
		public void requestFocus()
		{
			setFocus();
		}

	} // end class _LyJChoicePanel


	/**
	 * Special panel where its dimensions are based on those of another component.
	 */
	public class _SpecialPanel extends JPanel
	{
		private static final long serialVersionUID = 3624545191805197602L;

		/** Reference component which indicates the height of this panel. */
		protected Component		  _firstPanel;

		/**
		 * Creates and initializes a new SpecialPanel.
		 *
		 * @param c reference component
		 */
		public _SpecialPanel(Component c)
		{
			_firstPanel = c;
		}

		/**
		 * Computes the preferred size of this component. Its height is forced to be the same as the
		 * reference component.
		 */
		public Dimension getPreferredSize()
		{
			Dimension d = super.getPreferredSize();

			return new Dimension(d.width, _firstPanel.getPreferredSize().height);
		}
	} // end class SpecialPanel

	/**
	 * Gets the checkbox group of the current component.
	 *
	 * @return the checkbox group
	 */
	public Object getCheckboxGroup()
	{
		return _checkboxGroup;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _panel;
	}

	/**
	 * Adds a new checkbox with the given options to this current component. This method initializes
	 * the checkboxes list and their associations if they are empty.
	 *
	 * @param checkbox the new checkbox component to add
	 * @param option   the options of the new checkbox
	 */
	public void addCheckBox(Component checkbox, LyOption option)
	{
		if (_checkboxes == null)
		{
			_checkboxes = new Vector<Component>();
			_checkboxAssoc = new Hashtable<Component, LyOption>();
		}

		_checkboxes.addElement(checkbox);
		_checkboxAssoc.put(checkbox, option);
	}

	/**
	 * Gets the list of checkboxes that are selected.
	 *
	 * @return a String of the list of checkbox id separated with {@link LyString#SEP}
	 * @see    #setValue(String)
	 */
	public String getValue()
	{
		int				 i, n = _choice.getOptionCount();
		StringBuffer     sb = new StringBuffer();
		LyOption		 option;
		LySwingComponent internal;
		Component		 checkbox;
		char			 sep = getEnvironment().getString().SEP;

		for (i = 0; i < n; i++)
		{
			option = _choice.getOption(i);
			internal = (LySwingComponent)option.getInternalView();
			checkbox = internal.getComponent();

			if (isChecked(checkbox))
			{
				if (sb.length() > 0)
					sb.append(sep);
				sb.append(option.getId());
			}
		}
		if (sb.length() == 0)
			return null;
		return (sb.toString());
	}

	/**
	 * Sets or unsets the checkbox status of the specified elements.
	 *
	 * @param value the list of checkboxes to set their status. This list is a string separated with
	 *              {@link LyString#SEP} of the checkboxes id to set. If <code>null</code>, then
	 *              every checkbox is unselected.
	 * @see   #getValue()
	 */
	public void setValue(String value)
	{
		// XXX
		if (value != null)
		{
			char		   sep = getEnvironment().getString().SEP;
			String[]	   selected = LyString.stringToList(value, sep);
			Vector<String> selList = new Vector<String>(selected.length);
			int			   i;

			// create a vector of selected items
			for (i = 0; i < selected.length; i++)
				selList.addElement(selected[i]);

			// go through selection and (select) what is requested
			int		 n = _choice.getOptionCount();
			LyOption option;
			boolean  status;

			for (i = 0; i < n; i++)
			{
				option = _choice.getOption(i);
				status = selList.contains(option.getId());
				option.setChecked(status);
			}
		}
		else if (value == null)
		{
			// we must unselect everything
			LyOption option;
			int		 i, n = _choice.getOptionCount();

			for (i = 0; i < n; i++)
			{
				option = _choice.getOption(i);
				option.setChecked(false);
			}
		}
	} // end method setValue
} // end class LySwingChoice
