/*
 * LySwingCommand.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.graphic.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Command object (as described in the dtd) for swing viewer.
 *
 * <p>A command button allows to execute an action. There are predefined types of command giving
 * information about the nature of the action to execute (closing, cancellation, help, etc.).</p>
 *
 * @version $Id: LySwingCommand.java 34988 2012-03-23 08:33:20Z jlsaouli $
 * @author  Lyria
 * @see     LyCommand
 */
public class LySwingCommand extends LySwingObject
	implements LySwingComponent,
		LyGraphicalCommand,
		LyActionListener,
		LyEventTransmitter
{
	public static final String    __VERSION =
		"$Id: LySwingCommand.java 34988 2012-03-23 08:33:20Z jlsaouli $";

	/** The view independent command description. */
	protected LyCommand			  _command;

	/** The swing button. */
	protected Component			  _button;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter _actionTransmitter;

	/**
	 * Class constructor of a command object for the Swing target.
	 *
	 * @param command a COMMAND component defined for every type of view targets. This command
	 *                component is initialized specificaly for the Swing view.
	 */
	public LySwingCommand(LyCommand command)
	{
		super(command);
		_command = command;

		initialize();
	}

	/**
	 * Initializes this object. Builds the button and registers events.
	 */
	public void initialize()
	{
		String commandLabel = _command.getLabel();
		String label = translate(commandLabel);
		Image  image = getImage(_command.getImage());
		Font   font = getFont("LY_COMMAND_FONT");

		if (testResource("LY_USE_IMAGE_FOR_COMMAND") && (image == null))
		{
			image = (Image)getEnvironment().getImage(commandLabel, null, false);
			if (image != null)
				_button = createButton(label, image, font);
			else
				_button = createButton(label, null, font);
		}
		else
			_button = createButton(label, image, font);

		_button.setEnabled(_command.isEnabled());

		registerEvents();
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetEnabledInterface#setEnabled(boolean)
	 */
	public void setEnabled(boolean status)
	{
		_button.setEnabled(status);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (_actionTransmitter != null)
		{
			LyActionEvent evt = new LyActionEvent(this);

			evt.setSourceId(_command.getId());

			_actionTransmitter.processActionEvent(evt);
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return (_actionTransmitter);
	}

	/**
	 * This button doesn't handle selection events, so it has no selection transmitter.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		return null;
	}

	/**
	 * Creates and configures a button for the triggering of this object's command.
	 *
	 * @param  label the label to display on the button
	 * @param  image the image to display on the button
	 * @param  font  the font to used to draw the label
	 * @return the new button
	 */
	public Component createButton(String label, Image image, Font font)
	{
		LyJButton button;

		if ((label != null) && (label.length() != 0))
		{
			boolean useRollOver = true;

			button = new LyJButton(image,
					(((image != null)
								&& testResource("LY_USE_IMAGE_FOR_COMMAND")) ? null : label), label,
					_command.getId(), useRollOver);
		}
		else
			button = new LyJButton(image, null);

		button.setFont(font);
		button.setDefaultCapable(true);
		return button;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LySetLabelInterface#setLabel(java.lang.String)
	 */
	public void setLabel(String label)
	{
		label = translate(label);

		((LyJButton)_button).setText(label);
	}

	/**
	 * Register the events for this command.
	 */
	public void registerEvents()
	{
		LyActionTransmitter transmitter = ((LyJButton)_button).getActionTransmitter();

		if (transmitter != null)
			transmitter.addActionListener(this);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _button;
	}
} // end class LySwingCommand
