/*
 * LySwingCommands.java
 * Copyright (c) 2000-2009, Lyria-W4.
 * All rights reserved.
 */

package leon.view.win.swing;

import java.awt.*;
import javax.swing.*;
import leon.view.*;
import leon.view.event.*;
import leon.view.graphic.*;
import leon.view.win.swing.component.*;


/**
 * This class implements the Command bar object (as described in the dtd) for swing viewer.
 *
 * <p>A command bar allows to display a set of command buttons (Validate, Cancel, Close, etc.). This
 * bar is generally located at the bottom of a page or window.</p>
 *
 * @version $Id: LySwingCommands.java 28422 2010-11-03 12:03:38Z ehelmste $
 * @author  Lyria
 * @see     LyCommands
 * @see     LyCommand
 * @see     LySwingCommand
 */
public class LySwingCommands extends LySwingObject
	implements LySwingComponent,
		LyActionListener,
		LyGraphicalCommands,
		LyEventTransmitter
{
	public static final String    __VERSION =
		"$Id: LySwingCommands.java 28422 2010-11-03 12:03:38Z ehelmste $";

	/** The view independent command bar description. */
	protected LyCommands		  _commands;

	/** The panel that contains the command buttons. */
	protected Container			  _panel;

	/** The last command sub-panel added after a separator. */
	protected Container			  _lastBorderPanel;

	/** Provides notification events on component's actions. */
	protected LyActionTransmitter _actionTransmitter;

	/**
	 * Class constructor of a command bar object for the Swing target.
	 *
	 * @param commands a COMMANDS component defined for every type of view targets. This command bar
	 *                 component is initialized specificaly for the Swing view.
	 */
	public LySwingCommands(LyCommands commands)
	{
		super(commands);
		_commands = commands;

		initialize();
		setContent();
	}

	/**
	 * Initializes the current commands panel.
	 */
	public void initialize()
	{
		_panel = createPanel();
		addSep();
	}

	/**
	 * Creates and configures the panel that contains all the commands.
	 *
	 * @return a new commands panel
	 */
	public Container createPanel()
	{
		Image			   image = (Image)getEnvironment().getImage("LY_COMMANDS_BACKGROUND_IMAGE",
				null, false);

		// return new JPanel(new LySplitterLayout(LySplitterLayout.HORIZONTAL));
		LyJBackgroundPanel panel = new LyJBackgroundPanel(image);

		BoxLayout		   layout = new BoxLayout(panel, BoxLayout.X_AXIS);

		panel.setLayout(layout);
		return panel;
	}

	/**
	 * Creates and configures the last border panel.
	 *
	 * @param  align Specifies the alignment of the commands using the FlowLayout alignment values.
	 * @return the last border panel
	 */
	public Container createLastBorderPanel(int align)
	{
		Image		    image = (Image)getEnvironment().getImage("LY_COMMANDS_BACKGROUND_IMAGE",
				null, false);       
		LyJBorderPanel lastBorderPanel = new LyJBorderPanel(_commands.getBorder(), 1);

		lastBorderPanel.setLayout(new FlowLayout(align, 5, 3));
		lastBorderPanel.setGap(0);
		lastBorderPanel.setOpaque(image == null);

		Color col = getColor("LY_FORM_BACKGROUND");

		if (col == null)
			col = getColor("LY_DEFAULT_BACKGROUND");

		if (col != null)
			lastBorderPanel.setBackground(col);

		return lastBorderPanel;
	}

	/**
	 * Creates and configures all the command buttons for this command bar.
	 */
	public void setContent()
	{
		int i, n = _commands.getCommandCount();

		for (i = 0; i < n; i++)
		{
			LyCommand command = _commands.getCommand(i);

			if ((command == null)
			        || (command.getId() == LyComponent.SEP_ID))
				addSep(); // Sep
			else
			{
				LySwingCommand internal = (LySwingCommand)command.getInternalView();

				JButton		   button = (JButton)internal.getComponent();

				addCommand(button);

				// register events
				LyActionTransmitter transmitter = internal.getActionTransmitter();

				transmitter.addActionListener(this);
			}
		}
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyActionListener#actionPerformed(leon.view.event.LyActionEvent)
	 */
	public void actionPerformed(LyActionEvent actionEvent)
	{
		if (_actionTransmitter != null)
			_actionTransmitter.processActionEvent(actionEvent);
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.event.LyEventTransmitter#getActionTransmitter()
	 */
	public LyActionTransmitter getActionTransmitter()
	{
		if (_actionTransmitter == null)
			_actionTransmitter = new LySimpleActionTransmitter();

		return _actionTransmitter;
	}

	/**
	 * This command bar doesn't handle selection events, so it has no selection transmitter.
	 *
	 * @return <code>null</code>
	 * @see    leon.view.event.LyEventTransmitter#getSelectionTransmitter()
	 */
	public LySelectionTransmitter getSelectionTransmitter()
	{
		return null;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.setinterface.LyFreeInterface#free()
	 */
	public void free()
	{
		_panel = null;
		_lastBorderPanel = null;
		_commands = null;
		_actionTransmitter = null;
	}

	/**
	 * (non-Javadoc).
	 *
	 * @see leon.view.win.swing.LySwingComponent#getComponent()
	 */
	public Component getComponent()
	{
		return _panel;
	}

	/**
	 * Adds a button component to the panel.
	 *
	 * @param button the button component
	 */
	public void addCommand(Component button)
	{
		_lastBorderPanel.add(button);
	}

	/**
	 * Adds a separator to the panel.
	 */
	public void addSep()
	{
		int align = FlowLayout.CENTER;

		if (_commands.getAlign() == LyCommands.Align.LEFT)
			align = FlowLayout.LEFT;
		else if (_commands.getAlign() == LyCommands.Align.RIGHT)
			align = FlowLayout.RIGHT;
		else if (_commands.getAlign() == LyCommands.Align.CENTER)
			align = FlowLayout.CENTER;

		_lastBorderPanel = createLastBorderPanel(align);

		_panel.add(_lastBorderPanel);
	}

} // end class LySwingCommands
